(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[  17202320,     371222]*)
(*NotebookOutlinePosition[  17225955,     371820]*)
(*  CellTagsIndexPosition[  17225550,     371805]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[TextData[{
  "Short manual for the OpenGL Viewer for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"]
}], "Title"],

Cell["Jens-Peer Kuska", "Author"],

Cell["e-mail: kuska@informatik.uni-leipzig.de", "Address"],

Cell[TextData[{
  "The MathGL3d package offers the speed and quality of OpenGL together with \
",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " 3.0. Its core application mathview3d is a MathLink program for animation \
of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " 3D graphics. MathGL3d supports textures and colors per face. MathGL3d is \
also able to read 3Scripts and to generate POVRay 3.0 input files, VRML 2.0, \
QuickDraw3D and DXF files. The images can be saved as smooth shaded  \
PostScript and as PNG bitmaps.  "
}], "Abstract"],

Cell[TextData[{
  StyleBox["Important Notice: ",
    FontWeight->"Bold"],
  "The package is called MathGL3d ",
  StyleBox["not",
    FontWeight->"Bold"],
  " MathView3D. The reason is that the company Waterloo Maple Inc. has a \
program with a name similar to MathView3D that has nothing to do with the \
OpenGL viewer for ",
  StyleBox["Mathematica. ",
    FontSlant->"Italic"],
  "The correct name of the package is \"OpenGL Viewer for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "\" the sequence \"MathGL3d\" is a short cut for this long name. It is ",
  StyleBox["not",
    FontWeight->"Bold"],
  " a trade mark."
}], "Text"],

Cell["\<\
The name of the binary depends on the operating system and on the \
OpenGL implementation used. Usualy the binary name will reflect the OpenGL \
implementations of different vendors.  The documentation below refers to the \
binary  as mathview3d, but the name may be mv3dmesa.exe (Windows OS and Mesa \
library), mv3dsgi.exe (Windows OS and Silicon Graphics OpenGL),  mv3dmesa \
(Linux and Mesa) or mv3dxacc (Linux and XAccelerated OpenGL). The dependence \
on the particular name of the binary will be removed in future versions.\
\>",
   "Text"],

Cell[TextData[{
  StyleBox["The ",
    FontFamily->"Times New Roman"],
  StyleBox["MathLink",
    FontFamily->"Times New Roman",
    FontSlant->"Italic"],
  StyleBox[" program",
    FontFamily->"Times New Roman"],
  StyleBox[" mathview3d",
    FontFamily->"Arial"],
  " can run as stand alone program as well as a ",
  StyleBox["MathLink",
    FontSlant->"Italic"],
  " child of the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " kernel. Called from the command line it reads 3Scripts written by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " can display and animate the objects as well as write POVRay,VRML, \
QuickDraw3D, PostScript and DXF files with fixed names. Most information in \
stand alone mode is passed by command line arguments."
}], "Text"],

Cell[TextData[{
  "For MS Windows an OpenGL library is needed. Usually this comes with the \
drivers for the graphics card. If the graphic card has no such support the \
OpenGL implementation  from  ",
  ButtonBox["Microsoft for Windows 95",
    ButtonData:>{
      URL[ "ftp://ftp.microsoft.com/Softlib/MSLFILES/Opengl95.exe"], None},
    ButtonStyle->"Hyperlink"],
  " may work. The distribution includes the executable that is linked with \
the OpenGL 1.1 from Microsoft. Windows 98 and Windows NT come with OpenGL \
librarys.\n\nFor all other systems the program is compiled with Brian Paul's \
",
  ButtonBox["Mesa 3D",
    ButtonData:>"http://www.mesa3d.org/",
    ButtonStyle->"Hyperlink"],
  " graphics library. If hardware support for OpenGL on another platform is \
needed the binary must be recompiled with the system dependent OpenGL \
library. The source code is very portable and uses with the exception of ca. \
20 lines only OpenGL and GLUT functions. The Macintosh version of the Mesa 3D \
library can be found at ",
  ButtonBox["http://www.mesa3d.org/mac",
    ButtonData:>{
      URL[ "http://www.mesa3d.org/mac"], None},
    ButtonStyle->"Hyperlink"],
  ".\n\nAll intel-386 executables are compiled with Pentium optimization.\n\n\
The www-page for MathGL3d is ",
  ButtonBox["http://www.mpae.gwdg.de/~kuska/mview3d.html.",
    ButtonData:>{
      URL[ "http://www.mpae.gwdg.de/~kuska/mview3d.html"], None},
    ButtonStyle->"Hyperlink"],
  " There you can find the latest version of the program."
}], "Text"],

Cell[CellGroupData[{

Cell["Changes from version 1.0", "Subsection"],

Cell["\<\
In version 1.1 I support was added for the export of DXF files.\
\>", "Text"],

Cell[TextData[{
  "Version 1.2 was extensively changed, the source code increases by 3500 \
lines. The program uses now the GLUT 3.6 library. Some bugs in the lighting \
calculation were fixed. The reading of DXF and OFF files were added. Thanks \
to the ",
  StyleBox["rendereps",
    FontFamily->"Helvetica"],
  " example the export of smooth rendered PostScript were included. The GLUT \
3.6 library comes with the MUI interface. Using the MUI routines the \
interactive dialogs for the viewpoint and the light sources were created."
}], "Text"],

Cell[TextData[{
  "The menu got the free rotation item.  The export of VRML 2.0 files \
(Virtual Reality Modelling Language) and of QuickDraw3D metafiles was added. \
\nThe version 2.0 adds the ",
  StyleBox["MeshGraphics3D[]",
    FontFamily->"Courier"],
  " primitive and redefines ",
  StyleBox["ParametricPlot3D[]",
    FontFamily->"Courier"],
  ". Colors per face in ",
  StyleBox["SurfaceGraphics[]",
    FontFamily->"Courier"],
  " and ",
  StyleBox["MeshGraphics3D[]",
    FontFamily->"Courier"],
  " are now supported.  \nLast but not least, now the graphics can be covered \
with bitmap textures. Possible sources of textures are PNG files,  any ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " graphics as well as ",
  StyleBox["RasterArray[]",
    FontFamily->"Courier"],
  " primitives. "
}], "Text"],

Cell["\<\
In version 2.1 the speed for large polygon sets where increased, I \
add the polygon simplification and fixed the bug in the light source \
calculation.\
\>", "Text"],

Cell[TextData[{
  "The version 2.2 got a axis editor and the first Macintosh version where \
released. The Macintosh version where developed with the support of ",
  ButtonBox["Wolfram Research",
    ButtonData:>{
      URL[ "http://www.wolfram.com"], None},
    ButtonStyle->"Hyperlink"],
  "."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Interactivity in the OpenGL Viewer Window", "Section",
  CellTags->"WindowCommand"],

Cell[TextData[{
  StyleBox["The right mouse button inside the viewer window opens a menu",
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[1, 0, 0],
    FontVariations->{"Underline"->True}],
  StyleBox[".  ",
    FontColor->RGBColor[1, 0, 0],
    FontVariations->{"Underline"->True}]
}], "Text",
  FontSize->13,
  CellTags->"WindowCommand"],

Cell[TextData[{
  StyleBox[
  "Macintosh computers have only one mouse button and the mapping is\n\n",
    FontVariations->{"CompatibilityType"->0}],
  Cell[BoxData[
      StyleBox[
        FrameBox[GridBox[{
              {\(Left\ Button\), "\[LeftRightArrow]", 
                \(\[CloverLeaf]\  + \ MacMouse\)},
              {\(\ Middle\ Button\), "\[LeftRightArrow]", 
                \(\[OptionKey]\  + \ MacMouse\)},
              {\(Right\ Button\), "\[LeftRightArrow]", 
                \(\[ControlKey] + MacMouse\)}
              }]],
        FontFamily->"Helvetica"]]]
}], "Text",
  FontSize->13,
  FontColor->RGBColor[1, 0, 0],
  CellTags->"WindowCommand"],

Cell[BoxData[
    StyleBox[GridBox[{
          {
            \(\(Files ... 
                  \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \
\ \ \ \ \ \) \[RightTriangle] \)},
          {
            \(\(Spin ... 
                  \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \
\ \ \ \ \ \ \) \[RightTriangle] \)},
          {
            \(\(Rotate\  ... 
                  \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ 
                \) \[RightTriangle] \)},
          {
            \(\(Appearance ... \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \) 
              \[RightTriangle] \)},
          {\(Change\ view\ point\)},
          {
            StyleBox[\(Change\ lights\[IndentingNewLine]Axis\ Editor\),
              FontColor->GrayLevel[0]]},
          {\(Reset\ view\)},
          {\(Zoom\ in\)},
          {\(Zoom\ out\)},
          {\(Flip\ faces\)},
          {\(Clear\ window\)}
          },
        ColumnAlignments->{Left},
        GridFrame->True,
        RowLines->False,
        ColumnLines->True],
      FontFamily->"Arial"]], "Text",
  CellTags->"WindowCommand"],

Cell[TextData[{
  "The \"Files\" menu allows to write files with fixed names",
  StyleBox[".",
    FontFamily->"Arial"],
  StyleBox[
  " The file names are fixed because the GLUT library has no file name I/O \
dialog. You can still set other names with the command line options or with ",
    
    FontFamily->"Times"],
  StyleBox["Mathematica",
    FontFamily->"Times New Roman",
    FontSlant->"Italic"],
  StyleBox[
  " commands. With version 1.2 the ``Change view point\.b4\.b4 dialog has an \
entry ``Export Prefix\.b4\.b4. This string will be used to create the files \
with the corresponding extension. This means, if a PostScript file is saved \
the name is ",
    FontFamily->"Times"],
  StyleBox["\[LeftAngleBracket]",
    FontFamily->"Arial"],
  StyleBox["export prefix",
    FontFamily->"Arial",
    FontSlant->"Italic"],
  StyleBox["\[RightAngleBracket].eps ",
    FontFamily->"Arial"],
  StyleBox["and the Portable Network Graphics file is named ",
    FontFamily->"Times"],
  StyleBox["\[LeftAngleBracket]",
    FontFamily->"Arial"],
  StyleBox["export prefix",
    FontFamily->"Arial",
    FontSlant->"Italic"],
  StyleBox["\[RightAngleBracket].png",
    FontFamily->"Arial"],
  StyleBox[".",
    FontFamily->"Times New Roman"]
}], "Text",
  CellTags->"WindowCommand"],

Cell[BoxData[
    StyleBox[GridBox[{
          {\(Save\ \ POVRay\ file\)},
          {\(Save\ QuickDraw\ 3 D\)},
          {\(Save\ VRML\ 2.0\)},
          {\(Save\ \ PostScript\ file\)},
          {\(Save\ DXF\ file\)},
          {\(Save\ OFF\ file\)},
          {\(Save\ PNG\ bitmap\)},
          {\(Save\ EPS\ bitmap\)}
          },
        ColumnAlignments->{Left},
        GridFrame->True,
        RowLines->False,
        ColumnLines->True],
      FontFamily->"Arial"]], "Text",
  Background->None,
  CellTags->"WindowCommand"],

Cell[TextData[{
  "The \"Spin\" item allows simple animations of the object in the window. \
With the \"Track ball\" item you can choose an arbitrary rotation axis for \
the animation. The track ball works as follows. The window area of the OpenGL \
viewer is mapped onto a sphere ",
  Cell[BoxData[
      \(TraditionalForm
      \`\((x, y)\)\[LongRightArrow]\((\[Phi], \[Theta])\)\)]],
  ". When the left mouse button is pressed the vector from the origin to the \
point on the unit sphere is stored as ",
  Cell[BoxData[
      \(TraditionalForm\`v\&\[RightVector]\)]],
  ", the second vector, ",
  Cell[BoxData[
      \(TraditionalForm\`u\&\[RightVector]\)]],
  ", is obtained when the left mouse button is released. The cross product of \
the two vectors gives the rotation axes.  "
}], "Text",
  CellTags->"WindowCommand"],

Cell[BoxData[
    StyleBox[GridBox[{
          {"Trackball"},
          {\(X\ axes\ spin\)},
          {\(Y\ axes\ spin\)},
          {\(Z\ axes\ spin\)},
          {\(Spin\ slower\)},
          {\(Spin\ faster\)},
          {"Halt"}
          },
        ColumnAlignments->{Left},
        GridFrame->True,
        RowLines->False,
        ColumnLines->True],
      FontFamily->"Arial"]], "Text",
  CellTags->"WindowCommand"],

Cell["\<\
The \"Rotation\" item works similar to the \"Spin\" menu except \
that the object doesn't start spinning when the left mouse button is \
released.\
\>", "Text",
  CellTags->"WindowCommand"],

Cell[BoxData[
    StyleBox[GridBox[{
          {\(Free\ rotation\)},
          {\(X\ axes\ rotate\)},
          {\(Y\ axes\ rotate\ \)},
          {\(Z\ axes\ rotate\)}
          },
        GridFrame->True],
      FontFamily->"Arial",
      FontWeight->"Plain"]], "Text",
  CellTags->"WindowCommand"],

Cell["\<\
In the \"Appearence\" item the display can be modifyed between \
smooth, flat, and wire frame drawing of the 3d object in the window. For an \
export of a PostScript file the black background may be inappropriate and the \
``Toggle black/gray background\.b4\.b4 item may be used to switch to a light \
gray background. \
\>", "Text",
  CellTags->"WindowCommand"],

Cell[BoxData[
    StyleBox[GridBox[{
          {\(Smooth\ surface\)},
          {\(Flat\ surface\)},
          {\(Wire\ frame\)},
          {\(Toggle\ b/gray\ background\)},
          {\(Toggle\ polygon\ outlines\)}
          },
        ColumnAlignments->{Left},
        GridFrame->True,
        RowLines->False,
        ColumnLines->True],
      FontFamily->"Arial"]], "Text",
  CellTags->"WindowCommand"],

Cell[TextData[{
  "The main menu items",
  StyleBox[" Change viewpoint",
    FontFamily->"Helvetica"],
  ",  ",
  StyleBox["Change lights",
    FontFamily->"Helvetica"],
  " and ",
  StyleBox["Axis Editor",
    FontFamily->"Helvetica"],
  " open new windows. Use always the "
}], "Text",
  CellTags->"WindowCommand"],

Cell[BoxData[
    StyleBox[GridBox[{
          {"Hide"}
          },
        GridFrame->True],
      FontFamily->"Helvetica"]], "Text",
  CellTags->"WindowCommand"],

Cell[TextData[{
  "button to close the window. The ",
  StyleBox["Apply",
    FontFamily->"Helvetica"],
  " button may be used to  set the data for the main window. Only the",
  StyleBox[" MVClose[]",
    FontFamily->"Courier"],
  " function may be used to quit the OpenGL viewer."
}], "Text",
  CellTags->"WindowCommand"]
}, Closed]],

Cell[CellGroupData[{

Cell["The MathLink mode of the OpenGL Viewer", "Section",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  "When the OpenGL viewer is  called with the ",
  StyleBox["-mathlink",
    FontFamily->"Courier"],
  " option it communicates directly with the kernel. If the automatic texture \
file searching is used the ",
  StyleBox["$MVTexturePath",
    FontFamily->"Courier"],
  " variable must be set to a list of directories where the PNG files for \
texturing can be found. The help (",
  StyleBox["Help | Rebuild Help Index",
    FontFamily->"Helvetica"],
  ") index must be rebuild to use the ",
  ButtonBox["reference manual",
    ButtonData:>{"MVReference.nb", None},
    ButtonStyle->"Hyperlink"],
  " from the online help. If the ",
  StyleBox["OpenGLViewer.m",
    FontFamily->"Helvetica"],
  " file is placed correctly the command"
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Get["\<MathGL3d`OpenGLViewer`\>"]\)], "Input",
  CellLabel->"In[4]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    InterpretationBox[
      \("MathGL3d installed, Version 2.0\n"\[InvisibleSpace]"for "\
\[InvisibleSpace]"Windows95"\[InvisibleSpace]" system \
"\[InvisibleSpace]"Microsoft Windows"\[InvisibleSpace]"\n"\[InvisibleSpace]"\
\[Copyright] by Jens-Peer Kuska 1997,98\n"\[InvisibleSpace]"with MathLink \
form Wolfram Research."\),
      SequenceForm[ 
      "MathGL3d installed, Version 2.0\n", "for ", "Windows95", " system ", 
        "Microsoft Windows", "\n", 
        "\[Copyright] by Jens-Peer Kuska 1997,98\n", 
        "with MathLink from Wolfram Research."],
      Editable->False]], "Print",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  StyleBox["brings MathGL3d to live. Now the program ",
    FontFamily->"Times New Roman"],
  "is installed and appears on the screen. With the ",
  StyleBox["MVCreatePalette[]",
    FontFamily->"Courier"],
  " command a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " 3.0 palette is pasted in the current notebook."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(MVCreatePalette[]\)], "Input",
  CellLabel->"In[10]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[GridBox[{
        {" ", 
          StyleBox[\(OpenGL Viewer\),
            FontFamily->"Helvetica"], " ", " "},
        {
          FrameBox["Actions"], 
          FrameBox["Main"], 
          FrameBox["Spin"], 
          FrameBox["Window"]},
        {
          ButtonBox[\(MVShow3D[\[Placeholder]]\),
            ButtonData:>BoxData[ 
                
                RowBox[ {"MVShow3D", "[", "\[UnknownGlyph]", ",", 
                  "MVLineTubeSize", "->", "0.01", ",", "MVPolygonShading", 
                  "->", "MVSmooth", ",", "MVTubeSegments", "->", "12", ",", 
                  "MVTubeAngle", "->", "30", ",", "MVPointSphereSize", "->", 
                  "0.01", "]"}]]], 
          ButtonBox[\(Clear\ Window\),
            ButtonFunction:>MVClear[ ],
            ButtonEvaluator->Automatic], 
          ButtonBox[\(X - axis\),
            ButtonFunction:>CompoundExpression[ 
                MVPopWindow[ ], 
                MVSpin[ 
                  N[ 
                    Times[ Pi, 
                      Power[ 10, -1]]], {1.0, 0.0, 0.0}]],
            ButtonEvaluator->Automatic], 
          ButtonBox["Show",
            ButtonFunction:>MVShowWindow[ ],
            ButtonEvaluator->Automatic]},
        {
          ButtonBox[\(Smooth\ MVWritePOVRay[\[Placeholder]]\),
            ButtonData:>BoxData[ 
                
                RowBox[ {"MVWritePOVRay", "[", "\[Placeholder]", ",", 
                  "MVPolygonShading", "->", "MVSmooth", "]"}]]], 
          ButtonBox[\(Reset\ view\),
            ButtonFunction:>MVReset[ ],
            ButtonEvaluator->Automatic], 
          ButtonBox[\(Y - axis\),
            ButtonFunction:>CompoundExpression[ 
                MVPopWindow[ ], 
                MVSpin[ 
                  N[ 
                    Times[ Pi, 
                      Power[ 10, -1]]], {0.0, 1.0, 0.0}]],
            ButtonEvaluator->Automatic], 
          ButtonBox["Hide",
            ButtonFunction:>MVHideWindow[ ],
            ButtonEvaluator->Automatic]},
        {
          ButtonBox[\(Flat\ MVWritePOVray[\[Placeholder]]\),
            ButtonData:>BoxData[ 
                
                RowBox[ {"MVWritePOVRay", "[", "\[Placeholder]", ",", 
                  "MVPolygonShading", "->", "MVFlat", "]"}]]], 
          ButtonBox[\(Zoom\ in\),
            ButtonFunction:>MVZoom[ .5],
            ButtonEvaluator->Automatic], 
          ButtonBox[\(Z - axis\),
            ButtonFunction:>CompoundExpression[ 
                MVPopWindow[ ], 
                MVSpin[ 
                  N[ 
                    Times[ Pi, 
                      Power[ 10, -1]]], {0.0, 0.0, 1.0}]],
            ButtonEvaluator->Automatic], 
          ButtonBox["Iconify",
            ButtonFunction:>MVIconifyWindow[ ],
            ButtonEvaluator->Automatic]},
        {
          ButtonBox[\(MVWriteEPS[\[Placeholder]]\),
            ButtonData:>BoxData[ 
                
                RowBox[ {"MVWriteEPS", "[", "\[UnknownGlyph]", ",", 
                  "\[UnknownGlyph]", ",", "MVPolygonShading", "->", 
                  "MVSmooth", ",", "MVGrayBackground", "->", "True", 
                  "]"}]]], 
          ButtonBox[\(Zoom\ out\),
            ButtonFunction:>MVZoom[ 2.0],
            ButtonEvaluator->Automatic], 
          ButtonBox["Stop",
            ButtonFunction:>MVStopSpin[ ],
            ButtonEvaluator->Automatic], 
          ButtonBox["Quit",
            ButtonFunction:>MVClose[ ],
            ButtonEvaluator->Automatic]},
        {
          
          FrameBox[
            \(\[Copyright]\ \(1997, 98\)\ by\ \(Jens - Peer\)\ Kuska\)], 
          ButtonBox[\(MVWriteVRML[\[Placeholder]]\),
            ButtonData:>BoxData[ 
                
                RowBox[ {"MVWriteVRML", "[", "\[UnknownGlyph]", ",", 
                  "\[UnknownGlyph]", ",", "VRMLHeadLight", "->", "False", 
                  ",", "VRMLExtrusion", "->", "True", ",", 
                  "MVPolygonShading", "->", "MVSmooth", "]"}]]], 
          ButtonBox[\(MVWrite3DMF[\[Placeholder]]\),
            ButtonData:>BoxData[ 
                
                RowBox[ {"MVWrite3DMF", "[", "\[UnknownGlyph]", ",", 
                  "\[UnknownGlyph]", ",", "MVQuickDrawLights", "->", "True", 
                  ",", "MVQuickDrawViewPoint", "->", "True", ",", 
                  "MVPolygonShading", "->", "MVSmooth", "]"}]]], 
          ButtonBox[\(MVWritePNG[\[Placeholder]]\),
            ButtonData:>BoxData[ 
                
                RowBox[ {"MVWritePNG", "[", "\[UnknownGlyph]", ",", 
                  "\[UnknownGlyph]", ",", "MVDisplayGamma", "->", "True", 
                  ",", "MVInterlacePNG", "->", "True", ",", 
                  "MVPolygonShading", "->", "MVSmooth", "]"}]]]}
        },
      RowSpacings->0,
      ColumnSpacings->0,
      GridDefaultElement:>ButtonBox[ "\[UnknownGlyph]"]]], "Input",
  CellLabel->"In[56]:=",
  CellFrame->False,
  Active->True,
  GeneratedCell->True,
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LinkPatterns[$MVLink]\)], "Input",
  CellLabel->"In[57]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        TagBox[\(MViewPoolSize[]\),
          HoldForm], ",", 
        TagBox[\(MVPutGraphics3D[l_List]\),
          HoldForm], ",", 
        TagBox[\(MVClear[]\),
          HoldForm], ",", 
        TagBox[\(MVReset[]\),
          HoldForm], ",", 
        TagBox[\(MVShowWindow[]\),
          HoldForm], ",", 
        TagBox[\(MVHideWindow[]\),
          HoldForm], ",", 
        TagBox[\(MVIconifyWindow[]\),
          HoldForm], ",", 
        TagBox[\(MVPopWindow[]\),
          HoldForm], ",", 
        TagBox[\(MVPushWindow[]\),
          HoldForm], ",", 
        TagBox[\(MVStopSpin[]\),
          HoldForm], ",", 
        TagBox[\(MVZoom[zoom_Real]\),
          HoldForm], ",", 
        TagBox[
          \(MVSpin[angle_Real, 
            l_ /; VectorQ[l, Real == Head[#1]&] && 3 == Length[l]]\),
          HoldForm], ",", 
        TagBox[\(MVReshapeWindow[width_Integer, height_Integer]\),
          HoldForm], ",", 
        TagBox[\(MVGetPlotRange[]\),
          HoldForm], ",", 
        TagBox[\(MVGetViewPoint[]\),
          HoldForm], ",", 
        TagBox[\(mvRawGetLightSources[]\),
          HoldForm], ",", 
        TagBox[
          \(mvRawWritePOVRay[fname_String, ptexpre_String, mtexpre_String]\),
          HoldForm], ",", 
        TagBox[
          \(mvWrite3DMF[fname_String, savelight_Integer, saveview_Integer]\),
          HoldForm], ",", 
        TagBox[
          \(mvWriteDXF[fname_String, ts_Real, tseg_Integer, psize_Real, 
            tangle_Real]\),
          HoldForm], ",", 
        TagBox[\(mvWritePostScript[fname_String, wbitmap_Integer]\),
          HoldForm], ",", 
        TagBox[
          \(mvRawWriteVRML[fname_String, headl_Integer, useext_Integer, 
            ptpre_String, mtpre_String]\),
          HoldForm], ",", 
        TagBox[\(MVGetGraphics3D[]\),
          HoldForm], ",", 
        TagBox[\(MVGetRasterImage[]\),
          HoldForm], ",", 
        TagBox[\(mvPolygonReduce[tol_Real, iter_Integer]\),
          HoldForm], ",", 
        TagBox[\(MVLoadDXF[fname_String]\),
          HoldForm], ",", 
        TagBox[\(MVLoadOFF[fname_String]\),
          HoldForm], ",", 
        TagBox[\(mvWriteOFF[fname_String]\),
          HoldForm], ",", 
        TagBox[\(mvRawWritePNG[fname_String, gamma_Real, inter_Integer]\),
          HoldForm], ",", 
        TagBox[\(mvRawLoadPNGTexture[fname_String, gamma_Real]\),
          HoldForm], ",", 
        TagBox[\(mvRawPutMathTexture[clip_Integer, dispstr_String]\),
          HoldForm], ",", 
        TagBox[
          \(mvRawPutRasterTexture[bytec_Integer, w_Integer, h_Integer, data_]
            \),
          HoldForm], ",", 
        TagBox[\(MVGetTextureInformation[]\),
          HoldForm], ",", 
        TagBox[
          \(mvRawMapTexture[gid_Integer, texid_Integer, maptype_Integer, 
            move_List /; 3 == Length[move], axis_List /; 3 == Length[axis], 
            angle_Real, scale_List /; 3 == Length[scale], scaleU_Real, 
            scaleV_Real]\),
          HoldForm], ",", 
        TagBox[\(mvRotate3D[nx_Real, ny_Real, nz_Real, th_Real]\),
          HoldForm], ",", 
        TagBox[\(mvTranslate3D[nx_Real, ny_Real, nz_Real]\),
          HoldForm]}], "}"}]], "Output",
  CellLabel->"Out[57]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "The binary adds a new context path and additional ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " definitions to the current session. Allmost all functions and symbols \
added start with ",
  StyleBox["MV*",
    FontFamily->"Courier"],
  " or ",
  StyleBox["mv*",
    FontFamily->"Courier"],
  ". Only identifiers that begin with the capital letters MV are designed for \
direct use, the other identifiers are used for a more C-like communication \
between the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " kernel and the OpenGL viewer. The direct use of functions starting with \
mv* and mvRaw* is discouraged."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Names["\<MathView3D`*\>"]\)], "Input",
  CellLabel->"In[58]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({"EPSBitmap", "ExpandTextureFileName", "MeshGraphics3D", 
      "MVApplyTexture", "MVClipBorderColor", "MVClose", "MVCreatePalette", 
      "MVCylinderMapping", "MVDisplayGamma", "MVDisplayLightSources", 
      "MVFlat", "MVGetLightSources", "MVGrayBackground", "MVInterlacePNG", 
      "MVLineTubeSize", "MVMeshUVMapping", "MVMeshVUMapping", 
      "MVPasteGraphics", "MVPlaneMapping", "MVPointSphereSize", 
      "MVPolygonReduce", "MVPolygonShading", "MVPutTexture", 
      "MVQuickDrawLights", "MVQuickDrawViewPoint", "MVRotate3D", 
      "MVRotateTextureTarget", "MVScaleTexture", "MVScaleTextureTarget", 
      "MVShow3D", "MVSmooth", "MVSphereMapping", "MVTexture", 
      "MVTextureFilePrefix", "MVTextureMapType", "MVTextureSize", 
      "MVTorusMapping", "MVTranslate3D", "MVTranslateTextureTarget", 
      "MVTubeAngle", "MVTubeSegments", "MVWireFrame", "MVWorldLightSources", 
      "MVWrite3DMF", "MVWriteDXF", "MVWriteEPS", "MVWriteOFF", "MVWritePNG", 
      "MVWritePOVRay", "MVWriteVRML", "ReductionLoops", "VRMLExtrusion", 
      "VRMLHeadLight"}\)], "Output",
  CellLabel->"Out[58]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "If ",
  StyleBox["the OpenGL viewer",
    FontFamily->"Times New Roman"],
  " runs under Windows 95/NT the window is iconified on startup. The ",
  StyleBox["MVShowWindow[]",
    FontFamily->"Courier"],
  " command can be used to make it visible. The ",
  StyleBox["OpenGLViewer.m",
    FontFamily->"Helvetica"],
  " package calls ",
  StyleBox["MVShowWindow[]",
    FontFamily->"Courier"],
  " automatical."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  "The three dimensional graphics is passed to the OpenGL viewer with the ",
  StyleBox["MVShow3D[]",
    FontFamily->"Courier"],
  " command. ",
  StyleBox["MVShow3D[]",
    FontFamily->"Courier"],
  " has a couple of options to influence the appearance of the output."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[BoxData[{
    \(Off[CompiledFunction::cfse]\), 
    \(bx\  = \ 6\ \ Cos[u]\ \((1\  + \ Sin[u])\); \nby\  = \ 16\ \ Sin[u]; \n
    rad\  = \ 4\ \((1\  - \ Cos[u]\ /\ 2)\); \n
    X\  = \ If[Pi\  < \ u\  <= \ 2\ Pi, \ bx\  + \ rad\ Cos[v\  + \ Pi], \ \n
        \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ 
        bx\  + \ rad\ Cos[u]\ Cos[v]]; \n
    Y\  = \ If[Pi\  < \ u\  <= \ 2\ Pi, \ by, \ 
        by\  + \ rad\ Sin[u]\ Cos[v]]; \nZ\  = \ rad\ Sin[v]; \n\n
    kleinb = ParametricPlot3D[{X, \ Y, \ Z}, \ {u, \ 0, \ 2\ Pi}, 
        \ {v, \ 0, \ 2\ Pi}, \ \n\ \ \ \ \ \ \ \ PlotPoints\  -> \ {48, 12}, 
        \ Axes\  -> \ False, \ Boxed\  -> \ False, \ \n
        \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ \ 
        ViewPoint -> \ {0.893496, \(-1.73314\), \(-1.9499\)}, \n\t\t
        LightSources -> {{{0.335180, \(-1.1415\), \(-0.76458\)}, 
              RGBColor[1. , 0. , 0. ]}, {{1.5922, 0. , \(-0.681822\)}, 
              RGBColor[0. , 1. , 0. ]}, {{0.5276, \(-0.82099\), 
                \(-1.0235144\)}, RGBColor[0. , 0. , 1. ]}, {{0. , 0. , 1.5}, 
              RGBColor[0.25, 0.25, 0.25]}}, DisplayFunction -> Identity]; 
    \[IndentingNewLine]On[CompiledFunction::cfse]\)}], "Input",
  CellLabel->"In[128]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; MVShow3D[kleinb, MVPolygonShading -> MVSmooth]\)], "Input",
  CellLabel->"In[137]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVMesh3D[3], 
      PlotRange 
        \[Rule] {{\(-13.090739968238584`\), 9.754283822034724`}, {
            \(-15.991065019684074`\), 20.055623908350213`}, {
            \(-5.934507868443346`\), 5.934507868443346`}}, 
      ViewPoint \[Rule] {0.893496`, \(-1.73314`\), \(-1.9499`\)}, 
      AmbientLight \[Rule] {0.`, 0.`, 0.`}}\)], "Output",
  CellLabel->"Out[137]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[6]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000A04F6FAF4F70AF4F70B0507FAF4F70B0507FB04F7F\
BF5080B05080BF5080BF5080BF508FBF5080BF508FBF508FC0508FBF508FC0508F
C0508FC05090BF4F8FC0508FC04F8FC04F8FBF4F8F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000A04F6FAF5070AF5070B0507FAF5070B05F7FB0507FBF5F80B05080BF5F8FBF5F8FC05F8F\
BF5F8FC05F90C05F90C05F90C05F90CF609FC05F9FCF5F9FC05F9FCF609FCF5F9F
CF5FA0CF5F9FCF5FA0CF5F9FCF5F9FCF509FCF5F9FCF509FCF509FCF509FCF509FCF4F9FCF509F\
CF4F90000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000A04F60A04F60AF506F
AF4F6FAF5070AF507FB05F80B05080BF5F80BF5F80BF5F8FBF5F8FC06090C05F90C0609FC05F9F\
CF609FCF609FCF60A0CF60A0CF60A0CF60A0D060AFCF60A0D060AFCF60AFD060AF
CF60AFD060AFD060AFD060AFCF5FA0D060AFCF5FA0D060AFCF5FA0D05FAFCF5FA0D05FAFCF50A0\
D050A0CF50A0D050A0CF4F9FCF4F9F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000A05060A04F60AF5070AF5070B05F7FB0507F
BF5F80B05F80BF5F8FBF5F8FC06090C06090CF609FC0609FCF6FA0CF6FA0D06FA0CF6FA0D06FAF\
D06FAFDF6FB0D06FB0DF70B0D06FB0DF70B0D06FB0DF70BFDF6FBFDF70BFDF6FBF
E070BFDF6FBFDF70BFDF6FBFDF70BFDF6FBFDF6FBFDF6FB0DF6FBFDF60B0DF6FBFDF60B0DF60BF\
D060B0DF60B0D05FAFDF5FB0D050AFD05FAFD050A0D050AFD04FA0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000009F4F609F4F60A0506FA0506FAF5F70AF507FB05F80B05F80BF5F8F
BF5F8FC06090C06090CF609FCF609FCF6FA0CF60A0D06FAFD06FAFD070B0D06FAFDF70B0DF70B0\
DF70BFDF70BFE070C0DF70BFE07FC0DF70C0E07FC0DF70C0E07FCFDF70C0E07FCF
E070CFE07FCFE070C0E070CFE070C0E070CFDF70C0E070C0DF70C0E070C0DF6FC0E06FC0DF6FC0\
E06FC0DF60BFDF60BFDF60B0DF60BFD05FB0DF5FB0D05FB0DF5FB0D050AFDF50AF
D04FAFDF50AF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
9F4F5FA0506FA0506FAF5F70AF5070B05F7FB05F80BF608FBF5F8FC06090C06090
CF6F9FCF609FD06FA0CF6FA0D070AFD06FAFDF70B0DF70B0DF70BFDF70BFE07FC0E07FC0E07FCF\
E07FC0EF80CFEF7FCFEF80D0EF7FCFEF80D0EF7FD0EF80D0EF7FD0EF80DFEF80D0
EF80DFEF80DFF080DFEF7FD0EF80DFEF7FD0EF80DFE07FD0EF7FD0E07FD0EF7FD0E070D0EF7FD0\
E070CFE070CFE070C0E070CFDF6FC0E06FC0DF60C0E060C0DF60BFE060C0DF5FBF
DF5FBFDF50B0DF5FB0DF50AF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000009F50609F4F60\
A0506FA0506FAF5F70AF507FB05F80B05F80BF608FBF608FC06090C0609FCF6FA0
CF6FA0D070AFD06FAFDF70B0DF70B0E070BFE070BFE07FC0E07FC0EF7FCFE07FCFEF80D0EF7FD0\
EF80D0EF80DFF08FDFF080DFF08FE0F080DFF08FE0F080E0F08FE0F080E0F08FEF
F080E0F08FEFF080E0F08FEFF080E0F08FE0EF80E0F08FE0EF80DFF080E0EF80DFF080E0EF7FDF\
EF7FDFEF7FD0EF7FD0E070CFE070D0E070CFE070CFE06FCFE06FCFE06FC0E06FC0
DF60BFDF60BFDF5FBFDF5FBFDF50B0DF50B0DF50B0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000009F505FA0506FA0506FAF5F70\
AF5F70B05F7FB05F80BF608FBF608FC06090C06090CF6FA0CF6FA0D070AFD06FAF
DF70B0DF70B0E07FBFE07FBFEF80CFEF7FCFEF80D0EF80D0F080DFF080DFF08FDFF080DFFF8FE0\
FF8FE0FF8FEFFF8FEFFF90EFFF8FEFFF90EFFF8FEFFF90F0FF90EFFF90F0FF90EF
FF90F0FF90F0FF90F0FF90F0FF90F0FF90F0FF90F0FF8FF0FF90F0F08FF0FF90F0F08FEFFF8FEF\
F08FEFF08FEFEF80E0F080E0EF80DFEF80DFEF7FDFEF7FDFEF70D0EF70D0E070CF
EF70D0E06FCFE06FCFE060C0E060C0E05FC0E05FC0E05FBFE05FBFE050BF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000009F505F9F5060A0506FA0506FAF5F70AF5F7F\
B06080B05F80BF608FBF608FC06F9FC0609FCF6FA0D06FA0DF70AFD070AFDF70BF
DF70BFE07FC0E07FC0EF80CFEF7FCFF080DFF080DFFF8FE0F080E0FF8FEFFF8FE0FF8FEFFF8FE0\
FF90EFFF8FE0FF90EFFF90EFFF90EFFF90EFFF90EFFF90EFFF9FF0FF90EFFF9FF0
FF90EFFF9FF0FF90EFFF9FF0FF90F0FF9FF0FF90F0FF90F0FF90F0FF90FFFF90F0FF90FFFF90F0\
FF90FFF08FF0FF8FF0F08FEFF08FEFEF80E0F080EFEF80E0F080E0EF7FDFEF7FDF
EF70D0EF70D0E06FCFEF6FCFE06FCFEF6FCFE060C0E060CFE05FC0E060C0E050BFE05FBF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000904F5F9F50609F5060A05F6FAF5070B05F7FB05F7FBF608F\
BF608FC06F90C06F90CF6F9FCF6FA0D070AFD070AFDF70B0DF70B0E07FC0E07FC0
EF7FCFEF7FCFF080D0F080D0FF8FDFFF80E0FF8FEFFF8FE0FF90EFFF8FE0FF90EFFF90EFFF90EF\
FF90EFFF9FEFFF90EFFF9FEFFF9FEFFF9FF0FF9FEFFF9FF0FF9FEFFF9FF0FF9FF0
FF9FF0FF9FF0FF9FF0FF9FF0FF9FF0FF9FF0FF9FFFFF9FF0FF9FFFFF9FF0FF9FFFFF9FFFFF9FFF\
FF90FFFF9FFFFF90FFFF90FFFF90FFFF90FFF08FF0FF90F0F08FEFF08FEFF080E0
F080E0EF7FDFF07FDFEF7FDFEF7FDFEF70D0EF70D0EF6FD0EF6FD0E060CFEF60CFE05FC0EF5FC0\
E05FC0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000904F5F9F50609F5060A05F70A05070B05F7FB05F7FBF608FBF608F\
C06090C0609FCF6FA0CF6FA0D070AFD070B0DF7FBFDF70BFE07FC0EF7FC0F080D0
EF7FD0F080DFF080DFFF8FE0FF8FE0FF8FE0FF8FE0FF90EFFF90E0FF90EFFF90E0FF9FEFFF90E0\
FF9FEFFF9FEFFF9FEFFF9FEFFF9FEFFF9FEFFFA0F0FF9FEFFFA0F0FF9FEFFFA0F0
FF9FEFFFA0F0FF9FF0FFA0F0FF9FF0FFA0F0FF9FF0FFA0FFFF9FF0FFA0FFFF9FF0FF9FFFFF9FF0\
FF9FFFFF9FF0FF9FFFFF90F0FF9FFFFF90FFFF9FFFFF90FFFF90FFF08FF0FF90F0
F08FEFF08FEFF080E0F080E0EF7FDFEF7FDFEF70DFEF7FDFEF70D0EF70D0EF6FCFEF6FCFE060CF\
EF60CFE05FC0EF5FCFE050C0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000008F505F905060A05F6FA0506FAF5F70AF5F7FB06080B06080BF608FBF608FCF6F9F\
CF6F9FD070A0D06FAFDF70B0DF70B0E07FBFE07FC0EF80CFEF80CFF080D0F080DF
FF8FE0FF8FE0FF8FE0FF8FE0FF90EFFF90E0FF90EFFF90E0FF9FEFFF9FEFFF9FEFFF9FEFFFA0EF\
FF9FEFFFA0F0FFA0EFFFA0F0FFA0EFFFAFF0FFA0EFFFAFF0FFA0F0FFAFF0FFA0F0
FFAFF0FFA0F0FFAFF0FFA0F0FFAFFFFFA0F0FFAFFFFFA0F0FFAFFFFFA0FFFFA0FFFFA0FFFFA0FF\
FFA0FFFFA0FFFF9FFFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF90FF
FF9FFFFF90F0FF90F0F08FEFFF8FEFF080EFF080EFF07FE0F080E0EF7FDFF07FDFEF70D0EF70D0\
EF6FD0EF6FD0EF60CFEF60D0EF60CFEF60CF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000905060905060A0506FA0506FAF5F7FAF5F7FB06080B06080BF608FBF6090CF6F9FCF6F9F\
D070A0D06FAFDF70B0DF70B0E07FC0E07FC0EF7FCFEF7FCFF080DFF080DFFF8FE0
FF8FDFFF90E0FF8FE0FF90E0FF90E0FF9FEFFF90E0FF9FEFFF9FE0FFA0EFFF9FEFFFA0EFFFA0EF\
FFA0EFFFA0EFFFAFF0FFA0EFFFAFF0FFA0EFFFAFF0FFAFEFFFAFF0FFAFF0FFAFF0
FFAFF0FFAFF0FFAFF0FFAFF0FFAFF0FFAFFFFFA0F0FFAFFFFFA0F0FFAFFFFFA0F0FFAFFFFFA0F0\
FFA0FFFFA0F0FFA0FFFFA0FFFFA0FFFF9FFFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFF
FF90FFFF9FFFFF90FFFF90FFF090F0FF90F0F08FEFF08FEFF080EFF080EFEF7FDFF07FE0EF70DF\
EF70DFEF6FD0EF6FD0EF6FD0EF6FD0EF60CFEF60D0EF5FCF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000804F5F
9F50609F5060A05F70A05F70AF5F7FAF5F7FBF608FBF608FC06F90C06090CF6FA0CF6FA0D070AF\
D070AFE07FB0E070BFEF7FC0EF7FC0F080D0F080D0FF80DFFF80DFFF8FE0FF8FE0
FF90E0FF90E0FF9FEFFF90E0FF9FEFFF9FE0FFA0EFFF9FEFFFA0EFFFA0EFFFAFEFFFA0EFFFAFF0\
FFAFEFFFAFF0FFAFEFFFB0F0FFAFEFFFB0F0FFAFF0FFB0F0FFAFF0FFB0F0FFB0F0
FFB0F0FFB0F0FFB0FFFFAFF0FFB0FFFFAFF0FFB0FFFFAFF0FFB0FFFFAFF0FFB0FFFFAFFFFFAFFF\
FFAFFFFFAFFFFFA0FFFFAFFFFFA0FFFFAFFFFFA0FFFFA0FFFF9FFFFFA0FFFF9FFF
FFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF90FFFF9FFFFF90F0FF90F0F08FEFFF8FEFF080E0F080E0\
EF7FDFF07FE0EF70DFF070DFEF6FDFF070DFEF6FD0F06FDFEF60D0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006F4F508F4F5F9F5060
9F5060A05F6FA05F70AF5F7FAF5F7FB0608FBF608FC06F90C06090CF6FA0CF6FA0D070AFD070AF\
DF70B0E070BFEF7FC0EF7FC0F080D0F080D0FF80DFFF80DFFF8FE0FF8FDFFF90E0
FF8FE0FF90E0FF90E0FF9FEFFF9FE0FFA0EFFF9FE0FFA0EFFFA0EFFFAFEFFFA0EFFFAFEFFFAFEF\
FFAFEFFFAFEFFFB0F0FFAFEFFFB0F0FFB0EFFFB0F0FFB0EFFFB0F0FFB0F0FFBFF0
FFB0F0FFB0F0FFB0F0FFB0F0FFB0F0FFB0F0FFB0F0FFB0FFFFB0F0FFB0FFFFAFF0FFB0FFFFAFF0\
FFAFFFFFAFF0FFAFFFFFAFF0FFAFFFFFA0FFFFAFFFFFA0FFFFA0FFFFA0FFFFA0FF
FF9FFFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF90FFFF9FFFFF90F0FF90F0F08FF0FF8FF0F080EF\
F080EFF07FE0F080E0EF7FDFF07FE0EF70DFF070DFEF6FD0F06FDFEF60D0F06FDF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000080505F90505F9F50609F5060
A05F70A05F70AF607FB05F80BF608FBF608FC06F90C06F9FCF70A0D06FA0DF70B0DF70B0E07FBF\
E07FC0EF7FCFEF7FCFFF80D0FF80DFFF8FDFFF80DFFF8FE0FF8FE0FF90E0FF90E0
FF9FEFFF9FE0FFA0EFFF9FE0FFA0EFFFA0EFFFAFEFFFA0EFFFAFF0FFAFEFFFB0F0FFB0EFFFB0F0\
FFB0F0FFBFF0FFB0F0FFBFF0FFBFF0FFBFF0FFBFF0FFBFF0FFBFF0FFBFFFFFBFF0
FFC0FFFFBFF0FFBFFFFFBFF0FFBFFFFFBFF0FFBFFFFFB0F0FFBFFFFFBFFFFFBFFFFFB0FFFFB0FF\
FFB0FFFFB0FFFFB0FFFFB0FFFFAFFFFFB0FFFFAFFFFFAFFFFFAFFFFFAFFFFFA0FF
FFAFFFFFA0FFFFA0FFFFA0FFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF90F0FF90FF\
F08FF0FF8FF0F080EFFF8FEFF080EFF080EFF07FE0F07FE0F070DFF070DFF06FDF
F06FDFF06FDF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000604050804F5F8F505F9F50609F5060A05F70
A05F70AF5F7FAF5F7FB06080BF608FC06F90C06090CF6FA0CF6FA0D070AFD070B0E070BFE070BF\
EF7FCFEF7FCFF080D0F080D0FF8FDFFF80DFFF8FDFFF8FDFFF90E0FF90DFFF9FE0
FF9FE0FF9FEFFF9FE0FFA0EFFFA0E0FFAFEFFFA0EFFFAFEFFFAFEFFFB0F0FFB0EFFFB0F0FFB0EF\
FFBFF0FFB0EFFFBFF0FFBFF0FFC0F0FFBFF0FFC0F0FFBFF0FFC0F0FFBFF0FFC0F0
FFC0F0FFC0FFFFBFF0FFC0FFFFBFF0FFC0FFFFBFF0FFBFFFFFBFF0FFBFFFFFBFF0FFBFFFFFB0FF\
FFBFFFFFB0F0FFB0FFFFB0FFFFB0FFFFB0FFFFB0FFFFAFFFFFAFFFFFAFFFFFAFFF
FFA0FFFFAFFFFFA0FFFFA0FFFFA0FFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF90FF\
FF90FFF08FF0FF8FF0F08FEFFF8FF0F080EFF080EFF07FE0F07FE0EF70DFF070E0
F06FDFF070E0F06FDF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F4F5080505F8F505F9F50609F506FA05F70A05F70
AF607FAF5F7FBF608FBF608FC06F90C06F90CF6FA0D06FA0DF70AFDF70B0E07FBFE07FBFEF7FCF\
EF7FCFFF80D0FF80D0FF8FDFFF8FDFFF90E0FF8FDFFF90E0FF90E0FF9FE0FF9FE0
FFA0EFFF9FE0FFA0EFFFA0EFFFAFEFFFAFEFFFB0F0FFB0EFFFB0F0FFB0EFFFBFF0FFBFF0FFBFF0\
FFBFF0FFC0F0FFC0F0FFC0F0FFC0F0FFC0FFFFC0F0FFCFFFFFC0F0FFCFFFFFC0F0
FFCFFFFFC0F0FFCFFFFFC0F0FFC0FFFFC0F0FFC0FFFFC0FFFFC0FFFFC0FFFFC0FFFFBFFFFFC0FF\
FFBFFFFFBFFFFFBFFFFFBFFFFFB0FFFFBFFFFFB0FFFFB0FFFFAFFFFFB0FFFFAFFF
FFAFFFFFAFFFFFAFFFFFAFFFFFAFFFFFA0FFFFA0FFFFA0FFFFA0FFFF9FFFFFA0FFFF9FFFFF9FFF\
FF9FFFFF9FFFFF90FFFF90FFF08FF0FF8FF0F080EFFF80EFF07FE0F07FEFF07FE0
F07FE0F070E0FF70E0F070E0FF70E0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005040506F4050804F5F8F505F9050609F5060A05F70A05F70AF607F
AF5F7FB0608FBF608FC06F90C06090CF6F9FCF6FA0D070AFD070AFE070BFE070BFEF7FC0EF7FC0\
F080D0F080D0FF80DFFF80DFFF8FDFFF8FDFFF90E0FF90DFFF9FE0FF9FE0FFA0E0
FF9FE0FFA0EFFFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFB0F0FFB0EFFFBFF0FFBFEFFFC0F0FFBFEF\
FFC0F0FFC0F0FFC0F0FFC0F0FFCFF0FFC0F0FFCFFFFFC0F0FFCFF0FFCFF0FFCFFF
FFCFF0FFCFFFFFCFF0FFCFFFFFC0F0FFCFFFFFC0F0FFCFFFFFC0F0FFC0FFFFC0FFFFC0FFFFBFF0\
FFC0FFFFBFFFFFC0FFFFBFF0FFBFFFFFBFFFFFBFFFFFB0FFFFB0FFFFB0FFFFB0FF
FFAFFFFFB0FFFFAFFFFFAFFFFFAFFFFFAFFFFFA0FFFFA0FFFF9FFFFFA0FFFF9FFFFFA0FFFF9FFF\
FF9FFFFF9FFFFF9FFFFF90FFFF90FFF08FF0FF8FF0F080EFF080EFF07FE0F080EF
F07FE0F07FEFF070E0FF7FEFF070E0FF70E0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F4F50704F5080505F90505F9F5F609F506FA05F70A05F70AF607FAF6080
BF608FBF608FC06F90C06F9FCF70A0D06FA0DF70AFDF70B0E07FBFE07FBFEF7FC0EF7FCFFF80D0\
FF80D0FF8FDFFF80DFFF90E0FF8FDFFF90E0FF90E0FF9FE0FF9FE0FFA0EFFFA0E0
FFAFEFFFAFEFFFAFEFFFAFEFFFB0F0FFB0EFFFBFF0FFBFEFFFC0F0FFBFF0FFC0F0FFC0F0FFCFF0\
FFC0F0FFCFF0FFCFF0FFD0FFFFCFF0FFD0FFFFCFF0FFD0FFFFD0F0FFD0FFFFD0F0
FFD0FFFFD0FFFFD0FFFFCFFFFFD0FFFFCFFFFFD0FFFFCFFFFFD0FFFFCFFFFFCFFFFFC0FFFFCFFF\
FFC0FFFFC0FFFFC0FFFFC0FFFFC0FFFFC0FFFFBFFFFFBFFFFFBFFFFFBFFFFFB0FF
FFBFFFFFB0FFFFB0FFFFB0FFFFB0FFFFAFFFFFAFFFFFA0FFFFAFFFFFA0FFFFA0FFFFA0FFFFA0FF\
FFA0FFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF90F0FF90FFFF8FF0FF8FF0F080EF
FF8FF0F080EFFF80EFF07FEFFF7FEFFF7FEFFF7FEFF070EF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F4F50704F5080505F8F505F9F5F609F5060A05F70A05F70AF607FAF5F7FB0608F
BF608FC06F90C06090CF6FA0CF6FA0D070AFD070AFE070BFE070BFEF7FC0EF7FC0F080D0FF80D0\
FF80DFFF80DFFF8FDFFF8FDFFF90E0FF90DFFF9FE0FF9FE0FFA0E0FFA0E0FFA0EF
FFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFEFFFC0F0FFBFF0FFC0F0FFC0F0FFCFF0FFC0F0\
FFCFF0FFCFF0FFD0F0FFCFF0FFD0FFFFD0F0FFD0FFFFD0F0FFD0FFFFD0F0FFDFFF
FFD0F0FFDFFFFFD0F0FFD0FFFFD0FFFFD0FFFFCFF0FFD0FFFFCFFFFFD0FFFFCFFFFFCFFFFFCFFF\
FFCFFFFFC0FFFFCFFFFFC0FFFFC0FFFFC0FFFFC0FFFFBFFFFFBFFFFFBFFFFFBFFF
FFB0FFFFBFFFFFB0FFFFB0FFFFAFFFFFB0FFFFAFFFFFAFFFFFA0FFFFAFFFFFA0FFFFA0FFFFA0FF\
FFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF90FFFF9FFFFF90F0FF90FFF08FF0FF8FF0
F08FEFFF8FF0F080EFFF80F0F07FEFFF7FEFF070EFFF7FEFF070EF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000604F50\
704F5080505F8F505F9F5F609F506FA05F70A05F70AF607FAF6080BF608FBF608F
C06F90C06F9FCF70A0CF6FA0D070AFDF70B0E07FBFE07FBFEF7FC0EF7FCFFF80D0FF80D0FF80DF\
FF80DFFF8FDFFF8FDFFF90E0FF90DFFF9FE0FF9FE0FFA0EFFFA0E0FFAFEFFFAFEF
FFAFEFFFAFEFFFB0F0FFB0EFFFBFF0FFBFEFFFC0F0FFC0F0FFCFF0FFC0F0FFCFFFFFCFF0FFD0FF\
FFCFF0FFD0FFFFD0F0FFDFFFFFD0F0FFDFFFFFDFFFFFDFFFFFDFFFFFDFFFFFDFFF
FFDFFFFFDFFFFFDFFFFFDFFFFFDFFFFFD0FFFFDFFFFFD0FFFFDFFFFFD0FFFFD0FFFFD0FFFFD0FF\
FFCFFFFFD0FFFFCFFFFFCFFFFFCFFFFFCFFFFFC0FFFFC0FFFFC0FFFFC0FFFFBFFF
FFBFFFFFBFFFFFBFFFFFB0FFFFB0FFFFB0FFFFB0FFFFAFFFFFAFFFFFAFFFFFAFFFFFA0FFFFAFFF\
FFA0FFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF90F0
FF90FFFF8FF0FF8FF0FF80F0FF80F0FF7FEFFF80F0FF7FEFFF7FF0FF7FEFBF5FAF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000604F506F4F50\
80505F8F505F9050609050609F5F6FA05F70AF607FAF5F7FB06080B0608FC06F90
C06090CF6F9FCF6FA0D070AFD070AFDF70B0E070BFEF7FC0EF7FC0F07FCFF07FD0FF80DFFF80D0\
FF8FDFFF8FDFFF90E0FF90DFFF9FE0FF9FE0FF9FE0FF9FE0FFA0EFFFA0E0FFAFEF
FFAFEFFFB0EFFFB0EFFFBFF0FFBFEFFFC0F0FFC0F0FFC0F0FFC0F0FFCFF0FFCFF0FFD0FFFFCFF0\
FFD0FFFFD0F0FFDFFFFFD0F0FFDFFFFFDFF0FFE0FFFFDFF0FFE0FFFFDFFFFFE0FF
FFDFF0FFE0FFFFDFFFFFE0FFFFDFF0FFDFFFFFDFFFFFDFFFFFD0FFFFDFFFFFD0FFFFDFFFFFD0FF\
FFD0FFFFCFFFFFD0FFFFCFFFFFCFFFFFCFFFFFCFFFFFC0FFFFC0FFFFC0FFFFC0FF
FFBFFFFFC0FFFFBFFFFFBFFFFFB0FFFFB0FFFFAFFFFFB0FFFFAFFFFFAFFFFFAFFFFFAFFFFFA0FF\
FFA0FFFF9FFFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF90FFFF9FFF
F090F0FF90FFFF8FF0FF8FF0F080EFFF80F0F07FEFFF80F0F07FEFFF7FF0F07FEFD06FCF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000050404F604F50704F5080505F\
8F505F905F60905060A05F70A05F70AF607FAF5F7FB0608FB0608FBF6F90C06F90
CF6FA0CF6FA0D070AFD070AFE07FB0E070BFEF7FC0EF7FC0F080CFF080D0FF80DFFF80D0FF8FDF\
FF8FDFFF90E0FF90DFFF9FE0FF9FE0FFA0E0FF9FE0FFAFEFFFA0EFFFAFEFFFAFEF
FFBFF0FFB0EFFFBFF0FFBFEFFFC0F0FFC0F0FFCFF0FFCFF0FFD0FFFFD0F0FFD0FFFFD0F0FFDFFF\
FFDFFFFFDFFFFFDFFFFFE0FFFFE0FFFFE0FFFFE0FFFFEFFFFFE0FFFFEFFFFFE0FF
FFEFFFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFDFFFFFE0FFFFDFFFFFDFFFFFDFFFFFDFFF\
FFDFFFFFDFFFFFD0FFFFDFFFFFD0FFFFD0FFFFCFFFFFD0FFFFCFFFFFCFFFFFC0FF
FFCFFFFFC0FFFFC0FFFFBFFFFFBFFFFFB0FFFFB0FFFFB0FFFFB0FFFFAFFFFFB0FFFFAFFFFFAFFF\
FFA0FFFFA0FFFF9FFFFFA0FFFF9FFFFFA0FFFF9FFFFFA0FFFF9FFFFF9FFFFF9FFF
FF9FFFFF90FFFF90FFFF8FF0FF90FFFF8FF0FF8FF0FF80F0FF80F0FF80F0FF80F0FF7FF0EF7FEF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000050404F604F506F4F507F505F80505F\
9050609050609F5F6F9F5F6FAF5F7FAF5F7FB06080B06080BF6F8FBF6090C06F9F
CF6F9FD070A0D06FAFDF70B0DF70B0E07FBFE07FC0F07FCFF07FCFFF80D0FF80D0FF8FDFFF8FDF\
FF90DFFF90DFFF9FE0FF9FE0FF9FE0FF9FE0FFA0E0FFA0E0FFAFEFFFAFEFFFB0EF
FFB0EFFFBFF0FFBFEFFFC0F0FFC0F0FFCFF0FFC0F0FFCFF0FFCFF0FFD0FFFFD0F0FFDFFFFFDFF0\
FFDFFFFFDFF0FFE0FFFFE0FFFFE0FFF0E0FFFFE0FFFFE0FFFFEFFFF0E0FFFFE0FF
F0E0FFFFE0FFF0E0F0FFE0FFF0E0FFFFE0FFF0DFF0FFE0FFF0DFFFFFDFFFF0DFF0FFDFFFF0DFF0\
FFDFFFF0D0F0FFDFFFF0D0FFFFD0FFFFD0FFFFD0FFFFCFFFFFD0FFFFCFFFFFCFFF
FFC0FFFFCFFFFFC0FFFFC0FFFFBFFFFFBFFFFFB0FFFFB0FFFFB0FFFFB0FFFFAFFFFFB0FFFFAFFF\
FFAFFFFFA0FFFFA0FFFF9FFFFFA0FFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFF
FF90FFFF9FFFFF90F0FF90FFFF8FF0FF8FF0F080F0FF8FF0FF80F0FF8FF0F080F0FF80FFFF80F0\
FF80FF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000504050604F506F4F5080505F80505F905060\
9050609F5F6F9F5F70A0607FAF5F7FB06080B06080BF6F8FBF6090C06F9FCF6F9F
D070A0D06FA0DF70B0DF70B0E07FBFE07FBFF080CFF07FCFFF80D0FF80D0FF8FDFFF8FDFFF90DF\
FF8FDFFF9FE0FF9FE0FF9FE0FF9FE0FFA0EFFFA0E0FFAFEFFFAFEFFFB0F0FFB0EF
FFBFF0FFBFEFFFC0F0FFC0F0FFCFF0FFCFF0FFD0FFFFD0F0FFD0FFFFD0F0FFDFFFFFDFFFFFE0FF\
FFE0FFFFE0FFFFE0FFFFEFFFFFE0FFFFEFFFF0E0FFFFEFFFF0E0FFFFEFFFF0E0FF
FFEFFFF0E0FFFFEFFFF0E0FFF0E0FFF0E0FFFFE0FFF0E0FFF0E0FFF0DFF0F0E0FFF0DFF0F0DFFF\
F0DFF0FFDFFFF0D0FFFFDFFFF0D0F0FFD0FFF0D0FFFFD0FFFFCFFFFFD0FFFFCFFF
FFCFFFFFCFFFFFCFFFFFC0FFFFC0FFFFBFFFFFBFFFFFB0FFFFBFFFFFB0FFFFB0FFFFB0FFFFAFFF\
FFAFFFFFAFFFFFA0FFFFA0FFFFA0FFFFA0FFFF9FFFFFA0FFFF9FFFFF9FFFFF9FFF
FF9FFFFF9FFFFF9FFFFF90FFFF90FFFF8FF0FF90FFFF8FF0FF8FFFFF8FF0FF8FFFFF80FFFF8FFF\
FF80FFFF8FFFAF60AF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000050404F604F506F4F507F505F80505F8F5060905060\
9F5F6F9F5F6FA05F70A05F70AF6080B06080BF608FBF608FC06F90C06F9FCF6FA0
CF6FA0DF70AFDF70B0E07FBFE070BFEF7FC0EF7FCFFF80D0FF7FD0FF80DFFF80D0FF8FDFFF8FDF\
FF90DFFF90DFFF9FE0FF9FE0FFA0E0FFA0E0FFAFEFFFAFE0FFB0EFFFB0EFFFBFF0
FFB0EFFFBFF0FFC0F0FFCFF0FFC0F0FFCFF0FFCFF0FFD0FFFFD0F0FFDFFFFFDFFFFFE0FFF0DFFF\
FFE0FFF0E0FFFFE0FFF0E0FFFFEFFFF0E0FFF0EFFFF0E0FFF0EFFFF0E0FFF0EFFF
F0E0FFF0EFFFF0E0F0F0EFFFEFE0F0F0E0FFEFE0F0F0E0FFEFDFF0F0DFF0EFDFF0F0DFF0EFDFF0\
F0DFF0EFD0F0F0DFFFEFD0F0F0D0F0F0D0F0F0D0FFF0CFF0F0D0FFF0CFF0FFD0FF
F0CFF0FFCFFFFFC0FFFFCFFFFFC0FFFFC0FFFFBFFFFFBFFFFFB0FFFFBFFFFFB0FFFFB0FFFFAFFF\
FFAFFFFFA0FFFFA0FFFF9FFFFFA0FFFF9FFFFFA0FFFF9FFFFF9FFFFF9FF0FF9FFF
FF90F0FF9FFFFF90FFFF9FFFFF90F0FF90FFFF8FF0FF90FFF08FF0FF8FFFFF8FF0FF8FFFFF80FF\
FF8FFFFF80FFFF8FFFC070C0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000504050604F506F4F507F505F80505F8F50609050609F5F6F\
9F5F6FA06070A05F7FB06080B06080BF6F8FBF608FC06F9FC06F9FCF70A0D06FA0
DF70AFDF70B0E07FBFE070BFEF7FC0F07FCFFF80D0FF80D0FF8FDFFF8FDFFF8FDFFF8FDFFF90E0\
FF90DFFF9FE0FF9FE0FFA0EFFFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFEF
FFC0F0FFC0F0FFCFF0FFCFF0FFD0FFFFD0F0FFD0FFFFD0FFFFDFFFFFDFFFFFE0FFFFE0FFFFEFFF\
FFE0FFFFEFFFF0E0FFFFEFFFF0EFFFFFEFFFF0EFFFF0EFFFF0EFFFF0EFFFF0EFFF
F0EFFFF0EFFFF0EFFFEFE0FFF0EFFFEFE0FFF0EFFFEFE0FFF0E0FFEFDFF0F0E0F0EFDFF0F0E0FF\
EFDFF0F0DFFFEFDFF0F0DFFFEFD0F0F0D0F0EFD0F0F0D0FFF0D0F0F0D0FFF0CFF0
FFD0FFF0CFF0FFCFFFFFC0FFFFCFFFFFC0FFFFC0FFFFBFFFFFBFFFFFBFFFFFBFFFFFB0FFFFB0FF\
FFAFFFFFAFFFFFA0FFFFAFFFFFA0FFFFA0FFFFA0FFFFA0FFFF9FFFFF9FFFFF90FF
FF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF9FFFFF90FFFF90FFFF90FFFF90FFFF8FFFFF90FF\
FF8FFFFF8FFFFF8FFFFF8FFFCF70CF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000050404F604F506F4F507F505F7F4F5F8F50608F505F905F6F9F5F6F\
A05F70A05F70AF607FAF6080B0608FB0608FC06F90C06F90CF6FA0CF6FA0D070AF
DF70AFE07FB0E070BFEF7FC0EF7FC0F080CFF07FCFFF80D0FF80D0FF8FDFFF8FD0FF90DFFF90DF\
FF9FE0FF90DFFF9FE0FF9FE0FFA0EFFFA0E0FFAFEFFFB0EFFFB0EFFFB0EFFFBFF0
FFBFF0FFC0F0FFC0F0FFCFF0FFCFF0FFD0FFFFD0F0FFDFFFFFDFFFFFE0FFF0DFFFFFE0FFF0E0FF\
FFEFFFF0E0FFF0EFFFF0E0FFF0EFFFF0E0FFF0EFFFEFEFFFF0EFFFEFEFFFF0EFFF
EFEFFFF0EFFFEFE0F0EFEFFFEFE0F0EFEFFFEFE0F0EFE0FFEFE0F0EFE0F0E0DFEFEFDFF0EFDFF0\
EFDFF0EFDFF0EFDFF0EFD0EFEFD0F0EFD0EFEFD0F0EFD0EFF0D0F0EFCFEFF0D0F0
EFCFF0F0CFF0EFC0F0F0CFF0F0C0F0FFC0FFF0BFF0FFC0FFFFBFFFFFBFFFFFB0FFFFB0FFFFAFFF\
FFB0FFFFAFFFFFAFFFFFA0FFFFA0FFFFA0FFFFA0FFFF9FFFFF9FFFFF90F0FF9FFF
FF90F0FF9FFFFF90F0FF9FFFFF90FFFF9FFFFF90FFFF9FFFFF90FFFF90FFFF8FFFFF90FFFF8FFF\
FF8FFFFF80FFFF8FFFFF8FFFF08FF0BF70C0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000504050604F506F4F507F505F7F505F8F50608F505F9F5F6F9F5F6FA05F70\
A05F70AF6080AF6080B0608FB0608FC06F90C06F9FCF70A0CF6FA0D070AFDF70AF
E07FB0E070BFEF7FC0EF7FC0F080CFF07FCFFF80D0FF80D0FF8FDFFF8FDFFF90DFFF90DFFF9FE0\
FF90DFFFA0E0FF9FE0FFA0EFFFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFF0
FFC0F0FFC0F0FFCFFFFFCFF0FFD0FFFFD0FFFFDFFFFFDFFFFFE0FFF0E0FFFFE0FFF0E0FFFFEFFF\
F0E0FFFFEFFFF0EFFFF0EFFFF0EFFFF0EFFFF0EFFFF0EFFFEFEFFFF0F0FFEFEFFF
F0EFFFEFEFFFEFEFFFEFEFF0EFEFFFEFE0F0EFEFFFEFE0F0EFE0FFE0E0F0EFE0F0E0DFF0EFE0F0\
E0DFF0EFDFF0EFDFF0EFDFF0E0D0EFEFD0F0E0D0EFEFD0F0EFCFEFEFD0F0EFCFF0
EFCFF0EFCFEFF0CFF0EFC0F0F0C0F0F0C0F0FFC0FFF0BFF0FFC0FFFFBFFFFFBFFFFFB0FFFFB0FF\
FFAFFFFFAFFFFFAFFFFFAFFFFFA0FFFFAFFFFFA0FFFFA0FFFF9FF0FF9FFFFF90F0
FF9FFFFF90F0FF9FFFFF9FFFFF9FFFFF90FFFF9FFFFF9FFFFF9FFFFF90FFFF90FFFF90FFFF90FF\
FF8FFFFF90FFFF8FFFFF90FFFF8FFFEF8FEFBF70BF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005F4F50604F50704F5F7F4F5F80505F8F505F905F6090506F9F5F709F5F70\
AF607FAF607FB06080B06080BF6F90BF6F90CF6F9FCF6F9FD070A0D070AFDF70B0
DF70B0E07FBFE07FBFF07FCFF07FCFFF80D0FF80D0FF8FDFFF80D0FF8FDFFF8FDFFF90DFFF90DF\
FF9FE0FF9FE0FFA0E0FFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFEFFFC0F0
FFC0F0FFCFF0FFCFF0FFD0FFFFD0F0FFDFFFFFDFFFFFE0FFF0DFFFFFE0FFF0E0FFF0E0FFF0E0FF\
F0EFFFF0E0FFF0EFFFEFE0FFF0EFFFEFEFFFF0EFFFEFEFFFEFEFFFEFEFFFEFEFFF
EFEFFFEFEFFFE0EFF0EFEFFFE0E0F0EFEFF0E0E0F0EFEFFFE0E0F0E0E0F0E0DFEFE0E0F0E0DFEF\
E0DFF0E0DFEFE0DFF0E0D0EFE0D0EFE0D0E0E0D0EFE0CFEFE0D0EFE0CFEFEFD0EF
E0CFEFE0CFEFE0C0EFEFC0EFEFC0EFEFC0F0EFBFEFF0C0F0EFBFF0F0BFF0F0B0F0FFBFFFFFB0FF\
FFB0FFFFAFFFFFAFFFFFA0FFFFAFFFFFA0FFFFA0FFFF9FF0FF9FFFFF90F0FF9FFF
FF90F0FF9FFFFF90F0FF9FFFFF90F0FF9FFFFF90F0FF9FFFFF90FFFF90FFFF90FFFF90FFFF8FFF\
FF8FFFFF8FFFFF90FFFF8FFFFF90FFFF8FFFE08FE0B070BF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005F4F50604F5070505F7F505F8050608F505F905F60905F6F9F5F709F5F70AF607F\
AF607FB06080B06080BF6F90BF6F90C06F9FC06F9FD070A0D070AFDF70B0DF70B0
E07FBFE07FBFEF7FC0F07FCFFF80D0FF80D0FF8FDFFF80D0FF8FDFFF8FDFFF90E0FF90DFFF9FE0\
FF9FE0FFA0E0FFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0
FFCFFFFFCFF0FFD0FFFFD0FFFFDFFFFFDFFFFFE0FFF0DFFFFFE0FFF0E0FFFFE0FFF0E0FFF0EFFF\
F0EFFFF0EFFFEFEFFFF0F0FFEFEFFFF0F0FFEFEFFFF0F0FFEFEFFFEFF0FFEFEFFF
EFF0FFEFEFFFEFF0FFE0EFF0EFF0FFE0EFF0EFEFFFE0EFF0EFEFFFE0E0F0E0E0F0DFDFEFE0E0F0\
E0DFEFE0E0F0E0DFEFE0DFEFDFD0EFE0D0EFDFD0E0E0D0EFE0D0EFE0D0EFE0CFEF
E0CFEFE0CFE0E0CFEFE0C0E0EFC0EFEFC0EFEFC0F0EFBFEFF0BFF0EFBFEFF0BFF0F0B0F0FFBFFF\
FFB0FFFFB0FFFFAFFFFFAFFFFFAFFFFFAFFFFFA0FFFFA0FFFF9FFFFF9FFFFF90F0
FF9FFFFF90FFFF9FFFFF90F0FF9FFFFF90FFFF90FFFF90FFFF9FFFFF9FFFFF9FFFFF90FFFF90FF\
FF90FFFF90FFFF90FFFF90FFFF90FFFF90FFFF90FFDF8FE0B07FBF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F4F50604F506F4F5F704F5F80505F80505F8F5F609050609F5F6F9F5F6FA05F70A05F7F\
AF6080AF6080B0608FBF608FC06F90C06F90CF6FA0CF6FA0D070AFD070AFE070B0
E070BFEF7FC0EF7FC0F07FCFF07FCFFF80D0FF80D0FF8FDFFF8FD0FF90DFFF90DFFF9FE0FF9FDF\
FFA0E0FF9FE0FFA0E0FFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFEFFFC0F0
FFCFF0FFCFFFFFCFF0FFD0FFFFD0FFFFDFFFF0DFFFFFE0FFF0E0FFF0E0FFF0E0FFF0EFFFF0E0FF\
F0EFFFEFEFFFF0EFFFEFEFFFF0F0FFEFEFFFEFF0FFEFEFFFEFF0FFE0EFFFEFF0FF
E0EFFFEFF0FFE0EFF0E0EFF0E0EFF0E0EFFFDFEFF0E0EFF0DFE0F0E0E0F0DFE0EFDFE0EFDFDFEF\
E0E0EFDFDFEFDFDFEFDFD0E0DFDFEFD0D0E0DFD0E0DFCFE0DFD0EFDFCFE0E0CFE0
DFCFE0E0CFE0DFC0DFE0C0E0E0C0E0E0C0EFE0BFE0EFBFEFE0BFE0EFBFEFEFB0EFF0B0F0F0B0F0\
F0B0F0F0AFF0FFAFFFFFAFFFFFAFFFFFA0F0FFA0FFFF9FF0FF9FFFFF90F0FF9FF0
FF90F0FF9FFFFF90F0FF9FFFFF90F0FF90FFFF8FF0FF90FFFF90F0FF9FFFFF90FFFF90FFFF90FF\
FF90FFFF90FFFF90FFFF90FFFF90FFFF90FFFF90FFFF90FFDF8FDFAF7FB0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F4F50604F506F505F704F5F7F505F80505F8F5F609050609F5F6F9F5F6FA06070A05F70AF607F\
AF6080B06F8FBF608FC06F90C06F90CF70A0CF6FA0D070AFD070AFE07FB0E070B0
EF7FBFEF7FC0F080CFF07FCFFF80D0FF80D0FF8FDFFF8FDFFF90DFFF90DFFF9FE0FF9FDFFF9FE0\
FF9FE0FFAFEFFFA0E0FFAFEFFFAFEFFFB0F0FFB0EFFFBFF0FFBFF0FFC0F0FFC0F0
FFCFFFFFCFF0FFD0FFFFD0FFFFDFFFF0DFFFFFE0FFF0DFFFF0E0FFF0E0FFF0EFFFF0E0FFF0EFFF\
EFEFFFF0F0FFEFEFFFEFF0FFEFEFFFEFF0FFEFF0FFEFF0FFE0F0FFEFF0FFE0F0FF
EFF0FFE0EFFFE0F0FFE0EFF0E0F0FFDFEFF0E0F0FFDFEFF0E0EFF0DFE0EFDFE0F0DFE0EFDFE0EF\
DFDFEFDFE0EFDFDFEFDFDFEFD0D0E0DFD0E0D0D0E0DFD0E0DFCFE0DFD0E0DFCFE0
DFCFE0DFC0DFDFC0E0DFC0DFE0C0E0DFBFE0E0C0EFE0BFE0E0BFE0E0B0E0EFB0EFEFB0EFEFB0EF\
EFAFEFF0B0F0F0AFF0FFAFFFFFA0FFFFA0FFFFA0FFFFA0FFFF9FF0FF9FFFFF90F0
FF9FFFFF90F0FF9FFFFF90F0FF90FFFF8FF0FF90FFFF90F0FF90FFFF90FFFF9FFFFF90FFFF90FF\
FF90FFFF9FFFFF90FFFF9FFFFF90FFFF9FFFFF9FFFFF9FFFF090F0D090DFAF7FBF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000504F50
5F40506F4F5F6F4F507F505F7F505F8F50608F505F905F6F905F6F9F5F70A05F70AF607FAF5F7F\
B06080B0608FBF6F90BF6090C06F9FCF6F9FD070A0D06FA0DF70AFDF70B0E07FBF
E070BFEF7FC0F07FC0FF80D0FF7FCFFF80D0FF80D0FF8FDFFF8FD0FF90DFFF90DFFF9FE0FF9FDF\
FFA0E0FFA0E0FFAFEFFFAFE0FFB0EFFFB0EFFFBFF0FFB0EFFFBFF0FFC0F0FFCFF0
FFCFF0FFD0FFFFD0FFFFDFFFF0D0FFFFDFFFF0DFFFF0E0FFEFDFFFF0E0FFEFE0FFF0EFFFEFE0FF\
EFEFFFEFEFFFEFF0FFE0EFFFEFF0FFE0EFFFEFF0FFE0EFFFE0F0FFE0F0FFE0F0FF
E0EFFFE0F0FFDFEFF0E0F0FFDFEFF0DFEFF0DFEFF0DFEFF0DFE0EFDFEFEFD0E0EFDFE0EFD0DFE0\
DFE0EFD0DFE0DFDFEFD0D0E0D0D0E0D0D0DFD0D0E0D0CFDFD0D0E0D0CFDFD0CFE0
D0C0DFD0C0DFD0C0D0DFC0DFD0BFDFDFBFDFDFBFDFDFBFE0DFB0DFDFB0DFDFAFDFE0B0E0E0AFE0\
EFAFEFEFAFEFF0AFF0EFA0EFFFA0F0FFA0F0FFA0FFFF9FF0FF9FF0FF90F0FF9FF0
FF90F0FF90F0FF90F0FF90F0FF8FF0FF8FF0FF8FF0FF90F0FF8FF0FF90FFFF90F0FF9FFFFF90FF\
FF90FFFF90FFFF9FFFFF90FFFF9FFFFF90FFFF9FFFFF90FFFF9FFFE090EFD090DF
A07FB0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000504F505F4F50
6F505F6F4F5F7F505F7F505F8F5F608F5060905F60905F6F9F6070A05F70AF607FAF607FB06F80\
B0608FBF6F90BF6F90C0709FCF6F9FD070A0D070A0DF70B0DF70B0E07FBFE07FBF
F07FC0F07FC0FF80CFFF80D0FF8FD0FF8FD0FF8FDFFF8FDFFF90DFFF90DFFF9FE0FF9FE0FFA0E0\
FFA0E0FFAFEFFFAFE0FFB0EFFFB0EFFFBFF0FFBFEFFFC0F0FFC0F0FFCFFFFFCFF0
FFD0FFFFD0FFFFDFFFFFD0FFFFDFFFF0DFFFF0E0FFF0DFFFF0E0FFF0E0FFF0EFFFEFEFFFF0EFFF\
EFEFFFEFF0FFEFEFFFEFF0FFEFF0FFEFF0FFE0F0FFEFF0FFE0F0FFE0F0FFE0F0FF
E0FFFFE0F0FFE0F0FFDFF0F0DFF0FFDFF0F0DFF0F0DFEFF0DFEFF0D0EFEFDFEFEFD0E0EFD0E0EF\
D0E0EFD0E0EFD0DFE0D0DFE0D0D0E0D0DFE0CFD0DFD0D0E0D0D0DFD0D0E0D0CFDF
D0CFDFD0C0D0D0C0DFD0C0D0D0C0DFD0BFDFDFBFDFD0B0DFDFBFDFDFB0DFDFB0DFDFAFDFE0B0E0\
E0AFE0EFAFEFEFA0E0F0A0EFF0A0EFFFA0F0FF9FF0FFA0FFFF9FF0FF9FFFFF90F0
FF9FFFFF90F0FF90F0FF8FF0FF90F0FF8FF0FF8FF0FF8FF0FF90FFFF90F0FF90FFFF90FFFF9FFF\
FF90FFFF9FFFFF90FFFF9FFFFF90FFFF9FFFFF90FFFF9FFFFF9FFFFFA0FFDF90E0
C090D0A080B0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000504F505F4050604F50
6F4F5070505F7F505F80506080505F905F609050609F5F6F9F5F70A0607FA05F7FAF6080B06080\
BF6F8FBF608FC06F90C06F9FCF70A0CF6FA0D070AFDF70AFE07FB0E070B0EF7FBF
EF7FC0F080CFF07FCFFF80D0FF80D0FF8FDFFF8FD0FF90DFFF90DFFF9FE0FF90DFFF9FE0FF9FE0\
FFA0E0FFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFF
FFCFF0FFD0FFF0D0FFFFDFFFF0DFFFF0DFFFF0DFFFF0E0FFEFE0FFF0EFFFEFE0FFEFEFFFEFEFFF\
EFEFFFE0EFFFEFF0FFE0F0FFEFF0FFE0F0FFE0F0FFE0F0FFE0F0FFDFF0FFE0F0FF
DFF0FFE0F0FFDFF0F0DFF0F0D0F0F0DFF0F0D0EFF0DFEFF0D0EFEFD0EFEFD0E0E0D0E0EFCFE0E0\
D0E0EFCFDFE0D0DFE0CFDFDFCFDFE0CFD0DFCFD0DFCFCFDFCFD0DFCFCFD0CFCFDF
CFC0D0CFC0D0CFBFCFCFC0D0CFBFD0D0BFDFCFB0D0D0BFD0D0B0D0D0B0D0D0AFD0DFAFDFDFAFDF\
E0AFE0DFA0DFE0A0E0E09FE0EFA0EFEF9FE0F09FEFF09FF0FF9FF0FF90F0FF90F0
FF90F0FF90F0FF8FEFFF8FF0FF8FEFFF8FF0FF8FEFFF8FF0FF8FF0FF90F0FF8FF0FF90FFFF90F0\
FF9FFFFF90FFFF9FFFFF90FFFF9FFFFF90FFFF9FFFFF9FFFFF9FFFFF9FFFEF9FEF
CF90DFBF8FCF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000504F50604F5F6F4F5F
70505F7F505F805F608050608F5F60905F609F5F6F9F5F6FA06070A05F7FAF6080AF6080B06F8F\
BF608FC06F90C06F9FCF70A0CF6FA0D070AFD070AFDF7FB0E070B0EF7FC0EF7FC0
F080CFF07FCFFF80D0FF80D0FF8FD0FF8FD0FF90DFFF90DFFF90DFFF90DFFF9FE0FF9FE0FFA0E0\
FFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFF
FFD0FFFFD0FFFFDFFFF0D0FFF0DFFFF0DFFFF0E0FFEFE0FFF0E0FFEFE0FFF0EFFFEFEFFFEFEFFF\
EFEFFFEFF0FFE0F0FFEFF0FFE0F0FFEFFFFFE0F0FFE0FFFFDFF0FFE0FFFFDFF0FF
E0FFFFDFF0FFDFFFFFDFF0F0DFF0F0D0F0F0DFF0F0D0EFF0D0F0F0D0EFEFD0EFEFCFE0E0D0E0E0\
CFE0E0D0E0E0CFDFE0CFDFE0CFD0DFCFDFDFC0D0D0CFD0DFC0CFD0CFD0DFCFCFD0
CFCFD0C0C0CFCFC0D0C0BFCFCFC0D0CFBFD0D0BFD0CFB0CFD0B0D0CFAFCFD0AFD0D0AFD0DFAFD0\
D0A0D0DFAFDFDFA0DFE09FDFE09FDFEF9FE0EF9FE0F09FEFF090EFFF9FF0FF90F0
FF90F0FF8FEFFF90F0FF8FEFFF8FF0FF80EFFF8FF0FF8FF0FF8FF0FF8FF0FF90FFFF90F0FF90FF\
FF90FFFF9FFFFF90FFFF9FFFFF90FFFF9FFFFF90FFFF9FFFFF9FFFFFA0FFF0A0FF
E09FE0C090CFAF90C0000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000050404F5F4F50604F5070505F
704F5F7F505F80505F8F5F608F5060905F6F905F6FA06070A05F70AF607FAF607FB06F8FB0608F\
BF6F90BF6F90CF6F9FCF6F9FD070A0D070AFDF70B0DF70B0E07FBFE07FBFEF7FC0
EF7FC0FF80CFFF80CFFF80D0FF80D0FF8FDFFF8FD0FF90DFFF90DFFF9FDFFF9FDFFFA0E0FF9FE0\
FFA0E0FFA0E0FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFF
FFCFF0FFD0FFF0D0FFF0DFFFF0D0FFF0DFFFEFDFFFF0E0FFEFE0FFEFE0FFEFE0FFEFEFFFE0EFFF\
EFF0FFE0EFFFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FFDFF0FFDFF0FFDFF0FFDFFFFF
DFF0F0DFF0FFD0F0F0DFF0FFD0F0F0D0F0F0D0F0F0D0F0F0CFEFEFCFEFEFCFE0E0CFEFEFC0E0E0\
CFE0E0C0DFDFCFDFE0C0DFDFC0DFDFC0D0D0C0D0D0C0CFD0C0D0D0C0CFD0C0CFD0
C0C0CFC0C0CFC0BFCFC0BFCFC0BFCFC0BFCFC0B0CFCFB0CFC0AFC0CFAFCFC0A0C0CFAFCFCFA0CF\
D0A0D0D09FD0DF9FD0D090CFDF9FDFDF90DFEF90E0EF90E0F090EFF090EFFF90F0
FF8FEFFF8FEFFF80EFFF8FEFFF80EFFF80EFFF80EFFF8FF0FF80EFFF8FF0FF8FF0FF8FF0F08FF0\
FF90FFFF90FFFF90FFFF90FFFF90FFFF90FFFF9FFFFF9FFFFFA0FFFFA0FFFFA0FF
EF9FEFD09FDFB090CF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005040505F4F5F604F506F505F70505F
7F506080505F8F5F608F5060905F6F905F6F9F6070A05F70AF607FAF6080B06F8FB0608FBF6F90\
BF6F90C0709FCF6F9FD070A0D070AFDF70B0DF70B0E07FBFE07FBFEF7FC0EF7FC0
FF80CFFF80CFFF80D0FF80D0FF8FDFFF8FD0FF90DFFF90DFFF9FE0FF9FDFFFA0E0FFA0E0FFAFEF\
FFA0E0FFAFEFFFAFEFFFB0F0FFB0EFFFBFF0FFBFF0FFC0FFFFC0F0FFCFFFFFCFFF
FFD0FFFFD0FFFFDFFFF0D0FFF0DFFFF0DFFFF0E0FFEFE0FFF0EFFFEFE0FFEFEFFFE0EFFFEFF0FF\
E0F0FFEFF0FFE0F0FFE0FFFFE0F0FFE0FFFFDFF0FFE0FFFFDFFFFFDFFFFFDFFFFF
DFFFFFDFFFFFDFFFFFD0F0F0D0FFFFD0F0F0D0F0F0CFF0EFD0F0F0CFEFEFCFEFEFCFE0E0CFEFE0\
C0E0E0CFE0E0C0DFDFCFDFDFC0DFDFC0DFDFBFD0D0C0D0D0C0CFD0C0CFD0BFCFCF
C0CFCFBFC0CFC0C0CFBFBFC0C0BFCFC0BFC0C0BFCFC0B0C0C0B0CFC0AFC0C0AFC0C0A0C0CFAFCF\
CFA0CFD0A0CFCF9FCFD09FD0D090D0DF90D0DF90D0EF90E0EF90E0F090E0F08FE0
FF90F0FF8FEFFF8FEFFF80EFFF80EFFF80EFFF8FF0FF80EFFF8FF0FF8FF0FF8FF0FF8FF0FF90FF\
FF90F0FF90FFFF90FFFF9FFFFF90FFFF9FFFFF9FFFFFA0FFFFA0FFFFA0FFFFA0FF
FFAFFFDFA0E0CF9FDFAF90C0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F4F505F4F506F505F6F4F5F7F505F
7F505F805F6080505F905F60905F609F5F709F5F70A0607FA0607FB06080B06080BF6F8FBF6F8F\
C06F9FC06F9FCF70A0CF6FA0D070AFD070AFDF7FB0E070B0EF7FBFEF7FBFF07FC0
F07FCFFF80D0FF80CFFF8FD0FF8FD0FF90DFFF8FD0FF90DFFF90DFFF9FE0FF9FDFFFA0E0FFA0E0\
FFAFEFFFAFE0FFB0EFFFB0EFFFBFF0FFB0EFFFC0F0FFC0F0FFCFFFFFCFF0FFD0FF
FFD0FFFFD0FFF0D0FFF0DFFFEFDFFFF0DFFFEFDFFFEFE0FFEFE0FFEFEFFFE0E0FFEFEFFFE0EFFF\
E0F0FFE0EFFFE0F0FFDFF0FFE0FFFFDFF0FFDFFFFFDFF0FFDFFFFFD0F0FFDFFFFF
D0FFFFDFFFFFD0F0F0D0FFF0CFF0F0D0FFF0CFF0EFCFF0EFC0EFEFCFEFEFC0E0E0C0EFE0C0E0DF\
C0E0DFBFDFD0C0DFDFBFDFD0C0DFD0BFD0D0BFD0D0BFCFCFBFD0CFBFCFCFBFCFCF
BFC0C0BFC0C0B0BFC0BFBFC0BFBFC0BFBFC0BFB0BFBFB0C0BFAFBFBFAFBFBFA0BFC0A0C0C0A0C0\
CFA0C0C09FC0CF9FC0CF90C0D090CFD08FCFDF90D0DF8FD0E08FDFE08FD0EF8FE0
F080E0FF8FEFFF80E0FF80EFFF7FE0FF80EFFF80EFFF80EFFF80EFFF80EFF080EFFF8FF0FF8FF0\
FF90FFFF8FF0FF90FFFF90FFFF90FFFF90FFFF9FFFFF9FFFFFA0FFFFA0FFFFAFFF
FFAFFFEFAFF0D09FDFBF9FD0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F4F505F4F506F505F6F4F5F7F505F7F505F
805F60805060905F60905F609F5F6F9F5F70A0607FA0607FAF6080AF6080BF6F8FBF6F8FC06F90\
C06F90CF70A0CF70A0D070AFD070AFDF7FB0DF70B0E07FBFEF7FBFF080C0F07FC0
FF80CFFF80CFFF8FD0FF8FD0FF8FDFFF8FD0FF90DFFF90DFFF9FE0FF9FDFFFA0E0FFA0E0FFAFEF\
FFAFE0FFB0EFFFB0EFFFB0F0FFB0EFFFBFF0FFBFF0FFC0FFFFC0F0FFCFFFFFCFFF
FFD0FFF0D0FFFFDFFFF0D0FFF0DFFFEFDFFFF0E0FFEFE0FFEFE0FFE0E0FFEFEFFFE0EFFFE0F0FF\
E0F0FFE0F0FFE0F0FFE0FFFFDFF0FFE0FFFFDFFFFFDFFFFFDFFFFFDFFFFFD0FFFF
D0FFFFD0F0F0D0FFF0CFF0F0CFF0F0C0F0EFC0F0EFC0EFE0C0EFE0BFEFDFC0EFE0BFE0DFBFE0DF\
BFDFD0BFE0D0BFDFCFBFDFD0B0D0CFBFD0CFB0D0CFBFD0CFB0CFC0BFCFCFB0C0C0
BFC0C0B0C0C0BFC0C0B0BFBFBFBFC0B0B0BFBFB0BFB0AFBFBFAFBFB0A0B0BFAFBFBFA0BFC0A0C0\
BF9FBFC09FC0C090BFCF90BFCF8FBFD090CFD08FCFDF8FD0DF8FCFE08FD0E080D0
F08FE0F080E0FF80EFFF7FE0FF80EFFF7FE0FF80EFFF80EFFF80EFF080EFFF80EFFF80EFFF8FF0\
FF8FF0FF90FFFF90FFFF90FFFF90FFFF9FFFFF9FFFFFA0FFFFA0FFFFAFFFFFA0FF
FFAFFFF0AFFFE0AFF0CFA0DFB0A0D0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000504F505F4F50604F5F6F4F5070505F70505F805060
80505F8F5F608F5060905F6F905F6F9F6070A05F70AF607FAF607FB06080B0608FBF6F90BF6F90\
C06F9FC06F9FD070A0CF6FA0DF70AFDF70AFE07FB0E070B0EF7FC0EF7FC0F080CF
F07FCFFF80D0FF80D0FF8FD0FF8FD0FF90DFFF90DFFF90DFFF90DFFF9FE0FF9FE0FFA0E0FFA0E0\
FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFF0FFD0FF
F0CFFFF0D0FFF0D0FFF0DFFFEFD0FFEFDFFFEFDFFFEFE0FFE0E0FFE0E0FFE0EFFFE0EFFFDFEFFF\
E0F0FFDFF0FFE0F0FFDFF0FFDFFFFFDFF0FFDFFFFFD0F0FFDFFFFFD0FFF0D0FFF0
CFF0EFCFF0EFC0F0EFC0F0EFBFEFE0BFF0E0BFEFDFBFEFDFB0E0D0BFE0D0B0E0D0B0E0D0AFDFCF\
B0DFCFAFD0C0B0DFCFAFD0C0B0D0C0AFCFC0B0CFC0AFCFBFAFCFC0AFC0BFB0C0BF
AFBFBFB0C0BFAFBFBFB0BFBFAFB0B0B0B0BFAFAFB0B0AFB0AFA0B0B0A0B0B0A0AFB0A0B0B09FB0\
BF9FB0BF90B0BF90B0BF8FB0C08FBFC08FBFCF8FC0CF80C0D080CFD080C0E080D0
E07FD0EF80DFF07FDFFF7FE0FF7FE0FF7FE0F07FE0FF7FEFF07FE0FF7FEFF07FE0FF80EFF080EF\
FF8FF0FF8FF0FF90FFFF90FFFF90FFFF90FFFF9FFFFF9FFFFFA0FFFFA0FFFFAFFF
FFA0FFFFB0FFEFAFEFDFAFEFBFA0D0AFA0CF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F4F5060505F604F5F70505F70505F7F506080505F
8F5F608F5F60905F6F905F6F9F6070A06070AF607FAF607FB06F80B06080BF6F8FBF6F90C0709F\
C06F9FCF70A0CF70A0DF7FAFDF70AFE07FB0E07FB0EF7FC0EF7FC0F080CFF07FCF
FF80D0FF80D0FF8FD0FF8FD0FF90DFFF90DFFF90DFFF90DFFF9FE0FF9FE0FFA0E0FFA0E0FFAFEF\
FFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFD0FFFFCFFF
FFD0FFF0D0FFF0DFFFEFD0FFF0DFFFEFDFFFEFE0FFE0E0FFEFEFFFE0E0FFE0EFFFE0EFFFE0F0FF\
E0F0FFE0FFFFDFF0FFE0FFFFDFFFFFDFFFFFD0FFFFDFFFFFD0FFF0D0FFF0CFF0EF
CFFFF0C0F0EFC0F0EFBFEFE0BFF0E0B0EFDFB0EFDFB0EFD0B0EFD0AFE0CFB0E0CFAFDFCFAFE0CF\
AFDFC0AFDFC0AFD0BFAFD0C0AFCFBFAFD0BFA0CFB0AFCFBFAFC0B0AFC0BFA0BFB0
AFC0BFAFBFB0AFBFB0AFB0B0AFBFB0AFB0B0AFB0B0AFAFAFB0AFB0AFA0AFB0A0AFAF9FAFB09FB0\
B090AFBF90B0B08FAFBF8FB0BF8FB0C08FBFC080BFCF8FC0CF80BFD080C0DF7FC0
E080D0E07FD0F07FDFF07FDFFF7FEFFF7FE0FF7FE0FF7FE0FF7FEFF07FE0FF80EFF080EFFF8FF0\
FF8FF0FF8FF0FF8FF0FF90FFFF90FFFF9FFFFF9FFFFFA0FFFFA0FFFFA0FFFFA0FF
FFAFFFFFB0FFF0B0F0DFAFEFCFAFE0B0A0D0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000504F50604F5F604F506F505F70505F7F50607F505F805F60
8F5060905F60905F609F5F6F9F5F70A0607FA0607FAF6080AF6080B06F8FB0608FBF6F90C06F90\
CF709FCF6F9FD070A0D070AFDF7FB0DF70B0E07FBFE07FBFF07FC0F07FC0FF80CF
FF80CFFF8FD0FF80D0FF8FD0FF8FD0FF90DFFF90DFFF9FDFFF9FDFFFA0E0FF9FE0FFA0E0FFA0E0\
FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFD0FF
F0CFFFF0D0FFEFD0FFF0DFFFEFD0FFEFDFFFE0DFFFE0E0FFE0E0FFE0EFFFDFE0FFE0EFFFDFEFFF\
E0F0FFDFF0FFDFFFFFDFF0FFDFFFFFD0F0F0D0FFF0CFF0EFCFFFF0C0F0EFC0F0EF
BFF0E0BFF0E0B0EFDFB0EFDFAFEFD0B0EFD0AFE0CFAFE0CFA0E0C0AFE0CFA0DFC0A0DFC0A0D0BF\
A0DFBF9FD0B0A0D0B09FCFB0A0CFB09FC0AFA0CFB09FC0AFA0C0B09FBFAFA0BFB0
9FBFAFA0BFAF9FB0AFA0B0AFA0AFAFA0B0AFA0AFAFA0AFAFA0A0A0AFA0AFA09FA0AF9FAFAF90A0\
AF90AFAF8FA0B08FAFB080A0BF80AFBF80AFC080B0C080B0CF80BFCF7FB0D07FC0
D07FC0E07FCFE070CFEF7FDFF070DFFF7FE0F070DFFF7FE0F070E0FF7FE0F07FE0FF80EFF080EF\
FF8FF0F080EFFF8FF0FF8FF0FF90FFFF90FFFF9FFFFF9FFFFFA0FFFF9FFFFFA0FF
FFAFFFFFB0FFF0B0F0E0B0EFCFAFE0BFAFDFAFA0CF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000504F505F4F5F604F506F505F6F505F7F50607F505F805F60805F60
8F5F608F5F609F606F9F5F6FA06070A06070AF6080AF6080B06F8FB0608FBF6F90BF6F90C0709F\
CF6F9FD070A0D070A0DF7FAFDF70AFE07FBFE07FBFEF7FC0EF7FC0FF80CFFF80CF
FF80D0FF80D0FF8FD0FF8FD0FF90DFFF90DFFF9FDFFF9FDFFF9FE0FF9FE0FFA0E0FFA0E0FFAFEF\
FFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFD0FFFFCFFF
FFD0FFF0D0FFF0DFFFEFD0FFEFDFFFE0DFFFEFE0FFE0E0FFE0EFFFE0E0FFE0EFFFDFEFFFE0F0FF\
DFF0FFE0FFFFDFFFFFDFFFFFD0F0F0D0FFF0CFF0EFC0FFEFC0F0E0C0F0EFBFF0E0
BFF0E0B0EFDFB0EFDFAFEFD0AFEFD0A0E0CFA0EFCFA0E0C0A0E0C09FDFBFA0DFBF9FDFBF9FDFBF\
9FD0B09FD0B09FCFAF9FCFAF9FC0AF9FCFAF90C0A09FC0AF90BFA09FBFAF9FBFA0
9FBFAF9FB0A09FB0AF9FAFA09FAFA09FAFA0A0AFAF9FA0A0A0A0A0A09FA0A0A0A0A09FA0AF90A0\
A08F9FAF90A0AF8FA0AF80A0AF80A0BF80AFBF7FAFBF80AFBF7FAFC07FB0CF7FB0
D07FC0DF70C0E07FCFE070D0F070DFF070DFFF70E0F070DFFF70E0F070E0FF7FEFF07FE0FF80EF\
F080EFFF8FF0FF8FF0FF90FFFF90FFFF9FFFFF90FFFF9FFFFF9FFFFFA0FFFFA0FF
FFB0FFFFB0FFFFBFFFE0B0EFD0B0E0C0AFDFB0AFD0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F4F505F4F5060505F6F4F5F70505F70505F7F506080505F8F5F60
8F5060905F6F905F6F9F60709F5F70A0607FA0607FB06F80B06080BF6F8FBF6F8FC06F9FC06F90\
CF70A0CF70A0D070AFD070AFDF7FB0E070B0EF7FBFEF7FBFF07FC0F07FC0FF80CF
FF80CFFF8FD0FF8FD0FF90DFFF8FD0FF90DFFF90DFFF9FE0FF9FDFFFA0E0FFA0E0FFAFE0FFA0E0\
FFAFEFFFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0FFFFC0F0FFCFFFFFCFFFFFD0FF
F0CFFFF0D0FFEFD0FFEFDFFFE0D0FFEFDFFFE0DFFFE0E0FFDFE0FFE0EFFFDFEFFFE0EFFFDFEFFF\
DFF0FFDFF0FFDFFFFFCFF0F0CFF0F0C0F0EFC0F0EFBFEFE0BFF0E0B0EFDFB0F0DF
AFEFD0AFEFD0A0E0CFA0EFCF9FE0C09FE0C09FE0C09FE0C090DFB090DFB090D0B090D0B08FCFAF\
90D0AF8FCFA090CFA08FC09F90C0A08FC09F90C09F8FBF9F90BF9F90B09F90B0A0
90B09F90B09F90AF9F90AF9F90A09F9FA0A090A09F9FA09F909F9F9F9F9F9F909F9F909F9F8F90\
A08F9FA08090A0809FA07F90AF7F9FAF7F9FB07FA0B07FA0BF7FA0BF70A0C07FB0
CF70B0D070BFD070C0E070CFE06FCFF070DFF06FDFFF70DFF06FDFF070E0F070E0FF7FE0F07FE0\
FF80EFF080EFFF8FF0F08FF0FF90FFFF90FFFF9FFFFF90FFFF9FFFFFA0FFFFAFFF
FFAFFFFFB0FFFFB0FFEFBFF0D0B0E0C0B0DFB0AFD0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F4F5F5F4F5060505F6F505F70505F70505F7F5F608050608F5F608F5F60
905F6F905F6F9F60709F5F70A0607FA0607FAF6F80AF6080BF6F8FBF6F8FC06F90C06F90CF70A0\
CF70A0D070AFD070AFDF7FB0E07FB0EF7FBFEF7FBFF080C0F080C0FF80CFFF80CF
FF8FD0FF8FD0FF8FDFFF8FD0FF90DFFF90DFFF9FDFFF9FDFFFA0E0FFA0E0FFAFE0FFA0E0FFAFEF\
FFAFEFFFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFD0FFF0CFFF
FFD0FFF0D0FFF0D0FFEFD0FFEFDFFFE0DFFFEFE0FFE0E0FFE0EFFFDFEFFFE0EFFFDFEFFFE0F0FF\
DFF0FFD0FFFFCFF0F0CFFFF0C0F0EFC0F0EFBFF0E0BFF0E0B0EFDFB0F0DFAFEFD0
AFF0D0A0EFCFA0EFCF9FE0C09FE0C090E0BF90E0BF90DFB090DFB08FD0AF90DFAF8FD0AF8FD0AF\
8FCFA08FCFA08FC09F8FCF9F8FC09F8FC09F8FBF908FBF9F8FBF908FBF9F8FB090
90B09F8FAF9090AF9F8FA09090AF9F90A09F90A09F909F909F9F9F9090909F909F9090909F909F\
9F8F909F8F909F8090A08090A07F90AF7F9FAF7F9FB07F9FB0709FBF7FA0BF70A0
CF70B0CF70B0D070BFDF70C0EF70CFEF6FD0FF70DFF06FDFFF70DFF070DFFF7FE0F07FE0FF80EF\
F080EFFF8FEFF080F0FF90FFFF8FFFFF90FFFF90FFFF9FFFFF9FFFFFA0FFFFAFFF
FFB0FFFFB0FFFFBFFFEFBFF0DFBFEFCFB0E0BFB0DF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504F50504F50604F5F604F506F505F6F505F7F505F7F505F805F608050608F5F60
8F5F6090606F905F6FA06070A0607FAF6080AF6080B06F8FB06F8FBF6F90BF6F90C0709FCF6F9F\
D070A0D070A0DF70AFDF70AFE07FBFE07FB0EF7FC0EF7FC0F080CFF07FCFFF80D0
FF80CFFF8FD0FF8FD0FF90DFFF90D0FF9FDFFF90DFFF9FE0FF9FDFFFA0E0FFA0E0FFAFEFFFAFE0\
FFB0EFFFAFEFFFB0F0FFBFEFFFBFF0FFBFF0FFC0FFFFC0F0FFCFFFF0CFFFF0D0FF
F0CFFFF0D0FFEFD0FFEFD0FFE0D0FFEFDFFFE0DFFFE0E0FFDFE0FFE0EFFFDFEFFFDFEFFFDFEFFF\
D0F0FFCFF0EFCFF0EFC0F0EFC0F0EFB0EFDFB0F0DFAFEFD0AFEFD0A0EFCFA0EFD0
9FE0CF9FEFCF90E0BF90E0BF8FDFBF8FE0BF8FDFB08FDFB080D0AF8FD0AF80CFA080D0A080CF9F\
80CF9F7FC09080C09080C09080C0907FBF8F80BF8F80B08F80B08F80AF8F80B08F
80AF8F8FAF9080A08F80A08F809F8F8FA0908F9F908F9F908F908F9090908F8F8F908F90908F90\
908F90907F8F9F808F907F8F9F7F8F9F708FA07090A07090AF7090AF6F90BF70A0
BF6FA0C070AFCF6FB0D070BFDF6FC0EF6FCFEF6FD0F06FDFF06FD0F070DFF070DFF07FE0F07FE0\
F080EFF080EFFF8FF0F08FF0FF90FFFF8FFFFF90FFFF90FFFFA0FFFFA0FFFFAFFF
FFAFFFFFBFFFFFBFFFF0BFF0DFBFEFD0BFEFBFB0DFAFBFDF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504F5060505F604F5F6F505F6F505F7F5F607F505F805F608050608F5F608F5F60
905F6F905F6F9F6070A06070AF607FAF607FB06F8FB06F8FBF6F90BF6F90C0709FC0709FCF70A0\
CF70A0DF7FAFDF70AFE07FB0E07FB0EF80BFEF7FBFF080C0F080C0FF80D0FF80CF
FF8FD0FF8FD0FF90DFFF90D0FF90DFFF90DFFF9FE0FF9FDFFFA0E0FFA0E0FFAFEFFFAFE0FFB0EF\
FFAFEFFFB0F0FFB0EFFFBFF0FFBFF0FFC0FFFFC0F0FFCFFFFFCFFFFFD0FFF0CFFF
F0D0FFEFD0FFF0D0FFEFD0FFEFDFFFE0DFFFE0E0FFE0E0FFE0E0FFDFE0FFDFEFFFDFEFFFDFF0FF\
CFF0F0CFF0F0C0F0EFBFF0E0B0EFDFB0F0E0AFEFD0AFEFD0A0EFCFA0EFCF9FEFCF
9FEFC090E0BF90E0BF8FDFB08FE0B080DFAF80DFAF80D0A080D0A07FD09F80D09F7FCF9F7FCF9F\
7FC0907FCF907FC08F7FC08F7FBF807FBF8F7FB0807FB0807FB08080B08F7FAF80
80AF8F7FA08080A08F7FA08080A08F809F8F8F9F8F80908F8F908F808F8F8F908F8F8F8F908F90\
8F808F90808F907F8F907F8F9070809F708F9F708FA0708FA06F8FAF7090AF6F90
BF6FA0BF6FA0CF70B0CF6FB0DF6FC0DF6FC0EF6FCFF06FD0F06FDFF06FD0F070DFF070DFF07FE0\
F07FE0FF80EFF080EFFF8FF0FF8FF0FF90FFFF90FFFF9FFFFF9FFFFFAFFFFFAFFF
FFB0FFFFB0FFFFC0FFF0BFFFEFC0F0D0BFEFC0C0E0B0BFDF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F50\
5F4F505F4F506F505F6F4F5F70505F70505F7F5F607F505F805F608F5F60905F6F
905F6F9F60709F5F70A0607FA0607FAF6080AF6080B06F8FB06F8FC06F90C06F90CF709FCF6F9F\
D070A0D070AFDF7FB0DF70B0E07FBFE07FBFEF80C0EF7FC0F080CFFF80CFFF8FD0
F080CFFF8FD0FF8FD0FF90DFFF90D0FF9FDFFF9FDFFFA0E0FF9FDFFFA0E0FFA0E0FFAFEFFFAFE0\
FFB0EFFFB0EFFFBFF0FFB0EFFFBFF0FFC0F0FFC0FFFFC0F0FFCFFFF0CFFFF0D0FF
EFCFFFEFD0FFEFD0FFEFD0FFE0D0FFE0DFFFDFDFFFE0E0FFDFE0FFDFEFFFD0EFFFDFEFFFCFEFF0\
CFF0EFBFEFE0BFF0E0B0EFDFB0EFDFAFEFD0AFEFD09FE0CF9FEFCF90E0C090EFC0
8FE0BF8FE0BF80DFB080DFB07FD0A07FDFAF7FD0A07FD0A070CF9F70CF9F70CF9070CF906FC08F\
70C08F6FBF8070C0806FBF7F70BF7F70B07F70B07F6FAF7070AF7F70AF7F70AF7F
70A07F7FA07F709F7F7F9F8070907F7F90807F90808090807F8F80808F8080808080808F807F80\
8F7F8F8F7F808F7F808F707F907080906F7F907080906F7F9F6F809F6F80AF6F90
AF6F90B06FA0BF6FA0CF6FB0CF60B0DF6FBFE060C0F06FD0EF60D0F06FD0F070D0F070DFF070DF\
F07FE0F07FE0FF80EFF080EFFF8FFFFF8FFFFF90FFFF90FFFFA0FFFFA0FFFFAFFF
FFAFFFFFB0FFFFBFFFFFC0FFE0BFF0DFC0EFC0BFE0BFC0E0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F505F\
5F4F5060505F6F505F70506070505F7F5F607F5060805F608F5F608F5F6F905F6F
9F60709F6070A0607FA0607FAF6F80AF6080B06F8FB06F8FC07090C06F90CF709FCF709FD070A0\
D070A0DF7FAFDF70AFE07FBFE07FBFEF80C0EF7FC0F080CFF080CFFF80D0FF80CF
FF8FD0FF8FD0FF90DFFF90D0FF9FDFFF9FDFFF9FE0FF9FDFFFA0E0FFA0E0FFAFEFFFAFE0FFB0EF\
FFB0EFFFBFF0FFBFEFFFC0F0FFBFF0FFC0FFFFC0F0FFCFFFFFCFFFFFD0FFF0CFFF
F0D0FFEFD0FFEFD0FFE0D0FFEFDFFFE0DFFFE0E0FFDFDFFFE0E0FFDFE0FFDFEFFFD0EFF0CFF0F0\
C0EFEFBFF0E0B0EFDFB0F0DFAFEFD0AFEFD09FEFCF9FEFCF90E0C090EFC08FE0BF
8FE0BF80E0B080E0B07FDFAF7FDFAF70D0A070D0A070D09F70D09F6FCF9070CF906FC08F6FCF8F\
6FC0806FC0806FBF7F6FBF7F6FB0706FB0706FAF706FB0706FAF7070AF706FA070
70A07F709F70709F7F709F707F9F7F70907F7F90807F8F7F7F8F807F807F8080807F8080808080\
807F808F7F8080707F8F70808F70809070808F6F7F906F7F906F7F9F6F80A06080
AF6F90B06F9FBF6FA0BF6FA0CF6FB0D060B0E06FC0EF60CFF06FD0F06FD0F070DFF070D0F07FE0\
F07FE0FF80EFF080EFFF8FF0FF8FF0FF90FFFF90FFFF9FFFFF9FFFFFAFFFFFAFFF
FFB0FFFFB0FFFFC0FFF0C0FFEFCFFFDFC0F0D0CFEFBFC0E0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F4F50\
60505F604F506F505F6F505F7F50607F505F805F608050608F5F608F5F60905F6F
905F6F9F60709F6070A0607FAF607FB06F8FB06080BF6F8FBF6F90C0709FC06F9FCF70A0CF70A0\
D07FAFD070AFDF7FB0DF7FB0E07FBFE07FBFEF80C0F07FC0FF80CFF080CFFF8FD0
FF8FD0FF90D0F08FD0FF90DFFF90DFFF9FDFFF9FDFFFA0E0FFA0E0FFAFE0FFA0E0FFAFEFFFAFEF\
FFB0EFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFCFFFF0CFFFF0D0FF
EFCFFFEFD0FFE0D0FFEFD0FFE0D0FFE0DFFFDFDFFFDFE0FFDFE0FFDFEFFFD0E0F0CFEFF0C0EFEF\
BFEFEFB0EFDFB0EFDFA0EFD0A0EFD09FE0CF9FEFCF90E0C090E0BF80E0B080E0B0
7FDFAF7FDFAF70DFA070DFA06FD09F6FD09F60CF906FCF9060CF8F6FCF8F60C08060C08060BF7F\
60BF7F5FBF7060BF7060B06F60B07060AF6F60AF6F60AF6060AF6F60A06F6FA06F
609F6F6F9F7060906F6F90706F90707090706F8F70708F7F70807070807F707F707F7F7F7F7F7F\
807F807F707F80707F806F7F8F707F806F7F8F6F7F8F6070906F709060709F6F80
A0608FAF6F90AF6090BF6FA0C060A0CF60B0D060BFE060C0EF60CFF06FD0EF6FD0F070DFF070DF\
F07FE0F07FE0FF80EFF080EFFF8FFFFF8FFFFF90FFFF90FFFFA0FFFFA0FFFFAFFF
FFAFFFFFBFFFFFBFFFF0C0FFE0C0F0DFCFF0CFC0EFC0CFE0AFC0DF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504F5060505F\
604F5F6F505F6F505F70506070505F805F60805F608F5F608F5F6090606F905F6F
9F60709F6070A0607FA0607FAF6F80AF6080BF6F8FBF6F8FC07090C06F90CF70A0CF70A0D070A0\
D070AFDF7FB0DF7FB0E07FBFE07FBFEF80C0EF80C0F080CFF080CFFF8FD0FF8FD0
FF8FD0FF8FD0FF90DFFF90DFFF9FDFFF9FDFFFA0E0FFA0DFFFA0E0FFA0E0FFAFEFFFAFEFFFB0EF\
FFB0EFFFBFF0FFBFF0FFC0F0FFBFF0FFC0FFFFC0FFFFCFFFFFCFFFFFD0FFF0CFFF
F0D0FFEFD0FFEFD0FFE0D0FFE0DFFFE0DFFFE0E0FFDFDFFFDFE0FFD0E0FFD0EFFFC0E0EFC0EFEF\
B0EFE0B0EFDFAFEFD0A0EFD09FE0CF9FEFCF90E0C08FE0C080E0B080E0B07FDFAF
7FE0AF70DFA070DFA06FD09F6FD09F60CF9060CF9060C08F60C08F5FBF8060BF805FB07F60B07F\
5FAF7060B07F5FAF7060AF705FA06F60A06F5FA06060A0605F9F6060A06F609F60
609F6F609060609F6F60906F6F906F608F6F6F8F706F80707080706F7F70707F70707F707F7F7F\
70707F7F707F7F6F70806F7F7F6F7F806F7F806F708F6F7080606F906F7090607F
9F6F80A0608FAF6F90B0609FBF60A0C060AFDF6FBFE060C0F06FCFEF6FCFF070D0F070D0F070DF\
F070DFFF80EFF080EFFF80F0FF80F0FF90FFFF90FFFF9FFFFF9FFFFFAFFFFFAFFF
FFB0FFFFBFFFFFC0FFF0C0FFEFCFF0DFC0EFD0CFEFC0CFE0B0CFE0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005F4F5F5F4F50\
60505F6F505F70505F70505F7F50607F505F805F608050608F5F608F5F6090606F
905F6F9F6070A0607FAF6F80AF6080B06F8FB06F8FBF6F90BF6F90C0709FC06F9FCF70A0CF70A0\
D070AFD070AFDF7FB0DF7FB0E07FBFEF7FBFF080C0F07FC0FF80CFF080CFFF8FD0
F08FCFFF90D0F090D0FF90DFF090D0FF9FDFFF9FDFFFA0E0FFA0DFFFAFE0FFAFE0FFB0EFFFAFE0\
FFB0EFFFB0EFFFBFF0FFBFEFFFC0F0FFC0F0FFCFFFFFCFF0FFCFFFF0CFFFF0D0FF
EFCFFFEFD0FFE0D0FFEFD0FFE0D0FFE0DFFFDFDFFFDFE0FFD0DFFFD0E0FFC0E0EFC0EFEFB0E0DF\
B0EFDFA0E0D0A0EFD09FE0CF90E0CF8FE0C08FE0BF80DFB080E0B07FDFAF70DFAF
6FDFA06FDFA060D0906FCF905FC08F60C0805FBF7F5FBF805FB07F5FB07F5FAF705FAF7050A06F\
5FA06F509F6F5F9F6F509F605F9F6F5F90605F9060508F5F5F90605F8F5F5F9060
5F8F5F5F8F605F8060608F605F806060806060806060806F607F606F7F6F6F706F6F70706F6F6F\
706F70706F70706F707060707F6F707F6070806F707F606F80606F80606F8F6070
90607F9F6080A0608FAF6090B0609FC060AFCF60B0DF60BFEF60C0EF6FCFEF6FCFF070D0EF70D0\
F07FDFF07FE0F080EFF080EFFF8FFFFF8FFFFF90FFFF90FFFF9FFFFFA0FFFFAFFF
FFB0FFFFBFFFFFC0FFF0CFFFE0C0F0DFCFF0CFCFEFC0CFE0B0CFDF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F505F5F4F5060505F\
6F505F70506070505F7F5F607F5060805F60805F608F5F6F8F5F6090606F905F6F
9F60709F6070A06F7FA0607FB06F8FB06F8FBF6F90BF6F90C0709FC0709FCF70A0CF70A0D07FAF\
D07FAFDF7FB0DF7FB0E080BFEF7FBFEF80C0F080C0FF80CFFF80CFFF8FD0FF8FD0
FF90D0FF90D0FF90DFFF90DFFF9FDFFF9FDFFFA0E0FFA0E0FFAFE0FFAFE0FFAFEFFFAFEFFFB0EF\
FFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFD0FFFFCFFFFFD0FFF0CFFF
F0D0FFEFD0FFEFD0FFE0D0FFE0DFFFE0DFFFE0DFFFDFDFFFDFE0FFCFE0F0C0E0EFBFE0E0B0EFE0\
AFE0DFA0EFD09FE0CF9FEFCF8FE0C08FE0C080E0BF80E0B07FDFAF70E0AF6FD0A0
70D09F6FCF906FCF9060BF8060BF805FB07F5FB0705FAF705FAF705FA0705FA070509F6F5F9F6F\
5090605F9060508F605F9060508F605F8F6050805F5F8F5F50805F5F805F50805F
5F80605F805F5F80605F7F5F6080605F7F60607F605F7F60607F6F60706060706F60706F6F7070\
6F6F6F706F706F606F7060707060707F6F707F60708060707F606F80607080606F
90607F90607FA0608FA0608FB0609FBF60A0CF60B0DF60BFEF6FC0EF6FCFF06FD0EF6FD0F07FDF\
F07FDFF07FE0F07FEFFF8FF0FF80FFFF8FFFFF90FFFF9FFFFF9FFFFFAFFFFFAFFF
FFBFFFFFBFFFFFCFFFF0CFFFEFD0FFDFCFEFD0D0EFBFCFE0B0D0E0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F5060505F604F5F\
6F505F6F505F70506070505F7F5F607F5060805F608F5F60905F6F905F6F9F6070
9F6070A0607FA0607FAF6F80AF6080B06F8FB06F8FBF6F90BF6F90C0709FC0709FCF70A0D070A0\
DF7FAFDF70AFE07FB0E07FB0EF80BFEF7FBFF080C0F080CFFF8FCFF080CFFF8FD0
F08FD0FF90D0F090D0FF9FDFFF9FDFFFA0DFF09FDFFFA0E0FFA0E0FFAFE0FFAFE0FFB0EFFFB0EF\
FFBFEFFFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFCFFFF0CFFFF0D0FF
EFCFFFEFD0FFE0CFFFE0D0FFE0D0FFE0DFFFDFD0FFDFDFFFD0DFFFCFE0F0BFDFE0B0E0E0AFE0D0\
A0E0D09FE0CF9FE0CF8FE0C08FE0BF80DFB07FE0B070DFAF70DFAF6FCF9F6FCF90
60C08F60BF805FB07F5FB07F5FA0705FA06F509F60509F60509060509060508F5F508F6050805F\
50805F4F7F5050805F507F50507F5F4F7F50507F5F50705050705F50705050705F
50705050705F50705050705F506F5F5F705F506F5F5F70605F6F5F5F70605F6F60606F60606F60\
6F6F6F60606F6F606F6F5F6F70607070606F706070705F6F7F606F7F5F6F80606F
80606F8F607F905F7FA0608FAF6090BF60A0C060AFD060B0E060BFEF6FCFEF6FCFF070D0EF70D0\
F07FDFF07FE0F080EFF080F0FF8FFFFF8FFFFF90FFFF90FFFFA0FFFFA0FFFFB0FF
FFB0FFFFC0FFFFC0FFF0CFFFE0CFF0DFD0F0CFCFEFC0D0EFB0D0E0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000504F5060505F60505F6F505F\
6F505F705F6070505F7F5F607F5060805F60805F608F5F6F8F5F6F9060709F6070
9F607FA0607FAF6F80AF6F80B06F8FB06F8FBF7090BF6F90C0709FC0709FCF70A0CF70A0D07FAF\
D07FAFDF7FB0E07FB0E080BFEF7FBFF080C0F080C0FF80CFF080CFFF8FD0FF8FD0
FF90D0F090D0FF9FDFFF9FDFFF9FDFFF9FDFFFA0E0FFA0E0FFAFE0FFAFE0FFB0EFFFB0EFFFB0EF\
FFB0EFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFCFFFF0CFFFFFD0FFF0CFFF
F0D0FFEFCFFFEFD0FFE0D0FFE0DFFFDFD0FFDFDFFFDFDFFFD0E0FFC0DFEFBFE0E0AFE0DFAFE0DF\
9FE0CF9FE0CF90E0C08FE0C080DFB080E0B070DFAF70D0A06FCF906FC08F60B07F
60B07F5FAF705FAF705F9F6F5F9F6F50905F50905F508F5F508F5F4F80505080504F7F50000000\
0000000000000000000000000000000000000000000000000000000000004F6050
506F504F6050506F5F506050506F5F5060505F6F5F50605F5F6F5F50605F5F60605F605F606F60\
5F606060606F605F606F606F6F5F6F70607070606F70606F705F6F7F606F7F5F6F
80606F805F7090607F9F6080AF6090B0609FC060AFD060B0E060C0EF60C0F06FD0EF6FD0F070DF\
F070DFF07FEFF07FEFFF80F0FF80FFFF8FFFFF90FFFF9FFFFFA0FFFFAFFFFFB0FF
FFBFFFFFC0FFFFD0FFF0CFFFEFD0F0DFD0EFD0D0F0BFD0EFB0DFE0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F4F5F5F4F5060505F604F5F\
6F505F70505F7F5F607F505F805F60805F608F5F608F5F60905F6F905F6F9F6070
9F6070A0607FA0607FAF6F80AF6080B06F8FB06F8FBF7090BF6F90C0709FCF709FD07FA0D070A0\
DF7FAFDF7FB0E07FBFE07FBFEF80C0EF80C0F080CFF080CFFF8FCFF08FCFFF90D0
F08FD0FF90D0F090D0FF9FDFF09FDFFFA0DFF0A0DFFFA0E0F0A0E0FFAFE0FFAFE0FFB0EFFFB0EF\
FFB0EFFFBFEFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFCFFFF0CFFFF0D0FF
EFCFFFEFD0FFE0CFFFE0D0FFDFD0FFDFD0FFDFD0FFDFDFFFCFDFF0BFDFEFB0DFDFAFE0DF9FDFCF\
9FE0CF8FDFC08FE0C080DFB07FDFB070D0A070CF9F60BF8F60BF805FAF705FAF6F
509F60509F6050905F50905F4F8050000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F5050505F504F5050505F5F505F505F5F5F505F5F\
5F60605F5F5F605F605F5060605F6F605F606F606F6F5F60705F6F705F607F5F6F
7F5F6080606F8F5F709F6080A05F8FB0609FBF5F9FD060B0E060BFEF60C0EF6FCFEF70D0EF70D0\
F07FDFF07FE0F080EFF07FEFFF80FFFF8FFFFF90FFFF9FFFFFA0FFFFAFFFFFB0FF
FFBFFFFFC0FFFFCFFFF0D0FFE0D0F0DFD0F0CFD0EFC0DFEFB0D0E0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F505F5F4F5F60505F60505F6F5060\
6F505F705F6070505F805F60805F608F5F608F5F6090606F905F6F9F60709F6070
A06F7FA0607FAF6F80AF6F80B06F8FB06F8FBF7090BF6F90C0709FC0709FCF7FA0CF70A0DF7FAF\
DF7FB0E07FB0E07FB0EF80C0EF80C0F080CFF080CFFF8FD0F08FCFFF8FD0F08FD0
FF90DFFF90D0FF9FDFFF9FDFFFA0E0FFA0DFFFA0E0FFA0E0FFAFEFFFAFE0FFB0EFFFB0EFFFBFEF\
FFBFEFFFBFF0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFD0FFFFCFFFFFD0FFF0CFFF
F0D0FFEFD0FFEFD0FFE0D0FFE0DFFFDFD0FFDFDFFFD0DFFFCFE0F0BFDFE0B0E0DFA0DFD09FE0D0\
90DFC08FE0C080DFBF80E0BF70D0A06FC09060BF8060B0805FAF705FA06F50905F
50905F4F8F50000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000505050504F5050505F\
505F5F5F605F5F5F5F605F60605F60605F6F605F606F5F6F6F5F60705F6F705F60
7F606F7F5F6F8F6070905F7FA0608FAF6090C060A0D060AFEF60C0EF60C0EF6FCFEF6FCFF070DF\
F070DFF07FEFF07FEFFF80F0FF80FFFF90FFFF90FFFFA0FFFFA0FFFFB0FFFFB0FF
FFC0FFFFCFFFFFD0FFF0D0FFEFDFFFDFDFF0CFDFF0BFDFEFB0E0EF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000504F5060505F604F506F505F6F505F\
70506070505F7F5F607F5060805F60805F608F5F608F5F6090606F905F6F9F6070
9F607FA06F80A0607FAF6F8FB06F8FBF7090BF6F90C0709FC0709FCF70A0CF70A0D07FAFD070AF\
DF7FB0DF7FB0E07FBFE07FBFEF80C0EF80C0F080CFF080CFFF8FD0F08FCFFF90D0
F090D0FF9FDFF090D0FF9FDFF09FDFFFA0E0F0A0DFFFAFE0FFAFE0FFAFEFF0AFE0FFB0EFFFB0EF\
FFBFF0FFBFEFFFC0F0FFC0F0FFC0FFFFC0F0FFCFFFFFCFFFFFD0FFF0CFFFF0D0FF
EFCFFFEFD0FFE0CFFFE0D0FFDFD0FFDFDFFFD0D0FFD0DFFFC0DFEFB0DFE0A0DFD09FDFD090DFC0\
8FDFC080DFBF7FDFBF70CF9F6FBF8F5FB0705FAF70509F60509F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504F50504F50505F5F50505F5F50605F505F605F606050606F5F606F5060705F60
705F607F5F6F805F6F905F7F9F5F80AF6090BF5F9FD060B0E060BFEF6FC0EF6FC0EF70D0EF70D0\
F07FE0F070E0F07FEFF080F0FF8FFFFF8FFFFF9FFFFF9FFFFFAFFFFFAFFFFFBFFF
FFC0FFFFCFFFF0D0F0EFDFFFE0D0F0DFDFF0CFDFEFC0E0EFB0DFE0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000504F505F505F5F4F5F6F505F6F505F705060\
70505F7F5F607F5F60805F60805F608F606F8F5F6090606F905F6F9F60709F6070
A06F7FA0607FAF6F80AF6F80B06F8FB06F8FC0709FC0709FCF70A0CF70A0D07FAFD07FAFDF7FB0\
DF7FB0E080BFE07FBFEF80C0EF80C0F080CFF080CFFF8FCFF08FCFFF90D0F090D0
FF90D0F090D0FF9FDFFF9FDFFFA0DFF0A0DFFFAFE0FFAFE0FFAFE0FFAFE0FFB0EFFFB0EFFFBFEF\
FFBFEFFFC0F0FFBFF0FFC0F0FFC0F0FFCFFFFFCFFFFFD0FFF0CFFFFFD0FFF0CFFF
F0D0FFEFCFFFEFD0FFE0D0FFE0D0FFDFD0FFDFDFFFD0DFFFC0DFEFAFD0DFA0DFD09FDFCF90DFC0\
80DFBF80DFBF70CF9F6FBF805FAF6F5FA06F50905F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F4F5050505F504F505F50605F5060605F606050606F5F606F5F60
705F6F705F6080606F8F5F709F6080AF5F8FC060A0D060AFEF60C0E060C0EF6FCFEF6FD0F070DF\
F070DFFF7FEFF07FF0FF80FFFF8FFFFF90FFFF9FFFFFA0FFFFAFFFFFB0FFFFBFFF
FFCFFFFFCFFFFFDFFFEFDFF0E0DFF0D0DFF0CFE0F0BFE0EFB0E0EF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F505F5F4F5060505F60505F6F505F6F505F\
70506070505F7F5F607F5060805F60805F608F606F8F5F6F9060709060709F607F
9F607FA06F80AF6080B06F8FB06F8FBF6F90BF6F90C0709FC0709FCF70A0CF70A0D07FAFD070AF\
DF7FB0DF7FB0E080BFE07FBFEF80C0EF80C0F08FCFF080CFF08FCFF08FCFFF90D0
F090D0FF9FD0F09FD0FF9FDFF09FDFFFA0DFF0A0DFFFAFE0F0AFE0FFB0E0F0B0E0FFB0EFF0B0EF\
FFBFEFFFBFEFFFC0F0F0C0F0FFC0F0FFC0F0FFCFFFF0CFFFFFCFFFF0CFFFF0D0FF
EFCFFFEFD0FFE0CFFFE0D0FFDFD0FFDFD0FFD0D0FFD0DFFFB0D0E0AFDFDF9FD0CF90DFCF80D0BF\
80DFBF6FC09F6FB08050A060509F5F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000504F505F505F50505F6050606050606F5F60
605060705F607F5F608F5F70905F7FAF5F90BF5F9FD060B0E05FBFEF60C0EF6FCFEF70D0EF70D0\
F07FE0F07FEFFF80F0FF80FFFF8FFFFF90FFFF9FFFFFA0FFFFAFFFFFB0FFFFC0FF
FFCFFFFFD0FFEFD0F0EFDFFFDFDFF0D0E0F0C0E0EFBFE0EFAFE0E0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F4F5060505F60505F6F505F6F505F705F60\
7050607F5F607F5F60805F6F805F608F606F8F5F6F9060709060709F607F9F607F
A06F80AF6F80AF6F8FB06F8FBF7090BF7090C0709FC0709FCF7FA0CF70A0D07FAFD07FAFDF7FB0\
DF7FB0E080BFE07FBFEF80C0EF80C0F08FCFF080CFFF8FD0F08FCFFF90D0F090D0
FF9FDFFF9FD0FF9FDFF09FDFFFA0E0FFA0DFFFAFE0FFAFE0FFB0EFFFB0E0FFB0EFFFB0EFFFBFF0\
FFBFEFFFC0F0FFC0F0FFCFFFFFC0F0FFCFFFFFCFFFFFD0FFF0CFFFF0D0FFF0CFFF
F0D0FFEFD0FFEFD0FFE0D0FFE0D0FFDFD0FFDFDFFFCFD0F0BFDFE0A0D0D09FDFCF8FD0C08FDFC0\
6FC09060B07F50A060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F4F505F505F5F505F605F60605060
6F5F606F5F607F5F6F805F709F6080AF5F90C060A0D05FAFEF60C0EF60C0EF6FD0EF6FD0F07FE0\
F07FE0FF80F0FF80F0FF8FFFFF8FFFFF9FFFFF9FFFFFAFFFFFB0FFFFBFFFFFC0FF
FFD0FFFFD0FFF0DFFFE0DFF0DFE0F0CFE0F0CFE0F0BFE0EFB0EFE0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F505F5F4F5F60505F60505F6F505F70505F\
7F5F6070505F7F5F60805F608F5F60805F608F606F905F6F9F60709F6070A0607F
A0607FAF6F80AF6F80B06F8FB06F8FBF7090BF6F90C0709FC0709FCF7FA0CF70A0D07FAFD07FAF\
DF80B0DF7FB0E080BFEF80BFF080C0F080C0F08FCFF08FCFFF90D0F08FCFFF90D0
F090D0FF9FDFF09FD0FFA0DFF0A0DFFFA0E0F0A0DFFFAFE0F0AFE0FFB0EFF0B0E0FFBFEFFFBFEF\
FFBFF0F0BFEFFFC0F0FFC0F0FFCFFFFFCFF0FFCFFFFFCFFFFFD0FFF0CFFFF0D0FF
EFCFFFEFD0FFE0CFFFE0D0FFDFD0FFDFD0FFD0D0FFCFD0F0AFD0DFA0D0D08FD0C08FD0C06FC090\
60AF70509F5F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000504F5F50505F605060
60505F6F5F60705060805F709F5F7FB05F90BF5F9FDF60B0E060BFEF6FCFEF6FCFEF70DFEF70DF\
F07FEFF07FEFFF80FFFF80FFFF90FFFF90FFFFA0FFFFAFFFFFBFFFFFBFFFFFCFFF
FFD0FFF0DFFFEFDFF0E0E0F0D0DFF0CFE0F0BFE0EFBFEFEF9FE0D0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F505F5F505F60505F60505F6F5F606F505F705F60\
70505F7F5F607F5F60805F60805F608F606F8F606F9060709060709F6F7FA0607F
A06F80AF6F80B06F8FB06F8FBF7090BF6F90C0709FC0709FCF7FA0CF70A0D07FAFD07FAFDF7FB0\
DF7FB0E080BFE080BFEF80C0EF80C0F08FCFF08FCFFF8FD0F08FCFFF90D0F090D0
FF9FDFF09FD0FFA0DFF0A0DFFFA0E0F0A0DFFFAFE0FFAFE0FFB0EFFFB0E0FFBFEFFFB0EFFFBFF0\
FFBFEFFFC0F0FFC0F0FFCFF0FFCFF0FFD0FFFFCFFFFFD0FFF0CFFFF0D0FFF0D0FF
F0D0FFEFD0FFEFD0FFE0D0FFE0D0FFDFD0FFDFDFFFC0D0EFAFD0DF9FD0CF8FD0C080D0BF5FAF6F\
4F9050000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000504F5F5F505F
60506060505F705F60805F709F5F80AF5F8FC060A0E060BFEF6FC0EF6FCFEF70D0EF70DFF07FE0\
F07FEFFF80F0FF80FFFF8FFFFF90FFFFA0FFFFA0FFFFB0FFFFBFFFFFCFFFFFCFFF
FFDFFFF0DFF0EFE0FFDFE0F0D0E0F0C0E0EFC0EFEFB0EFEF9FE0CF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000504F5060505F5F4F506F505F6F505F705F6070505F\
7F5F607F5060805F60805F608F5F608F5F6090606F905F6F9F60709F6070A06F7F
A0607FAF6F80AF6F80B0708FB06F8FBF7090BF7090C0709FC0709FCF7FA0CF70A0D07FAFD07FAF\
DF7FB0E07FB0EF80BFEF80BFF080C0F080C0F08FCFF08FCFF090D0F090CFFF90D0
F090D0F09FDFF09FD0FFA0DFF0A0DFFFAFE0F0AFDFFFAFE0F0AFE0FFB0EFF0B0E0FFBFEFF0BFEF\
FFBFF0FFBFEFFFC0F0F0C0F0FFCFFFFFCFF0FFD0FFF0CFFFF0D0FFF0CFFFF0D0FF
EFCFFFEFD0FFE0CFFFE0D0FFDFD0FFDFD0FFD0D0FFC0D0EFA0CFD090D0CF80CFBF5FAF70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F505F6050607050608F5F709F507FB05F90CF5FAFE060BFE060C0EF6FCFEF6FD0F070DFF070E0\
F07FEFF07FF0FF80FFFF8FFFFF9FFFFF9FFFFFAFFFFFB0FFFFBFFFFFC0FFFFD0FF
F0D0F0EFDFFFE0DFF0DFE0F0CFE0EFC0EFEFBFEFEFB0F0E08FE0BF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F505F5F505F6F50606F505F705F606F505F7F5F60\
7F5F607F5F607F5F608F606F8F5F608F606F8F606F9F60709F6070A06F7FA0607F
AF6F80AF6F80B0708FB06F8FBF7090BF7090C0709FC0709FCF7FA0CF7FA0D07FAFD07FAFDF80B0\
DF7FB0E080BFE080BFF08FC0F080C0F08FCFF08FCFFF90D0F090D0FF90D0F090D0
FF9FDFF09FDFFFA0DFF0A0DFFFAFE0FFAFE0FFAFE0FFAFE0FFB0EFFFB0EFFFBFEFFFBFEFFFC0F0\
FFBFF0FFC0F0FFC0F0FFCFFFFFCFF0FFD0FFFFD0FFFFD0FFF0D0FFF0D0FFEFD0FF
F0D0FFEFD0FFEFD0FFE0D0FFE0D0FFDFD0FFDFDFFFBFD0E0A0D0D090D0C06FB08F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F506060505F7F5F608F5070A05F80BF5F9FDF60B0E060BFEF6FCFEF6FD0F070DFF070E0FF7FEF\
F07FF0FF80FFFF80FFFF90FFFF9FFFFFAFFFFFAFFFFFBFFFFFC0FFFFD0FFFFD0FF
FFE0FFEFE0F0E0E0FFD0E0F0D0EFF0C0EFEFBFF0EFA0EFDF80E0B0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F505F5F4F5060505F60505F6F505F6F505F705F6070505F\
7F5F607F5F60805F60805F608F606F8F5F6090606F9060709F607F9F6070A06F7F
A06F80AF6F8FAF6F80B0708FB06F90BF709FBF7090C0709FC070A0D07FAFCF7FA0D07FAFDF7FB0\
E080BFE07FB0EF80BFEF80C0F08FCFF080C0F08FCFF08FCFF090D0F090CFF09FD0
F09FD0FF9FDFF09FD0FFA0DFF0A0DFFFAFE0F0AFDFFFB0E0F0B0E0FFB0EFF0B0E0FFBFEFFFBFEF\
FFC0F0F0C0EFFFC0F0FFC0F0FFCFFFFFCFF0FFD0FFF0D0FFFFD0FFF0D0FFF0D0FF
EFD0FFEFD0FFE0CFFFE0D0FFDFD0FFDFD0FFD0D0FFB0D0E09FCFCF8FD0C0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000060505F7050608F5F70AF5F8FCF5FA0E05FB0E060C0E060CFEF6FD0EF6FDFF070E0F070EF\
FF7FFFFF80FFFF8FFFFF90FFFFA0FFFFAFFFFFB0FFFFBFFFFFCFFFFFCFFFFFDFFF
F0DFF0EFE0FFDFE0F0D0EFF0CFEFEFC0F0EFB0F0E09FEFCF70DFA0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F4F5060505F60505F6F505F6F505F705F607050607F5F60\
7F5F60805F60805F608F606F8F5F6090606F90606F9F60709F6070A06F7FA06F7F
AF6F80AF6F80B0708FB06F8FBF7090BF7090C07F9FC0709FCF7FA0CF7FA0D07FAFD07FAFDF80B0\
DF7FB0EF80BFEF80BFEF8FC0F080C0F08FCFF08FCFF090D0F090CFFF9FD0F09FD0
FF9FDFF09FD0FFA0DFF0A0DFFFAFE0F0AFDFFFB0E0FFAFE0FFB0EFFFB0E0FFBFEFFFBFEFFFC0F0\
FFC0EFFFCFF0FFC0F0FFCFFFFFCFF0FFD0FFFFD0FFFFD0FFF0D0FFF0D0FFF0D0FF
F0D0FFEFD0FFEFD0FFE0D0FFE0D0FFDFD0FFDFDFFFB0CFDF9FD0CF80CFBF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F4F5F7F5F60905F7FBF5F90D05FAFE060C0E060C0EF6FD0EF6FD0F070E0F070EFFF7FF0\
FF7FFFFF8FFFFF90FFFFA0FFFFA0FFFFB0FFFFBFFFFFC0FFFFCFFFFFDFFFF0DFFF
F0E0FFE0E0F0DFEFF0D0EFF0CFF0F0BFF0EFB0FFEF90EFBF6FE09F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F505F5F505F60505F60505F6F505F6F505F705F6070505F\
7F5F607F5F60805F60805F608F606F8F5F6F9060709060709F607F9F607FA06F80
A06F80AF6F8FAF6F8FB07090B06F90BF709FC0709FCF7FA0C070A0D07FAFD07FAFDF7FB0DF7FB0\
E080BFE080BFEF80C0EF80C0F08FCFF08FCFF08FCFF08FCFF090D0F090D0F09FD0
F09FD0F0A0DFF0A0DFFFA0DFF0A0DFFFAFE0F0AFE0FFB0E0F0B0E0FFB0EFF0B0EFFFBFEFF0BFEF\
FFC0F0FFC0F0FFCFF0F0C0F0FFCFFFFFCFFFFFD0FFF0CFFFF0D0FFF0D0FFF0D0FF
EFD0FFEFD0FFE0D0FFE0D0FFDFD0FFDFD0FFD0D0FFAFD0DF8FCFC0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000050505F80506FA05F80BF5F9FDF5FB0E060BFEF6FCFEF6FCFF070DFF070E0FF7FF0F07FF0\
FF80FFFF8FFFFF9FFFFF9FFFFFAFFFFFB0FFFFC0FFFFC0FFFFD0FFFFDFFFF0E0FF
E0E0F0E0EFF0D0EFEFCFF0F0C0EFEFBFF0EFA0F0DF80EFB05FDF80000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005F505F5F505F60505F60505F6F5F606F505F705F60705F607F5F60\
7F5F60805F6F805F608F606F8F606F9060709060709F6F7F9F607FA06F80A06F80
AF708FAF6F8FB07090B07090BF709FBF709FC07FA0C070A0D07FAFD07FAFD080B0D07FB0E080BF\
E080BFEF80C0EF80C0F08FCFF08FCFF08FCFF08FCFF090D0F090D0F09FD0F09FD0
FFA0DFF0A0DFFFA0DFF0A0DFFFAFE0F0AFE0FFB0E0F0B0E0FFBFEFFFB0EFFFBFEFFFBFEFFFC0F0\
FFC0F0FFCFF0FFCFF0FFD0FFFFCFFFFFD0FFFFD0FFFFDFFFF0D0FFF0D0FFF0D0FF
F0DFFFEFD0FFEFD0FFE0D0FFE0DFFFDFD0FFDFDFFFA0CFD08FCFC0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000504F5F905F7FAF5F8FCF60AFE060BFEF6FCFEF6FCFF070DFF070E0FF7FEFFF7FF0FF80FF\
FF8FFFFF9FFFFF9FFFFFAFFFFFB0FFFFBFFFFFC0FFFFD0FFFFDFFFFFE0FFEFE0F0
EFEFFFDFEFF0D0F0F0CFF0EFC0FFF0B0FFEFA0F0D07FE0A05FE080000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000504F505F505F5F505F60505F6F505F705F606F505F705F607F5F60\
7F5F607F5F60805F60805F608F606F8F5F6F9060709060709F6F7F9F607FA06F80
AF6F80B0708FB06F8FBF7090BF7090C0709FC0709FCF7FA0CF7FA0D07FAFD07FAFDF80B0DF7FB0\
E080BFE080BFEF80C0EF80C0F08FCFEF8FCFF090CFF090CFF090D0F090D0F09FD0
F09FD0F0A0DFF0A0DFF0AFDFF0AFDFFFAFE0F0AFE0FFB0E0F0B0E0FFBFEFF0BFEFFFBFEFFFBFEF\
FFC0F0F0C0F0FFCFF0FFCFF0FFD0FFFFD0FFFFD0FFF0D0FFF0D0FFF0D0FFF0D0FF
EFD0FFEFDFFFE0D0FFE0D0FFDFD0FFE0D0FFD0D0FFA0CFD080C0BF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000007F50609F507FBF5F90E05FB0E060C0E060CFEF6FD0EF6FDFF070EFF070EFFF7FFFFF80FF\
FF90FFFF90FFFFA0FFFFAFFFFFBFFFFFC0FFFFCFFFFFD0FFFFE0FFEFE0F0EFEFF0
DFEFF0DFEFF0CFEFEFC0F0EFBFF0EFB0FFEF90EFC070EF9F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005F505F5F505F60505F60505F6F5F606F505F705F60705F607F5F60\
7F5F6080606F805F608F606F8F606F9060709060709F6F7F9F607FA06F80A06F80
AF708FAF6F8FB07090B07090BF709FBF709FCF7FA0CF7FA0D07FAFD07FAFDF80B0DF7FB0E080BF\
E080BFEF8FC0EF80C0F08FCFF08FC0F090CFF090CFF090D0F090D0F09FD0F09FD0
F0A0DFF0A0D0FFAFDFF0A0DFFFAFE0F0AFDFFFB0E0F0B0E0FFBFEFF0BFE0FFC0EFFFBFEFFFC0F0\
FFC0EFFFCFF0FFCFF0FFD0FFFFCFF0FFD0FFFFD0FFFFD0FFF0D0FFF0DFFFF0D0FF
F0DFFFEFD0FFEFDFFFE0D0FFE0D0FFDFD0FFDFDFFF9FCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000008F506FAF5F8FD05FAFE060C0E060C0EF6FD0EF6FDFF070E0F070EFFF7FFFFF80FFFF8FFF\
FF90FFFFA0FFFFAFFFFFB0FFFFBFFFFFCFFFFFD0FFFFE0FFF0E0FFEFEFFFE0EFF0
DFEFF0CFEFEFCFF0F0BFF0EFBFFFEFA0FFDF8FF0BF60E08F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005F4F5060505F60505F6F505F6F505F705F6070505F7F5F607F5F60\
7F5F60805F6080606F805F608F606F90606F9F60709060709F6F7FA06F7FAF6F80
AF6F80B0708FB06F8FBF7090BF7090C0709FC0709FCF7FA0CF7FA0D07FAFD07FAFDF80B0DF7FB0\
E080BFE080BFEF8FC0EF80C0F08FCFEF8FCFF090CFEF90CFF09FD0F09FD0F09FD0
F09FD0F0A0DFF0A0DFF0AFDFF0AFDFF0AFE0F0AFE0FFB0E0F0B0E0FFBFEFF0BFEFFFC0EFF0BFEF\
FFC0F0F0C0F0FFCFF0F0CFF0FFD0FFFFD0FFFFD0FFF0D0FFF0D0FFF0D0FFF0DFFF
EFD0FFEFDFFFE0D0FFE0DFFFDFD0FFDFD0FFDFD0FF9FCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
704F5F90507FC05F9FE060BFE060C0EF6FD0EF6FD0F070E0F070E0FF7FF0FF7FFFFF8FFFFF8FFF\
FF9FFFFFA0FFFFB0FFFFBFFFFFC0FFFFCFFFFFDFFFF0DFF0EFE0FFE0E0F0DFEFF0
D0EFEFCFF0EFC0F0EFBFFFEFAFFFE09FFFCF7FEFAF5FE080000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005F505F60505F60505F6F50606F505F705F60705F607F5F60705F6080606F\
7F5F6080606F805F608F606F8F606F9060709060709F6F7FA06F7FA06F80A06F80
B0708FB0708FBF7090BF7090C07F9FC0709FCF7FA0CF7FA0D080AFD07FAFDF80B0DF80B0E080BF\
E080BFEF8FC0EF80C0F08FCFF08FCFF090CFF090CFF09FD0F09FD0F09FD0F09FD0
FFA0DFF0A0DFFFAFDFF0AFDFFFB0E0F0AFE0FFB0E0F0B0E0FFBFEFFFBFEFFFC0EFFFC0EFFFCFF0\
FFC0F0FFCFF0FFCFF0FFD0FFFFD0FFFFDFFFFFD0FFFFDFFFF0D0FFF0DFFFEFD0FF
F0DFFFEFDFFFEFDFFFE0D0FFE0DFFFDFD0FFDFDFFF90C0C0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
80506FBF5F90E060BFE060C0EF6FD0EF6FD0F070E0F070E0FF7FF0FF7FFFFF80FFFF8FFFFF9FFF\
FFA0FFFFB0FFFFB0FFFFC0FFFFCFFFFFDFFFFFE0FFF0EFFFEFEFF0E0EFF0D0EFF0
D0F0F0C0F0EFC0FFEFB0FFE0AFFFDF90F0BF7FF0A050E070000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005F505F5F505F60505F60505F6F505F6F505F705F6070505F7F5F607F5F60\
805F6F805F608F606F8F5F6F9060709060709F607F9F607FA06F80A06F80AF6F80
AF6F8FB07090B06F8FBF7090BF709FC07F9FC0709FCF7FA0CF7FA0D080AFD07FAFDF80B0DF80B0\
E080BFE080BFEF8FC0EF8FC0F08FCFEF8FCFF090CFF090CFF09FD0F09FD0F0A0D0
F0A0D0F0A0DFF0A0DFF0AFDFF0AFDFFFB0E0F0B0E0FFB0E0F0B0E0FFBFEFF0BFEFFFC0EFF0C0EF\
FFCFF0F0CFF0FFCFF0FFCFF0FFD0FFFFD0FFFFDFFFF0D0FFF0DFFFEFD0FFF0DFFF
EFD0FFEFDFFFE0D0FFE0DFFFE0D0FFE0DFFFDFD0FFDFD0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
A05080DF5FB0E05FBFE060CFEF60CFF06FDFEF6FE0F070EFF070F0FF80FFFF80FFFF90FFFF9FFF\
FFAFFFFFB0FFFFBFFFFFC0FFFFD0FFFFDFFFF0E0FFEFE0F0E0EFF0DFEFEFD0F0F0
CFF0EFC0FFEFB0F0E0B0FFE09FF0CF80F0B06FEF90000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005F505F60505F60505F6F5F606F505F705F6070505F7F5F607F5F60805F60\
7F5F608F606F80606F8F606F8F606F9F6F709F6070A06F7FA06F7FAF6F80AF6F80
B0708FB06F8FBF7090BF7090C07F9FC0709FCF7FA0CF7FA0D07FAFD07FAFDF80B0DF80B0E080BF\
E080BFEF8FC0EF8FC0F08FCFEF8FCFF090CFF090CFF09FD0F09FD0F0A0D0F09FD0
F0A0DFF0A0DFFFAFE0F0AFDFFFB0E0F0B0E0FFBFEFF0B0E0FFBFEFF0BFEFFFC0F0FFC0EFFFCFF0\
FFC0F0FFCFFFFFCFF0FFD0FFFFD0FFFFDFFFF0D0FFFFDFFFF0D0FFF0DFFFEFDFFF
EFDFFFEFDFFFEFDFFFE0DFFFE0DFFFDFD0FFDFDFFFD0D0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000905070
CF5FA0E05FBFEF60CFEF60CFEF6FDFEF6FDFFF7FEFFF7FF0FF80FFFF80FFFF90FFFF9FFFFFAFFF\
FFAFFFFFBFFFFFC0FFFFD0FFFFDFFFFFE0FFF0E0FFEFEFFFDFEFF0DFF0F0CFF0EF
CFF0EFBFF0E0BFFFEFA0FFD090FFBF70F0A060F08F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000504F505F505F5F505F60505F60505F6F5F606F505F705F60705F607F5F607F5F60\
805F60805F608F606F8F5F6F9060709060709F6F7F9F607FA06F80A06F80AF708F
AF6F8FB07090B07090BF709FBF7090C07F9FC0709FCF7FA0CF7FA0D07FAFD07FAFDF80B0DF80B0\
E080BFE080BFEF8FC0EF8FC0F08FCFEF8FCFF090CFEF90CFF09FD0F09FD0F0A0DF
F0A0D0F0A0DFF0A0DFFFAFE0F0AFDFF0B0E0F0B0E0FFBFEFF0B0E0FFBFEFF0BFEFFFC0F0F0C0EF\
FFCFF0FFCFF0FFD0FFF0CFF0FFD0FFF0D0FFFFDFFFF0D0FFF0DFFFEFD0FFF0DFFF
EFD0FFEFDFFFE0D0FFE0DFFFDFD0FFE0DFFFDFD0FFDFDFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000805060BF5F90
DF5FB0E060C0E060CFEF6FDFEF6FDFF070EFF070F0FF7FFFFF80FFFF90FFFF9FFFFFA0FFFFAFFF\
FFBFFFFFC0FFFFCFFFFFD0FFFFE0FFF0E0F0EFEFFFE0E0F0DFEFF0D0EFEFCFF0EF
BFF0E0BFFFE0AFFFDF9FFFC07FF0AF6FF09050EF7F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005F505F5F505F605F6060505F6F5F606F505F705F60705F607F5F607F5F6080606F\
805F608F606F8F606F9060709060709F6F7F9F607FA06F80A06F80AF7080AF6F80
B07090B0708FBF7090BF7090C07F9FC07F9FCF7FA0CF7FA0D080AFD07FAFDF80B0DF80B0E08FBF\
E080BFEF8FC0EF8FC0F090CFEF8FCFF090CFF090CFF09FD0F09FD0F0A0D0F0A0D0
FFAFDFF0A0DFFFAFE0F0AFDFFFB0E0F0B0E0FFBFEFF0BFE0FFC0EFFFBFEFFFC0F0FFC0EFFFCFF0\
FFCFF0FFD0FFFFCFF0FFD0FFFFD0FFFFDFFFF0D0FFFFDFFFF0DFFFF0DFFFEFDFFF
F0DFFFEFDFFFEFDFFFE0DFFFE0DFFFE0DFFFDFDFFFD0DFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000070505FAF5F8FDF5FB0
E060C0E060CFEF6FDFEF6FDFFF70EFFF70F0FF7FFFFF80FFFF90FFFF9FFFFFA0FFFFAFFFFFBFFF\
FFC0FFFFCFFFFFD0FFFFE0FFF0E0FFF0EFFFE0EFF0E0F0F0D0EFEFD0F0EFC0F0EF
C0FFEFB0FFE0A0FFD08FF0B07FF0A060EF80000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000504F505F505F60505F6F505F60505F6F5F606F505F705F60705F607F5F607F5F60\
80606F805F608F606F8F606F9060709060709F6F7F9F6F7FA06F80A06F80AF708F
AF6F8FB07090B07090BF709FBF709FC07FA0C07FA0CF7FAFCF7FAFD080B0D07FAFDF80B0DF80B0\
E08FBFE080BFEF8FC0EF8FC0F090CFEF90CFF090CFEF90CFF09FD0EF9FD0F0A0D0
F0A0D0F0AFDFF0AFDFF0AFE0F0AFDFFFB0E0F0B0E0FFBFEFF0BFE0FFC0EFF0C0EFFFC0F0F0C0EF\
FFCFF0F0CFF0FFD0FFFFD0F0FFDFFFF0D0FFFFDFFFF0DFFFF0DFFFEFDFFFEFDFFF
EFDFFFEFDFFFE0DFFFE0DFFFE0DFFFE0DFFFDFDFFFDFDFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000009F507FDF5FB0E060C0
E060CFEF6FD0EF6FDFF070EFF070EFFF7FFFFF80FFFF8FFFFF90FFFFA0FFFFAFFFFFB0FFFFBFFF\
FFCFFFFFD0FFFFDFFFF0E0FFF0EFFFE0E0F0E0EFF0D0EFEFD0F0EFC0F0E0C0F0EF
B0F0E0AFFFDF90F0BF80F0AF6FEF8F50F07F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005F505F5F505F60505F60505F6F5F606F5F60705F60705F607F5F607F5F6080606F\
805F608F606F8F606F9060709060709F6F7F9F6F7FA06F80A06F80AF708FAF6F8F
B0708FB0708FBF7F9FBF7090C07F9FC07F9FCF7FA0CF7FA0D080AFD07FAFDF80B0DF80B0E08FBF\
E080BFEF8FC0EF8FC0F090CFEF8FCFF090CFEF90CFF09FD0F09FD0F0A0DFF0A0D0
F0AFDFF0A0DFFFAFE0F0AFDFFFB0E0F0B0E0FFBFEFF0BFE0FFC0EFF0BFEFFFC0F0FFC0EFFFCFF0\
FFCFF0FFD0FFFFD0F0FFDFFFFFD0FFFFDFFFF0DFFFF0DFFFF0DFFFF0E0FFEFDFFF
EFDFFFE0DFFFEFE0FFE0DFFFE0DFFFDFDFFFE0E0FFDFDFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000008F506FD05FAFE060C0E060CF
EF6FD0EF6FDFF070EFF070F0FF7FFFFF80FFFF90FFFF90FFFFA0FFFFAFFFFFBFFFFFBFFFFFCFFF\
FFD0FFFFE0FFFFE0FFF0EFFFEFE0F0E0EFF0DFEFEFD0F0EFCFF0EFC0FFEFBFF0E0
B0FFE09FF0CF8FFFB070F09F60F080000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F505F5F505F60505F60505F6F505F6F505F705F606F505F705F60705F607F5F607F5F60\
80606F805F608F606F8F606F9060709060709F6F7F9F6F7FA06F80A06F80AF708F
AF6F8FB07090B07090BF7F9FBF709FC07FA0C07FA0CF7FAFCF7FAFD080AFD080B0DF80BFDF80B0\
E08FBFE080BFEF8FC0EF8FC0EF90CFEF90CFF09FCFEF90CFF09FD0EF9FD0F0A0DF
EFA0D0F0AFDFF0AFDFF0B0E0F0AFDFF0B0E0F0B0E0FFBFEFF0BFE0FFC0EFF0C0EFFFCFF0F0C0EF\
FFCFF0FFCFF0FFD0FFF0D0F0FFDFFFF0D0FFFFDFFFF0DFFFF0DFFFEFDFFFEFDFFF
E0DFFFEFDFFFE0DFFFE0DFFFDFDFFFE0DFFFDFDFFFDFDFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000007F5060CF50A0E060C0E060C0EF6FD0
EF6FDFF070E0F070EFFF7FFFFF7FFFFF8FFFFF90FFFFA0FFFFA0FFFFB0FFFFBFFFFFCFFFFFCFFF\
FFDFFFF0E0FFF0E0FFE0E0F0E0EFF0DFEFEFD0F0EFCFEFE0C0F0EFBFF0E0B0FFE0
A0F0D090FFBF7FF0A06FF08F50EF70000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F505F60505F60505F6F5F606F505F6F5F606F5F60705F60705F607F60607F5F6080606F\
80606F8F606F8F606F906F709060709F6F7F9F6F7FA07080A06F80AF708FAF708F
B07090B07090BF7F9FBF709FC07FA0C07FA0CF80A0CF7FA0D080B0D080AFDF80B0DF80B0E08FBF\
E08FBFEF8FC0EF8FC0F090CFEF90CFF09FCFEF90CFF09FD0F09FD0F0A0D0F0A0D0
F0AFDFF0AFDFF0B0E0F0AFDFFFB0E0F0B0E0FFBFEFF0BFE0FFC0EFFFC0EFFFCFF0FFC0EFFFCFF0\
FFCFF0FFD0FFFFD0F0FFDFFFFFDFFFFFDFFFF0DFFFF0E0FFF0DFFFF0E0FFEFDFFF
EFE0FFEFDFFFEFE0FFE0DFFFE0E0FFDFDFFFE0E0FFDFDFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000CF5FA0E060C0E060CFEF6FD0EF6FDF
F070EFF070EFFF7FFFFF7FFFFF8FFFFF90FFFFA0FFFFA0FFFFB0FFFFBFFFFFCFFFFFCFFFFFDFFF\
FFE0FFF0EFFFEFE0F0EFEFF0DFEFF0DFF0F0CFF0EFCFFFEFBFF0E0BFFFE0AFFFDF
A0FFCF8FFFB070FF9F5FF07F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F505F5F505F60505F60505F6F5F606F505F705F60705F607F5F607F5F60805F60805F60\
80606F805F6F8F606F8F6070906F7F9060709F6F7F9F6F7FA07080A06F80AF708F
AF708FB07090B07090BF7F9FC07F9FCF7FA0C07FA0CF80AFD07FAFDF80B0D080B0DF80BFE080BF\
EF8FC0E08FC0EF8FC0EF8FC0F090CFEF90CFF09FCFEF9FCFF0A0D0EF9FD0F0A0DF
F0A0D0F0AFDFF0AFDFF0B0E0F0B0DFF0BFE0F0B0E0FFBFEFF0BFEFFFC0EFF0C0EFFFCFF0FFCFF0\
FFD0F0F0CFF0FFD0FFFFD0FFFFDFFFF0D0FFF0DFFFF0DFFFF0DFFFEFDFFFEFE0FF
EFDFFFEFE0FFE0DFFFE0E0FFDFDFFFE0E0FFDFDFFFDFE0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000C0509FE060BFE060C0EF6FD0EF6FD0F070E0
F070EFFF7FFFFF7FFFFF8FFFFF90FFFF9FFFFFA0FFFFB0FFFFB0FFFFC0FFFFCFFFFFDFFFFFDFFF\
F0E0FFEFE0F0E0EFF0DFEFEFDFF0F0CFEFEFCFF0EFBFF0E0BFFFE0AFFFDFA0FFCF
90FFBF7FFFA060F080000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F505F60505F60505F6F5F606F505F705F60705F607F5F60705F607F606F7F5F6080606F\
8060608F606F8F606F906F709060709F6F7F9F6F7FA06F80A06F80AF708FAF708F
B07090B07090BF7F9FBF7F9FC07FA0C07FA0CF80AFCF7FAFD080B0D080AFDF8FBFDF80BFE08FBF\
E08FBFEF8FC0EF8FC0EF90CFEF90CFF09FD0EF9FCFF09FD0EF9FD0F0A0DFF0A0D0
F0AFDFF0AFDFF0B0E0F0B0DFF0B0E0F0B0E0FFBFEFF0BFEFFFC0EFF0C0EFFFCFF0FFCFF0FFD0F0\
FFCFF0FFD0FFFFD0FFFFDFFFFFDFFFFFDFFFF0DFFFF0DFFFEFDFFFF0E0FFEFDFFF
EFE0FFE0E0FFEFE0FFE0E0FFE0E0FFDFDFFFDFE0FFDFE0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000C0509FE060C0E060C0EF6FD0EF6FDFF070E0F070EF
FF7FF0FF7FFFFF8FFFFF8FFFFF9FFFFFA0FFFFB0FFFFB0FFFFC0FFFFCFFFFFDFFFFFDFFFFFE0FF\
EFE0F0EFEFFFDFEFF0DFF0F0CFF0EFCFF0EFC0F0E0BFFFE0B0FFDFAFFFD090FFBF
8FFFAF70F09060FF80000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
504F505F505F5F505F60505F60505F6F5F606F505F705F60705F5F7F5F607F5F6080606F805F60\
80606F8F606F9060708F6070906F7F9F607FA06F809F6F7FA06F80AF6F80B0708F
AF708FB07090BF7090C07F9FC0709FCF7FA0CF7FA0D080AFD07FAFDF80B0DF80B0E08FBFE080BF\
EF8FC0EF8FC0EF90C0EF8FC0EF90CFEF90CFF09FD0EF9FCFF0A0D0EFA0D0F0AFDF
EFA0D0F0AFDFF0AFDFF0B0E0F0B0DFF0BFE0F0BFE0FFC0EFF0BFE0FFC0EFF0C0EFFFCFF0F0CFEF\
FFD0F0FFD0F0FFDFFFFFD0FFFFDFFFF0DFFFF0DFFFEFDFFFF0DFFFEFDFFFEFE0FF
E0DFFFE0E0FFE0E0FFE0E0FFDFDFFFDFE0FFDFE0FFDFE0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000BF5090E05FBFE05FC0EF6FD0EF60D0F06FE0F070EFFF7FFF
FF7FFFFF80FFFF8FFFFF9FFFFF9FFFFFAFFFFFB0FFFFC0FFFFC0FFFFD0FFFFDFFFFFE0FFEFE0F0\
EFEFF0DFEFEFDFF0EFCFEFE0CFF0EFC0F0E0BFFFE0AFF0DFAFFFD090FFBF8FFFB0
70F09F60FF8F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F505F5F505F60505F60505F6F5F606F5F60705F60705F607F60607F5F6080606F7F5F608F606F\
80606F8F60708F6070906F7F906F7F9F6F7F9F6F7FA07080A06F80AF708FAF708F
BF7F90BF7090BF7F9FBF7F9FCF7FA0CF7FA0CF80AFCF7FAFDF80B0DF80B0DF8FBFDF80BFE08FC0\
E08FC0EF8FC0EF8FC0F090CFEF90CFF09FD0EF9FCFF0A0D0F0A0D0F0AFDFF0A0D0
F0AFDFF0AFDFF0B0E0F0B0DFFFBFE0F0BFE0FFC0EFF0BFE0FFC0EFF0C0EFFFCFF0F0CFF0FFD0F0\
FFD0F0FFDFFFFFD0FFFFDFFFFFDFFFFFDFFFF0DFFFF0E0FFEFDFFFF0E0FFEFE0FF
EFE0FFE0E0FFE0E0FFE0E0FFE0E0FFDFE0FFDFE0FFDFE0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000BF5090E060C0E060C0EF6FD0EF6FDFF070EFF070EFFF7FFFFF7FFF
FF8FFFFF90FFFF9FFFFFA0FFFFB0FFFFB0FFFFC0FFFFCFFFFFD0FFFFDFFFFFE0FFF0E0F0EFEFFF\
E0EFF0DFF0F0D0EFEFCFF0EFC0F0E0BFFFE0B0FFDFAFFFDF9FFFC08FFFB07FFF9F
70FF905FFF7F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F505F60505F5F505F60505F60505F6F5F606F505F705F60705F607F5F607F5F6080606F805F60\
8F606F8F606F9060709060709F6F7F9F6F7FA06F80A06F80AF708FAF6F8FB07090
B07090BF7F9FBF7090C07F9FC07F9FCF7FA0CF7FA0D080AFD080AFDF80B0DF80B0E08FBFE080BF\
EF8FC0E08FC0EF90C0EF90C0EF9FCFEF90CFF09FD0EF9FCFF0A0D0EFA0D0F0AFDF
F0AFDFF0B0DFF0AFDFF0B0E0F0B0E0F0BFE0F0BFE0F0C0EFF0C0EFFFCFEFF0C0EFFFCFF0F0CFF0\
FFD0F0F0D0F0FFDFFFFFDFFFFFDFFFF0DFFFF0DFFFEFDFFFF0E0FFEFDFFFEFE0FF
E0E0FFEFE0FFE0E0FFE0E0FFDFE0FFDFE0FFD0E0FFDFE0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000B0508FE060C0E060C0EF6FD0EF6FDFF070E0F070EFFF7FFFFF7FFFFF8FFF
FF90FFFF9FFFFFA0FFFFAFFFFFB0FFFFC0FFFFC0FFFFD0FFFFDFFFFFE0FFEFE0F0EFEFF0E0EFEF\
DFEFEFD0EFE0CFF0E0C0F0E0BFFFE0B0F0DFAFFFDF9FFFC090FFB07FFFA070FF90
5FFF7F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F505F5F505F605F6060505F6F5F606F505F705F60705F607F5F607F5F6080606F80606080606F\
80606F906F70906070906F7F906F7FA06F80A06F80A07080A06F80B07090B0708F
B07F90B07090C07F9FC07F9FC07FA0C07FA0D080AFD080AFD080B0D080B0E08FBFE080BFE08FC0\
E08FBFEF90C0EF90C0EF90CFEF90CFF09FD0EF9FCFF0A0D0EFA0D0F0AFDFF0AFD0
F0AFDFF0AFDFF0B0E0F0B0E0F0BFE0F0BFE0FFC0EFF0C0EFFFC0EFF0C0EFFFCFF0FFCFF0FFD0F0\
FFD0F0FFDFFFFFDFFFFFDFFFF0DFFFFFE0FFF0DFFFF0E0FFEFDFFFF0E0FFEFE0FF
EFE0FFE0E0FFE0EFFFE0E0FFE0EFFFDFE0FFDFEFFFD0E0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
DF5FB0E060C0E060CFEF6FD0EF6FDFF070EFF070EFFF7FFFFF7FFFFF8FFFFF90FF
FF9FFFFFA0FFFFB0FFFFBFFFFFC0FFFFCFFFFFD0FFFFDFFFFFE0FFF0E0F0EFEFF0E0EFEFE0EFEF\
D0EFEFD0F0EFC0F0E0C0FFE0B0F0DFB0FFDFA0FFCF90FFBF80FFA070FF9060FF80
5FFF70000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F505F
5F505060505F60505F6F5F6060505F6F5F606F5F5F705F60705F607F5F607F5F6080606F805F60\
8F606F8F606F906F709060709F6F7F9F6F7FA06F80A06F80AF708FAF708FB07090
B07090BF7F9FBF709FC07FA0C07F9FCF7FA0CF7FA0D080AFD080AFDF80B0DF80B0E08FBFE080BF\
EF8FC0E08FC0EF90CFEF90C0EF9FCFEF9FCFF09FD0EF9FCFF0A0D0EFA0D0F0AFDF
EFAFD0F0B0DFF0B0DFF0B0E0F0B0DFF0BFE0F0BFE0FFC0EFF0C0EFFFCFEFF0CFEFFFCFF0F0CFF0\
FFD0F0FFD0F0FFDFFFF0DFFFFFDFFFF0DFFFF0E0FFEFDFFFEFE0FFEFDFFFEFE0FF
E0E0FFE0E0FFE0E0FFE0EFFFDFE0FFDFE0FFD0E0FFDFEFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000DF5FB0\
E060C0E060CFEF6FDFEF6FDFF070EFF070EFFF7FFFFF7FFFFF8FFFFF90FFFF9FFF
FFA0FFFFB0FFFFB0FFFFC0FFFFC0FFFFD0FFFFDFFFFFE0FFEFE0F0EFEFF0E0EFEFDFEFEFD0EFE0\
D0F0E0C0F0E0C0F0E0B0F0DFB0FFDFA0FFCF90FFBF80F0A070FF9060FF805FFF70
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F505F
605F5F60505F605F6060505F705F606F5F60705F60705F607F606F7F5F6080606F8060608F606F\
8F606F906F709060709F6F7F9F6F7FA06F80A06F80AF708FAF708FB07090B07090
BF7F9FBF7F9FC07F9FC07F9FCF80A0CF7FA0D080AFD080AFDF8FB0DF80B0E08FBFE08FBFEF8FC0\
EF8FC0EF90C0EF90C0EF9FCFEF9FCFF09FD0EF9FCFF0A0D0F0A0D0F0AFDFF0AFDF
F0B0DFF0B0DFF0B0E0F0B0E0FFBFE0F0BFE0FFC0EFF0C0EFFFCFF0F0CFEFFFCFF0FFCFF0FFD0FF\
FFD0F0FFDFFFFFDFFFFFE0FFF0DFFFF0E0FFF0DFFFF0E0FFEFE0FFEFE0FFEFE0FF
EFEFFFE0E0FFE0EFFFDFE0FFE0EFFFDFEFFFDFEFFFD0E0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000DF5FBFE060CF\
E060CFEF6FDFEF6FDFF070EFF070F0FF7FFFFF80FFFF8FFFFF90FFFFA0FFFFA0FF
FFB0FFFFBFFFFFC0FFFFCFFFFFDFFFFFDFFFFFE0FFF0E0F0EFEFF0E0EFF0E0EFEFD0EFEFD0F0EF\
CFF0E0C0FFE0BFF0DFB0FFDFAFFFCF9FFFC08FFFAF7FFF9F6FFF8F60FF7F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F505F
5F505F60505F60505F6F5F606F505F705F60705F60705F60705F607F606F7F5F6080606F80606F\
8F60708F606F906F709060709F6F7F9F6F7FA07080A06F80AF708FAF708FB07090
B07090BF7F9FBF709FC07FA0C07FA0CF80AFCF7FA0D080AFD080AFDF8FB0DF80B0E08FBFE08FBF\
EF8FC0E08FC0EF90C0EF90C0EF9FCFEF9FCFEFA0D0EFA0D0F0A0D0EFA0D0F0AFDF
F0AFDFF0B0DFF0B0DFF0BFE0F0BFE0F0BFEFF0BFE0F0C0EFF0C0EFFFCFF0F0CFEFFFD0F0F0D0F0\
FFDFFFFFD0F0FFDFFFF0DFFFFFE0FFF0DFF0F0E0FFEFDFFFEFE0FFEFE0FFEFE0FF
E0E0FFE0EFFFDFE0FFE0EFFFDFE0FFDFEFFFD0E0FFD0EFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000DF5FBFE060C0E060CF\
EF6FDFEF6FDFF070EFF070F0FF7FFFFF7FFFFF8FFFFF90FFFFA0FFFFA0FFFFB0FF
FFB0FFFFC0FFFFCFFFFFD0FFFFDFFFFFE0FFEFE0F0EFEFF0E0EFEFDFEFEFD0EFE0D0F0E0C0F0E0\
C0F0E0B0F0DFB0FFDFAFFFD09FFFC08FFFAF7FFF9F6FFF8F60FF7F50D05F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F505F
60505F60505F6F5F606F505F6F5F606F5F60705F60705F607F606F7F5F6080606F80606F8F606F\
8F606F906F70906F709F6F7F9F6F7FA07080A06F80AF708FAF708FB07F90B07090
BF7F90BF7F90C07F9FC07F9FCF80A0CF7FA0D080AFD080AFDF80B0DF80B0E08FBFE08FBFE08FC0\
E08FC0EF90C0EF90C0EF9FCFEF9FCFF0A0D0EF9FCFF0A0D0EFA0D0F0AFDFF0AFDF
F0B0DFF0B0DFF0BFE0F0BFE0F0BFE0F0BFE0FFC0EFF0C0EFFFCFEFF0CFEFFFD0F0FFD0F0FFD0F0\
FFD0F0FFDFFFFFDFFFFFE0FFF0DFFFF0E0FFF0E0FFF0E0FFEFE0FFEFE0FFE0E0FF
EFEFFFE0E0FFE0EFFFDFEFFFDFEFFFDFEFFFDFEFFFD0EFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000BF5F90DF5FBFE060CFE060CFEF6FDF\
EF6FE0F070EFF070F0FF7FFFFF80FFFF90FFFF90FFFFA0FFFFA0FFFFB0FFFFBFFF
FFC0FFFFCFFFFFD0FFFFDFFFFFE0FFF0E0F0EFEFF0E0EFEFDFEFEFD0EFE0D0F0EFCFF0E0C0F0E0\
BFF0DFB0FFDFAFFFD0A0FFC08FFFB080FFA070FF8F60FF8050F06F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F505F
5F505F60505F60505F6F5F606F505F705F60705F607F5F60705F607F60607F5F6080606F80606F\
8F60708F6070906F709060709F6F7F9F6F7FA07080A06F80AF708FAF708FB07F90
B07090BF7F9FBF7F9FC07FA0C07F9FCF80A0CF80A0D080AFD080AFDF8FB0DF80B0E08FBFE08FBF\
EF8FC0E08FC0EF90CFE090C0EF9FCFEF9FCFEFA0D0EFA0CFEFA0D0EFA0D0F0AFDF
EFAFDFF0B0DFF0B0DFF0BFE0F0BFE0F0C0E0F0C0E0FFC0EFF0C0EFFFCFEFF0CFEFFFD0F0F0D0F0\
FFDFFFFFDFF0FFDFFFF0DFF0F0E0FFF0DFFFF0E0FFEFDFF0EFE0FFE0E0FFEFE0FF
E0E0F0E0E0FFDFE0FFE0EFFFDFE0FFDFEFFFD0EFFFD0EFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000D05FAFDF5FBFE060CFE060CFEF6FDFEF6FDF\
F070EFF070F0FF7FFFFF80FFFF8FFFFF90FFFFA0FFFFA0FFFFB0FFFFB0FFFFC0FF
FFC0FFFFD0FFFFDFFFFFE0FFEFE0F0EFEFF0E0E0EFDFEFEFD0EFE0CFF0E0C0F0DFC0F0DFB0F0D0\
B0FFD0AFFFCFA0FFC090FFB080FFA070FF906FFF8050F06F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000005F505F5F505F
605F6060505F6F5F606F5F5F705F60705F60705F60705F607F606F7F606080606F80606F8F6F70\
8F6070906F70906F709F6F7F9F6F7FA07080A06F80AF708FAF708FB07F90B07090
BF7F9FBF7F9FC07F9FC07F9FCF80A0CF80A0D080AFD080AFDF8FB0DF8FB0E08FBFE08FBFEF90C0\
E08FC0EF90C0EF90C0EF9FCFEF9FCFEFA0D0EFA0CFF0AFD0EFAFD0F0AFDFEFAFDF
F0B0DFF0B0DFF0BFE0F0BFE0FFC0EFF0C0E0FFC0EFF0C0EFFFCFF0FFCFEFFFD0F0FFD0F0FFDFFF\
FFDFFFFFDFFFFFDFFFFFE0FFF0E0FFF0E0FFEFE0FFF0E0FFEFE0FFEFEFFFE0E0FF
EFEFFFE0EFFFE0EFFFDFEFFFE0EFFFDFEFFFDFEFFFD0EFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000E05FBFE05FC0EF60CFEF60D0F06FDFF06FE0FF70F0\
FF70F0FF80FFFF80FFFF90FFFF90FFFFA0FFFFA0FFFFB0FFFFBFFFFFC0FFFFCFFF
FFDFFFFFDFFFFFE0FFF0E0F0EFEFF0E0EFEFDFEFEFD0EFE0D0F0EFCFF0E0C0FFE0BFF0DFB0FFDF\
AFFFD0A0FFC090FFB08FFFAF7FFF906FFF8F5FFF70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000005050505F505F
5F505F605F6060505F6F5F606F5F60705F60705F607F5F607F5F6080606F7F5F6080606F80606F\
8F6F708F6070906F7F906F7F9F6F809F6F7FA07080A07080AF708FAF708FB07F90
B07090BF7F9FBF7F9FC07FA0C07FA0CF80AFCF80A0D080AFD080AFDF8FB0DF80B0E08FBFE08FBF\
EF90C0E090C0EF90C0EF90C0EF9FCFEF9FCFEFA0D0EFA0D0F0AFD0EFAFD0F0AFDF
EFAFDFF0B0E0F0B0DFF0BFE0F0BFE0F0C0EFF0C0E0FFCFEFF0CFEFFFCFF0F0CFEFFFD0F0FFD0F0\
FFDFFFFFDFF0FFDFFFF0DFFFF0E0FFEFDFF0F0E0FFEFE0FFEFE0FFE0E0F0E0E0FF
E0E0FFE0EFFFDFE0F0DFEFFFDFEFFFDFEFFFD0EFFFD0EFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000E05FBFDF5FC0E060CFEF60D0F06FE0EF6FE0F070F0FF70F0\
FF80FFFF80F0FF90FFFF90FFFFA0FFFFA0FFFFB0FFFFBFFFFFC0FFFFCFFFFFD0FF
FFDFFFFFE0FFEFE0F0EFEFF0E0E0EFDFEFEFD0EFE0CFF0E0C0F0DFC0F0DFB0F0D0B0FFD0AFFFCF\
A0FFC090FFB08FFFA07FFF9070FF8F5FFF7050CF5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000005F505F5F505F
605F5F60505F6F5F606F5F60705F60705F607F606F7F5F607F606F7F5F6080606F80606F8F6F70\
8F6070906F7F906F7F9F6F7F9F6F7FA07080A07080AF708FAF708FB07F90B07090
BF7F9FBF7F9FC080A0C07FA0CF80A0CF80A0D080AFD080AFDF8FB0DF80B0E08FBFE08FBFEF90C0\
E08FC0EF90CFEF90C0EF9FCFEF9FCFEFA0D0EFA0CFEFAFD0EFAFD0F0B0DFEFAFDF
F0B0DFF0B0DFF0BFE0F0BFE0F0C0E0F0C0E0FFCFEFF0C0EFFFCFF0F0CFEFFFD0F0FFD0F0FFDFFF\
FFDFF0FFE0FFFFDFFFFFE0FFF0DFFFF0E0FFEFE0FFEFE0FFEFE0FFEFEFFFE0E0FF
E0EFFFDFE0FFE0EFFFDFEFFFDFEFFFD0EFFFDFF0FFD0EFF0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CF5FAFE060C0E060C0EF6FD0EF6FDFF070E0F070EFFF7FF0FF7FFFFF80FF\
FF8FFFFF90FFFF9FFFFFA0FFFFAFFFFFB0FFFFBFFFFFC0FFFFCFFFFFDFFFFFDFFF
FFE0FFEFE0F0EFEFF0E0EFEFDFEFEFD0EFE0D0F0E0C0F0DFC0F0DFBFF0D0B0FFD0AFFFCFA0FFC0\
90FFB08FFFAF7FFF9070FF8F60FF7F5FF06F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000005F505F60505F
5F505F605F5F60505F6F5F606F5F5F705F60705F607F60607F5F6080606F8060608F606F80606F\
8F6F708F6070906F7F906F7F9F6F809F6F80A0708FA07080AF708FAF708FB07F90
B07090BF7F9FBF7F9FC080A0C07FA0CF80AFCF80AFD080B0D080AFDF8FB0DF8FB0E08FBFE08FBF\
E090C0E090C0EF9FCFE090C0EF9FCFEF9FCFEFA0D0EFA0CFEFAFD0EFAFD0F0B0DF
EFAFDFF0B0DFEFB0DFF0BFE0F0BFE0F0C0E0F0C0E0FFCFEFF0CFEFFFCFF0F0CFF0FFD0F0F0D0F0\
FFDFFFFFDFFFFFE0FFF0DFF0F0E0FFEFE0FFF0E0FFEFE0F0EFE0FFE0E0F0E0EFFF
DFE0F0E0EFFFDFEFF0DFEFFFD0EFF0DFEFFFD0EFFFCFEFF0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000DF5FB0E060C0E060C0EF6FD0EF6FDFF070E0F06FEFFF70F0FF7FF0FF80FFFF8FF0\
FF90FFFF9FF0FFA0FFFFAFF0FFB0FFFFBFF0FFCFFFFFCFF0FFD0FFFFDFFFFFE0FF
EFE0F0EFE0F0E0E0EFDFEFEFD0EFE0CFF0E0C0F0DFC0F0DFB0F0D0B0FFD0AFFFCFA0FFC090FFAF\
8FFFA07FFF9070FF8F60FF7F5FFF6F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F505F60505F5F505F
605F60605F5F6F5F606F5F60705F60705F607F606F7F5F6080606F80606F80606F80606F8F6F70\
8F6F70906F7F906F7F9F70809F6F80A07080A07080AF7F8FAF708FB07F90B07F90
BF7F9FBF7F9FC080A0C07FA0CF80AFCF80AFD080AFD080AFDF8FB0DF8FB0E08FBFE08FBFEF90C0\
E090C0EF9FC0E090C0EF9FCFEF9FCFEFA0D0EFA0CFF0AFD0EFAFD0F0B0DFEFB0DF
F0BFDFF0B0DFF0BFE0F0BFE0F0C0EFF0C0E0FFCFEFF0CFEFFFD0F0FFCFF0FFD0F0FFD0F0FFDFFF\
FFDFFFFFE0FFFFDFFFFFE0FFF0E0FFF0E0FFEFE0FFEFEFFFEFE0FFEFEFFFE0E0FF
E0EFFFDFEFFFE0EFFFDFEFFFDFF0FFD0EFFFD0F0FFC0E0EF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000DF5FBFE060C0E060CFEF6FDFEF6FDFF070EFF070EFFF7FFFFF7FFFFF8FFFFF8FF0FF9FFF\
FFA0FFFFAFFFFFB0FFFFBFFFFFC0FFFFCFFFFFCFFFFFDFFFFFE0FFFFE0FFEFE0F0
EFEFF0E0EFEFDFEFEFD0EFE0D0F0E0C0F0DFC0F0DFBFF0D0B0FFD0AFFFCFA0FFC090FFB08FFFAF\
7FFF9070FF8F60FF7F5FFF70000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F505F5F505F60505F
60505F6F5F606F505F6F5F60705F60705F60705F607F60607F5F6080606F8060608F606F8F606F\
906F709060709F6F7F9F6F7FA070809F6F80A0708FAF708FB07F90AF708FB07F90
B07F90BF7F9FBF7F9FC080A0C080A0CF80AFCF80AFD08FB0D080B0DF8FBFDF8FB0E08FBFE08FBF\
EF90C0E090C0EF9FCFE09FC0EFA0CFE09FCFEFA0D0EFA0D0EFAFD0EFAFD0F0B0DF
EFB0DFF0BFE0EFB0DFF0BFE0F0BFE0F0C0EFF0C0E0F0CFEFF0CFEFFFD0F0F0D0F0FFD0F0FFD0F0\
FFDFFFFFDFFFFFE0FFF0DFFFF0E0FFEFE0F0EFE0FFEFE0FFEFEFFFE0E0F0E0EFFF
DFEFF0E0EFFFDFEFF0DFEFFFD0EFF0DFF0FFD0EFF0C0E0EF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000DF5FB0
DF5FBFE060CFE060CFEF6FDFEF6FDFF070EFF070EFFF7FFFFF7FF0FF8FFFFF8FF0FF9FFFFFA0F0\
FFAFFFFFB0F0FFBFFFFFC0F0FFCFFFFFCFF0FFDFFFFFDFFFF0E0FFEFE0F0EFEFF0
DFE0EFDFEFEFD0EFE0CFF0E0C0F0DFC0F0DFB0F0D0B0FFD0AFFFCFA0FFC090FFAF8FFFA07FFF90\
70FF8F60FF7F5FFF6F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F505F605F5F60505F
605F6060505F6F5F606F5F60705F60705F607F606F7F5F6080606F80606F8F606F8F606F8F6F70\
8F6070906F7F906F7F9F70809F6F80A0708FA0708FAF708FAF708FB07F90B07F90
BF7F9FBF7F9FC080A0C080A0CF80AFCF80AFD08FB0D080B0DF8FB0DF8FB0E090BFE08FBFE090C0\
E090C0EF9FCFE09FC0EF9FCFEF9FCFEFA0D0EFA0D0EFAFD0EFAFD0F0B0DFEFB0DF
F0BFDFEFB0DFF0BFE0F0BFE0F0C0E0F0C0E0FFCFEFF0CFEFFFD0F0F0CFEFFFD0F0FFD0F0FFDFFF\
FFDFFFFFE0FFFFE0FFFFE0FFF0E0FFF0E0FFEFE0FFEFEFFFE0E0FFEFEFFFE0EFFF
E0EFFFDFEFF0DFF0FFDFEFFFDFF0FFD0EFF0D0F0FFBFE0E0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DF60BFDF60C0
E060CFE060D0F06FDFF06FE0F070EFF070EFFF7FFFFF80F0FF8FFFFF90F0FFA0FFFFA0F0FFAFFF\
FFB0F0FFBFFFFFC0FFFFCFFFFFD0FFFFDFFFFFDFFFF0E0FFEFE0F0EFEFF0DFEFEF
DFEFEFD0EFE0D0F0E0C0F0DFC0F0DFBFF0D0B0FFD0AFFFCFA0FFC090FFAF8FFFA07FFF9070FF8F\
60FF7F5FFF7050D05F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F505F5F505F605F5F
60505F6F5F606F5F5F705F606F5F60705F60705F607F606F7F5F6080606F80606F8F60708F606F\
906F70906F709F6F7F9F6F7FA07080A06F80AF708FAF708FB07090B07090BF7F9F
B07F90BF7F9FC07F9FC080A0C07FA0CF80AFCF80AFD08FB0D080B0DF8FBFDF8FBFE090BFE08FBF\
E090C0E090C0EF9FCFE09FC0EFA0CFEFA0CFEFAFD0EFA0CFEFAFD0EFAFD0F0B0DF
EFB0DFF0BFDFEFBFDFF0C0E0F0BFE0F0C0EFF0C0E0F0CFEFF0CFEFFFD0F0F0D0F0FFDFF0F0D0F0\
FFDFFFFFDFFFFFE0FFF0DFF0F0E0FFEFE0FFEFE0FFE0E0F0EFEFFFE0E0F0E0EFFF
DFEFF0DFEFFFDFEFF0DFF0FFD0EFF0D0F0FFD0EFF0B0E0E0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B0508FDF5FBFDF5FC0E060CF
E060CFEF6FDFEF6FE0F070EFF070EFFF7FF0FF80F0FF90F0FF90F0FF9FF0FFA0F0FFAFFFFFB0F0\
FFBFFFFFC0F0FFCFFFFFD0F0FFDFFFF0DFF0F0E0F0EFE0EFE0E0EFDFE0E0DFEFE0
CFEFDFCFF0DFC0F0D0C0F0D0B0F0CFB0FFCFAFFFC0A0FFBF90FFAF8FFFA07FFF9070FF8F60FF70\
5FFF6F50DF5F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F505F5F505F605F5F60505F
6F5F606F5F60705F606F5F6070606F705F607F606F7F606080606F80606F8F6F6F8F606F906F70\
906F709F6F7F9F6F7FA07080A07080A0708FA0708FB07F90B07090B07F90B07F90
BF809FBF7F9FC080A0C080A0CF80AFCF80AFD08FB0D080B0DF8FBFDF8FB0E090BFE08FBFEF90C0\
E090C0EF9FCFE09FC0EFA0CFEFA0CFEFA0D0EFA0D0F0AFD0EFAFD0F0B0DFEFB0DF
F0BFE0F0BFDFF0C0E0F0BFE0F0C0EFF0C0EFFFCFEFF0CFEFFFD0F0F0D0F0FFDFF0FFD0F0FFDFFF\
FFDFFFFFE0FFF0E0FFFFE0FFF0E0FFF0EFFFEFE0FFEFEFFFE0EFFFEFEFFFE0EFFF
E0F0FFDFEFFFDFF0FFD0EFF0DFF0FFD0F0FFD0F0FFAFE0DF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F9F4F50A05FA04F8FB0509FAF509F
B05FA0B05FA06FB0806FB080FF80FFFF8FF0FF90FFFF90F0FFA0FFFFA0FFFFB0FFFFB0FFFFC0FF\
FFC0FFFFD0FFFFD0FFFFE0FFF0E0FFF0E0F0EFE0EFE0EFF0DFEFEFDFEFE0CFEFDF
CFF0E0C0F0DFC0F0D0B0F0CFB0FFD0AFFFCFA0FFBF90FFAF8FFFA07FFF9070FF8F60FF7F60FF70\
50E060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F505F5F505F5F505F605F5F
60505F6F5F606F5F5F705F60705F607F6060705F607F606F7F606080606F80606F8F6F708F6070\
906F7F906F709F6F7F9F6F7FA07080A07080AF708FAF708FB07F90B07090BF7F9F
BF7F9FC080A0BF7F9FC080A0CF80A0CF80AFCF80AFD08FB0D08FB0DF8FBFDF8FBFE090BFE090BF\
E09FC0E090C0EF9FCFE09FCFEFA0CFEFA0CFEFAFD0EFAFD0EFB0DFEFAFD0F0B0DF
EFB0DFF0BFE0EFBFDFF0C0E0F0C0E0F0CFEFF0C0EFF0CFEFF0CFEFFFD0F0F0D0F0FFDFF0FFDFF0\
FFE0FFF0DFFFF0E0FFF0E0FFF0E0FFEFE0F0EFE0FFE0E0F0EFEFFFE0E0F0E0EFFF
DFEFF0DFEFFFD0EFF0DFF0FFD0EFF0D0F0FFCFEFF0AFE0D0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000008F4F708F407F9F4F8FA04F90AF509FAF509FB05F9F
B05F9FB060A0B060A0BF6FAFBF6FA0BF70AFFF90F0FFA0FFFFA0F0FFB0FFFFB0F0FFC0FFFFC0FF\
FFD0FFFFD0FFFFDFFFF0DFF0F0E0F0E0E0EFE0EFEFDFE0E0D0EFE0CFEFDFCFF0DF
C0F0D0BFF0D0B0F0CFB0FFCFA0FFC0A0FFBF90FFAF8FFFA07FFF9070FF8F60FF705FFF6F50E05F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F505F5F505F605F6060505F
6F5F606F5F5F705F60705F60706060705F607F606F7F606F80606F80606F8F6F708F6070906F70\
906F709F707F9F6F7FA07080A07080AF708FAF708FAF7F90AF7090B07F9FB07F90
BF809FBF7F9FC080A0C080A0CF80AFCF80AFD08FB0D08FB0DF8FB0DF8FB0E090BFE090BFE090C0\
E090C0EF9FCFE09FC0EFA0CFEFA0CFEFAFD0EFAFD0EFAFD0EFAFD0F0B0DFEFB0DF
F0BFE0EFBFDFF0C0E0F0C0E0F0C0EFF0C0E0FFCFEFF0CFEFFFD0F0F0D0F0FFDFFFFFDFF0FFDFFF\
FFDFFFFFE0FFF0E0FFF0E0FFEFE0FFF0EFFFEFE0FFEFEFFFE0EFFFE0EFFFE0EFFF
E0EFFFDFEFF0DFF0FFD0F0F0D0F0FFD0F0F0CFF0F0A0DFD0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000080406F8F4F7F904F80A05090AF4F9FB0509FAF509FB05FA0B05FA0
B060A0B060A0BF6FAFBF6FAFC070AFBF70AFC07FB0C07FB0C080B0FFB0F0FFC0FFFFC0FFFFD0FF\
FFD0FFFFE0FFF0E0F0F0E0F0E0E0EFE0EFEFDFEFE0D0EFE0CFEFDFCFF0DFC0F0D0
BFF0D0B0F0CFB0FFCFA0FFC0A0FFBF90FFAF8FFFA07FFF9070FF8F6FFF7F60FF7050E05F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F50505F505F5F505F605F60
60505F6F5F606F5F60705F60705F607F60607F5F607F606F7F606080606F80606F8F6F708F6070\
906F7F906F7F9F70809F6F7FA07080A07080AF708FAF708FB07F90B07F90BF7F9F
BF7F9FC0809FC07F9FCF80A0C080A0CF80AFCF80AFD08FB0D08FB0DF8FBFDF8FBFE090BFE090BF\
E09FC0E09FC0EF9FCFE09FC0EFA0CFEFA0CFEFAFD0EFAFD0EFB0D0EFB0D0F0B0DF
EFB0DFF0BFE0EFBFE0F0C0E0F0C0E0F0CFEFF0CFEFFFD0EFF0CFEFFFD0F0F0D0F0FFDFFFF0DFF0\
FFE0FFF0DFFFF0E0FFEFE0F0F0E0FFEFE0F0EFEFFFE0E0F0E0EFFFE0EFF0E0EFFF
DFEFF0DFEFFFD0EFF0D0F0FFD0EFF0D0F0FFC0EFEF9FDFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000080406F904F7F904F80A04F90AF4F90AF509FAF509FB05F9FB05F9FB060A0
B060A0BF6FA0BF6FA0BF70AFBF70AFC07FB0BF7FAFC080B0BF80AFBF8FAFFFC0F0FFD0FFFFD0F0\
FFDFFFF0DFF0EFE0F0E0E0EFE0EFEFD0E0E0D0EFE0CFEFDFC0F0DFBFF0D0BFF0D0
B0F0CFAFFFCFA0FFBF9FFFB090FFA08FFFA07FFF9070FF8F60FF705FFF6F50E05F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000605F5F5F505F605F6060505F
6F5F606F5F60705F60705F607F606F7F60607F606F7F606F80606F80606F8F6F708F6070906F7F\
906F7F9F707F9F6F7FA07080A07080AF708FAF708FB07F90B07F90B07F9FB07F90
C0809FBF809FC080A0C080A0CF8FAFCF80AFD08FB0D08FB0DF8FBFDF8FBFE090BFE090BFE09FC0\
E09FC0EF9FCFE09FC0EFA0CFEFA0CFEFAFD0EFAFD0F0B0DFEFB0D0F0B0DFEFB0DF
F0BFE0F0BFE0F0C0E0F0C0E0FFCFEFF0CFEFFFCFEFF0CFEFFFD0F0FFD0F0FFDFFFFFDFF0FFE0FF\
FFE0FFFFE0FFF0E0FFF0EFFFEFE0FFEFEFFFEFE0F0EFEFFFE0EFFFE0EFFFDFEFF0
E0F0FFDFEFFFDFF0FFD0F0F0D0F0FFD0F0F0C0EFEF90DFC0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000804F6F\
8F4F70904F809F4F8FAF509FAF509FB05F9FAF5F9FB060A0B060A0BF6FA0B06FA0
BF70AFBF70AFC07FAFBF7FAFC080B0BF80AFBF80B0BF80AFBF8FAFB08FAFB090AFAFA0B0FFE0FF\
F0E0F0EFE0F0E0E0EFE0EFEFDFEFE0D0EFE0CFEFDFCFF0DFBFF0D0BFFFD0B0F0CF
B0FFCFA0FFBF9FFFB090FFA08FFFA07FFF907FFF8F6FFF7F60FF7050EF60000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F505F5F505F605F5F60505F605F60
605F5F6F5F606F5F60705F60705F607F606F7F5F6080606F80606F80606F80606F8F6F708F6F70\
906F7F906F7F9F70809F6F80A07080A07080AF708FAF708FB07F90B07F90BF7F9F
BF7F9FC080A0C07F9FCF80A0CF80A0D08FAFCF80AFD08FB0D08FB0DF90BFDF8FBFE090BFE090BF\
E09FC0E09FC0E0A0CFE0A0CFEFA0CFE0A0CFEFAFD0EFAFD0EFB0DFEFB0D0F0BFDF
EFBFDFF0BFE0EFBFE0F0C0E0F0C0E0F0CFEFF0CFEFF0D0EFF0D0EFFFDFF0F0D0F0FFDFFFFFDFFF\
FFE0FFF0DFF0F0E0FFEFE0F0F0EFFFEFE0F0EFEFFFE0E0F0E0EFFFDFEFF0E0EFFF
DFEFF0DFF0FFD0EFF0D0F0FFD0F0F0D0F0FFB0E0E090DFC0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000804F6F8F4070\
9F4F80A04F8FAF509FAF5090B05F9FAF5F9FB0609FB0609FB06FA0B06FA0BF70AF
BF70A0C07FAFBF7FAFC080AFBF80AFBF80AFB080AFB08FAFB08FA0B090AFAF90A0AF9FA0A09FA0\
EFE0F0E0E0EFDFEFE0D0EFE0D0EFE0CFEFD0C0F0D0BFF0CFBFFFCFAFF0C0AFFFC0
A0FFBF9FFFB08FFFA08FFF9F7FFF8F70FF8060FF7060FF6F50E05F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F505F605F5F5F505F605F60605F5F
6F5F606F5F60706060705F607F606F7F5F6080606F7F606F80606F80606F8F6F708F6F70906F7F\
906F7F9F70809F6F7FA07080A07080AF7F8FAF708FB07F90B07F90BF7F9FBF7F9F
BF809FBF7F9FC080A0C080A0CF8FAFCF80AFD08FB0D08FB0DF8FBFDF8FB0E090BFE090BFE09FC0\
E09FC0EFA0CFE09FC0EFA0CFE0A0CFEFAFD0EFAFD0EFB0D0EFB0D0F0BFDFEFBFDF
F0BFE0F0BFDFF0C0E0F0C0E0F0CFEFF0CFEFFFD0F0F0D0EFFFD0F0F0D0F0FFDFFFFFDFF0FFE0FF\
FFE0FFFFE0FFF0E0FFF0EFFFEFE0F0EFEFFFEFEFF0EFEFFFE0EFF0E0F0FFDFEFF0
DFF0FFDFEFF0DFF0FFD0F0F0D0F0FFCFF0F0BFEFE08FD0BF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007F40608F4F70904F7F9F4F8F\
A04F90AF509FAF509FB05F9FAF5F9FB060A0B060A0BF6FA0B06FA0BF70AFBF70AF
C07FAFBF7FAFC080B0BF80AFBF8FAFB08FAFB090AFB08FA0B090AFAF90A0AF9FAFA09FA0A0A0A0\
9FA09FE0EFEFD0EFE0D0EFDFCFEFD0C0F0D0BFF0CFBFFFCFAFF0C0AFFFC0A0FFBF
9FFFB08FFFA08FFF9F7FFF8F70FF8060FF7060FF6F50EF60000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F505F5F505F60505F60505F6F5F60
60505F6F5F606F5F60706060705F607F606F7F606080606F7F606080606F8F606F8F6F708F6F70\
906F7F906F7F9F70809F6F80A07080A07080AF7F8FAF708FB07F90B07F90BF7F9F
BF7F9FC080A0C080A0CF80A0C080A0CF8FAFD080AFD08FB0D08FB0DF8FBFDF8FBFE090BFDF90BF\
E09FC0E09FC0E0A0CFE0A0C0EFA0CFE0A0CFEFAFD0EFAFD0EFB0D0EFB0D0F0BFDF
EFBFDFF0BFE0F0BFE0F0C0E0F0C0E0F0CFEFF0CFEFFFD0F0F0D0EFFFDFF0F0DFF0FFDFFFF0DFF0\
FFE0FFF0E0F0F0E0FFEFE0F0EFE0FFEFE0F0EFEFF0E0E0F0E0EFF0DFEFF0DFEFF0
DFEFF0DFF0F0D0EFF0D0F0F0CFF0F0CFF0F0AFE0DF80D0BF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000007F406F8F4F7F904F7FA04F8FA04F90\
AF509FAF5090B05F9FAF5F9FB0609FB0609FB06FA0B06FA0BF70A0BF70A0BF7FAF
BF7FAFBF80AFB080AFBF8FAFB080A0B08FAFAF8FA0AF90A0A090A0A09FA0A09F9FA0A0A09FA09F\
9FAF9F90AF9FD0EFDFC0EFD0C0F0D0BFF0CFB0FFCFAFF0C0AFFFC09FFFB09FFFAF
8FFF9F80FF907FFF8F70FF8060FF705FFF6F50EF5F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F505F605F6060505F6F5F60605F5F
6F5F606F5F60706060705F607F606F7F606080606F80606F8F6F7080606F8F6F708F6F7090707F\
906F7F9F70809F7080A0708FA07080AF7F8FAF708FB07F90B07F90BF809FBF7F9F
C080A0C0809FC080A0C080A0CF8FAFCF8FAFD08FB0D08FB0DF90BFDF8FBFE090BFDF90BFE09FC0\
E09FC0E0A0CFE0A0C0EFAFCFE0A0CFEFAFD0EFAFD0EFB0DFEFB0D0F0BFDFEFBFDF
F0C0E0F0C0E0F0C0E0F0C0E0FFCFEFF0CFEFFFD0F0F0D0EFFFDFF0FFDFF0FFDFFFFFDFF0FFE0FF\
FFE0FFFFE0FFF0E0FFF0EFFFEFE0F0EFEFFFE0EFF0EFEFFFE0EFF0E0F0FFDFEFF0
DFF0FFDFF0F0DFF0FFD0F0F0D0FFFFCFF0F0AFE0DF7FD0B0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000804F6F804070904F809F4F80AF5090AF5090AF5F9F\
AF5F9FB0609FAF609FB06FA0B06FA0BF6FA0B06FA0BF7FAFBF7FAFC07FAFBF7FAF
BF80AFBF80AFBF8FAFB08FA0B090AFAF90A0AF9FA0A09FA0AFA0A0A0A0A0A0A0A09FA09F9FAFA0\
90AF9F90B09F8FAF90C0F0D0BFF0CFBFFFCFAFFFC0AFFFC09FFFB09FFFAF8FFF9F
80FF907FFF8F70FF8060FF7060FF6F50EF5F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005050505F505F5F505F605F6060505F6F5F60
6F5F60705F606F5F60706060705F607F606F7F606080606F80606F8F6F708F606F906F70906F70\
9F707F906F7F9F70809F7080A0708FA0708FAF7F8FAF7F8FB07F90B07F90BF809F
BF7F9FC080A0C080A0CF80A0CF80A0D08FAFD080AFD08FB0DF8FB0E090BFDF90BFE090BFDF90BF\
E09FC0E09FC0E0A0CFE0A0CFEFAFCFE0AFCFEFAFD0EFAFD0EFB0DFEFB0D0EFBFDF
EFBFDFF0C0E0EFC0E0F0C0E0F0C0E0F0CFEFF0CFEFFFD0F0F0D0F0FFDFF0F0DFF0FFE0FFFFE0FF\
FFE0FFF0E0F0F0E0FFEFE0F0EFEFFFE0E0F0EFEFF0E0EFF0E0EFF0DFEFF0DFF0F0
D0EFF0DFF0F0D0EFF0D0F0F0CFF0F0CFFFF0A0DFCF7FD0AF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000804F6F8F4070904F809F4F8FAF5090A05090AF5F9FAF5F90\
B0609FAF5F9FB060A0B06F9FBF70A0B06FA0BF70A0BF70A0BF7FAFBF7FA0BF80AF
B080A0B08FAFAF8FA0AF90A0AF90A0AF9FA0A0909FA09FA09F9F9F9FA0A090A09F90AF9F90AF9F\
90B09F8FAF908FB09080AF8FB0FFCFAFF0C0AFFFBF9FFFB090FFAF80FF9F80FF90
70FF8070FF7F60FF6F5FFF6F50E05F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F5F5F5F505F605F5F60505F6F5F606F5F60
6F5F606F5F6070606F705F607F606F7F606080606F80606F806F7080606F8F6F708F6F70906F7F\
906F7F9F70809F7080A0708FA07080AF7F8FAF7F8FB07F90B07F90BF809FBF7F9F
C080A0C0809FCF8FA0CF80A0CF8FAFCF8FAFD08FB0D08FB0DF90BFDF8FBFE090BFDF90BFE09FC0\
E09FC0E0A0CFE0A0CFEFAFCFE0A0CFEFAFD0EFAFD0EFB0D0EFB0D0F0BFDFEFBFDF
F0C0E0EFC0E0F0CFE0F0C0E0F0CFEFF0CFEFFFD0F0F0D0F0FFDFF0F0DFF0FFE0FFFFDFFFFFE0FF\
F0E0FFFFEFFFF0E0FFF0EFFFEFE0F0EFEFFFE0EFF0E0EFF0DFEFF0E0F0F0DFEFF0
DFF0F0D0F0F0D0F0F0D0F0F0D0FFF0C0F0EFA0E0D070CFAF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007F40608F4F708F4F7F9F508FA04F90AF5090AF5090AF5F9FAF5F9FB0609F\
AF609FB06FA0B06FA0BF70A0B070A0BF7FAFBF7FAFBF80AFB080A0BF8FAFB080A0
B08FAFAF8FA0B090A0AF90A0AF9FA0A09F9FA0A0A09FA09F9FAF9F90A09F9FB09F90AF9F90B09F\
8FAF908FB09080B08F7FB08FBF70BFAFFFBF9FFFB090FFAF8FFF9F80FF9070FF80
70FF7F60FF6F5FFF6F50E05F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F505F5F505F5F505F605F5F60505F6F5F60
6F5F5F705F60705F607F6060705F607F606F7F606080606F80606F8F6F708F6070906F7F906F70\
9F6F7F9F6F7FA070809F7080A0708FA0708FAF7F90AF708FB07F90B07F90BF809F
BF7F9FC080A0C080A0CF8FAFCF80A0D08FAFD08FAFDF8FB0D08FB0DF90BFDF90BFE09FC0DF90BF\
E09FC0E09FC0E0A0CFE0A0C0E0AFCFE0AFCFEFB0D0E0AFD0EFB0DFEFB0D0EFBFDF
EFBFDFF0C0E0EFC0E0F0CFEFF0C0E0F0CFEFF0CFEFFFD0F0F0D0EFFFDFF0F0DFF0FFE0FFF0DFFF\
FFE0FFF0E0F0F0EFFFEFE0F0EFEFF0E0E0F0E0EFF0DFEFF0E0EFF0DFEFF0DFF0F0
D0EFEFD0F0F0D0F0F0D0F0F0CFF0EFC0F0EF90DFC0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000007F406F8F4F7F904F7FA04F8FA04F90AF5090A05090AF5F9FAF5F90B0609FAF609F\
B06FA0B06F9FBF70A0B070A0BF7FA0BF7FA0BF80AFB07FA0B080A0AF80A0B08FA0
AF8F9FAF90A0A0909FA09FA09F9F9FA0A09F9FA09F9FAF9F90A09090AF9F8FAF908FB09080AF8F\
80B08F7FAF807FB08070AF7F70B07F9FFFAF90FFA080FF9080FF8F70FF806FFF7F
60FF6F5FFF6050DF5F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F5F5F5F505F605F60605F5F6F5F606F5F60
706060705F6070606F705F607F606F7F606F80606F80606F8F6F708F6F70906F7F906F709F707F\
9F6F7F9F70809F7080A07F8FA0708FAF7F90AF7F8FB07F90B07F90BF809FBF7F9F
C080A0C080A0CF8FAFCF80A0D08FAFD08FAFD08FB0D08FB0DF90BFDF90BFE09FBFDF90BFE09FC0\
E09FC0E0A0CFE0A0CFEFAFCFE0AFCFEFB0D0E0AFD0EFB0DFEFB0D0EFBFDFEFBFDF
F0C0E0F0C0E0F0CFE0F0CFE0F0D0EFF0CFEFFFD0F0F0D0EFFFDFF0FFDFF0FFE0FFFFE0FFFFE0FF\
F0E0FFF0EFFFF0E0F0F0EFFFEFEFF0EFEFFFE0EFF0E0F0FFDFEFF0DFF0F0DFEFF0
DFF0F0D0F0F0D0F0F0CFF0F0D0FFF0BFF0E090DFC0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
804F6F804F70904F809F4F80AF5090A05090AF5F9FAF5F90B0609FAF609FB0609FB0609FB06FA0\
B06FA0BF70A0B070A0BF7FAFBF7FA0BF80AFB080A0B08FAFB08FA0B090A0AF8FA0
AF9FA0A090A0A09FA0A09F9FA0A0A09FA09F9FAF9F90AF9F90B09F8FAF908FB09080AF8F80B08F\
7FB08F7FB08070B07F70B0806FB07F90FFA080FF9080FF9070FF806FFF7F60FF6F
5FFF6050DF5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F505F60505F60505F605F6060505F6F5F60
6F5F60706060705F607F606F7F60607F606F7F606080606F80606F8F6F708F6F70906F7F906F7F\
9F70809F6F7FA07080A07080AF7F8FA0708FAF7F90AF7F90B07F9FB07F90BF809F
BF809FC080A0C080A0CF8FAFCF80AFD08FB0D08FAFDF8FB0DF8FB0DF90BFDF90BFE09FC0DF9FC0\
E0A0C0DF9FC0E0A0CFE0A0CFEFAFD0E0AFCFEFB0D0E0B0D0EFBFDFEFB0D0EFBFDF
EFBFDFF0C0E0EFC0E0F0CFE0F0CFE0F0D0EFF0D0EFF0D0F0F0D0F0FFDFF0F0DFF0FFE0FFFFE0FF\
FFE0FFF0E0F0F0EFFFEFE0F0EFEFFFE0EFF0E0EFF0E0EFF0E0F0F0DFEFEFDFF0F0
D0F0F0D0F0F0CFF0EFD0F0F0CFF0EFBFF0E08FD0BF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000804F6F
8F40709F4F809F4F8FA05090A05090AF5F90A05F90AF609FAF609FB06F9FAF609FB06FA0B06F9F\
BF70A0B070A0BF7FAFB07FA0BF80A0B080A0B08FA0AF8FA0AF90A0A08F9FAF90A0
A0909FA09F9F9F9F9F9FA09F9FA09F9FAF9F90AF9090AF908FAF908FB09080AF8F80B08F7FAF80\
7FB0806FAF7F6FB07F60B07060B0705FB06F7FFF8F70FF7F6FFF7F5FFF6F5FFF60
50D05F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F505F605F5F5F505F605F60605F5F6F5F606F5F60
705F60705F607F606F7060607F606F7F606F806F6F80606F8F6F708F6F70906F7F906F7F9F707F\
9F6F7FA07080A07080A07F8FA0708FAF7F90AF7F90B07F90B07F90BF809FBF809F
C080A0C080A0CF8FAFCF80AFD08FAFCF8FAFDF90B0D08FB0DF90BFDF90BFE09FC0DF9FBFE09FC0\
DF9FC0E0A0CFE0A0CFE0AFCFE0AFCFEFB0D0E0B0D0EFB0DFEFB0D0EFBFDFEFBFDF
F0C0E0EFC0E0F0CFEFF0CFE0F0D0EFF0CFEFFFD0F0F0D0F0FFDFF0F0DFF0FFE0FFFFE0FFFFE0FF\
F0E0FFF0EFFFF0E0FFF0EFFFEFEFF0EFEFFFE0EFF0E0F0F0DFEFF0DFF0F0DFF0F0
DFF0F0D0F0EFD0FFF0CFF0F0CFFFF0B0EFDF8FDFBF000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007F406F8F4F70904F7F
9F4F8FA04F8FAF5090A05090AF5F9FAF5F90B0609FAF609FB06F9FB06F9FB070A0B070A0BF7FA0\
B07FA0BF80AFB080A0B080A0B080A0B08FA0AF8FA0AF90A0A0909FAF9FA0A09F9F
A09F9F9F9F9F9FA09F9FA09F9FAF9F90AF9090B09F8FAF908FB09080AF8F80B08F7FB0807FB080\
6FB07F6FB07F60B07060B0705FB06F5FBF6F70FF7F6FFF705FFF6F5FFF6050D05F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F505F5F505F605F5F60505F605F60605F5F6F5F60
6F5F60705F60705F607F606F7F5F607F606F7F606F806F6F80606F8F6F708F6F70906F7F906F7F\
9F70809F7080A07080A07080AF7F8FAF708FB07F90AF7F90B07F9FB07F9FBF809F
BF809FC080A0C080A0CF8FAFCF80AFD08FAFD08FAFDF90B0D08FB0DF90BFDF90BFDF9FC0DF9FBF\
E0A0C0DFA0C0E0AFCFE0A0CFE0AFCFE0AFCFEFB0D0E0B0D0EFBFDFEFBFDFEFC0DF
EFBFDFEFC0E0EFC0E0F0CFEFEFCFE0F0D0EFF0D0EFF0DFF0F0D0F0FFDFF0F0DFF0FFE0FFF0E0F0\
F0E0FFF0E0F0F0EFFFEFE0F0EFEFF0E0EFF0E0EFF0DFEFF0DFF0F0DFEFF0DFF0F0
D0EFEFD0F0F0CFF0EFCFFFF0C0F0EFAFEFDF80D0B0000000000000000000000000000000000000\
0000000000000000000000000000000000007F406F80406F8F4F7F904F80A0508F
A04F8FA05090A05090AF5F90AF5F90AF609FAF609FB06F9FAF6F9FB070A0B0709FBF7FA0B07FA0\
B07FA0B07FA0B080A0AF809FAF8FA0AF8F9FAF90A0A0909FA0909F9F9F9FA09F9F
9F9F909FA09F90A09090AF9F8FAF908FAF9080AF8F80B08F7FAF807FB08070AF8070B0806FAF7F\
6FB07F60B07060B0705FB06F5FB06F50AF606FFF705FFF605FF05F50CF50000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F505F605F5F60505F6F5F60605F606F5F606F5F60
70606F705F607F606F7F606080606F7F606F806F6F80606F8F6F708F6F70906F7F906F7F9F7080\
9F7080A07080A07080AF7F8FAF7F8FAF7F90AF7F90B0809FB07F90BF809FBF809F
C080A0C080A0CF8FAFCF80AFD08FB0D08FAFD090B0D08FB0DF90BFDF90BFDF9FC0DF9FBFE0A0C0\
E0A0C0E0A0CFE0A0CFE0AFD0E0AFCFEFB0D0E0B0D0EFBFDFEFBFDFEFBFDFEFBFDF
F0C0E0EFC0E0F0CFEFEFCFE0F0D0EFF0D0EFF0DFF0F0D0F0FFDFFFF0DFF0FFE0FFFFE0FFFFEFFF\
F0E0FFF0EFFFEFEFF0F0EFFFEFEFF0EFEFFFE0EFF0E0F0FFDFF0F0DFF0F0D0F0F0
DFF0F0D0F0F0D0FFF0CFF0EFCFFFF0A0EFD080D0B0000000000000000000000000000000000000\
000000000000000000000000000000804F6F8F40709F4F809F4F8FA05090A05090
AF5F90AF5F90AF609FAF6090B06F9FAF6F9FB06F9FB06F9FB070A0B070A0BF7FA0B07FA0BF80A0\
B080A0B08FA0AF80A0B08FA0AF8F9FAF90A0A0909FA09FA09F9F9FA0A09F9FA09F
9FAF9F90AF9090AF9F8FAF908FB09080AF8F80B08F7FAF807FB08F70B08070B0806FB07F6FB07F\
60B07060B0705FB06F5FB06F50AF6050AF605FFF605FF06050C050000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F505F5F505F605F5F60505F6F5F606F5F5F6F5F60
6F5F60706060705F607F606F7F606080606F7F606F806F708060708F6F708F6F70906F7F906F7F\
9F70809F7080A07080A07080AF7F8FAF708FB07F90B07F90BF809FB07F90BF809F
BF809FC08FA0C080A0CF8FAFCF8FAFD08FB0D08FAFDF90B0DF90B0DF90BFDF90BFDF9FC0DF9FC0\
E0A0C0DFA0C0E0AFCFE0AFCFE0B0D0E0AFCFE0B0D0E0B0D0EFBFDFE0BFDFEFC0DF
EFC0DFEFCFE0EFC0E0F0CFEFEFCFEFF0D0EFF0D0EFF0DFF0F0DFF0FFE0FFF0DFF0FFE0FFF0E0FF\
F0EFFFEFE0F0EFEFF0EFEFF0EFEFF0E0EFF0E0EFF0DFEFF0DFF0F0D0EFEFDFF0F0
D0F0EFD0F0F0CFF0EFCFFFF0C0F0EFA0EFD070CFAF000000000000000000000000000000000000\
0000000000000000007F40608F4F708F407F9F4F80A04F8FA05090A0508FAF5F90
A05F90AF609FAF6090AF609FAF6F9FB06F9FAF6F9FB070A0B070A0BF7FA0B07FA0B080A0AF809F\
B08FA0AF809FAF8FA0A08F9FAF909FA0909FA09F9F9F9F9F9FA09F90A09090A09F
90A09090AF908FAF8F8FAF9080AF8F80B08F7FAF807FB08070AF7F70B07F6FAF706FB07060B06F\
60B06F50B06050B06050AF5F50AF5F4FA0505FEF5F50BF50000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F505F5F505F605F60605F5F6F5F60605F5F6F5F606F5F60
706060705F607F606F7F606F7F606F7F606F806F70806F6F8F6F708F6F7090707F906F7F9F7080\
9F707FA07080A07080A07F8FA0708FAF7F90AF7F90B08090B07F90BF809FBF809F
C080A0C080A0CF8FAFCF8FAFD08FAFD08FAFDF90B0DF8FB0DF90BFDF90BFDF9FC0DF9FBFE0A0C0\
DFA0C0E0AFCFE0AFCFE0AFCFE0AFCFEFB0D0E0B0D0EFBFDFE0BFDFEFC0E0EFC0DF
EFC0E0EFC0E0F0CFEFEFCFEFF0D0EFF0D0EFF0DFF0F0DFF0FFDFF0F0DFF0FFE0FFF0E0FFF0EFFF\
F0E0F0F0EFFFEFEFF0EFEFFFE0EFF0EFF0F0E0EFF0E0F0F0DFEFF0DFF0F0D0F0F0
D0F0F0CFF0EFD0FFF0CFF0EFC0FFEF9FE0C070D0AF000000000000000000000000000000000000\
0000007F406F80406F8F4F7F904F7FA0508FA0508FA05090A05090AF5F90AF5F90
AF609FAF6090B06F9FAF6F9FB0709FB0709FB07FA0B07FA0B080A0B07FA0B080A0AF809FAF8FA0\
AF8F9FAF90A0A0909FA0909FA0909FA09F9F9F9F9F9FA09F90A0909FAF9F90AF90
90AF908FAF8F8FB09080AF8F80B08F7FAF807FB08070B07F70B07F6FB0706FB07060B06F5FB06F\
50B0605FB06050AF5F50AF5F4FA0505FE06050BF50000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005050505F505F5F505F605F6060505F6F5F606F5F60705F60
6F5F60706060705F607F606F7F606080606F80606F8F6F7080606F8F6F708F6F7090707F906F7F\
9F70809F7080A0708FA07080AF7F8FAF7F8FB07F90AF7F90B0809FB07F9FBF809F
BF809FC080A0C080A0CF8FAFCF8FAFD08FB0D08FAFDF90B0D090B0DF90BFDF90BFDF9FC0DF9FBF\
DFA0C0DFA0C0E0AFCFDFAFCFE0B0D0E0B0D0E0B0D0E0B0D0EFBFDFE0BFDFEFC0E0
EFC0DFEFCFE0EFCFE0F0CFEFEFCFEFF0D0EFF0D0EFF0DFF0F0DFF0F0E0FFF0E0FFFFE0FFF0E0F0\
F0EFFFEFE0F0EFEFF0E0E0F0E0EFF0E0EFF0E0F0F0DFEFEFDFF0F0D0F0EFD0F0F0
D0F0EFD0F0F0CFF0EFCFFFEFBFF0E090E0C0000000000000000000000000000000000000000000\
7F406F804070904F7F9F4F80A0508FA0508FA05090A0508FAF5F90A05F90AF609F
AF6090AF6F9FAF6F9FB0709FAF709FB07FA0B07F9FB080A0AF7F9FAF80A0AF809FAF8F9FA08F9F\
A0909FA0909FA0909F9F90909F9F9F9F9F909FA09F90A09090AF908FAF908FAF90
80AF8F80AF8F7FAF807FB08070AF7F70B07F6FAF7F6FB07F60AF7060B0705FB06F5FB06F50AF60\
50AF604FA05F50A05F4F9F504F9F50000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F5F5F5F505F605F60605F5F6F5F606F5F606F60606F5F60
70606F7060607F606F7F606F806F6F80606F806F70806F708F6F708F6F7090707F906F7F9F7080\
9F7080A07F80A07080AF7F8FAF7F8FAF7F90AF7F90B0809FB0809FBF809FBF809F
C08FA0C080A0CF8FAFCF8FAFD090B0D08FAFDF90B0D090B0DF9FBFDF9FBFDF9FC0DF9FBFE0A0C0\
DFA0C0E0AFCFDFAFCFE0B0D0E0B0D0E0B0D0E0B0D0EFBFDFE0BFDFEFC0E0EFC0DF
EFCFE0EFCFE0F0CFEFEFCFEFF0D0F0F0D0EFF0DFF0F0DFF0FFE0FFF0E0FFFFE0FFF0E0FFF0EFFF\
F0EFF0F0EFFFEFEFF0EFEFF0E0EFF0E0F0F0DFEFF0E0F0F0DFF0F0DFF0F0D0F0EF
D0FFF0CFF0EFD0FFF0CFFFEFBFF0E090E0BF0000000000000000000000000000007F40608F4F70\
8F4F7F9F4F80A04F8FA05090A0508FAF5F90A05F90AF609FAF6090AF6F9FAF6F9F
B0709FAF709FB070A0B0709FBF7FA0B07FA0B080A0AF809FB08FA0AF8F9FAF8FA0A08F9FAF909F\
A0909FA09F9F9F9F9FA0A09F9FA0909FA09F90A09090AF908FAF908FAF9080AF8F
8FB08F80AF807FB08070AF7F70B0806FB07F6FB07F60B07060B0705FB06F5FB06F50AF605FB060\
50AF5F50A05F4F9F504F9F50000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F505F605F5F5F505F605F60605F5F6F5F606F5F60705F60
705F607F606F705F607F606F7F606F806F6F80606F8F6F708F6F70906F7F8F6F7090707F90707F\
9F70809F7080A07F8FA0708FAF7F8FAF7F8FB07F90B07F90BF809FB07F9FBF809F
BF809FC08FA0C080A0CF8FAFCF8FAFD090B0D08FAFDF90B0D090B0DF9FBFDF9FBFDF9FC0DF9FC0\
E0A0C0DFA0C0E0AFCFDFAFCFE0B0D0E0B0D0E0BFD0E0BFD0EFBFDFE0BFDFEFC0E0
EFC0E0EFCFE0EFCFE0EFD0EFEFD0EFF0D0F0EFD0EFF0DFF0F0DFF0F0E0FFF0E0FFF0E0FFEFE0F0\
F0EFFFEFE0F0EFEFF0E0EFF0E0EFF0DFEFEFE0F0F0DFEFEFDFF0F0D0F0EFD0F0F0
CFF0EFD0FFEFCFF0EFCFFFEFB0EFDF8FE0BF0000000000000000007F40607F406F8F4F70904F7F\
A0508F9F4F8FA0508FA0508FA05F90A05F90AF6090AF6090AF6F9FAF6F90AF6F9F
AF6F9FB070A0B0709FB07FA0B07F9FB080A0AF809FAF8F9FA0809FAF8F9FA08F9FA0909F9F9090\
A09F9F9F90909F9F90909F9090A0908FA09090AF908FAF8F8FAF9080AF8F80AF8F
7FAF807FB08070AF7F70B07F6FAF706FB07060AF6F60B06F5FB06F5FB06F50AF6050AF604FA05F\
50A05F4F9F4F4F9F4F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F5F5F5F505F605F60605F5F6F5F606F5F60706060705F60
70606F7060607F606F7F606F806F6F80606F8F6F708F6F708F6F708F6F7090707F90707F9F7080\
9F7080A07F8FA07080AF7F8FAF7F8FB08090AF7F90B0809FB08090BF809FBF809F
C08FA0C080A0CF8FAFCF8FAFD08FAFD08FAFDF90B0D090B0DF9FBFDF9FBFDFA0C0DF9FBFDFA0C0\
DFA0C0E0AFCFDFAFCFE0B0D0E0B0CFE0BFD0E0B0D0EFBFDFE0BFDFEFC0E0EFC0E0
EFCFE0EFCFE0F0D0EFEFD0EFF0D0EFEFD0EFF0DFF0F0DFF0F0E0FFF0E0FFFFEFFFF0E0FFF0EFFF\
EFE0F0EFEFFFEFEFF0EFEFF0E0EFF0E0F0F0DFEFF0DFF0F0D0F0EFDFF0F0D0F0EF
D0FFF0CFF0EFCFFFF0C0FFEFB0F0DF80DFB0000000000000804F6F804070904F809F4F80A0508F\
A0508FA05F90A05F90AF5F90A05F90AF609FAF6090AF6F9FAF6F9FB0709FAF709F
B07FA0B07F9FB080A0B07F9FB080A0AF809FAF8F9FA08F9FAF909FA08F9FA0909F9F9090A09F9F\
9F9F909FA09F90A09090A0908FA09090AF908FAF8F8FAF8F80AF8F80B08F7FAF80
7FB08070AF7F70B07F6FAF706FB07060AF6F60B0705FB06F5FB06F50AF6050AF604FA05F4FA050\
4F9F4F4F9F4F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F505F5F505F605F5F60505F605F60605F5F6F5F606F5F60706060
705F6070606F7060607F606F7F606F806F6F80606F8F6F708F6F70906F7F906F7F90707F906F7F\
9F70809F7080A07F8FA0708FAF7F8FAF7F8FB08090B07F90B0809FBF809FBF80A0
BF809FC08FA0C080A0CF8FAFCF8FAFD08FB0D08FB0DF90B0D090B0DF9FBFD09FBFDFA0C0DF9FBF\
DFA0C0DFA0C0E0AFCFDFAFCFE0B0D0E0B0D0E0BFD0E0BFD0E0BFDFE0BFDFEFC0E0
E0C0E0EFCFE0EFCFE0EFD0EFEFD0EFF0D0F0EFD0F0F0DFF0EFDFF0F0E0FFF0E0FFF0EFFFEFE0F0\
EFEFF0EFE0F0EFEFF0E0EFEFE0EFF0DFEFEFDFF0F0D0EFEFDFF0EFD0F0EFD0F0EF
CFF0EFCFF0EFC0F0EFCFFFEFA0EFD080DFB07F4060804F708F4070904F809F4F8FA0508F9F508F\
A05F90A05F8FA05F90A05F90AF6090A06090AF6F9FAF6F90AF709FAF709FB07FA0
AF7F9FB07F9FAF7F9FAF809FA0809FAF8F9FA08F9FA0909F9F8F90A0909F9F90909F9F90909F90\
9FA09090A09090A0908FA08F8FAF9080AF8F80AF8F7FAF807FAF807FAF807FAF80
6FAF7F6FAF7F60AF7060B0705FAF6F5FB06F50AF6050B06050AF5F50AF5F4FA0504FA05040904F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F505F605F60605F5F605F60605F5F6F5F606F5F60706060705F60
7F606F70606F7F606F7F606F806F70806F6F8F6F708F6F7090707F906F7F90707F90707F9F7080\
9F7080A07F8FA0708FAF7F90AF7F8FB08090B07F90BF809FB0809FBF809FBF809F
C08FA0C08FA0CF8FAFCF8FAFD090B0D08FB0DF90B0D090B0DF9FBFDF9FBFDFA0C0DFA0BFE0AFC0\
DFA0C0E0AFCFDFAFCFE0B0D0E0B0D0E0BFD0E0BFD0EFC0DFE0BFDFEFC0E0E0C0E0
EFCFE0EFCFE0EFD0EFEFD0EFF0DFF0EFDFF0F0DFF0F0DFF0F0E0FFF0E0FFF0EFFFF0E0FFF0EFFF\
EFEFF0EFEFFFE0EFF0EFF0F0E0EFF0E0F0F0DFEFEFDFF0F0D0F0EFD0F0F0D0F0EF
D0FFF0CFF0EFCFFFEFC0FFEF7F4F6F80406F8F4F7F904F7FA0508FA0508FA05090A0508FAF5F90\
A05F90AF6090A06090AF6F9FAF6F90AF709FAF6F9FB07F9FB0709FB07FA0AF7F9F
B080A0AF809FAF809FAF809FAF8F9FA08F9FA0909F9F9090A09F9F9F9F909F9F9F909F909FA090\
90A09090AF908FAF8F8FAF9080AF8F80AF8F7FAF8080B0807FAF807FB08070AF7F
70B07F6FAF7060B0705FAF6F60B06F5FB0605FB06050AF5F50AF5F4FA0504FA050409F4F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F505F5F505F605F6060505F6F5F606F5F5F6F5F606F5F60706060
705F607F606F7060607F606F7F606F806F7080606F8F6F708F6F7090707F906F7F9F70809F7080\
A070809F7080A07F8FA07F8FAF7F90AF7F8FB08090B07F90BF809FBF809FC080A0
C080A0C08FA0C08FA0CF8FAFCF8FAFD090B0D08FB0D090B0D090B0DF9FBFD09FBFDFA0C0DFA0C0\
DFAFCFDFA0C0E0AFCFDFAFCFE0B0D0DFB0D0E0BFD0E0BFD0E0C0DFE0C0DFEFC0E0
E0C0E0EFCFEFEFCFE0EFD0EFEFD0EFF0DFF0EFDFF0F0DFF0F0E0F0F0E0FFEFE0F0F0EFFFEFE0F0\
EFEFFFE0EFF0E0EFF0E0EFF0E0F0F0DFEFEFDFF0EFD0EFEFDFF0EFD0F0EFD0F0EF
CFF0EFCFFFEFC0F0E0C0FFEF804070904F7F904F80A0508F9F508FA05F8FA0508FA05F90A05F8F\
AF6090A06090AF6F90AF6F90AF709FAF6F90B0709FAF709FB07F9FAF7F9FAF809F
AF809FAF8F9FA08090A08F9FA08F90A0909F9F90909F9F909F90909F9F90909F9090A0908FA08F\
90AF908FA08F8FAF8F80AF8080AF8F7FAF807FAF8070AF7F70AF7F6FAF706FAF70
60AF7060B0705FAF6F5FB06F50AF6050AF604FA05F50A05F4F9F504F9F5040904F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F505F605F5F60505F6F5F60605F606F5F606F5F6070606F706060
7F606F7060607F6F6F7F606F806F6F80606F8F6F708F6F7090707F8F6F7F9F708090707F9F7080\
9F7080A07F8FA07F8FAF7F90AF7F8FB08090B07F90BF809FB0809FBF8FA0BF80A0
C08FA0C08FA0CF8FAFCF8FAFD090B0D08FAFDF90B0D090B0DF9FBFD09FBFDFA0C0DFA0C0DFAFC0\
DFA0C0E0AFCFDFAFCFE0B0D0DFB0CFE0BFD0E0BFD0E0C0DFE0BFDFEFC0E0E0C0E0
EFCFE0EFCFE0EFD0EFEFD0EFEFDFF0EFDFEFF0E0F0F0DFF0F0E0FFF0E0FFF0EFFFEFE0FFEFEFFF\
EFEFF0EFEFFFE0EFF0E0F0F0DFEFEFE0F0F0DFF0EFDFF0EFD0F0EFD0FFEFCFF0EF
CFFFEFCFF0EFCFFFEF8F4F7F9F4F809F4F8FA0508FA0508FA05F90A05F8FAF6090A06090AF6090\
A06090AF6F9FAF6F90AF709FAF709FB07F9FAF7F9FB07F9FAF7F9FAF809FAF809F
AF8F9FA08F9FAF8F9FA08F90A0909F9F90909F9F9F9F9F909FA090909F9090A0908FA08F90AF90\
8FAF8F8FAF8F80AF8080AF807FAF807FB08070AF7F70AF7F6FAF706FB07060AF70
60B0705FAF6F5FB06F50AF6050AF604FA05F50A05F4F9F504F9F5040904F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000050505F5F5F5F5F505F605F5F60505F6F5F60605F5F6F5F606F5F60706060
705F607F606F7F60607F606F7F606F806F70806F708F6F708F6F7090707F906F7F9F70809F7080\
9F7080A07080A07F8FA07F8FAF7F90AF7F90B0809FB07F90BF809FBF809FC08FA0
BF80A0C08FA0C08FA0CF8FAFCF8FAFD090B0D090B0D09FB0D090B0D09FBFD09FBFDFA0C0D0A0C0\
DFAFC0DFAFC0DFB0CFDFAFCFE0B0D0DFB0D0E0BFDFE0BFD0E0C0DFE0C0DFEFCFE0
E0C0E0EFCFE0EFCFE0EFD0EFEFD0EFEFDFF0EFDFF0F0E0FFEFDFF0F0E0FFEFE0F0EFEFFFE0E0F0\
EFEFF0E0EFF0E0EFF0DFEFEFDFEFF0DFEFEFDFF0EFD0EFEFD0F0EFCFF0EFD0F0EF
CFF0E0CFFFEF904F7F9F4F8F9F4F80A0508F9F508FA05F8FA05F8FA06090A06090AF6F90A06090\
AF6F90AF6F90AF709FAF7090B07F9FAF7F9FAF809FA07F90AF809FA08090AF8F9F
A08090A08F9F9F8F90A090909F90909F9F90909F9090A090909F8F90A0908FA08F8FAF8F80A08F\
80AF8F7FAF807FAF8070AF7F70AF7F70AF7F70AF7F6FAF706FAF7060AF6F60B06F
5FAF605FB06050AF5F50AF5F4FA0504FA0504F9F4F4F9F4F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F5F5F5F505F605F60605F5F6F5F606F5F606F60606F5F6070606F706060
7F606F7F606F7F6F6F7F606F806F70806F708F6F708F6F7090707F90707F9F70809F7080A07F8F\
9F7080A07F8FA07F8FAF8090AF7F90B08090B08090BF809FBF809FBF8FA0BF80A0
C08FA0C08FA0CF8FAFCF8FAFD090B0D090B0D090B0D090B0DF9FBFD09FBFDFA0C0DFA0C0DFAFCF\
DFAFC0DFB0CFDFAFCFE0B0D0E0B0D0E0BFD0E0BFD0E0C0DFE0C0DFEFCFE0E0C0E0
EFCFEFEFCFE0EFD0EFEFD0EFF0DFF0EFDFF0F0E0F0EFE0F0F0EFFFF0E0FFF0EFFFEFEFF0EFEFFF\
E0EFF0EFEFF0E0EFF0E0F0F0DFEFF0DFF0F0D0F0EFDFF0F0D0F0EFD0FFEFCFF0EF
CFFFEF9F4F80A0508F9F508FA05F8FA05F8FA05F90A05F8FAF6090A06090AF6F90AF6F90AF709F\
AF709FAF709FAF709FB07FA0AF7F9FAF809FAF809FAF8F9FA0809FA08F9FA08F90
A0909F9F9090A0909F9F90909F9F90909F9090A09090A08F90AF908FAF8F8FAF8F80AF8F8FAF8F\
80AF807FAF8070AF7F7FB08070AF7F70B07F6FAF706FB07060AF6F60B06F5FAF60
5FB06050AF6050AF5F4FA05050A05F4F9F504F9F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F505F5F5F5F5F505F605F60605F5F6F5F606F5F607060606F5F6070606F
7060607F606F7F606F806F6F80606F806F70806F708F6F7F8F6F7F90707F906F7F9F70809F7080\
A07F8FA07080A07F8FA07F8FAF8090AF7F90B08090B08090BF809FBF809FC08FA0
C080A0CF8FAFC08FA0CF8FAFCF8FAFD090B0D090B0D09FB0D09FB0DFA0BFD09FBFDFA0C0D0A0C0\
DFAFCFDFAFC0DFB0CFDFB0CFE0BFD0DFB0D0E0BFDFE0BFDFE0C0DFE0C0DFE0CFE0
E0CFE0EFD0EFE0CFE0EFD0EFEFD0EFEFDFF0EFDFF0F0E0FFEFE0FFF0EFFFEFE0F0EFEFFFE0EFF0\
EFEFF0E0EFF0E0EFF0DFEFEFDFF0F0D0EFEFDFF0EFD0F0EFD0F0EFCFF0E0CFF0EF
CFF0E0A0508F9F508FA05F8F9F5F8FA05F90A05F8FA06090A06090AF6F90A06F90AF709FAF7090\
AF7F9FAF709FAF7F9FAF7F9FAF809FA08090A0809FA08090A08F9F9F8F90A09090
9F90909F90909090909F9F90909F8F90A0908FA08F8FA08F8FA08F8FAF8F80A08080AF807FAF80\
7FAF8070AF7F70AF7F6FAF706FAF7060AF6F60AF705FAF6F5FB06F50AF6050AF60
50A05F50A05F4F9F504F9F5040904F4F904F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F5F5F5F505F605F60605F5F6F5F606F5F606F60606F5F6070606F706060
7F606F7F606F806F6F7F606F806F70806F708F707F8F6F7090707F90707F9F70809F70809F7F80\
9F7080A07F8FA07F8FAF7F90AF7F90B08090B08090BF809FBF809FC08FA0BF80A0
C08FA0C08FA0CF90AFCF8FAFD090B0D090AFD09FB0D09FB0D09FBFD09FBFDFA0C0D0A0C0DFAFC0\
DFAFC0DFB0CFDFB0CFE0B0D0DFB0D0E0BFDFE0BFD0E0C0DFE0C0DFEFCFE0E0CFE0
EFCFE0E0CFE0EFD0EFEFD0EFEFDFF0EFDFF0F0E0FFEFE0F0F0EFFFEFE0FFF0EFFFEFEFF0EFEFFF\
E0EFF0E0F0F0DFEFF0E0F0F0DFEFEFDFF0F0D0F0EFD0F0EFD0F0EFD0FFEFCFF0E0
A0508F9F508FA05F90A05F8FA06090A0608FAF6F90A06F90AF6F90AF6F90AF709FAF709FAF7F9F\
AF7F9FAF809FAF7F9FAF809FA08090AF8F9FA08F90A08F9F9F8F90A0909F9F9090
9F9F909090909F9F90909F9090A0908FA08F90AF908FAF8F8FAF8F80AF8080AF8F7FAF807FAF80\
70AF7F70AF7F6FAF706FAF7060AF6F60B0705FAF6F5FB06F50AF605FAF6050A05F
50A05F4F9F504F9F504F904F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F505F605F5F5F505F605F60605F5F6F5F606F5F60706060705F6070606F
7060607F606F7F606F806F6F7F606F806F70806F708F707F8F6F7090707F90707F9F70809F7080\
A07F80A07080AF7F8FA07F8FAF7F90AF7F90B0809FB08090BF809FBF809FC08FA0
BF80A0C08FA0C08FA0CF90AFCF8FAFD090B0CF90B0D09FBFD09FB0D0A0BFD0A0BFDFA0C0D0A0C0\
DFAFCFDFAFCFDFB0CFDFB0CFDFBFD0DFBFD0E0C0DFDFBFD0E0C0DFE0C0DFE0CFE0
E0CFE0EFD0EFE0D0EFEFDFEFE0D0EFEFDFF0EFDFF0F0E0FFEFE0F0EFEFFFEFE0F0EFEFF0E0E0F0\
E0EFF0DFEFF0E0F0F0DFEFEFDFF0F0D0EFEFD0F0EFCFF0E0D0F0EFCFF0E0CFFFEF
9F508FA05F8F9F5F8FA06090A0608FA06F90A06F90AF6F90A06F90AF709FAF7090AF7F9FAF7F90\
AF7F9FA07F90AF809FA08090A08F909F8F90A08F909F8F909F90909F90909F9F90
90908F909F90909F8F90A0908FA08F8FAF8F80A08F80AF8F7FA0807FAF807FA07F7FAF7F70A070\
70AF7F6FAF706FAF7060AF6F60AF6F5FAF605FAF6050AF5F50AF5F4FA05F50A05F
4F9F504F9F5040904F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F505F605F5F5F505F605F60605F606F5F606F5F6070606F70606070606F706060
7F6F6F7F606F806F6F7F606F806F70806F708F707F8F6F7090707F90707F9F70809F7080A07F8F\
A07F80A07F8FA07F8FAF8090AF7F90B08090B08090BF8F9FBF809FC08FA0BF8FA0
C08FA0C08FA0CF90AFCF8FAFD090B0D090B0D09FB0D09FB0DFA0BFD0A0BFDFA0C0D0A0C0DFAFCF\
DFAFCFDFB0CFDFB0CFE0BFD0DFBFD0E0BFDFDFBFDFE0C0DFE0C0DFE0CFE0E0CFE0
EFD0EFE0D0EFEFD0EFEFD0EFEFDFF0EFDFF0EFE0FFEFE0F0F0EFFFEFEFFFEFEFFFE0EFF0EFEFFF\
E0EFF0E0F0F0DFEFF0DFF0F0DFF0EFDFF0F0D0F0EFD0FFEFCFF0EFCFFFEFCFF0E0
A06090A0608FA06090A06090AF6F90A06F90AF709FAF7090AF7F9FAF7F9FAF7F9FAF7F9FAF809F\
A0809FAF809FA08090A08F9FA08F90A090909F90909F9F909F90909F9F90909F8F
9FA09090A08F90A0908FA08F8FAF8F80AF8F80AF8F7FAF8080AF807FAF7F7FAF7F70AF7F70AF7F\
6FAF706FAF7060AF6F60B06F5FAF605FB06050AF6050AF6050A05F50A05F4F9F50
4F9F5040904F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F505F5F505F605F5F605F5F605F60605F5F6F5F606F5F60706060705F6070606F
7060607F6F6F7F606F806F70806F6F8F6F70806F708F707F8F6F7F90707F90707F9F70809F7080\
A07F8FA07080A07F8FA07F8FAF8090AF7F90B0809FB08090BF8F9FBF809FC08FA0
C08FA0CF8FAFC08FA0CF90AFCF90AFD090B0CF90B0D09FBFD09FBFD0A0BFD0A0BFD0AFC0D0AFC0\
DFAFCFD0AFCFDFB0CFDFB0CFDFBFD0DFBFD0E0C0DFDFC0DFE0CFE0E0C0DFE0CFE0
E0CFE0EFD0EFE0D0EFEFDFEFE0DFEFEFE0F0EFDFF0EFE0FFEFE0FFEFEFFFE0E0F0EFEFF0E0EFF0\
E0EFF0DFEFEFDFEFF0DFEFEFDFF0EFD0EFEFD0F0EFCFF0EFD0F0EFCFF0E0A0608F
A0608FA06090A0608FA06F90A06F90AF709FA07090AF7F9FAF7F90AF7F9FA07F90AF809FA08090\
A0809F9F8090A08F909F8F909F90909F8F8F9F909090908F909F90909F8F90A090
8F9F8F90A08F8FA08F8FAF8F80A08080AF807FA07F7FAF8070A07F70AF7F6FA0706FAF7060AF70\
60AF705FAF6F5FAF6F5FAF605FAF6050A05F50A05F4FA0504FA0504F904F4F904F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F505F605F60605F5F605F60605F5F6F60606F5F60706060705F6070606F70606F
7F606F7F606F806F70806F6F806F70806F708F707F8F6F7F90707F90707F9F7F809F70809F7F80\
9F7F80A07F8FA07F8FAF8090AF7F90B08090B08090BF809FBF809FC08FA0C08FA0
C08FA0C08FA0CF90AFCF8FAFD090B0CF90B0D09FB0D09FB0D0A0BFD0A0BFDFAFC0D0AFC0DFAFCF\
D0AFC0DFB0CFDFB0CFDFBFD0DFBFD0E0C0DFDFC0DFE0C0DFDFC0DFE0CFE0E0CFE0
E0D0EFE0D0EFEFDFF0E0DFEFEFDFF0EFDFF0EFE0FFEFE0FFEFEFFFEFE0F0EFEFFFE0EFF0E0EFF0\
DFEFEFE0F0F0DFEFEFDFF0F0D0EFEFDFF0EFD0F0EFD0F0EFCFF0E0CFFFEFA0608F
A06F90A06F90AF7090A06F90AF709FAF7090AF7F9FAF7F9FAF809FA07F90AF809FA08090A08F9F\
A08090A08F9F9F8F909F90909F90909F9F90909F8F909F90909F8F90A0908FA08F
8FAF8F8FA08F8FAF8F80A08080AF807FA07F7FAF8070AF7F70AF7F6FAF706FAF7060AF7060AF70\
5FAF6F60B06F5FAF605FAF6050A05F50A05F4F9F504F9F504F904F4F9F4F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F505F5F505F605F6060505F605F60605F5F6F60606F5F607060607060607F606F
7060607F606F7F606F806F70806F6F8F6F708F6F708F707F8F6F7F90707F90707F9F7F809F7080\
A07F8FA07F8FAF7F8FA07F8FAF8090AF7F90B0809FB08090BF809FBF809FC08FA0
C08FA0C08FA0C08FA0CF90AFCF90AFCF90B0CF90B0D09FBFCF9FB0D0A0BFD0A0BFD0AFC0D0AFC0\
DFAFCFD0B0CFDFB0D0D0B0CFDFBFD0DFBFD0E0C0DFDFC0DFE0CFE0DFCFDFE0CFE0
E0CFE0E0D0EFE0D0EFEFDFF0E0DFEFEFE0F0EFE0F0EFE0FFE0E0F0EFEFFFE0E0F0E0EFF0DFEFEF\
E0EFF0DFEFEFDFF0EFD0EFEFD0F0EFD0F0EFD0F0EFCFF0E0CFF0E0C0F0E0A06F90
A06F8FA06F90A06F90AF709FA07090AF7F9FA07F90AF809FA07F90A08090A08090A08F909F8090\
9F8F909F8F909F909090908F90909090908F909F908F9F8F8FA08F8FA08F8FAF8F
80A08080AF807FA08080AF807FA07F7FAF7F70A07F70AF7F6FA0706FAF7060AF6F60AF6F5FAF60\
5FAF6050AF6050AF604FA05F50A05F4F9F504F9F5040904F40904F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5F5F\
5F505F605F60605F5F6F5F60605F606F60606F5F6070606F7060607F606F70606F
7F6F6F7F606F806F70806F708F70708F6F708F707F8F707F90708090707F9F7F809F7080A07F8F\
A07F8FA07F8FA07F8FAF8090AF8090B08090B08090BF8F9FBF809FC08FA0BF8FA0
C090AFC08FA0CF90AFCF90AFD09FB0CF90B0D09FB0D09FB0D0A0BFD0A0BFD0AFC0D0AFC0DFB0CF\
D0B0CFDFB0CFDFB0CFDFBFD0DFBFD0E0C0DFDFC0DFE0CFE0E0CFE0E0CFE0E0CFE0
E0D0EFE0D0EFEFDFF0E0DFEFEFE0F0EFE0F0EFE0FFEFE0FFEFEFFFE0EFF0EFEFFFE0EFF0E0F0F0\
DFEFF0DFF0F0DFEFEFDFF0EFD0F0EFD0F0EFCFF0E0D0FFEFCFF0E0A06F90A06F90
AF709FAF7090AF7F9FAF7F90AF7F9FA07F90AF809FA08090A08F9FA08090A08F909F8F90A09090\
9F8F909F909090908F9F9F90909F8F90A0908F9F8F90A0908FA08F8FAF8F80A080
80AF8F80A08080AF807FA07F7FAF8070AF7F70AF7F6FAF706FAF7060AF6F60AF6F5FAF605FB06F\
50AF6050AF6050A05F50A05F4F9F504F9F5040904F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F505F\
5F5F5F5F505F605F60605F5F6F5F606F5F606F60606F5F6070606F7060607F606F
7F606F7F6F6F7F606F806F70806F708F707F8F6F7090707F90707F90708090707F9F7F809F7080\
A07F8FA07F8FAF7F8FAF7F8FB08090AF8090B0809FB0809FBF8F9FBF809FC08FA0
C08FA0CF90AFC08FA0CF90AFCF90AFD09FB0CF9FB0D09FBFCF9FBFD0A0BFD0A0BFD0AFC0D0AFC0\
DFB0CFD0B0CFDFBFD0DFBFCFDFBFD0DFBFD0DFC0DFDFC0DFE0CFE0DFCFE0E0D0E0
E0D0E0E0D0EFE0D0EFE0DFF0E0DFF0EFE0F0E0E0F0EFEFFFE0E0F0EFEFFFE0E0F0E0EFF0DFEFEF\
DFEFF0DFEFEFDFF0EFD0EFEFD0F0EFCFF0E0D0F0EFCFF0E0CFFFE0A06F90AF7090
A07090AF7F9FA07090AF7F90A07F90A08090A08090A080909F8090A08F909F8F909F8F909F8F8F\
9F909090908F909F90909F8F90A08F8F9F8F8FA08F8FA08F8FAF8F80A08080AF80
7FA07F7FAF8070A07F70AF7F70A07070AF706FA0706FAF7060AF6F60AF6F5FAF605FAF6050A05F\
50AF604F9F504FA0504F9F504F9F5040904F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5F5F\
5F505F605F60605F5F6F5F60605F606F60606F5F6070606F7060607F606F70606F
7F6F6F7F606F806F70806F708F6F708F6F7090707F8F707F90708090707F9F7F809F7080A07F8F\
A07F8FAF8090A07F8FAF8090AF8090B0809FB0809FBF8F9FBF809FC08FA0C08FA0
C08FA0C08FA0CF90AFCF90AFCF9FB0CF90B0D09FBFCF9FB0D0A0BFD0A0BFD0AFC0D0AFC0DFB0CF\
D0B0C0DFBFCFDFB0CFDFBFD0DFBFD0DFC0DFDFC0DFE0CFE0DFCFDFE0D0E0E0CFE0
E0D0EFE0D0EFEFDFF0E0DFF0EFE0F0E0E0F0EFEFFFEFE0FFEFEFFFE0EFF0E0EFF0E0EFF0E0EFF0\
DFEFEFDFF0F0D0EFEFDFF0EFD0F0E0D0F0EFCFF0E0CFFFE0C0F0E0AF709FAF7090
AF7F9FA07F90AF809FA07F90A08090A08090A08F909F8F90A08F909F8F909F90909F90909F9090\
90908F909F90909F8F90A08F8FA08F90A08F8FA08F8FAF8F80A08080AF807FA080
7FAF8070A07F70AF7F6FAF706FAF706FAF706FAF7060AF6F60AF6F5FAF605FAF6050A05F50A05F\
4F9F5050A05F4F9F504F9F5040904F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F505F\
5F5F5F5F505F605F60605F5F6F5F606F5F606F60606F5F6070606F7060607F606F
70606F7F6F6F7F6F6F806F70806F708F6F7F8F6F7090707F8F707F9070809070809F7F809F7080\
A07F8FA07F8FAF8090AF7F8FAF8090AF8090B0809FB0809FBF8F9FBF809FC08FA0
C08FA0CF8FA0C08FA0CF90AFCF90AFCF9FB0CF9FB0D0A0BFCF9FB0D0A0BFD0A0BFD0AFC0D0AFC0\
D0B0CFD0B0CFDFBFD0D0BFCFDFBFD0DFBFD0DFC0DFDFC0DFDFCFE0DFCFE0E0D0E0
DFD0E0E0D0EFE0D0EFE0DFF0E0DFF0EFE0F0E0E0F0EFEFFFE0E0F0E0EFF0DFEFF0E0EFF0DFEFEF\
DFEFEFD0EFEFDFF0EFD0EFE0D0F0EFCFF0E0CFF0E0C0F0E0AF7090A07090AF7F90
A07F90A07F90A07F90A080909F80909F8F909F80909F8F90908F8F9F909090908F9F909090908F\
909F8F8F9F8F90A08F8F9F8F8FA08F80A08080A08F7FA08080A0807FA07F7FAF7F
70A07070AF7F6FA0706FAF7060A06F60AF6F5FA06F5FAF6F50A06050AF6050A05F50A05F4F9F50\
4F9F504F904F4F904F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005F505F605F60\
5F5F5F605F60605F5F6F60606F5F606F60606F5F6070606F70606F7F6F6F70606F
7F6F707F6F6F806F70806F708F707F8F6F7F90707F8F707F907F809070809F7F809F7F80A07F8F\
A07F8FAF8090AF7F8FB08090AF8090B0809FB0809FBF8F9FBF8F9FC08FA0C08FA0
CF90AFC08FA0CF90AFCF90AFCF9FB0CF9FB0D0A0B0CF9FB0D0A0BFD0A0BFD0AFC0D0AFC0DFB0CF\
D0B0CFDFBFCFD0BFCFDFC0D0DFBFD0DFC0DFDFC0DFE0CFE0DFCFE0E0D0E0DFD0E0
E0DFEFE0D0EFE0DFF0E0DFF0EFE0FFE0E0F0EFEFFFE0E0F0EFEFFFE0EFF0E0EFF0DFEFF0E0F0F0\
DFEFEFDFF0EFD0EFEFD0F0EFD0F0E0D0F0EFCFF0E0CFFFE0A07F90AF7F9FA07F90
AF809FA08090A080909F8090A08F909F8F909F90909F8F8F9F909090908F9F9F9090908F909F90\
909F8F90A08F8FA08F8FAF8F80A08080AF8F80A08080AF807FA07F7FAF7F70A07F
70AF7F6FAF706FAF7060AF6F60AF7060AF6F5FAF6F50AF605FAF6050A05F50A05F4F9F50509F50\
4F904F4F904F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005F505F5F505F\
605F5F5F505F605F60605F5F6F60606F5F607060606F606070606F7060607F606F
7F606F806F707F6F6F806F70806F708F707F8F6F7F90707F90707F9F7F809070809F7F809F7F80\
A07F8FA07F8FAF8090AF7F90B08090AF8090B0809FB0809FBF8F9FBF8F9FC08FA0
C08FA0CF90AFC08FA0CF90AFCF90AFCF9FB0CF9FB0CFA0B0CFA0B0D0AFBFCFA0BFD0AFC0D0AFC0\
D0B0CFD0B0CFDFBFCFD0BFCFDFC0D0D0C0D0DFC0DFDFC0DFE0CFE0DFCFE0E0D0E0
DFD0E0E0DFEFE0DFEFE0DFF0E0DFF0EFE0FFE0E0F0E0EFFFE0E0F0E0EFF0DFEFEFDFEFF0DFEFEF\
DFF0EFD0EFEFD0F0EFD0EFE0D0F0EFCFF0E0CFF0E0C0F0E0A07F90A07F90A08090
9F8090A080909F808F9F8F909F8F8F9F9090908F8F9F909090908F909F9090908F909F8F8F9F8F\
8FA08F8FA0808FA08F80A08080AF807FA07F7FA0807FA07F7FAF7F70A07070AF70
6FA0706FAF7060A06F60AF6F5FA0605FAF6F50A05F50A06050A05F50A05F4F9F504F9F5040904F\
4F904F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005F505F605F5F\
5F505F605F60605F606F5F606F5F6070606F6F606070606F7060607F6F6F7F606F
7F6F707F6F6F806F70806F708F707F8F6F7090707F90707F9070809070809F7F809F7F80A07F8F\
A07F8FAF8090AF7F8FAF8090AF8090B08F9FB0809FBF8F9FBF8F9FC08FA0C08FA0
C090A0C08FA0CF90AFCF90AFCF9FB0CF9FAFD0A0B0CFA0B0D0AFBFCFA0BFD0AFC0D0AFC0D0B0C0\
D0B0C0DFBFCFD0BFCFDFC0D0D0BFD0DFC0DFDFC0DFDFCFDFDFCFDFE0D0E0DFD0E0
E0DFEFE0D0EFE0DFF0E0DFF0E0E0F0E0E0F0EFEFFFE0EFF0E0EFFFDFEFF0E0EFF0DFEFF0DFF0F0\
D0EFEFDFF0EFD0F0EFD0F0EFCFF0E0D0F0E0CFF0E0CFFFE0A07F90A08090A08090
A08F909F80909F8F909F8F909F909090908F9F909090908F909F90909F8F90A08F8F9F8F8FA08F\
8FA08F8FAF8F80A08080AF807FA07F7FAF807FA07F7FAF7F70A07070AF7F6FA070
6FAF7060A06F60AF6F5FAF605FAF6050A06050AF6050A05F50A05F4F9F504F9F5040904F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005F5F5F5F505F\
605F5F605F5F605F60605F5F6F5F606F5F6070606F6F5F6070606F70606F7F6F6F
7F606F7F6F707F6F6F806F70806F708F707F8F707F90707F90707F9070809070809F7F8F9F7F80\
A07F8FA07F8FAF8090AF7F8FAF8090AF8090B08F9FB08090BF8F9FBF8F9FC08FA0
C08FA0C090A0C090A0CF9FAFC090AFCF9FB0CF9FB0CFA0B0CFA0B0CFAFBFCFAFBFD0B0C0CFAFC0\
D0B0C0D0B0C0D0BFCFD0BFCFDFC0D0D0C0D0DFCFDFDFC0DFDFCFDFDFCFDFE0D0E0
DFD0E0E0DFEFDFDFEFE0E0F0E0DFF0E0E0FFE0E0F0E0EFFFDFE0F0DFEFF0DFEFEFDFEFF0D0EFEF\
D0F0EFD0EFEFD0F0EFCFEFE0CFF0E0CFF0E0CFF0E0C0F0DFA080909F8090A08F90
9F808F9F8F90908F8F9F9090908F8F90908F90908F909F8F8F9F8F8F9F8F8F9F8F8FA08F80A080\
80A08080A08080A0807FA07F7FA07F70A07F70AF7F6FA0706FAF7060A06F6FAF70
5FA06F5FAF6F5FA0605FAF6050A05F50A05F4F9F50509F504F904F4F904F408F4F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F5F5F5F505F605F60\
605F5F605F60605F606F60606F5F6070606F6F606070606F70606F7F6F6F7F606F
806F707F6F70806F70806F708F707F8F707F90707F8F707F907F809070809F7F809F7F80A0808F\
A07F8FAF8090AF7F8FB08090AF8090B08F9FB08090BF8F9FBF8F9FC08FA0C08FA0
CF90AFC090A0CF90AFC090AFCF9FB0CF9FB0CFA0B0CFA0B0D0AFBFCFAFBFD0B0BFCFAFBFD0B0C0\
D0B0C0D0BFCFD0BFCFDFC0D0D0C0D0DFCFDFDFC0D0DFCFDFDFCFDFE0D0E0DFD0E0
E0DFEFE0DFEFE0E0F0E0E0EFE0EFF0E0E0F0E0EFFFE0EFF0E0EFFFDFEFF0DFEFF0DFEFEFDFF0F0\
D0EFEFD0F0EFCFF0E0D0F0EFCFF0E0CFF0E0C0F0E0CFFFE09F8090A08F909F8F90
9F90909F8F8F9F909090908F9F9F90909F8F909F8F8F9F8F90A08F8FA08F8FA08F80A0808FAF8F\
80A08080AF807FA07F7FAF8070A07F70AF7F6FA07070AF706FA0706FAF7060A06F
60AF6F5FAF605FAF6050A05F50A05F509F5F509F504F90504F9F5040904F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000050505F5F5F5F5F505F\
605F60605F5F605F60605F606F60606F5F6070606F70606070606F70606F7F6F6F
7F6F6F806F707F6F70806F70806F708F707F8F707F90707F90707F9F7F809070809F7F809F7F80\
A0808FA07F8FAF8090AF808FB08090AF8090B08F9FB0809FBF8F9FBF8F9FC08FA0
C08FA0C090AFC090A0CF9FAFC09FAFCFA0B0C09FAFCFA0B0CFA0B0D0AFBFCFAFBFD0B0C0CFB0C0\
D0BFC0D0B0C0D0BFCFD0BFCFDFC0D0D0C0D0DFCFDFDFCFDFDFD0DFDFCFDFDFD0E0
DFD0E0E0DFEFDFDFEFE0E0F0E0E0F0E0EFF0DFE0F0E0EFF0DFEFF0DFEFF0D0EFEFDFEFEFD0EFEF\
D0F0EFCFEFE0D0F0EFCFF0E0CFF0E0C0F0DFC0F0E0C0F0DF9F8F909F808F9F8F90
908F8F9F909090908F90908F90908F909F8F8F9F8F8FA08F8FA0808FA08F80A08080A0807FA080\
7FA08070A07F7FA07F70A07070AF7F6FA0706FAF7060A06F60AF6F5FA0605FAF6F
50A0605FAF6050A05F50A05F4F9F504F9F5040904F4F904F408F40000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F5F5F5F505F605F60\
605F5F6F5F60605F606F60606F5F6070606F70606070606F70606F7F6F6F7F6F6F
7F6F707F6F70807070806F708F707F8F707F90707F90707F907F809070809F7F809F7F80A07F8F\
A07F8FAF8090AF808FAF8090AF8090B08F9FB08090BF8F9FBF8F9FC090A0C08FA0
C090A0C090A0CF9FAFC09FAFCF9FB0C09FAFCFA0B0CFA0B0CFAFBFCFAFBFD0B0C0CFB0BFD0B0C0\
D0B0C0D0BFCFD0BFCFD0C0CFD0C0CFDFCFD0DFCFD0DFCFDFDFCFDFDFD0E0DFD0E0
E0DFEFDFDFE0E0E0EFE0E0EFE0EFF0E0E0F0E0EFF0DFEFF0DFEFF0DFEFEFDFF0F0D0EFEFD0F0EF\
D0EFEFD0F0EFCFF0E0CFF0E0CFF0E0CFFFE0C0F0DFC0FFE09F8F8F9F9090908F8F
9F909090908F909F90909F8F909F8F8F9F8F8FA08F8FA08F8FA08F80A08080AF807FA0807FA080\
70A07F7FAF7F70A07F70AF7F6FA0706FAF7060A06F60AF6F5FA06F5FAF6F5FAF60
5FAF6050A05F50A05F4F9F504F9F504F904F4F904F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F505F5F5F5F5F505F\
605F60605F5F6F5F60605F606F60606F5F6070606F6F606070606F70606F7F6F6F
7F606F7F6F707F6F7080707F806F708F707F8F707F90708090707F907F809070809F7F8F9F7F80\
A07F8FA07F8FAF8090AF7F8FAF8090AF8090B08F9FB08090BF8F9FBF8F9FC08FA0
BF8FA0C090A0C090A0C09FAFC09FAFCFA0B0C0A0B0CFA0B0CFA0B0CFAFBFCFAFBFD0B0C0CFB0BF\
D0BFC0CFBFC0D0C0CFD0BFCFD0C0D0D0C0D0DFCFD0D0CFD0DFD0DFDFD0DFDFDFE0
DFD0E0DFDFEFDFDFEFE0E0EFDFE0EFE0EFF0DFE0EFDFEFF0DFE0EFDFEFEFD0EFEFD0F0EFCFEFE0\
D0F0EFCFEFE0CFF0E0C0F0E0CFF0E0C0F0DFC0FFE0BFF0DF9F8F8F908F8F90908F
90908F909F8F8F908F909F8F8F9F808FA08F80A0808FA08F80A08080A0807FA07F7FA07F70A07F\
70A07F6FA07070A0706FA06F6FA07060A06F60AF6F5FA0605FAF6050A06050A060
4F9F5F509F5F4F90504F9F50408F4F40904F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F5F5F5F505F605F60\
605F606F5F60605F606F606F6F606070606F6F6060706F6F70606F7F6F707F6F6F
806F707F6F7080707F806F708F707F8F707F90708090707F9F7F80907F809F7F8F9F7F80A0808F\
A07F8FAF8090A0808FAF8090AF8090B08F9FB08090BF8F9FBF8F9FC090A0C08FA0
C090A0C090A0C09FAFC09FAFCFA0B0C0A0AFCFA0B0CFA0B0CFAFBFCFAFBFD0B0BFCFB0BFD0BFC0\
D0BFC0D0BFCFD0BFC0D0C0CFD0C0CFDFCFD0D0CFD0DFD0DFDFD0DFDFDFE0DFD0DF
E0DFEFDFDFE0E0E0EFDFE0EFE0EFF0DFEFEFE0EFF0DFEFEFDFEFF0DFEFEFDFF0EFD0EFEFD0F0EF\
D0F0E0D0F0EFCFF0E0CFF0E0C0F0E0CFFFE0C0F0DFC0FFE090908F9F909090908F
909F90909F8F90A08F8F9F8F8FA08F8FA08F8FA08F80A08080AF807FA07F7FA08070A07F70AF7F\
70A07070AF706FA0706FAF7060A06F60AF6F5FA0605FAF6F50A06050A060509F5F
50A05F4F9F504F9F5040904F4F904F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F5F5F5F505F5F5F5F5F5F5F\
605F60605F5F6F5F606F5F606F606F6F606070606F70606F706F6F70606F7F6F70
7F6F6F806F707F6F7080707F8070708F707F8F707F90708090707F9F7F80907F809F7F8F9F7F80\
A0808FA07F8FAF8090AF808FB08090AF8090B08F9FB08F9FBF8F9FBF8F9FC090A0
BF90A0C090AFC090A0C09FAFC09FAFCFA0B0C0A0AFCFAFB0CFAFB0CFB0BFCFAFBFCFB0BFCFB0BF\
D0BFC0CFBFC0D0C0CFD0C0CFD0CFCFD0C0CFD0CFD0D0CFD0DFD0DFD0D0DFDFDFE0
DFDFE0DFE0EFDFDFE0E0E0EFDFE0EFE0EFF0DFE0EFDFEFEFD0EFEFDFEFEFD0EFE0D0EFEFCFEFE0\
D0F0EFCFEFE0CFF0E0C0F0DFCFF0E0C0F0DFC0FFDFBFF0DF90908F8F908F909F8F
8F9F8F8FA08F8F9F808FA08F80A08080A0807FA08080A0807FA07F7FA07F70A07070A07F6FA070\
6FA07060A06F60A06F60A06F60AF6F5FA0605FAF6050A05F50A05F4F9F504F9F50
4F90504F9050408F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F505F605F605F5F5F605F60\
605F5F6F60606F5F606F60606F606070606F6F606F706F6F70606F7F6F707F6F6F
7F6F707F6F7080707F8070708F707F8F707F907F8090707F907F80907F809F7F8F9F7F80A0808F\
A07F8FAF8090AF808FAF8090AF8090B08F9FB08F90BF8F9FBF8F9FC090A0BF8FA0
C090A0C090A0C09FAFC09FAFCFA0AFC0A0AFCFAFB0CFAFB0CFAFBFCFAFBFCFB0BFCFB0BFD0BFC0\
CFBFC0D0C0CFD0C0CFD0C0CFD0C0CFDFCFD0D0CFD0DFD0DFD0D0D0DFDFDFDFDFDF
DFDFE0DFDFE0E0E0EFDFE0EFE0EFEFDFE0EFDFEFEFDFEFEFDFEFEFD0EFEFD0F0EFD0EFE0D0F0EF\
CFEFE0CFF0E0CFF0E0CFF0E0C0F0DFC0FFE0C0F0DFC0FFDF90908F909F8F8F9F8F
8FA08F8F9F8F8FA08F80A08080A08080A08080A0807FA07F7FA07F70A07F70AF7F6FA0706FAF70\
60A06F6FAF7060A06F60AF6F5FA0605FAF6050A05F50A05F4F9F50509F5F4F9050
4F9050408F4F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F505F5F505F605F5F5F505F\
605F60605F606F6060605F606F60606F606070606F6F6060706F6F70606F7F6F70
7F6F6F806F70806F7080707F806F708F707F8F707F907F8090707F9F7F809F7F809F7F8F9F7F80\
A0808FA07F8FAF8090AF808FAF8090AF8090B08F9FB08090BF8F9FBF8F9FC090A0
BF909FC090A0C090A0C09FAFC09FAFC0A0B0C0A0AFCFAFB0C0AFB0CFB0BFCFB0BFCFB0C0CFB0BF\
CFBFC0CFBFC0D0C0CFCFC0C0D0CFCFD0CFCFD0CFD0D0CFD0D0D0DFD0D0DFDFDFDF
D0DFDFDFE0E0DFE0E0DFEFEFDFE0E0DFEFEFD0E0E0DFEFEFD0EFE0D0EFEFCFEFE0D0F0E0CFEFE0\
CFF0E0C0F0DFCFF0E0C0F0DFC0F0DFBFF0DFC0FFDFBFF0D0909F8F8F9F8F8FA08F
809F808FA08F80A08080A0807FA07F7FA0807FA07F7FA07F70A07070A0706FA0706FA07060A06F\
60A06F5FA0605FAF6F50A0605FA06050A05F50A05F4F9F504F9F504F904F4F904F
408F40000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F505F605F605F5F5F605F60\
605F606F6060605F606F606F6F606070606F6F606F706F6F706F6F7F6F707F6F70
8070707F6F7080707F80707F8F707F8F707F907F8090707F9F7F80907F809F7F8F9F7F80A0808F\
A0808FAF8090A0808FAF8F90AF8090B08F9FB08F90BF8F9FBF8F9FC090A0BF909F
C09FA0C090A0C09FAFC09FAFCFA0B0C0A0AFCFAFB0C0AFB0CFB0BFCFB0BFCFB0BFCFB0BFD0BFC0\
CFBFC0D0C0CFCFC0C0D0CFCFD0CFCFD0CFD0D0CFD0DFD0DFD0D0D0DFDFDFDFDFDF
DFE0E0DFE0E0DFE0EFDFE0EFDFEFEFDFEFEFDFEFEFD0EFE0DFF0EFD0EFE0D0F0EFCFEFE0D0F0E0\
CFF0E0CFF0E0C0F0DFCFFFE0C0F0DFC0FFDFBFF0DFC0FFDF8F9F8F8FA08F8FA08F
8FA08F80A08080A0807FA08080A0807FA07F7FA07F70A07F70AF7F6FA0706FAF7060A06F60AF70\
60A06F60AF6F5FA0605FAF6050A05F50A05F4F9F50509F504F904F4F904F408F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000050505F5F5F5F5F505F605F60605F5F\
605F60605F606F60606F60606F606F6F606070606F70606F706F6F70606F7F6F70
7F6F708070707F6F7080707F80707F8F707F8F707F907F809070809F7F80907F809F7F8F9F7F80\
A0808FA07F8FAF8090AF808FB08F90AF8090B08F9FB08F90BF8F9FBF8F9FBF90A0
BF909FC09FA0BF9FA0C09FAFC09FAFC0A0B0C0A0AFC0AFB0C0AFB0CFB0BFC0B0BFCFBFBFCFBFBF\
CFBFC0CFBFC0D0C0CFCFC0CFD0CFCFCFCFCFD0D0D0D0D0D0DFDFDFD0D0D0DFDFDF
D0DFDFDFE0E0DFE0E0DFEFEFDFE0E0DFEFEFD0E0E0D0EFE0D0EFE0D0EFE0CFEFE0CFF0E0CFEFDF\
CFF0E0C0F0DFC0F0DFC0F0DFC0FFDFBFF0D0BFFFDFBFF0D08FA08F809F8080A080
80A08080A0807FA07F7FA07F70A07F7FA07F70A07070A0706FA0706FA07060A06F60A06F5FA060\
5FAF6F50A06050A060509F5F509F5F4F90504F905040904F4F904F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F5F5F5F505F605F605F5F5F605F60\
605F606F60606F60606F606F6F6060706F6F6F606F706F6F706F6F7F6F707F6F70
7F70707F6F7080707F8070708F707F8F707F907F8090707F907F80907F809F7F8F9F7F80A0808F\
A07F8FAF8090A0808FAF8090AF8090B08F9FB08F90BF8F9FBF8F9FBF90A0BF909F
C09FA0BF90A0C09FAFC09FA0C0A0AFC0A0AFCFAFB0C0AFB0CFB0B0C0B0B0CFBFBFCFB0BFCFBFC0\
CFBFC0D0C0CFCFC0C0D0CFCFCFCFCFD0D0D0D0D0D0D0D0D0D0D0D0DFDFDFD0DFDF
DFE0E0DFE0E0DFEFEFDFE0E0DFEFE0D0EFE0DFEFE0D0EFE0D0EFE0CFEFE0D0F0E0CFEFE0CFF0E0\
C0F0DFCFF0E0C0F0DFC0F0DFBFF0DFC0FFDFBFF0D0BFFFD080A0808FA08F80A080
80A0807FA07F7FA08070A07F7FA07F70A07070AF7F6FA0706FA07060A06F60AF6F5FA06F5FAF6F\
50A0605FA060509F5F509F5F4F90504F9F504F904F4F904F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000050505F5F5F5F5F505F605F605F5F5F\
605F60605F606F6060605F606F606F6F6060706F6F6F606F706F6F706F6F7F6F70
7F6F6F7F70707F6F7080707F8070708F707F8F707F907F808F707F907F80907F809F7F809F7F80\
A0808FA07F8FAF808FA0808FAF8090AF8090B08F90B08090BF8F9FB08F9FBF909F
BF909FBF9FA0BF9FA0C0A0AFBFA0A0C0A0AFC0A0AFC0AFB0C0AFB0C0B0B0C0B0B0CFBFBFC0BFBF\
CFBFC0CFBFC0CFC0C0CFC0C0D0CFCFCFCFCFD0D0D0D0D0D0D0DFD0D0DFD0DFDFDF
D0DFDFDFE0E0D0E0E0DFEFE0D0E0DFD0EFE0D0EFDFD0EFE0CFEFDFCFEFE0CFEFDFCFF0E0C0EFDF\
C0F0DFC0F0DFC0F0DFBFF0D0BFF0DFBFF0D0BFFFD0B0F0D080A0807FA08080A080
7F9F7F7FA07F70A07F70A07F6FA07070A0706FA0706FA07060A06F60A06F5FA0605FA06050A05F\
50A05F509F5F509F5F4F90504F9050408F4F408F4F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F5F605F5F5F605F60605F5F606060\
605F606F60606F60606F606F6F606F706F6F6F606F706F70706F6F7F6F707F6F70
7F70707F707080707F8070708F7F7F8F707F8F7F7F8F707F907F80907F809F7F809F7F80A0808F\
A0808FA0808FA0808FAF8F90AF8090B08F90B08F90BF8F9FB08F9FBF909FBF909F
BF9FA0BF9FA0C0A0A0BFA0A0C0AFAFC0A0AFC0AFB0C0AFAFCFB0B0C0B0B0CFBFBFC0BFBFCFC0C0\
CFC0BFCFC0C0CFC0C0D0CFCFCFCFCFD0D0D0D0D0CFD0DFD0D0DFD0DFDFDFD0DFDF
DFE0E0DFE0E0DFEFE0D0E0E0DFEFE0D0EFE0D0EFE0CFEFDFD0F0E0CFEFDFCFF0E0CFEFDFCFF0E0\
C0F0DFC0F0DFC0F0DFC0FFDFBFF0D0BFFFDFBFF0D0BFFFD0B0FFD080A0807FA07F
7FA08070A07F70A07F70A07070AF7F6FA0706FA07060A06F60AF6F5FA0605FAF6F50A0605FA060\
509F5F509F5F4F90504F9050408F4F4F8F4F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F505F5F5F5F5F5F5F605F60605F60\
606060605F606F60606F60606F606F6F606F6F6F6F6F606F706F70706F6F7F6F70
7F6F708070707F6F7080707F80707F8F707F8F707F907F7F90707F907F80907F809F7F809F7F80\
A0808FA07F80A0808FA0808FAF8F90AF808FB08F90B08F90BF909FB08F90BF909F
BF909FBF9FA0BF9FA0BFA0A0BFA0A0C0AFAFBFAFAFC0AFAFC0AFAFC0B0B0C0B0B0CFBFBFC0BFBF\
CFC0C0CFC0BFCFCFC0CFCFC0D0CFCFCFCFCFD0D0CFD0D0CFD0DFD0D0DFD0D0E0DF
D0E0DFDFE0E0D0E0DFD0EFE0D0E0DFD0EFE0CFEFDFCFEFDFCFEFDFCFF0DFC0EFDFCFF0DFC0EFDF\
C0F0DFBFF0D0C0F0DFBFF0D0BFFFDFB0F0D0BFFFD0B0FFCFB0FFD07FA07F7FA07F
70A07070A07F6FA0706FA0706FA06F6FA07060A06F60A06F5FA0605FA06050A05F50A05F4F9F50\
509F504F904F4F904F408F4F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F505F5F5F5F5F5F5F605F60605F60606060\
605F606F606F6F60606F606F6F606F706F6F6F6F6F706F70706F6F7F70707F6F70
7F70707F6F7080707F80707F8F707F8F707F907F808F7F7F907F80907F809F80809F7F80A0808F\
9F7F80A0808FA0808FAF8F90AF808FB08F90B08F90B08F9FB08F90BF909FB0909F
BF9FA0BF9F9FBFA0A0BFA0A0C0AFAFBFA0AFC0AFAFC0AFAFC0B0B0C0B0B0CFBFBFC0BFB0CFC0BF\
C0C0BFCFCFC0CFC0C0CFCFCFCFCFC0D0D0CFCFD0CFD0DFD0D0DFD0D0E0DFD0DFD0
DFE0DFD0E0DFDFEFE0D0EFDFD0EFE0CFEFDFD0EFDFCFEFDFCFF0DFCFEFDFCFF0DFC0F0DFC0F0DF\
C0F0D0C0F0DFBFF0D0BFFFDFBFF0D0BFFFD0B0FFD0B0FFD0B0FFCF7FA07F70A07F
70A07F6FA07070A0706FA0706FA07060A06F60AF6F5FA0605FAF6050A05F50A060509F5F509F50\
4F904F4F9050408F4F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F5F5F5F505F5F5F5F5F5F5F605F60605F5F\
606060605F606F606F6060606F606F6F606F6F6F6F6F606F706F70706F6F7F7070
706F707F70707F707080707F8070708F707F8F707F907F808F707F907F80907F809F80809F7F80\
A0808FA08080A0808FA0808FAF808FAF808FB08F90B08F90B09090B09090BF909F
B0909FBF9FA0BF9FA0BFA0A0BFA0A0BFAFAFBFAFAFC0B0AFBFAFAFC0B0B0C0B0B0C0BFBFC0BFB0\
CFC0BFC0C0BFCFCFC0CFCFC0CFCFC0CFCFC0D0D0CFCFD0CFD0DFD0D0DFD0D0E0DF
D0E0D0D0E0DFD0E0DFD0EFDFCFE0DFD0EFDFCFEFDFCFEFDFC0EFD0CFEFDFC0EFD0C0F0DFBFEFD0\
C0F0D0BFF0D0BFF0D0BFF0D0BFFFD0B0F0CFB0FFD0B0F0CFB0FFCF70A07070A070
6FA0706FA07060A06F60A06F5FA06F60A06F5FA0605FA060509F5F509F5F4F9F504F9F50408F4F\
4F904F408F40000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F505F605F605F5F5F605F60605F60606060\
6060606F606F6060606F6F6F6F606F6F6F6F6F6F6F706F70706F707F7070706F70
7F707F7F707080707F80707F8F7F7F8F707F8F7F808F7F7F907F80907F809F80809F7F80A0808F\
A0808FA0808FA0808FAF8F90AF808FB08F90B08F90B0909FB09090BF909FB0909F
BF9FA0BF9F9FBFA0A0BFA0A0C0AFAFBFAFAFC0B0AFBFAFAFC0B0B0C0B0B0C0BFB0C0BFB0CFC0BF\
C0C0BFCFCFC0CFCFBFCFD0C0CFD0C0D0D0CFCFD0CFD0DFD0D0DFD0D0E0D0D0E0D0
D0EFDFD0E0DFD0EFDFD0EFDFD0EFDFCFEFDFCFEFDFCFEFD0CFF0DFC0EFD0CFF0DFC0F0D0C0F0DF\
BFF0D0C0F0D0BFF0D0BFFFD0BFF0D0BFFFD0B0FFCFB0FFD0AFFFCF70A07F6FA070
6FAF7060A06F6FA07060A06F60AF6F5FA0605FA060509F5F50A05F4F9F50509F504F904F4F904F\
408F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F5F5F5F505F605F605F5F5F605F60605F60\
606060605F606F606F6060606F6F6F6F606F6F6F6F6F6F6F706F70706F6F7F7070
7F6F707F70707F707080707F80707F8F7F7F8F707F8F7F7F8F7F7F907F80907F809F80809F7F80\
A0808F9F8080A0808FA0808FAF8F90AF808FB08F90AF8F90B0909FB09090B09F9F
B09F9FBF9FA0B09F9FBFA0A0BFA0A0BFAFAFBFAFA0C0B0AFBFB0AFC0BFB0BFBFB0C0BFB0C0BFB0\
CFC0BFC0C0BFCFCFC0C0CFBFCFD0C0CFD0C0CFDFCFCFDFCFD0DFD0CFDFCFD0E0D0
D0E0D0D0EFDFCFE0D0D0EFDFCFEFD0CFEFDFCFEFD0CFEFDFC0EFD0C0F0D0C0EFD0C0F0D0BFF0D0\
C0F0D0BFF0CFBFF0D0B0F0CFBFFFD0B0F0CFB0FFCFAFFFCFAFFFC06FA0706FA070
60A06F60A06F5FA0605FA06F5FA0605FA060509F5F509F5F4F90504F9050408F4F4F8F4F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F5F5F5F5F605F5F5F606060605F60606060\
6060606F606F60606F6F6F6F6F606F6F6F706F6F6F706F70706F6F707070706F70
7F70707F6F7080707F80707080707F80707F8F7F7F8F7F7F907F80907F7F9F8080907F809F8080\
9F7F80A0808FA0808FAF808FA0808FAF8F90AF8F8FB09090B09090B09F9FB0909F
BF9F9FB09F9FBFA0A0BFA0A0BFAFA0BFAFA0C0B0AFBFB0AFC0B0B0BFB0AFC0BFB0C0BFB0C0C0BF\
C0C0B0CFCFBFC0CFBFCFD0C0CFD0C0CFDFCFCFD0CFD0DFCFCFDFCFD0E0D0D0E0D0
D0EFDFCFE0D0D0EFDFCFEFD0CFEFDFCFEFD0CFF0DFC0EFD0CFF0D0C0EFD0C0F0D0C0F0D0C0F0D0\
BFF0CFBFFFD0BFF0CFBFFFD0B0F0CFB0FFD0B0FFCFB0FFCF6FA0706FA07060A06F
60A06F5FA06F60AF6F5FA0605FA060509F5F509F5F4F90504F9050408F4F4F904F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005050505F5F5F5F505F5F5F605F5F5F606060605F60\
6060606060606F606F6060606F6F6F6F606F6F6F6F6F6F6F706F6F706F6F707070
706F6F7F70707F6F708070707F707080707F8070708F7F7F8F707F907F7F907F7F907F80907F7F\
9F80809F7F80A0808FA08080A0808FA0808FAF8F8FAF8F8FB09090AF9090B09F9F
B09F90B0A09FB0A09FBFA0A0B0A0A0BFAFA0BFAFA0BFB0AFBFB0AFC0BFB0BFBFAFC0C0B0BFBFB0\
C0C0BFC0C0BFCFCFBFC0CFBFCFD0C0C0D0C0CFDFCFCFDFC0CFDFCFCFDFCFD0E0D0
CFE0CFCFEFD0CFE0D0CFEFD0C0E0D0CFEFD0C0EFD0C0F0D0C0EFD0C0F0D0BFEFCFC0F0D0BFF0CF\
BFF0D0B0F0CFBFFFCFB0F0CFB0FFCFB0F0CFB0FFCFAFFFC06FA07060A06F60A06F
5FA0605FA06F5FA0605FA060509F5F509F5F4F90504F9050408F4F408F4F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F5F5F5F5F5F605F605F5F60606060605F606F606F\
6060606F606F60606F6F6F6F6F6F6F6F6F706F6F6F706F70706F6F707070706F6F
7F70707F70707F70707F7070807F7F8070708F7F7F8F70708F7F7F8F7F7F907F80907F7F9F8080\
9F8080A080809F8080A08F8FA08080AF8F8FAF8F8FAF9090AF9090B09F90B09F90
B0A09FB0A09FBFA0A0B0A09FBFAFA0BFAFA0BFB0AFBFB0AFC0BFB0BFBFAFC0C0B0C0BFB0C0C0BF\
C0C0BFC0CFBFC0CFBFCFD0C0CFD0C0CFDFC0CFDFC0CFE0CFCFDFCFD0E0D0CFE0CF
D0EFD0CFEFD0CFEFD0CFEFD0CFEFD0C0EFD0CFF0D0C0EFD0C0F0D0C0F0D0C0F0D0BFF0CFC0F0D0\
BFF0CFBFFFD0B0F0CFBFFFCFB0FFCFB0FFCFAFFFCFAFFFCF60A06F60A07060A06F
60AF6F5FA0605FA060509F5F509F5F4F90504F9050408F4F4F8F4F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F505F5F5F5F5F5F5F605F605F5F5F606060605F60\
60606F6060606F606F60606F6F6F6F6F606F6F6F6F6F6F6F706F70706F6F707070
706F6F7F70707F6F6F7F70707F70708070708070708F7F708F70708F7F7F8F7F70907F7F907F7F\
9F80809F7F7F9F80809F8080A08080A08080AF8F8FA08F8FAF9090AF908FB09F90
AF9F90B0A09FB0A09FB0AFA0B0AF9FBFB0A0B0AFA0BFB0AFBFB0AFBFBFAFBFBFAFC0C0B0BFC0B0\
C0CFBFC0C0B0C0CFBFC0CFBFCFD0C0C0D0BFCFDFC0CFDFC0CFE0CFCFE0CFCFE0D0
CFE0CFCFEFD0C0E0CFCFEFD0C0EFCFC0EFD0C0EFCFC0F0D0BFEFCFC0F0D0BFEFCFBFF0CFBFF0CF\
BFF0CFB0F0CFB0FFCFB0F0CFB0FFCFAFF0C0AFFFCFA0F0C060A06F5FA0605FA06F
5FA0605FA060509F5F509F5F4F90504F9050408F4F408F4F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F5F605F5F5F605F605F5F60606060605F6060606F\
6060606F6F6F60606F6F6F6F6F606F706F706F6F6F706F70706F6F7F7070706F70
7F70707F6F708070707F70708070708070708F7F7F8070708F7F7F8F7F70907F7F907F7F90807F\
907F7F9F80809F8080A08080A08080A08F8FA08F8FAF908FA0908FAF9F90AF9F90
B0A09FAFA090B0AF9FB0AF9FB0AFA0B0AFA0BFB0AFBFB0A0BFBFAFBFBFAFC0C0B0BFC0B0C0C0B0\
BFC0B0C0CFBFC0CFBFC0D0C0C0D0BFCFDFC0CFDFC0CFE0CFCFE0C0CFEFCFCFE0CF
CFEFCFC0E0CFCFEFD0C0EFCFC0EFD0C0EFCFC0F0D0C0EFCFC0F0D0BFEFCFC0F0D0BFF0CFBFF0CF\
B0F0CFBFFFCFB0F0CFB0FFCFAFF0C0B0FFCFAFFFC0A0FFC05FA06F60AF6F5FA060
5FA060509F5F509F5F4F90504F9050408F4F408F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F505F5F5F605F5F5F605F605F5F60606060605F60\
6060606060606F606F6060606F606F60606F6F6F6F6F606F706F6F706F6F707070
706F6F7F70707F70708070707F70708070708070708F7F708070708F7F7F8F7F70907F7F8F7F70\
907F7F907F7F9F807F9F7F7FA080809F8080A08F8FA08F80A0908FA0908FAF9F90
AF9F90AFA090AFA090B0AF9FB0AF9FB0B0A0B0B0A0BFB0A0B0B0A0BFBFAFBFBFAFBFC0B0BFC0AF\
C0CFB0BFCFB0C0D0BFC0CFBFC0D0BFC0D0BFCFDFC0C0DFC0CFE0CFCFE0CFCFEFCF
C0E0C0CFEFCFC0E0CFC0EFCFBFEFCFC0EFCFBFEFCFC0F0CFBFEFCFBFF0CFBFF0CFBFF0CFB0F0C0\
B0F0CFB0F0C0B0FFCFAFF0C0AFFFC0AFFFC0AFFFC09FF0B05FA06F5FA0605FA060
4F9F50509F504F904F4F904F408F40408F40000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F5F605F5F5F606060605F60606060605F606F606F\
6060606F606F6060606F606F6060606F6F6F6F606F6F6F706F6F6F706F706F6F6F
7070707070707F70707F707080707F807070807F7F8070708F7F7F8F7F70907F7F8F7F7090807F\
907F7F9F807F9F7F7FA080809F8080A08F80A08F80A0908FA0908FAF9F90A09F8F
AFA090AFA090B0AF9FAFAF9FB0B0A0B0B09FBFB0A0B0B0A0BFBFAFBFBFAFBFC0AFBFC0AFC0CFB0\
BFCFB0C0D0BFC0CFB0C0D0BFC0D0BFCFDFC0C0DFC0CFE0CFCFE0C0CFEFCFC0E0C0
CFEFCFC0EFCFC0EFCFC0EFCFC0F0CFC0EFCFC0F0CFBFEFCFC0F0CFBFF0CFBFF0CFB0F0CFBFFFCF\
B0F0CFB0FFCFB0F0C0B0FFCFAFFFC0AFFFC0A0FFBF9FF0BF5FA0605FA060509F5F
509F5F4F90504F9050408F4F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F505F5F5F5F5F5F5F605F605F5F5F606060605F60\
6060605F5F60606060605F60606060605F6060606F6060606F6F6F6F606F6F6F6F
6F6F6F707070706F6F7070707070707F70707F70708070708070708F7F7F8070708F7F7F8F7F70\
907F7F907F7090807F90807F9F8F809F807FA08F809F8F80A0908FA0908FA09F8F
A09F8FAFA090AFA090AFAF9FAFAF9FB0B09FB0B09FB0BFA0B0BFA0BFC0AFB0BFA0BFC0AFBFC0AF\
BFCFB0BFCFB0C0D0B0BFD0B0C0DFBFC0DFBFC0DFC0C0DFC0CFE0C0C0E0C0C0EFCF
C0E0C0C0EFCFBFE0C0C0EFCFBFEFC0C0EFCFBFEFC0BFF0CFBFEFC0BFF0CFB0F0C0B0F0CFB0F0C0\
B0FFCFAFF0C0B0FFC0AFF0C0AFFFC0A0FFC0A0FFBF5FA0605FA060509F50509F50
4F904F4F904F408F40000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F5F605F5F5F605F605F5F5F6060605F5F60605F60\
5F5F5F6060605F5F606060605F5F5F5F5F6060606060606F6060606F6F6F6F606F
6F6F6F6F6F6F707070706F707070707070707F707F7F7070807F7F7F7070807F7F807F708F7F7F\
8F7F7F90807F907F7F90807F90807F9F8F809F8F80A090809F9080A09F8FA09F8F
AFA090A0A090AFAF90AFAF90B0B09FAFB09FB0BFA0B0BFA0B0BFA0B0BFA0BFC0AFBFC0AFBFCFB0\
BFCFAFC0D0B0BFD0B0C0DFBFBFD0BFC0DFC0C0DFBFC0E0C0C0E0C0CFEFC0C0E0C0
C0EFCFC0EFC0C0EFCFBFEFC0C0EFC0BFEFC0BFF0CFBFEFC0BFF0CFB0F0C0BFF0CFB0F0C0B0FFCF\
B0F0C0B0FFCFAFFFC0AFFFC0AFFFC0AFFFC09FF0B090F0B0509F5F509F5F4F9050
4F904F408F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F505F5F5F5F5050505F5F5F50505F5F5F5F505050\
505F5F5050505050504F4F4F4F4F504F9F5050A05F4F4F4F6060606060606F606F
6060606F6F6F6F6F6F706F706F6F6F707070706F707F70707070707F70707F7070807F7F807070\
807F7F807F708F7F7F8F7F7F90807F90807F9F8F809F8F7F9F90809F9080A09F8F
A09F8FA0A090A0A090AFAF90AFAF90AFB09FAFB09FB0BFA0AFBF9FB0C0A0B0C0A0BFCFAFB0C0AF\
BFCFAFBFCFAFBFD0B0BFD0B0C0DFBFBFDFBFC0E0BFC0DFBFC0E0C0C0E0BFC0EFC0
BFE0C0C0EFC0BFEFC0BFEFC0BFEFC0BFEFC0B0EFC0BFF0C0B0EFC0B0F0C0B0F0C0B0F0C0AFF0C0\
B0F0C0AFF0C0AFFFC0A0F0BFAFFFC0A0FFBF9FF0B080EFA04F90504F904F4F904F
408040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F4F4F5F5F5F5F5F5F5F5F5F505F50506F50507050508050\
508F50509050509050509F5F50A0505FAF5F50AF5F5FB0605FB05F5FBF605F5F5F
6F606F60606F6F6F6F6F6F6F706F706F6F70707070706F707F707F7070707F7F7F7F7070807F7F\
807F7F807F7F807F7F8F80808F807F908F80908F7F9F90809F9080A09F8F9F9F80
A0A08FA0A08FAFAF90A0AF90AFB09FAFB09FB0BF9FAFBF9FB0C0A0B0C0A0B0C0AFB0C0A0BFCFAF\
BFCFAFBFD0B0BFD0B0C0DFBFBFDFB0C0E0BFBFE0BFC0E0C0C0E0BFC0EFC0BFE0BF
C0EFC0BFEFC0C0EFC0BFEFC0BFF0C0BFEFC0BFF0C0B0EFC0BFF0C0B0F0C0B0F0C0B0F0C0B0FFC0\
AFF0C0B0FFC0AFFFC0AFFFC0A0FFBFA0FFBF90F0AF8FEFAF4F90504F904F408F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000050505050505F5F605F505F5050605F506F50507F50508050\
509050509050509F5F50A0505FAF5F50AF5F5FB05F5FB05F5FC0605FBF6060C06F
60C06F505050605F6060606F6060606F6F6F6F6F6F706F706F6F707070707070707F707F7F7070\
7F7F7F7F7070807F7F807F7F8F807F8F807F908F80908F7F9F90809F907F9F9F80
9F9F80A0A08FA0A08FA0AF90A0AF90AFB09FAFB090AFBF9FAFBF9FB0C0A0AFC0A0B0CFA0B0CFA0\
BFD0AFB0CFAFBFD0B0BFD0B0BFDFB0BFDFB0C0E0BFBFE0BFC0E0BFBFE0BFBFEFBF
BFE0BFBFEFC0BFEFBFBFEFC0B0EFBFBFF0C0B0EFBFB0F0C0B0F0BFB0F0C0AFF0BFB0F0C0AFF0BF\
AFFFC0AFF0BFAFFFC0A0FFBFAFFFC09FF0B08FF0AF7FE09F4F904F408F40000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000504F505F5F5F5F605F5F605F5060505F7F5F508050508F5F\
5090505FA05F50A05F5FAF5F50AF5F5FBF605FBF605FC0605FBF6060C06F60C06F
6FCF7060C06F6FCF706FCF70605F606060606F6F6F6F606F6F6F706F6F6F7070707070707F707F\
7070707F7F7F7F7F7F80807F80807F8F8F808F8F7F90908090907F9F9F809F9F80
9FA08F9FA080A0AF8FA0AF8FA0B090A0B090AFBF9FAFBF9FAFC09FAFBF9FB0CFA0B0CFA0B0CFAF\
B0CFAFBFD0B0BFD0AFBFDFB0BFDFB0C0E0BFBFE0BFC0E0BFBFE0BFBFEFBFBFE0BF
BFEFBFBFEFBFBFEFC0BFEFBFBFF0C0B0EFBFBFF0C0B0F0BFB0F0C0AFF0BFB0FFC0AFF0BFAFFFC0\
AFF0BFAFFFC0A0FFBFAFFFC0A0FFBF9FF0B08FEFA07FEF9F408F40000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000050505050505F5F605F5060505F6F5F507F505F8F5F509050\
509F5F50A05F5FAF5F50AF5F5FBF605FBF5F60C0605FBF6060C06F60C06F6FC070
60C06F6FCF706FC07070CF7F70CF7070CF7F5050506060606060606F6F6F6F606F6F6F706F6F6F\
7070707070707F7F7F7F7F7F80807F80807F8F8F80808F7F8F90808F907F909F80
909F809FA0809FA080A0AF8F9FAF8FA0B090A0B090AFBF9FA0BF90AFC09FAFC09FB0CFA0AFCFA0\
B0D0AFB0D0AFBFDFAFB0D0AFBFDFB0BFDFB0BFE0BFB0E0B0BFE0BFB0E0B0BFEFBF
B0E0B0BFEFBFB0EFBFB0EFBFB0EFBFB0F0BFB0EFBFB0F0BFAFEFBFAFF0BFAFF0BFAFF0BFA0F0BF\
AFFFBFA0F0BFAFFFBFA0F0BFA0FFBF8FEFAF80EFA070E09060DF8F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000504F505F5F5F5F605F5F6F5F5F705F5F805F5F905F5FA05F\
5FA05F5FAF605FB05F60BF605FBF6060C06060BF606FC06F60C06F6FCF706FC070
70CF7F70CF7070CF7F70CF7F7FD0807FCF8080D0807FD0805050506060606F606F6F606F6F6F70\
6F6F6F7070707070707F7F7F7F807F808F80808F7F8F90808F9080909F808F9F80
9FA08F90A0809FAF8F9FAF8FA0B090A0B08FA0B090A0B090AFC09FAFC09FAFC0A0AFC09FB0D0A0\
B0D0A0B0D0AFB0D0AFBFDFB0BFDFB0BFE0B0BFE0B0BFEFBFBFE0B0BFEFBFB0E0B0
BFEFBFB0EFBFB0EFBFB0EFBFB0F0BFB0F0BFB0F0BFAFF0BFB0F0BFAFF0BFAFFFBFAFF0BFAFFFBF\
A0FFBFAFFFBFA0FFBFA0FFBF9FF0B08FF0AF7FEF9F70E09F60DF8F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005050505050505F6F5F5F6F5F5F7F5F5F805F5F9F5F5FA05F\
5FAF5F5FAF5F5FB0605FB06060BF6060BF6060C06F60C06F6FC0706FC06F70CF70
70C07070CF7F70CF7F7FCF807FCF8080D0807FCF8080D08F80D08F8FDF908FD08F5F5F5F606060\
6F6F6F6F6F6F707070707070707F7F707F707F8F7F7F8F7F80908080907F8F9F80
8F9F8090A08090A0809FAF8F9FAF80A0B08F9FB08FA0BF90A0BF90AFC09FA0C09FAFCF9FAFCF9F\
B0D0A0AFD0A0B0DFAFB0DFAFBFE0B0B0DFAFBFE0B0B0E0B0BFEFB0B0E0B0B0EFB0
B0EFB0B0EFBFAFEFB0B0EFBFAFEFB0B0F0BFAFEFB0AFF0BFAFF0B0AFF0BFA0F0B0AFF0BFA0F0B0\
A0FFBFA0F0B0A0FFBF9FFFB090F0B080EFA07FEF9F6FDF8F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F5F5F5F605F5F705F5F7F5F5F90605F9F5F5FAF60\
5FAF5F60B06060B06060BF6F60BF606FC06F6FBF6F6FC0706FC06F70CF7070C070
7FCF7F70CF7F7FD0807FCF8080D08080D0808FD08F8FD08F8FDF908FD09090DF9F90DF909FDF9F\
50505F6F6F6F6F6F6F7070706F707070807F70807F7F8F7F7F8F7F809080809F80
8F9F808F9F8090A08F90A0809FAF8F9FAF8FA0B0909FB08FA0BF90A0BF90AFC09FA0C09FAFCFA0\
AFCF9FB0D0A0AFD0A0B0DFAFB0DFAFBFE0B0B0E0AFB0E0B0B0E0B0B0EFB0B0EFB0
B0EFB0AFEFB0B0F0BFAFEFB0B0F0BFAFEFB0B0F0BFAFF0BFAFF0BFAFF0B0AFFFBFA0F0B0A0FFBF\
A0F0B0A0FFBFA0FFB09FFFB08FF0AF80EFA070E0906FE08F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000505F505F6F5F5F705F5F805F5F8F5F5F9F605FA05F\
60B06060B06060B06F60B0606FBF6F60BF6F6FC0706FBF6F70C07070C07070CF7F
70C07F7FCF7F7FCF7F80D08080CF8080D08F80D08F8FDF908FD08F90DF9090D0909FDF9F90D090\
9FD09F9FD09F606F6F606F606F706F6F706F707F70707F707F8F7F7F8F7F809080
80907F8FA0808FA08090AF8F90A0809FAF8F9FB08F9FBF8F9FB08FA0BF90A0C090A0CF9FA0C09F\
AFCF9FAFCF9FAFD0A0AFD0A0B0DFAFAFDFAFB0E0AFAFE0AFB0E0B0AFE0AFB0EFB0
AFE0AFAFEFB0AFEFB0AFEFB0AFEFB0AFF0B0AFEFB0AFF0BFA0F0B0AFF0B0A0F0B0A0FFBFA0F0B0\
A0FFB09FFFB0A0FFB090F0AF8FF0AF7FE09F70E0905FDF80000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F505F5F605F6070605F7F60608F605F906060A060\
60AF6060B06F60B06F6FBF706FBF6F70C0706FBF7070C07F70C0707FCF7F70CF7F
7FCF807FCF8080D08080CF808FD08F8FD08F8FDF908FD09090DF9F90D0909FDF9F9FD09F9FDFA0\
9FD09FA0D0A0A0D09F6F606F606F6F6F7F6F6F7F6F7080707080707F8F7F7F8F7F
809F80809F808FA0808FA08090AF8F90AF8F9FB08F9FB08FA0BF909FBF90A0C090A0C090AFCF9F\
A0CF9FAFD0A0AFD0A0B0DFAFAFDFA0B0E0AFAFE0AFB0EFB0AFE0AFB0EFB0AFEFAF
B0EFB0AFEFB0AFF0B0AFEFB0AFF0B0AFF0B0AFF0B0A0F0B0AFFFBFA0F0B0A0FFBFA0F0B0A0FFBF\
A0FFB0A0FFB09FFFB090FFAF80EFA07FEF9F6FDF8F60DF80000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F5F5F5F6F5F607F605F805F609060609F60\
60AF6060AF606FB06F6FB06F70BF706FBF7070C07070C0707FC07F70C07F7FCF7F
7FCF7F80CF8080CF8080D08F80D08F8FD0908FD08F90D09090D0909FD09F90D0909FD09F9FD09F\
A0D0A09FCF9FA0D0A0A0CFA05050506060606F706F6F7F6F7080706F8070708F7F
708F707F9F7F7F9F7F80A0808FA08090AF8F8FAF8090B08F90B08F9FBF8F9FBF8FA0C090A0C090\
A0CF9FA0CF9FAFD0A0A0D09FAFDFA0AFDFA0AFE0AFAFE0A0AFE0AFAFE0AFAFEFAF
AFEFAFAFEFB0A0EFAFAFEFB0A0EFAFAFF0B0A0EFAFA0F0B0A0F0B0A0FFB09FF0B0A0FFB09FF0B0\
A0FFB09FFFB090FFAF8FF0A080EFA070E09060DF8F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F605F5F6F5F608060608F60609F6F\
609F606FB06F60B06F6FBF6F6FBF6F70C07070BF707FC07F7FC07F7FCF807FCF80
80CF8080CF808FD08F8FD08F8FD0908FD09090D09090D0909FD09F90D0909FD09F9FD09FA0D0A0\
A0CF9FA0D0A0A0CFA0AFD0A0AFCFA0B0CFAF6060606F706F6F7F6F708070708070
708F7F708F707F9F7F7F9F7F80A08080A0808FB08F8FB08F90BF8F90BF8F9FC0909FC090A0CF90\
A0CF90A0D09FA0D09FAFDFA0A0DF9FAFE0A0A0DFA0AFE0AFAFE0A0AFEFAFAFEFAF
AFEFAFA0EFAFAFF0B0A0EFAFAFF0B0A0EFAFAFF0B0A0F0B0A0F0B0A0F0B0A0FFB09FFFB0A0FFB0\
9FFFB09FFFB090F0AF8FF0A07FEF9F70E09060DF80000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000504F505F605F5F705F608060608F60\
609F6F60A0606FB06F60B06F6FBF706FBF7070C07F70BF7F7FC07F7FC07F80CF80
7FCF8080CF8F80CF8F8FD0908FCF8F90D09090CF9090D09090CF909FD09F9FCF9FA0D09F9FCF9F\
A0CFA0A0CF9FAFCFA0A0CFA0AFCFA0AFCFA0B0CFAF5F5F5F606F6F60706F6F806F
6F806F708F70708F707F9F7F7F9F7F80A08080A0808FB0808FB08090BF8F90BF8F9FC0909FC08F\
9FCF909FCF90A0D09F9FD09FA0DF9FA0DF9FA0DFA0A0DFA0A0E0A0A0E0A0AFEFAF
A0E0A0A0EFAFA0EFAFA0EFAFA0EFAFA0F0AF9FF0AFA0F0AF9FF0AFA0F0B09FF0AF9FFFB09FF0AF\
9FFFB090F0AF8FF0A07FEF9F7FE09060DF8F5FDF80000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000505F5F606F60607F606F8F6F\
6F906F6FA06F6FAF6F6FBF706FBF6F70BF7070BF707FC07F7FC07F80CF8080CF80
8FD08F8FCF8F8FD0908FCF8F90D09090CF9090D09F90CF909FD09F9FCF9FA0D09F9FCF9FA0D0A0\
A0CFA0AFCFA0AFCFA0B0CFAFAFCFA0B0CFAFB0C0AFBFCFB0BFC0AF6F6F6F60706F
6F80706F8F6F7090707090707F9F7F7F9F7F80A08080AF808FB08F8FB08090BF8F90C08F9FCF90\
9FCF90A0D0909FD090A0DF9FA0D09FA0DF9FA0DF9FA0E0A0A0E0A0A0EFA0A0E0A0
A0EFAFA0EFAFA0EFAFA0EFAFA0F0AFA0F0AFA0F0AF9FF0AFA0FFB09FF0AFA0FFB09FFFAF9FFFB0\
9FFFAF90FFAF80F0A080EF9F70E0906FE08F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000505F5F607060607F60\
6F8F6F60906F6FA06F6FAF6F70BF706FB07070BF7F70BF7F7FC0807FC07F80CF80
80CF8F8FD08F8FCF8F90CF9090CF9090CF9090CF909FCF9F9FCF909FCF9F9FCF9FA0CF9FA0CF9F\
AFCFA0A0C0A0AFCFA0AFC0A0B0CFAFAFC0A0B0C0AFB0C0AFBFC0B0BFC0AF606060
606F606F7F6F6F806F708F707090707F9F7F709F707FA07F80AF7F8FB08080B0808FBF8F90BF8F\
9FCF9090CF8F9FD0909FD0909FDF9F9FD0909FDF9F9FDF9FA0E0A09FE09FA0EFA0
9FE0A0A0EFAF9FEFA0A0EFAF9FEFA0A0F0AF9FEFA09FF0AF9FF0AF9FFFAF9FF0AF9FFFAF90FFAF\
9FFFAF8FF0A080EF9F70E09070E0905FDF80000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000050505F5F605F607F60\
6080606F906F6F9F6F70AF706FB06F70BF7070BF707FC07F7FC07F80CF8080CF80
8FCF8F8FCF8F90D09090CF9090CF9090CF909FCF9F9FCF909FCF9F9FCF9FA0CFA0A0CF9FAFCFA0\
A0C0A0AFCFA0AFC0A0B0CFAFB0C0A0B0CFAFB0C0AFBFC0B0BFC0AFC0C0B0BFC0B0
5F5F5F606F606F7F6F6F806F7090706F9070709F7F709F707FA07F7FAF7F80B08080B0808FBF8F\
8FC08F90CF9090CF8F9FD09090D0909FDF9F9FDF909FDF9F9FDF9FA0E0A09FE09F
A0EFA09FEFA0A0EFAF9FEFA09FF0AF9FEFA09FF0AF9FF0A09FF0AF90F0A09FFFAF90FFAF9FFFAF\
8FF0A08FF0A07FEF9F70EF9060DF8F5FDF80000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005F5F5F5F6F5F\
607F606080606F9F6F6F9F6F70B0706FB07070BF7F70BF707FC07F7FC07F80C080
80C0808FCF8F8FC08F90CF9090C09090CF9090C0909FCF9F9FC090A0CF9FA0C09FA0CFA0A0C09F\
AFC0A0AFC0A0B0C0AFAFC0A0B0C0AFB0C0AFBFC0AFB0BFAFBFC0B0BFBFAFC0BFB0
BFB0AFC0B0B06060606F706F607F6F6F8F6F6F906F709F70709F707FA07F7FAF7F80B08080B07F\
8FBF808FC08090CF8F90CF8F90D09090D0909FDF9090DF909FDF9F90DF9F9FE09F
90E09F9FEFA09FEFA09FEFA090EFA09FF0A090EFA09FF0A090F0A090F0A090F0A090FFAF90F0A0\
8FF0A07FEF9F7FEF906FE08F60DF80000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F605F\
60706060806F608F6F709F7070AF7070B07070BF707FC07F7FC07F80C08080C080
8FCF8F8FC08F90CF9090C08F9FCF909FCF909FCF9F9FC09FA0CF9FA0C09FA0CFA0A0C09FAFCFA0\
AFC0A0B0C0AFAFC0A0B0C0AFB0C0AFBFC0AFBFC0AFC0C0B0BFBFB0C0BFB0BFB0B0
C0BFB0C0B0B0C0B0B06060606F706F6F7F6F6F8F706F906F709F7070A0707FAF7F7FAF7F80B080\
80BF808FC0808FC08090D08F90D08F90DF9090D0909FDF9F90DF909FE09F90E09F
9FEFA09FEF9F9FEFA090EFA09FF0A090F0A09FF0A090F0A09FFFAF90F0A090FFAF90FFA090FFAF\
80F09F80F09F70E08F6FE08F5FDF80000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F50\
5F6F5F60706060806F6F906F70A0706FAF6F70BF7070BF707FBF7F7FBF7F80C080
80BF808FC08F8FC08F90C09090C0909FCF909FC0909FC09F9FC09FA0C09FA0C09FAFC0A0AFC0A0\
AFC0A0AFBFA0B0C0AFB0BFA0BFC0AFB0BFAFBFC0B0BFBFAFBFBFB0BFB0AFC0B0B0
BFAFAFC0AFB0BFA0AFC0AFB0505050606F6F6070606F806F6F8F6F709F706F9F7070AF707FAF7F\
80B07F7FBF7F80C0808FCF808FD08F8FD08F90DF8F8FD08F90DF908FDF9090E09F
90E09090EF9F90E09F90EFA090EF9F90F0A090EF9F90F0A090F0A090FFA08FF0A090FFA08FF09F\
80F09F70E0906FE08F60DF805FDF7F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
505050607060607F606F8F6F6F906F70A07070AF707FBF7F7FBF7F80BF8080BF80
8FC0808FBF8090C08F90C08F90C09090C0909FCF9F9FC09FA0C09FA0C09FAFC0A0AFC0A0AFC0A0\
AFC0A0B0C0AFB0C0A0B0C0AFB0BFAFBFC0B0BFBFAFBFBFB0BFB0AFC0B0B0BFB0AF
C0B0B0BFAFAFC0AFB0C0A0AFC0A0B0C09FAF6F6F6F60706F6F7F6F6F806F709F7070A07070AF7F\
70AF707FBF7F7FBF7F80CF8080CF808FD08F8FD08F8FDF908FDF8F90E0908FE090
90E09F8FE09090EF9F90EF9F90F0A090EF9F90F0A090F09F90F0A08FF09F90FFA08FF09F80F09F\
7FEF9070EF906FE08060DF8050D070000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000505F506070606080606F8F6F6F906F70A07070AF707FB07F7FB07F80BF80
80B0808FBF808FBF8090BF8F90BF8F9FC0909FBF90A0C09F9FBF9FA0C09FA0BF9FAFC0A0AFBF9F\
AFC0A0AFBFA0B0C0AFB0BFA0BFBFAFB0BFAFBFBFAFB0B0AFBFB0AFBFAFAFBFAFAF
BFA0AFBFA0AFBFA0AFC0A0AFBF9FAFC09FAFC090AF606060606F606F7F6F6F8F6F6F9F706FA06F\
70AF7070B0707FBF7F7FBF7F80CF8080CF808FD08F80D0808FDF8F8FDF8F8FE090
8FDF8F8FE0908FE09090EF9F8FEF9090F09F8FEF9F90F09F8FF09F8FF09F8FF09F80F09F7FEF90\
7FEF906FE08060E08050D07F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F5F5F607F606F806F708F7070906F7FA07F7FAF7080B07F80B07F
8FBF808FBF8090BF8F90BF8F9FC0909FBF909FC09F9FBF90A0C09FA0C09FAFC0A0AFBFA0B0C0A0\
AFBFA0B0C0AFB0BFA0BFC0AFBFBFAFBFBFAFBFB0AFBFB0AFBFAFAFBFAFAFBFAFAF
C0AFB0BFA0AFC0A0AFBF9FAFC09FB0C090AFC090AFC090AF6060606070606F806F6F8F6F70A070\
6FA07070B07070B0707FC07F7FC07F80CF8080D0808FDF8F80DF8F8FDF8F80DF8F
8FE0908FE0908FEF908FEF9090F09F8FEF9F90F09F8FF09F90FFA08FF09F8FF09F80EF907FF090\
70E08F6FE08F5FDF7F5FDF7F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000505F5F607060607F60708F6F7090707F9F707FA07080B07F
80B07F8FBF808FB08090BF8F90B08F9FBF909FBF90A0BF9FA0BF9FAFC09FA0BF9FAFBFA0AFBFA0\
B0BFA0B0BFA0B0BFAFB0B0A0BFBFAFB0B0A0BFB0AFB0AFAFBFAFAFBFA0A0BFA0AF
BFA0AFBFA0AFBF9FAFC09FAFBF90AFC090AFBF8FAFC08FAFC08FAF6060606070606F806F6F906F\
6FA0706FA06F6FB0706FB07070C07F70C07F7FCF807FD08080DF8080D08080DF8F
80DF8F8FE09080E08F8FEF9080EF908FEF9F80EF908FF09F80F09F8FF09F7FEF907FEF9070E08F\
70E08F60DF7F5FDF7F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005050506070606F7F6F708F70708F707F9F7F7FA07F
80B08080B07F8FBF8F8FB08090BF8F90B08F9FBF909FBF90A0BF9FA0BF9FAFC0A0AFBFA0B0BFA0\
B0BFA0B0BFAFB0BFA0BFBFAFB0B0A0BFB0AFB0AFA0BFAFAFBFA0A0BFAFAFBFA0AF
BFA0AFBF9FA0C09FAFBF90AFC090AFBF8FAFC090AFC08FAFC08FAFC080AF606F6F6070606F8F6F\
6F906F70AF706FAF7070B0706FBF7070C07F70C07F7FD0807FD08080DF8080DF80
80E08F80E08F80EF9080E08F8FEF9080EF908FF09F80F0908FF09F80F09080F09070EF8F70EF8F\
6FE08060E08050D070000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000505050606F6060706070806F708F707F9F7F
7F9F7F80AF7F80B0808FB0808FB08090B08F90B08F9FBF909FB090A0BF9FA0B09FAFBFA0AFB0A0\
B0BFA0B0B0A0B0B0A0B0AFA0B0AFA0B0AFA0BFAFAFB0A0A0BFA0A0B09FA0BF9FAF
B090A0BF9FAFBF90A0BF90AFBF8FA0BF8FAFBF80A0C08FAFBF80A0C080AF4F4F4F606F60607F60\
608F6F60906F6FAF6F6FAF6F6FB0706FBF7070C07F70C0707FD07F7FD07F7FDF80
7FDF8080E08F7FE08080E08F7FE08F80EF9080EF8F80F09080EF907FEF9070EF8F70EF8F6FDF80\
60DF805FD07050D070000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000505050606F6060706F708070708F70
809F7F809F7F8FAF808FB08090B08F90B08F9FB0909FB090A0BF9FA0B09FAFBF9FAFB09FB0BFA0\
B0B0A0B0B0A0B0B0A0B0B0AFB0AFA0BFAFAFB0A0A0BFA0AFB0A0A0BFA0AFBF9FA0
BF9FAFBF90A0BF90AFBF8FA0C090AFBF8FA0C08FAFBF80A0C080AFBF7FA0BF7FA050505060706F\
607F606F906F6F9F6F6FAF706FB06F6FBF706FBF7070CF7F70CF7F7FD07F7FDF7F
7FE0807FDF807FE08F7FE08080EF8F7FEF8F80F09080EF9080F0907FEF8F70EF8F6FE0806FE080\
60DF7F5FDF70000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000505050606060607060707F6F
70807080907F809F7F8FA0808FAF8090B08F90AF8F9FB0909FB090A0B09FA0B090AFB09FAFB09F\
B0B0A0AFAF9FB0AFA0B0AFA0B0AFA0B0A09FB0A0A0B09FA0BF9FA0B090A0BF9FA0
B090A0BF90A0B08FA0BF8FA0BF80A0BF80A0BF80A0BF80A0BF7FA0BF7FA0B0709FBF7F9F505050\
60706060806060906F609F6F6FAF6F6FAF6F6FBF706FBF706FCF706FCF7070D07F
70DF7F7FE08070DF807FE0807FE0807FEF8F7FEF8F7FEF8F7FEF8F70EF8F6FE0806FE0805FDF7F\
5FDF7F4FD06F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000504F50606060607060
707F7070807080907F809F7F90A08090AF809FB08F9FB08FA0B090A0B090AFB09FA0B09FAFB09F\
AFAF9FB0AFA0AFAF9FB0AFA0AFA09FB0A0A0B09FA0B09FA0B09F9FBF9FA0B090A0
BF90A0B08FA0BF8FA0BF80A0BF8FA0BF80A0BF80A0BF7FA0BF7FA0B0709FBF7F9FB0709FB0709F\
5F5F5F6070606080606F9F6F60A06F6FB06F6FB06F6FBF706FC0706FCF7F6FD070
70DF7F70DF7F7FE08070E0807FEF8070E0807FEF8F70EF807FEF8F6FE0806FE08060DF7F5FDF7F\
50D070000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F5F5F
606F606F7F6F7080707F8F7F80907F8FA08090AF809FAF8F9FAF8FA0B090A0AF90A0AF90A0A090\
AFAF9FA0A090AFA09FAF9F9FAFA09FAF9F9FB09FA0AF909FB0909FB0909FB090A0
B08F9FB08FA0B0809FBF80A0B07F9FBF7FA0B07F9FBF7F9FB0709FB0709FAF7090B07090AF6F8F\
AF6F905F5F606070605F8060609F6060A0606FB06F60B06F6FBF706FC06F6FCF70
6FD07070DF7F6FDF7F70E07F70E07F70E08070E07F70E0806FE07F6FE07F60DF7060DF7050D06F\
50D06F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F5F5F606F606F7F6F707F70808F7F80907F90A08090A08F9FB0909FAF8FA0AF909FA090A0AF90\
A0A090AFA09FA09F90AFA09FAF9F9FAF9F9FAF909FB09FA0B0909FB0909FB08F9F
B08FA0B0809FBF80A0B0809FBF80A0B07F9FBF7F9FB0709FB07F9FB07090B07090AF6F8FAF7090\
AF6F8FAF6F8F606060607F60608F60609F6F60AF6F6FB06F6FBF6F6FC0706FC070
6FD0706FD07070E07F6FDF7F70E08070E07F70E0806FE07F70E0806FDF7F60DF7F5FD07050D070\
4FD06F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F5F5F6060606F706F707F707F8F7F808F7F8F9F8090A0809FAF8F9FA08F9FA08F9F9F8F\
A0A0909F9F8FA09F90A09090AF9F90A09090AF909FAF8F90B0909FAF8F90B08F9F
AF809FB0809FB07F9FB07F9FB07F9FB07F9FAF7090B07090AF7090AF7090A06F8FAF6F8FA06F80\
A06F8F9F60805050505F6060607F605F8F6060A06060AF606FB06F60BF6F6FC070
60CF6F6FD07060D0706FE07F6FDF706FE07F6FDF706FDF7F6FDF7060DF705FD06F50D06F4FCF60\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005F5F5F6060606F706F707F707F8F7F808F7F8F90808F90809FA08F9F9F8F9F9F8F\
9F9F8FA09F90A0908FA09090A0908FAF9090A08F90AF8F90AF8F90B08F9FAF8090
B0809FAF7F90B0809FB07F9FB07F9FAF7090B07F90AF7090AF7090A06F8FAF708FA06F80A06F8F\
9F6F80A06F809F607F50505F5F6F6060806060906060A06F60AF606FBF6F60BF6F
6FC07060CF6F6FD0706FDF706FDF7060DF706FDF7060DF706FDF705FD06F5FD06F50D060000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000050505F5F5F5F6F6F6F6F706F7F7F707F7F70808F7F808F7F909080909080\
9F9F809F90809F908F9F8F80A08F8F9F8F8FA08F90A0808FAF8090A08090AF8090
AF7F90AF7F90AF7F90B07F9FAF7090AF7090A0708FAF708FA06F80A06F8FA06F80A06F809F607F\
9F6F7F9F607F9F607F9060705F5F5F5F6F606080605F906060AF6060AF6060BF6F
60BF6F60CF6F60CF6F6FD07060D06F60DF6F5FD06F60DF6F5FD06F5FD06F50CF6050CF60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000050505F5F5F606F6F6F6F6F6F7070707F7070807F7F80807F8F8F80\
908F809F9080908F809F908F9F8F80A08F8F9F808FA08F8FA0808FAF8090A07F8F
AF8090AF7F90B07F90AF7090AF7F90AF708FAF708FA0708FA0708FA06F80A06F809F6F7F9F6F80\
9F6F7F9F6F7F90607090607F8F606F5F5F605F7060608F60609F6060AF6060B060
6FBF6F60C06F60CF6F60D06F60DF705FD06F60D06F5FD06060D06F5FD06050D060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000505050606060606F6F706F6F706F6F7F70707F7070\
807F7F8F7F7F908F7F908F7F9F8F809080809F80809F7F80A0808F9F7F80A07F8F
A0708FAF7F8FA0708FA0708FA07080A0708F9F6F80A070809F6F809F6F809F6F7F9F6F7F906070\
906F7F9060709060708F606F806060705F5F6060605F7060608F605F9F6060AF60
60B06060BF6F5FC06060CF6F5FCF605FD06F50CF605FD06050CF6050D0604FC05F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005050505F5F5F6060606F6F6F6F6F6F706F70\
7F6F6F8070708070708F7F7F90807F9F808090807F9F80809F7F80A07F8F9F7080
A07F8FA07080A0708F9F7080A070809F6F809F70809F6F7F9F6F7F906F7F9F6F7F906F70906F70\
906070906F708F606F80606F705F606F5F605050506060605F706060906060A060
60B0605FB06060C06F5FC06060D06F5FCF605FD06050CF5F50CF604FCF5F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005050505F5F5F6F6F6F6F6060\
706F6F706F6F7F6F707F6F6F8070708F7070907F7F90707F907F7F90707F9F7080
9F6F7F9F7080906F7F9F707F906F7F906F7F906F70906F7F8F6070906F708F6070906F708F606F\
8F606F8F606F80606F705F606F5F605F505F5050505F5F5F5F70605F805F60A060
5FAF5F5FB0605FBF5F5FC0605FC05F5FCF6050C05F50CF5F4FC0504FC05F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000050505F606060\
6F6F6F706F6F706F6F7F6F707F6F70806F70806F7090707F90707F90707F90707F
9F7080906F7F90707F906F70906F7F906F70906F708F6070906F708F606F8F6F708F606F8F6F70\
80606F8F606F7F6060706060605F5F5F5F5F5050505F5F605F6F5F5F80605F9F5F
60AF605FAF5F5FBF605FBF5F5FCF6050C05F50CF5F4FC05050CF5F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F4F50\
5F5F5F6060606F6F6F6F606F706F6F706F6F7F6F70806F708F6F708F6F70906F7F
8F6F70906F708F6F708F6F708F6F708F6F7080606F8F6F7080606F8F6F6F80606F80606F80606F\
80606F7F6060706060605F5F605F5F5F505F50505050505F5F6F5F5F7F5F5F905F
5FA05F5FAF5F50AF505FBF5F50BF5F50C05F4FBF504FC050000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000050505F5F5F5F6F6F6F6F606F706F70706F6F7F6F707F6F6F8F6F70806F70
8F6F70806F708F6F70806F708F6F70806F6F806F70806F6F806F707F606F806F6F7F606F7F606F\
7F606F70606F605F606060605F5F5F5F5F5F5050505F5F5F5F605F5F7F5F5F8F5F
5FA05F50A05F5FB05F50B05050BF5F50BF5050C05F4FBF50000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000505050605F606060606F6F6F6F606F7F6F6F7F6F6F7F6F70
7F606F7F6F707F606F806F707F606F7F6F6F7F606F7F6F6F7F606F7F606F70606F7F606F706060\
70606F605F606060605F5F5F5F5F5F50505F5F5F5F5050505F605F506F505F805F
50905050A05F50A05050B0504FB0504FBF504FB04F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005050505F5F5F6060606F6F6F6F6F6F706F70706F6F
706F70706F6F706F6F70606F7F6F70706F6F7F6F6F70606F7F6F6F70606F706F6F70606F706F6F\
6F60606F6060605F606060605F5F5F5F5F5F50505F5F5F5F5F5F5F5F705F507F50
5F9F5F509F5050A0504FAF5050B0504FB050000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005050505F5F5F60606F6060606F6F6F
6060606F6F6F6060606F606F6F60606F6F6F6060606F606F6F60606F606F6F60606F606F6F6060\
6F6060605F5F605F605F5F5F5F5F5F50505F5F5F5F5050505F5F5F506050507F5F
508F50509F504F9F5050AF504FAF4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000505050606060606060
60606F6060606F6F6F6060606F6F6F6060606F6F6F6060606F6F6F6060606F6F6F6F60606F606F\
6060606060605F5F5F5F5F605F5F5F5F5F605F5F5F5F5F5F505F5F5F6F5F507050
508F5F509050509F504FA04F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000505050
5050505F5F5F5F5F5F60606060606060606060606060606F6060606060606060606F6060606060\
6060605F5F5F5F5F605F5F5F5F5F5F5F5F5F5F5F5F50505F5F5F5F50605050705F
4F7F505090504F904F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005050505050505F5F5F5F5F5F6060606060606F6F6F6060606F606F6060606F606F\
6060606060605F5F606060605F5F5F5F5F5F505F5F5F5F5F5F5F5F5F605F507050
508050000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005050505050505F5F5F5F5F5F5F60605F5F60606060606060\
6060605F5F5F5F5F5F5050505F5F5F5050505F5F5F505050506050000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005050505F5F5F5F5F5F5F6060\
5F5F5F5F5F5F5050505F5F5F505050505050000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000505050\
5050505050504F5050000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[6]:=",
  ImageSize->{288, 286.812},
  ImageMargins->{{1, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P00017PP10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`N90003RZB:01BTL0091`001`N4000RRjVTZJBYY:VTZZBZY<6T`JC1Y<6T`JC1Y<6T`JC1
Y;bT_6X02@L000L7P@00:HZTRZB;ZJBYY:VTZJBZY:ZTZZBZY:ZT`JC1Y<6T`JC1Y:ZTYJBUY:FT06H0
2@L000L7OP00;hZTY:VTZJBZY<6T`JC1ZL6Z`J[1Z/6Z`J[1Z/6Z`J[1Z/6Z`J[1YL6U`JG1Y<6T06<0
2@L000L7N`001HZTRZB:00:T02nYY:VTZZBZY<6T`JG1YL6U`J[1Z/6U`JG1YL6U`JG1YL6U`JG1YL6U
`JG1Y;bT_01O00T700071gX003V;ZJBYY:VTZZC1Y<6Y`J[1Z/6Z`/72`L;1`/72`L;1`/72`L;1`/72
`L;1`/72`L:U`ZG2YL6U`JD0G@091`001`Mh0004Y8ZTRP:T03NYY:VTZZC1Y<6U`JG1Z/6Z`Z[2`L;1
`/72`L;1`/72`L:Z`/72`L:Z`Z[1YL6U`JG1YL6U`JFl05/02@L000L7M0003hZYRjVTZJBZY<6T`JW1
ZP02`@04`/72`@C201g7`/O2al;7`/O2al;7`/O2al;7`/O2al;7`/O2a`04`P0<`L;1`/72_<:l`[c1
EP091`001`Mb0004RhZTRP:T00bYY:VTZZC1Y<6U`JX2`@0Q`/72`L;1`/72`LO1al77`LO2al;7`/O2
al;7`/O2al;700K200o1`/72`L:Z`ZG2YL6U`JD0E0091`001`Ma000DY8ZYRjVTZJBZY<6Y`J[1Z/;1
`/42`P0Yal;7`/O2al;7`/O2al;7`/O2al;7`/O2al;7`/O2al;7`/O2al;7`/L01<802/72`L;1`[c2
_<9B00T700071g0003f;RZB:ZJBYY:VT`JC1YL6Z`J[2`L;1`/72`L;1al77`LO1al77`LO2al;7`/O2
al;7`/O2al;7`/O2al;700K200g1`/72Z/:U`ZG2YL6U05002@L000L7K`009ZV;ZJBYY:VTZZW1Z/6Z
`/72`L;1al;7`/O2al;7`/O2al;7`/O23LL05<;7`/O2al;7`/O2al;7`/O2al;70/802/72`L;1`/72
_<9>00T700071fd002V:Y8ZYRjVTZJBYY:ZT`JG1Z/;1`/72`L;1`/77`LO1al77`LO1al77`P0;a`0D
`/O2al;7`/O2al;7`/O2al;7`/L2`P0<`L;1`Z[2YL:l`[c2C@091`001`M[000>RhZYRZVTZJBZZL6Y
`JX2`@0@`/72`L;1al77`/O2al;7`R7700k2al;7`/O2al;7`/O2a`K200G1`/72`@1;00T700071fX0
02B;RZB:ZHZYY:VTZZC1YL6Z`J[2`L;1`/72`LO1al77`LO2alHKa`0@`/O2al;7`/O2al;7`/O2a`C2
00[1`/72`L:l`[c2B@091`001`MX000?LH^:ZHZYRjVTZZW1ZL6Z00C100c2`LO1al77`LO2al8<a`03
1lL700?700<7alL05/L03<;7`/O2al;7`/O2a`[2B0091`001`MW000QLH^:Y8ZYRZV;ZJBYY<6T`J[1
Z/;1`/72`L;1al77`LO102O700k2al;7`/O2al;7`/O2a`C200S1`/72`L:l`TL02@L000L7I@0048^:
ZHZYRjVTZJBZZL6Y`JX2`@04`/72`@;200K7`/O2al88a`0O1lL7a`O71lL7a`O71lL7a`O71lL7a`O7
1lL7a`O71`0>a`0>`/O2al;7`/O2al;7`/L6`T@02@L000L7I0009XYaRhZTRZV;ZJBYY:ZT`JC1YL6Z
`/72`L;1`/77`LO1al77a/O62/L03@O71lL7a`O71lL7a`L04/L03/;7`/O2al;7`/O2al;71<802<72
`L;1`/72@`091`001`MS000@RhZ@RZV;ZJBYY:ZY`JW1ZP;100c2`L;1`/77`/O2al88a`0U1lL7a`O7
1lL7a`O71lL7a`O71lL7a`O71lL7a`O71lL7a`O71`0>a`0<`/O2al;7`/O2al;71/800jD0000o00T7
00071f8001B:LH^:RhZYRjVTZJBYY<6T`JG1ZP;10132`L;1`/77`LO1al77a/O62<L05@O71lL7a`O7
1lL7a`O71lL7a`O71`0Ba`0>`/O2al;7`/O2al;7`/L:`P03Y@0003h02@L000L7H0004G6;RY2:ZH^Y
Y:VTZZW1ZL6Z00;100c2`L;1`/77`/O2al88a`0:1lL7a`O71lL7a`l701;71lL7a`O71lL7a`O71lL7
a`L:a`0>`/O2al;7`/O2al;7`/L7`Sl02@L000L7G`005G6:LH^:RhZYRjVTZJBYY:ZT`JG1ZP02`@0@
`/72`L;1al77`LO1alK7aPK70247a`O71lL7a`O71lL7a`O71lL7a`O71lL7a`O71lL7a`L03/L03<;7
`/O2al;7`/O2a`[2?`091`001`MN000;LH^:T8ZYRZV;ZJ@00ZT01ZZY`JW1ZP;100c2`L;1al77`/O2
al86a`081lL7a`O71lLG1`0@a`O71lL7a`O71lL7a`O71`[70132al;7`/O2al;7`/O2al;70l800lNU
000l00T700071ed002EaRW6;RX^:ZHZYRjVTZJBZY:ZT`JG1Z/;1`/72`L;1al77`LO600S702D7a`O7
1lL7a`O71lL7a`O71lL7/0N`1lL7a`O71lL7a`O71lL700c700c2al;7`/O2al;7`/L<`P03Y@0003X0
2@L000L7G@002X^:T8ZYRZV;ZJ@2Z@06ZZW1ZL6Z0/403<;1`/77`LO1al;7`PS700P7a`O71lL7aaT7
00k71lL7a`O71lL7a`O71`[701O2al;7`/O2al;7`/O2al;7`/O2al;7ZP0k00T700071e`00W4058^:
RhZ@RZV:ZJBYY:VTZZC1Z/6Z0/403/;1`/72`L;1al77`LO61/L0:PO71lL7a`O71lL7a`O71k07/0N`
1k07/0N`1k07/0N`1k07[`N_1k37/0_700[2al;7`/O2al;73/800jD0000h00T700071e/001B;LH^:
ZHZYRjVTZJBZZL6Y`J[1ZP;100c2`L;1`/77`LO2al86a`061lL7a`O76@L05<L7a`O71lL7a`O71lL7
a`O71lL72/L06<;7`/O2al;7`/O2al;7`/O2al;7`/O2ZSP02@L000L7FP005XYaRhZ;RY2:ZH^YY:VT
ZZC1Y<6U`JX2`@0>`/72`L;1`/77`LO1alH6a`0b1lL7a`O71lL7a`N`1k07/0Ne1kD7]@Ne1k07/0N_
1jl7[`N_1jl7[lN_ajo7[lN_ajl5a`06`/O2al;71L801lO2`/;7`/L02<82ZSL02@L000L7F@0058]a
RhZ@RZV;ZH^YY:VT`JW1ZL6Z0/403/;1`/72`LO1al;7`/O21/L01PO71lL7aaT701:e1k07/0N`1k07
/0O71lL7a`L<a`0H`/O2al;7`/O2al;7`/O2al;7`/O2al;70ZXf00T700071eP001R:LH]aRhZ;RZV:
ZH^YY:VTZZC1Y<6U`JX2`@0>`/72`L;1al77`LO1alH6a`0m1lL7a`O71lL7/0N`1k07]@Ne1kD7]@Ne
1kD7[`N_1jl7[`N_1jo7[lN_ajo7[lN_ajo7[lNZal;7`/O2a`0;`P03al;200G200?7`/800ZXe00T7
00071eL001J;LH^:T8ZYRjV;ZJBYY:ZT`JW1ZL6Z0/403<;1`/72`LO1al;7`PS700H7a`O71lLE1`0K
]@Ne1kD7/0N`1k07/0N`1k07[`N_ajo7[lN_00G701g2al;7`/O2al;7`/O2al;7`/O2al;7`/O2al;7
ZP0e00T700071eH00W405X]aRhZ@RY2:ZH^YY:VTZZC1Y<6U`JX2`@0@`/72`L;1`/77`LO1alK7aPC7
03l7a`O71lL7a`N`1k07/0Ne1kD7]@Ne1kD7]@Ne1jl7[`N_1jl7[kF_1jo7[lN_ajo7[lN_aj[7Z/NZ
al;7`/L03L800lO2`P05`P04al;2TC@02@L000L7E@003X]aRhZ@RY2:ZH^YRjVT0ZT01ZZY`JW1ZPC1
00c2`L;1al77`/O2al86a`081lL7a`O71lLC1`10]@Ne1kD7]@Ne1k07[`N_1jl7[`N_ajo7[lN_ajo7
[lNZal;7`/O2al;7`/O2al;7`/O2al;7`/O2al;7`/O2/3@02@L000L7E@006G6;LH^:T8Z@RZV:ZH^Y
Y:VTZZC1Y<6U`JX00/403/;1`/72`L;1al77`LO61/L0:`O71lL7a`O71k07/0Ne1kD7]@Ne1kD7]@Ne
1kD7[`N_1jl7[kF_]Jn`[k001Jl03/N_aj[7Z/NZaj[7Z/O11l800l72`P0<`P03al;700:Z<`091`00
1`MD000=LH]aT8Z@RZV:ZH^YY002Z@08ZZVZZL6Z`JX2`@0>`/72`L;1al77`LO2al86a`081lL7a`O7
1lL=1`0`]@Ne1kD7]@Ne1kD7]@Ne1jl7[`N_1jl7[`N_ajo7[lN_aj[7Z/NZaj[7`/O2al;71L804/O2
al;7`/O2al;7`/O2al;7ZS<02@L000L7D`006g6;LH]aRhZ@RY2:ZH^YY:VTZJC1Y<6U`J[1ZP02`@0>
`/72`L;1al77`LO1alH6a`0K1lL7a`O71lL7/0Ne1kD7]@Ne1kD7]@Ne1kD700:e00b_]Jne[kF_]Jne
[k0:[`0=ZZnZ[j[6Z/:Z`Z[2`@03`P05`L;1`/402/802<O2al;7Z[2A<P091`001`MC000>Rg6@RY2:
T8ZYRZV;ZJ@2Z@06ZZW1ZL6Z1<403<;1`/77`LO1al;7`PK700X7a`O71lL7a`O72@L09;D7]@Ne1kD7
]@Ne1kD7]@Ne1k@7[`N_1jne[kF_]Jn`[k2_/0:_00bZaj[7Z/NZaj[7`LL;`P0:al;7`/O2al;7`P?7
00>Z/000<@091`001`MB009a01R:LH^:RhZ;RY2:ZH^YY:VTZZBZY<6Z`JX4`@0<`/72`L;1al77`LO6
2<L09@O71k07/0N`1k07]@Ne1kD7]@Ne1kBe];F_]Jne[kF_]:nd[k@00Zl049J_UZnF[iJ_UZnZ[jZ_
ZZl7ZP0>`J[2Z/;1`/72`L;1`/46`P04al;7`P?700>Z/940<0091`001`MB000ELH^:T8Z@RZV;ZH^Y
Y:VTZZW1ZL6Z00C100k2`L;1`/77`LO1al;7`PK700X7a`O71lL7a`O72`L02;D7]@Ne1kD70[D04[Be
];Fd]KBe[kB_]9Jd[kBF]0R_00nZ[jZ_ZZnZaZ[7Z/:Z`/402/802/O2al;7`/O2al83a`:`<0091`00
1`MA000ILH]aRhZ;RX^:T8ZYRjV;ZJBYY:ZT`J[1ZP06`@0<`/72`LO1al77`LO62<L04@N`1k07/0N`
1k07]@Ne1kD700:e02Vd]KBe[kFF]IJdU[BF]9JdUZnF[iF_UJnE[iF_T:n@[i2_T:Z@ZZVZZ@04ZP0>
`J[1YL;1`/72`L;1`/44`P04al;7`P?700B_/:Z_;`091`001`M@000GLH]aT8Z@RY2:ZH^YRjVTZZBZ
ZL6Y`JX01<403/;1`/72`LO1al77`/O21/L02PO71lL7a`O71lL91`06]@Ne1kD70[D09KBe];FK]I^d
U[BF]9JdU[BE]9F_UJnE[iF_UZn@[i2_ZJnY[jT01:X01/6Z`Z[2`@S200[7`/O2al;7`/O20lL00k37
[`0_00T700071e0001Z:LH]aRhZ;RX^:T8ZYRjVTZJBZY:ZT`J[1ZPK100c2`L;1al77`LO1alH8a`0=
1k07/0N`1k07/0Ne1`04]@0>];F_]IJeU[BF]9JdUY/2UP0]UIJEUYFFUIJEUYFFT9J@UY2FT:Z@ZI2Y
T:Z@ZZVZY:ZU`JG1_<:l`[c2`L;100C200C7`/O20lL01;37[k0_00T700071dl001R;LI2:T8Z@RZV;
ZH^YRjVTZZW1ZL6Y`JX4`@0>`/72`LO1al77`LO2al86a`0:1lL7a`O71lL7a`<700Re1`L7]@Ne1`Be
02nd]I^eVkFK]9JdU[BF]9FdUKBE[iF_UJnE[i2_T:n@[i2_T:n@ZY2ZZJZYZZW1ZP02`@04`/72`@C2
00S7`/O2al;7`PG700>`ajl0;`091`001`M>009a01Z;LH^:RhZ;RY2:T8ZYRjVTZJBZY<6T`J[1ZPK1
00[2`LO1al77`LO12/L02`N`1k07/0N`1k0700:e01Bd]KBe];FF]IJdUY^FViJKUIJEUPRE00R@UI2E
T9F@U@V@01BYRjV;ZH^ZY:ZU`JG2_<:l`/72`@K200?7`/L01<L00jn`UP0^00T700071dh001MaT76@
RY2:T8ZYRjV;ZJBZZJZY`JW1Z@06`@0<`/72`LO1al77`LO22<L02PO71lL7a`O71lL31`06]@Ne1kD7
1;D04[BeVkFK]9JdUY^EViFKUI^EV`FE02:FUIJ@UY2FT9J@UY2_T:Z@ZY2ZT:Z@ZZVZZL6Z`/72`L;1
1<801/O2al;7`PO70[0^00T700071dd001eaRW6;LH^:RhZ@RY2:ZH^YRjVTZJBZY<6T`J[1ZP06`@0:
`/77`LO1al77`@[70307/0N`1k07/0N`]KBe];F_]IJdUY^FViJKUI^EVgbEO9ElUGbEO9EfUGNEMiEg
UGL;T00ERi2;ZH^YRjZTZZG1_<:l`[c2`L;100C200?7`/L01<L00jn`[`0^00T700071dd000j@LI2:
T8Z@RZV;ZH^YR`:Y00RZZL6Y`JW1Z@C100k2`L;1`/77`LO1al77`PS70147a`O71lL7a`O71k07]@Ne
1`02]@08];Fd]I^eVk@2V`0:UY^EViFKUI^EVPRE00R@UI2ET9F@U@V@00RYT:V@ZH^ZZ@:Z0/401<;1
`/44`P06al;7`/O21<L01@O71k2e02h02@L000L7C002L@0LRg6;LH^:T8Z@RZV:ZH^YY:VTZJBZY<6Z
`J[1ZP;100k2`L;1`/77`LO1al77aPS70507a`N`1k07/0N`1jne[kFF]IJdUY^FViFKO9]lUGbEO9El
UGJEMYEfUGJEMYEfT7J@MY1fT7J@MY1bT8Z@Ri2:T8Z@RjV;ZZC1YL6U`[c2`@K200?7`/L01<L01:o7
[k0]00T700071d`001]aRg6@RY2:T8ZYRZV;ZH^YY:ZYZZW1ZL6Z`JX00/403<;1`/77`LO1al77`PS7
03X7a`O71lL7a`O71lL7]@Ne1kBeVkFK]I^dUY^FViFKUI]lUGbEO9EgUGNEMYEfUGJEMYEfT7J@MY1f
2i003h^@RjV;ZZC1YL6l`/72`@04`P04al;7`PC700H7a`N`1jl]00T700071d/001eaRW6;LH^:RhZ@
RY2:ZHZYRjVTZJBZY:ZY`J[1ZP04`@0>`/72`L;1al77`LO1alH6a`1B1lL7a`N`1k07/0N`]JneU[FF
]9JKUY]lVgbEO9ElUG^EMYEfUGJ@MY1fMgIgMWMfT7J@MY1fT76@LI1aT76@RY2:T8Z@Ri2;ZZBZYL6U
`ZG2`@K200?7`/L00lL01@N_1jne02d02@L000L7B`006X]aRhZ@RY2:T8ZYRjV;ZH^YY:ZYZZW1ZL6Z
1<403<;1al77`LO1al77`PS700l7a`O71lL7a`O71lL7/0L00[D0@kBeVkFK]9JKUY^EVgbEO9EgUGJE
MYEfT7J@MY1fT7J@MY1fT7J@MY1fT7J@MY1bT8^@Ri2;T8^@RjVTZZG1_<;1`/401<801<O2al82a`08
1lL7a`N`1k0]00T700071dX00W4078]aRg6;RX^:T8Z@RZV;ZH^YY:VTZZBZZL6Z`JX4`@0>`/72`LO1
al77`LO6alH8a`0O1lO7/0N`1k07/;FF]IJdUY^FVgbKO9ElUGJEMWMfM`03MPX002AfLGIaMW6@LI1a
T76@LI1aT76@LI2:T8Z@RjVTZZG1YL:U`/46`P0;al;7a`O71jl7[kD0;@091`001`M:000KLH]aRhZ@
RY2:ZHZYRjV;ZH^YY:ZYZZW1ZL6Z00;100k2`L;1al77`LO1al77`PS70247a`O71lL7a`O71k07/0N`
]IJeU[FFViFKO9ElUGJEMWL05P007G6@LI1aT76@LI1bT8^@Ri2;T8^YRjZT`Kc2`L;100C200C7`/O2
0/L020O71lL7/0N`;@091`001`M9000OLH]aRg6;LH^:T8Z@RZV:ZH^YRjVTZJBZZJZY`J[1ZP02`@0@
`/72`L;1al77`LO1alK7aPK701T7a`O71k07/0N`1k2eU[FFViJKO9ElUGIg009f7@02L@0HLW6@LI1a
T76@LI2:T8ZYRjZT`JG1YL;11/84a`071lL7/;F_]@0]00T700071dT001b;LH^:T8Z@RY2:ZHZYRjV;
ZH^ZZJZY`JW1Z/6Z0/403<;1`/77`LO1al77`PS701`7a`O71lL7a`O71lL7/0N`1k2eU[FFViFFO9Ef
90005X]aT76@LI2:T8Z@Ri2;ZZC1_<;1`/44`P03al;700?700P7a`O71k07/2d02@L000L7B@007g6;
LH]aRhZ@RY2:T8ZYRZV;ZH^YY:ZTZZW1Z/6Z`JX01<403/;1`/77`LO1al77a/O61/L00`O71`02a`0A
/0N`1k07[kFFViJKO9EfUGH09`03L@0BRg6;LI1aT8ZYRjZT`JG1YL;11<800lO2a`03a`071k07/;F_
]@0]00T700071dT001b;LI2:T8Z@RY2:ZH^YRjV;ZJBZZJZY`JW1Z/6Z0/403<;1al77`LO1al77`@S7
01D7a`O71lL7a`O71lL7/0N`]JneUY/00YD00gJE000Z000BRg6@LI2:T8Z@RjZT`Kc2`L;11<800lO2
a`03a`051lL7/0L00kD]00T700071dT001eaRg6;LI1aT8Z@RY2:ZH^YRjVTZJBZY:ZT`JW1ZP08`@0<
`/77`LO1al77a/O62/L03@N`1k07/;FF]IJEMW`0<@00479aRg6@LJVTZZG1YL6U`/44`P03al;700?7
00L7/0N`]Jne02d02@L000L7B0004G6@LI2:T8Z@RY2:ZH^YT:V@00:Y00ZZZL6Y`JW1Z/6Z0/403<;1
al77`LO1al77`@S701<7a`O71lL7a`O71lL7/0N`]IJK009l<`0028]aT76@RjZT0/400l;1`P03`P04
al;7`P;700<7a`L00PL01KD7]KFK02d02@L000L7B002L@0LRg6@LI2:T8Z@RY2:ZH^YRjVTZJBZY:ZT
`JW1ZPS100c2`LO1al77`LO6alH:a`0:1k07/0N`]IJKMSL000=aRg400Z@01ZZU`JG2_0C200?7`/L0
0lL02@N`1k2e[kFdV`0]00T700071dP000maT8Z@RY2:T8ZYRZV;ZI001:T02ZZY`JW1ZL6Z`JX2`@0:
`/77`LO1al77`@S701@7a`O71lL7a`O71lL7a`N`1k2eUSP000J@LI2;ZZD2`@03`/7200?200?7`/L0
0lL00`O71`021`05]@Ne]I/0;@091`001`M7000OLHYaRg6@LI2:T8Z@RY2;ZH^YRjVTZJBZY<6Y`J[1
ZP08`@0<al77`LO1alK7a/O61/L03@O71lL7/0N`1k2eUY/0>@002X]aZJC1YL6U`[`4`P03al;700?7
00T7/0N`]KBeUY/0;@091`001`M7000@Rg6@RY2:T8Z@RZV:ZH^YT0BY00ZZZL6Y`JW1Z/6Z0/403<;1
al77`LO1al77aPK701@7a`O71lL7a`O71lL7a`N`1jneUST000MaT:C1_<;100C200C7`/O20/L01`O7
1lL7]@L00kD00kBJ000/00T700071dL001maRg6;LI1aT8Z@RY2:T8^YRjV;ZJBYY:ZT`JW1Z/6Z00K1
00k2`LO1al77`LO6alK7aPS700X7a`N`1k07/;FF>P0027::ZZG1YL6l1<800lO2a`03a`051lL7/0L0
0[D01;BeUYX]00T700071dH0015aT76@RY2:T8Z@RZV;ZI2YT002Z@0:ZZVZZL6Y`J[1ZP;100c2`LO1
al77`LO1al48a`0C1lL7a`O71lL7a`O71lL7/0NF]@0j0003LJVU00;100?2`L800l800lO2a`03a`03
1lL7008700Be1kFe0Y/00h40000[00T700071dH00W407X]aT76@RY2:T8Z@RY2;ZH^YY:VTZJBZZL6Y
`J[1ZPK100k2`LO1al77`LO6alK7aPC700/7a`O71lL7/0N`1`02/003UP0003P00X/01/6U`JG2_0C2
00?7`/L00lL02PN`1k2e];FdVg`^00T700071dH0015aT76@RY2:T8Z@RZV;ZI2YT002Z@0:ZZVZZL6Y
`J[1ZP;100k7`LO1al77`LO1al77aPC701@7a`O71lL7a`O71lL7a`O71k07US/000F;Z[c2`@06`PC7
00<7a`L010L4]@:K;P091`001`M5000QLH]aRg6@RY2:T8Z@RY2:ZH^YRjVTZJBZZJZY`J[1Z/6Z00C1
00k7`LO1al77a/O6alK7aPC700l7a`O71lL7/0N`1k07/9H0>P001hZYYL:U`[`01<86a`091lL7/0Nd
]KBe00:K00=k0000;0091`001`M5000BT76@LI2:T8Z@Ri2;ZH^YRjV@0ZT02ZZYZZW1ZL6Z`JX4`@0:
al77`LO1al77aPS70187a`O71lL7a`O71lL7a`O71k0j0004ZJG1_0K200K7`/O71lL51`04]@Ne]@:K
00>J0000;@091`001`M5000SLH]aT76;RY2:T8Z@RY2:ZH^YRjVTZJBZZJZY`J[1Z/6Z`JX00/403<O1
al77`LO6alK7aPK700l7a`O71lL7/0N`1k07/0L0>@001X^T`JG2Y@K21<L01@O71k0700:e00Jd]I^K
PIX_00T700071d@001maT76@LI2:T8Z@RY2;ZH^YRjV@ZZVZZJZY`JW1Z/6Z00C100S7`LO1al77`@[7
0187a`O71lL7a`O71lL7a`O71k0h0006T:C1_<;11<801/O2alL7a`D700Je1kFe];D2V`03P@0002d0
2@L000L7A00098]aT76@LI2:T8Z@RY2:T8^YRjV;ZJBYY:ZYZZW1Z/6Z`J[1ZP;100c7`LO1al77a/O6
alH4a`0A1lL7a`O71lL7/0N`1k07/0L0=`02R`06`JG1YL:m1<803lO2alL7a`O71kD7];Fd]@02V`03
O00002h02@L000L7A0007g6@LI1aT8Z@RY2;T8^YRjV@ZI2ZZJZYZZW1Z/6Z`JX01<402<O1al77`LO1
2<L050O71lL7a`O71lL7a`O71lL7a`N`=`001H_1_<;100C200?7`/L00lL00`O71`041`04]@Ne]@:K
0YX`00T700071d@003:;LH]aT76@RY2:T8Z@RY2;ZH^YRjVTZZBZZJZY`J[1Z/6Z`J[1Z/O1al77`LO6
alK7aPC70147a`O71lL7a`N`1k07/0N`1`0f0007R/6U`JG2_004`P03al;700?700T7a`N`1kBe];D0
0Y/00h6J000`00T700071d<0022@LI1aT8Z@RY2:T8^@RjV;ZI2YT:ZYZZVZZL6Z`J[1ZPC100S7`LO1
al77`@S701@7a`O71lL7a`O71lL7a`O71lL7/3D000F:`Kc2`@04`P04al;7`P;700<7a`L010L00kD7
]@02]@04]9ZKPC402@L000L7@`006g6;LI1aT76@RY2:T8Z@RY2;ZH^YRjVTZZBZZ@02ZP0F`J[1Z/6Z
`J[6Z/O1al77`LO6alK7aPC70147a`O71lL7a`N`1k07/0N`1`0e0006ZZG1YL:l1<800lO2a`03a`05
1lL7/0L00[D01;Be];D2V`03P@0003002@L000L7@`00891aT76@RY2:T8Z@Ri2;ZI2YT:V@ZZVZZJZY
`J[1Z/6Z0/402/O1al77`LO1al48a`0<1lL7a`O71lL7a`O71`L00kD0000b0004`Kc2`@K200S7`/O7
1lL7a`<700>e1kD00kD2V`03P@0003402@L000L7@P02L@0FRg6;LH]aT8Z@RY2:T8Z@RjV;ZJBYY0:Y
00bZZL6Y`J[1Z/6Z`JX2`@0Qa/77`LO1alK7a/O6alK7aPO71lL7a`O71k07/0N`1k0703<000K1YL:U
`[`4`P03al;700?700/7a`N_1jne];FK]@02V`03P@0003802@L000L7@P004g6@LI2:T8Z@RY2:ZH^Y
RjV@ZI000ZT06ZZYZZVZZL6Z`J[6Z/O1al77`LO1alK7a/O60/L030O71lL7a`O71lL7a`T700>e1kD0
<P001:Zl`/44`P03al;700?700<7a`L010L02;D7]KFd]I^d0YXd00T700071d8001J;LH]aRg6;LI2:
T8Z@RY2:ZH^YRjVT1:T03:ZY`JW1Z/6Z`J[1ZP;100k6`LO1alK7a/O6alK7aP;70147a`O71lL7a`N`
1k07/0N`1`0a0006Y:G2YL:l1<800lO2a`03a`0;1lL7/0N_]KBe];D00Y/2VSD02@L000L7@P004g6@
LI2:T8Z@RY2:T8^YRjV@ZI000ZT06:ZYZZW1ZL6Z`J[6`LO1al77`LO1alK7aPC700X7a`O71lL7a`O7
2`L00kD7]@0`0004`Kc2`@C200C7`/O20/L00`O71`061`09]@Nd]I^dVY^103D02@L000L7@@005g6:
LH]aRg6@LI2:T8Z@RY2:ZH^YRjVT00:Y01jZZJZY`JW1Z/6Z`J[1Z/K1a/77`LO6alK7a/O6alH2a`0A
1lL7a`O71k07/0N`1k07/0L0;`001/6U`JG2_0C200?7`/L00lL02`O71k07[`Nd]KBe00:K0YX00h6J
000d00T700071d4001:@LI1aT8Z@RY2:T8ZYRjV;ZI02Z@0JZZVZZJZY`JW1Z/JZa/77`LO1al77`LO6
alH4a`081lL7a`O71lL;1`05]@Ne1kD0;P001<6l`/44`P04al;7`P;700<7a`L01PL01[D7];Fd]@:K
0YXf00T700071d4001MaRg6;LH]aT76@RY2:T8Z@RZV;ZH^YY002Z@0aZZVZZL6Y`J[1Z/6Z`J[6Z/O1
al77a/O6alK7a/O61lL7a`O71lL7/0N`1k07/0N`1`0]0006`JG1YL:l1<800lO2a`03a`0B1lL7/0N_
1kBe];FK]9ZKPIYk=P091`001`M1000BT76@RY2:T8Z@RY2;ZH^YT:V@0ZT06ZZYZZVZZL6Y`J[1Z/K1
al77`LO1al77a/O61<L020O71lL7a`O72`L01KD7]@Ne02/000FT`Kc2`@04`P04al;7`P;700<7a`L0
10L01;D7]@L2]@04];FKV`:J00>10000=@091`001`M1000ELH]aRg6;LI1aT8Z@RY2:T8ZYRjV;00BY
036ZZJZY`JW1Z/6Z`J[1Z/JZaZ[7`LO6alK7a/O6alH7a`O71lL7a`N`1k07/0N`1k0702X000NT`JG1
YL:l00C200?7`/L00lL03@O71k07[`Nd]KBeVk@00Y/00h6JP00h00T700071d0001=aT76@RY2:T8Z@
RY2;ZH^YT:V@00:Y01ZZZJZYZZW1ZL6ZaZ[6`LO1al77`LO6alK7aP;700X7a`O71lL7a`O72`L01KD7
]@Ne02T000Fl`[c2`@06`P08al;7a`O71lL31`04]@Ne1`:e00Nd]I^dVY^103T02@L000L7@000=8]a
T76@LI1aT8Z@RY2:T8^@RjV;ZH^YY:ZYZZVZZL6Z`J[1Z/6ZaZ[7Z/NZaj[7a/O6alH4a`0C1lL7a`O7
1k07/0N`1k07/0Ne1`0W0006ZZG2YL:l1/86a`0=1lL7/0N_1kBe];FK]002VP03PIZ003X02@L000L7
@0003G6@LI1aT8Z@RY2;T8/00Y008:V@ZI2ZT:ZYZZVZZL6Y`J[1Z/K1a/77`LO1al77a/O61<L02PO7
1lL7a`O71lL91`05]@Ne1kD09P001<6l`/46`P03al;700?700<7a`L010L02;D7]0Nd]KBe0Y/2VP03
P@0003T02@L000L7@0005X]aT76@LI1aT8Z@RY2:T8^@RjV;ZH/2Z@0LZZVZZJZY`J[1Z/6Z`J[6Z/NZ
aj[7[lO6alK7aPC70147a`O71lL7/0N`1k07/0Ne1`02]BD000K1YL6U`[`4`P03al;700G700d7a`N`
1jl7];Fd]I^d00:K00>1VX00?0091`001`M0000_LI1aT8Z@RY2:T8^@RjV@ZI2YT:ZYZZVZZJZY`JW1
Z/6Za/76`LO1al77`LO6alH01<L02PO71lL7a`O71lL91`05]@Ne1kD08`001JG1_<;100C200C7`/O2
1<L00`O71`041`0:]@Nd1kBe];FK]0:J00>10000>`091`001`M0000FT76@LI1aT76@RY2:T8Z@Ri2;
ZH^YR`:Y00BZZJZY0ZX05/6Z`J[1Z/6ZaZ[7Z/NZajo7[lO6alH4a`0A1lL7a`O71k07/0N`1kD7]@L0
0[DR0007YL6U`[c2_004`P03al;700?70147a`O71k07[`Nd]KBeVkBJV`02VP03P84003`02@L000L7
@000;g6@LI2:T8Z@RY2;T8^YT:V@ZI2ZZJZYZZVZZL6Y`J[1Z/K1a/77`LO1al77a/O600C700X7a`O7
1lL7a`O71`L01kD7]@Ne1kD080001ZZU`[c2`@C200C7`/O21<L00`O71`041`06]@Nd1kBe0[@01Y^d
VY^1VSh02@L000L7?`005g6@LI1aT76@LI2:T8Z@RY2;ZH^YRjVT00:Y00BZZJZY0ZX05/6Z`J[1Z/6Z
aZ[7Z/NZajo7a/O6alH2a`0C1lL7a`O71lL7/0N`1k07]@Ne1`02]Al000UkMZBTYJBZO9D00l800lO2
a`03a`0A1lL7a`N`1jne];Fd]I^dVY/00YX00h6J000n00T700071cl000b@LI1aT8Z@RY2:T8/2T00R
ZI2YT:V@ZZVZZJZY`JW1ZL6ZaZ[6`LK1al77`LO6alK7aP;700X7a`O71lL7a`O72@L01kD7]@Ne1k@0
7P002:B;YJBZY:ZT0jX00l;7`P02a`041lL7a`D700ne1k@7];FK]9^dVY^1VX00?`091`001`Lo000M
LI1aT76@LI1aT8Z@RY2:T8^YRjV;ZH^ZZJZYZZT00ZX05/6Z`J[1Z/6ZaZ[7Z/NZajo7[lO6alH2a`0C
1lL7a`O71k07/0N`1k07]@Ne1`02]A`000f:Rh^UY:FTZZBZY:ZT00>Z0lL02`O71lL7[`N_]KBe00Bd
00NJViZJPIYk04002@L000L7?`00391aT76@RY2:T8Z@R`:@02:YT:V@ZI2ZZJZYZZW1ZL6Y`J[6Z/K1
a/77`LO1alK7a/O60/L02PO71lL7a`O71lL91`07]@Ne1kD7UP0K000;RjB;YJBZY:ZTZZT01:X02<6Z
[lL7a`O70`L01[D7];Fd]@:d00NJ]9ZKPIZ004402@L000L7?P005X]aRg6@LI1aT8Z@RY2:T8Z@RjV;
ZH/4Z@0cZZVZZJZY`J[1Z/6ZaZ[6Z/JZajo7a/O6alK7aPO61lL7a`O71lL7/0N`1k07/0Ne1jnK01P0
00F:Rh^TR`02Y00GZZBZY:ZTZZVZZJZYZZVZT:n`1jl7[kD00[@019^dVk@2VP04PIZ0PD<02@L000L7
?P00;g6@LI2:T8Z@Ri2;T8^YT:V@ZI2YT:ZYZZVZZJZY`J[6Z/JZa/77`LO1alK7a/O600C700P7a`O7
1lL7a`/700Be1kD70[D00iH0000D00:;00ZTRjFTZZBZY:ZY0ZX03L6Z`JZ_ZZnZ[jZ_T:l00PL02[@7
];Fd]I^dV[@2VP03PIX004<02@L000L7?P005X]aRg6;LI1aT8Z@RY2:T8Z@RjV;ZI04Z@0_ZZVZZJZY
`J[1Z/6ZaZ[6Z/JZajo7a/O6alK7aPO61lL7a`O71k07/0N`1k07/0L00[D00iJK000E00>;01^TRjBT
ZZBZY:ZTZZVZZJZYZY2ZT:n@[i2F[kD00[@019^dVk@4VP03P84004@02@L000L7?P00;g6@RY2:T8Z@
Ri2;ZH^YT:V@ZI2ZT:ZYZZVZZL6Y`J[6Z/JZaZ[7`LO6alK7a/O600;700X7a`O71lL7a`O72`L01;D7
]@L2]@03UP0001800X/2Y008ZZBZY:ZYZZT2ZP0I`JZZZZnY[jV_ZJn@[iF_UJnd]KBeVkBJV`02VP03
PIX004D02@L000L7?@005g6;LH]aT76@LI2:T8Z@RY2:ZH^YRjV@00BY02nZZJZYZZW1Z/6Z`J[6Z/JZ
aZ[7[lN_alK7a/O61lL7a`O71lL7/0N`1k07]@N`1`02]@03UY/0018001f;RX^;Y8^YY:ZTZZBZY:ZY
ZZVZT:Z@ZY2_T9J@UP03U@03]9^d00:K0YX0186JP85700T700071cd0032@LI2:T8Z@Ri2;T8^@RjV@
ZI2YT:Z@ZZVZZJZY`JW1Z/JZaZ[6Z/O1alK7a/O6alH2a`0:1lL7a`O71lL7a`/700Be1kD70[D00iH0
000?000=RZB;ZX^ZY:ZTZZVZZ@03ZP0HZJnY[jV_ZJn@[iF_UJnE[iFFZ[BK]9ZK0YX00h6J001700T7
00071cd001MaRg6;LI1aT8Z@RY2:T8Z@RZV;ZH^YT004Z@0bZZVZZJZY`J[1Z/JZaZ[6Z/J_ajo7[lO6
alK7aPO61lL7a`O71k07/0N`1kD7]@Nd]IHA000MRX^;Y8^TRjVTZZBZY:ZYZZVZT:Z@ZY2_T:n@UI00
1iD02IZKVYZ1VX2JN`1800T700071cd0032@LI2:T8Z@Ri2;T8^YT:V@ZI2YT:Z@ZZVZZJZY`J[1Z/JZ
aZ[6Z/O1alK7a/O6alH2a`081lL7a`O71lL;1`06]@Ne1kD70[D?00:;00bTRjZ;ZZBZY:ZYZZT3ZP0B
ZJnY[jV_T:n@[iF_UJnEUYFF0iD00kBJV`02VP03PIYk04T02@L000L7?@005g6;LH]aT76@RY2:T8Z@
RY2;ZH^YRjV@00BY03:ZZJZYZZW1Z/6ZaZ[6Z/JZaZo7[lN_alK7aPO61lL7a`O71lL7/0N`1kD7]@Nd
1kBeUPh001j;RZB;Y8^YRjZTZZBZY:ZYZZVZT:Z@ZY2_T:n@UI07U@09O9EkVX6JP9Yk04X02@L000L7
?0002g6@LI2:T8Z@Ri2;00:@02BYT:V@ZI2YT:Z@ZZVZZJZY`J[1Z/JZaZ[7Z/NZalK7a/O6alH2a`08
1lL7a`O71lL;1`06]@Ne1kD70[@=000>Y8^URjZ;ZZBZY:ZYZZT3ZP0@ZJZY[i2_T:n@[i2_UJnEUPJE
00MlUIZJPIYk04/02@L000L7?000BH]aRg6@LI1aT8Z@RY2:T8^@RjV;ZI2YT:ZYZZVZZJZY`J[1Z/6Z
aZ[6Z/NZajo7[lN_ajo7aPO71lL7a`O71k07/0N`1kD7]@L01;D00iH00008000NRhZTRjB;ZH^YY:ZT
ZZBZZJZYZY2ZT:Z@[i2_T9F@0iD03Y2EO9ElUG^ENgafVX1kC@091`001`Ll000=LI1aT8Z@RY2:T8Z@
R`02T006ZI2YT:V@0ZT08ZZYZZVZZLJYaZ[6Z/JZaZ[7`LO6alK7a/O61lH7a`O71lL;1`08]@Ne1kD7
]@L2V`P002::Y8^TRjZTZZBZY:ZYZZVZZJZY[jV_T:n@[i2_T:nE[iFF1YD027bENiEkO82ECP091`00
1`Lk000ILH]aT76@LI1aT8Z@RY2:T8^@RjV;ZI2YT002Z@0]ZZVZZJZY`J[1Z/6ZaZ[6Z/JZajo7[lN_
ajo7aPO71lL7a`O71k07/0N`1k0700Be0[@00g`00004000PRhZ;RjB;Y8^YRjVTZZBZZJZYZY2ZT:Z@
ZY2_T9F@UI05U@09O9EkUG^ENiEf00=kC`091`001`Lk000>T76@LI2:T8Z@RY2:T8/2T006ZI2YT:V@
0ZT08ZZYZZVZZL6YaZ[6Z/JZaZ[6Z/N_alK7a/O61lH7a`O71lL;1`06]@Ne1kD70[D00i/0000400:;
01nTRjZ;ZX^ZY:ZYZZVZZJZYZZV_T:n@[i2_T:nE[iFF00JE00UlUGbENiEfUGH0D@091`001`Lk000e
LH]aT76@LI1aT8Z@RY2:T8^@RjV@ZI2YT:ZYZZVZZJZY`J[1Z/JZaZ[6Z/N_ajo7[lN_ajl00/L04PO7
1lL7a`O71k07/0N`1kD7]0>e00>dV`000`008HZ;RjB;Y8^YRjZ;ZZBZY:ZYZZVZT:Z@ZY2_T:n@UI2E
T005U@0;O9EkUG^EMYEfO7H0DP091`001`Lk000jT76@LI2:T8Z@RY2;T8^YT:V@ZI2YT:ZYZZVZZJZY
aZW6Z/JZaZ[6Z/O1ajo7a/O6alH7a`O71lL7a`T700Re1kD7]@Ne1`:e02JK0000RZB;ZX^ZRjZ;ZZBZ
ZJZYZZVZZJZY[i2_T:n@[i2_UJnEUPJE00YlUGbENiEfUGIkDP091`001`Lk0019LH]aT76@LI1aT8Z@
RY2:T8^@RjV@ZI2YT:ZYZZVZZJZY`J[1Z/JZaZ[6Z/N_ajo7[lN_ajl7aPO71lL7a`O71k07/0N`1k@7
]003]@05]9/008X00X/07ZB;ZH^YRjZ;ZZBZZJZYZZVZT:Z@ZY2_T:n@UI2ET0FE00alUG^ENiEfO7Ik
MW]C00T700071cX003]aT76@LI2:T8Z@RY2;T8^YT:V@ZI2YT:ZYZZVZZJZYaZW6Z/JZaZ[6Z/JZajo7
a/O6alH7a`O71lL7a`091`08]@Ne1kD7]@L2]@0RUX^;Y8^ZRjZ;ZX^ZRjZYZZVZZJZYZZV_T:n@[i2_
T:n@[`RE00UlUG^ENiEfO7H0E@091`001`Lj000HRg6@LI1aT76@LI2:T8Z@Ri2;T8^YT:V@0ZT0=:ZY
ZZVZZJZY`J[1Z/JZaZ[6Z/N_ajo7[lN_ajl7a`O71lL7a`O71k07/0N_1k@7];Fd]HX4R`0NY8^YRjV;
ZX^ZRjZYZZVZT:Z@ZY2YT:n@[Y2ET9F@1ID02gbENiEkUGIlMW]f05H02@L000L7>P00>g6@LI2:T8Z@
RY2:T8^@RjV@ZI2YT:V@ZZVZZJZY`JW6ZLJZaZ[6Z/JZajo7[lN_alH7aPO71lL7a`O700T700Ve1kD7
]@Ne1kD00X/08JB;ZX^ZRjZ;ZZVZZJZYZZVZZJZ@[i2_T:n@[i2_T:nEUP06U@09O9ElUGJEMWaf05L0
2@L000L7>@006G6@LI1aT76@LI2:T8Z@RY2:T8^YRjV;ZI001:T01ZZYZZVZZ@:Z03C6Z/JZaZ[6Z/J_
ajo7[lN_ajl7[`N`1k07/0N`1k07/0Nd1kBe];F;Y8^TRjV;ZH^YRjV;0ZT08ZZ@ZI2ZT:Z@ZI2^T9F@
UI2ET9F@UGNEMYEfUGJEMWafNgIJ00T700071cT000Z@LI1aT8Z@Ri2;1Y00:ZV@ZI2ZT:Z@ZZVZZJZY
ZZW6Z/JZaZ[6Z/NZajo7a/O6alH7a`O71lL7a`/702Je1kD7]@N;ZX^ZRjZ;ZX^ZZJZYZZVZZJZY[i2_
T:n@[i2_T:nEUPRE00UlUG^ENiEfUGH0F`091`001`Li000ILI1aT76@LI1aT8Z@RY2:T8^@RjV;ZH^Y
T004Z@06ZZVZZJZY0ZX0E/JZaZ[6Z/JZaZo7[lN_ajo7[`N_1k07/0N`1k07/0Ne1k@7];FdY8^YRjV;
ZH^YRjV@ZY2ZT:V@ZI2YT:V@[Y2ET9F@UI2ET9EgUGJEMYEfO7IlMW]fG0091`001`Li000:T76@LI2:
T8^@R`J@02ZYT:V@ZI2ZT:ZYZZVZZJZYaZ[6Z/JZaZ[7Z/N_alK7a/O61lL7a`O71lL;1`0R]@Ne1kD7
RjZ;ZX^ZT:ZYZZVZZJZYZZV_T:n@[i2_T:n@[`ZE00UlUG^EMYEfO7H0G@091`001`Li000ILI1aT76@
LI1aT8Z@RY2:T8Z@RjV;ZH^YT004Z@06ZZVZZJZY0ZX0E<JZaZ[6Z/JZaZo7[lN_ajo7[`N_1k07/0N`
1k07]@Ne1kBe];FdZH^YRjV;ZH^ZT:V@ZI2YT:V@ZI2^T9F@UI2ET9F@UI2EMiEfUGJEMWafO7IkMUh0
2@L000L7>0002G6@LI1aT8Z@R`08T00ZZI2YT:Z@ZY2ZZJZYZZV_ZLJZaZ[6Z/JZaj[7[lO6alK7aPO7
1lL7a`O72@L09;D7]@Ne1kD7RjZ;ZY2ZZJZYZZVZZJZ@[i2_T:n@[i2_T:nEUPRE00UlUG^EMYEfO7H0
G`091`001`Lh000JRg6@LI1aT76@LI2:T8Z@RY2;T8^YRjV;ZI04Z@06ZZVZZJZY0ZX0=lJZaZ[6Z/JZ
aZo7[lN_ajo7[`N_1k07/0N`1k07/0Nd1kBe];F;ZH^YRjZ;ZY2ZT:V@ZI2YT:T00i005IF@UI2ET9F@
UGNEMYEfUGJEMWafN`02MV402@L000L7>0002g6@LI1aT8Z@Ri2;00J@02ZYT:V@ZI2ZT:ZYZZVZZLJY
aZ[6Z/JZaZ[7[lN_ajo7aPO61lL7a`O71lL91`0S]@Ne1kD7]JZ@ZY2ZZJZYZY2ZT:Z@[i2_T:n@[i2_
T:n@UI001ID02WbEO9EkUGIlMW]R00T700071cP001Z;LI1aT76@LI2:T8Z@RY2:T8^@RjV;ZI2YT0BY
00JZZJZYZZT2ZP0caZ[6Z/JZaZo6[lN_ajo7[lN_1jl7/0N`1k07[`Nd1k@7];Fd]H^YRjZ;ZY2ZT:V@
ZI2Y00F@01FET9F@UI2ET9EgUGJEMYEfO7IlMW/00WIS00T700071cL000Z@LI1aT76@RY2;1Y00:ZV@
ZI2YT:Z@ZY2ZZJZYZZV_Z/JZaZ[6Z/JZajo7[lN_alH7a`O71lL7a`T702>e1kD7]@Ne1i2ZZJZYZY2Z
T:n@[i2_T:n@[i2_T:n@UY2ET005U@09O9ElUG^EMWaf06D02@L000L7=`006g6@LI1aT76@LI2:T8Z@
RY2;T8^@RjV;ZI2YT002Z@08ZZVZZJZYZZT2ZP0aaZ[6Z/JZaZo6[lN_ajo7[`N_1k07/0N`1k07[`Nd
1kBe];FdZH^ZRjV@ZI2YT:V@Z@03T00HUI2ET9F@UI2EMiEfUGJEMYEfO7IkMW]fIP091`001`Lg000<
T76@LI2:T8Z@RY2;1900:ZV@ZI2YT:V@ZZVZZJZY[jW6ZLJYaZ[6Z/JZaZo7[lN_ajl7[`N`1k07a`T7
01fe1kD7]@NeZY2ZT:Z@ZY2ZT:n@[i2_T:n@[i2FT007U@0:O9ElUGJEMWafNfP02@L000L7=P004X]a
T76@LI1aT76@LI2:T8Z@R`B@00BYT:V@1:T0=jZYZZVZZJZY[jZ_Z/JZaZo6[lN_ajo7[`N_1jl7/0N`
1k07/0Ne1jl7]0Nd]I2ZT:V@ZI2YT:T00i005iF@UI2ET9F@UI2EMiEfUGJEMWafO7Ik009fJ@091`00
1`Lf000=LI1aT76@RY2:T8Z@R`04T00ZZI2YT:V@ZI2ZZJZYZZV_ZLJYaZW6Z/JZaZo6[lN_ajl7[`N_
1k07/0O71`L07kD7]@Ne1kD7]JZ@ZY2ZT:n@[i2_T:n@[i2_T9F@UI001ID02WbEO9EfUGIlMW]Z00T7
00071cH001B;LH]aT76@LI1aT8Z@RY2:T8Z@R`:@00JYT:V@ZI02Z@0cZZVZZJZYZZV_ZZnZ[j[6[lJ_
ajo7[lN_1jl7[`N`1k07/0N`1kD7]0Nd1kBeT:V@ZI2Y00F@01RET9F@UI2ET9F@UGNEMYEfUGIlMWaf
NgI/00T700071cH000eaT76@LI2:T8^@Ri2;00B@02RYT:V@ZI2YT:ZYZZV_ZJnYaZW6ZLJZaZ[6[lJ_
ajo7[`N_1jl7[`N`1`L08KD7]@Ne1kD7]@NdZY2_T:n@[i2_T:n@[i2ET9F@UI2ET003U@09O9EkUGIl
MW]f06d02@L000L7=P004X]aT76@LI1aT76@RY2:T8Z@R`B@00BYT:V@1:T03:ZYZZVZZJZY[jZ_ZPB_
0276[lN_ajo7[`N_1jl7/0N`1k07/0Ne1k@7];Fd]I2YT:T00i005iF@UI2ET9F@UI2EMiEfUGJEMYEf
O7Il00=fKP091`001`Lf000;LI1aT76@RY2;T8/01Y00::V@ZI2YT:V@ZZVZZJnY[jW6ZLJYaZ[6Z/J_
aZo7[lN_1jl7[`N_1k071`0[]@Ne1kD7]@Nd1kB_T:n@[i2_T:n@[Y2ET9F@UI2ET9EgUGbENiEfO7Ik
MP1_00T700071cH001:;LI1aT76@LI1aT8Z@RY2:T8/4T006ZI2YT:V@0ZT03:ZYZZVZZJnY[jZ_ZPB_
01o6[lJ_ajo7[`N_1jl7/0N`1k07]@Nd1kBe];Fd]I2Y00>@01NET9F@UI2ET9EgUGNEMYEfUGJEMWaf
N`03MW002@L000L7=P003G6@LI2:T8Z@Ri2;T8/01900::V@ZI2YT:V@ZZVZZJnY[jW6ZLJYaZ[6Z/J_
aZo6[lN_1jl7[`N_1k051`0[]@Ne1kD7]@Ne1k@7]:n@[i2_T:j@UI2ET9F@UI2ET9EgUGJEMWafO7Ik
MP1a00T700071cH001:;LI1aT76@LI1aT8Z@RY2:T8/4T006ZI2YT:V@0ZT03:ZYZZVZZJnY[jZ_ZPB_
03O6[lJ_ajl7[`N_1jl7/0N`1k07]0Nd1kBe];Fd]I2YT9F@UI2ET9F@UGNEMYEfUGJEMYEfO7Ik00=f
LP091`001`Le000:T76@LI2:T8Z@R`R@04JYT:V@ZY2ZT:ZYZZV_ZJnY[jW6Z/JZaZo6[lJ_ajl7[`N_
1jl7[`N`1kD7]@Ne1kD7]@Ne1kD7]@Nd[i2_T:n@UI2ET9F@1ID02GbENiEfUGIlMP1e00T700071cD0
01eaT76@LI1aT76@LI2:T8Z@Ri2;T8^YT:V@ZI2YT004Z@0:ZZVZZJZY[jV_ZPR_02o6[lN_1jl7[`N_
1jl7[`N_1jne];Fd]KBeT9F@UI2ET9F@UGJEMYEfUGIlMWafN`03MWH02@L000L7=@002Y1aT76@LI2:
T8/8T01BZI2YT:Z@ZY2ZT:nY[jV_ZJnYaZ[6Z/J_aZo6[lN_1jl7[`N_1jl7]@Ne1kD7]@Ne1kD7]@Ne
1kBe]:n@[i2ET9F@UI2ET9ElUGbENiEfO7IkMWL02@L000L7=0007W9aT76@LI1aT76@LI2:T8Z@Ri2;
T8^YT:V@ZI2YT0BY00ZZZJZY[jV_ZJnZ2Zl0:@N_1jl7[`N_1jl7[`N_]Jne];Fd]KBeT9F@UI2EMiEf
UGJEMYEfO7Il00AfN@091`001`Ld000;LI1aT76@RY2:T8/02900D:V@ZI2ZT:Z@ZY2ZZJnY[jV_ZJnZ
aZ[6[lJ_aZo7[`N_1jl7[`N_1kD7]@Ne1kD7]@Ne1kD7];Fd]KBET9F@UI2ET9F@UGNEO9EfUGIlMW]f
N@091`001`Ld000FLW6@LI1aT76@LI1aT8Z@RY2;T8^@R`:@00JYT:V@ZI04Z@0:ZZVZZJnY[jV_ZPZ_
02L7[`N_1jl7[`N_1jne[kF_]KBe];Fd]I2ET9EgUGJEMYEfUGIlMW`01GIj00T700071c@000]aT76@
LI2:T8^@R`08T00DZI2YT:V@ZY2ZT:nY[jV_ZJnY[jT2[`0haZo6[lJ_1jl7[`N_1jl7[`Ne1kD7]@Nd
1k@7]0Nd1kBe];Fd]I2ET9F@UI2EMiEfUGJEMWafNgIk00T700071c@001J;LI1aT76@LI1aT76@RY2:
T8^@Ri2;0Y001ZV@ZI2YT0JY00RZZJnY[jV_Z@Z_01G:[`N_1jl7[`N_]Jne[kF_]KBe];D00k@03IF@
UGJEMYEfUGJEMW/017Im00T700071c<000Z@LI1aT76@RY2;2Y005:V@ZI2YT:Z@ZY2_ZJnY[jV_ZJnY
0Zl0=LJ_aZo6[lZ_1jl7[`N_1k@7]0Nd1k@7]0Nd1k@7];Fd]KBe];F@UI2ET9EgUGJEMYEfNgIk07h0
2@L000L7<`005g6;LI1aT76@LI1aT76@RY2:T8^@Ri2;00:@00JYT:V@ZI06Z@08ZZV_ZJnY[jT:[`0C
]:l7[`N_1jl7[kF_]Jne[kFd]@05]00;UGJEMYEfUGJEMW`017Io00T700071c<000R@LI1aT76@R`b@
01JYT:V@ZI2YT:V@ZZV_ZJnY[jV_ZJnY1Zl0:lJ_bZl7[`N_1jl7]0Nd1k@7]0Nd1kBe];Fd]KBe];F@
UGNEMiEfUGIlMW/0PP091`001`Lc000ELI1aT76@LI1aT76@LI1aT8Z@RY2;00J@00JYT:V@ZI02Z@08
[jV_ZJnY[jh8[`0I]:nd[lZ_1jl7[kF_]KBe];Fd]KBe];Fd]@02]008MYEfUGJEMW`4MX<02@L000L7
<P002G6@LI1aT76@R`0>T00DZI2YT:V@ZI2YT:nY[jV_ZJnY[jT8[`0XbZl7[`N_1jl7]0Nd1k@7]0Nd
]KBe];Fd]KBe]9EgUGNEMYEfO7IkMX<02@L000L7<P005Y1aT76@LI1aT76@LI1aT76@RY2:T8X6T00@
ZI2YT:V@ZI2^ZJnY[jV_[PR_016d[kB_]:l7[`N_]Jne];Fd]@03]003]KBd00:d00RK]7JEMYEfO0Ef
Q0091`001`Lb0009LI1aT76@LY2;00j@01BYT:V@ZI2YT:V@[i2_ZJnY[jV_[PJ_02Rd[lZ_1jl7[`Nd
1k@7]0Nd1kBe];Fd]KBe];Fd]I^EMYEfUGIlMW]fQ@091`001`Lb000FT76@LI1aT76@LI1aT76@LI2:
T8Z@RPR@012YT:V@ZI2^T:n@[jV_[Zn^1Zl03;B_]:nd[kB_]:nd[`^d00RK]9^dMYEfO0EfQP091`00
1`Lb000;LI1aT76@LI2;T8/03Y005:V@ZI2YT:V@[i2_T:nY[jV_[Zn^1:l09[B_]:o:[`N_1k@7]0Nd
]KBe];Fd]KBe];Fd]KBeViEfUGIkMW]fQ`091`001`Lb000FT76@LI1aT76@LI1aT76@LI1aT8Z@RPR@
01:YT:V@ZI2^T:j@[i2_[Zn^[jh4[`0:]:nd[kB_]:nd[`fd00JK]9^dMYD4MXT02@L000L7<P003g6@
LI1aT76@LI1aT7:@R`0<T00BZI2YT:V@[Y2_T:n@[jV_[Zn^1:l08kB_]:o:[`N_1k@7]0Nd]KBe];Fd
]KBe];Fd]I^dMYEfUGIk08X02@L000L7<P005Y1aT76@LI1aT76@LI1aT76@LI1aT74<T00>ZI2^T:j@
[Y2_[Zn^[jh4[`0<]:nd[kB_]:nd[kB_2K@019^dVk@2V`04MWafMX/02@L000L7<@02L@0@T76@MY1f
UGJEMYEkUG^@R`j@01JYT:j@[Y2^T:n^[jj_[[B^]:jd[kB_1;@04LZd];Be];Fd]KBe];Fd]KBe00:d
00JK]9]lMW^<00T700071c4000EaLW5fL@04MP0FO7JEMYEkUG^ELI1aT76@LI1aT8Z@RPJ@01jYT:j@
[Y2^UJnE[iF_[[B^]:nd[kB_]:nd[kB_]:l3]00:VkBK]9^dVkBJV`9fS@091`001`La000FRg6@LI1f
UGJEMYEkUG^JO9ZEVYFJL@b@01B^T:j@[Y2^UJnE[jjd[[B^]:nd[`Zd00Ve];Fd]KBe];D01;@019^d
VW^>00T700071c4001MaLW5fLI1fO7JEMYEkUG^ENiEkUGbEO002L@06T76@LI1a29009IF@[Y2^UJnE
[iF_UKB^]:nd[kB_]:nd[kB_]:ndU[BK]9^dVk@00Y/00iFJO02>00T700071c4001b;LI1fT7JEMYEk
UG^JNiZEVYFJUIZEVYF@LI2;1Y0059F@[Y2^T:jE[iF_UKB^]:jd[[B_5K@01Y^dVi^JVhh02@L000L7
<@03L@0@T7JEMYEfUG^ENiEkUG^EO0JE00RJUI1aT76@L@J@02FET9F@[YF^UJnE[iFdUKB_]:nd[kB_
]:ndU[BF]9^dVkBK]9^d00:K00>EVP00SP091`001`Lb000MLI1fT7JENiEkUG^JUIZEVYFJUIZEVYFd
UKBE]7401Y004YF@UI2^T:jE[iF_UKBE]:jd[a6d00VK]9^dVkBJViX0S`091`001`Lb000@MW6@MYEf
UG^ENiEkUG^EO0JE00bJUIZE]9FdUI1aT7H6T00QUI2EUJjE[YF_UJnE]:jd[[B_]:ndU[BF]9JdVkBK
]9^d00:K00BJVgbJS`091`001`Lb0009LI1fT7JENiEl00:E01JJUIZEVYFJUIZE]9FdUKBE]9FdUI1g
190049F@UI2^UJnE[iFdUKBE]:l=]008VkBK]9^dVk@2V`03VY]l08l02@L000L7<`003G6@MY1fUGJE
NiEkUG/01YD039ZEVYFdUKBE]9FdU@9a00B@MY1g0Y0019F@UI02U@0P[YF^UJnE]9FdUKBE]9FdU[BE
]9FdVkBK]9^dVY^EVWb@00T700071c@000EaT7JEM`08U@0FVYFKUKBE]9FdUKBE]9Fd[[B^]76@M`:@
00BET9F@0YD08JjE[iFdUKBE]9FdUKBJ]9ZdVkBK]9^dVkBK]9^dVY]lVP2@00T700071cD000UaT7J@
MYEkUG/019D079ZEVYFJUI^E[iF_UJnE[jj_[Zn^T76@MY1fUI06U@0E[YFdUKBE]9FdUKBE]9ZdV[BK
]9^d00:K00BJVgbJT@091`001`Lf0005MY2@UI001YD069ZE]9FdUKBE]9FdUKB^]:jd[kB_]:n@MP:@
00BET9F@19D07:nE]9FdUKBE]9ZdV[BK]9^dVkBK]9^dVY^JVgbA00T700071cH000R@MY1fUGJEN`JE
01:JUIFE[iF_UJnE[iF_[Zn^[jh2[`08T76@MY1gUI06U@0I/iFdUKBE]9FdUKBJ]9ZdVY^JViZKUIYl
VP2A00T700071cH000=aT7H00Y06U@0PVYFdUKBE]9FdUKBE]:jd[kB_]:nd[lJ_aW6@MiF@UI04U@0G
/iFdUKBE]9FdV[BJ]9ZdV[BJ]9^dVY/00YZB00T700071cL000MaT7J@MYEk00RE00j^UJnE[iF_UJn^
[jj_[PN_00MaT7J@MYF@00BE01RJUIZE]9FdUKBEViZKVY^JViZKVY]lVWbB00T700071cL000J;MY1g
UI04U@0>VYFdUKBE]9FdUKB^]:h4[`0:]:n_[lJ_aZ[6L@:@00>ET9D01ID05kBE]9FdUKBJ]9ZdV[BJ
]9ZdV[BJVgbJ09802@L000L7>002L@04T7J@MPZE01b_UJnE[iF_[Zn^[jj_[Zn^[jV_ZJmaT7J@MYEg
19D04IZEVYFJUI^EViZKVY^JViZK00:J00=lVW/0TP091`001`Li0006T7J@T9F@19D03:jE[iFdUJnE
[jj_[PV_00NZ[j[6Z/Ia00:@00>ET9D00iD04iZE]9FdUKBE]9ZdV[BJ]9ZdVY/00YX00g`0002A00T7
00071cX000B@MY1f0Y06U@0P[YF_UJnE[jj_[Zn^[jj_ZJnY[jV_ZJnYZW6@MY1fUG/4U@06VYFJUIZE
0YX01I^JViZK00BJ00=lVP00TP091`001`Ll0005LI2@UI0019D03ZjE[iF_UKB^[jj_[Zn^1Jl03JZ_
ZZnZaZ[6ZY1fUI0019D019ZEVYD2VP0=ViZdV[BJ]9ZKVY^1VP2D00T700071cd000=aT7400Y000iF@
U@05U@0P[iF_UJn^[jV_ZJnY[jVZZJZYZZVZZJYaT7J@MYEkUG`2U@0@VYFJUIZEVYFJUIYlVWbJNi@0
2@L000L7?P001G6@T9F@00:E022^UJnE[iF_[Zn^[jj_[ZnZ[jZ_ZZnZ[j[6ZL6YZW6@MPJE00>JUIX0
0iX01i^JViZKVY/00YX00g`0002C00T700071cl000=aT7H00Y000iF@U@03U@0a[iF_UJn@[jV_ZJnY
ZZVZZJZYZZVZZJZTZW6@MYEfUG^ENiElVYFJUIZEVYFJO9YlP@2E00T700071d0000=aT9000Y0089F@
[YF_UJnE[jj_[ZnY[jV_ZJnY[jVZZJZYZZVZZJYa0Y04U@04VYFJU@BJ00>KVY/00YX00gbJ002E00T7
00071d4000=aT7H01900<YF@[Y2^T:n@[i2_ZJZYZZVZZJZYZZVZY:Z;ZH^YLI1fUG^ENiEkVWbJO9Yl
VWbJNiYkUP091`001`M20003LI2@00:@02:ET:nE[i2_T:n@[jV_ZJnY[jVZZJZYZZVZZJZYZZVZMYEg
19D019ZEVYD5VP04UIYlVYH02@L000L7@`000g6@L@04T00:[Y2^T:j@[Y2^T0BY02:ZZJZYZX^YRjV;
ZH^YRi1fT7JENiEkUG^JNiYkVW^JNh5kUP091`001`M40003LI2@00:@02:^T:n@[i2_T:n@[jV_ZJZY
ZZVZZJZYZY2ZT:Z@ZG6@MYEg19D02YZEVWbJO9YlVW^G00T700071dD000EaT76@S`04T00<[Y2YT:V@
ZI2YT:V@0jT078^YRjV;ZH^YRi1aT7J@MYEkUG^JNiYkVW^JNgbG00T700071dP000=aT9001900:JV@
[Y2YT:Z@ZY2ZT:Z@ZI2YT:V@ZI2YRi2:T7JEMYEkUGbJO9YkVW^109P02@L000L7B@001g6@LI1aT8X0
19008jV@ZH^YRjV;ZH^YRi2;T8Z@RY1aT76@LI1fUG^ENiEkO7]l09T02@L000L7BP000g6@T006T00T
ZI2YT:V@ZI2YT:V@ZI2YT:V;ZHZ@LI1aT7JENiEkVW^JNh5kV@091`001`M<000WLW6@LI2:T8Z@Ri2;
ZH^@RY2:T8Z@RY2:T8Z@LI1aT76@MYEfUG]l00=kVP091`001`M=0003LW6@00F@00NYT:V@ZI2Y00J@
01>;T8^@LI1aT76@MYEfUG^ENgak09/02@L000L7C`008G6@LI1aT8Z@RY2:T8Z@RY2:T8Z@RY1aT76@
LI1aT7JEMP05Ni/02@L000L7D@00191aT8/3T005Ri2;T8/00i004X^@Ri2;T76@LI1aT7J@MYEkO09k
W0091`001`MB009a01b@LI1aT76@LI2:T8Z@LI1aT76@LI1aT76@MYEf0g^M00T700071e<001f@LI1a
T76@LI2:T8^@Ri2;T8Z@LI1aT76@LI1fU@02Nih02@L000L7E@00776@LI1aT76@LI1aT76@LI1aT76@
LI1aT75gMW^O00T700071eX001EaT76@LI1aT76@LI1aT76@LI1aT7H0X@091`001`MM0008LI1aT76@
LI07LJ@02@L000L7H0002G6@LI1aT76@L@2W00T700071f<00g6Z00T700071ol04@091`001`Oo0140
2@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo0140
2@L000L7o`0A00T700071ol04@091`001`Oo01402@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00001\
\>"],
  ImageRangeCache->{{{0, 287}, {285.812, 0}} -> {-8.15597, -7.50157, 1.10213, 
  1.10213}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell["\<\
X =  2*(-1 + E^(u/(6*Pi)))*Cos[u]*Cos[v/2]^2;
Y =  2*(-1 + E^(u/(6*Pi)))*Cos[v/2]^2*Sin[u];
Z = 1/2 - E^(u/(3*Pi))/2 - Sin[v] + E^(u/(6*Pi))*Sin[v];
spiral=ParametricPlot3D[{X, Y, Z}, {u, 0, 6 Pi},
                {v, 0, 2 Pi},
           PlotPoints -> {50, 20},
           PlotRange -> All, Axes -> False, 
        Boxed->False,DisplayFunction->Identity];\
\>", "Input",
  CellLabel->"In[96]:=",
  AspectRatioFixed->True,
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(MVClear[]; \ MVShow3D[Graphics3D[spiral]]; \)], "Input",
  CellLabel->"In[100]:=",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  "The ",
  StyleBox["MVPolygonShading",
    FontFamily->"Courier"],
  " option to ",
  StyleBox["MVShow3D[]",
    FontFamily->"Courier New"],
  " controls how the surface normals of the polygons are calculated. The \
default setting ",
  StyleBox["MVPolygonShading->MVSmooth ",
    FontFamily->"Courier"],
  "is shown here:"
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[12]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F2F30302F2F3F30303F2F2F3F30303F2F2F3F30303F2F2F3F30303F2F2F4030303F2F2F403030\
3F2F2F4030303F2F2F3F2F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F2F302F2F\
302F2F302F2F3F2F2F302F2F3F2F2F302F2F3F302F302F2F3F2F2F3F2F2F3F302F
3F2F2F3F302F3F2F2F40302F3F2F2F40302F3F2F2F4030303F2F2F40302F402F2F403030402F2F\
40302F402F2F4F3030402F2F40302F3F2F2F3F2F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F2F303030302F2F
3F3030302F2F3F3030302F2F3F3030302F2F3F3030302F2F3F30303F2F2F3F30303F2F2F403030\
3F302F4030303F2F2F4030303F302F4030303F302F40303040302F40303040302F
4F303040302F4F303040302F4F303040302F4F303040302F4F30304F302F4F30304F302F503F30\
4F302F5030304F302F503F304F302F4F303040302F4030303F2F2F3F30303F2F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F2F302F2F\
302F30302F2F302F2F302F2F3F3030302F2F3F2F2F302F2F3F3030302F2F3F302F
3F2F2F3F30303F2F2F3F302F3F2F2F4030303F2F2F40302F3F2F2F4030303F2F2F40302F40302F\
403030402F2F40302F40302F4F303040302F4F302F40302F4F303040302F4F302F
4F302F4F30304F302F4F302F4F302F5030304F302F50302F4F302F503F304F302F50303050302F\
503F3050302F503F3050302F5F3F3050302F50302F4F302F4F3030402F2F40302F
3F2F2F3F3030302F2F3F2F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F30302F2F3F30303F2F2F3F3030\
3F2F2F3F30303F2F2F3F30303F2F2F4030303F2F2F4030303F2F2F40303040302F
403030402F2F40303040302F40303040302F4F303040302F4F303040302F4F30304F302F4F3030\
4F302F4F30304F302F4F30304F302F503F304F302F5030304F302F503F3050302F
503F3050302F503F3050302F503F3050302F5F3F3050302F5F3F3050302F5F3F3050302F5F3F30\
50302F5F3F305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F3050302F503F304F302F
4F303040302F4F303040302F4030303F2F2F3F3030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F302F30302F2F3F2F303F2F2F4030303F2F2F\
402F30402F2F403030402F2F403030402F2F4F3030402F2F4F3030402F2F4F3030
402F2F4F30304F2F2F4F30304F2F2F4F30304F302F5030304F2F2F5030304F302F5030304F302F\
50303050302F50303050302F50303050302F50303050302F50303050302F503F30
50302F503F3050302F5F3F3050302F5F3F3050302F5F3F3050302F5F3F3050302F5F3F3050302F\
5F3F305F3F305F3F305F3F2F5F3F305F3F305F3F305F3F2F5F3F3050302F5F3F30
50302F503F2F4F302F4F303040302F40302F3F2F2F40302F3F2F2F3F2F2F302F2F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000303030302F2F3F30303F2F2F403030402F304F30304F2F2F4F3030\
4F30304F30304F2F2F5030304F30305030304F302F5030304F30305030304F3030
503030503030503030503030503030503030503030503030503F30503030503F305030305F3F30\
5030305F3F305030305F3F305030305F3F305030305F3F30503F305F3F30503F30
5F3F305F3F305F3F305F3F305F3F305F3F305F3F305F3F305F3F3F5F3F305F3F305F3F3060403F\
5F3F306040305F3F3060403F5F3F3060403F5F3F3060403F5F3F305F3F305F3F30
5F3F305F3F305F3F30503F2F5F3F3050302F503F304F302F4F303040302F4030303F2F2F403030\
3F2F2F3F3030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000302F30302F2F3F2F303F2F2F403030402F2F4F30304F2F2F4F30304F2F2F4F30304F2F30\
5030304F2F2F5030304F30305030304F302F5030304F30305030304F3030503030
503030503030503030503030503030503F30503030503F30503030503F30503030503F30503030\
5F3F305030305F3F305030305F3F305030305F3F30503F305F3F30503F305F3F30
5F3F305F3F305F3F305F3F305F3F305F3F3F5F3F305F3F305F3F305F403F5F3F305F40305F3F30\
60403F5F3F3060403F5F3F3060403F5F3F3060403F5F3F306040305F3F305F3F30
5F3F305F3F305F3F2F5F3F30503F2F5F3F3050302F5F3F2F50302F503F304F302F4F302F40302F\
40302F3F2F2F3F302F3F2F2F3F2F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F2F
3F30303F2F2F403030402F2F4F30304F2F2F4F30304F30305030304F3030503030503030503030\
50303050303F503030503030503030503F3F503030503F305030305F3F3F503030
503F3F5030305F3F3F5030305F3F3F5030305F3F3F503F305F3F3F503F305F3F3F5F3F305F3F3F\
503F305F3F3F5F3F305F3F3F5F3F305F3F3F5F3F305F3F3F5F3F305F403F5F3F30
5F403F5F3F3060403F5F3F3060403F5F3F3060403F5F3F3060403F5F3F3060403F60403F60403F\
5F403060403F60403F60403F60403060403F60403F60403F60403060403F5F4030
6040305F3F306040305F3F305F3F305F3F2F5F3F305F3F2F5F3F30503F2F5F3F3050302F503F30\
4F302F4F303040302F4030303F2F2F4030303F2F2F3F3030302F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F302F2F302F2F3F2F30
3F2F2F402F30402F2F4F30304F2F2F4F30304F30305030304F303050303050303050303F4F3030\
503030503030503F3F503030503030503030503F3F503030503F3F5030305F3F3F
5030305F3F3F5030305F3F3F5030305F3F3F503F305F3F3F503F305F3F3F5F3F305F3F3F503F30\
5F3F3F5F3F305F3F3F5F3F305F3F3F5F3F305F403F5F3F305F3F3F5F3F3060403F
5F3F3060403F5F3F3060403F5F3F3060403F5F3F3F60403F5F3F3060403F60403F60403F5F4030\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403060403F
5F3F306040305F3F306040305F3F305F3F305F3F2F5F3F30503F2F5F3F30503F2F5F3F3050302F\
503F2F50302F503F304F302F4F302F40302F40302F3F2F2F40302F3F2F2F3F302F
302F2F302F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F30302F2F3F30303F2F2F403030402F2F
4F30304F2F304F30304F303050303F4F303050303F503030503F3F50303F503F3F50303F5F3F3F\
503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F
5F3F3F5F3F3F5F3F3F5F3F3F5F403F5F3F3F5F403F5F3F3F60403F5F3F3F5F403F5F3F3F60403F\
5F3F3F60403F5F3F3F60403F5F403F60403F5F403F60403F60403F60403F60403F
60404060403F60403F60403F60404060403F60403F60403F6F4F4060403F604F4060403F6F4F40\
60403F6F4F4060403F6F4F4060403F6F4F4060403F6F4F4060403F6F4F3F60403F
6F4F3F60403F60403F60403060403F5F40306040305F3F306040305F3F305F40305F3F305F4030\
5F3F305F3F30503F2F5F3F30503F2F503F304F302F503F304F302F4F303040302F
4030303F302F3F30303F2F2F3F3030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000302F2F3F2F30302F2F3F2F30402F2F403030402F2F4F3030
4F2F304F30304F30305030304F3030503F3F503030503F3F50303F5F3F3F50303F5F3F3F503F3F\
5F3F3F50303F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F
5F3F3F5F403F5F3F3F5F3F3F5F3F3F5F403F5F3F3F5F403F5F3F3F60403F5F3F3F60403F5F3F3F\
60403F5F3F3F60403F5F403F60403F5F403F60403F60403F60404060403F60403F
60403F60404060403F60403F60403F604F4060403F60404060403F6F4F4060403F6F4F4060403F\
6F4F4060403F6F4F4060403F6F4F4060403F6F4F406F4F3F6F4F4060403F6F4F3F
60403F6F4F3F60403060403F60403060403F5F40306040305F3F306040305F3F305F40305F3F30\
5F3F30503F2F5F3F30503F2F5F3F3050302F503F3050302F503F304F302F4F302F
4F302F4F302F402F2F40302F3F2F2F3F302F3F2F2F3F2F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F2F303030302F2F3F30303F2F2F403030402F304F3030402F304F30304F3030
4F30304F3030503F3F50303F503F3F50303F5F3F3F503F3F5F3F40503F3F5F3F405F3F3F5F3F40\
5F3F3F5F40405F3F3F5F40405F3F3F5F40405F3F3F5F40405F3F3F6040405F3F3F
6040405F3F3F6040405F403F6040405F403F6040405F403F6040405F403F60404060403F604040\
60403F60404060403F60404060403F604F4060403F604F4060403F6F4F40604040
6F4F4060403F6F4F406040406F4F4060403F6F4F40604F406F4F40604F3F6F4F406F4F406F4F40\
6F4F406F4F406F4F406F4F406F4F406F4F406F4F406F4F406F4F406F4F406F4F3F
6F4F4060403F6F4F4060403F6F4F3F60403F604F3F60403060403F5F403060403F5F4030604030\
5F3F306040305F3F305F40305F3F305F40305F3F305F3F30503F2F5F3F30503F2F
503F3050302F503F304F302F4F303040302F40303040302F4030303F2F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F30302F2F\
3F2F2F3F2F2F4030303F2F2F402F30402F2F4F3030402F304F30304F30304F303F
4F303050303F4F3030503F3F50303F503F3F503F3F5F3F40503F3F5F3F405F3F3F5F40405F3F3F\
5F3F405F3F3F5F40405F3F3F5F40405F3F3F5F40405F3F3F5F40405F3F3F604040
5F3F3F6040405F403F6040405F3F3F6040405F403F6040405F403F60404060403F60404060403F\
60404060403F604F4060403F60404060403F604F4060403F604F406040406F4F40
60403F6F4F406040406F4F4060403F6F4F40604F406F4F406040406F4F406F4F406F4F406F4F40\
6F4F406F4F406F4F406F4F406F4F406F4F406F50406F4F406F4F406F4F406F4F40
6F4F3F6F4F40604F3F6F4F4060403F604F3F60403F604F3F60403060403F60403060403F5F3F30\
6040305F3F305F40305F3F2F5F40305F3F2F5F3F30503F2F5F3F30503F2F5F3F30
50302F503F3050302F503F3050302F50302F4F302F4F302F40302F40302F402F2F40302F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F2F3F30303F2F2F403030\
402F30403030402F2F4F30304F30304F30304F3030503F3F4F3030503F3F4F303F
503F3F503F3F503F3F503F3F5F40405F3F3F5F40405F3F406040405F40406040405F3F40604040\
5F40406040405F40406040405F40406040405F4040604040604040604040604040
604F40604040604F40604040604F40604040604F406040406F4F4F6040406F4F406040406F4F4F\
604F406F4F406040406F4F4F604F406F4F40604F406F4F4F6F4F406F4F4F6F4F40
6F4F4F6F4F406F4F4F6F4F406F504F6F4F406F504F6F4F4070504F6F4F4070504F6F4F4070504F\
6F4F4070504F6F4F4070504F6F504070504F6F4F4070504F70504070504F6F4F40
7050406F4F406F50406F4F406F50406F4F3F6F4F40604F3F6F4F40604F3F6F4F3F60403F604F3F\
60403060403F5F403060403F5F40306040305F3F306040305F3F305F40305F3F30
5F40305F3F305F3F30503F2F5F3F30503F2F503F3050302F503F304F302F4F30304F302F4F3030\
40302F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F302F2F302F2F3F2F303F2F2F402F30402F2F\
403030402F2F4F30304030304F30304F30304F303F4F3030503F3F4F3030503F3F
503F3F503F3F503F3F5F3F40503F3F5F40405F3F405F40405F40406040405F3F406040405F4040\
6040405F40406040405F40406040405F4040604040604040604040604040604040
604040604F4F604040604F406040406F4F4F604040604F406040406F4F4F6040406F4F40604F40\
6F4F4F6040406F4F4F604F406F4F4F604F406F4F4F6F4F406F4F4F6F4F406F4F4F
6F4F406F4F4F6F4F406F4F4F6F4F4070504F6F4F406F504F6F4F4070504F6F4F4070504F6F4F40\
70504F6F4F4070504F6F504070504F6F4F4070504F70504070504F6F504070504F
6F4F407050406F4F406F50406F4F406F4F406F4F3F6F4F40604F3F6F4F3F60403F604F3F60403F\
604F3F60403060403F6040306040305F40306040305F40306040305F3F305F4030
5F3F305F4030503F2F5F3F30503F2F5F3F30503F2F503F3050302F503F3050302F503F2F4F302F\
4F302F40302F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F30302F2F3F30303F2F2F4030303F2F2F403030402F304F3030\
4030304F303F4F30304F303F4F3030503F3F4F303F503F3F503F3F503F40503F3F
5F3F40503F3F5F40405F3F405F40405F404060404F5F404060404F604040604F4F60404060404F\
604040604F4F604040604F4F6040406F4F4F604040604F4F6040406F4F4F604F40
6F4F4F6040406F4F4F604F4F6F4F4F604F406F4F4F6F4F4F6F4F4F6F4F406F4F4F6F4F4F6F4F4F\
6F4F406F504F6F4F4F6F504F6F4F4F70504F6F4F4F6F504F6F4F4F70504F6F4F4F
70504F6F4F4F70504F6F504F70504F6F4F4F70504F70504F70504F70504F70504F70504F70504F\
70504F70504F70504F70504F70504F7F5F4F70504F705F4F70504F7F5F4F70504F
70504F70504070504F7050407050406F4F407050406F4F406F50406F4F3F6F4F406F4F3F6F4F3F\
604F3F6F4F3F604F3F604F3F604030604F3F60403060403F5F403060403F5F4030
6040305F3F306040305F3F305F40305F3F305F40305F3F305F3F30503F2F5F3F30503F2F503F30\
50302F503F3050302F503F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000302F2F302F2F3F2F30302F2F3F2F303F2F2F402F303F2F2F403030402F304F3030402F30\
4F30304F30304F303F4F30304F3F3F4F303F503F3F4F303F503F3F503F3F5F3F40
503F3F5F40405F3F405F40405F3F406040405F404060404F60404060404F604040604F4F604040\
604F4F604040604F4F604040604F4F60404F6F4F4F6040406F4F4F604F4F6F4F4F
6040406F4F4F604F4F6F4F4F604F406F4F4F6F4F4F6F4F4F6F4F406F4F4F6F4F4F6F4F4F6F4F4F\
6F4F4F6F4F4F6F504F6F4F4F6F504F6F4F4F70504F6F4F4F70504F6F4F4F70504F
6F4F4F70504F6F504F70504F6F4F4F70504F70504F70504F70504F70504F70504F70504F70504F\
70504F70504F70504F70504F70504F70504F7F5F4F70504F705F4F70504F7F5F4F
70504070504F70504070504F6F4F407050406F4F407050406F4F406F50406F4F3F6F4F40604F3F\
6F4F3F604F3F6F4F3F60403F604F3F604030604F3F6040306040305F4030604030
5F40306040305F40305F40305F3F305F40305F3F305F4030503F2F5F3F30503F2F5F3F30503F2F\
503F3050302F503F3050302F503F2F4F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F2F
3F3030302F2F3F30303F2F2F403030402F30403030402F304F30304030304F303F4030304F3F3F\
4F303F503F3F4F303F503F40503F3F503F40503F3F5F4040503F405F40405F3F40
5F404F5F404060404F5F4040604F4F60404F604F4F60404F6F4F50604F4F6F4F4F604F4F6F4F50\
604F4F6F4F4F604F4F6F4F506F4F4F6F4F506F4F4F6F50506F4F4F6F4F506F4F4F
6F50506F4F4F6F50506F4F4F7050506F4F4F7050506F4F4F7050506F4F4F7050506F4F4F705050\
6F504F7050506F504F70505070504F70505070504F70505070504F70505070504F
70505070504F70505070504F705F5070504F705F5070504F7F5F5070504F7F5F5070504F7F5F50\
70504F7F5F5070504F7F5F50705F4F7F5F5070504F7F5F507F5F4F7F5F5070504F
7F5F5070504F7F5F4F70504F7F5F4F70504F70504F70504070504F7050407050406F4F40705040\
6F4F406F50406F4F3F6F50406F4F3F6F4F3F604F3F6F4F3F604F3F604F3F604030
604F3F604030604F3F5F403060403F5F40306040305F40306040305F40305F40305F3F305F4030\
5F3F305F3F30503F2F5F3F30503F2F503F30503F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F30302F2F3F2F30
302F2F3F2F303F2F2F4030303F2F30403030402F304030304030304F303F4030304F303F4F303F\
503F3F4F303F503F3F503F3F503F40503F3F503F40503F405F4040503F405F4040
5F404060404F5F404060404F60404F604F4F60404F6F4F4F604F4F6F4F50604F4F6F4F4F604F4F\
6F4F50604F4F6F4F506F4F4F6F4F506F4F4F6F4F506F4F4F6F50506F4F4F6F5050
6F4F4F6F50506F4F4F6F50506F4F4F7050506F4F4F7050506F4F4F7050506F4F4F7050506F504F\
7050506F4F4F7050506F504F7050506F504F70505070504F70505070504F705050
70504F70505070504F70505070504F705F5070504F70505070504F7F5F5070504F705F5070504F\
7F5F5070504F7F5F5070504F7F5F5070504F7F5F50705F4F7F5F5070504F7F5F4F
70504F7F5F4F70504F705F4F70504F705F4F70504070504F70504070504F6F50407050406F5040\
7050406F4F3F6F50406F4F3F6F5040604F3F6F4F3F604F3F6F4F3F604030604F3F
604030604F3F6040306040305F403060403F5F40306040305F40305F40305F3F305F40305F3F30\
5F4030503F2F5F3F30503F2F5F3F30503F2F503F3050302F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F30302F2F302F30302F2F3F30303F2F2F
4030303F2F2F4030304030304030304030304F303F4030304F3F3F4F303F4F3F3F4F3F3F503F3F\
4F3F3F503F40503F3F504040503F405F40405F40405F40405F404060404F5F4040
604F4F5F404F604F4F604F4F604F50604F4F6F4F506F4F506F4F506F4F4F6F50506F4F506F5050\
6F4F4F6F50506F4F506F50506F4F4F7050506F4F506F50506F4F4F7050506F5050
7050506F4F4F7050506F50507050506F504F7050506F50507050506F504F705050705050705050\
70504F70505070505070505070504F705F50705050705F5070504F705F50705050
705F5070504F7F5F507050507F5F5070504F7F5F507050507F5F5070504F7F5F50705F507F5F50\
705F4F7F5F507F5F507F5F50705F4F7F5F507F5F507F5F507F5F4F7F5F507F5F4F
7F5F50705F4F7F5F50705F4F7F5F4F70504F7F5F4F70504F705F4F70504F705F4F70504070504F\
70504070504F6F50407050406F4F407050406F4F3F6F50406F4F3F6F50406F4F3F
6F4F3F604F3F6F4F3F604F30604F3F604030604F3F604030604F3F5F403060403F5F4030604030\
5F40305F40305F40305F40305F3F305F4030503F305F3F30503F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F2F302F2F302F2F3F2F30302F2F3F2F303F2F2F403030
3F2F2F4030304030304030304030304F303F4030304F3F3F4F30304F3F3F4F303F503F3F4F3F3F\
503F40503F3F504040503F405F4040503F405F404F5F40405F404F5F4040604F4F
5F404F604F4F60404F604F5060404F6F4F50604F4F6F4F50604F4F6F4F506F4F4F6F50506F4F4F\
6F4F506F4F4F6F50506F4F4F6F50506F4F4F6F50506F4F4F6F50506F4F4F705050
6F4F4F6F50506F504F7050506F4F4F7050506F504F7050506F504F7050506F504F7050506F504F\
70505070504F70505070504F70505070504F705F5070504F70505070504F705F50
70504F705F507050507F5F5070504F705F507050507F5F5070504F7F5F507050507F5F5070504F\
7F5F50705F507F5F50705F4F7F5F507F5F507F5F50705F4F7F5F507F5F4F7F5F50
705F4F7F5F50705F4F7F5F5070504F7F5F4F70504F7F5F4F70504F705F4F705040705F4F705040\
70504F70504070504F6F50407050406F50407050406F4F3F6F50406F4F3F6F5040
604F3F6F4F3F604F3F6F4F3F604030604F3F604030604F3F5F40306040305F40306040305F4030\
5F40305F40305F40305F3F305F40305F3F305F4030503F2F5F3F3050302F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F30302F2F3F3030302F2F3F30303F2F2F4030303F3030403030403030
40303F4030304F303F4030304F3F3F4F303F4F3F3F4F3F3F503F404F3F3F503F40503F3F504040\
503F405F40405040405F404F5F40405F404F5F404F604F4F60404F604F4F60404F
6F4F50604F4F6F4F50604F4F6F50506F4F506F50506F4F506F50506F4F506F50506F4F50705050\
6F50507050506F50507050506F50507050506F50507050506F50507050506F5050
705050705050705050705050705F50705050705050705050705F50705050705F50705050705F50\
705050705F507050507F5F507050507F5F507050507F5F50705F507F5F50705050
7F5F50705F507F5F50705F507F5F507F5F507F5F50705F507F5F507F5F507F5F507F5F507F5F50\
7F5F507F5F507F5F507F60507F5F507F60507F5F508060507F5F507F60507F5F50
7F60507F5F507F5F507F5F4F7F5F507F5F4F7F5F50705F4F7F5F50705F4F7F5F4F70504F7F5F4F\
70504F705F4F705040705F4F705040705F4F70504070504F6F50407050406F5040
7050406F4F3F6F50406F4F3F6F50406F4F3F6F4F3F604F3F6F4F3F604F3F604F3F604030604F3F\
604030604F3F5F4030604F3F5F40306040305F40305F40305F40305F4030503F2F
503F30000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F30302F2F3F2F30302F2F3F30303F2F2F3F30303F2F304030303F3030403030
4030304F3F3F4030304F3F3F4F303F4F3F3F4F3F3F4F3F404F3F3F504040503F3F504040503F40\
5F4040503F405F404F5F40405F404F5F404060404F5F404F604F4F60404F604F50
604F4F6F4F50604F4F6F4F50604F506F50506F4F506F50506F4F506F50506F4F506F50506F4F50\
7050506F4F506F50506F50507050506F50507050506F50507050506F5050705050
6F50507050506F50507050507050507050506F5050705050705050705F50705050705F50705050\
705F50705050705F50705050705F50705050705F507050507F5F507050507F5F50
705F507F5F507050507F5F50705F507F5F50705F507F5F50705F507F5F50705F507F5F507F5F50\
7F5F507F5F507F5F507F5F507F60507F5F507F5F507F5F507F60507F5F507F5F50
7F5F507F60507F5F4F7F5F507F5F4F7F5F50705F4F7F5F4F705F4F7F5F4F70504F705F4F70504F\
7F5F4F705040705F4F705040705F4F70504070504F7050407050406F5040705040
6F50407050406F4F3F6F50406F4F3F6F503F604F3F6F4F3F604F3F6F4F3F604F30604F3F604F30\
604F3F5F4030604F3F5F4030604F3F5F40305F40305F40305F40305F3F305F3F30
50302F503F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F2F302F30\
302F2F3F30303F2F303F30303F2F304030303F303040303F4030304F3F3F40303F
4F3F3F40303F4F3F404F3F3F4F3F404F3F3F504040503F40504040503F405F404F5040405F404F\
5040405F404F5F404F60404F5F404F604F5060404F604F5060404F6F4F50604F50
6F4F50604F506F50506F4F506F50506F4F506F505F6F505070505F6F505070505F6F505070505F\
6F505070505F6F505070505F6F505070505F70505070505F6F5050705F5F705050
705F5F705050705F5F705050705F5F705050705F5F705050705F5F7050507F5F5F705F507F5F5F\
7050507F5F5F705F507F5F5F705F507F5F5F705F507F5F5F705F507F5F5F7F5F50
7F5F5F705F507F5F5F7F5F507F5F5F7F5F507F605F7F5F507F605F7F5F507F605F7F5F507F605F\
7F5F5080605F7F5F507F605F7F5F5080605F7F5F5080605F7F5F5080605F7F5F50
7F60507F5F507F60507F5F507F60507F5F507F60507F5F4F7F5F50705F4F7F5F50705F4F7F5F4F\
705F4F7F5F4F705F4F7F5F4F70504F7F5F4F705040705F4F705040705F4F705040
705F406F50407050406F50407050406F503F6F50406F4F3F6F503F6F4F3F6F503F604F3F6F4F3F\
604F3F6F4F3F604F3F604F3F604030604F3F604030604F3F5F4030604F3F5F4030
5F4030503F2F503F30000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000302F2F3F2F30302F2F\
3F2F303F2F2F3F30303F2F304030303F303040303F40303040303F40303F4F3F3F
40303F4F3F3F4F3F3F4F3F404F3F3F503F404F3F40504040503F40504040503F405F404F503F40\
5F404F5F404F5F404F5F404F60404F5F404F604F5060404F604F50604F4F6F4F50
604F4F6F4F50604F506F50506F4F506F50506F4F506F505F6F4F506F505F6F505070505F6F5050\
70505F6F505070505F6F505070505F6F505070505F6F505070505F705050705F5F
6F505070505F705050705F5F705050705F5F705050705F5F705050705F5F705050705F5F705050\
705F5F705F507F5F5F7050507F5F5F705F507F5F5F705F507F5F5F705F507F5F5F
705F507F5F5F705F507F5F5F705F507F5F5F7F5F507F605F7F5F507F5F5F7F5F507F605F7F5F50\
7F605F7F5F507F605F7F5F507F605F7F5F5080605F7F5F5080605F7F5F5080605F
7F5F507F60507F5F507F60507F5F507F5F507F5F4F7F5F50705F4F7F5F50705F4F7F5F50705F4F\
7F5F4F705F4F7F5F4F70504F705F4F705040705F4F705040705F4F705040705F4F
6F50407050406F50407050406F50407050406F503F6F50406F4F3F6F503F6F4F3F6F503F604F3F\
6F4F3F604F306F4F3F604F30604F3F604F30604F3F5F4030604F3F5F4030604030
503F2F503F3050302F503F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F3030302F2F3F30303F2F303F3030\
3F30304030303F303040303F403030403F3F40303F4F3F3F40303F4F3F404F3F3F
4F3F404F3F40503F404F3F40504040503F4050404F503F405F404F5F404F5F404F5F404F604F4F\
5F404F604F505F404F604F50604F4F604F50604F4F6F4F50604F506F4F50604F50
6F505F6F4F506F505F6F4F5070505F6F505070505F6F505F705F5F6F505F70505F6F505F705F5F\
70505F705F5F70505F705F5F70505F705F5F70505F705F5F70505F705F5F70505F
7F5F5F705F5F705F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F\
705F5F7F5F5F705F5F7F605F7F5F5F7F5F5F7F5F5F7F605F7F5F5F7F605F7F5F5F
7F605F7F5F5F7F605F7F5F5F7F605F7F5F5F7F605F7F5F5F80605F7F605F80605F7F5F5F80605F\
7F605F80605F7F605F80605F7F605F80605F7F605F80605F80605F80605F7F6050
80605F7F605080605F7F605080605F7F60507F60507F5F507F60507F5F507F60507F5F4F7F6050\
7F5F4F7F5F50705F4F7F5F50705F4F7F5F4F705F4F7F5F4F705F4F705F4F705040
705F4F705040705F4F705040705F4F705040705F406F5040705F406F50407050406F503F6F5040\
6F503F6F503F6F4F3F6F503F604F3F6F4F3F604F3F6F4F3F604F3F604F3F604F30
604F3F5F40305F4030503F2F503F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F30302F2F3F3030302F303F30303F2F30\
40303F3F303040303F40303040303F40303F4F3F3F40303F4F3F404F303F4F3F40
4F3F3F503F404F3F40503F40503F4050404F503F405F404F503F405F404F5F404F5F404F5F404F\
604F505F404F604F5060404F604F5060404F604F50604F506F4F50604F506F4F50
604F506F505F6F4F506F505F6F505070505F6F505070505F6F505F70505F6F505070505F6F505F\
705F5F6F5050705F5F70505F705F5F705050705F5F70505F705F5F705050705F5F
70505F705F5F705050705F5F705F5F7F5F5F705F507F5F5F705F5F7F5F5F705F507F5F5F705F5F\
7F5F5F705F507F5F5F705F5F7F5F5F705F507F5F5F7F5F5F7F605F7F5F507F605F
7F5F5F7F605F7F5F507F605F7F5F5F7F605F7F5F507F605F7F5F5F80605F7F5F507F605F7F605F\
80605F7F5F5080605F7F605F80605F7F605080605F7F605F80605F7F605080605F
7F605080605F7F5F508060507F5F508060507F5F507F60507F5F507F60507F5F4F7F60507F5F4F\
7F6050705F4F7F5F4F705F4F7F5F4F705F4F7F5F4F705F4F7F5F4F705040705F4F
705040705F4F705040705F4F705040705F406F50407050406F50407050406F503F7050406F503F\
6F50406F4F3F6F503F6F4F3F6F503F604F3F6F4F3F604F3F6F4F3F604F30604F3F
604030604030503F2F503F30503F2F503F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F30303F2F303F30303F2F3040303F3F303040303F\
3F3030403F3F40303F403F3F40303F4F3F404F3F3F4F3F404F3F3F503F404F3F40
503F40503F4050404F503F405F404F503F405F404F5F404F5F404F5F404F604F505F404F604F50\
5F404F604F50604F50604F50604F506F4F50604F506F4F50604F506F505F6F4F50
6F505F6F4F5070505F6F505F70505F6F505F705F5F70505F705F5F70505F705F5F70505F705F5F\
70505F705F5F705F5F705F5F70505F7F5F5F705F5F705F5F705F5F7F5F5F705F5F
7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F605F7F5F5F7F605F\
7F5F5F7F605F7F5F5F7F605F7F5F5F7F605F7F5F5F7F605F7F5F5F7F605F7F605F
7F605F7F5F5F80605F7F605F80605F7F605F80605F7F605F80605F7F605F80605F7F605F80605F\
7F605F80605F80605F80605F80605F806F5F80605F806F5F80605F806F5F80605F
80605F7F605F80605F7F605080605F7F605080605F7F60508060507F5F508060507F5F507F6050\
7F5F4F7F60507F5F4F7F6050705F4F7F6050705F4F7F5F4F705F4F7F5F4F705F4F
7F5F4F705F4F7F5F4F705F40705F4F705040705F4F705040705F406F5040705F40705040705F40\
6F5040705F406F503F6F50406F503F6F50406F503F6F503F604F3F6F503F604F3F
6F4F3F604F30604F3F5F40305F4030503F2F503F30000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F30302F303F30303F2F303F30303F2F3040303F3F3030\
40303F40303F403F3F40303F4F3F3F40303F4F3F404F303F4F3F404F3F40503F40
4F3F40503F40503F405F404F503F405F404F5F404F5F404F5F404F5F404F5F404F604F505F404F\
604F5060404F604F5060404F604F50604F506F4F50604F506F4F50604F506F505F
6F4F506F505F6F505070505F6F505F70505F6F505F705F5F6F505F705F5F70505F705F5F70505F\
705F5F70505F705F5F70505F705F5F705F5F705F5F70505F705F5F705F5F7F5F5F
705F5F705F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F605F705F5F\
7F5F5F7F5F5F7F605F7F5F5F7F605F7F5F5F7F605F7F5F5F7F605F7F5F5F7F605F
7F5F5F7F605F7F605F80605F7F5F5F7F605F7F605F80605F7F605F80605F7F605F80605F7F605F\
80605F7F605F80605F7F605F80605F80605F806F5F7F605F80605F80605F806F5F
7F605F80605F7F605080605F7F605080605F7F605080605F7F5F507F60507F5F507F60507F5F4F\
7F60507F5F4F7F6050705F4F7F5F50705F4F7F5F50705F4F7F5F4F705F4F7F5F4F
705F4F705F4F705F40705F4F705040705F4F705040705F4F6F5040705F406F5040705F406F5040\
7050406F50407050406F503F6F50406F503F6F50406F4F3F6F503F604F3F6F503F
604F3F604F3F5F40305F4030503F2F503F304F302F503F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F30303F2F303F303F3F303040303F40303F40303F40303F4F3F3F\
40303F4F3F4040303F4F3F404F3F404F3F404F3F40503F40503F4050404F503F40
5F404F503F4F5F404F50404F5F404F5F404F5F40505F404F604F505F404F604F505F4050604F50\
604F50604F50604F506F505F604F506F505F604F506F505F6F505F6F505F6F505F
70505F6F505F70505F6F505F705F6070505F705F60705F5F705F60705F5F705F60705F5F7F5F60\
705F5F7F5F60705F5F7F5F60705F5F7F5F60705F5F7F6060705F5F7F5F60705F5F
7F60607F5F5F7F60607F5F5F7F60607F5F5F7F60607F5F5F7F60607F5F5F7F60607F5F5F7F6060\
7F605F7F60607F605F8060607F605F8060607F605F8060607F605F8060607F605F
8060607F605F8060607F605F806F6080605F8060607F605F806F6080605F806F6080605F806F60\
80605F806F6080605F806F6080605F806F6080605F8F6F60806F5F806F6080605F
806F6080605F806F5F80605F806F5F80605F80605F7F6050806F5F7F605080605F7F605080605F\
7F60507F60507F60507F60507F60507F60507F5F4F7F60507F5F4F7F6050705F4F
7F6050705F4F7F5F4F705F4F7F5F4F705F4F705F4F705F40705F4F705F40705F4F705040705F4F\
705040705F406F5040705F406F5040705F406F503F705F406F503F6F50406F503F
6F50406F503F6F503F604F30604F3F5F40305F4030503F2F503F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F2F303F2F303F303F3F2F3040303F3F303F40303F40303F40303F40303F\
4F3F4040303F4F3F404F30404F3F404F3F40503F404F3F40503F4F503F4050404F
503F4F5F404F503F4F5F404F5F404F5F40505F404F5F40505F404F604F505F404F604F50604F50\
604F50604F50604F50604F506F505F604F506F505F6F4F506F505F6F4F506F505F
6F505F70505F6F505F70505F6F505F705F6070505F705F60705F5F705F60705F5F705F60705F5F\
7F5F60705F5F705F60705F5F7F5F60705F5F7F5F60705F5F7F5F60705F5F7F5F60
705F5F7F6060705F5F7F60607F5F5F7F6060705F5F7F60607F5F5F7F60607F5F5F7F60607F5F5F\
7F60607F5F5F7F60607F605F8060607F605F7F60607F605F8060607F605F806060
7F605F8060607F605F8060607F605F8060607F605F80606080605F806F607F605F806F6080605F\
806F6080605F806F6080605F806F6080605F806F6080605F806F6080605F806F60
80605F806F5F80605F806F5F80605F806F5F7F605080605F7F605080605F7F60508060507F6050\
8060507F5F507F60507F60507F60507F5F4F7F60507F5F4F7F6050705F4F7F604F
705F4F7F604F705F4F7F5F4F705F4F7F5F4F705F40705F4F705F40705F4F705040705F4F705040\
705F406F5040705F406F5040705F406F503F7050406F503F7050406F503F6F5040
6F503F6F5040604F3F604F3F5F40305F4F30503F2F503F304F3F2F503F2F40302F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F30303F2F303F303F3F303040303F40303F40303F40303F4F3F4040303F4F3F40\
4F303F4F3F404F3F40503F404F3F40503F4F503F4050404F503F405F404F50404F
5F404F5F404F5F40505F404F5F40505F404F604F505F4050604F505F4050604F50604F50604F50\
604F506F505F604F506F505F604F506F505F6F505F6F505F6F505F70505F6F505F
705F5F6F505F705F6070505F705F60705F5F705F60705F607F5F60705F607F5F60705F607F5F60\
705F607F6060705F607F6060705F607F60607F5F607F6060705F607F60607F5F60
7F60607F5F607F60607F60607F60607F5F607F60607F60607F60607F60608060607F6060806060\
7F60608060607F60608060607F60608060607F60608060607F605F806F60806060
806F607F605F806F60806060806F6080605F806F60806060806F6080605F806F60806F60806F60\
80605F8F6F60806F60806F60806F5F8F6F60806F608F6F60806F608F6F60806F60
8F6F60806F5F8F6F60806F5F806F5F80605F806F5F80605F806F5F7F605F806F5F7F605080605F\
7F605080605F7F60508060507F60508060507F60507F60507F5F4F7F60507F5F4F
7F6050705F4F7F6050705F4F7F604F705F4F7F604F705F4F7F5F4F705F4F7F5F4F705F40705F4F\
705F40705F4F705F40705F40705040705F40705040705F406F5040705F406F5040
705F406F503F705F406F503F6F503F604F30604F3F5F4030504030503F2F503F304F302F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000302F30302F303F303F3F2F3040303F40303F40303F40303F40304040303F4F3F4040303F\
4F3F404F3040503F404F3F40503F4F503F40503F4F503F4050404F503F4F5F404F
503F4F5F404F5F404F5F40505F404F5F40505F404F604F505F404F604F50604F50604F50604F50\
604F50604F506F505F604F506F505F604F506F505F6F505F6F505F6F505F70505F
6F505F705F5F6F505F705F606F505F705F60705F60705F60705F5F705F60705F607F5F60705F5F\
7F5F60705F607F6060705F5F7F5F60705F607F6060705F5F7F60607F5F607F6060
705F5F7F60607F5F607F60607F5F5F7F60607F60607F60607F5F5F7F60607F60607F60607F605F\
7F60607F60608060607F605F8060607F60608060607F605F8060607F6060806F60
7F605F806060806060806F607F605F806F60806060806F6080605F806F60806060806F6080605F\
806F60806F60806F6080605F806F60806F608F6F60806F5F8F6F60806F608F6F60
80605F806F60806F5F806F6080605F806F5F80605F806F5F7F605F806F5F7F6050806F5F7F6050\
80605F7F605080605F7F60507F60507F60507F60507F5F4F7F60507F5F4F7F6050
705F4F7F6050705F4F7F6050705F4F7F5F4F705F4F7F5F4F705F4F705F4F705F40705F4F705F40\
705F4F705F40705F4F6F5040705F406F5040705F406F5040705F406F5040705F40
6F503F6F50406F503F6F503F604F3F604F3F5F40305F4030503F2F503F2F4F3F2F4F3F2F40302F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F30303F2F303F303F3F303F40304040303F403F4040303F4F3F404F30404F3F404F3F40503F4F\
4F3F40503F4F503F4050404F503F4F5F404F503F4F5F40505F404F5F40505F404F
5F40505F404F5F4F505F4050604F505F4F50604F50604F50604F5F604F50604F5F604F506F505F\
604F5F6F505F604F5F6F505F6F505F6F505F6F505F705F606F505F705F606F505F
705F60705F60705F60705F607F5F60705F607F6060705F607F606F705F607F6060705F607F606F\
7F5F607F60607F5F607F606F7F60607F60607F60607F606F7F60607F60607F6060
80606F7F60607F60607F606080606F7F60608060607F6060806F6F7F60608060607F6060806F6F\
7F6060806F607F6060806F6F806060806F60806060806F6F806060806F60806060
806F6F806F60806F60806F608F6F6F806F60806F60806F608F6F6F806F608F6F60806F608F6F6F\
806F608F6F60806F608F706F806F608F6F60806F608F706F8F6F608F7060806F60
8F7060806F608F6F60806F5F8F6F60806F5F806F60806F5F806F60806F5F806F5F80605F806F5F\
80605F806F5F7F6050806F5F7F6050806F5F7F6050806F5F7F60507F60507F6050
7F60507F60507F60507F604F7F60507F604F7F6050705F4F7F6050705F4F7F604F705F4F7F604F\
705F4F705F4F705F40705F4F705F40705F4F705F40705F4F705F40705F406F5040
705F406F5040705F406F503F6F5F406F503F6F503F604F30604F3F5F4030504030503F2F503F30\
4F302F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F2F302F30
302F303F303F3F2F3040303F40303F40304040303F4F3F404030404F3F404F30404F3F404F3F40\
503F4F4F3F40503F4F503F4F50404F503F4F5F404F503F4F5F405050404F5F4050
5F404F5F4F505F404F5F4F505F4050604F505F4050604F50604F50604F5F604F50604F5F604F50\
6F505F604F506F505F6F505F6F505F6F505F6F505F6F505F705F606F505F705F60
6F505F705F60705F5F705F60705F607F5F60705F607F6060705F607F606F705F607F6060705F60\
7F606F705F607F60607F5F607F606F7F5F607F60607F60607F606F7F60607F6060
7F60607F606F7F60607F60607F60608060607F60608060607F60608060607F60608060607F6060\
806F607F6060806F607F6060806F607F6060806F60806060806F60806060806F60
806060806F60806060806F60806F60806F60806F60806F60806F608F6F60806F608F6F60806F60\
8F6F60806F608F6F60806F608F6F60806F608F6F60806F608F7060806F608F6F60
806F608F6F60806F5F8F6F60806F5F8F6F60806F5F806F60806F5F806F5F80605F806F5F80605F\
806F5F7F6050806F5F7F6050806F5F7F60508060507F60508060507F60507F6050
7F60507F60507F604F7F60507F604F7F6050705F4F7F604F705F4F7F604F705F4F7F604F705F4F\
7F604F705F40705F4F705F40705F4F705F40705F4F705F40705F406F5040705F40
6F5040705F406F503F6F5F406F503F6F503F604F3F604F3F5F4F305F4F30503F2F503F304F3F2F\
4F3F2F40302F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F2F3F30303F2F30
3F303F3F303F40304040303F4F3F404030404F3F404F30404F3F404F3F40503F4F4F3F40503F4F\
503F4F50404F503F4F5F404F503F4F5F405050404F5F40505F404F5F4F505F4050
5F4F505F4050604F505F4F50604F505F4F50604F5F604F50604F5F604F506F505F604F5F6F505F\
604F5F6F505F6F505F6F505F6F505F705F606F505F705F606F505F705F60705F60
705F60705F607F5F60705F607F606F705F607F606F7F60607F606F7F5F607F606F7F60607F606F\
7F60607F606F7F60607F606F7F60607F606F7F60607F606F7F606080606F7F6060
80606F7F6060806F6F7F6060806F6F7F6060806F6F7F6060806F6F7F6060806F6F806060806F6F\
806060806F6F806F60806F6F806060806F6F806F60806F6F806F60806F6F806F60
806F6F806F608F6F6F806F608F6F6F806F608F6F6F806F608F6F6F806F608F706F806F608F706F\
806F608F706F8F6F608F706F8F6F608F706F8F6F608F706F8F6F608F706F8F6F60
8F7060806F608F7060806F608F7060806F608F7060806F5F8F6F60806F5F8F6F60806F5F806F5F\
806F5F806F5F806F5F806F5F7F6050806F5F7F6050806F5F7F6050806F5F7F6050
8060507F60508060507F60507F60507F604F7F60507F604F7F6050705F4F7F605070604F7F604F\
705F4F7F604F705F4F7F604F705F4F7F604F705F40705F4F705F40705F4F705F40
705F40705F40705F406F5040705F406F503F6F50406F503F60503F5F4F30604F305F4030504030\
4F3F2F503F304F302F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F2F302F30302F303F303F
3F2F3040303F40303F4F304040303F4F3F404F30404F3F4F4F3040503F4F4F3F40503F4F4F3F40\
503F4F503F4F50404F503F4F50404F50404F5F405050404F5F40505F404F5F4F50
5F404F5F4F505F4050604F505F4050604F50604F50604F5F604F50604F5F604F506F505F604F5F\
6F505F6F505F6F505F6F505F6F505F6F505F705F606F505F705F606F5060705F60
705F60705F60705F607F6060705F607F606F705F607F606F705F607F606F7F60607F606F7F5F60\
7F606F7F60607F606F7F60607F606F7F60607F606F7F60607F606F7F606080606F
7F60607F606F7F6060806F6F7F606080606F7F6060806F6F7F6060806F6F7F6060806F6F7F6060\
806F6F806060806F6F806060806F6F806F60806F6F806060806F6F806F60806F6F
806F60806F6F806F608F6F6F806F60806F6F806F608F6F6F806F608F6F6F806F608F706F806F60\
8F6F6F806F608F706F806F608F706F8F6F608F706F8F6F608F706F8F6F608F706F
806F608F7060806F608F7060806F608F6F60806F5F8F6F60806F5F806F60806F5F806F6080605F\
806F5F806F5F806F5F7F605F806F5F7F6050806F5F7F6050806F5F7F6050806F5F
7F60507F60507F60507F60507F604F7F60507F604F7F6050705F4F7F6050705F4F7F6050705F4F\
7F604F705F4F7F604F705F4F705F4F705F40705F4F705F40705F4F705F40705F4F
6F5F40705F406F5F40705F406F503F6F50406F503F6F503F604F3F604F3F5F40305F4030503F2F\
503F2F4F3F2F4F3F2F40302F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F2F3F30303F2F303F303F3F303F
4030404030404F3F404F30404F3F4F4F3F40503F4F4F3F40503F4F503F4F50404F503F4F504050\
503F4F5F405050404F5F40505F404F5F40505F40505F4F505F40505F4F505F4050
604F5F5F4F50604F5F604F5060505F604F5F60505F604F5F6F505F60505F6F505F60505F6F5060\
6F505F6F50606F505F705F606F5060705F606F5F60705F60705F60705F60705F60
7F606F705F607F606F705F607F606F7F606F7F606F7F606F7F606F7F606F7F606F7F606F806F6F\
7F606F80606F7F606F806F6F7F606F806F6F7F606F806F6F7F606F806F6F7F606F
806F6F80606F806F6F7F606F806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F\
806F6F806F6F806F6F8F6F6F806F6F8F6F6F806F6F8F706F806F6F8F6F6F806F6F
8F706F806F6F8F706F806F6F8F706F8F6F6F8F706F806F6F8F706F8F6F6F8F706F8F6F6F8F706F\
8F706F8F706F8F706F8F706F8F706F8F706F8F706F90706F8F706F8F706F8F7060
8F706F8F70608F706F8F6F608F706F8F6F608F7060806F608F7060806F608F7060806F5F8F7060\
806F5F806F60806F5F806F60806F5F806F5F806F5F806F5F806F5F806F5F7F6050
806F5F7F6050806F5F7F6050806F5F7F60507F6F507F60507F6F507F60507F60507F604F7F6050\
7F604F7F605070604F7F605070604F7F604F705F4F7F604F705F4F7F604F705F40
7F604F705F40705F4F705F40705F406F50406F5F406F503F6F50406F503F60503F5F4F30604F30\
5F40305040304F3F2F503F304F302F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F2F302F30302F303F303F3F2F3F40303F
40303F4F3F404030404F3F404F3040503F4F4F3F40503F4F4F3F4F503F4F503F4F50404F503F4F\
504050503F4F5F405050404F5F405050404F5F40505F40505F4F505F40505F4F50
5F4F50604F5F5F4F50604F5F604F5060505F604F5060505F604F5F6F505F604F5F6F505F6F505F\
6F50606F505F6F50606F505F705F606F505F705F60705F60705F60705F60705F60
705F607F606F705F607F606F705F607F606F7F606F7F606F7F606F7F606F7F606F7F606F7F606F\
7F606F7F606F7F606F7F606F806F6F7F606F806F6F7F606F806F6F7F606F806F6F
7F606F806F6F7F606F806F6F80606F806F6F7F606F806F6F806F6F806F6F806F6F806F6F806F6F\
806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F8F6F6F806F6F8F6F6F
806F6F8F706F806F6F8F706F806F6F8F706F806F6F8F706F8F6F6F8F706F806F608F706F8F6F6F\
8F706F8F6F608F706F8F706F8F706F8F6F608F706F8F706F8F706F8F70608F706F
8F70608F706F8F6F608F706F8F6F608F7060806F608F7060806F608F7060806F5F8F6F60806F5F\
8F6F60806F5F806F5F806F5F806F5F806F5F806F5F806F5F806F5F7F6050806F5F
7F6050806F5F7F6050806F507F6050806F507F60507F60507F60507F60507F604F7F60507F604F\
7F605070604F7F604F70604F7F604F705F4F7F604F705F4F7F604F705F4070604F
705F4070604F705F40705F406F5F40705F406F503F6F50406F503F6F503F604F30604F3F5F4F30\
5F4F30503F2F503F304F3F2F4F3F3040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000302F2F3F30303F2F303F303F3F303F403040403040
4F3F404F3040503F4F4F3F4F503F4F4F3F4F503F4F503F4F50404F503F4F504050503F4F5F4050\
50404F5F40505F40505F40505F40505F4F505F40505F4F505F4050604F5F5F4F50
604F5F604F5060505F604F5F60505F604F5F6F505F60505F6F505F60505F6F50606F505F6F5F60\
6F505F705F606F5060705F606F5F60705F60705F60705F6F705F607F606F705F60
7F606F705F607F606F7F606F7F606F7F606F7F6F707F606F7F6F6F7F606F806F707F606F806F6F\
7F606F806F707F606F806F6F7F606F806F70806F6F806F6F7F606F806F70806F6F
806F6F806F6F806F70806F6F806F6F806F6F806F70806F6F806F6F806F6F8F6F70806F6F806F6F\
806F6F8F7070806F6F8F706F806F6F8F7070806F6F8F706F806F6F8F706F806F6F
8F706F806F6F8F706F8F6F6F8F706F8F6F6F8F706F8F706F8F706F8F706F8F706F8F706F8F706F\
8F706F8F706F8F706F8F706F8F706F90706F8F706F90706F8F706F90706F8F706F
8F706F8F70608F706F8F70608F706F8F70608F706F8F70608F7060806F608F7060806F608F7060\
806F608F7060806F5F8F7060806F5F8F7060806F5F806F5F806F5F806F5F806F5F
806F5F7F6F50806F5F7F6F50806F5F7F6050806F5F7F6050806F507F6050806F507F60507F6050\
7F604F7F60507F604F7F605070604F7F605070604F7F604F705F4F7F604F70604F
7F604F705F4F7F604F705F40705F4F6F5F40705F406F50406F5F406F503F6F503F60503F60503F\
5F4F305F4F305040305040304F3F2F503F304F302F4F302F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F2F302F30302F303F303F3F2F3040303F40303F4F3F40
4030404F3F4F4F3040503F4F4F3F40503F4F4F3F4F503F4F4F3F4F50404F503F4F504050503F4F\
5F405050404F5F40505040505F40505F40505F4F505F40505F4F505F4F50604F5F
5F4F50604F5F604F5F60505F604F5F60505F604F5F6F505F604F5F6F505F6F505F6F50606F505F\
6F5F606F5060705F606F5060705F606F5F60705F60705F60705F6F705F607F606F
705F607F606F70606F7F606F70606F7F606F7F606F7F6F707F606F7F606F7F606F806F707F606F\
806F6F7F606F806F707F606F806F6F7F606F806F6F7F606F806F6F7F606F806F6F
7F606F806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F\
806F6F806F6F8F6F6F806F6F806F6F806F6F8F706F806F6F8F706F806F6F8F706F
806F6F8F706F806F6F8F706F806F6F8F706F8F6F6F8F706F806F6F8F706F8F706F8F706F8F6F6F\
8F706F8F706F8F706F8F706F8F706F8F706F8F706F8F706F8F706F8F706F8F706F
8F70608F706F8F70608F706F8F6F608F706F8F70608F706F806F608F7060806F608F7060806F60\
8F7060806F5F8F7060806F5F806F60806F5F806F60806F5F806F5F806F5F806F5F
7F6F50806F5F7F6F50806F5F7F6050806F5F7F6050806F5F7F60507F6F507F60507F6F507F604F\
7F60507F604F7F605070604F7F605070604F7F6050705F4F7F604F705F4F7F604F
705F4F70604F705F4070604F6F5F40705F406F50406F5F406F503F6F503F60503F6F503F604F30\
604F3F5F40305F4030503F2F503F2F4F3F2F4F3F2F40302F40302F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000302F2F3F30303F2F303F303F3F303F4030404030404F3F404F3040
4F3F4F4F3F4F503F4F4F3F4F504050503F4F504050503F4F5F40505040505F40505040505F4050\
5F40505F4F505F40505F4F5F5F4F50604F5F5F4F50604F5F5F4F5F604F5F604F5F
60505F604F5F60505F604F5F6F506060505F6F50606F505F6F5F606F50606F5F606F5060705F60\
6F5F60705F606F5F60705F6F705F6070606F705F607F606F705F6F7F606F70606F
7F606F7F606F7F606F7F606F806F707F606F806F707F606F806F707F6F6F806F707F606F806F70\
806F6F806F70806F6F806F70806F6F806F70806F6F806F70806F6F806F70806F6F
806F70806F6F806F70806F6F8F7070806F6F8F7070806F6F8F7070806F6F8F7070806F6F8F7070\
806F6F8F7070806F6F8F70708F6F6F8F7070806F6F8F70708F706F8F70708F706F
8F70708F706F8F70708F706F8F70708F706F8F70708F706F9070708F706F8F70708F706F907F70\
8F706F9070708F706F907F708F706F907F708F706F907F708F706F907F6F8F706F
907F6F8F706F8F706F8F706F90706F8F706F8F706F8F70608F706F8F70608F706F8F70608F706F\
8F70608F7060806F608F7060806F608F7060806F5F8F7060806F5F807060806F5F
807060806F5F806F5F806F5F806F5F806F5F806F5F7F6F50806F5F7F6F50806F5F7F6050806F5F\
7F60507F6F507F60507F6F507F60507F6F507F604F7F6F507F604F7F605070604F
7F605070604F7F604F705F4F70604F705F40705F4F6F5F40705F406F50406F5F406F503F6F503F\
60503F60503F5F4F305F4F305040305040304F3F2F503F304F302F4F302F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F2F302F30302F303F303F3F2F3F40303F40303F4F3F404030404F3F4F
4F3F40503F4F4F3F4F503F4F4F3F4F504050503F4F504050503F4F50405050404F5F4050504050\
5F40505040505F4F505F40505F4F5F5F4050604F5F5F4F50604F5F5F4F50604F5F
604F5F60505F604F5F60505F60505F6F506060505F6F50606F505F6F5F606F50606F5F606F5060\
705F606F5F60705F60705F60705F6F705F6070606F705F6F7F606F705F6F7F606F
70606F7F606F7F606F7F606F7F606F7F6F707F606F7F6F707F606F806F707F606F806F707F6F6F\
806F707F606F806F707F6F6F806F707F6F6F806F70806F6F806F70806F6F806F70
806F6F806F70806F6F806F70806F6F807070806F6F806F70806F6F8F7070806F6F8F7070806F6F\
8F7070806F6F8F7070806F6F8F7070806F6F8F7070806F6F8F7070806F6F8F706F
8F706F8F70708F706F8F706F8F706F8F70708F706F8F706F8F706F8F70708F706F8F706F8F706F\
9070708F706F8F706F8F706F907F708F706F907F6F8F706F907F6F8F706F90706F
8F706F8F706F8F706F8F706F8F70608F706F8F70608F706F8F70608F706F8070608F7060807060\
8F7060806F608F7060806F608F7060806F5F8F7060806F5F8F7060806F5F806F5F
806F5F806F5F806F5F806F5F806F5F806F5F7F6F50806F5F7F6F50806F5F7F6050806F507F6050\
806F507F60507F6F507F60507F6F507F604F7F60507F604F7F605070604F7F604F
70604F7F604F70604F70604F705F4070604F6F5F40705F406F50406F5F406F503F6F503F60503F\
6F503F604F30604F3F5F4F305F4F30503F2F5040304F3F2F4F3F3040302F40302F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F2F3F30303F2F303F303F3F303F4030404030404F3F404030404F3F4F4F3F4F
503F4F4F3F4F504050503F4F504050503F4F5040505040505F40505040505F4F505F40505F4F50\
5F40505F4F5F5F4F505F4F5F5F4F50604F5F5F4F5F604F5F604F5F60505F604F5F
605060604F5F6F506060505F6F506060505F6F5F606F50606F5F606F5060705F6F6F5F60705F6F\
6F5F60705F6F705F6070606F705F607F606F705F6F7F606F70606F7F60707F606F
7F60707F606F7F6F707F606F7F6F707F606F806F707F6F70806F707F6F70806F70806F70806F70\
806F6F806F70806F70806F70806F6F806F70806F70806F70806F6F807070806F70
807070806F6F8F7070806F708F7070806F6F8F7070806F6F8F7070806F6F8F7070806F6F8F7070\
806F6F8F70708F706F8F707080706F8F70708F706F8F70708F706F8F70708F706F
8F70708F706F8F70708F706F8F70708F706F907F708F706F9070708F706F907F708F706F907F70\
8F706F907F708F706F907F708F706F907F7090706F907F708F706F907F708F706F
907F6F8F706F907F6F8F706F8F706F8F706F8F7F6F8F706F8F706F8F70608F706F8F70608F706F\
8F70608F706F8F70608F70608070608F70608070608F7060806F608F7060806F5F
8F7060806F5F8F7060806F5F80705F806F5F80705F806F5F806F5F7F6F50806F5F7F6F50806F5F\
7F6F50806F5F7F6F50806F507F6050806F507F60507F6F507F604F7F6F507F604F
7F60507F604F7F605070604F7F604F705F4070604F705F40705F406F5F40705F406F503F6F5040\
60503F6F503F604F3F60503F5F4F305F4F3050402F5040304F3F2F4F3F304F302F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F2F\
302F30302F303F303F3F2F3F40303F40303F4030404030404F3F404F30404F3F4F
4F3F4F503F4F4F3F4F5040504F3F4F504050503F4F5040505040505F40505040505F4F50504050\
5F4F505F40505F4F5F5F40505F4F5F5F4F5F604F5F5F4F5F604F5F604F5F60505F
604F5F60506060505F6F506060505F6F50606F50606F5F606F50606F5F606F5060705F6F6F5F60\
705F6F6F5F60705F6F705F6070606F705F6F7F606F705F6F7F606F70606F7F6070
70606F7F60707F606F7F6F707F606F7F6F707F606F806F707F606F806F707F6F70806F707F6F6F\
806F707F6F6F806F707F6F6F806F70806F6F806F70806F6F806F70806F6F807070
806F6F806F70806F6F807070806F6F807070806F6F8F7070806F6F8F7070806F6F8F7070806F6F\
8F7070806F6F8F7070806F6F8F707080706F8F707080706F8F70708F706F8F7070
8F706F8F70708F706F8F70708F706F8F70708F706F8F70708F706F8F70708F706F907F708F706F\
907F708F706F907F708F706F907F708F706F907F708F706F907F708F706F907F6F
8F706F8F7F6F8F706F8F7F6F8F706F8F706F8F70608F706F8F70608F706F8F70608F706F807060\
8F70608070608F7060806F608F7060806F608F7060806F5F8F7060806F5F8F7060
806F5F807060806F5F807060806F5F806F5F806F5F806F5F7F6F50806F5F7F6F50806F5F7F6F50\
806F5F7F6F50806F507F60507F6F507F60507F6F507F604F7F6F507F604F7F6F50
70604F7F605070604F7F604F705F4F70604F705F40705F4F6F5F406F5F406F503F6F5F4060503F\
6F503F60503F60503F604F30604F305F4F305F4F30503F2F503F2F4F3F2F4F3F2F
40302F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F2F3F3030\
3F2F303F303F3F303F4030404030404F3F404030404F3F4F4F3F4F4F3F4F4F3F4F
504050503F505040505040505F40505040505F40505040505F4F5F5F40505F4F5F5F40505F4F5F\
5F4F5F604F5F5F4F5F60505F604F5F60505F604F5F60506060505F60506060505F
6F50606050606F50606F50606F5F606F50606F5F606F5060705F6F6F5F60705F6F6F5F6070606F\
705F6F70606F705F6F7F606F70606F7F606F70606F7F60707F606F7F60707F606F
7F6F707F60707F6F707F6070806F707F6F70806F707F6F70806F70806F70806F70806F70807070\
806F70806F70806F70807070806F70807070806F708F7070806F708F7070806F70
8F70708070708F7070806F708F70708070708F70708070708F70708F70708F70708F70708F7070\
8F70708F70708F70708F70708F70708F70708F70708F7F708F70708F7F708F7070
907F708F7070907F708F7070907F708F7070907F708F7070907F708F7070907F708F706F907F70\
907F70907F708F7F6F907F70907F70907F70907F6F907F70907F70907F708F7F6F
907F708F7F6F907F708F706F907F708F706F8F7F6F8F706F8F7F6F8F706F8F7F6F8F70608F7F6F\
8F70608F706F8F70608F706F8F70608F70608070608F70608F70608F7060807060
8F70608070608F7060806F5F8F7060806F5F807060806F5F807060806F5F80705F806F5F80705F\
806F5F806F5F7F6F50806F5F7F6F50806F5F7F6F50806F5F7F6F507F6F507F6050
7F6F507F60507F6F507F604F7F6F5070604F70604F705F4070604F705F40705F406F5F406F5F40\
6F503F6F504060503F6F503F604F3F60503F5F4F305F4F3050402F5040304F3F2F
503F30000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F2F302F30302F30\
3F303F3F2F3F40303F40303F4030404030404F3F404030404F3F4F4F3F4F4F3F4F
4F3F4F504050503F4F5040505040505F40505040505F40505040505F4F5F5040505F4F5F5F4050\
5F4F5F5F4F50604F5F5F4F5F60505F5F4F5F60505F604F5F605060604F5F605060
60505F6F50606050606F50606F50606F5F606F50606F5F606F5F60705F6F6F5F60705F6F705F6F\
70606F705F6F70606F705F6F7F606F705F6F7F606F70606F7F607070606F7F6070
7F606F7F6F707F606F7F6F707F6070806F707F6070806F707F6F70806F707F6F70806F70806F70\
806F70806F70806F70806F70807070806F70807070806F70807070806F70807070
806F708F7070806F708F7070806F708F7070806F708F70708070708F707080706F8F70708F7070\
8F707080706F8F70708F70708F70708F706F8F70708F70708F70708F706F8F7070
8F70708F7F708F706F8F7F708F7070907F708F706F907F708F706F907F708F706F907F708F706F\
907F708F706F907F708F7F6F907F708F706F907F70907F6F907F708F706F907F70
8F706F907F708F706F907F6F8F706F8F7F6F8F706F8F706F8F706F8F7F6F8F70608F706F8F7060\
8F706F8070608F706F8070608F706F8070608F70608070608F7060806F608F7060
8070608F7060806F5F807060806F5F807060806F5F80705F806F5F80705F806F5F806F5F806F5F\
806F5F7F6F50806F5F7F6F50806F5F7F6F50806F507F6F50806F507F60507F6F50
7F60507F6F507F604F7F605070604F7F604F705F4F70604F705F40705F4F6F5F406F5F406F503F\
6F5F4060503F6F503F60503F60503F604F30604F305F4F305F4F30503F2F504030
4F3F2F4F3F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F2F3F3030302F303F303F\
3F303F40304040303F4030404030404F3F4F4F3F404F3F4F4F3F4F5040504F3F4F
5040505040505040505040505F40505040505F4F5F5F40505F4F5F5F40505F4F5F5F4F5F5F4F5F\
5F4F5F60505F5F4F5F605060604F5F60506060505F60506060505F6F5060605060
6F50606050606F5F6F6F50606F5F6F6F5F60705F6F6F5F60705F6F6F5F6F70606F705F6F70606F\
705F6F7F607070606F7F607070606F7F607070606F7F60707F606F7F6F707F6070
7F6F707F6070806F707F6F70806F707F6F70806F7F806F70806F7F7F6F7080707F806F70807070\
806F7080707F806F70807070806F708F707F806F708F7070806F708F7070807070
8F70708070708F70708070708F70708070708F70708F70708F70708F70708F70708F70708F7070\
8F70708F7F708F70708F7F708F70708F7F708F70708F7F708F7070907F708F7070
907F708F7070907F708F7070907F708F7070907F708F7F70907F708F7F70907F70907F70907F70\
907F70907F70907F70907F70907F70907F70907F70907F70907F70907F70907F70
907F708F7F6F907F708F7F6F907F708F706F907F708F706F8F7F6F8F706F8F7F6F8F706F8F7F6F\
8F70608F7F6F8F70608F706F8F70608F706F8F70608F70608070608F7060807060
8F70608070608F70608070608F706080705F8F706080705F807060806F5F8F7060806F5F80705F\
806F5F80705F806F5F80705F7F6F5080705F7F6F50806F5F7F6F50806F5F7F6F50
806F507F6F50806F507F60507F6F5070604F7F605070604F70604F705F4070604F6F5F40705F40\
6F503F6F5F406F503F6F503F60503F60503F604F3060503F5F4F305F4F3050402F
5040304F3F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F2F302F30302F303F303F3F2F3F\
3F303F3F303F4030404030404030404030404F3F4F403F404F3F4F4F3F4F504050
4F3F4F5040505040505040505040505F405F5040505F4F5F5040505F4F5F5F4F5F5F4F5F5F4F5F\
5F4F5F5F4F5F6050605F4F5F605060604F5F605060604F5F6050606050606F5060
6050606F50606F50606F5F6F6F50606F5F6F6F5F60705F6F6F5F60705F6F6F5F6F70606F705F6F\
70606F705F6F70606F705F6F70606F70606F7F607070606F7F60707F606F7F6F70
7F606F7F6F707F60707F6F707F6070806F707F6F70806F7F7F6F70806F707F6F70806F7F7F6F70\
806F70806F70807070806F70807070806F70807070806F70807070806F708F7070
806F708F70708070708F70708070708F70708070708F70708070708F70708F70708F7070807070\
8F70708F70708F70708F70708F70708F70708F7F708F70708F7F708F70708F7F70
8F70708F7F708F7070907F708F7070907F708F7070907F708F7070907F708F7F70907F708F7070\
907F70907F70907F708F7F6F907F70907F70907F70907F6F907F70907F6F907F70
8F7F6F907F708F7F6F907F708F706F8F7F6F8F706F8F7F6F8F706F8F7F6F8F706F8F7F6F807060\
8F706F8F70608F706F8070608F706F8070608F706F8070608F70608070608F7060
8070608070608070608F7060806F5F807060806F5F807060806F5F80705F806F5F80705F806F5F\
80705F806F5F80705F7F6F50806F5F7F6F50806F5F7F6F50806F5F7F6F50806F50
7F6F507F6F507F6F507F6F5070604F7F605070604F70604F705F4070604F6F5F40705F406F5040\
6F5F406F503F6F5F4060503F60503F604F3F60503F5F4F30604F305F4F305F4F30
503F2F50402F4F3F2F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F2F3F30303F30303F303F3F303F403040\
40303F4030404030404F3F4F403F404F3F4F4F3F4F5040504F3F4F504050504050
50405F5040505F4F5F5040505F4F5F5F405F5F4F5F5F4F5F5F4F5F5F4F5F604F5F5F4F5F605060\
604F5F605060604F5F6050606050606050606050606F5F606050606F5F6F6F5060
6F5F6F6F5F606F5F6F6F5F60705F6F6F5F6F705F6F6F5F6F70606F705F6F70606F705F6F7F6070\
70606F7F607070606F7F60707060707F60707060707F6F707F60707F6F707F6070
806F7F7F6F70806F7F7F6F70806F7F7F6F70806F7F806F7080707F806F7080707F806F7080707F\
806F7080707F806F708F707F8070708F707F8070708F707F8070708F707F807070
8F707F8F70708F707F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F7070907F7F\
8F70708F7F7F8F7070907F7F8F7070907F7F8F7070907F7F8F7F70907F708F7F70
907F7F8F7F70907F708F7F70907F7F907F70907F70907F70907F7F907F70907F70907F7090807F\
907F70907F70907F7090807F907F70908070907F709F807F907F70908070907F70
908070907F70907F708F7F6F907F708F7F6F907F708F7F6F907F708F7F6F8F7F6F8F706F8F7F6F\
8F706F8F7F6F8F706F8F7F6F8F706F8F7F6F8F70608F7F6F8F70608F706F807060
8F7F6F8070608F70608070608F70608070608F706080705F8F706080705F8F706080705F8F7060\
80705F807060806F5F807060806F5F80705F806F5F80705F806F5F80705F7F6F50
80705F7F6F50806F5F7F6F507F6F507F60507F6F5070604F7F605070604F70604F705F4070604F\
6F5F406F5F406F503F6F5F4060503F6F503F60503F60503F604F3060503F5F4F30
5F4F3050402F5040304F3F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F302F30302F303F303F3F2F3F3F303F3F303F\
4030403F303F4030404030404F3F4F403F404F3F4F4F3F4F5040504F3F4F504050
5040505F405F5040505F4F5F5040505F4F5F50405F5F4F5F5F4F5F5F4F5F5F4F5F604F5F5F4F5F\
6050605F4F5F6050606050606050606050606050606050606F5F606050606F5F6F
6F50606F5F6F6F50606F5F6F6F5F6F705F6F6F5F6F705F6F6F5F6F70606F6F5F6F70606F705F6F\
706070705F6F70607070606F7F607070606F7F60707060707F6F707F60707F6F70
7F60707F6F707F6F707F6F707F6F70806F7F7F6F70806F7F806F7080707F7F6F7080707F806F70\
80707F806F7080707F806F7080707F806F7080707F8070708F707F8070708F707F
8070708F707F8070708F70708F70708F707F8070708F70708F70708F7F7F8F70708F7F708F7070\
8F7F7F8F70708F7F708F7070907F7F8F70708F7F708F7070907F708F7070907F70
8F7F70907F708F7070907F708F7F70907F708F7F70907F70907F70907F708F7F70907F70907F70\
907F70907F70907F70907F70908070907F70907F70907F70908070907F70907F70
907F70907F708F7F6F907F708F7F6F907F708F706F8F7F6F8F706F8F7F6F8F706F8F7F6F8F706F\
8F7F6F8F70608F7F6F8F70608F7F6F8070608F706F8070608F706F8070608F7060
8070608F70608070608F70608070608F706080705F80706080705F807060806F5F807060806F5F\
807060806F5F80705F806F5F80705F7F6F5F80705F806F5F80705F7F6F50806F5F
7F6F50806F5F7F6F507F6F507F6F507F6F5070604F7F605070604F70604F705F4070604F6F5F40\
705F406F50406F5F406F503F6F5F4060503F60503F604F3F60503F5F4F30604F30
5F4F305F4F3050402F504030000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F2F3F3030302F303F303F3F2F3F4030403F303F403040\
403040403F404030404F3F4F403F4F4F3F4F4F3F4F5040504F3F50504050504050
5F405F5040505F4F5F5F405F5F4F5F5F4F5F5F4F605F4F5F604F605F4F5F6050605F4F5F605060\
604F606050606050606050606050606F5F6F6050606F5F6F6050606F5F6F6F5F6F
6F5F6F6F5F6F705F6F6F5F6F705F6F6F5F6F706070705F6F706070705F6F70607070606F706070\
70606F7F60707060707F60707060707F6F707F60707F6F707F60707F6F7F7F6F70
7F6F7F7F6F70806F7F7F6F70806F7F7F6F7080707F806F7F80707F806F7F80707F80707F80707F\
806F708F707F80707F8F707F8070708F707F80707F8F707F8070708F707F8F7070
8F707F8070708F7F7F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F7070907F7F8F7070907F7F\
8F7070907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F907F70
907F7F8F7F70907F7F907F70907F7F907F7090807F907F70907F7F907F7090807F907F7090807F\
907F709F807F907F7090807F907F709F807F907F709F807F907F709F807F907F70
908070907F70908070907F70907F708F7F6F907F708F7F6F907F708F7F6F907F708F7F6F8F7F6F\
8F706F8F7F6F8F706F8F7F6F8F706F8F7F6F8F706F8F7F6F8070608F7F6F8F7060
8F7F6F8070608F7F6F8070608F70608070608F70608070608F706080705F8F706080705F807060\
80705F80706080705F807060806F5F80706080705F80705F806F5F80705F806F5F
80705F7F6F5080705F7F6F50806F507F6F507F6F507F604F7F6F5070604F7F604F70604F70604F\
6F5F40705F4F6F5F406F5F406F503F6F5F4060503F60503F604F3F60503F604F30
604F305F4F305F4F3050402F504030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F2F302F30302F303F303F302F303F303F3F303F4030403F303F\
403040403040403F4F4030404F3F4F4F3F4F4F40504F3F4F50405050405050405F
5040505F4F5F50405F5F4F5F50405F5F4F5F5F4F5F5F4F605F4F5F5F4F605F4F5F6050605F4F5F\
6050606050606050606050606050606050606F5F6F6050606F5F6F6F50606F5F6F
6F50606F5F6F6F5F6F705F6F6F5F6F705F6F6F5F6F70606F6F5F6F70606F705F6F706070705F6F\
70607070606F7F607070606F7F60707060707F6F707060707F6F707F60707F6F7F
7F60707F6F7F7F6F70806F7F7F6F70806F7F7F6F7080707F7F6F7080707F806F7F80707F806F70\
80707F806F7080707F806F7080707F8070708F707F8070708F707F8070708F707F
8070708F707F8070708F707F8070708F707F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F7070\
8F7F7F8F7070907F7F8F7070907F7F8F7F70907F7F8F7070907F7F8F7F70907F7F
8F7F70907F708F7F70907F7F8F7F70907F70907F70907F7F907F70907F70907F7090807F907F70\
907F70907F7090807F907F70908070907F709F807F907F70908070907F70908070
907F70907F70907F70907F708F7F6F907F708F7F6F907F708F7F6F8F7F6F8F7F6F8F7F6F8F706F\
8F7F6F8F706F8F7F6F8070608F7F6F8070608F7F6F8070608F706F8070608F7F6F
8070608F70608070608F70608070608070608070608F706080705F80706080705F807060806F5F\
80706080705F807060806F5F80705F806F5F80705F7F6F5F80705F7F6F5F80705F
7F6F5080705F7F6F5080705F7F6F507F6F507F60507F6F5070604F70604F70604F70604F6F5F40\
705F4F6F5F406F5F406F503F6F5F4060503F6F503F60503F60503F604F3060503F
5F4F305F4F305F4F305F4F30503F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000302F303F303F302F303F303F3F2F3F3F303F3F303F4030403F3040403F40\
4030404F3F4F403F4F4F3F4F4F3F4F4F40504F3F5050405050405050405F504050
5F4F5F5F4F5F5F4F5F5F4F5F5F4F605F4F5F6050605F4F5F605060604F60605060605060605060\
60506060506F6050606F5F6F6050606F5F6F6F50606F5F6F6F5F6F6F5F6F6F5F6F
705F6F6F5F6F70606F6F5F6F706070705F6F706070705F6F7060707060707060707060707F6070\
7060707F6F707060707F6F707F60707F6F707F60707F6F7F7F6F707F6F7F7F6F70
806F7F7F6F7F80707F7F6F7F80707F806F7F80707F806F7F80707F80707F80707F80707F8F707F\
80707F8F707F80707F8F707F80707F8F707F80707F8F7F7F8F707F8F7F7F8F707F
8F7F7F8F707F8F7F7F8F707F8F7F7F8F707F8F7F7F8F707F907F7F8F7F7F907F7F8F7F7F907F7F\
8F7F7F907F7F8F7F70907F7F907F7F907F7F8F7F7090807F907F7F907F7F907F70
90807F907F7F90807F907F7090807F907F7F90807F907F709F807F907F7090807F907F709F807F\
9080709F807F907F709F807F9080709F807F9080709F807F9080709F807F907F70
9F807F907F70908070907F70908070907F709080708F7F709080708F7F6F907F708F7F6F907F70\
8F7F6F8F7F708F7F6F8F7F6F8F7F6F8F7F6F8F706F8F7F6F8F706F8F7F6F8F7060
8F7F6F8F70608F7F6F8070608F7F6F8070608F7F608070608F7F608070608F70608070608F7060\
80706080706080705F8F706080705F80706080705F80706080705F80705F80705F
80705F80705F80705F7F6F5080705F7F6F507F6F507F6F507F6F507F604F7F6F5070604F70604F\
70604F70604F6F5F40705F4F6F5F406F5F4060503F6F5F4060503F60503F604F3F
60503F5F4F30604F305F4F305F4F3050402F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F2F302F30302F303F303F302F303F303F3F2F3F3F30403F303F403040403040\
403F4F4030404F3F4F4F3F4F4F40504F3F4F5040504F405050405F504050504F5F
50405F5F4F5F504F5F5F4F605F4F5F5F4F605F4F5F6050605F4F606050605F4F60605060605060\
60506F60506060506F6050606F5F6F6050606F5F6F6F506F6F5F6F6F5F6F6F5F6F
6F5F6F705F6F6F5F6F705F6F6F5F6F7060706F5F6F706070705F6F70607070606F706070706070\
7F60707060707F60707060707F6F707060707F6F707F60707F6F7F7F6F707F6F7F
7F6F70806F7F7F6F70806F7F7F6F7F80707F7F6F7F80707F806F7F80707F806F7F80707F80707F\
80707F80707F80707F80707F8F707F80707F8F707F80707F8F7F7F80707F8F707F
80707F8F7F7F80707F8F7F7F8F707F8F7F7F8F70708F7F7F8F707F8F7F7F8F70708F7F7F8F7F7F\
907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F
907F7090807F907F70907F7F907F7090807F907F7090807F907F7090807F907F7090807F907F70\
9F807F907F7090807F907F709F807F907F709F807F9080709F807F907F70908070
907F70908070907F70908070907F709080708F7F6F907F708F7F6F907F708F7F6F8F7F708F7F6F\
8F7F708F706F8F7F6F8F7F6F8F7F6F80706F8F7F6F8F706F8F7F6F8070608F7F6F
8070608F7F6F8070608F70608070608F7F608070608070608070608F7060807060807060807060\
80706080705F80706080705F80706080705F80705F80705F80705F7F6F5F80705F
806F5F80705F7F6F5080705F7F6F50806F5F7F6F507F6F507F60507F6F5070604F70604F70604F\
70604F6F5F40705F406F5F406F5F406F503F6F5F4060503F6F503F60503F60503F
604F3060503F5F4F305F4F305F4F305F4F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F302F2F2F303030302F303F303F302F303F303F3F303F3F30403F303F403F40403040403F4F\
403F404F3F4F4F3F4F4F40504F3F4F50405050405050405F5040505F4F5F504F5F
5F4F605F4F5F5F50605F4F606050605F4F606050605F506060506060506060506F60506060506F\
6050606F5F6F60506F6F5F6F60506F6F5F6F6F5F6F6F5F6F6F5F6F6F5F6F6F5F6F
705F6F6F5F6F7060706F5F6F7060706F5F6F7060707060707060707060707F60707060707F6070\
7060707F6F7F7060707F6F7F7F60707F6F7F7F6F707F6F7F7F6F70806F7F7F6F7F
806F7F7F6F7F80707F806F7F80707F806F7F80708080707F80708080707F8F708080707F807080\
80707F8F7F8080707F8F707F80707F8F7F8080707F8F7F7F80707F8F7F7F8F707F
8F7F7F8F707F8F7F7F8F7F7F8F7F7F8F707F907F7F8F7F7F907F7F8F7F7F907F7F8F7F7F907F7F\
8F7F7F907F7F8F7F7F907F7F8F7F7F90807F907F7F90807F907F7F90807F907F7F
90807F907F7F90807F907F7F90807F907F7F9F807F90807F9F807F907F7F9F807F90807F9F807F\
90807F9F807F90807F9F807F9080709F807F9F807F9F807F9080709F807F908070
9F807F907F709F807F907F70908070907F70908070907F709080708F7F709080708F7F6F907F70\
8F7F6F907F708F7F6F8F7F708F7F6F8F7F6F8F7F6F8F7F6F8F706F8F7F6F8F7F6F
8F7F6F8070608F7F6F8070608F7F6F8070608F7F6F8070608F7F608070608F7F608070608F7060\
8070608F7F6080706080706080705F80706080705F80706080705F80706080705F
80705F80705F80705F806F5F80705F7F6F5080705F7F6F507F6F507F60507F6F5070604F7F604F\
70604F70604F705F4070604F6F5F406F5F406F5F406F5F4060503F6F5F4060503F
60503F604F3060503F5F4F305F4F305F4F305F4F30000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F30
2F2F2F302F30302F3030303F302F303F303F302F3F3F30403F2F3F4030403F3040403F40403040\
403F4F403F4F4F3F4F4F3F4F4F40504F40505040504F405050405F50405F5F4F5F
504F5F5F4F605F4F5F5F50605F4F5F6050605F4F606050605F4F6060506060506060506F605060\
60506F6050606F5F6F6050606F5F6F60506F6F5F6F60506F6F5F6F6F5F6F6F5F6F
6F5F6F6F5F6F6F5F6F7060706F5F6F7060706F5F6F706070705F6F706070706070706070706070\
7060707060707F6F707060707F6F707F60707F6F7F7F60707F6F7F7F6F707F6F7F
7F6F707F6F7F7F6F7F80707F7F6F7F80707F806F7F807080806F7F80707F80707F80708080707F\
80707F80707F8F707F80707F8F707F80707F8F7F7F80707F8F7F7F80707F8F7F7F
80707F8F7F7F8F707F8F7F7F8F707F8F7F7F8F707F8F7F7F8F707F8F7F7F8F7F7F907F7F8F7F7F\
907F7F8F7F7F907F7F8F7F7F907F7F8F7F7F90807F8F7F70907F7F907F7F90807F
907F7090807F907F7F90807F907F7090807F907F7F90807F907F7090807F90807F9F807F907F70\
9F807F9080709F807F9080709F807F9080709F807F9080709F807F9080709F807F
907F70908070907F70908070907F70908070907F709080708F7F6F907F708F7F6F907F708F7F6F\
8F7F708F7F6F8F7F708F7F6F8F7F6F8F7F6F8F7F6F80706F8F7F6F80706F8F7F6F
8070608F7F6F8070608F7F6F8070608F7F608070608F7F60807060807060807060807060807060\
80706080706080706080705F80706080705F80706080705F80705F80705F80705F
7F6F5F80705F7F6F5F80705F7F6F507F6F5F7F6F507F6F507F6F507F6F5070604F7F6F5070604F\
70604F70604F70604F6F5F406F5F406F5F406F5F4060503F6F5F4060503F60503F
604F3060503F5F4F30604F3F5F4F305F4F305F402F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F2F
303030302F3030303F302F303F303F3F303F3F30403F303F403F40403040403F4F403F404F3F4F\
403F4F4F40504F3F4F5040504F405050405F5040505F4F5F504F5F5F4F5F5F4F5F
5F50605F4F606050605F4F6060506F60506060506F605060605F6F605060605F6F6050606F5F6F\
60506F6F5F6F60506F6F5F6F6F5F6F6F5F6F6F5F6F6F60706F5F6F7060706F5F6F
7060706F5F6F7060706F5F70706070706070706070706070706F707060707F6F707060707F6F7F\
7060707F6F7F7060707F6F7F7F6F7F7F6F7F7F6F7F80707F7F6F7F80707F7F6F7F
8070807F6F7F807080806F7F80708080707F80708080707F8F7F8080707F8F708080707F8F7F80\
80707F8F7F8080707F8F7F808F707F8F7F808F707F8F7F808F7F7F8F7F808F707F
8F7F808F7F7F8F7F808F7F7F907F808F7F7F907F808F7F7F907F808F7F7F907F808F7F7F908080\
907F7F90807F907F7F908080907F7F90807F907F7F90808090807F90807F907F7F
9F807F90807F9F807F90807F9F807F90807F9F807F90807F9F807F90807F9F807F90807F9F807F\
9F807F9F807F90807F9F8F7F9F807F9F8F7F9F807F9F8F7F9F807F9F807F908070
9F807F9080709F807F9080709F807F908070908070907F70908070907F709080708F7F70908070\
8F7F709080708F7F6F9080708F7F6F8F7F708F7F6F8F7F708F7F6F8F7F6F8F7F6F
8F7F6F8F7F6F8F7F6F80706F8F7F6F8F7F6F8F7F6F8070608F7F6F8070608F7F6F8070608F7F6F\
8070608F7F608070608F7F60807060807F6080705F807F6080705F80706080705F
80706080705F80705F7F705F80705F7F6F5F80705F7F6F5080705F7F6F507F6F507F60507F6F50\
70604F70604F70604F70604F6F5F4070604F6F5F406F5F406F5F406F5F4060503F
6F5F4060503F60503F5F4F3060503F5F4F305F4F305F4F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F2F302F30
2F2F3030303F302F303F303F302F3F3F303F3F2F3F3F30403F3040403F40403040403F4F403F4F\
4F3F4F403F4F4F40504F40505040504F405050405F50405F5F4F5F504F5F5F4F60
5F4F605F50605F4F606050605F4F6060506F5F506060506F60506060506F60506060506F60506F\
6F5F6F60506F6F5F6F605F6F6F5F6F605F6F6F5F6F6F5F6F6F5F706F5F6F6F5F70
6F5F6F7060706F5F6F7060706F60707060706F6070706070706070706070706070706070706070\
7F6F7F7060707F6F7F706F707F6F7F7F6F707F6F7F7F6F7F7F6F7F7F6F7F7F6F7F
7F6F7F80707F7F6F7F80707F80707F80708080707F80708080707F80708080707F80708080707F\
8F7F8080707F8F7F8080707F8F7F8080707F8F7F8080707F8F7F8080707F8F7F80
8F7F7F8F7F808F707F8F7F808F7F7F8F7F808F7F7F8F7F7F8F7F7F907F808F7F7F907F7F8F7F7F\
9080808F7F7F907F7F8F7F7F90807F8F7F7F90807F907F7F90807F907F7F90807F
907F7F90807F907F7F90807F90807F9F807F90807F9F807F90807F9F807F90807F9F807F90807F\
9F807F90807F9F807F90807F9F8F7F90807F9F807F90807F9F807F9080709F807F
9080709F807F90807090807F90807090807F907F70908070907F709080708F7F709080708F7F70\
9080708F7F6F8F7F708F7F6F8F7F708F7F6F8F7F6F8F7F6F8F7F6F807F6F8F7F6F
807F6F8F7F6F8070608F7F6F8070608F7F6F8070608F7F6F8070608F7F6F807060807F60807060\
8F7F60807060807060807060807F6080705F80706080705F80706080705F807060
80705F80705F7F6F5F80705F7F6F5F80705F7F6F507F6F5F7F6F507F6F507F6F507F6F5070604F\
7F6F5070604F70604F70604F70604F6F5F406F5F406F5F406F5F4060503F6F5F40
60503F60503F604F3060503F5F4F30604F3F5F4F305F4F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000303030302F30
30303F302F303F303F302F3F3F30403F303F4030403F3040403F40403040403F4F403F4F4F3F4F\
4F3F4F4F40504F40505040504F4050504F5F50405F5F4F5F504F5F5F4F605F4F60
5F50605F4F6060506F5F506060506F60506060506F60506060506F60506F6F5F6F60506F6F5F6F\
60506F6F5F6F6F5F6F6F5F6F6F5F6F6F5F706F5F6F6F60706F5F6F7060706F5F70
7060706F5F7070607070607070607070607070607F706070706F7F7060707F6F7F7060707F6F7F\
7060707F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F707F7F6F7F8070807F6F7F
8070807F6F7F80708080707F80708080707F807080807080807F808070808F7F808070808F7F80\
80707F8F7F808070808F7F8080707F8F7F808F7F808F7F808F7F7F8F7F808F7F7F
8F7F808F7F7F907F808F7F7F8F7F808F7F7F9080808F7F7F907F808F7F7F9080808F7F7F908080\
8F7F7F908080907F7F908080907F7F90808090807F90808090807F9F808090807F
90808090807F9F808090807F9F808090807F9F808090807F9F807F90807F9F8F809F807F9F8F7F\
9F807F9F8F809F807F9F8F7F9F807F9F8F809F807F9F8F7F9F807F9F8F7F9F807F
9F807F9080709F807F9080709F807F9080709F807F908070908070907F70908070908070908070\
8F7F709080708F7F709080708F7F6F9080708F7F6F8F7F708F7F6F8F7F708F7F6F
8F7F6F8F7F6F8F7F6F8F7F6F8F7F6F807F6F8F7F6F807F6F8F7F6F8070608F7F6F8070608F7F6F\
8070608F7F6F807060807F608070608F7F60807060807F6080705F807F6080705F
80706080705F80706080705F80705F7F6F5F80705F7F6F507F705F7F6F507F6F5F7F6F507F6F50\
70604F7F6F5070604F70604F70604F70604F6F5F406F5F406F5F406F5F406F5F3F
6F5F4060503F60503F60503F60503F5F4F3060503F5F4F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F30302F30
302F30302F3F302F3F3F303F302F3F3F30403F3040403F403F3040403F4F403F404F3F4F403F4F\
4F40504F3F4F4F40504F405050405F504050504F5F50405F5F4F5F5F4F5F5F5060
5F4F605F50605F4F6060506F5F506060506F60506060506F60506060506F60506F605F6F60506F\
6F5F6F60506F6F5F6F605F6F6F5F6F6F5F6F6F5F706F5F6F6F5F706F5F6F706070
6F5F6F7060706F5F707060706F60707060707060707060707060707060707060707F6F7F706070\
7F6F7F7060707F6F7F706F707F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F6F7F807080
7F6F7F8070807F6F7F8070807F707F80708080707F80708080707F8070808070808F7F8080707F\
8F7F8080707F8F7F8080707F8F7F8080707F8F7F8080707F8F7F808F7F7F8F7F80
8F7F7F8F7F808F7F7F8F7F808F7F7F8F7F808F7F7F907F808F7F7F907F808F7F7F9080808F7F7F\
9080808F7F7F9080808F7F7F908080907F7F908080907F7F90807F90807F908080
907F7F90807F90807F9F808090807F9F807F90807F9F807F90807F9F807F90807F9F8F7F90807F\
9F807F9F807F9F8F7F90807F9F8F7F9F807F9F8F7F90807F9F807F90807F9F8F7F
9080709F807F9080709F807F90807090807F90807090807F907F70908070907F709080708F7F70\
9080708F7F709080708F7F6F8F7F708F7F6F8F7F708F7F6F8F7F6F8F7F6F8F7F6F
807F6F8F7F6F807F6F8F7F6F8070608F7F6F8070608F7F6F8070608F7F6F8070608F7F6F807060\
807F60807060807F60807060807060807060807F6080705F80706080705F807060
7F705F80705F7F705F80705F7F6F5F80705F7F6F507F705F7F6F507F6F507F6F507F6F5070604F\
7F6F5070604F706F5070604F70604F6F5F4070604F6F5F406F5F406F5F406F5F40
60503F60503F60503F60503F5F4F3060503F5F4F305F4F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F30303030302F30
3F303F302F3F3F303F3F303F3F30403F3040403F403F3040403F4F403F4F4F3F4F403F4F4F4050\
4F40505040504F4050504F5F50405F504F5F504F5F5F4F605F4F605F50605F4F60
60506F5F506060506F605060605F6F60506F605F6F60506F6F5F6F60506F6F5F6F60506F6F5F70\
6F5F6F6F5F706F5F6F6F60706F5F6F6F60706F5F6F7060706F5F707060706F6070
706070706070706070706070706F7F706070706F7F7060707F6F7F70607F7F6F7F706F7F7F6F7F\
7F6F7F7F6F7F7F6F7F7F707F7F6F7F7F707F7F6F7F8070807F6F7F8070807F707F
807080807080807080807080807F80807080807F808070808F7F808070808F7F808070808F7F80\
807F808F7F80807F808F7F808F7F808F7F808F7F808F7F808F7F808F7F808F7F80
9080808F7F80907F808F7F809080808F7F809080808F7F80908080907F809080808F7F80908080\
90808090808090807F90808090808090808090807F9F80809080809F808090807F
9F8F8090807F9F808090807F9F8F809F807F9F8F8090807F9F8F809F807F9F8F809F807F9F8F80\
9F8F7F9F8F809F807F9F8F809F8F7F9F8F809F8F7F9F8F809F8F7F9F8F7F9F807F
9F8F7F9F807F9F8F7F90807F9F8F7F90807F9F807F9080709F807F90807090807F90807090807F\
9080709080708F7F709080708F7F709080708F7F6F9080708F7F6F8F80708F7F6F
8F80708F7F6F8F7F6F8F7F6F8F7F6F8F7F6F8F7F6F807F6F8F7F6F807F6F8F7F6F807F608F7F6F\
807F608F7F6F8070608F7F6F807F608F7F608070608F7F60807060807F60807060
807F6080705F80706080705F8070607F705F80705F7F6F5F80705F7F6F507F705F7F6F507F6F50\
7F6F507F6F5070604F7F6F5070604F70604F6F604F70604F6F5F406F5F406F5F40
6F5F40605F3F6F5F4060503F60503F60503F60503F5F4F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F30302F302F2F30302F3F
302F3F3F303F302F3F3F30403F3040403F403F3040403F4F403F40403F4F403F4F4F40504F3F4F\
4F40504F405050405F4F4050504F5F50405F5F4F5F504F5F5F4F605F4F605F5060
5F4F6060506F5F506060506F60506060506F605060605F6F60506F605F6F60506F6F5F6F605F6F\
6F5F70605F6F6F5F706F5F6F6F5F706F5F6F6F60706F5F707060706F5F70706070
6F60707060706F607070607070607070607F706070706F7F7060707F6F7F7060707F6F7F706F7F\
7F6F7F706F7F7F6F7F7F6F7F7F6F7F7F6F7F7F707F7F6F7F8070807F6F7F807080
7F707F8070807F707F807080807080807080807080807F808070808F7F808070808F7F80807080\
8F7F808070808F7F80807F808F7F80807F808F7F808F7F808F7F808F7F808F7F80
8F7F808F7F808F7F808F7F808F7F809080808F7F7F9080808F7F809080808F7F7F9080808F7F80\
9080808F7F7F90808090807F908080907F7F90808090807F90808090807F908080
90807F9F808090807F9F808090807F9F8F8090807F9F8F8090807F9F8F8090807F9F8F809F807F\
9F8F809F807F9F8F809F807F9F8F809F807F9F8F7F9F807F9F8F7F9F807F9F8F7F
90807F9F8F7F90807F9F807F9080709F807F90807090807F90807090807F908070908070908070\
9080708F7F709080708F7F709080708F7F6F8F80708F7F6F8F80708F7F6F8F7F70
8F7F6F8F7F70807F6F8F7F6F807F6F8F7F6F807F6F8F7F6F807F6F8F7F6F8070608F7F6F807F60\
8F7F6F807060807F60807060807F60807060807F60807060807F6080705F807F60
80705F8070607F705F80705F7F705F80705F7F6F507F705F7F6F507F705F7F6F507F6F507F6F50\
7F6F5070604F7F6F5070604F706F5070604F70604F6F5F4070604F6F5F406F5F40
605F406F5F4060503F60503F60503F60503F5F4F3060503F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F302F2F3030303F302F3F
3F303F30303F3F30403F3040403F403F3040403F4F403F4F403F4F403F4F4F40504F40504F4050\
4F4050504F5F50405F504F5F50405F5F4F60504F5F5F4F605F4F6060506F5F5060
60506F5F506060506F60506F605F6F60506F605F6F60506F605F6F60506F6F5F70605F6F6F5F70\
605F6F6F5F706F5F6F6F60706F5F706F60706F5F707060706F5F707060706F6070
7060706F6070706F7F706070706F7F7060707F6F7F70607F7F6F7F70607F7F6F7F706F7F7F6F7F\
7F6F7F7F70807F6F7F7F70807F6F7F8070807F6F7F8070807F6F7F8070807F7080
807080807080807F80807080807F808070808F7F8F8070808F7F8F8070808F7F8F807F808F7F8F\
807F808F7F8F8F7F808F7F808F7F808F7F8F8F7F808F7F808F7F808F808F8F7F80
8F80808F7F809080808F7F809080808F7F809080808F7F809080808F7F80908080908080908080\
9080809080809080809080809080809080809080809080809080809F8F80908080
9F8F809080809F8F809080809F8F8090807F9F8F809F80809F8F809F807F9F8F809F8F809F8F80\
9F8F7F9F8F809F8F7F9F8F809F8F7FA08F809F8F7F9F8F809F8F7F9F8F809F8F7F
9F8F7F9F807F9F8F7F9F807F9F8F7F90807F9F8F7F90807F9F807F9080709F8F7F90807090807F\
90807090807F9080709080708F80709080708F80709080708F7F6F9080708F7F6F
8F80708F7F6F8F80708F7F6F8F7F6F807F6F8F7F6F8F7F6F8F7F6F807F6F8F7F6F807F6F8F7F6F\
807F608F7F6F807F608F7F6F8070608F7F6F807F60807F608070608F7F60807060
807F6080705F807F6080705F8070607F705F80705F7F705F80705F7F6F507F705F7F6F507F6F50\
706F507F6F50706F507F6F5070604F706F5070604F70604F6F5F4070604F6F5F40
6F5F40605F406F5F4060503F605F4060503F60503F5F503F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F302F2F30302F3F302F3030303F
302F3F3F303F3F303F3F30403F3040403F403F3040403F4F403F4F403F4F403F4F4F40504F3F50\
4F40504F4050504F5F50405F504F5F504F5F5F4F60504F5F5F4F605F4F605F506F
5F4F6060506F5F506060506F5F506060506F60506F605F6F60506F605F6F60506F6F5F70605F6F\
6F5F70605F6F6F5F70605F6F6F5F706F5F706F60706F5F706F60706F6070706070
6F60707060706F607070607F70607070607F706070706F7F706070706F7F706F7F7F6F7F706F7F\
7F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F70807F6F7F7F70807F707F807080
7F707F807080807080807080807080807080807080807F8F807080807F808070808F7F8F807080\
8F7F80807F808F7F80807F808F7F808F7F808F7F808F7F808F7F808F7F808F7F80
8F7F808F7F808F7F809080808F7F809080808F7F809080808F7F809080808F7F809080808F7F80\
90808090808090808090808090808090808090808090807F9080809080809F8080
90807F9F80809080809F8F8090807F9F8F8090807F9F8F8090807F9F8F809F807F9F8F809F807F\
9F8F809F8F7F9F8F809F807F9F8F809F8F7F9F8F809F807F9F8F7F9F807F9F8F7F
90807F9F8F7F90807F9F8F7F9080709F8F7F9080709F8F7F90807090807F90807090807F908070\
9080709080709080708F7F709080708F7F709080708F7F6F8F80708F7F6F8F8070
8F7F6F8F7F708F7F6F8F7F70807F6F8F7F6F807F6F8F7F6F807F6F8F7F6F807F6F8F7F6F807060\
807F6F807F608F7F6F807060807F60807060807F60807060807F60807060807F60
80705F8070607F705F8070607F705F80705F7F6F5F7F705F7F6F507F705F7F6F507F6F5F706F50\
7F6F50706F507F6F5070604F706F5070604F70604F6F604F70604F6F5F406F604F
6F5F406F5F40605F3F6F5F4060503F60503F60503F60503F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F3030303F302F3F30303F30303F
3F30403F30403F3F403F3040403F4F403F4F403F4F403F4F4F40504F3F504F40504F405050405F\
4F4050504F5F50405F5F4F60504F5F5F4F605F4F605F506F5F4F6060506F5F5060
60506F60506F605F6F60506F605F7060506F605F70605F6F6F5F70605F6F6F5F70605F6F6F6070\
6F5F706F60706F5F706F60706F5F706F60706F607070607F6F607070607F6F6070
706F7F706070706F7F706070706F7F706F7F7F6F7F706F7F7F6F7F706F7F7F6F7F706F7F7F7080\
7F6F7F7F70807F6F7F8070807F70808070807F70808070807F7080807080807080
807F8F807080807F8F8070808F7F8F807F808F7F8F807F808F7F8F807F808F7F8F807F808F7F8F\
8F7F808F7F8F8F7F808F808F8F7F808F808F8F7F8090808F8F7F8090808F8F7F80
90808F8F7F8090808F8F7F8090808F8F808090808F8F808090808F90808090808F90808090808F\
9080809080809080809F8F8F9080809F8F809080809F8F8F9080809F8F80908080
9F8F809080809F8F809080809F8F809F8F809F8F809F80809F8F809F8F809F8F809F8F809F8F80\
9F8F809F8F809F8F80A08F809F8F80A08F809F8F7FA08F809F8F7F9F8F809F8F7F
9F8F809F8F7F9F8F809F8F7F9F8F809F8F7F9F8F7F90807F9F8F7F90807F9F8F7F9080709F8F7F\
90807090807F90807090807F9080709080708F80709080708F80709080708F7F70
9080708F7F708F80708F7F6F8F80708F7F6F8F80708F7F6F8F80708F7F6F8F7F6F807F6F8F806F\
807F6F8F7F6F807F608F7F6F807F608F7F6F807F608F7F6F807F60807F6F807060
807F60807060807F6080705F807F607F705F80705F7F705F80705F7F705F7F705F7F6F507F705F\
7F6F507F6F50706F507F6F50706F507F6F5070604F706F5070604F70604F6F5F40
70604F6F5F406F5F40605F406F5F4060503F605F4060503F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F30302F30302F3030303F302F3F3F303F
30303F3F30403F30403F3F403F3040403F4F403F4F403F4F403F4F4F4050403F504F40504F4050\
50405F4F405F504F5F50405F5F4F60504F5F5F4F605F4F605F506F5F4F605F506F
5F506060506F5F506F605F6F60506F605F6F60506F605F6F605F6F6F5F70605F6F6F5F70605F6F\
6F5F70605F6F6F60706F5F706F60706F5F706F60706F607070607F6F607070607F
6F6070706F7F6F6070706F7F70607F706F7F70607F706F7F706F7F7F6F7F706F7F7F6F7F706F7F\
7F70807F6F7F7F70807F6F7F7F70807F6F7F7F70807F70808070807F7080807080
807080807F80807080807F80807080807F8F8070808F7F8F807F808F7F8F807F808F7F8F807F80\
8F7F8F807F808F7F8F8F7F808F7F8F8F7F808F7F8F8F7F808F808F8F7F808F808F
8F7F8090808F8F7F809080808F7F8090808F8F7F809080808F808090808F8F7F80908080908080\
9080808F8080908080908080908080908080908080908080908F80908080908080
9080809F8F809080809F8F809080809F8F809080809F8F809080809F8F8090807F9F8F809F8F80\
9F8F809F8F7F9F8F809F8F7F9F8F809F8F7F9F8F809F8F7F9F8F809F8F7F9F8F80
9F8F7F9F8F8090807F9F8F7F90807F9F8F7F90807F9F8F7F90807F9F8F7F90807090807F908070\
90807F9080709080709080709080708F80709080708F80709080708F7F6F8F8070
8F7F6F8F80708F7F6F8F80708F7F6F8F8070807F6F8F7F6F807F6F8F7F6F807F6F8F7F6F807F6F\
8F7F6F807F608F7F6F807F608F7F6F807F60807F6F807F60807F6F807060807F60
807060807F607F705F8070607F705F8070607F705F7F705F7F6F5F7F705F7F6F507F6F5F7F6F50\
7F6F50706F507F6F50706F507F6F5070604F706F5070604F70604F6F604070604F
6F5F406F604F605F406F5F40605F3F6F5F4060503F60503F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030303F302F3030303F302F3F3F30403F303F
3F30403F3040403F4F3F3F40403F4F403F4F4F4050403F4F4F40504F3F504F405F4F405050405F\
4F405F504F60504F5F5F4F60504F605F50605F4F605F506F5F4F6060506F5F506F
60506F5F506F605F7060506F605F7060506F6F5F70605F6F6F5F70605F6F6F60706F5F706F6070\
6F5F706F60706F5F706F60706F5F7070607F6F607070607F6F6070706F7F706070
706F7F706070706F7F70607F706F7F70607F7F6F7F706F7F7F6F7F706F7F7F70807F6F7F7F7080\
7F6F7F7F70807F6F807F70807F70808070807F70808070807F7080807F8F807080
807F8F807080807F8F807080807F8F807F808F7F8F807F8F8F7F8F807F8F8F7F8F8F7F8F8F7F8F\
8F7F808F808F8F7F8F8F808F8F7F808F808F8F7F808F808F8F7F8090808F8F7F80
8F808F8F7F8090808F8F808090808F8F808090808F8F808090808F8F808090808F90808090808F\
90808090808F90808090808F908080908F8F908080908F809080809F8F8F908080
9F8F809080809F8F809080809F8F809080809F8F80908F809F8F80908F809F8F809F8F809F8F80\
9F8F809F8F809F8F809F8F809F8F809F8F809F8F809F8F809F8F7F9F8F809F8F7F
9F8F809F8F7F9F8F809F8F7F9F8F7F908F7F9F8F80908F7F9F8F7F90807F9F8F7F90807F9F8F7F\
908070908F7F908070908F7F908070908F7F9080709080708F80709080708F8070
9080708F80709080708F80708F80708F7F6F8F80708F7F6F8F8070807F6F8F80708F7F6F8F806F\
807F6F8F806F807F6F8F7F6F807F608F7F6F807F608F7F6F807F608F7F6F807F60
807F60807060807F60807060807F607F705F8070607F705F80705F7F6F5F7F705F7F6F507F705F\
7F6F507F705F706F507F6F50706F507F6F5070604F706F5070604F706F4F6F604F
70604F6F5F406F604F6F5F406F5F40605F3F6F5F4060503F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F302F2F3030303F302F3F30303F30303F3F3040
3030403F30403F3040403F4F3F3040403F4F403F4F4F4050403F4F4F40504F40504F405F4F4050\
50405F50405F504F6050405F5F4F60504F605F5060504F605F506F5F506060506F
5F506060506F60506F605F6F60506F605F6F605F6F605F70605F6F6F5F70605F6F6F5F70605F6F\
6F5F706F5F706F60706F5F706F60706F607070607F6F607070607F6F607070607F
6F607070607F70607F706F7F70607F706F7F706F7F7F6F7F706F7F7F6F7F706F7F7F6F80706F7F\
7F70807F6F7F7F70807F6F7F7F70807F70808070807F70808070807F7080807080
7F7080807F80807080807F8F807080807F8F807F808F7F8F807F808F7F8F807F808F7F8F807F80\
8F7F8F807F808F7F8F807F808F7F8F8F7F808F808F807F808F7F8F8F7F808F808F
8F7F808F808F8F7F808F808F8F7F808F80808F7F8090808F8F7F809080808F808090808F8F8080\
9080808F80809080808F8080908080908080908080908080908080908080908F80
908080908F80908080908F80908080908F809080809F8F809080809F8F809080809F8F8090807F\
9F8F80908F809F8F80908F7F9F8F809F8F7F9F8F80908F7F9F8F809F8F7F9F8F80
90807F9F8F7F90807F9F8F7F90807F9F8F7F90807F9F8F7F90807F908F7F90807F9F8F7F908070\
908F7F908070908F7F9080709080709080709080708F80709080708F8070908070
8F7F6F8F80708F7F6F8F80708F7F6F8F80708F7F6F8F8070807F6F8F7F6F807F6F8F7F6F807F6F\
8F7F6F807F6F8F7F6F807F60807F6F807F60807F6F807F60807F6F807F60807F60
807060807F607F7060807F607F705F8070607F705F80705F7F6F5F7F705F7F6F507F705F706F50\
7F6F50706F507F6F50706050706F5070604F706F5070604F70604F70604F70604F
6F5F406F604F6F5F406F5F40605F406F5F4060503F605F40000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F3030303F302F3F3F304030303F3F30403F3040
403F4F3F3F40403F4F403F4F404050403F4F4F4050403F504F405F4F405050405F4F405F504F60\
50405F504F60504F5F5F4F60504F605F50605F4F605F506F5F506F60506F5F506F
605F7060506F605F70605F6F6F5F70605F706F5F70605F706F6070605F706F60706F5F706F607F\
6F60706F607F6F607070607F6F607070607F6F6070706F7F70607F706F7F70607F
706F7F706F7F706F7F706F7F7F6F80706F7F7F6F80706F7F7F70807F6F7F7F70807F6F7F7F7080\
7F70807F70807F70808070807F70808070807F7080807F8F807080807F8F807080
807F8F807F8F807F8F807F808F7F8F807F8F8F7F8F807F8F8F808F807F8F8F7F8F807F8F8F808F\
8F7F8F8F808F807F8F8F808F8F7F8F8F808F8F7F8F8F808F8F7F8F8F808F8F7F80
8F808F8F808F8F808F8F808090808F8F808090808F8F808090808F8F808090808F8F808090808F\
90808090808F908080908F8F908080908F8F908080908F8F908080908F8F908080
9F8F8F9080809F8F809080809F8F8F908F809F8F80908F809F8F8F908F809F8F80908F809F8F80\
9F8F809F8F809F8F809F90809F8F809F8F809F8F809F8F809F8F809F8F809F8F7F
9F8F809F8F7F9F8F80908F7F9F8F809F8F7F9F8F80908F7F9F8F80908F7F9F8F7F90807F9F8F7F\
90807F9F8F7F90807F9F8F7F90807F908F7F908070908F7F90807090807F8F8070
90807F8F80709080708F80709080708F80708F80708F7F6F8F80708F806F8F80708F7F6F8F8070\
8F7F6F8F806F807F6F8F806F807F6F8F806F807F6F8F806F807F6F8F7F6F807F60
807F6F807F60807F60807060807F608070608070607F705F8070607F705F7F705F7F6F5F7F705F\
7F6F507F705F706F507F6F5F706F507F6F50706F507F6F5070604F706F5070604F
706F4F6F604F70604F6F5F406F604F6F5F406F5F40605F3F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F3030303F2F2F3F30303F30303F3F304030303F3F3040
3F3040403F4F3F3040403F4F403F4F404050403F4F4F40504F3F504F405F4F405050405F4F405F\
504F6050405F504F60504F605F4F60504F605F50605F4F605F506F5F506060506F
5F506F605F6F60506F605F70605F6F605F70605F6F605F70605F706F6070605F706F60706F5F70\
6F60706F5F706F60706F60706F607F6F607070607F6F6070706F7F6F6070706F7F
70607F706F7F70607F706F7F706F7F7F6F80706F7F7F6F80706F7F7F7080706F7F7F70807F6F7F\
7F70807F6F807F70807F70808070807F70808070807F7080807F8F7F7080807F8F
807080807F8F807080807F8F807F80807F8F807F80807F8F807F8F8F7F8F807F8F8F7F8F807F8F\
8F7F8F807F808F7F8F807F808F808F807F808F808F807F808F808F807F808F808F
8F7F808F808F8F7F808F808F8F7F808F808F8F7F808F808F8F808090808F8F808090808F8F8080\
90808F8F80809080808F808090808F8F8080908080908080908F8F8F8080908F80
908080908F80908080908F80908080908F80908080908F809080809F8F809080809F8F80908F80\
9F8F80908F809F8F80908F809F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80
908F7F9F8F8090807F9F8F7F908F7F9F8F7F90807F908F7F90807F9F8F7F90807F908F7F90807F\
908F7F908070908F7F908070908F7F90807090807F908070908F7F8F8070908070
8F80709080708F80708F80708F80708F80708F7F6F8F80708F7F6F8F8070807F6F8F806F807F6F\
8F806F807F6F8F7F6F807F6F8F806F807F6F807F6F807F608F7F6F807F60807F6F
807F60807F607F7060807F607F705F807F607F705F7F705F7F705F7F705F7F6F5F7F705F7F6F50\
7F705F706F507F6F50706F507F6F5070604F706F5070604F706F506F604F70604F
6F604F70604F6F5F406F604F6F5F406F5F40605F406F5F40000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000302F3F302F3F30304030303F3F30403030403F3F403F3040
403F4F3F304F403F4F403F4F403F50403F504F40504F40504F405F4F4050504F5F50405F504F60\
50405F5F4F60504F605F4F60504F605F506F5F50605F506F5F506F605F6F5F506F
605F6F60506F605F70605F70605F70605F706F6070605F706F6070605F706F607F6F5F706F607F\
6F5F706F607F6F607070607F6F6070706F7F6F607F706F7F6F607F706F7F70607F
706F7F706F7F7F6F80706F7F7F6F80706F7F7F7080706F7F7F7080706F7F7F70807F6F807F7080\
7F70808070807F70808070807F7080807F8F7F7080807F8F7F7080807F8F807080
807F8F807080807F8F807F8F807F8F807F8F8F7F8F807F8F8F7F8F807F8F8F808F807F8F8F808F\
807F8F8F808F807F8F8F808F807F8F8F808F8F7F8F8F808F807F8F8F808F8F7F8F
8F808F8F7F808F808F8F808F8F808F8F808090808F8F80808F808F8F808090808F8F808090808F\
8F8080908F8F8F808090808F8F8080908F8F908080908F8F908080908F8F908080
908F8F908080908F8F908F80908F809080809F8F8F908F809F8F80908F809F8F8F908F809F8F80\
908F809F8F80908F809F8F80908F809F8F80908F809F8F80908F809F8F80908F80
9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F7F90807F9F8F7F908F7F908F7F\
90807F908F7F90807F908F7F908070908F7F908070908F7F908070908F7F908070
908F7F8F8070908F7F8F80709080708F80709080708F80708F80708F806F8F80708F806F8F8070\
807F6F8F8070807F6F8F806F807F6F8F806F807F6F8F806F807F6F8F806F807F60
807F6F807F60807F6F807F60807F607F705F807F607F705F8070607F705F80705F7F705F7F705F\
7F6F507F705F7F6F507F6F50706F507F6F50706F50706F5070604F706F5070604F
706F4F6F604F70604F6F604F6F604F6F5F406F6040605F406F5F40000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F30302F3F302F3F303040302F3F3F30403F30403F3F4F
3F3040403F4F3F3F4F403F50403F4F403F50403F504F405F4F3F504F405F4F405F504F5F4F405F\
504F60504F5F504F60504F605F4F60504F605F506F5F4F605F506F5F506F60506F
5F506F605F6F60506F605F70605F6F605F70605F706F5F70605F706F6070605F706F6070605F70\
6F60706F60706F607F6F60706F607F6F607070607F6F6070706F7F6F607F706F7F
6F607F706F7F706F7F706F80706F7F706F80706F7F7F6F80706F7F7F7080706F807F70807F6F80\
7F70807F70807F70807F70807F70807F708080708F7F708080708F7F7080807F8F
7F7080807F8F807080807F8F807080807F8F807F8F807F8F807F8F807F8F807F8F8F7F8F807F8F\
807F8F807F8F8F808F807F808F7F8F807F808F808F807F808F808F807F808F808F
807F808F808F8F7F808F808F807F808F808F8F80808F808F8F7F808F808F8F80808F808F8F8080\
8F808F8F808090808F8F80809080808F808090808F8F80809080808F8080908F80
8F8080908F80908080908F80908080908F80908080908F80908080908F80908F80908F80908080\
908F80908F809F8F80908F809F8F80908F809F8F80908F7F9F8F80908F7F9F8F80
90807F908F80908F7F9F8F8090807F908F7F90807F908F7F90807F908F7F90807F908F7F908070\
908F7F908070908F7F8F807090807F8F8070908F7F8F80709080708F8070908070
8F80709080708F80709080708F80708F80708F80708F80708F7F6F8F80708F7F6F8F8070807F6F\
8F806F807F6F8F806F807F6F8F806F807F6F8F806F807F60807F6F807F60807F6F
807F60807F60807060807F607F705F8070607F705F8070607F705F7F705F7F705F7F705F706F50\
7F705F706F507F705F706F50706F50706F50706F5070604F706F5070604F706F50
6F604F70604F6F604F70604F6F5F406F60406F5F406F5F40605F3F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F3030303F30303F3030403030403F3F403F30403F3F4F3F304F
403F4F403F4F403F50403F4F4F40504F40504F405F4F4050504F5F4F405F504F6050405F504F60\
504F605F4F60504F605F506F5F4F605F506F5F506060506F5F506F605F6F5F506F
605F70605F70605F70605F706F6070605F706F6070605F706F607F6F60706F607F6F60706F607F\
6F607F6F607F6F607F706F7F6F607F706F7F6F607F706F7F706F7F706F80706F7F
706F80706F7F7F6F80706F7F7F7080706F807F7080706F807F70807F70807F70807F708080708F\
7F708080708F7F7080807F8F7F7080807F8F7F7080807F8F80708F807F8F80708F
807F8F807F8F807F8F807F8F8F7F8F807F8F8F7F8F807F8F8F8090807F8F8F808F807F8F8F8090\
807F8F8F808F807F8F8F808F807F8F8F808F807F8F8F808F8F808F8F808F807F8F
8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F80808F808F8F808F8F808F8F8080908F8F\
8F808F908F8F8F8080908F8F8F8080908F8F8F8080908F8F908080908F8F908080
908F8F908F80908F8F908080908F8F908F80908F8F908F809F8F8F908F809F8F8F908F809F8F8F\
908F809F8F80908F809F908F908F809F8F80908F809F8F80908F809F8F80908F80
9F8F80908F809F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80908F7F908F7F90807F908F7F\
908F7F908F7F90807F908F7F90807F908F7F908070908F7F908070908F7F8F8070
908F7F8F8070908F7F8F8070908F7F8F80709080708F80709080708F80708F80708F80708F8070\
8F80708F80708F806F8F80708F806F8F8070807F6F8F8070807F6F8F806F807F6F
8F806F807F60807F6F807F60807F60807060807F607F705F807F607F705F7F70607F705F7F705F\
7F705F7F705F7F6F507F705F706F507F6F50706F507F6F50706F50706F5070604F
706F5070604F70604F6F604F70604F6F604F6F604F6F5F406F6040000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000302F3F302F3F303040302F3F3F30403030403F3F4F3F30403F3F4F
3F3F4F403F50403F4F403F50403F504F405F403F504F405F4F405F50405F4F405F504F6050405F\
504F60504F60504F60504F605F506F504F605F506F5F506F5F506F5F506F605F6F
5F506F605F70605F6F605F70605F706F6070605F706F6070605F706F607F605F706F607F6F6070\
6F607F6F60706F607F6F607F706F7F6F607F706F7F6F607F706F7F6F607F706F7F
706F7F706F80706F7F706F80706F7F7F7080706F7F7F7080706F807F70807F6F807F70807F7080\
7F70807F70807F70807F708080708F7F708080708F7F7080807F8F7F7080807F8F
7F708F807F8F7F7080807F8F807F8F807F8F807F8F807F8F807F8F807F8F807F8F807F8F807F8F\
8F808F807F8F80808F807F8F8F808F807F8F8F808F807F8F8F808F807F808F808F
807F8F8F808F807F808F808F8080808F808F807F808F808F8F80808F808F8F80808F808F8F8080\
8F808F8F80808F808F8F8080908F8F8F808090808F8F8080908F8F8F8080908F8F
8F8080908F8F8F8080908F80908080908F8F8F8080908F80908F80908F80908F80908F80908F80\
908F80908F80908F80908F809F8F80908F80908F80908F80908F80908F7F908F80
908F7F908F8090807F908F80908F7F908F8090807F908F7F90807F908F7F90807F908F7F90807F\
908F7F8F807F908F7F8F807F908F7F8F8070908F7F8F8070908F7F8F807090807F
8F807090807F8F80708F80708F80708F80708F80708F80708F80708F80708F806F8F80708F806F\
8F8070807F6F8F8070807F6F8F8070807F6F8F806F807F6F8F806F807F60807F6F
807F60807F6F7F7F60807F607F7060807F607F705F8070607F705F7F70607F705F7F705F7F705F\
7F705F706F507F705F706F507F6F5F706F50706F50706F50706F5070604F706F50
6F604F706F4F6F604F70604F6F604F6F604F6F5F406F6040605F40000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000302F3F30304030303F3F30403030403F304F3F3040403F4F3F3F4F
403F50403F4F4F4050403F504F405F4F3F504F405F4F405F50405F4F405F504F60504F5F504F60\
504F605F5060504F605F506F504F605F506F5F506F5F506F5F506F605F7060506F
605F70605F70605F70605F706F6070605F706F607F6F5F706F607F6F60706F607F6F607F6F607F\
6F607F706F7F6F607F706F7F6F607F706F8070607F706F80706F7F706F80706F7F
706F80706F7F7F7080706F807F7080706F807F70807F70807F70807F70807F708F7F70807F708F\
7F708080708F7F708080708F7F7080807F8F7F708F807F8F7F7080807F8F807F8F
807F8F807F8F807F8F807F8F807F8F807F8F808090807F8F808090807F8F8F8090807F8F8F808F\
807F8F8F8090807F8F8F808F807F8F8F808F807F8F8F808F807F8F8F808F80808F
8F808F80808F8F808F8F808F8F808F80808F8F808F8F808F8F808F8F80808F808F8F808F8F808F\
8F8080908F8F8F808F8F8F8F8F8080908F8F8F8080908F8F8F8080908F8F8F8080
908F8F8F8080908F8F908F80908F8F908F80908F8F908F80908F8F908F80908F8F908F80908F8F\
908F809F8F8F908F80908F80908F809F8F80908F80908F80908F809F8F80908F80
908F80908F7F9F8F80908F7F908F80908F7F908F80908F7F908F8090807F908F80908F7F908F7F\
90807F908F7F90807F908F7F8F807F908F7F8F807F908F7F8F8070908F7F8F8070
908F7F8F8070908F7F8F80708F80708F8070908F708F80708F80708F80708F80708F80708F8070\
8F806F8F80708F806F8F807080806F8F807080806F8F8070807F6F8F806F807F6F
807F6F807F60807F6F807F60807F607F7060807F607F7060807F607F705F807F607F705F7F705F\
7F705F7F705F7F6F5F7F705F706F507F705F706F50706F50706F50706F50706F4F
706F506F604F706F506F604F70604F6F604F70604F6F60406F604F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F3030303F302F3F3030403030403F30403030403F304F3F304F403F4F
3F304F403F50403F504F4050403F504F405F4F405F4F405F4F405F50405F4F405F504F604F405F\
504F60504F605F5060504F605F506F5F50605F506F5F506F5F506F5F506F605F70
5F506F605F70605F70605F70605F706F6070605F706F607F605F706F607F6F60706F607F6F6070\
6F607F6F607F70607F6F607F706F7F6F607F706F7F6F607F706F7F706F7F706F80
706F7F706F80706F7F7F7080706F7F7F7080706F807F7080706F807F70807F6F807F7080706F80\
7F70807F70807F708F7F70807F708F7F7080807F8F7F7080807F8F7F7080807F8F
7F7080807F8F7F7F8F807F8F7F7F8F807F8F807F8F807F8F807F8F807F8F807F8F80808F807F8F\
807F8F807F8F80808F807F8F80808F807F8F8F808F807F8F80808F807F8F8F808F
807F808F808F807F8F8F808F807F808F808F8080808F808F8080808F808F8080808F808F808080\
8F808F8F80808F808F8F80808F808F8F80808F8F8F8F80808F8F8F8F8080908F8F
8F8080908F8F8F8080908F8F8F8080908F808F8080908F8F8F8080908F80908F80908F80908F80\
908F80908F80908F80908F80908F80908F80908F80908F80908F80908F80908F80
90807F908F80908F7F908F8090807F908F8090807F908F808F807F908F7F90807F908F7F8F807F\
908F7F8F807F908F7F8F8070908F7F8F8070908F7F8F80708F807F8F80708F807F
8F80708F80708F80708F80708F80708F80708F80708F80708080708F80708F80708F8070807F6F\
8F807080806F8F8070807F6F8F8070807F6F8F8070807F6F80806F807F6F80806F
807F60807F6F807F60807F6F7F7060807F607F7060807F607F705F7F70607F705F7F705F7F6F5F\
7F705F706F5F7F705F706F507F6F50706F50706F50706F50706F5070604F706F50
6F604F70604F6F604F70604F6F604F6F604F6F60406F604F605F40000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030303F302F3F3F30403030403F304F3F30403F3F4F3F3F4F403F503F3F4F
404050403F504F405F403F504F405F4F405F4F405F4F405F504F6050405F504F60504F60504F60\
504F605F506F504F605F506F5F506F5F506F5F506F605F705F506F605F70605F6F
605F70605F70606070605F706F607F6060706F607F6F60706F607F6F607F6F607F6F607F706F7F\
6F607F706F7F6F607F706F80706F7F706F80706F7F706F80706F80706F80706F80
7F7080706F807F7080706F807F70807070807F70807070807F708F7F70807F708F7F70807F708F\
7F70807F708F7F7080807F8F7F708F807F8F7F708F807F8F7F7F8F807F8F7F7F8F
807F8F807F8F807F8F807F8F808090807F8F808090807F8F8F8090807F8F8F8090807F8F8F8090\
807F8F8F8090807F8F8F809080808F8F8090807F8F8F809080808F8F808F80808F
8F808F80808F8F808F80808F8F808F8F808F8F808F80808F8F808F8F808F8F808F8F808F8F8F8F\
8F808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F8080908F8F8F808F908F8F8F8080
908F8F8F8F8F908F8F8F8F80908F8F908F80908F8F908F80908F8F908F80908F8F908F8090908F\
908F80908F8F908F8090908F908F80908F80908F80909080908F80908F80908F80
908F80908F80908F80908F7F908F80908F7F908F80908F7F908F80908F7F908F80908F7F908F80\
908F7F908F7F8F807F908F7F8F807F908F7F8F807F908F7F8F807F908F7F8F8070
908F7F8F80708F8F7F8F8070908F7F8F80708F80708F80708F8F708F80708F80708F80708F8070\
8F80708F807080806F8F807080806F8F807080806F8F807080806F8F806F807F6F
8F806F807F6F807F6F807F60807F6F807F60807F607F7060807F607F7060807F607F705F7F7060\
7F705F7F705F7F705F7F705F706F5F7F705F706F507F705F706F50706F50706F50
706F5070604F706F506F604F706F506F604F70604F6F6040000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F303030403030403F30403030403F304F3F304F403F4F3F304F403F50
403F50404050403F504F405F4F405F4F405F4F405F4F405F4F405F504F604F405F504F60504F60\
504F60504F605F506F504F605F506F5F506F5F506F5F506F605F705F506F605F70
605F70605F70605F70606070605F706F607F605F706F607F6F60706F607F6F60706F607F6F607F\
6F6F7F6F607F706F7F6F607F706F806F607F706F80706F7F706F80706F7F706F80
706F80707080706F80707080706F807F7080706F807F7080706F807F7080706F807F70807F7080\
7F708F7F70807F708F7F70807F7F8F7F70807F7F8F7F708F807F8F7F708F807F8F
7F7F8F807F8F7F7F8F807F8F807F8F807F90807F8F807F8F807F8F808090807F8F808090807F8F\
808090807F8F80808F807F8F8F808F807F8F80808F807F8F8F808F807F8F80808F
80808F8F808F807F8F8F808F80808F8F808F80808F8F808F80808F8F808F8080808F808F80808F\
8F808F8080808F808F8F80808F8F8F8F80808F8F8F8F80808F8F8F8F80808F8F8F
8F8080908F8F8F8080908F8F8F8080908F8F8F8080908F8F8F8F80908F8F8F8F80908F8F908F80\
908F8F8F8F80908F80908F80908F808F8F80908F80908F80908F808F8F80908F80
908F80908F808F8F7F908F80908F7F908F808F807F908F8090807F908F808F807F908F7F8F807F\
908F7F8F807F908F7F8F807F908F7F8F80708F8F7F8F80708F8F7F8F80708F807F
8F80708F8F7F8F80708F80708F80708F80708F80708F80708F80708F80708080708F8070808070\
8F8070807F6F8F807080806F8F8070807F6F808070807F6F8F806F807F6F80806F
807F6F807F6F807F60807F6F7F7F60807F607F7060807F607F7060807F607F705F7F70607F705F\
7F705F706F5F7F705F706F507F705F706F50706F50706F50706F50706F50706F50
70604F706F506F604F70604F6F604F70604F6F60406F604F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F3030303F3030403F30403030403F3F4F3F304F3F3F4F3F3F4F403F50403F50
403F50403F504F405F4F405F4F405F4F405F504F604F405F504F604F405F504F60504F60504F60\
504F605F506F5050605F506F5F506F5F506F5F506F605F705F506F605F70605F70
605F70605F706F607F605F706F607F6060706F607F6F607F6F607F6F607F6F6F7F6F607F6F6F7F\
6F607F706F806F6F7F706F806F6F7F706F80706F80706F80706F80707080706F80
707080706F807F7080706F807F7080706F807F708F7070807F708F7070807F708F7F70807F708F\
7F70807F7F8F7F708F7F7F8F7F708F807F8F7F708F807F8F7F708F807F8F7F7F8F
807F8F7F7F8F807F90807F8F807F90807F8F808090807F8F808090807F8F8F8090807F8F808090\
807F8F8F809080808F8F8090807F8F8F809080808F8F809080808F8F809080808F
8F808F80808F8F808F80808F8F808F80808F8F808F80808F8F808F80808F8F8F8F8F808F8F808F\
80808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F8080908F8F8F8F8F
8F8F8F8F8080908F8F8F8F80908F8F8F8F80908F8F8F8F80908F8F8F8F80908F8F908F80908F8F\
908F8090908F908F80908F8F908F80908F8F908F80908F80908F80908F80908F80
908F80908F80908F80908F80908F808F8F7F908F80908F7F908F808F8F7F908F80908F7F908F7F\
8F807F908F808F807F908F7F8F807F908F7F8F807F908F7F8F8070908F7F8F807F
8F8F7F8F80708F8F7F8F80708F807F8F80708F8F7F8F80708F80708080708F80708F80708F8070\
8080708F80708080708F807080806F8F807080806F8F8070807F6F8F8070807F6F
80806F807F6F80806F807F60807F6F7F7F60807F6F7F7F60807F607F7060807F607F705F7F7060\
7F705F7F705F7F705F7F705F706F507F705F706F507F705F706F50706F50706F50
706F5070604F706F506F604F706F4F6F604F70604F6F604F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F302F2F3F3030403030403F304F3F30403F3F4F3F304F403F503F3F4F403F50
403F50403F5F403F504F405F403F5F4F405F4F405F4F4F604F405F504F604F4F60504F60504F60\
504F60504F605F506F504F605F506F5F506F5F506F5F506F5F5F705F506F605F70
5F5F70605F70605F706F607F605F706F607F60607F6F607F60607F6F607F6F607F6F6F7F6F607F\
6F6F7F6F607F706F806F607F706F806F6F7F706F806F6F7F706F80706F7F706F80
706F7F706F80706F80707080706F80707080706F807F7080706F807F70807070807F708F707080\
7F708F7F70807F708F7F70807F708F7F708F7F7F8F7F70807F7F8F7F708F807F8F
7F708F807F8F7F7F8F807F8F7F7F8F807F8F807F8F808090807F8F808090807F8F808090807F8F\
808090807F8F808090807F8F80808F807F8F80808F807F8F80808F807F8F80808F
807F8F80808F80808F8F808F80808F80808F80808F8F808F80808F8F808F80808F8F808F808080\
8F808F80808F8F808F8080808F808F8080808F8F8F8080808F8F8F8F80808F8F8F
8F80808F8F8F8F80808F8F8F8F80808F8F8F8F8F80908F8F8F8080908F8F8F8F80908F8F8F8F80\
908F8F8F8F80908F8F8F8F80908F808F8F80908F808F8F80908F808F8F80908F80
8F807F908F808F8F7F908F808F807F908F808F807F908F808F807F908F7F8F807F908F7F8F807F\
908F7F8F807F908F7F8F807F8F8F7F8F807F8F8F7F8F80708F807F8F80708F8F7F
8F80708F807F8F80708F807F8080708F80708080708F80708080708F80708080708F807080806F\
8F807080806F8F8070807F6F808070807F6F808070807F6F80806F807F6F80806F
807F6F807F6F807F60807F6F7F7F60807F607F7F60807F607F70607F7F607F705F7F70607F705F\
7F705F7F705F7F705F706F507F705F706F507F705F706F50706F50706F50706F50
6F604F706F506F604F706F506F604F6F604F6F604F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F3F3030403030403F3F4F3F304F3F3F4F3F304F403F50403F50403F50403F50
4F405F40405F4F405F4F405F4F40604F405F504F604F405F504F60504F60504F60504F605F506F\
504F605F506F50506F5F506F5F506F5F5F705F506F605F705F5F70605F70605F70
60607F605F706F607F605F706F607F6F607F6F607F6F607F6F6F806F607F6F6F806F607F706F80\
6F6F7F706F806F6F7F706F80706F80706F80706F80707080706F80707080706F80
7F7080706F807F7080706F807F708F7070807F708F7070807F708F7F70807F708F7F70807F7F8F\
7F708F7F7F8F7F708F7F7F8F7F708F7F7F8F7F708F807F8F7F7F8F807F8F7F7F8F
807F90807F8F807F90807F8F808090807F8F808090807F8F808090807F90808090807F8F8F8090\
80809080809080808F8F809080808F8F809080808F8F809080808F8F809080808F
8F809080808F8F809080808F8F8F9080808F8F808F80808F8F8F9080808F8F8F8F80808F8F8F8F\
8F808F8F8F8F80808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F8F8F8F8F8F8F808F
8F8F8F8F8F8F8F8F8F8F8F8F908F8F8F8F8F908F8F8F8F80908F8F8F8F8F908F8F8F8F8090908F\
908F8090908F8F8F8090908F908F80908F8F8F8F8090908F908F80908F808F8F80
908F80908F80908F808F8F80908F808F8F80908F808F8F7F908F808F8F7F908F808F8F7F908F80\
8F8F7F908F808F807F908F808F8F7F908F7F8F807F908F7F8F807F8F8F7F8F807F
8F8F7F8F807F8F8F7F8F80708F8F7F8F80708F8F7F8F80708F8F7F8F80708F80708080708F8070\
8080708F80708080708F80708080708F807080806F8F807080806F808070807F6F
80806F807F6F80806F807F6F80806F807F60807F6F7F7F60807F6F7F7F60807F607F705F807F60\
7F705F7F70607F705F7F705F7F705F7F705F706F507F705F706F50706F5F706F50
706F50706F50706F506F604F706F506F604F706F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F30\
30303F302F3F3030403030403F3F4F3F304F3F3F503F3F4F403F50403F50403F50
403F504F405F403F504F405F4F405F4F40604F405F4F4F604F4060504F604F4F60504F60504F60\
5F506F504F605F506F50506F5F506F5F506F5F5F705F506F605F705F5F70605F70
605F7060607F605F706F607F60607F6F607F60607F6F607F6F607F6F6F7F6F607F6F6F7F6F607F\
706F806F607F706F806F6F7F706F806F6F7F706F806F6F80706F806F6F80706F80
706F80707080706F80707080706F807F708F706F807F708F7070807F708F7070807F708F707080\
7F708F7070807F708F7F708F7F7F8F7F708F7F7F8F7F708F7F7F8F7F708F7F7F8F
7F7F8F807F907F7F8F807F907F7F8F8080907F7F8F808090807F8F808090807F8F808090807F8F\
808090807F8F808090807F8F808090807F8F80809080808F80809080808F80808F
80808F8F809080808F80808F80808F8F808F80808F80808F80808F8F808F80808F8F808F80808F\
8F8F8F80808F8F808F80808F8F8F8F8080808F8F8F80808F8F8F8F8080808F8F8F
8F80808F8F8F8F80808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F80908F8F8F8F808F8F8F8F8F80\
908F8F8F8F80908F8F8F8F80908F8F8F8F80908F808F8F80908F808F8F80908F80
8F8F80908F808F8F7F908F808F8F7F908F808F807F8F8F808F8F7F908F808F807F8F8F7F8F807F\
908F7F8F807F8F8F7F8F807F8F8F7F8F807F8F8F7F8F807F8F8F7F8F80708F807F
8F80708F8F7F8080708F807F8080708F807F8080708F80708080708F80708080708F8070808070\
8F807080806F80807080806F808070807F6F808070807F6F808070807F6F80806F
807F6F80806F7F7F60807F6F807F60807F6F7F7F60807F607F7060807F607F705F7F7F607F705F\
7F706070705F7F705F70705F7F705F706F5070705F706F5070705F706F50706F50
706F50706F506F604F706F506F604F706F50000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F3F\
302F3F3F30403030403F3F4F3F304F403F503F3F50403F50403F5040405F403F50
4F405F403F5F4F405F4F405F4F40604F405F504F604F4F60504F604F4F6050506F504F6050506F\
504F605F506F5F506F5F506F5F506F5F5F705F506F605F705F5F70605F70605F70
60607F605F706F607F60607F6F607F6F607F6F6F806F607F6F6F806F607F706F806F607F706F80\
6F607F706F806F6F80706F806F6F80706F80706F80706F80706F80707080706F80
707080706F807F708F70708070708F7070807F708F7070807F708F7070807F708F7F708F7F708F\
7F708F7F7F8F7F708F7F7F8F7F708F7F7F8F7F7F8F7F7F8F7F7F8F807F907F7F8F
807F907F7F8F808090807F8F808090807F8F808090807F90808090807F90808090808090808090\
807F8F80809080809080809080808F8F809080808F80809080808F8F809080808F
8F809080808F8F809080808F8F809080808F8F8F9080808F8F8F8F80808F8F8F9080808F8F8F8F\
80808F8F8F8F80808F8F8F8F80808F8F8F8F8F808F8F8F8F80808F8F8F8F8F8F8F
8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F80908F8F8F8F8F908F8F\
8F8F8090908F8F8F80908F8F8F8F8090908F8F8F80908F8F8F8F80908F8F8F8F80
908F808F8F80908F808F8F80908F808F8F80908F808F8F80908F808F8F7F908F808F8F7F908F80\
8F807F908F808F8F7F8F8F7F8F807F8F8F7F8F807F8F8F7F8F807F8F8F7F8F807F
8F8F7F8F80708F8F7F8F80708F8F7F8080708F8F7F8080708F807F8080708F807F8080708F8070\
8080708F807080807080807080806F8F807080806F808070807F6F808070807F6F
80806F807F6F80806F807F6F807F6F7F7F60807F6F7F7F60807F607F7F60807F607F70607F7F60\
7F705F7F7F607F705F7F705F70705F7F705F70705F7F705F706F507F705F706F50
706F50706F50706F506F6F50706F506F604F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F30\
30303F302F403F304F30304F3F3F4F3F304F403F503F3F50403F50403F5040405F
403F504F405F40405F4F405F4F405F4F40604F405F504F604F4060504F60504F6050506F504F60\
50506F504F6F5F506F50506F5F506F5F506F5F5F705F506F605F705F5F70605F70
605F7060607F605F706F607F60607F6F607F6F607F6F607F6F607F6F607F6F607F6F6F806F607F\
6F6F806F607F706F806F607F706F806F6F80706F806F6F80706F806F6F80707080
6F6F80707080706F80707080706F80707080706F807F708F7070807F708F7070807F708F707080\
7F708F70708F7F7F8F70708F7F7F8F7F708F7F7F8F7F708F7F7F8F7F7F8F7F7F90
7F7F8F7F7F8F7F7F8F807F907F7F8F807F907F7F8F8080907F7F8F808090807F8F808090807F8F\
808090807F8F808090807F8F80809080808F808090807F8F80809080808F808090
80808F80808F80808F80809080808F80808F80808F80808F80808F80808F80808F8F808F80808F\
80808F80808F8F8F8F80808F8F8F8F80808F8F8F8F8080808F8F8F80808F8F8F8F
8080808F8F8F8080808F8F8F8080808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F80\
8F8F8F8F8F808F8F8F8F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F80
8F8F808F8F808F8F808F8F808F807F8F8F808F8F7F8F8F808F807F8F8F808F807F8F8F808F807F\
8F8F7F8F807F8F8F7F8F807F8F8F7F8F807F8F8F7F8080708F807F8080708F8F7F
8080708F807F8080708F807F8080708F80708080708F8070808070808070808070808070808070\
808070808070808070807F6F808070807F6F808070807F6F80806F807F6F80806F
7F7F6F807F6F7F7F6F807F6F7F7F60807F6F7F7F60807F6F7F70607F7F607F70607F7F607F705F\
7F70607F705F7F705F70705F7F705F706F5F7F705F706F5070705F706F50706F50
6F6F50706F506F6F50706F506F604F6F6F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F2F3030303F\
30303F3F304F3F304F3F3F503F3F50403F503F3F5040405F403F5040405F403F5F
4F405F4F405F4F405F4F405F4F4F604F4060504F604F4F60504F6F504F6050506F504F605F506F\
50506F5F506F5F506F5F5F705F506F5F5F705F5F70605F70605F7060607F605F70
6F607F60607F6F607F60607F6F6F806F607F6F6F806F607F6F6F806F607F6F6F806F607F706F80\
6F6F80706F806F6F80706F806F6F80706F806F6F8070708F706F80707080706F80
70708F70708070708F7070807F708F7070807F708F7070807F708F70708F7F708F70708F7F7F8F\
7F708F7F7F8F7F708F7F7F907F7F8F7F7F907F7F8F807F907F7F8F807F907F7F8F
8080907F7F8F8080907F7F8F808090807F90808090807F90808090808090808090808090808090\
8080908080908080908080908080908080908080908F809080809080809080808F
8F8F9080808F8F809080808F8F8F9080808F8F8F9080808F8F8F9080808F8F8F9080808F8F8F90\
80808F8F8F9080808F8F8F90808F8F8F8F8F80808F8F8F8F8F8F8F8F8F8F808F8F
8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F908F8F8F8F8F8F8F8F8F8F90908F\
8F8F8F90908F8F8F8090908F8F8F80908F8F8F8F8090908F8F8F80908F8F8F8F80
908F8F8F8F808F8F808F8F80908F808F8F808F8F808F8F80908F808F8F808F8F808F8F7F908F80\
8F8F7F8F8F808F807F8F8F808F8F7F8F8F7F8F807F8F8F7F8F807F8F8F7F8F807F
8F8F7F8F807F8F8F7F8080708F8F7F8080708F8F7F8080708F8F7F8080708F807F8080708F8F7F\
8080708080708080708F807080807080807080806F80807080806F808070807F6F
808070807F6F80806F807F6F80806F807F6F807F6F7F7F60807F6F7F7F60807F607F7060807F60\
7F70607F7F607F705F7F7F607F705F7F705F70705F7F705F70705F70705F706F50
70705F706F50706F506F6F50706F50000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F2F302F2F3F\
3030403030403F304F3F304F403F503F3050403F50403F5040405F403F5040405F
403F5F4F405F40405F4F405F4F405F4F4F604F405F504F604F4F60504F6F504F6050506F504F60\
5F506F50506F5F506F5F506F5F5F705F506F5F5F705F5F70605F705F5F70605F7F
605F7060607F605F7F6F607F60607F6F607F60607F6F6F806F607F6F6F806F607F6F6F806F607F\
6F6F806F607F6F6F806F6F80706F806F6F80706F806F6F807070806F6F80707080
706F8070708F706F8070708F706F8070708F70708070708F7070807F708F7070807F708F70708F\
7F7F8F70708F7F7F8F7F708F7F7F8F7F708F7F7F8F7F7F8F7F7F907F7F8F7F7F90
7F7F8F807F907F7F8F807F907F7F8F8080907F7F8F808090807F908080907F7F90808090807F90\
808090807F8F80809080809080809080808F80809080808F80809080808F808090
80808F80809080808F80809080808F80809080808F80809080808F8F8F9080808F808F8F80808F\
8F8F8F80808F8F8F8F80808F8F8F8F80808F8F8F8F80808F8F8F8F80808F8F8F8F
80808F8F8F8F80808F8F8F8F808F8F8F8F8F808F808F8F8F8F8F8F8F8F8F8F8F808F8F8F8F8F80\
8F8F8F8F8F808F8F8F8F8F808F8F8F8F8F808F8F808F8F808F8F8F8F8F808F8F80
8F8F808F8F808F8F808F8F808F8F808F8F808F807F8F8F808F8F7F8F8F808F807F8F8F808F807F\
8F8F808F807F8F8F7F8F807F8F8F7F80807F8F8F7F8F807F8F8F7F80807F8F8F7F
80807F8F8F7F8080708F807F8080708F807F80807080807F80807080807F808070808070808070\
808070808070808070808070808070807F6F808070807F6F8080707F7F6F80806F
7F7F6F80806F7F7F6F807F6F7F7F60807F6F7F7F60807F6F7F7F60807F607F70607F7F607F7060\
7F7F607F705F7F706070705F7F705F70705F7F705F706F5F70705F706F50706F50
706F50706F506F6F50706F506F6F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F2F3030303F\
3030403F304F3F304F3F3F503F3F50403F5F403F50403F5F403F504F405F40405F
4F405F40405F4F4F604F405F4F4F604F4060504F604F4F60504F6F504F6050506F504F6F5F506F\
50506F5F50705F506F5F5F705F506F605F705F5F70605F705F5F7060607F605F70
60607F60607F6F607F60607F6F607F60607F6F6F806F607F6F6F806F607F6F6F806F6F806F6F80\
6F6F80706F806F6F80706F806F6F807070806F6F807070806F6F8070708F706F80
70708F706F8070708F70708070708F7070807F708F70708F7F708F70708F7F7F8F70708F7F7F8F\
70708F7F7F8F7F708F7F7F8F7F708F7F7F907F7F8F7F7F907F7F8F807F907F7F8F
807F907F7F8F8080907F7F908080907F7F90808090807F90808090808090808090808090808090\
808090808090808090808090808090808090808090808090808090808090808090
80809080808F8F8F9080808F808F9080808F8F8F9080808F8F8F9080808F8F8F9080808F8F8F90\
80808F8F8F9080808F8F8F9080808F8F8F90808F8F8F8F8F808F8F8F8F8F808F8F
8F8F8F808F8F8F8F8F8F8F8F8F8F8F808F8F8F8F8F8F8F8F8F8F8F8F8F8F8F908F8F8F8F8F908F\
8F8F8F8F908F8F8F8F8F8F8F8F8F808F908F8F8F808F8F8F8F8F808F8F8F8F8F80
8F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F7F8F8F808F8F7F8F8F80\
8F807F8F8F808F8F7F8F8F808F807F8F8F808F807F8F8F7F80807F8F8F7F8F807F
8F8F7F80807F8F8F7F80807F8F8F7F8080708F8F7F8080708F807F8080708F8F7F808070808070\
808070808070808070808070808070808070808070808070807F6F808070807F6F
80806F7F7F6F80806F7F7F6F80806F7F7F6080806F7F7F60807F6F7F7F60807F6F7F7F607F7F60\
7F70607F7F607F705F7F7F6070705F7F706070705F7F705F706F5F7F705F706F50
70705F706F50707050706F50706F50000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F3F2F2F3F\
3030403030403F3F503F304F3F3F503F3F50403F503F3F50403F5F403F5040405F
403F5F4F405F40405F4F4F604F405F4F4F604F4F60504F604F4F60504F6F504F6050506F504F60\
5F506F50506F5F507050506F5F5F705F5070605F705F5F70605F70605F7060607F
605F7060607F60607F6F607F60607F6F607F60607F6F6F8060607F6F6F806F607F6F6F806F607F\
6F6F806F6F806F6F806F6F806F6F806F6F80706F806F6F80706F806F6F8070708F
6F6F8070708F706F8070708F706F8070708F70708070708F70708070708F70708F7F708F70708F\
7F708F70708F7F7F8F70708F7F7F8F7F708F7F7F907F708F7F7F907F7F8F7F7F90
7F7F8F7F7F907F7F8F8080907F7F8F8080907F7F908080907F7F908080907F7F908080907F7F90\
8080908080908080907F7F8F80809080809080809080808F80809080808F808090
80808F80809080808F80809080808F80809080808F808F9080808F80809080808F808F9080808F\
808F8F80808F808F8F80808F808F8F80808F8F8F8F80808F808F8F80808F8F8F8F
80808F8F8F8F808F8F8F8F8F80808F8F8F8F808F8F8F8F8F808F808F8F8F808F8F8F8F8F808F80\
8F8F8F8F8F808F8F8F808F808F8F8F8F8F808F8F8F808F808F8F808F8F808F8F80
808F808F8F80808F808F8F808080808F8F80808F808F8F8080807F8F8F8080807F8F8F8080807F\
8F8F8080807F8F8F8080807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F
80807080807F8080708F807F80807080807F80807080807F808070808070808070808070808070\
8080708080708080707F7F6F8080707F7F6F8080707F7F6F80806F7F7F6F80806F
7F7F6F807F6F7F7F6F807F6F7F7F607F7F6F7F7F607F7F6F7F7F607F7F607F7F607F7F607F7060\
7F7F607F705F7F7F6070705F7F705F70705F7F705F706F5070705F706F5070705F
6F6F50706F506F6F50706F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000302F30302F3F3F3040\
3030403F3F4F3F3050403F503F3F50403F5F403F5040405F403F5F4F405F40405F
4F40604F405F4F4F604F4060504F604F4F60504F604F4F6050506F504F6F50506F50506F5F5070\
50506F5F5F705F506F5F5F705F5F70605F705F5F7060607F605F7060607F605F7F
6F607F60607F6F607F60607F6F6F806F607F6F6F806F607F6F6F806F6F806F6F806F6F80706F80\
6F6F80706F806F6F807070806F6F807070806F6F8070708F706F8070708F706F80
70708F70708070708F70708070708F70708F70708F70708F7F7F8F70708F7F7F8F70708F7F7F90\
70708F7F7F9070708F7F7F907F7F8F7F7F907F7F8F7F7F907F7F907F7F907F7F90
8080907F7F908080907F7F908080907F7F908080907F809080809F80809080809F80809080809F\
808090808090808090808090808090808090808090808F90808090808F90808090
808F90808090808F908080908F8F90808090808F9080808F8F8F908080908F8F9080808F8F8F90\
808F8F8F8F9080808F8F8F90808F8F8F8F90808F8F8F8F90808F8F8F8F90808F8F
8F8F90808F8F8F8F8F808F8F8F8F908F8F8F8F8F8F808F8F8F908F8F8F8F8F908F8F8F8F8F908F\
8F8F8F8F908F8F8F8F8F908F8F8F8F8F8F8F8F8F808F908F8F8F808F8F8F8F8F80
8F8F8F8F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F7F8F8F80\
8F8F7F8F8F80808F7F8F8F808F8F7F8F8F8080807F8F8F808F807F8F8F7F80807F
8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F8080708F8F7F80807080807F808070808F7F\
80807080807080807080807080807080807080806F80807080806F8080707F7F6F
808070807F6F80806F7F7F6F80806F7F7F6F807F6F7F7F60807F6F7F7F607F7F6F7F7F60807F6F\
7F7F607F7F607F70607F7F607F705F7F706070705F7F706070705F7F705F706F50
70705F706F5070705F706F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F3030303F302F3F\
3F304F30304F3F3F503F3F50403F503F3F50403F5F403F5040405F403F5F4F405F
40405F4F406040405F4F4F604F4060504F604F4F60504F60504F6050506F504F6050506F50506F\
5F507050506F5F5F705F506F5F5F705F5F70605F705F5F7060607F605F7060607F
605F7F60607F60607F6F607F60607F6F6F8060607F6F6F806F607F6F6F806F607F6F6F806F6F80\
6F6F806F6F806F6F806F6F80706F806F6F80706F806F6F8070708F6F6F8070708F
6F6F8070708F6F6F8070708F70708070708F70708070708F70708F70708F70708F70708F70708F\
7F7F8F70708F7F7F8F70708F7F7F9070708F7F7F907F7F8F7F7F907F7F8F7F7F90
7F7F8F7F80907F7F8F7F80907F7F908080907F7F908080907F7F908080907F7F908080907F8090\
8080907F80908080908080908080907F8090808090808090808090808090808090
808090808F9080808F80809080808F808F9080808F808F9080808F808F9080808F808F9080808F\
808F9080808F808F9080808F8F8F9080808F808F8F808F8F8F8F9080808F8F8F8F
808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F808F8F\
8F8F8F808F808F8F8F808F808F8F8F808F808F8F8F808F808F8F8F808F808F8F80
808F808F8F80808F808F8F80808F808F8F808080808F8F80808F808F8F8080807F8F8F8080807F\
8F8F8080807F8F8F8080807F8F8F8080807F8F8F7F80807F8F8F7F80807F808F7F
80807F8F8F7F80807080807F808070808F7F80807080807F80807080807F808070808070808070\
8080707F80708080707F80708080707F7F6F8080707F7F6F8080707F7F6F80806F
7F7F6F80806F7F7F6F807F6F7F7F6F807F6F7F7F607F7F6F7F7F607F7F6F7F70607F7F607F7060\
7F7F607F705F7F7F6070705F7F7F6070705F7F705F70705F7F705F706F5070705F
706F5070705F6F6F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000030303F302F3F3F3040\
3030403F3F503F3F50403F503F3F5040405F403F5F40405F403F5F4F406040405F
4F40604F405F4F4F604F4F604F4F604F4F6050506F504F6050506F504F6F5F506F50506F5F5070\
50506F5F5F705F50705F5F705F5F70605F7F5F5F7060607F605F7060607F60607F
60607F60607F6F608060607F6F6F8060607F6F6F806F60806F6F806F6F806F6F806F6F806F6F80\
6F6F80706F806F6F80706F806F6F8070708F6F6F8070708F6F6F8070708F707080
70708F70708070708F70708F70708F70708F707F8F70708F707F8F70708F7F7F8F70708F7F7F8F\
70708F7F7F90707F8F7F7F90707F8F7F7F907F7F8F7F7F907F7F8F7F80907F7F90
7F80907F7F908080907F7F908080907F7F9080809F7F809080809F7F809080809F80809080809F\
80809080809F808090808090808090808F90808090808090808090808F90808090
808F90808090808F90808090808F90808090808F90808090808F9080808F8F8F90808F90808F90\
80808F8F8F90808F8F8F8F90808F8F8F8F90808F8F8F8F90808F8F8F8F90808F8F
8F8F90808F8F8F8F90808F8F8F8F8F808F8F8F9090808F8F8F8F8F808F8F8F908F8F8F8F8F908F\
808F8F8F908F8F8F8F8F8F8F808F808F908F808F808F8F8F808F808F8F8F808F80
8F8F8F808F808F8F8F808F808F8F80808F808F8F80808F808F8F80808F808F8F80808F808F8F80\
808F7F8F8F80808F7F8F8F8080807F8F8F8080807F8F8F7F80807F8F8F7F80807F
8F8F7F80807F8F8F7F80807F808F7F808070808F7F808070808F7F808070808F7F80807080807F\
80807080807F8080708080707F80708080708080708080707F7F6F8080707F7F6F
8080707F7F6F80806F7F7F6F80806F7F7F6F80806F7F7F607F7F6F7F7F607F7F6F7F7F607F7F60\
7F70607F7F607F70607F7F6070705F7F7F6070705F7F705F70705F7F705F70705F
70705F706F5070705F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F302F2F30303040302F40\
3F304F3F304F3F3F503F3050403F503F3F5040405F403F5040405F40405F4F4060
40405F4F40604F405F4F4F604F40604F4F604F4F60504F6F4F4F6050506F504F6F50506F50506F\
5F507050506F5F5F705F50705F5F705F5F70605F7F5F5F70605F7F605F7060607F
605F7060607F60607F6F607F60607F6F607F60607F6F6F8060607F6F6F806F6F7F6F6F806F6F7F\
6F6F806F6F806F6F806F6F806F6F806F6F807070806F6F807070806F6F8070708F
6F6F8070708F70708070708F6F708070708F70708F70708F70708F70708F70708F7F7F8F70708F\
707F8F70708F7F7F9070708F7F7F90707F8F7F7F90707F8F7F7F907F7F8F7F7F90
7F7F8F7F7F907F7F907F80907F7F907F80907F7F908080907F7F908080907F80908080907F8090\
8080907F80908080907F80908080907F80908080907F8090808090808090808090
7F8090808090808090808F9080808F808F9080808F808F9080808F808F9080808F808F9080808F\
808F9080808F808F9080808F808F9080808F808F9080808F808F8F808F8F808F90
808F8F808F8F808F8F8F8F8F808F8F808F8F808F8F8F8F8F808F8F808F8F808F8F8F8F8F808F80\
8F8F8F808F8F8F8F8F808F808F8F8F808F808F8F8F808F80808F8F808F808F8F8F
808F80808F80808F808F8F80808080808F80808F808F8F8080807F808F8080807F8F8F8080807F\
808F8080807F808F8080807F808F7F80807F808F7F80807F808F7F80807F808F7F
80807F80807F80807F808F7F80807080807F80807080807F7F807080807F7F807080807F7F8070\
8080707F80708080707F7F6F8080707F7F6F8080707F7F6F8080707F7F6F80806F
7F7F6F7F7F6F7F7F6F7F7F6F7F7F607F7F6F7F7F607F7F6F7F7F607F7F607F7F607F7F60707060\
7F7F607070607F7F6070705F7F706070705F7F706070705F70705F70705F70705F
706F5070705F6F6F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F3030303F30303F3F3F4F\
3F304F3F3F503F3F50403F5F403F5040405F403F5F4F405F40405F4F406040405F
4F4F604F40604F4F604F4F60504F6F4F4F6050506F504F6F50506F50506F5F506F50506F5F5F70\
5F50705F5F705F5F70605F7F5F5F70605F7F5F5F7060607F605F7F60607F60607F
60607F60607F6F607F60607F6F6F8060607F6F6F806F607F6F6F806F6F806F6F806F6F80706F80\
6F6F80706F806F6F8070708F6F6F8070708F6F6F8070708F70708070708F707080
70708F70708F70708F70708F707F8F70708F707F8F70708F7F7F9070708F7F7F9070708F7F7F90\
707F8F7F7F90707F8F7F7F907F7F907F7F907F7F907F80907F7F907F80907F7F90
8080907F7F908080907F7F9080809F7F809080809F7F809080809F80809080809F7F8090808F9F\
80809080809F808090808F9F808090808F9F808090808F9F808090808F90808090
808F9F808090808F90808090808F90808090808F90808090808F90808F90808F908080908F8F90\
808F90808F90808F8F8F8F90808F8F808F90808F8F8F8F90808F8F8F8F90808F8F
8F8F90808F8F8F8F90808F8F8F8F90808F8F8F8F8F808F8F8F9090808F8F8F8F8F808F8F8F908F\
808F8F8F8F8F808F8F8F908F808F8F8F8F8F808F808F908F808F8F8F8F8F808F80
8F8F8F808F808F8F8F808F808F8F8F808F808F8F80808F808F8F80808F808F8F80808F808F8F80\
808F808F8F80808F7F8F8F80808F7F8F8F8080807F8F8F80808F7F8F8F8080807F
8F8F8080807F808F7F80807F808F7F80807F808F7F80807F808F7F80807F808F7F808070808F7F\
8080708080707F80708080708080708080707F80708080707F80708080707F7F6F
8080707F7F6F80806F7F7F6F80806F7F7F6F7F7F6F7F7F607F7F6F7F7F607F7F6F7F7F607F7F6F\
7F7F607F7F607F70607F7F607F70607F7F6070705F7F7F6070705F7F705F70705F
7F705F70705F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F3F2F2F3F303040303040\
3F304F3F304F3F3F503F3F50403F5F3F3F5040405F403F5F40405F40405F4F4060
40405F4F4F604F40604F4F604F4F60504F6F4F4F6050506F504F6F50506F50506F50506F50506F\
5F5F7050506F5F5F705F5F70605F7F5F5F70605F7F5F5F7060607F5F5F7060607F
60607F60607F60607F60607F60607F6F6F8060607F6F6F806F607F6F6F806F6F806F6F806F6F80\
6F6F806F6F806F6F806F6F8070708F6F6F8070708F6F6F8070708F6F6F8070708F
6F708070708F6F708070708F70708F70708F70708F70708F70708F707F8F70708F707F8F70708F\
7F7F9070708F7F7F90707F8F7F7F90707F8F7F7F907F7F8F7F7F907F7F8F7F7F90
7F7F907F80907F7F907F80907F7F908080907F7F908080907F809080809F7F809080809F7F8090\
80809F7F80908080907F80908080907F80908080907F80908080907F8090808090
7F8090808F907F8090808090808090808F907F808F808F90808090808F9080808F808F9080808F\
808F9080808F808F9080808F808F9080808F808F9080808F808F9080808F808F8F
808F8F808F8F80808F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F\
808F8F808F80808F8F808F80808F8F808F80808F8F808F80808F8F808F80808F8F
808F80808F8F808F80808F80808F80808F80808080808F80808F80808F8080807F808F8080807F\
808F8080807F808F8080807F808F8080807F808F7F80807F808F7F80807F808F7F
80807F808F7F80807F808F7F80807F808F7F7F807080807F7F807080807F7F807080807F7F8070\
8080707F80708080707F80708080707F7F6F8080707F7F6F8080707F7F6F7F806F
7F7F6F80806F7F7F6F7F7F6F7F7F6F7F7F6F7F7F607F7F6F7F7F607F7F6F7070607F7F607F7060\
7F7F607070607F7F6070705F7F7F6070705F7F706070705F7F705F70705F70705F
70705F70705F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F302F2F303030403030403F304F\
3F304F3F3F503F3F50403F5F3F3F5040405F40405F40405F40405F4F4F604F4060
4F4F604F40604F4F6F4F4F60504F6F4F4F6050506F504F6F50506F50506F5F507050506F5F5F70\
5F50705F5F705F5F705F5F705F5F7060607F5F5F7060607F5F5F7060607F60607F
60607F60607F6F6F8060607F6F6F8060607F6F6F806F6F806F6F806F6F806F6F806F6F806F6F80\
6F6F8070708F6F6F8070708F6F6F8070708F6F6F8070708F6F6F8070708F70708F
70708F70708F70708F70708F70708F70708F707F8F70708F707F8F70708F7F7F9070708F7F7F90\
70708F7F7F90707F8F7F7F90707F8F7F7F907F7F907F7F907F7F907F80907F7F90
7F80907F7F9080809F7F809080809F7F809080809F7F809080809F7F809080809F7F809080809F\
7F809080809F7F809080809F7F8090808F9F7F80908080907F8090808F90808090
808F907F8090808F90808090808F907F8090808F90808090808F9080808F808F90808090808F90\
80808F808F90808F8F808F9080808F808F90808F8F808F90808F8F808F90808F8F
808F8F808F8F808F90808F8F808F8F808F8F808F8F808F8F808F8F808F8F8F8F8F808F8F808F8F\
808F8F808F8F808F8F808F8F808F80808F8F808F80808F8F808F80808F8F808F80
808F8F808F80808F8F808F80808F80808F80808F8F808F80808F80808F80808F80808F80808F80\
808F80808F80808F80808F8080807F808F80808F7F808F8080807F808F8080807F
808F7F80807F808F7F80807F808F7F80807F808F7F80807F808F7F7F8070808F7F80807080807F\
7F807080807F7F80708080707F80708080707F80708080707F7F6F8080707F7F6F
8080707F7F6F8080707F7F6F7F806F7F7F6F7F806F7F7F6F7F7F6F7F7F607F7F6F7F7F607F7F60\
7070607F7F607F70607F7F6070705F7F7F6070705F7F7F6070705F7F7F6070705F
70705F70705F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F30302F3F302F3F303040303040\
3F304F3F304F3F3F503F3F50403F5F403F5F40405F403F5F40405F40405F4F4F60
40405F4F4F604F4F604F4F6F4F4F60504F6F4F4F6050506F504F6F50506F50506F5F507050506F\
5F5F705F50705F5F705F5F705F5F705F5F70605F7F5F5F7060607F5F5F7060607F
5F5F7F60607F60607F60607F60607F60608060607F6F6F8060607F6F6F806F6F806F6F806F6F80\
6F6F806F6F806F6F806F6F806F6F806F6F8070708F6F6F8070708F6F6F8070708F
6F6F8070708F6F708F70708F6F708F70708F70708F707F8F70708F707F8F70708F707F9070708F\
707F9070708F7F7F9070708F7F7F90707F8F7F7F90707F8F7F7F907F7F907F8090
7F7F907F80907F7F907F80907F7F907F80907F7F907F80907F7F907F80907F809080809F7F7F90\
7F80907F80908080907F80907F80907F80908080907F80908080907F8090808090
7F808F8080907F80908080907F808F8080907F808F808F907F808F8080907F808F808F907F808F\
808F907F808F808F907F808F808F8F7F808F808F8F7F808F808F8F7F808F808F8F
7F808F808F8F80808F808F8F7F808F808F8F808F8F808F8F7F8080808F8F808F80808F8F7F8080\
808F8F808F80808F8F7F8080808F8F808F80808F8F7F8080808F80808F80808F8F
7F8080808F80808F80808F807F8080808F80808080808F807F8080808F80808080808F807F807F\
808F8080807F808F807F807F808F8080807F808F807F807F808F7F80807F808F7F
7F807F808F7F80807F808F7F7F807080807F7F807080807F7F807080807F7F807080807F7F8070\
8080707F80708080707F7F707F80707F7F707F80707F7F6F7F80707F7F6F7F8070
7F7F6F7F7F6F7F7F6F7F7F6F7F7F607F7F6F7F7F607F7F6F707F607F7F6F707F607F7F60707060\
7F7F607070607F7F6070705F70706070705F7F706070705F70705F70705F70705F
706F5F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000030303F302F3F3030403030403F3F4F\
3F304F3F3F503F3F5040405F403F5F40405F40405F4F406040405F4F4F604F4060
4F4F604F4F604F4F6F4F4F6050506F504F6F50506F50506F5F507050506F5F5F705050705F5F70\
5F5F705F5F705F5F70605F7F5F5F7060607F5F5F7060607F60607F60607F60607F
60608060607F60608060607F6F6F806060806F6F80606F806F6F806F6F806F6F806F6F8070708F\
6F6F806F708F6F6F8070708F6F6F8070708F6F6F8070708F6F708F70708F6F708F
70708F70708F70708F70708F707F9070708F707F8F70708F7F7F90707F8F7F7F90707F8F7F7F90\
707F8F7F7F90707F8F7F7F907F7F907F7F907F7F907F80907F7F907F80907F7F90
7F809F7F80907F809F7F7F9080809F7F809080809F7F809080809F7F809080809F7F809080809F\
7F809080809F7F809080809F7F80908080907F8090808F9F7F80908080907F8090
808F907F8090808F907F8090808F907F8090808F907F8090808F907F8090808F907F808F808F90\
80808F808F907F808F808F9080808F808F907F808F808F90808F8F808F9080808F
808F90808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F\
808F8F808F8F808F80808F8F808F8F808F8F808F80808F8F808F80808F8F808F80
808F8F808F80808F8F808F80808F8F808F80808F80808F80808F80808F80808F80808F80808F80\
808F80808F80808F80808F80808F80808F8080807F808F80808F7F808F8080807F
808F8080807F808F7F80807F808F7F80807F808F7F7F807F808F7F80807F808F7F7F8070808F7F\
7F807080807F7F807080807F7F80708080707F80708080707F80708080707F7F6F
8080707F7F6F7F80707F7F6F7F806F7F7F6F7F7F6F7F7F6F7F7F6F7F7F607F7F6F7F7F607F7F6F\
7F7F607F7F607070607F7F607070607F7F6070705F7F7F6070705F7F706070705F
7F7060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F302F2F3030303F302F3F30304030304F\
3F3F503F304F3F3F503F3F5040405F403F5F40405F40405F4F406040405F4F4F60
4F40604F4F604F4F604F4F6F4F4F6050506F4F4F6F50506F50506F50507050506F5F5F70505070\
5F5F705050705F5F705F5F705F5F7F5F5F705F5F7F5F5F7060607F5F5F7060607F
60607F60607F60607F60607F60607F6F6F8060607F6F6F80606F806F6F80606F806F6F806F6F80\
6F70806F6F806F70806F6F8070708F6F6F8070708F6F6F8070708F6F6F8070708F
6F708F70708F6F708F70708F70708F707F8F70708F707F8F70708F707F9070708F707F9070708F\
7F7F9070708F7F7F90707F8F7F7F90707F8F7F7F90707F907F8090707F907F8090
7F7F907F80907F7F907F80907F7F907F80907F7F907F80907F80907F80907F7F907F80907F8090\
7F80907F80907F80907F80907F80907F80907F80907F80907F80907F80907F8090
7F80907F80907F808F7F80907F808F8080907F808F7F80907F808F8080907F808F8080907F808F\
808F907F808F808F8F7F808F808F907F808F808F8F7F808F808F8F7F808F808F8F
7F808F808F8F7F808F808F8F7F808F808F8F7F8080808F8F7F808F808F8F7F8080808F8F7F8F80\
808F8F7F8080808F8F7F8080808F8F7F8080808F8F7F8080808F8F7F8080808F80
7F8080808F807F8080808F807F8080808F807F8080808F807F8080808F807F807F808F807F807F\
808F807F807F808F807F807F808F807F807F808F807F807F808F807F807F808F7F
7F807F808F7F7F807F808F7F7F807F808F7F7F807080807F7F807080807F7F807080807F7F8070\
80807F7F80707F80707F80707F80707F7F6F7F80707F7F6F7F80707F7F6F7F8070
7F7F6F7F806F7F7F6F7F7F6F7F7F6F7F7F6F707F607F7F6F707F607F7F6F7070607F7F60707060\
7F7F607070607F7F607070607F7F6070705F70706070705F70706070705F70705F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F3030303F302F3F3F30403030403F304F\
3F304F3F3F503F3F5040405F3F3F5F40406040405F4F406040405F4F4F604F4F60
4F4F604F4F60504F6F4F4F6F50506F4F4F6F50507050506F50507050506F5F5F705050705F5F70\
5F50705F5F7F5F5F705F5F7F5F5F7060607F5F5F7F60607F5F5F7F60607F60607F
60607F60607F606F8060607F6F6F8060607F6F6F80606F806F6F806F6F806F708F6F6F806F708F\
6F6F8070708F6F6F8070708F6F6F8070708F6F708F70708F6F708F70708F70708F
70708F70708F707F8F70708F707F8F70708F707F9070708F707F9070708F7F7F90707F8F7F7F90\
707F8F7F7F90707F907F7F90707F907F80907F7F907F80907F7F907F80907F7F90
7F80907F7F907F80907F7F907F80907F7F907F80907F80907F80907F80907F809F7F80907F8090\
7F80907F80907F80907F80907F80907F80907F80907F80907F80908080907F8090
7F80907F80908080907F80907F80907F808F808F907F808F8080907F808F808F907F808F808F90\
7F808F808F907F808F808F907F808F808F907F808F808F8F7F808F808F8F7F808F
808F8F7F808F808F8F7F8F8F808F8F7F808F808F8F7F8F8F808F8F7F8F8F808F8F7F8F8F808F8F\
7F8080808F8F7F8F80808F8F7F8080808F8F7F8F80808F8F7F8080808F8F7F8F80
808F807F8080808F8F7F8F80808F807F8080808F807F8F80808F807F8080808F807F8080808F80\
7F807F808F807F8080808F807F807F808F807F807F808F807F807F808F807F807F
808F807F807F808F807F807F808F7F7F807F808F7F7F807F808F7F7F8070808F7F7F807080807F\
7F807080807F7F80708080707F80708080707F80707F80707F7F6F7F80707F7F6F
7F80707F7F6F7F80707F7F6F7F806F7F7F6F7F806F7F7F6F7F7F6F707F607F7F6F707F607F7F6F\
7070607F7F607070607F7F607070607F7F60707060707F6070705F707F60000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F302F2F3030303F302F403F304F30304F\
3F304F3F304F3F3F503F3F5040405F403F5F40406040405F4040604040604F4F60
4040604F4F604F4F60504F6F4F4F6050506F504F6F50507050506F50507050506F5F5F7050506F\
5F5F705050705F5F705F5F705F5F705F5F705F607F5F5F7060607F5F5F7F60607F
5F607F60607F60607F60608060607F606F8060607F6F6F8060607F6F6F80606F806F6F806F6F80\
6F6F806F6F806F708F6F6F806F708F6F6F8070708F6F6F8070708F6F708F70708F
6F708070708F6F708F70708F6F708F70708F70708F707F8F70708F707F8F70708F707F9070708F\
707F90707F8F7F7F90707F8F7F7F90707F8F7F7F90707F8F7F7F90707F907F8090
707F8F7F7F90707F907F8090707F907F80907F7F907F8090707F907F80907F7F907F80907F7F90\
7F80907F7F907F80907F7F8F7F80907F80907F80907F7F8F7F80907F808F7F8090
7F7F8F7F80907F808F7F80907F808F7F80907F808F7F80907F808F7F80907F808F7F80907F808F\
7F808F7F808F7F808F7F808F7F808F7F808F7F8F8F7F808F7F808F7F808F7F8F8F
7F808F7F8F8F7F808F808F8F7F80807F8F8F7F8080808F8F7F80807F8F8F7F8080808F8F7F8080\
7F8F8F7F8080808F8F7F80807F8F807F8080808F8F7F80807F8F807F8080808F80
7F80807F8F807F8080808F807F80807F8F807F8080808F807F807F7F8F807F807F808F807F807F\
7F8F807F807F808F807F807F7F80807F807F808F807F807F7F807F7F807F808F7F
7F807F7F807F7F807F808F7F7F807F7F807F7F807F80807F7F80707F807F7F80707F807F7F8070\
7F80707F80707F80707F7F707F80707F7F707F80707F7F6F7F80707F7F6F7F8070
707F6F7F7F6F7F7F6F7F7F6F707F6F7F7F6F707F6F7F7F6F707F607F7F6F707F607F7F6F707060\
707F60707060707F6070705F70706070705F70706070705F70705F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F302F2F3030303F30303F3F30403030403F3F4F\
3F304F3F3F503F3F5040405F40405F40406040405F4F4F604040604F4F604F4F60
4F4F6F4F4F6050506F4F4F6F50506F50506F50507050506F5F5F705050705F5F705050705F5F70\
5F5F705F5F7F5F5F705F607F5F5F7060607F5F5F7F60607F5F607F60607F60607F
60608060607F606F8060607F6F6F80606F806F6F80606F806F6F806F6F806F6F8F6F6F806F708F\
6F6F806F708F6F6F8070708F6F708F70708F6F708F70708F6F708F70708F6F708F
707F8F70708F707F8F70708F707F9070708F707F9070708F7F7F90707F8F7F7F90707F8F7F7F90\
707F907F7F90707F907F8090707F907F7F90707F907F80907F7F907F8090707F90
7F80907F7F907F80907F7F907F80907F80907F80907F7F907F80907F80907F80907F80907F8090\
7F80907F80907F80907F80907F80907F80907F80907F80907F80907F80907F8090
7F80907F80907F80907F808F7F80907F80907F80907F808F7F8F907F808F7F80907F808F808F90\
7F808F7F8F907F808F808F907F808F7F8F8F7F808F808F907F808F808F8F7F808F
808F8F7F808F808F8F7F808F808F8F7F808F808F8F7F808F808F8F7F8F8F808F8F7F8080808F8F\
7F8F80808F8F7F8080808F8F7F8F80808F8F7F8080808F8F7F8F80808F8F7F8080
808F8F7F8080808F807F8080808F807F8080808F807F8080808F807F8080808F807F8080808F80\
7F8080808F807F807F808F807F807F808F807F807F808F807F807F808F807F807F
808F807F807F808F7F7F807F808F7F7F807F808F7F7F807F808F7F7F807F808F7F7F8070808F7F\
7F807080807F7F807080807F7F80707F80707F80707F80707F80707F80707F7F6F
7F80707F7F6F7F80707F7F6F7F80707F7F6F7F7F6F707F6F7F7F6F7F7F6F7F7F6F707F607F7F6F\
707F607F7F607070607F7F60707060707F6070705F7F7F6070705F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F302F302F2F3F303040302F4030304F30304F\
3F3F503F304F3F3F503F3F5040405F403F5F4040604040604F4F604040604F4F60
4F4F604F4F6F4F4F6050506F4F4F6F50506F50506F50507050506F505F7050506F5F5F70505070\
5F5F705050705F5F705F5F705F5F7F5F5F705F607F5F5F7F60607F5F5F7F60607F
5F607F60608060607F60608060607F606F8060607F6F6F80606F806F6F80606F806F6F806F6F80\
6F708F6F6F806F708F6F6F806F708F6F6F806F708F6F708F70708F6F708F70708F
6F708F70708F6F708F70708F70708F707F9070708F707F9070708F707F9070708F707F90707F8F\
707F90707F8F707F90707F8F7F7F90707F8F707F90707F8F7F7F90707F8F7F7F90
707F8F7F8090707F8F7F8090707F907F8090707F8F7F8090707F907F8090707F907F80907F7F90\
7F8090707F8F7F80907F80907F8090707F8F7F80907F808F7F8090707F8F7F8090
70808F7F8090707F8F7F80907F808F7F809070808F7F80907F808F7F809070808F7F808F7F808F\
7F808F7F808F7F808F7F808F7F808F7F808F7F808F7F808F7F8F8F7F808F7F808F
7F808F7F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F8080\
7F8F8F7F80807F8F807F80807F8F8F7F80807F8F807F80807F8F807F80807F8F80
7F80807F8F807F80807F8F807F80807F8F807F807F7F80807F807F7F8F807F807F7F80807F807F\
7F8F807F807F7F80807F807F7F8F807F807F7F807F7F807F7F8F7F7F807F7F807F
7F807F7F807F7F807F7F807F7F807F7F807F7F80707F807F7F807F7F807F7F80707F807F7F8070\
7F807F7F80707F80707F80707F80707F7F707F80707F7F707F8070707F6F7F8070
7F7F6F7F8070707F6F7F7F6F707F6F7F7F6F707F607F7F6F707F607F7F6F7070607F7F6F707060\
7F7F60707060707F60707060707F6070705F70706070705F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F302F2F3030303F302F3F3F304F30304F3F3F4F\
3F304F3F3F503F3F5040405F403F5F4040604040604F4F604040604F4F6F4F4F60
4F4F6F4F4F6F50506F4F4F6F50506F50506F50507050506F505F705050705F5F705050705F5F70\
505F705F5F7F5F5F705F5F7F5F5F7060607F5F5F7F60607F5F5F7F60607F60607F
60608060607F606F8060607F606F806060806F6F80606F806F6F80606F806F708F6F6F806F708F\
6F6F806F708F6F708F6F708F6F708070708F6F708F70708F6F708F70708F6F708F
70708F6F708F707F9070708F707F9070708F707F90707F8F707F9070708F707F90707F8F707F90\
707F8F7F7F90707F8F707F90707F8F7F7F90707F907F7F90707F8F7F8090707F90
7F8090707F8F7F8090707F907F8090707F907F80907F7F907F8090707F907F80907F80907F8090\
7080907F80907F80907F80907080907F80907F80907F809070808F7F80907F8090
7F809070808F7F80907F808F7F809070808F7F80907F808F7F80907F808F7F80907F808F7F8090\
7F808F7F80907F808F7F808F7F808F7F8F8F7F808F7F808F7F808F7F8F8F7F808F
7F8F8F7F808F7F8F8F7F808F7F8F8F7F80807F8F8F7F808F7F8F8F7F80807F8F8F7F80807F8F8F\
7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F807F80807F8F8F7F8080
7F8F807F80807F8F807F80807F8F807F80807F8F807F80807F8F807F807F7F8F807F80807F8F80\
7F807F7F8F807F807F7F8F807F807F7F8F807F807F7F8F807F807F7F8F807F807F
7F807F7F807F7F8F7F7F807F7F807F7F807F7F8F7F7F807F7F807F7F8070808F7F7F80707F807F\
7F80707F807F7F80707F80707F80707F80707F80707F80707F7F707F80707F7F6F
7F80707F7F6F7F80707F7F6F7F806F707F6F7F806F707F6F7F7F6F707F607F7F6F707F607F7F6F\
7070607F7F6F707060707F60707060707F60707060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F30302F3F2F2F3F303040302F403F304F30304F\
3F3F503F3F503F3F503F3F5040405F40405F4040604040604F4F60404F604F4F6F
4F4F604F4F6F4F4F6F50506F4F4F6F50506F50506F50507050506F5050705050705F5F70505070\
5F5F70505F705F5F7F505F705F5F7F5F5F705F607F5F5F705F607F5F607F60607F
5F607F60607F60607F606F8060607F606F806060806F6F80606F806F6F80606F806F6F8F606F80\
6F708F6F6F806F708F6F6F806F708F6F70806F708F6F70806F708F6F708F70708F
6F708F70708F6F708F707F8F6F708F707F8F70708F707F906F708F707F8F70708F707F9070708F\
707F8F707F8F707F9070708F707F90707F8F707F90707F8F707F90707F8F707F90
707F8F707F90707F8F7F8090707F8F707F90707F8F7F8090707F8F7F8090707F8F7F8090707F8F\
7F8090707F8F7F8090707F8F7F8090707F8F7F8090707F8F7F8090707F8F7F8090
707F8F7F8090707F8F7F8090707F8F7F808F707F8F7F8090707F8F7F808F70808F7F808F707F8F\
7F808F70808F7F808F70808F7F808F70808F7F808F7080807F808F70808F7F808F
7080807F808F7F80807F808F7080807F808F7F80807F8F8F7080807F808F7F80807F8F8F708080\
7F80807F80807F8F807080807F80807F80807F80807080807F80807F80807F8080
7080807F80807F80807F808070807F7F80807F807F7F808070807F7F80807F807F7F808070807F\
7F80807F807F7F808070807F7F807F7F807F7F807F70807F7F807F7F807F7F807F
70807F7F807F7F807F7F807F7080707F807F7F80707F807F7F80707F807F7F80707F807F7F8070\
7F80707F80707F8070707F707F80707F7F707F8070707F6F7F8070707F6F7F8070
707F6F7F7F6F707F6F7F7F6F707F6F7F7F6F707F6F7F7F6F707F60707F6F707F60707F6F707060\
707F60707060707F6070706070706070705F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F302F2F303030403030403F304F30304F3F3F50\
3F3F4F3F3F503F3F5040405F40405F4040604040604F4F6F4F4F604F4F6F4F4F60
50506F4F4F6F50506F4F506F50507050506F50507050506F5F5F705050705F5F705050705F5F7F\
5F5F705F5F7F5F5F705F607F5F5F7F5F607F5F5F7F60607F5F607F60607F5F607F
606F8060607F606F8060607F6F6F80606F806F6F80606F806F6F8F6F6F806F708F6F6F806F708F\
6F708F6F708F6F708F70708F6F708F70708F6F708F707F8F6F708F707F8F6F708F
707F9070708F707F9070708F707F9070708F707F9070708F707F90707F8F707F90707F8F707F90\
707F8F707F90707F8F707F90707F8F707F90707F8F7F8090707F8F7F7F90707F8F
7F8090707F907F8090707F8F7F8090707F907F8090707F8F7F80907080907F8090707F907F8090\
7080907F809070808F7F80907080907F809070808F7F809070808F7F809070808F
7F809070808F7F809070808F7F809070808F7F809070808F7F809070808F7F809070808F7F8090\
7F808F7F808F70808F7F808F7F808F7F808F70808F7F8F8F7F808F7F808F7F808F
7F8F8F7F808F7F8F8F7F80807F8F8F7F808F7F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F\
7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F807F8080
7F8F807F80807F8F807F80807F8F807F80807F8F807F80807F8F807F80807F8F807F807F7F8F80\
7F807F7F80807F807F7F8F807F807F7F80807F807F7F8F807F807F7F807F7F807F
7F8F7F7F807F7F807F7F807F7F8F7F7F807F7F807F7F807F7F8F7F7F807F7F807F7F80707F807F\
7F80707F807F7F80707F807F7F80707F80707F80707F80707F80707F80707F7F6F
7F80707F7F6F7F8070707F6F7F8070707F6F7F806F707F6F7F806F707F6F7F7F6F707F607F7F6F\
707F60707F6F707060707F60707060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F30302F3F2F2F3F303040302F403F304F30304F\
3F3F503F3F4F3F3F503F3F5040405F40405F4040604040604F4F6F404F604F4F6F
4F4F6F50506F4F4F6F50506F4F506F5050704F506F50507050506F505F70505070505F70505070\
5F5F70505F705F5F7F5F5F705F607F5F5F705F607F5F5F7F60607F5F5F7F60607F
5F607F6060805F607F606F8060607F606F80606080606F80606F806F6F80606F806F6F8F6F6F80\
6F708F6F6F806F708F6F708F6F708F6F70806F708F6F708F70708F6F708F70708F
6F708F707F8F6F708F70708F6F708F707F8F6F708F707F8F6F708F707F8F6F708F707F8F70708F\
707F906F708F707F8F707F8F707F90707F8F707F8F707F8F707F90707F8F707F90
707F8F707F90707F8F707F90707F8F708090707F8F708090707F8F7F8090707F8F708090707F8F\
7F8090707F8F708090707F8F7F8090707F8F708090707F8F7F8090707F8F70808F
707F8F7F8090707F8F70808F707F8F7F808F707F8F70808F707F8F7F808F707F8F70808F70808F\
7F808F707F8F7F808F70808F7F808F7080807F808F70808F7F808F7080807F808F
7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F7080807F8080708080\
7F808F7080807F80807080807F80807080807F80807080807F80807080807F8080
7080807F808070807F7F808070807F7F808070807F7F808070807F7F808070807F7F808070807F\
7F808070807F7F807F70807F7F807F70807F7F807F70807F7F807F70807F7F807F
70807F7F807F7080707F807F70807F7F807F7080707F807F7080707F807F7080707F807F7F8070\
7F807F707F707F80707F80707F8070707F707F8070707F707F8070707F6F7F8070
707F6F7F8070707F6F7F7F6F707F6F7F7F6F707F6F707F6F707F6F7F7F6F707F60707F6F707F60\
707F6F707060707F60707060707F60000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000030303F2F2F3F303040302F403F304F30304F3F3F4F\
3F3F4F403F503F3F5040405F40405F404F604040604F4F604F4F604F4F6F4F4F6F
4F506F4F4F6F5050704F506F5050704F506F505070505070505F705050705F5F70505F705F5F70\
505F705F5F7F5F5F705F607F5F5F705F607F5F5F7F5F607F5F607F6060805F607F
60608060607F606F80606080606F806060806F6F80606F806F6F80606F806F708F6F6F806F708F\
6F6F806F708F6F708F6F708F6F708F70708F6F708F70708F6F708F707F8F6F708F
70708F6F708F707F8F6F708F707F8F6F708F707F906F708F707F8F6F708F707F90707F8F707F8F\
70708F707F90707F8F707F90707F8F707F90707F8F707F90707F8F707F90707F8F
707F90707F8F708090707F8F708090707F8F7F8090707F8F708090707F8F7F8090707F8F7F8090\
707F8F7F809070808F7F8090707F8F7F809070808F7F8090707F8F7F809070808F
7F8090707F8F7F809070808F7F8090707F8F7F809070808F7F808F70808F7F809070808F7F808F\
70808F7F808F70808F7F808F70808F7F808F70808F7F808F70808F7F808F70808F
7F808F7080807F808F7080807F808F7080807F8F8F7080807F808F7080807F8F8F7080807F808F\
7080807F8F8F7080807F80807080807F8F807080807F80807080807F8F80708080
7F80807080807F80807080807F808070807F7F80807080807F808070807F7F808070807F7F8080\
70807F7F808070807F7F808070807F7F808070807F7F807F70807F7F807F70807F
7F807F70807F7F807F70807F7F807F70807F7F807F7F807F7F807F7080707F807F7F80707F807F\
7080707F807F7F80707F80707080707F80707F80707F8070707F707F8070707F70
7F8070707F6F7F8070707F6F7F806F707F6F7F806F707F6F7F7F6F707F607F7F6F707F60707F6F\
707F60707F6F707F60707F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F30302F3F2F2F3F3030403030403F304F30304F\
3F3F503F3F50403F503F3F5040405F40405F404F604040604F4F6F4F4F604F4F6F
4F4F6F4F506F4F4F6F50506F4F4F6F5050704F506F50507050506F505070505070505F70505070\
5F5F70505F705F5F7F505F705F5F7F5F5F705F607F5F5F7F5F607F5F607F60607F
5F607F6060805F607F606F8060607F606F80606F80606F80606F806F6F80606F806F708F606F80\
6F708F6F6F806F708F6F6F806F708F6F708F6F708F6F70806F708F6F708F6F708F
6F70806F708F6F708F70708F6F708F6F708F6F708F707F8F6F708F707F8F6F708F707F8F6F708F\
707F8F6F708F707F8F6F708F707F8F6F7F8F707F8F6F708F707F8F707F8F707F8F
6F7F8F707F8F707F8F707F90707F8F707F8F707F8F708090707F8F707F8F707F8F708090707F8F\
70808F707F8F708090707F8F70808F707F8F70808F707F8F70808F707F8F70808F
707F8F70808F707F8F70808F707F8F70808F707F8F70808F707F8F70808F707F8F70808F707F8F\
70808F707F8F70808F707F8070808F707F8070808F707F8070808F70808070808F
70808070808F7080807F808F7080807080807080807F808F7080807080807080807F8080708080\
7080807080807F808070808070808070808070808070807F708080708080708080
70807F70808070807F70808070807F70808070807F70808070807F70808070807F708080707F7F\
70807F70807F70807F707F7F70807F70807F70807F707F7F70807F70807F70807F
707F7070807F7080707F807F707F707F807F7080707F807F707F707F807F707F707F807F707F70\
7F8070707F707F8070707F707F8070707F707F8070707F6F7F8070707F6F7F8070
707F6F707F6F707F6F7F7F6F707F6F707F6F707F6F707F6F707F60707F6F707F60707F6F707060\
707F60707060707F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F3F2F2F3F3030403030403F304F30304F3F3F50\
3F3F50403F503F3F5040405F40405F404F604040604F4F6F4F4F6F4F506F4F4F6F
50506F4F506F5050704F506F50507050506F505070505070505F70505070505F70505F705F5F7F\
505F705F5F7F5F5F705F607F5F5F7F5F607F5F5F7F60607F5F607F6060805F607F
606F8060607F606F80606080606F80606F806F6F80606F806F708F606F806F708F6F6F806F708F\
6F708F6F708F6F708F6F708F6F708F6F708F6F708F70708F6F708F6F708F6F708F
707F8F6F708F707F8F6F708F707F8F6F708F707F8F6F708F707F8F6F708F707F8F6F708F707F90\
6F7F8F707F8F6F7F8F707F90707F8F707F8F707F8F707F90707F8F707F90707F8F
708090707F8F707F90707F8F708090707F8F708090707F8F708090707F8F708090707F8F7F8090\
707F8F708090707F8F7F809070808F708090707F8F7F809070808F708090707F8F
7F809070808F70808F707F8F7F809070808F70808F70808F7F808F70808F70808F70808F7F808F\
70808F70808F70808F7F808F70808F70808F70808F7F808F70808F7F808F708080
7F808F70808F7F808F7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F\
7080807F80807080807F808F7080807F80807080807F80807080807F8080708080
7F80807080807F80807080807F80807080807F808070807F7F808070807F7F808070807F7F8080\
70807F7F808070807F7F808070807F7F807F70807F7F807F70807F7F807F70807F
7F807F70807F7F807F70807F7F807F70807F7F807F7080707F807F7080707F807F7080707F807F\
7080707F807F7080707F807F7080707F80707080707F80707080707F8070707F70
7F8070707F6F7F8070707F6F7F8070707F6F7F806F707F6F7F806F707F6F707F6F707F60707F6F\
707F60707F6F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F30302F3F302F403030403030403F304F30304F\
3F3F503F3F503F40503F3F5040405F40405F404F60404F604F4F6F404F604F4F6F
4F4F6F4F506F4F4F6F50506F4F506F5050704F506F505070505070505F70505070505F70505F70\
5F5F7F505F705F5F7F505F705F607F5F5F705F607F5F5F7F5F607F5F607F60607F
5F607F606F805F607F606F80606080606F80606F80606F80606F806F6F8F606F806F708F606F80\
6F708F606F806F708F6F70806F708F6F6F806F708F6F70806F708F6F70806F708F
6F708F6F708F6F70806F708F6F708F6F7F8F6F70806F708F6F708F707F8F6F708F6F7F8F6F708F\
707F8F6F708F707F8F6F708F707F8F6F708F707F8F6F7F8F707F8F6F7F8F707F8F
6F7F8F707F8F6F7F8F707F8F707F8F707F8F6F7F8F707F8F707F8F708090707F8F70808F707F8F\
708090707F8F70808F707F8F70808F707F8F70808F707F8F70808F707F8F70808F
707F8F70808F707F8F70808F707F8F70808F707F8F70808F707F8F70808F707F8070808F707F8F\
70808F707F8070808F707F8070808F707F8070808F707F8070808F707F8070808F
70808070808F707F8070808F70808070808F708080708080708080708080708080708080708080\
708080708080708080708080708080707F80708080708080708080707F7F708080
70807F708080707F7F70808070807F708080707F7F70808070807F708080707F7F70807F70807F\
70807F707F7F70807F707F7F70807F707F7F70807F707F7F70807F707F7070807F
707F7F70807F707F7070807F707F7070807F707F7070807F707F707F807F707F70708070707F70\
7F8070707F70708070707F707F8070707F70708070707F707F8070707F6F708070
707F6F7F8070707F6F707F6F707F6F707F6F707F6F707F6F707F6F707F6F707F60707F6F707F60\
707F6F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000030303F302F3F3030403030403F3F4F30304F3F3F50\
3F3F4F4040503F3F5040405F40405F404F604040604F4F6F404F604F506F4F4F6F
4F506F4F4F6F5050704F506F5050704F506F505F70505070505F705050705F5F70505F705F5F7F\
505F705F607F5F5F705F607F5F5F7F5F607F5F607F5F607F5F607F6060805F607F
606F805F607F606F80606F80606F80606F806F6F80606F806F6F8F606F806F708F6F6F806F708F\
6F6F806F708F6F708F6F708F6F708F6F708F6F708F6F708F6F708F6F708F6F708F
6F708F6F708F6F7F8F6F708F6F708F6F708F707F8F6F708F6F7F8F6F708F707F8F6F708F707F8F\
6F708F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F
707F8F6F7F8F707F90707F8F707F8F707F8F708090707F8F708090707F8F708090707F8F708090\
707F8F708090707F8F70808F707F8F708090707F8F70808F707F8F70808F707F8F
70808F707F8F70808F707F8F70808F707F8F70808F70808F70808F707F8F70808F70808F70808F\
707F8F70808F70808F70808F707F8070808F70808F70808F70808070808F708080
70808F70808070808F70808070808F70808070808F70808070808070808070808F708080708080\
708080708080708080708080708080708080708080708080708080708080708080
70808070807F70808070808070808070807F70808070807F70808070807F70808070807F708080\
70807F70808070807F70807F707F7F70807F70807F70807F707F7F70807F70807F
70807F707F7F70807F70807F70807F707F707F807F70807070807F707F707F807F7080707F807F\
707F707F807F707F707F8070707F707F8070707F707F8070707F707F8070707F70
7F8070707F6F7F8070707F6F708070707F6F7F806F707F6F707F6F707F6F707F6F707F6F707F6F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F3F3030402F2F3F3030403030403F3F4F30304F\
3F3F503F3F5040405F3F3F5040405F40405F404F604040604F4F6F404F604F506F
4F4F6F4F506F4F4F6F5050704F506F5050704F506F505F704F506F505F70505070505F70505070\
505F70505F705F5F7F505F705F607F5F5F7F5F607F5F5F7F5F607F5F607F606080
5F607F6060805F607F606F805F607F606F80606F80606F80606F80606F80606F806F708F606F80\
6F708F606F806F708F606F806F708F6F70806F708F606F806F708F6F70806F708F
6F70806F708F6F70806F708F6F70806F708F6F70806F708F6F70806F708F6F70806F7F8F6F7080\
6F7F8F6F708F6F7F8F6F70806F7F8F6F708F707F8F6F70806F7F8F6F7F8F707F8F
6F708F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F70808F6F7F8F\
707F8F6F7F8F70808F6F7F8F707F8F6F7F8F70808F6F7F8F70808F6F7F8F70808F
6F7F8070808F6F7F8F70808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F80\
70808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F
6F7F807080806F7F807080806F7F80708080707F807080806F7F807080806F7F807080806F7F80\
7080806F7F807080806F7F7F7080806F7F807080806F7F7F7080806F7F7F708080
6F7F7F7080806F7F7F7080806F7F7F7080806F7F7F7080806F7F7F70807F6F7F7F70807F6F7F7F\
70807F6F7F7F70807F6F7F7F70807F6F7F7F70807F6F7F7070807F6F7F7070807F
6F7F7070807F707F7070807F707F7070807F707F7070807F707F70708070707F70708070707F70\
708070707F70708070707F70708070707F70708070707F6F707F70707F6F708070
707F6F707F70707F6F707F70707F6F707F6F707F6F707F6F707F60707F6F707F60707F6F6F7060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F3030303F2F2F3F3030403030403F3F4F30304F3F3F50\
3F3F504040503F3F5040405F40405F404F604040604F4F6F4F4F6F4F506F4F4F6F
5050704F506F5050704F506F505F70505070505F70505070505F70505F70505F7F505F705F5F7F\
505F705F607F505F7F5F607F5F607F5F607F5F607F6060805F607F6060805F607F
606F80606080606F80606F80606F80606F806F6F8F606F806F708F606F806F708F606F806F708F\
6F708F6F708F6F70806F708F6F708F6F708F6F708F6F708F6F708F6F708F6F708F
6F7F8F6F708F6F708F6F708F6F7F8F6F708F6F7F8F6F708F707F8F6F708F6F7F8F6F708F707F8F\
6F708F6F7F8F6F708F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F
707F8F6F7F8F707F8F6F7F8F70808F6F7F8F707F8F6F7F8F708090707F8F70808F707F8F708090\
707F8F70808F707F8F708090707F8F70808F707F8F70808F707F8F70808F707F8F
70808F707F8F70808F707F8F70808F707F8F70808F707F8F70808F70808F70808F707F8070808F\
70808F70808F707F8070808F70808070808F707F8070808F70808070808F707F80
70808F70808070808F70808070808F70808070808F70808070808F708080708080708080708080\
70808070808070808070808070808070808070808070808070808070808070807F
7080807080807080806F7F7F70808070807F7080806F7F7F70808070807F7080806F7F7F708080\
70807F70807F6F7F7F70807F70807F70807F6F7F7F70807F70807F70807F707F7F
70807F70807F70807F707F7070807F70807070807F707F7070807F70807070807F707F7070807F\
707F70708070707F707F8070707F70708070707F707F8070707F70708070707F6F
7F8070707F6F708070707F6F7F8070707F6F70806F707F6F70806F707F6F707F6F707F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F3F2F2F3F303040302F4030304F3030403F3F4F30304F\
3F3F503F3F503F405F3F3F5F40405F40405F404F604040604F4F6F404F604F4F6F
4F4F6F4F50704F4F6F4F50704F506F5050704F506F505F70505070505F70505070505F70505F70\
5F5F7F505F705F5F7F505F7F5F607F5F5F7F5F607F5F607F5F60805F607F606080
5F607F606F805F607F606F80606F80606F80606F80606F80606F806F6F8F606F806F708F606F80\
6F708F606F806F708F606F806F708F606F806F708F6F70806F708F6070806F708F
6F70806F708F6070806F708F6F70806F708F6F70806F708F6F70806F7F8F6F70806F7F8F6F7080\
6F7F8F6F70806F7F8F6F708F6F7F8F6F70806F7F8F6F708F6F7F8F6F70806F7F8F
6F7F8F707F8F6F7F806F7F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F\
70808F6F7F8F707F8F6F7F8F70808F6F7F8F707F8F6F7F8F70808F6F7F8070808F
6F7F8F70808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F80\
70808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F80708080
6F7F8070808F6F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80\
7080806F7F7F7080806F7F807080806F7F7F7080806F7F7F7080806F7F7F708080
6F7F7F7080806F7F7F7080806F7F7F7080806F7F7F70807F6F7F7F70807F6F7F7F70807F6F7F7F\
70807F6F7F7F70807F6F7F7F70807F6F7F7070807F6F7F7F70807F6F7F7070807F
6F7F7070807F6F7F7070807F6F7F7070807F6F7F70708070707F7070807F6F7F70708070707F70\
708070707F70708070707F70708070707F6F707F70707F6F708070707F6F707F70
707F6F708070707F6F707F6F707F6F707F6F707F6F707F6F707F6F707F6F6F7F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F30303040302F403030403030403F3F4F30304F3F3F50\
3F3F5040405F3F3F5040405F3F405F404F604040604F4F6F404F604F506F4F4F6F
4F506F4F506F5050704F506F5050704F5070505F70505070505F705050705F5F7F505F705F5F7F\
505F705F607F5F5F7F5F607F5F5F7F5F607F5F607F5F60805F607F606F805F607F
606F805F6080606F80606F80606F80606F80606F8F606F80606F8F606F806F708F606F806F708F\
606F806F708F6070806F708F6070806F708F6F708F6F708F6F70806F708F6F708F
6F708F6F70806F7F8F6F708F6F708F6F708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F708F6F7F8F\
6F708F6F7F8F6F708F6F7F8F6F708F707F8F6F7F8F6F7F8F6F7F8F707F8F6F7F8F
6F7F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F707F8F6F7F8F70808F6F7F8F70808F\
6F7F8F70808F6F7F8F70808F6F7F8F70808F6F7F8F70808F6F7F8F70808F6F7F8F
70808F6F7F8F70808F6F7F8F70808F6F7F8070808F6F7F8F70808F6F7F8070808F6F7F8070808F\
6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F80
7080806F7F8070808F6F7F807080806F7F8070808F6F7F807080806F7F807080806F7F80708080\
6F7F807080806F7F807080806F7F807080806F7F807080806F7F7F7080806F7F7F
7080806F7F7F7080806F7F7F7080806F7F7F7080806F7F7F7080806F7F7F7080806F7F7F70807F\
6F7F7F70807F6F7F7F70807F6F7F7F70807F6F7F7F70807F6F7F7F70807F6F7F7F
70807F6F7F7070807F707F7070807F6F7F7070807F707F7070807F707F7070807F707F70708070\
707F70708070707F70708070707F70708070707F70708070707F6F708070707F6F
708070707F6F708070707F6F707F70707F6F708070707F6F707F6F707F6F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F3F2F2F3F3030402F2F403030403030403F3F4F30304F\
3F3F503F3F503F405F3F3F5040405F3F405F404F60404060404F60404F604F4F6F
404F6F4F506F4F506F5050704F506F5050704F5070505F704F5070505F70505F70505F7F505F70\
505F7F505F705F607F505F705F607F5F5F7F5F607F5F607F5F607F5F607F606080
5F607F606F805F6080606F805F6080606F80606F80606F80606F80606F80606F80607080606F80\
607080606F806F708F606F806F708F6070806F708F606F806F708F6070806F708F
6070806F708F6070806F708F6070806F708F6F70806F708F6070806F708F6F70806F7F8F607080\
6F7F8F6F70806F7F8F6F70806F7F8F6F70806F7F8F6F70806F7F8F6F70806F7F8F
6F70806F7F8F6F7F806F7F8F6F70806F7F8F6F7F806F7F8F6F7F806F7F8F6F7F8F707F8F6F7F80\
6F7F8F6F7F8F707F8F6F7F806F7F8F6F7F80707F8F6F7F806F7F8F6F7F8070808F
6F7F806F7F8F6F7F8070808F6F7F806F7F8F6F7F8070808F6F7F806F7F8F6F7F8070808F6F7F80\
6F7F806F7F806F80806F7F806F7F806F7F806F80806F7F806F7F806F7F806F8080
6F7F806F7F806F7F806F80806F7F7F6F7F806F7F806F80806F7F7F6F7F806F7F806F80806F7F7F\
6F80806F7F7F6F80806F7F7F6F80806F7F7F6F80806F7F7F6F7F806F7F7F6F8080
6F7F7F6F7F7F6F7F7F6F80806F7F7F6F7F7F6F7F7F6F807F6F7F7F6F7F7F6F7F7F6F807F6F7F7F\
6F7F7F6F7F7F6F807F6F7F706F7F7F6F7F706F807F6F7F706F7F7F6F7F7070807F
6F7F706F7F7F6F7F7070807F6F7F70707F706F7F707080706F7F70707F706F7F70707F706F7F70\
707F706F7F70707F706F7F6F707F706F7F6F707F706F7F6F707F706F7F6F707F70
6F7F6F707F6F6F7F6F707F6F6F7F6F707F6F6F7F6F707F6F6F7F6F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F3F303040302F4030304F3030403F3F4F30304F3F3F50\
3F3F503F405F3F3F5040405F40405F404F604040604F4F6F404F604F4F6F4F4F6F
4F50704F506F5050704F506F505F70505070505F70505070505F7F505F70505F7F505F705F607F\
505F7F5F607F505F7F5F607F5F607F5F607F5F607F6060805F607F606F805F607F
606F80606F80606F80606F80606F80606F80606F80606F806F708F606F8060708F606F806F708F\
606F806F708F606F806F708F60708F6F708F6070806F708F6F708F6F708F6F708F
6F7F8F6F708F6F708F6F708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F708F6F7F8F\
6F708F6F7F8F6F708F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F
707F8F6F7F8F6F7F8F6F7F8F707F8F6F7F8F707F8F6F7F8F70808F6F7F8F707F8F6F7F8F70808F\
6F7F8F70808F6F7F8F70808F6F7F8F70808F6F7F8F70808F6F7F8F70808F6F7F80
70808F6F7F8070808F6F7F8070808F6F7F806F808F6F7F8070808F6F7F806F808F6F7F8070808F\
6F7F806F80806F7F8070808F6F7F806F80806F7F806F80806F7F806F80806F7F80
6F80806F7F806F80806F7F806F80806F7F806F80806F7F806F80806F7F806F80806F7F806F8080\
6F7F806F80806F7F7F7080806F7F806F80806F7F7F7080806F7F7F6F80806F7F7F
7080806F7F7F6F80806F7F7F7080806F7F7F6F80806F7F7F7080806F7F7F6F807F6F7F7F708080\
6F7F7F6F807F6F7F7F70807F6F7F7F70807F6F7F7F70807F6F7F7F70807F6F7F70
70807F6F7F7070807F6F7F7070807F6F7F7070807F6F7F7070807F707F7070807F6F7F7070807F\
707F707080706F7F70708070707F707080706F7F70708070707F70707F706F7F6F
708070707F6F707F70707F6F708070707F6F707F6F707F6F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F3F2F2F3F30304030304030304F3030403F3F4F30304F\
3F3F503F3F503F405F3F3F5F4040603F405F404F60404060404F6F404F604F4F6F
4F4F6F4F506F4F4F6F4F50704F506F505F704F5070505F70505070505F70505F70505F7F505F70\
5F5F7F505F705F607F505F7F5F607F5F5F7F5F607F5F607F5F60805F607F606080
5F607F606F805F607F606F805F6F80606F805F6F80606F80606F80606F80606F80606F80606F80\
60708F606F8060708F606F806F708F606F806F708F6070806F708F6070806F708F
6070806F708F6070806F708F6070806F708F6070806F708F6070806F7F8F6070806F7F8F6F7080\
6F7F8F6070806F7F8F6F70806F7F8F6F70806F7F8F6F70806F7F8F6F70806F7F8F
6F7F806F7F8F6F70806F7F8F6F7F806F7F8F6F7F806F7F8F6F7F8F6F7F8F6F7F806F7F8F6F7F80\
6F7F8F6F7F806F7F8F6F7F806F7F8F6F7F806F7F8F6F7F806F7F8F6F7F806F7F80
6F7F806F7F8F6F7F806F7F806F7F806F7F80607F806F7F806F7F806F7F80607F806F7F806F7F80\
6F7F80607F806F7F80607F806F7F80607F7F6F7F80607F806F7F80607F7F6F7F80
607F7F6F7F80607F7F6F7F80607F7F6F7F80607F7F6F7F80607F7F6F7F80607F7F6F7F80607F7F\
6F7F80607F7F6F7F806F7F7F6F7F80607F7F6F7F806F7F7F6F7F806F7F7F6F7F7F
6F7F7F6F80806F7F7F6F7F7F6F7F7F6F807F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F\
6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F6F7F706F7F7F6F7F706F7F7F6F7F706F7F7F
6F7F706F7F7F6F7F706F7F7F6F7F70707F7F6F7F706F7F706F7F70707F706F7F70707F706F7F70\
707F706F7F70707F706F7F70707F706F7F6F707F706F7F6F707F706F7F6F707F70
6F7F6F707F706F7F6F707F6F6F7F6F707F6F6F7F6F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F3F303040302F403030403030403F3F4F30304F3F3F50\
3F3F503F405F3F3F5F40405F3F405F404F60404060404F60404F604F4F6F404F6F
4F506F4F4F6F5050704F506F505F704F5070505F70505F70505F70505F70505F7F505F705F5F7F\
505F705F607F5F5F7F5F607F5F607F5F60805F607F5F60805F607F606F805F6080
606F805F6080606F80606F80606F80606F80606F80606F80606F80606F8060708F606F8060708F\
606F806F708F606F806F708F606F806F708F6070806F708F6070806F708F60708F
6F708F6070806F7F8F6F708F6F708F6070806F7F8F6F708F6F7F8F6F70806F7F8F6F708F6F7F8F\
6F70806F7F8F6F708F6F7F8F6F70806F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F
6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F806F7F8F6F7F8F6F7F8F\
6F7F806F7F8F6F7F806F7F8F6F7F806F7F8F6F7F806F7F8F6F7F806F7F8F6F7F80
6F7F80607F806F7F8F6F7F806F7F80607F806F7F806F7F806F7F80607F806F7F80607F806F7F80\
607F806F7F80607F806F7F80607F7F6F7F80607F806F7F80607F7F6F7F80607F7F
6F7F80607F7F6F7F80607F7F6F7F80607F7F6F7F80607F7F6F7F80607F7F6F7F80607F7F6F7F80\
607F7F6F7F80607F7F6F7F80607F7F6F7F806F7F7F6F7F80607F7F6F7F806F7F7F
6F7F7F6F7F7F6F80806F7F7F6F7F7F6F7F7F6F80806F7F7F6F7F7F6F7F7F6F807F6F7F7F6F7F7F\
6F7F7F6F807F6F7F7F6F7F7F6F7F7F6F807F6F7F7F6F7F7F6F7F706F807F6F7F70
6F7F7F6F7F7070807F6F7F706F7F7F6F7F7070807F6F7F70707F706F7F707080706F7F70707F70\
6F7F707080706F7F70707F706F7F6F707F706F7F70707F706F7F6F707F706F7F6F
707F706F7F6F707F706F7F6F707F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F3F3030402F2F4030304F30304F303F4F30304F\
3F3F503F3F503F405F3F3F5F3F405F3F405F404F603F4060404F60404F604F4F6F
404F604F506F4F4F6F4F50704F506F5050704F5070505F704F5070505F70505F70505F7F505F70\
505F7F505F705F607F505F7F5F607F5F607F5F607F5F607F5F60805F607F606F80
5F607F5F6F805F607F606F805F607F606F805F6F80606F805F6F80606F80606F80606F80606F80\
606F80606F80607080606F80607080606F806F708F606F806F708F6070806F708F
6070806F708F6070806F708F6070806F708F6070806F708F6070806F708F6070806F7F8F607080\
6F7F8F6070806F7F8F6070806F7F8F6070806F7F8F6070806F7F8F6F70806F7F8F
6070806F7F8F6F7F806F7F8F6F70806F7F8F6F7F806F7F8F6070806F7F8F607F806F7F8F607080\
6F7F80607F806F7F806070806F7F806070806F7F806070806F7F806070806F7F80
607080607F806070806F7F8060707F607F806070806F7F8060707F607F8060707F607F8060707F\
607F8060707F607F8060707F607F8060707F607F8060707F607F7F60707F607F80
60707F607F7F60707F607F7F60707F607F7F60707F607F7F60707F607F7F60707F607F7F60707F\
607F7F60707F607F7F60707F607F7F60707F6F7F7F60707F607F7F607F7F6F7F7F
60707F607F7F607F7F6F7F7F60707F6F7F7F607F7F6F7F7F607F706F7F7F607F7F6F7F7F607F70\
6F7F7F607F706F7F7F607F706F7F7F607F706F7F7F607F706F7F7F6F7F706F7F7F
6F7F706F7F706F7F706F7F706F7F706F7F706F7F706F7F706F7F706F7F706F7F706F7F706F7F6F\
6F7F706F7F6F6F7F706F7F6F6F7F706F7F6F6F7F706F7F6F6F7F6F6F7F6F6F7F70
6F706F6F7F6F6F7F6F6F7F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003030402F2F4030304F30304F303F4F30304F3F3F50\
3F3F503F405F3F3F5F4040603F405F404060404060404F6F404F604F4F6F404F6F
4F506F4F506F4F50704F506F505F704F5070505F70505F70505F7F505F70505F7F505F705F607F\
505F7F5F607F505F7F5F60805F607F5F60805F607F606F805F6080606F805F607F
606F805F6F80606F805F6F80606F80606F80606F80606F80607080606F80607080606F8060708F\
606F8060708F606F806F708F6070806F708F6070806F708F60708F6F708F607080
6F7F8F60708F6F708F6070806F7F8F6F708F6F7F8F6070806F7F8F6F708F6F7F8F6F70806F7F8F\
6F708F6F7F8F6F708F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F
6F7F8F6F7F8F6F7F8F6F7F806F7F8F6F7F8F6F7F8F6F7F806F7F8F6F7F806F7F8F607F806F7F8F\
6F7F806F7F8F607F806F7F8F607F806F7F806070806F7F80607F806F7F80607080
6F7F80607F806F7F806070806F7F80607F80607F8060707F6F7F80607080607F8060707F6F7F80\
60707F607F8060707F607F8060707F607F8060707F607F8060707F607F8060707F
607F8060707F607F7F60707F607F8060707F607F7F60707F607F8060707F607F7F60707F607F80\
60707F607F7F60707F6F7F7F607F7F607F7F60707F6F7F7F607F7F607F7F607F7F
6F7F7F607F7F6F7F7F607F7F6F7F7F607F7F6F7F7F607F7F6F7F7F607F7F6F7F7F607F7F6F7F7F\
607F7F6F7F7F607F7F6F807F6F7F7F6F7F7F6F7F706F807F6F7F7F6F7F7F6F7F70
6F807F6F7F706F7F7F6F7F706F807F6F7F706F7F7F6F7F7070807F6F7F706F7F706F7F70707F70\
6F7F706F7F706F7F70707F706F7F706F7F706F7F6F707F706F7F6F707F706F7F6F
707F706F7F6F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F3F3030402F2F4030304F30304F303F4F30304F\
3F3F50303F503F405F3F3F5F4040603F405F404060404060404F60404F60404F6F
404F6F4F506F4F4F6F4F50704F506F505F704F5070505F70505F70505F7F505F70505F7F505F70\
5F607F505F7F5F607F505F7F5F607F50607F5F607F5F607F5F60805F607F5F6080
5F607F606F805F607F5F6F805F6080606F805F607F606F805F6F80606F805F6F80606F80606F80\
607080606F80607080606F8060708F606F8060708F6070806F708F6070806F708F
6070806F708F6070806F708F6070806F708F6070806F708F6070806F7F8F6070806F7F8F607080\
6F7F8F6070806F7F8F6070806F7F8F6070806F7F8F6070806F7F8F6070806F7F8F
6070806F7F8F6070806F7F8F6070806F7F8F6070806F7F806070806F7F806070806F7F80607080\
6F7F80607080607F80607080607F8060707F607F80607080607F8060707F607080
60707F607F8060707F60708060707F607F805F707F60708060707F607F805F707F60707F60707F\
607F805F707F60707F5F707F607F7F5F707F60707F5F707F607F7F5F707F60707F
5F707F607F7F5F707F60707F5F707F607F7F5F707F60707F5F707F60707F5F707060707F5F707F\
607F7F5F707060707F60707F607F7F5F7070607F7F607070607F7F5F7070607F7F
607070607F7F607070607F7F607070607F7F607070607F7F607070607F7F607070607F7F607070\
6F7F7F6070706F7F7F607F706F7F7F607F706F7F7F607F706F7F7F607F706F7F7F
607F706F7F7F607F706F7F706F7F706F7F70607F706F7F706F7F706F7F706F7F706F7F706F7F70\
6F7F706F7F6F6F7F706F7F6F6F7F706F706F6F7F706F7F6F6F7F706F706F6F7F6F
6F7F6F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F3F2F2F4030304F303040303F4F30304F3F3F50\
303F503F405F3F3F5F3F405F3F405F404F603F4060404F60404F604F4F6F404F6F
4F506F404F6F4F50704F506F5050704F5070505F70505F70505F7F505F7050607F505F7F5F607F\
505F7F5F607F50607F5F607F5F607F5F60805F607F5F60805F607F5F6F805F607F
5F6F805F607F606F805F6F80606F805F6F80606F805F6F80606F805F6F80607080606F80607080\
606F8060708F606F8060708F606F806F708F60708060708F6070806F708F607080
6F708F6070806F7F8F6070806F708F6070806F7F8F6070806F7F8F6070806F7F8F60708F6F7F8F\
6070806F7F8F6F708F6F7F8F6070806F7F8F6F7F8F6F7F8F6070806F7F8F607F80
6F7F8F6070806F7F8F6070806F7F8F6070806F7F8F6070806F7F806070806F7F80607080607F80\
6070806F7F80607080607F80607080607F80607080607F8060707F607F8060707F
607F805F707F607F8060707F6070805F707F607F8060707F6070805F707F607F805F707F60707F\
5F707F607F805F707F60707F5F707F607F7F5F707F60707F5F707F607F7F5F707F
60707F5F707F607F7F5F707F60707F5F707F607F7F5F707F60707F5F707F60707F5F707F60707F\
5F7070607F7F5F707F60707F5F7070607F7F5F707F607F7F5F7070607F7F607070
607F7F5F7070607F7F607070607F7F607070607F7F607070607F7F607070607F7F607070607F7F\
6070706F7F7F607F706F7F7F607F706F7F7F607F706F7F7F607F706F7F7F6F7F70
6F7F7F607F706F7F7F6F7F706F7F706F7F706F7F7F6F7F706F7F706F7F706F7F706F7F706F7F70\
6F7F706F7F706F7F706F7F706F7F6F6F7F706F7F6F6F7F706F7F6F6F7F706F7F6F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F30402F2F4030304F30304F303F5030304F\
303F50303F503F405F3F3F5F3F405F3F405F4040603F4060404F60404F60404F6F
404F604F506F404F6F4F50704F506F4F50704F5070505F704F5070505F7F505F70505F7F505F70\
50607F505F7F5F607F505F7F5F607F50607F5F607F50607F5F607F5F607F5F6080
5F607F5F6F805F607F5F6F805F607F5F6F805F607F606F805F607F606F805F6F80606F805F6F80\
606F805F6F806070805F6F80607080606F8060708F606F8060708060708060708F
60708060708F6070806F708F60708060708F6070806F708F60708060708F6070806F7F8F607080\
6F7F8F6070806F7F8F6070806F7F8F6070806F7F8F607080607F806070806F7F80
607080607F80607080607F80607080607080607080607F8060707F607080607080607F8060707F\
60708060707F6070805F707F6070805F707F6070805F707F60707F5F707F607080
5F707F60707F5F707F60707F5F707F5F707F5F707F60707F5F707F5F707F5F707F5F707F5F707F\
5F707F5F707F5F707F5F70707F605F7F5F5F7F605F7F5F5F7F605F7F605F7F605F
7F605F7F605F7F605F80605F7F605F7F605F7F605F80605F7F605F80605F7F605F80605F7F6050\
5F707F5F70705F707F5F70705F707F5F70705F707F5F70705F70705F707060707F
5F70705F70705F707060707F5F70706070705F70706070705F70706070705F7070607F7F5F7070\
607F70607070607F7F607070607F70607070607F7F607070607F70607F706F7F70
6070706F7F70607F706F7F706070706F7F70607F706F7F706070706F7F70607F706F7F7060706F\
6F7F7060706F6F7F7060706F6F7F7060706F6F7F7060706F6F7F6F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F4030304F30304F303F4F30304F3F3F50\
303F503F405F3F3F5F3F40603F405F4040603F4060404F60404F60404F6F404F60
4F506F4F4F6F4F50704F506F505F704F5070505F704F5F70505F7F505F7050607F505F7F5F607F\
505F7F5F607F50607F5F60805F607F5F60805F607F5F6F805F607F5F6F805F607F
5F6F805F60805F6F805F607F606F805F6F80606F805F6F80606F805F6F80606F805F6F80607080\
606F80607080606F8060708F60708060708F6070806F708F60708060708F607080
6F7F8F60708F6F708F6070806F7F8F60708F6F7F8F6070806F7F8F60708F6F7F8F6070806F7F8F\
60708F6F7F8F6070806F7F8F6070806F7F8F6070806F7F8F6070806F7F8F607080
6F7F8F607080607F806070806F7F80607080607F80607080607F80607080607F80607080607F80\
607080607F8060707F607F8060707F6070805F707F607F805F707F607080705F60
7F60607F5F607F60607F5F607F60607F60607F60607F60607F60607F60607F60607F6060806060\
7F60608060607F6060806F607F60608060607F605F806F607F6060806F607F605F
806F6080605F806F607F605F806F60806F5F806F6080605F806F60806F5F806F607F605F806F5F\
7F605F806F5F7F605080605F7F60505F707F5F707060707F5F70705F707F5F7070
60707F5F707060707F5F707060707F5F707060707F5F7070607F7F5F7070607F7F5F7070607F7F\
607070607F7F607070607F7F607F70607F7F6070706F7F7F607F706F7F7F607F70
6F7F7F607F706F7F7F607F706F7F7F607F706F7F70607F706F7F706F7F706F7F706F7F706F7F70\
6F7F706F7F706F7F6F6F7F706F7F6F6F7F706F7F6F6F7F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F3F2F2F4030304F2F304F303F50303050\
3F3F50303F503F405F3F3F5F3F40603F405F4040603F4060404F60404060404F6F
404F604F506F404F6F4F506F4F506F4F5F704F5070505F704F5F70505F7F505F70505F7F505F7F\
5F607F505F7F5F607F50607F5F607F50607F5F607F50607F5F608050607F5F6080
5F607F5F6F805F607F5F6F805F607F5F6F805F607F5F6F805F6F80606F805F6F7F606F805F6F80\
606F805F6F80606F805F6F806070805F6F80607080606F8060708F606F8060708F
60708060708F60708060708F6070806F708F60708060708F6070806F7F8F60708060708F607080\
607F8F607080607080607080607F80607080607080607080607080607080607080
60708060708060707F6070806070806070805F707F6070805F707F6070805F707F6070805F707F\
6070805F707F60707F5F707F6070805F707F705F60705F607F5F60705F607F6060
7F5F607F606F7F5F607F60607F60607F60607F60607F60607F60608060607F60607F60607F6060\
8060607F60608060607F6060806F607F6060806F607F6060806F607F6060806F60
806060806F607F605F806F60806F60806F60806F5F806F60806F5F806F60806F5F806F5F7F605F\
806F5F7F605F806F5F7F60507F605F7F60507F60507F60507F6050506F705F7070
5F6F705F70705F6F705F70705F6F705F70705F6F705F70705F70706070705F70706070705F7070\
6070705F70706070705F7070607F705F7070607F70607070607F70607070607F70
6070706F7F70607070607F706070706F7F706070706F7F706070706F7F7060706F6F7F7060706F\
6F7F7060706F6F7F7060706F6F7F7060706F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000030304F2F304F303F5030304F303F50\
303F503F405F3F3F5F3F405F3F405F4040603F4060404F60404060404F6F404F60
4F506F404F6F4F50704F506F4F50704F5070505F704F5F70505F7F505F7050607F505F7F5F607F\
505F7F5F607F50607F5F607F50607F5F60805F607F5F60805F607F5F6F805F607F
5F6F805F607F5F6F805F607F5F6F805F607F606F805F6F80606F805F6F80606F805F6F80606F80\
5F6F80607080606F80607080606F8060708F60708060708F60708060708F607080
60708F6070806F7F8F6070806F708F6070806F7F8F607080607F8F6070806F7F8F60708060708F\
607080607F8F607080607080607080607F80607080607080607080607F80607080
6070806070806070806070806070805F707F6070805F707F6070805F707F6070805F707F607080\
5F707F6070805F707F705F60705F607F5F60705F607F6060705F607F606F7F6060
7F606F7F606F806F6F7F606F806F6F7F6060806F6F7F6060806F6F7F6060806F6F7F6060806F6F\
7F6060806F6F806F60806F6F806060806F6F806F60806F6F806F60806F6F806F60
806F60806F60806F6F806F60806F60806F608F7060806F608F7060806F608F7060806F60806F60\
806F5F806F60806F5F806F5F7F605F806F5F7F60507F6F5F7F60507F60507F6050
7F605070604F5F70705F6F705F70705F6F705F70705F70705F70705F70706070705F7070607070\
5F7070607F705F70706070705F7070607F70607070607F70607070607F70607070
607F706070706F7F70607F706F7F706070706F7F70607F706F7F706070706F7F70607F706F7F70\
60706F6F7F70607F6F6F7F7060706F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F3F2F304F2F304F303F5030304F\
303F50303F503F405F303F5F3F405F3F405F3F40603F405F404F603F4060404F6F
404F60404F6F404F6F4F506F40506F4F50704F5070505F704F5070505F7F505F7050607F505F70\
50607F505F7F50607F505F7F50607F50607F5F607F50607F5F607F50607F5F6080
50607F5F607F5F607F5F6F805F607F5F6F805F607F5F6F805F607F5F6F805F6F7F606F805F6F7F\
5F6F805F6F806070805F6F806070805F6F806070805F6F80607080606F8060708F
60708060708060708060708F6070806070806070806070806070806070806070806070805F7080\
6070806070806070805F70806070805F70806070805F707F6070805F707F607080
5F707F6070805F707F6070805F6F7F5F707F5F707F6070805F6F7F5F707F5F707F5F707F6F505F\
705F606F505F705F60705F60705F60705F607F606F705F607F606F7F60607F606F
7F60607F606F7F606F806F6F7F606080606F7F6060806F6F7F6060806F6F7F6060806F6F7F6060\
806F6F7F6060806F6F7F6060806F6F806F60806F6F7F6F60806F6F806F60806F6F
806F60806F60806F60806F60806F60806F60806F60807060806F60807060806F60806F60806F5F\
806F60806F5F806F607F6F5F806F5F7F605F806F5F7F60507F6F5F7F60507F6050
7060507F605070604F7F6050705F4F70604F705F4F70604F506F6F5F70705F6F705F70705F6F6F\
5F70705F70706070705F70705F70705F70706070705F70706070705F7070607F70
5F7070607070607070607F70607070607F70607070607F7060706F607F7060706F607F7060706F\
607F7060706F6F7F70000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000030303F2F304F303F5030304F303F50\
303F503F3F5F303F5F3F40603F405F3F40603F4060404F60404F60404F6F404F60
4F506F404F6F4F506F4F506F4F50704F5070505F704F5070505F7F505F7050607F505F7F5F607F\
505F7F5F607F50607F5F608050607F5F607F50607F5F60805F607F5F6F805F607F
5F6F805F607F5F6F805F607F5F6F805F6F805F6F805F6F80606F805F6F80606F805F6F80607080\
5F6F806070805F6F8060708F606F80607080606F8060708F60708060708F607080
60708F60708060708F607080607F8F60708060708F60708060708F607080607080607080607080\
6070806070805F70806070806070806070805F70806070805F70806070805F707F
6070805F707F6070805F707F6070805F707F6070805F707F705F606F5060705F606F5060705F60\
705F60705F6F705F607F606F705F6F7F606F7F606F7F606F7F606F7F6F6F7F606F
806F707F606F806F707F606F806F70806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F\
806F6F806F6F806F6F80706F806F6F806F6F806F6F8F706F806F6F80706F806F60
8F706F806F6F8F706F806F608F706F806F608F706F806F608F706F8070608F706F806F608F7060\
806F60807060806F5F807060806F5F806F60806F5F806F5F7F6F5F806F5F7F6050
806F5F7F60507F6F507F60507F6F507F604F7F605070604F7F605070604F70604F705F4070604F\
5F70705F70705F70706070705F70706070705F7070607F705F7070607F705F7070
607F70607070607F70607070607F70607070607F706070706F7F70607F706F7F706070706F7F70\
6070706F7F70000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030304F2F304F303050303050\
303F5F303F503F3F5F303F5F3F405F3F405F3F40603F4060404F603F4060404F6F
404F6040506F404F6F4F506F4F506F4F50704F506F4F5F704F5F70505F7F505F70505F7F505F7F\
50607F505F7F50607F505F7F5F607F505F7F5F607F50607F5F607F50607F5F607F
50607F5F6F8050607F5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F80\
606F805F6F80606F805F6F806070805F6F806070805F6F806070805F6F80607080
60708060708F6070806070806070806070805F70806070805F70806070805F6F806070805F7080\
6070805F6F7F6070805F707F6070805F6F7F5F70805F6F7F5F70805F6F7F5F707F
5F6F7F5F70805F6F7F5F707F604F5F6F50606F505F6F50606F5060705F606F5060705F60705F60\
705F6F705F6070606F705F607F606F705F607F606F7F606F7F606F7F606F7F6F6F
7F606F806F707F606F806F6F7F6F6F806F707F606F806F6F806F6F806F6F7F6F6F806F6F806F6F\
806F6F806F6F806F6F806F6F80706F806F6F806F6F806F6F80706F806F6080706F
806F6F8F706F806F6080706F806F608F706F806F608F706F8070608F706F806F60807060806F60\
807060806F60807060806F5F8070607F6F5F806F5F7F6F5F806F5F7F6F5F806F5F
7F6F50806F5F7F60507F6F507F60507F6F5070604F7F605070604F7F604F705F4F70604F705F4F\
705F4F5F6F6F5F70705F6F6F5F70705F6F6F5F70705F706F6070705F706F607070
5F706F6070705F706F6070705F706F607F705F706F607F7060706F607F7060706F607F7060706F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F40303F50303050303F50\
303F503F3F5F303F5F3F405F3F3F5F3F40603F4060404F603F4060404F6F404F60
404F6F404F6F4F506F40506F4F50704F506F505F704F5F70505F7F4F5F7050607F505F7F50607F\
505F7F5F607F50607F5F607F50607F5F608050607F5F607F50607F5F6F805F607F
5F6F805F607F5F6F805F607F5F6F805F607F5F6F805F6F805F6F805F6F80606F805F6F80606F80\
5F6F806070805F6F806070805F6F8060708F60708060708060708060708F607080
60708F60708060708F6070806070806070806070806070806070805F70806070805F7080607080\
5F707F6070805F70806070805F6F7F6070805F707F5F70805F6F7F5F7080604F5F
6F5060604F5F6F50606F50606F50606F5060705F606F5F60705F6F6F5F60705F6F705F6070606F\
705F607F606F70606F7F606F7F606F7F60707F606F7F6F707F606F806F707F6F6F
806F707F6F6F806F70806F6F806F70806F6F806F70806F6F806F70806F6F807070806F6F807070\
806F6F807070806F6F807070806F6F8F7070806F6F8F706F806F6F8F707080706F
8F706F80706F8F706F80706F8F706F80706F8F706F8F706F8F706F80706F8F706F8070608F706F\
8070608F706F807060807060806F60807060806F5F8070607F6F5F8070607F6F5F
806F5F7F6F5F806F5F7F6F507F6F5F7F60507F6F507F60507F6F5070604F7F6F5070604F7F604F\
705F4F70604F6F5F405F70705F6F6F5F70705F70705F70705F706F6070705F7070
6070705F706F6070705F70706070705F706F607F70607070607F7060706F607F70607070000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F3F2F304F303F502F3050\
303F50303F50303F5F303F5F3F405F3F405F3F40603F405F3F4F603F4060404F6F
404F60404F6F404F6F4F506F404F6F4F50704F506F4F5F704F5070505F704F5F70505F7F4F5F70\
50607F505F7F50607F505F7F50607F50607F5F607F50607F5F607F50607F5F607F
50607F5F607F50607F5F6F8050607F5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F805F6F7F\
5F6F805F6F806070805F6F806070805F6F806070805F6F806070805F6F80607080
5F6F806070805F6F806070805F6F806070805F70806070805F6F7F6070805F6F7F6070805F6F7F\
5F70805F6F7F5F70805F6F7F5F707F5F6F7F5F707F5F4F5F604F5F604F5F6F5060
604F5F6F50606F50606F50606F50606F5F606F5060705F6F6F5060705F6F705F60705F6F705F60\
70606F705F6F7F606F70606F7F606F7F606F7F6F707F606F7F6F707F606F806F70
7F6F6F806F707F6F70806F707F6F6F806F70806F6F806F70806F6F806F70806F6F807070806F6F\
807070806F6F807070806F6F80706F806F6F8F7070806F6F80706F80706F8F706F
806F6F8F706F80706F8F706F80706F8F706F80706F8F706F80706F8F706F80706F8F706F807060\
80706F807060807060806F60807060806F5F8070607F6F5F806F5F7F6F5F806F5F
7F6F5F7F6F5F7F6F507F6F5F7F60507F6F507F60507F6F507060507F6F5070604F7F6F5070604F\
70604F705F40705F406F503F5F7070506F6F5F7070506F6F5F70705F6F6F5F7070
5F6F6F5F70705F706F6070705F706F5F70705F706F6070705F706F607070000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F3030304F2F3050303F5F\
303F50303F5F303F5F3F405F3F405F3F40603F405F404F603F4060404F6F404F60
40506F404F6F4F506F40506F4F50704F50704F5F704F5070505F7F505F70505F7F505F7050607F\
505F7F50607F50607F5F607F50607F5F607F50607F5F6F8050607F5F6F8050607F
5F6F805F607F5F6F805F607F5F6F805F6F805F6F805F6F7F606F805F6F80606F805F6F80607080\
5F6F806070805F6F806070806070806070805F708060708F6070806070805F7080
6070806070806070805F70806070806070806070805F70806070805F70806070805F707F607080\
5F707F6070805F6F7F6070805F707F604F60604F5F6F50606050606F5060605060
6F5F6F6F50606F5F6F6F5060705F6F6F5F60705F6F705F6F70606F705F6F70606F705F6F7F6070\
70606F7F60707F606F7F6F707F606F7F6F707F6070806F707F6F70806F70806F70
80707F806F70807070806F70807070806F70807070806F708F70708070708F70708070708F7070\
8070708F70708070708F70708070708F707080706F8F70708070708F707080706F
8F7F708F706F8F7F708F706F8F7F708F706F8F7F708F706F8F7F708F706F8F7F6F8F706F8F7F6F\
80706F8F706F8070608F706F807060807060807060807060807060807060806F5F
8070607F6F5F80705F7F6F5F80705F7F6F5F7F6F5F7F6F507F6F5F7F6F507F6F507F60507F6F50\
7F604F7F604F70604F70604F6F5F406F5F405F6F6F5F70705F706F5F70705F6F6F
5F70705F706F5F70705F706F6070705F706F607070000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F403030502F3050\
303F5F303F50303F5F303F5F3F405F303F5F3F40603F405F404F603F4060404F60
404F6040506F404F6F4F506F40506F4F50704F506F4F5F704F5070505F7F4F5F70505F7F505F70\
50607F505F7050607F505F7F50607F50607F50607F50607F5F607F50607F5F607F
50607F5F6F8050607F5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F80\
606F805F6F80606F805F6F806070805F6F806070805F6F806070805F6F80607080
5F6F806070805F6F7F6070805F6F7F6070805F6F7F5F70805F6F7F6070805F6F7F5F70805F6F7F\
5F70805F6F7F5F707F5F4F5F604F60604F5F605060604F606F50606050606F5060
6F50606F5F6F6F50606F5F6F6F5F60705F6F6F5F60705F6F705F6F70606F705F6F70606F70606F\
7F607070606F7F60707F606F7F6F707F606F7F6F707F6F70806F707F6F70806F70
806F7080707F806F70807070806F70807070806F70807070806F70807070806F70807070807070\
8F70708070708F70708070708F707080706F8F70708070708F707080706F8F7070
80706F8F7F7080706F8F7F708F706F8F7F7080706F8F7F708F706F8F7F7080706F8F7F6F80706F\
8F7F6F8070608F706F80706080706F8070608070608070608070607F6F5F807060
7F6F5F8070607F6F5F806F5F7F6F5F806F5F7F6F507F6F5F7F6F507F6F5F7F60507F6F50706050\
7F6F5070604F70604F6F5F406F5F4060503F605F3F506F6F5F7070506F6F5F706F
5F6F6F5F70705F6F6F5F706F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000030303F2F3050303F50\
303F50303F5F303F5F3F405F303F5F3F40603F405F3F4F603F4060404F6F404F60
404F6F404F6F4F506F40506F4F50704F506F4F5F704F5070505F704F5F70505F7F505F7050607F\
505F7050607F505F7F50607F50607F5F607F50607F5F607F50607F5F6F8050607F
5F6F8050607F5F6F805F607F5F6F805F607F5F6F805F6F805F6F805F6F7F606F805F6F80606F80\
5F6F806070805F6F806070805F6F806070805F6F806070805F6F806070805F6F80
6070805F6F806070805F70806070805F6F7F6070805F707F6070805F6F7F6070805F707F5F7080\
5F6F7F5F4F5F5F4F5F604F60604F606F50606050606F506F6050606F5F6F6F5060
6F5F6F6F5060705F6F6F5F6F705F6F6F5F6F70606F705F6F706070705F6F7F607070606F7F6070\
70606F7F6F707F60707F6F707F6F70806F707F6F70806F7F7F6F7080707F806F70
80707F806F708F707F8070708F707F8070708F707F8070708F707F8070708F707F8070708F707F\
8070708F7F7F8070708F7F708070708F7F708F70708F7F708F70708F7F708F7070
8F7F708F70708F7F708F7F708F7F708F70708F7F708F7F708F7F708F7F6F8F7F708F7F6F8F7F70\
8F706F8F7F6F80706F8F7F6F8070608F7F6F807060807060807060807060807060
8070607F6F5F8070607F6F5F80705F7F6F5F80705F7F6F5F7F6F5F7F6F507F6F5F7F6F507F6F50\
706F507F6F5070604F70604F6F5F4070604F6F5F406F5F4060503F5F7070506F6F
5F706F504030504F30000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F3030304F2F3050\
303F50303F50303F5F303F503F405F303F5F3F40603F405F3F40603F4060404F60
3F4F60404F6F404F6F4F506F404F6F4F50704F506F4F5F704F50704F5F704F5F70505F7F4F5F70\
505F7F4F5F7050607F505F7050607F50607F50607F50607F50607F50607F5F607F
50607F5F607F50607F5F6F8050607F5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F805F6F7F\
5F6F805F6F806070805F6F7F5F6F805F6F805F70805F6F7F5F70805F6F7F5F7080
5F6F7F5F70805F6F7F5F70805F6F7F5F70805F6F7F5F70805F6F7F5F707F5F6F7F5F707F5F6F7F\
5F4F5F5F4F5F604F605F4F5F6050606050606F506F6050606F506F6F50606F5F6F
6F50606F5F6F6F5F6F705F6F6F5F6F705F6F705F6F706070705F6F70607070606F7F607070606F\
7F60707F60707F6F707F60707F6F707F6F70806F7F7F6F70806F7F806F7080707F
806F7080707F80707080707F80707080707F8070708F707F8070708F707F8070708F707F807070\
8F70708070708F7F7F8070708F70708070708F7F708070708F7F708F70708F7F70
8070708F7F708F70708F7F708F70708F7F708F7F708F7F708F7F6F8F7F708F7F6F8F7F7080706F\
8F7F6F80706F8F7F6F80706080706F80706080706F807060807060807060807060
7F6F5F8070607F6F5F8070607F6F5F7F705F7F6F5F7F705F7F6F507F6F5F7F6F507F6F5F706F50\
7F6F5070604F706F5070604F70604F6F5F406F5F4060503F60503F5F503060503F
5F4F305F4F30504030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F3040303F50\
303F50303F5F303F503F405F30405F3F40603F405F3F4F603F4060404F603F4F60
40506F404F6F40506F40506F4F50704F50704F5F704F5070505F704F5F70505F7F4F5F7050607F\
505F7050607F505F7F50607F50607F5F607F50607F5F6F8050607F5F6F8050607F
5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F805F6F805F6F80607080\
5F6F806070805F6F806070805F6F806070805F6F806070805F6F806070805F6F7F
6070805F6F7F5F70805F6F7F6070805F6F7F5F70805F6F7F6070805F6F7F5F70805040505F4F5F\
5F4F5F604F60604F6060506F6050606F506F6050606F5F6F6F506F6F5F6F6F5F6F
705F706F5F6F7060706F5F6F70607070606F70607070606F7F60707060707F6F707060707F6F70\
7F6F707F6F7F7F6F70806F7F7F6F7080707F7F6F7080707F80707F80707F80707F
8F7F7F80707F8F707F80707F8F7F7F80707F8F7F7F80707F8F7F7F8F707F8F7F7F80707F8F7F7F\
8F707F8F7F7F8F70708F7F7F8F7F7F8F7F7F8F7F708F7F7F8F7F708F7F7F8F7F70
907F7F8F7F708F7F7F8F7F7090807F8F7F70907F7F8F7F7090807F8F7F70907F708F7F708F7F70\
8F7F708F7F708F7F6F8F7F708F7F6F8F7F6F80706F8F7F6F807060807F6F807060
807F6F80706080706080706080706080705F8070607F6F5F80705F7F6F5F80705F7F6F507F705F\
7F6F507F6F5070604F7F6F5070604F70604F6F5F406F604F605F40605F3F60503F
60503F5F4F305F4F30504030000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F303F2F304F\
303F502F3F50303F5F303F5F3F405F303F5F3F40603F405F3F4F603F4060404F60
3F4F60404F6F404F6F40506F40506F4F507040506F4F5F704F50704F5F704F50704F5F704F5F70\
505F7F4F5F7050607F505F7050607F505F7050607F50607F5F607F50607F5F607F
50607F5F6F8050607F5F6F8050607F5F6F8050607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F80\
5F6F805F6F7F5F6F805F6F7F5F70805F6F7F5F6F805F6F7F5F70805F6F7F5F6F80
5F6F7F5F70805F6F7F5F6F7F5F6F7F5F707F5F6F7F5F707F5F6F7F50405F5040505F4F5F5F405F\
604F605F4F5F6050606050606F506F6050606F5F6F6F506F6F5F6F6F506F6F5F6F
6F5F6F705F706F5F6F706070705F6F706070705F6F7060707060707F6F707060707F6F707F6070\
7F6F7F7F6F707F6F7F7F6F7080707F7F6F7080707F806F7F80707F80707F80707F
80707F8F7F7F80707F8F707F80707F8F7F7F80707F8F7F7F80707F8F7F7F80707F8F7F7F8F707F\
8F7F7F8070708F7F7F8F7F7F8F7F7F8F70708F7F7F8F7F708F7F7F8F7F708F7F7F
8F7F708F7F7F8F7F708F7F7F8F7F70907F7F8F7F708F7F7F8F7F7090807F8F7F708F7F708F7F70\
8F7F708F7F6F8F7F70807F6F8F7F6F80706F8F7F6F80706F8F7F6F807060807060
8070608070607F70608070607F705F8070607F6F5F80705F7F6F5F80705F7F6F5F7F705F7F6F50\
7F6F5F706F50706F5070604F70604F6F60406F604F6F5F406F5F4060503F605F3F
5F50305F503F504F305F4F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F2F3030304F\
2F3050303F5F303F5F30405F303F5F3F40603F405F3F40603F4060404F603F4F60
404F6F404F6040506F40506F4F507040506F4F5F704F50704F5F704F5070505F704F5F70505F7F\
4F5F7050607F505F7050607F505F7F50607F50607F50607F50607F5F6F8050607F
5F6F8050607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F805F6F805F6F7F5F70805F6F805F6F80\
5F6F7F5F70805F6F805F6F805F6F7F5F70805F6F7F5F70805F6F7F5F70805F6F7F
5F70805F6F7F5F70805F6F7F5F707F5F6F7F5F708050405050405F50405F5F4F605F4F5F5F4F60\
5F4F6060506060506060506F60506F6F5F6F6F5F6F6F5F6F6F5F6F705F706F5F6F
7060706F5F6F7060707060707060707060707F6F707060707F6F707060707F6F7F7F6F707F6F7F\
7F6F7080707F7F6F7F80707F7F6F7F80707F80707F80707F80707F8F7F8080707F
8F7F8080707F8F7F808F707F8F7F8080707F8F7F808F7F7F8F7F808F7F7F8F7F808F7F7F8F7F7F\
8F7F7F8F7F808F7F7F8F7F7F8F7F7F8F807F8F7F7F8F7F7F8F7F7F90807F8F7F7F
90807F8F7F7F90807F8F7F7F90807F8F7F7090807F8F7F7F90807F8F7F7090807F8F7F70908070\
8F7F708F80708F7F708F7F708F7F6F8F7F70807F6F8F7F6F80706F8F7F6F80706F
807F6F807060807F6F8070608070607F70608070607F705F8070607F705F8070607F705F7F705F\
7F6F507F705F706F50706F5070604F706F4F6F604F6F604F6F5F406F6040605F3F
605F3F5F50305F503F5F4F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F40\
3030502F3050303F5F303F5030405F303F5F3F406030405F3F40603F4060404F60
3F4F60404F6F404F6F40506F404F6F4F507040506F4F5F7040506F4F5F704F50704F5F704F5F70\
4F5F704F5F70505F7F4F5F7050607F505F7050607F505F7050607F50607F5F607F
50607F5F6F7F50607F5F6F8050607F5F6F8050607F5F6F80506F7F5F6F805F6F7F5F6F80506F7F\
5F6F805F6F7F5F6F80506F7F5F6F7F5F6F7F5F6F80506F7F5F6F7F506F7F5F6F7F
506F7F5F6F7F506F7F5F6F7F506F7F5F6F7F4F3F5050405F5040505F405F50405F5F4F605F4F5F\
604F605F4F6060506F6050606F506F60506F6F5F706F5F6F6F5F706F5F6F706070
6F5F6F7060706F5F707060707060707060707060707F6F7F7060707F6F7F7060707F6F7F7F6F70\
7F6F7F7F6F7F80707F7F6F7F80707F80707F80707F80707F80708080707F8F7F80
80707F8F7F8080707F8F7F8080707F8F7F80807F7F8F7F8080707F8F7F808F7F7F8F7F808F7F7F\
8F7F7F8F7F7F8F7F7F8F7F7F8F7F7F8F7F7F8F7F7F8F7F7F8F7F7F8F7F7F8F807F
8F7F7F8F807F8F7F7F90807F8F7F708F807F8F7F7F90807F8F7F7090807F8F7F7090807F8F7F70\
8F80708F7F708F7F70807F6F8F7F70807F6F8F7F70807F6F8F7F6F80706F807F6F
807060807F6F807060807F607F70608070607F70608070607F705F8070607F705F7F705F7F6F5F\
7F705F706F507F6F5070604F706F506F604F70604F6F5F406F6040605F406F5F40
60503F60503F5F50305F503F504F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303040\
2F304F303F50303F503F405F30405F3F40603F405F3F4F603F4060404F603F4F60
404F6F404F6F40506F404F6F4F50704F506F4F5F704F50704F5F704F5F704F5F704F5F70505F7F\
4F5F7050607F4F5F7050607F50607F50607F50607F5F607F50607F5F607F50607F
5F6F8050607F5F6F8050607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F805F6F805F6F7F5F6F80\
5F6F7F5F6F805F6F7F5F70805F6F7F5F6F805F6F7F5F70805F6F7F5F6F805F6F7F
5F70805F6F7F5F707F506F7F50405050405050405F50405F5F4F605F4F5F5F4F605F4F6060506F\
60506060506F60506F6F5F6F6F5F6F6F5F706F5F6F7060706F5F707060706F5F70
70607070607070607F7060707F6F7F7060707F6F7F7060707F6F7F7F6F7F7F6F7F7F6F7F80707F\
7F6F7F8070807F707F80708080707F80708080707F8F7F808070808F7F80807F80
8F7F808F7F808F7F808F7F808F7F808F7F808F7F808F7F808F80808F7F808F80808F7F7F8F8080\
8F7F7F8F80808F7F7F9080808F7F7F8F80808F7F7F9080808F807F9080808F7F7F
9080808F807F90807F8F807F9080808F807F90807F8F807F90807F8F807F90807F8F807090807F\
8F807090807F8F7F709080708F7F708F80708F7F6F8F80708F7F6F8F7F70807F6F
8F7F6F807F6F807F6F807060807F6F807060807F60807060807F60807060807F607F705F807060\
7F705F7F705F706F507F6F5F706F50706F506F604F70604F6F604F6F604F605F40
6F5F40605F3F605F3F5F50305F503F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F30\
30304F2F304F303F50303F5030405F303F5F3F405F3F405F3F4F603F40603F4F60
3F4F60404F6F404F6040506F40506F4F507040506F4F507040506F4F5F704F506F4F5F704F5F70\
4F5F704F5F70505F704F5F7050607F4F5F7050607F50607F50607F50607F50607F
50607F5F6F8050607F5F6F8050607F5F6F8050607F5F6F80506F7F5F6F80506F7F5F6F7F506F7F\
5F6F80506F7F5F6F7F506F7F5F6F7F506F7F5F6F7F506F7F5F6F7F506F7F5F6F7F
506F7F5F6F7F403F4F4F3F504F3F5050405F50405050405F50405F5F4F605F4F5F5F4F605F4F60\
60506F60506060506F60506F6F5F706F5F6F6F5F706F5F707060706F5F70706070
6F607070607F70607070607F7060707F6F7F7060707F6F7F7F6F7F7F6F7F7F6F7F7F6F7F7F6F7F\
8070807F6F7F80708080707F80708080707F8070808070808F7F808070808F7F80
807F808F7F80807F808F7F80807F808F7F80807F7F8F7F808F7F7F8F8080807F7F8F7F808F7F7F\
8F80808F7F7F8F80808F7F7F8F80808F7F7F8F80808F7F7F8F80808F7F7F8F807F
8F807F90807F8F7F7F8F807F8F807F90807F8F807F90807F8F807F90807F8F80708F807F8F7F70\
8F807F8F7F708F80708F7F708F80708F7F6F8F7F708F7F6F8F7F70807F6F8F7F6F
807F6F8F7F6F807F60807F6F807060807F6F7F7060807F607F7060807F607F705F8070607F705F\
7F705F7F6F507F705F706F50706F506F604F706F4F6F604F6F604F605F406F6040
605F40605F405F503F605F3F5F5030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F30\
2F2F40303F50303F50303F5F303F5F3F40603F405F3F40603F4060404F603F4F60
404F6F404F6040506F40506F4F507040506F4F507040506F4F5F704F506F4F5F704F5F704F5F70\
4F5F70505F7F4F5F7050607F4F5F7050607F50607F50607F50607F50607F50607F
5F6F7F50607F5F6F80506F7F5F6F80506F7F5F6F805F6F7F5F6F80506F7F5F6F805F6F7F5F6F80\
506F7F5F6F80506F7F5F6F7F506F7F5F6F7F506F7F5F6F7F506F7F5F6F7F506F7F
4F3F504F3F5050405F4F3F5050405F50405F5F4F6050405F5F4F605F4F6060506F5F4F6060506F\
60506F6F5F6F60506F6F5F706F5F706F60706F5F707060706F607070607F706070
70607F7060707F6F7F70607F7F6F7F706F7F7F6F7F7F6F7F7F6F7F7F6F7F8070807F6F7F807080\
7F707F807080807080807080807080807F80807080807F80807F808F7F8F807F80
8F7F8F807F808F808F8F7F808F80808F7F808F80808F7F808F80808F7F808F80808F7F808F8080\
8F7F808F80808F80808F80808F7F7F8F80808F807F8F80808F807F9080808F807F
9080808F807F9080808F807F9080808F807F908F808F807F908F7F8F807F90807F8F807F90807F\
8F807090807F8F80708F807F8F7F708F80708F7F708F8070807F6F8F8070807F6F
8F8070807F6F8F806F807F6F807F6F807F60807F6F807F60807F6F807060807F607F7060807F60\
7F705F7F705F7F705F7F705F706F50706F5070604F706F4F6F604F6F604F6F5F40
6F6040605F3F605F4060503F605F3F5F5030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F3F2F304F303F502F3050303F5F303F5F3F406030405F3F40603F40603F4F60
3F4F60404F6F404F6040506F404F6F40506F40506F4F507040506F4F507040506F4F5F704F506F\
4F5F704F5F704F5F704F5F70505F704F5F7050607F4F5F7050607F50607F50607F
50607F50607F50607F5F6F8050607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F\
5F6F7F506F7F5F6F7F50607F5F6F7F506F7F5F6F7F506F7F5F6F7F40304F4F3F50
40304F4F3F504F3F5050405F4F3F5050405F50405F5F4F6050405F5F4F605F4F6060506F5F5060\
60506F60506F6F5F70605F6F6F5F706F5F7070607F6F5F7070607F6F607070607F
706070706F7F70607F7F6F7F70607F7F6F7F706F7F7F6F7F706F7F7F70807F6F7F7F70807F6F7F\
8070807F70808070807F7080807080807080807F80807080807F80807F808F7F8F
807F808F7F8F807F808F808F807F808F7F80807F808F8080807F808F8080807F808F8080807F7F\
8F80808F7F7F8F80808F7F7F8F80808F807F8F80808F807F8F80808F807F8F8080
8F807F8F80808F807F8F80808F807F8F807F8F807F908F7F8F807F8F807F8F807F8F807F8F8070\
8F807F8F7F708F807F807F708F8070807F708F8070807F6F8F7F70807F6F8F8070
807F6F807F6F807F6F807F6F807F60807F6F807F60807F6F7F7060807F607F7060807F607F705F\
7F70607F705F7F705F706F50706F50706F50706F506F604F6F604F6F60406F604F
605F40605F40605F3F605F3F5F50305F503F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F3030304F2F304F303F50303F503F405F30405F3F4F603F40603F4F603F4F60
404F6F404F6F40506F404F6F4F507040506F4F507040506F4F5F704F50704F5F704F50704F5F70\
4F5F704F5F704F5F7050607F4F5F7050607F4F5F7050607F50607F50607F50607F
5F6F7F50607F5F6F8050607F5F6F805F6F7F5F6F80506F7F5F6F805F6F7F5F6F80506F7F5F6F80\
5F6F7F5F6F7F506F7F5F6F80506F7F5F6F7F506F7F5F6F7F40304F4F3F50403F50
4F405F4F3F5050405F4F405F504F6050405F5F4F60504F605F506F5F4F6060506F5F506F605F70\
60506F6F5F70605F706F60706F5F7070607F6F6070706F7F70607F706F7F70607F
7F6F7F706F7F7F6F80706F7F7F70807F6F7F7F70807F6F7F7F70807F70808070807F7080807080\
7F7080807F80807080807F8F807F80807F8F807F808F7F8F807F8F8F7F8F807F8F
8F808F8F7F8F8F808F8F7F8F8F808F8F80808F808F8F80808F808F8F80808F80808F80808F8080\
8F80808F80808F80808F80808F80808F80808F8080908F808F80808F8F808F8080
908F808F8080908F808F807F908F808F807F908F808F807F908F808F807F908F7F8F807F908F7F\
8F807F8F807F8F80708F807F8F80708F807F8F7F708F80708F80708F8070807F6F
8F8070807F6F8F8070807F6F8F8070807F6F807F6F807F6F807F6F807F60807F6F807F60807F6F\
7F70607F7F607F705F7F705F706F5070705F706F50706F506F604F6F6F4F6F604F
6F604F605F40606040605F3F605F405F503F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F302F2F3F30304F2F304F303F50303F5030405F30405F3F4F603F40603F4F60
3F4F60404F6F3F4F6040506F404F6F40506F40506F40506F40506F4F5F7040506F4F5F704F5070\
4F5F704F5F704F5F704F5F705060704F5F7050607F4F5F7050607F4F607050607F
50607F506F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F506F7F5F6F7F50607F5F6F7F506F7F\
5F6F7F50607F5F6F7F506F7F5F6F7F506F7F5F6F7F40304F403F504030504F3F50
4F3F504F405F4F3F5F50405F50405F504F6050405F5F4F605F4F605F506F5F4F6060506F60506F\
605F70605F6F6F5F706F5F706F607F6F607070607F6F607F706F7F70607F706F7F
70607F706F80706F7F7F6F80706F7F7F7080706F7F7F70807F6F807F70807F6F807F70807F7080\
8070807F7080807F8F7F7080807F8F807080807F8F807F8F8F7F8F807F8F8F7F8F
807F8F8F808F807F8F8F808F807F8F8F808F807F808F808F8080808F808F8080808F80808F8080\
8F80808F80808F80808F80808F80808F80808F80808F80808F8F808F807F8F8080
8F807F8F8F808F807F8F8F808F807F908F808F807F8F8F808F807F908F808F807F8F807F8F807F\
8F807F8F80708F807F8F80708F807F807F708F8070807F708F8070807F6F8F8070
807F6F8F8070807F6F807F6F807F6F807F6F807F6F807F6F807F60807F6F7F7F60807F607F7F60\
807F607F705F7F705F70705F7F705F706F50706F506F6F4F706F506F604F6F604F
6060406F6040605F40605F405F5F3F605F3F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030303F2F304F303F50303050303F5F303F5F3F405F3F405F3F4F603F4F60
404F6F3F4F6040506F404F6F40506F40506F4F507040506F4F507040506F4F5F704F50704F5F70\
4F5F704F5F704F5F704F5F704F5F7050607F4F5F7050607F50607050607F50607F
50607F50607F5F6F8050607F5F6F7F50607F5F6F80506F7F5F6F7F506F7F5F6F80506F7F5F6F7F\
506F7F5F6F7F506F7F5F6F7F506F7F5F6F7F40304F403F50403F504F3F5F4F3F50
4F405F4F405F50406050405F504F60504F605F4F6F5F4F605F506F5F506F605F6F60506F605F70\
605F706F60706F5F706F607F6F607F706F7F70607F706F7F70607F706F80706F7F
706F80706F7F7F6F80706F807F7080706F807F70807F70807F70807F708080708F7F7080807F8F\
7F7080807F8F80708F807F8F807F8F807F8F807F8F807F8F807F8F8F8090807F8F
8F8090807F8F8F80908F808F8F808F80808F8F808F8F808F8F808F8F80808F808F8F80808F808F\
8F80808F8F8F8F80808F8F808F80808F8F808F80808F8F808F8080908F808F8080
8F8F808F8080908F808F8F80908F808F8F80908F808F8F80908F808F807F908F808F807F8F8F7F\
8F807F8F8F7F8F807F8F807F8F80708F807F8F80708F8070807F708F8070807F70
8F8070807F6F8F8070807F6F808070807F6F80806F807F6F807F6F807F60807F6F807F60807F6F\
7F7F60807F607F705F7F706070705F7F705F706F50706F506F6F4F706F506F604F
6F604F6060406F604F605F40605F405F503F50503F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F302F304F2F304F303F50303F50303F5F303F503F405F3F40603F4F60
3F4060404F6F3F4F6040506F3F4F6040506F40506F40506F40506F40506F40506F4F5F7040506F\
4F5F704F5F704F5F704F5F704F5F704F5F7050607F4F5F7050607F4F607050607F
50607050607F50607F506F7F50607F506F7F50607F5F6F7F50607F506F7F50607F5F6F7F50607F\
5F6F7F506F7F5F6F7F5060705F6F7F506F7F40305040304F4F3F50403F504F405F
4F3F5050405F4F405F504F6050405F504F60504F605F506F5F4F605F506F5F506F605F7060506F\
605F70605F706F607F6F5F706F607F6F607F706F7F6F607F706F7F70607F706F80
706F7F706F80706F807F7080706F807F7080706F807F7080706F807F70807F70807F708F7F7080\
7F7F8F7F7080807F8F7F708F807F8F807F8F807F8F807F8F807F8F807F8F8F8090
807F8F8F809080808F8F8090807F8F8F808F80808F8F808F80808F8F808F8080808F808F808080\
8F808F8F80808F8F8F8080808F8F808F80808F8F808F80808F8F808F80808F8F80
8F80808F8F808F80808F8F808F807F8F8F808F8F7F908F808F807F8F8F808F807F8F8F7F8F807F\
8F807F8F807F8F807F8080708F807F8080708F807F807F708F8070807F708F8070
807F6F808070807F6F808070807F6F807F6F807F6F807F6F7F7F60807F6F7F7F60807F6F7F7060\
7F7F607F70607F7F6070705F7F705F706F5070705F6F6F50706F506F604F6F6F4F
6060406F604F605F406060405F5F3F5F5F3F504F30000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F4030304F30304F303F50303F503F405F3F405F3F4F603F4F60
404F6F3F4F6040506F404F6F40506F40506F40506F40506F4F5F7040506F4F5F7040506F4F5F70\
4F5F704F5F704F5F7050607F4F5F7050607F4F5F7050607F50607F50607F50607F
506F7F50607F506F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F506F7F5F6F7F506F7F5F6F7F\
506F7F5F6F7F506F7F5F6F7F506F7F40304F40304F4F3F5F4F3F504F405F4F405F
50406050405F504F60504F605F4F6F5F4F605F506F5F506F60507060506F605F70605F706F607F\
6F5F706F607F6F607F70607F6F607F706F806F607F706F80706F7F706F80706F80
7F7080706F807F7080706F807F708F7F70807F708F7F70807F7F8F7F708F7F7F8F7F708F807F8F\
7F7F8F807F8F807F8F807F90807F8F808090807F8F8F80908080908F8090808090
8F80908F80908F80908F808F8F8F908F808F8F8F908F808F8F8F908F808F8F8F8F8F808F8F8F8F\
8F808F8F8F8F8F80808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F808F8F8F8F8F80
908F8F8F8F80908F808F8F80908F808F8F80908F808F8F80908F808F8F808F8F808F8F7F8F8F80\
8F8F7F8F8F7F8F807F8F8F7F8F807F8F807F8080708F807F8080708F807F808070
8F8070808070808070807F6F808070807F6F808070807F6F808070807F6F80806F7F7F6F807F6F\
7F7F607F7F607F70607F7F607F705F7F705F706F5F70705F706F50706F506F604F
6F6F506F604F6F604F605F406060405F5F3F50503F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F3F2F2F4030304F2F304F303F50303F503F405F30405F3F4F60
3F4F60404F6F3F4F60404F6F3F4F6F40506F404F6F40506F40506F4F507040506F4F5F7040506F\
4F5F7040506F4F5F704F5F704F5F704F5F705060704F5F7050607F4F5F7050607F
50607F50607F50607F50607F50607F506F7F506070506F7F50607F5F6F7F506070506F7F50607F\
5F6F7F5060705F6F7F506F7F3F304F3F304F403F50403F504F405F4F3F5F4F405F
4F405F504F60504060504F60504F605F506F5F4F605F506F5F506F605F7060506F605F70605F70\
6F607F605F706F607F6F607F6F607F6F607F706F806F607F706F80706F7F706F80
706F80707080706F807F7080706F807F708F7070807F708F7F708F7F7F8F7F708F7F7F8F7F708F\
807F8F7F708F807F8F7F7F8F807F90807F8F808090807F8F808090807F908F8090
8080908F80908080908F809080808F8F809080808F8F809080808F8F8F9080808F8F8F8F8F808F\
8F8F8F80808F8F8F8F8F80808F8F8F8F80808F8F8F8F8F808F8F8F8F8F808F8F80
8F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F808F8F7F8F8F808F807F\
8F8F8080807F8F8F7F80807F8F807F8080708F807F8080708F807F807F70808070
807F70808070807F6F808070807F6F8080707F7F6F80806F7F7F6F80806F7F7F6F807F6F7F7F60\
7F7F607F70607F7F6070705F7F706070705F70705F706F507070506F6F4F6F6F50
6F604F6F604F606040606040605F40605F40504F30000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F3030304F2F304F303F50303F50303F50303F503F40603F405F
3F4F603F4F6040506F404F6F40506F404F6F40506F40506F40506F40506F4F5F7040506F4F5F70\
40506F4F5F704F5F704F5F704F5F7050607F4F5F7050607F4F607050607F50607F
50607F50607F506F7F50607F506F7F50607F5F6F7F50607F506F7F50607F5F6F7F506F7F5F6F7F\
506F7F5F6F7F506F7F3F304F3F304F403F50403F504F405F4F3F5F5040604F405F
504F60504060504F6F504F605F506F5F4F6F5F50705F506F605F70605070605F70605F706F607F\
605F706F607F6F607F6F607F6F607F706F80706F7F706F80706F80706F80706F80
707080706F807F708F7070807F708F7070807F708F7F708F7F7F8F7F708F807F907F7F8F807F90\
7F7F8F807F90807F90808090807F908080908080908080908080908F809F808090
8F809F8080908F8F9F8F80908F8F908080908F8F908F80908F8F908F808F8F8F908F8F8F8F8F90\
8F8F8F8F8F908F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F8F808F908F8F8F80
8F8F8F8F8F8090908F8F8F808F908F8F8F8090908F8F8F808F90808F8F808F8F808F8F808F8F80\
8F8F7F8F8F808F8F7F8F8F7F80807F8F8F7F80807F8F807F8080708F807F808070
808070808070808070808070808070807F6F808070807F6F8080707F7F6F80806F7F7F6F807F6F\
7F7F607F7F6F7F70607F7F6070705F7F706070705F70705F706F5070705F6F6F50
6F6F5060604F6F6F4F60604060604F5F5F3F5F503F4F4030000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F3F2F304F2F304F303F50303F4F303F50303F5F3F405F
3F405F3F4F603F4F6040506F3F4F6040506F404F6F40506F40506F40506F40506F4F5F7040506F\
4F5F7040506F4F5F704F5F704F5F704F5F705060704F5F7050607F4F607050607F
4F607050607F50607050607F50607050607F506070506F7F506070506F7F506070506F7F506070\
506F7F506070506F7F3F304040304F40304F403F50403F504F405F4F405F504060
4F4060504F60504F605F4F6F504F605F506F5F506F605F705F506F605F70605F70605F7F605F70\
6F607F605F7F6F607F6F607F6F607F6F607F706F806F607F706F80706F80706F80
706F80707080706F807F708F706F807F708F70708F7F708F70708F7F7F8F7F708F7F7F907F708F\
7F7F907F7F8F807F907F7F8F808090807F90808090807F908080908080908F809F
8080908F80908080908F8F908080908F8F908080908F8F9080808F8F8F9080808F8F8F9080808F\
8F8F908F8F8F8F8F8F808F8F8F8F8F8F8F8F8F8F8F8F8F808F8F8F8F8F808F8F8F
8F8F808F8F8F8F8F808F908F8F8F808F90808F8F808F90808F8F808F8F808F8F808F8F80808F7F\
8F8F8080807F8F8F7F80807F8F8F7F80807F8F8F7F80807080807F80807080807F
807F70808070807F708080707F7F6F8080707F7F6F8080707F7F6F80806F7F7F6F807F6F7F7F60\
7F7F607F70607F7F6070705F70705F70705F70705F6F6F5070705F6F6F506F6F50
6F604F6F6F4F60604F60604F605F405F5F405050304F4F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000030303F2F304030304F30304F303F50303F503F405F3F405F
3F4F603F4F60404F6F3F4F6040506F40506F40506F40506F4F507040506F4F5F7040506F4F5F70\
4F5F704F5F704F5F704F60704F5F7050607F4F5F7050607F50607050607F50607F
50607F50607F50607F50607F506F7F50607F506F7F5060705F6F7F50607F506F7F5060705F6F7F\
506F7F5F6F7F3F3040403F4F403F4F403F50403F504F40604F405F504F604F4060
504F6F504F605F506F504F6F5F50705F506F605F705F5070605F70605F70605F7F605F706F607F\
60607F6F607F6F607F706F806F607F706F806F6F80707080706F80707080706F80
7F708F7070807F708F70708F7F7F8F7F708F7F7F8F7F708F7F7F907F7F8F7F7F907F7F8F807F90\
7F7F908080907F7F9080809F80809080809F8080908F809F80809F8F8F9F80809F
8F8F9F8F809F8F8F9F8080908F8F9F8F8F908F8F9F8F8F908F8F9F8F8F908F8F908F8F908F8F90\
8F8F908F8F908F8F8F8F90908F8F8F8F90908F8F8F8F90908F8F8F8F908F8F8F8F
90908F8F8F8F8F908F8F8F8090908F8F908090908F8F8F8090908F8F8F808F90808F8F808F9080\
8F8F808F8F808F8F7F8F8F808F8F7F8F8F7F80807F8F8F7F80807F8F8F7F808070
8F807F808070808070808070808070808070808070807F6F80807080806F8080707F7F6F80806F\
7F7F607F7F6F7F70607F7F607070607F706070705F70705F70705F70705F6F6F50
7070506F6F506F6F5060604F6F604F5F5F405F503F4F4F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F30402F2F4030304F30304F303F50303F503F405F
30405F3F4F603F40603F4F603F4F6040506F3F4F6040506F40506F40506F40506F40506F40506F\
4F5F7040506F4F5F704F5F704F5F704F5F704F60704F5F7050607F4F5F7050607F
4F607050607F4F607050607F50607050607F4F607050607F506070506F7F506070506F7F506070\
506F7F302F3F3F304F3F304F403F50403F504F405F4F405F4F40604F4060504F60
504F60504F6F504F6F5F506F5F506F5F50705F506F605F705F5F70605F70605F70605F7F605F70\
6F607F60607F6F607F6F607F6F6F806F607F706F806F6F80706F80706F80707080
706F8070708F706F807F708F70708F7F708F70708F7F7F8F7F708F7F7F907F708F7F7F907F7F8F\
807F907F7F908080907F7F90808090807F9080809F80809080809F8080908F8F9F
80809F8F8F9F8080908F8F9F8080908F8F9F8080908F8F9F808F908F8F90808F908F8F908F8F90\
8F8F90808F8F8F8F908F8F8F8F90908F8F8F8F8F908F8F8F8F90908F8F8F8F908F
8F8F8F8F908F8F8F808F908F8F8F808F908F8F8F808F908F8F8F808F90808F8F808F8F808F8F80\
8F8F80808F7F8F8F80808F7F8F8F7F80807F808F7F80807F808F7F80807080807F
80807080807F8080708080708080708080707F7F6F8080707F7F6F8080707F7F6F7F7F6F7F7F60\
7F7F6F7070607F7F607070607F7F6070705F70705F6F705F70705F6F6F506F6F50
6F6F506F6F5060604F6F6F4F60604F6060405050304F4F303F3F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F3F30304F30304F303F4F303F4F3F405F304050
3F405F3F405F3F4F603F4F60404F6F3F4F6040506F40506F40506F40506F4F5F7040506F4F5F70\
40506F4F5F704F5F704F5F704F5F7050607F4F5F7050607F4F5F7050607F506070
50607F4F607050607F50607050607F506070506F7F506070506F7F506070506F7F506070506F7F\
302F3F3F30403F304F403F50403F504F405F4F405F4F40604F4060504F6F504F60
504F6F504F6F5F50705F506F5F50705F506F605F705F5F70605F70605F7060607F605F7F6F607F\
60607F6F6F806F607F6F6F806F607F706F806F6F80707080706F8070708F707080
70708F7070807F708F70708F7F7F8F70708F7F7F907F7F8F7F7F907F7F8F807F907F7F90808090\
7F7F9080809080809080809F80809080809F80809F808F9F80809F8F8F9F80809F
8F8F9F80809F8F8F9F808F9F8F8F9F808F908F8F9F808F908F8F9F808F908F8F9F8F8F908F8F90\
808F908F90908F8F908F8F908F8F8F8F90908F8F8F8F90908F8F8F8F90908F8F8F
8F908F8F8F8F8F908F8F908F8F908F8F90808F908F8F90808F908F8F8F808F908F8F8F808F9080\
808F808F9080808F808F8F80808F7F8F8F80808F7F8F8F7F80807F8F8F7F80807F
808F7F80807080807F8080708080707F80708080708080708080707F7F6F8080707F7F6F7F7F6F\
7F7F607F7F6F707F607F7F6070705F707F6070705F70705F6F6F5070705F6F6F50
6F6F506F6F506F6F50606F4F6F6F4F5F5F405F5F3F4F4F30404030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F3F2F2F4030304F2F3040303F4F303F50304050
303F503F405F30405F3F4F603F40603F4F603F4F6040506F3F4F6040506F40506F40507040506F\
4F5F7040506F4F5F70405F6F4F5F704F5F704F60704F5F704F60704F5F7050607F
4F5F705060704F607050607F4F60705060704F607050607F4F6070506070506070506F7F4F6070\
3030403030403F3F4F3F304F403F50403F504F405F4F405F4F4F604F4F60504F6F
504F6F50506F504F6F5F507050506F5F50705F5070605F705F5F70605F7F605F7060607F605F7F\
6F607F60607F6F6F806F607F6F6F806F6F80706F806F6F80707080706F8070708F
706F8070708F70708F7F708F70708F7F7F8F70708F7F7F9070708F7F7F907F7F8F7F7F907F7F90\
7F80907F7F908080907F7F9080809080809080809F80809080809F808090808F9F
808090808F9F808090808F9F808090808F9F8080908F8F9F808090808F90808F908F8F90808090\
8F8F90808F908F8F90808F8F8F8F90808F8F8F8F90808F8F8F8F8F808F8F8F908F
808F8F8F908F808F8F8F908F808F808F908F808F808F908F808F808F9080808F808F9080808F80\
8F8F80808F7F8F8F80808F7F808F80808F7F808F7F80807F808F7F80807F808F7F
80807080807F8080708080707F80708080707F80708080707F7F6F7F80707F7F6F7F806F707F60\
7F7F6F707F607F7F6070706070706070705F70705F6F6F5F6F705F6F6F506F7050
606F506F6F5060604F6F6F5060604F60604050503F4F4F303F3F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003030402F3040303F4F30304F303F50303F50
3F405F30405F3F4F603F40603F4F603F4F6040506F3F4F6040506F40506F40506F40506F4F5F70\
405F6F4F5F704F5F704F5F704F5F704F60704F5F7050607F4F607050607F4F6070
50607F4F607050607F4F607050607F50607050607F506070506F7F506070506F7F5060703F3040\
3030403F3F4F3F3F4F404050403F504F405F4F405F504F6F504F6F504F6F504F6F
5F507050506F5F50705F506F5F5F705F5070605F705F5F7060607F605F7F60607F60607F6F6080\
60607F6F6F806F607F6F6F806F6F80706F806F6F8070708F70708F70708F70708F
7F708F70708F7F7F8F70708F7F7F907F7F8F7F7F907F7F8F7F7F907F7F907F80907F7F9080809F\
7F7F9080809F7F809080809F80809080809F808090808F9F80809F808F9F80809F
8F8F9F80809F8F8F9F80809F8F8F9F808F9F8F8F9F808F908F8F9F808F908F8F9F808F908F8F90\
808F908F8F90808F908F9090808F908F8F90808F8F8F9090808F8F8F9090808F8F
8F908F8F8F8F8F908F808F8F8F908F8F908F8F908F808F808F908F8F8F808F908F808F808F9080\
808F808F9080808F808F9080808F7F8F8F80808F7F8F8F80808F7F808F7F80807F
808F7F80807F808F7F808070808F7F8080708080707F80708080707F80708080707F7F6F7F806F\
7F7F6F7F7F6F707F607F7F6F707060707F6070705F7070606F705F70705F6F6F50
6F705F6F6F506F6F50606F4F6F6F505F60405F5F3F4F4F30404030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F3F2F30402F304030304F2F304F303F50
303F503F405F30405F3F4F603F40603F4F603F4F603F4F603F4F6040506F40506F40506F40506F\
4F5F7040506F4F5F704F5F704F5F704F5F704F5F704F5F705060704F5F704F6070
4F5F705060704F5F705060704F60705060704F60705060704F607050607F4F607030303F30303F\
3F30403F30403F3F4F3F3F5040405F40405F4F40604F4060504F6F4F4F60504F6F
504F6F5F507050506F5F50705F50705F5F705F50705F5F705F5F7060607F605F7060607F60607F\
6F608060607F6F6F806F6F806F6F806F6F80706F8F6F6F8070708F706F8070708F
70708F70708F70708F7F7F8F70708F7F7F9070708F7F7F907F7F8F7F7F907F7F8F7F7F907F7F90\
7F80907F7F907F80907F7F908080907F80908080907F809080809F7F809080809F
808090808F9F808090808F9F808090808F9F808090808F90808090808F90808090808F90808F90\
808F9080808F808F90808F8F808F90808F8F808F90808F8F8F8F8F808F8F808F8F
808F8F8F908F808F808F908F808F808F908F808F808F908F808F808F9080808F80808F80808F80\
808F80808F7F808F80808F7F808F80808F7F808F7F808F7F808F7F80807F808F7F
80807F808F7F7F807080807F7F807080807F7F80708080707F80707F80707F7F6F7F806F7F7F6F\
7F7F6F707F60707F60707060707F606F705F7070606F705F70705F6F6F506F705F
6F6F506F6F5060604F606F5060604F60604F50503F4F4F303F402F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F403030402F3040303F50303F50
303F50303F503F405F30405F3F4F603F4F603F4F603F4F6040506F3F4F6040506F40506F40506F\
40506F4F5F704F5F704F5F704F5F704F60704F5F704F60704F5F7050607F4F6070
5060704F5F7050607F4F60705060704F607050607F50607050607F50607030303F30303F3F3040\
303040403F4F3F3F5040405F40405F4F40604F4060504F6F4F4F6F50506F504F6F
50507050506F5F5F705F50705F5F705F5070605F7F5F5F7060607F605F7F60607F60607F6F6080\
60607F6F6F806F6F806F6F806F6F8070708F6F6F8070708F70708F70708F70708F
70708F70708F7F7F9070708F7F7F9070708F7F7F907F7F8F7F7F907F7F8F7F80907F7F907F8090\
7F7F908080907F7F908080907F809080809F7F809080809F7F8090808F9F808090
808F9F808090808F9F808090808F9F808090808F9F808090808F9F808090808F9F808F90808F90\
808F90808F90808F90808F90808F8F8F8F90808F8F808F90808F8F8F9090808F8F
8F8F8F808F8F8F908F808F8F8F908F808F808F908F808F808F908F808F808F908F808F80809080\
808F80809080808F80808F80808F7F808F80808F7F808F8080807F808F7F808F7F
808F7F7F807F808F7F808070808F7F7F8070808F7F7F80708080707F80707F80707F7F6F7F806F\
707F6F7F7F6F707F60707F60707060707F6070705F7070606F705F70705F6F6F50
6F705F606F506F6F50606F4F606F4F5F60405F5F404F4F30404030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F402F2F3F2F30402F304F303F50
2F3F4F303F50303F5030405F30405F3F4F603F40603F4F603F4F603F4F603F4F6040506F3F506F\
40506F40506F4F5F7040506F4F5F704F5F704F5F704F5F704F5F704F5F704F6070
4F5F704F60704F5F705060704F5F705060704F60705060704F60705060702F2F3F303040303040\
3F304F3F304F403F503F3F5040405F40405F4F4F604F4060504F6F4F4F6F50506F
504F6F50507050506F5F5F705050705F5F705F5F70605F7F5F5F7060607F605F7F60607F60607F\
6F608060607F6F6F806F6F806F6F806F6F8070708F6F6F8070708F6F708F70708F
70708F70708F70708F707F8F70708F707F8F70708F7F7F9070708F7F7F90707F8F7F7F90707F8F\
7F7F907F7F907F80907F7F907F80907F7F907F80907F80907F80907F809080809F
7F80908080907F8090808F9F7F8090808F907F8090808F907F8090808F907F8090808F907F8090\
808F9080808F808F907F808F808F90808F8F808F9080808F808F8F808F8F808F8F
808F8F808F8F808F80808F8F808F80808F8F808F80808F8F808F80808F80808F80808F80808F80\
808F80808F7F808F807F807F808F807F807F808F7F7F807F808F7F7F807F808F7F
7F807080807F7F807080807F7F80707F80707F80707F80707F80707F80707F7F6F7F8070707F6F\
7F7F6F707F60707F6F707060707F606F705F7070606F705F6F705F6F6F506F705F
606F506F6F50606F50606F505F60405F604050503F4F50303F402F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003030402F304030304F2F304F
303F50303F5030405F30405F3F4F603040603F4F603F4F6040506F3F4F6040506F3F506040506F\
40506F4F5F7040506F4F5F704F5F704F5F704F5F705060704F5F705060704F5F70
50607F4F60705060704F607050607F50607050607F506070506F7F5060703030403030403F3F4F\
3F304F3F3F503F3F5040405F40405F4F40604F4060504F6F504F6F50507050506F
5F50705050705F5F705F50705F5F7F5F5F70605F7F5F5F7F60607F60607F60608060607F6F6F80\
6060806F6F806F6F806F708F6F6F8070708F6F6F8F70708F70708F70708F70708F
707F8F70708F707F8F70708F7F7F90707F8F7F7F90707F8F7F7F90707F8F7F7F90707F8F7F8090\
7F7F907F80907F7F907F80907F80907F80907F809080809F7F8090808F9F7F8090
808F9F7F8090808F9F7F8090808F9F808090808F9F7F8090808F9F808090808F907F8090808F90\
808F90808F90808F90808F90808F90808F90808F8F808F90808F8F808F8F808F8F
80908F808F8F808F8F808F8F80908F808F8F808F8F808F8080908F808F80808F80808F80809080\
808F80808F80808F80808F80808F7F808F807F8F7F808F807F8F7F808F7F7F807F
808F7F7F807F808F7F7F8070808F7F7F807080807F7F80708080707F80707F80707F7F6F7F8070\
707F6F7F7F6F707F60707F6F707F60707F606F705F707F606F705F6F705F6F6F5F
6F705F606F506F6F5060604F606F4F5F5F405F5F404F4F30404F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F30402F2F4030304F
2F304F303F502F3F5030405F30405F3F406030405F3F4F603F4F603F4F603F4F6040506F3F4F60\
40506F40506F40506F40506F4F5F70405F6F4F5F704F5F704F60704F5F704F6070
4F5F705060704F5F705060704F60705060704F6070506070506070506F7F2F2F3F303040303040\
3F3F4F3F304F403F50403F5F4040604040604F4F604F4F6050506F4F4F6F505070
50506F5F50705050705F5F705F5F705F5F7F5F5F7060607F5F5F7F60607F60607F60608060607F\
6F6F806060806F6F806F6F806F708F6F6F806F708F6F708F70708F6F708F70708F
6F708F70708F6F708F70708F70708F707F8F70708F707F8F707F8F7F7F90707F8F7F7F90707F8F\
7F7F90707F8F7F80907F7F907F80907F7F907F80907F80907F80907F80907F8090
7F8090808F907F80907F80907F8090808F907F8090808F907F8090808F907F8090808F907F8090\
808F907F808F808F907F808F808F907F808F808F8F7F8F8F808F8F7F808F808F8F
7F8F8F808F8F7F8F80808F8F7F8F80808F8F7F8F80808F807F8F80808F807F8F80808F807F8F80\
808F807F8F7F808F807F8F7F808F807F807F808F7F7F807F808F7F7F807F7F8F7F
7F807F7F8F7F7F80707F807F7F80707F807F7F80707F80707F80707F8070707F6F7F8070707F6F\
7F7F6F707F60707F6F707060707F606F70607070606F705F6F705F6F6F5F6F705F
606F506F6F50606F50606F4F5F60405F604050503F4F503040402F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030304F2F304F
303F4F2F3F4F304050303F5030405F30405F3F4F6030405F3F4F603F4F6040506F3F4F6040506F\
3F506040506F40506F4F5F6F40506F4F5F704F5F704F60704F5F705060704F5F70
5060704F607050607F4F60705060704F6070506F7F506070506F7F5060703030403030403F304F\
30304F3F3F503F3F5F40406040405F4F4F604F40604F4F6F4F4F6F505070505070
5F50705050705F5F7F5F5F705F5F7F5F5F7060607F5F5F7F60607F60607F606F806060806F6F80\
606F806F6F8F6F6F806F6F8F6F6F8070708F6F708F70708F6F708F70708F70708F
70708F6F708F707F8F70708F707F8F70708F707F90707F8F707F90707F8F7F7F90707F8F7F7F90\
707F8F7F8090707F907F80907F7F907F80907F80907F80907F80907F809F7F8090
7F8F907F8090808F9F7F8090808F907F8090808F907F8090808F907F8090808F907F8090808F90\
7F808F808F907F8090808F907F808F808F907F8F8F808F907F8F8F808F907F8F8F
808F8F7F8F8F808F8F808F8F808F8F7F8F80808F8F7F8F80808F8F7F8F80808F8F7F8F80808F80\
7F8F80808F807F8F80808F807F807F808F807F8F7F808F807F807F808F7F7F807F
7F8F7F7F807F7F8F7F7F80707F807F7F80707F807F7F80707F8070707F707F8070707F6F7F8070\
707F6F7F806F707F6F707F6F6F7060707F606F70606F70606F705F6F705F6F705F
6F705F606F50606F505F604F5F604F505F40505F404F5030404F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F304F
2F30402F3F4F2F3F503040502F3F5030405F30405F3F4F6030405F3F4F603F4F603F4F603F4F60\
40506F3F506040506F40506F405F6F40506F4F5F70405F6F4F5F704F5F704F6070
4F5F6F4F60704F5F705060704F5F705060704F60705060704F60705060702F2F4030304F30304F\
3F3F503F3050403F5F3F3F5F4040604040604F4F6F4F4F60504F6F4F4F6F505070
5050705050705050705F5F7F505F705F5F7F5F5F7F60607F5F5F7F60608060607F606F8060607F\
606F80606F806F6F80606F806F6F8F6F6F806F708F6F6F806F708F6F708F70708F
6F708F70708F6F708F707F8F6F708F707F8F70708F707F8F70708F707F8F707F8F707F90707F8F\
707F90707F8F7F8090707F8F7F8090707F8F7F8090707F8F7F80907F80907F8090
7F80907F80907F80907F80907F808F7F80907F80907F80907F808F7F80907F808F7F8F907F808F\
7F8F907F808F7F8F907F808F7F8F8F7F808F7F8F907F808F7F8F8F7F808F808F8F
7F80807F8F8F7F8F80808F8F7F80807F8F8F7F8F80808F8F7F80807F8F807F80807F8F807F807F\
7F8F807F807F7F8F807F807F7F8F807F807F7F8F807F807F7F807F7F807F7F807F
7080707F807F7080707F807F7080707F80707080707F8070707F6F708070707F6F708070707F6F\
707F6F707F6F707F6F6F7060707F606F70606F7F606F705F6F705F60705F6F705F
606F50606F505F604F5F604F505F40505F404F503F4F503F3F402F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
30304F2F304F303F502F3F5030405F30405F30405F30405F3F4F603F4F603F4F603F4F6040506F\
3F4F6040506F40506F405F6F40506F4F5F70405F6F4F5F704F5F704F60704F5F70
5060704F5F705060704F6070506070506070506070506070506F7F50607030304F30304F3F3F50\
3F30503F3F5F3F3F5F4040604040604F4F604F4F6050506F4F4F6F505070505070
5F5F7F505F705F5F7F5F5F705F607F5F5F7F60607F5F607F606080606080606F806060806F6F80\
606F806F6F8F606F806F708F6F6F806F708F6F6F8F70708F6F708F70708F6F708F
707F9070708F707F8F6F708F707F90707F8F707F90707F8F707F90707F8F707F90707F8F7F8090\
707F8F7F8090707F8F7F8090707F907F80907080907F80907F80907F80907F8090
7F8F907F80907F8F907F80907F8F907F80907F8F907F80907F8F907F80907F8F907F808F808F90\
7F808F7F8F907F808F808F907F808F7F8F907F808F808F907F8F8F808F8F7F8F8F
808F8F7F8F8F808F8F7F8F8F808F8F7F8F80808F8F7F8F80808F8F7F8F807F8F807F8F80808F80\
7F8F807F8F807F8080808F807F8F807F8F807F807F7F8F807F807F7F8F7F7F807F
7F8F7F7F807F7F807F7080707F807F7F80707F807F7080707F80707080707F8070707F6F708070\
707F6F707F6F707F6F707F6F6F7F60707F606F7060707F606F705F6F706060705F
6F705F606F50606F505F604F5F604F505F40505F404F5030404F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F304F2F304F2F3F502F3F5030405F2F3F5030405F30405F3F4F6030405F3F4F603F4F60\
3F50603F4F6040506F40506F405F6F40506F4F5F70405F6F4F5F70405F6F4F5F70
4F5F6F4F60704F5F6F4F60704F5F705060704F60705060704F60705060702F2F4030304F30304F\
3F3F503F30503F3F5F3F3F5F4040604040604F4F6F4F4F6F5050704F506F505070
5050705F5F7F5050705F5F7F5F5F7F5F607F5F5F7F6060805F607F6060805F607F606F80606080\
606F80606F806F6F80606F806F708F606F806F708F6F708F6F708F6F708F6F708F
6F708F707F8F6F708F707F8F6F708F707F8F6F708F707F8F707F8F707F90707F8F707F8F707F8F\
707F90707F8F707F90707F8F708090707F8F708090707F8F7F809070808F7F8090
7080907F809070808F7F80907F808F7F809070808F7F80907F808F7F8F9070808F7F80907F808F\
7F8F907F808F7F8F8F7F808F7F8F8F7F808F7F8F8F7F808F7F8F8F7F808F7F8F8F
7F808F7F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F807F80807F8F807F80807F8F807F8080\
7F8F8070807F7F8F807F807F7F8F8070807F7F8F7F70807F7F8F7F70807F7F807F
70807F7F807F7080707F807F7080707F807F707F70708070707F70708070707F6F707F6F707F6F\
707F6F6F7F60707F6F6F7060707F6F6F70606F7F606F705F6F706060705F6F705F
606F50606F505F604F5F604F505F40505F404F503F4F503F404F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F304F2F3F50303F502F3F5030405F30405F30405F30405F3F4F603F4F603F4F60\
3F4F6040506F3F506040506F40506F4F5F6F405F6F4F5F70405F6F4F5F704F5F70
4F60704F5F6F4F60704F5F705060704F60705060704F60705060704F607030304F30304F3F3050\
3030503F3F5F3F3F5F4040604040604F4F6F4F4F6F4F4F6F4F4F6F505070505070
5F5F7F505F705F5F7F5F5F7F5F607F5F5F7F6060805F607F6060805F607F606F80606080606F80\
606F806F6F8F606F806F708F606F806F708F6F708F6F708F6F708F707F8F6F708F
707F8F6F708F707F906F708F707F8F6F708F707F90707F8F707F90707F8F707F90707F8F707F90\
707F8F708090707F8F708090707F8F7F80907080907F809070808F7F8090708090
7F80907080907F8F907F80907F809070808F7F8F907F808F7F8F9070808F7F8F907F808F7F8F90\
7F808F7F8F907F808F7F8F8F7F808F7F8F907F808F7F8F8F7F808F7F8F8F7F808F
7F8F8F7F808F7F8F8F7F8F8F7F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F807F80807F8F80\
7F80807F8F807F80807F8F8070807F7F8F807F807F7F8F8070807F7F8F7F70807F
7F807F70807F7F807F7080707F807F7080707F807F707F70708070707F6F708070707F6F707F6F\
6F7F6F707F6F6F7F6F707F6F6F70606F7F6F6F70606F7F606F705F6F706060705F
60705F5F6F50606F505F604F5F604F505F40505F404F5030404F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F30402F304F2F3F502F3F5030405F2F405030405F30405F3F4F6030405F\
3F4F603F4F604050603F4F6040506F40506F405F6F40506F4F5F6F405F6F4F5F70
405F6F4F5F704F5F6F4F60704F5F6F4F60704F5F6F5060704F5F6F5060702F2F4F303050303050\
3F3F5F3F3F5F3F405F3F3F5F4040604040604F4F6F4F4F6F4F50704F5070505F70
5050705F5F7F505F7F5F5F7F505F7F5F607F5F5F7F5F60805F607F6060805F607F606F805F607F\
606F80606F80606F80606F806F6F8F606F806F708F606F806F708F6F708F6F708F
6F708F6F708F6F708F707F8F6F708F707F8F6F708F707F8F6F708F707F8F6F7F8F707F906F7F8F\
707F90707F8F708090707F8F708090707F8F708090707F8F70809070808F7F8090
70808F70809070808F7F809070808F70809070808F7F809070808F7F808F70808F7F808F70808F\
7F808F70808F7F808F70808F7F808F70808F7F8F8F7080807F808F7080807F8F8F
7080807F8F8F7080807F8F8F7080807F8F8F7080807F8F807080807F80807080807F8F8070807F\
7F808070807F7F808070807F7F808070807F7F808070807F70807F70807F70807F
70807070807F70807070807F707F70708070707F707080706F7F6F707F706F7F6F707F6F6F7F6F\
6F7F6F6F70606F7F6F6F70606F7F606F70606F706060705F6F706060705F6F705F
606F50606F505F604F5F604F505F40505F404F503F4F503F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003030402F3F5030405F2F3F5030405F30405F3F4F6030405F3F4F60\
3F4F603F50603F4F6040506F40506F40506F40506F4F5F70405F6F4F5F70405F6F
4F5F704F5F6F4F60704F5F6F4F60704F5F705060704F60705060704F60703030503030503F3F5F\
3F3F5F3F3F5F3F3F5F4040604040604F4F6F404F6F4F50704F5070505070505070
5F5F7F505F7F5F607F5F5F7F5F60805F607F5F60805F607F606F805F6080606F80606080606F80\
606F80606F80606F806F708F606F806F708F60708F6F708F6F708F6F708F6F708F
707F906F708F707F906F708F707F906F7F8F707F906F7F8F707F90707F8F707F90707F8F708090\
707F8F708090707F8F708090708090708090707F8F7F809070809070809070808F
7F809070809070809070808F7F809070808F7F809070808F7F809070808F7F809070808F7F8F90\
70808F7F808F70808F7F8F8F70808F7F8F8F70808F7F8F8F70808F7F8F8F708080
7F8F8F70808F7F8F8F7080807F8F8F7080807F8F8F7080807F8F8F7080807F8F807080807F8F80\
7080807F8F807080807F8F8070807F7F8F8070807F7F8F8070807F7F807F70807F
7F807F70807F70807F70807070807F708070708070707F70708070707F707080706F7F6F707F70\
6F7F6F6F7F6F6F70606F7F6F6F70606F7F606070606F7F6060705F6F706060705F
60705F5F6F505F6F5050604F50604F4F5F404F503F404F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F30402F3F5030405F2F3F5030405F30405F30405F30405F\
3F4F603F4F5F3F4F603F4F6040506F3F506040506F40506F4F5F6F40506F4F5F6F
405F6F4F5F70405F6F4F5F704F5F6F4F60704F5F6F4F60704F5F6F5060702F2F4F303050303050\
3F3F5F303F5F3F405F3F405F4040604040604F4F6F4F4F6F5050704F5070505F70
505F705F5F7F505F7F5F5F7F505F7F5F607F5F5F7F5F60805F607F6060805F607F606F805F6080\
606F80606F80606F80606F8060708F606F806F708F6070806F708F60708F6F708F
6F708F6F7F8F6F708F6F7F8F6F708F707F8F6F708F6F7F8F6F7F8F707F906F7F8F707F906F7F8F\
707F906F7F8F707F906F7F8F7080906F7F8F708090707F8F7080906F7F8F708090
707F8F7080906F7F8F70808F707F8F70808F6F7F8F70808F70808F70808F707F8F70808F70808F\
70808F70808070808F70808F70808F70808070808F70808070808F70808070808F
70808070808F708080708080708080708080708080708080708080708080708080708080708080\
70808070807F70808070807F70808070807F70807F70807F70807F70807F70807F
70807F70807F6F7F7070807F6F7F707080706F7F70707F706F7F6F707F706F7F6F6F7F6F6F7F6F\
6F7F6F6F70606F7F6F6070606F7F606070606F706060705F607060606F5F60705F
5F6F505F6F5050604F50604F4F5F404F5F40405030405030000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F30402F3F5030405F30405030405F30405F3F4F60\
304F5F3F4F603F4F603F50603F4F6040506F40506F405F6F40506F4F5F70405F6F
4F5F70405F6F4F5F704F5F6F4F60704F5F6F4F60704F5F6F4F60704F606F303050303050303F5F\
303F503F405F3F3F5F4040604040604F4F6F404F6F4F50704F506F505F70505070
505F7F505F7F5F607F505F7F5F607F5F5F7F5F60805F607F5F60805F607F606F805F6080606F80\
5F6F80606F8F606F8060708F606F806F708F60708F6F708F60708F6F7F8F6F708F
6F7F8F6F708F6F7F906F708F6F7F906F7F8F707F906F7F8F707F906F7F8F707F906F7F8F707F90\
6F7F8F7080906F7F8F7080906F7F8F7080906F7F8F7080906F7F8F708090707F8F
7080906F7F8F7080906F7F8F70808F6F7F8F70808F6F808F70808F6F7F8F70808F70808F70808F\
6F808F70808F70808F70808F6F808070808F70808070808F70808070808F708080
70808F70808070808F708080708080708080708080708080708080708080708080708080708080\
70807F70808070807F70808070807F70808070807F70808070807F70807F70807F
70807F70807F70807F70807070807F6F7F707080706F7F70707F706F7F6F707F706F7F6F6F7F6F\
6F706F6F7F6F6F70606F7F6F6070606F7F6060706060706060705F607060606F5F
606F5F5F60505F6050505F4F505F404F503F4F503F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F30402F3F5030405F2F405030405F30405F\
304F5F30405F3F4F603F4F603F50603F4F6040506F3F506040506F40506F405F6F
40506F405F6F405F6F4F5F6F405F6F4F5F6F4F5F6F4F60704F5F6F4F60702F2F4F303050303050\
3F3F5F3F3F5F3F40603F405F4040604040604F4F6F404F6F4F50704F5070505F7F
505070505F7F505F7F5F607F505F7F5F607F5F607F5F60805F607F5F60805F607F606F805F6080\
606F805F6F80606F80606F80606F80606F8060708F606F806F708F60708F6F708F
60708F6F708F6F708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F7F8F6F7F8F6F708F\
6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F70808F
6F7F8F6F7F8F6F7F8F6F808F6F7F806F7F8F6F7F8F6F808F6F7F806F808F6F7F806F808F6F7F80\
6F808F6F7F8070808F6F7F806F80806F7F8070808F6F7F806F80806F7F80708080
6F7F807080806F7F807080806F7F807080806F7F807080806F7F7F7080806F7F7F7080806F7F7F\
7080806F7F7F7080806F7F7F70807F6F7F7F70807F6F7F7F70807F6F7F7F70807F
6F7F7070807F6F7F7070807F6F7F706F7F706F7F706F7F706F7F6F6F7F706F7F6F6F7F6F60706F\
6F7F6F6070606F7F6F607060607060607060607060606F5F60705F5F6F5F606F5F
5F60505F6F5050604F50604F4F50404F503F404F30000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003030402F3F5030405F30405F3F4F60\
30405F3F4F603F4F603F50603F4F6040506F3F5060405F6F40506F405F6F40506F
4F5F70405F6F4F5F6F405F6F4F5F704F5F6F4F60704F5F6F5060704F607030304F3030503F3F5F\
303F5F3F40603F405F404F604040604F4F6F404F6F4F50704F5070505F70505070
505F7F505F7F5F607F505F7F5F60805F607F5F60805F607F5F6F805F6080606F805F6080606F80\
606F80606F8F606F8060708F606F8060708F60708F6F708F60708F6F7F8F6F708F
6F7F906F708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F\
6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F808F6F7F8F6F808F6F7F8F
70808F6F7F8F6F808F6F7F8F70808F6F7F8F6F808F6F7F8F70808F6F7F8F6F808F6F7F8070808F\
6F7F806F808F6F7F8070808F6F7F806F808F6F7F8070808F6F80807080806F7F80
7080806F80807080806F7F807080806F80807080806F7F807080806F80807080806F7F7F708080\
6F807F7080806F7F7F7080806F807F70807F6F7F7F70807F6F807F70807F6F7F7F
70807F6F807F70807F6F7F7070807F6F7F706F80706F7F706F80706F7F6F6F7F706F7F6F6F7F6F\
6F7F6F6F7F6F6070606F7F6F607060607060607060607060606F5F60705F5F6F50
5F6F5F5F605050604F505F40505F404F503F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F30402F3F50304050304050\
30405F30405F3F4F603F4F5F3F50603F4F604050603F506040506F405060405F6F
40506F405F6F40506F4F5F6F405F6F4F5F6F405F6F4F5F704F5F6F4F60702F2F40303050303050\
3F3F5F303F5F3F40603F4060404F6F404F604F4F6F4F4F6F4F50704F5070505F70
505070505F7F505F7F5F607F505F7F5F608050607F5F60805F607F5F6F805F607F5F6F805F6080\
606F805F6F80606F80606F8060708F606F8060708F60708F6F708F6070806F708F
60708F6F708F6070806F708F60708F6F7F8F6070806F7F8F60708F6F7F8F6070806F7F8F60708F\
6F7F8F6070806F7F8F607F8F6F7F8F6070806F7F8F607F806F7F8F607F806F7F8F
6F7F806F7F8F607F806F7F8F6F7F806F7F8F607F806F7F8F607F806F7F8F607F806F7F80607F80\
6F7F8F607F806F7F806F7F806F8080607F806F7F806F7F806F8080607F806F7F80
6F7F806F80806F7F7F6F7F806F7F7F6F80806F7F7F6F80806F7F7F6F80806F7F7F6F807F6F7F7F\
6F807F6F7F7F6F807F6F7F7F6F807F6F7F7F6F807F6F7F7F6F807F6F7F706F807F
6F7F706F807F6F7F706F7F706F7F706F7F70607F6F6F7F70607F6F6F7F7060706F6F7F6F60706F\
6F7F6F60706060706F6070606070605F70606070605F6F5F60705F5F6F505F6F50
5060505F6050505F4F505F4F4F50404F503F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F30402F3F50304050\
3040503F4F5F30405F3F4F603F4F604050603F506040506F405060405F6F40506F
405F6F40506F4F5F6F405F6F4F5F6F405F6F4F60704F5F6F4F60704F5F6F30304F303050303F5F\
303F5F3F40603F4060404F604040604F4F6F404F6F4F50704F5070505F704F5070
505F7F505F705F607F505F7F5F608050607F5F60805F607F5F6F805F6080606F805F6F80606F80\
5F6F8060708F606F8060708F606F8060708F60708F60708F6070806F708F60708F
6F708F60708F6F7F8F60708F6F7F8F60708F6F7F8F60708F6F7F8F6070806F7F8F60708F6F7F8F\
6070806F7F8F607F8F6F7F8F607F806F7F8F607F8F6F7F8F607F806F7F8F607F80
6F7F8F607F806F7F8F607F806F7F8F607F806F7F8F607F806F7F8F607F806F7F8F607F806F7F80\
607F806F7F8F607F806F7F80607F806F8080607F806F7F80607F806F80806F7F80
6F7F80607F7F6F80806F7F806F8080607F7F6F80806F7F7F6F8080607F7F6F80806F7F7F6F807F\
607F7F6F807F6F7F7F6F807F6F7F7F6F807F6F7F7F6F807F6F7F706F807F6F7F70
6F807F6F7F706F807F6F7F706F7F706F7F706F7F706F7F706F7F70607F6F6F7F70607F6F6F7F6F\
60706F6F7F6F60706060706F6070606070605F705F60705F5F6F5F5F6F5F5F6F50
5F6F5050604F50604F4F5F404F5F40000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F303F2F304F\
303F502F3F5030405F30405F3F4F5F3F4F5F3F50603F4F6040506F3F506040506F
3F5060405F6F40506F4F5F6F40506F4F5F6F405F6F4F5F704F5F6F4F60702F2F40303050303050\
3F3F5F303F5F3F40603F4060404F6F404F6F4F506F404F6F4F50704F50704F5F70
4F5070505F7F505F7F50607F505F7F5F607F50607F5F608050607F5F60805F60805F6F805F6080\
5F6F805F6F80606F8F5F6F806070805F6F8060708F5F6F8060708F60708060708F
60708060708F60708060708F60708060708F60708060708F60708060708F607080607F8F607080\
607080607080607F8F607080607F80607080607F8F607080607F80607080607F8F
607080607F80607F80607F80607080607F80607F80607F80607080607F80607F80607F80607080\
607F80607F80607F8060707F607F80607F7F607F80607F7F607F80607F7F607F80
607F7F607F80607F7F607F80607F7F607F7F607F7F6F7F7F607F7F607F7F607F7F6F7F7F607F7F\
607F7F607F7F6F7F7F607F70607F7F607F706F7F7F607F706F7F7F607F706F7F7F
607F706F7F70607F706F7F7060706F6F7F70607F6F6F7F7060706F607F6F60706F607F6F607060\
60706F60706060706F5F70606070605F6F5F5F70605F6F5F5F6F5F5060505F6F50
50604F50604F4F5F4F4F5F4040503F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F3040\
2F3F50304050303F503F405F30405F3F4F603F4F5F4050603F506040506F405060
405F6F40506F4F5F6F405F6F4F5F704F5F6F4F5F704F5F6F4F60704F5F6F30304F30304F3F3F5F\
303F5F3F40603F4060404F6F404F6F404F6F404F6F4F50704F50704F50704F5070
505F7F505F7F50607F505F7F5F608050607F5F60805F607F5F6F805F60805F6F805F6F80606F8F\
5F6F80606F8F5F6F8060708F606F8060708F606F8060708F60708060708F607080
60708F60708060708F607080607F8F607080607F8F607080607F8F607080607F8F607080607F8F\
607080607F8F607080607F8F607080607F8F607080607F8F607F80607F8F607F80
6F7F8F607F80607F8F607F80607F8F607F80607F80607F80607F80607F80607F80607F80607F80\
607F80607F80607F80607F80607F80607F80607F7F607F80607F80607F80607F7F
607F80607F7F607F80607F7F6F7F80607F7F607F80607F7F6F7F80607F7F607F7F607F7F6F7F7F\
607F7F607F7F607F7F6F7F7F607F7F607F7F607F706F7F7F607F706F7F7F607F70
6F7F7F607F706F7F70607F706F7F70607F706F7F70607F6F6F7F70607F6F607F7060706F607F6F\
60706F607F6F60706060706F5F70606070605F6F5F5F705F5F6F5F5F6F5F506050
50605050604F50604F4F5F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F303F2F304F303F50303F5030405F3040503F4F5F3F4F5F3F50603F4F6040506F
3F5060405F6F405060405F6F40506F4F5F6F405F6F4F5F6F4F5F6F4F60702F2F4030304F303050\
3F4060303F5F3F40603F4060404F6F404F6F404F6F404F6F4F507040506F4F5070
4F5070505F7F4F5F7050607F505F7F5F608050607F5F608050607F5F6F805F60805F6F805F6F80\
5F6F805F6F805F6F805F6F806070805F6F806070805F6F8060708F5F6F80607080
5F708060708F5F708060708060708060708F5F70806070805F708060708F5F70806070805F7080\
6070805F70806070805F7080607F805F7080607F805F7080607F805F7080607F80
607080607F805F7080607F805F7080607F805F707F607F805F7080607F805F707F607F805F707F\
607F805F707F607F805F707F607F805F707F607F805F707F607F805F707F607F7F
5F707F607F7F5F707F607F7F5F707F607F7F5F707F607F7F60707F607F7F5F707F607F7F60707F\
607F7F5F7070607F7F607070607F7F5F7070607F70607070607F7F607070607F70
607070607F70607070607F7060706F607F7060706F607F7060706F607F6F5F706F60706F60706F\
60706F5F70606070605F6F605F70605F6F5F5F6F5F506F5F5F6F5F506050506050
4F604F50604F4F5F404F5F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F303F2F3F4F303F50303F5030405F30405F3F4F5F3F4F5F4050603F5060
40506F3F5060405F6F40506F405F6F40506F4F5F6F405F6F4F5F6F4F5F6F30304030304F303F50\
303F5F3F40603F4060404F6F3F4F60404F6F404F6F4F507040506F4F50704F5070
4F5F7F4F5F7050607F505F7F50607F50607F5F6F805F60805F6F805F60805F6F805F6F805F6F80\
5F6F805F6F805F6F805F6F805F6F806070805F6F806070805F6F8060708F5F7080
60708F5F708060708F60708060708F5F708060708F6070806070805F7080607F8F5F7080607080\
5F7080607F805F7080607F805F7080607F805F7080607F805F7080607F805F7080
607F805F7080607F805F7080607F805F707F607F805F707F607F805F707F607F805F707F607F80\
5F707F607F805F707F5F7F7F5F707F607F805F707F607F7F5F707F607F805F707F
607F7F5F707F607F7F5F707F607F7F5F707F607F7F5F707F607F7F5F707F607F7F5F707F607F7F\
5F7070607F7F5F7070607F7F5F7070607F7F607070607F705F7070607F7F607070
607F70607070607F70607070607F705F706F607F7060706F607F6F5F706F607F6F5F706F60706F\
5F706060706F5F70605F70605F6F5F5F70605F6F5F5F6F5F506050506F5050604F
50604F4F5F4F4F5F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F303F2F304F303F502F3F503040503040503F405F3F405F3F4F60
3F4F603F50603F506040506F3F506040506F405060405F6F4050604F5F6F2F2F3F30304F30304F\
3F3F5F3040603F40603F4060404F6F3F4F6F404F6F404F6F40507040506F4F5070
4F50704F5F704F5F7050607F505F7F50607F50607F5F608050607F5F608050607F5F6F8050607F\
5F6F80506F7F5F6F80506F7F5F6F805F6F805F6F805F6F805F6F805F6F805F7080
5F6F805F70805F6F806070805F6F805F70805F70806070805F70805F70805F70806070805F7080\
5F70805F70806070805F707F5F70805F707F5F70805F707F5F70805F707F5F7080
506F7F5F707F506F7F5F707F506F7F5F707F506F7F5F707F506F7F5F707F506F7F5F707F506F70\
5F707F506F705F707F506F705F707F506F705F707F506F705F707F5070705F707F
5070705F707F5F70705F707F5F70705F707F5F70705F707F5F70705F707F5F70705F707F5F7070\
5F707F5F70705F707F5F70705F70705F70706070705F70705F70705F7070607070
5F706F6070705F706F6070705F706F5F706F5F706F60706F5F706F5F706F5F706F5F706F5F7060\
5F706F5F70605F7060506F5F5F6F5F506F5F5F6F5F5060505060504F604F506050
4F5F4F4F5F4F405040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F303F2F3F4F303F50303F503F405F3040503F4F5F3F4F5F
3F50603F4F6040506F3F5060405F6F405060405F6F40506F4F5F6F405F6F3030403030403F3F50\
303F5F3F40603F4060404F6F3F4F6F404F6F404F6F4F507040506F4F50704F5070
4F5F7F4F5F70505F7F505F7F50608050607F5F608050607F5F6F805060805F6F8050607F5F6F80\
5F6F805F6F805F6F805F6F805F6F805F6F805F6F805F70805F6F805F70805F6F80
6070805F70806070805F70806070805F70806070805F70806070805F70806070805F7080607080\
5F70805F70805F707F5F70805F707F5F70805F707F5F70805F707F5F707F506F7F
5F707F506F7F5F707F506F7F5F707F506F7F5F707F506F705F707F506F7F50707F506F705F707F\
506F7050707F506F705F707F506F7050707F506F705F707F506F705F707F506F70
5F707F5070705F707F5070705F707F5F70705F707F5F70705F707F5F70705F707F5F70705F7F7F\
5F70705F707F5F7070607F7F5F70705F70705F7070607F705F70706070705F7070
607F705F70706070705F706F607F705F706F5F706F5F706F60706F5F706F5F706F5F706060706F\
5F70605F706F5F6F605F7060506F5F5F6F5F50605F506F5F5060505060504F5F4F
4F604F405F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F403F2F304F303F4F303F4F304050303F503F405F
30405F3F4F603F4F5F3F50603F506040506F3F506040506F405060405F6F2F2F3F30304030304F\
3F3F5F303F5F3F40603F4060404F6F3F4F60404F6F404F6F4F507040506F4F5070
4F50704F5F7F4F5F70505F7F505F7F50607F50607F50608050607F5F6F8050607F5F6F8050607F\
5F6F8050607F5F6F80506F7F5F6F80506F7F5F6F80506F7F5F6F80506F7F5F6F80
5F6F805F70805F6F7F5F70805F6F805F70805F6F7F5F70805F6F7F5F7080506F7F5F7080506F7F\
5F7080506F7F5F707F506F7F5F707F506F7F506F7F506F7050707F506F70506F7F
506F70506F7F4F6F70506F704F6F70506F704F6F70506F704F6F70506F704F6070506F704F6F70\
506F704F6070506F704F6F70506F704F6F6F506F704F6F70506F704F6F70506F70
506F70506F704F6F70506F70506F70507070506F70507070506F70507070506F70507070506F70\
5F7070506F705F70705070705F70705070705F70705F70705F70705F706F5F7070
5F706F5F70705F6F6F5F706F5F706F5F706F506F6F5F706F5F6F6F5F706F506F605F706F5F6F60\
5F7060506F605F6F60506F5F506F5F506050506F5F4F60505060504F5F4F4F5F4F
405F40000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F304F303F50303F4F304050304050
3F4F5F30405F3F4F603F4F604050603F506040506F405060405F6F40506030303F303040303F4F\
303F503F40603F40603F4F6F3F4F60404F6F404F6F40507040506F4F50704F5070
4F5F704F5F7050607F4F5F7F50607F50607F50608050607F506F8050607F5F6F8050607F5F6F80\
50607F5F6F80506F7F5F6F80506F7F5F6F805F6F805F6F80506F7F5F70805F6F80
5F70805F6F805F70805F70805F7080506F7F5F7080506F7F5F7080506F7F5F7080506F7F50707F\
506F7F50707F506F7F506F7F506F70506F7F506F70506F7F4F6F70506F7F4F6F70
506F704F6F70506F704F6F70506F704F6070506F704F60704F6F704F606F4F6F704F606F4F6F70\
4F606F4F6F704F606F4F6F704F606F4F6F704F606F4F6F704F606F506F704F6F70
506F704F6F6F506F704F6F70506F704F6F6F506F70506F70506F704F6F6F507070506F70507070\
506F705F70705070705F70705070705F70705070705F707050706F5F70705F706F
5F707050706F5F70705F706F5F706F506F6F5F706F506F6F5F706F506F605F706F5F6F605F7060\
506F605F7060506F60506F6050605F506F5F5060505060504F5F4F4F6050405F4F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F3040303F4F2F3F4F304050
303F5030405F30405F3F4F603F4F5F3F50603F4F604050603F50604050602F2F303030402F3040\
303F50303F5F3F40603F40603F4F6F3F4F60404F6F3F4F6F40507040506F4F5F70
4050704F5F704F5F7050607F4F5F7050607F4F5F7F50607F50607F50607F50607F506F8050607F\
506F7F50607F5F6F8050607F506F80506F7F5F6F80506F7F5F6F80506F7F5F7080
506F7F5F6F80506F7F5F7080506F7F506F7F506F7F506F7F506F70506F7F506F70506F7F4F6070\
506F704F6070506F704F60704F6F704F60704F6F704F606F4F6F704F606F4F6F70
40606F4F60704F606F4F607040606F4F606F40606F4F606F40606F4F606F40606F4F606F40606F\
40606F40606F4F606F40606F40606F40606F4F606F40606F4F606F40606F4F606F
40606F4F606F40606F4F6F6F40606F4F606F4F606F4F6F6F4F606F4F6F6F4F606F506F704F6F6F\
506F70506F6F506F70506F6F506F70506F6F507070506F6F507070506F6F5F7070
506F6F50706F506F6F5F706F506F6F506F6F506F605F706F506F60506F60506F605F6F60506F60\
506F60506F5F506F6050605F506F5F4F60505060504F5F4F4F6050405F403F503F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F304F303F4F303F4F
3040503040503F405F30405F3F4F603F4F603F50603F50604050603F506040506F2F2F3F30304F\
3030503F3F5F303F5F3F4F603F4F60404F6F404F6F40507040506F4F5070405070
4F5F7F4F5F70505F7F4F5F7050607F4F607F50607F50607F506F8050607F506F8050607F5F6F80\
506F7F5F6F80506F7F5F6F80506F7F5F6F80506F7F5F7080506F805F7080506F7F
5F7080506F7F5F7080506F7F5F707F506F7F506F7F506F7F506F7F506F70506F7F4F6F70506F7F\
4F6F70506F704F6070506F704F60704F6F704F606F4F6F704F606F4F607040606F
4F6F7040606F4F606F40606F4F606F40606F40606F40606F4F606F40606F40606F405F6F40606F\
405F6040606F405F6040606F40606F40606F40606F40606F40606F40606F40606F
4F606F40606F4F606F40606F4F606F40606F4F606F40606F4F6F704F606F4F6F704F606F506F70\
4F6F6F506F704F6F6F506F70506F6F506F70506F6F507070506F6F507070506F6F
5F7070506F6F50706F506F6F5F706F506F6F50706F506F605F706F506F605F6F6F506F605F6F60\
506F60506F60506F5F506F6050605F506F5F4F60504F6050405F4F405040000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F3040303F4F
2F3F4F304050303F5030405F30405F3F4F5F304F5F3F50603F4F604050603F4F602F2F3F2F3040\
303F503030503F3F5F30405F3F4F603F4F60404F6F404F6F40507040506F4F5070
4050704F5F7F4F5F704F5F7F4F5F7050607F4F5F7050607F4F607F50607F4F607F506F7F50607F\
506F8050607F506F7F506F7F506F80506F7F506F80506F7F506F80506F7F506F7F
506F7F506F7F506F70506F7F506F70506F7F4F6070506F704F6070506F704F60704F6F704F6070\
4F6F704F606F4F607040606F4F607040606F4F606F40606F40606F405F60405F60
3F5F60405F603F505F3F5F603F505F3F505F3F505F806F707F6F6F8070707F6F6F80706F80706F\
80706F80706F80706F80706F3F50603F505F3F5F603F5F60405F603F5F60405F6F
3F5F6040606F3F5F6040606F405F6040606F405F6040606F40606F4F606F40606F4F606F4F606F\
4F6F6F4F606F4F6F6F4F606F506F6F4F606F506F6F4F6F6F506F6F4F6F6F506F6F
506F6F506F6F4F6F6F506F6F506F60506F6F506F60506F60506F60506F60506F60506F604F605F\
506F604F605F506F5F4F605F50605F4F60504F6050405F4F3F5040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F3040
303F4F2F3F4F30405030405030405F30405F3F4F603F4F5F3F50603F4F604050603F5060303040\
30304F303F5F303F5F3F40603F4060404F6F3F4F6F40507040506F4F5F70405070
4F5F704F5F704F5F7F4F5F704F607F4F5F7050607F4F607F50607F4F607F506F7F50607F506F7F\
50607F506F80506F7F506F80506F7F5F6F80506F7F506F7F506F7F506F7F506F7F
506F7F4F6F70506F7F4F6F70506F7F4F6070506F7F4F6070506F704F60704F6F704F60704F6F70\
4F606F4F6F7040606F4F606F40606F40606F405F60405F603F505F7F607070606F
7F6F707F6F7080707F806F70807070806F70807070807070807070807070807F70807070807F70\
8070708F7F70807070807F7080706F807F6F80706F80706F7F70608070607F705F
3F50603F505F405F603F5F60405F6F3F5F6040606F405F6040606F405F604F606F40606F4F606F\
40606F4F6F6F4F606F4F6F6F4F606F4F6F6F4F606F4F6F6F4F606F506F6F4F6F6F
506F6F4F6F6F506F6F506F6F506F6F4F6F60506F6F506F60506F60506F60506F604F6F60506F60\
4F605F506F5F4F605F4F605F4F60504F605F405F4F405040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F30402F304F303F502F3F4F3040503040503F4F5F304F5F3F4F603F4F5F3F50602F2F3F\
30304F303050303F5F303F5F3F40603F4060404F6F3F4F6F40506F40506F405F70
40506F4F5F704050704F5F70405F704F5F704F5F704F607F4F5F704F607F4F607F50607F4F607F\
506F7F50607F506F7F4F607F506F7F4F607F506F7F4F6070506F7F4F6070506F7F
4F60704F6F704F60704F6F704F60704F60704F606F4F6F7040606F4F607040606F4F606F405F6F\
40606F405F60405F603F505F3F505F6F5F6F70606F705F6F7060707060707F6F70
7F6F70806F7F7F6F7080707F7F6F7080707F80707080707F807070807070807070807F70807070\
807F708070708F7F7080706F807F7080706F807F6F7F706080706F7F70607F7060
7F6F5F7F70607F6F5F7F6F5F3F505F3F5F603F5F60405F603F5F60405F60405F6040606F405F60\
40606F405F6040606F405F604F606F4060604F606F4060604F606F4F60604F6F6F
4F60604F6F6F4F6060506F6F4F6060506F604F6060506F604F60604F6F604F605F4F6F604F605F\
4F605F4F605F4F605F405F504F6050405F4F3F50402F4030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F3040303F4F303F5030405F3040503F4F5F3F4F5F3F50603F4F5F405060\
2F304F303050303F503F405F3F405F3F4F603F4F6040506F40506F405070405070
4F5F70405F704F5F70405F704F607F4F5F704F607F4F5F7050607F4F607F50607F4F607F506F7F\
50607F506F7F50607F506F7F506F7F506F7F4F6070506F7F4F6070506F7F4F6070
506F7F4F60704F6F704F60704F6F704F60704F6F704F606F4F6F7040606F4F606F405F60406060\
3F5F606F506F6F5F6F706070705F6F7060707060707F6F7F7F6F707F6F7F7F6F70
80707F80707F80707F80707F8F7F7F80707F8F7F7F80707F8F7F7F807F7F8F7F7F807F708F7F7F\
807F708F7F7F807F708F807F807F70807F70807F6F807F7080706F807F6F7F7060
807F6F7F70607F7060706F5F7F705F706F503F50603F5060405F603F5F60405F6F3F5F6040606F\
405F6040606F405F604F606F4060604F606F4060604F6F6F4F606F4F6F6F4F6060
506F6F4F6F60506F6F4F6060506F6F4F6F60506F604F6060506F604F60604F6F604F605F4F6F60\
4F605F4F605F4F60504F605F405F4F405040304030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F30402F3040303F503040503F4F5F304F5F3F4F5F3F4F5F\
30304F2F304F303F50303F503F405F30405F3F4F603F4F6040507040506F405070
40506F4F5F704050704F5F70405F704F5F704F5F704F607F4F5F704F607F4F607050607F4F607F\
506F7F4F6070506F7F4F6070506F7F4F6070506F7F4F6070506F704F60704F6F70
4F60704F6F704F606F4F60704F606F4F607040606F4F606F405F604060603F5F5F3F5F605F4F5F\
60506F60506F6F5F706F5F70706070706070706F707060707F6F7F7F6F7080707F
7F707F80708080707F807F8080707F8F7F8080707F8F7F7F807F7F8F7F7F807F7F8F7F7F807F7F\
8F807F807F708F807F807F708F7F70807F70807F70807F6F807F6F7F706F807F6F
7F70607F7F607F70607F705F706F5F706F506F604F6F604F605F403F5F603F505F3F5F603F5F60\
405F603F5F60405F60405F6040606F405F604F606F4060604F606F4060604F606F
4F60604F6F6F4F60604F6F604F60604F6F604F60604F60604F605F4F606040605F4F605F40605F\
4F605F405F504F6050405F4F3F5040304030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F3040303F4F303F4F30405F304F5F3F4F60\
3F4F5F303050303050303F5F303F5F3F40603F4060404F6F3F4F6F40507040506F
405F704050704F5F70405F704F5F704F5F704F607F4F5F704F607F4F607050607F4F607F506F7F\
4F6070506F7F4F6070506F7F4F6070506F7F4F6070506F704F6070506F704F6070
4F6F704F60704F6F704F606F4F607040606F4F606F405F60405F603F5F5F5F4F605F4F6060506F\
60506F6F60706F607070607F7060707F6F7F706F7F7F6F7F7F6F7F80708080707F
807F808070808F7F80807F808F7F80807F808F80808F7F808F8080807F7F8F80808F7F7F8F8080\
8F807F8F80808F807F8F807F807F708F807F807F70807F70807F6F807F707F7F6F
807F6F7F70607F7F6F7F706070705F706F50706F506F604F6F604F605F403F5F603F5F60405F60\
3F5F60405F6F405F60406060405F604F606F4060604F606F4060604F606F406060
4F60604060604F6F6F4F60604F6F604F60604F6F604F60604F606040605F4F605F40605F4F605F\
405F504F6050405F503F504F30403F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F30402F3040303F50303F50\
304F5F2F2F4F3030502F3050303F5F303F5F3F40603F40603F4F6F3F4F6F40506F
40506F405F7040506F4F5F70405F704F5F70405F704F60704F5F704F607F4F60704F607F4F6070\
4F607F4F60704F607F4F60704F60704F60704F6F704F60704F60704F60704F6F70
4F606F4F607040606F4F606F405F60405F603F5F60405F604F3F505040605F4F6060506F60506F\
6F5F706F5F7070607F6F6070706F7F706F7F7F6F7F7F6F7F7F70807F707F807F80
807080807F80807F808F808F807F808F8080807F808F8080807F808F80808F80808F808080807F\
8F80808F807F8F808080807F8F807F807F7080807F807F70807F707F7F6F807F70
7F7F6F7F7F6F7F70607F7060706F5F706F5F6F6F506F6F5060604F60604F5F5F3F3F5F603F505F\
3F5F603F5F60405F603F5F60405F60405F60406060405F60406060405F60406060
405F604060604060604F606040605F4F606040605F4F6060405F5F40605F405F5040605F405F50\
405F503F504F3F504F2F4030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F3040303F4F\
303F4F3040502F304F303F50303F503F405F30405F3F4F6F3F4F6F40506F3F506F
405F7040506F4F5F70405F704F60704F5F704F607F4F5F704F607F4F60704F607F4F607050607F\
4F60704F607F4F6070506F7F4F60704F6F704F6070506F704F60704F6F704F606F
4F6F704F606F4F606F4060604F606F4060604F3F504F405F5F4F605F4F6060506F60506F6F5F70\
6F5F7070607F70607F7F6F80706F7F7F70807F70808070807F7080807F8F807F80
8F808F807F8F8F808F80808F8F808F8F80808F808F8080808F808F8F80808F80808F80808F8F80\
8F80808F8F8080807F8F8F8080807F8F807F80807080807F8080708080707F7F70
8080707F7F6F7F7F6F7F70607F7F6070705F70705F6F6F506F6F5060604F5F5F403F50603F5F60\
3F5F60405F603F5F6040606F405F60406060405F60406060405F60406060405F60
4F60604060604F60604060604F60604060604F6060405F5F4F605F405F5F40605F405F50405F50\
3F5F4F3F504F304F3F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F30402F3040303F4F2F304F303F5F303F5030405F3040603F4F6F3F4F603F506F
3F506F40507040506F405F70405F704F5F70405F704F60704F5F704F607F4F5F704F60704F6070\
4F60704F60704F60704F60704F60704F606F4F60704F606F4F6F704F606F4F606F
4F606F4F606F4060604F6060405F60403F4F4F3F50504060504F605F506F5F506F605F70605F70\
6F607F6F607F706F80706F807F7080706F807F70807F7080807F8F7F7080807F8F
807F8F80809080808F8F808F80808F8F808F8080808F808F8080808F808F8080808F8080808080\
8F8F808080808F8F8080807F80807F80807F80807F807F7080807F7F7F70808070
7F7F6F7F80707F7F6F7F7F6F7070607070606F6F506F6F5060604F606F4F5F5F3F50503F3F505F\
3F5F603F505F3F5F603F5F60405F603F5F5F405F603F5F5F405F603F5F5F405F60
405F5F406060405F5F406060405F5F406060405F5F405F5F405F50405F503F504F405F503F504F\
3F504F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F30403030402F305030405F30405F3F4F60304F603F4F6F3F4F6F
40506F3F506F405F7040506F4F5F70405F704F60704F5F704F60704F5F704F60704F60704F6070\
4F60704F60704F60704F60704F60704F6F704F60704F6F704F606F4F6F704F606F
4F606F4060604F606F4060603F3F4F403F50504060504F605F506F5F506F605F70605F706F607F\
6F607F706F80706F8070708F7070807F7F8F7F708F7F7F8F7F7F8F808090807F8F
8080908080908F8F908080908F8F9080808F8F8F8F80808F8F8F8F8080808F8F8F808F808F8F8F\
808F808F8F80808F80808F8080807F808F8080807F80807F7F807080807F7F7F70
7F80707F7F6F7F806F707F6F707F6070705F7070606F6F5F6F6F5060604F5F60404F503F3F5F60\
3F5060405F603F5F60405F603F5F60405F60405F60405F60405F60406060405F60
406060405F5F406060405F5F40605F405F5F40605F405F50405F503F5050405F503F504F3F504F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F30402F3F5030405F30405F304F60304F603F4F6F
3F4F603F506F3F506F405F6F40506F405F70405F704F6070405F704F6070405F704F6070405F6F\
4F60704F60704F60704F606F4F60704F606F4F60704F606F4F606F4F606F4F606F
405F604F60604060603030403F304F40405F4F4060504F6F504F6F5F50705F5F7060607F60607F\
6F6F806F6F80707080706F8070708F70708F7F7F8F707F8F7F7F907F7F90808090
7F8090808090808090808F90808090808F9080808F808F9080808F808F8F808080808F8F808080\
808F80808F80808F807F807F8080807F807F80807F7F7F707F807F7F7F707F8070
7F7F6F7F7F70707F6F707F6F6F70607070606F705F6F705F606F50606F50505F404F503F3F505F\
3F5F603F5060405F603F505F405F603F5F5F405F603F5F5F405F603F5F5F405F60
3F5F5F405F5F405F5F405F5F3F5F50405F5F3F5F50405F503F504F3F504F30504F3F504F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030304F2F3F5F3040603040603F4F60304F60
3F506F3F506F40506F3F506F405F70405F704F5F70405F704F60704F60704F60704F60704F6070\
4F60704F60704F6070506F704F6070506F704F606F506F704F606F4F606F4F606F
506F6F4F606F506F6F30304F403F5F4F405F504F6F50506F5F5F705F5F7060607F60607F6F6F80\
6F6F80707080706F8070708F70708F7F7F90707F8F7F80907F7F907F809F7F8090
808F9F808F9F808F9F808F9F808F9F808F90808F90808F90809090808F8F808F90808F8F80908F\
808F8F808F8F808F80808F80808F80808F807F807F808F7F7F807F7F807F7F8070
7F807F7F7F707F7F70707F6F707F6F6F70606F70606F6F5F6F705F606F505F604F4F503F405F60\
3F5F60405F603F5F60405F60405F60405F603F5F5F405F60405F5F405F603F5F5F
405F60405F5F405F5F3F5F5F405F5F3F5F50405F503F504F3F50503F504F3F504F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F304F2F3F5F30406030405F304F60
304F603F506F3F4F603F506F3F506F405F70405F6F405F70405F6F4F6070405F6F4F60704F6070\
4F60704F606F4F60704F606F4F6F704F606F4F60704F606F4F6F6F4F606F506F6F
4F606F506F6F2F2F403F304F3F3F504F4F604F4F6F5F50705F5070605F7F5F5F7F60607F60607F\
6F6F806F6F8070708F70708F707F8F70708F7F7F90707F907F80907F80907F809F
7F8090808F9F7F809F808F9F7F8F90808F9F7F8F90808F907F8F8F808F907F8F8F808F8F7F8F80\
808F8F7F8F80808F807F8F80808F807F807F7F8F7F7F807F7F807F7080707F807F
707F707F8070707F6F707F6F6F70606F70606F6F5F6F705F606F50606F50505F404F503F3F5060\
405F603F505F405F603F505F405F603F505F405F5F3F505F405F5F3F505F405F5F
3F5F5F405F5F3F50503F5F5F3F50503F505030504F3F504F304F40000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000303F502F405F304F60304F60
3F4F60304F603F506F3F506F405F6F40506F405F70405F6F4F5F70405F6F4F60704F5F704F6070\
4F606F4F60704F60704F60704F6070506F704F606F506F704F606F506F704F606F
506F6F4F606F30304F3F3F5040405F4F40605050705F50705F5F7F5F5F7F60608060607F6F6F80\
6F6F8070708F6F708F707F8F70708F7F7F90707F907F8090707F907F80907F8090
7F809F7F8090808F9F7F80907F8F9F7F8090808F907F8F907F8F907F8F8F808F8F7F8F8F7F8F8F\
7F8F80808F8F7F8F807F8F807F807F7F8F807F807F7F807F70807F7F807F708070
7F807F707F707080706F7F6F707F6F6F70606F706060705F606F5F5F604F5F604F4F503F405F60\
3F505F405F603F505F405F603F505F405F603F505F405F5F3F505F405F5F3F5F5F
405F5F3F5050405F5F3F50503F5050304F40000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F302F3F502F405F304F60
30405F3F4F603F4F603F506F3F506040506F40506F405F70405F6F405F6F405F6F4F6070405F6F\
4F606F405F6F4F60704F606F4F60704F606F506F704F606F506F6F4F606F506F6F
4F606F506F6F30304F3F3F5F403F5F4F4F6F4F4F6F5F5F70505F705F607F5F607F606F80606080\
6F6F806F6F8070708F6F708F70708F6F708F707F8F6F7F8F707F90707F8F7F8090
707F8F7F80907080907F809070808F7F809070808F7F8F9070808F7F808F70808F7F8F8F708080\
7F8F807080807F8F8070807F7F808070807F7F807F70807F70807F707F7070807F
707F70707F706F7F6F707F6F6F70606F706060705F60705F5F60505F604F505F404F503F3F505F\
3F5F603F505F3F5F5F3F505F3F5F5F3F505F3F5F5F3F50503F505F3F50503F5F5F
3F50503F5050304F4F304F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F303F303F5030405F
3F4F60304F603F506F3F506040506F40506F405F70405F6F405F70405F6F4F6070405F6F4F6070\
4F606F4F60704F606F506F704F6070506F704F606F506F704F606F506F70506F6F
5F6F70506F6F3030503F3F5F4F4F6F4F4F6F505070505F705F607F5F607F606F80606F806F708F\
6F6F806F708F6F708F707F8F6F708F707F8F6F7F8F707F90707F8F708090707F8F
7080907080907F809070808F7F809070808F7F809070808F7F8F9070808F7F8F8F7080807F8F8F\
7080807F8F807080807F8F8070808070808070807F70807F70807F70807F707F70
70807F6F7F70707F706F7F6F6F7F6F6F70606F7060606F5F606F5F50604F505F4040503F405F60\
3F505F405F603F505F405F5F3F505F3F5F5F3F505F3F5F5F3F50503F5F5F30504F
3F504F304040000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F3F303F50
2F3F5F304F5F30405F3F50603F506040506F3F506F405F6F40506F405F6F405F6F405F6F405F6F\
4F6070405F6F4F60704F606F506F704F606F506F704F606F506F7050606F506F70
506F6F5F70703030503F3F5F4040604F4F6F4F506F505F7F5F5F7F5F60805F607F606F80606F80\
60708F606F806F708F6F708F6F7F8F6F708F6F7F8F6F7F8F707F8F6F7F8F707F8F
6F7F8F70808F6F7F8F70808F6F7F8F70808F6F7F8070808F70808070808F6F8080708080708080\
7080806F7F7F7080806F807F7080806F7F7F70807F6F7F7F70807F6F7F706F7F70
6F7F706F7F7060706F6F7F6F60706060706060705F606F5F50605050604F4F503F3F505F3F505F\
3F5F5F3F50503F505F3F50503F505F3050503F505030505030504F2F4040304040
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F3040
303F50303F503F4F60304F603F50603F5060405F6F40506F405F6F405F6F4F6070405F6F4F6070\
405F6F4F60704F6070506F704F606F506F704F606F506F70506F705F7070506F70
5F70705F6F703F3F503F3F5F404F6F404F6F505F70505F7F5F60805F607F606F805F6F80606F80\
606F806F708F6070806F7F8F60708F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F
6F7F8F6F7F806F808F6F7F806F808F6F7F8070808F6F7F806F80806F7F807080806F7F80708080\
6F7F807080806F7F807080806F7F7F7080806F7F7F6F807F6F7F706F807F6F7F70
6F7F7060706F6F7F7060706F60706F6070606070605F6F5F5F6F504F5F4F4F5F403F505F3F505F\
3F50503F505F3F50503F50503050503F5050304F4F304F402F403F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F40303F502F3F5030405F30405F3F4F603F506040506F3F5060405F6F405F6F405F6F405F6F\
4F606F4F606F4F60704F606F4F60704F606F506F704F606F506F70506F6F506F70
506F6F5F6F7030304F3F40603F40604F506F4F5070505F7F505F7F5F607F5F607F5F6F805F6F80\
606F80606F8060708F60708060708F60708060708F607080607080607080607F80
607080607F80607080607F80607080607F806070806F7F8060707F607F80607F7F6F7F80607F7F\
6F7F7F607F7F6F7F7F607F7F6F7F7F607F7F6F7F7F607F706F7F70607F706F7F70
60706F607F6F60706F60706F5F6F605F70605F6F5F5F6F5F4F604F4F5F4F304F503F505F304F50\
3F5050304F503F5050304F4F304F4F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F3040303F50303F503F4F5F304F5F3F50603F5060405F6F3F506F405F6F405F6F4F6070\
405F6F4F60704F606F506F704F6070506F704F6070506F70506F70506F70506F70
5F70705F6F70303F503F405F404F6F40506F4F5F704F5F7050607F50607F5F6F805F6F80607080\
5F6F806070805F6F8060708F6070806070805F70806070805F70806070805F7080
607F80607080607F805F707F607F8060707F607F805F707F607F8060707F607F8060707F607F80\
60707F607F7F60707F607F7F607F7F607F7F6070706F7F7F607F70607F70607070
607F7060706F60706F5F706060706F5F6F605F6F5F5060505060504F5F403F5050304F503F5050\
304F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000303F50303F5030405F30405F3F4F603F50603F506F3F5060405F6F405F6F\
405F6F405F6F4F606F4F606F4F60704F606F506F704F606F506F7050606F506F70
506F705F70703030403F405F3F4F6F40507040506F4F5F704F5F7F50607F50607F5F6F80506F7F\
5F6F80506F7F5F6F805F6F7F5F70805F6F7F5F70805F6F7F5F7080506F7F5F707F
5F6F7F5F707F506F7F5F707F5F707F5F707F506F7F5F707F50707F5F707F5070705F707F5F7070\
5F707F5F70705F707F5F70705F70705F70705F70705F70706070705F706F5F7070
5F706F60706F5F6F605F7060506F605F6F5F5060505060504F5F4F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000303F50303F5030405F304F5F3F50603F50603F506F3F5060405F6F\
405060405F6F405F6F4F60704F606F506F704F6070506F704F6070506F70506F70
5F70705F6F7030304F303F5F404F6F404F6F4F5F704F5F7050607F50607F506F8050607F506F80\
506F7F5F6F80506F7F5F6F7F506F7F5F7080506F7F506F7F506F70506F7F506F70
506F704F6070506F704F6070506F704F6070506F704F6F70506F704F6F70506F70506F70507070\
506F705F707F506F705F7070506F705F70705F70705F70705F706F5F70705F706F
5F706F506F6F5F706F506F60506F6050605F50605F4F5F4F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F3F5030405F30405F3F4F60304F603F50603F5060\
4050603F5060405F6F405F6F4F606F4F606F4F60704F606F506F70506070506F70
506F705F6F70506F70303F5F30405F40506F40506F4F5F704F5F704F607F4F6070506F7F4F6070\
506F7F50607F506F7F4F60704F6F704F60704F607040606F40606F405F6F40606F
3F5F60405F6F3F5F60405F6F3F5F60405F603F5F6040606F405F6040606F405F6F4F606F40606F\
4F606F4F606F506F704F6F6F506F70506F6F506F70506F6F506F6F506F6F506F6F
506F60506F6F506F60506F604F605F4F605F405F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F304F304050304F5F3F5060304F603F506F\
3F5060405F6F405F6F4F606F405F6F4F60704F606F506F70506070506F70506F70
5F6F70506F705F6F7F3030503F40603F4F6F405F70405F704F607F4F5F704F607F4F6070506F7F\
4F60704F6F704F60704F607040606F40606F405F6040606F3F505F3F4F502F4040
5F6F7F50607F50607F40506F40506F3F4F60303F4F30404F3F4F5F3F505F3F5F603F5F6040606F\
405F6040606F40606F4F6F6F4F606F506F6F4F6F6F506F704F6F6F506F6F4F6F60
506F6F506F60506F604F605F506F5F4F60503F5040000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030304F2F3F4F304F5F304F5F\
3F50603F50603F50603F5060405F6F405F6F4F606F4F606F4F60704F606F506070
4F6070506F70506070506F70303F5F3F4F603F4F6F405F70405F704F6070405F704F60704F6070\
4F607040606F40606F405F60405F60305050606F5F7F7F7080707F6F5F4F60503F
50607F50607F4F5F704F5F7040506F40506F3F4F603040602F405F3040502F4050304F5F30505F\
3F5F603F5060405F603F5F6040606F405F604F606F4F60604F6F6F4F60604F6F60
4F60604F60604F605F4F605F405F503F5040000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F3040304050\
304F5F3F50603F50603F50603F5060405F6F405F60405F6F405F6F4F60704F6070
506F704F6070506F70506F7030304F30405F3F506F3F506F405F70405F704F6070405F6F4F6070\
405F6F4F606F405F60405F5F405F505F6F5F707F708F8080706F7F7F6F60705F50
604F3F50607050607F4F5F704F5F7040506F40506F3F4F603F4F6030405F304F60304F5F305060\
30505F3F5F603F5060405F603F5F6040606F4060604F606F4060604F606F4F6060
4F606040605F4F605F405F503F5040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30304F2F3F4F304F5F304F5F3F50603F505F3F50603F5060405F6F405F6F4F606F
4F5F6F4F60704F606F506F704F6070303F503040603F506F3F506F405F70405F6F4F6070405F6F\
40606F405F604F60604F605F5F6F606F706F7F80807F7F80605F707F606070605F
60504050607F4F60704F6070405F704F5F7040506F40506F3F4F603F5060304F603F5060304F60\
3F506030505F3F50603F50603F5F603F5F60405F603F5F60406060405F60406060
405F5F405F5F3F5F503F5040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F3040304050304F5F3F5060304F5F3F50603F5060405F6F405F60
4F606F4F606F4F60704F606F506F7050607030405F304F60405F6F405F6F405F6F405F6F4F6F70\
4F606F4F6F6F506F6F607070607F707F80807F8F8F8080905F50707F6F6F70605F
7F605F60504050607F4F60705060704F5F704F5F70405F6F405F7040506F405F6F3F50603F5F6F\
3F50603F5F603F5060405F603F5F60405F60405F60406060405F60406060405F5F
405F5F3F50503F504F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F3F50304F5F304F5F3F506030505F3F5060
3F5060405F6F405F60405F6F405F6F4F60704F606F303F50304F5F405F6F405F6F4F606F4F606F\
506F70506F6F5F70706070706F7F80707F807F8F90707F8F5F5F7F70606F7F6F6F
70605F6F60504F5F704F60704F5F704F6070405F6F405F6F405F6F405F6F3F5060405F6F3F5060\
3F5F603F50603F5F603F5060405F603F505F405F603F5F5F405F5F3F50503F5050
30504F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000304F5F304F5F
3F5060304F5F3F50603F5060405F60405F604F606F405F6F4F6070304F5F3F5F6F405F6F4F6070\
4F606F5F70705F7070607F7F6F7F7F70808F70808F707F905F607F806F70807070
7F7F6F6F6F5F4F60704F5F704F6070405F6F4F6070405F6F405F6F405F6F405F6F3F5F60405F6F\
3F5060405F603F50603F5F603F505F3F5F603F505F3F5F5F3F5050304F4F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000304F502F4F50304F5F304F5F3F505F3F505F405F603F5060405F6F405F6F3050603F5060\
40606F4F606F506F705F6F7060707F6070806F7F80606F805F607F3F4F50708080
60707050605F405F6F4F606F405F6F405F6F405F6F405F6F3F5060405F6F3F5060405F603F5060\
3F5F603F505F3F5F6030505F3F505F30505030504F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000304F50304F5F304F5F3F50603F505F3F50603F5060405F60\
304F5F3F5060405F604F6F704F6070506F7F506F70506F7F4F60704F60703F5F6F
406060405F6F40606F405F6F40606F405F6F405F6F3F5F60405F603F50603F5F603F505F3F5F60\
3F505F3F505F3050503F505F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000304F502F4050304F50304F50\
304F5F304F503F505F304F5F3F505F3F505F3F50603F505F3F50603F505F3F5060
3F505F3F50603F505F3F50603F505F3F506030505F3F505F304F503F505F304F50305050304F50\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F4050304F50304F50304F502F4F50304F50304F50
304F502F4F50304F502F4F50304F502F4F50304F50000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[12]:=",
  ImageSize->{288, 286.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P00017PP10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`NJ000?L@1a0740L@1a0740L@1a06L02@L000L7P@03L@09Jg5[LF]aJg5[00]a00U`LG1a
L75`LG002G401G1aJg5[06402@L000L7N000>g40L@1a0740L@1a0740L@1a0740LF]aJg5[LF]aJg5[
LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg40L@1a05d02@L000L7M@00175[LF/;L@09Jg5`LG1aL75`00Ya
01b:L8Y`RW2:L8Y`RW6:LHYaRW6:L8Y`RW1aL75`0W5K00T700071g@004Ea0740L@1aJg5[LF]aJg5[
LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[RV^:JhY[LF]aJhY[RV^:JhY[RV]aJg5[LF]a0740E`091`00
1`Ma0008LF]aJg5/LF/<L@0_RW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW6:L8X02G5C00T700071g0004ma075[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[
RV^:JhY[RV^:L8Y`RW2:L8Y`RV^:JhY[RV^:Jg5[LF]aJg5[L@1a05402@L000L7K@03L@03Jg5[009a
04F:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW6:LHYaRW6:LHX017401F]aJg5[04d02@L000L7JP00GG40L@1aJg5[LF]aJg5[LF]aJg5[LF]aJg5[
RV^:JhY[RV^:Jg5[RV^:JhY`RV^:JhY[RV^:JhY`RW2:L8Y`RW2:L8Y`RV^:JhY[RV]aJg5[LF]aJg5[
L@1a0740L@1900T700071fP000=/LG401W40<XYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW6:LHYaRW6:LHYaRW6:LHYa1hX07G6:LHYaRW6:LHYaRW6:LHYaRW6:L8Y`RW1aL75`009aA`091`00
1`MV001UL@1a075[LF]aJg5[LF]aJhY[RV]aJg5[LF^:JhY[RV^:JhY[RV^:L8Y`RW2:L8Y`RW2:L8Y`
RW2:L8YaRW6:LHYaRW6:L8Y`RW2:L8Y`RW2:L8Y`RV]aJhY[LF]aJg5[LF]a0740A@091`001`MR0009
K75[LF]aJg5/00Aa02R:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa7HX0776:
LHYaRW6:LHYaRW6:LHYaRW6:L8Y`LG1aJg5000T700071f00011a075[LF]aJg5[LF]aJg5[0W40GhYa
RW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW6:LHY`RW2:L8Y`RW2:L8Y`RW2:L8Y[RV]aJg5[LF]a03l02@L000L7G@001W5[LFaaK0Qa01b:LHYa
RW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa0hX02g6:RXYaRW6:LHYa00b:00f?RXn:ShZ?RXn:ShZ?00R:
01aaRW6:LHYaRW6:LHYaRW6:LHY`RW2:L8Y`RW1a?P091`001`ML001@L@1aJg5[LF]aJg5[LF]aJg5[
RV^:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW43RP0ULHYaRW6:LHY`RW2:L8Y`RW2:L8Y`RW2:L8Y[RV^:JhY[LF]aJ`0l00T700071eT000IaJg5[
LF`:L@0JRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW4>RP05ShZ?RXl01HX05hn:RXZ?RXn:ShZ?RXn:
ShZ?RXn:ShZ?00>:00>?RXl00XX0776:LHYaRW6:LHYaRW6:LHYaRW2:L8Y`RW6:L74j00T700071eP0
059a0740LF]aJg5[LF]aJg5[LF^:K8Y/RW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW6:LHYaRW6:LHYaRW6:LHmaSg6?LHma0XX00hn:RP06RP0RLHYaRW6:LHY`RW2:L8Y`RW2:L8Y`RW2:
L8Y[RV]aJg5[LCT02@L000L7E@00275[LF]aJg5/2W4038YaRW6:LHYaRW6:L@N:00YaRhZ;RX^:T8Z@
2HX0<I2:T8Z@RY2:T8Z?RXn:ShZ?RXn:T8Z@RY2:T8Z@RXn:ShZ?RXn:ShZ?RXn:ShZ?RXl01XX0576:
LHYaRW6:LHYaRW6:LHYaRW6:>0091`001`MD0016L@1a0740LF]aJg5[LF]aJg5/LFb:LHYaRW6:LHYa
RW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHmaSg6:L@N:03QaSg6?LHn:Sg6?RXn:ShZ?
RXn:Sg6?LHmaRW6:LHYaRW6:LHY`RW2:L8Y`RW2:L8Y`RW2:L75[LCL02@L000L7D@00275[LF]aJg5[
374038YaRW6:LHYaRW6:L@Z:00^@RY2:T8Z@RY2:S`03RP0gShZ?RY2:T8Z@RY2:T8Z?RXn:T8Z?RY2:
T8Z@RY2:T8Z@RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:S`08RP0BLHYaRW6:LHYaRW6:LHYaRW6:=P091`00
1`M@000BL@1a0740LF]aJg5[LF]aK75/1740<HYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW6:LHYaRW6?LHmaSg6?LHl00hX0@Hn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:Sg6?LHmaSg6?LHYa
RW6:LHYaRW6:L8Y`RW2:L8Y`RW2:L8Y`RW1a03D02@L000L7C0001f]aJg5/LF`02W403XYaRW6:LHYa
RW6:LHYa0XX0GH^:RhZ@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:S`08RP0;LHYaRW6:LHYaRW40
<`091`001`M<0015L@1a0740LF]aJg5[LF]aK75/LFaaK8Y/RW6:LHYaRW6:LHYaRW6:LH]aRg6;LHYa
RW6:LHYaRW6;LI1aT76@LI1aRW6?00>:04Z?RXmaT76@LI1aT8Z@RXn:ShZ?RXn:T8Z@RY2:ShZ?RXn:
ShZ?RXn:ShZ?LHmaSg6?LHYaRW6:LHY`RW2:L8Y`RW2:L8Y`RV]aJc802@L000L7B@00175[LF`@L@0<
RW6:LHYaRW6;LHYa0XX0M8^:RhZ;RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:
Sg6:LHYaRW6:LHYa<@091`001`M:002FL@1a0740LF]aJg5/LFaaK75/LFb:K8YaRW6:LHYaRW6:LHYa
RW6;LH]aRg6;LH]aRW6;LHYaRg6:LI1aT76@LI1aT8Z@RXn:T8Z?RY2:T8Z@RY2:T8Z@RXn:T8Z?RY2:
T8Z@RY2:T8Z?RXn:ShZ?RXn:ShZ?LHmaSg6?LHmaRW6:LHYaRW6:L8Y`RW2:L8Y`RW1aJg5[<0091`00
1`M8009a00=/LG403740R8YaRW6:LHYaRg6;LH]aRhZ;RX^:RhZ;RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:ShZ?RY2:
ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXmaSg6:LHYaRW4_00T700071dP0019a0740L@1aJg5[LFaaK75/
LF`4L@24RW6:LHYaRW6:LHYaRW6:LHYaRg6;LI1aT76@LH]aRg6@LH]aT8Z@LI1aT76@LI2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z?RXn:ShZ?RXn:Sg6?LHmaSg6?LHYaRW6:LHYa
RW6:L8Y`RW2:L8Y`RW1aJg5[;P091`001`M600ma08N:LHYaRW6:LH]aRg6;LH]aRhZ;RX^:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8n@Si2?T8n@RZV?
T8Z@RY2:T8Z@Si2:T8Z@RY2:T8Z@RY2:T8Z?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXl00XX01G6:LHYa02d0
2@L000L7AP004740L@1aJg5[LF]aK75/LF`6L@29RW6:LHYaRW6:LHYaRW6:LHYaRg6;LI1aT76@LI1a
T76@RY2:T8Z@RY1aT8Z@LI2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RXn:
ShZ?RXn:ShZ?RXn:Sg6?LHmaShZ?LHmaRW6:LHmaSg2:L8Y`RW6:L8Y`LF]aJg40:`091`001`M400ea
08f:LHYaRW6:LHYaRW6;LH]aRg6;LI2:RhZ@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2?T8n@Si2?T8nYSjV?ZHnYSi2?T8n@Si2:T8Z@RY2:T8Z@RY2:
T8Z?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXl00XX01G6:LHYa02/02@L000L7A0004740LF]aJg5[LF]aK75/
LF`6L@2<RW6:LHYaRW6:LHYaRW6:LH]aRg6;LI1aT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RXn:ShZ?RXn:ShZ?RXmaSg6?RXn:
Sg6?LHmaSg6?LHmaSg6?LHY`RW2:L75[LF/Z00T700071d0000E[LFaaK00<L@2GRW6:LH]aRg6;LH]a
Rg6;LH]aRhZ;RY2:T8Z@RY2:T8Z@RY2:T8^@RY2:T8Z@RY2:T8Z@Ri2:T8Z@RY2:T8Z@RY2:T8Z@RY2?
T8n@Si2?T8nYSjV?ZHnYSjV?T8n@Si2?T8n@Si2?T8Z@RY2:T8n@RY2:T8Z@RXn:T8Z?RXn:ShZ?RXn:
ShZ?RXn:ShZ?RXmaRW6:L75`L@0X00T700071d0000ia0740LF]aK75/LFaaK0Ia09F:LHYaRW6:LHYa
RW6:LH]aRg6;LI1aT76@LI1aT8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:ShZ?RXn:ShZ?RXn:ShZ?LHmaShZ?RXn:
Sg6?LHmaSg6?L8Y`RW2:L75[L@0W00T700071cl000AaK75/3740=XYaRW6;LH]aRg6;LH]aRg6;LH^:
RhZ@RY2:T8Z@RY2:T8Z@Ri2;T8^@Ri2:T8Z@RY2:T8Z@R`>@00>;T9000i002HZ@RY2:T8n@RP0:T00=
ZHnYSjV?ZHnYT:V@Z@02T00jSi2?T8nYSi2?T8Z@Si2?T8n@Si2?T8Z@RY2:T8Z@RY2:ShZ?RXn:ShZ?
RXn:ShZ?RXmaRW6:L75`LBH02@L000L7?P003740L@1aK75/LFaaK0Qa09V:LHYaRW6;LH]aRg6;LH]a
Rg6;LI1aT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z?RXn:T8Z?RXn:ShZ?RXn:ShZ?RXn:
Sg6?LHmaSg2:L8Y`RW1aJg409@091`001`Lm0004LFaaK0Ia03b:LG5aRW6:LHYaRW6;LH]aRg6;LH]a
Rg6;RX^:T8Z@RY2:T8Z@RY2:T8^@Ri2;T8^@Ri2;T8^@Ri2;T8/NT017ZHnYT:V?ZI2YT:V@ZI2YT:V@
ZHnYSjV?ZHnYSi2?T8n@Si2?T8n@RY2:T8Z@RY2:ShZ?RXn:ShZ?RXn:ShZ?RXn:Sg6:LHX0174T00T7
00071c`000aa0740LFaaK75/LF`8L@2NRW6:LHYaRg6;LH]aRg6;LH]aRg6;LI1aT76@LI1aT76@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RXn:T8Z@RXn:ShZ?RXn:ShZ?RXn:Sg6?LHmaSg6?L8Y`RW1aL75[
LF/R00T700071c/000=aK7401g400hYaRP03L@0XRW6:LHYaRg6;LH]aRg6;LI1aRhZ;RY2:T8Z@RY2:
T8Z@RY2:T8^@R`>@00f;T92@Ri2;T8^@Ri2;00j@00FYT92@Z@09T01?ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YSjV?ZHnYSjV?ZHn@Si2?T8n@Si2:T8Z@RY2:T8Z?RXn:ShZ?RXn:ShZ?RXn:ShZ?LHYaRW6:L75`
L@0R00T700071cX000ia0740LFaaK75/LFaaK0Ia0:6:LHYaRW6;LH]aRg6;LH]aRg6;LH]aT76@LI1a
T76@LI2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2?T8n@Si2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z?RXn:ShZ?RXmaSg6?LHma
Sg6?L8Y`RW1aL75[L@0Q00T700071cT000=aK7401g40;XYaRW6;LHYaRW6;LH]aRg6;LH]aRg6@LI1a
T8Z@RY2:T8Z@RY2:T8^@Ri2;T8/9T003Ri2@012@00VYT:V@ZI2YT:T00i00>jV@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZHnYSjV?ZHnYSjV?T8n@Si2?T8n@Si2?T8n@RY2:T8n@00:?012:ShZ?RXn:ShZ?
RXn:ShZ?0XX0176:LHX2L@03L74001l02@L000L7>0004740L@1aK75/LFaaK75/LF`4L@2TRW6:LHYa
RW6;LH]aRg6;LH]aRg6;LI1aT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2?T8Z@RY2:T8Z@RY2?T8n@Si2?T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z?RXn:ShZ?RXn:Sg6?LHmaSg6?LHY`RW2:L75`LF/P00T700071cH000=[LF`02W40:X]a
Rg6;LH]aRg6;LH]aT76;LI1aT8Z@RY2:T8Z@RY2:T8^@Ri2;T8^@R`>@00>;T9004Y00EjV@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZHnYSjV?ZHnYSjV?
T8n@Si2?T8n@Si2?T8n@Si2:T004S`0>RXn:ShZ?RXmaSg6?LHX3LAl02@L000L7=P003740L@1aJg5/
LFaaK0Qa0:R;LH]aRg6;LH]aRg6;LH]aRg6;LI1aT76@LI1aT8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2?T8n@Si2?T8n@Si2:T8Z@RY2:T8Z@Si2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RXn:ShZ?RXn:ShZ?LHmaSg2?L8Y`RW2:L75`LF/N00T7
00071cD000=aK7402g40:8]aRg6;LH]aRg6;LI1aT76@LI2:T8Z@Ri2;T8Z@Ri2;T8^@Ri2;T8/JT01I
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YSjV?
ZHnYSjV?ZHn@Si2?T8n@Si2?T8n@Si2?T8n@Si000Xl048Z?RXn:ShZ?RXmaSg6?LHX2LAh02@L000L7
=0003740L@1a075/LFaaK0Ya0:V;LH]aRg6;LH]aRg6;LH]aRg6;LI1aT76@LI1aT76@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2?T8Z@RY2:T8Z@RY2?T8n@Si2?T8n@Si2?
T8n@Si2:T8n@Si2?T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z?RXn:ShZ?LHmaSg6?L8m`
RW2:L8Y`LG1a01d02@L000L7<`0<L@0ZRg6;LH]aRg6;LH]aRg6@LI1aT76@RY2;T8^@Ri2;T8^@Ri2;
T8^@Ri2;7900FjV@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YSjV?ZHnYSjV?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si000Xl048Z?RXn:ShZ?LHmaSg6:
LHX2LA`02@L000L7<P003W40L@1a0740LFaaK75/2740AW9aRg6;LH]aRg6;LH]aRg6;LH]aT76@LI1a
T76@LI2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8^@Ri2;T8^@Ri2:T8Z@RY2:T8l3T01SSi2@T8n@RY2?T8Z@
Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@
RY2:ShZ?RXn:Sg6?LHmaSg2?L8Y`RW2:L75`01`02@L000L7<@0>L@0XRg6;LH]aRg6;LH]aT76@LI1a
T76@Ri2;T8^@Ri2;T8^@Ri2;T8^@Rab@05fYT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V?ZHnYSjV?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si00
18l04HZ?RXn:ShZ?LHmaRW6:LHYa01/02@L000L7<0004740L@1a0740LFaaK75/LF`4L@18LW5bLH]a
Rg6;LH]aRg6;LH]aRg6;LI1aT76@LI1aT76@LI2:T8Z@RY2;T8^@RY2:T8Z@Ri2:T8^@Ri2;T8^@Ri2:
T8Z@RY2:2i002hn@Si2?T8n@Si2?00>@052?T92@SjV?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@RY2:T8Z@
RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:ShZ?RXn:Sg6?LHmaSg2?L8Y`RW2:L8Y`LA/02@L000L7;`0>L@0X
LW6;LH]aRg6;LH]aT76@LI1aT76@LI2;T8^@Ri2;T8^@Ri2;T8^@RaB@00>YT9001I00AjV@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V?
ZHnY00:@01R?T8n@Si2?T8nYSi2?T8n@Si2?T8n@Si04S`0@RXn:ShZ?LHmaRW6:LHYaRQX02@L000L7
;P004740L@1a074HLFaaK75/LF`4L@0nLW5bLH]aRg6;LH]aRg6;LI1aRg6@LI1aT76@LI1aT76@LI1a
T8Z@RY2;T8^@Ri2;T8^@Ri2;T8^@Ri2;T8/3T009Ri2;T8^@RY2;00^@00V?T8n@Si2?T8l02Y00C:V?
ZHn@Si2?T8n@Si2?T8n@Si2?T8Z@RY2:T8Z@RY2?T8Z@RY2:T8Z@RY2:T8Z@RY2:ShZ?RXn:ShZ?LHma
Sg6?L8m`RW2:L8Y`RW0J00T700071bd000QaK75/LFaaK0Qa02ibLG9aRg6;LI1aT76;LI1bT7:@Ri2;
T7:@Ri2;T8^@Ri2;T8^@Ri2;T8^@Ri2;4900AJV@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@Z@08T00_SjV?ZHn@Si2?T8n@Si2?T8n@
Si2?T8n@Si2?T8Z?RXn:ShZ?RXmaSg6?LHYaRW006@091`001`L/000<L@1a0740L@1aK75/2740=8]a
Rg6;LH]aRg6;LH]aRg6;LI1aT76@LI1aT76@LI1aT8Z@Ri2;T8^@Ri2;T8^@Ri2;T8/3T009Ri2@T8^@
Ri2;00>@00>;T90079002jV@ZI2YT:V@ZI2Y00R@04>?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2:T8Z@
Si2?T8n@Si2:T8Z@RY2:T8Z@RXn:Sg6?LHmaSg6?L8m`RW2:L8Y`01P02@L000L7:`00274HLFaaK75/
2740:79aLW6;LH]aT76@LI1aT7:@LY2;T8^@Ri2;T8^@Ri2;T8^@Ri2;T8/HT017ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:j@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:T0
1Y0078nYSjV?ZHn@Si2?T8n@Si2?T8n@Si2?T8n@Si04S`0@RXn:ShZ?LHmaSg6?LHYaRQP02@L000L7
;0003740L@1a074HLFaaK0Ia03IbLH]aRg6;LH]aRg6;LH]aRg6@LI1aT76@LI1aT76@LY1bT8^@Ri2;
T8^@Ri2;T8^@Ri2;T8/7T003Ri2;02B@00fYT:V@ZI2YT:V@ZI2Y00F@04BYSi2?T8n@Si2?T8n@Si2?
T8n@Si2?T8n@RY2:T8n@Si2?T8n@Si2:T8Z@RY2:T8Z@RXn:Sg6?LHmaSg2?L8Y`RW2:L1P02@L000L7
:P001aQaK75/LF`02W409W9aRg6;LI1aT76@LI1aT7:@LY2;T8^@Ri2;T8^@Ri2;T8^@Ri2;6Y00BJV@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2^T:V@[Y2^T:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:T0190078nYSjV?ZHnYSi2?T8n@Si2?T8n@Si2?T8n@Si06S`0>RXn:ShZ?LHmaSg6?
LHXH00T700071bX000aa0740L@1a674HLF`6L@0dLW5bLH]aRg6;LI1aT76;LH]aT76@LI1aT76@LI1a
T76@LY2;T8^@Ri2;T8^@Ri2;T8^@RbJ@00>YT9001I004jV@T92YT:V@ZI2YT:V@ZI2YT:T00i00A:V?
ZHn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@RY2:T8Z@RY2:ShZ?RXmaSg6?LHm`
Sg2:L8Y`60091`001`LY0006LFaaK75/2W40979aLW6;LH]aT76@LI1aT76@LY2;T8^@Ri2;T8^@Ri2;
T8^@Raj@04^YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:j@[Y2^T:j@[Y2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:T00i007JV@ZI2YSjV?T8n@Si2?T8n@Si2?T8n@Si2?T8n@00J?
00b:ShZ?LHmaSg6?LHlH00T700071bX000aa0740LAQa675/LF`4L@0fLW5bLG9aRg6;LH]aT76@LH]a
T76@LI1aT76@LI1aT76@LI2;T8^@Ri2;T8^@Ri2;T8^@Ri2;<900F:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YSjV?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2:T8Z@RY2:T8Z?RXn:Sg6?LHma
Sg6?L8m`RW0H00T700071bP000E/LFaaK00:L@0VLW5bLH]aRg6;LI1aT76@LI1bT7:@Ri2;T8^@Ri2;
T8^@Ri2;T8/NT01FZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2^T:j@[Y2^T:j@
ZI2YT:V@ZI2YT:j@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZHl3T00DSi2?T8n@Si2?T8n@Si2?T8n@
Si04S`0<RXn:ShZ?LHmaSg6?60091`001`LX000>L@1a074HLAQaK75/LF`4L@0fLW5bLH]aRg6;LI1a
T76@LI1aT76@LI1aT76@LI1aT76@LI2;T8^@Ri2;T8^@Ri2;T8^@Ri2;8Y001JV@ZI2Y00N@05ZYT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V?ZHn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2:
T8Z@RY2:T8Z?RXn:ShZ?LHmaSg6?LHm`Sg0H00T700071bL000=aK7402g40:79aLW5bLH]aRg6;LI1a
T7:@LY1bT8^@Ri2;T8^@Ri2;T8^@Ri2;T8/NT01EZY2ZT:Z@ZY2ZT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2^T:j@[Y2^T:j@[Y2^T:j@ZI2YT:j@[Y2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@Z@0:T00@
Si2?T8n@Si2?T8n@Si2?T0B?00Z:ShZ?LHmaSg6?60091`001`LV000<L@1a0740LAQaK75/1740>79a
LW5bLH]aRg6;LH]aRg6;LI1aT76@LI1aT76@LY1bT7:@LY1bT7:@LY1bT7:@Ri2;T8^@Ri2;0i000h^@
T00/T003ZI2@00N@00VYT:V@ZI2YT:T03900>Hn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@
RY2:T8Z?LHmaSg6?LHmaSg6?LHmaSg6:L@0H00T700071bH000<HLF`02W40879aLW6;LH]aRg6@LI1b
T7:@LY2;T8^@LY2;T8^@Ri2;7900Oi6@ZY2ZT:Z@ZY2ZT96@ZY2ZT:Z@ZI2ZT:Z@ZY2ZT:V@ZI2YT:V@
ZI2_T:n@[Y2_T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@ZI2YT:V@[Y2^T:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YSi2?T8n@Si2?T8n@Si2?T8n@Si0018l01XZ?RXn:SaP02@L000L79P002W40L@1a075/
LF`6L@0dLW5bLG9aRg6;LH]aRg6;LI1aT76@LI1aT76@LI1bT7:@LY1bT7:@LY1bT7:@Ri2;T8^@Rd2@
00VYT:V@ZI2YT:T03900>Hn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2:T8Z@RY1aSg6?
LHmaSg6?LHmaSg6?L@0H00T700071bD000=aK7402G408W9aLW5bLH]aRg6;LI1aT7:@LY1bT7:@Ri1b
T8^@Ri2;T8/NT01ZTI2ZT:Z@ZY2ZT96@TI2ZT96@ZY2ZT:Z@ZY2ZT:V@ZI2YT:V@ZI2_T:n@[i2^T:n@
[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:V@[Y2YT:V@[Y2^T:j@[Y2^T:V@ZI2YT:V@ZI2YT:V@ZI2YS`>@
012?T8n@Si2?T8n@Si2?T8n@18l01XZ?RXn:SaP02@L000L79P003740L@1a675/LFaaK0Aa03IbLG9a
LW6;LH]aRg6;LH]aT76@LI1aT76@LI1aT7:@LY1bT7:@LY1bT7:@LY2;T8^@Ri2;T8/5T003Ri2@03J@
00FYT:V@Z@03T003ZI2@00Z@03V?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@RY2:T76@
LI1aSg6?LHmaSg6?LHmaSg4060091`001`LT0003675a00Ya029bLG9aLW6;LH]aT76@LI1bT7:@LY2;
T8^@Ri2;T8^@Ri2;49000i6@T@05T01_TI2AT96@ZY2ZT:Z@ZY2ZT96@TI2AT:Z@TI2ZT:Z@ZY2ZT:V@
ZY2YT:V@ZI2_T:n@[i2_T:n@[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@ZI2^T:V@[Y2^T:j@[Y2^T:V@
ZI2YT:V@ZI2YT:V@ZI2Y00B@012?T8n@Si2?T8n@Si2?T8n@18l01XZ?RXn:SaP02@L000L790003740
L@1a674HLFaaK0Aa03IbLG9aLW5bLH]aRg6;LH]aT76@LI1aT76@LI1aT7:@LY1bT7:@LY1bT7:@LY2;
T8^@Ri2;T8]ET00iSi2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@RY2:T76@LI1aSg6?
LHmaSg6?LHma01P02@L000L790000faaL@08L@0TLW5bLG9aLW6;LI1aT76@LY1bT7:@LY2;T8^@Ri2;
T8^@Ri2;3Y001I6@TI2A00>@076AT92@TI2AT96@TI2AT:Z@ZY2AT96@TI2AT96@ZY2ZT:Z@ZY2ZT:V@
ZY2YT:V@[i2_T:n@[i2_T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2YT:j@ZI2YT:j@[Y2^T:j@[Y2YT:V@
ZI2YT:V@ZI2YT:V@Z@04T00@Si2?T8n@Si2?T8n@Si2?T0J?00B:ShZ?60091`001`LT000:L@1a674H
LAQaK0Aa03IbLG9aLW5bLH]aRg6;LH]aRg6@LI1aT76@LI1aT7:@LY1bT7:@LY1bT8^@LY1bT8^@LY2;
T8]GT00iSi2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2:T76@LHmaSg6?LHmaSg6?
LHma01P02@L000L78`000g5/L@05L@0VLW5bLG9aLW5bLH]aRg6@LI1bT7:@LY1bT7:@Ri2;T8^@Ri2;
T8/>T01kTI2AT96@TI2AT96@TI2AT96@TI2AT:Z@ZY2ZT96@TI2AT96@TI2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@
ZI2_T:n@[i2_T:n@[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:V@[Y2^T:j@[Y2^T:V@ZI2YT:V@
ZI2YT:V@ZI2Y00J@012?T8n@Si2?T8n@Si2?T8n@1Xl00hZ?000G00T700071b80049a074HLAQa675/
LFabLG9aLW5bLG9aLW5bLH]aRg6@LI1aT76@LI1bT7:@LY1bT7:@LY1bT7:@LY1bT8^@Ri2;T8/3T005
Ri2;T8/0GI00;hn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2:T8Z@RXn:ShZ?LHmaSg6?LHma01P0
2@L000L78P000aQaL@08L@0NLW5bLG9aLW6;LH]aRg:;LY1bT7:@LY1bT7:@LY2;7900JI6@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@ZY2ZT:Z@ZY2_T:V@ZI2YT:V@[Y2_T:n@[i2_T:j@
[Y2^T:V@[Y2YT:j@[Y2^T:j@[Y2YT:V@ZI2YT:V@ZI2YT:V@Z@0<T00KSi2?T8n@Si2?T8n@Si2?T8n@
Si2:T8Z@RXn:01T02@L000L78P00@74HLAQa675/LFabLG9aLW5bLG9aLW5bLG9aRg6;LI1aT76@LI1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@Ri2;T8/5T003Ri2;05n@02j?T8n@Si2?T8n@Si2?T8n@Si2?T8n@
Si2?T8n@Si2:T8Z@RXn:T76?LHmaSg6?6@091`001`LR00Ua01ibLG9aLW5bLG9aRg6;LX]bT7:@LY1b
T7:@LY1bT78NT00_TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2ZT:n@ZY2_T:l0
0i00;JV@ZI2^T:j@[i2_T:n@[i2_T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@ZI2YT:V@Z@03T007ZI2YT:V@
Z@0<T00JSi2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@RY0J00T700071b8000Qa674HLFaaK09a03QbK79a
LW5bLG9aLW5bLH]aRg6;LI1aT76@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@Ri2;T8^@R`>@00>;T900
Gi00;Hn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2:T8Z@LHmaSg6?L@0J00T700071b80
1g408W9aLW5bLG9aLW6;LH]bRg:;LY1bT7:@LY1bT7:@LY2;T8/JT00cTI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT:Z@TI2_T:n@[i2A00N@026^T:n@[i2_T:n@[Y2^T:j@[Y2^T:V@
[Y2^T:j@[Y2^T:T02I000jV@Z@0BT00HSi2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z?6P091`001`LR0006
LAQa675/1740=W9aLW5bLG9aLW5bLG9aRg6;LI1aT76@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY2;
T8^@RfN@02b?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2:T76@LHmaSa/02@L000L7
8@06L@0RLW5bLG9aLW5bLG9aRg6;LX]bT7:@LY1bT7:@LY1bT7:@LQ2@00>AT9401i00Fi6@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT9J@[i2_T:n@[i2AT96@TI2FT9F@[i2_T:n@
[i2_T:j@[Y2^T:j@[Y2^T:j@[Y2^T:h00i000jV@T005T003ZI2@01:@01N?T8n@Si2?T8n@Si2?T8n@
Si2?T8n@S`0K00T700071b8000Qa674HLFaaK09a03AbLG9aLW5bLG9aLW5bLH]aRg6@LI1aT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@Ri2;1i000h^@T01OT00/Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@
Si2?T8n@Si2?T8Z@RY1aT76?LHlK00T700071b401W408W9aLW5bLG9aLW5bLH]aRg:;LY1bT7:@LY1b
T7:@LY1bT78>T005TI2AT9401I00DI6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@UY2FT9J@UY2_T96@TI2FT96@UY2_T:n@[i2_T:n@UI2^T9F@U@03T009[Y2^T:j@[Y2^022@
01J?T8n@Si2?T8n@Si2?T8n@Si2?T8n@70091`001`LP0016L@1a674HLAQbK79/LW5bLG9aLW5bLG9a
Rg6@LI1aT76@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LV6@00>?T9001I0098n@
Si2?T8n@Si2?T8n@Si2?T8n@Si2:T8Z@LI1aT76?LHmaSad02@L000L78003L@0>LW5bLG9aLW5bLG9a
LW42LP0BRg:;LY1bT7:@LY1bT7:@LY2;5Y00KI6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT9J@UY2FT9J@UY2AT9J@UY2FT:n@[i2_T:n@[i2FT:n@UY2_T9F@[i2_T:n@[i2^T9F@
UI2ET9F@UI2^T:j@[Y2^T:h07Y003Hn@Si2?T8n@Si2?T8l07@091`001`LP0016LAQa674HLAQbK79a
LW5bLG9aLW5bLG9aRg6@LI1aT76@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LV^@
02:?T8n@Si2?T8n@Si2?T8n@Si2?T76@LI1aT76@LI1aSg6?7@091`001`LO00Aa00abLG9aLW5bLG9a
LW44LP0BRg:@LY1bT7:@LY1bT7:@LY2;39000i6@T005T01_TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2FT96@TI2AT96@TI2FT9J@UY2FT96@UY2FT9J@UY2_T:n@[i2_T:n@UY2FT9J@UY2FT:n@
[i2_T:n@[i2ET9F@UI2ET9F@[Y2^T:j@[Y2E01j@00b?T8n@Si2?T8n@Si0N00T700071b0004Ia674H
LAQb679/LW5bLG9aLW5bLG9aLW6@LI1aT76@LI1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1b49000i6@T@1HT00QSi2?T8n@Si2?T8n@Si2?T8n@Si2:T76@LI1aT76@LI1a01h02@L000L7
7P000faaL@02L@0:LW5bLG9aLW5bL@Ib012;LY1bT7:@LY1bT7:@LY1b39000i6@T@05T01aTI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT9J@UY2FT9J@TI2AT9J@TI2FT9J@
UY2FT9J@UY2FT9J@UY2FT9F@UY2FT9J@UY2ET9F@UI2ET9F@UI2ET9F@UI2ET9D07Y0038n@Si2?T8n@
Si2?T1h02@L000L77P00B740LAQa674HLVabK79aLW5bLG9aLW5bLH]aT76@LI1aT76@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LQ:@00>AT900EY0088n@Si2?T8n@Si2?T8n@Si2?T8n@
LI1aT76@LI1aT76@7`091`001`LN00Ea00YbLG9aLW5bLG9a178058]bT7:@LY1bT7:@LY1bT7:@LY1b
29001I6@TI2A00>@07>AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2FT9J@UY2FT96@TI2AT96@TI2FT9J@UY2FT9J@UY2FT9J@UI2FT9J@UY2FT9J@UY2ET9F@UI2ET9F@
UI2ET9F@UI2E01j@00^?T8n@Si2?T8n@S`0O00T700071ah004Qa074HLAQa679/LW5bLG9aLW5bLG9a
LW6@LI1aT76@LI1aT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT79[T00OSi2?T8n@
Si2?T8n@Si2?T8n@Si1aT76@LI1aT76@L@0P00T700071ad01W402W9aLW5bLG9aLW46LP0BT7:@LY1b
T7:@LY1bT7:@LY1g1Y00FI6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@UY2AT96@TI2AT96@TI2FT9J@UY2FT9J@UY2F00>@00nET92@UY2FT9F@
UY2ET9D02I002IF@UI2ET9F@U@0NT009Si2?T8n@Si2?02402@L000L770002740L@1a674H0W40>79a
LW5bLG9aLW5bLG9aT75bLI1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1g2I000gN@
M`03T003Mi2@05n@00>?T9000i006Hn@Si2?T8n@Si2?T7J@Si2?T8n@Si1fT8l08P091`001`LM0004
LAQaK0Aa02AbLG9aLW5bLG9aT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT78:T017TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT940
1i002IJ@UY2FT9F@UP0=T005UI2@T9D099003Xn@Si2?T8n@Si2?T8n@8P091`001`LL0008L@1a074H
LAP2L@0nLW5bLG9aLW5bLG9aLW5gLG9aMg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7N@Mi1gT7L3T009Mi1gT7N@Mi1g00f@00EgT92@M`1;T003Si2?00>@01Z?T8n@Si2?T8n@Si1fT7J@
MY2?T8n@MY1fT2<02@L000L770001AQa675/009a00YbLG9aLW5bLG9a0W806Y1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1b2Y00AI6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@T@09T003TI2F00>@00>ET900>9002Xn@Si2?T8n@Si0T00T7
00071a`000Qa074HLAQa609a011bLG9aLW5bLG9aLW5bLI1a0W80>91bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1g2i002GN@Mi1gT7N@M`09T005Mi2@T7L0
0i000gN@T00gT003Si2@00>@01J?T8n@Si2?T7J@MY1fT7J@Si1fT7J@9@091`001`LL0005674HLF`0
1740179aLW44LP0LT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LPR@04FAT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT9402i001I6@
T92A03j@00N?T92@Si2?02D02@L000L770002740LAQa675/0W40C79aLW5bLG9aLW5bLG9aMg5gLGMb
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7L3T003
Mi2@00>@00UgT7N@Mi1gT7L02I002GN@Mi1gT7N@M`0;T009Mi1gT7N@Mi1g00^@00UgT7N@Mi1gT7L0
2i001gN@T91fT8l00i005hn@Si2?T8n@MY1fT7J@MY1fT7J@MY1f02H02@L000L770000aQa6004L@06
LW5bLG9a178077MbT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7L:T013TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@T@1<T007Si2?T8n@
S`0W00T700071a`000Qa074HLAQaK09a059bLG9aLW5bLG9aLW5bLGMaMg9gLY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g0i002gN@Mi1gT7N@
Mi1g00>@00egT92@Mi1gT7N@Mi1g00^@00UgT7N@Mi1gT7L02i002GN@Mi1gT7N@M`0;T007Mi1fT7J@
MP05T00CSi2?T8n@MY1fT7J@MY1fT7J@MP0X00T700071a`000<HLF`017401W9aLW5bL@Ab01agLY1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1g2Y002I6@TI2AT96@T@03T00gTI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@T@1>T003Si2@02T02@L000L77000bW4H
LAQa674HLQQbLG9aLW5bLG9aLW5bLG9aMg5gLI1aT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7J@MY1fT7J@Mi1gT7J@MY1f
T7J@MY1fT7J@MY1fT7J@MY1fT74Z00T700071a/027401W9aLW5bL@9b01agLWMbMg:@LY1bT7:@LY1b
T7:@LY1bT7N@Mi1g0Y00@i5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT9400i000i6@T005T009TI2AT96@TI2A04F@:`091`001`LL0039
LAQa674HLAQb679aLW5bLG9aLW5bLG9aLW5gLI1aT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7J@Mi1fT7J@MY1gT7J@Mi1f
T7J@MY1fT7J@MY1fT7J@MY1fT7J@02/02@L000L76`08L@0aLW5bLG9aLW5gLWMbMg9gLY1bT7:@LY1b
T7:@LY1gT7N@Mi1gT7NAMi5gTGNAMi5gT@03T00eTI2@T96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT9402i001I6@T92A00>@00>AT900@Y0/00T700071a`00<Ma674H
LAQa678HLTibLG9aLW5bLG9aLW5bLGMaMg6@LI1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1gT7N@Mi1gT7N@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@MY1fT7J@MY1fT7J@MY1fT7J@MY1f
T7J@MY1fT7J@MY1fT7J@02d02@L000L76`08L@08LW5bLG9aLW42LP0VMg9gLY1bT7:@LY1bT7:@LY1b
T7N@Mi1gT7N@Mi5gTGNAMi5gTGL2T00]TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2A00>@00NAT96@TI2A05J@;P091`001`LL0035LAQa674HLAQb679>LW5bLG9aLW5bLG9a
LW5gLI1aT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7J@MY1fT7J@MY1gT7J@MY1fT7J@MY1fT7J@MY1fT7J@MY00;`091`00
1`LK00Qa00QbLG9aLW5bL@Ab02AgLWMbT7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gTGNAMi5gTGL4T00?
TI2@T96@TI2AT96@TI2A00>@016AT96@TI2AT96@TI2AT96@T@0:T005Mi6@T7L00i000gN@T009T009
Mi1gT7N@Mi1g00^@00=gT900<90a00T700071a`00<9a674HLAQa678HLTibCW9aLW5bLG9aLW5bLGMa
Mg6@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1gT7N@Mi1gT7N@LY1g
T7:@Mi1bT7N@Mi1gT7N@Mi1bT7:@LY1bT7N@Mi1gT7N@Mi1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7:@Mi1b
T7N@LY1gT7N@Mi1gT7N@MY1fT7J@MY1fT7J@MY1fT7J@MY1fT7J@MY1fT7J@MS802@L000L77007L@08
LW5bLG9aLW42LP0VMg9gLWMbT7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7NAMi5gTGL2T005TI2@T940
0i007I5gTI2AT96@TI2AT96@TI2AT96@TI2AT96@TGNA00Z@00UgT7N@Mi1gT7L00i005gN@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1g00>@00UgT7N@Mi1gT7L0:I0d00T700071a`00;ma674HLAQaCW9>LW5bLG9a
LW5bLG9aMg5gLGMbMg9gLY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1b
T7N@LY1gT7:@Mi1gT7N@Mi1gT7:@LY1bT7:@LY1gT7:@Mi1bT7:@LY1bT7:@LY1gT7:@Mi1bT7N@LY1b
T76@LY1aT7N@Mi1gT7N@Mi1aT76@LI1aT7J@Mi1gT7N@Mi1gT7J@MY1fT7J@MY1fT7J@MY1fT00e00T7
00071ad0041a675aLW5bLG9aLW5bLG9aMg9gLWMbT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1gT7N@Mi1g
T7NAMi1gTGNAMi5g0i000gN@T003T01MMi5gTGN@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gT8Z@RY2:T8Z@RY2:T8Z@RY2:T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g00F@
00UgT7N@Mi1gT7L0290g00T700071ad00;/HLAQa675>LTibLG9aLW5bLG9aMg5gLGMaMg9gLWMbT7:@
LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7:@LY1bT7:@LY1gT7N@Mi1gT7N@
LY1bT7:@LY1aT8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@LY1bT7:@LY1bT76@
LI1aT76@LI1gT7N@Mi1gT7N@MY1fT7J@MY1fT7J@MY1f03P02@L000L77@00@W4HLDibLG9aLW5bLG9a
LW5gLWMbMg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gTGNAMi5gTGN@M`>@06=gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi2:T8Z@Ri2:T8Z@RY2:T8Z@RY2:T8Z@RY2?T8n@Si2?T8n@
Si2:T8Z@RY2:T8Z@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7L00i002GN@Mi1gT7N@M`05T3X02@L000L7
7P00]g4HLAQaCW9>LW5bLG9aLW5bLG9aMg5gLGMbMg9gLWMbT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7N@Mi1gT7:@Mi1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:ShZ?RXn:Sg6@LI1aT76@LI1aT76@MY1fT7J@MY1fT76@MY1a
T7J@MY1fT00k00T700071ah000lHLDibLG9aLW5bLG9aLW400W80EWMbMg:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7N@Mi1gT7N@Mi1gT7NAMi1gTGN@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT8Z@RY2:
T8Z@RY2;0i000h^@T003T009RY2:T8Z@Si2?00>@03N?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@RXn:ShZ?
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g00:@?@091`001`LN002dLAQa674HLTibCW9aLW5bLG9a
LW5gLGMaMg9gLWMbMg9gLY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LH]aRg6@LI1aT8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:ShZ?RXn:Sg6?LHmaSg6@LI1aT7J@MY1fT7J@LI1aT7J@LI1f?P091`001`LO001TLDibLG9a
LW5bLG9aLW5gLWMbMg9gLY1bT7:@LY1bT7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi5gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mh^:T8Z@RY2:T8Z@RY2:T8^@Raj@032YSi2?T8n@Si2?T8n@Si2?T8n@RY2:
ShZ?RXmgT7N@Mi1gT7N@Mi1gT7N@Mi1gT7Lo00T700071al00;4HLAQa679>LTibLG9aLW5bLG9aMg5g
LGMbMg9gLWMbMg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1aRg6;
LH]aT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@
RY2:T8Z?RXn:Sg6?LHmaSg6@LI1aT76@LI1aT76@LI1aT740@0091`001`LO000>LDibLG9aLW5bLG9a
LW42LP1DMg9gLWMbT79gLY1bT7:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1aRg6@LI2:T8Z@Ri2;T8^@Ri2;T8^@Ri2;4Y000jV@T009T00]ZI2YSi2?T8n@Si2?T8n@Si2?
T8Z@RY2:T8Z?RXmgT7N@Mi1gT7N@Mi1gT7N@04802@L000L78000574HLQQb678HLTibCW9aLW5bLG9a
0W80TWMaMg5gLWMbMg9gLWMbMg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LX]aRg6;LH]a
Rg6@LI1aT76@LI1aT8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@Si2?T8n@RY2:
T8Z@RY2:T8Z@RXn:ShZ?RXn:Sg6?L8maT76@LI1aB0091`001`LP000;LG9aLW5bLG9aLW401780BWMb
Mg9gLWMbT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mh]aRhZ@Ri2;
T8^@Ri2;T8^@Ri2;39001jV@ZI2YT:T02I004JV@ZI2YT:V@ZI2YT:V@ZI2Y00B@01:?T8n@Si2?T8n@
Si2?T8n@Si02S`0:RXn:Sg6@Mi1aRTP02@L000L78000L74HLAQb678HLTibCW9aLW5bLG9aLW5gLGMa
Mg5gLWMbMg9gLWMbMg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bRg6;LH]aRg6;LH]aT76@LI1a
T76@LI2:T8Z@Ri2;T8^@RY2:T8Z@RY2;T8X3T00eRY2?T8Z@RY2:T8Z@Si2?T8n@Si2?T8Z@RY2:T8Z@
RY2:T8Z?RXn:ShZ?LHmaSg2:L8Y`RW00B0091`001`LQ009a00QbLG9aLW5bL@Ab04YgLWMbMg9gLY1b
T7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1aRg6;LY2;T8^@Ri2;T8^@Ri2;
T8^@R`b@00NYT:V@ZI2Y00>@02fYT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZHn@Si2?T8n@Si2?
T8n@Si000Xl02XZ?RXmaSg6?LHY800T700071b8003ea678HLQQbCW9>LW5bLG9aLW5bLGMaMg5gLGMa
Mg9gLWMbMg9gLWMbT7:@LY1bT7:@LY1bT7:@LY1bT7:@009b02]aRg6;LH]aRg6;LH]aT76@LI1aT76@
LI2:T8^@Ri2;T8^@Ri2;T8^@Ri2;00^@00V?T8n@Si2?T8l00i009Hn@Si2?T8n@RY2:T8Z@RY2:T8Z@
RY2:ShZ?LHmaSg6?L8m`RW00B0091`001`LQ000:LDibLG9aLW5bL@Ab04QgLWMbMg9gLY1bT7:@LY1b
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT76;LH]aRg6@LY2;T8^@Ri2;T8^@Ri2;T8/>T00Y
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:T00Y0068n@Si2?T8n@Si2?T8n@
RXn:ShZ?LHmaSdP02@L000L78P00H74HLQQb679>LTibLG9aLW5bLG9aMg5gLGMaMg5gLGMbMg9gLWMb
Mg:@LY1bT7:@LY1bMg9gLWMbMg9gK79aRg6;LH]aRg6;LI1aT76@LI1aT76@LI1aT8Z@Ri2;T8^@R`>@
00>;T8/00i000h^@T00?T003Si2@00F@02B?T8n@Si2?T8Z@RY2:T8Z@RY2:T8Z@RXn:Sg6?LHmaSg2:
L75700T700071b8000U>LW5bLG9aLW401W80A7MbMg9gLY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1aRg6;LH]aT76@Ri2;T8^@Ri2;T8^@Ri2;4900:jV@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:T00Y005hn@Si2?T8n@Si2?T8n@RXn:Sg6?LHma04L02@L000L7
8P00>W4HLAQb679>LTibLG9aLW5bLG9aMg5gLGMaMg5gLWMbMg9gLWMbMg:@LY1bT7:@LY1bMg9gLWMb
LF`2L@0RLW6;LH]aRg6;LI1aT76@LI1bT76@LI2;T8^@Ri2;T8^@RbN@02:?T8n@Si2?T8n@Si2:T8n@
Si2:T8Z?RXmaSg6?LHmaSg2:A`091`001`LR0009679aLW5bLG9a00Ab049gLWMbMg9gLWMbT7:@LY1b
T7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT75bLG9aRg6;LI1aT7:@Ri2;T8^@Ri2;T8/DT00]ZY2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2^T:V@ZI2YT:V@ZI2YT:V@ZI2Y00J@01:?T8n@Si2?T8Z?RXn:Sg6?
LHm600T700071b8005Ya674HLDibCW9>LW5bLG9aLW5gLG9aMg5gLWMbMg9gLY1bT7:@LY1bT79gLY1b
Mg9gLY1/LFabK79aLW5bLH]aRg6;LI1aT76@LY1bT7:@Ri2;T8^@Ri2;T8/[T00SSi2?T8n@Si2?T8n@
Si2?T8n@Si2:T8Z?LHmaSg6?LHmaSg00AP091`001`LR0003675a009a04QbLG9aLW5bLGMbMg9gLWMb
Mg9gLWMbMg9gLY1bT7:@LY1bT7N@Mi1gT7N@Mi1aLW5bLG9aRg6;LH]bRg:@LY2;T8^@Ri2;T8/<T00=
TI2AT96@ZY2ZT96@T@05T00MZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2^T:V@ZI2YT:T039002Xn@Si2?T8n@
Si04S`06RXn:Sg6?AP091`001`LT001DLAQa679>LTibCW9aLW5bLG9aLW5gLGMaMg9gLWMbMg:@LY1b
T79gLWMbMg9gLWM/LFabK79aLW5bLH]aRg6;LI1aT7:@LY1bT7:@LY2;T8^@Ri2;0i000h^@T00[T00R
Si2?T8n@Si2?T8n@Si2?T8n@Si2:T8Z?LHmaSg6?LHm`LDD02@L000L78`00BW5>LG5bLG9aLW5bLGMa
Mg9gLWMbMg9gLWMbMg9gLY1bT7:@LY1bT7N@Mi1gT7N@Mg9aLW5bLH]aRg6;LX]bT7:@LY2;T8^@Ri2;
3Y00;i6@TI2AT96@ZY2AT96@TI2AT:Z@ZY2_T:j@[Y2YT:V@ZI2YT:j@[Y2^T:V@ZI2Y00b@00b?T8n@
Si2?T8n@Si04S`05RXmaSg40A@091`001`LT001@LAQa675>LTibCW9aLW5bLGMaLW5gLGMbMg9gLWMb
Mg9gLY1bT79gLWMbMg9gLWM/LVabLG9aLW6;LH]aRg6@LY1bT7:@LY1bT7:@LY2;T8/kT00LSi2?T8n@
Si2?T8n@Si2:T8Z@LHmaSg6?LHm`LDD02@L000L790000aQaL@02L@06LW5bLG9a0W80>WMbMg9gLWMb
Mg9gLWMbMg:@LY1bT7:@LY1gT7N@Mi1gT75bLG9aLW6;LH]aRg:@LY1bT8^@Ri2;T8/:T00eTI2AT96@
TI2AT96@TI2ZT96@TI2AT96@[i2_T:n@[i2_T:j@[Y2^T:j@[Y2^T:j@[Y2^T:T0490028n@Si2?T8n@
18l01HZ?LHma04D02@L000L79000D74HLAQa675>LTibLG9aLW5bLG9aLW5gLGMbMg9gLWMbMg9gLY1b
Mg9gLWMbMaQaK79/LW5bLG9aRg6;LH]aT7:@LY1bT7:@LY1bT7:@Ri2;3Y000i6@T00ZT00MSi2?T8n@
Si2?T8n@Si2:T8Z@LHmaSg6?LHm`RW00A0091`001`LU0003LDia00=a02IbLG9aMg5gLWMbMg9gLWMb
Mg9gLWMbT7:@LY1bT7:@Mi1gT7N@M`9a01QbLG9aRg6;LH]aT7:@LY1bT8^@Ri2;T8/8T00gTI2AT96@
TI2AT96@TI2AT:Z@TI2AT96@TI2_T:n@[i2_T:n@[Y2^T:j@[Y2^T:j@[Y2^T:j@[P0@T00:Si2?T8n@
Si2?T0B?00AaSg5aA0091`001`LV001:LAQa675>LTibCW9aLW5bLG9aMg5gLGMbMg9gLWMbMg9gLWMb
Mg9gLWMbLFaaK79aLW5bLH]aRg6;LI1bT7:@LY1bT7:@LY1bT78<T009TI2AT96@TI2A02b@01^?T8n@
Si2?T8n@Si2:T8Z@LI1aSg6?LHmaRW00A0091`001`LV0003CW5a009a02AbLG9aLW5gLGMbMg9gLWMb
Mg9gLWMbT7:@LY1bT7:@Mi1gT7L4L@0FLW5bLH]aRg:@LY1bT7:@LY2;T8^@R`J@03^AT96@TI2AT96@
TI2AT96@TI2AT:n@TI2AT96@UY2_T:n@[i2_T:n@[i2_T:j@[Y2^T:j@[Y2^T:j@[P0@T00<Si2?T8n@
Si2?T8n@0Xl01G6?LG5`04<02@L000L79P009g4HLAQa675>LTibCW9aLW5bLGMaMg5gLGMaMg9gLGMb
Mg5gLGMaM`02L@03674H009a029bLG9aLW5bLG9aT76@LI1bT7:@LY1bT7:@LY1bT7:@LY1b3Y001i6@
TI2AT94099006Xn@Si2?T8n@Si2?T8n@Si1aT76@LHmaSg6?0W400g1a001200T700071bL000YaCW5a
LW5bLG9a178067MbMg9gLWMbMg:@LWMbMg9gLY1bT7:@LPAa01AbLG9bRg:;LY1bT7:@LY1bT7:@LPJ@
04VAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT:n@[i2_T:n@UY2FT9J@UY2FT:n@[i2_T:n@[i2^T9F@
UI2ET9F@[Y2ET9F@UI2E00Z@00Z?T8n@Si2?T8n@0Xl3LD<02@L000L7:0008g4HLAQaCW9>LTibCW9a
LW5gLGMaMg5gLGMaMg5gLGMaMg5g009a02L0LAQa679aLW5bLG9aLW6;LI1aT76@LY1bT7:@LY1bT7:@
LY1bT78049001I6@TI2A02R@01R?T8n@Si2?T8n@Si1fT76@LI1aT76?LHl2L@03L74004802@L000L7
:0002aQaCW9aLW5bLG9a00Ab01AgLWMbMg9gLWMbMg9gLWMbMg:@LPIa00=bLG800g803X]bT7:@LY1b
T7:@LY1b1Y00BI6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@UY2FT9J@UY2FT9J@UY2FT9J@[i2_T:n@
[i2_T9F@UI2ET9F@UI2ET9F@UI2ET9D02Y0038n@Si2?T8n@MY1fS`=a@`091`001`LX001BL@1a674H
LDibCW9>LTibCW9aMg5gLGMaMg5gLGMaMg5gLGL0LAQa674HLW5bLG9aLW5bLG9aT76@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7N@M`b@00>AT9009Y006Xn@Si2?T8n@Si2?T8n@MY1aT76@LI1aSg6?0W400g1a
001200T700071bT000YaCW5aLW5bLG9a0g805G5gLWMbMg9gLY1bMg9gLWMbT7:@LP04L@04LW5bL@9b
012;LY1bT7:@LY1bT7:@LY1b1900Bi6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2FT9J@UY2FT96@
UY2FT9J@UY2FT9J@UY2FT9J@UI2ET9F@UI2ET9F@UI2ET9F@U@0:T00<Si2?T8n@Si1fT7J?0g5300T7
00071bX0051a674HLDibCW9>LTibCW9aLW5gLGMaMg5gLGMaMg5gLGL0LAQa675/LW5bLG9aLW5bLI1a
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1gT7N@McN@01R?T8n@Si2?T8n@Si1fT76@LI1aT76?LHl2L@03
L74004802@L000L7:`04L@0HLW5bLG9aLW5gLGMbMg9gLWMbMg9gLWMb0WL00i1bL@03L@04LW5bL@Ab
00n@LY1bT7:@LY1bT7:@Mi400i00?i6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2FT9J@UY2FT96@
TI2AT96@UY2FT9J@UY2FT9J@UI2ET9F@U@03T007UI2ET9F@U@0<T00?Si2?T8n@Si1fT76?LGIa04<0
2@L000L7;000;W4HLDiaCW9>LTibCW9>LW5gLGMaMg5gLGMaMg5gLGLHLAQa678HLW5bLG9aLW42LP0N
Mg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@Mi1g=I006WJ@MY1fT8n@MY1fT7J@MY1aT76@LI1aSg5f
0W400g1a001200T700071b`00g40779aLW5bLG9aMg5gLWMbMg9gLWMbMg9gLY1bT782L@06LW5bLG9a
17803i1bT7:@LY1bT7:@LY1gT@03T00mTI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT9J@UY2FT9J@
TI2AT96@TI2AT9J@UY2FT9J@UY2ET9F@U@03T007UI2ET9F@U@0>T00?Si2?T8n@MY1fT76?LGIa04<0
2@L000L7;@00:aQa675>LDibCW9>LTibLGMaMg5gLGMaMg5gLGMaMaQa674HLQQbLG9aLW401780991b
T7:@LY1bT7:@LY1bT7:@LY1gT7:@LY1gT7N@Mi1gT7N@M`F@00EgT92@M`0QT00NMi1gT7N@MY1fT7J@
MY1fT7J@MY1aT76@LI1aSg5f0W400g1a001200T700071bl000aaCW9>LTibLG9aLW42LP0>Mg9gLWMb
Mg9gLWMbMg82L@06LW5bLG9a0W80@GMbT7:@LY1bT7:@LY1gT7N@Mi5gTGNAT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2F00f@00>ET9005Y0048n@Si2?T8n@Si1fSg6?LGI400T7
00071c0005Ia678HLTibCW9>LW5bLG9aMg5gLGMaMg5gLGLHLAQb679aLW5bLG9aT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@M`N@00UgT7N@Mi1gT7L02i002GN@Mi1gT7N@
M`07T00HMi1gT7J@MY1fT7J@LI1fT7J@LI1aSg5f0W400g00001200T700071c4000YaCW9>LW5bLG9a
0W803WMbMg9gLWMbMg9gLWMb0W401W9aLW5bL@9b046@LY1bT7:@LY1bT7:@Mi1gT7NAMi5gTGNAT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@T@0VT00@Si2?T8n@MY1fT7J?LGIaMT@0
2@L000L7<P00VG4HLTibCW9>LTibLG9aMg5gLGMaMg5gLGLHLAQaCW9aLW5bLGMaMg:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@MY1fT7J@LI1aT7J@LI1aT76?LGI`L@1500T700071c<000YaCW9a
LW5bLG9a0W8037MbMg9gLWMbMg9gLP9a00IbLG9aLW42LP0aMg:@LY1bT7:@LY1bT7N@Mi1gT7NAMi5g
TGNAT96@TI2AT96@TI2AT96@TI2AT96@T@03T009TI2@T96@TI2A00>@00>AT9008Y004GN@MY1fT7J@
MY1fT7J?LGIa04D02@L000L7=000UW4HLTibCW9>LW5bLGMaMg5gLGMaMg5g674HLTibLG9aLW5gLGMb
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7J@MY1fT7J@LI1aT76@LHmaSg5fL4H02@L000L7
=@02L@0DLW5bLG9aLW5gLGMbMg9gLWMbMg82L@04LW5bL@Ab01egLY1bT7:@LY1bT7N@Mi1gT7NAMi5g
TGNAMi6@T@03T00?TI2@T95gTGNAMi5gTGNA00Z@00=gTGL00i000gN@T00MT00FMi1gT7N@Mi1fT7J@
MY1fT7J@LHmaMTH02@L000L7=P00Tg5>LDibCW9aLW5gLGMaMg5gLGMaMaQa679>LW5bLG9aMg5gLI1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@MY1fT7J@MY1aT76@LI1aT76?LGIaMP1700T700071cH0
01E>LG5bLG9aLW5bLGMaMg9gLWMbMg800W40179aLW42LP1<Mg9gLWMbT7:@LY1bT7:@Mi1gT7N@Mi5g
TGNAMi5gT7N@Mi1gT7N@Mi5gT7NAMi1gTGN@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@M`>@00MgT7N@
Mi1g00^@01EgT7N@Mi1fT7J@MY1fT7J@MXmaMW40A`091`001`Lg002A675>LDibCW9aLW5gLGMaMg5g
LGLHLAQbCW9aLW5bLG9aMg9gLY1bT7:@LY1bT7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1bT7:@LY1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1fT7J@MY1fT76@LI1aT76@
LI1aMW5fL@1800T700071cT0019aCW9aLW5bLGMbMg9gLWMbMg82L@06LW5bLG9a0W80LgMbMg9gLY1b
T7:@LY1gT7N@Mi1gT7N@Mi1gT7NAMi5gT7NAMi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7J@MY1aT76?LGH0B0091`001`Lj002>
LAQa675>LTibCW9aLW5bLGL0LAQbCW9>LW5bLG9aLW5gLWMbT7:@LY1bT79gLY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bMg9gLWMbMg9gLI1aT76@LI1aT7:@LY1bT7:@LI1aT76@LI1a
T76@LI1aT76@LI1aT76@LGIaMW5fLDP02@L000L7>`00275>LW5bLG9a0W801WMbMg9gLP9a00AbLG9a
1780LWMbMg9gLY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7J@LI1aT75fLDT0
2@L000L7?00078XHLAQaCW9>LTibLG9aM`1a678HLTibLG9aLW42LP1/Mg9gLWMbT7:@LY1bMg9gLWMb
Mg:@LY1bT7:@LY1bT79gLWMbMg9gLWMbMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMa
Mg6@LI1aT76@LI1aMg6@LGMaT75gLGMaMg5fLGIaBP091`001`Ln000<CW9aLW5bLG9aMg9g1G40179a
LW42LP0[Mg9gLWMbMg:@LY1bT7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT005M`08T7N@LY1b
Mg8:M`09T7N@Mi1gT7N@00=g02:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7J@MY1aT76@LGMaB`091`00
1`Lo0024675>LDibCW9>LTib074HLAQbCW9aLW5bLG9aMg9gLWMbMg9gLWMbMg9gLWMbMg9gLY1bMg9g
LGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaT75g
LI1aMg5gLGMaMg5gLGMaMW5f0W5;00T700071d000g402W9aLW5bLGMaMaP2L@04LW5bL@9b02EgLWMb
Mg9gLY1bT7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@00Qg01mbMg9gLWMbMg9gLWMbMg9gLWMbMg9g
LWMbMg9gLWMb00Qg01^@Mi1gT7N@Mi1gT7N@Mi1gT7J@MY1aT75gLGH0C0091`001`M0000HLAQaCW5>
LTibCW9>LAQa679>LW5bLG9a0W80IgMbMg5gLWMbMg9gLWMbMg9gLWMbMg9gLGMaMg5gLGMaMg5gLGMa
Mg5gLGM>MdibCY2:T8Z@RY2:T4ibCWM>MdigCWMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMa
Mg5gLGH00W5=00T700071d4017401g9aLW5gLGL00g40179aLW42LP0IMg9gLWMbMg:@LY1bT7:@LY1g
T7N@Mi1gT003M`08T7:@LY1bMg83M`07LWMbMg5gR`0ET00=SgMaMg5gLGMaMg9gLP05M`0HLWMbT7N@
Mi1gT7N@MY1fMgIgMWMaMg5fCP091`001`M3001l674HLDiaCW9>LP1a678HLW5bLG9aLW5bLGMaMg5g
LWMbMg9gLWMbMg9gLWMaMg5gLGMaMg5gLGMaLTj@LI1aT8Z@RY2:T8Z@RY2?T8n@Si2?T8n@RY2:T8Z@
CWM>Mg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMP9a00==0000C0091`001`M70007LDibLG9a
LP03L@0BLW5bLGMbMg9gLWMbMg9gLWMb0gL06g:@LY1bT79gLWMbMg9gLWMaMg6;LY2;T8^@R`0<T007
ZI2YT:V@Z@06T004Si2?T0:?00n:Mg5gLGMaMg5gLWMbMg800gL037IgLGMaMg5gLGMaM`9aD0091`00
1`M9000k675>LTib674HLTibCW9aMg5gLGMbMg9gLWMbMg9gLGMaMg5gLGMaMg5gLGMaMg6;LI1aT76@
LY2;T8/03i00:Xn@Si2?T8n@Si1aSg6?LGM>MdigLGMaMg5gLGMaMg5gLGMaMg5gLGMaMP9aD@091`00
1`M:0004CW5aLP=a00AbLG9a0W809WMbMg9gLWMbMg9gLWMbMg9gLWMbMg9gLWMbMg5gLG9aRg:@Ri2;
49001jV@ZI2YT:T0290018n@Si02S`0LLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMU802@L000L7
C000=W4HLDia675>LTibCWMaMg5gLWMbMg9gLWMbMg5gLGMaMg5gLGMaMfabLH]aRg6@LY1bT8^@Ra>@
02V?T8n@Si2?T8n@LHmaSdigCWM>Mg5gLGMaMg5gLGMaMg5gLGMaMTiaC@1B00T700071dh000=>LG40
0W40:79aLW5gLWMbMg9gLWMbMg9gLWMbMg9gLWMbMg9gLGMaLW6;LY1bT8/6T007TI2AT96@T@05T005
ZI2^T:h039007Xn@ShmaSg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMU@02@L000L7D0008W4HLDibCW9>
LTigLGMaMg5gLGMaMg5gLGMaMg5gLGMaMaP2L@0<LW6;LI1bT7:@LY2;6I008Xn@Si1fT76@LHm`Mdig
CWM>Mg5gLGMaMg5gLGMaMTifCW5E00T700071e8001i>LW5bLG9aMg9gLWMbMg9gLWMbMg9gLWMbMg9g
LGL3L@9b00B;LY1b19005I6@TI2AT96@TI2AT96@TI2FT9F@U@0<T00LSi2?T75fLGMaMg5gLGMaMg5g
LGMaMg5gLGIaMUH02@L000L7D`0071QbCW9>LTibCWMaMg5gLGMaMg5gLGMaMg5gLGL2L@0?K79aLW6@
LY1bT7:@LY1g00:@00FAT96@T@0DT00NSi1fT76@LHm`MdigCWMaMg5gLGMaMg5gLGI>LDiaE`091`00
1`MC000PLDibLG9aMg5gLWMbMg9gLWMbMg9gLWMbMg:@LY1aLW42LP04T7:@LP:@01^AT96@TI2AT96@
TI2FT9J@UY2FT9J@UI2ET9D029006Hn@Si1aMW5gLGMaMg5gLGMaMg5gLGMaMW40F@091`001`MD000K
LDibCW9>LTigLGMaMg5gLGMaMg5gLGMaMg5g00=a011bLH]bT7:@LY1bT7:@Mi1g6I0057J@MY1aT76?
L7M>MdigCW9aLW5g17400dh0001J00T700071eD00W403W9aLW9gLWMbMg9gLWMb27L0191gT742LP0I
T7:@LY1gTI2AT96@TI2AT96@TI2AT96@UP0BT00CSi1fSg5fLGMaMg5gLGMaMg5fL@1O00T700071eH0
04]a679>LTibCWMaMg5gLGMaMg5gLGMaMg5gLG9aLW5gLY1bT7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7J@MY1aT76@LI000W4034ibCW5>LDiaCW5>LF402@L000L7E`02L@0<LW5bLGMbMg9gLWMb
1WL01Y1gT7N@L@9b00J@LY1bT7L4T00=TI2AT96@TI2AT96@T@0@T006Mi1gT7J@0WH0275gLGMaMg5g
0W400dh0001Q00T700071eP004Ua675>LTibCWMaMg5gLGMaMg5gLGMaMg5bCW9aMg:@LY1bT7:@LY1b
T7:@Mi1bT7:@LY1bT7N@Mi1gT7N@Mi1fT76@LI1aT76@009a00M>LDiaCW5>06H02@L000L7F@02L@0:
LW5gLGMbMg9gLPIg00J@Mi1gT742LP0RMg:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@M`>@
00igT7N@MY1fT75fLG9aLVX02@L000L7F`00AQQaCW9>LTigLGMaMg5gLGMaMg5gLG9>LW5gLWMbT79g
LY1bT7:@LWMbT7:@LY1bT7:@LY1bT7:@LI1aT76@LI1aT76@LGI_00T700071e`000]aLW5gLGMaMg5g
LP06M`04T7N@L@9b02igLY1bT7:@Mi1gT7N@Mi1gT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7J@LI1a
K`091`001`MM0013679>LTibCW9>Mg5gLGMaMg5gLG9>LW5gLGMaMg5gLWMaMg5gLGMaMdigCWM>Mg5g
LGMaMg5gLGMaMg5gLGMaMg5gL@1`00T700071eh000]>LW5gLGMaMg9gLP02M`06T7N@Mi1a0W801gMb
Mg9gLY000WL01W9gLWMbM`9a00AbT79g0g401W9aMg9gLPIg00V@MY1fT76@LGH0L0091`001`MP000M
LAQbCW9>LW5gLGMaMg5gLGLHLTigLGMaMg5gLGL00g407I2;Sg1gLGMaLTibCW5>LTibLGMaMg5gLGMa
Mg5g009aL@091`001`MT00=a00QgLGMaMg9gLP=g00]aLW9gLWMbMg5gL@05T008RY1bMg9gLWL2LP0A
LGMaMg5gLGMaMg5gLGMaMg40L`091`001`MV000HLDibCW9>Mg5gLGMaMdibLGMaMg5gLI1f0i00676@
LHYbMg5gLGMaLTigCWM>MdigLGMaM`9a00=>0000LP091`001`MX000=CW9aMg5gLGMbMg9gL@02LP06
Mg9gMi1g0Y006iJ@T8^@RWMbMg9gLWMbMg5gLGMaMg5gLGMaMP1f00T700071f`002abCW9>LW5gLGMa
LTigLI1aT7N@Mi1bT8Z@LGMaMg5gLGM>MdigCWM>MdigCWP02@L000L7K`009g9aLW5gLGMbMg5gLY1g
T7N@Mi1aT75gLWMbMg9gLWMaMg5gLGMaMP1j00T700071g80025aCW9>LTibLG9aMg5gLGMaMdigLGMa
Mg5gLG9>MdibCW80O@091`001`Mf000ILG9aLW5gLGMaMg5gLGMaMg5gLGMaLW5gL@2100T700071gd0
00e>LDiaCW5>LDiaCW5>08H02@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T7
00071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T7
00071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T7
00071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T7
00071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T7
00071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T7
00071ol04@091`001`Oo01402@L000L7o`0A00T7003o1b47003o1b47003o1b47003o1b47003o1b47
003o1b470000\
\>"],
  ImageRangeCache->{{{0, 287}, {285.875, 0}} -> {-8.12139, -7.50157, 1.10189, 
  1.10189}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "The ",
  StyleBox["MVFlat",
    FontFamily->"Courier"],
  " value looks better for polyhedra."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(Needs["\<Graphics`Polyhedra`\>"]\)], "Input",
  CellLabel->"In[9]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; \n
    MVShow3D[\n\tGraphics3D[Stellate[Icosahedron[]]], \n\t
      MVPolygonShading -> MVFlat\n\t]; \nMVPasteGraphics[]; \)], "Input",
  CellLabel->"In[14]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F505F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F5F5F705F606F5F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F5F606F505F6F5F5F6F505F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F5F606F505F705F606F5F5F705F606F5F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F505F6F5F5F6F505F6F5F606F505F6F5F5F6F505F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F5F5F705F606F5F5F705F606F5F5F705F606F5F5F705F606F5F5F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F505F6F5F606F505F6F5F606F505F6F5F606F5F5F705F606F505F6F5F60
6F5F5F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000705F606F5F5F705F606F5F5F705F606F5F5F705F606F5F5F705F606F5F5F
705F60000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
6F5F5F6F505F6F5F606F505F6F5F606F5F5F6F5F606F505F6F5F606F5F5F705F60
6F505F705F60000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
6F5F5F705F606F5F5F705F606F5F5F705F606F5F5F705F606F5F60705F606F5F5F
705F606F5F60705F60000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000006F5F5F\
6F5F606F505F6F5F606F5F5F705F606F505F6F5F606F5F5F705F606F5F5F705F60
6F5F5F705F606F5F5F705F60000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008070608F70600000000000006F5F5F705F60\
6F5F5F705F606F5F5F705F606F5F5F705F606F5F5F705F606F5F5F705F606F5F60
705F606F5F5F705F606F5F60705F60000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F30508070608070608070600000006F5F606F505F\
6F5F606F5F5F6F5F606F505F6F5F606F5F5F705F606F505F705F606F5F5F705F60
6F5F5F705F606F5F5F705F606F5F5F705F60000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F3F5F3F3F508F706F8070608F7060807060705F60\
6F5F60705F606F5F5F705F606F5F60705F606F5F5F705F606F5F60705F606F5F60
706060705F60705F606F5F60706060705F60705F60000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F3F5F3F30503F3F5F807060807060807060807060807060\
705F606F5F5F6F5F606F5F5F705F606F5F5F705F606F5F5F705F606F5F5F705F60
6F5F5F705F606F5F5F705F606F5F5F705F606F5F5F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F3F503F3F5F3F3F508F70608070608F7060807060807060\
8070608070606F5F60705F606F5F5F705F606F5F60705F606F5F5F705F606F5F60
705F606F5F60706060705F60705F60705F60706060705F60000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003030503F3F503F30503F3F5F807060807060807060807060806F5F\
8070608070608070606F5F5F6F5F606F5F5F705F606F5F5F705F606F5F5F705F60
6F5F5F705F606F5F5F705F606F5F5F705F606F5F5F705F606F5F5F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F3F5F3F3F503F3F5F3F3F508F7F6F8070608F70608070608F7060\
807060807060807060807060807060705F606F5F60706060705F60705F606F5F60
706060705F60706060705F60706060705F60706060705F60706060705F60000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F3F503F30503F3F5F3F30503F3F5F807060807060807060807060807060\
807060807060807060807060807060806F5F705F606F5F60705F606F5F5F705F60
6F5F60705F606F5F5F705F606F5F60705F606F5F5F705F60705F607060606F5F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F3F503F3F5F3F3F503F3F5F3F3F508F70608070608F7060807060807060\
8070608070608070608070608070608070608070608070606F5F60706060705F60
705F60705F60706060705F60706060705F60706060705F60706060705F60706060705F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003030503F3F503030503F3F503F30503F3F5F807060807060807060807060807060\
807060807060807060806F5F80706080705F8070607F6F5F8070606F5F60705F60
6F5F5F705F606F5F60705F606F5F5F705F60705F60705F606F5F5F705F60705F60706060705F5F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F8070608F70608070608F7F6F\
807060807060807060807060807060807060807060807060807060807060807060
706060705F60706060705F60706060705F60706060705F60706060705F60706060705F60706060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F3F503030503F3F503F30503F3F5F3030503F3F5F3F3F50807060807060807060807060\
807060807060807060807060807060806F5F8070608070608070607F6F5F807060
7F6F5F705F606F5F60705F60705F607060606F5F60705F60705F60706060705F60706060705F60\
706060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F30503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F8070608F7F6F807060807060\
8070608F7060807060807060807060807060807060807060807060807060807060
80706080705F807060705F60706060705F60706060705F60706060705F60706060705F60706060\
705F607F6060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3030503F3F503030503F3F503030503F3F503F30503F3F5F303050807060807060807060807060\
80706080706080706080705F8070608070608070607F6F5F8070607F6F5F807060
7F6F5F8070607F6F5F8070606F5F60705F60705F60706060705F60705F60705F60706060705F60\
706060705F60706060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F8070608F7F6F8070608F7F6F\
8070608070608070608F7060807060807060807060807060807060807060807060
80706080706080706080705F807060807060706060705F6070606F705F60706060705F607F606F\
706060706060705F607F606F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000605F405F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F3F50
3030503F3F503030503F3F503F30503F3F5F3030503F3F5F3F3F508F7060807060807060807060\
8070608070608070608070608070608070608070608070608070607F6F5F807060
7F705F8070607F6F5F8070607F6F5F8070607F6F5F706060705F60706060705F60706060705F60\
706060705F60706060705F60706060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F605F405F5040605F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F3050
3F3F503F30503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F8070608F7F6F807060807060\
8070608F7F60807060807060807060807060807060807060807060807060807060
80706080705F8070608070608070607F6F5F8070607F705F706060705F607F606F705F60706060\
705F607F606F7060607F6060705F607F606F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F2F2F2F2F605F4F5F5040605F405F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3F30503F3F5F303050807060807060807060807060\
8070608070608070608070608070608070608070607F6F5F80706080705F807060
7F6F5F8070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F807060705F60706060705F60\
706060705F60706060705F60706060705F60706060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F2F5F5040605F405F5040605F405F50405F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F3F503F3050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F8F7F6F8070608F7F6F\
8070608070608070608F7F6F807060807060807060807060807060807060807060
8070608070608070608070608070608070608070607F705F8070607F705F807060705F607F606F\
7060607F606F7060607F606F7060607F606F706060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F302F2F605F40605F4F605F40605F405F5040605F4F5F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F3F503030503F3F50
3030503F3F503030503F3F503F30503F3F5F3030503F3F5F3F3F503F3F5F807060807060807060\
8070608070608070608070608070608070608070608070608070607F705F807060
8070608070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F\
706060705F60706060705F607F6060705F60706060705F60000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F2F
2F2F2F302F2F605040605F40605040605F405F5040605F405F5040605F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003030503F3F5F3F3050
3F3F503F30503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F508F7F6F807060807F6F\
8070608F7F6F807060807060807060807060807060807060807060807060807060
8070608070608070608070608070607F705F8070607F705F8070607F6F5F8070607F705F807060\
7F6F5F7F606F7060607F606F7060607F606F7060607F606F706060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F2F2F2F2F
302F2F605F40605F4F605F40605F4F605040605F4F605040605F405F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3F30503F3F5F3030503F3F5F807060807060807060\
8070608070608070608070608070608070608070607F7060807060807060807060
7F6F5F8070607F705F8070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F\
7F705F7F6F5F7F705F705F60706060705F607F6060705F60706060705F60000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F2F2F2F2F2F2F302F2F
605040605F40605040605F405F5040605F405F5040605F405F50405F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F3F503F30503F3F503F3050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F8F7F6F8070608F7F6F\
807060807F6F8070608F7F6F807060807060807060807F60807060807060807060
8070608070608070608070608070608070608070607F705F8070607F70608070607F6F5F807060\
7F705F8070607F6F5F8070607060607F606F7060607F606F7060607F606F706060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F2F2F2F2F302F302F2F2F6F5F4F605F40
6F5F4F605F40605F4F605F40605F4F605F40605F4F5F5040605F4F5F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F3F503030503F3F503030503F3F50
3030503F3F503030503F3F503F30503F3F5F3030503F3F5F3F3F503F3F5F807060807060807060\
8070608070608070608070608070608070608070608070608070607F7060807060
8070608070607F6F5F8070607F705F8070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F\
8070607F6F5F7F705F7F6F5F7F70607F6F5F7F60607060607F606F705F607F606F
706060000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F2F2F2F2F2F2F2F2F2F2F6F5F4F605F40605F4F
605F40605F4F605040605F40605040605F405F5040605F405F5040605F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003030503F3F503030503F3F5F3F3050
3F3F5F3F30503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F8F7F6F8070608F7F6F\
8070608F7F6F807060807060807060807F6F807060807060807060807060807060
8070608070608070608070608070607F70608070607F70608070607F705F8070607F705F807060\
7F6F5F8070607F6F5F8070607F6F5F8070607F6F5F7F606F7060607F606F706060
7F606F706060000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F2F302F2F2F2F2F302F302F2F2F6F5F4F605F406F5F4F605F40
605F4F605F40605F4F605F40605F4F605040605F4F605F40605F405F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003030503F3F503030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3F30503F3F5F3030503F3F5F80706080706F807060\
8070608070608070608070608070608070608070607F7060807060807060807060
7F705F8070607F70608070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F8070607F6F5F\
7F70607F6F5F7F70607F6F5F7F705F7F6F5F7F705F7F6F5F7F6F5F7060607F606F
705F607F606F000000000000000000000000000000000000000000000000000000000000000000\
2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F6F5F4F605F40605F4F605F40605F4F
605F40605F40605F40605F405F5040605F405F5040605F405F5040605F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F3F503030503F3F503F30503F3F503F3050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F8070608F7F6F\
8070608F7F6F8070608F7F6F807060807F60807060807F6F807060807060807060
8070608070608070608070608070608070608070607F70608070607F70608070607F705F807060\
7F705F8070607F6F5F8070607F705F8070607F6F5F8070607F6F5F7F7060706060
7F606F7F60607F606F000000000000000000000000000000000000000000000000000000302F2F\
2F2F2F302F2F2F2F2F3030306F5F4F6F604F605F4F6F604F605F4F6F5F4F605F40
6F5F4F605F40605F4F605F40605F4F605F40605F4F5F5040605F4F5F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F3F5030304F3F3F503030503F3F503030503F3F50
3030503F3F5F3030503F3F503F30503F3F5F3030503F3F5F3F3F503F3F5F3F3050807060807060\
807F6F8070608070608070608070608070608070608070608070607F7060807060
8070608070607F705F8070607F70608070607F6F5F8070607F705F8070607F6F5F8070607F6F5F\
8070607F6F5F7F70607F6F5F7F70607F6F5F7F705F7F6F5F7F705F7F6F5F7F705F
7F6F5F7F606F7060607F606F0000000000000000000000000000000000002F202F2F2F2F2F2F2F\
302F2F2F2F2F302F2F605F4F6F604F605F406F5F4F605F406F5F4F605F40605F4F
605F40605F4F605F40605F40605F40605F405F5040605F405F5040605F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000402F2F402F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003030503F3F503030503F3F503030503F3F5F3F3050
3F3F5F3030503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F8070608F7F6F\
8070608F7F6F807060807F6F807060807F6F807060807060807060807060807060
8070608070608070608070608070607F70608070608070608070607F705F8070607F7060807060\
7F6F5F8070607F705F8070607F6F5F8070607F6F5F7F70607F6F5F8070607F6F5F
7F705F7F6F5F7F606F7F606F7F606F0000000000000000002F2F2F2F2F2F302F2F2F2F2F302F2F\
2F2F2F3030306F5F4F6F604F605F4F6F604F605F4F6F604F605F406F604F605F40
605F4F605F40605F4F605F40605F4F605F40605F4F605F40605F405F5040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2030402F2F
3F2F20402F2F3F2F2F000000000000000000000000000000000000000000000000000000000000\
00000000000000000030304F3F30503030503F3F503030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3F30503F3F5F3030503F3F5F3F3F50807F6F807060\
807060807060807060807060807060807060807060807060807060807060807060
7F70608070607F70608070607F6F5F8070607F705F8070607F6F5F8070607F6F5F8070607F6F5F\
7F70607F6F5F7F70607F6F5F7F705F7F6F5F7F70607F6F5F7F705F7F6F5F7F705F
7F6F5F7F6F5F7F6F5F7F6F5F7060607F606F0000002F2F2F2F202F2F2F2F2F2F2F302F2F2F2F2F\
302F2F6F5F4F6F604F605F406F5F4F605F406F5F4F605F40605F4F605F40605F4F
605F40605F40605F40605F4F5F5040605F405F5040605F405F5040605F405F5040000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F3F2F2F3F
40302F402F2F402F2F402F2F40302F402F2F402F2F000000000000000000000000000000000000\
0000000000000000003F3F503030503F3F503030503F3F503F30503F3F50303050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F8070608F7F6F\
8070608F7F6F8070608F7F6F807060807F6F807060807F6F807060807060807060
807F608070608070608070608070608070608070607F70608070608070608070607F705F807060\
7F70608070607F705F8070607F705F8070607F6F5F8070607F6F5F7F70607F6F5F
8070607F6F5F7F705F7F6F5F7F70607F606F7F6F6F2F2F2F302F302F2F2F302F302F2F2F6F604F\
6F604F6F604F6F5F4F6F604F6F5F4F6F604F605F4F6F604F605F4F6F604F605F40
6F604F605F40605F4F605F40605F4F605F40605F4F605F40605F4F5F5F40605F40000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F3F
2F203F402F2F3F2F2F402F2F402F2F402F2F3F2F2F402F2F3F2F2F402F2F000000000000000000\
0000000000003F30503030503F3F503030503F3F503030503F3F503030503F3F50
3030503F3F5F3030503F3F5F3F30503F3F5F3030503F3F5F3F3F503F3F5F3F3050807F6F807060\
807F6F807060807060807060807060807060807060807060807060807060807060
8070608070607F70608070607F70608070607F705F8070607F705F8070607F6F5F8070607F6F5F\
8070607F6F5F7F70607F6F5F7F70607F6F5F7F705F7F6F5F7F70607F6F5F7F705F
7F6F5F7F705F7F6F5F7F6F5F7F6F5F7F705F706F5F7F606F2F2F2F302F2F2F2F2F6F604F6F5F4F\
6F604F605F4F6F604F605F4F6F604F605F406F5F4F605F406F5F4F605F40605F4F
605F40605F4F605F40605F40605F40605F4F5F5040605F405F5040605F405F5040000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F3F
2F2F3F2F2F3F40302F402F2F402F2F402F2F40302F402F2F402F2F402F2F402F2F402F2F402F2F\
0000000000003030503F3F503030503F3F503030503F3F503030503F3F5F3F3050
3F3F5F3030503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F8070608F7F6F\
8070608F7F6F807060807F6F807060807F6F807060807060807060807F6F807060
8070608070608070608070608070607F70608070608070608070607F70608070607F7060807060\
7F705F8070607F705F8070607F6F5F8070607F705F7F70607F6F5F8070607F6F5F
7F70607F6F5F7F70607F6F5F7F705F7F6F5F7F705F7F6F5F302F302F2F2F7060506F604F6F604F\
6F5F4F6F604F6F5F4F6F604F605F4F6F604F605F4F6F604F605F406F604F605F40
605F4F605F40605F4F605F40605F4F605F40605F4F605F40605F405F5040605F40000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F3F
2F20302F2F3F402F2F402F2F402F2F402F2F402F2F402F2F3F2F2F402F2F402F2F402F2F3F2F2F\
402F2F30304F3F3F5030304F3F30503030503F3F503030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F50807F6F807060\
807060807060807F6F807060807060807060807060807060807060807060807060
7F70608070607F70608070607F705F8070607F70608070607F6F5F8070607F705F8070607F6F5F\
7F70607F6F5F8070607F6F5F7F70607F6F5F7F70607F6F5F7F705F7F6F5F7F705F
7F6F5F7F6F5F7F6F5F7F705F706F5F7F6F5F7F6F5F7F6F5F706F5F402F30806F6F6F604F6F5F4F\
6F604F6F5F4F6F604F605F406F604F605F406F604F605F40605F4F605F40605F4F
605F40605F4F605F40605F4F5F5040605F405F5040605F405F5040605F405F5040000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F3F
2F2F3F2F2F3F2F2F3F402F2F4F3030402F2F40302F402F2F4F3030402F2F40302F402F2F40302F\
402F2F3F3F503030503F3F503030503F3F503030503F3F5F3F3F503F3F5F303050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F8F7F6F\
8070608F7F6F8070608F7F6F807060807F6F807060807F6F807060807F60807060
807F6F807060807060807060807F608070608070607F70608070608070608070607F7060807060\
7F70608070607F705F8070607F705F8070607F6F5F8070607F705F7F70607F6F5F
8070607F6F5F7F70607F6F5F7F70607F6F5F7F705F7F6F5F403030402F30403030402F30403030\
402F308F7070806F6F8F70706F5F4F6F604F605F4F6F604F605F4F6F604F605F40
6F604F605F40605F4F605F40605F4F605F40605F4F605F40605F4F605F40605F40000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F3F
2F203F2F2F3F2F203F2F2F3F402F2F402F2F402F2F402F2F402F2F402F2F402F2F402F2F402F2F\
3F3F5030304F3F30503030503F3F503030503F3F503030503F3F503030503F3F50
3030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F503F3F5F3F30503F3F5F807060\
807F6F807060807060807060807F6F807060807060807060807060807060807060
8070608070607F70608070607F70608070607F705F8070607F70608070607F6F5F8070607F705F\
8070607F6F5F7F70607F6F5F8070607F6F5F7F70607F6F5F7F70607F6F5F7F705F
7F6F5F7F705F7F6F5F7F705F7F6F5F7F705F706F5F402F303F2F304030303F2F30402F30402F30\
403030402F30402F30402F30403030806F6F807070806F6F807070806F6F807070
605F40605F4F605F40605F4F605F40605F4F5F5040605F405F5040605F405F5040000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F3F
2F2F3F2F2F3F2F2F3F2F2F3F2F2F3F402F2F4F3030402F2F4F302F402F2F4F3030402F2F40302F\
30304F3F3F503030503F3F503030503F3F503030503F3F503030503F3F5F3F3050
3F3F5F3030503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F8F7F6F\
8070608F7F6F807060807F6F807060807F6F807060807F6F807060807F6F807060
807060807060807F608070608070607F70608070608070608070607F70608070607F7060807060\
7F705F8070607F70608070607F705F8070607F705F7F70607F6F5F8070607F6F5F
7F70607F6F5F7F70607F6F5F7F705F7F6F5F403030402F30403030402F30403030402F30403030\
402F30403030402F30403030402F304F3030402F30403030806F6F8F707080706F
8F7070806F6F8F707080706F8F7070806F6F8F7070605F40605F405F5040605F4F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F20302F2F3F2F203F2F2F3F2F203F2F2F3F402F2F4F2F2F402F2F402F2F402F2F402F2F30304F\
30305030304F3F3F5030304F3F30503030503F3F503030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F503F3F5F807060\
807F6F807060807F6F807060807060807060807060807060807060807060807060
7F70608070607F70608070607F70608070607F70608070607F705F8070607F705F8070607F6F5F\
7F70607F6F5F8070607F6F5F7F70607F6F5F7F70607F6F5F7F705F7F6F5F7F7060
7F6F5F7F705F7F6F5F7F705F706F5F402F303F2F30402F303F2F2F402F30402F30402F30402F30\
402F30402F30403030402F30402F30402F30403030402F30402F30402F304F3030
402F3080706F806F6F807070806F6F80706F806F6F8F7070806F6F80706F806F6F8F7070806F6F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F3F2F2F3F2F2F3F2F2F3F302F402F2F3F302F40402F2F4F3030402F2F4F3030402F2F3F3F50\
3030503F3F503030503F3F503030503F3F503030503F3F5F3F3F503F3F5F303050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F8F7F6F\
8070608F7F6F8070608F7F6F807060807F6F807060807F6F807060807F6F807060
807F6F807060807F60807060807F608070608070607F70608070608070608070607F7060807060\
7F70608070607F705F8070607F70608070607F705F8070607F705F7F70607F6F5F
8070607F705F7F70607F6F5F7F7060402F30403030402F30403030402F30403030402F30403030\
402F30403030402F304F303F402F304F3030402F304F303F402F304F3030402F30
4F303F402F304F303F402F304F303F80706F8F707080706F8F707080706F8F707080706F8F7070\
80706F8F707080706F8F707080706F8F707080706F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F20302F2F3F2F203F2F2F3F2F2F3F2F2F3F2F2F3F4F2F2F402F2F4F302F402F2F30305030304F\
3F3F5030304F3F30503030503F3F503030503F3F503030503F3F503030503F3F50
3030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F503F3F5F3F30503F3F5F807060\
807F6F807060807F6F807060807F6F807060807060807060807F60807060807060
8070608070607F70608070607F70608070607F70608070607F70608070607F705F8070607F705F\
8070607F6F5F7F70607F705F8070607F6F5F7F70607F6F5F7F70607F6F5F7F705F
7F6F5F7F70607F6F5F7F705F3F2F304030303F2F30402F30402F30403030402F30402F30402F30\
403030402F30403030402F30403030402F30403030402F304F3030402F304F3030
402F304F3030402F304F3030402F304F3030402F304F3030402F308F7070806F6F8F7070806F6F\
8F7070806F6F8F706F80706F8F7070806F6F8F706F80706F8F7070806F6F8F706F
80706F8F7070000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F3F2F2F3F2F2F3F2F2F3F2F2F3F2F2F3F302F402F2F3F4F3030402F2F4F303030304F3F3F50\
30304F3F3F503030503F3F503030503F3F503030503F3F503030503F3F5F3F3050
3F3F5F3030503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F8F7F6F\
8070608F7F6F807060807F6F8070608F7F6F807060807F6F807060807F6F807060
807F60807060807F6F8070608070608070608070608070608070607F70608070607F7060807060\
7F70608070607F70608070607F705F8070607F705F8070607F6F5F8070607F705F
7F70607F6F5F7F7060402F30403030402F30403030402F30403030402F30403030402F30403030\
402F304F3030402F304F3030402F304F303F402F304F3030402F304F303F402F30
4F303F402F304F303F4F2F304F303F402F304F303F4F30304F303F4F2F304F303F4F30308F7070\
80706F8F707080706F8F707080706F8F707080706F8F707080706F8F707080706F
8F707080706F8F707080706F8F707080706F8F707080706F8F7070000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2020302F2F3F2F203F2F2F3F2F203F2F2F3F2F2F3F2F2F402F2F3F4F2F2F30304F30305030304F\
30305030304F3F3F5030304F3030503030503F3F503030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F807060\
807F6F807060807F6F807060807060807060807F6F807060807060807060807060
7F70608070607F70608070607F70608070607F70608070607F705F8070607F70608070607F6F5F\
7F70607F705F8070607F6F5F7F70607F6F5F7F70607F6F5F7F70607F6F5F7F7060
7F6F5F7F705F3F2F30402F303F2F30402F30402F30403030402F30402F30402F30403030402F30\
402F30402F30403030402F30403030402F304F3030402F30403030402F304F3030
402F304F3030402F304F3030402F304F3030402F304F3030402F304F30304F2F304F303F4F2F30\
4F30304F2F307F8F8F806F6F8F706F80706F8F7070806F6F8F706F80706F8F7070
806F6F8F706F80706F8F706F806F6F8F706F80706F8F706F806F6F8F706F80706F8F706F806F6F\
8F706F80706F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F3F2F2F3F2F2F3F2F2F3F302F402F2F3F302F402F2F3F302F402F2F403F3F5030304F3F3F50\
3030503F3F503030503F3F503030503F3F503030503F3F5F3F3F503F3F5F303050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
807F6F8F7F6F8070608F7F6F807060807F6F8070608F7F6F807060807F6F807060
807F6F807060807F6F807060807F6F807060807060807060807F608070608070607F7060807060\
7F70608070607F70608070607F70608070607F705F8070607F705F8070607F705F
807060402F30403030402F30403030402F30403030402F304F303F402F30403030402F304F303F\
402F304F3030402F304F303F402F304F303F402F304F303F4030304F303F402F30
4F303F4F30304F303F4F2F304F303F4F30304F303F4F2F304F303F4F30304F303F4F30304F303F\
4F30307F8F8F7F80808F707080706F8F707080706F8F70708F706F8F707080706F
8F70708F706F8F707080706F8F70708F706F8F707080706F8F70708F706F8F707080706F8F7070\
8F706F8F707080706F8F70708F706F8F70708F706F8F70708F706F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F3F2F203F2F2F3F2F2F3F2F2F3F2F2F3F5F3F3F503F305F403F503F3030305030304F\
3F3F5030304F3030503030503F3F503030503F3F503030503F3F503030503F3F50
3030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F503F3F5F3F30503F3F5F3F3F5F\
8F7F6F807060807F6F807060807F6F807060807F6F807060807F6F807060807060
8070608070607F70608070607F70608070607F70608070607F70608070607F70608070607F7060\
8070607F705F7F70607F705F8070607F6F5F7F70607F6F5F7F70607F6F5F7F7060
402F304030303F2F30402F30402F30403030402F30402F30402F30403030402F30403030402F30\
403030402F30403030402F304F3030402F304F3030402F304F3030402F304F3030
402F304F303F402F304F30304F2F304F303F402F304F30304F2F304F303F4F2F304F303F7F8080\
7F8F8F7F80807F8F8F80706F8F707080706F8F707080706F8F707080706F8F7070
80706F8F707080706F8F706F80706F8F707080706F8F706F80706F8F707080706F8F706F80706F\
8F707080706F8F706F80706F8F706F80706F8F706F80706F8F706F80706F8F706F
80706F8F706F80706F8F706F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F3F2F2F3F2F2F3F2F2F3F5F3F305F403F5F3F305F403F5F3F305F403F30304F3F3F50\
30304F3F3F503030503F3F503030503F3F503030503F3F503030503F3F5F3F3050
3F3F5F3030503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
8070608F7F6F807060807F6F8070608F7F6F807060807F6F807060807F6F807060
807F6F807060807F6F807060807060807060807F608070608070607F70608070607F7060807060\
7F70608070607F70608070607F705F8070607F70608070607F705F807060402F30
403030402F30403030402F30403030402F30403030402F30403030402F304F303F402F304F3030\
402F304F303F402F304F3030402F304F303F402F304F303F402F304F303F4F2F30
4F303F4F2F304F303F4F30304F303F4F2F304F303F4F30304F303F4F2F307F8F8F7F80807F8F8F\
7F80807F8F8F7F80808F707080706F8F707080706F8F707080706F8F70708F706F
8F707080706F8F70708F706F8F707080706F8F70708F706F8F707080706F8F70708F706F8F7070\
80706F8F70708F706F8F707080706F8F70708F706F8F70708F706F8F70708F706F
8F70708F706F8F70708F706F8F706F8F706F8F7F70000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F3F5F3F305F403F503F305F3F3F5F3F305F403F503F305F3F3F503F305F403F30304F\
30305030304F3F3F5030304F3030503030503F3F503030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F30305F\
807F6F807060807F6F807060807F6F807060807F6F807060807060807060807F60
7F70608070608070608070607F70608070607F70608070607F70608070607F70608070607F705F\
8070607F705F8070607F6F5F7F70607F705F7F70607F6F5F7F70603F2F30402F30
3F2F30402F30402F30403030402F30402F30402F30403030402F30402F30402F30403030402F30\
403030402F304F3030402F304F3030402F304F3030402F304F3030402F304F3030
402F304F3030402F304F303F402F304F30304F2F304F303F7080807F808F7F80807F808F708080\
7F808F7F80807F8F8F806F6F8F707080706F8F7070806F6F8F707080706F8F7070
80706F8F706F80706F8F707080706F8F706F80706F8F707080706F8F706F80706F8F707080706F\
8F706F80706F8F707080706F8F706F80706F8F706F80706F8F706F80706F8F706F
80706F8F706F80706F8F706F80706F8F706F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F3F30
5F403F5F3F3F5F403F5F3F305F403F5F3F3F5F403F5F3F305F403F5F3F3F5F403F5F3F305F403F\
3030503F3F503030503F3F503030503F3F503030503F3F5F3F3F503F3F5F303050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
807F6F8F7F6F8070608F7F6F807F6F807F6F8070608F7F6F807060807F6F807060
807F6F807060807F6F807060807F6F807060807F60807060807F6F8070608070607F7060807F60\
7F70608070607F70608070607F70608070607F70608070607F7060403030402F30
40303F402F30403030402F304F303F402F30403030402F304F303F402F304F303F402F304F303F\
402F304F303F402F304F303F4030304F303F402F304F303F4F30304F303F4F2F30
4F303F4F30304F303F4F30304F303F4F30307F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F\
7F80807F8F8F7F80808F70708F70708F707080706F8F70708F706F8F707080706F
8F70708F706F8F70708F706F8F70708F706F8F70708F706F8F7F708F706F8F70708F706F8F7F70\
8F706F8F70708F706F8F7F708F706F8F70708F706F8F7F708F706F8F70708F706F
8F7F708F706F8F70708F706F8F7F70000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F403F5F3F305F403F
5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F30\
5F403F30304F3030503030503F3F503030503F3F503030503F3F503030503F3F50
3030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F503F3F5F3030503F3F5F3F3F5F\
8F7F6F807060807F6F807060807F6F807060807F6F807060807F6F807060807060
807060807F6F7F70608070608070608070607F70608070607F70608070607F70608070607F7060\
8070607F705F8070607F70608070607F6F5F7F70607F705F4030303F2F30402F30
402F30403030402F30403030402F30403030402F30403030402F304F3030402F30403030402F30\
4F3030402F304F3030402F304F303F402F304F3030402F304F303F402F304F3030
4F2F304F303F402F304F303F7F80807F808F7080807F808F7F80807F808F7080807F808F7F8080\
7F8F8F7080807F808F80706F8F707080706F8F707080706F8F707080706F8F7070
80706F8F707080706F8F707080706F8F707080706F8F707080706F8F707080706F8F707080706F\
8F707080706F8F706F80706F8F707080706F8F706F8F706F8F707080706F8F706F
8F706F8F707080706F8F706F8F706F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F403F5F3F3F5F403F5F3F305F403F5F3F3F
5F403F5F3F305F403F5F3F3F5F403F5F3F305F403F5F3F3F5F403F5F3F305F403F5F3F305F403F\
5F3F305F403F3030503F3F503030503F3F503030503F3F503030503F3F5F3F3F50
3F3F5F3030503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
8070608F7F6F807F6F807F6F8070608F7F6F807060807F6F807060807F6F807060
807F6F807060807F6F807060807F60807060807F6F8070608070607F7060807F607F7060807060\
7F70608070607F70608070607F70608070607F7060403030402F30403030402F30
403030402F3040303F402F30403030402F304F303F402F304F3030402F304F303F402F304F303F\
402F304F303F402F304F303F402F304F303F4F30304F303F4F2F304F303F4F3030
4F303F4F2F307F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F\
7F80807F8F8F7F80808F707080706F8F70708F70708F707080706F8F70708F706F
8F707080706F8F70708F706F8F70708F706F8F70708F706F8F70708F706F8F7F708F706F8F7070\
8F706F8F7F708F706F8F70708F706F8F7F708F706F8F70708F706F8F7F708F706F
8F70708F706F8F7F708F706F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F
5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F30\
5F403F5F3F305F403F30304F3030503030503F3F503030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F30305F\
807F6F807060807F6F807060807F6F807060807F6F807060807F60807060807F6F
7F70608070608070608070607F70608070607F70608070607F70608070607F70608070607F705F\
8070607F70608070607F705F7F70607F705F4030303F2F30402F30402F30403030
402F30402F30402F30403030402F30402F30402F30403030402F30403030402F304F3030402F30\
4F3030402F304F3030402F304F3030402F304F303F402F304F3030402F304F303F
402F307F808F7080807F808F7080807F808F7080807F808F7080807F80807080807F808F708080\
7F80807080807F808F80706F8F707080706F8F707080706F8F707080706F8F7070
80706F8F707080706F8F707080706F8F707080706F8F707080706F8F707080706F8F707080706F\
8F707080706F8F707080706F8F706F80706F8F707080706F8F706F8F706F8F7070
80706F8F706F8F706F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F403F60403F5F3F3F60403F5F403F60403F5F3F3F60403F5F403F5F403F5F3F3F
60403F5F403F5F403F5F3F3F60403F5F403F5F403F5F3F3F60403F5F403F5F403F5F3F3060403F\
5F403F5F403F5F3F3060403F3030503F3F503030503F3F5F3F3F503F3F5F303050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F3F5F8F7F6F807F6F8F7F6F807F6F807F6F8070608F7F6F807060807F6F807060
807F6F807060807F6F807060807F6F807060807F6F807060807F6F807060807F607F7060807F60\
7F70608070607F70608070607F7060403030402F3040303F402F30403030402F30
4F303F402F304F303F402F304F303F402F304F303F402F304F303F402F304F303F402F304F303F\
4F30304F303F402F304F303F4F30304F303F4F2F304F303F4F30304F303F7F8080
7F8F8F7F80807F808F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F7F80807F8F8F\
7F80807F8F8F7F80808F7F708F70708F70708F70708F7F708F70708F70708F706F
8F7F708F70708F70708F706F8F7F708F70708F70708F706F8F7F708F70708F7F708F706F8F7F70\
8F706F8F7F708F706F8F7F708F706F8F7F708F706F8F7F708F706F8F7F708F706F
8F7F708F706F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F3F305F403F5F3F30\
5F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F
5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F30\
5F403F5F3F305F403F5F3F303F3F503030503F3F503030503F3F503030503F3F50
3030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F3030503F3F5F3F3F5F\
3F3F5F807060807F6F807060807F6F807060807F6F807060807F6F807060807F6F
807060807F6F7F7060807060807060807F607F70608070607F70608070607F70608070607F7060\
8070607F70608070607F7060403030402F30402F30402F30403030402F30403030
402F30403030402F30403030402F304F3030402F30403030402F304F303F402F304F3030402F30\
4F303F402F304F3030402F304F303F402F304F303F4F2F307F808F7080807F808F
7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F7080807F8080708080\
7F808F7080807F808080706F8F707080706F8F707080706F8F707080706F8F7070
80706F8F707080706F8F70708F706F8F707080706F8F70708F706F8F707080706F8F70708F706F\
8F707080706F8F70708F706F8F707080706F8F70708F706F8F707080706F8F7070
8F706F8F7070000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005F403F60403F5F3F3F60403F5F403F60403F\
5F3F3F60403F5F403F60403F5F3F3F60403F5F403F60403F5F3F3F60403F5F403F
60403F5F3F3F60403F5F403F60403F5F3F3F60403F5F403F60403F5F3F3060403F5F403F60403F\
5F3F3060403F5F403F60403F5F3F303F3F503030503F3F503030503F3F5F303F50
3F3F5F3030503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F3F5F8F7F6F807F6F8F7F6F8070608F7F6F807060807F6F807060807F6F807060
807F6F807060807F6F807060807F6F807060807F6F807060807F607F7060807F607F7060807060\
7F70608070607F7060403030402F3040303F402F30403030402F304F303F402F30
403030402F304F303F402F304F303F402F304F303F402F304F303F402F304F303F402F304F303F\
402F304F303F4F30304F303F4F2F304F303F7080807F808F7080807F808F708080
7F808F7080807F8F8F7F80807F808F7080807F8F8F7F80807F808F7080807F8F8F7F80807F808F\
7080807F8F8F7F80808F70708F70708F7F708F70708F70708F706F8F7F708F7070
8F70708F706F8F7F708F70708F70708F706F8F7F708F70708F70708F706F8F7F708F706F8F7070\
8F706F8F7F708F706F8F7F708F706F8F7F708F706F8F7F708F706F8F7F708F706F
8F7F70000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F403F5F3F3F60403F5F3F305F403F5F3F3060403F5F3F30\
5F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F
5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F305F403F5F3F30\
5F403F5F3F305F403F5F3F305F403F5F3F303F3F503030503F3F503030503F3F50
3030503F3F503030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F30305F\
3F3F5F807060807F6F807060807F6F807060807F6F807060807F6F807060807F6F
7F7060807060807060807F607F70608070607F70608070607F70608070607F70608070607F7060\
8070607F70608070603F2F30402F30402F30403030402F30402F30402F30403030
402F30403030402F30403030402F30403030402F304F3030402F304F3030402F304F303F402F30\
4F3030402F304F303F402F307F808F7080807F808F7080807F808F7080807F808F
7080807F80807080807F808F7080807F80807080807F808F7080807F80807080807F8080708080\
7F80807080807F808080706F8F707080706F8F707080706F8F707080706F8F7070
80706F8F707080706F8F707080706F8F70708F706F8F707080706F8F70708F706F8F707080706F\
8F70708F706F8F707080706F8F70708F706F8F707080706F8F70708F706F8F7070
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000060403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F\
5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F
60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F\
5F403F60403F5F403F60403F5F403F60403F5F403F3F3F5F303F503F3F5F303050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F3F5F8F7F6F807F6F8F7F6F807F6F8F7F6F8070608F7F6F807F6F807F6F807060
807F6F807060807F6F807060807F6F807060807F6F807060807F6F807060807F607F7060807F6F\
7F7060807060402F3040303F402F3040303F402F304F303F402F304F303F402F30
4F303F402F304F303F402F304F303F4030304F303F402F304F303F4F30304F303F402F304F303F\
4F30304F303F7080807F808F7080807F808F7080807F808F7080807F808F708080
7F8F8F7080807F808F7080807F8F8F7080807F808F7080807F8F8F7F80807F808F7080807F8F8F\
7F80807F808F7080808F7F708F70708F7F708F70708F7F708F70708F7F708F7070
8F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F706F8F7F70\
8F70708F7F708F706F8F7F708F70708F7F708F706F8F7F708F70708F7F708F706F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F403F60403F5F3F3F60403F5F403F60403F5F3F3F60403F5F403F60403F5F3F3060403F5F403F\
60403F5F3F3060403F5F403F60403F5F3F3060403F5F3F3F60403F5F3F3060403F
5F3F3F60403F5F3F3060403F5F3F3F60403F5F3F3060403F5F3F3F60403F5F3F3060403F5F3F30\
60403F5F3F3060403F5F3F3060403F5F3F3060403F5F3F303F3F503030503F3F50
3030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F30305F3F3F5F3F3F5F\
3F3F5F807060807F6F807060807F6F807060807F6F807060807F6F807060807F6F
807060807F6F7F7060807F60807060807F6F7F70608070607F70608070607F70608070607F7060\
807060402F30402F30402F30403030402F30403030402F30403030402F30403030
402F304F3030402F304F3030402F304F303F402F304F3030402F304F303F402F304F303F402F30\
7F808F70808070808F7080807F808F7080807F808F7080807F808F7080807F808F
7080807F808F7080807F80807080807F808F7080807F80807080807F808F7080807F8080708080\
7F80807080807F80808F70708F707080706F8F70708F70708F707080706F8F7070
8F706F8F707080706F8F70708F706F8F707080706F8F70708F706F8F70708F706F8F70708F706F\
8F7F708F706F8F70708F706F8F7F708F706F8F70708F706F8F7F708F706F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F403F60403F5F403F
60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F\
5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F
60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F\
5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F3F3F5F303F50
3F3F5F3030503F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F3F5F8F7F6F807F6F8F7F6F8070608F7F6F807F6F807F6F807060807F6F807060
807F6F807060807F6F807060807F6F807060807F6F807060807F607F7060807F6F807060807060\
402F3040303F402F30403030402F304F303F402F3040303F402F304F303F402F30
4F303F402F304F303F402F304F303F402F304F303F4F30304F303F402F304F303F4F30307F808F\
7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F708080
7F808F7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F\
7080807F808F7080808F7F708F70708F7F708F70708F7F708F70708F7F708F7070
8F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F706F8F7F708F70708F7F70\
8F706F8F7F708F70708F7F708F706F8F7F708F70708F7F708F706F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F403F60403F5F3F3F60403F5F403F60403F
5F3F3F60403F5F403F60403F5F3F3F60403F5F403F60403F5F3F3060403F5F403F60403F5F3F30\
60403F5F403F60403F5F3F3060403F5F403F60403F5F3F3060403F5F403F60403F
5F3F3060403F5F403F60403F5F3F3060403F5F3F3F60403F5F3F3060403F5F3F3060403F5F3F30\
60403F5F3F3060403F5F3F3060403F5F3F3060403F5F3F3060403F5F3F303F3F50
3030503F3F503030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F30305F\
3F3F5F807060807F6F807060807F6F807060807F6F807060807F6F807060807F6F
807060807F60807060807F6F7F70608070607F7060807F607F70608070607F7060807060402F30\
402F30402F30403030402F30402F30402F30403030402F30403030402F304F3030
402F30403030402F304F303F402F304F3030402F304F303F402F304F303070808070808F707F80\
70808F70808070808F707F8070808070808070808F707F807080807080807F808F
707F807080807080807F80807080807080807080807F80807080807080807080807F8080708080\
7080807080807F808080706F8F70708F70708F707080706F8F70708F706F8F7070
80706F8F70708F706F8F707080706F8F70708F706F8F707080706F8F70708F706F8F70708F706F\
8F70708F706F8F7F708F706F8F70708F706F8F7F708F706F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000060404060403F60403F60403F60404060403F60403F60403F
60403F60403F60403F5F403F60403F60403F60403F5F403F60403F60403F60403F5F403F60403F\
60403F60403F5F403F60403F60403F60403F5F403F60403F60403F60403F5F403F
60403F60403F60403F5F403F60403F60403F60403F5F403F60403F60403F60403F5F403F60403F\
60403F60403F5F403F60403F60403F60403F5F403F60403F60403F60403F303050
3F3F5F3F3F503F3F5F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F60\
3F3F5F3F3F5F807F6F8F7F6F807F6F8F7F6F807F6F8F7F6F807F6F807F6F807060
807F6F807F60807F6F807060807F6F807060807F6F807060807F6F807060807F6F402F3040303F\
402F3040303F402F304F303F402F304F303F402F304F303F402F304F303F402F30
4F303F4030304F303F402F304F303F4F30304F303F4F2F307F808F7080807F808F7080807F808F\
7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F708080
7F808F7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F7080807F808F\
7080807F80807080808F7F708F70708F7F708F70708F7F708F70708F7F708F7070
8F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F70\
8F70708F7F708F70708F7F708F70708F7F708F7070907F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F
5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F\
60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F
5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F3F3060403F5F403F\
60403F5F3F3060403F5F403F60403F5F3F3060403F5F403F60403F5F3F3060403F
3030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F30305F3F3F5F3F3F5F\
3F3F5F3F3F5F807F6F807F6F807F6F807060807F6F807060807F6F807060807F6F
807060807F6F807060807F6F807060807F6F7F70608070607F7060807F60402F30403030402F30\
403030402F30403030402F30403030402F30403030402F304F303F402F304F3030
402F304F303F402F304F3030402F304F303F707F8070808F70808070808F707F8070808F708080\
70808F707F8070808070808070808F707F8070808070808070808F707F80708080
7080807F80807080807080807080807F80807080807080807080807F8080708080708080708080\
7F80807080807080808F70708F7F708F70708F70708F70708F7F708F706F8F7070
8F70708F7F708F706F8F70708F70708F7F708F706F8F70708F70708F7F708F706F8F70708F706F\
8F7F708F706F8F7F708F706F8F7F708F706F8F7F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005F6F8060403F\
60403F60404060403F60403F60403F60404060403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F\
60403F60403F60403F60403F5F403F60403F60403F60403F5F403F60403F60403F
60403F3030503F3F5F3F3F503F3F5F303F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F3F5F3F3F60807F6F8F7F6F807F6F8F7F6F807F6F807F6F807060807F6F807F60
807F6F807060807F6F807060807F6F807060807F6F807060807F6F402F3040303F402F3040303F\
402F304F303F402F304F303F402F304F303F402F304F303F402F304F303F402F30
4F303F402F304F303F4F303070808F7080807F808F70808070808F7080807F808F70808070808F\
7080807F808F70808070808F7080807F808F70808070808F7080807F808F708080
7F808F7080807F808F7080807F80807080807F808F7080807F80807080807F808F7080807F8080\
7080807F80807080808F7F708F70708F7F708F70708F7F708F70708F7F708F7070
8F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F70\
8F70708F7F708F70708F7F708F70708F7F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F607F\
5F6F805F607F5F6F805F403F60403F5F403F60403F5F403F60403F5F403F60403F
5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F\
60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403F60403F
5F403F60403F5F403F60403F5F403F60403F5F403F60403F5F403060403F5F403F60403F5F4030\
60403F5F403F60403F5F403060403F5F403F60403F5F3F3060403F5F403F60403F
5F3F3060403F3030503F3F5F3030503F3F5F3030503F3F5F3030503F3F5F3F3F5F3F3F5F30305F\
3F3F5F3F3F5F807F6F807060807F6F807060807F6F807060807F6F807060807F6F
807060807F6F807060807F6F7F7060807F607F7060807F607F7060402F30402F30403030402F30\
403030402F30403030402F30403030402F304F3030402F30403030402F304F303F
402F304F3030707F8070808F707F8070808F707F8070808F707F80708080707F8070808F707F80\
708080707F8070808F707F80708080707F80708080707F80708080707F80708080
707F80708080708080708080707F80708080708080708080707F80708080708080708080707F80\
7080807080807080808070708F70708F70708F7F7080706F8F70708F70708F7F70
8F706F8F70708F70708F7F708F706F8F70708F70708F7F708F706F8F70708F706F8F7F708F706F\
8F70708F706F8F7F708F706F8F70708F706F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F6F80\
5F6F805F6F805F6F7F606F805F6F805F6F805F6F80604F4060403F60403F60403F
604F4060403F60403F60403F60404060403F60403F60403F60404060403F60403F60403F604040\
60403F60403F60403F60404060403F60403F60403F60404060403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F3F3F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F60\
3F3F5F3F3F60807F6F8F7F6F807F6F8F7F6F807F6F8F7F6F807F6F807F6F807F60
807F6F807F6F807F6F807060807F6F807060807F6F8070604F303F402F3040303F402F304F303F\
402F304F303F402F304F303F4030304F303F402F304F303F4030304F303F402F30
70808F70808070808F7080807F808F70808070808F7080807F808F70808070808F7080807F808F\
70808070808F7080807F808F70808070808F7080807F808F70808070808F708080
7F808F7080807F808F7080807F808F7080807F80807080807F808F7080807F80807080807F808F\
7080807F80807080808F7F708F70708F7F708F70708F7F708F70708F7F708F7070
8F7F708F70708F7F708F7070907F708F70708F7F708F7070907F708F70708F7F708F7070907F70\
8F70708F7F708F7070907F708F7070000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5060\
5F6F805F607F5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F80
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F5F403F60403F60403F\
60403F5F403F60403F60403F60403F5F403F60403F60403F60403F5F403F60403F
60403F60403F5F403F60403F3030503F3F5F3030503F3F5F3F3F5F3F3F5F30305F3F3F5F3F3F5F\
3F3F5F3F3F5F807F6F807F6F807F6F807060807F6F807060807F6F807060807F6F
807060807F6F807060807F6F807060807F6F7F7060403030402F30403030402F30403030402F30\
40303F402F30403030402F304F303F402F304F3030402F304F303F402F3070808F
707F8070808F707F8070808F707F8070808F707F80708080707F8070808F707F80708080707F80\
70808F707F80708080707F80708080707F80708080707F80708080707F80708080
708080708080707F80708080708080708080707F80708080708080708080707F80708080708080\
708080707F807080808F70708F7F708F70708F7F708F70708F7F708F70708F7F70
8F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F706F8F7F708F7070\
8F7F708F706F8F7F708F7070000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000605F6F605060605F6F\
5F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F805F6F805F6F7F5F6F805F6F80
5F6F805F6F80606F805F6F8060403F60403F604F4060403F60403F60403F604F4060403F60403F\
60403F60404060403F60403F60403F60404060403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F60403F3F3F5F303F503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F3F5F3F3F60807F6F8F7F6F807F6F8F7F6F807F6F807F6F807F6F807F6F807F6F
807F6F807060807F6F807060807F6F80706040303F402F3040303F402F304F303F402F304F303F\
402F304F303F402F304F303F402F304F303F402F304F303F707F8070808F708080
70808F707F8070808F70808070808F707F8070808F70808070808F707F8070808F70808070808F\
707F8070808F70808070808F70808070808F70808070808070808070808F708080
70808070808070808F7080807080807080807F80807080807080807080807F8080708080708080\
7080807F80807080808F7F708F70708F7F708F70708F7F708F70708F7F708F7070
8F7F708F70708F7F708F70708F7F708F7070907F708F70708F7F708F7070907F708F70708F7F70\
8F7070907F708F70708F7F70000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F50606050605F5060605F6F5F5060\
605F6F5F607F5F6F8050607F5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F80
5F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F60403F60403F60403F60403F\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F3F3F5F3030503F3F5F3030503F3F5F303F5F3F3F5F30305F\
3F3F5F3F3F5F807F6F807060807F6F807F6F807F6F807060807F6F807060807F6F
807060807F6F807060807F6F7F7060402F30402F30403030402F30403030402F30403030402F30\
403030402F304F303F402F304F3030402F3070808F6F7F8070808F707F8070808F
6F7F80708080707F8070808F6F7F80708080707F8070808F6F7F80708080707F807080806F7F80\
708080707F80708080707F80708080707F80708080707F80708080707F80708080
707F80708080707F80708080707F80708080707F80708080707F80708080707F80708080707F80\
708080707F807080808F70708F70708F70708F7F708F70708F7F708F70708F7F70
8F70708F7F708F70708F7F708F70708F7F708F70708F7F708F706F8F7F708F70708F7F708F706F\
8F7F708F70708F7F70000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000605060605F6F5F5060605F6F605060605F6F605060605F6F\
605F6F605F6F5F6F7F5F6F805F6F805F6F805F6F7F5F6F805F6F805F6F805F6F7F
606F805F6F805F6F805F6F80606F805F6F80606F805F6F806070805F6F80606F805F6F80607080\
60403F604F4060403F6F4F4060403F60403F60403F6F4F4060403F60403F60403F
6F4F4060403F60403F60403F6F4F4060403F60403F60403F6F4F4060403F60403F60403F6F4F40\
60403F60403F60403F6F4F4060403F60403F60403F6F4F4060403F60403F60403F
6F4F3F60403F60403F60403F6F4F3F60403F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F3F3F60\
3F3F5F3F3F603F3F5F8F7F6F807F6F8F7F6F807F6F8F7F6F807F6F807F6F807F6F
807F6F807F6F807F6F8070604F303F402F3040303F402F304F303F402F304F303F402F304F303F\
4030304F303F402F304F303F707F8070808F707F8070808F70808070808F707F80
70808F70808070808F707F8070808F70808070808F707F8070808F70808070808F707F8070808F\
70808070808F70808070808F70808070808F70808070808F708080708080708080
70808F70808070808070808070808F7080807080807080807F80807080807080807080807F8080\
708080708080708080907F7F8F70708F7F708F7070907F7F8F70708F7F708F7070
907F7F8F70708F7F708F7070907F708F7F708F7F708F7070907F708F7F70907F708F7070907F70\
8F7F70907F70000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006050605F5060605F6F5F5060605F6F5F5060605F6F5F5060605F6F5F5060\
605F6F5F50605F6F805F607F5F6F805F607F5F6F805F607F5F6F805F607F5F6F80
5F6F7F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F80\
606F805F6F7F5F6F805F6F8060403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F3F3F5F303F5F3F3F5F30305F3F3F5F3F3F5F\
3F3F5F3F3F5F3F3F5F807F6F807F6F807060807F6F807F6F807F6F807060807F6F
807060807F6F807060403030402F3040303F402F30403030402F304F303F402F30403030402F30\
4F303F402F3070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F80708080
707F8070808F6F7F80708080707F8070808F6F7F80708080707F807080806F7F80708080707F80\
7080806F7F80708080707F80708080707F80708080707F80708080707F80708080
707F80708080707F80708080707F80708080707F80708080707F80708080707F80708080707F80\
708080707F807080808F70708F7F708F70708F7F708F70708F7F708F70708F7F70
8F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F7070\
8F7F70000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F5060605F6F5F5060605F6F5F5060605F6F605060605F6F5F5060605F6F605060605F6F\
605060605F6F605F6F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F
5F6F805F6F7F5F6F805F6F805F6F805F6F7F606F805F6F805F6F805F6F80606F805F6F80606F80\
5F6F806070805F6F80606F805F6F806070805F6F8060708060403F6F4F4060403F
60403F60403F6F4F4060403F60403F60403F6F4F4060403F60403F60403F6F4F4060403F60403F\
60403F6F4F4060403F60403F60403F6F4F4060403F60403F60403F6F4F3F60403F
60403F60403F6F4F3F60403F60403F60403F6F4F3F60403F3F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F3F5F3F3F603F3F5F8F7F6F807F6F8F7F6F807F6F807F6F807F6F8F7F6F807F6F
807F6F807060807F6F402F3040303F402F304F303F402F304F303F402F304F303F402F304F303F\
6F7F8070808F707F8070808F707F8070808F707F8070808F707F8070808F707F80
70808F707F8070808F707F8070808F707F8070808F707F8070808F707F8070808F707F80708080\
707F8070808F708080708080707F8070808F708080708080707F80708080708080
708080707F80708080708080708080707F80708080708080708080708080708080708080708080\
7080807080807080808F7F708F7070907F7F8F70708F7F708F7070907F7F8F7070
8F7F708F7070907F708F70708F7F708F7070907F708F7F708F7F708F7070907F708F7F70907F70\
8F7070000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005F5060605F6F5F50606050605F5060605F6F5F5060605F6F5F5060605F6F5F5060\
605F6F5F5060605F6F50607F5F6F7F5F607F5F6F8050607F5F6F805F607F5F6F80
5F607F5F6F805F6F7F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F\
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F606F805F6F7F5F6F805F6F80606F80
5F6F7F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F\
60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F60403F
60403F60403F60403F60403F60403F60403F60403F60403F60403F3F3F5F303F5F3F3F5F30305F\
3F3F5F3F3F5F3F3F5F807F6F807F6F807F6F807F6F807060807F6F807060807F6F
807060807F6F402F30403030402F30403030402F3040303F402F30403030402F3070808F6F7F80\
707F8F6F7F8070808F6F7F80707F806F7F8070808F6F7F807080806F7F8070808F
6F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80\
7080806F7F807080806F7F807080806F7F807080806F7F80708080707F80708080
6F7F80708080707F807080806F7F80708080707F807080806F7F7F708080707F807080806F7F7F\
708080707F807080808F70708F7F708F70708F7F708F70708F7F708F70708F7F70
8F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F7070\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005F5060605F6F605060605F6F5F5060605F6F60506F605F6F605060605F6F\
605F6F605F6F60506F605F6F5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F
5F6F805F6F805F6F805F6F7F5F6F805F6F805F6F805F6F7F606F805F6F805F6F805F6F80607080\
5F6F80606F805F6F806070805F6F806070805F6F806070805F6F806070805F6F80
6070805F6F806070805F6F8060708060403F6F4F4060403F6F4F4060403F6F4F4060403F6F4F40\
60403F6F4F4060403F6F4F4060403F6F4F3F60403F6F4F4060403F6F4F3F60403F
6F4F4060403F6F4F3F60403F6F4F4060403F6F4F3F60403F6F4F4060403F3F3F5F3F3F5F3F3F60\
3F3F5F3F3F603F3F5F8F7F6F807F6F8F7F6F807F6F8F7F6F807F6F807F6F807F6F
8F7F6F402F304F303F402F304F303F402F304F303F402F304F303F40303070808F6F7F8070808F\
707F8070808F6F7F8070808F707F8070808F6F7F8070808F707F8070808F707F80
70808F707F8070808F707F8070808F707F8070808F707F8070808F707F8070808F707F8070808F\
708080708080707F8070808F708080708080707F8070808F708080708080707F80
708080708080708080707F80708080708080708080707F80708080708080708080708080708080\
708080708080708080907F7F8F7F70907F7F8F7070907F7F8F7F70907F7F8F7070
907F7F8F7F70907F708F7070907F7F8F7F70907F708F7070907F7F8F7F70907F708F7F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005F5060605F6F5F5060605F6F5F5060605F6F5F5060605F6F605060\
605F6F5F5060605F6F6050605F6F8050607F5F6F805F607F5F6F8050607F5F6F80
5F607F5F6F805F607F5F6F805F6F7F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F\
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F80606F805F6F7F5F6F80
5F6F806070805F6F7F606F805F6F806070805F6F806070805F6F806F4F3F60403F60403F60403F\
6F4F3F60403F60403F60403F6F4F3F60403F60403F60403F6F4F3F60403F60403F
60403F6F4F3F60403F60403F60403F6F4F3F60403F60403F60403F6F4F3F30305F3F3F5F3F3F5F\
3F3F5F3F3F5F3F3F5F807F6F8F7F6F807F6F807F6F807F6F807F6F807060807F6F
402F3040303F402F30403030402F304F303F402F3040303F6F7F8070808F6F7F80707F8F6F7F80\
70808F6F7F80707F806F7F8070808F6F7F807080806F7F8070808F6F7F80708080
6F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80\
7080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80708080
6F7F807080806F7F80708080707F807080806F7F80708080707F807080806F7F7F708080707F80\
7080806F7F7F7080808F70708F7F708F70708F7F708F70708F7F708F70708F7F70
8F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F5060605F6F605060605F6F5F5060605F6F60506F605F6F\
605060605F6F605F6F605F6F60506F5F6F805F6F7F5F6F805F607F5F6F805F6F7F
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F805F6F80\
5F6F7F606F805F6F805F6F805F6F806070805F6F80606F805F6F806070805F6F80
6070805F6F806070805F6F806070805F6F806070805F6F806070805F6F806070805F6F80607080\
60403F6F4F4060403F6F4F3F60403F6F4F4060403F6F4F3F60403F6F4F4060403F
6F4F3F60403F6F4F4060403F6F4F3F60403F6F4F4060403F6F4F3F60403F6F4F403F3F5F3F3F5F\
3F3F5F3F3F603F3F5F8F7F6F807F6F8F7F6F807F6F807F6F807F6F8F7F6F402F30
40303F402F304F303F402F304F303F402F3070808F6F7F8070808F6F7F8070808F6F7F8070808F\
6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F80
70808F6F7F8070808F707F807080806F7F8070808F707F807080806F7F8070808F707F80708080\
6F7F80708080707F807080806F7F80708080707F807080806F7F80708080707F80
708080707F80708080707F80708080707F80708080708080708080707F80708080708080708080\
707F807080807080808F7F7F8F7070907F7F8F7F708F7F7F8F7070907F7F8F7F70
907F7F8F7070907F7F8F7F70907F708F7070907F7F8F7F70907F708F7070907F7F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F5060605F6F5F5060605F6F5F5060605F6F5F5060\
605F6F605060605F6F5F5060605F6F50607F5F6F8050607F5F6F7F50607F5F6F80
50607F5F6F7F5F607F5F6F8050607F5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F805F607F\
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80
5F6F7F5F6F805F6F7F606F805F6F7F5F6F805F6F806070805F6F7F606F805F6F806070805F6F80\
6070805F6F806070805F6F806F403F60403F6F4F3F60403F6F403F60403F6F4F3F
60403F6F403F60403F6F4F3F60403F60403F60403F6F4F3F60403F60403F60403F6F4F3F30305F\
3F3F5F3F3F5F3F3F5F807F6F807F6F807F6F807F6F807060807F6F402F3040303F
402F30403030402F3040303F6F7F806F7F8F6F7F80707F8F6F7F806F7F806F7F80707F8F6F7F80\
707F806F7F80707F8F6F7F80707F806F7F807080806F7F80707F806F7F80708080
6F7F80707F806F7F807080806F7F80707F806F7F807080806F7F80707F806F7F807080806F7F80\
707F806F7F807080806F7F807080806F7F807080806F7F807080806F7F80708080
6F7F7F7080806F7F807080806F7F7F7080806F7F807080806F7F7F7080806F7F7F7080806F7F7F\
7080806F7F7F7080808F70708F7F708F70708F7F708F70708F7F708F70708F7F70
8F70708F7F708F70708F7F708F70708F7F708F70708F7F708F70708F7F70000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000605060605F6F605F6F605F6F60506F605F6F\
605F6F605F6F605F6F605F6F605F6F605F6F5F607F5F6F805F6F7F5F6F805F607F
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F805F6F805F6F7F606F80\
5F6F805F6F805F6F7F6070805F6F80606F805F6F806070805F6F806070805F6F80
6070805F6F806070805F6F806070805F6F806070805F6F806070805F6F806070805F6F80607080\
606F806070805F6F806070806070806070805F6F8060708F6F403F6F4F4060403F
6F4F4060403F6F4F4060403F6F4F4060403F6F4F4060403F6F4F4060403F6F4F4060403F6F4F40\
3F3F5F3F3F603F3F5F3F4060807F6F8F7F6F807F6F8F7F6F402F304F303F402F30
4F303F40303070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F\
6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F80
70808F6F7F8070808F6F7F8070808F707F807080806F7F8070808F707F807080806F7F8070808F\
707F807080806F7F80708080707F807080806F7F80708080707F807080806F7F80
708080707F80708080707F80708080708080708080707F80708080708080708080707F80708080\
708080708080707F80907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70
907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000605F6F5F5060605F6F5F5060605F6F605060\
605F6F5F5060605F6F60506F605F6F605060605F6F50607F5F6F8050607F5F6F7F
5F607F5F6F8050607F5F6F805F607F5F6F8050607F5F6F805F6F7F5F6F805F607F5F6F805F6F7F\
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80
5F6F7F5F6F805F6F7F5F6F805F6F806070805F6F7F5F6F805F6F806070805F6F7F6070805F6F80\
6070805F6F806070805F6F806070805F6F806070805F6F806070805F6F80607080
5F6F806F4F3F60403F6F4F3F60403F6F4F3F60403F6F4F3F60403F6F4F3F60403F6F4F3F60403F\
6F4F3F3F3F5F3F3F5F3F3F5F8F7F6F807F6F807F6F807F6F40303F402F3040303F
6F7F80707F8F6F7F806F7F8F6F7F80707F8F6F7F806F7F806F7F80707F8F6F7F806F7F806F7F80\
707F8F6F7F80707F806F7F807080806F7F80707F806F7F807080806F7F80707F80
6F7F807080806F7F80707F806F7F807080806F7F80707F806F7F807080806F7F80707F806F7F80\
7080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80708080
6F7F807080806F7F7F7080806F7F807080806F7F7F7080806F7F807080806F7F7F7080806F7F7F\
7080806F7F7F7080808F70708F7F7F8F70708F7F708F7070907F7F8F70708F7F70
8F7070907F7F8F70708F7F708F7070907F708F70708F7F708F7F70000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000605F6F605060605F6F605F6F605F6F\
60506F605F6F605F6F605F6F605F6F605F6F605F6F5F6F8050607F5F6F805F607F
5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80\
5F6F7F5F6F805F6F7F5F6F805F6F7F606F805F6F805F6F805F6F7F6070805F6F80
606F805F6F806070805F6F806070805F6F806070805F6F806070805F6F806070805F6F80607080\
5F6F806070805F6F806070805F6F806070805F70806070805F6F80607080607080
6070805F708060708F60708060708060403F6F4F406F403F6F4F4060403F6F4F406F403F6F4F40\
60403F6F4F403F3F5F3F3F60807F6F8F7F6F807F6F40303F402F304F303F6F7F80
707F8F6F7F8070808F6F7F80707F8F6F7F8070808F6F7F80707F8F6F7F8070808F6F7F8070808F\
6F7F8070808F6F7F8070808F6F7F8070808F6F7F807080806F7F8070808F6F7F80
7080806F7F8070808F6F7F807080806F7F807080806F7F807080806F7F807080806F7F80708080\
6F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80
7080806F7F807080806F7F807080806F7F80708080707F807080806F7F80708080707F80708080\
6F7F7F708080707F80907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70
907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000605F6F5F5060605F6F5F5060\
605F6F60506F605F6F5F5060605F6F60506F605F6F60506F5F6F7F50607F5F6F7F
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F8050607F5F6F7F5F607F5F6F8050607F\
5F6F805F6F7F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80
5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F6070805F6F7F\
5F6F805F6F806070805F6F7F6070805F6F806070805F6F806070805F6F80607080
5F6F806070805F6F806070805F6F806070805F6F806070805F6F806F4F3F60403F6F4F3F60403F\
6F4F3F60403F6F4F3F303F5F807F6F807F6F40303F402F306F7F806F7F806F7F8F
6F7F806F7F806F7F806F7F8F6F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F80\
6F7F806F7F806F7F806F7F806F7F806F7F80707F806F7F806F7F806F7F80707F80
6F7F806F7F806F7F80707F806F7F806F7F806F7F80707F806F7F806F7F806F7F807080806F7F80\
6F7F806F7F807080806F7F7F707F806F7F807080806F7F7F707F806F7F80708080
6F7F7F707F806F7F7F7080806F7F7F707F806F7F7F7080806F7F7F7080806F7F7F7080806F7F7F\
7080806F7F7F7080808F70708F7F708F70708F7F7F8F70708F7F708F7070907F7F
8F70708F7F708F7070907F708F70708F7F708F7070000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000605F6F605F6F605F6F\
605F6F605F6F605F6F605F6F605F6F605F6F605F6F6F5F6F5F607F5F6F805F607F
5F6F805F6F7F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80\
5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F606F805F6F805F6F805F6F7F
6070805F6F80606F805F6F806070805F6F806070805F6F806070805F6F806070805F6F80607080\
5F6F806070805F6F806070805F6F806070805F6F806070806070806070805F6F80
60708060708060708060708060708F60708060708060708060708F60708060708F60708060708F\
6070806F4F406F403F3F4060807F6F4F303F6F7F8070808F6F7F80707F8F6F7F80
70808F6F7F80707F8F6F7F8070808F6F7F80707F8F6F7F8070808F6F7F8070808F6F7F8070808F\
6F7F8070808F6F7F8070808F6F7F8070808F6F7F8070808F6F7F807080806F7F80
70808F6F7F807080806F7F8070808F6F7F807080806F7F807080806F7F807080806F7F80708080\
6F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80
7080806F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80708080\
707F807080806F7F7F907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70
907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000605F6F60506F\
605F6F5F5060605F6F605F6F605F6F60506F605F6F605F6F605F6F50607F5F6F7F
50607F5F6F7F50607F5F6F7F50607F5F6F8050607F5F6F7F5F607F5F6F8050607F5F6F7F5F6F7F\
5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80
5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F6070805F6F7F5F6F805F6F80\
6070805F6F7F6070805F6F806070805F6F7F6070805F6F806070805F6F80607080
5F6F806070805F6F806070805F6F806070805F6F806070805F70806070805F6F80607080607080\
6070805F70806070806070807050406F7F806F7F8F6F7F806F7F8F6F7F806F7F80
6F7F806F7F8F6F7F806F7F806F7F806F7F8F6F7F806F7F806F7F806F7F806F7F806F7F806F7F80\
6F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F80
6F7F80707F806F7F806F7F806F7F80707F806F7F806F7F806F7F80707F806F7F806F7F806F7F80\
7080806F7F7F6F7F806F7F807080806F7F7F6F7F806F7F807080806F7F7F707F80
6F7F807080806F7F7F707F806F7F7F7080806F7F7F7080806F7F7F7080806F7F7F7080806F7F7F\
7080806F7F7F7080808F7F70907F7F8F70708F7F7F8F7F70907F7F8F70708F7F7F
8F7F70907F7F8F7070907F708F7F70907F7F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000605F6F\
605F6F605F6F605F6F605F6F605F6F605F6F605F6F605F6F605F6F5F6F805F607F
5F6F7F50607F5F6F805F607F5F6F805F607F5F6F805F6F7F5F6F805F607F5F6F805F6F7F5F6F80\
5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F
5F6F805F6F7F6070805F6F805F6F805F6F7F6070805F6F806070805F6F7F6070805F6F80607080\
5F6F806070805F6F806070805F6F806070805F6F806070805F6F806070805F6F80
6070805F6F806070806070806070805F6F8060708060708060708060708060708F607080607080\
60708060708F60708080807F3F406070504F7050407050406F7F80707F8F6F7F80
6F7F8F6F7F80707F8F6F7F806F7F8F6F7F80707F8F6F7F806F7F8F6F7F8070808F6F7F806F7F80\
6F7F8070808F6F7F806F7F806F7F8070808F6F7F80707F806F7F807080806F7F80
707F806F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80708080\
6F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80
7080806F7F807080806F7F807080806F7F7F7080806F7F807080806F7F7F7080806F7F80708080\
6F7F7F7080806F7F7F907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70
907F7F8F7F70907F7F8F7F70907F7F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
605F6F60506F605F6F5F506F605F6F605F6F605F6F60506F605F6F605F6F5F6F7F
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F\
5F6F7F5F607F5F6F8050607F5F6F7F5F6F7F5F6F805F607F5F6F805F6F7F5F6F80
5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F\
5F6F805F6F7F6070805F6F7F5F6F805F6F7F6070805F6F7F6070805F6F80607080
5F6F7F6070805F6F806070805F6F806070805F6F806070805F6F806070805F6F806070805F6F80\
6070803F2F3F40303F807F708080703F40607050406F5040705040705040705040
6F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F80\
6F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F80
6F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F7F6F7F806F7F806F7F806F7F7F\
6F7F806F7F806F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F80
6F7F7F6F7F806F7F7F707F806F7F7F6F7F806F7F7F7080806F7F7F6F7F806F7F7F7080806F7F7F\
6F7F806F7F7F7080808F70708F7F7F8F7F70907F7F8F70708F7F7F8F7F70907F7F
8F70708F7F708F7F70907F7F8F7070000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000605F6F605F6F605F6F605F6F605F6F605F6F6F5F70605F6F605F6F50607F
5F6F805F607F5F6F8050607F5F6F805F6F7F5F6F805F607F5F6F805F6F7F5F6F805F6F7F5F6F80\
5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F
606F805F6F7F5F6F805F6F7F6070805F6F805F6F805F6F7F6070805F6F806070805F6F80607080\
5F6F806070805F6F806070805F6F806070805F6F806070805F6F806070805F6F80
6070805F70806070805F6F806070806070806070805F708060708060708060708060708060708F\
3F2F3F40303F8080708F807F808070404F6F3F406F705F4F70504070504F705040
705F4F70504070504F705040707F8F6F7F806F7F8F6F7F8070808F6F7F806F7F8F6F7F8070808F\
6F7F806F7F806F7F8070808F6F7F806F7F806F7F8070808F6F7F806F7F806F7F80
7080806F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80708080\
6F7F807080806F7F807080806F7F807080806F7F807080806F7F807080806F7F80
7080806F7F807080806F7F807080806F7F807080806F7F7F7080806F7F807080806F7F7F708080\
6F7F807080806F7F7F907F7F8F7F7F907F7F8F7F70907F7F8F7F70907F7F8F7F70
907F7F8F7F70907F7F8F7F70000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000605F6F605F6F605F6F60506F605F6F605F6F605F6F605F6F605F6F
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F\
5F6F8050607F5F6F7F5F6F7F5F6F8050607F5F6F7F5F6F7F5F6F805F6F7F5F6F80
5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F\
5F6F805F6F7F5F6F805F6F7F6070805F6F7F5F70805F6F806070805F6F7F607080
5F6F806070805F6F7F6070805F6F806070805F6F806070805F6F806070805F6F803F303F3F2F3F\
40303F807F708080708080708080703F4060404F6F3F406F705040705040705040
7050407050407050407050407050407050406F50406F7F806F7F806F7F806F7F806F7F806F7F80\
6F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F80
6F7F806F7F806F7F806F7F806F7F806F7F806F7F7F6F7F806F7F806F7F806F7F7F6F7F806F7F80\
6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F80
6F7F7F6F7F806F7F7F6F7F806F7F7F7080806F7F7F6F7F806F7F7F7080806F7F7F6F7F806F7F7F\
7080806F7F7F6F7F808F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F
8F7F70907F7F8F7F70000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000605F6F605F6F605F6F605F70605F6F605F6F605F6F6F5F70605F6F
605F7050607F5F6F8050607F5F6F7F50607F5F6F805F607F5F6F7F50607F5F6F805F6F7F5F6F80\
5F607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F6070805F6F805F7080\
5F6F7F6070805F6F806070805F6F7F6070805F6F806070805F6F806070805F6F80
6070805F6F806070805F6F806070805F6F806070805F70806070805F6F8040303F3F2F3F40303F\
3F2F3F8F807F80807080807F808070404F6F40406F404F6F3F406F705F4F705040
70504F705040705F4F70504070504070504070504F70504070504070504070504F6F7F806F7F80\
6F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F80
6F7F806F7F806F80806F7F806F7F806F7F807080806F7F806F7F806F7F807080806F7F806F7F80\
6F7F807080806F7F806F7F806F7F807080806F7F806F7F806F7F7F7080806F7F80
7080806F7F7F7080806F7F807080806F7F7F7080806F7F7F7080806F7F7F7080806F7F7F708080\
6F7F7F7080806F7F7F907F7F8F7F70907F7F8F7F7F907F7F8F7F70907F7F8F7F70
907F7F8F7F70907F7F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000060506F605F6F605F6F605F6F60506F605F6F605F6F605F6F
605F6F5F607F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F\
5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F5F6F7F5F6F80
50607F5F6F7F5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F\
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F70805F6F7F5F6F805F6F7F607080
5F6F7F5F70805F6F7F6070805F6F7F6070805F6F806070803F2F303F303F3F2F3F40303F3F2F3F\
80807080807080807F807F708080703F4060404F6F3F4060404F6F705040705040
7050407050407050407050407050407050407050407050406F50407050407050407050406F5040\
7050406F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F80
6F7F7F6F7F806F7F806F7F806F7F7F6F7F806F7F806F7F806F7F7F6F7F806F7F7F6F7F806F7F7F\
6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F80
6F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F\
6F7F7F6F7F7F6F7F808F7070907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F
8F7F70907F7F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000605F6F605F6F605F6F6F5F70605F6F6F5F70605F6F
6F6070605F6F5F6F7F50607F5F6F805F607F5F6F7F50607F5F6F805F6F7F5F6F7F50607F5F6F80\
5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F6070805F6F7F5F6F805F6F7F607080\
5F6F806070805F6F7F6070805F6F806070805F6F806070805F6F806070805F6F80
6070805F6F806070805F6F806070805F70806070803F2F3F40303F3F2F3F40303F3F2F3F8F807F\
80807080807F8080708F807F808070404F6F3F406F404F6F40406F404F6F705040
7F5F4F705040705F4F7050407F5F4F70504070504F705040705F4F70504070504F705040705F4F\
70504070504070504070504F6F7F806F7F806F7F806F7F806F7F806F7F806F7F80
6F80806F7F806F7F806F7F806F80806F7F806F7F806F7F807080806F7F806F7F806F7F80708080\
6F7F806F7F806F7F807080806F7F806F7F806F7F807080806F7F806F80806F7F7F
7080806F7F806F80806F7F7F7080806F7F807080806F7F7F7080806F7F7F7080806F7F7F708080\
6F7F7F7080806F7F7F907F7F8F7F7F907F7F8F7F70907F7F907F7F907F7F8F7F70
907F7F503F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000605F6F605F6F605F6F605F6F605F6F605F6F
605F6F605F7050607F5F607F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F\
5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F5F607F5F6F8050607F5F6F7F
5F6F7F5F6F805F6F7F5F6F7F5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F\
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F6070805F6F7F5F6F80
5F6F7F6070805F6F7F5F70805F6F7F3F303F3F2F3F3F303F3F2F3F40303F3F2F3F80807F808070\
80807F807F7080807080807080807F807F70404F6F3F4060404F6F3F4060404F6F
70504070504F705040705040705040705040705040705040705040705040705040705040705040\
7050406F50407050407050407050406F50407050406F50406F7F806F7F7F6F7F80
6F7F806F7F806F7F7F6F7F806F7F806F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F\
6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F80
6F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F\
6F7F806F7F7F6F7F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F
503F2F50402F503F2F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000605F6F605F70605F6F6F5F70605F6F
6F5F70605F6F6F607050607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F\
50607F5F6F805F6F7F5F6F7F50607F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80\
5F6F7F6070805F6F7F5F6F805F6F7F6070805F6F806070805F6F7F6070805F6F80
6070805F6F7F6070805F6F803F303F3F2F3F40303F3F2F3F40303F3F2F3F40303F80807080807F\
80807080807F80807080807F80807080807F3F406F404F6F3F406F404F6F40406F
404F6F7050407F5F4F705040705F4F7050407F5F4F70504070504F705040705F4F705040705040\
705040705F4F70504070504070504070504F705040705040705040705040705040
6F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F806F7F80\
6F7F7F6F7F806F7F806F7F806F7F7F6F80806F7F806F7F806F7F7F6F80806F7F7F
6F7F806F7F7F6F80806F7F7F6F7F806F7F7F6F80806F7F7F6F7F806F7F7F7080806F7F7F6F7F80\
6F7F7F7080806F7F7F907F7F8F7F7F907F7F8F7F7F907F7F8F7F70907F7F907F7F
5F4030503F2F5F4030503F2F504030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000605F6F605F6F605F6F605F6F
605F6F605F6F605F6F5F607F50607050607F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F\
5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F
50607F5F6F7F5F6F7F5F6F7F50607F5F6F7F5F6F7F5F6F805F6F7F5F6F7F5F6F7F5F6F805F6F7F\
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80
5F6F7F5F6F803F2F3F3F303F3F2F303F303F3F2F3F3F303F3F2F3F3F303F80807080807F807F70\
80807080807080807F807F70808070807F70404F6F3F4060404F6F3F4060404F6F
3F4060404F6F70504070504F705040705040705040705040705040705040705040705040705040\
7050407050407050406F50407050407050407050406F50407050406F5040705040
6F50407050406F5040705040607F7F6F7F806F7F7F6F7F80607F7F6F7F806F7F7F6F7F80607F7F\
6F7F806F7F7F6F7F80607F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F80
6F7F7F6F7F806F7F7F6F7F806F7F7F6F7F7F6F7F7F6F7F806F7F7F6F7F7F6F7F7F6F7F806F7F7F\
6F7F7F6F7F7F6F7F7F8F7F70907F7F8F7F70907F7F8F7F70907F7F8F7F705F402F
503F2F50402F503F2F50402F503F2F503F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000605F6F6F5F70605F6F
6F6070605F6F6F6070605F6F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F80\
5F6F7F5F6F7F50607F5F6F805F6F7F5F6F7F5F6F7F5F6F805F6F7F5F6F805F6F7F
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F7080\
5F6F7F5F6F805F6F7F6070805F6F7F5F70805F6F7F6070805F6F806070805F6F7F
6070803F2F3F40303F3F2F3F40303F3F2F3F40303F3F2F3F40304080807F80807F8080708F807F\
80807080807F80807080807F80807080807F3F4060404F6F40406F404F6F3F406F
404F6F404F6F404F6F7050407F5F4F7050407F5F4F7050407F5F4F705040705F4F7050407F5F4F\
705040705F4F705040705F4F70504070504F705040705F4F705040705040705040
705F4F7050407050407050407050407050407050406F7F806F7F806F7F806F7F806F7F806F7F80\
6F7F806F7F806F7F7F6F7F806F7F806F7F806F7F7F6F80806F7F806F7F806F7F7F
6F80806F7F7F6F7F806F7F7F6F80806F7F7F6F7F806F7F7F6F80806F7F7F6F7F806F7F7F708080\
6F7F7F6F80806F7F7F90807F907F7F907F7F8F7F7F90807F907F7F5F403050402F
5F403050402F5F4030503F2F5F4030503F2F504030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000605F6F605F6F
605F6F6F5F70605F6F605F7050607F5F6F7F50607F5F607F50607F5F6F7F50607F5F6F7F50607F\
5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F
50607F5F6F7F50607F5F6F7F5F6F7F5F6F8050607F5F6F7F5F6F7F5F6F805F6F7F5F6F7F5F6F7F\
5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F3F303F
3F2F3F3F303F3F2F3F3F303F3F2F3F40303F3F2F3F3F303F3F2F3F80807F80807080807F808070\
80807F807F7080807080807080807F807F7040406F3F4060404F6F3F4060404F6F
3F406F404F6F3F406F404F6F7050407F5F4F705040705040705040705F4F705040705040705040\
7050407050407050407050407050407050407050407050407050406F5040705040
7050407050406F50407050406F50407050406F50407050406F50407050406F50406F7F806F7F7F\
6F7F80607F7F6F7F806F7F7F6F7F80607F7F6F7F806F7F7F6F7F806F7F7F6F7F80
6F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F7F6F7F7F6F7F806F7F7F6F7F7F6F7F7F\
6F7F806F7F7F6F7F7F8F7F7F907F7F8F7F70907F7F8F7F70907F7F503F2F5F402F
503F2F5F4030503F2F50402F503F2F504030503F2F50402F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000605F6F
6F5F70605F6F6F6070605F6F6F607050607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F\
50607F5F6F7F50607F5F6F7F50607F5F6F7F5F6F7F5F6F7F50607F5F6F805F6F7F
5F6F7F5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80\
5F6F7F5F6F805F6F7F5F6F805F6F7F5F70805F6F7F5F6F805F6F7F3F303F3F2F3F
3F303F3F2F3F40303F3F2F3F40303F3F2F3F40303F3F2F3F80807F80807080807F80807080807F\
80807080807F80807080807F80807080807F808070404F6F3F406F404F6F40406F
404F6F3F406F404F6F404F6F404F6F7050407F5F4F7050407F5F4F7050407F5F4F705040705F4F\
7050407F5F4F705040705F4F705040705F4F705040705040705040705F4F705040
705040705040705F4F7050407050407050407050407050407050406F5040705040705040705040\
6F50406F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F
6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F80\
6F7F7F6F80806F7F7F907F7F8F7F7F90807F907F7F907F7F5F402F5F40305F402F
5F4030503F2F5F403050402F5F4030503F2F5F4030503F2F504030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000605F70
605F6F605F70605F6F6F5F70605F6F605F7050607F5F607F50607050607F50607F5F6F7F506070\
5F607F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F5F6F7F5F6F7F506F7F\
5F6F7F5F6F7F5F6F805F6F7F5F6F7F5F6F7F5F6F803F2F303F2F3F3F2F3F3F303F
3F2F303F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F80807F807F7080807F80807080807F807F70\
808070807F7080807F7F7F70808070807F708080703F406040406F3F4060404F6F
3F4060404F6F3F406F404F6F3F406F7050407050407F5F4F705040705040705040705F4F705040\
705040705040705040705040705040705040705040705040705040705040705040
6F50407050407050407050406F50407050406F50407050406F50407050406F50407050406F5040\
7050406F50407050406F503F6F7F80607F7F6F7F80607F7F6F7F7F607F7F6F7F80
607F7F6F7F7F6F7F7F6F7F80607F7F6F7F7F6F7F7F6F7F7F607F7F6F7F7F6F7F7F6F7F7F607F7F\
6F7F7F6F7F7F6F7F7F8F7F70907F7F8F7F7F907F7F503F2F5F402F50402F5F4030
503F2F5F402F503F2F5F4030503F2F50402F503F2F50402F503F2F50402F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6F6070605F6F6F6070605F6F6F6070605F705F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F\
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F805F6F7F5F6F7F50607F
5F6F805F6F7F5F6F7F5F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F805F6F7F5F6F80\
5F6F7F5F6F805F6F7F5F70805F6F7F5F6F803F2F3F40303F3F2F3F3F303F3F2F3F
40303F3F2F3F40303F3F2F3F4030403F303F80807F80807F8F807F80807F80807F80807080807F\
80807080807F80807080807F80807080807F808070404F6F3F406F404F6F3F406F
404F6F404F6F404F6F40406F404F70404F6F7F5F4F7050407F5F4F7050407F5F4F7050407F5F4F\
7050407F5F4F7050407F5F4F705040705F4F7050407F5F4F705040705F4F705040
705F4F705040705F4F705040705F4F705040705040705040705F4F705040705040705040705F40\
7050407050407050407050407050407050406F50407050406F7F7F6F7F806F7F7F
6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F8080\
6F7F7F6F7F806F7F7F90807F907F7F90807F907F7F6040305F402F5F40305F402F
5F40305F402F5F403050402F5F403050402F5F4030503F2F5F403050402F5F4030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006F5F70605F6F605F70605F6F6F6070605F6F50607F50607F5F6F7F50607050607F50607F\
5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F506F7F5F6F7F50607F5F6F7F5F6F7F\
5F6F7F506F7F5F6F7F5F6F7F3F303F3F2F303F2F3F3F2F3F3F303F3F2F3F3F303F
3F2F3F3F303F3F2F3F3F303F3F2F3F40303F80807080807F80807080807F80807080807F808070\
80807F807F7080807080807080807F7F7F708080703F4060404F6F3F4060404F6F
3F406F404F6F3F406F404F6F3F406F404F6F3F406F7F5F4F7050407F5F4F705040705F4F705040\
7F5F4F705040705040705040705F4F705040705040705040705F40705040705040
7050407050407050407050407050407050407050407050407050407050406F5040705040705040\
7050406F50407050406F50407050406F50407050406F50407050406F503F705040
607F7F6F7F80607F7F6F7F7F607F7F6F7F80607F7F6F7F7F6F7F7F6F7F7F607F7F6F7F7F6F7F7F\
6F7F7F607F7F6F7F7F8F7F7F907F7F8F7F7F5F40305F402F5F40305F402F5F4030
5F402F5F4030503F2F5F4030503F2F5F4030503F2F5F402F503F2F5F4030503F2F50402F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006F6070605F6F6F6070605F6F6F607050607F5F6F7F50607F5F6F7F50607F5F6F7F\
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F506F7F
5F6F7F50607F5F6F7F5F6F7F5F6F7F506F7F5F6F805F6F7F5F6F7F5F6F7F5F6F805F6F7F5F6F80\
5F6F7F5F6F805F6F7F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F40303F3F2F3F
40303F3F2F3F40303F3F2F3F40303F80807F80807F80807F80807F80807080807F80807080807F\
80807080807F80807080807F80807080807F808070404F6F3F4060404F6F3F406F
404F6F3F406F404F6F404F6F404F6F40406F404F70404F6F7F5F4F7050407F5F4F7050407F5F4F\
7050407F5F4F7050407F5F4F7050407F5F4F705040705F4F7050407F5F4F705040
705F4F705040705F4F705040705F40705040705F4F705040705040705040705F4F705040705040\
705040705F407050407050407050407050407050407050406F5040705040705040
7050406F50407050406F50406F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F806F7F7F6F7F7F\
6F7F7F6F7F806F7F7F90807F907F7F604F305F402F5F40305F402F5F40305F402F
5F40305F402F5F40305F402F5F403050402F5F403050402F5F4030503F2F5F4030503F2F5F4030\
503F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006F5F70605F6F605F70605F6F6F607050607050607F50607050607F506070\
50607F50607F5F6F7F50607050607F50607F5F6F7F5060705F6F7F50607F5F6F7F
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F\
5F6F7F3F2F303F303F3F2F303F2F3F3F2F3F3F303F3F2F303F2F3F3F2F3F3F303F
3F2F3F3F303F3F2F3F40303F80807080807F80807080807F807F7080807F80807080807F7F7F70\
808070807F7080807F7F7F708080707F7F708080703F406040406F3F4060404F6F
3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F7F5F4F7050407F5F4F705040\
705F407050407F5F4F705040705040705040705F4F705040705040705040705F40
7050407050407050407050407050407050407050407050407050407050407050407050406F5040\
7050407050407050406F50407050406F50407050406F50407050406F5040705040
6F503F7050406F50407050406F503F7050406F503F6F7F7F60707F6F7F7F607F7F6F7F7F60707F\
6F7F7F607F7F6F7F7F8F7F7F6040305F402F6040305F402F5F40305F402F5F4030
5F402F5F40305F402F5F4030503F2F5F402F503F2F5F4030503F2F5F402F503F2F5F402F503F2F\
50402F503F2F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F6070605F706F6070605F705F6F7F50607F5F6F7F50607F5F6F7F\
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F
5F6F7F5F6F7F5F6F7F50607F5F6F805F6F7F5F6F7F5F6F7F5F6F805F6F7F5F6F7F5F6F7F5F6F80\
3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F40303F3F2F3F40303F3F2F3F
4030403F2F3F40303F3F2F3F8F8F7F80807F80807F80807F80807F80807F80807F80807080807F\
80807080807F80807080807F80807080807F80807080807F3F406F404F6F3F406F
404F6F404F6F404F6F3F406F404F70404F6F404F70404F6F404F70404F6F7F5F4F7050407F5F4F\
705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040
7F5F4F7050407F5F4F7050407F5F4F705040705F4F7050407F5F4F705040705F4F705040705F4F\
705040705F40705040705F4F705040705F40705040705F40705040705040705040
705F407050407050406F5040705F407050407050406F50407050406F50407050406F7F7F6F7F80\
6F7F7F6F7F7F6F7F7F90807F604030604F305F402F604F305F40306040305F402F
6040305F402F5F40305F402F5F40305F402F5F40305F402F5F40305F402F5F403050402F5F4030\
50402F5F4030503F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000006F6070605F6F6F6070605F6F50607F50607050607F506070\
5F6F7F50607050607F50607F5F6F7F506070506F7F50607F5F6F7F50607F5F6F7F
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F3F2F3F3F2F30\
3F303F3F2F303F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F
3F2F3F40303F3F2F3F80807F80807F80807F80807080807F80807080807F807F7080807F808070\
80807F7F7F70808070807F7080807F7F7F708080707F7F70404F6F3F4060404F6F
3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F4F6F404F703F406F7F5F4F705040\
7F5F4F7050407F5F4F7050407F5F4F705040705F4F7050407F5F4F705040705F40
705040705F4F705040705040705040705F40705040705040705040705F40705040705040705040\
7050407050407050407050407050406F50407050407050407050406F5040705040
6F50407050406F50407050406F50407050406F503F7050406F50407050406F503F7050406F5040\
705040607F7F6F7F7F5F402F604F305F402F6040305F402F6040305F402F5F4030
5F402F5F40305F402F5F40305F402F5F4030503F2F5F403050402F5F4030503F2F5F402F503F2F\
5F4030503F2F50402F503F2F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F6070605F706F60705060705F6F7F50607F5F6F7F\
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F
5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F5F6F7F5F6F7F506F7F3F303F3F2F3F3F303F\
3F2F3F3F303F3F2F3F3F303F3F2F3F40303F3F2F3F3F303F3F2F3F40303F3F2F3F
40303F3F2F3F80807F80807F80807F80807F80807F80807F80807F80807080807F80807080807F\
80807080807F80807080807F80807080807F8080708080703F4060404F6F3F406F
404F6F3F406F404F6F404F6F404F6F3F406F404F70404F6F404F70404F6F404F707F5F407F5F4F\
7050407F5F4F705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040
7F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040705F4F7050407F5F4F705040705F40\
705040705F4F705040705F40705040705F40705040705F40705040705F40705040
705040705040705F407050407050406F50407050407050407050406F50407050406F5040705040\
6F5040705040908F8F5F707060402F604F30604030604F305F402F604F305F402F
6040305F402F6040305F402F5F40305F402F5F40305F402F5F40305F402F5F40305F402F5F4030\
503F2F5F403050402F5F4030503F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000605F6F6F6070605F6F6F607050607050607F506070\
50607F50607050607F50607050607F506070506F7F50607050607F50607F5F6F7F
506070506F7F50607F5F6F7F5060705F6F7F50607F5F6F7F302F303F2F3F3F2F303F303F302F30\
3F2F3F3F2F3F3F303F3F2F303F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F
3F2F3F3F303F80807F80807F80807080807F80807080807F7F7F7080807F80807080807F7F7F70\
8080707F7F7080807F7F7F708080707F7F708080707F7F703F406F3F4060404F6F
3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F70705040\
7F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040705F407050407F5F4F
705040705F40705040705F4F705040705040705040705F40705040705040705040705F40705040\
7050407050407050407050407050407050407050406F5040705040705040705040
6F50407050406F50407050406F503F7050406F50407050406F503F7050406F50407050406F503F\
7050406F503F908F8F506F6F5F70705F6F6F5F70705F402F6040305F402F604030
5F402F5F40305F402F5F40305F402F5F40305F402F5F4030503F2F5F402F50402F5F4030503F2F\
5F402F503F2F5F4030503F2F50402F503F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000605F706F60706F5F706F60705060705F6F7F\
50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F5F6F7F50607F
5F6F7F50607F5F6F7F50607F5F6F7F506F7F5F6F7F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F\
3F2F3F3F303F3F2F3F40303F3F2F3F40303F3F2F3F4030403F2F3F40303F3F2F3F
40304080807F80807F80807F80807F80807F80807F80807F80807F80807F80807F80807080807F\
80807080807F80807080807F80807080807F80807080807F3F406F404F6F3F406F
404F6F40406F404F6F3F406F404F6F404F6F404F6F404F6F404F70404F6F404F70404F6F404F70\
7F5F4F7F5F4F705F407F5F4F7F5F4F7F5F4F7050407F5F4F705F407F5F4F705040
7F5F4F705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F\
705040705F4F7050407F5F4F705040705F40705040705F4F705040705F40705040
705F4F705040705F40705040705F40705040705F40705040705F407050407050406F5040705F40\
705040908F8F90808F5F70705F706F5F70705F6F6F5F70705F706F5F70705F402F
604F305F4030604F305F402F604F305F402F5F40305F402F5F40305F402F5F40305F402F5F4030\
5F402F5F403050402F5F403050402F5F4030503F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000605F706F6070605F6F50607F506070\
50607F50607050607F50607050607F50607050607F5060705F6F7F506070506F7F
50607F5F6F7F5060705F6F7F50607F3F2F3F302F303F2F3F3F2F303F303F3F2F303F2F3F3F2F3F\
3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F
80807F80807F80807080807F80807F80807F80807080807F80807080807F7F7F7080807F808070\
80807F7F7F708080707F7F7080807F7F7F708080707F7F70404F6F3F4060404F6F
3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F404F6F\
404F707050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F
7050407F5F4F7050407F5F4F7050407F5F4F705040705F407050407F5F4F705040705F40705040\
705F40705040705F40705040705F40705040705040705040705F40705040705040
705040705F406F50407050407050407050406F50407050406F50407050406F50407050406F5040\
705040908080908F8F506F6F5F7070506F6F5F706F506F6F5F7070506F6F5F706F
506F6F6040305F402F5F40305F402F5F40305F402F5F40305F402F5F40305F402F5F40305F402F\
5F4030503F2F5F402F503F2F5F4030503F2F50402F503F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000605F706F60706F5F7050607F\
5060705F6F7F50607F506F7F5060705F6F7F50607F5F6F7F50607F5F6F7F50607F
5F6F7F50607F5F6F7F50607F3F303F3F2F303F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F\
3F2F3F3F303F3F2F3F3F303F3F2F3F40303F3F2F3F40303F3F2F3F40304080807F
80807F80807F80807F80807F80807F80807F80807F80807F80807F80807080807F80807080807F\
80807080807F80807080807F7F807080807F8080708080707F7F70404F6F3F406F
404F6F3F406F404F6F3F406F404F6F3F406F404F6F404F6F404F6F404F6F404F70404F6F404F70\
404F6F404F707F5F4F7F5F4F705F407F5F4F7F5F407F5F4F7050407F5F4F705F40
7F5F4F7050407F5F4F705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F\
7050407F5F4F705040705F4F7050407F5F4F705040705F40705040705F4F705040
705F40705040705F40705040705F40705040705F40705040705F40705040705F40705040705040\
90808F908F8F908F8F5F7070506F6F5F70705F6F6F5F7070506F6F5F70705F6F6F
5F706F506F6F5F70705F6F6F6040305F402F604F305F402F5F40305F402F5F40305F402F5F4030\
5F402F5F40305F402F5F403050402F5F403050402F5F4030503F2F5F4030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000605F706F6070506070\
50607F50607050607F50607050607F50607050607F50607050607F506070506F7F
50607050607F302F303F2F3F302F303F2F3F3F2F303F2F3F302F303F2F3F3F2F3F3F303F3F2F30\
3F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F80807F
80807F80807F80807F80807F7F807080807F80807080807F7F7F7080807F7F807080807F7F7F70\
8080707F7F7080807F7F7F708080707F7F708080707F7F708080703F4060404F6F
3F4060404F6F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F703F406F\
404F70404F6F404F707050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F
7050407F5F4F7050407F5F4F7050407F5F407050407F5F4F705040705F407050407F5F4F705040\
705F40705040705F40705040705F40705040705F40705040705040705040705F40
7050407050407050407050406F50407050407050407050406F50407050406F50407050406F503F\
908F8F908080908F8F9080805F6F6F506F6F5F706F506F6F5F6F6F506F6F5F706F
506F6F5F6F6F506F6F5F706F506F6F5F6F6F506F6F5F40305F402F5F40305F402F5F40305F402F\
5F402F5F402F5F4030503F2F5F402F503F2F5F4030503F2F50402F503F2F50402F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006F5F706F6070\
50607F50607F5060705F6F7F50607F5F6F7F5060705F6F7F50607F5F6F7F50607F
5F6F7F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F40303F\
3F2F3F3F303F3F2F3F4030403F2F3F40303F3F2F3F4030403F303F80807F80807F
808F7F80807F80807F80807F80807F80807F80807F80807F80807F80807F80807F80807080807F\
80807080807F80807080807F80807080807F7F807080807F808070404F6F3F4060
404F6F3F406F404F6F3F406F404F6F404F6F404F6F404F6F404F70404F6F404F70404F6F404F70\
404F6F404F70404F6F404F707F5F4F7F5F4F7F5F407F5F4F7F5F4F7F5F4F7F5F40
7F5F4F7F5F4F7F5F4F705F407F5F4F7F5F407F5F4F7050407F5F4F705F407F5F4F7050407F5F4F\
705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040
7F5F4F705040705F40705040705F4F705040705F40705040705F4F705040705F40705040908F8F\
908F8F908F8F908F8F9F8F8F5F6F6F5F7070506F6F5F70705F6F6F5F706F506F6F
5F70705F6F6F5F706F506F6F5F706F506F6F5F706F506F6F5F706F506F6F6040305F402F604F30\
5F402F5F40305F402F5F40305F402F5F403050402F5F403050402F5F403050402F
5F4030000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000605F70\
50607F50607050607F50607050607F50607050607F50607050607F5060703F2F3F
302F303F2F3F302F303F2F3F3F2F303F303F302F303F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F\
3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F80807F80807F80807F
80807F80807F80807F80807F80807F80807F7F807080807F80807080807F7F7F7080807F7F8070\
80807F7F7F708080707F7F7080807F7F7F708080707F7F708080703F406040406F
3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F4F6F\
404F703F4F6F404F70404F6F404F707050407F5F4F7F5F407F5F4F7050407F5F4F
705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040\
7F5F4F705040705F407050407F5F4F705040705F40705040705F40705040705F40
705040705F40705040705F40705040705F40705040705040705040705F40705040908F8F90808F\
908F8F908080908F8F90808F5F706F506F6F5F6F6F506F6F5F706F506F6F5F6F6F
506F6F5F6F6F506F6F5F6F6F506F6F5F6F6F506F605F6F6F506F6F5F6F6F506F60506F6F506F60\
5F40305F402F5F40305F402F5F4030503F2F5F403050402F5F4030503F2F5F402F
503F2F5F4030000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
6F5F7050607F50607050607F5060705F6F7F50607050607F5060703F303F3F2F3F
3F303F3F2F303F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F\
3F2F3F40303F3F2F3F40303F3F2F3F4030403F2F3F40303F80807F808F7F80807F
80807F80807F80807F80807F80807F80807F80807F80807F80807F80807080807F80807080807F\
7F807080807F80807080807F7F807080807F7F80708080707F7F70404F6F3F4060
404F6F3F4060404F6F3F406F404F6F3F406F404F6F404F6F404F6F3F4F6F404F70404F6F404F70\
404F6F404F70404F6F404F70404F6F7F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F4F
7F5F4F7F5F407F5F4F7F5F407F5F4F705F407F5F4F7F5F407F5F4F7050407F5F4F705F407F5F4F\
7050407F5F4F705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040
7F5F4F7050407F5F4F705040705F40705040705F4F705040705F40705040705F40908F8F908F8F\
90808F908F8F908F8F908F8F506F6F5F706F506F6F5F706F506F6F5F706F506F6F
5F6F6F506F6F5F706F506F6F5F6F6F506F6F5F706F506F6F5F6F6F506F6F5F706F506F6F5F6F6F\
506F605F6F6F5F402F5F40305F402F5F40305F402F5F403050402F5F403050402F
5F403050402F5F4030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F2F605F7050607F50607050607F50607050607F302F303F2F3F302F303F2F3F
302F303F2F3F302F303F2F3F302F303F2F3F3F2F3F3F303F3F2F303F2F3F3F2F3F3F303F3F2F3F\
3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F80807F80807F80807F80807F
7F807F80807F80807F80807F7F807080807F7F807080807F7F7F7080807F7F807080807F7F7F70\
80807F7F7F7080807F7F7F708080707F7F708080707F7F708080707F7F70404F6F
3F40603F406F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F\
404F6F3F4F6F404F703F4F6F404F70404F6F7F5F4F7050407F5F4F7F5F407F5F4F
7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040\
7F5F407050407F5F4F705040705F407050407F5F4F705040705F40705040705F40
705040705F40705040705F40705040705F40705040705F4070504070504090808F908F8F8F8080\
908F8F90808F908F8F908080506F6F506F6F5F6F6F506F6F506F6F506F6F5F6F6F
506F60506F6F506F6F5F6F6F506F60506F6F506F60506F6F506F60506F6F506F60506F6F506F60\
506F6F506F60506F6F506F60506F6F5F402F5F4030503F2F5F402F50402F5F4030
503F2F5F402F503F2F5F402F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F30\
2F2F2F6F607F506070506F7F50607050607F302F303F303F3F2F3F3F303F3F2F3F
3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F403040\
3F2F3F40303F3F2F3F4030403F303F40304080807F808F8080807F80807F80807F
808F7F80807F80807F80807F80807F80807F80807F80807F80807F80807F80807F80807080807F\
80807080807F7F807080807F80807080807F7F807080807F7F80708080703F4060
404F6F3F406F404F6F3F406F404F6F404F6F404F6F3F406F404F70404F6F404F70404F6F404F70\
404F6F404F70404F6F404F70404F70404F70404F707F5F4F7F5F4F7F5F4F7F5F4F
7F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F4F7F5F4F705F407F5F4F\
7F5F407F5F4F705F407F5F4F705F407F5F4F705F407F5F4F705F407F5F4F705040
7F5F4F705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040908F8F908F8F908F8F\
908F8F908F8F908F8F908F8F908F8F5F6F6F506F6F5F706F506F6F5F6F6F506F6F
5F706F506F6F5F6F6F506F6F5F706F506F6F5F6F6F506F6F5F6F6F506F6F5F6F6F506F605F6F6F\
506F6F5F6F6F506F605F6F6F506F605F6F6F506F605F6F6F5F402F5F40305F402F
5F40305F402F5F403050402F5F4030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F2F\
2F2F2F2F2F2F6F60705060703F2F3F302F303F2F3F302F303F2F3F302F303F2F3F
3F2F303F303F302F303F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F\
3F303F3F2F3F3F303F3F2F3F40303F80807F80807F80807F80807F80807F80807F
80807F80807F7F807F80807F80807F80807F7F807080807F7F807080807F7F7F7080807F7F8070\
80807F7F7F708080707F7F7080807F7F7F708080707F7F708080707F7F703F406F
3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F4F6F\
404F703F406F404F70404F6F404F703F4F6F404F70404F707F5F4F7F5F407F5F4F
7F5F407F5F4F705F407F5F4F7F5F407F5F4F7050407F5F4F705F407F5F4F7050407F5F4F705F40\
7F5F4F7050407F5F4F7050407F5F4F7050407F5F4F7050407F5F4F705040705F40
7050407F5F4F705040705F40705040705F40705040705F40705040908F8F8F808F908F8F90808F\
908F8F90808F908F8F908F8F908F8F506F6F506F6F506F6F5F6F6F506F60506F6F
506F60506F6F506F60506F6F506F60506F6F506F60506F6F506F60506F6F506F60506F6F506F60\
506F6F506060506F6F506F60506F6F506060506F60506F60506F6F506060506F60
50402F5F4030503F2F5F402F503F2F5F4030503F2F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F2F2F2F2F30\
2F2F2F302F302F2F2F3F2F3F302F303F303F3F2F3F3F303F302F303F303F3F2F3F
3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F40303F3F2F3F3F303F\
3F2F3F4030403F2F3F40303F3F2F3F808F8080807F80807F80807F808F7F80807F
80807F80807F80807F80807F80807F80807F80807F80807F80807F7F807080807F80807080807F\
7F807080807F7F807080807F7F807080807F7F80708080707F7F7080807F3F4060
404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F3F406F404F70404F6F404F70\
404F6F404F70404F6F404F70404F6F404F70404F70404F70404F707F5F4F7F5F4F
7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F407F5F4F\
705F407F5F4F7F5F407F5F4F705F407F5F4F705F407F5F4F7050407F5F4F705F40
7F5F4F7050407F5F4F705F407F5F4F7050407F5F4F7050407F5F4F90808F908F8F908F8F908F8F\
908F8F908F8F908F8F908F8F908F8F5F6F6F506F6F5F6F6F506F6F5F6F6F506F6F
5F6F6F506F605F6F6F506F6F506F6F506F605F6F6F506F60506F6F506F605F6F6F506F60506F6F\
506F605F6F6F506F60506F6F506F60506F6F506F60506F6F506F60506F6F506F60
506F605060605F40305F402F5F403050402F5F403050402F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F2F2F2F2F2F2F2F\
2F2F2F2F2F2F405F6F302F303F2F3F302F303F2F3F302F303F2F3F302F303F2F3F
302F303F2F3F3F2F3F3F303F302F303F2F3F3F2F3F3F303F3F2F3F3F2F3F3F2F3F3F303F3F2F3F\
3F303F3F2F3F3F303F3F2F3F80807F80807F80807F7F807F80807F80807F80807F
7F807F80807F7F807F80807F7F807080807F7F807080807F7F7F7080807F7F807080807F7F7F70\
80807F7F7F7080807F7F7F708080707F7F708080707F7F707F80707F7F7040406F
3F40603F406F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F\
404F6F3F4F6F404F703F406F404F70404F6F404F703F4F6F404F70404F707F5F4F
7F5F407F5F4F7F5F407F5F4F705F407F5F4F7F5F407F5F4F7050407F5F4F705F407F5F4F705040\
7F5F4F705F407F5F4F7050407F5F4F7050407F5F4F7050407F5F407050407F5F4F
705040705F407050407F5F4F705040705F40705040705F408F808F908F8F8F808F908F8F8F808F\
908F8F90808F908F8F8F808F908F8F506F60506F6F506060506F6F506F60506F6F
506060506F6F506F60506F6F506060506F6F506F60506F6F506060506F60506060506F6F506060\
506F60506060506F60506060506F60506060506F60506060506F60506060506F60
4F6060506F60506060506F604F60605F402F503F2F5F4030503F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F2F2F2F302F2F2F303030\
2F2F2F405F6F40506F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F
3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F4030403F2F3F40303F3F2F3F403040\
3F303F4030403F2F3F808F8080807F808F8080807F808F8080807F80807F80807F
808F7F80807F80807F80807F80807F80807F80807F80807F80807F80807F80807F7F807080807F\
80807080807F7F807080807F7F807080807F7F807080807F7F80708080703F4060
404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F3F406F404F70404F6F404F70404F6F404F70\
404F6F404F70404F6F404F70404F70404F70404F70405070404F70405070404F70
80604F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F407F5F4F\
7F5F4F7F5F4F7F5F407F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F407F5F4F705F40
7F5F4F705F407F5F4F705F407F5F4F705F407F5F4F90808F908F90908F8F908F8F908F8F908F90\
908F8F908F8F908F8F908F8F908F8F5F6F6F506F6F5F6F6F506F6F506F6F506F60
5F6F6F506F6F506F6F506F605F6F6F506F60506F6F506F605F6F6F506F60506F6F506F60506F6F\
506F60506F6F506F60506F6F506F60506F6F506F60506F6F506F60506F60506060
506F6F506F60506F60506060506F60506F60506F605060605F403050402F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F2F2F2F2F2F2F2F2F2F2F\
405F6F405060405F6F302F303F2F3F302F303F2F3F302F303F303F302F303F2F3F
3F2F3F3F303F3F2F3F3F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F\
3F303F3F2F3F3F303F80807F80807F80807F80807F80807F80807F7F807F80807F
80807F80807F7F807F80807F7F807F80807F7F807080807F7F807080807F7F7F7080807F7F8070\
80807F7F7F708080707F7F7080807F7F7F707F80707F7F708080707F7F707F8070
3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F4F6F\
404F703F406F404F70404F6F404F703F4F6F404F70404F6F404F70404F6F404F70
7F5F4F7F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F40\
7F5F4F705F407F5F4F705F407F5F4F7050407F5F4F705F407F5F4F7050407F5F4F
705F407F5F4F7050407F5F4F7050407F5F4F705040908F8F8F808F908F8F8F808F908F8F90808F\
908F8F8F808F908F8F908F8F908F8F506060506F6F506F60506F6F506060506F6F
506F60506F6F506060506F6F506060506F6F506060506F60506060506F6F506060506F60506060\
506F60506060506F60506060506F604F6060506F60506060506F604F6060506F60
506060506F604F6060506F60506060506F604F605F506F604F6060506F604F605F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F2F2F2F2F302F302F2F2F405F6F\
405060405F6F40506F3F303F302F303F303F3F2F3F3F303F3F2F3F3F303F3F2F3F
3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F4030403F2F3F40303F\
3F2F3F40304080807F808F8080807F808F8080807F80807F80807F808F7F80807F
80807F80807F80807F80807F80807F7F807F80807F80807F80807F7F807080807F7F807080807F\
7F807080807F7F807080807F7F7F7080807F7F80708080707F7F7080807F7F7F70
404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F3F406F404F70404F6F404F70\
3F4F6F404F70404F6F404F70404F6F404F70404F70404F70404F70405070404F70
404F707F5F4F80604F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F7F5F4F\
7F5F407F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F40
7F5F4F705F407F5F4F705F407F5F4F705F40908F90908F8F908F8F8F8F8F908F90908F8F908F8F\
908F8F908F8F908F8F908F8F908F8F908F8F506F6F506F6F506F60506F6F506F60
506F6F506F60506F6F506F60506F6F506060506F6F506F60506F6F506060506F6F506F60506F60\
506060506F6F506060506F60506060506F60506060506F60506060506F60506060
506F60506060506F60506060506F60506060506F60506060506F60000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F2F2F202F2F2F2F2F2F2F40506F3F5060\
40506F3F506040506F302F303F2F3F302F303F2F3F302F303F2F3F3F2F3F3F303F
302F303F2F3F3F2F3F3F303F3F2F3F3F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F\
3F303F80807F8080807F807F80807F80807F80807F7F807F80807F7F807F80807F
7F807F80807F7F807F80807F7F807080807F7F807080807F7F7F7080807F7F807080807F7F7F70\
80807F7F7F7080807F7F7F707F80707F7F708080707F7F707F80707F7F707F8070
3F40603F406F3F4060404F6F3F40603F4F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F\
404F6F3F4F6F404F703F406F404F703F4F6F404F703F4F6F404F70404F70404F70
404F70404F707F5F4F7F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F40\
7F5F4F7F5F407F5F4F705F407F5F4F705F407F5F4F7050407F5F4F705F407F5F4F
7050407F5F4F705F407F5F4F7050407F5F408F808F908F8F8F808F908F8F8F808F908F8F8F808F\
908F8F8F8F8F908F8F8F808F908F8F908F8F506F6F506060506F6F506060506F6F
4F6060506F60506060506F6F4F6060506F60506060506F604F6060506F60506060506F604F6060\
506F604F6060506F604F6060506F604F6060506F604F60605060604F6060506F60
4F605F5060604F605F506F604F605F506060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F302F2F2F302F302F2F2F405F6F405060405F6F\
405060405F6F405060706F803F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F
3F303F3F2F3F3F303F3F2F3F4030403F2F3F3F303F3F2F3F4030403F2F3F4030403F2F3F403040\
80807F808F8080807F808F8080807F808F8080807F808F8080807F80807F80807F
808F7F80807F80807F80807F80807F80807F80807F7F807F80807F80807F80807F7F807080807F\
7F807080807F7F807080807F7F807080807F7F807080807F7F80708080707F7F70
404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F404F6F404F6F3F4F6F404F70\
404F6F404F70404F6F404F70404F6F404F70404F6F405070404F70404F70404F70
405070404F704050707F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F7F5F4F7F5F4F80604F\
7F5F4F7F5F4F7F5F4F7F604F7F5F4F7F5F4F7F5F407F5F4F7F5F4F7F5F4F7F5F40
7F5F4F7F5F4F7F5F4F7F5F407F5F4F908F8F908F908F8F8F908F90908F8F908F90908F8F908F90\
908F8F908F8F908F8F908F90908F8F908F8F506F60506F6F506F60506F6F506060
506F6F506F60506F6F506060506F6F506F60506F6F506060506F6F506F60506F60506060506F6F\
506060506F60506060506F60506060506F60506060506F60506060506F60506060
506F60506060506F60000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F2F202F2F2F2F2F2F2F4050603F50604050603F5060\
40506F3F50604050606F6F7F3F303F302F303F2F3F3F2F3F3F303F302F3F3F2F3F
3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F\
808F8080807F80807F80807F8080807F807F80807F80807F80807F7F807F80807F
7F807F80807F7F807F80807F7F807F80807F7F807080807F7F807080807F7F7F7080807F7F8070\
80807F7F7F707F80707F7F7080807F7F7F707F80707F7F707F80707F7F707F8070
3F4060404F6F3F40603F4F6F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F\
404F6F3F406F404F6F3F4F6F404F703F4F6F404F70404F6F404F70404F6F404F70
404F70404F70404F70404F707F5F4F7F5F4F7F5F407F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F40\
7F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F705F407F5F4F
7F5F407F5F4F705F407F5F4F705F40908F8F8F808F908F8F8F808F908F8F8F808F908F8F8F8F8F\
908F8F8F808F908F8F8F8F8F908F8F8F8F8F506F6F506060506F6F4F6060506F60
506060506F6F4F6060506F60506060506F604F6060506F60506060506F604F6060506F604F6060\
506F604F6060506F604F6060506F604F60605060604F6060506F604F605F506060
4F605F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F2F2F302F2F2F4050603F506040506F40506040506F\
3F5060405F6F405060706F803F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F
3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F4030403F2F3F40303F3F2F3F4030403F2F3F808F80\
80807F808F8080807F808F8080807F808F8080807F80807F80807F808F7F80807F
80807F7F807F80807F80807F80807F7F807F80807F7F807F80807F7F807080807F7F807080807F\
7F807080807F7F807080807F7F7F7080807F7F80708080707F7F7080807F7F7F70
7F80703F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F404F6F404F6F\
3F4F6F404F70404F6F404F70404F6F404F70404F70404F70404F70405070404F70
404F70404F70405070404F704050707F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F7F5F4F\
7F5F4F80604F7F5F4F7F5F4F7F5F4F7F604F7F5F4F7F5F4F7F5F407F5F4F7F5F4F
7F5F4F7F5F407F5F4F7F5F40908F908F808F908F908F8F8F908F908F8F8F908F90908F8F908F8F\
8F8F8F908F90908F8F908F8F908F8F908F90506060506F6F506060506F6F506060
506F6F506060506F6F506060506F60506060506F6F506060506F60506060506F60506060506F60\
4F6060506F60506060506F604F6060506F60506060506F604F6060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F202F2F2F2F2F2F2F4050603F50604050603F50604050603F5060\
4050603F50604050606F607F3F2F3F302F3F3F303F302F303F2F3F3F2F3F3F303F
3F2F3F3F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F8080807F807F\
80807F80807F8080807F807F80807F7F807F80807F7F807F80807F7F807F80807F
7F807F80807F7F807F80807F7F807080807F7F807080807F7F7F7080807F7F807080807F7F7F70\
7F807F7F7F7080807F7F7F707F80707F7F707F80707F7F707F80707F7F707F8070
7F7F703F406F3F4060404F6F3F40603F4F6F3F4060404F6F3F4060404F6F3F406F404F6F3F406F\
404F6F3F406F404F703F406F404F703F4F6F404F703F4F6F404F70404F6F404F70
3F4F6F404F70404F70404F70404F70404F707F5F4F7F5F4F7F5F407F5F4F7F5F4F7F5F4F7F5F40\
7F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F
705F407F5F4F7F5F40908F8F8F808F908F8F8F808F908F8F8F808F908F8F8F808F908F8F8F808F\
908F8F8F8F8F908F8F8F808F908F8F8F8F8F506F604F6060506F604F6060506F60
4F60605060604F6060506F604F60605060604F6060506F604F60605060604F60605060604F605F\
5060604F60605060604F605F5060604F605F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F2F302F302F2F2F4050603F506040506F4050604050603F506040506F\
40506040506F405060706F803F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F
3F30403F2F3F3F303F3F2F3F4030403F2F3F4030403F2F3F4030403F303F40304080807F808F80\
80807F808F8080807F808F8080807F808F8080807F808F8080807F80807F80807F
808F7F80807F80807F7F807F80807F80807F80807F7F807F80807F7F807F80807F7F807080807F\
7F807080807F7F807080807F7F807080807F7F807080807F7F80708080707F7F70
80807F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F3F4F6F404F70\
404F6F404F70404F6F404F70404F6F404F70404F6F404F70404F70404F70404F70
405070404F70405070404F704F507F404F7040507F7F5F4F80604F7F5F4F80604F7F5F4F80604F\
7F5F4F80604F7F5F4F80604F7F5F4F7F604F7F5F4F80604F7F5F4F7F5F4F7F5F4F
7F604F7F5F4F7F5F4F8F8F8F908F908F8F8F908F908F8F8F908F90908F8F908F908F8F8F908F90\
908F8F908F90908F8F908F90908F8F908F8F908F8F506F6F506060506F6F506060
506F6F506060506F60506060506F6F506060506F604F6060506F60506060506F604F6060506F60\
506060506F604F6060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F202F2F2F2F2F2F2F4050603F4F604050603F50604050603F50604050603F5060\
4050603F50604050606F6F7F3F303F302F3F3F2F3F3F2F3F3F303F3F2F3F3F303F
3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F40304080807F80808080807F\
8080807F807F80807F80807F8080807F807F80807F7F807F80807F7F807F80807F
7F807F80807F7F807F80807F7F807F80807F7F807080807F7F807080807F7F7F7080807F7F8070\
80807F7F7F707F807F7F7F707F807F7F7F707F80707F7F707F80707F7F707F8070
7F7F70404F6F3F40603F406F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F\
404F6F3F406F404F6F3F4F6F404F703F4F6F404F70404F6F404F703F4F6F404F70
404F70404F70404F70404F70404F70405070404F70805F4F7F5F4F80604F7F5F4F7F5F4F7F5F4F\
7F604F7F5F407F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F407F5F4F7F5F407F5F4F
7F5F407F5F4F8F808F908F8F8F808F908F908F808F908F8F8F8F8F908F8F8F808F908F8F8F8F8F\
908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F4F6060506F604F6060506060
4F6060506F604F60605060604F6060506F604F60605060604F60605060604F605F5060604F6060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F302F2F2F4050603F50604050603F50604050603F50604050603F5060405060\
3F50604050603F50604050603F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F
3F303F3F2F3F4030403F2F3F3F303F3F2F3F4030403F2F3F40304080807F808F8080807F808F80\
80807F808F8080807F80808080807F808F8080807F80807F7F807F80807F80807F
80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807080807F7F807080807F\
7F807080807F7F807080807F7F7F7080807F7F80707F80707F7F7080807F7F7F70
7F80703F4060404F6F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F\
3F4F6F404F70404F6F404F703F4F6F404F70404F6F404F70404F6F404F70404F70
404F70404F70405070404F70405070404F7040507F404F7080604F7F5F4F80604F7F5F4F80604F\
7F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F7F604F7F5F4F80604F7F5F4F
7F5F4F7F5F40908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F90\
8F8F8F908F90908F8F908F8F8F8F8F908F90908F8F506F604F6060506F60506060
506F604F6060506F604F6060506F604F6060506F604F60605060604F6060506F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F2F2F2F4050603F4F5F3F50603F4F5F4050603F4F5F4050603F4F604050603F4F5F\
4050603F50604050603F4F603F2F3F3F2F3F3F303F302F3F3F2F3F3F2F3F3F303F
3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F3F2F3F80808080807F8080807F807F\
80807F7F807F8080807F807F80807F7F807F80807F7F807F80807F7F807F80807F
7F807F80807F7F807F80807F7F7F7080807F7F807080807F7F7F707F807F7F7F7080807F7F7F70\
7F807F7F7F707F807F7F7F707F80707F7F707F80707F7F707F80707F7F707F8070
7F7F703F40603F40603F4F6F3F40603F406F3F4060404F6F3F4060404F6F3F406F404F6F3F406F\
404F6F3F406F404F6F3F406F404F6F3F4F6F404F703F4F6F404F703F4F6F404F70
3F4F6F404F70404F70404F70404F70404F70404F70405070404F70805F4F7F5F4F80604F7F5F40\
7F5F4F7F5F4F7F604F7F5F407F5F4F7F5F4F7F5F4F7F5F407F5F4F7F5F407F5F4F
7F5F408F8F8F8F808F908F908F808F8F8F8F8F808F908F8F8F808F908F8F8F8F8F908F8F8F808F\
908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F4F60605060604F6060506060
4F60605060604F60605060604F605F5060604F605F506060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F302F2F2F4050603F4F604050603F50604050603F50604050603F50604050603F5060405060\
3F50604050603F50604050603F2F3F3F303F3F2F3F3F30403F2F3F3F303F3F2F3F
4030403F2F3F4030403F2F3F4030403F303F4030403F2F3F808F80808080808F8080807F808F80\
80807F808F8080807F808F8080807F808F8080807F808F8080807F80807F7F807F
808F7F80807F80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807080807F\
7F807080807F7F807080807F7F807080807F7F7F7080807F7F80707F80707F7F70
80807F7F7F70404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F3F406F404F70\
404F6F404F703F4F6F404F70404F6F404F70404F6F404F70404F70404F70404F70
405070404F70405070404F7040507F404F7040507F404F704F507F404F7080604F7F5F4F80604F\
7F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F
80604F8F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F90\
908F8F908F908F8F8F909090908F8F908F90908F8F506F604F6060506F604F6060
506F604F60605060604F6060506F60000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F2F
2F2F2F4050603F4F5F3F50603F4F5F4050603F4F5F3F50603F4F5F4050603F4F5F4050603F4F5F\
4050603F4F5F4050603F50603F303F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F
3F2F3F3F303F3F2F3F3F303F3F2F3F3F30403F2F3F80808080807F808F807F807F80808080807F\
8080807F807F80807F7F807F8080807F807F80807F7F807F80807F7F807F80807F
7F807F80807F7F807F80807F7F807F80807F7F807080807F7F807080807F7F7F707F807F7F8070\
7F807F7F7F707F807F7F7F707F807F7F7F707F80707F7F707F80707F7F707F8070
7F7F707F80703F40603F406F3F4060404F6F3F40603F4F6F3F406F404F6F3F406F404F6F3F406F\
404F6F3F406F404F6F3F4F6F404F703F406F404F703F4F6F404F703F4F6F404F70
404F70404F70404F70404F70404F70405070404F70404F70404F7040507F404F7080604F7F5F4F\
80604F7F5F4F80604F7F5F4F80604F7F5F4F7F5F4F7F5F4F80604F7F5F407F5F4F
8F808F908F908F808F8F8F908F808F908F908F808F908F8F8F8F8F908F908F8F8F908F8F8F8F8F\
908F908F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F5060604F605F506060
4F6060506060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F302F2F2F
4050603F4F5F4050603F4F5F4050603F4F5F4050603F50604050603F4F5F4050603F5060405060\
3F505F4050603F50604050603F2F3F3F303F3F2F3F3F303F3F2F3F3F30403F2F3F
3F303F3F2F3F4030403F2F3F4030403F2F3F808F80808080808F8080807F808F8080807F808F80\
80807F808F8080807F8080807F807F808F8080807F80807F7F807F8080807F807F
80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807080807F7F807080807F\
7F807080807F7F80707F807F7F7F7080807F7F80707F80707F7F707F807F7F7F70
7F80707F7F70404F6F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F\
3F406F404F70404F6F404F703F4F6F404F70404F6F404F70404F6F404F70404F70
404F70404F70405070404F70405070404F7040507F404F7040507F404F704F507F404F7080604F\
7F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F80604F8F8F8F
908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F90\
8F8F8F908F908F8F8F908F908F8F8F909090908F8F908F904F6060506060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F2F3F5060
3F4F5F3F4F5F3F4F5F3F50603F4F5F3F50603F4F5F3F50603F4F5F3F50603F4F5F4050603F4F5F\
3F50603F4F5F4050603F4F5F3F2F3F3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F
3F2F3F3F303F3F2F3F3F303F3F2F3F3F303F80807F808F807F807F8080807F807F8080807F807F\
80807F7F807F8080807F807F80807F7F807F80807F7F807F80807F7F807F80807F
7F807F80807F7F807F80807F7F7F707F807F7F807080807F7F7F707F807F7F7F707F807F7F7F70\
7F807F7F7F707F807F7F7F707F80707F7F707F80707F7F707F80707F7F707F8070
7F7F707F80703F40603F406F3F40603F406F3F4060404F6F3F40603F4F6F3F406F404F6F3F406F\
404F6F3F406F404F6F3F406F404F6F3F4F6F404F703F406F404F703F4F6F404F70
3F4F6F404F70404F70404F703F4F70404F70404F70405070404F70404F70404F7040507F404F70\
80604F7F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F407F5F4F7F5F4F8F8F90
8F808F8F8F908F808F8F8F908F808F8F8F8F8F8F8F908F908F808F8F8F8F8F8F8F908F908F8F8F\
908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F2F3F50603F4F5F
4050603F4F5F4050603F4F5F4050603F4F5F4050603F4F5F4050603F505F4050603F4F5F405060\
3F50604050603F505F4050603F2F3F3F303F3F2F3F4030403F2F3F3F30403F2F3F
4030403F303F4030403F2F3F403040808080808F80808080808F80808080808F8080807F808F80\
80807F808F8080807F808F8080807F8080807F807F808F8080807F80807F7F807F
80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807080807F\
7F807080807F7F807080807F7F80707F807F7F7F7080807F7F80707F80707F7F70
7F807F7F7F70404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F3F406F404F70\
3F4F6F404F6F3F4F6F404F70404F6F404F70404F6F404F70404F70404F70404F70
405070404F70405070404F70405070404F70405070404F704F507F404F7040507F404F704F507F\
40507F80604F7F5F4F80605080604F80604F7F5F4F80604F7F5F4F908F908F8F8F
908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F90\
8F8F8F908F908F8F8F909090908F8F908F908F8F8F909090000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F3F4F5F3F4F5F3F4F5F
3F4F5F3F505F3F4F5F3F4F5F3F4F5F3F50603F4F5F3F505F3F4F5F3F50603F4F5F3F50603F4F5F\
4050603F4F5F3F50603F4F5F6F6F803F2F3F3F303F3F2F3F3F303F3F2F3F3F303F
3F2F3F3F30403F2F3F3F303F808080808F807F807F80808080807F808F807F807F8080807F807F\
8080807F807F80807F7F807F8080807F807F80807F7F807F80807F7F807F80807F
7F807F80807F7F807F80807F7F807F80807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70\
7F807F7F7F707F807F7F7F707F807F7F7F707F80707F7F707F80707F7F707F8070
7F7F707F80707F7F703F406F3F4060404F6F3F40603F4F6F3F4060404F6F3F4060404F6F3F406F\
404F6F3F406F404F6F3F4F6F404F703F406F404F703F4F6F404F703F4F6F404F70
404F6F404F703F4F6F404F70404F70404F70404F70404F70404F7040507F404F7040507F404F70\
40507F7F5F4F80604F7F5F4F80604F7F5F4F80604F7F5F4F80604F8F808F8F8F90
8F808F8F8F908F808F8F8F908F8F8F908F908F808F8F8F908F8F8F908F908F8F8F908F908F8F8F\
908F908F8F8F908F8F8F8F8F908F908F8F8F908F8F8F8F8F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F2F3F4F5F3F4F5F3F50603F4F5F
3F505F3F4F5F4050603F4F5F3F50603F4F5F4050603F4F5F3F50603F4F5F4050603F4F5F405060\
3F4F5F4050603F4F5F4050606F6F7F3F30403F2F3F3F303F3F2F3F4030403F2F3F
4030403F2F3F4030403F303F808F80808080808F80808080808F8080807F808F8080807F808F80\
7F807F808F8080807F8080807F807F808F807F807F80807F7F807F8080807F807F
80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807080807F7F80707F807F\
7F807080807F7F80707F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70
7F80707F7F707F80703F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F3F4F6F404F6F\
3F406F404F703F4F6F404F703F4F6F404F70404F6F404F70404F6F404F70404F70
404F70404F70405070404F70404F70404F7040507F404F70405070404F7040507F404F7040507F\
404F704F507F80604F80604F7F5F4F80604F805F4F80604F8F808F908F908F8F8F
8F8F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F90\
8F8F8F908F908F8F8F908F908F8F8F9090908F8F8F908F90000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F4F5F3F4F5F3F4F5F3F4F5F
3F4F5F3F4F5F3F4F5F3F4F5F3F4F5F3F4F5F3F4F5F3F505F3F4F5F3F4F5F3F4F5F3F505F3F4F5F\
3F505F3F4F5F3F50603F4F5F6F6F803F2F3F3F303F3F2F3F3F303F3F2F3F3F303F
3F2F3F3F303F3F2F3F808F807F807F8080807F807F808F807F807F8080807F807F8080807F807F\
80807F7F807F8080807F807F80807F7F807F80807F7F807F80807F7F807F80807F
7F807F7F807F7F807F80807F7F7F707F807F7F80707F807F7F7F707F807F7F7F707F807F7F7F70\
7F807F7F7F707F807F7F7F707F80707F7F707F80707F7F707F80707F7F707F8070
707F707F80707F7F703F406F3F40603F406F3F40603F4F6F3F40603F4F6F3F406F404F6F3F406F\
404F6F3F406F404F6F3F406F404F6F3F4F6F404F703F406F404F703F4F6F404F70
3F4F6F404F70404F6F404F703F4F6F404F70404F70404F70404F70404F70404F7040507F404F70\
40507F404F7040507F7F5F4F80604F7F5F4F80604F7F5F4F8F8F908F808F8F8F90
8F808F8F8F908F808F8F8F908F808F8F8F908F8F8F8F8F908F808F8F8F908F8F8F908F908F8F8F\
8F8F8F8F8F8F908F908F8F8F908F8F8F8F8F908F908F8F8F908F8F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F505F3F4F5F4050603F4F5F3F50603F4F5F4050603F4F5F3F50603F4F5F405060\
3F4F5F4050603F4F5F4050606F6F803F30403F2F3F4030403F303F4030403F2F3F
4030403F303F808F80808080808F80808080808F80808080808F80808080808F8080807F808F80\
80807F808F807F807F808F8080807F8080807F807F808F807F807F80807F7F807F
80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807080807F\
7F80707F807F7F807080807F7F80707F807F7F7F707F807F7F80707F80707F7F70
7F807F7F7F707F80703F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F\
3F4F6F404F6F3F4F6F404F70404F6F404F703F4F6F404F70404F6F404F70404F6F
405070404F70404F70404F70405070404F70405070404F7040507F404F7040507F404F704F507F\
404F704F507F404F704F507F80604F80604F80604F908F908F8F908F8F908F8F8F
908F908F8F90908F908F8F8F908F908F8F90908F908F8F8F908F908F8F8F908F908F8F8F909090\
8F8F8F908F908F8F8F9090908F8F8F9090908F8F8F909090908F8F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F4F5F3F505F3F4F5F3F4F5F3F4F5F\
3F505F3F4F5F3F505F3F4F5F6F6F803F2F3F3F303F3F2F3F3F30403F2F3F3F303F
3F2F3F808F807F807F808F80808080808F807F807F8080807F807F808F807F807F8080807F807F\
8080807F807F80807F7F807F8080807F807F80807F7F807F80807F7F807F80807F
7F807F80807F7F807F7F807F7F807F80807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70\
7F807F7F7F707F807F7F7F707F807F7F7F707F80707F7F707F80707F7F707F8070
7F7F707F8070707F703F406F3F40603F4F6F3F40603F4F6F3F4060404F6F3F4060404F6F3F406F\
404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F4F6F404F703F4F6F404F70
3F4F6F404F703F4F6F404F70404F70404F70404F70404F70404F70405070404F70405070404F70\
40507F404F7040507F404F7040507F7F5F4F8F8F908F808F8F8F908F808F8F8F90
8F8F8F8F8F908F808F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F908F908F8F8F8F8F908F8F8F\
908F908F8F8F8F8F908F8F8F908F908F8F8F908F8F8F8F8F908F90000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004050603F4F5F3F505F6F6F7F4030403F2F3F3F30403F2F3F4030403F303F
403040808080808F80808080808F80808080808F80808080808F807F807F808F8080807F808F80\
7F807F808F807F807F8080807F807F808F807F807F80807F7F807F8080807F807F
80807F7F807F80807F7F807F80807F7F807F80807F7F807F7F807F7F807080807F7F80707F807F\
7F80707F807F7F80707F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70
7F80707F7F707F80703F40603F4F6F3F4060404F6F3F406F404F6F3F406F404F6F3F406F404F6F\
3F406F404F6F3F4F6F404F6F3F4F6F404F70404F6F404F703F4F6F404F70404F70
404F70404F70405070404F70404F70404F70405070404F70405070404F7040507F404F7040507F\
404F704F507F404F7040507F404F704F507F8F8F908F8F908F8F8F8F8F908F8F90
8F8F908F8F8F908F908F8F908F8F908F8F8F908F908F8F8F908F908F8F8F908F908F8F8F908F90\
8F8F8F9090908F8F8F908F908F8F8F9090908F8F8F9090908F8F8F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F803F2F3F3F303F3F2F3F3F303F3F2F3F3F3040
7F8080808F807F8080808F807F807F8080807F807F808F807F807F8080807F807F8080807F807F\
80807F7F807F8080807F807F80807F7F807F80807F7F807F7F807F7F807F80807F
7F807F7F807F7F807F7F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F707F807F7F7F70\
7F807F7F7F707F807F7F7F707F80707F7F707F8070707F707F80707F7F707F8070
707F707F8070707F707F80703F40603F406F3F40603F4F6F3F40603F4F6F3F4060404F6F3F4060\
3F4F6F3F406F404F6F3F406F404F6F3F406F404F703F406F404F703F4F6F404F70
3F4F6F404F703F4F6F404F703F4F6F404F70404F70404F703F4F70404F70404F70405070404F70\
405070404F7040507F80808F8F8F908F808F8F8F908F808F8F8F908F808F8F8F90
8F808F8F8F908F8F8F8F8F908F808F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F\
8F8F908F8F8F908F908F8F8F8F8F908F8F8F908F908F8F8F8F8F8F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F804030403F2F3F4030403F303F403040808080
808F80808080808F80808080808F80808080808F80808080808F80808080808F807F807F808F80\
80807F808F807F807F808F807F807F8080807F807F808F807F807F80807F7F807F
80807F7F807F80807F7F807F80807F7F807F80807F7F807F80807F7F807F7F807F7F807080807F\
7F80707F807F7F80707F807F7F80707F807F7F7F707F807F7F80707F80707F7F70
7F807F7F7F707F80707F7F70404F6F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F\
3F4F6F404F6F3F4F6F404F70404F6F404F703F4F6F404F70404F6F404F70404F6F
404F70404F70404F70404F70405070404F70405070404F7040507F404F7040507F404F704F507F\
404F708F8F908F80908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F90
908F908F8F908F8F908F8F90908F908F8F90908F908F8F8F908F908F8F90908F908F8F8F909090\
8F8F909090908F8F8F9090908F8F8F9090908F8F8F9090908F8F8F909090000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F803F2F3F3F303F3F2F3F3F30403F2F3F808F80
808080808F807F8080808F807F8080808F807F807F8080807F807F808F807F807F8080807F807F\
8080807F807F80807F7F807F8080807F807F80807F7F807F80807F7F807F7F807F
7F807F80807F7F807F7F807F7F807F7F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70\
7F807F7F7F707F807F7F7F707F807F7F7F707F80707F7F707F8070707F707F8070
7F7F707F8070707F707F80703F40603F4F6F3F40603F406F3F4060404F6F3F40603F4F6F3F406F\
404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F4F6F404F703F4F6F404F70
3F4F6F404F703F4F6F404F70404F70404F703F4F70404F70404F70405070404F708F8F908F8090\
8F8F9080808F8F8F908F80908F8F908F808F8F8F908F808F8F8F908F808F8F8F90
8F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F\
8F8F908F8F8F8F8F908F8F8F908F908F8F8F8F8F908F8F8F9090908F8F8F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F7F4030403F2F3F4030403F2F3F808F80808080
808F80808080808F80808080808F807F8080808F80808080808F807F807F808F807F807F808F80\
7F807F808F807F807F8080807F807F808F807F807F80807F7F807F8080807F807F
80807F7F807F80807F7F807F7F807F7F807F80807F7F807F7F807F7F80707F807F7F80707F807F\
7F80707F807F7F80707F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70
7F80707F7F707F80707F7F703F4F6F3F4060404F6F3F4060404F6F3F4060404F6F3F406F404F6F\
3F406F404F6F3F4F6F404F6F3F4F6F404F703F4F6F404F703F4F6F404F70404F6F
404F70404F6F404F70404F70404F70404F70405070404F704050708F80908F8F908F80908F8F90\
8F80908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F90
8F8F908F8F908F8F908F8F908F8F908F8F8F908F908F8F908F8F908F8F8F9090908F8F90908F90\
8F8F8F9090908F8F8F9090908F8F8F9090908F8F8F9090908F8F8F909090000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F803F2F3F3F30403F2F3F808F807F8080808F80
7F8080808F807F8080808F807F807F8080807F807F808F807F807F8080807F807F8080807F807F\
8080807F807F8080807F807F7F807F7F807F80807F7F807F7F807F7F807F7F807F
7F807F7F807F7F807F7F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F707F807F7F7F70\
7F807F7F7F707F807F707F707F80707F7F707F8070707F707F8070707F707F8070
707F707F8070707F707F80703F40603F40603F40603F4F6F3F40603F406F3F40603F4F6F3F4060\
3F4F6F3F406F404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F4F6F404F70
3F4F6F404F703F4F6F404F703F4F6F404F708080908F8F9080808F8F8F908080908F8F9080808F\
8F8F908080908F8F9080808F8F8F908F808F8F8F9080808F8F8F908F8F8F8F8F90
8F808F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F\
8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F804030403F2F3F808F8F808F80808F80808080
808F80808080808F80808080808F80808080808F807F8080808F80808080808F807F807F808F80\
7F807F808F807F807F808F807F807F8080807F807F808F807F807F80807F7F807F
8080807F807F80807F7F807F80807F7F807F7F807F7F807F80807F7F807F7F807F7F80707F807F\
7F80707F807F7F80707F807F7F80707F807F7F7F707F807F7F80707F80707F7F70
7F807F7F7F707F8070707F70404F6F3F4060404F6F3F4060404F6F3F406F404F6F3F406F404F6F\
3F4F6F404F6F3F4F6F404F703F4F6F404F703F4F6F404F70404F6F404F703F4F6F
404F70404F70404F70404F708F8F908F80908F8F908080908F8F908F8F908F8F908F80908F8F90\
8F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F90
8F8F908F8F908F8F908F8F90908F908F8F908F8F908F8F909090908F8F908F8F908F8F8F909090\
8F8F909090908F8F8F9090908F8F909090908F8F8F9090908F8F8F909090000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F803F2F3F3F3040808080808F807F8080808F80
7F8080808F807F8080808F807F8080808F807F807F8080807F807F808F807F807F8080807F807F\
8080807F807F80807F7F807F8080807F807F7F807F7F807F80807F7F807F7F807F
7F807F7F807F7F807F7F807F7F807F7F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70\
7F807F7F7F707F807F7F7F707F807F707F707F80707F7F707F8070707F707F8070
707F707F8070707F707F8070707F703F4F6F3F40603F406F3F40603F4F6F3F40603F4F6F3F406F\
404F6F3F406F404F6F3F406F404F6F3F406F404F6F3F4F6F404F703F406F404F70
3F4F6F8F8F908080908F8F908080908F8F908080908F8F908080908F8F908080908F8F90808090\
8F8F908080908F8F908F80908F8F9080808F8F8F908F8F908F8F908F808F8F8F90
8F8F908F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F\
8F8F908F8F8F8F8F908F8F8F8F90908F8F8F8F8F908F8F8F9090908F8F8F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F7F6F6F80808F80808F80808080808F80808080
808F807F8080808F80808080808F807F8080808F807F8080808F807F807F808F807F807F808F80\
7F807F808F807F807F8080807F807F808F807F807F80807F7F807F8080807F807F
7F807F7F807F80807F7F807F7F807F7F807F7F807F7F807F7F807F7F80707F807F7F807F7F807F\
7F80707F807F7F80707F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70
7F8070707F707F80707F7F707F80703F4060404F6F3F4060404F6F3F4060404F6F3F406F404F6F\
3F406F404F6F3F4F6F404F6F3F406F404F703F4F6F404F703F4F6F8F8F9F808090
8F8F908080908F8F908080908F8F908080908F8F908F80908F8F908080908F8F908F8F908F8F90\
8F80908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F90
8F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F8F9090908F8F908F8F90\
8F8F8F9090908F8F908F90908F8F8F9090908F8F8F9090908F8F8F9090908F8F8F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F806F6F7F6F6F806F6F7F6F6F807F8080808F80
7F80808080807F8080808F807F807F8080807F807F808F807F807F8080807F807F8080807F807F\
7F80807F807F8080807F807F7F807F7F807F7F807F7F807F7F807F7F807F7F807F
7F807F7F807F7F807F7F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F707F807F707F70\
7F807F7F7F707F807F707F707F8070707F707F8070707F707F8070707F707F8070
707F707F8070707F707F8070707F6F3F40603F40603F406F3F40603F406F3F40603F4F6F3F4060\
3F4F6F3F406F404F6F3F406F3F4F6F3F406F404F6F8080908080908080908F8F90
8080908F80908080908F8F908080908F8F908080908F8F908080908F8F908080908F8F90808090\
8F8F908080908F8F9080808F8F8F908080908F8F9080808F8F8F908F8F908F8F90
80808F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F\
8F8F908F8F8F8F8F908F8F8F8F8F908F8F8F8F90908F8F8F8F8F908F8F8F8F9090
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F806F6F806F6F807070806F6F806F6F806F6F80
7070806F6F807070806F6F807070806F6F807070806F6F80808F807F8080808F807F807F808F80\
7F807F808F807F807F808F807F807F8080807F807F808F807F807F80807F7F807F
8080807F807F7F807F7F807F80807F7F807F7F807F7F807F7F807F7F807F7F807F7F80707F807F\
7F80707F807F7F80707F807F7F80707F807F7F7F707F807F7F80707F807F7F7F70
7F807F7F7F707F8070707F707F80703F40603F4F6F3F4060404F6F3F406F404F6F3F406F404F6F\
3F4F6F404F6F3F406F404F6F8080908F8F9F8080908F8F9F8080908F8F9F808090
8F8F9F8080908F8F908080908F8F9F808F908F8F908080908F8F9F8F8F908F8F90808F908F8F90\
8F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F908F8F90
8F8F908F8F908F8F908F8F908F90908F8F908F8F908F8F908F90908F8F908F90908F8F90909090\
8F8F908F90908F8F909090908F8F909090908F8F8F9090908F8F909090908F8F8F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F806F6F7F6F6F806F6F7F6F6F806F6F7F6F6F80
6F6F806F6F806F6F7F6F6F806F6F807070806F6F806F6F806F6F807070806F6F807070806F6F80\
7070806F6F807070806F6F807070806F6F807F807F7F807F7F807F7F807F7F807F
7F807F7F807F7F807F7F807F7F807F7F807F7F7F707F807F7F80707F807F7F7F707F807F7F7F70\
7F807F707F707F807F7F7F707F807F707F707F8070707F707F8070707F707F8070
707F707F8070707F707F8070707F703F406F3F40603F406F3F40603F4F6F3F40603F4F6F3F4060\
404F6F3F40608080908080908F8F908080908080908080908F8F90808090808F90
8080908F8F908080908F8F908080908F8F908080908F8F908080908F8F908080908F8F90808090\
8F8F908080908F8F90808F908F8F908080908F8F90808F908F8F90808F8F8F8F90
8F8F908F8F90808F8F8F8F908F8F908F8F908F8F8F8F8F908F8F908F8F908F8F8F8F8F908F8F8F\
8F8F908F8F8F8F8F908F8F8F8F90908F8F8F8F90908F8F8F8F90908F8F8F8F9090
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F7F6F6F806F6F806F6F806F6F807070806F6F80
6F6F806F6F807070806F6F807070806F6F807070806F6F807070806F6F807070806F6F80707080\
6F6F8070708F6F70807070806F6F8070708F70708070708F70708070708F707080
70708F70708070708F70708070708F7F807F7F807F7F807F7F807F7F80707F807F7F807F7F807F\
7F7F707F807F7F80707F807F7F7F707F807F7F7F707F807F707F707F807F7F7F70
7F8070707F707F8070707F707F8070707F70404F6F3F40603F4F6F3F4060404F6F3F406F80809F\
8080908F8F9F8080908F8F9F8080908F8F9F8080908F8F9F8080908F8F9F808090
8F8F908080908F8F9F8080908F8F908080908F8F9F808F908F8F908080908F8F90808F908F8F90\
808F908F8F908F8F908F8F90808F908F8F908F8F908F8F908F8F908F8F908F8F90
8F8F908F8F908F8F908F8F908F8F908F8F908F90908F8F908F8F908F8F908F90908F8F908F9090\
8F8F908F90908F8F908F90908F8F8F9090908F8F908F90908F8F8F9090908F8F90
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F806F6F7F6F6F806F6F7F6F6F806F6F7F6F6F80
6F6F7F6F6F806F6F806F6F806F6F7F6F6F806F6F807070806F6F806F6F806F6F807070806F6F80\
7070806F6F807070806F6F807070806F6F807070806F6F807070806F6F8070708F
6F6F8070708070708070708F6F6F8070708F70708070708F70708070708F70708070708F707080\
70708F7070807F807F707F707F8070707F707F8070707F707F8070707F707F8070
707F707F7F70707F707F8070707F6F707F703F40603F406F3F40603F406F80809080809F808090\
808090808090808090808090808090808090808F908080908080908080908F8F90
808090808F908080908F8F90808090808F908080908F8F908080908F8F908080908F8F90808090\
8F8F908080908F8F908080908F8F90808F908F8F908080908F8F90808F908F8F90
808F8F8F8F908F8F908F8F90808F8F8F8F908F8F908F8F90808F8F8F8F908F8F8F8F8F908F8F8F\
8F8F908F8F8F8F90908F8F8F8F8F908F8F8F8F90908F8F8F8F90908F8F8F8F9090
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F806F6F806F6F807070806F6F806F6F806F6F80
7070806F6F807070806F6F807070806F6F807070806F6F8070708F6F6F807070806F6F8070708F\
6F708070708F6F6F8070708F70708070708F70708070708F70708070708F707080
70708F70708070708F70708070708F70708F70708F707080707F8F70708F70708F70708F7F7F8F\
70708F7F7F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F70708F
7F7F8F707F707F8070707F707F8070707F703F4F6F8080908F8F9F808090808F9F8080908F8F9F\
8080908F8F9F8080908F8F9F8080908F8F9F8080908F8F9F8080908F8F9F808090
8F8F9F8080908F8F9F8080908F8F9F808F908F8F9F8080908F8F9F808F908F8F90808F908F8F9F\
8F8F908F8F90808F908F8F9F8F8F908F8F908F8F908F8F908F8F908F8F908F8F90
8F8F908F8F908F8F908F8F908F90908F8F908F90908F8F908F90908F8F908F90908F8F908F9090\
8F8F908F90908F8F909090908F8F908F90908F8F909090908F8F909090908F8F90
909090000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F806F6F7F6F6F806F6F7F6F6F806F6F7F6F6F80
6F6F806F6F806F6F7F6F6F806F6F807070806F6F806F6F806F6F807070806F6F807070806F6F80\
7070806F6F807070806F6F807070806F6F807070806F6F8070708F6F6F80707080
70708070708F6F708070708F70708070708F70708070708F70708070708F70708070708F707080\
70708F70708070708F70708F707F8F70708070708F70708F7F7F8F70708F7F7F8F
70708F7F7F8F70708F7F7F8F70708F303F5F6070606F706080809080809F80809080809F808090\
80809F808090808090808090808F9F808090808090808090808F9F808090808F90
8080908F8F90808090808F908080908F8F908080908F8F908080908F8F908080908F8F90808090\
8F8F908080908F8F90808F908F8F90808F908F8F90808F908F8F90808F908F8F90
808F908F8F90808F908F8F908F8F908F8F90808F8F8F8F908F8F908F8F908F8F8F8F8F908F8F90\
8F90908F8F8F8F8F908F8F8F8F90908F8F8F8F90908F8F8F8F90908F8F8F8F9090
8F8F8F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F6F806F6F806F70806F6F80
6F6F806F6F807070806F6F807070806F6F807070806F6F807070806F6F8070708F6F6F80707080\
6F6F8070708F6F708070708F6F6F8070708F70708070708F6F708070708F707080
70708F70708070708F70708070708F70708070708F70708F70708F707080707F8F70708F70708F\
70708F7F7F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F70708F
7F7F8F70708F7F7F8F2F3050303F5F2F30506F7F6F6F70606F70606070606F7060808090808F9F\
8080908F8F9F808090808F9F8080908F8F9F8080908F8F9F8080908F8F9F808090
8F8F9F8080908F8F9F8080908F8F9F8080908F8F9F808F908F8F90808F908F8F9F808F908F8F90\
808F908F8F9F808F908F8F90808F908F8F908F8F908F8F90808F908F8F908F8F90
8F8F908F8F908F90908F8F908F8F908F8F908F90908F8F908F90908F8F908F90908F8F908F9090\
8F8F908F90908F8F908F90908F8F908F90908F8F908F90908F8F909090908F8F90
8F9090000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F7F6F6F806F6F7F6F6F806F6F7F6F6F80
6F6F7F6F6F806F6F806F6F806F6F7F6F6F806F6F806F70806F6F806F6F806F6F807070806F6F80\
7070806F6F807070806F6F807070806F6F807070806F6F807070806F6F8070708F
6F6F807070806F708070708F6F6F8070708F70708070708F70708070708F70708070708F707080\
70708F70708070708F70708070708F70708F707F8F70708070708F70708F7F7F8F
70708F2F30502F3050303F502F305030305F2F30506F70606070606F70606070606F7060606F5F\
6F706080809080809F808090808090808090808F9F808090808F90808090808F90
808090808F90808090808F90808090808F908080908F8F90808090808F908080908F8F90808090\
8F8F90808F908F8F908080908F8F90808F908F8F90808F908F8F90808F908F8F90
808F908F8F90808F908F8F90808F8F8F8F90808F908F8F90808F8F8F8F908F8F908F8F90808F8F\
8F8F908F8F8F8F90908F8F8F8F8F908F8F8F8F90908F8F8F8F90908F8F8F8F9090
8F8F8F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F70806F6F806F6F806F6F80
7070806F6F807070806F6F807070806F6F807070806F6F8070708F6F6F807070806F6F8070708F\
6F708070708F6F6F8070708F70708070708F6F708070708F70708070708F707080
70708F70708070708F70708070708F70708F70708F707080707F8F70708F70708F70708F7F7F8F\
70708F707F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F70708F7F7F8F2F3050
303F502F3050303F502F3050303F5F2F3050303F5F6F70606F7F6F6F70606F70606070606F7060\
6070606F70606070606F7060808090808F9F8080908F8F9F8080908F8F9F808090
8F8F9F808F908F8F9F8080908F8F9F808F908F8F9F808F908F8F9F808F908F8F9F808F908F8F9F\
808F908F8F90808F908F8F9F8F8F908F8F90808F908F8F9F8F8F908F8F908F8F90
8F90908F8F908F8F908F8F908F90908F8F908F90908F8F908F90908F8F908F90908F8F908F9090\
8F8F908F90908F8F908F90908F8F908F90908F8F909090908F90908F90908F8F90
909090000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F7F6F6F806F6F806F6F806F6F7F6F6F80
6F6F806F6F806F6F806F6F806F6F806F70806F6F806F6F806F6F807070806F6F807070806F6F80\
7070806F6F807070806F6F8070708F6F6F807070806F6F8070708F6F6F8070708F
6F708070708F6F708070708F70708070708F70708070708F70708070708F70708070708F707080\
70708F70708070708F70708F707F8F70708070708F70708F2F30502F304F2F3050
2F30503030502F30503030502F3050303F5F2F30506F70606070606F70606070606F7060607060\
6F7060606F5F607060606F5F607060606F5F607060606F5F808F9F808090808F90
808090808F9F808090808F908080908F8F90808090808F908080908F8F908080908F8F90808F90\
8F8F908080908F8F90808F908F8F90808F908F8F90808F908F8F90808F908F8F90
808F908F8F90808F908F8F90808F908F8F90808F908F8F908F8F908F9090808F908F8F908F8F90\
8F90908F8F8F8F90908F8F908F90908F8F8F8F90908F8F908F90908F8F8F8F9090
8F8F8F8F9090000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F6F806F6F806F70806F6F80
6F6F806F6F807070806F6F806F70806F6F807070806F6F807070806F6F8070708F6F6F80707080\
6F6F8070708F6F708070708F6F6F8070708F70708070708F6F708070708F707080
70708F70708070708F70708070708F70708070708F70708F70708F707080707F8F70708F70708F\
70708F7F7F8F70708F707F8F70708F7F7F8F2F304F3030502F304F303F502F3050
303F502F3050303F502F3050303F5F2F3050303F5F6F70606F70606070606F70606070606F7060\
6070606F70606070606F7060606F5F6F7060606F5F607060606F5F607060606F5F
808F9F8080908F8F9F808F908F8F9F8080908F8F9F808F908F8F9F808F908F8F9F808F908F8F90\
808F908F8F9F808F908F8F90808F908F8F9F808F908F8F90808F908F8F9F8F8F90
8F8F90808F908F90908F8F908F8F908F8F908F90908F8F908F90908F8F908F90908F8F908F9090\
8F8F908F90908F8F908F90908F8F908F90908F8F908F90908F8F908F90908F9090
8F90908F8F90000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F7F6F6F806F6F7F6F6F806F6F806F6F80
6F6F7F6F6F806F6F806F6F806F6F806F6F806F6F806F70806F6F806F6F806F6F807070806F6F80\
6F70806F6F807070806F6F807070806F6F8070708F6F6F807070806F6F8070708F
6F6F8070708F6F708070708F6F708070708F70708070708F6F708070708F70708070708F707080\
70708F70708070708F70708070708F2F304F2F30502F304F2F30502F304F2F3050
2F304F2F30502F30503030502F30502F30502F3050303F5F6070606F70606070606F7060606F5F\
607060606F5F607060606F5F607060606F5F607060606F5F606F5F606F5F606F5F
5F6F5F606F5F5F6F5F808F90808090808F90808090808F90808090808F908080908F8F90808090\
808F90808F908F8F908080908F8F90808F908F8F90808F908F8F90808F908F8F90
808F908F8F90808F908F8F90808F908F8F90808F908F8F90808F908F8F90808F908F9090808F8F\
8F8F908F8F908F9090808F8F8F90908F8F908F90908F8F8F8F90908F8F8F8F9090
8F8F8F8F9090000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F70806F6F806F70806F6F80
7070806F6F806F70806F6F807070806F6F807070806F6F8070708F6F6F807070806F6F8070708F\
6F708070708F6F6F8070708F70708070708F6F708070708F70708070708F707080
70708F70708070708F70708070708F70708F70708F707080707F8F70708F70708F70708F7F7F8F\
70708F707F8F70708F0000002F304F2F30502F304F3030502F304F3030502F3050
303F502F3050303F502F3050303F5F2F3050303F5F2F30506F706F6070606F70606070606F7060\
6070606F7060606F5F6F7060606F60607060606F5F607060606F5F607060606F5F
607060606F5F606F5F5F6F5F606F5F606F5F8F8F9F808F908F8F9F808F908F8F9F808F908F8F9F\
808F908F8F9F808F908F8F9F808F908F8F9F808F908F909F8F8F908F8F90808F90
8F909F8F8F908F90908F8F908F909F8F8F908F90908F8F908F90908F8F908F90908F8F908F9090\
8F8F908F90908F8F908F90908F8F908F90908F8F908F90908F90908F90908F9090
9090908F9090000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F7F6F6F806F6F806F6F806F6F7F6F6F80
6F6F806F6F806F6F806F6F806F6F806F70806F6F806F70806F6F807070806F6F806F70806F6F80\
7070806F6F807070806F6F8070708F6F6F807070806F6F8070708F6F6F8070708F
6F708070708F6F708070708F70708070708F6F708070708F70708070708F70708070708F707080\
70708F0000000000000000000000002F304F2F304F2F304F2F30502F304F2F3050
2F304F3030502F304F2F30502F3050303F502F3050303F506070606F7060606F60607060606F60\
6F7060606F5F607060606F5F607060606F5F606F60606F5F606F605F6F5F606F5F
5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6050606F5F5F6050606F5F808090808F90808F90\
8F8F90808F908F8F90808F908F8F90808F908F8F90808F908F8F90808F908F8F90
808F908F8F90808F908F8F90808F908F8F90808F908F8F90808F908F9090808F908F90908F8F90\
8F9090808F908F90908F8F908F90908F8F908F90908F8F908F90908F8F8F8F9090
8F8F908F9090000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F6F806F6F806F70806F6F80
6F70806F6F807070806F6F806F70806F6F807070806F6F807070806F6F8070708F6F6F80707080\
6F6F8070708F6F708070708F6F6F8070708F6F708070708F6F708070708F707080
70708F70708070708F70708070708F70708070708F70708F70708F707080707F8F70708F000000\
0000000000000000000000000000003030502F304F2F30502F304F3030502F304F
3030502F3050303F502F3050303F502F3050303F5F2F30506F70606070606F70606070606F7060\
606F606F7060606F60607060606F5F607060606F5F607060606F5F607060606F5F
606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6050606F5F000000000000\
808F908F8F9F808F908F8F9F808F908F8F9F808F908F8F90808F908F909F808F90
8F8F90808F908F909F8F8F908F9090808F908F90908F8F908F90908F8F908F90908F8F908F9090\
8F8F908F90908F8F908F90908F8F908F90908F90908F90908F8F908F90908F9090
8F90908F9090000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F7F6F6F80606F7F6F6F806F6F806F6F80
6F6F7F6F6F806F6F806F6F806F6F806F6F806F6F806F70806F6F806F70806F6F807070806F6F80\
6F70806F6F807070806F6F807070806F6F8070708F6F6F807070806F6F8070708F
6F6F8070708F6F708070708F6F708070708F70708070708F6F708070708F000000000000000000\
0000000000000000000000000000000000002F304F2F304F2F304F2F304F2F3050
2F304F2F30502F304F2F30502F30502F30502F3050303F502F3050607060606F60607060606F5F\
607060606F5F607060606F5F606F60606F5F606F605F6F5F606F5F5F6F5F606F5F
5F6F5F606F5F5F6F5F606F5F5F60505F6F5F5F6050606F5F5F60505F6F5F000000000000000000\
000000000000000000808F90808F90808F908F8F90808F908F8F90808F908F8F90
808F908F8F90808F908F8F90808F908F8F90808F908F9090808F908F8F90808F908F9090808F90\
8F9090808F908F9090808F908F90908F8F908F9090808F8F8F90908F8F908F9090
8F8F8F8F90908F8F90000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F70806F6F806F70806F6F80
7070806F6F806F70806F6F807070806F6F807070806F6F8070708F6F6F8070708F6F6F8070708F\
6F708070708F6F6F8070708F6F708070708F6F708070708F70708070708F707080
70708F70708070708F70708070708F70708F70708F70708F000000000000000000000000000000\
0000000000000000000000000000000000000000003030502F304F3030502F304F
303F502F3050303F502F3050303F502F3050303F502F3050303F5F6070606F70606070606F7060\
607060607060606F5F6F7060606F5F607060606F5F607060606F5F606F60606F5F
607060606F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F000000000000000000\
000000000000000000000000000000000000808F908F909F808F908F909F808F90
8F909F8F8F908F909F808F908F909F8F8F908F9090808F908F909F8F8F908F90908F8F908F909F\
8F8F908F90908F8F908F90908F90908F90908F8F908F90908F90908F90908F9090
8F90908F90908F9090000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F7F6F6F806F6F806F6F806F6F7F6F6F80
6F6F806F6F806F6F806F6F806F6F806F70806F6F806F70806F6F807070806F6F806F70806F6F80\
7070806F6F807070806F6F8070708F6F6F807070806F6F8070708F6F6F8070708F
6F708070708F6F708070708F70708070708F000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F304F2F304F2F304F2F3050
2F304F2F30502F304F2F30502F30503030502F30502F30502F30506F7060606F5F607060606F5F\
607060606F5F607060606F5F607060606F5F606F5F606F5F606F5F5F6F5F606F5F
5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F60505F6F5F5F6050000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000008F8F90
808F908F8F90808F908F8F90808F908F9090808F908F8F90808F908F9090808F908F9090808F90\
8F9090808F908F90908F8F908F9090808F908F90908F8F908F9090808F908F9090
8F8F908F90908F8F90000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F6F806F6F806F70806F6F80
6F70806F6F806F70806F6F806F70806F6F807070806F6F807070806F6F8070708F6F6F8070708F\
6F6F8070708F6F708070708F6F6F8070708F6F708070708F6F708070708F707080
70708F6F708070708F707080000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F304F3030502F304F
2F30502F304F303F502F30503030502F3050303F502F3050303F502F30506F7060607060607060\
606F5F607060606F5F607060606F5F607060606F5F606F60606F5F607060606F5F
606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000008F909F808F908F9090808F908F909F8F8F908F9090808F908F909F8F8F908F9090\
808F908F90908F8F908F90908F8F908F90908F90908F90908F8F908F90908F9090
8F90908F90908F9090000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F7F6F6F80606F7F6F6F806F6F806F6F80
6F6F7F6F6F806F6F806F6F806F6F806F6F806F6F806F70806F6F806F70806F6F806F70806F6F80\
6F70806F6F807070806F6F807070806F6F8070708F6F6F807070806F6F8070708F
6F6F8070708F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F304F2F304F
2F304F2F30502F304F2F30502F304F2F30502F30503030502F30502F3050606F5F607060606F5F\
607060606F5F6070605F6F5F606F5F606F5F606F5F5F6F5F606F5F5F6F5F606F5F
5F6F5F606F5F5F6F5F606F5F5F60505F6F5F5F60505F6F5F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008F8F90808F908F9090808F908F8F90808F908F9090808F90\
8F9090808F908F9090808F908F9090808F908F9090808F908F90908F8F908F9090
808F908F90908F8F90000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F70806F6F806F70806F6F80
6F70806F6F806F70806F6F8070708F6F6F807070806F6F8070708F6F6F8070708F6F6F8070708F\
6F708070708F6F6F8070708F6F708070708F6F708070708F70708070708F6F7080
70708F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F30502F304F
3030502F304F3030502F304F303F502F3050303F502F3050303F5F2F30506F7060606F606F7060\
607060607060606F5F607060606F5F607060606F5F607060606F5F606F5F5F6F5F
606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6050000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008F909F8F8F908F909F808F908F909F\
8F90908F90908F8F908F909F8F90908F90908F90908F909F8F90908F90908F9090
8F90908F90908F90908F9090000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F7F6F6F806F6F806F6F806F6F806F6F80
6F6F806F6F806F6F806F6F806F6F806F70806F6F806F70806F6F806F70806F6F806F70806F6F80\
7070806F6F807070806F6F8070708F6F6F8070708F6F708070708F6F6F80000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F304F
2F304F2F30502F304F2F30502F304F3030502F304F2F30502F3050303F50606F5F607060606F5F\
607060606F5F607060606F5F6070605F6F5F606F5F5F6F5F606F5F5F6F5F606F5F
5F6F5F606F5F5F60505F6F5F5F60505F6F5F5F6050000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000808F90\
8F9090808F908F9090808F908F9090808F908F9090808F908F9090808F908F9090
8F90908F9090808F908F9090000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F6F806F6F806F70806F6F80
6F70806F6F806F70806F6F806F70806F6F8070708F6F6F806F70806F6F8070708F6F6F8070708F\
6F6F8070708F6F708070708F6F6F8070708F6F708070708F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F30502F304F3030502F304F3030502F3050303F502F3050303F502F3050303F5F607060607060\
606F5F607060606F5F607060606F5F607060606F5F606F5F5F6F5F606F5F5F6F5F
606F5F5F6F5F606F5F5F6F5F606F5F5F6050606F5F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000008F90908F9090808F908F909F8F90908F90908F90908F90908F9090
8F90908F90908F90908F9090000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F7F6F6F80606F7F6F6F806F6F806F6F80
6F6F806F6F806F6F806F6F806F6F806F6F806F6F806F70806F6F806F70806F6F806F70806F6F80\
6F70806F6F807070806F6F806F70806F6F80000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F304F2F304F2F304F2F30502F304F2F30502F304F2F30502F304F2F30502F3050607060606F5F\
606F60606F5F6070605F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F
5F60505F6F5F5F60505F6F5F5F60505F6F5F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000808F908F9090808F908F9090808F908F9090
808F908F90908090908F9090000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F70806F6F806F70806F6F80
6F70806F6F806F70806F6F8070708F6F6F806F70806F6F8070708F6F6F8070708F6F6F8070708F\
6F708070708F6F708070708F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F304F3030502F304F303F502F3050303F502F3050303F502F3050303F50606F60607060\
606F60607060606F5F607060606F5F607060606F5F607060606F5F606F5F5F6F5F
606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000008F90908F90908F9090
8F909F8F90908F90908F9090000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F7F6F6F806F6F806F6F806F6F806F6F80
6F6F806F6F806F6F806F6F806F6F806F70806F6F806F70806F6F806F70806F6F806F70806F6F80\
70708F6F6F80000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F304F2F304F2F304F2F30502F304F2F30502F30503030502F3050607060606F5F\
6070605F6F5F606F60606F5F606F605F6F5F606F5F5F6F5F606F5F5F6F5F606F5F
5F6F5F606F5F5F60505F6F5F5F6050000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8090908F90908090908F9090000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F6F806F6F806F70806F6F80
6F70806F6F806F70806F6F806F70806F6F8070708F6F6F806F70806F6F8070708F6F6F8070708F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003030502F304F2F30502F304F303F502F30503030502F3050303F502F3050607060\
606F5F607060606F5F6070605F6F5F607060606F5F606F5F5F6F5F606F5F5F6F5F
606F5F5F6F5F606F5F5F6F5F5F6F5F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008F9090000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F806F6F80606F7F6F6F806F6F806F6F80
606F806F6F806F6F806F6F806F6F806F6F806F6F806F70806F6F806F70806F6F80000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F304F2F304F2F30502F304F2F30502F304F2F30502F30503030505F6F5F\
606F605F6F5F606F605F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F
5F60505F6F5F5F60505F6F5F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F70806F6F806F70806F6F80
6F70806F6F806F70806F6F8070708F6F6F806F708F6F6F8070708F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003030502F304F3030502F304F303F502F3050303F502F3050607060\
606F5F607060606F5F607060606F5F606F605F6F5F606F605F6F5F606F5F5F6F5F
606F5F5F6F5F606F5F5F605F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F7F6F6F806F6F806F6F80606F806F6F80
6F6F806F70806F6F806F6F806F6F806F70806F6F806F7080000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F304F2F30502F304F2F30502F304F2F30502F304F2F3050606F5F\
6070605F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F605F5F6F5F
5F605F5F6F5F5F6050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F70806F6F806F6F806F6F806F70806F6F80
6F70806F6F806F70806F6F806F70806F6F80000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F304F3030502F304F3030502F304F303F502F3050303F50\
606F5F607060606F5F606F605F6F5F606F605F6F5F606F5F5F6F5F606F5F5F6F5F
606F5F5F605F606F5F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F806F6F80606F7F6F6F806F6F806F6F80
606F806F6F806F6F806F7080000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F304F2F30502F304F2F30502F304F2F30502F304F\
606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F605F5F6F5F5F605F5F6F5F
5F60505F6F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F806F6F806F70806F6F806F70806F6F80
6F708F6F6F80000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F30502F304F303F502F3050303F502F3050303F50\
606F5F6070605F6F5F6070605F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F
606F5F5F6F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F806F6F806F6F806F6F80606F806F6F80
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F304F2F304F2F30502F304F2F30502F3050\
606F605F6F5F606F5F5F6F5F606F5F5F6F5F5F6F5F5F6F5F606F5F5F60505F6F5F
5F6050000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F6F806F6F806F6F80000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F30502F304F303F502F3050303050\
2F30506070605F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F
5F6F5F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F6F80000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F304F2F304F2F304F2F30502F304F\
2F30505F6F5F606F5F5F6F5F5F6F5F5F6F5F5F6F5F5F60505F6F5F5F605F5F6F5F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F304F3030502F304F303F50\
2F3050606F605F6F5F606F605F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F304F2F30502F304F\
2F30505F6F5F606F5F5F6F5F606F5F5F605F5F6F5F5F6F5F5F6F5F5F6050000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003030502F304F2F3050\
2F304F303F505F6F5F606F5F5F6F5F606F5F5F6F5F5F6F5F5F6F5F606F5F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F30502F304F\
2F30502F304F606F5F5F605F5F6F5F5F6F5F5F6F5F5F60505F6F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303F50\
2F3050303F505F6F5F606F605F6F5F606F5F5F6F5F606F5F5F6F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F304F\
2F30502F304F5F6F5F5F6F5F5F6F5F5F605F5F6F5F5F605F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F304F303F502F3050606F5F5F6F5F606F5F5F6F5F5F6F5F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F304F2F30505F605F5F6F5F5F605F5F6F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003030502F304F606F5F5F6F5F5F6F5F5F6F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F30502F304F5F6F5F5F6050000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F30505F6F5F5F6F5F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F6F5F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F6F5F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[14]:=",
  ImageSize->{288, 286.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P00017PP10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol0
4@091`001`Oo01402@L000L7o`0A00T700071gh000>;0000S`091`001`Mm0003T8Z@09002@L000L7
O@00176;LH^?00T700071g`000N:T8Z@RY2:08d02@L000L7N`002G6@LI1aT76@L@2<00T700071gX0
00Z:T8Z@RY2:T8Z@S0091`001`Mj000:T76@LI1aT76@LH`02@L000L7NP002hZ@RY2:T8Z@RY2:08/0
2@L000L7NP00391aT76@LI1aT76@LHX02@L000L7N@003Y2:T8Z@RY2:T8Z@RY2:R@091`001`Md000D
T8X0091aT76@LI1aT76@LI1aT76800T700071g<001IbSi2?08Z@RY2:T8Z@RY2:T8Z@RY2:Q`091`00
1`Mc000G676:T8Z@LI1aT76@LI1aT76@LI1aT740QP091`001`Mb000ILG9aZHnYSi2?T8Z@RY2:T8Z@
RY2:T8Z@RP2500T700071g4001/HLAQbRY2:T8Z@RY1aT76@LI1aT76@LI1aT740Q0091`001`Ma000L
LW5bLJV?ZHnYSi2?T8Z@RY2:T8Z@RY2:T8Z@RX<02@L000L7L0007G4HLQQbRY2:T8Z@RY2:T76@LI1a
T76@LI1aT76@08<02@L000L7L0007W5bLG9aZHnYSi2?T8n@Si2?T8Z@RY2:T8Z@RY2:T8802@L000L7
L0007g4HLAQbRY2:T8Z@RY2:T8Z@RY1aT76@LI1aT76@LI00P@091`001`M`000OLG9aLW5bSjV?ZHn@
Si2?T8n@Si2:T8Z@RY2:T8Z@RP2100T700071fl0024HLAQb678HT8Z@RY2:T8Z@RY2:T8Z@LI1aT76@
LI1aT740P0091`001`M_000RLW5bLG9aLXnYSi2?T8n@Si2?T8n@Si2:T8Z@RY2:T8Z@RWl02@L000L7
KP00974HLAQa678HT8Z@RY2:T8Z@RY2:T8Z@RY1aT76@LI1aT76@LBD000=a0000EP091`001`M]000W
LW5bLG9aLW5bSi2?T8n@Si2?T8n@Si2?T8n@Si2:T8Z@RY2:T8Z@024000>?LHl0F0091`001`M]000X
674HLAQa674HT8n@RY2:T8Z@RY2:T8Z@RY2:T8Z@LI1aT76@LI1aT1l000B?LHmaF0091`001`M/000Z
LG9aLW5bLG9aLXn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@RY2:T8Z@7@001HmaSg6?05P02@L000L7
K000:W4HLQQa674HLAQbSi2?T8Z@RY2:T8Z@RY2:T8Z@RY2:T76@LI1aT76@LAX000Qa08maSg6?LEP0
2@L000L7J`00;79aLW5bLG9aLW5bLI2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2:T8Z@RY2:60002G5[
Sg6?LHmaS`1H00T700071f/002`HLQQa674HLAQb676?T8n@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T76@
LI1aT1L000Ya08maSg6?LHmaF0091`001`MZ000^LG9aLW5bLG9aLW5bLI2?T8n@Si2?T8n@Si2?T8n@
Si2?T8n@Si2?T8Z@RY2:T1<000ea075[Sg6?LHmaSg6?05P02@L000L7JP00;g4HLAQb674HLAQa67:?
T8n@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@LI1aT8Z@014000ia075aSg6?LHmaSg6?LEP02@L000L7
J@00<G9aLW5bLG9aLW5bLG9aT8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z@Ri003`003g40
LHZ?LHmaSg6?LHmaS`1H00T700071fT0038HLAQb678HLAQa678HLXn@Si2:T8Z@RY2:T8Z@RY2:T8Z@
RY2:T8Z@RY2:T8Z@RY2:T0d0011a075aSg6?LHmaSg6?LHmaF0091`001`MX000BLG9aLW5bLG9aLW5b
LG9aLXnY19007Xn@T92?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2:T0P000=[LF/00W403hmaSg6?LHma
Sg6?LHmaS`1H00T700071fL003HHLAQa674HLAQa678HLQQb692:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@
RY2:T8Z@RY2:T8Z@LI05000FL@1a0740Sg6?LHmaSg6?LHmaRW6:LEP02@L000L7EP000f]a000>000C
LW5bLG9aLW5bLG9aLW5bLG:?Z@06T00NSi2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8^@0`00175[
LF/2L@0AShZ?LHmaSg6?LHmaSg6?LHl0F0091`001`MF0005LF]aJg402`00Dg4HLAQa674HLAQb678H
LQQb692:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY1aT01a0740L@2?LHmaSg6?LHma
Sg6?LHYaRW6:05L02@L000L7EP0021QaL75[LG1a2000575bLG9aLW5bLG9aLW5bLG9aLXnY2900=hn@
Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2;T01aK76:ShZ?RXmaSg6?LHmaSg6?LHmaSg40E`091`00
1`MH0009LF]aJg40L@1a00D003aa674HLAQa674HLQQb678HLQR@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY1aT002L@0ESg6?LHmaSg6?LHmaSg6?LHYaRW6:05L02@L000L7E`003G4H
LG1aJg5[LF]aJg400P00575bLG9aLW5bLG9aLW5bLG9aLXnY1Y00>Hn@T92?T8n@Si2?T8n@Si2?T8n@
Si2?T8n@Si2?T8n@Sg6:ShZ?RXn:Sg6?LHmaSg6?LHmaSg6?L@1G00T700071eP0065a075[LF]a075[
L@1a674HLAQa674HLAQb678HLQQb692:T8n@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY1a
T8Z@Ji1aSg6?LHmaSg6?LHmaSg6?LHYaRW6:05L02@L000L7E`00174HLAP3L@0LJg5[LF]aJg9aLW5b
LG9aLW5bLG9aLW5bLG9aZ@R@01n?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?00Ea0i0048ma
Sg6?LHmaSg6?LHmaSg5G00T700071eP004Ma0740LF]aJg5[LF]a674HLAQa674HLAQb678HLQQb67::
T8Z@Si2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T002L@0HJg5[LF]aJg5[T8Z@RY1aSg6?
LHYaRW6:E`091`001`MH000;674HLAQaJg5[LF/01740@W9aLW5bLG9aLW5bLG9aLW5bLI2?T8n@Si2?
T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si1/LFaaK75/LFaaK0=a00A/LG5a1I001ZV@T92YT5D0
2@L000L7F000JG40L@1a675[LF]a674HLAQa678HLQQb674HLQQb678HLXn@RY2:T8Z@RY2:T8n@Si2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8YaJg5[LF]aJg5[LF]aJg5[LF]aJg5[T8Z@RY2:T8Z@RY2:T01?00T7
00071eP000THLAQa675[LG001W40=W9aLW5bLG9aLW5bLG9aLW5bLI2?T8n@Si2?T8n@Si2?T8n@Si2?
T8n@Si2?T8n@Si2?T8n@S`=a00U/LFaaK75/LF`02g402VaaK:V@ZI2YT:T9T4T02@L000L7F000MG40
L@1a675[LF]a674HLAQa674HLQQb674HLAQb678HLXn@RY2:T8Z@RY2:T8n@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[T8Z@RY2:T8Z@RY2:T8Z@RY2:T01300T7
00071eP000THLAQa674HLF/01W40>79aLW5bLG9aLW5bLG9aLW5bLI2?T8n@Si2?T8n@Si2?T8n@Si2?
T8n@Si2?T8n@Si2?T8maK75/0g401faaK75/LF`02g402VaaK75/LFaaK9H;T007ZI2YT:V@Z@02T3h0
2@L000L7F000I740L@1a674HLAQa674HLAQa678HLQQb674HLQQb678HLQR@RY2:T8Z@RY2:T8n@Si2:
T8n@RY2:T8Z@RY2:T8Z@RW5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF/3T00JRY2:T8Z@
RY2:T8Z@RY2:T8Z@RY2:T8Z@RY0g00T700071eT000Aa674H0W400hYaRP05L@0bLW5bLG9aLW5bLG9a
LW5bLG9aLXn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8l3L@0=K75aLFaaK75/LFaaK00;L@0:
K75/LFaaT9J@UP^@00NYT:V@ZI2Y00^@00>YT:T0<P091`001`MJ001NL@1aL8Y`RW1a674HLAQa674H
LQQb674HLAQb678HLQR@RY2:T8Z@RY2:T8n@RY2?T8Z@RY2:T8Z@RY2:T8YaJg5[LF]aJg5[LF]aJg5[
LF]aJg5[LF]aJg5[LF]aJ`N@02::T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@;`091`00
1`MI009a00N:LHYaRW6:00Ea03AbLG9aLW5bLG9aLW5bLG9aLW5bSi2?T8n@Si2?T8n@Si2?T8n@Si2?
T8n@Si2?T8maK75/0g402faaLG5/LFaaK75/00]a00Y/LI2AT96@TI2F2i001jV@ZI2YT:T02i001JV@
ZI2Y03002@L000L7E`00GF^:JhY[RW2:L8Y`RW1a674HLAQb678HLQQa678HLQQb678HT8Z@Si2:T8Z@
RY2?T8n@Si2?T8Z@RY2:T8Z@RY1[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJ`0;T00PRY2:T8Z@
RY2:T8Z@RY2:T8Z@RY2?T8Z@Si2:T8Z@RY0a00T700071eD000n:LHYaRW6:LHYaRW6:LHX00g40779a
LW5bLG9aLW5bLG9aLW5bLG:?T8n@Si2?T8l3T00CSi2?T8n@Si2?T8n@Si2?T8maK009L@07K75/LFaa
K00:L@:@00VAT96@TI2FT9402i001jV@ZI2YT:T01i002:V@T92YT:V@<@091`001`M@001TRV^:L8Y`
RW2:L8Y`RV^:JhY[RV^:Jg4HLAQa674HLAQb678HLQQb678HT8n@Si2?T8n@Si2:T8Z@RY2:T8Z@RY2:
T8Z@Jg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[T7N@M`^@01j:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@
RY2:T8Z@RY0c00T700071dh002aaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaLW5bLG9aLW5bLG9aLW5b
LG9aLPZ@012?T8n@Si2?T8n@Si2?T8n@17402FaaK75/LFaaK009L@0:K75aLI2FT96@UP^@01fYT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@Z@0d00T700071d/006U[RV^:JhY[RW2:L8Y`RW2:L8Y[RV^:
JhY[RV^:674HLAQa674HLQQb678HLQQb692?T8n@Si2?T8n@RY2:T8Z@RY2:T8Z@RY1[LF]aJg5[LF]a
Jg5[LF]aJg5[LF]aJg5[T7N@Mi1gT7L02i007HZ@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:03@0
2@L000L7B@00<HYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaLW5bLG9aLW5bLG9aLW5bLG80
2Y003Xn@Si2?T8n@Si2?T8n@1W402FaaK75/LFaaK009L@0:T96@TI2AT9J@T@^@01bYT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@=@091`001`M7001RL8Y[RV^:JhY[RV^:L8Y`RW2:L8Y`RV^:JhY[RV^:
JhY`RQQa674HLAQb678HLQQb678HT8n@Si2?T8n@Si2:T8Z@RY2:T8Z@RY2:LF]aJg5[LF]aJg5[LF]a
Jg5[LF]aJg42T009Mi1gT7N@Mi1g00^@01b:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@=@091`00
1`M4000fLHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaLW5bLG9aLW5bLG9aLW5b
2Y002Hn@Si2?T8n@S`03T003Sg5a00Ea00E/LFaaK003L@03K75a00Aa19002I6@TI2FT96@UP0;T00L
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT3D02@L000L7@P00HhY`RW2:L8Y`RV^:L8Y[RW2:L8Y`
RW2:L8Y`RV^:JhY[RV^:JhY`RW2:674HLAQb678HLQQb678HT8n@Si2?T8n@Si2:T8Z@RY2:T8Z@RW5[
LF]aJg5[LF]aJg5[LF]aJg5[L@02T00=Mi2@T7N@Mi1gT7N@M`0;T00KRY2:T8Z@RY2:T8Z@RY2:T8Z@
RY2:T8Z@RY2:03H02@L000L7@000>W6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:
LHYaRW6:LG9aLW5bLG9aLW5bLG8:T00;Si2?T8n@Si2?T8l02G401faaK75/LF`01G47T009TI2AT96@
TI2A00^@01ZYT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT3L02@L000L7?@00Mg2:JhY`RW2:L8Y`RV^:
L8Y[RW2:JhY`RW2:L8Y`RW2:JhY[RV^:JhY[RW2:L8Y`RQQa678HLQQb678HLQR@Si2?T8n@Si2?T8Z@
RY2:T8Z@RW5[LF]aJg5[LF]aJg5[LF]aJg5[T7N@Mi1gT7N@Mi1gT7N@Mi1g00>@00EgT92@M`03T00I
RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RP0h00T700071c/0042:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LG9aLW5bLG9aLW5bLG9a2I002Hn@Si2?T8n@S`0;L@05
K75aLF`00g400fb@T008T009TI2AT96@TI2A00^@01VYT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2Y03P0
2@L000L7=P00J91`RW6:L8YaRW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`
RW2:L8Y`RQQa674HLQQb678HLXn@Si2?T8Z@Si2?T8n@Si2?LF]aJg5[LF]aJg5[LF]aJg5g0i002gN@
Mi1gT7N@Mi1g00V@025gT7N@Mi1gT7N@Si2?T8n@Si2?T8Z@RY2:T8Z@Si2?T8l0>P091`001`Lg001<
T7N@MhYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LG9a
LW5bLG9aLW6@Si2?T8n@S`F@00n?T8n@K75/LFaaK75/LF`01748T005TI2@T9405i005jV@T92YT:V@
ZI2YT:V@ZI2YT:V@ZI2Y03X02@L000L7=`00IG:@LY1bT7:@LHY`RW2:L8Y`RW6:L8YaRW2:LHY`RW2:
L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L74HLQQb678HLQQbSi2?T8n@Si2?T8n@Si2?T6]aJg5[LF]a
Jg5[LF]aMi1g00>@00]gT7N@Mi1gT7N@M`03T003Mi2@00F@021gT7N@Mi1gT7N@Si2?T8n@Si2?T8Z@
RY2:T8Z@Si2?T3/02@L000L7=`00A91gT7N@Mi1gT7N@LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa
RW6:LHYaRW6:LHYaRW6:LHYaRW5bLG9aLW5bLG9a0i001Hn@Si2?00N@17402VaaK75/LFaaK749T005
TI2@T9406I005JV@ZI2YT:V@ZI2YT:V@ZI2YT:V@Z@0l00T700071cH009j;LI1bT7:@LY1bT7:@LY1b
RW2:L8YaRW2:LHY`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`RQQa678HLQQb67:?T8n@Si2?
T8n@Si2?T6]aJg5[LF]aJg5[LF]aMi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T8n@Si2?T8n@Si2:T8Z@RY2:T8n@Sc`02@L000L7=000BG6@LY1bT7N@Mi1gT7N@LY1gT7N@Mi1aRW6:
LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LG9aLW5bLG9aT8l00i000hn@S`05T003
K75a00=a00Q/LFaaK75/TBV@01BYT:V@ZI2YT:V@ZI2YT:V@ZI2YT3d02@L000L7<@00Sg6@LI1aRg6@
LI1bT7:@LY1bT7:@LY1bT7:@LY1bRW6:LHY`RW2:L8Y`RW2:L8YaRW2:LHY`RW2:L8Y`RW2:L8Y`RQQb
678HLQQb692?T8n@Si2?T8n@Jg5[LF]aJg5[LF]aMi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi2?00>@00n?T92@Si2?T8Z@Si2:T8l0?P091`001`L_001=T76@LI1aT7:@LY1g
T7N@Mi1bT7N@LY1gT7N@Mi1gT7N@Mi1aRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LG9a
LW5bLG800Y001Hn@T92?00>@274026aaK96@TI2A:I004ZV@ZI2YT:V@ZI2YT:V@ZI2YT3l02@L000L7
;@00TG6@LI1aT76@LH]aRg6@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1`RW2:L8Y`RW6:L8YaRW2:
LHY`RW2:L8Y`RW2:L8Y`RQQb678HLQR@Si2?T8n@Si2?LF]aJg5[LF]aMi1gT7N@Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7L00i0048n@T92?T8n@RY2?T8Z@Si0o00T700071bh0
04iaT76@LI1aT7:@LY1bT7N@Mi1bT7:@Mi1bT7N@Mi1gT7N@Mi1gT7N@Mi1gRW6:LHYaRW6:LHYaRW6:
LHYaRW6:LHYaRW6:LHYaLW5bLG84T003Si2?00:@27401I6@TI2A00>@00>AT9009i004ZV@ZI2YT:V@
ZI2YT:V@ZI2YT3l02@L000L7<000X8]aRg6;LI1aT76@LI1bT79gLY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LXY`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:678HLQR@Si2?T8n@Sg5[LF]aJg5gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@RY2:T8Z@RY2:T8n@Si2?
T8m000T700071c4003B@LY1bT7:@LI1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1a2hX0376:LHYaRW6:LG9aLPN@00UaK75/LFbAT9402i001I6@TI2A00>@00>AT9007I004:V@
ZI2YT:V@ZI2YT:V@ZI1100T700071c8009b;LH]aT76@LI1aT7:@LWMbMg:@LWMbT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT72:L8Y`RW2:L8Y`RW2:L8Y`RQQb692?T8n@Sg5[LF]aMi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi2:T8Z@RY2:T8Z@Si2?
T8m200T700071c<003R@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@M`F:00eaRW6:LHYaRW6:LG9a00B@17401i6@TI2AT9402i001i6@TI2AT9408I003jV@
ZI2YT:V@ZI2YT:V@Z@1200T700071c<009YaRg6@LI1aT76@LI1bMg:@LWMbT79gLY1bT7:@LY1bT7:@
LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1`RW2:L8Y`RW2:L78HLXn@Si1[LF^@Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT8Z@RY2:T8Z@RY2?T8n@
@`091`001`Ld000oLY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1a00>:00=aRXX00XX00g9aT002T0=a0Y001i6@TI2AT9402i001I6@T92A02>@
00fYT:V@ZI2YT:V@ZI2Y04@02@L000L7=@00UW6@LI1aT76@LI1aMg9gLWMbMg9gLY1bT7:@LY1bT7:@
LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bRW2:L8Y`LXn@Jg5gT7N@Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT8Z@RY2:T8Z@RY2?T4D0
2@L000L7=P00A7:@LY1bT7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gTGN@Mi5gTGNA0XX00g6@L@06T007TI2AT96@T@0=T005TI2@T9408I003:V@ZI2YT:V@
ZI2YT4D02@L000L7=P00TY1aT76@LI1aT75gLWMbMg9gLWMbT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LI1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@RY2?T8Z@Si2:0Y1600T700071cL004B@LY1b
T7:@LY1gT7N@Mi1gT7N@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
TGN@M`:@00AbRXn:19001i6@TI2AT940<i002ZV@ZI2YT:V@ZI1700T700071cT008EaT76@LI1aT7:@
LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LW6?T8mb
CXmaRW6:LHYgT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@
Mi1g00Z@B0091`001`Lj000nLY1bT7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi02L@B@00aaLW:?RXn:ShZ?RXleT009ZI2YT:V@ZI2Y04P02@L000L7
>P00Q91aT76@LI1aT7:@LY1bT79gLY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1[LF^@Si2?LW5bCXmaRW6:LHYaRW6:Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@M`Z@B0091`001`Lk000jT7:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7:@Mi1b
T7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@M`Ea19004W5bLG::ShZ?RXn:ShZ?RXn:Sbn@00RYT:V@
ZI2YT4T02@L000L7?000?I1aT76@LI1aT7:@LY1bT79gLY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bLF]aJg400Y002hn@Sg9>LW5bLHYa00>:03EaRW6:LHmaRW6:Mi1gT7N@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@M`08T003L@0004L02@L000L7?@00=I1bT7:@LY1b
T7:@LY1gT7N@Mi1gT7N@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@00Ia1i00679aLW5bRXn:
ShZ?RXn:ShZ?RXn:ShZ?RRR@00VYT:V@ZI2:L8X0B0091`001`Ln000jT76@LI1aT7:@LY1bMg9gLWMb
T79gLY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@075[LF]aJ`>@00V?T8n@CW9>LW400hX0=g6:
RXYaRW6:LHmaRW6:LHYaRWN@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7L01i00171a
L75700T700071cl0036@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1bT7:@LY1gT7:@Mi1gT7N@Mi1gT7N@Mi1g
T7N@00Ma2900179aLW42LP0HShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:1i001GN@T91g01J@00^YT:V@
ZG2:L8Y`RP1600T700071d0003J@LI1aT76@LY1bMg9gLY1bMg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT01aJg5[LF`5T00;Si2?T4ibLG9>LW401hX0<G6?LHmaSg6:LHmaRW6:LHYaRW6@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7L01I00271aL75`RV]aA@091`001`M0000_LY1bT7:@LY1bT7N@
Mi1gT7N@Mi1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gLF`01g49T004LW5bL@Ab01b?RXn:ShZ?RXn:
ShZ?RXn:ShZ?RXn:ShZ?RXn:0i000gN@T00FT00=ZI2YT8Y`RW6:L8Y`RP1400T700071d4003UaT76@
LI1aMg5gLY1aT7:@LY1bMg9gLWMbMg9gLY1bT7:@LY1bT7:@LW40L@1a0740L@2@Si2?T8l00i00@Hn@
Sg9aLW5bCW9>RW6?LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:Mi1gT7N@Mi1gT7N@Mi1gT7N@
Mi1g00>@00e[RW2:JhY[LG1aL75`04802@L000L7@P009h^@Ri2;T7N@LY1bT7:@LY1bT7N@Mi1gT7N@
Mi1gT7N@Mi1gT7N@M`07L@03K75a00:@00>YT9002909LP0SShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:
ShZ?RXn:ShZ?RXl04i004:V@RW2:L8Y`RW6:LHYaRW5100T700071d<003UaT76@LI1bT75gLY1bT7:@
LWMbMg9gLWMbMg:@LY1bT7:@LY00L@1a0740L@1a076?T8n@Si2?T8l00i003hmbLG9aLTibCW9>Sg6?
L@03RP0]LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW2:Mi1gT7N@Mi1gT7N@Mi1g00:@012:L8Y`
RW2:Jg5`LG1aL75`@0091`001`M4000RRi2;T8^@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT0Ya
00=/LI003I0;LP0XShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RPb@01::L8Y`
RW2:L8YaRW6:LHYaRW4o00T700071dD003EaT76@LGMaMg5gLY1aMg9gLWMbMg9gLWMbT7:@LY1bT01a
0740L@1a0740LHn@Si2?T8n@S`03T00@Si2?LW5bLG9>LTibCW9>S`::04=aRXZ:LHYaRW6?LHYaRW6:
LHYaRW6:LHYaRW6:LHYaRW2:LHY`RWN@Mi1gT7N@MhY`RW2:L8Y`RW1aL75`LG1aL75[03h02@L000L7
AP007h^@Ri1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT6`02g400fb^T00>T0ab02b?RXn:ShZ?RXn:
ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RPF@01EaRW2:LHY`RW2:LHYaRW6:LHYaRW00
?@091`001`M7000YLI1aMg5gLI1aMg:@LWMbMg9gLWMbMg:@LY1bL@1a0740L@1a0740LF/00i002Hn@
Si2?T8n@S`03T00>Si1aLW5bCW9>LTibCW88RP0oLHYaSg6?LHYaSg6:LHYaRW6:LHYaRW6:LHYaRW6:
LHYaRW6:LHYaT7N:L8Y`RW2:L8Y`RW1aL75`LG1aJg5[03`02@L000L7B0007H^@LY1bT7:@LY1bT7N@
Mi1gT7N@Mi1gT7MaK75/00]a4I0<LP18ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:
ShZ?RXn:ShZ?RXn:[g2:LHY`RW2:L8YaRW6:LHYaRW6:L8Y`>`091`001`M8000bT76@LGMaMg5gLGMb
Mg9gLWMbMg9gLY00L@1a0740L@1a0740L@1a092?T8n@Si2?T8l3T00ASi2?T4ibLG9>LTibCW9>LW40
2HX09W6?LHmaSg6:LHmaRW6:LHYaRW6:LHYaRW6:LHY`RW6:LHYaRW6:0Y005g6@L8Y`RW2:L8Y`RW1a
L75`LF^:Jg5[03X02@L000L7B@00792;T7:@LY1bT7:@Mi1gT7N@Mi1gT6aaK75/LF`:LA:@3W80B8n:
ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:Si2ZMi1gT7N@L8YaRW6:LHYa
RW6:LHY`RW2:L3T02@L000L7BP008i1aT79gLWMbMg:@LY1bT7:@LW40L@1a0740L@1a0740L@1a00j@
01j?T8n@Si2?LTibLG9aLW5bLG9aLW5bRXmaShZ?RXl3RP03ShZ:00::01YaSg6?LHmaSg6:LHYaRW6:
LHYaSg6?LHmaRP>@01agLI1aMg6@LGMaRW2:L8Y`RW2:L8Y`LG1aL75`=P091`001`M;000FT8^@LY1b
T7:@LY1bT7:@K75/LFaaK0=a00U/LFaaK75/LF`05I000g5bLP0=LP19RXn:ShZ?RXn:ShZ?RXn:ShZ?
RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ_T:mgT7N@Mi1gT7N@Mi1gRW2:L8Y`RW6:L8YaRW2:L00e00T7
00071d`0022@LGMbMg9gLI1bT7:@0740L@1a0740LF]a0740L@1a00n@02B?T8n@Si2?LTibLG9aLW5b
LG9aLW5bLG::ShZ?RXn:ShZ?RXl2RP09LHmaSg6?LHma00N:00UaRW6?LHmaSg401I00775gLI1aMg6@
LI1aT76@LHY`RW2:L8Y`LG1aL74e00T700071dd001:@LY1bT7:@LY1bLFaaK75/LF`7L@05K75/LF`0
5i000g5bLP0?LP18RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RZn@[i2_Mi1gT7N@
Mi1gT7N@MY1gT7J@MXYaRW6:LHYaRW2:=0091`001`M>0007T79gLWMaM`02L@0D0740L@1a0740L@1a
075[L@1a074@T00_Si2?T8n@Sg9>LW5bLG9aLW5bLG9aLW5bLHn:ShZ?RXn:ShZ?RXmaSg6?LHmaSg40
2HX01W6?LHmaS`F@01mgLGMaMg5gLGMaT76@LI1aT76@LHY`RW2:L75`LG2:03<02@L000L7C`003Y1b
T7:@K75/LFaaK75/0g400faaL@03L@04K75/LAV@4W80BHn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:
ShZ?RXn:ShZ?T:n@[i2_Mi1gT7N@Mi1gT7J@MY1fT7J@MY1gT7N:LHY`RW6:L8X0<P091`001`M>000K
LG9gLWL0L@1a0740L@1a0740L@1aK75[L@1a01:@03:?T8n@Si2?T4ibLG9aLW5bLG9aLW5bLG9aLW6?
RXn:ShZ?RXn:ShZ?LHn:Sg6?LHn:S`>:00R?RXZ:Sg6?L@N@021aMg5gLGMaMg6@LI1aT76@LI1aMg5g
LGMaRW2:L8Y`RS402@L000L7C@004740LH^@K75/LFaaK75/LF`9L@03K76@01Z@57809hn:ShZ?RXn:
ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:[i2_T:n@[`02T00QMi1gT7N@Mi1fT7J@MY1fT7J@Mi1gT7J@
Mi1fRW2:L8Y`02l02@L000L7C0007740L@1a0740L@1a0740L@1a0740LF]a075/L@0CT00`Si2?T8n@
Si1>LW5bLG9aLW5bLG9aLW5bLG9aLXZ?RXn:ShZ?RXn:ShZ?LHn:Sg6?0hX02Hn:RXZ?RXmaS`08T00S
LGMaMg5gLGMaMg6@LGMaT75gLGMaMg5gLGMaMg6:L8Y`RW00;P091`001`M;000BL@1a075bLFaaK75/
LFaaK75/0g400faaL@03L@03K92@01Z@5G809XZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:Si2_
T:n@[i2_0Y008gN@Mi1gT7N@LI1fT76@LI1aT7J@MY1fT7J@MY1aT76:L8Y`02d02@L000L7BP007740
L@1gLGL0L@1a0740L@1a0740L@1a0740L@0=T009Si2?T8n@Si2?00>@03J?T8n@Sg9>LTibCW9>LW5b
LG9aLW5bLG9aLXZ?RXn:ShZ?RXn:ShZ?RXn:Sg6?LHmaSg6?LHX:T00VLGMaMg5gLGMaMg5gLGMaMg5g
LGMaMg5gLGMaMg5gLGMaMg5gLGL[00T700071dT000Ya0740Mg5gLW5/2g4026aaK75/LFaa7Y000g5b
LP0@LP1<Mg9bRXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn@[i2_T:n@[i2_T:mgT7J@MY1fT7N@Mi1g
T7N@MY1aT76@LI1aT76@MY1fT7J@MRd02@L000L7BP006g40LTibCW80L@1a0740L@1a0740L@1a0740
L@0<T009Si2?T8n@Si2?00>@03N?T8n@Si2?LTibCW9>LTibLG9aLW5bLG9aLW5bLG::ShZ?RXn:ShZ?
RXn:ShZ?LHmaSg6?LHma00b@021gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLC002@L000L7
B0002aQa07MaMg5gLI1/00]a00Q/LFaaK76@UAj@5G80A7MbMhZ?RXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:
Si2_T:n@[i2_T:n@[gJ@MY1aT7J@Mi1gT7N@MY1gT76@LI1aT76@LI1f<`091`001`M80008L@1bCW9>
LTh2LP0BL@1a0740L@1a0740L@1a07403i002Hn@Si2?T8n@S`03T00eSi2?T8mbCW9>LTibCW9aLW5b
LG9aLW5bLG9aLW5bRXn:ShZ?RXn:ShZ?RXn:Sg6?RXmaSg4039006gMaMg5gLGMaMg5gLGMaMg5gLGMa
Mg5gLGMaM`0e00T700071dL000aa67MaMg5gLGMaT6`;L@08K75/LFbET9DNT003LG9b011b049gLW9b
Mg9gRXn:ShZ?RXn:ShZ?RXn:ShZ?RXn:[i2_T:n@[i2_T:n@[g6@LI1aT76@MY1gT7J@MY1fT76@LI1a
T74g00T700071dH000Ya075>LTibCW9>0W804740L@1a0740L@1a0740L@0?T009Si2?T8n@Si2?00>@
03F?T8n@Si2?LTibCW9>LTibLG9aLW5bLG9aLW5bCW9aLW5bRXn:ShZ?RXn:ShZ?RXn:Sg6?RP0>T00F
Mg5gLGMaMg5gLGMaMg5gLGMaMg5gLCX02@L000L7A@003W5/LG5gLGMaMg5gLI1/2g4026aaLG6@UI2E
7Y000g5bL@0BLP0WMg9gLWMbMhZ?RXn:ShZ?RXn:ShZ?RXn:Si2_T:n@[i2_T:n@[i2_00:@015aT76@
LI1gT7IgMY1fMgJ@L@0m00T700071d@000aa075aLTibCW9>LTh2LP0?L@1a0740L@1a0740L@1a01:@
00N?T8n@Si2?00F@03:?T8n@Sg9>LTibCW9>LW5bLG9aLW5bLG9>LW5bLG9aLW6?RXn:ShZ?RXn:ShZ?
RXn:Sa2@011aMg5gLGMaMg5gLGMaMg5g?`091`001`M4000?675aMg5gLGMaMg5gLGM/00]a00Q/LI2E
T9F@UAj@00EaLW5bL@0BLP0UMg9gLWMbShZ?RXn:ShZ?RXn:ShZ?RZn@[i2_T:n@[i2_T:n@[`02T00=
LI1aT76@MWMfMg5gMP1100T700071d8000ma079>LTibLG9aLW5bLG800W40301a0740L@1a075/LAb@
026?T8n@Si2?T8n@LG9aLW5bLG9>LW5bCW9aLW5bLG9aLW401G803g6?RXn:ShZ?RXn:ShZ?RP0AT008
LGMaMg5gLGM700T700071d40015a07MaMg5gLG9aLW5bLG9aM`02L@0:K75/LFaaK75/LBN@77809WMb
ShZ?RXn:ShZ?RXn:ZY2ZT:Z@[i2_T:n@[i2_T:n@[g5gLGMaB@091`001`M0000AL@1bCW9>LTibLG9a
LW5bLG800W402P1a0740L@1a074NT00SSi2?T8n@Si2?T75bLG9aLW5bCW9>LW5bCW9aLW5bLG9aLW40
0g803G5bLHn:ShZ?RXn:ShX059000gMa001;00T700071d4001agLG9aMg5gLG9aLW5bLG9aLVaaK75/
LFaaK75/:90NLP0OMg:?RXn:ShZ?RXn@ZY2ZT:Z@[i2_T:n@[i2_T:n@[`1>00T700071d00015bCW9>
LTibCW9aLW5bLG9aLP02L@0:0740L@1a0740LAj@02V?T8n@Si2?T8n@LG9aLW5bLG9>LTibCW9aLW5b
LG9aLW5bLG9aLW5bL@02LP07ShZ?RXn:S`0ET4h02@L000L7?`02L@0KLW5gLG9aMg5bLG9aLW5bLG:@
LFaaK75/LFaa02Z@6g808GMbMg9gRXn:ShZ?RZZ@ZY2ZT:n@[i2_T:n@[i2_T:n@[`1>00T700071ch0
01iaCW9>LTibCW9>LTibLG9aLW5bLI00L@1a0740L@0OT00SSi2?T8n@Si2?T8mbLG9aLW5bCW9>LTib
CW9aLW5bLG9aLW400g802g5bLG9aLXZ?RXn:01J@CP091`001`Lo000MLW5bLG9aLW5bLG9aLW5bLG9a
LY1aK75/LFaaK9D0:Y0MLP0PMg9gLWN:ShZ?T:Z@ZY2ZT:Z@[i2_T:n@[i2_T:n@[i1=00T700071d@0
01IbCW9>LW5bCW9aLW5bLI00L@1a07408I00<8n@Si2?T8n@Si2?LW5bLG9aLTibCW9>LW5bCW9aLW5b
LG9aLW5bLG9aLW5bLG::SaR@C@091`001`M9000CLW5bLG9aLW5bT75/LFaaK9F@U@0ZT1]b029gLWMb
Mg9gLWN@ZY2ZT:n@ZY2_T:n@[i2_T:n@[i2_T:n@C@091`001`MB0006T01a074H6I002Hn@Si2?T8n@
S`03T00YSi2?T8n@Si1fT4ibCW9>LW5bCW9aLW5bLG9>LTibCW9>LW5bLG9aLW400W8KT4d02@L000L7
DP000i1aL@02L@J@00VET9F@UI2ET9D08I00179aLW4BLP03Mg9g00Ib01j@ZY2_T:n@[i2_T:n@[i2_
T:n@[i2_T:n@[i2_T:m<00T700071e8000F@074HL@0LT007Si2?T8n@S`05T00TSi2?T8n@MY1>LTib
CW9>LW5bLG9aLW5bCW9>LTibLG9>LW5b8I1<00T700071e8000B@LG5a1i002IF@UI2ET9F@U@0QT006
LW5bLG9a4W809GMbZY2ZT:Z@[i2_T:n@[i2_T:n@[i2_T:n@[i2_T:n@[i2_T:l0C0091`001`MB0003
T1Qa01b@00V?T8n@Si2?T8l00i008Xn@Si2?T8n@MY1>LTibCW9>LTibLG9aLW5bCW9>LTibCW8UT4`0
2@L000L7DP00191aT9D7T009UI2ET9F@UI2E026@00AbLG9a4W809jn@[i2_T:Z@[i2_T:n@[i2_T:n@
[i2_T:n@[i2_T:n@[i2_T:n@[`1<00T700071e8000>@69007Y001hn@Si2?T8l01I006hn@Si2?T7J@
MW9>LTibCW9aLW5bLG9aLTibCP0ZT4`02@L000L7DP000i2ET008T009UI2ET9F@UI2E02:@00EaLW5b
L@0:LP0^[i2_T:n@[i2_T96@[i2ZT:n@[i2_T:n@[i2_T:n@[i2_T:n@[i2_T:n@[i2_T4/02@L000L7
DP00191bT78KT009Si2?T8n@Si1f00>@01V?T8n@Si2?T7J@MW9>LTibCW9>LW5bCW9a02n@B`091`00
1`MB00f@00NET9F@UI2E02:@00=aLW402780<Y6@TI2_T:n@[i2_T:n@[i2FT:n@[i2_T:n@[i2_T:n@
[i2_T:n@[i2_T:n@[i2_T:n@B`091`001`MB000FT7:@LY1bT7:@LY1bT7:@LY1bT7:@LP^@00N?T8n@
Si2?00F@01:?T8n@Si1fT7IbCW9>LTibLG8dT4/02@L000L7DP0KT003TI2@01V@1780>96@TI2AT96@
UY2AT96@TI2AT96@[i2_T:n@[i2_T96@UY2FT:n@UY2_T:n@[i2_T:n@[i2_T:n@B`091`001`MB000V
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1gT783T00?LY1bT7:@LY1bT8n@Si1>00B@
00VAT96@TI2AT9403I001I6@T92A01f@BP091`001`MB01V@00FAT92@T@0;T01;TI2AT96@TI2AT96@
TG6@Si6@TI2AT96@UY2AT96@TI2AT96@TI2_T:n@[i2_T:n@UY2FT9J@UY2_T:n@[i2_T:n@[i2_T:n@
[i2_04X02@L000L7D`00<G:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7N@LY1bT7:@
LY1bT7800Y001W4HLGJ@L@:@00VAT96@TI2AT9402i001I6@T92A01n@BP091`001`MC01Z@00>AT900
2i00596@TI2AT96@TDibLG9aLXn@Si1f0Y00=I6@TI2FT96@TI2AT96@TI2FT:n@[i2_T:n@UY2FT9J@
UY2FT:n@[i2_T:n@[i2_T:n@[i2_04X02@L000L7D`009G:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@
LY1bT7:@LY1gT7800i005g:@LY1bT7:@674HLAQa691fT76@LI1a00:@00>AT9402i000i6@T@03T003
TI2@01^@BP091`001`MC01B@00VAT92@TI2AT9402i00C96@TI2AT79>LW5bLG9aLXn@Si1fT7J@MY1f
T7JAT96@TI2AT96@[i2_T:n@[i2_T9J@UY2FT9J@UY2_T:n@[i2_T:n@[i2_T:n@[i1900T700071e<0
04EbT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1gT7:@LY1bT1Qa674HLAQa674HT76@
LI1aT76@LI1aT7402Y001I6@T92A022@B@091`001`MC01J@00NAT92@TI2A00^@022AT95>LTibCW9>
LW5bCW9aT8n@MY1fT7J@MY1fT7J@MP:@02ZAT96@UY2_T9J@[i2_T9J@UY2FT9J@UY2_T:n@[i2_T:n@
[i2_T:n@[i1900T700071e<002EbT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@Mi1b00>@
00mb01Qa674HLAQa674HLAP00W404Y1aT76@LI1aT76@LI1aT76@L@B@00NAT96@TI2A01j@B@091`00
1`MC00j@00FAT92@T@0;T008TI2AT96@TI05009a02=>LG5aCW9>LTibCY1fT8n@Si1fT7J@MY1fT7J@
MY1fT7J@MP02000RUY2FT9J@[i2_T:n@[i2_T:n@[i2_T:n@[i2_T:n@[i2_T4T02@L000L7D`0097:@
LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT0P002<HLAQa674HLAQa674HT76@LI1aT76@
LI1aT76@LI1aT76@L@06022@B0091`001`MC00b@00NAT96@TI2A00^@00BAT96@2`04L@0NCW9>LTib
CW9fT8n@MY2?T7J@MY1fT7J@MY1fT7J@20007Y2_T:n@[i2_T:n@UY2_T:n@[i2_T:n@[i2_T:n@[dP0
2@L000L7D`0087:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@3@008G4HLAQa674HLAQa691a
T76@LI1aT76@LI1aT76@LI1aT00=01Z@B0091`001`MC00j@00FAT96@T@0;T1000g407DibCW9>LTib
CY1fT7J@MY1fT7J@MY1fT7J@MY1f010001N_T:n@[i2FT:n@UY2_T:n@[i2_T:n@[`1800T700071e<0
01ebT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LP0B000:LAQa674HLAQa609a01>@LI1aT76@LI1a
T76@LI1aT76@01<0591800T700071e<039001i6@TI2AT9402I0C000OCW5>LTibCW9>LTj@Si1fT7J@
MY1fT7J@MY1fT7J@L@0E000BT:n@UY2_T:n@[i2_T:n@[i2_B0091`001`MC000JLY1bT7:@LY1bT7:@
LY1bT7:@LY1bT7:@LY0F0009674HLAQa674H009a01:@LI1aT76@LI1aT76@LI1aT74J00j@B0091`00
1`MC00Z@00VAT92@TI2AT9401I0I000LCW9>LTibCW9>T7J@MY1fT7J@MY1fT7J@MY1aT1d000^_T:n@
[i2_T:n@[`1800T700071e<001IbT7:@LY1bT7:@LY1bT7:@LY1bT7:@6`00274HLAQa674H0W404I1a
T76@LI1aT76@LI1aT76@02002I1800T700071e<04i0O000ILDiaCW5>LW6@MY1fT7J@MY1fT7J@MY1f
T00W0003UY2F04P02@L000L7D`004G:@LY1bT7:@LY1bT7:@LY1b024001THLAQa674HLAR@LI1aT76@
LI1aT76@LI1a02X000>@0000A@091`001`MC0003T7N@00:@00=gT9001i0T000GCW5>LDiaCW9fT7J@
MY1fT7J@MY1fT7H0L`091`001`MC000=LY1bT7:@LY1bT7:@LP0W000F674HLAQa691aT76@LI1aT76@
LI1aT7<02@L000L7D`03T003Mi2@00J@:0005G5>LDiaCW9fT7J@MY1fT7J@MY1fT01d00T700071e<0
00YbT7:@LY1bT7:@:`001W4HLAQa609a00b@LI1aT76@LI1aT75d00T700071e<000>@Mi000Y0017N@
T90]000BLDiaCW9>T7J@MY1fT76@MY1fM@091`001`MC0007LY1bT7:@LP0_0005674HLAP00W402i1a
T76@LI1aT76@07D02@L000L7D`03T003Mi000340011>LDibCY1fT7J@MY1fT7J@MP091`001`MC0003
LY1b03D000lHLAQa691aT76@LI1aT740MP091`001`N<000=CW5aLWJ@MY1fT7J@MP1g00T700071hd0
00/HLAR@LI1aT76@L@1h00T700071hd000]aCW5aT7J@MY1fT01h00T700071hd000<HLAP00W401I1a
T76?07T02@L000L7SP002G5bLI1fT7J@MP1i00T700071hh000Ua675aT76@LI00N@091`001`N?009a
00IbMY1fT7Ii00T700071i0000Ia691aT75j00T700071i000W4017J@MY1j00T700071i4000@HLG6?
N`091`001`NC0003T00007X02@L000L7T`000g40001j00T700071ol04@091`001`Oo01402@L000L7
o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7
o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7
o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7
o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7
o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7
o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`001`Oo01402@L000L7
o`0A00T700071ol04@091`001`Oo01402@L000L7o`0A00T700071ol04@091`00o`LQ1`00o`LQ1`00
o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00\
\>"],
  ImageRangeCache->{{{0, 287}, {285.875, 0}} -> {-8.12139, -7.50157, 1.10189, 
  1.10189}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "Surface graphics can be sent to the OpenGL viewer as well. We use a \
surface graphics to show the third possible value for ",
  StyleBox["MVPolygonShading:",
    FontFamily->"Courier"]
}], "Text",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(\(surf = 
      Plot3D[Sin[x*y], {x, 0, Pi}, {y, 0, Pi}, DisplayFunction -> Identity]; 
    \)\)], "Input",
  CellLabel->"In[10]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; 
    MVShow3D[surf, MVPolygonShading -> MVWireFrame, MVGrayBackground -> True]
      \)], "Input",
  CellLabel->"In[11]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVMesh3D[2], 
      PlotRange 
        \[Rule] {{0.`, 3.14159265358979311`}, {0.`, 3.14159265358979311`}, {
            \(-0.992603329887604068`\), 0.999998994939753238`}}, 
      ViewPoint 
        \[Rule] {1.30000000000000004`, \(-2.39999999999999991`\), 2.`}, 
      AmbientLight \[Rule] {0.`, 0.`, 0.`}}\)], "Output",
  CellLabel->"Out[11]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[17]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFB090B0BF90B0C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFBF90B0BF9FBFBF90B0CFCFCFCFCFCFBF9FBFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0BF90B0BF90B0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFBF90B0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
BF90B0BF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0B090B0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFBF90B0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFBF90B0BF9FBFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFBF90B0BF90B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0BF90B0
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFBF9FB0BF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0BF90B0BF90B0BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0
BF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
BF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFB090B0AF8FB0CFCFCF
CFCFCFBF9FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
BF90B0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0B090B0CFCFCFCFCFCFCFCFCF
CFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
BF9FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFAF8FB0B090B0AF8FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
B090B0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFB090B0B090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFAF8FB0\
B090B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFBF9FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAF8FB0B090B0B090B0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B090B0AF8FB0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFBF90B0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0B090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0AF8FB0AF8FB0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0BF90B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
AF8FB0A08FB0AF90B0CFCFCFCFCFCFAF8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFA08FAFAF8FB0
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B090B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFBF90B0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAF90B0A08FB0CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFA08FAFAF8FB0CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFB090B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0BF90B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0BF90B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFA08FB0AF8FB0A08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFBF9FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
AF8FB0A08FAFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAF8FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFB090B0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFB090B0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0AF90B0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF9F80AFA08FB0A08FAFCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFAF8FB0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFBF90B0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFBF90B0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB09F8FAFCFCFCFCFCFCFA08FB0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFBF9FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF9F80AF9F8FB0C0C0C0CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B090B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFA08FB09F8FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0A08FB0
9F80AFA08FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF9F80AFCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0B090B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFBF9FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFBF9FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F80AFA08FB0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0A08FB09F80AFCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0A08FB0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAF8FB0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFBF90B0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFBF90B0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF9F8FB0A08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF9080AF9F80AF9F80AFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0A08FB0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFAF8FB0CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FB0\
9080AFCFCFCF9F8FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9080AF9F80B0C0C0C0\
CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F80AF\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B090B0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
2F2F2F90A090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9080AF9F8FB09080AFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F9F90
3030308F9F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF9F80B09080AFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF9F80AF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFB090B0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0BF90B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0BF90B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF8F9F90CFCFCF
2F2F2FCFCFCF8F9F90CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF9080AF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAF8FB0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FB0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F9F90CFCFCFCFCFCF
CFCFCF302F2F90A09FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C09F80AF9080AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F80AF\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B090B0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFB090B0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFB090B0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8F9F90CFCFCFC0C0C0CFCFCF
C0C0C0302F2FC0C0C090A090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FB09080AF
9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F9F90CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF302F2FCFCFCF8F9F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF8F80AF9080AF9080AFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFA08FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
A08FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFAF8FB0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFBF90B0CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFBF90B0CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF8F9F90CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF302F2FC0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9080B08F80AFCFCFCFCFCFCF9F8FB0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF9F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
9F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FB0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF8F9F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
908FA0302F2F908FA0908FA0908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF8F7FAF9080AFC0C0C0CFCFCFC0C0C0CFCFCF9080AFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C09F8FB0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
9F8FB0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAF8FB0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF8F9F90CFCFCFC0C0C0CFCFCFC0C0C0908FA0908FA0908FA0
C0C0C0302F2FC0C0C0CFCFCFC0C0C0908FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
9080AF9080B08F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF8F9F90CFCFCFCFCFCF908FA0908FA0908FAFCFCFCFCFCFCFCFCFCF
CFCFCF302F2FCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C09080AF8F80AF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C09F80AF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF9F80AFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF9F80AFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0B090B0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFBF9FB0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFBF9FB0C0C0C0CFCFCFCFCFCF\
CFCFCF8F9F90908FA0908FA0908FA0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF302F2FC0C0C0CFCFCFCFCFCF908FA0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AF9080B0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9080AF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCF\
8F9F90908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF302F2FCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8F80AF9080AF8F7FAFCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C09F80B0
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF9F80AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B090B0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFBF90B0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFBF90B0CFCFCF8F9F90\
908FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0302F2FC0C0C0CFCFCFC0C0C0CFCFCF908FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF8F80AF8F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9080AF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBF8F9F90CFCFCF\
CFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF302F2FCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF807FAF8F80AFCFCFCF8F80AFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
9080AFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0A08FB0CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0A08FB0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFAF8FB0\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFBF90B0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF302F2FBF90B0506F50CFCFCF\
CFCFCF908FA0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF302F2FC0C0C0CFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
8F80AF807FAFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF9F8FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F2FCFCFCF5F6F5FBF90B0CFCFCF\
CFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
40303F402F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C08F80AF807FAF8F80AF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807FAFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
9080AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0A08FB0C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0A08FB0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAF8FB0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0BF90B0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F2F2F302F2FC0C0C0506F50C0C0C0BF90B0C0C0C0\
CFCFCFC0C0C0908FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C040303F
C0C0C040303FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0908FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7FAF8F80AFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF9F8FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB090B0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F2FCFCFCFCFCFCF5F705FCFCFCFCFCFCFBF90B0CFCFCF\
CFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3F
CFCFCF402F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF8F80AF807FAFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF807FAFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
9080AFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0A08FB0CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0A08FB0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
B090B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0BF90B0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCF506F50CFCFCFC0C0C0BF90B0CFCFCF\
CFCFCFC0C0C0CFCFCF908FA0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF402F3FCFCFCF
CFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF908FA0CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8F80AF8F80AF8F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF9080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
AF8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF90B0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF302F2F2F2F2FCFCFCFCFCFCFCFCFCF506F50CFCFCFCFCFCFCFCFCF302F30BF9FB0\
CFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCFCFCFCF
CFCFCFCFCFCF4F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FA0\
807FAFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C08F80AFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C09F80AFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
B090B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFB090B0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C02F2F2FC0C0C0CFCFCFC0C0C0CFCFCF506F50CFCFCFC0C0C0CFCFCFC0C0C03F2F30B090B0\
CFCFCFC0C0C0CFCFCF8F8FA0CFCFCFC0C0C0CFCFCF402F3FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF402F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8F8FA0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AF807FAFCFCFCF\
807FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF9080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
B090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FBF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
303030CFCFCFCFCFCFCFCFCFCFCFCF506F50CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F30BF9FBF\
CFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCF4F3040CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF40303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807FAF8080AFC0C0C0CFCFCFCFCFCF\
CFCFCF807FA0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C08F80AFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF9F80B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFAF8FB0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFBF90B0\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C02F2F2F
90A0AFCFCFCFC0C0C0CFCFCFCFCFCF506F50C0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F2F30BF90B0\
CFCFCFC0C0C0CFCFCFCFCFCF908FA0C0C0C040303FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF908FA0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FAF8080AF807FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F80AFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF9F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFBF9FB0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030302F2F2F
CFCFCF90A0AFCFCFCFCFCFCF5F6F50CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F30CFCFCF\
BF90B0CFCFCFCFCFCFCFCFCFCFCFCF4F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF4F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C08080AF7F7FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF807FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807FA0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF9080AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C09F8FB0C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C09F8FB0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFAF8FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
BF90B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C02F2F2FC0C0C0302F2F
C0C0C0CFCFCF90A0AF506F50C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAFC0E0B0CFE0AFC0E0\
BF90B0C0C0C0CFCFCFC0C0C040303F908FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF402F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0908FA0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF807FAF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
BF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303030CFCFCFCFCFCF2F2F2F
CFCFCFCFCFCF5F6F5F90A0AFCFCFCFCFCFCFB0CFE0AFCFE0B0CFEFB0CFE0CFCFCF302F30CFCFCF\
B0CFE0CFCFCFCFCFCFCFCFCF402F3F908FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF40303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C08080AF
807FAFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C08080AFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807FAFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF9080AFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFA08FB0C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFA08FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFAF8FB0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
BF9FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF2F2F2FCFCFCF2F2F2FCFCFCF
CFCFCFCFCFCFAFC0E0B0CFE0AFC0E0B0CFE0909FAF9FA0AF90A0AF9FA0AF909FAF3F2F3090A0AF\
9FA0AFAFC0E0CFCFCF402F3FCFCFCFC0C0C0908FA0CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80AF7F7FA08080AFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF807FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F80AFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF9F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
BF90B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F2FCFCFCFAFC0E0B0CFE0B0CFE0
B0CFE0AFC0E09090AFCFCFCF8F90AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F30CFCFCF\
CFCFCFB0CFE0402F3F9FA0AF90A0AF9FA0B0807F908F7F9FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FA07F7FAFC0C0C0CFCFCF7F7FA0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C08080AFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C08F80AFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C09F80B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F80AFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F80AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0B090B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFBF90B0CFCFCFC0C0C0B0CFE0AFC0E0B0CFE0AFC0E0CFCFCF2F2F2F8F90AF
506F50CFCFCFC0C0C0CFCFCF8F8FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F2F30C0C0C0\
CFCFCFAFC0E0CFCFCFC0C0C0CFCFCFC0C0C0807F9FC0C0C0807F90807090807F9F
C0C0C0CFCFCFC0C0C040303FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0908FA0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF8080AF7F7FAFCFCFCFCFCFCFCFCFCFCFCFCF7F80AFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF807FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF9080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
AFBFE0B090B0AFCFE0B0CFEF2F2F2FCFCFCFCFCFCFCFCFCF8F90A0302F2F506F50
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F90A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F30CFCFCF\
CFCFCFB0CFE0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCFCFCFCF
8F7F9F807F90CFCFCF402F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
7F7FAF7F7FA07F7FAFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C07F7FAF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF807FAFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF8F80AFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C09F80AFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF9F80AFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF9F80AFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0B090B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFAFBFDFB0BFE0AFBFDFAFBFE0AFBFDF\
CFCFCFAFC0E0CFCFCF2F2F2FCFCFCFC0C0C08F90AF8F90A0CFCFCF2F2F2F506F50
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF8F90AFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F2F30CFCFCF\
40303FC0C0C0B0CFE0CFCFCFCFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF80709040303FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0908FA0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7FAF7F80AF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7FAF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FAFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF9080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFB0BFE0AFBFE0B0BFE0AFBFE0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFB0CFE02F2F2FCFCFCF8F90A09090AFCFCFCFCFCFCF2F2F2F5F6F5FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F90A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F3040303F\
CFCFCFCFCFCFB0CFE0CFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF402F3F8F7F9F8070908F7F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF908FA0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07F7FAF7F7FA0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07F7FAF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF7F7FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807FAFCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF9080AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F80AFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0A08FB0C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFAF8FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
AFB0DFAFBFE0AFBFDFAFBFE0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0B0CFE08F8FA0CFCFCFC0C0C0CFCFCFC0C0C02F2F2FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8F90A0CFCFCFC0C0C0CFCFCFC0C0C040303FC0C0C0\
CFCFCFC0C0C0B0CFE0C0C0C0CFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C040303FC0C0C0CFCFCFC0C0C0807F9F807090CFCFCFC0C0C0CFCFCF8F8FA0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7FA08080AFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFB090B0CFCFCFCFCFCFAFBFE0B0BFE0AFBFE0B0BFE0AFBFE0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF302F2FAFCFE0CFCFCFCFCFCFCFCFCFCFCFCF303030302F30CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F90AFCFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCF\
CFCFCFCFCFCFCFCFCFB0CFEFCFCFCFCFCFCFCFCFCFCFCFCF807F90CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F908F7F9FCFCFCF908FAF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF707FA07F7FAF707FA07F7FAFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
7F7FA0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF8080AFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF807FAFCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF9080AFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0A08FB0
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0A08FB0CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0AF8FB0A0AFD0B0BFE0AFBFDFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
302F2FC0C0C0B0CFE0CFCFCFCFCFCFC0C0C0CFCFCF2F2F2F3F2F30C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF8F90A0CFCFCFCFCFCFCFCFCF302F3040303FCFCFCF\
CFCFCFC0C0C0CFCFCFAFC0E0CFCFCFC0C0C0CFCFCFCFCFCF807F9FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C040303FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0807F9F807090\
908FA0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF7F7FAF707FA0CFCFCFCFCFCFCFCFCFCFCFCF7F80AFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
7F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF807FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAF90B0A08FB0AF8FB0\
A08FB0AF90B0CFCFCFCFCFCFAFBFE0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F2F\
CFCFCFCFCFCFCFCFCFB0CFE0CFCFCFCFCFCF2F2F2FCFCFCF302F30CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F90A0CFCFCF302F30CFCFCF402F3FCFCFCF\
CFCFCFCFCFCFCFCFCFB0CFE0CFCFCFCFCFCFCFCFCFCFCFCF807090CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF40303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
807090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
707FA0707FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FA0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
7F7FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0807FAFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807FAFCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF9080AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0A08FB0
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0A08FB0C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFA08FAFAF8FB0A08FAFAF8FB0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0AFBFE0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0302F2FC0C0C0\
CFCFCFC0C0C0CFCFCFAFC0E08F90AF2F2F2FCFCFCFC0C0C0302F30C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C08F90AFC0C0C0302F30C0C0C040303FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0B0CFE0C0C0C0CFCFCFC0C0C0CFCFCF807090CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF402F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
807F90C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7FAF707FA0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF7F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF807FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7FAFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
A08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA08FB0CFCFCFCFCFCFCFCFCF
CFCFCFA08FB0AF90B0A08FB0AF90B0A08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFAFBFE0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F2F2F2F2FCFCFCF\
CFCFCFCFCFCFCFCFCFB0CFEF8F90A0302F2FCFCFCFCFCFCF302F30CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F90A03F303FCFCFCFCFCFCF402F3FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFB0CFE0CFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF4F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
807F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0707FA0707FA07F7FAFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF7F7FAFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF7F7FAFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C08F80AFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF9F80B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
9F80AFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF9F80AFAF8FB0A08FAFAF8FB0
A08FB0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFB0BFE0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF2F2F2F2F2F2FCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFAFC0E0302F2F8F90A0CFCFCFCFCFCF3F2F30C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F2F308F90A0CFCFCFC0C0C040303FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFB0CFE0C0C0C0CFCFCFCFCFCFCFCFCF807090CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF402F3FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF807090CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF707FA07F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF7F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
A08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFA08FB09F80AFA08FB09F8FAFAF8FB0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0BFE0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F2F4F5F40303030CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF302F2FAFC0E09090AFCFCFCFCFCFCF302F30CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3030CFCFCF8F90AFCFCFCFCFCFCF402F3FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFAFC0E0CFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF40303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8F7F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0707FA06F70A06F7FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07F7FAFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF7F7FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C08F80AFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C09F80AFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
9F80AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F80AF\
9F8FB09F80AFA08FB0C0C0C0CFCFCFC0C0C0CFCFCFA0AFD0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAFBFDFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C02F2F2FC0C0C04F5F402F2F2FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF2F2F2FB0CFE0C0C0C08F90AFC0C0C0302F30C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF302F30CFCFCF8F8FA0CFCFCFC0C0C040303FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAFC0E0CFCFCFC0C0C0CFCFCFC0C0C0807F9F
C0C0C0CFCFCFC0C0C0CFCFCF402F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF807090CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FAF\
6F7FA0707FA0CFCFCFCFCFCF6F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80AFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF7F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
A08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FAFA08FB09F8FB0A08FB09F8FAFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0AFDFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0BFE0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F2FCFCFCF4F604FCFCFCF303030CFCFCFCFCFCF\
CFCFCFCFCFCF2F2F2FCFCFCFB0CFE0CFCFCF8F90A0707090302F30CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCFCFCFCF8F90A0CFCFCF402F3FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0CFE0CFCFCFCFCFCFCFCFCFCFCFCF807090
CFCFCFCFCFCFCFCFCFCFCFCF4F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8F7F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F70A0707FA0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F70A0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF7F7FA0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF7F7FA0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C08080AFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807FAF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF9080AFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFA08FB0C0C0C0A08FB09F80AFA08FB09F80AFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFAFAFDFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFAFBFDFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF2F2F2F4F5F40CFCFCFCFCFCF2F2F2FCFCFCFCFCFCF\
CFCFCFC0C0C02F2F2F7F8F807F9080AFC0E0CFCFCF706F8F3F2F30706F8F706F8F
CFCFCFCFCFCFC0C0C0CFCFCF302F30CFCFCFC0C0C0CFCFCFCFCFCF8F90AFC0C0C040303FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFAFC0E0CFCFCFCFCFCFCFCFCFC0C0C0807F9F
CFCFCFCFCFCFC0C0C0CFCFCF402F3FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0807F9FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA06F7FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090CF9090C0
9090CF9080AFA08FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0AFD0CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFBFE0CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F2FCFCFCF4F5F40CFCFCFCFCFCF3030307F8F807F9080\
7F8F80303030CFCFCFCFCFCFCFCFCFB0CFE0CFCFCF3F3030CFCFCFCFCFCFCFCFCF
707090706F8F707090CFCFCF3F3030CFCFCFCFCFCFCFCFCFCFCFCF8F90A0CFCFCF402F3FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0CFE0CFCFCFCFCFCFCFCFCFCFCFCF
8F7F9FCFCFCFCFCFCFCFCFCF4F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF807F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F7FA0707FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0707FA0C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07F7FAFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C07F7FAFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C08080AFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807FA0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9080AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C09090C09090C09090C08F90C09090C0C0C0C0CFCFCF
C0C0C09F9FCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0A0AFDFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0AFBFE0C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF2F2F2FCFCFCF4F5F40CFCFCF708F802F2F30708F80CFCFCFC0C0C0\
2F2F2FC0C0C0CFCFCFC0C0C0CFCFCFAFC0E0CFCFCF302F30CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C03F2F306F6F8F706F8FC0C0C0CFCFCFC0C0C0CFCFCF8F90A040303FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B0CFE0C0C0C0CFCFCFC0C0C0CFCFCF
807090CFCFCFC0C0C04F30404F2F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0807F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF60709F707FAF6F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FA0CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7FA0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF7F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
8F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F8FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
9090C09090CF9090C09090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF9F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFAFDFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFBFE0CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF2F2F2F506F4F4F5F4050604F7F8F80CFCFCF2F2F2FCFCFCFCFCFCFCFCFCF\
2F2F2FCFCFCFCFCFCFCFCFCFCFCFCFB0CFE0CFCFCF3F303F706F8FCFCFCFCFCFCF
CFCFCFCFCFCF3F3030CFCFCFCFCFCFCFCFCF707090706F8F707090CFCFCF8F90AF402F3FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFCFE0CFCFCFCFCFCFCFCFCFCFCFCF
8F7F9FCFCFCF4F3040CFCFCFCFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF6F7FA060709F707FA0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F7FA0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C07F7FAFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C07F7FAFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807FAFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
8F80AFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C09F80AF9090C09090C08F90C09090C09090C0\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFA0AFD0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFAFBFDFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF2F2F2F50604FCFCFCF4F6040C0C0C0708F80CFCFCF302F30C0C0C0CFCFCF2F2F2F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0B0CFE0302F3070708FC0C0C0CFCFCF
CFCFCF3F2F30C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF70708F8F90A040303FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFAFC0E0CFCFCFC0C0C0CFCFCF
8070904F3040C0C0C0CFCFCFCFCFCF4F3040C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60709F
6F7FA0CFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80AF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF7F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
807FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF8F8FBF8F8FBF808FBF8F8FC09090C0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFA0A0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFAFDFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0BFE0CFCFCFCFCFCFCFCFCF
CFCFCF2F2F2F302F30CFCFCF4F6040CFCFCFCFCFCFCFCFCF7F8F802F2F2FCFCFCF2F2F2FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0CFE03F3030706F8FCFCFCFCFCFCF
CFCFCF302F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3F7F7090\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0CFE0CFCFCFCFCFCFCFCFCF
4F3040807090CFCFCFCFCFCFCFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0607FA060709FCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF6F70A0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F70A0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FA0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF7F7FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FA0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
8F80AFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C08F8FBF
8080BF8F8FBF8080BFCFCFCFC0C0C0CFCFCFC0C0C09090C0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF9F9FCFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFA0AFD0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAFB0DFCFCFCFC0C0C0CFCFCF
C0C0C02F2F2FC0C0C02F2F2F405F40CFCFCFC0C0C0CFCFCF70807F2F2F30C0C0C0302F2FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B0CFE0302F30CFCFCF706F8FCFCFCF
302F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C040303FC0C0C0\
7F7090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFAFC0E0CFCFCFC0C0C04F3040
C0C0C0807F90C0C0C0CFCFCFC0C0C04F303FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FA060709F6F7FA0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FA0CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
8F80AFCFCFCFCFCFCFCFCFCFCFCFCF8F8FBF808FBF8F8FC08F8FBF8F8FBFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090C0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFA0A0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0AFD0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0BFE0CFCFCFCFCFCFCFCFCF
3030302F2F2F2F2F2F4F5F40CFCFCFCFCFCFCFCFCFCFCFCF303030708F80302F2FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0CFEFCFCFCF707090302F30
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCF\
CFCFCF7F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0CFE04F3040CFCFCF
CFCFCF807F90CFCFCFCFCFCFCFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF807F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF60709F6F7FA0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F70A0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF707FA0C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
7F7FAFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF7F7FA0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF8080AFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
8F80AF807FAF8F8FBF808FBF8F8FBFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF9090C0CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C09F9FD0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0A0AFDF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFB0BFE0C0C0C02F2F2F
CFCFCF302F2F405F40CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF2F2F2F302F2FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFAFC0E0CFCFCF302F30CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C040303FCFCFCF\
CFCFCFC0C0C07F7090706F8FCFCFCFC0C0C0CFCFCFCFCFCFB0CFE0C0C0C0CFCFCF
CFCFCFCFCFCF807090CFCFCFCFCFCF4F3040C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF807F9FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF60709F60709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
7F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FAFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AF807FAF8F80AF8F80AF8F80AF\
CFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090CFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0AFDF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFBFE0CFCFCF2F2F2F
302F2F2F2F2F4F604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F302F2F2F7F8F80CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0CFE0CFCFCF3F3030706F8F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3F7F7090\
CFCFCFCFCFCFCFCFCFCFCFCF706F8FCFCFCFCFCFCF4F3040AFC0E0CFCFCFCFCFCF
CFCFCFCFCFCF8F7F9FCFCFCFCFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF807090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F709F5F7090\
5F709FC0C0C05F709FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0707FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F7FA0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
7F7FAFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C07F7FAFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807FAFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C08F80AF807FA08F80AF807FAF8F80AFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF7F7FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8F90C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C09FA0D0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
A0AFD0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0AFBFE02F2F2FCFCFCF
2F2F2F4F5F40C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F2F2FCFCFCF70807FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF302F30B0CFE0C0C0C0706F8F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C040303F706F8F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF706F8F4F3040C0C0C0B0CFE0C0C0C0CFCFCF
C0C0C0CFCFCF807090CFCFCFC0C0C04F303FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F709F60709FCFCFCFCFCFCF\
CFCFCFCFCFCF5F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF7F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F80AF8F7FAF
8F80AF8F80AF8F80AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8F8FBF7F80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090CFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFA0A0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
AFB0DFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFB0BFE02F2F2F
4F604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F2F3F303FCFCFCFCFCFCF708F80CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F303FB0CFE0CFCFCF706F8F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3F7F7090\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F3040706F907F7090CFCFCFB0CFE0CFCFCF
CFCFCFCFCFCF8F7F9FCFCFCFCFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F709F5F7090CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF5F7090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F7FA0CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0707FA0
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF7F7FA0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF7F7FA0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF807FA08080AF807FAF8080AF7F7FB0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C08080B0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF9090C0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF9F9FCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
A0AFD0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF302F2FAFBFDF2F2F2F
4F5F40CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F2F30CFCFCFCFCFCFC0C0C0708F80CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF402F3FB0CFE0C0C0C0CFCFCF
706F8FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C040303FCFCFCF\
7F7090C0C0C0CFCFCFCFCFCF4F3040C0C0C0CFCFCFCFCFCF7F7090AFC0E0CFCFCF
CFCFCFCFCFCFC0C0C0807F9FCFCFCF4F3040C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF807090CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F709060709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF5F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FA0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
807FAF8080AF807FAF8080AF807FAFCFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090C0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF9FA0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
AFAFDFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F2F2F2F2FB0BFE04F5F40
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303030302F307F8F80708F807F8F80708F80808FA0\
7F809FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F3040303FAFC0E0CFCFCFCFCFCF
707090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCF\
706F8FCFCFCFCFCFCFCFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFB0CFE0CFCFCF
CFCFCFCFCFCFCFCFCF807F90CFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCF5F70905F709F5F7090CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C05F709FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F7FA0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
6F70A0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C07F7FAFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FA0CFCFCF7F7FA0807FAF807FA08080AF7F7FA0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FB0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C08080B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C09090C0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF9F9FCFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0A0AFDFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F2F2FCFCFCFAFBFDFCFCFCF
C0C0C0CFCFCF70807F708F8070807F2F2F3070807F3F2F3FC0C0C0CFCFCFC0C0C0CFCFCF7F8090\
CFCFCF7F809F808FA0C0C0C0CFCFCF302F30CFCFCF3F2F3FCFCFCFAFC0E0CFCFCF
6F6F8FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C040303FC0C0C0\
7F7090C0C0C0CFCFCF4F2F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0B0CFE0
706F8FCFCFCFC0C0C0807F9FC0C0C04F303FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807F9FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
507090506F8F507090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
707FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF7F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF6F70A0707FAF6F70A06F70AF807FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80B0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF8080BFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090CFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFA0AFDFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F2F5F6F5050604F302F2FAFBFE0
7F8F80708F80CFCFCFCFCFCFCFCFCF2F2F2FCFCFCF302F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
7F809FCFCFCFCFCFCF808FA0302F30CFCFCF402F3FCFCFCFCFCFCFB0CFE0CFCFCF
707090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCF\
7F6F90CFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFCFE0
CFCFCF7F6F90CFCFCF807090CFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807090CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
506F8FCFCFCF506F8FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F709FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F70A0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF707FA0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C07F7FAFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F70A06F70A0
6F70A06F70A06F70A0CFCFCFCFCFCFCFCFCFC0C0C06F70A0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C08080B0
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF7F7FB0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF8F90C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF9FA0D0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFAFAFDFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF4F5F3F2F2F2FCFCFCF302F302F2F2FAFBFE0
6F7F6FCFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C03F2F30CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
7F809FC0C0C0CFCFCFCFCFCF3F303F7F809F40303FCFCFCFCFCFCFAFC0E0CFCFCF
CFCFCF70708FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C040303FCFCFCF\
CFCFCF4F2F3FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0B0CFE0
CFCFCFCFCFCF706F8FCFCFCF8070904F3040C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0807F9FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
50708FCFCFCFCFCFCF506F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F709FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF6F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF7F80AF6F70A06F7FAF6F70A06F70A06F70A0CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F70A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80B0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090CFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFA0AFD0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCF2F2F2F303030AFBFE0
6F8070CFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCF302F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
7F809FCFCFCFCFCFCF3F303FCFCFCFCFCFCF402F3F808FA0CFCFCFCFCFCFB0CFE0
CFCFCF706F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCF\
4F303F7F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
B0CFE0CFCFCFCFCFCF706F908F7F9FCFCFCF4F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
4F6F8FCFCFCFC0C0C0CFCFCF4F6F8FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F7090CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0707FA0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C06F7FA0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF707FA0\
707FA0707FA0707FAF707FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F70A07070AFCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
7F7FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF7F7FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9090C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C09F9FCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFA0AFD0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C02F2F2FC0C0C0CFCFCF2F2F2F302F30C0C0C0AFBFE0
C0C0C06F7F70C0C0C0302F30C0C0C0CFCFCFC0C0C03F2F3FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
7F809FC0C0C03F2F3FC0C0C0CFCFCFC0C0C040303FC0C0C0808F9F7F809FB0CFE0
C0C0C0706F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C040303F4F2F3F\
CFCFCF706F8FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
AFC0E0CFCFCFC0C0C0CFCFCF807F9F807F904F2F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
507090CFCFCFCFCFCFCFCFCFCFCFCF506F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F709FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FA0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF707FA0CFCFCFCFCFCFCFCFCF7F7FAF707FA07F80AF707FA07F7FAF707FA0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF606F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FAFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
8080BFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090C0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0A0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFAFAFDFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF5060402F2F2FCFCFCFCFCFCF3030302F2F2FCFCFCFCFCFCF
B0BFE0CFCFCF303030CFCFCFCFCFCFCFCFCF3F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF3F303FCFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFAFCFE0
808FA0CFCFCF707090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F303FCFCFCF\
CFCFCF7F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
B0CFEFCFCFCFCFCFCFCFCFCFCFCFCF807F9F4F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
506F8FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF506F8FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF5F709FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F7FA0\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF5F6F906F70A0707FA0707FA07F7FAFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF60709FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F70A0CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
7F7FB0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF8080B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C09090C0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFA0AFD0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C0302F30CFCFCF302F30C0C0C0CFCFCF
AFBFDFCFCFCF2F2F2FCFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF302F30CFCFCFCFCFCFCFCFCFC0C0C040303FCFCFCFCFCFCFC0C0C0B0CFE0
CFCFCF808FA0706F8FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F2F3F4F303FCFCFCF\
CFCFCFC0C0C07F7090CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
AFC0E0CFCFCFC0C0C0CFCFCFCFCFCF807FA04F2F3F807F9FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
50708FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF506F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60709FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F9F5060905F6F9F5F6F90
5F6F9FCFCFCFCFCFCF5F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF606F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F70A0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090C0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0A0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFA0AFDFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF2F2F2FCFCFCFCFCFCF6F80702F2F2FCFCFCF2F2F2FCFCFCFCFCFCF
B0BFE02F2F2FCFCFCF6F7F70CFCFCFCFCFCF3F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
302F307F8F9FCFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFCFCFCF
B0CFE0CFCFCF6F6F8F6F6080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F3FCFCFCF4F303FCFCFCF\
CFCFCFCFCFCF706F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFAFC0E0CFCFCFCFCFCFCFCFCF807F9F4F3040CFCFCF8F7F9FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807090CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
4F6F8FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F6F8FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F709F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
6F7FA0C0C0C05F6F905060905F6F905060905F6F90C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C05F6F90C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF606F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F70A0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C07F80B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C07F80B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C09090C0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F9FCFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0AFAFDFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C02F2F2FC0C0C06F7F702F2F2FCFCFCF2F2F2FCFCFCFC0C0C0CFCFCF
2F2F2FAFBFE0C0C0C06F7F70C0C0C0CFCFCF302F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF302F30\
CFCFCFC0C0C07F809FC0C0C0CFCFCFC0C0C040303FC0C0C0CFCFCFC0C0C0CFCFCF
AFC0E0CFCFCFC0C0C06F6080606080CFCFCFC0C0C0CFCFCF3F2F3FCFCFCFC0C0C04F3040C0C0C0\
CFCFCFC0C0C07F7090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0B0CFE0C0C0C0CFCFCFC0C0C0807F9F4F2F3FCFCFCFC0C0C0807F90C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807F90C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
507090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF506F8FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
607F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF50608F4F608F506090\
5F6F905F6F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6070A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F70A0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090CF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFA0AFD0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF50604FCFCFCFCFCFCF2F2F2FCFCFCFCFCFCF303030CFCFCFCFCFCFCFCFCF
303030AFBFE0CFCFCFCFCFCF6F8070CFCFCF3F303FCFCFCFCFCFCFCFCFCFCFCFCF302F30CFCFCF\
CFCFCFCFCFCF7F809FCFCFCFCFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFCFCFCF
B0CFEFCFCFCFCFCFCF606080CFCFCF6F608FCFCFCF402F3FCFCFCFCFCFCFCFCFCF4F303FCFCFCF\
CFCFCFCFCFCF706F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFB0CFE0CFCFCFCFCFCFCFCFCFCFCFCF4F3040CFCFCFCFCFCFCFCFCF8F7F9FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F90CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
506F8FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF506F8FCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
5F7090CFCFCFC0C0C0CFCFCFCFCFCF50608F4F608F4F608F4F608F50608FC0C0C0CFCFCFCFCFCF\
CFCFCF506090CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF506090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F70A0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF8080B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF7F7FB0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF9090C0\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C09F9FD0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0A0AFDFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF2F2F2FCFCFCF6F7F6F302F30C0C0C0CFCFCF2F2F2FCFCFCFC0C0C0302F30
CFCFCFB0BFE07F8F907F8F9F7F8F905F6080302F30CFCFCFCFCFCFCFCFCFC0C0C03F2F30CFCFCF\
CFCFCFC0C0C07F809FCFCFCFCFCFCFC0C0C040303FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFB0CFE0C0C0C0606080CFCFCFCFCFCF3F2F3FCFCFCFCFCFCFCFCFCFC0C0C04F303FCFCFCF\
CFCFCFC0C0C0CFCFCF706F8FCFCFCFC0C0C0CFCFCFCFCFCFB0CFE0AFC0E0B0CFE0
AFC0E0CFCFCF4F8F704F907FCFCFCFCFCFCF4F2F3FCFCFCFCFCFCFCFCFCFC0C0C0807F9FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF807090CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
50708FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF506F8FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F709F
60709F405F805060904F608F50608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F9FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FAFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090CF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFAFAFDFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
2F2F2FCFCFCF6F7F70303030CFCFCFCFCFCFCFCFCF303030CFCFCFCFCFCF2F2F2F
7F8F9FAFBFE0CFCFCFCFCFCFCFCFCF7F8F903F303F5F5F7FCFCFCFCFCFCF3F303FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF7F8F9FCFCFCFCFCFCF402F3FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFAFC0E0CFCFCF606080CFCFCF402F3FCFCFCF6F6080CFCFCFCFCFCFCFCFCF4F303FCFCFCF\
CFCFCFCFCFCFCFCFCF7F7090AFC0E0B0CFE0B0CFE0B0CFE0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF50907F4F8F704F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7F9F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
4F6F8FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F6F8F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F70905F709F5F70905F709F5F7090CFCFCF
C0C0C0CFCFCF405F7FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C05F6F90C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F6F90C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C060709FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F70A0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF7F7FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C08080B0C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8F90C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C09FA0D0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFA0AFD0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F202F\
2F2F2F607F6FCFCFCF2F2F2FCFCFCFC0C0C0302F307F8F907F8F9F2F2F2FCFCFCF
C0C0C0CFCFCFAFBFDFCFCFCFC0C0C03F2F30C0C0C0CFCFCF5F5F7F3F2F30C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF7F8090CFCFCF402F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0B0CFE0C0C0C0CFCFCF3F2F3FCFCFCFC0C0C0CFCFCF606080CFCFCF4F2F3FCFCFCFC0C0C0\
B0CFE0AFC0E0B0CFE0AFC0E0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F8F70CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
807F90C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807090CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
507090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
506F8F60709F5F7090607F9F5F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F9FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F70AFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
9090C0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0A0D0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFAFB0DFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF506040\
2F2F2FCFCFCF2F2F2FCFCFCF7F8F9F7F8F9F2F2F2FCFCFCF2F2F2FCFCFCFCFCFCF
CFCFCFCFCFCFB0BFE0CFCFCFCFCFCF302F307F8F9FCFCFCFCFCFCF302F30606080CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF808F9FCFCFCF40303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFB0CFE0CFCFCF402F3F6F6F8FCFCFCFCFCFCFCFCFCFB0CFEFB0CFE0B0CFE0AFCFE0B0CFEF\
CFCFCFCFCFCFCFCFCFCFCFCF7F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF50907F4F8F7FCFCFCF807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8F7F9F807090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7F9FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
506F8FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF506F8F506F8F506F8F506F8F506F8F\
CFCFCF305070CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0405F80CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF506090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF506090CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF606F9FCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F70A0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF7F7FB0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C08080B0CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
9090C0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF9F9FCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFA0AFD0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C02F2F2F2F2F2F\
6F7F70607F6F302F306F7F6F7F8F9FC0C0C0302F30CFCFCF302F30C0C0C0CFCFCF
CFCFCFCFCFCFAFBFDFCFCFCFCFCFCF3F303F7F8F90CFCFCF302F30CFCFCFC0C0C05F607F5F5F7F\
CFCFCFC0C0C0CFCFCFCFCFCF7F809F3F2F3FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFAFC0E0CFCFCFAFC0E0B0CFE0AFC0E0B0CFE0CFCFCFCFCFCF4F2F3FCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF7F7090C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF4F8F704F907FCFCFCFCFCFCFC0C0C0807F9FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0807F9F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF506F8F507090506F8F50708F506F8F507090506F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF3F5070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF506090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6070A0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F70A0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
9090C0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9FA0D0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0AFD0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5F3F302F30\
CFCFCF303030CFCFCF60705FCFCFCF7F8F9F2F2F2F302F30CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFAFBFE0CFCFCF302F307F8F9F302F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
5F5F7FCFCFCFCFCFCFCFCFCF7F809F403040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6FAF9F60A09FB0CFE0B0CFE06F6080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F30406F6080CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF706F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF4F8F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF807090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F90
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C03F5070C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF305070CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF405F7FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C05F6F90C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F6F90C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0606F9FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F70A0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C08080B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FB0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
9090C0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C09F9FCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0A0AFDFC0C0C0CFCFCFC0C0C0CFCFCF2F2F2F4F5F402F2F2F\
302F30C0C0C060705FC0C0C0CFCFCF2F2F2F302F30C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0AFBFE0C0C0C03F2F30302F307F8F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF5F5F7F5F607FC0C0C07F809F402F3F6FAF9F60A09F6FA09F60A09F6FAF9F
3F2F3FCFCFCFC0C0C050907FC0C0C0606080C0C0C0CFCFCFC0C0C0CFCFCF4F2F3FCFCFCF606080\
CFCFCFC0C0C0CFCFCFC0C0C07F7090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C050907F
4F8F70CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807F9FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCFC0C0C0807F9F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF3F5F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF3F5070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F9FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FAF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080BFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF9090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0AFDFCFCFCFCFCFCFCFCFCFCFCFCF4F5F402F2F2FCFCFCF\
2F2F2F60705FCFCFCFCFCFCFCFCFCF3030302F2F2F808F9FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFAFBFE0CFCFCF302F303F303F7F8F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6FAFA060A09F6FAFA06FA09F6FAFA0CFCFCFCFCFCFCFCFCFCFCFCF402F3F
CFCFCFCFCFCFCFCFCFCFCFCF50907F606080CFCFCFCFCFCFCFCFCFCFCFCF4F3040CFCFCFCFCFCF\
6F608FCFCFCFCFCFCFCFCFCFCFCFCF7F7090CFCFCFCFCFCF4F8F7F50907FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7FA0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807090CFCFCFCFCFCFCFCFCF807090
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F5070305070CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF305070CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF405F7FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF506090CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F6F90CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF60709FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F70A0\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF7F7FB0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF7F7FB0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF8F90C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF9FA0D0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFAFAFDFC0C0C0CFCFCFCFCFCF2F2F2F4F5F3F2F2F2F2F2F2F\
CFCFCF5F6F5FCFCFCFCFCFCFCFCFCF2F2F2FCFCFCFCFCFCF7F8F9FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0AFBFE0CFCFCF3F303FC0C0C07F8F9FCFCFCF6FAF9F60A09F6FA09F60A09F\
6FAF9FC0C0C0CFCFCFCFCFCF5F60803F2F3FCFCFCFCFCFCFCFCFCF3F2F3FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF4F8F706F6080C0C0C0CFCFCFCFCFCFCFCFCF4F2F3FCFCFCFCFCFCF\
CFCFCF606080CFCFCFCFCFCFCFCFCF706F8FCFCFCF4F8F70CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807F9FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807090CFCFCFC0C0C0CFCFCF
807090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F5070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F5070CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF406080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F90CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6070A0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F70A0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF9090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFAFDFCFCFCF2F2F2F4F5F3F5F604F2F2F2F303030\
5F6F5FCFCFCFCFCFCFCFCFCFCFCFCF303030CFCFCFCFCFCF6F807F606F80CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF6FAF9F60A09F6FAFA06FA09F6FAF9FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF403040606080CFCFCF402F3FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F8F70CFCFCFCFCFCFCFCFCFCFCFCF4F3040CFCFCFCFCFCF\
CFCFCFCFCFCF6F6080CFCFCFCFCFCF50907F4F8F70CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807090CFCFCFCFCFCF
8F7F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF305070CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F5070C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0405F80C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F6F90C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF506090CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF606F9FCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
6F70A0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF7F7FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07F80B0C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C09090C0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF9F9FCFCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFA0AFD0CFCFCF2F202F4F5F402F2F2F302F305F6F50\
CFCFCFC0C0C0CFCFCFC0C0C0302F30C0C0C0CFCFCF6F7F7FCFCFCFA0AFD0AFAFDF
A0AFD0A0AFDFA0AFD0AFAFDF60A09FCFCFCFC0C0C0CFCFCF7F8F90CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF402F3F6F60803F2F3FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F8F70CFCFCFC0C0C0CFCFCF4F2F3FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF4F8F704F907F706F8FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807F9FC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807090CFCFCF
807090CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F50703F5F7FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F5F7FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF406080CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F90CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6070A0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
6F70A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF8080BFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80B0CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF9090C0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0A0D0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFAFAFDF2F2F2F50604050604F302F302F2F2F60705F\
CFCFCFCFCFCFCFCFCFAFB0DFA0AFDFAFAFDFA0AFD0AFB0DFA0AFDF606080CFCFCF
CFCFCF7F8F9F3F3030CFCFCF3F30306FA09FCFCFCFCFCFCF808F9FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40303F402F3F6F6F8FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F804F8F70CFCFCFCFCFCF4F3040CFCFCFCFCFCF\
CFCFCF50907F4F8F70CFCFCF6F608FCFCFCF706F90CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807090
CFCFCF807F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F5070CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F5070C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF405F80C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF506090CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF5F6F90C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C060709F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
6F70A0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF8080B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF8080B0C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C09090C0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF9F9FCFCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0A0AFDF4F5F3FCFCFCFA0AFD0A0AFDFA0AFD0\
AFAFDFA0AFD0A0AFDF2F2F2F3030306F7F7FCFCFCFCFCFCFCFCFCFC0C0C05F6080
CFCFCFCFCFCF302F2F7F8F9F302F306FAF9FC0C0C0CFCFCFCFCFCF7F8F9FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F2F3F5F607FCFCFCF6F6080C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF606080CFCFCF4F8F704F3040C0C0C0CFCFCFCFCFCF\
50907FC0C0C0CFCFCFCFCFCFCFCFCF6060807F7090CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF807FA0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
807090807F9FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F5070\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F5F7FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF506090CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F9F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF9090C0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA0A0D0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF80B0CF80B0CF80B0CF80B0D09F9FCFAFAFDFA0AFD0303030CFCFCFCFCFCF\
CFCFCFCFCFCF302F307080802F2F2FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F607F
CFCFCF302F30CFCFCF3030307F8F9F7F8F906FAFA0CFCFCFCFCFCF7F8F90CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF303030CFCFCF5F5F7FCFCFCFCFCFCF6F6F8FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608050907F4F303FCFCFCF4F8F7050907F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF706F90CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
3F50703050703F5070C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F5070
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF405F7FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF506090CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF506090CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C060709F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF6F70A0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C07F80B0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FB0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF8F90C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C09FA0D0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C080B0CF7FAFCF80B0CF80AFCF80B0CF
7FAFCF80B0CF80AFCFCFCFCFC0C0C0CFCFCF2F2F2F80B0CFC0C0C02F2F2F5F6F5FCFCFCFC0C0C0\
CFCFCF2F2F2F6F80802F2F2FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
302F2FCFCFCFC0C0C03F3030C0C0C0CFCFCF60A09F7FB0BFC0C0C07F8F9FC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF3F2F303F3030C0C0C0CFCFCF5F5F7FCFCFCFC0C0C06F6080
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F6080C0C0C04F907F4F8F70CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF706F8FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807F9FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0807F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF303F5F3F5F7F3F50703F5F7F3F50703F5F7F3F50703F5F7F3F5070
3F5F7F405F80405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F80CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6070A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6F70AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF9090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9090C080B0D0\
80B0CF80B0D080B0CF80B0D080B0CF80B0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5060404F5F4050604F80B0D02F2F2FCFCFCFCFCFCFCFCFCF\
6F8080303030CFCFCF303030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F30
606F80CFCFCFCFCFCF302F30CFCFCFCFCFCFCFCFCF60A09F808F9F7F8F9F7F8F9FCFCFCFCFCFCF\
CFCFCFCFCFCF3F2F3F3F3030CFCFCFCFCFCFCFCFCF606080CFCFCFCFCFCFCFCFCF
6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40303040303FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7090CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF303F5F304060C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF304F6FC0C0C0405F80405F80405F80405F7F405F80405F80405F80405F7F405F804F608F\
50608FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F6F90
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F6F90CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
606F9FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C06F70A0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF7F7FB0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF7F7FB0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF9090C0CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF9090C08F90C09090C09090C09090C08F90C09090C09090C09090C08F90C080B0CFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF2F2F2F4F5F3FCFCFCF80AFCF60705FC0C0C0CFCFCF6F807F\
303030C0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF302F2FCFCFCF
CFCFCF606080C0C0C0303030CFCFCFCFCFCFC0C0C0CFCFCF60A09FCFCFCF5F5F7FCFCFCFCFCFCF\
40303F3F2F30CFCFCF303030CFCFCFC0C0C0CFCFCF5F5F7FCFCFCFC0C0C0CFCFCF
606080CFCFCFC0C0C0CFCFCFCFCFCF40303F402F30606080CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F60807F7090C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF304060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF304F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F507F\
CFCFCF5060904F608F50608F4F608F5060904F608F50608F4F608F5060905F6F90
5F6F9F5060905F6F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6F90CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF606F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF6F70A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80B0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF8F8FC08F8FBF9090CF9090C09090CF9090C09090CF9090C09090CF9090C0
9090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80B0CF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5F3F5F604F2F2F2F30303080B0CFCFCFCF6F807F303030\
CFCFCFCFCFCFCFCFCF303030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3030CFCFCF
CFCFCF5F607F3F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6FAF9FCFCFCF5050705F5F7F40303F\
CFCFCFCFCFCFCFCFCF3F3030CFCFCFCFCFCFCFCFCF5F6080CFCFCFCFCFCFCFCFCF
CFCFCF606080CFCFCF40303040303FCFCFCFCFCFCFCFCFCF6F6080CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCF7F7090CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0304060C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0304F6FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
3F507FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
405080CFCFCFC0C0C05F6F905060905F6F905060905F6F905060905F6F905060905F6F90506090\
606F9F606F9F60709F5F6F9F606F9F606F9F60709F5F6F9FCFCFCFC0C0C0CFCFCF
C0C0C0606F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C06F70A0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07F80B07F7FB0708FBF
8080BF8F8FBF8080BF8F8FBF8080BF8F8FBF8080BF8F8FBF8080BF8F8FBF8080BF8F8FBF8080BF\
8F8FBFC0C0C0CFCFCFC0C0C080AFD0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7FAFCF\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF2F202F4F5F40C0C0C0302F302F2F2F80B0CF5F6F5FCFCFCF2F2F2F\
5F604FC0C0C0303030C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0303030C0C0C0CFCFCF
C0C0C0CFCFCF302F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06FA09FC0C0C03F303F5F5F7F\
CFCFCFC0C0C03F3030C0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F607FC0C0C0CFCFCF
C0C0C0403030402F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF606080CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F608FC0C0C07F7090
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807F9FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0807F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF304060303F5FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF304F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
4F5F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5F8F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F9F6070A0606F9F
6070A06F70A06070A05F6FA06070A05F6FA06070A05F6FA06070A05F6FA06070A05F6FA06070A0\
5F6FA06070A05F6FA07F7FB0707FAF7F7FAF707FAF7F7FB0707FAF7F7FAF707FAF
7F7FB0707FAF7F7FAF707FAF7F7FB0707FAF7F7FAF707FAF8080BF7F80B08080B07F80B08080BF\
7F80B08080B07F80B08080BF7F80B08080B07F80B08080BFCFCFCFCFCFCF7080BF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF80A0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF50604050604F3030302F2F2F60705FCFCFCF80B0D0302F2FCFCFCF\
CFCFCF5F6F50302F2F607F80607080CFCFCFCFCFCF3F3030CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6FA09F40303F4F5070606080\
CFCFCFCFCFCF303030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F7FCFCFCF403030
40303FCFCFCFCFCFCF606080CFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F608FCFCFCFCFCFCF
7F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF2F3F5FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF304F6FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF304F70CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF4F5F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
4F5F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF5F6F9FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF6F7FAFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F7FB0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
7080BFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF80A0CFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
80B0CFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C02F2F2F4F5F3FCFCFCF2F2F2F302F30CFCFCF5F6F5080AFCF303030CFCFCF\
CFCFCFC0C0C03030305F7080CFCFCF5F707F3030305F70806070805F707F607080
CFCFCFCFCFCF302F2FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF40303F60A09F505070CFCFCF\
5F6080302F30CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF40303F402F30CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF606080CFCFCFC0C0C0CFCFCF606080CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F608FC0C0C0CFCFCF
CFCFCF7F7090C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0807F9FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0807F9FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303F5F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF304F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF3F4F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF6070A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF6F7FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
7F8FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8FAFD0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF4F5F3F5F604F2F2F2FCFCFCF2F2F2FCFCFCF5F604F30303080B0CFCFCFCF\
CFCFCF506050302F30CFCFCF5F70803F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
607080302F306070806F807FCFCFCFCFCFCFCFCFCFCFCFCF40303FCFCFCFCFCFCF6FA09F505070\
CFCFCF3F3030CFCFCFCFCFCFCFCFCFCFCFCF40303040303F5F5F7FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606080CFCFCFCFCFCFCFCFCF606080CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6080CFCFCFCFCFCF
CFCFCFCFCFCF7F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
2F3F5F30405FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0304F6FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF304F70CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF405080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF405080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C06070A0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C06F7FB0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6070AFCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
7080BFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF80A0CFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF80AFCFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF2F2F2F4F5F40C0C0C0302F302F2F2FCFCFCF5F604F302F30C0C0C080B0CFC0C0C0\
5060502F2F2FCFCFCFC0C0C0607080302F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0303030C0C0C0404F6F5F607FCFCFCFC0C0C03F303FC0C0C0CFCFCFC0C0C06FAF9F4F4F6F\
CFCFCF3030305F607FC0C0C0403030402F30CFCFCFC0C0C0CFCFCF5F5F7FCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0606080C0C0C0CFCFCFC0C0C0606080C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF606080CFCFCF
C0C0C0CFCFCFC0C0C07F7090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807F9F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF807F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF304060303F5FCFCFCFCFCFCFCFCFCFCFCFCF304F70CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF4F5F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF4F5F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF5F6FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF6F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FB0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF708FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80A0D0CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF80B0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF4F5F4050604F3030305F604F303030CFCFCF5F6F50302F30CFCFCFCFCFCF80B0D0\
50604F303030CFCFCFCFCFCF303030607080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF303030CFCFCF404060606F803F303F40303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6FAFA0\
3F3030CFCFCF40303040303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606080CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F80CFCFCFCFCFCF606080CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCF
CFCFCFCFCFCFCFCFCF706F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8F7FA0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF8F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C03F2F2F303F5F3040602F3F5FCFCFCFCFCFCF304F6FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF3F507FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0405F8FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0405F8FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C05F6FA0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF6F7FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F7FB0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF708FBFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C080AFD0CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C080B0CFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF5060402F2F2F302F305F604FCFCFCF2F2F2F5F604F2F2F2FCFCFCFC0C0C0CFCFCF80AFCF\
CFCFCF2F2F2FCFCFCF302F30CFCFCFC0C0C0607080CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF3F3030C0C0C0CFCFCF3F2F30606080C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF60A09F\
40303F402F30CFCFCF5F5F7FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F5F7FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0606080CFCFCF6F6080C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF606080CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF706F8FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807F9F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF807F9FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F2F2FCFCFCF303F5F304060CFCFCF304F70CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF4F5F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF6070A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FAFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF7F8FC0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80AFD0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF2F2F2F5F604F2F2F2F5F604F2F2F2FCFCFCF302F30CFCFCFCFCFCFCFCFCF50604FCFCFCF\
80B0CF303030303030CFCFCFCFCFCFCFCFCFCFCFCF607080CFCFCFCFCFCFCFCFCF
3F3030CFCFCFCFCFCF3F303F404F6FCFCFCF606F80CFCFCFCFCFCFCFCFCF7FB0BF7FB0BF7FB0BF\
4F4F70CFCFCFCFCFCFCFCFCF5F5F7FCFCFCFCFCFCFCFCFCFCFCFCF606080CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6080CFCFCF6F6F8FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF706F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F2F2FCFCFCFC0C0C030405F303F5F304F6FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF304F70CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF405080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0405F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF5F6F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F7FAFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F7FB0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF7080BFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF80A0CFCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF7FAFCFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
505F402F2F2F2F2F2F5F604FC0C0C0302F30C0C0C0303030C0C0C0CFCFCFC0C0C0506050C0C0C0\
80B0CF302F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F5F4FCFCFCFC0C0C0CFCFCF
302F2FCFCFCF3F2F30CFCFCF3F4060CFCFCFC0C0C05F608070AFB07FB0BFC0C0C0CFCFCFC0C0C0\
505070C0C0C0CFCFCFC0C0C0CFCFCF5F5F7FCFCFCFC0C0C0CFCFCFC0C0C05F607F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0606080606080CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F608F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF706F8FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
807F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F2F40\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F2F2F402F30CFCFCFCFCFCF303F5F304F70\
30406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F4F7FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF4F5F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FB0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF7F8FC0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8FAFD0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF80B0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
30303050604F3030305F604FCFCFCF2F2F2F303030CFCFCFCFCFCFCFCFCF506050CFCFCFCFCFCF\
302F3080B0D0CFCFCFCFCFCFCFCFCF50605050604FCFCFCF607080CFCFCFCFCFCF
3F30303F303FCFCFCFCFCFCFCFCFCF7FB0BF7FB0BF7FB0BFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
505070CFCFCFCFCFCFCFCFCFCFCFCF606080CFCFCFCFCFCFCFCFCFCFCFCF5F5F7F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F80CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6080
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF706F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40304F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF402F2FC0C0C0CFCFCFCFCFCF\
30304F30406F304F6FCFCFCFCFCFCFC0C0C0CFCFCF304F7FCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF4F5F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF405080CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6070A0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F7FB0CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6070AFCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0708FBFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF80AFD0C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF80B0CFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
2F2F2F303030C0C0C05F604F2F2F2F3030305F604FCFCFCFCFCFCFCFCFCF4F5F4FCFCFCFCFCFCF\
3F303080AFCFCFCFCFCFCFCF506050C0C0C0CFCFCFCFCFCFCFCFCF5F707F3F6F4F
4F806FCFCFCFC0C0C07FB0BF70AFBF404F6FC0C0C0CFCFCF5F607FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF4F4F6FCFCFCFCFCFCFCFCFCFC0C0C05F607FCFCFCFCFCFCFC0C0C05F607F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF606080CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F608F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF706F8FCFCFCFC0C0C0CFCFCFCFCFCF\
807FA0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF807FA0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF3F3040\
40304FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F30CFCFCFCFCFCF\
302F4FCFCFCFCFCFCF304F6F30406F304F70CFCFCF3F507FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FB0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF708FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80A0D0CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF80B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F2F
5F6F4F2F2F2F5F6F4FCFCFCF3030302F2F2FCFCFCFCFCFCFCFCFCF50604FCFCFCFCFCFCF3F3030\
CFCFCFCFCFCF80B0CF506050CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF406F4F4F806F
7FB0BF7FB0BF7FB0BFCFCFCFCFCFCF404060CFCFCFCFCFCFCFCFCF5F607FCFCFCFCFCFCFCFCFCF\
CFCFCF505070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6080CFCFCFCFCFCFCFCFCF
5F6080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6080
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF706F8FCFCFCFCFCFCFCFCFCF\
CFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40304F\
3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F2F2F3F2F2F\
CFCFCF2F2F4FCFCFCFC0C0C0CFCFCFC0C0C0304F6F30406F3F507FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF405080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0405F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F6F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6070AFCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F7FB0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0708FBFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF80A0CFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C080B0CFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0302F30
2F2F2FCFCFCF5F6040CFCFCF2F2F2F5F604FC0C0C0CFCFCF4F5F4FCFCFCFC0C0C0303030302F2F\
CFCFCF505F4F80B0CFC0C0C0CFCFCFC0C0C0CFCFCF3F604F3F6F4F70AFBF7FB0BF
5F707FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F4060CFCFCFC0C0C0CFCFCF5F607FCFCFCFC0C0C0\
CFCFCF4F4F70CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F607FC0C0C0CFCFCF
5F5F7FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F6F8F60608F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
606080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF706F8FCFCFCFC0C0C0\
CFCFCF807F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F2F40\
3F3040C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
3F2F2FCFCFCF302F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30406F304F6F30406F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6070A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FB0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FAFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF7F8FC0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80AFD0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80B0D0CFCFCFCFCFCFCFCFCF302F302F2F2F
3030305F604FCFCFCF302F2F5F6F50CFCFCFCFCFCFCFCFCF506050CFCFCF303030302F30506050\
50604FCFCFCFCFCFCF80B0D0CFCFCF3F6F4F7FB0BF7FB0BF7FB0BFCFCFCFCFCFCF
CFCFCF607080CFCFCFCFCFCFCFCFCFCFCFCF404F6FCFCFCFCFCFCFCFCFCF606080CFCFCFCFCFCF\
CFCFCF505070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F7FCFCFCFCFCFCF
606080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608FCFCFCF\
60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF706F90CFCFCF\
CFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40304F\
3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF3F2F2F40302F302F4FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C030406F
303F6F30406F303F60CFCFCFCFCFCFCFCFCFC0C0C0405F8FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF405080CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F6FA0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F7FAFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F7FAF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF7080BFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF80A0CFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF80AFCFCFCFCFCFCFCFCFCFCF2F2F2F302F30
CFCFCF5F604FC0C0C03030305F604FCFCFCFC0C0C0506050CFCFCF303030C0C0C0303030CFCFCF\
CFCFCFC0C0C0CFCFCF80AFCF7FB0BF70AFBFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF5F707FCFCFCFCFCFCFCFCFCF3F4060CFCFCFCFCFCFCFCFCFC0C0C05F6080CFCFCF\
CFCFCFC0C0C0505070CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F5F7FCFCFCF
5F5F7FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C060608FCFCFCF\
6F6F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
6F6080CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C07F7090CFCFCF\
CFCFCF807F9FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0807F9FCFCFCFCFCFCFC0C0C03F3040CFCFCF\
CFCFCF3F2F40CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF3F2F2F402F2F302F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30406F
CFCFCFCFCFCFCFCFCF30406F30406F303F6FCFCFCFCFCFCF4F5F8FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F80CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FAFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FB0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7080BFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80A0D0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80B0CFCFCFCF2F2F2F303030CFCFCF
CFCFCF5F604F303030302F30CFCFCFCFCFCF506050CFCFCF3030305F604F3F3030CFCFCFCFCFCF\
80B0CF80B0D080B0CFCFCFCF4F806FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF607080CFCFCFCFCFCFCFCFCF404060CFCFCFCFCFCFCFCFCFCFCFCF606080\
CFCFCFCFCFCF4F5070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F7F
CFCFCF5F5F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8F\
CFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F7090\
CFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCF3F3040CFCFCF\
CFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C03F2F2F2F2F4FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
303F60CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF303F6030406F303F60405F8FC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0405F8FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06070A0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F7FB0C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6070AF\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0708FBFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C080AFD0
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C080B0CFC0C0C02F2F2F2F2F2FCFCFCF
506040302F302F2F2F5F6050C0C0C0CFCFCF505F4F3030305F604F5F604F80AFCF80B0CF7FAFCF\
CFCFCFC0C0C04F80704F7F6FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF5F707FCFCFCFC0C0C0404F6FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
5F6080C0C0C0505070C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
5F5F7F5F607FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60608F\
CFCFCFC0C0C06F6F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C06F608FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
7F7090C0C0C0807F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807F9FC0C0C0CFCFCFC0C0C03F304FC0C0C0\
CFCFCF3F3040CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F30302F4FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF30406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F804F5F8F405F80CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5F8FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6070A0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FB0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
6F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF708FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
8FAFD0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF80B0D02F2F2F5F6F4FCFCFCF
5F6F4F2F2F2F303030CFCFCFCFCFCF606F4F30303080B0CF80B0D080B0CFCFCFCFCFCFCFCFCFCF\
CFCFCF4F8070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF607080CFCFCF404060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
5F6080CFCFCFCFCFCF505070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6060805F5F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F90\
CFCFCFCFCFCFCFCFCF6F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF6F608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF7F7090807F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCFCFCFCF3F3040CFCFCF\
CFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F2F2F302F4FCFCFCFC0C0C0CFCFCF
CFCFCF30406FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF304F7FCFCFCF405080\
4F5F8F405080CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF405080CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF5F6F9FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6070AFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
6F7FB0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF7080BFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
80A0CFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF80AFCF303030C0C0C05F604F
2F2F2F3030305F604F80B0CF80AFCF80B0CF80AFCFCFCFCFCFCFCF3F3030C0C0C0CFCFCFCFCFCF\
4F8070C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F707FCFCFCF404060CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF5F607FCFCFCF4F4F70CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF5F6080C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF60608F\
CFCFCFC0C0C0CFCFCF60608FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF6F608FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C07F7090807F9FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF807F9FCFCFCFC0C0C03F3040CFCFCF\
CFCFCF3F2F40CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F4FCFCFCFCFCFCF
CFCFCFCFCFCF30406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F507FCFCFCFCFCFCF\
CFCFCFCFCFCF405F804F5F8F405F80CFCFCFCFCFCFCFCFCF405F80CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F6FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FB0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
6F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F8FC0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
80AFD0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F2F80B0CF5F6F4FCFCFCF2F2F2F
80B0CF80B0CF80B0D0606F4F5F604F2F2F2FCFCFCFCFCFCF3F3030CFCFCFCFCFCF4F806F4F8070\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF607080404F6FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5F607F505070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF5F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
60608FCFCFCFCFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF6F6080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF7F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCF3F3040CFCFCF\
CFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF302F4F40302F
C0C0C0CFCFCFC0C0C030406FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F4F7FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF405080405F8F405080405F8FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F6FA0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F7FB0
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF6F7FAFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7080BFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C080AFD0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C02F2F2F506F4F80B0CF80AFCF80B0CF
2F2F2F5F604F5F604FCFCFCF2F2F2FCFCFCFC0C0C03F30305F604FCFCFCF4F806FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F4060CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF4F4F6FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF50507F5F5F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
60608FC0C0C0CFCFCFC0C0C0CFCFCF60608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF606080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF7F80AF807F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF807F9FCFCFCFC0C0C03F304FC0C0C0\
CFCFCFC0C0C03F304FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F
30304FCFCFCFCFCFCFCFCFCF3F406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF304F7FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F6090405F8F
4F6090405F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6070A0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7FAF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6F7FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7080BFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF80A0D0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF8FAFD080A0D080AFD080A0D08FAFD0CFCFCF609F9F609090
609F9F5F604FCFCFCF302F2F606F4FCFCFCFCFCFCF302F30606F4F4F806FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404060CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF606F80505070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF5F5F805F5F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF808FAFCFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9F40304FCFCFCFCFCFCF\
CFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
302F4F30304F40302FCFCFCF303F6FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF304F7F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF304F7FCFCFCF
CFCFCFCFCFCF405F8F4F5F8F405F8F4F5F90C0C0C0CFCFCFCFCFCFCFCFCF5F6F9FCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F7FB0
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C06F7FB0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0708FBFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF80A0CFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C080AFD080A0CF\
80AFD080A0CF80AFD0CFCFCFCFCFCF506F4F302F305F6F4F3030305F6040303030
5F604F609F9F5F9090609F9FCFCFCFCFCFCF302F2F4F806F4F806FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF404F6FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF505070C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF50507FCFCFCF5F5F7FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF60608FCFCFCFCFCFCFCFCFCFC0C0C060608FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF606080CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF808FAFC0C0C0807F9FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF807FA03F2F40CFCFCFCFCFCF\
CFCFCFC0C0C03F3040CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF30304F30304F30304FCFCFCF303F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
304F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF304F7F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F6090405F8F4F5F90405F8F6070A0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F7FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF6F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F8FBFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF8FAFD0CFCFCFCFCFCFCFCFCFCFCFCF80A0D080AFD080A0D08FAFD0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF2F2F2F3030305F6F4F3030305F604FCFCFCF302F30
6070502F2F2FCFCFCF5F604F609F9F609090609F9F5F604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404F6FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF4F4F6F505070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF505F7F5F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F9F40304FCFCFCFCFCFCF\
CFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF302F4FCFCFCF2F2F4FCFCFCF303F60CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
3F4F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F4F7F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05070A0506F9F\
506FA0506F9F5070A0506F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
6070AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C06F7FB0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7080BFCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C07F9FCF80A0CF80AFD080A0CF80AFD080A0CFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C05070502F2F2F5F6F4F2F2F2F5F604F2F2F2F5F604F
2F2F2FCFCFCFC0C0C05F604F2F2F2F3030305F604FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F4060CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C04F4F704F4F70CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C05F5F80C0C0C05F5F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF60608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C060608FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C06F608FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0808FAFC0C0C0CFCFCFC0C0C0807F9FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807F9F3F3040CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF3F2F40CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF30304FCFCFCF30304F30406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F8FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF506FA05F70A0506F9F5F70A0506FA05F70A0506F9F
CFCFCF6F7FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF6F7FB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F8FC0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF7F9FCF809FCF7F90CF809FCF7F9FCF809FCF7F90CF
809FCFCFCFCFCFCFCF609090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3030302F2F305F6F4F2F2F2FCFCFCF5F604F303030302F2F
CFCFCFCFCFCF303030302F30CFCFCFCFCFCF60604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404060CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F50704F5070CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF505F7FCFCFCFCFCFCF5F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCF
CFCFCFCFCFCFCFCFCF6F6080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF302F4FCFCFCFC0C0C030406FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF304F7FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
304F7FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF405F8F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
506F9F6070A0607FAF6080B06080B06080B0607FAF6080B06080B06080B0607FAF6080B06080B0\
CFCFCFC0C0C0CFCFCF6F7FAFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF7080BF708FBF7080BF708FBF7080BF7090BF7F90C0809FCF7F90CF\
809FCF7F90C0809FCF7F90CFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF609090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF506F503030302F2F2FCFCFCF5F604F3030302F2F2F606F50
302F303030305F604FCFCFCFCFCFCFCFCFCF5F604FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0404F60CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF4F4F6F505070CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C05F5F7FCFCFCFCFCFCF5F507FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C060608FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF60608FCFCFCF
CFCFCFCFCFCFC0C0C06F608FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF7F80AFCFCFCFCFCFCFCFCFCFC0C0C0807F9F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF807F9FCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF3F3040CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304FCFCFCFCFCFCF30306030305FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF304F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF304F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5F90\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF5F70A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
6080B06F80B06080B0709FBF7080BF7F8FC0708FBF7F8FBF7080BF7F8FC0708FBF
7F8FBF7080BF7F8FC0708FBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F90BFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF609090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5070503030302F2F2F606F4FCFCFCF5F604F2F2F2F607050303030
CFCFCFCFCFCF606F4FCFCFCFCFCFCF5F604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404F6F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5070CFCFCF505070CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF5F5F80CFCFCFCFCFCF5F5F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608F
CFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF302F4FCFCFCFC0C0C030304F30305FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF304F7FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C03F4F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
405F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF506F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C06F90BFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F90BFCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF609090CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0303030C0C0C0302F30C0C0C05F604F2F2F2F3030302F2F30CFCFCF
C0C0C05F604FC0C0C0CFCFCFC0C0C05F604FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F4060\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F4F6FCFCFCFC0C0C0505070C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF50507FCFCFCFC0C0C0CFCFCF5F507FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C06F6F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
60608FCFCFCFC0C0C0CFCFCF606080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8080AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0807F9FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F3040807F9FC0C0C0\
CFCFCFC0C0C0CFCFCF3F2F40CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304FCFCFCF30304FCFCFCF30305F3F3F60\
CFCFCFCFCFCFCFCFCF3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
405F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF506F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF709FC0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F90BFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF609090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF5070502F2F2F3030305F6F4F303030302F30303030CFCFCF607050CFCFCF
CFCFCF5F604FCFCFCFCFCFCFCFCFCF5F604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
404060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F6FCFCFCF505070CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF5F5F80CFCFCFCFCFCFCFCFCF605F80CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF60608FCFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3040CFCFCF807F9FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF302F4FCFCFCF2F2F4FCFCFCF4F3F2F\
30305F30305FCFCFCFCFCFCF3F507FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF304F7FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF405F8FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0506FA0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F90BFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F90BFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF609090609F9FC0C0C0CFCFCFCFCFCF\
CFCFCF2F2F2F302F305F6F4F3030302F2F2F607050CFCFCF606F50C0C0C0CFCFCF
CFCFCF5F604FC0C0C0CFCFCF5F604FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF3F4060CFCFCFCFCFCFCFCFCFC0C0C04F4F70CFCFCFCFCFCF4F4F6FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF50507FCFCFCFC0C0C0CFCFCFCFCFCF5F5F7FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF60608FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF6F6F8FC0C0C0CFCFCF6F6080CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0808FAFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0807F9FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF40304FC0C0C0807F9FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF40304FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F3F3030304FCFCFCF30304FCFCFCF\
CFCFCF4F403030305FCFCFCF304F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF304F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF4F6090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF5F70A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F90BFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF709FBF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF609F9FCFCFCFCFCFCF\
506F50303030303030303030302F306F705FCFCFCFCFCFCF606F50CFCFCFCFCFCF
5F604FCFCFCFCFCFCFCFCFCF5F604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF404F6FCFCFCFCFCFCFCFCFCFCFCFCF4F4F70CFCFCFCFCFCF505070CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F7FCFCFCFCFCFCFCFCFCFCFCFCF5F5F80CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF6F6F8FCFCFCFCFCFCF6F6080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCFCFCFCFCFCFCF3F3040CFCFCF807F9FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C040302F302F4FCFCFCF2F2F4F\
CFCFCFC0C0C04F3F3030305F30305F304F7FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C03F4F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C04F5F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF506F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07090BFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
6F90BFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0609F9F506F4F\
302F302F2F30303030508070607050607050CFCFCFC0C0C0606F50C0C0C0CFCFCF
5F604FCFCFCFC0C0C05F604FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0404F6FC0C0C0CFCFCFC0C0C0CFCFCF404F6FCFCFCFC0C0C0505070
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F5F7FC0C0C0CFCFCFC0C0C0CFCFCF5F5F7FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF60608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C06F6F8FC0C0C06F608FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0808FAFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0807F9FC0C0C0CFCFCFC0C0C03F3040C0C0C0CFCFCF807F9F\
CFCFCFC0C0C0CFCFCFC0C0C03F3040C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F3F3030304F30304F\
CFCFCFCFCFCFCFCFCF5040304F3F303F507F304F7FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF405F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF506FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF709FC0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF709FC0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303030\
3030305F80705080706F705FCFCFCF6F705FCFCFCF607050CFCFCFCFCFCFCFCFCF
606F4FCFCFCFCFCFCF5F604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF404060CFCFCFCFCFCFCFCFCFCFCFCF4F5070CFCFCFCFCFCF4F5070
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F7FCFCFCF\
CFCFCFCFCFCFCFCFCF6F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF6F6F906F608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA0CFCFCFCFCFCF3F3040CFCFCFCFCFCF8F7FA0\
CFCFCFCFCFCFCFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C040302F302F4F\
30304FC0C0C0CFCFCFCFCFCF4F3F304F3F3030305F304F7F3F507FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF304F7FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF405F8FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF5F70A0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F90BFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF6F90BFCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F9FB070A0B04F6F4F\
5F8070C0C0C05F6F4F60705060705FC0C0C0CFCFCF606F4FCFCFCFC0C0C05F604F
CFCFCFCFCFCFC0C0C05F604FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF404060CFCFCFC0C0C0CFCFCF4F4F6FCFCFCFC0C0C0CFCFCF
4F4F70CFCFCFC0C0C0CFCFCFCFCFCF5F5F80C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF5F5F7F\
CFCFCFC0C0C0CFCFCFCFCFCF6F6F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF60608F6F608FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF8080AFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0807F9FCFCFCF40304FC0C0C0CFCFCF807F9F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F2F40CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F3F30\
30304F30304FCFCFCFCFCFCFCFCFCF4F40304F4030303060CFCFCF3F507F304F7F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF304F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF4F5F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F70A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F90BFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF6F90BFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F9FB070A0B06F9FB05F70504F6F4F5F8070\
CFCFCF606F4F607050CFCFCF607050CFCFCFCFCFCF607050CFCFCFCFCFCF5F604F
CFCFCFCFCFCF606F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF404F6FCFCFCFCFCFCFCFCFCF4F4F70CFCFCFCFCFCFCFCFCF
CFCFCF4F4F70CFCFCFCFCFCFCFCFCFCFCFCF5F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F80\
CFCFCFCFCFCFCFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608F6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807FA03F3040CFCFCFCFCFCF807FA0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
40302F302F4FCFCFCFC0C0C0CFCFCF4F3F2FCFCFCF30305F4F4030C0C0C0CFCFCF
304F7F3F4F7FC0C0C0CFCFCFC0C0C03F4F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0405F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF506F9FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F90BFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF6F90BFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C06FA0B06F9FB0CFCFCFC0C0C0CFCFCF406F4F5F80705F604F\
CFCFCF5F604F607050606F50CFCFCFC0C0C0606F50C0C0C0CFCFCFC0C0C05F604F
C0C0C0CFCFCF5F604FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0404F60C0C0C0CFCFCFC0C0C04F4F70C0C0C0CFCFCF
C0C0C0505070C0C0C0CFCFCFC0C0C0CFCFCF50507FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
5F5F7FC0C0C0CFCFCFC0C0C060608FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0808FAFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F2F40807F9FC0C0C0CFCFCFC0C0C0\
807F9FC0C0C0CFCFCFC0C0C0CFCFCF3F3040CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF4F3F3030304FCFCFCFCFCFCFCFCFCF4F3F30CFCFCF30305F504030CFCFCF
CFCFCFCFCFCF3F507F304F7FCFCFCFCFCFCF3F507FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF405F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF506F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF709FC0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF709FC0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF70A0BF6F9FB070A0BFCFCFCFCFCFCFCFCFCFCFCFCF4F6F4F5F80705F6F4FCFCFCF\
CFCFCF5F6F4FCFCFCF6F705FCFCFCFCFCFCF606F50CFCFCFCFCFCF60604FCFCFCF
CFCFCFCFCFCF60604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF404060CFCFCFCFCFCFCFCFCF4F4F70CFCFCFCFCFCF
CFCFCFCFCFCF505070CFCFCFCFCFCFCFCFCF5F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF5F5F80CFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80A0707090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40304FCFCFCF807FA0CFCFCFCFCFCF\
807F9FCFCFCFCFCFCFCFCFCFCFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C040302F302F4FCFCFCFC0C0C0CFCFCF4F3F2FCFCFCF30305F4F4030
4F3F30CFCFCFC0C0C0CFCFCF304F7F3F507F304F7FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF405F8FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0506FA0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F90BF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF7090BFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
6F9FB070A0B0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF4F6F4F5F8070C0C0C05F6F4FCFCFCF\
6F7F5FC0C0C0CFCFCF607050CFCFCFC0C0C0606F50CFCFCFCFCFCF5F604FCFCFCF
CFCFCF5F604FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F4060CFCFCFCFCFCF4F4F70C0C0C0CFCFCF
CFCFCFCFCFCF4F4F6FCFCFCFCFCFCFCFCFCF50507FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF5F507FCFCFCFCFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF707FA0CFCFCF707090CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF808FAFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF3F3040CFCFCFC0C0C0807F9FCFCFCF\
807FA0C0C0C0CFCFCFCFCFCFCFCFCF3F2F40CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF4F3F3030304FCFCFCFCFCFCF4F4030CFCFCF303F60CFCFCF
CFCFCF4F3F30CFCFCFCFCFCFCFCFCFCFCFCF3F507F304F7F405F8FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F6090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F70A0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F90BF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF709FC0CFCFCFCFCFCFCFCFCF70A0BF6F9FB0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F6F4F5F80705F705FCFCFCF5F6F4FCFCFCF\
5F6F4FCFCFCF607050CFCFCFCFCFCF607050CFCFCFCFCFCFCFCFCF606F4FCFCFCF
CFCFCF5F604FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404F6FCFCFCFCFCFCF4F4F6FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF4F5070CFCFCFCFCFCFCFCFCF505F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5F5F7FCFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA07F709FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40304FCFCFCFCFCFCFCFCFCF807FA0\
CFCFCF807FA0CFCFCFCFCFCFCFCFCF40304F402F40CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C030305030305FCFCFCFC0C0C04F3F30C0C0C030305F
C0C0C0CFCFCF4F3F30CFCFCFC0C0C0CFCFCFC0C0C0304070C0C0C0405F8F3F5F8F405F8FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C04F5F8FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF506F9FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
6F90BFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F90BF70A0B06F9FB06FA0B0C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5080705F80704F6040CFCFCF5F604FCFCFCF5F604F\
CFCFCFC0C0C0607050C0C0C0CFCFCF606F4FCFCFCFC0C0C05F604FC0C0C0CFCFCF
5F604FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0404F6FC0C0C0CFCFCF4F4F6FCFCFCF
C0C0C0CFCFCFC0C0C0505070C0C0C0CFCFCFC0C0C05F5F80C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C05F5F7FC0C0C0CFCFCF60608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07080A0C0C0C0707090C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8080AFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F2F40CFCFCFC0C0C0CFCFCFC0C0C0\
807F9F807F9FCFCFCFC0C0C0CFCFCF3F3040403040C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30305030305FCFCFCF4F3F30CFCFCF30305F
CFCFCFCFCFCF5040304F403050403FCFCFCFCFCFCFCFCFCF3F4F70CFCFCFCFCFCFCFCFCF40608F\
405F8F405F8F405F8FCFCFCFCFCFCFCFCFCF405F8FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF506FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
709FC0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF60909F6F9F9F60909F6F9F9FCFCFCF6F9F9F60909F6F9F9F60909F
6F9F9F60909FCFCFCFCFCFCFCFCFCF508070303030506F50607F60CFCFCF5F6F4F5F6F4F6F705F\
CFCFCFCFCFCF607050CFCFCFCFCFCF607050CFCFCFCFCFCF5F604FCFCFCFCFCFCF
6F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404060CFCFCFCFCFCF4F5070CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF505070CFCFCFCFCFCF505F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF605F80CFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCF7F709FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF807FA0CFCFCFCFCFCF3F3040CFCFCF402F40CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C030305030305FCFCFCF4F3F2FCFCFCF
30305FCFCFCFC0C0C04F4030CFCFCF50403FC0C0C0CFCFCFCFCFCF304F70C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF3F5F8F405F8F3F5F8FCFCFCF405F8FCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF5F70A0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF7090BFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF60909F60909F60909F\
60909F60909FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF60909F60909F2F2F2F303030506F4F607F5FCFCFCF5F6F4FC0C0C05F6F4F607050\
CFCFCF606F50CFCFCFCFCFCF606F50C0C0C0CFCFCFCFCFCF5F604FC0C0C06F705F
607050CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF404060CFCFCF4F4F6FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF505070C0C0C05F5F7FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF5F5F7FC0C0C060608FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C07080A0CFCFCFCFCFCFC0C0C070709FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF7F80AFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF3F3040CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF7F80AFCFCFCFCFCFCF40304FC0C0C0403040CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40302F30305030305F303F60CFCFCF
CFCFCF30305FCFCFCF4F4030CFCFCFCFCFCF50403FCFCFCFCFCFCF304070CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F8F405F8F506F9F5F70A0506FA0
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F70A0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF709FC0CFCFCFCFCFCFCFCFCF6F9F9F60909F6F9F9F60909FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF6F9F9F60909F303030CFCFCF5F70505F7F5FCFCFCF5F6F4F606F4F607050CFCFCF\
CFCFCF6F705FCFCFCFCFCFCF606F50CFCFCFCFCFCF5F604FCFCFCF6F705FCFCFCF
6F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4040604F5070CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F70CFCFCFCFCFCF5F5F80CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF5F5F7FCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80A0CFCFCFCFCFCFCFCFCF7F709F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF808FAFCFCFCF3F3040CFCFCF402F40CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF302F504F3F2FC0C0C030305F
C0C0C030305FC0C0C0CFCFCF4F3F30CFCFCFC0C0C050403F4F3F30CFCFCF304070CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0405F8FC0C0C0CFCFCF
506F9F506FA0506F9F5070A0506F9F506FA0506F9F5070A0C0C0C0CFCFCF506F9FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06080B0607FAF6080B0
607FAF6080B0608FAF60909F60909060909FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C060909FC0C0C0303030C0C0C05070505F705FCFCFCF5F604F5F6F4FC0C0C0607050C0C0C0\
CFCFCF607050CFCFCFC0C0C0606F50C0C0C0CFCFCF5F604F6F705FC0C0C06F705F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0404F60C0C0C04F4F70
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F4F70CFCFCF50507FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F5F7F6F6F8FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF707F9FCFCFCFC0C0C0CFCFCFC0C0C0\
707090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0808FAF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F304FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0808FAFC0C0C03F3040C0C0C0CFCFCF3F2F40CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030504F3F304F3F30
3F3F60CFCFCF303F60CFCFCF50403FCFCFCFCFCFCFCFCFCFCFCFCF4F40303F4F70CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F8FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F70A0506FA05F70A0506F9F6F80B0\
6080B06080B06080B06F80B06080B06080B06080B06F80B0CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF60909FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60909F
6F9F9FCFCFCF303030506F505F70505F7F5FCFCFCFCFCFCF606F4FCFCFCF6F705FCFCFCFCFCFCF\
607050CFCFCFCFCFCF607050CFCFCFCFCFCFCFCFCF606F4FCFCFCFCFCFCF6F7050
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404F6F4F4F6F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505070CFCFCF5F5F80CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80A0CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF7F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF8080AF40304FCFCFCFCFCFCFCFCFCF40304FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF302F504F3F2F
4F3F2F30305FC0C0C030305FCFCFCF504030C0C0C0CFCFCFCFCFCFCFCFCF4F3F30304070CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F5F8FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05070A0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF60909F60909FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF60909F60909FCFCFCF
CFCFCFCFCFCF2F2F2F608070CFCFCF607F5FC0C0C05F6F4F5F6F4FCFCFCF606F50CFCFCFCFCFCF\
60705FC0C0C0CFCFCF606F4FCFCFCFC0C0C05F604FCFCFCFCFCFCFC0C0C06F705F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F40604F4F70
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C050507050507FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F5F7F60608FCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF7080A0C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C070709FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0808FAF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F3040CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF8080AFCFCFCFC0C0C0CFCFCF3F2F40CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F3F30303050
4F3F2FCFCFCF303F6030305FCFCFCF4F3F30CFCFCFCFCFCFCFCFCFCFCFCF50403FCFCFCF304F70\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF405F8F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F7FA0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F9F9FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F9F9FCFCFCFCFCFCFCFCFCF
CFCFCF2F2F2FCFCFCF608070607F5FCFCFCFCFCFCF5F6F4FCFCFCF607050CFCFCFCFCFCF6F705F\
CFCFCFCFCFCFCFCFCF607050CFCFCFCFCFCF606F4FCFCFCFCFCFCF6F705FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F70
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505070505F7FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF7F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
808FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCFCFCFCF403040CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0303050
C0C0C04F3F2FC0C0C0CFCFCF302F504F4030C0C0C0CFCFCFC0C0C0CFCFCFC0C0C050403F30406F\
50403FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0405F8F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
5070A0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C060909FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60909F60909FC0C0C0CFCFCFC0C0C0CFCFCF
2F2F2FCFCFCF60806FCFCFCF5F705FCFCFCF5F6F4FCFCFCFC0C0C0607050C0C0C0CFCFCF606F50\
CFCFCFC0C0C0606F50C0C0C0CFCFCF5F604F6F705FC0C0C0CFCFCF606F50CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0404F6F
404F6FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F4F6F5F5F7FC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF606F9060608FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07080A0C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0707090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
7F80AFCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C02F2F30C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF7F80AFCFCFCFC0C0C0CFCFCF3F2F40CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
30305FCFCFCF4F3F30CFCFCF30305F30406F304070CFCFCFCFCFCFCFCFCFCFCFCF4F4030CFCFCF\
30407050403F4F4030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
40608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF5F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F9F9FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF60909FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F2F
CFCFCF608070CFCFCF5F705FCFCFCF5F70506F7F60CFCFCFCFCFCF607050CFCFCFCFCFCF6F705F\
CFCFCFCFCFCF606F50CFCFCF606F4F6F705FCFCFCFCFCFCFCFCFCF6F705FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
4F5070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F7FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F709FCFCFCF6F6F8FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
808FAFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F50302F2F2FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF40304F808FAFCFCFCFCFCFCFCFCFCF403040CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF30305040302F4F3F30CFCFCF3030504F3F3030406F30406FCFCFCFC0C0C0CFCFCF4F4030\
304F70C0C0C0CFCFCF4F4030CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF405F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C05070A0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF60909F60909FCFCFCFCFCFCF\
CFCFCFC0C0C060909F60909FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF2F2F2FCFCFCF
60806FCFCFCFC0C0C0607F5FCFCFCF607050C0C0C0CFCFCF607050CFCFCFC0C0C0607050CFCFCF\
CFCFCFC0C0C0606F505F604FCFCFCF606F50CFCFCFCFCFCFCFCFCF606F50CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
4F4F6FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF5F5F80C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F709FC0C0C060608F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF706F90CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
8080AFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3050302F2F30CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C03F3040CFCFCF808FAFC0C0C0CFCFCFCFCFCF403040C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3030504F3F2F4F4030CFCFCF3030504F4030CFCFCF30406FCFCFCFCFCFCF50403F\
CFCFCF3F4F70CFCFCFCFCFCF4F4030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF405F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5070A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60909FCFCFCF\
60909F6F9F9FCFCFCFCFCFCF608F906F90906090906F90902F2F2FCFCFCF608070
CFCFCFCFCFCF607F5FCFCFCF607050607F5FCFCFCFCFCFCF6F705FCFCFCFCFCFCF607050CFCFCF\
CFCFCF607050606F4FCFCFCFCFCFCF6F705FCFCFCFCFCFCF607050CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF4F4F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F90CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF607090CFCFCF60608F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80A0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707090CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF8080AFCFCFCFCFCFCFCFCFCFCFCFCF3F5030CFCFCF2F2F30CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF3F3040CFCFCF8080AFCFCFCFCFCFCFCFCFCF3F2F40CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF30304F30304F40302F4F3F2F3030504F4030C0C0C0CFCFCF30406F3040704F3F30\
CFCFCF304070CFCFCFC0C0C0CFCFCF4F403050403FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF3F5F8FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF50709FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06090AF608FAF\
609090608F8F609090608F8FCFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C0608070
C0C0C0CFCFCF5F705F5F705060705FCFCFCFC0C0C0607050C0C0C0CFCFCFC0C0C0607050C0C0C0\
CFCFCF606F4F606F50C0C0C06F705FC0C0C0CFCFCFC0C0C06F705FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C04F4F70C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60608F5F5F80C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F7090C0C0C0CFCFCF
60608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF707FA0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF706F90CFCFCFC0C0C0CFCFCF
C0C0C0808FAFC0C0C0CFCFCFC0C0C0CFCFCF3F5030CFCFCF2F2F2FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C03F304FC0C0C0CFCFCF8080AFCFCFCFC0C0C0403040C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF30304F303050CFCFCF4F403030305050403FCFCFCFCFCFCFCFCFCF3F4070\
30406FCFCFCF304070CFCFCFCFCFCF50403FCFCFCF504F3FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF405F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF5070A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F90B0608FAF6F90B0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030302F2F2FCFCFCFCFCFCF608070CFCFCF
CFCFCF5F7F5F607050CFCFCF6F7F60CFCFCFCFCFCF607050CFCFCFCFCFCF6F705FCFCFCFCFCFCF\
CFCFCF607050CFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF4F5070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F90505F7FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7090CFCFCFCFCFCF
CFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707090CFCFCFCFCFCF
CFCFCF808FAFCFCFCFCFCFCFCFCFCFCFCFCF3F5030CFCFCF303030CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF3F3040CFCFCFCFCFCF808FAFCFCFCFCFCFCF403040CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF30304F30304FC0C0C04F3F30303050CFCFCFC0C0C0CFCFCFCFCFCF\
50403F30406F30406F304070CFCFCFC0C0C050403FCFCFCF50403F4F4030CFCFCF
CFCFCFCFCFCFC0C0C0405F8FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF50709FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6090AF608FAF6090AFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF303030C0C0C0CFCFCFCFCFCF608070C0C0C0CFCFCF
5F705F607050C0C0C06F7F5FCFCFCFCFCFCF606F50CFCFCFCFCFCFCFCFCF606F50CFCFCFCFCFCF\
606F50C0C0C0CFCFCFCFCFCF6F705FC0C0C0CFCFCF607050CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C04F4F70CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0606F8F50507F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF606F90CFCFCF
CFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C07080A0CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF706F90CFCFCF
CFCFCF808FAFC0C0C0CFCFCFCFCFCF3F5030C0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF3F2F40CFCFCFCFCFCFCFCFCF7F80AFCFCFCFCFCFCFCFCFCF3F2F40CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F303050CFCFCF4F4030303050CFCFCFCFCFCFCFCFCF\
CFCFCF50403FCFCFCF405F8F405F8F405F8F4F4030CFCFCFCFCFCFCFCFCF4F403F
CFCFCFCFCFCFCFCFCFCFCFCF405F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F7FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF6F90B0608FAF6090B0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF303030CFCFCFCFCFCFCFCFCF608070CFCFCFCFCFCFCFCFCF
607F5FCFCFCFCFCFCF607F5FCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF6F705FCFCFCF607050\
6F705FCFCFCFCFCFCFCFCFCF607050CFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF4F4F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F8FCFCFCF\
505F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F709FCFCFCF
CFCFCFCFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707090
CFCFCFCFCFCF808FAFCFCFCFCFCFCF3F5030CFCFCFCFCFCF2F2F30CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCF40304FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF30304F30304FC0C0C0303050C0C0C0CFCFCFC0C0C0\
CFCFCF4F4030CFCFCFC0C0C030406FC0C0C0405F8F3F5F80405F8F3F5F8FCFCFCF
4F4030CFCFCFC0C0C0CFCFCF3F5F80CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0507FA0C0C0C0CFCFCFC0C0C0CFCFCF
608FAF608FAFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0303030C0C0C0CFCFCFC0C0C0CFCFCF5F806FCFCFCFC0C0C0607F5F
C0C0C0CFCFCF60705FCFCFCFC0C0C0CFCFCF607050CFCFCFC0C0C0607050C0C0C0606F50C0C0C0\
6F705FC0C0C0CFCFCF606F50CFCFCFC0C0C0CFCFCF606F50CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF404F6FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0606F8FC0C0C0\
5F5F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF606F90CFCFCF
C0C0C0CFCFCFC0C0C06F6F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C07080A0C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
706F90CFCFCF8080AFCFCFCFC0C0C03F5030C0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF3F3040CFCFCFC0C0C0CFCFCFC0C0C0808FAFC0C0C0CFCFCF3F2F40CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030504F3F2F3030504F3F2F303050CFCFCFCFCFCF\
CFCFCF50403FCFCFCFCFCFCFCFCFCF304070CFCFCF504F3FCFCFCFCFCFCF405F8F
40608F405F8F405F8FCFCFCFCFCFCF405F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5070A05F7FA0608FAF6F90B0608FAF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
2F2F30303030CFCFCFCFCFCFCFCFCFCFCFCF608070CFCFCFCFCFCF6F8F7F6F8F7F
CFCFCFCFCFCF6F7F60CFCFCFCFCFCF607050CFCFCFCFCFCFCFCFCF607050CFCFCF606F506F7F60\
CFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF4F5070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F8FCFCFCF\
CFCFCF5F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7090
CFCFCFCFCFCFCFCFCFCFCFCF6F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80A0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF707090808FAFCFCFCFCFCFCF3F5030CFCFCFCFCFCF303030CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCFCFCFCF40304FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C030304F403F2F30304F4F3F2F303050CFCFCF\
CFCFCFC0C0C0504030CFCFCFCFCFCF30406FCFCFCF4F4030CFCFCFC0C0C0CFCFCF
CFCFCF50403FC0C0C0405F8F3F5F8F405F8F3F5F8F5070A05070A0507FA050709F5070A05070A0\
507FA050709F5070A05070A0507FA050709FCFCFCFCFCFCF609090608F8FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF2F2F2F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F806FCFCFCF608F7F6F8F7F60807FCFCFCF
CFCFCF6F7F5FC0C0C0CFCFCFCFCFCF60705FC0C0C0CFCFCF607050CFCFCF606F4FCFCFCF6F705F\
CFCFCFC0C0C0CFCFCF607050CFCFCFC0C0C06F705FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF4F4F70C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF606F8F\
CFCFCFC0C0C05F5F7FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F7090
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F6F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF707FA0\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF706F90808FAFCFCFCF3F5030C0C0C0CFCFCFCFCFCF2F2F30C0C0C0CFCFCFCFCFCF\
CFCFCF3F2F40CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF8080AFCFCFCF3F2F40CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303050CFCFCF303050303050CFCFCF\
CFCFCFCFCFCF4F4030CFCFCFCFCFCFCFCFCF30406F50403FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF504F3FCFCFCF504F3FCFCFCF406090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF609090
6F9090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F30CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF60806FCFCFCF6F8F7FCFCFCF6F8F7FCFCFCFCFCFCF
CFCFCF607F5FCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF6F705F606F50CFCFCFCFCFCF6F7F5F\
CFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F90\
CFCFCFCFCFCF505F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7090
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80A0\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF707FA03F5030CFCFCFCFCFCFCFCFCFCFCFCF2F2F2FCFCFCFCFCFCFCFCFCF\
3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCF402F40
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C030304FC0C0C030304F302F50\
CFCFCFC0C0C0CFCFCF4F3F30CFCFCFC0C0C0CFCFCF30406F50403FC0C0C0CFCFCF
C0C0C0CFCFCF4F4030CFCFCFC0C0C04F403F4F403040608FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0609090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C02F2F302F2F2FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF5F806FCFCFCF608F7FCFCFCF60807FCFCFCFC0C0C0CFCFCF
60705FCFCFCFC0C0C0CFCFCF606F50CFCFCFC0C0C0CFCFCF606F50CFCFCFC0C0C06F7F5FC0C0C0\
CFCFCFC0C0C06F705FC0C0C0CFCFCFC0C0C06F705FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F4F6FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60608F\
CFCFCFC0C0C0CFCFCF50507FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F709F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F6F8FC0C0C0CFCFCFC0C0C0CFCFCF707F9F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C07080A07F80AF808FAFC0C0C0CFCFCFC0C0C02F2F30C0C0C0CFCFCFC0C0C0\
3F3040C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F80AFCFCFCFC0C0C0403040
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F4F3F30CFCFCF303050\
303050CFCFCFCFCFCF50403FCFCFCFCFCFCFCFCFCFCFCFCF30406FCFCFCFCFCFCF
CFCFCFCFCFCF504F3FCFCFCFCFCFCFCFCFCFCFCFCF4F403F406090CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF6F9090609090CFCFCFCFCFCFCFCFCFCFCFCF303030CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF608F706F8F7FCFCFCFCFCFCF6F8F80CFCFCFCFCFCF607F5F
CFCFCFCFCFCFCFCFCF607050CFCFCFCFCFCFCFCFCF6F806FCFCFCFCFCFCFCFCFCF6F705FCFCFCF\
CFCFCFCFCFCF6F7050CFCFCFCFCFCFCFCFCF6F7050CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F6FCFCFCFCFCFCFCFCFCFCFCFCF6F6F90\
CFCFCFCFCFCFCFCFCFCFCFCF5F5F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
707FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF7F80A0CFCFCF808FAF8080AFCFCFCF2F2F2FCFCFCFCFCFCFCFCFCF\
3F3040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAFCFCFCFCFCFCF402F40
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF302F4FCFCFCFCFCFCF\
304F7030406FCFCFCF4F4030CFCFCFC0C0C0CFCFCFCFCFCF304070C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C050403FCFCFCFCFCFCFC0C0C0CFCFCF4F403F40608FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF608F90609090C0C0C0302F30CFCFCFCFCFCF5F806F60807060806F\
6080705F806F609090CFCFCFCFCFCFC0C0C06F8F7FCFCFCFCFCFCFC0C0C06F7F5F
CFCFCFCFCFCFC0C0C0607050CFCFCFCFCFCF6F806F6F7F60CFCFCFCFCFCF60705FCFCFCFCFCFCF\
CFCFCF606F50CFCFCFCFCFCFCFCFCF606F50CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C04F4F70CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
6F6F8FC0C0C0CFCFCFCFCFCF5F5F80C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
607090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF60608FCFCFCFC0C0C0CFCFCFCFCFCF\
7080A0C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF707FA0CFCFCFC0C0C0CFCFCF8080AF808FAFC0C0C0CFCFCFCFCFCF\
40304FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF808FAFC0C0C0403040
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F3F2FCFCFCF\
4F3F2F30305030406F3040704F4030CFCFCFCFCFCFCFCFCFCFCFCF3F4070CFCFCF
CFCFCFCFCFCFCFCFCF4F403FCFCFCFCFCFCFCFCFCFCFCFCF504F3FCFCFCF406090CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF608F90608FA060807060807060806FCFCFCFCFCFCFCFCFCF\
CFCFCF6F8F80CFCFCFCFCFCFCFCFCF6F8F80CFCFCFCFCFCFCFCFCF6F7F60CFCFCF
CFCFCFCFCFCFCFCFCF607050CFCFCF6F806F708060CFCFCFCFCFCFCFCFCF6F7F60CFCFCFCFCFCF\
CFCFCF6F705FCFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F70CFCFCFCFCFCFCFCFCFCFCFCF\
606F8FCFCFCFCFCFCFCFCFCFCFCFCF5F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608FCFCFCFCFCFCFCFCFCF\
707FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF7F80A0CFCFCFCFCFCFCFCFCFCFCFCF2F2F2F808FAF808FAF40304F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF8080AFCFCFCF402F40
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F2F4F\
CFCFCF4F3F2F30304F4F3F2F30406F30406FCFCFCFC0C0C0CFCFCF4F4030304070
C0C0C0CFCFCFC0C0C050403FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C04F403FC0C0C040608FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF608F90609090608F90CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60807F\
6F8F7FC0C0C0CFCFCFC0C0C06F8F7FC0C0C0CFCFCFC0C0C0CFCFCF60705FCFCFCF
C0C0C0CFCFCF607050CFCFCF6F806FCFCFCF6F7F60CFCFCFC0C0C0CFCFCF6F705FCFCFCFC0C0C0\
CFCFCF606F50CFCFCFC0C0C0CFCFCF606F50CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF404F6FCFCFCFC0C0C0CFCFCFC0C0C0\
606F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F5F80C0C0C0CFCFCFC0C0C0CFCFCF
606F90CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60608FCFCFCFC0C0C0\
CFCFCF707FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C03F5030C0C0C07080A0C0C0C0CFCFCFC0C0C02F2F30C0C0C0CFCFCF7F80AF\
808FAFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0808FAFC0C0C0CFCFCF
3F2F40CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
30304F4F4030CFCFCF3030504F403050403F30406F3F4070CFCFCFCFCFCF30406F
CFCFCFCFCFCFCFCFCFCFCFCF504F3FCFCFCFCFCFCFCFCFCFCFCFCF4F403FCFCFCFCFCFCF406090\
4F403FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F9090608F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F7FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCF607F5FCFCFCFCFCFCF
CFCFCFCFCFCF6F705F6F806FCFCFCFCFCFCF6F7F60CFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF\
6F705FCFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F70CFCFCFCFCFCFCFCFCF\
606F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F7FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF6F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60608FCFCFCF\
CFCFCF7F80A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF3F5030CFCFCF707FA0CFCFCFCFCFCFCFCFCFCFCFCF303030CFCFCF40304F\
CFCFCF808FAF8080AFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAF2F2F2F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
30304F302F4F4F3F3030304FCFCFCF4F3F30CFCFCFCFCFCF30407030406F50403F
30406FCFCFCFC0C0C0CFCFCF4F4030CFCFCFC0C0C0CFCFCFCFCFCF50403FC0C0C0CFCFCFCFCFCF\
40608F4F4030CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F9090608F90609090
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF60807FCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF608F7FCFCFCFC0C0C0CFCFCFCFCFCF6F7F5FC0C0C0CFCFCF
CFCFCF60705F6F806FCFCFCFCFCFCF6F7F60C0C0C0CFCFCFCFCFCF6F7F5FC0C0C0CFCFCFCFCFCF\
6F705FC0C0C0CFCFCFCFCFCF6F705FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF4F4F70C0C0C0CFCFCFCFCFCF\
CFCFCF60608FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF50507FCFCFCFC0C0C0CFCFCF
CFCFCF6F709FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C060608FCFCFCF\
CFCFCF707FA0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C03F5030CFCFCF7080A0C0C0C0CFCFCFCFCFCFCFCFCF2F2F2FCFCFCF3F3040\
CFCFCFC0C0C0CFCFCF8080AF808FAFC0C0C0CFCFCFCFCFCFCFCFCF7F80AF30302F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF30304F30304F4F403030304FCFCFCF4F3F30CFCFCFCFCFCFCFCFCF30406F
30407030406FCFCFCFCFCFCF50403FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF504F3FCFCFCFCFCFCF\
CFCFCF4060904F403FCFCFCFCFCFCF6F9090608F906F9090CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCFCFCFCF6F7F60CFCFCFCFCFCFCFCFCF
CFCFCF607050CFCFCFCFCFCFCFCFCF6F7F60CFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF\
607050CFCFCFCFCFCFCFCFCF607050CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5070CFCFCFCFCFCF\
CFCFCF6F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F7FCFCFCFCFCFCF
CFCFCF607090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F8F\
CFCFCF7F80A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF3F5030CFCFCFCFCFCF7F80A0CFCFCFCFCFCFCFCFCF303030CFCFCF40304F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808FAF808FAFCFCFCF3F2F40808FAF2F2F2F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C030304F2F2F4FCFCFCF30304F4F40304F3F2FCFCFCFC0C0C0CFCFCF
4F403030406F4F708F4F70904F6F8F4F70904F708F4F70904F6F8F4F70904F708F4F70904F6F8F\
4F70904F708F40608F608F90608F90C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60807FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF608F7FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60705FCFCFCFC0C0C0CFCFCF
6F8070CFCFCFC0C0C0CFCFCF6F7F60CFCFCFC0C0C0CFCFCF60705FCFCFCFC0C0C0CFCFCF606F50\
CFCFCFC0C0C0CFCFCFC0C0C06F705FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F4F6FCFCFCFC0C0C0\
CFCFCF60608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F5F7FC0C0C0CFCFCF
C0C0C0CFCFCF606F90CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
60608FC0C0C07080A0C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF3F5030CFCFCFC0C0C0CFCFCF707FA0CFCFCFC0C0C0CFCFCF2F2F2F3F304FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F80AF808FAFC0C0C0808FAF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF30304F30304FCFCFCF30304F5040304F4030CFCFCFCFCFCF
504F3FCFCFCFCFCFCF30406FCFCFCF4F403FCFCFCFCFCFCFCFCFCFCFCFCF504F3FCFCFCFCFCFCF\
CFCFCFCFCFCF4F7090507090CFCFCF504F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F7F6F8F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6F8F7FCFCFCFCFCFCFCFCFCFCFCFCF607F5FCFCFCFCFCFCF6F8F7F6F8F7F
708F7FCFCFCFCFCFCFCFCFCF70806FCFCFCFCFCFCFCFCFCF6F7F60CFCFCFCFCFCFCFCFCF6F705F\
CFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F6FCFCFCF\
CFCFCFCFCFCF606F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F80CFCFCF
CFCFCFCFCFCF6F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6F6F8F707FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3F5030CFCFCFCFCFCFCFCFCF7F80A0CFCFCFCFCFCFCFCFCF3030303F3040CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF402F40CFCFCFCFCFCF808FAF8080AF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF30304F302F4F30304F4F3F2FCFCFCF4F3F30CFCFCF
4F4030CFCFCFC0C0C030406FCFCFCF50403FC0C0C0CFCFCFCFCFCFCFCFCF4F4030CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF4F403F5070904F6F8F4F403F4F403FCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F8F7FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
6F8F7FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F7F5FCFCFCF6F8F7FC0C0C06F8F7F
CFCFCFCFCFCFC0C0C06F7F60CFCFCFCFCFCFC0C0C0CFCFCF6F705FCFCFCFC0C0C0CFCFCF607050\
CFCFCFC0C0C0CFCFCF607050CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF4F4F6F\
CFCFCFC0C0C0606F8FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F5F7F
CFCFCFCFCFCF606F90CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C07080A0CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF305030CFCFCFCFCFCFCFCFCFC0C0C07080A0CFCFCFCFCFCF2F2F2F3F3040CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F2F40CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF30304F4F403030304F3030504F3F2FCFCFCF4F4030
CFCFCF4F4030CFCFCFCFCFCF304F70CFCFCF504F3FCFCFCFCFCFCFCFCFCFCFCFCF4F403FCFCFCF\
CFCFCFCFCFCFCFCFCF504F3FCFCFCFCFCFCF4F70905070904F403FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F7F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7F606F8F7F6F8F7FCFCFCFCFCFCF6F8F7F
CFCFCFCFCFCFCFCFCF6F7F60CFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCF6F705F\
CFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4F70\
CFCFCFCFCFCF606F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F7F
CFCFCFCFCFCF6F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF707FA07F80A0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3F5030CFCFCFCFCFCFCFCFCFCFCFCF707FA0CFCFCFCFCFCF3030303F3040CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF3F2F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF302F404F3F302F2F4F30304FC0C0C0CFCFCF
4F3F2F50403FC0C0C0CFCFCFC0C0C030406F4F4030CFCFCFC0C0C0CFCFCFC0C0C050403FC0C0C0\
CFCFCFC0C0C0CFCFCF4F4030CFCFCFC0C0C0CFCFCFC0C0C04F7090C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C06F8F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F8F7FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60705FCFCFCFC0C0C0CFCFCF6F807FCFCFCF
C0C0C0CFCFCF6F7F60CFCFCFC0C0C0CFCFCFC0C0C06F7F5FC0C0C0CFCFCFC0C0C06F705FC0C0C0\
CFCFCFC0C0C06F705FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
4F4F70C0C0C0606F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
50507FCFCFCFC0C0C06F709FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF5F80807080A0707FA0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF3F5030CFCFCFC0C0C0CFCFCFC0C0C07080A0C0C0C0CFCFCFC0C0C03F304FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0403040C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304FCFCFCF30304F303050CFCFCF
CFCFCF4F4030CFCFCFCFCFCFCFCFCFCFCFCF304F70CFCFCFCFCFCFCFCFCFCFCFCF4F403FCFCFCF\
CFCFCFCFCFCFCFCFCF504F3FCFCFCFCFCFCFCFCFCFCFCFCF4F403F5070904F7090
CFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F7FCFCFCF\
CFCFCFCFCFCF6F806F6F806F6F806FCFCFCFCFCFCFCFCFCF6F8F7FCFCFCFCFCFCF
CFCFCFCFCFCF6F7F60CFCFCFCFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCF6F7050CFCFCF\
CFCFCFCFCFCF6F7050CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
4F4F70CFCFCFCFCFCF6F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF5F5F7FCFCFCF6F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF5F8F8FCFCFCFCFCFCF707FA07F80A0CFCFCFCFCFCFCFCFCF
CFCFCF3F5030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80A0CFCFCF303030CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF303040CFCFCF30406F30406F
30406FCFCFCF4F3F30CFCFCFCFCFCFCFCFCFC0C0C030406FCFCFCFCFCFCFC0C0C0CFCFCF4F403F\
CFCFCFC0C0C0CFCFCFCFCFCF50403FC0C0C0CFCFCFCFCFCFCFCFCF4F40304F7090
5F80906F8F7F5F80905F80905F8090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF60807F6F806F607F6F\
6F806F607F60CFCFCF608F8FCFCFCFC0C0C0CFCFCFCFCFCF6F8F7FC0C0C0CFCFCF
CFCFCFCFCFCF6F7F60CFCFCFCFCFCFCFCFCF60705FCFCFCFCFCFCFCFCFCFC0C0C06F705FCFCFCF\
CFCFCFC0C0C06F705FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF4F4F6FCFCFCF606F8FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF5F5F80C0C0C06F7090CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF5F8F80C0C0C0CFCFCFCFCFCF7080A0707FA0CFCFCF
CFCFCF3F5030C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF707FA02F2F302F2F2FCFCFCFCFCFCF\
CFCFCFC0C0C0403040CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303040CFCFCF30304F
CFCFCF30406F304F7030406FCFCFCFCFCFCFCFCFCF4F403F304F70CFCFCFCFCFCFCFCFCF504F3F\
CFCFCFCFCFCFCFCFCFCFCFCF4F403FCFCFCF4F70905070904F7090507090CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F80905F80905F80905F8090608F8F6F806FCFCFCFCFCFCF\
CFCFCFCFCFCF608F8FCFCFCFCFCFCFCFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCF
CFCFCF6F7F60CFCFCFCFCFCFCFCFCFCFCFCF6F7F60CFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCF\
CFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF4F5070CFCFCF6F6F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF5F5F806F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF5F8080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF707FA0
7F80A0707FA0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F80A0CFCFCF303030CFCFCFCFCFCF\
CFCFCF40304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F2F40CFCFCF
2F2F40CFCFCFC0C0C04F403030406F30406F30406F50403FC0C0C030406FC0C0C0CFCFCF4F4030\
CFCFCFC0C0C0CFCFCF4F6F8F4F70904F708FCFCFCFC0C0C0CFCFCF4F4030CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF608080CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF608F80CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F8070CFCFCFC0C0C0CFCFCF
C0C0C06F7F60C0C0C0CFCFCFC0C0C06F705FC0C0C0CFCFCFC0C0C0CFCFCF606F50CFCFCFC0C0C0\
CFCFCF606F50CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C04F4F7060608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C05F5F80606F90CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C05F8080C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C03F5030707FA07080A0C0C0C0CFCFCFC0C0C02F2F30C0C0C07080A02F2F2FCFCFCFC0C0C0\
403040C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F
50403030304FCFCFCFCFCFCF5040304F4030CFCFCF30406F3F4F7030406F304F70CFCFCF507090\
4F70905070904F7090CFCFCFCFCFCF504F3FCFCFCFCFCFCFCFCFCFCFCFCF4F403F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F8F608F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
608F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCFCFCFCF
70806FCFCFCFCFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCF\
CFCFCF6F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF4F5070606F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF505F7F6F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F8F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF3F5030CFCFCFCFCFCF7F80A0707FA0CFCFCF2F2F2FCFCFCF707FA0303030CFCFCF40304F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C030304F
30304030304F2F2F40CFCFCF4F3F30CFCFCF4F3F30CFCFCF4F4030CFCFCF30406F4F7090CFCFCF\
50403FC0C0C0CFCFCFCFCFCFCFCFCF4F4030CFCFCFCFCFCFCFCFCFC0C0C04F403F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F8F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
6F8F8FC0C0C0CFCFCFCFCFCFCFCFCF6F807FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
6F7F60CFCFCFC0C0C0CFCFCFCFCFCF6F7F5FC0C0C0CFCFCFCFCFCF6F705FC0C0C0CFCFCF607050\
6F705F6F7F6FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF4F4F6F6F6F8FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF607090CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F8080CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
3F5030CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF707FA07080A0CFCFCF7080A02F2F2F403040CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
30304F4F3F2F30304F30304FCFCFCF4F3F30CFCFCF4F403050403FCFCFCFCFCFCFCFCFCF305070\
4F403FCFCFCFCFCFCFCFCFCFCFCFCF504F3FCFCFCFCFCFCFCFCFCFCFCFCF4F403F
CFCFCFCFCFCFCFCFCF608F8F6F8F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F8F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF708F7FCFCFCFCFCFCFCFCFCFCFCFCF6F7F60
CFCFCFCFCFCFCFCFCFCFCFCF6F7F5FCFCFCFCFCFCFCFCFCFCFCFCF607050CFCFCF6F705FCFCFCF\
6F806FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF606F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6070906F709FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F8F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
3F5030CFCFCFCFCFCFCFCFCFCFCFCF2F2F2FCFCFCFCFCFCF7F80A0707FA07F80A0CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C030304F4F3F2F30304F2F2F404F4030C0C0C0CFCFCF4F3F3050403FC0C0C0CFCFCFC0C0C0\
304F70304F70CFCFCFC0C0C0CFCFCFC0C0C04F403FC0C0C0CFCFCFC0C0C04F403F
C0C0C0CFCFCF608F80CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F8F8FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C06F8F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F7F60
C0C0C0CFCFCFC0C0C0CFCFCF60705FCFCFCFC0C0C0CFCFCFC0C0C06F705F606F50CFCFCFC0C0C0\
6F8070C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0606F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C090AFBFC0C0C06F709FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F8F80C0C0C0CFCFCFC0C0C0CFCFCF
305030CFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF30304FCFCFCF30304FCFCFCF4F4030CFCFCFCFCFCF4F403FCFCFCFCFCFCFCFCFCF\
CFCFCF504F3F304F70CFCFCFCFCFCFCFCFCF4F403FCFCFCFCFCFCFCFCFCFCFCFCF
6F8F8F608F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F8FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCFCFCFCF6F7F60CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF6F7F60CFCFCFCFCFCFCFCFCF6F705F6F705FCFCFCFCFCFCFCFCFCF\
6F8070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF90AFB0CFCFCFCFCFCF6F709F6F7090CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F8080CFCFCFCFCFCFCFCFCFCFCFCF
3F5030CFCFCFCFCFCFCFCFCF303030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C030304FCFCFCF30304F4F3F30CFCFCFCFCFCFCFCFCF4F3F30CFCFCFCFCFCF\
CFCFCF4F4030CFCFCF304F70CFCFCFC0C0C04F403FCFCFCFCFCFCFC0C0C06F8F8F
4F403FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F8F8FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF6F807FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F7F60CFCFCF
CFCFCFCFCFCFC0C0C06F705FCFCFCFCFCFCFC0C0C06F8070608070CFCFCFC0C0C0CFCFCF6F806F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0606F8FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF90A0B0CFCFCFCFCFCFCFCFCF606F906F7090CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F8080CFCFCFCFCFCFC0C0C0CFCFCF
3F5030CFCFCFC0C0C02F2F30CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF30304FCFCFCF30304F4F4030CFCFCFCFCFCF4F40304F403FCFCFCF\
CFCFCFCFCFCF4F403FCFCFCF304F70CFCFCF4F403FCFCFCF608F8F6F8F8FCFCFCF
504F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F8FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF6F8F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF708060CFCFCF
CFCFCFCFCFCFCFCFCF6F705FCFCFCFCFCFCF6F807FCFCFCF6F8070CFCFCFCFCFCFCFCFCF6F8070\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F90CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9FAFBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F709F\
607090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F8F80CFCFCFCFCFCF3F5030
CFCFCFCFCFCF303030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C030304FC0C0C0305070304F70304F70304F70305070304F70\
304F70304F70305070304F70304F70304F70305070608080CFCFCFC0C0C0CFCFCF
C0C0C04F403FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F8F8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C06F8F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F7F60C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C06F7F5F607F70608070C0C0C06F8070C0C0C0CFCFCFC0C0C0CFCFCF6F7F6F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0606F8FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF90AFB0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF606F90CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F8080CFCFCF90AFB03F5030
C0C0C02F2F30C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F30304FCFCFCF30304FCFCFCFCFCFCF4F4030CFCFCF\
4F403FCFCFCF4F403FCFCFCFCFCFCFCFCFCF304F704060803F5F7FCFCFCFCFCFCF
CFCFCF4F403FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF608F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF6F8F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7F60CFCFCFCFCFCF
CFCFCFCFCFCF6F807F608070CFCFCFCFCFCFCFCFCF6F8070CFCFCFCFCFCFCFCFCF6F8070CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6F90CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF90AFB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF6F70906F709FCFCFCFCFCFCFCFCFCFCFCFCF5F8080CFCFCF3F5030
303030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF2F2F4F30304F2F2F4F30304FC0C0C0CFCFCF4F4030\
CFCFCF4F403050403FCFCFCFCFCFCFC0C0C0CFCFCF4F403FCFCFCF3F5F7F3F5F7F
CFCFCF4F403FC0C0C0CFCFCFCFCFCFCFCFCF608F80CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF6F807FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F7F60CFCFCFC0C0C0CFCFCF
6F7F606F7F605F7F7FCFCFCFCFCFCFCFCFCF6F7F6FCFCFCFCFCFCFCFCFCFC0C0C06F8070CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0606F8F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C09FAFBFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF6F709F606F90CFCFCF90AFB05F8F80305030CFCFCF
2F2F2FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304FCFCFCF30304F30304FCFCFCF4F4030\
CFCFCFCFCFCF4F403F504F3FCFCFCFCFCFCFCFCFCFCFCFCF4F403FCFCFCFCFCFCF
40607F3F5F7FCFCFCFCFCFCFCFCFCF608F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
6F8F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F7F60CFCFCF7080606F7F60
CFCFCF5F7F7FCFCFCFCFCFCFCFCFCFCFCFCF708070CFCFCFCFCFCFCFCFCFCFCFCF6F8070CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F6F90606F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF90AFB0CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F70906F709F5F80803F50302F2F2F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F2F404F40302F2F4F30304FC0C0C0\
4F4030C0C0C0CFCFCF4F4030CFCFCFC0C0C0CFCFCFC0C0C04F403FC0C0C0CFCFCF
C0C0C0CFCFCF3F5F7F3F607F608080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
6F8F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06080806F7F606F7F60C0C0C0CFCFCF
C0C0C060807FC0C0C0CFCFCFC0C0C06F8070C0C0C0CFCFCFC0C0C0CFCFCF6F7F6FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF60608FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF90A0B0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF90AFB0CFCFCFC0C0C06F70905F8080CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F4F403030304F30304F\
4F4030CFCFCFCFCFCFCFCFCF4F4030CFCFCFCFCFCFCFCFCF4F403FCFCFCFCFCFCF
4060803F5F7F40607F3F5F7FCFCFCF6F8F7F6F8F7F6F8F7F708F7F6F8F7F6F8F7F6F8F7F708F7F\
5F7F805F7F80507F805F7F805F7F805F7F8060807FCFCFCFCFCFCFCFCFCFCFCFCF
608080CFCFCFCFCFCFCFCFCFCFCFCF6F8070CFCFCFCFCFCFCFCFCFCFCFCF708070CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF606F8FCFCFCFCFCFCFCFCFCFCFCFCF9FAFBFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9FAFBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF2F2F4F4F4030CFCFCF\
30304F4F3F30CFCFCFCFCFCF4F4030403F30CFCFCF3F5F7F3F607F3F5F7F3F5F7F
CFCFCFCFCFCF4F4030CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F7F80CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F7F7FCFCFCFCFCFCFCFCFCFC0C0C060807F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F8070C0C0C0CFCFCFCFCFCF6F8070C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF606F8FCFCFCFC0C0C0CFCFCFCFCFCF9FAFBFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF9FAFBFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304FCFCFCF\
CFCFCF30304FCFCFCF40607F3F5F7F40607F3F5F7FCFCFCFCFCFCF504F3FCFCFCF
CFCFCFCFCFCFCFCFCF4F403FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF507F805F7F80CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF5F7F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60807F
CFCFCFCFCFCFCFCFCFCFCFCF6F8070CFCFCFCFCFCFCFCFCFCFCFCF6F806FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F8FCFCFCFCFCFCFCFCFCF90AFB0CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF9FAFBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F2F4F\
303050C0C0C030304FC0C0C0CFCFCF4F3F30CFCFCF403F30CFCFCF4F4030CFCFCF
C0C0C0CFCFCFC0C0C04F403FC0C0C0CFCFCFC0C0C0CFCFCF507F80CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF5F7F7FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F7F7FCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF6F7F6FCFCFCFC0C0C0CFCFCFC0C0C06F8070C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60608FCFCFCFC0C0C090AFBFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C09FAFBFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
4F4030303050CFCFCF5070804F6F80CFCFCF4F4030CFCFCF4F4030504F3FCFCFCF
CFCFCFCFCFCFCFCFCF4F403FCFCFCFCFCFCF5F7F80507F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF60807FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF608080CFCFCF
CFCFCFCFCFCFCFCFCF6F8070CFCFCFCFCFCFCFCFCFCFCFCF708070CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F8FCFCFCFCFCFCF9FAFBFCFCFCFCFCFCF\
CFCFCF9FAFBFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF4F3F303030502F2F4FCFCFCF4F6F804F4030CFCFCFCFCFCF403F304F403F
CFCFCFCFCFCFC0C0C04F403FCFCFCF5F7F80C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF5F7F7FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF608080C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C06F8070CFCFCFCFCFCFC0C0C0CFCFCF6F806FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF606F8FCFCFCF90A0B0CFCFCFCFCFCF\
9FAFBFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF4F4030CFCFCF2F2F4F3030504F6F804F4030CFCFCFCFCFCF4F4030
CFCFCFCFCFCFCFCFCF5F7F805F7F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
5F7F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF608080CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF708070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8070CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F8F9FAFBFCFCFCF9FAFBF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF4F3F304F4030C0C0C03030504F6F7F4F4030C0C0C0CFCFCF
403F2FCFCFCF507F80CFCFCF4F4030CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
60807FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F7F7FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF6F7F6FCFCFCFC0C0C0CFCFCFC0C0C06F8070C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60608F9FAFBFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF4F40304F4030CFCFCFCFCFCF4F6F804F7080CFCFCF
5F7F805F7F80CFCFCFCFCFCF504F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF608080\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60807FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF6F8070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF4F3F30CFCFCFCFCFCFCFCFCF2F2F4F30304F
CFCFCF4F4030403F30CFCFCFCFCFCF4F403FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C060807FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C060807FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF6F8070C0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F8070CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303050CFCFCFCFCFCF4F4030CFCFCF
3030502F2F4FCFCFCF4F4030CFCFCF4F403FCFCFCFCFCFCFCFCFCFCFCFCF608080CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF608080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F8070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF708070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF2F2F4F3030502F2F4F4F4030
C0C0C0CFCFCF2F2F4FCFCFCF403F2F4F403FC0C0C0CFCFCFC0C0C060807FC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C060807FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
6F7F6FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F7F6FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303F5FCFCFCF30304F
30305030305030305030304F3030504F4030CFCFCFCFCFCF608080CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF60807FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
708070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF708070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF2F3F50CFCFCF
4F3F30CFCFCFC0C0C03030502F2F4F303050403F3060807FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF5F7F7FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F8070
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F8070C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303F5F
4F4030CFCFCFCFCFCFCFCFCF4F4030CFCFCF303050303F5F30405FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF608080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8070
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F806FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
2F3F504F4030C0C0C0CFCFCF403F2FCFCFCFC0C0C04F4030C0C0C0303F5F2F3F50CFCFCFC0C0C0\
CFCFCFC0C0C0608080C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F7F6FCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F8070C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF303F5FCFCFCFCFCFCFCFCFCF4F4030CFCFCFCFCFCF4F4030CFCFCFCFCFCF2F3F5F30405F\
CFCFCF608080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF708070CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF708070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF403F30303F5F2F3F50303F5F2F3F50303F5F2F3F50303F5F2F3F50303F5F2F3F50\
303F5F2F3F50304F602F4F60CFCFCFC0C0C0CFCFCFCFCFCF6F8070C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF6F806FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF4F4030CFCFCFCFCFCF4F4030CFCFCFCFCFCF403F30CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF2F4F60305060304F60304F606F806FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF6F8070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF403F30CFCFCFC0C0C04F4030C0C0C0CFCFCF403F30CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0304F602F4F606F80706F7F6F6F8070
6F7F6F6F80706F7F6F6F8070C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4030CFCFCF403F30CFCFCFCFCFCF4F4030CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF30506F304F60305060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF405F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF403F304F4030403F30CFCFCFC0C0C04F4030CFCFCF\
CFCFCF2F4F60304F60CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF405F70C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4030CFCFCFCFCFCFCFCFCF304F60\
2F4F60CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F5F6F
405F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C04F40302F4F60304F60403F2F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F5F6FCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4030CFCFCFCFCFCF\
4F4030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F5F6FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C04F4030CFCFCF\
CFCFCF403F30CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF3F5F6FCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4030\
403F304F4030CFCFCFCFCFCFCFCFCFCFCFCF3F5F6FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF403F304F4030C0C0C03F5F6F3F5F6FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF403F30406070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[17]:=",
  ImageSize->{288, 286.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P00017PP10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7[`N_1jl72PL000L7
ojlA[`T700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`N_1`X700071on_4Jl91`00
1`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7[`N_1jl7[`L:1`001`Oo[a6_2@L000L70?n_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7[`N_1jl72PL000L7ojlA[`T700071`3o[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_000@[`N_1jl7[`N_1jl7[`N_1`X700071on_4Jl91`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004:l7
[`N_1jl7[`N_1jl7[`L:1`001`Oo[a6_2@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7
[`N_1jl72PL000L7ojlA[`T700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`N_1`X7
00071on_4Jl91`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7[`N_1jl7[`L:1`001`Oo[a6_
2@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7[`N_1jl72PL000L7ojlA[`T700071`3o
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`N_1`X700071l6_0ZY=[`T700071`32[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jZ_ZPL3[`1:1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`Nm[`03ZZn_00>_00>Z[jl0BZl91`001`L0^Zl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1`:_00JZ1jl7[`L3[`1:1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`Ni[`:Z2:l00jZ_[`1:[`T700071`2e
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
ZP02[`1H1jl7[`N_1jl7[`N_1jX7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071k>_00>Z[jl03Zl00jZ_[`19[`T7
00071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`NZ[jX7[`N_1jl7[`N_1jl7[`N_1jX7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`N_1`X700071jj_0ZXE[`03ZZn_04R_
2@L000L70:f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
ZP02[`0E1jX7[`N_1jl7[`N_1jl7[`N_1jl700>_04P7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`NZ[`:Z1:l00jZ_
[`0B[`03ZZn_04R_2@L000L70:V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`NZ00:_01T7[`N_1jX7[`N_1jl7[`N_1jl7[`N_1jl700>_04P7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`00
1`NV[`:Z2:l00jZ_[`0C[`03ZZn_04N_2@L000L70:F_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jX00Zl02@N_1jl7[`N_1`03[`1L1jl7[`N_1jl7[`N_1jl7[`N_1jX7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl:1`001`NQ[`03TJZZ00f_00>Z[jl04Zl00jZ_[`17[`T700071`2O[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`NZ00:_00l7[`N_1jl7[`N_1jl7[`L00jl04`N_1jl7[`N_
1jl7[`N_1jl7[`L00jl0APN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`NL[`03ZY6_01>_00>Z[jl04Zl00jZ_[`16[`T7
00071`2H[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L2[`0/TJn_1jX7[`N_1jl7[`N_1jl7
[`N_1jl7ZPN_1jl7[`N_1jl7[`N_1jl7[`L3[`161jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071iN_00>@ZZl01:l00jZ_
[`0A[`03ZZn_01:_00>Z[jl0AZl91`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jX7
[`N_1jl7ZPN_1jl7[`N_1jl7[`N_1jl7[`NZ1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7[`N_1jl7
[`L:1`001`ND[`03ZY6_00N_00>Z[jl04Jl00jZ_[`1K[`T700071`2@[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl70Zl03:Z_[`N_1jl7[`N_1`>_0147[`N_1jl7[`N_1jl7[`N_1`03[`1J1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7Sjl00i6Z[`0=[`03TJn_016_00>Z[jl04Zl00jZ_[`15[`T7
00071`2<[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L2[`0@ZPN_1jl7[`N_1jl7[`N_1`>_0147[`N_1jl7[`N_
1jl7[`N_1`03[`1J1jl7[`N_1jl7[`N_1jl7[`N_1jX7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7Rjl00i2Z[`0A[`03
TJn_016_00>Z[jl04Zl00jZ_[`15[`T700071`38[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_ZY47[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jX7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L3[`141jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl:1`001`N5[`03TJZ_01R_00>Z[jl09Zl00jZ_[`14[`T700071`22[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`:_04BA1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7ZPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1`>_04@7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`X700071h6_00>@ZZl07:l00jZ_[`0V[`03ZZn_04B_2@L000L707j_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`L2[`0<T@N_1jl7[`N_1jl70jl0P0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`NZ
1jl7[`N_1jl7[`N_1jl7[`N_1jl7ZPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7OJl00i6Z[`0=[`03T:n_082_2@L000L7
08b_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1i6ZT@N_1jl7[`N_1jl7[`N_1`>_0807[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7ZPN_1jl7[`N_1jl7[`N_1jl7[`N_1jX7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071gV_00>A[i004Jl00i2_[`0U[`03
ZZn_05R_2@L000L70:2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[jZA1jl7T@N_1jl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_1jl70jl04@N_
1jl7[`N_1jl7[`N_1jl700>_01<7[`N_1jl7[`N_1jl7[`N_1jl700>_00d7[`N_1jl7[`N_1jl700>_
0387[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071gB_
0i45[`03ZZn_00n_00>Z[jl04:l00jZ_[`0A[`03ZZn_01>_00>Z[jl03:l00iF@[`0c[`T700071`1a
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1i400Zl0;@N_
1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_1`03[`0W1jl7[`N_1jl7[`N_
1jl7[`N_1jX7[`N_1jl7[`N_1jl7[`N_1jl700>_0487[`N_1jl7[`N_1i07[g6@1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`M_[`03T:n_00Z_00>Z[jl0
3jl00jZ_[`0@[`03ZZn_01:_00>Z[jl04jl00jZ_[`08[`07T:n_[`2_T00b[`T700071`1l[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7TJZA1jl7[`N_1jl7[`N_1`>_
00l7[`N_1jl7[`N_1jl7[`L00jl0B`N_1jl7[`N_1jl7[`N_1jl7ZPN_1jl7[`N_1jl7[`N_1jl7[`NZ
1jl7[`N_1jl7[`N_1jl7[`N_1jl7ZPN_1jl7[`N_1i07[`N_L@03[`0`1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7JZl00i6@T@0@[`03TJn_00n_00>A[jl0
4:l00jZ_[`0A[`03ZZn_01>_00>Z[jl01Zl00i2_[`03[`040:n_LC6_2@L000L70;J_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1i6Z[`NA1jl7[`N_1jl7[`N_1jl7[jZ_1jl7
[`N_1jl7[`N_1jl7[jZ_1jl7[`N_1jl7[`N_1jl7[`NZ1jl7[`N_1jl7[`N_1jl7[`N_1`>_05H7[`N_
1jl7[`N_1jl7[`N_1jl7ZPN_1jl7[`N@1jl7[`N_LI2_T0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071fJ_00>AT:l00jl00i6_[`0>[`03TJn_00n_00>A[jl0
4:l00jZ_[`0B[`03ZZn_01>_00>Z[jl00jl00i2_[`02[`04TI2A00>_00>@[jl0;Zl91`001`L0XZl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`NAZZl7[`N_1i47[`N_1jl7[`N_1jl7
[`N_1i47[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_1`>_0147[`N_1jl7[`N_1jl7[`N_
1`03[`0C1jl7[`N_1jl7[`N_1jl7[`N_1`03[`051jl7T0L00Zl0292_[`N_LJl70jl0;PN_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`MR[`03TI2_00R_00>@[jl0
3Zl00jZ_[`0?[`03ZZn_012_00>Z[jl04Jl00jZ_[`0C[`08ZZn_[i2AT946[`050:n_[i00<:l91`00
1`L0XZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1i6Z[`N_1jl7[`N_1jnZ[`N_1jl7
[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_1`>_02/7[`N_1jl7[`N_
1jl7[`N_1jl7ZPN_1jl7[`N_1jl7[`N_1jl7[`N_1jX700>_03`7[`N_1jl7[g6_1jl7T0N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`MM[`03T96@00f_00>A[jl0
3Zl00jZ_[`0?[`03ZZn_012_00>Z[jl04Zl00jZ_[`0C[`03ZZnE00Z_00<0[jl00Zl00i6_[`0][`T7
00071`1n[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7TJZ_1jl7[`N_1jl7[`N_1jl7[`NA1jl7
[`N_1jl7[`N_1jl70jl03`N_1jl7[`N_1jl7[`N_1`03[`0m1jl7[`N_1jl7[`N_1jl7[`NZ1jl7[`N_
1jl7[`N_1jl7[`N_1jX7[`N_1jl7[`N_1jl7[`N_1jl7[g6ZT003[`0l1jl7[`N_1jma[`N_1i07[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7F:l00i6@[`02[`03
T:n_00j_00>Z[jl03Zl00i2_[`0?[`03T:n_012_00>Z[jl09Jl0176_T:X2[`03TJn_00J_00=[[jl0
0jl00i2_[`0/[`T700071`1n[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[jZAZZl7[`N_TJl7[`N_1jl7
[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl70jl03`N_1jl7[`N_1jl7[`N_1`03[`0U1jl7[`N_1jl7[`N_
1jl7[`NZ1jl7[`N_1jl7[`N_1jl7[`N_1`03[`0C1jl7[`N_1jl7[`N_1jl7076_T003[`091i07[`N_
1jl7009a00B_1jl70jl0;0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
2PL000L7Djl00i2A[`07[`03T:n_00j_00>Z[jl03Zl00i2_[`0?[`03T:n_012_00>Z[jl04Zl00jZ_
[`0=[`040:n_T0:_00BZ[jnA1jl00g5[[`04[`03T:n_02b_2@L000L707j_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7T@N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`L3[`0?1jl7[`N_
1jl7[`N_1jl700>_0147[`N_1jl7[`N_1jl7[`N_1`03[`0A1jl7[`N_1jl7[`N_1jl7[`L00jl04`N_
1jl7[`N_1jl7[`L01jl7MPL00jl0?PN_[jl7[`N_LJl7K0N_1jl7T0N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7Cjl00i2AT00;[`03T:n_00n_00>@[jl03Zl00jZ_
[`0?[`03ZZn_012_00>A[jl04Jl00jZ_[`0;[`050:n_[g400jl01@2_[jn@00B_00Ea[jn_L@05[`03
TJn_02^_2@L000L704f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02T@1E1jl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_
1jl7[`N_1jnZ[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_
1`03[`0[1jl7[`N_1jl7[`N_1jl7[`N_1jX7[`N_1jl7[`N_1`07[`Mf1jl7[g6Z1`03[`0j1jma[`N_
1f`7[`N_1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071d^_
00B@TJnA3jl00i6_[`0>[`03T:n_00j_00>Z[jl03jl00jZ_[`0@[`03ZZn_01f_00D0[jn_L@05[`03
0:n_00:_00BA[jm/1:l00g6_[`04[`03T:n_02Z_2@L000L704V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00:A07X7[`N_
TJl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_1jnZ[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_
1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_1jl7ZPN_1jl7[`N_1jl7[`N_1jl7[`NZ1jl7[`N_1jl700:_
01/7[i2_1jl7[g6Z1jl7T0Ma1jl7[`M/1jl7[`L00jl0:PN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`X700071dN_00>@TJl01Jl00i2_[`0<[`03TJn_022_00>@[jl03jl00i2_
[`0@[`03ZZn_01Z_00H0LJnF[hl6[`030:n_00>_00=/[jl01:l00g6_[`04[`03T:n_02Z_2@L000L7
04>_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00:A03/7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[jZ_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1`03[`0?1jl7[`N_1jl7[`N_1jl700>_02D7[`N_1jl7[`N_1jl7[`N_1jX7[`N_
1jl7[`N_1jl7[`N_1jl700>_01P7[`N_LJl700N_T:l7/;F`]Jma[`N_1f`2[`0h1jl7[`Ma1jl7[`N_
1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`M1[`03T96_00b_
00>A[jl03:l00i2_[`0O[`03TJn_00n_00>A[jl04:l00jZ_[`0B[`03ZZn_00>_00<0[`000jl02IJ`
[iJ@UY2F0002UP05[f^_[i401Jl00g6_[`05[`03TJn_02V_2@L000L703n_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00Y40I@N_1jl7[`N_
1jl7[`N_1i47[`N_1jl7[`N_1jl7[jZ_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jnZ[`N_
1jl7[`N_1jl7[`N_1jnZ[`N_1jl7[`N_1jl7[`N_1jl700>_0147[`N_1jl7[`N_1jl7[`N_1`03[`1B
1jl700N`]K2e/:nA1jl7[`N_LJneK:nA[i2Z[`N_1jma[`N_1jl7T0N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071cf_00B@TJnA3jl00i6_[`0=[`03TJn_00f_00>Z[jl0
3jl00jZ_[`0?[`03ZZn_012_00>Z[jl04Zl02ZZ_/9J`UZl0TG42[`03TJn_00B_00<0[iH01:l01I6_
T8^A00:_00=[[jl0<:l91`001`L0>jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_00:A06T7[`NA1jl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_
1i47[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_
1jl7[`N_1jl7[`L00jl07`N_1jl7[`N_1jl7[`N_1jl7[k2Z]@07[`N@LGH7[`L00jl0A0N_1jma[kF_
1jl7[jZ_1jnA[g6_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
2PL000L7>:l00i6@T@06[`03T:n_00f_00>@[jl03:l00i6_[`0=[`03ZZn_00n_00>Z[jl03jl00jZ_
[`0@[`03ZZn_00b_00N`[k2F/:n`00>_00FAT:l0MP04[`03T:n_00>_00@0[jn`1:l00h^_[`02[`03
T6^_036_2@L000L703J_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`L2T@1k[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_1i47[`N_1jl7
[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7
[`N_1jX7[`N_1jl7[`N_00>`00V_1jl7[`N`LI000Zl01`L0T:l7[`L00jl010N_1jl2L@0C1k07[`N_
ZZl7[`N_LI2AT0N_1`03[`0X1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X7
00071cF_00>@[jl02Jl00i2_[`0=[`03T:n_00f_00>@[jl03Jl00i2_[`0>[`03ZZn_012_00>A[jl0
4:l00jZ_[`04[`04U[2F/0n_00=a[jl01Jl00i6_[`02[`04Jjn_/0B_00>;[jl00jl016^_[jl2T0:_
00>@[jl0::l91`001`L0<jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`02T@0k1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_1jnZ[`N_
1jl7[`N_1jl7[`L00jl03`N_1jl7[`N_1jl7[`N_1`03[`0?1jl7[`N_1jl7[`N_1jl700>_01<7[`N_
1jl7[`N_1jl7[`N_1jX700F`03H7[`N_1jl7[`N_1jl7JkF_1jmaK0N_1jl7[`NA1jl7[g6_1jne[`N_
1i07[`N_LJl7[`N_TI02[`0X1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X7
00071c2_00BAT96@3Zl00i6_[`0=[`03TJn_00b_00>@[jl03Jl00i2_[`0?[`03TJn_00n_00>A[jl0
3jl00i6`[`0A[`030:n_00>_00<0LJl01Zl01Y6_[jmaJ`>_00>F[jl00Zl00i6_[`02[`03Jjn_00F_
00>;TJl0::l91`001`L0PZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
T96_1jl7[`N@1jl7[`N_1jl7[`N_1i07[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jnA[`N_1jl7
[`N_1jl7[`N_ZZl7[`N_1jl7[`N_1jl7[`L3[`0?1jl7[`N_1jl7[`N_1jl700>_00L7[`N_1jl700:_
02ZZ[jl7[`N_1jn`[`N_1jl7[`N_1jl7[`N_LJl7[`NALJl7K0N_1jl7[`L3[`12LJma[`N_1k07[`N_
TJl7[`M/1jl7[`N_1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7
:jl00i2_[`06[`03TJn_00b_00>@[jl03:l00i6_[`0=[`03TJn_00f_00>A[jl08Jl02JZ_[jnZTJZA
ZP0:[`03UZn_00b_00=a0:l00jl01I40[jma00R_00F@LJn_J`04[`03/:n_00:_00>@[jl00Zl00g6_
[`05[`03T:n_02N_2@L000L702V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
T002T@1^1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1i47[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`NA[jX2[`0T
1jl7[`N_1jl7[`N_1jn`[`N_1jl7[`N_1jl7[g401jl7[`M[0Zl0CPM/1jl7[`N_1jmaT0N_LJl7[`N`
1jl7[i6_1jl7K0N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X7
00071bJ_00>AT:l03:l00i2_[`0;[`03T:n_01b_00>A[jl03Jl00i6_[`0M[`05ZY2ZT:X07jl3L@F_
00F`T:n_L@07[`03LJnA00:_00=[[jl00Zl00k2_[`02[`03T:n_00:_00=a[jl01Zl00i2_[`0V[`T7
00071`2A[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7T96@1jl7[`N_1jl7[`N_1jnA
[`N_1jl7[`N_1jl7[`N@1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7
T@N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1i6_T@02[`1l1jl7UPN_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[g6_MP07[`N_1k07T@M/1jl7[`N_1f`7T0N_LJl7[`N_]Jl7[`N@1jl7K0N_1jl7[`N_
TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071b:_00FAMjn_T@0M[`03
TJn_01b_00>@[jl03Jl00i6_[`0D[`04ZY6ZT2J_00Ea[gJ_L@04[`060;2_TG9a1Zl01`2_[jn@[f/0
2Zl01I2_[jma00R_00>@[jl09Zl91`001`L0c:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
T96_1jl7[i6_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7T0N_1jl7[`N_1jl7[`N_TJl7[`N_1jl7
[`N_1jl7[i6_1jl7[`N_1jl7[`N_1jnZ[`N_1jl7[`N_1jl7[`N_1i47TJZAZZl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`1f[`L01jl7[g6@]JnAK96@1jl7K0N_1`>_041a[`N_
1jne[`N_1i07[`M/1jl7[`N_1jl7T0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
2PL000L77Zl00i5g[`06[`03Mjn_01Z_00>A[jl03:l00i2_[`0=[`03T:n_00f_00>A[jl03:l4TBB_
00>`[jl02:l0176_[g43T003LJn_00:_00P0[jn_LY1bL@B_00>@[f/01Zl00k2_[`05[`03LJn_00N_
00>@[jl09Jl91`001`L0/jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N@TJl7[`N_1jl7[`N@1jl7
[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[i6_1jl7
[`N_1jl7[`N_1jnZ[`N_1jl7[`N_/96`TK2_1i47[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7/0N_1jl7[`N_009a05X7T76@1jma[`N_]Jma[`N_1jl7T96_1jl7T76_1jl7[`N`1jl7[`N_
LF`7[`N_1jl7[`N@1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`LI[`03
LY5g00b_00>A[jl02jl00i6_[`0<[`03T:n_00^_00>@[jl03Zl00i6_[`0<[`03TJn_00B_0i4J[`03
TJn_012_00>`[jl01Jl2L@04MY2_00>_00<0[jl00jl00`2@[`03[`090:n_T7:@[i5[00J_00>`[jl0
0jl0176_[f/9[`03T:n_02B_2@L000L70;6_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N@TJl7[`N_1jl7
[`N_1jl7[i6_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7
[`N_1jl7T@N_1jl7[`N_1jl7[`N_1i6`TK2A/:l7[`N_1jl7[`N_/:l7[`N_1jl7[`N_1jl7[`N_/:l7
[`N_1jl7[`N_1jl7[`N_1jn`[`N_1jl00W40G0N_UJma[`L01jl7[`N`LI07[`M/1jl7[`N_TI1a[`N_
1jl7[kF_1jma[`N_LJl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
2PL000L75:l00i5b[`03[`03TJn_00b_00=b[jl02jl00i2_[`0<[`03TJn_00^_00>A[jl03Jl00i6_
[`06[`>A1:l00i6_[`0>[`03TJn_00n_00>A[jl04Jl00iJ_[`02[`04LJmaL@:_00B@0:l01Zl01[00
[g:_00V_00=[[jl01Jl01IJ_[g6@00>_00=[[jl01jl00h^_[`0T[`T700071`2a[`N_1jl7[`N_1jl7
[`N_1jnAMi6_1jl7[`N_TJl7[`N_1jl7[`N_1jl7T0N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_1i47
[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_1i47[`NAUY6_T@N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7
[k2_1jl7[`N_1jl7[`N_1jl7[`NF1jl7[`N_1jl7[`N_1jl7[`N_/:l7[g40009a05`7[`MaUF/7[`N_
1jl7[kF_TF`7[`N_1jl7[`N_LJnA[`N_1jl7/76_TJl7[g6_1jl7[`N_1jl7T0N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071`n_00=bTJl02Zl00i6_[`0;[`03TJn_00^_00>A[jl0
2jl00i6_[`0<[`03T:n_00b_00BAT96A3Jl00i6_[`0>[`03/:n_00n_00>`[jl04Jl01;2_LJl2L@B_
00=a0:l02:l00iJ_000;[`05Jjn_T7801:l01K2_[jn@00:_00=/[jl02:l00i6_[`0S[`T700071`3o
[`N_1jl7[`N_1jl7[`Mg1jl7[`N_1jl7[`N_1i07[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N@1jl7
[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jnZTJZAZZnA[`N_1jl7[`N_1jl7[`N_/:l7
[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`NF1jl7[`N_1jl7[`N_1jl7[`N_1k070740MZl7
[`M/LI07[`N_1jl7[kF_LI07[`N_1jl7[`N_LI07[`N@1jma/0N_1i07[g6_1jl7[`N_1jl7T0N_1jl7
[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`Mg1`X700071`^_00=bT7803jl00gN_[`0;[`03
Mjn_00Z_00>A[jl03:l00i2_[`0;[`03T:n_00>_00B@TI2A1Zl00i6_[`0=[`03TJn_00j_00>`[jl0
4:l00i6_[`0@[`05/02_07H01Jl00g6_T008[`040;2_T0Z_00=[T:l00jl0179a[k03[`04T:n_J`^_
00>;[jl08Zl91`001`L0I:l7[`N_1jl7[i5g1jl7[`N_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jnA
[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7TJZAZY47[`N_1jl7[`L3[`1>
1jl7[`N_1jl7[`N_1jn`[`N_1jl7[`N_1jl7[`N_1jn`[`N_1jl7[`N_1jl7[`N_1jn`[`N_1jl7[`N_
1jl7[`N_1jl7[k00MZl7[`N_0W40F0N_UJl7[`N_1jma/0N@1jl7[`N_1jl7[g6@1jl7[g6@TJne[`N@
1jma[`N_1jl7[`N_1jnZ[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`L7[`03
LY2_00F_00=b[jl03:l00i6_[`0H[`03T:n_00b_00>A[jl02:l0192AT94O[`03TJn_00j_00>A[jl0
3jl00i6_[`0@[`04LIIaL@F_00D0[jn_M`07[`05Jk2_[g802Jl00f^_[`02[`06LJn_[i2F0jl00i2_
K00;[`03Rjn_02:_2@L000L704j_1jl7[i5g1jl7[`N_1jl7M`N_1jl7[`N_1jl7[`N@1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`NA1jl7[`L5T@1Q1jl7T@N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[k2_1jl7[`N_1jl7[`N_1jl7[k2_1jl7[`N_
1jl7[`N_1jl7[g40/747[`N_1`02L@1IT9F@UI2A[`N_1faa/0N_TJl7[`N_1jl7[g6_1jl7L@N_1jne
[`N_TJma[`N_1jl7[`N_1jnZ[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl02PL000L7
0jl00g:A[`0:[`03TJn_00b_00=g[jl02Zl00i6_[`0;[`03TJn_00^_00JAT96@TI08[`03TJn_01f_
00>A[jl03Zl00i6_[`0@[`03/:n_00j_00<0[iH00jl01GN@Mi0000B_00b@[i6@[`2_0:nF[g89[`05
Jjn_[f/01Zl01[2_[i2_J`b_00>A[jl08Jl91`001`L0AGNAMi6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7
[`N_1jnA[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N@1jl7[`N_1jl7[`04T@1S1jl7[`N_1jl7[`N_
1jnA[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[k2_1jl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_
1jl7[`N_1jl7UPN_1jl7[`N_1jl7[`N_1`2@009a066`UI07[`N_LJl7[`N_TJl7K0M/1jne[i6_1jl7
[`N_1jma[`Ma1jl7[`N_1k2A[i6_LJl7[`N_1jl7[`N_1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00X700071`05Mg:_[gL03Jl00i6_[`0;[`03TJn_00Z_00=g[jl02jl01I2AMi5b01N_
00>A[jl03:l00i6_[`0>[`03TJn_022_00>`[jl03jl01740/7H3[`030:l000F_00>@[jl00Zl01g6@
[jn`[i002:l00f^_R`0;[`04LY6_L@b_00>@[jl08:l91`001`L0>GMh[`N_1jl7[`N_1jl7[`N_1jnA
[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[i6_1jl7[`N_1jnAT003T@3C[`N_1jl7[`N_1jl7[i6_1jl7
[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_1jn`[`N_1jl7[`N_1jl7[`N_1i47
[`N_1jl7[`N_1jl7[`N_1jn`[`N_1jl7[`N_1jl7[g6_LF`7/0N@1f`7[g6_1jl7[i6_LJl7L@N@TK07
T0N_1jl7[`N_LFbA[`N_1jl7[`N_1jl7T95/1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`0:1`001`L00gMb[`02[`03LZn_00b_00>A[jl02Zl00gN_[`0:[`09TJn_[i5g
TGNA00F_00>A[jl02Zl00gN_[`0<[`03TJn_00b_00>A[jl03Zl00i6_[`0>[`03TJn_00n_00>A[jl0
2Zl01GH0[`1a00B_00@0[jma1jl01G6_[jma00>_00B`T:mb1jl00f^_R`0=[`03T76_00^_00>@[jl0
8:l91`001`L0ogNA[`N_1gL7[`N_1jl7[`N_1jl7M`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[i6@TI2A
[`N_1jl7[`N_1gL7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jnA[`N_1jl7
[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_/:l7[`N_1jl7[`N_1jl7[`N_/:l7[`N_1jl7[`N_1`07
K0M/1jn`[g6_1f`7[`N_1jl7K0N_1g47[`N`1i2A[`N_1jl7K76_1i07[`N_1jl7[`N_1jnAK96_1jl7
[`N_1jl7[i6_1jl7[`N_1jl7[`N_1jl7[`0047NA[`N_1gL7[`N_1jl7[`L:1`001`L00gMb[`04[`03
LZn_00^_00=b[jl04Jl0179gLY42[`03LZn_00Z_00>A[jl02Zl00g:_[`0<[`03TJn_00b_00>A[jl0
3Zl00i6_[`0>[`03TJn_012_00>`[jl02:l01P2_[g6_L@>_00Ea[i2_L@06[`03096_00:_00=a[jl0
1:l00g:_[`03[`03Jjm[00:_00>@[jl01Zl029J_[jn@LJnA;Jl91`001`L0ogMh[`N_1jl7M`N_1jl7
[`N_1jl7[`Mg1jl7[`N_1jl7[`N@1gNAMi5g1jl7[`N_TJl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7
T0N_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7
[`N_/:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LJn@[g6_1jma/0N@1f`7[`N_1f`7[i6_1g47
[`N_]Jl7LPN_1f`7[g6_1i07[`N_1jl7[kF_1jnAK0N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`0047Mh[`N_1jl7M`N_1jl7[`L:1`001`L00gMb[`06[`03LZn_00^_00=g[jl01Zl017MbT78<[`03
TJn_00Z_00=b[jl02Zl00i6_[`0<[`03TJn_00b_00>A[jl03Zl00i6_[`0P[`03/:n_00N_00If[jl0
[`03[`060;2_[gIa1:l01@2_[jn@00:_00=a[jl01Zl00g:_J`03[`04Jjn_T0f_00Ea[jn_T@0[[`T7
00071`3oMi6_1jl7[`N_1gL7[`N_1jl7[`N_1jnA[`N_TG9hLY6_1jnA[`N_1jl7[`N_1jl7LPN_1jl7
[`N_1jl7[i6_1jl7[`N_1jl7[`N@1jl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7
[`N_1jl7[k2_1jl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_1jl7UPN_1jl7[`N_1`07T76_LJl7
K0N`UY2FK0N_1jma[`N_TJl7L@N_1jl7/0N_1f`7[`N_LJl7[i6_1jl7/;F`1gMl[`M/1jl7[`N_1jl7
[`N_1i07[`N_1jl7[`N_1jl7[`N_000@Mi6_1jl7[`N_1gL7[`N_1`X700071`03Mg:_00R_00=b[jl0
2:l01I5bMg9g00V_00>@[jl02Jl00i6_[`0;[`03TJn_00Z_00=g[jl03:l00i6_[`0;[`03TJn_00j_
00>A[jl08Jl00i6_[`05[`040:mfL@:_00H0[jmaT;03[`05LG:_[g401jl00g6_[`03[`06/:n_Jjmb
0jl026^_[jn;/:n`1jl00gMfL@04[`03T:n_00J_00>A[jl07jl91`001`L0ogMh[`N_1jl7[`N_1gL7
[`N_TGNAMi6_1jl7[`N_1jl7[`N_1gL7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7M`N_1jl7[`N_1jl7
[i6_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_1jn`[`N_1jl7[`N_1jl7
[`N_1i47[`N_1jl7[`N_1jl7[`N_1jn`[`N_1jl7075g1`07K9J@LJl7[k2_1jl7LW6_1jl7[`N_LJl7
[`N_1k07K0N_1g87K0N_]K2e[`N_1jl7[`N_1jl7M`N_1jl7T0N_1jl7[`N@1jl7[`N_1jl7[`N_1jl7
[`0047Mh[`N_1jl7[`N_1gL7[`L:1`001`L00gMb[`05[`07Mg9gLWN_LP0:[`03LZn_00V_00=b[jl0
2Zl00gN_[`0:[`03LZn_00Z_00>A[jl02jl00i6_[`0<[`03TJn_01n_00>A[jl03jl00i6_[`03[`05
075fT0000Y000g6_L@04[`06UZn_T:ma0Zl00g:@[`02[`03T@2_00J_00BF/:n`0jl00g6_[`03[`03
T:n_00F_0WL3[`03T:n_00F_00>;[jl00jl00i2_[`0N[`T700071`3o[`MgTGNAMi6_1jl7[`N_1jl7
[`N_1jl7[`Mb1jl7[`N_1jl7[`Mb1jl7[`N_1jl7[`Mg1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1i07
[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[k2_1jl7[`N_1jl7[`N_1jl7[`NF1jl7[gI/1g47MPL0LJl7[`N_1k07T76_1jl7[i2_1i1a[`N_1jnF
MkF;1jl7[`MaT:l7[`N@1jl7[`N_O:l7[`N_TJl7[`N_1jl7T0N_1jnA[`N_1jl7[`N_1jl7[`N_000@
[`MgTGNAMi6_1jl7[`N_1`X700071`B_00=b[jl02:l00dj_[`08[`03Mjn_00V_00>@[jl02Zl00g:_
[`0F[`03TJn_00b_00>A[jl03:l00i6_[`0a[`04/:n_[`=a00<0[g400Zl00g6_[`05[`04/:maT0J_
00Z@LY5[UWNFMiH00jl00g:_[`02[`03LJn@00>_00>@[jl00jl00gN_[`05[`03TJn_00J_00F;[jn_
T00P[`T700071`2R[`N_1jl7[`N_1jl7[`N_1g87[`N_1jl7[`N_Mjl7[`N_1jl7[`N_TJl7[`N_1jl7
[`N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_1i47
[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`NF1jl70W40Jjma
MPN_1`2F[`N_1jl7/0MaUZl7[`N_1iFFT9J_1jl7K0N_1gN@[`N_1g47[`N_1jnA[`MgUJl7[`N_1jl7
T0N_1jl7[`N_1i07[i6_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00X700071`F_00=>LZl0
2Jl00dj_[`08[`03Mjn_00R_00=b[jl02Jl00g:_[`0G[`03LZn_00b_00>A[jl02jl00i6_[`0=[`03
TJn_00n_00>`[jl03jl02K2_[`1f076_T005[`03TJn_00>_00Z`[jn@[gNFMiIg0Zl00i00[`02[`03
0:n_00:_00=gLZl00jl0176_[g83[`03LZmf00Z_00>A[jl02:l018^_[i0O[`T700071`2R[`N_1jl7
[gN_1jl7[`N_1jl7LPN_1jl7[`N_1g87[`N_1jl7[`N_1gL7[`N_1jl7[`N_1gL7[`N_1jl7[`N_1gL7
[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`NA1jl7
[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`N_/:mf0g401i2_1jl7[`L00Y00HJl7[`N_UYFF
U@N_1jl7[`N_LG87K0N_1jl7[gb_1jl7L@N_1h/7[iEg1jl7[`N_1jl7[`N@1jl7[`N_1jl7[`N_1i07
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl02PL000L72:l00g:_[`07[`03LZn_00R_00=b[jl0
2:l00i2_[`09[`03Mjn_00V_00>A[jl02jl00i6_[`0;[`03TJn_00b_00>A[jl03Jl00i6_[`0Q[`06
TJmaL@2?1:l02`2_MjnA/96`TK1g00:_00>@[jl01Zl00f^@0008[`05Mjn_[g401:l00gIgLP0=[`03
T:n_00R_00>@[i007jl91`001`L0ojl7[`N_1jl7[gMb1jl7[`N_1jmg[`N_1jl7[`N_1g87[`N_1jl7
[`N_TJl7[`N_1jl7[`N_TJl7[`N_1jl7[`Mg1jl7[`N_1jl7[`N_1i07[`N_1jl7[`N_1jl7T@N_1jl7
[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[k2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[k00SfaaMPN_1jn`U[2FT:l7T76_LJl7T0N_1jl7[`N_LFbA[`N_1jl7[`N@O:l7L@N_UGL7R`N@1jl7
[`N_1jl7[`N_TJl7[`N_1jl7[`N_TJnA[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7[gMb1jl7[`L:1`00
1`L;[`03CZn_00J_00=b[jl02:l00g:_[`07[`03LZn_00V_00=b[jl02Jl00g:_[`0:[`03TJn_01Z_
00>A[jl03Jl00i6_[`0R[`0:/:n`TK2A/95a00B_00J@[jl0T@03[`03T:n_00F_00@0T:n@1jl0192_
Mf`7[`03LY2_00^_00>@[jl02Zl2T1j_2@L000L70?n_1jl7[`N_1jl7[`Mb1jl7[`N_1jmg[`N_1jl7
[`N_Mjl7[`N_1jl7[`Mb1jl7[`N_1jl7[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jn`[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`NFUiJGTK1a1jl7[g6@LJl7[`Mb1f`7LIJE1jnF[`N_1jl7L@Mb1jnA[`N_1jl7[i1gLJml
M`N_1jl7T0N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7
[`Mb1jl72PL000L73jl02TmbCg9?LTmbCgL8[`03LZn_00N_00=b[jl05:l00i6_[`0:[`03LZn_00^_
00>A[jl03:l00i6_[`0=[`03TJn_00j_00RAUiJGUYNFU`V_00AfLGH00jl017Ma[g45[`05LG:_[g40
0Zl017NAT944[`050:n_[g800jl00g:_[`03[`03Jg6_00V_00>;[jl02Zl00i6_[`0;[`03TJn_01^_
2@L000L70=N_1jl7[`N_1jl7[`N_1jmbL@N_1jl7[`MbMg9gLWMbMg9gLPN_1jl7[`N_1jl7[`N_1jl7
[`Mg1jl7[`N_1jl7[`Mg1jl7[`N_1jl7[`N_1i07[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`NA1jl7
[`N_1jl7[`N_1jl7T@N_1jl7[`NA/96`TK2A/96G[`N_1jl7[`N_1jl7[`N_1`1f[i2_1i1a[`L01jl7
[g6_1g87L@N_1gL7LPN_LF`7[`N_T:l7[i2_1jmaK92_1jl7[`N_1jl7[`02T00f1jl7[`N_1jl7[`N@
1jl7[`N_1jl7[`N_1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L74Jl00dj_[`0?[`0<
LWMbMg9gLWMbTG:A1jl00i2_[`0:[`03LZn_00Z_00>A[jl02Zl00i6_[`0=[`03TJn_00Z_2Y4:[`03
UZn_00f_00MfLG6F[gMa00>_00=a[jl00Zl01@2_[i0000B_00FF[g9bL@06[`03LZn_00:_00>@[g40
0jl00i2_[`08[`03LZn;00Z_00>A[jl02jl00i6_[`0K[`T700071`1[[`N_1jl7[`N_1jl7[`N_1jl7
L@N_1jl7[g:_1jl7[`N_1g87[`N_1jl7[`Mb1jnALY5bTG:AMi5gTGNAMi5gTJl7[i6_1jl7[`N_1jl7
[`N@1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl00Y400iJA[`0:T@0S[`N_Ujl7[`N_1jl7[`N_1jl7
[`NF1jl7[`N_1jl7[`N_1jl00W409`M/UgH708n_LJl7[`Ma1jl7[g6_1jl7[iJ_LG87[g6_1jl7LPN_
1`02L@17[`N_1g87[`N_1jl7[`N_1h/7T0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl02PL000L74jl00dib[`04[`03LZn_00F_00=b[jl01jl00gN_[`08[`03
LZn_00J_012ALY5bTG:ALY5bTG:ALY5b0i402gNAMi5gTGNAMi6@00fA0Zl00gR_[`0=[`03TJn_00n_
00>F[jl02jl01g5f076_U`000Zl01Xl0T7:_00F_00<0[jl00jl017MaLI02[`030:n_00:_00B@[g5[
1Zl00i2_[`09[`03T:n_01R_00>@[jl06Zl91`001`L0`Zl7[`N_1jl7[`N_1jl7[`N_1jl7[g:_1jl7
[g:_1jl7[`N_LZl7[`N_1jl7[gR_1jl7[`N_1jl7LPN_1jl7[`N_1jl7M`N_1jl7[`N_1jl7[i6_1jl7
[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_Ujl7
[`N_1jl7[`N_1`1f[g5/T9Ia[`N_LGL7LI1gT7L7[g6_1jl7[g5gT:n@K0N_1jl70W40Bjl7[`Mb1jl7
LPN_1jl7[`N_1jl7R`N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`0:1`001`LF[`03LZn_00>_00=b[jl01Jl00g:_[`06[`03LZn_00R_00=g[jl02Jl00i6_
[`0:[`03N:n_00Z_00>A[jl03Jl00i6_[`0=[`03Ujn_00j_00>F[jl02Zl017IaLJl3L@07UZn_MP2_
LP05[`04LW5bT0>_00Ma[jmgLZma00>_00=aJi001Zl01G:_[jmb00Z_00F@[jn_T008[`03T:n_00^_
00>@[jl06Zl91`001`L0Vjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LW47[`N_LZl7[`N_1jmb[`N_1jl7
[`Mb1jl7[`N_1jl7[gN_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_TJl7[`N_1jl7
[`N_1jl7T@N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`N_1iH7[`N_1jl7[`N_009a0247K76_
LJnG[i001jn@[`N_1jl7L@MbT:ma[`N_UW87LI000W40Cjl7[i2_1jl7[i2_1jn@[`N_1jl7[`N_1jn@
[`N_TJl7[`N_1jnA[`N_1jl7[`N_1jl7[`N@1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl02PL000L7
6Jl00aR_[`02[`03LZn_00B_00=?[jl01Zl00gN_[`08[`03LZn_00V_00=b[jl02Zl00i6_[`0:[`03
N:n_00f_00=h[jl03Jl00i6_[`0>[`03Ujn_00R_00MaMP2?0:l000:_00>GLG400Zl00`2_[`04[`05
LJmbLW401Jl019H0[f/5[`03LZn_00B_00B@[jmb2jl01G:_[jmb00N_00>@[jl03:l00i6_[`0I[`T7
00071`0I[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@1oLTh7[g:_1jl7[`Mb1jl7[`N_1jmh[`N_
1jl7[`N_1g87[`N_1jl7[`N_TJl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_
UZl7[`N_1jl7[`N_1jl7[iN_1jl7[`N_1jl7[`N_1jl7[iN_1jl7[`N_1`02L@03K8n_009a06h7[`NF
1`07K0Mg1jl7[`Ma1jma[`N_1jl7Mg5/1jl7[`N_T:l7[`N_1jn@[i2_1jl7[`N_1jl7[i2_1jl7T0N_
1jl7[i6_1jl7[`N_1jl7[`N_1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`X700071ab_00Ia[jm>
LTh5[`03CZn_00J_00=b[jl01jl00gN_[`09[`03TJn_01N_00>A[jl07Jl00iJ_[`0?[`03UZn_00B_
0W401P2?[g6_L@:_00Aa[iIa1:l02g6_[jl0[`2_CZmb00:F0jl01G6_[jmb00B_00>@[jl01Zl00g:@
[`0:[`03T:n_00>_00>@[jl00jl00i6_[`0<[`03T:n_00>_00<H[jl04Zl91`001`L08Jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_LF`7[`02LP1f1jl7[`Mb1jl7[`N_1jmg[`N_1jl7[`N_N:l7[`N_
1jl7[`N_TJl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7UPN_1jl7[`N_1jl7[`N_1jl7UPN_1jl7[`Aa0187076_1g47[g6F1jl7L@Mg1jl2L@031jl700:F
05V_1jl7[i2_1jn@[`N_1g87[`N_1jl7[`N@1jl7[`N_1jl7[`N;1jl7[`N_TJl7[`N@1jl7[`N_1jl7
[`N_1jnA[`N_1jma[`N_1jl7[`N_1jl7[`N_1jl7[`0:1`001`LO[`06Jjn_LDib1:l00dn_[`05[`03
LZn_00N_00=b[jl02Jl00g:_[`09[`03N:n_00^_00=b[jl03:l00i6_[`0>[`03TJn_00n_00>F[jl0
0jl01@1a[hl0009a0Zl01GJ_[`2F00:_00Ea[jn_LP02MP:_00FFN:n_T005[`05LZn_[i000jl00i2_
[`07[`03T:n_00V_00>@[jl01Jl01I2_[jnA00j_00>@[jl00jl00aQa[`0B[`T700071`0U[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`M/1f`7[`03LP27[gN_1jl7[`N_1g87[`N_1jl7[`Mb1jl7
[`N_1jl7[i6_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7
[`N_1jnG[`N_1jl7[`N_1jl7[`N_1jnG[`N_1`2?08n_L@07[hn_1g47[iMa1jl7[gIg00>F04V_1jl7
[i2_1jl7LPN_1jn@[`N_T:l7[`N_1jl7[i6_1jl7[`N_1jl7R`N_1jl7[`N_TJl7[i6_1jl7[`N_1jl7
[`N_1i07[`N_009a0187[`N_1jl7[`N_1jl7[`N_1jl:1`001`LQ[`030:lH00B_00=bCW801jl00g:_
[`07[`03Mjn_00V_00>A[jl02Jl00i6_[`0;[`03TJn_00^_00>A[jl03jl00iJ_[`0>[`03UZn_00:_
00Ma0:ma[`2?00B_00D0LJmfUP03[`05CWIhUW801:l01G:_[jn@00>_00=b[jl00jl0192_[g89[`03
LY2_01:_00B@[jn@3Zl00i6_[`02[`03676_01:_2@L000L702Z_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7J`M/1jl7[`L2LP1][`N_1jl7[gN_1jl7[`N_1jmh[`N_1jl7[`N_1gL7[`N_1jl7
[`N_1i47[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N_Ujl7[`N_1jl7[`N_1jl7[`NF1jl7[`N_1jl7
[`N_1jl7[`NF1jma0002L@071f^@[`N_1`03L@05Sjl7UPL00iH0CPN_1gL7[`Mb1jl7LPN_1g87[`N_
1g87[i2_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[`N@1jl7[`9a
0187[`N_1jl7[`N_1jl7[`N_1jl:1`001`LS[`03074H00J_00=bCW801:l00g:_[`07[`03LZn_01B_
00>A[jl03:l00gN_[`0;[`03N:n_00j_00>F[jl03jl029J_[`1a[hn_0W42[`04MP2_00B_0YH7[`04
LZn_LPB_00Eb[jn_L@05[`03LZmb00V_00=b[jl05Jl00i2_[`0=[`04T:n_L@:_00<H[jl04:l91`00
1`L0;jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1f]aK0N_1jl7[`N_00=b06R_1g87
[`N_1jl7[`Mb1jl7[`N_1jl7[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7
T@N_1jl7[`N_1jl7[`N_1jnG[`N_1jl7[`N_1jl7[`N_1jnG[g6_1`9a07=/1jn@[hma1jnGUYN_Mjl7
[`N_1jl7M`N_LZl7[`Mb1jmg[`N_1jl7LPMb1jl7[`N_1jl7T96_1jl7[`N_1jl7[`N_1jl7[`N_TJl7
T0N_1jl7[`N_1jl7[`N_TJl7L@N_LJl7[`N_1jl7[`N_1jl7[`N_00X700071bJ_00=a6:l00jl00g:_
[`03[`04LTibLPR_00=g[jl05:l00gR_[`0;[`03LZn_00b_00>A[jl03Zl00i6_[`0?[`04UZl0[`9a
00@0Sjn_0g43UP:_00=fMjl02Jl00g:_L@05[`03LZma00N_00=bT:l02:l00i2_T008[`03LZn_00R_
00>@[i403Zl0192_[g42[`036:n_012_2@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jmaK0N_1jmb[`N_1jl7[gMb1jl7[`N_1jmh[`N_1jl7[`N_TJl7[`N_1jl7[`N_1i47
[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`NFSjn?076_1gIaUYN_1jl7M`N_1jl7[`N_1jl7[`N_LZl7[`N_T:l7[`N_1jl7[i1b1jl7[`N_1jl7
T0N_TJl7[`N_1jnA[`N_1jl7[`N_1jnAT0N_1jl7[`N_1jl7[`N_TJl7L@N_LJl0012_1jl7[`N_1jl7
[`N_1jl72PL000L7:Zl01G6_[jmb00^_00=gLWL03Jl00i6_[`09[`03TJn_00^_00>A[jl03:l00i6_
[`0O[`0409J_L@>F00>?LG401:l00gIg[`0?[`9b1Zl00g:_[`07[`03LZn_00Z_00>@[jl01:l00g:_
[`0;[`03Rjn_00j_00Ea[jn_L@0A[`T700071`2I[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[g6_1jmb[`N_1jl7LPN_1jl7[gMbMjl7[`N_1jl7[`Mg1jl7[`N_1jl7[`N_TJl7
[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_1jl7UPN_1jl7[`N_1jl7[`N_1jma
MYNF009a07B?[g6_1g47M`N_1jl7[`N_1jl7[`N_1jl7[`Ma1jl7[`N_1jl7[`N_1jl7[i2_1jl7[`N_
1jl7[`N_TJl7[`N_1g87[`N_1jl7[`N_1jnAT0N_1jl7[`N_1jl7[`N_1g47[`Ma1jl7[`N_1jl7[`N_
1jl7[`X700071bb_00Ea[jn_LP05[`03LZn_00N_00AgLWMb1jl00g:_[`09[`03Mjn_00^_00=h[jl0
3:l00i6_[`0=[`03Ujn_00^_00ZGTINAUgN@Mhn_0W42[`03LGJ_01B_00=a[jl01:l00g:_[`06[`03
LZn_00b_00>@[jl00jl00i2_[`0:[`03T:n@00b_00>AK:l00jl00g6_[`0?[`T700071`3o[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`M/LG07LPN_1jl7[g:_1jl7[`N_1g87
[`MbN79h[`N_TJl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7
[`N_1jl7[iN_1jl7[`N_Ui6GTIN_1gIaMXmaSgNFM`N_LGL7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LZl7
[`N_1g87[`N_1jl7[i1b1jl7[`N_1jnA[`N_1g87[`N_1g87[`N_1jl7[`N_1jl7T@N@1jl7[`N_1jl7
[`N@LJl7[`Ma000@[`N_1jl7[`N_1jl7[`N_1`X700071bf_00DHLAR_6006[`03LZn_00J_00=?[jl0
1:l017MbMg8<[`03TJn_01Z_00>A[jl03Jl00i6_[`02[`03Ui6G00J_00D0LG5a[`03L@06[hmgT7N?
5jl00g:_[`03[`03LW6_00N_00>@[jl01Zl00g:_[`03[`03LZn_00:_00>@[jl02jl0196_[i0:[`03
TFb_00>_00=a[jl03jl91`001`L0Ujl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`M/1aP7L@N_1jl7LPN_1jl7[`N_LZl7[`N_1jl7[`MgTGNAMi6_1jl7[`Mh1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[i6_1jl7[`N_1jl7[`N_1i6GUYNFUjl7[`N_1jl7[i00S`000g400jl7
L@02L@0OSjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g47[`N_1`02LP1?[`N_1jl7[`N_T:l7[`N_1jnA
[`N_1jl7LPN_1jnA[`N_1jl7[`N_1jl7T@N_1jl7[`N_1jl7[`N@LJl7[`N_LJl7[`N_1jl7[`N_1jl7
[`0:1`001`L_[`046:lHLPJ_00=?[jl01Zl00g:_[`07[`03Mjn_00>_00QhLY5bTG:_M`b_00>A[jl0
3:l00i6_[`05[`07TIJAUY6FT@02[`03T:n_00V_00Ma08l0[`1a00:_00D0LJn_L@0I[`03LJn_00>_
00=aLZl02:l00i2_[`04[`03LZn_00F_00Eb[jn_LP0<[`05TJn_[i402Zl00i2_[`03[`03K:n_00j_
2@L000L7076_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`M/1jmb
[`N_1jmb[`N_1jl7[`Mb1jl7[`N_1jl7[gR_1jl7[`N_1jl7Mi5hTGRAN95hTGRA[`NA1jl7[`N_1jl7
[`06T@0OUi6GTINA1jl7[`N_1jl7[iJ_1jl7[`N_1jl7MW401`02L@0WMW5aSjl7[hn_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_LZl7[`N_009b04h7[`N_1jl7[`Mb1jl7[`N_1g87[`N_1jnA[`N_TJl7[`N_
1jl7[`N_1jnA[`N_TJl7[`N_1jl7[i6_1jl7[g6_1jl7[`N_1jl7[`N_1jl:1`001`La[`056:n_LQP0
1:l00g:_[`06[`03Cjn_00N_00=b[jl02Zl00i6_[`07[`0=TGRFN95hTGRAN95hT@05[`03TJn_00n_
00>@[jl01Zl2L@0408n_S`9a0Zl0176_[hlK[`03LZn_00:_00=a[g401Zl0192_[g86[`03T:n_00F_
00F@[jn_T00<[`03T:n_00:_00>A[jl01Zl00i2_[`03[`03K:n_00j_2@L000L70?n_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1f`7[g6_1jl7[g:_1jl7[`N_LZl7
[`N_1jl7[`Mb1jl7[`N_1jl7[`Mg1jl7[`N_1jl7[`N_UZl7[`N_1jl7[`N_1jl7[iJ_1jl7[`N_1jl7
[`N_1jl7[iJ_1jl7[`N_LJma[hl0LJl7L@N_1g47[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Ma1jl7
[g:_1g87[`N_1g87[`Mb1jl7[`N@1jl7[`N_1jl7[`Mb1jl7[`N_1jl7[`N_TJl7[`N_TJl7[`N_1jma
T0N_1jl0012_1jl7[`N_1jl7[`N_1jl72PL000L7=:l01W6_[f]b60:_00=b[jl01Jl00g:_[`08[`03
LZn_00V_00=b[jl02Zl00i6_[`0=[`03TJn_012_00=gUZl00Zl01740S`02L@B_00B?[jma7Zl01G:_
[jmb00:_00=a[jl00Zl00g:_[`02[`03LZn_00>_00=b[jl01Jl017:_[i0=[`03TJn_00B_00>@[jl0
0Zl00g6_T@05[`03LJn_00f_2@L000L703B_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl70g40F@N_LG47LPN_1jl7[`N_LZl7[`N_1jl7[gR_1jl7[`N_1jl7[`Mg
1jl7[`N_1jl7[`N_UZl7[`N_1jl7[`N_1jl7[iJ_1jl7[`N_1jl7[`N_1jl7[`N_1jn@00Aa07^?1jl7
[hn_1jn?[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LPN_1g87[i2_1jl7[i2_1jl7LPN_1jnA
[`N_1jl7[`N@1jn@[`N_1jl7[`N_1jl7T@N_1jl7[`N@1jl7L@N@1jl7[`Ma1jl7[`N_1jl7[`N_1jl0
2PL000L7=Jl00f^_6002[`04LAQbC`N_00=?[jl02:l00gN_[`09[`03N:n_00Z_00=h[jl08Jl027Mf
0740Mg6?1:l0176_[hlP[`05LJn_[g400Zl00g:_[`02[`03LZn_00:_00>@[jl00Zl00g:_[`07[`03
LZmb00b_00>A[jl01Zl0192_[g42[`03T:n_00:_00=a[jl03Jl91`001`L0=Zl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`9a05J_1jn:LGMb1jl7[`N_1g87
[`N_1jl7[`Mb1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7
[`N_1jl7[`N_1`9a07mgT8l7S`Mf1jn?[`Ma1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jmb
[`N_Mjl7LPN_1jn@[`N_1jn@[`N_TJl7[`N_1jl7[`N@TJl7[`N_1jl7[`N_1jnF[`N_1jl7[`N@1g47
[jZ_1jl7L@N_1jl7[`N_1jl7[`N_00X700071cJ_00=a67400jl0175[LTl6[`03LZn_00R_00=b[jl0
5Zl00i6_[`0=[`03TJn_012_00MhMWJ_LHma00:_00>?[hl00jl00hn_[`0R[`05LJn_[g400jl00g:_
[`04[`05T:n_[i002Jl00i1b[`0<[`03T:n_00N_00F@LJn_T005[`036:n_00b_2@L000L703N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00=a00>_1jl0
0W43LP3=[`N_1jmb[`N_1jl7[`N_N:l7[`N_1jl7[`N_TJl7[`N_1jl7[`N_1jnF[`N_1jl7[`N_1jl7
[`N_UZl7[`N_1jl7[`N_1jl7TIJAMWL7MY2_T:l7MPMa1jn?[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[g:_1jl7LPN_1g87[`N_1g87[`Mb1jl7[`N_1jl7[i6@1jl7[`N_1jl7[`N_UZl7
[`N_1jl7[`Ma1jnA[`N_1jma[`N_1jl7[`N_1jl7[`0:1`001`Lh[`03LAR_00:_00=[[aP00Zl00g9?
[`02[`03Cjn_00R_00=g[jl02:l00g:_[`0<[`03TJn_00f_00>A[jl02Jl00iIh[`02[`06MXn_Sg6?
0Zl0176_[hl2[`03LJn_022_00Eb[jn_LP03[`05LZn_[g802Jl00i2_[`09[`03T:n_01F_00FA[jn_
T@04[`036:n_00b_2@L000L703V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl00W402@N_1g47L@N_1`02LP39[`Mb1jl7[`N_1jl7LPN_1jl7[`N_1jl7
M`N_1jl7[`N_1jl7[`NA1jl7[`N_1jl7[`N_1jl7T@N_1jl7[`N_UY6F[`N_1gJ?[`Mf1hl7[i2_Sjl7
[hn_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7L@N_1g87[`N_T:l7LPN_1jl7[i2_1g87
[`N_1jl7[`N_1i2A[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jnA[`N@1jl7[g6_1jl7[`N_1jl7[`N_00X7
00071cZ_00=a6:l00Zl016]bL743[`03LTmb01F_00=h[jl03:l00gR_[`0=[`03N:n_00B_00>F[jl0
1:l01G6@LJmf00:_00J?[jma[g42[`03Sjn_02:_00A>[jmb1:l0176_[g86[`05LZn_[g802Zl00i6_
T00G[`03TJnA00B_00<H[jl03:l91`001`L0>jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_009a0=87[g5a1jma[`N_1g9h[`N_1jl7[`N_1jl7[`N_
1jl7[`N_TJl7[`N_1jl7[`N_1jnF[`N_1jl7[`N_1jl7[`N_1i47[`NAUZl7[`N_1gJ@MXn_Sjn@[`Mf
1jn?[`Ma1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LZl7LPN_1jl7[`N_T:l7[`N_
1g87[i6_1jl7[`N_1jl7[i6@1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_TJnA[`N_LJl7[`N_1jl7[`N_
1jl:1`001`Lm[`0567:_[aP00Zl00g5`L@02[`08CZn_[g9gLWL2[`03LZn_00Z_00=h[jl02jl00iJ_
[`0:[`0:MiIgUZnFMiIgT@>_00Mg07If[g6@00:_00Aa[jn?1:l00g6_[`0T[`04LZn_L@B_00Ab[jmb
1jl00g:_LP0;[`04Mjn_T@V_00>A[jl03:l0192_[g4?[`T700071`27[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`MaLW07L@N_LJn:[`Mb1jl7
[`N_N787LPN_1jl7[`N_1jl7M`N_1jl7[`N_1jl7[`N_UZl7[`N_1jnFMiIg1jl7[`N_1jl7[iIg009a
08J?MPMfSgH7MPN_T:l7[`N?T:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[g:_
LZl7[`N_T:n@[`N_1jl7[`Mb1g87[`N_1jl7[`N_TJl7[`N_1jl7[`N_1i47[`N_1jl7[`N_1jl7[i6_
1g47[`N_1jl7[`N_1jl7[`X700071cj_00=a67800Zl00aR_L002[`03LJmb00R_00EbMg:ALP09[`03
LZn_00n_00AgTGNA3Zl02GNA0:maT:n?L@02[`04T:n_L@:_00B?LJma::l00g9a[`04[`04LJn_T0N_
00>@[i002jl0196_[h/8[`03TJn_00J_00<H[jl01:l00i6_L@0?[`T700071`3E[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LZmb[g:_1g47
[g6_LZl7[`N_1jl7[gR_1gNAMi5gTGL7[i6_1jl7[`N_1jnAN95hTGNFM`N_1jl7[`N_1jl7[`N_1jl7
M`Ma1gJ@[hmfT:l7MPN_T:l7L@N?1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`Ma1g87[`N_1g87LPN_1jl7[`N_00:@03P7[`N_1jl7[`N_1i07[`N@1jl7[`N_1jnF[`N_1jl7[g6_
1jl7[`NA1g47K0N_1jl7[`N_1jl7[`X700071d2_00MaJg:_LZma00>_00=aCZl02:l00g:_[`07[`03
LWQb009h00NAN95hTGRA00F_00>A[jl04:l01WN_07IaT0:_00Ea[jn_L@02[`07Sjn_[hn_S`0Z[`03
LG:_00B_00=b[i002Jl00i2_[`09[`03TJn_00:_00>;[jl01Jl00i2_[`05[`036:n_00B_00FAK:n_
L@0=[`T700071`33[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1faaL7:_LZn:[`N_LG87[`N_1jl7[`Mb1jl7[`N_1jl7[`Mg1jl7[`N_1jl7
[`N_1jnFM`N_1jl7[`N_1jl7[`N_UWL7[g5g1gH7MXn_1jn@[`Mf1jn?[`N?1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_009b04X7[`N_1jn@LPN_1jl7[`N_1g:A[`N_1jl7[`N_
1jnA[`N_1i07[`N_1jnF[`N_1jl7[g6_1jl7[`N_TJl7K0N_1jl7[`N_1jl7[`X700071d2_00Qa676_
LQR_J`>_00=`[dh02Jl00g:_[`09[`03Mjn_00f_00=g[jl02Zl029J_[jma[i1f0Zl01Hn_Sjn@00>_
00J?[jma[g4/[`03LZn_00F_0W89[`03LZn_00V_00=g[jl00jl00h^_[`04[`03TJn_00B_00<H[jl0
1Jl0192_[g4=[`T700071`34[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jmb[g6_1faa[`N_1g5bL@N_1jl7[`N_N:l7[`N_1jl7[`N_1gL7
[`N_1jl7[`N_1jl7[`Mg1jl7[`N_1jl7MiJ_1jl7[i2_T:n?[`Mf1gH7[i2_1g6@[i2_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`9b04V_1jl7[`MbT:l7[`N_1jl7[i5b1jl7
[`N_1jl7[i6_1jl7[`N@1jl7[`NA1jl7[`N_LJl7[`N_1jnA[`M/1jl7[`N_1jl7[`N_00X700071d:_
00Mb[g6_LTib00>_00Ea[g9`L@07[`03Mjn_00Z_00=g[jl03Jl00gN_[`05[`03Mjn_00>_00L0T:n@
[hmf00:_00>?[i000Zl0176_LI02[`03LJn_02^_00=a[jl01Jl00g:_[`07[`03LZmb00V_00=g[jl0
1Jl00h^_[`02[`03TJn_00>_00=a0:l01Zl01FbA[jm/00b_2@L000L70?J_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LW1a[g9aLZl7
[g5b1jma[`N_1jl7[gR_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7MiJ_1jnFM`N_1jl700N_1gH7
MPN_T:n@[`N_T747MPN_T:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[g:_1jl7[`N_T:l7[`N_1jl7[`N@1g87[`N_1jl7[`N@1jl7[`N_TJl7[`NA1jl7[`L2L@0G[`N_1jl7
L@NA1jma[`N_1jl7[`N_1jl02PL000L7A:l01G4HLF]a00:_00E>LV^_CP03[`03L:n_00B_00=g[jl0
2Jl00gR_[`0?[`05TGN@Mi000jl0276_[gJ_LHma0Zl018n_[g42[`04Sg6_LC:_00=b[jl01Jl00i1b
[`07[`03T:n_00V_00>A[jl01Zl01G:_[jn@00B_00=a[jl01Jl01W6_[i2_K0b_2@L000L704F_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl00W401PMaRZl7[`9b02l7LPN_1g47[`N_1g87[`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl7[iIh
UZl7[`N_1`02L@1>[`Mg1gJ@[i2_1gH7[`N_1gH7[i2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jmg[`N_1jl70Y00?jl7[`N_1jl7T0N_TJl7[`N_1jl7[`N_1jl7
[`N_1jl7[iJ_1jl7L@N_1jl7[`Ma1jnF[g6_1jl7[`N_1jl7[`0:1`001`M6[`04LAR_60B_00UaCW9>
[g2_L7401:l00g:_[`0:[`03N:n_00N_00>AMi402:l01@2_[jmf00:_00>?[i000Zl00g6_[`03[`05
LJn_[g400Zl00hn_[`0f[`9b2:l017:_[i0A[`04T:n_T0>_00==[jl01Jl01AR_[jn@00:_00=a[jl0
2Jl91`001`L0Ajl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_00Aa0<@7[`N_RZmhLW6_1jma[`N_1g87[`N_1jl7[`N_1gL7[`N_
1jl7N9J_1jl7[`N_1jl7L@N_1gL7[`N_1gH7[i2_1jl7MY2_1jn@[`N?1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jnA[`N_1jl7[`N_TJl7[i6_1jl7[`N_
1jl7[`N_1jl7[i6_1i47[gJ_1jl7[`N_LJl7[`N_1g47[`N_1jl7[`N_2PL000L7B:l00g6_L@04[`03
L:n_009b00EgLWN_L@03[`03Mjn_00Z_00=g[jl00Zl00i5g[`0:[`030:n_00:_00J@[jmf[g43[`07
LJn_Sg6_L@02[`05Sjn_[hl0<Jl00g:_[`04[`03LZn_00N_00Eb[jn_LP0B[`03T:nA00:_00==[jl0
1Jl00aR_[`04[`03LJn_00V_2@L000L70?6_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`MaLW1b[`N_RZl7[`Ma1jmhLWQb1jl7
[`N_1jl7[`N_1jnAMiJAUZl7[`N_1jl7[`N_1g47[`N_T:l7T9F_T:l7MPN_1gJ@S`N_1hl7[`N?1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g87[`N_1jnA[i2_1jl7
[`N_1i07[`N@1jl7[`N_TJl7[`N_1jl7[`N_00:A01`7[gJ_1jl7[`N_LJl7[`N_1g47[`N_1jl7[`N_
2PL000L7BJl016]aJg43[`05LJn_[g401:l0472_LWM?N79hLWQbN79hMgP2[`03Mi2_00^_00<0[jl0
0Zl01i2_T7N@[i000jl0192_[g42[`05T:n_[hl00Zl00g6_[`0a[`03LJn_00B_00>@[i001jl00i2_
[`02[`03LZn_00>_00=g[jl02Zl2T005[df_[g401:l00aR_[`02[`03TJma00^_2@L000L70?n_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_LZmbL@N_1g47[g9a1jl7[`Ma1g5h[`N_1jl7[`N_1jl7[`N_UY07[`N_1jl7[`LH1jl7
[i2_UJnE[`Mf1jn@[`N_T7H7S`N_T:l7[i2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jmb[`N_1jl7T0Mb1jl7[`N_1i07[`N_1i07[`N_TJl7[`N_1jl7[`N_1jnA
L@N_1`07[`Ma1jl0012_1jl7[`N_1jl7[`N_1jl72PL000L7BZl0176_LAP3[`05Jjn_LW001:l01G6_
[g1g012_00=g[jl01:l00g40[`03[`05T:n@[i001Zl01g6_[jma[gH00jl01G6_[jma03B_00=b[jl0
0jl0192_[g86[`03TJn_00B_00Eb[jn_M`0=[`03Mi6@00:_00Ea[jn_L@05[`04TJn_K0Z_2@L000L7
04Z_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1`9a0:Z_LW47[hZ_1jmb[`N_1g47[`MaN:l7[`N_1jl7[`N_1jl7[`Mg1jl7
[g6_1jl7[`MgUJl7T0N_T:l7MPN_1i07[i2_1jn@[`N_T:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LZl7[`N@1jl7LPN_1jl7[`N_1jl7[i6_1jl7T0N_1jl7
[`N_1jl7[`N@1`:A01L01jl7L@N_1jl7T@N_LJl7[`N_1jl7[`0:1`001`M;[`076:n_LTj_L003[`03
CZn_00>_00I`[jn_L7L@[`04Mi2_L@:_00EfT7J@M`03[`05Mjn_[gH00Zl01Hn_[gJ@00F_00F?[jn_
S`0f[`03LJn_00J_00>@[jl02jl01I2_[jnA00f_00FA[jn_T@03[`03LJn_00B_00>@[f`02Zl91`00
1`L0Cjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_LJl00g40ZG87[`MaLZl7[hZ_1jl7L@Mb1jl7[`N_1jl7[`N_1gNF
[`N_1jl7[i2_1jnE[`N_1gH7[i2_T:n@[`N_1jl7MPN_1gH7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LPN_1jl7[`N_1g87[`N_1jl7[`N_1jl7LPN_TJl7
[`N_1jl7[`N_MZnA[`LH1jl00Y404PN_1jl7[iJ_1f`7[`N_1jl7[`X700071dj_00MaL:lHLDib00:_
00=>[jl00Zl01G6_[jma00:_00=bLJl02:l00i1g[`07[`03T:n_00:_00Eg[jn_MP03[`05LI2_[gH0
0Zl0192_[g43[`03LJn_03N_00=a[jl01Zl00gN_[`02[`03LZn_00V_00>@[jl02Jl00df_M`03[`06
0:m/[i2A1Jl00i40[`09[`T700071`1>[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl717400jl7[`02LP2XRW87[`Ma
1jl7L@N_1g9a[`N_1jnFMiJ_1jl7[`N_1jn@[`N_1gL7[`N_T:l7MY2_1hl7[`N?1jn@[`N_T:l7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[g:_1jnA[`N_
1jn@[`N_TJl7[`N_1jl7[`N_TJl7[`N_1jl7[`N_MZnA[`N_LJma[`N_0Y403PN_1i5a[`N_1jl7[`N_
2PL000L7Cjl01AQaL75[00>_00=>LTh00Zl00g6_[`02[`05L:n_[g800Zl00i1gT00;[`03T:n_00:_
00=g[jl00Zl01I2_[jn?00>_00F@[jn_T002[`05LJn_[g40>Zl017:_[g86[`04Mjn_LPV_00=b[i00
2jl014f_[i42[`030:m/00F_00F@[aRA000:[`T700071`1@[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L2L@2W
[g5`1jl7LG9gN7MhMgQgN7MhMgQgUZl7[`N_1jl7[`N_1jn@[`N_1gL7[`N_1gH7[`N_1jn@[`N_T:l7
MPN_1jn@[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1g87[i6_1jl7[`Mb1jl7M`N_1jl7[`N_1g87T0N_1jl7[`N_1g47[`N@1jl00W4050N_1jl7[`NA
UZnF[`N_1jl7[`N_2PL000L7DJl01AQa675`00:_00Aa[jm>1Jl00g2_[`02[`03Mjma00Z_00=gT:l0
1:l00i2_[`02[`05MZn_T7L00jl01GJ_[jmf00:_00F@[jn_T00l[`04LJn_T0J_00Ab[jnA2Zl00g:A
[`08[`05LJn_[i400Zl00`1a[`08[`03TI2_00V_2@L000L705:_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
0W405G07L@Ma1jmb[`N_1jma[`N_1g6AM`02L@2>1jl7[`N_1i07[`N_1jl7[`N_1jn@[iF_UJl7S`N_
1jn@[`Mf1jl7MPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`Mb1gL7[`N_1jl7LPMg1jl7[`N_1jl7[`N@1jl7[`N_1jl7L@N_1jnA[`9a01@7[`N_
1jma[`N_1jl7[`N_1jl7[`X700071e:_00Qa675[[g2_L0:_00=bL:l00jl01G6_[jma00:_00=bMg00
1Jl00gN_[`05[`03Mjn_00:_00>@Mi000Zl017N_[i03[`05T:n_[i000jl00i2_[`0k[`03LZn@00N_
00=g[i402jl00i6@[`07[`03LJn_00:_00Ag[`1a1Jl00g6_[`0=[`T700071`3Y[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jma676_1g2:[`N_LJl7[`Ma1jl7L@N_1jmh[`N_1i07[`N_1jl7T0N_1jl7MPN_1i07
[i2_1jl7S`N_T:l7[i2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7M`N_1jl7[`N_1jnA[`N_1jl7[`N_1jl00Y00996_1jl7[`Ma1jl7[i6_
1g47[`N_1g47[`N_1jl7[`N_1jl7[`X700071e>_00@H[aQa0Zl01G2_[jm>00B_00Ea[jn_L@04[`03
L7Mb00Z_00J@[jmfT7H3[`05Mjn_[gH01:l0192_[g43[`03LJn_03n_00=b[jl01Jl00g:_LP0;[`05
Mjn_TI000jl00df_[`05[`030:n_00>_00=/[jl03Zl91`001`L0FJl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jma[g6_00=b0;>_1g5b[`N_RZl7[g6_N7MhM`N_1jl7Mi5gTGN@[`N_UZl7[`N@1jl7S`N_1jn@
[`N_1jl7[i2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jmg[i6_1jl7[`N_1g:A[`N_1jl7[`N_1jl7M`N_1jnAT96_1jl7[`N_LJl7[g6_1jl7
[`N_1jl7[`N_1jl7[`0:1`001`MF[`03LJma00:_00I`LTia[g82[`06L:n_Mg9g0jl00g6_[`06[`03
Mjn_00:_00>@[jl00jl01I2_[jn@00>_00>@[jl01Jl00hn_[`0o[`9b2Jl00g:@[`0:[`03T:n_00B_
00==TGL00jl00`2A0002[`03LJn_016_2@L000L705N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0
0g401:l7L@L4LP1a[i5gTJl7L@N_1jma[`N_1jnFM`N_1jl7[`N_1jnE[`N_1jl7[`N?1jl7[`N_1hl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl00WL020N_1jl7[`N_0Y00<0N_1jl7[`N_1jl7[iF_1jl7[gJ_1i2A[g6@LJma[`N_1jl7[`N_
1jl7[`N_1jl7[`X700071eN_00hHLAQa[g5[[g2_CWN_L@>_00=a[jl00Zl00g2_[`02[`03T:n_00V_
00>@[jl01jl02I2_[jn@[hmaT012[`03LI2_00R_00>@[jl02Zl00gN_[`02[`03LJn_00>_00FAMi40
L@0E[`T700071`1H[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`Aa0;>_1g6:[`N_Mg47[`Ma1jl7
[g6_1gNF[`N_1jl7[iJ_1jl7T0N_1jn@[`N_T:l7[`MfT:n@[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Mg1jl7[`N_1jl7[i6@1jl7
[`N_1jl7[iJ_1jl7L@N_1jma[`N@TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`0:1`001`MI[`04674HL@:_
00=[LJl00Zl00g9>[`02[`05L:n_[g002Zl00i2_[`02[`03Mjn_00>_00Ef[jn_L@04[`04T:n_MTB_
00>@[jl02:l00i6_LP0:[`05T:n_[g400jl00g6_[`0I[`T700071`1L[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7L@L2L@20[`Ma1jl7[hZ_1jma[`N_1i2F[`N_1jl7[iJ_1jl7[iF_1jl7S`N_1jn@
[`N_T8l7[i2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jmg[`N_1jl7[`L3[`0^1i2A[`N_1jl7[`Mg1jl7L@N_1g47[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`X700071e^_00<H[f/00jl01V^_[jm`LP:_00I`[jn_Mg47[`03T:n_00>_
00>@[jl00jl01I2_[jn@00>_00F@MZn_T017[`03T:n_00N_00FF[jn_T007[`05T:n_[g400Zl00`2_
[`0J[`T700071`1L[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L3L@2`1jmaL@N_1jmb[g6_
UY07L@N_1jl7[iJ_1jl7[`N@1jl7[`N?1jl7S`N_T:n@[`N@1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_TJl7[`N_1jl7UPN_
1jnAM`N_1jl7[iF_MZl7L@N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`MN[`0=LTibCW9>
LTibCW9>T004[`03L:n_00:_00>@[jl01:l00gN_[`03[`03MZn_00:_00F@MY2_T003[`03T:n_04N_
00>@[jl01Jl00iJ_[`04[`03LZn_00>_00EgUTf_000N[`T700071`1M[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_00Aa07l7[g6_LJl7[`Ma1g9g[g6_1jnE[`N_1jl7T0N_1jl7S`N_1hn@[`N_
T:l7[i2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7M`N_1jl700>_00T7[`N_1jl7T9400Zl08G4700N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`0:1`001`MN[`06LJma6:m`0Zl00g6_[`02[`05L:n_Mdl00Zl00i2_
[`0;[`05T:n@MY001:l01GJ_[jmf04f_00=b[jl00jl00i6_[`08[`04T79a022_2@L000L705n_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`04L@271g47L@N_1jma[`N_1g:@[`N_1jl7[`N_
1jl7[i2?T:l7M`N_1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g87[`N_1i47[`N_1jl700>_0287M`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7H:l01G5`LAQa00:_00E`[jn_L002[`0B
Mdmg[i1gT7N@Mi1gT7N@Mi1g2:l00i2_[`1B[`03LZn_00^_00>A[jl08Zl91`001`L0HJl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W4010MaLJl2L@1i1g9gLWN_1jl7[`N_1jl7M`N_
1jl7M`N_1jn@[`N_1i07[`N@1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7M`N_1`03[`0Z1jl7[`NF1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7HZl0276_[aQgCgM?0Zl01G6_[jma00F_
0WL5[`03Mjn_00>_00Eg[jn_MP04[`03T:n_052_00Eb[jn_T@05[`03TJn_02B_2@L000L706>_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W40O@N_1g47L@Ma1jl7L@N_1jn@[`N_
1jl7M`N_1jl7M`N_1jn@[`N_1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g8700>_02X7[`NF1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl:1`001`MT[`09LAQgLZm`[g1a00>_00=a[jl0
0WL6[`03Mjn_00B_00F@[jn_T004[`03MZn_05>_00=b[jl00jl00i6_[`0V[`T700071`1/[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jma60MgLJl70W40XJl7L@Mg1jl7[`N_1gL7
[`N_1jn@[`N_1i07[`N_T:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7M`NA1iH7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00X700071fN_00@HLG9a0Zl01G2_[i1g00j_00>@[jl0
0Zl00gJ_[`03[`03MZn_05F_00=b[i40:Zl91`001`L0Ijl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl00W40YPN_MgL7Mi2_RZl7[`N_1jn@[`N_1jn@[`N_1i07[`N_1i07[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl:1`001`MX[`08LJn_[g4H[g02[`03LJn_00>_00>@[jl01:l00gN_[`02[`03T:n_00>_
00>@[jl0Pjl91`001`L0KZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[g:_1g470g40WPMa1jl7[i2_1jl7[`Mg1jl7[`N_1jl7[i2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7JZl0174H
LG03[`03Jg6_00:_00>@[jl01:l00gN_[`09[`03MZn_08B_2@L000L706f_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[g:_009a0:1bLG9a1jn@[`N_1jl7[i2_1jl7[`N_
1jl7[i2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7K:l02G6_LJma675[T007[`03T:n_00B_00=f[jl00Zl00i2_
[`25[`T700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[g9a1jma[g9aLZl7[`N_1gL7[`N_1jn@[`N_1i07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`N_1`X700071fj_00AaL:m[
0Zl0176_LQP3[`03Mjn_00B_00>@[jl00jl00i2_[`25[`T700071`3o[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LZl7L@N_LJl7CW9g1jl7[`N_1i07[`N_T:l7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@
[`N_1jl7[`N_1jl7[`N_1`X700071g2_00eb678HLQQb678HLDib00B_00=f[jl00jl00gJ_[`26[`T7
00071`21[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jma
[`Ma1jma[`N_1jl7[g9>00J@08P7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_2PL000L7LZl00g6_L@02[`03L:n_00:_
00=bCW801Zl00g6_[`28[`T700071`1c[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`03L@2I[`Ma1dib[`N_1jl7[`Mb1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00X7
00071gF_00E`[jn_CP09[`03CZn_08Z_2@L000L707R_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LG82L@2E[`N_1jl7[`Mb1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl02PL000L7Mjl00g2_[`07[`03CZn_08b_2@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Ma1g47[`N_1g87[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7
[`N_1jl7[`N_1jl72PL000L7NJl00g1a[`03[`03CZn_08j_2@L000L707Z_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`9a09>_Mg87[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl02PL000L7Njl00f]g[`2B[`T700071`3o[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@
[`N_1jl7[`N_1jl7[`N_1`X700071on_4Jl91`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7
[`N_1jl7[`L:1`001`Oo[a6_2@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7[`N_1jl7
2PL000L7ojlA[`T700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`N_1`X700071on_
4Jl91`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7[`N_1jl7[`L:1`001`Oo[a6_2@L000L7
0?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7[`N_1jl72PL000L7ojlA[`T700071`3o[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`N_1`X700071on_4Jl91`001`L0ojl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`004:l7[`N_1jl7[`N_1jl7[`L:1`001`Oo[a6_2@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7
[`N_1jl7[`N_1jl72PL000L7ojlA[`T700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7
[`N_1`X700071on_4Jl91`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7[`N_1jl7[`L:1`00
1`Oo[a6_2@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7[`N_1jl72PL000L7ojlA[`T7
00071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@[`N_1jl7[`N_1jl7[`N_1`X700071on_4Jl91`001`L0ojl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7[`N_1jl7[`L:1`001`Oo[a6_2@L000L70?n_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl0012_1jl7[`N_1jl7[`N_1jl72PL000L7ojlA[`T700071`3o[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000@
[`N_1jl7[`N_1jl7[`N_1`X700071on_4Jl91`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004:l7[`N_1jl7
[`N_1jl7[`L:1`001`Oo[a6_2@L000L70?n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0012_1jl7[`N_1jl7[`N_1jl7
2PL00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00001\
\>"],
  ImageRangeCache->{{{0, 287}, {285.875, 0}} -> {-8.12139, -7.50157, 1.10189, 
  1.10189}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "The 3D graphics data passed to the viewer can be returned to ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " using the ",
  StyleBox["MVGetGraphics3D[]",
    FontFamily->"Courier"],
  " function. The polygons are triangulated and a ",
  StyleBox["Graphics3D[] ",
    FontFamily->"Courier"],
  "object is returned. You can manipulate the graphics data inside ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " using it's high level functions. You will get back the ",
  StyleBox["Point[]",
    FontFamily->"Courier"],
  ", ",
  StyleBox["Line[]",
    FontFamily->"Courier"],
  " and ",
  StyleBox["Polygon[]",
    FontFamily->"Courier"],
  " primitives, lines are not expanded to tubes and points not expanded to \
spheres. This feature will be of limited interest for graphics generated by ",
  
  StyleBox["Mathematica. ",
    FontSlant->"Italic"],
  "But the ",
  StyleBox["MVGetGraphics3D[]",
    FontFamily->"Courier"],
  " can be used to import  DXF files."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[MVGetGraphics3D[]]; \)\)], "Input",
  CellLabel->"In[23]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .92286 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
5.55112e-017 1.05085 -0.0150009 1.05085 [
[ 0 0 0 0 ]
[ 1 .92286 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
.05779 .25176 m
0 .6164 L
s
0 .6164 m
.39882 .92286 L
s
.39882 .92286 m
.41034 .59312 L
s
.41034 .59312 m
.05779 .25176 L
s
.68211 0 m
.95235 .42431 L
s
.95235 .42431 m
1 .77293 L
s
1 .77293 m
.70428 .3819 L
s
.70428 .3819 m
.68211 0 L
s
.05779 .25176 m
0 .6164 L
s
0 .6164 m
.70428 .3819 L
s
.70428 .3819 m
.68211 0 L
s
.68211 0 m
.05779 .25176 L
s
.41034 .59312 m
.95235 .42431 L
s
.95235 .42431 m
1 .77293 L
s
1 .77293 m
.39882 .92286 L
s
.39882 .92286 m
.41034 .59312 L
s
0 0 m
1 0 L
1 .92286 L
0 .92286 L
closepath
clip
newpath
.5 Mabswid
.6623 .51529 m .62594 .53999 L .60814 .54566 L p 1 .997 .99 r
F P
0 g
s
.62594 .53999 m .59104 .62885 L .57232 .64682 L p 1 1 .839 r
F P
0 g
s
.57232 .64682 m .60814 .54566 L .62594 .53999 L p 1 1 .839 r
F P
0 g
s
.44287 .83539 m .40985 .73965 L .38865 .72096 L p 1 .675 .319 r
F P
0 g
s
.70265 .56034 m .6623 .51529 L .64413 .49135 L p .784 .725 .982 r
F P
0 g
s
.60814 .54566 m .64413 .49135 L .6623 .51529 L p 1 .997 .99 r
F P
0 g
s
.38865 .72096 m .42148 .8106 L .44287 .83539 L p 1 .675 .319 r
F P
0 g
s
.42148 .8106 m .38865 .72096 L .36691 .7018 L p 1 .642 .268 r
F P
0 g
s
.64413 .49135 m .68264 .49831 L .70265 .56034 L p .784 .725 .982 r
F P
0 g
s
.68264 .49831 m .64413 .49135 L .6261 .48511 L p .977 .892 1 r
F P
0 g
s
.64413 .49135 m .60814 .54566 L .58977 .56157 L p 1 1 .926 r
F P
0 g
s
.58977 .56157 m .6261 .48511 L .64413 .49135 L p 1 1 .926 r
F P
0 g
s
.59104 .62885 m .55517 .74997 L .53503 .76137 L p 1 1 .771 r
F P
0 g
s
.53503 .76137 m .57232 .64682 L .59104 .62885 L p 1 1 .771 r
F P
0 g
s
.60814 .54566 m .57232 .64682 L .55268 .66705 L p 1 1 .774 r
F P
0 g
s
.55268 .66705 m .58977 .56157 L .60814 .54566 L p 1 1 .774 r
F P
0 g
s
.36691 .7018 m .39962 .78499 L .42148 .8106 L p 1 .642 .268 r
F P
0 g
s
.39962 .78499 m .36691 .7018 L .34461 .68214 L p 1 .6 .204 r
F P
0 g
s
.7481 .65397 m .70265 .56034 L .68264 .49831 L p .496 .483 .883 r
F P
0 g
s
.68264 .49831 m .72557 .56115 L .7481 .65397 L p .496 .483 .883 r
F P
0 g
s
.72557 .56115 m .68264 .49831 L .66353 .45688 L p .696 .638 .944 r
F P
0 g
s
.6261 .48511 m .66353 .45688 L .68264 .49831 L p .977 .892 1 r
F P
0 g
s
.66353 .45688 m .6261 .48511 L .60786 .49541 L p 1 1 .994 r
F P
0 g
s
.6261 .48511 m .58977 .56157 L .57056 .58509 L p 1 1 .817 r
F P
0 g
s
.57056 .58509 m .60786 .49541 L .6261 .48511 L p 1 1 .817 r
F P
0 g
s
.66353 .45688 m .70408 .48168 L .72557 .56115 L p .696 .638 .944 r
F P
0 g
s
.70408 .48168 m .66353 .45688 L .64507 .43869 L p .887 .805 .995 r
F P
0 g
s
.60786 .49541 m .64507 .43869 L .66353 .45688 L p 1 1 .994 r
F P
0 g
s
.47918 .88387 m .44287 .83539 L .42148 .8106 L p .821 .335 .08 r
F P
0 g
s
.34461 .68214 m .37728 .75856 L .39962 .78499 L p 1 .6 .204 r
F P
0 g
s
.37728 .75856 m .34461 .68214 L .32173 .66198 L p .979 .543 .122 r
F P
0 g
s
.57232 .64682 m .53503 .76137 L .51398 .76979 L p 1 1 .762 r
F P
0 g
s
.51398 .76979 m .55268 .66705 L .57232 .64682 L p 1 1 .762 r
F P
0 g
s
.55517 .74997 m .51745 .85128 L .49622 .84482 L p 1 1 .863 r
F P
0 g
s
.49622 .84482 m .53503 .76137 L .55517 .74997 L p 1 1 .863 r
F P
0 g
s
.58977 .56157 m .55268 .66705 L .532 .68705 L p 1 1 .707 r
F P
0 g
s
.532 .68705 m .57056 .58509 L .58977 .56157 L p 1 1 .707 r
F P
0 g
s
.64507 .43869 m .68394 .42566 L .70408 .48168 L p .887 .805 .995 r
F P
0 g
s
.62682 .44358 m .66502 .3986 L .68394 .42566 L p 1 .998 1 r
F P
0 g
s
.64507 .43869 m .62682 .44358 L p .64982 .43636 L 1 .998 1 r
F P
0 g
s
.58899 .51939 m .62682 .44358 L .64507 .43869 L p 1 1 .872 r
F P
0 g
s
.62682 .44358 m .58899 .51939 L .58057 .53341 L p .963 1 .608 r
F P
0 g
s
.58057 .53341 m .56914 .55244 L p .59375 .50599 L .963 1 .608 r
F P
0 g
s
.59375 .50599 m .62682 .44358 L .58057 .53341 L .963 1 .608 r
F
.60826 .46846 m .62249 .44373 L p .62384 .44337 L 1 1 .9 r
F P
0 g
s
.56914 .55244 m .60826 .46846 L .62682 .44358 L p .963 1 .608 r
F P
0 g
s
.62794 .44226 m .62682 .44358 L .61817 .45518 L p 1 1 .9 r
F P
0 g
s
.61817 .45518 m .60826 .46846 L p .62794 .44226 L 1 1 .9 r
F P
0 g
s
.60826 .46846 m .62057 .4533 L .62794 .44226 L 1 1 .9 r
F
.66502 .3986 m .62962 .44027 L p .62057 .4533 L F P
0 g
s
.62962 .44027 m .62794 .44226 L p .62057 .4533 L 1 1 .9 r
F P
0 g
s
.61689 .45784 m .62384 .44337 L .63 .44171 L 1 1 .9 r
F
.60826 .46846 m .62384 .44337 L .61689 .45784 L F
.62249 .44373 m .64682 .40145 L p .63 .44171 L F P
0 g
s
.60826 .46846 m .56914 .55244 L .548 .58828 L p .002 .002 .002 r
F P
0 g
s
.64682 .40145 m .60826 .46846 L .59723 .49051 L p .773 .893 .46 r
F P
0 g
s
.59723 .49051 m .58882 .50732 L p .61034 .46804 L .773 .893 .46 r
F P
0 g
s
.61034 .46804 m .64682 .40145 L .59723 .49051 L .773 .893 .46 r
F
.62865 .43056 m .64682 .40145 L p .61034 .46804 L F P
0 g
s
.64682 .40145 m .66502 .3986 L p .63 .44171 L 1 1 .9 r
F P
0 g
s
.66502 .3986 m .64682 .40145 L p .66744 .39768 L 1 .985 1 r
F P
0 g
s
.70512 .3908 m .67668 .39633 L p .66744 .39768 L 1 .985 1 r
F P
0 g
s
.67668 .39633 m .66502 .3986 L p .66744 .39768 L 1 .985 1 r
F P
0 g
s
.64682 .40145 m .68593 .36046 L .70512 .3908 L p 1 .985 1 r
F P
0 g
s
.68593 .36046 m .64682 .40145 L .63545 .41967 L p 1 1 .852 r
F P
0 g
s
.63737 .41989 m .6506 .4037 L .63816 .41681 L 1 1 .852 r
F
.6506 .4037 m .68593 .36046 L .63545 .41967 L F
.64109 .41534 m .64989 .40025 L .65389 .39967 L F
.64989 .40025 m .66939 .36682 L .65389 .39967 L F
.63545 .41967 m .62865 .43056 L p .63737 .41989 L F P
0 g
s
.63737 .41989 m .63816 .41681 L .63545 .41967 L 1 1 .852 r
F
.62865 .43056 m .66806 .36734 L .66939 .36682 L p F P
0 g
s
.62865 .43056 m .66939 .36682 L .64109 .41534 L 1 1 .852 r
F
.65054 .40598 m .66806 .36734 L p .63464 .43045 L .048 .048 .048 r
F P
0 g
s
.60972 .4775 m .63303 .43667 L p .63464 .43045 L .048 .048 .048 r
F P
0 g
s
.63303 .43667 m .65054 .40598 L p .63464 .43045 L .048 .048 .048 r
F P
0 g
s
.66806 .36734 m .70795 .32426 L .71585 .33492 L p 1 1 1 r
F P
0 g
s
.68593 .36046 m .72692 .34987 L .74858 .41952 L p .86 .766 .971 r
F P
0 g
s
.66939 .36682 m .68593 .36046 L p .65389 .39967 L 1 1 .852 r
F P
0 g
s
.74858 .41952 m .70512 .3908 L .68593 .36046 L p .86 .766 .971 r
F P
0 g
s
.68593 .36046 m .66806 .36734 L p .6847 .3624 L 1 1 1 r
F P
0 g
s
.72692 .34987 m .68985 .35944 L p .6847 .3624 L 1 1 1 r
F P
0 g
s
.68985 .35944 m .68593 .36046 L p .6847 .3624 L 1 1 1 r
F P
0 g
s
.66806 .36734 m .71585 .33492 L .69177 .3603 L 1 1 1 r
F
.70795 .32426 m .66806 .36734 L .65558 .39487 L p .795 .957 .573 r
F P
0 g
s
.65558 .39487 m .65054 .40598 L p .66886 .37991 L .795 .957 .573 r
F P
0 g
s
.66886 .37991 m .70795 .32426 L .65558 .39487 L .795 .957 .573 r
F
.69092 .3433 m .70795 .32426 L p .67877 .3658 L F P
0 g
s
.77104 .36947 m .72692 .34987 L .70795 .32426 L p .899 .802 .981 r
F P
0 g
s
.70795 .32426 m .74945 .30495 L .77104 .36947 L p .899 .802 .981 r
F P
0 g
s
.74945 .30495 m .70795 .32426 L .70012 .33301 L p 1 1 1 r
F P
0 g
s
.70012 .33301 m .69092 .3433 L p .70177 .33619 L 1 1 1 r
F P
0 g
s
.70177 .33619 m .74945 .30495 L .70012 .33301 L 1 1 1 r
F
.69092 .3433 m .73155 .29453 L .74018 .29956 L p F P
0 g
s
.74018 .29956 m .74945 .30495 L p .71585 .32697 L 1 1 1 r
F P
0 g
s
.69092 .3433 m .74018 .29956 L .71585 .32697 L 1 1 1 r
F
.71612 .33466 m .73155 .29453 L p .70497 .34235 L .095 .095 .095 r
F P
0 g
s
.71585 .33492 m .72692 .34987 L p .69177 .3603 L 1 1 1 r
F P
0 g
s
.66502 .3986 m .70512 .3908 L .72618 .45679 L p .852 .766 .977 r
F P
0 g
s
.67448 .39724 m .70762 .34744 L p .70497 .34235 L .095 .095 .095 r
F P
0 g
s
.70762 .34744 m .71612 .33466 L p .70497 .34235 L .095 .095 .095 r
F P
0 g
s
.65054 .40598 m .67506 .36793 L p .67877 .3658 L .795 .957 .573 r
F P
0 g
s
.67506 .36793 m .69092 .3433 L p .67877 .3658 L .795 .957 .573 r
F P
0 g
s
.58882 .50732 m .60584 .47451 L p .61034 .46804 L .773 .893 .46 r
F P
0 g
s
.60584 .47451 m .62865 .43056 L p .61034 .46804 L .773 .893 .46 r
F P
0 g
s
.68394 .42566 m .66105 .43333 L p .64982 .43636 L 1 .998 1 r
F P
0 g
s
.66105 .43333 m .64507 .43869 L p .64982 .43636 L 1 .998 1 r
F P
0 g
s
.64507 .43869 m .60786 .49541 L .58899 .51939 L p 1 1 .872 r
F P
0 g
s
.42148 .8106 m .45764 .8622 L .47918 .88387 L p .821 .335 .08 r
F P
0 g
s
.45764 .8622 m .42148 .8106 L .39962 .78499 L p .809 .33 .073 r
F P
0 g
s
.32173 .66198 m .35444 .7313 L .37728 .75856 L p .979 .543 .122 r
F P
0 g
s
.35444 .7313 m .32173 .66198 L .29825 .64128 L p .915 .465 .1 r
F P
0 g
s
.77317 .65445 m .72557 .56115 L .70408 .48168 L p .507 .48 .874 r
F P
0 g
s
.70408 .48168 m .74911 .54983 L .77317 .65445 L p .507 .48 .874 r
F P
0 g
s
.74911 .54983 m .70408 .48168 L .68394 .42566 L p .679 .611 .922 r
F P
0 g
s
.68394 .42566 m .72618 .45679 L .74911 .54983 L p .679 .611 .922 r
F P
0 g
s
.72618 .45679 m .68394 .42566 L .66502 .3986 L p .852 .766 .977 r
F P
0 g
s
.60786 .49541 m .57056 .58509 L .55031 .61268 L p 1 1 .669 r
F P
0 g
s
.55031 .61268 m .58899 .51939 L .60786 .49541 L p 1 1 .669 r
F P
0 g
s
.79689 .75015 m .7481 .65397 L .72557 .56115 L p .294 .339 .816 r
F P
0 g
s
.72557 .56115 m .77317 .65445 L .79689 .75015 L p .294 .339 .816 r
F P
0 g
s
.39962 .78499 m .4356 .83827 L .45764 .8622 L p .809 .33 .073 r
F P
0 g
s
.4356 .83827 m .39962 .78499 L .37728 .75856 L p .777 .304 .061 r
F P
0 g
s
.51745 .85128 m .47918 .88387 L .45764 .8622 L p .883 1 1 r
F P
0 g
s
.45764 .8622 m .49622 .84482 L .51745 .85128 L p .883 1 1 r
F P
0 g
s
.29825 .64128 m .3311 .70322 L .35444 .7313 L p .915 .465 .1 r
F P
0 g
s
.3311 .70322 m .29825 .64128 L .27415 .62003 L p .819 .355 .071 r
F P
0 g
s
.53503 .76137 m .49622 .84482 L .47433 .83439 L p 1 1 .92 r
F P
0 g
s
.47433 .83439 m .51398 .76979 L .53503 .76137 L p 1 1 .92 r
F P
0 g
s
.55268 .66705 m .51398 .76979 L .49205 .77381 L p 1 1 .771 r
F P
0 g
s
.49205 .77381 m .532 .68705 L .55268 .66705 L p 1 1 .771 r
F P
0 g
s
.37728 .75856 m .41308 .81208 L .4356 .83827 L p .777 .304 .061 r
F P
0 g
s
.41308 .81208 m .37728 .75856 L .35444 .7313 L p .721 .255 .043 r
F P
0 g
s
.57056 .58509 m .532 .68705 L .51025 .70409 L p 1 1 .648 r
F P
0 g
s
.51025 .70409 m .55031 .61268 L .57056 .58509 L p 1 1 .648 r
F P
0 g
s
.49622 .84482 m .45764 .8622 L .4356 .83827 L p .625 .966 1 r
F P
0 g
s
.7728 .52228 m .72618 .45679 L .70512 .3908 L p .699 .616 .914 r
F P
0 g
s
.70512 .3908 m .74858 .41952 L .7728 .52228 L p .699 .616 .914 r
F P
0 g
s
.79838 .63712 m .74911 .54983 L .72618 .45679 L p .551 .504 .876 r
F P
0 g
s
.72618 .45679 m .7728 .52228 L .79838 .63712 L p .551 .504 .876 r
F P
0 g
s
.91097 .71256 m .94399 .60605 L .94522 .51517 L p .325 .192 .535 r
F P
0 g
s
.91683 .64045 m .91097 .71256 L p .93691 .56306 L .325 .192 .535 r
F P
0 g
s
.84412 .79304 m .83489 .76812 L p .85666 .76047 L .792 .269 .169 r
F P
0 g
s
.88452 .75069 m .86221 .77407 L p .85666 .76047 L .792 .269 .169 r
F P
0 g
s
.86221 .77407 m .84412 .79304 L p .85666 .76047 L .792 .269 .169 r
F P
0 g
s
.85608 .74393 m .88452 .75069 L .83489 .76812 L .792 .269 .169 r
F
.83489 .76812 m .82302 .73607 L p .85608 .74393 L F P
0 g
s
.82302 .73607 m .87042 .7621 L .88452 .75069 L p .792 .269 .169 r
F P
0 g
s
.84937 .70823 m .87042 .7621 L p .82567 .68447 L .155 .282 .818 r
F P
0 g
s
.79838 .63712 m .81689 .66293 L p .82567 .68447 L .155 .282 .818 r
F P
0 g
s
.81689 .66293 m .84937 .70823 L p .82567 .68447 L .155 .282 .818 r
F P
0 g
s
.89531 .71387 m .87869 .74607 L p .91683 .64045 L .523 .255 .547 r
F P
0 g
s
.87849 .74093 m .91683 .64045 L .87869 .74607 L .523 .255 .547 r
F
.87042 .7621 m .91097 .71256 L .91377 .67809 L p F P
0 g
s
.87869 .74607 m .87042 .7621 L p .87849 .74093 L .523 .255 .547 r
F P
0 g
s
.87042 .7621 m .91377 .67809 L .89375 .70094 L .523 .255 .547 r
F
.91097 .71256 m .87042 .7621 L .85945 .73404 L p .76 .252 .187 r
F P
0 g
s
.85945 .73404 m .84937 .70823 L p .88016 .7104 L .76 .252 .187 r
F P
0 g
s
.88016 .7104 m .91097 .71256 L .85945 .73404 L .76 .252 .187 r
F
.84937 .70823 m .89782 .72883 L .91097 .71256 L p F P
0 g
s
.84795 .54849 m .90201 .61314 L .91092 .64237 L p .426 .471 .904 r
F P
0 g
s
.92881 .65503 m .91097 .71256 L .90143 .72436 L p .566 .293 .566 r
F P
0 g
s
.90209 .71866 m .92881 .65503 L .90143 .72436 L .566 .293 .566 r
F
.94399 .60605 m .92881 .65503 L p .92533 .66331 L F P
0 g
s
.89782 .72883 m .92104 .66707 L .92533 .66331 L .566 .293 .566 r
F
.92104 .66707 m .94399 .60605 L .92533 .66331 L F
.90143 .72436 m .89782 .72883 L p .90209 .71866 L F P
0 g
s
.89782 .72883 m .92486 .69982 L p .92882 .65483 L .566 .293 .566 r
F P
0 g
s
.89782 .72883 m .92882 .65483 L .92326 .66118 L .566 .293 .566 r
F
.91377 .67809 m .91683 .64045 L p .89375 .70094 L .523 .255 .547 r
F P
0 g
s
.92486 .69982 m .93986 .68374 L p .9298 .6438 L .566 .293 .566 r
F P
0 g
s
.93986 .68374 m .89782 .72883 L .87586 .66811 L p .719 .165 .146 r
F P
0 g
s
.92637 .69308 m .93986 .68374 L p .91027 .67651 L .719 .165 .146 r
F P
0 g
s
.87586 .66811 m .90399 .68202 L p .91027 .67651 L .719 .165 .146 r
F P
0 g
s
.90399 .68202 m .92637 .69308 L p .91027 .67651 L .719 .165 .146 r
F P
0 g
s
.91092 .64237 m .92637 .69308 L p .9088 .66068 L .426 .471 .904 r
F P
0 g
s
.84795 .54849 m .91092 .64237 L .9088 .66068 L .426 .471 .904 r
F
.92326 .66118 m .92882 .65483 L .9298 .6438 L .566 .293 .566 r
F
.87586 .66811 m .89782 .72883 L p .87664 .69273 L .247 .385 .88 r
F P
0 g
s
.82344 .60201 m .87569 .66789 L p .87664 .69273 L .247 .385 .88 r
F P
0 g
s
.87569 .66789 m .87586 .66811 L p .87664 .69273 L .247 .385 .88 r
F P
0 g
s
.88717 .74164 m .88452 .75069 L p .87042 .7621 L .523 .255 .547 r
F P
0 g
s
.93986 .68374 m .94399 .60605 L p .9298 .6438 L .566 .293 .566 r
F P
0 g
s
.94522 .51517 m .93853 .5447 L p .93691 .56306 L .325 .192 .535 r
F P
0 g
s
.93853 .5447 m .91683 .64045 L p .93691 .56306 L .325 .192 .535 r
F P
0 g
s
.27415 .62003 m .30724 .6743 L .3311 .70322 L p .819 .355 .071 r
F P
0 g
s
.30724 .6743 m .27415 .62003 L .24939 .5982 L p .673 .203 .031 r
F P
0 g
s
.4356 .83827 m .47433 .83439 L .49622 .84482 L p .625 .966 1 r
F P
0 g
s
.35444 .7313 m .3901 .78366 L .41308 .81208 L p .721 .255 .043 r
F P
0 g
s
.3901 .78366 m .35444 .7313 L .3311 .70322 L p .636 .179 .018 r
F P
0 g
s
.58899 .51939 m .55031 .61268 L .52884 .64002 L p .868 .94 .507 r
F P
0 g
s
.52884 .64002 m .56914 .55244 L .58899 .51939 L p .868 .94 .507 r
F P
0 g
s
.82302 .73607 m .77317 .65445 L .74911 .54983 L p .362 .388 .841 r
F P
0 g
s
.74911 .54983 m .79838 .63712 L .82302 .73607 L p .362 .388 .841 r
F P
0 g
s
.47433 .83439 m .4356 .83827 L .41308 .81208 L p .334 .754 1 r
F P
0 g
s
.51398 .76979 m .47433 .83439 L .45181 .81964 L p 1 1 1 r
F P
0 g
s
.45181 .81964 m .49205 .77381 L .51398 .76979 L p 1 1 1 r
F P
0 g
s
.24939 .5982 m .28284 .64454 L .30724 .6743 L p .673 .203 .031 r
F P
0 g
s
.28284 .64454 m .24939 .5982 L .22395 .57578 L p .021 .021 .536 r
F P
0 g
s
.532 .68705 m .49205 .77381 L .46929 .77217 L p 1 1 .81 r
F P
0 g
s
.46929 .77217 m .51025 .70409 L .532 .68705 L p 1 1 .81 r
F P
0 g
s
.3311 .70322 m .36667 .75307 L .3901 .78366 L p .636 .179 .018 r
F P
0 g
s
.36667 .75307 m .3311 .70322 L .30724 .6743 L p .014 .014 .476 r
F P
0 g
s
.41308 .81208 m .45181 .81964 L .47433 .83439 L p .334 .754 1 r
F P
0 g
s
.79626 .47704 m .74858 .41952 L .72692 .34987 L p .738 .638 .913 r
F P
0 g
s
.72692 .34987 m .77104 .36947 L .79626 .47704 L p .738 .638 .913 r
F P
0 g
s
.84412 .79304 m .79689 .75015 L .77317 .65445 L p .14 .196 .753 r
F P
0 g
s
.77317 .65445 m .82302 .73607 L .84412 .79304 L p .14 .196 .753 r
F P
0 g
s
.45181 .81964 m .41308 .81208 L .3901 .78366 L p .118 .559 .979 r
F P
0 g
s
.82344 .60201 m .7728 .52228 L .74858 .41952 L p .608 .537 .881 r
F P
0 g
s
.74858 .41952 m .79626 .47704 L .82344 .60201 L p .608 .537 .881 r
F P
0 g
s
.548 .58828 m .58882 .50732 L .60826 .46846 L p .002 .002 .002 r
F P
0 g
s
.55031 .61268 m .51025 .70409 L .48745 .71536 L p .908 1 .616 r
F P
0 g
s
.48745 .71536 m .52884 .64002 L .55031 .61268 L p .908 1 .616 r
F P
0 g
s
.30724 .6743 m .34279 .72039 L .36667 .75307 L p .014 .014 .476 r
F P
0 g
s
.34279 .72039 m .30724 .6743 L .28284 .64454 L p .054 .071 .625 r
F P
0 g
s
.22395 .57578 m .25789 .61395 L .28284 .64454 L p .021 .021 .536 r
F P
0 g
s
.25789 .61395 m .22395 .57578 L .19781 .55273 L p .083 .247 .772 r
F P
0 g
s
.49205 .77381 m .45181 .81964 L .42872 .80031 L p .955 1 1 r
F P
0 g
s
.42872 .80031 m .46929 .77217 L .49205 .77381 L p .955 1 1 r
F P
0 g
s
.3901 .78366 m .42872 .80031 L .45181 .81964 L p .118 .559 .979 r
F P
0 g
s
.42872 .80031 m .3901 .78366 L .36667 .75307 L p .107 .439 .91 r
F P
0 g
s
.84937 .70823 m .79838 .63712 L .7728 .52228 L p .449 .444 .863 r
F P
0 g
s
.7728 .52228 m .82344 .60201 L .84937 .70823 L p .449 .444 .863 r
F P
0 g
s
.66806 .36734 m .62865 .43056 L .60972 .4775 L p .048 .048 .048 r
F P
0 g
s
.56914 .55244 m .52884 .64002 L .50612 .66229 L p .655 .819 .384 r
F P
0 g
s
.50612 .66229 m .548 .58828 L .56914 .55244 L p .655 .819 .384 r
F P
0 g
s
.84795 .54849 m .79626 .47704 L .77104 .36947 L p .667 .572 .884 r
F P
0 g
s
.77104 .36947 m .81915 .41446 L .84795 .54849 L p .667 .572 .884 r
F P
0 g
s
.81915 .41446 m .77104 .36947 L .74945 .30495 L p .787 .675 .92 r
F P
0 g
s
.74945 .30495 m .79365 .31019 L .81915 .41446 L p .787 .675 .92 r
F P
0 g
s
.79365 .31019 m .74945 .30495 L .73155 .29453 L p .978 .898 1 r
F P
0 g
s
.28284 .64454 m .31849 .68569 L .34279 .72039 L p .054 .071 .625 r
F P
0 g
s
.31849 .68569 m .28284 .64454 L .25789 .61395 L p .1 .24 .782 r
F P
0 g
s
.51025 .70409 m .46929 .77217 L .44578 .76376 L p .977 1 .897 r
F P
0 g
s
.44578 .76376 m .48745 .71536 L .51025 .70409 L p .977 1 .897 r
F P
0 g
s
.19781 .55273 m .23236 .5825 L .25789 .61395 L p .083 .247 .772 r
F P
0 g
s
.23236 .5825 m .19781 .55273 L .17092 .52904 L p .147 .492 .968 r
F P
0 g
s
.36667 .75307 m .40511 .77628 L .42872 .80031 L p .107 .439 .91 r
F P
0 g
s
.40511 .77628 m .36667 .75307 L .34279 .72039 L p .114 .395 .892 r
F P
0 g
s
.62865 .43056 m .58882 .50732 L .56803 .55121 L p .075 .075 .075 r
F P
0 g
s
.56803 .55121 m .60972 .4775 L .62865 .43056 L p .075 .075 .075 r
F P
0 g
s
.87042 .7621 m .82302 .73607 L .79838 .63712 L p .155 .282 .818 r
F P
0 g
s
.46929 .77217 m .42872 .80031 L .40511 .77628 L p .771 1 1 r
F P
0 g
s
.40511 .77628 m .44578 .76376 L .46929 .77217 L p .771 1 1 r
F P
0 g
s
.87586 .66811 m .82344 .60201 L .79626 .47704 L p .537 .496 .875 r
F P
0 g
s
.79626 .47704 m .84795 .54849 L .87586 .66811 L p .537 .496 .875 r
F P
0 g
s
.25789 .61395 m .29376 .6491 L .31849 .68569 L p .1 .24 .782 r
F P
0 g
s
.29376 .6491 m .25789 .61395 L .23236 .5825 L p .145 .416 .921 r
F P
0 g
s
.34279 .72039 m .38105 .74752 L .40511 .77628 L p .114 .395 .892 r
F P
0 g
s
.38105 .74752 m .34279 .72039 L .31849 .68569 L p .132 .412 .914 r
F P
0 g
s
.52884 .64002 m .48745 .71536 L .46369 .71824 L p .795 .992 .65 r
F P
0 g
s
.46369 .71824 m .50612 .66229 L .52884 .64002 L p .795 .992 .65 r
F P
0 g
s
.87138 .47522 m .81915 .41446 L .79365 .31019 L p .723 .609 .887 r
F P
0 g
s
.79365 .31019 m .84144 .33788 L .87138 .47522 L p .723 .609 .887 r
F P
0 g
s
.77329 .26299 m .81709 .25028 L .84144 .33788 L p .852 .737 .943 r
F P
0 g
s
.79166 .30559 m .77329 .26299 L p .79318 .28485 L .852 .737 .943 r
F P
0 g
s
.73155 .29453 m .77329 .26299 L .79365 .31019 L p .978 .898 1 r
F P
0 g
s
.77329 .26299 m .73155 .29453 L .7192 .32666 L p .674 .899 .532 r
F P
0 g
s
.7192 .32666 m .71612 .33466 L p .73067 .31642 L .674 .899 .532 r
F P
0 g
s
.73067 .31642 m .77329 .26299 L .7192 .32666 L .674 .899 .532 r
F
.71612 .33466 m .75768 .27981 L .77329 .26299 L p F P
0 g
s
.74468 .34747 m .78783 .29014 L .78964 .28183 L p .183 .158 .255 r
F P
0 g
s
.81709 .25028 m .77329 .26299 L .7622 .27494 L p 1 1 1 r
F P
0 g
s
.76571 .27582 m .81709 .25028 L .7622 .27494 L 1 1 1 r
F
.7622 .27494 m .75768 .27981 L p .76571 .27582 L F P
0 g
s
.75768 .27981 m .79525 .25662 L .79508 .25897 L 1 1 1 r
F
.75768 .27981 m .79654 .23904 L p .79525 .25662 L F P
0 g
s
.78938 .26405 m .79508 .25897 L .79491 .26131 L 1 1 1 r
F
.75768 .27981 m .79508 .25897 L .78938 .26405 L F
.79654 .23904 m .79974 .23567 L p .79491 .26131 L F P
0 g
s
.78964 .28183 m .79974 .23567 L p .7784 .27901 L .183 .158 .255 r
F P
0 g
s
.74468 .34747 m .78964 .28183 L .7784 .27901 L .183 .158 .255 r
F
.79974 .23567 m .81709 .25028 L p .79491 .26131 L 1 1 1 r
F P
0 g
s
.84144 .33788 m .79365 .31019 L .79166 .30559 L p .852 .737 .943 r
F P
0 g
s
.79318 .28485 m .84144 .33788 L .79166 .30559 L .852 .737 .943 r
F
.17092 .52904 m .20625 .55021 L .23236 .5825 L p .147 .492 .968 r
F P
0 g
s
.20625 .55021 m .17092 .52904 L .14327 .50466 L p .431 .685 1 r
F P
0 g
s
.548 .58828 m .50612 .66229 L .49135 .66994 L p .018 .018 .018 r
F P
0 g
s
.49135 .66994 m .48224 .67466 L p .51325 .63393 L .018 .018 .018 r
F P
0 g
s
.51325 .63393 m .548 .58828 L .49135 .66994 L .018 .018 .018 r
F
.48224 .67466 m .52543 .61948 L .548 .58828 L p F P
0 g
s
.548 .58828 m .52543 .61948 L p .57007 .54049 L .064 .064 .064 r
F P
0 g
s
.58882 .50732 m .57881 .52717 L p .57007 .54049 L .064 .064 .064 r
F P
0 g
s
.57881 .52717 m .548 .58828 L p .57007 .54049 L .064 .064 .064 r
F P
0 g
s
.52543 .61948 m .56803 .55121 L .58381 .5179 L p .064 .064 .064 r
F P
0 g
s
.52543 .61948 m .58381 .5179 L .57007 .54049 L .064 .064 .064 r
F
.56803 .55121 m .52543 .61948 L .50849 .63274 L p .086 .086 .086 r
F P
0 g
s
.50849 .63274 m .50149 .63822 L p .52798 .60358 L .086 .086 .086 r
F P
0 g
s
.52798 .60358 m .56803 .55121 L .50849 .63274 L .086 .086 .086 r
F
.54563 .58903 m .56803 .55121 L p .53288 .59717 L F P
0 g
s
.56803 .55121 m .54563 .58903 L p .58392 .52239 L .124 .124 .124 r
F P
0 g
s
.60972 .4775 m .59352 .50614 L p .58392 .52239 L .124 .124 .124 r
F P
0 g
s
.59352 .50614 m .56803 .55121 L p .58392 .52239 L .124 .124 .124 r
F P
0 g
s
.54563 .58903 m .58933 .52924 L .60605 .4868 L p .124 .124 .124 r
F P
0 g
s
.54563 .58903 m .60605 .4868 L .59409 .50469 L .124 .124 .124 r
F
.50149 .63822 m .51809 .61972 L p .53288 .59717 L .086 .086 .086 r
F P
0 g
s
.51809 .61972 m .54563 .58903 L p .53288 .59717 L .086 .086 .086 r
F P
0 g
s
.54563 .58903 m .50149 .63822 L .47646 .63772 L p .044 .044 .044 r
F P
0 g
s
.47646 .63772 m .52164 .60914 L .54563 .58903 L p .044 .044 .044 r
F P
0 g
s
.58933 .52924 m .54563 .58903 L .52164 .60914 L p .126 .126 .126 r
F P
0 g
s
.56705 .56954 m .58933 .52924 L p .5581 .56611 L .126 .126 .126 r
F P
0 g
s
.52164 .60914 m .54441 .58929 L p .5581 .56611 L .126 .126 .126 r
F P
0 g
s
.54441 .58929 m .56705 .56954 L p .5581 .56611 L .126 .126 .126 r
F P
0 g
s
.60605 .4868 m .60972 .4775 L p .59409 .50469 L .124 .124 .124 r
F P
0 g
s
.58381 .5179 m .58882 .50732 L p .57007 .54049 L .064 .064 .064 r
F P
0 g
s
.73155 .29453 m .69092 .3433 L .67448 .39724 L p .095 .095 .095 r
F P
0 g
s
.44578 .76376 m .40511 .77628 L .38105 .74752 L p .573 .903 1 r
F P
0 g
s
.69092 .3433 m .65054 .40598 L .63229 .46365 L p .13 .13 .152 r
F P
0 g
s
.63229 .46365 m .67448 .39724 L .69092 .3433 L p .13 .13 .152 r
F P
0 g
s
.23236 .5825 m .2686 .61074 L .29376 .6491 L p .145 .416 .921 r
F P
0 g
s
.2686 .61074 m .23236 .5825 L .20625 .55021 L p .343 .574 1 r
F P
0 g
s
.31849 .68569 m .35659 .71415 L .38105 .74752 L p .132 .412 .914 r
F P
0 g
s
.35659 .71415 m .31849 .68569 L .29376 .6491 L p .182 .47 .958 r
F P
0 g
s
.48745 .71536 m .44578 .76376 L .42164 .74774 L p .873 1 1 r
F P
0 g
s
.42164 .74774 m .46369 .71824 L .48745 .71536 L p .873 1 1 r
F P
0 g
s
.90201 .61314 m .84795 .54849 L .81915 .41446 L p .617 .539 .877 r
F P
0 g
s
.81915 .41446 m .87138 .47522 L .90201 .61314 L p .617 .539 .877 r
F P
0 g
s
.38105 .74752 m .42164 .74774 L .44578 .76376 L p .573 .903 1 r
F P
0 g
s
.89782 .72883 m .84937 .70823 L .82344 .60201 L p .247 .385 .88 r
F P
0 g
s
.14327 .50466 m .17952 .51707 L .20625 .55021 L p .431 .685 1 r
F P
0 g
s
.17952 .51707 m .14327 .50466 L .11482 .47958 L p .645 .812 1 r
F P
0 g
s
.65054 .40598 m .60972 .4775 L .58933 .52924 L p .136 .136 .172 r
F P
0 g
s
.58933 .52924 m .63229 .46365 L .65054 .40598 L p .136 .136 .172 r
F P
0 g
s
.42164 .74774 m .38105 .74752 L .35659 .71415 L p .452 .761 1 r
F P
0 g
s
.29376 .6491 m .3318 .67639 L .35659 .71415 L p .182 .47 .958 r
F P
0 g
s
.3318 .67639 m .29376 .6491 L .2686 .61074 L p .326 .549 1 r
F P
0 g
s
.20625 .55021 m .24302 .57075 L .2686 .61074 L p .343 .574 1 r
F P
0 g
s
.24302 .57075 m .20625 .55021 L .17952 .51707 L p .541 .698 1 r
F P
0 g
s
.50612 .66229 m .46369 .71824 L .43914 .71053 L p .718 1 .809 r
F P
0 g
s
.43914 .71053 m .48224 .67466 L .50612 .66229 L p .718 1 .809 r
F P
0 g
s
.35659 .71415 m .39699 .72365 L .42164 .74774 L p .452 .761 1 r
F P
0 g
s
.39699 .72365 m .35659 .71415 L .3318 .67639 L p .43 .684 1 r
F P
0 g
s
.2686 .61074 m .30672 .63459 L .3318 .67639 L p .326 .549 1 r
F P
0 g
s
.30672 .63459 m .2686 .61074 L .24302 .57075 L p .474 .631 1 r
F P
0 g
s
.46369 .71824 m .42164 .74774 L .39699 .72365 L p .739 1 1 r
F P
0 g
s
.11482 .47958 m .15216 .48306 L .17952 .51707 L p .645 .812 1 r
F P
0 g
s
.15216 .48306 m .11482 .47958 L .08553 .45376 L p .792 .887 1 r
F P
0 g
s
.92637 .69308 m .87586 .66811 L .84795 .54849 L p .426 .471 .904 r
F P
0 g
s
.39699 .72365 m .43914 .71053 L .46369 .71824 L p .739 1 1 r
F P
0 g
s
.17952 .51707 m .217 .52927 L .24302 .57075 L p .541 .698 1 r
F P
0 g
s
.217 .52927 m .17952 .51707 L .15216 .48306 L p .693 .785 1 r
F P
0 g
s
.79974 .23567 m .75768 .27981 L .74468 .34747 L p .183 .158 .255 r
F P
0 g
s
.75768 .27981 m .71612 .33466 L .70113 .40691 L p .275 .174 .342 r
F P
0 g
s
.70113 .40691 m .74468 .34747 L .75768 .27981 L p .275 .174 .342 r
F P
0 g
s
.24302 .57075 m .28138 .58917 L .30672 .63459 L p .474 .631 1 r
F P
0 g
s
.28138 .58917 m .24302 .57075 L .217 .52927 L p .609 .703 1 r
F P
0 g
s
.3318 .67639 m .37199 .69139 L .39699 .72365 L p .43 .684 1 r
F P
0 g
s
.37199 .69139 m .3318 .67639 L .30672 .63459 L p .479 .665 1 r
F P
0 g
s
.71612 .33466 m .67448 .39724 L .65701 .46569 L p .299 .177 .358 r
F P
0 g
s
.65701 .46569 m .70113 .40691 L .71612 .33466 L p .299 .177 .358 r
F P
0 g
s
.43914 .71053 m .39699 .72365 L .37199 .69139 L p .629 .912 1 r
F P
0 g
s
.15216 .48306 m .19054 .48645 L .217 .52927 L p .693 .785 1 r
F P
0 g
s
.19054 .48645 m .15216 .48306 L .12414 .44818 L p .805 .843 1 r
F P
0 g
s
.08553 .45376 m .12414 .44818 L .15216 .48306 L p .792 .887 1 r
F P
0 g
s
.12414 .44818 m .08553 .45376 L .05536 .42717 L p .89 .928 1 r
F P
0 g
s
.52543 .61948 m .48224 .67466 L .45741 .67292 L p .441 .814 .559 r
F P
0 g
s
.45741 .67292 m .50149 .63822 L .52543 .61948 L p .441 .814 .559 r
F P
0 g
s
.30672 .63459 m .34675 .65127 L .37199 .69139 L p .479 .665 1 r
F P
0 g
s
.34675 .65127 m .30672 .63459 L .28138 .58917 L p .563 .681 1 r
F P
0 g
s
.217 .52927 m .25581 .54065 L .28138 .58917 L p .609 .703 1 r
F P
0 g
s
.25581 .54065 m .217 .52927 L .19054 .48645 L p .72 .762 1 r
F P
0 g
s
.67448 .39724 m .63229 .46365 L .61227 .52086 L p .281 .173 .324 r
F P
0 g
s
.61227 .52086 m .65701 .46569 L .67448 .39724 L p .281 .173 .324 r
F P
0 g
s
.48224 .67466 m .43914 .71053 L .41403 .69078 L p .695 1 1 r
F P
0 g
s
.37199 .69139 m .41403 .69078 L .43914 .71053 L p .629 .912 1 r
F P
0 g
s
.41403 .69078 m .45741 .67292 L .48224 .67466 L p .695 1 1 r
F P
0 g
s
.41403 .69078 m .37199 .69139 L .34675 .65127 L p .598 .795 1 r
F P
0 g
s
.12414 .44818 m .16361 .44246 L .19054 .48645 L p .805 .843 1 r
F P
0 g
s
.16361 .44246 m .12414 .44818 L .09544 .41243 L p .884 .88 1 r
F P
0 g
s
.05536 .42717 m .09544 .41243 L .12414 .44818 L p .89 .928 1 r
F P
0 g
s
.28138 .58917 m .32139 .60399 L .34675 .65127 L p .563 .681 1 r
F P
0 g
s
.32139 .60399 m .28138 .58917 L .25581 .54065 L p .654 .711 1 r
F P
0 g
s
.19054 .48645 m .23 .48958 L .25581 .54065 L p .72 .762 1 r
F P
0 g
s
.23 .48958 m .19054 .48645 L .16361 .44246 L p .808 .806 1 r
F P
0 g
s
.56705 .56954 m .61227 .52086 L .63229 .46365 L p .218 .159 .23 r
F P
0 g
s
.58933 .52924 m .56705 .56954 L p .59678 .52129 L .218 .159 .23 r
F P
0 g
s
.56705 .56954 m .52164 .60914 L .49645 .60204 L p .024 .024 .024 r
F P
0 g
s
.49645 .60204 m .54293 .58233 L .56705 .56954 L p .024 .024 .024 r
F P
0 g
s
.61227 .52086 m .56705 .56954 L .54293 .58233 L p .174 .131 .131 r
F P
0 g
s
.58991 .5547 m .61227 .52086 L p .58458 .54541 L .174 .131 .131 r
F P
0 g
s
.54293 .58233 m .57564 .5631 L p .58458 .54541 L .174 .131 .131 r
F P
0 g
s
.57564 .5631 m .58991 .5547 L p .58458 .54541 L .174 .131 .131 r
F P
0 g
s
.63229 .46365 m .60435 .5063 L p .59678 .52129 L .218 .159 .23 r
F P
0 g
s
.60435 .5063 m .58933 .52924 L p .59678 .52129 L .218 .159 .23 r
F P
0 g
s
.25581 .54065 m .296 .55056 L .32139 .60399 L p .654 .711 1 r
F P
0 g
s
.296 .55056 m .25581 .54065 L .23 .48958 L p .738 .745 1 r
F P
0 g
s
.34675 .65127 m .38858 .65839 L .41403 .69078 L p .598 .795 1 r
F P
0 g
s
.38858 .65839 m .34675 .65127 L .32139 .60399 L p .629 .739 1 r
F P
0 g
s
.45741 .67292 m .41403 .69078 L .38858 .65839 L p .68 .954 1 r
F P
0 g
s
.16361 .44246 m .20395 .43657 L .23 .48958 L p .808 .806 1 r
F P
0 g
s
.20395 .43657 m .16361 .44246 L .13618 .39745 L p .876 .84 1 r
F P
0 g
s
.09544 .41243 m .13618 .39745 L .16361 .44246 L p .884 .88 1 r
F P
0 g
s
.23 .48958 m .27061 .49225 L .296 .55056 L p .738 .745 1 r
F P
0 g
s
.27061 .49225 m .23 .48958 L .20395 .43657 L p .809 .777 1 r
F P
0 g
s
.32139 .60399 m .36305 .61379 L .38858 .65839 L p .629 .739 1 r
F P
0 g
s
.36305 .61379 m .32139 .60399 L .296 .55056 L p .687 .726 1 r
F P
0 g
s
.50149 .63822 m .45741 .67292 L .43195 .65417 L p .594 1 1 r
F P
0 g
s
.38858 .65839 m .43195 .65417 L .45741 .67292 L p .68 .954 1 r
F P
0 g
s
.43195 .65417 m .47646 .63772 L .50149 .63822 L p .594 1 1 r
F P
0 g
s
.13618 .39745 m .17759 .38222 L .20395 .43657 L p .876 .84 1 r
F P
0 g
s
.20395 .43657 m .2452 .43048 L .27061 .49225 L p .809 .777 1 r
F P
0 g
s
.2452 .43048 m .20395 .43657 L .17759 .38222 L p .867 .806 1 r
F P
0 g
s
.78783 .29014 m .74468 .34747 L .73103 .43209 L p .473 .201 .474 r
F P
0 g
s
.73103 .43209 m .77746 .3821 L .78783 .29014 L p .473 .201 .474 r
F P
0 g
s
.43195 .65417 m .38858 .65839 L .36305 .61379 L p .681 .819 1 r
F P
0 g
s
.296 .55056 m .33761 .55836 L .36305 .61379 L p .687 .726 1 r
F P
0 g
s
.33761 .55836 m .296 .55056 L .27061 .49225 L p .75 .736 1 r
F P
0 g
s
.74468 .34747 m .70113 .40691 L .6842 .47846 L p .455 .197 .442 r
F P
0 g
s
.6842 .47846 m .73103 .43209 L .74468 .34747 L p .455 .197 .442 r
F P
0 g
s
.27061 .49225 m .31238 .49425 L .33761 .55836 L p .75 .736 1 r
F P
0 g
s
.31238 .49425 m .27061 .49225 L .2452 .43048 L p .808 .755 .999 r
F P
0 g
s
.17759 .38222 m .2197 .36674 L .2452 .43048 L p .867 .806 1 r
F P
0 g
s
.36305 .61379 m .40624 .61734 L .43195 .65417 L p .681 .819 1 r
F P
0 g
s
.40624 .61734 m .36305 .61379 L .33761 .55836 L p .713 .752 1 r
F P
0 g
s
.2452 .43048 m .28738 .42415 L .31238 .49425 L p .808 .755 .999 r
F P
0 g
s
.28738 .42415 m .2452 .43048 L .2197 .36674 L p .858 .778 .987 r
F P
0 g
s
.70113 .40691 m .65701 .46569 L .63709 .51975 L p .416 .189 .378 r
F P
0 g
s
.63709 .51975 m .6842 .47846 L .70113 .40691 L p .416 .189 .378 r
F P
0 g
s
.47646 .63772 m .43195 .65417 L .40624 .61734 L p .711 .933 1 r
F P
0 g
s
.2197 .36674 m .26252 .35099 L .28738 .42415 L p .858 .778 .987 r
F P
0 g
s
.33761 .55836 m .38063 .56344 L .40624 .61734 L p .713 .752 1 r
F P
0 g
s
.38063 .56344 m .33761 .55836 L .31238 .49425 L p .759 .732 1 r
F P
0 g
s
.65701 .46569 m .61227 .52086 L .58991 .5547 L p .338 .172 .254 r
F P
0 g
s
.58991 .5547 m .63709 .51975 L .65701 .46569 L p .338 .172 .254 r
F P
0 g
s
.52164 .60914 m .47646 .63772 L .45079 .61363 L p .613 1 1 r
F P
0 g
s
.40624 .61734 m .45079 .61363 L .47646 .63772 L p .711 .933 1 r
F P
0 g
s
.31238 .49425 m .35536 .4954 L .38063 .56344 L p .759 .732 1 r
F P
0 g
s
.35536 .4954 m .31238 .49425 L .28738 .42415 L p .807 .738 .98 r
F P
0 g
s
.28738 .42415 m .33054 .41757 L .35536 .4954 L p .807 .738 .98 r
F P
0 g
s
.33054 .41757 m .28738 .42415 L .26252 .35099 L p .851 .755 .966 r
F P
0 g
s
.45079 .61363 m .49645 .60204 L .52164 .60914 L p .613 1 1 r
F P
0 g
s
.26252 .35099 m .30606 .33498 L .33054 .41757 L p .851 .755 .966 r
F P
0 g
s
.45079 .61363 m .40624 .61734 L .38063 .56344 L p .736 .793 1 r
F P
0 g
s
.38063 .56344 m .42501 .5652 L .45079 .61363 L p .736 .793 1 r
F P
0 g
s
.42501 .5652 m .38063 .56344 L .35536 .4954 L p .768 .736 1 r
F P
0 g
s
.37469 .4107 m .33054 .41757 L .30606 .33498 L p .844 .735 .948 r
F P
0 g
s
.30606 .33498 m .35035 .31869 L .37469 .4107 L p .844 .735 .948 r
F P
0 g
s
.39956 .4955 m .35536 .4954 L .33054 .41757 L p .806 .724 .966 r
F P
0 g
s
.33054 .41757 m .37469 .4107 L .39956 .4955 L p .806 .724 .966 r
F P
0 g
s
.35536 .4954 m .39956 .4955 L .42501 .5652 L p .768 .736 1 r
F P
0 g
s
.49645 .60204 m .45079 .61363 L .42501 .5652 L p .754 .857 1 r
F P
0 g
s
.41987 .40351 m .37469 .4107 L .35035 .31869 L p .837 .718 .932 r
F P
0 g
s
.35035 .31869 m .39541 .30212 L .41987 .40351 L p .837 .718 .932 r
F P
0 g
s
.42501 .5652 m .47068 .56314 L .49645 .60204 L p .754 .857 1 r
F P
0 g
s
.47068 .56314 m .42501 .5652 L .39956 .4955 L p .776 .746 1 r
F P
0 g
s
.54293 .58233 m .49645 .60204 L .47068 .56314 L p .746 .95 1 r
F P
0 g
s
.44499 .49433 m .39956 .4955 L .37469 .4107 L p .805 .715 .955 r
F P
0 g
s
.37469 .4107 m .41987 .40351 L .44499 .49433 L p .805 .715 .955 r
F P
0 g
s
.39956 .4955 m .44499 .49433 L .47068 .56314 L p .776 .746 1 r
F P
0 g
s
.4661 .39597 m .41987 .40351 L .39541 .30212 L p .832 .704 .919 r
F P
0 g
s
.39541 .30212 m .44125 .28526 L .4661 .39597 L p .832 .704 .919 r
F P
0 g
s
.58991 .5547 m .54293 .58233 L .51752 .55682 L p .595 .99 1 r
F P
0 g
s
.47068 .56314 m .51752 .55682 L .54293 .58233 L p .746 .95 1 r
F P
0 g
s
.77746 .3821 m .73103 .43209 L .71338 .48474 L p .504 .197 .404 r
F P
0 g
s
.71338 .48474 m .76357 .45511 L .77746 .3821 L p .504 .197 .404 r
F P
0 g
s
.63709 .51975 m .58991 .5547 L .56804 .5469 L p .029 .029 .029 r
F P
0 g
s
.57771 .54144 m .63709 .51975 L .56804 .5469 L .029 .029 .029 r
F
.56804 .5469 m .56538 .54595 L p .57771 .54144 L F P
0 g
s
.56538 .54595 m .6141 .53032 L .63709 .51975 L p .029 .029 .029 r
F P
0 g
s
.6842 .47846 m .63709 .51975 L .6141 .53032 L p .264 .138 .138 r
F P
0 g
s
.6141 .53032 m .6635 .50988 L .68116 .48308 L p .264 .138 .138 r
F P
0 g
s
.68116 .48308 m .6842 .47846 L p .66766 .4907 L .264 .138 .138 r
F P
0 g
s
.6141 .53032 m .68116 .48308 L .66766 .4907 L .264 .138 .138 r
F
.51752 .55682 m .56538 .54595 L .58991 .5547 L p .595 .99 1 r
F P
0 g
s
.49165 .49172 m .44499 .49433 L .41987 .40351 L p .804 .708 .947 r
F P
0 g
s
.41987 .40351 m .4661 .39597 L .49165 .49172 L p .804 .708 .947 r
F P
0 g
s
.51752 .55682 m .47068 .56314 L .44499 .49433 L p .785 .764 1 r
F P
0 g
s
.51342 .38803 m .4661 .39597 L .44125 .28526 L p .827 .691 .908 r
F P
0 g
s
.44125 .28526 m .48791 .26811 L .51342 .38803 L p .827 .691 .908 r
F P
0 g
s
.73103 .43209 m .6842 .47846 L .6635 .50988 L p .427 .18 .272 r
F P
0 g
s
.6635 .50988 m .71338 .48474 L .73103 .43209 L p .427 .18 .272 r
F P
0 g
s
.44499 .49433 m .49165 .49172 L .51752 .55682 L p .785 .764 1 r
F P
0 g
s
.56186 .37968 m .51342 .38803 L .48791 .26811 L p .822 .681 .898 r
F P
0 g
s
.48791 .26811 m .53539 .25065 L .56186 .37968 L p .822 .681 .898 r
F P
0 g
s
.53951 .48748 m .49165 .49172 L .4661 .39597 L p .804 .704 .943 r
F P
0 g
s
.4661 .39597 m .51342 .38803 L .53951 .48748 L p .804 .704 .943 r
F P
0 g
s
.56538 .54595 m .51752 .55682 L .49165 .49172 L p .793 .794 1 r
F P
0 g
s
.49165 .49172 m .53951 .48748 L .56538 .54595 L p .793 .794 1 r
F P
0 g
s
.61143 .37088 m .56186 .37968 L .53539 .25065 L p .818 .671 .889 r
F P
0 g
s
.53539 .25065 m .58372 .23287 L .61143 .37088 L p .818 .671 .889 r
F P
0 g
s
.6141 .53032 m .56538 .54595 L .53951 .48748 L p .801 .841 1 r
F P
0 g
s
.58855 .48143 m .53951 .48748 L .51342 .38803 L p .805 .703 .941 r
F P
0 g
s
.51342 .38803 m .56186 .37968 L .58855 .48143 L p .805 .703 .941 r
F P
0 g
s
.66217 .36158 m .61143 .37088 L .58372 .23287 L p .815 .663 .881 r
F P
0 g
s
.58372 .23287 m .63292 .21478 L .66217 .36158 L p .815 .663 .881 r
F P
0 g
s
.53951 .48748 m .58855 .48143 L .6141 .53032 L p .801 .841 1 r
F P
0 g
s
.6635 .50988 m .6141 .53032 L .58855 .48143 L p .798 .912 1 r
F P
0 g
s
.63874 .47342 m .58855 .48143 L .56186 .37968 L p .806 .704 .941 r
F P
0 g
s
.56186 .37968 m .61143 .37088 L .63874 .47342 L p .806 .704 .941 r
F P
0 g
s
.71411 .35176 m .66217 .36158 L .63292 .21478 L p .812 .656 .874 r
F P
0 g
s
.63292 .21478 m .68303 .19635 L .71411 .35176 L p .812 .656 .874 r
F P
0 g
s
.71338 .48474 m .6635 .50988 L .63874 .47342 L p .737 .995 1 r
F P
0 g
s
.58855 .48143 m .63874 .47342 L .6635 .50988 L p .798 .912 1 r
F P
0 g
s
.76357 .45511 m .71338 .48474 L .69002 .46331 L p .438 .909 .981 r
F P
0 g
s
.63874 .47342 m .69002 .46331 L .71338 .48474 L p .737 .995 1 r
F P
0 g
s
.69002 .46331 m .63874 .47342 L .61143 .37088 L p .808 .709 .945 r
F P
0 g
s
.61143 .37088 m .66217 .36158 L .69002 .46331 L p .808 .709 .945 r
F P
0 g
s
.69002 .46331 m .74233 .45098 L .76357 .45511 L p .438 .909 .981 r
F P
0 g
s
.74233 .45098 m .69002 .46331 L .66217 .36158 L p .811 .716 .951 r
F P
0 g
s
.66217 .36158 m .71411 .35176 L .74233 .45098 L p .811 .716 .951 r
F P
0 g
s
.25 Mabswid
.68211 0 m
.95235 .42431 L
s
.95235 .42431 m
1 .77293 L
s
1 .77293 m
.70428 .3819 L
s
.70428 .3819 m
.68211 0 L
s
.05779 .25176 m
0 .6164 L
s
0 .6164 m
.70428 .3819 L
s
.70428 .3819 m
.68211 0 L
s
.68211 0 m
.05779 .25176 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[23]:=",
  ImageSize->{288, 265.75},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P00012@P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/870P1L1`00`0L200031`0005/7002m1`<0
0`L205/7002k1`801PL00`07001J1`00^0L300P700@01`L0F@L00;H70P0;1`0400L705T7002c1`<0
3@L01@071`L005P7002a1`8040L00`071`021`0300L705D7002^1`<04`L01@071`L005L7002/1`80
5PL00`071`021`0300L705@7002Y1`<060L00`071`021`0300L705@7002W1`806`L00`071`031`03
00L705<7002T1`<07@L00`071`041`0300L70587002R1`8080L00`071`041`0300L70587002O1`<0
8PL00`071`051`0300L70547002M1`809@L00`071`061`0300L70507002J1`<09`L00`071`061`03
00L70507002H1`80:PL00`071`071`0300L704l7002F1`80;0L00`071`081`0300L704h7002C1`<0
;PL00`071`081`0300L704h7002A1`80<@L00`071`091`0300L704d7002>1`<0<`L00`071`0:1`03
00L704`7002<1`80=PL00`071`0:1`0300L704`700291`<0>0L00`071`0;1`0300L704/700271`80
?0L00`071`0:1`0300L704/700241`<0?PL00`071`0;1`0300L704X700221`80@@L00`071`0<1`03
00L704T7001o1`<0@`L00`071`0<1`0300L704T7001m1`80APL00`071`0=1`0300L704P7001j1`<0
B0L00`071`0>1`0300L704L7001h1`80B`L00`071`0>1`0300L704L7001e1`<0C@L00`071`0?1`03
00L704H7001c1`80D0L00`071`0@1`0300L704D7001`1`<0DPL00`071`0@1`0300L704D7001^1`80
E@L00`071`0A1`0300L704@7001[1`<0E`L00`071`0B1`0300L704<7001Y1`80FPL00`071`0B1`03
00L704<7001V1`<0G0L00`071`0C1`0300L70487001T1`80G`L00`071`0C1`0300L70487001Q1`<0
HPL00`071`0C1`0300L70447001O1`80I@L00`071`0D1`0300L70407001L1`<0I`L00`071`0D1`03
00L70407001J1`80JPL00`071`0E1`0300L703l7001G1`<0K0L00`071`0F1`0300L703h7001E1`80
K`L00`071`0F1`0300L703h7001B1`<0L@L00`071`0G1`0300L703d7001@1`80M0L00`071`0H1`03
00L703`7001=1`<0MPL00`071`0H1`0300L703`7001;1`80N@L00`071`0I1`0300L703/700181`<0
N`L00`071`0J1`0300L703X700161`80OPL00`071`0J1`0300L703X700131`<0P0L00`071`0K1`03
00L703T700111`80P`L00`071`0L1`0300L703P7000n1`<0Q@L00`071`0L1`0300L703P7000l1`80
R0L00`071`0M1`0300L703L7000j1`80R`L00`071`0L1`0300L703L7000g1`<0R@L2008700<01`L0
7@L00`071`0f1`00=@L208T70`001K001`L0020700<01`L0=@L003870`281`<000G7[lN_00021`03
00L701h700<01`L0=@L003070P281`<000^`Z[2Z/:Z`ZP07000Q1`0300L703@7000]1`<0Q`L3000>
ajo7[lN_ajo7[lL01`0R1`0300L703<7000[1`80Q`L3000A/:Z`Z[2Z/:Z`Z[2Z/:X01`008PL00`07
1`0c1`00:0L308L70P005<L0ajo7[lN_ajo7[lN_ajo700L08`L00`071`0b1`009PL208P70P005ZZ`
ZP2Z/:Z`Z[2Z/:Z`Z[2Z/:X01`0T1`0300L70347000S1`<0Q`L3000Hak37/<N`0:o7[lN_ajo7[lN_
ajo700L090L00`071`0a1`008@L208P70P006K2Z/:Z`Z[2Z/02`Z[2Z/:Z`Z[2Z/:Z`Z[000P0U1`03
00L70307000N1`<0R0L2000K/<N`ak37/<N`ak00[lN_ajo7[lN_ajo7[lN_00809PL00`071`0_1`00
70L208P70`007K2Z/:Z`Z[2Z/:Z`Z[00/:Z`Z[2Z/:Z`Z[2Z/:Z`00807@L00`071`061`0300L702l7
000I1`<0Q`L3000Pa`37/<N`ak37/<N`ak37/<L0ajo7[lN_ajo7[lN_ajl201`700<01`0020L00`07
1`0^1`005`L208L70`008k2Z/:X0Z[2Z/:Z`Z[2Z/:Z`Z[2Z0:Z`Z[2Z/:Z`Z[2Z/:Z`00806`L01@07
1`L000L700<01`L0;PL001@70`271`8002K7/<N`ak37/02`ak37/<N`ak37/<N`a`37[lN_ajo7[lN_
ajo7[`806PL00`1/00031`0300L700D700<01`L0;@L001870P271`<002VZ/:Z`Z[2Z/:Z`0;2Z/:Z`
Z[2Z/:Z`Z[2Z0:Z`Z[2Z/:Z`Z[2Z/:Z`ZP0201P700@060L010L00`071`051`0300L702`7000@1`80
QPL3000//<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`a`37[lN_ajo7[lN_ajo7[lL201L700D0K75/
00051`0300L700@700<01`L0;0L001070P241`8002nZ0:Z`Z[2Z/:Z`Z[2Z/:Z`0;2Z/:Z`Z[2Z/:Z`
Z[2Z/02`Z[2Z/:Z`Z[2Z/:Z`ZP0201H700H06747L@031`03/<L000L700<01`L0:`L0010700<01`00
P0L3000aak370<N`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak37/02_ajo7[lN_ajo7[lN_a`0201H7
00d0LFaaK007ak37/<L000L700<01`L0:PL0010700@01`L0O0L3000daj[7Z/NZa`2`Z[2Z/:Z`Z[2Z
/:Z`ZP2Z/:Z`Z[2Z/:Z`Z[2Z/02`Z[2Z/:Z`Z[2Z/:Z`ZP8050L2000=LAP7602`ak37/<N`00071`03
00L702X7000?1`0300L7008700<01`L0M`L2000iak37/<N`ak37/02`ak37/<N`ak37/<N`a`37/<N`
ak37/<N`ak37/02_ajo7[lN_ajo7[lN_ajl001870`003W5/LFaa0<N`ak37/<L020L00`071`0Y1`00
3`L00`071`031`0300L707<70`00>j[7Z/NZaj[7Z/NZa`2`Z[2Z/:Z`Z[2Z/:Z`ZP2Z/:Z`Z[2Z/:Z`
Z[2Z/02`Z[2Z/:Z`Z[2Z/:Z`Z[0001470`003g4HLK1a602`ak37/<N`00091`0300L702P7000?1`03
00L700@700<01`L0K`L3000n/02`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak370<N`ak37/<N`ak37
/<N`a`37[lN_ajo7[lN_ajo7[`0@1`0D0;@0LFaaK75/L@37/<N`ak37/0081`0300L702P7000?1`03
00L700D700<01`L0J`L30011Z/NZa`37Z/NZaj[7Z/NZaj[7ZP2Z/:Z`Z[2Z/:Z`Z[2Z0:Z`Z[2Z/:Z`
Z[2Z/:Z`ZP2Z/:Z`Z[2Z/:Z`Z[2Z/0003PL2000D/`1a674H01Qa0<N`ak37/<N`a`091`0300L702L7
000?1`0300L700H700<01`L0J0L20014/<N`ak37/<L0ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak37
/02`ak37/<N`ak37/<N`a`37[lN_ajo7[lN_ajo7[`0=1`0G0;>d075/LFaaK75/0;37/<N`ak37/000
2PL00`071`0V1`003`L00`071`071`0300L706@70`00A/NZaj[7Z/NZajX0Z/NZaj[7Z/NZaj[7ZP2Z
/:Z`Z[2Z/:Z`Z[2Z/02`Z[2Z/:Z`Z[2Z/:Z`ZP2Z/:Z`Z[2Z/:Z`Z[2Z/00<1`0309Fc008001@HLAQa
ag4HL@37/<N`ak37/<N`00T700<01`L09PL000h700<01`L02@L00`071`1P1`<004S7/<N`ak37/<N`
ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/<N`0:o7[lN_ajo7[lN_
ajl200T70P000k>d/`02000DLFaaK75/LF`0/<N`ak37/<N`a`0:1`0300L702D7000>1`0300L700X7
00<01`L0G0L3001;a`37Z/NZaj[7Z/NZaj[7ZP2Zaj[7Z/NZaj[7Z/NZ0:Z`Z[2Z/:Z`Z[2Z/:Z`0;2Z
/:Z`Z[2Z/:Z`Z[2Z/02`Z[2Z/:Z`Z[2Z/:Z`008020L01@2cUK>E008001Ea674H1aQa6740ak37/<N`
ak37/0002`L00`071`0T1`003PL00`071`0;1`0300L705L71000C/N`ak00/<N`ak37/<N`ak37/<L0
ak37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`ak00[lN_ajo7[lN_ajo7[`80
1`L0402d/kBc06`0K75/LFaaK742000;/<N`ak37/<N`a`002`L00`071`0T1`003PL00`071`0<1`03
00L705<70`00DZo7[lN_ajo7[`2Zaj[7Z/NZaj[7Z/NZa`37Z/NZaj[7Z/NZaj[70;2Z/:Z`Z[2Z/:Z`
Z[2Z0:Z`Z[2Z/:Z`Z[2Z/:Z`0;2Z/:Z`Z[2Z/:Z`Z[0200H70P007IFcU@1/L@1a6777LAQa08/0ak37
/<N`ak37/<L000/700<01`L08`L000h700<01`L03@L00`071`1?1`<005N`ak37/<N`ak37/<L0ak37
/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak37/02_ajo7
[lN_ajo7[lN_a`001@L2000N/kBc]01bK01/LFaaK75/08/0/<N`ak37/<N`ak002`L00`071`0S1`00
3PL00`071`0>1`0300L704`70P00FJl0[lN_ajo7[lN_ajo7[`2Zaj[7Z/NZaj[7Z/NZa`37Z/NZaj[7
Z/NZaj[70;2Z/:Z`Z[2Z/:Z`Z[2Z0:Z`Z[2Z/:Z`Z[2Z/:Z`ZP2Z/:Z`Z[2Z/:Z`Z[2Z008010L2000N
/iFc075/L@1a60LHLAP0QX/0ak37/<N`ak37/<L030L00`071`0R1`003PL00`071`0?1`80B@L3001K
ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`
a`37/<N`ak37/<N`ak37/<L0ajo7[lN_ajo7[lN_a`0200<70`003[>d079/LP1aK75/LF`00X/03P07
0<N`ak37/<N`ak003@L00`071`0Q1`003@L00`071`0B1`0300L704<70`00G/N_ajo7[lL0ajo7[lN_
ajo7[lN_a`37Z/NZaj[7Z/NZaj[70<NZaj[7Z/NZaj[7Z/NZ0:Z`Z[2Z/:Z`Z[2Z/:Z`0;2Z/:Z`Z[2Z
/:Z`Z[2Z0:Z`Z[2Z/:Z`Z[2Z/:X200870`004IFc075/LF`0676ZLAP0QX^600<000b`ak37/<N`ak37
/00<1`0300L70247000=1`0300L701<700<01`L0@0L2001Qak37/<N`ak37/02`ak37/<N`ak37/<N`
0;37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/<L0
ajo7[lN_ajo7[lN_a`02000B1`00L@2d/`1/LVab075/LF`018/00`370002000:ak37/<N`ak3700d7
00<01`L080L000d700<01`L050L00`071`0l1`<006>_ajo7[lN_ajo7[lL0ajo7[lN_ajo7[lN_a`37
Z/NZaj[7Z/NZaj[7ZP2Zaj[7Z/NZaj[7Z/NZ0:Z`Z[2Z/:Z`Z[2Z/:Z`0;2Z/:Z`Z[2Z/:Z`Z[2Z0:Z`
Z[2Z/:Z`Z[2Z/:X00P000`L00002000FU@1/LFaaK00H/1Qa08^6RhH0Z[2Z/08000R`ak37/<N`00h7
00<01`L07`L000d700<01`L05@L00`071`0i1`8006J`0;37/<N`ak37/<N`ak00/<N`ak37/<N`ak37
/02`ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak37
/02_ajo7[lN_ajo7[lL3000@L@00/`1/LVabK01/LFaa00F;00L0ak00Zk2[008000O7/<N`ak0000d7
00<01`L07`L000d700<01`L05PL00`071`0e1`<006[7/<N`0:o7[lN_ajo7[lN_ajl0[lN_ajo7[lN_
ajo70<NZaj[7Z/NZaj[7Z/NZ0:[7Z/NZaj[7Z/NZaj[70;2Z/:Z`Z[2Z/:Z`Z[00/:Z`Z[2Z/:Z`Z[2Z
/:Z`0;2Z/:Z`Z[2Z/:Z`ZP2Z1@008FaaK75/L@1a/740RhJ;QX^60:Z`0;2Z/:Z`0<N`ak37000>1`03
00L701h7000=1`0300L701L700<01`L0<@L3001_ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`
ak37/<L0ak37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`a`37/<N`ak37/<N`
ak37/02_ajo7[lN_ajo7[lL0a`1a00<000YbK79/LV`0K7401h/02P37/02[/:^`Zk020004/<N`00l7
00<01`L07@L000`700<01`L06@L00`071`0^1`800737/<N`ak37/<N`0:o7[lN_ajo7[lN_ajl0[lN_
ajo7[lN_ajo7[`2Zaj[7Z/NZaj[7Z/NZa`37Z/NZaj[7Z/NZaj[70;2Z/:Z`Z[2Z/:Z`Z[2Z0:Z`Z[2Z
/:Z`Z[2Z/:Z`0;2Z/:Z`Z[2Z/:Z`ZP2Z10007W5/LFaaK740a`2;QX^6RhJ;QP2Z/:X0Z[2Z/:Z`ZP80
00?700L03PL00`071`0M1`0030L00`071`0J1`0300L702X70`00OK37/<N`ak37/<N`a`37/<N`ak37
/<N`ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37
/<N`a`37/<N`ak37/<N`ak37/<L0ajo7[lN_ajo7[lL0a`1a079/LVabK79/00802H/04@37/<L0/:^`
Zk2[/:^`0<L000l700<01`L070L000`700<01`L06`L00`071`0V1`<007F`0;37/<N`ak37/<N`ak37
0<N_ajo7[lN_ajo7[lL0ajo7[lN_ajo7[lN_0:[7Z/NZaj[7Z/NZaj[70<NZaj[7Z/NZaj[7Z/L0/:Z`
Z[2Z/:Z`Z[2Z/:X0Z[2Z/:Z`Z[2Z/:Z`Z[2Z0:Z`Z[2Z/:Z`Z[2Z0:X00`0096aaK75/LF`0/026RhJ;
QX^6R`2`Z[2Z0:Z`Z[2Z/:Z`Z[2Z010700<01`L06`L000`700<01`L070L00`071`0R1`<007R`ak37
/02`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`0;37
/<N`ak37/<N`ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak370<N_ajo7[lN_ajo70<L30007
LVabK79/LP0200V;0180/<N`a`2`Zk2[/:^`Zk2[/00@1`0300L701/7000<1`0300L701d700<01`L0
7PL3002R/<N`ak37/<L0ak37/<N`ak37/<N`a`37[lN_ajo7[lN_ajo70<N_ajo7[lN_ajo7[lL0aj[7
Z/NZaj[7Z/NZajX0Z/NZaj[7Z/NZaj[7ZP2Z/:Z`Z[2Z/:Z`Z[2Z/02`Z[2Z/:Z`Z[2Z/:Z`ZP2Z/:Z`
Z[2Z/:Z`ZP2Z/01aK75/LFaaK026RhJ;QX^6RhJ;0;2Z/:Z`0;2Z/:Z`Z[2Z/:X04@L00`071`0J1`00
30L00`071`0N1`0300L701X70`00Qk37/<N`ak37/<N`0;37/<N`ak37/<N`ak00/<N`ak37/<N`ak37
/02`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak00
/<N`ak37/<N`ak37/<L0ajo7[lN_ajo7[lL0ajl0K79/LVabK00200Z;01<0/<N`ak00Zk2[/:^`Zk2[
/:/0010700<01`L06PL000`700<01`L07`L00`071`0G1`8007n`0;37/<N`ak37/<N`a`37/<N`ak37
/<N`ak37/02_ajo7[lN_ajo7[lN_0:o7[lN_ajo7[lN_a`37Z/NZaj[7Z/NZaj[7ZP2Zaj[7Z/NZaj[7
Z/NZ0:Z`Z[2Z/:Z`Z[2Z/:Z`0;2Z/:Z`Z[2Z/:Z`Z[2Z/02`Z[2Z/:Z`Z[2Z00@000MaK75/LFaa0080
01f6RhJ;QX^6RhJ;0;2Z/:Z`0;2Z/:Z`Z[2Z/:Z`000A1`0300L701T7000;1`0300L7024700<01`L0
4`L3001j1k07/02`ak37/<N`ak37/<L0ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`
ak37/<L0ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<N`0;37/<N`ak37/<N`
ak37/02_ajl7000=/02`06abK79/LP2;000:R`0C0;37/<N`0:^`Zk2[/:^`Zk2[000B1`0300L701P7
000;1`0300L7028700<01`L040L2001gak37/<N`a`37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02_
ajo7[lN_ajo7[lN_0:o7[lN_ajo7[lN_ajl0Z/NZaj[7Z/NZaj[7ZP2Zaj[7Z/NZaj[7Z/NZa`2`Z[2Z
/:Z`Z[2Z/:Z`0;2Z/:Z`Z[2Z/:Z`Z[001P08/00Y0;2`06aaK740R`2;QX^6RhJ;QX^6R`2`Z[2Z/:X0
Z[2Z/:Z`Z[2Z/0004PL00`071`0H1`002`L00`071`0S1`0300L700d70P00M;07/0N`1k07/02`ak37
/<N`ak37/<L0ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37
/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`a`37/<N`ak371@005/N`0;37/<N`ak37/<N`
0;37079/LP02R`0308^;00V;01@0/<N`ak370;2[/:^`Zk2[/:^`018700<01`L05`L000/700<01`L0
90L00`071`091`<00777/<N`ak37/<N`ak00/<N`ak37/<N`ak370<N`ak37/<N`ak37/<L0ajo7[lN_
ajo7[lN_0:o7[lN_ajo7[lN_ajl0Z/NZaj[7Z/NZaj[7Z/L0aj[7Z/NZaj[7Z/NZa`2`Z[2Z/:Z`Z[2Z
/:Z`ZP2Z/00500R`00<0/;002;00:P2`/01/LF`0K8/0RhJ;QX^6RhJ;QP2Z/:Z`Z[2Z0:Z`Z[2Z/:Z`
Z[2Z01<700<01`L05PL000/700<01`L09@L00`071`051`<006l700N`1k07/0N`1k07/0L0ak37/<N`
ak37/<N`0;37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<L0ak37/<N`ak37/<N`
ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<L01@007lN`ak37/<N`ak37/<L0ak37/<N`ak37/02`
ak00K0000h/00`2;R`08R`0E0<N`ak37/<N`0:^`Zk2[/:^`Zk0001<700<01`L05PL000/700<01`L0
9PL010071`L3001]ak37/<L0ak37/<N`ak37/<N`0;37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N_
ajo7[lN_ajo7[lL0ajo7[lN_ajo7[lN_a`37Z/NZaj[7Z/NZaj[70<NZaj[7Z/NZaj[7Z/NZ0:Z`Z[2Z
/00500:`00<0/;003K000`2`/008/0040;2`/08002A/Rfb;08^6RhJ;QX^6RhH0Z[2Z/:Z`Z[00/:Z`
Z[2Z/:Z`ZP0D1`0300L701D7000;1`0300L702L70`00K0N`1k07/0N`0;07/0N`1k07/0N`1k00/<N`
ak37/<N`ak370<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak00/<N`ak37/<N`
ak37/02`ak37/<N`ak37/<N`a`37/0@002K7/<N`ak37/02`ak37/<N`ak37/<N`a`37/<N`ak37/<N`
0;37/08018/00`2;R`08R`0F0<N`ak37/<N`0:^`Zk2[/:^`Zk2[01@700<01`L050L000X700<01`L0
9@L3001Yak37/<N`ak37/<N`0;37/<N`ak37/<N`a`37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02_
ajo7[lN_ajo7[lL0ajo7[lN_ajo7[lN_a`37Z/NZaj[7Z/NZaj[7ZP2Zaj[7Z/NZaj[700H03;000`2`
/00=/0030;2`00J`0P04/00@06b;K026RhJ;QX^6RhJ;QP8001B`Z[2Z/:Z`0;2Z/:Z`Z[2Z/:Z`01@7
00<01`L050L000X700<01`L08`L2001U1`07/0N`1k07/0N`1k0700N`1k07/0N`1k07/02`ak37/<N`
ak37/<L0ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`
a`37/<L01`00:<N`a`37/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`ak00/<N`ak030007/<L0ak37
0003R`0308^;00R;0P005K00ak37/<N`a`2`Zk2[/:^`Zk2[000E1`0300L701<7000:1`0300L70207
0`00H;07/0L0ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak00/<N`ak37/<N`ak00/<N`ak37/<N`ak37
/02_ajo7[lN_ajo7[lN_0:o7[lN_ajo7[lN_ajl0Z/NZaj[7Z/NZa`L000^`Zk2[/:^`Zk2[000>/003
0;2`00f`00<0/;000`06/00Y0;2`06b;K026RhJ;QX^6RhH0Q@2AZ`2`Z[2Z/:X0Z[2Z/:Z`Z[2Z/000
5PL00`071`0B1`002PL00`071`0N1`8005b`1k07/0N`0;07/0N`1k07/0N`1k00/0N`1k07/0N`1k07
0<N`ak37/<N`ak37/02`ak37/<N`ak37/<L0ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`a`L0
0337/<L0ak37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak03000>/<N`ak37
/<N`0;37/002R`0308^;00J;00L0RP2[/:^`0080012`ak370;2[/:^`Zk2[/:/05PL00`071`0B1`00
2PL00`071`0L1`8005P7/0N`1k07/0N`0;37/<N`ak37/<N`a`37/<N`ak37/<N`ak00/<N`ak37/<N`
ak370<N`ak37/<N`ak37/<L0ajo7[lN_ajo7[lN_0:o7[lN_ajo7[lN_1P0:/00@0:^`Zk2[/:^`Zk2[
/:^`00j`00<0/;002K030003/02`00Z`0140/;00Rf`0QX^6RhJ;QP25R`02000FTJ^AZi40Z[2Z0:Z`
Z[2Z/:Z`Z[2Z01H700<01`L04@L000X700<01`L06@L3001F/0N`1k07/0N`1k0700N`1k07/0N`1k07
/0L01k07/0N`1k07/0N`0;37/<N`ak37/<N`0;37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37
/<L4000g/<L0ak37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`a`37/<N`ak37
/003000D/<N`a`37/<N`ak37/<N`ak00/002R`0308^;00B;00@0RX^:0P005[2[/:^`Z`2`ak00Zk2[
/:^`Zk2[/00F1`0300L7014700091`0300L701L70`00EK0700N`1k07/0N`1k07/0L0ak37/<N`ak37
/<N`0;37/<N`ak37/<N`a`37/<N`ak37/<N`ak00/<N`ak37/<N`ak370<N_ajo7[lN_ajo7[lL0ajl0
1007/0030;2`00^`00l0/:^`Zk2[/:^`Zk2[/0003[000`2`/003/0<01k000`2`/00;/00?0;0008^6
RhJ;QP25RhF;008001JAZi6[TJ^A0:Z`0;2Z/:Z`Z[2Z/:X05`L00`071`0@1`002@L00`071`0D1`<0
1@L0C@2`1k07/0N`1k07/0N`0;07/0N`1k07/0N`1`07/0N`1k07/0N`1k00/<N`ak37/<N`ak370<N`
ak37/<N`ak37/02`ak37/<N`ak3700H003V`ak37/<N`ak37/02`ak37/<N`ak37/<N`0;37/<N`ak37
/<N`ak370<N`ak37/<N`ak37/<N`0;000`006[37/<N`ak37/<L0ak37/<N`ak37/<N`a`370P05R`06
08Z;RX^:0P005[2[/:^`Zk2[0;00Zk2[/:^`Zk2[/00H1`0300L700l700091`0300L701470`00Ck07
/0N`1k07/02`1k07/0N`1k07/0N`0;37/<N`ak37/<N`a`37/<N`ak37/<N`ak00/<N`ak37/<N`ak00
/<N`ak37/<N`ak37/02_ajl01P03/0030;2`00^`00<0/;002k00402`Zk2[/:^`Zk2[/:^`Z`0=/0<0
3[000`2`/00;/0<001J;QX^6R`2;QH^50:X0TJ^AZi6[TJ^A0P003;2Z/:Z`Z[2Z/:Z`01L700<01`L0
3`L000T700<01`L03PL300`703l01k07/0N`1k07/0N`1`07/0N`1k07/0N`1k00/0N`1k07/0N`1k07
0<N`ak37/<N`ak370<N`ak37/<N`ak001`00?k37/<N`ak37/<L0ak37/<N`ak37/<N`0;37/<N`ak37
/<N`ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/003000P/02`ak37/<N`ak37/<N`ak00/<N`ak37
/<N`ak37/<L200B;01<0RhZ;RX/0/:/0Zk2[/:^`Zk2[008000^`Zk2[/:^`Zk2[000H1`0300L700h7
00091`0300L700/70`00B;0700N`1k07/0N`1k07/0L01k07/0N`1k07/0N`0;37/<N`ak37/<N`a`37
/<N`ak37/<N`ak00/<N`ak37/<N`ak370<N`a`H00k000`2`/00;/0030;2`00^`00<0/;002k003`2[
/:^`Zk2[/:^`Zk2[0007/0<01;000`2`/00=/0030;2`00f`02D0RhJ;08^5RhF;Q@2Z/02[TJ^AZi6[
TJ^A0:Z`Z[2Z/:Z`Z[0001T700<01`L03@L000T700<01`L020L300H700<01`L02PL0;P2`1k07/0N`
1k07/0N`0;07/0N`1k07/0N`1`07/0N`1k07/0N`1k00/<N`ak0:0016ak37/<N`ak37/02`ak37/<N`
ak37/<L0ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<N`0;37/0<002R`ak37
/<N`a`37/<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak000X/05P2;RX^:RhZ;0;2[0:^`Zk2[/:^`
Zk02000:Zk2[/:^`Zk2[01T700<01`L03@L000T700<01`L01@L3000k/0N`1k07/0N`0;07/0N`1k07
/0N`1k00/0N`1k07/0N`1k070<N`ak37/<N`ak37/02`ak37/<N`ak002@06/0030;2`00^`00<0/;00
2[000`2`/00;/0030;2`00^`00l0Zk2[/:^`Zk2[/:^`Zk00100;/0030;2`00f`00<0/;003;03000S
RhF;QH^5RhD0Z[00Zi6[TJ^AZi6[T@2A0:Z`Z[2Z/:Z`ZP0060L200h700081`0300L700@70P0=1`03
00L700X701h01k07/0N`1k07/0N`1`07/0N`1k07/0N`1k00/0L6001C/<N`0;37/<N`ak37/<N`a`37
/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak00/<N`ak37
/<N`ak000`00;[37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak37/<L2000S
RX^:RhZ;RX/0/:^`0;2[/:^`Zk2[/:/0T@2[/:^`Zk2[/0005@L300031`0700d700081`0300L700D7
03<01k07/0N`1k07/0N`1k00/0N`1k07/0N`1k0700N`1k07/0N`1k07/02`ak37/<N`ak001P0:/003
0;2`00V`00<0/;002[000`2`/00;/0030;2`00^`00<0/;002k002@2`Zk2[/:^`Z`030003Zk0000n`
00<0/;003;000`2`/00=/00U08D0QH^5RhF;0;2Z/:X0TJ^AZi6[TJ^AZ`2[T@2Z/:Z`Z[2Z000B1`<0
10L00`071`0<1`0020L00`071`061`0300L700`700<01`L02@L04@2`1k07/0N`1k07/0N`0;0700H0
05V`1k00/<N`ak37/<N`ak370<N`ak37/<N`ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`0;37
/<N`ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/003001I/<N`ak370<N`ak37/<N`ak37/<N`0;37
/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`a`2;08^:RhZ;RP2[/:^`0;2[/:^`Zk2[/:/0TK00/:^`
Zk2[/0003PL400L700<01`L030L000P700<01`L01`L09@07/0N`1k07/0N`1k07/02`1k07/0N`1k07
/02`1k07/0N`1k001P0:/0030;2`00V`00<0/;002[000`2`/00:/0030;2`00Z`00<0/;002k000`2`
/00;/0030;2[00<000Z[/:^`Zk2[/:/03[000`2`/00=/0030;2`00f`02@0R`2;QH^5R`2`Z[2Z096[
TJ^AZi6[TJ/0Zi6[0;2Z/:Z`ZP0;1`<030L00`071`0;1`0020L00`071`081`8030L00`071`091`04
00N`1`H0062`1k0700N`1k07/0N`1k070<N`ak37/<N`ak37/02`ak37/<N`ak37/02`ak37/<N`ak37
/<L0ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<L2001P/<N`ak37/<N`ak37
0<N`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`ak00RX/0RhZ;RP2[/:^`
Z`2[/:^`Zk2[/:/0TK2A/02`Zk2[/:/01`L300l700<01`L02`L000P700<01`L02PL0502`1k07/0N`
1k07/0L01k07/0N`200:/0030;2`00Z`00<0/;002K000`2`/00:/0030;2`00Z`00<0/;002[000`2`
/00;/0030;2`00R`10003Z^`Zk2[/:^`Zk2[/:/03k000`2`/00</0030;2`00f`02D0RhF;08^5R`2`
Z[2Z/02[TJ^AZi6[TJ^A096[TJ/0/:Z`Z[0000@70`0B1`0300L700/700071`0300L700`700<01`L0
20L8001X/0N`1k0700N`1k07/0N`1k07/02`1k07/0N`1k07/02`ak37/<N`ak37/02`ak37/<N`ak37
/<L0ak37/<N`ak37/<N`0;37/<N`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<L3001Vak370<N`
ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<N`a`37/<N`ak37/<N`ak37/<L0
RhX0RX/0/02`Zk2[0:^`Zk2[/:^`Zk00/96`TK00/:^`Z`070`0E1`0300L700/700071`0300L700d7
00@0/0N`1P003kF`]@2e/;F`]K2e/;F`000</0030;2`00V`00<0/;002[000`2`/009/0030;2`00Z`
00<0/;002k000`2`/00;/0040;2`/0<01[00402[/:^`Zk2[/:^`Zk2[/00>/0030;2`00f`00<0/;00
3;06000NQ@2A0:Z`Z[00Zi6[TJ^AZi6[T@2AZi6[TJ/0/:Z`0P0H1`0300L700/700071`0300L700h7
0`00KkD7]@Ne1kD7]@L01k07/0N`1k07/02`1k07/0N`1k07/0L01k07/0N`1k07/0L0ak37/<N`ak37
/<L0ak37/<N`ak37/<N`0;37/<N`ak37/<N`a`37/<N`ak37/<N`ak00/<N`ak37/<N`ak37/0040012
ak37/<N`ak37/02`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<N`a`37/<N`
ak37/<N`1005W0<001j`TP2[/:/0Zk2[/:^`Zk2[/02`TK2A/96`0;2[/00H1`0300L700/700071`03
00L700l701T0/;F`]K2e/;F`]K00/;F`]K2e/;F`]K0000^`00<0/;002K000`2`/009/0030;2`00Z`
00<0/;002[000`2`/00:/0030;2`00V`0`0=/00?0;2[/:^`Zk2[/:^`Zk0000n`00<0/;003;000`2`
/004/0@000UnW7jLOYanW7h00`007Y6[T@2Z/:X0TJ^AZi6[TJ^AZ`2[TJ^AZi6[0;2Z01P700<01`L0
2`L000L700<01`L040L2001Y1kD7]@Ne1kD7]@2`1k07/0N`1k0700N`1k07/0N`1k07/02`1k07/0N`
1k07/02`ak37/<N`ak37/02`ak37/<N`ak37/02`ak37/<N`ak37/<L0ak37/<N`ak37/<N`0;37/<N`
ak3700<00437/02`ak37/<N`ak37/<N`0;37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<N`0;37/<N`
ak37/<N`ak37/02`100;W0<0022`0;2B/980Zk00/:^`Zk2[/:^`Z`2A/96`TK2A0:^`01P700<01`L0
2`L000L700<01`L04PL05`2e/;F`]K2e/;F`0;2e/;F`]K2e/;D000^`00<0/;002K000`2`/009/003
0;2`00Z`00<0/;002K000`2`/00;/0030;2`00:`0`06/0030;2`00Z`0100/:^`Zk2[/:^`Zk2[/:/0
3[000`2`/00;/0@0039nW7jLOYanW7jLOYanW02`T@2AZ`2[TJ^A0:X0TJ^AZi6[TJ^AZ`2[TJ^AZi6[
T@2Z01T700<01`L02PL000L700<01`L04`L0HP07]@Ne1kD7]@Ne0;07/0N`1k07/0L01k07/0N`1k07
/0L01k07/0N`1k07/02`ak37/<N`ak37/02`ak37/<N`ak37/<L0ak37/<N`ak37/<N`0;37/<N`ak37
/<N`a`370`00?LN`ak37/<N`a`37/<N`ak37/<N`ak370<N`ak37/<N`ak37/<L0ak37/<N`ak37/<N`
ak00/<N`ak37/<L01@03]@0309bL00ZL0P003;2[/02`T[00/9:`TP8001N`Zk2[/:^`Zk2[096`TK2A
/96`T@2[000H1`0300L700X700061`0300L701D70P005;2e/;F`]K2e/02`]K2e/;F`]K002k000`2`
/009/0030;2`00V`00<0/;002K000`2`/00:/0030;2`00V`0`0</0030;2`00^`00l0Zk2[/:^`Zk2[
/:^`Z`003k000`2`/00500Ve00X0W7jLOYanW7jL0P009I6`TK2A/02[TJ/0Zi6[0;00Zi6[TJ^AZi6[
0:^AZi6[TJ^AZi400P0H1`0300L700X700061`0300L701L705T0]@Ne1kD7]@L01k07/0N`1k07/02`
1k07/0N`1k07/02`1k07/0N`1k070<N`ak37/<N`ak370<N`ak37/<N`ak37/02`ak37/<N`ak37/02`
ak37/<N`a`03000ha`37/<N`ak37/<N`ak37/02`ak37/<N`ak37/<L0ak37/<N`ak37/<N`ak00/<N`
ak37/<N`ak0600ne00<0W9`01I`04P2[0;2[/:^`Z`2B/980T[2B/08001J[/:^`Zk2[/:^`0;2A/96`
TK2A/94060L00`071`0:1`001PL00`071`0H1`H000ne/;D0]K2e/;F`]K2e/0002[000`2`/008/003
0;2`00V`00<0/;002[000`2`/00:/0030;2`00:`0`05/0030;2`00^`00<0/;002[00402[/:^`Zk2[
/:^`Zk2[/003/0P000F`]K2e000A]@0509anW7h00`00:I6B0;2A/96`0:^AZi40TJ^AZ`2[TJ^AZi6[
TJ^A096[TJ^AZi6[TJ/001P700<01`L02PL000H700<01`L06@L01@2`]K2e00X004:`1k0700N`1k07
/0N`1k0700N`1k07/0N`1k00/<N`ak37/<N`ak00/<N`ak37/<N`ak370<N`ak37/<N`ak370<L3000_
ak37/<N`ak00/<N`ak37/<N`ak37/02`ak37/<N`ak37/<N`0;37/<N`ak37/<L01`003@Ne1kD7]@Ne
1kD7]@004[D06@2L002[/:/0TJ/0/:^`Z`2B/9:`TP2B/9800P005K2[/:^`Zk2[/02`TK2A/96`TK2A
000H1`0300L700X700061`0300L701X700h0Uk2G/9N`Uk00/9N`U``01;000`2`/008/0030;2`00V`
00<0/;002[000`2`/008/0<02k000`2`/00;/0030;2`00^`00@0/:^`1`005KF`]K00/;F`]K2e/;F`
]K2e/;F`000A]@8002n`TK2A/02BT@2A/96`0:^AZi6[T@2AZ`2B0:^AZi6[TJ^A096[TJ^AZi6[TJ^A
000G1`0300L700X700061`0300L701/701T0/;F`]K2e/;D0]K2e/;F`]K2e0;F`]K2e00/002D7/0N`
1`37/<N`ak37/<N`0;37/<N`ak37/<N`a`37/<N`ak3700<002C70<N`ak37/<N`ak37/02`ak37/<N`
ak37/<N`a`37/<N`ak0:000M1k07/0N`1k07/0L01kD7]@Ne1kD7]@Ne1kD7]@003[D3000I/:^`Zk2[
096[T@2[/:^`0;2B/9:`0;2B/002000EZk2[/:^`Zk00/96`TK2A/96`TK0001L700<01`L02PL000D7
00<01`L07@L04`2G/9N`Uk2G09N`Uk2G/9N`U`002[000`2`/003/0`01K000`2`/00:/0040;2`/0<0
1K000`2`/00:/0030;2`00B`3P07/00Q0;2e/;F`]K2e/;F`]K2e0;F`]K2e/;F`]K2e/;F`]K0000^e
0P00<k2B002`TK2A/02BTI:A096`T@2AZi6[TJ/0Zi40T@2AZi6[TJ^AZ`2[TJ^AZi6[TJ^A000H1`03
00L700T700051`0300L701h701h0/;F`]K2e/02`]K2e/;F`]K00/;F`]K2e/;F`]@09/0030;2`00B`
<P00=PN`1k07/0L01k07/0N`1k07/0N`1`07/0N`1k07/0N`1k07/02e1kD7]@Ne1kD7]@Ne1kD700Ve
0P00=J^`Zk00TP2`Zk2[/02[TJ^A0:^`0;2B/9:`T[00/02[T@2[/:^`Zk2[096`TK2A/96`TK0001P7
00<01`L02@L000D700<01`L07`L04P2G/9N`Uk2G09N`Uk2G/9N`00Z`00<0/;002;000`2`/008/003
0;2`00R`0`0;/0030;2`00Z`00<0/;003;000`2`/00;/00Q0;F`]K2e/;F`]K2e/;F`0;2e/;F`]K2e
/;F`]K2e/;D000Je0P007k2B/9:`TP2A096`TK2A096BTI80/940TJ^AZi6[TJ/00P005Y6BT@2AZi6[
TJ/0Zi6[TJ^AZi6[T@0H1`0300L700T700051`0300L7020701d0/;F`]K2e/02`]K2e/;F`]@2e/;F`
]K2e/;F`0009/0030;2`00V`00<0/;001;0300>`05L0ak37/<N`ak37/<L0ak37/<N`ak37/<N`0;07
/0N`1k07/0N`1k00/0N`1k07/0N`1k07/02`1k07/0N`1k07/0N`1`07]@Ne1kD7]@Ne1kD7]@Ne1`00
0kD4000hZk2[/:^`Z`2A0:^`Zk00Zi6[TJ/0/:/0T[2B/9:`TP2B096[0;2[/:^`Z`2A/96`TK2A/96`
T@0G1`0300L700T700051`0300L702470P004;2G/9N`U`2G/9N`Uk2G/009/0030;2`00R`00<0/;00
2;00102`/;0300J`00<0/;002K000`2`/00:/0030;2`00b`00<0/;002k008P2`]K2e/;F`]K2e/;F`
]@2e/;F`]K2e/;F`]K2e/;F`0;D200>`0240/9:`T[2B/02B0;2A/940TI:ATY6B0;00Zi6[TJ^AZi40
0P005Y:ATP2[TJ^AZ`2[TJ^AZi6[TJ^AZ`0G1`0300L700T700051`0300L702871@005k2e0;F`]K2e
/;F`0;2e/;F`]K2e/;D000V`00<0/;002K0300V`05H0ak37/<N`ak37/<L0ak37/<N`ak37/<N`a`07
/0N`1k07/0N`1k0700N`1k07/0N`1k07/0L01k07/0N`1k07/0N`1k0700Ne1kD7]@Ne1kD7]@Ne1`80
00F`ak37/002000g/:^`Zk2[096B0;2[/02[TJ^AZi40Z`2B/9:`T[2B/980TJ^AZ`2`Zk2[/02`TK2A
/96`TK2A000G1`0300L700T700051`0300L702<700@0NIMi10002K2G/9N`Uk2G0009/0030;2`00N`
00<0/;001[0400b`00<0/;002K000`2`/00:/0030;2`00^`00<0/;003;007@2e/;F`]K2e/;F`]K2e
/02`]K2e/;F`]K2e/;F`00<01[005`2B0;2B/9:`09:ATP2`TK00TY6BTI:A008000V[TJ^AZi6[TJ/0
0P005I6BTI80Zi6[T@2AZi6[TJ^AZi6[000G1`0300L700T700041`0300L702D70P04U`<001:`]K2e
/;F`0;2e/;F`]K2e/009/0030;2`00:`0`03/0030;2`00Z`0540/<N`ak37/<N`ak00/<N`ak37/<N`
ak37/02`1k07/0N`1k07/0N`0;07/0N`1k07/0N`1k00/0N`1k07/0N`1k07/0N`0;D7]@Ne1kD7]@L0
0P031`0N0;37/<N`a`2B0;2[/:^`09:A0:^`Z`2AZi6[TJ^A0P007k2B/9:`T[2B/02[TJ^A0:^`Zk00
/96`TK2A/96`T@004PL00`071`031`0300L700P700041`0300L702L700H0UgVGNIL40005Uk2G/000
2K000`2`/007/0030;0000801k000`2`/00:/0030;2`00R`00<0/;002k000`2`/00;/0030;2`00b`
01L0/;F`]K2e/;F`]K2e/;D0]K2e/;F`]@0300<70P000`L0/004/00G0940T[2B/980TI:A096`09:A
TY6BTI800P002I6[TJ^AZi6[T@02000ETY6BT@2AZi40TJ^AZi6[TJ^AZi40014700<01`L00`L00`07
1`081`0010L00`071`0X1`0309NG00>G00<0/9L00P003[F`]@2e/;F`]K2e/;D01k0300Z`00<0/;00
2[00B`37/<N`ak37/<N`a`37/<N`ak37/<N`ak00/0N`1k07/0N`1k07/0L01k07/0N`1k07/0N`1`07
/0N`1k07/0N`1k07/0L01kD7]@0200D700<01`L00PL01P2`ak3700:B03H0Zk2[/02BTI:A0:/0TJ^A
Zi6[TJ/0/9:`T[2B/9:`0:/0Zi6[T@2[/:/0TK2A/96`TK2A/00A1`0300L700<700<01`L020L000@7
00<01`L0:@L02P2GNIMi09:GTYL400V`00<0/;000[0300:`00<0/;002K000`2`/009/0030;2`00V`
00<0/;002k000`2`/00;/0030;2`00b`0180]K2e/;F`]K2e/;F`]K00/;D300@70P061`F`01L0TY40
T[2B096BTI80/940TI:ATY6BT@02000OZi6[TJ^AZi40TI80TY6BT@2AZ`2[TJ^AZi6[TJ^A000@1`80
1@L00`071`081`0010L00`071`0Z1`0309NG00:G00L0Uk2G/9N`00<000Re/;F`]K2e/0@01[000`2`
/008/0030;2`00Z`04D0/<N`ak37/<N`ak00/<N`ak37/<N`ak3700N`1k07/0N`1k07/0L01k07/0N`
1k07/0N`1k00/0N`1k07/0N`1k07/0L010041`802@L01LN`ak0000>B01<0Zk2[096BTI80/02[TJ^A
Zi6[008001nB/9:`T[2B/02[T@2AZi6[0;2[096`TK2A/96`TK0001070P051`0300L700P700041`03
00L702/70P002gVGN@2BUi:GTYNB00D00[0300V`00<0/;002;000`2`/00:/0030;2`00V`00<0/;00
2k000`2`/00;/0030;2`00b`00X0/;F`]K2e/;F`1`041`0300L700X71;005`2ATP2`T[00TY6BT@2A
096BTI:ATY6B008001jAZi6[TJ^AZ`2BT@2ATY6B0:/0Zi6[TJ^AZi6[T@0@1`801@L00`071`081`00
10L00`071`0/1`<00YL02`2G/9N`Uk00/9N`00D000>e0;001k000`2`/009/0030;2`00V`03h0/<N`
ak37/<N`ak00/<N`ak37/<N`ak37/02`1k07/0N`1k07/0N`0;07/0N`1k07/0N`1k0700N`1k07/0T0
0`L200h700C7/<L00i802@2`Z`2ATY6BT@02000KZi6[TJ^AZi40T@2B/9:`T[2B096[T@2AZi6[0080
00bA/96`TK2A/96`T@0?1`801@L00`071`081`000`L00`071`0^1`0<07T0UgT0TYNBUi:G0`001K2B
/9:`00<02;000`2`/008/0030;2`00Z`00<0/;002K000`2`/00:/0030;2`00^`00<0/;003K000`2e
000;00041`00010702h0/;00TI:ATP2`09:ATY6BTK00TY6BTI:ATP2B0:^AZi6[TJ/0TY6BT@2ATY6B
0P002i6[TJ^AZi6[TJ/000h700<0K@001@L00`071`081`000`L00`071`0`1`07N@000;2G/003000:
/02`Uk2G/9N`U`D01;000`2`/008/0030;2`00Z`03D0ak37/<N`ak37/<L0ak37/<N`ak37/<N`a`07
/0N`1k07/0N`1k0700N`1k07/0N`1k07/00A0003bPL0014700@0ak000i802@2[/02BTI:ATP02000K
TJ^AZi6[T@2ATP2`T[2B/980TJ^AZ`2[TJ^A008000^`TK2A/96`TK2A000>1`0308`000H700<01`L0
1`L000<700<01`L0<@L00gUF0003000BUi:GTP2B/9:`T[2B/02`Uk2G10000k00/009/0030;2`00Z`
00<0/;002;000`2`/00;/0030;2`00^`00<0/;001K050004W;FL]@/000@7bPL04`L0>@2`09:ATY40
TP2ATY6BTI80TY6BTI:ATY40Li80Zi6[TJ^A096BTI80TY6BT@2AZi6[TJ^AZi6[000>1`0306d000H7
00<01`L01`L000<700<01`L0<@L3000BN@00Uk2G/9L0Uk2G/9N`Uk001k0500N`00<0/;002[00:`2`
ak37/<N`ak37/02`ak37/<N`ak37/<N`0;07/0N`1k07/0N`1k00/0L010002ibeW;FL]IbeW;FL00D0
00P70<X7bPO:01@70P05TP8000JBTI:ATY42000KZi6[TJ^AZ`2BT@2B/9:`T[00Zi6[TJ/0Zi6[0080
00ZA/96`TK2A/9403PL00`2<00061`0300L700L700031`0300L702h70`021`03NEIi00<001VGTYL0
/9:`T[2B/980Uk2G/9N`Uk2G09N`00H00[000`2`/009/0030;2`00V`00<0/;002k000`2`/008/0H0
016eW;FL]IbeW;FL]IbeW;FL]@040006bP3:1lX05PL2000<TI:ATY40TI:ATY6B0P00:96BTI:ATY40
Li9c096[TJ^A096BTI:ATP2BT@1h096[TJ^AZi6[T@0<1`0400M]00H700<01`L01`L000<700<01`L0
;0L200H700D0EWT0U`02000<Uk00/9N`Uk2G/9L02;000`2`/006/0T01[007P37/<N`ak37/<N`0;37
/<N`ak37/<N`ak3700N`1`H01;D04P2eW;FL]IbeW;FL]IbeW;FL]@@000O:1`07bPL001L700<0TY80
0i820006TI:ATY6B0P006i6[TJ^AZ`2BTI:A09:`T[00Zi6[TJ^A096[T@02000:/96`TK2A/96`00`7
00@0LX`01PL00`071`071`000PL00`071`0Z1`<02@L00`1F0003N@8001nG0;2B/9:`T[00/9N`Uk2G
/9N`0;2G/9N`Uk2G/9L000F`3@05/0030;2`00B`2`0;]@0?0;FL]IbeW;FL]IbeW;FL00800/X02@3:
bP07bPO:000G1`0>096BTI:ATP2BTI:ATY42000WTY6BTI:ATP2BLi9c096[T@2ATY6BTI:A096B0980
Zi6[TJ^AZi4000`700@0LVd01PL00`071`071`000PL00`071`0W1`<03@L01`1F07VGNIL00P002IN`
Uk2G/9N`0008/0030;2`00R`00<0/;002K001`2e/;F`]K0030002KF`]K2e/;F`000>]@0<0;FL]Ibe
W;FL]Ibe0P000lX0bP0200071lX0bPO:000I1`0309:B00>B0P008i:ATY6B0980Zi6[TJ^A096BTI80
/9:`TP2AZi6[TJ^A094000:B00T0/96`TK2A/00030L00`1b00071`0300L700L700021`0300L702@7
0`0A1`03D5H000Ei01h0T[2B/9:`TP2G/9N`Uk2G/9L0Uk2G/9N`Uk2G/00;/0030;2`00V`00l0]K2e
/;F`]K2e/;F`]@003kD02P2eW;FL]IbeW;D20004b@3:00C:00D01lX7000I1`0609:ATY6B0P008I6B
TI:A07<0TI:ATY6B099cTW>B0:^AZ`2BTI:ATY6BT@02000;TWRB0:^AZi6[T@002`L01007LP071`03
00L700L700021`0300L702470`0E1`08M5H0NIMiUgT20007/9N`Uk2G0008/0030;2`00R`00<0/;00
2K00702`]K2e/;F`]K2e/02`]K2e/;F`]K2e/;F`]@0>]@0809beW;FL]I`300;:0P002LX0bPL01lX7
000K1`0309:B00>B0P008i6BTI:A07<0TJ^AZi40TI:ATY6B0;2B096[TJ^AZi6[0:/000>B00P0/96`
TK2A00X700@01g8020L00`071`061`000PL00`071`0N1`<06@L00e00N@04N@0N09L009:`T[00/9N`
Uk2G/9N`0;2G/9N`Uk2G/9L02k000`2`/009/00@0;2e/;F`]K2e/;F`]K2e00je00D0W;FL]@030008
b@3:0<X0bP02bP050<X7bP006`L01P2ATY6BT@80026BTI:ATP2BL`2ATY6BT@1cTW>BL`2AZ`2BTI:A
TY6BTI800P002gRBN02AZi6[TJ/000X700@0KG8020L00`071`061`000PL00`071`0K1`<07@L02W@0
UgVGNIL0Ui820004/9N`00R`00<0/;002;000`2`/009/00L0;F`]K2e/;F`]K2e0;F`]K2e/;F`]K2e
/;F`00ne00<0W0000`001<T0b/X2000:bP07bPL01lX701`700<0TY800Y809@2B09:ATY40Li80Zi6[
TJ/0TY6BTI:A09:`0:^AZi6[TJ^AZ`0019801`2A/96`T@002PL0102<LP081`0300L700H700021`03
00L701P70`0Q1`03D01i00Ai00@0Ui:G0P005i80Uk2G/9N`Uk2G09N`Uk2G/9N`Uk0000^`00<0/;00
2K00402e/;F`]K2e/;F`]K2e/00>]@H000>c0<T00P02bP080<[:00O:1`0M1`0Y09:ATY40L`2ATY6B
099cTP2BTI:A07>BLi9cTW<0T@2ATY6BTI:ATY400P002Y9hTWP0TJ^AZ`0:1`0406eb00P700<01`L0
1PL000031`0701L70`0U1`0?M000NIMi09:GTYNB09L000R`00<0/;002;000`2`/009/00L0;2e/;F`
]K2e/;F`0;2e/;F`]K2e/;F`]K2e00be1`001K?90<X000;:00T0bPO:0<X7bP007PL00`2BTP02TP0U
07<0TI:ATP2BL`2AZi6[09:ATY6BTI80/02[TJ^AZi6[TJ^A0005TP05096`T@002@L01@07S78000P7
00<01`L01PL000031`0701@70`0Y1`9@00/0NGUi09:GTYNBU`02000DUk2G/9N`Uk00/9N`Uk2G/9N`
U`0;/0030;2`00V`0100/;F`]K2e/;F`]K2e/;D02;D:0003b@3:00800/X0203:b/X01lX07`L07@2A
TY6B07<0TY6BT@1cTW<0TI:ATP2BLi9cTW>B008000ZATY6BTI:ATY6B0P002WRBN980Zi6[T@081`05
00M]LP0020L00`071`061`0000<700L04@L302d700edEP1iUgT0TYNBUi:G00@01K000`2`/008/003
0;2`00V`01d0]K2e/;F`]K2e/;F`0;2e/;F`]K2e/;F`]K2e0004]@`000G9/`3:0003bP080<X700O:
1`0P1`0M09:BTP1cTP2BTI80TW>BL`2AZi40TI:ATY6BTI800P00396[TJ^AZi6[T@2B00BB00D0/96`
00081`0500N<LP0020L00`071`061`0000<700L03PL303470U02000<N@2GTYNBUi80TYNB0`003iN`
U`2G/9N`Uk2G/9N`000:/0030;2`00Z`0100]K2e/;F`]K2e/;F`]K000[D>0003/`3900800/X02@3:
b/X0bPO:000P1`0509:ATP000W<05P2ATY40Li9cTP2BTI80TW>BLi9cTW<2000ETY6BTI:ATY6B099h
07RBN980Zi4000P700D0LVeb00091`0300L700D700000`L01`0;1`<0=@L04P1FN@2G09NBUi:GTP2G
/9N`U`@02[000`2`/009/00K0;2e/;F`]K2e/;F`]@2e/;F`]K2e/;F`]K2e010000>cb@000P03bP08
00O:0<X7bP0R1`0[09:BTP2BL`2ATY40Li9cTP2[T@2ATY6BTI:ATY40TJ^AZi6[TJ^AZ`2B0005TP03
0;0000L700H0S7:<LP091`0300L700D700000`L01`081`<0>@L0501@N@1i09NBUi:G09NBUi:GTYNB
10002K2G/9N`Uk2G000:/0030;2`00Z`00/0/;F`]K2e/;F`]@030003]Ibe00d000G900390003bP08
0<[:00O:1`0R1`05096BT@000W<05P2BTI80TW>BL`2ATY40Li9cTW>BLi82000BTI:ATY6BTI:A07RB
099hTWRB0P071`0606ebKG802@L00`071`051`0000<01`L01@L303d700<0EWT00P004Y:GTYNBU`2`
Uk2G/9N`Uk00/0@01;000`2`/009/00C0;F`]K2e/;F`]K2e/02`]K2e/004000:]IbeW;FL]IbeW0P0
00K9/`00bP02bP080<X7bP3:1`0T1`0X09:B099cTP2BT@1cTW>BL`2AZ`2BTI:ATY6BT@1d096[TJ^A
Zi6[00>B00<01i800Y800`2A00061`0600MbS7802@L00`071`051`0000<01`L00PL3044701X0D7U@
09:GTYNBU`2GTYNBUi:GTP2G/9N`U`@000>`0;002K000`2`/00:/0030;F`00<000fL]IbeW;FL]Ibe
W;FL00T000Fc003:0003bP080<[:0<X7bP0T1`0409:A00=c01H0TI80TW>BLi80TY40Li9cTW>BLi9c
0P004Y:ATY6BTI:A07RBN980TWRBN0801PL01P07LVeb00T700<01`L01@L0000300L00080A@L0101F
NEH2000GUi:GTP2G/9N`Uk2G/9L0Uk2G/9N`Uk002005/0070;2e/;F`]@08000?W;FL]IbeW;FL]Ibe
W;FL00T000>cb@000/X0303:b/X0bPL01lX702D702L0TP2BLi9c0940Li9cTW>B0:/0TY6BTI:ATY6B
07T0Zi6[TJ^AZ`0019800`07TP03TP0300L700@700@0S7:<0P091`0300L700D7000204T702@0D7T0
N@2GTYNB09:GTYNBUi:G0;2G/9N`Uk2G/9L0Uk2G/9L<000E]K2e/;F`]@2eW;FL]IbeW;FL]Ibe00T0
00>^0<T00P000lW:0002bP080<[:bP3:1`0V1`04096B00=c02/0TY40Li9cTW>B0980TW>BLi9cTW<0
LgT0TY6BTI:ATP2BN99h07RBN98000D700L0LVebKH/000T700<01`L01@L000031`0704T700D0EWT0
U`02000STYNB09N`Uk2G/9N`0;2G/9N`Uk2G/02e/;F`]K2e/;F`]@003KD0302eW;FL]IbeW;FL]@L0
00D70000b@0200;:00`0b/[:00O:0<X7bP0W1`0W0980Li9cTP2B099cTW>BL`2A096BTI:ATY6B07Ud
N@2[TJ^AZi4000FB00<0TY800Y800`071`021`0707:<LXb;00091`0300L700D700021`80B0L20003
D7T0009i0P00>IL0Ui:GTYNBUi80Uk2G/9N`Uk2G/02`Uk2G/9N`Uk2G0;2e/;F`]K2e/;F`]K2e0;FL
]IbeW;FL]@0:0008U000bP3:b@02bP080<[:bP07bP0X1`03098000Ac00X0T@1cTW>BLi9c0P007i9c
TW>BLi9cTP1i1gT0TY6BTI80TWRBN99h07RBN00010L0101bKG82R`0300L700P700<01`L010L000@7
00<01`L0APL2000XEP1iUgVG09L0/9N`Uk2G/9L0Uk2G/9N`Uk2G0;2e/;F`]K2e/;F`00je00L0]Ibe
W;FL00d000>c00000/X0303:b/X0bPO:0<X702T700NB07>BLi9c008000NBLi9cTW>B008001BATY6B
TI:ATY40M7Ud096[TJ^A00JB00<0TY800Y802`071`L01habZX/000X700<01`L010L000D700<01`L0
APL20003D01i00=i0P00<i:GTYNBUi:G0;2G/9N`Uk2G/9L0Uk2G/9N`Uk2G/9L0/;F`]K2e/;F`]K2e
/;D0]IbeW0070003O000008000R^002c0<W:00?:00H0b/X0bPL202T700<0T@0017<02P2B099cTW>B
Li82000OLi9cTW>BLi80NG=iL`2ATY6BT@1hTWRBN980TWRB00021`0506d7KG800X/00`071`081`03
00L700@700061`80A`L30006UgVGNIMi0P007k2G/9N`Uk00/9N`Uk2G/9N`0;F`]K2e/;F`]K2e/000
3[D00`2e00070003U@0000D000?90<X00/X02P3:bP07bPL01`0/1`07TP2BLi9cTP020007Li9cTW>B
L`02000DTY6BTI:ATY40M7UdNG@0TJ^AZ`07TP0409:B008700P0S7:<RjZ;00X700<01`L010L000P7
00<01`L0A0L0101@0006N@<002^BUi:GTP2G/9N`Uk2G/9N`0;2G/9N`Uk2G/9N`0;F`]K2e/;F`]K2e
/;F`00T000=lU@000`0029@0/lX0b/T00lX0103:b/X20003bP0702X70P05L`8002jBLi9cTW>BL`1c
TW>BLi9cTP1iLgUcNG<0TI:A07RBN99hTWRB099h00L0KG9]0h/00`071`081`0300L700@700091`03
00L704@700X0D000UgVGNIL00YL2000L/9N`U`2G/9N`Uk2G/9L0/;F`]K2e/;F`]K2e00Ze0`002Ybe
W;FL]I`0UI/40006[@39b/T00/X01`3:b/X01lX00P000`O:000/1`0@099cTW>BL`1cTW>BLi9cTP80
01>ATY6BTI:ATP1iM7UdN@2[TJ/000RB00d0TY801`2<LZZ;ZX/000X700<01`L010L000X70P151`03
050000Ii00@0Ui:G0P0089:G0;2G/9N`Uk2G/9L0Uk2G/9N`Uk2G/9N`0;F`]K2e100029beW;FL]Ibe
0P000gbE00060007b[<0bL[90002bP050<[:bP000lX00`071`0Z1`801G<2000/Li9cTW>BLi80TW>B
Li9cTW<0LgUcNG=i09:ATP2BN99hTWRBN01h00L0KG84R`0300L700P700<01`L010L000`700<01`L0
@`L02@1@07VGNIMi0005U`8001V`0;2G/9N`Uk2G/02e/;F`]K2e/;F`]K0000:e0`003kFL]IbeW;FL
]Ibe09^EV`070004b/W:00;:00L0b/[:0<X7008000?:1`00;PL0401cTW>BLi80TW>BLi9cTW<2000C
TY6BTI:ATP1iM7Ud1gAi0:^A0009TP0:098000MbZX^Z00/700<01`L010L000d700<01`L0@`L0101@
0004N@0709:GTYNBU`05000A/9N`Uk2G/02`Uk2G/9N`Uk001@004[FL]IbeW;FL]IbeW;D0UGbEO0L0
00NcbP3:bLX000;:00@0b/X01<X00`071`0/1`0307=c00Ac02P0Li9cTW>BLi80L`2BLi9cTW<0LgUc
NG=iLgT0TP2BN99hTWRBN99h0`000`N;R`02R`0300L700X700<01`L00`L000h70P141`8000Mi09Mi
UgT000NG00<0UkD01P001IL0/;F`00H01;D2000@]IbeW;FL]IbeW;D0ViFKU@L000C9b/T00/X03@3:
b/X01`0700O:1`00;`L01P2BLi9cTP8001YcTW>BLi9c051d09:ATY6BT@1dNGAiM7UdN@802Y830006
S8^ZRjX030L00`071`031`0040L00`071`121`8000=i07T00WT04@2GTYNBUi:G0;2G/9N`Uk2G00D0
01^L]IbeW;FL]IbeW;D0]IbeW;FL]Ibe09ElUG`01P002K@0b[?:0<[90003bP800/X0103:bP0`1`03
07=c00Ac02X0TW>BLi9cTW<0D01cTW>BLi80NG=iLgUcNG<0T@1hTWRBN99hTWRB0785R`0300L700X7
00<01`L00`L0014700<01`L0@@L0101d07T20003NIL000NG00`0]INeUkFG]INeU`0>]@070;FL]Ibe
W0020004ViFKU@H000K:0<[9bP03bP030<[:008000L7bP3:1lX0034700H0Li9cTW<2000JTW>BLi9c
TP1dD02ATY6BT@1dNGAiM7UdNG@200^B00L0ZX^ZRjX000`700<01`L00`L001870P121`0Z0500NEH0
N@2BUi:GTYNBU`2`Uk2G/9N`Uk2G09beW;FL]IbeW;FL]Ibe0P002YbeW01lUGbEO9D5000:];<0/lZc
0<W:00?:00D0b/[:0003bP0300L702l700<0Lg<00g<0:@1c099cTW>BL`1@Le00Li9cTP1iLgUcNG=i
LgUc07RBN99hTWRBN99h008018/00`071`0:1`0300L700<7000D1`0300L704070P03N@0307T000NG
00`0UkFG]INeUkFG]@0@]@0909`0UI^EViFK00<000Ve]02d0<W:b@000lX00`3:bP020008bPO:0<X7
bP0b1`0S099cTW<0D01cTW>BLi80M51d09:ATY6B07UdNGAiM7UdNG@00P09TP8000F;ZX^Z000<1`03
00L700<7000E1`0300L704070P000eIiEP02000SUi:GTYNBUi80Uk2G/9N`Uk2G/02eW;FL]IbeW;FL
]IbeW;D02P03]00=0;>d0<ZcbP3:bLX0bP0200?:00D0b/[:000c1`0307=c00=c01l0C01cTW>BLi80
Le1c099cTW<0LgUcNG=i1gUcN@2S008000YhTWRBN99h08/00h/00`071`0:1`0300L700<7000F1`03
00L704070P04N@801YL03@2eUkFG]INeUkFG]@003;D200Ne00d0]KBe]02dbP3:bLX000?:00d0bP07
00O:1`07bPL003@702H0Li9cTP1c099cTW>BL`1@M51d09:ATP1iM7UdNGAiM7Ud09fSW@801Y801P2;
08^Z00`700<01`L00`L001L70P111`0607UFNEIi10006I:GTYL0/9N`Uk2G/9N`Uk00]IbeW;FL]I`0
0`002IbeW;FL]Ibe0004]00<0;>d0<Zcb[<0bL[90P02bP050<[:bP000lX00`071`0b1`0307=c009c
02l0C7<0TW>BLi80Le1cD01cTW<0LgUcNG=iLgUcN@2SWJ>MX`2BN99hTP2SR`2;000=1`0300L70087
000I1`0300L703h700=@07T00gT0102GUiL200>G00`0]INeUkFG]INeU`05]@<02[D03`2e];Fd]@3:
]039b/W:0002bP8000/7bP3:1lX0bPO:000f1`0X099cTP1cD7<0TW>BL`1@M51dD02ATP1iM7UdNGAi
M7UdN@2SWJ>MX`800i801P2MX`2S00d700<01`L00PL001X700<01`L0?PL02d`0NEIiEP1iUgVG00@0
00jG/9N`Uk2G/9N`U`2L]@8000fL]IbeW;FL]IbeW;D000Fd00d0];<0/lZcbP3:bLX000?:00D0b/[:
0003bP0300L703@700<0Lg<00W<09@1cC01cTW>BL`1@Le1cD01cTP1iLgUcNG=iLgUc09fSWJ>MXid0
0P001980Xid200h700<01`L00PL001/70P0o1`<00gT00`2GU`04U`l03;D2000?]KBe];D0b[C:0<[9
b/T000;:0P003<X700O:1lX0bPO:03L702/0Li9c051cD01cTW>B07A@M51dD02A07AiM7UdNGAiM7T0
XifSWJ>MXifS00D000<700L030L00`071`021`007@L00`071`0m1`0J0500EWT0UgVGNIMiUgT0UkFG
]INeUkFG]IL2000;]IbeW;FL]IbeW0001[@02`2d/k@0b[?:/`3900<00lX01@3:b/X000C:00<01`L0
=@L0;@1cLg<0LdacC01cTW<0D7=@Le1c0980NG=iLgUcNG=iLgT0XifSWJ>MXifSW@03008700<01`L0
2PL00`071`021`007PL00`071`0l1`8000EF07Ui0008U`0>09beW;FL]IbeW;FL]@08]@0A0;Be];Fd
]K@0]<Zd0<W:bLX00P003lX700O:1`07bPL01lX7000i1`0D099c051cD7<0TW>B07A@M51dD7@2000F
M7UdNGAiM7Ud1g@0WJ>MXifSWJ>MX`800`L00`071`0:1`0300L70087000O1`80?@L094/0D01F07VG
NIMiUgVG0;FG]INeUkFG]INeUkD0]IbeW;FL00Nd00T0/kBc0;?:/lX00P02bP050<[:bP001<X01@3:
b/X003T701D0Lg=c04acC7<0TW>B07=@Le1cD7<00P0057=iLgUcNG=iLgT0XifSWJ>MXifS0P051`03
00L700T700<01`L00PL0024700<01`L0>`L01P1iEP1i00RG00h0]IbeW;FL]IbeW;FL]@800kD04@2d
]KBe];Fd0;C:]<X0b/W:0080017:1lX0bPO:0<X7bP3:1lX7000j1`0D07>B07=@Le00Li9c051dD7A@
M502000CNGAiM7UdNGAiM7T0XifSWJ>MX`0200L700<01`L02@L00`071`011`008PL00`071`0k1`03
07U@00<001YiUgVGNIMiU`2eUkFG]INeUkFG]INeU`2L00Nd00P0/kBc]03:/`<00lX00`3:bP02bP05
0<[:bP000lX00`071`0i1`0E07=c04acC7=<07>B07=@Le1cD7=@008001=iLgUcNG=iLgUc09fSWJ>M
Xid000T700<01`L020L00`071`011`008`L203/70P0017UFN@09U`0=0;FL]IbeW;FL]IbeW004000<
]KBe];Fd]@3:]<Zd0`021`0B00O:1`07bPL01lX7bP3:1lX0?0L05P2B07=@Le1cD01c051dD7A@M51d
0:<2000@M7UdNGAiM7T0XifSWJ>M00X700<01`L020L00`071`011`009@L00`071`0i1`05051iD7T0
0`003iMiUgVGN@2G]INeUkFG]@0400Be1002]0030;@000@000@7bPL00lX00`3:bP02bP050<[:bP00
1<X00`071`0j1`0H07=c07=<Ldac099c051cD7=@Le1c0:>R0P003GUcNG=iLgT0XifSW@0030L00`07
1`071`0300L70047000V1`0300L703T700L0EWUFN@2G00H000FG09beW00400be10000`Ne00041`0D
00O:1lX0bPO:1`07bPL01lX7bP0m1`0E07<0D7=@Le1c0980M51dD7A@M50000BS0`002GUdNG@0WJ>M
000>1`0300L700H700<01`L00@L002L700<01`L0>@L2000:NE1i09MnUgjGOPD03kD2000;1kD7]@3:
1lX7bP001<X01@3:b/X000C:00D0b/[:000n1`0:07<0LdacC7=<L`80011@Le1cD7=@Le00XZ>RXj:S
0P001W=i0:>M010700<01`L01@L00`071`011`00:0L203T700H0N@1iEP09U`030;Fe00fe00L0]@Ne
1kD000D701@0bPO:1`07bPO:0<X7bP3:1lX703l70P001e1cD7=@Le000P002WA@M51dD7A@M008X`80
00=i00004@L00`071`051`0300L70047000Z1`0300L703L70100EP1iD01nUgjGOYMnUgh03;D2000=
]@Ne1kD700O:1lX70004bP050<[:bP001<X00`3:bP02bP0300L703d700X0L`1<LdacC7=<0P0027=@
Le1cD7=@0P002J>RXj:SXZ>RX`0201<700<01`L010L00`071`011`00:`L00`071`0g1`0607T005H0
2IL200Ve00T01kD7]@Ne1`001@L05007bPO:0<X7bPL01lX7bP3:1lX0@@L20007Le1cD7=@L`02000:
D7A@M51dD7A@00VS00<01`L04`L00`071`041`400@L11`00;0L203P70140EU00D01nUgjGOYMnUgjG
0006]@8000h7]@Ne1kD700O:1lX700G:00<0b/X00/X01@3:b/X000C:00<01`L0?`L2000LC7=<Ldac
C7<0Le1cD7=@Le1c0:>RXj:SXZ>R01L700<01`L00`L100470@L002h700<01`L0=@L0101ANE4200^G
00<0]KD00[D02P2e1kD7]@Ne1`061`0E0<X7bPL01lX7bP3:1lX700O:1lX0048700T0Le1cD7=@Le00
0`002E1dD7A@M51d0006X`0300L701H700<01`L00`L100470@L002l700<01`L0=@L0101@EU02000=
UgjGOYMnUgjGOYL0]@02000@]@Ne1kD7]@L01lX7bPO:00G:00<0b/X00/X00`3:bP02bP030<[:00;:
00<01`L0@0L2000:LdacC7=<Ld`0XP8000acD7=@Le00XZ>RXj8201T700<01`L00PL100470@L00307
0P0f1`05055iDGT03@002PNe1kD7]@Ne1`061`0F0<X7bPO:1`07bPO:0<X7bP3:1lX704@700d0D7=@
Le1cD7<0^Z:j008000A@M51d0P03X`0308/001X700D01`L700021`00<PL00`071`0c1`0S05I@EU1F
09bGW9NLUibGW9NL00Ne1kD7]@L01lX7bPO:1`001LX00`3:bP02bP030<[:00;:00@0b/[:0P141`0?
07=<LdacC7=<L`2SXZ>R008000UcD7<0Xj:SR`0070L010071`021`00<`L00`071`0c1`0B07UANE5i
0;FL]IbeW;FL]Ibe0P001[D7]@Ne00H701<01lX7bPO:0<X7bPL01lX7bP3:008000<700L0A0L0401c
D7=@Le1cD02R^Z:jX[X20006D02SXh/07@L010071`021`00=0L203@70P007U1FD5H0W9NLUibGW9NL
UibG0;D7]@3:1lX7bPO:00G:00<0b/X00/X00`3:bP02bP<000?:1`00A`L04@1cC7=<Ldac0:>RXj:S
XZ>R00<000>;00L07PL00`0700021`00=PL00`071`0a1`0G05H007UAN@2eW;FL]IbeW;FL]I`0]@00
1`L0403:1lX7bPL01lX7bP3:1lX200051lX7bP00B0L2000@D7=@Le1c0;ZR^Z:jX[ZR^P808@L00`07
00011`00=`L00`071`0a1`03051F008000i@09NLUibGW9NLUibGW0<000L7bPO:1lX000K:00<0b/X0
0lX02P3:003:0<X7bP1;1`0B04acC7=<0::SXZ>RXj:SXZ<08PL20047000h1`80<PL0101FNEH30007
W;FL]IbeW00400031kD700@00PL03007bPO:1lX0bPO:1`<00PL01P3:1lX704`70100D7=@L`2jX[ZR
^Z:jX[X08`L20047000j1`0300L702l700H0EU1FD5H20003Ui`000<02kD500?:00<0b/X010002/X7
bPL01lX7bP1=1`0>04acC7<0Xj:SXZ>RX`0S1`<00@L003/700<01`L0;`L01`1iEWUFNEH00P003PNe
1kD7]@Ne1kD7]@Ne0P041`H00PL00`071`021`0600O:1lX0C`L0301@Le00X[ZR^Z:j020710041`00
?0L00`071`0_1`0805I@EU1FD00=]@0300L700H700<01`L00`L02`07bPO:0<X7bPL0050700X0C7=<
0::SXZ<07@L400P7000m1`80;`L05@1FNEIiEWUF0;D7]@Ne1kD7]@Ne00081`0300L700<700<01`L0
0PL01`3:1lX7bP00D@L20006L`2jX[X06PL400`7000o1`0300L702d70P001eI@EU1FD0002;D200P7
00<01`L010L02`3:1lX700O:1lX005@70P000j:S000G1`@040L0040700<01`L0;@L4000<NEIiEP2e
1kD7]@L02@L00`071`041`0300L7008700L01lX7bPL005D70`0D1`@050L004470P0^1`0305Ii00<0
00=FD0001;D00`071`071`0300L700@700`0bPO:1lX0bPO:1`1G1`0300L700l7100H1`00@`L00`07
1`0/1`0505InEWh010000kD000091`0300L700D700<01`L00PL01`3:1lX7bP00I@L401`700141`03
00L702`700L0EWUFNEIi00802`L00`071`041`0=00O:1lX700O:1lX7001Q1`@080L004D70P0/1`0:
07iFOUInEWiF00T700<01`L01@L00`071`031`060<X7bPL0GPL402@700171`0300L702T70P002GUF
NEIiEWUF00071`0300L700D700d0bPO:1lX0bPO:1lX005X7100X1`00B0L00`071`0X1`0<00L0OUIn
EWiFOUH01@L00`071`051`0300L700@700L01lX7bPL005H7100/1`00B@L202P700<01`L00`001gUF
NEIiEP000`L00`071`061`0=00O:1lX700O:1lX7001C1`@0<0L004/700<01`L09@L01P071`L0U`H0
00AF00L020L00`071`041`070<X7bPO:001?1`@0=0L004`700<01`L090L00`071`021`0307in00An
0`081`0>00O:1lX7bP3:1lX7bP1<1`@0>0L004d700<01`L08`L00`071`031`0909MnUgjGOYMn0080
1PL00`071`041`0700O:1lX700181`@0?0L004h70P0S1`0300L700@700<0OWh01gh200<700l0bPO:
1lX700O:1lX7bP00A0L40407001@1`0300L7020700<01`L01@L02`2GOYMnUgjGOYMn008000<700L0
1@L01`3:1lX7bP00@@L404@7001A1`0300L701l700<01`L01PL00`1nOP04OPL000g:1lX7bP3:1lX7
bPL003d710181`00DPL201l700<01`L01`L60007W;FL]IbeW00200<700P01lX7bPO:03T7101<1`00
E0L00`071`0L1`0300L700P700h0]IbeW;FL]IbeW;FL]@<000L7bPO:1lX003H7101@1`00E@L00`07
1`0K1`0300L700T700h0]IbeW;FL]IbeW;FL]@<000H7bPO:1`0b1`@0E0L005H70P0K1`0300L700X7
00/0]IbeW;FL]IbeW00200031kD700<000<700L0;PL405P7001H1`0300L701P700<01`L02PL02@2L
]IbeW;FL]@0200P70P0[1`@0G0L005T700<01`L05`L00`071`0;1`0609beW;FL0P002`Ne1kD7]@Ne
1kD002P7101P1`00FPL201L700<01`L030L00`2L]@0200X70P0U1`@0I0L005`700<01`L050L00`07
1`0=1`8000`7]@Ne1kD7]@Ne1`0S1`@0J0L005d700<01`L04`L00`071`0>1`0300L700P700<01`L0
7PL406`7001N1`804PL00`071`0@1`0:0;D7]@Ne1kD701d7101`1`00H0L00`071`0?1`0300L70147
00<01`L010L00`071`0H1`@0M0L0064700<01`L03PL00`071`0B1`040;D7]@805`L407P7001R1`03
00L700d700<01`L04`L00`07000E1`@0O0L006<70P0=1`0300L701@700<01`L040L40807001U1`03
00L700X700<01`L08`L408@7001V1`0300L700T700<01`L07`L408P7001W1`802@L00`071`0K1`@0
S0L006T700<01`L01PL00`071`0G1`@0T0L006X700<01`L01@L00`071`0C1`@0U0L006/70P051`03
00L700l7102H1`00K@L00`071`021`0300L700/7102L1`00KPL01@071`L000T7102P1`00K`L20003
1`0700@7102T1`00L@L200031`000080Z0L007870P2/1`00\
\>"],
  ImageRangeCache->{{{0, 287}, {264.75, 0}} -> {-0.000203924, 0.0142706, 
  0.00331714, 0.00331714}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Three dimensional line rendering", "Subsection",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  StyleBox["The OpenGL viewer ",
    FontFamily->"Times New Roman"],
  "can render lines as tubes. The rendering is controlled by the ",
  StyleBox["MVShow3D ",
    FontFamily->"Courier"],
  "options ",
  StyleBox["MVLineTubeSize",
    FontFamily->"Courier"],
  " and ",
  StyleBox["MVTubeSegments",
    FontFamily->"Courier"],
  ". The ",
  StyleBox["MVLineTubeSize ",
    FontFamily->"Courier"],
  StyleBox[
  "must be a floating point number and is given in the original dimensions of \
the graphics. Here is an example:",
    FontFamily->"Times New Roman"]
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(knot1 = 
      ParametricPlot3D[{\n\t\t\t
          \(-10\)\ Cos[t]\  - \ 2\ Cos[5\ t]\  + \ 15\ Sin[2\ t], \ \n\t\t\t
          \(-15\)\ Cos[2\ t]\  + \ 10\ Sin[t]\  - \ 2\ Sin[5\ t], \n\t\t\t
          10\ Cos[3\ t]\n\t\t}, \n\t\t{t, \ 0, \ 2\ Pi}]; \)\)], "Input",
  CellLabel->"In[138]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .84509 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-1.54261e-16 1.04136 -0.0512965 1.04136 [
[.05724 .23902 -18 -8.82381 ]
[.05724 .23902 0 .17619 ]
[.19683 .18415 -17.5176 -9 ]
[.19683 .18415 .48242 0 ]
[.3442 .12626 -5.55614 -9 ]
[.3442 .12626 .44386 0 ]
[.5 .06508 -10.5462 -9 ]
[.5 .06508 1.45381 0 ]
[.66499 .00034 -9.98009 -9 ]
[.66499 .00034 2.01991 0 ]
[.75899 .0665 0 -6.19153 ]
[.75899 .0665 18 2.80847 ]
[.83027 .1767 0 -6.08758 ]
[.83027 .1767 6 2.91242 ]
[.89475 .27639 0 -5.99567 ]
[.89475 .27639 12 3.00433 ]
[.95336 .36702 0 -5.91382 ]
[.95336 .36702 12 3.08618 ]
[.03084 .26796 -18 -2.74511 ]
[.03084 .26796 0 6.25489 ]
[.02098 .33245 -12 -2.79839 ]
[.02098 .33245 0 6.20161 ]
[.0107 .39963 -6 -2.85413 ]
[.0107 .39963 0 6.14587 ]
[-1e-05 .46969 -6 -2.91249 ]
[-1e-05 .46969 0 6.08751 ]
[-0.01119 .54281 -12 -2.97366 ]
[-0.01119 .54281 0 6.02634 ]
[ 0 0 0 0 ]
[ 1 .84509 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.04393 .25692 m
.70078 0 L
s
.06662 .24804 m
.07132 .25255 L
s
[(-20)] .05724 .23902 1 .96085 Mshowa
.20578 .19361 m
.21026 .19834 L
s
[(-10)] .19683 .18415 .9464 1 Mshowa
.35264 .13617 m
.35686 .14112 L
s
[(0)] .3442 .12626 .85205 1 Mshowa
.50786 .07546 m
.51179 .08064 L
s
[(10)] .5 .06508 .7577 1 Mshowa
.67218 .01119 m
.67578 .01661 L
s
[(20)] .66499 .00034 .66335 1 Mshowa
.125 Mabswid
.09387 .23738 m
.09666 .24011 L
s
.1214 .22661 m
.12417 .22937 L
s
.14923 .21573 m
.15197 .21851 L
s
.17735 .20473 m
.18007 .20754 L
s
.23452 .18237 m
.23717 .18523 L
s
.26357 .17101 m
.26619 .1739 L
s
.29293 .15952 m
.29553 .16244 L
s
.32262 .14791 m
.32519 .15085 L
s
.383 .1243 m
.3855 .1273 L
s
.41369 .11229 m
.41615 .11532 L
s
.44473 .10015 m
.44716 .10321 L
s
.47612 .08787 m
.47851 .09096 L
s
.53998 .0629 m
.5423 .06604 L
s
.57246 .05019 m
.57474 .05336 L
s
.60531 .03734 m
.60756 .04054 L
s
.63855 .02434 m
.64075 .02756 L
s
.25 Mabswid
.70078 0 m
.96253 .40426 L
s
.7468 .07108 m
.74071 .07337 L
s
[(-10)] .75899 .0665 -1 .37589 Mshowa
.81799 .18103 m
.81185 .18319 L
s
[(0)] .83027 .1767 -1 .3528 Mshowa
.88239 .2805 m
.87622 .28255 L
s
[(10)] .89475 .27639 -1 .33237 Mshowa
.94094 .37092 m
.93473 .37287 L
s
[(20)] .95336 .36702 -1 .31418 Mshowa
.125 Mabswid
.76164 .09399 m
.75797 .09535 L
s
.77616 .11642 m
.77249 .11777 L
s
.79039 .1384 m
.78672 .13973 L
s
.80433 .15993 m
.80065 .16124 L
s
.83138 .20171 m
.82769 .203 L
s
.84451 .22199 m
.84082 .22326 L
s
.85738 .24187 m
.85369 .24313 L
s
.87001 .26137 m
.86631 .26261 L
s
.89455 .29927 m
.89084 .30048 L
s
.90647 .31768 m
.90276 .31889 L
s
.91817 .33576 m
.91446 .33695 L
s
.92966 .3535 m
.92594 .35468 L
s
.73166 .04769 m
.72801 .04908 L
s
.71619 .02379 m
.71254 .0252 L
s
.95202 .38803 m
.94829 .38919 L
s
.25 Mabswid
.04393 .25692 m
0 .54614 L
s
.04297 .26323 m
.04903 .26087 L
s
[(-10)] .03084 .26796 1 -0.38998 Mshowa
.03315 .32784 m
.03924 .32554 L
s
[(-5)] .02098 .33245 1 -0.37814 Mshowa
.02293 .39516 m
.02904 .39292 L
s
[(0)] .0107 .39963 1 -0.36575 Mshowa
.01227 .46536 m
.01841 .46319 L
s
[(5)] -1e-05 .46969 1 -0.35278 Mshowa
.00114 .53863 m
.0073 .53654 L
s
[(10)] -0.01119 .54281 1 -0.33919 Mshowa
.125 Mabswid
.04104 .27595 m
.04468 .27454 L
s
.03909 .28876 m
.04273 .28736 L
s
.03713 .30168 m
.04077 .30029 L
s
.03515 .31471 m
.0388 .31332 L
s
.03114 .34108 m
.0348 .33971 L
s
.02911 .35443 m
.03277 .35307 L
s
.02707 .3679 m
.03073 .36654 L
s
.02501 .38147 m
.02867 .38012 L
s
.02083 .40896 m
.0245 .40763 L
s
.01872 .42288 m
.02239 .42156 L
s
.01659 .43692 m
.02026 .4356 L
s
.01444 .45108 m
.01812 .44977 L
s
.01008 .47976 m
.01377 .47847 L
s
.00787 .49429 m
.01156 .49301 L
s
.00565 .50894 m
.00934 .50767 L
s
.0034 .52372 m
.0071 .52246 L
s
.25 Mabswid
.04393 .25692 m
0 .54614 L
s
0 .54614 m
.37653 .84509 L
s
.37653 .84509 m
.38673 .58135 L
s
.38673 .58135 m
.04393 .25692 L
s
.70078 0 m
.96253 .40426 L
s
.96253 .40426 m
1 .683 L
s
1 .683 m
.71971 .30302 L
s
.71971 .30302 m
.70078 0 L
s
.04393 .25692 m
0 .54614 L
s
0 .54614 m
.71971 .30302 L
s
.71971 .30302 m
.70078 0 L
s
.70078 0 m
.04393 .25692 L
s
.38673 .58135 m
.96253 .40426 L
s
.96253 .40426 m
1 .683 L
s
1 .683 m
.37653 .84509 L
s
.37653 .84509 m
.38673 .58135 L
s
0 0 m
1 0 L
1 .84509 L
0 .84509 L
closepath
clip
newpath
.5 Mabswid
.69509 .49975 m
.68463 .53215 L
s
.70515 .46986 m
.69509 .49975 L
s
.68463 .53215 m
.67448 .56668 L
s
.71339 .44269 m
.70515 .46986 L
s
.67448 .56668 m
.66456 .60265 L
s
.71775 .4185 m
.71339 .44269 L
s
.66456 .60265 m
.65408 .63885 L
s
.65408 .63885 m
.64178 .67355 L
s
.71579 .39769 m
.71775 .4185 L
s
.64178 .67355 m
.62621 .70454 L
s
.62621 .70454 m
.60614 .72927 L
s
.70497 .38092 m
.71579 .39769 L
s
.60614 .72927 m
.58092 .74515 L
s
.2761 .39916 m
.23043 .37977 L
s
.32885 .42174 m
.2761 .39916 L
s
.38732 .44531 m
.32885 .42174 L
s
.23043 .37977 m
.19254 .36562 L
s
.58092 .74515 m
.5507 .74981 L
s
.44951 .46777 m
.38732 .44531 L
s
.19254 .36562 m
.16241 .35838 L
s
.5507 .74981 m
.51657 .7414 L
s
.16241 .35838 m
.13936 .35911 L
s
.68302 .36907 m
.70497 .38092 L
s
.51657 .7414 m
.48048 .71879 L
s
.51298 .48717 m
.44951 .46777 L
s
.48048 .71879 m
.44508 .68173 L
s
.13936 .35911 m
.12222 .36803 L
s
.44508 .68173 m
.41337 .63095 L
s
.57493 .50183 m
.51298 .48717 L
s
.6484 .36317 m
.68302 .36907 L
s
.12222 .36803 m
.10962 .38438 L
s
.41337 .63095 m
.38842 .56822 L
s
.10962 .38438 m
.10027 .40649 L
s
.63246 .51043 m
.57493 .50183 L
s
.38842 .56822 m
.37292 .49632 L
s
.60073 .36412 m
.6484 .36317 L
s
.37292 .49632 m
.36883 .41893 L
s
.10027 .40649 m
.0934 .43187 L
s
.68275 .51199 m
.63246 .51043 L
s
.36883 .41893 m
.37701 .34046 L
s
.0934 .43187 m
.08894 .45756 L
s
.54115 .37241 m
.60073 .36412 L
s
.37701 .34046 m
.39706 .26561 L
s
.08894 .45756 m
.08775 .48045 L
s
.72346 .50594 m
.68275 .51199 L
s
.39706 .26561 m
.42726 .19894 L
s
.08775 .48045 m
.09152 .49771 L
s
.47253 .38778 m
.54115 .37241 L
s
.42726 .19894 m
.4648 .14429 L
s
.4648 .14429 m
.50627 .10436 L
s
.09152 .49771 m
.10257 .50715 L
s
.75303 .49216 m
.72346 .50594 L
s
.39919 .409 m
.47253 .38778 L
s
.50627 .10436 m
.54817 .0804 L
s
.10257 .50715 m
.12338 .50751 L
s
.54817 .0804 m
.58756 .07225 L
s
.58756 .07225 m
.62239 .07853 L
s
.32632 .43381 m
.39919 .409 L
s
.12338 .50751 m
.156 .49869 L
s
.77099 .47095 m
.75303 .49216 L
s
.62239 .07853 m
.65176 .09705 L
s
.156 .49869 m
.20142 .48183 L
s
.25902 .45919 m
.32632 .43381 L
s
.65176 .09705 m
.67582 .12525 L
s
.20142 .48183 m
.25902 .45919 L
s
.67582 .12525 m
.69556 .16054 L
s
.77813 .44303 m
.77099 .47095 L
s
.69556 .16054 m
.71239 .20056 L
s
.71239 .20056 m
.72766 .24325 L
s
.77633 .40942 m
.77813 .44303 L
s
.72766 .24325 m
.74222 .28691 L
s
.76811 .37133 m
.77633 .40942 L
s
.74222 .28691 m
.75607 .33006 L
s
.75607 .33006 m
.76811 .37133 L
s
.25 Mabswid
.70078 0 m
.96253 .40426 L
s
.96253 .40426 m
1 .683 L
s
1 .683 m
.71971 .30302 L
s
.71971 .30302 m
.70078 0 L
s
.04393 .25692 m
0 .54614 L
s
0 .54614 m
.71971 .30302 L
s
.71971 .30302 m
.70078 0 L
s
.70078 0 m
.04393 .25692 L
s
.04393 .25692 m
.70078 0 L
s
.06662 .24804 m
.07132 .25255 L
s
[(-20)] .05724 .23902 1 .96085 Mshowa
.20578 .19361 m
.21026 .19834 L
s
[(-10)] .19683 .18415 .9464 1 Mshowa
.35264 .13617 m
.35686 .14112 L
s
[(0)] .3442 .12626 .85205 1 Mshowa
.50786 .07546 m
.51179 .08064 L
s
[(10)] .5 .06508 .7577 1 Mshowa
.67218 .01119 m
.67578 .01661 L
s
[(20)] .66499 .00034 .66335 1 Mshowa
.125 Mabswid
.09387 .23738 m
.09666 .24011 L
s
.1214 .22661 m
.12417 .22937 L
s
.14923 .21573 m
.15197 .21851 L
s
.17735 .20473 m
.18007 .20754 L
s
.23452 .18237 m
.23717 .18523 L
s
.26357 .17101 m
.26619 .1739 L
s
.29293 .15952 m
.29553 .16244 L
s
.32262 .14791 m
.32519 .15085 L
s
.383 .1243 m
.3855 .1273 L
s
.41369 .11229 m
.41615 .11532 L
s
.44473 .10015 m
.44716 .10321 L
s
.47612 .08787 m
.47851 .09096 L
s
.53998 .0629 m
.5423 .06604 L
s
.57246 .05019 m
.57474 .05336 L
s
.60531 .03734 m
.60756 .04054 L
s
.63855 .02434 m
.64075 .02756 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[138]:=",
  ImageSize->{288, 243.375},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000l`P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00;X710031`80G@L00;X700<01`L00`L010071`1L1`00^`L00`071`021`0400L705`7
002l1`0500L71`000PL00`071`1J1`00_@L010071`021`0300L705X7002j1`0400L7008700@01`L0
20L20587002k1`8010L200L70P000`L0001A1`00a`L300<70P1A1`00a@L200H700<01`00D0L00<87
0`000`L01`051`0400L704l7002o1`<02`L010071`1?1`00_0L300h700D01`L7001>1`00^PL400l7
00H01`L01`1=1`00^0L201<700<01`L00PL00`071`1;1`00]@L301H700<01`L00PL00`071`1:1`00
/`L201T700<01`L00PL00`071`1:1`00/0L300031`0701P700<01`L00`L00`071`191`00S@L400<7
0P0H1`807PL00`071`041`0300L704P7002?1`0500L71`000PL00`071`0B1`<080L00`071`041`80
B@L008l700D01`L700021`0300L700l7100R1`0300L700D700<01`L0A`L008l700D01`L700021`03
00L700d70P0V1`0300L700H700<01`L02PL400<70P0c1`00S`L01@071`L0008700<01`L02PL302P7
00<01`L01PL00`071`0<1`0500L71`000PL00`071`0`1`00S@L300<700@01`L02PL202/700<01`L0
1`L00`071`0;1`0500L71`000PL00`071`0`1`00S`L00`071`021`8020L300031`0702X700<01`L0
20L00`071`021`@010L01@071`L0008700<01`L0<0L009`70P0`1`0300L700P70P0;1`0500L71`00
0PL00`071`0`1`00V@L301071`0K1`0300L700L700<01`002@L300<700@01`L0<PL009L70`0=1`D0
1`L601D700<01`L02PL00`071`081`0300L700870P0c1`00U0L300<700<01`L01PL401870P0C1`03
00L700X700<01`L0@PL009470`0=1`8060L00`071`0@1`0300L700/700<01`L0@@L008l70P0>1`80
6`L2010700<01`L02`L00`071`111`00S0L3008700<01`L02@L201l70P0?1`0300L700/70P111`00
RPL200l70P0S1`0300L700`700<01`L02`L2044700271`<03`L202H700<01`L02`L00`071`0<1`03
00L703l700251`8000<700L03PL00`071`0V1`0300L700/700<01`L03@L00`071`0n1`00J@L201L7
0`0B1`0300L702P700<01`L02PL00`071`0>1`0300L703d7001X1`0400L701<70`0D1`0300L702X7
00<01`L02@L00`071`0>1`0300L703d7001X1`0400L70147100D1`0300L702`700<01`L020L00`07
1`0?1`0300L703`7001X1`0400L700h70`0G1`0300L702d700<01`L020L00`071`0?1`80?@L006P7
00@01`L030L201T700<01`L0;`L00`071`071`0300L7010700<01`L0>`L006P700@01`L02@L301X7
00<01`L0<@L00`071`061`0300L7014700<01`L0>PL006T70P081`8070L00`071`0c1`0300L700D7
00<01`L04PL00`071`0i1`00L0L301d700<01`L0=0L00`071`051`0300L7018700<01`L0>@L006h7
0P0O1`0300L703H700<01`L010L00`071`0C1`0300L703P7001[1`<000<700007@L00`071`0g1`03
00L700@700<01`L04PL00`07000i1`00J0L3028700<01`L0>@L00`071`031`0300L701@700<01`L0
=`L006H70P0T1`0300L703/700<01`L00PL00`071`0E1`0300L703H7001S1`<000<700L08`L00`07
1`0k1`0300L7008700<01`L05@L00`071`0f1`00H@L202P700<01`L0?@L00`071`021`0300L701D7
00<01`L0=@L005h70`0Y1`0300L703h700<01`L00PL00`071`0F1`0300L703@7001L1`8000<700L0
:@L00`071`0o1`0500L71`005`L203H7000j1`@00`L201H70`0]1`0300L7040700D01`L7000I1`03
00L700D70P0/1`00?0L01@071`L0008700<01`L040L302l700<01`L0@PL010071`0J1`0300L700<7
00@01`L0:`L003`700D01`L700021`0300L700h7100`1`0300L7048700@01`L06PL00`071`031`04
00L702/7000d1`@010L01@071`L0008700<01`L02`L303<700<01`L0A0L00`07000K1`0300L70087
00@01`L0:`L003`700D01`L700021`0300L700P70`0e1`0300L704D700<01`006`L200<700@01`L0
:`L003X70`031`0400L700P7100f1`0300L704D700<01`006PL00`0700031`0400L702/7000l1`03
00L700870P071`80>@L00`071`171`807@L010071`L202`700171`<0>PL00`071`181`807PL00`07
1`0^1`00A@L303`700<01`L0B@L00`071`0L1`0300L702h700121`<00`L00`071`0h1`0300L704X7
00<01`L07@L00`071`0]1`00@0L2044700<01`L0BPL00`071`0M1`80;PL003d710111`0300L704/7
0P0O1`0300L702`7000j1`<0A@L00`071`1;1`8080L00`071`0[1`00>0L204L700<01`L0C@L00`07
1`0N1`0300L702/7000e1`@0B0L00`071`1=1`8080L00`071`0Z1`004`L400<70P0G1`80B`L00`07
1`1>1`808@L00`071`0Y1`004`L00`071`031`0400L701<70`1=1`0300L704h70P0P1`80:`L001@7
00<01`L00PL010071`0A1`<0CPL00`071`1?1`0300L0024700<01`L0:0L000d710041`0500L71`00
0PL00`071`0<1`<0D@L00`071`1?1`0300L0028700<01`L09`L001H700@01`L00PL00`071`091`<0
E0L00`071`1?1`0300L0028700<01`L09`L001<700@01`L00PL010071`091`@0E0L00`071`1@1`04
00L7028700<01`L09PL001@70P041`801`L305P700<01`L0D0L010071`0R1`809`L002470P1J1`03
00L7054700@01`L08`L00`071`0U1`007PL405/700<01`L0D@L01@071`L002<700<01`L090L001`7
0P041`0300L705L700<01`L0DPL01@071`L002@700<01`L08`L001X70P1P1`0300L7058700D01`L7
000T1`0300L702<7000J1`<0GPL00`071`1C1`0300L7008700<01`L08@L202@7000J1`0300L005h7
00<01`L0D`L00`071`021`0300L7028700<01`L01@L400<70P0D1`0030L400<70P051`0400L705d7
00<01`L0D`L00`071`021`0300L702<700<01`L01PL01@071`L0008700<01`L04@L000h700D01`L7
00021`0500L71`0010L00`071`1I1`0300L705D700<01`L00PL00`071`0S1`0300L700D700D01`L7
00021`0300L70147000>1`0500L71`000PL010071`L300<700<01`L0F0L00`071`1E1`0300L70087
00<01`L08`L00`071`051`0500L71`000PL00`071`0A1`001PL400@700D01`L700021`0500L71`00
1PL00`071`1G1`0300L705D700<01`L00PL00`071`0T1`801@L01@071`L0008700<01`L04@L000h7
00D01`L700021`0500L71`001`L205L700<01`L0E@L00`071`031`0300L7028700<01`000`L300<7
00@01`L04`L000`70`031`0400L700<70P081`0300L705<700<01`L0EPL00`071`031`0300L702@7
00<01`L00`L00`071`021`8050L000h700<01`L00PL200@700<01`L020L00`071`1B1`0300L705H7
00<01`L00`L00`071`0U1`0300L701d7000I1`0300L700T700<01`L0D@L00`071`1F1`0300L700@7
00<01`L09@L00`071`0L1`0060L00`071`0;1`0300L704l700<01`L0E`L00`071`041`0300L702D7
0P0M1`0060L300`700<01`L0CPL00`071`1F1`801PL00`071`0U1`807@L001P700<01`L03@L00`07
1`1=1`0300L705<70`021`0300L700@700<01`L09PL00`071`0J1`0060L00`071`0>1`0300L704/7
00<01`L0D@L300D700<01`L010L00`071`0V1`0300L701X7000H1`8040L00`071`1:1`0300L704h7
0`091`0300L700<700<01`L09`L00`071`0I1`0060L00`071`0@1`0300L704T700<01`L0B`L300d7
00<01`L00`L00`071`0W1`0300L701P7000H1`0300L7014700<01`L0B0L00`071`181`<04@L00`07
1`021`0300L702H70P0J1`005`L00`07000C1`0300L704H700<01`L0APL301@700<01`L00PL00`07
1`0X1`0300L701L7000G1`805@L00`071`151`0300L704<70`0H1`0300L7008700<01`L09`L00`07
1`0G1`003`L300D700<01`L05@L00`071`141`0300L704070`0L1`0500L71`00:PL00`071`0F1`00
4PL00`071`021`0300L701H700<01`L0@PL00`071`0n1`<080L010071`0Z1`805`L0018700<01`L0
0PL201P70P121`0300L703/70`0S1`0500L71`00:PL00`071`0E1`002@L400870`051`0300L701T7
00<01`L0?`L00`071`0i1`809`L010071`0[1`0300L701@7000?1`0300L700D700<01`L06PL00`07
1`0n1`0300L703H70`0Z1`0300L002`700<01`L04`L000l700<01`L010L00`071`0L1`0300L703d7
00<01`L0<`L302h70P0/1`0300L701<7000?1`@00`L201h700<01`L0?0L00`071`0`1`<0<@L00`07
000[1`8050L001H700<01`L07PL603P700<01`L0;@L303D70P0]1`0300L700@710031`8010L001H7
00<01`L06PL600@7100c1`0300L702/70`0i1`0300L702/700<01`L010L00`071`031`0400L700<7
000F1`0300L701P70P061`0300L700D7100_1`0300L702P70`0<1aH06PL202d700<01`L010L00`07
1`021`0400L700<7000F1`8060L00`071`071`0300L700P70P0]1`0300L702D70`071`P05PL501H7
00<01`L0;0L00`071`041`0500L71`000PL00`071`011`005PL00`071`0F1`0300L700T700<01`L0
2@L302X700<01`L08PL300@71P0S1`8050L202`70P071`0400L7008700<01`L00@L001D700<01`L0
5PL00`071`0;1`0300L700/70P0X1`0300L701l70`021`D0:`L2018700<01`00;@L01@071`L00087
00@01`L00PL00`071`011`005@L201H700<01`L03@L00`071`0<1`<09@L00`071`0L1`<000<70000
0P0b1`804@L00`07000/1`0300L700870P041`8010L001D700<01`L050L00`071`0?1`0300L700h7
0P0S1`0300L701T71`0h1`0300L700h700<01`00;@L00`071`0=1`005@L00`071`0D1`0300L70107
00<01`L03`L3020700<01`L05PL503d700<01`L03PL010071`0]1`0300L700`7000E1`0300L701<7
00<01`L04PL00`071`0A1`<07@L00`071`0C1`@0@PL00`071`0=1`0500L71`00:`L200h7000E1`80
50L00`071`0C1`8050L201/700<01`L040L304H700<01`L03PL01@071`L002`700<01`L02`L001D7
00<01`L04`L00`071`0E1`0300L701<70P0I1`0300L700d70`1:1`0300L700d700D01`L7000]1`03
00L700X7000=1`801PL00`071`0B1`0300L701L700<01`L050L201H700<01`L02PL404d700<01`L0
3@L00`071`021`0300L702X700<01`L02PL000`700@01`L010L00`071`0C1`0300L701P700<01`L0
5@L201@700<01`L01`L4050700<01`L03PL00`071`021`0300L702T70P0;1`0030L010071`041`80
50L00`071`0I1`0300L701H70P0B1`0300L700@7101C1`0300L700h700<01`L00PL00`071`0V1`<0
00<700L02PL000`700@01`L010L00`071`0B1`0300L701/700<01`L05`L2010700@01`L7101G1`03
00L700d700<01`L00`L00`071`0Q1`@010L00`071`091`0030L010071`041`0300L7018700<01`L0
70L00`071`0H1`803PL505X700<01`L03@L00`071`041`0300L701d70`081`0300L700T7000<1`04
00L700@70P0C1`0300L701d700<01`L06@L200X7101M1`0300L700d700<01`L01@L00`071`0I1`<0
30L00`071`081`003@L200D700<01`L04PL00`071`0N1`0300L701X70`051`<000<700L0G@L00`07
1`0=1`0300L700D700<01`L05PL300l700<01`L020L001@700<01`L04@L00`071`0P1`0300L701`7
1P031`0300L705`700<01`L03@L00`071`061`0300L70147100B1`0300L700P7000C1`0300L70187
00<01`L08@L00`071`0J1`H010L00`071`1L1`0300L700d700<01`L01`L00`071`0=1`<05PL00`07
1`081`004`L201<700<01`L08PL00`071`0F1`@01@L300031`0705`700<01`L03PL00`071`081`03
00L700T70`0I1`0300L700P7000C1`0300L7018700<01`L08`L00`071`0C1`<02`L205d700<01`L0
3PL00`071`081`0300L700H70`0L1`0300L700P7000C1`0300L7018700<01`L090L00`071`0?1`<0
3`L405X700<01`L03PL00`071`091`0400L71`@07`L00`071`081`004`L00`071`0A1`0300L702H7
0P0<1`<04PL010071`L305L700<01`L03PL00`071`0:1`<090L00`071`071`004`L00`071`0A1`03
00L702P700<01`L01PL301D700<01`L010L305<700<01`L03`L00`071`071`<000<700L090L00`07
1`071`004PL201<700<01`L0:@L00`071`021`@05`L00`071`071`<0D0L00`071`0>1`0300L700D7
0`041`0300L702<700<01`L01`L0018700<01`L04PL00`071`0Z1`D06PL00`071`0:1`80CPL00`07
1`0>1`0400L71`@020L00`071`0R1`0300L700L7000B1`0300L7018700<01`L0:0L501d700<01`L0
2`L304X700<01`L03`L400d700<01`L08@L00`071`071`004PL3018700<01`L09@L5008700<01`L0
6`L00`071`0>1`<0A`L00`071`0=1`<04@L00`071`0P1`0300L700L7000B1`0300L7018700<01`L0
8PL500H700<01`L06PL00`071`0A1`<0@`L00`071`0;1`<000<700L04@L00`071`0P1`0300L700L7
000:1`<01@L00`071`0A1`0300L702071P091`0300L701T700<01`L050L403l700<01`L01`L400@7
00<01`L04@L00`071`0P1`0300L700H7000=1`0300L7008700<01`L04@L00`071`0M1`H03@L00`07
1`0H1`0300L701P70`0l1`0300L700@70`071`0300L701<700<01`L07`L00`071`061`003@L01007
1`L201<700<01`L06PL300031`00014700<01`L05`L00`071`0K1`<0>@L010071`L300T700<01`L0
5@L00`071`0N1`0300L700H7000:1`<010L00`071`0B1`0300L701L70`0010L0000D1`0300L701H7
00<01`L07PL403@700<01`000P0<1`0300L701D700<01`L07PL00`071`061`002PL00`071`041`03
00L7018700<01`L050L300041`0001P700<01`L05@L00`071`0R1`@0;PL400h700<01`L05`L00`07
1`0M1`0300L700H7000:1`0300L700@700<01`L04PL00`071`0A1`<000@7000070L00`071`0D1`03
00L702H7100W1`<00PL00`071`0<1`0300L701T700<01`L070L00`071`061`002PL400<70P0D1`03
00L700d70`0010L0000P1`0300L701<700<01`L0:PL402070`051`0300L700X70P0L1`0300L701`7
00<01`L01PL0014700<01`L04`L00`071`0:1`<000@7000090L00`071`0B1`0300L702h7100I1`<0
1`L00`071`081`<07`L00`071`0L1`0300L700D7000A1`0300L701@700<01`L01PL300041`0002P7
0P0B1`0300L703871@0@1`@02PL00`071`061`808`L00`071`0K1`0300L700D7000A1`0300L701D7
00<01`L00`L200041`0002d700<01`L03`L00`071`0g1`D020L300h700<01`L00PL402H700<01`L0
6PL00`071`051`0040L201P7200a1`0300L700l700<01`L0>`L801071P0Z1`0300L701X700<01`L0
1@L0010700<01`L05PL303P700<01`L03PL00`071`0m1`<000<700004@0`1`0300L701T700<01`L0
1@L0010700<01`L04`L303`700<01`L03@L00`071`0i1`@05PL00`071`0_1`0300L701P700<01`L0
1@L0010700<01`L040L3040700<01`L030L00`071`0f1`<06@L00`071`0a1`0300L701L700<01`L0
1@L000l70P0?1`<0A0L00`071`0;1`0300L703<70`0L1`0300L7034700<01`L060L00`071`041`00
40L00`071`0:1`<0B0L00`071`0;1`0300L702l70`0O1`0300L7038700<01`L05`L00`071`041`00
40L00`071`071`<0C0L00`071`0:1`0300L702/7100Q1`0300L703@700<01`L05PL00`071`041`00
3`L00`071`051`<0D0L00`071`091`0300L702P70`0U1`0300L703D700<01`L05@L00`071`041`00
3`L300870`1D1`0300L700P700<01`L09@L302P700<01`L0=@L00`071`0E1`0300L700@700000`L0
000200<70P051`0300L00080F0L00`071`081`0300L702470`0Z1`0300L703L700<01`L050L00`07
1`041`000`L01@071`L0008700<01`L00PL205`700<01`L01`L00`071`0M1`@0;@L00`071`0h1`03
00L701<700<01`L010L000<700D01`L700021`0300L700<700<01`L0F`L00`071`061`0300L701X7
0`0a1`0300L703T700<01`L04`L00`071`031`000`L01@071`L0008700<01`L010L205`700<01`L0
1@L00`071`0G1`<0<`L00`071`0j1`0300L701<700<01`L00`L000<700D01`L700021`0300L700H7
00<01`L0FPL200D700<01`L050L303H700<01`L0>`L00`071`0B1`0300L700<7000010L000031`04
00L700T700<01`L0F`L00`071`031`0300L700l7100i1`0300L703`700<01`L04@L00`071`031`00
0`L00`071`021`802`L00`071`1K1`0300L7008700<01`L030L303`700<01`L0?PL00`071`0@1`03
00L700<7000F1`80G0L01@071`L000/70`0o1`0300L703h700<01`L040L00`071`031`0060L00`07
1`1J1`0400L700P70`121`0300L703l700<01`L040L00`071`021`006@L00`071`1J1`0400L700<7
10141`0300L7044700<01`L03`L00`071`021`006PL00`071`1J1`0300L000<0B0L00`071`121`03
00L700h700<01`L00PL001/70P1K1`<0BPL00`071`121`0300L700h700<01`L00PL001d700<01`L0
F0L00`07001:1`0300L704<700<01`L03@L00`071`021`007PL00`071`1G1`0300L004T700<01`L0
A@L00`071`0<1`0300L70087000O1`0300L705H700@01`L0B0L00`071`151`0300L700`700<01`L0
0PL002070P1F1`0400L704P700<01`L0APL00`071`0<1`0300L70047000R1`0300L705<700D01`L7
00161`0300L704P700<01`L02`L00`071`011`008`L00`071`1B1`0500L71`00APL00`071`191`03
00L700X700<01`L00@L002@700<01`L0D@L00`071`021`0300L704<700<01`L0B@L00`071`0:1`03
00L70047000U1`80D@L00`071`021`0300L7048700<01`L0B`L00`071`091`0300L70047000W1`03
00L704h700<01`L00PL00`071`121`0300L704`700<01`L020L00`071`011`00:0L00`071`1=1`03
00L700<700<01`L0@@L00`071`1=1`0300L700L700<01`L00@L002T700<01`L0C0L00`071`031`03
00L7044700<01`L0C@L00`071`081`400@L11`00:PL00`071`1;1`0300L700@700<01`L0?`L00`07
1`1?1`0300L700L70@011`47000[1`80BPL00`071`051`0300L703l700<01`L0D0L00`071`061`40
0@L11`00;@L00`071`171`0300L700H700<01`L0?PL00`071`1A1`0300L700D70@011`47000^1`03
00L704H700<01`L01PL00`071`0m1`0300L7058700<01`L01@L100470@L002l700<01`L0A@L00`07
1`071`0300L703`700<01`L0D`L00`071`041`400@L11`00<0L204D700<01`L020L00`071`0k1`03
00L705@700<01`L00`L100470@L0038700<01`L0@PL00`071`081`0300L703X700<01`L0EPL00`07
1`031`400@L003<700<01`L0@@L00`071`091`0300L703T700<01`L0EPL00`071`031`400@L003@7
00<01`L0@0L00`071`091`0300L703T700<01`L0E`L00`071`021`400@L003D70P101`0300L700X7
00<01`L0=`L00`071`1I1`0500L71`000@L003L700<01`L0?@L00`071`0;1`0300L703H700<01`L0
FPL010071`011`00>0L00`071`0l1`0300L700/700<01`L0=PL00`071`1J1`0400L70047000i1`03
00L703/700<01`L030L00`071`0d1`0300L705`700<01`000@L003X70P0k1`0300L700d700<01`L0
<`L00`071`1M1`400@L10000?0L00`071`0h1`0300L700d700<01`L0<`L00`071`1N1`80000m1`03
00L703L700<01`L03PL00`071`0a1`0300L705l70P0003h700<01`L0=PL00`071`0?1`0300L70307
00<01`L0G`L20000?`L203H700<01`L040L00`071`0^1`0300L705`710021`00@@L00`071`0c1`03
00L7014700<01`L0;@L00`071`1H1`@01PL0048700<01`L0<PL00`071`0B1`0300L702/700<01`L0
E@L400X700131`0300L7034700<01`L04PL00`071`0[1`0300L70547100>1`00A0L2034700<01`L0
4`L00`071`0Y1`0300L704h7100B1`00APL00`071`0^1`0300L701@700<01`L09`L00`071`1;1`@0
5PL004L700<01`L0;@L00`071`0E1`0300L702D700<01`L0B@L301X700181`0300L702`700<01`L0
5PL00`071`0S1`0300L704H7100M1`00B@L202/700<01`L060L00`071`0R1`0300L70487100Q1`00
B`L00`071`0X1`0300L701T70P0Q1`0300L703l7100U1`00C0L00`071`0W1`0300L701/70P0N1`03
00L703`7100Y1`00C@L00`071`0V1`0300L701d700<01`L06PL00`071`0i1`@0;@L004h70P0V1`03
00L701h70P0H1`80>0L40347001@1`0300L702<700<01`L080L201D700<01`L0=0L403D7001A1`03
00L7028700<01`L08PL201470P0d1`<0>@L0058700<01`L08@L00`071`0T1`<02`L30387100l1`00
D`L2024700<01`L09`L300@7100a1`@0@0L005D700<01`L07PL00`071`0Z1`@0<@L404@7001F1`03
00L701d700<01`L0F`L404P7001G1`0300L701`700<01`L0E`L404`7001H1`8070L00`071`1C1`@0
D0L005X700<01`L06@L00`071`1@1`<0E0L005/700<01`L060L00`071`1<1`@0E`L005`700<01`L0
5`L00`071`181`@0F`L005d700<01`L05PL00`071`141`@0G`L005h70P0F1`0300L70407101S1`00
H0L00`071`0C1`0300L703`7101W1`00H@L00`071`0B1`0300L703P7101[1`00HPL00`071`0A1`03
00L703D70`1_1`00H`L2014700<01`L0<@L40787001U1`0300L700h700<01`L0;@L407H7001V1`03
00L700d700<01`L0:@L407X7001W1`0300L700/700<01`L09PL407h7001X1`802`L00`071`0R1`@0
PPL006X700<01`L020L00`071`0N1`@0QPL006/700<01`L01`L00`071`0J1`@0RPL006`700<01`L0
1PL00`071`0G1`<0SPL006d70P061`0300L701<7102A1`00K`L00`071`031`0300L700l7102E1`00
L0L00`071`021`0300L700/7102I1`00L@L01@071`L000T7102M1`00LPL200031`0700@7102Q1`00
M0L200031`000080Y@L007D70P2Y1`00o`LQ1`00\
\>"],
  ImageRangeCache->{{{0, 287}, {242.375, 0}} -> {-0.053123, 0.0113765, 
  0.00353105, 0.00353105}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell["\<\
Since the size of the knot is large we can use a large tube size.\
\
\>", "Text",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(MVClear[]; \ 
    MVShow3D[knot1, MVLineTubeSize -> 2.0, MVPolygonShading -> MVSmooth]; 
    \)], "Input",
  CellLabel->"In[139]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00238095 0.0238095 0.00238095 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
1200 string
400 400 8 [400 0 0 400 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000423939423131\
4A39394231314A3939423131423939423131423939423131423131423131000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003931314A39394A39395239395A3939634242\
5A39395A39395239395A39395A39395A39395231395A3939523939523939523131
5239395231315239394A31315239394A31315231314A3131000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003931314231314A39394A313952393952394263424A5A39426342425A4242\
6342425A4242634242634242634242634242634242634242634242634239634242
6342426342425A42396342426342396342426342396342395A39395A39394A31314A39314A3131\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4231394231314A39394A39395239425A394263424A63424A634A4A63424A6B4A4A634A4A634A4A\
63424A6B4A4A634A4A6B4A4A63424A6B4A4A6B4A4A6B4A4A634A426B4A4A6B4A4A
6B4A4A6B4A426B4A4A6B4A4A6B4A4A6342426B4A426342426B4A426342396342395A3939523939\
4A31314A3939000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003931314231394231394A3939
4A31395239425A394263424A5A4242634A4A634A4A6B4A52634A4A6B4A52634A526B5252634A4A\
6B4A526B4A526B52526B4A4A6B4A526B4A4A6B52526B4A4A6B52526B4A4A735252
6B4A4A6B524A6B4A4A7352526B4A4A6B4A4A6B4A426B4A4A6342426B4A426342426B4A425A3939\
5A42395239395239394A3131000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003931393931314231394231395239425239395A39425A3942
63424A63424A634A4A63424A6B4A526B4A526B52526B4A526B525A6B52526B525A6B4A5273525A\
6B525273525A6B525273525A6B52527352526B525273525A6B52527352526B4A52
7352526B52527352526B4A527352526B524A73524A6B4A4A73524A6B4A426B4A426B4A426B4A42\
6342396342395A39315A4239523931523931000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003931314231394A31395239424A31395A39425A394263424A5A424263424A
63424A6B4A52634A4A6B4A526B4A526B525A6B4A526B525A6B525A735A636B525A735A5A73525A\
735A636B525A73525A73525A735A5A6B525A73525A73525A735A5A6B525273525A
73525A735A5A6B525273525A7352527352526B4A4A7352526B4A4A73524A6B4A426B4A4A6B4A42\
6B4A426B42396B4A426342396342395A3931523931523131000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3931395239425239425A42425A394263424A5A424A63424A63424A6B4A52634A52
6B52526B4A526B525A6B525A73525A6B525A735A63735A63735A63735A637B5A6B735A63735A63\
735A637B5A63735A63735A63735A5A7B5A63735A63735A63735A5A7B5A63735A63
735A63735A5A7B5A63735A5A735A5A7352527B5A5A73525273525273524A73525273524A73524A\
6B4A4273524A6B4A426B4A426B42396B4A426342396342395239315A3939000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000423939423139\
5239425A39425A424A5A424263424A63424A634A52634A4A6B4A526B4A526B525A
6B525A6B525A6B525A735A636B525A735A63735A637B636B735A63735A6B735A637B636B735A63\
735A6B735A637B636B735A63735A63735A637B636B735A637B5A63735A637B636B
735A63735A63735A637B5A6373525A735A5A73525A7B5A5A73525273525273525273525273524A\
73525273524A73524A6B4A42734A4A6B4A42734A426342396342395A42395A4239
523131000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003931314239394A393952424A\
5A424A5A424A5A424A634A52634A526B4A52634A526B525A6B525A6B525A6B525A
735A63735A63735A63735A637B636B735A6B7B636B735A6B7B636B735A6B73636B735A6B7B636B\
735A6B7B636B735A637B636B735A6B7B636B735A637B636B735A6B7B636B735A63
7B636B735A63735A63735A637B5A63735A637B5A6373525A7B5A5A73525A735A5A735252735A5A\
73525273525273524A73525273524A73524A6B4A4273524A6B4A426B4A42634239
6342395A39395A3931000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003931314A39424A39425A424A5A424A\
634A525A424A634A52634A526B525A6B4A5A6B525A6B525A735A636B5263735A63
735A6373636B735A6B73636B73636B7B637373636B73636B73636B7B6373735A6B73636B73636B\
7B6373735A6B7B636B73636B7B636B735A6B7B636B735A6B7B636B735A63735A6B
735A637B636B735A63735A63735A637B5A6373525A735A63735A5A7B5A6373525A735A5A735252\
735A5A73525273525273524A7352526B4A4A73524A734A4A73524A6B4A42734A42
6B4A396B4A425A39315A3939000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003939394231394A3942523942634A52634A52634A52\
634A526B525A6B525A6B52636B525A735A63735A63735A63735A637B636B73636B
7B636B735A6B7B63737B63737B637373636B7B6B737B63737B637373636B7B6B737B63737B6373\
73636B7B63737B63737B637373636B7B637373636B7B636B735A6B7B637373636B
7B636B735A637B636B735A637B636B735A637B636B735A637B5A63735A5A7B5A637B5A5A7B5A5A\
73525A7B5A5A735252735A527352527B5A5273525273524A734A4A73524A734A42
7352426B4A396B4A426342395A4239523931000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000423139423939524242523942634252634A526B525A634A52\
6B525A6B525A735A636B5263735A63735A637B636B735A6B7B636B73636B7B6373
73636B7B63737B63737B6B737363737B63737B63737B6B737363737B63737B63737B6B73736373\
7B63737B63737B6B7373636B7B63737363737B6373735A6B736373735A6B7B6373
735A6B73636B735A6B7B636B735A637B5A6B735A637B636B735A637B5A63735A637B5A63735A5A\
7B5A63735A5A7B5A5A7352527B5A5A7352527B5A5273524A73524A73524A73524A
6B4A427352426B4A426B4A426342395A42395A3931000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003931314239394A394252424A5A424A634A52634A526B525A634A5A6B5A63\
6B5263735A6B735A637B636B73636B73636B73636B7B637373636B7B637373636B
7B6B737B63737B6B737363737B6B7B7B6B737B6B7B7363737B6B7B7B6B737B6B7B7363737B6B7B\
7B63737B6B737363737B6B7373637373637373636B7B6373736373736373735A6B
7B637373636B73636B735A6B7B636B735A6B7B636B735A637B636B735A637B5A63735A637B6363\
7B5A637B5A63735A5A7B5A637B5A5A7B5A5A7352527B5A5273525273525273524A
73524A7352427352426B4A396B4A42634239634239523931000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003931314239394A39425A424A5A424A634A52634A526B525A634A5A6B52636B5A63\
73636B735A6B73636B73636B7B637373636B7B63737363737B6B737363737B6B73
7B6B737B6B7B7363737B6B7B7B6B737B6B7B736B737B6B7B7B6B7B7B6B7B7363737B6B7B736373\
7B6B7B7363737363737363737B6B7373637373637373637373637373636B736373
73636B7B6373735A6B73636B735A6B7B636B735A6B735A6B735A637B636B735A637B5A63735A63\
7B5A6373525A7B5A63735A5A7B5A6373525A7B5A5A7352527B5A5273524A735252
73524A73524A734A427352426B4A426B4A426342396342395A3931000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3931394231394A394252394A5A42525A4252634A5A634A5A6B525A6B525A735A63735A6373636B\
735A6B7B63737363737B63737363737B6B737B63737B6B737B6373846B7B7B6B7B
7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B73847B6B7B7B6B7B7B6B7B7B6B847B6B7B7B6B7B\
73637B7B6B7B73637B736B7B7363737B6B7B73637373637B7363737B637B736373
736373735A6B7B637373636B736373735A6B7B6373735A6B7B636B735A637B636B735A637B5A63\
735A637B63637B5A637B5A637B5A5A7B5A637B5A5A7B5A5A7352527B5A5A7B5252
7B5A5273524A7B5A4A73524A73524A6B4A42734A426342396342395A3931000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000393131423939
4231394A394252424A634A525A4252634A5A634A5A6B52636B525A6B5A63735A6373636B735A6B\
7B63737363737B6B737363737B6B737B6B737B6B7B7B6B737B6B7B7B6B7B84737B
7B6B7B7B6B7B7B6B7B7B73847B6B7B7B6B7B7B6B7B7B7384736B7B7B6B7B736B7B7B6B8473637B\
736B7B73637B7B6B7B73637373637B73637B73637B6B637373637373637373637B
735A73736373735A73736373735A6B73636B735A6B73636B735A6B735A6B735A6B7B636B735A63\
735A63735A637B5A63735A5A7B5A637B5A5A7B5A63735A5A7B5A5A7B5A527B5A5A
7352527B5A5273524A7B524A735242735242734A42734A426342396342395A4231000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003931314239394A3942
52424A5A424A5A4A525A4A526B525A63525A6B52636B5263735A6B735A6B73636B73636B7B6B73\
7B6B737B6B737B63737B6B7B7B6B7B7B6B7B7B6B7B84737B7B6B7B7B737B7B6B7B
8473847B6B7B7B73847B6B7B7B73847B6B847B7384736B7B7B7384736B7B7B6B7B736B7B7B6B84\
736B7B736B7B73637B736B7B73637B73637B6B637373637B6B63737363736B5A73
7363737363737363736B5A6B736373735A6B736373735A6B736373735A6B735A6B735A637B636B\
735A63735A6B735A637B5A637B5A637B5A63735A5A7B5A637B5A5A7B5A5A735252
7B5A5A7B52527B5A5273524A7B5A5273524A73524A734A427352426B4A396342395A3931000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003931313931394A39424A394252424A
5A424A634A525A4A5263525A63525A6B5A636B5263735A6B735A6B7B637373636B7B6B737B6B73\
7B6B7B736B737B6B7B7B6B7B7B737B7B6B7B7B737B7B6B7B8473847B6B7B7B7384
7B6B7B7B73847B6B7B7B73847B6B847B7384736B7B736B84736B7B7B6B84736B7B736B7B736B7B\
736B7B6B637B736B7B6B637B736B7B6B63736B637B6B637373637B6B5A73736373
6B63737363736B5A6B7363736B5A737363736B5A6B735A736B5A6B735A736B5A6B735A6B735A6B\
735A6B735A63735A63735A637B5A63735A5A7B5A637B5A5A7B5A6373525A7B5A5A
7B5A527B5A5A7352527B5A5273524A7B5A5273524A73524A7352427352426B4239634A395A4231\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004239394239394A39424A39425A42525A4252
634A5A634A526B52636B52636B5A636B5A6373636B73636B7B63737B63737B6B7B7B6B7B7B6B7B\
7B6B7B8473847B6B7B8473847B6B7B8473848473848473847B73848473847B7384
7B73847B738484738C7B73847B7384736B847B7384736B84736B84736B7B7B7384736B7B736B84\
736B7B736B84736B7B736B7B6B637B6B637B6B637B6B637B6B5A7373637B6B6373
6B63736B5A7373637B6B63737363736B5A73736373735A737363736B5A6B736373735A6B735A6B\
735A6B7B636B735A637B5A6B735A637B636B7B5A637B5A637B5A5A8463637B5A5A
7B5A5A7B5A527B5A5A7B5A527B5A5273524A7B5A527B524A7B524A7352427352426B4A396B4A39\
634231000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000042393942394252424A52394A5A42525A4252634A5A
634A5A6B525A6B5263735A636B5A6373636B73636B7B6B737363737B6B7B7B6B7B84737B7B6B7B\
7B737B7B737B8473847B6B7B8473847B73848473847B73848473847B738484738C
7B73847B73847B73847B7384736B847B7384736B84737384736B7B736B84736B7B736B846B6B7B\
736B7B6B637B736B7B6B637B6B637B63637B6B637B635A736B637B6B5A736B637B
6B5A736B63736B5A736B63736B5A736B63736B5A737363736B5A6B735A73735A6B736373735A6B\
735A6B735A6B7B636B735A637B5A63735A637B6363735A637B5A637B5A5A7B6363
7B5A5A7B5A5A7B5A5A7B5A5A7352527B5A527B524A7B5A5273524A7B524A7352427352426B4A39\
6B4A39634231000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004239394A394252424A52424A5A4A525A4A52634A5A634A5A
6B52636B5263735A6B6B5A6B7363737363737B6B737B63737B6B7B7B6B7B7B737B7B6B7B847384\
7B737B8473847B7384847B8C8473848473847B7384847B8C7B73847B73847B7384
7B738C7B73847B73847373847B7384737384737384736B7B737384736B84736B846B6B7B736B84\
6B6B7B6B6B7B6B637B6B6B7B63637B6B637B635A736B637B6B63736B637B635A73
6B637B6B5A736B63736B5A736B63736B5A736B5A736B5A6B6B5A736B5A6B735A736B5A6B736373\
735A6B735A6B735A6B7B636B735A6B7B5A6B735A637B636B7B5A637B5A637B5A5A
7B63637B5A5A7B5A5A7B5A527B5A5A7B5A527B5A5273524A7B5A527B524A7B524A73524273524A\
6B4A426B4A39634231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000039313942394252424A52424A5A4A525A4A5263525A634A5A6B5263
6B5263735A6B6B5A6B7363737363737B6B7B7363737B6B7B7B6B7B8473847B6B7B7B73847B737B\
8473847B737B847384847384847B8C7B73847B73847B7384847B8C7B73847B7384
7B73847B73847373847373847373847373846B6B7B7373846B6B847373846B6B7B6B6B846B6B7B\
6B6B8463637B6B637B63637B63637B635A7363637B635A736B637B635A73636373
635A736B6373635A736B5A73635A736B5A7363526B6B5A736B5A6B6B5A736B5A6B735A736B5A6B\
735A736B5A6B735A6B735A6B735A6B735A63735A6B735A637B5A6B735A637B5A63
7B5A637B5A637B5A5A7B5A5A7B5A5A7B5A5A7352527B5A527B524A7B5A5273524A7B524A735242\
7352426B4A396B4A39634231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000042393942394252424A5A4252634A5A634A5A63525A63525A6B5A636B5A63
73636B735A6B7B63737B63737B6B7B7B6B7384737B7B6B7B8473847B6B7B847384847384847384\
7B7384847B8C847B8C847B8C7B7384847B8C7B7384847B8C7B7384847B8C7B7384
7B7B8C7373847B7B8C7373847373846B6B8473738C736B847373846B6B84736B846B6B846B6B84\
63637B6B6B8463637B63637B5A5A7363637B63637B63637B635A7363637B635A73
636373635A73636373635A73635A73635A6B6B5A73635A736B5A73635A6B6B5A736B5A6B735A73\
6B5A6B736373735A6B735A6B735A6B7B636B735A6B7B5A6B735A637B636B7B5A63
7B5A637B5A5A8463637B5A5A7B5A5A7B5A5A845A5A7B5A527B5A527B524A7B5A527B5A4A7B5A4A\
7352427352426B4A396B4A39634231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39313942394252424A52424A636B6B634A5A6B526363525A6B5A636B5A6373636B
735A6B7B63737B63737B6B7B7B6B737B6B7B7B6B7B8473847B6B7B8473847B7384847B8C7B7384\
847B8C847B84847B8C7B7384847B8C7B7384847B8C7B73847B7B8C7B73847B7B8C
7373847B73847373847373846B6B847373846B73847373846B6B846B6B846B6B7B6B6B8463637B\
63637B63637B63637B5A5A7363637B5A5A7363637B5A5A73635A7B5A5A7363637B
5A5A73635A735A5A73635A735A526B635A7363526B635A7363526B635A73635A6B6B5A736B5A6B\
6B5A6B6B5A6B735A736B5A6B735A6B735A6B7B5A6B735A637B5A63735A637B6363
735A637B5A637B5A637B63637B5A5A7B5A5A7B5A5A7B5A5A7B52527B5A527B5A527B5A527B524A\
7B524A7352427352426B4A396B4A39634231000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000423939\
42394252424A52424A63525A63525A6B5263635263735A6B6B5A6B73637373636B
7B6B7B7B6B737B6B7B7B6B7B8473847B6B7B8473847B73848473848473848473847B7384847B8C\
847B8C847B8C847B8C847B8C7B7B847B7B8C7B73847B7B8C7B73847B7B8C737384
7B7B8C7373847373846B6B847373846B6B846B73846B6B7B6B7384636B7B6B6B8463637B636B84\
63637B63637B5A5A7363637B5A5A735A5A735A5A735A5A7B5A5A735A5A735A5A73
635A735A5A735A5A735A526B635A735A526B635A735A526B635A73635A6B6B5A7363526B6B5A73\
6B5A6B735A736B5A6B735A73735A6B735A6B735A63735A6B735A63735A6B735A63
7B5A6B7B5A637B5A637B5A5A8463637B5A5A7B5A5A7B5A52845A5A7B5A527B5A527B524A7B5A52\
7B524A7B524A7352427352426B4A396B4A394A3931000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000393139423939\
52424A52424A634A5A634A5A6B52636352636B5A6B6B5A6B73637373636B7B6B73
7B6B737B6B7B7B6B7B7B6B7B7B6B7B8473847B6B84847384847384847B8C7B7384847B8C847B8C\
847B8C7B7B8C847B8C7B7B84847B8C7B73847B7B8C7B73847B7B8C737384737384
73738473738C6B6B847373846B6B846B73846B5A63636B7B636B7B636B7B5A637B63637B5A637B\
63637B5A5A735A6373525A735A5A7352526B525A73525A735A5A7352526B5A5A73
5A52735A5A7352526B5A52735A526B5A5A735A526B5A52735A526B635A7363526B635A7363526B\
6B5A736B526B6B5A6B6B5A6B735A6B6B526B735A6B735A6B735A6B735A63735A63
735A637B5A63735A637B5A637B5A637B5A637B5A5A7B5A5A7B5A527B5A5A7B52527B5A527B5A4A\
7B5A5273524A7352427352427352426B4A396B4A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004239394239424A424A\
52424A634A5A63525A6B52636B526373636B73636B7363737363737B6B7B7B6B7B
7B6B7B7B6B7B8473847B73848473847B7384847B8C84738C847B8C84738C847B8C847B8C847B8C\
847B8C84848C847B8C847B8C7B7B84847B8C7B7B8C7B7B8C7373847B7B8C737384
737B8C73738473738C6B73846B7384636B846B7384636B7B636B7B5A637B636B7B5A637B5A637B\
5A5A735A637B525A73525A73525A73525A73525A73525A7352526B5A5A73525273
525A7352526B5A5A735252735A527352526B5A5A735A526B635A735A526B635A73635A6B635A73\
63526B6B5A736B5A6B6B5A6B6B5A6B735A73735A6B735A6B735A637B5A6B735A63
7B5A6B735A637B636B7B5A637B63637B5A5A8463637B5A5A845A5A7B5A5284635A7B5A527B5A52\
7B5A4A845A527B524A7B52427352397352426B4A396B4A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003931394239394A424A52424A\
634A5A634A5A6B52636352636B5A6B735A6B7B63737363737B6B7B7B6B7B7B6B7B
7B6B7B7B73847B73848473847B7384847B8C84738C847B8C84738C847B8C847B8C848494847B8C\
847B8C847B8C847B8C7B7B847B7B8C7B7B8C7B7B8C7373847B7B8C737384737B8C
6B73847373846B73846B7384636B7B636B84636B7B636B7B5A63735A637B5A63735A637B525A73\
526373525A73525A734A526B4A5A734A526B525A734A526B5252734A526B525273
4A526B52527352526B52527352526B52526B52526B5A52735A526B5A52735A526B635A7363526B\
635A6B63526B6B5A736B526B6B5A6B6B5A6B735A6B6B526B735A6B735A63735A6B
735A637B5A637B5A637B63637B5A637B5A637B5A5A8463637B5A5A7B5A5A7B5A52845A5A7B5A52\
7B5A527B5A4A845A527352427352427352427352426B4A31634239000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003931394239394A424A52424A634A5A\
634A5A6B52636B5263735A6B6B5A6B7363737363737B6B7B7B6B7B7B6B7B7B6B7B
8473847B73848473847B7384847B8C847B8C847B8C847B8C847B94847B8C847B8C847B8C84848C\
847B8C847B8C7B7B847B7B8C7B7B8C7B7B8C737B847B7B8C737B84737B8C6B7384
7373846B73846B7384636B7B6B7384636B7B5A6B7B5A63735A6B7B5263735A6373525A73526373\
525A73525A734A526B4A5A734A526B4A526B42526B4A52734A526B4A526B4A4A6B
5252734A526B52526B4A4A6B52527352526B52526B52526B5A52735A526B5A52735A526B635A73\
63526B635A6B63526B6B5A736B5A6B6B5A6B6B526B735A6B6B5A6B735A6B735A63
7B5A6B735A637B5A6B735A637B636B7B5A637B5A637B5A5A8463637B5A5A845A5A7B5A52845A5A\
7B5A527B5A527B5A4A845A527B524A735242734A397352426B4A39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003931314231394A424A52424A5A4A5A634A5A\
6B52636352636B5A6B6B5A6B7363737363737B6B7B7B6B7B7B73847B6B7B7B7384
7B738484738C7B7384847B8C847B8C847B8C847B8C847B8C847B8C8484947B7B8C847B8C7B7B8C\
847B8C7B7B847B7B8C7B7B847B7B8C737B847B7B8C737384737B8C6B73846B7384
6B737B6B7384636B7B636B7B5A6B7B5A6B7B5263735263735263735263734A5A6B525A734A5A6B\
4A5A7342526B4A526B42526B42526B394A6342526B424A6B4A526B424A634A526B
424A6B4A526B424A634A4A6B4A4A6B4A526B4A4A6B52526B52526B5A5273524A6B5A526B5A526B\
6352735A526B63526B63526B6B5A6B63526B6B5A6B6B526B735A6B6B5263735A6B
735A637B5A6B735A637B5A637B5A637B63637B5A637B5A637B5A5A8463637B5A5A7B5A5A7B5A52\
845A5A7B5A4A7B5A527B5A4A7B5A4A735242735242734A397352396B4A31000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003939394239394A424A52424A634A5A634A5A635263\
635263735A6B735A6B7363737363737B6B7B7B6B7B7B73847B6B7B847384847384
847B8C8473848C7B8C847B8C847B94847B8C8C8494847B8C848494847B8C848494847B8C84848C\
7B7B8484848C7B7B8C7B7B8C737B847B7B8C737B84737B846B7384737B846B737B
6B7384636B7B63737B636B7B5A6B7B5263735A63735263735263734A5A6B5263734A5A6B4A5A6B\
4252634A5A6B42526342526B394A6342526B5A4242424A63424A6342526B424A63
424A6B424A634A4A6B424A634A4A6B424A634A526B4A4A6B52526B524A6B5A52735A526B5A526B\
5A526B635A7363526B63526B63526B6B5A736B5A6B6B5A6B6B5A6B735A6B735A6B
735A6B735A637B636B7B5A637B636B7B5A6384636B7B5A638463637B5A5A8463637B5A5A84635A\
7B5A5284635A7B5A52845A527B5A4A845A4A7B5242735242734A397352396B4A31
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003931314239394A424A4A424A5A4A5A634A5A6B5263635263\
6B5A6B6B5A6B7363737363737B6B7B7B6B7B8473847B6B7B847384847384847B8C
847384847B8C847B8C8C8494847B8C847B94847B8C848494847B8C847B8C847B8C84848C7B7B84\
7B7B8C7B7B8C7B7B8C737B847B7B8C737B84737B846B737B6B73846B737B6B737B
636B7B636B7B5A6B735A6B7B526373526373525A6B526373425A634A5A6342525A425263394A5A\
424A5A394A5A394A5A394252394A5A39425A39425A39395242425A39425A424A63
394263424A63424263424A63424263424A63424A634A4A6B4A4A63524A6B524A6B5A526B524A6B\
5A526B5A526B63526B5A526B63526B63526B6B5A6B6B526B6B5A6B6B5A6B735A6B
6B5A63735A6B735A637B636B735A637B5A637B5A638463637B5A5A8463637B5A5A8463637B5A5A\
845A5A7B5A5284635A7B5A52845A527B5A4A7B5A4A735242735242735239735239
4A3931000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004231394A39424A424A5A4A5A634A5A635263635263735A6B\
6B5A6B7363737363737B6B7B7B6B7B8473847B6B7B847384847384847B8C84738C
847B94847B8C847B94847B8C8C8494847B8C84848C847B8C84848C847B8C84848C7B7B8484848C\
7B7B8C7B7B8C737B847B7B8C737B84737B846B737B6B73846B737B63737B636B73
63737B5A6B735A6B7352636B52637352636B4A5A6B42525A425263424A52394A52314252394A52\
39425239425231424A39425239394A39424A31394A39395239394A393952393952
42426339426342426339425A424A63394263424A634242634A4A6B4A4A6B524A6B524A6B5A526B\
52526B5A526B5A4A6B63526B5A526B63526B63526B6B5A736B5A6B735A6B6B526B
735A6B735A6B735A6B735A637B636B7B5A637B5A6B7B5A6384636B7B5A638463637B5A5A846363\
7B5A5A845A5A7B5A5284635A7B5A52845A527B524A7B5A4A7352427352426B4A39
735239000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003931394A394A4A394A5A4A525A4A5A6352636352636B5A6B6B5A6B\
7363737363737B6B7B7B6B7B8473847B6B7B847384847384847B8484738C847B8C
847B8C8C8494847B8C847B94847B8C8484947B7B8C847B8C847B8C84848C7B7B847B7B8C7B7B84\
7B7B8C737B84737B84737B84737B846B737B6B737B636B7B63737B5A6B73636B73
5A6B735A6B7352636B52636B4A5A63425263394A52394A5239424A39424A000000000000000000\
00000000000000000000000000000000000000000000000000000000000039394A
39394A39395231395239425A39425A39426339425A4242634242634A4A6B4A4A63524A6B524A6B\
52526B524A635A526B5A4A6B5A526B5A4A6B63526B63526B6B5A736B526B6B5A6B
6B5A6B735A6B735A63735A6B735A637B636B735A637B5A637B5A638463637B5A5A845A637B5A5A\
84635A7B5A52845A5A7B5A528463527B5A4A845A527B5A4A7B5A4A735242735242
6B4A39735239000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003939394A39424A394A5A4A5A634A5A6352636352636B5A6B6B5A6B736373\
7363737B6B7B7B6B7B8473847B737B847B84847384847B848473848C7B8C847B94
8C8494847B8C8C849484848C848494847B8C84849484848C84848C7B7B8484848C7B7B8C7B848C\
737B847B848C737B84737B846B737B6B738463737B63737B5A6B7363737B5A6B73
5A63734A5A634A5A6342525A42525A394A4A394A4A000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000039395239395239425A5A39314242633942634242634242634A4A6B4A4A63524A6B\
4A4A6B5A526B524A6B5A526B5A4A6B63526B63526B63526B63526B6B5A736B5A6B
735A6B6B5A6B735A6B735A6B7B5A6B735A637B636B7B5A637B63637B5A63846363845A63846363\
7B5A5A846363845A5A84635A7B5A5284635A845A52845A527B5A4A7B5A4A735242
7352426B4A39634A39000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003931394A42424A394A5A4A525A4A5A6352636352636B5A6B6B5A6B736373736373\
7B6B7B7B6B7B8473847B6B7B847384847384847B8C847384847B8C847B8C8C8494
847B8C848494847B8C848494847B8C84848C847B8C84848C7B7B8484848C7B7B8C7B848C737B84\
7B7B8C737B84737B846B737B6B737B636B7363737B5A6B6B5A6B7352636B526363
42525A42525A394A52394A52000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031314239394A39395239426339395A4242634242634A4A63424263\
4A4A6B4A4A6B524A6B524A635A526B5A4A6B63526B5A526B63526B6B526B6B5A6B
6B526B735A6B735A6B735A6B735A637B5A6B7B5A637B636B7B5A637B63637B5A638463637B5A5A\
846363845A5A84635A7B5A52845A5A845A528463527B5A4A845A4A7B5A4A7B5A4A
7352397352396B4A39000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003931394A39424A394A5A4A5A5A4A5A6352636352636B5A6B6B5A6B7363737363737B6B7B\
7B6B7B7B73847B6B7B847384847384847B84847384847B8C847B8C847B8C847B8C
8C8494847B8C848494847B8C848494847B8C84848C7B7B8484848C7B7B8C7B848C737B847B848C\
737B84737B846B737B6B737B636B73636B735A6B6B5A6B7352636B4A5A5A395252
395252000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039394A31394A39426339425A4242634242634A4A63\
4A42634A4A6B4A4A6352526B524A6B5A526B5A4A6B63526B63526B6B526B6B526B
735A736B5A6B735A6B6B5A6B735A6B735A6B7B5A6B735A637B636B7B5A638463637B5A63846363\
845A638463637B5A5A84635A845A5A84635A7B5A52846352845A52845A4A7B5242
7B5A4A7352427352396B4A31000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3931394A39424A394A5242525A4A5A6352636352636B5A6B6B5A6B73637373636B736B737B6B7B\
8473847B6B7B847384847384847B8C7B7384847B8C847B8C84848C847B8C84848C
847B8C848494847B8C84848C847B8C84848C7B7B847B848C7B7B847B848C737B84737B84737B84\
737B846B737B6B737B636B736373735A636B5A6B6B4A63634A525A394242000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039314A39425A39395A424263424263\
4A42634242634A4A6B4A4A63524A6B524A635A526B5A4A6B63526B63526B6B526B
6B526B735A6B6B526B735A6B735A6B735A6B735A637B5A6B7B5A637B636B7B5A638463637B5A63\
8463637B5A5A84635A845A5A84635A7B5A52845A52845A528463527B5A4A7B5A4A
7B52427B5A427352397352396B4A31000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000393139
4239424A394A5A4A5A5A4A5A6352636352636B5A6B6B5A6B7363737363737B6B7B7B6B7B7B7384\
7B6B7B847B84847384847B8C8473848C7B8C847B8C84848C847B8C8C849484848C
8C8494847B8C84849484848C84848C7B7B8C84848C7B848C7B848C7B7B847B848C737B84737B84\
6B737B6B737B637373636B735A6B6B5A6B734A5A5A425252000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039314A42426339395A424263\
42395A4A42634A4263524A6B4A4A635A526B5A4A6B5A526B5A526B63526B63526B
6B5A6B6B526B735A736B5A6B735A6B735A6B7B636B735A6B7B5A6B7B5A6384636B7B6363846363\
7B5A63846363846363846363845A5A84635A845A5A84635A845A52846352845A4A
7B5A4A7B52427B5A42735239735239523931000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000393139423942
4A394A5242525A4A5A6352636352636B5A6B6B5A6B736373736373736B737B6B7B8473847B6B7B\
847384847384847B8C847384847B8C847B8C8C848C847B8C8C848C84848C8C8494
84848C84848C84848C84848C7B7B8484848C7B848C84848C737B84737B84737B7B737B846B7373\
6B737B636B736373735A636B52636B425252424A4A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039314A39395A39395A\
42425A42425A4A42634A4263524A6B524A635A526B524A635A526B5A526B63526B
63526B6B5A6B6B526B735A6B6B526B735A6B735A6B7B5A6B735A637B5A6B7B5A6384636B7B5A63\
846363845A638463637B5A5A84635A845A5A84635A845A52846352845A52846352
7B5A4A7B5A4A7B52427B5A426B4A39735239000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003931394239424A394A
524A525A4A5A6352636352636B5A6B6B5A6B73637373636B7B6B7B7B6B7B7B737B7B6B7B847384\
847384847B84847384847B8C847B8C847B8C847B8C8C849484848C8C849484848C
84848C84848C84848C7B7B8484848C7B848C7B848C737B847B7B84737B7B737B7B6B73736B737B\
636B73636B735A6B6B526363425252394A42000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000031314A39395A\
39395A42395A42395A4A4A634A4A63524A6B524A635A526B5A4A6B5A526B5A526B
6B527363526B6B5A6B6B526B735A6B735A6B735A6B735A637B5A6B7B5A6B7B5A6B7B5A6384636B\
8463638463637B5A6384636384635A84635A845A5A84635A845A52846352845A52
8463527B5A4A7B5A4A735239735242735239634A39000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004239424A3942524252
5A4A5A635263634A636B52636B5A6B7363736B5A6B736373736B7B7B73847B6B7B7B73847B7384\
847B8C7B7384847B8C847B8C8C848C847B8C84848C84848C8C8494847B8C84848C
84848C84848C7B7B8484848C7B84847B848C737B7B737B8473737B737B7B6B73736B7373636B73\
636B735263634A5A5A394A4A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393142\
39395A39315242395A42425A4A4A634A4263524A6B524A635A526B524A635A526B
63526B6B526B63526B6B5A6B6B526B735A6B6B526B735A6B735A6B7B5A6B735A637B5A6B7B5A63\
84636B7B5A638463638463638463637B5A5A84635A845A5A84635A845A52846352
845A4A845A527B52427B5A427352397352396B4A31523931000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000423942423942524A525A4A5A
6352636352636B5A6B6B5A6B7363736B63737B6B7B7B6B7B7B73847B6B7B847384847384847B8C\
8473848C7B8C847B8C8C848C847B8C8C84948C848C8C849484848C8C849484848C
84848C7B7B8484848C7B84847B8484737B847B7B84737B7B737B7B6B73736B737B636B73636B73\
5263634A5A52394A42000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39314239395A42395A42426342425A4A4A6B4A4A63524A6B524A635A526B5A526B
63526B63526B6B5A736B526B6B5A6B6B526B735A6B735A6B735A6B735A637B636B7B5A6B84636B\
7B5A6384636B846363846363845A5A8C636384635A84635A845A5A8C635A846352
846352845A4A84634A7B5A427B5A42735239735239634A31000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004231424A424A5A4A5A635263
5A4A5A6352636B5A6B7363736B5A6B736373736B7B7B73847B6B7B8473847B7384847B8C847384\
847B8C847B8C8C848C847B8C8C849484848C8C849484848C84848C84848C84848C
7B7B8484848C7B7B847B848C737B84737B7B73737B737B7B637373637373636B6B636B6B4A5A5A\
42524A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000039314242395A39395A4242634242634A4A6B4A4263524A6B524A635A526B
5A4A6363526B63526B6B5A6B6B526B6B5A6B6B5A6B735A6B735A6B735A6B735A6B7B636B7B5A63\
84636B84636384636B845A6384636384635A846363845A5A84635A845A5A8C635A
845A52846352845A4A845A4A7B52427B5A42735239735239523931000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003931394A424A5242525A4A5A5A4A63
6B5A6B6B5A6B6B5A736B5A6B736B7B736B7B7B6B7B7B6B7B8473847B7384847384847384847B8C\
847B8C847B8C847B8C8C849484848C8C849484848C8C849484848C84848C7B7B84
84848C7B84847B8484737B847B8484737B7B6B737B6373736B7373636B6B5A6B634A5252424A4A\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000039315242395A42395A424263424263524A6B4A4A63524A6B524A63
5A526B5A526B63526B63526B6B5A736B526B735A6B6B526B735A6B735A6B7B5A6B735A637B636B\
7B5A6B84636B84636384636B846363846363845A5A8C636384635A84635A845A52
8C635A8463528463527B5A4A845A4A7B52427B52427352396B4A39000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000039313942394A4A424A5A4A5A5A4A5A635263
63526B6B5A736B5A6B736373736B737B737B7B6B7B7B73847B7384847B847B7384847B8C847B8C\
8C848C847B8C84848C84848C8C849484848C84848C84848C84848C7B7B8484848C
7B7B847B8484737B847B7B84737B7B737B7B636B6B637373636B6B52636342524A000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000039315242395A39395A4242634A4263524A6B4A4263524A6B
524A635A526B5A4A6363526B63526B6B5A736B526B735A6B735A6B735A6B735A6B7B5A6B7B5A6B\
84636B7B5A6384636B84636384636B845A6384636384635A846363845A5A84635A
8463528C635A845A4A845A527B5A4A7B5A4A7352397352397352395A4239000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004239424A424A524A5A5A4A5A6B5A6B6B5A6B
6B5A6B6B5A6B736B7B736B7B7B6B7B7B6B7B8473847B7384847B84847384847B8C847B8C8C848C\
847B8C8C84948C848C8C849484848C8C8C9484848C84848C84848484848C7B8484
7B84847B7B847B8484737B7B737B7B6373736373735A6B6B5A6B63313931000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000039315242425A42395A4A42634A4263524A6B524A63
524A6B524A6363526B63526B63526B63526B735A736B5A6B735A6B735A6B7B636B7B5A6B7B636B\
7B5A6B84636B84636B84636B8463638C636B84636384636384635A8C636384635A
8C635A845A528C635A84635284634A7B5A427B5A427B52427352396B4A31000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000042314242394A524252524A5A635263635A6B6B6373
6B5A6B73637373637B7B73847B6B7B7B73847B7384847B8C7B7384847B8C847B8C8C848C847B8C\
8C849484848C8C8C9484848C84848C84848C84848C847B8484848C7B848484848C
7B7B847B8484737B7B737B7B63737363736B5A6B635A6B63393939000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000039315239395242425A42395A4A42634A4263524A6B
4A4263524A6B5A4A6B63526B5A526B6B526B6B526B735A736B526B735A6B735A6B7B636B735A6B\
7B636B7B5A6B84636B845A6384636B8463638C636384635A84636384635A8C6363
845A5A8C635A8463528C6352845A4A845A4A7B5A427B5A427352397352395A4231000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000042314242394252425252425263526363526B6B5A6B6B5A6B
73637B73637B7B6B7B7B6B7B8473847B73848473847B7384847B8C847B8C847B8C847B8C8C8494\
8484948C849484848C8C849484848C84848C84848484848C7B84847B84847B7B84
7B8484737B7B737B7B6B73736B7373636B6B5A6B63394239000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000039395239395242425A4242634A42634A4263
524A6B524A635A4A6B5A4A6363526B63526B6B5A6B6B526B735A73735A6B735A6B735A6B7B636B\
7B5A6B7B636B7B5A6B84636B84636B84636B8463638C636384636384636384635A
8C635A84635A8C635A8463528C6352845A4A845A4A7B52427B5A42735239735239000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004231424A424A4A42525A4A5A6352636B5A6B6B5A6B736373
7363737B6B7B7B6B7B7B73847B7384847B847B7384847B8C847B8C84848C847B8C84849484848C\
8C849484848C84848C84848C84848C7B7B8484848C7B84848484847B7B847B8484
737B7B737B7B6B737363736B5A6B6B5A6B63394239000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000039395239395242426342395A4A4263
4A4263524A6B524A635A4A6B5A4A6B63526B63526B6B5A6B6B526B735A736B5A6B735A6B735A6B\
7B636B7B5A6B7B636B7B636B84636B8463638463638463638C636384635A84635A
84635A8C635A8463528C63528463528463527B5A4A7B5A4A7B5A427B5A42735231634239000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003931424A394A4A424A5A4A5A6352636B5A6B6B5A6B736373736373
736B7B736B7B8473847B73848473847B7384847B8C847B8C84848C847B8C8C84948C84948C8494\
84848C8C8C9484848C84848C84848C84848C8484848484847B7B84848484737B7B
737B7B6B73736B7373636B6B5A6B63424A42000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000039315242395A39395A4A4263424263
4A42634A42635A4A6B5A4A6B5A526B5A526B6B5A736B526B6B5A736B5A6B735A73735A6B7B636B\
735A6B84636B7B636B84636B84636B8C636B84636B8C636B8463638C6363846363
8C636384635A8C635A8C635A8C635A8463528C6352845A4A7B5A4A7B52397B5A42735239523939\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000042394A4A394A524A5A5A4A5A6B5A6B63526B6B63736B6373736B7B
73637B7B6B847B7384847B8C7B7384847B8C847B8C8C848C847B8C8C849484848C8C8C9484848C\
8C849484848C8C848C84848484848C84848484848C7B7B847B8484737B7B737B7B
6B73736B7373636B6B636B63424A42000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000039315242395A42395A4A4263
4242634A4A63524A635A4A6B5A4A6363526B63526B6B5A736B526B735A6B735A6B735A73735A6B\
7B636B7B5A6B84636B7B5A6B84636B84636B8C636B8463638C63638463638C6363
84635A8C635A84635A8C635A8463528C63528463528463527B5A427B5A427B52397B5A39634231\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000042394242394A524A525A4A5A63526B63526B6B63736B637373637B73637B
7B73847B738484738C7B738C847B8C847B8C847B8C847B8C8C84948484948C849484848C8C8C94\
84848C84848C84848C84848C8484848484847B7B848484847B7B7B737B7B6B7373
6B7373636B6B636B6342524A000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039395242395A42395A
4A42634A4263524A6B524A635A526B5A526B63526B63526B6B5A736B5A6B735A736B5A6B7B6373\
735A6B7B636B7B5A6B84637384636B84636B84636B8C6B6B8463638C6363846363
8C6B6384635A8C635A84635A8C6B5A8C63528C6352845A4A84634A7B5A427B5A427352397B5A39\
523931000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3931424239424A4252524A5A6352636352636B5A6B6B5A6B73637B7363737B6B7B
7B6B8484738C7B7384847B8C847B8C847B8C847B8C84849484848C8C849484848C84849484848C\
84848C84848484848C84848484848C7B7B847B84847B7B7B737B7B6B73736B7373
636B6B636B6B42524A000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039395239315242395A
42395A4A42634A4263524A6B524A6B5A526B5A4A6B63526B63526B6B5A736B526B735A6B735A6B\
7B6373735A6B7B636B7B5A6B84636B7B636B84636B84636B8C6B6B8463638C6363
8463638C636384635A8C635A8C635A5A4A63845A52846352845A4A84634A7B5A427B5A42735239\
6B4A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3931424A42525242525A52636352636B5A736B5A737363737363737B6B847B6B84
7B73847B738C847B8C847B8C847B8C847B8C8C84948C84948C84948484948C8C9484848C84848C\
84848C84848C84848C84848C7B84848484847B7B84737B7B6B73736B7373636B6B
636B634A5A52000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000423952423952
42425A42425A4A4A6B4A4A63524A6B524A6363526B63526B6B526B6B526B735A73735A6B735A73\
735A6B7B63737B636B7B636B7B5A6B84637384636B8C636B8463638C6B6B8C6363
8C63638C63638C6B638C63638C635A8C635A8C635A846352846352845A4A84634A845A427B5A42\
3931525A4239000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393139\
42394A4A42525A4A5A5A5263635A6B6B5A6B7363736B6373736B7B736B7B7B7384
7B7384847B8C847B8C847B8C847B8C847B8C847B8C8C849484848C8C849484848C8C849484848C\
84848C84848484848C7B7B847B84847B7B847B7B7B6B73736B7373636B6B636B6B
525A52000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000423952
393952424263424263524A6B4A42635A4A6B5A4A6B63526B63526B6B526B6B5A6B735A736B5A6B\
735A6B735A6B7B63737B5A6B7B636B7B636B84636B84636B735A638463638C6B6B
8463638C63638C63638C6B638C635A8C635A8C63528C635A845A4A84634A845A4A84634A7B5A42\
7352426B4A39000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039313942394A\
4A4252524A5A5A4A5A6B5A6B6B5A6B6B63736B6373736B7B736B7B7B73847B7384
847B8C847B8C847B8C847B8C8C848C847B8C84848C84848C8C849484848C84848C84848C84848C\
84848484848C7B7B848484847B7B847B7B7B6B7373737373636B6B636B6B52635A
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000039314A
42395A42395A4A42634A4263524A6B524A6B5A4A6B5A4A6B63527363526B6B5A736B526B735A73\
735A6B735A73735A6B7B63737B636B84636B7B636B84636B84636B84636B846363
8C6B6B8C63638C6B638C635A946B638C635A8C635A8463528C635284635284634A7B5A4284634A\
735242735239634A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000042394242394A\
524A52524A5A63526B63526B6B5A736B5A6B73637B736B7B7B73847B6B8484738C
84738C847B8C847B8C847B8C847B8C8C848C847B8C8C848C84848C8C8C8C84848C84848C848484\
84848C7B7B847B84847B7B7B7B7B7B6B73736B7373636B6B636B6B5A635A000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
39315242395A42395A4A42634A4263524A6B524A635A526B5A4A6B63527363526B6B5A736B5A6B\
735A73735A6B735A6B735A6B7B63737B636B84636B84636B84636B84636384636B
8463638C6B6B8463638C6B638C63638C6B6384635A8C635A8463528C6352845A4A84634A845A42\
7B5A426B52396B4A395A4231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000042394242394A524252\
524A5A63526B63526B6B5A736B5A7373637B73637B7B6B847B6B84847B8C847B8C
847B8C847B8C8C8494847B8C8C848C84848C8C84948C848C8C8C9484848C8C8C8C84848484848C\
7B848484848C7B84847B7B7B737373737B736B736B636B6B5A635A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3939523939524242634242634A426B4A42635A4A6B5A4A6B63526B5A526B6B5A736B5A6B735A73\
6B5A6B7B6373735A6B7B63737B5A6B84637384636B84636B84636B8C6B6B84636B
8C6B6B8C63638C6B6B8C6B638C6B638C635A946B638C635A8C635A8463528C635284634A84634A\
845A427B5A426B5239634A39000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004239424A4252524252\
5A52636352636B5A6B6B5A6B73637373637B7B6B84736B7B7B738484738C847B8C
847B8C847B8C847B8C8C848C847B8C8C848C84848C8C8C9484848C84848C84848484848C7B7B84\
8484847B84847B84847373737373736B736B63736B5A635A000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000039395242395A4242634242634A4A63524A635A4A6B5A4A6363526B63526B6B5A736B526B\
735A73735A6B736373735A6B7B63737B636B84637384636B84636B84636B8C6B6B
8463638C6B6B8C63638C6B6B8C63638C6B638C635A8C6B638C63528C63528463528C6352845A4A\
84634A7B5A42735A42634A39000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003931424A424A4A42525A5263\
635263635A6B635A6B736373736373736B7B736B7B7B73847B7384847B8C84738C
847B8C847B8C84848C847B8C8C849484848C8C848C84848C8C8C8C84848C84848C7B7B84848484\
7B7B7B7B7B7B737373737B736B736B63736B5A6363000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000039314A42395242395A4A42634A4263524A6B524A635A526B5A526B63526B63526B6B5A73\
7B5A52735A73735A6B7B63737B636B7B63737B636B846B7384636B846B6B84636B
8C6B6B8C636B8C6B6B8C6363946B6B8C6B63946B638C635A8C6B5A8C63528C635284634A8C6352\
84634A84634A735239735242634A39000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039313942394A4A394A524A5A5A4A63\
635A6B63526B6B5A736B6373736B7B736B7B7B73847B7384847B8C7B7384847B8C
847B8C84848C847B8C84848C84848C8C848C84848C8C848C84848C84848C7B7B847B84847B7B7B\
7B8484737B7B737B736B736B6B736B5A635A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000039314A42395A42395A4A426B4A4263524A6B524A6B5A526B5A4A6B63526B7B5A52\
6B5A736B526B735A73735A6B7B63737B5A6B84637384636B846B7384636B84636B
84636B8C6B6B8463638C6B6B8C6B63946B638C63638C6B638C635A8C6B5A8463528C635284634A\
84634A7B5A427B5A427352396B5239000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000042394242394A524A5A5A4A6363526B\
63526B6B63736B637373637B73637B7B73847B7384847B8C7B738C847B8C847B8C
84848C847B8C8C849484848C8C848C84848C8C8C9484848C84848C84848484848C7B84847B8484\
7B7B7B7B847B6B737363736B5A635A313131000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000042395242395242425A4242634A4A6B4A4A63524A6B524A6B63526B63526B6B5A73\
6B526B735A73735A73736373735A6B8463737B637384637384636B8C6B7384636B
8C6B6B84636B8C6B6B8C6B6B8C6B6B8C6B63946B6B8C6B63946B638C635A8C6B5A8C63528C6352\
84634A8C634A8463427B5A426B52396B4A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003939424239424A4252524A5A63526B63526B\
6B5A736B5A7373637B73637B7B6B847B6B8484738C7B738C847B8C847B8C84848C
847B8C84848C84848C8C849484848C8C8C8C84848C8C8C8C8484848484847B84848484847B7B7B\
7B7B7B7373736B736B5A635A5A635A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000423952393952424263424263524A6B4A4263524A6B5A4A6B63527363526B\
6B5A736B5A6B735A73735A6B7B63737B5A738463737B636B84637384636B8C6B73
84636B8C6B6B8C636B8C6B6B8C63638C6B638C6B63946B638C635A8C6B5A8C635A8C6B5A846352\
8C635284634A84634A7B5A427352426B4A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003931423939424A4252524A5A5A52636352636B5A73\
6B5A737363736B63737B6B847B6B847B73847B7384847B8C847B8C847B8C847B8C
8C848C84848C8C848C84848C8C8C9484848C84848C84848484848C7B84847B84847B7B7B7B847B\
737B736B7373636B635A6B63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000039314A42395A42395A4A42634A4263524A6B524A6B5A526B5A4A6B6B5A73\
6B526B6B5A736B5A6B736373735A737B63737B636B846B73846373846B7384636B
8C6B738C6B6B8C6B6B8C636B946B6B8C6B63946B6B8C6B63946B638C6B5A8C6B5A8C63528C6B5A\
8463528C6352845A4A84634A735A42735239634A31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039314242394A4A42525A52635A5263635A6B635A6B\
6B63736B6373736B7B736B7B7B73847B7384847B8C847B8C847B8C847B8C847B8C
847B8C8C848C84848C8C848C84848C8C8C8C8484848484847B84848484847B7B7B7B7B7B737B7B\
737B73636B63636B63526352000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000039315242395A42395A4A42634A4263524A6B524A635A526B63526B\
6B5A7363526B6B5A736B5A6B736373735A6B7B63737B6373846B7384636B846B73
84636B8C6B7384636B8C6B6B8C6B6B946B6B8C6B63946B63946B63946B638C635A8C6B5A8C6352\
8C6B5284634A84634A84634A7B5A426B52396B4A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000039313942424A4A4252524A5A5A4A63635A6B635A6B6B6373\
6B6373736B7B736B7B7B73847B7384847B8C847B8C847B8C847B8C8C849484848C
8C848C84848C8C8C948C848C8C8C8C8484848C8C8C8484848484847B7B7B7B84847B7B7B737B73\
6B6B6B636B635A635A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003939523939524242634242637352424A42635A4A6B5A4A6B635273\
63526B6B5A736B5A73735A73735A737B63737B63738463738463738C6B73846373
8C6B7384636B8C6B738C6B6B8C6B6B8C6B63946B6B8C6B63946B638C6B63946B638C6B5A8C6B5A\
8C63528C6B5284634A8C634A845A427B5A426B5239634A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004A524242394242424A524A5A524A5A63526B635A6B6B63736B5A73\
73637B736B7B7B73847B6B8484738C84738C847B8C847B8C847B8C847B8C8C8494
84848C8C848C8C848C8C8C9484848484848C8484848484847B7B7B7B84847B7B7B7B7B7B6B736B\
636B635A635A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000039395242395A42426342395A4A4263524A6B5A526B5A4A6B\
6352736352736B5A736B5A6B735A73735A737B63737B63738463738463738C6B73
84636B8C6B738C6B6B8C6B738C636B8C6B6B8C6B6B946B6B8C6B63946B638C6B63946B638C635A\
8C6B5A8C63528C6B5284634A84634A7B5A42735A42634A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000039394242394A524252524A5A63526B63526B6B5A736B5A73736B7B\
736B7B7B6B847B6B84847B8C7B738C847B8C847B8C84848C847B8C84848C84848C
8C849484848C8C8C8C84848C8C8C8C8484848484847B7B7B8484847B7B7B7B7B7B6B736B6B736B\
636B635A635A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000039314A42395242395A4A42634A4263524A6B524A6B635273\
63526B6B5A736B526B735A73735A73736373735A737B63737B6373846B73846373
8C6B73846B738C6B738C6B6B8C6B738C6B6B946B6B8C6B6394736B946B63946B638C6B5A946B63\
8C6B5A8C6B5284634A8C6B5284634A84634A735242735242634A39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031313142314A4A42524A42525A526363526B6B5A73635A6B6B63736B637B\
736B84736B7B7B73847B738C847B8C7B7384847B8C847B8C8C848C847B8C84848C
84848C8C8C9484848C84848C8484848484847B7B7B7B84847B7B7B7B847B7373736B736B636B63\
5A635A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039314A39395A4239634A42634A4263524A6B5A4A6B\
63527363526B6B5A736B5A73735A73735A6B7363737B63738463737B6373846B73
846B738C6B7384636B8C6B738C6B6B8C6B6B8C6B6B946B6B8C6B6394736B946B63946B638C6B5A\
946B5A8C63528C6B5284634A8C63528463427B5A427352396B5239000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003931424A39524A42525A526363526B635A6B635A6B73637B73637B736B7B\
736B7B7B738C7B738C847B8C7B7B8C84848C847B8C84848C847B8C8C84948C848C
8C8C9484848C8C8C8C84848C84848C7B84848484847B847B7B847B737B73737B73636B63636B63\
525A52000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000423952423952423963423963524A6B524A6B5A526B\
5A526B6B5A736B5A736B5A736B5A737363737363737B63737B6373846B7B846373
8C6B738463738C6B738C6B738C6B738C6B6B9473738C6B6B946B6B946B6394736B946B63946B63\
8C6B5A946B5A8C6B528C6B5284634A8C6B528463427B5A426B5239634A39000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000042394A42394A52425A5A5263635A6B635A6B6B63736B6373736B7B736B7B\
7B73847B7384847B8C7B738C847B8C847B8C84848C847B8C8C848C84848C8C8C94
84848C8C848C84848C84848C7B84848484847B7B7B7B8484737B7B737B736B736B636B635A6352\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000042395239395A4242634A4263524A6B524A6B\
5A526B6352736B5A7363526B6B5A73735A737B6373735A737B63737B6373846B7B
8463738C6B738C6B738C6B738C6B6B8C6B738C6B6B94736B8C6B63946B6B946B6394736B8C6B5A\
946B638C6B5A946B5A8C63528C63528C634A8C634A7B5A42735242634A39000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000039313942314A52425A524A5A5A526B63526B6B63736B6373736B7B736B7B7B7384\
7B7384847B8C7B738C847B8C847B8C847B8C847B8C8C849484848C8C848C84848C
8C8C9484848C84848C8484848484847B847B7B847B737B7B7B7B7B6B736B636B635A635A000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039314A42395A4239634A42634A4263524A6B\
5A4A6B63527363526B6B5A736B5A73735A73735A737B63737B63737B63737B6373
8C6B7B846B738C6B73846B6B8C6B738C6B6B8C6B738C6B6B94736B946B6B94736B946B6394736B\
946B63946B5A8C6B52946B5A8C63528C6B5284634A84634A735A426B5239634A31
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003931424A39524A42525A52635A52636B5A736B637373637B6B637B736B847B6B84\
84738C7B7384847B8C847B8C847B8C847B8C84848C84848C8C849484848C8C848C
84848C84848C7B84848484847B847B848484737B7B7B7B7B6B736B6B736B5A635A5A6352000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039314A4239634239634A42634A4A6B\
5A526B5A4A6B6352736352736B5A736B5A6B735A73735A737B63737B6373846373
8463738C6B7B846B738C6B738C6B738C6B738C6B6B8C6B6B8C6B6B94736B8C6B6394736B63526B\
6B526B8C6B5A946B5A8C6B5A946B5A8C63528C634A42395A7B634A6B52396B5239
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003131314239524A4252524A5A5A4A636B5A736B5A736B637B6B637B7B6B847B6B847B738C\
7B738C847B8C847B8C847B8C847B8C8C849484848C8C848C84848C8C8C948C848C
8C848C84848484848C8484848484847B7B7B7B847B737B736B736B636B63636B5A000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039394239395A4242634A4263524A6B\
524A6B6352736352736B5A736B527373637B735A73736373735A7384637B846373
846B7B846B738C6B7B8C6B738C6B738C6B739473738C6B6B947373946B6B94736B946B6B94736B\
8C6B63947363946B5A946B5A8C6B52946B5A8C635284634A7B5A427B5A42735239
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000039314242394A52425A524A5A63526B635A736B637B6B6373736B84736B847B738C7B7384\
847B8C847B8C847B8C847B8C84848C84848C8C849484848C8C8C8C84848C8C8C8C
8484848484848484848484847B7B7B7B847B737B73737373636B6363635A000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039395239395A4A42634A4263\
524A6B5A4A6B6352736352736B5A736B5A73736373735A737B63737B6373846B7B
7B6373846B7B8C6B738C6B7B8C6B738C6B738C6B739473738C6B6B94737394736B94736B8C6B63\
947363946B639473638C6B5A946B5A8C6B52946B5284634A84634A7B5A427B5A42
313142000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000042314A4A39524A425A5A526363526B6B5A736B6373736B7B736B847B73847B738C847B8C\
7B7B8C847B8C847B8C84848C84848C84848C84848C8C8C948C848C8C848C848484
84848C8484848484847B7B7B7B84847B7B7B737B736B6B63636B635A6352000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039314A3939634239634A4A6B\
524A6B5A4A6B5A4A6B635A736352736B5A736B5A7373637B7363737B63737B6373
846B7B846B738C6B7B846B738C6B738C6B738C73738C6B6B94737394736B947373946B6B94736B\
946B639473638C6B5A947363946B5A946B5A8C6B528C6B5284634A7B6342735A39
735A39000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3931394239524A3952524A5A5A52636B5A73635A736B637B6B637B736B84736B847B738C7B7394\
7B7B847B7B8C847B8C847B8C8C848C84848C8C848C84848C8C8C8C848484848484
8484848484847B7B7B7B847B7B7B7B737B736B6B6B636B635A635A000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000393152423963424263\
524A6B524A6B5A526B5A526B6B5A736352736B5A736B5A73736373735A737B6373
7B6373846B7B846373846B73846B738C73738C6B6B947373946B73947373946B6B94736B946B6B\
94736B946B63947363946B5A9473638C6B52946B5A8C6B528C6B4A7B5A427B5A42
735239000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
42314A42394A524A5A5A4A63635A6B635A736B637B6B637B736B84736B847B738C7B738C847B94\
847B948C849C847B8C84848C84848C8C8C948C848C8C8C8C84848C8C8C8C848484
8484847B847B8484847B7B7B7B7B7B6B736B6B736B63635A000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000423952393963424263\
4A4263524A6B5A4A6B6352736352736B5A7B6B5A73735A73735A737B637B7B6373
846B7B8463738C6B7B846B738C6B7B8C6B7394737B8C6B739473739473739C737394736B94736B\
946B6394736B947363947363946B5A947363946B5A946B5A8C634A8C634A7B5A42
7B5A42735239000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000393139
39314A4A425252425A5A526B63526B6B63736B6373736B84736B7B7B738C7B738C847B947B7B94\
847B94847B948C8494847B8C8C848C84848C8C8C9484848C84848C848484848484
7B7B7B8484847B7B7B7B847B73736B6B736B636B5A5A635A000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000042395A39395A\
4A42634A426B5A4A6B5A4A6B6352736352736B5A7B6B5A73736373735A737B637B
7B6373846B7B846B738C6B7B846B738C6B738C6B739473738C6B739473739473739C737394736B\
94736B94736394736B946B63947363946B5A94735A8C6B528C6B5284634A84634A
7B5A39735A39000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000393142
4A39524A42525A4A635A52636B63736B637373637B6B637B7B73847B738C7B738C7B739484849C\
847B94848494847B8C8C849484848C8C8C8C84848C8C8C8C8484848484847B8484
8484847B7B7B7B847B737B73737373636B63636B5A000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039314A424263\
4242634A426B4A4A6B5A52735A5273635A73635273735A7B6B5A7373637B736373
846B7B846B7B846B7B846B738C737B8C6B738C73738C6B7394737B94737394737394736B9C7373\
94736B94736B946B6394736B946B63947363946B5A94735A8C6B528C6B52846342
846342735A39000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000042314A
4A3952524A5A524A63635A6B635A736B637B6B637B736B84736B847B738C7B738C847B94847B94\
848494847B8C84849484848C8C849484848C8C848C84848484848C7B7B84848484
7B7B7B7B847B737B737373736B6B63636B63525A52000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039394239395A\
424263424263524A6B524A6B5A52736352736B5A736B5A73735A7B6B5A737B637B
7B637B846B7B846373846B7B846B738C737B8C6B738C73738C6B73947373946B739C7373947373\
9C7373946B6B94736B947363947363946B5A94735A946B5A94735A8C634A84634A
7B5A427B5A42735231000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000039314242394A
4A425A524A5A63526B635A736B63736B6373736B84736B847B738C7B738C847B94847B94848494\
847B948C84948484948C849484848C8C8C948C848C8C8C8C8484848484847B8484
7B847B7B7B7B7B7B736B736B636B635A6352000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393952\
39395A4A426B4A426B524A6B5A4A6B635A736352736B5A736B5A7373637B73637B
7B637B7B637B846B7B846B7B8C6B7B8C6B7394737B8C737394737B9473739C737B9473739C7B73\
94736B9C7B7394736B94736B9473639C736394736394735A946B5A94735A8C634A
84634A7B5A427B5A42000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000039314A4A4252
4A425A5A4A635A526B6B63736B5A736B637B736B7B7B738C736B847B7394847B9484849C847B94\
8484948484948C849484848C8C848C84848C8C8C8C8484848484847B847B848484
7B7B7B7B7B736B736B6B73635A635A000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393142\
3939634242634A426B4A42635A4A6B5A5273635A736352736B5A736B5A737B637B
736373846B7B846B7B8C6B7B846B738C6B7B8C6B7B94737B8C6B739473739473739C7B7B947373\
9C737394736B9C736B947363947363947363947363946B5A94735A946B5A8C6B52
8463428463427B5A39735A39000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000031313942394A4A3952
524A635A5263635A73635A7373637B6B637B736B84736B847B738C7B7394847B94847B94848494\
847B9484849484848C8C849484848C8C8C8C84848484848C8484847B84847B7B7B
7B7B7B737B736B736B63635A5A635A000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
393952424263424263524A6B524A6B5A52735A52736B5A736B5A7373637B735A73
7B6B7B7B637B846B7B846B7B8C737B8C6B7B8C737B8C6B7394737B94737394737B9473739C7B7B\
9C73739C737394736B9C7B6B94736B947363946B639C736394735A94735A8C6B52
8C6B528463427B6342735A39000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039314A42314A4A425A
524A6363526B635A6B6B63736B637B736B84736B847B738C7B738C847B947B7B94847B94847B94\
848494847B8C84848C84848C8C8C8C84848484848C8484848484847B7B7B7B7B7B
737B73737B73636B635A635A525A4A000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39395239395A4242634A4263525273524A6B6352736352736B5A7B6B5A7373637B
73637B7B6B7B7B637B846B7B846B7B8C737B8C6B738C737B8C737394737B947373947373947373\
9C7B7394736B9C737394736B9C736B947363947363947363947363946B5A94735A
8C6B4A846B4A7B5A427B5A39000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039314A4A425A52425A
5A52635A526B6B63736B637B736B7B736B7B7B738C7B738C7B7B947B7B9484849C847B94848494\
847B8C8C849484848C8C849484848C8C8C8C8484848484847B7B7B8484847B7B7B
7B7B736B736B6B6B635A6352000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3931424239634242634A426B524A6B5A52735A5273635A73635A7373637B73637B
73637B7B637B846B84846B7B8C6B7B846B7B8C737B8C737B94737B8C7373947B7B94737B9C7B7B\
9C73739C7B7B9C73739C7B7394736B9C7B6B94736394736394735A9C736394735A
946B5284634A84634A7B5A427B5A39000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003131394239524A4252524A63
5A5263635A73635A736B637B6B637B736B84736B847B73947B738C847B94847B94848494847B8C\
84849484848C8C849484848C8C8C8C84848C84848C7B7B7B7B847B7B7B7B7B7B7B
7373736B736B5A635A5A6352000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003939526B5239424263524A6B524A6B6352735A5273635A736B5A7373637B
735A737B637B7B637B846B84846B7B8C6B7B8C6B7B94737B8C6B7B94737B94737B9C7B7B947373\
9C7B7B9C7B739C7B7394736B9C7B6B94736B9C7B6B94736394736394735A947363
8C6B528C6B5284634A846342735A39000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000042394A42395252425A524A63
635A6B635A736B637B6B637B736B84736B847B738C7B738C847B94847B94847B94847B948C8494\
84849484849484848C8C8C8C84848C84848C7B84848484847B7B7B7B7B7B737B73
737B73636B635A6352000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000039394A42395A4A42634A426B524A735A4A73635A73635A736B5A7B6B5A73
73637B73637B7B6B7B7B6B7B8C7384846B7B8C737B8C6B7B94737B94737B94737B9473739C7B7B\
9C73739C7B7B9C73739C7B739C736B9C7B6B9473639C7B6B94736394736394735A
9C735A8C6B528C6B4A7B63427B6342735A39000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000039313942314A4A425A4A425A5A5263
635A6B6B63736B5A73736B7B736B847B738C73738C7B7B947B7B94847B947B7B9484849484848C\
8C849484848C84848C84848C8C8C8C8484848484847B7B7B7B847B737B73737B73
6B736B636B5A525A4A000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000042395A42395A4A4A6B4A4A6B5A52735A5273635A7B6352736B5A7B
6B5A7B7B637B73637B7B6B7B846B7B8C7384846B7B8C737B8C737B94737B8C737394737B947373\
9C7B7B9C73739C7B739C73739C7B7394736B9C736B9473639C736B94735A947363
94735A94735A846B4A84634A7B5A427B5A39000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000039314A4239524A425A5A52635A526B
635A73635A73736B7B736B7B736B84736B84847B947B7B94847B94847B948C8494848494848494\
84848C8C8C9484848C8C8C8C848484848C8C84848484847B7B7B737B847B737B73
6B73635A6352525A4A000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000039395A4242634A4263524A6B524A735A52735A52736B5A7B6B5A7B
73637B73637B7B6B847B6B7B846B84846B7B8C73848C737B94737B8C737B947B8494737B9C7B7B\
94737B9C7B7B9C7B7B9C7B7B9C7B739C7B739C7B6B9C7B6B9473639C7B6B947363
9C736394735A94735A8C6B4A8463427B5A397B5A39000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000042394A423952524A5A5A4A63635A73
635A736B637B6B637B736B84736B847B738C7B738C847B947B7B94848494847B948C849484848C\
84849484848C8C8C8C8484848484848484848484847B7B7B7B7B7B737B7373736B
63635A5A6352000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000039395A4A42634A42635A52735A5273635273635A736B637B
6B5A7373637B73637B7B6B847B6B7B846B84846B7B8C73848C6B7B94737B94737B947B7B94737B\
9C7B7B9C7B7BA57B7B9C7B739C7B739C7B739C7B7394736B9C7B6B9473639C7B63
94735A94735A8C6B528C6B528463427B6342735A39000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000039314242314A4A425A524A635A526B635A6B
6B637B6B637B736B84736B847B738C7B738C7B7B947B7B94848494847B94848494847B8C8C8494\
84848C8C848C84848C8C8C8C8484848484847B7B7B7B847B7B7B73737B736B6B63
636B5A525A4A000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000042395A4242634A4A6B4A4A6B5A52735A5273635A7B635A73
73637B73637B73637B7B637B846B84846B848C7384846B7B8C73848C737B94737B94737B9C7B7B\
9C7B7B9C7B7B9C7B73A57B7B9C7B739C7B739C736B9C7B739C736B9C7B6B947363
9C7B6394735A94735A8C6B528C6B528463427B6342000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003931424239524A425A5A52635A526B635A73
6B6373736B7B6B637B736B8473738C7B7B947B7394847B94847B94848494847B8C84849484848C\
8C849484848C84848C8484848484847B7B7B7B7B7B7B7B737B7B736B736B636B63
5A6352000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000424263424263524A6B524A6B5A52735A5273635A7B
6B5A7B73637B73637B7B6B7B7B6B7B847384846B7B8C73848C737B9473848C737B94737B94737B\
9C7B7B9473739C7B7B9C7B7BA57B7B9C7B739C7B739C7B6B9C7B6B9473639C7B63
9473639C7B6394735A9473528C6B4A846B4A7B5A39735A39000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003131394239524A4252524A635A4A63635A73635A73
6B637B6B637B736B84736B847B738C7B738C84849C847B94848494847B8C8C849484848C8C8494\
84848C8C8C9484848C8484847B84848484847B7B7B7B7B7B737B7373736B63635A
5A6352000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000039395A4A42634A4A6B524A735A5273635A7B635A7B
6B5A7B6B5A7B7363847B637B7B6B847B6B7B8C73848C73848C73848C737B947B8494737B947B84\
94737B9C7B849C7B7BA57B7B9C7B7BA5847B9C7B739C7B739C7B6B9C7B6B9C7B6B
9C7B639473639C7B6394735A8C735284634A846B4A7B5A39000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000039314A4239524A425A524A5A5A526B635A736B637B
6B637B736B84736B847B738C7B738C847B94847B94848494847B8C84849484848C8C849484848C\
8C848C84848C8C8C8C8484848484847B7B7B7B847B737B73737B73636B6363635A
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000042395A42425A4A4A6B4A4A6B5A52735A5273635A7B
635A736B637B73637B7B6B847B637B846B84846B848C73848C6B7B8C73848C737B947B8494737B\
947B7B9C7B7B9C7B7B9C7B7BA57B7B9C7B73A5847B9C7B6B9C7B739C7B6B9C7B6B
9473639C7B6394735A9C7B638C6B528C6B5284634A7B6342735A39000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000039313939314A4A42524A425A5A526B63526B635A73635A73
736B84736B8473738473738C7B7B947B7B94847B947B7B9484849484849484849484848C8C8C94\
84848C84848C8484848484847B847B7B847B737B737B7B736B736B636B5A525A4A
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000423963424263524A6B524A735A52735A5273
6B5A7B6B5A7B73637B73637B7B6B847B6B84847384846B848C73848C73849473848C737B947B84\
947B7B9C7B7B9C7B7BA5847B9C7B7BA5847B9C7B73A5847B9C7B73A57B739C736B
9C7B6B9C73639C7B6394735A94735A8C6B52846B4A7B63427B6342000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000313142423952423952524A635A526B635A73635A736B637B
6B637B736B84736B847B738C7B7394847B947B7B94847B94847B94848494847B8C84848C84848C\
8C8C8C8484848484847B847B84847B7B7B737B7B7373736B6B73635A6352525A4A
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000039395A4A42634A426B525273524A73635273
635A736B637B6B5A7B73637B73637B7B6B847B6B7B8473848473848C73848C737B94738494737B\
9C7B8494737B9C7B7B9C7B7BA5847B9C7B73A5847B9C7B73A57B739C7B6B9C7B6B
9C7B6B9C7B6B9473639C736394735A947352846B4A84634A7B5A39735A39000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000039314A423952524A63524A63635A6B635A736B637B6B637B
736B84736B847B738C7B738C847B947B7B9484849484849484849484848C8C8C9484848C8C8C8C\
84848C8C8C8C8484848484847B7B7B7B847B7B7B7373736B636B5A636352000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000042395A4242634A426B524A6B5A52735A5273
635A7B635A7B73638473637B7B6B847B637B8473848473848C73848C7384947B848C7384947B84\
94737B9C7B849C7B7B9C7B849C7B7BA58484A5847BA5847B9C7B73A584739C7B6B
9C7B6B9C7B639C7B6B9C7B639C7B639473528C735284634A7B6342735A39000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000039314239314A4A425A524A5A5A526B5A526B6363736B637B736B84
6B6B8473738C7B738C847B947B7B94847B94847B94848494847B8C84849484848C8C8C9484848C\
84848C8484848484847B7B7B7B7B7B7B7B737B7B736B6B6363635A525A4A000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000424263424263524A6B524A735A5A7B
5A5273635A7B6B5A7B73638473637B7B6B847B6B84847384846B848C73848C7384947B848C737B\
947B84947B849C7B849C7B7BA5847BA57B7BAD847BA57B73A584739C7B73A58473
9C7B6B9C7B6B9C7B639C7B6394735A94735A8C6B528C6B4A7B63427B6342000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003931424239524A425A524A635A526B635A736363736B637B6B637B
73738C73738C7B738C7B738C847B94847B94848494847B8C84849484848C84848C84848C8C8C8C\
8484848484847B7B7B84847B7B7B737B7B7373736B6B73635A6352000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000039395A4A42634A426B525273525273
635A7B635A7B6B5A7B6B5A7B736B8473637B7B6B847B6B848C738C8C73848C73848C7384947B84\
947B849C7B849C7B7BA58484A57B7BA5847BA57B7BAD847BA58473A584739C7B6B
A584739C7B6B9C7B6B9C73639C7B6394735A94735A846B4A846B4A7B6342735A39000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000039314A423952524A63524A63635A73635A736B637B6B637B736B84
736B847B738C7B738C847B947B7B948484947B7B8C84849484848C8C849484848C84848C848484\
84848C7B7B7B7B847B7B7B7B7B7B7B73736B6B736B63635A5A6352000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000042395A4A4A6B4A4A6B5A5273
5A5273635A7B635A736B637B6B637B7B6B8473637B7B6B84846B848C738C8C7384947B84947384\
9C7B84947B7B9C7B849C7B7BA58484A57B7BAD847BAD847BAD847B9C7B73A57B73
9C7B6BA57B739C7B639C7B639C7B639C7B639473528C7352846B4A846342735A39000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004239524A425A52425A5A526B635A736B637B6B637B736B84736B84
7B738C7B738C847B94847B94847B94847B8C84849484848C84849484848C8C8C9484848C848C8C\
8484848484847B847B7B847B737B737B7B736B6B6363635A525A4A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000423963424263524A6B524A73
5A527B5A52736B637B6B637B73638473637B7B6B847B6B84847384847384947B8C947B84947B84\
947B849C848C9C7B84A58484A57B7BAD8484AD847BAD8484AD847BAD847BA58473
A584739C7B6BA5846B9C7B6B9C7B63947B5A9C7B638C73528C6B4A7B63427B6342000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003131394239524A395A524A635A526B636373635A736B637B6B6B8473738C
736B847B73947B7B948484947B7B8C84849484848C84849484848C84848C84848C8C8C8C848484\
8484847B847B84847B7B7B737B7B736B736B6B6B635A5A4A000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000039315A4242634A4263525273
525273635A7B635A7B6B637B6B5A7B7363847363847B6B847B6B848473848C7384947B8C947384\
9C7B849C7B84A584849C7B84A58484A58484AD8484AD847BAD847BAD847BAD847B
9C7B73A584739C7B6B9C846B947B639C7B63947B5A947B5A8C6B4A846B4A7B63427B6339000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000393142423952524A63524A635A526B635A736B637B6B637B736B84736B84
7B738C7B738C847B947B7B94848494847B9484849484848C8C849484848C8C8C8C84848484848C\
8484847B847B7B7B737B7B7B737B736B73635A63525A5A52000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000042395A4A42634A4A6B
5A52735A527B635A7B635A7B73638473637B736B847B6B848473848473848C738C8C7384947B8C\
9C7B849C7B849C7B84A58484A58484AD8484A5847BB58C84AD847BAD847BA58473
A5847B9C7B73A584739C7B6B9C846B9C7B639C7B6394735A947352846B4A846342735A39000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003931424239524A425A5A526B5A526B6B63736B637B736B846B6B8473738C
7B738C847B947B7B94847B94847B94848494847B8C84848C84848C8C8C8C848484848484848484\
8484847B7B7B7B7B7B737B73737B73636B5A5A6352000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000042395A4239634A4A6B
524A6B5A527B5A5273635A7B6B5A7B73638473637B7B6B847B6B8484738C8473848C7B8C947B84\
9C7B8C9C7B849C8484A57B84A58484A5847BAD8484AD847BB58C84AD847BAD847B
A58473A584739C7B6B9C7B6B9C7B6B9C7B6B947B639C7B638C73528C73527B63427B6342000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004239524A425A524A63525263635A736B637B6B637B6B6B8473738C73738C
7B7B947B7B94848494847B94848494847B8C8C849484848C8C8C8C84848C8C8C8C848484848484\
7B7B7B84847B7B7B737B7B736B7363636B5A5A5A4A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039315A4A42634A4263
524A6B525273635A7B635A7B6B63846B637B736B847B6B847B6B84846B848C7B8C8C7B8C947B8C\
947B849C848CA58484A5848CA58484AD8C8CAD8484B58C84AD847BB58C84AD847B
AD847BA58473A584739C846B9C846B9C7B639C846B9C7B63947B5A8C6B528C6B4A7B63427B6339\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3931394239524A425A4A4A5A5A526B635A736B637B6B637B736B84736B847B7394
7B738C847B947B7B948484947B7B8C84849484848C8C8C9484848C8C8C8C8484848484847B7B7B\
7B847B7B7B7B7B7B7B73736B6B6B635A6352000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000042395A4A426B
4A426B5252735A5273635A84635A7B6B63847363847B6B847B6B848473848473848C7B8C8C7384\
947B8C9C7B84A5848CA57B84AD8484AD8484AD8C84AD8484B58C84B58C84B58C84
AD847BAD847BA58473A584739C7B6B9C846B9C7B639C846394735A947352846B4A846B4A735A39\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3931394239524A425A52526B5A526B636373636373736B84736B8473738C73738C
7B7B947B7B9484849C847B9484849484848C84848C84848C8C8C8C84848C8484847B847B848484\
7B7B7B7B7B7B737B7373736B636B5A5A6352000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000039395A423963
524A6B524A6B5A527B635A7B6B63846B637B7363847363847B738C7B6B8484738C847384947B8C\
947B8C9C7B8C9C7B84A5848CA58484AD8484AD8484B58C84B58C84B58C84B58C7B
B58C84AD847BAD8473A5846BA584739C7B6B9C846B9C7B639C846394735A8C73528463427B6342\
735A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39394A4239524A425A524A63635A73635A736B637B6B637B737384736B847B738C
7B7B9484849C7B7B9484849484849484849484848C84848C84848C848C8C7B84848484847B7B7B\
7B847B737B73737B736B6B63636B5A52524A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000423963
4A4263524A73524A73635A7B635A7B6B63846B637B736B84736B847B738C7B6B8484738C8C738C\
947B8C947B849C848CA58484AD848CA58484AD8484AD8484B58C84B5847BBD8C84
B58C7BB58C7BA58473A584739C846BA584739C7B639C7B639C7B63947B5A8C6B52846B4A7B6342\
7B6339000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39314A4A425A4A4A635A526B5A5A6B6B637B6B637B736B84736B847B738C7B738C
7B7B947B7B9484849C84849484849484848C8C8C9484848C8C8C8C6B63738484847B847B7B847B\
7B7B737B7B7B73736B6B6B635A5A52000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000423963
4A426B4A4A6B5A52735A5A7B635A84635A7B736B84736B847B6B847B6B8484738C84738C8C7B8C\
8C7B8C9C848C9C848CA5848CA58484AD8C8CAD8C84B58C8CB58C84BD948CBD8C84
BD8C84B58C7BB58C7BAD8473A584739C846BA5846B9C846B9C8463947B5A947B5A8C6B4A846B4A\
7B6339000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000313131\
42395242425A524A63525263635A7363637B6B6B846B637B73738C73738C7B7B94
7B7B9484849C84849484849C847B9484849484848C8C8C9484848C84848C8484848484847B7B73\
7B7B73737B734A425A636352525A4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000423963
4239634A4A6B524A735A5A7B5A5A7B6B63846B6384736B847363847B6B847B7384847B8C84738C\
8C7B8C947B8C9C848C9C7B84A5848CAD848CB58C8CAD8484B58C84B58C84BD948C
BD8C84B58C84B58C7BB58C7BA58473A584739C846B9C846B9C7B639C8463947B5A8C7352846B42\
7B6342735A39000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393142\
4239524A425A524A635A5A73635A736B637B6B637B736B84736B847B738C7B738C
84849C847B9C84849C847B9484849484849484849484848C8C8C8C8484848484847B7B7B7B847B\
737B73737B736B6B63636B5A525242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4242634A426B524A73525273635A7B635A7B6B63846B6384736B8C7B6B847B738C8473848C7B8C\
8C7B8C947B8C947B8CA5848CA5848CAD848CAD8484B58C8CB58C84BD8C84BD8C84
C6948CBD8C84B58C7BAD8473AD8C7BA58473A584739C7B6BA5846B9C7B63947B638C73528C7352\
7B6B427B6339000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393142\
4239524A425A524A635A526B6B637B6363736B6B7B6B6B8473738C73738C7B7B94
7B7B9484849C7B7B9484849484849484849484848C84848C8484848484847B7B7B7B7B7B7B7B73\
7B7B7373736B6B6B635A6352525242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
39395A4A426B4A4A6B5252735A527B636384635A7B6B6384736B847B738C7B6B8484738C84738C\
8C7B8C8C7B8C947B8C9C848CA5848CA58484AD8C8CB58C84BD8C8CB58C84BD8C84
BD8C84C69484B58C7BB58C7BAD8C73AD8C73A5846BA5846B9C846B9C846B947B5A947B5A8C6B4A\
846B4A7B6339000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000039394A\
423952524A63525263635A73635A736B6B846B6B84736B84736B8C7B7B947B7B94
84849C847B9484849C84849484849484848C8C8C9484848C8C8C8C8484848484847B7B7B7B7B7B\
737373737B73636B5A5A6352000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000423963524A735252735A5A7B635A7B6B63846B6384736B84736B847B738C84738C847B8C\
847B8C948494947B8C9C848C9C848CAD8C8CAD8C8CB58C8CB58C84BD948CBD8C8C
C6948CBD8C84BD9484B58C7BB58C7BAD8473AD8C73A5846BA5846B9C84639C846B947B5A8C7352\
846B4A846B42736339000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000039314A\
4A425A4A4A635A526B5A5A6B6B637B6B637B736B846B6B8473738C7B738C847B94
7B7B9484849C84849484849484848C84849484848C8C8C8C8484848484847B847B7B847B737B73\
737B7373736B636B5A525A4A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004242634A426B525273525273635A7B635A7B6B63846B6384736B847B6B8484738C84738C\
8C7B8C8C7B8C948494947B8C9C848CA5848CAD8C8CAD848CB58C8CBD8C8CC6948C
BD8C84C6948CBD8C84BD9484B58C7BB58C7BAD8C73AD8C739C846B9C846B9C84639C84638C7352\
8C7352846B427B6342000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000313139423952\
4A425A524A6352526B63637363637B6B6B7B6B6B7B73738C73738C7B7B947B7B94
84849C84849484849484849484849484848C84848C8484848C8C8C84848484847B7B7B737B7B73\
73736B73736B5A6352525A4A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003939634A426B4A4A6B5A527B5A5A7B63638463637B736B84736B847B738C7B6B84847B8C\
847B8C8C7B8C8C7B8C9C84949C848CA5848CA5848CAD8C8CB58C8CBD8C8CBD8C84
C6948CC69484C69484BD8C7BBD9484B58C7BB58C7BA58473AD8C73A5846B9C846B947B63947B5A\
8C7352846B4A7B6342000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000393142423952\
4A425A4A4A635A526B635A736B637B6B637B736B84736B847B738C73738C7B7B9C
7B7B9484849C7B7B9484849484848C84849484848C84848C8484848484847B7B7B7B7B7B737B73\
737B736B6B5A5A63524A5242000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004239634A4A6B524A735A5A7B5A5A7B6B63846B6384736B8C736B847B738C7B738C\
847B8C84738C8C7B94947B8C9C84949C848CA5848CAD8C8CB58C8CB58C84BD948C
BD948CC6948CC69484C69484BD8C7BBD947BAD8C73AD8C73A58473A58C739C84639C846B947B5A\
947B5A846B4A846B42736339000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039314A42425A\
4A425A5A526B5A5A6B63637B63637B736B84736B8473738C73738C847B94847B9C
84849C847B9484849C84849484849484848C8C8C8C84848C8484847B847B84847B7B7B737B7B73\
73736B6B6B635A5A4A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004242634A426B5252735252736363846363846B63846B63847B738C7B738C84738C\
84738C8C7B948C7B8C94849494848CA58C94A5848CAD8C8CAD8C8CB59494BD8C8C
C6948CC6948CCE9C8CC69484C69484BD8C7BBD947BB58C73AD8C73A5846BA58C739C846B9C8463\
947B5A8C7B5A846B4A7B6342000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000423952423952\
4A4A6352526B635A73635A736B637B6B6B8473738C736B847B73947B7B9484849C
7B7B9484849484849484849484848C84848C84848C8C8C8C8484848484847B7B7B7B7B7B73736B\
73736B63635A5A5A4A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000039395A4A426B4A4A6B5A52735A5A7B6363846363846B6B84736B847B738C7B738C\
84738C847B8C8C84948C7B8C9484949C848CA58C94A5848CAD8C8CB58C8CBD948C
BD8C8CC6948CC6948CCE9C8CC69484BD9484BD8C7BB5947BAD8C73B58C73AD846BA5846B9C7B63\
9C84638C7352847352736342735A39000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000031313939314A4A425A\
524A635A526B5A5A736B637B6B637B736B846B6B847B738C7B73947B7B947B7B94
84849C84849484849484848C84849484848C84848C8484848484847B7B7B7B7B7B737B737B7B73\
6B736363635A525242000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000423963524A735252735A5A7B635A7B6B63846B6384736B8C736B847B738C\
7B738C847B8C847B8C9484949484949C84949C848CAD8C94AD8C8CB58C8CB58C8C
BD948CC6948CCE948CCE9484CE9C8CCE9484C69484BD8C7BBD947BB58C73AD8C73A5846BA5846B\
9C7B63947B5A846B4A7B6B4A735A39000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003131394239524A425A\
524A6352526B635A7363637B6B6B846B6B7B73738C73738C7B7B947B7B94847B9C
7B7B948484947B7B8C84849484848C84848C8484848484847B847B84847B737B737B7B7373736B\
6B73635A5A4A525242000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004239634A426B525273525273635A7B6363846B6B846B6384736B8C7B6B8C\
84738C84738C8C7B948C7B9494849494848C9C8494A5848CAD8C94AD8C8CB58C8C
BD948CC6948CBD8C84C6948CC69484CE9C84C6947BC6947BBD8C73B58C73AD846BA5846B9C8463\
9C84638C7B5A8C73527B6B4A736342000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039394A4239524A4263\
524A63635A736363736B637B6B637B73738C73738C7B73947B739484849C848494
84849484849484849484848C84849484848C8C8C8C8484848484847B7B7B7B847B737B73737B6B\
636B5A636352000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004239634A426B4A4A6B5A527B5A5A7B636384636384736B8C736B8C7B738C\
7B738C847B94847B8C8C84948C84949C8C949C8494A58C94A58C8CB59494AD8C8C
B59494B58C8CBD948CBD948CC6948CC69484CE9C84C6947BBD947BB58C73B58C73AD8C6BA5846B\
9C84639C84638C7352846B4A736342736339000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039314A42425A4A425A\
52526B5A5A736B637B63637B6B6B84736B8473738C73738C7B7B947B7B9484849C
7B7B9484849484849484849484848C84848C8484848484847B7B7B7B7B7B7B7B737B7B736B736B\
636B5A5A5A4A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004239634A4A6B524A735A5A7B5A5A7B6B63846B6384736B8C736B84\
7B738C7B738C8C7B94847B8C9484949484949C8C949484949C8C94A58C94AD9494
AD8C8CB5948CB5948CBD948CBD9484C69484C69484C69484BD8C7BBD8C7BAD8C73AD8C73A58463\
9C8463948463947B5A846B4A7B6B426B5A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000313131393952423952524A63\
52526B635A73635A736B6B846B6B8473738473738C7B7B947B7B9484849C7B7B94
84849C84849484849484848C8C8C9484848C84848C7B847B8484847B7B7B7B7B7373736B73736B\
636352525A4A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004242634A426B5252735252736363846363846B6B846B6B847B738C\
7B738C84738C84738C8C7B948C7B949484949484949C8C9C9C8C94A58C94A58C8C
AD9494AD948CB5948CB5948CC69C8CC69484C69C84C6947BC6947BBD8C73B58C73AD846BAD8C6B\
9C84639C84638C7B5A8C73527B6B4A736342000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039313939314A4A425A4A4A63\
5A526B5A5A7363637B6B637B736B846B6B8473738C7B738C7B7B9C7B7B94848494
847B94848494847B8C84848C84848C848C8C8484848484847B7B7B7B7B7B73736B737B736B6B63\
6363524A5242000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000039395A4A426B4A4A6B5252735A5A7B6363846363846B6B84736B84\
7B738C7B738C847B94847B948C84948C7B949484949484949C8C9C9C8C94A58C94
A58C8CAD9494AD8C8CB5948CBD948CC69C8CBD9484C69C84BD947BBD947BB58C73AD8C73A5846B\
A58C6B947B63947B638C7352846B4A736339000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000031314242425A4A425A524A63\
5A526B63637B63637B6B6B846B6B8473738C73738C7B7B947B7B9484849C848494
84849484848C8C8C9484848C848C8C848484848C8C84848484847B7B7B737B7B7373736B6B6B63\
5A5A4A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004239634A4A735252735A5A7B635A846B6B8C6B6B8C736B8C\
736B8C847B94847B8C847B94847B9494849C948494948C9C948C949C949C9C8C94
A59494A58C8CB59494B5948CBD9C8CBD9484C69C8CC69C84C69484BD8C7BBD947BAD8C73AD8C73\
A5846B9C8C6B9484638C7B5A846B4A7B6B4A6B5A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039314A4239524A4A63524A63\
5A5A73635A736B637B6B637B736B8473738C7B7B947B738C847B9C7B7B94848494
7B7B8C84849484848C848C9484848C84848C8484848484847B7B737B7B73737B73737B6B63635A\
5A5A4A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004239634A426B5252735252735A5A7B636384736B8C6B6B8C\
73738C7B738C847B9484738C8C7B948C8494948C9C948494948C9C948C949C9494
9C8C8CA59494AD948CB59C94B5948CBD9C8CBD9C84C69C8CBD947BBD947BB58C73B59473A58C6B\
A58C6B9C84639C84638C7B5A8473527B6B42736342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039315242425A4A425A5A526B\
5A5A7363637B63637B736B846B6B8473738C73738C7B7B947B7B948484947B7B94
84849484849484849484848C8C8C8C8484848484847B7B7B7B847B7B7B73737B736B7363636B5A\
525A4A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000052423142426B4A4A6B52527B5A5A7B6363846B638473738C\
736B8C7B738C7B738C847B94847B948C84948C8494948C9C948C9C9C8C9C948C94
A59494A59494AD9494AD948CB59C8CB59C8CBD9C8CBD9C84C69C84BD947BBD947BAD8C73AD8C73\
A58C6BA5846B9484639484638473527B6B4A736339000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000313139423952423952524A6352526B\
635A73635A736B637B6B6B8473738C736B847B73947B7B9484849C7B7B94848494
84849484849484848C84848C84848C84848C7B7B7B7B847B7B7B737B7B7373736B6B73635A6352\
525A42000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003939634A4A734A4A735A5A7B5A5A7B6B63846B6B8C\
73738C736B8C7B73947B7394847B94847B948C84949484949C8C9C948C949C9494
9C9494A59494A5948CAD948CAD948CB59C8CB59C84BD9C8CBD9C84BD9C84B58C73B59473AD8C73\
AD8C739C84639C84639484638C7B5A7B6B4A7363426B5A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003931423939524A4263524A635A526B\
5A5A736B637B6B637B736B84736B847B7B947B7B947B7B947B7B9484849C848494
84849484848C8C8C9484848C848C8C8484848484847B7B7B7B7B7B737B737B7B736B6B63636352\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000042426B42426B525273525273636384636384736B8C\
736B8C7B73947B738C847B94847B948C849C8C8494948C9C948C9C9C949C948C94
9C94949C9494A59C94A5948CAD9C94AD9C8CB59C8CBD9C8CBD9C8CB59484BD9C84B5947BB59473\
A58C6BA58C6B9C84639484638C7B5A8473527B6B42736339000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031314242395A4A425A52526B5A526B\
63637B63637B6B6B846B6B8473738C73738C7B7B947B7B948484947B7B94848494
7B848C84849484848C8C8C8C8484848484847B847B84847B737B737B7B7373736B6B6B635A5A4A\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039395242426B4A4A6B52527B5A5A7B6363846B6384\
736B8C73738C7B7B947B738C847B94847B948C849C8C8494948C9C948C9C9C949C
948C949C94949C9494A59C94A5948CAD9C8CAD9C8CBDA58CB59C84BD9C84B5947BB5947BAD8C73\
AD8C73A58C6B9C8C6B948463947B5A8473527B6B4A6B5A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039395242395A4A4A63524A63635A73\
63637B6B637B6B637B73738C73738C7B738C7B739484849C7B7B948484947B7B8C
84849484848C84848C84848C848C8C84848484847B7B7B737B7B7B737B6B73736B6363525A5A4A\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000393131313131393939\
393131393131393131423939423131000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039393939426B4A4A735252735A5A7B5A5A7B6B6B8C\
6B6B8C73738C73738C7B7B947B7B94848494847B94948C9C948C9C948C9C948C94
9C949C9C94949C9C949C948CA59C94AD9C8CAD9C8CB59C84BDA58CB59C84B59C84AD947BB5947B\
AD8C73A58C739C846B9C8C6B9484638C7B527B6B4A736B42000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003131313931524A425A4A425A52526B5A5A73\
63637B63637B6B6B846B6B8473738C73738C7B7B947B7B948484947B7B94848494
84848C84849484848C84848C8484848484847B7B7B7B7B7B737B737B7B736B6B63636352525242\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000313131393131393131423939393131423131
4231314239394231314A39314A39314A39394A31314A39314A39315239394A3131523931523931\
5239395231315A39315239315A39395231315A39315239315A3939523131523931
5231315239394A3131523131423131393131000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039426B42426B525273525273635A84636384\
736B8C6B6B8C7373947B7394847B947B7B948C84948C8494948C9C948C9C948C9C
948C949C949C9494949C9C949C9C8CAD9C94A59C8CB59C8CB59C8CBDA58CAD9C7BB59C7BAD9473\
AD9473A58C6BA58C6B9C84639484638473528473527363426B6339000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031313942395242425A524A6B5A526B636373\
63637B6B6B846B6B8473738C73738C7B7B947B7B948484947B7B94848494848494
84849484848C8C8C9484848C8484847B847B8484847B7B737B7B7373736B6B73635A6352000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000423939423139423939\
4231314A39394A39315239394A39315A42395239315A42395A39316342395A3931
6342395A39316342395A39396342395A39316342395A42396342395A39316342395A4239634239\
5A39396342395A42396342395A39396342425A42396342395A39396342395A4239
5A42395A39395A42395239395A39395239315239395239315239394A31315239394A31314A3931\
4A31314A3939423131000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039395A424A6B4A4A735A5A7B5A5A8463638C\
6B6B8C73739473738C7B7B947B7B9484849C8484948C849C8C849C948CA5948C9C
9C949C9494949C9C9C9C9C94A59C94A59C8CADA594AD9C8CB5A58CB59C84B5A584AD9C7BAD947B\
AD8C73AD9473A58C6B9C8C6B9484639484638473527B6B4A6B5A39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003131313931314231393931314A39394A39315239394A31314A39314A31314A39394A3131\
4A31314A31314A31314231314A3131423131000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003931393939524A4263524A635A5A735A5A73\
63637B6B637B736B846B6B8473738C7B7B8C84849C7B7B94848494848494848494
84848C84848C84848C8C8C8C7B84848484847B7B7B7B7B7B73736B737B6B636B5A5A6352000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003931314239394231314A39394239314A39394A3939524239523931\
5A42395A39316342395A3931634239634239634239634231634239634239634A39
634239634239634239634A42634239634239634239634A42634239634A42634239634A42634239\
634A42634242634A42634239634A42634239634A42634239634A42634A39634A42
634239634A395A4239634A395A42395A42395A42395A42395239315A42395242395A4239523931\
5239394A39315239394A31314A3931423131000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000042394A39426B4A4A735252735A5A845A5A84\
6B6B8C6B6B8C73739473738C7B7B947B7B948C849C8484948C8C9C948C9C94949C
948C949C949C9494949C9C949C9C8CA59C94A59C8CADA594AD9C8CB5A58CAD9C84B59C84AD947B\
AD947BA58C73A58C739C84639484638C845A8C7B5A7B6B42736342000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000393131393131
4239394239394A39394A39315242395239395242395239395A42425A42395A42425242395A4242\
5242395242395239395242425239395239394A3939524239423931423131000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031313942425A4A426352526B5A526B63637B\
63637B6B6B846B6B8473738C73738C7B7B947B7B9484849C7B7B948484947B848C
84849484848C84848C8484848484847B847B7B7B7B737B737B7B7373736B636B5A525A42000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000393139393131
4A39394239394A39394A39315242395239395242395239315A42395A42396342396342316B4A39\
6342396B4A396342396B4A426B4A396B4A396B4A396B4A426B4A396B4A426B4A39
6B52426B4A426B4A426B4A426B52426B4A426B52426B4A426B524A6B4A426B52426B4A426B524A\
6B4A426B52426B4A426B52426B52426B52426B4A426B52426B52426B5242634A39
6B52426B52426B5242634A396B5242634A39634A425A4A39634A425A4A395A4A395A42315A4A39\
5A42395A42395239315A42395239315239314A3931423931393131000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000042426B4A4A6B52527B52527B636384\
6B6B8C736B8C6B6B8C7B7B947B7B94847B94847B948C849C8C849C948C9C948C9C
9C949C94949C9C949C9494949C9C949C9C94A5A594A59C8CADA58CADA58CB5A58CAD9C7BAD9C84\
AD947BA59473A58C6BA58C73948463948463847352847352736342000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003931313931314239394231314A3939
4A39395A42425239395A42395242395A42425242395A42425A4242634A425A4A42634A4A5A4A42\
634A4A5A4A425A4A425A4A425A4A425A42425A4A425A4242424A525A4239524239
4A39394A3939392931000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003931424239524A4A63524A635A5A7363637B\
6B6B846B637B736B8473738C7B7B9473738C7B7B947B7B948484947B7B8C848494
84848C84848C8484848484847B848484847B737B737B7B7373736B73736B5A6352525A42000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003931314239394231314A39394A39395239394A3931524239
5239395A42395A39396342396342396B42396342396B42396342396B4A426342396B4A426B4A39\
6B4A426B4A396B4A426B4A427352426B4A427352426B524273524A6B524273524A
6B524273524A6B524273524A6B524273524A6B524273524A6B524A735A4A6B524273524A6B5242\
735A4A6B524273524A6B5242735A4A6B52427352426B5242735A4A6B5242735A42
6B5242735A426B52426B52426B52426B52426B52396B5239634A396B5239634A39634A39634A39\
634A395A42315A4A395A42315A4A39524231425252524231524231423131423131
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003939634A4A734A4A73525A7B5A5A7B\
6B6B8C6B6B8C73738C73738C7B7B947B7B94847B948484948C8C9C8C849C948C9C
948C9C9C949C9494949C9C949C9C949CA5949C9C8CA5A58CA5A58CADA58CAD9C84AD9C84AD9C7B\
AD9C7BA58C73A58C739C8C6B9C8C6B8C84638C7B5A7B734A7B6B4A6B5A39000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002931313931314A39394A39395239395239395A42425A4242
5A42425A4242634A425A42425A4A425A4242634A4A634A42634A4A5A4A4263524A63524A63524A\
63524A6B525263524A63524A63524A6B524A63524A63524A63524A6B524A634A42
634A425A4A425A4A425242394A3939423131423139000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039314242425A4A425A5A526B5A5A7363637B\
63637B736B84736B8473738C73738C8484947B7B948484947B7B8C84849484848C
84848C84848C8C8C8C8484848484847B7B7B7B847B7B7B73737B736B6B5A636352000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000423939423939423939\
4239395242394A39395242395239395A42425A42396342425A42396B4A42634239
6B4A426342396B4A426B4A426B4A426B4A397352426B4A426B52426B4A4273524A6B5242735242\
6B524273524A73524273524A7352427B5A4A735A4A735A4A735A4A7B5A4A735A4A
7B5A4A735A4A7B6352735A4A7B5A52735A4A7B6352735A4A7B5A4A735A4A7B6352735A4A7B634A\
735A4A7B634A735A4A7B634A735A4A7B634A735A4A735A4A735A427B634A735A4A
735A4A735A4273634A735A42735A426B5242735A426B52426B5A426B52396B5A426B52396B5242\
634A396B5242634A39635239634A31635239634A31634A395A42315A4239524231
4A3931423131423931000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000042395242426B4A527352527B5A6384\
63638473739473738C737394737394847B9C847B9484849C84849C948CA5948C9C
94949C948C9C9C9C9C9C94949C9C949C9C94A5A594A5A58CA5A594ADA58CB5AD8CADA584AD9C84\
A5947BAD9C7BA59473A58C739C8C6B9C8C6B8C845A847B52736B42736342000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004231314231314A39394A39395242425239395A42425A4242634A42
5A4242634A425A4A42634A4A5A4A42634A4A634A4A63524A634A4A63524A63524A6B5A5263524A\
6B5A5263524A6B5A5263524A6B525263524A6B5A5263524A63524A63524A6B5252
63524A63524A634A4263524A5A4A425A4A425A4242524242423931313131000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039394A4239524A4A6352526B635A7363637B\
6B6B846B6B8473738C73738C7B7B947B7B948484947B7B9484849484848C848494
84848C84848C84848C8484847B7B7B7B847B7B7B737B7B7373736B6B6B5A5A5A4A000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004231394A39394A39395239395239395A4242523939\
5A42395A42396342426342396B4A426B42396B4A426B42396B4A426B4A42734A42
6B4A426B4A426B4A4273524A6B4A4273524A73524273524A73524273524A73524A7B5A4A73524A\
735A4A735A4A7B5A52735A4A7B5A4A7B5A4A7B6352735A4A7B63527B5A527B6352
7B5A4A7B63527B63528463527B634A7B63527B63527B63527B5A4A7B63527B634A7B6352735A4A\
7B63527B634A7B6352735A4A7B634A735A4A7B6352735A4A73634A735A4A7B634A
735A4A73634A735A4A73634A6B5A42735A4A6B5A42735A4A6B52426B5A426B52426B5A42635239\
6B52426352396B5242634A39635239634A396352395A4A31634A395A4A315A4A39
5242315242314A39314A3931393131000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000042426B4A4A7352527B525A7B\
6363846B6B8C73739473738C7B73947B7B9484849C847B948C849C8C849C948CA5
948C9C94949C94949C9C9C9C9494949C9C949C9C94A5A5949CA58CADA58CADA58CADA58CA59C7B\
A59C7BA5947BA5947B9C8C6B9C8C6B9484639484638473527B734A736342000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3129314231314231314A39394A39395242425239395A4A425A42425A4A425A4A42
63524A634A4A63524A634A4A6B524A63524A6B524A63524A6B5A526B52526B5A5263524A6B5A52\
6B5A526B5A526B524A6B5A526B5A526B5A5263524A6B5A526B5A526B5A526B5A52
6B5A526B52526B525263524A63524A63524A63524A5A4A42634A4A5A4A425A4239523931000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031314242425A4A4A6352526B5A5A736B637B\
6B637B6B6B846B6B847373947373947B7B947373848484947B84948484947B848C
848C8C84848C84848C7B84848484847B7B7B7B7B7373736B73736B63635A5A5A4A000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004239394231394A39394A39395242395239395A42425A42396342426342396B4A42\
6B4A426B4A426B4A427352426B4A427352426B4A4273524A73524273524A6B5242
73524A73524A73524A73524A7B5A5273524A7B5A4A73524A7B5A527B5A4A7B5A52735A4A7B6352\
7B5A527B63527B5A4A8463527B63527B63527B6352846B5A7B6352846B527B6352
846B5A7B6352846B527B6352846B5A7B63527B6B527B6352846B527B63527B63527B63527B6B52\
7B63527B635273634A7B6B527B63527B635273634A7B63527363527B635273634A
7B635273634A73635273634A73635273634A73634A6B5A4273634A6B5A426B5A4A6B5A426B5A4A\
6B5A426B5A426352396B5A426352396352426352396352426352396352395A4A31
5A4A395A4231524231523931524231423931000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039396B4A4A734A4A735A5A84\
5A63846B6B8C6B6B8C7373947373947B7B947B7B9484849C84849C8C8C9C8C8C9C
9494A594949C94949C9494949C9C9C9C9C949CA5949C9C8CA5A594A5A58CADAD8CA5A584ADA584\
A59C7BA59C7B9C9473A594739C8C6B9C8C6B8C84638C7B5A7B7352736B425A5239
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000392931\
4231394231314A39394A39395242425242395A42425A4242634A4A5A4A42634A4A
634A4A63524A634A4A6B524A5A5A5A6B5A5263524A6B5A526B524A6B5A5263524A6B5A526B5A52\
6B5A526B5A526B5A526B5A526B5A526B5A526B5A526B5A5273635A6B5A526B5A5A
6B5A5273635A6B5A526B5A526B52526B5A5263524A63524A63524A63524A5A4A425A4A42524239\
5A4239423131000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039394A42395A524A63524A6B5A5A7363637B\
6B6B846B6B8473738C73738C7B7B947B7B947B849C7B7B948484947B7B8C84848C
84848C848C8C7B84848484847B7B7B7B847B737373737B736B6B63636B5A525242000000000000\
0000000000000000000000000000000000000000000000000000004231314A3939
4A39395242424A39395242395A42396342425A4239634A42634A426B4A426B4A42734A426B4A42\
73524A6B4A4273524A73524A73524A73524A73524A73524A7B5A5273524A7B5A4A
735A4A7B5A52735A4A7B5A527B5A527B63527B5A4A7B63527B63528463527B63527B63527B6352\
846B5A7B6352846B5A846352846B5A7B6352846B5A846B52846B5A7B6352846B5A
846B52846B5A7B6352846B5A7B6B52846B5A7B63527B6B527B6B52846B5A7B63527B6B527B6352\
7B6B5A7B63527B6B527B63527B6B527363527B6B527363527B6B5273634A7B6B52
7363527B6B5273634A73635273634A73635273634A73634A73634A73634A6B5A426B5A4A6B5A42\
735A4A6B5A426B5A426B5A426B5A426352396B5A426352396B5A42634A39635239
5A4A395A4A395242315242395242315242314A39294A3931000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000042425A42426B4A4A7352527B\
5A63846363846B6B8C6B6B8C7373947373947B7B947B7B9484849C84849C8C8C9C
948C9C9494A594949C94949C949C949C9C9C949C949CA5949CA58CA5AD94A5A58CADA58CA5A584\
A5A5849C9C7B9C9C7B9C94739C9473948463948C638C7B5A847B5273634A635A39
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003131314231314A3139\
4A31395239425239425A42425A4242634A4A5A4A42634A4A5A4A4263524A634A4A
63524A63524A6B5A526B52526B5A526B5A526B5A526B5A526B5A526B5A526B5A526B5A526B5A52\
6B5A5273635A6B635A73635A6B5A5273635A6B635A73635A6B635A73636373635A
636B6B73635A73636373635A73635A6B5A5273635A6B5A526B5A526352526B5A5263524A63524A\
5A4A425A4A425A4239524239312931000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039394A42425A4A4A635A52735A5A7363637B\
63637B73738C73738C73738C73738C84849C7B84948484947B849484849484848C
84848C84848C8C8C8C84848484847B7B7B737B7B7373736B73736B6363525A5A4A000000000000\
0000000000000000000000004231314A39394A39394A39394A39395A42425A4242
5A42425A4239634A42634A426B4A426B4A4273524A6B4A4273524A6B4A4273524A73524A73524A\
73524A7B5A527B5A4A7B5A52735A4A7B63527B5A527B63527B5A5284635A7B6352
8463527B5A5284635A84635284635A7B6352846B5A84635A846B5A84635A8C6B5A846B5A846B5A\
846B5A8C6B63846B5A846B5A846B5A8C7363846B5A846B5A846B5A847363846B5A
846B5A846B5A847363846B5A846B5A7B6B5A84735A846B5A846B5A7B6B5A84735A7B6B5A846B5A\
7B6B52846B5A7B6B5A7B6B5A7B6B52846B5A7B6B527B6B5A7B6B527B6B5A7B6B52
7B6B5A736B527B6B5A7B6B527B6B527363527B6B5273635273635273634A73635273634A73634A\
6B5A4A73634A6B634A6B634A6B5A426B634A6B5A426B5A42635A426B5A42635242
6352425A4A396352425A4A395A4A39524231524239524231523931000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039393942426B4A4A7352527B\
5A5A7B636B8C6B6B8C73739473738C7B7B947B7B9484849C84849C8C8CA58C8C9C
9494A5948C9C9C94A594949C9C9C9C949C949CA59C9CA5949CA5949CA58CA5AD94A5AD8CA5A58C\
A59C84A5A5849C9C7B9C9473948C6B9C9473948C6B8C846342426B7B7352635A42
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003931314231314A39394A3139\
5239425239425A42425A4242634A4A5A4A42634A4A5A4A4263524A634A4A6B524A
63524A6B52526B52526B5A526B524A6B5A526B5A526B5A52635A526B5A526B5A526B635A6B5A52\
6B635A6B635A73635A6B635A73635A73635A736B636B635A73636373635A736B63
73635A736B63736363736B6373635A73636373635A7363636B5A5A6B635A6B5A5273635A635252\
63525263524A63524A5A42425A4242524239524239000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039395242395A4A4A6352526B63637B63637B\
6B6B846B6B8473738C73738C7B7B947B7B9484849C7B7B9484849484848C848494
84848C84848C8484848484847B7B7B7B7B7B737B73737B736B6B63636352525242000000000000\
4231394231394A39394A39395239425239395A42425A42396342426342426B4A42
6342426B4A426B4A4273524A6B4A4273524A73524A73524A73524A735A4A73524A7B5A52735A4A\
7B5A527B5A528463527B5A527B63527B635284635A7B635284635A84635A846B5A
84635A846B5A846B5A8C6B6384635A846B5A846B5A8C6B63846B5A846B5A846B5A8C7363846B5A\
846B63846B5A8C7363846B5A846B63846B5A8C7363846B5A847363846B5A847363
846B5A84736384735A847363846B5A847363846B5A8473637B6B5A84735A846B5A8473637B6B5A\
84735A7B6B5A84735A7B6B5A7B6B5A7B6B5A84735A7B6B527B6B5A7B6B5A7B735A
7B6B527B6B5A7B6B527B735A736B527B6B5A736B527B6B5A736352736B52736352736B526B634A\
7363526B634A7363526B5A4A6B634A6B5A4A6B634A635A426B5A4A4A5A526B5A4A
6352426352425A52426352425A4A395A4A39524239524A394A39314A3931000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039426B4A4A734A4A73\
5A5A7B5A63846B6B8C6B6B8C7373947373947B7B9C7B7B9484849C84849C8C8CA5
8C8C9C9494A594949C9C9C9C9494949C9C9C9C9C949CA5949CA58C9CA5949CAD8CA5AD94A5A584\
A5A5849C9C7B9C9C7B9494739C9473948C6B948C6B8C845A8C845A736B4A6B6342
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000393139424A424A524A4A3939524242\
5A42425A42425A4242634A4A634A4A634A4A634A4A6B525263524A6B524A63524A
6B5A526B5A526B5A526B524A6B5A526B5A526B5A526B5A526B635A6B5A526B63526B5A5273635A\
6B635A73635A6B635A736B636B635A736B636B635A736B63736363736B63736363
7B6B6B736B63736B637363637B6B63736B63736B637363637B6B6373636373636373635A736363\
6B5A5A6B5A5263524A6B525263524A634A425A42395A4A42524239423931000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039314A4A42634A4A635A526B5A5A736B6B7B\
6B6B7B6B6B846B6B847B7B947B7B947B7B947B7B9484849C8484948484947B848C
84848C84848C84848C7B847B84847B7B7B737B7B7373736B7373634231394A39394A3139523942\
5239395242425239395A42425A4242634A426342426B4A4A6B4A4273524A6B4A42
73524A73524A73524A73524A7B5A527352527B5A5273524A7B5A527B5A527B5A527B5A5284635A\
7B635284635A7B6352846B5A84635A846B5A84635A8C6B63846B5A8C6B63846B5A
8C7363846B638C7363846B5A8C7363846B638C7363846B5A8C7363846B638C7363846B5A8C7363\
846B63847363846B5A8C7363847363847363846B5A8C7363847363847363847363
8C73638473638473638473638C736384736384736384735A8473638473638473637B735A847363\
84735A8473637B6B5A8473637B735A84735A7B6B5A8473637B735A7B735A7B6B5A
84735A7B735A7B735A7B6B5A7B735A7B6B5A7B735A736B527B735A736B52736B52736B52736B52\
736B52736B526B634A736B526B634A7363526B634A7363526B634A6B634A6B5A42
6B634A635A42635A425A52396352425A52395A4A39524A395A4A395242314A3931000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039426342426B52527B\
52527B5A63846363847373946B738C73739473739484849C84849C84849C84849C
9494A5948CA59494A594949C9C9C9C949C9C9C9C9C9C9C949CA5949CA594A5AD949CAD8CA5AD8C\
9CA5849CA5849C9C7B9C9C7B949473948C738C84638C8C63847B5A7B7352635A42
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003931314231314A39394A3939524242524242\
5A4A4A5A42425A4A4A5A4A4A63524A634A4A63524A63524A6B525263524A6B5252
6B52526B5A526B524A6B5A526B5A526B5A526B5A526B5A526B5A526B635A6B5A526B635A6B635A\
73635A6B635A6B635A6B635A736B636B635A736B636B635A736B636B635A736B63
736B63736B6B736B63736B63736B63736B6B736363736B63736B637B6B6B736363736363736363\
7363636B5A5A73635A6B5A526B5A5263524A63524A634A42634A425242395A4239
5239314A3131000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039395242425A524A6B52526B635A7363637B\
6B6B846B6B8473738C73738C7B7B947B7B947B849C7B84948484947B7B8C84848C
84848C84848C8484848484847B7B7B4239394231314A39394A39395242425239395A42425A4242\
634A426342426B4A426B4A4273524A6B4A4273524A734A4A73524A73524A73524A
73524A7B5A5273524A7B5A527B5A527B5A527B5A527B63527B5A5284635A7B635284635A84635A\
846B6384635A846B63846B5A8C7363846B5A8C73638C6B638C7363846B638C7363
846B638C7363846B638C7363846B638C7363846B638C73638473638C7363846B638C7363847363\
8C73638473638473638473638C73638473638473638473638C7B6B847363847363
8473638C7B6B847363847B638473638C7B6B847363847363847363847B63847363847363847363\
847B637B735A8473637B7363847B637B735A8473637B735A8473637B735A7B735A
7B735A7B7363736B5A7B735A736B5A7B735A736B5273735A736B5273735A736B52736B52736B52\
736B526B6352736B526B6352736B526B634A6B63526B634A6B63526B634A6B634A
635A4A6B634A635A42635A426352426352425A4A395A5239524A395A4A394A4231423931000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039394242426B4A4A73\
525A7B525A7B63638C6B6B8C7373946B738C737B947B7B9484849C84849C8C8C9C
8C8C9C9494A58C8C9C94949C94949C9C9C9C949C949CA5949CA5949CA5949CA58C9CAD8C9CAD8C\
A5AD8C9C9C7B6B6B8494947B9C94738C8C6B8C8C6B8C84638C845A7B734A6B6342
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003129314231394231394A39394A39395242425A42425A4242\
5A4242634A4A634A4A63524A634A4A6B52526B52526B52526B524A6B5A526B5A52
6B5A526B5A526B5A526B5A526B5A526B5A526B635A6B5A526B6B636B5A5273635A6B635A6B635A\
6B635A6B635A6B635A6B63636B635A736B636B6363736B636B6363736B6B736B63
736B6B736B637B736B736B6B7B6B6B736B637B736B736B6B7B6B6B736B637B736B736B637B6B63\
7363637B6B6373636373636373635A7363636B5A526B5A5263524A63524A634A42
5A4A425242395242394A3131000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003939524A42634A4A635A5A73635A736B637B\
6B637B73738C73738C737B94737B8C84849C7B84948484947B849484849484848C
84848C4231314A39394A39395239425239395A42425A42426342425A42426B4A4A6B4A426B4A4A\
6B4A4273524A73524A73524A73524A7B5A5273524A7B5A527352527B5A527B5A52
7B5A527B5A5284635A7B635A84635A7B635A846B6384635A846B6384635A8C6B63846B638C6B63\
846B638C73638C73638C73638C73638C736B8C73638C73638C73638C736B8C7363
8C736B8C73638C7B6B8C73638C736B8C73638C7B6B8C73638C736B8473638C7B6B8C73638C7B6B\
8473638C7B6B8C73638C7B6B8473638C7B6B8C7B638C7B6B8473638C7B6B8C7B6B
8C7B6B847B638C7B6B8C7B6B8C7B6B847B638C7B6B847B638C7B6B847B638C7B6B847B63847B63\
847363847B6B847363847B637B7363847B637B73637B73637B735A847B637B7363
7B73637B735A7B73637B735A7B7363736B5A7B736373735A73735A736B5A73735A736B5A736B5A\
6B6B52736B5A736B52736B526B6352736B526B63526B6B526B634A6B6B526B634A
6B634A635A426B634A635A42635A42635A42635A425A52395A52395A4A39394A5A4A4231423931\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000042424A42426B4A4A73\
4A52735A63846363846B6B8C6B6B8C737B94737B947B7B9C7B7B94848C9C848C9C
8C8CA58C8C9C9494A594949C949C9C949C9C9CA59C9C9C949CA5949CA594A5AD949CAD8CA5AD94\
9C9C8CA5A58C9C9C849C9C7B9494739494738C8C6B8C8C63847B5A847B5A6B6342
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003139393931314231314A39394A39395242425242425A4A4A5A4242\
634A4A5A4A4A63524A634A4A63524A6B524A6B5A526B524A6B5A526B5A526B5A52
6B5A526B5A526B5A526B5A52635A526B5A526B5A526B635A6B5A526B635A6B63526B635A635A52\
6B635A6B635A6B636363635A6B63636B635A736B636B635A736B63736B63736B6B
736B6373736B736B6B7B736B736B6B7B6B6B736B6B7B736B736B6B7B736B7B6B6B7B736B736B63\
7B6B6B736B637B6B6B7363637363637363637363636B5A5A73635A6B5A526B5A52
63524A63524A5A4A425A42425239394A3931000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000042395A42425A524A6352526B63637B63637B\
6B6B846B6B8473738C73738C7B7B947B7B9484849C7B7B948484944A3139523939
4A39395242425242425A42425A42426342426342426B4A4A6B4A42734A4A734A4A73524A734A4A\
73524A73524A7B5A5273524A7B5A527B5A527B5A5A7B5A527B5A5A7B5A5A84635A
7B5A5A84635A84635A846B6384635A846B63846B5A8C6B63846B5A8C6B638C6B638C736B8C6B63\
8C736B8C73638C736B8C73638C736B8C736B8C7B6B8C73638C736B8C736B8C7B6B
8C73638C736B8C736B8C7B6B8C73638C7B6B8C736B8C7B6B8473638C7B6B8C7B6B8C7B6B847363\
8C7B6B8C7B6B8C7B6B847B638C7B6B8C7B6B8C7B6B847B6B8C7B6B8C7B6B8C846B
847B6B8C7B6B8C7B6B8C846B847B63847B6B847B6B8C7B6B847B63847B6B847B63847B6B7B7363\
847B637B7363847B637B73637B7B637B73637B7B637B735A7B73637B73637B7B63
73735A7B736373735A7B736373735A73736373735A7B7363736B5A73735A736B5A73735A6B6B52\
736B5A6B6B52736B5A6B63526B6B526B6B526B6B526B63526B6B526B63526B6352
63634A63634A635A4A63634A635A42635A425A5A42635A425A52395A52395A52395A52394A4231\
423931000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039394242426B42426B\
4A5273525A7B63638463638C6B73947373947B7B94737B947B849C84849C8C8C9C
84849C8C8CA59494A59494A594949C949C9C949C9C9CA59C949C949C9C9C9C9C949C9C949C9C8C\
9C9C8C9C9C849C9C8494947B9494738C8C6B948C6B8C84638C845A7B7352736B4A
5A5239000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003931314231314A39394A39395242425242425A424A5A4242634A4A\
634A4A634A4A634A4A6B52526B524A6B5A526B524A6B5A526B5A526B5A526B5A52
735A526B5A526B5A526B5A526B63526B5A526B63526B5A526B635A6B63526B635A635A526B635A\
63635A6B635A63635A6B6B636B635A6B6B636B6363736B6B6B6B63736B6B6B6B63
73736B73736B737373736B6B7B737373736B7B7373736B6B7B73737B736B7B73737B6B6B7B7373\
7B6B6B7B6B6B736B6B7B6B6B736B63736B637363637B6B637363637363636B5A5A
73635A6B5A526B5A5263524A63524A5A42425242394A39314A3931000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003131393939524A4A63524A635A5A735A5A736B6B84\
6B6B847373846B6B847B7B947B7B947B849C4A3139523942523939524242523942
634242634242634A4A634A4273524A6B4A4A73524A6B4A4A73525273524A73525273524A7B5A52\
735A527B5A52735A527B635A7B5A5A7B635A7B5A5A84635A84635A84636384635A
846B63846B63846B63846B638C736B8C6B638C736B8C6B6394736B8C736B8C736B8C7363947B6B\
8C736B8C736B8C736B947B6B8C736B8C7B6B8C736B947B738C736B8C7B6B8C736B
947B738C7B6B8C7B6B8C736B8C7B738C7B6B8C7B6B8C7B6B8C7B738C7B6B8C7B6B8C7B6B8C8473\
8C7B6B8C7B6B8C7B6B8C84738C7B6B8C7B738C7B6B8C84738C7B6B8C8473847B6B
8C8473847B6B8C7B6B847B6B8C846B847B6B847B6B847B6B847B6B847B6B847B6B7B7363847B6B\
7B7B637B7B6B7B7363847B6B7B73637B7B637B73637B7B637B73637B7363737363
7B7B637373637B736373735A7B7363737363737363736B5A737363736B5A73735A6B6B5A73735A\
6B6B5A736B5A6B6B5273735A6B6B526B6B526B6B526B6B5A6B6B526B6B5263634A
6B6B5263634A63634A635A4A63634A635A42635A425A5242635A425A52395A5242524A395A5239\
4A4231423931000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003939524A4A73\
4A52737B7B5A5A5A7B6B6B8C6B6B8C7373947373947B7B9C7B7B9484849C84849C
8C8CA58C8CA59494A58C949C9C9CB59494AD9C9CAD9494A59C9CA59C9C9C9C9C9C9C9C949C9C94\
9C9C8C9C9C8C9C9C7B9C9C7B9494739494738C8C6B8C8C6B84845A847B5A6B634A
635A42000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003931314231314A39394A31395239425242425A4A4A5A42425A4A4A5A4A4A\
63524A634A4A63524A63524A6B5A526B524A6B5A526B52526B5A526B524A6B5A52
6B5A526B5A52635A526B5A526B5A526B6352635A526B6352635A526B635A635A52636352635A52\
6B635A635A5263635A63635A6B6B6363635A6B6B636B6B63736B6B6B6B6373736B
736B6B737373736B6B7373737373737B737373736B7B73737B73737B737373736B7B73737B7373\
7B7373736B6B7B736B7B6B6B7B736B736B637B6B6B736B637B6B6B736363736363
73635A73635A6B5A526B5A526B52526B52525A4A425A4A425242395242394A3131000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000031313139395242425A52526B52526B63637B63637B\
6B6B846B6B844A31394A31395239424A39395242425A42426342425A4242634A4A
634A4A6B524A6B4A4A73525273524A73525273524A7352527352527B5A527352527B5A5A7B5A52\
7B635A7B5A527B635A7B635A8463637B635A846B638463638C6B63846B638C6B63
846B638C736B846B638C736B8C736B8C736B8C73638C736B8C736B947B6B8C736B8C736B8C736B\
947B738C736B8C7B6B8C736B947B738C736B8C7B738C7B6B947B738C7B6B8C7B73
8C7B6B947B738C7B6B8C7B738C7B6B8C84738C7B6B8C7B738C7B6B8C84738C7B6B8C7B738C7B6B\
8C84738C7B6B8C84738C7B6B8C8473847B6B8C8473847B6B8C8473847B6B847B6B
847B6B8C8473847B6B847B6B847B6B847B6B7B7B6B847B6B7B7B6B847B6B7B73637B7B6B7B7B63\
7B7B6B7B73637B7B6B7B73637B7B6B7373637B7B637373637B7B6B737363737363
7373637B7B637373637373637373637373636B6B5A73736373735A7373636B6B5A73735A6B6B5A\
73735A6B6B526B6B5A6B6B5A6B6B5A6B6B526B6B5A6B6B526B6B5A6363526B6B52
6363526B6B5263634A63634A635A4A63634A5A5A42635A425A5A42635A425A52395A5239524A39\
524A39423931423931000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000042425242426B\
4A527B4A527B5A5A845A63846B6B8C6B6B8C737B94737B947B849C7B7B9C84849C
84849C8C8CA58C8CA59494AD9494AD9C9CB59494A5949CA5949C9C9C9CA59494949C9C949C9C8C\
9C9C8C949C849C9C8494947B94947B8C8C6B8C8C6B8C84638C84637B7B52736B4A
5A5239000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003931394231314239394A39395242425242425A424A5A4242634A4A634A4A634A4A\
634A4A6B52526B524A6B5A526B5252735A526B5A526B5A526B5A527363526B5A52
6B63526B5A526B635A6B5A526B63526B5A526B635A635A52636352635A5263635A63635263635A\
635A5A6B635A63635A63636363635A6B6B636B6B636B6B6B6B6B6373736B6B6B6B
7373736B6B6B7373737373737B73737373737B7B7B7B73737B7B7B7B7373847B7B7B7373847B7B\
7B7373847B7B7B73738473737B6B6B8473737B6B6B7B6B6B7B6B6B7B6B6B7B6B63
7B6B6373635A73636373635A73635A6B5A52735A5A6B525263524A5A4A425A4A42524239523939\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039394242395A4A42634A4A635A5A7363637B4A3939\
4A31395242425239425A42425A4242634A4A634242634A4A634A4A73525273524A
7352527352527B5A527B5A527B5A52735A527B635A7B5A5A7B635A7B5A5A84636384635A846363\
84635A846B63846363846B63846B638C6B6B8C6B638C736B8C6B638C736B8C736B
8C736B8C736B947B738C736B947B6B8C736B947B738C736B947B738C736B947B738C7B6B947B73\
8C7B6B947B738C7B73947B738C7B6B9484738C7B739484738C7B739484738C7B73
9484738C7B739484738C84739484738C7B739484738C84739484738C84739484738C84738C8473\
8C84738C84738C84738C8473847B6B8C84738C84738C8473847B6B8C8473847B6B
848473847B6B848473847B6B84846B7B7B6B8484737B7B6B847B6B7B7B6B847B6B7B7B6B7B7B6B\
7B7B637B7B6B7B7B6B7B7B6B7373637B7B6B7B7B637B7B6B7373637B7B6B737363
737B637373637B7B6B737363737B63737363737B637373637373636B735A7373636B7363737363\
6B6B5A7373636B6B5A6B735A6B6B5A6B735A6B6B5A6B6B5A636B526B6B5A6B6B52
6B6B526363526B6B5263634A6363525A5A4A63634A635A4A635A4A5A5A42635A4A5A52425A5239\
524A31524A394A4231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000423952424A6B\
424A73525A7B525A84636384636B8C737394737B947B7B947B7B9484849C84849C
848CA5848CA58C94AD8C94AD9494AD9494AD9C9CAD949CA59C9CA594949C9C9C9C949C949C9C94\
949C8C9C9C8C9C9C849C9C849494739494738C8C6B8C8C6384845A84845A736B4A
635A42000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003931314231314A39394231395239425242425A424A5A4242634A4A5A4A4A63524A634A4A\
63524A63524A6B5A526B524A6B5A526B5A526B5A526B5A526B5A526B5A526B6352
6B5A526B63526B5A526B635A635A526B6352635A526363525A5A52635A525A5A5263635A5A5A52\
63635A63635A6363635A635A6363636363636B6B636363636B6B6B6B6B6B6B736B
6B6B6B73736B6B736B73737373736B7373737373737B7B7B7373737B7B7B7B7373847B7B7B7373\
847B7B847373847B7B7B73738473738473738473737B6B6B8473737B6B6B7B7373
736B637B6B6B736363736363735A5A73635A6B5A5A735A5A6B52526B5A5263524A634A4A5A4242\
5242394A39394A3931000000000000000000000000000000000000000000000000
00000000000000000000000000000000000031313942425A42425A4A39394A3939523942523939\
5A42425A4242634A4A5A4242634A4A6B4A4A7352526B4A4A7352527352527B5A52
7352527B5A5A7B5A527B635A7B5A5A7B635A7B635A8463637B635A846363846363846B63846363\
846B63846B638C736B846B638C736B8C736B94736B8C736B8C736B8C736B947B73
8C736B947B6B8C736B947B738C736B947B738C7B6B947B738C7B6B947B738C7B73947B738C7B6B\
947B738C7B739484738C7B739484738C7B739484738C7B739484738C847394847B
8C7B739484738C847394847B8C84739484738C847394847B8C84738C84738C84738C84738C7B73\
8C84738C84738C8473847B6B8C8473847B738C8473847B6B848473847B6B848473
7B7B6B848473847B6B8484737B7B6B847B6B7B7B6B8484737B7B6B7B7B6B7B7B6B7B7B6B737B6B\
7B7B6B7B7B6B7B7B6B7373637B7B6B737B6B7B7B6B737363737B6B737363737B6B
737363737B6B737363737B6B6B7363737363737363737B636B73637373636B73637373636B6B5A\
6B73636B735A6B73636B6B5A6B735A6B6B5A6B735A636B5A6B6B5A636B5A6B6B5A
636352636B52636352636B5263634A63635263634A63634A5A5A425A5A4A42525A5A5A42525239\
525239524A39524A39423929000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000042424A394263\
4A4A734A527B5A5A845A5A84636B8C6B6B8C737B94737B947B7B947B849C84849C
84849C848CA58C8CA59494AD8C94A59494AD9494AD949CAD94949C949C9C949C9C9C9C9C94948C\
9C9C8C949C8C9C9C8494947B94947B8C8C738C8C6B848463848463847B5A737352
635A39000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3931313931314231394A31395239425239425A424A5A4242634A4A634A4A634A4A634A4A6B5252\
6B524A6B52526B524A735A526B5A526B5A526B5A52735A526B5A526B5A526B5A52
73635A6B5A526B6352635A526B6352635A52635A525A5A526363525A5A52635A525A5A5263635A\
5A5A5A5A635A5A5A5A6363635A635A6363635A6363636B6B636B636B6B6B636B6B
6B73736B6B6B7373736B6B6B7373737373737373737373737B7B7B7B737B7B7B7B7B7373847B7B\
7B737B847B7B7B7373847B7B847B7B847B7B847373847B7B8473738473737B6B73
8473737B6B6B7B6B6B7363637B6B6B7363637363636B5A5A73635A6B5A5A6B5A5A6B52526B5252\
634A4A5A4A425242395242394A3131000000000000000000000000000000000000
0000000000000000000000000000000000004A39394A31395242425239425A42425A4242634A4A\
634A4A6B4A4A6B4A4A7352527352527352527352527B5A5A735A527B5A5A7B5A52
7B635A7B5A5A84635A7B635A846B63846363846B638463638C6B6B846B638C6B6B846B638C736B\
8C6B6B8C736B8C736B9473738C736B9473738C736B947B738C736B947B738C736B
947B738C7B73947B738C7B6B947B738C7B73947B738C7B739484738C7B739484738C7B7394847B\
8C7B739484738C7B7394847B94847394847B8C847394847B94847394847B8C8473
948C7B94847394847B8C847394847B8C84738C84738C84738C847B8C84738C8473847B738C8473\
8484738C8473847B738C8473848473848473847B6B848473847B738484737B7B6B
8484737B7B6B8484737B7B6B8484737B7B6B7B7B737B7B6B7B84737B7B6B7B7B6B737B6B7B7B73\
737B6B7B7B6B737B6B7B7B6B737B6B737B6B73736B7B7B6B737B6B737B6B6B7363
737B6B6B736373736B6B7363737B6B6B73636B73636B73637373636B73636B73636B6B5A6B7363\
6B6B5A6B7363636B5A6B73636B6B5A6B735A636B5A6B735A636B5A636B5A636B52
6B6B5A636B52636B52636352636B526363526363525A634A6363525A5A4A63634A5A5A425A5A42\
5252395252394A4A31524A39000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000393942424A63\
424A734A527B52527B5A63846363846B738C6B738C737B94737B947B849C7B849C
848CA5848C9C8C8CA58C8CA59494AD9494AD949CAD9494A5949CA5949C9C949C9C9494949C9C94\
949C8C9C9C8C949C849C9C8494947B9494738C8C6B8C8C6B84846384845A736B4A
6B634A000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000313131
39293152524A4231395239425239425A424A5242425A4A4A5A4A4A63524A634A4A6B525263524A\
6B52526B524A6B5A526B52526B5A526B524A6B5A526B5A526B5A526B5A526B5A52
6B5A526B6352635A4A635A52635A52635A525A5A4A5A5A525A5A525A5A52525A525A5A525A5A52\
5A5A5A525A525A5A5A5A5A5A5A63635A5A5A5A63635A6363636B6B636363636B6B
636B6B6B73736B6B6B6B73736B6B737373736B6B737373737373737B7B7B7373737B737B7B737B\
7B7B7B7B7373847B7B7B737B847B7B847373847B7B847B7B8C7B7B8473737B7373
7B73738473737B6B6B7B6B6B7B6B6B7B6B6B7363637363637363637363636B5A5A735A5A6B5A52\
6B5A5263524A63524A5A4A425A4A424A39394A3939393131000000000000000000
0000000000004231314A39394A39395242425239395A42425A4242634A4A6342426B4A4A6B4A4A\
7352526B4A4A7352527352527B5A5A7352527B5A5A7B5A5A7B635A7B5A5A7B635A
7B635A8463637B635A846B63846B638C6B6B846B638C6B6B8C6B6B8C736B8C6B6B8C736B8C736B\
947B738C736B8C73738C736B947B738C736B947B738C7B73947B738C736B947B73
8C7B73947B738C7B73947B738C7B739484738C7B739484738C7B7394847B8C7B739484738C8473\
94847B8C7B7394847B8C847394847B8C847394847B94847B948C7B8C847394847B
8C847394847B8C84738C847B8C84738C847B8484738C84738484738C8473847B73848473848473\
848473847B73848473847B738484737B7B6B8484737B7B738484737B7B6B7B8473
7B7B738484737B7B6B7B84737B7B6B7B8473737B6B7B7B737B7B6B7B8473737B6B7B7B6B737B6B\
7B7B73737B6B737B6B737B6B737B6B6B7363737B6B6B7363737B6B6B73636B736B
6B73636B736B636B636B73636B73636B7363636B636B7363636B636B7363636B5A637363636B5A\
6B7363636B5A636B5A636B5A637363636B5A636B5A636B5A636B5A5A6352636B5A
636B52636B5A5A6352636B52636352636B525A634A6363525A634A6363525A5A4A5A5A4A525A42\
5A52424A4A39524A394A4231423931000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000042425A\
4A4A734A4A73525A7B525A84636B8C636B8C6B73947373947B7B94737B947B849C
7B849C848CA5848C9C8C94A58C94A5949CAD8C94A59494A59494A5949CA5949494949C94949C94\
9C9C94949484949C8494947B94947B8C8C6B8C8C6B8484638484637B7B5A737352
636342000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000393131
4231394231395239425239425A4242524242634A4A5A4A4A634A4A634A4A6B52526B52526B5252\
6B524A735A526B5A526B5A526B5A52735A526B5A526B5A526B5A5273635A6B5A52
6B6352635A526B6352635A52635A525A5A4A635A525A5A525A5A52525A4A5A5A52525A52525A52\
525A525A5A5A525A5A5A635A525A5A5A63635A63636363635A6363636B6B6B6363
6B6B6B636B6B6B73736B6B736B73736B737373737B73737373737B73737B7B7B7B7B737B7B7B7B\
7B737B847B7B7B7B7B847B7B84737B847B7B847B7B8C7B7B847B7B8C7B7B7B7B7B
847B7B7B7373847B7B8473738473737B6B6B7B73737B6B6B7B6B6B7363637B6363736363736363\
6B5A5A73635A6B5A526B525263524A63524A5A4A425A42424A39394239394A3139
4A39394A39395242425242425A42425A4242634A4A634A4A6B4A4A6B4A4A735252735252735252\
7352527B5A5A7B5A5A7B5A5A7B5A5A8463637B635A8463637B635A846B63846363
846B638463638C6B6B8C6B6B8C736B8C6B6B9473738C736B9473738C736B947B738C7373947B73\
8C7373947B738C7B73947B738C7B73947B7B947B73947B738C7B7394847B947B73
94847B8C7B7394847B94847394847B8C7B7394847B94847B94847B8C8473948C7B94847B94847B\
8C847B948C7B94847B948C7B8C847B948C84948C7B948C7B8C847B948C7B8C847B
8C8C7B8C84738C8C7B8C847B8C847B8484738C8C7B8484738C847B8484738C847B848473848473\
84847384847B8484738484737B7B7384847B7B84738484737B7B738484737B8473
7B84737B7B737B84737B84737B8473737B6B7B8473737B737B7B73737B6B7B7B73737B6B737B6B\
6B736B737B6B6B736B6B7B6B6B73636B7B6B6B73636B736B6373636B736B637363
6B736B6373636B736B637363637363636B63637363636B63637363636B5A637363636B63636B63\
5A6B5A637363636B5A636B635A6B5A636B63636B5A636B5A5A6B5A636B5A636B5A
636B5A5A6352636B5A636352636B525A6352636B525A63526363525A634A6363525A634A5A5A4A\
5252425A5A42525239524A39424231000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000042425A\
424A6B4A527B525A7B5A63845A63846B738C6B738C737B94737B947B849C7B8494
84849C84849C8C94A58C8CA58C94A58C94A5949CAD9494A5949CA594949C949C9C949C94949C94\
94948C9C9C8C949C84949C848C947B94947B8C8C6B8C8C6B848463848463737352
6B634A525239000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000313131423139
4231394A39425239425A42425239425A424A5A4242634A4A5A4A4A63524A63524A6B525263524A\
6B5A526B5252635A526B5A526B5A526B5A526B63526B5A526B63526B5A526B6352
635A52635A52635A52635A525A5A4A5A5A525A5A4A5A5A5252524A52524A4A524A525A524A524A\
525A524A5252525A5A4A5A52525A5A525A5A5A6363525A635A63635A6363636B6B
5A636B636B6B636B6B6B7373636B736B73736B737373737B6B7373737B7B73737B7B7B7B73737B\
7B7B7B7B737B7B7B7B7B737B847B7B7B7B7B847B8484737B847B7B7B7B7B847B84
7B737B847B7B7B737B847B7B7B73738473737B6B738473737B6B6B7B6B6B7363637B6B6B736363\
736363735A5A73635A6B52526B5A526352526B52524231394A39394A3939524242
5239425A42425A4242634A4A63424A6B4A4A6B4A4A7352527352527352527352527B5A5A735252\
7B5A5A7B5A5A8463637B5A5A846363846363846B63846363846B63846B638C6B6B
846B638C736B8C6B6B8C73738C736B9473738C7373947B738C7373947B738C7B73947B738C7B73\
947B738C7B73947B7B8C7B73947B73947B7394847B8C7B7394847B947B7394847B
8C7B7394847B94847B94847B8C847394847B94847B948C7B8C847B94847B94847B948C848C847B\
948C7B948C7B948C848C847B948C7B8C847B948C7B8C847B8C8C7B8C847B8C8C7B
8484738C847B8C847B8C8C7B8484738C847B8484738C847B84847384847B84847384847B7B8473\
8484737B847384847B7B7B737B84737B847384847B7B7B737B84737B84737B8473
737B737B8473737B737B8473737B6B737B73737B6B737B736B736B737B6B6B736B737B6B6B7363\
6B736B6B736B6B736B6373636B736B6373636B736B636B63637363636B6363736B
5A6B63637363636B636373635A6B63636B635A6B63636B635A6B5A636B635A6B5A636B635A6B5A\
5A6B635A6B5A636B635A635A5A6B635A6B5A636B635A635A5A6B5A5A635A636B5A
5A63525A635A5A6352636B5A5A63525A63525A63526363525A63525A63525A5A4A5A634A5A5A42\
5A5A4A5252425252424A4A394A4A39000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393952\
424A6B4A4A73525A7B525A7B5A6384636B8C6B73946B7394737B94737B947B849C
7B8494848C9C848C9C8C94A58C8CA59494AD8C94A5949CAD8C94A59494A594949C949C9C8C948C\
949C94949C8C949C8C94947B94947B8C8C738C94738484638484637B7B5A7B7B5A
6363425A5239000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000313131393139393131
4A39424A39425242425239425A424A5A424A634A4A5A4A4A6B525263524A6B52526B524A6B5A52\
6B52526B5A526B5252735A526B5A526B5A526B5A527363526B5A526B6352635A52
6B6352635A52635A525A5A4A5A5A525A5A4A5A5A5252524A5252524A524A4A524A4A524A4A5252\
4A52524A52524252524A5A5A4A525A525A5A4A5A5A526363525A63526363526363
5A6B6B5A636B636B73636B6B6B73736B6B736B737B6B737373737B6B737B73737B73737B7B7B84\
73737B7B7B7B73737B7B7B847B7B7B7B7B7B7B737B847B84847B7B7B84847B7B7B
847B847B7B7B847B7B7B7373847B7B7B73738473737B6B738473737B6B737B6B73736B6B7B6B6B\
7363637363637363637363634A31394A39394A39395242425242425A424A5A4242
634A4A634A4A6B4A526B4A4A7352527352527352527352527B5A5A7B5A5A7B5A5A7B5A5A846363\
7B63638463637B6363846B6B846363846B6B846B638C736B8C6B6B8C736B8C6B6B
8C73738C73739473738C7373947B7B947B73947B738C7B73947B7B947B73947B7B8C7B7394847B\
947B7394847B8C7B7394847B947B7B94847B947B7394847B94847B94847B94847B
94847B94847B94847B94847B948C8494847B948C7B94847B948C84948C7B948C8494847B948C84\
948C7B948C7B8C847B948C848C847B8C8C7B8C847B8C8C7B8C847B8C847B84847B
8C8C7B84847B8C847B8484738C847B84847B84847B84847384847B84847384847B7B847384847B\
7B84737B847B7B7B7384847B7B84737B84737B7B737B847B737B73737B73737B6B
737B73737B6B737B736B736B737B736B736B6B7B6B6B736B6B7B6B6B736B6B736B6373636B736B\
63736B63736B636B6363736B636B6363736B5A6B6363736B5A6B636373635A6B63
63736B5A6B635A6B635A6B635A73635A6B635A6B635A6B635A6B635A6B635A6B635A6B5A5A6B63\
5A6B635A6B6352635A5A6B635A6B5A5A6B635A635A5A6B635A6B5A5A6B5A5A635A
5A6B5A5A635A5A6B5A5A63525A6B5A5A63525A63525A63525A63525A63525A6352525A4A5A634A\
5A5A4A5A5A4A5252425252424A4A39393131000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000424252\
42426B4A52734A527B5263845A6384636B8C636B8C737B94737B94737B94737B94
7B849C7B849C848C9C848C9C8C94A58C94A59494AD8C94A5949CA594949C949C9C8C9494949C94\
949C8C949C8C949484949C848C947B8C9473848C6B8C8C6B84846384845A737352
6B6B4A5A5239000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000394239393139393131423939
4A39395242425239425A42425A4242634A4A5A4A4A634A4A634A4A6B525263524A6B52526B524A\
6B5A526B524A6B5A526B5A527363526B5A526B5A526B5A526B6352635A4A635A52
635A4A635A525A524A5A5A5252524A5A5A524A524A4A524A424A424A524A424A4A424A4A394A4A\
424A4A394A4A425252424A524A525A4252524A525A4A525A525A634A5A5A525A63
525A635A636B5A636B5A6B6B5A6B6B636B73636B736B73736B6B736B737B6B6B7373737B73737B\
737B8473737B737B7B73737B7B7B8473737B7B7B847B737B8484847B7B847B8484
7B7B7B847B847B7B7B7B7B7B7B737B847B7B7B73737B737B7B73738473737B6B737B73737B6B6B\
4A39394231394A39395239395242425239425A424A5A424A634A4A634A4A6B4A52
6B4A527352527352527352527352527B5A5A73525A7B5A5A7B5A5A7B63637B5A5A8463637B6363\
846B6B846363846B6B846B6B8C6B6B846B6B8C736B8C6B6B8C73738C736B8C7373
8C7373947B7B8C7373947B73947B73947B7B8C7B73947B7B947B7394847B8C7B7394847B947B7B\
94847B8C7B7394847B94847B94847B8C847B94847B94847B94847B94847B94847B
94847B948C8494847B948C7B94847B948C8494847B948C84948C7B948C848C847B948C7B8C847B\
948C848C847B8C8C7B8C847B8C8C7B84847B8C847B84847B8C8C7B84847B84847B
84847B84847B84847384847B84847B84847B7B847384847B7B847384847B7B7B737B847B7B8473\
7B847B7B7B737B8473737B737B8473737B6B737B73737B73737B736B736B6B7B73
6B736B6B7B7363736B6B736B63736B6B736B63736B63736B63736B63736B5A6B6363736B5A6B63\
63736B5A6B635A736B5A6B635A736B5A6B635A6B635A6B635A6B6B526B635A6B63
526B635A6B635263635A6B63526B635A6B6352635A526B63526B635A6B6352635A526B6352635A\
5A6B6352635A526B6352635A5A6B6352635A5A6B5A52635A5A6B6352635A5A6B5A
52635A5A6B5A5263525A635A5263525A6B5A5263525A63525A63525A6352525A4A5A634A525A4A\
5A5A4A525242525A42525242525242393931000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393942\
424A63424A734A527B525A7B5A63845A63846B738C6B7394737B94737B947B8494
7B8494848C9C7B849C848C9C848C9C8C94A58C94A58C94A58C949C949CA58C949C949C9C8C9494\
949C948C9484949C8494948494947B8C8C738C8C73848C6B848C637B7B5A7B7B52
6B6B4A5A5A42000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003931313931314239394A3939
5239425239425A424A5A42425A4A4A5A424A635252634A4A6B525263524A6B5A526B52526B5A52\
6B5252735A526B5A52735A526B5A5273635A6B5A526B63526B5A526B6352635A52
635A525A5A4A635A525A524A5A5A4A4A524A4A524A424A42424A4239424239424231424231424A\
314242394A4A394A4A394A52394A4A42525A424A5242525A42525A4A5A634A5A63
525A63525A635A636B5A636B5A636B5A636B636B73636B73636B73636B736B737B6B737B6B737B\
6B737B737B8473737B737B8473737B7B7B847B7B847B7B847B8484848C8C7B8484
8484847B7B848484847B7B84847B847B7B7B847B847B7B7B847B7B7B737384737B4A3139523939\
5239395A42425A42425A424A5A424A6B4A52634A4A6B52526B4A52735A5A735252
735A5A73525A7B5A5A7B5A5A7B5A637B5A5A8463637B63638463637B6363846B6B84636B846B6B\
846B6B8C73738C6B6B8C73738C6B6B9473738C7373947B738C7373947B7B947B7B
947B7B947B7394847B947B7B94847B947B7B94847B94847B94847B947B7B94848494847B94847B\
94847B948C8494847B94848494847B948C8494847B948C8494847B948C84948C84
948C84948C849C8C84948C84948C84948C7B948C84948C84948C848C847B948C848C8C7B8C8C84\
8C847B8C8C848C8C7B8C8C8484847B8C8C8484847B7B635A84847B8C8C8484847B
848C7B84847B848C7B84847B84847B7B847B848C7B7B847B84847B7B847384847B7B847B7B847B\
737B737B847B737B73737B736B7B6B737B736B7B736B7B736B736B6B7B7363736B
6B736B63736B6B737363736B63736B5A6B6363736B63736B63736B5A6B6363736B5A6B635A6B6B\
5A6B635A736B5A6B635A6B635263635A6B635263635263634A5A5A4A63634A635A
4A63634A5A5A4A63634A635A4A635A4A5A5A4A63634A635A4A635A4A5A5A4A63634A635A4A635A\
4A5A5A5263634A635A4A635A4A5A5A52636352635A52635A52635A5A6B6352635A
5A6B6352635A5A6B635A635A5A635A52635A5A6B5A5A635A5A635A5263525A635252634A5A634A\
525A4A5A5A4A525A42525A424A5239424239000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39425A4A527B4A527B525A84525A84636B8C636B8C6B73946B7B947B8494738494
7B84947B8494848C9C848C9C8C94A58C94A5949CAD8C94A59494A58C949C949C9C8C949C949C94\
8C948C949C94949C84949C848C947B8C947B8C8C73848C6B8484638484637B7B5A
6B6B4A5A5A39000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000313131312931423139423939523942
5239425A42425A42425A4A4A5A4242634A4A634A4A6B525263524A6B52526B52526B5A526B524A\
6B5A526B5A527363526B5A526B5A526B5A526B63526B5A526B5A52635A52635A52
5A5A4A635A525A5A4A4A39394A4A42424A42424A42394242000000000000000000000000000000\
00000000000000000000000031313931394239424A39424A424A52424A524A525A
42525A4A5A634A5A6352636B525A6352636B635252636B735A636B636B73636B736B737B636B73\
6B737B6B737B7373846B737B7373847373847B7B847373847B848C7B848C848C94
7B848C7B848C7B848C84848C7B7B847B7B847B7B84847B847B737B5239395239395A4242523942\
5A42425A424A6B4A52634A4A6B52526B525273525A735252735A5A735A5A7B5A5A
735A5A7B5A637B5A5A8463637B5A63846363846363846B6B846363846B6B846B6B8C7373846B6B\
8C73738C73739473738C7373947B738C7373947B7B8C7B73947B7B947B7B94847B
8C7B7B94847B94847B94847B8C7B7B94847B94847B94848494847B94848494847B948C8494847B\
948484948484948C8494847B948C84948C84948C84948484948C84948C84948C84
948C84948C84948C84948C848C847B948C848C8C84948C848C847B8C8C848C847B8C8C8484847B\
8C8C8484847B8C8C8484847B848C7B84847B848C84735A52848C7B84847B848C84
7B847B84847B7B847B848C7B7B847B7B847B7B847B7B847B737B737B847B737B737384736B7B73\
737B736B7B736B7B736B736B6B7B7363736B6B7B7363736B63736B63736B63736B
5A6B6B63736B5A6B6B63736B5A6B635A6B6B5A6B635A736B526B635A6B63526363526B634A635A\
4A63634A5A5A4A6363425A5A425A5A42525A425A5A395252425A5A39525A425A5A
395252425A5A39525A425A5A395252425A5A42525A425A5A395252425A5A42525A425A5A395252\
425A5A425A5A4A5A5A425A5A4A5A5A4A5A5A4A635A4A5A5A4A635A4A5A5A52635A
4A5A5A52635A52635A52635A52635A52635A52635A5A635A5263525A63525263525A6352525A4A\
525A4A525A4A525A4A525242525242424231000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
42425A424A734A527B4A527B5263845A638463738C6B738C737B94737B947B8494
737B947B84947B8494848C9C848C9C8C94A58C94A5949CA58C949C8C949C8C949C949C9C8C9494\
949C948C948C949C8C8C94848C947B8C94738C947384846B8484637B845A7B845A
6B6B4A635A42000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003129314231394231394A39424A3942
5A424A5A42425A424A5A4242634A4A634A4A635252634A4A6B52526B52526B52526B524A735A52\
6B5A52735A526B5A5273635A6B5A526B5A526B5A525A4242635A52635A525A5A4A
635A525A5A4A52524A4A4A424A4A4A394A42000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000031314231394231394A
424A52424A524A525A4A525A525A6B4A5A6352636B52636B5A6B735A63735A6B735A6373636B7B\
636B736B737B636B736B737B6B737B7373846B737B737B84737B8473848C737B84
7B84947B848C7B84947B848C8484947B7B8C5239394A31395A42425A39425A424A5A4242634A52\
634A4A6B52526B4A5273525A73525A735A5A73525A7B5A637B5A5A7B5A637B5A5A
8463637B63638463637B6363846B6B846B6B846B6B846B6B8C73738C6B6B8C73738C6B6B947373\
8C7373947B738C7373947B7B947B7B947B7B947B7B94847B947B7B94847B94847B
94848494847B94847B94847B948C8494847B94848494847B948C84948484948C8494847B9C8C84\
948C84948C849484849C8C84948C84948C84948C849C8C8C948C84948C848C8C84
948C84948C84948C848C847B948C848C8C848C8C848C847B8C8C848C847B8C8C8484847B8C8C84\
84847B848C8484847B848C8484847B848C7B7B847B7B635284847B84847B7B847B
848C847B847B7B847B737B737B847B737B73738473737B73737B736B7B736B7B736B736B6B7B73\
6B736B6B737363736B6B737363736B63736B5A6B6B63736B5A6B6B5A736B5A6B6B
5A736B526B63526B63526363526B634A63634A63634A5A5A4A5A5A425A5A425A5A39525242525A\
395252395252000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000031394A394A52\
394A5242525A395252425A5A425A5A425A5A425A524A5A5A4A5A5A4A5A5A4A5A5A
52635A4A5A5A52635A4A5A5A52635A52635A52635A5263525A635A526352526352525A525A6352\
525A4A525A4A525A4A525A4A4A52394A5239000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
393952424A73424A73525A7B525A845A6384636B8C6B73946B7B94737B94737B94
7B849C7B84947B8C9C7B8C9C8494A584949C8C94A58C949C8C9CA58C949C8C949C8C9494949C9C\
8C9494949C8C8C9484949C848C947B8C947B848C6B848C6B8484637B8463737B52
7373525A5A42000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003129313931314231394A39424A3939524242
5242425A424A5A4242634A4A634A4A635252634A4A6B52526B524A6B5A526B524A6B5A526B5A52\
735A526B5A52735A526B5A526B6352635A4A6B5A52635A52635A525A524A635A52
5A5A4A5A5A4A4A4A42424A42394239000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003139423139424A3931394A5242525A424A5A4A5A634A5A6352636B52636B5A63735A6373\
636B735A6373636B73636B736B737B636B7B6B737B6B737B737B846B7384737B8C
737B8C7B8494737B8C4A31394A39395239425239425A42425A424A634A4A63424A6B4A526B4A52\
73525A6B5252735A5A735A5A7B5A63735A5A7B5A637B5A638463637B5A63846363
846363846B6B84636B846B6B846B6B8C7373846B6B8C73738C73739473738C73739473738C7373\
947B7B8C7B73947B7B947B7B94847B8C7B7B94847B94847B94848494847B94847B
94847B948C8494847B94848494847B948C8494847B948C849484849C8C84948484948C84948C84\
9C8C8C948C84948C84948C84948C8C948C84948C84948C84948C848C8484948C84
8C8C84948C848C847B8C8C848C84848C8C8484847B8C8C8484847B8C8C8484847B848C8484847B\
848C847B847B84847B84847B848C847B847B7B847B7B847B7B847B737B7B7B847B
737B7B7B847B737B73737B736B7B73737B736B736B6B7B736B736B6B7B7363736B63737363736B\
6373735A6B6B63736B5A6B6B5A736B526B635A6B6B526B63526B634A5A5A4A5A5A
425A5A4A5A5A42525242525A39525239525A314A52394A52000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000031424A39424A424A52394A52425A5A42525A425A5A425252425A5A
425A5A4A5A5A525A5A4A5A5A4A5A5A52635A4A5A5252635A52635252635A526352526352525A52\
526352525A4A525A4A4A52425252424A4A31393931000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
42425242426B4A52734A527B526384526384636B8C63738C737B946B7394737B94
737B947B849C7B84947B8C9C848C9C8C94A584949C8C94A58C949C8C9C9C8C94948C949C8C9494\
8C9C948C948C8C9C8C8C94848C9C84848C73848C73848C6B848C6B7B7B5A7B7B5A
6B734A63634A000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003931394231394A39394A3939524242524242
5A424A5A4242634A4A634A4A634A4A634A4A6B52526B52526B52526B524A735A526B5A526B5A52\
6B5A5273635A6B5A526B63526B5A526B63526B5A526B5A52635A4A635A525A5A4A
5A5A4A4A4A424A5242000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031424A394A5A394A5A42526342525A525A6B525A6B5A6373\
5A636B636B735A6373636B73636B736B737B6B737B6B73846B737B737B8C6B7384
4A39394A31395242425239425A424A5A4242634A4A634A4A6B4A526B4A5273525A73525A735A5A\
735A5A7B63637B5A637B63637B5A6384636B7B636384636B846363846B6B846B6B
8C6B6B846B6B8C73738C6B738C73738C7373947B7B8C7373947B7B8C737B947B7B947B7B947B7B\
947B7B94848494847B94848494847B948C84948484948C8494847B948C84948C84
948C849484849C8C84948C84948C84948C849C8C8C948C84948C8C948C849C948C948C8C9C948C\
948C849C948C948C84948C8C8C8C8494948C8C8C84948C848C8C84948C8C8C8C84
8C8C848C84848C8C848C8C848C8C848484848C8C84848C84848C8484847B8C8C84848484848C84\
7B847B848C8484847B848C847B847B848C847B847B7B847B737B7373847B737B73
737B7B6B7B73737B7B6B7B736B7B7363736B6B7B7363736B6373735A736B6373735A736B5A736B\
5A6B6B5A736B5A6B6B526B634A5A5A4A636342525A42525A39525239525A394A52
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000314252394A52394A5242525A42525A
425A5A425A5A4A63634A5A5A4A635A4A5A5A52635A4A635A52635A525A525A635A526352526352\
525A524A5263525A4A525A4A4A52394A5239424231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4A4A31424A6B424A734A527B525A7B5A638C5A6B8C6B73946B7394737B94737B94
7384947384947B8C9C7B8C94848C9C848C9C8C94A58C94A58C94A58C949C8C9C9C8C94948C9494\
8C948C949C948C948C949C848C947B8C947B848C73848C6B7B84638484637B7B5A
7373525A6342000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003931313931314239394231395239425239425A424A
5242425A424A5A4A4A634A4A634A4A6B52526B524A6B5A526B524A6B5A526B5A52735A526B5A52\
6B5A526B5A527363526B5A526B5A526B5A526B6352635A4A635A525A5A4A635A52
4A52424A4A42000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031424A394A52394A5A425263425263\
525A6B525A6B5A63735A636B5A63735A6373636B7B636B7B6B73844A3139523942
5239395A42425A4242634A4A5A424A634A526B4A5273525A6B5252735A5A735A5A7B5A63735A5A\
7B63637B5A6384636B7B636384636B84636B846B6B84636B846B6B846B6B8C7373
846B738C73738C7373947B7B8C7373947B7B947B7B947B7B8C7B7B947B7B947B7B9484848C7B7B\
94848494847B948C8494847B948C84948C84948C84948484948C84948C84948C84
948C84948C84948C84948C8C948C84948C8C948C8494948C948C84948C8C948C8C94948C948C84\
948C8C948C84948C8C8C8C84948C848C8C84948C8C8C8C848C8C848C8C848C8C84
8484848C8C84848C848C8C8484847B848C84848484848C8484847B848C84848484848C847B847B\
848C847B847B7B848473847B73847B737B7B73847B6B7B736B7B736B7B736B7B73
6373736B7B73637373637B735A736B63736B5A736B6373735A6B6B5A736B5A6B6B5A6B6B4A635A\
4A5A5A42525A42525A314A4A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000031424A394A52394A5242525A
395252425A5A425A5A4A6363425A5A4A5A5A4A5A5A52635A4A5A5252635A525A5252635A525A52\
526352525A525263524A52424A52424A4A394A4A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4239524242634A527B4A5273525A84526384636B8C636B8C6B73946B7B94738494
737B947B84947B84947B8C9C7B8C9484949C84949C8C94A584949C8C949C8C94948C9C9C8C9494\
8C94948C948C8C9C8C8C94848C948484947B8C94737B8C6B848C6B7B84637B845A
6B735263634A000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000313131393131423139423139523942523942524242524242
5A4A4A5A424A634A4A634A4A6B52526B524A6B52526B524A6B5A526B52526B5A526B5A52735A52\
6B5A52735A526B5A526B63526B5A526B5A52635A4A6B6352635A4A5A5A4A525242
4A524A000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000314252394A5A\
424A5A4A52634A526352636B52636B5A63734A31395239425239425A42425A4242
634A4A63424A634A52634A526B525A6B525273525A73525A7B5A637B5A637B63637B5A6384636B\
7B636384636B84636B8C6B73846B6B846B73846B6B8C73738C6B738C73738C7373
947B7B8C737B947B7B8C737B947B84947B7B947B7B947B7B94848494847B94848494847B948C84\
948484948C849484849C8C84948C84948C84948C849C8C8C948C84948C84948C84
9C948C948C84948C8C948C849C948C948C8C94948C948C849C948C948C8C948C8C8C8C8494948C\
8C8C84948C8C8C8C84948C8C8C8C848C8C848C84848C8C8C8C8C848C8C84848484
8C8C84848C84848C84848484848C84848484848C847B847B848C847B84847B8C847B847B7B8484\
73847B73847B737B7373847B6B7B736B7B736B73736B7B7363737363737363736B
63737363736B6373735A6B6B5A73735A6B6B5A6B6B4A6363526363425A5A42525A394A52000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000313142394231394A5A394A52
42525A42525A42525A42525A4A63634A5A5A4A5A5A4A5A5A52635A4A5A5A52635A4A5A5252635A\
525A52526352525A4A525A4A4A52424A5239424A31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
393952424263424A734A527B525A7B5A63845A638463738C6B7394737B946B7B94
73849C7384947B84947B8494848C9C848C9C84949C84949C8C9CA58C949C8C949C8494948C9C94\
8C94948C948C8C94848C9C8C8C947B8C947B848C73848C737B846B7B8463737B5A
737B5A313142000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003131314231394231314A39394A39395242425239425A4242
5A4242634A4A5A4A42634A4A634A4A6B525263524A6B52526B524A6B5A526B524A6B5A526B5A52\
7363526B5A526B5A526B5A526B5A52635A4A635A52635A4A5A5A4A5252424A5242
424A42000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
314252314252394A5A4A31395239425239395239425A42425A424A5A424A634A52
634A526B52526B4A5273525A73525A7B5A63735A5A7B5A637B5A637B636B7B5A6384636B84636B\
846B6B84636B8C6B738C6B6B8C7373846B738C73738C737394737B8C7373947B7B
947B7B947B848C7B7B947B7B947B7B9484848C7B7B94848494847B948C8494847B948C84948484\
948C84948C84948C84948C84948C8C948C84948C84948C849C948C948C8494948C
948C849C948C948C8494948C948C8C9C948C948C84948C8C948C8C94948C8C8C848C8C8C8C8C84\
8C8C8C8C8C848C8C848C8C848C8C8C8484848C8C84848C848C8C84848484848C84
848484848C847B847B848C847B8484848C847B847B7B84847B847B7B8484737B7B73847B6B7B73\
737B7B6B73736B7B736373736B7B7363736B63737363736B6373735A6B6B5A736B
5A6B6B5A73735263635263634A5A5A425A5A394A52394A52000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039424A31394A394252
394A5242525A394A5242525A42525A4A5A5A4A5A5A4A5A5A4A5A5A52635A525A6352635A525A52\
5263524A5A4A525A524A5A4A525A4A4A4A39424231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
42395A39425A424A734A5273525A7B525A7B5A63845A6B8C6B73946B738C737B94
737B947384947384947B8C947B8C9484949C848C9C84949C84949C8C949C8494948C9494849494\
8C949484948C8C948C8C94848C9C84848C73848C737B8C6B848C6B737B5A7B845A
737B52000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003131313931393931314A39424A39395239425239425A424A5A4242
634A4A5A424263524A634A4A6B525263524A6B5A526B52526B5A526B524A735A526B5A52735A52\
6B5A527363526B5A526B5A52635A4A6B6352635A52635A5252524A52524A424A42
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4A31395239425239425A424252394263424A5A424A634A52634A526B525A6B525A
73525A73525A7B5A637B5A637B5A637B5A6384636B7B636384636B7B636B846B73846B6B8C6B73\
846B6B8C73738C6B739473738C7373947B7B8C7B7B947B7B8C7B7B948484947B7B
948484947B7B948484948484948484948484948C84948484948C84948C849C8C8C948C8494948C\
948C849C948C948C8494948C948C849C948C94948C94948C948C8C9C948C94948C
94948C94948C9C949494948C94948C8C8C8C94948C8C8C8C948C8C8C8C8494948C8C8C8C8C8C8C\
8C8C848C8C8C8C8C848C8C8C848C848C8C8C848C84848C84848484848C8C848C84
848C847B8484848C8C7B84847B8C847B847B7B848473847B73847B6B7B7373847B6B7B736B7B7B\
6373736B7B736373736373735A736B6373735A736B5A7373526B6B526B6B4A5A63
4A5A5A395252395252000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000031314239425A394252
394A5A394A5A42525A42525A425A5A42525A4A5A634A5A5A4A63634A5A5A52636352635A52635A\
525A5252635A525A52525A524A52424A5242424231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
393952424263424A6B4A527B4A527B5A63845A6384636B8C63738C737B946B7B94
738494737B947B8C9C7B84947B8C9C7B8C9484949C84949C8C949C8494948C9C9C8494948C9494\
84948C8C9C948C948C8C948C84947B8C947B848C73848C737B84637B846373845A
737B5A000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003931393931314231394A39395239425239425A42425A42425A4A4A
5A4242634A4A634A4A6B524A634A4A6B524A6B524A6B5A526B524A6B5A526B5A52735A526B5A52\
6B5A526B5A526B6352635A4A6B5A52635A4A635A525A524A52524A4A4A42000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A31395239424A3939\
5239425242425A424A5A424A634A52634A526B525A6B4A526B525A73525A7B5A63
735A5A7B5A637B5A637B636B7B5A637B636B7B636B846B6B84636B846B73846B738C73738C6B73\
8C73738C737394737B8C7373947B7B947B7B947B848C7B7B948484947B84948484
94847B948484948484948C84948484948C84948C84948C8C948C84948C8C948C849C948C948C84\
94948C948C8C9C948C948C8494948C94948C9C948C94948C94948C94948C9C9494
948C8C94948C948C8C94948C8C8C84948C8C8C8C8C94948C8C8C848C8C8C8C8C8C8C8C8C848C84\
8C8C8C848C848C8C8C848484848C84848C84848C8C7B8484848C847B8C84848C84
7B84847B848473847B7B8484737B7B73847B6B7B7B737B7B6373736B7B73637373637B735A736B\
6373735A736B5A7373526B6B526B6B4A5A634A5A63395252394A52000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039394A31394A39425A
394252424A5A424A5A42526342525A425A5A425A5A4A5A634A5A5A5263634A5A5A52635A4A5A52\
526352525A52526352525A4A525A4A4A5242424231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
42395A42425A4A4A6B4A4A734A527B525A7B5A638C5A63846373946B7394737B94
6B7B947384947384947B8C9C7B84947B8C94848C9484949C84949484949C8494948C949484948C\
8C949484948C8C948C8494848C948484947B84947B7B8C6B7B8C6B7B8463737B5A
6B734A000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003131313931314239394231394A39424A39395A42425242425A42425A4242
634A4A634A4A634A4A634A4A6B52526B524A6B52526B524A6B5A526B5A526B5A526B5A4A736352\
6B5A526B5A52635A4A6B6352635A52635A525A524A5A5A4A525242424A42000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005239395239395239425239425A424A\
5A424A634A52634A4A6B52526B52526B525A6B525A735A63735A637B5A63735A63
7B636B7B63637B636B7B636B846B7384636B846B73846B6B8C73738C6B738C73738C6B7394737B\
94737B947B7B94737B9C7B84947B849C7B849C7B84948484948484948484948484
948C8C948484948C849484849C8C8C948C84948C8C948C849C948C94948C94948C9494849C948C\
94948C94948C94948C9C9C9494948C9C948C94948C9C9C9494948C94948C8C8C8C
94948C8C8C8C948C8C8C8C8494948C8C8C8C8C8C8C848C848C8C8C848C8C8C8C8C848C848C8C8C\
848C84848C8C848484848C8C848C84848C847B8484848C8C7B84847B848473847B
7B8484737B7B737B7B6B7B736B7B7B6B7B736B7B73637373637B736373736373735A6B6B5A7373\
526B6B52636342525242525A394A52000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039394A393952394252
424A5A394A5A42525A42525A4A5A6342525A4A5A634A5A5A5263634A5A5A5263634A5A5A52635A\
525A525263524A5A4A526352525A4A4A5242000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
393952424263424A6B4A527B4A5A7B525A84526384636B8C636B8C6B7B946B7394
738494737B947384947384947B8C9C7B8C9484949C848C9C8C949C8494948494948494948C9C94\
84948C84948C8494848C9C8C84948484947B7B8C73848C73424A6B7B84636B7352
6B7352000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3131313129313931394231394A39394A31395239425239425A424A5A42425A4A4A
5A4A42634A4A634A4A6B524A6B524A6B525263524A6B52526B524A6B5A526B524A6B5A526B5A52\
6B5A52635A4A6B5A52635A4A6B5A52635A4A5A5A4A525242525242000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003931395239394A31395239425239425A424A5A4242634A4A634A4A\
6B4A52634A526B525A6B525A735A6373525A735A63735A637B636B735A637B636B
7B636B846B6B7B636B846B73846B6B8C6B73846B738C73738C6B7394737B8C737394737B94737B\
947B8494737B9C7B849C7B849C8484947B849C8484948484948C8C948484948C84
948C84948C8C948C84948C8C948C849C948C948C8494948C94948C9C948C94948C94948C94948C\
9C9C9494948C94948C94948C9C9C9494948C94948C94948C94948C8C8C8C8C8C8C
8C8C8C8C8C8C8C8C848C8C8C8C8C8C8C8C8C848484848C8C848C84848C8C848484848C8C848C84\
848C8C7B8484848C847B84847B8C8473847B7B848473847B7384846B7B7B737B7B
6B7B736B7B7B6373736373736373736373735A6B6B5A73735A6B6B5A73735263634A5A5A394A52\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039394A31394A394252
394A5A424A5A424A5A42526342525A4A5A6342525A4A5A5A4A5A5A5263634A5A5A52635A52635A\
52635A4A5A52525A52525A4A525A4A424A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
42395A42425A4A4A6B4A5273525A844A5A7B5A63845A638463738C63738C6B7B94
6B7B94738494737B947384947384947B8C9C7B8C9484949C84949C84949C84949484949484948C\
84948C84948484948C8494848C9484848C7B848C737B8C737B8C6B737B5A6B7B52
636B4A000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3129313931394231394A39394A39395242425239425A42425A4242634A4A634A42
634A4A634A4A6B52526B524A6B524A63524A6B5A526B524A6B5A526B524A735A526B5A526B5A52\
6B5A4A6B6352635A52635A52635A4A635A525A524A52524A000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003131314239394A31395239425239395A424A5A424A634A4A5A424A6B5252634A526B525A\
6B525A735A63735A637B5A63735A637B636B7B63637B636B7B636B846B7384636B
846B7384636B8C6B73846B738C737B8C6B7394737B8C737B94737B8C737B947B84947B849C7B84\
947B849C848C9C84849C848C9484849C8C8C9C848C948C8C948C849C948C948C8C
94948C948C8C9C948C94948C94948C94948C9C9C9494948C9C9C9494948C9C9C9494948C9C9C94\
94948C9C9C9494948C9494948C948C9494948C948C94948C8C8C8C94948C8C8C8C
8C8C8C8C8C8C8C948C8C8C8C8C8C8C848C848C8C8C848C8C848C8C848C847B5A5A848C8C848C8C\
7B8484848C8C7B84847B848473847B738484737B7B737B7B6B7B736B7B7B637B73
637B73637373637B735A73735A73735A6B6B5A73734A636342525A000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039394A42425A39425A
424A5A394A5A4252634252634A5A63425A634A63634A5A634A63634A5A5A526363526363526363\
525A5A52635A525A52525A524A524A4A4A42000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
393952424A63424A6B4A527B4A527B5263845263845A6B8C5A6B8C6B7B946B7B94
6B7B946B7B9473849C7384947B8C947B8C9484949C84949484949C849494849494849494849494\
84948C8C948C84948C849484849484849484848C7B848C737384637384636B7352
6B734A000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000312931\
3931313931314239394231394A39425239425A42425239425A42425A4242634A4A
5A4A42634A4A634A4A6B524A63524A6B524A6B524A6B5A526B524A6B5A526B5A52735A526B5A4A\
6B5A52635A4A6B5A52635A4A635A4A5A524A52524A4A4A42000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000423939
4231395231395239395A39425A394263424A5A424A634A52634A526B525A6B525A735A6373525A\
735A63735A637B636B735A637B636B7B636B84636B7B636B846B73846373846B73
846B738C6B738C6B738C737B8C737394737B8C737B947B8494737B947B84947B849C848C947B84\
9C848C9484849C848C9484849C848C94848C9C8C8C94848494948C948C8C9C948C
948C8C94948C94948C9C9C9494948C949C9494948C9C9C9494948C949C94949C8C9C9C9494948C\
94949494948C9494948C8C8C94948C8C8C8C94948C8C8C8C8C8C8C8C8C8C8C8C8C
848C848C8C8C848C8C8C8C8C848484848C8C848C8C848C8C7B8484848C8C7B84847B8C8C738484\
73848473847B7384846B7B7B6B7B7B6B7B7B6B7B7B637373637373637373637373
5A6B6B5A73735A6B6B526B6B42525A394A52000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000031293939394A394252424A5A
394A5A4252634252634A526342525A4A5A634A5A635263634A5A5A52636352635A52636352635A\
52526352635A5263524A5A4A525A4A424A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
42395A4242634A4A6B4A5273525A7B525A845A63845A6384636B8C63738C6B7B94
6B7B947384947384945A737B7384947B8C947B8C9484949C7B8C9484949484949484949484948C\
52637B84948C84948C7B8C7B849484848C7B84947B7B8C6B7384636B7B5A6B7B52
636B42000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393139\
3931314231394231395239425239425239425239425A424A5A4242634A4A5A4242
6B524A634A4A6B524A634A4A6B52526B524A6B52526B524A735A526B5A526B5A526B5A4A6B5A52\
6B5A526B5A52635A4A6B5A525A5A4A5A524A4A4A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000393939423139523139523139
5A39425A39425A424A5A424A634A52634A526B525A6B525A735A63735263735A63735A637B636B\
735A637B636B7B5A6B8463737B636B846B737B636B846B73846B73846B73846B73
8C737B8C737B8C737B8C737B947B84947B84947B84947B849C848C947B849C848C947B849C848C\
94848C9C848C94848C9C8C8C948C8C9C8C8C948C8C9C8C8C94948C94948C94948C
9C9C9494948C9C9C9494948C9C9C94949C949C9C94949C8C9C9C94949C94949C9494948C949C94\
8C948C94948C8C8C8C9494948C8C8C8C8C8C848C8C8C8C8C848C8C848C8C848C84
848C8C848C8C848C8C7B8484848C8C7B8C8C848C8C7B84847B8C8C7B84847384846B7B7B738484\
6B7B7B6B7B7B6373736B7B7B6373736373735A73736373735A73735A7373526B6B
526363425252000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000039314A39394A42425A42425A
4A52634252634A52634252634A5A6B4A5A63525A6363636B636B7352636352636352635A5A6363\
52635A52635A4A5A52525A524A524A424A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39395A424263424A6B52527B4A5A7B525A845263845A6B8C5A6B8C63738C63738C
737B946B84947384947384947B8C9C738C947B8C947B8C9484949C7B9494849494849494849C94\
84948C84948C7B8C848494847B8C7B84947B7B8C737B8C73738463737B5A63734A
63734A000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000313131313131\
4231395252424A39394A31395239425239395242425A42425A42425A4242634A4A
634A4A6B524A634A426B524A6B524A6B525263524A6B524A6B524A6B5A526B524A6B5A526B5A4A\
6B5A5263524A635A4A635A4A5A5A4A524A424A4A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003931314A31395231395239425239395A3942
5A394A63424A5A424A634A52634A526B525A6B525A735A63735A637B5A6B735A637B636B7B636B\
7B636B7B636B7B63737B636B846B73846B73846B73846B738C737B846B738C737B
8C737B947B848C737B947B84947B849C848C947B849484849484849C848C94848494848C94848C\
9C8C8C94848C948C8C948C8C9C8C8C948C8C948C8C948C8C9C9C9494948C949C94
949C8C9C9C94949C8C9C9C94949C949C9C9494948C949C9494948C949C948C948C8C948C8C948C\
8C94948C8C8C8C8C8C848C8C8C8C8C84848C848C8C848C8C848C8C7B8484848C8C
7B8C8C848C8C7B84847B8C8C7B84847B8C8C738484738484737B7B7384846B7B7B6B7B7B637B7B\
6B7B7B6373736373735A73735A7373526B6B5A73735263634A5A5A39424A000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000313142424252424252424A5A
4A4A634A52634A52634A5A634A5A63525A6B4A5A6352636B526363526363525A6352636352635A\
5A6363525A52525A524A5A4A525A4A292129000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
42395A4242634A4A6B4A5273525A844A5A7B5263845A6384636B8C5A6B8C6B7394
6B7B9473849C6B7B947384947384947B8C947384947B8C947B8C948494947B94948494947B948C\
84948C7B8C847B94847B8C7B84947B7B8C737B8C7373846B7384636B7352637352
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000313131525242\
3931394A39424A39395239425239395A42425242425A42425A4242634A4A634A42
634A4A634A426B524A63524A6B524A63524A6B5A526B524A6B5A526B5A4A6B5A526B5A4A6B5A52\
635A4A6B5A52635A4A635A4A52524252524A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003931394231394A31395239425239425A39425A394263424A63424A
634A52634A526B525A6B525A7352636B525A7B636B735A637B636B7B5A6B7B63737B636B846373\
7B636B846B73846B73846B7B846B738C737B8C737B8C737B8C737B947B848C7B84
947B84947B849C848C947B8494848C947B849C848C94848C94848C94848C9C8C94948C8C948C94\
948C8C9C9494948C8C949494948C8C9C9494948C8C949494948C8C9C9C94949C94
9C9C94949C949CA594949C94949C9494948C949C948C94949494948C948C9494948C8C8C8C9494\
848C8C8C9494848C8C8C8C8C848C8C8C8C8C848C8C848C8C7B8C8C848C8C7B8C8C
848C8C7B84847B8C8C7B84847B8484737B7B7384846B7B7B6B7B7B63737B6B7B7B637373637373\
5A73736373735A73735A6B734A5A5A42525A634A425A4A425A42395A4242523939
523931000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000039394A4239524A4A634A4A63
4A52634A5263525A6B4A5A63525A6B4A5A6352636B52636B52636B5263635A63635263635A6363\
52635A5A635A525A52525A4A4A5242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393942\
39395A424A63424A6B4A527B4A5A7B5263845263845A6B8C5A6B8C63739463738C
6B7B946B7B94738494738494738C947384947B8C947B8C947B94947B8C9484949484949484948C\
7B8C8484948C7B8C847B94847B8C7B7B947B7B8C73738C6B6B7B5A6B7B5A63734A
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000313131393139393131\
4231394231395239424A31395239425239395A42425239395A42425A4242634A4A
634A42634A4A634A426B524A634A426B524A6B524A6B5A526B524A6B5A526B5A4A6B5A52635A4A\
635A4A635A4A635A525A52425252424A4A42000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3931394231395239395231395239425A39425A424A5A424A634252634A526B525A
634A526B525A6B525A735A63735A637B636B7B636B7B636B7B636B7B63737B6373846B737B6373\
846B73846B738C737B846B738C737B8C737B947B848C737B947B84947B84947B8C
947B8494848C947B8C94848C94848C94848C94848C948C9494848C948C94948C8C948C94948C8C\
948C94948C8C949494948C8C94949494949494949494948C949494949C949CA594
949C94949C94949C94949C948C948C9494948C948C9494948C8C8C8C94948C8C8C8C9494848C8C\
848C8C848C8C848C8C84848C848C8C848C8C848C8C7B84847B8C8C7B8C8C7B8C8C
7384847384847384847384846B7B7B6B7B7B637B7B637B7B5A73736373735A73735A7373526B6B\
52636B42525A394A526B52526B5A5A6B52526B5252634A4A634A4A5A4242634A42
5239395A4239523931523931000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000031313939394A42425A4A4A5A4A5263
4A5263525A6B525A6B52636B525A6352636B52636B5A6B6B5263635A63635263635A6B6352635A\
5A636352635A5263524A524A4A5242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393942\
42395A4242634A52734A5273525A7B525A845263845263845A6B8C63738C6B7B94
637B8C6B84946B84947384946B848C738C94738C947B8C94738C8C7B94947B94948494947B8C8C\
7B948C7B8C847B94847B8C7B7B8C7B7B8C737B8C737384636B7B6363735263734A
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000313131312931423139\
4231394A39394A31395239425239395239425239395A42425A4242634A42634242
634A4A634A4263524A634A426B524A6B524A6B524A6B524A6B5A526B5A4A6B5A5263524A6B5A52\
635A4A635A4A5A524A5A5A4A4A5242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000393139423131523939\
5231395239425239425A424A5A424A6342525A424A6B4A5A634A526B525A6B525A
735A63735A63735A6B735A6B7B63737B636B7B63737B636B846B73846B73846B73846B738C737B\
846B7B8C737B84737B8C73848C7384947B848C7384947B8C947B84947B8C947B84
94848C94848C94848C94848C9C8C9494848C948C94948C8C948C94948C94948C94948C8C949494\
948C94949494948C94949494949494949494949494949C9C949494949C94949C94
949C94949494949C948C948C9494948C948C8C9494848C8C8C9494848C8C848C8C848C8C848C94\
848C8C848C8C7B848C848C8C7B8C8C7B8C8C7B84847B8C8C7384847384846B7B7B
7384846B7B7B6B7B7B637373637B7B6373735A73735A6B6B5A7373526B6B4A5A63394A5273636B\
6B5A63735A636B5A5A735A636B5A5A6B5A5A6B52526B5A5A6B5252635252634A4A
634A4A5A42425A42425239395A42395239314A3931000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003131424242524A4A5A4A52634A5263
525A6B525A6B525A6B525A6B5A636B52636B5A636B52636B5A6B6B5A636B5A6B6B5263635A6B63\
5A63635A635A525A52525A524A5242000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000042394A\
39395A424263424A6B52527B4A5A7B525A845263845A6B8C5A6B8C63738C63738C
6B7B946B7B946B84946B7B94738C94738494738C94738C8C7B94947B8C947B94947B8C8C7B948C\
7B8C847B9484738C7B7B947B7B8C7B7B8C7373846B73846B6B7B5A6373525A6B4A
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000313131312931393131393131\
4A31394A31395239394A39395239425239395A42425A42395A42425A4242634A42
634A42634A4A634A4263524A634A426B524A6352426B524A6B524A6B5A5263524A6B5A4A635A4A\
635A4A5A524A5A524A5252424A5242000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003931394A31394A3139523942523942\
5A424A5239425A424A5A424A634A52634A526B525A6B525A7352636B5263735A6B
735A6B7B636B735A6B7B63737B636B846B737B6373846B73846B7384737B846B738C737B84737B\
8C73848C737B8C73848C7384947B8C947B84947B8C947B8494848C947B8494848C
94848C948C9494848C948C94948C8C948C94948C8C948C94948C949494948C8C8C949494949494\
9494948C9494949494949494949C9C8C9494949C94949C94949C948C94948C9C94
8C94948C9C948C948C8C94948C948C8C9494848C8C848C8C848C8C848C8C7B848C848C8C7B8C8C\
848C8C7B848C7B8C8C7B848C7B8C8C738484738484737B847384846B7B7B6B7B7B
63737B637B7B5A73735A73735A6B735A7373526B6B42313139424A736B7373636B73636B73636B\
73636B6B5A63735A636B5A63735A636B525A6B5A5A6B525A6B5A5A6B52526B5252
6352526B52525A4A4A634A4A5A42425A4242523939523939523931523931000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000039314A4242524A4A634A4A63525263
52526B5A5A6B525A6B5A636B52636B5A637352636B5A6B6B5A6B6B5A6B6B5263635A6B635A6363\
5A6B635263525263524A524A4A5239000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039394239394A\
42395A4242634A4A6B4A5273525A844A5A7B5263845263845A6B8C5A6B8C637394
6B7B946B84946B7B8C6B84946B8494738C9473848C738C94738C8C7B94947B8C8C7B948C7B8C8C\
7B948C738C7B7B8C7B738C7B7B8C7B7384737384736B7B636B7B5A636B4A5A6B4A
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000312931393139393131423139\
4231395239425239395239425239395A42425A42425A42425A4239634A42634A42
634A42634A426B524A634A426B524A6352426B5A4A6B524A6B5A4A63524A6B5A52635A4A635A4A\
63524A635A4A5A524A525242000000000000000000000000000000000000000000
0000000000000000000000000000003931394239394A31395239425239395A424A5A42425A424A\
5A424A634A52634A52634A5A634A5A6B52636B5263735A63735A637B636B73636B
7B637373636B7B6B737B6373846B7B846B738C737B8C6B7B8C737B84737B8C73848C73848C7B84\
8C7384947B8C947B84947B8C947B8494848C94848C94848C94848C948C9494848C
948C94948C8C948C94948C94949494948C94949494948C949494948C8C9494949C949494949494\
8C9494949C9C949494949C9C8C9494949C9C8C9494949C948C9494949C948C9494
8C94948C948C8C9494848C8C8C9494848C8C8C9494848C8C848C8C7B8C8C848C947B8C8C848C8C\
7B848C7B8C8C7B848C73848C737B847384846B7B7B6B7B7B63737B637B7B637373
6373735A6B735A7373526B6B42525A73736B73736B63635A5A5A4A7363737B6B73736373736373\
73636B73636B73636B73636B6B5A6373636B6B5A63735A636B5A5A735A636B5A5A
6B5A5A6B52526B5A5A635252635252634A4A634A4A5A42425A42425239395A4242523939523939\
4A3131000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003931394A524A4A425A4A4A5A525A6B525A6B
5A5A6B525A6B5A63735A63735A63735A636B636B735A6B73636B735A6B6B636B6B5A6B635A6B63\
5A635A5A6B63525A524A5242000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000393942423952\
39395A424A6B4A4A6B4A527B4A5A7B5263845263845A6B8C5A6B8463739463738C
6B7B946B7B946B84946B84947384946B848C738C94738C947B8C94738C8C7B94947B8C8C7B948C\
738C847B9484738C7B7B8C7B738C737B8C7373846B6B846363735A6373525A6B42
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000424239313131393131423139423131\
4A31394A31395239424A39395239395239395A42425A39395A42425A4242634A42
5A4242634A42634A4263524A634A426B524A6352426B5A4A6352426B5A4A63524A6B5A4A63524A\
635A4A5A524A5A524A4A4A39000000000000000000000000000000000000000000
0000000000003931394239394231395231395239395239425239425A424A5A424A634A525A4252\
634A5A634A5A6B52636B525A6B5A63735A6373636B735A6B73636B73636B7B6373
7363737B6B737B6B73846B7B846B738C737B8C737B9473848C737B8C73848C7384947B8C8C7B84\
947B8C947B8C94848C8C7B8C94848C94848C948C948C848C948C94948C8C948C94
8C8C8C948C94948C949494948C8C9494949494949494949C8C94949494948C9494949C9C8C9494\
949C9C8C9C94949C9C8C94948C94948C94948C94948C948C8C94948494948C9494
848C8C849494848C8C849494848C8C848C8C7B8C8C848C947B848C7B8C8C7B8C8C7B8C8C738484\
7384847384847384846B7B7B6B7B7B637B7B637B7B5A73735A73735A6B735A7373
526B6B42525A7B7B7B7B847B73736B737B6B7373636B6B5A5A5A42736B73736B737B6B73736373\
736B7373636B73637373636B73636B735A6373636B6B5A63735A636B5A63735A63
6B525A6B5A5A6B525A6B5A5A6352526B52526352526B5252634A4A634A4A5A42425A4242523939\
5242425239395239394A3131000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003931394A42524242525252635252635A5A6B
5A5A6B5A63735A6373636B735A636B636B735A6B6B636B735A6B6B636B6B5A6B6B636B6B5A6363\
5A6B635A635A525A52424A39000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000042424A393952\
393963425A5A4A52734A5273525A7B525A845263845263845A6B8C5A738C637B94
637B8C6B7B946B7B946B84946B848C73849473848C738C94738C8C7B8C947B8C8C7B948C738C84\
738C84738C7B7B8C84738473738C7373847373846B637B5A63735A5A6B4A5A6B42
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
0000000000000000000000000000000000000000000000003131393131313931393931314A3139\
4A31395239394A31395239425239396B5A5263524A6B5A5263524A634A425A4239
634A42634A42634A42634A426B524A6352426B524A6352426B5A4A63524A635A4A63524A635A4A\
635A4A5A524A4A4A42000000000000000000000000000000000000000000000000
3939394231394A31394A31395239425239425A424A52394A634A525A4252634A52634A526B525A\
6B525A6B52636B5263735A6B735A6B73636B735A6B7B63737363737B63737B6373
846B7B7B6B7B846B7B846B7B8C73848C737B8C73848C7384947B8C8C7B84947B8C8C7B8494848C\
947B8C94848C8C848C94849494848C948C948C848C948C94948C94948C948C8C94
94949C94949494949C8C8C9494949C94949494949C8C9494949C9C8C949C949C9C8C9494949C9C\
8C9C9C8C9C9C8C94948C9C948C94948C94948494948C9494849494849494848C8C
849494848C8C848C947B8C8C848C947B8C8C7B8C8C7B848C7B8C8C73848C73848C6B7B84738484\
6B7B7B6B7B7B63737B637B7B6373735A73735A6B735A7373526B6B4A5A5A7B8484
848C8C8484847B847B7B7B737B7B7373736B73736B636352635A4A736B73736B73736373736B73\
736B73736B7373636B736B7373636B73636B6B636B73636B6B5A6373636B6B5A63
7363636B5A636B5A636B525A6B5A5A6B525A6B525A6352526B525A6352526B5252634A4A634A4A\
5A424A5A424A5239425A42425239395239394A3131000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000042394A424252524A635252635A5A6B5A5A6B
6363735A6373636B735A6373636B73636B73636B73636B6B637373636B6B636B6B5A6B63636B63\
5A635A5A635A4A524A4A5242000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039394A39394A424263\
39396342426B424A6B525A7B4A5A7B5263845263845A6B8C5A6B8463738C63738C
6B7B946B7B946B84946B7B8C7384946B848C738C946B848C738C94738C8C7B948C738C847B948C\
738C84738C8473847B738C7B738473738C736B7B636B7B636373525A6B4A526342
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
0000000000000000000000000000000000000000000000003131313931393929314231395A524A\
4A39394A31315231394A31395239395239395A39395A39395A42425A42395A4242
5A4239634A425A4A39634A42634A426B5242634A4263524A6352426B5A4A635242635A4A63524A\
635A4A5252424A5242000000000000000000000000000000393139393139423939
4A39425239424A394252394252394A5A42525A424A634A52634A5263525A634A5A6B52636B5263\
735A6B6B5A63735A6B735A6B7B637373636B7B63737363737B6B7B7B63737B6B7B
846B7B847384846B7B8C73848C7384947B848C7384947B8C947B8C94848C8C7B8C94848C8C848C\
9484948C848C9484948C8494948C948C848C948C948C8C9494949C8C8C9494949C
8C949494949C8C949494949C8C949C949C9C8C9494949C9C8C9C9C949C9C8C94948C9C9C8C9494\
8C9C9C8494948C94948494948C9494848C8C849494848C8C8494947B8C8C848C94
7B8C8C848C947B848C7B8C8C7B848C7B8C8C7384847384846B7B847384846B7B7B6B7B7B63737B\
637B7B5A73735A73735A6B735A7373526B6B4A5A5A7B848C84848C7B7B8484848C
7B84848484847B7B7B7B7B7B737B737B7B736B73636B6B5A5A5A4A736B7B6B6373736B73736373\
736B736B6373736B73736373736B7373636B73637373636B73636B6B5A6373636B
6B5A6373636B6B5A636B5A636B5A5A6B5A636B525A6B525A6B525A6B525A634A526B5252634A52\
6B5252634A4A634A4A5A424A5A4A4A5239425A42425239395239394A31314A3131
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000031313942394A4A425A524A5A5A526B5A5A6B5A636B
63636B636B73636373636B73636B73637373636B6B637373636B736373735A6B63636B635A6B63\
636B63525A524A524A000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000042394239394A423952393963\
42426B39426B424A734A527B525A844A5A7B5263845263845A6B8C5A6B8463738C
637B946B7B94637B8C6B84946B848C6B84946B848C738C8C738C8C738C94738C8C738C8C738C84\
738C846B847B738C7B6B8473738C736B846B6B846B63735A637352526B4A313142
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
0000000000000000000000000000000000000000003131313931393931314239394231394A3139\
4A31395239394A31395239394A31395239395239395A42395A39396342425A4239
6342425A4239634A42634A42635242634A426B524A63524263524A635242635A4A63524A635A4A\
5A524252524A4A4A420000000000000000003931394239424231394A3942523942
5A424A5A394A5A424A5A424A634A525A425263525A634A5A6B52636B5263735A6B735A6B735A6B\
735A6B7B63737363737B63737363737B6B7B7B6B7B7B6B7B7B6B7B847384846B7B
8C73848473848C7B8C8C7384947B8C947B849C848C9C7B8C9C84949C848C9484948C848C948C94\
8C848C948C948C8C94948C948C8C9494949C8C8C9494949C8C8C9494949C8C949C
94949C8C949C949C9C8C949C949C9C8C949C949CA58C9C9C8C9C9C8C94948C9C9C8C94948C9494\
8494948C9494849494849494848C8C849494848C8C848C947B8C8C848C947B8C8C
7B8C947B848C7B8C9473848C73848C6B7B847384846B7B846B7B8463737B637B7B63737B63737B\
5A6B735A7373526B734A5A5A737B8C8484947B84948484947B848C84848C84848C
84848C7B84848484847B7B7B7B7B7373736B737B6B6B6B5A63634A524A39736B7B736B7B736B7B\
6B6373736B7B736B73736B73736373736B73736373736B7373636B736B7373636B
73636B6B5A6373636B735A637363636B5A637363636B5A636B5A636B525A6B5A636B525A6B525A\
6352526B525A6352526B5252634A4A634A52634A4A63424A5A42425A4242523939
5239394A3131523939000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003931394A425A4A4A5A5252635A5263636373636373
636B736363736B737B6B6B736B737B636B736B737B6B73736B7373636B6B63736B636B63636B63\
5A635A5A63524A5242000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000042394239394242425242395A42426B\
394263424A73424A734A5A7B4A5A7B5263845263845A6B8C5A6B8463738C63738C
6B7B94637B8C6B84946B848C6B84946B848C738C946B8C8C738C8C738C8C7B9494738C8C738C84\
6B8C7B738C846B8C7B738C7B6B8473738C736B7B63637B5A5A6B525A6B4A313139
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000313139312931393139423131423139423131\
4A31394A31395239394A31315239395231395239395239395A39395A39395A4242
5A42395A42425A4239634A425A4A39635242635242635242635242635242635242635A4A5A5242\
5A52424A4A420000003131313931394231394A39424A314252394A52394A5A424A
5A424A5A42525A4252634A5A634A5A63525A63525A6B5A636B5263735A6B735A6B736373735A6B\
7363737363737B6B7B7363737B6B7B7B6B7B7B6B7B7B6B7B8473848473848C7384
8473848C7B8C8C7B84947B8C947B8C94848C9C848C9C84949C848C9C84948C8494948C948C8494\
948C948C8C9494949C8C8C9494949C8C8C9494949C8C949494949C8C949C949C9C
8C949C8C9C9C8C9C9C949CA58C949C8C9C9C8C949C8C9C9C8494948C94948494948C9494848C8C\
849494848C948494947B8C8C848C947B8C8C848C947B848C7B8C8C7B848C7B8C8C
73848C73848C7384847384846B7B7B6B7B84637B7B637B7B5A73735A73735A6B735A7373526B6B\
4A5A5A6B6B8473738C73738C7B7B947B7B947B84947B7B8C7B848C7B848C84848C
7B84848484847B847B7B847B737B73737B6B73736B6B6B635A5A4A5252396B6B7B736B7B6B6373\
736B7B6B6373736B7B6B6373736B73736373736B73736373736373736373736B73
73636B73637373636B73636B6B5A6373636B6B5A6373636B6B5A636B5A636B5A5A6B5A636B525A\
6B525A6B525A6B525A6352526B5252634A526B5252634A4A634A4A5A424A634A4A
5A42425A42425239395239424A31315231394A3131000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000042394A4A4252524A63525263635A6B6363736B6B7B
6363736B6B7B6B6B7B6B737B6B6B736B737B6B73736B737B6373736B7373636B6B636B6B5A635A\
5A635A525A4A4A5242000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000042394239314242394A423952424263393963\
42426B424A6B4A52734A527B525A7B525A845263845263845A6B8C5A6B8C637B94
63738C637B94637B8C6B8494637B8C6B848C6B848C738C946B848C738C8C738C8C738C8C6B8484\
738C846B847B738C7B6B84736B84736B846B6B84635A73525A7352526342000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
0000000000000000000000000000000000003131313129313931393931314231394231314A3139\
4A31394A31394A31315239394A31395239394A31315239395239395A3939523939
5A42395A42395A42395A4239634A42634A42635242634A4263524A63524263524A5A52425A5A4A\
5252423931393931394239424A39425239425239425A424A5A424A5A42525A4252
634A5A634A5A63525A634A5A6B52636B5263735A6B6B5A6B736373735A6B73637373636B7B6373\
7363737B6B7B7363737B6B7B7B6B7B7B6B847B6B7B8473848473848C73848C7384
947B8C947B8C947B8C947B8C9C84949C84949C84949C84949C8C949C84949C8C948C8C9494949C\
8C8C948C949C8C8C9494949C8C949C8C949C8C949C949CA58C949C8C9C9C8C949C
949CA58C9C9C8C9C9C8C949C8C9C9C8C949C8C949C8494948C949C849494849494848C8C849494\
848C94848C947B8C8C848C947B8C8C7B8C8C73848C7B8C9473848C73848C738484
7384846B7B846B7B8463737B637B7B5A737B5A737B5A6B735A7373526B734A5A5A5A637B6B6B84\
6B6B846B73846B7384737B94737B947B84947B7B8C8484947B848C84848C7B8484
84848C8484848484847B7B737B7B7373736B73736B6B6B5A6363525252396B6B7B6B6373736B7B\
6B6B7B736B7B6B6373736B7B6B6B73736B736B6373736B7B736373736B736B6373
736B7373637373637373636B73637373636B73636B6B5A6B73636B6B5A63735A636B5A63735A63\
6B5A636B5A636B525A6B5A5A6B525A6B525A634A526B525A6352526B5252634A4A
634A52634A4A634A4A5A42425A42425239395239424A39395239394A31314A3131000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000393139423952524A5A524A5A5A5A6B635A6B636373636373
6B6B7B6B6B7B6B737B6B737B73737B6B73736B737B6B73737373736B736B6B736B636B63636B63\
5A635A525A4A394231000000000000000000000000000000000000000000000000
00000000000000000000000000000000000042394242314242395239394A42425A42396342426B\
42426B4A4A73424A73525A7B4A5A7B5263845263845A6B8C5A6B845A738C5A738C
6B7B94637B8C6B7B94637B8C6B84946B848C6B848C6B848C738C8C738C8C738C8C6B8484738C84\
6B847B6B8C7B6B84736B8C736B846B6B846B637B5A63735A5A6B4A526B42000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000312931393131393131393139392931423131423131\
4A31394231314A31394A31315239394A31315231394A3131523939523131523939
5239395A42395239395A42395A4239634A425A4A396352425A52426352425A5242393139393139\
4239424231394A39424A394252394A52394A5A424A5A424A5A4A525A4252634A5A
634A5A6B52636352636B5A6B6B5A6B735A6B6B5A6B736373735A73736373736373736373736373\
7B6B7B73637B7B6B7B7B6B7B8473847B6B7B8473848473848C7B8C8C73848C7B8C
947B8C948494947B8C9C84949C84949C849494848C9C84949484949C8C949484948C8C9C8C8C9C\
8C949C8C8C9C8C949C8C949C8C9CA58C949C8C9C9C8C9C9C8C9CA58C9C9C8C9C9C
8C9C9C8C9C9C8494948C949C8494948C949C848C94849494848C948494947B8C8C848C947B8C8C\
7B8C947B848C7B8C8C7B848C7B8C8C73848C73848C73848C73848C6B7B846B7B84
63737B637B7B5A73735A73735A6B735A7373526B6B4A5A5A42425A52526B52526B5A637B63637B\
6B6B846B6B8473738C73738C7B7B94737B8C7B84947B848C8484947B7B8C84848C
7B84848484847B7B7B7B7B7B7B7B737B7B737373636B6B63635A4A5252426B637B6B6B7B6B6B7B\
736B7B6B63736B6B7B6B6373736B7B6B6373736B736B6373736B7B6B6373736373
6B6373736B736B636B73637373636B7363736B636B73637373636B73636B6B5A63735A6B6B5A63\
735A6B6B5A636B5A636B525A6B5A636B525A6B525A6B525A6B525A634A526B5252
634A526B5252634A4A634A4A634A4A634A4A5A42425A42425A42425A42425239395239394A3939\
5239394A31314A3131000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003931394A42524A425A5A52635A5A6B6363736363736B6B7B
6B6B7B7373846B6B7B73737B6B737B73737B6B73736B73736B7373635A6B636B6B636B635A635A\
5A635A4A4A42000000000000000000000000000000000000000000000000000000
00000000000000000000000039313142394242314242395239395242425242395A42426B42426B\
424A73424A6B4A52734A527B525A840000005263845263845A6B8C5A6B8463738C
63738C637B94637B8C637B8C637B8C6B848C63848C6B848C6B848C738C8C6B84846B8C846B847B\
6B8C7B6B84736B84736B84736B8473637B63637B635A73525A6B52000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
0000000000000000000000000000000000003931393931313931393931314231394231314A3139\
4231314A31394A31314A31394A31315239394A3131523939523131523939523939
5239395239395A42395A42395A42395A4239635242634A426352423931394239424231394A3942\
4A314252394A52394A5A424A52394A5A42525A4252634A5A5A4A5A635263635263
6B5263635263735A6B6B5A6B7363736B5A6B73637373637373637B7363737B6B7B7B6B7B7B6B7B\
7B6B7B8473847B6B848473848473848C7B8C8C738C8C7B8C8C7B8C948494947B8C
94849494848C9C84949484949C84949484949C8C94948494948C949484949C8C9C8C949C8C949C\
8C949C949CA58C949C8C9CA58C949C949CA58C9CA58C9CA58C949C8C9CA58C949C
8C949C8494948C949C849494849494848C94849494848C94848C947B8C8C848C947B8C8C7B8C94\
7B848C7B8C9473848C7B8C8C73848C73848C6B7B846B7B84637B7B637B845A737B
5A737B5A6B735A7373526B734A5A630000000000003939524242634A4A635A5A735A637B636B84\
636B846B738C6B738C737B8C737B8C7B84947B84947B84947B848C84849484848C
84848C7B84848484847B7B7B7B7B7373736B7B7B6B6B6B636363525252396B6B7B6B6B7B6B6B7B\
6B6B7B736B7B6B6B7B6B6B7B6B637B736B7B6B6B7B736B7B6B6373736B7B6B6373
736B736B6373736B7B7363737363736B6373736B737363737363736B636B73637373636B73636B\
6B5A6B73636B6B5A63735A636B5A63735A636B5A636B5A636B525A6B5A636B525A
6B525A6352526B525A634A52635252634A4A6B5252634A4A634A4A5A424A634A4A5A42425A4242\
5239425A42425239395239394A31395239394A3131000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004239424A42525A52635A526363636B6363736B6B7B6B6B7B
7373846B7384737B84737384737B7B6B7373737B7B737373737B736B736B6B736B636B63636B5A\
525A524A4A39000000000000000000000000000000000000000000000000000000
00000000000042393942393942394242394A42395242395242395A4239524A426B42426B424A6B\
424A6B4A527B4A527B525A7B4A5A7B5A638C5263845A6B8C5A6B8463738C63738C
637B94637B8C6B8494637B8C6B848C637B8C6B848C6B848C6B8C8C6B848C738C8C6B8C846B8C84\
6B847B6B8C7B6B84736B847363846B6B846B63735A5A73524A5A39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
0000000000000000000000000000003131313131313931393929314231394231314231395A524A\
4231394231314A31394231314231394231314A31394A31314A31394A3131523939
5231315239395239395A42395239395A42393931394239393931394231424A314252394A523942\
52394A52394A5A42525242525A42525A4252634A5A5A4A5A6352636352636B5A6B
6B52636B5A6B6B5A6B7363736B5A737363737363737B6B7B7363737B6B7B7B6B7B7B6B847B6B7B\
7B73847B738484738C8473848C7B8C8C7B8C8C7B948C7B8C948494947B94948494
947B8C9484949484949C8494948494948494948494948C9C8C8494948C94948C94948C9C8C949C\
8C9CA58C9CA58C9CA58C9C9C8C9CA58C9C9C8C9CA584949C8C949C84949C8C949C
848C94849494848C948494947B8C8C848C947B8C8C848C947B848C7B8C947B848C7B8C9473848C\
73848C73848C73848C6B7B846B7B84637B7B637B845A73735A737B5A6B735A6B73
526B6B4A5A6300000000000000000000000000000039395A39425A4A4A6B52526B63637B5A637B\
636B846B6B8473738C73738C737B947B7B947B84947B7B8C7B848C7B848C84848C
7B84848484847B847B84847B737B737B7B7373736B737363635A4A5A5242636373636B7B636373\
6B6B7B6B637B6B6B7B6B637B6B6B7B6B6B7B6B6B7B6B63736B6B7B6B6373736B7B
6B6373736B736B6373736B7B6B63737363736B63737363736B636B7363736B636B7363736B5A6B\
73636B73636B73636B6B5A6373636B6B5A63735A6B6B5A636B5A636B5A5A6B5A63
63525A6B525A6B525A6B525A634A526B5252634A526B5252634A4A634A52634A4A634A525A424A\
634A4A5A424A634A4A5A42425A42425239425A42425239395239394A31394A3939
423131000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000031313942394A524A5A524A5A635A6B635A6B6B63736B6B7B737384
6B6B847373847373847B7B8473737B737B7B737373737B7B6B73736B736B636B63636B635A635A\
525A4A394231000000000000000000000000000000000000000000000000000000
3931394239424231394A394242314242395242395242425A39395242425A4242634A4A6B42426B\
424A734A52734A5A7B4A5A7B5263845263845A6B8C5263845A6B8C5A6B8C637394
5A738C637B8C637B8C638494637B8C63848C63848C6B848C6B84846B8C8C6B84846B8C8463847B\
6B847B6384736B8473637B6B63846B637B63637B63526B52000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000313131393139312931423139393131423139423131423139\
4231314231394231314231394231314231394231314231394231314A3139423131
4A39395239315239393931314239393931394231424231394A394252394252394A52394A5A4252\
5242525A42525A4252634A5A5A4A5A635263634A5A6B526B6B52636B5A6B6B5A6B
736373736373736373736373736B7B73637B736B7B73637B7B6B847B6B847B73847B6B8484738C\
7B738484738C84738C8C7B948C7B8C8C7B948C7B8C948494947B94948494947B94
948494948494948494948494948C9C948494948C9C8C8494948C9C8C8C94948C9C8C8C948C9CA5\
8C9CA58C9CA5849C9C8C9CA5849C9C8C9C9C84949C8C9C9C84949C84949C7B8C94
84949C7B8C94848C947B8C8C848C947B8C8C7B8C9473848C7B8C9473848C73848C73848C73848C\
6B84846B7B84637B7B6B7B8463737B5A737B526B735A7373526B6B526B6B000000
00000000000000000000000000000000000000000031314A4242634A4A635252735A5A73636B84\
636B846B73846B6B84737B8C737B8C7B84947B7B8C8484947B848C84848C7B8484
84848C8484848484847B7B7B7B7B7B737B6B73736B6B6B636B63525252425A63735A6373636B7B\
63637363637B6363736B6B7B6B6B7B6B6B7B6B637B6B6B7B6B6B7B6B6B7B6B6373
736B7B6B6373736B7B6B6373736B7B6B63737363736B6373736B736B63737363736B636B736373\
6B636B7363736B5A6B73637373636B73636B6B5A6B73636B6B5A63735A636B5A63
735A636B5A636B5A636B525A6B5A636B525A6B525A634A526B525A6352526B5252634A526B5252\
634A52634A52634A4A634A5263424A634A4A5A42425A424A5A42425A4242523939
5239425239395239394A31314A39394A3131000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003131394A4252524A5A5A52635A5A636B6B736B6B73736B7B6B6B84
7B7B8C73738C7B7B8C7373847B7B84737373737B7B6B7373737B736B736B6B6B635A635A636B5A\
4A5242000000000000000000000000000000000000000000423139423939423139
4A394242314239525A42314A4A395242395242395A42395A4A42634242634A426B42426B4A4A73\
424A734A527B4A527B5263845263845263845263845A6B8C5A6B845A738C5A738C
637B94637B8C637B8C637B8C6B848C63848C6B848C6384846B8C8C6B848C6B8C8463847B6B8C84\
6B847B6B8473637B6B6B8473637B6B637B633931424A5242000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000313131393139312931393131393131423139392931393131393131\
423139392931423131393131423139392931393131393131423139392931423131
3931313931393931394231394231394A39424A314252394A52394A52424A52394A524252524252\
5A4A5A5A42525A4A5A5A4A5A6352636352636B526B6B526B6B5A736B5A6B736373
73637373637B73637373637B73637B7B6B84736B7B7B6B847B6B847B73847B6B847B738C7B738C\
847B8C84738C8C7B948C7B8C8C84948C7B8C8C84948C7B949484948C7B94948494
9484949484948C84949484948C8494948C9C8C84948C8C9C8C8C948C8C9C8C8C948C8C9C8C8C9C\
8C9CA584949C849C9C84949C849C9C84949C84949C84949484949C7B8C947B8C94
7B8C947B8C947B848C7B8C947B848C7B8C9473848C73848C73848C73848C6B7B846B7B846B7B84\
6B7B845A737B5A737B5A6B735A73734A6B6B526B6B4A636B000000000000000000
00000000000000000000000000000000000000000031394A39425A4A4A6B4A526B5A5A7B63637B\
6B6B84636B846B738473738C7B7B94737B8C7B84947B7B8C7B84947B7B8C7B848C
7B848484848C7B7B7B7B847B7B7B737B7B7373736B7373636363525A5A42525A6B5A63735A6373\
6363735A6373636373636373636B7B6363736B6B7B6B637B6B6B7B6B63736B6B7B
6B637B6B6B7B6B63736B6B7B6B6373736B7B6B63736B63736B6373736B736B63736B63736B6373\
7363736B636B7363736B636B7363736B5A6B73636B6B5A6B73636B6B5A6B735A6B
6B5A63735A6B6B52636B5A636B52636B5A636B525A6B525A6B525A6B525A634A526B525A634A52\
6B525A634A52634A52634A52634A5263424A634A4A63424A634A4A5A42425A424A
5A42425A42425239395239425239395239394A31394A39394A31314A39394231314A3131000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004239424A42525A52635A5263635A6B6B636B736B7B6B6B7B737384
73738C7B7B947373847B7B84737B7B737B7B737373737B736B736B6B736B636B5A636B5A525A52\
424A390000000000000000000000000000004231394231394A39424231394A3942
4A39424A394A4231424A39524239524A425A42395242425A42425A4A426B42426B4A4A6B424A6B\
4A52734A52734A5A7B525A7B5263844A6384526384526B845A738C526B845A738C
5A738C637B8C5A738C637B8C637B8C6B848C637B846B848C6384846B8C8C6384846B848463847B\
6B847B637B73638473637B6B63846B5A735A526B52000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000313139313131393139393131393139393131423139393131393139\
39313142313963524A393139393131393139393139393139393131423939393139
4231393931394239424239425239424A394252424A52394A5242525242525A4A5A5A425A5A4A5A\
5A4A5A6352636352636352636352636B5A736B5A6B7363736B5A7373637B736373
73637B73637B7B6B84736B7B7B6B84736B847B738C7B73847B738C7B738C847B8C84738C847B94\
847B8C8C84948C7B948C84948C7B949484948C84949484948C849494849C948494
9484948C8494948C9C8C8494948C9C8C8494948C9C8C8C9C8C8C9C8C8C948C949C8C8C9C8C949C\
84949C8C9CA584949C84949C7B8C9C84949C848C948494947B8C948494947B8C94
7B8C947B848C7B8C9473848C7B8C9473848C7B8C9473848C6B848C6B7B846B7B8463737B63737B\
5A6B735A737B526B73526B734A636B000000000000000000000000000000000000
00000000000000000000000000000000000000000031314A42425A424A63525A735A5A7363637B\
63637B6B738C6B7384737B8C737B8C7B84947B84947B84947B848C8484947B848C
84848C7B84848484847B847B7B7B7B73736B7B7B6B7373636B6B5A5A52425A6373525A735A6373\
5A5A7363637B5A637363637B5A6373636B7B63637B63637B63637B6B6B7B6B637B
6B6B7B6B637B736B7B6B6B7B6B6B7B6B6373736B7B6B637B6B6B7B6B6373736B7B6B6373736373\
6B63737363736B63737363736B636B7363736B63737363736B5A6B73637373636B
73636B6B5A6B73636B6B5A6B735A6B6B5A63735A6B6B5A636B5A636B525A6B5A636B525A6B525A\
63525A6B525A63525A6B525A634A526B525A634A52634A52634A4A634A52634A4A
634A4A5A424A634A4A63424263424A5A424263424A5A42425A42425239395A4242523939523942\
4A39395239424A39394A39394231314A3939423131000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000039313942394A524A5A5A4A5A635A6B63636B6B6B73736B737B73847B7384
7B7B8C7B7B8C7B84947B7B847B7B84737B7B7B7B7B737B737373736B6B636B6B6363635A525A4A\
0000000000000000004A39394231394A39424A39424A39424A31424A394A4A3942
4A394A4A394A4A395A4A395A4A425A42395A4A42634242634A426342426B4A4A734A4A734A5273\
425273525A7B525A7B526384526384526B8C5263845A6B8C526B845A738C5A738C
637B8C5A7B8C6384944A5273525273637B8C6B848C63848C6B848C6384846B8C8463847B63847B\
6384734A4263637B6B637B6B5A73635A7B63000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000313139312931393139393131393139393131393139393131423139393131\
393139393131393139313131393139393139393139393139423139393139423942
4231424A39424A394A52424A4A394A52424A5242525A42525242525A4A5A5A4A5A634A635A4A5A\
6352636352636B5A6B63526B6B5A736B5A737363736B5A7373637B73637B736B7B
73637B736B84736B847B7384736B847B738C7B738C84738C7B738C847B8C847B8C8C7B94847B8C\
8C7B948C7B948C84948C7B948C84948C84949484948C84948C84948C8494948C9C
8C84948C8C948C84948C8C9C8C84948C8C9C8C8C9C8C8C9C848C948C8C9C848C9C8C949C848C94\
84949C84949C84949C7B8C947B949C7B8C947B8C947B8C947B8C947B848C7B8C94
73848C73849473848C7384946B7B8C6B848C6B7B846B7B8463737B63737B5A737B5A737B526B6B\
526B734A636B000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039394239395A4A4A634A526B5A637B5A637B\
636B84636B8473738C6B7384737B8C7B7B8C7B84947B7B8C7B848C7B848C84848C
7B84848484847B84848484847B7B737B7B7373736B73736B6363525A5A424A5A6B525A73525A6B\
525A73525A735A63735A5A735A63735A637363637B5A637363637B636373636B7B
6363736B637B63637B6B6B7B63637B6B6B7B6B637B6B6B7B6B63736B637B6B63736B6B7B6B6373\
6B63736B63737363736B63736B63736B63737363736B5A6B6B63736B636B736373
6B5A6B6B636B6B5A6B73636B6B5A6B6B5A6B6B5A6B73636B6B5A636B5A636B5A636B5A636B525A\
6B5A636B525A6B526363525A6B525A63525A6B525A634A52634A52634A52634A52
634A4A634A5263424A634A525A424A63424A5A424A63424A5A424263424A5A42426342425A3942\
5A42425A39425A42425239395239425239395239424A39394A39394A39394A3939
4231314A3939423131423939000000000000000000000000000000000000000000000000000000\
0000003931394A4252524A5A5A5263635A636B63736B63737B6B7B7B737B7B7B84
7B7B8C848494848C947B848C737B7B7B7B7B737B737B7B736B736B4A39394A31394A39394A3139\
4A39394A39394A39424A31394A39424A314252394A4A39424A394A4A394A4A424A
4A394A4A39524A395A4A425A42395A4A42634242634A4A6B4242634A4A6B4A4A6B4A5273424A73\
4A527B4A5A7B526384525A84526384526384526B845263845A6B8C5A738463738C
5A738C637B8C637B8C63848C5A7B8463848C637B8463848C637B846B84846384846B8C8463847B\
6384735A7B6B637B6B5A73635A735A4A634A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000312931393139393131393139393131423139393131393139393131423139\
393131393139393131423139393139393139393139423942423142423942423142
52394A4A394A52425252394A5A42525242525A425A5242525A4A635A4A5A6352635A4A6363526B\
63526B6B5A6B6B5A6B7363736B5A737363736B6373736B7B73637B736B7B73637B
7B6B84736B847B738C736B8C7B738C7B738C7B738C7B738C847B94847B94847B94847B948C8494\
847B948C8494847B948C849C8C84948C84948C849494849C8C84948C849C8C8494
8C8C9C8C849C8C8C9C8484948C8C9C848C9C848C9C848C9C8C949C848C9C848C9C7B8C94848C9C\
7B8C947B949C7B8C947B949C7B8C947B8C9473848C7B8C9473848C73849473848C
73849473848C73848C6B7B846B7B8463737B63737B5A73735A737B526B73526B734A636B4A6B6B\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031313942425A424A6352526B525A73636384\
636B7B6B6B846B6B84737B8C737B8C7B7B94737B8C7B84947B848C7B848C7B8484
84848C8484848484847B7B7B7B847B737B73737B6B7373636B6B5A5A5A42425263425263525A6B\
525A6B525A73525A6B5A6373525A735A63735A5A7363637B5A63736363735A6373
63637B63637363637B6363736B6B7B63637B6B6B7B6363736B6B7B6B637B6B6B7B6B6373736B7B\
6B63736B637B6B6373736B7B6B63736B63736B63737363736B63737363736B5A6B
7363736B636B7363736B5A6B7363736B5A6B73636B6B5A6B73636B6B5A6B735A6B6B5A63735A6B\
6B5A636B5A636B52636B5A636B52636B526363525A6B526363525A6B525A634A52
6B525A634A52634A52634A52634A52634A52634A525A424A634A5263424A634A4A5A424A634A4A\
5A424A63424A5A394263424A5A42425A42425A39425A42425A39425A3942523939
5239425239395239394A31395239394A31394A39394A31315239394A31394A39394A31314A3939\
4A31394A39394231314A39394A31394A39394231314A3939423139423939423131
4A39394231314231394A31315239394A31395239394A31395239424A39425239424A3939523942\
4A39425239424A314252394A52394A52394A4A394A5242524A394A4A42524A394A
5242524A395A4A425A4A395A4A42634A42634A4A6B4242634A4A734A4A6B4A4A734A4A734A527B\
4A527B525A7B4A5A7B5263845263845263845263845A6B8C526B845A738C5A7384
637B8C637B8C637B8C5A7B8463848C637B84638484637B8463848463848463848463847B63847B\
5A7B735A7B6B5A73635A7B6352735A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
0000000000000000003131314A4A3939313931293139313939313163524A392931393139393139\
4231393931313931393931394231423931394231424231424A394A4A394A52394A
52394A5242525239525242525242525A4A5A5A425A5A4A635A4A6363526B5A526363526B635A6B\
6B5A736B5A6B6B63736B5A7373637B6B637373637B6B637B736B8473637B736B84
736B847B738C736B8C7B73947B738C847B947B738C847B94847B94847B94847B8C847B94847B94\
8C8494847B948C8494847B948C849C847B948C84948C84948C849C8484948C849C
8484948C8C9C848494848C9C848C9C848C9C7B8C9C848C9C848C9C848C9C7B84947B8C947B8C94\
7B8C94738C947B8C94738C947B8C9473848C73849473848C7384946B7B8C73848C
6B7B846B7B8463737B63737B5A737B5A737B526B73526B734A636B4A6B6B000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031313939395A4A4A634A4A6B5A5A735A637B\
636B84636B7B6B73846B7384737B8C737B8C7B84947B7B8C7B84947B7B8C7B848C
7B848484848C7B847B7B847B7B7B737B7B7373736B73736B6B6B5A5A5A4A314252394A5A394A5A\
4252634252634A5A6B4A5A6B525A73525A6B525A73525A735A6373525A735A6373
5A637363637B5A5A7363637363637363637B63637363637B6363736B6B7B6363736B6B7B6B637B\
6B6B7B6B63736B637B6B63736B6B7B6B63736B63736B63736B63736B63736B6373
6B63737363736B5A6B6B63736B5A6B7363736B5A6B6B5A6B6B5A6B73636B6B5A6B6B5A6B6B5A6B\
735A6B6B5A636B5A6B6B5A636B5A6B6B52636B52636B52636B526363525A6B525A
63525A6B525A634A5263525A634A526B525A634A52634A52634A52634A525A424A634A5263424A\
634A525A424A63424A5A424A63424A5A42425A424A5A42425A424A5A39425A4242
5A39425A42425239395A39425A39395A39425239395A39395239395A3942523139523939523939\
5A3939523139523939523139523939523131523939523139523939523131523939
5231395239424A31395239425239425239424A394252394252394252394A4A394252394A523942\
52394A4A314252394A4A394A5239524A394A5242524A39525242524A39524A4252
4A395A5242634A395A4A42634A42634A4A6B4242634A4A6B4A4A6B4A527342526B4A52734A5273\
525A7B4A5A7B5263845263845263844A637B526B84526B845A738C526B845A738C
5A738C637B8C5A73845A7B8C5A7B84637B845A7B7B637B84637B7B638484637B7B63847B5A7B73\
5A7B735273635A7363526B5A4A5A4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000313131393139313131393139393139635242393139423139393139393139\
39313942394263524A4239424231424A394A4A394A4A394A4A394A524252524252
5242525242525A4A5A5A4A5A5A4A635A4A5A63526363526363526B63526B6B5A736B5A736B6373\
6B5A7373637B6B637373637B6B637B736B7B73637B736B84736B847B738C736B8C
7B739473738C7B7B9C7B7394847B947B738C847B94847B94847B94847B94848494847B94848494\
847B948C849C8484948C849C8484948C8C9C8C84948C8C9C8484948C8C9C848C9C
848C9C848C9C848C9C848C9C848C9C7B8C9C848C9C7B8C94848C9C7B84947B8C9C7B8C947B8C94\
73848C7B8C94738C94738C9473848C738C9473848C73848C6B7B846B848C637B84
637B845A737B63737B5A6B73526B734A636B526B73000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031313939425A42426352526B525A7363637B\
63637B6B6B846B738473738C73738C7B84947B84947B84947B7B8C8484947B848C
84848C7B8484848C847B847B7B847B737B737B7B7373736B6B6B5A5A5A4A00000031425239425A\
39425A424A63424A634A52634A5263525A73525A73525A73525A735A63735A6373
5A63735A5A7363637B5A637363637B5A637363637B63637363637B6363736B6B7B63637B6B637B\
63637B6B6B7B6B637B6B6B7B6B63736B6B7B6B637B6B637B6B637373637B6B6373
6B63736B63737363736B63736B63736B5A6B7363736B63736B63736B5A6B7363736B5A6B6B636B\
6B5A6B73636B6B5A6B6B5A6B6B5A63735A6B6B5A6B6B5A6B6B52636B5A6B6B5263
6B5A6363525A6B5A636B52636B526363525A6B526363525A6B525A634A526B525A634A52634A5A\
634A52634A5A634A52634A525A424A634A5263424A63424A5A424A63424A5A424A
63424A5A394263424A5A42425A424A5A394263424A5A42425A42425A39425A424A5A39425A4242\
5239425A42425A39425A39425239395A42425239425A39425239395A4242523942
5A39425239425A424A52394A5A424A52394A5A424A52394A52394A52394A5A425252394A52394A\
52394A5242525239525242524A395252425A52425252425A4A395252425A4A425A
5242634A395A5242634A42634A4A6B4A426B4A4A734A4A734A4A734A4A7352527B4A52734A5A7B\
4A5A7B526384526384526384526384526B84526B845A6B84526B845A738C5A738C
637B8C5A7B84637B8C5A7B84637B845A7B84638484637B846384845A7B7B63847B5A7B735A7B73\
5A73635A7B6352735A526B5A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000313131393139313131393139393139393139393139393139393139423142393139\
42314242314242394A4231424A394A4A394A5242524A394A52425252425252425A
5242525A4A5A5A4A5A5A52635A4A6363526B63526B635A6B63526B6B5A73635A736B6373635A73\
6B637B6B637373637B6B637B73637B73637B736B84736B84736B8C736B8C7B7394
73738C7B73947B739C7B7BA57B7B9C7B7B947B7B94847B947B7B94847B94847B9484849C847B94\
8484948484948C849C8484948C849C8484948C8C9C848494848C9C84849C848C9C
7B8494848C9C7B8C9C848C9C7B84947B8C9C7B8C947B8C9C7384947B8C947384947B8C9473848C\
73849473848C7384946B848C73848C6B7B8C6B7B8C637384637B845A737B63737B
526B73526B73526B73526B73426363000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003939524242634A4A6B5A5A735A5A7B\
636B84636B846B738C6B7384737B8C737B8C7B84947B7B8C7B848C7B848C84848C
7B848484848C7B84848484847B7B737B7B7373736B73736B6B6B5A635A4A000000000000000000\
000000314252394A5A39425A424A63424A634A526B4A5263525A6B5A4A4A5A6373
525A735A6373525A735A63735A5A735A63735A5A7363637B5A5A7363637363637363637B636373\
63637B63637B6B6B7B6363736B637B6B637B6B6B7B6363736B637B6B63736B637B
6B63736B63736B63736B637B6B5A736B63736B63736B63736B5A6B6B63736B5A6B6B63736B5A6B\
6B5A6B6B5A6B6B63736B5A6B6B5A6B6B5A6B6B5A6B6B5A636B5A6B6B5A636B5A6B
6B52636B5A636B52636B5A6B6352636B52636B52636B526363525A6B526363525A6B5263634A5A\
63525A634A5A63525A634A52634A5A634A52634A5A5A4252634A525A4252634A52
5A424A6342525A424A634A525A424A63424A5A424A6342525A424A5A424A5A424A63424A5A394A\
5A424A5A424A5A424A5A39425A424A5A39425A424A5239425A424A5A394A5A424A
52394A5A424A5A424A5A425252394A5A425252394A5A425252394A5A39525239525A395252394A\
52395252395252425A52395252425A52425A52425A4A395A5242634A425A524263
4A395A4A42634A42634A4A6B4A426B4A4A6B4A4A735252734A4A734A52734A5273525A7B4A5273\
525A7B525A845263844A637B526384526384526B84526B845A73845A73845A7B8C
5A73845A7B8C5A7B845A7B845A737B5A7B845A7B7B6384845A7B7B63847B5A7B735A7B73527363\
52736352735A52735A425239000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000031313931313139313939313939313939313939314239313939314239314242394A\
4231424A394A42314A4A39524A394A5239524A395252425A52425252425A52425A
5A4A635A4A635A52635A4A63635A6B63526B635A73635A6B6B5A736B5A736B63736B5A736B637B\
6B637B6B637B6B637B736B846B6B84736B84736B8473738C73738C7B7394737394
7B7B9C7B739C7B7BA57B7B9C8484AD7B7B947B7B947B7B9484849C7B7B948484947B7B9484849C\
84849484849C8484948C849C84849C84849C848494848C9C84849C848C9C7B849C
848C9C7B8C9C7B8C9C7B84947B8C9C7B84947B8C947384947B8C947384947384946B848C738494\
6B848C73848C6B7B8C6B848C637B84637B845A737B63737B5A6B735A6B73526B6B
526B734A636B000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039395A4242634A526B52526B63637B\
63637B6B6B846B6B8473738C73738C737B94737B8C7B84947B848C7B848C7B848C
84848C7B84848484847B7B7B7B847B7B7B73737B737373637373635A5A4A525239000000000000\
000000000000000000000000314252394A5A394A5A4252634252634A526B4A526B
525A73525A6B525A73525A735A6373525A735A6373525A735A637B5A63735A63735A5A7363637B\
63637363637B6363736B637B6363736B637B6363736B6B7B6B637B6B637B6B6373
6B6B7B6B63736B637B6B63736B637B6B63736B63736B63736B63736B63736B63736B5A6B6B6373\
6B5A6B6B63736B5A6B6B63736B5A6B6B5A6B6B5A6B6B5A736B5A6B6B5A6B6B5A6B
6B5A6B6B5A6B6B5A6B6B52636B5A6B6B52636B5A6B6352636B5A6B6B52636B52636352636B5263\
6352636B5263634A5A6B5263634A5A63525A634A5A6B525A634A5A634A5A634A52
634A5A634A52634A525A4252634A5A634A52634A525A4252634A525A4252634A525A424A634A52\
5A42525A42525A424A6342525A424A5A42525A394A5A42525A42525A42525A4252
5A42525A42525A42525A39525A42525A42525A42525239525A425A5239525A42525239525A425A\
52425252425A524252524A5A52425A52426352425A5242635242635242634A4263
524A6B4A42634A4A6B4A4A6B5252734A4A735252734A4A7352527B4A52734A527B4A5273525A7B\
525A7B5263844A637B526B84526384526B84526B845A738C526B845A738C5A7384
5A7B8C5A73845A7B845A737B5A7B845A7B7B5A7B7B5A7B7B63847B5A7B7B5A7B7352736B5A736B\
52735A52735A4A634A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000031293131313931313939314231293939314239314242314239314242394A42314A\
4A394A42314A4A39524A394A5242524A3952524252524252524A5A52425A5A4A63
5A4A635A526B5A4A6363526B63526B635A73635A6B635A73635A736B637B635A736B637B6B637B\
736B7B6B637B736B846B6B84736B8C6B6B8473738C73738C7B73947373947B739C
7B739C7B7BA57B7BA57B7BA57B7BA58484AD7B7B947B7B947B7B9484849C7B7B9484849C848494\
84849C7B849484849C84849C848C9C7B8494848C9C7B849C848C9C7B84947B8C9C
7B84947B8C9C7384947B84947384947B849473848C73849473848C7384946B7B8C6B848C6B7B8C\
6B848C637B846B7B8C63737B63737B526B73526B73526B73526B734A636B000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000313952424263424A63525A735A637B\
636B8463637B6B6B846B7384737B8C737B8C7B7B947B7B8C7B84947B7B8C7B848C
7B848484848C7B84847B847B7B7B737B7B7373736B73736B6B6B5A636352524A39000000000000\
000000000000000000000000000000000000314252314252394A5A394A5A425263
4252634A526B4A52634A5A6B4A5A6B525A73525A6B525A73525A735A6373525A735A63735A5A73\
63637B5A5A736363735A637363637B5A637363637B6363736B637B6363736B637B
6363736B637B6363736B637B6B63736B637B6363736B63736B63736B637B635A736B63736B6373\
6B6373635A6B6B63736B5A6B6B6373635A6B6B5A736B5A6B6B5A73635A6B6B5A6B
6B5A6B6B5A6B63526B6B5A6B6B5A6B6B5A6B6352636B5A6B6B52636B5A6B6352636B5263635263\
6B52636352636B52636352636B5263634A5A635263634A5A6B5263634A5A63525A
634A5A63525A634A5A634A5A634A5A634A5A5A4A52634A5A634A5A634A5A5A4252634A5A5A4252\
634A5A5A4252634A525A4252634A5A5A42525A42525A425263425A5A42525A4252
5A42525A425A5242525A425A5A42525A425A5239525A425A5242525A425A5239525A425A52425A\
5A425A52425A52426352425A524A63524263524263524263524A6B4A426352426B
4A4263524A6B4A4A6B4A4A734A4A735252734A4A734A52734A527352527B4A52734A5A7B4A5A7B\
5263844A5A7B526384526384526B844A637B526B84526B845A738C5273845A738C
5A73845A7B8452737B5A7B845A737B5A7B7B52737B5A7B7B5A7B735A7B7352736B52736B526B63\
52735A4A6B52394A39000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000039313931313939314239313939394239314242394A39314242394A42394A4A3952\
42394A4A42524A39524A42524A395252425A52425A524A63524A5A5A52635A4A63
5A526B5A526B635A73635A73635A73635A736B637B635A736B637B635A7B6B637B6B637B6B6B84\
6B637B736B84736B84736B8C736B8C7B73947373947B73947373947B7B9C7B7B9C
7B7BA57B7BA58484AD7B7BA57B84A57B7BA584849C7B7B9484849C7B7B9484849C84849C84849C\
7B8494848C9C84849C848C9C7B849C848C9C7B849C7B8C9C7B84947B8C9C7B8494
7B84947384947B8C9473849473849473848C7384946B848C73848C6B7B8C6B848C6B7B8C6B7B8C\
637B84637B845A737B5A737B526B73526B734A636B4A636B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039395A39425A4A526B5252735A637B\
5A637B6B6B846B6B846B738C6B738C7B7B947B7B947B84947B7B8C8484947B848C
84848C7B8484848C8C7B847B7B847B737B737B7B7373736B7373635A5A4A5A5242000000000000\
00000000000000000000000000000000000000000000000000000039425A39425A
394A63394A5A4252634252634A526B4A526B525A73525A6B525A73525A735A63735A5A735A6373\
5A5A735A637B5A63735A637B5A5A7363637B63637363637B63637363637B63637B
63637B6363736B637B63637B6B637B6363736B6B7B6B637B6B637B6363736B637B6B63736B637B\
635A736B637B6B63736B63736B5A736B63736B63736B63736B5A6B6B63736B5A6B
6B5A73635A6B6B63736B5A6B6B5A6B635A6B6B5A6B6B5A6B6B5A6B6352636B5A6B6B526B6B5A6B\
6352636B5A6B6352636B52636352636B5A6B6352636B52636352636B5263635263
635263634A5A6B5263634A63635263634A5A635263634A5A634A63634A5A635263634A5A634A5A\
5A4A5A634A63634A5A634A5A5A425A634A5A5A425A634A5A5A425A634A5A5A425A
5A4A5A5A425A634A635A425A5A425A5A425A5A4A635A425A5A425A52425A5A4A635A425A5A4A63\
52425A5A4A635242635A4A635242635A4A6B524263524A6B4A4263524A6B524A6B
524A6B4A4A6B525273524A735252734A4A7352527B4A52734A527B4A5273525A7B4A5A7B525A7B\
4A5A7B526384526384526384526384526B84526B84527384526B845A7B8C5A7384
5A7B845273845A7B845A737B5A7B7B52737B5A7B7B5A7B7B5A7B7B52736B5A7B6B527363526B5A\
4A6B524A634A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000031313939314231314239314239314242394A39314A42394A42394A4A3952423952\
4A39524A395252425A4A425252425A52425A524A63524A5A5A4A635A4A635A526B
5A526B635A73635A7363637B635A7363637B635A736B637B635A7B6B637B6B637B6B6B846B637B\
6B6B846B6B8473738C6B6B8C7373947373947B7B9C7373947B7B9C7B7B9C7B7BA5
737B9C7B7BA57B7BA57B84A57B7BA57B84A57B84A57B849C7B7B947B849C7B849484849C7B8494\
7B849C7B849C848C9C7B84947B849C7B84947B8C9C7384947B84947384947B8494
73848C73849473848C7384946B7B8C6B848C6B7B8C6B848C637B846B7B8C637B84637B845A737B\
5A737B526B73526B734A63634A636B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031314A424263424A63525A735A5A73\
63637B636B846B73846B6B8473738C737B8C7B8494737B8C7B848C7B848C84848C
7B848484848C8484848484847B7B737B7B7373736B737B6B6B6B5A636352525239000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000031425A31425A394A63394A5A4252634252634A526B4252634A5A6B4A5A6B526373525A73\
526373525A735A6373525A735A63735A5A7363637B5A5A7363637B5A637363637B
5A5A7363637B63637363637B635A7363637B6363736B637B6363736B637B63637B6B637B636373\
6B637B6B63736B637B635A736B63736B5A736B6373635A736B63736B5A736B6373
635A6B6B5A736B5A6B6B6373635A6B6B5A6B635A6B6B5A7363526B6B5A6B635A6B6B5A6B635263\
6B5A6B63526B6B5A6B6352636B526B6352636B5A6B6352636352636352636B5263
634A63635263635263635263634A63635263634A636352635A4A5A634A63634A5A6352635A4A5A\
634A635A4A5A634A635A4A5A634A635A4A5A634A635A425A5A4A635A4A5A634A63
5A425A5A4A635A425A5A4A635A425A5A4A635A425A5A4A6352425A5A4A635A42635A4A63524263\
5A4A63524A635A4A6B524263524A6B524A63524A6B4A4263524A6B524A6B524A6B
4A4A6B525273524A7352527B4A4A734A52734A5273525A7B4A52734A5A7B4A5A7B5263844A5A7B\
5263845263845263844A637B526B84526B84526B84526B845A73845273845A7B84
52737B5A738452737B5A7B7B52737B5A7B7B527B735A7B7B52736B52736B5273635273634A6352\
4A6B52000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000031313931314239314A39314239314A39314A42395242394A4239524239524A425A\
4A42524A425A4A425A524A6352425A524A63524A635A526B5A526B5A526B5A526B
635A73635A73635A7B635A736B637B63637B6B637B63637B6B6B846B637B6B6B846B6384736B8C\
6B6B8473738C6B6B8C73739473739473739C73739C7B7BA5737B9C7B7BA5737B9C
7B84A57B7BA57B7BA5737B9C7B84A57B7B9C7B84A57B7B947B849C7B849C7B849C7B849C7B8C9C\
7B849C7B849C7384947B849C7384947384947384947384947384947384946B7B8C
7384946B7B8C73848C6B7B8C6B848C6B7B8C6B7B8C637B846B7B8C63737B5A737B526B73526B73\
4A636B4A636B425A63000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039395239425A4A4A6B4A526B5A637B\
5A637B636B84636B8473738C6B738C737B8C737B8C7B84947B7B8C7B848C7B7B8C
84848C7B84848484847B7B7B7B847B7B7B737B7B7373736B7373636363525A5242000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000394A5A314A5A394A5A394A5A4252634252634A526B42526B525A73\
525A6B525A73525A735A6373525A735A6373525A735A637B5A63735A63735A5A73
63637B5A637363637B5A5A7363637B63637363637B63637363637B63637363637B6363736B637B\
6363736B637B6363736B637B6B63736B637B635A736B637B6B63736B6373635A73
6B63736B5A736B6373635A6B6B63736B5A736B5A73635A6B6B5A73635A6B6B5A73635A6B6B5A73\
635A6B6B5A6B63526B6B5A6B63526B6B5A6B6352636B5A6B63526B63526B635263
6B526B63526363526B63526363526B635263635263634A6363526B6352636352635A4A63635263\
634A636352635A4A63635263634A63634A635A4A636352635A4A635A4A635A4A63
634A635A4A635A4A635A4A635A4A635A4A635A4A635A4A635A4A6B5A4A635A4A6B524A635A4A6B\
524A6B5A4A6B524A635A4A6B524A6B524A6B4A4A6B525273524A6B524A734A4A6B
5252735252735252734A527352527B4A527352527B4A5273525A7B4A5A7B525A7B4A5A7B526384\
526384526384524A6B526B84526B84526B84526B845A73845273845A738452737B
5A738452737B52737B52737B5A7B7B527373527B73527373527373527363526B634A6B524A6B52\
313139000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000031293931314252423939314A39314A39394A39314A4239524239524A42524A425A\
52425A4A425A524A634A4A63524A63524A635A526B5A526B5A52735A526B635A73
635A73635A7B635A7363637B3929316B637B63637B6B63846B637B6B6B846B63846B6B846B6B8C\
73738C6B6B8C737394737394737B9C737394737B9C73739C7B7BA573739C737BA5
737B9C7B7BA5737B9C7B7B9C737B9C7B84A5737B9C7B849C7B849C7B849C7384947B849C738494\
7B849C737B947384947384947384946B7B8C7384946B7B8C7384946B7B8C6B7B8C
6B7B8C6B848C637B846B7B8C637B846B7B8C5A737B5A737B526B73526B734A636B4A636B425A63\
425A63000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031314A42425A42426352526B525A73\
63637B63637B6B6B846B6B84737B8C73738C7B7B947B7B8C7B84947B7B8C7B848C
7B848484848C7B84847B84847B7B7B7B7B7373736B73736B6B6B636B6B52525242000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031425A31425A394A63394A5A425263424A63\
4A526B4252634A5A6B4A5A6B525A73525A6B525A73525A735A6373525A735A6373
5A5A735A637B5A5A735A63735A5A7363637B5A5A736363735A5A7363637B5A5A73636373636373\
63637B635A7363637B6363736B637B635A736B637B6363736B637B635A736B6373
6363736B637B635A736B6373635A736B6373635A6B6B5A73635A736B5A73635A6B6B5A73635A6B\
6B5A7363526B635A6B635A6B6B5A7363526B635A6B63526B6B5A6B63526B63526B
63526B635A6B63526363526B63526B63526B634A6363526B63526363526B5A4A6363526B634A63\
63526B5A4A636352635A4A6363526B5A4A636352635A4A6363526B5A4A635A4A63
5A4A635A526B5A4A635A4A6B5A4A635A526B5A4A635A4A6B5A4A6B5A526B524A635A4A6B524A6B\
5A526B524A6B524A6B524A6B52526B4A4A6B524A6B524A6B5252734A4A6B525273
5252735252734A52735252734A5273525A7B4A52734A5A7B4A5A7B525A7B4A5A7B5263844A637B\
526B844A637B526384526384526B844A6B7B526B84526B845A7384526B7B527384
52737B52737B52737352737B527373527B7B5273735273734A6B6352736342635A4A6B5242634A\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000039314231314239314A39314A42395239394A4239524239524A425A4A425A4A4263\
4A425A524A63524A63524A6B524A635A526B5A526B5A5A735A5273635A73635A73
635A7B635A736B637B63637B6B638463637B6B6B846B6B846B6B846B6B8473738C6B6B8C737394\
6B7394737B9C737394737B9C7373947B7B9C737B9C737B9C737B9C7B7BA5737B9C
7B7BA5737B9C7B84A5737B9C7B849C737B9C7B849C73849C7B849C7384947B849C738494738494\
737B947384947384947384946B7B8C7384946B7B8C7384946B7B8C6B84946B7B8C
6B7B8C637B846B7B8C6373846373845A6B7B5A737B526B734A636B425A63426363000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031313939395A4A4A6B4A526B5A5A7B\
5A637B636B84636B846B738C6B738C737B947B7B8C7B84947B7B8C8484947B848C
84848C7B8484848C8C7B84847B847B7B7B737B7B7373736B73736363635263634A524A39000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000394A5A394A5A394A63\
394A5A4252634252634A526B42526B525A734A5A6B525A73525A735A637B525A73
5A6373525A735A637B5A63735A637B5A5A7363637B5A63735A637B5A5A7363637B63637363637B\
5A5A7363637B63637363637B635A736B637B63637B63637B635A736B637B63637B
6B637B635A736B637B6363736B637B635A736B637B6363736B6373635A736B637B635A736B6373\
635A736B6373635A736B5A73635A6B6B5A73635A6B6B5A73635A6B6B5A73635A6B
635A7363526B6B5A7363526B635A6B63526B635A7363526B63526B63526B635A6B63526B63526B\
5A526B63526B63526B63526B5A526363526B63526B63526B5A4A6B63526B5A526B
63526B5A4A6B63526B5A526B5A526B5A4A6B6352735A526B5A526B5A4A6B5A52735A4A6B5A526B\
524A6B5A5273524A6B525273524A6B52527352527352527352527352527B525273
525273525273525A7B52527B525A7B4A5273525A7B4A5A7B525A7B4A5A7B526384526384526384\
4A6384526B84526384526B84526B7B527384526B845273845273845A738452737B
52737B52737B52737B527373527373527373527B734A736B4A73634A6B5A4A6B5A426352000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000031313939314A31314A39394A39314A42395239395242425A42425A4A42634A425A\
524A63524A6352526B524A6B5A526B5A526B5A5A735A526B5A5A735A5A7363637B
5A5A7363637B63637B6B638463637B6B6B846B6B846B6B8C6B6B846B6B8C6B6B8C7373946B738C\
737394737394737B9C6B7394737B9C73739C737B9C73739C737B9C737B9C737B9C
737B9C737B9C737B9C7B849C737B9473849C737B9C7B849C737B94738494737B947384946B7B94\
7384946B7B947384946B7B8C6B7B8C6B7B8C6B848C637B8C6B7B8C637B8C6B7B8C
6373846373845A737B5A737B5263734A636B4A636B42636B000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039425A424A6352526B525A73\
63637B63637B6B6B84636B8473738C737B8C7B8494737B8C7B848C7B848C84848C
7B848484848C7B84848484847B7B7B7B7B7B737B737B7B6B6B73636B6B5A5A5A4A5A5242000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
31394A314A5A394A63394A5A4252634252634A526B4252634A5A6B4A5A6B525A73
4A5A6B525A73525A735A637B525A735A63735A5A735A637B525A735A63735A5A7363637B5A5A73\
5A637B5A5A7363637B5A5A7363637B635A7363637B5A5A7363637B63637363637B
635A7363637B635A7363637B635A7363637B635A736B637B635A7363637B635A736B637B635A73\
636373635A736B6373635A73635A73635A736B5A73635A6B635A73635A73635A73
63526B635A73635A6B635A7363526B635A7363526B635A735A526B635A6B63526B635A735A526B\
63526B63526B6352735A526B63526B5A526B6352735A526B63526B5A526B635273
5A526B5A52735A526B6352735A4A6B5A526B5A526B5A5273524A6B5A52735A4A6B5A5273524A6B\
525273524A6B5A5273524A6B525273525273525273525273525273525273525A7B
4A527352527B525273525A7B4A52734A5A7B4A5A7B52637B4A5A7B5263845263845263844A637B\
5263844A637B526B844A637B526B84526B7B527384526B8452738452737B527384
4A6B7B52737B52737352737B4A73735273734A736B4A736B42635A426B5A426352000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000031314231314239395239315239395239395242425A42425A4A425A4A425A524A6B\
524A6B52526B524A6B5A527352526B5A527352526B5A5A735A5A735A5A7B5A5A73
63637B63637B63637B63637B6B6B846B6B846B6B8C6B6B8473738C6B6B8C7373946B6B8C737394\
6B73947373946B7394737B9C73739C737B9C6B739C737B9C737B9C737B9C737B9C
737B9C737B9C737B9C737B9473849C737B9473849C737B9473849C6352637384946B7B8C738494\
6B7B8C6B7B946B7B8C6B84946B7B8C6B7B8C637B846B7B8C637B846373845A6B7B
5A737B526B73526B734A636B4A636B425A63000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039395A424A634A4A6B5A5A73\
5A637B636B84636B7B6B738C6B738C737B8C737B8C7B84947B7B8C7B848C7B7B8C
84848C7B84848484847B84848484847B7B737B7B7373736B73736B6B6B5A636352525239000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000313942314A5A394A63394A5A4252634252634A526B42526B
4A5A6B4A526B525A734A5A6B525A73525A73526373525A735A637B525A735A6373525A735A637B\
5A5A735A63735A5A7363637B5A637363637B5A5A7363637B63637363637B5A5A73
63637B63637363637B5A5A7363637B635A7363637B635A7363637B635A7363637B635A736B637B\
635A7363637B635A736B637B635A73635A73635A736B637B635A73635A73635A73
6B5A73635A73635A73635A73635A73635A73635A7363526B635A73635A73635A7363526B635A73\
635273635A735A526B635A736352736352735A526B635A735A526B5A52735A526B
6352735A526B5A52735A526B5A52735A526B5A5273524A6B5A52735A526B5A5273524A6B5A5273\
52527352527352526B525273525273525273525273525A7B525273525A7B525273
525A7B525A7B525A7B4A5273525A7B4A5A7B525A7B4A5A7B5263845263845263844A637B526B84\
4A637B526B844A637B526B84526B7B526B844A6B7B527384526B7B5242634A6B7B
52737B5273735273734A6B735273734A736B4A736B426B5A4A6B5A000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000031313931314A31314A39395239315242395A42395A4A426342425A4A4A634A4A63\
52526B4A4A6B52526B52526B5A527352526B5A5A735A5A735A5A7B5A5A735A5A7B
5A5A7B63638463637B6363846363846B6B8C6B6B846B738C6B6B8C6B73946B6B8C6B73946B7394\
7373946B73947373946B7394737B9C6B7394737B9C6B7B9C737B9C6B7394737B9C
6B7B94737B9C6B7B94737B94737B947384946B7B94737B946B7B947384946B7B8C6B7B8C6B7B8C\
6B7B8C63738C6B7B8C637B8C6B7B8C6373846373845A737B5A737B526B73526B73
4A636B4A636B425A63000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039394A39425A4A4A6B52526B\
5A637B5A637B636B846B6B8473738C6B7384737B8C737B8C7B8494737B8C7B848C
7B848484848C7B7B848484847B7B7B7B847B737373737B6B73736B7373636363525A5A42000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000313139314A5A394A63394A5A425263
424A6342526B4252634A526B4A526B4A5A734A526B525A73525A73526373525A73525A73525A73\
5A637B525A735A63735A5A735A637B5A5A735A63735A5A7363637B5A5A735A6373
5A5A7363637B5A5A736363735A5A7363637B5A5A736363735A5A7363637B5A5A73635A73635A73\
63637B5A5A73635A73635A7363637B635A73635A73635A7363637B635A73635A73
635A73635A735A5A73635A73635A73635A735A5273635A73635A73635A735A5273635A73635273\
635A735A526B635A735A5273635A735A526B5A52735A5273635A735A526B5A5273
5A526B5A527352526B5A52735A526B5A527352526B5A52735252735A527352526B5A5273525273\
5A5273525273525273525273525A7B525273525A7B525273525A7B4A5273525A7B
4A527B525A7B4A5273525A7B4A5A7B525A7B4A5A7B52637B4A637B5263844A637B52637B4A637B\
526B844A637B4A6B7B4A6B7B526B844A6B7B526B7B4A6B7B5273844A6B7B52737B
4A6B735273734A6B6B4A73734A6B6B4A736B426B63425A4A000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000031314A39395239395242425A42425A42426342425A4A4A6B4A4A635A4239\
4A4A6B52527352526B5252735252735A5A735A5A735A5A7B5A5A7363637B5A637B
6363845A637B636B84636B846B6B8C636B846B738C6B6B8C6B738C6B6B8C6B73946B73946B7394\
6B7394737B946B7394737B9C6B7394737B9C6B7B94737B9C6B7B94737B9C6B7B94
737B946B7B9473849C737B947384946B7B8C7384946B7B8C6B7B946B7B8C6B7B946B7B8C6B7B8C\
637B8C6B7B8C637B8C637B8C5A73845A7384526B7B526B734A636B4A636B425A63
425A63000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000313131424263424A63525273\
5A5A73636B84636B846B73846B7384737B8C737B8C7B84947B7B8C7B84947B848C
7B848C7B848484848C8484848484847B7B7B7B847B737B6B73736B7373636B6B5A5A5A4A525239\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000394A5A
394A63394A5A4252634252634252634252634A5A6B4A526B4A5A6B4A526B525A73525A73525A73\
525A735A637B525A735A637B525A735A637B5A63735A637B5A5A735A637B5A6373
5A637B5A5A7363637B5A5A735A637B5A5A7363637B5A5A7363637B5A5A7363637B5A5A7363637B\
5A5A7363637B635A7363637B5A5A7363637B635A7363637B5A5A7363637B635A73
635A7B5A5A7363637B635A73635A735A5A73635A7B635A73635A735A5A73635A7B635A73635A73\
5A5273635A7B635A73635A735A5273635A735A5A735A5A735A5273635A735A5273
5A5A735A52735A5A735A52735A5A735252735A5A735A52735A5A735252735A5A7B5252735A5A73\
5252735A5A7B525A7B525A7B52527B5A5A7B525A7B525A7B4A5273525A7B525A7B
525A7B4A5A7B52637B525A7B52637B4A5A7B5263845263845263844A637B526B844A637B4A637B\
4A637B526B844A6B7B526B7B4A6B7B5273844A6B7B52737B4A6B7B52737B4A6B73
4A6B734A6B6B4A73734A6B6B4A6B6B42635A425242000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000039395231315239395A39395A42426342425A4A4A634A4A6352526B523939\
52526B52526B52527352526B525A73525A735A5A7B525A735A5A7B5A5A7B636384
5A637B636384636384636B8C636B846B6B8C6B6B8C6B738C636B8C6B738C6B738C6B73946B738C\
6B73946B73946B7B946B73946B7B946B7B94737B9C6B73946B7B946B7B94737B94
6B7B8C6B7B946B7B8C6B7B946B7B8C6B7B8C6B7B8C6B7B8C63738C637B8C637384637B8C637384\
637B8C5A7384637384526B7B526B734A63734A6373425A63425A63000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039395A424A634A526B\
5A5A735A637B636B84636B846B73846B738C737B94737B8C7B84947B848C7B848C
7B7B847B848C7B848484848C7B847B7B847B7B7B737B7B7373736B7373636B6B5A63634A524A39\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000314252394A63394A5A394A63394A63425263394A6342526B42526B4A5A6B42526B\
4A5A6B4A5A6B525A734A5A73525A73525A735A637B525A735A637B525A735A637B
525A735A637B5A5A735A637B525A735A637B5A5A735A637B5A5A735A637B5A5A7363637B5A5A73\
5A637B5A5A7363637B5A5A7363637B5A5A7363637B5A5A73635A7B5A5A7363637B
5A5A73635A735A5A73635A7B5A5A735A5A735A5A73635A7B5A5A735A5A735A5A73635A7B5A5273\
5A5A735A5A73635A7B5A52735A5A735A5A735A5A735A52735A5A735A52735A5A73
5252735A5A735A52735A5A7B5252735A5A735252735A5A7B5252735A5A735252735A5A7B525273\
525A7B525A7B5A5A7B525273525A7B525A7B525A7B4A5273525A7B4A5A7B525A7B
4A5A73525A7B4A5A7B5263844A5A7B52637B4A637B5263844A637B4A637B4A637B4A637B4A637B\
4A637B4A637B526B7B4A637B4A6B7B4A6B7B52737B4A6B734A6B734A6B734A6B73
426B6B4A6B6B426B634A6B63425A4A000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000039395A39395239425A39395A424A634242634A4A6B4A4A6B525273\
52526B5252734A526B525A73525A735A5A7B525A735A637B5A5A7B5A637B5A637B
636B84636384636B84636B846B738C636B8C6B738C636B8C6B738C6B738C6B739463738C6B7394\
6B73946B73946B73946B7B9C6B73946B7B946B7394737B946B7B946B7B946B738C
6B7B946B7B8C6B7B8C63738C6B7B8C63738C637B8C637384637B8C6373846373845A7384637384\
5A6B7B526B7B4A6373526B734A636B42636B395A5A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039395A42425A4A526B\
525A735A637B63637B6B73846B6B8473738C6B738C7B84947B7B8C7B84947B7B8C
7B848C7B848C7B848C7B84848484847B847B7B7B7B73736B7B7B73737363736B63635A4A5A5242\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031394A394A5A314A5A395263394A63425263394A6342526B\
42526B4A526B42526B4A5A734A5A6B525A734A5A6B526373525A73525A73525A73
5A637B525A735A637B525A735A637B5A5A735A637B525A735A637B5A5A735A637B5A5A735A637B\
5A5A735A637B5A5A735A637B5A5A735A637B5A5A7363637B5A5A735A5A7B5A5A73
63637B5A5A735A5A735A5A735A5A7B5A5A735A5A735A5A735A5A7B5A5A735A5A735A5A735A5A7B\
5A5A735A5A735A5A73635A7B5A5A735A5A735A52735A5A7B5A5A735A5A73525273
5A5A7B525A735A5A735252735A5A7B525A735A5A7B5252735A5A7B525A735A5A7B5252735A5A7B\
525A7B525A7B525273525A7B525A7B525A7B4A5273525A7B4A5A7B525A7B4A5A73
52637B4A5A7B52637B4A5A7B5263844A637B52637B4A637B5263844A637B4A637B4A637B4A6B7B\
4A637B4A6B7B4A6373526B7B4A6B7B4A6B7B4A6B734A6B734A6B6B4A6B6B42636B
4A6B6B426B63426B63395242000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
0000000000000000000000004A392939395A31395239425A42425A424A634242634A4A6B4A4A6B\
5252734A526B525273525273525A7B525A735A5A7B5A5A7B5A63845A5A7B5A6384
5A6384636B8C5A6384636B8C636B8C6B738C636B8C6B738C636B8C6B738C636B8C6B739463738C\
6B739463738C6B73946B73946B7B9463738C6B7B946B738C6B7B9463738C6B7B8C
63738C6B7B8C63738463738C63738463738C5A73845A73845A73845A7384526B7B5A737B526B7B\
526B73425A6B42636B425A63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003131424242634A4A63\
525A735A5A73636B7B636B846B73846B6B84737B8C737B8C7B8494737B8C7B848C
7B848C7B848C7B7B847B84847B84848484847B7B737B7B7373736B73736B6B6B5A6363525A5242\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000313139394252314A5A394A63394A63\
425263394A634252634252634A526B42526B4A526B4A526B525A734A526B525A73
4A5A73525A734A5A73525A73525A735A637B525A73525A73525A735A637B525A735A5A7B525A73\
5A637B525A735A5A7B5A5A735A637B525A735A5A7B5A5A735A637B525A735A5A73
5A5A735A5A7B525A735A5A735A5A735A5A7B525A735A5A735A5A735A5A7B525A735A5A735A5A73\
5A5A7B525A735A5A735A5A735A5A7B5252735A5A73525A735A5A7B5252735A5A73
525A735A5A7B525273525A7B525A735A5A7B525273525A7B525A735A5A7B525273525A7B525A7B\
525A7B4A5273525A7B4A5A73525A7B4A5273525A7B4A5A73525A7B4A5A734A5A7B
4A5A7B52637B4A5A7B4A637B4A637B5263844A5A7B4A637B4A637B4A637B4263734A637B4A6373\
4A6B7B4263734A6B7B4A6B734A6B7B42636B4A6B7342636B4A6B6B426363426363
42636342635A000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000031395239425A39425A4242634242634A4A6B4A4A6B4A526B\
4A526B525A73525A73525A7B525A735A637B5A5A7B5A63845A637B636B845A6384
636B845A6384636B8C636B8C636B8C636B8C6B738C63738C63738C63738C6B739463738C6B7394\
63738C6B7B946B73946B7B9463738C6B7B9463738C6B7B8C63738C6B7B8C63738C
63738C5A7384637B8C6373846373845A73846373845A73845A7384526B7B526B7B4A636B425A63\
39525A39525A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039313939425A4A4A6B\
4A526B5A637B63637B636B84636B8473738C73738C737B8C737B8C7B84947B848C
7B848C7B848C84848C7B84848484847B847B84847B7B7B737B7B7373736B73736B6B6352635A4A\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000313942\
314252395263394A63425263394A6342526B42526342526B42526B4A5A734A526B
4A5A734A526B525A734A5A73525A734A5A73526373525A73525A73525A735A637B525A735A637B\
525A735A637B5A5A7B5A637B525A735A637B5A5A7B5A637B525A735A637B5A5A7B
5A637B525A735A637B5A5A7B5A637B525A735A637B5A5A7B5A5A7B525A735A637B5A5A7B5A5A7B\
525A735A637B5A5A7B5A5A7B525A735A637B5A5A7B5A5A7B525A735A637B525A7B
5A5A7B525A7B5A637B525A7B5A5A7B525A7B5A637B525A7B5A5A7B525A7B5A637B5A4A63525A7B\
525A7352637B525A7B525A7B4A5A7352637B4A5A7B525A7B4A5A7B52637B4A5A7B
4A637B4A5A7B5263844A637B4A637B4A637B5263844A637B4A637B4A63734A6B7B4A637B4A6B7B\
4263734A6B7B4A6B7B4A6B7342636B4A6B7342636B42636B426363426B63426363
42524A000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000031395A39395A39426339425A424263424A634A526B424A6B\
4A52734A5273525A7B525A73525A7B525A7B5A6384525A7B5A63845A63845A6B84
5A63845A6B845A6B84636B8C5A6B84636B8C636B8C63738C636B8C63738C63738C6B739463738C\
63739463738C6B7B9463738C63738C63738C63738C5A738463738C5A738463738C
5A6B845A73845A73845A7384526B7B5A6B7B526B7B526B7B4A636B4A636B425A5A39525A000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039425A424263\
4A526B525A736363845A637B636B846B7384737B8C73738C7B7B947B7B8C7B8494
7B7B8C7B848C7B848484848C7B84848484847B7B7B7B7B7B73736B73736B7373636B6B5A5A5A4A\
525239000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000031394239425A394A5A395263394A6342526342526342526B
42526342526B42526B4A5A6B42526B4A5A6B4A526B4A5A734A526B4A5A734A5A73525A734A5A73\
525A73525A7352637B525A7352637B525A735A637B525A735A637B525A7B5A637B
525A735A637B525A7B5A637B525A735A637B525A7B5A637B525A735A5A7B525A7B5A637B525A73\
5A5A7B525A7B5A637B525A735A5A7B525A7B5A637B525A735A5A7B525A7B5A637B
525A7B525A7B525A7B5A637B525A73525A7B525A7B525A7B525A73525A7B525A7B52637B4A5A73\
525A7B4A5A7B52637B4A5A734A5A7B4A5A7B52637B4A5A734A637B4A5A7B52637B
4A5A7B4A637B4A637B4A63844A5A7B4A637B4A637B4A637B4263734A637B4263734A6B7B426373\
4A6B7B4263734A6B7342636B42636B42636342636B395A5A426363395A5A394A42
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000000000039395A39395A424263424263424A6B424A6B4A5273\
4A5273525A7B4A5273525A7B525A7B52637B525A7B5A638452637B5A638452637B
5A6B845A63845A6B845A6B84636B8C5A6B84636B8C5A6B8463738C636B8C63738C63738C6B7394\
63738C63738C5A738463738C63738463738C5A6B8463738C5A73845A73845A6B7B
5A73845A6B7B5A737B526B7B5A737B4A63734A636B425A6342525A000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000313952424A63\
4A526B525A735A637B636B84636B846B73846B7384737B8C737B8C7B8494737B8C
7B848C7B848C7B848C7B848484848C7B84847B847B7B7B737B7B7373736B73736B6B6B5A636352\
525239000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000039424A394A5A395263394A63
42526B39526342526B394A6342526B42526B42526B42526B4A5A7342526B4A5A6B42526B4A5A73\
4A5A6B4A5A734A526B525A73525A73525A734A5A7352637B525A73525A73525A73
52637B525A73525A7B525A735A637B525A73525A7B525A735A637B525A7B52637B525A735A637B\
525A7B52637B525A735A637B525A7B525A7B525A735A637B525A7B525A7B525A73
52637B525A7B525A7B4A5A73525A7B525A73525A7B4A5A73525A7B525A7B525A7B4A5A7352637B\
4A5A7B525A7B4A5A7352637B4A5A7B4A5A7B4A5A7352637B4A5A7B4A637B4A5A7B
52637B4A637B4A637B425A734A637B4A63734A637B4263734A637B4263734A63734263734A6B7B\
42637342637342636B42636B426363426363395A5A42636339524A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000000000031395239425A39425A424A63394263424A6B424A6B\
4A52734A5273525A7B4A5A73525A7B4A5A7352637B525A7B5A638452637B5A6384
52637B5A6B8452637B5A6B845A6B845A6B8C5A6B845A6B845A6B8463738C5A6B8463738C5A6B84\
63738C5A6B845A73845A6B845A73845A6B7B5A73845A6B7B5A7384526B7B526B7B
526B7B526B7B4A63734A636B425A63425A5A314A52000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000031394A39425A\
4A526B4A526B5A637B5A637B636B84636B846B73846B738C737B8C737B8C7B848C
7B7B8C7B848C7B7B847B84847B84848484847B7B7B7B7B7B737B737B7B73737363737363636352\
5A5A42000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000313142
31424A394A5A314A5A395263394A6342526B394A6342526B42526342526B42526342526B42526B\
4A5A6B42526B4A5A6B4A526B4A5A734A526B4A5A734A5A73525A734A526B525A73
4A5A73525A734A5A73525A734A5A73525A734A5A73525A734A5A73525A7B4A5A73525A73525A73\
525A7B4A5A73525A7B525A7352637B4A5A73525A7B525A73525A7B4A5A73525A73
4A5A73525A7B4A5A73525A734A5A73525A7B4A5A734A5A734A5A73525A7B4A5A734A5A7B4A5A73\
52637B4A5A734A5A7B4A5A734A637B4A5A734A637B4A5A7B4A637B425A734A637B
425A734A637B425A73426373425A734A6373425A734263734263734A637342637342637342636B\
426373395A6342636B395A63426363395A5A395A5A393131000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000000000039425A39425A394263394263424A6B424A6B425273\
42526B4A5A7B4A5A7B525A7B4A5A7352637B525A7B52637B525A7B5A638452637B
5A638452637B5A6B845A6B845A6B845A6B845A6B8C5A6B845A738C5A6B8463738C5A738C5A738C\
5A6B845A738C5A6B845A7384526B7B5A7384526B7B5A6B7B526B7B526B7B526B73
4A6373425A63425A63395252000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000031313939425A\
424A63525A735A5A7363637B636B7B6B73846B7384737B8C737B8C7B84947B848C
7B848C7B7B8C84848C7B848C84848C7B84848484847B7B7B7B7B7373736B7B7B6B7373636B6B5A\
5A5A42000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031394239424A31424A394A5A394A5A395263394A6342526B42526342526B\
395263425A6B42526B425A6B42526B4A5A73425A6B4A5A734A526B4A5A734A5A73
4A5A734A5A6B4A5A734A5A734A5A734A5A73525A734A5A734A5A734A5A73525A7B4A5A73525A73\
4A5A7352637B4A5A73525A7B4A5A7352637B4A5A73525A7B4A5A7352637B4A5A73
4A5A7B4A5A73525A7B4A5A734A5A7B4A5A7352637B4A5A734A5A7B4A5A7352637B4A5A734A637B\
4A5A7352637B4A5A7B4A637B4A5A734A637B4A637B4A637B425A734A637B426373
426373425A734A6373426373426373425A6B42637342637342637342636B42637342636B4A3142\
4231394A3142395A63396363395A5A395A52000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000000000000000039425A31395A394263394263424A6B424A6B\
4252734A52734A5A7B4229294A5A7B4A5A7B52637B4A5A7B52637B52637B526384
52637B52638452637B5A6B8452637B5A6B845A6B845A6B84526B845A6B845A6B845A7384526B84\
5A6B84526B7B5A6B84526B7B526B7B526B7B526B7B4A63734A6373425A6B425A63
395252000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000031395A\
424A634A526B5A5A7B5A637B636B84636B846B738C6B7384737B8C737B8C7B8494
737B8C7B848C7B848C84848C7B84848484847B84847B847B737B737B7B7373736B73736B636352\
635A4A524A39000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031314239394A31424A394A5A39425A\
39526339526342526B39526342526B42526B425A6B42526B42526B42526B4A5A73
42526B425A6B42526B4A5A7342526B4A5A7342526B4A5A7342526B4A5A734A5A734A5A7342526B\
4A5A734A5A734A5A734A5A734A5A734A5A734A5A7B4A5A734A5A734A5A734A5A73
425A734A5A734A5A734A5A7B425A734A5A734A5A734A637B425A734A5A734A5A734A637B425A73\
4A5A7B4A5A734A637B425A734A6373425A734A637B425A73425A73425A73426373
395A6B425A73425A6B426373395A6B42636B425A6B426373395A6B42636B425A63425A6339525A\
395A5A395252000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000000000000000000000039313131425A394263394263424A6B\
424A6B42527342526B4A5A7B4229314A5A7B4A5A7352637B4A5A7B52637B4A637B
52638452637B52638452637B5A6B84526B84526B8452637B5A6B84526B7B526B7B52637B526B7B\
526B7B526B7B4A637B526B7B526B7B4A6373425A6B425A6339525A395252000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000039425A\
4242634A526B52527363637B63637B636B84636B8473738C73738C737B8C737B8C
7B84947B848C7B848C7B7B8484848C7B84848484847B7B7B7B847B737B73737B6B737363737363\
635A4A5A5242000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
31394239424A31424A394A5A394A5A395263394A6342526B39526B42526B39526B
425A6B42526B42526B42526B425A7342526B425A6B42526B4A5A7342526B425A7342526B4A5A73\
425A6B4A5A7342526B4A5A734A5A734A5A73425A734A5A734A5A734A5A73425A73
4A5A73425A734A5A73425A734A5A73425A734A5A73425A734A637B425A734A5A73425A734A637B\
425A734A6373425A734A6373425A73425A73425A6B426373425A6B425A73395A6B
426373395A6B425A6B395A6B42636B395A6B42636B395A63425A63395252395252314A4A000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000000000000000000000000000000000031395239425A394263\
424A6B394A6342526B42526B4A5A734252734A5A7B4A5A7352637B4A5A734A637B
4A637B52637B4A637B52637B4A637B526B7B4A637352637B4A637B526B7B4A63734A63734A6373\
4A63734A5A6B4A63734A636B4A6373000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000031314A\
424263424A635252735A5A73636B8463637B6B73846B7384737B8C737B8C7B7B8C
737B8C7B848C7B7B847B848C7B84848484847B847B7B847B7B7B737B7B7373736B73736B6B6B63\
6B6352525239000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000031394A31424A394A52314252394A63
394A63425263394A6339526339526342526B39526342526B39526B425A6B39526B42526B42526B\
425A7342526B425A6B42526B425A7342526B425A73425A6B4A5A7342526B425A73
425A734A5A73425A6B425A73425A734A5A73425A6B425A73425A6B425A73425A6B425A73425A6B\
425A73425A6B425A73425A6B425A73395A6B425A6B395A6B425A6B395263395A6B
395A6B395A6B395A63395A6339525A39525A394A4A39524A000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
00000000000000000000000000000000000000000000000000000000000000000031425A39425A\
39425A424A63424A6342526B42526B4A5A734A5A734A637B4A5A734A637B4A637B
4A637B4A5A734A637B4A63734A637B4A5A734A637B4A63734A6373425A6B4A637342636B4A6373\
425A6B4A636B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000313139\
39395A4A4A6B4A526B5A637B5A637B636B84636B846B738C6B7384737B94737B8C
7B84947B7B8C7B848C7B848C7B848C7B8484848C8C7B847B7B847B737B737B7B7373736B737363\
636352635A4A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
39424A31424A394A5A394252394A5A394252394A5A394A5A394A63394A5A425263395263425263\
39526342526B42526B42526B39526B425A73425A6B425A7342526B425A73425A6B
425A6B39526B425A6B42526B425A6B39526B425A6B42526B425A6B395263425A6B42526B425A6B\
395263425A6B425A6B425A6B395263425A6B395A6B395A6B395A63395A6B395A63
395A6339525A39525A394A4A394A4A000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39425A39425A394A63394A6342526B42526B425A7342526B425A73425A734A6373
425A6B425A73425A734A6373425A6B425A73425A6B426373425A6B425A6B425A6B000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39425A4242634A526B525A735A637B63637B6B7384636B846B738C737B8C7B8494
737B8C7B848C7B848C7B848C7B7B847B84847B84848484847B7B7B7B7B73737B6B7B7B6B737363\
6B6B5A5A5A4A5A5239000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031313942394A524A5A31424A394252314252394A52314252\
394A52394A52394A5A394A5A394A5A394A5A425263394A5A3952634A3131425263
395263395263395263425263394A5A39526339526342526339525A39526339526339526339525A\
395263395263395A6339525A395263395263395A6331525A393131392931000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000312929393131392931314252314252425263395263425A73312929
425A73425A6B425A6B39526B425A6B395A6B425A6B000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
313952424A63424A63525A735A637B636B7B636B7B6B73846B7384737B8C737B8C
7B84947B7B8C7B848C7B7B8484848C7B84848484847B847B84847B7B7B737B7B7373736B73736B\
6B6B5A63634A525239000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A4A394A39524A4252635263635A636B5A6B6B5A63736B73\
736B6B7B6B737B6B6B7B73737B737384737331424A394A52314A52394A52314A52
39525A394A5A394A5A314A5239525A394A5A39525A314A5239525A000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000395263000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
31394A39425A424A6B4A526B5A637B5A637B636B84636B846B738C6B7384737B8C
737B8C7B848C737B8C7B848C7B84847B848C7B7B847B84847B7B7B7B847B73736B73736B737363\
737363635A4A5A5242000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000042394A4A394A524A5A5A525A6B5A6B635A636B636B73636B\
7B6B73736B6B7B73737B6B6B7B73737B6B6B7B736B7B736B7B736B736B63737363
6B635A636352525242000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
313131394263424A634A5273525A73636B84636B846B6B846B7384737B8C737B8C
7B8494737B8C7B84947B848C7B848C7B848484848C7B84848484847B7B737B7B7B737B6B73736B\
7373636B6B5A5A5A42000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031313142314A5242525A4A5A635263635A636B636B6B636B736B6B\
736B6B7B73737B73737B73737B736B8473737B736B7B736B736B637B7363736B63
6B6B5A5A5A4A5A5A4A000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000039395A424A634A4A6B525A735A637B636B84636B7B6B73846B7384737B8C
737B8C7B848C7B7B8C7B848C7B7B847B848C7B84848484847B7B7B7B847B7B7B737B7B73737363\
7373636B635A635A4A524A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031293142394A4A42525A525A6352636B5A636B5A6373636B73636B\
7B6B737B6B737B73737B6B6B7B73737B736B8473737B736B737363736B63736B63
6363525A5A4A424239000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000031395A39425A4A526B525A735A637B5A637B636B846B6B846B738C6B7384
737B94737B8C7B848C737B8C7B848C7B84847B84847B84848484847B7B7B7B7B7B73736B7B7B73\
7373636B6B5A5A5A4A5A5242000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004231424A394A524A525A4A5A6B5A6B6B5A636B636B6B636B7B6B73\
7B6B737B73737B6B6B7B73737B736B7B73737B736B7B736B736B63736B636B635A
6B6B5A5A5A4A000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000031314239425A424A63525A73525A735A637B636B7B6B73846B6B8473738C
737B8C7B8494737B8C7B848C7B7B8C7B848C7B7B847B84847B84847B847B737B737B7B7373736B\
73736B6B6B5A6363525A5242000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004231424A4252524252635263635A636B636B6B5A6373636B736B6B\
7B7373736B6B7B73737B736B7B73737B736B7B736B7373637373636B6B5A6B6B5A
636352525242000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000031313939425A424A634A526B5A637B5A637B636B84636B846B738C6B738C
737B8C737B8C7B84947B848C7B848C7B848484848C7B84848484847B7B7B7B847B7B7B737B7B73\
7373637373636363525A5A42524A31000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031313142394A4A39525A4A5A6352636B5A636B5A6373636B73636B7B6B73\
7B6B6B7B73737B73737B73737B736B847B737B736B7B736B736B63736B636B6B5A
6363524A4A39000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000039425A39425A4A526B525A735A637B5A637B636B84636B84737B8C
6B7384737B8C737B8C7B848C737B847B848C7B848484848C7B84847B84847B7B7B7B7B7B73736B\
73736B7373636B6B5A5A5A42525239000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000039313942394A5242525A4A5A635A63635A636B636B6B636B736B73736B6B\
7B73737B6B6B7B73737B6B6B7B736B7B736B7B736B736B63736B636B6B5A6B6B5A
5A5A4A4A4A39000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000313952424A634A4A6B525A735A637B636B84636B846B73846B7384
737B8C737B8C7B848C737B8C7B848C7B848C7B848C7B84848484847B84847B847B737B737B7B73\
73736B73736B6B6B5A6B63525A5242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003931424A42525242525A525A6352636B636B6B636B73636B73636B7B7373\
7B6B737B73737B6B6B7B73737B736B7B736B7373637B736B736B636B6B5A636352
5A5A4A424A31000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000031314A39425A4A526B4A526B5A637B5A637B636B84636B7B6B7384
6B7384737B8C737B8C737B8C737B8C7B848C737B847B84847B84848484847B7B7B7B7B7B737B73\
7B7B736B6B637373636B6352635A4A524A31000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000042314A4A394A524A5A5A4A5A635A636B5A6373636B6B636B736B6B7B6B6B\
7B73737B6B6B7B73737B736B7B73737B736B7B736B736B63736B636B6352636352
525242000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004A4A3139425A424A63525A73525A735A637B5A637B6B73846B7384
6B7B8C6B7B8C7B8494737B8C7B848C7B7B847B848C7B84847B84847B84848484847B7B7B7B7B73\
73736B7B7B6B7373636B6B5A635A4A5A5A42000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000031313942394A524252524A5A635A63635A636B636B6B636B736B737B6B6B7B7373\
7B6B6B8473737B736B7B73737B736B847B737373637373636B6B5A6B6B5A63634A
525242000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000313952424A634A526B525A735A637B636B7B636B846B7384
6B7384737B8C737B8C7B8494737B8C7B848C7B84847B848C7B84847B84847B847B7B847B737B73\
7B7B7373736B73736B6B6B5A6B63525A5A42525239000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003131394A394A4A42525A4A5A6352636B5A6B6B5A6373636B73636B7B6B737B6B6B\
7B73737B736B7B73737B736B7B736B7B736B7B736B6B6B5A6B6B5A6B63525A634A
424A39000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000031395A39425A4A526B4A526B5A637B5A637B636B84636B84
6B738C6B738C737B8C737B8C7B848C7B7B8C7B848C7B7B847B848C7B84847B84847B7B7B7B847B\
7B7B73737B737373637373636B6B5A63634A525239000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000039314242394A5A4A5A5A4A5A635263635A6373636B73636B736B73736B6B7B7373\
7B6B6B7B73737B6B6B7B73737B736B7B736B736B637373636B6B5A6B6B5A525242
4A4A39000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000394263424A634A5273525A735A6B7B5A637B636B84
6B7384737B8C6B7384737B8C737B8C7B848C737B847B84847B84847B84847B7B7B7B84847B7B7B\
7B7B7373736B73736B73736373736363634A5A5A42524A31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000393142524252524A5A63526363525A6B5A6B6B636B736B7373636B7B6B6B7B6B6B\
7B73737B736B7B736B7B736B7B736B736B637373636B6B5A6B6B5A636352525242
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000031395A424A634A526B525A735A637B636B84636B84
6B73846B7384737B8C737B8C7B848C737B8C7B848C7B848C7B848C7B84848484847B84847B847B\
7B7B737B7B7373736B73736B6B6B636B6B5A635A4A5A5239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000031313142424A5242525A4A5A5A525A6B5A6B6B636B73636B73636B7B6B737B6B6B7B7373\
7B736B847B737B736B7B736B7B736B7B736B736B63736B636B635A6363524A5242
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000039425A4A526B4A526B5A637B5A637B636B84
636B846B73846B7384737B8C737B8C7B848C7B848C7B848C7B7B847B84847B84848484847B7B7B\
7B847B7B7B737B7B737373637373636B6B5A6B63525A5239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000039313942394A524A5A5A4A5A635263635A6373636B73636B736B6B736B6B7B73737B6B6B\
7B73737B736B8473737B6B6B7B736B7373637373636B635A6B6B5A5A5A4A4A5242
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000039425A3942634A5273525A735A637B5A637B
6B73846B73846B738C6B738473848C737B8C7B848C737B847B848C7B84847B84847B7B84848484\
7B7B7B7B7B7B73736B7B7B73737363737363636352635A4A524A39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003931424A42525242526352636352636B5A6B6B636B736B73736B6B7B6B73736B6B7B7373\
7B73737B73737B736B7B736B73736B7B736B736B63736B5A6B63525A5A4A424A31
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000292931394263424A63525A73525A735A6B7B
636B7B6B7384636B846B738C737B8C7B848C737B8C7B848C7B7B847B848C7B7B847B84847B847B\
8484847B7B737B7B7373736B73736B6B6B636B6B5A635A4A5A5242000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004239424A424A5A4A5A5A4A5A635A636B5A6373636B73636B7B6B73736B6B7B73737B6B6B\
8473737B736B7B736B7B736B7B736B736B63736B636B635A636352525242000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000039425A424A634A526B5A637B5A637B
636B84636B846B738C6B7384737B8C737B8C7B84947B848C7B848C7B7B847B848C7B84847B8484\
7B847B84847B7B7B737B7B7373736B73736B7373636B63525A5A425A5239000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
313139423942524A5A5A4A5A63526363526373636B73636B736B6B736B6B7B73737B6B73847373\
7B736B8473737B736B7B736B7373637B736B736B636B6B5A5A5A4A525242000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000031394A39425A4A526B4A5A735A637B
5A637B636B84636B846B738C6B7384737B8C737B8C7B848C737B847B848C7B84847B848C7B7B84\
7B84847B847B7B847B737B737B7B6B7373637373636B635263634A5A5239000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3131394A424A5242525A525A5A525A6B5A636B636B736B6B73636B7B6B6B7B6B6B7B73737B736B\
7B73737B736B7B736B736B637B736B737363736B636363525A634A424A39000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000394263424A634A5273525A73
636B84636B7B637384636B84737B8C737B8C737B8C737B8C7B848C7B848C7B848C7B7B847B8484\
7B84847B84847B7B737B7B7B737B6B73736B736B637373636363525A5A42000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4239424A4252524A5A5A4A5A635A636B5A6373636B73636B7B6B73736B6B7B73737B6B6B847373\
7B736B7B73737B736B7B736B7373637373636B6B5A6B63525A5A4A393931000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000313952424A63424A63525A73
5A637B636B84636B7B6B73846B7384737B8C6B7B8473848C737B8C7B848C737B847B84847B8484\
7B84847B7B7B7B847B7B7B737B7B7373736B73736B736B636B6B5A5A5A425A5239
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000313131
393942524252524A525A525A6352636B636B6B636B73636B736B6B7B6B73736B6B7B73737B7373\
8473737B6B6B7B736B7373637B736B6B6B5A6B6B5A6363525A5A4A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000039425A42526B4A526B
5A637B5A637B636B84636B846B73846B7384737B8C737B8C7B848C737B8C7B848C7B84847B8484\
7B84848484847B847B7B847B7B7B737B7B7373736B7373636B6B5A6B63525A5242
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000313139
42394A5242525A525A6352636B5A636B5A63736B7373636B7B6B73736B6B8473737B7373847373\
7B736B8473737B736B7B736B736B637373636B635A636352525242000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000031425A3942634A526B
4A5A735A637B5A6B7B637384636B846B73846B7B8473848C737B8C7B848C7B848C7B848C7B7B84\
7B84847B84848484847B7B7B7B7B7B7B7B737B7B6B7373637373636B6352635A4A
524A39000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000393942
42394A524A5A5A4A5A635263635A6373636B73636B7B6B73736B6B7B6B737B73738473737B736B\
7B736B7B736B7B736B736B637373636B6B5A6B6B5A5A5A4A424A39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000394263424A63
525A735263735A6B7B5A6B7B6B73846B73846B7B8C6B7B847B848C737B8C7B848C737B847B848C\
7B84847B84847B847B8484847B7B7B7B7B7373736B7B7B6B737363736B5A635A4A
5A5242000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000393139
524252524A525A525A63525A6B636B6B636B736B6B73636B7B6B737B6B6B7B73737B736B847373\
7B736B7B736B7373637B736B736B636B6B5A636352636352313931000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000313952424A63
4A526B525A7352637B636B7B636B7B6B73846B7384737B8C737B8C7B848C737B847B848C7B7B84\
7B84847B7B847B84847B847B7B847B737B737B7B7373736B73736B6B6B5A6B6352
5A5A425A5239000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000292931423942
5A5A4A5A4A5A5A525A635A636B5A63736B6373636B736B6B736B6B7B73737B6B6B7B73737B736B\
7B7373736B6B7B736B7373637373636363526363525A5A4A000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000313942394263
42526B4A526B5A637B5A637B636B84636B846B738C6B7384737B8C737B8C7B848C7B848C7B848C\
7B7B8484848C7B84848484847B7B7B7B847B7B7B737B7B737373637373636B6B5A
63634A5A5239000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000424A3942394A
5242525A4A5A635A63635A636B636B6B636B7B6B737B6B737B73737B6B6B8473737B73737B7373\
7B736B7B736B7B736B7B736B6B6B5A6B6B5A6363525A5A4A000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000039425A
3942634A526B525A735A6B7B5A637B636B846373846B7B8C6B7384737B8C737B8C7B848C737B84\
7B848C7B84847B84847B7B7B7B84847B7B7B7B7B7B73736B73736B737363737363
63634A5A5A42524A31000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003931424A4252
5242526352636352636B5A6B6B5A6373636B736B6B7B73737B6B6B7B73737B736B8473737B736B\
7B736B7B736B7B736B736B636B6B63636352636352424A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000031314A
424A6342526B525A73526373636B84636B7B6373846B7384737B8C737B8C73848C737B847B848C\
7B84847B84847B7B848484847B84847B847B7B7B737B7B7373736B73736B736B63
6B6B5A635A4A5A5239000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000042314A4A394A
5A4A5A63525A635A636B5A6373636B73636B7B6B73736B6B7B73737B736B7B73737B736B847B73\
7B736B7B736B736B6373736B6B6B5A6363525A5A4A394239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
31425A424A634A526B525A735A637B636B84636B7B6B73846B7384737B8C737B84737B8C737B84\
7B848C737B847B84847B84847B84847B7B7B7B7B7B737B737B7B73737363737363
6B6B5A6B63525A5242524A39000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000393139423142524252
5A4A5A6352636352636B5A6B6B636B736B73736B6B7B6B737B6B6B7B73737B6B6B7B73737B736B\
7B7373736B63737363736B636B6B5A5A63525A5A4A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
31425A39425A4A526B525A735A637B5A637B6373846373846B73846B738473848C737B8C7B848C\
737B847B848C7B84847B84847B84848484847B847B7B7B7B737B6B7B7B73737363
7373636B63526363525A5242524A39000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003931424A424A524252
63526363525A6B5A6B6B636B73636B73636B7B73737B6B737B73737B736B847B737B7373847B73\
7B736B7B736B73736373736363635A636352525A4A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000039425A424A634A5A73525A735A637B5A6B7B637384636B846B73846B7B8473848C737B84\
7B848C737B847B848C7B7B847B84847B84848484847B7B737B7B7373736B7B7B6B
736B63736B636B6352635A4A524A39000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004231424A394A5A4A5A
5A4A5A635A636B5A6373636B73636B7B6B737B6B6B7B73737B6B6B7B73737B736B847B737B736B\
7B736B7373637373636B6B5A6B6B5A5A634A4A5242000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000031395A424A63424A6B525A735A637B5A6B7B5A6B7B6B73846B7384737B8C6B7B847B848C\
737B8C7B848C737B847B848C7B84847B84847B847B7B847B7B7B737B7B7373736B
73736B736B63736B5A63634A635A4A524A39000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003931394239425242525A4A52
635263635A636B5A6B6B636B7B6B737B6B6B7B73737B6B6B7B73737B73737B73737B736B7B7B73\
7B736B73736B6B6B637373636363525A5A4A313931000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000031313931425A424A634A526B5A637B5A637B636B7B636B7B6B73846B7384737B8C737B8C\
7B848C737B847B84847B7B847B84847B7B7B7B84847B7B7B7B847B5A6B73737B6B
73736B7373636B6B5A6B6B5A635A4A5A52424A4231000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003931394A394A4A42525A525A
63525A6B5A636B5A6373636B736B6B7B7373736B6B7B73737B736B7B73737B736B7B73737B736B\
7B736B736B637373636B6B5A636352525242000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000039425A424A634A526B525A735A6B7B5A6B7B636B84636B846B7B8C6B7B84737B8C\
737B8C7B848C7B84847B848C7B7B847B84847B84847B8484636B7B7B847B7B7B73
7B7B6B737363737363736B5A6B63525A5A425A5242000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004239424A394A5A4A5A63525A
635A63635A6373636B73636B7B6B737B6B6B7B73737B73737B73737B736B847B737B736B7B736B\
7373637B736B736B636B6B5A5A5A4A525A4A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000313142394263424A634A5A735263735A6B7B5A6B7B6373846373846B7B8C6B7B84\
73848C737B8C7B848C737B847B84847B84847B84847B7B7B7B847B7B7B7B7B7B7B
73736B73736B7373637373636B6B5A6B63525A52425A5239000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003131394231424A42525A4A5A635263
63525A6B5A636B636B736B73736B6B7B6B737B6B737B73737B736B7B73737B7373847B7373736B\
73736B736B637373636363525A5A4A4A5242000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031395A424A634A526B525A7352637B636B84636B7B6B73846B7384737B8C\
737B8C73848C737B847B848C7B7B847B84847B7B7B7B84847B847B7B847B7B7B73
7B7B7373736B73736B736B637373636B6B5A63634A5A5239524A39000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003929394A394A5242525A525A63525A
6B5A636B5A6373636B73636B7B73737B6B6B7B73737B6B6B8473737B7373847B737B736B7B736B\
7373637373636B6B5A636B5A525A4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031395239425A4A526B4A526B5A637B5A637B636B84636B7B6B73846B7384\
737B8C737B84737B8C737B847B848C737B847B84847B7B847B84847B7B7B7B847B
7B7B737B7B7373736B73736B736B63737363636352635A4A5A4A39000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003931424A394A5A4A525A4A5A635263
6B5A6373636B6B636B736B737B6B6B7B73737B6B6B7B73737B736B847B737B736B7B736B73736B\
73736B6B6B5A6B6B5A5A6352525A4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000039425A394A634A5A73525A735A637B5A637B6373846373846B7384\
6B738473848C73848C7B848C737B847B848C7B84847B84847B7B7B8484847B847B
7B847B7B7B737B7B7373736B73736B736B637373636B6352635A42524A39000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003131394239425242525A4A52635263635A63
6B5A636B5A637B6B737B6B737B73737B6B6B8473737B7373847B737B736B847B737B736B7B736B\
736B637373636B6B5A5A63524A5242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000394A63424A63525A73525A735A6B7B5A6B7B637384637384\
6B73846B7B8473848C737B847B848C7384847B848C737B847B84847B847B7B8484
7B7B7B7B7B7B7B7B737B7B73737363737363736B5A736B5A635A4A5A5242524A31000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003931394A424A52425263525A63525A6B5A63
6B5A6373636B736B6B7B73737B6B6B7B73737B73738473737B736B847B737B736B7B736B737363\
7373636B6B5A636B5A525A4A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031394A424A6342526B525A735A637B5A6B7B5A6B7B6B7384\
6B73846B7B846B7B8473848C73848C7B848C737B847B848C7B84847B84847B7B7B
8484847B7B7B7B7B7B73736B7B7B7373736B7373636B6B5A6B6B5A635A4A5A5239000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004239424A394A5A4A5A63525A635A636B5A63
73636B73636B7B6B73736B6B7B73737B73737B73737B736B847B737B73737B736B73736B7B736B\
736B636B73635A63525A5A4A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000031425A424A634A526B52637352637B5A6B7B636B7B\
6B73846373846B7B846B7B8473848C737B847B848C737B847B8484737B7B7B8484
7B7B7B7B847B7B7B737B7B7373736B7B7B6B736B637373636B6B5A6B63525A52425A4A39000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003931394231425242525A4A5A63526363525A6B5A63
6B636B7B6B73736B6B7B73737B6B6B7B73737B736B7B73737B73736B5A6373736B73736B737363\
737363636B5A5A63524A5242000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039426342526B4A526B5A637B5A6B7B636B84\
636B7B6B73846B7384737B8C737B847B848C73848C7B848C737B847B848C7B8484
7B84847B7B7B8484847B7B7B7B7B7373736B7B7B6B7373637373636B635A6B63525A5242524A39\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000039313952424A5A4A5263525A63525A6B5A636B6363
73636B736B6B7B73737B6B737B73737B736B847B737B7373847B737B736B7B7B7373736B737363\
6B6B5A6B6B5A5A5A4A4A5242000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031394A39425A42526B4A5A735A637B5A637B\
636B84636B846B73846B7384737B8C737B847B848C737B847B84847B84847B8484
7B7B7B7B84847B847B84847B7B7B737B7B7373736B7B736B736B63736B636B6352635A4A5A4A39\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000031293142394252424A634A5A5A4A5A6B5A636B5A6373636B
73636B7B6B737B6B6B7B73737B6B6B7B73737B7373847B737B736B7B736B73736B7B736B6B6B63\
6B73636363525A5A4A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000394263424A634A526B4A5A735A6B7B\
5A6B7B637384636B7B6B7B846B7384737B8C737B847B848C7384847B8484737B84
7B84847B84847B84847B7B7B7B847B7B7B737B7B7373736B7B736B736B63736B5A63634A635A4A\
524A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004239424A394A524A525A4A526B52636B5A6373636B6B636B
7B6B737B6B6B7B73737B6B6B8473737B7373847B737B736B847B737B736B7B736B736B63737363\
6B6B5A6363524A5242000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000394A63394A634A5A6B525A73\
5A6B7B5A6B7B6373846373846B7B846B7384737B8C737B847B848C737B847B8484
7B7B847B84847B7B7B7B847B7B7B7B7B847B7373737B7B7373736B73736B6B6B5A6B6B5A635A4A\
635242524239000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000031313942394252424A52425263525A63525A6B5A636B5A6373636B
736B6B7B7373736B6B7B73737B7373847B737B736B7B7B737B736B7B736B7373637373636B6B63\
6B6B5A525A4A4A5242000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000313142424A6342526B525A73\
526373636B84636B7B6373846373846B7B8C6B7B8473848C737B847B848C7B8484
7B84847B7B847B84847B84847B84847B7B7B7B847B7B7B737B7B6B7373637B736B736B636B6B5A\
635A4A63524A524A39000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003931394A394252424A63525A63525A6B5A636B5A6373636B736B6B
7B6B737B6B6B8473737B73738473737B7373847B737B7B737B7B7373736B7B736B7373636B7363\
5A63525A6352000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000031425A424A6342526B\
525A735A637B636B845A6B7B6373846B73846B7B8C6B7B8473848C737B847B848C
737B847B84847B84847B84847B7B7B7B847B7B7B7B7B7B7B73736B7B736B7373637373636B6B5A\
4A5A6B635A4A635242524239000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004239424A39425A4A525A4A5A6B5A63635A5A73636373636B7B6B73
736B6B7B73737B73738473737B7373847B737B7B737B7B7373736B73736B737363737363636B5A\
5A6352525A42000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000031425A42526B\
4A526B5263735A637B636B84636B7B6B73846B7384737B8C737B8473848C737B84
7B848C7384847B8484737B7B7B84847B847B7B847B7B7B737B7B7373736B73736B736B63737363\
6B6B5A6B6B5A635A4A63524A524A39000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004231394239425242525A4A5263525A63525A736363736363736B6B736B6B
7B73737B6B6B7B73737B7373847B737B7B737B7B737B736B7B7B7373736B73736B6B6B63636B5A\
5A5A4A525242000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000031314239425A\
42526B4A526B5263735A637B636B84636B7B6B73846B7384737B8C6B7B8473848C
737B847B8484737B847B84847B7B7B7B84847B7B7B7B847B7B7B737B7B7373736B73736B736B63\
7373636B63526B6B52635A4A63524A524A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000042313952424A5A424A63525A63525A6B5A636B5A6373636B73636B7B6B6B
7B6B6B7B73737B736B7B73737B7373847B737B736B7B736B73736B73736B6B6B636B6B5A5A6352\
525A4A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000039425A\
394A634A5A6B525A735A637B5A637B6373846373846B73846B7384737B8C737B8C
73848C737B847B848C7B84847B84847B7B7B8484847B847B7B847B7B7B737B7B7373736B73736B\
736B637373636B6B5A736B5A635A4A635A4A5A4A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3931394A424A52424A634A5A63525A6B5A636B5A6373636B73636B7B6B737B6B73
7B73737B736B847B737B7373847B737B7373847B737B736B7B736B737363737363636B5A5A6352\
4A5242000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
39425A424A634A5A6B4A5A735A6B7B5A6B7B637384636B7B6B73846B7B84737B8C
737B8473848C737B847B8484737B847B84847B847B7B84847B7B7B7B7B7B7B7B737B7B7373736B\
73736B736B637373636B635A6B6B5A635A4A635A4A524A39524231000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000393131\
4A394252424A634A52634A526B5A5A6B5A637363637363637B6B6B7B6B6B7B7373
7B6B6B7B73737B7373847B737B73737B7B737B736B7B7B6B7373637373636B6B5A636B5A525A4A\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
313139394A63424A63525A735263735A6B7B5A6B7B6373846373846B73846B7384
73848C737B84738484737B847B84847B7B847B84847B7B7B8484847B7B7B7B847B7B7B737B7B73\
73736B73736B736B637373636B6B5A736B5A635A4A635A4A5A4A39524231000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000312929423942\
4A39425A4A525A4A526B525A6B5A5A7363636B63637B6B6B7B6B6B7B73737B6B6B
8473737B7373847B737B7373847B737B736B7B7B7373736B73736B6B7363636B5A5A6352525A4A\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000313952394A6342526B525A735263735A6B7B5A6B7B63738463737B6B7384
6B7B8473848C737B84738484737B847B8484737B7B7B84847B7B7B7B847B7B7B737B7B7B7B7B73\
7B7B7373736B73736B736B637373636B6B5A736B5A635A4A635A4A524A39524231
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004231394A3942\
52424A5A4A4A63525A63525A735A636B5A6373636B736B6B7B73737B6B6B7B7373
7B7373847B737B736B7B7B737B73737B7B7373736B73736B7373636B73635A6352525A52000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000039425A424A6342526B5263735A637B5A6B7B5A6B7B6B73846B7384
6B7B846B7B847B848C7384847B8484737B847B84847B84847B84847B7B7B7B84847B7B7B7B847B\
7B7B737B7B737B736B7B736B736B63737363736B5A736B5A635A4A635A4A5A4A39
524231000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000042313942393952424A\
5A424A63525A63525A6B5A636B5A63736B6B736B6B7B6B6B7B6B6B8473737B7373
8473737B7373847B737B7B73847B737B736B7B7B7373736B73736B636B5A636B5A525A4A000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000031393931425A4252634A5A6B52637B5263735A6B7B636B7B6B7384
63737B6B7B84737B8473848C737B847384847384847B8484737B7B7B84847B7B7B7B847B7B7B73\
7B847B7B7B737B7B7373736B7B736B7373637373636B6B5A736B5A635A52635A4A
524A39524231000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A5A4A423139524242524242\
634A526352526B5A5A6B5A5A73636373636B7B6B6B7B6B6B7B73737B7373847B73
7B7373847B737B7B73847B737B736B7B7B6B73736B737B6B6B6B63636B5A525A52000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000031425A394A6342526B4A5A6B5A6B7B5A6B7B636B7B636B7B
6B73846B7384737B846B7B8473848C7384847B8484737B847B84847B84847B84847B7B7B7B847B\
7B7B7B7B847B7B7B737B7B737B736B7B736B736B63737363736B5A736B5A635A4A
635A4A5A4A395242394A3931000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000423139523942524242634A52\
6352526B525A6B5A5A7363637363637B6B6B7B6B6B7B73737B73737B73737B7373
847B737B7B73847B737B736B7B7B7373736B73736B6B73636B73635A6352525A4A000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000039425A394A634A526B4A5A735A6B7B5A6373636B7B
636B7B6B73846B737B6B7B846B7B8473848C737B847B84847384847B8484737B7B7B847B7B7B7B\
7B847B7B7B737B7B7B7B7B737B7B7373736B73736B736B637373636B6B5A736B5A
6B5A52635A4A5A4A42524A424A42394A3931000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003931314231315239425242425A4A4A634A52\
6B5A5A6B525A6B5A637363637B6B6B736B6B7B736B7B6B6B7B73737B736B7B7B73
7B7373847B737B736B7B7B7373736B737B6B6B73636B7363636B5A5A6352000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000394A63424A634A5A6B525A735A6B7B5A6B7B
6B6B6363737B6B7B846B7B84737B846B7B847B848C7384847B84847384847B84847B84847B8484\
7B7B7B7B84847B847B7B847B7B7B737B7B737B7B6B7B736B7373637B7363736B5A
736B5A6B5A526B635263524A5A52425242395242394A3931000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000423139423939524242524242634A526352526B5A5A\
6B5A5A73636373636373736B7B6B6B7B73737B736B7B73737B7373847B737B7B73
847B737B7B73847B737B7B6B7B7B6B73736B73736B6B7363636B5A525A4A000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000424A63424A634A5A6B525A735A6B7B
5A6B7B63738463737B6B7B8463737B6B7B846B7B8473848C737B847B84847384847B8484737B7B\
7B84847B7B7B7B847B7B7B7B7B847B7B7B737B7B7373736B7B736B737363737363
6B6B5A736B5A6B63526B6352635A4A63524A5A4A425A4A424A3931000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000423131423939524242524242634A4A6352526B5A5A6B5A5A\
7363637363637B6B6B7363637B6B6B7B6B6B7B73737B736B847B737B7B73847B73
7B7B737B7B737B7B737B7B7373736B73736B6B73636B6B635A6352000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000394A634252634A5A6B4A5A73
5A6B7B5A6B7B63737B636B7B6B7B846B73846B7B846B7B8473848C7384847B8484737B847B8484\
7B84847B84847B7B7B7B847B7B7B7B7B847B7B7B737B7B737B736B7B736B736B63
737363736B5A736B5A6B63527363526B63526B5A4A5A52425A5242524239524239000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004A31314A3131524242524242634A4A6352526B52526B5A52736363\
736363736B637363637B736B7B736B7B73737B736B847B737B7B73847B737B7B73
847B737B7B737B7B7373736B737B6B6B73636B73635A63525A6352000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000394A63394A634A5A6B
4A5A735A6B7B5A6B7B63737B63737B6B738463737B6B7B846B7B84738484737B84738484737B84\
7B8484737B7B7B847B7B7B7B7B847B7B7B737B7B7B7B7B737B7B7373736B73736B
7373637373636B6B5A736B5A6B635A736B5A6B5A526B6352635A4A63524A524A395242394A3931\
4A3939423131000000000000000000000000000000000000000000000000000000
3931314A31314231314A39395239395A4242524242636B63634A4A6B52526B5252735A5A736363\
736B6B7363637B6B6B7B6B6B7B73737B736B7B73737B7373847B737B73737B7B73
7B7B737B7B7373736B73736B6B736B6B73635A635A5263524A5A4A000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000394A63425263
4A5A6B4A5A735A6B7B5A6B7B637384636B7B6B7B846B73846B7B846B7B8473848C7384847B8484\
737B847B84847B84847B84847B7B7B8484847B847B7B847B7B7B737B7B737B7B6B
7B736B7373637B736B736B63736B636B6B5A736B5A6B63526B63526B5A4A6B5A4A6352425A4A42\
5242395242425242395242394A39315239394A31314A313142525A4239394A3939
4A39394A39395A42425A42425A4A425A4A426B52526B52526B52526B5A527363637363637B6B63\
736B637B736B7B736B7B73737B736B847B737B7373847B737B7B73847B7B7B7B73
7B7B7373736B7B7B7373736B6B736B636B5A5A635A525A4A000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000394A63
394A634A5A6B4A5A735A6B7B5A6B7B63737B63737B6B738463737B6B7B846B7B84738484737B84\
7384847384847B8484737B7B7B84847B7B7B7B847B7B7B7B7B847B7B7B737B7B73
6373737B7B6B7B736B7B736B736B63736B63736B5A736B636B635A73635A6B63526B635263524A\
6352425A4A426352425A4A395A4A425A42395A4A42524239524239523939524242
5242395A4A425A4A4263524A63524A6B52526B52526B5A5A6B5A5273635A73635A7B6B63736B63\
7B736B7B736B8473737B736B7B73737B7373847B737B7373847B737B7B737B7B73
73736B737B7373736B737B6B636B5A5A635A525A52000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
394A634252634A5A6B4A5A6B5A6B7B5A6B7B63737B636B7B6B73846B73846B7B846B7B84738484\
738484738484737B847B84847B847B7B8484737B7B7B847B7B7B7B7B847B7B7B73
7B7B737B7B737B7B7373736B7B7B6B73736B737363736B63737363736B5A736B5A6B635273635A\
6B5A526B5A4A63524A6B5A4A63524A63524A63524263524A634A42634A4A5A4A42
63524A63524A6B524A63524A6B5A526B5A5273635A6B5A5A73636373635A736B63736B637B736B\
7B736B7B73737B736B847B737B73737B7B737B7373847B737B7B737B7B7373736B
7B7B7373736B73736B636B63636B635263524A5A4A000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000394A63394A634A526B4A5A6B5A6B7B5A6B73636B7B636B7B6B738463737B6B7B846B7B84\
738484737B84738484737B847B8484737B7B7B847B7B7B7B7B847B7B7B737B7B7B
7B7B737B7B737B7B737B7B737B7B6B7B7B7373736B7B736B736B637373636B6B5A736B5A6B635A\
73635A6B5A4A6B5A526B5A526B63526B5A4A6B5A526B5A4A6B5A5263524A6B5A52
63524A6B5A526B524A6B5A526B5A5A73635A6B635A7363637363637B6B63736B637B736B7B6B6B\
7B73737B736B7B73737B7373847B737B73737B7B737B7B73847B7373736B737B73
73736B737B6B6B6B63636B63526352525A52000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000394A634252634A5A6B4A5A6B5A6B7B5A6B7B63737B636B7B6B73846B73846B7B84\
6B7B8473848C7384847B8484737B847B84847B84847B84847B7B7B7B847B7B7B7B
7B847B7B7B7384847B7B7B737B7B737B7B737B7B737B736B7B736B7373637B736B736B63736B63\
73635A736B5A73635A73635A6B635273635A6B635A73635A6B5A5273635A6B6352
73635A6B5A5273635A73635A73636373635A7B6B63736B637B6B6B736B637B736B7B736B7B7373\
7B736B847B737B7B73847B737B7B7384847B7B7B737B7B737B7B737B7B73737B6B
737B736B736B6B73635A635A526352000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000394A63394A63425A6B4A5A6B5A637B5A6B73636B7B636B7B6B738463737B\
6B7B846B7B84738484737B847384847384847B8484737B7B7B847B7B847B7B847B
7B7B7B7B847B7B7B7B7B847B7B7B737B7B737B7B73847B737B736B7B736B7B736B7B736B736B63\
736B63736B5A736B6373635A736B63736B5A736B6373635A73635A73635A736B5A
6B635A73635A73635A7B6B63736B637B6B637B6B637B736B7B6B6B7B736B7B736B7B73737B736B\
7B7B737B7B73847B7B7B7B73847B737B7B737B847B737B737B7B73737B6B737B73
6B736B6373635A635A52635A000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000394A63394A634A526B4A5A6B5A6B7B5A6B7B636B7B5A6B7B6B7384\
63737B6B7B846B7B7B738484737B84738484737B7B7B84847B84847B8484737B7B
7B84847B7B7B7B847B7B7B737B847B7B7B737B7B737B7B73847B737B7B737B7B7373736B7B736B\
7373637B736B736B637B736B736B637B7363736B637B7363736B63736B63736B5A
7B6B63736B637B6B63736B637B736B7B736B7B736B7B736B847B737B736B7B73737B736B847B73\
7B7B73847B737B7B7384847B7B7B737B7B73737B737B7B73737B6B737B736B736B
6B736B5A6B5A52635A000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000394A5A394A5A42526B4A5A6B5263735263735A6B7B636B7B\
63738463737B6B7B846B7B84737B84737B847384847384847B8484737B7B7B8484
7B847B7B8484737B7B7B847B7B7B7B7B847B7B7B737B7B737B7B737B7B737B7B737B7B737B736B\
7B7B6B73736B7B736B7B736B7B736B7373637B736B7B736B7B736B736B637B7363
736B63737373736B637B736B7B736B7B73737B736B7B73737B736B847B737B736B7B7B737B7B73\
847B7B7B7B737B7B737B7B737B7B73737B73737B7373736B737B736B736B6B736B
636B635A6B5A4A5A4A000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000394A63394A6342526B4A5A6B5A637B5A6B7B636B7B\
5A6B7B63738463737B6B7B846B7B7B738484737B84738484737B847B84847B8484
7B848473847B7B84847B847B7B847B7B7B7B7B847B7B7B7B7B847B7B7B7384847B7B7B737B7B73\
7B7B6B7B7B737B7B6B7B7B737B736B847B737B736B7B7B737B736B847B737B736B
7B736B7B736B847B737B736B7B7B737B736B847B737B7B73847B737B7373847B737B7B73847B7B\
7B7B7384847B7B7B737B7B7B737B737B7B73737B73737B736B736B6B7B6B6B736B
5A6B63525A52000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000394A5A394A5A42526B4A5A6B526373526373\
5A6B7B5A6B7B63738463737B6B7B846B7B7B737B846B7B7B737B84737B847B8484
737B7B7B84847B84847B8484737B7B7B847B7B7B7B7B847B7B7B737B847B7B7B737B847B7B7B73\
7B7B737B7B737B7B737B736B7B7B737B7B73847B737B736B7B7B737B7B73847B73
7B736B7B7B737B736B847B737B736B7B7B737B7B73847B737B7B737B7B737B7B73847B7B7B7B73\
7B7B7B7B7B737B7B7B737B73737B73737B73737B736B736B6B736B637363636B63
525A52000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000394A5A394A6342526B4A5A6B526373\
5A6B7B5A6B7B635A526B738463737B6B73846B737B6B7B846B7B7B737B846B7B7B
7384847384847B848473847B7B84847B847B7B847B737B7B7B847B7B7B7B7B847B7B7B737B847B\
7B7B737B7B737B7B737B7B7B7B7B737B7B737B7B73847B737B7B737B7B737B7373
847B737B7B737B7B737B7373847B7B7B7B737B7B737B7B7384847B7B7B737B7B7B7B7B7384847B\
7B7B737B7B73737B73737B73737B73737B736B736B6B736B636B635A6B63392931
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000314A5A394A5A4252634A5A6B\
52637352636B5A6B735A6B7B63737B636B7B63737B63737B6B7B846B737B6B7B84
6B7B7B7384846B7B7B7384847384847B8484737B7B7B847B737B7B7B847B737B737B7B7B7B7B7B\
7B847B7B7B737B7B7B7B7B737B847B7B7B737B7B737B7B737B7B7B7B7B737B7B73
7B7B737B7B737B7B737B7B737B7B7384847B7B7B737B7B7B7B7B737B847B7B7B737B847B7B7B73\
7B7B7B737373737B736B7373737B736B736B6B736B5A6B635A6B634A5A52312929
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000394A5A425263\
4252635263735263735A63735A6B7363737B63737B63737B63737B6B7B846B7B7B
6B7B846B7B7B738484737B84738484737B7B7B84847B84847B8484737B7B7B84847B847B7B847B\
7B7B7B8484847B847B7B847B7B7B7B84847B7B7B7B7B847B7B7B7384847B7B7B73
7B7B7B7B7B7384847B7B7B7B84847B7B7B7384847B7B847B7B847B7B7B737B847B7B7B7B7B7B7B\
737B73737B7B737B73737B736B736B6B736B5A6B635A63634A5A524A5A52000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3952634252634A5A6B4A5A6B5263735263735A6B7B5A6B7B63737B63737B6B7B84
63737B6B7B7B6B7B7B737B846B7B7B737B84737B7B7B8484737B7B7B848473847B7B84847B7B7B\
7B84847B847B7B84847B7B7B7B847B7B847B7B847B7B7B7B7B847B7B7B7B7B847B
7B7B737B847B7B847B84847B7B7B737B847B7B7B7B7B847B737B737B7B7B737B737B7B7B737B73\
737B736B7B73737B7363736B63736B5A6B6352635A4A5A52392929000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000313142394A63425A6B4A5A6B4A636B4A636B5A6B735A6B7363737B636B7B
6B737B63737B6B737B6B737B737B846B7B7B737B846B7B7B7384847384847B8484737B7B7B8484\
737B7B7B847B737B7B7B84847B847B7B847B7B7B7B7B84847B847B7B847B7B7B7B
7B847B7B7B7B7B847B737B737B847B737B7B7B7B7B737B737B7B7B737B73737B736B7373635252\
6B7B736B7B73636B6B63736B52636352635A4A5A52000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000031394A4252634252634A5A6B4A5A6B5263735263735A6B7B
5A6B7B63737B636B7363737B63737B6B7B7B6B737B6B7B7B6B7B7B7384846B7B7B737B7B737B7B\
738484737B7B737B7B737B7B7B847B737B7B7B847B737B7B7B847B737B7B7B847B
7B847B7B847B737B73737B7B737B73737B7B6B7B73737B736B7B73737B736B73736B7B735A4A4A\
6B7B735A6B6B5A6B6352635A52635A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000394A5A425263425A6B425A63526373526373
5A6B735A6B7363737B63737B63737B63737B6B7B846B7B7B6B7B7B6B7B7B737B846B7B7B737B7B\
6B7B7B738484737B7B737B7B737B7B738484737B7B73847B737B7B7B8484737B7B
737B7B737B7B73847B737B7B737B7B6B7B73737B7B6B7B736B7B736B73736B7B736B73736B7373\
5A6B635A6B6B52635A4A635A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000394A5A4252634252634A636B
4A5A6B5263735A6B7363737B5A6B7363737B63737B6B7B7B63737B6B7B7B6B7B7B6B7B846B737B\
6B7B7B6B7B7B737B7B6B737B6B7B7B6B7B7B737B7B6B7B7B737B7B737B7B737B7B
6B7B73737B7B6B7B7B737B7B6B73736B7B736B73736B7B736373736B73736373736373735A6B63\
526B634A635A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000313139395263395263
4A5A6B4A5A6B52636B52636B5A6B7B5A6B7363737B636B736B737B63737B6B737B63737B6B7B7B\
6B737B6B7B7B6373736B7B7B6B737B6B7B7B6B7373737B7B6B7B7B6B7B7B635252
737B7B6B7B7B6B7B7B6B73736B7B7B6B737363737363736B6373735A6B6B5A6B6B526B63526B63\
4A5A5A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000031424A
394A5A4252634252634A5A6B4A636B5A6B7352636B5A6B735A6B735A73735A6B736373735A6B73\
63737B63737363737B6373736B7B7B6373736B737B6373736B7B7B6373736B7B7B
6B73736B7B7B63736B6373735A6B6B637373526B6B5A6B6B526363526B634A635A526363000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
423129314A524252634252634A5A6B4A5A6352636B4A636B52636B52636B526B7352636B5A6B73\
526B6B5A73735A6B736373735A6B7363737B63737363737B6373736B7B7B637373
6373735A6B6B5A6B735A6B6B526B6B5263635263634A63634A6363425A5A4A635A000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031394A39525A425A634252634A5A634A5A634A636B4A5A63\
4A636B4A636B52636B52636B526B6B526B6B5A6B735A6B6B5A6B735A6B6B5A6B6B
52636352636B4A6363526363425A5A4A5A5A425A5A425A5A313939000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000031313939424A39525A425A63\
425A63425A63425A634A636B4A5A634A636B4A636352636B52636B52636B4A5A63
4A6363425A5A394A4A000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000313139394A5239525A425A6339525A425A63425A63313942000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[13]:=",
  ImageSize->{288, 286.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P00017PP10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol0
4P081`001`Oo018020L000L7o`0B00P700071i/000Q[LF]aJg5[LFh020L000L7V003L@0;RW6:LHYa
RW6:LHX00g5X00P700071i<001][LF]aK8YaRW6:LHYaRW6:LHYaRW6:LHY[LF/0H`081`001`NA00Aa
00>:LHX00XX01W6@RX^:R`>:00>@RXX01XX00g6:L@1R00P700071hd002E[LF^:LHYaRW6:LH]aRg6@
LH]aRg6:LI1aRg6:LHYaRW2:L8Y[05l020L000L7S0001G6:LHYa00::01^@RY2:T8Z@Ri2:T8Z@RY2:
T8Z@RY2:T8Z?RXl018X00g6:001L00P700071hX00W40:XYaRW6:LH]aT76@LI2:T76@RY1aT8Z@LI1a
T76@RX]aT76:LHYaRW6:Je/020L000L7R003L@0URW6;LH^:T8Z@Ri2;T8^@Ri2;T8^@Ri2:T8^@RY2:
T8Z@RY2:T006RP03L@0005P020L000L7QP00:G5[LG6:LH]aRg6@RY1aT8Z@RY1aT8Z@RY2:T76@RY1a
T8Z@LI2:T76@00::00QaRW6:LHY`RUP020L000L7Q@02L@0>RW6;RY2:T8^@Ri2;T8/3T005Ri2@T8/0
0i005X^@Ri2;T8^@Ri2:T8Z@RY2:T8Z@RXl2RP03LHX005H020L000L7Q000;G5/RW6;LI1aT76@RY2:
T8^@Ri2;T8^@Ri2;T76@LY1aT8Z@LI2:T76@RY2:T003RP07Sg6:LHY`RP1F00P700071h800g402X]a
RhZ@RY2;T8/?T00JRi2;T8^@Ri2;T8^@Ri2:T8Z@RY2:T8Z@RXl2RP03LHX005@020L000L7P000?G5/
LG6:LI1aT8Z@RY2;T8^@Ri2;T8^@Ri1bT7:@LY1bT76@LI1aT76@LI1aT76@RY2:T76?RXmaRW6:L8X0
E0081`001`Mo00Aa00J;RY2:T8/GT00LRi2;T8^@Ri2;T8^@Ri2:T8^@RY2:T8Z@RXn:S`::D`081`00
1`Mn009a03Z:LH]aT76@RY2;T8^@Ri2;T8^@Ri1bT7:@LY1bT7:@LY1bT76@LI1aT76@LI1aT76@RY2:
T8Z?RXma0hX00g6:001A00P700071gd00W401XYaT8Z@R`b@00>AT9002Y0087:@LY1bT8^@Ri1bT8^@
Ri2;T8Z@Ri2:T8Z@RY2:ShZ?0XYA00P700071g`00W403Y1aRg6@LI2;T8^@Ri2;0i00;8^@T91bT7:@
LY1bT7:@LY1bT76@LY1aT76;LI1aT76@LI1aT76@RY2:T8Z?18X00g2:001?00P700071g/00W401XYa
T8^@R`R@00FAT96@T@03T003TI2@00J@02IbT7:@LY1bT7:@LY1bT7:@Ri2;T8^@RY2;T8Z@RY2:T8Z?
RXn:LDl020L000L7NP04L@0:Rg6@LI2:T8^@R`N@02j;T92@LY1bT7:@LY1bT75gLGMaLW6@LH]aRg6@
LH]aT76;LI1aT8Z@RY2:T8Z?0XX00g6:L01?00P700071gT00W401X]aT8^@R`J@00^AT96@TI2AT96@
T@04T00]Mi1gT7:@LY1bMg9gLY1bT7:@LY1bT7:@Ri2;T8^@RY2:T8Z@RY2:T8Z?RXn:04h020L000L7
N004L@0:Rg6@LI2;T8^@R`^@02igT7:@LY1bT7:@LGMaMg5bLG9aLW5bLG9aRg6@LH]aT76@LI1aT8Z@
RY2:T8Z?0XX00g2:001=00P700071gL00W401X]aT8^@R`B@00fAT96@ZY2AT96@TI2A00B@00YgT7N@
LY1bMg5g0W801g5bLW9aLW400W80791bT7:@LY2;T8^@Ri2:T8^@RY2:T8Z@RXn:ShY<00P700071gH0
17402X]aT76@Ri2;T8/;T00bMi1bT7:@LWMaMg5bLG9>LDib679aLW5bLG9aLW5bLH]aRg6@LI1aT76@
RY2:T8Z@RXl2RP03L8X004/020L000L7M@02L@06Rg6@Ri2;1Y001i6@TI2AT94029001WN@Mi1aM`9a
2@02L@04LW5bL@9b01^;LX]bT7:@Ri2;T8^@RY2:T8Z@RY2:T8Z?RXl0B`081`001`Me0009K75aRg6@
LI2;00>@00>;T9002I002gN@LY1aT75gLG9>00d001hHLF]bLG9aLW5bLH]aRg6@LI2:T8Z@RY2:T8Z@
RXl2RP03L8X004T020L000L7L`02L@06Rg6@Ri2;1Y002i6@ZY2AT96@TI2A00B@00AgT75g50007g5b
LG9aLW6;LY1bT8^@Ri2;T8Z@RY2:T8Z@RXn:ShX0BP081`001`Mc000=K75aRg6@LI2;T8^@R`0;T006
Mi1bT76@0W4F000L679aLW5bLG9aRg6;LI1aT8Z@RY2:T8Z?RXn:S`::00=`0000A`081`001`Ma009a
00J;LI2;T8/4T00;TI2AT:Z@TI2AT9401Y000g9gL@0J000KLW5bLH]aRg:@LY2;T8^@Ri2:T8Z@RY2:
T8Z?00>:B0081`001`Ma000=K75aRg6@LI2;T8^@R`0;T007Mi1bT75gCP0K000N679/LW5bLH]aRg6;
LI1aT8Z@RY2:T8Z@RXn:Sg6:B0081`001`M`0007LH]aT8^@R`06T007TI2ZT96@T@08T09a7`006g9a
LW:;LI1bT8^@Ri2:T8^@RY2:T8ZYRXn:S`02RTL020L000L7L002L@08Rg6@LI2;T8/=T004LY1aT09a
80007QQbLG9aLW6;LI1aT76@LI2:T8Z@RY2:T8Z?RXm`LDH020L000L7K`02L@04Rg6@R`J@00NZT96@
TI2A00J@00=gT74090000g9aLP02LP0GLI1bT8^@Ri2;T8^@RY2:T8ZYRXn:ShX0AP081`001`M_0009
LG9aT76@Ri2;00f@00AgT76@9P006QQbLG9aLW6;LI1aT8Z@RY2:T8Z@RY2:T8Z?0XY600P700071fd0
0W4018]aT8/6T007TI2AT96@T@06T003Mi0002L000=aLW400W805h]aT8^@Ri2;T8^@Ri2:T8Z@RZV:
ShZ?04D020L000L7K@002FaaLH]aT8^@R`0=T005Mi1aT4d0:@0071QbK79aRg6;LI1aT8Z@RY2:T8Z@
RXn:ShZ?L8Y400P700071f`00g400i1bT005T007ZY2AT:Z@ZP06T003Mi0002X000AbLG9a0W805X]b
T8^@Ri2;T8^YRY2:ZHZ@RY2:ShY400P700071f`00W401X]aT7:@R`n@00=aT740;0006G5/LW5bLH]a
T76@LI2:T8Z@RY2:T8Z?RXl00XX00f/0001100P700071f/00W4018]aT8/2T00;TI2AT96@TI2AT940
1Y000g40000[0003LG9a009b01F;LY2;T8^@Ri2;T8^@RZV:ZG6@RXl00XY300P700071fX000QaK8]a
T76@LPn@00=aT740<002L@0ILW6;LH]aT76@RY2:T8Z@RY2:T8Z?RXmaRP1200P700071fX000EaLW6@
R`04T007TI2ZT96@T@08T003L@0002l001YaLW6;LX]bT8^@Ri2;T8^@RZV?ZHZ@RXn:S`::@@081`00
1`MY0007K75aRg6@LP0?T003MY1a038001aa679aRg6;LI1aT8Z@RY2:T8Z@RY2:T8Z?RXm`@@081`00
1`MY009a00B@Ri2;0Y001i6@TI2AT:l029000g40000b000>LW5bLH]bT8^@Ri2;T8/2T00;ZHZYSjV:
T8Z?RXl0@@081`001`MX000:LFb;LI1aT8^@R`^@00=fT740=P006G9/LW6;LHmaT76@RY2:T8Z@RY2:
T8Z?RXl00XY000P700071fP000MaRg:@Ri2;00:@00VAT96@ZY2AT9401902LCH000=aLW800W802Y2;
T8^@Ri2;ZH/2T009ZHnYSjV:ShZ?00::?`081`001`MV000:LAQbLI1aT8^@R`^@00=fT740>000775/
LW5bLI1aT76@RY2:T8Z@RY2:T8Z?RXn:Sg0o00P700071fH00g400i1bT003T007TI2AT:Z@T@06T003
MY0003P000abLG9bRg:@Ri2;T8/2T00=ZH^YT:V?ZHnYRXn:S`0o00P700071fD00g401X]aT7:@R`f@
00=aT000>P006G9aLW6;LI1aT8Z@RY2:T8Z@RY2:T8Z?RXl00XXn00P700071fD00W400h]bT003T009
TI2AT96@TI2_00J@00=a0000>@003G5bLH]bT8^@Ri2;T8/00Y001jV@ZI2YSjT00Xl018Z?RXXm00P7
00071f@000QaK8]aT7:@LPf@00=fT740?000775/LW6;LI1aT76@Ri2:T8Z@RY2:T8Z@RXn:Sg0m00P7
00071f@00g400i2;T003T007TI2AT96@ZP06T003MP0003`000YbLG9bT8^@Ri2;0Y003jV@ZI2YSjV@
ZHnYRZR:S`0m00P700071f@00W401X]aT7:@R`f@00=aS`00?P006G9/LW6;LI1aT8Z@Ri2:T8Z@RY1a
T8Z@RW800XXl00P700071f<00W42R`0;T8^@T96@TI2AT:l02910000;LG9bRg:@Ri2;T8/00Y003ZV@
ZI2YT:V?ZHnYRXn:?0081`001`MS0007K79aT76@R`0=T003LI00040001aa679aRg6@LI2:T8Z@RY2:
T8Z@RY2:T8Z?RXm`>`081`001`MR0005LG9aT7801Y000i6@ZP08T003L@00040000abLH]bT7:@Ri2;
T8/2T00=ZI2YT:V?ZHnYSjR:S`0k00P700071f800W401X]aT7:@R`^@00=aT740A0006W9aRg6;LY1a
T8^@RY2:T8ZYRY2:T8Z?RXn9>`081`001`MR0005LG9aT8/00Y001i6@TI2ZT9401Y000gH000130009
LG9bT7:@Ri2;00:@00nYT:V@ZI2YT:V?ZHnYRXl0>`081`001`MR009a00J@LI1bT8/;T003LI0004@0
01UbLG9aRg6@LI2;T8Z@RY2:ZHZ@RY2:T8Z?00::>P081`001`MQ009a0X/2T009TI2AT96@TI2_00J@
AP000g5bLP02LP04T7:@R`B@00^YT:V@ZI2YSjV?Z@02S`03RXl003P020L000L7H000275/LW6@LY1b
2i000gJ@L@17000K679aRg6@LY2;T8^@RY2;T8ZYRY2:ZHZ?RXm`03T020L000L7H0001G5bLI2;00:@
00NAT96@TI2Z00J@00>?0000AP02L@9b00B@LY2;0Y004JV@ZI2YT:V@ZI2YSjV?ZHZ?03T020L000L7
H002L@06Rg6@LY2;2i000g6@0018000JRW5bLH]aT8^@Ri2;T8^YT:V?ZHZ@RY2:ShXi00P700071el0
0W42LPB@00NAT96@ZY2A00J@00=a0000A`002G5bLY1bT8^@R`02T00=ZI2YT:V@ZI2YSjV?Z@02S`03
RP0003H020L000L7G`001fabLI1aT7803I000g5f0018000<LW5bLH]bT7:@Ri2;0i003H^YT:V?ZHZ@
RY2:ShT0=`081`001`MN0005LG9aT8/00Y002I6@TI2AT96@[`04T0:?B`04LP04T7:@R`B@00nYT:V@
ZI2YT:V?ZHnYRXl0=`081`001`MN009a00J;LI1bT78;T003LI0004`000abLG9aT7:@Ri2;T8/4T00:
ZHnYSjV:T8Z?RCL020L000L7G@02L@04Rg:@R`:@00NAT96@TI2Z00J@00=a0000B`001g5bLY1bT8/0
0Y001ZV@ZI2YT0>Y00J@ZI2YSjT2S`03RP0003@020L000L7G@001aQaLH]aT7802i0017J@LGI=000<
LW5bLH]bT8^@Ri2;0Y003JV@ZI2YSjV?ZHZ?RXl0=P081`001`MM0004LW6@LPB@00NAT96@[i2A00J@
CP002G5bLY1bT8^@R`02T006ZI2YT:V@0jT0292YT:V?ZHZ?=@081`001`ML009a00IbLI1bT78;T003
Si1a04l000]aLW6@LY1bT8^@R`04T00<ZI2YSjV?ZHn@RXn:=@081`001`ML0005LH]aT8/00Y002I6@
TI2AT96@T@04T003L@0004l000EaLW:@LP04T004ZI2YT0NY00B?ZHnY0Xl00hX0000b00P700071e`0
0W401X]bT7:@R`^@00=aT000D000279aLW6@LY2;19002jV@ZI2YT:V?ZHn@00:?00>:S`00<`081`00
1`MK009a00B;LY2;0Y001i6@TI2AT:l01Y000g40001@009b00B@LY2;0Y001:Z@ZI09Z@05SjV?ShX0
=0081`001`MK0007675aRg:@LP0;T003MY1a05<000U/LW:;LY2;T8/00Y001:V@ZI03Z@04SjV?Z@:?
00>:S`00<`081`001`MK0004LW6@LPB@00FAT96@[`06T003S`0005800W80191bT8/4T003ZY2Z00>Y
00>^ZJT00ZT0192YSjT2Sc<020L000L7F`001fabLI1bT7802i000g6@001C0009679aRg:@LY2;00:@
00BYT:V@1JT028nYSjV:ShZ?<P081`001`MK0004LW6@LPB@00FAT96@T@06T003MP0005<000EaRg:@
R`04T003ZY2Z00>Y00>^ZJh01:T01HnYShn:038020L000L7FP02L@06LW6@LY1b0Y000i6@T004T004
Sg9aMUD000QbLG9bT7:@R`B@00>YT:T01ZT01XnYSi2:Sc8020L000L7FP001G5bLI2;00:@00NAT96@
TI2A00B@00>?T740EP02LP0?T7:@T96@ZY2ZZJZY[ZW600JY00>?ZHl0<P081`001`MJ009a00AbLI1b
0Y000i6@T006T003LI1a05L000M/LW:@LY2;00B@00>YT:T01ZT01hnYSjV:Sg00<@081`001`MJ0005
LG9bT8/00Y001i6@TI2AT:l019000gJ?001F009b00B@LY2;0Y004JZ@ZY2ZZJnY[ZV^ZJj@ZHnY00:?
<@081`001`MI0009675aRg:@LY1b00V@00=fSg40F0001W9aT7:@LPB@00>YT:T01ZT028nYSjV?T8Z?
<0081`001`MI009a00f@LY2@TI2AT96@TI2A00B@00=a0000E`00576@LY2;TI2AT:Z@[jV_ZLJYaJV^
1:T2S`03RP0002h020L000L7F@001aQbLI1bT8/00Y000i6@T006T09aFP001g5bLY1bT8/019000jV@
Z@06Z@06Z:V?ZHZ?<0081`001`MI0004LW6@LP:@00NAT96@TI2A00J@00=a0000F002LP0CT7:@T96@
ZY2ZZJnY[jW6ZLFY[P02Z@03SjV?030020L000L7F@001g5bLI1bT7802I000hn@L@1K0007K79bT7:@
R`06T003ZI2Y00JY00F?ZHn?L00_00P700071eP000EaLW6@LP02T007TI2AT96@[`06T5/00W805i1b
T92AT96@ZY2_ZJnY[jW5ZJjY[XnY00:?;`081`001`MH009a0W800i1bT00:T003Si1a05`000IbLI1b
T784T004TI2YT0NY00J?ZHn@RXl^00P700071eP000EaLW6@R`02T007TI2AT96@T@04T003LHl005/0
00=aT7800Y004I6@TI2ZT:n@[jV_ZJjY[ZV^00:Y0Xl00hX0000/00P700071eP00W40179aT782T003
TI2@00J@00=aSg40G@001g5bLY1bT8/019001jV@ZI2^ZJh00ZT01XnYSi2:Sbh020L000L7E`02L@04
Rg:@R`:@00NAT96@TI2F00B@00=a0000G002LP0CT7:AT96@ZY2_T:nY[jV_ZJjY[P02Z@03Si2?02h0
20L000L7E`001aQbLI1bT7802I0018n@LHl_0006L@1a075[:@001dibLY1bT7800Y002i6@T92^T:j@
[ZV^00:Y00B?ZHn?;P081`001`MG0004LW6@LP:@00NAT96@TI2A00J@00=a00008P0<L@09RW6:LHYa
RW6:00Aa8@02LP0GT7:AT96@TI2_T:n@[jV_ZJnY[ZV^Si000Xl]00P700071b8000aa075[LF]aJg5[
LF/Y0007679aT7:@LP0;T003L@0001X002]aJg5[LF^:L8Y`RW2:L8YaRW2:L8YaRW6:LHY`RW6:L8Y`
RV^:L75[LF]a01P000M/LW6@LY1b00B@012AT92@[Y2^ZJj@ZHnYSi1a;@081`001`LO00Aa00^:LHYa
RW6:LHYaRP02LBL000AbLI1b0Y001i6@TI2AT9401Y0H00Ea00J:LHYaRW46RP03ShZ:00>:00V?RXZ:
ShZ?RXl00hX02hmaRW6:LHYaRW6:00=a5`004W:@LY6@TI2AT:Z@[i2_T:n^[`:^00BY[ZVY0Xl]00P7
00071ah000=aJg400W40372:LHYaRW6:LHYaRP9a00=[L@008`001FabLI1b00^@00>?T7405000=W5[
LF]aJhY`RW2:LHY`RW6:LHYaRW6:LHYaRW6:LHmaRW6:LHYaSg6:LHY`RW2:L8Y`LG1aJaH000IbLGMb
T784T00ATI2@T:j@[Y2^T:j?ZHn@RXl0;0081`001`LJ00=a01V:LHYaRW6:LHmaShZ@LI2:Sg6?RXma
Sg6:009a7`00179aT782T007TI2AT96@T@04T0:?3@02L@04RW6:L@R:036?RXn:ShZ?RY2:ShZ@RXn:
T8Z@RXn:T8Z?RXn:ShZ?RXn:ShZ?RXmaShZ?LHYaRW6:009a4`02LP:@00fAT96@ZY6_T:n@[iF_00:^
00BY[Y2Y0Xl00hX0000Z00P700071aX001maJg5[RW6:LHYaRW6?LHYaSg6?LHmaSg6:LHYaRW6:01h0
00MaLW6@LY1b00V@00=aT7402P004W5[LF^:LHYaRW6:LHYaRW6:L@::032?RXn:ShZ?RXn:ShZ?RY2:
ShZ?RXn:ShZ?RXmaSg6?LHmaSg6?LHY`RW6:L8Y`LG0C0005LG9bT7800Y004i6@TI2AT:j@[Y2^T:j?
ZHn@RXl0;0081`001`LG00Aa02::LHYaSg6?RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z?LHYa6P003G9a
T7:AT96@TI2AT94019000g6?000300=a00B:LHYa18X0@Hn:ShZ?RY2:T8Z@RY2:T8n@Si2?T8nYSi2?
T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2:ShZ?RXn:Sg6?LHYaRW6:01400W80691gTI2AT96@[i6_T:nE
[jj_ZJj@[Y1bSb`020L000L75P00:G5[LFb:LHYaRW6:LHYaSg6?LI1aT76@LI1aT76@LI1aT76?LHYa
RW1a01P000E/LW:@LP0;T007MXma06]aJ`02L@1;RW6:LHYaRW6:LHn:ShZ?RXn:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z?LHmaSg6?LHmaSg6:LHY`LG1a010000M>LW:@LY1b00:@
00^AT96@[i2^T:j@[P02T004Si2?Sb`020L000L75@04L@0JRW6:LHmaShZ@RY2:T8Z@RY2:T8n@RY2?
T8l3T00=Si2:T8Z@RXmaRW6:L@0D009b00^@LY2@TI2AT96@T@02T0=a00>:LHX00hX098n:ShZ@RY2:
T8Z@RY2:T8n@SjV?T8nYSi2?ZHn@Si2?T8nYS`>@02B?ZHn@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8Z?
RXmaSg6?LHl2L@l001AaLY1bTI2AT96@[i2_T:nE[iF_[P:@00B^T:V?;0081`001`LD000]LF]aLHYa
RW6:LHYaSg6?LHmaT76@LI1aT76@LI1aT76@RY1aT8Z@LI1aSg6:009a00=[00004@00275bLI1bT7NA
1900FV]aK8YaRW6:LHYaRW6?RXn:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8n@RY2?T8Z@Si2:T8n@Si2?T8n@
Si2?T8Z@RY2:T8Z@LI1aT76@LHmaSg6?LHmaSg6?L8Y>L@h001YaLW6@LY1gTI2AT96@UY2^T:j@[Y2^
Si2?T2`020L000L74`04L@0FRW6:LI2:ShZ@RY2:T8Z@RY1aT8Z@RP>@00>?T9002i0028Z@RY2:Sg6:
0W4>0003LG9a00B@00VAT8YaRW6:LHl00hX0:i2:T8Z@RY2:T8Z@SjV?ZI2YSjV@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:T0290068n@T92?T8n@Si2?T8n@Si2?T8n@Si2?T0:?00IaSg6?LHl2L@h000=aMhl0
0Y004I6@TI2_TJnA[i6_T:n@[Y2^00:@0Xl[00P700071a8003MaJg5/RW6:LHYaSg6:LHmaSg6?LHma
Sg6?LI1aT76@LI1aT76@RY2:T8Z@LI2:T76@LHYaRW1a00d00W42LP04T79aK09a05n:LHYaRW6:LHma
ShZ@RY2:T8Z@RY2:T8Z@RY2?T8n@RY2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T8n@RY2?T76@LI1a
T76@LI1aT76@LHmaT76?LHmaSg6?L75`L@0=0006LG9bLY1b0Y001i6@TI2_T9H00i001Jj@T8n@00:?
00=`0000:@081`001`LA00Aa01Z:LHYaShZ?RY2:T8Z@RY2:T76@LI1aT76@LAF@00J:T8Z@LHX2L@X0
00MaLW6:LHYa00B:00Z@RY2:T8Z@RY2:19007jV@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:T0
49001Hn@T92?00F@01J?T8n@Si2?T8n@Si1aT7J?LHmaMg6?0W4<0009LG9bT7NAT96@00:A00^_TJn@
[i2_T:j@[P02T0:?:`081`001`L@000XL@1aK8YaRW6:LHYaSg6?LHmaSg6?LHmaSg6@LI1aT76@LI1a
T76@R`>@012:T92@RY2:T76@LI1aT76:0W400f/0000600Aa00B:LHYa0XX07Xn:T8Z@RY2:T8Z@RY2:
T8Z@Si2?T8n@Si2?T8n@S`>@03n?T92@Si2?T8n@Si2?T8n@Si2?T8n@Si2?T7J@Si1fT8n@LI1aT76@
LI1aT76@LI1aT76@LHmaSg6?LHm`LG003002L@9b01J@LY1gTI2AT96@UY2FT9F@[Xn@Si1a:`081`00
1`L@00=a01j:LHYaT8Z?RY2:T8Z@RY2:T76@LI1aT76@LI1aT8XGT008RY2:T8Z?LHX3L@04RW6:L@::
00Z@RY2:T8Z@RY2?0Y008jV@ZI2YT:V@ZI2YT:V@ZI2YT:V@[Y2YT:V@[Y2YT:j@ZI2^01R@022?T92@
MY1fT7J@MY1fT7J@MY1fT7J@LI1fT76@LHmaRPd00W801Y1bTI2AT0:A00^_TJnA[i2_T:j@[P02T0:?
00=a0000:0081`001`L@000ELFb:K8YaRW6:LHmaSg6?LHmaSg6?00=a019fLG5aMg5gLI1aT76@LI1g
T787T00?Ri2@T8Z@RY2:T76:LHY[009a00J:LHYaRW44RP0BT8Z@RY2:T8Z@RY2?T8n@Si2?0i001Hn@
T92?00f@00F?T92@S`03T00bSi2@T8n@Si1fT8n@MY1aT7J@LI1aT76@LI1aT76@LI1aT76@LI1aT76?
LI1aSg6?LHl2L@03L00000/0019aLW:@LY1gTI2AT96@TI2AT9D2T006Si2?T76::P081`001`L=00Aa
00B:LHYa0XX03hn:T8Z@RY2:Sg6@LHmaMP05L@0:Mg5gLI1aT7:@MaF@0g403XYaRW6;RY2:T8Z@RY2:
0Y008JV@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2^T:V@[Y2YT:j@[P03T003[Y2@01:@02]gT7N@MY1gT7J@
MY1fT7J@MY1aT7J@LI1fT76@MY1aT76@LI1aSg6?LHma00`00W805GMbT92AT96@UY6_T:n@[i2^T:j@
U@02S`03L@0002P020L000L73P005G5[LFb:LHYaRW6?LHmaSg6?LHmaS`04L@05CW5>LDh00W403W9a
Mg6@LI1aT76@LY1b3I005f]aLHYaRW6@LH^:T8Z@RY2:T8Z@RY2?00>@00>?T9003Y000jV@T00BT00f
Si2@T7J@MY1aT76@LI1aT76@LI1aT75gLGMaMg5gLGMaMg5gLGMaMg5gLI1aT76@LGIaSg5f0W4<0007
CW9bMg:@M`02T007TI2AT96@UP03T007UI2@Si1aS`0Z00P700071`d0174018YaRW42RP0;T8Z@RY2:
T76@LHl00g4:009a00YbLGMaT7N@Mi1g1Y00596@TI2ALHYaRW6@RY2:T8Z@RY2:0Y008jV@ZI2YT:V@
ZI2YT:V@[Y2YT:j@[Y2^T:j@[Y2^T:j@ZI2^00F@00FERY2@U@0<T00>Mi1gT7J@MWMaMg5gLGL=009a
01AgLGMaMg5gLGMaT76@LI1aSg6?L@`01782T00?TI2AT9J@[i6_T:n@[Y2^00:@00>?T740:P081`00
1`L<0004LF]aJ`9a00n:LHYaSg6?LHmaSg6?LHl00W400dd0000<000ZLF]bLGMaMg6@LI1bT7N@Mi1[
LFb:LHYaRg6@RY2:T8Z@RY2:T8Z@RY2?39000jV@T00HT00BMi1fT76@MY1aT76@LI1aMg5g0W400dia
000D00Aa00egLGMaMg5fLGMaMW5f009a30002G5bLGMbT7N@M`02T003TI2A00>@00VET92?T8n@LHl0
:P081`001`L;00Aa00B:LHYa0XX02i2:T8Z@RY2:T76?009a50005g5bLGMbT7:@Mi1aRW6:LH^:T8Z@
RY2:00B@02>YT:V@ZI2YT:V@ZI2^T:V@[Y2^T:j@[i2_T:n@[Y2_T:j@[P03T005UI2@T9D029001WN@
Mi1gT09g00=aMg408P02L@0=LW5gLGMaT76@LI1aMP02L@/001YbLGMbT7NAT96@UY2FT:n@[i2_T9F@
UHn@LBX020L000L730004g5/LG6:LHYaRW6:LHmaSg6?LHl00W4H0004CW9aM`9a01=/RW6:LH]aRhZ@
RY2:T8Z@RY2:012@00>YT9005Y0047N@MY1fT76@LI1aT75gLG8V000>675>LW5bLGMaMg5gLGH2L@03
L00000T000THLW5gLY1bT7L00Y001I6@UY2A00>@00NET92?T7Ia02X020L000L72P03L@0DRW6:LHYa
T8Z@RY2:T8Z@LI1aSg4I000=LHYaRW6;RY2:T8Z@RP04T00WZI2YT:V@ZI2YT:V@[Y2^T:n@[Y2_T:n@
[i2_T:n@[i2^T:n@UI2F00>@00>ET9001Y001gN@Mi1aMg40;@003W5bLG9aMg5gLI1aT76@0W4;000E
LW5bLY1gTI2AT96@UY2FT:n@[i2^00:@00>?T000:P081`001`L:000FLF]aK8YaRW6:LHYaSg6?LHma
Sg6:LAP001AaJg5aRW6;LH]aT8Z@RY2:T8Z@RQ2@00>YT9005Y002WN@Mi1aT76@LGL2L@03CP0002h0
00ea675aLW5gLGMaMW5f009a00=`00002@000fabL@02LP04T7:@M`f@00>?T740:`081`001`L900Aa
016:LHYaRW6@RXn:T8Z@LI1aS`02LAH000EaRW6:L@02RP06T8Z@RY2:19009jV@ZI2YT:V@ZI2YT:j@
[Y2_T:j@[i2_T:n@[i2_T:n@UI2_T9F@UP0:T007Mi1gT75gL@0c000?LG9aLW5gLGMaT76@LI1a00`0
01UbLG9bT7NAT96@TI2FT9J@[i2ET9F@LXn@02/020L000L72P004g5[LG6:LHYaRW6?LHYaSg6?LHl0
0W4C000FLF]aLHYaRW6;LI2:T8Z@RY2:T8Z@R`B@00FYT:V@Z@05T005[Y2@T:h03Y000hZ@T003T00:
Mi1gT76@LI1aLST00W402g9>LW5gLGMaMg5f009a30000aQbL@02LP06T7:@Mi5g0Y000i6@T006T003
MY1a02/020L000L720000faaL@02L@04RW6:L@::00b?RY2:T8Z@RXmaSg4A00=a00j:LHYaT8Z@RY2:
T8^@R`:@026YT:V@ZI2ZT:V@ZY2YT:n@[i2_T:n@[i2_T:n@[i2_T:l00i000iJ@T006T004Mi1gT09g
00=a0000>002L@0>LW5gLGMaT76@LI1aMW4<0009LW5bLY1bTI2A00>@00VFT9J@UW:FT9D00Y02Sb/0
20L000L72003L@0CK75aRW6:LHYaRW6?LHmaSg6?L@0@000DLFb:K8YaRg6@RY2:T8Z@RY2;T8/8T003
ZI2@00F@00>^T:h00i000iF@T00<T00;Mi1gT7N@LI1aMdh0?`003QQaLG9aMg5gLGMaMg5f0W4<000=
K79aMg:@LY1gT7N@M`07T004Si1fSb`020L000L71`04L@0BRW6:LHYaRW6?RXn:T8Z?LHma3P02L@0>
RW6:LH^:T8Z@Ri2;T8/2T00UZI2YT:V@ZI2YT:Z@[i2_T:n@[i2_T:n@[i2_T:n@[i2FT9J@UP06T00;
Mi1gT7N@LHYaRW40?P02L@0=LW5gLI1bT76@LI1aS`0=000ELW5bLY1gTGNAT96@UY2FT9J@UI2E00:@
00>?0000:P081`001`L80004LF]aJ`9a00f:LHYaRW6:LHmaSg6:009a2`00575[LFb:LH]aRg6@LI2:
T8Z@Ri2;29000jV@T009T003UI2@00j@01IgT7N@Mi1aMg6@LI1aT76@LHYaRW1a>P003g5bLGMaMg6@
LI1aT75fL@0=000=679aMg:@LY1gT7N@M`05T006Mi2?T7J?;0081`001`L600Ea01::LHYaRW6:LHYa
ShZ?LHmaSg4900=a00Z:LHYaRhZ@Ri2;29008jV@ZI2ZT:Z@[i2_T:n@[i2_T:n@[i2_T:n@[i2FT9J@
TI2F00B@00IgT7N@Mg42T00?Ri2:T8Z@RY2:T76:LHYa03P000mbLG9bT7:@LY1gT76@LGH030000g5b
L@02LP0?Mg:AMi6@TI2AT9J@UY2E00B@00=f0000:P081`001`L5000GL75[LF]aJg5[RW6:LHYaRW6?
LHYaRW001P00175[LF`2L@0@RW6;LI1aT76@LI2;T8^@R`j@00FAT92@UP0@T00PMi1gT76@LGMaT76@
LI1aT76@LI1aT76@LH]aRW6:LHX2L@03J`0003000W403W9aT76@LI1aT76@LGI`30003W4HMg5gLWMb
T7N@Mi1g0i0027N@T91fT76?;0081`001`L500Aa01::LHZ:T76:LHYaShZ?LHmaSg46009a00j:LHYa
Rg6@RY2;T8^@R`B@02>YT:V@ZY2ZT:n@[i2_T:n@[i2_T:n@[i2_T:n@UY2FT9J@UP04T004Mi1gT09g
2Y0058^@Ri2;T8Z@RY2:T8Z@LHYaRW6:<000476;LI1bT7:@Mi1gT7J@LGH;009a17803I1gT7NAT96@
UY2FT9H00i0019F?T7H/00P700071`H001=aJg5[LF]aL8YaRW2:LHYaRW6?009a0`005W5/LFb:LHYa
Rg6@LI1aT8^@Ri2;T8/2T003ZI2@00N@00VAT96@UY2FT9H039001gN@Mi1bT7400WL7T00LSi1aLG:@
LI1aT76@LI1aT76@LI1aRg6;LHYaRP9a00=[L@00:P003G9aRg6@LI1bT76@LI000W4;009a01UbCW9a
Mg:@LY1gT7N@Mi1gT7N@Mi1fT76?02`020L000L71@06L@0@RW6:LHYaRW6:LHmaSg6?00Aa00b:LH]a
Rg6@RY2;T8/8T00OZY2ZT:Z@ZY2_T:n@[i2_T:n@[i2FT:n@UY2FT96@UP04T005Mi1gT7L00Y001I6@
TI2A00J@00=aT8/01I005X^@T92;T8^@RY2:T8Z@RX]aRW6:LHX2LBL00W40191bT787T003LI1a00X0
0W404G9aLW9gLY1bTGNAMi6@TI2F00>@00JET92?T74/00P700071`@001=a075[LF]aJg5[LF]aL8Y`
RW6:009a00=/LF`00W4048YaRg6;LI1aT76@Ri2;T8/6T003ZI2@00>@00VAT96@UY2FT9H039002GN@
Mi1aT75gL@02LP03T7:@00R@02B?T75aLY1bT76@LI1aT76@LI1aT76@LI1aT76;LHYaRW6:LHX2L@03
Jg5[024000maLW6@LI1bT7:@LI1aT7402P00775/LFabLG9aM`1gLY1bT7N@Mi1gT7N@MY1fT74]00P7
00071`@02G402HYaRW6:LHYaRP05L@0<RW6;LI1aT8^@Ri2;1Y008JZ@ZY2ZT:Z@[i2ZT:n@[i6_TJnA
[i2FT9J@UY2FT9J@T@02T005Mi1gT7L00P002g9aT7NAT96@TI2A00J@00=aT9000i001H^@T92;00>@
01^;T8^@Ri2;T8Z@RY2:T8Z;LI1aRW6:LHYaRW407P00179aT789T003MY0000L00g400g9aLP03LP0?
Mg:AMi1gTI2AMi6@TI2E00B@0WH]00P700071`@002QaJg5[LF]aJg5[LF]aJg5[LF]aK8YaRW6;LH]a
T76@LY1bT8^@Ri2;2Y002i6@TI2FT9J@TI2A00R@00QgT7N@LWMaM`H000MaLW6@LY1g00V@0g40;Y1a
T7:@LY1bT76@LY1aT76@LI1aT76@LI1aT76;LH]aRW6:LHYaRW6:K75[LF/F009a00J;LI1bT783T005
LI1aT7401@002f]aK79/LFabLG9a009b01=gLY1bT7N@Mi1gT7N@Mi1fT7Ia02h020L000L70`000g5/
L@06L@03K75a00Ia00j:LHYaRg6;LI2;T8^@R`J@01nAT:Z@ZY2ZT:Z@ZY2ZT:n@[i6FT9J@UY2FT9J@
UY2A00B@00AgT7Mg1`02LP0;T7:@T96@TI2AT94019002G6@LY1bT7N@R`03T00TRi2@T8^@Ri2;T8^@
Ri2;T8Z@RY2:T76;RX]aRW6:LHYaRW6:174B0003LI1a00B@00>AT9001902L@<01W40179aLW42LP04
Mg9gLP9g00FAMi5gT@02T003MiIg00>@00=gT7H0;P081`001`L2000BL@1a075[L@1a0740L@1aK75/
0W4048]aRg6@LI1aT7:@LY1bT8/:T005TI2@T9400i000i6@T@06T00;Mi1gT7N@LY1bMg402`001aQb
LY1bT7L01i00@Xn@LHmaMg6@LI1bT76@LY1bT7:@LI1aT76@LI1aT76@LI1aT76@LI1aRg6;LHYaRW6:
LHYaRW6:K75/LF]aJg5[L@L00W40191aT8/3T00EMi2?LF]aJg5[LFaaK75/LVabLG9a009b01AgLWMb
Mg:@LY1gT7N@Mi1gT7J@LBl020L000L70P000faaL@0>L@0:Rg6;LI1bT8^@R`>@02:;T92AT96@TI2A
T96@ZY2ZT:n@TI2FT9J@UY2FT9J@TI2A190017N@MgL=0004LW6@LP:@00FAT96@T@06T00?LGMbT7:@
LY1bT7:@LY1b00>@02b;T8^@Ri2;T8^@Ri2;T8^@Ri2:T8Z@LI2:T76;LH]aRg6:LHYaRW6:LHYaRP=a
00>:LG402g401HYaLG6:00=a00Z:LG5aLW5bLG9a178057MbMg:@Mi1gTGN@Mi1gT7N@Mi1f<0081`00
1`L20003L@1a009a00L0LFb:K75/00Aa016;LH]aRg6@LI1bT7:@Ri1bT@07T00?TI2@T96@TI2AT96@
TI2A00:@00]gT7N@Mi1gT75gL@0A0007CW9aT7:@LP09T01;LHl0LTibLG9aMg6@LI1bT76@LY1bT7:@
LY1aT76@LI1aT76@LI1aT76@LI1aRg6;LH]aRg6:LHYaRW6:LHYaRVb:LHY/RW6:JhY/009a00J:K75a
RV`3L@0=K75aLW5bLG9/LW5bL@04LP0CMg9gLY1bT7N@Mi1gT7N@MY1aM`0`00P700071`03075a00aa
00j;LH]aRg:@Ri1bT8^@R`:@01nAT96@TI2AT96@TI2ZT96@UY2AT9J@UY2FT96@TI2A00:@00EgT7N@
L@0B0004LW6@LP:@00NAT96@TI2A00B@05ma0000LG9aMg6:LY1bT7:@LY1bT7:@Ri1bT8^@Ri2;T8^@
Ri2;T8^@Ri2:T8Z@LI2:T76@LH]aRg6;LH]aRW6:LHYaRW6:LHYaRW6:LHYaRW6;LH]aRg6;LH]aLW5b
L@02LP0HT79gLY1bT7:@Mi1gT7N@Mi1gT7N@MY1a<0081`001`L2000:L@1a674HLFaaK0Aa01>;LH]a
T76@LI1bT7:@LY2;T8^A00>@00fAT92@TI2AT96@TI2A00B@00agT7N@Mi1gT76@LGLD0007679aT7:@
LP09T003LHm`00@004aaCW9aLW5gLGMaT76@LI1aT7:@LY1bT76@LI1aT76@LI1aT76@LI1aT76;LH]a
Rg6;LH]aRg6:LHYaRW6:LHYaRW6:LHYaRW6:LHY/0W403X]aLG5bLG9aLW5bLG9a17804WMbMg:@LY1g
T7:@Mi1gT75gLC4020L000L700<0LG401W403W9aLW6;LI1bT7:@Ri1b19001Y6@TI2AT0FA01V@TI2A
T96@UY2AT96@TI2AT96@TGN@Mi1b01P000AbLI1b0Y001I6@TI2A00J@00=aS`001P00Dg9aLW5gLY1b
T7:@LY1bT7:@LY2;T7:@Ri2;T7:@Ri2;T8^@LY2;T8^@Ri1aT8Z@LI2:T76@LH]aRg6;LH]aRg6;LH]a
Rg6;LH]aRg6;LH]bRg:;00=b01R@LWMbT79gLY1bT7N@Mi1gT7N@Mi1gT7Hb00P700071`8001Ya674H
LFaaK79aLW6;LH]aT76@LY1bT7:@LPB@00FAT96@T@03T005TI2AT940190037N@Mi1gT7N@LY1aMaX0
00THLW6@LY1bT7L01I001Hn@MXm`00T005E>LTibLG9aMg5gLI1aT7:@LI1bT76@LY1aT7:@LI1aT76@
LI1aT76@LI1aRg6@LH]aRg6;LH]aRg6;LH]aRg6;LH]aRg6;LH]aLW6;LG9aLW5bLG9a00Ab00YgLWMb
Mg9gLY1b0WL01i1fMg5gLGH0<P081`001`L00`1aL@04L@0>LW5bLH]aT7:@LY1bLG84T006TI2AT96@
0i404Y2ATI6@TI2AT96@TI2AT95gT@:@00EgT79gL@0L009a00^@LY2@TI2AT96@T@04T0:?3@00Bg9a
LW5gLGMbT7:@LY1bT7:@LY1bT7:@LY2;T7:@Ri1bT8^@LY2;T7:@LY1bT7:@LY1aT7:@LH]aT76;LX]a
T7:;LH]bRg:@LX]bR`03LP0;T79gLWMbMg:@LY000gL02i1gT7N@Mi1fT75f03<020L000L700H0674H
LF`2L@0OLW5bLH]aRg6@LI1bT7:@LY1bT7N@Mi6@TGNAT95gT@02T00@Mi1gT7N@Mi1gT7:@LWMaMal0
00IbLI1bT789T003LHma00l004U>LTibLG9aMg5gLI1aT7:@LI1bT76@LY1aT7:@LI1aT76@LI1aT76@
LI1aRg6;LI1aRg6;LH]aRg6;LH]aRg6;LH]aRg5bLG9a00=b00=aMg800W804gMbMg9gLWMbMg9gLWMa
MgIgLGL0=0081`001`L01`1aLG9aLW400W802Y1bT7:@LY1bT784T00GTI2AT96@TI2AT96@TI2AT96@
TI2AT9400Y001GN@MgMa02<000YaLW:@Mi6@TI2A1Y000hn@000C001?LG9aMg9gLWMbT7:@LY1bT7:@
LY1bT7:@LY1bT7:@LY2;T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:;LY1bT7:@LY1bT7:@LY1bMg:@
LP04M`05T7N@Mi0017L00gH0000d00P700071`80035a675/LW5bLG9aLW6@LI1bT7:@LY1bT7:@LY5g
TG:AMi1gT7N@Mi1gT7N@LY1bMg5g02D00W42LP06T7:@Mi1g1I000gJ@L@0F0010LTibLG9aMg5gLGMa
T76@LI1bT76@LI1bT76@LI1aT7:@LI1aT76@LH]aT76;LI1aRg6@LH]aRg5bLH]aLW5bL@=b00=aMg80
0W804WMbMg9gLWMbMg9gLWMaMg5gLCL020L000L70P004G5bLG9aT76@LY1bT7:@LY1g00:@01NAT96@
TI2AT96@TI2AT96@TGN@Mi5gT002M`03LGL002P00W802I1gT92AT96@T@04T003Si1a01T004aaLW5b
LGMbMg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT79g
LY1bMg:@LWMbMg:@1WL00gIg000h00P700071`<002/HLW5bLG9aLW6@LY1bT7:@LY1bT7:@LY1gT7:@
Mi1gT7N@LY1bT79gLGMa02/000EaLW:@LP09T004LI1aRQ/0031aCW9>LW5bLGMaMg5gLWMaT75gLY1a
T7:@LI1bT76@LI1aT7:@LI1aT76@LY1aT742LP05T75gLY000g8067MaMg9gLWMbMg9gLWMbMg5gLWMa
Mg5gCST020L000L70`00179aLW42LP0RT7:@LY1bT7N@Mi6@TGNAT96@TGNAT95gT7N@Mi1gT79gLBh0
0W802I1bT92AT96@T@06T003L@0001h00W40@79aMg5gLWMbMg9gLY1bT7:@LY1bT7:@LY1bT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT79gLY1bMg9gLWMbMg83M`06LWMbMg5f>`081`001`L3000V679>LW5b
LGMbMg:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT79gLG8`0007CW9aT7:@LP09T003LHm`028004EaCW9a
LW5bLG9aMg5gLGMaMg:@LGMbT7:@LWMaT79gLY1bMg6@LWMbT75gLWMbMg9gLGMbMg9gLWMbMg9gLGMb
Mg5gLGL00W4l00P700071`@000=aLW400W8077MbT7:@LY1bT7N@Mi1gT7NAMi1gT7N@Mi1bMg4d009a
00UgLY1gTI2AT9401Y000g6?000V009a03igLGMbMg5gLWMbMg9gLY1bMg:@LWMbT79gLY1bT79gLY1b
Mg9gLWMbT79gLWMbMg9gLWMbMg9gLWMaMg5gLCh020L000L71000179>LW42LP0IMg5gLWMbT7:@LY1b
T7:@LY1bMg:@LWMaM`0g0006LW6@LY1g2I000hn@L@0Y000m675>LTibLG9aMg5gLGMaMg5gLGMaMg5g
LWMaMg9gLWMaMg9gLGMbMg5gLWMbMg9gLGMaMg5gLGM[LDigCP0o00P700071`D00W42LP0CMg9aLY1b
T7:@LY1gT7N@Mi1bT002M`03LGL003T000aaMg:@Mi6@TI2AT944T0:?<003L@0^LW5gLWMbMg9gLWMb
Mg9gLWMbMg9gLWMbMg9gLWMbMg9gLWMbMg9gLWMbMg5gLD@020L000L71`005TibCW9aLW5gLGMbMg9g
LWMaMg5gLGLm009a0W80191bT7L7T003LI1a03@002YaCW5>LTibLG9>Mg5bLGMaLW5gLGMaMg5gLGMa
Mg5gLGMaMdigCWM>LTi700P700071`T000mbLGMbMg9gLWMbMg9gLWL0@`02LP07T7:@T96@T@06T003
Si1a03X017401g9aLW5bLG800g403GMaLW5gLGMaMg5gLGL00W5=00P700071`X000]a075aLW5gLG9a
M`160007LG9bT7:@M`07T004Si1aLCX001=`LG6@LI2:T8Z@CW5>LTiaCW9>05L020L000L7F`00179a
T7L2T005TI2AT9401Y000g40000h009a00>@LI001Y0018n@ShmK00P700071e/000LHLW6@LY1b00N@
00F?T76?L00i000?K75aT76@RY2:T8Z@LHma05/020L000L7F`00179aT782T005TI2AT9401Y000g6?
000h009a00>@RY001Y0018n@ShmK00P700071e`000QbLI1bT7:@M`N@00AaSg5a=`00475/Rg6@LI2:
T8Z@RY1aSg1K00P700071e`000EaMg:@M`02T005TI2@T9401902S`03L@0003D000EaRg6@R`07T004
Si1aLE/020L000L7G002L@9b00J@LY1gT7L5T004Si1aScL00W402h]aT76@RY2:T76@009aG0081`00
1`ML0005LG9bT7800Y000i6@T@06T003Si1a03H00W4018]aT8/7T003Si1a05`020L000L7G@001dib
LY1bT7801i001Hn@LI1`03D000hHLG6;LI1aT8Z@RY1aT09aG0081`001`MM0004LW6@M`:@00>AT940
2902S`03L@0003800W400i1aT006T004Si2?T5d020L000L7GP00279aT7:@LY1g1i00176@LHXc000@
LFabLI1aT8Z@RY2:T76@LEd020L000L7GP000g5gLP04T003TI2@00J@00>?T740<`001G5bLI2;00V@
00>:S`00G0081`001`MO0007LG9bT7:@M`07T005Si1aSg00<P02L@0>Rg6@LI2:T8Z@LI1aSg5M00P7
00071el00W800gMbT003T003TI2@00J@00>:S`00<@001G6;LI2;00N@00B?T75aG@081`001`MO000;
679aT7:@LY1gT7L00i001Xn@LI1aRS400W403H]aT76@RY2:T8Z@LHl0GP081`001`MP0005LGMbT7L0
0Y000i6@T@06T003Si1a03000W400i1aT008T003Si1a05h020L000L7H@001digLI1bT7L01i00176@
LHla009a00^;LI2:T8Z@RY1aT002LEh020L000L7H@02LP05T7N@T9402Y000hZ?000]0007675aT76@
R`07T003LHl005h020L000L7HP00279aT7:@Mi1g1i00176@LG4]009a00j;LI1aT8Z@RY2:T76?L5l0
20L000L7HP001G5gLY1g00:@00>AT9001Y02S`03L@0002/00g400i2:T006T004Si1aLEl020L000L7
HP002W9>Mg:@LY1gT7L5T005Si1aSg00;002L@0=Rg6@LI2:T8Z@RY1aS`1P00P700071f<00W800i1g
T003T003TI2@00J@0XX/0005LH]aT8/01i000hn@L@1P00P700071f<000THLW6@LY1gT7L01I0018n@
LI02LB/00W403I1aT8Z@RY2:T76@LHl0H0081`001`MT000:LGMgT7NAT96@T@J@00>?T740:P02L@04
T76@R`N@0Xl00g40001N00P700071f@000YbLGMbT7:@Mi1g1I001G6@LHm`02P000maK8]aT76@RY2:
T8n@LI000W5P00P700071f@000EaLW:@M`02T003TI2@00R@0Xl00g40000U00=a00>@Ri002902Sf40
20L000L7I@001digLI1bT7801i001Xn@LI1aSbL00W403X]aT8Z@RY2?T8Z@LI1aH@081`001`MV0003
LGMb00B@00>AT9002902SbH000EaRg6@R`09T003L@00060020L000L7IP00279aMg:@LY1g0i002WN@
T91aT76@LHXU009a00f;LI1aT8Z@RY1aT76?068020L000L7I`00175bMg84T003TI2@00J@00>?T740
8`001FaaLH^:00V@00>?T740HP081`001`MX0008LW6@LY1gT7L7T005LI1aSg008P003g5/RW6@LI2:
T8Z@Si1aT002LF8020L000L7J0001g5bLY1gT7L00Y000i6@T006T0:?00=aRP007`001G6:LI2;00V@
00>?T000HP081`001`MY0009CW9aT7:@Mi1g00N@00EaT76?L@0P009a00j@LI2:T8Z@Si1aT76@LF<0
20L000L7JP001G5gLY1g00f@00B:T76:7@02L@04T8Z@R`V@00=a0000HP081`001`MZ000:LTigLY1b
T7N@M`F@00N?T76@LHma01`00W403h]aT76@RY2?T8Z@RY1aS`1T00P700071f/00W40191bT7L=T004
Si1aRQT00W4018^:T8X9T003Si1a06@020L000L7K0002W9aMg:@LY1gT7L3T008Si2?T76@LHl2LAP0
015[LG6;LI1aT8Z@RY2?T76@L@1U00P700071f`000EaLW6@LP0?T004Si1aRQH000EaRW6@RP0;T003
MP0006@020L000L7K@002dibLI1bT7:@Mi1g00F@00J?T8Z@LHl2L@03J`0001800W403hYaT76@RY2:
T8n@LI1aS`1V00P700071fl00W800i1gT00>T006Si2:Sg6:40001G6:LI2:00f@00=a0000I0081`00
1`M_000;LG9aT7:@LY1gT7L01I0028n@RY1aSg6:0W4>000CJg5aRW6@LI2:T8Z@Si2?T76@L@1W00P7
00071g400W800i1gT00>T006Si2?T76?0g4600Aa00B:MXZ:3I000g5f001V00P700071g4000]>LW6@
LY1bT7N@M`05T00:Si2:T76@LHmaS`9a00=`LF/00g405V]aLHYaRW6@LI2:T8Z@Si2?T8n@LI1X00P7
00071g8000MaMg:@Mi1g00f@01F?T8Z@RXmaSg6:LHYaRW6?LI2:T8X03I000g40001V00P700071g<0
00e>LW6@LY1bT7N@Mi1g00>@02F?T8n@LI1aT76@LHmaSg6:LHmaT76@LI2:T8Z@Si2?T7J@LI1a06T0
20L000L7M0001G5gLY1g01>@00f:T8Z@RY2:T8Z@RY2?00f@00=a0000J0081`001`Me000;CW9aT7:@
LY1gT7L01I008Hn@Si2?T8Z@LI2:T76@RY1aT8Z@Si2:T8n@Si2?T76@L@1[00P700071gH000MaMg:@
Mi1g01F@00>?T9003i02MV/020L000L7M`003DibLGMbT7:@Mi1gT7L01i003hn@Si2?T8n@RY2?T8n@
S`03T007Si1fT76@L@1]00P700071gP000MaMg:@Mi1g02>@00=fL@00K0081`001`Mj000<LW5gLI1b
T7:@Mi1g0i000gN@T005T003Si2@00>@00>?T9000i001WN@LI1aMVh020L000L7N`001WMaMg:@Maf@
00AgT75gK`081`001`Mn000BLW5gLI1bT7:@LY1gT7N@Mi1g1I002gN@Mi1fT7J@LGMa074020L000L7
O`02LP06Mg:@Mi1g5I000gJ@L@1b00P700071h8001]bLGMaT7:@LY1gT76@Mi1gT7N@LI1aT76@LGL0
M0081`001`N20003LG9a009g00R@Mi1gT7N@M`>@00YgT92@Mi1gT76@M@081`001`N6000CLDibLGMa
Mg5gLI1aT75gLGMaM`1h00P700071hT00W402WMaMg9gLI1bMg5l00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P700071ol04P081`001`Oo018020L000L7o`0B00P700071ol04P081`001`Oo0180
20L000L7o`0B00P7003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b470000\
\>"],
  ImageRangeCache->{{{0, 287}, {285.875, 0}} -> {-10.8285, -10.0021, 1.46919, 
  1.46919}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell["Here we have a smaller box and the tube size is reduced.", "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\ \(knot3 = 
      ParametricPlot3D[{4\ Cos[t\  + \ N[Pi]]\ /\ 3\  + \ 2\ Cos[3\ t], \n
          \ \ \ \ \ \ \ 4\ Sin[t]\ /\ 3\  + \ 2\ Sin[3\ t], \n\ \ \ \ \ \ \ 
          Sin[4\ t]\  + \ Sin[2\ t]\ /\ 2}, \ \n
        \ \ \ \ \ \ \ \ \ \ \ \ \ \ {t, \ 0, \ 2\ Pi}]; \)\)\)], "Input",
  CellLabel->"In[14]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .86373 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-1.11022e-16 1.05047 -0.0427704 1.05047 [
[.12147 .21063 -11.958 -9 ]
[.12147 .21063 .04197 0 ]
[.33802 .12428 -5.54843 -9 ]
[.33802 .12428 .45157 0 ]
[.57395 .03029 -5.11784 -9 ]
[.57395 .03029 .88216 0 ]
[.74593 .0708 0 -6.21085 ]
[.74593 .0708 12 2.78915 ]
[.85032 .23239 0 -6.05427 ]
[.85032 .23239 6 2.94573 ]
[.94016 .37149 0 -5.92395 ]
[.94016 .37149 6 3.07605 ]
[.02785 .2994 -12 -2.74888 ]
[.02785 .2994 0 6.25112 ]
[.01208 .39843 -6 -2.83339 ]
[.01208 .39843 0 6.16661 ]
[-0.00474 .50407 -6 -2.92414 ]
[-0.00474 .50407 0 6.07586 ]
[ 0 0 0 0 ]
[ 1 .86373 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.04662 .25333 m
.68486 0 L
s
.13073 .21995 m
.13535 .22461 L
s
[(-2)] .12147 .21063 .993 1 Mshowa
.34652 .13429 m
.35077 .1393 L
s
[(0)] .33802 .12428 .84948 1 Mshowa
.58153 .04101 m
.58531 .04638 L
s
[(2)] .57395 .03029 .70595 1 Mshowa
.125 Mabswid
.18301 .1992 m
.18573 .20204 L
s
.23637 .17801 m
.23904 .18091 L
s
.29087 .15638 m
.29348 .15933 L
s
.40338 .11173 m
.40586 .11478 L
s
.46147 .08867 m
.46389 .09178 L
s
.52084 .0651 m
.52318 .06827 L
s
.07951 .24028 m
.08233 .24302 L
s
.64358 .01638 m
.64578 .01966 L
s
.25 Mabswid
.68486 0 m
.96158 .42795 L
s
.73366 .07546 m
.72752 .0778 L
s
[(-2)] .74593 .0708 -1 .38019 Mshowa
.8379 .23668 m
.8317 .23882 L
s
[(0)] .85032 .23239 -1 .34539 Mshowa
.92764 .37545 m
.92138 .37743 L
s
[(2)] .94016 .37149 -1 .31643 Mshowa
.125 Mabswid
.76127 .11816 m
.75757 .11953 L
s
.78781 .1592 m
.7841 .16054 L
s
.81333 .19868 m
.80962 .19999 L
s
.86157 .27328 m
.85784 .27454 L
s
.88439 .30857 m
.88065 .3098 L
s
.9064 .3426 m
.90265 .34381 L
s
.70491 .031 m
.70124 .03243 L
s
.94816 .40718 m
.94439 .40834 L
s
.25 Mabswid
.04662 .25333 m
0 .54809 L
s
.04008 .29464 m
.0462 .29226 L
s
[(-1)] .02785 .2994 1 -0.38914 Mshowa
.02439 .39387 m
.03055 .39159 L
s
[(0)] .01208 .39843 1 -0.37036 Mshowa
.00765 .49973 m
.01384 .49756 L
s
[(1)] -0.00474 .50407 1 -0.35019 Mshowa
.125 Mabswid
.03702 .31398 m
.0407 .31257 L
s
.03393 .33357 m
.03761 .33217 L
s
.03079 .35342 m
.03447 .35202 L
s
.02761 .37351 m
.0313 .37213 L
s
.02113 .41449 m
.02483 .41313 L
s
.01783 .43538 m
.02153 .43404 L
s
.01448 .45655 m
.01819 .45522 L
s
.01109 .478 m
.0148 .47668 L
s
.04311 .27553 m
.04677 .27409 L
s
.04609 .25667 m
.04975 .25521 L
s
.00416 .52176 m
.00789 .52047 L
s
.00063 .54409 m
.00436 .54281 L
s
.25 Mabswid
.04662 .25333 m
0 .54809 L
s
0 .54809 m
.3991 .86373 L
s
.3991 .86373 m
.40834 .59708 L
s
.40834 .59708 m
.04662 .25333 L
s
.68486 0 m
.96158 .42795 L
s
.96158 .42795 m
1 .70975 L
s
1 .70975 m
.70262 .30855 L
s
.70262 .30855 m
.68486 0 L
s
.04662 .25333 m
0 .54809 L
s
0 .54809 m
.70262 .30855 L
s
.70262 .30855 m
.68486 0 L
s
.68486 0 m
.04662 .25333 L
s
.40834 .59708 m
.96158 .42795 L
s
.96158 .42795 m
1 .70975 L
s
1 .70975 m
.3991 .86373 L
s
.3991 .86373 m
.40834 .59708 L
s
0 0 m
1 0 L
1 .86373 L
0 .86373 L
closepath
clip
newpath
.5 Mabswid
.38993 .70551 m
.35584 .65896 L
s
.43029 .74649 m
.38993 .70551 L
s
.71662 .48238 m
.65965 .48418 L
s
.77114 .48346 m
.71662 .48238 L
s
.35584 .65896 m
.3304 .61057 L
s
.47428 .77802 m
.43029 .74649 L
s
.82085 .48721 m
.77114 .48346 L
s
.65965 .48418 m
.60248 .48851 L
s
.3304 .61057 m
.3157 .56379 L
s
.51899 .79641 m
.47428 .77802 L
s
.86318 .49293 m
.82085 .48721 L
s
.60248 .48851 m
.54739 .49452 L
s
.3157 .56379 m
.31359 .52163 L
s
.56123 .79849 m
.51899 .79641 L
s
.89533 .49948 m
.86318 .49293 L
s
.31359 .52163 m
.32556 .48656 L
s
.5977 .78206 m
.56123 .79849 L
s
.54739 .49452 m
.4968 .50081 L
s
.91437 .50539 m
.89533 .49948 L
s
.62515 .74634 m
.5977 .78206 L
s
.32556 .48656 m
.35262 .4604 L
s
.4968 .50081 m
.45336 .50548 L
s
.64071 .69233 m
.62515 .74634 L
s
.91758 .50902 m
.91437 .50539 L
s
.35262 .4604 m
.39497 .44416 L
s
.45861 .58247 m
.38597 .60522 L
s
.52975 .54641 m
.45861 .58247 L
s
.64216 .62297 m
.64071 .69233 L
s
.59619 .5006 m
.52975 .54641 L
s
.38597 .60522 m
.31533 .61192 L
s
.65514 .44906 m
.59619 .5006 L
s
.31533 .61192 m
.25038 .60106 L
s
.45336 .50548 m
.41994 .50627 L
s
.70413 .396 m
.65514 .44906 L
s
.62829 .54307 m
.64216 .62297 L
s
.25038 .60106 m
.19479 .57264 L
s
.39497 .44416 m
.45171 .43794 L
s
.90285 .50885 m
.91758 .50902 L
s
.19479 .57264 m
.15191 .52843 L
s
.74106 .34556 m
.70413 .396 L
s
.59906 .45883 m
.62829 .54307 L
s
.15191 .52843 m
.12445 .47198 L
s
.12445 .47198 m
.11421 .40835 L
s
.5557 .37721 m
.59906 .45883 L
s
.45171 .43794 m
.52044 .44078 L
s
.86938 .50391 m
.90285 .50885 L
s
.76413 .30166 m
.74106 .34556 L
s
.11421 .40835 m
.12183 .34367 L
s
.50066 .3051 m
.5557 .37721 L
s
.41994 .50627 m
.39942 .5008 L
s
.12183 .34367 m
.14676 .28454 L
s
.43735 .24856 m
.50066 .3051 L
s
.14676 .28454 m
.18724 .23727 L
s
.36988 .21208 m
.43735 .24856 L
s
.18724 .23727 m
.24045 .20721 L
s
.52044 .44078 m
.59719 .45072 L
s
.30274 .19819 m
.36988 .21208 L
s
.24045 .20721 m
.30274 .19819 L
s
.81814 .49408 m
.86938 .50391 L
s
.77199 .26778 m
.76413 .30166 L
s
.59719 .45072 m
.67652 .46496 L
s
.7522 .48039 m
.81814 .49408 L
s
.67652 .46496 m
.7522 .48039 L
s
.39942 .5008 m
.39435 .48699 L
s
.76384 .24674 m
.77199 .26778 L
s
.73975 .24041 m
.76384 .24674 L
s
.39435 .48699 m
.40629 .46365 L
s
.70103 .24934 m
.73975 .24041 L
s
.40629 .46365 m
.43521 .43104 L
s
.65052 .27243 m
.70103 .24934 L
s
.43521 .43104 m
.479 .39127 L
s
.59272 .30684 m
.65052 .27243 L
s
.479 .39127 m
.53344 .34819 L
s
.53344 .34819 m
.59272 .30684 L
s
.25 Mabswid
.68486 0 m
.96158 .42795 L
s
.96158 .42795 m
1 .70975 L
s
1 .70975 m
.70262 .30855 L
s
.70262 .30855 m
.68486 0 L
s
.04662 .25333 m
0 .54809 L
s
0 .54809 m
.70262 .30855 L
s
.70262 .30855 m
.68486 0 L
s
.68486 0 m
.04662 .25333 L
s
.04662 .25333 m
.68486 0 L
s
.13073 .21995 m
.13535 .22461 L
s
[(-2)] .12147 .21063 .993 1 Mshowa
.34652 .13429 m
.35077 .1393 L
s
[(0)] .33802 .12428 .84948 1 Mshowa
.58153 .04101 m
.58531 .04638 L
s
[(2)] .57395 .03029 .70595 1 Mshowa
.125 Mabswid
.18301 .1992 m
.18573 .20204 L
s
.23637 .17801 m
.23904 .18091 L
s
.29087 .15638 m
.29348 .15933 L
s
.40338 .11173 m
.40586 .11478 L
s
.46147 .08867 m
.46389 .09178 L
s
.52084 .0651 m
.52318 .06827 L
s
.07951 .24028 m
.08233 .24302 L
s
.64358 .01638 m
.64578 .01966 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[14]:=",
  ImageSize->{288, 248.75},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000n0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00<D70P1I1`00
``L200031`0005P700301`<00`L205P7002S1`@05`L200H700<01`00E`L00:<700<01`L05@L300P7
00@01`L0EPL00:@700<01`L04PL300X700@01`L0EPL00:D700<01`L03PL3008700<01`L020L01@07
1`L005D7002V1`0300L700/70P0@1`0300L7008700<01`L0DPL00:<700@01`L02PL3018700<01`L0
0PL00`071`1B1`00Y0L200T70P0F1`0400L71`80D`L00:`70`0H1`0300L7008700<01`L0D@L00:X7
0P0K1`0300L700<700<01`L0D0L00:L70`0M1`0300L700@700<01`L0C`L00:D70P031`806`L00`07
1`041`0300L704l7002R1`<08PL00`071`051`0300L704h7002P1`809@L00`071`061`0300L704d7
002M1`<09`L00`071`061`0300L704d7002K1`80:PL00`071`071`0300L700X7100n1`00V0L302`7
00<01`L020L00`071`091`0300L703l7002F1`8000<700L0;0L00`071`081`0300L700X700<01`L0
?PL009<70`0a1`0300L700T700<01`L00PL400@700<01`L0?@L009470P0d1`0300L700T70P0<1`03
00L703`7002>1`<0=PL00`071`081`0300L000T700@01`L0?PL008`70P0i1`0300L700/700<01`L0
1`L203l700291`<0>`L00`071`0<1`0300L704L700271`80?PL00`071`0<1`0300L704L700241`<0
0PL00`071`0l1`0300L700`700<01`L0APL008870P141`0300L700d700<01`L0A@L007l70`161`03
00L700d700<01`L0A@L007d70P191`0300L700h700<01`L0A0L006870P0F1`<0B`L00`071`0>1`03
00L704@7001Q1`0400L701<70P1>1`0300L700l700<01`L0@`L0064700@01`L040L404l700<01`L0
40L00`071`121`00H@L010071`0>1`80D`L00`071`0@1`80@`L0064700@01`L02`L305D700<01`L0
40L204<7001Q1`0400L700P70`1H1`0300L7018700<01`L0@0L006870P071`80F`L00`071`0B1`03
00L70407001X1`<0G@L00`071`0C1`0300L703l7001V1`<0G`L00`071`0D1`0300L703h7001S1`<0
0`L00`071`1L1`0300L701@700<01`L0?PL006470P1U1`0300L701D700<01`L0?@L005h70`1W1`03
00L701H700<01`L0?0L005`70P1Z1`0300L701H700<01`L0?0L005T70`1/1`0300L701L700<01`L0
>`L005L70P1`1`0300L701L700<01`L0>PL005@70`021`0300L706d700<01`L05PL203`7001B1`80
M@L00`071`0H1`0300L703T7001?1`<0M`L00`071`0I1`0300L703P7001=1`80NPL00`071`0I1`03
00L703P7001:1`<0O0L00`071`0J1`0300L703L700181`80O`L00`071`0J1`0300L703L700151`<0
0PL00`071`1l1`0300L701/700<01`L0=PL004<70P241`0300L701`700<01`L0=@L004070`261`03
00L701`700<01`L0=@L002@7100F1`80R@L00`071`0M1`0300L703@7000T1`0300L701@70`2;1`03
00L701h700<01`L0<`L002D700<01`L04@L201`7201Z1`0300L701d70P0e1`007PL400@700<01`L0
3@L3008700<01`L04@L800P71P1T1`0300L701l700<01`L0<PL002L700<01`L02PL201@71@0F1`H0
GPL00`071`0P1`0300L70347000T1`0400L700T70`0D1`808@L405X700<01`L080L00`071`0a1`00
9@L200P70P0E1`809`L205P700<01`L08@L00`071`0`1`00;0L301D70P0[1`80E`L00`071`0Q1`03
00L702l7000Z1`805PL202l70P1E1`0300L7024700<01`L01PL202L7000W1`<00PL201870P0c1`03
00L7058700<01`L08PL00`071`041`0400L702H7000U1`805`L203H70P1B1`0300L702<700<01`L0
0`L010071`0V1`008PL301L70P0j1`80D0L00`071`0S1`0300L700<700@01`L09PL002070P0I1`03
00L703`700<01`L0C@L00`071`0S1`8010L010071`0V1`007@L301X700<01`L0?PL204d700<01`L0
20L301L700@01`L00`L010071`0V1`006`L200031`0701T700<01`L0@@L204/700<01`L010L400<7
0`0G1`0300L700870P0W1`0060L301h700<01`L0@`L204T700<01`L0100:1`805PL00`071`0Z1`00
5PL2020700<01`L0APL00`071`161`<03`L00`071`0D1`0300L702X7000D1`808@L00`071`181`03
00L704<70`0B1`0300L701@700<01`L0:@L001@70P0P1`0300L704X700<01`L0@0L2008700<01`L0
40L00`071`0E1`0300L702P7000D1`0300L001l700<01`L0B`L203d70`041`0300L7014700<01`L0
50L00`071`0X1`0050L010071`0M1`0300L704h700<01`L0>0L200L700<01`L04@L00`071`0E1`03
00L702L7000C1`0300L7008700<01`L06@L00`071`1@1`0300L703D70P091`0300L7018700<01`L0
5@L00`071`0V1`004`L00`071`031`0300L701L700<01`L0DPL00`071`0b1`802`L00`071`0B1`03
00L701D700<01`L09PL001<70P051`0300L701D700<01`L0E0L00`071`0_1`803@L00`071`0B1`03
00L701D70P0W1`004`L00`071`051`0300L701@700<01`L0E@L202d70P0@1`0300L7010700<01`L0
60L00`071`0T1`004`L00`071`061`0300L7018700<01`L0F0L00`071`0Y1`0300L7010700<01`L0
40L00`071`0H1`0300L702@7000C1`0300L700L700<01`L040L00`071`1J1`0300L702H70P0C1`03
00L7010700<01`L06@L00`071`0S1`004PL00`071`091`8040L00`071`1K1`0300L702@700<01`L0
4`L00`071`0@1`0300L701X700<01`L08PL001870`0;1`0300L700`700<01`L0G@L00`071`0Q1`80
5PL00`071`0?1`0300L701/700<01`L08PL0018700<01`L030L00`071`0;1`0300L705h70P0O1`80
60L00`071`0?1`0300L701`700<01`L08@L000X710041`0300L700d700<01`L02@L00`071`1Q1`03
00L701/700<01`L060L00`071`0?1`0300L701d700<01`L080L000`700<01`L00`L00`071`0>1`03
00L700P700<01`L0HPL00`071`0H1`806`L00`071`0>1`0300L701h700<01`L080L000`700<01`L0
0`L00`071`0?1`0300L700H700<01`L0I0L00`071`0F1`0300L701X70P0@1`0300L701l700<01`L0
7`L000@710041`0300L7008700<01`L04@L00`071`051`0300L706@700<01`L050L201X70`021`03
00L700`700<01`L07`L00`07000P1`0030L00`071`021`804`L00`071`041`0300L706D700<01`L0
4PL00`071`0G1`<01@L00`071`0<1`0300L7024700<01`L07PL000X70`041`0300L701<700<01`L0
0PL00`071`1W1`0300L700l70P0G1`<02@L00`071`0:1`0300L702<700<01`L07@L000`700<01`L0
0PL00`071`0D1`0500L71`00JPL00`071`0=1`0300L701@70`0=1`0300L700T700<01`L08`L00`07
1`0M1`004@L00`071`0E1`0300L006/700<01`L02`L201@70`0A1`0300L700L700<01`L09@L00`07
1`0L1`004@L00`071`0F1`80K0L00`071`081`804`L301@700<01`L01`L00`071`0V1`0300L701/7
000@1`8060L206d700<01`L01PL00`071`0A1`8060L00`071`051`0300L702L700<01`L06`L00107
00<01`L05`L00`07001/1`0300L700@70P0A1`<06`L00`071`041`0300L702P700<01`L06PL00107
00<01`L05PL01@071`L006`700<01`L00PL00`071`0>1`<07`L00`071`021`0300L702X700<01`L0
6@L0010700<01`L05PL00`071`021`0300L706X700@01`003PL3028700<01`L00PL00`071`0Y1`80
6`L0010700<01`L05PL00`071`031`0300L706X700<01`L02`L302H700@01`L0;PL00`071`0H1`00
40L00`071`0F1`0300L700@700<01`L0J0L200X70`0Z1`80<0L00`071`0G1`003`L201P700<01`L0
1@L00`071`1V1`0400L700H70`0^1`0300L702h700<01`L05`L000l700<01`L05`L00`071`061`80
I0L200@700<01`L00`0`1`80<@L00`071`0F1`003`L00`071`0F1`0300L700T700<01`L0H0L00`07
1`041`<0<PL010071`0a1`0300L700@7100=1`003`L00`071`0F1`0300L700X700<01`L0GPL00`07
1`021`<000<700L0<@L00`071`021`0300L702h700<01`L010L00`071`0>1`003`L00`071`0F1`03
00L700/700<01`L0G0L00`071`0300D700<01`L0;`L00`071`031`0300L702h700<01`L010L00`07
1`0=1`003`L201L700<01`L030L00`071`1J1`@02@L00`071`0]1`0300L700@700<01`L0;`L00`07
1`041`0300L700`7000?1`0300L701H700<01`L03@L00`071`1G1`<030L00`071`0/1`0300L700H7
00<01`L0;PL00`071`051`0300L700/7000>1`0300L701L700<01`L03PL00`071`1D1`<03`L00`07
1`0[1`0300L700L700<01`L0;0L00`0700031`0400L700d7000>1`0300L701L700<01`L03`L00`07
1`1@1`D04@L00`071`0Y1`0300L700T700<01`L0;PL010071`L200h7000>1`0300L701L700<01`L0
40L00`071`1<1`<00PL00`071`0A1`0300L702P700<01`L02PL00`071`0^1`0300L70147000>1`03
00L701L700<01`L04@L00`071`181`<010L00`071`0C1`0300L702H700<01`L030L00`071`0^1`03
00L70107000>1`805`L00`071`0C1`0300L704@70`061`0300L701@700<01`L09PL00`071`0=1`03
00L702d700<01`L040L000h700<01`L05PL00`071`0D1`0300L704070`081`0300L701H700<01`L0
90L00`071`0?1`0300L702d700<01`L03`L000H70P061`0300L701H700<01`L05@L00`071`0l1`<0
2PL00`071`0G1`0300L702<700<01`L040L00`071`0^1`0300L700h700051`0400L700@700<01`L0
5`L00`071`0F1`0300L703P70`0<1`0300L701T700<01`L08@L00`071`0B1`0300L702d700<01`L0
3PL000D700@01`L010L00`071`0G1`0300L701L700<01`L0=0L300h700<01`L06PL00`071`0P1`03
00L701@700<01`L0;0L200l700051`0400L700@700<01`L05`L00`071`0H1`0300L703070`0?1`80
7PL00`071`0N1`0300L701H700<01`L0;@L00`071`0<1`001@L010071`041`8060L00`071`0I1`03
00L702`70`0A1`0300L701h700<01`L07@L00`071`0G1`0300L702d700<01`L030L000D700@01`L0
10L00`071`0G1`0300L701X700<01`L0:0L301<700<01`L080L00`071`0K1`0300L701T700<01`L0
;@L00`071`0;1`001PL200D700<01`L060L00`071`0J1`0300L702@70`0E1`0300L7028700<01`L0
6@L00`071`0K1`0300L702d700<01`L02PL000`700<01`L06@L00`071`0K1`808@L301L700<01`L0
8`L00`071`0I1`0300L701`700<01`L0;0L00`071`0:1`0030L00`071`0I1`0300L701d700<01`L0
70L201T700<01`L09@L00`071`0G1`0300L701d700<01`L0;0L200/7000<1`806PL00`071`0N1`03
00L701P70`0J1`0300L702H700<01`L05PL00`071`0O1`0300L702P70`000`L01`0:1`0030L00`07
1`0I1`0300L701l700<01`L050L301`700<01`L0:0L00`071`0D1`0300L7024700<01`L08`L400@7
00<01`L02@L000`700<01`L06@L00`071`0P1`0300L701070`0L1al03`L00`071`0C1`0300L702<7
00<01`L07`L300P700<01`L02@L000`700<01`L06PL00`071`0P1`0300L700`70`0F1`T000<700L0
70L700T700<01`L04@L00`071`0T1`0300L701`70`0<1`0300L700P7000;1`0300L701/700<01`L0
8@L00`071`081`<05`L200T700<01`L08`L800031`07014700<01`L09PL00`071`0G1`@03`L00`07
1`081`002`L00`071`0K1`0300L7028700<01`L010L301L70`0:1`0300L702`71P0=1`0300L702P7
00<01`L04`L301<700<01`L020L000/70P0L1`0300L702<700<01`L00`0H1`8030L00`071`0_1`04
00L71`H01PL00`071`0Y1`0300L701070`0F1`0300L700P7000;1`0300L701/700<01`L08`L301T7
0P0=1`0300L7034700<01`L01PL602d700<01`L02`L401T700<01`L020L000/700<01`L06`L00`07
1`0P1`<00PL00`071`0E1`803PL00`071`0b1`0300L700T7200Y1`0300L700L70`0M1`0300L700P7
000;1`0300L701`700<01`L070L300H700<01`L04`L00`071`0=1`0300L703@700<01`L01`L00`07
1`061`D09@L00`071`031`<080L00`071`081`002PL00`071`0M1`0300L701T70`0:1`0300L70147
00<01`L03PL00`071`0d1`0300L700H700<01`L030L5020700<01`L00`0T1`0300L700L7000:1`03
00L701d700<01`L05PL300h700<01`L03`L00`071`0>1`0300L703D700<01`L01@L00`071`0B1`H0
6@L402L700<01`L01`L000X70P0N1`0300L701@70P0B1`0300L700d700<01`L03`L00`071`0f1`03
00L700<700<01`L06@L501470`031`0300L702D700<01`L01`L000X700<01`L07PL00`071`0@1`<0
5@L00`071`0;1`0300L700l700<01`L0=`L00`071`021`0300L701l71@091`<01`L00`071`0T1`03
00L700L7000:1`0300L701h700<01`L03@L301T700<01`L02@L00`071`0@1`0300L703P700@01`L0
4PLN00X700<01`L090L00`071`071`002PL00`071`0O1`0300L700T70`0M1`0300L700L700<01`L0
40L00`071`0i1`<00`L@01L74@000`L01`0T1`0300L700L7000:1`0300L701l700<01`L01PL30247
00<01`L01PL00`071`0@1`0300L703073`0T1`<01`L500D71P0Q1`0300L700L700091`0300L70247
00<01`L00PL302D70P051`0300L7018700<01`L090L;00P700<01`0090L400l71@031`0300L700H0
70L00`071`061`002@L3024700<01`000P0Z1`0300L7008700<01`L04PL00`071`0M1`L04PL01007
1`0Q1`<060L400P7100H1`0300L700H700091`0300L702070`0^1`0500L71`005@L201<72P0H1`05
00L71`007PL301l71@071`D04`L00`071`061`0000<700000P041`0300L701d70`021`0300L702d7
00@01`L05`L200H72`0P1`801@L00`071`0H1`<08`L010071`L700D7100?1`0300L700H700031`03
00L700<700<01`L06PL300D700<01`L0;PL201X71P0Z1`0300L700D700<01`L050L402L700<01`L0
1`L900l700<01`L01PL000<700<01`L00`L00`071`0G1`<02@L00`071`0^1`0300L704H70P091`03
00L701070`0[1`0300L701l700<01`L01PL000<700<01`L00PL00`071`0E1`<030L00`071`0^1`80
APL00`071`091`0300L700d70`0_1`0300L701h700<01`L01PL000<700<01`L00PL00`071`0B1`<0
40L00`071`0/1`0400L704<70P0<1`0300L700T7100c1`0300L701d700<01`L01PL000041`0000@7
0P0A1`804`L00`071`0/1`0500L71`00@@L00`071`0=1`0300L700D70`0h1`0300L701d700<01`L0
1@L000<700<01`L00PL00`071`0=1`<05PL00`071`0[1`0300L7008700<01`L0?@L00`071`0>1`03
00L700870`0k1`0300L701d700<01`L01@L000P700<01`L02PL301T700<01`L0:`L00`071`031`03
00L703X70P0A1`0300L00080?`L00`071`0L1`0300L700D700081`0300L700L70`0M1`0300L702X7
00<01`L010L00`071`0h1`0300L700l710131`0300L701/700<01`L01@L000L700<01`L01@L30247
00<01`L0:@L00`071`051`0300L703D70P0?1`<00`L00`071`121`0300L701X700<01`L01@L000L7
00<01`L00PL302D700<01`L0:0L00`071`061`0300L703<700<01`L030L300L700<01`L0@@L00`07
1`0J1`0300L700D700071`D0:@L00`071`0W1`0300L700L700<01`L0<0L200/7100:1`0300L70487
00<01`L06@L00`071`051`001`L202d700<01`L09PL00`071`081`0300L702d70P0:1`<03PL00`07
1`131`0300L701T700<01`L010L000P700<01`L0;0L00`071`0U1`0300L700T700<01`L0:PL200T7
0`0A1`0300L704@700<01`L060L00`071`041`002@L202d700<01`L090L00`071`0:1`0300L702L7
0P071`@050L00`071`141`0300L701P700<01`L010L000/700<01`L0:`L00`071`0S1`0300L700/7
00<01`L090L200H70`0I1`0300L704@700<01`L05`L00`071`041`0030L00`071`0[1`0300L70287
00<01`L030L202870P051`<070L00`071`151`0300L701H700<01`L010L000d700<01`L0:`L00`07
1`0Q1`0300L700h700<01`L07@L200@70`0O1`0300L704H700<01`L05@L00`071`041`003PL202`7
00<01`L080L00`071`0?1`0300L701X70P021`@08PL00`071`161`0300L701D700<01`L010L00107
00<01`L0:PL00`071`0P1`0300L700l700<01`L05`L200041`0002H700<01`L0A`L00`071`0E1`03
00L700<7000A1`0300L702X70P0P1`0300L7010700<01`L050L502T700<01`L0B0L00`071`0D1`03
00L700<7000B1`0300L702/70P0N1`0300L7014700<01`L04@L402`700<01`L0B0L00`071`0D1`03
00L700<7000C1`80;@L201d700<01`L04@L00`071`0=1`<0<0L00`071`191`0300L701<700<01`L0
0`L001D700<01`L0;0L201/700<01`L04PL00`071`091`@0<`L00`071`191`0300L7018700<01`L0
0`L001H700<01`L0;@L201T700<01`L04`L00`071`041`@0=`L00`071`1:1`0300L7014700<01`L0
0`L001L700<01`L0;PL201L700<01`L050L00`071`0403X700<01`L0BPL00`071`0A1`0300L700<7
000H1`80<0L201H700<01`L04`L403d700<01`L0B`L00`071`0A1`0300L70087000J1`0300L702l7
100B1`0300L701070`000`L01`0n1`0300L704`700<01`L040L00`071`021`006`L00`071`0b1`H0
30L00`071`091`L010L00`071`0m1`0300L704d700<01`L03`L00`071`021`0070L00`071`0g1`H0
1`L;00/700<01`L0?@L00`071`1=1`0300L700l700<01`L00PL001d70P0m1`P05@L00`071`0n1`03
00L704d700<01`L03PL00`071`021`007`L00`071`121`0300L7018700<01`L0?PL00`071`1>1`03
00L700d700<01`L00PL0020700<01`L0@@L00`071`0B1`0300L703h700<01`L0C`L00`071`0<1`03
00L70087000Q1`0300L7044700<01`L04@L00`071`0n1`0300L704l700<01`L03@L00`071`011`00
8PL2044700<01`L04@L00`071`0n1`0300L7050700<01`L030L00`071`011`0090L00`071`0o1`03
00L7010700<01`L0?PL00`071`1A1`0300L700/700<01`L00@L002D700<01`L0?PL00`071`0@1`03
00L703h700<01`L0DPL00`071`0:1`0300L70047000V1`0300L703h700<01`L03`L00`071`0n1`03
00L7058700<01`L02PL00`071`011`009`L203h700<01`L03`L00`071`0n1`0300L705<700<01`L0
2@L00`071`011`00:@L00`071`0l1`0300L700d700<01`L0?`L00`071`1D1`0300L700P700<01`L0
0@L002X700<01`L0>`L00`071`0=1`0300L703l700<01`L0E@L00`071`071`0300L70047000[1`03
00L703/700<01`L030L00`071`0o1`0300L705D700<01`L020L100470@L002`70P0k1`0300L700`7
00<01`L0?`L00`071`1F1`0300L700L70@011`47000^1`0300L703T700<01`L02`L00`071`0o1`03
00L705L700<01`L01PL100470@L002l700<01`L0>@L00`071`0:1`0300L703l700<01`L0F0L00`07
1`051`400@L11`00<0L00`071`0i1`0300L700T700<01`L0?`L00`071`1H1`0300L700D70@011`47
000a1`80>@L00`071`091`0300L703l700<01`L0F@L00`071`041`400@L11`00<`L00`071`0g1`03
00L700P700<01`L0?`L00`071`1J1`0300L700<70@011`47000d1`0300L703L700<01`L01`L00`07
1`0o1`0300L705/700<01`L00`L10047000e1`0300L703L700<01`L01PL00`071`0o1`0300L705/7
00<01`L00`L10047000f1`80=`L00`071`061`0300L703l700<01`L0G0L00`071`021`400@L003P7
00<01`L0=@L00`071`051`0300L703l700<01`L0G@L01@071`L00047000i1`0300L703D700<01`L0
10L00`071`0o1`0300L705h700@01`L00@L003X700<01`L0=@L00`071`031`0300L703l700<01`L0
GPL010071`011`00>`L203D700<01`L00`L00`071`0o1`0300L705l700<01`000@L003d700<01`L0
<`L00`071`021`0300L703h700<01`L0H@L100470@0003h700<01`L0<`L01@071`L0040700<01`L0
HPL20000?`L00`071`0c1`0400L7040700<01`L0HPL20000@0L00`071`0c1`0300L0040700<01`L0
HPL20000@@L203@70P0o1`0300L705l710021`00@`L00`071`0b1`0300L703d700<01`L0F`L400H7
00141`0300L703470P0n1`0300L705L7100:1`00A@L00`071`0`1`80?PL00`071`1C1`@03PL004H7
0P0`1`0300L003`700<01`L0D0L4018700181`0300L702`700D01`L7000k1`0300L704`7100F1`00
B@L00`071`0[1`0300L7008700<01`L0>0L00`071`181`@06PL004X700<01`L0:PL00`071`031`03
00L703L700<01`L0A0L401h7001;1`80:PL00`071`041`0300L703D700<01`L0@@L40287001=1`03
00L702L700<01`L01@L00`071`0d1`0300L703d7100V1`00CPL00`071`0V1`0300L700H700<01`L0
<PL00`071`0j1`@0:PL004l700<01`L09@L00`071`071`80<@L00`071`0g1`@0;PL005070P0U1`03
00L700T700<01`L0;PL00`071`0c1`@0<PL0058700<01`L08PL00`071`0:1`0300L702`700<01`L0
<@L303H7001C1`0300L7024700<01`L02`L202/700<01`L0;PL403T7001D1`0300L7020700<01`L0
3@L00`071`0W1`0300L702/7100m1`00E@L2020700<01`L03PL00`071`0U1`0300L702P710111`00
E`L00`071`0M1`0300L700l70P0U1`0300L702@710151`00F0L00`071`0L1`0300L701470P0R1`03
00L7024710191`00F@L00`071`0K1`0300L701<70P0N1`8080L404d7001J1`806`L00`071`0E1`<0
6@L201h7101A1`00G0L00`071`0H1`0300L701P70`0C1`<070L405D7001M1`0300L701L700<01`L0
6`L200l70P0K1`@0F@L005h700<01`L05PL00`071`0M1`l06@L405d7001O1`805PL00`071`111`@0
H@L0064700<01`L04`L00`071`0m1`@0I@L0068700<01`L04PL00`071`0i1`@0J@L006<700<01`L0
4@L00`071`0e1`@0K@L006@70P0A1`0300L703870`1a1`00IPL00`071`0>1`0300L702h7101d1`00
I`L00`071`0<1`0300L702/7101h1`00J0L00`071`0;1`0300L702L7101l1`00J@L200/700<01`L0
8`L40807001[1`0300L700P700<01`L07`L408@7001/1`0300L700L700<01`L06`L408P7001]1`03
00L700H700<01`L05`L408`7001^1`801PL00`071`0C1`@0T0L0070700<01`L00`L00`071`0?1`@0
U0L0074700<01`L00PL00`071`0;1`@0V0L0078700D01`L700091`@0W0L007<70P000`L01`041`@0
X0L007D70P000`L000020:@7001f1`80Z0L00?l78@L00?l78@L00001\
\>"],
  ImageRangeCache->{{{0, 287}, {247.75, 0}} -> {-0.02494, 0.0292819, 
  0.00340383, 0.00340383}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[BoxData[
    \(MVClear[]; \ MVShow3D[knot3, MVLineTubeSize -> 0.3]; \)], "Input",
  CellLabel->"In[15]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[22]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040303F4F3F3F5F40405F403F503F3F403030000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F30303F30304F303F
4F3F3F5F404060403F704F40704F407F504F7F504F7F504F7F504F805F4F7F5040704F40604040\
604040503F3F4F3F3F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F3030\
3F3030403F3F4F303F503F3F503F3F6040406F4040704F40704F407F504F70504F
7F50507F504F805F50805F50805F5F805F5F8F60608F605F8F606080605F8F605F805F50805F50\
805F4F805F507F5040704F4060403F5F4040503F3F4F303F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F3F3F50303F5F3F3F5F3F3F
603F3F603F3F5F403F503F3F503F3F4F3F304F3F3F403F3F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F3F3F5F3F3F60404060403F\
6F40406F4040704F4F6F4F4F70504F704F4F7F50507F50507F5F5F7F5F5080605F
805F5F8F60608F60608F6F6F8F6060906F6F906F6F907070906F6F906F6F8F6F60906F608F605F\
8F605F8F605F8F605F805F50805F50805F4F805F4F704F406F40405F3F3F5F3F3F
4F3030000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000403F3F4F3030503F3F503030603F3F603030603F3F603F3F6F403F
6F403F6F403F6F403070403F6F40306F403F60403060403F5F3F305F403F503F30503F3F4F3F30\
403F3F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000603F3F6F4F4F6F4F4F704F50704F507F5050\
7050507F5F5F7F5F5F805F60805F5F8060608060608F6F6F8F6F6F906F6F8F6F6F
9070709070709F707F9F70709F7F7F9F7F7FA080809F7F7FA07F7F9F7F7F9F7F7F9F70709F7070\
90706F9F706F906F60906F60906F60906F608F605F90605F8F60508F6050805F4F
805F4F704F406F4F405F3F3F503F3F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F3F3F\
503F3F5F3F3F5F3F3F6F3F3F6F3F3F6F403F6F403F704F40704F3F7F4F40704F3F
7F4F407F4F3F7F4F3F7F4F3F80503F506F60804F3F7F4F3F7F4F3F704F3F704F3F60403060403F\
5F3F3F5F3F3F503F30000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F3F3F604040704F5070505F7F5F5F7F505F805F607F5F60\
80606F8060608F6F6F8F606F8F6F708F6F6F9070709070709F707F90707F9F7F80
9F7F7FA080809F7F80A0808FA0808FAF8F8FA0808FAF8F8FA0808FAF808FA08080A08080A07F7F\
A07F7F9F7F709F7F709F70709F707090706F9F706F906F60906F60906F5F906F5F
8F605F90605F8F60508F6050805F4F7F504F6F404060403F503F30503030000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F303F4F3F3F503030603F3F60303F\
6F3F3F6F3F3F70403F70403F7F4F4070403F7F4F407F4F408050407F4F3F805040
804F3F8F504080503F8F503F8F503F90503F8F50308F503F8F4F3F8F503F7F4F307F4F3F704030\
70403F6F3F306F403F603F30603F3F30304F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000040303F6040406F4F4F70505F7F5F6080606F80606F8F6F708F6F70907070\
8F6F7090707F90707F9F7F809F7F7F9F80809F7F80A0808FA0808FA0808FA0808F
A08F90A0808FAF8F90AF8F90AF8F90AF8F90AF909FAF8F90AF9090AF8F90AF8F90AF8F8FAF8F8F\
AF808FAF8F8FA08080AF8080A07F7FA0807FA07F70A07F709F7070A07F709F706F
9F706F906F609F7060906F5F906F5F90605F906F5F8F60508F60507F504F6F4F405F3F3F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F303F503F3F5F3F3F603F3F6F3F3F703F3F703F3F7F403F\
7F403F7F4F407F4F40805040804F408F504F8050408F5F408F5040905F408F5040
905F40905F3F905F40905F3F9F5F3F905F3F9F6040905F3F905F408F503F8F5F4080503F80503F\
7F4F3F7F4F3F70403F70403F6F403F6F403F603F3F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F30305F3F406F4F4F6F4F507F5F5F7F5F608F6F708F6F7F90707F90707F9F7F80907F80\
9F7F809F7F809F808F9F7F8FA0808F9F808FA08090A0808FA08F90A08F90AF8F9F
A08F90AF8F9FAF8F9FAF909FAF8F9FAF90A0AF909FAF90A0AF8F9FAF909FAF8F9FB0909FAF8F90\
AF9090AF8F90B08F90AF8F8FAF8F8FAF8080AF8F8FA08080AF8080A07F7FA0807F
9F7F70A07F709F7F6FA07F709F706F9F706F9070609F7060906F5F906F5F805F507F504F60403F\
5F3030000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000040303F503F3F5F30306F3F3F70303F7F3F3F703F3F7F403F7F403F804F4080403F\
804F40804F408F50408F5040905F4F8F5F40905F40905F40905F40905F409F6040
9F5F3F9F60409F5F3F9F60409F5F3F9F60409F603F9F604090603F9060408F5F3F8F5F4080503F\
8050407F503F7F5040704F3F704F3F6F403F6F403F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000303030
60404060404F704F5070505F80606F80607090707F9070809F808F9F7F8F9F808F9F808FA08090\
9F8090A08F90A08090A08F9FA08F90AF8F9FA08F9FAF90A0AF8F9FAF90A0AF909F
B090A0AF90A0B09FAFAF90A0B09FAFB09FAFB09FAFAF9FAFB09FAFB090A0B09FA0B090A0B09FA0\
B0909FB0909FB0909FB0909FB09090B0909FB09090B09090B08F8FB0908FAF8F8F
B08F8FAF8080AF8F80A0807FAF807FA08070A08070A07F70A07F709F70609F7060906F5F8F6050\
704F406F4040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000403F3F
503F3F5F3F3F603F3F6F3F3F7F3F3F803F3F8F403F80403F8F4F408F4F408F50408F5040905F4F\
905F409F5F4F905F4F9F604F9F604FA0604F9F6040A06F4FA06040A06F4FA06040
A06F4FA06040A06F4FA06040A06F4FA06F40A06F4F9F6F40A0704F9F6F409F6F40906040906040\
8F60408F6040805F40805F407F503F7F5040704F3F70503F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000050303F60404F
604F4F70505F705F6080606F8F6F70907080907F809F808F9F8090A080909F8090A08F9FA08090\
A08F9FA08F9FA08F9FA08F9FAF90A0A08F9FAF90A0AF90A0AF9FAFAF90A0B09FAF
AF90AFB09FB0AF9FAFB09FB0B09FB0B0A0B0B09FB0B09FB0B09FAFB09FB0B09FAFB09FAFB09FAF\
B09FAFB090A0B09FA0B090A0B09FA0B0909FB0909FB0909FB0909FB09090B09090
B09090B09090B08F8FB0908FAF8F80B08F8FAF8080AF807FA07F70A07F709F706F9F7060906F5F\
906F5F805F4F704F40603F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000403F3F4F3F3F5F3F3F
603F3F6F3F3F703F3F803F3F803F3F8F40408F403F904F40905040905F4F9050409F5F4F9F5F40\
A0604F9F604FA06F4FA0604FAF6F4FA06040AF6F4FA06F40AF6F4FA06F40AF6F4F
A06F4FAF704FA06F4FA0704FA06F4FA0704FA06F40A0704FA07040A0704F9F6F409F704F906F40\
906F4F8F60408F6040805F408060407F5F3F7F5F3F70503F70503F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F3F405F3F406F4F506F4F5F
7F5F607F5F6F8F6F7F8F707F907F8F907F8F9F80909F8090A08F9FA08F9FA090A0A08FA0AF90A0\
A090A0AF90AFAF90A0AF9FAFAF90AFB09FB0AF9FAFB09FB0AF9FB0B0A0B0B09FB0
B0A0BFB09FB0BFA0BFB0A0BFBFA0C0B0A0BFBFAFC0B0A0BFBFA0BFB0A0BFBFA0BFB0A0B0BFA0B0\
B09FB0BFA0B0B09FAFBFA0AFB09FAFBFA0AFB09FA0BF9FA0B09FA0BF9FA0B09F9F
BF9FA0B0909FBF9F9FB09090BF9F9FB09090BF9090B08F8FB08F80AF807FAF807FA07F70A07F6F\
9F70609F7060906050805F4F6F4F40603F3F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000403F3F503F3F503F3F6040406F3F3F
7F3F3F7F3F3F8F40408F4040904F40904F409F504F9F5040A05F4F9F5F4FAF604FA0604FAF6F50\
AF6F4FB07050AF6F4FB07050AF6F4FB07050AF704FB07050AF704FB07050AF704F
AF7050AF704FAF7F50AF704FAF7F50AF704FAF7F50AF7F4FAF7F4FA07F4FAF7F4FA07F4FA07F4F\
9F704F9F704F906F40906F408F60408F6F4080603F80603F7F5F3F604F3F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F303F5F3F405F404F6F4F5070505F7F606F
80606F8F707F907080907F8F907F8F9F80909F809FA08FA0A08FA0A090AFA090A0AF90AFA090AF\
AF9FAFAF90AFB09FB0AF9FB0B09FB0AF9FB0B0A0BFAF9FB0B0A0BFB0A0BFB0A0C0
B0A0BFB0A0C0B0A0C0BFAFCFB0A0C0BFAFCFB0A0C0BFAFCFB0A0C0BFAFC0B0A0BFBFAFC0B0A0BF\
BFA0BFB0A0B0BFA0BFB09FB0BFA0B0B09FB0BFA0B0B09FAFBFA0AFB09FAFBFA0AF
B09FA0BF9FA0B09FA0BF9FA0B0909FBF9F9FBF9090BF9F90B08F8FB08F8FAF8080AF807FA07F70\
A07F709F70609F7060906F5F8F60507F50406F4F405F3030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000403F3F4F3F3F5F403F603F3F6F3F3F703F3F804040
803F3F8F40408F4040904F4F904F409F504F9F504FA05F4FA05F4FAF604FAF6F4FB07050B0704F\
B07050B0704FB07050B0704FB07050B07050B07F50AF7050B07F50AF7050B07F50
AF7050B07F50AF7F50B07F50AF7F50AF7F50AF7F50AF8050AF7F4FAF8050AF804FAF8050A07F4F\
A07F4F9F70409F7F4F9070409070408F6F3F8F6F4080603F80603F7F5F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040303F503F4060404F604F507050607F5F60806070806F7F
907F80907F8F9F80909F8090A08F9FA08F9FA090A0A090AFAF9FB0AF9FB0AF9FB0AF9FB0B0A0BF\
AF9FBFB0A0BFB0A0BFB0A0C0B0A0C0B0AFC0B0A0C0BFAFCFB0AFC0BFAFCFB0A0C0
BFAFCFB0AFC0BFAFCFB0AFC0BFAFCFB0AFCFBFAFCFB0AFCFBFAFCFBFAFCFBFAFCFBFAFCFBFAFCF\
BFAFC0BFAFC0BFA0C0BFAFC0BFA0BFBFAFBFBFA0BFBFAFBFBFA0B0BFA0BFBFA0B0
BFA0B0BFA0B0BFA0B0BF9FAFBFA0AFBF9FA0BFA0A0BF9F9FBF9F9FBF9090BF9090B08F80B08F80\
AF807FAF807FA07F6FA07F6F9F70609F70608F60507F504F6F403F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000403F3F5040405F3F3F6040406F3F3F7F404080403F8F40408F4040
904F4F904F4F9F50509F504FA05F50A05F50AF6050AF6050B06F50B06F50BF705FBF7050C0805F\
BF7F50BF7F5FBF7F50BF805FBF7F5FBF805FB07F5FBF805FB0805FBF805FB0805F
BF805FB0805FB0805FB08050B08F5FB08050B08F5FB08050B08F5FB08F50B08F50AF8F50AF8F50\
A0804FA0804F9F7F4F9F7F4F9070409070408F70409070408F6F3F806F3F7F603F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F2F30503F405F3F4F604F506F4F5F705F607F5F6F806F7F8F707F907F8F
907F8F9F80909F8090A08F9FA08FA0A090AFA090AFAF9FB0AF9FB0B0A0BFAF9FBFB0A0C0B0A0BF\
B0A0C0B0A0C0B0A0C0B0A0C0B0AFC0B0A0C0B0AFC0B0A0C0B0AFC0B0A0C0B0AFC0
B0A0C0B0AFC0AFA0C0B0AFC0AFA0C0B0AFCFAFA0C0B0AFCFB0AFC0B0AFCFB0AFC0BFAFCFB0AFCF\
BFAFCFBFAFCFBFAFCFBFAFCFBFAFCFBFAFC0BFAFC0BFAFC0BFAFC0BFA0BFBFAFBF
BFA0BFBFAFBFBFA0B0BFA0B0BFA0B0BFA0B0BF9FAFBFA0A0BF9FA0BF9F9FBF9F90BF9F90B0908F\
B0908FAF807FAF807FA08070A07F709F70609F7060906F5F8F6050704F406F403F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
403F3F4F3F3F5F4040603F3F704040703F3F8040408F4040904F408F4F4090504F
90504F9F5F509F5F4FA06050A06050AF6F50AF6F50B0705FB07050BF7F5FBF7F50C0805FBF805F\
C0805FBF7F5FBF805FBF805FBF805FBF805FBF805FBF805FBF8F60B0805FBF8F60
B0805FBF8F5FB0805FBF8F5FB08F5FBF8F5FB08F50B08F5FB08F50B0905FAF8F50AF8F50AF8F50\
AF8F50A0804FA0804F9F7F4F9F7F4F9070409070408F70408F7040806F3F806F3F
6F5030000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F303F5F404F6040506F505F7050607F606F8060708F70808F7080907F8F907F90
9F809F9F809FA08FA0A08FA0AF90AFA090B0AF9FBFAF9FBFB0A0C0AFA0C0B0AFC0B0A0C0B0AFC0\
AFA0C0B0AFC0B0A0C0B0AFC0B0A0C0B0AFCFB0AFC0B0AFC0B0A0C0B0AFCFB0AFC0
B0AFC0AFA0C0B0AFC0AFA0C0AFAFC0AFA0C0AFAFCFAFA0C0B0AFCFAFA0C0B0AFCFB0AFCFB0AFCF\
B0AFCFBFAFCFB0AFCFBFB0CFBFAFCFBFB0D0BFAFCFBFB0D0BFAFCFC0B0CFBFAFC0
BFAFCFBFAFC0C0AFC0BFAFC0BFAFBFBFA0B0BFA0B0BFA0AFBFA0AFBF9FA0C0A0A0BF9F9FC09F9F\
BF9090BF9090B08F80B08F80AF807FAF8070A07F6FA07F6F9F70609F705F805F4F
704F40603F3F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F3F4F3F3F\
503F3F6F40406F40407F40408040408F4F408F4F40904F4F90504F9F5F509F5F50
A06050A06050AF6F5FAF6F5FB0705FB0705FBF7F5FBF7F5FC08060C0805FC08060C0805FC08F60\
C08F60C08F60C08F60C08F60BF8F60C08F60BF8F60C08F60BF8F60C09060BF8F60
C09060BF8F60C09060BF9060BF9060BF905FBF9060B0905FB0905FB0905FB0905FAF905FAF905F\
A08F50A08F509F80509F8050907F4F90804F8F7F4F8F7F4F8F70408F7040806F40
7F603F4F4030000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000503040\
5F404F604F506F5060705F607F6070806F7F8F70808F7080907F8F907F909F809F
9F809FA08FA0A08FA0AF90B0A090B0AF9FBFAF9FBFB0A0C0AFA0C0AFA0C0AFA0C0B0A0C0AFA0BF\
AFA0C0AFA0C0B0AFC0AFA0C0AFAFC0AFA0C0B0AFC0AFA0C0AFAFC0AFA0C0AFAFC0
A0A0BFAFA0C0A0A0BFAFA0C0A0A0BFA0A0C0A0A0C0AFAFC0A0A0C0AFAFC0AFA0C0AFAFCFAFA0C0\
B0AFCFAFAFC0B0AFCFB0AFC0B0AFCFB0AFCFBFAFCFB0AFCFBFAFCFBFAFCFBFB0D0
BFAFCFBFB0CFBFAFCFC0B0CFBFA0C0BFAFC0BFA0BFBFA0BFBFA0AFBFA0AFBF9FAFBFA0A0BF9F9F\
BF9F9FBF9F90BF9F90B08F8FB08F80AF8F7FAF8F7FA07F6FA07F6F9F70609F7060
8F6050805F4F6F4F3F603F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F3F3F5F4040603F3F\
7040407040408040408F4040904F40904F4F9F504F9F504FA05F50A05F50AF605F
A06050AF6F5FAF6F5FB0705FB0705FBF7F60BF7F5FC08060C0805FC08060C08060C08F60C08F60\
C08F60C08F60CF906FC08F60C0906FC09060C0906FC09060C0906FC09060C0906F
BF9060C09F60BF9060C09F60BF9060BF9F60B0905FB09060B0905FB0905FAF905FB09060A08F5F\
A08F5FA08F50A08F509F80509F8050907F4F90804F8F7F4F8F7F4F807040807040
70603F6F503F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000503040604F5F\
6F505F705F6F7F6070806F7F8F7080907F8F907F8F9F80909F809F9F8FA09F8FA0
A090AFA090AFAF9FB0AF9FBFB0A0C0B0A0C0B0AFC0AFA0C0B0AFC0AFA0C0AFAFC0AFA0C0B0AFC0\
AFA0C0B0AFC0AFA0C0B0AFC0AFAFC0B0AFC0AFAFC0AFAFC0AFAFC0AFAFC0A0A0C0
AFAFC0A0A0C0A0AFC0A0A0BFA0AFC0A0A0C0A0A0C0A0A0C0AFAFC0A0A0C0AFAFC0AFA0C0AFAFCF\
AFAFC0B0AFCFAFAFC0B0AFCFB0AFCFB0AFCFB0AFCFBFB0CFB0AFCFBFB0D0BFAFCF
BFB0D0BFAFCFBFB0D0BFAFD0C0B0D0BFAFCFBFAFC0BFAFBFC0AFBFBFA0B0C0AFB0BFA0AFC0A0AF\
BFA0A0C0A0A0BF9F9FC09F9FBF9090BF908FB08F80B08F80AF807FAF8070A07F6F
A07F6F9F705F906F5F7F504F704F405F3F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004F4040503F3F6040406F40407F4F40\
804040904F40904F409F504F9F504FA05F509F5F50AF605FAF605FAF6F60AF6F5F
B07060B07060BF7F60BF7F60C08060C08060CF8F6FC08060CF8F6FC08F60CF906FCF8F6FCF906F\
CF906FCF9F6FCF906FCF9F70CF9F6FCF9F70C09F6FCF9F70C09F6FCF9F6FC09F6F
CFA06FC09F6FC09F6FBF9F60C09F6FBF9F60BF9F60B09060B09F60B09F60B09F60AF9060AF9F60\
A0905FA0905F9F8F5FA08F5F9F80509F8F509080509080508F7F4F8F804F80704F
7F6F406F5F3F60503F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F30406040506F5060\
7F60707F60708F70808F7080907F8F907F8F9080909F809FA08FA09F8FA0A090AF
A090AFAF9FBFAF9FBFB0A0C0AFA0C0B0AFC0AFA0C0AFAFC0AFA0C0AFAFC0A0A0BFAFAFC0AFA0C0\
AFAFC0A0A0BFAFAFC0AFA0C0AFAFC0A0A0BFAFAFC0A0A0BFA0AFC0A0A0BFA0A0C0
A0A0BFA0A0C09FA0BF9FA0BF9FA0BFA0A0C09FA0BFA0A0C09FA0BFA0A0C0A0A0BFA0A0C0A0A0C0\
AFAFC0A0A0C0AFAFC0AFA0C0AFAFCFAFA0C0AFAFCFAFAFC0B0AFCFAFAFC0B0AFCF
B0AFCFB0AFCFB0AFCFBFB0D0BFAFD0BFB0D0BFAFCFBFAFCFBFAFC0C0AFC0BFA0BFBFAFBFBFA0B0\
C0A0B0BFA0A0C0A0A0BF9F9FC0A09FBF9F90BF9F90B0908FB09080AF807FAF807F
A0806FA0806F9F70609F70608F60507F504F6F403F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000403F3F503F40603F3F7040407F4040804F408F4040\
904F4F904F4F9F50509F5050A05F5FA05F50AF605FAF6F5FB07060B06F5FBF7060
BF7F60C08060C07F60CF806FC08060CF8F6FC08F60CF8F6FCF8F6FCF906FCF906FCF906FCF906F\
CF9F70CF9F6FD09F70CF9F70D0A070CF9F6FCFA070CF9F6FCFA070C09F6FCFA06F
C09F6FC0A06FBF9F6FBF9F6FBF9F60BF9F6FB09F60B09F6FB09F60B09F6FAF9F60AF9F6FAF9060\
AF9F60A0905FA0905F9F8F5F9F8F5F908F50908F509080509080508F804F8F7F4F
7F6F4070604060503F5F4F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F303F5F40506F50607F6070\
806F7F8F708F8F708F907F90907F909F809F9F8F9FA08FA0A090AFAF9FB0AF9FB0
AFA0BFAFA0BFB0A0C0AFA0C0B0AFC0AFA0C0B0AFC0AFA0C0AFAFC0AFA0C0AFAFC0AFA0C0AFAFC0\
A0A0C0AFAFC0A0AFC0AFAFC0A0A0BFAFAFC0A0A0BFA0AFC09FA0BFA0AFC09FA0BF
9FA0BF9FA0BF9FA0BF909FBF909FBF909FB09FA0BF9FA0BF9FA0BF9FA0BFA0A0C09FA0BFA0A0C0\
9FA0BFA0A0C0A0A0C0A0A0C0A0A0C0AFAFC0A0A0C0AFAFC0AFA0C0AFAFCFAFA0C0
AFAFCFAFA0C0B0AFD0B0AFCFBFB0D0BFAFD0C0B0DFBFB0D0C0B0CFBFAFC0C0B0C0BFAFBFC0AFBF\
BFA0B0C0AFB0C0A0AFC0A0AFBFA0A0C0A0A0C09F9FC09F9FBF908FBF908FB08F80
B08F7FAF8070AF8070A07F6FA07F609F705F8F60507F50406F403F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F3F3F503F3F6040406F40407F4F40804040904F4F904F4F9F504F\
9F504FA05F50A05F50AF605FAF605FB06F60B06F60BF706FBF7060C07F6FC07F6F
C0806FC0806FCF8F6FCF8F6FCF906FCF8F6FCF9070CF906FCF9F70CF906FD09F70CF9F70D09F70\
CF9F70D0A07FD0A070D0AF7FD0A070D0AF7FCFA070CFAF70CFA070CFA070C0A070
C0A070BFA06FC0A070BFA06FBFA06FB09F6FBFA06FB09F6FB0A06FAF9F6FB0A06F606F6FAFA06F\
A09F60A09F6FA09060A090609F905F9F9060908F5F908F5F908F50908F508F8050
807F4F7060406F5F405F4F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000402F3F5F3F4F604F5F705F707F6F7F\
8F708F8F708F907F90907F909F809F9F8FA0A090AFA090AFA09FB0A09FB0AFA0C0
AF9FBFAFA0C0AFA0C0B0AFC0AFA0BFAFAFC0AFA0BFAFAFC0A0A0BFAFAFC0A0A0BFA0AFC0A0A0BF\
A0AFC0A0A0BFA0AFC09FA0BFA0A0BF9FA0BF9FA0BF9F9FBF9FA0BF909FB09FA0BF
909FB0909FB08F9FB08F9FB08090AF8F90B08F90B0909FB08F90B0909FB0909FB09F9FBF909FB0\
9FA0BF9F9FBF9FA0BF9F9FBF9FA0BF9F9FBFA0A0C09F9FBFA0A0C0A0A0BFA0A0C0
A0A0BFAFA0C0A0A0C0AFAFCFAFA0CFB0AFD0B0AFD0BFB0D0BFAFD0BFB0D0BFAFCFC0B0CFBFAFC0\
C0AFBFBFA0BFC0AFB0BFA0AFC0A0AFBFA0A0C0A0A0BF9F9FC0A09FBF9F90BF9F90
B09080B08F80AF8F7FAF8F70A07F6FA07F6F9F70609F705F805F4F7050406F4030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000403F3F5040405F3F3F6F4040704040804F4F8F4040904F4F90504F9F5F509F5050\
A0605FA0605FAF6F60AF6F5FB07060B07060BF7F6FBF7F60C0806FC0806FCF8F6F
C08F6FCF8F6FCF8F6FCF9070CF8F6FCF9070CF9070D09F70CF9070D09F70CF9F70D0A07FCF9F70\
D0A07FD0A070D0AF7F605F3FD0AF7FCFA070CFAF7FCFA070CFAF70C0A070C0A070
BFA06FBFA070BFA06FBFA070B09F6FB0A06FB0A06FB0A070AF9F6FB0A06FAFA06F60706FA09F6F\
AFA06FA09F60A09F6F9F90609F9F609F905F9F9060908F5F90905F8F8F508F8F50
7F7F4F706F4F6F5F4060503F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F4050604F5F705F6F7F6F7F8F7F8F\
8F7F9090809F90809F9F8FA09F8FA0A090AFA090B0AF9FBFAF9FBFB0A0C0AFA0C0
B0AFC0AFA0C0B0AFC0AFA0C0B0AFC0AFAFC0AFAFC0AFAFC0AFAFC0A0AFC0A0AFC0A0AFBFA0AFC0\
A0A0BFA0AFC09FA0BF9FA0BF9FA0BF9FA0BF909FB09FA0BF909FB0909FB08F9FB0
8F9FB08F90AF8F90B08090AF8090AF7F8FAF8090AF8090AF8F90B08090B08F9FB08F90B0909FBF\
8F9FB0909FBF909FB09FA0BF909FBF9FA0BF9F9FBF9FA0C09FA0BFA0A0C09FA0BF
A0A0C0A0A0C0A0A0C0A0A0C0AFAFCFAFAFCFB0AFD0B0AFD0BFB0DFBFB0D0C0B0DFBFB0D0C0B0D0\
BFAFCFC0B0C0BFAFBFC0AFBFC0AFB0C0AFB0C0A0AFC0AFAFC0A0A0C0A0A0C09F9F
C0A09FBF908FBF908FB08F80B08F7FAF8070AF8070A07F60A07F608F6050805F4F6F4030000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4F3F40503F3F6040406F40407F4F4F804F40904F4F904F4F9F50509F5050A05F5FA05F50AF6F60\
AF6F5FB07060B07060BF7F6FBF7F6FC08070C0806FCF8F70CF8F70CF9070CF8F70
D09070CF9070D09070CF9070D09F7FD09F70D09F7FD09F70D0A07FD0A07FD0A07FD0A07FDFAF80\
D0AF7FDFAF80D0AF7FDFB080D0AF7FD0AF7FCFAF7FCFAF7FC0AF7FCFAF7FC0A070
C0AF7FBFA070BFAF70BFA070BFAF70B0A070B0AF70B0A070B0AF70B0A070B0AF70AFA070AFAF70\
AFA06FAFA06FA09F6FA0A06F9F9F609F9F609090609F906090905F90905F8F8F50
8F8F507F704F706F4F605F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000503F4F5F4050705F6F7F60708F708F8F708F\
90809F90809F9F8FA09F8FAFA090B0A090B0AF9FBFAF9FBFAFA0C0AFA0BFAFA0C0
AFA0C0B0AFC0AFA0C0AFAFC0AFA0C0AFAFC0A0A0BFAFAFC0A0A0BFA0AFC09FA0BF9FA0BF9FA0B0\
9FA0BF909FB090A0B0909FB0909FB08F9FB08F9FB08F90AF8F9FB08090AF8090AF
7F8FAF7F90AF708FA07F8FA07080A0708FA07080A07F8FA07F8FA07F8FAF7F8FA08090AF808FAF\
8090B08090AF8F90B08F90B08F9FB08F90B0909FB0909FB0909FBF909FB09F9FBF
909FBF9FA0C09F9FBFA0A0C0A0A0C0AFAFCFAFA0CFB0AFCFB0AFCFBFB0D0B0AFD0BFB0DFBFB0D0\
C0B0D0BFAFCFC0B0CFBFAFC0C0AFC0BFAFB0C0AFB0C0A0AFC0AFAFC0A0A0C0A0A0
C0A09FC0A09FBF9F90BF9F8FB09080B09080AF8F70AF8F70AF806FA07F608F60507F5F406F4030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000004F3F40
503F406040406F40407F4F4F804F4F9050509F504F9F5F50A05F50AF605FA0605FAF6F60AF6F60\
B0706FB07060BF7F6FBF7F6FC08070C08070CF8F70CF8F70D09070CF8F70D09070
CF9070D09F7FCF9070D09F7FD09F7FD0A07FD09F7FD0A07FD0A07FD0AF80D0A07FDFAF80D0AF7F\
DFB080D0AF7FD0B080D0AF7FD0B080CFAF7FCFAF7FC0AF7FCFAF7FC0AF70C0AF7F
BFAF70BFAF7FB0A070BFAF70B0A070BFAF7FB0A070B0AF70AFAF70B0AF7FAFA070AFA070A0A070\
A0A0709F9F6F9F9F6F9F9F609F9F6F9090609090608F905F8F90608F8F5F8F8F5F
80805080804F706F406F6040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000503F4F5F40506F506F7F60708F708F8F708F90809F\
90809F9F90AFA090AFA09FB0A09FBFAFA0C0AFA0C0AFA0C0AFA0C0B0AFC0AFA0C0
B0AFC0AFAFC0B0AFC0AFAFC0AFAFC0A0AFBFAFAFC0A0AFC0A0AFC09FA0BF9FA0BF909FB090A0B0\
8F9FB0909FB08F9FB08F9FB08090AF8F9FB08090AF8090AF7F8FA07F90AF7F8FA0
7F8FA07080A0708FA06F809F6F80A0607F9F6F80A06F809F7080A06F80A0708FA07080A07F8FAF\
708FA07F8FAF7F8FAF808FAF7F8FAF8090B08090AF8F90B08090AF8F90B08F90B0
909FBF909FBF9FA0C09F9FBFA0A0C0A0A0C0AFAFCFAFA0CFAFAFCFAFAFCFB0B0D0B0AFD0BFB0DF\
BFB0D0C0B0DFBFB0D0C0B0CFBFAFC0C0B0C0C0AFBFC0AFBFC0AFB0C0AFB0C0A0AF
C0AFAFC0A0A0CFA0A0C09F9FC09F90BF908FBF908FB08F7FB08F7FAF806FA07F6F906F50805F4F\
6F4F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000403F3F5040405F4040
6F4F4F704F4F80504F8F504F9F50509F5F50A0605FA0605FAF6F60AF6F5FB0706FB07060BF7F6F\
BF7F6FC08070C08070CF8F7FCF8F70D0907FCF907FD0907FCF907FD09F7FD09F7F
D09F7FD09F7FDFA080D0A07FDFA080D0A07FDFAF80D0AF80DFAF80D0AF80DFB080DFB080DFB08F\
D0B080DFB080D0B080D0B080CFB080CFB080CFB07FCFB080C0AF7FC0B080BFAF7F
BFAF7FBFAF7FBFB07FBFAF7FBFAF7FB0AF7FB0B07FB0AF7FB0AF7FAFA070AFAF7FA0A070A0A070\
9F9F6F9FA070909F6F909F6F8F9060909F6F8F90608F906F8F90608F9060808F5F
808F507F804F7F7F4F6F6F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F30405F40506F5060705F708070808F708F907F9090809F\
9F8FA09F8FAFA09FBFA09FBFAFA0C0AFA0BFAFA0C0AFA0BFAFA0C0AFA0C0AFAFC0
AFA0BFAFAFC0AFAFBFAFAFC0A0A0BFA0AFC0A0A0BFA0AFBF9FA0B09FA0B0909FB08F9FB08090AF\
8F9FAF8090AF8090AF7F90A08090AF7F8FA07F8FA0708FA0708FA06F809F6F80A0
607F90607F905F708F5F708F4F6080506080506F805F708F5F7090607F9F607F9F6F7F9F607F9F\
6F809F6F7F9F6F80A06F7F9F7080A07080A0708FA07080A07F8FAF7F80A07F8FAF
7F8FAF8F90B08F90B0909FBF909FBF9FA0C09FA0C0A0A0C0A0A0C0AFAFCFAFAFCFB0AFD0B0AFCF\
BFB0D0BFB0D0C0B0DFBFB0D0C0B0D0BFAFCFC0B0CFBFAFBFC0AFBFC0AFB0C0AFB0
C0A0AFC0AFAFC0A0A0C0A0A0C0A09FC0A09FBF908FBF908FB08F7FB08F7FAF806FA07F6F8F6F50\
805F4F6F4F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F3F3F5F4040604040704F4F
7F4F4F80504F90504F9F5F509F5F50A0605FA0605FB06F60B06F60BF706FBF706FC07F70BF7F6F\
C08070CF8F70D0907FCF8F70D0907FCF907FD09F7FD0907FD09F7FD09F7FD0A080
D09F7FD0A080D0A07FDFAF80D0A080DFAF80D0AF80DFB080D0AF80DFB08FDFB080DFB08FD0B080\
D0B080D0B080D0B080CFB080CFB080C0B080CFB080C0AF7FC0B080BFAF7FC0B080
BFAF7FBFB07FB0AF7FBFB080B0AF7FB0B07FAFAF7FAFAF7FA0A070A0AF7F9FA0709FA07F909F70\
90A070909F6F909F6F8F906F8F9F6F8F906F8F9F6F8090608F906F808F60808F5F
7F80507F8050707F40707040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F3F405040506F5060705F708070808F708F907F9090809F9F8FAF\
9F90AFA09FBFAF9FBFAFA0C0AFA0C0AFA0C0AFA0C0B0AFC0AFAFC0B0AFC0AFAFC0
B0AFC0AFAFC0AFAFC0A0AFC0AFAFC0A0AFBFA0AFBF9FA0BF9FA0BF909FB08F9FB08090AF809FAF\
8090AF8090AF7F8FA07F90AF708FA0708FA06F809F6F809F607F905F7F8F506F80
506F804F607F4F607040506F40506F304F603F4F603F4F6F405F70405F704F607F4F60805F6F8F\
5F708F607F9F5F7090607F9F607F9F6F7F9F607F9F6F80A06F7F9F7080A06F80A0
7F8FAF7F8FAF8090B08F90B0909FBF909FBF9FA0C09FA0C0A0A0CFA0A0C0AFAFCFAFAFCFB0B0D0\
B0AFD0BFB0D0BFB0D0C0BFDFBFB0DFC0B0DFC0B0D0C0B0D0C0B0C0C0B0C0C0AFBF
CFB0BFC0AFB0CFAFB0C0AFAFCFAFAFCFA09FCFA09FC09F8FC09F8FB09080B08F7FAF8070AF806F\
906F5F80604F704F3F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F3F3F503F4060404F6F404F7F4F50804F4F
9050509F5050A05F5FA05F5FAF6F60AF6F60B0706FB0706FBF7F70BF7F6FC08070C08070D0907F\
D0907FD09080D0907FDF9F80D09F80DF9F80D09F80DFA080D0A080DFA080D0A080
DFAF8FDFAF80DFAF8FDFAF80DFB08FDFB08FDFB08FDFB08FE0BF8FDFB08FDFBF8FD0B08FDFBF8F\
D0B08FD0BF8FCFB080D0BF8FCFB080CFBF8FC0B080C0BF8FC0B080C0B080BFB080
C0BF80BFB080BFB080B0B07FB0B080AFAF7FAFAF7FA0AF7FA0AF7F9FA07F9FAF7F90A0709FA07F\
909F70909F708F9F6F8F9F708F9F708F9F708F9F6F8F9F7080906F8090607F8F5F
808F5F7F80507F804F707040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000403040503F4F604F606F5F6F7F6F808070808F7F90907F909F8FA09F8FAF\
A09FB0A09FBFAFA0C0AFA0C0AFA0C0AFA0BFAFAFC0AFA0C0AFAFC0AFAFBFAFAFC0
AFAFC0AFAFC0A0AFBFA0AFC0A0A0BFA0AFBF9FA0B090A0B08F9FAF8F9FAF8090A08090AF7F8FA0\
7F8FA06F809F6F8090607F8F607F8F507080506F804F607F4F607F3F5F6F3F506F
304F6000000000000000000000000000000000000000000000000000000000000030405F3F4060\
40506F3F506F405F7F4F5F7F50608050608F5F70905F709060709F5F7090607F9F
607F9F7080A07080A07F8FAF808FAF8F90B08F90B0909FBF909FBF9FA0C09FA0C0A0A0CFA0A0CF\
AFAFCFAFAFCFB0B0D0B0AFD0BFB0DFBFB0D0C0B0DFBFB0D0C0B0D0BFAFCFC0B0CF
C0AFC0C0B0C0C0AFB0C0AFB0CFAFAFD0AFAFCFA09FCFA09FC09F8FBF9F8FB08F7FB08F7FAF8070\
AF806F906F5080604F704F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F3F40503F4060404F6F404F7F4F50804F4F905050
9F5050A05F5FA05F5FAF6F60AF6F60B0706FB0706FBF7F70BF7F70C0807FC080708F806FD0907F\
D09F80D0907FD09F80D09F80DFA080D09F80DFA080D0A080DFAF8FD0A080DFAF8F
DFAF80DFAF8FDFAF8FDFB08FDFB08FE0B08FDFB08FE0BF8FDFB08FDFBF8FD0B08FD0BF8FD0B08F\
D0BF8FCFB080CFBF8FCFB08FCFBF8FC0B080C0BF8FC0B080C0BF8FBFB080BFBF80
BFB080B0B080AFAF7FAFAF80A0AF7FA0AF7F9FA07F9FAF7F90A07F9FAF7F90A07F90A07F8F9F70\
8FA0708F9F708F9F70809F708F9F70809F708F9F70809F7080906F7F8F607F8F60
7080507F8050707F4F707F4F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000403040503F4F604F5F6F506F7F6F8080708F8F7F90907F909F8FA09F8FAFA090B0\
A09FBFAFA0CFAFA0C0AFAFC0AFA0C0AFAFC0AFAFC0AFAFC0AFAFC0B0B0C0AFAFC0
AFAFC0A0AFBFAFAFC0A0AFBFA0AFBF9FA0B09FA0B08F9FAF8F9FAF8090A08090AF708FA06F8090\
5F70805F70804F6F7F4F6F7F406070405F6F3F5060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000303F5F404F60404F6F405070405070506080506F8F
5F70905F70906F7FA06F80A07F8FAF7F8FAF8F90B08F90B0909FBF909FBF9FA0C09FA0C0A0A0C0\
A0A0C0AFAFCFAFAFCFB0B0D0B0AFD0BFB0DFBFB0DFC0BFDFBFB0DFC0BFDFC0B0D0
C0B0CFC0B0C0C0B0C0CFAFB0CFAFB0CFAFAFD0AFAFCFA0A0CFA09FC09F90C09F8FBF9080B0907F\
AF8070AF806F90705F8F604F704F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000403F3F5040405F404F6F4F50704F4F8050508F50509F5F5F9F5F5F
A06060A06060AF6F6FB06F6FB0706FBF7F6FC08070C08070CF8F7FCF8F7FD09080D09080DF9F8F\
D09F80DFA08FDFA080DFA08FDFA080DFAF8FDFA08FDFAF8FDFAF8FE0B090DFAF8F
E0B090DFB08FE0BF90DFB090E0BF90DFBF90E0BF90DFBF8FDFBF90D0B08FD0BF8FD0BF8FD0BF8F\
CFBF8FCFBF8FCFBF8FCFBF8FC0BF8FCFBF8FC0BF8FC0BF8FBFBF8FBFBF8FB0B080
B0B080AFAF80AFB080A0AF7FA0AF809FA07F9FAF8090A07F90AF808FA07F90A0808FA07F8FA07F\
8F9F7F8FA07F8F9F7F8FA07F809F7F8FA07F809F7F809F707F906F80906F7F8F60
7F8F607080507F8050707F4F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004F3F4F5F4F5F6050607F607F806F808F7F908F7F909080A09F8FA0A090B0A090B0\
AFA0C0AFA0C0AFA0C0AFA0C0AFAFC0AFA0BFAFAFC0AFAFBFAFAFC0AFAFBFAFAFC0
A0AFBFA0AFBF9FA0B09FA0BF90A0B090A0B08F9FAF8F9FAF8090A07F90A0708090607F8F506F7F\
4F60703F5F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003030503F3F5F
404F70506F8F5F6F90607F9F607F9F7080A07080A0808FAF808FAF8F90B09090B09F9FBF909FBF\
9FA0C09FA0C0A0AFCFA0A0C0AFAFCFAFAFCFB0B0D0B0B0D0BFB0DFBFB0DFC0BFDF
BFB0D0C0B0D0C0AFC0C0AFBFC0AFB0CFAFB0CFAFAFD0AFAFCFA09FCFA09FC09F90C09F8FB09080\
B08F7FAF8070AF806F90705F8F604F70503F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F3F3F5F404F5F404F6F4F50704F508050508F50509F5F5F9F5F5FA06060
A06060AF6F6FB0706FBF7F70BF7F70C08070C0807FCF8F7FCF8F7FD09080D09F80DFA08FDF9F80\
DFA08FDFA08FDFA08FDFA08FDFAF8FDFAF8FDFAF90DFAF8FE0B090DFB08FE0B090
DFB08FE0BF90E0BF90E0BF90DFBF90DFBF90DFBF8FDFBF90D0B08FD0BF8FCFBF8FD0BF90CFB08F\
CFBF8FC0BF8FCFBF8FC0BF8F604030C0BF8FC0BF8FBFB080B0B08FAFB080AFB080
A0AF7FA0AF809FAF7F9FAF8090A07F90AF808FA07F90A080606F6F8FA0808FA07F8FA080809F7F\
8FA07F809F7F80A07F7F9F7F809F7F7F9F7F809F7F809F70809F707F906F7F906F
708F5F7F805F708050707F4F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F3F4F5F4F5F6050607F607F806F808F7F908F7F909F8FA09F8FA0A090B0A090BFAFA0C0\
AFA0C0B0AFCFAFA0C0B0AFCFAFAFC0AFAFC0AFAFC0B0B0C0AFAFC0AFB0C0A0AFBF
A0AFBF9FA0BF9FAFBF90A0B090A0B08F9FAF8F9FAF8090A07F90A06F8090607F8F506F7F4F6070\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F3F5F4F5F7F506F8F5F709F60709F6F7FA07080A07F8FAF7F8FAF8F90B08F90B0909FBF\
909FBF9FA0C09FA0C0A0AFCFA0A0CFAFAFCFAFAFCFB0B0D0B0B0D0BFB0DFBFB0DF
C0BFE0BFB0CFC0B0CFC0AFC0CFB0C0CFAFBFCFB0B0CFAFAFD0B0AFCFAFA0CFA09FC09F90C09F90\
BF9080B09080AF8F70AF8F709F705F8F6F5070503F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
503F405F404F604F4F7F50507F50508F505F90505FA06060A06060AF6F6FAF6F6F
B07070B07070BF7F70BF7F70C08F7FC08F7FCF9080CF9080DF9F8FDF9F8FDFA08FDFA08FE0AF90\
DFAF90E0AF90DFAF8FE0B090DFAF90E0B090E0B090E0BF9FE0B090E0BF9FE0BF90
EFC09FE0BF9FE0C09FDFBF90E0C09FDFBF90DFBF90D0BF90D0C090D0BF90D0BF90CFBF8FCFC090\
CFBF90CFC090C0BF8FC0C090C0BF90BFBF90B0B08FB0BF8FAFB080AFB08FA0AF80
A0B0809FAF809FAF8090A08090AF808FA0808FAF8080A0808FA08080A08080A0807F9F7F809F80\
7F9F7F7F9F7F70907F7F9F807F9F7F7F9F807F9F7F809F7F809F70809F707F906F
7F906F7F8F6070805F6F7040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4030405F405F604F60705F707F6F808F708F8F7F9090809F9080A09F90B09F90B0AF9FC0AFA0C0\
AFA0C0AFA0C0AFAFC0AFAFC0AFAFC0AFAFBFAFAFC0AFAFC0AFAFC0A0A0BFA0AFBF
9FA0B09FA0B0909FAF90A0B08F9FAF8F9FAF7F90A0708F9F6F7F8F5F70804F6070000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F3F5F405070506F90506F9060709F6F7F9F7080A07080A0808FAF808FAF\
909FBF9090B09F9FBF9F9FBFA0A0C0A0A0C0AFAFCFAFA0CFB0AFD0AFAFCFB0B0D0
BFB0DFBFB0D0BFAFCFC0B0CFC0AFC0CFB0C0C0AFB0CFAFB0CFAFAFD0AFAFCFA0A0CFA09FC09F90\
C09F8FB09080B0907FAF8F70AF8F709F705F8F6F507F503F000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F3F40503F40\
604F4F6F4F507F5050805050905F5F9F505FA06060A06060AF6F6FAF6F6FB07070
B07070BF7F7FBF7F7FC08F7FC08F80CF9080CF9080D09F8FDF9F8FDFA090DFA08FDFAF90DFAF90\
E0AF90DFAF90E0B090E0B090E0B09FE0B090E0BF9FE0B090E0BF9FE0BF90E0C09F
E0BF9FE0C09FDFBF90DFC09FDFBF90DFC090D0BF90D0BF90CFBF90D0C090CFBF8FCFBF90C0BF90\
CFC090C0BF8FC0C090BFBF90BFBF90B0B08FAFB08FA0B080A0B08F9FAF809FAF80
90AF8090AF808FA0808FAF808FA0808FAF807F9F7F7F9F8070907F709F806F907F70907F6F907F\
7090806F907F709080709080709F8070907F7F9F7F7F907F7F9F7F7F90707F9070
7F8F6F7F8F6F6F7F506F7040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000403040
5040505F4F60705F707F6F8080708F8F7F9090809F9080A09F90AF9F90B0AFA0C0AFA0C0AFA0C0\
AFA0C0B0AFCFAFAFC0AFAFC0AFAFC0AFB0C0AFAFC0AFAFC0A0AFBFA0AFBF9FA0B0
9FA0B090A0AF90A0B08F9FAF8F9FAF7F909F708F90607F805F707F4F606F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F3F5F4F5F7F506F8F5F709060709F6F7FA07080A07F8FAF\
7F8FAF8F90B08F90B0909FBF909FBF9FA0C09FA0C0A0A0CFA0A0C0AFAFCFAFAFCF
B0B0D0BFB0DFC0B0DFBFB0CFC0B0CFC0AFC0CFB0C0CFAFBFCFB0BFCFAFAFD0B0AFCFAFA0CFAF9F\
C09F90C09F90BF9080BF9080AF8F70AF8F709F7F608F6F507F5040000000000000
0000000000000000000000000000000000000000000000000000000000005040405F404F6F4F50\
704F5080505F8F505F9F5F609F5F60A0606FA0606FAF6F70AF7070B07F7FBF7F7F
BF807FC0807FCF8F80CF8F80D0908FD09F8FDFA090DFA090DFAF90DFAF90E0B090DFAF90E0B09F\
DFB090E0B09FE0B09FE0BF9FE0BF9FEFBF9FE0BF9FEFC09FE0BF9FEFC0A0E0C09F
E0C09FDFBF9FE0C09FDFBF90DFC09FD0BF90D0C09FCFBF90D0C090CFBF90CFC09FCFC090CFC090\
C0BF90C0C09FBFBF90BFBF90AFB08FAFBF8FA0B08FA0B08F9FAF809FAF8090AF80
90AF808FA0808FAF8F80A080709F806F907F6F90806F907F6F907F608F7F6F90806090806F9080\
6090806F908F6F90807090806F907F709F8070907F7F9F7F7090707F9F7F7F9070
7F9070708F60708050606F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F3040504050
5F4F5F6F5F7070607F80708F80708F8F809F9080A09F8FAF9F90B0A09FBFA09FC0AFA0C0AFA0C0\
AFAFC0AFA0C0AFAFC0AFAFBFAFAFC0AFAFBFAFAFC0A0A0B09FAFB09FA0B09FA0B0
909FAF909FAF8F9FA0809FA0708F906F808F5F707F506F70000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F406040507F506F905F6F9060709F607F9F\
7080A07F80A0808FAF808FAF909FB0909FBF9F9FBF9F9FBFA0A0C0A0A0C0AFAFCF
AFA0CFB0AFD0B0B0D0BFB0DFBFAFCFC0B0CFC0AFC0CFB0C0C0AFBFCFB0B0CFAFAFD0AFAFCFAFA0\
CFAF9FC09F90C09F90BF9080B09080AF8F70AF8F709F7F608F6F507F5F40000000
000000000000000000000000000000000000000000000000403F3F5040405F404F6F4F50704F50\
80505F8F505F9F5F609F5F60A0606FA0606FAF7070AF7070B07F7FBF7F7FC08080
C0807FCF8F80CF9080D09F8FD09F8FDFA090DFA090DFAF90DFAF90DFB090DFAF90E0B09FDFB090\
E0B09FE0B09FEFBF9FE0BF9FEFBF9FE0BF9FEFC0A0E0C09FE0C0A0E0C09FE0C09F
DFBF9FDFC09FD0BF90DFC09FD0BF90D0C090CFBF90CFC09FCFBF90CFC090C0BF90C0C09FC0BF90\
BFC090B0BF8FB0BF90A0B08FA0B08F9FAF809FB08F90AF8090AF808FA0808FAF8F
80A0807F9F806F907F6090805F8F7F5F8F7F5F8F7F608F805F8F7F608F805F8F8060908F5F8F80\
60908F6090806F908F6F8F806F90806F907F70908070907F70907F7090707F907F
7F9070708F6060704F606F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F30404F40505F4F5F
6F5F7070607F80708F80709090809F9080A09F8FAF9F90B0A09FBFAFA0C0AFA0CFAFA0C0B0AFCF\
AFAFC0B0AFC0AFAFC0B0B0C0AFAFC0AFB0C0A0AFBFA0AFBF9FA0B09FA0B0909FAF
90A0AF8F9FA08090A0708F906F808F5F707F506F70000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003F40604F5F80506F8F5F709F\
60709F6F7FA07080A07F8FAF808FAF8F90B09090B09F9FBF9F9FBF9FA0C0A0A0C0
AFAFCFAFAFD0B0B0DFBFB0DFC0B0DFBFB0D0C0B0CFC0B0C0CFB0C0CFB0BFCFB0BFCFAFB0D0B0B0\
D0AFA0CFAFA0C0A090C0A090BF9F8FBF9080B08F7FB08F70A07F6090705F7F5F40
704F3F0000000000000000000000000000000000004F3F405F404F604F5070505F7F505F8F5F60\
905F60A06060A06060AF6F70AF6F70B07070B07070BF8080BF807FC08F80C08F80
CF908FCF908FD09F90D09F90DFAF9FDFAF90E0B09FDFB090E0B09FE0B09FE0BF9FE0B09FE0BFA0\
E0BF9FEFC0A0E0BFA0EFC0A0EFC0A0EFCFA0E0C0A0EFCFA0E0C0A0E0C0A0DFC09F
DFC0A0DFC09FDFC09FD0C09FD0C09FCFC09FD0C09FCFC09FCFC09FC0C09FCFC09FC0C090BFC090\
B0BF90B0BF90A0B08FA0B0909FB08F9FB08F90AF8090AF8F8FAF8F8FAF8F709F7F
6F90805F8F7F5F8F8050807F405F8F405F804070804070804F7F8F4F7F805F808F5F808F608F8F\
608F8F6F90906F908F709F8F70908F709F8F7090807F9F807F90807F9F807F907F
7F9F706F80606F7F505F603F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004F3F4F50405F60506F
70607F80708F80708F8F7F9F8F809F9F8FAF9F8FAFA09FBFA09FC0AFA0C0AFA0C0AFAFC0AFAFC0\
B0AFC0AFAFC0AFAFC0AFAFC0AFB0C0A0AFBFA0AFBF9FA0B09FA0B08F9FAF8F9FAF
809FA07F909F6F808F607F80506F70000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004040604F5F80\
506F905F6F9060709F6F7F9F7080A07080A0808FAF8F90B0909FBF9090BF9F9FC0
9FA0C0AFAFCFAFA0CFB0AFD0B0B0D0BFB0DFBFB0D0C0B0CFC0B0C0CFB0C0C0AFBFCFB0BFCFAFB0\
D0B0AFCFAFA0CFAFA0C0A090C0A090BF9080BF9080B08F7FB08F70A07F60907050
7F5F40704F3F0000000000000000004F3F40503F4060404F6F4F507F505F80505F905F609F5F60\
A0606FA0606FAF6F70AF6F70B0707FB0707FBF8080BF8080C08F8FC08F80CF908F
CF908FDFA090D09F90DFAF9FDFAF90DFB09FDFB090E0B09FDFB09FE0BFA0E0BF9FE0BFA0E0BF9F\
EFC0A0E0BFA0EFC0A08F6F50EFCFAFE0C0A0E0C0A0E0C0A0E0C0A0DFC09FDFC0A0
D0C09FD0C09FD0C09FD0C09FCFC09FCFC09FC0C09FCFC09FC0C09FC0C09FBFBF90B0BF90AFB08F\
AFBF909FB08F9FB08F90AF8F90AF8F8FAF808FAF8F7FA080709F805F8F70508F7F
3F303F403F60404070505F80405F8040608F40608F4F6F8F4F708F507F8F507F8F5F808F5F808F\
608F8F608F8F6F90906F908F709F9070908F709F8F7090807F9F8070907F7F907F
7F90706F80605F6F4F5F603F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000504050605F6F70607F
7F6F8F80708F8F7F9F8F809F9F8FAF9F8FAFA09FBFA09FC0AFA0C0AFA0C0B0AFCFAFAFC0B0AFC0\
AFAFC0B0B0C0AFAFC0AFB0C0AFAFBFAFAFBF9FA0B09FA0B0909FAF90A0AF809FA0
7F8F906F7F80607F7F506F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F40604F608F506F8F5F709060709F6F7FA07080A0808FAF8F90B0909FBF909FBF
9FA0C09FA0C0AFAFCFAFAFCFB0B0D0B0B0D0C0BFDFBFB0D0C0B0D0C0B0C0CFB0C0CFB0BFCFB0BF\
CFAFB0D0B0B0D0AFAFD0AFA0C0A09FC0A090BF9F8FBF9F80B08F7FB08F70A0806F
90705F805F4070503F0000005040405F40406F4F50704F5080505F8F50609F5F609F5F60A0606F\
A0606FAF7070AF7070B07F7FB07F7FC08080C08080CF8F8FCF908FD09F90D09F90
DFA09FDFA090DFAF9FDFAF9FE0B09FDFB09FE0BFA0E0BF9FE0BFA0E0BFA0EFC0A0E0C0A0EFC0A0\
E0C0A0EFCFAFEFCFAFEFCFAFE0C0A0EFCFAFE0C0A0E0CFA0DFC0A0DFCFA0D0C0A0
D0CFA0D0C09FD0CFA0CFC09FCFCFA0CFC09FCFCFA0C0C09FC0C09FB0BF90B0BF90AFB08FA0B090\
9FB08F9FB09090AF8F90AF8F8FAF8F7FA0806F907F5F8F7F4F8070000000000000
403F4F3F3F60404F7F50508050608F4F608F506F904F6F90507F90507F905F809F5F8090608F90\
608F906F90906F9090709F90709F907F9F90709F8F7F9F8F7090807F9F807F9070
7F90706F7F5F60704F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000040304F5F4F606F5F707F6F80
7F6F8F8F7F908F7F9F908FA0908FAFA090B0A09FBFAFA0C0A0A0C0AFA0C0AFA0C0B0AFC0AFAFC0\
AFAFC0AFAFC0AFB0C0A0AFBFA0AFBFA0A0B09FA0B0909FAF90A0AF8F9FA07F909F
6F7F805F7070000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030304F404F6F4F5F80506F905F6F9060709F6F7F9F7F8FAF808FB0909FBF
9090BF9F9FC09FA0C0A0AFCFAFA0CFB0AFD0B0B0D0BFB0DFBFB0D0C0B0CFC0B0C0CFB0C0C0AFBF\
CFB0BFCFAFB0D0B0B0CFAFA0CFAFA0C0A09FC0A090BF9F80BF9080B08F7FB08F70
A0806090705F403F3F5040405F404F6F4F50704F508050608F50609F5F6F9F5F6FA06F70A06F70\
AF707FAF707FB07F80B07F80BF808FC08F8FCF908FCF908FD09F90D09F90DFA09F
D0A09FDFAF9FDFAF9FE0B0A0DFB09FE0BFA0DFBF9FE0C0A0E0BFA0E0C0A0E0C0A0EFCFAFE0C0A0\
EFCFAFEFCFAFEFCFAFE0CFA0E0CFAFDFC0A0E0CFA0DFC0A0DFCFA0D0C0A0D0CFA0
CFC09FD0C0A0CFC09FCFCFA0C0C09FCFCFA0C0C09FBFC09FB0BF90AFBF90A0B08FA0B09090AF8F\
90AF8F8FAF8F80AF8F70907F60907F5080704F807F000000000000000000000000
403F5040407040407050508050608F506F90506F905070905070905F7F9F5F809F608F9F608090\
6F8F9F6F8F906F909F6F9090709F90709F90709F8F7090807F908070907F7F907F
708F6F6F7F5F5F6F40000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005040505F506070607F7F708F
8F7F9F8F7F9F908FA0908FAFA09FBFA09FBFAFA0CFAFA0C0AFAFCFAFAFC0B0AFC0AFAFC0B0B0C0\
AFAFC0B0B0C0AFAFBFAFB0BFA0AFB0A0AFB09FA0AF9FA0AF8F9FA080909F70808F
607F7F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F3F50404F6F4F608F50608F60709F6F7FA07F8FAF8090B0
909FBF909FBF9FA0C09FA0C0AFAFCFAFAFD0B0B0D0B0B0DFC0BFDFBFB0D0C0B0D0C0B0CFCFBFCF\
CFB0BFCFB0BFCFB0B0D0B0B0D0B0AFD0AFA0CFA09FCFA09FC09F8FBF9F80B0907F
B0907F4F3F405F404F60404F70505F7F505F8F5F60905F60A0606FA0606FAF6F70AF6F70B07F7F\
B07F7FBF8080BF8080C08F8FC08F8FCF9090CF9090D0A09FD0A09FDFAFA0DFAF9F
E0B0A0DFB0A0E0BFA0DFBFA0E0C0AFE0C0A0E0C0AFE0C0A0EFCFAFE0CFAFEFCFAFEFCFAFEFD0B0\
EFCFAFEFD0B0E0CFAFEFD0AFE0CFAFE0CFAFDFCFA0DFCFAFD0CFA0D0CFA0D0C0A0
D0CFA0CFCFA0CFCFA0CFCFA0CFCFA0BFC09FBFC09FAFBF90AFBF90A0B08F9FB09090AF8F90B090\
8FAF8F709F805F8F7F50807F0000000000000000000000000000000000003F3030
4F40604F40705050805050805F6F90506F905F709F5F7090607F9F60809F608FA0608F9F6F8F9F\
6F8F9F70909F6F909F709F9F709F907FA09F709F8F7F9F8F7090807F907F709070
7F90706F7F5F5F6F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000040304F50405F605F7070607F807090
8F7F9F9080A0908FAF9F90B0A090BFAFA0C0A0A0C0AFA0C0AFA0C0AFAFC0AFAFC0B0AFC0AFAFC0\
B0B0C0AFAFBFAFAFBFA0AFB0A0AFB09FA0AF90A0AF8F9FA08F9FA070808F6F7F7F
506F6F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030304F405F8050608F60709F6F7F9F7F80AF
808FB08F90BF9090BF9F9FC09FA0C0A0AFCFAFA0CFB0AFD0B0B0D0BFB0DFBFB0D0C0B0D0C0B0CF\
CFB0CFC0B0BFCFB0BFCFB0B0D0B0B0CFAFAFD0AFA0CFA09FCFA090BF9F8FBF9F80
503F4060404F6F40507F505F80505F905F6F905F609F606FA06070AF6F7FAF6F70B07F7FB07F80\
BF808FBF8080C08F8FC08F90CF9F90CF9F90D0A09FD0A09FDFAFA0DFAF9FDFB0A0
DFB0A0E0BFA0DFBFA0E0C0A0E0C0A0E0CFAFE0C0A0E0CFAFE0CFAFEFCFAFE0CFAFEFD0B0EFD0AF\
EFD0B0E0CFAFE0CFAFDFCFAFE0CFAFDFCFA0DFCFAFD0CFA0D0CFAFCFC0A0CFCFA0
CFCFA0CFCFA0C0C0A0C0C0A0B0BF9FB0BF9FA0B090A0B0909FB08F9FB0908FAF8F8FAF8F709F7F\
5F8F7F407F70000000000000000000000000000000000000000000000000403F40
4F406F4F4F7F4F507F505F8F5F608F60709F5F7090607F9F5F7F9F6080A060809F6F8FA06F8F9F\
6F909F6F909F70909F7090907F9F9F7F9F907F9F8F7090807090807090707F9070
708F6F6F7F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F3F4F5F4F606F5F7080708F807090
8F80A0908FAF9F90B0A090BFAFA0C0A0A0C0AFAFCFAFA0C0AFAFC0AFAFC0B0B0C0AFAFC0B0B0C0\
AFB0C0AFB0C0A0AFBFA0AFBF9FA0AF9FA0B0909FAF8F9FA07F8F907080805F7070
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000405F804F608F60709F6F7F9F
7F80AF808FAF909FBF909FBF9FA0C09FA0C0AFAFCFAFAFCFB0B0D0B0B0D0BFBFE0BFB0D0C0BFD0\
C0B0CFCFBFCFCFB0C0CFB0BFCFB0B0D0B0B0D0B0AFD0B0A0CFAF9F4F3F40504040
604F506F4F507F505F8F5060905F6F9F5F6FA06F70A06F70AF707FAF707FB07F80B07F80BF808F\
BF8F8FCF9090CF9090D09F9FD09F9FDFA0A0DFA0A0DFAFA0DFAFA0E0BFAFDFBFA0
E0C0AFE0C0A0E0CFAFE0C0AFE0CFAFE0CFAFEFD0B0E0CFAFEFD0B0EFD0AFEFD0B0EFD0B0EFD0B0\
E0CFAFE0D0B0DFCFAFDFCFAFDFCFAFDFD0AFD0CFAFD0CFAFCFCFA0D0CFAFCFCFA0
CFCFAFC0C0A0BFC0A0B0BF9FAFBF9FA0B090A0B0909FB08F90B0908FAF8F8FAF8F508070407F70\
000000000000000000000000000000000000000000000000000000000000403F50
4F4070504F7F5F5F8F5F608F606F9F60709F6F7FA0607F9F6F80A06F80A0708FA06F90A07090A0\
6F909F709FA070909F7F9F9F7F9F9F80A09F7F9F8F7F9F8F70907F7F907F709070
708F6F5F704F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000050405F5F4F6070607F7F708F8F7F9F
8F80A0908FAF9F90BFA09FC0A09FC0AFA0C0AFA0C0AFAFC0AFA0C0AFAFC0AFAFC0B0B0C0AFAFBF\
AFB0BFA0AFBFA0AFBF9FA0AF9FA0AF909FA090A0A07F8F90708080607070000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000405F804F608F5F7090
60709F7080A0808FAF8F90BF8F90BF9F9FC09FA0C0A0AFCFA0A0CFB0AFD0B0AFD0BFB0DFBFB0D0\
C0B0D0C0B0CFCFB0CFC0B0BFCFB0BFCFB0B0D0B0B0CFAFAF4F3F40503F40604F50
6F4F507F505F80506090606F9F5F6FA06F70A06F70AF707FAF707FB07F80B08080BF8F8FBF8F8F\
C09090CF9090D09F9FD09F9FDFA0A0D0AFA0DFB0A0DFB0A0DFBFA0DFBFA0E0C0AF
DFC0A0E0CFAFE0CFAFE0CFAFE0CFAFEFD0B0E0CFAFEFD0B0E0D0B0EFD0B0E0D0B0E0D0B0DFCFAF\
E0CFAFDFCFAFDFCFAFD0CFAFD0CFAFD0CFAFD0CFAFCFCFA0CFCFAFCFCFA0C0CFA0
BFC09FB0C09FAFBF90AFBF909FB08F9FB09090AF8F8FAF8F80A08F508070000000000000000000\
0000000000000000000000000000000000000000000000000000000000004F405F
4F40705050805F5F80606F90606F906F709F6F7F9F6F80A06F80A06F8FA06F8FA07090AF7090A0\
7090A070909F7F9F9F709F9F7F9F9F7F9F907F9F907F90807F908070907F709070
6F806050603F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000403F4F50405F6F5F707F6F80807F908F7F9F
908FAF9F90B0A09FC0A0A0CFAFA0CFAFA0C0AFAFC0AFAFC0B0B0C0AFAFC0B0B0C0AFB0C0B0B0C0\
AFAFBFAFAFBFA0AFB0A0AFB09FA0AF90A0AF80909F7F8F8F6F7F7F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000004050704F608F
5F709060709F7080AF808FAF8F90BF909FBF9FA0C09FA0C0AFAFCFAFAFCFB0B0D0B0B0D0BFBFE0\
BFB0DFC0BFD0C0B0CFCFBFCFCFB0C0CFBFC0CFB0B05040405F404F6F4F50704F5F
805F608F5F6F9F60709F6070AF707FAF707FB07F80B07F80BF808FBF808FC08F90C08F90CF9F9F\
CF9F9FD0A0A0D0A0A0DFAFAFDFAFA0DFB0AFDFB0A0E0BFAFE0BFAFE0C0AFE0C0AF
EFD0B0E0CFAFEFD0B0E0D0B0EFD0B0EFD0B0EFDFBFEFD0B0EFDFBFEFD0B0EFD0B0E0CFAFE0D0B0\
DFCFAFDFD0B0D0CFAFDFD0B0D0CFAFD0D0AFCFCFAFD0D0AFCFCFAFC0CFA0BFC09F
B0C0A0AFBF9FA0BF909FB0909FB0908FAF8F8FAF906F907F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000403F40504F6F
5F50805F5F80606090606F906F7FA06F7FA07080A07080A0708FAF708FAF7F90AF7F90AF7F9FAF\
7F9FA07F9FA07F9F9F80A0A07FA09F80A09F7F9F90809F8F7F9F807F907F403F4F
607F5F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000403F505F506F6F5F7F80708F8070908F80A0
908FAF9F90BF9F9FC0AFA0CFAFA0CFAFAFC0AFA0C0AFAFC0AFAFC0B0B0C0AFAFBFAFB0C0AFAFBF\
AFAFBFA0A0B0A0AFB09FA0AF90A0AF8F909F7F8F8F6F7F7F5F706F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F4F60
4F5F805F6F9060709F7080A07F8FAF8F90B08F90B09F9FC09F9FC0A0AFCFA0A0CFB0AFD0B0B0D0\
BFBFDFBFB0D0C0B0D0C0B0CFCFBFCFC0B0C0503F405F404F6F4F50704F5F805060
8F5F6F9F60709F6070A06F7FAF707FB07F80B07F80BF808FBF808FC08F90C09090CF9F9FCF9F9F\
D0A0A0D0A0A0DFAFAFDFAFA0DFB0AFDFB0AFE0BFAFDFC0AFE0CFB0E0CFAFE0D0B0
E0CFB0EFD0B0E0D0B0EFD0B0EFD0B0EFDFBFEFDFB0EFDFBFE0D0B0E0D0B0DFCFAFDFCFB0DFCFAF\
DFD0B0D0CFAFD0CFAFCFCFAFD0D0AFCFCFAFCFCFAFC0CFA0BFCFA0B0BF9FAFBF9F
A0B090A0B09090AF8F90B0908FAF8F7FA08F000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F303F403F4F504F70
5F507F60608F60608F6F709F6F7FA07080A06F7FA0708FAF708FA07F90AF7F90AF7F9FAF7F9FAF\
809FAF7F9FA080A0A07F9F9F80A09F80A09F80A0907F9F8F7F9F80709070708F6F
4F603F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F3040504F5F60506F706F807F708F8F809F8F80A0
9F90B09F90BFAFA0CFAFA0CFAFAFCFAFAFCFB0AFCFAFAFC0B0B0C0AFB0C0B0B0C0AFAFBFAFB0BF\
A0AFB0A0AFB09FA0AF9FA0AF8F9FA0809090708080607070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F405F405F805F6F9060709F7080A07F8FAF8F90BF8F90BF9F9FC09F9FC0A0AFCFAFAFCFB0B0D0\
B0B0D0BFBFE0BFB0DFC0BFDFC0B0CF50404F5F404F6F4F5F704F5F805F6F8F5F6F
9F60709F607FAF7080AF7080B07F8FB07F8FBF8F90BF8F90C0909FC0909FD09FA0D09FA0DFA0AF\
D0AFA0DFB0AFDFB0AFDFBFAFDFBFAFE0C0B0E0C0AFE0CFB0E0CFB0EFD0B0E0D0B0
EFDFBFE0D0B0EFDFBFEFDFBFEFE0BFEFDFBFEFDFBFE0D0B0E0D0B0DFCFB0E0D0B0DFCFB0DFD0B0\
D0CFAFD0D0B0CFCFAFD0D0B0CFCFAFCFD0AFBFCFA0BFC0A0AFBF9FAFBF9FA0B090
9FB09090AF8F90B09080AF8F4F7F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000403F405040605F4F7F
605F8060608F6F709F6F709F7080A07080A07F8FAF708FAF7F90AF7F90AF809FB0809FAF80A0AF\
809FAF80A0AF80A0A08FA0A080A09F8FAFA0809F90809F8F7F907F7F907F60705F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000403F50504F606F5F7F706F808070908F7F9F908FAF
908FB0A09FC0A0A0CFAFAFCFAFA0C0AFAFCFAFAFC0B0B0C0AFAFBFB0B0C0AFAFBFAFB0BFA0AFB0\
A0AFB09FA0AF9FA0AF909FA08090907080806F7F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F3F4F405F80506F905F7090707FA07F80AF8F90B08F90B0909FBF9F9FC0A0A0CFA0A0CF\
AFAFD0B0AFD0BFB0DF40303F50404F5F40506F4F5F704F5F805F6F8F5F6F9F607F
9F607FA06F80A07080B07F8FB07F8FBF8090BF8F90C0909FC0909FCF9FA0D0A0A0DFAFAFD0AFAF\
DFB0AFDFB0AFDFBFB0DFBFAFE0C0B0DFC0B0E0CFB0E0CFB0E0D0B0E0D0B0EFDFBF
E0DFB0EFDFBFEFDFBFEFE0BFE0DFBFE0DFBFE0D0B0E0D0B0DFCFB0DFCFB0D0CFAFD0D0B0D0CFAF\
D0CFB0CFCFAFCFD0B0C0CFAFC0CFAFBFC0A0B0C0A0AFBF9FA0BF9F9FB09090B090
8FAF8F8FAF8F5F8070000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F30304F405050406F5F5080
605F806F6F906F709F707FA0707FA07F80AF7080A07F90AF7F8FAF809FB0809FAF809FB0809FAF\
80A0AF809FA080A0A080A09F8FA0A0809F90809F8F7F90807F907F708F6F506040
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F405F60506F6F607F807090807F9F9080A0908FAF
A09FBFA09FC0AFAFD0AFA0CFB0AFCFAFAFCFB0B0CFAFAFC0B0B0C0AFB0BFAFB0BFA0AFB0A0AFB0\
A0AFAF9FAFAF90A0A08F9F9F7F8F8F70807F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000405F80506F9060709F7080A07F80AF8F90B08F90BF9F9FC09F9FC0A0AFCF\
A0AFCFB0B0D04F3F3F5F404F5F40506F4F5F7050608F5F6F905F709F607F9F6F7F
A07080AF7080B07F8FB0808FBF8F9FC08F9FCF90A0CF9FA0D0A0AFD0A0AFDFAFB0DFAFAFDFBFB0\
DFBFB0E0C0B0DFC0B0E0CFBFE0CFB0E0D0BFE0D0B0EFDFBFE0DFBFEFE0BFEFDFBF
EFE0C0EFE0BFEFE0C0E0DFBFEFDFBFE0D0BFE0D0BFDFCFB0DFD0BFD0CFB0DFD0B0D0CFB0D0D0B0\
CFCFB0CFD0B0C0CFAFC0D0AFBFC0A0B0C0A0A0BF9FA0BF9F9FB09090B0908FAF8F
709F80000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F404F5040605F507F605F80
6F6F906F6F90707FA07F7FA07F8FAF7F8FAF8090AF7F90AF809FB0809FB08FA0B0809FAF8FA0B0\
80A0AF8FA0AF80A0A08FAFA080A09F80A0907F9F8F7F9F80709070607F60000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F304F504F6060506F706F807F70908F7F9F8F80A09F90B0
9F90BFA0A0CFA0A0CFAFAFCFAFAFC0B0AFCFAFAFC0B0B0C0AFB0C0B0B0C0A0AFB0A0AFB09FA0AF\
9FAFAF90A0A0909F9F7F8F8F70808060706F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000040508050608F5F6F906F7FA07080A08F90B08F90B0909FBF9F9FC0\
A0A0CF4F3F3F50404F5F40506F4F60705060805F6F905F709F607F9F6F7FA07080
A07080B07F8FB0808FBF8F90BF8F9FC090A0CF9FA0D0A0AFD0A0AFDFAFB0D0AFAFDFB0B0DFBFB0\
E0C0B0DFC0B0E0CFBFE0CFB0E0D0BFE0D0B0EFDFBFE0DFBFEFE0BFE0DFBFEFE0C0
EFE0BFEFE0C0E0DFBFE0DFBFDFD0BFE0D0BFDFCFB0DFD0B0D0CFB0D0D0B0CFCFB0CFD0B0CFCFAF\
CFD0B0C0CFAFBFCFAFB0C0A0AFC0A09FB09F9FB09F90AF908FB0907FA08F407060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F3F3F4F40505F4F705F507F6F5F8F
6F608F70709F707FA07F80AF7F80AF808FAF7F8FAF8090B0809FAF8FA0B0809FAF80A0AF809FAF\
8FA0AF80A0A08FA0A080A0A080A09F7F9F8F7F9F8070907F7090704F6040000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004F4050504F606F607F7F6F8F807F9F8F7F9F908FAF9F90B0
A09FC0A0A0CFAFAFCFAFAFCFB0AFCFAFAFC0B0B0CFAFB0C0B0B0C0AFAFBFAFB0BFA0AFB0A0AFAF\
9FA0A09FA0A08090907F8F80607070000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F508050608F5F6F906F7FA07080A08F90B08F90B0909FBF\
4F3F405F4050604F5F7050607F50608F5F709060709F6F7F9F6F80A0708FAF708F
B08090B08090BF8F9FBF8F9FC09FA0CF9FA0D0A0AFD0A0AFDFB0B0DFB0B0DFBFBFDFBFB0E0C0BF\
DFC0B0E0CFBFE0CFBFE0D0BFE0D0BFEFE0C0EFE0BFEFE0C0EFE0C0EFEFC0EFE0C0
EFE0C0E0DFBFE0E0C0DFDFBFDFD0BFDFD0B0DFD0BFD0CFB0D0D0B0CFCFB0D0D0B0CFCFB0CFD0B0\
C0CFAFBFCFAFB0C0A0AFBFA09FB09F9FB09F8FAF908FAF90507F6F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F404F5F4F6F5F507F6F5F8F6F608F
70709F70709F8080AF8080AF808FAF808FAF8090B0809FB08FA0B08FA0B08FA0B08FA0AF8FA0AF\
80A0A08FAFAF8FA0A08FAFA0809F90809F8F7F90807F907F6F8060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F304F4F405F60506F6F607F7F7090807F908F80A0908FAF9F90BF
A09FC0AFA0CFAFA0CFAFAFCFAFAFC0AFAFC0AFAFC0B0B0C0AFAFBFAFB0BFA0AFB0A0AFB09FA0A0\
9FA0A08F90907F8F806F7F70000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F50704F608F506F906F7F9F7080A0808FB04F3F40\
5F405060405F7050607F50608F5F70905F709F6F809F6F80A0708FA0708FAF8090
B08090BF8F9FBF8F9FC09FA0C09FA0D0A0AFD0A0AFDFAFB0DFAFB0DFBFBFDFBFB0E0C0BFDFC0B0\
E0CFBFE0D0BFE0DFC0E0DFBFEFE0C0E0E0BFEFEFC0EFE0C0EFEFC0EFE0C0EFE0C0
E0DFBFE0DFBFDFD0BFDFD0BFD0CFB0D0D0BFD0CFB0D0D0B0CFCFB0CFD0B0C0CFB0CFD0B0BFCFAF\
B0C0A0AFBF9FA0BF9F90B0908FB09080AF8F6F8F7F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F303F504F5F5F4F70605F806F5F80706F90
7070907F7FA08080AF808FAF808FAF8090B08090AF8F9FB08F9FB08FA0B08FA0AF8FA0AF80A0A0\
8FA0AF8FA0A08FAFA080A09F809F907F90807F907F708F6F506F4F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000403F505F4F6F605F707F708F7F70908F80A09080AF9F90B09F9FBF
AFA0CFAFA0CFAFAFCFAFAFCFB0B0CFAFAFC0B0B0C0B0B0C0B0B0C0AFAFBFAFAFB0A0A0AFA0AFAF\
90A0A08F909070807F6F706F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F4F6F4F608F5F6F904F3F404F3F4F5F4F50\
604F5F70506F7F506F8F607F90607F9F6F809F6F80A07F8FA07F8FB08090B08090
BF90A0BF90A0C09FAFCF9FAFD0AFB0D0AFB0DFB0BFDFB0B0E0BFBFDFBFBFE0C0BFE0CFBFE0D0C0\
E0D0C0EFDFC0E0DFC0EFE0C0EFE0C0EFEFCFEFEFC0F0F0CFEFEFC0EFEFC0E0E0C0
E0E0C0DFDFBFDFDFBFD0D0BFDFD0BFD0D0B0D0D0BFCFCFB0CFD0BFCFD0B0C0D0B0BFCFAFB0C0AF\
A0BFA0A0BF9F90B0908FB09080AF8F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F40505F4F6F605F806F5F80706F90707090
807FA08080AF8F8FB0808FAF8F90B08F90B08F9FB08F9FB090AFBF90A0B090AFB08FA0AF8FAFAF\
8FA0A08FAFAF8FAFA08FA09F809F8F7F9F80708F70708F6F40503F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F404F405F5F4F6F6F60807F6F8F807F9F8F80A0908FAF9090B0A09FC0
A0A0CFAFAFCFAFA0C0AFAFCFAFAFC0B0B0C0AFAFC0B0B0C0AFAFBFAFAFBFA0A0AFA0AFAF9FA0A0\
90A0A0808F80707F7F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F405F4F3F404F3F4F5F4F50604F5F\
70506F7F506F8F5F7F90607F9F6F809F6F80A0708FA07F90AF809FAF8090BF8FA0
BF90A0C09FAFC09FAFD0A0B0D0AFB0DFB0BFD0AFB0DFBFBFDFBFBFE0C0C0DFC0BFE0CFC0E0D0C0\
EFDFC0E0DFC0EFE0C0EFEFC0EFEFCFEFEFC0EFEFCFE0E0C0E0E0C0DFDFC0E0DFC0
DFDFBFDFDFBFD0D0BFD0D0BFCFCFB0D0D0BFCFCFB0CFD0B0C0CFB0C0CFB0B0C0A0AFC0A0A0BF9F\
9FBF9F8FAF9080AF904F7060000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000403F4050405F60507F605F8070608F706F907F7F9F
8080A08F8FAF808FAF8F90B08F90B0909FB08F9FB090A0BF90A0B090AFB08FA0AF8FA0AF8FA0AF\
8FAFAF8FA0A08FAFA0809F90809F8F708F7F708F705F7050000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000403F504F4060605F70706080807090807F9F908FAF908FB0A09FC0A09FC0
AFAFCFAFA0CFB0AFCFAFAFC0B0B0CFAFAFC0B0B0C0B0B0BFAFB0BFA0AFB0A0AFB09FA0AF9FA0A0\
8F9090808F806F7F6F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004F3F4050404F604F5F6F4F6070506F\
7F507090607F90607F9F6F809F6F8FA07F90A07F90AF809FB0809FBF90A0BF90A0
C09FAFC09FAFD0AFB0D0AFB0DFB0BFD0B0BFDFBFC0DFBFBFDFC0C0DFC0C0E0D0C0E0D0C090807F\
E0E0C0EFEFCFEFEFCFEFF0CFEFEFCFEFF0CFE0EFC0E0EFCFDFE0C0E0E0C0DFDFBF
DFDFC0D0D0BFD0D0BFCFD0BFCFD0BFCFD0B0CFD0BFC0D0B0BFCFB0B0C0A0AFC0A09FBF9F90B09F\
8FAF9080AF90000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F405060507060507F70608F706F907F7F9F807FA0
8F8FAF8F8FAF8F90B08F90B0909FB0909FB090A0BF90A0BF9FAFBF90A0B090AFB090A0AF90AFAF\
8FA0A090AFA08FA09F80A0907F90807F907F708F6F303030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000403F50504F6F605F707F6F8F7F70908F80A08F80AF9F90B09F9FBFAFA0CF
A0A0CFAFAFCFAFAFC0B0AFCFAFAFC0B0B0C0AFB0BFB0B0BFAFAFB0AFAFB0A0A0AF9FA0AF909F9F\
8F909070807F60706F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F404050404F604F5F6F4F6070506F7F5070\
8F607F90607F9F6F8F9F6F8FA07F90A07F90AF809FAF809FBF8FA0BF90A0C09FAF
C09FAFCFA0B0D0AFB0D0B0BFD0AFBFDFB0C0D0BFBFDFC0C0DFC0C0DFCFC0DFD0C0E0DFCFE0DFC0\
EFEFCFEFEFCFEFF0CFEFEFCFEFEFCFE0EFC0E0EFCFDFE0C0DFE0C0D0DFBFDFDFC0
D0D0BFD0D0BFCFCFBFCFD0BFC0CFB0C0D0BFBFCFB0BFCFAFAFBFA0A0BFA090B09F90B09F80AF8F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000403F405F4F6060507F6F608070608F7F70907F7F9F8F80AF
808FAF8F90AF8F90AF909FB0909FB090A0BF90A0B09FAFBF90A0B090AFB090A0AF90AFAF8FA0A0\
90AFA08FAFA08FA09F7F9F8F7F9080708F70607F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F304F4F40605F506F706F807F7090807F9F8F80A09090B09F90BFA0A0CFA0A0CF
AFAFCFAFAFCFB0AFCFAFAFC0B0B0CFB0B0C0B0B0C0AFB0BFAFB0BFA0AFB0A0AFAF9FA0A09FA0A0\
808F8F70807F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000403F3F4F404F504050604F606F4F607F5F70805F70906080\
9060809F6F8F9F708FA07F90A07F9FAF809FB08FA0BF90AFBF90AFC09FB0C0A0B0
D0AFBFD0AFBFD0B0C0D0B0BFDFBFC0DFBFC0DFC0C0DFC0C0E0D0CFDFD0CFE0DFCFE0DFCFEFEFD0\
E0EFCFF0FFD0EFF0CFEFF0D0E0EFCFE0EFCFDFE0C0DFE0CFDFDFC0DFDFC0D0D0BF
D0D0C0CFD0BFCFD0BFC0D0BFCFD0BFBFCFB0B0CFAFAFBFA0A0BFA090B09F8FB09F80AF8F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F3030504F5F6050706F5F8070608F7F709F807F9F8F80AF8F8FAF
9090B08F90B0909FB0909FB09FA0BF9FA0BF9FAFBF9FAFBF9FB0BF9FAFB09FAFB090AFAF90AFAF\
90AFA090AFA080A090809F8F7F907F708F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
403F504F4060605F7F706F8F7F7090807F9F8F80AF908FB09F9FC0A09FC0AFA0CF
AFA0CFAFAFCFAFAFC0B0B0C0AFAFC0B0B0C0AFB0BFAFB0BFAFAFB0A0AFB09FA0A09FA0A08F9090\
808F806F706F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000403F3F50404F504050604F606F4F607F5F70805F708F6080906080\
9F6F8F9F708FA07F90A07F90AF80A0AF8FA0B090AFBF90AFC09FB0C0A0B0CFAFBF
CFAFBFD0B0C0D0B0BFD0B0C0D0BFC0DFC0CFD0C0C0DFCFCFDFD0CFE0DFCFDFDFCFE0E0CFE0E0CF\
EFF0D0E0F0CFE0F0CFE0EFCFE0EFCFDFE0C0DFE0C0D0DFC0D0DFC0CFD0BFCFD0BF
CFD0BFCFD0BFC0CFBFC0D0BFB0C0AFB0C0AFA0BF9F9FBF9F8FB0908FAF907FA08F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000403F4F5F4F606F5F806F5F807F6F907F70908080A08F80A08F8FAF
8F90AF909FB0909FB09FA0BF90A0BF9FAFBF9FAFBF9FAFBF90AFB09FAFB090AFAF9FAFAF90AFA0\
90AFA08FA09F80A0907F90807F907F4F5F40000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000030304F\
403F5F5F506F6F607F7F70907F70908F80A08F8FAF9F90BF9F9FC0AFAFD0AFA0CF
AFAFCFAFAFC0B0B0CFB0B0C0B0B0C0B0B0C0B0B0C0AFAFB0AFAFB0A0A0AFA0AFAF9FA0A08F9090\
7F807F6F706F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000403F4050404F5F40506F50606F506F7F5F70805F7F906080906F809F708F\
9F7090A07F9FA0809FAF8FA0AF8FA0BF90AFBF90AFC0A0B0C0A0B0D0AFBFCFAFBF
D0B0C0D0B0C0D0BFCFD0BFC0DFC0CFDFC0CFDFD0CFDFD0CFE0DFCFDFDFCFE0E0D0E0EFD0EFF0D0\
E0F0CFE0F0D0E0EFCFE0EFCFDFE0C0DFE0CFD0DFC0D0DFC0CFD0BFD0D0C0CFD0BF
CFD0C0C0D0BFBFD0BFB0C0AFAFC0AF9FBF9F9FBF9F8FAF9080AF90000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000403F4050405F6F507F6F5F807F6F907F7090807FA08F80A09090B09090AF
909FB0909FB09FA0BF9FA0BF9FAFBF9FAFBF9FAFBF9FAFB09FAFB090AFB09FB0B090AFAF9FB0AF\
8FA09F8FA0907F9F807F907F607F5F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F2F4F\
4F4F605F5070706F80706F8F807F9F8F80A0908FB09090BFA09FC0A0A0CFAFAFCF
AFA0C0AFAFCFAFAFC0B0B0C0AFB0BFB0B0C0AFAFBFAFAFB0A0A0AFA0AFAF9FA0A0909F9F808F80\
707F70000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000403F405040505F4050604F606F506F7F5F7F805F7F8F60809060809F70909F7090\
A07F9FA07F9FAF8FA0AF8FA0B090AFBF90AFC0A0B0C0A0B0CFAFBFCFAFBFD0B0C0
CFB0C0D0B0C0D0BFC0DFC0CFD0C0CFDFCFCFDFCFCFDFDFD0DFDFCFE0E0D0E0E0D0E0F0D0E0EFCF\
E0EFD0DFEFCFDFEFCFD0E0C0D0E0CFD0DFC0D0DFC0CFD0BFCFD0C0C0D0BFC0D0C0
BFCFBFBFCFB0AFC0AFA0C0A09FB09F90B09F80AF9080AF90000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F3F3F4F3F5060506F6F507F70608F7F6F8F807F9F807FA0908FAF908FAF9090B0
9090B09F9FB0909FB09FAFBF9FAFBF9FAFBF9FAFB09FAFB090AFAF9FAFB090AFAF9FB0AF8FA09F\
8FA090809F8F7F9080708F70000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000403F50\
504F606F607F706F8F807F9F807FA0908FAF908FB0A09FC0A0A0CFAFAFCFAFA0CF
B0AFCFAFAFC0B0B0CFB0B0C0B0BFC0B0B0BFB0B0BFAFAFB0AFAFB0A0A0AFA0A0A08F9090808F80\
6F7F6F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F3F405040505F4F5F6F506F705070805F7F8F5F80906F8F906F8F9F70909F7090A0809F\
A0809FAF8FAFAF8FAFBF9FB0BF9FB0C0A0BFC0A0BFCFAFC0CFAFC0D0B0C0CFB0C0
D0BFCFD0BFCFDFC0CFD0C0CFDFD0D0DFD0D0DFDFD0DFDFD0E0E0D0E0EFD0E0F0DFE0EFD0E0F0D0\
DFEFCFDFEFCFD0E0CFDFE0CFD0E0C0D0DFCFCFD0C0CFDFC0C0D0C0C0D0C0BFCFBF
B0CFB0AFC0AFA0BFA090B09F90B09F80AF90000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F404F604F6F6F507F70608F7F6F8F807F9F807FA0908FAF9090AF9F90B09090B0
9FA0BF9FA0BF9FAFBF9FAFBFA0B0C09FAFBF9FAFBF9FAFB09FB0B09FAFAF9FB0AF90AFA08FAF9F\
809F8F809F80708F70506F4F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F3050\
504F6F605F7F7F70907F709F8F80A08F8FAF9F90BF9F9FC0AFA0CFA0A0CFAFAFCF
AFAFC0B0B0CFAFAFC0B0B0C0B0B0BFB0B0BFAFAFB0AFAFB0A0A0AFA0AFAF9F9F9F8F908F7F8070\
6F7060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
403F405040505F405F6F506F705070805F7F805F808F608F906F8F9F70909F7090A07F9FA0809F\
AF8FAFAF8FAFB090B0BF9FB0C0A0BFC0A0BFCFAFC0CFAFC0CFB0C0CFB0C0D0BFCF
D0BFCFD0C0D0D0C0CFDFCFD0D0CFD0DFDFD0DFDFD0DFE0D0DFE0D0E0EFDFDFEFD0DFEFD0DFEFCF\
DFEFCFD0E0CFD0E0CFCFDFC0D0DFCFCFD0C0CFD0C0C0D0BFC0D0C0B0CFB0AFC0B0
A0BFA09FBFA090B0908FAF907FA08F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000403F40\
5F4F60604F70705F8070608F807090807F9F8F80A0908FAF9F90B09090B09F9FB0
9F9FB09FA0BF9FA0BFA0AFBF9FAFBF9FAFBF9FAFB09FAFB090AFAF9FB0AF90AFA090AFA0809F90\
809F8F70907F607F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000403F5F\
504F6F6F60807F7090807FA08F80AF9090B09F90BFA0A0CFA0A0CFAFAFCFAFAFCF
B0AFCFAFAFC0B0B0C0B0B0C0B0B0C0AFB0BFAFB0B0A0AFAFA0AFAF9FA0A09FA09F808F807F8070\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000403F40
5040505F4F5F6F506F705070805F80806080906F8F906F8F9F70909F709FA080A0A080A0AF8FAF\
AF8FAFB09FB0BF9FB0C0A0BFC0A0BFCFAFC0CFAFC0CFB0CFCFB0CFD0BFCFD0BFCF
D0C0D0D0C0D0DFCFD0D0D0D0DFDFD0DFDFD0E0E0DFDFE0DFE0EFDFDFEFD0DFEFD0DFEFCFDFEFCF\
D0E0CFD0E0CFCFDFCFCFDFCFCFDFC0CFDFC0C0D0C0BFD0BFB0CFB0AFC0B0A0BFA0
9FBFA08FAF908FAF90000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000403F405F4F5F\
604F70705F8070608080709080709F9080A0908FAF9F90B09090B09F9FBF9F9FB0
A0A0BF9FA0BFA0B0C0A0AFBFA0B0BF9FAFB09FB0B09FAFB09FB0B090AFAF90AFAF8FA090809F8F\
7F907F70907F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F3050\
504F6F605F7F7F70907F7F9F8F80AF908FB09F9FC0A09FCFAFA0CFA0A0CFAFAFCF
AFAFC0B0B0C0B0B0C0B0B0C0AFB0BFAFB0BFAFAFAFAFAFAFA0A0A09FA0A08F908F808F806F7060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000403F40504050
5F405F6F506F705070805F80805F808F6F8F8F6F8F9F709F9F709FA080A0A080A0AF8FAFAF8FAF\
B090B0B09FB0C0A0BFC0A0BFC0AFC0C0AFC0CFB0CFCFB0C0CFBFCFCFBFCFD0C0D0
CFC0D0D0CFD0D0CFD0DFDFDFD0DFD0DFE0DFDFE0DFDFEFDFDFE0D0DFEFD0D0E0CFD0E0CFCFDFCF\
CFE0CFCFDFC0CFDFCFC0D0C0C0D0C0BFD0C0BFCFBFAFC0B0A0C0AF9FBFA090B09F
80AF9080AF90000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F3030504050604F6F\
6F5F7F705F807F6F8F8070908F7FA09080A09F90B09090AF9F9FB09F9FB0A0A0BF
9FA0B0A0AFBFA0AFBFA0B0BF9FAFB09FAFB09FAFAF9FB0B090AFAF9FB0AF8FA09F809F907F9080\
70907F4F604F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000403F5F\
504F6F6F60807F70908080A08F8FAF9090BF9F9FC0AFAFD0AFAFD0AFAFCFAFAFC0
B0B0CFB0B0C0B0B0C0B0B0BFB0B0BFAFAFB0AFAFB0A0AFAFA0AFAF9FA09F8F908F7F80706F7060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000403F404F40505F405F
6F506F70507080608080608F906F90906F909F7F9F9F7F9FA080A0A080A0AF8FAFAF90B0B09FBF\
BF9FBFC0A0C0C0A0C0CFB0CFC0AFCFCFB0CFCFB0CFD0BFD0CFBFD0D0C0D0D0C0D0
D0CFDFD0D0DFDFDFDFD0DFDFDFE0DFDFE0DFDFEFDFDFE0D0DFEFD0D0E0CFD0E0D0CFE0CFD0E0CF\
CFDFCFCFDFCFC0D0C0C0DFCFBFD0C0B0CFBFAFC0B0A0C0AF9FBFA090B09F80AF90
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005040505F4F60705F7F\
7060807F6F8F8070908F80A0908FAF9F90B09F90B0A09FBF9F9FB0A0A0BFA0A0BF
A0AFC0A0AFBFAFB0C0A0AFBFA0B0BF9FAFB09FB0B09FAFAF9FB0AF90AFA08FA0907F9F807F9080\
6F806F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000403F5F\
5F50706060807F7090807F9F8F8FAF908FB09F9FC0A0A0CFAFAFDFAFAFD0AFB0D0
AFAFCFB0B0C0B0B0BFB0B0BFAFAFB0AFB0B0A0AFAFA0AFAF9FA0A09F9F9F808F807F7F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000403F404F405050405F604F6F
6F5070805F808060808F6F908F6F909F7F9F9F7F9FA080A0A080A0AF8FB0AF90B0B09FBFB09FBF\
C0A0C0BFA0C0C0AFCFC0AFC0CFB0CFC0B0CFCFBFD0CFBFD0D0C0D0CFC0D0D0CFDF
D0CFD0D0DFDFD0DFDFDFE0DFDFE0DFDFEFDFD0E0D0D0EFD0D0E0CFD0E0D0CFDFCFCFDFCFC0DFC0\
CFDFCFC0D0C0C0D0C0BFD0C0B0CFBFA0BFAF9FBFAF90B09F8FB09F7FA090806F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F3F4F5F405F6F5070705F7F\
7F6F8F806F908F7F9F9080A09F90AF9090AF9F9FB09F9FB0A0A0BF9FA0B0A0AFBF
A0AFBFAFB0C0A0AFBFA0B0BF9FAFB09FB0B09FAFAF9FB0AF90AFA08FA09F7F9F8F7F9080708F70\
3F503F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000404060\
50506F6F6F8F7F70908080A08F8FAF9F90BF9F9FC0AFAFD0AFAFD0B0B0D0AFAFCF
B0B0CFB0B0C0B0B0C0B0B0BFB0B0BFAFAFB0AFAFAFA0A0A0A0A0A090908F808F80707F6F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000403F404F405050405F604F6F6F5070
7F5F8080608F8F6F90906F909F7F9F9F7FA0A080AFA08FAFAF90B0AF90B0B09FBFB09FBFC0AFC0\
BFAFC0C0B0CFC0AFCFCFB0D0C0B0CFCFBFD0CFBFD0CFC0DFCFC0D0D0CFDFD0CFDF
D0DFDFD0DFDFDFE0E0D0E0DFDFE0DFD0E0D0D0EFD0D0E0CFD0E0D0CFDFCFCFE0CFC0DFCFC0DFCF\
C0D0C0C0D0CFBFCFC0B0CFBFA0BFAF9FBFAF8FB09F8FB09F7FA090A08F608F704F
80603F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000403F4050405F6F5070705F7F7F6F8F\
806F8F8F7F9F9080A09F90AF9F90AFA09FB09F9FB0A0A0BFA0A0BFAFAFBFA0AFBF
AFB0C0A0B0BFAFB0BFA0B0B0A0B0B09FAFAF9FB0B09FAFAF90AFA0809F907F9F8F708F7F506F50\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000403F5F\
5F50706F60807F7F9F807F9F8F8FAF9090BF9F9FCFA0A0CFAFAFD0AFAFCFB0B0D0
AFAFC0B0B0C0B0B0C0B0B0C0AFAFB0AFAFB0A0A0A0A0A0A09F9F9F8F908F7F80706F7060000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F3F3F4F3F4F4F40505F4F606F50707F5F80
7F5F808F6F908F6F9090709F907F9FA080AFA080AFAF90B0AF90B0B09FBFB09FBFBFAFC0BFA0C0\
C0AFCFBFAFCFC0B0D0C0B0CFCFBFD0C0BFD0CFC0DFCFC0D0CFCFDFCFCFDFD0D0E0
D0D0DFD0E0E0D0E0DFD0E0DFCFDFD0D0E0D0CFE0CFCFE0D0C0DFCFCFDFCFC0D0C0C0DFCFBFD0C0\
BFD0C0B0CFBFAFCFB09FBFAF90BFA080AF9F80AF90BFA080BFA080A08F709F8060
806F40000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F303F4F4050604F6F6F507F7F60807F6F8F\
8F7F9F8F809F9F8FAF908FAF9F90B09F9FB0A0A0BFA0A0B070707FA0AFBFAFB0C0
A0B0BFA0B0BFA0AFB0A0B0B09FAFAF9FB0AF9FAFAF90AFA0809F90809F8F708F7F6F8F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F4060\
5F5070706F8F7F70908F80A08F8FAF9F90BF9F9FC0AFAFDFAFAFD0B0B0D0AFAFCF
B0B0CFB0B0C0B0B0C0B0B0BFB0B0BFAFAFB0AFAFAFA0A0A0A0A0A08F908F7F8070606F5F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000403F404F40504F405F604F6F6F50707F5F807F5F8F
8F6F908F6F9090709F907FA0A080AFA08FAFAF90B0AF90BFB0A0C0B0A0C0BFAFCFBFAFCFC0B0D0\
C0B0CFC0B0D0C0B0D0CFBFDFCFBFDFCFC0DFCFC0DFD0CFE0CFCFDFD0DFE0D0DFE0
DFE0E0D0E0E0D0E0DFD0E0D0D0E0DFCFE0D0CFE0D0CFDFCFCFDFCFC0DFCFC0DFCFBFD0C0BFD0CF\
B0CFBFAFC0B09FBFAF90BFA080AF9F7FAF90C0B09FC0AF90BFA08FBFA080A08F6F
9F805F806F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F3F4F60506F6F507F7F60807F6F8F8F7F9F\
8F809F9F8FAF9F90AFA09FB09F9FB0A0A0BFA0A0BFAFAFC0AFAFC0AFB0C0AFB0C0
AFBFC0A0B0BFA0B0BF9FB0B0A0B0B09FB0AF9FB0AF8FA09F8FA0907F908070907F3F503F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000404060\
5F5F7F6F60807F7F9F807FA0908FB09090BF9FA0CFA0A0D0AFAFD0AFAFCFB0B0CF
AFAFCFB0B0CFB0B0BFB0B0BFAFAFB0AFAFB0A0A0A0A0A0A09F9F9F8F908F707F6F6F6F5F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F3F3F4F3F4F4F3F505F4F60604F70705F807F5F80806090
806F9090709F90709F9F80AF9F80AFAF90B0AF90BFB09FC0B0A0C0BFAFCFBFA0CFBFAFCFBFAFCF\
C0B0D0C0B0D0C0BFDFC0BFD0CFC0DFC0C0DFCFCFE0CFCFDFD0D0E0CFD0E0D0E0E0
CFDFDFD0E0DFCFDFD0CFE0D0CFDFCFCFE0D0C0DFCFC0DFCFBFD0CFC0D0CFBFD0C0BFD0C0AFC0BF\
AFC0B090BFA08FB0A080AF907FA090CFBFA0CFB0A0C0AF90C0AF90B0A080B09F7F
9F8F60907F5F50403F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000403F405F4F606F50707F60807F60808070908F7F9F\
9F8FAF9F8FAF9F90B09F90B0A0A0BFA0A0B0AFAFBFAFAFBFAFB0C0AFAFBFAFB0BF
A0B0BFAFB0BFA0AFB0A0B0B09FAFAF9FB0AF90A09F8FA0907F908070907F506F50000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003030504F4060\
5F5070706F8F7F7F9F8F80AF8F8FAF9F9FC09F9FCFAFAFDFAFAFD0B0B0D0AFAFCF
B0B0CFB0B0C0B0B0C0B0B0BFB0B0BFAFAFAFAFAFAFA0A0A09FA09F8F8F807F8070000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000403F3F4F404F4F3F505F4F6F604F70705F807F5F80806090806F90
8F709F9070A09F80AF9F80AFA090B0AF90BFB0A0C0B0A0C0BFAFCFBFAFCFC0B0D0BFB0D0C0B0D0\
C0B0D0C0BFDFC0BFDFCFC0E0CFC0DFCFCFE0CFCFE0D0D0E0CFDFE0D0E0EFCFDFE0
D0E0E0CFDFDFCFE0DFCFDFD0CFE0D0C0DFCFC0DFCFBFD0CFBFD0CFBFD0C0BFD0C0AFC0BFA0C0B0\
90B0A08FB0A07FAF90D0CFBFD0C0B0D0C0B0CFBFA0CFB0A0C0AF90C0AF90B0A080
AF90709F805F907F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F303F5F4F5F6F50707060807F60808F70908F7F9F908FA0\
9F8FAFA09FB09F9FB0A0A0BFA0A0BFAFAFC0AFAFBFAFB0C0AFB0BFAFB0C0AFB0BF
AFB0BFA0B0B06F60709FB0AF9FB0AF90AFA090AF9F809F8F7F90806F8F70000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F30504F4060\
5F5F7F6F6F8F7F7F9F807FA08F8FB09090BFA0A0CFA0A0D0AFAFD0AFAFCFB0B0CF
AFAFC0B0B0C0B0B0BFB0B0BFAFAFB0AFAFAFA0A0A0A0A0A09F9F908F9080707F6F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F303F403F4F4F3F50504060604F6F70507F705F807F608F8060908F709F
8F709F907FAF9F80AFA08FB0A090B0AF9FC0B0A0C0BFAFCFB0A0CFBFAFD0BFAFD0C0B0D0BFB0D0\
C0BFDFC0BFDFC0C0E0C0C0DFCFCFE0CFCFE0CFD0EFCFD0E0CFDFE0CFDFDFCFDFE0
C0DFD0CFDFD0C0DFD0C0DFD0C0D0CFC0DFCFBFD0CFBFD0CFB0CFC0B0CFC0A0C0B0A0BFAF90B0A0\
8FAF9F7FA090CFC0C0CFC0BFD0CFBFD0C0B0D0C0AFC0B0A0C0B09FBFAF90BFA08F
AF9070A08F6F907F505F403F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000303030504050604F6F705F7F7F608080708F8F7F909080A09F8FAF\
9F90B09F90B0A09FBFA09FB0A0A0BFA0AFBFAFB0C0AFAFBFAFB0BFA0B0BFAFB0BF
A0B0B0A0B0B0A0B0B0A0B0B090AFA090AF9F809F8F7F9F806F8F70405F3F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3050504F6F\
605F7F7070907F7F9F8F8FAF8F8FB09F9FC09FA0CFAFAFDFAFAFD0B0B0D0AFAFCF
B0B0CFB0B0C0B0B0C0B0B0BFB0B0BFAFAFAFAFAFAFA0A09FA0A09F8F8F807F7F6F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F3F3F4F404F4F3F50504F6F604F70705080705F80806090806F908F709F8F70A0
9080AF9F80AFA08FB0A090BFAF9FC0B0A0C0BFAFCFB0AFCFBFB0D0BFB0D0C0BFDFBFB0DFC0C0E0\
C0BFDFCFC0E0C0C0E0CFCFEFCFCFE0CFD0EFCFDFEFD0DFEFCFDFE0CFDFE0C0DFDF
CFDFDFC0DFD0C0DFD0C0DFCFC0DFD0BFD0CFBFD0CFB0CFC0B0CFC0A0C0B09FBFAF8FB0A08FB09F\
7FA090C0C0CFCFC0C0D0CFC0D0C0BFDFCFBFD0C0B0D0C0AFCFBFA0C0B09FBFAF90
BFA080A090709F80608F704F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F404F604F60705F7F7F60808070908F7F909080A09F8FAFA090B0\
A090B0A09FBFA0A0BFAFAFC0AFAFBFAFB0C0AFB0C0AFB0C0AFB0BFAFB0BFA0B0B0
AFBFBFA0B0B0A0B0B09FAFA090AFA08FA090809F8F7F907F4F6F4F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F305F4F406F\
605F7F706F8F7F7F9F8080A09090B09090BFA0A0CFA0AFD0AFAFD0AFAFCFB0B0CF
AFB0C0B0B0C0B0B0BFB0B0BFAFAFB0AFAFAFA0A0A0A0A0A09F9F908F8F80707F6F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F3F3F\
403F4F4F3F505040605F4F6F6F507F7050807F608F7F6090806F9F80709F907FAF
9080AF9F8FB09F8FB0AF9FC0AF9FC0B0A0CFB0A0CFBFAFD0BFAFD0BFB0DFBFB0DFC0BFDFBFBFDF\
C0C0E0C0C0E0CFCFEFC0CFE0CFD0EFCFD0EFCFDFEFC0D0E0CFDFE0C0D0DFC0DFDF
C0DFD0C0DFD0BFD0CFBFD0CFBFD0CFBFD0CFB0CFC0AFCFBFA0BFB09FBFAF8FAF9F80AF9F70A08F\
BFBFCFBFBFC0C0C0C0CFC0C0D0C0C0CFC0BFD0D0BFCFC0AFCFC0AFC0B09FC0B09F
BFA08FAF9F7F9F8060907F5F50403F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F3F4F5F4F5F705F7F705F7F806F8F8F70909080A0908FA09F90AF9F90B0\
A09FB0A09FB0AFA0BFAFAFBFAFB0C0AFB0C0AFB0C0AFB0BFAFB0BFA0B0B0A0B0B0
A0B0B0A0B0B09FAFA090AFA08FA090809F8F70907F506F50000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F5F504F6F\
605F7F7F70907F7F9F8F8FAF8F8FB09F9FC0A0A0CFAFAFDFAFAFD0B0B0D0AFB0CF
B0B0CFB0B0C0BFBFC0B0B0BFB0B0B0AFAFAFAFAFAFA0A09F9FA09F8F8F7F7F7F6F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F3F403F4F\
4F3F505040605F4F6F6F507F7050807F60907F6090806F9F80709F907FAF9080AF
9F8FB09F8FBFAF9FC0AF9FC0B0A0CFB0A0CFBFAFD0B0AFD0BFB0DFBFB0DFC0C0E0BFBFE0C0C0E0\
C0C0E0CFCFEFC0CFEFCFD0EFCFD0EFCFDFEFC0D0E0CFDFE0C0D0DFC0DFDFC0DFD0
C0DFD0BFD0CFBFDFD0BFD0CFBFD0CFAFCFC0AFCFBF9FBFB09FBFAF8FAF9F80AF9FAFAFCFBFBFD0\
BFB0CFC0BFCFC0BFC0CFC0C0D0D0C0DFE0C0D0D0BFD0CFB0CFC0AFCFC0A0C0B09F
C0AF90B09F80AF90709F805F8F7F50000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004F3F4F5F4060705070705F7F806F8F8F709090809F9080A0A090B0A090B0A09FB0\
A09FB0AFAFBFAFAFBFAFB0C0AFB0C0B0B0C0AFB0BFAFB0BFA0B0BFAFB0BFA0B0B0
A0B0B09FB0AF9FAFA08FA090809F8F7F9080607F60000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F5F4F4F6F\
606080707090807FA08080A09090B09090BFA0A0CFA0AFD0AFAFD0AFAFCFB0B0CF
AFAFC0B0B0C0B0B0BFB0B0BFAFAFB0AFAFAFA0A0A0A0A0A09F9F908F8F7F707060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F303F403F4F403F50\
50406050406F6F507F6F5080705F8F7F6090806F9F806F9F8F7FA08F7FAF9F8FB0
9F8FB0A090C0A09FC0AFA0CFAFA0CFB0AFD0B0AFD0BFB0D0B0B0D0BFBFDFBFBFDFC0C0E0BFC0E0\
C0CFEFC0CFEFCFD0F0C0D0EFC0D0EFC0D0E0C0D0E0BFD0DFC0D0DFBFD0D0BFDFD0
BFD0CFBFD0CFB0CFCFB0D0CFAFC0C0A0C0BF9FBFAF90B0A080AF9F7FA0909FA0C0AFAFCFAFAFCF\
B0B0D0BFB0CFC0CFC0C0CFC0D0D0C0D0DFC0DFDFC0CFCFB0CFCFB0C0BFA0C0BFA0
BFAF90BFA08FAF907FA08F6F8F7F505F4F3F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000403F405F405F6F506F7050707F6F80806F8F907F9F90809FA090AF9F90AFA09FB0A09FB0\
AFA0BFA0A0BFAFAFC0AFAFC0B0B0C0AFB0BFAFB0BFA0B0B0A0B0BFA0B0B0A0B0B0
9FB0AF90AFA08FA090809F907F908060806F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F5F505070\
6060807F709F7F7FA08F8FB09090B09F9FC0A0A0CFAFAFDFAFAFD0B0B0D0AFB0CF
B0B0CFB0B0C0BFBFC0B0B0BFB0B0B0AFAFAFAFAFAFA0A09F9FA09F8F8F7F7F7F6F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003F3F3F403F4F4F3F50504060\
50406F60507F6F50807060907F6090806F9F806F9F8F7FAF8F7FAF908FB09F8FBF
A09FC0A09FC0AFA0CFAFA0CFB0AFD0B0AFD0B0B0DFB0B0D0BFBFE0BFBFDFC0C0E0BFC0E0C0CFF0\
C0CFEFCFD0F0C0D0EFCFD0EFC0D0E0C0D0E0BFD0DFC0DFDFBFD0D0BFDFD0BFD0CF
BFD0D0B0D0CFB0D0CFAFC0C0A0C0BF9FBFAF90B0A080AF9F7FAF908F90BF9FA0C0A0A0C0AFB0CF\
B0B0C0BFBFCFBFC0C0CFCFCFCFD0C0D0DFCFD0DFC0DFDFC0D0CFB0CFCFAFC0BFA0
C0BF9FBFAF90B0A080A090709F8F608F704F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4F3F405F405F6F506F705F70806F8080708F907F9F90809FA090AFA090AFA09FB0A09FB0AFA0BF\
AFA0BFAFAFC0AFB0C0B0BFCFAFB0C0AFB0C0A0B0BFAFB0BFA0B0B0A0B0B0A0B0B0
9FB0AF8FA09F8FA0907F908070907F3F5030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000403F5F504F6F\
6F6080707090807FA08080AF9090BF9090C0A0A0D0A0AFD0AFAFD0AFAFCFB0B0CF
B0B0C0B0B0C0B0B0BFB0B0BFAFAFB0AFAFAFA0A0A0A0A09F9F9F908F8F7F707060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F303F403F4F403F5050406050406F\
604F7F6F5080705F8F7060907F6F9F7F6F9F8070A08F7FAF9080B0908FB09F90C0
A09FC0AFA0CFA0A0CFAFAFCFAFAFCFB0B0D0B0B0D0B0BFDFB0BFDFBFC0E0BFC0E0C0CFEFC0CFEF\
C0D0F0C0CFEFC0D0EFBFD0E0C0D0E0BFD0DFBFD0DFBFD0D0BFD0D0B0D0CFB0D0CF
B0CFCFB0CFCFA0C0BF9FC0B090B0AF8FB0A07FA09070A0907F8FAF8F90B08F90BF9FA0BFA0AFC0\
B0B0CFB0B0C0BFC0C0C0C0C0CFCFC0CFD0C0D0DFC0D0D0BFD0D0BFCFC0AFCFC0AF
C0B09FC0B09FB0A08FAF9F7F9F8F60907F5F5F4F3F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000403F40
5F405F6F4F6F7050707F6080806F8F907F9F90809F9F8FAFA090AFA09FB0A09FB0AFA0BFAFA0BF\
AFAFC0AFAFC0B0B0C0AFB0C0AFB0C0A0B0BFAFB0BFA0B0B0A0B0B09FB0AF9FB0AF
8FA09F8FA0907F90807F907F405F40000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000403F60505070\
6060807F7F9F807FA08F8FB09090BF9F9FCFA0A0D0AFAFDFAFAFD0B0B0D0AFB0CF
B0B0CFB0B0C0BFBFC0B0B0BFB0B0B0AFAFAFAFAFA0A0A09F9F9F9080807F7F7F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000403F4F403F5050406050406F604F7F\
6F5080705F8F7060907F6F9F7F6F9F807FAF8F7FAF9080B0908FB09F90C0A09FC0
A0A0CFA0A0CFAFAFD0AFAFCFB0B0D0B0B0D0B0BFDFB0BFDFBFC0E0BFC0E0C0CFEFBFCFEFC0D0F0\
C0CFEFC0D0EFBFD0E0BFD0E0BFCFDFBFD0DFBFD0D0BFD0D0B0D0CFB0D0D0B0CFCF
AFCFCFA0C0BF9FC0B090B0AF8FB0A07FA09050608F606F9F7080AF7F8FAF8F90B0909FBFA0AFC0\
AFAFC0B0BFCFB0BFC0C0C0CFC0CFC0CFD0CFD0D0C0DFDFC0D0D0BFD0D0B0CFC0AF
CFC0A0C0B09FC0B090B0A080AF907090805F8F7F50000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000403F405F405F
6F4F6F705070806080806F8F907F9F9080A0A08FAFA090AFA09FB0A09FB0AFA0BFAFA0BFAFAFC0\
AFAFC0B0BFCFB0B0C0B0B0C0AFB0BFAFB0BFA0B0B0A0B0B0A0B0B0A0B0AF90AF9F
8FA0907F9F8F7F90804F6F4F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000404060505070\
6F608F7070908080A08080AF9090BF9F9FC0A0A0D0A0AFD0AFAFD0AFAFCFB0B0CF
AFAFC0B0B0C0B0B0BFB0B0BFAFAFAFAFAFAFA0A0A0A0A09F90908F8F8F7F70705F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F3F40403F504F406050406F5F4F7F604F80\
6F5F8F6F5F8F7F609F7F6F9F8070A0807FAF8F80B08F80B09F90BF9F90C0A09FCF
A09FC0AFA0CFAFA0CFAFAFD0AFAFD0B0B0DFB0B0DFBFBFE0B0BFE0BFC0E0BFC0E0C0CFEFBFCFEF\
BFD0EFBFCFE0BFD0E0B0CFDFBFD0DFB0CFD0B0D0D0B0CFCFB0D0CFAFCFCFAFCFC0
9FBFBF9FBFB08FB0A080AF9F70A0903F30604F4F70606F906F7FA07F80AF808FAF909FBF9F9FBF\
A0AFC0AFAFC0B0BFC0B0BFC0C0C0C0C0CFC0CFD0C0D0D0C0D0DFC0CFD0B0CFCFB0
C0BFA0C0BFA0BFAF90BFAF8FAF907F9F8F6F907F505F4F3F000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F3F3F503F506F4F6F
6F50707F6080806F8F907F9090809F9F8FA09F8FAFA090B0A09FB0AFA0BFAFA0BFAFAFC0AFAFC0\
B0B0C0AFB0C0B0B0C0AFB0BFAFB0BFA0B0B0A0B0B0A0B0AF9FB0AF90A09F8FA090
7F908F7F9080507050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004040605F5F7F\
6F6F8F7F7F9F807FA08F8FB09090BF9FA0CFA0A0D0AFB0DFAFAFD0B0B0D0AFB0CF
B0B0CFB0B0C0BFBFC0B0B0B0B0B0B0AFAFAFAFAFA0A0A09F9F9F90808070707060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F3F40403F4F50406050406F5F4F7F604F806F5F8F\
6F5F90706F9F7F6F9F8070AF807FAF8F80B08F80B09F90C09F90C0A09FCFA09FCF
AFA0D0A0A0CFAFAFD0AFAFD0B0B0DFB0B0DFB0BFE0B0BFE0BFCFEFBFCFEFC0CFEFBFCFEFC0D0EF\
BFCFE0BFD0E0B0CFDFBFD0E0B0CFDFB0D0DFB0CFD0B0D0D0AFCFCFAFCFC09FBFBF
9FBFB08FB0A080AF9F70A090000000000000404070505F806F7FA07080A0808FB08F90B09FA0BF\
A0A0C0AFB0CFAFB0C0BFBFCFBFC0C0C0CFCFCFCFC0D0DFCFD0DFC0D0DFBFCFCFB0
CFCFB0C0BFA0C0BF9FBFAF90B0A08FA090709F8F608F704F000000000000000000000000000000\
000000000000000000000000000000000000000000403F3F5040506F4F6F705070
806080806F8F907F9F90809FA08FAFA090AFA09FB0A09FB0AFA0BFAFA0BFB0AFC0AFAFC0B0B0CF\
B0B0C0B0BFC0AFB0BFAFBFC0A0B0BFAFB0BFA0B0B0A0BFB090AFA08FA09F809F8F
7F9F80607F6F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000302F4F404060505070\
6F6F8F70709F8080A08F8FAF9090BF9F9FC0A0AFD0A0AFD0AFB0D0AFAFCFB0B0CF
B0B0C0B0B0C0B0B0BFB0B0BFAFAFAFAFAFAFA0A09FA0A09F90908F8F807070705F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F30404F405F50406F5F4F7F5F4F806F508F6F5F8F\
70609F70609F7F70A07F70AF8F80B08F80B0908FBF9090C09F9FCF9F9FC0A0A0CF
A0A0CFAFAFD0A0AFD0AFB0DFAFB0DFB0BFE0B0BFDFB0C0E0B0C0E0BFCFEFBFCFE0BFCFEFB0CFE0\
BFCFE0B0CFDFB0CFDFB0CFD0B0CFDFAFCFD0AFCFD0AFCFCFA0C0C090BFB090BFAF
80AFA080AF9F709F8F604040603F3F6F3F4060303F603F3F603030603F3F7080A0808FB08F90B0\
9FA0BFA0A0C0AFB0C0B0B0C0BFBFC0BFC0C0C0CFC0CFD0C0D0DFC0D0D0BFD0D0BF
CFCFAFCFC0AFC0B09FC0B09FB0A08FAF9F7F9F8F60907F5F5F4F3F000000000000000000000000\
0000000000000000000000000000000000003F303F503F506F4F6F6F50707F607F
806F80907F9090809F9F8FA09F8FAFA09FB0A09FB0AFA0BFAFA0BFB0AFC0AFAFBFB0B0C0B0B0C0\
B0BFC0AFB0BFAFB0BFA0B0B0AFB0B0A0B0B0A0B0B090AFA090A09F809F8F7F9080
6F8F70000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F305F4040605F5F7F\
6F6F8F7F7FA08080A08F8FB09090BF9FA0CFA0A0D0AFAFD0AFAFCFB0B0D0B0B0CF
B0B0CFB0B0C0BFBFC0B0B0B0B0B0B0AFAFA0AFAFA0A0A09F9F9F8F808070707060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F304040405F4F406F5F4F7F5F4F806F5F8F6F5F9070609F\
70609F7F70A07F70AF8080B08F80B0908FC09090C09F9FCF9F9FCFA0A0CFA0A0CF
A0AFD0A0AFD0AFB0DFAFB0DFB0BFE0AFBFDFB0C0E0B0C0E0BFCFEFB0CFE0BFCFEFB0CFE0B0CFE0\
B0CFDFB0CFDFAFCFDFB0CFDFAFCFD0AFD0D0A0CFCFA0C0C090BFB090BFAF80AFA0
7FAF9F80505080505F7F4F507F4F50704F4F7F4F5070404F704F4F70404070404F704040704040\
6F3F407040406F3F3F6F3F3F6F303F6F3F3FC0C0C0CFD0C0CFD0C0D0DFC0D0D0BF
D0D0B0CFC0AFCFC0A0C0B09FBFB090B0A080AF90709080608F7F50000000000000000000000000\
0000000000000000000000000000003F3F3F50404F6F4F6F7050707F6080806F80
907F9090809FA08FAF9F90AFA09FB0A09FB0AFA0BFAFA0BFB0AFC0B0AFC0B0B0CFB0B0CFB0BFCF\
AFB0C0B0BFC0AFB0BFAFB0BFA0B0B0A0BFB09FAFAF90AFA0809F90809F8F70907F
405F40000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F406F50507F\
6F6F9070709F8080AF808FB09090BF9F9FC0A0AFD0A0AFD0AFAFD0AFAFCFB0B0CF
AFAFC0B0B0C0B0B0BFB0B0BFAFAFAFAFAFAFA0A09FA0A09F909080808070706F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F303F3F3F4F403F5F504F705F4F8060508F605F8F6F609F6F609F\
7F6FA07F70A0807FB08080B08F8FBF908FC09F9FCF9090C09F9FCF9FA0CFA0AFD0
A0A0CFA0AFD0A0B0D0AFBFDFAFB0DFB0C0E0B0C0E0B0CFEFB0C0E0B0CFE0B0CFE0B0CFE0AFC0DF\
B0CFDFAFCFD0AFCFD0A0C0CFAFCFCFA0C0CF9FC0BF8FB0B08FB0AF7FAF9F7FA090
805F5F8F5F608F5F608F5F60805F5F8F5F6080505F805F5F80505F80505F7F505080505F7F4F50\
7F4F507F4F4F7F4F5070404F7F4F4F7040407F404F704040704040703F40704040
6F3F3F703F3F6F3F3F703F3F6F3030BFAF8FAF9F7FA08F6F8F7F50605040000000000000000000\
0000000000000000000000000000004F3F4F6F4F606F506F7F607F806F80907F90
90809F9F8FA09F8FAFA090B0A090B0AFA0BFAFA0BFB0AFC0B0AFC0B0B0C0B0B0C0B0B0C0AFB0C0\
AFB0C0AFB0BFAFB0BFA0B0B0A0B0B09FAFAF90AFA0809F907F9F8F708F7F4F604F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004040605F5F7F\
6F6F8F7F7FA08080A09090BF9090C0A0A0CFA0AFD0AFB0DFAFAFD0B0B0D0AFB0CF
B0B0CFB0B0C0BFBFC0B0B0B0B0B0B0AFAFA0AFAFA0A0A09F9F9F8F80807070705F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F3F4F403F5F5040705F4F8060508F605F906F609F6F609F7F70A0\
7F70AF807FB08080B08F8FC09090C0909FCF909FC09FA0CF9FA0CFA0AFD0A0AFD0
AFB0DFA0B0DFAFBFDFAFBFDFB0C0E0B0C0E0B0CFEFB0C0E0B0CFEFB0CFE0B0CFE0AFCFDFB0CFDF\
AFCFD0AFCFD0AFCFD0AFCFD0A0C0CF9FC0C090B0B08FB0AF7FAF9F7FA0908F606F
906F70906F70906F70906070906F70906F70906F7090606F9F6F7090606F906F6F90606F90606F\
9060609060608F5F609060608F5F5F8F5F608F505F8F5F5F8050508F505F805050
805050804F508050507F4F4F804F4F7F4F4F804F407F4F3F804F406F403F5F403F4F3F303F3F3F\
000000000000000000000000503F4F6F4F6F70506F7F607F806F80907F9090809F
A08FAFA090AFA09FB0A09FB0AFA0BFAFA0BFB0AFC0B0AFC0BFB0CFB0B0C0B0B0C0AFB0BFB0BFC0\
AFB0BFAFBFBFA0B0B0AFBFB0A0B0AF9FAFA08FA090809F8F7090805F705F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F604F4F70\
60608F7070908080A08F8FAF909FBF9F9FC0A0AFD0AFB0DFAFB0DFAFB0D0B0B0CF
B0B0C0B0B0C0B0B0BFB0B0B0AFAFAFAFAFAFA0A09FA0A09F90908F8F80706F6F50000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F3F403F3F4F4F40605040705F4F8060508F6F609F6F609F706FA07070A0\
7F7FB07F7FB08F8FBF8F8FC09090CF9090C09F9FCF9F9FCF9FA0D09FA0CFA0AFD0
A0AFD0AFB0DFA0B0DFAFBFE0AFBFE0B0C0EFAFC0E0B0C0E0AFC0E0B0CFE0AFC0DFAFCFDFAFC0D0\
AFCFD0A0C0CFA0CFCFA0C0C09FC0BF8FB0AF80B0A07FA09F70A090906F7F906F7F
906F7F9F707F906F7F9F707F906F7F9F707F906F709F707F906F709F707F906F709F70709F6F70\
9F70709F6F709F70709F6F709F70709F6F6F9F6F709F6F6F9F6F6F9060609F606F
906060906060905F5F9060608F5F5F905F5F8F5F508F5F508F504F8F504F804F40805040804F40\
804F407F403F6F403F5F3F3F4F3F3F3F30307F607F806F80907F90907F9F9F8FA0
9F8FAFA09FB0A09FB0AFA0BFAFA0BFB0AFC0B0AFC0B0B0C0B0B0C0B0B0C0AFB0BFB0B0BFAFB0BF\
AFBFBFA0B0B0A0B0B0A0B0AF9FAFA08FA090809F8F709080607F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F5F505070\
5F5F8070709F7F80A08F8FB08F90BF9FA0CFA0AFD0AFB0DFAFB0DFB0BFDFB0B0CF
B0B0CFB0B0C0BFBFC0B0B0BFB0B0B0AFAFA0AFAFA0A0A09FA09F908F80707F7F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F3F403F3F4F4040604F40705F4F805F508F605F906F609F706FA07070AF7F7FB0\
7F7FB08F8FBF8F8FC09090CF9090C09F9FCF909FCF9FA0D09FA0CFA0AFD0A0AFD0
A0B0DFA0B0DFAFBFE0AFBFE0B0C0EFAFC0E0B0CFE0AFC0E0AFCFE0AFC0DFAFCFDFA0C0D0AFCFD0\
A0C0CFA0CFD09FC0C09FC0BF8FB0AF8FB0AF7FA09F70A09090708F9F7F8F9F7F8F
5F607F9F70809F7F8F9F70809F7F8F9F7080A07F809F7080A07F809F7080A07F809F7080A07F80\
9F707FA07F80A0707FA07F7F9F707FA07F7FA0707FA07F7FA07070A07F7FA07070
A07F7FA07070AF7F7FA07070A070709F6F6F9F6F6F9F6F609F6F6090605F90605F905F50905F50\
8F5F50905F508F504F8F504F8050408F5040804F3F804F3F6F403F60403F4F3F3F
403F3FA09FB0AFA0BFAFA0BFB0AFC0B0AFC0BFB0CFB0B0C0B0B0C0AFB0BFB0BFC0AFB0BFAFBFBF\
AFB0BFAFBFBF9FB0AF9FB0AF8FA09F8FA0907F90806F8F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000030304F40406F\
5F5F806F6F907F80A08080AF8F90BF909FC0A0AFD0A0AFD0AFB0DFAFB0D0B0B0D0
AFB0C0B0B0C0B0B0BFBFB0BFAFAFAFAFAFAFA0A09FA0A09F9F9F8F908F7F7F7060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F3040403F50403F6050407F504F80605F90605F906F6FA0706FA07F7FB07F7FB0\
8080BF808FBF8F90C08F90C0909FCF909FCF9FA0D090A0CF9FAFD09FAFD0A0B0DF
A0B0DFA0BFE0A0BFE0AFC0E0A0BFE0AFC0E0AFC0DFAFC0DFA0C0DFAFC0DFA0C0D0A0C0D09FC0CF\
A0C0CF9FC0C09FBFBF8FB0AF80AFA070A090709F8F9F7F909F7F909F7F90A08090
9F7F8F9F7F909F7F8FA080909F7F8FA07F8F9F7F8FA0808F9F7F8FA07F8FA07F8FA07F8F9F7F80\
A07F8FA07F80A07F8FA07F80A07F80A07F80A07F80A07F80A07F80A07F80AF7F80
A07F7FAF7F80A07F7FAF7F80A07F7FAF7F7FA07070AF7F7FA07070A070709F6F6FA0706F9F6F60\
9F6F6090605F9F605F905F5090605F905F50905F508F504F8F5F4F8F50408F5040
804F3F804F40804F3F704F3F5F3F30503F3F3F3030B0B0C0AFB0BFAFB0BFAFB0BFAFBFBFAFB0B0\
A0B0B09FAFAF9FAFA08FA09F8FA0907F908070907F40604F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F4F6F\
5F5F7F6F6F907F7F9F8F8FAF8F90B09F9FC09FA0CFAFB0DFAFB0DFB0B0DFAFB0D0
B0BFD0B0B0C0BFBFC0B0B0BFB0B0BFAFAFAFAFAFA0A0A09FA0A09090908080806F6F6F50000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F3F403F3F50403F60504F7F504F805F508F605F906F6FA06F6FA07F7FB07F7FB08080BF\
808FC08F90C08F90C0909FCF909FCF90A0D090A0CF9FAFD09FAFD0A0B0DFA0B0DF
AFBFE0A0BFE0AFC0E0A0BFE0AFC0E0AFC0DFAFC0E0A0C0DFAFCFDFA0C0D0A0C0D09FC0CFA0CFCF\
9FC0C09FBFBF8FB0AF80B0A07FA090709F8F9F809FA0809F9F809FA0809F9F809F
A08FA0A0809FA08F9FA0809FAF8F9FA0809FAF8F9FA08090AF8F9FA08090AF8F9FA08090AF8F9F\
AF8090AF8F90AF8090AF8F90AF8090AF8F90AF808FAF8F90AF808FAF8F90AF808F
B08F90AF808FB08F8FAF808FB08F90B0808FB08F8FB08080B08F8FB08080B08080AF7F7FAF807F\
AF7F70AF7F70A0706FA07070A0706FA0706F9F6F60A06F609F605F9F6F5F906050
9F605F905F50905F508F5F4F905F4F8F50408F5040804F3F8F5040704F3F604040503F3F4F3F3F\
9FB0AF9FAFA090A09F8FA09F809F8F6F8F7040604F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F4060\
50507F60608F7F7F9F8080A08F90B09090BF9FA0CFA0AFD0AFB0DFAFB0D0B0B0D0
AFB0CFB0B0CFB0B0BFB0B0BFB0B0B0B0AFAFAFA0A0AFA09FA09F909F908F807F6F706F50000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F303F3F3F4F3F30504F406F4F407F504F805F508F60609F6F609F7070AF707FB07F80BF8080BF\
8F8FC0808FC08F90CF8F90CF90A0CF909FCF9FA0D09FAFD09FB0DF9FAFD0A0B0DF
A0BFDFA0BFE0A0BFDFA0C0E0A0BFDFA0C0DFA0BFD0A0C0DFA0C0D0A0C0D09FC0CF9FC0CF90BFC0\
90BFBF8FB0AF80AFA070A0906F9F8F90809F9F80A09F809FA08FA09F809FA08FA0
A08FA0A08FA0A08FA0AF8FA0A08FA0AF90A0A08FA0AF8FA0AF8FA0AF90A0AF8F9FAF8FA0AF8F9F\
AF90A0AF8F9FAF8F9FAF8F9FB0909FAF8F9FB08F9FAF8F90B0909FAF8F90B08F9F
B08F90B08F9FB08F90B08F90B08F90B09090B08F8FB08F90B08F8FB08F8FB08F8FB08F8FB08F80\
B08F8FB08080B08080AF807FB0807FAF7F70AF7F70AF7F70AF7F70A0706FA0706F
A06F60A070609F6F5F9F6F5F9F605F9F6F5F906050906050905F4F905F4F905F4F905F4F805040\
7F50406F403F60403F503F3F4F3F3F40604F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000404060\
4F5070606F8F70709F8080AF808FAF909FBF9F9FC0A0AFD0AFB0DFB0BFDFAFB0D0
B0B0D0B0B0C0BFBFC0B0B0BFBFB0BFB0AFAFB0AFAFAFA09FA0A09F9F9F8F8F8F7F70705F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F303040303F
3F3F4F3F30504F406F4F4070504F80504F8F605F9060609F6F6FAF7070AF7F80BF8080C0808FC0\
808FC08F9FCF8F9FCF909FCF909FCF9FAFD090AFD09FAFDF9FAFDFA0BFE0A0BFDF
A0BFE0A0BFDFA0C0E0A0BFDFA0C0DFA0BFD0A0C0DF9FC0D0A0C0D09FC0CF9FC0CF90BFC090BFBF\
8FB0AF80AFA070A0909F8FA09080A09F8FA09F8FA09F8FAF9F8FA0A08FAFA08FAF
A090AFA08FA0AF90AFA090AFAF90AFA090AFAF90AFAF90AFAF90AFAF90AFB09FAFAF90AFB09FAF\
B090A0B09FAFB090A0B09FAFB090A0B09FAFB090A0B09FA0B090A0BF9FA0B090A0
BF9FA0B0909FBF9FA0B0909FBF909FB0909FBF909FB0909FBF909FB09090BF909FB09090BF9090\
B08F8FBF9090BF8F8FBF908FB08F8FBF908FBF8F80BF8F8FB08080BF8F80B0807F
B0807FAF7F70B08070AF7F70AF7F70A0706FAF7F6FA07060A070609F6F5FA06F5F9F6F5F9F6F5F\
9060509F6050905F4F905F4F8050407F50406F4F40604040503F3F4F3F3F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303F5F\
4F4F705F5F8070709F7F7FA0808FAF8F90B09F9FC09FA0CFAFB0DFAFB0D0B0B0D0
AFB0CFB0B0CFB0B0C0B0B0C0B0B0B0B0AFB0AFAFA0AFA0A0A09F90A09F908F80707F7F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F3040303F40303F3F3F40
30304F3F3F5F403F70504F7F4F4F8050508F5F5F90606FA06F6FA0707FB07F80BF808FC0808FC0\
8090CF8090C08F9FCF8F9FCF90A0D090A0D09FAFD090AFD09FB0DF9FB0DFA0BFE0
9FB0DFA0BFDF9FBFDFA0C0DF9FBFD0A0C0D09FBFD09FC0D090BFCF9FC0CF90BFBF90BFB080AFA0\
7FAFA0709F909F8FAF9080A09F8FAF9F8FA09F8FAF9F8FAF9F8FAF9F8FAFA090AF
9F8FAFA090AFA090AFA090AFA090AFAF90AFA090AFAF90AFA090AFAF90AFAF90AFB09FAFAF90AF\
B09FAFB090AFB09FAFB09FAFB09FAFB09FAFBF9FB0B09FAFBF9FAFBF9FAFBF9FAF
BF9FAFBF9FAFBF9FAFBF9FAFB090A0BF9FA0BF90A0BF9FA0B0909FBF9F9FBF909FBF9F9FB09090\
BF909FBF9090BF909FBF9090BF9090BF908FBF9090BF8F8FBF908FBF8F8FBF908F
BF8F80BF908FBF8F80BF8F80B0807FB08F7FB08070B0807FAF7F70AF7F70AF7F6FAF7F6FA07060\
A070609F6F5FA06F5F9F60509F605090604F90604F905F408F5F408050407F5040
6F403F60403F503F3F4F3F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303F50\
4F4F6F5F60806F6F908080A0808FAF9090BF909FC0A0AFD0A0B0DFAFB0DFAFB0D0
B0BFD0B0B0CFB0BFCFB0B0BFBFBFBFB0B0B0B0AFAFAFA0A0AFA09FA09F90909080807F60706F50\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000403F404030404F3F503F304F
3F3F5F403F6F4F407F4F4F8050508F5F5F9060609F606FA0707FB0707FBF7F8FC07F8FC08090CF\
8090CF8F9FCF8F9FCF90A0D090A0D090AFDF90AFD09FB0DF9FB0DFA0BFE09FBFDF
A0BFE09FBFDFA0BFDF9FBFD0A0C0DF9FBFD09FC0D090BFCF9FC0CF90BFC08FBFBF80AFAF80AFA0\
70A0909F8FB0908FAF9F90B09F8FAF9F90B09F90AFA090B09F90B0A090B09F90B0
A090B0A090B0A09FB0A090B0AF9FB0A090B0AF9FB0A090B0AF9FB0AF9FB0AF9FB0AF9FB0B09FB0\
AF9FB0B09FB0B09FB0B0A0B0B09FB0BFA0B0B09FB0BFA0B0BFA0B0BFA0B0BFA0B0
C0A0B0BFA0B0C0A0B0BFA0AFC0A0B0BFA0AFC0A0AFBF9FAFC0A0AFBF9FA0C0A0AFBF9FA0C0A0A0\
BF9FA0C09FA0BF9F9FC09FA0BF9F9FC09F9FBF9F9FC09F9FBF9F90C09F9FBF9090
C09F90BF9090C09F90BF908FC09F90BF908FC0908FBF9080C0908FBF8F80BF8F7FB08070B0807F\
AF8070B08070AF7F6FAF7F6FA07060A07060A06F5FA0705F9F6F509F6F5090604F
9F604F905F4F905F4F8050407F50406F4F3F60403F503F3F4F3F3F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
404F6F50507F6F6F907F7F9F808FAF8F8FB0909FC09FA0CFA0B0DFAFB0DFAFB0D0
AFB0CFB0B0CFB0B0C0B0B0C0B0B0BFB0B0B0AFAFAFAFAFA0A0A09FA0A09090908080806F706F50\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000040304040304F4F3F503030403F3F50
3F3F604F40704F407F504F8050508F5F5F9F5F609F6F70AF6F70B07F80BF7F80BF808FC08090C0\
8F9FCF809FCF8FA0D08FA0D090AFD090AFD09FB0DF90B0DF9FBFE090B0DF9FBFDF
9FBFD09FBFDF90BFD09FBFD09FBFCF9FC0D090BFCF90BFCF8FB0BF8FB0B080AFA07FAF9F709F90\
908FB0908FAF9090B0908FB09F90B09090B09F90B09F90B0A09FBF9F90B09F90B0
9F90B0A09FB09F90B0A09FB0A090B0A09FB0A090B0A09FB0A09FB0AF9FB0A09FB0AF9FB0AF9FB0\
B0A0BFAF9FB0B0A0B0AF9FB0B0A0BFB09FB0B0A0B0B0A0B0BFA0BFB0A0B0BFA0BF
BFA0B0BFAFBFBFA0B0C0AFBFBFA0B0C0AFBFBFA0B0C0A0B0BFA0AFC0A0B0BFA0AFC0A0AFBFA0AF\
C0A0AFBF9FA0C0A0AFBF9FA0C0A0A0BF9FA0C0A0A0C09F9FC0A0A0C09F9FC09F9F
C09F9FC09F9FC09F90C09F9FC09F90C09F90C09F90C09F90C09F8F707F80BF908FC0908FBF8F80\
BF9080B08F7FB08F7FB08070B08070AF7F6FAF7F6FAF7F60AF7F60A0705FA07060
9F6F5FA06F5F9F60509F6F5090604F9F604F905F408F5F407F503F7F504060403F60403F503F30\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F4060505F7F606F8F7F7FA07F80A08F90B09090BF9FA0CFA0AFD0AFB0E0AFB0D0
B0B0D0B0B0CFB0BFCFB0B0C0BFB0BFB0B0B0B0B0B0AFAFA0AFA0A0A09F90A09F8F8F80707F705F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000403040403F504F405F50405F3F3F503F3F5F
4040704040704F4F804F508F505F905F5F9F606FAF6F70AF7080BF7080BF7F8FC07F8FC08090CF\
8090CF8FA0D08FA0D090AFD090AFD09FB0DF90B0DF9FB0DF90B0DF9FBFDF90B0D0
9FBFDF90BFD09FBFD090BFD09FBFD090BFCF90C0CF8FB0BF8FB0B080AFA07FAFA0709F90908FB0\
8F8FAF9090B0908FB09090B09090B09F90BF9090B09F90BF9090B09F9FBF9F90BF
9F9FBF9F90B0A09FBFA09FBFA09FBFA09FBFA0A0BFA09FBFAFA0BFA09FBFAFA0BFAF9FBFAFA0BF\
AF9FBFB0A0BFAFA0BFB0A0BFAFA0BFB0A0BFB0A0BFB0AFBFB0A0BFBFAFBFB0A0BF
BFAFBFBFA0BFBFAFBFC0AFBFC0AFC0C0AFBFCFB0C0C0AFBFCFB0BFC0AFBFCFB0BFCFAFB0CFAFB0\
C0AFB0CFAFB0CFAFB0CFAFB0CFA0AFCFAFB0CFA0AFCFAFAFCFA0A0CFAFAFCFA0A0
CFAFA0CFA0A0CFAFA0CFA09FCFA0A0CFA09FD0A09FCFA09FD0A09FCFA09FCFA09FCFA090CFA090\
C09F8FC09F90C0908FC0908FBF9080BF9080B08F7FBF8F7FB08070B08070AF806F
AF806FAF7F60AF7F60A07060A0705F9F6F5FA0705F9F6F509F6F5090604F9F604F905F408F5F40\
7F5040704F4060403F5F403F503F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F40604F4F7060608F7070907F80A0808FAF9090BF909FC0A0AFD0A0B0DFAFB0DF
AFB0D0B0B0CFB0B0C0B0B0C0B0B0BFB0B0BFB0AFAFB0AFAFA0A09FA0A0909F9F8F908F707F705F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000030303F403F504F405F4F4060504F6F3F3050403F6F
403F704F407F4F4F8050508F505F905F60A0606FA06F7FB06F7FB07080BF7080BF7F90C07F90C0\
809FCF809FCF8FA0D08FA0D090AFDF90AFDF90B0DF90B0D090B0DF90B0D090BFD0
90B0CF90BFD090BFCF90BFCF8FBFC090BFC08FB0BF80B0B07FA0A07FA09F6F9F8F8F8FB0808FAF\
8F8FB08F8FAF8F90B08F8FB09090B08F90B09090B08F90B09090B09090B09F9FBF
9090BF9F9FBF9F9FBFA09FBF9F9FBFA09FBF9F9FBFA0A0BFA09FBFA0A0BFA09FBFAFA0C0A09FBF\
AFA0BFAFA0BFAFA0C0AFA0BFAFA0C0AFA0BFB0AFC0AFA0BFB0AFC0B0A0BFB0AFC0
B0A0BFBFAFC0B0AFBFBFAFC0BFAFBFC0AFC0C0AFBFC0B0C0C0AFBFCFB0BFCFB0BFD0B0C0CFB0BF\
D0B0BFCFB0BFD0B0BFCFAFB0D0B0BFCFAFB0D0B0B0CFAFAFD0AFB0D0AFAFD0AFB0
D0AFAFD0AFAFD0AFAFD0AFAFD0AFA0D0AFA0D0AFA0D0AFA0D0A09FD0AFA0CFA09FCFAFA0CFA09F\
CFA09FCFA090CFA090C09F90C0A090C09F8FC09F8FBF9080BF9080BF907FBF907F
B08F70B08F70B08070B08070AF7F6FAF7F6FA07F60AF7F60A0705FA0705F9F6F50A06F509F604F\
9F6F4F90604F90604F805F40805040704F3F6F4F405F3F3F503F3F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F3F604F50705F6080707F9F7F80A08F8FB08F90BF9FA0C09FA0CFAFB0DFAFB0DF
B0BFDFB0B0CFB0BFCFB0B0C0BFBFC0B0B0BFB0B0B0AFAFAFAFAFA0A0A09FA09F90908F7F807F60\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F303F4F3F505040605040605F4F6F504F703F3F60403F6F
4F4F7F4F4F8050508F5050905F609F606FA06F70B06F7FB07080BF7080BF7F8FC07F8FC0809FCF\
809FCF8FA0D08FA0D090AFDF90AFDF90B0DF90B0DF90B0DF90B0D090BFD090B0D0
90BFD090BFCF90BFD08FBFCF90BFCF8FB0BF80B0B07FA0A07FA09F6F9F8F808FB0808FAF8F90B0\
808FB08F90B0808FB08F90B08F90B08F90BF8F90B09090BF8F90B0909FBF9090BF
9F9FBF909FBF9F9FC09F9FBFA0A0C09F9FC0A0A0C0A0A0C0A0A0C0A0A0C0A0A0C0A0A0C0AFAFC0\
AFA0C0AFAFC0AFA0C0AFAFCFAFAFC0B0AFC0AFAFC0B0AFCFB0AFC0B0AFCFB0AFC0
BFB0CFB0AFC0BFB0C0BFAFC0C0B0C0BFAFC0C0B0C0C0B0C0CFB0C0C0B0C0CFB0C0CFB0C0D0BFC0\
CFB0C0D0BFC0D0B0BFDFBFC0DFBFBFDFBFC0DFB0BFDFBFC0DFB0BFDFBFBFDFB0B0
DFBFBFDFB0B0DFB0B0DFB0B0E0B0B0DFB0AFE0B0B0DFB0AFE0B0AFD0B0AFD0B0AFD0AFA0D0B0AF\
D0AFA0D0AFA0CFAF9FD0AFA0CFA09FCFAF9FCFA090CFA090CFA090CFA090C09F8F
C09F8FC09F80C09F80BF907FBF907FB08F70BF8F70B0806FB08F70AF806FAF806FAF7F60AF7F60\
A0705FA0705F9F6F50A070509F6F4F9F6F4F9060408F604F8050407F50406F4F3F
6F4F405F3F3F503F3F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
303F50404F6F5F5F806F6F907F80A08080AF8F90BF909FBF9FA0CFA0AFD0AFB0DF
AFB0D0B0B0D0AFB0CFB0B0C0B0B0BFBFB0BFB0AFB0B0AFAFAFA0A0AFA09FA09F8F9F9080807F60\
706F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000030303040304F4F40604F4060504F6F504F6F5F50703F3060404070
40407F4F4F804F508F505F9F5F609F6070AF6070AF6F7FB06F80BF708FC0708FBF7F90C07F90C0\
809FCF809FCF8FAFD08FAFD090B0DF8FAFD090B0D08FB0D090B0D08FB0CF90B0D0
8FB0CF90BFCF8FB0C08FBFC080B0BF80AFB07FA0A070A09F6F9F8F7F8FB07F80AF808FB07F8FAF\
808FB07F8FAF808FB0808FB08090B0808FB08F90B08090B08F90BF8090B08F90BF
8F90B0909FBF8F90BF909FBF909FBF9F9FC0909FBF9FA0C09F9FBFA0A0C09FA0C0A0A0C0A0A0C0\
A0AFCFA0A0C0AFAFCFA0A0C0AFAFCFAFAFC0AFAFCFAFAFC0B0AFCFAFAFC0B0AFCF
B0AFC0B0B0CFB0AFC0B0B0C0B0AFC0BFB0C0BFAFC0C0B0C0BFB0C0C0B0C0C0B0C0CFB0C0CFB0C0\
CFBFC0CFB0C0D0BFC0D0B0C0D0BFC0D0B0BFDFBFC0DFBFBFE0BFC0DFBFBFE0BFC0
E0BFBFE0BFC0E0BFBFE0BFBFE0BFBFE0BFBFE0B0B0E0BFB0E0B0B0DFBFBFD0B0B0DFBFB0D0B0AF\
DFB0AFD0B0A0D0B0AFD0AFA0D0B0A0CFAF9FD0AFA0CFAF9FCFAF9FCFA090CFA090
CFA090CFA090C09F8FC0A08FC09F80C09F8FBF907FBF9080BF907FBF907FB08F70B08F70B0806F\
B0806FAF7F60AF7F60A07F5FA07F5F9F7050A070509F6F4F9F6F4F906040906040
8F5F40805F407F4F3F704F3F60403F60403F503F30000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000404F6F505F7F6F6F90707F9F808FAF8F8FB0909FC09FA0CFA0B0DFAFB0DF
AFB0DFAFB0CFB0BFCFB0B0C0BFB0C0B0B0BFBFB0BFB0AFAFB0AFA0AFA09FAFA0909F90808F8070\
706F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F30404F3F5F4F4060504F70504F705F507F4F503F404070404070
4F4F804F4F8F505F90505F9F5F6FA06070AF6F7FB06F7FB07080BF7080BF7F90C07F90C0809FCF\
809FCF8FA0D08FA0D08FAFDF8FAFD090B0DF8FB0D08FB0D08FB0CF90BFD08FB0CF
8FBFCF8FB0C08FBFC080B0BF80AFB070A0A070A09F6F9F8F7F8FAF7080AF7F8FB07F8FAF7F8FB0\
7F8FAF8090B07F8FB08090B07F8FB08090BF8090B08090BF8090B08F90BF8090B0
8F90BF8090B08F9FBF8F90BF909FBF8F9FBF909FC0909FBF90A0C0909FBF9FA0C09FA0C09FA0C0\
9FA0C0A0AFCFA0A0C0A0AFCFA0A0C0AFAFCFAFAFCFAFAFCFAFAFCFB0B0D0AFB0CF
B0B0D0B0B0CFB0B0CFB0AFC0B0B0CFB0B0C0BFB0CFBFB0C0C0B0CFBFB0C0C0BFCFC0B0C0CFBFCF\
C0B0C0CFBFCFCFBFC0D0BFC0D0BFC0D0BFCFD0BFC0DFBFC0DFBFC0E0C0C0DFBFC0
E0C0C0E0BFC0EFC0C0E0C0C0EFC0C0EFC0C0F0C0C0EFC0BFEFC0C0DFC0BFE0C0C0DFBFBFE0C0BF\
DFBFB0DFBFB0DFBFB0DFBFB0D0B0AFDFBFAFD0B0A0D0B0AFD0B0A0D0B0A0D0AF9F
D0AF9FCFAF9FD0AF9FCFAF90CFAF90CFA08FCFAF90C0A08FCFA08FC09F80C0A080BF9F7FC09F7F\
BF9070BF9070B08F6FB08F6FAF8060AF806FAF7F60AF7F60A0705FA0705F9F6F50
9F6F50906F4F9F6F4F906040906040805F3F80503F704F3F6F4F3F6040305F403F503030000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F406050507F60608F707F9F7F80A08F8FB08F90BF9FA0CFA0AFD0AFB0DF
AFB0D0AFB0D0AFB0CFB0B0CFB0B0BFBFB0BFB0B0B0B0AFAFAFA0A0AFA09FA09F90A09F8F8F806F\
7F705F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F3F403F50403F5F4F4F6F504F7050507F50507F3F3F603F3F70404F7F
404F804F508F4F50905F60A05F6FA06070B06070B06F80BF6F80BF708FBF708FC07F90CF7F90C0\
809FCF80A0CF8FAFD080AFD08FAFD08FAFCF8FB0D080AFCF8FB0CF80B0CF8FB0CF
80B0C08FB0C07FAFB07FAFAF70A09F709F906F908F6F80AF6F80A07080AF6F80AF708FAF7080AF\
708FAF7080AF7F8FB0708FAF7F8FB07F8FB07F90B07F8FB07F90B07F90B07F90B0
7F8FB08090B08090B08090BF8090B08F9FBF8090BF8F9FBF8F9FBF909FBF8F9FBF90A0C0909FBF\
90A0C090A0C09FA0C09FA0C09FA0C09FA0C0A0AFCFA0A0C0A0AFCFA0AFCFAFAFCF
A0AFCFAFB0CFAFAFCFB0B0CFAFAFC0B0B0CFB0B0C0BFB0CFB0B0C0BFB0CFBFB0C0C0BFCFC0B0C0\
C0BFCFC0B0C0CFBFCFCFB0C0CFBFC0CFBFC0D0BFCFD0BFC0D0BFC0D0BFC0DFC0C0
DFBFC0E0C0C0DFBFC0E0C0C0E0BFC0EFC0C0E0C0C0EFC0C0E0CFC0E0CFCFE0C0C0E0CFC0DFC0BF\
E0C0C0DFC0BFE0C0BFDFBFB0DFBFB0DFBFB0DFBFB0D0B0AFDFBFAFD0B0AFDFB0AF
D0B0A0D0B0A0D0AF9FD0B0A0CFAF9FD0AF9FCFAF90CFAF90CFA08FCFAF90C0A08FCFA08FC0A080\
C0A080BF9F7FC09F7FBF9070BF9070B08F70B08F70AF8060AF806FAF7F60AF7F60
A0705FA0705F9F70509F7050906F4F906F4F9060409060408F5F3F8F5F3F7F503F70503F6F4030\
60403F5F3F30503F3F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F406F4F5070606F8F707F9F8080AF808FB0909FC09F9FC0A0AFD0AFB0DF
AFBFDFAFB0D0B0B0D0B0B0CFBFBFCFB0B0BFBFB0BFB0AFAFB0AFAFAFA09FAFA09FA09F8F908F7F\
7F705F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30303F403F50403F5F504F6F504F7050507F50507F5F5F803F3F6040407F404F7F
4F508F4F509050609F5060A06070B06070B06F7FB06F7FB0708FC0708FBF7F90C07F90C0809FCF\
80A0CF80AFD080A0D08FAFD08FAFD08FB0D080AFCF8FB0CF80B0CF8FB0CF80B0C0
8FB0C080AFB07FAFAF70A0A070A09F6F908F607FA0607F9F6F80A0607FA06F80A06F80A06F80AF\
6F80A06F80AF6F80A0708FAF7080AF708FAF708FAF708FB0708FAF708FB0708FAF
7F90B07F90B07F90B07F90B08090BF8090B0809FBF8090BF8F9FBF8F9FBF8F9FBF8F9FBF90A0C0\
8F9FBF90A0C0909FBF9FA0C090A0C09FA0C09FA0C09FAFCF9FA0C0A0AFCF9FAFC0
A0AFCFA0AFCFAFB0CFAFB0CFB0B0D0AFB0CFB0B0CFB0B0CFBFBFCFB0B0CFBFBFCFBFB0CFC0BFCF\
BFBFCFC0BFCFC0BFCFCFBFCFC0BFCFCFBFCFCFBFCFD0C0CFCFBFCFD0C0CFD0BFC0
DFC0CFD0C0CFDFC0CFDFC0C0E0C0CFDFC0C0E0C0CFE0C0C0EFD0D0EFD0D0EFD0D0E0CFCFEFD0CF\
E0CFC0E0CFCFE0CFC0E0CFC0E0C0BFE0CFBFDFC0BFE0C0BFDFC0B0DFC0B0DFBFAF
DFC0B0D0BFAFDFBFAFD0B0A0D0BFA0D0B0A0D0B0A0CFB09FD0B09FCFAF90CFB09FCFAF90CFAF90\
CFAF8FCFAF90C0A08FCFA08FC09F80C09F80BF907FBF9F7FB09070B09070B08F6F
B08F6FAF8060AF8060A07F5FA07F5FA07050A070509F6F4F9F704F906F40906F408F603F906040\
805F3F805F3F704F3F704F3F60403F5F403F503F30000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000303F5F4F50705F5F8070709F7F7FA0808FB08F90BF9F9FC09FA0CFAFB0DF
AFB0D0AFB0D0AFB0CFB0B0CFB0B0C0BFB0C0B0B0B0B0B0B0AFAFA0AFAFA0A09F90A09F90908F7F\
807F606F604F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F30\
3F304F3F3F504F40604F4F6F50507F50507F5F5F80505F803F3F6F3F407F404F80
40508F4F5F9F4F5F9F5F6FAF5F6FAF607FB0607FB06F80BF6F80BF708FC0708FC07F9FCF7F9FCF\
80A0D07FA0CF80AFD080AFCF80AFCF80AFCF80AFCF80AFC080B0CF80AFC080B0C0
7FAFB07FAFAF70A09F709F9F6F908F5F70905070905F7F9F5F70905F7F9F5F70905F7F9F5F7F9F\
607F9F5F7F9F607FA0607F9F6080A0607F9F6080A06080A06F80A06080A06F80A0
6F80AF708FB0708FAF7090B0708FB07F90B0708FB07F90B07F90B0809FBF7F90B0809FBF8090BF\
8F9FBF809FBF8F9FBF8F9FBF8FA0C08F9FBF90A0C090A0BF90A0C090A0C09FA0C0
90A0C09FAFCF9FAFC0A0AFCFA0AFCFAFB0CFA0AFCFAFB0CFAFB0CFB0BFD0AFB0CFB0BFCFB0B0CF\
BFBFCFB0B0CFBFBFCFBFBFCFC0BFCFBFBFCFC0BFCFC0BFCFCFC0CFC0BFCFCFC0CF
CFBFCFD0C0CFCFBFC0D0C0CFD0C0C0DFC0CFD0C0C0DFC0CFDFCFCFE0D0D0DFCFCFE0D0CFE0CFCF\
E0D0CFE0CFCFE0D0CFE0CFC0E0D0C0DFCFBFE0CFC0DFC0BFE0CFBFDFC0B0DFC0B0
DFC0B0DFC0B0D0BFAFDFC0AFD0BFAFD0BFAFD0BFA0D0BFA0CFB09FD0BFA0CFB09FCFB09FCFB090\
CFB090C0AF90CFAF90C0AF8FCFAF8FC0A080C0A080C0A080C0A080BF9F7FBF9F7F
B09070B09070AF8F6FAF8F6FAF8060AF8060A07F5FA07F5F9F70509F7050906F4F9F704F906F40\
906F408F603F8F6040805F3F805F3F704F30704F3F60403060403F503F30503F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000303F5F404F705F60806F6F907F80A0808FAF8F90BF909FC0A0AFD0A0AFD0
AFB0DFAFB0D0B0B0D0B0B0CFB0B0CFB0B0BFBFBFBFB0B0B0B0AFAFAFA0A0AFA09FA09F90A09F8F\
8F806F7F705F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F304F\
3F30504F40604F406F50507F50507F5F5F80505F805F608F3F3060404F7F40508F
4F5F904F5F9F506FA05F6FAF5F70AF5F70AF6F80BF6F80BF6F8FBF6F8FBF7090CF709FC07FA0CF\
7FA0CF80A0D080A0CF80AFCF80AFCF80AFCF80AFC080B0C080AFC080B0C080AFBF
7FAFB070A0A070A09F6F908F4F708F4F6F8F5070904F708F5070904F708F507090507090507090\
507090507F9F5070905F7F9F5070905F7F9F507F9F5F7F9F5F7F9F5F7F9F5F7F9F
5F7F9F5F7F9F6080A06080A06F8FAF6F80A0708FAF708FAF7090B07090B07F90B07F90B07F9FBF\
7F90B0809FBF809FBF809FBF809FBF8FA0BF809FBF8FA0C08F9FBF90A0C08FA0C0
90A0C090A0C09FAFC090A0C09FAFC09FAFC0A0B0CFA0AFC0A0B0CFA0B0CFAFB0CFAFB0CFAFBFD0\
AFB0CFB0BFD0B0BFCFB0BFD0B0BFCFBFBFD0BFBFCFBFC0D0BFBFCFC0C0D0C0BFCF
C0C0CFC0BFCFCFC0D0CFC0CFCFC0CFCFC0CFD0CFCFD0C0CFDFCFD0D0CFCFDFD0D0DFCFCFDFD0CF\
DFCFCFE0D0CFDFCFCFE0D0CFDFCFC0E0D0CFE0D0C0E0D0C0E0CFC0E0D0C0DFCFBF
E0CFBFDFCFB0DFCFBFDFC0B0DFC0B0D0C0AFDFC0B0D0C0AFD0C0AFD0BFA0D0C0A0D0BFA0D0BFA0\
CFB09FCFBF9FCFB090CFB090C0AF90CFB090C0AF8FC0AF8FC0A080C0AF8FBFA080
C0A080BF9F7FBF9F7FB09070B09070AF8F6FB0906FAF8F60AF8F60A0805FA0805F9F7F509F7F50\
9F704F9F7050906F4F906F4F8F60408F6040805F3F805F3F7F50307F503F6F4F30
6F403F5F3F305F403F503F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000404F6F50507F606F90707F9F808FAF808FB0909FC09FA0CFA0AFD0
A0B0D0AFB0D0AFB0CFB0B0CFB0B0C0B0B0C0B0B0BFB0B0B0AFAFAFAFAFA0AFA09FAFA0909F908F\
9F8F7F807F60706F4F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000030304030304F\
403F5F4040604F4F704F5070505F7F505F7F5F60803030503F3F6F3F407F40508F
4050904F60A05060A05F70AF5F70AF607FB0607FB06F80BF6F80BF708FC07090C07F9FCF709FC0\
7FA0CF7FA0C080A0CF7FA0C080AFC07FA0C080AFC07FAFBF80AFC07FAFB07FAFB0
70A0A0709F9F6F908F5F7F605F3F30603F303F50603F5F6F3F506F405F703F5F704060703F5F70\
40607F40607F4F6F804060804F6F8F406F8F4F708F4F6F8F4F70904F708F507090
4F708F4F708F4F708F5070905070905F7F9F5F7F9F5F809F5F7F9F6080A06080A06F8FAF6F8FA0\
708FAF708FAF7090B07090B07F9FB07F90B0809FBF7F90B0809FBF809FBF8FA0BF
809FBF8FA0BF809FBF8FA0BF8FA0BF90A0C090A0BF90AFC090A0C09FAFC09FAFC0A0B0CF9FAFC0\
A0B0CFA0B0CFAFBFCFAFB0CFAFBFCFAFBFCFB0BFD0B0BFCFB0BFD0B0BFCFBFC0D0
BFBFCFBFC0D0BFBFCFC0C0D0C0BFCFC0C0CFC0C0CFCFC0CFCFC0CFCFCFCFCFCFCFD0CFCFCFCFCF\
D0CFCFD0CFC0D0CFCFD0CFC0DFCFCFD0CFC0DFD0C0D0CFC0DFD0C0DFCFC0E0D0C0
DFCFBFDFD0C0DFCFBFDFCFBFDFCFB0DFCFB0D0C0B0DFCFB0D0C0AFD0C0AFD0C0A0D0C0AFCFBFA0\
CFC0A0CFBF9FCFBF9FC0B090CFB09FC0B090CFB090C0AF8FC0AF8FC0AF8FC0AF8F
BFA080BFA080BFA07FBFA07FB09F7FBF9F7FB09070B09070AF8F6FAF8F6FAF8F60AF8F60A0805F\
A0805F9F7F509F7F5090704F90704F8F6F4F8F6F4F806040806040805F3F805F3F
7F50307F50306F4F306F4F30603F305F3F30503030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F406F505F80606F8F7F7FA07F80AF8F90BF8F90BF9FA0CFA0AFD0
AFB0E0AFB0D0B0B0D0AFB0CFB0BFCFB0B0C0BFBFC0B0B0BFBFB0B0B0AFAFB0AFA0AFA09FAFA090\
A0908F908F707F705F706040000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000030304F403F5F\
4040604F4F704F5070505F80505F805F60805060803F3F5F3F3F6F404F7F40508F
4F609F4F60A0506FAF506FAF5F7FB05F7FB06080BF6080BF6F8FC06F90C0709FCF709FCF7FA0CF\
7FA0C07FA0CF7FA0C080AFCF7FA0C080AFC07FAFBF80AFC07FAFBF7FAFB070A0A0
70A09F6F9F906090806040306F403F6F40306F4F3F6F4030704F3F6F4F30704F3F6F4F3070503F\
704F3070503F704F3070503F70503F7F5F3F3040503F4F5F3F4F5F3F4F5F3F4F5F
4050603F50603F6F803F6080406F8F406F8F4F708F4F708F507F90507090507F9F507F905F809F\
5F809F6080A06080A06F8FAF6F8FA07090AF6F8FAF7090B07090B07F9FB07F9FB0
809FBF809FBF809FBF809FB08FA0BF80A0BF8FA0BF8FA0BF90AFC090A0BF90AFC090AFC09FB0C0\
9FAFC09FB0CF9FB0C0A0B0CFA0B0CFAFBFCFA0B0CFAFBFD0AFBFCFB0BFD0B0BFCF
BFC0D0B0C0D0BFC0D0BFC0D0C0C0D0BFC0D0C0C0D0C0C0CFCFCFD0C0C0CFCFCFCFC0C0CFCFCFCF\
CFCFCFCFCFCFCFCFC0D0CFCFCFCFC0D0CFCFD0CFC0D0D0CFD0CFC0D0D0C0D0CFC0
DFD0C0DFD0C0DFD0C0DFD0BFDFD0C0DFD0BFDFD0BFDFCFB0DFD0BFDFCFB0DFCFB0D0CFAFDFCFB0\
D0C0AFD0CFAFCFC0A0D0C0A0CFBF9FCFBF9FC0BF9FCFBF9FC0B090C0B090C0B08F
C0B090C0AF8FC0AF8FBFA080C0AF80BFA07FBFA07FB09F7FBFA07FB09F70B09F70B0906FB0906F\
AF8F60AF8F60A0805FA080609F7F5F9F7F5F9070509070508F6F4F8F6F4F806040
806040805F3F7F5F3F70503070503F6F4F306F403F5F3F305F3F3F2F3F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F40604F507060608F6F709F7F80AF808FB08F9FBF909FC0A0AFD0
A0B0DFAFBFDFAFB0D0B0BFD0B0B0CFB0BFCFB0B0BFBFB0BFB0B0B0B0B0B0AFA0A0AFA09FA09F90\
A09F8F908F7F8F7F6F706F4F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F503F3F5F\
4F4F6F4F4F7050507F505F7F506080505F8050608F3030503F3F6F3F407F405090
4F5F9F506FA04F6FA05070AF5F70AF607FB05F7FB0608FBF6F8FBF7090C06F90C0709FC0709FC0\
7FA0C0709FC07FA0C07FA0BF7FAFC070A0BF7FAFBF70A0B07FAFB070A0A0709F9F
6F908F60908F6040306F4F3F6F4F30505F506F4F306F4F3F6F4F3070503F6F4F3070503F705030\
70503F70503F7F5F3F705F3F7F5F407F5F3F7F5F407F5F407F60407F5F4080604F
7F604080604F7F604F806F4F806F4F8F6F50806F508F7050806F503F6F803F6080406F80406F80\
4F708F4F708F4F70904F7090507F90507F905F809F5F7F9F6080A060809F608FA0
608FA06F90AF7090AF7F9FB07090B07F9FB07F9FB0809FBF809FB080A0BF809FBF8FA0BF8FA0BF\
90AFC08FA0BF90AFC090AFC09FB0C09FAFC09FB0C09FB0C0A0B0CFA0B0C0AFBFCF
AFBFCFAFBFCFAFBFCFB0C0D0B0BFCFBFC0D0B0C0D0BFC0CFBFC0CFBFC0CFBFC0CFC0C0CFBFC0CF\
C0CFCFC0C0C0C0CFCFC0C0C0CFCFCFC0C0C0CFCFC0C0CFC0CFCFC0CFCFBFCFCFC0
CFCFBFD0CFC0CFCFBFD0CFBFD0CFBFD0D0BFD0CFBFD0D0BFD0CFB0DFD0BFD0CFB0DFD0B0D0CFB0\
D0D0B0CFCFAFD0CFAFCFC0A0CFC0A0CFC09FCFC0A0C0BF9FCFBF9FC0B090C0BF90
C0B090C0B090BFAF8FC0B08FBFAF80BFAF80BFA080BFAF80B0A07FBFA07FB09F70B09F70B09F70\
B09F70AF906FAF906FA08F60A08F609F805F9F805F907F50907F508F704F8F704F
806F408F6F407F5F3F7F5F3F7F503F70503F6F4F306F4F3F6040306040305F30302F3F4F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000404F70505F806F70907F80A0808FAF8F90B090A0C09FA0CF
A0B0D0AFB0DFB0BFDFAFBFD0B0BFD0B0B0CFBFBFCFB0B0BFBFB0BFB0AFB0B0AFAFAFA0A0AFA09F\
A09F90A09F8F9080708070606F6040000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F5F40406F\
404F6F50507F50507F505F80505F805F608F50608F3F3F5F3F3060404F80405090
4F609F4F60A05070AF5070AF5F7FB05F7FB06080BF608FBF6F90C06F90C0709FC0709FC0709FC0\
709FC07FA0C070A0BF7FA0C070A0BF7FAFBF70A0B07FA0B070A0A070A0A06F9F90
6F908F6F4F306F503F6F503070503F6F503070503F70503F705F3F70503F7F5F3F705F3F7F5F3F\
705F3F7F60407F5F407F60407F604080604F7F6040806F4F7F604F806F4F806F4F
806F50806F4F8F7050806F508F70508070508F705F8F705F8F7F5F8F705F8F7F608F7F5F907F60\
8F7F609080608F7F605F40403F6F8040708F406F8F4F708F4F708F507F90507F90
507F9F507F906080A06080A06F8FAF6F8FAF709FB07090B07F9FB07F9FB0809FBF7F9FB080A0BF\
809FB08FA0BF8FA0BF8FAFBF8FA0BF90AFC090AFC090AFC090AFC09FB0C09FB0C0
A0B0CF9FB0C0A0BFCFA0BFCFAFBFCFAFBFCFB0C0D0B0C0CFB0C0CFB0C0CFBFC0CFB0C0CFBFC0CF\
B0C0C0BFCFCFBFC0C0BFCFCFBFC0C0C0CFCFBFC0C0C0CFC0C0C0C0C0CFC0C0CFC0
CFCFC0C0CFBFCFCFC0CFCFBFCFCFBFCFCFBFD0D0BFCFCFBFD0D0BFCFCFB0D0D0BFD0D0B0D0D0BF\
D0D0B0D0D0B0D0D0B0D0D0B0CFCFAFD0CFAFCFC0A0CFCFA0C0C09FCFC0A0C0BF9F
C0BF9FC0BF90C0BF90C0B090C0B090BFAF8FBFB08FBFAF80BFAF80B0A07FBFAF80B0A07FB0A07F\
B09F70B09F70AF9F6FAF9F6FAF9060AF906FA08F60A08F609F805F9F805F907F50
907F508F6F4F8F704F806F408060407F5F3F7F5F4070503F70503F604F3F604F3F5F3F30000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003040604F507F5F608F707F9F7F80A0808FB08F90BF90A0C0
9FA0CFA0B0D0AFBFDFB0BFDFAFB0CFB0BFCFB0B0C0B0B0C0B0B0BFBFB0B0B0AFAFB0AFAFAFA09F\
AFA090A09F8F9F90808F7F607F6F50000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F3F40406040406F\
4F4F704F5070505F7F505F7F506080506080506F8F3030503F3F6F3F407F405090
405F904F60A04F6FA05070AF5070AF5F7FB05F80B0608FBF608FBF6F90C06F90BF709FC06F90BF\
709FBF709FBF70A0BF709FBF70A0BF70A0B070A0B06F9FA0709F9F6F90906F908F
4F6F406F503F6F5030705F3F6F5030705F3F705F30705F3F705F3F7F5F3F705F3F7F60407F5F3F\
7F60407F60408060407F6040806F4F7F6040806F4F7F6F4F806F4F806F4F807050
806F508F70508070508F705F8070508F7F5F8F705F8F7F608F7F5F8F7F608F7F609080608F7F60\
90806F4F303F5F404F604F507F5F6F80607F3F3050404F705F6F9060709F7F80AF
3F6080406F804060804F708F4F6F8F507F90507F9060809F60809F6F8FA06F8FA07090B07090AF\
7F9FB07F9FB07F9FB07F9FB080A0BF80A0B08FA0BF80A0BF8FAFBF8FA0BF90AFC0
8FAFBF90AFC090AFC09FB0C09FB0C09FB0C09FB0C0AFBFCFA0BFC0AFBFCFAFBFCFB0C0CFAFBFC0\
B0C0CFB0C0C0B0C0CFB0C0C0BFC0C0B0C0C0BFC0C0B0C0C0BFCFC0BFC0C0C0CFC0
BFC0BFC0CFC0BFC0BFC0CFC0C0C0BFC0CFBFC0CFBFCFCFBFC0CFB0CFCFBFCFCFB0CFD0BFCFCFB0\
CFD0BFCFD0B0D0D0B0CFCFAFCFD0B0CFCFAFCFCFAFC0C0A0CFCFA0C0C0A0C0C0A0
C0BF9FC0BF9FBFBF90C0BF90BFB090BFB090BFB08FBFB08FB0AF80BFAF80B0A080B0AF80B0A07F\
B0A07FAF9F70B09F70AF906FAF9F6FA0906FA0906F9F8F60A08F609F805F9F805F
907F50907F508F704F8F704F806F40806F407F603F7F604070503F6F503F604F30604F3F503F30\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000404F705F6080606F907F80A0808FAF8F90BF909FBF
9FAFCFA0AFD0AFBFDFAFBFDFB0BFDFB0BFD0B0BFCFB0B0C0BFBFC0BFB0BFBFB0B0B0AFAFB0AFA0\
AFA09FAFA090A0908F9F8F7F80706070604F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000030304F40406F4F5070\
4F5070505F7F505F7F5F608F5060805F6F8F506F8F3F3F5F3F3F6F40408040508F
4F609F4F609F5070AF5070AF5F7FB05F80B0608FBF608FBF6F90C06F90BF6F9FC06F90BF709FC0\
709FBF70A0BF709FBF70A0BF70A0B070A0B06F9FA0709FA06F9F906F908F507050
6F5F3F6F5F3F705F3F6F5F3F70603F70603F7F603F7F603F7F6F407F603F806F407F6040806F4F\
806F40806F4F806F4F8F705080704F80705080704F8F70508070508F7F5F807050
8F7F5F8F7F5F8F7F5F8F7F5F9080608F7F609080608F7F6090806F8F806090806F90806F908F6F\
503F40604F507050608060708F6F7F90708090708F4F4F6F50608F6F709F707FA0
808FB08F90B09FA0C09FAFC0AFB0C03F5F7040607F406F7F4F708F4F708F507F905F7F9060809F\
60809F6F8FA06F8FA0709FB0709FB07F9FB07F9FB080A0BF80A0B080A0BF80A0BF
8FAFBF8FAFBF8FAFBF8FAFBF90B0C090AFC09FB0C09FB0C09FBFC09FB0C0A0BFC0A0BFC0AFC0CF\
AFBFC0AFC0CFAFC0CFB0C0CFB0C0CFB0C0CFB0C0C0BFCFCFB0C0C0BFCFC0BFC0C0
BFCFC0BFC0C0BFCFC0BFCFBFC0CFC0C0CFBFC0CFC0C0CFBFC0CFC0C0CFBFCFCFBFC0CFBFCFD0BF\
CFD0BFCFD0BFC0CFB0CFD0B0CFCFB0CFD0B0CFCFAFCFD0B0CFCFAFCFCFAFC0C0A0
CFCFA0C0C0A0C0C0A0C0BF9FC0C09FBFBF90C0BF90BFB090BFB090BFB08FBFB08FB0AF80BFAF80\
B0AF80B0AF80AFA07FB0A07FAF9F70AF9F70A0906FA0906FA09060A090609F8F5F
9F8F5F9080509080508F7F4F907F508F704F80704F7F6F407F6F40705F3F6F5F3F604F3F604F3F\
4F3F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F406F4F50705F608F6F709F7F80AF808FAF8F90BF
909FC09FAFCFA0AFD0AFBFDFAFBFD0B0BFD0AFB0CFB0BFCFB0B0C0BFB0BFB0B0B0B0B0B0B0AFA0\
B0A0A0A09F90A09F8F9F908090806F7F6F50000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000303050404F6F404F70\
4F507F4F507F506080506080506F8F506080506F8F3030503F3F6F3F407F40508F
405F904F609F4F6FA05070AF507FAF5F80B05F80B0608FBF608FBF6F90BF6090BF6F90BF6F90B0\
709FBF6F9FB070A0BF6F9FB070A0B06F9FA06F9F9F6090906F908F507F605F7050
605F3F6F60406F5F3F70603F70603F70603F70603F7F60407F603F7F6F407F6F40806F4F7F6F40\
80704F7F6F4F80704F80704F8070508070508F7F508070508F7F5F807F508F7F5F
8F7F5F8F80608F7F5F8F80608F7F608F80608F7F6090806F8F8060908F6F8F806F4F303F5F4040\
6F505F70506080607F8F6F7F90708F90708F9F80903F3F5F505F805F6F906F7FA0
7F80AF8F90B0909FBF9FAFC0A0AFBFAFB0C0AFBFC0B0C0CFB0CFCFC0D0CF305F7040607F40607F\
4F6F804F6F80507F8F507F905F809F5F809F6F8FA06F8FA07090AF7090AF7F9FB0
7F9FB080A0B07F9FB080A0BF80A0B08FAFBF80A0B08FAFBF8FAFBF90AFBF90AFBF90B0BF90B0BF\
9FB0C09FB0BFA0BFC0A0BFC0AFBFC0A0BFC0AFC0C0AFC0C0B0C0CFAFC0C0B0C0C0
B0C0C0B0CFC0B0C0BFBFCFC0B0C0BFBFCFC0BFC0BFBFCFBFBFC0BFC0CFBFBFCFBFC0CFBFBFCFBF\
C0D0BFBFCFB0C0CFB0C0CFB0C0CFB0BFCFAFC0CFB0C0CFAFC0CFAFC0C0A0C0CFAF
C0C0A0C0CFA0C0C0A0C0C0A0BFBF9FC0C09FBFBF90BFBF90BFB090BFBF90B0B08FB0B08FB0AF80\
B0AF8FAFA07FB0A080AFA07FAFA07FA09F70A09F70A0906FA0906F9F8F609F9060
908F5F9F8F5F9080509080508F7F508F7F5080704080704F7F6F407F6F406F5F3F6F5F3F605030\
604F302F3040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000404F70505F80606F90707FA0808FAF8F90B0
909FBF90A0C0A0B0D0AFB0DFB0BFDFAFBFD0B0BFD0B0B0CFB0BFCFB0B0BFBFBFBFB0B0B0B0AFAF\
AFA0A0B0A09FAF9F90A09F8F908F7F8F7F6070604F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F4060404F6F4F507F\
4F507F506080506080506F8F5060805F6F8F506F8F3F3F5F3F306F404080404F8F
405F90405F9F4F6FA04F70A05F7FB05F80B0608FBF608FBF6090BF608FB06F90BF6F90B06F9FBF\
6F9FB070A0BF6F9FB06FA0B06F9FA06F9FA06090906F908F5F80705F7F60606040
6F6F406F603F706F4070603F706F40706F3F7F6F407F6F407F6F4F7F6F4080704F7F704F807050\
7F704F807050807050807F508070508F7F5F807F5F8F7F5F807F5F8F80608F7F5F
8F80608F7F608F80608F80608F806F8F80608F8F6F8F806F908F6F8F806F503F4060404F705060\
7F5F6F8F6F808F7080907F8F907F8FA08F9FA08F9FAF90A04F4F70606F9F6F709F
7F80AF808FAF909FBF9FA0C0A0AFC0A0B0C0AFBFCFB0BFC0BFCFCFBFCFCFC0D0CFBFCFBFBFCFBF\
B0BFAFBFBFAFB0B09F3F607F40607F4F6F804F7080507F8F507F8F5F809F60809F
6F8FA06F8FA0709FAF709FAF7FA0B07F9FB07FA0B07FA0B080A0B080A0B08FAFBF8FAFB08FAFBF\
8FAFB090B0BF90B0BF9FB0BF90B0BF9FBFC09FB0BFA0BFC0A0BFBFAFC0C0AFBFC0
AFC0C0AFC0C0B0CFC0B0C0C0B0CFC0B0C0BFBFCFC0B0C0BFBFCFC0B0C0BFBFCFC0BFCFBFBFCFBF\
BFCFB0BFCFBFBFCFB0BFCFB0BFC0B0BFCFB0BFC0AFBFCFB0BFC0AFC0CFAFBFC0AF
BFC0AFBFC0A0C0C0AFBFC0A0C0C0A0BFC09FC0C0A0BFBF9FBFBF9FBFBF90BFBF90B0B090BFB090\
B0AF8FB0B08FAFAF80AFAF80AFA07FAFA07FA09F70A09F709F9F6FA09F6F9F906F
9F9060908F5F908F6090805F90805F8F7F508F805080704F80704F706F40706F406F603F6F5F3F\
6050305F4F3F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000303F604F507F5060806F709F7F80A0808FB0
8F90B0909FC09FA0CFA0B0D0AFB0DFAFBFDFAFB0D0B0BFCFB0B0C0B0B0C0B0B0BFBFB0BFB0AFAF\
B0AFAFAFA09FAFA09FA0908FA090808F806F80705F6F5F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030303F3F3F5F404F6F405070\
4F5F7F4F5F80506080506080506F8F506F805F6F8F506F803F3F6F3F407F404F8F
3F508F40609040609F4F70A04F70AF5F80B05F80B0608FB05F8FB06090B06090B06090B06090B0\
6F9FB06F9FAF6F9FAF6090A060909F60909060908F5F8F7F60806F5F7050606F4F
6060406F6F406F6F3F706F40706F40706F40706F407F704F7F6F407F704F7F704F8070507F704F\
8070507F7050807F5F807050807F5F807F5F807F5F807F5F807F60807F5F8F8060
807F608F80608080608F806F8080608F806F8F806F8F8F6F4F30305F4040604F507F5F60806070\
8F6F808F7080907F8F907F8FA08F9FA08F9FAF90A0AF90AF404060505F8060709F
6F7FA07F8FAF8F90B09FA0C09FA0BFA0AFC0A0B0C0B0BFC0B0C0C0BFCFCFBFCFC0BFCFC0B0C0B0\
BFC0B0B0BFA0BFBFA0AFAF90A0A0809F9070908F606F5F403F60703F607F406F80
406F804F708F50708F5F80905F80906F90A06F90A0709FAF6F90AF709FAF709FAF7FA0B07F9FAF\
80A0B080A0AF8FAFB080A0B08FAFB08FAFB090B0BF90AFB090B0BF90B0B09FBFBF
9FB0BFA0BFBFA0BFBFAFC0C0A0BFBFAFC0C0AFC0BFB0C0C0AFC0BFB0C0BFB0C0BFB0CFBFB0C0B0\
B0C0BFB0C0B0B0C0B0B0C0B0B0C0B0B0C0AFB0C0B0B0BFAFB0C0AFB0BFAFB0C0AF
B0BFA0B0C0A0B0BFA0BFC0A0B0BF9FB0BFA0B0BF9FBFBF9FB0BF9FBFBF9FB0BF90BFBF9FB0B090\
B0B090AFB08FB0B08FAFAF80AFAF80A0A07FA0A07F9F9F70A09F709F9F6F9F9F6F
909060909060908F5F908F5F8F80508F8050807F50807F507F704F7F704F706F406F6040605F3F\
605F3F5F5030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F4F6F505F80606F907F80A07F80AF
8F90B08F9FBF9FA0CF9FAFCFAFBFDFAFBFDFB0BFDFB0B0CFB0BFCFB0B0C0BFBFC0B0B0BFBFB0BF\
B0AFAFB0AFA0AFA09FAFA090A09F8F9F90808F7F607F6F50000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030303F404060404F6F4F5F7F\
505F5F50608F5060805F6F8F506F8F5F708F506F8F5F708F3F3F6F4040803F4F80
405F90405F904F6F9F4F6FA0507FAF5080B05F80B05F80B0608FB0608FB06090B06090B06F9FB0\
609FB06F9FB06090A06F9FA0609090609090608F80608F7F608060607F50606F4F
6F704F6F6F40707040706F407F704F70704F7F704F7F704F807F507F7050807F507F7050807F5F\
807F50807F5F807F5F808060807F5F808060807F5F8F80608080608F8060808060
8F8F6F80806F8F8F6F80806F8F8F708F8F6F8F8F70503F3F604F506F505F80606F8F6F7F907080\
90708F9F80909F8090A08F9FA090A0AF9FAFB09FAFB0A0B03F3F5F505F80606F90
707FA07F80AF8F90B0909FBFA0AFC0A0AFC0AFB0C0AFBFC0B0C0CFBFCFCFBFD0CFBFCFC0BFCFBF\
BFC0B0BFC0AFB0BFA0BFB09FAFA08FA09F7F908F6F807F5F7F908F80908F7F908F
80908F7F908F80908F3F6070406F804F6F7F50708F507F8F60809F608F9F6F90A06F90A0709FAF\
709FAF7FA0AF7F9FAF7FA0B07FA0AF80AFB080A0AF8FAFB08FAFB08FAFB08FAFB0
90B0BF90B0B09FB0BF9FB0B0A0BFBF9FBFBFA0BFBFA0BFBFAFC0C0AFC0BFAFC0C0AFC0BFB0CFBF\
AFC0BFB0C0BFAFC0B0B0C0BFAFC0B0B0C0B0AFBFAFB0C0B0AFBFAFB0C0AFAFBFAF
B0C0AFAFBFA0B0BFAFAFBFA0B0C0A0B0BFA0B0BFA0AFBF9FB0BFA0B0BF9FB0BF9FB0BF90B0BF9F\
B0BF90B0BF90AFB08FB0B08FAFAF80AFAF80A0A07FA0AF809FA07FA0A0709F9F70
9F9F7090906F90906F8F8F609090608F8F5F8F8F5F807F50807F507F704F70704F6F6F406F6F40\
605F3F5F5F3F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003040604050705F608F6F70907F80AF
808FAF8F90BF909FBF9FAFCFA0AFD0AFBFDFAFB0D0B0BFD0B0B0CFB0BFCFB0B0C0BFB0BFB0B0B0\
BFB0B0B0AFA0B0A0A0AF9F90AF9F90A0908090807080705F706040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030304F3F3F60404F704F507F\
5060804F6080506F80506080506F8F506F8050708F506F803F3F6F3F3F703F4F80
3F508F405F904060904F6FA04F70A0507FAF507FAF5F80B05F80AF5F8FB05F8FAF6090AF6090AF\
609FAF6090A06090A06090906090905F8F80608F805F8070607F605F704F6F704F
6F704F70704F6F6F4070704F70704F7F7F5070704F7F7F507F70507F7F507F70507F7F5F7F7F5F\
807F5F7F7F5F80805F807F5F808060807F5F80806080806080806F80806080806F
80806F808F6F80806F808F6F80806F8F8F70503F406F4F507050608060708F6F7F907080907F8F\
9F80909F8090A08F9FA090A0B09FAFB09FAFB0A0B0B0AFB0B0B0BF404F6F5F6F90
60709F7080AF7F8FAF909FBF9FA0BFA0AFC0A0AFBFAFBFC0AFBFC0B0C0CFBFCFC0BFCFC0B0C0BF\
BFC0BFB0BFAFB0BFA0B0B09FAFAF909F9F7F90906F8F805F504F3F6F807F6F807F
6F807F6F80806F807F6F80806F807F6F80806F7F7F3F506F3F5F6F40607F4F607F507080507F8F\
5F809060809F6F90A06F90A0709FAF709FA0709FAF709FAF7FA0AF7FA0AF80A0AF
80A0AF8FAFB080AFAF8FAFB08FAFB090B0B090B0B09FB0B090B0B09FBFBF9FBFB0A0BFBFA0BFBF\
AFC0BFA0BFB0AFC0BFA0BFB0AFC0B0A0BFAFAFBFB0A0BFAFAFC0B0A0BFAFAFBFAF
AFBFA0AFBFAFA0BFA0AFBFA0AFBFA0AFBFA0AFB09FAFBF9FAFB09FAFBF9FAFB090AFBF9FAFB090\
AFBF90A0B08FAFB090A0AF8FAFB08FA0AF80A0AF809FA07FA0A07F9F9F709FA070
909F6F909F6F8F90609090608F8F608F8F6080805F80805F7F7F507F7F506F704F6F6F4F60603F\
60603F5F5030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000404F704F5F7F606F90707FA0
808FAF808FB0909FBF90A0C0A0AFCFA0B0D0B0BFDFAFBFD0B0BFD0B0B0CFBFBFCFB0B0C0BFB0BF\
B0B0B0BFB0B0B0AFA0B0A09FAF9F90AF9F8F9F8F7F8F7F6F7F6F4F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030304F40406F404F704F5F7F\
506080506F8F506F80506F8F506F805F708F507080507080506F7F3F407F3F407F
3F508F3F5F8F40609F406F9F4F70A04F7FA05080AF5080AF5F8FAF5F8FAF6090AF5F8FAF6090AF\
6090A0609FA060909F6090905F8F8F60908F5F8F7F60806F5F5040607F5F607050
6F7F506F704F707F50707050707F507070507F7F5F7F7F507F7F5F7F7F5F80805F7F7F5F808060\
7F7F5F808060808060808060808060808F6F808060808F6F80806F808F6F80806F
808F6F808F6F808F70808F6F503F3F5F404F70505F7F5F6F8F6F7F8F6F7F907F8F9F7F8FA08090\
A08F9FAF90A0AF90A0B0A0B0B0A0B0B0AFBFB0AFBFB0B0C0B0B0C0404F6F505F8F
60709F6F7FA0808FB08F90B09FA0C09FAFBFA0B0C0A0B0C0B0BFC0B0C0C0BFCFCFBFCFC0BFCFC0\
B0C0B0BFC0B0B0BFAFB0BFA0AFAF90A0A0809F9070908F60605F405F7F705F7070
5F70705070705F70705070705F70705070705F70705070705F70705070705F7070506F6F3F506F\
3F5F6F4F607F4F6F7F507080507F8F608090608F9F6F90A06F90A0709FAF709FA0
7FA0AF709FAF7FA0AF7FA0AF80AFAF80A0AF8FAFB08FAFAF8FB0B08FAFAF90B0B090B0B09FBFB0\
90B0B09FBFB09FBFB0A0BFB0A0BFB0A0C0B0A0BFB0A0BFB0A0BFAFA0C0B0A0BFAF
A0BFAFA0BFA0AFBFAFA0BFA0A0BFA0A0BFA0AFBFA0A0BF9FAFBFA0A0B09FAFBF9FA0B09FAFBF9F\
A0B090A0B090A0B08FA0B08F9FAF8FA0AF8F9FAF809FAF809FA07F9FA07F90A07F
9FA07F909F70909F708F906F8F906F8F8F608F906080805F7F805F707F50707F506F6F406F6F40\
5F603F5F603F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F304F405070505F806F709F
707FA0808FAF8090B0909FC090A0C0A0B0D0A0B0DFAFBFDFAFB0CFB0BFCFB0B0C0B0BFC0B0B0BF\
BFB0BFB0AFAFB0AFAFAFA09FAFA09FA09F8FA09F8F90807080705F705F40000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030305F3F40604F50704F507F\
506080506080506F8F506F80506060506F8050708F506F7F50707F506F703F4F7F
304F803F5F8F3F5F90406F9F406F9F4F70A04F7FA05080AF5080A05F8FA0508FA05F8FAF5F8FA0\
6090A05F8F9F60909F5F8F8F608F8F5F8F7F6080705F7F5F607F5F607F506F7F5F
6070506F7F506F7050707F5F707050707F5F707F5F7F7F5F707F5F7F7F5F7F7F5F7F80607F7F5F\
7F80607F80607F80607F80607F80607F8060808F6F7F8060808F6F7F806F808F6F
7F806F7F8F703F30305F3F4060404F70506080606F8F6F7F8F7080907F8F9F7F8FA08F9FA08F9F\
AF90A0AF90A0B0A0B0B0A0B0B0AFBFB0AFBFB0B0C0B0B0C0B0BFCF3F3050505F7F
5F6F906F7FA07080A08F90B0909FBF9FAFC09FAFBFA0B0C0A0B0BFB0C0C0B0C0C0BFCFCFB0C0BF\
BFC0BFB0C0B0B0BFAFB0B0A0B0B09FA0A0809F9F7F8F8F607F7050000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F506F3F5F6F4F607F4F6F7F507080507F8F5F8090
608F906F90A06F90A0709FA0709FA070A0A0709FA07FA0AF7FA0A080A0AF80A0A080AFAF80A0A0\
8FAFAF8FAFAF8FB0AF8FAFAF90B0AF90B0AF9FB0AF90B0AF9FBFAF9FBFAFA0BFAF
9FB0A09FBFAF9FB0A0A0BFAF9FB0A0A0BFA09FB09FA0BFA09FB09FA0B09F9FB09FA0B09F9FB090\
9FB0909FAF8F9FB09090AF8F9FAF8F90A0809FAF8090A07F90A07F909F7F90A07F
8F9F708F9F708F906F8F9F6F8090608F906F808F60808F60708050707F506F704F6F704F606F3F\
5F603F505F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F4F6F505F805F6F8F
707FA07F80AF8F90B08F90BF9FA0C09FAFCFAFB0DFAFBFDFB0BFDFB0BFD0B0BFCFB0B0C0BFBFC0\
BFB0BFBFB0BFB0AFAFB0AFAFAFA09FAFA09FA09F8FA09080907F6F80705F706040
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000303F5F404F6F4F5070506080\
5060805F6F8F506F805F708F5F708F5F708F5070805F708050707F4F5F60506F70
3F50803F508F3F60904060904F70A04F709F507FA04F7FA05080A05080A05F8FA05F8FA05F90A0\
5F8F9F60909F5F8F906090905F8F80608F805F7F6F60806F607F606F8060607F5F
6F7F5F6F7F5F707F5F707F5F708060707F5F708060707F5F7F80607080607F80607080607F806F\
7F80607F806F7F80607F806F7F806F7F8F6F7F806F7F8F707F806F7F8F707F806F
403F3F4F3F40604F506F4F507F606F8F6F709070809070809F80909F8090A08F9FA08F9FAF9FAF\
B09FAFB0A0B0B0A0B0BFB0BFB0B0BFBFBFC0B0B0C0BFC0CFBFC0CFBFC0D04F4F6F
606F906F709F7F80AF808FAF909FBF9FA0C0A0AFC0A0AFC0AFBFC0AFBFC0B0C0CFBFCFC0BFCFC0\
BFC0BFBFC0BFB0BFAFBFBFAFB0B09FAFAF909F9F7F9F90708F805F50403F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F5F6F405F6F
4F6F7F4F6F7F507F805F7F8F608F90608F9F6F9FA06F9FA0709FA0709FA070A0A0709FA07FA0A0\
7FA0A07FA0A07FA0A080AFA080AFA08FAFA080AFA08FB0AF8FAFA090B0AF8FB0A0
90B0AF90B0A09FBFAF9FB0A09FBFAF9FB0A09FBFA09FB0A09FBFA09FB09F9FBF9F9FB09F9FB09F\
9FB0909FB09090AF8F9FAF9090AF8F90AF8F8FA08090A0808FA07F8FA07F8F9F70
8FA07F8F9F708F9F7080906F8F9F6F80906F8090607F8F5F7F8F5F708050707F50607040607040\
5F6F3F5F603F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000303040404F70506080
6070907F80AF7F8FAF8F90B08F9FBF9FA0CF9FAFCFAFBFDFAFBFD0B0BFD0AFB0CFB0BFCFB0B0C0\
BFB0C0B0B0B0BFB0B0B0AFAFB0AFA0AF9F90AF9F90A09080A090808F7F6F80705F
706040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003F3F5F3F40604F507F4F5F7F\
5060805060805F6F8F506F805F708F5070805F708050707F5F707F506F70507070
506F6F3F508F3F5F8F406090406F904F709F40709F4F7F9F4F7F9F5080A050809F5F8FA0508F9F\
5F8F9F5F808F5F8F8F5F8F805F8F805F807060806F5F7F6F60806F607F606F7F60
6F7F5F6F80606F7F5F707F606F7F5F7080606F7F5F708060707F60708060707F6070806F708060\
70806F70806070806F70806F60504070806F70806F70806F70807070806F4F3F40
5F404F70505F7F506080606F8F6F7F9070809070809F80909F8090A08F9FA090A0AF9FAFB09FAF\
B0A0B0B0A0B0B0AFBFB0B0BFBFBFC0B0BFC0BFBFCFB0C0CFBFCFD0B0C0D0404060
505F8060709F6F7FA07F8FAF8F90B09FA0C09FA0BFA0AFC0A0B0C0AFBFC0AFBFC0BFCFCFB0C0C0\
BFCFC0B0C0B0B0C0B0B0BFA0B0BFA0AFAF90A0A0809090708F8F60605F40000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F506F405F6F4F6F7F4F6F7F507F805F808F608F9F60909F6F909F60909F\
6F9F9F6F909F709F9F709F9F709F9F709F9F7FA09F7FA09F7FA09F7FA09F80A09F
80A09F8FAFA080AF9F8FAFA08FAF9F90B0A08FAF9F90B0A090B09F9FB0A090B09F90B09F90AF90\
90B0908FAF8F90AF908FA08F8FAF8F8FA0808FA080809F7F8FA080809F7F809F7F
809F70809F707F906F80906F7F906F7F906F708F60708F5F6F7F506F7F5060704F60704F5F6F40\
5F6F3F405030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F4F704F5F7F
606F906F7F9F808FAF808FB08F9FBF909FC0A0AFCFA0B0D0AFBFDFAFBFD0B0BFD0B0B0CFB0BFCF\
B0B0C0BFBFC0BFB0B0BFB0B0B0AFA0B0AFA0AF9F90AF9F90A09080A0908090806F
8F705F70604F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000303F5F404F6F4F5070506080\
5060805F6F8F5F6F8F5F708F5F708F607F8F5F70805F7F805F707F5F7F7F507070
5F706F506F603F5F8F3F6090406F90406F904F709F4F709F4F7F9F4F7F9F50809F50809F5F8F9F\
5080905F8F905F80805F8F805F807F60807F5F807060807060806F6F806F607F60
6F80606F7F606F80606F7F607080606F7F6070806F6F7F6070806F6F806070806F70806F70806F\
6F806F70806F70806F70806F6F806F7080706F806F7080706F806F604F506F505F
7F5F6F805F6F8F6F7F8F6F7F907F8F907F8FA08090A08F9FAF90A0AF90A0B0A0AFB0A0B0B0AFBF\
B0AFBFBFB0C0B0B0C0BFBFCFB0BFCFBFC0D0BFC0D0BFCFDFBFCFD0BFCFD03F3F50
505F7F5F6F90707FA07F80AF8F90B0909FBF9FAFC0A0AFC0A0B0C0AFB0C0B0C0C0B0C0C0BFCFCF\
B0C0BFBFC0BFB0C0B0B0BFAFB0B0A0B0B09FA0A0809F9F7F8F8F607F7050000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F50603F506F4F60704F6F7F507F80\
50808F608F9060909060909F6090906F9F9F6F9090709F9F6F9F90709F9F709F90
7FA09F709F907FA09F7FA09F80A09F7FA09F80AF9F80AF9F8FAF9F8FAF9F90B09F8FAF9F90B09F\
8FAF908FAF908FAF8F8FAF8F80A08F8FA08F80A08080A080809F7F80A0807F9F7F
809F7F7F9070809F707F906F7F906F708F60708F606F805F6F805F607F50607F505F6F405F6F40\
50603F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000405070
505F806070907080A0808FAF8090B0909FBF90A0C0A0AFD0A0B0D0AFBFDFAFB0D0B0BFCFAFB0C0\
B0B0C0B0B0BFBFB0BFB0AFB0B0AFAFB0A0A0B0A09FAF9F90AF9F90A09080A09080
907F6F8F705F70604F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003F3F60404F6F4F5F7F4F5F7F\
506080506F805F708F5F708F5F708F5F7080607F8F5F70805F70805F707F5F7070
506F6F506F605060503F60903F608F406F90406F904F70904070904F7F9F4F7F9050809F4F808F\
50808F5080805F808050807F5F807F5F80705F80705F7F6F60806F607F6060806F
607F6060806F607F606F806F607F606F806F6F7F606F806F6F7F606F806F6F7F6F6F806F6F7F6F\
6F806F6F7F6F6F806F6F7F6F6F806F607F6F6F806F607F6F6F505F7F506080606F
8060708F6F7F8F707F907F8F9F7F8FA08090A08F9FAF90A0AF90A0B0A0AFB0A0B0B0AFBFB0AFBF\
B0B0C0B0B0C0BFBFCFB0BFCFBFC0D0B0C0D0BFCFDFB0C0D0B0C0D0AFBFCFAFBFCF
404F6F5F609060709F7080AF7F8FAF8F9FBF90A0BFA0AFC09FAFBFAFB0C0AFBFC0B0C0C0B0C0C0\
BFCFC0B0C0BFB0C0B0B0BFAFB0BFA0AFB09FAFAF8F9F9F7F90906F80805F4F403F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F50603F5060406070406F7050707F507F805F8F905F8F8F6090906090906F9090
60908F6F90906F9090709F906F9F90709F90709F907FA090709F907FA0907FA0907FA0907F9F8F\
7FA0907FA08F80A08F7FA08F80A08F7F9F8080A0807F9F7F7F9F807F9F7F7F9F7F
7090707F9070709070709070708F6F6F8F6060805F60805F5F7F505F7050506F4F506F4F4F603F\
3F4F3F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F406F
4F5F805F6F8F707FA07F80AF8090B08F9FBF90A0C09FAFCFA0B0D0AFBFDFB0BFDFAFBFCFB0BFCF\
B0B0C0BFBFC0B0B0BFBFBFBFBFB0B0BFB0AFB0A0A0B0AFA0AF9F90AF9F90A0908F
A090809080708F70607F604F705F3F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003F3F5F4050704F5F7F506080\
5060805F708F5F708F607F905F708F607F8F607F80607F8F5F7F80607F7F5F7070
5F70705F6F605060504F50403F5F8F3F608F406F904070904F7F904F7F905080904F7F9050808F\
50808F50808F5080805F80805F807F5F807F5F80706080705F7F6F60806F607F6F
60806F607F6F60806F607F6F6F806F607F6F6F806F60806F6F806F60806F6F806F607F6F6F8070\
607F6F607F6F5F7F6F608070607F6F607F6F6F4F5F805F6F805F6F8F6F708F6F7F
907080907F809F80909F8090AF909FA08F9FAF9FA0AF9FAFBFA0B0B0A0B0BFAFBFB0AFBFBFBFC0\
B0BFC0BFC0CFB0C0CFBFCFD0BFCFD0BFCFDFB0C0D0B0C0D0AFBFCFAFBFC0A0B0BF
404060505F8060709F6F7FA0808FB08F90B09FA0C09FAFBFA0B0C0A0B0C0AFBFC0AFBFC0BFCFCF\
BFC0C0BFC0BFB0C0B0B0C0B0B0BFA0B0BFA0AFAF90A0A0809090708F8F605F5040
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F5060405F6F4F60704F6F70507F7F507F80
5F8F905F8F8F609090608F8F6F909060908F6F9F906F908F6F9F906F908F6F9F8F6F908F709F8F\
6F908F709F8F6F9080709F8F709F80709F807090807F9F80709F7F7F9F8070907F
7F9F7F7090707090706F906F70906F6080606080605F7F5F5F7F5F5070505070504F6F404F6F40\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F4F70506080606F907080A07F8FAF8F90BF8F9FBF9FA0C09FAFCFAFBFDFAFB0D0AFBFD0AFB0CF\
B0BFCFB0B0C0B0B0C0B0B0BFBFB0BFB0AFB0B0AFAFB0A0A0B0A0A0AF9F90AF9F90
A09080A090809080708F70607F604F705F3F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003F3F5F404F6F4F5F7F4F607F\
5F6F805F6F805F708F5F708F607F8F5F708F607F8F607F80607F805F707F607070
5F706F5F706F50604F505F400000003F507F305F7F3F60803F6F8040708F40708F4F7F8F4F7F8F\
507F8F4F7F80507F80507F7F50807F507F705F80705F7F705F7F705F7F6F5F7F70
5F7F6F6080705F7F6F607F6F5F7F6F607F6F5F7F6F5F7F6F5F7F6F5F7F6F5F706F5F7F6F50706F\
5F706F402F2F50706F4F6F606F505F7F506080606F8060708F6F7F8F6F7F907F8F
9F7F8FA08090A08F90AF90A0AF90A0B09FAFB09FAFB0A0B0B0A0B0BFB0BFB0AFBFBFBFC0B0BFC0\
BFC0CFB0C0CFBFCFD0BFCFD0BFCFDFB0C0CFB0C0CFAFBFC0AFBFC0A0B0BFA0B0BF
30305050507F5F6F906F7FA07080A08090B08F9FBF9FAFC09FAFBFA0B0C0A0B0BFAFBFC0B0C0C0\
BFCFC0B0C0BFB0C0BFB0BFAFB0BFAFAFB09FAFB09FA0A0809F9F7F8F8F60707050
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F5060
3F5F604F6070406F7050707F507F805F8F8F5080805F8F8F5F8F805F8F805F8F805F8F805F8F7F\
608F805F8F7F608F7F5F8F7F608F7F608F7F608F7F608F70608F7F608F706F8F70
608F706F9070608F6F6F8F7060806F60806F5F7F5F5F7F5F5070505070504F6F4F4F6F4F304F30\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F4F6F40507F5F6F906F7F9F7F8FAF808FB08F9FBF90A0C09FAFCFA0B0D0AFBFDFAFBFD0B0BFD0\
B0BFCFB0BFCFB0B0C0BFBFC0B0B0BFBFB0BFBFB0B0BFB0AFB0A0A0B0A0A0AF9F90
AF9F90A0908FAF90809F80708F7F607F604F705F40000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000303050404F704F5F7F506080\
506F805F708F5F708F607F90607F8F607F8F607F8F60808F607F80607F80607F70
607F705F70605F6F5F505F4F0000000000003F4F703050703F60803F608040708F40708F4F7F8F\
4070804F7F804F7F7F507F80507F7F50807F507F705F7F70507F705F7F70507F70
5F7F70507F6F5F7F70507F6F5F7F6F50706F5F7F7050706F50706F4F706F50706F4F6F6F4F6F6F\
406060406F603F6060805F60805F6F8F6F708F6F7F9070809070809F8090A08090
AF8F9FAF909FB09FAFB09FAFB0A0B0B0A0B0BFAFBFB0AFBFBFB0BFB0B0BFBFBFCFB0BFCFBFC0D0\
BFC0D0BFCFDFBFCFDFBFCFD0B0C0CFB0C0CFAFBFC0A0BFC0A0B0BFA0B0BF9FAFAF
9FAFAF404F6F5F6F906F709F7080AF7F8FAF909FBF9FA0BFA0AFC0A0AFBFAFBFC0AFBFC0B0C0C0\
B0C0C0BFCFC0B0C0BFB0C0B0B0BFAFB0BFAFAFB09FAFAF8F9F9F7F90906F80805F
303040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F4F5F3F505F405F6F40606F4F70704F7F7F50808050807F508080\
50807F50807F50807F5F807F5080705F8F7F5F80705F807F5F80705F8F705F8070
5F8F705F806F5F806F507F60507F6050705F507F5F4F70504F7050406F4F406F4F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000404F704F5F80607090707FA07F8FAF8090B08F9FBF90A0C09FAFCFA0B0D0AFBFDFAFBFD0\
B0BFD0AFBFCFB0BFCFB0B0C0B0B0C0B0B0BFBFB0BFB0AFB0BFAFAFB0A0A0B0A0A0
AF9F90AF9F90A0908FA090809F80708F7F607F604F705F40000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030304F3F40604F5F7F4F607F\
5F6F805F6F8F607F905F708F607F8F607F8F6F808F607F80608080607F7F607F70
60706F6070605F60505F604F3F302F000000000000000000304F6F3F5070305F703F6F803F6F7F\
40707F406F7F40707F4070704F30304F70704F7F704F70704F7F704F7070507F70
4F706F5070704F706F4F706F4F6F6F4F706F406F6F4F6F6F4060603F505F2F2F2F4F3F405F4040\
6F50507F506080606F80606F8F6F7F8F707F9F7F809F7F8FA08090A08F90AF90A0
AF90A0B09FAFB09FAFB0A0B0B0A0B0BFAFBFB0AFBFBFB0C0B0B0BFBFBFC0B0BFCFBFC0D0B0C0CF\
BFCFD0B0CFD0B0CFD0AFBFCFAFBFCFA0B0C0A0BFBF9FB0B09FB0B090AFAF90AFA0
809F903F4060505F8060709F6F7F9F7F8FAF8090B090A0BF9FA0BFA0AFC0A0B0BFAFBFC0AFBFBF\
B0C0C0B0C0BFB0C0BFB0BFB0B0BFAFAFB0A0B0B0A0A0AF8F9FA08090906F8F8F60
5F503F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000304F5F3F50603F5F5F\
406060406F6F4F7F70407F704F7F704F7F704F7F704F7F6F4F7F704F7F6F507F70
4F70604F70604070604070603F6F50406F5F3F6050406F503F604F3F604F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030406F4F5F7F5F608F6F7FA07F80AF8090B08F90BF90A0C090A0C0A0B0CFA0B0D0AFC0DF\
AFBFD0B0BFD0B0BFCFB0BFCFB0BFC0BFBFC0B0B0BFBFBFBFBFB0B0BFB0B0B0AFA0
B0AFA0AF9F9FAF9F90A0908FAF908F9F8F70907F6F806F507F5F40000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F4F404F6F4F5F7F506080\
5F6F8060708F60708F607F90607F8F6F808F607F8F6F808F6F80806F8080607F70
6F7F7060706F60705F5F604F40403F0000000000000000000000000000000000000000003F607F\
3F6F7F3F6F7F3F6070406F7F3F6070406F703F606F406F7040606F40606F40606F
403030304F5F0000000000000000000000000000000000000000004F3F3F5F4F4F6F4F50805F60\
805F6F8F6F708F6F70907080907F809F808F9F8090AF909FAF909FB09FAFB09FAF
B0A0B0B0A0B0BFAFBFBFAFBFC0B0C0BFB0C0C0BFD0B0BFC0BFC0CFBFC0CFBFCFD0BFCFD0BFD0DF\
B0CFD0B0CFD0AFC0CFAFC0CFA0BFC0A0BFBF9FB0B09FB0B090AFAF8FA0A07F9090
70908F3030504F507F5F6F90707FA07F80AF8090B0909FBF9FAFC09FAFC0A0B0C0A0B0C0B0BFC0\
B0C0C0BFCFC0B0C0BFB0C0BFB0BFAFB0BFAFAFB0A0AFB09FA0A0809F9F7F8F8F60
6F604F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000304F5F3F50603F50604060603F606040706F40706F40706F406F60
40706F3F6F60406F603F605F406F5F3F60503F60503F5F4F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F4F70505F805F6F90707FA07F80AF8090B08090B090A0C090A0C09FAFCFA0B0D0\
AFBFDFAFBFD0B0BFD0AFBFCFB0BFCFB0B0C0B0B0C0B0B0BFBFB0BFBFAFB0BFAFAF
B0A0A0B0A0A0AF9F90AF9F90A0908FAF90809F8070907F6F806F507F5F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F3F5F4F5F7F50607F\
5F6F8F5F6F8F607F90607F8F6F8090607F8F6F808F6F808F6F808F6F7F7F6F807F
6F7F706F7F70606F5F5F6F504F4F3F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F3F3F5F40406F50507F505F80606F80606F\
8F6F7F8F6F7F907F809F7F8FA08090A08F90AF90A0AF90A0B09FAFB09FAFB0A0B0
B0AFB0BFB0BFBFAFBFC0B0CFBFBFCFC0C0D0C0BFD0BFC0CFB0C0CFBFCFD0B0CFD0BFCFDFB0C0D0\
B0C0CFAFBFC0AFBFC0A0B0BFA0B0BF9FAFB09FAFAF8FA0A0809F9F70908F6F8F80
5F7F7000000040406F5F609060709F7080AF7F8FAF8F9FBF90A0BF9FAFC09FAFBFA0B0C0AFB0C0\
B0C0C0B0C0C0B0C0BFB0BFB0B0BFB0AFB0A0B0BFA0AFB09FA0AF8F909F7F90906F
6F604F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F2F2F403030000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F50705060805F6F907080A07F8FAF8090B08090B090A0C090A0C0A0B0CF\
A0B0CFAFC0DFAFBFD0B0BFD0B0BFCFB0BFCFB0B0C0BFBFC0B0B0BFBFB0BFBFB0B0
BFB0B0B0AFA0B0AFA0AF9F9FAF9F90AF908FAF908F9F8F7F907F6F806F507F604F70503F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F405F405070506080\
506F8060708F607F906F8090607F8F6F80906F808F708F906F80807080806F807F
6F807F6F7F706F7060606F505F5F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000403F3F503F406F4F50704F507F5F60805F608F6F708F6F70907080\
9070809F808FA08090AF8F9FAF909FB09FAFB09FAFB0A0B0B0A0B0BFAFBFBFAFBF
C0B0C0BFB0C0C0BFCFC0BFCFCFC0DFC0C0DFCFCFE0C0C0DFBFCFD0BFCFD0BFCFDFB0C0CFB0C0CF\
AFBFC0AFBFC0A0B0BF9FB0BF9FAFAF9FAFAF8FA09F809F90708F806F8F80000000
000000000000404060505F8060709F6F7FA0808FAF8F90B090A0C09FA0BFA0B0C0A0B0C0AFBFC0\
AFBFC0B0C0C0B0C0C0B0C0BFB0BFB0B0BFAFAFB0A0B0B0A0A0AF8F9FA08090906F
707050000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F4F704F5F805F6F8F6F7FA07080A08090B08090B0909FBF90A0C0\
9FAFCFA0B0CFAFBFDFAFBFD0B0BFD0AFBFCFB0BFCFB0B0C0B0B0C0B0B0BFBFB0BF
B0AFB0BFAFB0B0A0A0B0A0A0AF9F90AF9F90A0908FAF90809F807F907F6F8F705F806F507F5F40\
705F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000303F504050704F5F7F\
5F6F805F6F8F607F90607F8F6F80906F808F7080906F808F70808F6F8080708080
6F7F7F708070607060606F5F5F604F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F30403F3F503F406F4F4F704F507F505F7F505F805F6080606F8F6F7090707F9F7F809F7F80\
A08090A08F90AF90A0AF90A0B09FAFB09FAFBFA0B0B0A0B0BFAFBFBFB0BFC0BFC0
BFB0C0C0C0D0C0C0D0CFCFDFC0C0DFC0CFDFBFC0D0C0C0D0B0CFD0B0CFD0AFC0CFAFC0CFA0BFBF\
A0BFBF9FB0B09FB0B090AFAF90A0A0809F907F90906F808060807F000000000000
00000000000030304F4F50705F6F906F7FA07080A08090B08F9FB09FA0C09FAFBFA0B0C0A0B0BF\
AFBFC0AFBFC0B0C0C0B0BFBFB0BFB0AFBFAFB0BFAFAFB09FA0AF909F9F7F909070
707050000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F50705060805F6F907080A07F8FAF8090B08090B090A0C0\
90A0C09FB0CFA0B0CFAFC0DFAFBFD0B0BFD0B0BFCFB0BFCFB0BFC0BFBFC0B0B0BF
BFBFBFBFB0B0BFB0B0B0AFAFB0AFA0B0A09FB09F90AF908FAF908FA08F80A08F7F907F6F907F60\
806F5F806050705F4070503F000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F3F50404F6F506080\
5F6F8060708F60708F6F80906F80907080906F8090708F90708F8F708F8F708080
708F8070807F707F6F60705F606F50403F30000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003030304F3F405F4040
6F4F50704F507F505F7F5F5F80606F80606F8F6F708F6F70907080907F809F808FA0808FAF8F9F\
AF909FB09FA0B09FA0BFA0B0B0A0B0BFAFBFBFAFBFC0B0C0C0B0C0C0BFCFC0BFCF
CFCFD0CFCFDFCFCFE0C0C0DFC0CFDFBFC0D0BFC0D0B0BFCFBFC0CFAFC0CFAFC0CFA0BFBFA0BFBF\
9FB0B09FB0B090AFAF8FA0A07F9F9070908F60807F5F8070000000000000000000
000000000000000000404F6F5F60906F709F7080AF7F8FAF909FBF9FA0BF9FAFC0A0AFBFAFBFC0\
AFBFC0B0C0C0B0C0C0B0C0C0B0BFB0B0BFB0AFBFAFB0BFA0A0AF909FA080909070
808060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F4F704F5F805F6F8F6F7FA07080AF8090B08090B0\
909FBF90A0BF9FAFCFA0B0CFAFBFD0AFBFD0B0BFD0AFBFCFB0BFCFB0B0C0B0BFC0
B0B0BFBFB0BFBFAFB0BFB0B0B0AFA0B0A0A0AF9F9FAF9F90AF908FAF908FA08F80A08F809F8070\
9F80708F70608F6F5F7F604F7F5F406F4F3F000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000030304F404F6F4F5F7F\
5F6F805F6F8F607F90607F906F80906F8090708F90708F90708F9070808F708F8F
70807F70807F6F7F6F6F7060606F50504F3F000000000000000000000000000000000000000000\
0000000000000000000000000000003030304F3F3F5F40406F4F4F704F507F505F
7F5F5F80606080606F8F6F708F6F7090707F90707F9F7F809F808FA08090A08F90AF909FAF90A0\
B09FAFB09FAFBFA0B0BFAFB0BFB0BFBFB0BFC0BFC0C0BFC0CFC0CFC0C0CFCFCFDF
CFCFDFCFCFDFBFC0D0BFC0D0BFBFCFBFC0D0B0BFC0B0BFCFAFBFC0AFC0C0A0B0BFA0B0BF9FAFB0\
9FAFB08FA0A0809F9F70908F6F8F805F7F70000000000000000000000000000000
000000000000000000404F6F505F8F60709F6F7F9F7F8FAF70706F90A0BF90A0BF9FAFC0A0B0C0\
AFBFC0AFBFC0B0C0C0AFBFBFB0BFBFAFBFAFB0BFAFAFB0A0A0AF909F9F80909070
808060504F3F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F50704F60805F6F8F6F7FA07F8FAF8090B0\
8090B090A0BF90A0C09FAFCFA0B0CFAFBFDFAFBFD0B0BFD0AFBFCFB0BFCFB0BFC0
BFBFC0B0B0BFBFB0BFBFB0B0BFB0B0B0AFAFB0AFA0B0A09FB0A09FAF9F90AF9F90AF908FAF908F\
A08F80A08F7F9F7F70907F6F8F6F5F806F507F5F40705040000000000000000000
00000000000000000000000000000000000000000000000000000000000030304F3F40604F5F7F\
50608060708F6070906F80906F8090708F9F708F907F8F90708F907F8F907F8F8F
7F8F8070807F7F807F6F70606F705F505040000000000000000000000000000000000000000000\
0000000000003F30304F3F405F4040704F507050507F5F5F7F5F5F80606F80606F
8F6F708F6F7090707F907F809F808F9F808FAF8F90A08F90AF909FAF909FB09FA0B09FA0B0A0AF\
B0A0AFBFAFBFBFAFBFC0B0C0C0B0C0CFC0CFC0C0CFCFC0D0CFC0D0D0D0DFC0CFDF
C0CFDFBFC0D0BFC0D0B0BFCFB0C0CFB0BFC0B0BFC0AFBFC0AFBFBFA0B0BFA0B0BF9FAFB090AFAF\
8FA09F809F9F708F8F6F80805F7F70000000000000000000000000000000000000
00000000000000000040405F50608F5F6F90707FA07F80AF8F9FBF8F9FBF9FAFC09FAFBFA0B0C0\
A0B0C0B0C0CFAFBFC0B0C0BFAFBFB0B0BFB0AFBFAFB0BFA0A0AF90A0A08F909070
8F8F6F50503F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F4F704F5F805F6F8F6F7F9F7080A0\
8090B08090B08F9FBF90A0BF9FAFCF9FB0CFAFBFD0AFBFD0AFBFD0AFBFCFB0BFCF
B0B0C0B0B0C0B0B0BFBFB0BFBFAFB0BFB0B0B0AFA0B0AFA0B0A09FB0A09FAF9F90AF9F90AF908F\
AF908FA08F80AF90809F807F9F80709070608F70608060507F5F4F70503F000000
0000000000000000000000000000000000000000000000000000000000000000003F4060405070\
5060805F6F80607F90607F906F80907080907F8F9F708F907F8F907F8F8F7F8F8F
7F8F807F8F80708070707F6F6F705F60604F404030000000000000000000000000000000303030\
4F3F3F5F4040704F4F704F507F5F5F7F5F5F80606080606F8F6F708F6F7090707F
90707F9F7F809F808FA08F90A08F90AF8F9FAF8F9FB090A0AF909FB09FA0B09FAFBFA0B0B0A0B0\
BFAFBFBFB0BFC0BFC0C0BFC0C0C0CFC0C0CFCFCFD0CFCFD0CFCFDFC0C0D0C0C0D0
B0BFCFB0C0CFB0BFCFB0BFCFAFB0C0AFBFC0A0B0BFA0B0BF9FAFB09FAFB090AFAF90AFAF809F9F\
7F90906F808060807F000000000000000000000000000000000000000000000000
00000000000000000040405F4F50805F6F906F709F7F80AF8090B0909FBF90A0BF9FAFC09FAFBF\
A0B0C0AFBFC0B0C0C0AFBFBFAFBFBFAFBFAFAFBFAFAFB0A0A0AF9F9FA080909F7F
808F605F5F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003F50704F60805F6F8F6F7FA0\
7F8FAF8090B08090B090A0BF90A0C09FAFCFA0B0CFAFBFD0AFBFD0B0C0D0AFBFCF
B0BFCFB0BFC0BFBFC0B0B0BFBFBFBFBFB0B0BFB0B0BFAFAFBFAFAFB0A0A0B0A0A0B0A09FB0A09F\
AF9F90AF9F90AF908FAF908FA08F80AF8F809F80709F7F6F8F70608F6F5F7F5F4F
7F5F4070503F70503F60403F4F403F0000000000000000000000000000000000003F3F504F5F7F\
50607F5F6F8F60708F6F809F6F8090708F9F708F907F909F7F909F8090907F8F8F
80908F7F8F807F8F80707F6F707F6F303030403F3F4F3F3F5F40406040406F4F406F4040705050\
7F5050805F5F805F5F8F606F8F606F906F709070709F7F809F7F80A0808FA0808F
AF8F90AF8F90AF909FAF909FB09FA0B090A0BF9FA0B09FA0BFA0AFB0A0AFBFAFB0BFAFB0C0B0C0\
C0BFC0CFC0CFC0C0CFCFCFD0CFCFD0D0D0DFCFCFD0CFCFDFC0C0D0BFC0D0B0BFCF
BFC0CFB0BFC0B0BFC0AFB0C0AFBFC0A0B0BFA0B0BF9FAFAF90AFAF8FA0A08FA0A07F909070908F\
6080805F7F7F000000000000000000000000000000000000000000000000000000
0000000000000000003F3F50505F805F60906F7F9F7080A08F90B08F9FBF90A0BF9FA0BFA0B0C0\
A0B0C0AFBFCFAFBFC0B0C0C0AFBFBFB0BFB0AFB0AFB0BFAFA0B09FA0AF8F909F7F
8F906F60604F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F4F704F5F805F6F8F\
6F7F9F7080A08090B08090B08F9FBF90A0BF9FAFC09FB0CFAFBFD0AFBFD0AFBFD0
AFBFCFB0BFCFB0B0C0B0BFC0B0B0BFBFB0BFB0B0B0BFB0BFBFAFB0BFB0B0B0AFAFB0AFAFB0A0A0\
B0A0A0AF9F90B09F90AF9090AF908FA08F80AF8F80A08F7FA0807F90706F907060
80605080605F805F508060507F5F4F7F5F4F7050407F5040704F40704F406F403F5F403F3F3030\
506080506F806070906F7F9070809F7080907F8F9F7F8F9F3F303F403030503F3F
5F403F6F4F406F4040704F40704F4070504F704F4F7F50507F50507F5F507F5050805F5F805F5F\
8F60608F606F906F709070709F7F7F907F7FA0808FA0808FAF8F90A08F90AF909F
AF909FB090A0B0909FBF9FA0B09FA0BFA0AFBF9FAFC0A0B0C0A0B0C0AFB0BFB0BFC0BFC0C0BFC0\
CFC0CFC0C0CFCFCFD0CFCFD0CFCFDFC0C0D0C0C0D0BFC0CFBFC0D0B0BFCFB0BFCF
AFB0C0AFBFC0A0B0BFA0B0BF9FB0B09FAFB08FA0A08FA0A0809F9F809F9F70908F6F8F805F7F7F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F3F4F4F4F705F6090606F907080A07F8FAF8F9FBF8F9FBF9FAFBF9FAFBF\
A0B0C0A0B0C0AFBFC0AFBFBFAFBFBFAFB0AFAFBFAFAFB0A0AFB09F9FA08F909F7F
8F8F6F6F6F50000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F50704F6080\
5F6F8F6F7FA07F8FAF8090B08090B090A0BF90A0C09FAFCF9FB0CFAFBFD0AFBFD0
B0BFD0AFBFCFB0BFCFB0BFC0B0BFC0B0B0BFBFBFC0BFB0BFBFB0BFBFB0B0C0B0BFBFAFB0BFAFAF\
B0A0A0BFAFA0B0A09FB0A09FAF9F90B09F90AF908FAF908FA08F80AF8F80A0807F
9F7F7090706F9F7F6F9070609070608F6F608F6F6080605F8F605F805F508060507F5F507F5F50\
7F504F7F504F704F40704F406F4F3F7F504F704F407F504F70504F7F5F507F5050
805F507F5F50805F5F805F5F806060805F5F8F60608F60608F6F6F8F606F906F708F6F6F907070\
9070709F7F809F7F80A0808FA0808FAF8F90AF8F90B0909FAF909FB09FA0B09FA0
BF9FA0BF9FA0C0A0AFBFA0AFC0AFB0C0A0B0CFAFBFCFAFB0CFB0BFCFB0BFCFC0CFC0C0CFCFCFD0\
CFCFD0D0D0DFCFCFD0CFCFD0C0C0D0C0C0D0BFC0CFB0C0CFAFBFC0B0BFCFAFB0C0
AFBFC0A0B0BFA0B0BF9FAFB090AFAF8FA0A08FA0A07F9F9F7F90906F8F8F6F808040605F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F4F7050608F60709F707FA0808FB08F9FB090A0BF90A0BF9FAFC0\
A0B0BFAFBFC0AFBFC0AFBFC0AFBFB0AFBFB0AFB0AFAFBFAFA0B09FA0AF90909F7F
8F9070707050000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000304F70\
4F5F80506F8F6F7F9F7080A07F90B08090B08F9FBF8FA0BF9FAFC09FAFCFAFBFD0
AFBFD0AFBFD0AFBFCFB0BFCFAFB0C0B0BFC0B0B0C0BFBFC0B0B0BFBFB0BFBFB0B0BFB0BFBFAFB0\
BFAFB0B0AFAFBFAFAFB0A0A0B0A0A0B09F9FB09F9FAF9090AF908FAF8F8FAF908F
A08F80A08F80A0807FA08F809F807F9F807F9F7F709F7F7090706F90706F906F6090706F8F6F60\
8F6F6080605F80605F805F50805F50805F5080605F80605F8F60608060608F6F60
8F60608F6F6F8F6F6F907070906F6F9070709070709F7F7F90707F9F7F7F9F7F7FA080809F7F80\
A0808FA0808FAF8F90AF8F90AF909FAF909FB09FA0B0909FBF9FA0BF9FA0C0A0AF
BFA0AFC0A0AFC0A0AFCFAFB0C0AFB0CFB0BFCFB0BFD0B0C0CFB0C0D0BFC0CFC0CFCFCFD0CFCFD0\
CFCFD0C0CFD0C0CFD0BFC0CFBFC0CFB0BFCFB0BFCFAFB0C0AFBFC0A0B0BFA0B0BF
9FAFB09FAFB08FA0AF8FA0A0809F9F7F9F9F70909070908F607F7F40605F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000040406050608F5F6F90707FA07F80AF8F9FBF8F9FB090A0BF90A0BF\
A0B0C0A0B0C0AFBFC0AFBFBFAFBFBFAFB0B0AFBFAFA0B0A0AFB0A09FA08F909F80
8F8F6F7F7F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F50704F6080506F8F6F7FA0708FAF8090B08090B090A0BF90A0C09FAFCF9FB0CF
AFBFD0AFBFD0B0C0D0AFBFCFB0BFD0B0BFCFB0BFCFB0B0C0BFBFCFB0B0C0BFBFC0BFB0BFBFB0BF\
BFB0BFC0B0BFBFAFB0BFB0B0BFAFAFBFAFAFB0A0A0B0A0A0B09F9FB09F90AF908F
B09F90AF908FAF908FAF8F8FAF908FAF8F8FAF8F8FA08F80AF8F8FA08F80A08F80A0807FA0807F\
9F7F7F9F7F7F9070709F7F7090706F90707090706F9F7F709F70709F7F7F9F7F7F
A080809F7F7FA08080A08080A08F8FA0808FAF8F8FA0808FAF8F90A08F8FAF8F90AF8F8FAF9090\
AF9090B0909FB0909FBF9FA0B09FA0BFA0A0BF9FA0C0A0AFC0A0AFC0AFB0C0AFB0
CFB0BFCFAFB0D0B0BFCFB0BFD0BFC0D0BFC0D0BFCFD0BFC0D0BFCFCFBFC0CFBFC0CFCFD0CFCFD0\
C0CFD0C0CFD0BFC0CFBFC0CFB0BFC0AFBFCFAFB0C0AFBFC0A0B0BFA0B0BF90AFB0
90AFAF8FA0A080A0A07F909F7F90906F8F8F608080405F5F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000404F60505F8F606F906F7F9F7F8FAF8090B090A0BF90A0BF9FAFC0\
9FAFBFAFBFC0AFBFC0AFC0C0AFBFBFAFBFB0AFB0AFAFBFAFAFB0A0A0AF90909F80
9090708080604F4F3F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000304F704F5F80506F8F6F7F9F7080A07F90B08090B08F9FBF8FA0BF9FAFC0
9FB0CFA0BFD0A0BFD0AFC0D0AFBFD0B0BFD0AFBFCFB0BFCFB0B0C0B0BFCFB0B0C0BFB0C0B0B0BF\
BFB0BFBFB0B0BFB0BFBFAFB0BFB0B0BFAFAFBFAFAFB0A0A0BFA0A0B09F9FB09F9F
B09F90B09F9FAF9090B09F90AF9090B09090AF908FAF9090AF8F8FAF908FA08F8FAF8F8FA08F80\
AF8F8FA08080A08F80A08080A080809F807FA08080A08080AF8F8FA08080AF8F8F
A08F8FAF8F90A08F8FAF8F90AF8F8FAF9090AF8F90AF9090AF8F90B0909FAF8F90B0909FB0909F\
B09FA0B09F9FBF9FA0BF9FA0C0A0AFBFA0AFC0AFAFC0AFAFCFAFB0CFAFB0CFB0BF
CFB0BFD0BFC0D0B0C0D0BFC0D0BFC0D0BFC0CFB0C0CFBFC0C0B0C0C0BFC0BFB0C0C0CFD0BFC0CF\
BFC0CFB0BFC0B0BFCFAFB0C0AFBFC0A0B0BFA0B0BF9FAFB09FAFB08FA0AF8FA0A0
809F9F7F9F9F708F90708F8F60807F4F605F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F3F50505F805F60906F7F9F7080A08090B08F9FB090A0BF90A0BF\
9FB0C0A0B0BFAFBFC0AFBFBFAFBFBFA0B0B0AFBFAFA0B0A0AFB0A09FA09090A080
8F9070808F604F4F3F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F50704F6080506F8F6F7F9F7080AF8090B08090B08FA0BF90A0BF
9FAFC09FAFC0A0B0CFA0B0CFAFBFD0AFBFD0B0C0D0AFBFD0B0BFD0B0BFCFB0BFCFB0BFC0BFBFC0\
B0B0C0BFBFC0BFB0BFBFB0BFBFB0B0C0B0BFBFAFB0BFAFB0BFAFAFBFAFAFB0A0A0
BFA0A0B0A0A0BFA0A0B09F9FB0A0A0B09F9FB09F9FB09F9FB09F9FAF9090B09F9FAF9090B09090\
AF9090B09090AF8F8FAF9090AF8F8FB09090AF9090B09090AF9090B09090AF9090
B0909FAF9090B09F9FB0909FB09F9FB0909FBF9FA0B09F9FBF9FA0B09F9FBF9FA0BF9FA0BF9FA0\
BF9FA0C0A0AFC0A0AFCFAFB0C0AFB0CFB0B0CFAFB0D0B0BFD0B0BFDFBFC0D0BFC0
DFBFC0D0BFC0D0BFC0CFBFC0CFBFC0C0B0C0C0BFC0C0BFC0C0BFC0BFB0C0BFBFC0BFC0CFBFC0CF\
B0BFC0B0BFC0AFB0C0AFB0C0A0B0BFA0B0BF90AFB090AFAF8FA0A0809FA07F909F
7090906F8F8F608080405F5F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F3F504F507F5F6F9060709F7080AF808FB0909FBF90A0BF9FAFBF\
9FAFBFA0B0C0A0BFC0AFBFC0A0BFBFAFBFBFA0B0AFAFBFAFA0B0A0A0AF9F90A080
8F9F7F808F605F5F40000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000304F704F5F80506F8F607F9F7080A07F90B08090B08F9FBF
8F9FBF90A0BF90A0C09FAFCF9FAFC0A0B0CFA0BFCFAFC0D0AFBFCFAFBFD0AFBFCFB0BFCFAFB0C0\
B0BFC0B0B0C0BFBFC0B0B0BFBFB0BFBFB0BFBFB0BFBFAFB0BFB0B0BFAFAFBFAFAF
BFA0AFBFAFAFBFA0A0BFA0AFB0A0A0BFA0A0B0A0A0BFA0A0B09F9FB09FA0B09F9FB09F9FB09F9F\
B09F9FB09090B09F9FAF9090B09090B09090B09F9FB09090B09F9FB0909FBF9F9F
B0909FBF9FA0B09F9FBF9FA0B09FA0BF9FA0BF9FA0BFA0AFBF9FA0C0A0AFBFA0A0C0A0AFBFA0AF\
C0AFAFC0AFAFCFB0B0CFAFB0D0B0BFD0B0BFD0BFC0D0BFBFD0BFC0D0BFC0D0BFC0
CFB0C0CFBFC0C0B0C0C0BFC0BFB0C0BFBFC0BFB0C0BFBFC0B0B0BFB0BFC0B0B0BFB0BFC0AFBFC0\
AFBFC0A0B0BFA0B0BF9FAFB09FAFB08FA0AF8FA0A07F9F9F7F909F708F906F8F8F
5F7F7F405F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003030404F507050608F60709F6F7FA0808FB08F9FB090A0BF90A0B0\
9FAFBF9FAFBFAFBFC0A0BFBFAFBFBFA0B0B0AFBFB0A0B0A0A0B0A09FAF909FA08F
8F9070808F6F5F5F40000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F50704F6080506F8F6F7F9F7080A07F8FAF7F8FAF
809FB0809FB08FA0BF8FA0BF9FAFC09FAFC0A0B0CFA0B0CFAFBFD0AFBFD0B0C0D0AFBFCFB0C0D0\
B0BFCFB0BFCFB0BFC0BFBFCFB0B0C0BFBFC0BFB0BFBFB0BFBFB0BFC0B0BFBFB0B0
C0B0BFC0B0B0C0B0B0BFAFB0C0B0B0BFAFB0C0AFB0BFAFAFC0AFB0BFA0AFBFAFAFBFA0A0BFA0AF\
BFA0A0BFA0A0B09FA0BFA0A0BF9FA0BFA0A0BF9FA0C0A0AFBFA0A0C0A0AFBFA0A0
C0A0AFC0A0AFC0A0AFC0A0AFC0AFB0C0A0AFCFAFB0C0AFAFCFAFB0CFAFB0CFAFB0CFAFB0CFB0BF\
CFB0B0D0B0BFD0B0BFDFBFC0DFBFC0DFBFC0D0BFC0D0BFC0CFBFC0CFBFC0C0BFC0
C0BFC0BFB0C0C0BFC0BFB0C0BFBFC0B0B0C0B0BFC0B0B0C0B0BFC0AFB0C0AFBFC0AFB0BFAFBFC0\
A0B0BFA0B0BF90AFB090AFB08FA0A0809FA07F909F7F90906F8F8F608080405F5F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F4F6F5F6090606F907080A07F8FAF8F9FBF8F9FBF90AFBF\
9FAFBFA0B0C0A0B0BFAFBFC0AFBFBFAFBFBFA0B0AFAFBFAFA0B0A0A0B09F90A08F
909F7F808F6F6F6F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000304F70405F7F4F60805F709060709F7080A0
708FAF7F90B07F90B0809FB08F9FB090A0BF90A0BF9FAFC09FAFC0A0B0CFA0B0CFAFBFD0AFBFD0\
B0BFD0AFBFCFB0BFCFB0BFC0B0BFCFB0B0C0BFBFC0B0B0BFBFB0BFBFB0BFBFB0BF
BFB0BFC0B0BFBFAFB0C0B0BFBFAFB0C0B0BFBFAFB0C0B0B0C0AFB0C0B0B0BFAFB0C0AFB0BFAFAF\
C0AFB0BFA0AFC0AFAFBFA0AFC0AFAFBFA0AFC0AFAFC0A0AFC0AFB0C0A0AFC0AFB0
C0AFAFCFAFB0C0AFB0CFAFB0CFAFB0CFB0BFCFAFB0CFB0BFCFB0BFD0B0BFCFAFB0CFB0BFCFB0BF\
DFBFC0D0BFBFDFBFC0D0BFC0D0BFC0CFB0BFCFBFC0C0B0C0C0BFC0BFB0C0BFBFC0
BFB0C0BFBFC0B0B0BFB0B0C0AFB0BFB0BFC0AFB0BFAFB0C0A0B0BFA0B0BF9FAFB09FB0BF9FAFB0\
9FAFB08FA0AF8FA0A07F9F9F7F9F9F708F906F8F8F6080804F5F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000404F60505F8F606F906F7F9F7F8FAF8090B08F9FBF90A0BF\
9FAFBF9FAFBFA0B0C0A0BFC0AFBFBFA0B0B0A0B0B0A0B0AFA0B0A09FAF909FA08F
8F907F80906F6F6F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F5070405F7F506F8F5F6F90
607F9F6F80A07F90AF7F90AF809FB0809FB08FA0BF8FA0BF90AFC090AFC0A0B0CFA0B0CFAFBFD0\
AFBFD0B0C0D0AFBFCFB0BFD0B0BFCFB0BFCFB0BFC0BFBFC0B0B0C0BFBFC0BFB0BF
BFB0C0BFB0BFBFB0C0BFB0BFBFB0BFBFB0BFC0B0BFBFB0BFC0B0BFBFB0BFC0B0BFC0B0BFC0B0BF\
C0AFB0CFB0BFC0AFB0C0B0B0C0AFB0CFB0B0CFAFB0CFB0BFCFAFB0D0B0BFCFB0BF
D0B0BFCFB0BFD0B0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BFC0B0BFCFB0C0C0B0BFCFB0BF\
D0BFC0D0BFC0CFBFC0CFBFC0C0BFC0CFBFC0C0B0C0C0BFC0BFB0C0BFBFC0B0B0C0
B0BFC0AFB0BFB0BFC0AFB0BFAFBFC0A0B0BFA0B0C09FB0BF9FB0BF90AFB09FAFB090A0AF90A0AF\
809FA0809FA07F909F7090906F8F8F608080405F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F405F5F608F606F906F7FA07080AF8F9FBF8F9FBF90A0BF\
90A0BF9FB0C0A0B0BFAFBFC0A0BFBFAFBFBFA0B0B0A0B0AFA0B0A0A0B0A09FA08F
909F7F808F6F707F5F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000304F70405F7F
4F60805F6F906070907080A0708FAF7F90AF7F90B0809FB0809FB08FA0BF90A0BF9FAFC09FAFC0\
A0B0CFA0B0CFAFBFD0AFBFCFAFBFD0AFBFCFB0BFCFAFB0C0B0BFC0B0B0C0B0BFC0
B0B0C0B0B0C0B0B0C0BFB0C0B0B0BFBFB0C0BFB0BFBFB0C0BFB0BFBFB0BFBFB0BFBFB0BFBFAFBF\
C0B0BFBFAFBFC0B0BFBFAFB0CFB0BFCFB0BFCFB0BFC0B0B0CFB0BFC0B0BFCFB0BF
C0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0C0BFB0BFC0B0BFBFB0BFC0B0C0BFB0BF\
CFBFC0C0B0BFC0BFC0BFB0BFBFBFC0BFB0BFBFBFC0B0B0BFB0B0C0AFB0BFAFB0C0
A0B0BFAFB0C0A0B0BFA0B0BF9FAFB09FAFB090AFB090AFB08FA0AF8FA0AF809FA080A0A07F9F9F\
7F909F708F906F8F905F7F7F405F5F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F3F504F507F606F9060709F7080AF8090B08F9FBF8F9FB0\
90A0BF90AFBFA0B0BFA0B0BFA0BFBFA0B0B0A0B0B0A0B0A0A0B0A0A0AF9F9FAF90
8F907F80906F707F5F404030000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F5070405F7F4F6F8F506F8F607F9F6F80A07F90AF7F90AF809FB0809FB08FA0BF8FA0BF90AFC0\
9FAFC09FB0CF9FB0CFAFBFD0AFBFD0B0C0D0AFBFCFB0BFCFAFBFCFB0BFCFB0BFCF
B0BFCFB0BFCFB0BFCFB0B0C0BFBFCFB0B0C0BFBFC0B0B0C0BFBFC0BFB0C0BFBFC0BFB0BFBFBFC0\
BFB0BFBFB0C0BFB0BFC0BFC0C0B0BFC0B0C0C0B0BFC0BFC0C0B0BFC0B0C0C0B0BF
C0BFC0BFB0BFC0B0C0BFB0BFC0BFC0BFB0C0BFBFC0BFB0BFBFBFC0BFB0C0BFBFC0BFB0C0BFBFC0\
BFB0C0C0BFC0BFB0C0BFBFC0B0B0C0B0BFC0AFB0C0B0BFC0AFB0C0AFBFC0A0B0BF
A0B0C09FAFBF9FAFBF90AFB090AFB08FA0AF8FA0AF809FA08FA0AF809FA0809FA07F909F7F9F9F\
6F8F8F608080405F5F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F3F4F505F7F5F60906F7F9F7080A08090B08F9FB090A0BF\
90A0BF9FAFBF9FB0BFAFBFC0A0BFBFAFBFBFA0B0B0A0BFAFA0B0A0A0B0A09FA08F
90A0808090707F8F604F4F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000304F70405F7F4F60805F70905F70906F80A0708FAF7F90B07F90AF809FB08F9FB0\
90A0BF90A0BF9FAFC09FAFC0A0B0CFA0B0CFAFBFD0AFBFCFAFBFD0AFBFCFAFBFCF
AFBFCFB0BFCFAFB0CFB0BFCFAFBFCFB0BFCFAFB0C0B0BFCFB0B0C0B0BFCFB0B0C0B0BFC0B0B0C0\
BFBFC0B0B0BFBFBFC0BFB0BFC0B0C0BFB0BFBFB0C0BFB0BFBFB0C0BFB0BFBFB0C0
B0B0BFBFB0C0B0B0BFBFB0C0B0B0C0BFBFC0B0B0BFB0BFC0B0B0C0B0BFC0B0B0C0B0BFC0B0B0C0\
B0BFC0B0B0BFB0B0C0AFB0BFAFB0C0A0B0BFAFB0C0A0B0BFA0B0BF9FAFB09FAFB0
90A0B090AFB08FA0AF8FA0AF809FA0809FA07F909F7F9F9F70909F70909F60808F5F7F804F6F70\
405F5F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F4F705F609060709F7080A07F8FAF8F9FBF8F9FB0\
90A0BF90A0BF9FB0BFA0B0BFA0BFBFA0B0B0A0B0B0A0B0AFA0B0A09FAF9F9FAF90
8F9F808F90707F805F505040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F50703F5F7F4F6F8F506F8F607F9F6F80A07F90AF7F90AF809FB0\
809FB08FA0BF8FA0BF90AFC090AFC09FB0CF9FB0C0A0B0CFA0B0CFA0BFCFA0B0CF
AFBFCFA0BFCFAFC0D0AFBFCFAFBFD0AFBFCFB0BFCFAFBFCFB0BFCFAFBFCFB0BFCFB0BFCFB0BFCF\
B0BFC0B0BFCFB0BFC0BFBFC0B0B0C0BFBFC0B0B0C0BFBFC0B0B0C0B0BFC0B0B0C0
B0BFC0B0B0C0B0BFC0B0B0C0B0BFC0AFB0C0B0BFCFAFBFC0B0BFC0AFB0C0B0BFCFAFBFC0AFBFCF\
AFBFC0AFBFC0AFB0C0AFB0C0A0B0BFA0B0BF9FAFBF9FAFBF90A0B090AFB08FA0AF
8FA0AF809FA0809FA07F9FA07F9FA070909F70909F60808F5F7F804F6F6F406060305050000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F4F6F5F608F60709F6F7FA0808FB08F9FB090A0BF\
90A0BF9FAFBF9FAFBFA0B0C0A0B0BFA0BFBFA0B0B0A0B0AF9FB0A0A0B0A09FAF90
90A080809070808F6F505F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000305070405F7F4F608050708F5F70906F80A0708FA0\
7F90AF7F90AF809FB0809FB08FA0BF8FA0BF90AFC090A0BF90AFC090AFBF9FAFC0
9FAFC09FB0C09FB0C0A0B0CF9FB0C0A0B0CFA0B0CFAFBFCFA0BFCFAFBFCFAFBFCFAFBFCFAFBFCF\
AFBFCFAFBFC0B0BFCFAFB0C0B0B0C0AFB0C0B0BFC0AFB0C0AFB0C0AFB0C0AFBFC0
AFB0C0AFBFC0AFB0C0AFBFCFA0B0C0AFBFC0A0B0C0AFBFC0A0B0C0A0B0C0A0B0BFA0B0C09FB0BF\
A0B0C09FAFBF9FB0BF90AFB090AFB08FA0B08FA0B0809FAF809FAF809FA0809FA0
7F909F7F909F70909F709090607F8050707F40606F3F6060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004040605F608F5F6F906F7FA07080AF8F9FB08F9FB0\
90A0BF90A0B09FAFBF9FAFBFA0BFBFA0B0B0A0B0B09FB0AFA0B0A09FAF9F9FAF90
8F9F808F90707F806060604F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F5070405F7F4F6F8F506F8F607F9F\
6F80A07F90AF7F90AF809FB0809FB08F9FBF809FB08FA0BF8FA0BF90A0BF8FA0BF
90AFBF90A0BF90AFC090AFBF9FB0C09FAFC09FB0C09FAFC0A0B0CFA0B0C0A0BFCFA0B0C0AFBFCF\
A0BFCFAFBFCFAFBFCFAFBFCFAFB0C0AFBFCFAFB0C0AFBFCFA0B0C0AFBFCFA0B0C0
A0BFC0A0B0C0A0B0C09FB0BFA0B0C09FB0BFA0B0C09FAFBF9FB0C09FAFBF9FB0BF90AFBF9FB0BF\
90AFB09FAFBF90A0B090A0B08FA0AF8FA0AF809FA0809FA07F90A07F90A070909F
708F90607F8050707F40606F3F5F60000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040405F505F80606F9F6F7F9F7F8FAF8090B090A0BF\
90A0BF90AFBF9FAFBFA0B0C0A0BFBFA0BFBFA0B0B0A0B0B09FB0A0A0B0A09FAF90
90A08F8F907F808F6F606F4F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000030506F405F7F4F6080\
5070905F70906F80A0708FA07F90AF7F8FAF7F90AF7F90AF809FB07F90AF809FB0
809FB08F9FB0809FB08FA0BF8FA0B090A0BF8FA0B090AFBF90A0BF90AFBF90AFBF9FAFC09FAFBF\
9FB0C09FB0C0A0B0C09FB0C0A0B0C09FAFBF9FB0C09FAFBF9FB0C090AFBF9FAFBF
90AFBF9FAFBF90A0B090AFBF90A0B090AFBF8FA0B090AFB08FA0B090A0B08FA0AF8FA0B08FA0AF\
8FA0B0809FAF809FAF7F9FA07F9FA070909F70909F6F8F9F6F8F905F7F80506F70
4060603F5F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F3F50505F805F60906F7F9F7080A08090B08F9FB0\
90A0BF90A0B09FAFBF9FAFBFA0BFC09FB0B0A0B0B09FB0AFA0B0AF9FAF9F9FAF9F
90A08F8F9F7F7F8F606F7050000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000305070\
3F5F7F4F6F8F50708F5F7F90607F906F809F6F809F708FA0708FA07F90AF7F90AF
7F90AF7F90AF809FB0809FAF809FB0809FB08FA0B0809FB08FA0B08FA0B090A0BF8FA0BF90AFBF\
90A0BF90AFBF90AFBF90AFBF90A0BF90AFBF90A0B090AFBF8FA0B090A0BF8FA0B0
8FA0B08FA0B08FA0B08FA0B08FA0B0809FAF8FA0B0809FAF80A0AF809FAF80A0AF809FAF809FAF\
7F9FA07F9FA070909F70909F6F8F9F6F8F905F7080506F703F5F603F505F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040405F4F507F606F906F709F7F80AF7F8FAF8F9FBF\
8F9FB090AFBF90AFBF9FB0BF9FB0BFA0BFBFA0B0B0A0B0B09FB0A0A0B0A09FAF9F
90A08F80907F80906F707F5F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030506F3F5F703F5F7040607F40607F4F6F804F6F805F708F5F708F607F90
607F906F809F6F809F708FA0708FA07F90AF7F90AF7F9FAF7F90AF809FAF809FAF809FB0809FAF\
8FA0B0809FB08FA0B0809FAF8F9FB0809FAF809FB0809FAF809FAF7F9FAF809FAF
7F90A0809FAF7F90A07F9FAF7F90A07F9FA07F90A07F9FA07090A07F90A07090A07090A06F8F9F\
6F8F9F6F8F9060808F50707F4F6F703F5F60000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F3F504F50805F608F6F709F707FA0808FB08090B0\
8F9FBF8FA0B090AFBF90AFB09FB0BF9FB0BFA0B0B09FAFAF9FB0AF9FAFA09FB09F
90A08F8F9F7F7F8F6F607050000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F5F703F5F7040607F40607F
4F6F804F6F8050708F50708F5F7F90607F9060809F6F809F708FA0708FA07F90AF7F90AF7F9FAF\
7F90AF809FB07F90AF809FAF7F90AF7F9FAF7F90A07F9FAF7F90A07F90AF7090A0
7F90A0708F9F708F9F6F8F9F6F8F9F6080906080905F7F8F60808F5F7F805F7F8050707F507080\
4F707F4F60703F505F6F6050504030000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040405F4F507F5F609060709F7080A07F8FAF8F9FBF\
8F9FB090A0BF90A0BF9FB0BF9FB0BFA0BFBF9FB0B0A0B0B09FB0AF9FB0A09FAF9F
90AF908F9F7F8090705F604F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030506F3F5F703F5F7040607F40607F4F6F80506F805F708F5F708F607F90607F90\
6F809F6F809F708FA06F809F6F8F9F6080906080905F7F8F5F7F8F507080507080
5070805070804F6F7F4F6F7F406070406F703F606F3F606F3F5F6F3F5F6F305060A0A09F9F9F90\
9F908F8F8F7F8F80707F7060605F4F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F3F505050805F608F6F709F6F7FA07F8FAF8090B0\
8F9FBF8F9FB090AFBF90AFBF9FB0BF9FB0BFA0B0B09FB0AF9FB0AF9FAFA09FAF9F
8FA08F8F9F7F7F8F6F5F604F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F5F703F5F7040607F40607F4F6F80\
4F6F805070804F6F7F4F6F7F40607F4F607F4060704060703F5F6F3F5F6F305060
707F907F8F9F8090AF8F9FAF90A0B090A0B0A0B0BF9FAFB0A0AFB0A0A0AFA0AFAFA0A0A0A0A09F\
9F9F8F9F908F908F7F8F80707F6F5F605040000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040405F5050805F608F606F907080A07F8FAF8090B0\
809FB090A0BF90A0BF9FAFBF9FB0BFA0B0BF9FB0B09FB0B09FB0AF9FB0AF9FAF9F
90A08F80907F808F6F505F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000405070506080
606F8F70809F7F8FA08F90AF8F9FAF90A0B09FAFB0A0B0B09FAFAFA0AFAFA0A0A0A0AFA0A0A09F\
A0A09F9F908F9090808F80708F7F6F7060505F4F3F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040405F505F8050608F60709F6F7FA07F8FB08090AF\
8F9FB08F9FB090AFBF90AFB09FB0BF9FAFB09FB0B09FAFAF9FB0AF9FAFA09FAF9F
8F9F8080907F7F8F60505040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F404F4F5F70
606F8F6F7F907F8FA0808FA08F9FAF90A0AF9FAFB09FAFB0A0B0BFA0AFB0A0B0AFA0AFA0AFAFA0\
A0A09FA0A09F9F908F9F90808F80708F7F6F6F5F4F504030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004040605050805F609060709F7080A07F8FAF809FB0\
8F9FB090A0BF90A0B09FB0BF9FB0BF9FB0BF9FAFB09FB0B09FAFA09FB0A090AF90
8FA08F8090707F8F6F404F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000040506F
505F7F60708F707F90808FA08090A08F9FAF90A0AF9FAFB0A0B0B0A0B0B0A0AFAFA0AFAFA0A0A0\
A0AFA0A0A09FA09F909F9080908F7F8F7F6F807F60605040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040405F505F805F608F6F709F6F7FA07F8FAF8090AF\
8F9FB08F9FB090A0BF90AFB09FB0BF90AFB09FB0B090AFAF9FB0AF90AF9F90AF90
809F80809070708060404F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000303F40
4F5F7F5F60806F7F907F809F8090A08F90AF90A0B09FA0B0A0B0BFA0B0BFAFB0B0A0AFAFAFAFAF\
A0A0A0AFAFA0A0A09FA09F90909080908F7F8F806F7F705F5F4F3F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F4F60505F805F60906070907080A07F8FAF8090B0\
809FB08FA0BF90A0B09FAFBF9FAFBF9FB0BF90AFAF9FB0AF90AFA09FB0A08FA090
8F9F807F8F70708F60000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F406050607F5F6F80707F9F7F8F9F8090AF8F9FAF90A0B09FA0B0A0B0BFA0B0BFAFB0B0A0AFAF\
A0AFAFA0A0A0AFA0A0A09F909F9F8F908F80908F7F807F6F706050504030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040405F505F805F608F60709F707FA07F90B07F90AF\
809FB08F9FB090A0BF90A0B09FB0BF90AFB090AFAF90AFA090AFA090AF9F8FA090
80907F7F8F706F805F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004050705F608060708F7F809F808FA08F9FAF909FAF9FAFBFA0AFBFAFBFBFAFB0BFAFB0BF\
A0AFAFAFAFAFA0A0A0AFAFA0A09F90A09F8F908F80908F7F8F7F6F706050000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F4F60505F80606F9060709F7080A07F8FAF809FB0\
809FB08FA0B08FA0B09FAFBF9FAFBF9FB0B090AFAF9FB0AF90AFA090AFA08FA08F
809F807F8F6F708060000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F4F50705F6F8060708F7F809F808FA0909FAF90A0B09FAFBFA0AFBFAFBFC0AFB0BF\
AFB0B0A0AFAFAFAFAFA0A0A0AFA0A0A09F909F908F908F7F9080708070606F5F4F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000404060505F805F608F6F709F707FA07F90B07F90AF\
809FB0809FB08FA0B090AFB09FB0BF90AFAF90AFAF90AFA090AFA08FA09F8FA08F
7F907F7F8F6F607050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F40504F5F7F606F8F6F7F907F8FA08090AF909FB09FA0B0A0B0BFA0B0BFB0BFC0\
AFB0BFAFB0BFAFAFAFAFAFAFA0A0A0AFA0A0A09F909F908F90807F9080707F6F5F
605040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F4F6F505F80606F9060709F7080AF7F8FAF8090B0\
8090AF8FA0B08FA0B090AFBF90AFB090AFB090AFAF90AFAF8FAF9F90AF9F809F8F
80907F70806F60705F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000404F604F5F7F606F8F6F7F90808FA08090A090A0B09FA0B0A0B0BFAFB0BF\
B0BFC0AFB0BFAFB0B0A0AFAFAFAFA0A0A09FAFA09F9F908F9F90809080708F7F6F
7F6F5F605040000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004040605F5F8F5F608F6F709F707FA08090B07F90AF\
809FB0809FAF8FA0B08FA0B090AFB090AFAF90AFAF8FA0A090AF9F8FA090809F8F
7F8F7070806F5F6F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F5F7F5060806F709070809F8090AF8F9FAF9FA0B0A0AFBFAFB0C0\
AFBFC0B0BFC0AFB0BFB0B0BFAFAFAFAFAFAFAFA09FAFA09F9F908F9F9080908070
8F80707F6F5F605040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F4F6F505F80606F906F709F7080AF7F8FAF809FB0\
809FB08FA0B08FA0B090AFBF90AFB090AFB08FA0A090AFA08FA09F90AF9F809F80
7F907F7080605F6F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000030303F4F5F7F5F60806F709070809F8F90AF8F90AF9FA0B0A0AFBF\
AFBFC0B0BFC0B0BFC0B0B0BFB0B0BFAFAFAFAFAFAFAFA09FAFA09F9F908F9F8F80
9080708F7F6F7F6F5F604F40000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004040605F5F8F5F608F6F709F707FA07F90AF7F90AF\
809FB0809FAF8FA0B08FA0B090AFB08FA0AF8FAFAF8FA0A08FAF9F8FA090809F80
708F7070806F4F5F40000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F405050607F5F6F80707F907F809F8F90AF909FB09FAFBF\
A0AFBFB0BFCFB0C0CFBFBFC0B0B0BFB0B0BFB0B0AFB0AFAFAFA09FAFA09F9F908F
9F8F809080708F7F6F7F6F5F604F40000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F4F6F5F5F8F606F906F709F7080AF7F8FAF8090B0\
8090AF8FA0B08FA0B090AFB08FAFB090AFB08FA0A08FAFA08FA09F8FA09F7F9080
7F8F706F8060505F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F4F605060805F6F80707F9F7F8FA08F90AF909FAF\
A0AFBFA0B0BFB0BFCFB0BFCFBFBFC0BFBFBFBFB0B0B0AFAFB0AFA0AFA09FAF9F9F
9F908F9F8F808F80708F7F6F7F6F5F5F4F40000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F3F4F40406F5F608F5F608F6F709F7080A07F90B07F90AF\
809FB0809FAF8FA0B08FA0B090AFB08FA0AF8FAFA080A09F8FA09F809F907F9F80
70806F6F8060405040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003030304050705F6F8060708F7F809F808FA0909FB0\
9FA0B0A0AFBFAFB0C0BFC0CFC0C0C0C0C0C0BFBFBFBFBFBFB0AFAFB0AFAFAFA09F
AFA09FA0908F9F8F809080708F7F6F7F6F5F604F40000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F3F504F5F4F5F5F8F606F906F709F7080AF7F8FAF8090B0\
809FAF8F9FB08FA0B090AFBF8FAFB090AFAF8FA0A08FAFA08FA09F8FA0907F907F
708F706F7F60405040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000303F404F5F705F6F806F708F7F809F808FA0\
909FB09FA0B0A0AFBFAFB0BFBFBFC0BFBFC0C0C0C0BFBFBFBFB0B0B0AFAFB0AFA0
AFA09FAF9F9F9F908F9F8F808F7F708F7F6F7F60505F4F40000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000040405F4F4F705F608F606F906F7F9F7080A07F90AF7F90AF\
809FB0809FAF8FA0B08FA0B08FAFB08FA0A08FA0A080A09F8FA09F809F8F7F907F
6F806F6F7F603F403F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F405F4F5F7F60708F6F7F90808FA0\
8F90AF909FB09FA0B0A0AFBFAFAFBFBFBFC0BFBFC0C0BFC0BFB0BFBFB0B0B0AFAF
B0AFAFAFA09FAF9F9F9F908F9F8F808F7F708F7F6F7F6F5F5F4F40000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F3F50404060505F805F608F606F906F709F7F8FAF7F8FAF8090B0\
8090AF8FA0B080A0B08FAFB08FA0AF8FAFAF8FA0A08FA09F80A09080A09070907F
708F6F607F5F303040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000040506F50607F60708F6F7F90\
7F8FA0808FA0909FAF909FAFA0A0B0AFAFB0B0B0BFBFBFBFBFBFBFB0B0B0BFB0B0
AFAFAFAFA0A0AF9F9FAF9F909F8F809F8F808F7F708F7F6F7F60505F403F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F3F504F4F6F5050805F608F606F906F7F9F7080A07F90AF7F90AF\
809FAF809FAF8FA0B08FA0AF8FA0AF80A0A08FA0A0809F9F80A0907F9F8F70907F
6F8060607F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030303F4F5F7F5060806F7090\
707F90808FA08F90AF909FAF9F9FAFA0A0B0AFAFB0BFB0C0BFBFC0BFBFBFB0B0B0
BFB0B0B0AFAFB0AFA0AF9F9FAF9F9F9F908F9F8F80907F708F7F6F7F605F5F4F40000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F40605050705F5F8F5F608F6F709F6F7F9F7F8FAF7F8FAF8090B0\
809FAF809FB080A0AF8FAFB08FA0AF8FA0AF80A09F8FA09F809F90809F8F708F7F
6F806F60705F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003030404F5F70506080\
6F708F707F90808FA08F90A09090AF9090AFA0A0B0AFA0B0B0B0BFBFB0BFBFB0BF
B0AFB0B0AFB0AFA0A0AFA0A0AF9F9FAF9F909F8F809F8F7F8F7F708F706F7F6F5F6F504F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F3050404060505F80505F80606F906070907080A07080A07F90AF7F90AF\
809FB0809FAF8FA0B08FA0AF8FA0AF80A0A080A0A0809F9080A0907F9080708F7F
607F60000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003030404F5F7F\
5F60806F7090707F90808FA08F90A0909FAF9F9FAFA0A0B0AFA0B0B0B0BFBFB0BF
BFB0BFB0AFB0B0AFB0AFA0A0AFA0A0AF9F9FAF9F9FA0908F9F8F80907F708F7F6F806F5F7F6050\
4F3F3F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000403F505050705F5F8060608F606F906F7F9F707FA07F8FAF7F8FAF8090AF\
8090AF80A0B080A0AF8FAFB080A0AF8FA0AF80A09F80A09F7F9F907F9F8F6F8F70
608060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303040\
4F5F7F5060806F7090707F90808FA08F8FA09090AF9090AFA09FB0A0A0B0B0AFBF
B0AFBFB0AFB0B0A0AFB0A0AFAF9FA0AF9FA0A09090AF9F909F8F8F9F8F80907F708F7F6F806F5F\
7F6F5F5F4040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F40604F4F705F5F805F608F6070906F709F7080A07080A07F90AF7F90AF\
809FAF809FAF8FA0B080A0AF80A0AF809FA080A09F7F9F907F9F90708F7F6F806F
5F7050000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
303F4F50607F5F60806F7090707F9F808FA08F90AF909FAF9F9FAFA0A0B0A0A0B0
B0AFBFBFAFBFBFAFBFB0A0AFB0A0AFAF9FA0B09FA0AF9F9FAF9F9FA09090A08F8F90807F907F70\
807060806F5F6F504F403030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F40605F5F805F608F606F906070907080A07080AF7F8FAF7F8FAF809FAF\
809FAF80A0B080A0AF8FA0AF80A0A080A0A07F9F9F80A09F7F908F708F7F607F60
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000303F4F505F7F5F60806F7090707F90808FA08F8FA09090AF9090AFA09FB0
A09FB0B0A0B0B0AFB0B0AFB0B0A0AFB0A0AFAF9FA0AF9FA0AF909FAF9F9FA0908FA08F8F90807F\
907F70807060806F5F705F4F503F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000403050504F6F5F5F8060608F5F6F8F6F709F6F7FA07F8FAF7080A07F90AF7F90AF\
809FAF809FAF80A0AF809FA080A0A07F9F9F7F9F9F7F9F9070908060806F5F705F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F3F5050607F5F6F806F7F90707F9F808FA08F90A0909FAF9F9FAF
A0A0B0A0A0B0B0AFBFB0A0B0BFAFB0B0A0AFB0A0AFAF9FA0AF9FA0AF909FAF9F9FA09090A0908F\
9F807F90807F8F706F806F607F6050604F40000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F3F404F406050507060608F5F608F6F70906F7F9F7080A07080A07F8FAF7F8FAF7F90AF\
7F9FAF80A0B0809FAF809FA07F9F9F809F9F7F9F907F9F8F6F807060806F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F3F5050607F5F6F806F7090707F90808FA08F8FA09090AF
9090AFA09FB0A09FAFAFA0B0AFA0AFB0A0B0B0A0AFB0A0AFAF9FA0AF9FA0AF909FAF909FA08F8F\
A08F8F9F808090807F8F706F806F607F60506F504F3F2F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004030504F40605F5F7F5F5F80606F90606F906F7FA07080A0708FA0708FA07F90AF7F90A0\
809FAF809FAF809FA07F909F7F9F9F7090907F9F906F8F7F60807050705F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F40505060805F6F806F7F9070809F808FA08F90AF
909FB09F9FAFA09FB0A09FAFAFA0B0AFA0AFB0A0B0B0A0AFB0A0AFAF9FA0AF9FA0AF909FAF9F9F\
A09090A090909F8F809F80808F7F708F706F7F605F7F60505F403F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4F3F5F4F3F60504F705F5F80606F90606F906F709F6F7FA07F8FAF708FA07F8FAF708FA07F9FAF\
7F9FAF80A0AF7F9FA07F9FA07F90907F9F9070908F6F8F7F5F7F60507050000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F405050607F5F6F806F7F90707F9F808FA0
8F90AF9090AF9090AF9F9FAF9F9FAFA09FAFA09FAFAFA0B0AFA0AFB0A0AFAF9FA0AF9FA0A0909F\
AF909FA08F90A08F909F8F809F80808F70708F706F806F607F605F705F4F6F5040
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000004F4050
4F3F5F50406F504F7060608F60608F6F70906070907080A07080A0708FA0708FA07F90A07F90A0\
809FAF7F90A07F9F9F70909070909070908F6F8F80607F6F5F705F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F405F5060805F6F806F7F90707F9F
808FA0808FA09090AF9090AF9F9FAF9F90AFA09FB0A09FAFAFA0B0AF9FAFB0A0AFAF9FA0AF9FA0\
A0909FAF909FA09090A090909F80809F8F8F90807F907F708F6F6F806F607F6050
705F506F5040000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000503F4F504060
504F6F504F705F5F8060608F6F6F90607090707FA07080A0708FAF708FA07F90AF7090A07F9FAF\
7F9FA07F9FA070909F70909070908F70908F60807F5F7F6F4F6F50000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F405F5060805F6F8F6F7090
6F7F907F809F808FA08F90AF8F90A09090AF9090AF9F9FAF9F90AFA09FAFA09FAFAFA0AFA09FA0\
A09FA0A0909FA0909F9F8F909F8F909F808F9F808F907F7F907F7F8F70708F706F
7F605F7F605F70504F6F5040303030000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004F303F504060504F6F
504F6F5F507F3F4F50606F8F606F8F6F709F6F7F9F708FA07080A0708FA0708FA07F90A07090A0\
7F9FA070909F70909F6F8F8F70908F6F8F8060807050705F303F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000404F605060805F6F80
6F7090707F907F809F808FA09090AF8F90AF9090AF9090AF9F9FAF9F9FAFA09FAFA09FAFAFA0B0\
A09FA0AF9FA0A0909FA0909FA08F90A08F909F808F9F8F8F9080809F8080907F7F
907F70806F6F806F607F5F50705F506F4F4050403F000000000000000000000000000000000000\
000000000000000000000000000000000000503F40503F4F5F40605F4F6F5F4F70
605F80605F8060608F606F8F6F709F6F7F9F7080A07080A07F8FAF708FA07090A07090A07F9FAF\
7F90A070909F6F8F907090906F8F8F6F8F7F5F7F6050705F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F405F505F7F
5F608060708F6F709070809F7F809F8F90AF8F8FA09090AF9090AF9F9FAF9F90A09F9FAF9F9FAF\
A09FAFA090A0A09FA0A0909FA0909F9F8F909F8F909F808F9F808F907F80907F80
907F7F907F7F8F6F6F806F6F7F605F7F5F5F7050507050506F4F4F6F4F4F6040406040405F3F3F\
5F3F4050303F50303F4F303060404F5F404F5F40505F405F5F4F6F5F4F70605F7F
605F8060608F60608F6F6F906F70906F7F9F7080A0708FA07080A0708FA0708F9F7090A07090A0\
70909F6F8F906F8F906F8F8F6F8F805F7F7050706040604F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F3F50
4F507050608060708F6F7090707F9F7F809F8F8FAF8F90AF9090AF9090AF9F9FAF9F90AF9F9FAF\
9F9FAFA09FAFA09FA0A09FAFA090A0A090A09F8F9F9F8F9F9F80909F8F9090808F
908080907F7F907F7F8F70708F7070806F6F8F6F7080606F80606F7F5F607F5F6070505F70505F\
6F4F506F4F5F6F4F50604F5F604F5F604F60604F606050705F5070605F7F606080
6F6F8F60608F6F6F906F70906F7F9F6F7F9F708FAF7080A0708FA0708FA07090A07090A07F90A0\
70909F7090906F8F8F6F8F8F60807F5F7F6F4F6F50000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
30304F404F6F506080606F8F60708F707F90707F9F808FA08F8FA08F90AF8F90A09090AF9090A0\
9F90AF9F90A09F90AF9F90A09F90A09F90A0A090A09F8F9F9F8F9F9F80909F8090
907F8F907F8F907F80907F808F707F8F707F806F70806F707F606F7F606F7F5F607F5F6070505F\
7050606F50606F50606050606F506F60506F6F5070605070605F7F605F806F6F8F
60608F606F90606F906F709F6F709F7080A07080A0708FA06F809F708FA0708F9F7090A06F8F9F\
6F8F906F8F8F6F8F8F60807F5F80704F6F5F40604F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F3040404F60505F7F606F8F60708F707F90707F9F808FA08F8FA09090AF8F90AF9F9FAF\
9090AF9F90AF9090A09F90AF9F90A09F90A09F90A0A090A09F90A09F909F9F8F90
9F8F9090808F90808F907F80907F8F8F70808F7080806F7F806F7F8060708060707F606F7F6070\
705F6F706070705F70705F7F6F5F706F5F7F6F50706F60806060806F6F90606F8F
6F6F906F6F906F709F6F7F9F7080A07080A0708FA07080A0708FA0708FA07090A070909F70909F\
6F8F906F8F8F608080608F80507F6F4F6F5F30403F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F405F4F50705F6F8F606F8F6F7F90707F907F80A0808FA08F90AF8F8FAF\
9090AF8F8FA09090AF908FA09090A0908FA09F90A0908F9F9F90A0908F9F9F8F9F
9F8F9F9F8F9F9080909080908F7F8F907F8F8F70808F7080806F808070807F6F7F806F7F7F607F\
7F6F8070607F70607F6F5F7F6F5F7F6F5F7F6F60806060806F6F906F6F8F6F6F90
606F906F709F6F7090707FA06F80A07080A06F809F708FA06F8F9F708FA06F8F9F70909F6F8F90\
6F8F906080806080805F7F7050706040604F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F3F5040506F5F6F8F606F8F6F7F90707F907F80A08080A08F8FA0\
8F8FA09090AF8F8FA09090AF908FA09090AF908FA09090A0908FA09F90A0908F9F
9F8FA0908F9F9F8F9F9080909080908F7F909080908F7F8F8F7F8F80708F8F708F806F80806F80\
7F60807F6F807060807060806F6080706F8F6F6080706F906F6F8F6F70906F6F90
6F709F6F709F7080A07080A07080A06F809F708FA06F8F9F708FA06F8F9F7090A06F8F9F6F8F90\
60808F608F805F807F507F6F406F5F40604F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000030304F404F605060805F6F806070906F7F907F809F7F7F9F\
8080A08080A08F8FAF8F8FA08F8FAF8F8FA0908FA08F8FA0908FA0908F9F908FA0
90809F908F9F8F809F90809F8F7F908F7F908F7F908F7F9080708F80708F80708F80708F7F6F80\
7F6F8F706080706F8F6F6080706F8F6F6F8F706F906F6F8F6F70906F6F906F7090
6070906F7F9F6F7F9F6F80A06F7F9F6F809F6F809F6F8F9F6F809F708F9F6F8F9F6F8F9060808F\
60808F5F807F5F807F4F6F603F5F40000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F406050607F5F6F8F6F70906F7090707F9F\
707F9F7F809F7F809F808FA08F8FA08F90AF8F8FA09090AF8F8FA0908FA0908FA0
908FA08F809F908F9F8F809F8F809F8F7F908F7F9F807F908F7F90807090807F907F708F807090\
706F8F7F7090706F8F7F7090707090707090706F9070709F6F70906F709F6F7090
707FA06F80A07080A06F7F9F6F809F6F809F6F8F9F6F809F708FA06F8F9F6F8F9060808F608F8F\
5F80805F807F4F6F5F3F5F40000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F3F5F4F5F70506080606F8F606F8F\
6F70906F7090707F9F707F9F7F80A07F809F8F8FA08F8FA08F8FAF8F8FA08F8FA0
8F80A08F8FA080809F8F809F807F9F8F809F807F90807F9F807F90807F908070908070907F7090\
7F70907070907F70907070907070906F6F907070906F70906F709F6070906F7F9F
6F7F9F6F809F607F9F6F7F9F607F906F809F6080906F8F9F6080906080905F7F805F80805F7F7F\
5F7F7F40605F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000303F5040506F4F5F7F5F6F8F\
5F6F8F6F70906F7090707F90707F907F809F7F809F8080A08080A08F8FAF8F8FA0
8F8FA08F80A08F8FA08080A08F80A080809F8F80A0807F9F80809F807F9F807F9F7F7F90807F9F\
7F709F7F7F9F7070907070906F709070709F6F7090707F9F6F70906F7F9F6F7F9F
6F809F6F7F9F6F809F607F906F809F6080906F8F9F60809060808F5F7F805F7F80507F7F507F70\
405F5F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F4060404F6F\
4F5F7F50608060708F606F8F6F70906F7090707F9F707F9F7F809F7F809F8080A0
7F809F8080A080809F8080A080809F8080A080809F8080A07F7F9F80809F7F7F9F7F7F9F707090\
707F9F7070907070906F70906F70906F7090707F9F6F70906F7F9F6F7F9F6F7F9F
607F90607F9F607F90607F90607F90608090607F8F5F7F8F50707F507F7F507070507070405F50\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303F5F\
3F405F40506F4F5F7F5F6F8F60708F6F70906F7090707F9F707F9F7F809F7F7F9F
7F80A07F7F9F7F809F7F7F9F7F80A07F7F9F7F809F7F7F9F8080A07F7F9F7F809F7F7F9F7F7F9F\
70709070709F6F7090707F9F6F709F707F9F6F7F9F707F9F6F7F9F6F809F607F9F
6F809F607F90607F90607F906080905F7F8F5F7F8F50707F507F7F4F70704F7070405050000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F3F503F4060404F6F505F7F50608060708F60708F6F7F906F7090707F9F
6F7F90707F9F6F7090707F9F707090707F9F707090707F9F707F9F7F7F9F707F9F707F9F6F7090\
6F709F6F70906F7F9F6F70906F7F9F6F70906F7F9F6F7F9F6F7F9F607F90607F90
607F90607F905F7F8F607F905F7F8F5F7F804F707F4F70704F6F6F4F6F6F3F4F4F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000030405F3F40604F50704F5F7F5F6F8F606F8F6F7F90607090
6F70906F70906F7F9F6F70906F7F9F6F7090707F9F6F7F90707F9F707F9F707F9F6F7F906F7F9F\
6F70906F7F9F6F7F9F6F7F9F6F7F9F6F7F9F607F906F7F9F607F906F7F9F607F90
607F905F708F607F905F7F8F507F804F6F7F4F7070406F6F406F6F3F4F4F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F3F5F3F4F6040506F50607F5F6F805F6F8F
5F6F8F6070906070906F70906070906F70906070906F7F9F6F70906F7F9F6F70906F7F9F607090\
6F7F9F6070906F7F9F6070906070905F708F6070905F708F6070905F708F607F90
5F708F5F7F8F5070805070804F6F70406F703F6060406060304040000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003040603F4F6F4F5F704F5F7F
506080506080506F80506F805F708F5F6F8F6070906070906F7F9F6070906F7F9F6070906F7F9F\
6070906070905F708F5F70905F6F8F5F708F50708F5F708F5F708F5F7F8F5F708F
5F7F8F5070805070804F6F704F6F703F60603F5F60304040000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030405F3F506F
3F506F405070405070405F70405F704F607F4F607F506F80506F805F708F5F6F8F5F708F5F6F8F\
5F708F506F80506F80506F80506F804F6F80506F80506F80507080506F80507080
4F6F7F4F6F7F40607040606F3F5F603F5F5F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030405030506F30506F3F506F3F506F405F70405F704F607F4F607F506F80506F80506F80\
4F607F4F6F804F607F4F6F804F607F4F6F804F6F7F4F607F4F5F6F4F6070405F6F
405F603F5F603F5F60000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000030303030404F3F506F3F506F3F5F70405F7040607F405F70\
40607F405F7040607F405F6F405F6F3F50604050603F4F5F3F4F5F304F50000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000303F4F3F5060\
3F506F3F50603F4F5F3F4F5F3F404F3F404F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[22]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001PM70006LF^:L75[a@081`00
1PLo00Ea0XX01Hn:ShZ?00::00=aRW40`P081`001PLg0004L@1aJ`9a00>:LHX01HX02I2:T8Z@RY2:
S`02RP05LHYaLF/0K@002f^:JhY[RV]aJg5[04D020L000H7=00018YaRW46RP0ET8Z@RY2:ZHZYT:V@
ZHnYRZV:T8Z?00B:0W5T00=a00B:LHYa1hX0176:LHX3LD0020L000H7<000;6^:LH^:RhZ@RY2:T8Z@
RZV;ZI2YT:V@ZI2YT:V@ZI2YRZV:ZHZ@RY2:ShZ?0XX00g1aJ`1F000:LF^:JhY[RW2:L0N:00n9RXV:
RHZ9RW2:JhY[RQP0>`081`001PL_0005LHZ:T8/00Y00:jV@ZI2YT:V@ZY2ZT:n@[jV_ZJnY[jV_ZJZ@
ZI2YT:V@ZI2YSjV?ZHnYRXl00XYA009a00>:LHX01HX04hn:Z8ZXRZR:Z8ZXRZR:Z8ZXRXl01XX00g40
000h00P700061bd000QaK8Z:T8Z@R`B@01>YT:V@ZI2YT:Z@ZY2ZT:Z@ZY2Z00:Y01B@ZI2YT:V@ZI2Y
SjV?ZHZ@RY2:S`::C@002W5[RV^:Q8Z9RXT4RP0IShZ?RXn:Z8ZXRZR:Z8Z?RHn9RXV:L8Y`RP0j00P7
00061b/000=aRW400X/2T00WZI2ZT:Z@ZY2_T:n@[jV_ZJnY[jZ_ZZnZ[jV_ZJnY[jV_ZJnY[jV_00BY
00J@ZI2YSjT2RTT00W400hYaRP05RP0PZ8ZXRZR:Z8ZXRZR?Z8nXSjR?Z8nXSjR:Z8ZXRXn:ShXj00P7
00061bX000M[RW6;LI2;00:@02>YT:V@ZI2ZT:Z@ZY2_T:n@[jZ_ZZnZ[jZ_ZJnY[jV_ZJZYZP08Z@06
T:V?ZHZ@0XX00f/000140006LF^:JhZ41XX08:R:Z8ZXRZR:Z8ZXRZR:Z8ZXRZR:Z8ZXRXn:ShZ?RHm`
>@081`001PLX000NRW6;RY2;ZI2ZT:Z@[i2_ZJnY[jZ_ZZnZ[jZ_ZZnZ0jl05ZZ_ZZnZ[jZ_ZZnY[jV_
ZJnY[jV_ZJl4Z@04T:V:S`::@P000g6:L@04RP0?XhZSRZR:Z8ZYZ:VXZJRY00bX00f?Z8nXSjR?Z8ZX
RXma03P020L000H79P002F^:LHYaT8Z@R`02T00[ZY2ZT:Z@ZY2_TJnA[i6_ZZnZ[jZ_ZZnZ[jZ_ZZnZ
[jZ_ZZnY[jV_ZJnYZP06Z@06T:V:ZHZ?0XX00f/0000l0004LF^:J`R:00bSRZR:Z:>XXjRSZ:<9Z00@
SjR?Z8nXSjR:Z8Z?RHn9ScP020L000H79002L@04RW6@RP:@03NYT:Z@[i2_ZZnZ[jZ_Z[2Z/:Z`[k2Z
/:n`[k2_/:n`[k2_/:Z`ZZnZ[jZ_ZZnZ[jZ_ZJnY[jV^00:Y00N@ZHnYRXn:03/000=aRW4018X0:j>:
XhZXRZVSZJRYZ<2XZJRYZ:VXZJRYZ:VXZJRYZ:VX[ZR^SjR?Z8nXSjP00Xlg00P700061b<000QaK8Ya
T8Z@R`:@03FYT:Z@ZY6_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZJnY
[P02Z@0:T:V?ZHZYRXmaRSL000IaJhY[RV/6RP06Z8ZXRZRS0ZP04JVXZJRYZ:VXZJRYZ:VXZJRY00>X
00nYSjR?Z8nXSjR:ShZ?RHX0=P081`001PLR009a03N@RY2;ZI2ZT:Z@[jZ_ZZnZ/:n`[k2Z/:Z`[k2Z
/:n`Z[2_/:n`[k2_/:n`[k2_/:n`[lN_/:n`00:_00ZZ[jZ_ZJnY[jV^0ZT0192YSjT2RS@00g44RP0_
XhZSRZV:ZJRYZ:VXZJS0Z<FXaJS5Z<FX[ZR^Z:jX[ZR^Z:jX[ZR^Sjj?ZHnXSjP00Xl00hYa000d00P7
00061b0000E/Rg6@R`04T00ZZY2ZT:ZA[jZ_TJnA[i6_TJnZ[jZ_TJnA[i6_TJnA[i6_TJnA[i6_ZZnZ
1Zl03K2Z/:Z_ZZnZ[jV_ZJl01ZT018nYRY02RP03J`0002`000=aJhX01HX0;J>:Z8ZYRZVSZJRYZ:VX
ZJRYZ:VXZJR^Z:jX[ZR^Z:jX[ZR^Z:jXZHnYSjV?Z004S`04RXm`RS@020L000H77`00776;RY2;ZI2Z
T:Z@[i6_ZZnZ/:Z`Z[2A/:Z`[k04[`0/UZnA[i6_TK2A/96`TK2A/:n`[k2_/:n`[k2_/:n`[k2_ajZ_
ZZnY[jV_ZJl4Z@05SjV:ShX0:`000g6:L@02RP0gXhZSRZV:ZJ>YZ:VX`:S0Z<FXaJW5ZLFYaJW5ZLFY
aJW5ZLFY[ZR^Z:jY[ZV^Z:j?[Xn^SjV?Z002S`03LHX0038020L000H77P001F^:LI2;00B@03NZT:ZA
[i6_ZZnA[i6_TJnA[i6_TJnA[i6_TJnA[i6FTJnA[i6_TJnA[i6_TJnA[i6_TK2Z/:Z`00:_00ZZ[jZ_
ZZnY[jV_1:T01XnYSjV:S`::9`00175[RW44RP0cXhZXRZVSZJ>YZ:VXZJRYZ:VX[ZR^Z:jYaJW5ZJjX
[ZR^Z:jX[ZR^Z:jXT8n^SjV?ZHn@00B?00=aSg00<`081`001PLM0003K8Ya00:@046ZT:Z@[i6_TJnZ
[jZ`Z[2Z/96`TK2A/9J_TJnA[i6_TJnA[i6_TJnA[i6_TJnA[i6_TJnA[i6_TK2Z/:n`[k2_a`04[`0A
ZZnY[jV_ZJjY[Y2YSjV:ShX08`000g6:L@04RP08XhZYRZVXZJP4Z@0]aJW5ZLFYaJW5ZLFYaJW5[Xn^
aJk5ZLFY[ZV^ZJjY[ZV^MjjY[Y2^Sjj?[Xn^00:?00>:S`00<P081`001PLM0004RW6@R`:@01nAT:Z@
ZY6_ZZnZ[jZ_TJnA[i6_TJnA[i6_TIJAUY6F00:A00F@TI2AT002T@0IUY6FTIJAUY6_TJnA[i6_TJnA
/:Z`[k2_/002[`08ZZnZ[jV_ZJl4Z@07SjV?ZHZ?L00P0004LF^:L@B:03VSRZR:ZJ>YZ:VXZJRYZ<2X
`:S5Z<FXaJW5ZLFYaJk5ZJjY[ZR^Z:jX[ZR^Z:jX[Xn^Sjj?[Xn@Si000Xl00hZ?L00c00P700061a`0
04^:LI2;ZY2ZT:nA[jZ_ZZnZ/:n`Z[2A/96`TJnA[i6_TJnA[i6FTIJAUY2FT9J@UY2FTIJA[i6_TJnA
[i6_TK2A/:Z`[k2_/:n`[k000jl02lJ_aZW6ZJnY[ZV^00:Y00>:S`007P02L@B:00JYRZV:ZJP6Z@0K
`:W5ZLFYaJW5ZLFYaJW5[/F^aJk5[/F^aJk500:^00FY[ZV^Z@03[P0<ZJj@[Y2^T:j?[Xn^18lc00P7
00061a/00W407H^:T8^AT:ZA[i6_TJnZ[jZ_ZZnA[i6_TJnAUY6F00BA01>@TI2AT95gTGNAMi5gTGNA
Mi6@00RA00f_TJnA[i6_TK2Z/:n`00B_00RZ[jZ_ZJnY[PBY00B?ZHZ:7@02L@0>RW6:RZ>:ZHZYRZVX
ZJP6Z@0AaJW5ZLFYaJW5ZLF^aJk5[/D01:h06:V^ZJjY[ZV^ZJjY[Y2^Sjj?UHn@Si2?T0B?00=a0000
<@081`001PLJ009a00n@Ri2@ZY2_TJnZ[jZ_Z[001:l09I6`TJnA[i6_TJnAUY6FT9J@TI2AMi5bT7NA
Mi5gTGNAT96@TI0019404IJA[i6`TK2A/:n`[k2_/:n`00>_00_6[lJYaZW6ZJnY[P02Z@::6`02L@::
00R?RZV:ZHZYZ0BY01o5ZLFYaJW5ZLFYaJW5ZLF^aJk5[/F^aJk5[/F^aJk500R^012E[Y2^T:j@[Y2^
T:j?UHn@0Xlc00P700061aX000EaRg6@R`02T00OZY2_TJnA[jZ_ZZnZ[i6_TJnA[i6FT9J@TI2AMi5g
T002M`0GLWMaLTibCWMaMg:@LY1gTGNAMi5gTI000Y403iJA[i6_TJnA/:n`[k2_/004[`08ZZnY[jV_
ZJh2Z@04Si2:RQP000=[LG4018X02:V:ZHZYXjVX0ZT05l2Y`:W5ZLFYaJW5ZLFYaJk5[/F^aJk500j^
01:@[Y2^T9F@UHnESiF?UHn@Si02S`03L@00034020L000H76@000g6;R`02T008ZY2_TJnA/:X2[`05
/:n_[k000Zl03Y6_TJnA[i2FT9J@TGN@0WL0179gLW89000]LW5gLWMbTGNAMi5gTI2FTJnA[i6`TK2A
/:n`[k2_/:n`[lN_aZo6ZLJY[jV^00:Y0XXF0003LHYa00::00JYRZV:ZJP4Z@0Q`935ZLFYaJW5ZLFY
aZk6[/J^aZk6[/J^aZk5[/J^aJk600J^01BE[Y2^UJj@[Y2^T9F@UI2ESiF?T0:?<`081`001PLH0005
K75aT8/00Y002ZZ@ZY6_TJnA[i43[`0BTJnA[i6_T9J@TGN@LWMbMg5b4`00179aLW42LP06Mg:AMi6@
0Y401YJA[i6_T@:_00N`[k2_/:n`00:_00RZ[jW6ZJnY[P:Y00B?T8Z:4`02L@03RW6:00>:00RYRZVS
ZJRYZ0:Y01O0ZLFYaJW5ZLFYaJW5[/F^aJk5[/F^a@0<[P0IT:j@[Y2^T9F@UI2ET9F@UI2ESi2?T7J?
MP0c00P700061aL000=aRg400Y007:Z@[i6_Z[2_/:n`[k2_/9J_TJnA[i6FT95gT78Q0005LY5gTI00
0Y4089JA[i6`TK2A/:n`[k2_/:o7[lJ_aZo6ZLJY[ZV^SjV:3`002G6:LI2:ZHZYRP06Z@0SaJW5ZLJY
aZk6[/J^a/G6aLK5a/G6aLJ^aZk6[/Z^bZk:[Zl01:h01YF^UJjE[P:E012@UI2ET9F@UI2ET9F@UI2E
0Xlc00P700061aH000E/Rg6@R`02T0:A01NZTJnA[i6_TJnA[i6_TJn@UY2AMi1bM`0T0009679bMg:A
Mi6@00:A00V_TJnA[i6`TK001Zl01ZZ_ZJnY[P:Y00F?ZHZ?R@0=0003K8Ya00::00J@RZV:ZJ<8Z@0E
aJW5ZLFYaJk6[/J^aJk6[/F^aJk500b^01JE[Y2ET9F@UI2ET9EgT7N@Mi1gUGNE0Y001Hn@Shma03<0
20L000H75@000g6;L@02T00GZY2ZTJnZ/:Z`[k2_/:n`UZnA[i6_TIH00Y02MbL00W80195gTI02T@0N
[i6_TK2A/96`[k2_/:o7[lJ_aZo6ZLJY[ZV^SjV:2`002G6;RY2:ZHZYR`04Z@0]`:W5ZLJYaZk6[/J^
a/G6aLK5b/G:aLK5aZk6[/J^bZk:[/Z^[jj_UJjE[YF^00:E01B@UI2EMiEgUGNFT9F@UI2ET9F@U@:?
<`081`001PLD0005675aT7800Y004i6@ZY6_TJnA[i6_TJnA[i6_T9H00Y000gN@L@0[000BCW9bT7:A
T96@UY6_TJnA[jZ`2:l01:V_ZJh2Z@05SjV?ShX02002L@0:RW6;RZV:ZHZYRPJY01O5ZLFYaJk6[/J^
aZk6[/J^aZk6[/F^a@05[P03[jj^00:^00RE[Y2^T9F@U@:@015gT7N@Mi1gT7N@Mi1gT7N@M`03T003
MXma03<020L000H74`00775bLI2;TI2ZTJnA/:Z`[k2_/:n`UZnA[i6_T9H2T003M`0002`00W808Y5g
TI2FTJnA[i6`TK2_/:n`[lN_aZo6[/JYaZV^ZJj?ZHX6009a00J;RZV:ZHX6Z@0[aJW6ZLJ^a/G6aLK5
a/G:aL[5b/G:aLZ^bZk:[/Z^bZk:[[B^[iF_UJnE[`02U@05T9EgT78017L03I5gUWNFT9J@UY2ET9D0
0Y02Sc<020L000H74`02L@0IRg:@Ri6@ZY6_TJnA[jZ_TJnA[i6_T9J@T@02Mc8000MaMg:AMi5g00:A
00VFTJnA[jZ`[k001Jl01LJY[jV^00BY0Xl01G00001/009a0XX028^:ZHZYRZVT1ZT07LFYaJk5[/J^
aZk6[/J:aZk6[/J^aZk6[Zn^[jj_00:^00fE[YF^T9F@UGN@Mg5/009b00AgLWMb0WL02I1gT7N@MiEg
U@04T003MXm`03<020L000H74`004W6@Ri6@ZY2_TK2Z/:n`[k2_/0:_00JA[i6_T9H2T3D001mhMi5g
TI2FTJnA/:n`[k2_/:o7[lJ_aZo6ZLJY[ZV^00:?00>:07400XX01X^:ZHZYR`BY02_0ZLJYaZW6[/K5
a/G6aLK5b/G:a/[5b/G:[/Z^bZk:[/Z^bZj_[ZnE[iF_00:E00F@M`00L@02LP0CT79hMgQgTGNFMiJ@
UY2FT9F@U@02T003MP00038020L000H74P001Fb;LI2;00:@0Y404jnA[i6_ZZnA[i6_TJn@UY2AMi00
=P001aQbLWMbTGL00Y402IJA[i6_Z[2_/004[`06Z/JY[jV^1:T01HYaLHYa00::00J;RZV:ZH/6Z@0O
aJW6ZLJ^aZk6[/J^aZk6[/J^aZk6[/J^[jj_[Zn^[`02[P0:UJj@UI2EMi1fM`<00W42LP0?Mg9gLY5g
TGNAMi5gUGNE00B@00=fS`00<`081`001PLB000ARg6@T:Z@[i6_Z[2_/:n`[k000Zl01YJ_TJn@UP:@
>@02LP0LN7:AT9JA[i6`[k2_/:n`[lN_aZo6[lJYaZV^Z@9a0XX0192:ZH/6Z@0:aZW6ZLJYa/G6a@;6
01g:a/[6b/K:a/[6b/K:[/Z^bZk:[/Z^]:j_[ZnE[`02U@03Mi0000@00W42LP0CT7NAMi5gUWNFT9J@
UY2FT9F@U@02T003MP00038020L000H74@02L@0>T7:@T96@ZY6_TJnA[jX3[`09TJn@UY2FMi1a03/0
00DHMg:AM`04T@07[i6_Z[2_/005[`06aZW6ZJjY0W42RP06RhZ@RZV;1ZT08JjYaZW6[/J^aZk6[/J^
aZk:[/Z^aZk6[/J^[jj_[Zn^[`02[P07UJj@UGN@MP070003K79b009b01B@LY1gTGNAMi5gTGNFT9Eg
T8n@MSD020L000H74@004W6;Ri6@ZY6_TK2_/:n`[k2_/0:_00BA[i6_0Y000gL0000l000<Mg:AT9JA
[i6`[k2_0k002Jo7[lJ_aZo6[P02L@06RhZYRjV@1:T02ZZYaZW6ZLJYaZh2aP0Ob/K:a/[6b/K:a/[6
bZo:[lZ_bZo:[lZ^[jj_UJnE[`02M`T00W805h]bTGNAT96@UY2FT9J@UY2FT9F@UI2?03D020L000H7
40000fabL@02T00CTI2ZTJnA[i6_TJnF[i6_TJn@T@02T48000EaMg:AT002T@07[i6_TK2_/006[`09
LHYaT8ZYRZV;00JY02R_ZLJYaZk6[/J^bZk:[lZ_bZo:[lJ_aZo6[lJ^[jj_[ZnE[YF^T9Eg3002L@0C
Rg:@LY5gTGNAT96@UY2FT9J@U@02T003K90003D020L000H74002LP:@00fAT:nA/:Z`Z[2_/:n`00:_
00BA[i6_0Y000gL000110003LI5g00:A00j_TK2A/96`[k2_ajo7L@::0X/00jV;Z@03Z@0:ZZW6ZLJY
aZo6[`;601o:a/[6b/K:a/[6b/K:[lZ_bZo:[lZ_bZj_UJnE[iFF00d00W4068]bTG:AT96@UY6FTIJA
UY2FT9J@UI2ELCH020L000H73`02L@04Rg:@T0:A00nZTJnA[i6_UZnA[i6_T9H00Y000g4000130005
LY1bTI000Y401jnA[i6`[k000Zl02G6:LH^:ZH^YR`04Z@0KZZW6ZLJ^aZk6[lJ_aZo:[lZ_bZo:[lZ_
aZo600B_00Z^[jj_UJj@UI1g3P002fabLY1bTG:AMi5g00:A00VFT9J@UY2FT9D00Y000g40000e00P7
00061`l000=aRh/00Y002jZA[i6`[k2_/:n`00:_00BF[i6_0i160004LG:AM`:A00R_TJnA/:n`[`9a
01B;RX^;ZH^YT:ZYZZW6ZLJYaZo6[`;601k:a/[6b/K:a/[6b/K:[lZ_bZo:[lZ_]:j_UJnE[gL@009b
01N@LY6@TI2AT9JAUY6FTIJ@UY2FT9F@S`0g00P700061`l000IbLI2;TI02T@0=[i6_TJnA[iJ_TJn@
[`02T003M`0004L000EbMg:AT002T@04[i6_[`9a00R:LH^:T8^YR`:Y01fZZJZYaZW6[lJ_aZo6[lZ_
bZo:[lZ_bZo:[lZ_aP04[`04[Zn^[`:E00>@U@004@00676;LY1bTI2AT96@UY6FTIJ@UY2FT9EfT3P0
20L000H73`005W6@T96@ZY6_TK2_/:n`[k2F[i6_T9H2T4X000UbTGNAT:nA[i400W404X^:T8^YRjV@
ZZV_ZLJYaZ[6[`;600K:a/[6b/H2bP0G1lX7a/Z_bZo:[lZ_bZo:[kB^[iF_T7L04P001g6;LY6;TI00
0Y4049JAUY6FTIJAUY2FT9J@UG4h00P700061`l00W82T00ATI2ZTJnA[i6_UZnA[i6_T:l00Y1<0005
CgMbTGL00Y42L@08Rg6;RY2;ZH/2Z@0MZZV_ZLJYaZo6[lJ_aZo:[lZ_bZo:[lZ_bZo:[lX01:l01:j_
UJl2U@03T000018000YaLY1bTG:AT96@0Y402IJAUY2FT9J@U@02T003MP0003L020L000H73P003g5b
T92AT:nA/:Z`[k2_/002[`04UZn@[`:@CP007g9hMi6@UW6;RY2;ZI2YT:ZY[jW6ZLJZaZo6[l[6b/H0
1<X05PO:1lK:[lZ_bZo:[lZ_]:j_UJnEUY0D009a00B@Ri6@0Y404IJAUY6_TJnAUY6_UIJ@UI2?03T0
20L000H73P000fb@R`02T0:A00J_TJnA[i43[`04TJn@[`:@00=a0000C@001digLW5/Rg400X/0192;
ZI02Z@0BZZV_ZLJZaZo6[lJ_bZo:[lZ_0lX02Zo:[lZ_bZo:[k@2[`03[ZnE00>E5P002G6;LY1bTI2A
T002T@0;UY6FTIJ@UY2FT9D00Y000g00000g00P700061`d0011/LX^AT:ZA[i6`[k2_/:n`0Zl01I6_
T:n@05400g42R`0_T8^YT:V@ZZVZZLJZaZo6[lJ_1jl7aPO:1lX7bPO:1jo:[lZ_bZo:[lZ^[iF_UGL0
5P006G6;LY6@TI2_TJnA[i6_UZnA[iJ_UIJ@UI00>P081`001PL=009a00Z@Ri6@ZY6_TJnA1Jl0196_
T9H2T5000g408h]aT8^YRjV@ZY2ZZJnZaZo6[lJ_aZo:[lZ_bZo:[lZ_bZo:00J_00F^[iFFT00I0009
K8]bT8^AT96@00:A00^FTIJAUY2FT9J@UP02T003MY0003T020L000H73002L@:@00^AT:nA/96`[k2_
/004[`05UZn@UY00CP02L@0[Rg6;Ri2;ZY2ZT:ZY[j[6[lJ_aZo7[`O61lH7bPO:1lX7]0N_bZo:[lZ_
]002[`03UJnE01T000EaRg:@R`02T@0C[i6_TJnA[iJ_UZnF[iF_T9J@U@0k00P700061``000=/LW80
0Y02T@06[i6_TJnA0jl019J_TJl4T4d017403X^:T8^YRjV@ZY2ZZJnZ0Zl05LJ_aZo6[lZ_bZl7]<Z_
bZo:[lZ_]004[`05[ZnEUI006P02L@06T7:@T96@0Y403IJAUY6_TJn@[i2FT9D00Y000g40000i00P7
00061`/000=aLW800Y002jZA[i6`Z[2_/:n`00B_00>F[i000Y1<009a02^;RY2;T8^YT:Z@[jV_Z/JZ
aZo7[lN_1jl7aPO:1lX7bPNd1jl7[lZ_bZnd00:_00>E[iD06P02L@0HT8^AT:ZA[i6_TJnF[iJ_UZnF
[i2_T9Ef?0081`001PL;009a00B;LY6@0Y401ZnA[i6_T@>_00BA[i2_0Y000gH0001:009a01::LH^;
T8^@RjZ@ZY2ZZJnZ[jX2[`0CaZo6[lJ_bZl7[`N_bZo:[lZ_]004[`05UJnEUI006P00275/Rg:@Ri6@
19402iJA[i6_TJn@[i2E00:@00=f0000>`081`001PL;000>LH^;TI2_TK2A/:Z`[k06[`04T:n@T4/0
0W42R`0]T8^YT:Z@ZY2_ZZnZaZ[7[lN_ajl7[`N_1lX7bPNd1k@7]0N_1jnd[kBE[iFF01`001YaRh^A
T96@[i6_TJnF[iJ_UZnF[iF_T9F@Scd020L000H72`02L@0<T8^AT:ZA[i6_TJnA0jl0196_T:l2T003
L@0004T000=aRg400X/0292;T8^YT:Z@0ZX01:nZ[jX2[`0GaZo6[lN_1jl7[`N_bZo:[lZ_]:ndUZl0
0YD00i2E000K0005K8]aT7800Y04T@0?[i6_TJnA[i6_T9J@UGN@03h020L000H72`000g6;R`02T@09
[i6`TK2_/:n`00B_00FF[i2ET01;0005LH^;T8/00Y001:Z@ZY02ZP0M[j[6Z/N_ajo7[`N_1jl7bPO:
1k@7]0N_1jl7[k@00Zl00iF_U@0M000JLH]bT8^AT:ZA[i6_TJnF[iJ_UZnF[i2FT9Do00P700061`/0
0W402Y2;TI6_TJnA[i43[`04UZn@[`B@BP000faaL@02R`08T8^@RjZ@ZY02ZP04[jZ_ZPB_016`[lN_
1jl7[`N_bZnd[kB_]002[`05UZnEUI007P02L@0ET7:@Ri6@ZY6_TJnA[iJ_TJnF[i2E00:@00=a0000
?@081`001PL;0003LH^;00:A00V_TK2_/:n`[k001:l01IJ_T92?04T000]aLW6@Ri6@ZY2ZT002ZP0Q
[j[7Z/N_ajo7[`N_1jl7]0Nd1k@7]0N_1jl7[kB_]9FF01l000j;LI2;TI2ZTJnA[i6_UP>_00VF[iJ_
UJn@UI00@0081`001PL;0004LW6@R`:A00J_TK2A/943[`04UZn@[`:@00>?0000A`000faaL@02R`08
T8^@T:Z@ZY42ZP04[jZ_ZP:_01>`[lN_ajl7[`N_1jo:[kB_]:nd00:_00FE[iFFRP0O000FLH]bT8^A
T:Z@[i6_TJnF[iJ_UZnEUP:@00=fL@00?`081`001PL;0003LI2;00:A00V_TK2_/:n`[k000Zl0192_
T:l2T4L00W409g9aT8^ZT:Z@ZY6_ZZnZ/:n`[lN_ajl7[`N_1k@7[`Nd1jl7[kF_]002[`04UIJE[P:Y
00>?00006P02L@0HT8^AT:ZA[i6_TI2F/9J_UZnF[iJ_T9F@@P081`001PL;0004LW6@R`:A00>_TK00
1Zl0196_T:l2T003L@0004@00W40479aRh^@Ri6@ZY2ZTJnZ[jX2[`0C/:n`[k2_/:l7[`N_1jne[kB_
]002[`0;UZnEUY2_ZJj?ZHX06`000fb;R`02T00ATI2ZTJnA[i6_UZnF[iJ_UJl00Y000gMa001100P7
00061`/000=aT8/00Y401jnA/:n`[k001:l0192_T:l2T4D00W409g9aRh^@RjZ@ZY6_Z[2Z/:Z`[lN_
ajl7/0N`1jl7[`N_1jl7[`N_]002[`06UIJ@aZk60Zh00hnYL@0I0003LH^;00:@00JZTJnA[i42[`0;
/:n_UZnF[iJ_T:l00Y1300P700061`X000<HLW400Y02T@03[i6`00J_00B@[i2_0Y1500Aa02F;LY2;
T8^ZT:ZA[jZ_Z[2_/:n`[k2_/:ne[`N_]Jne[kB_]:nd00:F0YD01:o6[jl2[P:Y0XlI000GRg6@Ri6@
ZY6_TJnA[iJ_UZnFT9F_T9H00Y1400P700061`X000AbLI2@0Y401jnA/:n`[k001:l01I2_T:j@04@0
0W40=g9aRh^@Ri6@ZY6_TK2Z/:Z`[lN_ak07/0N`1k07[`N_1jl7[kF_]IJ_UIJ_bZo:[lJ^[jV^ShX0
5P003G6;LY2;ZY2_TJnA[i400jl02iJ_UZnF[iF_T9Ea04@020L000H72P000aQbLP02T0:A00>_TK00
1Zl0192_T:l2T4@00g408h]bT8^@Ri6@ZY6_ZZnZ/:n`[k2_/:n`[kF_]Jne[kF_]Jnd00:_00BF[iFE
0jl02/J_aZj_ZJj?ZHXF009a01R@Ri2@ZY6ZTJnA[iJ_UZnF[iJ_T9J@UI1500P700061`X00W82T0:A
00N_TK2_/:n`00B_00F@[i2^T0130003LG9a00:;00B@Ri6@0Y40<:nA/:Z`Z[2_/:o7/0N`1k07/0N`
1jl7[kF_]JndUZnEUZn`[`N_bZo:[lJY[XnYLA@00W402Y2;TI2_TJnA[jX2[`0;/:n_UZnF[iF_T9H0
0Y1500P700061`X000<HLW800Y02T@03[i6`00J_00B@[i2_0Y13009a00R;LX^;T8^AT0:A01n_TJnZ
/:Z`[k2_/:n`[kF_]Jne[kF_]JndUZnF[i2F00F_00?:[lX00Zl2[P:Y0XlC009a00B;LY2;0Y401ZZA
[i6_UP>_00RF[iJ_UJn@U@:@AP081`001PL:009b0Y02T@07[i6`[k2_/004[`05T:n@[Xl0@@03L@:;
03V@Ri6@TI2ZTK2A/:Z`[k2_/:l7/0N`1k07/;F_]Jne[kF_]9J_UIJA/:n`[`N_1jo:[lJ^[jV^ShX0
4@02L@0:T8^@T:Z@[i6_ZP:_00>`[jl00Zl01YJ_UZn@UP:@A`081`001PL:0005679bTI000Y400jnA
/006[`04T:n@[`:@@@002W5/LW:;LY2;TI02T@0<ZY6_TK2Z/:n`[k2_0[005;F_]Jne[kF_]JndUZnE
UY2ETJnF1:l03;B_bZo6[ZnY[XnYRQ000W4038]aT8^AT:ZA[i6_T@>_00RF[iJ_UZn@U@:@00==0000
A@081`001PL:009b0Y02T@09/96`[k2_/:n`00:_00F^[i2@S`0o0009LG9aRg:ARi6@00:A00ZZTK2A
/:n`[k2_0[005;F`]K07/;F_]Jne[kF_]9J_UIIb0Y403iJA/:n`[`N_1jl7[lZ^aP02[P:Y00>?0000
2`004W6;Ri2;TI2ZTJnA[jZ`[k2_/0:_00RF[iJ_T9J@MTT020L000H72P000g5bLP02T0:A00J`TK2A
[i43[`03T:n@00>@?P000aQaL@02LP06Rg:ARi6@0Y406JZA[i6_TK2_/:n`[k2_]Jne[kF_]JndUZl0
0YH00i2E6002LP06TI2ATJnA1:l00kB_bP02[`:^0ZT2S`03L00000X00W42R`:@016AT:ZA[i6_TJnF
[iJ_UZnEUP03T003MP0004P020L000H72P02LP0=TI2ATK2A/:n`[k2_/002[`05[Zn@T8l0?@03L@06
Rg:;Ri6@0Y402ZZA/96`TK2_/:l4/00A]K2e/;F_]Jne[kF_]IJ_UIH00P000g:@T002T@0C[i6`[k2_
1jl7[lZ_aZj_ZJj?Z00;000ILH^@Ri6@ZY6_TJnZ/:n`[k2_/9J_UZn@UP02T4/020L000H72@02L@9b
00RAT96A/96`T@F_00J@[i2@Si0m009a0W801X]bT7:AR`BA01N_TJnA/9J`[k2_/:n`[k2_/:n`U[2F
[`02UP09T9EaRVb:JhY[00:A00>_TJl01Jl01LZ_[jj_00:^00B@ZHn:2@00175/Rh/2T008TI2ZTJnA
[i43[`08UZnF[iJ_T9D2T4`020L000H72@001G5bLY6@00:A00V`TK2_/:n`[k000Zl01Jj_T92?03`0
0W42LP:;00>ARi400i402;2A/96`TK2_1;004kF`]K2e[kF_]JneU[FF[iF@RX/03XX02:l7[lZ_bZk6
0Zh2Z@03S`0000H00W42R`:@00JZTJnA[jX2[`05/:n`[k000Zl01iJ_UYJ@UG40C0081`001PL:0005
LG9bTI000Y401[2A/96_T@>_00J@[i2@Si0k000:LAQbLH]bT7:AR`BA01N_TJnA/96`U[2_/:n`[k2_
/9J`UZnF[`02UP:@00R:T8Z@RX^:R`J:00iaRW6:LHYaRV^:JhZY[P:?00=a00001@000fb;L@02T008
TI2ZTJZA[jX3[`08UZnF[iJ_T9H2T003L@0004/020L000H72P02LP0=TI2ATK2F/:n`[k2_/002[`05
[Zn@T8l0>`002G5bLX]bTH^AT002T@0:/96`TK2A/9J`[`:`02Ve/;F`]K2e[kF_]IJeUZnEUY2YT:V@
ZI2YT:V@ZHZYRZV:ZHZYRY2:R`08RP9a10000g6;R`02T006ZY6_TJnZ0Zl01K2_/:n`00:_00NF[iJ_
T9Ef04h020L000H72P000aQbLP02T006UY6`U[2F1Jl0192_T902ScX000<HLG400W8018]bTG86T@0g
[i6`TK2F/9J`[k2_/:n`U[2F[iJ_UIJ@UH^@Ri2;T8^@Ri2;T8ZYRZV:ZHZYRZV:T8Z@RY2:S`06RP0?
RHY[L@2@Ri6@ZY6_TJnZ00>_00^F[iJ_UZnE[i2EMP1?00P700061`X00W82T0:A00V`U[2_/:n`[k00
0Zl01Jj_T:V?03T00g42LP04TG:AT0:A00Z_TK2A/96`TK2F1;00:;F`]K2e[kFF]IJdUZn@UY2ZLZV@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI03Z@0>T:V@ZHnYRZV:ZHZXRZP2RP=a00V_ZZn_/:n`[k001:l01IJ_
T:n@050020L000H72P000aQbLP02T0:A00J`TK2F/9H3[`04T:n@[P:@>@000g5/LP03LP03T7:A00FA
04FFTK2A/96`U[2F/9J`U[2F/9J_UZnEUWNET:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZHZYRZV:ZHZ@RXl01XX0172:Jg42[`08UZnF[i2_T9D2T50020L000H72`000g:@T002T@09/96`[k2_
/:n`00:_00B^[i2^0Y0h0007LG9bRg:AT004T@0:/96`TK2A/9J`UPB`00Ve/;F_]IJeUZl00YH08Y2F
T:Z@ZY2ZT:ZYZZVZZJZY[jV_ZJnY[jV_ZJZYZZVZZJX8Z@0>T:V@ZHnYSjV:ZHZXRZP4RP9a00NF[i2_
T9Ea050020L000H72`00179aT782T@05[i6`[k001:l01:V_T:l2T003L@0003@000Aa679a0W80191b
TG84T@0CUY6FTINA/9J`U[2F/9J`U[2F[`02UP0XT9F@TI2ZT:Z@ZY2ZT:Z@ZY2ZT:nYZZVZZJZYZZVZ
ZJnY[jVZZJZYZPjY00j?ZHnYRZV:ZHZYRXn:S`B:00=[L@00C@081`001PL;0003LG9b00:A00V_TK2_
/:n`[k000Zl01:j_T:l2T3<01G401g:;LY5bTI000Y40DiNAUi6`TK2F/9J`Uk2F]IJeU[FF]IJ_UIJA
ZY6_TJnA[i6_TJnA[i6_TJnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[ZV^00JY
00Z?ZHnYRZR:Z8Z?0XX2LDT020L000H72`02L@09T7:ATIJA/9J`00J_00B@[i2@0Xlb0005K75/LAP0
1780195bTGP4T@0AUY6GTINF/9J`U[2F/9J`UZl00YH00i2ET@02T@0gZY6ZTJZAZY6_TJnA[i6_TJnZ
[jZ_ZZnZ[jZ_ZZnZ[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV^ZJjY[P08Z@0<Z:V?ZHZYRZR:Z8Z?0XX00g2:
J`1500P700061``00W82T@09[i6`U[2_/:n`00B_00F@[i2@S`0a00Ea0W8018]bTG84T@1KUi6GTK2F
/9J`Uk2F/9J`U[FF/9J_UIJA[i6_TJnA[i6_TJnA[i6_TJnA[i6_ZZnZ[jZ_ZZnZ[jZ_Z/JZaZ[6Z/JZ
[jZ_ZZnY[jW6ZLJYaZW6ZLJYaZV_ZJjY[P06Z@0:SjV?ZHZYRZR:Z0::00=aRP00@0081`001PL<0003
CWMb00BA00>`U[001Zl01Y2_T:j?T30000M/LG5b679a009b00IgLY5bTGP4T@0=Ui6GUYNFUiJGU[2F
/004UP03T9Ig00ZA03V_TJnA[i6_TJnA[i6_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jV_ZJnY
[jV_ZLJYaZV^ZJh01ZT03ZRYSjV?ZHZYRZV:ShZ?0XX0176:L74l00P700061``00W802i6@UY6`U[2_
/:n`00B_00F@[i2^S`0_0003LG9a00Ib00BALY5g19404INAUi6`U[2F/9J`U[2F/9J`00:F01n@UY2_
TJnA[i6_TJnA[i6_TJnA[i6_TJnA[jZ_ZZnZ00B_02n`Z[2_/:o7[lN_aZo6[lJ_aZo6[lJ_aZo6[lJ^
aZk6ZLJ^aZW6ZLJYaZV^ZJjY[P04Z@0:SjV?ZHnXSjR:S`B:>@081`001PL<0003679b00BA00>_U[00
1Zl0192_T:h2T2l00W40179aLQP4LP06TG:AN95h0Y403IJAUi6GUYNFUiJGUYL019H01I2FMi6@00ZA
01b_TJnA[i6_TJnA[i6_TJnA[jZ_ZZnZ[jZ_ZZnZ2:l08LJ_aZo6[lJ_aZo6[lJ^aZk6[/J^aZV_ZJnY
[ZV^ZJjY[P02Z@0>Z:VXZHnYSjV:Z8ZXRXl2RP04L8Y[LC@020L000H73002LP:@0Y401jnA/:n`[k00
1:l01ZV_T:j?T2`00g42LP03Rg9b00=b00BALY5h0Y404INAUi6GU[2F/9J`U[2F/9J`00:F03j@UY2F
TIJAUY6FTIJAUY6FTJnA[i6`TK2A/96`U[2_/:n`[k2_/:n`[k2_/:o7[lN_ajo7[lN_aZo6[lO6a`S6
01J_aZo6[lJ^aZk6ZLJYaZW6ZLJY[ZV^1:T02ZRYSjV?Z8ZXRXl2RP03LHYa030020L000H73@02LP08
TI2ATJnF/9H5[`04ZJn@[P:@00=a0000:P00175/LW42LP9a0W8027MbTG:AN95h0Y401iJAUi6GUYL0
29H01gNFMi5gTI0039405IJAUY6FTIJA[i6_TJnA[i6_TJnF/002[`03ZZn_00Z_02O6[lJ_aZo6[lJ_
aZo6[lJ_aZo6[lJ_aZk6[/JYaZW6ZJjY[ZV^ZJh00ZT05:RYZ:V?ZHZYRZR:Z8Z?RHY`RW1a;0081`00
1PL=0003LI2@00:A00N_TK2_/:n`00B_00BY[i2_0Y0/0007LG9bRg:@L@04LP04TGNAN0:A00nFTINA
UiJGU[2FUiJGU[000YH01I2FT96@00:A04FFTI6AUY6FTIJAUY6FTIJA[i6_TJnA[i6`TK2A/9J`U[2F
/:n`[k2_/:n`[lN_ajo7[lN_ajl7[`N_1lH7aPO61lK:a/X00/H06:o6[lJ_aZk6[/J^aZk6[/JYaJV^
ZJjY[P:Y00bXZHnYSjV?Z8ZXRXl2RP03LHYa02P020L000H73@00179aTI02T@06[i6`U[2F1Jl0192^
T902SbT000Aa679a0W80191bLTl2LP08N7:AN95hTGP2T@0@UY6GTIJAUY6FTIJ@UWN@LP9g01:@Mi1g
TGNAMi5gTGNAMi5hTGP8T@0BUY6FTIJAUY6FTIJA[i6_UZnF4Zl07LJ_aZo6[lZ_bZo:[lZ_aZo6[lJ_
aZk6[/J^aZj_00:^00RY[ZV^ZJjY[P:Y01FXZHnYSjV:Z8ZXRXn9ShV?L8Y`LF/08@081`001PL=0003
LG:@00:A00N_TK2F/:n`00J_00FY[i2YS`0Y0009LG9bT7:@LY5a009b00AhMi5h19403IJAUi6GTINF
UiJGUYL00YH05i2FMgQgN7MhMi5gTGNAMi5gTGNAMi5g00BA04nFTIJAUY6FTIJA[i6`TK2A/9J`U[2F
/9J`[k2_/:n`[k2_]Jo7[lN_1jl7[`N_1jl7[`N_1jl7aPO6b/K:[lZ_bZo:[/Z^aZk6[/J^aZk500:^
01BY[ZV^ZJjY[XnYSjV?ZHnXRZR:S`B:00=aRP0070081`001PL>0003LI1g00:A00JFTK2F/9H5[`04
T:n@Z@:@:0002aQaK79aMg:@679?009b00QhLY5hTGRAN0:A02BFTIJAUY6FTIJAUWN@MXY[LTibLGMa
Mg9gLWMbMg9gLWMbMg82M`0STGNAMi5gTGNAN9JAUY6FTIJAUY6FTIJAUY6_TJnF[iJ_U[003Jl07LJ_
aZo6[lJ_bZo6[lJ_aZo6[lJ_aZk6[Zn^[jj_00:^00RY[ZV^ZJjY[P:Y01F?ZHnYSjV?Z8Z?RXn:ShV:
RHY`RV/06`081`001PL>000=LWNATJnA/9J`[k2_/004[`06[ZnY[i2?9`02L@9b00>@LY001G82N004
TGRAN0:A00^GTINAUi6GTINFU`02UP03T9Ig00Z:00Z?RXn:Sg5bLG9a17L0EGQgN7NAMi5gTGNFT9JA
UY6FTIJAUY6FTJnF[iJ`U[2F/9J`U[2_/:ne[kF_]Jl7[`N_1jl7[`N_1jl7[`N_1jo:[lZ_bZo:[lZ_
bZk:[/J^aZk6[/H01:h03ZV^ZJjY[ZR^SjV?ZHnX0Xl018Z?RXl2RP03LHY>01P020L000H73P000dj@
LP02T@07UY6`U[2F/006[`05T:j@T7409`000aQbL@02LP06Mg9g679?0W80>WQbTGRAN95hTGRFN9JA
UWRFN9IgUWN:L8YaRW2:L8Y`RW2?L8n9ShZ?RXn:ShZ?RXn:SdmgLWMbMg82M`06TGNAMi5g19404YJA
UY6FTIJAUY6_TJnF[iJ_UPB_00>`[jl04Jl03lJ_aZo:[lZ_bZk6[Zn^[`08[P0<ZJjY[ZR^Z:j?ZHnY
18l038Z?RXn9RW2:L8Y[LAH020L000H73`0037:AT9JA/96`[k2_/0J_00FY[i2@RP0V00Ab00>@LY40
0W800g5bLP02N00DTGRAN9JAUi6GTINAUi6FTIJ@UY02RP0BShZ?RXn:ShZ?RXn:ShZ?RXn:18l05Y2?
T8nYSjV?ZHnYShYgN7MhMi5gTGL2T@13UY6FTIJAUY6_TJnA/9J`U[2F/9J`[k2_]Jne[kF_]Jne[`N_
1jl7[lZ_bZo:[lZ_bZo:[lZ_bZo:[/Z^aZk6[/J^[`02[P0@ZJjY[ZV^ZJjX[XnYSjV?Z0:?00N:ShZ?
LHYa01H020L000H73`02LP09TI2ATIJA/9J`00J_00BY[i2^0XlU009a00YbLGMbMg9g679?0W80;WQb
N7RAN95hTGRAN9IhUWRFMiIgT72:L8Y`Sg2?RHn:ShZ?RXn:ShZ?RXn:ShX2S`0<T8n@Si2?T6^:LI2;
0W42T004Mg9gLPAg012AMi5gUY6FTIJAUY6FTIJA0YH02:nF[iJ_UZnF5:l03KB_]:o:[[B^[jj_[Zl0
1Zh03ZV^ZJjY[Xn^Sjj?ZHnY18l02HZ?RXm`RW2:J`0E00P700061`l000=bT7L00Y402JnA/9J`[k2_
/004[`04[ZnY[P:?90002G5bLWMbTG:AM`04LP9h01ZAN95hUY6FTIJAUY6FTIJ@UY2FMXn:ShZ?RPR?
01JYSi2?ZHnYSjV?ZHnYSjV@ZI2^LH^:0Y001:V@LW82T@06[i6`[gMb0WL037QgTGNFT9JAUY6FT@:F
03F_UZnF/9J`U[2F]Jne[kF_]Jne[kF_]Jnd[lZ_bZo:[lZ_bZo:[lZ_bZo:[/Z^aZj_[Zn^[`04[P0>
ZJjY[Y2^Sjj?ZHnXSjP2S`05RXmaRW4050081`001PL?0006LW6@LY6@0Y400jnA/008[`05ZJj@T8X0
9002L@9b00QgLWMbMaQbC`9b03ihLWQhTGRAN95hTGRFN9IgUWN@MXmaSg2?L8n:ShZ?RXn:ShZ?RXn:
T8n@Si2?T8n@Si2?T76;LI2;T8^Y609g0Y4019JA[iH4[`04Mg9gLP9g00BAMi5g0Y402YJAUY6FTIJA
UY42UP06[iJ_UZnF4:l00kB_[`02[`0:[Zn^[jj_[Zn^[`J^00bY[Y2^T:j?ZHnYSi06S`07RXm`Sg2:
600C00P700061a000W802i6@UY6_TK2_/:n`00J_00FY[i2YRP0S0009LG9bMg:ALY5g00Ab01IhMgQh
TGRFN9JAUY6FTIJ@UWNFMi1a38l07Y2?T8nYSjV?ZHnYSjV@ZI2YT:UaT8^YT:Z@[i2_LP:A00^FTK2F
/:n`[kF_]002[`03[WMb009g01JAMi6@UY6FTIJAUY6_UZnF[iJ_U[2F0Zl09K2_]Jne[kB_]:ne[kB_
]:nd[kB_]:nd[kB_]:jd[Zn^[jj_[Zl00Zh03:V^T:j@[Xn^SjV?Z@B?00B:Sg6:4`081`001PL@0005
679bTI000Y401IJA/9J`00J_00NY[i2YSi1`028000UbLG9bMg9gLY000W801DmbCgQb009h02jAN95h
TGRAMiIgTGN@Mi1aSg6?LHmaShZ?RXn:ShZ@RY2?T8n@Si2?T8n@Si2?0W42RP08T8^@T:Z@ZY02LP06
TI2ATJnA1jl02jj_T:j?T75gLWMb009g00j@Mi5gTGNFT9J@UY6FT@BF00J_UZnF[iH;[`0@[Zn^[jj_
[Zn^[jj_[Zn^[`:^00bE[Y2^T:j?[Xn@Si04S`07RXmaSg2?L00C00P700061a4000=aMg800Y402IJA
[iJ`[k2_/004[`04[ZnY[P:@00>:00007P02L@9b00N@Mi1gT7N@00=b01AhLWQgTGRAN9IhUWRFMiIg
UGN@MPJ?00j@Si2?T8n@Si2?T8n@S`B@00RET9F@[W6;RP:@00RZT:Z@[jZ_ZP:_00fAMi6A[i6`U[2_
]Jnd00:_00J^[jj^T:T9T00JMg9gMi5gTI2FTIJAUY6FTJnF[iJ_UZnF[iH8[`0C]9J_[ZnE[jj_[Zn^
[jj_[Zn^[`02[P0<T:j@[Y2^Sjj?[Xn@18l00g6?000B00P700061a8000EaMg:AT002T@:F00>`U[00
1Zl01:V^T:T2Sb0003dHLW5gLWMbMg9gLWM?LTmhLWQgN7NAMi5gTGNAMi1gT7J:MXmaSg6?LI2?T8n@
Si2?T8n@Si2?T8n@Si2?009a00B;RY2;0Y001ZZ@ZZZ_ZP>_00=bTGL00Y401:nA[iH7[`:^01F@ZHn@
LI1aT76@LI1aT75gCWMbMg800WL0395gUWNFMiJ@UY2FT0JF02F_UZnF[iF_UJnE[iF_UJnE[iF_UJjE
[YF^T:j@[Y2^SiF?T8n@00:?00EfSg6?L00C00P700061a8000AaLY2@0Y402IJA/9J`[k2_/004[`04
ZJn@Z@:?7`003G9aMg:@Mi1aT7N@Mi000g8097QgN7NAMiIgUWNFMiIgUGN@MY2?T8n@Si2?T8n@Si2?
T8n@S`J@00jELHYaT8^YT:Z@[jV_ZPB_016`[i1gTI2_TK2F/9Je[kB_]002[`:^00>@[P004@003g9g
LY1gTGNFT9J@UY2FT@04UP0W[iJ_UZnF[iJ_UJnE[iF_UJnE[iF_UJnE[iF^T:j@[Y2^T:j?[Xn@00B?
00=aS`004P081`001PLC009b00>@Mi400i400k2F/006[`04ZJn@[P:@00>:S`007@00AaQbLGMbMg:@
LWMbMg5gCgMbN79hMgQgTGNAMi1gT7N@MY1fT7J@MY1fT7J@MY1fT8n@Si2?T8n@Sg5/RW6@RY2;ZI2Z
T:nZ00N_00]aT7:AT9JA[i6_UP05[`06[Zn@[Xma5@0017MaMg82M`0OT7NFMiIgUY2FT9J@UY2FT9JE
UYF_UJnE[iF_UJnE[P02U@0>T9F@UI2ESiF?UHnESi02S`05MXmaSg004`081`001PLC0003LG9b00:A
00^FTK2F/:n`[k2_/002[`06[ZnY[Y2Y0XlM000^LW5gLY1gT7N@Mi1gT7N@LWQgN7MhMi5gUWNAMiEg
T7N@Mi2?T8n@Si2?T8n@S`F@00=aT9000i000g6@RP02T008ZI2ZT:n@[jX4[`04/:ne[`9b00fAT9JA
/96`U[2_]Jnd00:_00F^[i2^L@0J000aMg9gMi1gUY2FT9J@UY2FT9J@UY2_UJnE[iF_UJnE[iF_UJnE
[Y2^T9F@UI2ET9F?T002S`04MXmfMQ<020L000H750001DigLY6@00:A00F_U[2F/006[`04ZJj@Z@:@
00>:S`006`0031QbLGMbT7:@Mi1gT09g02maMdmgLWQbN7MhMi5gT7N@Mi1fT7J@MY1fT7J@MY1fT7J@
MY1fT7J@MY1aT8Z@R`02T006ZI2ZT:nZ1Zl02;2_/1QgLY6@0Y401:nA[iH5[`06[Zn^[Y2?7`000g9a
M`03M`0CT7N@Mi1gUWNFMiJ@UY2FT9JEUP02U@0FT9F@UI2ET9F@UHnESi1fT7J?MXmaMQ@020L000H7
5@000g:@M`02T@09[i6`U[2_/:n`00B_00JY[jV^T:T2Sa/00W80:WMbT7NAMi1gT7N@Mi1aN79hMgQg
N7NAMi1gT7N@Mi1gT7N@MY1fT8n@S`^@00f:T8^YT:Z@ZY2_ZJnZ00:_00b`[k2_]Jne[k1bTGL2T@09
[i6`U[2_]Jnd00:_00J^[jj^Sg4R0003LGMb009g02F@MiIgUWNFT9J@UY2FT9J@UY2FT9J@UI2ET9F@
UI2ET9EfT7J?009f00=a00004P081`001PLE009b00Z@Mi6AUY6`U[2F1Jl01Zj_ZJj@Z@:@00>:S`00
6@005AQbLWMbT7N@Mi1gT7N@LGIaMdmgLP08M`0LT7N@MY1fT7J@MY1fT7J@MY1fT7J@MY1fRhZ@RP:@
00RYT:Z@[i2_ZPJ_00>`[k000Zl01IIbLY5g00:A00B_TJnF1jl2[P:@00=a00009@000digL@02M`0M
T7N@Mi1gUGN@Mi1gT7N@MiEgUGNEMiEfT7J@MY000WH00g5fL@0E00P700061aH00W802i6@[i6`U[2_
/:n`00J_00RY[jV^T:V:SaT00W804WMbT7NAMi5gT7N@Mi1aS`1gLP9g00=hMgP00gL05I1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gL@02M`03RY2;00:@00ZYT:Z@[jV_ZZnZ0Zl02K2_]Jne[kF_/002[`0>LI1gTI2F
TK2F/:ne[k@4[`:^00>?T000:P007gMaMgN@MiIgUGNEMiEgUGNEMiEgUGNEMiEgUGN@MY000gH00g40
000C00P700061aH000=>LW800Y401iJA[iJ`U[001Zl01Zj_ZJj@Z@:@00>:S`005`004AQbLWMbT7N@
Mi1gT7N@MY1a008000IbCWM?Mg87M`0GMWMfMgJ@MY1aT75gLGMaMg5gRY2:T8/00Y002:Z@[jV_ZZnZ
1:l00k2_/004[`0=UZn@[g6@LY6@UY6_T@07[`06[Zn@T8ma;`00179>Mg4:M`0>T7J@MY1fT7IgLGIa
MW4G00P700061aL000=bMg800Y402jnA/9J`[k2_/:n`00B_00J^[jV_ZJh2T003RP0001@000]aLWMb
T7NAMi5gT@02T005Mi1fSg401`002W9gLWMbMg9gLWL2LPL00W42R`0:ZI2YT:Z@[jV_ZPB_00n`[k2_
]Jne[kF_]IJ_UZl00YH01I1bLY6@00:A00V_U[2F/:ne[k@00Zl01Jj_T:j:03T000=aMg401gL01GIg
MWMa01X020L000H760000digM`04T@06UY6`U[2F1Jl02:Z_ZJn@ZI2Y0Xl00hX0000D009b00egLY1g
T7N@Mi1gT7J@009a60001f^:LH^:T8/00Y001ZV@ZY2_Z@J_00F`[jn_/002[`0;UZnF[i6_T9IgT000
0W80295gTI6FTJnF1Jl019F_UID2Sd4000=aJ`007P081`001PLI000@LY6@UY6FTK2F/:n`[k2_/0B_
00J^[jV_ZJh2T003RXl001<000YaMg:@Mi5gTI2F19000hn@L@0E0005LHZ:RhX00Y002ZV@ZY2_ZJnZ
[jX2[`0D/:o7[`N`1jne[kF_/9J_UZn@UY030003LI1g00:A00V_TJnF/:ne[k@00Zl01Jj_T:j?0680
20L000H76P000g:@M`04T@06UY6_U[2F1Jl03JZ_ZJj@ZI2YSi2:Sg404P02L@0?Mg:@Mi1gT7N@Mi1g
T76?018000=[LG400XX018^:T8X2T008ZI2ZT:nY[jX4[`07/:n`[k2_/002[`06UZnF[i2F0Y000gN@
0002009a00Z@Mi6AUY6_UZnF0jl019F_UJh2Sf8020L000H76P004GMbTI2FTIJA/9Je[k2_/:n`00B_
00b^[jV_ZJj@ZHn@RXl@000>LG9bT7NAT96@UY2FT9D2T003MW4000d00g4018^:T8/2T00:ZI2YT:ZY
[jV_ZP:_00g7[lN_1k07[kF_/:ne00:_00RF[iJ_T9IgT0D00W82T@09UY6_U[2_]Jne00B_00>@[Xl0
HP081`001PLK009b00>@Mi400i42UP04/9J`UPF_00nZ[jV_T:V@ZI2YSi2:Sg003P02L@0@Mg:@Mi1g
T7N@Mi1gT76?L0`00W4028YaRhZ@RY2;0Y002:V@ZI2_ZJnZ1Zl00k2_/006[`09UZnA[i2FMi1f00L0
00AaMg:A0Y003i6_UZnF[iJ_UJnE[Xn@L01Q00P700061a`0015gLY6@UY6FTK2F/:n`[k2_/004[`0>
[ZnY[jV_ZJj@ZHn@RXl<000>LG9bT7NAT9J@UY2FT9D2T0:?2003L@0BRhZ@RY2;ZI2YT:ZY[jV_ZJnZ
0Zl03lN_1jl7[`N_]Jn`[k2_/002[`04UZn@UP:@00=g00001P02LP0;TI2ATJnF/9Je[k@01:l0192^
T8mQ00P700061ad00W800i1gT@03T@06[iJ`U[2F1jl03jV_ZJj@ZI2YT:V:T8Z?L00;0009CW9bT7N@
Mi5g00F@00AfT75a1@02L@08RW6@RY2:T8X2T00:ZI2YT:ZYZZV_ZPJ_00>`[k001Zl01YJ_UZnAUP:@
00=gT0002@001G5bLY6@00:A00^_TJnF[iJ_UJnE[P02T003L@0005l020L000H77P004GMbTI2FTJnA
/9J`[kF_/:n`00J_00jY[jV_ZJn@[Y2YSjV:S`::00=a00001002LP0;T7NAT9J@UY2FT9H00i04L@B:
01:@RZV;ZI2YT:Z@[jV_ZJnY[jX4[`0=1jl7[`N_1jne[k2_/002[`08UZnF[i2FMi0;009b00^AT96A
[i6`U[2_]@04[`04UJj@Sf4020L000H77`02LP0<T7NATIJA[iJ`U[2F2Jl04:V_ZJj@ZI2YT:V:T8Z@
RXl4RP0CLHY`LG:@Mi5gTI1aJg5`RW6:L@06RP0DT8Z@RZV@ZI2YT:ZYZZV_ZJnZ[jX4[`07/:n`[k2_
/002[`08UZnF[i6FT9H2T003M`0000/000DHLW:@M`02T@0;[i6_UZnF[iF_UJh00Y000g40001O00P7
00061b0000mgLY6@UY6_TK2F/:n`[k002Zl05ZV_ZJnY[i2^T:V@ZHnYSi2:T8Z@RXl3RP0NShZ?RY2:
T8Z@RY2:ZI2YT:V@ZI2ZT:nY[jV_ZJnZ0Zl04LJ_aZo6[`N_1jl7[`N_]Jn`00:_00VF[iJ_T9J@UY00
3`000g:AM`02T@07[i6`U[2_/004[`04UJj@T64020L000H78P000digM`04T@07UY6_U[2_/00<[`04
ZJnY[`BY00Z@ZI2YT:V@ZI2Y0Y004hn@RY2?ZI2YT:V@ZI2YT:V@ZI000ZT04JZY[jV_ZJnZ[j[6[lJ_
aZo600>_00>`[jl00Zl019J_UZl2UP07T9J@T7N@L@0B009b00n@Mi6AUY6_UZnF[iJ_UJh00Y000hma
001O00P700061b<000=bTI000Y403iJA[iJ`U[2_/:n`[k2_/008[`0EZJnY[jV_ZJnY[jV_T:n@[i2Y
T:V@00BY012_T:nY[jV_ZJnY[jV_ZJnZ0Zl04lJ_aZo7[lN_ajo7[lN_1jn`[k000Zl029J_UZn@UY2F
0WLD0009LI1bTI2FTJnF00N_00F@[Y2EL@1P00P700061b@000=bMgL019402IJA[iJ`U[2F/00:[`0C
ZZnY[jV_ZJnY[jV_T:n@ZI2YT008Z@0>[jV_ZJnY[jV_ZJnY[jX4[`05aZo6[lH03:l01YJ_UYJ@UP:@
00=gT7405P00175bT782T@0=UY6_UZnF[iJ_UJn@U@02Sf0020L000H79002LP0CTGNATIJA[i6`U[2F
]Jn`[k2_/00<[`0dZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnYaZW6Z/J_aZo6[lJ_ajo7[lN_/:n`
[k2_/0:_00RF[iJ_T9J@UP9g60002G5gLY6@UY6_UP07[`05UJn@UG40H0081`001PLU009b0WL03i6@
TI6FTJnA[iJ`U[2F/00<[`0OZZnZ[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jZ_ZP02[`05aZo6[lH0
3:l019J_UZl2UP04TIJ@T@:@00=g00006@004W6@LY6@UY6_UZnF[iJ_UJn@U@:?H0081`001PLV009b
0WL04i6@UY6FTJnA/9J`U[2_/:n`[k001jl02lJ_aZo6[lJ_aZo600F_02_6[lJ_aZo6[lJ_aZo6[lJ_
aZo7[lJ_ajn`[k2_/:n`[k2F[iJ_UZn@UY2F00:@00=a00006P02LP09TI2FTJnF[iJ`00:_00NF[iF_
T9F?060020L000H7:0004DigLY1gTI2FTIJAUY6_U[2F01F_00>Z[jl02:l00lJ_[`0B[`0@UZnF[iJ_
UZnA[i2FT95gT09g7P001dj@LY5gUY400YH02jnF[iJ_UJn@UHn@060020L000H7:P02LP9g0Y406IJA
UY6_TK2F/:n`[k2_/:n`[k2_/:n`[k001Jl05lJ_[jo6[lJ_aZo6[lJ_aZo7[lJ_ajo600V_00f`UZnF
[iJ_TJn@UY2F00:@00=a00007P02LP0CTGNATJnA[iJ_UZnF[iF_T9F?L@1O00P700061b`0011>Mg:A
Mi6@UY6FTIJA[iJ`<:l039J_TJnAUY2FT9J@UP:@0WLR000E67MbTGNFTIJA[iJ_UZnE[i2ESi1`05l0
20L000H7;P02LP0OTGNATIJAUY6_U[2F/:n`[k2_/:n`[k2_/:n`[k2_/00;[`09/:n`[k2_/:n`00R_
00nF[iJ_TJn@UY2FT9IgT7409@000g:AM`02T@0=[i6_UZnF[iJ_UJn@[P02Sel020L000H7<@02LP12
Mg:@Mi6@UY6FTIJAUY6_TJnF[iJ_UZnF[iJ_UZnF[iJ_UZnF[iJ_UZnF[iJ_UZnF[i6_TIJAUY2FT9J@
TGN@LWM>:P02LP0BT7NFTIJA[i6_UZnE[i2ESi1aG`081`001PLc0005LWMgTGL00Y40<iJA[i6_TJnA
[i6_U[2F/9J`U[2_/:n`[k2_/9J`U[2F[iJ_UZnF[iJ_UZnA[i6_T9J@UP02T09g00=a0000:`000g6@
M`02T@0?[i6_U[2F[iJ_UJn@[Y2?05l020L000H7=@02LP0eMg:AMi6@TI2FT9JAUY6FTIJAUY6FTJnF
[iJ_TJnA[i6_TIJAUY6FTIJAUY6FT9J@UWNAMi000WL00dh0000^0004LG:@M`:A00^FTJnA[iJ_UJn@
U@02T003L@0005d020L000H7=`00=W9gMi5gTI2FT9JAUY6FTIJAUY6_TJnF[iJ_TJnA[i6_TJnA[i6F
TIJAUY6FTIJ@UY2FMi1aMc8000=aMgL00Y403JnA[iJ_UZnF[iF_T9D00Y1O00P700061cT0031gCWMb
Mg:@LY1gTGNAT96@TI2FTIJAUY2FT9J@UY2FT96@TI2AT96@TGNAMi1bMdhf0004LG:@LP:A00VFTIJA
[iJ_T:l00YD2T003L@0005d020L000H7?P0017MbMg82M`0OTGNAMi6@UY6FTIJAUY6FT9J@UY2FT9J@
TGNAMi1gT002M`03LI1a03L000=aMg800Y403jnA[i6_UZnF[iF_UJj@S`1O00P700061d@0021>Mg9g
LWMbT7N@Mi5gTGN@Mi1bMg9gLWMaMdigCW:@[P:@00>:S`00=P00175bT782T@0;UY6FTJnF[iF_T9D0
0Y000g40001M00P700061dT000=bMg801GL01g9gLWMbMg400Y403JnA[iJ_UZnF[i2_T:T00Xlf0003
LI1b00:A00nFTJnA[iJ_UZnF[i2ET8l0G`081`001PMF000>LGMbTI2FTJnF[i2_T:l2T004Si1aRSD0
0W803gMbTI2FTIJA[i6_T:n@U@02T003L@0005d020L000H7EP004g5bTI2FTJnF[iJ_UZn^[i2^T:T0
0XXd0003LI1b00:A00nFTJnA[iJ_UZnE[i2ET740G`081`001PMG009b0Y002i6@UY6_UZn@[i2^00:@
00>?T740=002LP0BT7:AT9JAUY6_TJn@UY2ESi1`G`081`001PMH000ALWNAT9JA[i6_UZnF[jV_T:h0
0Y02RS8000=aT7800Y403YJA[i6_UZnF[iF_T9F?H0081`001PMI000?T7:AT9J@[i6_UZn@[i2^00B@
00=a0000<002LP09T7:ATIJAUY6_00:F00JEUY2ESi1P00P700061eT000YbT7NAT:nA[iJ`1:l01:V_
T:h2T003RP0002l000=aT7800Y403YJAUY6_UZnF[i2_T9EfH0081`001PMI000ALG:@Mi6@UY6_UZnF
[i2_T:h00Y000hn@000`009b00>@LY400i401iJAUY6FT9H00Y000gJ@001O00P700061eX00W801Y6@
UY2_T@F_00R^[jV_T:V@Sc0000=aTG800Y402iJAUY6_UZnF[i2F00:@00=f0000GP081`001PMK0008
Mg:@Mi6@[i45[`04T:n@Z@:@0XX_009b00f@LY6AUY6FTIJAUY2F00:@00=fS`00G`081`001PMK000:
LGMbTI2_TJnF/0J_00JY[i2YShX^0003LI5b00:A00^FTIJA[iJ_UZn@U@02T003L@0005h020L000H7
G0000digLP02T0:A1jl01ZV_T:V?T0::;@02LP03T7:A00>A00NFTIJAUY2F00:@0WIP00P700061ed0
00MaT7NAT:nA00:_00>`[jl00Zl02:j_ZJj@ZHn::`00175bTG82T@0;UY6FTJnF[iFFT9D00Y000g40
001N00P700061ed00W401Y1bTI2FT@N_00JY[i2YSi02RRX000DHMW:@LP02T@09UY6FTIJ@UY2E00:@
00=fL@00G`081`001PMO0007LGMgTI2_T@02[`03/:o700B_00BY[i2Y0Y000g40000U0005LG9bTGL0
0Y402iJA[i6_TJn@[i2E00:@00=a0000GP081`001PMP0003LGMb00:@00BFT:nA1Jl0192_T:T2T0:?
00=`00009002L@9b012@Mi6@UY2FT9J@UY2EMi1fH@081`001PMP0008LG:@T9JA[i42[`03/:n_00:_
00J^[jV_T:T2T003L@0002<000EaT7:AT002T@0;UY6FTJnA[i2_T9D00Y1Q00P700061f400W42M`06
TI2AT:nA0jl01ZZ_T:n@Z@:@00>:T7408`0041QbLWMbT7NAT9J@UY2FT9H2T003Mi00064020L000H7
HP002W5bTI2FT:nA[jX5[`06ZJnY[i2Y0Y000g40000P009b00B@LY6@0Y4039JAUY6_TJn@UY2EMV80
20L000H7H`004g5bT7NAT96@[i6_ZZnY[i2YT:T00Y000hn@L@0Q000@LG9bT7:AMi6@UY2FT9J@UP:@
00=fS`00H@081`001PMT000:T7:AT9JA[i6_T@>_00bZ[jV_ZJn@ZHn@RW4O009b00BAMi6@0Y402IJA
UY6_T9J@UP02T6<020L000H7I00051QgLY1gTI2AT:nZ[jZ_ZJn@ZY2Y0Y0018Z@LG4M009a00n@LY1b
TI2AT9J@UY2FT9H00Y000g40001Q00P700061fD000UaT7NAT9J@[i401Jl02:Z_ZJnY[i2Y0Y02RQ/0
0W401Y1bTI2AT0:A00ZFTIJAUY2FT9IgI0081`001PMV000767MbT7NAT002T@0;[i6_ZZnY[i2ZT:T0
0Y0018Z@RXXJ000BK79bT7:@Mi5gTI2FT9J@UWNE0WMT00P700061fL001]aT7NAT9JA[i6_ZZnZ[jZ_
ZJn@[i2YT:V:T7405`001g5bLY1gTI000Y4039J@UY6FT9J@UY2EMV@020L000H7J@02LP0CT7NAT96@
[i2_T:n@[i2ZT:V@Z@02T005Si2:Sg404`005G5bLI1bT7:AMi5gTI2FT95gT7N@MP1U00P700061fX0
0W80796@TI2_T:nA[i6_TJn@[i2_T:Z@ZI2YRY2:Sg4?000FLH]bT76AMi6@UY2FT9J@UY2FT9EgT6D0
20L000H7J`02LP0CT7NAT96@TI2_T:n@[i2ZT:V@Z@04T004RY2:S`9a2`00675/Rg6@LY1bTGNAMi5g
TI2AMi1gT7N@LFD020L000H7K002LP0GTI2AT:n@[i6_T:nA[i2_T:n@ZY2YT:T00Y0018Z@RY02RP04
LHYaRP=a01F:LH]bT7:ALY6@TI2FT9J@UY2FT9H00Y02MVD020L000H7K@005G5bT7:AT96@TI2_T:n@
[i2ZT:Z@Z@04T00TRY2:T8Z@RY1aRW6:LH]aRg6;LY1bT7:AMi5gTGNAT95gT7N@0WIV00P700061fh0
0W806I5gTI2FT:n@[i6_TJn@[i2_T:Z@ZY2YT:T00Y0078^@RY2:T8Z@Ri2;T7:ALY5gTI2AT9J@UY2F
T9H2T09g00=a0000I0081`001PM^000F679bT7:@Mi6@TI2AT96@ZY2ZT:Z@Z@B@022;T8^@RY2:T76@
LY1bRg:@LY1bTGNAMi5gTGNAMi1gT09g0W5V00P700061g000W806I5gTI2AT:nA[i6_T:n@[i2_T:n@
ZY2ZT:T01Y001H^@Ri2;00:@00fAMi6@TI2AT9J@UY2F00:@00AgT7IfI`081`001PMa009a01>@LY1g
TI2AT96@TI2AT96@ZY2Z00J@01n;T8^@Ri1bT7:@LY1bTG:AMi5gTGNAMi5gT7N@MWMa06L020L000H7
LP02L@06T7NAT96@0Y404JnA[i2_T:n@[i2ZT:Z@ZY2A00N@01>AT96@TGNAT96@TI2FT9J@TGN@009g
00=a0000IP081`001PMd000FCWMbT7:AT96@TI2AT96@TI2AT96@T@:@01Z;T8^@Ri2;T7:@LY1bTGNA
Mi5gTGNAMi1gT09g00=>0000I`081`001PMf009b0WL0;96@TI2AT96@[i2ZT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@TGN@Mi1aK0081`001PMh000XCWMbT7:@LY5gTI2AT96@TI2AT96@TG:ALY1bT7:A
LY5gT7N@Mi1gT0Ag00=a0000J`081`001PMi000XLGMbTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TGNAMi5gT09g00=a0000K0081`001PMj000T679aMg:@LY1bT7N@Mi1gT7NAT95bTG:ALY5bTGNA
Mi1gT7N@0WL00g5gL@1_00P700061g`0025bLGMbTGNAT96@TI2AT96@TI2AT96@TI2AMi5gTGNAMi40
1GL00g40001]00P700061gh001`HLW5gLY1bT7:@LY1gT7N@LY5bTG:@LY1bT7N@0WL0179gLGL2LG00
20L000H7P002LP0IMg:@Mi1gTGNAT96@TGNAMi5gTGNAMi5gT004M`9aL@081`001PN2000MCW9aLW5g
LWMbMg:@LY1bMg9gLWMbMg9gLWMaMdh0L`081`001PN4009b00IgLWMbMg82M`06T79gLWMb0gL01G9g
LGMa07H020L000H7R@003g5>LTigLGMaMg5gCW9>LP1j00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Colors", "Subsection",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  StyleBox["The OpenGL Viewer ",
    FontFamily->"Times New Roman"],
  "support the ",
  StyleBox["SurfaceColor[RGBColor[_,_,_]]",
    FontFamily->"Courier"],
  " graphics primitives. Unfortunately, the 3Script package cannot store \
surface colors. The binary change the definitions inside the ",
  StyleBox["Graphics`ThreeScript",
    FontFamily->"Courier"],
  " context when it get's loaded. The surface color is now marked inside the \
3Script by the \"",
  StyleBox["surfacecolor",
    FontFamily->"Courier"],
  "\" directive. It doesn't matter if the 3Script file is loaded before or \
after the binary. If both are active in one session, the surface colors are \
properly exported. Now let's create an example:"
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(d1[x_List, t_] := \(D[#, t]\ &\)\ /@\ x\n\), 
    \(d2[x_List, t_] := \(D[#, t, t]\ &\)\ /@\ x\n\), 
    \(norm[x_List] := x/Sqrt[x . x]\n\), 
    \(FrenetFrame[x_List, t_, phi_, r_] := \n\t\ \ 
      Module[{xp, xpp, tangent, binormal, normal}, \n\t\t\ \ xp = d1[x, t]; \n
        \t\t\ \ xpp = d2[x, t]; \n\t\t\ \ tangent = norm[xp]; \n\t\t\ \ 
        binormal = norm[Cross[xp, xpp]]; \n\t\t\ \ 
        normal = Cross[binormal, tangent]; \n\t\t\ \ 
        x + normal*r*Cos[phi] + r*binormal*Sin[phi]\n\t\t\t]\)}], "Input",
  CellLabel->"In[122]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(General::"spell1" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(normal\)\" is similar \
to existing symbol \"\!\(Normal\)\"."\)], "Message",
  CellLabel->"From In[122]:=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[BoxData[
    \(r = \@3/3; \nc1 = {Cos[t], Sin[t] + r, \(-Cos[3*t]\)/3}; \n
    c2 = {Cos[t] + 1/2, Sin[t] - r/2, \(-Cos[3*t]\)/3}; \n
    c3 = {Cos[t] - 1/2, Sin[t] - r/2, \(-Cos[3*t]\)/3}; \n
    gr = \(ParametricPlot3D[
            Evaluate[FrenetFrame[#, t, phi, 0.2]], {phi, 0, 2  \[Pi]}, {t, 0, 
              2  \[Pi]}, PlotPoints -> {16, 48}, DisplayFunction -> Identity]
            \ &\)\ /@\ {c1, c2, c3}; \)], "Input",
  CellLabel->"In[11]:=",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  "Since ",
  StyleBox["ParametricPlot3D",
    FontFamily->"Courier"],
  " returns now a ",
  StyleBox["MeshGraphics3D",
    FontFamily->"Courier"],
  " object we must convert it explicitly to a ",
  StyleBox["Graphics3D",
    FontFamily->"Courier"],
  " object."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(threerings = 
      Show[Graphics3D[
          \(Apply[Join, #]\ &\)\ /@
            Transpose[{{{SurfaceColor[RGBColor[1, 0, 0]]}, {
                    SurfaceColor[RGBColor[0, 1, 0]]}, {
                    SurfaceColor[RGBColor[0, 0, 1]]}}, 
                \(First[Graphics3D[#]]\ &\)\ /@\ gr}]]]; \)\)], "Input",
  CellLabel->"In[16]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .76381 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
1.44654e-16 1.04543 -0.0910361 1.04543 [
[ 0 0 0 0 ]
[ 1 .76381 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02912 .25753 m
0 .44491 L
s
0 .44491 m
.38779 .76381 L
s
.38779 .76381 m
.39403 .59362 L
s
.39403 .59362 m
.02912 .25753 L
s
.69808 0 m
.9756 .42021 L
s
.9756 .42021 m
1 .60007 L
s
1 .60007 m
.70987 .19617 L
s
.70987 .19617 m
.69808 0 L
s
.02912 .25753 m
0 .44491 L
s
0 .44491 m
.70987 .19617 L
s
.70987 .19617 m
.69808 0 L
s
.69808 0 m
.02912 .25753 L
s
.39403 .59362 m
.9756 .42021 L
s
.9756 .42021 m
1 .60007 L
s
1 .60007 m
.38779 .76381 L
s
.38779 .76381 m
.39403 .59362 L
s
0 0 m
1 0 L
1 .76381 L
0 .76381 L
closepath
clip
newpath
.5 Mabswid
.50541 .5254 m .54147 .51709 L .542 .53174 L p .767 0 0 r
F P
0 g
s
.542 .53174 m .5083 .53983 L .50541 .5254 L p .767 0 0 r
F P
0 g
s
.57817 .53409 m .542 .53174 L .54147 .51709 L p .697 0 0 r
F P
0 g
s
.5083 .53983 m .542 .53174 L .54343 .54655 L p .751 0 0 r
F P
0 g
s
.542 .53174 m .57817 .53409 L .57504 .54862 L p .701 0 0 r
F P
0 g
s
.57504 .54862 m .54343 .54655 L .542 .53174 L p .701 0 0 r
F P
0 g
s
.54147 .51709 m .58022 .51954 L .57817 .53409 L p .697 0 0 r
F P
0 g
s
.61044 .54622 m .57817 .53409 L .58022 .51954 L p .623 0 0 r
F P
0 g
s
.57817 .53409 m .61044 .54622 L .60378 .55961 L p .647 0 0 r
F P
0 g
s
.60378 .55961 m .57504 .54862 L .57817 .53409 L p .647 0 0 r
F P
0 g
s
.5055 .51296 m .54192 .50503 L .54147 .51709 L p .772 0 0 r
F P
0 g
s
.54147 .51709 m .50541 .5254 L .5055 .51296 L p .772 0 0 r
F P
0 g
s
.58022 .51954 m .54147 .51709 L .54192 .50503 L p .681 0 0 r
F P
0 g
s
.54343 .54655 m .51369 .55384 L .5083 .53983 L p .751 0 0 r
F P
0 g
s
.51369 .55384 m .54343 .54655 L .54551 .55904 L p .72 0 0 r
F P
0 g
s
.54343 .54655 m .57504 .54862 L .57134 .56073 L p .696 0 0 r
F P
0 g
s
.57134 .56073 m .54551 .55904 L .54343 .54655 L p .696 0 0 r
F P
0 g
s
.54192 .50503 m .58084 .50741 L .58022 .51954 L p .681 0 0 r
F P
0 g
s
.61472 .53202 m .58022 .51954 L .58084 .50741 L p .582 0 0 r
F P
0 g
s
.58022 .51954 m .61472 .53202 L .61044 .54622 L p .623 0 0 r
F P
0 g
s
.57504 .54862 m .60378 .55961 L .59583 .56995 L p .663 0 0 r
F P
0 g
s
.59583 .56995 m .57134 .56073 L .57504 .54862 L p .663 0 0 r
F P
0 g
s
.47569 .54276 m .50541 .5254 L .5083 .53983 L p .812 0 0 r
F P
0 g
s
.5083 .53983 m .48097 .55685 L .47569 .54276 L p .812 0 0 r
F P
0 g
s
.48097 .55685 m .5083 .53983 L .51369 .55384 L p .775 0 0 r
F P
0 g
s
.50859 .50463 m .5433 .4976 L .54192 .50503 L p .746 0 0 r
F P
0 g
s
.54192 .50503 m .5055 .51296 L .50859 .50463 L p .746 0 0 r
F P
0 g
s
.58084 .50741 m .54192 .50503 L .5433 .4976 L p .62 0 0 r
F P
0 g
s
.54551 .55904 m .52069 .56509 L .51369 .55384 L p .72 0 0 r
F P
0 g
s
.52069 .56509 m .54551 .55904 L .54791 .5671 L p .635 0 0 r
F P
0 g
s
.54551 .55904 m .57134 .56073 L .56769 .56836 L p .658 0 0 r
F P
0 g
s
.56769 .56836 m .54791 .5671 L .54551 .55904 L p .658 0 0 r
F P
0 g
s
.5433 .4976 m .57991 .49976 L .58084 .50741 L p .62 0 0 r
F P
0 g
s
.61588 .51938 m .58084 .50741 L .57991 .49976 L p .471 0 0 r
F P
0 g
s
.58084 .50741 m .61588 .51938 L .61472 .53202 L p .582 0 0 r
F P
0 g
s
.51369 .55384 m .48934 .56933 L .48097 .55685 L p .775 0 0 r
F P
0 g
s
.48934 .56933 m .51369 .55384 L .52069 .56509 L p .712 0 0 r
F P
0 g
s
.57134 .56073 m .59583 .56995 L .58792 .57547 L p .662 0 0 r
F P
0 g
s
.58792 .57547 m .56769 .56836 L .57134 .56073 L p .662 0 0 r
F P
0 g
s
.63589 .567 m .61044 .54622 L .61472 .53202 L p .558 0 0 r
F P
0 g
s
.61044 .54622 m .63589 .567 L .62648 .57833 L p .598 0 0 r
F P
0 g
s
.62648 .57833 m .60378 .55961 L .61044 .54622 L p .598 0 0 r
F P
0 g
s
.47439 .52942 m .5055 .51296 L .50541 .5254 L p .838 0 0 r
F P
0 g
s
.50541 .5254 m .47569 .54276 L .47439 .52942 L p .838 0 0 r
F P
0 g
s
.64272 .55338 m .61472 .53202 L .61588 .51938 L p .491 0 0 r
F P
0 g
s
.61472 .53202 m .64272 .55338 L .63589 .567 L p .558 0 0 r
F P
0 g
s
.60378 .55961 m .62648 .57833 L .61607 .58548 L p .628 0 0 r
F P
0 g
s
.61607 .58548 m .59583 .56995 L .60378 .55961 L p .628 0 0 r
F P
0 g
s
.52069 .56509 m .49942 .57812 L .48934 .56933 L p .712 0 0 r
F P
0 g
s
.49942 .57812 m .52069 .56509 L .52812 .57167 L p .554 0 0 r
F P
0 g
s
.54791 .5671 m .52812 .57167 L .52069 .56509 L p .635 0 0 r
F P
0 g
s
.47733 .51909 m .50859 .50463 L .5055 .51296 L p .843 0 0 r
F P
0 g
s
.5055 .51296 m .47439 .52942 L .47733 .51909 L p .843 0 0 r
F P
0 g
s
.57991 .49976 m .61371 .51044 L .61588 .51938 L p .471 0 0 r
F P
0 g
s
.64584 .53977 m .61588 .51938 L .61371 .51044 L p .327 0 0 r
F P
0 g
s
.61588 .51938 m .64584 .53977 L .64272 .55338 L p .491 0 0 r
F P
0 g
s
.59583 .56995 m .61607 .58548 L .60642 .58719 L p .643 0 0 r
F P
0 g
s
.60642 .58719 m .58792 .57547 L .59583 .56995 L p .643 0 0 r
F P
0 g
s
.51419 .50188 m .54537 .49616 L .5433 .4976 L p .401 0 0 r
F P
0 g
s
.5433 .4976 m .50859 .50463 L .51419 .50188 L p .401 0 0 r
F P
0 g
s
.57991 .49976 m .5433 .4976 L .54537 .49616 L p .218 0 0 r
F P
0 g
s
.52812 .57167 m .54791 .5671 L .55022 .56928 L p .198 0 0 r
F P
0 g
s
.55022 .56928 m .54791 .5671 L p .55712 .56876 L .355 0 0 r
F P
0 g
s
.56471 .57012 m .55022 .56928 L p .55712 .56876 L .355 0 0 r
F P
0 g
s
.55022 .56928 m .55025 .56922 L p .54981 .56931 L .651 0 0 r
F P
0 g
s
.53473 .57239 m .52812 .57167 L p .54127 .57025 L .198 0 0 r
F P
0 g
s
.55022 .56928 m .53473 .57239 L p .54127 .57025 L .198 0 0 r
F P
0 g
s
.5095 .58169 m .52812 .57167 L .53473 .57239 L p 0 0 0 r
F P
0 g
s
.53473 .57239 m .53935 .56702 L p .52378 .57596 L .787 0 0 r
F P
0 g
s
.53473 .57239 m .51787 .57936 L .5095 .58169 L p 0 0 0 r
F P
0 g
s
.51787 .57936 m .53473 .57239 L p .52378 .57596 L .787 0 0 r
F P
0 g
s
.53473 .57239 m .55022 .56928 L p .54981 .56931 L .651 0 0 r
F P
0 g
s
.54791 .5671 m .56534 .56821 L p .55712 .56876 L .355 0 0 r
F P
0 g
s
.55025 .56922 m .55204 .56509 L p .54074 .56985 L .651 0 0 r
F P
0 g
s
.54074 .56985 m .53473 .57239 L .55025 .56922 L .651 0 0 r
F
.56769 .56836 m .56471 .57012 L p .55712 .56876 L .355 0 0 r
F P
0 g
s
.56534 .56821 m .56769 .56836 L p .55712 .56876 L .355 0 0 r
F P
0 g
s
.54537 .49616 m .57757 .49796 L .57991 .49976 L p .218 0 0 r
F P
0 g
s
.61371 .51044 m .57991 .49976 L .57757 .49796 L p 0 0 0 r
F P
0 g
s
.56769 .56836 m .58792 .57547 L .58274 .57522 L p .489 0 0 r
F P
0 g
s
.5794 .57416 m .56769 .56836 L .58274 .57522 L .489 0 0 r
F
.58141 .57516 m .56471 .57012 L .56769 .56836 L p F P
0 g
s
.57745 .56898 m .56294 .56563 L .56443 .56941 L p .305 0 0 r
F P
0 g
s
.56443 .56941 m .56471 .57012 L p .56626 .56999 L .305 0 0 r
F P
0 g
s
.57745 .56898 m .56443 .56941 L .56626 .56999 L .305 0 0 r
F
.58274 .57522 m .58141 .57516 L p .5794 .57416 L .489 0 0 r
F P
0 g
s
.45431 .57112 m .47569 .54276 L .48097 .55685 L p .796 0 0 r
F P
0 g
s
.48097 .55685 m .4635 .58339 L .45431 .57112 L p .796 0 0 r
F P
0 g
s
.4635 .58339 m .48097 .55685 L .48934 .56933 L p .723 0 0 r
F P
0 g
s
.4488 .55709 m .47439 .52942 L .47569 .54276 L p .843 0 0 r
F P
0 g
s
.47569 .54276 m .45431 .57112 L .4488 .55709 L p .843 0 0 r
F P
0 g
s
.48934 .56933 m .47484 .59181 L .4635 .58339 L p .723 0 0 r
F P
0 g
s
.47484 .59181 m .48934 .56933 L .49942 .57812 L p .57 0 0 r
F P
0 g
s
.64945 .59633 m .63589 .567 L .64272 .55338 L p .476 0 0 r
F P
0 g
s
.63589 .567 m .64945 .59633 L .6379 .6016 L p .522 0 0 r
F P
0 g
s
.6379 .6016 m .62648 .57833 L .63589 .567 L p .522 0 0 r
F P
0 g
s
.52812 .57167 m .5095 .58169 L .49942 .57812 L p .554 0 0 r
F P
0 g
s
.66015 .58627 m .64272 .55338 L .64584 .53977 L p .392 0 0 r
F P
0 g
s
.64272 .55338 m .66015 .58627 L .64945 .59633 L p .476 0 0 r
F P
0 g
s
.48405 .51357 m .51419 .50188 L .50859 .50463 L p .619 0 0 r
F P
0 g
s
.50859 .50463 m .47733 .51909 L .48405 .51357 L p .619 0 0 r
F P
0 g
s
.57757 .49796 m .60853 .50678 L .61371 .51044 L p 0 0 0 r
F P
0 g
s
.6447 .52845 m .61371 .51044 L .60853 .50678 L p 0 0 0 r
F P
0 g
s
.61371 .51044 m .6447 .52845 L .64584 .53977 L p .327 0 0 r
F P
0 g
s
.58792 .57547 m .60642 .58719 L .59918 .58309 L p .475 0 0 r
F P
0 g
s
.59918 .58309 m .58141 .57516 L .58792 .57547 L p .475 0 0 r
F P
0 g
s
.62648 .57833 m .6379 .6016 L .62748 .60114 L p .539 0 0 r
F P
0 g
s
.62748 .60114 m .61607 .58548 L .62648 .57833 L p .539 0 0 r
F P
0 g
s
.44792 .54364 m .47733 .51909 L .47439 .52942 L p .88 0 0 r
F P
0 g
s
.47439 .52942 m .4488 .55709 L .44792 .54364 L p .88 0 0 r
F P
0 g
s
.66822 .57312 m .64584 .53977 L .6447 .52845 L p .199 0 0 r
F P
0 g
s
.64584 .53977 m .66822 .57312 L .66015 .58627 L p .392 0 0 r
F P
0 g
s
.52138 .50528 m .54779 .50104 L .54537 .49616 L p .487 0 0 r
F P
0 g
s
.54537 .49616 m .51419 .50188 L .52138 .50528 L p .487 0 0 r
F P
0 g
s
.57757 .49796 m .54537 .49616 L .54779 .50104 L p .569 0 0 r
F P
0 g
s
.49942 .57812 m .48644 .59495 L .47484 .59181 L p .57 0 0 r
F P
0 g
s
.48644 .59495 m .49942 .57812 L .5095 .58169 L p .056 0 0 r
F P
0 g
s
.55022 .56928 m .56471 .57012 L .56294 .56563 L p .494 0 0 r
F P
0 g
s
.56294 .56563 m .55204 .56509 L .55022 .56928 L p .494 0 0 r
F P
0 g
s
.54779 .50104 m .5742 .50242 L .57757 .49796 L p .569 0 0 r
F P
0 g
s
.60853 .50678 m .57757 .49796 L .5742 .50242 L p .639 0 0 r
F P
0 g
s
.56471 .57012 m .58141 .57516 L .57745 .56898 L p .305 0 0 r
F P
0 g
s
.61607 .58548 m .62748 .60114 L .61998 .59495 L p .445 0 0 r
F P
0 g
s
.61998 .59495 m .60642 .58719 L .61607 .58548 L p .445 0 0 r
F P
0 g
s
.55204 .56509 m .53935 .56702 L .53473 .57239 L p .651 0 0 r
F P
0 g
s
.49344 .51389 m .52138 .50528 L .51419 .50188 L p .319 0 0 r
F P
0 g
s
.51419 .50188 m .48405 .51357 L .49344 .51389 L p .319 0 0 r
F P
0 g
s
.45185 .53306 m .48405 .51357 L .47733 .51909 L p .881 0 0 r
F P
0 g
s
.47733 .51909 m .44792 .54364 L .45185 .53306 L p .881 0 0 r
F P
0 g
s
.60853 .50678 m .63943 .52137 L .6447 .52845 L p 0 0 0 r
F P
0 g
s
.67232 .5591 m .6447 .52845 L .63943 .52137 L p 0 0 0 r
F P
0 g
s
.6447 .52845 m .67232 .5591 L .66822 .57312 L p .199 0 0 r
F P
0 g
s
.5742 .50242 m .6012 .50911 L .60853 .50678 L p .639 0 0 r
F P
0 g
s
.63943 .52137 m .60853 .50678 L .6012 .50911 L p .695 0 0 r
F P
0 g
s
.58141 .57516 m .59918 .58309 L .59566 .5738 L p .243 0 0 r
F P
0 g
s
.59566 .5738 m .57745 .56898 L .58141 .57516 L p .243 0 0 r
F P
0 g
s
.5095 .58169 m .4963 .59221 L .48644 .59495 L p .056 0 0 r
F P
0 g
s
.4963 .59221 m .5095 .58169 L .51787 .57936 L p .785 0 0 r
F P
0 g
s
.43606 .59532 m .4488 .55709 L .45431 .57112 L p .748 0 0 r
F P
0 g
s
.45431 .57112 m .44735 .60503 L .43606 .59532 L p .748 0 0 r
F P
0 g
s
.44735 .60503 m .45431 .57112 L .4635 .58339 L p .634 0 0 r
F P
0 g
s
.5289 .51434 m .55014 .51151 L .54779 .50104 L p .684 0 0 r
F P
0 g
s
.54779 .50104 m .52138 .50528 L .5289 .51434 L p .684 0 0 r
F P
0 g
s
.5742 .50242 m .54779 .50104 L .55014 .51151 L p .684 0 0 r
F P
0 g
s
.4635 .58339 m .45979 .60958 L .44735 .60503 L p .634 0 0 r
F P
0 g
s
.45979 .60958 m .4635 .58339 L .47484 .59181 L p .351 0 0 r
F P
0 g
s
.53935 .56702 m .55204 .56509 L .55304 .55518 L p .765 0 0 r
F P
0 g
s
.55204 .56509 m .56294 .56563 L .5627 .55556 L p .658 0 0 r
F P
0 g
s
.5627 .55556 m .55304 .55518 L .55204 .56509 L p .658 0 0 r
F P
0 g
s
.55014 .51151 m .57038 .51246 L .5742 .50242 L p .684 0 0 r
F P
0 g
s
.6012 .50911 m .5742 .50242 L .57038 .51246 L p .677 0 0 r
F P
0 g
s
.42782 .5821 m .44792 .54364 L .4488 .55709 L p .812 0 0 r
F P
0 g
s
.4488 .55709 m .43606 .59532 L .42782 .5821 L p .812 0 0 r
F P
0 g
s
.56294 .56563 m .57745 .56898 L .57678 .55793 L p .548 0 0 r
F P
0 g
s
.57678 .55793 m .5627 .55556 L .56294 .56563 L p .548 0 0 r
F P
0 g
s
.67883 .6088 m .66822 .57312 L .67232 .5591 L p .291 0 0 r
F P
0 g
s
.66822 .57312 m .67883 .6088 L .66701 .61588 L p .387 0 0 r
F P
0 g
s
.66701 .61588 m .66015 .58627 L .66822 .57312 L p .387 0 0 r
F P
0 g
s
.66015 .58627 m .66701 .61588 L .65527 .6172 L p .426 0 0 r
F P
0 g
s
.65527 .6172 m .64945 .59633 L .66015 .58627 L p .426 0 0 r
F P
0 g
s
.60642 .58719 m .61998 .59495 L .61676 .58403 L p .121 0 0 r
F P
0 g
s
.61676 .58403 m .59918 .58309 L .60642 .58719 L p .121 0 0 r
F P
0 g
s
.53935 .56702 m .52305 .57145 L .51787 .57936 L p .787 0 0 r
F P
0 g
s
.52305 .57145 m .53935 .56702 L .54115 .55642 L p .839 0 0 r
F P
0 g
s
.55304 .55518 m .54115 .55642 L .53935 .56702 L p .765 0 0 r
F P
0 g
s
.48405 .51357 m .45185 .53306 L .45997 .52718 L p .514 0 0 r
F P
0 g
s
.45997 .52718 m .49344 .51389 L .48405 .51357 L p .514 0 0 r
F P
0 g
s
.5039 .52006 m .5289 .51434 L .52138 .50528 L p .646 0 0 r
F P
0 g
s
.52138 .50528 m .49344 .51389 L .5039 .52006 L p .646 0 0 r
F P
0 g
s
.47484 .59181 m .47127 .60813 L .45979 .60958 L p .351 0 0 r
F P
0 g
s
.47127 .60813 m .47484 .59181 L .48644 .59495 L p 0 0 0 r
F P
0 g
s
.6012 .50911 m .6309 .51982 L .63943 .52137 L p .695 0 0 r
F P
0 g
s
.63943 .52137 m .67172 .54657 L .67232 .5591 L p 0 0 0 r
F P
0 g
s
.67172 .54657 m .63943 .52137 L .6309 .51982 L p .789 0 0 r
F P
0 g
s
.68876 .59719 m .67232 .5591 L .67172 .54657 L p .062 0 0 r
F P
0 g
s
.67232 .5591 m .68876 .59719 L .67883 .6088 L p .291 0 0 r
F P
0 g
s
.64945 .59633 m .65527 .6172 L .64564 .61245 L p .391 0 0 r
F P
0 g
s
.64564 .61245 m .6379 .6016 L .64945 .59633 L p .391 0 0 r
F P
0 g
s
.42528 .55425 m .43147 .54434 L p .4389 .53986 L .844 0 0 r
F P
0 g
s
.45185 .53306 m .44792 .54364 L p .44065 .55092 L .856 0 0 r
F P
0 g
s
.44013 .5476 m .45185 .53306 L p .44065 .55092 L .856 0 0 r
F P
0 g
s
.45185 .53306 m .42401 .5676 L .42528 .55425 L p .883 0 0 r
F P
0 g
s
.42528 .55425 m .45997 .52718 L .45185 .53306 L p .883 0 0 r
F P
0 g
s
.45997 .52718 m .43943 .54321 L p .4389 .53986 L .844 0 0 r
F P
0 g
s
.43943 .54321 m .42528 .55425 L p .4389 .53986 L .844 0 0 r
F P
0 g
s
.42401 .5676 m .44013 .5476 L p .44065 .55092 L .856 0 0 r
F P
0 g
s
.44792 .54364 m .42782 .5821 L .42401 .5676 L p .856 0 0 r
F P
0 g
s
.57038 .51246 m .59296 .51713 L .6012 .50911 L p .677 0 0 r
F P
0 g
s
.6309 .51982 m .6012 .50911 L .59296 .51713 L p .684 0 0 r
F P
0 g
s
.57745 .56898 m .59566 .5738 L .59653 .56091 L p .49 0 0 r
F P
0 g
s
.59653 .56091 m .57678 .55793 L .57745 .56898 L p .49 0 0 r
F P
0 g
s
.51787 .57936 m .5027 .58398 L .4963 .59221 L p .785 0 0 r
F P
0 g
s
.5027 .58398 m .51787 .57936 L .52305 .57145 L p .894 0 0 r
F P
0 g
s
.53542 .52749 m .552 .52576 L .55014 .51151 L p .736 0 0 r
F P
0 g
s
.55014 .51151 m .5289 .51434 L .53542 .52749 L p .736 0 0 r
F P
0 g
s
.57038 .51246 m .55014 .51151 L .552 .52576 L p .701 0 0 r
F P
0 g
s
.54115 .55642 m .55304 .55518 L .55303 .54124 L p .77 0 0 r
F P
0 g
s
.55304 .55518 m .5627 .55556 L .56406 .54164 L p .691 0 0 r
F P
0 g
s
.56406 .54164 m .55303 .54124 L .55304 .55518 L p .691 0 0 r
F P
0 g
s
.552 .52576 m .56678 .52636 L .57038 .51246 L p .701 0 0 r
F P
0 g
s
.59296 .51713 m .57038 .51246 L .56678 .52636 L p .664 0 0 r
F P
0 g
s
.5627 .55556 m .57678 .55793 L .57954 .54392 L p .612 0 0 r
F P
0 g
s
.57954 .54392 m .56406 .54164 L .5627 .55556 L p .612 0 0 r
F P
0 g
s
.59918 .58309 m .61676 .58403 L .61844 .57026 L p .455 0 0 r
F P
0 g
s
.61844 .57026 m .59566 .5738 L .59918 .58309 L p .455 0 0 r
F P
0 g
s
.6309 .51982 m .66649 .53766 L .67172 .54657 L p .789 0 0 r
F P
0 g
s
.69511 .58302 m .67172 .54657 L .66649 .53766 L p 0 0 0 r
F P
0 g
s
.67172 .54657 m .69511 .58302 L .68876 .59719 L p .062 0 0 r
F P
0 g
s
.6379 .6016 m .64564 .61245 L .63981 .6024 L p .074 0 0 r
F P
0 g
s
.63981 .6024 m .62748 .60114 L .6379 .6016 L p .074 0 0 r
F P
0 g
s
.52407 .55928 m .54115 .55642 L .53976 .54241 L p .82 0 0 r
F P
0 g
s
.55303 .54124 m .53976 .54241 L .54115 .55642 L p .77 0 0 r
F P
0 g
s
.54115 .55642 m .52407 .55928 L .52305 .57145 L p .839 0 0 r
F P
0 g
s
.48644 .59495 m .4798 .60086 L .47127 .60813 L p 0 0 0 r
F P
0 g
s
.4798 .60086 m .48644 .59495 L .4963 .59221 L p .931 0 0 r
F P
0 g
s
.51357 .53107 m .53542 .52749 L .5289 .51434 L p .742 0 0 r
F P
0 g
s
.5289 .51434 m .5039 .52006 L .51357 .53107 L p .742 0 0 r
F P
0 g
s
.53976 .54241 m .55303 .54124 L .552 .52576 L p .759 0 0 r
F P
0 g
s
.552 .52576 m .53542 .52749 L .53976 .54241 L p .759 0 0 r
F P
0 g
s
.56678 .52636 m .552 .52576 L .55303 .54124 L p .701 0 0 r
F P
0 g
s
.55303 .54124 m .56406 .54164 L .56678 .52636 L p .701 0 0 r
F P
0 g
s
.58528 .52946 m .56678 .52636 L .56406 .54164 L p .643 0 0 r
F P
0 g
s
.56678 .52636 m .58528 .52946 L .59296 .51713 L p .664 0 0 r
F P
0 g
s
.47093 .5271 m .5039 .52006 L .49344 .51389 L p .369 0 0 r
F P
0 g
s
.49344 .51389 m .45997 .52718 L .47093 .5271 L p .369 0 0 r
F P
0 g
s
.56406 .54164 m .57954 .54392 L .58528 .52946 L p .643 0 0 r
F P
0 g
s
.62055 .52414 m .59296 .51713 L .58528 .52946 L p .649 0 0 r
F P
0 g
s
.59296 .51713 m .62055 .52414 L .6309 .51982 L p .684 0 0 r
F P
0 g
s
.57678 .55793 m .59653 .56091 L .60168 .54668 L p .569 0 0 r
F P
0 g
s
.60168 .54668 m .57954 .54392 L .57678 .55793 L p .569 0 0 r
F P
0 g
s
.66649 .53766 m .6309 .51982 L .62055 .52414 L p .743 0 0 r
F P
0 g
s
.52305 .57145 m .50447 .57164 L .5027 .58398 L p .894 0 0 r
F P
0 g
s
.50447 .57164 m .52305 .57145 L .52407 .55928 L p .866 0 0 r
F P
0 g
s
.53976 .54241 m .52072 .54497 L .52407 .55928 L p .82 0 0 r
F P
0 g
s
.52072 .54497 m .53976 .54241 L .53542 .52749 L p .789 0 0 r
F P
0 g
s
.53542 .52749 m .51357 .53107 L .52072 .54497 L p .789 0 0 r
F P
0 g
s
.3446 .47026 m .38252 .46192 L .38327 .47712 L p 0 0 .73 r
F P
0 g
s
.38327 .47712 m .34779 .48522 L .3446 .47026 L p 0 0 .73 r
F P
0 g
s
.42052 .4795 m .38327 .47712 L .38252 .46192 L p 0 0 .745 r
F P
0 g
s
.34779 .48522 m .38327 .47712 L .38452 .49232 L p 0 0 .648 r
F P
0 g
s
.38327 .47712 m .42052 .4795 L .41707 .49443 L p 0 0 .664 r
F P
0 g
s
.41707 .49443 m .38452 .49232 L .38327 .47712 L p 0 0 .664 r
F P
0 g
s
.61019 .53365 m .58528 .52946 L .57954 .54392 L p .614 0 0 r
F P
0 g
s
.58528 .52946 m .61019 .53365 L .62055 .52414 L p .649 0 0 r
F P
0 g
s
.57954 .54392 m .60168 .54668 L .61019 .53365 L p .614 0 0 r
F P
0 g
s
.38252 .46192 m .42244 .46442 L .42052 .4795 L p 0 0 .745 r
F P
0 g
s
.45303 .49176 m .42052 .4795 L .42244 .46442 L p 0 0 .73 r
F P
0 g
s
.42052 .4795 m .45303 .49176 L .44596 .50551 L p 0 0 .647 r
F P
0 g
s
.44596 .50551 m .41707 .49443 L .42052 .4795 L p 0 0 .647 r
F P
0 g
s
.40576 .59567 m .42401 .5676 L .42782 .5821 L p .84 0 0 r
F P
0 g
s
.42782 .5821 m .41371 .6092 L .40576 .59567 L p .84 0 0 r
F P
0 g
s
.41371 .6092 m .42782 .5821 L .43606 .59532 L p .763 0 0 r
F P
0 g
s
.34413 .45722 m .3824 .44926 L .38252 .46192 L p 0 0 .813 r
F P
0 g
s
.38252 .46192 m .3446 .47026 L .34413 .45722 L p 0 0 .813 r
F P
0 g
s
.42244 .46442 m .38252 .46192 L .3824 .44926 L p 0 0 .828 r
F P
0 g
s
.62055 .52414 m .65747 .53396 L .66649 .53766 L p .743 0 0 r
F P
0 g
s
.66649 .53766 m .69683 .56867 L .69511 .58302 L p 0 0 0 r
F P
0 g
s
.69683 .56867 m .66649 .53766 L .65747 .53396 L p .84 0 0 r
F P
0 g
s
.43606 .59532 m .42499 .61879 L .41371 .6092 L p .763 0 0 r
F P
0 g
s
.42499 .61879 m .43606 .59532 L .44735 .60503 L p .61 0 0 r
F P
0 g
s
.38452 .49232 m .35317 .49962 L .34779 .48522 L p 0 0 .648 r
F P
0 g
s
.35317 .49962 m .38452 .49232 L .38608 .50502 L p 0 0 .536 r
F P
0 g
s
.38452 .49232 m .41707 .49443 L .41267 .50674 L p 0 0 .557 r
F P
0 g
s
.41267 .50674 m .38608 .50502 L .38452 .49232 L p 0 0 .557 r
F P
0 g
s
.59566 .5738 m .61844 .57026 L .62475 .55608 L p .552 0 0 r
F P
0 g
s
.62475 .55608 m .59653 .56091 L .59566 .5738 L p .552 0 0 r
F P
0 g
s
.62748 .60114 m .63981 .6024 L .63887 .58875 L p .379 0 0 r
F P
0 g
s
.63887 .58875 m .61998 .59495 L .62748 .60114 L p .379 0 0 r
F P
0 g
s
.3824 .44926 m .4225 .4517 L .42244 .46442 L p 0 0 .828 r
F P
0 g
s
.45727 .47703 m .42244 .46442 L .4225 .4517 L p 0 0 .816 r
F P
0 g
s
.42244 .46442 m .45727 .47703 L .45303 .49176 L p 0 0 .73 r
F P
0 g
s
.4963 .59221 m .48385 .58898 L .4798 .60086 L p .931 0 0 r
F P
0 g
s
.48385 .58898 m .4963 .59221 L .5027 .58398 L p .936 0 0 r
F P
0 g
s
.41707 .49443 m .44596 .50551 L .43721 .516 L p 0 0 .54 r
F P
0 g
s
.43721 .516 m .41267 .50674 L .41707 .49443 L p 0 0 .54 r
F P
0 g
s
.48283 .5329 m .51357 .53107 L .5039 .52006 L p .659 0 0 r
F P
0 g
s
.5039 .52006 m .47093 .5271 L .48283 .5329 L p .659 0 0 r
F P
0 g
s
.43147 .54434 m .47093 .5271 L .45997 .52718 L p .844 0 0 r
F P
0 g
s
.40248 .58047 m .42528 .55425 L .42401 .5676 L p .889 0 0 r
F P
0 g
s
.42401 .5676 m .40576 .59567 L .40248 .58047 L p .889 0 0 r
F P
0 g
s
.70586 .61876 m .68876 .59719 L .69511 .58302 L p .439 0 0 r
F P
0 g
s
.68876 .59719 m .70586 .61876 L .69463 .62719 L p .484 0 0 r
F P
0 g
s
.69463 .62719 m .67883 .6088 L .68876 .59719 L p .484 0 0 r
F P
0 g
s
.65747 .53396 m .62055 .52414 L .61019 .53365 L p .688 0 0 r
F P
0 g
s
.50125 .55733 m .52407 .55928 L .52072 .54497 L p .823 0 0 r
F P
0 g
s
.52407 .55928 m .50125 .55733 L .50447 .57164 L p .866 0 0 r
F P
0 g
s
.71427 .60584 m .69511 .58302 L .69683 .56867 L p .363 0 0 r
F P
0 g
s
.69511 .58302 m .71427 .60584 L .70586 .61876 L p .439 0 0 r
F P
0 g
s
.44735 .60503 m .43772 .62277 L .42499 .61879 L p .61 0 0 r
F P
0 g
s
.43772 .62277 m .44735 .60503 L .45979 .60958 L p .181 0 0 r
F P
0 g
s
.67883 .6088 m .69463 .62719 L .68246 .62969 L p .508 0 0 r
F P
0 g
s
.68246 .62969 m .66701 .61588 L .67883 .6088 L p .508 0 0 r
F P
0 g
s
.49358 .5436 m .52072 .54497 L .51357 .53107 L p .764 0 0 r
F P
0 g
s
.51357 .53107 m .48283 .5329 L .49358 .5436 L p .764 0 0 r
F P
0 g
s
.31273 .48778 m .3446 .47026 L .34779 .48522 L p 0 0 .681 r
F P
0 g
s
.34779 .48522 m .31832 .50237 L .31273 .48778 L p 0 0 .681 r
F P
0 g
s
.31832 .50237 m .34779 .48522 L .35317 .49962 L p 0 0 .593 r
F P
0 g
s
.52072 .54497 m .49358 .5436 L .50125 .55733 L p .823 0 0 r
F P
0 g
s
.59653 .56091 m .62475 .55608 L .63459 .544 L p .603 0 0 r
F P
0 g
s
.63459 .544 m .60168 .54668 L .59653 .56091 L p .603 0 0 r
F P
0 g
s
.3465 .44833 m .38295 .44129 L .3824 .44926 L p 0 0 .918 r
F P
0 g
s
.3824 .44926 m .34413 .45722 L .3465 .44833 L p 0 0 .918 r
F P
0 g
s
.4225 .4517 m .3824 .44926 L .38295 .44129 L p 0 0 .936 r
F P
0 g
s
.65747 .53396 m .6936 .55657 L .69683 .56867 L p .84 0 0 r
F P
0 g
s
.71845 .59065 m .69683 .56867 L .6936 .55657 L p .198 0 0 r
F P
0 g
s
.69683 .56867 m .71845 .59065 L .71427 .60584 L p .363 0 0 r
F P
0 g
s
.38608 .50502 m .35985 .51106 L .35317 .49962 L p 0 0 .536 r
F P
0 g
s
.35985 .51106 m .38608 .50502 L .38768 .51304 L p 0 0 .308 r
F P
0 g
s
.38608 .50502 m .41267 .50674 L .40803 .51431 L p 0 0 .344 r
F P
0 g
s
.40803 .51431 m .38768 .51304 L .38608 .50502 L p 0 0 .344 r
F P
0 g
s
.64618 .53618 m .61019 .53365 L .60168 .54668 L p .644 0 0 r
F P
0 g
s
.61019 .53365 m .64618 .53618 L .65747 .53396 L p .688 0 0 r
F P
0 g
s
.38295 .44129 m .42069 .4435 L .4225 .4517 L p 0 0 .936 r
F P
0 g
s
.45795 .4638 m .4225 .4517 L .42069 .4435 L p 0 0 .924 r
F P
0 g
s
.4225 .4517 m .45795 .4638 L .45727 .47703 L p 0 0 .816 r
F P
0 g
s
.42908 .54627 m .43147 .54434 L .4308 .5454 L p .917 0 0 r
F P
0 g
s
.47093 .5271 m .43147 .54434 L .44156 .53961 L p .346 0 0 r
F P
0 g
s
.43147 .54434 m .40444 .56617 L .41132 .5552 L p .87 0 0 r
F P
0 g
s
.41132 .5552 m .4389 .54098 L p .43147 .54434 L .87 0 0 r
F P
0 g
s
.4389 .54098 m .44156 .53961 L .44048 .54012 L p .87 0 0 r
F P
0 g
s
.44048 .54012 m .43147 .54434 L p .4389 .54098 L .87 0 0 r
F P
0 g
s
.40444 .56617 m .42908 .54627 L p .42528 .55425 L .917 0 0 r
F P
0 g
s
.4308 .5454 m .42528 .55425 L p .42908 .54627 L .917 0 0 r
F P
0 g
s
.45385 .54095 m .42471 .54875 L p .43467 .55005 L .417 0 0 r
F P
0 g
s
.44156 .53961 m .41388 .55388 L p .422 .54948 L .362 0 0 r
F P
0 g
s
.42471 .54875 m .422 .54948 L .42328 .54954 L p .417 0 0 r
F P
0 g
s
.42328 .54954 m .43467 .55005 L p .42471 .54875 L .417 0 0 r
F P
0 g
s
.41388 .55388 m .41132 .5552 L .41239 .55463 L p .362 0 0 r
F P
0 g
s
.41239 .55463 m .422 .54948 L p .41388 .55388 L .362 0 0 r
F P
0 g
s
.42528 .55425 m .40248 .58047 L .40444 .56617 L p .917 0 0 r
F P
0 g
s
.35317 .49962 m .32674 .51519 L .31832 .50237 L p 0 0 .593 r
F P
0 g
s
.32674 .51519 m .35317 .49962 L .35985 .51106 L p 0 0 .465 r
F P
0 g
s
.41267 .50674 m .43721 .516 L .42825 .52143 L p 0 0 .341 r
F P
0 g
s
.42825 .52143 m .40803 .51431 L .41267 .50674 L p 0 0 .341 r
F P
0 g
s
.47801 .51276 m .45303 .49176 L .45727 .47703 L p 0 0 .685 r
F P
0 g
s
.45303 .49176 m .47801 .51276 L .46801 .52437 L p 0 0 .598 r
F P
0 g
s
.46801 .52437 m .44596 .50551 L .45303 .49176 L p 0 0 .598 r
F P
0 g
s
.31092 .47385 m .34413 .45722 L .3446 .47026 L p 0 0 .764 r
F P
0 g
s
.3446 .47026 m .31273 .48778 L .31092 .47385 L p 0 0 .764 r
F P
0 g
s
.6936 .55657 m .65747 .53396 L .64618 .53618 L p .773 0 0 r
F P
0 g
s
.60168 .54668 m .63459 .544 L .64618 .53618 L p .644 0 0 r
F P
0 g
s
.61998 .59495 m .63887 .58875 L .64301 .57389 L p .521 0 0 r
F P
0 g
s
.64301 .57389 m .61676 .58403 L .61998 .59495 L p .521 0 0 r
F P
0 g
s
.5027 .58398 m .48266 .57454 L .48385 .58898 L p .936 0 0 r
F P
0 g
s
.48266 .57454 m .5027 .58398 L .50447 .57164 L p .887 0 0 r
F P
0 g
s
.45979 .60958 m .44972 .6204 L .43772 .62277 L p .181 0 0 r
F P
0 g
s
.44972 .6204 m .45979 .60958 L .47127 .60813 L p .666 0 0 r
F P
0 g
s
.48501 .49868 m .45727 .47703 L .45795 .4638 L p 0 0 .773 r
F P
0 g
s
.45727 .47703 m .48501 .49868 L .47801 .51276 L p 0 0 .685 r
F P
0 g
s
.66701 .61588 m .68246 .62969 L .67146 .62574 L p .451 0 0 r
F P
0 g
s
.67146 .62574 m .65527 .6172 L .66701 .61588 L p .451 0 0 r
F P
0 g
s
.44156 .53961 m .48283 .5329 L .47093 .5271 L p .346 0 0 r
F P
0 g
s
.44596 .50551 m .46801 .52437 L .45668 .53156 L p 0 0 .483 r
F P
0 g
s
.45668 .53156 m .43721 .516 L .44596 .50551 L p 0 0 .483 r
F P
0 g
s
.64618 .53618 m .68591 .54879 L .6936 .55657 L p .773 0 0 r
F P
0 g
s
.7177 .57574 m .6936 .55657 L .68591 .54879 L p 0 0 0 r
F P
0 g
s
.6936 .55657 m .7177 .57574 L .71845 .59065 L p .198 0 0 r
F P
0 g
s
.35985 .51106 m .33661 .52409 L .32674 .51519 L p 0 0 .465 r
F P
0 g
s
.33661 .52409 m .35985 .51106 L .36672 .51758 L p 0 0 .181 r
F P
0 g
s
.38768 .51304 m .36672 .51758 L .35985 .51106 L p 0 0 .308 r
F P
0 g
s
.31323 .46293 m .3465 .44833 L .34413 .45722 L p 0 0 .861 r
F P
0 g
s
.34413 .45722 m .31092 .47385 L .31323 .46293 L p 0 0 .861 r
F P
0 g
s
.50447 .57164 m .47641 .56008 L .48266 .57454 L p .887 0 0 r
F P
0 g
s
.47641 .56008 m .50447 .57164 L .50125 .55733 L p .821 0 0 r
F P
0 g
s
.42069 .4435 m .45492 .45429 L .45795 .4638 L p 0 0 .924 r
F P
0 g
s
.48785 .48449 m .45795 .4638 L .45492 .45429 L p 0 0 .869 r
F P
0 g
s
.45795 .4638 m .48785 .48449 L .48501 .49868 L p 0 0 .773 r
F P
0 g
s
.68591 .54879 m .64618 .53618 L .63459 .544 L p .707 0 0 r
F P
0 g
s
.45385 .54095 m .49358 .5436 L .48283 .5329 L p .448 0 0 r
F P
0 g
s
.48283 .5329 m .4429 .53939 L p .45385 .54095 L .448 0 0 r
F P
0 g
s
.4662 .54819 m .50125 .55733 L .49358 .5436 L p .714 0 0 r
F P
0 g
s
.49358 .5436 m .45385 .54095 L .4662 .54819 L p .714 0 0 r
F P
0 g
s
.46801 .52437 m .47801 .51276 L p .47757 .54641 L 0 0 .451 r
F P
0 g
s
.47755 .54764 m .477 .5463 L p .47757 .54641 L 0 0 .451 r
F P
0 g
s
.477 .5463 m .46801 .52437 L p .47757 .54641 L 0 0 .451 r
F P
0 g
s
.4429 .53939 m .44156 .53961 L .4468 .54018 L p .448 0 0 r
F P
0 g
s
.4468 .54018 m .45385 .54095 L p .4429 .53939 L .448 0 0 r
F P
0 g
s
.43721 .516 m .45668 .53156 L .44593 .53305 L p 0 0 .265 r
F P
0 g
s
.44593 .53305 m .42825 .52143 L .43721 .516 L p 0 0 .265 r
F P
0 g
s
.61676 .58403 m .64301 .57389 L .65151 .56047 L p .579 0 0 r
F P
0 g
s
.65151 .56047 m .61844 .57026 L .61676 .58403 L p .579 0 0 r
F P
0 g
s
.35134 .44517 m .38407 .43944 L .38295 .44129 L p 0 0 .934 r
F P
0 g
s
.38295 .44129 m .3465 .44833 L .35134 .44517 L p 0 0 .934 r
F P
0 g
s
.42069 .4435 m .38295 .44129 L .38407 .43944 L p 0 0 .933 r
F P
0 g
s
.36672 .51758 m .38768 .51304 L .38905 .51494 L p 0 0 0 r
F P
0 g
s
.38905 .51494 m .38768 .51304 L p .3966 .51455 L 0 0 0 r
F P
0 g
s
.40395 .5158 m .38905 .51494 L p .3966 .51455 L 0 0 0 r
F P
0 g
s
.38905 .51494 m .38906 .51488 L p .38859 .51497 L 0 0 .979 r
F P
0 g
s
.3726 .518 m .36672 .51758 L p .38 .51601 L 0 0 0 r
F P
0 g
s
.38905 .51494 m .3726 .518 L p .38 .51601 L 0 0 0 r
F P
0 g
s
.34627 .52752 m .36672 .51758 L .3726 .518 L p 0 0 0 r
F P
0 g
s
.3726 .518 m .37646 .51215 L p .36021 .52134 L 0 0 .923 r
F P
0 g
s
.35405 .52482 m .34627 .52752 L p .36611 .52035 L 0 0 0 r
F P
0 g
s
.3726 .518 m .35405 .52482 L p .36611 .52035 L 0 0 0 r
F P
0 g
s
.36672 .51758 m .34627 .52752 L .33661 .52409 L p 0 0 .181 r
F P
0 g
s
.32106 .54077 m .33661 .52409 L .34627 .52752 L p 0 0 0 r
F P
0 g
s
.34627 .52752 m .33046 .53767 L .32106 .54077 L p 0 0 0 r
F P
0 g
s
.33046 .53767 m .34627 .52752 L .35405 .52482 L p 0 0 .882 r
F P
0 g
s
.35405 .52482 m .3363 .52891 L .33409 .53223 L p 0 0 .882 r
F P
0 g
s
.33409 .53223 m .33046 .53767 L p .35405 .52482 L 0 0 .882 r
F P
0 g
s
.35405 .52482 m .3726 .518 L p .36021 .52134 L 0 0 .923 r
F P
0 g
s
.3726 .518 m .38905 .51494 L p .38859 .51497 L 0 0 .979 r
F P
0 g
s
.38768 .51304 m .40561 .51416 L p .3966 .51455 L 0 0 0 r
F P
0 g
s
.38906 .51488 m .38994 .51029 L p .37862 .51533 L 0 0 .979 r
F P
0 g
s
.37862 .51533 m .3726 .518 L .38906 .51488 L 0 0 .979 r
F
.40803 .51431 m .40395 .5158 L p .3966 .51455 L 0 0 0 r
F P
0 g
s
.40561 .51416 m .40803 .51431 L p .3966 .51455 L 0 0 0 r
F P
0 g
s
.65527 .6172 m .67146 .62574 L .66353 .61597 L p 0 0 0 r
F P
0 g
s
.66353 .61597 m .64564 .61245 L .65527 .6172 L p 0 0 0 r
F P
0 g
s
.38407 .43944 m .41727 .44128 L .42069 .4435 L p 0 0 .933 r
F P
0 g
s
.45492 .45429 m .42069 .4435 L .41727 .44128 L p 0 0 .817 r
F P
0 g
s
.47127 .60813 m .4589 .61204 L .44972 .6204 L p .666 0 0 r
F P
0 g
s
.4589 .61204 m .47127 .60813 L .4798 .60086 L p .922 0 0 r
F P
0 g
s
.40803 .51431 m .42825 .52143 L .42217 .52093 L p 0 0 0 r
F P
0 g
s
.41876 .51985 m .40803 .51431 L .42217 .52093 L 0 0 0 r
F
.42061 .5208 m .40395 .5158 L .40803 .51431 L p F P
0 g
s
.41563 .51411 m .40115 .51083 L .40351 .515 L p 0 0 .966 r
F P
0 g
s
.40351 .515 m .40395 .5158 L p .40537 .51559 L 0 0 .966 r
F P
0 g
s
.41563 .51411 m .40351 .515 L .40537 .51559 L 0 0 .966 r
F
.42217 .52093 m .42061 .5208 L p .41876 .51985 L 0 0 0 r
F P
0 g
s
.50125 .55733 m .4662 .54819 L .47641 .56008 L p .821 0 0 r
F P
0 g
s
.28914 .51657 m .31273 .48778 L .31832 .50237 L p 0 0 .579 r
F P
0 g
s
.31832 .50237 m .29849 .52919 L .28914 .51657 L p 0 0 .579 r
F P
0 g
s
.29849 .52919 m .31832 .50237 L .32674 .51519 L p 0 0 .462 r
F P
0 g
s
.38342 .59232 m .40248 .58047 L .40576 .59567 L p .897 0 0 r
F P
0 g
s
.40576 .59567 m .38682 .60784 L .38342 .59232 L p .897 0 0 r
F P
0 g
s
.38682 .60784 m .40576 .59567 L .41371 .6092 L p .843 0 0 r
F P
0 g
s
.63459 .544 m .67505 .54675 L .68591 .54879 L p .707 0 0 r
F P
0 g
s
.68591 .54879 m .71213 .56363 L .7177 .57574 L p 0 0 0 r
F P
0 g
s
.71213 .56363 m .68591 .54879 L .67505 .54675 L p .665 0 0 r
F P
0 g
s
.28327 .50202 m .31092 .47385 L .31273 .48778 L p 0 0 .665 r
F P
0 g
s
.31273 .48778 m .28914 .51657 L .28327 .50202 L p 0 0 .665 r
F P
0 g
s
.67505 .54675 m .63459 .544 L .62475 .55608 L p .66 0 0 r
F P
0 g
s
.61844 .57026 m .65151 .56047 L .66287 .55086 L p .62 0 0 r
F P
0 g
s
.66287 .55086 m .62475 .55608 L .61844 .57026 L p .62 0 0 r
F P
0 g
s
.32674 .51519 m .30976 .53774 L .29849 .52919 L p 0 0 .462 r
F P
0 g
s
.30976 .53774 m .32674 .51519 L .33661 .52409 L p 0 0 .233 r
F P
0 g
s
.3853 .57665 m .40444 .56617 L .40248 .58047 L p .922 0 0 r
F P
0 g
s
.40248 .58047 m .38342 .59232 L .3853 .57665 L p .922 0 0 r
F P
0 g
s
.41371 .6092 m .39493 .6206 L .38682 .60784 L p .843 0 0 r
F P
0 g
s
.39493 .6206 m .41371 .6092 L .42499 .61879 L p .724 0 0 r
F P
0 g
s
.422 .54948 m .45385 .54095 L .44156 .53961 L p .362 0 0 r
F P
0 g
s
.49006 .54239 m .47801 .51276 L .48501 .49868 L p 0 0 .567 r
F P
0 g
s
.47801 .51276 m .49006 .54239 L .47755 .54764 L p 0 0 .451 r
F P
0 g
s
.7355 .61404 m .71427 .60584 L .71845 .59065 L p .439 0 0 r
F P
0 g
s
.71427 .60584 m .7355 .61404 L .728 .62636 L p .452 0 0 r
F P
0 g
s
.728 .62636 m .70586 .61876 L .71427 .60584 L p .452 0 0 r
F P
0 g
s
.7385 .59867 m .71845 .59065 L .7177 .57574 L p .41 0 0 r
F P
0 g
s
.71845 .59065 m .7385 .59867 L .7355 .61404 L p .439 0 0 r
F P
0 g
s
.62475 .55608 m .66287 .55086 L .67505 .54675 L p .66 0 0 r
F P
0 g
s
.50144 .5321 m .48501 .49868 L .48785 .48449 L p 0 0 .658 r
F P
0 g
s
.48501 .49868 m .50144 .5321 L .49006 .54239 L p 0 0 .567 r
F P
0 g
s
.31931 .45694 m .35134 .44517 L .3465 .44833 L p 0 0 .914 r
F P
0 g
s
.3465 .44833 m .31323 .46293 L .31931 .45694 L p 0 0 .914 r
F P
0 g
s
.41727 .44128 m .44866 .45018 L .45492 .45429 L p 0 0 .817 r
F P
0 g
s
.48602 .47257 m .45492 .45429 L .44866 .45018 L p 0 0 .66 r
F P
0 g
s
.45492 .45429 m .48602 .47257 L .48785 .48449 L p 0 0 .869 r
F P
0 g
s
.4798 .60086 m .46363 .59908 L .4589 .61204 L p .922 0 0 r
F P
0 g
s
.46363 .59908 m .4798 .60086 L .48385 .58898 L p .919 0 0 r
F P
0 g
s
.64564 .61245 m .66353 .61597 L .66012 .60203 L p .337 0 0 r
F P
0 g
s
.66012 .60203 m .63981 .6024 L .64564 .61245 L p .337 0 0 r
F P
0 g
s
.42825 .52143 m .44593 .53305 L .4376 .52853 L p 0 0 0 r
F P
0 g
s
.4376 .52853 m .42061 .5208 L .42825 .52143 L p 0 0 0 r
F P
0 g
s
.70586 .61876 m .728 .62636 L .71722 .6335 L p .441 0 0 r
F P
0 g
s
.71722 .6335 m .69463 .62719 L .70586 .61876 L p .441 0 0 r
F P
0 g
s
.46801 .52437 m .47755 .54764 L .46604 .54691 L p 0 0 .254 r
F P
0 g
s
.46604 .54691 m .45668 .53156 L .46801 .52437 L p 0 0 .254 r
F P
0 g
s
.7365 .58287 m .7177 .57574 L .71213 .56363 L p .343 0 0 r
F P
0 g
s
.7177 .57574 m .7365 .58287 L .7385 .59867 L p .41 0 0 r
F P
0 g
s
.39214 .56349 m .41132 .5552 L .40444 .56617 L p .904 0 0 r
F P
0 g
s
.40444 .56617 m .3853 .57665 L .39214 .56349 L p .904 0 0 r
F P
0 g
s
.28188 .48796 m .31323 .46293 L .31092 .47385 L p 0 0 .75 r
F P
0 g
s
.31092 .47385 m .28327 .50202 L .28188 .48796 L p 0 0 .75 r
F P
0 g
s
.42499 .61879 m .40641 .6284 L .39493 .6206 L p .724 0 0 r
F P
0 g
s
.40641 .6284 m .42499 .61879 L .43772 .62277 L p .384 0 0 r
F P
0 g
s
.50979 .51853 m .48785 .48449 L .48602 .47257 L p 0 0 .727 r
F P
0 g
s
.48785 .48449 m .50979 .51853 L .50144 .5321 L p 0 0 .658 r
F P
0 g
s
.35784 .44837 m .38558 .44415 L .38407 .43944 L p 0 0 .01 r
F P
0 g
s
.38407 .43944 m .35134 .44517 L .35784 .44837 L p 0 0 .01 r
F P
0 g
s
.41727 .44128 m .38407 .43944 L .38558 .44415 L p 0 0 .055 r
F P
0 g
s
.33661 .52409 m .32106 .54077 L .30976 .53774 L p 0 0 .233 r
F P
0 g
s
.38905 .51494 m .40395 .5158 L .40115 .51083 L p 0 0 .997 r
F P
0 g
s
.40115 .51083 m .38994 .51029 L .38905 .51494 L p 0 0 .997 r
F P
0 g
s
.70263 .55641 m .71213 .56363 L p .68975 .55344 L .665 0 0 r
F P
0 g
s
.67505 .54675 m .70263 .55641 L p .68975 .55344 L .665 0 0 r
F P
0 g
s
.72984 .56928 m .71213 .56363 L .70263 .55641 L p .146 0 0 r
F P
0 g
s
.70263 .55641 m .71963 .56021 L .72984 .56928 L p .146 0 0 r
F P
0 g
s
.70263 .55641 m .69081 .55538 L p .71782 .55991 L .301 0 0 r
F P
0 g
s
.71963 .56021 m .70263 .55641 L p .71782 .55991 L .301 0 0 r
F P
0 g
s
.70263 .55641 m .67505 .54675 L .66287 .55086 L p .657 0 0 r
F P
0 g
s
.38558 .44415 m .41282 .44555 L .41727 .44128 L p 0 0 .055 r
F P
0 g
s
.44866 .45018 m .41727 .44128 L .41282 .44555 L p 0 0 .106 r
F P
0 g
s
.40395 .5158 m .42061 .5208 L .41563 .51411 L p 0 0 .966 r
F P
0 g
s
.43467 .55005 m .4662 .54819 L .45385 .54095 L p .417 0 0 r
F P
0 g
s
.69463 .62719 m .71722 .6335 L .705 .63419 L p .335 0 0 r
F P
0 g
s
.705 .63419 m .68246 .62969 L .69463 .62719 L p .335 0 0 r
F P
0 g
s
.48385 .58898 m .46303 .58377 L .46363 .59908 L p .919 0 0 r
F P
0 g
s
.46303 .58377 m .48385 .58898 L .48266 .57454 L p .877 0 0 r
F P
0 g
s
.71213 .56363 m .72984 .56928 L .7365 .58287 L p .343 0 0 r
F P
0 g
s
.63981 .6024 m .66012 .60203 L .66186 .58637 L p .459 0 0 r
F P
0 g
s
.66186 .58637 m .63887 .58875 L .63981 .6024 L p .459 0 0 r
F P
0 g
s
.40282 .55508 m .422 .54948 L .41132 .5552 L p .706 0 0 r
F P
0 g
s
.41132 .5552 m .39214 .56349 L .40282 .55508 L p .706 0 0 r
F P
0 g
s
.45668 .53156 m .46604 .54691 L .45751 .54025 L p 0 0 0 r
F P
0 g
s
.45751 .54025 m .44593 .53305 L .45668 .53156 L p 0 0 0 r
F P
0 g
s
.43772 .62277 m .41932 .62985 L .40641 .6284 L p .384 0 0 r
F P
0 g
s
.41932 .62985 m .43772 .62277 L .44972 .6204 L p .416 0 0 r
F P
0 g
s
.38994 .51029 m .37646 .51215 L .3726 .518 L p 0 0 .979 r
F P
0 g
s
.32817 .45697 m .35784 .44837 L .35134 .44517 L p 0 0 0 r
F P
0 g
s
.35134 .44517 m .31931 .45694 L .32817 .45697 L p 0 0 0 r
F P
0 g
s
.28524 .47679 m .31931 .45694 L .31323 .46293 L p 0 0 .848 r
F P
0 g
s
.31323 .46293 m .28188 .48796 L .28524 .47679 L p 0 0 .848 r
F P
0 g
s
.44713 .55688 m .47641 .56008 L .4662 .54819 L p .699 0 0 r
F P
0 g
s
.4662 .54819 m .43467 .55005 L .44713 .55688 L p .699 0 0 r
F P
0 g
s
.66287 .55086 m .69081 .55538 L .70263 .55641 L p .657 0 0 r
F P
0 g
s
.69081 .55538 m .66287 .55086 L .65151 .56047 L p .61 0 0 r
F P
0 g
s
.44866 .45018 m .47977 .46497 L .48602 .47257 L p 0 0 .66 r
F P
0 g
s
.51372 .50395 m .48602 .47257 L .47977 .46497 L p 0 0 .519 r
F P
0 g
s
.48602 .47257 m .51372 .50395 L .50979 .51853 L p 0 0 .727 r
F P
0 g
s
.41282 .44555 m .44019 .45228 L .44866 .45018 L p 0 0 .106 r
F P
0 g
s
.47977 .46497 m .44866 .45018 L .44019 .45228 L p 0 0 .128 r
F P
0 g
s
.48266 .57454 m .45718 .56882 L .46303 .58377 L p .877 0 0 r
F P
0 g
s
.45718 .56882 m .48266 .57454 L .47641 .56008 L p .813 0 0 r
F P
0 g
s
.42061 .5208 m .4376 .52853 L .4332 .51868 L p 0 0 .93 r
F P
0 g
s
.4332 .51868 m .41563 .51411 L .42061 .5208 L p 0 0 .93 r
F P
0 g
s
.47641 .56008 m .44713 .55688 L .45718 .56882 L p .813 0 0 r
F P
0 g
s
.63887 .58875 m .66186 .58637 L .66845 .57176 L p .521 0 0 r
F P
0 g
s
.66845 .57176 m .64301 .57389 L .63887 .58875 L p .521 0 0 r
F P
0 g
s
.68246 .62969 m .705 .63419 L .69345 .62822 L p .064 0 0 r
F P
0 g
s
.69345 .62822 m .67146 .62574 L .68246 .62969 L p .064 0 0 r
F P
0 g
s
.26831 .54092 m .28327 .50202 L .28914 .51657 L p 0 0 .51 r
F P
0 g
s
.28914 .51657 m .27962 .55084 L .26831 .54092 L p 0 0 .51 r
F P
0 g
s
.27962 .55084 m .28914 .51657 L .29849 .52919 L p 0 0 .359 r
F P
0 g
s
.67871 .5608 m .65151 .56047 L .64301 .57389 L p .566 0 0 r
F P
0 g
s
.65151 .56047 m .67871 .5608 L .69081 .55538 L p .61 0 0 r
F P
0 g
s
.3649 .45747 m .38723 .45468 L .38558 .44415 L p 0 0 .441 r
F P
0 g
s
.38558 .44415 m .35784 .44837 L .3649 .45747 L p 0 0 .441 r
F P
0 g
s
.41282 .44555 m .38558 .44415 L .38723 .45468 L p 0 0 .463 r
F P
0 g
s
.41554 .5529 m .43467 .55005 L .422 .54948 L p .039 0 0 r
F P
0 g
s
.41532 .55143 m .40282 .55508 L .41554 .5529 L p .039 0 0 r
F P
0 g
s
.4053 .55134 m .41554 .5529 L .40282 .55508 L p .441 0 0 r
F P
0 g
s
.41554 .5529 m .4053 .55134 L .41576 .55439 L p .332 0 0 r
F P
0 g
s
.41576 .55439 m .41745 .55489 L p .41554 .5529 L .332 0 0 r
F P
0 g
s
.422 .54948 m .41532 .55143 L p .41554 .5529 L .039 0 0 r
F P
0 g
s
.29849 .52919 m .2919 .55535 L .27962 .55084 L p 0 0 .359 r
F P
0 g
s
.2919 .55535 m .29849 .52919 L .30976 .53774 L p 0 0 .017 r
F P
0 g
s
.37646 .51215 m .38994 .51029 L .3902 .49979 L p 0 0 .867 r
F P
0 g
s
.38994 .51029 m .40115 .51083 L .40014 .50016 L p 0 0 .884 r
F P
0 g
s
.40014 .50016 m .3902 .49979 L .38994 .51029 L p 0 0 .884 r
F P
0 g
s
.38723 .45468 m .4081 .45565 L .41282 .44555 L p 0 0 .463 r
F P
0 g
s
.44019 .45228 m .41282 .44555 L .4081 .45565 L p 0 0 .456 r
F P
0 g
s
.44972 .6204 m .43143 .62465 L .41932 .62985 L p .416 0 0 r
F P
0 g
s
.43143 .62465 m .44972 .6204 L .4589 .61204 L p .844 0 0 r
F P
0 g
s
.25984 .52727 m .28188 .48796 L .28327 .50202 L p 0 0 .606 r
F P
0 g
s
.28327 .50202 m .26831 .54092 L .25984 .52727 L p 0 0 .606 r
F P
0 g
s
.64301 .57389 m .66845 .57176 L .67871 .5608 L p .566 0 0 r
F P
0 g
s
.40115 .51083 m .41563 .51411 L .41423 .50246 L p 0 0 .875 r
F P
0 g
s
.41423 .50246 m .40014 .50016 L .40115 .51083 L p 0 0 .875 r
F P
0 g
s
.37055 .57614 m .3853 .57665 L .38342 .59232 L p .913 0 0 r
F P
0 g
s
.38342 .59232 m .36608 .59145 L .37055 .57614 L p .913 0 0 r
F P
0 g
s
.36608 .59145 m .38342 .59232 L .38682 .60784 L p .916 0 0 r
F P
0 g
s
.75123 .57952 m .7365 .58287 L .72984 .56928 L p .484 0 0 r
F P
0 g
s
.7365 .58287 m .75123 .57952 L .75677 .59422 L p .452 0 0 r
F P
0 g
s
.75677 .59422 m .7385 .59867 L .7365 .58287 L p .452 0 0 r
F P
0 g
s
.7385 .59867 m .75677 .59422 L .75714 .60836 L p .402 0 0 r
F P
0 g
s
.75714 .60836 m .7355 .61404 L .7385 .59867 L p .402 0 0 r
F P
0 g
s
.51874 .5547 m .50979 .51853 L .51372 .50395 L p 0 0 .546 r
F P
0 g
s
.50979 .51853 m .51874 .5547 L .50603 .56186 L p 0 0 .442 r
F P
0 g
s
.50603 .56186 m .50144 .5321 L .50979 .51853 L p 0 0 .442 r
F P
0 g
s
.50144 .5321 m .50603 .56186 L .49324 .56299 L p 0 0 .276 r
F P
0 g
s
.49324 .56299 m .49006 .54239 L .50144 .5321 L p 0 0 .276 r
F P
0 g
s
.44593 .53305 m .45751 .54025 L .4535 .52874 L p 0 0 .914 r
F P
0 g
s
.4535 .52874 m .4376 .52853 L .44593 .53305 L p 0 0 .914 r
F P
0 g
s
.37969 .563 m .39214 .56349 L .3853 .57665 L p .877 0 0 r
F P
0 g
s
.3853 .57665 m .37055 .57614 L .37969 .563 L p .877 0 0 r
F P
0 g
s
.38682 .60784 m .36704 .60634 L .36608 .59145 L p .916 0 0 r
F P
0 g
s
.36704 .60634 m .38682 .60784 L .39493 .6206 L p .867 0 0 r
F P
0 g
s
.37646 .51215 m .35858 .51638 L .35405 .52482 L p 0 0 .923 r
F P
0 g
s
.35858 .51638 m .37646 .51215 L .37759 .50095 L p 0 0 .816 r
F P
0 g
s
.3902 .49979 m .37759 .50095 L .37646 .51215 L p 0 0 .867 r
F P
0 g
s
.31931 .45694 m .28524 .47679 L .29282 .47045 L p 0 0 .819 r
F P
0 g
s
.29282 .47045 m .32817 .45697 L .31931 .45694 L p 0 0 .819 r
F P
0 g
s
.74146 .56674 m .72984 .56928 L .71963 .56021 L p .502 0 0 r
F P
0 g
s
.72984 .56928 m .74146 .56674 L .75123 .57952 L p .484 0 0 r
F P
0 g
s
.33829 .46312 m .3649 .45747 L .35784 .44837 L p 0 0 .362 r
F P
0 g
s
.35784 .44837 m .32817 .45697 L .33829 .46312 L p 0 0 .362 r
F P
0 g
s
.7355 .61404 m .75714 .60836 L .75222 .61953 L p .304 0 0 r
F P
0 g
s
.75222 .61953 m .728 .62636 L .7355 .61404 L p .304 0 0 r
F P
0 g
s
.30976 .53774 m .30304 .55362 L .2919 .55535 L p 0 0 .017 r
F P
0 g
s
.30304 .55362 m .30976 .53774 L .32106 .54077 L p 0 0 .667 r
F P
0 g
s
.44019 .45228 m .47013 .46308 L .47977 .46497 L p 0 0 .128 r
F P
0 g
s
.47977 .46497 m .51254 .4908 L .51372 .50395 L p 0 0 .519 r
F P
0 g
s
.51254 .4908 m .47977 .46497 L .47013 .46308 L p 0 0 .127 r
F P
0 g
s
.51372 .50395 m .51254 .4908 L p .52291 .52311 L 0 0 .592 r
F P
0 g
s
.47013 .46308 m .50641 .48133 L .51254 .4908 L p 0 0 .127 r
F P
0 g
s
.51254 .4908 m .53572 .52812 L .52921 .54278 L p 0 0 .592 r
F P
0 g
s
.53572 .52812 m .51254 .4908 L .50641 .48133 L p 0 0 0 r
F P
0 g
s
.53714 .51317 m .53572 .52812 L p .51803 .49989 L 0 0 0 r
F P
0 g
s
.50641 .48133 m .51628 .49156 L p .51803 .49989 L 0 0 0 r
F P
0 g
s
.51628 .49156 m .53714 .51317 L p .51803 .49989 L 0 0 0 r
F P
0 g
s
.52921 .54278 m .52462 .53126 L p .52291 .52311 L 0 0 .592 r
F P
0 g
s
.52462 .53126 m .51372 .50395 L p .52291 .52311 L 0 0 .592 r
F P
0 g
s
.51372 .50395 m .52921 .54278 L .51874 .5547 L p 0 0 .546 r
F P
0 g
s
.67146 .62574 m .69345 .62822 L .68461 .61659 L p .334 0 0 r
F P
0 g
s
.68461 .61659 m .66353 .61597 L .67146 .62574 L p .334 0 0 r
F P
0 g
s
.49006 .54239 m .49324 .56299 L .48255 .55785 L p 0 0 0 r
F P
0 g
s
.48255 .55785 m .47755 .54764 L .49006 .54239 L p 0 0 0 r
F P
0 g
s
.25565 .5122 m .28524 .47679 L .28188 .48796 L p 0 0 .689 r
F P
0 g
s
.28188 .48796 m .25984 .52727 L .25565 .5122 L p 0 0 .689 r
F P
0 g
s
.4081 .45565 m .43097 .46031 L .44019 .45228 L p 0 0 .456 r
F P
0 g
s
.47013 .46308 m .44019 .45228 L .43097 .46031 L p 0 0 .433 r
F P
0 g
s
.39196 .55423 m .40282 .55508 L .39214 .56349 L p .774 0 0 r
F P
0 g
s
.39214 .56349 m .37969 .563 L .39196 .55423 L p .774 0 0 r
F P
0 g
s
.41563 .51411 m .4332 .51868 L .43355 .50517 L p 0 0 .844 r
F P
0 g
s
.43355 .50517 m .41423 .50246 L .41563 .51411 L p 0 0 .844 r
F P
0 g
s
.70758 .55726 m .71963 .56021 L p .70512 .55778 L .301 0 0 r
F P
0 g
s
.72913 .558 m .71963 .56021 L .70758 .55726 L p .466 0 0 r
F P
0 g
s
.70758 .55726 m .71628 .55482 L .72913 .558 L p .466 0 0 r
F P
0 g
s
.70758 .55726 m .69574 .56099 L p .70952 .55685 L 0 0 0 r
F P
0 g
s
.71628 .55482 m .70758 .55726 L p .70952 .55685 L 0 0 0 r
F P
0 g
s
.69081 .55538 m .70758 .55726 L p .70512 .55778 L .301 0 0 r
F P
0 g
s
.70758 .55726 m .69081 .55538 L .67871 .5608 L p .481 0 0 r
F P
0 g
s
.3363 .52891 m .35405 .52482 L .35858 .51638 L p 0 0 .799 r
F P
0 g
s
.42811 .55741 m .44713 .55688 L .43467 .55005 L p .598 0 0 r
F P
0 g
s
.43467 .55005 m .41554 .5529 L .42811 .55741 L p .598 0 0 r
F P
0 g
s
.4589 .61204 m .4406 .61363 L .43143 .62465 L p .844 0 0 r
F P
0 g
s
.4406 .61363 m .4589 .61204 L .46363 .59908 L p .914 0 0 r
F P
0 g
s
.39493 .6206 m .37332 .61828 L .36704 .60634 L p .867 0 0 r
F P
0 g
s
.37332 .61828 m .39493 .6206 L .40641 .6284 L p .635 0 0 r
F P
0 g
s
.37125 .47093 m .38871 .46925 L .38723 .45468 L p 0 0 .597 r
F P
0 g
s
.38723 .45468 m .3649 .45747 L .37125 .47093 L p 0 0 .597 r
F P
0 g
s
.4081 .45565 m .38723 .45468 L .38871 .46925 L p 0 0 .613 r
F P
0 g
s
.37759 .50095 m .3902 .49979 L .38977 .48525 L p 0 0 .773 r
F P
0 g
s
.3902 .49979 m .40014 .50016 L .40113 .48564 L p 0 0 .788 r
F P
0 g
s
.40113 .48564 m .38977 .48525 L .3902 .49979 L p 0 0 .788 r
F P
0 g
s
.38871 .46925 m .40395 .46986 L .4081 .45565 L p 0 0 .613 r
F P
0 g
s
.43097 .46031 m .4081 .45565 L .40395 .46986 L p 0 0 .599 r
F P
0 g
s
.40014 .50016 m .41423 .50246 L .41671 .48786 L p 0 0 .775 r
F P
0 g
s
.41671 .48786 m .40113 .48564 L .40014 .50016 L p 0 0 .775 r
F P
0 g
s
.71963 .56021 m .72913 .558 L .74146 .56674 L p .502 0 0 r
F P
0 g
s
.728 .62636 m .75222 .61953 L .74361 .62521 L p .046 0 0 r
F P
0 g
s
.74167 .6258 m .728 .62636 L .74361 .62521 L .046 0 0 r
F
.74279 .62576 m .71722 .6335 L .728 .62636 L p F P
0 g
s
.73044 .6259 m .705 .63419 L .71608 .63357 L p .394 0 0 r
F P
0 g
s
.71608 .63357 m .71722 .6335 L p .71825 .63291 L .394 0 0 r
F P
0 g
s
.73044 .6259 m .71608 .63357 L .71825 .63291 L .394 0 0 r
F
.74361 .62521 m .74279 .62576 L p .74167 .6258 L .046 0 0 r
F P
0 g
s
.4376 .52853 m .4535 .52874 L .45477 .51435 L p 0 0 .907 r
F P
0 g
s
.45477 .51435 m .4332 .51868 L .4376 .52853 L p 0 0 .907 r
F P
0 g
s
.40282 .55508 m .39196 .55423 L .4053 .55134 L p .441 0 0 r
F P
0 g
s
.66353 .61597 m .68461 .61659 L .68006 .60132 L p .419 0 0 r
F P
0 g
s
.68006 .60132 m .66012 .60203 L .66353 .61597 L p .419 0 0 r
F P
0 g
s
.43834 .56787 m .45718 .56882 L .44713 .55688 L p .788 0 0 r
F P
0 g
s
.44713 .55688 m .42811 .55741 L .43834 .56787 L p .788 0 0 r
F P
0 g
s
.50957 .41688 m .5479 .40853 L .54846 .42425 L p 0 .652 0 r
F P
0 g
s
.54846 .42425 m .51265 .43237 L .50957 .41688 L p 0 .652 0 r
F P
0 g
s
.58692 .42668 m .54846 .42425 L .5479 .40853 L p 0 .602 0 r
F P
0 g
s
.51265 .43237 m .54846 .42425 L .54998 .43985 L p 0 .568 0 r
F P
0 g
s
.54846 .42425 m .58692 .42668 L .5836 .44199 L p 0 .537 0 r
F P
0 g
s
.5836 .44199 m .54998 .43985 L .54846 .42425 L p 0 .537 0 r
F P
0 g
s
.47586 .54723 m .46604 .54691 L .47755 .54764 L p 0 0 .72 r
F P
0 g
s
.47755 .54764 m .48255 .55785 L .47586 .54723 L p 0 0 .72 r
F P
0 g
s
.35902 .5036 m .37759 .50095 L .37575 .48635 L p 0 0 .727 r
F P
0 g
s
.38977 .48525 m .37575 .48635 L .37759 .50095 L p 0 0 .773 r
F P
0 g
s
.37759 .50095 m .35902 .5036 L .35858 .51638 L p 0 0 .816 r
F P
0 g
s
.46363 .59908 m .44521 .5987 L .4406 .61363 L p .914 0 0 r
F P
0 g
s
.44521 .5987 m .46363 .59908 L .46303 .58377 L p .906 0 0 r
F P
0 g
s
.32106 .54077 m .31112 .54587 L .30304 .55362 L p 0 0 .667 r
F P
0 g
s
.31112 .54587 m .32106 .54077 L .33046 .53767 L p 0 0 .783 r
F P
0 g
s
.34789 .47435 m .37125 .47093 L .3649 .45747 L p 0 0 .542 r
F P
0 g
s
.3649 .45747 m .33829 .46312 L .34789 .47435 L p 0 0 .542 r
F P
0 g
s
.69574 .56099 m .67871 .5608 L .66845 .57176 L p .493 0 0 r
F P
0 g
s
.69574 .56099 m .68619 .57083 L p .69843 .56238 L .222 0 0 r
F P
0 g
s
.69574 .56099 m .70758 .55726 L p .69216 .55915 L .481 0 0 r
F P
0 g
s
.7051 .55776 m .71628 .55482 L p .70468 .5583 L 0 0 0 r
F P
0 g
s
.69574 .56099 m .7051 .55776 L p .70468 .5583 L 0 0 0 r
F P
0 g
s
.7051 .55776 m .69574 .56099 L p .69843 .56238 L .222 0 0 r
F P
0 g
s
.68619 .57083 m .69753 .56642 L .7051 .55776 L p .222 0 0 r
F P
0 g
s
.7051 .55776 m .69753 .56642 L p .70548 .55726 L .544 0 0 r
F P
0 g
s
.71104 .55087 m .7051 .55776 L p .70548 .55726 L .544 0 0 r
F P
0 g
s
.67871 .5608 m .69574 .56099 L p .69216 .55915 L .481 0 0 r
F P
0 g
s
.5479 .40853 m .58911 .41108 L .58692 .42668 L p 0 .602 0 r
F P
0 g
s
.62129 .43903 m .58692 .42668 L .58911 .41108 L p 0 .528 0 r
F P
0 g
s
.58692 .42668 m .62129 .43903 L .61422 .45314 L p 0 .478 0 r
F P
0 g
s
.61422 .45314 m .5836 .44199 L .58692 .42668 L p 0 .478 0 r
F P
0 g
s
.37575 .48635 m .38977 .48525 L .38871 .46925 L p 0 0 .691 r
F P
0 g
s
.38871 .46925 m .37125 .47093 L .37575 .48635 L p 0 0 .691 r
F P
0 g
s
.40395 .46986 m .38871 .46925 L .38977 .48525 L p 0 0 .706 r
F P
0 g
s
.40641 .6284 m .3839 .62518 L .37332 .61828 L p .635 0 0 r
F P
0 g
s
.3839 .62518 m .40641 .6284 L .41932 .62985 L p .088 0 0 r
F P
0 g
s
.25645 .49824 m .29282 .47045 L .28524 .47679 L p 0 0 .778 r
F P
0 g
s
.28524 .47679 m .25565 .5122 L .25645 .49824 L p 0 0 .778 r
F P
0 g
s
.38977 .48525 m .40113 .48564 L .40395 .46986 L p 0 0 .706 r
F P
0 g
s
.42265 .47292 m .40395 .46986 L .40113 .48564 L p 0 0 .691 r
F P
0 g
s
.40395 .46986 m .42265 .47292 L .43097 .46031 L p 0 0 .599 r
F P
0 g
s
.30338 .4701 m .33829 .46312 L .32817 .45697 L p 0 0 0 r
F P
0 g
s
.32817 .45697 m .29282 .47045 L .30338 .4701 L p 0 0 0 r
F P
0 g
s
.40113 .48564 m .41671 .48786 L .42265 .47292 L p 0 0 .691 r
F P
0 g
s
.45872 .46729 m .43097 .46031 L .42265 .47292 L p 0 0 .569 r
F P
0 g
s
.43097 .46031 m .45872 .46729 L .47013 .46308 L p 0 0 .433 r
F P
0 g
s
.74934 .55721 m .74146 .56674 L .72913 .558 L p .641 0 0 r
F P
0 g
s
.74146 .56674 m .74934 .55721 L .7623 .5667 L p .565 0 0 r
F P
0 g
s
.7623 .5667 m .75123 .57952 L .74146 .56674 L p .565 0 0 r
F P
0 g
s
.41423 .50246 m .43355 .50517 L .43863 .49038 L p 0 0 .746 r
F P
0 g
s
.43863 .49038 m .41671 .48786 L .41423 .50246 L p 0 0 .746 r
F P
0 g
s
.50641 .48133 m .47013 .46308 L .45872 .46729 L p 0 0 .413 r
F P
0 g
s
.75123 .57952 m .7623 .5667 L .7716 .5777 L p .473 0 0 r
F P
0 g
s
.7716 .5777 m .75677 .59422 L .75123 .57952 L p .473 0 0 r
F P
0 g
s
.50968 .40326 m .54841 .39529 L .5479 .40853 L p 0 .735 0 r
F P
0 g
s
.5479 .40853 m .50957 .41688 L .50968 .40326 L p 0 .735 0 r
F P
0 g
s
.58911 .41108 m .5479 .40853 L .54841 .39529 L p 0 .665 0 r
F P
0 g
s
.54998 .43985 m .51838 .44715 L .51265 .43237 L p 0 .568 0 r
F P
0 g
s
.51838 .44715 m .54998 .43985 L .55223 .45272 L p 0 .455 0 r
F P
0 g
s
.54998 .43985 m .5836 .44199 L .5797 .45447 L p 0 .449 0 r
F P
0 g
s
.5797 .45447 m .55223 .45272 L .54998 .43985 L p 0 .449 0 r
F P
0 g
s
.4444 .58248 m .46303 .58377 L .45718 .56882 L p .867 0 0 r
F P
0 g
s
.45718 .56882 m .43834 .56787 L .4444 .58248 L p .867 0 0 r
F P
0 g
s
.46303 .58377 m .4444 .58248 L .44521 .5987 L p .906 0 0 r
F P
0 g
s
.54841 .39529 m .58981 .39777 L .58911 .41108 L p 0 .665 0 r
F P
0 g
s
.62585 .4238 m .58911 .41108 L .58981 .39777 L p 0 .575 0 r
F P
0 g
s
.58911 .41108 m .62585 .4238 L .62129 .43903 L p 0 .528 0 r
F P
0 g
s
.5836 .44199 m .61422 .45314 L .6058 .46376 L p 0 .41 0 r
F P
0 g
s
.6058 .46376 m .5797 .45447 L .5836 .44199 L p 0 .41 0 r
F P
0 g
s
.35858 .51638 m .33751 .51595 L .3363 .52891 L p 0 0 .799 r
F P
0 g
s
.33751 .51595 m .35858 .51638 L .35902 .5036 L p 0 0 .707 r
F P
0 g
s
.66012 .60203 m .68006 .60132 L .68063 .58511 L p .456 0 0 r
F P
0 g
s
.68063 .58511 m .66186 .58637 L .66012 .60203 L p .456 0 0 r
F P
0 g
s
.37184 .55569 m .37969 .563 L .37055 .57614 L p .835 0 0 r
F P
0 g
s
.37055 .57614 m .35965 .56613 L .37184 .55569 L p .835 0 0 r
F P
0 g
s
.35965 .56613 m .37055 .57614 L .36608 .59145 L p .9 0 0 r
F P
0 g
s
.71722 .6335 m .74279 .62576 L .73044 .6259 L p .394 0 0 r
F P
0 g
s
.37575 .48635 m .35523 .4887 L .35902 .5036 L p 0 0 .727 r
F P
0 g
s
.35523 .4887 m .37575 .48635 L .37125 .47093 L p 0 0 .644 r
F P
0 g
s
.68619 .57083 m .66845 .57176 L .66186 .58637 L p .479 0 0 r
F P
0 g
s
.66845 .57176 m .68619 .57083 L .69574 .56099 L p .493 0 0 r
F P
0 g
s
.37125 .47093 m .34789 .47435 L .35523 .4887 L p 0 0 .644 r
F P
0 g
s
.73489 .5508 m .72913 .558 L .71628 .55482 L p .72 0 0 r
F P
0 g
s
.72913 .558 m .73489 .5508 L .74934 .55721 L p .641 0 0 r
F P
0 g
s
.38616 .54876 m .39196 .55423 L .37969 .563 L p .747 0 0 r
F P
0 g
s
.37969 .563 m .37184 .55569 L .38616 .54876 L p .747 0 0 r
F P
0 g
s
.66186 .58637 m .68063 .58511 L .68619 .57083 L p .479 0 0 r
F P
0 g
s
.75677 .59422 m .7716 .5777 L .77564 .58836 L p .321 0 0 r
F P
0 g
s
.77564 .58836 m .75714 .60836 L .75677 .59422 L p .321 0 0 r
F P
0 g
s
.41745 .55489 m .42811 .55741 L .41554 .5529 L p .332 0 0 r
F P
0 g
s
.44755 .47696 m .42265 .47292 L .41671 .48786 L p 0 0 .661 r
F P
0 g
s
.42265 .47292 m .44755 .47696 L .45872 .46729 L p 0 0 .569 r
F P
0 g
s
.41671 .48786 m .43863 .49038 L .44755 .47696 L p 0 0 .661 r
F P
0 g
s
.36608 .59145 m .35165 .57835 L .35965 .56613 L p .9 0 0 r
F P
0 g
s
.35165 .57835 m .36608 .59145 L .36704 .60634 L p .946 0 0 r
F P
0 g
s
.235 .54063 m .25565 .5122 L .25984 .52727 L p 0 0 .554 r
F P
0 g
s
.25984 .52727 m .24307 .55457 L .235 .54063 L p 0 0 .554 r
F P
0 g
s
.24307 .55457 m .25984 .52727 L .26831 .54092 L p 0 0 .428 r
F P
0 g
s
.41932 .62985 m .39699 .62578 L .3839 .62518 L p .088 0 0 r
F P
0 g
s
.39699 .62578 m .41932 .62985 L .43143 .62465 L p .655 0 0 r
F P
0 g
s
.45872 .46729 m .49631 .47724 L .50641 .48133 L p 0 0 .413 r
F P
0 g
s
.53714 .51317 m .50641 .48133 L .49631 .47724 L p 0 0 .178 r
F P
0 g
s
.26831 .54092 m .25432 .56435 L .24307 .55457 L p 0 0 .428 r
F P
0 g
s
.25432 .56435 m .26831 .54092 L .27962 .55084 L p 0 0 .206 r
F P
0 g
s
.4332 .51868 m .45477 .51435 L .46114 .49963 L p 0 0 .798 r
F P
0 g
s
.46114 .49963 m .43355 .50517 L .4332 .51868 L p 0 0 .798 r
F P
0 g
s
.47798 .43453 m .50957 .41688 L .51265 .43237 L p 0 .653 0 r
F P
0 g
s
.51265 .43237 m .48359 .44962 L .47798 .43453 L p 0 .653 0 r
F P
0 g
s
.48359 .44962 m .51265 .43237 L .51838 .44715 L p 0 .547 0 r
F P
0 g
s
.40022 .54651 m .4053 .55134 L .39196 .55423 L p .599 0 0 r
F P
0 g
s
.39196 .55423 m .38616 .54876 L .40022 .54651 L p .599 0 0 r
F P
0 g
s
.46604 .54691 m .47586 .54723 L .47439 .53295 L p 0 0 .975 r
F P
0 g
s
.47439 .53295 m .45751 .54025 L .46604 .54691 L p 0 0 .975 r
F P
0 g
s
.33046 .53767 m .31468 .53338 L .31112 .54587 L p 0 0 .783 r
F P
0 g
s
.31468 .53338 m .33046 .53767 L .3363 .52891 L p 0 0 .692 r
F P
0 g
s
.75408 .53814 m .74934 .55721 L .73489 .5508 L p .691 0 0 r
F P
0 g
s
.74934 .55721 m .75408 .53814 L .76907 .54381 L p .605 0 0 r
F P
0 g
s
.76907 .54381 m .7623 .5667 L .74934 .55721 L p .605 0 0 r
F P
0 g
s
.31509 .47589 m .34789 .47435 L .33829 .46312 L p 0 0 .358 r
F P
0 g
s
.33829 .46312 m .30338 .4701 L .31509 .47589 L p 0 0 .358 r
F P
0 g
s
.513 .39381 m .5499 .38677 L .54841 .39529 L p 0 .841 0 r
F P
0 g
s
.54841 .39529 m .50968 .40326 L .513 .39381 L p 0 .841 0 r
F P
0 g
s
.58981 .39777 m .54841 .39529 L .5499 .38677 L p 0 .744 0 r
F P
0 g
s
.55223 .45272 m .52584 .45875 L .51838 .44715 L p 0 .455 0 r
F P
0 g
s
.52584 .45875 m .55223 .45272 L .55482 .46069 L p 0 .226 0 r
F P
0 g
s
.55223 .45272 m .5797 .45447 L .57586 .46198 L p 0 .268 0 r
F P
0 g
s
.57586 .46198 m .55482 .46069 L .55223 .45272 L p 0 .268 0 r
F P
0 g
s
.29282 .47045 m .25645 .49824 L .26215 .48777 L p 0 0 .884 r
F P
0 g
s
.26215 .48777 m .30338 .4701 L .29282 .47045 L p 0 0 .884 r
F P
0 g
s
.23148 .52488 m .25645 .49824 L .25565 .5122 L p 0 0 .652 r
F P
0 g
s
.25565 .5122 m .235 .54063 L .23148 .52488 L p 0 0 .652 r
F P
0 g
s
.54475 .56434 m .52921 .54278 L .53572 .52812 L p 0 0 .555 r
F P
0 g
s
.52921 .54278 m .54475 .56434 L .53266 .57291 L p 0 0 .424 r
F P
0 g
s
.53266 .57291 m .51874 .5547 L .52921 .54278 L p 0 0 .424 r
F P
0 g
s
.5499 .38677 m .58887 .38902 L .58981 .39777 L p 0 .744 0 r
F P
0 g
s
.62713 .40998 m .58981 .39777 L .58887 .38902 L p 0 .626 0 r
F P
0 g
s
.58981 .39777 m .62713 .40998 L .62585 .4238 L p 0 .575 0 r
F P
0 g
s
.51838 .44715 m .49251 .46276 L .48359 .44962 L p 0 .547 0 r
F P
0 g
s
.49251 .46276 m .51838 .44715 L .52584 .45875 L p 0 .397 0 r
F P
0 g
s
.49631 .47724 m .45872 .46729 L .44755 .47696 L p 0 0 .544 r
F P
0 g
s
.33383 .50105 m .35902 .5036 L .35523 .4887 L p 0 0 .622 r
F P
0 g
s
.35902 .5036 m .33383 .50105 L .33751 .51595 L p 0 0 .707 r
F P
0 g
s
.71628 .55482 m .72137 .54855 L .73489 .5508 L p .72 0 0 r
F P
0 g
s
.73862 .53462 m .73489 .5508 L .72137 .54855 L p .777 0 0 r
F P
0 g
s
.73489 .5508 m .73862 .53462 L .75408 .53814 L p .691 0 0 r
F P
0 g
s
.5797 .45447 m .6058 .46376 L .59743 .46909 L p 0 .27 0 r
F P
0 g
s
.59743 .46909 m .57586 .46198 L .5797 .45447 L p 0 .27 0 r
F P
0 g
s
.64846 .46023 m .62129 .43903 L .62585 .4238 L p 0 .442 0 r
F P
0 g
s
.62129 .43903 m .64846 .46023 L .63848 .4721 L p 0 .402 0 r
F P
0 g
s
.63848 .4721 m .61422 .45314 L .62129 .43903 L p 0 .402 0 r
F P
0 g
s
.4766 .42002 m .50968 .40326 L .50957 .41688 L p 0 .755 0 r
F P
0 g
s
.50957 .41688 m .47798 .43453 L .4766 .42002 L p 0 .755 0 r
F P
0 g
s
.72137 .54855 m .71628 .55482 L .7051 .55776 L p .796 0 0 r
F P
0 g
s
.5537 .55108 m .53572 .52812 L .53714 .51317 L p 0 0 .664 r
F P
0 g
s
.53572 .52812 m .5537 .55108 L .54475 .56434 L p 0 0 .555 r
F P
0 g
s
.27962 .55084 m .26688 .56827 L .25432 .56435 L p 0 0 .206 r
F P
0 g
s
.26688 .56827 m .27962 .55084 L .2919 .55535 L p 0 0 0 r
F P
0 g
s
.705 .63419 m .73044 .6259 L .71733 .61986 L p .508 0 0 r
F P
0 g
s
.71733 .61986 m .69345 .62822 L .705 .63419 L p .508 0 0 r
F P
0 g
s
.7623 .5667 m .76907 .54381 L .78109 .55067 L p .491 0 0 r
F P
0 g
s
.78109 .55067 m .7716 .5777 L .7623 .5667 L p .491 0 0 r
F P
0 g
s
.36704 .60634 m .34922 .5903 L .35165 .57835 L p .946 0 0 r
F P
0 g
s
.34922 .5903 m .36704 .60634 L .37332 .61828 L p .879 0 0 r
F P
0 g
s
.42628 .56432 m .43834 .56787 L .42811 .55741 L p .782 0 0 r
F P
0 g
s
.42811 .55741 m .41745 .55489 L .42628 .56432 L p .782 0 0 r
F P
0 g
s
.75714 .60836 m .77564 .58836 L .77365 .59687 L p 0 0 0 r
F P
0 g
s
.77365 .59687 m .75222 .61953 L .75714 .60836 L p 0 0 0 r
F P
0 g
s
.65574 .44571 m .62585 .4238 L .62713 .40998 L p 0 .474 0 r
F P
0 g
s
.62585 .4238 m .65574 .44571 L .64846 .46023 L p 0 .442 0 r
F P
0 g
s
.51874 .5547 m .53266 .57291 L .51945 .57527 L p 0 0 .219 r
F P
0 g
s
.51945 .57527 m .50603 .56186 L .51874 .5547 L p 0 0 .219 r
F P
0 g
s
.61422 .45314 m .63848 .4721 L .62744 .47931 L p 0 .34 0 r
F P
0 g
s
.62744 .47931 m .6058 .46376 L .61422 .45314 L p 0 .34 0 r
F P
0 g
s
.38251 .53584 m .38616 .54876 L .37184 .55569 L p .714 0 0 r
F P
0 g
s
.37184 .55569 m .36692 .53985 L .38251 .53584 L p .714 0 0 r
F P
0 g
s
.36692 .53985 m .37184 .55569 L .35965 .56613 L p .799 0 0 r
F P
0 g
s
.32589 .48685 m .35523 .4887 L .34789 .47435 L p 0 0 .522 r
F P
0 g
s
.34789 .47435 m .31509 .47589 L .32589 .48685 L p 0 0 .522 r
F P
0 g
s
.35523 .4887 m .32589 .48685 L .33383 .50105 L p 0 0 .622 r
F P
0 g
s
.43143 .62465 m .4103 .61991 L .39699 .62578 L p .655 0 0 r
F P
0 g
s
.4103 .61991 m .43143 .62465 L .4406 .61363 L p .837 0 0 r
F P
0 g
s
.41166 .54936 m .41745 .55489 L .4053 .55134 L p .244 0 0 r
F P
0 g
s
.4053 .55134 m .40022 .54651 L .41166 .54936 L p .244 0 0 r
F P
0 g
s
.39674 .53494 m .40022 .54651 L .38616 .54876 L p .601 0 0 r
F P
0 g
s
.38616 .54876 m .38251 .53584 L .39674 .53494 L p .601 0 0 r
F P
0 g
s
.43355 .50517 m .46114 .49963 L .47148 .48721 L p 0 0 .711 r
F P
0 g
s
.47148 .48721 m .43863 .49038 L .43355 .50517 L p 0 0 .711 r
F P
0 g
s
.49631 .47724 m .53323 .50046 L .53714 .51317 L p 0 0 .178 r
F P
0 g
s
.55801 .53537 m .53714 .51317 L .53323 .50046 L p 0 0 .76 r
F P
0 g
s
.53714 .51317 m .55801 .53537 L .5537 .55108 L p 0 0 .664 r
F P
0 g
s
.75085 .50949 m .75408 .53814 L .73862 .53462 L p .719 0 0 r
F P
0 g
s
.75408 .53814 m .75085 .50949 L .76696 .51189 L p .633 0 0 r
F P
0 g
s
.76696 .51189 m .76907 .54381 L .75408 .53814 L p .633 0 0 r
F P
0 g
s
.48395 .47929 m .44755 .47696 L .43863 .49038 L p 0 0 .632 r
F P
0 g
s
.44755 .47696 m .48395 .47929 L .49631 .47724 L p 0 0 .544 r
F P
0 g
s
.35965 .56613 m .35255 .54632 L .36692 .53985 L p .799 0 0 r
F P
0 g
s
.35255 .54632 m .35965 .56613 L .35165 .57835 L p .88 0 0 r
F P
0 g
s
.23309 .50996 m .26215 .48777 L .25645 .49824 L p 0 0 .751 r
F P
0 g
s
.25645 .49824 m .23148 .52488 L .23309 .50996 L p 0 0 .751 r
F P
0 g
s
.72137 .54855 m .72526 .53382 L .73862 .53462 L p .777 0 0 r
F P
0 g
s
.7051 .55776 m .71104 .55087 L p .71091 .55447 L .796 0 0 r
F P
0 g
s
.70572 .55742 m .71104 .55087 L p .70184 .56146 L .544 0 0 r
F P
0 g
s
.69753 .56642 m .70572 .55742 L p .70184 .56146 L .544 0 0 r
F P
0 g
s
.70572 .55742 m .70641 .56714 L p .71002 .55295 L 0 0 0 r
F P
0 g
s
.71104 .55087 m .70572 .55742 L p .71122 .54617 L .9 0 0 r
F P
0 g
s
.71624 .5359 m .71104 .55087 L p .71122 .54617 L .9 0 0 r
F P
0 g
s
.71318 .54056 m .71624 .5359 L p .70841 .55192 L .9 0 0 r
F P
0 g
s
.70572 .55742 m .71318 .54056 L p .70841 .55192 L .9 0 0 r
F P
0 g
s
.72526 .53382 m .72137 .54855 L .71104 .55087 L p .883 0 0 r
F P
0 g
s
.71104 .55087 m .71624 .5359 L .72526 .53382 L p .883 0 0 r
F P
0 g
s
.72483 .50959 m .72526 .53382 L .71624 .5359 L p .922 0 0 r
F P
0 g
s
.71932 .5121 m .71624 .5359 L .71318 .54056 L p .924 0 0 r
F P
0 g
s
.71624 .5359 m .71932 .5121 L .72483 .50959 L p .922 0 0 r
F P
0 g
s
.71318 .54056 m .71669 .54703 L p .71924 .52609 L 0 0 0 r
F P
0 g
s
.72049 .5158 m .71318 .54056 L p .71924 .52609 L 0 0 0 r
F P
0 g
s
.71318 .54056 m .70572 .55742 L p .71002 .55295 L 0 0 0 r
F P
0 g
s
.71104 .55087 m .72137 .54855 L p .71091 .55447 L .796 0 0 r
F P
0 g
s
.52584 .45875 m .50324 .47175 L .49251 .46276 L p 0 .397 0 r
F P
0 g
s
.50324 .47175 m .52584 .45875 L .53378 .4652 L p 0 .082 0 r
F P
0 g
s
.55482 .46069 m .53378 .4652 L .52584 .45875 L p 0 .226 0 r
F P
0 g
s
.73598 .50867 m .73862 .53462 L .72526 .53382 L p .808 0 0 r
F P
0 g
s
.73862 .53462 m .73598 .50867 L .75085 .50949 L p .719 0 0 r
F P
0 g
s
.53323 .50046 m .49631 .47724 L .48395 .47929 L p 0 0 .418 r
F P
0 g
s
.43863 .49038 m .47148 .48721 L .48395 .47929 L p 0 0 .632 r
F P
0 g
s
.47974 .40852 m .513 .39381 L .50968 .40326 L p 0 .879 0 r
F P
0 g
s
.50968 .40326 m .4766 .42002 L .47974 .40852 L p 0 .879 0 r
F P
0 g
s
.45751 .54025 m .47439 .53295 L .47844 .51751 L p 0 0 .874 r
F P
0 g
s
.47844 .51751 m .4535 .52874 L .45751 .54025 L p 0 0 .874 r
F P
0 g
s
.3363 .52891 m .31305 .51832 L .31468 .53338 L p 0 0 .692 r
F P
0 g
s
.31305 .51832 m .33329 .52754 L p .33407 .51629 L 0 0 .598 r
F P
0 g
s
.33751 .51595 m .30648 .50333 L .31305 .51832 L p 0 0 .598 r
F P
0 g
s
.30648 .50333 m .33751 .51595 L .33383 .50105 L p 0 0 .489 r
F P
0 g
s
.29613 .4911 m .33383 .50105 L .32589 .48685 L p 0 0 .318 r
F P
0 g
s
.33383 .50105 m .29613 .4911 L .30648 .50333 L p 0 0 .489 r
F P
0 g
s
.32835 .53045 m .3277 .48744 L .3369 .48469 L p .922 0 0 r
F P
0 g
s
.33047 .52595 m .32835 .53045 L p .3328 .50665 L .922 0 0 r
F P
0 g
s
.3328 .50665 m .33373 .50437 L .33047 .52595 L .922 0 0 r
F
.33576 .49196 m .33047 .52595 L p .3359 .49004 L F P
0 g
s
.3363 .52891 m .33751 .51595 L p .33407 .51629 L 0 0 .598 r
F P
0 g
s
.33329 .52754 m .3363 .52891 L p .33407 .51629 L 0 0 .598 r
F P
0 g
s
.35074 .48269 m .33893 .52118 L .33047 .52595 L p .828 0 0 r
F P
0 g
s
.33047 .52595 m .3369 .48469 L .35074 .48269 L p .828 0 0 r
F P
0 g
s
.33653 .56212 m .33047 .52595 L .33893 .52118 L p .935 0 0 r
F P
0 g
s
.34922 .5903 m .33653 .56212 L .3418 .55418 L p .964 0 0 r
F P
0 g
s
.33653 .56212 m .34922 .5903 L .35286 .59993 L p .844 0 0 r
F P
0 g
s
.33893 .52118 m .3418 .55418 L .33653 .56212 L p .935 0 0 r
F P
0 g
s
.35286 .59993 m .33775 .56877 L .33653 .56212 L p .844 0 0 r
F P
0 g
s
.33047 .52595 m .33653 .56212 L .33775 .56877 L p .966 0 0 r
F P
0 g
s
.33775 .56877 m .32835 .53045 L .33047 .52595 L p .966 0 0 r
F P
0 g
s
.32835 .53045 m .33304 .53392 L p .33023 .50948 L .975 0 0 r
F P
0 g
s
.3277 .48744 m .328 .50712 L p .33023 .50948 L .975 0 0 r
F P
0 g
s
.328 .50712 m .32835 .53045 L p .33023 .50948 L .975 0 0 r
F P
0 g
s
.3248 .49047 m .3277 .48744 L p .33054 .51213 L .975 0 0 r
F P
0 g
s
.3288 .49326 m .3248 .49047 L p .33485 .51435 L 0 0 0 r
F P
0 g
s
.3418 .55418 m .33893 .52118 L .3522 .51696 L p .839 0 0 r
F P
0 g
s
.34385 .53571 m .33713 .51674 L p .33485 .51435 L 0 0 0 r
F P
0 g
s
.3328 .50665 m .3359 .49004 L .33373 .50437 L .922 0 0 r
F
.33713 .51674 m .3288 .49326 L p .33485 .51435 L 0 0 0 r
F P
0 g
s
.33304 .53392 m .32942 .51485 L p .33054 .51213 L .975 0 0 r
F P
0 g
s
.32942 .51485 m .3248 .49047 L p .33054 .51213 L .975 0 0 r
F P
0 g
s
.3369 .48469 m .33576 .49196 L p .3359 .49004 L .922 0 0 r
F P
0 g
s
.69753 .56642 m .68619 .57083 L .68063 .58511 L p .364 0 0 r
F P
0 g
s
.43022 .57806 m .4444 .58248 L .43834 .56787 L p .897 0 0 r
F P
0 g
s
.43834 .56787 m .42628 .56432 L .43022 .57806 L p .897 0 0 r
F P
0 g
s
.69345 .62822 m .71733 .61986 L .70578 .60866 L p .498 0 0 r
F P
0 g
s
.70578 .60866 m .68461 .61659 L .69345 .62822 L p .498 0 0 r
F P
0 g
s
.2919 .55535 m .27979 .56469 L p .26688 .56827 L 0 0 0 r
F P
0 g
s
.27859 .56561 m .2919 .55535 L .30304 .55362 L p 0 0 .845 r
F P
0 g
s
.28743 .55675 m .30172 .55388 L p .31112 .54587 L 0 0 .8 r
F P
0 g
s
.30172 .55388 m .30304 .55362 L .30387 .55282 L p 0 0 .8 r
F P
0 g
s
.30387 .55282 m .31112 .54587 L p .30172 .55388 L 0 0 .8 r
F P
0 g
s
.27979 .56469 m .27859 .56561 L .27734 .5659 L p 0 0 0 r
F P
0 g
s
.27734 .5659 m .26688 .56827 L p .27979 .56469 L 0 0 0 r
F P
0 g
s
.50603 .56186 m .51945 .57527 L .50739 .57096 L p 0 0 0 r
F P
0 g
s
.50739 .57096 m .49324 .56299 L .50603 .56186 L p 0 0 0 r
F P
0 g
s
.30338 .4701 m .26215 .48777 L .27183 .48263 L p 0 0 .803 r
F P
0 g
s
.27183 .48263 m .31509 .47589 L .30338 .4701 L p 0 0 .803 r
F P
0 g
s
.76907 .54381 m .76696 .51189 L .78164 .51547 L p .521 0 0 r
F P
0 g
s
.78164 .51547 m .78109 .55067 L .76907 .54381 L p .521 0 0 r
F P
0 g
s
.7716 .5777 m .78109 .55067 L .78809 .5576 L p .27 0 0 r
F P
0 g
s
.78809 .5576 m .77564 .58836 L .7716 .5777 L p .27 0 0 r
F P
0 g
s
.4406 .61363 m .42148 .60856 L .4103 .61991 L p .837 0 0 r
F P
0 g
s
.42148 .60856 m .4406 .61363 L .44521 .5987 L p .899 0 0 r
F P
0 g
s
.58887 .38902 m .62488 .3999 L .62713 .40998 L p 0 .626 0 r
F P
0 g
s
.65909 .43095 m .62713 .40998 L .62488 .3999 L p 0 .493 0 r
F P
0 g
s
.62713 .40998 m .65909 .43095 L .65574 .44571 L p 0 .474 0 r
F P
0 g
s
.6058 .46376 m .62744 .47931 L .61721 .48059 L p 0 .205 0 r
F P
0 g
s
.61721 .48059 m .59743 .46909 L .6058 .46376 L p 0 .205 0 r
F P
0 g
s
.40721 .53734 m .41166 .54936 L .40022 .54651 L p .382 0 0 r
F P
0 g
s
.40022 .54651 m .39674 .53494 L .40721 .53734 L p .382 0 0 r
F P
0 g
s
.37332 .61828 m .35286 .59993 L .34922 .5903 L p .879 0 0 r
F P
0 g
s
.35286 .59993 m .37332 .61828 L .3839 .62518 L p 0 0 0 r
F P
0 g
s
.51899 .39022 m .55215 .38451 L .5499 .38677 L p 0 .879 0 r
F P
0 g
s
.5499 .38677 m .513 .39381 L .51899 .39022 L p 0 .879 0 r
F P
0 g
s
.58887 .38902 m .5499 .38677 L .55215 .38451 L p 0 .748 0 r
F P
0 g
s
.38377 .51274 m .38251 .53584 L .36692 .53985 L p .671 0 0 r
F P
0 g
s
.36692 .53985 m .36802 .51399 L .38377 .51274 L p .671 0 0 r
F P
0 g
s
.36802 .51399 m .36692 .53985 L .35255 .54632 L p .757 0 0 r
F P
0 g
s
.53378 .4652 m .55482 .46069 L .55732 .46231 L p 0 0 0 r
F P
0 g
s
.55732 .46231 m .55482 .46069 L p .56464 .46205 L 0 0 0 r
F P
0 g
s
.57275 .46317 m .55732 .46231 L p .56464 .46205 L 0 0 0 r
F P
0 g
s
.55732 .46231 m .55735 .46224 L p .55688 .46232 L 0 .902 0 r
F P
0 g
s
.54085 .46532 m .53378 .4652 L p .54778 .46348 L 0 0 0 r
F P
0 g
s
.55732 .46231 m .54085 .46532 L p .54778 .46348 L 0 0 0 r
F P
0 g
s
.514 .47503 m .53378 .4652 L .54085 .46532 L p 0 0 0 r
F P
0 g
s
.54085 .46532 m .54581 .45898 L p .52923 .4684 L 0 .946 0 r
F P
0 g
s
.52294 .47197 m .514 .47503 L p .53423 .46771 L 0 0 0 r
F P
0 g
s
.54085 .46532 m .52294 .47197 L p .53423 .46771 L 0 0 0 r
F P
0 g
s
.53378 .4652 m .514 .47503 L .50324 .47175 L p 0 .082 0 r
F P
0 g
s
.48395 .47929 m .52458 .49218 L .53323 .50046 L p 0 0 .418 r
F P
0 g
s
.52458 .49218 m .48395 .47929 L .47148 .48721 L p 0 0 .538 r
F P
0 g
s
.48947 .48823 m .50324 .47175 L .514 .47503 L p 0 0 0 r
F P
0 g
s
.514 .47503 m .49999 .48478 L .48947 .48823 L p 0 0 0 r
F P
0 g
s
.49999 .48478 m .514 .47503 L .52294 .47197 L p 0 .995 0 r
F P
0 g
s
.52294 .47197 m .50683 .4755 L .50424 .47901 L p 0 .995 0 r
F P
0 g
s
.50424 .47901 m .49999 .48478 L p .52294 .47197 L 0 .995 0 r
F P
0 g
s
.52294 .47197 m .54085 .46532 L p .52923 .4684 L 0 .946 0 r
F P
0 g
s
.54085 .46532 m .55732 .46231 L p .55688 .46232 L 0 .902 0 r
F P
0 g
s
.55482 .46069 m .57336 .46183 L p .56464 .46205 L 0 0 0 r
F P
0 g
s
.55735 .46224 m .55931 .45718 L p .54725 .46249 L 0 .902 0 r
F P
0 g
s
.54725 .46249 m .54085 .46532 L .55735 .46224 L 0 .902 0 r
F
.57586 .46198 m .57275 .46317 L p .56464 .46205 L 0 0 0 r
F P
0 g
s
.57336 .46183 m .57586 .46198 L p .56464 .46205 L 0 0 0 r
F P
0 g
s
.39684 .51344 m .39674 .53494 L .38251 .53584 L p .551 0 0 r
F P
0 g
s
.38251 .53584 m .38377 .51274 L .39684 .51344 L p .551 0 0 r
F P
0 g
s
.55215 .38451 m .58643 .38638 L .58887 .38902 L p 0 .748 0 r
F P
0 g
s
.62488 .3999 m .58887 .38902 L .58643 .38638 L p 0 .546 0 r
F P
0 g
s
.7338 .47687 m .75085 .50949 L .73598 .50867 L p .768 0 0 r
F P
0 g
s
.75085 .50949 m .7338 .47687 L .74756 .47598 L p .682 0 0 r
F P
0 g
s
.74756 .47598 m .76696 .51189 L .75085 .50949 L p .682 0 0 r
F P
0 g
s
.75222 .61953 m .77365 .59687 L .76588 .60169 L p .707 0 0 r
F P
0 g
s
.76588 .60169 m .74279 .62576 L .75222 .61953 L p .707 0 0 r
F P
0 g
s
.58639 .46093 m .57091 .45772 L .57245 .4623 L p 0 .651 0 r
F P
0 g
s
.59056 .46812 m .57275 .46317 L .57586 .46198 L p 0 0 0 r
F P
0 g
s
.57245 .4623 m .57275 .46317 L p .5744 .4629 L 0 .651 0 r
F P
0 g
s
.57586 .46198 m .59743 .46909 L .59196 .46832 L p 0 0 0 r
F P
0 g
s
.5884 .46722 m .57586 .46198 L .59196 .46832 L 0 0 0 r
F
.58639 .46093 m .57245 .4623 L .5744 .4629 L 0 .651 0 r
F
.59196 .46832 m .59056 .46812 L p .5884 .46722 L 0 0 0 r
F P
0 g
s
.42853 .59373 m .44521 .5987 L .4444 .58248 L p .918 0 0 r
F P
0 g
s
.4444 .58248 m .43022 .57806 L .42853 .59373 L p .918 0 0 r
F P
0 g
s
.44521 .5987 m .42853 .59373 L .42148 .60856 L p .899 0 0 r
F P
0 g
s
.42298 .56079 m .41745 .55489 L p .41848 .55687 L .579 0 0 r
F P
0 g
s
.41745 .55489 m .41166 .54936 L .41848 .55687 L p .579 0 0 r
F P
0 g
s
.41209 .54268 m .41848 .55687 L .41166 .54936 L p 0 0 0 r
F P
0 g
s
.41166 .54936 m .40721 .53734 L .41209 .54268 L p 0 0 0 r
F P
0 g
s
.41848 .55687 m .41209 .54268 L .41049 .5499 L p .976 0 0 r
F P
0 g
s
.41049 .5499 m .41045 .55008 L p .41848 .55687 L .976 0 0 r
F P
0 g
s
.41848 .55687 m .42628 .56432 L .42298 .56079 L p .579 0 0 r
F P
0 g
s
.76696 .51189 m .74756 .47598 L .76367 .47624 L p .585 0 0 r
F P
0 g
s
.76367 .47624 m .78164 .51547 L .76696 .51189 L p .585 0 0 r
F P
0 g
s
.55695 .51986 m .53323 .50046 L .52458 .49218 L p 0 0 .652 r
F P
0 g
s
.53323 .50046 m .55695 .51986 L .55801 .53537 L p 0 0 .76 r
F P
0 g
s
.35165 .57835 m .3418 .55418 L .35255 .54632 L p .88 0 0 r
F P
0 g
s
.3418 .55418 m .35165 .57835 L .34922 .5903 L p .964 0 0 r
F P
0 g
s
.69494 .57934 m .68063 .58511 L .68006 .60132 L p .428 0 0 r
F P
0 g
s
.68063 .58511 m .69494 .57934 L .69753 .56642 L p .364 0 0 r
F P
0 g
s
.68461 .61659 m .70578 .60866 L .69785 .5943 L p .468 0 0 r
F P
0 g
s
.69785 .5943 m .68006 .60132 L .68461 .61659 L p .468 0 0 r
F P
0 g
s
.72526 .53382 m .72483 .50959 L .73598 .50867 L p .808 0 0 r
F P
0 g
s
.45523 .46371 m .47798 .43453 L .48359 .44962 L p 0 .556 0 r
F P
0 g
s
.48359 .44962 m .46502 .47666 L .45523 .46371 L p 0 .556 0 r
F P
0 g
s
.46502 .47666 m .48359 .44962 L .49251 .46276 L p 0 .404 0 r
F P
0 g
s
.23962 .49843 m .26899 .48402 L p .26215 .48777 L 0 0 .856 r
F P
0 g
s
.26215 .48777 m .23309 .50996 L .23962 .49843 L p 0 0 .856 r
F P
0 g
s
.27183 .48263 m .23962 .49843 L .24997 .49228 L p 0 0 .771 r
F P
0 g
s
.28385 .48378 m .25286 .49156 L p .26241 .49265 L 0 0 0 r
F P
0 g
s
.25286 .49156 m .24997 .49228 L .25123 .49232 L p 0 0 0 r
F P
0 g
s
.25123 .49232 m .26241 .49265 L p .25286 .49156 L 0 0 0 r
F P
0 g
s
.26899 .48402 m .27183 .48263 L .27079 .48318 L p 0 0 .856 r
F P
0 g
s
.27079 .48318 m .26215 .48777 L p .26899 .48402 L 0 0 .856 r
F P
0 g
s
.35255 .54632 m .3522 .51696 L .36802 .51399 L p .757 0 0 r
F P
0 g
s
.3522 .51696 m .35255 .54632 L .3418 .55418 L p .839 0 0 r
F P
0 g
s
.68006 .60132 m .69785 .5943 L .69494 .57934 L p .428 0 0 r
F P
0 g
s
.7247 .47879 m .73598 .50867 L .72483 .50959 L p .87 0 0 r
F P
0 g
s
.73598 .50867 m .7247 .47879 L .7338 .47687 L p .768 0 0 r
F P
0 g
s
.44937 .44864 m .4766 .42002 L .47798 .43453 L p 0 .673 0 r
F P
0 g
s
.47798 .43453 m .45523 .46371 L .44937 .44864 L p 0 .673 0 r
F P
0 g
s
.28385 .48378 m .32589 .48685 L .31509 .47589 L p 0 0 0 r
F P
0 g
s
.31509 .47589 m .27183 .48263 L .28385 .48378 L p 0 0 0 r
F P
0 g
s
.73945 .44092 m .76367 .47624 L .74756 .47598 L p .717 0 0 r
F P
0 g
s
.78164 .51547 m .76367 .47624 L .77946 .47762 L p .432 0 0 r
F P
0 g
s
.76367 .47624 m .73945 .44092 L .75464 .43899 L p .627 0 0 r
F P
0 g
s
.75464 .43899 m .77946 .47762 L .76367 .47624 L p .627 0 0 r
F P
0 g
s
.3954 .48345 m .38377 .51274 L .36802 .51399 L p .548 0 0 r
F P
0 g
s
.36802 .51399 m .38263 .48201 L .3954 .48345 L p .548 0 0 r
F P
0 g
s
.38263 .48201 m .36802 .51399 L .3522 .51696 L p .664 0 0 r
F P
0 g
s
.4535 .52874 m .47844 .51751 L .48732 .50366 L p 0 0 .775 r
F P
0 g
s
.48732 .50366 m .45477 .51435 L .4535 .52874 L p 0 0 .775 r
F P
0 g
s
.49251 .46276 m .47711 .48534 L .46502 .47666 L p 0 .404 0 r
F P
0 g
s
.47711 .48534 m .49251 .46276 L .50324 .47175 L p 0 .126 0 r
F P
0 g
s
.405 .51598 m .40721 .53734 L .39674 .53494 L p .298 0 0 r
F P
0 g
s
.39674 .53494 m .39684 .51344 L .405 .51598 L p .298 0 0 r
F P
0 g
s
.72794 .44377 m .74756 .47598 L .7338 .47687 L p .807 0 0 r
F P
0 g
s
.74756 .47598 m .72794 .44377 L .73945 .44092 L p .717 0 0 r
F P
0 g
s
.49324 .56299 m .50739 .57096 L .4986 .56066 L p 0 0 .835 r
F P
0 g
s
.4986 .56066 m .48255 .55785 L .49324 .56299 L p 0 0 .835 r
F P
0 g
s
.30304 .55362 m .28743 .55675 L .27859 .56561 L p 0 0 .845 r
F P
0 g
s
.66306 .49011 m .64846 .46023 L .65574 .44571 L p 0 .295 0 r
F P
0 g
s
.64846 .46023 m .66306 .49011 L .65082 .49533 L p 0 .24 0 r
F P
0 g
s
.65082 .49533 m .63848 .4721 L .64846 .46023 L p 0 .24 0 r
F P
0 g
s
.78109 .55067 m .78164 .51547 L .79243 .51966 L p .305 0 0 r
F P
0 g
s
.79243 .51966 m .78809 .5576 L .78109 .55067 L p .305 0 0 r
F P
0 g
s
.67443 .47961 m .65574 .44571 L .65909 .43095 L p 0 .318 0 r
F P
0 g
s
.65574 .44571 m .67443 .47961 L .66306 .49011 L p 0 .295 0 r
F P
0 g
s
.48691 .40205 m .51899 .39022 L .513 .39381 L p 0 .968 0 r
F P
0 g
s
.513 .39381 m .47974 .40852 L .48691 .40205 L p 0 .968 0 r
F P
0 g
s
.3839 .62518 m .36202 .60553 L .35286 .59993 L p 0 0 0 r
F P
0 g
s
.36202 .60553 m .3839 .62518 L .39699 .62578 L p .443 0 0 r
F P
0 g
s
.70572 .55742 m .69753 .56642 L .69494 .57934 L p .123 0 0 r
F P
0 g
s
.403 .48584 m .39684 .51344 L .38377 .51274 L p .307 0 0 r
F P
0 g
s
.40406 .48878 m .405 .51598 L .39684 .51344 L p 0 0 0 r
F P
0 g
s
.3954 .48345 m .403 .48584 L p .39758 .49343 L .307 0 0 r
F P
0 g
s
.403 .48584 m .40406 .48878 L p .40036 .50141 L 0 0 0 r
F P
0 g
s
.39684 .51344 m .39976 .50034 L p .40036 .50141 L 0 0 0 r
F P
0 g
s
.39976 .50034 m .403 .48584 L p .40036 .50141 L 0 0 0 r
F P
0 g
s
.405 .51598 m .40494 .51436 L p .40307 .50902 L .986 0 0 r
F P
0 g
s
.40494 .51436 m .40406 .48878 L p .40307 .50902 L .986 0 0 r
F P
0 g
s
.40406 .48878 m .39827 .49178 L p .40307 .50902 L .986 0 0 r
F P
0 g
s
.40307 .50902 m .40151 .50345 L .40603 .51744 L .986 0 0 r
F
.39827 .49178 m .40212 .50454 L p .40151 .50345 L F P
0 g
s
.40212 .50454 m .40603 .51744 L p .40151 .50345 L .986 0 0 r
F P
0 g
s
.40679 .51996 m .405 .51598 L p .40307 .50902 L .986 0 0 r
F P
0 g
s
.40603 .51744 m .40679 .51996 L p .40307 .50902 L .986 0 0 r
F P
0 g
s
.38377 .51274 m .39469 .48523 L p .39758 .49343 L .307 0 0 r
F P
0 g
s
.39469 .48523 m .3954 .48345 L p .39758 .49343 L .307 0 0 r
F P
0 g
s
.77946 .47762 m .79243 .51966 L .78164 .51547 L p .432 0 0 r
F P
0 g
s
.3522 .51696 m .36687 .48174 L .38263 .48201 L p .664 0 0 r
F P
0 g
s
.36687 .48174 m .3522 .51696 L .33893 .52118 L p .748 0 0 r
F P
0 g
s
.77564 .58836 m .78809 .5576 L .78877 .56341 L p 0 0 0 r
F P
0 g
s
.78877 .56341 m .77365 .59687 L .77564 .58836 L p 0 0 0 r
F P
0 g
s
.32589 .48685 m .28385 .48378 L .29613 .4911 L p 0 0 .318 r
F P
0 g
s
.58643 .38638 m .61943 .39535 L .62488 .3999 L p 0 .546 0 r
F P
0 g
s
.65793 .41843 m .62488 .3999 L .61943 .39535 L p 0 .341 0 r
F P
0 g
s
.62488 .3999 m .65793 .41843 L .65909 .43095 L p 0 .493 0 r
F P
0 g
s
.59743 .46909 m .61721 .48059 L .60956 .47564 L p 0 0 0 r
F P
0 g
s
.60956 .47564 m .59056 .46812 L .59743 .46909 L p 0 0 0 r
F P
0 g
s
.41945 .56783 m .43022 .57806 L .42628 .56432 L p .947 0 0 r
F P
0 g
s
.42628 .56432 m .41848 .55687 L .41945 .56783 L p .947 0 0 r
F P
0 g
s
.79243 .51966 m .77946 .47762 L .7923 .47989 L p .051 0 0 r
F P
0 g
s
.77946 .47762 m .75464 .43899 L .77097 .43828 L p .503 0 0 r
F P
0 g
s
.77097 .43828 m .7923 .47989 L .77946 .47762 L p .503 0 0 r
F P
0 g
s
.74279 .62576 m .76588 .60169 L .7536 .60193 L p .752 0 0 r
F P
0 g
s
.7536 .60193 m .73044 .6259 L .74279 .62576 L p .752 0 0 r
F P
0 g
s
.74022 .40463 m .75464 .43899 L .73945 .44092 L p .765 0 0 r
F P
0 g
s
.75464 .43899 m .74022 .40463 L .7559 .40099 L p .679 0 0 r
F P
0 g
s
.7559 .40099 m .77097 .43828 L .75464 .43899 L p .679 0 0 r
F P
0 g
s
.63848 .4721 m .65082 .49533 L .63977 .49433 L p 0 .118 0 r
F P
0 g
s
.63977 .49433 m .62744 .47931 L .63848 .4721 L p 0 .118 0 r
F P
0 g
s
.20992 .53653 m .23148 .52488 L .235 .54063 L p 0 0 .589 r
F P
0 g
s
.235 .54063 m .21326 .55253 L .20992 .53653 L p 0 0 .589 r
F P
0 g
s
.21326 .55253 m .235 .54063 L .24307 .55457 L p 0 0 .45 r
F P
0 g
s
.47148 .48721 m .51263 .48981 L .52458 .49218 L p 0 0 .538 r
F P
0 g
s
.52458 .49218 m .55068 .50716 L .55695 .51986 L p 0 0 .652 r
F P
0 g
s
.55068 .50716 m .52458 .49218 L .51263 .48981 L p 0 0 0 r
F P
0 g
s
.402 .45162 m .38263 .48201 L .36687 .48174 L p .435 0 0 r
F P
0 g
s
.40514 .45467 m .3954 .48345 L .38263 .48201 L p 0 0 0 r
F P
0 g
s
.402 .45162 m .40514 .45467 L p .39509 .46687 L 0 0 0 r
F P
0 g
s
.39124 .47156 m .39454 .46544 L .39509 .46687 L 0 0 0 r
F
.39454 .46544 m .402 .45162 L .39509 .46687 L F
.40514 .45467 m .40139 .45728 L p .39772 .47331 L .873 0 0 r
F P
0 g
s
.3954 .48345 m .39838 .47465 L p .39772 .47331 L .873 0 0 r
F P
0 g
s
.39838 .47465 m .40514 .45467 L p .39772 .47331 L .873 0 0 r
F P
0 g
s
.403 .48584 m .3954 .48345 L p .39899 .46778 L .873 0 0 r
F P
0 g
s
.40005 .47907 m .39796 .47429 L .40298 .48626 L .9 0 0 r
F
.38842 .47293 m .402 .45162 L p .39124 .47156 L 0 0 0 r
F P
0 g
s
.36687 .48174 m .38937 .45285 L p .39227 .45997 L .435 0 0 r
F P
0 g
s
.39129 .45899 m .39905 .47709 L p .39796 .47429 L .9 0 0 r
F P
0 g
s
.39905 .47709 m .40298 .48626 L p .39796 .47429 L .9 0 0 r
F P
0 g
s
.40139 .45728 m .4019 .46636 L p .39899 .46778 L .873 0 0 r
F P
0 g
s
.4019 .46636 m .403 .48584 L p .39899 .46778 L .873 0 0 r
F P
0 g
s
.38263 .48201 m .38842 .47293 L p .39124 .47156 L 0 0 0 r
F P
0 g
s
.40406 .48878 m .403 .48584 L p .40005 .47907 L .9 0 0 r
F P
0 g
s
.40298 .48626 m .40406 .48878 L p .40005 .47907 L .9 0 0 r
F P
0 g
s
.39263 .44866 m .402 .45162 L p .39227 .45997 L .435 0 0 r
F P
0 g
s
.38937 .45285 m .39263 .44866 L p .39227 .45997 L .435 0 0 r
F P
0 g
s
.39263 .44866 m .36687 .48174 L .35074 .48269 L p .601 0 0 r
F P
0 g
s
.33893 .52118 m .35074 .48269 L .36687 .48174 L p .748 0 0 r
F P
0 g
s
.44844 .43399 m .47974 .40852 L .4766 .42002 L p 0 .79 0 r
F P
0 g
s
.4766 .42002 m .44937 .44864 L .44844 .43399 L p 0 .79 0 r
F P
0 g
s
.51263 .48981 m .47148 .48721 L .46114 .49963 L p 0 0 .621 r
F P
0 g
s
.45477 .51435 m .48732 .50366 L .49944 .49386 L p 0 0 .695 r
F P
0 g
s
.49944 .49386 m .46114 .49963 L .45477 .51435 L p 0 0 .695 r
F P
0 g
s
.7923 .47989 m .77097 .43828 L .78575 .43892 L p .246 0 0 r
F P
0 g
s
.77097 .43828 m .7559 .40099 L .77194 .39935 L p .578 0 0 r
F P
0 g
s
.77194 .39935 m .78575 .43892 L .77097 .43828 L p .578 0 0 r
F P
0 g
s
.21176 .5203 m .23309 .50996 L .23148 .52488 L p 0 0 .704 r
F P
0 g
s
.23148 .52488 m .20992 .53653 L .21176 .5203 L p 0 0 .704 r
F P
0 g
s
.24307 .55457 m .22124 .56559 L .21326 .55253 L p 0 0 .45 r
F P
0 g
s
.22124 .56559 m .24307 .55457 L .25432 .56435 L p 0 0 .223 r
F P
0 g
s
.68303 .46562 m .65909 .43095 L .65793 .41843 L p 0 .301 0 r
F P
0 g
s
.65909 .43095 m .68303 .46562 L .67443 .47961 L p 0 .318 0 r
F P
0 g
s
.72208 .44707 m .7338 .47687 L .7247 .47879 L p .921 0 0 r
F P
0 g
s
.7338 .47687 m .72208 .44707 L .72794 .44377 L p .807 0 0 r
F P
0 g
s
.24997 .49228 m .28385 .48378 L .27183 .48263 L p 0 0 .771 r
F P
0 g
s
.52667 .3932 m .55477 .38901 L .55215 .38451 L p 0 0 0 r
F P
0 g
s
.55215 .38451 m .51899 .39022 L .52667 .3932 L p 0 0 0 r
F P
0 g
s
.58643 .38638 m .55215 .38451 L .55477 .38901 L p 0 .017 0 r
F P
0 g
s
.50324 .47175 m .48947 .48823 L .47711 .48534 L p 0 .126 0 r
F P
0 g
s
.5737 .55889 m .5537 .55108 L .55801 .53537 L p 0 0 .59 r
F P
0 g
s
.5537 .55108 m .5737 .55889 L .56545 .57146 L p 0 0 .434 r
F P
0 g
s
.56545 .57146 m .54475 .56434 L .5537 .55108 L p 0 0 .434 r
F P
0 g
s
.57706 .54309 m .55801 .53537 L .55695 .51986 L p 0 0 .718 r
F P
0 g
s
.55801 .53537 m .57706 .54309 L .5737 .55889 L p 0 0 .59 r
F P
0 g
s
.46114 .49963 m .49944 .49386 L .51263 .48981 L p 0 0 .621 r
F P
0 g
s
.55732 .46231 m .57275 .46317 L .57091 .45772 L p 0 .788 0 r
F P
0 g
s
.57091 .45772 m .55931 .45718 L .55732 .46231 L p 0 .788 0 r
F P
0 g
s
.35074 .48269 m .37869 .44628 L .39263 .44866 L p .601 0 0 r
F P
0 g
s
.37869 .44628 m .35074 .48269 L .3369 .48469 L p .697 0 0 r
F P
0 g
s
.55477 .38901 m .5829 .39044 L .58643 .38638 L p 0 .017 0 r
F P
0 g
s
.61943 .39535 m .58643 .38638 L .5829 .39044 L p 0 .103 0 r
F P
0 g
s
.72751 .40965 m .73945 .44092 L .72794 .44377 L p .863 0 0 r
F P
0 g
s
.73945 .44092 m .72751 .40965 L .74022 .40463 L p .765 0 0 r
F P
0 g
s
.72483 .50959 m .72273 .48964 L p .7247 .47879 L .87 0 0 r
F P
0 g
s
.72187 .48141 m .72483 .50959 L .71932 .5121 L p .987 0 0 r
F P
0 g
s
.72592 .4843 m .72124 .50401 L p .72049 .5158 L 0 0 0 r
F P
0 g
s
.72124 .50401 m .71932 .5121 L .72013 .51466 L p 0 0 0 r
F P
0 g
s
.72013 .51466 m .72049 .5158 L p .72124 .50401 L 0 0 0 r
F P
0 g
s
.72273 .48964 m .72187 .48141 L .72377 .47964 L p .87 0 0 r
F P
0 g
s
.72377 .47964 m .7247 .47879 L p .72273 .48964 L .87 0 0 r
F P
0 g
s
.57275 .46317 m .59056 .46812 L .58639 .46093 L p 0 .651 0 r
F P
0 g
s
.39699 .62578 m .37517 .60605 L .36202 .60553 L p .443 0 0 r
F P
0 g
s
.37517 .60605 m .39699 .62578 L .4103 .61991 L p .682 0 0 r
F P
0 g
s
.31112 .54587 m .29181 .54318 L .28743 .55675 L p 0 0 .8 r
F P
0 g
s
.29181 .54318 m .31112 .54587 L .31468 .53338 L p 0 0 .695 r
F P
0 g
s
.48255 .55785 m .4986 .56066 L .49465 .5461 L p 0 0 .892 r
F P
0 g
s
.49465 .5461 m .47586 .54723 L .48255 .55785 L p 0 0 .892 r
F P
0 g
s
.39605 .41424 m .37869 .44628 L .36251 .44486 L p .607 0 0 r
F P
0 g
s
.36251 .44486 m .38211 .40984 L .39605 .41424 L p .607 0 0 r
F P
0 g
s
.3369 .48469 m .36251 .44486 L .37869 .44628 L p .697 0 0 r
F P
0 g
s
.38211 .40984 m .36251 .44486 L .34679 .44465 L p .702 0 0 r
F P
0 g
s
.3248 .49047 m .33413 .44569 L .34679 .44465 L p .855 0 0 r
F P
0 g
s
.3369 .48469 m .3277 .48744 L p .34352 .46809 L .774 0 0 r
F P
0 g
s
.3277 .48744 m .34679 .44465 L .36251 .44486 L p .774 0 0 r
F P
0 g
s
.3277 .48744 m .3248 .49047 L p .33497 .46927 L .855 0 0 r
F P
0 g
s
.33497 .46927 m .3354 .46924 L .3277 .48744 L .855 0 0 r
F
.34109 .45742 m .3277 .48744 L p .3385 .46191 L F P
0 g
s
.3248 .49047 m .3288 .49326 L p .33215 .46337 L .95 0 0 r
F P
0 g
s
.33413 .44569 m .3314 .45877 L p .33215 .46337 L .95 0 0 r
F P
0 g
s
.3314 .45877 m .3248 .49047 L p .33215 .46337 L .95 0 0 r
F P
0 g
s
.32671 .44781 m .33413 .44569 L p .33167 .46763 L .95 0 0 r
F P
0 g
s
.32761 .46747 m .32671 .44781 L p .33167 .46763 L .95 0 0 r
F P
0 g
s
.33082 .47523 m .32761 .46747 L .33167 .46763 L .95 0 0 r
F
.3288 .49326 m .32818 .47981 L p .33082 .47523 L F P
0 g
s
.32818 .47981 m .32761 .46747 L p .33082 .47523 L .95 0 0 r
F P
0 g
s
.3288 .49326 m .33914 .49531 L p .33129 .46533 L .869 0 0 r
F P
0 g
s
.32671 .44781 m .32731 .46087 L p .33129 .46533 L .869 0 0 r
F P
0 g
s
.32731 .46087 m .3288 .49326 L p .33129 .46533 L .869 0 0 r
F P
0 g
s
.32589 .45067 m .32671 .44781 L p .33237 .46946 L .869 0 0 r
F P
0 g
s
.33497 .46927 m .3385 .46191 L .3354 .46924 L .855 0 0 r
F
.34748 .46824 m .3369 .48469 L p .34352 .46809 L .774 0 0 r
F P
0 g
s
.33195 .45379 m .32589 .45067 L p .34313 .47859 L .036 0 0 r
F P
0 g
s
.35403 .49624 m .34725 .4832 L p .34313 .47859 L .036 0 0 r
F P
0 g
s
.34725 .4832 m .33195 .45379 L p .34313 .47859 L .036 0 0 r
F P
0 g
s
.33914 .49531 m .33149 .46952 L p .33237 .46946 L .869 0 0 r
F P
0 g
s
.33149 .46952 m .32589 .45067 L p .33237 .46946 L .869 0 0 r
F P
0 g
s
.34679 .44465 m .34109 .45742 L p .3385 .46191 L .855 0 0 r
F P
0 g
s
.36251 .44486 m .34748 .46824 L p .34352 .46809 L .774 0 0 r
F P
0 g
s
.40679 .51996 m .41209 .54268 L .40721 .53734 L p 0 0 0 r
F P
0 g
s
.40178 .52472 m .41045 .55008 L .41082 .54839 L p .997 0 0 r
F P
0 g
s
.41082 .54839 m .41209 .54268 L p .40897 .53725 L .997 0 0 r
F P
0 g
s
.40721 .53734 m .405 .51598 L .40569 .51751 L p 0 0 0 r
F P
0 g
s
.40721 .53734 m .40569 .51751 L .40691 .52484 L 0 0 0 r
F
.40897 .53725 m .40178 .52472 L .41082 .54839 L .997 0 0 r
F
.40569 .51751 m .40679 .51996 L p .40691 .52484 L 0 0 0 r
F P
0 g
s
.54475 .56434 m .56545 .57146 L .55367 .57864 L p 0 0 .188 r
F P
0 g
s
.55367 .57864 m .53266 .57291 L .54475 .56434 L p 0 0 .188 r
F P
0 g
s
.70641 .56714 m .69494 .57934 L .69785 .5943 L p .388 0 0 r
F P
0 g
s
.69494 .57934 m .70641 .56714 L .70572 .55742 L p .123 0 0 r
F P
0 g
s
.40552 .41914 m .39263 .44866 L .37869 .44628 L p .446 0 0 r
F P
0 g
s
.37869 .44628 m .39605 .41424 L .40552 .41914 L p .446 0 0 r
F P
0 g
s
.4143 .58035 m .42853 .59373 L .43022 .57806 L p .934 0 0 r
F P
0 g
s
.43022 .57806 m .41945 .56783 L .4143 .58035 L p .934 0 0 r
F P
0 g
s
.575 .52675 m .55695 .51986 L .55068 .50716 L p 0 0 .842 r
F P
0 g
s
.55695 .51986 m .575 .52675 L .57706 .54309 L p 0 0 .718 r
F P
0 g
s
.21849 .50657 m .23962 .49843 L .23309 .50996 L p 0 0 .819 r
F P
0 g
s
.23309 .50996 m .21176 .5203 L .21849 .50657 L p 0 0 .819 r
F P
0 g
s
.25432 .56435 m .23253 .57347 L .22124 .56559 L p 0 0 .223 r
F P
0 g
s
.23253 .57347 m .25432 .56435 L .26688 .56827 L p 0 0 0 r
F P
0 g
s
.74355 .36702 m .7559 .40099 L .74022 .40463 L p .741 0 0 r
F P
0 g
s
.7559 .40099 m .74355 .36702 L .75939 .36303 L p .65 0 0 r
F P
0 g
s
.75939 .36303 m .77194 .39935 L .7559 .40099 L p .65 0 0 r
F P
0 g
s
.73044 .6259 m .7536 .60193 L .73897 .59749 L p .671 0 0 r
F P
0 g
s
.73897 .59749 m .71733 .61986 L .73044 .6259 L p .671 0 0 r
F P
0 g
s
.62744 .47931 m .63977 .49433 L .63184 .48719 L p 0 0 0 r
F P
0 g
s
.63184 .48719 m .61721 .48059 L .62744 .47931 L p 0 0 0 r
F P
0 g
s
.78809 .5576 m .79243 .51966 L .79743 .52375 L p 0 0 0 r
F P
0 g
s
.79743 .52375 m .78877 .56341 L .78809 .5576 L p 0 0 0 r
F P
0 g
s
.7923 .47989 m .79743 .52375 L .79243 .51966 L p .051 0 0 r
F P
0 g
s
.41045 .55008 m .41945 .56783 L .41848 .55687 L p .976 0 0 r
F P
0 g
s
.55931 .45718 m .54581 .45898 L .54085 .46532 L p 0 .902 0 r
F P
0 g
s
.72903 .37413 m .74022 .40463 L .72751 .40965 L p .832 0 0 r
F P
0 g
s
.74022 .40463 m .72903 .37413 L .74355 .36702 L p .741 0 0 r
F P
0 g
s
.78575 .43892 m .79998 .48268 L .7923 .47989 L p .246 0 0 r
F P
0 g
s
.79998 .48268 m .78575 .43892 L .79648 .44082 L p 0 0 0 r
F P
0 g
s
.78575 .43892 m .77194 .39935 L .78567 .39998 L p .413 0 0 r
F P
0 g
s
.78567 .39998 m .79648 .44082 L .78575 .43892 L p .413 0 0 r
F P
0 g
s
.79743 .52375 m .7923 .47989 L .79998 .48268 L p .895 0 0 r
F P
0 g
s
.54022 .49947 m .55068 .50716 L p .52772 .49669 L 0 0 0 r
F P
0 g
s
.51263 .48981 m .54022 .49947 L p .52772 .49669 L 0 0 0 r
F P
0 g
s
.56785 .5126 m .55068 .50716 L .54022 .49947 L p 0 0 .911 r
F P
0 g
s
.54022 .49947 m .55681 .50306 L .56785 .5126 L p 0 0 .911 r
F P
0 g
s
.54022 .49947 m .52733 .49818 L p .55496 .50275 L 0 0 0 r
F P
0 g
s
.55681 .50306 m .54022 .49947 L p .55496 .50275 L 0 0 0 r
F P
0 g
s
.54022 .49947 m .51263 .48981 L .49944 .49386 L p 0 0 .301 r
F P
0 g
s
.49693 .40179 m .52667 .3932 L .51899 .39022 L p 0 0 0 r
F P
0 g
s
.51899 .39022 m .48691 .40205 L .49693 .40179 L p 0 0 0 r
F P
0 g
s
.45262 .42224 m .48691 .40205 L .47974 .40852 L p 0 .93 0 r
F P
0 g
s
.47974 .40852 m .44844 .43399 L .45262 .42224 L p 0 .93 0 r
F P
0 g
s
.34679 .44465 m .36606 .40668 L .38211 .40984 L p .702 0 0 r
F P
0 g
s
.36606 .40668 m .34679 .44465 L .33413 .44569 L p .782 0 0 r
F P
0 g
s
.4103 .61991 m .39 .60135 L .37517 .60605 L p .682 0 0 r
F P
0 g
s
.39 .60135 m .4103 .61991 L .42148 .60856 L p .797 0 0 r
F P
0 g
s
.77365 .59687 m .78877 .56341 L .78292 .56706 L p .923 0 0 r
F P
0 g
s
.78292 .56706 m .76588 .60169 L .77365 .59687 L p .923 0 0 r
F P
0 g
s
.26241 .49265 m .29613 .4911 L .28385 .48378 L p 0 0 0 r
F P
0 g
s
.61943 .39535 m .65239 .41031 L .65793 .41843 L p 0 .341 0 r
F P
0 g
s
.68742 .45048 m .65793 .41843 L .65239 .41031 L p 0 .119 0 r
F P
0 g
s
.65793 .41843 m .68742 .45048 L .68303 .46562 L p 0 .301 0 r
F P
0 g
s
.5829 .39044 m .61168 .3972 L .61943 .39535 L p 0 .103 0 r
F P
0 g
s
.65239 .41031 m .61943 .39535 L .61168 .3972 L p 0 .165 0 r
F P
0 g
s
.40387 .59224 m .42148 .60856 L .42853 .59373 L p .874 0 0 r
F P
0 g
s
.42853 .59373 m .4143 .58035 L .40387 .59224 L p .874 0 0 r
F P
0 g
s
.77194 .39935 m .75939 .36303 L .77393 .36283 L p .528 0 0 r
F P
0 g
s
.77393 .36283 m .78567 .39998 L .77194 .39935 L p .528 0 0 r
F P
0 g
s
.59056 .46812 m .60956 .47564 L .60587 .46522 L p 0 .568 0 r
F P
0 g
s
.60587 .46522 m .58639 .46093 L .59056 .46812 L p 0 .568 0 r
F P
0 g
s
.3956 .37817 m .38211 .40984 L .36606 .40668 L p .663 0 0 r
F P
0 g
s
.36606 .40668 m .38061 .37178 L .3956 .37817 L p .663 0 0 r
F P
0 g
s
.38061 .37178 m .36606 .40668 L .35055 .40527 L p .746 0 0 r
F P
0 g
s
.33413 .44569 m .35055 .40527 L .36606 .40668 L p .782 0 0 r
F P
0 g
s
.53266 .57291 m .55367 .57864 L .54037 .57913 L p 0 0 0 r
F P
0 g
s
.54037 .57913 m .51945 .57527 L .53266 .57291 L p 0 0 0 r
F P
0 g
s
.71307 .57837 m .69785 .5943 L .70578 .60866 L p .511 0 0 r
F P
0 g
s
.69785 .5943 m .71307 .57837 L .70641 .56714 L p .388 0 0 r
F P
0 g
s
.71991 .41524 m .72794 .44377 L .72208 .44707 L p .984 0 0 r
F P
0 g
s
.72794 .44377 m .71991 .41524 L .72751 .40965 L p .863 0 0 r
F P
0 g
s
.33775 .56877 m .35286 .59993 L .36202 .60553 L p .071 0 0 r
F P
0 g
s
.40745 .38615 m .39605 .41424 L .38211 .40984 L p .553 0 0 r
F P
0 g
s
.38211 .40984 m .3956 .37817 L .40745 .38615 L p .553 0 0 r
F P
0 g
s
.42148 .60856 m .40387 .59224 L .39 .60135 L p .797 0 0 r
F P
0 g
s
.40887 .42373 m .402 .45162 L .39263 .44866 L p .01 0 0 r
F P
0 g
s
.40552 .41914 m .40887 .42373 L p .40079 .43613 L .01 0 0 r
F P
0 g
s
.40887 .42373 m .4054 .4272 L p .40373 .43919 L .9 0 0 r
F P
0 g
s
.402 .45162 m .40556 .43717 L p .40373 .43919 L .9 0 0 r
F P
0 g
s
.40556 .43717 m .40887 .42373 L p .40373 .43919 L .9 0 0 r
F P
0 g
s
.40514 .45467 m .402 .45162 L p .40359 .44022 L .9 0 0 r
F P
0 g
s
.40139 .45728 m .40514 .45467 L p .40066 .44259 L .918 0 0 r
F P
0 g
s
.39263 .44866 m .39915 .43373 L p .40079 .43613 L .01 0 0 r
F P
0 g
s
.39561 .42894 m .39881 .44464 L p .40066 .44259 L .918 0 0 r
F P
0 g
s
.39881 .44464 m .40139 .45728 L p .40066 .44259 L .918 0 0 r
F P
0 g
s
.4054 .4272 m .40525 .44265 L p .40359 .44022 L .9 0 0 r
F P
0 g
s
.40525 .44265 m .40514 .45467 L p .40359 .44022 L .9 0 0 r
F P
0 g
s
.39915 .43373 m .40552 .41914 L p .40079 .43613 L .01 0 0 r
F P
0 g
s
.72298 .45026 m .72208 .44707 L p .7223 .44965 L .921 0 0 r
F P
0 g
s
.7247 .47879 m .72359 .46047 L p .7223 .44965 L .921 0 0 r
F P
0 g
s
.72359 .46047 m .72298 .45026 L p .7223 .44965 L .921 0 0 r
F P
0 g
s
.72187 .48141 m .72592 .4843 L p .72621 .48232 L 0 0 0 r
F P
0 g
s
.73059 .45277 m .72491 .47144 L p .72621 .48232 L 0 0 0 r
F P
0 g
s
.72298 .45026 m .7247 .47879 L .72187 .48141 L p .959 0 0 r
F P
0 g
s
.72491 .47144 m .72187 .48141 L p .72621 .48232 L 0 0 0 r
F P
0 g
s
.31468 .53338 m .29091 .52725 L .29181 .54318 L p 0 0 .695 r
F P
0 g
s
.29091 .52725 m .31468 .53338 L .31305 .51832 L p 0 0 .595 r
F P
0 g
s
.71733 .61986 m .73897 .59749 L .7246 .58913 L p .595 0 0 r
F P
0 g
s
.7246 .58913 m .70578 .60866 L .71733 .61986 L p .595 0 0 r
F P
0 g
s
.55068 .50716 m .56785 .5126 L .575 .52675 L p 0 0 .842 r
F P
0 g
s
.47586 .54723 m .49465 .5461 L .49628 .52985 L p 0 0 .78 r
F P
0 g
s
.49628 .52985 m .47439 .53295 L .47586 .54723 L p 0 0 .78 r
F P
0 g
s
.22898 .49769 m .24997 .49228 L .23962 .49843 L p 0 0 .912 r
F P
0 g
s
.23962 .49843 m .21849 .50657 L .22898 .49769 L p 0 0 .912 r
F P
0 g
s
.43581 .48816 m .44937 .44864 L .45523 .46371 L p 0 .461 0 r
F P
0 g
s
.45523 .46371 m .44783 .49828 L .43581 .48816 L p 0 .461 0 r
F P
0 g
s
.44783 .49828 m .45523 .46371 L .46502 .47666 L p 0 .259 0 r
F P
0 g
s
.26688 .56827 m .24523 .57477 L .23253 .57347 L p 0 0 0 r
F P
0 g
s
.24523 .57477 m .26688 .56827 L .27859 .56561 L p 0 0 .878 r
F P
0 g
s
.70578 .60866 m .7246 .58913 L .71307 .57837 L p .511 0 0 r
F P
0 g
s
.5347 .40234 m .55731 .3996 L .55477 .38901 L p 0 .356 0 r
F P
0 g
s
.55477 .38901 m .52667 .3932 L .5347 .40234 L p 0 .356 0 r
F P
0 g
s
.5829 .39044 m .55477 .38901 L .55731 .3996 L p 0 .371 0 r
F P
0 g
s
.71826 .38318 m .72751 .40965 L .71991 .41524 L p .942 0 0 r
F P
0 g
s
.72751 .40965 m .71826 .38318 L .72903 .37413 L p .832 0 0 r
F P
0 g
s
.46502 .47666 m .4611 .50274 L .44783 .49828 L p 0 .259 0 r
F P
0 g
s
.4611 .50274 m .46502 .47666 L .47711 .48534 L p 0 0 0 r
F P
0 g
s
.54581 .45898 m .55931 .45718 L .5604 .4461 L p 0 .788 0 r
F P
0 g
s
.55931 .45718 m .57091 .45772 L .57069 .44647 L p 0 .707 0 r
F P
0 g
s
.57069 .44647 m .5604 .4461 L .55931 .45718 L p 0 .707 0 r
F P
0 g
s
.70641 .56714 m .71669 .54703 L .71318 .54056 L p 0 0 0 r
F P
0 g
s
.71318 .54056 m .72049 .5158 L .71932 .5121 L p .924 0 0 r
F P
0 g
s
.71932 .5121 m .72592 .4843 L .72187 .48141 L p .987 0 0 r
F P
0 g
s
.55731 .3996 m .57887 .40058 L .5829 .39044 L p 0 .371 0 r
F P
0 g
s
.61168 .3972 m .5829 .39044 L .57887 .40058 L p 0 .36 0 r
F P
0 g
s
.42702 .47411 m .44844 .43399 L .44937 .44864 L p 0 .593 0 r
F P
0 g
s
.44937 .44864 m .43581 .48816 L .42702 .47411 L p 0 .593 0 r
F P
0 g
s
.27477 .49954 m .30648 .50333 L .29613 .4911 L p 0 0 .302 r
F P
0 g
s
.29613 .4911 m .26241 .49265 L .27477 .49954 L p 0 0 .302 r
F P
0 g
s
.57091 .45772 m .58639 .46093 L .58571 .44868 L p 0 .612 0 r
F P
0 g
s
.58571 .44868 m .57069 .44647 L .57091 .45772 L p 0 .612 0 r
F P
0 g
s
.52733 .49818 m .51422 .50357 L p .53945 .50034 L 0 0 .073 r
F P
0 g
s
.52733 .49818 m .54022 .49947 L p .5154 .49605 L 0 0 .301 r
F P
0 g
s
.54372 .49979 m .55681 .50306 L p .54197 .5006 L 0 0 0 r
F P
0 g
s
.52733 .49818 m .54372 .49979 L p .54197 .5006 L 0 0 0 r
F P
0 g
s
.54372 .49979 m .52733 .49818 L p .53945 .50034 L 0 0 .073 r
F P
0 g
s
.53082 .50342 m .54372 .49979 L p .52797 .50181 L 0 0 .073 r
F P
0 g
s
.51422 .50357 m .53082 .50342 L p .52797 .50181 L 0 0 .073 r
F P
0 g
s
.53082 .50342 m .52035 .51339 L p .53295 .50462 L 0 0 0 r
F P
0 g
s
.54372 .49979 m .53082 .50342 L p .54511 .49914 L 0 0 .732 r
F P
0 g
s
.53982 .49984 m .55212 .49705 L p .54009 .50064 L 0 0 .732 r
F P
0 g
s
.53082 .50342 m .53982 .49984 L p .54009 .50064 L 0 0 .732 r
F P
0 g
s
.53982 .49984 m .53082 .50342 L p .53295 .50462 L 0 0 0 r
F P
0 g
s
.5312 .50852 m .53982 .49984 L p .52771 .50827 L 0 0 0 r
F P
0 g
s
.52035 .51339 m .5312 .50852 L p .52771 .50827 L 0 0 0 r
F P
0 g
s
.5312 .50852 m .52781 .52164 L p .53416 .50872 L 0 0 .235 r
F P
0 g
s
.53982 .49984 m .5312 .50852 L p .53957 .4991 L 0 0 .616 r
F P
0 g
s
.53896 .49895 m .54543 .49251 L p .53573 .50342 L 0 0 .616 r
F P
0 g
s
.5312 .50852 m .53896 .49895 L p .53573 .50342 L 0 0 .616 r
F P
0 g
s
.54543 .49251 m .53982 .49984 L p .53957 .4991 L 0 0 .616 r
F P
0 g
s
.55691 .4904 m .55212 .49705 L .53982 .49984 L p 0 0 .918 r
F P
0 g
s
.53982 .49984 m .54543 .49251 L .55691 .4904 L p 0 0 .918 r
F P
0 g
s
.5501 .47678 m .54543 .49251 L .53896 .49895 L p 0 0 .663 r
F P
0 g
s
.53896 .49895 m .5387 .50868 L p .54252 .49402 L 0 0 .18 r
F P
0 g
s
.54585 .48122 m .5501 .47678 L p .54181 .49328 L 0 0 .663 r
F P
0 g
s
.53896 .49895 m .54585 .48122 L p .54181 .49328 L 0 0 .663 r
F P
0 g
s
.54543 .49251 m .5501 .47678 L .56031 .47494 L p 0 0 .813 r
F P
0 g
s
.5604 .4461 m .54775 .44719 L .54581 .45898 L p 0 .788 0 r
F P
0 g
s
.55221 .45176 m .5501 .47678 L .54585 .48122 L p 0 0 .524 r
F P
0 g
s
.54585 .48122 m .54842 .48753 L p .55101 .46584 L 0 0 .339 r
F P
0 g
s
.5501 .47678 m .55221 .45176 L .55895 .44953 L p 0 0 .74 r
F P
0 g
s
.55229 .45518 m .54585 .48122 L p .55101 .46584 L 0 0 .339 r
F P
0 g
s
.54585 .48122 m .53896 .49895 L p .54252 .49402 L 0 0 .18 r
F P
0 g
s
.53896 .49895 m .5312 .50852 L p .53416 .50872 L 0 0 .235 r
F P
0 g
s
.55212 .49705 m .54372 .49979 L p .54511 .49914 L 0 0 .732 r
F P
0 g
s
.49944 .49386 m .52733 .49818 L p .5154 .49605 L 0 0 .301 r
F P
0 g
s
.52733 .49818 m .49944 .49386 L .48732 .50366 L p 0 0 .471 r
F P
0 g
s
.35055 .40527 m .36496 .36805 L .38061 .37178 L p .746 0 0 r
F P
0 g
s
.36496 .36805 m .35055 .40527 L .33816 .40586 L p .83 0 0 r
F P
0 g
s
.35055 .40527 m .33413 .44569 L .32671 .44781 L p .868 0 0 r
F P
0 g
s
.32671 .44781 m .33816 .40586 L .35055 .40527 L p .868 0 0 r
F P
0 g
s
.31305 .51832 m .28487 .51179 L .29091 .52725 L p 0 0 .595 r
F P
0 g
s
.28487 .51179 m .31305 .51832 L .30648 .50333 L p 0 0 .479 r
F P
0 g
s
.71669 .54703 m .70641 .56714 L .71307 .57837 L p .372 0 0 r
F P
0 g
s
.41414 .39442 m .40552 .41914 L .39605 .41424 L p .342 0 0 r
F P
0 g
s
.39605 .41424 m .40745 .38615 L .41414 .39442 L p .342 0 0 r
F P
0 g
s
.69453 .50223 m .68303 .46562 L .68742 .45048 L p 0 .18 0 r
F P
0 g
s
.68303 .46562 m .69453 .50223 L .68197 .50945 L p 0 .15 0 r
F P
0 g
s
.68197 .50945 m .67443 .47961 L .68303 .46562 L p 0 .15 0 r
F P
0 g
s
.41209 .54268 m .40679 .51996 L .40178 .52472 L p .997 0 0 r
F P
0 g
s
.67443 .47961 m .68197 .50945 L .66952 .5104 L p 0 .058 0 r
F P
0 g
s
.66952 .5104 m .66306 .49011 L .67443 .47961 L p 0 .058 0 r
F P
0 g
s
.61721 .48059 m .63184 .48719 L .62846 .47508 L p 0 .62 0 r
F P
0 g
s
.62846 .47508 m .60956 .47564 L .61721 .48059 L p 0 .62 0 r
F P
0 g
s
.30648 .50333 m .27477 .49954 L .28487 .51179 L p 0 0 .479 r
F P
0 g
s
.40246 .55828 m .4143 .58035 L .41945 .56783 L p .935 0 0 r
F P
0 g
s
.41945 .56783 m .41045 .55008 L .40246 .55828 L p .935 0 0 r
F P
0 g
s
.72855 .33961 m .74355 .36702 L .72903 .37413 L p .778 0 0 r
F P
0 g
s
.74355 .36702 m .72855 .33961 L .74356 .33344 L p .679 0 0 r
F P
0 g
s
.74356 .33344 m .75939 .36303 L .74355 .36702 L p .679 0 0 r
F P
0 g
s
.47439 .53295 m .49628 .52985 L .50322 .51478 L p 0 0 .678 r
F P
0 g
s
.50322 .51478 m .47844 .51751 L .47439 .53295 L p 0 0 .678 r
F P
0 g
s
.51945 .57527 m .54037 .57913 L .52784 .57276 L p 0 0 .881 r
F P
0 g
s
.52784 .57276 m .50739 .57096 L .51945 .57527 L p 0 0 .881 r
F P
0 g
s
.79648 .44082 m .78567 .39998 L .79477 .4028 L p 0 0 0 r
F P
0 g
s
.78567 .39998 m .77393 .36283 L .78468 .36651 L p .289 0 0 r
F P
0 g
s
.78468 .36651 m .79477 .4028 L .78567 .39998 L p .289 0 0 r
F P
0 g
s
.80111 .48552 m .79648 .44082 L .80128 .44367 L p .989 0 0 r
F P
0 g
s
.80128 .44367 m .79921 .447 L p .79895 .4497 L .864 0 0 r
F P
0 g
s
.79648 .44082 m .79943 .46931 L p .79966 .47883 L 0 0 0 r
F P
0 g
s
.79535 .48792 m .79919 .4593 L p .79895 .4497 L .864 0 0 r
F P
0 g
s
.79919 .4593 m .80128 .44367 L p .79895 .4497 L .864 0 0 r
F P
0 g
s
.80111 .48552 m .79998 .48268 L p .79966 .47883 L 0 0 0 r
F P
0 g
s
.79943 .46931 m .80111 .48552 L p .79966 .47883 L 0 0 0 r
F P
0 g
s
.79477 .4028 m .80128 .44367 L .79648 .44082 L p 0 0 0 r
F P
0 g
s
.54581 .45898 m .52848 .46299 L .52294 .47197 L p 0 .946 0 r
F P
0 g
s
.52848 .46299 m .54581 .45898 L .54775 .44719 L p 0 .812 0 r
F P
0 g
s
.48691 .40205 m .45262 .42224 L .46128 .41542 L p 0 .937 0 r
F P
0 g
s
.46128 .41542 m .49693 .40179 L .48691 .40205 L p 0 .937 0 r
F P
0 g
s
.78877 .56341 m .79743 .52375 L .79568 .52703 L p .975 0 0 r
F P
0 g
s
.79568 .52703 m .78292 .56706 L .78877 .56341 L p .975 0 0 r
F P
0 g
s
.76588 .60169 m .78292 .56706 L .77146 .56787 L p .825 0 0 r
F P
0 g
s
.77146 .56787 m .7536 .60193 L .76588 .60169 L p .825 0 0 r
F P
0 g
s
.50809 .40789 m .5347 .40234 L .52667 .3932 L p 0 .27 0 r
F P
0 g
s
.52667 .3932 m .49693 .40179 L .50809 .40789 L p 0 .27 0 r
F P
0 g
s
.7158 .3503 m .72903 .37413 L .71826 .38318 L p .87 0 0 r
F P
0 g
s
.72903 .37413 m .7158 .3503 L .72855 .33961 L p .778 0 0 r
F P
0 g
s
.47711 .48534 m .47335 .50072 L .4611 .50274 L p 0 0 0 r
F P
0 g
s
.47335 .50072 m .47711 .48534 L .48947 .48823 L p 0 .867 0 r
F P
0 g
s
.79998 .48268 m .79568 .52703 L .79743 .52375 L p .895 0 0 r
F P
0 g
s
.51422 .50357 m .48732 .50366 L .47844 .51751 L p 0 0 .582 r
F P
0 g
s
.48732 .50366 m .51422 .50357 L .52733 .49818 L p 0 0 .471 r
F P
0 g
s
.61168 .3972 m .64338 .40806 L .65239 .41031 L p 0 .165 0 r
F P
0 g
s
.65239 .41031 m .68684 .43673 L .68742 .45048 L p 0 .119 0 r
F P
0 g
s
.68684 .43673 m .65239 .41031 L .64338 .40806 L p 0 .262 0 r
F P
0 g
s
.41027 .35204 m .3956 .37817 L .38061 .37178 L p .626 0 0 r
F P
0 g
s
.38061 .37178 m .39708 .34196 L .41027 .35204 L p .626 0 0 r
F P
0 g
s
.39708 .34196 m .38061 .37178 L .36496 .36805 L p .71 0 0 r
F P
0 g
s
.79568 .52703 m .79998 .48268 L .80111 .48552 L p .942 0 0 r
F P
0 g
s
.70508 .49002 m .68742 .45048 L .68684 .43673 L p 0 .143 0 r
F P
0 g
s
.68742 .45048 m .70508 .49002 L .69453 .50223 L p 0 .18 0 r
F P
0 g
s
.24149 .49523 m .26241 .49265 L .24997 .49228 L p 0 0 0 r
F P
0 g
s
.24267 .49417 m .22898 .49769 L .24149 .49523 L p 0 0 0 r
F P
0 g
s
.2301 .49335 m .24149 .49523 L .22898 .49769 L p 0 0 .982 r
F P
0 g
s
.24149 .49523 m .2301 .49335 L .2403 .4963 L p 0 0 0 r
F P
0 g
s
.2403 .4963 m .24195 .49678 L p .24149 .49523 L 0 0 0 r
F P
0 g
s
.24997 .49228 m .24267 .49417 L p .24149 .49523 L 0 0 0 r
F P
0 g
s
.36202 .60553 m .34534 .57295 L .33775 .56877 L p .071 0 0 r
F P
0 g
s
.34534 .57295 m .36202 .60553 L .37517 .60605 L p .496 0 0 r
F P
0 g
s
.66306 .49011 m .66952 .5104 L .65931 .50484 L p 0 0 0 r
F P
0 g
s
.65931 .50484 m .65082 .49533 L .66306 .49011 L p 0 0 0 r
F P
0 g
s
.42296 .45848 m .45262 .42224 L .44844 .43399 L p 0 .706 0 r
F P
0 g
s
.44844 .43399 m .42702 .47411 L .42296 .45848 L p 0 .706 0 r
F P
0 g
s
.27859 .56561 m .25713 .56919 L .24523 .57477 L p 0 0 .878 r
F P
0 g
s
.25713 .56919 m .27859 .56561 L .28743 .55675 L p 0 0 .883 r
F P
0 g
s
.403 .48584 m .40139 .45728 L .39129 .45899 L p .9 0 0 r
F P
0 g
s
.47844 .51751 m .50322 .51478 L .51422 .50357 L p 0 0 .582 r
F P
0 g
s
.57887 .40058 m .60295 .40523 L .61168 .3972 L p 0 .36 0 r
F P
0 g
s
.64338 .40806 m .61168 .3972 L .60295 .40523 L p 0 .35 0 r
F P
0 g
s
.41958 .36446 m .40745 .38615 L .3956 .37817 L p .517 0 0 r
F P
0 g
s
.3956 .37817 m .41027 .35204 L .41958 .36446 L p .517 0 0 r
F P
0 g
s
.75939 .36303 m .74356 .33344 L .75832 .33288 L p .542 0 0 r
F P
0 g
s
.75832 .33288 m .77393 .36283 L .75939 .36303 L p .542 0 0 r
F P
0 g
s
.58639 .46093 m .60587 .46522 L .60683 .4511 L p 0 .543 0 r
F P
0 g
s
.60683 .4511 m .58571 .44868 L .58639 .46093 L p 0 .543 0 r
F P
0 g
s
.50683 .4755 m .52294 .47197 L .52848 .46299 L p 0 .866 0 r
F P
0 g
s
.1951 .51932 m .21176 .5203 L .20992 .53653 L p 0 0 .694 r
F P
0 g
s
.20992 .53653 m .19021 .53506 L .1951 .51932 L p 0 0 .694 r
F P
0 g
s
.19021 .53506 m .20992 .53653 L .21326 .55253 L p 0 0 .552 r
F P
0 g
s
.58906 .52288 m .575 .52675 L .56785 .5126 L p 0 0 .814 r
F P
0 g
s
.575 .52675 m .58906 .52288 L .59445 .53796 L p 0 0 .693 r
F P
0 g
s
.59445 .53796 m .57706 .54309 L .575 .52675 L p 0 0 .693 r
F P
0 g
s
.57706 .54309 m .59445 .53796 L .59424 .55238 L p 0 0 .531 r
F P
0 g
s
.59424 .55238 m .5737 .55889 L .57706 .54309 L p 0 0 .531 r
F P
0 g
s
.71881 .42045 m .71991 .41524 L p .72153 .43891 L .984 0 0 r
F P
0 g
s
.72208 .44707 m .71881 .42045 L p .72153 .43891 L .984 0 0 r
F P
0 g
s
.71307 .39266 m .71991 .41524 L .71881 .42045 L p .931 0 0 r
F P
0 g
s
.71991 .41524 m .71307 .39266 L .71826 .38318 L p .942 0 0 r
F P
0 g
s
.71881 .42045 m .72449 .42438 L p .71672 .40629 L 0 0 0 r
F P
0 g
s
.71443 .40097 m .71307 .39266 L p .71616 .40764 L .931 0 0 r
F P
0 g
s
.71881 .42045 m .71443 .40097 L p .71616 .40764 L .931 0 0 r
F P
0 g
s
.7049 .37765 m .71307 .39266 L .71443 .40097 L p .889 0 0 r
F P
0 g
s
.71443 .40097 m .72221 .40666 L p .71287 .39497 L 0 0 0 r
F P
0 g
s
.70867 .3897 m .7049 .37765 L p .71022 .39067 L .889 0 0 r
F P
0 g
s
.71443 .40097 m .70867 .3897 L p .71022 .39067 L .889 0 0 r
F P
0 g
s
.70867 .3897 m .71443 .40097 L p .71287 .39497 L 0 0 0 r
F P
0 g
s
.71818 .39779 m .70867 .3897 L p .71641 .3994 L 0 0 0 r
F P
0 g
s
.72221 .40666 m .71818 .39779 L p .71641 .3994 L 0 0 0 r
F P
0 g
s
.71818 .39779 m .73184 .40045 L p .71998 .39615 L .217 0 0 r
F P
0 g
s
.70867 .3897 m .71818 .39779 L p .7069 .38861 L .498 0 0 r
F P
0 g
s
.71139 .39303 m .70032 .38325 L p .71177 .39257 L .498 0 0 r
F P
0 g
s
.71818 .39779 m .71139 .39303 L p .71177 .39257 L .498 0 0 r
F P
0 g
s
.71139 .39303 m .71818 .39779 L p .71998 .39615 L .217 0 0 r
F P
0 g
s
.72559 .39654 m .71139 .39303 L p .72504 .39798 L .217 0 0 r
F P
0 g
s
.73184 .40045 m .72559 .39654 L p .72504 .39798 L .217 0 0 r
F P
0 g
s
.72559 .39654 m .74049 .39309 L p .7263 .39469 L .407 0 0 r
F P
0 g
s
.71139 .39303 m .72559 .39654 L p .71087 .39366 L .275 0 0 r
F P
0 g
s
.7122 .39627 m .69784 .39111 L p .71945 .39534 L .275 0 0 r
F P
0 g
s
.72559 .39654 m .7122 .39627 L p .71945 .39534 L .275 0 0 r
F P
0 g
s
.7122 .39627 m .72559 .39654 L p .7263 .39469 L .407 0 0 r
F P
0 g
s
.72689 .39442 m .7122 .39627 L p .73596 .3936 L .407 0 0 r
F P
0 g
s
.74049 .39309 m .72689 .39442 L p .73596 .3936 L .407 0 0 r
F P
0 g
s
.72689 .39442 m .73937 .3858 L p .7184 .39519 L .681 0 0 r
F P
0 g
s
.7122 .39627 m .72689 .39442 L p .70442 .39727 L 0 0 0 r
F P
0 g
s
.72689 .39442 m .70215 .40246 L .68736 .39943 L p 0 0 0 r
F P
0 g
s
.70215 .40246 m .72148 .39618 L p .7184 .39519 L .681 0 0 r
F P
0 g
s
.72148 .39618 m .72267 .3948 L .7184 .39519 L .681 0 0 r
F
.72148 .39618 m .72689 .39442 L p .72267 .3948 L F P
0 g
s
.73937 .3858 m .71585 .39858 L .70215 .40246 L p .681 0 0 r
F P
0 g
s
.7261 .38837 m .71585 .39858 L p .72903 .38743 L .877 0 0 r
F P
0 g
s
.74744 .37186 m .73248 .38344 L p .72903 .38743 L .877 0 0 r
F P
0 g
s
.73248 .38344 m .7261 .38837 L p .72903 .38743 L .877 0 0 r
F P
0 g
s
.68736 .39943 m .7122 .39627 L p .70442 .39727 L 0 0 0 r
F P
0 g
s
.69784 .39111 m .71139 .39303 L p .71087 .39366 L .275 0 0 r
F P
0 g
s
.70032 .38325 m .70867 .3897 L p .7069 .38861 L .498 0 0 r
F P
0 g
s
.71443 .40097 m .71881 .42045 L p .71672 .40629 L 0 0 0 r
F P
0 g
s
.71881 .42045 m .72208 .44707 L .72298 .45026 L p 0 0 0 r
F P
0 g
s
.20452 .50571 m .21849 .50657 L .21176 .5203 L p 0 0 .812 r
F P
0 g
s
.21176 .5203 m .1951 .51932 L .20452 .50571 L p 0 0 .812 r
F P
0 g
s
.36496 .36805 m .38223 .3359 L .39708 .34196 L p .71 0 0 r
F P
0 g
s
.38223 .3359 m .36496 .36805 L .35126 .36762 L p .794 0 0 r
F P
0 g
s
.33816 .40586 m .35126 .36762 L .36496 .36805 L p .83 0 0 r
F P
0 g
s
.54166 .41608 m .55931 .41446 L .55731 .3996 L p 0 .515 0 r
F P
0 g
s
.55731 .3996 m .5347 .40234 L .54166 .41608 L p 0 .515 0 r
F P
0 g
s
.57887 .40058 m .55731 .3996 L .55931 .41446 L p 0 .496 0 r
F P
0 g
s
.21326 .55253 m .19068 .5503 L .19021 .53506 L p 0 0 .552 r
F P
0 g
s
.19068 .5503 m .21326 .55253 L .22124 .56559 L p 0 0 .329 r
F P
0 g
s
.54775 .44719 m .5604 .4461 L .56041 .43096 L p 0 .693 0 r
F P
0 g
s
.5604 .4461 m .57069 .44647 L .57216 .43136 L p 0 .634 0 r
F P
0 g
s
.57216 .43136 m .56041 .43096 L .5604 .4461 L p 0 .634 0 r
F P
0 g
s
.72187 .48141 m .73059 .45277 L .72298 .45026 L p .959 0 0 r
F P
0 g
s
.55931 .41446 m .57506 .41508 L .57887 .40058 L p 0 .496 0 r
F P
0 g
s
.60295 .40523 m .57887 .40058 L .57506 .41508 L p 0 .454 0 r
F P
0 g
s
.57069 .44647 m .58571 .44868 L .58868 .4335 L p 0 .559 0 r
F P
0 g
s
.58868 .4335 m .57216 .43136 L .57069 .44647 L p 0 .559 0 r
F P
0 g
s
.32835 .53045 m .33775 .56877 L .34534 .57295 L p 0 0 0 r
F P
0 g
s
.35126 .36762 m .33816 .40586 L .33103 .40838 L p .932 0 0 r
F P
0 g
s
.32589 .45067 m .33103 .40838 L .33816 .40586 L p .974 0 0 r
F P
0 g
s
.32671 .44781 m .32589 .45067 L p .33204 .42822 L .974 0 0 r
F P
0 g
s
.33235 .42712 m .32671 .44781 L p .33204 .42822 L .974 0 0 r
F P
0 g
s
.32589 .45067 m .33195 .45379 L p .33149 .43094 L .806 0 0 r
F P
0 g
s
.33103 .40838 m .32846 .42953 L p .33149 .43094 L .806 0 0 r
F P
0 g
s
.32846 .42953 m .32589 .45067 L p .33149 .43094 L .806 0 0 r
F P
0 g
s
.33048 .41243 m .33103 .40838 L p .33147 .43018 L .806 0 0 r
F P
0 g
s
.33816 .40586 m .33235 .42712 L p .33204 .42822 L .974 0 0 r
F P
0 g
s
.33673 .41733 m .33048 .41243 L p .33687 .43344 L .136 0 0 r
F P
0 g
s
.34393 .45661 m .33994 .43486 L p .33687 .43344 L .136 0 0 r
F P
0 g
s
.33994 .43486 m .33673 .41733 L p .33687 .43344 L .136 0 0 r
F P
0 g
s
.33195 .45379 m .33115 .4313 L p .33147 .43018 L .806 0 0 r
F P
0 g
s
.33115 .4313 m .33048 .41243 L p .33147 .43018 L .806 0 0 r
F P
0 g
s
.72629 .55421 m .71307 .57837 L .7246 .58913 L p .54 0 0 r
F P
0 g
s
.71307 .57837 m .72629 .55421 L .71669 .54703 L p .372 0 0 r
F P
0 g
s
.579 .50967 m .56785 .5126 L .55681 .50306 L p 0 0 .924 r
F P
0 g
s
.56785 .5126 m .579 .50967 L .58906 .52288 L p 0 0 .814 r
F P
0 g
s
.5737 .55889 m .59424 .55238 L .5884 .56368 L p 0 0 .244 r
F P
0 g
s
.5884 .56368 m .56545 .57146 L .5737 .55889 L p 0 0 .244 r
F P
0 g
s
.70744 .36373 m .71826 .38318 L .71307 .39266 L p .963 0 0 r
F P
0 g
s
.71826 .38318 m .70744 .36373 L .7158 .3503 L p .87 0 0 r
F P
0 g
s
.38944 .56583 m .40387 .59224 L .4143 .58035 L p .846 0 0 r
F P
0 g
s
.4143 .58035 m .40246 .55828 L .38944 .56583 L p .846 0 0 r
F P
0 g
s
.41446 .40156 m .40887 .42373 L .40552 .41914 L p 0 0 0 r
F P
0 g
s
.40552 .41914 m .41414 .39442 L .41427 .39737 L p 0 0 0 r
F P
0 g
s
.40552 .41914 m .41427 .39737 L .41201 .40636 L 0 0 0 r
F
.40824 .40633 m .4054 .4272 L .40725 .42534 L p .957 0 0 r
F P
0 g
s
.40725 .42534 m .40887 .42373 L p .40872 .41952 L .957 0 0 r
F P
0 g
s
.40872 .41952 m .40824 .40633 L .40725 .42534 L .957 0 0 r
F
.41427 .39737 m .41446 .40156 L p .41201 .40636 L 0 0 0 r
F P
0 g
s
.50739 .57096 m .52784 .57276 L .5183 .5606 L p 0 0 .912 r
F P
0 g
s
.5183 .5606 m .4986 .56066 L .50739 .57096 L p 0 0 .912 r
F P
0 g
s
.72827 .52008 m .71669 .54703 L .72629 .55421 L p .41 0 0 r
F P
0 g
s
.71669 .54703 m .72827 .52008 L .72049 .5158 L p 0 0 0 r
F P
0 g
s
.42344 .37716 m .41414 .39442 L .40745 .38615 L p .315 0 0 r
F P
0 g
s
.40745 .38615 m .41958 .36446 L .42344 .37716 L p .315 0 0 r
F P
0 g
s
.21688 .49653 m .22898 .49769 L .21849 .50657 L p 0 0 .927 r
F P
0 g
s
.21849 .50657 m .20452 .50571 L .21688 .49653 L p 0 0 .927 r
F P
0 g
s
.60956 .47564 m .62846 .47508 L .63028 .46008 L p 0 .595 0 r
F P
0 g
s
.63028 .46008 m .60587 .46522 L .60956 .47564 L p 0 .595 0 r
F P
0 g
s
.64338 .40806 m .68134 .42672 L .68684 .43673 L p 0 .262 0 r
F P
0 g
s
.71186 .47487 m .68684 .43673 L .68134 .42672 L p 0 0 0 r
F P
0 g
s
.68684 .43673 m .71186 .47487 L .70508 .49002 L p 0 .143 0 r
F P
0 g
s
.73626 .48696 m .72049 .5158 L .72827 .52008 L p .294 0 0 r
F P
0 g
s
.72049 .5158 m .73626 .48696 L .72592 .4843 L p 0 0 0 r
F P
0 g
s
.7536 .60193 m .77146 .56787 L .75637 .56566 L p .728 0 0 r
F P
0 g
s
.75637 .56566 m .73897 .59749 L .7536 .60193 L p .728 0 0 r
F P
0 g
s
.38652 .4943 m .40178 .52472 L .40679 .51996 L p .888 0 0 r
F P
0 g
s
.40679 .51996 m .39827 .49178 L .38652 .4943 L p .888 0 0 r
F P
0 g
s
.25386 .49967 m .27477 .49954 L .26241 .49265 L p 0 0 .054 r
F P
0 g
s
.26241 .49265 m .24149 .49523 L .25386 .49967 L p 0 0 .054 r
F P
0 g
s
.37517 .60605 m .35807 .57389 L .34534 .57295 L p .496 0 0 r
F P
0 g
s
.35807 .57389 m .37517 .60605 L .39 .60135 L p .661 0 0 r
F P
0 g
s
.65082 .49533 m .65931 .50484 L .65316 .49366 L p 0 .602 0 r
F P
0 g
s
.65316 .49366 m .63977 .49433 L .65082 .49533 L p 0 .602 0 r
F P
0 g
s
.52959 .44959 m .54775 .44719 L .54629 .43199 L p 0 .702 0 r
F P
0 g
s
.56041 .43096 m .54629 .43199 L .54775 .44719 L p 0 .693 0 r
F P
0 g
s
.54775 .44719 m .52959 .44959 L .52848 .46299 L p 0 .812 0 r
F P
0 g
s
.48947 .48823 m .48245 .49249 L .47335 .50072 L p 0 .867 0 r
F P
0 g
s
.48245 .49249 m .48947 .48823 L .49999 .48478 L p 0 .96 0 r
F P
0 g
s
.28743 .55675 m .26616 .55764 L .25713 .56919 L p 0 0 .883 r
F P
0 g
s
.26616 .55764 m .28743 .55675 L .29181 .54318 L p 0 0 .764 r
F P
0 g
s
.39073 .52943 m .40246 .55828 L .41045 .55008 L p .898 0 0 r
F P
0 g
s
.41045 .55008 m .40178 .52472 L .39073 .52943 L p .898 0 0 r
F P
0 g
s
.51841 .41934 m .54166 .41608 L .5347 .40234 L p 0 .479 0 r
F P
0 g
s
.5347 .40234 m .50809 .40789 L .51841 .41934 L p 0 .479 0 r
F P
0 g
s
.22124 .56559 m .19649 .56244 L .19068 .5503 L p 0 0 .329 r
F P
0 g
s
.19649 .56244 m .22124 .56559 L .23253 .57347 L p 0 0 0 r
F P
0 g
s
.77393 .36283 m .75832 .33288 L .7703 .33806 L p .273 0 0 r
F P
0 g
s
.7703 .33806 m .78468 .36651 L .77393 .36283 L p .273 0 0 r
F P
0 g
s
.54629 .43199 m .56041 .43096 L .55931 .41446 L p 0 .61 0 r
F P
0 g
s
.55931 .41446 m .54166 .41608 L .54629 .43199 L p 0 .61 0 r
F P
0 g
s
.57506 .41508 m .55931 .41446 L .56041 .43096 L p 0 .571 0 r
F P
0 g
s
.42431 .44391 m .46128 .41542 L .45262 .42224 L p 0 .824 0 r
F P
0 g
s
.45262 .42224 m .42296 .45848 L .42431 .44391 L p 0 .824 0 r
F P
0 g
s
.34534 .57295 m .33304 .53392 L .32835 .53045 L p 0 0 0 r
F P
0 g
s
.56041 .43096 m .57216 .43136 L .57506 .41508 L p 0 .571 0 r
F P
0 g
s
.5948 .41809 m .57506 .41508 L .57216 .43136 L p 0 .511 0 r
F P
0 g
s
.57506 .41508 m .5948 .41809 L .60295 .40523 L p 0 .454 0 r
F P
0 g
s
.47297 .41481 m .50809 .40789 L .49693 .40179 L p 0 0 0 r
F P
0 g
s
.49693 .40179 m .46128 .41542 L .47297 .41481 L p 0 0 0 r
F P
0 g
s
.57216 .43136 m .58868 .4335 L .5948 .41809 L p 0 .511 0 r
F P
0 g
s
.6324 .41216 m .60295 .40523 L .5948 .41809 L p 0 .42 0 r
F P
0 g
s
.60295 .40523 m .6324 .41216 L .64338 .40806 L p 0 .35 0 r
F P
0 g
s
.80128 .44367 m .79477 .4028 L .79759 .40738 L p .944 0 0 r
F P
0 g
s
.79477 .4028 m .78468 .36651 L .78976 .37348 L p 0 0 0 r
F P
0 g
s
.78976 .37348 m .79759 .40738 L .79477 .4028 L p 0 0 0 r
F P
0 g
s
.58571 .44868 m .60683 .4511 L .61233 .43577 L p 0 .502 0 r
F P
0 g
s
.61233 .43577 m .58868 .4335 L .58571 .44868 L p 0 .502 0 r
F P
0 g
s
.68134 .42672 m .64338 .40806 L .6324 .41216 L p 0 .388 0 r
F P
0 g
s
.74035 .56081 m .7246 .58913 L .73897 .59749 L p .641 0 0 r
F P
0 g
s
.7246 .58913 m .74035 .56081 L .72629 .55421 L p .54 0 0 r
F P
0 g
s
.3737 .57137 m .39 .60135 L .40387 .59224 L p .761 0 0 r
F P
0 g
s
.40387 .59224 m .38944 .56583 L .3737 .57137 L p .761 0 0 r
F P
0 g
s
.78292 .56706 m .79568 .52703 L .78736 .5289 L p .864 0 0 r
F P
0 g
s
.78736 .5289 m .77146 .56787 L .78292 .56706 L p .864 0 0 r
F P
0 g
s
.73897 .59749 m .75637 .56566 L .74035 .56081 L p .641 0 0 r
F P
0 g
s
.80128 .44367 m .79535 .48792 L .80111 .48552 L p .989 0 0 r
F P
0 g
s
.79759 .40738 m .79921 .447 L .80128 .44367 L p .944 0 0 r
F P
0 g
s
.56596 .50054 m .55681 .50306 L .54372 .49979 L p 0 0 1 r
F P
0 g
s
.55681 .50306 m .56596 .50054 L .579 .50967 L p 0 0 .924 r
F P
0 g
s
.39 .60135 m .3737 .57137 L .35807 .57389 L p .661 0 0 r
F P
0 g
s
.56545 .57146 m .5884 .56368 L .57786 .56987 L p 0 0 0 r
F P
0 g
s
.57786 .56987 m .55367 .57864 L .56545 .57146 L p 0 0 0 r
F P
0 g
s
.35126 .36762 m .36819 .33491 L .38223 .3359 L p .794 0 0 r
F P
0 g
s
.36819 .33491 m .35126 .36762 L .34184 .37059 L p .888 0 0 r
F P
0 g
s
.33103 .40838 m .34184 .37059 L .35126 .36762 L p .932 0 0 r
F P
0 g
s
.52848 .46299 m .50874 .46192 L .50683 .4755 L p 0 .866 0 r
F P
0 g
s
.50874 .46192 m .52848 .46299 L .52959 .44959 L p 0 .734 0 r
F P
0 g
s
.80111 .48552 m .78736 .5289 L .79568 .52703 L p .942 0 0 r
F P
0 g
s
.78736 .5289 m .80111 .48552 L .79535 .48792 L p .821 0 0 r
F P
0 g
s
.54629 .43199 m .52603 .43413 L .52959 .44959 L p 0 .702 0 r
F P
0 g
s
.52603 .43413 m .54629 .43199 L .54166 .41608 L p 0 .601 0 r
F P
0 g
s
.54166 .41608 m .51841 .41934 L .52603 .43413 L p 0 .601 0 r
F P
0 g
s
.22898 .49769 m .21688 .49653 L .2301 .49335 L p 0 0 .982 r
F P
0 g
s
.4986 .56066 m .5183 .5606 L .51346 .54475 L p 0 0 .791 r
F P
0 g
s
.51346 .54475 m .49465 .5461 L .4986 .56066 L p 0 0 .791 r
F P
0 g
s
.26392 .51032 m .28487 .51179 L .27477 .49954 L p 0 0 .364 r
F P
0 g
s
.27477 .49954 m .25386 .49967 L .26392 .51032 L p 0 0 .364 r
F P
0 g
s
.37653 .45947 m .39827 .49178 L .40406 .48878 L p .805 0 0 r
F P
0 g
s
.40406 .48878 m .39129 .45899 L .37653 .45947 L p .805 0 0 r
F P
0 g
s
.40514 .45467 m .4054 .4272 L .39561 .42894 L p .918 0 0 r
F P
0 g
s
.29181 .54318 m .27068 .54211 L .26616 .55764 L p 0 0 .764 r
F P
0 g
s
.27068 .54211 m .29181 .54318 L .29091 .52725 L p 0 0 .651 r
F P
0 g
s
.3248 .49047 m .33304 .53392 L .34385 .53571 L p 0 0 0 r
F P
0 g
s
.53082 .50342 m .51422 .50357 L .50322 .51478 L p 0 0 .37 r
F P
0 g
s
.6214 .42197 m .5948 .41809 L .58868 .4335 L p 0 .464 0 r
F P
0 g
s
.5948 .41809 m .6214 .42197 L .6324 .41216 L p 0 .42 0 r
F P
0 g
s
.58868 .4335 m .61233 .43577 L .6214 .42197 L p 0 .464 0 r
F P
0 g
s
.23253 .57347 m .20668 .56934 L .19649 .56244 L p 0 0 0 r
F P
0 g
s
.20668 .56934 m .23253 .57347 L .24523 .57477 L p 0 0 .861 r
F P
0 g
s
.71307 .39266 m .7049 .37765 L .70744 .36373 L p .963 0 0 r
F P
0 g
s
.72449 .42438 m .72298 .45026 L .73059 .45277 L p .22 0 0 r
F P
0 g
s
.72298 .45026 m .72449 .42438 L .71881 .42045 L p 0 0 0 r
F P
0 g
s
.33304 .53392 m .34534 .57295 L .35807 .57389 L p .426 0 0 r
F P
0 g
s
.43418 .35091 m .41958 .36446 L .41027 .35204 L p .467 0 0 r
F P
0 g
s
.41027 .35204 m .42801 .33536 L .43418 .35091 L p .467 0 0 r
F P
0 g
s
.42801 .33536 m .41027 .35204 L .39708 .34196 L p .553 0 0 r
F P
0 g
s
.39708 .34196 m .41722 .32282 L .42801 .33536 L p .553 0 0 r
F P
0 g
s
.41722 .32282 m .39708 .34196 L .38223 .3359 L p .626 0 0 r
F P
0 g
s
.58648 .49958 m .579 .50967 L .56596 .50054 L p 0 0 .848 r
F P
0 g
s
.579 .50967 m .58648 .49958 L .59963 .5093 L p 0 0 .781 r
F P
0 g
s
.59963 .5093 m .58906 .52288 L .579 .50967 L p 0 0 .781 r
F P
0 g
s
.40349 .48722 m .42296 .45848 L .42702 .47411 L p 0 .582 0 r
F P
0 g
s
.42702 .47411 m .41196 .50155 L .40349 .48722 L p 0 .582 0 r
F P
0 g
s
.41196 .50155 m .42702 .47411 L .43581 .48816 L p 0 .417 0 r
F P
0 g
s
.58906 .52288 m .59963 .5093 L .60864 .52047 L p 0 0 .683 r
F P
0 g
s
.60864 .52047 m .59445 .53796 L .58906 .52288 L p 0 0 .683 r
F P
0 g
s
.6324 .41216 m .6718 .42222 L .68134 .42672 L p 0 .388 0 r
F P
0 g
s
.68134 .42672 m .71374 .45933 L .71186 .47487 L p 0 0 0 r
F P
0 g
s
.71374 .45933 m .68134 .42672 L .6718 .42222 L p 0 .346 0 r
F P
0 g
s
.74139 .52417 m .72629 .55421 L .74035 .56081 L p .572 0 0 r
F P
0 g
s
.72629 .55421 m .74139 .52417 L .72827 .52008 L p .41 0 0 r
F P
0 g
s
.43581 .48816 m .42399 .51151 L .41196 .50155 L p 0 .417 0 r
F P
0 g
s
.42399 .51151 m .43581 .48816 L .44783 .49828 L p 0 .141 0 r
F P
0 g
s
.69998 .33672 m .7158 .3503 L .70744 .36373 L p .904 0 0 r
F P
0 g
s
.7158 .3503 m .69998 .33672 L .71083 .32431 L p .819 0 0 r
F P
0 g
s
.71083 .32431 m .72855 .33961 L .7158 .3503 L p .819 0 0 r
F P
0 g
s
.7437 .45417 m .72592 .4843 L .73626 .48696 L p .405 0 0 r
F P
0 g
s
.72592 .4843 m .7437 .45417 L .73059 .45277 L p 0 0 0 r
F P
0 g
s
.60587 .46522 m .63028 .46008 L .63703 .44485 L p 0 .537 0 r
F P
0 g
s
.63703 .44485 m .60683 .4511 L .60587 .46522 L p 0 .537 0 r
F P
0 g
s
.42112 .38797 m .41446 .40156 L .41414 .39442 L p 0 0 0 r
F P
0 g
s
.42946 .38051 m .42112 .38797 L .42344 .37716 L p .067 0 0 r
F P
0 g
s
.42112 .38797 m .42946 .38051 L .4253 .3842 L p .477 0 0 r
F P
0 g
s
.41428 .39416 m .42344 .37716 L .42112 .38797 L p 0 0 0 r
F P
0 g
s
.4253 .3842 m .41929 .38951 L p .42112 .38797 L .477 0 0 r
F P
0 g
s
.41414 .39442 m .41428 .39416 L p .42112 .38797 L 0 0 0 r
F P
0 g
s
.26989 .52535 m .29091 .52725 L .28487 .51179 L p 0 0 .53 r
F P
0 g
s
.28487 .51179 m .26392 .51032 L .26989 .52535 L p 0 0 .53 r
F P
0 g
s
.34184 .37059 m .33103 .40838 L .33048 .41243 L p .914 0 0 r
F P
0 g
s
.29091 .52725 m .26989 .52535 L .27068 .54211 L p 0 0 .651 r
F P
0 g
s
.72855 .33961 m .71083 .32431 L .72486 .31735 L p .694 0 0 r
F P
0 g
s
.72486 .31735 m .74356 .33344 L .72855 .33961 L p .694 0 0 r
F P
0 g
s
.63977 .49433 m .65316 .49366 L .65219 .47876 L p 0 .677 0 r
F P
0 g
s
.65219 .47876 m .63184 .48719 L .63977 .49433 L p 0 .677 0 r
F P
0 g
s
.49999 .48478 m .48678 .4794 L .48245 .49249 L p 0 .96 0 r
F P
0 g
s
.48678 .4794 m .49999 .48478 L .50683 .4755 L p 0 .808 0 r
F P
0 g
s
.48567 .42058 m .51841 .41934 L .50809 .40789 L p 0 .278 0 r
F P
0 g
s
.50809 .40789 m .47297 .41481 L .48567 .42058 L p 0 .278 0 r
F P
0 g
s
.40887 .42373 m .41446 .40156 L .40824 .40633 L p .957 0 0 r
F P
0 g
s
.37554 .53325 m .38944 .56583 L .40246 .55828 L p .804 0 0 r
F P
0 g
s
.40246 .55828 m .39073 .52943 L .37554 .53325 L p .804 0 0 r
F P
0 g
s
.49465 .5461 m .51346 .54475 L .51419 .52802 L p 0 0 .669 r
F P
0 g
s
.51419 .52802 m .49628 .52985 L .49465 .5461 L p 0 0 .669 r
F P
0 g
s
.54372 .49979 m .55212 .49705 L .56596 .50054 L p 0 0 1 r
F P
0 g
s
.46128 .41542 m .42431 .44391 L .4309 .43288 L p 0 .954 0 r
F P
0 g
s
.4309 .43288 m .47297 .41481 L .46128 .41542 L p 0 .954 0 r
F P
0 g
s
.1956 .49788 m .20452 .50571 L .1951 .51932 L p 0 0 .793 r
F P
0 g
s
.1951 .51932 m .18273 .5086 L .1956 .49788 L p 0 0 .793 r
F P
0 g
s
.18273 .5086 m .1951 .51932 L .19021 .53506 L p 0 0 .695 r
F P
0 g
s
.55367 .57864 m .57786 .56987 L .56442 .56979 L p 0 0 .954 r
F P
0 g
s
.56442 .56979 m .54037 .57913 L .55367 .57864 L p 0 0 .954 r
F P
0 g
s
.52035 .51339 m .50322 .51478 L .49628 .52985 L p 0 0 .54 r
F P
0 g
s
.50322 .51478 m .52035 .51339 L .53082 .50342 L p 0 0 .37 r
F P
0 g
s
.39999 .47092 m .42431 .44391 L .42296 .45848 L p 0 .713 0 r
F P
0 g
s
.42296 .45848 m .40349 .48722 L .39999 .47092 L p 0 .713 0 r
F P
0 g
s
.72355 .51153 m .70508 .49002 L .71186 .47487 L p 0 .264 0 r
F P
0 g
s
.70508 .49002 m .72355 .51153 L .71162 .5202 L p 0 .196 0 r
F P
0 g
s
.71162 .5202 m .69453 .50223 L .70508 .49002 L p 0 .196 0 r
F P
0 g
s
.75115 .48891 m .72827 .52008 L .74139 .52417 L p .523 0 0 r
F P
0 g
s
.72827 .52008 m .75115 .48891 L .73626 .48696 L p .294 0 0 r
F P
0 g
s
.78468 .36651 m .7703 .33806 L .77741 .34818 L p 0 0 0 r
F P
0 g
s
.77741 .34818 m .78976 .37348 L .78468 .36651 L p 0 0 0 r
F P
0 g
s
.6718 .42222 m .6324 .41216 L .6214 .42197 L p 0 .433 0 r
F P
0 g
s
.57141 .49291 m .56596 .50054 L .55212 .49705 L p 0 0 .9 r
F P
0 g
s
.56596 .50054 m .57141 .49291 L .58648 .49958 L p 0 0 .848 r
F P
0 g
s
.21028 .49067 m .21688 .49653 L .20452 .50571 L p 0 0 .872 r
F P
0 g
s
.20452 .50571 m .1956 .49788 L .21028 .49067 L p 0 0 .872 r
F P
0 g
s
.50531 .44643 m .52959 .44959 L .52603 .43413 L p 0 .617 0 r
F P
0 g
s
.52959 .44959 m .50531 .44643 L .50874 .46192 L p 0 .734 0 r
F P
0 g
s
.77146 .56787 m .78736 .5289 L .77386 .52902 L p .761 0 0 r
F P
0 g
s
.77386 .52902 m .75637 .56566 L .77146 .56787 L p .761 0 0 r
F P
0 g
s
.43471 .36686 m .42344 .37716 L .41958 .36446 L p .337 0 0 r
F P
0 g
s
.41958 .36446 m .43418 .35091 L .43471 .36686 L p .337 0 0 r
F P
0 g
s
.69412 .35248 m .70744 .36373 L .7049 .37765 L p .959 0 0 r
F P
0 g
s
.70744 .36373 m .69412 .35248 L .69998 .33672 L p .904 0 0 r
F P
0 g
s
.73249 .49794 m .71186 .47487 L .71374 .45933 L p 0 .308 0 r
F P
0 g
s
.71186 .47487 m .73249 .49794 L .72355 .51153 L p 0 .264 0 r
F P
0 g
s
.37084 .49588 m .39073 .52943 L .40178 .52472 L p .792 0 0 r
F P
0 g
s
.40178 .52472 m .38652 .4943 L .37084 .49588 L p .792 0 0 r
F P
0 g
s
.38223 .3359 m .40363 .31541 L .41722 .32282 L p .626 0 0 r
F P
0 g
s
.40363 .31541 m .38223 .3359 L .36819 .33491 L p .694 0 0 r
F P
0 g
s
.49628 .52985 m .51419 .52802 L .52035 .51339 L p 0 0 .54 r
F P
0 g
s
.59445 .53796 m .60864 .52047 L .61195 .5312 L p 0 0 .489 r
F P
0 g
s
.61195 .5312 m .59424 .55238 L .59445 .53796 L p 0 0 .489 r
F P
0 g
s
.24195 .49678 m .25386 .49967 L .24149 .49523 L p 0 0 0 r
F P
0 g
s
.44783 .49828 m .43757 .51537 L .42399 .51151 L p 0 .141 0 r
F P
0 g
s
.43757 .51537 m .44783 .49828 L .4611 .50274 L p 0 0 0 r
F P
0 g
s
.35807 .57389 m .34385 .53571 L .33304 .53392 L p .426 0 0 r
F P
0 g
s
.74356 .33344 m .72486 .31735 L .7397 .31709 L p .464 0 0 r
F P
0 g
s
.7397 .31709 m .75832 .33288 L .74356 .33344 L p .464 0 0 r
F P
0 g
s
.19021 .53506 m .17383 .52104 L .18273 .5086 L p 0 0 .695 r
F P
0 g
s
.17383 .52104 m .19021 .53506 L .19068 .5503 L p 0 0 .532 r
F P
0 g
s
.69453 .50223 m .71162 .5202 L .69871 .52243 L p 0 .074 0 r
F P
0 g
s
.69871 .52243 m .68197 .50945 L .69453 .50223 L p 0 .074 0 r
F P
0 g
s
.24523 .57477 m .21955 .56974 L .20668 .56934 L p 0 0 .861 r
F P
0 g
s
.21955 .56974 m .24523 .57477 L .25713 .56919 L p 0 0 .978 r
F P
0 g
s
.79921 .447 m .79759 .40738 L .79352 .41293 L p .933 0 0 r
F P
0 g
s
.79759 .40738 m .78976 .37348 L .78818 .38259 L p .998 0 0 r
F P
0 g
s
.78818 .38259 m .79352 .41293 L .79759 .40738 L p .998 0 0 r
F P
0 g
s
.75755 .52735 m .74035 .56081 L .75637 .56566 L p .672 0 0 r
F P
0 g
s
.74035 .56081 m .75755 .52735 L .74139 .52417 L p .572 0 0 r
F P
0 g
s
.49713 .43178 m .52603 .43413 L .51841 .41934 L p 0 .484 0 r
F P
0 g
s
.51841 .41934 m .48567 .42058 L .49713 .43178 L p 0 .484 0 r
F P
0 g
s
.34184 .37059 m .35735 .33921 L .36819 .33491 L p .888 0 0 r
F P
0 g
s
.35735 .33921 m .34184 .37059 L .33838 .37651 L p .783 0 0 r
F P
0 g
s
.33048 .41243 m .33838 .37651 L .34184 .37059 L p .914 0 0 r
F P
0 g
s
.32589 .45067 m .33914 .49531 L .35403 .49624 L p .036 0 0 r
F P
0 g
s
.22432 .48818 m .2301 .49335 L .21688 .49653 L p 0 0 .942 r
F P
0 g
s
.21688 .49653 m .21028 .49067 L .22432 .48818 L p 0 0 .942 r
F P
0 g
s
.79535 .48792 m .77386 .52902 L .78736 .5289 L p .821 0 0 r
F P
0 g
s
.52603 .43413 m .49713 .43178 L .50531 .44643 L p 0 .617 0 r
F P
0 g
s
.34385 .53571 m .35807 .57389 L .3737 .57137 L p .612 0 0 r
F P
0 g
s
.60683 .4511 m .63703 .44485 L .6475 .43209 L p 0 .496 0 r
F P
0 g
s
.6475 .43209 m .61233 .43577 L .60683 .4511 L p 0 .496 0 r
F P
0 g
s
.75637 .56566 m .77386 .52902 L .75755 .52735 L p .672 0 0 r
F P
0 g
s
.79921 .447 m .78362 .48943 L .79535 .48792 L p .864 0 0 r
F P
0 g
s
.78362 .48943 m .79921 .447 L .7905 .45022 L p .764 0 0 r
F P
0 g
s
.79352 .41293 m .7905 .45022 L .79921 .447 L p .933 0 0 r
F P
0 g
s
.59082 .47966 m .58648 .49958 L .57141 .49291 L p 0 0 .787 r
F P
0 g
s
.58648 .49958 m .59082 .47966 L .60603 .48541 L p 0 0 .749 r
F P
0 g
s
.60603 .48541 m .59963 .5093 L .58648 .49958 L p 0 0 .749 r
F P
0 g
s
.6718 .42222 m .71035 .44602 L .71374 .45933 L p 0 .346 0 r
F P
0 g
s
.73695 .48172 m .71374 .45933 L .71035 .44602 L p 0 .333 0 r
F P
0 g
s
.71374 .45933 m .73695 .48172 L .73249 .49794 L p 0 .308 0 r
F P
0 g
s
.77386 .52902 m .79535 .48792 L .78362 .48943 L p .729 0 0 r
F P
0 g
s
.65982 .42409 m .6214 .42197 L .61233 .43577 L p 0 .464 0 r
F P
0 g
s
.6214 .42197 m .65982 .42409 L .6718 .42222 L p 0 .433 0 r
F P
0 g
s
.35891 .53547 m .3737 .57137 L .38944 .56583 L p .718 0 0 r
F P
0 g
s
.38944 .56583 m .37554 .53325 L .35891 .53547 L p .718 0 0 r
F P
0 g
s
.40207 .45541 m .4309 .43288 L .42431 .44391 L p 0 .841 0 r
F P
0 g
s
.42431 .44391 m .39999 .47092 L .40207 .45541 L p 0 .841 0 r
F P
0 g
s
.55212 .49705 m .55691 .4904 L .57141 .49291 L p 0 0 .9 r
F P
0 g
s
.5747 .47596 m .57141 .49291 L .55691 .4904 L p 0 0 .81 r
F P
0 g
s
.57141 .49291 m .5747 .47596 L .59082 .47966 L p 0 0 .787 r
F P
0 g
s
.35973 .45863 m .38652 .4943 L .39827 .49178 L p .717 0 0 r
F P
0 g
s
.39827 .49178 m .37653 .45947 L .35973 .45863 L p .717 0 0 r
F P
0 g
s
.71035 .44602 m .6718 .42222 L .65982 .42409 L p 0 .419 0 r
F P
0 g
s
.3288 .49326 m .34385 .53571 L .35891 .53547 L p .401 0 0 r
F P
0 g
s
.35891 .53547 m .33914 .49531 L .3288 .49326 L p .401 0 0 r
F P
0 g
s
.69423 .36891 m .7049 .37765 L .70867 .3897 L p .926 0 0 r
F P
0 g
s
.7049 .37765 m .69423 .36891 L .69412 .35248 L p .959 0 0 r
F P
0 g
s
.72449 .42438 m .72221 .40666 L .71443 .40097 L p 0 0 0 r
F P
0 g
s
.61233 .43577 m .6475 .43209 L .65982 .42409 L p 0 .464 0 r
F P
0 g
s
.38113 .4286 m .39129 .45899 L .40139 .45728 L p .815 0 0 r
F P
0 g
s
.40139 .45728 m .39561 .42894 L .38113 .4286 L p .815 0 0 r
F P
0 g
s
.3737 .57137 m .35891 .53547 L .34385 .53571 L p .612 0 0 r
F P
0 g
s
.54037 .57913 m .56442 .56979 L .5504 .56338 L p 0 0 .974 r
F P
0 g
s
.5504 .56338 m .52784 .57276 L .54037 .57913 L p 0 0 .974 r
F P
0 g
s
.63184 .48719 m .65219 .47876 L .65663 .46275 L p 0 .594 0 r
F P
0 g
s
.65663 .46275 m .62846 .47508 L .63184 .48719 L p 0 .594 0 r
F P
0 g
s
.50683 .4755 m .48552 .46372 L .48678 .4794 L p 0 .808 0 r
F P
0 g
s
.48552 .46372 m .50683 .4755 L .50874 .46192 L p 0 .666 0 r
F P
0 g
s
.59963 .5093 m .60603 .48541 L .61777 .49226 L p 0 0 .68 r
F P
0 g
s
.61777 .49226 m .60864 .52047 L .59963 .5093 L p 0 0 .68 r
F P
0 g
s
.73059 .45277 m .73609 .4263 L .72449 .42438 L p .22 0 0 r
F P
0 g
s
.73609 .4263 m .73059 .45277 L .7437 .45417 L p .498 0 0 r
F P
0 g
s
.42344 .37716 m .43471 .36686 L .42946 .38051 L p .067 0 0 r
F P
0 g
s
.76794 .48979 m .74139 .52417 L .75755 .52735 L p .64 0 0 r
F P
0 g
s
.74139 .52417 m .76794 .48979 L .75115 .48891 L p .523 0 0 r
F P
0 g
s
.19068 .5503 m .17044 .53313 L .17383 .52104 L p 0 0 .532 r
F P
0 g
s
.17044 .53313 m .19068 .5503 L .19649 .56244 L p 0 0 .147 r
F P
0 g
s
.76002 .45417 m .73626 .48696 L .75115 .48891 L p .574 0 0 r
F P
0 g
s
.73626 .48696 m .76002 .45417 L .7437 .45417 L p .405 0 0 r
F P
0 g
s
.25035 .50629 m .26392 .51032 L .25386 .49967 L p 0 0 .137 r
F P
0 g
s
.25386 .49967 m .24195 .49678 L .25035 .50629 L p 0 0 .137 r
F P
0 g
s
.59424 .55238 m .61195 .5312 L .60892 .53966 L p 0 0 0 r
F P
0 g
s
.60892 .53966 m .5884 .56368 L .59424 .55238 L p 0 0 0 r
F P
0 g
s
.4611 .50274 m .45134 .51153 L p .43757 .51537 L 0 0 0 r
F P
0 g
s
.45037 .5124 m .4611 .50274 L .47335 .50072 L p 0 .982 0 r
F P
0 g
s
.46017 .50304 m .47223 .50092 L p .48245 .49249 L 0 .927 0 r
F P
0 g
s
.47223 .50092 m .47335 .50072 L .47428 .49988 L p 0 .927 0 r
F P
0 g
s
.47428 .49988 m .48245 .49249 L p .47223 .50092 L 0 .927 0 r
F P
0 g
s
.45134 .51153 m .45037 .5124 L .44899 .51272 L p 0 0 0 r
F P
0 g
s
.44899 .51272 m .43757 .51537 L p .45134 .51153 L 0 0 0 r
F P
0 g
s
.68197 .50945 m .69871 .52243 L .68703 .51776 L p 0 0 0 r
F P
0 g
s
.68703 .51776 m .66952 .5104 L .68197 .50945 L p 0 0 0 r
F P
0 g
s
.36819 .33491 m .38951 .31445 L .40363 .31541 L p .694 0 0 r
F P
0 g
s
.38951 .31445 m .36819 .33491 L .35735 .33921 L p .692 0 0 r
F P
0 g
s
.47297 .41481 m .4309 .43288 L .44166 .42733 L p 0 .869 0 r
F P
0 g
s
.44166 .42733 m .48567 .42058 L .47297 .41481 L p 0 .869 0 r
F P
0 g
s
.2059 .4771 m .21028 .49067 L .1956 .49788 L p 0 0 .807 r
F P
0 g
s
.1956 .49788 m .18984 .48128 L .2059 .4771 L p 0 0 .807 r
F P
0 g
s
.18984 .48128 m .1956 .49788 L .18273 .5086 L p 0 0 .765 r
F P
0 g
s
.41295 .39501 m .40824 .40633 L .41446 .40156 L p .883 0 0 r
F P
0 g
s
.41446 .40156 m .42112 .38797 L .41295 .39501 L p .883 0 0 r
F P
0 g
s
.75832 .33288 m .7397 .31709 L .7528 .32364 L p 0 0 0 r
F P
0 g
s
.7528 .32364 m .7703 .33806 L .75832 .33288 L p 0 0 0 r
F P
0 g
s
.72221 .40666 m .72449 .42438 L .73609 .4263 L p .414 0 0 r
F P
0 g
s
.25713 .56919 m .23285 .56351 L .21955 .56974 L p 0 0 .978 r
F P
0 g
s
.23285 .56351 m .25713 .56919 L .26616 .55764 L p 0 0 .87 r
F P
0 g
s
.23536 .49086 m .24195 .49678 L .2301 .49335 L p 0 0 .935 r
F P
0 g
s
.2301 .49335 m .22432 .48818 L .23536 .49086 L p 0 0 .935 r
F P
0 g
s
.22011 .476 m .22432 .48818 L .21028 .49067 L p 0 0 .837 r
F P
0 g
s
.21028 .49067 m .2059 .4771 L .22011 .476 L p 0 0 .837 r
F P
0 g
s
.78362 .48943 m .75755 .52735 L .77386 .52902 L p .729 0 0 r
F P
0 g
s
.35403 .49624 m .37554 .53325 L .39073 .52943 L p .704 0 0 r
F P
0 g
s
.39073 .52943 m .37084 .49588 L .35403 .49624 L p .704 0 0 r
F P
0 g
s
.33838 .37651 m .33048 .41243 L .33673 .41733 L p 0 0 0 r
F P
0 g
s
.33048 .41243 m .33195 .45379 L .34393 .45661 L p .136 0 0 r
F P
0 g
s
.587 .44984 m .59082 .47966 L .5747 .47596 L p 0 0 .754 r
F P
0 g
s
.59082 .47966 m .587 .44984 L .60357 .45228 L p 0 0 .734 r
F P
0 g
s
.60357 .45228 m .60603 .48541 L .59082 .47966 L p 0 0 .734 r
F P
0 g
s
.78976 .37348 m .77741 .34818 L .77833 .36153 L p .985 0 0 r
F P
0 g
s
.77833 .36153 m .78818 .38259 L .78976 .37348 L p .985 0 0 r
F P
0 g
s
.39648 .40784 m .39561 .42894 L .4054 .4272 L p .878 0 0 r
F P
0 g
s
.4054 .4272 m .40824 .40633 L .39648 .40784 L p .878 0 0 r
F P
0 g
s
.18273 .5086 m .17458 .48785 L .18984 .48128 L p 0 0 .765 r
F P
0 g
s
.17458 .48785 m .18273 .5086 L .17383 .52104 L p 0 0 .699 r
F P
0 g
s
.75755 .52735 m .78362 .48943 L .76794 .48979 L p .64 0 0 r
F P
0 g
s
.55691 .4904 m .56031 .47494 L .5747 .47596 L p 0 0 .81 r
F P
0 g
s
.56031 .47494 m .55691 .4904 L .54543 .49251 L p 0 0 .813 r
F P
0 g
s
.65982 .42409 m .70223 .43723 L .71035 .44602 L p 0 .419 0 r
F P
0 g
s
.73619 .46562 m .71035 .44602 L .70223 .43723 L p 0 .266 0 r
F P
0 g
s
.71035 .44602 m .73619 .46562 L .73695 .48172 L p 0 .333 0 r
F P
0 g
s
.57125 .44886 m .5747 .47596 L .56031 .47494 L p 0 0 .761 r
F P
0 g
s
.5747 .47596 m .57125 .44886 L .587 .44984 L p 0 0 .754 r
F P
0 g
s
.7905 .45022 m .76794 .48979 L .78362 .48943 L p .764 0 0 r
F P
0 g
s
.4094 .44331 m .43882 .42873 L p .4309 .43288 L 0 .972 0 r
F P
0 g
s
.4309 .43288 m .40207 .45541 L .4094 .44331 L p 0 .972 0 r
F P
0 g
s
.44166 .42733 m .4094 .44331 L .42079 .43674 L p 0 .871 0 r
F P
0 g
s
.45477 .42829 m .42369 .43602 L p .43431 .4369 L 0 0 0 r
F P
0 g
s
.42369 .43602 m .42079 .43674 L .42216 .43676 L p 0 0 0 r
F P
0 g
s
.42216 .43676 m .43431 .4369 L p .42369 .43602 L 0 0 0 r
F P
0 g
s
.43882 .42873 m .44166 .42733 L .44051 .42792 L p 0 .972 0 r
F P
0 g
s
.44051 .42792 m .4309 .43288 L p .43882 .42873 L 0 .972 0 r
F P
0 g
s
.5312 .50852 m .52035 .51339 L .51419 .52802 L p 0 0 .407 r
F P
0 g
s
.2538 .52031 m .26989 .52535 L .26392 .51032 L p 0 0 .451 r
F P
0 g
s
.26392 .51032 m .25035 .50629 L .2538 .52031 L p 0 0 .451 r
F P
0 g
s
.52784 .57276 m .5504 .56338 L .53831 .55175 L p 0 0 .872 r
F P
0 g
s
.53831 .55175 m .5183 .5606 L .52784 .57276 L p 0 0 .872 r
F P
0 g
s
.76794 .48979 m .7905 .45022 L .77662 .45277 L p .676 0 0 r
F P
0 g
s
.7905 .45022 m .79352 .41293 L .78317 .4185 L p .816 0 0 r
F P
0 g
s
.78317 .4185 m .77662 .45277 L .7905 .45022 L p .816 0 0 r
F P
0 g
s
.79352 .41293 m .78818 .38259 L .78012 .39227 L p .894 0 0 r
F P
0 g
s
.78012 .39227 m .78317 .4185 L .79352 .41293 L p .894 0 0 r
F P
0 g
s
.60603 .48541 m .60357 .45228 L .61823 .4558 L p 0 0 .69 r
F P
0 g
s
.61823 .4558 m .61777 .49226 L .60603 .48541 L p 0 0 .69 r
F P
0 g
s
.60864 .52047 m .61777 .49226 L .62404 .49909 L p 0 0 .497 r
F P
0 g
s
.62404 .49909 m .61195 .5312 L .60864 .52047 L p 0 0 .497 r
F P
0 g
s
.50874 .46192 m .47885 .44823 L .48552 .46372 L p 0 .666 0 r
F P
0 g
s
.47885 .44823 m .50874 .46192 L .50531 .44643 L p 0 .516 0 r
F P
0 g
s
.26616 .55764 m .24422 .5517 L .23285 .56351 L p 0 0 .87 r
F P
0 g
s
.24422 .5517 m .26616 .55764 L .27068 .54211 L p 0 0 .753 r
F P
0 g
s
.77662 .45277 m .75115 .48891 L .76794 .48979 L p .676 0 0 r
F P
0 g
s
.75115 .48891 m .77662 .45277 L .76002 .45417 L p .574 0 0 r
F P
0 g
s
.33195 .45379 m .35403 .49624 L .37084 .49588 L p .453 0 0 r
F P
0 g
s
.37084 .49588 m .34393 .45661 L .33195 .45379 L p .453 0 0 r
F P
0 g
s
.70223 .43723 m .65982 .42409 L .6475 .43209 L p 0 .445 0 r
F P
0 g
s
.33838 .37651 m .35163 .34814 L .35735 .33921 L p .783 0 0 r
F P
0 g
s
.35163 .34814 m .33838 .37651 L .3416 .3844 L p .025 0 0 r
F P
0 g
s
.33673 .41733 m .3416 .3844 L .33838 .37651 L p 0 0 0 r
F P
0 g
s
.45477 .42829 m .49713 .43178 L .48567 .42058 L p 0 0 0 r
F P
0 g
s
.48567 .42058 m .44166 .42733 L .45477 .42829 L p 0 0 0 r
F P
0 g
s
.70867 .3897 m .70032 .38325 L .69423 .36891 L p .926 0 0 r
F P
0 g
s
.23006 .47819 m .23536 .49086 L .22432 .48818 L p 0 0 .835 r
F P
0 g
s
.22432 .48818 m .22011 .476 L .23006 .47819 L p 0 0 .835 r
F P
0 g
s
.19649 .56244 m .17322 .54278 L .17044 .53313 L p 0 0 .147 r
F P
0 g
s
.17322 .54278 m .19649 .56244 L .20668 .56934 L p 0 0 .694 r
F P
0 g
s
.33914 .49531 m .35891 .53547 L .37554 .53325 L p .596 0 0 r
F P
0 g
s
.37554 .53325 m .35403 .49624 L .33914 .49531 L p .596 0 0 r
F P
0 g
s
.62846 .47508 m .65663 .46275 L .6657 .44849 L p 0 .535 0 r
F P
0 g
s
.6657 .44849 m .63028 .46008 L .62846 .47508 L p 0 .535 0 r
F P
0 g
s
.20652 .45297 m .2059 .4771 L .18984 .48128 L p 0 0 .76 r
F P
0 g
s
.18984 .48128 m .19044 .45434 L .20652 .45297 L p 0 0 .76 r
F P
0 g
s
.19044 .45434 m .18984 .48128 L .17458 .48785 L p 0 0 .742 r
F P
0 g
s
.45143 .34491 m .43418 .35091 L .42801 .33536 L p .395 0 0 r
F P
0 g
s
.42801 .33536 m .44724 .32886 L .45143 .34491 L p .395 0 0 r
F P
0 g
s
.44724 .32886 m .42801 .33536 L .41722 .32282 L p .409 0 0 r
F P
0 g
s
.34393 .45661 m .37084 .49588 L .38652 .4943 L p .617 0 0 r
F P
0 g
s
.38652 .4943 m .35973 .45863 L .34393 .45661 L p .617 0 0 r
F P
0 g
s
.44985 .36177 m .43471 .36686 L .43418 .35091 L p .364 0 0 r
F P
0 g
s
.43418 .35091 m .45143 .34491 L .44985 .36177 L p .364 0 0 r
F P
0 g
s
.66952 .5104 m .68703 .51776 L .67863 .50692 L p 0 .537 0 r
F P
0 g
s
.67863 .50692 m .65931 .50484 L .66952 .5104 L p 0 .537 0 r
F P
0 g
s
.21937 .45347 m .22011 .476 L .2059 .4771 L p 0 0 .762 r
F P
0 g
s
.2059 .4771 m .20652 .45297 L .21937 .45347 L p 0 0 .762 r
F P
0 g
s
.47335 .50072 m .46017 .50304 L .45037 .5124 L p 0 .982 0 r
F P
0 g
s
.56835 .41572 m .587 .44984 L .57125 .44886 L p 0 0 .755 r
F P
0 g
s
.587 .44984 m .56835 .41572 L .58312 .41498 L p 0 0 .75 r
F P
0 g
s
.58312 .41498 m .60357 .45228 L .587 .44984 L p 0 0 .75 r
F P
0 g
s
.75162 .42585 m .7437 .45417 L .76002 .45417 L p .629 0 0 r
F P
0 g
s
.7437 .45417 m .75162 .42585 L .73609 .4263 L p .498 0 0 r
F P
0 g
s
.5884 .56368 m .60892 .53966 L .59996 .54434 L p 0 0 .832 r
F P
0 g
s
.59996 .54434 m .57786 .56987 L .5884 .56368 L p 0 0 .832 r
F P
0 g
s
.25164 .53639 m .27068 .54211 L .26989 .52535 L p 0 0 .625 r
F P
0 g
s
.26989 .52535 m .2538 .52031 L .25164 .53639 L p 0 0 .625 r
F P
0 g
s
.27068 .54211 m .25164 .53639 L .24422 .5517 L p 0 0 .753 r
F P
0 g
s
.41929 .38951 m .41295 .39501 L .42077 .38827 L p .477 0 0 r
F P
0 g
s
.41295 .39501 m .41929 .38951 L .40588 .39227 L p .701 0 0 r
F P
0 g
s
.41929 .38951 m .43137 .38692 L .41769 .39085 L p 0 0 0 r
F P
0 g
s
.41769 .39085 m .41754 .3909 L p .41929 .38951 L 0 0 0 r
F P
0 g
s
.42077 .38827 m .42112 .38797 L p .41929 .38951 L .477 0 0 r
F P
0 g
s
.24147 .49832 m .25035 .50629 L .24195 .49678 L p 0 0 0 r
F P
0 g
s
.24195 .49678 m .23536 .49086 L .24147 .49832 L p 0 0 0 r
F P
0 g
s
.60357 .45228 m .58312 .41498 L .59986 .41528 L p 0 0 .733 r
F P
0 g
s
.59986 .41528 m .61823 .4558 L .60357 .45228 L p 0 0 .733 r
F P
0 g
s
.36453 .42621 m .37653 .45947 L .39129 .45899 L p .727 0 0 r
F P
0 g
s
.39129 .45899 m .38113 .4286 L .36453 .42621 L p .727 0 0 r
F P
0 g
s
.73609 .4263 m .73515 .40867 L .72221 .40666 L p .414 0 0 r
F P
0 g
s
.71818 .39779 m .72221 .40666 L .73515 .40867 L p .416 0 0 r
F P
0 g
s
.67837 .34768 m .69412 .35248 L .69423 .36891 L p .953 0 0 r
F P
0 g
s
.69412 .35248 m .67837 .34768 L .68338 .3321 L p .934 0 0 r
F P
0 g
s
.68338 .3321 m .69998 .33672 L .69412 .35248 L p .934 0 0 r
F P
0 g
s
.7703 .33806 m .7528 .32364 L .76183 .33594 L p .759 0 0 r
F P
0 g
s
.76183 .33594 m .77741 .34818 L .7703 .33806 L p .759 0 0 r
F P
0 g
s
.17383 .52104 m .16266 .49572 L .17458 .48785 L p 0 0 .699 r
F P
0 g
s
.16266 .49572 m .17383 .52104 L .17044 .53313 L p 0 0 .558 r
F P
0 g
s
.35735 .33921 m .37729 .32017 L .38951 .31445 L p .692 0 0 r
F P
0 g
s
.37729 .32017 m .35735 .33921 L .35163 .34814 L p 0 0 0 r
F P
0 g
s
.52781 .52164 m .51419 .52802 L .51346 .54475 L p 0 0 .618 r
F P
0 g
s
.51419 .52802 m .52781 .52164 L .5312 .50852 L p 0 0 .407 r
F P
0 g
s
.5183 .5606 m .53831 .55175 L .53031 .53696 L p 0 0 .757 r
F P
0 g
s
.53031 .53696 m .51346 .54475 L .5183 .5606 L p 0 0 .757 r
F P
0 g
s
.69998 .33672 m .68338 .3321 L .69362 .32049 L p .851 0 0 r
F P
0 g
s
.69362 .32049 m .71083 .32431 L .69998 .33672 L p .851 0 0 r
F P
0 g
s
.46795 .43567 m .50531 .44643 L .49713 .43178 L p 0 .299 0 r
F P
0 g
s
.50531 .44643 m .46795 .43567 L .47885 .44823 L p 0 .516 0 r
F P
0 g
s
.41722 .32282 m .43797 .31635 L .44724 .32886 L p .409 0 0 r
F P
0 g
s
.43797 .31635 m .41722 .32282 L .40363 .31541 L p .391 0 0 r
F P
0 g
s
.56031 .47494 m .55895 .44953 L .57125 .44886 L p 0 0 .761 r
F P
0 g
s
.55895 .44953 m .56031 .47494 L .5501 .47678 L p 0 0 .74 r
F P
0 g
s
.49713 .43178 m .45477 .42829 L .46795 .43567 L p 0 .299 0 r
F P
0 g
s
.73515 .40867 m .73609 .4263 L .75162 .42585 L p .59 0 0 r
F P
0 g
s
.44278 .37661 m .42946 .38051 L .43471 .36686 L p .309 0 0 r
F P
0 g
s
.43471 .36686 m .44985 .36177 L .44278 .37661 L p .309 0 0 r
F P
0 g
s
.17458 .48785 m .1738 .45732 L .19044 .45434 L p 0 0 .742 r
F P
0 g
s
.1738 .45732 m .17458 .48785 L .16266 .49572 L p 0 0 .707 r
F P
0 g
s
.51346 .54475 m .53031 .53696 L .52781 .52164 L p 0 0 .618 r
F P
0 g
s
.77662 .45277 m .78317 .4185 L .76831 .42309 L p .723 0 0 r
F P
0 g
s
.76831 .42309 m .76002 .45417 L .77662 .45277 L p .723 0 0 r
F P
0 g
s
.67939 .36455 m .69423 .36891 L .70032 .38325 L p .915 0 0 r
F P
0 g
s
.69423 .36891 m .67939 .36455 L .67837 .34768 L p .953 0 0 r
F P
0 g
s
.37962 .48235 m .39999 .47092 L .40349 .48722 L p 0 .672 0 r
F P
0 g
s
.40349 .48722 m .38324 .4988 L .37962 .48235 L p 0 .672 0 r
F P
0 g
s
.38324 .4988 m .40349 .48722 L .41196 .50155 L p 0 .511 0 r
F P
0 g
s
.6475 .43209 m .69072 .43454 L .70223 .43723 L p 0 .445 0 r
F P
0 g
s
.70223 .43723 m .7303 .45234 L .73619 .46562 L p 0 .266 0 r
F P
0 g
s
.7303 .45234 m .70223 .43723 L .69072 .43454 L p 0 .048 0 r
F P
0 g
s
.40025 .397 m .39648 .40784 L .40824 .40633 L p .841 0 0 r
F P
0 g
s
.40824 .40633 m .41295 .39501 L .40025 .397 L p .841 0 0 r
F P
0 g
s
.55804 .41737 m .57125 .44886 L .55895 .44953 L p 0 0 .745 r
F P
0 g
s
.57125 .44886 m .55804 .41737 L .56835 .41572 L p 0 0 .755 r
F P
0 g
s
.69072 .43454 m .6475 .43209 L .63703 .44485 L p 0 .468 0 r
F P
0 g
s
.76002 .45417 m .76831 .42309 L .75162 .42585 L p .629 0 0 r
F P
0 g
s
.63028 .46008 m .6657 .44849 L .67778 .4385 L p 0 .496 0 r
F P
0 g
s
.67778 .4385 m .63703 .44485 L .63028 .46008 L p 0 .496 0 r
F P
0 g
s
.3416 .3844 m .33673 .41733 L .34878 .42222 L p .427 0 0 r
F P
0 g
s
.33673 .41733 m .34393 .45661 L .35973 .45863 L p .485 0 0 r
F P
0 g
s
.35973 .45863 m .34878 .42222 L .33673 .41733 L p .485 0 0 r
F P
0 g
s
.57409 .37847 m .59986 .41528 L .58312 .41498 L p 0 0 .756 r
F P
0 g
s
.61823 .4558 m .59986 .41528 L .61581 .41657 L p 0 0 .685 r
F P
0 g
s
.59986 .41528 m .57409 .37847 L .59011 .37658 L p 0 0 .753 r
F P
0 g
s
.59011 .37658 m .61581 .41657 L .59986 .41528 L p 0 0 .753 r
F P
0 g
s
.21737 .42225 m .20652 .45297 L .19044 .45434 L p 0 0 .68 r
F P
0 g
s
.19044 .45434 m .20486 .421 L .21737 .42225 L p 0 0 .68 r
F P
0 g
s
.20486 .421 m .19044 .45434 L .1738 .45732 L p 0 0 .702 r
F P
0 g
s
.78818 .38259 m .77833 .36153 L .77283 .37578 L p .928 0 0 r
F P
0 g
s
.77283 .37578 m .78012 .39227 L .78818 .38259 L p .928 0 0 r
F P
0 g
s
.38162 .46556 m .40207 .45541 L .39999 .47092 L p 0 .799 0 r
F P
0 g
s
.39999 .47092 m .37962 .48235 L .38162 .46556 L p 0 .799 0 r
F P
0 g
s
.38117 .40577 m .38113 .4286 L .39561 .42894 L p .782 0 0 r
F P
0 g
s
.39561 .42894 m .39648 .40784 L .38117 .40577 L p .782 0 0 r
F P
0 g
s
.41196 .50155 m .3919 .51215 L .38324 .4988 L p 0 .511 0 r
F P
0 g
s
.3919 .51215 m .41196 .50155 L .42399 .51151 L p 0 .25 0 r
F P
0 g
s
.2268 .45575 m .23006 .47819 L .22011 .476 L p 0 0 .699 r
F P
0 g
s
.22011 .476 m .21937 .45347 L .2268 .45575 L p 0 0 .699 r
F P
0 g
s
.56146 .38115 m .58312 .41498 L .56835 .41572 L p 0 0 .748 r
F P
0 g
s
.58312 .41498 m .56146 .38115 L .57409 .37847 L p 0 0 .756 r
F P
0 g
s
.42079 .43674 m .45477 .42829 L .44166 .42733 L p 0 .871 0 r
F P
0 g
s
.71083 .32431 m .69362 .32049 L .70738 .31486 L p .62 0 0 r
F P
0 g
s
.70738 .31486 m .72486 .31735 L .71083 .32431 L p .62 0 0 r
F P
0 g
s
.61777 .49226 m .61823 .4558 L .62851 .45981 L p 0 0 .563 r
F P
0 g
s
.62851 .45981 m .62404 .49909 L .61777 .49226 L p 0 0 .563 r
F P
0 g
s
.78317 .4185 m .78012 .39227 L .76695 .40078 L p .79 0 0 r
F P
0 g
s
.76695 .40078 m .76831 .42309 L .78317 .4185 L p .79 0 0 r
F P
0 g
s
.75538 .50532 m .73249 .49794 L .73695 .48172 L p 0 .291 0 r
F P
0 g
s
.73249 .49794 m .75538 .50532 L .74743 .51814 L p 0 .184 0 r
F P
0 g
s
.74743 .51814 m .72355 .51153 L .73249 .49794 L p 0 .184 0 r
F P
0 g
s
.75855 .48911 m .73695 .48172 L .73619 .46562 L p 0 .379 0 r
F P
0 g
s
.73695 .48172 m .75855 .48911 L .75538 .50532 L p 0 .291 0 r
F P
0 g
s
.63703 .44485 m .67778 .4385 L .69072 .43454 L p 0 .468 0 r
F P
0 g
s
.20668 .56934 m .18178 .54828 L .17322 .54278 L p 0 0 .694 r
F P
0 g
s
.18178 .54828 m .20668 .56934 L .21955 .56974 L p 0 0 .968 r
F P
0 g
s
.22418 .42436 m .21937 .45347 L .20652 .45297 L p 0 0 .567 r
F P
0 g
s
.22405 .42699 m .2268 .45575 L .21937 .45347 L p 0 0 0 r
F P
0 g
s
.23006 .47819 m .2268 .45575 L .22705 .45718 L p 0 0 .118 r
F P
0 g
s
.22705 .45718 m .22745 .45947 L p .22818 .46473 L 0 0 .118 r
F P
0 g
s
.23006 .47819 m .22705 .45718 L .22818 .46473 L 0 0 .118 r
F
.22109 .46401 m .23118 .49064 L .23183 .48897 L p 0 0 .558 r
F P
0 g
s
.23536 .49086 m .23006 .47819 L .23401 .48335 L p 0 0 .493 r
F P
0 g
s
.22745 .45947 m .23401 .48335 L .23006 .47819 L p 0 0 .118 r
F P
0 g
s
.23183 .48897 m .23401 .48335 L p .2301 .4775 L 0 0 .558 r
F P
0 g
s
.2301 .4775 m .22109 .46401 L .23183 .48897 L 0 0 .558 r
F
.22745 .45947 m .2268 .45575 L p .22394 .44826 L 0 0 .663 r
F P
0 g
s
.20652 .45297 m .21671 .42413 L p .2192 .43244 L 0 0 .567 r
F P
0 g
s
.22418 .42436 m .22405 .42699 L p .22165 .44055 L 0 0 0 r
F P
0 g
s
.21937 .45347 m .22165 .43966 L p .22165 .44055 L 0 0 0 r
F P
0 g
s
.22165 .43966 m .22418 .42436 L p .22165 .44055 L 0 0 0 r
F P
0 g
s
.22394 .44826 m .22165 .44226 L .2265 .45681 L 0 0 .663 r
F
.21686 .42968 m .22165 .44317 L p .22165 .44226 L F P
0 g
s
.22165 .44317 m .2265 .45681 L p .22165 .44226 L 0 0 .663 r
F P
0 g
s
.2265 .45681 m .22745 .45947 L p .22394 .44826 L 0 0 .663 r
F P
0 g
s
.21737 .42225 m .22418 .42436 L p .2192 .43244 L 0 0 .567 r
F P
0 g
s
.21671 .42413 m .21737 .42225 L p .2192 .43244 L 0 0 .567 r
F P
0 g
s
.61581 .41657 m .62851 .45981 L .61823 .4558 L p 0 0 .685 r
F P
0 g
s
.23401 .48335 m .24147 .49832 L .23536 .49086 L p 0 0 .493 r
F P
0 g
s
.1738 .45732 m .1888 .42081 L .20486 .421 L p 0 0 .702 r
F P
0 g
s
.1888 .42081 m .1738 .45732 L .15934 .46145 L p 0 0 .699 r
F P
0 g
s
.16266 .49572 m .15934 .46145 L .1738 .45732 L p 0 0 .707 r
F P
0 g
s
.73515 .40867 m .73184 .40045 L .71818 .39779 L p .416 0 0 r
F P
0 g
s
.61195 .5312 m .62404 .49909 L .62367 .50471 L p 0 0 0 r
F P
0 g
s
.62367 .50471 m .60892 .53966 L .61195 .5312 L p 0 0 0 r
F P
0 g
s
.34878 .42222 m .35973 .45863 L .37653 .45947 L p .631 0 0 r
F P
0 g
s
.37653 .45947 m .36453 .42621 L .34878 .42222 L p .631 0 0 r
F P
0 g
s
.40363 .31541 m .42514 .30953 L .43797 .31635 L p .391 0 0 r
F P
0 g
s
.42514 .30953 m .40363 .31541 L .38951 .31445 L p .254 0 0 r
F P
0 g
s
.48245 .49249 m .46522 .48887 L .46017 .50304 L p 0 .927 0 r
F P
0 g
s
.46522 .48887 m .48245 .49249 L .48678 .4794 L p 0 .787 0 r
F P
0 g
s
.65931 .50484 m .67863 .50692 L .67503 .49175 L p 0 .525 0 r
F P
0 g
s
.67503 .49175 m .65316 .49366 L .65931 .50484 L p 0 .525 0 r
F P
0 g
s
.68626 .3799 m .70032 .38325 L .71139 .39303 L p .754 0 0 r
F P
0 g
s
.70032 .38325 m .68626 .3799 L .67939 .36455 L p .915 0 0 r
F P
0 g
s
.751 .40659 m .75162 .42585 L .76831 .42309 L p .697 0 0 r
F P
0 g
s
.75162 .42585 m .751 .40659 L .73515 .40867 L p .59 0 0 r
F P
0 g
s
.24152 .50935 m .2538 .52031 L .25035 .50629 L p 0 0 .36 r
F P
0 g
s
.25035 .50629 m .24147 .49832 L .24152 .50935 L p 0 0 .36 r
F P
0 g
s
.62851 .45981 m .61581 .41657 L .62832 .41863 L p 0 0 .461 r
F P
0 g
s
.61581 .41657 m .59011 .37658 L .60685 .37579 L p 0 0 .733 r
F P
0 g
s
.60685 .37579 m .62832 .41863 L .61581 .41657 L p 0 0 .733 r
F P
0 g
s
.72355 .51153 m .74743 .51814 L .736 .52533 L p 0 .01 0 r
F P
0 g
s
.736 .52533 m .71162 .5202 L .72355 .51153 L p 0 .01 0 r
F P
0 g
s
.57786 .56987 m .59996 .54434 L .58657 .54435 L p 0 0 .932 r
F P
0 g
s
.58657 .54435 m .56442 .56979 L .57786 .56987 L p 0 0 .932 r
F P
0 g
s
.57444 .34074 m .59011 .37658 L .57409 .37847 L p 0 0 .74 r
F P
0 g
s
.59011 .37658 m .57444 .34074 L .59078 .33702 L p 0 0 .753 r
F P
0 g
s
.59078 .33702 m .60685 .37579 L .59011 .37658 L p 0 0 .753 r
F P
0 g
s
.3416 .3844 m .35212 .36021 L .35163 .34814 L p .025 0 0 r
F P
0 g
s
.35212 .36021 m .3416 .3844 L .35103 .39289 L p .406 0 0 r
F P
0 g
s
.34878 .42222 m .35103 .39289 L .3416 .3844 L p .427 0 0 r
F P
0 g
s
.43137 .38692 m .41929 .38951 L .42946 .38051 L p .189 0 0 r
F P
0 g
s
.42946 .38051 m .44278 .37661 L .43137 .38692 L p .189 0 0 r
F P
0 g
s
.2231 .38893 m .20486 .421 L .1888 .42081 L p 0 0 .545 r
F P
0 g
s
.22515 .39167 m .21737 .42225 L .20486 .421 L p 0 0 .133 r
F P
0 g
s
.2231 .38893 m .22515 .39167 L p .21609 .40477 L 0 0 .133 r
F P
0 g
s
.21262 .40979 m .21583 .40339 L .21609 .40477 L 0 0 .133 r
F
.21583 .40339 m .2231 .38893 L .21609 .40477 L F
.21031 .41142 m .2231 .38893 L p .21262 .40979 L F P
0 g
s
.22515 .39167 m .22004 .39394 L p .2184 .41128 L 0 0 .666 r
F P
0 g
s
.21737 .42225 m .21975 .41291 L p .2184 .41128 L 0 0 .666 r
F P
0 g
s
.21975 .41291 m .22515 .39167 L p .2184 .41128 L 0 0 .666 r
F P
0 g
s
.22418 .42436 m .21737 .42225 L p .21897 .4053 L 0 0 .666 r
F P
0 g
s
.22023 .41705 m .21744 .41188 L .22274 .42431 L 0 0 .81 r
F
.1888 .42081 m .21111 .39059 L p .21359 .39777 L 0 0 .545 r
F P
0 g
s
.20851 .39534 m .21796 .41458 L p .21744 .41188 L 0 0 .81 r
F P
0 g
s
.21796 .41458 m .22274 .42431 L p .21744 .41188 L 0 0 .81 r
F P
0 g
s
.22004 .39394 m .22136 .40361 L p .21897 .4053 L 0 0 .666 r
F P
0 g
s
.22136 .40361 m .22418 .42436 L p .21897 .4053 L 0 0 .666 r
F P
0 g
s
.20486 .421 m .21031 .41142 L p .21262 .40979 L 0 0 .133 r
F P
0 g
s
.22405 .42699 m .22418 .42436 L p .22023 .41705 L 0 0 .81 r
F P
0 g
s
.22274 .42431 m .22405 .42699 L p .22023 .41705 L 0 0 .81 r
F P
0 g
s
.21435 .3862 m .2231 .38893 L p .21359 .39777 L 0 0 .545 r
F P
0 g
s
.21111 .39059 m .21435 .3862 L p .21359 .39777 L 0 0 .545 r
F P
0 g
s
.21435 .3862 m .1888 .42081 L .17184 .4217 L p 0 0 .638 r
F P
0 g
s
.15934 .46145 m .17184 .4217 L .1888 .42081 L p 0 0 .699 r
F P
0 g
s
.75641 .47224 m .73619 .46562 L .7303 .45234 L p 0 .467 0 r
F P
0 g
s
.73619 .46562 m .75641 .47224 L .75855 .48911 L p 0 .379 0 r
F P
0 g
s
.38892 .45128 m .4094 .44331 L .40207 .45541 L p 0 .918 0 r
F P
0 g
s
.40207 .45541 m .38162 .46556 L .38892 .45128 L p 0 .918 0 r
F P
0 g
s
.76831 .42309 m .76695 .40078 L .751 .40659 L p .697 0 0 r
F P
0 g
s
.42399 .51151 m .40414 .5201 L .3919 .51215 L p 0 .25 0 r
F P
0 g
s
.40414 .5201 m .42399 .51151 L .43757 .51537 L p 0 0 0 r
F P
0 g
s
.73184 .40045 m .73515 .40867 L .751 .40659 L p .613 0 0 r
F P
0 g
s
.77741 .34818 m .76183 .33594 L .76519 .35187 L p .965 0 0 r
F P
0 g
s
.76519 .35187 m .77833 .36153 L .77741 .34818 L p .965 0 0 r
F P
0 g
s
.17044 .53313 m .15615 .50358 L .16266 .49572 L p 0 0 .558 r
F P
0 g
s
.15615 .50358 m .17044 .53313 L .17322 .54278 L p 0 0 .085 r
F P
0 g
s
.40588 .39227 m .40025 .397 L .41295 .39501 L p .701 0 0 r
F P
0 g
s
.62832 .41863 m .60685 .37579 L .62155 .37623 L p 0 0 .639 r
F P
0 g
s
.60685 .37579 m .59078 .33702 L .60705 .33519 L p 0 0 .753 r
F P
0 g
s
.60705 .33519 m .62155 .37623 L .60685 .37579 L p 0 0 .753 r
F P
0 g
s
.55437 .38418 m .56835 .41572 L .55804 .41737 L p 0 0 .712 r
F P
0 g
s
.56835 .41572 m .55437 .38418 L .56146 .38115 L p 0 0 .748 r
F P
0 g
s
.35163 .34814 m .36913 .33165 L .37729 .32017 L p 0 0 0 r
F P
0 g
s
.36913 .33165 m .35163 .34814 L .35212 .36021 L p .2 0 0 r
F P
0 g
s
.15934 .46145 m .16266 .49572 L .15615 .50358 L p 0 0 .628 r
F P
0 g
s
.72486 .31735 m .70738 .31486 L .72231 .31626 L p .058 0 0 r
F P
0 g
s
.72231 .31626 m .7397 .31709 L .72486 .31735 L p .058 0 0 r
F P
0 g
s
.35103 .39289 m .34878 .42222 L .36453 .42621 L p .595 0 0 r
F P
0 g
s
.72024 .44418 m .7303 .45234 L p .70641 .4416 L 0 .048 0 r
F P
0 g
s
.69072 .43454 m .72024 .44418 L p .70641 .4416 L 0 .048 0 r
F P
0 g
s
.74934 .45756 m .7303 .45234 L .72024 .44418 L p 0 .551 0 r
F P
0 g
s
.72024 .44418 m .7385 .44754 L .74934 .45756 L p 0 .551 0 r
F P
0 g
s
.72024 .44418 m .7077 .44261 L p .73656 .44723 L 0 0 0 r
F P
0 g
s
.7385 .44754 m .72024 .44418 L p .73656 .44723 L 0 0 0 r
F P
0 g
s
.72024 .44418 m .69072 .43454 L .67778 .4385 L p 0 .238 0 r
F P
0 g
s
.17184 .4217 m .20048 .38393 L .21435 .3862 L p 0 0 .638 r
F P
0 g
s
.20048 .38393 m .17184 .4217 L .1568 .42354 L p 0 0 .664 r
F P
0 g
s
.17184 .4217 m .15934 .46145 L .14952 .46603 L p 0 0 .676 r
F P
0 g
s
.14952 .46603 m .1568 .42354 L .17184 .4217 L p 0 0 .676 r
F P
0 g
s
.36506 .40047 m .36453 .42621 L .38113 .4286 L p .696 0 0 r
F P
0 g
s
.38113 .4286 m .38117 .40577 L .36506 .40047 L p .696 0 0 r
F P
0 g
s
.78012 .39227 m .77283 .37578 L .76184 .38841 L p .828 0 0 r
F P
0 g
s
.76184 .38841 m .76695 .40078 L .78012 .39227 L p .828 0 0 r
F P
0 g
s
.56072 .34573 m .57409 .37847 L .56146 .38115 L p 0 0 .705 r
F P
0 g
s
.57409 .37847 m .56072 .34573 L .57444 .34074 L p 0 0 .74 r
F P
0 g
s
.55895 .44953 m .55545 .4284 L p .55804 .41737 L 0 0 .745 r
F P
0 g
s
.55401 .41968 m .55895 .44953 L .55221 .45176 L p 0 0 .657 r
F P
0 g
s
.55705 .42225 m .55362 .44317 L p .55229 .45518 L 0 0 0 r
F P
0 g
s
.55362 .44317 m .55221 .45176 L .55226 .45413 L p 0 0 0 r
F P
0 g
s
.55226 .45413 m .55229 .45518 L p .55362 .44317 L 0 0 0 r
F P
0 g
s
.55545 .4284 m .55401 .41968 L .55673 .41812 L p 0 0 .745 r
F P
0 g
s
.55673 .41812 m .55804 .41737 L p .55545 .4284 L 0 0 .745 r
F P
0 g
s
.43431 .4369 m .46795 .43567 L .45477 .42829 L p 0 0 0 r
F P
0 g
s
.38522 .39351 m .38117 .40577 L .39648 .40784 L p .749 0 0 r
F P
0 g
s
.39648 .40784 m .40025 .397 L .38522 .39351 L p .749 0 0 r
F P
0 g
s
.21955 .56974 m .1947 .5486 L .18178 .54828 L p 0 0 .968 r
F P
0 g
s
.1947 .5486 m .21955 .56974 L .23285 .56351 L p 0 0 .916 r
F P
0 g
s
.21742 .3504 m .20048 .38393 L .18382 .38249 L p 0 0 .693 r
F P
0 g
s
.18382 .38249 m .2035 .34603 L .21742 .3504 L p 0 0 .693 r
F P
0 g
s
.1568 .42354 m .18382 .38249 L .20048 .38393 L p 0 0 .664 r
F P
0 g
s
.2035 .34603 m .18382 .38249 L .16712 .38212 L p 0 0 .716 r
F P
0 g
s
.18382 .38249 m .1568 .42354 L .14621 .42602 L p 0 0 .662 r
F P
0 g
s
.14621 .42602 m .16712 .38212 L .18382 .38249 L p 0 0 .662 r
F P
0 g
s
.71162 .5202 m .736 .52533 L .72303 .52561 L p 0 0 0 r
F P
0 g
s
.72303 .52561 m .69871 .52243 L .71162 .5202 L p 0 0 0 r
F P
0 g
s
.36453 .42621 m .36506 .40047 L .35103 .39289 L p .595 0 0 r
F P
0 g
s
.5387 .50868 m .52781 .52164 L .53031 .53696 L p 0 0 .597 r
F P
0 g
s
.52781 .52164 m .5387 .50868 L .53896 .49895 L p 0 0 .235 r
F P
0 g
s
.71139 .39303 m .69784 .39111 L .68626 .3799 L p .754 0 0 r
F P
0 g
s
.15615 .50358 m .14952 .46603 L .15934 .46145 L p 0 0 .628 r
F P
0 g
s
.22632 .35515 m .21435 .3862 L .20048 .38393 L p 0 0 .616 r
F P
0 g
s
.20048 .38393 m .21742 .3504 L .22632 .35515 L p 0 0 .616 r
F P
0 g
s
.74731 .39714 m .751 .40659 L .76695 .40078 L p .731 0 0 r
F P
0 g
s
.751 .40659 m .74731 .39714 L .73184 .40045 L p .613 0 0 r
F P
0 g
s
.23541 .52206 m .25164 .53639 L .2538 .52031 L p 0 0 .62 r
F P
0 g
s
.2538 .52031 m .24152 .50935 L .23541 .52206 L p 0 0 .62 r
F P
0 g
s
.48678 .4794 m .46459 .47234 L .46522 .48887 L p 0 .787 0 r
F P
0 g
s
.46459 .47234 m .48678 .4794 L .48552 .46372 L p 0 .652 0 r
F P
0 g
s
.38951 .31445 m .41094 .30966 L .42514 .30953 L p .254 0 0 r
F P
0 g
s
.41094 .30966 m .38951 .31445 L .37729 .32017 L p .064 0 0 r
F P
0 g
s
.7303 .45234 m .74934 .45756 L .75641 .47224 L p 0 .467 0 r
F P
0 g
s
.72559 .39654 m .73184 .40045 L .74731 .39714 L p .587 0 0 r
F P
0 g
s
.76695 .40078 m .76184 .38841 L .74731 .39714 L p .731 0 0 r
F P
0 g
s
.65316 .49366 m .67503 .49175 L .6769 .47492 L p 0 .461 0 r
F P
0 g
s
.6769 .47492 m .65219 .47876 L .65316 .49366 L p 0 .461 0 r
F P
0 g
s
.57697 .30145 m .59078 .33702 L .57444 .34074 L p 0 0 .776 r
F P
0 g
s
.59078 .33702 m .57697 .30145 L .59309 .29721 L p 0 0 .797 r
F P
0 g
s
.59309 .29721 m .60705 .33519 L .59078 .33702 L p 0 0 .797 r
F P
0 g
s
.40031 .44194 m .42079 .43674 L .4094 .44331 L p 0 .984 0 r
F P
0 g
s
.4094 .44331 m .38892 .45128 L .40031 .44194 L p 0 .984 0 r
F P
0 g
s
.56442 .56979 m .58657 .54435 L .57108 .53963 L p 0 0 .886 r
F P
0 g
s
.57108 .53963 m .5504 .56338 L .56442 .56979 L p 0 0 .886 r
F P
0 g
s
.43757 .51537 m .41792 .52123 L .40414 .5201 L p 0 0 0 r
F P
0 g
s
.41792 .52123 m .43757 .51537 L .45037 .5124 L p 0 .9 0 r
F P
0 g
s
.62404 .49909 m .62851 .45981 L .6326 .46364 L p 0 0 0 r
F P
0 g
s
.6326 .46364 m .62367 .50471 L .62404 .49909 L p 0 0 0 r
F P
0 g
s
.35103 .39289 m .35879 .37331 L .35212 .36021 L p .406 0 0 r
F P
0 g
s
.35879 .37331 m .35103 .39289 L .36506 .40047 L p .564 0 0 r
F P
0 g
s
.42522 .39458 m .41696 .39198 L p .41754 .3909 L .306 0 0 r
F P
0 g
s
.41754 .3909 m .40588 .39227 L .41816 .38975 L p 0 0 0 r
F P
0 g
s
.40588 .39227 m .41754 .3909 L .40365 .38779 L p .338 0 0 r
F P
0 g
s
.41696 .39198 m .40365 .38779 L .41754 .3909 L p .306 0 0 r
F P
0 g
s
.41816 .38975 m .41929 .38951 L p .41754 .3909 L 0 0 0 r
F P
0 g
s
.62832 .41863 m .6326 .46364 L .62851 .45981 L p 0 0 .461 r
F P
0 g
s
.23118 .49064 m .24152 .50935 L .24147 .49832 L p 0 0 .371 r
F P
0 g
s
.24147 .49832 m .23401 .48335 L .23118 .49064 L p 0 0 .371 r
F P
0 g
s
.56181 .30876 m .57444 .34074 L .56072 .34573 L p 0 0 .736 r
F P
0 g
s
.57444 .34074 m .56181 .30876 L .57697 .30145 L p 0 0 .776 r
F P
0 g
s
.63173 .37784 m .63559 .38036 L p .63556 .38373 L 0 0 .587 r
F P
0 g
s
.63523 .42112 m .62832 .41863 L p .62777 .41521 L 0 0 .639 r
F P
0 g
s
.63031 .40497 m .63523 .42112 L p .62777 .41521 L 0 0 .639 r
F P
0 g
s
.63523 .42112 m .62155 .37623 L .63173 .37784 L p 0 0 0 r
F P
0 g
s
.6326 .46364 m .62832 .41863 L .63523 .42112 L p 0 0 .382 r
F P
0 g
s
.63523 .42112 m .63268 .44204 L p .6326 .46364 L 0 0 .382 r
F P
0 g
s
.63268 .44204 m .62968 .46663 L .63167 .4646 L p 0 0 .382 r
F P
0 g
s
.63167 .4646 m .6326 .46364 L p .63268 .44204 L 0 0 .382 r
F P
0 g
s
.62968 .46663 m .63523 .42112 L .63526 .42364 L p 0 0 .685 r
F P
0 g
s
.63526 .42364 m .63523 .42112 L p .6342 .40836 L 0 0 0 r
F P
0 g
s
.63173 .37784 m .63398 .40703 L .6342 .40836 L 0 0 0 r
F
.63299 .39414 m .63173 .37784 L p .63556 .38373 L 0 0 .587 r
F P
0 g
s
.63559 .38036 m .63233 .38339 L p .63206 .38618 L 0 0 .696 r
F P
0 g
s
.62155 .37623 m .63031 .40497 L p .62777 .41521 L 0 0 .639 r
F P
0 g
s
.62826 .42574 m .633 .39639 L p .63206 .38618 L 0 0 .696 r
F P
0 g
s
.63398 .40703 m .63526 .42364 L .6345 .41267 L 0 0 0 r
F
.63398 .40703 m .6345 .41267 L .6342 .40836 L F
.63415 .40925 m .63299 .39414 L p .63524 .38504 L 0 0 .587 r
F P
0 g
s
.633 .39639 m .63559 .38036 L p .63206 .38618 L 0 0 .696 r
F P
0 g
s
.63526 .42364 m .63415 .40925 L p .63556 .38373 L 0 0 .587 r
F P
0 g
s
.63556 .38373 m .63415 .40925 L .63524 .38504 L 0 0 .587 r
F
.62155 .37623 m .60705 .33519 L .62055 .33556 L p 0 0 .72 r
F P
0 g
s
.62055 .33556 m .63173 .37784 L .62155 .37623 L p 0 0 .72 r
F P
0 g
s
.77833 .36153 m .76519 .35187 L .76222 .36862 L p .94 0 0 r
F P
0 g
s
.76222 .36862 m .77283 .37578 L .77833 .36153 L p .94 0 0 r
F P
0 g
s
.46381 .37453 m .44278 .37661 L .44985 .36177 L p .395 0 0 r
F P
0 g
s
.44985 .36177 m .47016 .35865 L .46381 .37453 L p .395 0 0 r
F P
0 g
s
.47016 .35865 m .44985 .36177 L .45143 .34491 L p .355 0 0 r
F P
0 g
s
.45143 .34491 m .47088 .3421 L .47016 .35865 L p .355 0 0 r
F P
0 g
s
.47088 .3421 m .45143 .34491 L .44724 .32886 L p .283 0 0 r
F P
0 g
s
.7397 .31709 m .72231 .31626 L .73582 .32452 L p .576 0 0 r
F P
0 g
s
.73582 .32452 m .7528 .32364 L .7397 .31709 L p .576 0 0 r
F P
0 g
s
.16712 .38212 m .18691 .34276 L .2035 .34603 L p 0 0 .716 r
F P
0 g
s
.18691 .34276 m .16712 .38212 L .15315 .38289 L p 0 0 .72 r
F P
0 g
s
.14195 .42873 m .15315 .38289 L .16712 .38212 L p 0 0 .633 r
F P
0 g
s
.14621 .42602 m .14195 .42873 L p .1536 .40716 L 0 0 .633 r
F P
0 g
s
.1536 .40716 m .15764 .39967 L .15436 .40724 L 0 0 .633 r
F
.1536 .40716 m .15436 .40724 L .14621 .42602 L F
.1461 .47029 m .14621 .42602 L .1568 .42354 L p 0 0 .612 r
F P
0 g
s
.14952 .46603 m .1461 .47029 L p .15166 .44596 L 0 0 .612 r
F P
0 g
s
.15625 .5101 m .1461 .47029 L .14952 .46603 L p 0 0 .348 r
F P
0 g
s
.1461 .47029 m .14978 .47349 L p .1479 .44852 L 0 0 .369 r
F P
0 g
s
.14621 .42602 m .14616 .44627 L p .1479 .44852 L 0 0 .369 r
F P
0 g
s
.14616 .44627 m .1461 .47029 L p .1479 .44852 L 0 0 .369 r
F P
0 g
s
.14195 .42873 m .14621 .42602 L p .14811 .45123 L 0 0 .369 r
F P
0 g
s
.1534 .44339 m .14952 .46603 L p .15166 .44596 L 0 0 .612 r
F P
0 g
s
.14489 .43121 m .14195 .42873 L p .15151 .45317 L 0 0 .476 r
F P
0 g
s
.16006 .47504 m .15328 .45546 L p .15151 .45317 L 0 0 .476 r
F P
0 g
s
.15328 .45546 m .14489 .43121 L p .15151 .45317 L 0 0 .476 r
F P
0 g
s
.14978 .47349 m .14634 .45384 L p .14811 .45123 L 0 0 .369 r
F P
0 g
s
.14634 .45384 m .14195 .42873 L p .14811 .45123 L 0 0 .369 r
F P
0 g
s
.1568 .42354 m .1534 .44339 L p .15166 .44596 L 0 0 .612 r
F P
0 g
s
.14195 .42873 m .14489 .43121 L p .15008 .40085 L 0 0 .538 r
F P
0 g
s
.15315 .38289 m .14988 .39628 L p .15008 .40085 L 0 0 .538 r
F P
0 g
s
.14988 .39628 m .14195 .42873 L p .15008 .40085 L 0 0 .538 r
F P
0 g
s
.14431 .38469 m .15315 .38289 L p .14802 .4129 L 0 0 .538 r
F P
0 g
s
.14472 .41744 m .14431 .38469 L p .14802 .4129 L 0 0 .538 r
F P
0 g
s
.14489 .43121 m .14472 .41744 L p .14802 .4129 L 0 0 .538 r
F P
0 g
s
.16088 .39522 m .14621 .42602 L p .15764 .39967 L 0 0 .633 r
F P
0 g
s
.14489 .43121 m .15465 .43301 L p .14812 .40251 L 0 0 .144 r
F P
0 g
s
.14431 .38469 m .14448 .39806 L p .14812 .40251 L 0 0 .144 r
F P
0 g
s
.14448 .39806 m .14489 .43121 L p .14812 .40251 L 0 0 .144 r
F P
0 g
s
.14221 .38721 m .14431 .38469 L p .14902 .40671 L 0 0 .144 r
F P
0 g
s
.14734 .39005 m .14221 .38721 L p .15896 .41575 L 0 0 .661 r
F P
0 g
s
.16956 .4338 m .16273 .42036 L p .15896 .41575 L 0 0 .661 r
F P
0 g
s
.16273 .42036 m .14734 .39005 L p .15896 .41575 L 0 0 .661 r
F P
0 g
s
.15465 .43301 m .14746 .40655 L p .14902 .40671 L 0 0 .144 r
F P
0 g
s
.14746 .40655 m .14221 .38721 L p .14902 .40671 L 0 0 .144 r
F P
0 g
s
.16712 .38212 m .16088 .39522 L p .15764 .39967 L 0 0 .633 r
F P
0 g
s
.44762 .44383 m .47885 .44823 L .46795 .43567 L p 0 .275 0 r
F P
0 g
s
.46795 .43567 m .43431 .4369 L .44762 .44383 L p 0 .275 0 r
F P
0 g
s
.23285 .56351 m .20972 .54362 L .1947 .5486 L p 0 0 .916 r
F P
0 g
s
.20972 .54362 m .23285 .56351 L .24422 .5517 L p 0 0 .838 r
F P
0 g
s
.60892 .53966 m .62367 .50471 L .6166 .50813 L p 0 0 .791 r
F P
0 g
s
.6166 .50813 m .59996 .54434 L .60892 .53966 L p 0 0 .791 r
F P
0 g
s
.7077 .44261 m .72024 .44418 L p .6944 .44073 L 0 .238 0 r
F P
0 g
s
.7257 .44394 m .7385 .44754 L p .72299 .44506 L 0 0 0 r
F P
0 g
s
.7077 .44261 m .7257 .44394 L p .72299 .44506 L 0 0 0 r
F P
0 g
s
.7257 .44394 m .7077 .44261 L p .72124 .44453 L 0 0 0 r
F P
0 g
s
.71314 .44746 m .7257 .44394 L p .70922 .4461 L 0 0 0 r
F P
0 g
s
.69484 .44798 m .71314 .44746 L p .70922 .4461 L 0 0 0 r
F P
0 g
s
.71314 .44746 m .703 .45755 L p .71608 .44848 L 0 0 0 r
F P
0 g
s
.7257 .44394 m .71314 .44746 L p .72785 .44306 L 0 .763 0 r
F P
0 g
s
.7077 .44261 m .69484 .44798 L p .72124 .44453 L 0 0 0 r
F P
0 g
s
.72321 .44353 m .73506 .4409 L p .72268 .4446 L 0 .763 0 r
F P
0 g
s
.71314 .44746 m .72321 .44353 L p .72268 .4446 L 0 .763 0 r
F P
0 g
s
.72321 .44353 m .71314 .44746 L p .71608 .44848 L 0 0 0 r
F P
0 g
s
.71521 .45221 m .72321 .44353 L p .71064 .45225 L 0 0 0 r
F P
0 g
s
.703 .45755 m .71521 .45221 L p .71064 .45225 L 0 0 0 r
F P
0 g
s
.71521 .45221 m .71252 .46552 L p .71908 .45217 L 0 0 0 r
F P
0 g
s
.72321 .44353 m .71521 .45221 L p .72369 .44253 L 0 .866 0 r
F P
0 g
s
.72962 .43577 m .72321 .44353 L p .72369 .44253 L 0 .866 0 r
F P
0 g
s
.74056 .43387 m .73506 .4409 L .72321 .44353 L p 0 .947 0 r
F P
0 g
s
.72321 .44353 m .72962 .43577 L .74056 .43387 L p 0 .947 0 r
F P
0 g
s
.72405 .44208 m .72962 .43577 L p .7198 .44697 L 0 .866 0 r
F P
0 g
s
.71521 .45221 m .72405 .44208 L p .7198 .44697 L 0 .866 0 r
F P
0 g
s
.73506 .4409 m .7257 .44394 L p .72785 .44306 L 0 .763 0 r
F P
0 g
s
.74872 .4447 m .7385 .44754 L .7257 .44394 L p 0 .799 0 r
F P
0 g
s
.7257 .44394 m .73506 .4409 L .74872 .4447 L p 0 .799 0 r
F P
0 g
s
.73526 .41927 m .72962 .43577 L .72405 .44208 L p 0 .948 0 r
F P
0 g
s
.72405 .44208 m .72489 .45178 L p .72874 .43667 L 0 0 0 r
F P
0 g
s
.73211 .42347 m .73526 .41927 L p .72691 .43624 L 0 .948 0 r
F P
0 g
s
.72405 .44208 m .73211 .42347 L p .72691 .43624 L 0 .948 0 r
F P
0 g
s
.74478 .41768 m .74056 .43387 L .72962 .43577 L p 0 .869 0 r
F P
0 g
s
.72962 .43577 m .73526 .41927 L .74478 .41768 L p 0 .869 0 r
F P
0 g
s
.74049 .39309 m .74731 .39714 L .76184 .38841 L p .763 0 0 r
F P
0 g
s
.74731 .39714 m .74049 .39309 L .72559 .39654 L p .587 0 0 r
F P
0 g
s
.76184 .38841 m .75347 .38322 L .74049 .39309 L p .763 0 0 r
F P
0 g
s
.74444 .39111 m .74478 .41768 L .73526 .41927 L p 0 .838 0 r
F P
0 g
s
.72689 .39442 m .74049 .39309 L .75347 .38322 L p .763 0 0 r
F P
0 g
s
.73868 .39303 m .73526 .41927 L .73211 .42347 L p 0 .861 0 r
F P
0 g
s
.73526 .41927 m .73868 .39303 L .74444 .39111 L p 0 .838 0 r
F P
0 g
s
.73211 .42347 m .736 .42961 L p .73873 .40718 L 0 0 0 r
F P
0 g
s
.74007 .39616 m .73211 .42347 L p .73873 .40718 L 0 0 0 r
F P
0 g
s
.73211 .42347 m .72405 .44208 L p .72874 .43667 L 0 0 0 r
F P
0 g
s
.72405 .44208 m .71521 .45221 L p .71908 .45217 L 0 0 0 r
F P
0 g
s
.67778 .4385 m .7077 .44261 L p .6944 .44073 L 0 .238 0 r
F P
0 g
s
.7077 .44261 m .67778 .4385 L .6657 .44849 L p 0 .317 0 r
F P
0 g
s
.37052 .38512 m .36506 .40047 L .38117 .40577 L p .663 0 0 r
F P
0 g
s
.38117 .40577 m .38522 .39351 L .37052 .38512 L p .663 0 0 r
F P
0 g
s
.39155 .38821 m .38522 .39351 L .40025 .397 L p .655 0 0 r
F P
0 g
s
.40025 .397 m .40588 .39227 L .39155 .38821 L p .655 0 0 r
F P
0 g
s
.22415 .53421 m .24422 .5517 L .25164 .53639 L p 0 0 .749 r
F P
0 g
s
.25164 .53639 m .23541 .52206 L .22415 .53421 L p 0 0 .749 r
F P
0 g
s
.60705 .33519 m .59309 .29721 L .60748 .29675 L p 0 0 .801 r
F P
0 g
s
.60748 .29675 m .62055 .33556 L .60705 .33519 L p 0 0 .801 r
F P
0 g
s
.35212 .36021 m .3665 .34695 L .36913 .33165 L p .2 0 0 r
F P
0 g
s
.3665 .34695 m .35212 .36021 L .35879 .37331 L p .397 0 0 r
F P
0 g
s
.48552 .46372 m .45835 .45638 L .46459 .47234 L p 0 .652 0 r
F P
0 g
s
.45835 .45638 m .48552 .46372 L .47885 .44823 L p 0 .501 0 r
F P
0 g
s
.2165 .31289 m .2035 .34603 L .18691 .34276 L p 0 0 .755 r
F P
0 g
s
.18691 .34276 m .20109 .30632 L .2165 .31289 L p 0 0 .755 r
F P
0 g
s
.20109 .30632 m .18691 .34276 L .17037 .34113 L p 0 0 .774 r
F P
0 g
s
.15315 .38289 m .17037 .34113 L .18691 .34276 L p 0 0 .72 r
F P
0 g
s
.65997 .35963 m .67939 .36455 L .68626 .3799 L p .923 0 0 r
F P
0 g
s
.67939 .36455 m .65997 .35963 L .66163 .34365 L p .969 0 0 r
F P
0 g
s
.66163 .34365 m .67837 .34768 L .67939 .36455 L p .969 0 0 r
F P
0 g
s
.5448 .52003 m .53031 .53696 L .53831 .55175 L p 0 0 .74 r
F P
0 g
s
.53031 .53696 m .5448 .52003 L .5387 .50868 L p 0 0 .597 r
F P
0 g
s
.36506 .40047 m .37052 .38512 L .35879 .37331 L p .564 0 0 r
F P
0 g
s
.55196 .35117 m .56146 .38115 L .55437 .38418 L p 0 0 .588 r
F P
0 g
s
.56146 .38115 m .55196 .35117 L .56072 .34573 L p 0 0 .705 r
F P
0 g
s
.17322 .54278 m .15625 .5101 L .15615 .50358 L p 0 0 .085 r
F P
0 g
s
.15625 .5101 m .17322 .54278 L .18178 .54828 L p 0 0 .755 r
F P
0 g
s
.22819 .32095 m .21742 .3504 L .2035 .34603 L p 0 0 .713 r
F P
0 g
s
.2035 .34603 m .2165 .31289 L .22819 .32095 L p 0 0 .713 r
F P
0 g
s
.24422 .5517 m .22415 .53421 L .20972 .54362 L p 0 0 .838 r
F P
0 g
s
.22862 .35948 m .2231 .38893 L .21435 .3862 L p 0 0 .26 r
F P
0 g
s
.22632 .35515 m .22862 .35948 L p .22152 .37278 L 0 0 .26 r
F P
0 g
s
.2204 .3705 m .22632 .35515 L p .22152 .37278 L 0 0 .26 r
F P
0 g
s
.22862 .35948 m .22381 .36265 L p .22346 .37555 L 0 0 .597 r
F P
0 g
s
.2231 .38893 m .22596 .37367 L p .22346 .37555 L 0 0 .597 r
F P
0 g
s
.22596 .37367 m .22862 .35948 L p .22346 .37555 L 0 0 .597 r
F P
0 g
s
.22515 .39167 m .2231 .38893 L p .22343 .37666 L 0 0 .597 r
F P
0 g
s
.22004 .39394 m .22515 .39167 L p .21925 .37871 L 0 0 .774 r
F P
0 g
s
.21435 .3862 m .2204 .3705 L p .22152 .37278 L 0 0 .26 r
F P
0 g
s
.21259 .36406 m .21671 .38061 L p .21925 .37871 L 0 0 .774 r
F P
0 g
s
.21671 .38061 m .22004 .39394 L p .21925 .37871 L 0 0 .774 r
F P
0 g
s
.22381 .36265 m .22456 .37897 L p .22343 .37666 L 0 0 .597 r
F P
0 g
s
.22456 .37897 m .22515 .39167 L p .22343 .37666 L 0 0 .597 r
F P
0 g
s
.47885 .44823 m .44762 .44383 L .45835 .45638 L p 0 .501 0 r
F P
0 g
s
.55415 .38705 m .55804 .41737 L .55401 .41968 L p 0 0 .495 r
F P
0 g
s
.55401 .41968 m .55705 .42225 L p .5573 .42017 L 0 0 .354 r
F P
0 g
s
.55804 .41737 m .55554 .3979 L p .55467 .38688 L 0 0 .712 r
F P
0 g
s
.56095 .38925 m .55642 .40908 L p .5573 .42017 L 0 0 .354 r
F P
0 g
s
.55642 .40908 m .55401 .41968 L p .5573 .42017 L 0 0 .354 r
F P
0 g
s
.55415 .38705 m .55437 .38418 L p .55467 .38688 L 0 0 .712 r
F P
0 g
s
.55554 .3979 m .55415 .38705 L p .55467 .38688 L 0 0 .712 r
F P
0 g
s
.45292 .38709 m .43137 .38692 L .44278 .37661 L p .419 0 0 r
F P
0 g
s
.44278 .37661 m .46381 .37453 L .45292 .38709 L p .419 0 0 r
F P
0 g
s
.5504 .56338 m .57108 .53963 L .55627 .53101 L p 0 0 .823 r
F P
0 g
s
.55627 .53101 m .53831 .55175 L .5504 .56338 L p 0 0 .823 r
F P
0 g
s
.44724 .32886 m .46584 .32768 L .47088 .3421 L p .283 0 0 r
F P
0 g
s
.46584 .32768 m .44724 .32886 L .43797 .31635 L p .125 0 0 r
F P
0 g
s
.67837 .34768 m .66163 .34365 L .66909 .33079 L p .954 0 0 r
F P
0 g
s
.66909 .33079 m .68338 .3321 L .67837 .34768 L p .954 0 0 r
F P
0 g
s
.65219 .47876 m .6769 .47492 L .68392 .45941 L p 0 .411 0 r
F P
0 g
s
.68392 .45941 m .65663 .46275 L .65219 .47876 L p 0 .411 0 r
F P
0 g
s
.69871 .52243 m .72303 .52561 L .71075 .51885 L p 0 .618 0 r
F P
0 g
s
.71075 .51885 m .68703 .51776 L .69871 .52243 L p 0 .618 0 r
F P
0 g
s
.66435 .37602 m .68626 .3799 L .69784 .39111 L p .813 0 0 r
F P
0 g
s
.68626 .3799 m .66435 .37602 L .65997 .35963 L p .923 0 0 r
F P
0 g
s
.14952 .46603 m .15615 .50358 L .15625 .5101 L p 0 0 .348 r
F P
0 g
s
.77283 .37578 m .76222 .36862 L .75347 .38322 L p .868 0 0 r
F P
0 g
s
.75347 .38322 m .76184 .38841 L .77283 .37578 L p .868 0 0 r
F P
0 g
s
.53831 .55175 m .55627 .53101 L .5448 .52003 L p 0 0 .74 r
F P
0 g
s
.37729 .32017 m .39782 .31678 L .41094 .30966 L p .064 0 0 r
F P
0 g
s
.39782 .31678 m .37729 .32017 L .36913 .33165 L p .263 0 0 r
F P
0 g
s
.69484 .44798 m .6657 .44849 L .65663 .46275 L p 0 .367 0 r
F P
0 g
s
.6657 .44849 m .69484 .44798 L .7077 .44261 L p 0 .317 0 r
F P
0 g
s
.55012 .31792 m .56072 .34573 L .55196 .35117 L p 0 0 .641 r
F P
0 g
s
.56072 .34573 m .55012 .31792 L .56181 .30876 L p 0 0 .736 r
F P
0 g
s
.41388 .43918 m .43431 .4369 L .42079 .43674 L p 0 0 0 r
F P
0 g
s
.41366 .43855 m .40031 .44194 L .41388 .43918 L p 0 0 0 r
F P
0 g
s
.40294 .43698 m .41388 .43918 L .40031 .44194 L p 0 .857 0 r
F P
0 g
s
.41388 .43918 m .40294 .43698 L .4141 .43982 L p 0 0 0 r
F P
0 g
s
.4141 .43982 m .4159 .44028 L p .41388 .43918 L 0 0 0 r
F P
0 g
s
.42079 .43674 m .41366 .43855 L p .41388 .43918 L 0 0 0 r
F P
0 g
s
.5387 .50868 m .54842 .48753 L .54585 .48122 L p 0 0 .18 r
F P
0 g
s
.54585 .48122 m .55229 .45518 L .55221 .45176 L p 0 0 .524 r
F P
0 g
s
.55221 .45176 m .55705 .42225 L .55401 .41968 L p 0 0 .657 r
F P
0 g
s
.45037 .5124 m .43084 .51528 L .41792 .52123 L p 0 .9 0 r
F P
0 g
s
.43084 .51528 m .45037 .5124 L .46017 .50304 L p 0 .96 0 r
F P
0 g
s
.65663 .46275 m .68392 .45941 L .69484 .44798 L p 0 .367 0 r
F P
0 g
s
.2268 .45575 m .22405 .42699 L .21686 .42968 L p 0 0 .663 r
F P
0 g
s
.40365 .38779 m .39155 .38821 L .40588 .39227 L p .338 0 0 r
F P
0 g
s
.17037 .34113 m .18453 .30233 L .20109 .30632 L p 0 0 .774 r
F P
0 g
s
.18453 .30233 m .17037 .34113 L .15665 .34143 L p 0 0 .778 r
F P
0 g
s
.17037 .34113 m .15315 .38289 L .14431 .38469 L p 0 0 .704 r
F P
0 g
s
.14431 .38469 m .15665 .34143 L .17037 .34113 L p 0 0 .704 r
F P
0 g
s
.3788 .37798 m .37052 .38512 L .38522 .39351 L p .584 0 0 r
F P
0 g
s
.38522 .39351 m .39155 .38821 L .3788 .37798 L p .584 0 0 r
F P
0 g
s
.36585 .46409 m .38162 .46556 L .37962 .48235 L p 0 .778 0 r
F P
0 g
s
.37962 .48235 m .36106 .48027 L .36585 .46409 L p 0 .778 0 r
F P
0 g
s
.36106 .48027 m .37962 .48235 L .38324 .4988 L p 0 .665 0 r
F P
0 g
s
.35879 .37331 m .36992 .36336 L .3665 .34695 L p .397 0 0 r
F P
0 g
s
.36992 .36336 m .35879 .37331 L .37052 .38512 L p .505 0 0 r
F P
0 g
s
.7528 .32364 m .73582 .32452 L .7455 .33824 L p .849 0 0 r
F P
0 g
s
.7455 .33824 m .76183 .33594 L .7528 .32364 L p .849 0 0 r
F P
0 g
s
.54842 .48753 m .5387 .50868 L .5448 .52003 L p 0 0 .61 r
F P
0 g
s
.23419 .32917 m .22632 .35515 L .21742 .3504 L p 0 0 .582 r
F P
0 g
s
.21742 .3504 m .22819 .32095 L .23419 .32917 L p 0 0 .582 r
F P
0 g
s
.77228 .46785 m .75641 .47224 L .74934 .45756 L p 0 .509 0 r
F P
0 g
s
.75641 .47224 m .77228 .46785 L .77825 .48329 L p 0 .381 0 r
F P
0 g
s
.77825 .48329 m .75855 .48911 L .75641 .47224 L p 0 .381 0 r
F P
0 g
s
.68338 .3321 m .66909 .33079 L .68114 .3233 L p .767 0 0 r
F P
0 g
s
.68114 .3233 m .69362 .32049 L .68338 .3321 L p .767 0 0 r
F P
0 g
s
.75855 .48911 m .77825 .48329 L .77872 .49798 L p 0 .224 0 r
F P
0 g
s
.77872 .49798 m .75538 .50532 L .75855 .48911 L p 0 .224 0 r
F P
0 g
s
.23401 .48335 m .22745 .45947 L .22109 .46401 L p 0 0 .558 r
F P
0 g
s
.67402 .39008 m .69784 .39111 L .7122 .39627 L p .542 0 0 r
F P
0 g
s
.69784 .39111 m .67402 .39008 L .66435 .37602 L p .813 0 0 r
F P
0 g
s
.37561 .45003 m .38892 .45128 L .38162 .46556 L p 0 .856 0 r
F P
0 g
s
.38162 .46556 m .36585 .46409 L .37561 .45003 L p 0 .856 0 r
F P
0 g
s
.43929 .3942 m .41754 .3909 L .43137 .38692 L p .416 0 0 r
F P
0 g
s
.43137 .38692 m .45292 .38709 L .43929 .3942 L p .416 0 0 r
F P
0 g
s
.22194 .49881 m .23541 .52206 L .24152 .50935 L p 0 0 .646 r
F P
0 g
s
.24152 .50935 m .23118 .49064 L .22194 .49881 L p 0 0 .646 r
F P
0 g
s
.43797 .31635 m .45583 .31786 L .46584 .32768 L p .125 0 0 r
F P
0 g
s
.45583 .31786 m .43797 .31635 L .42514 .30953 L p .196 0 0 r
F P
0 g
s
.38324 .4988 m .36207 .49583 L .36106 .48027 L p 0 .665 0 r
F P
0 g
s
.36207 .49583 m .38324 .4988 L .3919 .51215 L p 0 .46 0 r
F P
0 g
s
.37052 .38512 m .3788 .37798 L .36992 .36336 L p .505 0 0 r
F P
0 g
s
.55974 .2724 m .57697 .30145 L .56181 .30876 L p 0 0 .815 r
F P
0 g
s
.57697 .30145 m .55974 .2724 L .5749 .26587 L p 0 0 .86 r
F P
0 g
s
.5749 .26587 m .59309 .29721 L .57697 .30145 L p 0 0 .86 r
F P
0 g
s
.63173 .37784 m .62055 .33556 L .62896 .3381 L p 0 0 .49 r
F P
0 g
s
.62055 .33556 m .60748 .29675 L .6177 .30019 L p 0 0 .75 r
F P
0 g
s
.6177 .30019 m .62896 .3381 L .62055 .33556 L p 0 0 .75 r
F P
0 g
s
.62896 .3381 m .63559 .38036 L .63173 .37784 L p 0 0 .49 r
F P
0 g
s
.62367 .50471 m .6326 .46364 L .62968 .46663 L p 0 0 .693 r
F P
0 g
s
.62968 .46663 m .6166 .50813 L .62367 .50471 L p 0 0 .693 r
F P
0 g
s
.76186 .45422 m .74934 .45756 L .7385 .44754 L p 0 .643 0 r
F P
0 g
s
.74934 .45756 m .76186 .45422 L .77228 .46785 L p 0 .509 0 r
F P
0 g
s
.59996 .54434 m .6166 .50813 L .60397 .5087 L p 0 0 .834 r
F P
0 g
s
.60397 .5087 m .58657 .54435 L .59996 .54434 L p 0 0 .834 r
F P
0 g
s
.75538 .50532 m .77872 .49798 L .77357 .50938 L p 0 0 0 r
F P
0 g
s
.77357 .50938 m .74743 .51814 L .75538 .50532 L p 0 0 0 r
F P
0 g
s
.54659 .2835 m .56181 .30876 L .55012 .31792 L p 0 0 .744 r
F P
0 g
s
.56181 .30876 m .54659 .2835 L .55974 .2724 L p 0 0 .815 r
F P
0 g
s
.23029 .28533 m .2165 .31289 L .20109 .30632 L p 0 0 .796 r
F P
0 g
s
.20109 .30632 m .21662 .27489 L .23029 .28533 L p 0 0 .796 r
F P
0 g
s
.21662 .27489 m .20109 .30632 L .18453 .30233 L p 0 0 .833 r
F P
0 g
s
.68703 .51776 m .71075 .51885 L .70135 .50615 L p 0 .551 0 r
F P
0 g
s
.70135 .50615 m .67863 .50692 L .68703 .51776 L p 0 .551 0 r
F P
0 g
s
.18178 .54828 m .16306 .51409 L .15625 .5101 L p 0 0 .755 r
F P
0 g
s
.16306 .51409 m .18178 .54828 L .1947 .5486 L p 0 0 .867 r
F P
0 g
s
.22418 .42436 m .22004 .39394 L .20851 .39534 L p 0 0 .81 r
F P
0 g
s
.23954 .29805 m .22819 .32095 L .2165 .31289 L p 0 0 .729 r
F P
0 g
s
.2165 .31289 m .23029 .28533 L .23954 .29805 L p 0 0 .729 r
F P
0 g
s
.3887 .44044 m .40031 .44194 L .38892 .45128 L p 0 .908 0 r
F P
0 g
s
.38892 .45128 m .37561 .45003 L .3887 .44044 L p 0 .908 0 r
F P
0 g
s
.36913 .33165 m .38811 .32971 L .39782 .31678 L p .263 0 0 r
F P
0 g
s
.38811 .32971 m .36913 .33165 L .3665 .34695 L p .346 0 0 r
F P
0 g
s
.59309 .29721 m .5749 .26587 L .58953 .26501 L p 0 0 .888 r
F P
0 g
s
.58953 .26501 m .60748 .29675 L .59309 .29721 L p 0 0 .888 r
F P
0 g
s
.4273 .44356 m .44762 .44383 L .43431 .4369 L p 0 .041 0 r
F P
0 g
s
.43431 .4369 m .41388 .43918 L .4273 .44356 L p 0 .041 0 r
F P
0 g
s
.76183 .33594 m .7455 .33824 L .74962 .35502 L p .931 0 0 r
F P
0 g
s
.74962 .35502 m .76519 .35187 L .76183 .33594 L p .931 0 0 r
F P
0 g
s
.46017 .50304 m .44064 .5032 L .43084 .51528 L p 0 .96 0 r
F P
0 g
s
.44064 .5032 m .46017 .50304 L .46522 .48887 L p 0 .853 0 r
F P
0 g
s
.54971 .35614 m .55437 .38418 L .55415 .38705 L p 0 0 0 r
F P
0 g
s
.54971 .35614 m .55196 .35117 L p .55375 .37572 L 0 0 .588 r
F P
0 g
s
.5439 .32741 m .55196 .35117 L .54971 .35614 L p 0 0 .273 r
F P
0 g
s
.54429 .33563 m .5439 .32741 L p .54703 .3429 L 0 0 .273 r
F P
0 g
s
.54971 .35614 m .54429 .33563 L p .54703 .3429 L 0 0 .273 r
F P
0 g
s
.55196 .35117 m .5439 .32741 L .55012 .31792 L p 0 0 .641 r
F P
0 g
s
.53442 .31148 m .5439 .32741 L .54429 .33563 L p 0 0 .145 r
F P
0 g
s
.54429 .33563 m .55132 .34112 L p .54182 .32907 L 0 0 .704 r
F P
0 g
s
.53754 .32365 m .53442 .31148 L p .53993 .32495 L 0 0 .145 r
F P
0 g
s
.54429 .33563 m .53754 .32365 L p .53993 .32495 L 0 0 .145 r
F P
0 g
s
.53754 .32365 m .54429 .33563 L p .54182 .32907 L 0 0 .704 r
F P
0 g
s
.54653 .3317 m .53754 .32365 L p .54542 .33364 L 0 0 .704 r
F P
0 g
s
.55132 .34112 m .54653 .3317 L p .54542 .33364 L 0 0 .704 r
F P
0 g
s
.54653 .3317 m .55992 .33416 L p .54766 .32971 L 0 0 .949 r
F P
0 g
s
.53754 .32365 m .54653 .3317 L p .5348 .32217 L 0 0 0 r
F P
0 g
s
.53879 .3265 m .52796 .3166 L p .53986 .32628 L 0 0 0 r
F P
0 g
s
.54653 .3317 m .53879 .3265 L p .53986 .32628 L 0 0 0 r
F P
0 g
s
.53879 .3265 m .54653 .3317 L p .54766 .32971 L 0 0 .949 r
F P
0 g
s
.5528 .32986 m .53879 .3265 L p .55289 .33161 L 0 0 .949 r
F P
0 g
s
.55992 .33416 m .5528 .32986 L p .55289 .33161 L 0 0 .949 r
F P
0 g
s
.5528 .32986 m .5676 .32604 L p .55269 .32761 L 0 0 .965 r
F P
0 g
s
.53879 .3265 m .5528 .32986 L p .53732 .32679 L 0 0 0 r
F P
0 g
s
.53787 .32917 m .52362 .32407 L p .54634 .32858 L 0 0 0 r
F P
0 g
s
.5528 .32986 m .53787 .32917 L p .54634 .32858 L 0 0 0 r
F P
0 g
s
.52362 .32407 m .53787 .32917 L p .52167 .32647 L 0 0 0 r
F P
0 g
s
.49747 .32244 m .51689 .32365 L p .52167 .32647 L 0 0 0 r
F P
0 g
s
.51689 .32365 m .52362 .32407 L p .52167 .32647 L 0 0 0 r
F P
0 g
s
.53787 .32917 m .51085 .33196 L .49747 .32244 L p 0 0 0 r
F P
0 g
s
.51085 .33196 m .53787 .32917 L p .52877 .32981 L 0 0 .777 r
F P
0 g
s
.52548 .33483 m .51085 .33196 L p .52757 .32995 L 0 0 .777 r
F P
0 g
s
.53787 .32917 m .55238 .32698 L p .52877 .32981 L 0 0 .777 r
F P
0 g
s
.53787 .32917 m .5528 .32986 L p .55269 .32761 L 0 0 .965 r
F P
0 g
s
.55238 .32698 m .53787 .32917 L p .56284 .32654 L 0 0 .965 r
F P
0 g
s
.5676 .32604 m .55238 .32698 L p .56284 .32654 L 0 0 .965 r
F P
0 g
s
.53764 .29729 m .55012 .31792 L .5439 .32741 L p 0 0 .597 r
F P
0 g
s
.5439 .32741 m .53442 .31148 L .53764 .29729 L p 0 0 .597 r
F P
0 g
s
.55238 .32698 m .56463 .31778 L p .54258 .32739 L 0 0 .945 r
F P
0 g
s
.55238 .32698 m .53241 .33281 L p .52757 .32995 L 0 0 .777 r
F P
0 g
s
.53241 .33281 m .52548 .33483 L p .52757 .32995 L 0 0 .777 r
F P
0 g
s
.52548 .33483 m .54649 .3287 L p .54258 .32739 L 0 0 .945 r
F P
0 g
s
.54649 .3287 m .5475 .32718 L .54258 .32739 L 0 0 .945 r
F
.56463 .31778 m .53888 .3305 L .52548 .33483 L p F P
0 g
s
.54871 .31961 m .53888 .3305 L p .55289 .31907 L 0 0 .85 r
F P
0 g
s
.57246 .30311 m .55582 .31467 L p .55289 .31907 L 0 0 .85 r
F P
0 g
s
.55582 .31467 m .54871 .31961 L p .55289 .31907 L 0 0 .85 r
F P
0 g
s
.54649 .3287 m .55238 .32698 L p .5475 .32718 L 0 0 .945 r
F P
0 g
s
.52362 .32407 m .53879 .3265 L p .53732 .32679 L 0 0 0 r
F P
0 g
s
.52796 .3166 m .53754 .32365 L p .5348 .32217 L 0 0 0 r
F P
0 g
s
.54971 .35614 m .55448 .35978 L p .5466 .34111 L 0 0 .6 r
F P
0 g
s
.54429 .33563 m .54971 .35614 L p .5466 .34111 L 0 0 .6 r
F P
0 g
s
.55437 .38418 m .54971 .35614 L p .55375 .37572 L 0 0 .588 r
F P
0 g
s
.39211 .37805 m .3788 .37798 L .39155 .38821 L p .405 0 0 r
F P
0 g
s
.39155 .38821 m .40365 .38779 L .39211 .37805 L p .405 0 0 r
F P
0 g
s
.3919 .51215 m .36877 .50814 L .36207 .49583 L p 0 .46 0 r
F P
0 g
s
.36877 .50814 m .3919 .51215 L .40414 .5201 L p 0 0 0 r
F P
0 g
s
.69362 .32049 m .68114 .3233 L .69575 .32254 L p 0 0 0 r
F P
0 g
s
.69575 .32254 m .70738 .31486 L .69362 .32049 L p 0 0 0 r
F P
0 g
s
.18453 .30233 m .20083 .26846 L .21662 .27489 L p 0 0 .833 r
F P
0 g
s
.20083 .26846 m .18453 .30233 L .16957 .3016 L p 0 0 .853 r
F P
0 g
s
.15665 .34143 m .16957 .3016 L .18453 .30233 L p 0 0 .778 r
F P
0 g
s
.7122 .39627 m .68736 .39943 L .67402 .39008 L p .542 0 0 r
F P
0 g
s
.55401 .41968 m .56095 .38925 L .55415 .38705 L p 0 0 .495 r
F P
0 g
s
.73937 .3858 m .75347 .38322 L .76222 .36862 L p .895 0 0 r
F P
0 g
s
.75347 .38322 m .73937 .3858 L .72689 .39442 L p .763 0 0 r
F P
0 g
s
.41754 .3909 m .43929 .3942 L .42522 .39458 L p .306 0 0 r
F P
0 g
s
.1461 .47029 m .15625 .5101 L .16306 .51409 L p 0 0 .532 r
F P
0 g
s
.16957 .3016 m .15665 .34143 L .1481 .34362 L p 0 0 .751 r
F P
0 g
s
.14221 .38721 m .1481 .34362 L .15665 .34143 L p 0 0 .632 r
F P
0 g
s
.14431 .38469 m .14221 .38721 L p .14945 .36427 L 0 0 .632 r
F P
0 g
s
.1504 .36335 m .14431 .38469 L p .14945 .36427 L 0 0 .632 r
F P
0 g
s
.14221 .38721 m .14734 .39005 L p .14772 .36668 L 0 0 .231 r
F P
0 g
s
.1481 .34362 m .14515 .36542 L p .14772 .36668 L 0 0 .231 r
F P
0 g
s
.14515 .36542 m .14221 .38721 L p .14772 .36668 L 0 0 .231 r
F P
0 g
s
.14629 .34738 m .1481 .34362 L p .14773 .36591 L 0 0 .231 r
F P
0 g
s
.15665 .34143 m .1504 .36335 L p .14945 .36427 L 0 0 .632 r
F P
0 g
s
.15166 .35206 m .14629 .34738 L p .15229 .36892 L 0 0 .584 r
F P
0 g
s
.15892 .39269 m .1549 .37019 L p .15229 .36892 L 0 0 .584 r
F P
0 g
s
.1549 .37019 m .15166 .35206 L p .15229 .36892 L 0 0 .584 r
F P
0 g
s
.14734 .39005 m .14677 .36684 L p .14773 .36591 L 0 0 .231 r
F P
0 g
s
.14677 .36684 m .14629 .34738 L p .14773 .36591 L 0 0 .231 r
F P
0 g
s
.55747 .49465 m .5448 .52003 L .55627 .53101 L p 0 0 .727 r
F P
0 g
s
.5448 .52003 m .55747 .49465 L .54842 .48753 L p 0 0 .61 r
F P
0 g
s
.7385 .44754 m .74872 .4447 L .76186 .45422 L p 0 .643 0 r
F P
0 g
s
.42514 .30953 m .44251 .3144 L .45583 .31786 L p .196 0 0 r
F P
0 g
s
.44251 .3144 m .42514 .30953 L .41094 .30966 L p .413 0 0 r
F P
0 g
s
.74743 .51814 m .77357 .50938 L .7636 .51551 L p 0 0 0 r
F P
0 g
s
.7636 .51551 m .736 .52533 L .74743 .51814 L p 0 0 0 r
F P
0 g
s
.55012 .31792 m .53764 .29729 L .54659 .2835 L p 0 0 .744 r
F P
0 g
s
.20783 .50642 m .22415 .53421 L .23541 .52206 L p 0 0 .744 r
F P
0 g
s
.23541 .52206 m .22194 .49881 L .20783 .50642 L p 0 0 .744 r
F P
0 g
s
.23339 .33616 m .22862 .35948 L .22632 .35515 L p 0 0 0 r
F P
0 g
s
.2258 .34068 m .22381 .36265 L .22638 .36095 L p 0 0 .709 r
F P
0 g
s
.22638 .36095 m .22862 .35948 L p .22794 .35494 L 0 0 .709 r
F P
0 g
s
.22632 .35515 m .23419 .32917 L .23386 .33206 L p 0 0 0 r
F P
0 g
s
.22632 .35515 m .23386 .33206 L .23146 .34135 L 0 0 0 r
F
.22794 .35494 m .2258 .34068 L .22638 .36095 L 0 0 .709 r
F
.23386 .33206 m .23339 .33616 L p .23146 .34135 L 0 0 0 r
F P
0 g
s
.76519 .35187 m .74962 .35502 L .74744 .37186 L p .939 0 0 r
F P
0 g
s
.74744 .37186 m .76222 .36862 L .76519 .35187 L p .939 0 0 r
F P
0 g
s
.3665 .34695 m .38355 .3462 L .38811 .32971 L p .346 0 0 r
F P
0 g
s
.38355 .3462 m .3665 .34695 L .36992 .36336 L p .386 0 0 r
F P
0 g
s
.55929 .45923 m .54842 .48753 L .55747 .49465 L p 0 0 .652 r
F P
0 g
s
.54842 .48753 m .55929 .45923 L .55229 .45518 L p 0 0 .339 r
F P
0 g
s
.76222 .36862 m .74744 .37186 L .73937 .3858 L p .895 0 0 r
F P
0 g
s
.38498 .36333 m .36992 .36336 L .3788 .37798 L p .406 0 0 r
F P
0 g
s
.3788 .37798 m .39211 .37805 L .38498 .36333 L p .406 0 0 r
F P
0 g
s
.2428 .31093 m .23419 .32917 L .22819 .32095 L p 0 0 .557 r
F P
0 g
s
.22819 .32095 m .23954 .29805 L .2428 .31093 L p 0 0 .557 r
F P
0 g
s
.40031 .44194 m .3887 .44044 L .40294 .43698 L p 0 .857 0 r
F P
0 g
s
.56679 .42465 m .55229 .45518 L .55929 .45923 L p 0 0 .575 r
F P
0 g
s
.55229 .45518 m .56679 .42465 L .55705 .42225 L p 0 0 0 r
F P
0 g
s
.58657 .54435 m .60397 .5087 L .58797 .50629 L p 0 0 .814 r
F P
0 g
s
.58797 .50629 m .57108 .53963 L .58657 .54435 L p 0 0 .814 r
F P
0 g
s
.67863 .50692 m .70135 .50615 L .69651 .48974 L p 0 .45 0 r
F P
0 g
s
.69651 .48974 m .67503 .49175 L .67863 .50692 L p 0 .45 0 r
F P
0 g
s
.48611 .38104 m .45292 .38709 L .46381 .37453 L p .459 0 0 r
F P
0 g
s
.46381 .37453 m .49213 .36503 L .48611 .38104 L p .459 0 0 r
F P
0 g
s
.49213 .36503 m .46381 .37453 L .47016 .35865 L p .382 0 0 r
F P
0 g
s
.43823 .45438 m .45835 .45638 L .44762 .44383 L p 0 .398 0 r
F P
0 g
s
.44762 .44383 m .4273 .44356 L .43823 .45438 L p 0 .398 0 r
F P
0 g
s
.20374 .43197 m .22109 .46401 L .22745 .45947 L p 0 0 .756 r
F P
0 g
s
.22745 .45947 m .21686 .42968 L .20374 .43197 L p 0 0 .756 r
F P
0 g
s
.1947 .5486 m .17548 .51481 L .16306 .51409 L p 0 0 .867 r
F P
0 g
s
.17548 .51481 m .1947 .5486 L .20972 .54362 L p 0 0 .84 r
F P
0 g
s
.46522 .48887 m .44556 .48709 L .44064 .5032 L p 0 .853 0 r
F P
0 g
s
.44556 .48709 m .46522 .48887 L .46459 .47234 L p 0 .732 0 r
F P
0 g
s
.47016 .35865 m .49259 .34917 L .49213 .36503 L p .382 0 0 r
F P
0 g
s
.49259 .34917 m .47016 .35865 L .47088 .3421 L p .243 0 0 r
F P
0 g
s
.36992 .36336 m .38498 .36333 L .38355 .3462 L p .386 0 0 r
F P
0 g
s
.2087 .46857 m .22194 .49881 L .23118 .49064 L p 0 0 .703 r
F P
0 g
s
.23118 .49064 m .22109 .46401 L .2087 .46857 L p 0 0 .703 r
F P
0 g
s
.71314 .44746 m .69484 .44798 L .68392 .45941 L p 0 .163 0 r
F P
0 g
s
.60748 .29675 m .58953 .26501 L .60114 .27006 L p 0 0 .858 r
F P
0 g
s
.60114 .27006 m .6177 .30019 L .60748 .29675 L p 0 0 .858 r
F P
0 g
s
.40414 .5201 m .38005 .51503 L .36877 .50814 L p 0 0 0 r
F P
0 g
s
.38005 .51503 m .40414 .5201 L .41792 .52123 L p 0 .685 0 r
F P
0 g
s
.16306 .51409 m .14978 .47349 L .1461 .47029 L p 0 0 .532 r
F P
0 g
s
.63559 .38036 m .62896 .3381 L .63077 .3424 L p 0 0 .375 r
F P
0 g
s
.62896 .3381 m .6177 .30019 L .62193 .307 L p 0 0 .271 r
F P
0 g
s
.62193 .307 m .63077 .3424 L .62896 .3381 L p 0 0 .271 r
F P
0 g
s
.77035 .44358 m .76186 .45422 L .74872 .4447 L p 0 .652 0 r
F P
0 g
s
.76186 .45422 m .77035 .44358 L .78422 .45352 L p 0 .531 0 r
F P
0 g
s
.78422 .45352 m .77228 .46785 L .76186 .45422 L p 0 .531 0 r
F P
0 g
s
.70738 .31486 m .69575 .32254 L .71038 .32873 L p .443 0 0 r
F P
0 g
s
.71038 .32873 m .72231 .31626 L .70738 .31486 L p .443 0 0 r
F P
0 g
s
.77228 .46785 m .78422 .45352 L .79425 .46483 L p 0 .386 0 r
F P
0 g
s
.79425 .46483 m .77825 .48329 L .77228 .46785 L p 0 .386 0 r
F P
0 g
s
.57148 .5013 m .55627 .53101 L .57108 .53963 L p 0 0 .78 r
F P
0 g
s
.55627 .53101 m .57148 .5013 L .55747 .49465 L p 0 0 .727 r
F P
0 g
s
.63965 .36201 m .66435 .37602 L .67402 .39008 L p .905 0 0 r
F P
0 g
s
.66435 .37602 m .63965 .36201 L .64355 .34779 L p .988 0 0 r
F P
0 g
s
.64355 .34779 m .65997 .35963 L .66435 .37602 L p .988 0 0 r
F P
0 g
s
.19136 .5121 m .20972 .54362 L .22415 .53421 L p 0 0 .799 r
F P
0 g
s
.22415 .53421 m .20783 .50642 L .19136 .5121 L p 0 0 .799 r
F P
0 g
s
.47555 .39453 m .43929 .3942 L .45292 .38709 L p .516 0 0 r
F P
0 g
s
.45292 .38709 m .48611 .38104 L .47555 .39453 L p .516 0 0 r
F P
0 g
s
.6166 .50813 m .62968 .46663 L .62012 .46824 L p 0 0 .773 r
F P
0 g
s
.62012 .46824 m .60397 .5087 L .6166 .50813 L p 0 0 .773 r
F P
0 g
s
.44471 .4698 m .46459 .47234 L .45835 .45638 L p 0 .594 0 r
F P
0 g
s
.45835 .45638 m .43823 .45438 L .44471 .4698 L p 0 .594 0 r
F P
0 g
s
.46459 .47234 m .44471 .4698 L .44556 .48709 L p 0 .732 0 r
F P
0 g
s
.57108 .53963 m .58797 .50629 L .57148 .5013 L p 0 0 .78 r
F P
0 g
s
.63559 .38036 m .62826 .42574 L .63526 .42364 L p 0 0 .587 r
F P
0 g
s
.63077 .3424 m .63233 .38339 L .63559 .38036 L p 0 0 .375 r
F P
0 g
s
.20972 .54362 m .19136 .5121 L .17548 .51481 L p 0 0 .84 r
F P
0 g
s
.16957 .3016 m .18553 .26715 L .20083 .26846 L p 0 0 .853 r
F P
0 g
s
.18553 .26715 m .16957 .3016 L .15874 .30428 L p 0 0 .841 r
F P
0 g
s
.1481 .34362 m .15874 .30428 L .16957 .3016 L p 0 0 .751 r
F P
0 g
s
.47088 .3421 m .48736 .33613 L .49259 .34917 L p .243 0 0 r
F P
0 g
s
.48736 .33613 m .47088 .3421 L .46584 .32768 L p 0 0 0 r
F P
0 g
s
.64167 .37825 m .67402 .39008 L .68736 .39943 L p .713 0 0 r
F P
0 g
s
.67402 .39008 m .64167 .37825 L .63965 .36201 L p .905 0 0 r
F P
0 g
s
.41312 .38811 m .39211 .37805 L .40365 .38779 L p .014 0 0 r
F P
0 g
s
.40365 .38779 m .42522 .39458 L .41312 .38811 L p .014 0 0 r
F P
0 g
s
.63526 .42364 m .62012 .46824 L .62968 .46663 L p 0 0 .685 r
F P
0 g
s
.67503 .49175 m .69651 .48974 L .6971 .47252 L p 0 .363 0 r
F P
0 g
s
.6971 .47252 m .6769 .47492 L .67503 .49175 L p 0 .363 0 r
F P
0 g
s
.65997 .35963 m .64355 .34779 L .65278 .33803 L p .974 0 0 r
F P
0 g
s
.65278 .33803 m .66163 .34365 L .65997 .35963 L p .974 0 0 r
F P
0 g
s
.3672 .4417 m .37561 .45003 L .36585 .46409 L p 0 .781 0 r
F P
0 g
s
.36585 .46409 m .35418 .45267 L .3672 .4417 L p 0 .781 0 r
F P
0 g
s
.35418 .45267 m .36585 .46409 L .36106 .48027 L p 0 .762 0 r
F P
0 g
s
.62012 .46824 m .63526 .42364 L .62826 .42574 L p 0 0 .729 r
F P
0 g
s
.736 .52533 m .7636 .51551 L .75049 .5152 L p 0 .859 0 r
F P
0 g
s
.75049 .5152 m .72303 .52561 L .736 .52533 L p 0 .859 0 r
F P
0 g
s
.703 .45755 m .68392 .45941 L .6769 .47492 L p 0 .276 0 r
F P
0 g
s
.68392 .45941 m .703 .45755 L .71314 .44746 L p 0 .163 0 r
F P
0 g
s
.41094 .30966 m .42818 .31798 L .44251 .3144 L p .413 0 0 r
F P
0 g
s
.42818 .31798 m .41094 .30966 L .39782 .31678 L p .444 0 0 r
F P
0 g
s
.75494 .43664 m .74872 .4447 L .73506 .4409 L p 0 .784 0 r
F P
0 g
s
.74872 .4447 m .75494 .43664 L .77035 .44358 L p 0 .652 0 r
F P
0 g
s
.38249 .43421 m .3887 .44044 L .37561 .45003 L p 0 .776 0 r
F P
0 g
s
.37561 .45003 m .3672 .4417 L .38249 .43421 L p 0 .776 0 r
F P
0 g
s
.19254 .3956 m .21686 .42968 L .22405 .42699 L p 0 0 .824 r
F P
0 g
s
.22405 .42699 m .20851 .39534 L .19254 .3956 L p 0 0 .824 r
F P
0 g
s
.22515 .39167 m .22381 .36265 L .21259 .36406 L p 0 0 .774 r
F P
0 g
s
.14195 .42873 m .14978 .47349 L .16006 .47504 L p 0 0 .476 r
F P
0 g
s
.6769 .47492 m .6971 .47252 L .703 .45755 L p 0 .276 0 r
F P
0 g
s
.77825 .48329 m .79425 .46483 L .7987 .47561 L p 0 .151 0 r
F P
0 g
s
.7987 .47561 m .77872 .49798 L .77825 .48329 L p 0 .151 0 r
F P
0 g
s
.4159 .44028 m .4273 .44356 L .41522 .43962 L p 0 0 0 r
F P
0 g
s
.4273 .44356 m .4159 .44028 L .42534 .44986 L p 0 .261 0 r
F P
0 g
s
.4159 .44028 m .40971 .43396 L .41661 .44096 L p 0 0 0 r
F P
0 g
s
.41661 .44096 m .417 .44135 L p .4159 .44028 L 0 0 0 r
F P
0 g
s
.41522 .43962 m .41388 .43918 L p .4159 .44028 L 0 0 0 r
F P
0 g
s
.55448 .35978 m .55415 .38705 L .56095 .38925 L p 0 0 .662 r
F P
0 g
s
.55415 .38705 m .55448 .35978 L .54971 .35614 L p 0 0 0 r
F P
0 g
s
.14978 .47349 m .16306 .51409 L .17548 .51481 L p 0 0 .788 r
F P
0 g
s
.25311 .28343 m .23954 .29805 L .23029 .28533 L p 0 0 .748 r
F P
0 g
s
.23029 .28533 m .24675 .26734 L .25311 .28343 L p 0 0 .748 r
F P
0 g
s
.24675 .26734 m .23029 .28533 L .21662 .27489 L p 0 0 .848 r
F P
0 g
s
.36106 .48027 m .3456 .46532 L .35418 .45267 L p 0 .762 0 r
F P
0 g
s
.3456 .46532 m .36106 .48027 L .36207 .49583 L p 0 .69 0 r
F P
0 g
s
.21662 .27489 m .23528 .25423 L .24675 .26734 L p 0 0 .848 r
F P
0 g
s
.23528 .25423 m .21662 .27489 L .20083 .26846 L p 0 0 .916 r
F P
0 g
s
.41792 .52123 m .39402 .51523 L .38005 .51503 L p 0 .685 0 r
F P
0 g
s
.39402 .51523 m .41792 .52123 L .43084 .51528 L p 0 .901 0 r
F P
0 g
s
.57212 .46319 m .55747 .49465 L .57148 .5013 L p 0 0 .727 r
F P
0 g
s
.55747 .49465 m .57212 .46319 L .55929 .45923 L p 0 0 .652 r
F P
0 g
s
.52822 .26865 m .54659 .2835 L .53764 .29729 L p 0 0 .752 r
F P
0 g
s
.54659 .2835 m .52822 .26865 L .53911 .25564 L p 0 0 .859 r
F P
0 g
s
.53911 .25564 m .55974 .2724 L .54659 .2835 L p 0 0 .859 r
F P
0 g
s
.5737 .39038 m .55705 .42225 L .56679 .42465 L p 0 0 .655 r
F P
0 g
s
.55705 .42225 m .5737 .39038 L .56095 .38925 L p 0 0 .354 r
F P
0 g
s
.23946 .32178 m .23339 .33616 L .23419 .32917 L p 0 0 0 r
F P
0 g
s
.24723 .31373 m .23946 .32178 L .2428 .31093 L p 0 0 .115 r
F P
0 g
s
.23946 .32178 m .24723 .31373 L .24259 .31743 L p 0 0 .678 r
F P
0 g
s
.23432 .3289 m .2428 .31093 L .23946 .32178 L p 0 0 0 r
F P
0 g
s
.24259 .31743 m .2359 .32277 L p .23946 .32178 L 0 0 .678 r
F P
0 g
s
.23419 .32917 m .23432 .3289 L p .23946 .32178 L 0 0 0 r
F P
0 g
s
.68736 .39943 m .64924 .39377 L .64167 .37825 L p .713 0 0 r
F P
0 g
s
.64924 .39377 m .68736 .39943 L .70215 .40246 L p .235 0 0 r
F P
0 g
s
.43261 .39631 m .42522 .39458 L .42987 .39446 L p .566 0 0 r
F P
0 g
s
.42522 .39458 m .4622 .40322 L .44831 .40561 L p .6 0 0 r
F P
0 g
s
.44831 .40561 m .42012 .39159 L p .42522 .39458 L .6 0 0 r
F P
0 g
s
.42012 .39159 m .41312 .38811 L .41707 .39022 L p .6 0 0 r
F P
0 g
s
.41707 .39022 m .42522 .39458 L p .42012 .39159 L .6 0 0 r
F P
0 g
s
.40513 .37581 m .4297 .39588 L p .42809 .3907 L 0 0 0 r
F P
0 g
s
.41312 .38811 m .44098 .40196 L p .43623 .40121 L .488 0 0 r
F P
0 g
s
.4297 .39588 m .43623 .40121 L .43338 .39753 L p 0 0 0 r
F P
0 g
s
.42987 .39446 m .43929 .3942 L p .43261 .39631 L .566 0 0 r
F P
0 g
s
.43338 .39753 m .42809 .3907 L p .4297 .39588 L 0 0 0 r
F P
0 g
s
.44098 .40196 m .44831 .40561 L .44412 .40408 L p .488 0 0 r
F P
0 g
s
.44412 .40408 m .43623 .40121 L p .44098 .40196 L .488 0 0 r
F P
0 g
s
.4622 .40322 m .43261 .39631 L p .43929 .3942 L .566 0 0 r
F P
0 g
s
.43929 .3942 m .47555 .39453 L .4622 .40322 L p .566 0 0 r
F P
0 g
s
.3975 .43146 m .40294 .43698 L .3887 .44044 L p 0 .736 0 r
F P
0 g
s
.3887 .44044 m .38249 .43421 L .3975 .43146 L p 0 .736 0 r
F P
0 g
s
.15874 .30428 m .1481 .34362 L .14629 .34738 L p 0 0 .538 r
F P
0 g
s
.55974 .2724 m .53911 .25564 L .55309 .24821 L p 0 0 .941 r
F P
0 g
s
.55309 .24821 m .5749 .26587 L .55974 .2724 L p 0 0 .941 r
F P
0 g
s
.72231 .31626 m .71038 .32873 L .72243 .34083 L p .716 0 0 r
F P
0 g
s
.72243 .34083 m .73582 .32452 L .72231 .31626 L p .716 0 0 r
F P
0 g
s
.66163 .34365 m .65278 .33803 L .66582 .33449 L p .65 0 0 r
F P
0 g
s
.66582 .33449 m .66909 .33079 L .66163 .34365 L p .65 0 0 r
F P
0 g
s
.77549 .4228 m .77035 .44358 L .75494 .43664 L p 0 .634 0 r
F P
0 g
s
.77035 .44358 m .77549 .4228 L .79154 .42863 L p 0 .531 0 r
F P
0 g
s
.79154 .42863 m .78422 .45352 L .77035 .44358 L p 0 .531 0 r
F P
0 g
s
.22862 .35948 m .23339 .33616 L .2258 .34068 L p 0 0 .709 r
F P
0 g
s
.46584 .32768 m .47727 .3282 L .48736 .33613 L p 0 0 0 r
F P
0 g
s
.47727 .3282 m .46584 .32768 L .45583 .31786 L p .46 0 0 r
F P
0 g
s
.71585 .39858 m .73937 .3858 L .74744 .37186 L p .877 0 0 r
F P
0 g
s
.19243 .47234 m .20783 .50642 L .22194 .49881 L p 0 0 .753 r
F P
0 g
s
.22194 .49881 m .2087 .46857 L .19243 .47234 L p 0 0 .753 r
F P
0 g
s
.58158 .42644 m .55929 .45923 L .57212 .46319 L p 0 0 .69 r
F P
0 g
s
.55929 .45923 m .58158 .42644 L .56679 .42465 L p 0 0 .575 r
F P
0 g
s
.6177 .30019 m .60114 .27006 L .60766 .2802 L p 0 0 .403 r
F P
0 g
s
.60766 .2802 m .62193 .307 L .6177 .30019 L p 0 0 .403 r
F P
0 g
s
.40513 .37581 m .38498 .36333 L .39211 .37805 L p .244 0 0 r
F P
0 g
s
.39211 .37805 m .41312 .38811 L .40513 .37581 L p .244 0 0 r
F P
0 g
s
.73506 .4409 m .74056 .43387 L .75494 .43664 L p 0 .784 0 r
F P
0 g
s
.75899 .41894 m .75494 .43664 L .74056 .43387 L p 0 .738 0 r
F P
0 g
s
.75494 .43664 m .75899 .41894 L .77549 .4228 L p 0 .634 0 r
F P
0 g
s
.60397 .5087 m .62012 .46824 L .60554 .46814 L p 0 0 .773 r
F P
0 g
s
.60554 .46814 m .58797 .50629 L .60397 .5087 L p 0 0 .773 r
F P
0 g
s
.2533 .29982 m .2428 .31093 L .23954 .29805 L p 0 0 .564 r
F P
0 g
s
.23954 .29805 m .25311 .28343 L .2533 .29982 L p 0 0 .564 r
F P
0 g
s
.52222 .28499 m .53764 .29729 L .53442 .31148 L p 0 0 .588 r
F P
0 g
s
.53764 .29729 m .52222 .28499 L .52822 .26865 L p 0 0 .752 r
F P
0 g
s
.39782 .31678 m .41535 .32805 L .42818 .31798 L p .444 0 0 r
F P
0 g
s
.41535 .32805 m .39782 .31678 L .38811 .32971 L p .427 0 0 r
F P
0 g
s
.18698 .43344 m .2087 .46857 L .22109 .46401 L p 0 0 .779 r
F P
0 g
s
.22109 .46401 m .20374 .43197 L .18698 .43344 L p 0 0 .779 r
F P
0 g
s
.20083 .26846 m .22061 .24634 L .23528 .25423 L p 0 0 .916 r
F P
0 g
s
.22061 .24634 m .20083 .26846 L .18553 .26715 L p 0 0 .958 r
F P
0 g
s
.72303 .52561 m .75049 .5152 L .73652 .50843 L p 0 .728 0 r
F P
0 g
s
.73652 .50843 m .71075 .51885 L .72303 .52561 L p 0 .728 0 r
F P
0 g
s
.78422 .45352 m .79154 .42863 L .80449 .43546 L p 0 .394 0 r
F P
0 g
s
.80449 .43546 m .79425 .46483 L .78422 .45352 L p 0 .394 0 r
F P
0 g
s
.17548 .51481 m .16006 .47504 L .14978 .47349 L p 0 0 .788 r
F P
0 g
s
.5749 .26587 m .55309 .24821 L .56779 .24766 L p 0 0 .988 r
F P
0 g
s
.56779 .24766 m .58953 .26501 L .5749 .26587 L p 0 0 .988 r
F P
0 g
s
.36207 .49583 m .34298 .47752 L .3456 .46532 L p 0 .69 0 r
F P
0 g
s
.34298 .47752 m .36207 .49583 L .36877 .50814 L p 0 .408 0 r
F P
0 g
s
.42534 .44986 m .43823 .45438 L .4273 .44356 L p 0 .261 0 r
F P
0 g
s
.77872 .49798 m .7987 .47561 L .79671 .48401 L p 0 0 0 r
F P
0 g
s
.79671 .48401 m .77357 .50938 L .77872 .49798 L p 0 0 0 r
F P
0 g
s
.37858 .41999 m .38249 .43421 L .3672 .4417 L p 0 .674 0 r
F P
0 g
s
.3672 .4417 m .36193 .42433 L .37858 .41999 L p 0 .674 0 r
F P
0 g
s
.36193 .42433 m .3672 .4417 L .35418 .45267 L p 0 .714 0 r
F P
0 g
s
.63233 .38339 m .63077 .3424 L .62552 .34775 L p 0 0 .675 r
F P
0 g
s
.63077 .3424 m .62193 .307 L .61934 .31606 L p 0 0 .515 r
F P
0 g
s
.61934 .31606 m .62552 .34775 L .63077 .3424 L p 0 0 .515 r
F P
0 g
s
.58851 .46635 m .57148 .5013 L .58797 .50629 L p 0 0 .757 r
F P
0 g
s
.57148 .5013 m .58851 .46635 L .57212 .46319 L p 0 0 .727 r
F P
0 g
s
.15874 .30428 m .17334 .27125 L .18553 .26715 L p 0 0 .841 r
F P
0 g
s
.17334 .27125 m .15874 .30428 L .15398 .30999 L p 0 0 .61 r
F P
0 g
s
.14629 .34738 m .15398 .30999 L .15874 .30428 L p 0 0 .538 r
F P
0 g
s
.14221 .38721 m .15465 .43301 L .16956 .4338 L p 0 0 .661 r
F P
0 g
s
.62826 .42574 m .60554 .46814 L .62012 .46824 L p 0 0 .729 r
F P
0 g
s
.43084 .51528 m .40824 .50864 L .39402 .51523 L p 0 .901 0 r
F P
0 g
s
.40824 .50864 m .43084 .51528 L .44064 .5032 L p 0 .884 0 r
F P
0 g
s
.73582 .32452 m .72243 .34083 L .72976 .35672 L p .837 0 0 r
F P
0 g
s
.72976 .35672 m .7455 .33824 L .73582 .32452 L p .837 0 0 r
F P
0 g
s
.40971 .43396 m .4159 .44028 L .40294 .43698 L p 0 .579 0 r
F P
0 g
s
.40294 .43698 m .3975 .43146 L .40971 .43396 L p 0 .579 0 r
F P
0 g
s
.16006 .47504 m .17548 .51481 L .19136 .5121 L p 0 0 .796 r
F P
0 g
s
.39378 .41868 m .3975 .43146 L .38249 .43421 L p 0 .61 0 r
F P
0 g
s
.38249 .43421 m .37858 .41999 L .39378 .41868 L p 0 .61 0 r
F P
0 g
s
.58797 .50629 m .60554 .46814 L .58851 .46635 L p 0 0 .757 r
F P
0 g
s
.63233 .38339 m .61536 .42703 L .62826 .42574 L p 0 0 .696 r
F P
0 g
s
.61536 .42703 m .63233 .38339 L .6224 .38638 L p 0 0 .722 r
F P
0 g
s
.62552 .34775 m .6224 .38638 L .63233 .38339 L p 0 0 .675 r
F P
0 g
s
.7261 .38837 m .74744 .37186 L .74962 .35502 L p .914 0 0 r
F P
0 g
s
.70215 .40246 m .66109 .40598 L .64924 .39377 L p .235 0 0 r
F P
0 g
s
.66109 .40598 m .70215 .40246 L .71585 .39858 L p .443 0 0 r
F P
0 g
s
.4027 .35979 m .38355 .3462 L .38498 .36333 L p .344 0 0 r
F P
0 g
s
.38498 .36333 m .40513 .37581 L .4027 .35979 L p .344 0 0 r
F P
0 g
s
.38811 .32971 m .40632 .34286 L .41535 .32805 L p .427 0 0 r
F P
0 g
s
.40632 .34286 m .38811 .32971 L .38355 .3462 L p .395 0 0 r
F P
0 g
s
.77211 .39183 m .77549 .4228 L .75899 .41894 L p 0 .629 0 r
F P
0 g
s
.77549 .4228 m .77211 .39183 L .78933 .39432 L p 0 .539 0 r
F P
0 g
s
.78933 .39432 m .79154 .42863 L .77549 .4228 L p 0 .539 0 r
F P
0 g
s
.60554 .46814 m .62826 .42574 L .61536 .42703 L p 0 0 .733 r
F P
0 g
s
.35418 .45267 m .34656 .43098 L .36193 .42433 L p 0 .714 0 r
F P
0 g
s
.34656 .43098 m .35418 .45267 L .3456 .46532 L p 0 .741 0 r
F P
0 g
s
.17517 .47464 m .19136 .5121 L .20783 .50642 L p 0 0 .78 r
F P
0 g
s
.20783 .50642 m .19243 .47234 L .17517 .47464 L p 0 0 .78 r
F P
0 g
s
.66909 .33079 m .66582 .33449 L .68043 .33786 L p .032 0 0 r
F P
0 g
s
.68043 .33786 m .68114 .3233 L .66909 .33079 L p .032 0 0 r
F P
0 g
s
.74056 .43387 m .74478 .41768 L .75899 .41894 L p 0 .738 0 r
F P
0 g
s
.38355 .3462 m .4027 .35979 L .40632 .34286 L p .395 0 0 r
F P
0 g
s
.17494 .39466 m .20374 .43197 L .21686 .42968 L p 0 0 .827 r
F P
0 g
s
.21686 .42968 m .19254 .3956 L .17494 .39466 L p 0 0 .827 r
F P
0 g
s
.75627 .3907 m .75899 .41894 L .74478 .41768 L p 0 .721 0 r
F P
0 g
s
.75899 .41894 m .75627 .3907 L .77211 .39183 L p 0 .629 0 r
F P
0 g
s
.14489 .43121 m .16006 .47504 L .17517 .47464 L p 0 0 .775 r
F P
0 g
s
.17517 .47464 m .15465 .43301 L .14489 .43121 L p 0 0 .775 r
F P
0 g
s
.52212 .30193 m .53442 .31148 L .53754 .32365 L p 0 0 .26 r
F P
0 g
s
.53442 .31148 m .52212 .30193 L .52222 .28499 L p 0 0 .588 r
F P
0 g
s
.55448 .35978 m .55132 .34112 L .54429 .33563 L p 0 0 .6 r
F P
0 g
s
.45583 .31786 m .464 .32681 L .47727 .3282 L p .46 0 0 r
F P
0 g
s
.464 .32681 m .45583 .31786 L .44251 .3144 L p .53 0 0 r
F P
0 g
s
.7455 .33824 m .72976 .35672 L .73104 .37356 L p .895 0 0 r
F P
0 g
s
.73104 .37356 m .74962 .35502 L .7455 .33824 L p .895 0 0 r
F P
0 g
s
.19686 .36345 m .20851 .39534 L .22004 .39394 L p 0 0 .794 r
F P
0 g
s
.22004 .39394 m .21259 .36406 L .19686 .36345 L p 0 0 .794 r
F P
0 g
s
.19136 .5121 m .17517 .47464 L .16006 .47504 L p 0 0 .796 r
F P
0 g
s
.74962 .35502 m .73104 .37356 L .7261 .38837 L p .914 0 0 r
F P
0 g
s
.51285 .38381 m .47555 .39453 L .48611 .38104 L p .481 0 0 r
F P
0 g
s
.48611 .38104 m .51599 .36775 L .51285 .38381 L p .481 0 0 r
F P
0 g
s
.51599 .36775 m .48611 .38104 L .49213 .36503 L p .38 0 0 r
F P
0 g
s
.56095 .38925 m .56557 .36143 L .55448 .35978 L p 0 0 .662 r
F P
0 g
s
.56557 .36143 m .56095 .38925 L .5737 .39038 L p 0 0 .755 r
F P
0 g
s
.71521 .45221 m .703 .45755 L .6971 .47252 L p 0 .112 0 r
F P
0 g
s
.42955 .46412 m .44471 .4698 L .43823 .45438 L p 0 .572 0 r
F P
0 g
s
.43823 .45438 m .42534 .44986 L .42955 .46412 L p 0 .572 0 r
F P
0 g
s
.2428 .31093 m .2533 .29982 L .24723 .31373 L p 0 0 .115 r
F P
0 g
s
.59881 .42728 m .57212 .46319 L .58851 .46635 L p 0 0 .723 r
F P
0 g
s
.57212 .46319 m .59881 .42728 L .58158 .42644 L p 0 0 .69 r
F P
0 g
s
.71075 .51885 m .73652 .50843 L .72418 .49637 L p 0 .578 0 r
F P
0 g
s
.72418 .49637 m .70135 .50615 L .71075 .51885 L p 0 .578 0 r
F P
0 g
s
.5902 .39023 m .56679 .42465 L .58158 .42644 L p 0 0 .711 r
F P
0 g
s
.56679 .42465 m .5902 .39023 L .5737 .39038 L p 0 0 .655 r
F P
0 g
s
.18553 .26715 m .20521 .24505 L .22061 .24634 L p 0 0 .958 r
F P
0 g
s
.20521 .24505 m .18553 .26715 L .17334 .27125 L p 0 0 .903 r
F P
0 g
s
.79154 .42863 m .78933 .39432 L .80509 .39776 L p 0 .421 0 r
F P
0 g
s
.80509 .39776 m .80449 .43546 L .79154 .42863 L p 0 .421 0 r
F P
0 g
s
.79425 .46483 m .80449 .43546 L .81211 .44216 L p 0 .134 0 r
F P
0 g
s
.81211 .44216 m .7987 .47561 L .79425 .46483 L p 0 .134 0 r
F P
0 g
s
.44064 .5032 m .4202 .49638 L .40824 .50864 L p 0 .884 0 r
F P
0 g
s
.4202 .49638 m .44064 .5032 L .44556 .48709 L p 0 .819 0 r
F P
0 g
s
.22998 .32871 m .2258 .34068 L .23339 .33616 L p 0 0 .777 r
F P
0 g
s
.23339 .33616 m .23946 .32178 L .22998 .32871 L p 0 0 .777 r
F P
0 g
s
.50484 .39929 m .4622 .40322 L .47555 .39453 L p .545 0 0 r
F P
0 g
s
.47555 .39453 m .51285 .38381 L .50484 .39929 L p .545 0 0 r
F P
0 g
s
.58953 .26501 m .56779 .24766 L .58067 .25418 L p 0 0 .822 r
F P
0 g
s
.58067 .25418 m .60114 .27006 L .58953 .26501 L p 0 0 .822 r
F P
0 g
s
.55132 .34112 m .55448 .35978 L .56557 .36143 L p 0 0 .801 r
F P
0 g
s
.49213 .36503 m .51371 .35382 L .51599 .36775 L p .38 0 0 r
F P
0 g
s
.51371 .35382 m .49213 .36503 L .49259 .34917 L p .149 0 0 r
F P
0 g
s
.40495 .42065 m .40971 .43396 L .3975 .43146 L p 0 .475 0 r
F P
0 g
s
.3975 .43146 m .39378 .41868 L .40495 .42065 L p 0 .475 0 r
F P
0 g
s
.36877 .50814 m .34683 .48717 L .34298 .47752 L p 0 .408 0 r
F P
0 g
s
.34683 .48717 m .36877 .50814 L .38005 .51503 L p 0 .318 0 r
F P
0 g
s
.37991 .39485 m .37858 .41999 L .36193 .42433 L p 0 .593 0 r
F P
0 g
s
.36193 .42433 m .36309 .39633 L .37991 .39485 L p 0 .593 0 r
F P
0 g
s
.36309 .39633 m .36193 .42433 L .34656 .43098 L p 0 .653 0 r
F P
0 g
s
.61536 .42703 m .58851 .46635 L .60554 .46814 L p 0 0 .733 r
F P
0 g
s
.16956 .4338 m .19243 .47234 L .2087 .46857 L p 0 0 .791 r
F P
0 g
s
.2087 .46857 m .18698 .43344 L .16956 .4338 L p 0 0 .791 r
F P
0 g
s
.15398 .30999 m .14629 .34738 L .15166 .35206 L p 0 0 .339 r
F P
0 g
s
.14629 .34738 m .14734 .39005 L .15892 .39269 L p 0 0 .584 r
F P
0 g
s
.39386 .39513 m .39378 .41868 L .37858 .41999 L p 0 .502 0 r
F P
0 g
s
.37858 .41999 m .37991 .39485 L .39386 .39513 L p 0 .502 0 r
F P
0 g
s
.75401 .35623 m .77211 .39183 L .75627 .3907 L p 0 .676 0 r
F P
0 g
s
.77211 .39183 m .75401 .35623 L .76866 .35567 L p 0 .593 0 r
F P
0 g
s
.76866 .35567 m .78933 .39432 L .77211 .39183 L p 0 .593 0 r
F P
0 g
s
.77357 .50938 m .79671 .48401 L .78852 .48852 L p 0 .987 0 r
F P
0 g
s
.78852 .48852 m .7636 .51551 L .77357 .50938 L p 0 .987 0 r
F P
0 g
s
.62193 .307 m .60766 .2802 L .6079 .29375 L p 0 0 .41 r
F P
0 g
s
.6079 .29375 m .61934 .31606 L .62193 .307 L p 0 0 .41 r
F P
0 g
s
.6224 .38638 m .62552 .34775 L .61403 .35322 L p 0 0 .737 r
F P
0 g
s
.61403 .35322 m .60747 .38881 L .6224 .38638 L p 0 0 .737 r
F P
0 g
s
.64924 .39377 m .61908 .38377 L .61889 .36821 L p .94 0 0 r
F P
0 g
s
.61889 .36821 m .64167 .37825 L .64924 .39377 L p .94 0 0 r
F P
0 g
s
.61908 .38377 m .64924 .39377 L .66109 .40598 L p .841 0 0 r
F P
0 g
s
.42774 .4806 m .44556 .48709 L .44471 .4698 L p 0 .725 0 r
F P
0 g
s
.44471 .4698 m .42955 .46412 L .42774 .4806 L p 0 .725 0 r
F P
0 g
s
.21264 .3419 m .21259 .36406 L .22381 .36265 L p 0 0 .797 r
F P
0 g
s
.22381 .36265 m .2258 .34068 L .21264 .3419 L p 0 0 .797 r
F P
0 g
s
.44556 .48709 m .42774 .4806 L .4202 .49638 L p 0 .819 0 r
F P
0 g
s
.417 .44135 m .42534 .44986 L .4159 .44028 L p 0 0 0 r
F P
0 g
s
.58851 .46635 m .61536 .42703 L .59881 .42728 L p 0 0 .723 r
F P
0 g
s
.78933 .39432 m .76866 .35567 L .78586 .35601 L p 0 .502 0 r
F P
0 g
s
.78586 .35601 m .80509 .39776 L .78933 .39432 L p 0 .502 0 r
F P
0 g
s
.64167 .37825 m .61889 .36821 L .62415 .35596 L p .992 0 0 r
F P
0 g
s
.62415 .35596 m .63965 .36201 L .64167 .37825 L p .992 0 0 r
F P
0 g
s
.3456 .46532 m .33504 .43885 L .34656 .43098 L p 0 .741 0 r
F P
0 g
s
.33504 .43885 m .3456 .46532 L .34298 .47752 L p 0 .735 0 r
F P
0 g
s
.71252 .46552 m .6971 .47252 L .69651 .48974 L p 0 .306 0 r
F P
0 g
s
.6971 .47252 m .71252 .46552 L .71521 .45221 L p 0 .112 0 r
F P
0 g
s
.70135 .50615 m .72418 .49637 L .71568 .48116 L p 0 .446 0 r
F P
0 g
s
.71568 .48116 m .69651 .48974 L .70135 .50615 L p 0 .446 0 r
F P
0 g
s
.71585 .39858 m .67523 .41282 L .66109 .40598 L p .443 0 0 r
F P
0 g
s
.67523 .41282 m .71585 .39858 L .7261 .38837 L p .756 0 0 r
F P
0 g
s
.74478 .41768 m .74444 .39111 L .75627 .3907 L p 0 .721 0 r
F P
0 g
s
.6224 .38638 m .59881 .42728 L .61536 .42703 L p 0 0 .722 r
F P
0 g
s
.43623 .40121 m .40513 .37581 L .41312 .38811 L p .488 0 0 r
F P
0 g
s
.68114 .3233 m .68043 .33786 L .69404 .34764 L p .458 0 0 r
F P
0 g
s
.69404 .34764 m .69575 .32254 L .68114 .3233 L p .458 0 0 r
F P
0 g
s
.62463 .40003 m .66109 .40598 L .67523 .41282 L p .633 0 0 r
F P
0 g
s
.59881 .42728 m .6224 .38638 L .60747 .38881 L p 0 0 .725 r
F P
0 g
s
.66109 .40598 m .62463 .40003 L .61908 .38377 L p .841 0 0 r
F P
0 g
s
.34656 .43098 m .34617 .39932 L .36309 .39633 L p 0 .653 0 r
F P
0 g
s
.34617 .39932 m .34656 .43098 L .33504 .43885 L p 0 .704 0 r
F P
0 g
s
.69651 .48974 m .71568 .48116 L .71252 .46552 L p 0 .306 0 r
F P
0 g
s
.49331 .41161 m .44831 .40561 L .4622 .40322 L p .599 0 0 r
F P
0 g
s
.4622 .40322 m .50484 .39929 L .49331 .41161 L p .599 0 0 r
F P
0 g
s
.62552 .34775 m .61934 .31606 L .61026 .3258 L p 0 0 .709 r
F P
0 g
s
.61026 .3258 m .61403 .35322 L .62552 .34775 L p 0 0 .709 r
F P
0 g
s
.74439 .35761 m .75627 .3907 L .74444 .39111 L p 0 .774 0 r
F P
0 g
s
.75627 .3907 m .74439 .35761 L .75401 .35623 L p 0 .676 0 r
F P
0 g
s
.60747 .38881 m .58158 .42644 L .59881 .42728 L p 0 0 .725 r
F P
0 g
s
.58158 .42644 m .60747 .38881 L .5902 .39023 L p 0 0 .711 r
F P
0 g
s
.14734 .39005 m .16956 .4338 L .18698 .43344 L p 0 0 .815 r
F P
0 g
s
.18698 .43344 m .15892 .39269 L .14734 .39005 L p 0 0 .815 r
F P
0 g
s
.49259 .34917 m .50631 .34439 L .51371 .35382 L p .149 0 0 r
F P
0 g
s
.50631 .34439 m .49259 .34917 L .48736 .33613 L p .336 0 0 r
F P
0 g
s
.15398 .30999 m .16639 .28012 L .17334 .27125 L p 0 0 .61 r
F P
0 g
s
.16639 .28012 m .15398 .30999 L .15622 .31777 L p 0 0 .273 r
F P
0 g
s
.15166 .35206 m .15622 .31777 L .15398 .30999 L p 0 0 .339 r
F P
0 g
s
.44251 .3144 m .44983 .3323 L .464 .32681 L p .53 0 0 r
F P
0 g
s
.44983 .3323 m .44251 .3144 L .42818 .31798 L p .498 0 0 r
F P
0 g
s
.53754 .32365 m .52796 .3166 L .52212 .30193 L p 0 0 .26 r
F P
0 g
s
.76007 .31772 m .78586 .35601 L .76866 .35567 L p 0 .629 0 r
F P
0 g
s
.80509 .39776 m .78586 .35601 L .80279 .35719 L p 0 .363 0 r
F P
0 g
s
.78586 .35601 m .76007 .31772 L .77626 .31589 L p 0 .551 0 r
F P
0 g
s
.77626 .31589 m .80279 .35719 L .78586 .35601 L p 0 .551 0 r
F P
0 g
s
.39232 .36271 m .37991 .39485 L .36309 .39633 L p 0 .43 0 r
F P
0 g
s
.36309 .39633 m .37868 .36167 L .39232 .36271 L p 0 .43 0 r
F P
0 g
s
.37868 .36167 m .36309 .39633 L .34617 .39932 L p 0 .537 0 r
F P
0 g
s
.15465 .43301 m .17517 .47464 L .19243 .47234 L p 0 0 .796 r
F P
0 g
s
.19243 .47234 m .16956 .4338 L .15465 .43301 L p 0 0 .796 r
F P
0 g
s
.26907 .2765 m .25311 .28343 L .24675 .26734 L p 0 0 .766 r
F P
0 g
s
.24675 .26734 m .2644 .25975 L .26907 .2765 L p 0 0 .766 r
F P
0 g
s
.2644 .25975 m .24675 .26734 L .23528 .25423 L p 0 0 .892 r
F P
0 g
s
.40257 .39714 m .40495 .42065 L .39378 .41868 L p 0 .305 0 r
F P
0 g
s
.39378 .41868 m .39386 .39513 L .40257 .39714 L p 0 .305 0 r
F P
0 g
s
.15892 .39269 m .18698 .43344 L .20374 .43197 L p 0 0 .827 r
F P
0 g
s
.20374 .43197 m .17494 .39466 L .15892 .39269 L p 0 0 .827 r
F P
0 g
s
.74785 .32022 m .76866 .35567 L .75401 .35623 L p 0 .708 0 r
F P
0 g
s
.76866 .35567 m .74785 .32022 L .76007 .31772 L p 0 .629 0 r
F P
0 g
s
.26745 .29398 m .2533 .29982 L .25311 .28343 L p 0 0 .602 r
F P
0 g
s
.25311 .28343 m .26907 .2765 L .26745 .29398 L p 0 0 .602 r
F P
0 g
s
.63965 .36201 m .62415 .35596 L .63402 .34918 L p .928 0 0 r
F P
0 g
s
.63402 .34918 m .64355 .34779 L .63965 .36201 L p .928 0 0 r
F P
0 g
s
.58109 .36074 m .5737 .39038 L .5902 .39023 L p 0 0 .765 r
F P
0 g
s
.5737 .39038 m .58109 .36074 L .56557 .36143 L p 0 0 .755 r
F P
0 g
s
.80449 .43546 m .80509 .39776 L .81675 .40158 L p 0 .2 0 r
F P
0 g
s
.81675 .40158 m .81211 .44216 L .80449 .43546 L p 0 .2 0 r
F P
0 g
s
.24742 .31972 m .2359 .32277 L .24723 .31373 L p 0 0 0 r
F P
0 g
s
.22998 .32871 m .2359 .32277 L .22119 .32533 L p 0 0 .956 r
F P
0 g
s
.23234 .32357 m .22119 .32533 L .23466 .32299 L p 0 0 .697 r
F P
0 g
s
.23466 .32299 m .2359 .32277 L p .23234 .32357 L 0 0 .697 r
F P
0 g
s
.2359 .32277 m .22998 .32871 L .23905 .32208 L p 0 0 .678 r
F P
0 g
s
.23822 .32664 m .23014 .32412 L p .23234 .32357 L 0 0 0 r
F P
0 g
s
.22119 .32533 m .23234 .32357 L .21713 .32005 L p 0 0 .982 r
F P
0 g
s
.23014 .32412 m .21713 .32005 L .23234 .32357 L p 0 0 0 r
F P
0 g
s
.2359 .32277 m .24742 .31972 L .2325 .32353 L p 0 0 .697 r
F P
0 g
s
.2325 .32353 m .23234 .32357 L p .2359 .32277 L 0 0 .697 r
F P
0 g
s
.23905 .32208 m .23946 .32178 L p .2359 .32277 L 0 0 .678 r
F P
0 g
s
.38005 .51503 m .3566 .49256 L .34683 .48717 L p 0 .318 0 r
F P
0 g
s
.3566 .49256 m .38005 .51503 L .39402 .51523 L p 0 .683 0 r
F P
0 g
s
.1794 .36089 m .19254 .3956 L .20851 .39534 L p 0 0 .792 r
F P
0 g
s
.20851 .39534 m .19686 .36345 L .1794 .36089 L p 0 0 .792 r
F P
0 g
s
.40043 .36453 m .39386 .39513 L .37991 .39485 L p 0 .199 0 r
F P
0 g
s
.40154 .36682 m .40257 .39714 L .39386 .39513 L p 0 0 0 r
F P
0 g
s
.39232 .36271 m .40043 .36453 L p .39464 .37309 L 0 .199 0 r
F P
0 g
s
.40043 .36453 m .40154 .36682 L p .39761 .38132 L 0 0 0 r
F P
0 g
s
.39386 .39513 m .39698 .38061 L p .39761 .38132 L 0 0 0 r
F P
0 g
s
.39698 .38061 m .40043 .36453 L p .39761 .38132 L 0 0 0 r
F P
0 g
s
.40257 .39714 m .40249 .39458 L p .40049 .38912 L 0 .899 0 r
F P
0 g
s
.40249 .39458 m .40154 .36682 L p .40049 .38912 L 0 .899 0 r
F P
0 g
s
.40154 .36682 m .39532 .36919 L p .40049 .38912 L 0 .899 0 r
F P
0 g
s
.40049 .38912 m .39882 .38268 L .40365 .39777 L 0 .899 0 r
F
.39532 .36919 m .39947 .3834 L p .39882 .38268 L F P
0 g
s
.39947 .3834 m .40365 .39777 L p .39882 .38268 L 0 .899 0 r
F P
0 g
s
.40447 .40058 m .40257 .39714 L p .40049 .38912 L 0 .899 0 r
F P
0 g
s
.40365 .39777 m .40447 .40058 L p .40049 .38912 L 0 .899 0 r
F P
0 g
s
.37991 .39485 m .39156 .36467 L p .39464 .37309 L 0 .199 0 r
F P
0 g
s
.39156 .36467 m .39232 .36271 L p .39464 .37309 L 0 .199 0 r
F P
0 g
s
.56557 .36143 m .56389 .34288 L .55132 .34112 L p 0 0 .801 r
F P
0 g
s
.54653 .3317 m .55132 .34112 L .56389 .34288 L p 0 0 .894 r
F P
0 g
s
.50391 .27924 m .52222 .28499 L .52212 .30193 L p 0 0 .602 r
F P
0 g
s
.52222 .28499 m .50391 .27924 L .5087 .26294 L p 0 0 .752 r
F P
0 g
s
.5087 .26294 m .52822 .26865 L .52222 .28499 L p 0 0 .752 r
F P
0 g
s
.80279 .35719 m .81675 .40158 L .80509 .39776 L p 0 .363 0 r
F P
0 g
s
.60114 .27006 m .58067 .25418 L .58947 .2667 L p 0 0 0 r
F P
0 g
s
.58947 .2667 m .60766 .2802 L .60114 .27006 L p 0 0 0 r
F P
0 g
s
.17334 .27125 m .19172 .25067 L .20521 .24505 L p 0 0 .903 r
F P
0 g
s
.19172 .25067 m .17334 .27125 L .16639 .28012 L p 0 0 0 r
F P
0 g
s
.41016 .42561 m .417 .44135 L .40971 .43396 L p 0 .032 0 r
F P
0 g
s
.40971 .43396 m .40495 .42065 L .41016 .42561 L p 0 .032 0 r
F P
0 g
s
.34617 .39932 m .36185 .36155 L .37868 .36167 L p 0 .537 0 r
F P
0 g
s
.36185 .36155 m .34617 .39932 L .33196 .40334 L p 0 .607 0 r
F P
0 g
s
.33504 .43885 m .33196 .40334 L .34617 .39932 L p 0 .704 0 r
F P
0 g
s
.52822 .26865 m .5087 .26294 L .51869 .25066 L p 0 0 .878 r
F P
0 g
s
.51869 .25066 m .53911 .25564 L .52822 .26865 L p 0 0 .878 r
F P
0 g
s
.7987 .47561 m .81211 .44216 L .81301 .44758 L p 0 0 0 r
F P
0 g
s
.81301 .44758 m .79671 .48401 L .7987 .47561 L p 0 0 0 r
F P
0 g
s
.23528 .25423 m .25419 .24655 L .2644 .25975 L p 0 0 .892 r
F P
0 g
s
.25419 .24655 m .23528 .25423 L .22061 .24634 L p 0 0 .982 r
F P
0 g
s
.56389 .34288 m .56557 .36143 L .58109 .36074 L p 0 0 .813 r
F P
0 g
s
.41802 .45243 m .42955 .46412 L .42534 .44986 L p 0 .586 0 r
F P
0 g
s
.42534 .44986 m .417 .44135 L .41802 .45243 L p 0 .586 0 r
F P
0 g
s
.81675 .40158 m .80279 .35719 L .81662 .35902 L p 0 .031 0 r
F P
0 g
s
.80279 .35719 m .77626 .31589 L .79374 .31501 L p 0 .446 0 r
F P
0 g
s
.79374 .31501 m .81662 .35902 L .80279 .35719 L p 0 .446 0 r
F P
0 g
s
.7636 .51551 m .78852 .48852 L .77549 .48829 L p 0 .877 0 r
F P
0 g
s
.77549 .48829 m .75049 .5152 L .7636 .51551 L p 0 .877 0 r
F P
0 g
s
.25982 .30925 m .24723 .31373 L .2533 .29982 L p 0 0 .354 r
F P
0 g
s
.2533 .29982 m .26745 .29398 L .25982 .30925 L p 0 0 .354 r
F P
0 g
s
.76095 .27858 m .77626 .31589 L .76007 .31772 L p 0 .659 0 r
F P
0 g
s
.77626 .31589 m .76095 .27858 L .77769 .27479 L p 0 .593 0 r
F P
0 g
s
.77769 .27479 m .79374 .31501 L .77626 .31589 L p 0 .593 0 r
F P
0 g
s
.7261 .38837 m .68931 .41307 L .67523 .41282 L p .756 0 0 r
F P
0 g
s
.63458 .41426 m .67523 .41282 L .68931 .41307 L p .096 0 0 r
F P
0 g
s
.67523 .41282 m .63458 .41426 L .62463 .40003 L p .633 0 0 r
F P
0 g
s
.60747 .38881 m .61403 .35322 L .59828 .35783 L p 0 0 .759 r
F P
0 g
s
.59828 .35783 m .5902 .39023 L .60747 .38881 L p 0 0 .759 r
F P
0 g
s
.5051 .29678 m .52212 .30193 L .52796 .3166 L p 0 0 .393 r
F P
0 g
s
.52212 .30193 m .5051 .29678 L .50391 .27924 L p 0 0 .602 r
F P
0 g
s
.39936 .3279 m .37868 .36167 L .36185 .36155 L p 0 .25 0 r
F P
0 g
s
.4027 .33031 m .39232 .36271 L .37868 .36167 L p 0 0 0 r
F P
0 g
s
.39936 .3279 m .4027 .33031 L p .39197 .34432 L 0 0 0 r
F P
0 g
s
.38786 .34968 m .39138 .34301 L .39197 .34432 L 0 0 0 r
F
.39138 .34301 m .39936 .3279 L .39197 .34432 L F
.4027 .33031 m .39867 .33223 L p .39478 .35091 L 0 .952 0 r
F P
0 g
s
.39232 .36271 m .39549 .35281 L p .39478 .35091 L 0 .952 0 r
F P
0 g
s
.39549 .35281 m .4027 .33031 L p .39478 .35091 L 0 .952 0 r
F P
0 g
s
.40043 .36453 m .39232 .36271 L p .39612 .34446 L 0 .952 0 r
F P
0 g
s
.39923 .3425 m .40043 .36453 L p .39612 .34446 L 0 .952 0 r
F P
0 g
s
.39867 .33223 m .39923 .3425 L p .39612 .34446 L 0 .952 0 r
F P
0 g
s
.38486 .35158 m .39936 .3279 L p .38786 .34968 L 0 0 0 r
F P
0 g
s
.38935 .32542 m .39936 .3279 L p .38896 .33723 L 0 .25 0 r
F P
0 g
s
.38587 .33 m .38935 .32542 L p .38896 .33723 L 0 .25 0 r
F P
0 g
s
.40043 .36453 m .39867 .33223 L .38783 .3333 L p 0 .892 0 r
F P
0 g
s
.39726 .35666 m .39501 .3511 L .40038 .36398 L 0 .892 0 r
F
.38783 .3333 m .39617 .35367 L p .39501 .3511 L F P
0 g
s
.39617 .35367 m .40038 .36398 L p .39501 .3511 L 0 .892 0 r
F P
0 g
s
.37868 .36167 m .38486 .35158 L p .38786 .34968 L 0 0 0 r
F P
0 g
s
.40154 .36682 m .40043 .36453 L p .39726 .35666 L 0 .892 0 r
F P
0 g
s
.40038 .36398 m .40154 .36682 L p .39726 .35666 L 0 .892 0 r
F P
0 g
s
.36185 .36155 m .38587 .33 L p .38896 .33723 L 0 .25 0 r
F P
0 g
s
.38935 .32542 m .36185 .36155 L .34459 .36238 L p 0 .435 0 r
F P
0 g
s
.33196 .40334 m .34459 .36238 L .36185 .36155 L p 0 .607 0 r
F P
0 g
s
.68931 .41307 m .7261 .38837 L .73104 .37356 L p .837 0 0 r
F P
0 g
s
.21593 .33045 m .21264 .3419 L .2258 .34068 L p 0 0 .863 r
F P
0 g
s
.2258 .34068 m .22998 .32871 L .21593 .33045 L p 0 0 .863 r
F P
0 g
s
.5902 .39023 m .59828 .35783 L .58109 .36074 L p 0 0 .765 r
F P
0 g
s
.42809 .3907 m .4027 .35979 L .40513 .37581 L p 0 0 0 r
F P
0 g
s
.15622 .31777 m .15166 .35206 L .16337 .35685 L p 0 0 .684 r
F P
0 g
s
.15166 .35206 m .15892 .39269 L .17494 .39466 L p 0 0 .751 r
F P
0 g
s
.17494 .39466 m .16337 .35685 L .15166 .35206 L p 0 0 .751 r
F P
0 g
s
.34298 .47752 m .32937 .44663 L .33504 .43885 L p 0 .735 0 r
F P
0 g
s
.32937 .44663 m .34298 .47752 L .34683 .48717 L p 0 .444 0 r
F P
0 g
s
.61934 .31606 m .6079 .29375 L .60174 .30834 L p 0 0 .686 r
F P
0 g
s
.60174 .30834 m .61026 .3258 L .61934 .31606 L p 0 0 .686 r
F P
0 g
s
.69575 .32254 m .69404 .34764 L .70422 .36212 L p .654 0 0 r
F P
0 g
s
.70422 .36212 m .71038 .32873 L .69575 .32254 L p .654 0 0 r
F P
0 g
s
.81662 .35902 m .79374 .31501 L .80962 .31523 L p 0 .237 0 r
F P
0 g
s
.79374 .31501 m .77769 .27479 L .79488 .27277 L p 0 .514 0 r
F P
0 g
s
.79488 .27277 m .80962 .31523 L .79374 .31501 L p 0 .514 0 r
F P
0 g
s
.19618 .33954 m .19686 .36345 L .21259 .36406 L p 0 0 .801 r
F P
0 g
s
.21259 .36406 m .21264 .3419 L .19618 .33954 L p 0 0 .801 r
F P
0 g
s
.48016 .41871 m .43623 .40121 L .44831 .40561 L p .653 0 0 r
F P
0 g
s
.44831 .40561 m .49331 .41161 L .48016 .41871 L p .653 0 0 r
F P
0 g
s
.74172 .32296 m .75401 .35623 L .74439 .35761 L p 0 .809 0 r
F P
0 g
s
.74439 .35761 m .7415 .35958 L p .74157 .35785 L 0 .831 0 r
F P
0 g
s
.74616 .33499 m .74172 .32296 L p .74844 .32364 L 0 .708 0 r
F P
0 g
s
.74281 .32549 m .74172 .32296 L p .74193 .32578 L 0 .809 0 r
F P
0 g
s
.74439 .35761 m .74338 .33699 L p .74193 .32578 L 0 .809 0 r
F P
0 g
s
.74338 .33699 m .74281 .32549 L p .74193 .32578 L 0 .809 0 r
F P
0 g
s
.74444 .39111 m .74236 .3688 L p .74439 .35761 L 0 .774 0 r
F P
0 g
s
.74236 .3688 m .7415 .35958 L .74345 .35826 L p 0 .774 0 r
F P
0 g
s
.74345 .35826 m .74439 .35761 L p .74236 .3688 L 0 .774 0 r
F P
0 g
s
.7415 .35958 m .74444 .39111 L .73868 .39303 L p 0 .889 0 r
F P
0 g
s
.74598 .36182 m .7408 .38395 L p .74007 .39616 L 0 0 0 r
F P
0 g
s
.7408 .38395 m .73868 .39303 L .73964 .39519 L p 0 0 0 r
F P
0 g
s
.73964 .39519 m .74007 .39616 L p .7408 .38395 L 0 0 0 r
F P
0 g
s
.7415 .35958 m .74598 .36182 L p .7463 .35965 L 0 0 0 r
F P
0 g
s
.75401 .35623 m .74616 .33499 L p .74844 .32364 L 0 .708 0 r
F P
0 g
s
.7511 .32736 m .74484 .34836 L p .7463 .35965 L 0 0 0 r
F P
0 g
s
.74385 .34651 m .74439 .35761 L p .74157 .35785 L 0 .831 0 r
F P
0 g
s
.74484 .34836 m .7415 .35958 L p .7463 .35965 L 0 0 0 r
F P
0 g
s
.74281 .32549 m .74385 .34651 L p .74157 .35785 L 0 .831 0 r
F P
0 g
s
.74172 .32296 m .74785 .32022 L p .74844 .32364 L 0 .708 0 r
F P
0 g
s
.42818 .31798 m .43729 .34377 L .44983 .3323 L p .498 0 0 r
F P
0 g
s
.43729 .34377 m .42818 .31798 L .41535 .32805 L p .445 0 0 r
F P
0 g
s
.33196 .40334 m .33504 .43885 L .32937 .44663 L p 0 .749 0 r
F P
0 g
s
.53911 .25564 m .51869 .25066 L .53223 .24453 L p 0 0 .966 r
F P
0 g
s
.53223 .24453 m .55309 .24821 L .53911 .25564 L p 0 0 .966 r
F P
0 g
s
.61403 .35322 m .61026 .3258 L .59625 .33449 L p 0 0 .769 r
F P
0 g
s
.59625 .33449 m .59828 .35783 L .61403 .35322 L p 0 0 .769 r
F P
0 g
s
.48736 .33613 m .495 .34118 L .50631 .34439 L p .336 0 0 r
F P
0 g
s
.495 .34118 m .48736 .33613 L .47727 .3282 L p .544 0 0 r
F P
0 g
s
.34459 .36238 m .37446 .32327 L .38935 .32542 L p 0 .435 0 r
F P
0 g
s
.37446 .32327 m .34459 .36238 L .32977 .36403 L p 0 .533 0 r
F P
0 g
s
.34459 .36238 m .33196 .40334 L .32287 .40771 L p 0 .665 0 r
F P
0 g
s
.32287 .40771 m .32977 .36403 L .34459 .36238 L p 0 .665 0 r
F P
0 g
s
.74743 .28352 m .76007 .31772 L .74785 .32022 L p 0 .73 0 r
F P
0 g
s
.76007 .31772 m .74743 .28352 L .76095 .27858 L p 0 .659 0 r
F P
0 g
s
.64355 .34779 m .63402 .34918 L .64687 .34914 L p 0 0 0 r
F P
0 g
s
.64687 .34914 m .65278 .33803 L .64355 .34779 L p 0 0 0 r
F P
0 g
s
.56389 .34288 m .55992 .33416 L .54653 .3317 L p 0 0 .894 r
F P
0 g
s
.16337 .35685 m .17494 .39466 L .19254 .3956 L p 0 0 .782 r
F P
0 g
s
.19254 .3956 m .1794 .36089 L .16337 .35685 L p 0 0 .782 r
F P
0 g
s
.22061 .24634 m .24014 .2392 L .25419 .24655 L p 0 0 .982 r
F P
0 g
s
.24014 .2392 m .22061 .24634 L .20521 .24505 L p 0 0 .918 r
F P
0 g
s
.39402 .51523 m .37064 .49267 L .3566 .49256 L p 0 .683 0 r
F P
0 g
s
.37064 .49267 m .39402 .51523 L .40824 .50864 L p 0 .769 0 r
F P
0 g
s
.52592 .41445 m .49331 .41161 L .50484 .39929 L p .552 0 0 r
F P
0 g
s
.50484 .39929 m .53391 .39881 L .52592 .41445 L p .552 0 0 r
F P
0 g
s
.53391 .39881 m .50484 .39929 L .51285 .38381 L p .491 0 0 r
F P
0 g
s
.51206 .31262 m .52796 .3166 L .53879 .3265 L p 0 0 .031 r
F P
0 g
s
.52796 .3166 m .51206 .31262 L .5051 .29678 L p 0 0 .393 r
F P
0 g
s
.393 .28827 m .37446 .32327 L .35717 .32181 L p 0 .483 0 r
F P
0 g
s
.35717 .32181 m .37811 .28394 L .393 .28827 L p 0 .483 0 r
F P
0 g
s
.32977 .36403 m .35717 .32181 L .37446 .32327 L p 0 .533 0 r
F P
0 g
s
.37811 .28394 m .35717 .32181 L .34035 .32128 L p 0 .581 0 r
F P
0 g
s
.31673 .36861 m .32678 .32178 L .34035 .32128 L p 0 .659 0 r
F P
0 g
s
.32977 .36403 m .31989 .36623 L p .33684 .34603 L 0 .602 0 r
F P
0 g
s
.31989 .36623 m .34035 .32128 L .35717 .32181 L p 0 .602 0 r
F P
0 g
s
.31989 .36623 m .31673 .36861 L p .32766 .3467 L 0 .659 0 r
F P
0 g
s
.32766 .3467 m .32814 .34689 L .31989 .36623 L 0 .659 0 r
F
.33424 .3347 m .31989 .36623 L p .33146 .3391 L F P
0 g
s
.32055 .41171 m .31989 .36623 L .32977 .36403 L p 0 .719 0 r
F P
0 g
s
.32937 .44663 m .32287 .40771 L .33196 .40334 L p 0 .749 0 r
F P
0 g
s
.32287 .40771 m .32055 .41171 L p .32535 .3869 L 0 .719 0 r
F P
0 g
s
.32977 .36403 m .32655 .38445 L p .32535 .3869 L 0 .719 0 r
F P
0 g
s
.32655 .38445 m .32287 .40771 L p .32535 .3869 L 0 .719 0 r
F P
0 g
s
.34683 .48717 m .33062 .45298 L .32937 .44663 L p 0 .444 0 r
F P
0 g
s
.32287 .40771 m .32937 .44663 L .33062 .45298 L p 0 .674 0 r
F P
0 g
s
.33062 .45298 m .32055 .41171 L .32287 .40771 L p 0 .674 0 r
F P
0 g
s
.32256 .38917 m .31989 .36623 L .32019 .38703 L 0 .679 0 r
F
.32055 .41171 m .32552 .41463 L p .32256 .38917 L F P
0 g
s
.32256 .38917 m .32019 .38703 L .32055 .41171 L 0 .679 0 r
F
.31673 .36861 m .31989 .36623 L p .32288 .39193 L F P
0 g
s
.32096 .37076 m .31673 .36861 L p .32745 .39358 L 0 .005 0 r
F P
0 g
s
.33705 .41593 m .32986 .39575 L p .32745 .39358 L 0 .005 0 r
F P
0 g
s
.32986 .39575 m .32096 .37076 L p .32745 .39358 L 0 .005 0 r
F P
0 g
s
.32552 .41463 m .32166 .39443 L p .32288 .39193 L 0 .679 0 r
F P
0 g
s
.32166 .39443 m .31673 .36861 L p .32288 .39193 L 0 .679 0 r
F P
0 g
s
.31673 .36861 m .32096 .37076 L p .32462 .33998 L 0 .699 0 r
F P
0 g
s
.32678 .32178 m .32384 .33545 L p .32462 .33998 L 0 .699 0 r
F P
0 g
s
.32384 .33545 m .31673 .36861 L p .32462 .33998 L 0 .699 0 r
F P
0 g
s
.31879 .32323 m .32678 .32178 L p .3241 .34436 L 0 .699 0 r
F P
0 g
s
.31973 .34378 m .31879 .32323 L p .3241 .34436 L 0 .699 0 r
F P
0 g
s
.32316 .35219 m .31973 .34378 L .3241 .34436 L 0 .699 0 r
F
.32096 .37076 m .32032 .35669 L p .32316 .35219 L F P
0 g
s
.32032 .35669 m .31973 .34378 L p .32316 .35219 L 0 .699 0 r
F P
0 g
s
.32096 .37076 m .33198 .3723 L p .32365 .34132 L 0 .509 0 r
F P
0 g
s
.31879 .32323 m .31941 .33688 L p .32365 .34132 L 0 .509 0 r
F P
0 g
s
.31941 .33688 m .32096 .37076 L p .32365 .34132 L 0 .509 0 r
F P
0 g
s
.31786 .3254 m .31879 .32323 L p .3248 .34558 L 0 .509 0 r
F P
0 g
s
.32766 .3467 m .33146 .3391 L .32814 .34689 L 0 .659 0 r
F
.34109 .34659 m .32977 .36403 L p .33684 .34603 L 0 .602 0 r
F P
0 g
s
.3243 .32793 m .31786 .3254 L p .33627 .35452 L 0 .199 0 r
F P
0 g
s
.34792 .37294 m .34066 .35911 L p .33627 .35452 L 0 .199 0 r
F P
0 g
s
.34066 .35911 m .3243 .32793 L p .33627 .35452 L 0 .199 0 r
F P
0 g
s
.33198 .3723 m .32382 .3452 L p .3248 .34558 L 0 .509 0 r
F P
0 g
s
.32382 .3452 m .31786 .3254 L p .3248 .34558 L 0 .509 0 r
F P
0 g
s
.34035 .32128 m .33424 .3347 L p .33146 .3391 L 0 .659 0 r
F P
0 g
s
.35717 .32181 m .34109 .34659 L p .33684 .34603 L 0 .602 0 r
F P
0 g
s
.57981 .34054 m .58109 .36074 L .59828 .35783 L p 0 0 .798 r
F P
0 g
s
.58109 .36074 m .57981 .34054 L .56389 .34288 L p 0 0 .813 r
F P
0 g
s
.40447 .40058 m .41016 .42561 L .40495 .42065 L p 0 0 0 r
F P
0 g
s
.39909 .40486 m .40839 .43276 L .40879 .43112 L p 0 .806 0 r
F P
0 g
s
.40879 .43112 m .41016 .42561 L p .40681 .41934 L 0 .806 0 r
F P
0 g
s
.40495 .42065 m .40257 .39714 L .4033 .39846 L p 0 0 0 r
F P
0 g
s
.40495 .42065 m .4033 .39846 L .40461 .40622 L 0 0 0 r
F
.40681 .41934 m .39909 .40486 L .40879 .43112 L 0 .806 0 r
F
.4033 .39846 m .40447 .40058 L p .40461 .40622 L 0 0 0 r
F P
0 g
s
.72489 .45178 m .71252 .46552 L .71568 .48116 L p 0 .267 0 r
F P
0 g
s
.71252 .46552 m .72489 .45178 L .72405 .44208 L p 0 0 0 r
F P
0 g
s
.538 .38247 m .51285 .38381 L .51599 .36775 L p .389 0 0 r
F P
0 g
s
.51285 .38381 m .538 .38247 L .53391 .39881 L p .491 0 0 r
F P
0 g
s
.15622 .31777 m .166 .2923 L .16639 .28012 L p 0 0 .273 r
F P
0 g
s
.166 .2923 m .15622 .31777 L .16518 .32628 L p 0 0 .663 r
F P
0 g
s
.16337 .35685 m .16518 .32628 L .15622 .31777 L p 0 0 .684 r
F P
0 g
s
.40311 .29285 m .38935 .32542 L .37446 .32327 L p 0 .311 0 r
F P
0 g
s
.37446 .32327 m .393 .28827 L .40311 .29285 L p 0 .311 0 r
F P
0 g
s
.24723 .31373 m .25982 .30925 L .24742 .31972 L p 0 0 0 r
F P
0 g
s
.41251 .46531 m .42774 .4806 L .42955 .46412 L p 0 .742 0 r
F P
0 g
s
.42955 .46412 m .41802 .45243 L .41251 .46531 L p 0 .742 0 r
F P
0 g
s
.42538 .37583 m .40632 .34286 L .4027 .35979 L p .24 0 0 r
F P
0 g
s
.4027 .35979 m .42809 .3907 L .42538 .37583 L p .24 0 0 r
F P
0 g
s
.59828 .35783 m .59625 .33449 L .57981 .34054 L p 0 0 .798 r
F P
0 g
s
.55992 .33416 m .56389 .34288 L .57981 .34054 L p 0 0 .884 r
F P
0 g
s
.73104 .37356 m .70089 .40662 L .68931 .41307 L p .837 0 0 r
F P
0 g
s
.70089 .40662 m .71125 .39527 L p .71276 .38609 L .841 0 0 r
F P
0 g
s
.73104 .37356 m .72976 .35672 L p .71276 .38609 L .841 0 0 r
F P
0 g
s
.71125 .39527 m .73104 .37356 L p .71276 .38609 L .841 0 0 r
F P
0 g
s
.72976 .35672 m .70794 .39448 L .70089 .40662 L p .841 0 0 r
F P
0 g
s
.70914 .37872 m .70794 .39448 L p .71681 .36162 L .813 0 0 r
F P
0 g
s
.72243 .34083 m .71831 .35257 L p .71681 .36162 L .813 0 0 r
F P
0 g
s
.71831 .35257 m .70914 .37872 L p .71681 .36162 L .813 0 0 r
F P
0 g
s
.41535 .32805 m .42862 .3592 L .43729 .34377 L p .445 0 0 r
F P
0 g
s
.42862 .3592 m .41535 .32805 L .40632 .34286 L p .37 0 0 r
F P
0 g
s
.60766 .2802 m .58947 .2667 L .59259 .28306 L p 0 0 .438 r
F P
0 g
s
.59259 .28306 m .6079 .29375 L .60766 .2802 L p 0 0 .438 r
F P
0 g
s
.76463 .23764 m .77769 .27479 L .76095 .27858 L p 0 .669 0 r
F P
0 g
s
.77769 .27479 m .76463 .23764 L .78161 .23314 L p 0 .609 0 r
F P
0 g
s
.78161 .23314 m .79488 .27277 L .77769 .27479 L p 0 .609 0 r
F P
0 g
s
.75049 .5152 m .77549 .48829 L .75988 .4833 L p 0 .721 0 r
F P
0 g
s
.75988 .4833 m .73652 .50843 L .75049 .5152 L p 0 .721 0 r
F P
0 g
s
.22119 .32533 m .21593 .33045 L .22998 .32871 L p 0 0 .956 r
F P
0 g
s
.71038 .32873 m .70422 .36212 L .70914 .37872 L p .755 0 0 r
F P
0 g
s
.70914 .37872 m .72243 .34083 L .71038 .32873 L p .755 0 0 r
F P
0 g
s
.81211 .44216 m .81675 .40158 L .82227 .40514 L p 0 0 0 r
F P
0 g
s
.82227 .40514 m .81301 .44758 L .81211 .44216 L p 0 0 0 r
F P
0 g
s
.16639 .28012 m .18251 .26229 L .19172 .25067 L p 0 0 0 r
F P
0 g
s
.18251 .26229 m .16639 .28012 L .166 .2923 L p 0 0 .363 r
F P
0 g
s
.81662 .35902 m .82227 .40514 L .81675 .40158 L p 0 .031 0 r
F P
0 g
s
.40839 .43276 m .41802 .45243 L .417 .44135 L p 0 .671 0 r
F P
0 g
s
.417 .44135 m .41016 .42561 L .40839 .43276 L p 0 .671 0 r
F P
0 g
s
.60018 .40072 m .62463 .40003 L .63458 .41426 L p .826 0 0 r
F P
0 g
s
.62463 .40003 m .60018 .40072 L .59784 .38431 L p .88 0 0 r
F P
0 g
s
.59784 .38431 m .61908 .38377 L .62463 .40003 L p .88 0 0 r
F P
0 g
s
.55309 .24821 m .53223 .24453 L .54702 .2457 L p 0 0 .815 r
F P
0 g
s
.54702 .2457 m .56779 .24766 L .55309 .24821 L p 0 0 .815 r
F P
0 g
s
.68931 .41307 m .64731 .4241 L .63458 .41426 L p .096 0 0 r
F P
0 g
s
.64731 .4241 m .68931 .41307 L .70089 .40662 L p .583 0 0 r
F P
0 g
s
.40632 .34286 m .42538 .37583 L .42862 .3592 L p .37 0 0 r
F P
0 g
s
.51537 .4268 m .48016 .41871 L .49331 .41161 L p .599 0 0 r
F P
0 g
s
.49331 .41161 m .52592 .41445 L .51537 .4268 L p .599 0 0 r
F P
0 g
s
.74912 .24514 m .76095 .27858 L .74743 .28352 L p 0 .724 0 r
F P
0 g
s
.76095 .27858 m .74912 .24514 L .76463 .23764 L p 0 .669 0 r
F P
0 g
s
.80962 .31523 m .82501 .36121 L .81662 .35902 L p 0 .237 0 r
F P
0 g
s
.82501 .36121 m .80962 .31523 L .82124 .31654 L p 0 0 0 r
F P
0 g
s
.80962 .31523 m .79488 .27277 L .80968 .27287 L p 0 .382 0 r
F P
0 g
s
.80968 .27287 m .82124 .31654 L .80962 .31523 L p 0 .382 0 r
F P
0 g
s
.82227 .40514 m .81662 .35902 L .82501 .36121 L p 0 .75 0 r
F P
0 g
s
.16518 .32628 m .16337 .35685 L .1794 .36089 L p 0 0 .76 r
F P
0 g
s
.17934 .334 m .1794 .36089 L .19686 .36345 L p 0 0 .788 r
F P
0 g
s
.19686 .36345 m .19618 .33954 L .17934 .334 L p 0 0 .788 r
F P
0 g
s
.61026 .3258 m .60174 .30834 L .59022 .32137 L p 0 0 .789 r
F P
0 g
s
.59022 .32137 m .59625 .33449 L .61026 .3258 L p 0 0 .789 r
F P
0 g
s
.34035 .32128 m .36095 .28057 L .37811 .28394 L p 0 .581 0 r
F P
0 g
s
.36095 .28057 m .34035 .32128 L .32678 .32178 L p 0 .657 0 r
F P
0 g
s
.40824 .50864 m .38651 .48741 L .37064 .49267 L p 0 .769 0 r
F P
0 g
s
.38651 .48741 m .40824 .50864 L .4202 .49638 L p 0 .789 0 r
F P
0 g
s
.79671 .48401 m .81301 .44758 L .8069 .45076 L p 0 .969 0 r
F P
0 g
s
.8069 .45076 m .78852 .48852 L .79671 .48401 L p 0 .969 0 r
F P
0 g
s
.60665 .41709 m .63458 .41426 L .64731 .4241 L p .736 0 0 r
F P
0 g
s
.63458 .41426 m .60665 .41709 L .60018 .40072 L p .826 0 0 r
F P
0 g
s
.19973 .3266 m .19618 .33954 L .21264 .3419 L p 0 0 .854 r
F P
0 g
s
.21264 .3419 m .21593 .33045 L .19973 .3266 L p 0 0 .854 r
F P
0 g
s
.70794 .39448 m .72976 .35672 L .72243 .34083 L p .813 0 0 r
F P
0 g
s
.40135 .47771 m .4202 .49638 L .42774 .4806 L p 0 .783 0 r
F P
0 g
s
.42774 .4806 m .41251 .46531 L .40135 .47771 L p 0 .783 0 r
F P
0 g
s
.61908 .38377 m .59784 .38431 L .60008 .37061 L p .912 0 0 r
F P
0 g
s
.60008 .37061 m .61889 .36821 L .61908 .38377 L p .912 0 0 r
F P
0 g
s
.51599 .36775 m .5375 .36817 L .538 .38247 L p .389 0 0 r
F P
0 g
s
.5375 .36817 m .51599 .36775 L .51371 .35382 L p .146 0 0 r
F P
0 g
s
.79488 .27277 m .78161 .23314 L .79725 .2324 L p 0 .524 0 r
F P
0 g
s
.79725 .2324 m .80968 .27287 L .79488 .27277 L p 0 .524 0 r
F P
0 g
s
.43623 .40121 m .48016 .41871 L .4676 .41931 L p .695 0 0 r
F P
0 g
s
.4676 .41931 m .42809 .3907 L .43623 .40121 L p .695 0 0 r
F P
0 g
s
.39252 .24935 m .37811 .28394 L .36095 .28057 L p 0 .581 0 r
F P
0 g
s
.36095 .28057 m .37648 .24261 L .39252 .24935 L p 0 .581 0 r
F P
0 g
s
.37648 .24261 m .36095 .28057 L .34434 .27872 L p 0 .671 0 r
F P
0 g
s
.32678 .32178 m .34434 .27872 L .36095 .28057 L p 0 .657 0 r
F P
0 g
s
.1794 .36089 m .17934 .334 L .16518 .32628 L p 0 0 .76 r
F P
0 g
s
.73211 .46323 m .71568 .48116 L .72418 .49637 L p 0 .457 0 r
F P
0 g
s
.71568 .48116 m .73211 .46323 L .72489 .45178 L p 0 .267 0 r
F P
0 g
s
.73942 .28879 m .74785 .32022 L .74172 .32296 L p 0 .796 0 r
F P
0 g
s
.74785 .32022 m .73942 .28879 L .74743 .28352 L p 0 .73 0 r
F P
0 g
s
.33062 .45298 m .34683 .48717 L .3566 .49256 L p 0 .305 0 r
F P
0 g
s
.40518 .25747 m .393 .28827 L .37811 .28394 L p 0 .461 0 r
F P
0 g
s
.37811 .28394 m .39252 .24935 L .40518 .25747 L p 0 .461 0 r
F P
0 g
s
.4202 .49638 m .40135 .47771 L .38651 .48741 L p 0 .789 0 r
F P
0 g
s
.53879 .3265 m .52362 .32407 L .51206 .31262 L p 0 0 .031 r
F P
0 g
s
.40668 .29691 m .39936 .3279 L .38935 .32542 L p 0 0 0 r
F P
0 g
s
.40311 .29285 m .40668 .29691 L p .39806 .3111 L 0 0 0 r
F P
0 g
s
.40668 .29691 m .40296 .29974 L p .40119 .31357 L 0 .913 0 r
F P
0 g
s
.39936 .3279 m .40315 .31185 L p .40119 .31357 L 0 .913 0 r
F P
0 g
s
.40315 .31185 m .40668 .29691 L p .40119 .31357 L 0 .913 0 r
F P
0 g
s
.4027 .33031 m .39936 .3279 L p .40104 .31476 L 0 .913 0 r
F P
0 g
s
.39867 .33223 m .4027 .33031 L p .39789 .31647 L 0 .871 0 r
F P
0 g
s
.38935 .32542 m .39631 .30895 L p .39806 .3111 L 0 0 0 r
F P
0 g
s
.39245 .30082 m .3959 .31822 L p .39789 .31647 L 0 .871 0 r
F P
0 g
s
.3959 .31822 m .39867 .33223 L p .39789 .31647 L 0 .871 0 r
F P
0 g
s
.40296 .29974 m .40281 .31694 L p .40104 .31476 L 0 .913 0 r
F P
0 g
s
.40281 .31694 m .4027 .33031 L p .40104 .31476 L 0 .913 0 r
F P
0 g
s
.39631 .30895 m .40311 .29285 L p .39806 .3111 L 0 0 0 r
F P
0 g
s
.47727 .3282 m .4817 .34484 L .495 .34118 L p .544 0 0 r
F P
0 g
s
.4817 .34484 m .47727 .3282 L .464 .32681 L p .533 0 0 r
F P
0 g
s
.57539 .33052 m .57981 .34054 L .59625 .33449 L p 0 0 .846 r
F P
0 g
s
.57981 .34054 m .57539 .33052 L .55992 .33416 L p 0 0 .884 r
F P
0 g
s
.20521 .24505 m .22462 .23903 L .24014 .2392 L p 0 0 .918 r
F P
0 g
s
.22462 .23903 m .20521 .24505 L .19172 .25067 L p 0 0 0 r
F P
0 g
s
.73652 .50843 m .75988 .4833 L .7445 .47441 L p 0 .591 0 r
F P
0 g
s
.7445 .47441 m .72418 .49637 L .73652 .50843 L p 0 .591 0 r
F P
0 g
s
.5528 .32986 m .55992 .33416 L .57539 .33052 L p 0 0 .978 r
F P
0 g
s
.65278 .33803 m .64687 .34914 L .66047 .35593 L p .276 0 0 r
F P
0 g
s
.66047 .35593 m .66582 .33449 L .65278 .33803 L p .276 0 0 r
F P
0 g
s
.59625 .33449 m .59022 .32137 L .57539 .33052 L p 0 0 .846 r
F P
0 g
s
.72418 .49637 m .7445 .47441 L .73211 .46323 L p 0 .457 0 r
F P
0 g
s
.16518 .32628 m .17231 .30566 L .166 .2923 L p 0 0 .663 r
F P
0 g
s
.17231 .30566 m .16518 .32628 L .17934 .334 L p 0 0 .775 r
F P
0 g
s
.6079 .29375 m .59259 .28306 L .58944 .3004 L p 0 0 .676 r
F P
0 g
s
.58944 .3004 m .60174 .30834 L .6079 .29375 L p 0 0 .676 r
F P
0 g
s
.73766 .25439 m .74743 .28352 L .73942 .28879 L p 0 .773 0 r
F P
0 g
s
.74743 .28352 m .73766 .25439 L .74912 .24514 L p 0 .724 0 r
F P
0 g
s
.72489 .45178 m .736 .42961 L .73211 .42347 L p 0 0 0 r
F P
0 g
s
.73211 .42347 m .74007 .39616 L .73868 .39303 L p 0 .861 0 r
F P
0 g
s
.73868 .39303 m .74598 .36182 L .7415 .35958 L p 0 .889 0 r
F P
0 g
s
.27999 .30649 m .25982 .30925 L .26745 .29398 L p 0 0 .524 r
F P
0 g
s
.26745 .29398 m .28654 .29001 L .27999 .30649 L p 0 0 .524 r
F P
0 g
s
.28654 .29001 m .26745 .29398 L .26907 .2765 L p 0 0 .675 r
F P
0 g
s
.50399 .43377 m .4676 .41931 L .48016 .41871 L p .636 0 0 r
F P
0 g
s
.48016 .41871 m .51537 .4268 L .50399 .43377 L p .636 0 0 r
F P
0 g
s
.26907 .2765 m .28691 .27272 L .28654 .29001 L p 0 0 .675 r
F P
0 g
s
.28691 .27272 m .26907 .2765 L .2644 .25975 L p 0 0 .807 r
F P
0 g
s
.56779 .24766 m .54702 .2457 L .56047 .25404 L p 0 0 0 r
F P
0 g
s
.56047 .25404 m .58067 .25418 L .56779 .24766 L p 0 0 0 r
F P
0 g
s
.70089 .40662 m .66067 .42784 L .64731 .4241 L p .583 0 0 r
F P
0 g
s
.6162 .43071 m .64731 .4241 L .66067 .42784 L p .528 0 0 r
F P
0 g
s
.64731 .4241 m .6162 .43071 L .60665 .41709 L p .736 0 0 r
F P
0 g
s
.18427 .31782 m .17934 .334 L .19618 .33954 L p 0 0 .825 r
F P
0 g
s
.19618 .33954 m .19973 .3266 L .18427 .31782 L p 0 0 .825 r
F P
0 g
s
.20562 .32086 m .19973 .3266 L .21593 .33045 L p 0 0 .949 r
F P
0 g
s
.21593 .33045 m .22119 .32533 L .20562 .32086 L p 0 0 .949 r
F P
0 g
s
.34434 .27872 m .35973 .23835 L .37648 .24261 L p 0 .671 0 r
F P
0 g
s
.35973 .23835 m .34434 .27872 L .33106 .2787 L p 0 .759 0 r
F P
0 g
s
.34434 .27872 m .32678 .32178 L .31879 .32323 L p 0 .733 0 r
F P
0 g
s
.31879 .32323 m .33106 .2787 L .34434 .27872 L p 0 .733 0 r
F P
0 g
s
.166 .2923 m .17926 .27795 L .18251 .26229 L p 0 0 .363 r
F P
0 g
s
.17926 .27795 m .166 .2923 L .17231 .30566 L p 0 0 .668 r
F P
0 g
s
.736 .42961 m .72489 .45178 L .73211 .46323 L p 0 .269 0 r
F P
0 g
s
.41232 .26563 m .40311 .29285 L .393 .28827 L p 0 .227 0 r
F P
0 g
s
.393 .28827 m .40518 .25747 L .41232 .26563 L p 0 .227 0 r
F P
0 g
s
.61889 .36821 m .60008 .37061 L .60655 .36197 L p .88 0 0 r
F P
0 g
s
.60655 .36197 m .62415 .35596 L .61889 .36821 L p .88 0 0 r
F P
0 g
s
.41016 .42561 m .40447 .40058 L .39909 .40486 L p 0 .806 0 r
F P
0 g
s
.48286 .29086 m .5051 .29678 L .51206 .31262 L p 0 0 .463 r
F P
0 g
s
.5051 .29678 m .48286 .29086 L .48407 .27409 L p 0 0 .619 r
F P
0 g
s
.48407 .27409 m .50391 .27924 L .5051 .29678 L p 0 0 .619 r
F P
0 g
s
.66067 .42784 m .70089 .40662 L .70794 .39448 L p .81 0 0 r
F P
0 g
s
.17934 .334 m .18427 .31782 L .17231 .30566 L p 0 0 .775 r
F P
0 g
s
.26838 .3194 m .24742 .31972 L .25982 .30925 L p 0 0 .329 r
F P
0 g
s
.25982 .30925 m .27999 .30649 L .26838 .3194 L p 0 0 .329 r
F P
0 g
s
.39982 .44088 m .41251 .46531 L .41802 .45243 L p 0 .766 0 r
F P
0 g
s
.41802 .45243 m .40839 .43276 L .39982 .44088 L p 0 .766 0 r
F P
0 g
s
.5676 .32604 m .57539 .33052 L .59022 .32137 L p 0 0 .907 r
F P
0 g
s
.57539 .33052 m .5676 .32604 L .5528 .32986 L p 0 0 .978 r
F P
0 g
s
.2644 .25975 m .28101 .25751 L .28691 .27272 L p 0 0 .807 r
F P
0 g
s
.28101 .25751 m .2644 .25975 L .25419 .24655 L p 0 0 .885 r
F P
0 g
s
.51371 .35382 m .53243 .35834 L .5375 .36817 L p .146 0 0 r
F P
0 g
s
.51228 .35199 m .50631 .34439 L p .53243 .35834 L .42 0 0 r
F P
0 g
s
.50631 .34439 m .52359 .35474 L .53243 .35834 L p .42 0 0 r
F P
0 g
s
.53243 .35834 m .52359 .35474 L p .54399 .36198 L .438 0 0 r
F P
0 g
s
.55265 .36505 m .53243 .35834 L p .54399 .36198 L .438 0 0 r
F P
0 g
s
.53243 .35834 m .51371 .35382 L .51228 .35199 L p .42 0 0 r
F P
0 g
s
.50391 .27924 m .48407 .27409 L .49108 .26047 L p 0 0 .748 r
F P
0 g
s
.49108 .26047 m .5087 .26294 L .50391 .27924 L p 0 0 .748 r
F P
0 g
s
.74879 .20695 m .76463 .23764 L .74912 .24514 L p 0 .743 0 r
F P
0 g
s
.76463 .23764 m .74879 .20695 L .76491 .20004 L p 0 .697 0 r
F P
0 g
s
.76491 .20004 m .78161 .23314 L .76463 .23764 L p 0 .697 0 r
F P
0 g
s
.54997 .41859 m .52592 .41445 L .53391 .39881 L p .587 0 0 r
F P
0 g
s
.53391 .39881 m .5549 .40163 L .54997 .41859 L p .587 0 0 r
F P
0 g
s
.5549 .40163 m .53391 .39881 L .538 .38247 L p .538 0 0 r
F P
0 g
s
.82124 .31654 m .80968 .27287 L .81956 .27509 L p 0 .04 0 r
F P
0 g
s
.80968 .27287 m .79725 .2324 L .80889 .23559 L p 0 .354 0 r
F P
0 g
s
.80889 .23559 m .81956 .27509 L .80968 .27287 L p 0 .354 0 r
F P
0 g
s
.4876 .30793 m .51206 .31262 L .52362 .32407 L p 0 0 .234 r
F P
0 g
s
.51206 .31262 m .4876 .30793 L .48286 .29086 L p 0 0 .463 r
F P
0 g
s
.82639 .36338 m .82124 .31654 L .82655 .31872 L p 0 .806 0 r
F P
0 g
s
.82655 .31872 m .8245 .32142 L p .82423 .3243 L 0 .724 0 r
F P
0 g
s
.82124 .31654 m .82452 .34639 L p .82466 .3571 L 0 0 0 r
F P
0 g
s
.82039 .36516 m .82437 .33512 L p .82423 .3243 L 0 .724 0 r
F P
0 g
s
.82437 .33512 m .82655 .31872 L p .82423 .3243 L 0 .724 0 r
F P
0 g
s
.82639 .36338 m .82501 .36121 L p .82466 .3571 L 0 0 0 r
F P
0 g
s
.82452 .34639 m .82639 .36338 L p .82466 .3571 L 0 0 0 r
F P
0 g
s
.81956 .27509 m .82655 .31872 L .82124 .31654 L p 0 .04 0 r
F P
0 g
s
.60174 .30834 m .58944 .3004 L .58058 .31562 L p 0 0 .81 r
F P
0 g
s
.58058 .31562 m .59022 .32137 L .60174 .30834 L p 0 0 .81 r
F P
0 g
s
.54319 .4334 m .51537 .4268 L .52592 .41445 L p .62 0 0 r
F P
0 g
s
.52592 .41445 m .54997 .41859 L .54319 .4334 L p .62 0 0 r
F P
0 g
s
.42809 .3907 m .4676 .41931 L .4578 .41323 L p .528 0 0 r
F P
0 g
s
.4578 .41323 m .42538 .37583 L .42809 .3907 L p .528 0 0 r
F P
0 g
s
.81301 .44758 m .82227 .40514 L .82057 .40783 L p 0 .925 0 r
F P
0 g
s
.82057 .40783 m .8069 .45076 L .81301 .44758 L p 0 .925 0 r
F P
0 g
s
.78852 .48852 m .8069 .45076 L .79475 .45108 L p 0 .818 0 r
F P
0 g
s
.79475 .45108 m .77549 .48829 L .78852 .48852 L p 0 .818 0 r
F P
0 g
s
.73514 .21844 m .74912 .24514 L .73766 .25439 L p 0 .774 0 r
F P
0 g
s
.74912 .24514 m .73514 .21844 L .74879 .20695 L p 0 .743 0 r
F P
0 g
s
.19172 .25067 m .21032 .24615 L .22462 .23903 L p 0 0 0 r
F P
0 g
s
.21032 .24615 m .19172 .25067 L .18251 .26229 L p 0 0 .16 r
F P
0 g
s
.82501 .36121 m .82057 .40783 L .82227 .40514 L p 0 .75 0 r
F P
0 g
s
.40818 .22036 m .39252 .24935 L .37648 .24261 L p 0 .59 0 r
F P
0 g
s
.37648 .24261 m .39407 .20956 L .40818 .22036 L p 0 .59 0 r
F P
0 g
s
.39407 .20956 m .37648 .24261 L .35973 .23835 L p 0 .696 0 r
F P
0 g
s
.82057 .40783 m .82501 .36121 L .82639 .36338 L p 0 .813 0 r
F P
0 g
s
.464 .32681 m .46876 .35481 L .4817 .34484 L p .533 0 0 r
F P
0 g
s
.46876 .35481 m .464 .32681 L .44983 .3323 L p .484 0 0 r
F P
0 g
s
.59022 .32137 m .58058 .31562 L .5676 .32604 L p 0 0 .907 r
F P
0 g
s
.3566 .49256 m .3387 .45676 L .33062 .45298 L p 0 .305 0 r
F P
0 g
s
.3387 .45676 m .3566 .49256 L .37064 .49267 L p 0 .57 0 r
F P
0 g
s
.58535 .41999 m .60665 .41709 L .6162 .43071 L p .759 0 0 r
F P
0 g
s
.60665 .41709 m .58535 .41999 L .58216 .40261 L p .803 0 0 r
F P
0 g
s
.58216 .40261 m .60018 .40072 L .60665 .41709 L p .803 0 0 r
F P
0 g
s
.60018 .40072 m .58216 .40261 L .58146 .38625 L p .827 0 0 r
F P
0 g
s
.58146 .38625 m .59784 .38431 L .60018 .40072 L p .827 0 0 r
F P
0 g
s
.66582 .33449 m .66047 .35593 L .6724 .36842 L p .574 0 0 r
F P
0 g
s
.6724 .36842 m .68043 .33786 L .66582 .33449 L p .574 0 0 r
F P
0 g
s
.41814 .23336 m .40518 .25747 L .39252 .24935 L p 0 .45 0 r
F P
0 g
s
.39252 .24935 m .40818 .22036 L .41814 .23336 L p 0 .45 0 r
F P
0 g
s
.78161 .23314 m .76491 .20004 L .78079 .19889 L p 0 .624 0 r
F P
0 g
s
.78079 .19889 m .79725 .2324 L .78161 .23314 L p 0 .624 0 r
F P
0 g
s
.21713 .32005 m .20562 .32086 L .22119 .32533 L p 0 0 .982 r
F P
0 g
s
.19194 .31009 m .18427 .31782 L .19973 .3266 L p 0 0 .891 r
F P
0 g
s
.19973 .3266 m .20562 .32086 L .19194 .31009 L p 0 0 .891 r
F P
0 g
s
.17231 .30566 m .18259 .29489 L .17926 .27795 L p 0 0 .668 r
F P
0 g
s
.18259 .29489 m .17231 .30566 L .18427 .31782 L p 0 0 .807 r
F P
0 g
s
.58067 .25418 m .56047 .25404 L .57019 .26815 L p 0 0 .259 r
F P
0 g
s
.57019 .26815 m .58947 .2667 L .58067 .25418 L p 0 0 .259 r
F P
0 g
s
.55717 .38532 m .538 .38247 L .5375 .36817 L p .45 0 0 r
F P
0 g
s
.538 .38247 m .55717 .38532 L .5549 .40163 L p .538 0 0 r
F P
0 g
s
.5087 .26294 m .49108 .26047 L .50276 .25238 L p 0 0 .834 r
F P
0 g
s
.50276 .25238 m .51869 .25066 L .5087 .26294 L p 0 0 .834 r
F P
0 g
s
.73836 .29351 m .73942 .28879 L p .74113 .3142 L 0 .796 0 r
F P
0 g
s
.74172 .32296 m .73836 .29351 L p .74113 .3142 L 0 .796 0 r
F P
0 g
s
.73221 .26387 m .73942 .28879 L .73836 .29351 L p 0 .624 0 r
F P
0 g
s
.73942 .28879 m .73221 .26387 L .73766 .25439 L p 0 .773 0 r
F P
0 g
s
.73836 .29351 m .74458 .29685 L p .73624 .27761 L 0 .138 0 r
F P
0 g
s
.73379 .27196 m .73221 .26387 L p .73553 .27984 L 0 .624 0 r
F P
0 g
s
.73836 .29351 m .73379 .27196 L p .73553 .27984 L 0 .624 0 r
F P
0 g
s
.72358 .247 m .73221 .26387 L .73379 .27196 L p 0 .484 0 r
F P
0 g
s
.73379 .27196 m .74224 .27725 L p .73222 .26486 L 0 .256 0 r
F P
0 g
s
.72771 .25928 m .72358 .247 L p .72928 .26093 L 0 .484 0 r
F P
0 g
s
.73379 .27196 m .72771 .25928 L p .72928 .26093 L 0 .484 0 r
F P
0 g
s
.72623 .23257 m .73766 .25439 L .73221 .26387 L p 0 .767 0 r
F P
0 g
s
.73221 .26387 m .72358 .247 L .72623 .23257 L p 0 .767 0 r
F P
0 g
s
.71229 .23664 m .72358 .247 L .72771 .25928 L p 0 .507 0 r
F P
0 g
s
.71887 .25163 m .71229 .23664 L p .72196 .25084 L 0 .507 0 r
F P
0 g
s
.72771 .25928 m .71887 .25163 L p .72196 .25084 L 0 .507 0 r
F P
0 g
s
.72771 .25928 m .73799 .26727 L p .72592 .2574 L 0 .127 0 r
F P
0 g
s
.71887 .25163 m .72771 .25928 L p .72592 .2574 L 0 .127 0 r
F P
0 g
s
.72771 .25928 m .73379 .27196 L p .73222 .26486 L 0 .256 0 r
F P
0 g
s
.74224 .27725 m .73799 .26727 L .72771 .25928 L p 0 .256 0 r
F P
0 g
s
.71887 .25163 m .7308 .26162 L p .71452 .25275 L 0 .175 0 r
F P
0 g
s
.70395 .24699 m .71887 .25163 L p .71452 .25275 L 0 .175 0 r
F P
0 g
s
.73379 .27196 m .73836 .29351 L p .73624 .27761 L 0 .138 0 r
F P
0 g
s
.73836 .29351 m .74172 .32296 L .74281 .32549 L p 0 0 0 r
F P
0 g
s
.52362 .32407 m .49747 .32244 L .4876 .30793 L p 0 0 .234 r
F P
0 g
s
.25363 .32654 m .23234 .32357 L .24742 .31972 L p 0 0 .028 r
F P
0 g
s
.24742 .31972 m .26838 .3194 L .25363 .32654 L p 0 0 .028 r
F P
0 g
s
.70794 .39448 m .67242 .42477 L .66067 .42784 L p .81 0 0 r
F P
0 g
s
.62724 .43929 m .66067 .42784 L .67242 .42477 L p 0 0 0 r
F P
0 g
s
.66067 .42784 m .62724 .43929 L .6162 .43071 L p .528 0 0 r
F P
0 g
s
.35973 .23835 m .37817 .20276 L .39407 .20956 L p 0 .696 0 r
F P
0 g
s
.37817 .20276 m .35973 .23835 L .34505 .23731 L p 0 .803 0 r
F P
0 g
s
.33106 .2787 m .34505 .23731 L .35973 .23835 L p 0 .759 0 r
F P
0 g
s
.53569 .44358 m .50399 .43377 L .51537 .4268 L p .641 0 0 r
F P
0 g
s
.51537 .4268 m .54319 .4334 L .53569 .44358 L p .641 0 0 r
F P
0 g
s
.25419 .24655 m .26979 .24701 L .28101 .25751 L p 0 0 .885 r
F P
0 g
s
.26979 .24701 m .25419 .24655 L .24014 .2392 L p 0 0 .688 r
F P
0 g
s
.4676 .41931 m .50399 .43377 L .49371 .43411 L p .599 0 0 r
F P
0 g
s
.49371 .43411 m .4578 .41323 L .4676 .41931 L p .599 0 0 r
F P
0 g
s
.7415 .35958 m .7511 .32736 L .74281 .32549 L p 0 .831 0 r
F P
0 g
s
.18427 .31782 m .19194 .31009 L .18259 .29489 L p 0 0 .807 r
F P
0 g
s
.32055 .41171 m .33062 .45298 L .3387 .45676 L p 0 .06 0 r
F P
0 g
s
.34505 .23731 m .33106 .2787 L .32338 .28055 L p 0 .865 0 r
F P
0 g
s
.31786 .3254 m .32338 .28055 L .33106 .2787 L p 0 .812 0 r
F P
0 g
s
.31879 .32323 m .31786 .3254 L p .32448 .30199 L 0 .812 0 r
F P
0 g
s
.32484 .30127 m .31879 .32323 L p .32448 .30199 L 0 .812 0 r
F P
0 g
s
.31786 .3254 m .3243 .32793 L p .32384 .30408 L 0 .63 0 r
F P
0 g
s
.32338 .28055 m .32062 .30298 L p .32384 .30408 L 0 .63 0 r
F P
0 g
s
.32062 .30298 m .31786 .3254 L p .32384 .30408 L 0 .63 0 r
F P
0 g
s
.32274 .28399 m .32338 .28055 L p .32382 .30329 L 0 .63 0 r
F P
0 g
s
.33106 .2787 m .32484 .30127 L p .32448 .30199 L 0 .812 0 r
F P
0 g
s
.32939 .28844 m .32274 .28399 L p .32956 .30604 L 0 .146 0 r
F P
0 g
s
.33709 .33038 m .33283 .30715 L p .32956 .30604 L 0 .146 0 r
F P
0 g
s
.33283 .30715 m .32939 .28844 L p .32956 .30604 L 0 .146 0 r
F P
0 g
s
.3243 .32793 m .32345 .30403 L p .32382 .30329 L 0 .63 0 r
F P
0 g
s
.32345 .30403 m .32274 .28399 L p .32382 .30329 L 0 .63 0 r
F P
0 g
s
.74638 .43664 m .73211 .46323 L .7445 .47441 L p 0 .465 0 r
F P
0 g
s
.73211 .46323 m .74638 .43664 L .736 .42961 L p 0 .269 0 r
F P
0 g
s
.67242 .42477 m .70794 .39448 L .70914 .37872 L p .845 0 0 r
F P
0 g
s
.62415 .35596 m .60655 .36197 L .61623 .35995 L p .436 0 0 r
F P
0 g
s
.61623 .35995 m .63402 .34918 L .62415 .35596 L p .436 0 0 r
F P
0 g
s
.55238 .32698 m .5676 .32604 L .58058 .31562 L p 0 0 .936 r
F P
0 g
s
.5905 .43553 m .6162 .43071 L .62724 .43929 L p .666 0 0 r
F P
0 g
s
.6162 .43071 m .5905 .43553 L .58535 .41999 L p .759 0 0 r
F P
0 g
s
.59784 .38431 m .58146 .38625 L .5834 .37364 L p .83 0 0 r
F P
0 g
s
.5834 .37364 m .60008 .37061 L .59784 .38431 L p .83 0 0 r
F P
0 g
s
.73766 .25439 m .72623 .23257 L .73514 .21844 L p 0 .774 0 r
F P
0 g
s
.38588 .44854 m .40135 .47771 L .41251 .46531 L p 0 .745 0 r
F P
0 g
s
.41251 .46531 m .39982 .44088 L .38588 .44854 L p 0 .745 0 r
F P
0 g
s
.41265 .27244 m .40668 .29691 L .40311 .29285 L p 0 0 0 r
F P
0 g
s
.40311 .29285 m .41232 .26563 L .41246 .26844 L p 0 0 0 r
F P
0 g
s
.40311 .29285 m .41246 .26844 L .41005 .27802 L 0 0 0 r
F
.40599 .27669 m .40296 .29974 L .40495 .29823 L p 0 .948 0 r
F P
0 g
s
.40495 .29823 m .40668 .29691 L p .40651 .29203 L 0 .948 0 r
F P
0 g
s
.40651 .29203 m .40599 .27669 L .40495 .29823 L 0 .948 0 r
F
.41246 .26844 m .41265 .27244 L p .41005 .27802 L 0 0 0 r
F P
0 g
s
.45251 .40142 m .42862 .3592 L .42538 .37583 L p .007 0 0 r
F P
0 g
s
.42538 .37583 m .4578 .41323 L .45251 .40142 L p .007 0 0 r
F P
0 g
s
.74852 .39996 m .736 .42961 L .74638 .43664 L p 0 .323 0 r
F P
0 g
s
.736 .42961 m .74852 .39996 L .74007 .39616 L p 0 0 0 r
F P
0 g
s
.56831 .42049 m .54997 .41859 L .5549 .40163 L p .656 0 0 r
F P
0 g
s
.5549 .40163 m .56931 .40284 L .56831 .42049 L p .656 0 0 r
F P
0 g
s
.56931 .40284 m .5549 .40163 L .55717 .38532 L p .631 0 0 r
F P
0 g
s
.56727 .43654 m .54319 .4334 L .54997 .41859 L p .668 0 0 r
F P
0 g
s
.54997 .41859 m .56831 .42049 L .56727 .43654 L p .668 0 0 r
F P
0 g
s
.44983 .3323 m .45848 .36936 L .46876 .35481 L p .484 0 0 r
F P
0 g
s
.45848 .36936 m .44983 .3323 L .43729 .34377 L p .412 0 0 r
F P
0 g
s
.42227 .2464 m .41232 .26563 L .40518 .25747 L p 0 .194 0 r
F P
0 g
s
.40518 .25747 m .41814 .23336 L .42227 .2464 L p 0 .194 0 r
F P
0 g
s
.18251 .26229 m .19979 .25938 L .21032 .24615 L p 0 0 .16 r
F P
0 g
s
.19979 .25938 m .18251 .26229 L .17926 .27795 L p 0 0 .494 r
F P
0 g
s
.75717 .36394 m .74007 .39616 L .74852 .39996 L p 0 .204 0 r
F P
0 g
s
.74007 .39616 m .75717 .36394 L .74598 .36182 L p 0 0 0 r
F P
0 g
s
.77549 .48829 m .79475 .45108 L .77864 .44845 L p 0 .694 0 r
F P
0 g
s
.77864 .44845 m .75988 .4833 L .77549 .48829 L p 0 .694 0 r
F P
0 g
s
.56831 .42049 m .58535 .41999 L .5905 .43553 L p .724 0 0 r
F P
0 g
s
.58535 .41999 m .56831 .42049 L .56931 .40284 L p .74 0 0 r
F P
0 g
s
.56931 .40284 m .58216 .40261 L .58535 .41999 L p .74 0 0 r
F P
0 g
s
.58216 .40261 m .56931 .40284 L .57012 .38648 L p .741 0 0 r
F P
0 g
s
.57012 .38648 m .58146 .38625 L .58216 .40261 L p .741 0 0 r
F P
0 g
s
.38273 .37123 m .39909 .40486 L .40447 .40058 L p 0 .806 0 r
F P
0 g
s
.40447 .40058 m .39532 .36919 L .38273 .37123 L p 0 .806 0 r
F P
0 g
s
.37064 .49267 m .3523 .45726 L .3387 .45676 L p 0 .57 0 r
F P
0 g
s
.3523 .45726 m .37064 .49267 L .38651 .48741 L p 0 .661 0 r
F P
0 g
s
.52359 .35474 m .50631 .34439 L .495 .34118 L p .645 0 0 r
F P
0 g
s
.58947 .2667 m .57019 .26815 L .57444 .28553 L p 0 0 .518 r
F P
0 g
s
.57444 .28553 m .59259 .28306 L .58947 .2667 L p 0 0 .518 r
F P
0 g
s
.38725 .40926 m .39982 .44088 L .40839 .43276 L p 0 .768 0 r
F P
0 g
s
.40839 .43276 m .39909 .40486 L .38725 .40926 L p 0 .768 0 r
F P
0 g
s
.20446 .3097 m .19194 .31009 L .20562 .32086 L p 0 0 .948 r
F P
0 g
s
.20562 .32086 m .21713 .32005 L .20446 .3097 L p 0 0 .948 r
F P
0 g
s
.68043 .33786 m .6724 .36842 L .68052 .38441 L p .71 0 0 r
F P
0 g
s
.68052 .38441 m .69404 .34764 L .68043 .33786 L p .71 0 0 r
F P
0 g
s
.79725 .2324 m .78079 .19889 L .79374 .20378 L p 0 .465 0 r
F P
0 g
s
.79374 .20378 m .80889 .23559 L .79725 .2324 L p 0 .465 0 r
F P
0 g
s
.51869 .25066 m .50276 .25238 L .51712 .25127 L p 0 0 .648 r
F P
0 g
s
.51712 .25127 m .53223 .24453 L .51869 .25066 L p 0 0 .648 r
F P
0 g
s
.3387 .45676 m .32552 .41463 L .32055 .41171 L p 0 .06 0 r
F P
0 g
s
.55717 .38532 m .57012 .38648 L .56931 .40284 L p .631 0 0 r
F P
0 g
s
.5905 .43553 m .56727 .43654 L .56831 .42049 L p .724 0 0 r
F P
0 g
s
.82655 .31872 m .81956 .27509 L .82275 .2791 L p 0 .729 0 r
F P
0 g
s
.81956 .27509 m .80889 .23559 L .81448 .24222 L p 0 0 0 r
F P
0 g
s
.81448 .24222 m .82275 .2791 L .81956 .27509 L p 0 0 0 r
F P
0 g
s
.5375 .36817 m .55639 .37238 L .55717 .38532 L p .45 0 0 r
F P
0 g
s
.55639 .37238 m .5375 .36817 L .53243 .35834 L p .201 0 0 r
F P
0 g
s
.45273 .38588 m .43729 .34377 L .42862 .3592 L p .288 0 0 r
F P
0 g
s
.42862 .3592 m .45251 .40142 L .45273 .38588 L p .288 0 0 r
F P
0 g
s
.76149 .44333 m .7445 .47441 L .75988 .4833 L p 0 .587 0 r
F P
0 g
s
.7445 .47441 m .76149 .44333 L .74638 .43664 L p 0 .465 0 r
F P
0 g
s
.56463 .31778 m .58058 .31562 L .58944 .3004 L p 0 0 .82 r
F P
0 g
s
.58058 .31562 m .56463 .31778 L .55238 .32698 L p 0 0 .936 r
F P
0 g
s
.43729 .34377 m .45273 .38588 L .45848 .36936 L p .412 0 0 r
F P
0 g
s
.23234 .32357 m .25363 .32654 L .23822 .32664 L p 0 0 0 r
F P
0 g
s
.36903 .45435 m .38651 .48741 L .40135 .47771 L p 0 .71 0 r
F P
0 g
s
.40135 .47771 m .38588 .44854 L .36903 .45435 L p 0 .71 0 r
F P
0 g
s
.8069 .45076 m .82057 .40783 L .81181 .40914 L p 0 .796 0 r
F P
0 g
s
.81181 .40914 m .79475 .45108 L .8069 .45076 L p 0 .796 0 r
F P
0 g
s
.75988 .4833 m .77864 .44845 L .76149 .44333 L p 0 .587 0 r
F P
0 g
s
.82655 .31872 m .82039 .36516 L .82639 .36338 L p 0 .806 0 r
F P
0 g
s
.82275 .2791 m .8245 .32142 L .82655 .31872 L p 0 .729 0 r
F P
0 g
s
.70914 .37872 m .68047 .41531 L .67242 .42477 L p .845 0 0 r
F P
0 g
s
.68047 .41531 m .70914 .37872 L .70422 .36212 L p .828 0 0 r
F P
0 g
s
.24014 .2392 m .25511 .24309 L .26979 .24701 L p 0 0 .688 r
F P
0 g
s
.25511 .24309 m .24014 .2392 L .22462 .23903 L p 0 0 0 r
F P
0 g
s
.38651 .48741 m .36903 .45435 L .3523 .45726 L p 0 .661 0 r
F P
0 g
s
.34505 .23731 m .36312 .20113 L .37817 .20276 L p 0 .803 0 r
F P
0 g
s
.36312 .20113 m .34505 .23731 L .33493 .2397 L p 0 .929 0 r
F P
0 g
s
.32338 .28055 m .33493 .2397 L .34505 .23731 L p 0 .865 0 r
F P
0 g
s
.57012 .38648 m .55717 .38532 L .55639 .37238 L p .579 0 0 r
F P
0 g
s
.56637 .44829 m .53569 .44358 L .54319 .4334 L p .662 0 0 r
F P
0 g
s
.54319 .4334 m .56727 .43654 L .56637 .44829 L p .662 0 0 r
F P
0 g
s
.59259 .28306 m .57444 .28553 L .57246 .30311 L p 0 0 .686 r
F P
0 g
s
.57246 .30311 m .58944 .3004 L .59259 .28306 L p 0 0 .686 r
F P
0 g
s
.17926 .27795 m .19491 .27643 L .19979 .25938 L p 0 0 .494 r
F P
0 g
s
.19491 .27643 m .17926 .27795 L .18259 .29489 L p 0 0 .693 r
F P
0 g
s
.82639 .36338 m .81181 .40914 L .82057 .40783 L p 0 .813 0 r
F P
0 g
s
.52871 .44735 m .49371 .43411 L .50399 .43377 L p .597 0 0 r
F P
0 g
s
.50399 .43377 m .53569 .44358 L .52871 .44735 L p .597 0 0 r
F P
0 g
s
.58944 .3004 m .57246 .30311 L .56463 .31778 L p 0 0 .82 r
F P
0 g
s
.62724 .43929 m .59679 .44662 L .5905 .43553 L p .666 0 0 r
F P
0 g
s
.56727 .43654 m .5905 .43553 L .59679 .44662 L p .679 0 0 r
F P
0 g
s
.58146 .38625 m .57012 .38648 L .57059 .37414 L p .719 0 0 r
F P
0 g
s
.57059 .37414 m .5834 .37364 L .58146 .38625 L p .719 0 0 r
F P
0 g
s
.81181 .40914 m .82639 .36338 L .82039 .36516 L p 0 .706 0 r
F P
0 g
s
.19657 .29426 m .18259 .29489 L .19194 .31009 L p 0 0 .836 r
F P
0 g
s
.19194 .31009 m .20446 .3097 L .19657 .29426 L p 0 0 .836 r
F P
0 g
s
.69404 .34764 m .68052 .38441 L .68337 .40103 L p .784 0 0 r
F P
0 g
s
.68337 .40103 m .70422 .36212 L .69404 .34764 L p .784 0 0 r
F P
0 g
s
.67242 .42477 m .63793 .44131 L .62724 .43929 L p 0 0 0 r
F P
0 g
s
.63793 .44131 m .67242 .42477 L .68047 .41531 L p .754 0 0 r
F P
0 g
s
.30116 .31252 m .26838 .3194 L .27999 .30649 L p 0 0 .614 r
F P
0 g
s
.27999 .30649 m .30719 .29583 L .30116 .31252 L p 0 0 .614 r
F P
0 g
s
.30719 .29583 m .27999 .30649 L .28654 .29001 L p 0 0 .728 r
F P
0 g
s
.37201 .33333 m .39532 .36919 L .40154 .36682 L p 0 .782 0 r
F P
0 g
s
.40154 .36682 m .38783 .3333 L .37201 .33333 L p 0 .782 0 r
F P
0 g
s
.59679 .44662 m .62724 .43929 L .63793 .44131 L p .357 0 0 r
F P
0 g
s
.4027 .33031 m .40296 .29974 L .39245 .30082 L p 0 .871 0 r
F P
0 g
s
.60008 .37061 m .5834 .37364 L .58768 .36697 L p .735 0 0 r
F P
0 g
s
.58768 .36697 m .60655 .36197 L .60008 .37061 L p .735 0 0 r
F P
0 g
s
.31673 .36861 m .32552 .41463 L .33705 .41593 L p 0 .005 0 r
F P
0 g
s
.28654 .29001 m .30708 .27916 L .30719 .29583 L p 0 0 .728 r
F P
0 g
s
.30708 .27916 m .28654 .29001 L .28691 .27272 L p 0 0 .833 r
F P
0 g
s
.48632 .42764 m .45251 .40142 L .4578 .41323 L p 0 0 0 r
F P
0 g
s
.4578 .41323 m .49371 .43411 L .48632 .42764 L p 0 0 0 r
F P
0 g
s
.70422 .36212 m .68337 .40103 L .68047 .41531 L p .828 0 0 r
F P
0 g
s
.18259 .29489 m .19657 .29426 L .19491 .27643 L p 0 0 .693 r
F P
0 g
s
.55639 .37238 m .57059 .37414 L .57012 .38648 L p .579 0 0 r
F P
0 g
s
.59679 .44662 m .56637 .44829 L .56727 .43654 L p .679 0 0 r
F P
0 g
s
.74458 .29685 m .74281 .32549 L .7511 .32736 L p 0 .248 0 r
F P
0 g
s
.74281 .32549 m .74458 .29685 L .73836 .29351 L p 0 0 0 r
F P
0 g
s
.63402 .34918 m .61623 .35995 L .62748 .36504 L p .343 0 0 r
F P
0 g
s
.62748 .36504 m .64687 .34914 L .63402 .34918 L p .343 0 0 r
F P
0 g
s
.32552 .41463 m .3387 .45676 L .3523 .45726 L p 0 .448 0 r
F P
0 g
s
.43377 .21767 m .41814 .23336 L .40818 .22036 L p 0 .436 0 r
F P
0 g
s
.40818 .22036 m .42716 .20105 L .43377 .21767 L p 0 .436 0 r
F P
0 g
s
.42716 .20105 m .40818 .22036 L .39407 .20956 L p 0 .588 0 r
F P
0 g
s
.39407 .20956 m .41561 .18739 L .42716 .20105 L p 0 .588 0 r
F P
0 g
s
.41561 .18739 m .39407 .20956 L .37817 .20276 L p 0 .72 0 r
F P
0 g
s
.53223 .24453 m .51712 .25127 L .53169 .25745 L p 0 0 0 r
F P
0 g
s
.53169 .25745 m .54702 .2457 L .53223 .24453 L p 0 0 0 r
F P
0 g
s
.76266 .40378 m .74638 .43664 L .76149 .44333 L p 0 .487 0 r
F P
0 g
s
.74638 .43664 m .76266 .40378 L .74852 .39996 L p 0 .323 0 r
F P
0 g
s
.71843 .20229 m .73514 .21844 L .72623 .23257 L p 0 .826 0 r
F P
0 g
s
.73514 .21844 m .71843 .20229 L .7301 .18871 L p 0 .841 0 r
F P
0 g
s
.7301 .18871 m .74879 .20695 L .73514 .21844 L p 0 .841 0 r
F P
0 g
s
.45946 .29242 m .4876 .30793 L .49747 .32244 L p 0 0 .281 r
F P
0 g
s
.4876 .30793 m .45946 .29242 L .46275 .27738 L p 0 0 .465 r
F P
0 g
s
.46275 .27738 m .48286 .29086 L .4876 .30793 L p 0 0 .465 r
F P
0 g
s
.76524 .32822 m .74598 .36182 L .75717 .36394 L p 0 .322 0 r
F P
0 g
s
.74598 .36182 m .76524 .32822 L .7511 .32736 L p 0 0 0 r
F P
0 g
s
.41978 .25727 m .41265 .27244 L .41232 .26563 L p 0 0 0 r
F P
0 g
s
.42871 .24863 m .41978 .25727 L .42227 .2464 L p 0 0 0 r
F P
0 g
s
.41978 .25727 m .42871 .24863 L .42425 .25234 L p 0 .886 0 r
F P
0 g
s
.41247 .26534 m .42227 .2464 L .41978 .25727 L p 0 0 0 r
F P
0 g
s
.42425 .25234 m .41781 .25769 L p .41978 .25727 L 0 .886 0 r
F P
0 g
s
.41232 .26563 m .41247 .26534 L p .41978 .25727 L 0 0 0 r
F P
0 g
s
.29003 .32644 m .25363 .32654 L .26838 .3194 L p 0 0 .489 r
F P
0 g
s
.26838 .3194 m .30116 .31252 L .29003 .32644 L p 0 0 .489 r
F P
0 g
s
.33493 .2397 m .32338 .28055 L .32274 .28399 L p 0 .875 0 r
F P
0 g
s
.74879 .20695 m .7301 .18871 L .7452 .1808 L p 0 .821 0 r
F P
0 g
s
.7452 .1808 m .76491 .20004 L .74879 .20695 L p 0 .821 0 r
F P
0 g
s
.28691 .27272 m .30081 .26532 L .30708 .27916 L p 0 0 .833 r
F P
0 g
s
.30081 .26532 m .28691 .27272 L .28101 .25751 L p 0 0 .802 r
F P
0 g
s
.495 .34118 m .51245 .35812 L .52359 .35474 L p .645 0 0 r
F P
0 g
s
.51245 .35812 m .495 .34118 L .4817 .34484 L p .625 0 0 r
F P
0 g
s
.40668 .29691 m .41265 .27244 L .40599 .27669 L p 0 .948 0 r
F P
0 g
s
.46202 .30944 m .49747 .32244 L .51085 .33196 L p 0 0 0 r
F P
0 g
s
.49747 .32244 m .46202 .30944 L .45946 .29242 L p 0 0 .281 r
F P
0 g
s
.22478 .3196 m .20446 .3097 L .21713 .32005 L p 0 0 .861 r
F P
0 g
s
.21713 .32005 m .23822 .32664 L .22478 .3196 L p 0 0 .861 r
F P
0 g
s
.37097 .41298 m .38588 .44854 L .39982 .44088 L p 0 .709 0 r
F P
0 g
s
.39982 .44088 m .38725 .40926 L .37097 .41298 L p 0 .709 0 r
F P
0 g
s
.48286 .29086 m .46275 .27738 L .47141 .2669 L p 0 0 .577 r
F P
0 g
s
.47141 .2669 m .48407 .27409 L .48286 .29086 L p 0 0 .577 r
F P
0 g
s
.77319 .36555 m .74852 .39996 L .76266 .40378 L p 0 .422 0 r
F P
0 g
s
.74852 .39996 m .77319 .36555 L .75717 .36394 L p 0 .204 0 r
F P
0 g
s
.80889 .23559 m .79374 .20378 L .80148 .21393 L p 0 0 0 r
F P
0 g
s
.80148 .21393 m .81448 .24222 L .80889 .23559 L p 0 0 0 r
F P
0 g
s
.22462 .23903 m .23949 .24652 L .25511 .24309 L p 0 0 0 r
F P
0 g
s
.23949 .24652 m .22462 .23903 L .21032 .24615 L p 0 0 .216 r
F P
0 g
s
.79475 .45108 m .81181 .40914 L .79745 .40882 L p 0 .685 0 r
F P
0 g
s
.79745 .40882 m .77864 .44845 L .79475 .45108 L p 0 .685 0 r
F P
0 g
s
.43433 .23448 m .42227 .2464 L .41814 .23336 L p 0 .21 0 r
F P
0 g
s
.41814 .23336 m .43377 .21767 L .43433 .23448 L p 0 .21 0 r
F P
0 g
s
.71215 .21922 m .72623 .23257 L .72358 .247 L p 0 .753 0 r
F P
0 g
s
.72623 .23257 m .71215 .21922 L .71843 .20229 L p 0 .826 0 r
F P
0 g
s
.36592 .37256 m .38725 .40926 L .39909 .40486 L p 0 .722 0 r
F P
0 g
s
.39909 .40486 m .38273 .37123 L .36592 .37256 L p 0 .722 0 r
F P
0 g
s
.37817 .20276 m .40103 .179 L .41561 .18739 L p 0 .72 0 r
F P
0 g
s
.40103 .179 m .37817 .20276 L .36312 .20113 L p 0 .862 0 r
F P
0 g
s
.55265 .36505 m .55639 .37238 L p .53957 .36252 L .201 0 0 r
F P
0 g
s
.57059 .37414 m .55639 .37238 L .55265 .36505 L p .409 0 0 r
F P
0 g
s
.57065 .36796 m .58768 .36697 L .5834 .37364 L p .6 0 0 r
F P
0 g
s
.59361 .3675 m .58768 .36697 L p .5797 .36796 L .073 0 0 r
F P
0 g
s
.57065 .36815 m .57059 .37414 L p .55922 .36838 L .409 0 0 r
F P
0 g
s
.5834 .37364 m .57059 .37414 L .57065 .36796 L p .6 0 0 r
F P
0 g
s
.57027 .36913 m .59003 .36775 L p .5797 .36796 L .073 0 0 r
F P
0 g
s
.59003 .36775 m .59361 .3675 L p .5797 .36796 L .073 0 0 r
F P
0 g
s
.57065 .36796 m .57065 .36815 L p .5584 .36796 L .409 0 0 r
F P
0 g
s
.5584 .36796 m .57065 .36815 L .55922 .36838 L .409 0 0 r
F
.55265 .36505 m .57065 .36796 L p .5584 .36796 L F P
0 g
s
.57065 .36796 m .57027 .36913 L p .5797 .36796 L .073 0 0 r
F P
0 g
s
.58768 .36697 m .57065 .36796 L p .5797 .36796 L .073 0 0 r
F P
0 g
s
.55265 .36505 m .54652 .36469 L p .56076 .36662 L .264 0 0 r
F P
0 g
s
.57065 .36796 m .55265 .36505 L p .56076 .36662 L .264 0 0 r
F P
0 g
s
.53243 .35834 m .55265 .36505 L p .53957 .36252 L .201 0 0 r
F P
0 g
s
.56576 .45371 m .52871 .44735 L .53569 .44358 L p .555 0 0 r
F P
0 g
s
.53569 .44358 m .56637 .44829 L .56576 .45371 L p .555 0 0 r
F P
0 g
s
.3523 .45726 m .33705 .41593 L .32552 .41463 L p 0 .448 0 r
F P
0 g
s
.76491 .20004 m .7452 .1808 L .76119 .17996 L p 0 .74 0 r
F P
0 g
s
.76119 .17996 m .78079 .19889 L .76491 .20004 L p 0 .74 0 r
F P
0 g
s
.63793 .44131 m .60316 .45134 L .59679 .44662 L p .357 0 0 r
F P
0 g
s
.56637 .44829 m .59679 .44662 L .60316 .45134 L p .479 0 0 r
F P
0 g
s
.8245 .32142 m .82275 .2791 L .81855 .28423 L p 0 .791 0 r
F P
0 g
s
.82275 .2791 m .81448 .24222 L .81294 .25121 L p 0 .781 0 r
F P
0 g
s
.81294 .25121 m .81855 .28423 L .82275 .2791 L p 0 .781 0 r
F P
0 g
s
.78001 .40692 m .76149 .44333 L .77864 .44845 L p 0 .591 0 r
F P
0 g
s
.76149 .44333 m .78001 .40692 L .76266 .40378 L p 0 .487 0 r
F P
0 g
s
.33493 .2397 m .35148 .20501 L .36312 .20113 L p 0 .929 0 r
F P
0 g
s
.35148 .20501 m .33493 .2397 L .33118 .24516 L p 0 .918 0 r
F P
0 g
s
.32274 .28399 m .33118 .24516 L .33493 .2397 L p 0 .875 0 r
F P
0 g
s
.51085 .33196 m .46996 .32557 L .46202 .30944 L p 0 0 0 r
F P
0 g
s
.46996 .32557 m .51085 .33196 L .52548 .33483 L p 0 0 0 r
F P
0 g
s
.2457 .32836 m .23822 .32664 L .24331 .3266 L p 0 0 .328 r
F P
0 g
s
.23822 .32664 m .27564 .33525 L .26039 .33743 L p 0 0 .064 r
F P
0 g
s
.26039 .33743 m .23186 .32314 L p .23822 .32664 L 0 0 .064 r
F P
0 g
s
.23186 .32314 m .22478 .3196 L .22917 .3219 L p 0 0 .064 r
F P
0 g
s
.22917 .3219 m .23822 .32664 L p .23186 .32314 L 0 0 .064 r
F P
0 g
s
.2157 .30655 m .24034 .32706 L p .23748 .32126 L 0 0 .773 r
F P
0 g
s
.22478 .3196 m .25298 .33372 L p .24689 .33251 L 0 0 0 r
F P
0 g
s
.24034 .32706 m .24689 .33251 L .24359 .32857 L p 0 0 .773 r
F P
0 g
s
.24359 .32857 m .23748 .32126 L p .24034 .32706 L 0 0 .773 r
F P
0 g
s
.25298 .33372 m .26039 .33743 L .25571 .33572 L p 0 0 0 r
F P
0 g
s
.25571 .33572 m .24689 .33251 L p .25298 .33372 L 0 0 0 r
F P
0 g
s
.27564 .33525 m .2457 .32836 L p .25363 .32654 L 0 0 .328 r
F P
0 g
s
.24331 .3266 m .25363 .32654 L p .2457 .32836 L 0 0 .328 r
F P
0 g
s
.25363 .32654 m .29003 .32644 L .27564 .33525 L p 0 0 .328 r
F P
0 g
s
.31786 .3254 m .33198 .3723 L .34792 .37294 L p 0 .199 0 r
F P
0 g
s
.82039 .36516 m .79745 .40882 L .81181 .40914 L p 0 .706 0 r
F P
0 g
s
.68047 .41531 m .64643 .4363 L .63793 .44131 L p .754 0 0 r
F P
0 g
s
.64643 .4363 m .68047 .41531 L .68337 .40103 L p .864 0 0 r
F P
0 g
s
.48316 .41539 m .45273 .38588 L .45251 .40142 L p .297 0 0 r
F P
0 g
s
.45251 .40142 m .48632 .42764 L .48316 .41539 L p .297 0 0 r
F P
0 g
s
.52347 .44393 m .48632 .42764 L .49371 .43411 L p 0 0 0 r
F P
0 g
s
.49371 .43411 m .52871 .44735 L .52347 .44393 L p 0 0 0 r
F P
0 g
s
.33705 .41593 m .3523 .45726 L .36903 .45435 L p 0 .579 0 r
F P
0 g
s
.54702 .2457 m .53169 .25745 L .54388 .26988 L p 0 0 .265 r
F P
0 g
s
.54388 .26988 m .56047 .25404 L .54702 .2457 L p 0 0 .265 r
F P
0 g
s
.77864 .44845 m .79745 .40882 L .78001 .40692 L p 0 .591 0 r
F P
0 g
s
.8245 .32142 m .80794 .36622 L .82039 .36516 L p 0 .724 0 r
F P
0 g
s
.80794 .36622 m .8245 .32142 L .81532 .32416 L p 0 .642 0 r
F P
0 g
s
.81855 .28423 m .81532 .32416 L .8245 .32142 L p 0 .791 0 r
F P
0 g
s
.48407 .27409 m .47141 .2669 L .484 .26285 L p 0 0 .554 r
F P
0 g
s
.484 .26285 m .49108 .26047 L .48407 .27409 L p 0 0 .554 r
F P
0 g
s
.28101 .25751 m .28936 .25671 L .30081 .26532 L p 0 0 .802 r
F P
0 g
s
.28936 .25671 m .28101 .25751 L .26979 .24701 L p 0 0 0 r
F P
0 g
s
.53888 .3305 m .56463 .31778 L .57246 .30311 L p 0 0 .85 r
F P
0 g
s
.79745 .40882 m .82039 .36516 L .80794 .36622 L p 0 .619 0 r
F P
0 g
s
.60316 .45134 m .63793 .44131 L .64643 .4363 L p .484 0 0 r
F P
0 g
s
.60316 .45134 m .56576 .45371 L .56637 .44829 L p .479 0 0 r
F P
0 g
s
.60655 .36197 m .58768 .36697 L .59361 .3675 L p 0 0 0 r
F P
0 g
s
.59361 .3675 m .61623 .35995 L .60655 .36197 L p 0 0 0 r
F P
0 g
s
.64687 .34914 m .62748 .36504 L .63831 .37644 L p .642 0 0 r
F P
0 g
s
.63831 .37644 m .66047 .35593 L .64687 .34914 L p .642 0 0 r
F P
0 g
s
.35316 .41535 m .36903 .45435 L .38588 .44854 L p 0 .651 0 r
F P
0 g
s
.38588 .44854 m .37097 .41298 L .35316 .41535 L p 0 .651 0 r
F P
0 g
s
.2157 .30655 m .19657 .29426 L .20446 .3097 L p 0 0 .873 r
F P
0 g
s
.20446 .3097 m .22478 .3196 L .2157 .30655 L p 0 0 .873 r
F P
0 g
s
.21032 .24615 m .22568 .25677 L .23949 .24652 L p 0 0 .216 r
F P
0 g
s
.22568 .25677 m .21032 .24615 L .19979 .25938 L p 0 0 .45 r
F P
0 g
s
.35401 .33229 m .38273 .37123 L .39532 .36919 L p 0 .697 0 r
F P
0 g
s
.39532 .36919 m .37201 .33333 L .35401 .33229 L p 0 .697 0 r
F P
0 g
s
.4817 .34484 m .50096 .36798 L .51245 .35812 L p .625 0 0 r
F P
0 g
s
.50096 .36798 m .4817 .34484 L .46876 .35481 L p .581 0 0 r
F P
0 g
s
.32096 .37076 m .33705 .41593 L .35316 .41535 L p 0 .423 0 r
F P
0 g
s
.35316 .41535 m .33198 .3723 L .32096 .37076 L p 0 .423 0 r
F P
0 g
s
.72358 .247 m .71229 .23664 L .71215 .21922 L p 0 .753 0 r
F P
0 g
s
.74458 .29685 m .74224 .27725 L .73379 .27196 L p 0 .138 0 r
F P
0 g
s
.37693 .29993 m .38783 .3333 L .39867 .33223 L p 0 .76 0 r
F P
0 g
s
.39867 .33223 m .39245 .30082 L .37693 .29993 L p 0 .76 0 r
F P
0 g
s
.36903 .45435 m .35316 .41535 L .33705 .41593 L p 0 .579 0 r
F P
0 g
s
.7511 .32736 m .75712 .29819 L .74458 .29685 L p 0 .248 0 r
F P
0 g
s
.75712 .29819 m .7511 .32736 L .76524 .32822 L p 0 .462 0 r
F P
0 g
s
.42227 .2464 m .43433 .23448 L .42871 .24863 L p 0 0 0 r
F P
0 g
s
.7912 .36636 m .76266 .40378 L .78001 .40692 L p 0 .534 0 r
F P
0 g
s
.76266 .40378 m .7912 .36636 L .77319 .36555 L p 0 .422 0 r
F P
0 g
s
.78277 .32791 m .75717 .36394 L .77319 .36555 L p 0 .475 0 r
F P
0 g
s
.75717 .36394 m .78277 .32791 L .76524 .32822 L p 0 .322 0 r
F P
0 g
s
.36312 .20113 m .38589 .17738 L .40103 .179 L p 0 .862 0 r
F P
0 g
s
.38589 .17738 m .36312 .20113 L .35148 .20501 L p 0 .985 0 r
F P
0 g
s
.41102 .26406 m .40599 .27669 L .41265 .27244 L p 0 .989 0 r
F P
0 g
s
.41265 .27244 m .41978 .25727 L .41102 .26406 L p 0 .989 0 r
F P
0 g
s
.78079 .19889 m .76119 .17996 L .77531 .18641 L p 0 .471 0 r
F P
0 g
s
.77531 .18641 m .79374 .20378 L .78079 .19889 L p 0 .471 0 r
F P
0 g
s
.74224 .27725 m .74458 .29685 L .75712 .29819 L p 0 .457 0 r
F P
0 g
s
.56047 .25404 m .54388 .26988 L .55151 .28637 L p 0 0 .47 r
F P
0 g
s
.55151 .28637 m .57019 .26815 L .56047 .25404 L p 0 0 .47 r
F P
0 g
s
.80794 .36622 m .78001 .40692 L .79745 .40882 L p 0 .619 0 r
F P
0 g
s
.34792 .37294 m .37097 .41298 L .38725 .40926 L p 0 .649 0 r
F P
0 g
s
.38725 .40926 m .36592 .37256 L .34792 .37294 L p 0 .649 0 r
F P
0 g
s
.33118 .24516 m .32274 .28399 L .32939 .28844 L p 0 0 0 r
F P
0 g
s
.32274 .28399 m .3243 .32793 L .33709 .33038 L p 0 .146 0 r
F P
0 g
s
.54871 .31961 m .57246 .30311 L .57444 .28553 L p 0 0 .732 r
F P
0 g
s
.48487 .39944 m .45848 .36936 L .45273 .38588 L p .452 0 0 r
F P
0 g
s
.45273 .38588 m .48316 .41539 L .48487 .39944 L p .452 0 0 r
F P
0 g
s
.52548 .33483 m .48196 .33813 L .46996 .32557 L p 0 0 0 r
F P
0 g
s
.48196 .33813 m .52548 .33483 L .53888 .3305 L p 0 0 .978 r
F P
0 g
s
.21263 .28973 m .19491 .27643 L .19657 .29426 L p 0 0 .755 r
F P
0 g
s
.19657 .29426 m .2157 .30655 L .21263 .28973 L p 0 0 .755 r
F P
0 g
s
.19979 .25938 m .21615 .27208 L .22568 .25677 L p 0 0 .45 r
F P
0 g
s
.21615 .27208 m .19979 .25938 L .19491 .27643 L p 0 0 .616 r
F P
0 g
s
.81448 .24222 m .80148 .21393 L .80257 .22765 L p 0 .693 0 r
F P
0 g
s
.80257 .22765 m .81294 .25121 L .81448 .24222 L p 0 .693 0 r
F P
0 g
s
.39337 .2776 m .39245 .30082 L .40296 .29974 L p 0 .84 0 r
F P
0 g
s
.40296 .29974 m .40599 .27669 L .39337 .2776 L p 0 .84 0 r
F P
0 g
s
.6512 .42503 m .68337 .40103 L .68052 .38441 L p .855 0 0 r
F P
0 g
s
.68337 .40103 m .6512 .42503 L .64643 .4363 L p .864 0 0 r
F P
0 g
s
.49108 .26047 m .484 .26285 L .4984 .26605 L p 0 0 0 r
F P
0 g
s
.4984 .26605 m .50276 .25238 L .49108 .26047 L p 0 0 0 r
F P
0 g
s
.78001 .40692 m .80794 .36622 L .7912 .36636 L p 0 .534 0 r
F P
0 g
s
.46876 .35481 m .49117 .38262 L .50096 .36798 L p .581 0 0 r
F P
0 g
s
.49117 .38262 m .46876 .35481 L .45848 .36936 L p .529 0 0 r
F P
0 g
s
.19491 .27643 m .21263 .28973 L .21615 .27208 L p 0 0 .616 r
F P
0 g
s
.52359 .35474 m .54652 .36469 L .55265 .36505 L p .438 0 0 r
F P
0 g
s
.54652 .36469 m .52359 .35474 L .51245 .35812 L p .651 0 0 r
F P
0 g
s
.66047 .35593 m .63831 .37644 L .64677 .39217 L p .757 0 0 r
F P
0 g
s
.64677 .39217 m .6724 .36842 L .66047 .35593 L p .757 0 0 r
F P
0 g
s
.26979 .24701 m .27465 .25492 L .28936 .25671 L p 0 0 0 r
F P
0 g
s
.27465 .25492 m .26979 .24701 L .25511 .24309 L p 0 0 .105 r
F P
0 g
s
.57019 .26815 m .55151 .28637 L .55321 .30399 L p 0 0 .612 r
F P
0 g
s
.55321 .30399 m .57444 .28553 L .57019 .26815 L p 0 0 .612 r
F P
0 g
s
.57444 .28553 m .55321 .30399 L .54871 .31961 L p 0 0 .732 r
F P
0 g
s
.56556 .45173 m .52347 .44393 L .52871 .44735 L p .103 0 0 r
F P
0 g
s
.52871 .44735 m .56576 .45371 L .56556 .45173 L p .103 0 0 r
F P
0 g
s
.45848 .36936 m .48487 .39944 L .49117 .38262 L p .529 0 0 r
F P
0 g
s
.32694 .31466 m .29003 .32644 L .30116 .31252 L p 0 0 .721 r
F P
0 g
s
.30116 .31252 m .32969 .29781 L .32694 .31466 L p 0 0 .721 r
F P
0 g
s
.32969 .29781 m .30116 .31252 L .30719 .29583 L p 0 0 .826 r
F P
0 g
s
.81532 .32416 m .7912 .36636 L .80794 .36622 L p 0 .642 0 r
F P
0 g
s
.7912 .36636 m .81532 .32416 L .80053 .32647 L p 0 .566 0 r
F P
0 g
s
.81532 .32416 m .81855 .28423 L .80758 .28959 L p 0 .708 0 r
F P
0 g
s
.80758 .28959 m .80053 .32647 L .81532 .32416 L p 0 .708 0 r
F P
0 g
s
.81855 .28423 m .81294 .25121 L .80442 .26099 L p 0 .771 0 r
F P
0 g
s
.80442 .26099 m .80758 .28959 L .81855 .28423 L p 0 .771 0 r
F P
0 g
s
.64643 .4363 m .60852 .44874 L .60316 .45134 L p .484 0 0 r
F P
0 g
s
.56576 .45371 m .60316 .45134 L .60852 .44874 L p .316 0 0 r
F P
0 g
s
.52092 .43379 m .48316 .41539 L .48632 .42764 L p .375 0 0 r
F P
0 g
s
.48632 .42764 m .52347 .44393 L .52092 .43379 L p .375 0 0 r
F P
0 g
s
.80053 .32647 m .77319 .36555 L .7912 .36636 L p 0 .566 0 r
F P
0 g
s
.77319 .36555 m .80053 .32647 L .78277 .32791 L p 0 .475 0 r
F P
0 g
s
.31872 .33075 m .27564 .33525 L .29003 .32644 L p 0 0 .621 r
F P
0 g
s
.29003 .32644 m .32694 .31466 L .31872 .33075 L p 0 0 .621 r
F P
0 g
s
.3243 .32793 m .34792 .37294 L .36592 .37256 L p 0 .491 0 r
F P
0 g
s
.36592 .37256 m .33709 .33038 L .3243 .32793 L p 0 .491 0 r
F P
0 g
s
.33118 .24516 m .34533 .2138 L .35148 .20501 L p 0 .918 0 r
F P
0 g
s
.34533 .2138 m .33118 .24516 L .33459 .25281 L p 0 0 0 r
F P
0 g
s
.32939 .28844 m .33459 .25281 L .33118 .24516 L p 0 0 0 r
F P
0 g
s
.30719 .29583 m .32646 .28303 L .32969 .29781 L p 0 0 .826 r
F P
0 g
s
.32646 .28303 m .30719 .29583 L .30708 .27916 L p 0 0 .888 r
F P
0 g
s
.33198 .3723 m .35316 .41535 L .37097 .41298 L p 0 .566 0 r
F P
0 g
s
.37097 .41298 m .34792 .37294 L .33198 .3723 L p 0 .566 0 r
F P
0 g
s
.45225 .20978 m .43377 .21767 L .42716 .20105 L p 0 .412 0 r
F P
0 g
s
.42716 .20105 m .44776 .19233 L .45225 .20978 L p 0 .412 0 r
F P
0 g
s
.44776 .19233 m .42716 .20105 L .41561 .18739 L p 0 .554 0 r
F P
0 g
s
.65134 .40939 m .68052 .38441 L .6724 .36842 L p .818 0 0 r
F P
0 g
s
.68052 .38441 m .65134 .40939 L .6512 .42503 L p .855 0 0 r
F P
0 g
s
.33709 .33038 m .36592 .37256 L .38273 .37123 L p 0 .612 0 r
F P
0 g
s
.38273 .37123 m .35401 .33229 L .33709 .33038 L p 0 .612 0 r
F P
0 g
s
.60852 .44874 m .64643 .4363 L .6512 .42503 L p .789 0 0 r
F P
0 g
s
.45056 .22787 m .43433 .23448 L .43377 .21767 L p 0 .255 0 r
F P
0 g
s
.43377 .21767 m .45225 .20978 L .45056 .22787 L p 0 .255 0 r
F P
0 g
s
.61623 .35995 m .59361 .3675 L .60019 .37527 L p .545 0 0 r
F P
0 g
s
.60019 .37527 m .62748 .36504 L .61623 .35995 L p .545 0 0 r
F P
0 g
s
.6724 .36842 m .64677 .39217 L .65134 .40939 L p .818 0 0 r
F P
0 g
s
.77383 .29726 m .76524 .32822 L .78277 .32791 L p 0 .563 0 r
F P
0 g
s
.76524 .32822 m .77383 .29726 L .75712 .29819 L p 0 .462 0 r
F P
0 g
s
.54652 .36469 m .57027 .36913 L .57065 .36796 L p .264 0 0 r
F P
0 g
s
.46996 .32557 m .43665 .31443 L .43581 .29803 L p 0 0 .472 r
F P
0 g
s
.43581 .29803 m .46202 .30944 L .46996 .32557 L p 0 0 .472 r
F P
0 g
s
.43665 .31443 m .46996 .32557 L .48196 .33813 L p 0 0 .289 r
F P
0 g
s
.41781 .25769 m .41102 .26406 L .41978 .25727 L p 0 .886 0 r
F P
0 g
s
.60852 .44874 m .56556 .45173 L .56576 .45371 L p .316 0 0 r
F P
0 g
s
.35913 .2972 m .37201 .33333 L .38783 .3333 L p 0 .672 0 r
F P
0 g
s
.38783 .3333 m .37693 .29993 L .35913 .2972 L p 0 .672 0 r
F P
0 g
s
.75712 .29819 m .7562 .27874 L .74224 .27725 L p 0 .457 0 r
F P
0 g
s
.73799 .26727 m .74224 .27725 L .7562 .27874 L p 0 .56 0 r
F P
0 g
s
.46202 .30944 m .43581 .29803 L .44032 .28497 L p 0 0 .594 r
F P
0 g
s
.44032 .28497 m .45946 .29242 L .46202 .30944 L p 0 0 .594 r
F P
0 g
s
.69548 .21248 m .71215 .21922 L .71229 .23664 L p 0 .756 0 r
F P
0 g
s
.71215 .21922 m .69548 .21248 L .7009 .19546 L p 0 .874 0 r
F P
0 g
s
.7009 .19546 m .71843 .20229 L .71215 .21922 L p 0 .874 0 r
F P
0 g
s
.79374 .20378 m .77531 .18641 L .78508 .19913 L p 0 .185 0 r
F P
0 g
s
.78508 .19913 m .80148 .21393 L .79374 .20378 L p 0 .185 0 r
F P
0 g
s
.35148 .20501 m .37278 .18287 L .38589 .17738 L p 0 .985 0 r
F P
0 g
s
.37278 .18287 m .35148 .20501 L .34533 .2138 L p 0 0 0 r
F P
0 g
s
.53888 .3305 m .49601 .34498 L .48196 .33813 L p 0 0 .978 r
F P
0 g
s
.49601 .34498 m .53888 .3305 L .54871 .31961 L p 0 0 .937 r
F P
0 g
s
.71843 .20229 m .7009 .19546 L .71194 .18251 L p 0 .944 0 r
F P
0 g
s
.71194 .18251 m .7301 .18871 L .71843 .20229 L p 0 .944 0 r
F P
0 g
s
.41561 .18739 m .43782 .17846 L .44776 .19233 L p 0 .554 0 r
F P
0 g
s
.43782 .17846 m .41561 .18739 L .40103 .179 L p 0 .7 0 r
F P
0 g
s
.7562 .27874 m .75712 .29819 L .77383 .29726 L p 0 .578 0 r
F P
0 g
s
.24689 .33251 m .2157 .30655 L .22478 .3196 L p 0 0 0 r
F P
0 g
s
.50276 .25238 m .4984 .26605 L .51204 .27601 L p 0 0 .21 r
F P
0 g
s
.51204 .27601 m .51712 .25127 L .50276 .25238 L p 0 0 .21 r
F P
0 g
s
.44266 .3314 m .48196 .33813 L .49601 .34498 L p 0 0 0 r
F P
0 g
s
.48196 .33813 m .44266 .3314 L .43665 .31443 L p 0 0 .289 r
F P
0 g
s
.44298 .24354 m .42871 .24863 L .43433 .23448 L p 0 .04 0 r
F P
0 g
s
.43433 .23448 m .45056 .22787 L .44298 .24354 L p 0 .04 0 r
F P
0 g
s
.3064 .34341 m .26039 .33743 L .27564 .33525 L p 0 0 .512 r
F P
0 g
s
.27564 .33525 m .31872 .33075 L .3064 .34341 L p 0 0 .512 r
F P
0 g
s
.80053 .32647 m .80758 .28959 L .79173 .29421 L p 0 .636 0 r
F P
0 g
s
.79173 .29421 m .78277 .32791 L .80053 .32647 L p 0 .636 0 r
F P
0 g
s
.69657 .23068 m .71229 .23664 L .71887 .25163 L p 0 .556 0 r
F P
0 g
s
.71229 .23664 m .69657 .23068 L .69548 .21248 L p 0 .756 0 r
F P
0 g
s
.3974 .26553 m .39337 .2776 L .40599 .27669 L p 0 .878 0 r
F P
0 g
s
.40599 .27669 m .41102 .26406 L .3974 .26553 L p 0 .878 0 r
F P
0 g
s
.30708 .27916 m .31773 .27285 L .32646 .28303 L p 0 0 .888 r
F P
0 g
s
.31773 .27285 m .30708 .27916 L .30081 .26532 L p 0 0 0 r
F P
0 g
s
.78277 .32791 m .79173 .29421 L .77383 .29726 L p 0 .563 0 r
F P
0 g
s
.25511 .24309 m .25922 .26034 L .27465 .25492 L p 0 0 .105 r
F P
0 g
s
.25922 .26034 m .25511 .24309 L .23949 .24652 L p 0 0 .355 r
F P
0 g
s
.33459 .25281 m .32939 .28844 L .34227 .29312 L p 0 .356 0 r
F P
0 g
s
.32939 .28844 m .33709 .33038 L .35401 .33229 L p 0 .449 0 r
F P
0 g
s
.35401 .33229 m .34227 .29312 L .32939 .28844 L p 0 .449 0 r
F P
0 g
s
.81294 .25121 m .80257 .22765 L .79674 .24255 L p 0 .793 0 r
F P
0 g
s
.79674 .24255 m .80442 .26099 L .81294 .25121 L p 0 .793 0 r
F P
0 g
s
.37695 .27494 m .37693 .29993 L .39245 .30082 L p 0 .733 0 r
F P
0 g
s
.39245 .30082 m .39337 .2776 L .37695 .27494 L p 0 .733 0 r
F P
0 g
s
.51245 .35812 m .53904 .3715 L .54652 .36469 L p .651 0 0 r
F P
0 g
s
.53904 .3715 m .51245 .35812 L .50096 .36798 L p .645 0 0 r
F P
0 g
s
.7301 .18871 m .71194 .18251 L .72678 .17588 L p 0 .938 0 r
F P
0 g
s
.72678 .17588 m .7452 .1808 L .7301 .18871 L p 0 .938 0 r
F P
0 g
s
.45946 .29242 m .44032 .28497 L .4495 .27752 L p 0 0 .665 r
F P
0 g
s
.4495 .27752 m .46275 .27738 L .45946 .29242 L p 0 0 .665 r
F P
0 g
s
.80758 .28959 m .80442 .26099 L .79035 .26983 L p 0 .71 0 r
F P
0 g
s
.79035 .26983 m .79173 .29421 L .80758 .28959 L p 0 .71 0 r
F P
0 g
s
.52157 .41861 m .48487 .39944 L .48316 .41539 L p .517 0 0 r
F P
0 g
s
.48316 .41539 m .52092 .43379 L .52157 .41861 L p .517 0 0 r
F P
0 g
s
.7308 .26162 m .71887 .25163 L p .73113 .26166 L 0 .127 0 r
F P
0 g
s
.73799 .26727 m .7308 .26162 L p .73113 .26166 L 0 .127 0 r
F P
0 g
s
.7308 .26162 m .73799 .26727 L .75273 .26951 L p 0 .632 0 r
F P
0 g
s
.74611 .26482 m .7308 .26162 L p .74543 .26688 L 0 .632 0 r
F P
0 g
s
.75273 .26951 m .74611 .26482 L p .74543 .26688 L 0 .632 0 r
F P
0 g
s
.74611 .26482 m .76216 .26061 L p .74697 .26216 L 0 .861 0 r
F P
0 g
s
.7308 .26162 m .74611 .26482 L p .73036 .26156 L 0 0 0 r
F P
0 g
s
.73188 .2637 m .71641 .25867 L p .73953 .26346 L 0 0 0 r
F P
0 g
s
.74611 .26482 m .73188 .2637 L p .73953 .26346 L 0 0 0 r
F P
0 g
s
.71641 .25867 m .73188 .2637 L p .71549 .26078 L 0 0 0 r
F P
0 g
s
.69102 .25642 m .70987 .25809 L p .71549 .26078 L 0 0 0 r
F P
0 g
s
.70987 .25809 m .71641 .25867 L p .71549 .26078 L 0 0 0 r
F P
0 g
s
.73188 .2637 m .70538 .26609 L .69102 .25642 L p 0 0 0 r
F P
0 g
s
.70538 .26609 m .73188 .2637 L p .72365 .26395 L 0 .729 0 r
F P
0 g
s
.72132 .26879 m .70538 .26609 L p .72243 .2641 L 0 .729 0 r
F P
0 g
s
.73188 .2637 m .74772 .26114 L p .72365 .26395 L 0 .729 0 r
F P
0 g
s
.73188 .2637 m .74611 .26482 L p .74697 .26216 L 0 .861 0 r
F P
0 g
s
.74772 .26114 m .73188 .2637 L p .75732 .26111 L 0 .861 0 r
F P
0 g
s
.76216 .26061 m .74772 .26114 L p .75732 .26111 L 0 .861 0 r
F P
0 g
s
.74772 .26114 m .7612 .25136 L p .73873 .26118 L 0 .964 0 r
F P
0 g
s
.74772 .26114 m .72812 .26682 L p .72243 .2641 L 0 .729 0 r
F P
0 g
s
.72812 .26682 m .72132 .26879 L p .72243 .2641 L 0 .729 0 r
F P
0 g
s
.72132 .26879 m .74195 .26282 L p .73873 .26118 L 0 .964 0 r
F P
0 g
s
.74195 .26282 m .74325 .26116 L .73873 .26118 L 0 .964 0 r
F
.7612 .25136 m .73612 .26399 L .72132 .26879 L p F P
0 g
s
.74722 .25241 m .73612 .26399 L p .75023 .2523 L 0 .924 0 r
F P
0 g
s
.76992 .23597 m .75401 .24749 L p .75023 .2523 L 0 .924 0 r
F P
0 g
s
.75401 .24749 m .74722 .25241 L p .75023 .2523 L 0 .924 0 r
F P
0 g
s
.74195 .26282 m .74772 .26114 L p .74325 .26116 L 0 .964 0 r
F P
0 g
s
.71641 .25867 m .7308 .26162 L p .73036 .26156 L 0 0 0 r
F P
0 g
s
.7562 .27874 m .75273 .26951 L .73799 .26727 L p 0 .56 0 r
F P
0 g
s
.34227 .29312 m .35401 .33229 L .37201 .33333 L p 0 .581 0 r
F P
0 g
s
.37201 .33333 m .35913 .2972 L .34227 .29312 L p 0 .581 0 r
F P
0 g
s
.40103 .179 m .42407 .17056 L .43782 .17846 L p 0 .7 0 r
F P
0 g
s
.42407 .17056 m .40103 .179 L .38589 .17738 L p 0 .802 0 r
F P
0 g
s
.71887 .25163 m .70395 .24699 L .69657 .23068 L p 0 .556 0 r
F P
0 g
s
.57027 .36913 m .54652 .36469 L .53904 .3715 L p .634 0 0 r
F P
0 g
s
.56581 .44254 m .52092 .43379 L .52347 .44393 L p .524 0 0 r
F P
0 g
s
.52347 .44393 m .56556 .45173 L .56581 .44254 L p .524 0 0 r
F P
0 g
s
.77325 .27613 m .77383 .29726 L .79173 .29421 L p 0 .65 0 r
F P
0 g
s
.77383 .29726 m .77325 .27613 L .7562 .27874 L p 0 .578 0 r
F P
0 g
s
.6119 .43914 m .6512 .42503 L .65134 .40939 L p .817 0 0 r
F P
0 g
s
.6512 .42503 m .6119 .43914 L .60852 .44874 L p .789 0 0 r
F P
0 g
s
.54871 .31961 m .50976 .34493 L .49601 .34498 L p 0 0 .937 r
F P
0 g
s
.45281 .3461 m .49601 .34498 L .50976 .34493 L p 0 0 0 r
F P
0 g
s
.49601 .34498 m .45281 .3461 L .44266 .3314 L p 0 0 0 r
F P
0 g
s
.62748 .36504 m .60019 .37527 L .60624 .38903 L p .712 0 0 r
F P
0 g
s
.60624 .38903 m .63831 .37644 L .62748 .36504 L p .712 0 0 r
F P
0 g
s
.33459 .25281 m .34581 .22606 L .34533 .2138 L p 0 0 0 r
F P
0 g
s
.34581 .22606 m .33459 .25281 L .34466 .26132 L p 0 .327 0 r
F P
0 g
s
.34227 .29312 m .34466 .26132 L .33459 .25281 L p 0 .356 0 r
F P
0 g
s
.43075 .25416 m .41781 .25769 L .42871 .24863 L p 0 0 0 r
F P
0 g
s
.42871 .24863 m .44298 .24354 L .43075 .25416 L p 0 0 0 r
F P
0 g
s
.50976 .34493 m .54871 .31961 L .55321 .30399 L p 0 0 .833 r
F P
0 g
s
.79173 .29421 m .79035 .26983 L .77325 .27613 L p 0 .65 0 r
F P
0 g
s
.56556 .45173 m .60852 .44874 L .6119 .43914 L p .672 0 0 r
F P
0 g
s
.75273 .26951 m .7562 .27874 L .77325 .27613 L p 0 .67 0 r
F P
0 g
s
.59361 .3675 m .57027 .36913 L .56951 .37759 L p .605 0 0 r
F P
0 g
s
.56951 .37759 m .60019 .37527 L .59361 .3675 L p .605 0 0 r
F P
0 g
s
.80148 .21393 m .78508 .19913 L .78872 .21592 L p 0 .657 0 r
F P
0 g
s
.78872 .21592 m .80257 .22765 L .80148 .21393 L p 0 .657 0 r
F P
0 g
s
.23748 .32126 m .21263 .28973 L .2157 .30655 L p 0 0 .773 r
F P
0 g
s
.40343 .26003 m .3974 .26553 L .41102 .26406 L p 0 .944 0 r
F P
0 g
s
.41102 .26406 m .41781 .25769 L .40343 .26003 L p 0 .944 0 r
F P
0 g
s
.51712 .25127 m .51204 .27601 L .52251 .291 L p 0 0 .419 r
F P
0 g
s
.52251 .291 m .53169 .25745 L .51712 .25127 L p 0 0 .419 r
F P
0 g
s
.29202 .35051 m .24689 .33251 L .26039 .33743 L p 0 0 .364 r
F P
0 g
s
.26039 .33743 m .3064 .34341 L .29202 .35051 L p 0 0 .364 r
F P
0 g
s
.34533 .2138 m .36401 .19461 L .37278 .18287 L p 0 0 0 r
F P
0 g
s
.36401 .19461 m .34533 .2138 L .34581 .22606 L p 0 0 0 r
F P
0 g
s
.23949 .24652 m .24581 .2721 L .25922 .26034 L p 0 0 .355 r
F P
0 g
s
.24581 .2721 m .23949 .24652 L .22568 .25677 L p 0 0 .515 r
F P
0 g
s
.7452 .1808 m .72678 .17588 L .74288 .17681 L p 0 .702 0 r
F P
0 g
s
.74288 .17681 m .76119 .17996 L .7452 .1808 L p 0 .702 0 r
F P
0 g
s
.34466 .26132 m .34227 .29312 L .35913 .2972 L p 0 .532 0 r
F P
0 g
s
.30081 .26532 m .30491 .2691 L .31773 .27285 L p 0 0 0 r
F P
0 g
s
.30491 .2691 m .30081 .26532 L .28936 .25671 L p 0 0 .007 r
F P
0 g
s
.50096 .36798 m .53153 .38438 L .53904 .3715 L p .645 0 0 r
F P
0 g
s
.53153 .38438 m .50096 .36798 L .49117 .38262 L p .616 0 0 r
F P
0 g
s
.35968 .26917 m .35913 .2972 L .37693 .29993 L p 0 .64 0 r
F P
0 g
s
.37693 .29993 m .37695 .27494 L .35968 .26917 L p 0 .64 0 r
F P
0 g
s
.80442 .26099 m .79674 .24255 L .78498 .25598 L p 0 .771 0 r
F P
0 g
s
.78498 .25598 m .79035 .26983 L .80442 .26099 L p 0 .771 0 r
F P
0 g
s
.53904 .3715 m .56951 .37759 L .57027 .36913 L p .634 0 0 r
F P
0 g
s
.46275 .27738 m .4495 .27752 L .46183 .27708 L p 0 0 .491 r
F P
0 g
s
.46183 .27708 m .47141 .2669 L .46275 .27738 L p 0 0 .491 r
F P
0 g
s
.52532 .40109 m .49117 .38262 L .48487 .39944 L p .578 0 0 r
F P
0 g
s
.48487 .39944 m .52157 .41861 L .52532 .40109 L p .578 0 0 r
F P
0 g
s
.38128 .26133 m .37695 .27494 L .39337 .2776 L p 0 .756 0 r
F P
0 g
s
.39337 .2776 m .3974 .26553 L .38128 .26133 L p 0 .756 0 r
F P
0 g
s
.6119 .43914 m .56581 .44254 L .56556 .45173 L p .672 0 0 r
F P
0 g
s
.33905 .34592 m .3064 .34341 L .31872 .33075 L p 0 0 .598 r
F P
0 g
s
.31872 .33075 m .3474 .32971 L .33905 .34592 L p 0 0 .598 r
F P
0 g
s
.3474 .32971 m .31872 .33075 L .32694 .31466 L p 0 0 .699 r
F P
0 g
s
.35913 .2972 m .35968 .26917 L .34466 .26132 L p 0 .532 0 r
F P
0 g
s
.49117 .38262 m .52532 .40109 L .53153 .38438 L p .616 0 0 r
F P
0 g
s
.35122 .3126 m .32694 .31466 L .32969 .29781 L p 0 0 .799 r
F P
0 g
s
.32694 .31466 m .35122 .3126 L .3474 .32971 L p 0 0 .699 r
F P
0 g
s
.7308 .26162 m .71641 .25867 L .70395 .24699 L p 0 .175 0 r
F P
0 g
s
.76937 .26553 m .77325 .27613 L .79035 .26983 L p 0 .729 0 r
F P
0 g
s
.77325 .27613 m .76937 .26553 L .75273 .26951 L p 0 .67 0 r
F P
0 g
s
.2339 .30555 m .21615 .27208 L .21263 .28973 L p 0 0 .755 r
F P
0 g
s
.21263 .28973 m .23748 .32126 L .2339 .30555 L p 0 0 .755 r
F P
0 g
s
.38589 .17738 m .40883 .17007 L .42407 .17056 L p 0 .802 0 r
F P
0 g
s
.40883 .17007 m .38589 .17738 L .37278 .18287 L p 0 0 0 r
F P
0 g
s
.61268 .42413 m .65134 .40939 L .64677 .39217 L p .804 0 0 r
F P
0 g
s
.65134 .40939 m .61268 .42413 L .6119 .43914 L p .817 0 0 r
F P
0 g
s
.63831 .37644 m .60624 .38903 L .61067 .40636 L p .773 0 0 r
F P
0 g
s
.61067 .40636 m .64677 .39217 L .63831 .37644 L p .773 0 0 r
F P
0 g
s
.74611 .26482 m .75273 .26951 L .76937 .26553 L p 0 .796 0 r
F P
0 g
s
.55321 .30399 m .52084 .33787 L .50976 .34493 L p 0 0 .833 r
F P
0 g
s
.52084 .33787 m .55321 .30399 L .55151 .28637 L p 0 0 .737 r
F P
0 g
s
.22568 .25677 m .23684 .28813 L .24581 .2721 L p 0 0 .515 r
F P
0 g
s
.23684 .28813 m .22568 .25677 L .21615 .27208 L p 0 0 .643 r
F P
0 g
s
.79035 .26983 m .78498 .25598 L .76937 .26553 L p 0 .729 0 r
F P
0 g
s
.53169 .25745 m .52251 .291 L .5279 .30835 L p 0 0 .547 r
F P
0 g
s
.5279 .30835 m .54388 .26988 L .53169 .25745 L p 0 0 .547 r
F P
0 g
s
.34466 .26132 m .35295 .23966 L .34581 .22606 L p 0 .327 0 r
F P
0 g
s
.35295 .23966 m .34466 .26132 L .35968 .26917 L p 0 .523 0 r
F P
0 g
s
.41592 .25787 m .40343 .26003 L .4166 .25788 L p 0 .726 0 r
F P
0 g
s
.40343 .26003 m .41592 .25787 L .40101 .25394 L p 0 .772 0 r
F P
0 g
s
.42414 .26028 m .41528 .25785 L p .41592 .25787 L 0 0 0 r
F P
0 g
s
.41528 .25785 m .40101 .25394 L .41592 .25787 L p 0 0 0 r
F P
0 g
s
.4166 .25788 m .41781 .25769 L p .41592 .25787 L 0 .726 0 r
F P
0 g
s
.41781 .25769 m .43075 .25416 L .41592 .25787 L p 0 .726 0 r
F P
0 g
s
.41589 .33159 m .44266 .3314 L .45281 .3461 L p 0 0 .54 r
F P
0 g
s
.44266 .3314 m .41589 .33159 L .41299 .31438 L p 0 0 .652 r
F P
0 g
s
.41299 .31438 m .43665 .31443 L .44266 .3314 L p 0 0 .652 r
F P
0 g
s
.50976 .34493 m .46544 .35609 L .45281 .3461 L p 0 0 0 r
F P
0 g
s
.46544 .35609 m .50976 .34493 L .52084 .33787 L p 0 0 .971 r
F P
0 g
s
.21615 .27208 m .2339 .30555 L .23684 .28813 L p 0 0 .643 r
F P
0 g
s
.32756 .35854 m .29202 .35051 L .3064 .34341 L p 0 0 .475 r
F P
0 g
s
.3064 .34341 m .33905 .34592 L .32756 .35854 L p 0 0 .475 r
F P
0 g
s
.80257 .22765 m .78872 .21592 L .78556 .23382 L p 0 .803 0 r
F P
0 g
s
.78556 .23382 m .79674 .24255 L .80257 .22765 L p 0 .803 0 r
F P
0 g
s
.64677 .39217 m .61067 .40636 L .61268 .42413 L p .804 0 0 r
F P
0 g
s
.46554 .24009 m .44298 .24354 L .45056 .22787 L p 0 .211 0 r
F P
0 g
s
.45056 .22787 m .47234 .22302 L .46554 .24009 L p 0 .211 0 r
F P
0 g
s
.47234 .22302 m .45056 .22787 L .45225 .20978 L p 0 .311 0 r
F P
0 g
s
.45225 .20978 m .47313 .20498 L .47234 .22302 L p 0 .311 0 r
F P
0 g
s
.47313 .20498 m .45225 .20978 L .44776 .19233 L p 0 .406 0 r
F P
0 g
s
.76119 .17996 m .74288 .17681 L .75744 .18522 L p 0 0 0 r
F P
0 g
s
.75744 .18522 m .77531 .18641 L .76119 .17996 L p 0 0 0 r
F P
0 g
s
.56951 .37759 m .53904 .3715 L .53153 .38438 L p .673 0 0 r
F P
0 g
s
.56647 .42767 m .52157 .41861 L .52092 .43379 L p .619 0 0 r
F P
0 g
s
.52092 .43379 m .56581 .44254 L .56647 .42767 L p .619 0 0 r
F P
0 g
s
.42262 .34858 m .45281 .3461 L .46544 .35609 L p 0 0 .375 r
F P
0 g
s
.45281 .3461 m .42262 .34858 L .41589 .33159 L p 0 0 .54 r
F P
0 g
s
.36552 .25217 m .35968 .26917 L .37695 .27494 L p 0 .647 0 r
F P
0 g
s
.37695 .27494 m .38128 .26133 L .36552 .25217 L p 0 .647 0 r
F P
0 g
s
.52729 .32495 m .55151 .28637 L .54388 .26988 L p 0 0 .646 r
F P
0 g
s
.55151 .28637 m .52729 .32495 L .52084 .33787 L p 0 0 .737 r
F P
0 g
s
.38805 .25515 m .38128 .26133 L .3974 .26553 L p 0 .796 0 r
F P
0 g
s
.3974 .26553 m .40343 .26003 L .38805 .25515 L p 0 .796 0 r
F P
0 g
s
.54388 .26988 m .5279 .30835 L .52729 .32495 L p 0 0 .646 r
F P
0 g
s
.43665 .31443 m .41299 .31438 L .41442 .29982 L p 0 0 .752 r
F P
0 g
s
.41442 .29982 m .43581 .29803 L .43665 .31443 L p 0 0 .752 r
F P
0 g
s
.32969 .29781 m .34985 .29744 L .35122 .3126 L p 0 0 .799 r
F P
0 g
s
.34985 .29744 m .32969 .29781 L .32646 .28303 L p 0 0 .861 r
F P
0 g
s
.34581 .22606 m .36118 .21061 L .36401 .19461 L p 0 0 0 r
F P
0 g
s
.36118 .21061 m .34581 .22606 L .35295 .23966 L p 0 .326 0 r
F P
0 g
s
.24689 .33251 m .29202 .35051 L .27799 .35077 L p 0 0 .106 r
F P
0 g
s
.27799 .35077 m .23748 .32126 L .24689 .33251 L p 0 0 .106 r
F P
0 g
s
.67595 .22372 m .69657 .23068 L .70395 .24699 L p 0 .613 0 r
F P
0 g
s
.69657 .23068 m .67595 .22372 L .67778 .20618 L p 0 .796 0 r
F P
0 g
s
.67778 .20618 m .69548 .21248 L .69657 .23068 L p 0 .796 0 r
F P
0 g
s
.56581 .44254 m .6119 .43914 L .61268 .42413 L p .728 0 0 r
F P
0 g
s
.35968 .26917 m .36552 .25217 L .35295 .23966 L p 0 .523 0 r
F P
0 g
s
.60019 .37527 m .56951 .37759 L .5685 .39196 L p .704 0 0 r
F P
0 g
s
.5685 .39196 m .60624 .38903 L .60019 .37527 L p .704 0 0 r
F P
0 g
s
.28936 .25671 m .29019 .27256 L .30491 .2691 L p 0 0 .007 r
F P
0 g
s
.29019 .27256 m .28936 .25671 L .27465 .25492 L p 0 0 .312 r
F P
0 g
s
.45386 .25332 m .43075 .25416 L .44298 .24354 L p 0 .083 0 r
F P
0 g
s
.44298 .24354 m .46554 .24009 L .45386 .25332 L p 0 .083 0 r
F P
0 g
s
.76216 .26061 m .76937 .26553 L .78498 .25598 L p 0 .842 0 r
F P
0 g
s
.76937 .26553 m .76216 .26061 L .74611 .26482 L p 0 .796 0 r
F P
0 g
s
.44776 .19233 m .46772 .189 L .47313 .20498 L p 0 .406 0 r
F P
0 g
s
.46772 .189 m .44776 .19233 L .43782 .17846 L p 0 .497 0 r
F P
0 g
s
.47141 .2669 m .46183 .27708 L .47518 .28383 L p 0 0 .033 r
F P
0 g
s
.47518 .28383 m .484 .26285 L .47141 .2669 L p 0 0 .033 r
F P
0 g
s
.69548 .21248 m .67778 .20618 L .68585 .1918 L p 0 .936 0 r
F P
0 g
s
.68585 .1918 m .7009 .19546 L .69548 .21248 L p 0 .936 0 r
F P
0 g
s
.68064 .24145 m .70395 .24699 L .71641 .25867 L p 0 .341 0 r
F P
0 g
s
.70395 .24699 m .68064 .24145 L .67595 .22372 L p 0 .613 0 r
F P
0 g
s
.79674 .24255 m .78556 .23382 L .77614 .24965 L p 0 .843 0 r
F P
0 g
s
.77614 .24965 m .78498 .25598 L .79674 .24255 L p 0 .843 0 r
F P
0 g
s
.37278 .18287 m .39474 .17718 L .40883 .17007 L p 0 0 0 r
F P
0 g
s
.39474 .17718 m .37278 .18287 L .36401 .19461 L p 0 0 0 r
F P
0 g
s
.78498 .25598 m .77614 .24965 L .76216 .26061 L p 0 .842 0 r
F P
0 g
s
.53153 .38438 m .5685 .39196 L .56951 .37759 L p .673 0 0 r
F P
0 g
s
.31483 .36544 m .27799 .35077 L .29202 .35051 L p 0 0 .281 r
F P
0 g
s
.29202 .35051 m .32756 .35854 L .31483 .36544 L p 0 0 .281 r
F P
0 g
s
.52084 .33787 m .47842 .35964 L .46544 .35609 L p 0 0 .971 r
F P
0 g
s
.43204 .36254 m .46544 .35609 L .47842 .35964 L p 0 0 .036 r
F P
0 g
s
.46544 .35609 m .43204 .36254 L .42262 .34858 L p 0 0 .375 r
F P
0 g
s
.61268 .42413 m .56647 .42767 L .56581 .44254 L p .728 0 0 r
F P
0 g
s
.40101 .25394 m .38805 .25515 L .40343 .26003 L p 0 .772 0 r
F P
0 g
s
.37437 .24384 m .36552 .25217 L .38128 .26133 L p 0 .656 0 r
F P
0 g
s
.38128 .26133 m .38805 .25515 L .37437 .24384 L p 0 .656 0 r
F P
0 g
s
.35295 .23966 m .36484 .22806 L .36118 .21061 L p 0 .326 0 r
F P
0 g
s
.36484 .22806 m .35295 .23966 L .36552 .25217 L p 0 .513 0 r
F P
0 g
s
.77531 .18641 m .75744 .18522 L .76787 .19968 L p 0 .396 0 r
F P
0 g
s
.76787 .19968 m .78508 .19913 L .77531 .18641 L p 0 .396 0 r
F P
0 g
s
.43581 .29803 m .41442 .29982 L .42 .29041 L p 0 0 .847 r
F P
0 g
s
.42 .29041 m .44032 .28497 L .43581 .29803 L p 0 0 .847 r
F P
0 g
s
.5685 .39196 m .53153 .38438 L .52532 .40109 L p .669 0 0 r
F P
0 g
s
.56742 .40973 m .52532 .40109 L .52157 .41861 L p .653 0 0 r
F P
0 g
s
.52157 .41861 m .56647 .42767 L .56742 .40973 L p .653 0 0 r
F P
0 g
s
.7009 .19546 m .68585 .1918 L .69887 .18308 L p 0 .998 0 r
F P
0 g
s
.69887 .18308 m .71194 .18251 L .7009 .19546 L p 0 .998 0 r
F P
0 g
s
.47842 .35964 m .52084 .33787 L .52729 .32495 L p 0 0 .899 r
F P
0 g
s
.71641 .25867 m .69102 .25642 L .68064 .24145 L p 0 .341 0 r
F P
0 g
s
.43924 .26047 m .41592 .25787 L .43075 .25416 L p 0 0 0 r
F P
0 g
s
.43075 .25416 m .45386 .25332 L .43924 .26047 L p 0 0 0 r
F P
0 g
s
.43782 .17846 m .45698 .17781 L .46772 .189 L p 0 .497 0 r
F P
0 g
s
.45698 .17781 m .43782 .17846 L .42407 .17056 L p 0 .492 0 r
F P
0 g
s
.32646 .28303 m .34346 .28681 L .34985 .29744 L p 0 0 .861 r
F P
0 g
s
.32477 .28106 m .31773 .27285 L p .34346 .28681 L 0 0 0 r
F P
0 g
s
.31773 .27285 m .33306 .28261 L .34346 .28681 L p 0 0 0 r
F P
0 g
s
.34346 .28681 m .33306 .28261 L p .35419 .2901 L 0 0 0 r
F P
0 g
s
.36316 .29328 m .34346 .28681 L p .35419 .2901 L 0 0 0 r
F P
0 g
s
.34346 .28681 m .32646 .28303 L .32477 .28106 L p 0 0 0 r
F P
0 g
s
.36552 .25217 m .37437 .24384 L .36484 .22806 L p 0 .513 0 r
F P
0 g
s
.56647 .42767 m .61268 .42413 L .61067 .40636 L p .737 0 0 r
F P
0 g
s
.60624 .38903 m .5685 .39196 L .56742 .40973 L p .731 0 0 r
F P
0 g
s
.56742 .40973 m .61067 .40636 L .60624 .38903 L p .731 0 0 r
F P
0 g
s
.36317 .34991 m .33905 .34592 L .3474 .32971 L p 0 0 .661 r
F P
0 g
s
.3474 .32971 m .36829 .3323 L .36317 .34991 L p 0 0 .661 r
F P
0 g
s
.36829 .3323 m .3474 .32971 L .35122 .3126 L p 0 0 .757 r
F P
0 g
s
.3556 .36508 m .32756 .35854 L .33905 .34592 L p 0 0 .548 r
F P
0 g
s
.33905 .34592 m .36317 .34991 L .3556 .36508 L p 0 0 .548 r
F P
0 g
s
.52532 .40109 m .56742 .40973 L .5685 .39196 L p .669 0 0 r
F P
0 g
s
.23748 .32126 m .27799 .35077 L .26671 .34406 L p 0 0 0 r
F P
0 g
s
.26671 .34406 m .2339 .30555 L .23748 .32126 L p 0 0 0 r
F P
0 g
s
.74772 .26114 m .76216 .26061 L .77614 .24965 L p 0 .942 0 r
F P
0 g
s
.27465 .25492 m .27617 .28269 L .29019 .27256 L p 0 0 .312 r
F P
0 g
s
.27617 .28269 m .27465 .25492 L .25922 .26034 L p 0 0 .483 r
F P
0 g
s
.61067 .40636 m .56742 .40973 L .56647 .42767 L p .737 0 0 r
F P
0 g
s
.39979 .35132 m .42262 .34858 L .43204 .36254 L p 0 0 .538 r
F P
0 g
s
.42262 .34858 m .39979 .35132 L .39647 .33328 L p 0 0 .657 r
F P
0 g
s
.39647 .33328 m .41589 .33159 L .42262 .34858 L p 0 0 .657 r
F P
0 g
s
.41589 .33159 m .39647 .33328 L .39521 .3161 L p 0 0 .752 r
F P
0 g
s
.39521 .3161 m .41299 .31438 L .41589 .33159 L p 0 0 .752 r
F P
0 g
s
.484 .26285 m .47518 .28383 L .48717 .29665 L p 0 0 .348 r
F P
0 g
s
.48717 .29665 m .4984 .26605 L .484 .26285 L p 0 0 .348 r
F P
0 g
s
.36401 .19461 m .38431 .1907 L .39474 .17718 L p 0 0 0 r
F P
0 g
s
.38431 .1907 m .36401 .19461 L .36118 .21061 L p 0 .16 0 r
F P
0 g
s
.78508 .19913 m .76787 .19968 L .77229 .21767 L p 0 .661 0 r
F P
0 g
s
.77229 .21767 m .78872 .21592 L .78508 .19913 L p 0 .661 0 r
F P
0 g
s
.37012 .31518 m .35122 .3126 L .34985 .29744 L p 0 0 .857 r
F P
0 g
s
.35122 .3126 m .37012 .31518 L .36829 .3323 L p 0 0 .757 r
F P
0 g
s
.38861 .24296 m .37437 .24384 L .38805 .25515 L p 0 .632 0 r
F P
0 g
s
.38805 .25515 m .40101 .25394 L .38861 .24296 L p 0 .632 0 r
F P
0 g
s
.71464 .18163 m .72678 .17588 L .71194 .18251 L p 0 0 0 r
F P
0 g
s
.71194 .18251 m .69887 .18308 L .71464 .18163 L p 0 0 0 r
F P
0 g
s
.52729 .32495 m .48957 .35606 L .47842 .35964 L p 0 0 .899 r
F P
0 g
s
.47842 .35964 m .44263 .37112 L .43204 .36254 L p 0 0 .036 r
F P
0 g
s
.44263 .37112 m .47842 .35964 L .48957 .35606 L p 0 0 .674 r
F P
0 g
s
.34682 .37529 m .31483 .36544 L .32756 .35854 L p 0 0 .371 r
F P
0 g
s
.32756 .35854 m .3556 .36508 L .34682 .37529 L p 0 0 .371 r
F P
0 g
s
.27799 .35077 m .31483 .36544 L .30298 .36537 L p 0 0 0 r
F P
0 g
s
.30298 .36537 m .26671 .34406 L .27799 .35077 L p 0 0 0 r
F P
0 g
s
.7612 .25136 m .77614 .24965 L .78556 .23382 L p 0 .901 0 r
F P
0 g
s
.77614 .24965 m .7612 .25136 L .74772 .26114 L p 0 .942 0 r
F P
0 g
s
.41592 .25787 m .43924 .26047 L .42414 .26028 L p 0 0 0 r
F P
0 g
s
.48957 .35606 m .52729 .32495 L .5279 .30835 L p 0 0 .796 r
F P
0 g
s
.44032 .28497 m .42 .29041 L .42884 .28785 L p 0 0 .76 r
F P
0 g
s
.42884 .28785 m .4495 .27752 L .44032 .28497 L p 0 0 .76 r
F P
0 g
s
.42407 .17056 m .44269 .17342 L .45698 .17781 L p 0 .492 0 r
F P
0 g
s
.44269 .17342 m .42407 .17056 L .40883 .17007 L p 0 0 0 r
F P
0 g
s
.40461 .36725 m .43204 .36254 L .44263 .37112 L p 0 0 .33 r
F P
0 g
s
.43204 .36254 m .40461 .36725 L .39979 .35132 L p 0 0 .538 r
F P
0 g
s
.41299 .31438 m .39521 .3161 L .39622 .30263 L p 0 0 .849 r
F P
0 g
s
.39622 .30263 m .41442 .29982 L .41299 .31438 L p 0 0 .849 r
F P
0 g
s
.78872 .21592 m .77229 .21767 L .76992 .23597 L p 0 .809 0 r
F P
0 g
s
.76992 .23597 m .78556 .23382 L .78872 .21592 L p 0 .809 0 r
F P
0 g
s
.36118 .21061 m .37942 .20828 L .38431 .1907 L p 0 .16 0 r
F P
0 g
s
.37942 .20828 m .36118 .21061 L .36484 .22806 L p 0 .341 0 r
F P
0 g
s
.26021 .33143 m .23684 .28813 L .2339 .30555 L p 0 0 .746 r
F P
0 g
s
.2339 .30555 m .26671 .34406 L .26021 .33143 L p 0 0 .746 r
F P
0 g
s
.78556 .23382 m .76992 .23597 L .7612 .25136 L p 0 .901 0 r
F P
0 g
s
.38193 .35178 m .36317 .34991 L .36829 .3323 L p 0 0 .699 r
F P
0 g
s
.36829 .3323 m .38299 .33347 L .38193 .35178 L p 0 0 .699 r
F P
0 g
s
.38299 .33347 m .36829 .3323 L .37012 .31518 L p 0 0 .791 r
F P
0 g
s
.38026 .36822 m .3556 .36508 L .36317 .34991 L p 0 0 .59 r
F P
0 g
s
.36317 .34991 m .38193 .35178 L .38026 .36822 L p 0 0 .59 r
F P
0 g
s
.25922 .26034 m .26536 .29775 L .27617 .28269 L p 0 0 .483 r
F P
0 g
s
.26536 .29775 m .25922 .26034 L .24581 .2721 L p 0 0 .606 r
F P
0 g
s
.38095 .2268 m .36484 .22806 L .37437 .24384 L p 0 .488 0 r
F P
0 g
s
.37437 .24384 m .38861 .24296 L .38095 .2268 L p 0 .488 0 r
F P
0 g
s
.48949 .24559 m .45386 .25332 L .46554 .24009 L p 0 .318 0 r
F P
0 g
s
.46554 .24009 m .49596 .22824 L .48949 .24559 L p 0 .318 0 r
F P
0 g
s
.49596 .22824 m .46554 .24009 L .47234 .22302 L p 0 .371 0 r
F P
0 g
s
.38193 .35178 m .39979 .35132 L .40461 .36725 L p 0 0 .591 r
F P
0 g
s
.39979 .35132 m .38193 .35178 L .38299 .33347 L p 0 0 .701 r
F P
0 g
s
.38299 .33347 m .39647 .33328 L .39979 .35132 L p 0 0 .701 r
F P
0 g
s
.39647 .33328 m .38299 .33347 L .38327 .31628 L p 0 0 .791 r
F P
0 g
s
.38327 .31628 m .39521 .3161 L .39647 .33328 L p 0 0 .791 r
F P
0 g
s
.33306 .28261 m .31773 .27285 L .30491 .2691 L p 0 0 .102 r
F P
0 g
s
.47234 .22302 m .49646 .21076 L .49596 .22824 L p 0 .371 0 r
F P
0 g
s
.49646 .21076 m .47234 .22302 L .47313 .20498 L p 0 .424 0 r
F P
0 g
s
.36484 .22806 m .38095 .2268 L .37942 .20828 L p 0 .341 0 r
F P
0 g
s
.4984 .26605 m .48717 .29665 L .49564 .31327 L p 0 0 .508 r
F P
0 g
s
.49564 .31327 m .51204 .27601 L .4984 .26605 L p 0 0 .508 r
F P
0 g
s
.37012 .31518 m .38327 .31628 L .38299 .33347 L p 0 0 .791 r
F P
0 g
s
.40461 .36725 m .38026 .36822 L .38193 .35178 L p 0 0 .591 r
F P
0 g
s
.34985 .29744 m .36833 .30138 L .37012 .31518 L p 0 0 .857 r
F P
0 g
s
.36833 .30138 m .34985 .29744 L .34346 .28681 L p 0 0 .919 r
F P
0 g
s
.72678 .17588 m .71464 .18163 L .73042 .18777 L p 0 0 0 r
F P
0 g
s
.73042 .18777 m .74288 .17681 L .72678 .17588 L p 0 0 0 r
F P
0 g
s
.25971 .31503 m .24581 .2721 L .23684 .28813 L p 0 0 .711 r
F P
0 g
s
.23684 .28813 m .26021 .33143 L .25971 .31503 L p 0 0 .711 r
F P
0 g
s
.24581 .2721 m .25971 .31503 L .26536 .29775 L p 0 0 .606 r
F P
0 g
s
.65435 .22443 m .68064 .24145 L .69102 .25642 L p 0 .481 0 r
F P
0 g
s
.68064 .24145 m .65435 .22443 L .65861 .20856 L p 0 .721 0 r
F P
0 g
s
.65861 .20856 m .67595 .22372 L .68064 .24145 L p 0 .721 0 r
F P
0 g
s
.47815 .25993 m .43924 .26047 L .45386 .25332 L p 0 .26 0 r
F P
0 g
s
.45386 .25332 m .48949 .24559 L .47815 .25993 L p 0 .26 0 r
F P
0 g
s
.4969 .34586 m .5279 .30835 L .52251 .291 L p 0 0 .704 r
F P
0 g
s
.5279 .30835 m .4969 .34586 L .48957 .35606 L p 0 0 .796 r
F P
0 g
s
.38327 .31628 m .37012 .31518 L .36833 .30138 L p 0 0 .888 r
F P
0 g
s
.47313 .20498 m .49087 .19611 L .49646 .21076 L p 0 .424 0 r
F P
0 g
s
.49087 .19611 m .47313 .20498 L .46772 .189 L p 0 .441 0 r
F P
0 g
s
.37827 .38004 m .34682 .37529 L .3556 .36508 L p 0 0 .411 r
F P
0 g
s
.3556 .36508 m .38026 .36822 L .37827 .38004 L p 0 0 .411 r
F P
0 g
s
.70538 .26609 m .66461 .25895 L .65649 .24222 L p 0 .1 0 r
F P
0 g
s
.69102 .25642 m .70538 .26609 L p .67667 .25207 L 0 .1 0 r
F P
0 g
s
.66461 .25895 m .70538 .26609 L .72132 .26879 L p 0 .479 0 r
F P
0 g
s
.67734 .27184 m .66461 .25895 L p .69802 .26475 L 0 .479 0 r
F P
0 g
s
.72132 .26879 m .70655 .26981 L p .69802 .26475 L 0 .479 0 r
F P
0 g
s
.70655 .26981 m .67734 .27184 L p .69802 .26475 L 0 .479 0 r
F P
0 g
s
.65649 .24222 m .66816 .24702 L p .67667 .25207 L 0 .1 0 r
F P
0 g
s
.66816 .24702 m .69102 .25642 L p .67667 .25207 L 0 .1 0 r
F P
0 g
s
.69102 .25642 m .65649 .24222 L .65435 .22443 L p 0 .481 0 r
F P
0 g
s
.41114 .25267 m .38861 .24296 L .40101 .25394 L p 0 .547 0 r
F P
0 g
s
.40101 .25394 m .42414 .26028 L .41114 .25267 L p 0 .547 0 r
F P
0 g
s
.33829 .37874 m .30298 .36537 L .31483 .36544 L p 0 0 0 r
F P
0 g
s
.31483 .36544 m .34682 .37529 L .33829 .37874 L p 0 0 0 r
F P
0 g
s
.67595 .22372 m .65861 .20856 L .6686 .19734 L p 0 .883 0 r
F P
0 g
s
.6686 .19734 m .67778 .20618 L .67595 .22372 L p 0 .883 0 r
F P
0 g
s
.44263 .37112 m .41014 .37841 L .40461 .36725 L p 0 0 .33 r
F P
0 g
s
.38026 .36822 m .40461 .36725 L .41014 .37841 L p 0 0 .403 r
F P
0 g
s
.39521 .3161 m .38327 .31628 L .38273 .30308 L p 0 0 .887 r
F P
0 g
s
.38273 .30308 m .39622 .30263 L .39521 .3161 L p 0 0 .887 r
F P
0 g
s
.51204 .27601 m .49564 .31327 L .49908 .33071 L p 0 0 .614 r
F P
0 g
s
.49908 .33071 m .52251 .291 L .51204 .27601 L p 0 0 .614 r
F P
0 g
s
.40883 .17007 m .4273 .17667 L .44269 .17342 L p 0 0 0 r
F P
0 g
s
.4273 .17667 m .40883 .17007 L .39474 .17718 L p 0 .02 0 r
F P
0 g
s
.48957 .35606 m .45259 .37279 L .44263 .37112 L p 0 0 .674 r
F P
0 g
s
.45259 .37279 m .48957 .35606 L .4969 .34586 L p 0 0 .926 r
F P
0 g
s
.41014 .37841 m .44263 .37112 L .45259 .37279 L p 0 0 0 r
F P
0 g
s
.41442 .29982 m .39622 .30263 L .39937 .29521 L p 0 0 .945 r
F P
0 g
s
.39937 .29521 m .42 .29041 L .41442 .29982 L p 0 0 .945 r
F P
0 g
s
.29409 .35822 m .26021 .33143 L .26671 .34406 L p 0 0 .763 r
F P
0 g
s
.26671 .34406 m .30298 .36537 L .29409 .35822 L p 0 0 .763 r
F P
0 g
s
.52251 .291 m .49908 .33071 L .4969 .34586 L p 0 0 .704 r
F P
0 g
s
.36833 .30138 m .38273 .30308 L .38327 .31628 L p 0 0 .888 r
F P
0 g
s
.41014 .37841 m .37827 .38004 L .38026 .36822 L p 0 0 .403 r
F P
0 g
s
.4495 .27752 m .42884 .28785 L .43945 .29273 L p 0 0 0 r
F P
0 g
s
.43945 .29273 m .46183 .27708 L .4495 .27752 L p 0 0 0 r
F P
0 g
s
.43207 .26199 m .42414 .26028 L .42913 .26034 L p 0 .181 0 r
F P
0 g
s
.42414 .26028 m .46384 .26885 L .44892 .27079 L p 0 .042 0 r
F P
0 g
s
.44892 .27079 m .41865 .25627 L p .42414 .26028 L 0 .042 0 r
F P
0 g
s
.41865 .25627 m .41114 .25267 L .41539 .25515 L p 0 .042 0 r
F P
0 g
s
.41539 .25515 m .42414 .26028 L p .41865 .25627 L 0 .042 0 r
F P
0 g
s
.40255 .23887 m .42894 .25979 L p .4272 .25334 L 0 .352 0 r
F P
0 g
s
.41114 .25267 m .44106 .26702 L p .43595 .26535 L 0 0 0 r
F P
0 g
s
.46384 .26885 m .43207 .26199 L p .43924 .26047 L 0 .181 0 r
F P
0 g
s
.42913 .26034 m .43924 .26047 L p .43207 .26199 L 0 .181 0 r
F P
0 g
s
.42894 .25979 m .43595 .26535 L .43288 .26114 L p 0 .352 0 r
F P
0 g
s
.43288 .26114 m .4272 .25334 L p .42894 .25979 L 0 .352 0 r
F P
0 g
s
.44106 .26702 m .44892 .27079 L .44443 .26891 L p 0 0 0 r
F P
0 g
s
.44443 .26891 m .43595 .26535 L p .44106 .26702 L 0 0 0 r
F P
0 g
s
.43924 .26047 m .47815 .25993 L .46384 .26885 L p 0 .181 0 r
F P
0 g
s
.74288 .17681 m .73042 .18777 L .74341 .20051 L p 0 .262 0 r
F P
0 g
s
.74341 .20051 m .75744 .18522 L .74288 .17681 L p 0 .262 0 r
F P
0 g
s
.67778 .20618 m .6686 .19734 L .6827 .19279 L p 0 .859 0 r
F P
0 g
s
.6827 .19279 m .68585 .1918 L .67778 .20618 L p 0 .859 0 r
F P
0 g
s
.30491 .2691 m .32039 .2857 L .33306 .28261 L p 0 0 .102 r
F P
0 g
s
.32039 .2857 m .30491 .2691 L .29019 .27256 L p 0 0 .389 r
F P
0 g
s
.46772 .189 m .48004 .18683 L .49087 .19611 L p 0 .441 0 r
F P
0 g
s
.48004 .18683 m .46772 .189 L .45698 .17781 L p 0 0 0 r
F P
0 g
s
.73612 .26399 m .7612 .25136 L .76992 .23597 L p 0 .924 0 r
F P
0 g
s
.40255 .23887 m .38095 .2268 L .38861 .24296 L p 0 .475 0 r
F P
0 g
s
.38861 .24296 m .41114 .25267 L .40255 .23887 L p 0 .475 0 r
F P
0 g
s
.39474 .17718 m .41352 .1871 L .4273 .17667 L p 0 .02 0 r
F P
0 g
s
.41352 .1871 m .39474 .17718 L .38431 .1907 L p 0 .173 0 r
F P
0 g
s
.36316 .29328 m .36833 .30138 L p .35087 .29115 L 0 0 .919 r
F P
0 g
s
.38273 .30308 m .36833 .30138 L .36316 .29328 L p 0 0 .982 r
F P
0 g
s
.38144 .29615 m .39937 .29521 L .39622 .30263 L p 0 0 .989 r
F P
0 g
s
.40416 .29524 m .39937 .29521 L p .39031 .29595 L 0 0 0 r
F P
0 g
s
.38148 .29636 m .38273 .30308 L p .37033 .29687 L 0 0 .982 r
F P
0 g
s
.39622 .30263 m .38273 .30308 L .38144 .29615 L p 0 0 .989 r
F P
0 g
s
.3796 .29681 m .40039 .29548 L p .39031 .29595 L 0 0 0 r
F P
0 g
s
.40039 .29548 m .40416 .29524 L p .39031 .29595 L 0 0 0 r
F P
0 g
s
.38144 .29615 m .38148 .29636 L p .36943 .29642 L 0 0 .982 r
F P
0 g
s
.36943 .29642 m .38148 .29636 L .37033 .29687 L 0 0 .982 r
F
.36316 .29328 m .38144 .29615 L p .36943 .29642 L F P
0 g
s
.38144 .29615 m .3796 .29681 L p .39031 .29595 L 0 0 0 r
F P
0 g
s
.39937 .29521 m .38144 .29615 L p .39031 .29595 L 0 0 0 r
F P
0 g
s
.36316 .29328 m .35544 .29239 L p .37079 .29461 L 0 0 0 r
F P
0 g
s
.38144 .29615 m .36316 .29328 L p .37079 .29461 L 0 0 0 r
F P
0 g
s
.34346 .28681 m .36316 .29328 L p .35087 .29115 L 0 0 .919 r
F P
0 g
s
.37628 .38517 m .33829 .37874 L .34682 .37529 L p 0 0 0 r
F P
0 g
s
.34682 .37529 m .37827 .38004 L .37628 .38517 L p 0 0 0 r
F P
0 g
s
.45259 .37279 m .41547 .38284 L .41014 .37841 L p 0 0 0 r
F P
0 g
s
.37827 .38004 m .41014 .37841 L .41547 .38284 L p 0 0 0 r
F P
0 g
s
.46021 .36715 m .4969 .34586 L .49908 .33071 L p 0 0 .834 r
F P
0 g
s
.4969 .34586 m .46021 .36715 L .45259 .37279 L p 0 0 .926 r
F P
0 g
s
.28976 .3451 m .25971 .31503 L .26021 .33143 L p 0 0 .853 r
F P
0 g
s
.26021 .33143 m .29409 .35822 L .28976 .3451 L p 0 0 .853 r
F P
0 g
s
.75744 .18522 m .74341 .20051 L .75127 .2176 L p 0 .518 0 r
F P
0 g
s
.75127 .2176 m .76787 .19968 L .75744 .18522 L p 0 .518 0 r
F P
0 g
s
.33149 .3747 m .29409 .35822 L .30298 .36537 L p 0 0 .785 r
F P
0 g
s
.30298 .36537 m .33829 .37874 L .33149 .3747 L p 0 0 .785 r
F P
0 g
s
.74722 .25241 m .76992 .23597 L .77229 .21767 L p 0 .819 0 r
F P
0 g
s
.67734 .27184 m .72132 .26879 L .73612 .26399 L p 0 .866 0 r
F P
0 g
s
.39994 .22124 m .37942 .20828 L .38095 .2268 L p 0 .377 0 r
F P
0 g
s
.38095 .2268 m .40255 .23887 L .39994 .22124 L p 0 .377 0 r
F P
0 g
s
.38431 .1907 m .40382 .2029 L .41352 .1871 L p 0 .173 0 r
F P
0 g
s
.40382 .2029 m .38431 .1907 L .37942 .20828 L p 0 .281 0 r
F P
0 g
s
.41547 .38284 m .45259 .37279 L .46021 .36715 L p 0 0 .93 r
F P
0 g
s
.41547 .38284 m .37628 .38517 L .37827 .38004 L p 0 0 0 r
F P
0 g
s
.42 .29041 m .39937 .29521 L .40416 .29524 L p 0 0 0 r
F P
0 g
s
.40416 .29524 m .42884 .28785 L .42 .29041 L p 0 0 0 r
F P
0 g
s
.46183 .27708 m .43945 .29273 L .44998 .30429 L p 0 0 .338 r
F P
0 g
s
.44998 .30429 m .47518 .28383 L .46183 .27708 L p 0 0 .338 r
F P
0 g
s
.68585 .1918 m .6827 .19279 L .69848 .19579 L p 0 0 0 r
F P
0 g
s
.69848 .19579 m .69887 .18308 L .68585 .1918 L p 0 0 0 r
F P
0 g
s
.37942 .20828 m .39994 .22124 L .40382 .2029 L p 0 .281 0 r
F P
0 g
s
.29019 .27256 m .30767 .29565 L .32039 .2857 L p 0 0 .389 r
F P
0 g
s
.30767 .29565 m .29019 .27256 L .27617 .28269 L p 0 0 .542 r
F P
0 g
s
.45698 .17781 m .46579 .18461 L .48004 .18683 L p 0 0 0 r
F P
0 g
s
.46579 .18461 m .45698 .17781 L .44269 .17342 L p 0 .008 0 r
F P
0 g
s
.76787 .19968 m .75127 .2176 L .75261 .23599 L p 0 .688 0 r
F P
0 g
s
.75261 .23599 m .77229 .21767 L .76787 .19968 L p 0 .688 0 r
F P
0 g
s
.77229 .21767 m .75261 .23599 L .74722 .25241 L p 0 .819 0 r
F P
0 g
s
.51825 .24707 m .47815 .25993 L .48949 .24559 L p 0 .421 0 r
F P
0 g
s
.48949 .24559 m .52164 .22944 L .51825 .24707 L p 0 .421 0 r
F P
0 g
s
.52164 .22944 m .48949 .24559 L .49596 .22824 L p 0 .464 0 r
F P
0 g
s
.50964 .26377 m .46384 .26885 L .47815 .25993 L p 0 .38 0 r
F P
0 g
s
.47815 .25993 m .51825 .24707 L .50964 .26377 L p 0 .38 0 r
F P
0 g
s
.49596 .22824 m .51921 .2138 L .52164 .22944 L p 0 .464 0 r
F P
0 g
s
.51921 .2138 m .49596 .22824 L .49646 .21076 L p 0 .498 0 r
F P
0 g
s
.29084 .32828 m .26536 .29775 L .25971 .31503 L p 0 0 .756 r
F P
0 g
s
.25971 .31503 m .28976 .3451 L .29084 .32828 L p 0 0 .756 r
F P
0 g
s
.66461 .25895 m .63238 .24665 L .63223 .22941 L p 0 .641 0 r
F P
0 g
s
.63223 .22941 m .65649 .24222 L .66461 .25895 L p 0 .641 0 r
F P
0 g
s
.63238 .24665 m .66461 .25895 L .67734 .27184 L p 0 .405 0 r
F P
0 g
s
.4641 .35504 m .49908 .33071 L .49564 .31327 L p 0 0 .734 r
F P
0 g
s
.49908 .33071 m .4641 .35504 L .46021 .36715 L p 0 0 .834 r
F P
0 g
s
.27617 .28269 m .29719 .31074 L .30767 .29565 L p 0 0 .542 r
F P
0 g
s
.29719 .31074 m .27617 .28269 L .26536 .29775 L p 0 0 .653 r
F P
0 g
s
.65649 .24222 m .63223 .22941 L .63795 .21552 L p 0 .821 0 r
F P
0 g
s
.63795 .21552 m .65435 .22443 L .65649 .24222 L p 0 .821 0 r
F P
0 g
s
.33306 .28261 m .35544 .29239 L .36316 .29328 L p 0 0 0 r
F P
0 g
s
.35544 .29239 m .33306 .28261 L .32039 .2857 L p 0 0 .208 r
F P
0 g
s
.47518 .28383 m .44998 .30429 L .45852 .32053 L p 0 0 .523 r
F P
0 g
s
.45852 .32053 m .48717 .29665 L .47518 .28383 L p 0 0 .523 r
F P
0 g
s
.73612 .26399 m .69259 .2787 L .67734 .27184 L p 0 .866 0 r
F P
0 g
s
.69259 .2787 m .73612 .26399 L .74722 .25241 L p 0 .902 0 r
F P
0 g
s
.37463 .38259 m .33149 .3747 L .33829 .37874 L p 0 0 .896 r
F P
0 g
s
.33829 .37874 m .37628 .38517 L .37463 .38259 L p 0 0 .896 r
F P
0 g
s
.26536 .29775 m .29084 .32828 L .29719 .31074 L p 0 0 .653 r
F P
0 g
s
.43595 .26535 m .40255 .23887 L .41114 .25267 L p 0 0 0 r
F P
0 g
s
.69887 .18308 m .69848 .19579 L .71317 .20591 L p 0 .028 0 r
F P
0 g
s
.71317 .20591 m .71464 .18163 L .69887 .18308 L p 0 .028 0 r
F P
0 g
s
.63832 .26431 m .67734 .27184 L .69259 .2787 L p 0 .013 0 r
F P
0 g
s
.67734 .27184 m .63832 .26431 L .63238 .24665 L p 0 .405 0 r
F P
0 g
s
.49726 .27674 m .44892 .27079 L .46384 .26885 L p 0 .337 0 r
F P
0 g
s
.46384 .26885 m .50964 .26377 L .49726 .27674 L p 0 .337 0 r
F P
0 g
s
.46021 .36715 m .41967 .37964 L .41547 .38284 L p 0 0 .93 r
F P
0 g
s
.37628 .38517 m .41547 .38284 L .41967 .37964 L p 0 0 .945 r
F P
0 g
s
.32764 .36373 m .28976 .3451 L .29409 .35822 L p 0 0 .918 r
F P
0 g
s
.29409 .35822 m .33149 .3747 L .32764 .36373 L p 0 0 .918 r
F P
0 g
s
.49646 .21076 m .51129 .20285 L .51921 .2138 L p 0 .498 0 r
F P
0 g
s
.51129 .20285 m .49646 .21076 L .49087 .19611 L p 0 .371 0 r
F P
0 g
s
.44269 .17342 m .45057 .18994 L .46579 .18461 L p 0 .008 0 r
F P
0 g
s
.45057 .18994 m .44269 .17342 L .4273 .17667 L p 0 .15 0 r
F P
0 g
s
.46352 .33853 m .49564 .31327 L .48717 .29665 L p 0 0 .638 r
F P
0 g
s
.49564 .31327 m .46352 .33853 L .4641 .35504 L p 0 0 .734 r
F P
0 g
s
.41967 .37964 m .46021 .36715 L .4641 .35504 L p 0 0 .912 r
F P
0 g
s
.42884 .28785 m .40416 .29524 L .40979 .30285 L p 0 0 .13 r
F P
0 g
s
.40979 .30285 m .43945 .29273 L .42884 .28785 L p 0 0 .13 r
F P
0 g
s
.48717 .29665 m .45852 .32053 L .46352 .33853 L p 0 0 .638 r
F P
0 g
s
.65435 .22443 m .63795 .21552 L .64865 .20739 L p 0 .946 0 r
F P
0 g
s
.64865 .20739 m .65861 .20856 L .65435 .22443 L p 0 .946 0 r
F P
0 g
s
.35544 .29239 m .3796 .29681 L .38144 .29615 L p 0 0 0 r
F P
0 g
s
.41967 .37964 m .37463 .38259 L .37628 .38517 L p 0 0 .945 r
F P
0 g
s
.74722 .25241 m .7078 .27831 L .69259 .2787 L p 0 .902 0 r
F P
0 g
s
.69259 .2787 m .64903 .27946 L .63832 .26431 L p 0 .013 0 r
F P
0 g
s
.64903 .27946 m .69259 .2787 L .7078 .27831 L p 0 .601 0 r
F P
0 g
s
.7078 .27831 m .74722 .25241 L .75261 .23599 L p 0 .83 0 r
F P
0 g
s
.4272 .25334 m .39994 .22124 L .40255 .23887 L p 0 .352 0 r
F P
0 g
s
.71464 .18163 m .71317 .20591 L .72412 .2214 L p 0 .313 0 r
F P
0 g
s
.72412 .2214 m .73042 .18777 L .71464 .18163 L p 0 .313 0 r
F P
0 g
s
.44892 .27079 m .49726 .27674 L .48315 .28382 L p 0 .275 0 r
F P
0 g
s
.48315 .28382 m .43595 .26535 L .44892 .27079 L p 0 .275 0 r
F P
0 g
s
.32039 .2857 m .34647 .29901 L .35544 .29239 L p 0 0 .208 r
F P
0 g
s
.34647 .29901 m .32039 .2857 L .30767 .29565 L p 0 0 .472 r
F P
0 g
s
.4273 .17667 m .43709 .20197 L .45057 .18994 L p 0 .15 0 r
F P
0 g
s
.43709 .20197 m .4273 .17667 L .41352 .1871 L p 0 .232 0 r
F P
0 g
s
.49087 .19611 m .49915 .19856 L .51129 .20285 L p 0 .371 0 r
F P
0 g
s
.49915 .19856 m .49087 .19611 L .48004 .18683 L p 0 0 0 r
F P
0 g
s
.66255 .20653 m .6686 .19734 L .65861 .20856 L p 0 0 0 r
F P
0 g
s
.65861 .20856 m .64865 .20739 L .66255 .20653 L p 0 0 0 r
F P
0 g
s
.32749 .34766 m .29084 .32828 L .28976 .3451 L p 0 0 .816 r
F P
0 g
s
.28976 .3451 m .32764 .36373 L .32749 .34766 L p 0 0 .816 r
F P
0 g
s
.53233 .27889 m .49726 .27674 L .50964 .26377 L p 0 .366 0 r
F P
0 g
s
.50964 .26377 m .54091 .26213 L .53233 .27889 L p 0 .366 0 r
F P
0 g
s
.54091 .26213 m .50964 .26377 L .51825 .24707 L p 0 .408 0 r
F P
0 g
s
.3796 .29681 m .35544 .29239 L .34647 .29901 L p 0 0 .215 r
F P
0 g
s
.37362 .37259 m .32764 .36373 L .33149 .3747 L p 0 0 .951 r
F P
0 g
s
.33149 .3747 m .37463 .38259 L .37362 .37259 L p 0 0 .951 r
F P
0 g
s
.42198 .36923 m .4641 .35504 L .46352 .33853 L p 0 0 .808 r
F P
0 g
s
.4641 .35504 m .42198 .36923 L .41967 .37964 L p 0 0 .912 r
F P
0 g
s
.43945 .29273 m .40979 .30285 L .41526 .31684 L p 0 0 .458 r
F P
0 g
s
.41526 .31684 m .44998 .30429 L .43945 .29273 L p 0 0 .458 r
F P
0 g
s
.54533 .24425 m .51825 .24707 L .52164 .22944 L p 0 .441 0 r
F P
0 g
s
.51825 .24707 m .54533 .24425 L .54091 .26213 L p 0 .408 0 r
F P
0 g
s
.42428 .23679 m .40382 .2029 L .39994 .22124 L p 0 .337 0 r
F P
0 g
s
.39994 .22124 m .4272 .25334 L .42428 .23679 L p 0 .337 0 r
F P
0 g
s
.37463 .38259 m .41967 .37964 L .42198 .36923 L p 0 0 .949 r
F P
0 g
s
.75261 .23599 m .72035 .27065 L .7078 .27831 L p 0 .83 0 r
F P
0 g
s
.72035 .27065 m .75261 .23599 L .75127 .2176 L p 0 .735 0 r
F P
0 g
s
.40416 .29524 m .3796 .29681 L .37753 .30512 L p 0 0 .192 r
F P
0 g
s
.37753 .30512 m .40979 .30285 L .40416 .29524 L p 0 0 .192 r
F P
0 g
s
.41352 .1871 m .42777 .21858 L .43709 .20197 L p 0 .232 0 r
F P
0 g
s
.42777 .21858 m .41352 .1871 L .40382 .2029 L p 0 .291 0 r
F P
0 g
s
.73042 .18777 m .72412 .2214 L .72938 .23948 L p 0 .494 0 r
F P
0 g
s
.72938 .23948 m .74341 .20051 L .73042 .18777 L p 0 .494 0 r
F P
0 g
s
.61215 .26398 m .63832 .26431 L .64903 .27946 L p 0 .558 0 r
F P
0 g
s
.63832 .26431 m .61215 .26398 L .60967 .24597 L p 0 .704 0 r
F P
0 g
s
.60967 .24597 m .63238 .24665 L .63832 .26431 L p 0 .704 0 r
F P
0 g
s
.7078 .27831 m .66275 .28959 L .64903 .27946 L p 0 .601 0 r
F P
0 g
s
.66275 .28959 m .7078 .27831 L .72035 .27065 L p 0 .926 0 r
F P
0 g
s
.40382 .2029 m .42428 .23679 L .42777 .21858 L p 0 .291 0 r
F P
0 g
s
.52101 .29177 m .48315 .28382 L .49726 .27674 L p 0 .309 0 r
F P
0 g
s
.49726 .27674 m .53233 .27889 L .52101 .29177 L p 0 .309 0 r
F P
0 g
s
.30767 .29565 m .33784 .31209 L .34647 .29901 L p 0 0 .472 r
F P
0 g
s
.33784 .31209 m .30767 .29565 L .29719 .31074 L p 0 0 .612 r
F P
0 g
s
.34647 .29901 m .37753 .30512 L .3796 .29681 L p 0 0 .215 r
F P
0 g
s
.61911 .28158 m .64903 .27946 L .66275 .28959 L p 0 .354 0 r
F P
0 g
s
.64903 .27946 m .61911 .28158 L .61215 .26398 L p 0 .558 0 r
F P
0 g
s
.3311 .32935 m .29719 .31074 L .29084 .32828 L p 0 0 .716 r
F P
0 g
s
.29084 .32828 m .32749 .34766 L .3311 .32935 L p 0 0 .716 r
F P
0 g
s
.42198 .36923 m .37362 .37259 L .37463 .38259 L p 0 0 .949 r
F P
0 g
s
.72801 .25693 m .75127 .2176 L .74341 .20051 L p 0 .626 0 r
F P
0 g
s
.75127 .2176 m .72801 .25693 L .72035 .27065 L p 0 .735 0 r
F P
0 g
s
.74341 .20051 m .72938 .23948 L .72801 .25693 L p 0 .626 0 r
F P
0 g
s
.63238 .24665 m .60967 .24597 L .61213 .23055 L p 0 .838 0 r
F P
0 g
s
.61213 .23055 m .63223 .22941 L .63238 .24665 L p 0 .838 0 r
F P
0 g
s
.52164 .22944 m .54483 .22823 L .54533 .24425 L p 0 .441 0 r
F P
0 g
s
.54483 .22823 m .52164 .22944 L .51921 .2138 L p 0 .451 0 r
F P
0 g
s
.43595 .26535 m .48315 .28382 L .46966 .28372 L p 0 .15 0 r
F P
0 g
s
.46966 .28372 m .4272 .25334 L .43595 .26535 L p 0 .15 0 r
F P
0 g
s
.29719 .31074 m .3311 .32935 L .33784 .31209 L p 0 0 .612 r
F P
0 g
s
.48004 .18683 m .48487 .20179 L .49915 .19856 L p 0 0 0 r
F P
0 g
s
.48487 .20179 m .48004 .18683 L .46579 .18461 L p 0 .146 0 r
F P
0 g
s
.42196 .35332 m .46352 .33853 L .45852 .32053 L p 0 0 .713 r
F P
0 g
s
.46352 .33853 m .42196 .35332 L .42198 .36923 L p 0 0 .808 r
F P
0 g
s
.44998 .30429 m .41526 .31684 L .41958 .33474 L p 0 0 .61 r
F P
0 g
s
.41958 .33474 m .45852 .32053 L .44998 .30429 L p 0 0 .61 r
F P
0 g
s
.6686 .19734 m .66255 .20653 L .67726 .21321 L p 0 0 0 r
F P
0 g
s
.67726 .21321 m .6827 .19279 L .6686 .19734 L p 0 0 0 r
F P
0 g
s
.45852 .32053 m .41958 .33474 L .42196 .35332 L p 0 0 .713 r
F P
0 g
s
.50881 .29858 m .46966 .28372 L .48315 .28382 L p 0 .205 0 r
F P
0 g
s
.48315 .28382 m .52101 .29177 L .50881 .29858 L p 0 .205 0 r
F P
0 g
s
.72035 .27065 m .67719 .29293 L .66275 .28959 L p 0 .926 0 r
F P
0 g
s
.6294 .29585 m .66275 .28959 L .67719 .29293 L p 0 0 0 r
F P
0 g
s
.66275 .28959 m .6294 .29585 L .61911 .28158 L p 0 .354 0 r
F P
0 g
s
.37753 .30512 m .34647 .29901 L .33784 .31209 L p 0 0 .507 r
F P
0 g
s
.37344 .35682 m .32749 .34766 L .32764 .36373 L p 0 0 .844 r
F P
0 g
s
.32764 .36373 m .37362 .37259 L .37344 .35682 L p 0 0 .844 r
F P
0 g
s
.63223 .22941 m .61213 .23055 L .61918 .22035 L p 0 .97 0 r
F P
0 g
s
.61918 .22035 m .63795 .21552 L .63223 .22941 L p 0 .97 0 r
F P
0 g
s
.67719 .29293 m .72035 .27065 L .72801 .25693 L p 0 .896 0 r
F P
0 g
s
.37362 .37259 m .42198 .36923 L .42196 .35332 L p 0 0 .842 r
F P
0 g
s
.40979 .30285 m .37753 .30512 L .37559 .31972 L p 0 0 .506 r
F P
0 g
s
.37559 .31972 m .41526 .31684 L .40979 .30285 L p 0 0 .506 r
F P
0 g
s
.51921 .2138 m .53942 .21678 L .54483 .22823 L p 0 .451 0 r
F P
0 g
s
.53942 .21678 m .51921 .2138 L .51129 .20285 L p 0 .268 0 r
F P
0 g
s
.55821 .28271 m .53233 .27889 L .54091 .26213 L p 0 .442 0 r
F P
0 g
s
.54091 .26213 m .56352 .26448 L .55821 .28271 L p 0 .442 0 r
F P
0 g
s
.56352 .26448 m .54091 .26213 L .54533 .24425 L p 0 .489 0 r
F P
0 g
s
.55095 .29823 m .52101 .29177 L .53233 .27889 L p 0 .379 0 r
F P
0 g
s
.53233 .27889 m .55821 .28271 L .55095 .29823 L p 0 .379 0 r
F P
0 g
s
.4272 .25334 m .46966 .28372 L .45914 .27637 L p 0 0 0 r
F P
0 g
s
.45914 .27637 m .42428 .23679 L .4272 .25334 L p 0 0 0 r
F P
0 g
s
.46579 .18461 m .47095 .21207 L .48487 .20179 L p 0 .146 0 r
F P
0 g
s
.47095 .21207 m .46579 .18461 L .45057 .18994 L p 0 .233 0 r
F P
0 g
s
.33784 .31209 m .37559 .31972 L .37753 .30512 L p 0 0 .507 r
F P
0 g
s
.42196 .35332 m .37344 .35682 L .37362 .37259 L p 0 0 .842 r
F P
0 g
s
.59625 .28415 m .61911 .28158 L .6294 .29585 L p 0 .491 0 r
F P
0 g
s
.61911 .28158 m .59625 .28415 L .59282 .26546 L p 0 .625 0 r
F P
0 g
s
.59282 .26546 m .61215 .26398 L .61911 .28158 L p 0 .625 0 r
F P
0 g
s
.61215 .26398 m .59282 .26546 L .59211 .24745 L p 0 .733 0 r
F P
0 g
s
.59211 .24745 m .60967 .24597 L .61215 .26398 L p 0 .733 0 r
F P
0 g
s
.6827 .19279 m .67726 .21321 L .69014 .22634 L p 0 .201 0 r
F P
0 g
s
.69014 .22634 m .69848 .19579 L .6827 .19279 L p 0 .201 0 r
F P
0 g
s
.37559 .31972 m .33784 .31209 L .3311 .32935 L p 0 0 .648 r
F P
0 g
s
.37414 .33806 m .3311 .32935 L .32749 .34766 L p 0 0 .748 r
F P
0 g
s
.32749 .34766 m .37344 .35682 L .37414 .33806 L p 0 0 .748 r
F P
0 g
s
.56599 .24654 m .54533 .24425 L .54483 .22823 L p 0 .533 0 r
F P
0 g
s
.54533 .24425 m .56599 .24654 L .56352 .26448 L p 0 .489 0 r
F P
0 g
s
.72801 .25693 m .68992 .28883 L .67719 .29293 L p 0 .896 0 r
F P
0 g
s
.67719 .29293 m .64134 .30441 L .6294 .29585 L p 0 0 0 r
F P
0 g
s
.64134 .30441 m .67719 .29293 L .68992 .28883 L p 0 .729 0 r
F P
0 g
s
.54292 .30845 m .50881 .29858 L .52101 .29177 L p 0 .27 0 r
F P
0 g
s
.52101 .29177 m .55095 .29823 L .54292 .30845 L p 0 .27 0 r
F P
0 g
s
.46966 .28372 m .50881 .29858 L .49778 .29807 L p 0 0 0 r
F P
0 g
s
.49778 .29807 m .45914 .27637 L .46966 .28372 L p 0 0 0 r
F P
0 g
s
.37344 .35682 m .42196 .35332 L .41958 .33474 L p 0 0 .749 r
F P
0 g
s
.41526 .31684 m .37559 .31972 L .37414 .33806 L p 0 0 .65 r
F P
0 g
s
.37414 .33806 m .41958 .33474 L .41526 .31684 L p 0 0 .65 r
F P
0 g
s
.3311 .32935 m .37414 .33806 L .37559 .31972 L p 0 0 .648 r
F P
0 g
s
.68992 .28883 m .72801 .25693 L .72938 .23948 L p 0 .798 0 r
F P
0 g
s
.63795 .21552 m .61918 .22035 L .62968 .21723 L p 0 .898 0 r
F P
0 g
s
.62968 .21723 m .64865 .20739 L .63795 .21552 L p 0 .898 0 r
F P
0 g
s
.60183 .30045 m .6294 .29585 L .64134 .30441 L p 0 .262 0 r
F P
0 g
s
.6294 .29585 m .60183 .30045 L .59625 .28415 L p 0 .491 0 r
F P
0 g
s
.60967 .24597 m .59211 .24745 L .59426 .23311 L p 0 .842 0 r
F P
0 g
s
.59426 .23311 m .61213 .23055 L .60967 .24597 L p 0 .842 0 r
F P
0 g
s
.45346 .26292 m .42777 .21858 L .42428 .23679 L p 0 .297 0 r
F P
0 g
s
.42428 .23679 m .45914 .27637 L .45346 .26292 L p 0 .297 0 r
F P
0 g
s
.57794 .28456 m .55821 .28271 L .56352 .26448 L p 0 .527 0 r
F P
0 g
s
.56352 .26448 m .57902 .26559 L .57794 .28456 L p 0 .527 0 r
F P
0 g
s
.57902 .26559 m .56352 .26448 L .56599 .24654 L p 0 .588 0 r
F P
0 g
s
.57686 .30137 m .55095 .29823 L .55821 .28271 L p 0 .451 0 r
F P
0 g
s
.55821 .28271 m .57794 .28456 L .57686 .30137 L p 0 .451 0 r
F P
0 g
s
.41958 .33474 m .37414 .33806 L .37344 .35682 L p 0 0 .749 r
F P
0 g
s
.45057 .18994 m .45989 .22762 L .47095 .21207 L p 0 .233 0 r
F P
0 g
s
.45989 .22762 m .45057 .18994 L .43709 .20197 L p 0 .283 0 r
F P
0 g
s
.57794 .28456 m .59625 .28415 L .60183 .30045 L p 0 .499 0 r
F P
0 g
s
.59625 .28415 m .57794 .28456 L .57902 .26559 L p 0 .601 0 r
F P
0 g
s
.57902 .26559 m .59282 .26546 L .59625 .28415 L p 0 .601 0 r
F P
0 g
s
.59282 .26546 m .57902 .26559 L .57992 .24757 L p 0 .683 0 r
F P
0 g
s
.57992 .24757 m .59211 .24745 L .59282 .26546 L p 0 .683 0 r
F P
0 g
s
.51129 .20285 m .52995 .21197 L .53942 .21678 L p 0 .268 0 r
F P
0 g
s
.52995 .21197 m .51129 .20285 L .49915 .19856 L p 0 .101 0 r
F P
0 g
s
.69848 .19579 m .69014 .22634 L .69887 .24359 L p 0 .424 0 r
F P
0 g
s
.69887 .24359 m .71317 .20591 L .69848 .19579 L p 0 .424 0 r
F P
0 g
s
.56599 .24654 m .57992 .24757 L .57902 .26559 L p 0 .588 0 r
F P
0 g
s
.60183 .30045 m .57686 .30137 L .57794 .28456 L p 0 .499 0 r
F P
0 g
s
.54483 .22823 m .5652 .23187 L .56599 .24654 L p 0 .533 0 r
F P
0 g
s
.5652 .23187 m .54483 .22823 L .53942 .21678 L p 0 .55 0 r
F P
0 g
s
.4537 .24565 m .43709 .20197 L .42777 .21858 L p 0 .312 0 r
F P
0 g
s
.42777 .21858 m .45346 .26292 L .4537 .24565 L p 0 .312 0 r
F P
0 g
s
.43709 .20197 m .4537 .24565 L .45989 .22762 L p 0 .283 0 r
F P
0 g
s
.69868 .27786 m .72938 .23948 L .72412 .2214 L p 0 .691 0 r
F P
0 g
s
.72938 .23948 m .69868 .27786 L .68992 .28883 L p 0 .798 0 r
F P
0 g
s
.57992 .24757 m .56599 .24654 L .5652 .23187 L p 0 .649 0 r
F P
0 g
s
.57595 .31322 m .54292 .30845 L .55095 .29823 L p 0 .314 0 r
F P
0 g
s
.55095 .29823 m .57686 .30137 L .57595 .31322 L p 0 .314 0 r
F P
0 g
s
.53547 .31155 m .49778 .29807 L .50881 .29858 L p 0 0 0 r
F P
0 g
s
.50881 .29858 m .54292 .30845 L .53547 .31155 L p 0 0 0 r
F P
0 g
s
.64134 .30441 m .60866 .31163 L .60183 .30045 L p 0 .262 0 r
F P
0 g
s
.57686 .30137 m .60183 .30045 L .60866 .31163 L p 0 .323 0 r
F P
0 g
s
.59211 .24745 m .57992 .24757 L .58049 .2335 L p 0 .766 0 r
F P
0 g
s
.58049 .2335 m .59426 .23311 L .59211 .24745 L p 0 .766 0 r
F P
0 g
s
.71317 .20591 m .69887 .24359 L .70188 .26185 L p 0 .573 0 r
F P
0 g
s
.70188 .26185 m .72412 .2214 L .71317 .20591 L p 0 .573 0 r
F P
0 g
s
.68992 .28883 m .65294 .30571 L .64134 .30441 L p 0 .729 0 r
F P
0 g
s
.65294 .30571 m .68992 .28883 L .69868 .27786 L p 0 .972 0 r
F P
0 g
s
.60866 .31163 m .64134 .30441 L .65294 .30571 L p 0 0 0 r
F P
0 g
s
.61213 .23055 m .59426 .23311 L .59894 .22493 L p 0 .949 0 r
F P
0 g
s
.59894 .22493 m .61918 .22035 L .61213 .23055 L p 0 .949 0 r
F P
0 g
s
.48985 .29021 m .45346 .26292 L .45914 .27637 L p 0 .407 0 r
F P
0 g
s
.45914 .27637 m .49778 .29807 L .48985 .29021 L p 0 .407 0 r
F P
0 g
s
.72412 .2214 m .70188 .26185 L .69868 .27786 L p 0 .691 0 r
F P
0 g
s
.5652 .23187 m .58049 .2335 L .57992 .24757 L p 0 .649 0 r
F P
0 g
s
.60866 .31163 m .57595 .31322 L .57686 .30137 L p 0 .323 0 r
F P
0 g
s
.64865 .20739 m .62968 .21723 L .64187 .22187 L p 0 0 0 r
F P
0 g
s
.64187 .22187 m .66255 .20653 L .64865 .20739 L p 0 0 0 r
F P
0 g
s
.49915 .19856 m .518 .21475 L .52995 .21197 L p 0 .101 0 r
F P
0 g
s
.518 .21475 m .49915 .19856 L .48487 .20179 L p 0 .268 0 r
F P
0 g
s
.53942 .21678 m .56123 .223 L p .5471 .22128 L 0 .55 0 r
F P
0 g
s
.56123 .223 m .53942 .21678 L .52995 .21197 L p 0 0 0 r
F P
0 g
s
.52995 .21197 m .55469 .22155 L .56123 .223 L p 0 0 0 r
F P
0 g
s
.56123 .223 m .55469 .22155 L p .57 .22406 L 0 0 0 r
F P
0 g
s
.58062 .2258 m .56123 .223 L p .57 .22406 L 0 0 0 r
F P
0 g
s
.56123 .223 m .5652 .23187 L p .5471 .22128 L 0 .55 0 r
F P
0 g
s
.58049 .2335 m .5652 .23187 L .56123 .223 L p 0 .706 0 r
F P
0 g
s
.57537 .31803 m .53547 .31155 L .54292 .30845 L p 0 0 0 r
F P
0 g
s
.54292 .30845 m .57595 .31322 L .57537 .31803 L p 0 0 0 r
F P
0 g
s
.65294 .30571 m .61559 .31574 L .60866 .31163 L p 0 0 0 r
F P
0 g
s
.57595 .31322 m .60866 .31163 L .61559 .31574 L p 0 0 0 r
F P
0 g
s
.58062 .2258 m .59894 .22493 L .59426 .23311 L p 0 .852 0 r
F P
0 g
s
.56123 .223 m .58062 .2258 L .58062 .22603 L p 0 .706 0 r
F P
0 g
s
.58062 .22603 m .58049 .2335 L p .56828 .22685 L 0 .706 0 r
F P
0 g
s
.56123 .223 m .58062 .22603 L .56828 .22685 L 0 .706 0 r
F
.61918 .22035 m .59894 .22493 L .60541 .22442 L p 0 .718 0 r
F P
0 g
s
.59894 .22493 m .58062 .2258 L .58029 .22594 L p 0 .621 0 r
F P
0 g
s
.60541 .22442 m .59919 .22491 L p .59067 .22537 L 0 .621 0 r
F P
0 g
s
.58029 .22594 m .60156 .22466 L p .59067 .22537 L 0 .621 0 r
F P
0 g
s
.60156 .22466 m .60541 .22442 L p .59067 .22537 L 0 .621 0 r
F P
0 g
s
.59919 .22491 m .59894 .22493 L p .59828 .22496 L 0 .621 0 r
F P
0 g
s
.58029 .22594 m .59919 .22491 L .59828 .22496 L 0 .621 0 r
F
.59426 .23311 m .58049 .2335 L .58062 .22603 L p 0 .852 0 r
F P
0 g
s
.58062 .22603 m .58062 .2258 L p .58101 .22601 L 0 .852 0 r
F P
0 g
s
.58101 .22601 m .59426 .23311 L .58062 .22603 L 0 .852 0 r
F
.66218 .29941 m .69868 .27786 L .70188 .26185 L p 0 .869 0 r
F P
0 g
s
.69868 .27786 m .66218 .29941 L .65294 .30571 L p 0 .972 0 r
F P
0 g
s
.48647 .27624 m .4537 .24565 L .45346 .26292 L p 0 .463 0 r
F P
0 g
s
.45346 .26292 m .48985 .29021 L .48647 .27624 L p 0 .463 0 r
F P
0 g
s
.52988 .30686 m .48985 .29021 L .49778 .29807 L p 0 .497 0 r
F P
0 g
s
.49778 .29807 m .53547 .31155 L .52988 .30686 L p 0 .497 0 r
F P
0 g
s
.61559 .31574 m .65294 .30571 L .66218 .29941 L p 0 .919 0 r
F P
0 g
s
.61559 .31574 m .57537 .31803 L .57595 .31322 L p 0 0 0 r
F P
0 g
s
.60541 .22442 m .62968 .21723 L .61918 .22035 L p 0 .718 0 r
F P
0 g
s
.66255 .20653 m .64187 .22187 L .65359 .23357 L p 0 .247 0 r
F P
0 g
s
.65359 .23357 m .67726 .21321 L .66255 .20653 L p 0 .247 0 r
F P
0 g
s
.48487 .20179 m .50564 .22476 L .518 .21475 L p 0 .268 0 r
F P
0 g
s
.50564 .22476 m .48487 .20179 L .47095 .21207 L p 0 .345 0 r
F P
0 g
s
.48832 .25855 m .45989 .22762 L .4537 .24565 L p 0 .433 0 r
F P
0 g
s
.4537 .24565 m .48647 .27624 L .48832 .25855 L p 0 .433 0 r
F P
0 g
s
.66741 .28646 m .70188 .26185 L .69887 .24359 L p 0 .748 0 r
F P
0 g
s
.70188 .26185 m .66741 .28646 L .66218 .29941 L p 0 .869 0 r
F P
0 g
s
.47095 .21207 m .49512 .24029 L .50564 .22476 L p 0 .345 0 r
F P
0 g
s
.49512 .24029 m .47095 .21207 L .45989 .22762 L p 0 .394 0 r
F P
0 g
s
.55469 .22155 m .52995 .21197 L .518 .21475 L p 0 .169 0 r
F P
0 g
s
.67726 .21321 m .65359 .23357 L .66273 .25031 L p 0 .478 0 r
F P
0 g
s
.66273 .25031 m .69014 .22634 L .67726 .21321 L p 0 .478 0 r
F P
0 g
s
.57523 .31483 m .52988 .30686 L .53547 .31155 L p 0 .664 0 r
F P
0 g
s
.53547 .31155 m .57537 .31803 L .57523 .31483 L p 0 .664 0 r
F P
0 g
s
.45989 .22762 m .48832 .25855 L .49512 .24029 L p 0 .394 0 r
F P
0 g
s
.66218 .29941 m .62146 .31192 L .61559 .31574 L p 0 .919 0 r
F P
0 g
s
.57537 .31803 m .61559 .31574 L .62146 .31192 L p 0 .82 0 r
F P
0 g
s
.52717 .29505 m .48647 .27624 L .48985 .29021 L p 0 .573 0 r
F P
0 g
s
.48985 .29021 m .52988 .30686 L .52717 .29505 L p 0 .573 0 r
F P
0 g
s
.66763 .26907 m .69887 .24359 L .69014 .22634 L p 0 .626 0 r
F P
0 g
s
.69887 .24359 m .66763 .26907 L .66741 .28646 L p 0 .748 0 r
F P
0 g
s
.62146 .31192 m .66218 .29941 L .66741 .28646 L p 0 .891 0 r
F P
0 g
s
.62968 .21723 m .60541 .22442 L .61257 .23185 L p 0 .036 0 r
F P
0 g
s
.61257 .23185 m .64187 .22187 L .62968 .21723 L p 0 .036 0 r
F P
0 g
s
.69014 .22634 m .66273 .25031 L .66763 .26907 L p 0 .626 0 r
F P
0 g
s
.55469 .22155 m .58029 .22594 L .58062 .2258 L p 0 0 0 r
F P
0 g
s
.62146 .31192 m .57523 .31483 L .57537 .31803 L p 0 .82 0 r
F P
0 g
s
.518 .21475 m .54669 .22796 L .55469 .22155 L p 0 .169 0 r
F P
0 g
s
.54669 .22796 m .518 .21475 L .50564 .22476 L p 0 .343 0 r
F P
0 g
s
.5279 .27809 m .48832 .25855 L .48647 .27624 L p 0 .53 0 r
F P
0 g
s
.48647 .27624 m .52717 .29505 L .5279 .27809 L p 0 .53 0 r
F P
0 g
s
.58029 .22594 m .55469 .22155 L .54669 .22796 L p 0 .16 0 r
F P
0 g
s
.57556 .304 m .52717 .29505 L .52988 .30686 L p 0 .697 0 r
F P
0 g
s
.52988 .30686 m .57523 .31483 L .57556 .304 L p 0 .697 0 r
F P
0 g
s
.62519 .30068 m .66741 .28646 L .66763 .26907 L p 0 .778 0 r
F P
0 g
s
.66741 .28646 m .62519 .30068 L .62146 .31192 L p 0 .891 0 r
F P
0 g
s
.64187 .22187 m .61257 .23185 L .61915 .24604 L p 0 .389 0 r
F P
0 g
s
.61915 .24604 m .65359 .23357 L .64187 .22187 L p 0 .389 0 r
F P
0 g
s
.57523 .31483 m .62146 .31192 L .62519 .30068 L p 0 .813 0 r
F P
0 g
s
.60541 .22442 m .58029 .22594 L .57954 .23405 L p 0 .121 0 r
F P
0 g
s
.57954 .23405 m .61257 .23185 L .60541 .22442 L p 0 .121 0 r
F P
0 g
s
.50564 .22476 m .53863 .2412 L .54669 .22796 L p 0 .343 0 r
F P
0 g
s
.53863 .2412 m .50564 .22476 L .49512 .24029 L p 0 .425 0 r
F P
0 g
s
.54669 .22796 m .57954 .23405 L .58029 .22594 L p 0 .16 0 r
F P
0 g
s
.53195 .25898 m .49512 .24029 L .48832 .25855 L p 0 .481 0 r
F P
0 g
s
.48832 .25855 m .5279 .27809 L .53195 .25898 L p 0 .481 0 r
F P
0 g
s
.62519 .30068 m .57556 .304 L .57523 .31483 L p 0 .813 0 r
F P
0 g
s
.49512 .24029 m .53195 .25898 L .53863 .2412 L p 0 .425 0 r
F P
0 g
s
.62609 .28388 m .66763 .26907 L .66273 .25031 L p 0 .674 0 r
F P
0 g
s
.66763 .26907 m .62609 .28388 L .62519 .30068 L p 0 .778 0 r
F P
0 g
s
.65359 .23357 m .61915 .24604 L .62395 .26449 L p 0 .559 0 r
F P
0 g
s
.62395 .26449 m .66273 .25031 L .65359 .23357 L p 0 .559 0 r
F P
0 g
s
.66273 .25031 m .62395 .26449 L .62609 .28388 L p 0 .674 0 r
F P
0 g
s
.57954 .23405 m .54669 .22796 L .53863 .2412 L p 0 .392 0 r
F P
0 g
s
.57631 .28732 m .5279 .27809 L .52717 .29505 L p 0 .63 0 r
F P
0 g
s
.52717 .29505 m .57556 .304 L .57631 .28732 L p 0 .63 0 r
F P
0 g
s
.57556 .304 m .62519 .30068 L .62609 .28388 L p 0 .723 0 r
F P
0 g
s
.61257 .23185 m .57954 .23405 L .5785 .24886 L p 0 .418 0 r
F P
0 g
s
.5785 .24886 m .61915 .24604 L .61257 .23185 L p 0 .418 0 r
F P
0 g
s
.53863 .2412 m .5785 .24886 L .57954 .23405 L p 0 .392 0 r
F P
0 g
s
.62609 .28388 m .57631 .28732 L .57556 .304 L p 0 .723 0 r
F P
0 g
s
.5785 .24886 m .53863 .2412 L .53195 .25898 L p 0 .496 0 r
F P
0 g
s
.57736 .26776 m .53195 .25898 L .5279 .27809 L p 0 .567 0 r
F P
0 g
s
.5279 .27809 m .57631 .28732 L .57736 .26776 L p 0 .567 0 r
F P
0 g
s
.57631 .28732 m .62609 .28388 L .62395 .26449 L p 0 .641 0 r
F P
0 g
s
.61915 .24604 m .5785 .24886 L .57736 .26776 L p 0 .551 0 r
F P
0 g
s
.57736 .26776 m .62395 .26449 L .61915 .24604 L p 0 .551 0 r
F P
0 g
s
.53195 .25898 m .57736 .26776 L .5785 .24886 L p 0 .496 0 r
F P
0 g
s
.62395 .26449 m .57736 .26776 L .57631 .28732 L p 0 .641 0 r
F P
0 g
s
.25 Mabswid
.69808 0 m
.9756 .42021 L
s
.9756 .42021 m
1 .60007 L
s
1 .60007 m
.70987 .19617 L
s
.70987 .19617 m
.69808 0 L
s
.02912 .25753 m
0 .44491 L
s
0 .44491 m
.70987 .19617 L
s
.70987 .19617 m
.69808 0 L
s
.69808 0 m
.02912 .25753 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[16]:=",
  ImageSize->{287.938, 219.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004O0000f`P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/L70P1F1`00a@L200031`0005D700321`<0
0`L205D7002o1`<01PL00`07001D1`00_@L200T700@01`L0D`L00;X70`0;1`0400L705<7002g1`<0
3PL01@071`L00587002e1`804@L00`071`021`0300L704l7002b1`<04`L00`071`021`0300L704l7
002`1`805PL00`071`031`0300L704h7002]1`<06@L00`071`031`0300L704d7002Z1`<070L00`07
1`031`0300L704d7002X1`807`L00`071`041`0300L704`7002U1`<08@L00`071`051`0300L704/7
002R1`<090L00`071`051`0300L704/7002P1`809`L00`071`061`0300L704X7002M1`<0:@L00`07
1`071`0300L704T7002J1`<0;0L00`071`071`0300L704T7002H1`80;`L00`071`081`0300L704P7
002E1`<0<@L00`071`091`0300L704L7002C1`80=0L00`071`091`0300L704L7002@1`<0=PL00`07
1`0:1`0300L704H7002=1`<0>@L00`071`0:1`0300L704H7002;1`80?0L00`071`0;1`0300L704D7
00281`<0?PL00`071`0<1`0300L704@700251`<0@@L00`071`0<1`0300L704@700231`80A0L00`07
1`0=1`0300L704<700201`<0APL00`071`0>1`0300L70487001n1`80B@L00`071`0>1`0300L70487
001k1`<0C0L00`071`0>1`0300L70447001h1`<0C`L00`071`0?1`0300L70407001f1`80DPL00`07
1`0?1`0300L70407001c1`<0E0L00`071`0@1`0300L703l7001`1`<0E`L00`071`0A1`0300L703h7
001^1`80FPL00`071`0A1`0300L703h7001[1`<0G0L00`071`0B1`0300L703d7001X1`<0G`L00`07
1`0C1`0300L703`7001V1`80HPL00`071`0C1`0300L703`7001S1`<0I0L00`071`0D1`0300L703/7
001Q1`80I`L00`071`0E1`0300L703X7001N1`<0J@L00`071`0E1`0300L703X7001K1`<0K0L00`07
1`0F1`0300L703T7001I1`80K`L00`071`0G1`0300L703P7001F1`<0L@L00`071`0G1`0300L703P7
001C1`<0M0L00`071`0H1`0300L703L7001A1`80M`L00`071`0I1`0300L703H7001>1`<0N@L00`07
1`0I1`0300L703H7001<1`80O@L00`071`0I1`0300L703D700191`<0:0L804l700<01`L06PL00`07
1`0d1`00APL302L7401;1`0300L701X700<01`L0=0L004@70P0X1a0000A=0000B@L010071`L90147
00<01`L0<`L004470`0X1`D000==00001`001Dd0C@1=00@0A`L00`0700070003C@0000<03`L00`07
1`0b1`00?PL302X71`000dd0000300==00@0C@1=1`121`/014d600d700<01`L0<PL003`70P0/1`L0
00U=04d0C@1=04d00`000dd0C@0:03`71`02C@<01dd600`700<01`L0<@L003T70`0]1`L000E=04d0
C@0200E=0P000dd0C@02009=00@0C@1=0P0j1`P000==DTd00U800`1=C@03C@080580CE9=DP0<1`03
00L70307000g1`80;PL40003C@1=00<000==04d00P0024d004d0C@1=1@001dd0C@1=04d00P0e1``0
14d2DP0504e=C@000U801@1BCE9=00802`L00`071`0`1`00=0L302l7100014d0C@02C@8000==0000
0Td00`1=C@03C@800Td00`1=C@02C@8000==00000`0a1`D00Td70004CE9=00=B00<0C@0015801@1B
CE9=00802`L00`071`0_1`00<@L3034710000dd0C@0200A=00<0C@000dd0101=04d200A=00P0C@1=
04d0C@L0;@L70003C@0000D00dd00`1BDP02DP03059B00AB1`0;1`0300L702h7000_1`80<`L40003
C@1=00801Td00`1=C@02C@0404e=C@8014d0101=CDd20005C@00C@000dd402T72`02DP<000A=DTd0
15800`1BDP03DP8015T300X700<01`L0;PL002`70`0e1`D000A=00001Td200A=0P0014d0C@03C@04
04d0C@<014d0101=CDd6028730001dd004eBCE800P02C@07059BDP1I0003DP0905T0FE9IDP1B0080
2PL00`071`0]1`00:@L303L71`000e9=000200A=0P04C@0604d0DTd00dd00`1=C@02C@08059=DTeB
C@04C@0304d000@07@L70003C@00008014d00`1BDP02DP05F@00C@000e800`1B0002DP0505UIF@00
15T00`1I000:1`0300L702d7000W1`80>PL500A=00@0DTd00Td0101=C@04C@800dd02P1=C@1=04d0
DP05C@0304e=00==0`000dd0C@0201T72P02C@8000e=04d0DTeBC@1B04d000AB00<0FE800P001eUB
FE80DUT00`000eUB000:1`0300L702`7000T1`<0>`L4000:DP1BCE9=04eBC@<000A=DTd00dd01P1=
DTeB009=00h0CDd0DTeB059=DTeB00==00<0DP000P000dd0C@0301D72@05C@030580009=00<0DU80
0U800`1IDP0200=B00D0FE9I0005F@<000H2F@00F@0;1`0300L702/7000Q1`<0?PL300==00<0CDd0
0Td01P1=DTeB00==00@0CDe=0P03C@0404e=C@<000E=DTeB0005C@0304d000A=0P001Dd0C@1=00D0
2`L;00M=00@0DU9B0P001DeBCE8000AB0P02DP03059I008000ABFE9I0P04F@0305T000802PL00`07
1`0[1`007`L204470P003U9=DTd0CE9=DP1BCE9=0P00159=DP02C@8000EBCE9=0002C@8001A=DTeB
C@1=DTeBCE80DTeBCE9=DP<000M=0000C@1=01@000==DP0014d200EB00<0DU800U801`1BFE9I0580
0P000e9IDP02009I0P002@800UT2F@1I0P0200/700<01`L0:PL001`70`121`<014d60007CE9=DP1=
0003C@800Td01@1=CDd0009=00T0CE9=DTeBC@0014d0101B0008C@<000m=04d0C@1=04d0C@1=04d0
1P001TeBCE9=00==00<0DU8015801`1I059=DP000e800`1I00020007FE80F@1B0002F@0305UI009I
00<0FET00UT200/700<01`L0:@L001X70P151`8000M=04eB058000=B0`00259=DP1B059=0P001e9=
04eBC@000Td2000<CE9=DTeBC@1=DTd00e820009DTeBCE9=04eB00D000M=0000C@1=00H000a=DTeB
CE9=DTd0C@06DP0405UB009B00`0FE9IDP1I0580DP02F@8000eI0UT0F@9I0P02F@9I00802`L00`07
1`0Y1`005`L304H70`0014d0C@02DP03059B009B00H0DP00C@02C@0404e=00==01T0CDe=04d0C@1=
DTeBCE9=04d0CE9=DTeB00800dd02@1=DTeBCE9=DP09000=DTd0CE9=DTeBCE9=DP0200IB00P0FE9I
DP1=00=B00H0DUUB0582009I00<0FET00UT600=I00@0F@002`L00`071`0X1`0050L302P7200I1`<0
00==DP0015801@1BDU8000=B0P001E9=04eB008000I=DTd0CE8200=B00T0CE9=DTeBC@002E82000>
DP1BCE9=DTeBCE9=058300EB00H0CE9=DTd40003FE8000AB00H0FE9IDUT20005DUUBF@000eT01@1I
001I008000<2F@800P030P0905T000800P1I00`700<01`L09`L001870P0X1`<000<H00002@0D1`<0
0dd200AB1P03DP0504e=04d00`02C@0@04e=C@1BCE9=04eBCE9=DP<000UBCE9=DTeBCE800`00259=
DTeBCE9=0`002deBCE9=DTeB059=00<01U803@1BDU80FE9IDUUBF@000e80101IDUT20007DP0005T0
F@030009F@9I0P02F@8000=I0P0;1`0300L702L7000?1`<09`L400HH100011Q;01P301470`001E9=
058000=B00H0F@1BFE830007DP0004d0C@020007DTeB059=0005DP0604eBCE800e8200MB00<0DU80
2@05DPP001ABFE9IDUUBFE9I0580FE9IDUUBF@8000ABFE800UT01@02001I00<000EI008200030P04
009I0P8000=I0UT00P0<1`0300L702H7000<1`<0:0L500066000600H0`0014/HBaP601070P03C@0=
059B0580DUUB059IDP070003C@0000800dd03P1=001=DTeBC@1=DTd01E801`1=DTeBC@001E830003
DP0000H01U800`1BDP09DP8000YIDUUBFE9I05T00U80101BFE830004F@1I0P8000AI0P020`001`9I
0UT0F@000eT00`071`0;1`0300L702D7000:1`80:PL20004600H009;0P002QP0Bd/064/HBaP20003
600000803@L30003CE8000AB00<0FET00eT00`1BF@080003C@00008000MBC@1=058000EB00@0CE80
1U8200=B00<0DU800`06DP0E05UBFE9IDUUBFE9I05UBFE9IDUUB00<000UI059IDUUBF@000UT0101B
F@02F@0905T005T00UT000821@002`8005T0F@800UT2008030L00`071`0U1`0020L202/70P002QP0
601;64/HB`04B`0501Q;64/02P0:1`@000==05800e830004FE9IDP800eT;009=0P00259=04eBCE9=
0P09DPH02E800`1BDP0:DP03059B009B0`002E9IDUT0FE9IDP0200AI0P000eUB0002F@0405T00P80
00=I00000P801`1I0UT2F@0015T500d700<01`L090L000P70P0Z1`8000@H00001T/00`1;B`02B`05
04/HBaP00P02B`0501Q;64/00P091`D01E800`1BF@020004F@1I0Pl000=B00000e800`1BDP03DP80
2E8500UB00l0FE9IDUUBFE9IDUUB05800`07F@05059IDP001UT20004F@1I0P<000<200001P0015T2
F@8300070UT2F@1I0P0200h700<01`L08`L000P700<01`00:0L4009;0P001d/HBaP0B`000d/0100H
BaP200=;00<061P00QP300P70P0014eB0003DP06059IDUUB0`03F@<000<200L00PL2000AC@0004d0
DTeBC@1=DTeB05800P04DP<000=B05800U8600=B00<0DU801588000>DUUBFE9IDUUBFE80DP07F@04
0580F@P00UT30003F@1I00800UT00`1IF@03009I0P0>1`0300L702<700081`0400L702H70`04B`04
04`0009;00<0C0000d/01P1;601;00=;00D064/HB`0400H70`00159=05820008DP1IDUUBF@02F@@0
00@200801@L01@1=001=00801580201BDU80DUUB1004DP0:059IDUUBFE9IDPX000MIDUUBFE8000aI
0P06F@@000AI008220001UT2009I00EI0P02F@0305T200803`L00`071`0R1`001`L00`071`021`03
00L702870P000d/0B`02B`0504a;C4/00P0014a;C003B`8000A;C4/00T/00`0H600260H010L30003
CE8000=B10000e9I0004F@0405T2F@801`L00`1=00020009DP00CE9=04d0009B0`000e800003DP03
059B00]B01T0DUUBFE9IDUUBFE9I05UBFE9IDUUBFE9I00801UT00`02F@030009F@8000L0C@1B00H0
00=I05T00eT200=I00@0FEUI0P0?1`0300L7024700071`0300L700<700<01`L080L20003B`1;00=;
00/0Bda;C4/0Bda;0003B`0604]<Bd`00d/0101;64/2000364/H00@000h70000DTd0DUUBF@1IDP@0
0eT00`02F@0300X70`000e9=0003DP@000IIDUUBFE82000MDP1IDUUBFE9IDUUBFE9I05UBFE9IDUUB
FE9IDP0025T300030P1I00=I0P000`9I000200030P0000@700H0C@1BF@03F@0405T0F@8000=I05T0
0UT0101I0UT200l700<01`L08@L000L700<01`L010L2020700<0BaP01P001Da;C4]<008000A;C4/0
0T/01P1<Bda;00=;00X06000B`1;64/H1`000deB0004DP07059IDUT0F@0700030P0700P710000deB
C@0600UB100;DP0C059IDUUBFE9IDUUBFE80DUUBF@0300AI00T0FEUI05T2F@8010061`d000AIDUT0
15T00`1IF@02010700<01`L080L000L700<01`L01PL00`071`0L1`0304/H008014`20004C4]<B`80
00a<Bd`0B`1<04a;C003B`8000a;C4]<04/HBaQ;01P2000>CE9=059IDUT0FE9IDP03F@0405T2F@80
20L;009B00l0F@00DUUBFE9IDUT0FE8010005eUBFE9IDUT0FE9IDUUBFE9IDUUBF@1I00<000P2F@9I
0UT2009I00@0F@9I10081`<00e80101BFE8200AI00<0FET00UT00`02F@02014700<01`L07`L000L7
00<01`L01`L00`071`0K1`0;01P0B`1;C4]<04`00P0024/0C4`0C4]<0P001T`0Bda;00=;00<0Bd/0
0d/01`0HBaQ;64/01@000e80DP03DP06059IDUT00eT01002F@8200H72@0011P004d300=B0P06DP07
059IDUUBF@05009B00L0DUUBFE9I00T02UT30003F@8000<02PL01`1=059=DP0015801`1BFE9IDP00
0eT00`1IF@02014700<01`L07`L000L700<01`L020L00`071`0I1`<00d`01@1<C4`000=<0P02C005
04]<B`000T`01@1<Bd`0009;00`0C01;C4]<B`0HBaP20003B`00008H1000259I05UBFE800eT0101I
0UT200@72@0011Q;64/30007DP00DUUBF@03000?FE9I05UBFE9IDUUBFE9I00T01eT02`02F@9I0UT2
F@9I008000<2F@00100<1`800e801@1BFE9I008015T0101IFET6014700<01`L07PL000L700<01`L0
2@L00`071`0H1`8000U;C4]<04a;C00014`0201<C4`0Bd`00d`02@1<Bd`0B`1<0006B`0301P000E;
0P001AP0600H008000EBFE9I0003F@05009I00801P000aP0000200HH00<0Bd/01@06DP80011B059I
DUUBFE9IDUUBFE803ET00`1IF@05F@@000<200000`0>1`<000A=DTd00e820006FE9IDP1I0`02F@<0
4PL00`071`0M1`001PL00`071`0;1`0300L701L700<H04`00d`500A<0P02C00304a;00800T`0101;
C4/200=<00P0C4]<Bda;00Q;0P000aP0000260h000<H01P01P001AQ;64/H008014/60006FE9IDUUB
100035UBFE9IDUUBFE9I00eI00D00UT2F@0300<21@0@1`0304d000=B00H0FE9I05T60003DUT0009I
0`0B1`0300L701d700061`0300L700`700<01`L05@L20009C01<B`1<04`400<00T`00`1<0002C080
00m<04a<04a;C01<Bda;C00014/0101;0008B`@000<H01P01`001AP0600H00H000@HBaP00aP00`1;
B`05B`0304];00@01E800`1BF@040008DUUBFE9IDP0=F@0305T0008000AI0UT20P000`80000B1`h0
0U801@1BFE80009I0`0C1`0300L701`700061`0300L700d700<01`L050L2000<C01<04`400A<1004
0P000d`00002C0800T`01P1<C01<B`801D`0201<Bd`0C4]<0P06B`0304];00@00QP00`0H600260L0
24/00`0H0006B`8000A;04];1P000e9I0006F@D000=BF@001UT900@21@0E1`<00dd03P1BDU80FE80
DP1BFE800eT20003F@L7018700<01`L06`L000H700<01`L03PL00`071`0B1`0601P0C4/014`01`04
C01<14`00P001D`0C4`0009<00H0C4`0B`02C00604a;C4/00T/01`1<Bda;C4/00P03B`0304];00E;
2P0034/064/HBaQ;64/HB`801T/01@1<Bd`0009;00@0Bda;10003580FE9IDUUBFE9IDPT01eT0101I
0UT200030P0001L70`03C@0504eBC@000e820004FE9I00=I00<0F@0050L00`071`0K1`001PL00`07
1`0?1`0300L701470P03C08000A<14`410000d`4C0020003101<00800T`01@1<04`0009<00<0C4`0
0d`02`1;04]<Bda;C4]<008000=;04/024/00`1;B`0200E;00<0Bd/00d/40003C01;00A;00H0C4]<
Bd`20004Bda;009<1`09F@0?009I0UT2F@9I0UT2F@0200H06@L3009=00l0DU9B059IDUUB059IDP00
0UT00`02F@0201@700<01`L06PL000H700<01`L040L00`071`0@1`8000E;C01<0003C00300@00084
0P001@A<001<00@000E<04a<0004C00B04]<Bda;04]<Bda;C4]<Bda;0`08B`<01d/00`1;60030007
Bda;C4]<0002B`8000m<Bda;C4]<04a;C01<Bd`03000159IDP06F@X07PL2009=00D0DTeB0005DP80
00=BF@000UT00`1IF@0201D700<01`L06@L000D700<01`L04PL00`071`0?1`0:04a<C01<100400<4
00<010@01`000d`00002C00304`000E<00<0C4`00T`3000:C4]<Bda;C4]<009<2005B`P001A;C4]<
Bda;C4/0B`1;C4]<Bda;C08000=;C0000d`01P0404`0C1P05@L>009=00l0DU9B059IDUT0F@1IDP00
0UT00`02F@0201D700<01`L06@L000D700<01`L04`L200h70P0014a;C003C08000E<14`4000210X0
00=<04`00P05C00404]<00=<0P0034]<Bda;C01<Bda;C0/0015<Bda;C4/0Bda;C4]<Bda;C002000C
Bda;C4]<Bd`0C01<B`1;C4/0C0020007C004C01<100;00=<00H0C000B`0A1`D000M[Q6^409d000>4
0P02C@05059=DP0015800`1BF@0200=I00<0FET00P0F1`0300L701P700051`0300L701D700<01`L0
2`L5000614`4C01<0P0210<00P@90005C000C00014`0101<C005C08000E<Bda;0002C0<01D`00`1<
C009C00904a;C4]<Bda;008000Y<04a;C4]<Bd`00d`200=<00<0C0@00`0010@004`3008400<010@0
0`@01`04C0A<04`00P000d/01`0=1`H01H@0102MWId20005Q000C@000e802`1IDUUBF@1IDUT000=I
00<00P000P0G1`0300L701L700051`0300L701H700<01`L02PL20003C4/000A<00@014`410000`@0
000<000314`000A<0P08C08000=;00000P006da;C4]<Bda;04]<Bda;C4]<Bda;04]<Bda;C00200A<
00D0C4]<B`0200A<0P000da;0003C08000<4000020001d`4001<10000d`2009<0P0<1`@000=[Q6/0
0`000f^40006W@0305T000<000=BC@0015800`1IDP080003F@0001L700<01`L05`L000D700<01`L0
5`L00`071`091`800T`3000:1004C0A<1004C0L000A<000H100014`4C004C08024`500U<00<0C4`0
2T`0101;C4/200I<00D0Bda;0006C00704a<04`4C00:00@40P001PA<14`0C08000@404a<0P0;1`D0
0h@00`2MW@0200JM0P000`5I0P0<0008DP1B0580DUT301P700<01`L05PL000D700<01`L060L00`07
1`081`8000=<B`000T`3000?14`4C01<14`400A<04`000871@001@A<14`000=<00@0C00014`2009<
00<0C4`01P004T]<B`1;C4]<Bda;C4]<Bd`0C08024`0101<B`05C0<000A;04a<3`001`A<04`4C0@0
0P03C00304a<00802@L70009JhA[08BMQ9d000BM00D0W@400P0500A=00<0DU800U800`1BDP020004
FE8001X700<01`L05@L000D700<01`L06@L00`071`071`800T`01P1<C4`0108000Y<04`4C0A<04`4
0`041`036000008000A<14`00T`01004C0@400A<00<0C4`01T`B00]<0P05C00304`4008000A<1004
0`041`800P@00`041003100:04`000@014`4009<00<0B`001PL20004J`24W@<00h@00`2MW@03W@80
00NM09d1W@1I00D000E=DTeB0003DP0405UBF@8000AB05T06PL00`071`0E1`0010L00`071`0K1`03
00L700D70P0014]<B`03C00404a<C0@000@4C01<0`000d`01`051`8000U<04`4C01<04`00`000d`0
0003C00304a<00Y<00<0C4`02T`00`1<C009C08014`20006C0@0C01<0`071`@000A<04`41003C005
04a<C0000d`200D70P02J`0308B400:400RM08A[08BM00BM00L0W@6M0Id200@014d01@1BDU8000AB
0P0015T0F@0K1`0300L701@700041`0300L701`700<01`L010L200=<00`0C4a<00A<14`0C0@40004
C01<00T700H0C01<14`30006C0A<14`40P000d`0C00;C00304a<00Y<00<0C4`01T`30003101<009<
00D014`4C0020003100000T700@0B`1<10003@A<04`4C0@014`4C0000T`200@70P03J`0508BMQ000
0Yd300:M00<0WId00Yd01`01W@6M05T00`001DeBCE8000=B00P0FE9IDP1BF@8000=I00L06`L00`07
1`0C1`0010L00`071`0M1`0300L700<70P0014]<B`03C00304a<009<00<0C4`00T`01P1<C01<00X7
1@000d`00008C0@02T`00`1<C00;C00304a<009<0`04C00=04a<00A<14`404`0W@0200P700<06000
0T`00`1<C002C00304a<009<00<0C4`00T`00`1<C00200871002J`0308B400:400<0WId00Yd01024
W@03W@0709d00Id0W@0400A=00D0DU9B0004DP0305UB008000=I00L06`L00`071`0C1`0010L00`07
1`0N1`0300L700870P03C00D04a<C004C0A<04`4C0@014`0C0091`H000<400@00P002T`4C0A<14`0
C0@400I<00<0C4`02d`00`1<0002000714`4C0A<1003000414`4C0@000BM09f40P071`80015<1004
C0A<04`4C0@014`4C00:00=[00H0WHBMQ004W@0409fM00>M00L0W@2M0@0100<000EBCE9=0003DP09
059IDUT0FE9I008000=I00L070L00`071`0B1`0010L00`071`0O1`0400L71`<000=<B`000d`00`1<
C002C00304a<009<00H0C4`0C0061`<01@L4000414`000E<00L014`4C0A<00@00T`00`1<C004C0T0
2D`20004C0A<108000U<08BMQ9d0WH@00P061`<00T`01@1<C4`000A<00<0C4`00`02C0800f/01P2M
001[00>400<0WId00Yd0102M0007W@8000@1W@0014d01@1BDU8000AB00T0FE9IDP1I05T07@L00`07
1`0B1`0010L00`071`0P1`0300L0008000=<00000P02C00=00A<14`0C0A<1004C00200D70P0:1`<0
0`@2000>C0A<04`4C0A<14`4C0@9000?14`4C0A<1004C0A<14`400<000L400A<100400<000FMQ9f4
W@0200:M0`031`0<04`014`0C0A<04`41@02C0<000=[Q6/00P0019f4WH@20004WHBM00BM00@0W@00
0id01@01W@4000810`0014eBC@03DP0;059IDUUB059IDP000eT00`071`0L1`0300L7014700041`03
00L702471003C00304`000@00T`00`1<C002C00304a<00800PL300h70`000d`4C002000MC004C0A<
14`4C0A<14`0C0A<14`4C0A<14`4C00014`3000614`4C01<1003Q00309f4008000JM0024WH@20003
Q00000/00T`00`1<000400=[00X0Q9f4WHBM09d00X@00`2MW@02W@0409fM00>M00<0WId00P000id1
000200==00D0DU9B0004DP06059IDUT015T00`071`0L1`0300L7010700031`0300L702<70P000d/0
C002C00604`400A<3P0C1`<00`@5000I14`4C0A<14`400A<14`4C0A<14`4C0@0100300H41@003HBM
Q9f408@0Q9f4WH@00P04W@D000=<04`00T`0101<14`50003Q9f400<000F4WHBMQ00200>M0P04W@05
006M0@000Yd200040Id0008100<200002@002EUBF@1IDUUBF@0501h700<01`L040L000<700<01`L0
8`L20003B`1<009<00D0C4a<0003C00304a<009<1@0G1`<000=<10000`@6000:C0A<04`4C0A<10T0
00L4C0A<14`400@00h@00`24Q003Q00908BMQ9f4WHBM008000J408BMQ9d<00B40`02W@0508BMQ000
1Yd00`2MW@02W@0809fM0040W@02W@0609d0W@1I10001dd0C@1BCE802@001UT0F@0701h700<01`L0
3`L000<700<01`L08`L20003B`1<009<00/0C0A<04`4C0@01003000414`001P710000`@0100710X0
1`@00`04100210L01H@01P24WH@0Q08000RMQ9f4WHBMQ0<00164WHBMQ9d0WHBMQ9f4W@2MQ00200JM
00<0Q0001Yd00`010002W@0=09d1W@2M09d0W@400@0200030@0000<00dd00`1BDP02DP0905UBFE80
DUUB00<00PL00`071`0L1`0300L700h700031`0300L702<70P000d/0C002C00304a<009<00<0C4`0
0`03C00404`001P700@0F@1B1009100300@400X400<010001P001H@0Q6/000B400P0Q8B409f4W@80
00V4WHBMQ9d0WH@010001if408BMQ9d01P08W@801Yd0102MW@02W@0809fM09d1W@02W@06006M09d1
1@0024d004eBCE801E801@1IDP1I00800PL00`071`0L1`0300L700h700031`0300L7028700D0601;
0004C00400A<10<000T400A<14`0C0006PL3009B2003100300@400@4200010@00002J`0308@000>4
00<0Q8@00X@02@2M09f4WHBMQ0020004WHBM00NM1`04W@0309fM00FM0P000`40W@04W@0C006M0Id0
W@2M0@01W@400Id0W@0200030@0000<00Td00`1BDP03DP0;059IDUT0FE9I05T00P031`0300L701`7
00<01`L03@L000<700<01`L08PL01@0H04/000A<00<0C0000P04C00704a<C01<000J1`0605T0CE80
0e8E0003J`2400<00V/01`1[Q6^4J`0018@02P2MQ9f4WHBMQ9d2000EQ024WHBMQ9f4WH@0Q9f4WHBM
Q9d000BM0`04W@0509fMW@001Id200>M00<0W@400P001Id0W@2M008000d10000C@1=059=DTd000AB
00T0FE9I05T0F@0010L00`071`0L1`0300L700`700031`0300L7028700H0601;C002C0<000i<14`0
C0A<04`4C01<01X70`03DP0505UBF@0015T800050P0008@00P000f/0Q003009[00<0Q8@00h@00`24
Q002Q0<000V4WHBMQ9f4WH@00`08W@0309fM00NM00`0W@00W@6M0Id1W@02W@0?006M0Id1W@400Id0
W@6M008000<100400`000`400002009=00<0DU800U803P1BFE9I05UBFE80F@1I1PL00`071`0K1`03
00L700`700021`0300L702<710001T/0001<00A<00D0C4a<0004C0806PL40003DTd000AB00L0FE9I
DUT0009I00@0FEUI100018@00002Q0<000U[0024JhA[Q0000h@00`24Q002000:WHBMQ9f4W@2MQ0H0
00NMQ024WHBM00L02Yd200NM0P02W@05006M09d00P001@400@0100D000M=059=DTd000EB00D0DUUB
0002F@0305T700D700<01`L070L00`071`0;1`000PL00`071`0S1`0304`000D00T`03@1<14`400A<
1004C0@00P0K1`<00e801@1BFE8000EI00L0FEUI05T200D00X@01024Q8@20004J`2400F400l0Q02M
Q9f4W@2MQ9f4W@002Id700FM00/00Id1W@6M0Id1W@0200060Id1W@6M0P002@400@01W@2M0@0;0003
C@1B00=B00`0DUUBFE80DUUBF@02F@0305T700H700<01`L070L00`071`0:1`000PL00`071`0R1`<0
00I;04a<00@A01/70`00159=DP04DP07059IDUUB0002F@0405UIF@@000@708@00h@3009[00H0Q6^4
Jh@200J40P0048BMQ024WHBMQ9f4WHBMQ00:W@0309fM00FM0P000id0W@04W@0509d1W@400P001Id0
W@6M00L000B40P1=0P001DeBCE8000AB00H0DUUBF@02F@0305UI00P700<01`L06`L00`071`0:1`00
0PL00`071`0O1`<00PL05P1<00A<04`4C0@014`4C01<14`0C0@201/70`04DP0605UBFE8015T02@1I
FET00UT0W@0200<70P03Q08000B4Jh@01H@02@2MQ9f4WHBMQ00300bM00<0WId029d01@2M0Id100<0
01<1W@400Id1W@2M0Id1W@2M0@0100X000=I00000P03DP0<05UBFE80DUUBFE800UT00`02F@081`03
00L701`700<01`L02@L0008700<01`L070L300D700@0Bd`00T`03P1<14`400A<1004C0@00T`201/7
100014eBC@04DP0705UBFE9I0002F@0305T000<01PL200:400<0Jf/00V/01@24JhA[00<028@3000:
WHBMQ9f4WHBM00bM00@0W@001id0402M002M0Id1W@400@0109d50004Q6/0Q0<00Td500=B00H0FE9I
DP03F@0305UI00T700<01`L070L00`071`081`000PL00`071`0J1`8020L2000EC01<14`0C0A<04`4
C0@014`400@001`700AI000015801P1BFE9I00AI00<0FET00P0015T000081`03WH@0008000B4Jh@0
0h@00`24Q0020007WHBMQ9f40002W@H00id00`2MW@04W@P000VM0Id1W@6M0Id00P001P6M0Id1W@@0
00@100000PL50003Q000008000ABCE9=0e850004FE9I00AI0P0:1`0300L701/700<01`L020L00003
1`0701T70`0:1`8000=<B`000T`04004C0A<04`4C0@014`0C00L1`09DP0004eBCE8000AB00@0DUUB
10000eT0F@0300X71003J`0308@000F400<0Q8@00h@02@2MQ9f4WHBMQ00:00JM00<0WId01id20005
0Id1W@400P000id1000300D70`000f/0F@0300==00<0DU800e800eUBF@080003F@0700X700<01`L0
6`L00`071`071`0000<700L05PL300d70P02C00D00A<1004C0A<04`4C01<14`0C00K1`D015800`1I
DP05009I00@00UT2100<1`<000>4Jh@00P06Q08000>4W@0039d00`2MW@0:W@07006M0Id1W@03000:
W@2M0Id1006M0@<01`L300:400P0F@00DTeB00AB00L0DUUBFE8000=I00@0F@002`L00`071`0L1`03
00L700H700000`L01`0C1`<040L20004C4]<00=<00l014`4C01<14`0C0@0C0006`L20003C@0000L0
00=BF@001ET01P1IF@1I00h71002J`801h@3000=Q9f4WHBMQ9f4WHBM000<W@0309fM00:M0P04W@03
006M008000D1W@6M000:1`8000E[Q0020003C@03059B00=B00X0FE9IDP1BFE800UT200`700<01`L0
6`L00`071`061`0000<700L040L301<700<060000T`04@1<14`0C0A<14`0C0A<04`400806`L30007
DP1BFE9I0005F@0305UI009I00H0F@800P0<1`0405T2F@<000=[06/00P07Q0@02Id00`2MW@0:W@03
0040008000H1W@6M0Id30004W@6M0@802`L20003Q000008000A=DTd015801`1IDUUBF@000eT01002
F@0>1`0300L701X70P071`0000<700L03@L301D7100014]<B`03C00<04`4C0@014`4C01<0P0L1`80
00=BC@0015801P1IDUUB00AI00@0FEUI0P0:1`03008000T000A[Q0001H@01024WH@60004WHBM00JM
2008W@8000@1W@6M0`0<1`D000<204d00Td00`1BDP02DP0:059IDUT0FE9I009I00<0F@L03PL00`07
1`0F1`<000<700L01PL000031`0700X70`0H1`8000<H04`00T`04004C0A<04`4C0@014`400@201`7
0P02DP0605UBFE801ET00`1IF@02F@03009I00801`L@0003JhA[00800h@02P2MQ9f4WHBMQ9d:00JM
00L00Id1W@6M00<000FM09d1W@0600X70P0026/0001BCE8015801`1BFE9IDP000eT00`1I0P0200l7
00<01`L04PL400@700<01`L01@L000031`0700P70P0K1`<000HH04a;C003C00400A<10L000=<00L0
6`L30003CE8000=B00L0FE9IDUT000AI00D0FEUI00051`0305T000L00dd70003Q6^4008000n408BM
Q9f4WHBMQ9f4W@0039d00`2MW@02W@8019d00`0100030003CE8000802@L600==00<0DU800U802`1I
DUUB059IDUT0009I0P0@1`0300L700h70`081`0300L700D700000`071`051`<07@L20003100000/0
00<4C0000d`00`1<000L1`0304d0009B00H0DUUBF@04F@0305UI009I00@00UT20P021`l000EBCE9=
DP040003JhA[00D000Z4WHBMQ9f4WH@039d00`01000200080Id1W@400@04DPP01PL50004C@0000AB
00L0FE9IDUT000=I00@00UT20P0A1`0300L700X70`0<1`0300L700@700000`071`021`<08@L20003
C01<00=<00<0C4`00T`00`1<C002C00404a<C08070L30003CE8000AB00L0DUUBFE8000AI3`06C@<0
1Dd40003Q6/000>41P0019f4W@06W@P01id20005CE9=DTd01`02C@8010L00`40000200==00<0DU80
0U802`1BFE9I05UBFE80009I00<00P004@L00`071`061`@03`L00`071`041`0000<01`000P0T1`80
015<1004C0@014`4C01<14`40003C0<06`L00`1=0003DP0605UBFE8015T0101IFET20003F@9I00T0
24d00`1BDP030004CE80F@@000=[08@01h@:00JM00D00Id1W@040003DP0000@000MB001BCE9=00L0
0PL300==00`0DP00DP1BFE9IDP03F@0605T2F@024`L00`071`021`<04`L00`071`041`000P0W1`05
04/0C00014`00`1<C002C00304a<009<00D0C4`0B`0201X710000e9=0004DP0605UBFE9I0`02F@03
05UI00l00dd00`1BDP02DP8000E=05UBF@0500V400h0WHBMQ9f4WHBMQ9d0W@T01U800`1=C@06C@T0
14d0101=DTd40008FE80DUUBF@03F@03008701<7100F1`0300L700@700000`L01`0V1`0D04/014`0
C0A<1004C0@014`4C003C08000<7000060L400=B00D0DUUB0004F@8015T03`02F@80F@1=059=DTeB
0003DPD01E8@00>400<0WId00id:0003C01<00801U8D00==00<0DU800U80101IDUT400=I00@2F@9I
0P0A1`D060L00`071`041`000PL202H70P02C00304a<009<00<0C4`00T`00`1<C002C00304a<0080
0`L00`071`0E1`@000ABCE801580101B0002F@0305UI009I00@0FET00U830006DTeBCE801U8300AB
00<0FET015T00`1IF@0E0007C01<C01;0003DP@000eB059=DTeBCE80C@1=00H00dd01`1=DTeBC@00
15800`1IDP0700030P0700d70`041`0300L701H700<01`L010L000@700<01`L08`L0501;04`400A<
14`0C0A<1004C0@00d`00`071`021`0300L701@71@03DP0305T000<015T00`1IF@02F@0305T20080
00=B05802@05DP0405T0009B00<0FET015T02@02F@9I05T0B`08009<00<0C4`00T`0201<C4`0C1QB
0P00459IDUT0DTeBCE9=DP1=04d30007C@00C@1=0005DP07059IDUUB0004F@04009I0P8030L300P7
00<01`L05@L00`071`041`001@L00`071`0R1`0304/0009<00<0C4`00T`00`1<C002C00304a<009<
00@0C4`01@L00`071`0@1`H000==0000100015UBF@05F@0305UI009I00@0F@1I0P02DP08059IDUUB
058200=B00@0FE9I0P000e80F@04F@0305UI00@000PH0000Bda;00=<00<0C4`00d`01@1<C01<009;
1e800`1=DP0?000:C@1=059=DTeB00AB00H0DUUBF@03F@03008000P7100<1`0300L701@700<01`L0
10L000H700<01`L08@L05P1;04`4C01<14`400A<14`0C0A<04`200H700<01`L03@L80006DP1IDUT0
15T00`1IF@03F@0305UI009I0P03F@8000UIDP1BFE9IDP000UT30003DP1I00=I2P020P04009;609<
1003C00304a<009<00D0C4a<0002C00605UBFE9I1@001E9=DTd000A=00<0C@000P03C@03059B00=B
00L0FE9IDUT000AI00@0F@9I0P061`<040L00`071`0E1`0300L700<700071`0300L702070P001D/0
C4`000E<00D0C4a<0002C0<00T/00`071`041`0300L700X72`000dd0DP02DP0605UBFE801ET00`1I
F@02F@0305UI009I0P003E80FE9IDUT0FE9IDUT00P003E9IDUUB05T2F@9I00801@001QP0Bda;009<
1@03C00504a<C4/00`03DP8000UI059=DTeBC@0014d500==00P0DP1BCE9=00AB00H0FE9IDP03F@03
008000<70`0D1`0300L701@700<01`L00`L000P70P0P1`8000e;00A<1004C0A<04`400@00T`0101<
B`071`0300L700L71P001D/HB`01008000I=059IDP04F@0305UI00=I00L0FEUI008000=I00<0F@00
100015UBF@05F@<00eT01`02F@9I00800P031`0304/000=<00<0C4`01D`80004C000F@8000]BFE80
CE9=DTeB0005C@8000==04d00Td0101BDP02DP06059IDUT01ET01002F@8601P700<01`L04`L00`07
1`031`002PL00`071`0M1`<000E;04a<0003C0@00T`0201;C4]<04/020L00`071`041`0300@000=<
0P001AQ;602M00<000==05800U801P1BFE9I00EI00@0FET00UT0101IF@03F@0305UI00D000EIDUUB
0002F@8000AB05T210051`0601P0Bd`014`00`1<C003C00604a<C01<0P003U80FE9IDP1=DTeBCE80
0Td400==00/0CDd0CE9=DP1B0004DP06059IDUT00eT00`02000M1`0300L7018700<01`L00`L000/7
00<01`L070L20003100000H000=<100014`01`1<C4`0B`002@L00`071`021`8000A<14`41`002P40
001B05UBF@04F@0305UI00=I00L0F@1I0UT2008015T00`1IF@03F@030080008000II009I0UT40003
0PL700H70P03C00304a<009<00<0C4`00T`01`1<14`0C0@00P002U9IDUUBF@1BCE860008CE80DTeB
CE8200IB00H0FE9IDP05F@0405T2F@807PL00`071`0B1`0300L700<7000<1`0300L701/70P02C005
04a<C0001D`0301<C4`0C4]<B`1;00T700<01`0030000d/000040003C@1B009B00H0FE9IDP04F@80
1UT200=I00<0FET00eT00`1IF@02F@8015T00`020P091`0601P0C4/014`00`1<C003C00504a<C000
0T`20005FE9IDUT01@002E9=DP1BCE9=0005C@8000MBCE9=DP1I00800U801P1IDUUB00=I00<00P00
7`L00`071`0A1`0300L700<7000=1`8070L2000<C0@014`400A<14`014`01`1<C4`0C00020LC000:
W@0004d0DUUB00AI00<0FET00P002P9I05T2F@9I05T20003F@1I00AI00H00UT2F@8300040UT000X7
0P03C00304a<009<00<0C4`00T`02@04C0A<04`0C0020006F@0005T01U802P1=DTeB059=DTd200EB
00D0FE9IDP0300AI00@00UT20`0P1`0300L7010700<01`L00`L000l700<01`L06@L2009<00D0C4a<
0005C00904a<C01;C4]<00801PL:00<H20000`4000020003C@1B009B00@0DUUB0`05F@0505UIF@00
15T400=I00<0FET00eT00`020P0600P70P000d]<0004C00304a<00=<00D0C4a<0002C0D0019IDUUB
04eBCE9=04eBCE9=04d3000=C@1=DTeB05UBFE9I0002F@8000=IDUT00eT3020700<01`L04@L00`07
1`021`0040L00`071`0H1`8000e<1004C0A<04`4C0@000A<00L0C4a<01P000<73`001QQ;64/HB`80
00>M000010001EUBF@1I008001EI0P02F@9I0UT0F@800UT2F@800UT01P000`9I00020PP01PL01P0H
B`1<00A<00<0C4`00T`0201<14`400A<0P021`0305UI009I00<0DU800e800`1BC@0500EB00X0DU80
FE9IDUT01ET700030@07020700<01`L040L00`071`021`004@L00`071`0G1`800d`01@1<C4`000A<
00<0C4`00T`00`1<B`020003B`0000X01QP400<H0`000id0000800EI00<0FET015T200II00T0F@9I
0UT0F@800`000da;0005008H0P041`8000=<B`0014`00`1<C003C00504a<C0000T`00`1;00031`d0
01==DTd0CE9=DTeB0580FE9IDUT000II00H0F@9I09d3028700<01`L03`L00`071`021`004PL00`07
1`0F1`0@04`014`0C0A<1004C0A<00A<0P0014]<04/90009BaQ;64/HBaP000=;1@000f/0000600=I
00H00UT2F@820005F@9I0UT00P00309I0UT0F@9I0UT200<21P04B`P000<H04/00T`2009<00<0C4`0
0T`02`04C0A<04`404`000D70P000e80DP03DP03059B00AB00<0DU800e820007FE9IDUUB0005F@04
05T2F@<000>M00008`L00`071`0>1`0300L70087000C1`805PL00`1;0003C00504a<C00014`0101<
C4`20005Bda;01P03P04600304];00=;0`001V^4Q9f4W@<000AIDUT01UT01`1IFET0F@800`002ET0
0UT2F@800P030004C4]<00M;20000d]<0003C0<014`00`1<C002C00604a<04/01`L200AB00/0CE9=
DTeB059=DP02009B00H0FE9IDP05F@0405T2F@8000>M09d00P0S1`0300L700h700<01`L00PL001D7
00<01`L050L2000=14`400A<14`0C0A<1003009<00@0Bda;100021P0BaQ;64/00d/700A;0P0028A[
08BMQ02M0`02F@8000QI0UT2F@1I00H22@04C18000=<B`001D`03@@004`0C0A<1004C0@00`091`<0
0U800`1BDP04DP<000EIDUUBF@020003DP1I00AI00@0F@9I0P001Yd00@0102@700<01`L03@L00`07
1`021`005PL00`071`0C1`800d`00`1<C002C00304a<00<014`0101<Bd`2000;6000601;64/HB`00
1d/0101<0002B`80011[Q6/0Q9f408@0C@1BFE8015T300070UT2F@9I00040P8000=;C4/00`000d`0
B`05B`0504/HBaP00P001QP0C4]<00=<00<0C4`00T`:0003C00700X73008DP801ET0101I0UT200:M
00@0W@019PL00`071`0<1`0300L70087000G1`0300L7018700T0B`1<14`0C0@0100014`4C004C008
04a;C01;04/:00E;00@0Bda;1@001f^409f4W@000Yd300=I00@0F@000`83008200<00P000`000`@0
C002000?C0A<04]<Bda;C01;64/H00D000=;C00014`03P1<14`400A<14`400A<0`0=1`8000eI05UB
FE9IDP1BFE9I00<00UT2009I00@0F@9I0P02W@0509d100409PL00`071`0<1`0300L70087000H1`80
4`L8000414`400=<00<0C4`00d`0101<04`2000464/H00E;00<0B`000`02B`0504]<Bd`00`002hA[
Q02MQ9d0WHBM00H000QI0UT2F@9I0PX01T`00`1;B`09000864/004/0B`03C00304a<00=<00<0C4`0
0T`01@04C01<00<020L400870`05DP03059B00801eT700:M00@0WId00P4W1`0300L700`700<01`L0
0@L001X700<01`L040L04@1<C0A<04`4C0A<04`4C0@000A<00D0B`1;C00200=;00/0Bda;C4/0Bda;
C0040003B`0H00<000U[Q6/0Q9f4W@0019d40004F@0000820P000`80000200<400H0C0A<14`500=<
00L0Bda;C01;008000A;C00014`03P04C0A<04`4C0A<04`410000id000031`<020L900II0`001P80
002M00>M00@00Id00P4X1`0300L700/700<01`L00@L001/700<01`L040L2000:C0A<04`4C0A<00=<
00<0C4`00d`01@1;C4]<00<000=;04/014/01`1<Bda;C4/00`000d]<0003000>JhA[08BMQ024WHBM
Q002W@P000U<1004C0A<10000d`20008C01;C4]<B`04B`0904/004]<Bd`000=<00<0C4`00d`00`1<
C002C00404`4C08000>MQ0001@0>1`L00eT200080UT009f4W@03W@0509fMW@000P4Y1`0300L700X7
00<01`L00@L001`700<01`L03`L04@1<04`400A<14`400A<14`000=<00T0C01;C4/0Bd`01@002Ta;
C01<Bda;C002C0H000V4Jh@0WHBMQ00019d00`01W@0300D70P000`@01003100804`000A<1005C005
04a;C4/00P0024`0C4]<04`00T`0401<14`400A<14`400A<04`400:M0`0C1`/000B4WH@00id0102M
0Id30BT700<01`L02PL00`071`011`007@L00`071`0?1`8000X4C0@014`4C0@014`0101<C002C080
00U<04a;C4/0Bd`01`001da;C01<Bd`0100036^4Jh@0WHBM09f4W@@000BM00401PL80007C0A<101<
B`090007Bda;04]<0003C0801D`00`1<C002C00400A<108000B408BM0`0019d0000B1`<000=[Q6/0
18@>02X700<01`L02@L00`071`011`007PL200l701<0C004C01<14`4C01<14`404`000E<0P001D]<
Bd`000E<00@0C4a<1003C0D000I[Q6/0Q9d500030Id000<10P081`800P@01@1<14`400@000A<14`0
14`01@1<Bda;00801D`04004C0A<04`4C0A<04`400@200JM1@0A1`<000V4Jh@0WHBMQ00019d0202M
WId00Id0:`L00`071`081`0300L70047000P1`0300L700d700`0C004C01<14`4C003C0801d`2000@
Bda;04]<Bda;04]<Bda;C0<000=<00002@02W@0309fM00>M00<0W@400P091`060@00C01<0P000`@0
C004C00;04a;C4/0Bda;C0001D`0101<0002C00304a<009<00H0C0A<04`?00l70P001HA[Q6/000B4
00D0WHBM0003W@04006M02`700<01`L020L00`071`011`008@L00`071`0<1`0=04`0C0@014`4C0@0
10020003C0@000I<00H0C000C005C00304a<00A<00<0C0@00`02Q00608BMQ9d019d01P2M0Id10081
0`061`P00`@01P1<14`400E<00<0C4/00`04C00500A<14`00`0024`4C0@014`40P002XBMQ9f4WHBM
Q002W@D030L30009JhA[08BMQ9d000BM00P0WIfM09d102d700<01`L020L100470@L0028700<01`L0
30L02P1<04`400A<14`2000614`4C0@014`00`1<C002C0X000A;C4/00T`00`071`020008Q6/0Q9f4
W@03W@0309fM00:M00H0W@400@041`8000>4Jh@01@03100304a<009<00H0Bda;C4/300=<00D0C4a<
0004C00300A<0080015<00A<14`0Q9f4WHBMQ9f40006W@<02`L20005JhA[Q00018@01@24WH@000BM
00@00Id0;PL00`071`071`400@L11`008`L00`071`0<1`8000D4C01<1002008400H014`4C0@2009<
00<0C4`00d`00`1<C003C00300A<00D010L200:400H0WHBMQ004W@06006M0Id00P4200870P0016/0
0004Q1401D`02@1<C01<14`4000511`020L0304008A[Q02MQ9f400BM00L0WIfM006M0080;`L00`07
1`061`400@L11`0090L200d71P0030A<14`400A<14`4C0801T`00`1<C003C00304a<009<00<01`L0
10L20009Q6/0Q9f4WH@000>M00<0WId00Yd01@01W@0101<000A<14`01D`00`1<C004C08014`02`04
C0A<14`0C0A<008000nMQ9f4WHBMQ024WHBMQ9d010000hBMQ00200H70P001H@006/000B400D0WHBM
0004W@0309d10080;`L00`071`061`400@L11`009PL00`071`0;1`800P@00`041004100700A<14`4
C00300=<00<0C4`00d`00`1<100300L70`02Q00608BMQ9d00id02@01W@6M00400@040003WHBM0080
1id0102M0004100604`4C0@01D`20006C0A<14`01@@0100410@30009Q9f4WHBMQ02M00P000>MQ000
0`0019d000031`0;W@1[Q6/0Q9f4W@0019d02P2MWId0W@400@0`1`0300L700D70@011`47000W1`03
00L700`700<010@00P002@A<14`400A<10020003C0A<00`000<41`L01`L30008Jh@0WHBMQ004W@03
09fM00:M0P000`400@0<0004Q9f400:M0`000`@0C004C00304`0008000<4C0000d`02P1<14`4C01<
14`30003Q9f400L000j4WHBMQ024WHBMQ9f4W@T000B4J`240P02Q00508BMQ00019d01@01W@010387
00<01`L010L100470@L002P700<01`L030L2008400<010@00`@02`0400A<14`4C0@000D400<014`0
2`L400:400H0WHBMQ003W@0409d1W@800`43000:0@01W@6M0@01W@H000<100002@002@A<14`4C01<
0005100400@410L000Z4WHBM09f4WHBM4`001hA[Q02MQ9d00P04W@0309fM00:M00D0W@400@0b1`03
00L700@70@011`47000Y1`803@L2008400H014`4C0@2000>C0A<14`4C01<14`4C00>1`<000R4J`24
WHBM00BM0`001Id0W@6M008000@1000019d00`2MW@04W@0D004009d0C0@014`4C01<14`4C0@20009
C0A<14`0C0A<00800X@00`24Q004Q00409f4W@<000>408@01H@00`24Q002Q00306/000<000=[Q000
18@01P24002M00BM00@0W@6M0P0c1`0300L700@70@011`00:`L00`071`0;1`800P@2000310000084
10000`A<000210@03`L300:400P0WHBMQ9d0W@<000D1W@6M00030@0400400@@000X1W@2M0Id1W@40
0P4500@400D014`4C00200840P02100400@410800h@02@2MQ9f4WH@0Q002000;WHBM09f4WHBMQ000
18@6000<JhA[08BMQ9d0WHBM2P000`401`0c1`0300L700<70@011`00;0L00`071`0<1`D020@501<7
10001F^409f400@00id02001W@400Id10`001@6M09d100L00Yd01P2M0Id1W@D000A<10040P0610P0
0X@01024Q8@500J400<0Q8@02P0026^4001[06/018@01P2MQ9f400FM00<0W@400P0f1`0300L70087
0@011`00;@L00`071`0?1`T06@L800:M00<0WId00Yd0202M0Id100400P42008100T0W@6M0@01W@40
10050@03004100D01P@90003J`0000@00h@01P2MQ9f4W@D00h@00`24Q002Q00306^4009[00l0Q024
09f4WH@0Q9f4W@0019d01P2M0@0103H700<01`L00PL10047000^1`0300L703070`03Q00608BMQ9d0
19d06@2M0Id0W@2M0Id0W@400Id1W@2M0Id1W@400`00206M0@01W@6M1@0210@01`L20009J`24JhA[
Q6/000:40`000h@0Q004Q00308B400>400<0Q6/00P0018A[Q004Q00608BMQ9d01Id00`01W@0203P7
00D01`L700011`00;`L203470`02Q00309fM00:M00<0WId00Yd01P01W@6M00@10`003@400Id1W@2M
0Id1W@000P4200030@0100@010L300`70`04Q00408@000B400<0Q8@00h@00`24Q002Q00308@00080
00n4Jh@0Q02MQ024WHBMQ00019d01P01W@0103T700@01`L00@L0034700<01`L0;PL300>400H0WHBM
Q004W@0<006M09d0W@6M09d12@00206M0@01W@6M0`0I1`L01H@00`24Q003Q00308B400D000J4J`1[
Q003Q00808@0WHBMQ005W@0309d10080>PL010071`011`00<PL00`071`0^1`0306/000:400<0WId0
0Yd00`2MW@02W@0909d0W@6M004000@100<00@400P400`010@0500<100<01`L06PL20003J`0000F4
2P03Q00808B4Q024J`04Q00709f4WHBM0004W@0609d10040>`L00`0700011`00<`L00`071`0]1`<0
0h@01P24WHBM00BM00H0W@6M0Id2000A0Id1W@2M0Id1006M0Id100400`0O1`@001J40000J`1[Q6^4
J`1[Q6^4J`1[Q6^40P04Q00709d0WHBM0005W@03006M0080?@L100470@0003@70P0^1`<00X@00`2M
W@02W@0309fM008000NM09d1W@6M00<00P400`010@030@030041008100<00@408`L30005JhA[Q000
1H@00`24Q003Q00408B4Q08018@0102MQ9d20003Q02M00>M00H00Id00@0n1`80000f1`0300L702/7
0`03Q00509f4WH@00P04W@0809d1W@400Id:00040Id0WBL70P004F^4J`1[Q6^4J`1[Q6^4J`1[0080
00>408@00h@01P24WHBMQ0<00id00`2M0@0203l70P0003L700<01`L0:`L00`1[0002Q00309fM00<0
00X1W@6M0Id0W@401@400`010@030@0300400080:PLA00B400X0Q8@0Q9f4WH@00id704070P0003P7
00<01`L0:`L70004WHBM00JM0180W@2M0Id1W@2M0Id1006M0@0^1`8016/06`1[Q6^4J`1[Q6^4J`24
08BMQ9f408BMQ9f40002W@80?@L40087000i1`0300L702/70P03Q00309fM00:M00L00Id1W@6M00<0
0`400`010@030@@0<0L01P1[Q6^400F400<0Q8@00h@20006Q9f4WH@01Id00`2MW@0203X710061`00
>PL202`70P0026^4J`24WHBM0P03W@0409d1W@X0=0L3009[00P0Q6^4Jh@0Q0<00X@02`24WH@0Q9f4
WH@000:M0P0g1`@02PL003`700<01`L0:`L200:400<0WId00id03`2M09d1W@6M0@01W@6M000g1``0
1X@00`240005W@0509fMW@00=0L400h7000m1`0300L702/70P001hA[08BMQ9d00`002@6M0Id1W@2M
0@0203X71004J`0308B400F40P001XBMQ9f400>M00<01`L0<0L30187000n1`0300L702/70P02Q004
09d000>M0`000`6M000303h70P001XA[Q6^400B40P02Q0800Yd01@2MWId002l7100E1`00?`L202d7
100028BMQ9f4WHBM0`131`<00V/00`24Q00200J41P0/1`@06@L0044700<01`L0;@L904P71P04Q0L0
:`L401d700121`0300L702`700<01`L0D`L802/7100Q1`00@`L00`071`0[1`0300L708<70`0U1`00
A0L00`071`0Y1`0300L70807100X1`00A@L202T700<01`L0O0L402`700171`0300L702H700<01`L0
N0L4030700181`0300L702D700<01`L0M0L403@700191`0300L702@700<01`L0L@L303P7001:1`80
90L00`071`1]1`@0>`L004`700<01`L08@L00`071`1Y1`@0?`L004d700<01`L080L00`071`1U1`@0
@`L004h700<01`L07`L00`071`1Q1`@0A`L004l700<01`L07PL00`071`1N1`<0B`L005070P0N1`03
00L705X7101>1`00DPL00`071`0K1`0300L705H7101B1`00D`L00`071`0J1`0300L70587101F1`00
E0L00`071`0I1`0300L704h7101J1`00E@L00`071`0H1`0300L704X7101N1`00EPL201P700<01`L0
A`L30687001H1`0300L701D700<01`L0@`L406D7001I1`0300L701@700<01`L0?`L406T7001J1`03
00L701<700<01`L0>`L406d7001K1`804`L00`071`0g1`@0L@L005d700<01`L040L00`071`0d1`<0
M@L005h700<01`L03`L00`071`0`1`@0N0L005l700<01`L03PL00`071`0/1`@0O0L0060700<01`L0
3@L00`071`0X1`@0P0L006470P0<1`0300L702D710241`00H`L00`071`091`0300L702870`281`00
I0L00`071`081`0300L701h7102;1`00I@L00`071`071`0300L701X7102?1`00IPL200L700<01`L0
5PL409<7001X1`0300L700@700<01`L04PL409L7001Y1`0300L700<700<01`L03`L309/7001Z1`03
00L7008700<01`L02`L409h7001[1`0500L71`002@L40:87001/1`8000<700L010L40:H7001^1`80
00<700000P2Z1`00K`L20:h70000\
\>"],
  ImageRangeCache->{{{0, 286.938}, {218.875, 0}} -> {-0.000640078, 0.0870764, 
  0.00333809, 0.00333809}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[BoxData[
    \(MVClear[]; MVShow3D[threerings]; \)], "Input",
  CellLabel->"In[17]:=",
  CellTags->"MathLinkMode"],

Cell["\<\
Now the view point as well as the light positions changed to give \
the following image:\
\>", "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[18]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000312931
312931313131312931313131000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000292929313131312929313131312929313131312929313131312929
313131312929313131312929313131312929313131000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313131312931313131312931313131313131313131312931313131313131313131
313131313131313131313131313131313131313131313131313131313131313131000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292929
313131312929313131292929313131312929313131292929313131313131313131313129
313931313131313131313129393931313131313131313129393131313131313131312929
393131312931313131000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000312929313131313129393931313129313931313129
393931313129313931313129393931313129313931000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000313131312929313131
313131313131312931313131313131313131313131313131313131313131313131313131
313131313931313131313131313131313931313131313131313131313131313131313131
313131313131313131313131313131313131000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000312931313131313131313931313131313931313931313931313931313931
313931314231313931314231313931314231313929313931313931000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000292929313131292929313131312929
313131312929313131313129313131313129313931313129313931313129313931313131
313931313129313931313131313931313129313931313131313931313129393931313131
313931313129393131313131313131312929393131312931000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000313129
393131313129393131313129393931313129394231313929394231313929394231313929
394231314229394A31314229394A31314229394A31314229314231313929394231313129
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000312931313131312931313131313131313131
313131313131313131313131313131313931313131313931313131313931313931313931
313931313931313931313931313931313931313931313931313931313931313931313931
313131313931313131313131313131313131313131313131313131313131313131000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000313131313131313131313931
313131313931313931313931313931313931313931314231314231314231314231314231
314231314A31314A31314A31314A31314A31314A31314A31314A31314231314231314231
313929313131000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000292929313131292929313131313129313131293129
313931313129313931293129313931313131313931313129313931313931313931313929
394231313931314231313929394231313931313931313929394231313931313931313929
393931313131313931313129393931313131313931313129393131313131313131312929
393131312931000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000313131313129393931313131393931313129
394231313931394231313929394231313931394231314229394A31314A31314A31314229
394A31314A31394A31314A29395231315229315231314A29395231314A29314A31314229
394A31313929313931000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000312931313131312929313131313131313131313131313131
313131313931313931313931313931313931313931313931313931313931313931313931
313931314231313931314231313931314231313931313931313931314231313931313931
313931313931313931313931313931313931313931313931313131313931313131313131
313131313131313131313131313131000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000313131313131313131313931313931313931313931
314231314231314231314231314231394231314231314A31314A31315231314A31314A31
314A31315231314A31315231315231315231315231315A31315231315231314A31314A31
314A31314A31314229313931000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000292929313131292929313131293129313931293129313931313129
313931313929313931313929314231313929314231313931314231313929314231313931
314231313931394231313931314231313929314231313931314231313931394231313931
314231313931394231313931394231313931394231313931393931313131393931313131
393931313131393931313131393131313131393131000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
393131313131393131313131393931313129393931313931393931313931394231314231
394A31314229394A31314231394A31314229394A31314A31395231314A29395231315229
395231315229395231315229395A31315229395A31315A29395A31315229395231315229
315231315229315231314229314231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000312931313131313131313131313131313931313131313931313931313931
313931313931313931313931313931314231313931314231313931314231314231314231
314231314231314231314231314231314231314231314231314231314231314231314231
314231314231314231314231314231314231314231314231313931314231313931313931
313931313931313931313931313131313131313131313131313131313131313131000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000313131313131313131
313131313131313131393931313931313931313931314231314231314231314231394A31
314A31314A31314A31314A31314A31314A31314A31315231315231315231315231315A31
315231315A31315231315231315231315A31315A31315A31315A31315A31315A31315A31
315A31315A31315231315231314A31314231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000292929313131292929313931313129313931293129313931313929313931293929
314231313931314231313929314231313931314231313929394231314231314231313929
394231314231314231314229394A31314231314231314229394231314231314231314229
394A31314231314A31314231394A31314231314231314231394A31314231394231313931
394231313931394231313931394231313931393931313131393931313131393931313131
393131313131000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000393131313129393931313131393931313129
393931313931394231313929394231314231394A31314229394A31314231394A31314229
395231314A31395231314A29395231314A31395231315229395A31315231395A31315A29
396331315A31395A31315229395A31315A31316331315A29396331315A29316331315A29
396331315A29315A31315229395231314A29314231000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313131313129313131313131313931313931313931313931313931313931313931
313931314231313931314231314231314231314231314231314231314231314231314231
314231314231314231314231314231314A31314231314231314231314231314231314231
314231314A31314231314A31314A31314A31314231314A31314231314A31314231314231
314231314231314231314231314231394231313931313931313931313931313931313931
313931393931313131313131313131393131000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000313131313131313131393931313131313931313931393931313931313931
313931394231314231314231314A31394A31314A31314A31314A31395231314A31315231
315231395231315231315231315231395231315231315A31315A31395A31315A31315A31
316331316331315A31315A31315A31316331315A31316331316331316331316331316331
315A31315A31315A29315A31315231315231314A29000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
313131292929313931293129313931293929314231313929314231293929314231313929
314231313929314231313931314231313929314231314231314231314231314A31314231
314A31314231314A31314231314A31314231394A31314231314A31314231394A31314231
394A31314231394A31314A31394A31314231394A31314A31394A31314231394A31314231
394A31314231394A31314231394A31314231394A31314231394231313931394231313931
394231313931394231313931393931313131393931313131393931313131393931313131
393131000000000000000000000000000000393131313131393131313131393131313131
393931313131393931313931394231313931394231313931394231313931394231314231
394A31314231394A31314A31395231314A31395231315231395231315231395A31315231
395A31315231395A31315A31395A31315229395A31315A31396331315A29396331315A31
396331316329396331315A29396331316329396331316329396B31316329396B31316329
316331315A29396331315A29315A31315229315231314229000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000313131
312931313131313931313931313931314231314231314231313931314231314231314231
314231314231314231314231314231314231314231314231314231314231314231314A31
314231314A31314231314A31314231314A31314A31314A31314A31314A31314A31314A31
314A31314A31314A31314A31314A31314A31314A31314A31314A31314A31314A31314A31
314A31314A31314A31314A31314A31314A31314A31314A31314231314A31314231314231
314231314231314231314231314231314231314231314231313931313931313931393931
313931313931313931313931313931313931313931393931313931313931313931393931
313931314231314231394231314231314231314231394A31314231314A31314A31394A31
314A31315231315231395231315231315A31315231395A31315231315A31315A31395A31
315A31315A31315A31396331315A31315A31315A31316331316331316331316331316331
316331316B31316331316331316331316331316331316B31316331316B31316331316331
316331316331316331316331315A31315A31315A31315231313929000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000313131292929
313131313129313931293929314231314229314231293929314231314231314231294229
314231314231314231294229314A31314231314231314229314A31314231314A31314229
394A31314231314A31314231394A31314231314A31314229394A31314A31314A31314A31
395231314A31315231314A31395239314A31395231314A31395231314A31395231314A31
395239314A31395231314A31395239314A31395231314A31395239314A31394A31314A31
394A39314231394A31314231394A39314231394A31314231394A39314231394231314231
394A39314231394231314231394239314231394231314231394A39314231394A31314231
394A39314231394A31314231394A39314A31394A31314A31395231314A31395231315231
395A39315231395A31315231395A31315A31395A31315A29396331315A31396331315A31
396331316331396331316329396331315A31396331316329396B31316331396331316329
396B31316331396B31316329396B31316331316B31316329396B31316B31316B31316329
396B31316329316B31316329396B31316329316331315229314A31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000292929313131
313131313931313931314231314231314231314231314231314231314231314231314231
314231314231314231314231314231314231314231314231314231314A31314231314231
314A31314A31314A31314A31314A31314A31314A31314A31314A31314A31314A31314A31
314A31315231314A31315231315231315231315231315231315231315231315231315231
315231395231315231315231315231395231315231315231315231395231315231315231
315231395231314A31314A31314A31395231314A31314A31314A31394A31314A31314A31
314A31394A31314A31314A31314A31394A31314A31314A31314A31394A31314A31314A31
314A31395231314A31315231315231395231315231315231315231395A31315A31315A31
315A31396331315A31316331316331396331315A31316331316331396331316331316331
316331396B31316331316331316331396331316331316B31316331316331316331316B31
316B31396B31316B31316B31316B31316331316331316B31316B31316B31316B31316B31
316B31316B31316329316B31316331316331315A29316331315231314231000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000292929313131293129
313931293929314231294229314A31294229314A31294229314A31314229314A31314229
314A31314229314A31314229314A31314229314A31314231314A31314229314A31314231
314A31314231394A31314A31314A31314231394A31314A31395231314A31395231314A31
395231314A31395231315231395231315231395A31315231395A31315231395A31315231
395A31315231395A39315231395A31315231395A31315231395A31315231395A39315231
395A31315231395A31315231395231315231395239315231395231314A31395231314A31
395231314A31395239315231395231314A31395231315231395231315231395A39315231
395A31315231395A31315231395A31315A31396331315A31396331315A31396331316331
396331316331396B31316331396B31316331396B31316331396B31316331396B31316B31
396B31316331396B31316331396B31316331396B31316B31396B31316329396B31316B31
396B31316B29397331316B31396B31316329396B31316B29396B31316B29397331316B29
316B31316329396B31316329316B31316329316331315A29316331314A29000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000313131313131313931
313931314231314231314A31314A31314A31314A31314A31314231314A31314231314A31
314231314231314231314231314231314231314231314A31314231314A31314231314A31
314231314A31314A31314A31314A31314A31314A31314A31314A31315231315231315231
315231315231315231315231315231315231315231315A31315231315A31315A31315A31
315A31315A31315A31315A31315A31315A31315A31315A31315A31315A31315A31395A31
315A31315A31315A31315A31315A31315A31315A31395A31315A31315A31315A31395A31
315A31315A31315A31395A31315A31315A31315A31395A31315A31395A31315A31395A31
315A31315A31315A31396331316331396331316331396331316331396331316331396B31
316331396B31316B31396B31316B31316B31316B31396B31316B31396B31316B31397331
316B31316B31316B31316B31316B31316B31316B31396B31316B31316B31316B31317331
316B31316B31316B31316B31316B31316B31316B31316B31316B31317331316B31316B31
316B31316B31316B31316B31316331316B31316331316331315A29315A31314229000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000292929313931293929
314231294229314A31294A29315231314A29314A31294A29314A31314231314A31294229
314A31314231314A31294229314A31314231314A31314229394A31314231314A31314229
394A31314231314A31314A29395231314A31315231314A31395231314A31315231314A31
395A39315231395A31315231395A39315231395A31315231395A39315A31395A31315A31
396339315A31395A31315A31396339315A31396331315A31396339315A31396331315A31
396339315A31396331315A31396339315A31396331315A31396339315A31396331315A31
396339315A31396331315A31396339316331396331315A31396339316331396331316331
396B39316331396B31316331396B39316331396B31316331396B39316B31396B31316B31
397339316B31397331316B31397339316B31397331316B31397331317331397331316B31
397331316B31397331316B31397331317331397331316B29396B31316B31397331316B29
397331316B31397331316B29396B31316B31397331316B29397331316B29317331316B29
396B31316B29316B31316329396B31316B29316B31316329396331315A29315A31000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000312931313131313929314231
314231314A31314A31315231314A31315231314A31314A31314A31314A31314A31314A31
314231314A31314231314231314231314231314231314231314231314A31314231314A31
314A31314A31314A31314A31314A31315231314A31315231315231315231315231315231
315231315A31315231315A31315A31315A31315A31315A31315A31395A31315A31315A31
315A31396331315A31316331316331396331316331316331316331396331316331316331
316331396331316331316331316331396331316331316331316331396B31316331316331
316331396B31316331316B31316B31396B31316B31316B31316B31396B31316B31316B31
316B31396B31316B31396B31396B31397331316B31396B31396B31397331396B31397331
397331397331317331397331317331397331317331317331317331397331317331317331
317331397331317331317331317331397331317331317331316B31397331317331316B31
316B31317331317331317331316B31316B31316B31316B31316B31317331316B31316B31
316B31316B31316B31316B31316B31316B31316B29316B31316331316331315A29000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000313131293129314231294229
314A31294A29315231295229315231294A29315231314A29315231294A29314A31314229
314A31314229314A31314231314A31314229314A31314231314A31314231314A31314231
314A31314231314A31314A31314A31314A31395231314A31395231314A31395231315231
395A31315231395A31315231395A31315231395A31315A31396331315A31396339315A31
396331315A31396339316331396331316331396B39316331396B31316331396B39316331
396B31316331396B39316331396B31316331396B39316331396B31316331396B39316B31
396B31316B31396B39316B31397331316B31397339316B31397331316B31397339316B31
397331316B31397339316B31397331316B31397331317331397331397331397B31397331
397B31317331397B31317331397B31317331397B31317331397B31317331397B31317331
397B31317331397B31317331397B31317331397B31317331397331317331397331316B31
397331316B29397331316B31397331316B29396B31316B29397331316B29397331316B29
317331316B29397331316B29317331316B29317331316B29316B31316329316331314229
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000313131313931314231314A31
314A31315231315231315A31315231315231315231315231315231315231314A31314A31
314A31314A31314231314A31314231314231314231314231314231314A31314231314A31
314A31314A31314A31314A31314A31314A31314A31315231315231315231315231315231
315231315A31315A31315A31315A31315A31315A31315A31315A31316331316331316331
316331316331316331316331316331316B31316331396B31316331316B31316B31396B31
316B31316B31316B31396B31316B31316B31316B31396B31316B31396B31316B31397331
316B31397331316B31397331317331397331397331397331317331397331397331397331
397331397331397331397331397331397331397331397B31397331397B31397B31397B31
397B31397B31397B31397B31317B31397B31317B31397B31317B31317B31317B31397B31
317331317B31317B31397B31317B31317B31317331317331317331317331316B31317331
317331317331317331317331316B31316B31316B31317331316B31317331316B31316B31
316B31317331316B31317331316B31317331317331317331316B29316B31316331315A31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000292929313931293929314A31294A29
315231315229315A31295229315A31315229315A31295229315A31314A31315231294A29
315231314A31314A31294229314A31314231314A31314229314A31314231314A31314229
394A31314A31314A31314231394A31314A31315231314A31395239315231315A31315231
395A31315231315A31315A31396339315A31396331315A31396339315A31396331315A31
396B39316331396B31316331396B39316331396B31316B31397339316B31396B31316B31
397339316B31397331316B31397339316B31397331316B31397339316B31397331316B31
397339317331397331317331397B39317331397B31317331397B39317331397B31317331
397B39397331397B31317331427B39397B31397B31397B31428439397B31398431317B31
398439397B31398431317B31398439317B31398431317B31398439317B31397B31317B31
397B31317B31397B31317329397B31317331397B31317331397B31317331397331316B29
397331316B31397331316B29396B31316B31316B31316B29397331316B31317331316B29
397331316B29317331316B29397331317329317331316B29317331316B29316331295A29
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000313131313131314231294229315231
315231315A31315A31315A31315A31315A31315A31315A31315231315231315231315231
314A31314A31314231314A31314231314231314231314231314231314A31314231314231
314231314A31314A31314A31314A31314A31314A31315231315231315231315231315231
315A31315A31315A31315A31315A31316331315A31316331316331316331316331316331
316331396B31316331316B31316B31396B31316B31316B31316B31397331316B31317331
317331397331317331397331397331397331317331397331397331397331317331397331
397331397B31397331397B31397B31397B31397B31397B31397B31397B31397B31397B31
397B31397B31397B31397B31397B31398431397B31398431398431398431398431398431
398431398431318431398431318431398431317B31318431317B31398431317B31317B31
317B31398431317B31317B31317B31397B31317B31317B31317B31317331317331317331
317331397331317331317331316B31316B31316B31316B31316B31317331316B31316B31
316B31317331316B29317331317331317331317329317331317331317331316329315A31
314231000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000292929313131292929
313131292929000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000313131293129313931294229315231295229
315A31295A29316331296329316331295A29316331295A29315A31295229315A31315229
315231294A29314A31314229314A31314229314A31314231314231314229314A31314231
314A31314231394A31314A31314A31314A31395231314A31395231315231395A31315231
395A31315231395A31315A31396331315A31396339316331396331316331396B31316331
396B31316331396B39316B31396B31316B31397339316B31397339316B31397339317331
397331317331397B39317331397B39317331397B39317331397B39317331397B39317B31
397B39317B31398439397B31398439317B31398439397B31398439397B31428439397B31
428431397B31428439398431428439398431428C39398431428431398431398C31318431
398431318431398439318431398431318431398431318431398431317B31398431317B31
398431317B31398431317B31397B31317B31397B31317331397B31316B29397331316B31
397331316B31397331316331396B31316329396B31316B29397331316329396B31316B29
317331316B29317331317329317B31317329317B31317329317B31317329316B31315A29
315A31000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000292929312931292929312931292931312931
292929312931292931312931292931312931312931312931000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000313131313131313931314A31314A31315231
315A31316331316331316331316331316331316331316331316331316331315A31315A31
315231315231314A31314A31314A31314231314231314231314231314231314231314231
314231314A31314231314A31314A31314A31314A31315231315231315231315231315A31
315231315A31315A31315A31315A31316331316331316331316331316331316331316B31
316B31316B31316B31316B31316B31316B31316B31317331317331317331317331397331
317331397331317331397B31397B31397B31397B31397B31397B31397B31397B31398431
397B31398431398431398431398431398431398431398431398431398431398431398431
398431398431398431428C31398431428C31398431398C31398431398431398431398431
318431398431318431398431318431318431318431398431318431398431318431398431
317B31317B31317B31317B31317B31317B31317B31397B31317331317331317331316B31
316B31317331317331316B31316331316B31316B31316B31316331316B31316B31316B31
316B31317331317331317B31317331317B31317B31317B31317329317331316B29316331
315A29000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313131292931313131292929313131292931313131292931
313139292931313131292931313139292931313131292929313131292929313131292929
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000313131293129314231294A29315231295229
316331295A29316B31296329316B31316B29316B31296329316B31316329316331295A29
315A31315229315231294A29314A31314231314A31294229314A31314231314231314229
314A31314231314A31314229394A31314A31315231314A31395231315231315231315231
395A39315231395A31315A31396339315A31396331315A31396B39316331396B31316331
396B39316B31396B31316B31397339316B31397331316B31397339317331397331317331
397B39317331397B31317331397B39317B31397B31317B31398439397B31398439317B31
398439398431398439398431428C39398431398439398431428C39398431398C39398431
428C39398431428C39398431428C39398431398C31398431398C39398431398C31318431
398C39318431398C31318431398C39318431398C31318431398C39318431398431317B31
398439317B31398431317B31398431317B31398431317329397331317331396B31316B29
397331316B31316B31315A29396B31316331316B31316329396B31316329316B31316B29
397331317329317B31317329397B31317B29317B31317329317B31317329317331316329
316331000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000292929312931312931312939292931312939312939312939292931312939
312939313139292931312931312931313139292931312931312931313131292931312931
312931313131000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000293129313931314231315231315231315A31
315A31316331316331316B31316B31317331316B31316B31316B31316B31316331316331
316331315A31315231315231315231314A31314231314231314231314231314231314231
314231314231314231314231314231314A31314A31314A31315231315231315231315231
315231315A31315A31315A31315A31316331315A31316331316331316B31316331316B31
316B31396B31316B31316B31317331397331317331317331317331397331317331317331
317B31397B31317B31397B31397B31397B31397B31397B31397B31398439398431398431
398431398439398431398439398431398C39398431398439398431398C39398431398C39
398C31398C39398C31398C31398C31398C31398C31398C31398C31398C31318C31318C31
318C31398C31318C31318C31318C31398C31318C31318C31318C31398C31318431318431
318431398431317B31318431317B31397B31317B31317331317331396B31316B31316B31
316B31316B31315A31316331316331316331316331316331316331316B31316B29316B31
317331317331317329317B31317B31318431317B29317B31317B31317B31317329316B31
316331314A31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
312931292929313139292931313139292931313139292939313139292939313139292939
313139292939313139292931313139292931313139292931313139292931313139292931
313131292929313131292929000000000000000000000000000000000000000000000000
000000000000000000000000000000292929313931293929314A31294A29315A31295A29
316331296329316B31296B29317331296B29317331297329317331296B29317331296B29
316B31296329316331295A29315A31295229315231294A29315231294A29000000313931
314231313929314231313931314A31314231394A31314A31315231314A31395231315231
395A31315231395A31315A31396331315A31396331316331396331316331396B39316331
396B31316B31397339316B31397339316B31397339317331397B39317331397B39317331
397B39317331397B39317B31398439317B31398439317B31398439317B31398439318431
398C39318431398C39398431398C39398431398C39398431398C39398431398C39398C31
398C39398C31398C39398C31399439398C31399439318C31399431318C31399439318C31
398C31318C31399431318C31398C31318C31398C31318431398C31318431398431317B31
398431317B31398431317B31397B31317B31397B31317331396B31316B31397331316331
396B31316329396331315A29316331315A29396331315A29316331316329317331316B29
317331317329317B31317B29318431317B29318431317B29317B31317329317B31316B29
316B31315229000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000312931
292931312931292939312939292939312939312939312942292939312942312939312942
292939312939312939313139312939312939312939313139312939312939312931313131
312931312931312931313131312931000000000000000000000000000000000000000000
000000000000000000000000000000313131313131314231314A31315231315A31316331
316331316B31316B31317331317331317331317331317B31317331317331317331317331
316B31316B31316B31316B31316331316331316331316331315A31315231000000000000
000000313931313931314231313931314231314231314A31314A31314A31314A31315231
315231315A31315A31315A31315A31316331316331316331316331316331316331316B31
316B31316B31316B31317331317331317331317331397331317331317B31317B31397B31
317B31397B31317B31397B39317B31398439318431398439318431398439398439398C39
398439398C39398C39398C39398C39398C39398C39398C39398C39398C39398C39398C39
398C31398C39398C31399439398C31399431398C31399431318C31399431318C31399431
318C31398C31318C31398C31318C31398C31318C31398C31318431318431318431398431
318431318431317B31397B31317B31317B31317331317331316B31316B31316331316B31
316331315A31315A31315231315231315A31315A31316331316331316B31317331317331
317331317B31317B31318431318429318431318431318431317B29317B31317B31317331
316329315A31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000312931292929
313139292939313139292939313142292939313142292939313142292942313142292939
313142292942313142292939313142292939313142292939313142292939313139292931
313139292931313131292931313131292931313131000000000000000000000000000000
000000000000000000000000000000293129313931293929315231295229315A31295A29
316B31296B31317331296B29317B31297329317B31297329317B31317B29317B31297329
317B31297329317331296B29317331316B29316B31296329315A31000000000000000000
000000000000000000000000314231313931314231314229394A39314A31315231314A31
395231315231315A31315231396339315A31396331315A31396B39316331396B31316331
396B39316B31397331316B31397339317331397331317331397B39317331397B39317B31
398439317B31398439317B31398439318431398439318431398C39318439398C39318431
398C42318C39398C39318439398C42318C39398C39318C31399439318C39399439318C31
399439318C31399439318C31399439319431399431318C31399439318C31399431318C31
399439318C31398C31318C31398C39318431398C31318431398C39318431398431318431
398431317B31397B31317329397B31317331397B31316329396B31316331316331316329
395231314A29315231315229315A31315A29316331316329396B31316B29317331317329
318431317B29318431317B29398C31318429318431317B29318431317B29317B31297329
316B31315A29000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000292931292939
312939312942292939312942312942312942292942312942312942313142292942312942
312942313142292942312942312942313142312942312942312939313142292939312939
312939313139312931312931312931313131312931312931000000000000000000000000
000000000000000000000000000000313931293929314231314A31315A31295A31316331
316B31317331317331317331317331317B31317B31317B31317B31317B31317B31317B31
317B31317B31317329317331317331317331316B31316331315231000000000000000000
000000000000000000000000000000000000313931313931314231314A31314A31314A31
314A31315231315231315231315A31315A31315A31316331316331316331316331316B31
316B31396B31316B31316B31317331397331317331317331317331397B39317B31317B39
317B31398439317B31318439318439398439318439318439318439398C39318439398C39
398C39398C39318C39398C39398C39398C39318C39398C39398C39399439318C39399439
399439399439319431399431399431399431319431319431319431399431318C31318C31
318C31398C31318C31318C31318C31398C31318431318C31318431398431318431318431
317B31397B31317331317B31317331397B31316331316B31315A31316331000000313931
314231315231315229315A31315A31316331316329316B31317331317331317B29317B31
318431318431318429318431318431318C31318429318431318431318431317B29317331
316B31316331000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000292931313139292939
31314229293931314229294231314A29294231314A29294231314A29294231314A292942
31314A29294231314A29294231314A292942313142292942313142292939313142292939
313139292931313139292931313139292931313131292931313131292929313131000000
000000000000000000000000000000293129314231294229315231295229316331296331
316B31296B29317B31297329317B31297329318431297B29318431297B29318431297B29
318431297B29317B31297329317331296B29316B31295A29000000000000000000000000
000000000000000000000000000000000000000000000000313931313931394A31314A31
315231314A31395231315231395A31315231395A39315A31396331316331396B31316331
396B31316B31397339316B31397331316B31397339317331397B39317331397B39317B31
397B39317B31398439317B31398439318439398C39318439398C39318439398C42318C39
399442318C39399442318C39399442318C39399439318C39399439318C31399439319431
399439319431399439319431399439319431399C39319431399431318C31399439318C31
398C31318C31398C39318431398C31318431398C31318431398C31318431398431317B31
397B31317331397B31317331397B31316331395A31315A31000000000000000000313129
314A31314A29315A31315A29316331316329317331316B29317B31317329317B31317B29
318431318429318C31318429318C31318429318C31318429318431317B29317B31317329
317331316329314231000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000292931312939292939312942
29294231294229294231294A29294A31294A29294A31294A31294A31294A31294A31314A
31294A31294A31294A31314A31294231314A312942313142312942313142312942313142
312939313139312939313139312931313131312931313131312931313131312931313131
312931000000000000000000312931313931314231314A31315231315A31316331316331
316B31317331317331317B31317B31317B31317B31318431318431318431318431318431
317B31317B31317B31317B31317331317331316331315A31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000313931314231
314231314A31314A31315231315231315A31315A31315A31315A31316331316331316331
316331316B31316B31316B31316B31317331317331317331317331317B31317B31397B39
317B31317B39317B39318439318439318439318439398439318439318C39318C39398C42
318C39398C39318C39399439318C39319439318C39399439318C39399439319431399439
319431319431319431399431319431399431319431399431318C31319431318C31398C31
318C31398C31318C31398C31318C31318C31318431398C31318431318431317B31397B31
317331317331317331317331316331315A31000000000000000000000000000000313131
314229314A31315231315A31316329316B31316B31317331317331317B31317B31318431
318429318431318431318C31318431318C31318C31318C31318429318431317B31317B31
316B29316331314A29000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000313139292939312942292942
31314A29294231314A29294A31315229294A31314A29294A31315229294A31314A29294A
31315229294A31315229294A31314A29294231314A29294231314A29294231314A292942
313142312939313142292939313142312939313139292931313139312931313131292931
313131312931313131000000313131293929314231294229315A31295A29316331296331
317339297331317B31297329317B31297B31318431297B29318431298431318431297B29
318431297B29317B31297329317B31296B29316331295229000000000000000000000000
000000000000000000000000000000000000313131292931313139312931313139292931
393931314231314A31314A31395239315231315A31315231396339315A31396331316331
396B39316331396B31316B31397339316B31397331317331397B39317331397B39317331
397B39317B31398439317B31398439318439398439318439398C42318439398C39318439
399442318C39398C39318C31399439318C39399439318C31399439318C31399439318C31
399439318C31399431318C31399439319431399431318C31399439318C31398C31318C31
398C39318C31398C31318431398C39318431398C31318431398439317B31397B31317331
397B31316B31396B31316329000000000000000000000000000000000000000000292929
314231314A29315A31295A29316B31316B29317331317329317B31317B29318431317B29
318C31318429318C31318429318C31318C29318C31318429318C31317B29318431317329
317331316329315231000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000029293129294231294229294231294A
29294A31294A29294A31294A31294A31314A29294A31295231294A31315229294A312952
31294A31315231294A31294A31294A31314A31294A31294A31294A31314A31294A31314A
31294231314A312942313142312942313142312939313139312939313139312931313131
312931313131312931313131312931313931294229314A31315231316331316331316B31
316B31317339317331317B31317B31318431317B31318431318431318431318431318431
318431318431317B31317B31317331317331315A29000000000000000000000000000000
000000000000000000313131313131313139312931313139313139313139312939313139
313139313142312942314231314A31315231315231315231315A31315A31315A31316331
316331316331316331316B31316B31397331317331317331317331397B39317331317B31
317B39397B39317B31317B39317B39398439318439318439318439398C39318439318C39
318C39398C39318C39318C39318C39398C39318C31318C31318C31399431318C31318C31
318C31399431318C31319431319431399431318C31318C31318C31398C31318C31318C31
318C31398C31318C31318C31318431398431318431318431317B31397B31317331316B31
316B31396B31316331000000000000000000000000000000000000000000000000000000
313931314A31315229315A31316331316B31316B29317331317B31317B31317B29318431
318431318C31318C29318C31318C31318C31318C29318C31318431318431317B29317B31
317331316B31315A29000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000031313129293931314A29294231314A29294A
31315229294A31315229294A31315229294A31315229294A31315229294A31315229294A
31315229294A31315229294A31315229294A31315229294A31315229294A31315229294A
31314A29294A31314A29294231314A29294231314A312939313142292939313139312931
313139292931313139312931313131292931313131294A29315A31295A29316B31296331
317339297331317B39297331317B31297B29318431298429318431297B29318431297B29
318431297B29317B31297329317331296329315A31000000000000000000000000000000
313131312931313139312931313139312931313139312939313142312939313142312942
31314A31294231314A31294231314A314231394A31314A31395231315231395A39315A31
396331315A31396B39316331396B31316B31397339316B31397331317331397B39317331
397B39317331397B39317B31398439317B31398439318431398439318431398C39318431
398C39318431398C39318C31398C39318C31398C39318C31398C31318C31399439318C31
399439318C31399439318C31399431318C31399439318C31398C31318C31398C39318C31
398C31318431398431317B31398431317B31398431317331397B31317331396B31316331
396331000000000000000000000000000000000000000000000000000000000000000000
313931294229315231315A29316331316329317331317329317B31317B29318431317B29
318C31318429318C31318C29318C31318C29319431318C29318C31318429318431317B29
317331316B29316331000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000029293931294229294A31294A29294A312952
29294A31295229294A31295229294A31295231294A313152292952312952312952313152
31295231315231295231315231295231315231295231315231294A313152312952313152
31294A31315231294A31314A31294A31314A31294A31314A312942313142312942313142
312939313139312939313139312939313139312931313131312931313131316331316B39
316B31317339317B31317B31317B31318431318431318431318431318431318431318431
318431318431317B31317331317331316331315231000000000000313131313131313139
31293131313931313931313931293931314231314231314231294231314A31314A31314A
31294A31314A31314A31314A31294A313152313152314231314A31315231315231315A31
315A31315A31316331316331316331316B31316B31316B31316B31317331317331317331
317331317B31317B31397B31317B31317B31317B31398439318431318439318431398C39
318431318C39318C31398C31318C31318C31318C31398C31318C31318C31318C31398C31
318C31318C31318C31398C31318C31318C31318C31398C31318C31318C31318431398431
318431318431317B31397B31317B31317B31317B31397331316331316331315A31000000
000000000000000000000000000000000000000000000000000000000000000000000000
313129314231314A31315A31316329316B31316B31317331317331317B31317B31318431
318431318C31318C31318C31318C31318C31318C31318C31318431318431318431317B31
317329316B31316329000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000029293131294229294231315229294A31295229294A
31315229294A31315229294A31315229294A31315229294A31315229295231315A292952
31315A29295231315229295231315A312952313152292952313152292952313152292952
31315231295231315229294A31315231294A31315229294A31314A31294A31314A292942
31314A312942313142292939313142312939313139292931313139312931313139292931
313139317331317B39297B31318439297B31318431298429318C31318431318431297B29
318431297B29317B31297329317331312931313139312931313139312939313139292939
31314231293931314231293931314A31294231314A31294239315231294A31315231294A
31315231294A31315231294A39315231295231315A31295231315A314231314A31314A31
395239315231395A31315A31396339316331396B31316331396B39316B31397331316B31
397339317331397B31317331397B39317B31398431317B31398439318431398439318431
398C39318431398C31318431398C39318431398C31318431398C39318C31398C31318431
398C39318C31398C31318431399439318C31398C31318431398C39318431398431317B31
398439317B31397B31317331397B39317331396B31315A29000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
313131313929315231295229316331316329316B31296B29317B31317B31318431317B29
318C31318431318C31318429399431318C29319431318C29318C31318429318431297B29
317B31316B29316B31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031293929294229294A292952312952292952312952
312952313152292952312952312952313152292952312952312952313152292952312952
31295231315229295231295231295231315A31295231315231295231315A312952313152
31295231315A31295231315231295231315231295231315231315231315231294A31314A
31314A31314A312942313142313142313142312942313142313139313142312939313139
313139313139312931313139313131313131312931313131313131313131312931313131
313131313139312931313139313139313139312939313142313142313142312942313142
31314231314A31294231314A31314A31314A31294A313152313152313152312952313152
31315231315231295231315231315A31315A31295A31315A31315231315A31295231315A
314231315231315231315A31315A31316331316331316331316331396B31316B31316B31
316B31397331317331317331317331397B31317B31317B31317B31398431317B31318431
318431398431318431318431318431398431318431318431318431398C31318431318431
318431398C31318431318431318431398C31318431318431317B31397B31317B31317B31
317B31397B31317331317331316B31315A31000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
313131314231314A29315231315A31316331316B29317331317331317B31317B31318431
318431318C31318431318C31318C31319431318C31318C31318C31318C31318429317B31
317B31317331316329000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000029293931314A29294A31315229295231315A292952
31315A29295231315A29295231315229295231315A29295231315A29295231315A292952
31315A29295231315A29295231315A29295231315A29295231315A29295231315A292952
31315A29295231315A31295231315A29295231315A31295231315A31295231315A312952
31315229294A31315231294A31314A31294231314A31294231314A31294231314A312942
313142312939313142312939313142312939313142312939313142312939313142312939
31314231293931314231293931314231294231314A31294231314A31294231314A31294A
31314A31294A31315231294A31315231295231315A31295231315A31295231315A312952
31315A31295231315A31295A31316331295A31316331295A31315A31295231316331295A
31315A31295A313163314231395231315231395A31315A31396331316331396B39316331
396B31316B31397339316B31397339317331397B39317331397B31317331397B39317B31
398439317B31398439317B31398431317B31398439317B31398439317B31398439317B31
398431317B31398439317B31398431317B31398439317B31397B31317331397B31317331
396B31316331396331315231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000293929314A31315229315A31316329316B31316B29317B31317329318431317B31
318431318431318C31318C31319431318C29399431318C29318C31318429318431317B29
317B31316B29316B31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000029293131294229294A31295229295231295A29295A31295A
29295231295229295231295229295231295231295231295231295231295A31295231315A
31295231295A31295A31315A31295A31315A31295A31315A31295A31315A31295A31315A
31295A31315A31295A31315A31295A31315A31295A31315A31295A31315A31315A31315A
31295231315231295231315231295231315231314A31315231294A31314A31294A31314A
31294A31314A31314A31314A31294231314231314231314A31314231314A31314231314A
31294231314A31314A31314A31314A31314A31314A31314A31294A313152313152313152
31315231315231315231315A31295A31315A31315A31315A31315A31315A31315A31315A
31295A31315A31315A31316331316331316331316331316331295A31315A313163313163
31315A31315A31315A31316331295A313163314A31315231315231315A31315A31316331
316331316B31316B31316B31316B31316B31316B31397331317331317331317331317B31
317B31317B31317B31397B31317B31317B31317B31317B31317B31317B31317B31397B31
317B31317B31317B31317B31317B31317B31317331397B31316B31316331316331315A31
315A31000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313931314231315231315A29316331316B31317331317331317B31317B31318431
318431318431318431318C31318C31319431319431319431318C31318C31318431318431
317B29317331316B31314A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031314229294A31295229295231316329295A31295A29295A
31315A29295231315A29295231315A29295231315229295231315A29295231315A292952
31315A29295A31315A29295231315A29295A31315A29295A31316331295A31316329295A
31316331295A31316329295A31316331295A31316329295A31316331295A31316329295A
31316331295A31315A29295231315A31295231315A29295231315A31295231315229294A
31315231294A31315229294A31315231294A31315231294A31315231294A31315231294A
31315231294A31315231294A31315231295231315231295239315A31295231315A312952
31316331295A31316331295A39316331295A31316331295A39316331295A31316331295A
39316331296331316331296331316331296331316331296331316331296331316331295A
31316331295A31316331296331316331296331316331295A313163314A31395231314A31
395A39315231395A31315A31396B39316331396B31316B31397339316B31397331317331
397B39317331397331317331397B39317331397B31317331397B39317331397B31317331
397B39317331397B31317331397B39316B31396331315A31395A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313129314231294A29315A31315A29316B31296B29317331317331317B31297B31
318439318431318C39318431399439318C31319431318C29399431318429318431317B29
317B31317329316B31294A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000029294231295229295229295A29295A31296329295A312963
29295A31295A29295231295231295231315229295231295231295231315A29295231295A
31295A31315A31295A31295A31295A31315A31295A313163312963313163312963313163
312963313163312963313163313163313163312963313163313163313163312963313163
31316331316331295A31316331315A31316331295A31315A31315A31315A31295A31315A
31315A31315A31295231315231315231315A31295231315231315231315A31295231315A
31315A31315A31295A31315A31315A31315A31295A31315A31315A31316331295A313163
313163313163312963313163313163313163312963313163313163313163312963313163
31316331316B31296331316331316331316B31296331316331316331316B312963313163
31316331316331296331316331316331316331296331316331316331316331295A313163
31315A313163314A31314A31315231315A31315A31315A31316331396331316331316B31
316B31397331316B31316B31316B31396B31316B31316B31316B31396B31316331316331
316331396331316331315A31315231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313931314229315231315A31316331316329316B31317331317B31317B31317B31
318431318C39318431318C31318C31319431318C31319431318C31318C31318429318431
317B31317331316B29315231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029292931294A29294A31316329295A313163292963313163292963
31316329295A31315A29295231315A29295231315A29295231315A29295231315A292952
31315A29295231316329295A31316329295A31316329295A313163292963313163292963
31316329296331316B31296331316B29296331316B31296331316B29296331316B312963
31316B31296331316B31296331316331296331316B31296331316331296331316331295A
31316331295A31316331295A31316331295A31316331295A31316331295A31316331295A
31316331295A31316331295A31316331295A31316331296331316B31296339316B312963
39316B31296339316B31296339316B31296339316B31296331316B31296331316B31296B
31316B31296331316B31296331316B31296331316B31296331316B31296331316B312963
31316331296331316B31295A31316331296331316B31296331316331295A31316331295A
31316331295A31316331296331316329295A31316331295A313163314A31395239315231
395A31315231395A39315231395A31315231395A39315231395231314A31313139000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000293129314231314A29315A31315A29316B31316B29317331317329317B31317B31
318431318431318C39318C31318C39318C31319439318C29319431318429318C31317B29
317B31317329316B31315229000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031293929294A29295229295A31296329296331296B292963312963
29296331296329295A31295A29295231295231295231295229295231295231295231315A
31295A31315A31295A31315A31295A31316331296331316331296331316331296331316B
31296331316B31296331316B31296B31316B31296B31316B31296B31316B31296B31316B
31296B31316B31316B31316B31296B31316B31316B31316B31296B31316B31316B31316B
31296331316B31316331316B31316331316331316331316B31316331316B31316331316B
31316331316B31316B31316B31316B31316B31316B31316B31316B31316B31316B39316B
31316B39316B39316B39316B31316B31316B31316B31316B31316B31316B31316B313173
31316B31316B31316B31316B31316B31316B31316B31316B31316B31316B31316B313163
31296331316B31316331316331296331316331316331316B31296331316331295A313163
31295A31316331296331316331295A313163312963313163312963313163312963313163
31296331316B31296331316B31296331316B31296331315A31295231314A312939000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313131314231314A31315229315A31316331316B31317329317331317B31317B31
318431318439318439318C39318C31318C39318C31319431318C31318C31318431318431
317B29317331316B29315A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029293931315229295A31296329296331316B29296B31296B292963
31316B29296331316329295231315A29295231315229294A31315A29295231315A292952
31315A29295A31315A29295A31316331296331316329296331316B31296331316B292963
31316B31296331316B29296331316B31296B31316B29296B31317331296B31317329296B
31317331296B31316B29296B31317331296B31317331296B31317331296B31316B29296B
31317331296B31317331296B31317331296B31317331296B39317331296B31317331296B
31317331296B31317331296B39317331296B39317331296B39317331296B39317331296B
39317339316B39317331296B39317331297339317331296B39317331297331317331296B
39317331296B31317331296B39317331296B31317331296B31317331296B31316B292963
31316B31296331316331296331316B31296331316329296331316331296331316329295A
31316331295A31316329295A31316331296331316329295A313163312963313163292963
31316B31296B31316B29296331316B29296331316B29295A31315A29295231314A292931
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313129314231294A29315A31315A29316331296329317331317329317B31317B29
318439318431318C39318431399439318C31319439318C31399431318C29318C31317B29
318431317329316B31295A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029294229295231295A29296329296B29296B31297329296B29296B
29296B31296B29296331296329295A31295A292952312952312952313152292952312952
31295231315A29295A31295A31295A31316331296331316331296331316B31296331316B
31316B31316B31296B31316B31316B31317331296B313173313173313173312973313173
31317331317331296B31317331317331317331297331317331317331317331296B313173
313173313173312973313173313173313173312973313173313173313173313173313173
313173313173312973313173313173393173313173393173393173393173393173393173
39317339317339317339317339317339317B313173313173313173313173313173313173
31317331317331296B31317331317331317331296B31316B31316B31316B31296B31316B
31316B31316331296331316331316331316331296331316331316331315A31295A313163
31315A31316331296331316331316331316331295A31316331296331316331296331316B
31296B31316B31296B31316B31296B31316B31296B31296331295A31315231294A312939
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313131313929314A31315231315A31316329316B31317331317331317331317B31
318431318439318439318C39318C39318C39318C31318C31318C31318C31318429318431
317B31317B31316B29316331000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031293929294231315A29295A31316B29296B31297329296B313173292973
31317329296B31316B29296331316329295A31315A29295231315229294A31315A292952
31315A29295231315A29295A31316329295A31316329296331316B29296331316B29296B
31317329296B31317331296B31317329296B31317331296B31317331297331317B312973
31317B31297331317B31297331317B31297331317B31297331317B31297331317B312973
31317B31297331317B31297331317B31297331317B31297331317B31297339317B312973
31317B31297339317B31297339317B31297339317B39317339317B39297342317B393173
39317B39297339317B31317B39317B31297339317B31297331317B31297339317B312973
31317331297331317331297331317B31297331317331296B31317331296B31317331296B
31316B31296331316B31296331316331296331316B31296331316331295A31316331295A
31316329295A31316329295A31316329295A31316331296331316B29296331316B29296B
31317329296B31317329297331317329296B313173292963313163292952313152292942
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000292929314231314229315231295A29316331316329317331317329317B31317B31
318439318431318C39318439318C39318C31399439318C31319431318C29318C31317B29
318431317329317331315A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000029293931294A29295A31296329296B31296B29297331297329297331297B
29297331297329297331296B29296B31296329295A31295A29295231294A31294A313152
29295231295231295231315A31295A31316331296331316331296331316B31296B31316B
31296B31317331297331317331297331317331297331317331297331317331317331317B
31297331317B31317B31317B31317B31317B31317B31317B31297B31317B31317B31317B
31317B31317B31317B31317B31317B31317B31317B31317B31317B31317B31317B31317B
31317B39317B39317B39317B39317B39317B39317B39317B39317B42317B39317B39317B
39317B39317B39317B39317B31317B31317B31317B31317B31317B31317B31317B31317B
31317331317B31317331317B31317331317331317331317331316B31317331316B31316B
31316331316331316331316331296331316331316331316331295A31315A31315A31315A
31295A31316331295A31315A31295A31316331296331316331296331316B31296B313173
31297331317331297331317B31297331317331296B31316B31296331315A31295231314A
312931000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
3129293929313929294231314229294A3131000000000000000000000000000000000000
000000313131313931314A31315229315A31316331316B31316B31317331317B31317B31
317B31318439318439318C39318C39318C39318C31318C31318C31318C31318431318431
317B29317B31316B31316331000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031294229294A31316329296331296B29296B31317329297331297B292973
31317B29297B31297B29297331317329296B31316B29296331316329295A31315229294A
31315229295231315229295231316329295A31316329296331316B31296331316B29296B
31317331296B31317329296B31317B31297331317B29297331317B31297331317B292973
31317B31297B31317B31297B31318431297B31318431297B39318431297B31318431297B
39318431297B31318431297B39318431297B31318431297B39318431297B39318431297B
39318439317B39318439297B42318439317B42318439317B42398439318442318439297B
39318439317B39318431297B39318431317B31318431297B39318431297B31317B312973
39317B31297B31317B31297331317B31297331317331296B39317331296B313173312963
31316B31296331316329295A31316331296331316329295A31316331295231315A292952
31316331295231315A29295231316331295A31316329296331316B31296B31317329296B
31317B29297331317B29297331317B29297331317329296B31316B29296331315A29294A
313139000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004231314A2929
4A31314229294A31314A29295231315229295231315229294A3131392929313131000000
000000312929313931294229315231315A29316331296329317331317329317B31297B31
318439318431318C39318439398C42318C39318C39318C31399431318C29318C31317B29
318431317329317331295A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000029294229295229296331296B29296B29297329297331297B29297B29297B
29297B31298429297B29297B29297B31297B29297329296B29296B31296B292963292963
29295A31314A29295231295231295A31316331296331316331296331316B31296B31316B
31296B31317331297331317331317331317B31297331317B31317B31317B31297B31317B
31317B31317B31297B31317B31317B31318431297B313184313184313184312984313184
313184313184312984313184313184313184313184313184313184393184393184393184
39318439318C39318439318439318442318C423184423184423184423184393184393184
39318439318431318431318431318431318431317B31318431317B31318431297B31317B
31317B31317B31297331317331317331317331297331317331317331317331296B31316B
31316331316B31295A31316331316331316331295A31315A31315231315A31295231315A
31315A31315231295231315A31295A31316331296331316B31296B31317331297331317B
31297B31317B31297B31317B31297B31317B31297331317331296B31316331295A312952
312939000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003931313929294229314A29295231315229295A2931
5A29315A31315229295A29315A29295A31315229295A29315A29295A31315229294A2929
422929313131313129314231314A31315A31315A29316331316B31317331317331317B31
317B31318439318439318C39318C39318C39318C31318C31318C31318C31318429318431
317B31317B31316B29316331000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031314A29295231296329296331317329297331317B29297B31318429297B
31318429297B31318429297B31318429297B31317B29297331317329296B313173292963
31316B29295A31314A29295231315229295A31316329296331316B29296331316B29296B
31317329296B31317331297331317B29297331317B31297331317B29297B31318431297B
31318431297B31318431297B31318431297B31318431298431318431298439318C312984
31318C31298439318C31298439318C31298439318C31298439318C31298439318C393184
39318C39318442318C39318442318C39318442318C39318442318C39318439318C313184
39318C31318431318431318431318431297B31318431297B31318431297B31318431297B
31318431297331317B31297331317B31297331317B31297331317331296B313173312963
31316B31295A31316331295A31316331295A31315A31295231315229294A31315229294A
31315229295231315A29295A31316329296331316B29296B31317329297331317B29297B
31318429297B31318429297B31317B29297331317B292973313173292963313163292952
31314A000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003931313929294A31314229295231315229295A31315A2929
6331315A29296B31316329296331315A29296331315A29296331315A2929633131522929
523131292929313931314229315231315229316331316329316B31316B29317B31317B31
318439317B31318C42318439318C39318431318C39318C31319431318C29398C31318429
318431317329316B31315A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029293129294A31295229296329296B29296B31297329297B31297B29297B312984
29298431298429298431298429297B31298429297B31297B29297B31297329297331296B
29296331295A29295231294229294A31315231295A31316331296331316B31296B31316B
31296B31317331297331317331297331317B31297B31317B31297B31317B31317B313184
31297B31318431318431318431298431318431318431318431318431318431318431318C
31318C31318C31318C31318C31318C31318C31318C39318C31318C39318C39318C39318C
39318C39318C39318C42318C39318C39318C39318C39318C39318C39318C39318439318C
313184313184313184313184313184313184313184313184313184313184313184313184
31317B31317B31317331317B31317331317331317331317331316B31316B31316B31316B
31295A31316331315A31315A31295A31315A31295231314A000000000000000000000000
31295231315A31296331316B31296B31316B31297331317B31297B31317B31297B313184
31297B31318431298431318431297B31317B31297B31317331296B31316B31296331315A
31294A313131000000000000000000000000000000000000000000000000000000000000
0000000000003129293931314229294A31315229294A31315229295A31315A29295A3131
6329296331316B29296B31316B2929633131632929633131632929633131632929633131
5A2929313131313931314231314A29315A31315A31316331316B29317331317331317B31
317B31318439318439318C39318C39318C39318C31318C31318C31318C31318431318431
317B31317331316B31316331000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029293131295229295231316B29296B31297329297331317B29297B31298429297B
31318429298431318C29298431318429297B31298429297B31317B29297331297329296B
31316B29295A31295200000000000000000031315229295231316329296331316B292963
31317331296B31317329297331317B31297331317B29297331318431297B31318429297B
31318431298431318429298431318C31298431318C31298439318C31298431318C312984
39318C31298C31318C31298C39319431298C39319431298C39319431318C39319431298C
39319439318C39318C39298C42399439318C39318C39318C39398C31318C39318C313184
39398C31318431318C31298439318C31318431318C31298439318431298431318431297B
31318431297331317B31297339317B31297331317331296B31317331296B31316B29295A
31316331295A31315A29295231315A000000000000000000000000000000000000000000
31314A29295A31316329296331317329297331317B29297331318429297B31318429297B
31318429298431318429298431318429297B31317B29297331317329296B313163292952
313152292931000000000000000000000000000000000000000000000000000000000000
0000004229295231314A29295231315229296331315229295A31315A29296331315A2929
6331316329296B31316329297331316B29296B31316329296B3131632929633131632929
633131312929313931294229315231315229316331296329317331317329317B31317331
318439318439318C39318431398C39318431318C31318C29399431318C29318C31317B29
318431317329316B31294229000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031293929294A29295A29296331296B29297329297B29297B31298429297B292984
29298431298C29298429298429298431298429297B29297B29297B31297B29297329296B
29296331295A29294A00000000000000000000000031315231295A313163312963313163
31316B31317331297331317331317331317B31297331317B31317B31318431297B313184
31318431318431298431318431318431318C31298431318C31318C31318C31298C31318C
31318C31319431318C31319431319431319431318C31319439319439319439318C393194
39318C39319439318C39318C39318C39319439318C39318C39318C39318C31318C31318C
31318C31318C31318C31318C31318431318C31318431318431318431318431317B31317B
31317B31317B31297331317B31317331317331296B31316B31316B31316B31295A31315A
313152313152312952000000000000000000000000000000000000000000000000000000
00000031315229296331296B31297331317B31297B31317B31297B313184292984312984
31298431318C31298431318431318431318431297B31297B31297331317331296331315A
312952313142000000000000000000000000000000000000000000000000312929423131
4A31315231315229295231315A2931633131632929633131633131633131632929633131
6329316B31316B29296B31316B29317331316B29297331316B29296B31316329296B2931
6B2929313131313929314231315231315A31315A29316331316B31317331317331317B31
317B39318439318439318C39318C31318C31318C31318C31318C31318C31318431318431
317B31317331316329314A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029293931315229295A31316B29296B31317329297331318429297B313184292984
31318C29298431318C29298431318C29298431318429297B31317B29297331317329296B
31296329295231314A000000000000000000000000000000000000292952313163292963
31316B29296B31317331296B31317329297331317B31297331317B31297B31318431297B
31318431298431318C31298431318C31298431318C31298C31318C31298C31319431298C
31319431298C31319431299431319431298C39319431298C39319431298C39319431318C
39319431318C39319431318C39319431318C39399431318C39319431318C31319431318C
31319431318C31318C31318431318C31298431318431298431318431297B31318431297B
31317B31297331317B31297331317331296B31317331296B31316B31296331315A31294A
313152292942000000000000000000000000000000000000000000000000000000000000
00000029294A31316329296331317329297331318429297B31318429298431318C292984
31318C29298431318C29298431318C29297B31318429297331317B29296B31316B29295A
31315A292942000000000000000000000000000000000000000000392929423131422929
5231315229295A31315A29296331316329296B31316329296B31316329297331316B2929
6B31316B29297331316B29297331316B29297331316B29297331316329296B3131632929
6B3131292929313931314229315231295229316331316329317331317331317B39317B31
318439318431318C39318431318C39318431399431318C31399431318429398C31317B29
318431316B29316B31314A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294229295229295A29296B31297329297331297B29297B292984292984312984
29298431298C29298431298429298429298429298431298429297B31297329297331296B
29296329295229294200000000000000000000000000000000000000000031295231315A
31296331316331296B31316B31297331317331297331317B31297B31317B31317B313184
31298431318431318431318C31318431318C31318C31318C31298C31319431318C313194
313194313194313194313194313194313194313194313194313194313194313194313194
31319431319431319431319431319431319431319431319431319431319431318C31318C
31318C31318C31318C31318C31318431318431318431318431318431318431317B31317B
31317B31317B31317331317331316B31316B31316B31316331296331315A31294A523131
522931423131000000000000000000000000000000000000000000000000000000000000
00000031294A31295A31316B31296B31317331297B31318431298431318431298431318C
31318C31318C31318431318C31318431318431317B31317B31297331317331296B313163
31295A31294A0000000000000000000000000000000000004229294A31314A2931523131
5229315A31315A29316331316329296B31316B29316B31316B2931733131732931733131
6B29296B31317329296B31316B29297331317329317331316B29296B31316B29296B3131
6B29296B3131313931314A31315229315A31315A31316B31316B31317331317331317B39
317B31318439318439318C31318431318C31318C31399431318C31318C31318431318431
317B31317331316B31314A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A31295A29295A31316B29296B31297B29297331318429298431318429297B
31318C29298431318C29298431318429298431298429297B31317B292973312973292963
31316329294A000000000000000000000000000000000000000000000000000000000000
31316331295A31316B29296331317331296B31317329297331317B31297B31317B29297B
31318431298431318429298431318C31298431318C29298C31319431298C31319431298C
31319431299431319C31299439319C31299431319C31299439319C313194313194312994
39319C31319431319431318C39399431318C31319431318C39399431318C31319431298C
39318C31318431318C31298439318C31298431318431297B39318431297B31318431297B
31317B31297331317331296B31317331296331316329295A31315A5A29296331315A2929
5A3131522929523131422929000000000000000000000000000000000000000000000000
00000029294231315A29296331317329297331317B29297B31318429298431318C292984
31318C29318431318C29318431398C31318431318429297B31317B29297331316B292963
31315A29294A3131310000003131313929294231314A29295A31315229295A3131522929
6331316329296331315A29296B31316B29296B31316B29297331317329297B3131732929
7331316B29297331316B29297331317329297331316B29297331316B29297331316B2929
6B3131632929313931294229315231315229316331296329317331317331317B39297331
318439318431318C39318431398C31318C31399431319429399431318C31318C31317B29
317B31316B29316B31294A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031295229295229296329296B31297329297329297B292984312984292984292984
29298431318C29298429298429298431298429297B29297B29297B31297329296B29296B
29295A31294A000000000000000000000000000000000000000000000000000000000000
00000000000031295A31316331316B31316B31296B31317331317331317B31297B31317B
31318431318431298431318431318431318C31298431318C31318C31318C31298C313194
313194313194312994313194313194313194313194313194313194313194313194313194
31319431319431318C31319431319431319431318C31319431318C31318C31318C31318C
31318C31318C31298431318431318431318431318431318431317B31317B31297B313173
31317331317331296B31316B31315A31315A31295A633131633131633131632931633131
6329316331315A2931523131523131423131000000000000000000000000000000000000
00000000000029295231296331296B31317329297B31297B31298431318431298431298C
31318C31318C31318C31318C31318C31318C31318431317B31317B31317329296B312963
31295A3131522929313931314231314A31315229295A31315A3131633131632929633131
6331316B31316329296331316B31317331316B2929733131733131733131732929733131
7331317331317329296B31317329317331316B29297331317329317331316B29296B3131
6B2929313131313929314A31315231315A31315A29316B31316B31317331317331317B31
318431318439318431318431318C31319431319431319431318C31318C31318431318431
317B31317331316B29314A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A31315A29296331297329297331317B29297B31318429298431318C292984
31318C29298431318C29298431318429297B31318429297B31317B29296B31316B292963
31315A292942000000000000000000000000000000000000000000000000000000000000
00000000000000000029295231316329296331316B31296B31317329297331317B31297B
31318431297B31318431298431318C31298431318C31298431318C31298C31319431298C
31319431298C31319431299431319431299431319431299431319431298C31319431318C
31319431318C31319431318C31319431318C31319431318C31319431298C31318C312984
31318C31298431318C31298431318431297B31318431297B31317B31297331317B31296B
31317331296B31316331295A31315A6329296B31316329296B31316329296B3131632929
6B31316329296331315A29295A31315229295231314A2929000000000000000000000000
00000000000031315229295A31316B29296B31317B29297B31318429298431318C292984
31318C29298431318C29318431398C29318431318429317B31317B292973313173292963
3131634229294A31314A29295231315229296331315A29296331316329296B31316B2929
7331316B29297331316B29297331317329297331316B29297331317329297B3131732929
7B31317329297B31316B29297331317329297331316B29297331316B2929733131632929
6B3131292929314231314A29315231295229316331316B29317331317331317B39317B31
318439318431318C31318431398C31318C31399431319431399431318429398C31317B29
317B31296B29316B31314A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A29295A31296329296B29297329297B31298429298431298C29298431298C
29298431318C29298C31318C29298431298429297B31297B29297329297329296B312963
29295A29294A000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031315A31296331316B31296B31317331297331317B
31297B31318431318431318431298431318431318431318C31318431318C31318C31318C
31318C31318C31318C31319431319431319431319431319431318C31319431318C31318C
31318C31318C31318C31318C31318C31318C31318C31318C31318C31318C31318C31318C
31318431318431318431318431317B31317B31317B31317B31317331317331316B31316B
31315A31315A3131526B31316B29316B31316B29316B31316B29316B31316B29316B3131
6329316B31316329316331316329316331315A29315A3131522931523131422931000000
00000000000031294231316331296331317331297331317B31297B31318431298431318C
31298C31318C31318C31318C31318C31318C31318431318431317B31317B3129734A3131
4A29315231315231315A31316329316331316331316B31316B29316B31316B3131733131
7329317331317331317331317329317B31317331317331317329317B31317329317B3131
7B29317B31317329317331317329297331316B29297331317329297331316B29296B3131
632929313131313931314A31315229315A31316331316B31316B31317331317B31317B31
318431318431318431318C31318C31319431319431399431318C31318C31318431318431
317329317331316B31315231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294231295A29295A31317329297331317B29297B31318429298431318C292984
31318C29298431318C29298431318C29297B31298429297B31317B29296B31296B292963
31315A292942000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000029295231316331296331316B29296B
31317B31297331317B29297B31318431297B31318429298431318C31298431318C292984
31318C31298C31319431298C31319431298C31319429298C31319431318C31319431298C
31319431298C31318C31298C39319431318C31318C31298C31318C31298431318C312984
39318431297B31318431297331317B31297331317329296B31317331296B313163292952
31315A6329296B31316329296B31316B29316B31316B29297331316B29296B31316B2929
6B31316B29316B31316329296B31316329296B31316329296331315A29315A3131522929
52313131292931314229295231316B29296B31317B29297B31318429298431318C292984
31318C29318C31318C29318431398C3131843131843129295231314A2931523131522929
5A31315A29296331316329296B31316B29316B31316B29297331317329297B3131732929
7B31317329297B31317329297B31317B29297B31317329297B31317B29297B3131732929
7B31317B29297B31316B29297331316B29297331316B29297331316B29296B3131632929
6B3131313129314231294A29315A31315A29316331296B29317331317331317B31297B31
318439318431318C31318C31399439319431399C31318C31399431318431318431297B29
317B31317329316B31294A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294229295A29296329296B31297329297329297B29298431318C29298429298C
29298C31318C29298429318429298431318429297B29297B29297331297329296B292963
292952312942000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000031315231295A31316B
31316B31317331297331317331317B31317B31297B31317B313184313184312984313184
31318C31318C31298C31318C31318C31318C31318C31318C31318C31318C31318C31318C
31318C31318C31298C31318C31318C31318C31318C31318C31318C31318431298431317B
31317B31317B31317331317331317331317331296B31316B313163313152632931633131
6329316B31316329316B31316B31316B31316B29316B31316B31317331316B2931733131
6B31317331316B29316B31316B31316B31316B29316B31316B31316B3131632931633131
5A31315A313152293152313152313142313129297331297B31297B31318429298431298C
31318C31318C31318C3131314231315231315229315231315A31315A31315A2931633131
6B31316B31316B29316B31317331317331317329317331317331317B31317B29317B3131
7B31317B31317B29317B31317B31317B31317329297331317B31317B31317329297B3131
7B31317B31317329297331317331316B31316B29296B31316B29316B31316329296B3131
632931313131314229314A31315231315A31316329316B31317331317331317B31317B31
318431318C31318C31318C31319431399C31319431319431318C31318C31318429318431
317B31317331316B29315231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294231316329295A31316B29297331317B29297331318429298431318C292984
31318C29298431318C29298431318429297B31318429297B31317B29296B31316B29295A
312952292939000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31315A29296331317331296B31317331297331317B31297331317B29297B31318431297B
31318431298431318C31298431318C31298431318C31298431318C31298431318C312984
31318C31298431318C31298431318431298431318431298431318431297B31317B312973
31317B31297331317331296B31316B31295A31315A3129526331315A2929633131632929
6B31316329296B31316B29296B31316B29297331316B29297331316B29297331316B2929
7331316B29297331316B29297331316B29297331316B29297331316B2929733131632931
6B31316329296331315A29296331315A29295A31315229295A31315229295A3131522929
5A31315229295A31315229295A31315A29296331316329296B31316329297331316B2929
7331316B29297331317329297B31317329297B31317329298431317B29297B31317B2929
7B31317B29298431317B29298431317B29297B31317329297B31317329297B3131732929
7B31317329297B31316B29297331316329296B31316329296B31316329296B3131632929
633131293129314231314A29315A31295A29316B31316B29317331317331317B31317B31
318431318431398C39318C31399431319431399C39318C31398C31318429318431317B29
317B31316B29316B31314A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A29295A29296329296B29297329297B31317B29297B31318429298C31298C
29298431298429298431318C29298429297B29297B31297B29297331296B29296331295A
29294A292939000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000029293131316331296B31316B31316B31317331297331317331317331317B
31317B31317B31317B313184312984313184313184313184313184313184313184313184
31317B31317B31317B31317B31317B31317B31317B31317B31317B31317B313173313173
31317331316331315A31315A3129525231315229315A31315A2931633131632931633131
6329316B31316B29316B31316B29316B31316B29317331316B2931733131732931733131
732931733131733131733131732931733131732931733131732931733131733131733131
6B29317331316B31316B31316B31316B3131633131633131632931633131633131633131
6331316331316331316B31316B29316B31316B3131733131733131733131733131733131
7329317B31317B31317B31317B31317B31317B31318431317B29318431317B31317B3131
7B31318431318431318431318429318431317B31317B31317B29317B31317331317B3131
7B29297331317329317331316B29296B31316B29316331316329296B31316329295A3131
5A2929313931314231314A31315229315A31316B31316B31317331317331317B31318431
318431318C31318C31319431319431399C31319431319431318C31318431318431317B31
317331317331316331314A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31313129294A31295A29296331316B29297331297B29297331318429297B31318C292984
31318C29298431318429298431318429297B31317B29297331317B29296B312963292952
31314A292931000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029312931423131295231315A29295A31316B31296B31317329296B
31317B31297331317B29297331318431297B31317B29297331317B31297B31317B292973
31317B31297331317B31297331317B31297331317B29297331317B31296B313163292952
3131524229294A31314A29295231315229295A31315229295A31315A29296331315A2929
6B31316329316B31316329297331316B29317331316B29297331316B2931733131732929
7B31317329317B31317329297B31317329317B31317329297B3131732931733131732929
7B31317329317331317329297331317329317331316B29297331316B29317331316B2929
7331317329317331317329297B31317329317B31317329297B31317329317B3131732929
7B31317B29318431317B29298431318429318431318429298431318429318431317B2929
8431318429298431317B29298431317B29298431317B29298431317329297B3131732929
7331317329297331316B29296B31316329296331315A29296B31316329295A31315A2929
633131313129314A31294A29315A31315A29316B31296B29317331317331317B31297B31
398C39318431398C31318C31399C39319431399439318C31398C39318431318431297B29
317B31316B29316331000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
29293131295229295229296329296B31297329297B29297B29297B31318429298431318C
29298431298429298429298429298431318429297329297329297331296B29295A29295A
29294A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000313131313931314231294A31315A31316331316B31316B31313152
31315A31316331296B31317331317331317331297331317331317331317331296B31316B
31316B31316B31296331316331316331316331295A31315A313152393131392929422931
4229314A31314229294A29314A29315231315229315231315A29315A31315A2931633131
6331316B31316329316B31316B31317331316B2931733131733131733131732931733131
7331317B31317329317B31317B31317B31317B29317B31317B31317B31317329317B3131
7B31317B31317329317B31317331317B31317329317331317331317B31317329317B3131
7B31317B31317329317B31317B31317B31317B29317B31317B31317B31317B2931843131
8431318431318429318431318431318C3131842931843131843131843131842931843131
8431318C31318429318431318431318431318429317B31317B31317331317329297B3131
7331317331317329297331316331316B31315A29296331316329315A31315A29295A3131
5A2929313931314229314A31315231316331316B31316B31317331317B31317B31318431
318431318C31318C31319431319431399439318C31318C31318C31318431318429317B31
317331317331315229000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31293929294A31315A29296331297329296B31317B29297B31318429297B313184292984
31318C29298431318429297B31318429297B31317B29297331317329296B313163292952
313152000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000293129313931293929315231295229316331296331316B31296B31
317B31297331317B31317B3131843931295231315A31295231315229294A318431297B29
317B31296B29317331296329314A31312929393131312929393131312929393131392929
4231313929294A31314229294A31314A29295231315229295A31315229296331315A2929
6331316329296B31316329296B31316B29297331316B29297331317329297B3131732931
7B31317329297B31317329297B31317329297B31317B29317B31317B29297B31317B2931
7B31317B29297B31317B29317B31317B29297B31317B29317B31317B29297B31317B2931
7B31317B29298431317B29298431317B29298431317B29318431317B2929843131842929
8C31318429298C31318429298C31318429298C31318429298C31318429298C3131842929
8C31318429298431318429298431317B29298431317B29297B31317329297B31316B2929
7331316B29297331316B29296331315A29296331315A29295A31315229295A3131522929
313131293929314A31314A29315A31296329316B31316B31317331297331318431317B31
318C31318431399431319431399C39319431399439318C31318C31318429318431317329
317331296B29315A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
29294229295229295A31296B29296B29297329297B312984292984313184293184313184
29318C31318429298431298429298431298429297B31297329296B29296B292963312952
29294A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000313131293931314231314A31315231315A31316331316B31317331
317331317B31317B31317B31317B31318431318C31318C39318431318431318431318431
317B31317331316B31316B31294A31312931312929312931312929312931312931313131
3929313931314229314231314229314A31314A29315231315229315A31315A29315A3131
5A29316331316329316B31316B29316B31316B2931733131732931733131733131733131
7329317B31317B31317B31317B29317B31317B31317B31317B29318431317B3131843131
7B31318431318431318431317B31318431318431318431317B3131843131843131843131
8431318431318431318431318431318431318431318431318431318C31318C31318C3131
8C31318C31318C31318C31318C31318C31318C31318C31318C31318C31318C31318C3131
8431318431318431318431318431318431318431318431317B29317331317331316B3131
6B29316B31316B31316B31316329295A31315A29295A31315229294A31314A2929523131
313129313931314231315231315A31316331316B31317331317331317B31317B31318431
318431318C31318C31399439319431319439318C31318C31318431318431317B31317B31
317329316B31315231000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31314A29295231296329296331317329297331297B29297B313184292984313184292984
31318C29298431318C29297B31318429297B31317B29297331317329296B312963292952
31314A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000292929314231294231314A31294A29316331296331316B31296B31
317B39317B31318431297B31317B31317B31318431318431398C39318431318439297B31
318431297329316B31296329314A31312929313131292929313131292929313131292929
3931313129293931313929294A31314229294A31314A29295231315229315A31315A2929
6331315A29296331316329296B31316B29316B31316B2929733131732931733131732929
7B31317329317B31317329297B31317B29317B31317B29298431317B29318431317B2929
8431318429318431318429298C31318429318C31318429298C31318429318C3131842929
8C31318429318C31318429298C31318C29318C31318429299431318C29318C31318C2929
9431318C29319431318C29299431318C31318C31318C29298C31318C29318C3131842929
8C31318429318431317B29298431318429298431317B29297B31317329297331316B2929
6B31316329296B31315A29295A31315229295231314A29294A31314229294A31314A2929
313131313929314A31295229316331316329316B31296B29317B31317B31318431318429
398C31318C31399439319431399C39318C31318C39318431318C31318429317B31297329
317331316329315231000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
29294A31295A29296329296B29297331297B29297B292984292984313184292984313184
31318431318C29298431318429298431298429297B29297329297331296B29296329295A
29294A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000313131313931314231294A31315231315A31316331316B31316B31
317331317B39317B31317B31317B31317B31317B31318431318C31318C39318431317B31
317B31317B31296B29316331316331422931312929312931292931313131292929312931
3129313131313129313929313929314231314229314A31314A29315231315229315A3131
5A29316331316329316331316331316B31316B29316B31316B3131733131732931733131
7331317B31317B29317B31317B31317B31317B29317B31317B31318431317B2931843131
8431318431318429318431318431318C31318429318C31318C31318C31318C29318C3131
8C31318C31318C31318C31318C31318C31318C31318C31318C31319431318C3131943131
8C31319431318C31319431318C31319431318C31318C31318C31318C31318C29318C3131
8431318431318429318431318431318431317B29317B31317331317B31316329316B3131
6B31316331316329296331315231315231314A29294231314229294231314229294A2929
313131314231314A29315231316331316B31316B31317331317331317B31317B31318431
318C31318C31318C31319439319439319439318C31318C31318431318431317B29317331
317331316331315229000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31293929295231316329296B31317329297331318429298431318C29298431318C292984
31318C29318431318C29298431318429297B31318429297331317329296B31316B292952
31294A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313931294229314A31294A29315A31295A29316B31296B31
317339317331318439318431318431297B31318431297B31318C39318431318439317B31
317B312973293173312963293163314A29294A3131422929423131422929423131392929
3131312929293131313129293931313929294231314229294A31314A2929523131522929
5A31315A29296331316329296B31316329296B31316B29317331316B2929733131732929
7B31317329297B31317B29317B31317B29298431317B29318431317B2929843131842931
8431318429298C31318429318C31318429318C31318429318C31318C29318C31318C2931
9431318C29319431318C29319431318C29319431318C29319431318C29299431318C3131
9431318C29299431318C29299431318C29299431318C29318C31318C29298C3131842929
8C31318429298431318429298431317B29297B31317329297B31316B29296B3131632929
6331315A29296331314A29294A31314229294231313129294231314229294A31314A2929
313931293929315231315229316331296329316B31316B29317B31317B29318431318431
318C31318C31399439319431399439318C31399439318431318C31297B29317B31317329
317331295A29315231000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
29293929295229295A31296B29297329297329297B31298429298429298429298431318C
29318431318C29318431318429298431318429297B31297B29296B29296B292963312952
292942000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000293931314231314A31315231295231316331316B31317331
317331317B31317B31318439318431318431318431318431318431318C31318431318431
317B31317B31317331316B31315A315A29315A29295A29315229295229314A2929000000
0000000000003129313131313129313931313929314231314229314A31314A2931523131
5229315A31315A29316331316329316331316331316B31316B29316B31316B2931733131
7329317331317331317B31317B29317B31317B31318431317B3131843131843131843131
8431318431318431318C31318C31318C31318C31318C31318C31318C31318C31318C3131
8C31319431318C31319431318C3131943131943131943131943131943131943131943131
9431319431319431319431319431319431318C31318C31318C31318C31318C31318C3131
8431318431318431318431317B29317B31317B31317B31317331316331316331315A3131
5A29295A31314A29314231313929293131313129294231314229294A29315229295A3131
313929314231314A31315A31316331316B31316B31317331317B31317B31318431318C31
318C31319439319439319439319439318C39318C31318C31318431318431317B31317B31
317331316331000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294231315A29295A313173292973312973292973313184292984312984292984
31318C29298431318C29297B31318429297B31317B29297331317329296331296329294A
313142000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313931294229314A31294A29315A31295A31316B31296B31
317339317331318439298431318C39318431318431297B31318431298431318C31297B29
318431317B31317B31296B29316B312952296329315A29295A31315229294A2931000000
0000000000000000002929293131313129293931313929294231314229294A31314A2929
5231315229315A31315A29296331316329316331316329296B31316B29317331316B2929
7331317329317B31317329297B31317B29318431317B2929843131842931843131842931
8C31398431318C31318429318C31398C31318C31318C29319431398C31319431398C2931
9431398C31319431318C29319431399431319431319429319C31319431319C3131942929
9C31319431319431318C29299431318C29318C31318429298C31318429318C3131842929
8C31318429318431317329297B31317329297B31317329297331316329295A3131522929
5A31314A29293931313129294231314229294A31314A29295231315229296331315A2929
313931314229315231295A29316B31316329317331297329317B31317B29318431318431
399439318C31399439319439399439318C31318C31298429318C31317B29317B31297329
317331315A29000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294229295A29296329316B31317329297329297329297B312984292984292984
29298431318429298431318429317B31317B29297329297329296B31296329295A292952
292939000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313131314231294A31315231315A31316331296331316B31
317331317B31317B31318439318C39318C39318431318431318431318431318431318431
317B31317B31317331317331316331315A316329296329295A2929522931000000000000
0000000000000000000000000000003131313129313131313929314231314229314A3131
4A31315231315229315A31315A31316331316329316331316331316B31316B29316B3131
7331317331317329317331317B31317B31317B29317B3131843131843131843131843131
8C31318C31318C31318C31318C31319431318C3131943131943131943139943131943131
9431319431319431319431319431319C31319431319C31319C31319C31319C31319C3131
9431319431319431319431318C31318C31318C31318C31318C31318C3131842931843131
8431318431317B29317B31317331317331317329316B3131633131633131522929523131
4A31310000003129294229294229294A31314A29295229295A2929633131632929313131
313931314A31315229315A31316331316B31316B31317331317B31318431318431318C31
318C39319439319439319439319439319439318C31318C31318431318431317B29317331
316B31315A31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294231315229296331316B29296B31317B29297331317B29297B313184292984
31318429297B31318429297B31318429297331317329296B31317329295A31315A29294A
312939000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313931293929314A31295229315A31295A29316B31296B31
317331297331318439318431318C39318C31318C39318431318C39298431318C31298429
318431297B29317B31297329316B31295A296B31316329296331314A2929000000000000
000000000000000000000000000000000000000000292929393131312929423131422929
4A31314A29295231315229295A31315229296331315A29296331316329296B31316B2931
7331316B29297331317329317B31317329297B31317B29318431317B2931843131842931
8C31398429318C31398C31319431398C31319431399431319431399431319C3139943131
9C31399431319C31399431319C31319431319C31319431319C31319431299C3131943131
9431318C29299431318C29318C31318429298C31318429298C31318429298C3131842929
8431317329297B31317329297331316B29296B31316329296331314A29294A3131422929
0000000000003931313929294A31314A29295A31315A29296331316329296B3131293129
314231294A29315A31315A29316B31296B29317331317329317B31317B31318C39318431
319439318C39399C42319439319439318C31318C31318429318431317B29317B31317329
316B31295A29000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A29294A31296329296331316B29317331318429297B31297B29297B312984
29298429298429297B31317B29297B31317B29317331316B29296B29296B29295229294A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000293131314231314A31315231315A31316331316331316B31
317331317B31317B31318439318431318C39318C39318C39318C31318C31318431318431
318431317B31317B31317331316B31315A316B29296329315A29294A2931000000000000
000000000000000000000000000000000000000000000000292931313131312931393131
3929314231314A29315231315229315231315A29315A31315A29316331316331316B3131
6B29316B31317331317331317331317331317B31317B31317B31317B31317B3131843131
8431318431318431318C31318C31318C31318C3131943139943131943139943131943131
943131943131943131943131943131943131943131943131943131943131943131943131
8C31318C31318C31318C31318C31318C31318431318C3131843131843131843131843131
7B31317331317331317331316B31316B31316331316331314A2931423131000000000000
0000000000003129294229314A29295229295229295A31316329296B29316B2929313931
314229314A31315A31316331316331316B31317331317B31317B31318431318431318C39
318C39319439319439319439318C31318C31318C31318C31318431318431317B31317331
316B29315A31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A31295229295231316B29296B31317329297B31318429297B31317B29297B
31318429298431318429297331317B29297B31317B29296B31316B29296331295A292942
313142000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313931293929314A31295229315A31295A31316B31296B29
317331297331318439297B31318C39318C31319439318C31319439318C31318C31298429
318431297B29317B31297329317331295A293152316329295A3131422929000000000000
000000000000000000000000000000000000000000000000000000000000313131312929
3931313929294231314229295231315229315A31315229295A31315A2931633131632929
6B31316B29317331316B29297331317329317B31317329297B31397B29318431317B2931
8431398431318C31398429318C31398C31318C31398C29319431398C31319431398C2931
9431398C31319431318C29319431398C31319431318C29299431319431319431318C2929
9431318C29318C31318429298C31318429318C31318429298431317B2931843131732929
7B31317329317331316329296331315A2929633131422929000000000000000000000000
0000000000003131313929294A31314A29295A31315A29296B31316329296B3131293129
314A31314A29315A31295A29316B31316B29317331297329318431318431318C39318431
399439319439319439318C31399439318C31318C31298429318431317B29317B31297329
316B31315229000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A29295229295A29296B31317329297331317B31318431318429297B29297B
29297B31298429298429297B29297B31297B29297329297329296B31316329295A29294A
292939000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313131313931294231315231315A31316331296331316B31
317331317B31317B31318431318431318C39318C31319439318C39318C39318431318431
318431318431297331317331317331316331295231632931522929422931000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
3129313931313929314231314231314A31314A29315231315231315A31315A2931633131
6331316B31316B29316B31317331317331317329317331317B31317B31317B31317B3131
7B31318431318431318431318431318C31398431318C31318C31318C31398C31318C3131
8C31318C31318C31318C31318C31319431318C31318C31318C31319431318C31318C3131
8C31318C31318431318431318431318431318431317B31317B31317B31317B2931733131
7331317331316329316331315A31315A3131522929000000000000000000000000000000
0000000000002929293929294A29295231315229296329296329296B3131312929313931
314A31315231315A29316331316B31317331317329317B31317B31318431318431318C39
318C39319439319439319439318C31318C31318431318431318431317B31317329317331
316331000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294231315A29295A31316329296B31317B29297B31318429318431318C29297B
31318429297B31318429297B31317B29297331317B29297331317329296331315A292952
312939000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313131293129314231294A29315A31295A29316331296331
317331297331317B31297B31318439298431318C39318C31319439318C31318C39298431
318431297B29317B312973293173312963293152315A2929523131000000000000000000
000000000000000000000000000000000000000000000000000000000000313131292929
3131312931293931313129294231314229294A31314A29315231315229295A31315A2929
6331315A29296B31316329316B31316B29297331317329317B31317329317B31317B2931
7B31317B29318431397B29318431398429318C31398431318C31398429318C3139842931
8C31398429318C31398C31318C31318429318C31318C29318C31318C29318C3131843131
8C31318429298C31318429298431317B29297B31317329317B31317329297B31316B2929
7331316329296331315229295A3131422929000000000000000000000000000000000000
0000000000000000003929294A31314A29295A31315A29296B3131632929313131293929
314A31295229316331316329316B31296B29317B31317B29318431318431318C39318C31
319439318C31399439318C31319439318429318C31318429318431297B29317B31316B29
316331000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A29295A29296329296329297329297331317B29317B31318431318C31318C
29318431318429297B31298429297B29297B29297331297329297331296B29295A312952
292942000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000313931313931314A31295231315A31316331316B31
316B31317331317331317B31317B31318431318431318C39318C31318C39318C31318C31
318431318431317B31317331317331316B31315A316329314A2929000000000000000000
000000000000000000000000000000000000000000000000292929313131293129313931
2939293142312942293131313129313931314229314231314A29314A31315229315A3131
5A29315A31316331316331316329316B31316B31316B31316B31317331317331317B3131
7331317B31317B31317B31317B3131843131843131843139843131843131843131843131
8431318C31318431318C31318431318C31318431318C31318431318C31318C31318C3131
8431318431317B31317B31317B31317B31317331317331317331317331316B31316B3131
5A31315A31315231314A3131392931000000000000000000000000000000000000000000
0000000000000000003929314229295229315229296331316329296B2931313131314231
314A31315A31315A31316B31316B31317331317331317B31317B31318431318C31318C39
318C31319439319431319431318C31318C31318431318431317B29317B31317331316B31
315A29000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31314229294A31295A29295A31316B29296B31317B29297B31318429318431398C29318C
31318C29298431318429297B31318429297B31317B29297331317329296B312963292952
313152000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000293129314231294229315231295A31316331296329
317331297331317B31297329317B31297B31318431298431319439318C31318C39298431
318C39298431318431297329317331296B293163312942294A3131000000000000000000
000000000000000000000000000000000000000000292929313931293129313931293929
314A31294229314A31294A293152313129293931313929294231314229314A31314A2929
5A31315229315A31315A29296331316329316B31316329296B31316B29317331316B2931
7331397329317B31317329317B31397B29318431397B29318431398431318431397B2931
8431398429318431317B29318431398431318431317B29298431318431318431317B2929
8431317B29317B31317329297B31317329317331316B29297331316B29316B3131522929
5A3131422929393131292929000000000000000000000000000000000000000000000000
0000000000000000003129294A31314A29295A31315A29296B3131292929313931294229
315231315229316331296329317331317329317B31297B29318431318431318C39318C31
399439318C31319439318C31318C31318429318C31298429318431317B29317331296329
315231000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
29294231295229295A29296329296B31297329297329297B31318431318C31318C313194
31318C31318C29318431318429298431298429297B29297B29297331297329296329295A
292952000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000313131293931314231315231315A31295A31316331
316B31317331297331317B31317B31317B31318431318431318C31318C31318C31318C31
318431318431297B31317B31317331317331296331315231422929000000000000000000
000000000000000000000000292929313131293129293931293929314231294229294231
294A29314A31294A29315231295229315A31295A293131313931313931314229314A3131
4A31315231315229315A31315A31316331316329316331316331316B31316B29316B3131
6B31317331317331317331317331317B31317B31317B31317B31317B31317B31317B3131
7B31317B31317B31317B31317B31317B31317B31317B31317B31317B31317B31317B3131
7B31317B31317329317331317331317331316B2931633131633131633131522931523131
393131314231293929313929293129000000000000000000000000000000000000000000
0000000000000000003129294229295231315229295A2929632929313131313929314A31
315231315A31316329316B31316B31317331317331317B31318431318431318C31318C39
318C31319439318C31318C31318C31318C31318429318431317B31317B31316B29316331
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31293129294A31316329296331297329297331317B29297B313184292984313994293194
31399429318C31319429298431318C29298431318429297331317329296B31316B29295A
313152000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000292929313931294229315231295229316331296331
316B31296B31317B31297331317B31297B31318431298431318C31298C31318C39298431
318C31298431318431297B29317B31296B29316B31295A29314A31000000000000000000
000000000000313131293129313931293129313931293929314231294229314A31294A29
315231295229315A31295229315A31295A29316331296329316B31312929393131392929
4231314229294A31314A29315A31315229296331315A29316331315A29316B3131632931
6B31316329317331316B29317331316B29317B31397329317B31397329317B3139732931
7B31317329317B31317329317B31317329297B31317329317B31317329297B3131732931
7331316B29297331316B29316B31316329296331315A29295A3131522929393131294229
314A31294229314231293929313931293129313931000000000000000000000000000000
0000000000000000003129294231314A29295A31315A2929313131293129314231294A29
315231295A29316331316B29317331297329317B31317B31318431318431318C39318C31
319439318C31319439318C31318C31298429318C31317B29318431297329317331315A29
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A29295A31296329296B29297329297331297B29297B31318431318C31318C
31319431319431318C31318C29318431318429297B31297B29297329296B29296B29295A
29294A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000293931314231294A31315231315A31316331
296331316B31317331317B31317B31317B31317B31318431318431318C31318C31318C31
318431318431318431318431317331317331316B31316331295231000000000000000000
292929313131293129313931293929313931294229314231294229314A31294A29315231
295229315A31295A29315A31295A29316331296329316331296329316B31296B29313131
3129313931314231314231314A29315231315231315A31315A29315A31315A3131633131
6331316331316331316B31316B31316B31316B3131733131733131733131733131733131
733131733131733131733131733131733131733131733131733131733131733131733131
6B31316B31316331316331315A31315A31315231315231314A3131314A31294A29314A31
294229314231293929313931293929313931293929313931293129313131000000000000
0000000000000000003129313929294A2929522929632931313129313931314A31315231
315231315A31316331316B31317331317331317B31318431318431318C31318C31318C31
318C31319431318C31318C31318C31318C31318431318431317B31317331316B31315A31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294231295229295A31316B29296B31317B29297B31318429297B313184292984
31399431319431399429318431318C29298431318429297331317329296B31296B29295A
313152292939000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000313931293929314A31294A29315A31296329
316B31296B31317331297331317B31297B31318431297B29318431298431318C31298431
318C31298431318431297B29317B31297329317331296329315A31294229313131293129
313931293129313931293929314231294229314A31294A29315231295229315A31295229
315A31295A29316331295A29316331296329316B31296329316B31296B29316B31296B29
3173312973293931313129294231314229314A31314A29295231315229315A3131522929
5A31315A29316331315A29296B31316329316B31316B29317331316B29316B31316B2929
7331316B29316B31316B29297331316B29316B31316B29297331316B29316B3131632929
6331315A29315A31314A29295231314A2931423131294A29315231294A29314A31294229
314A31294229314A31294229314231293929314231293929313931293129313931293129
3131310000000000002929294231314A29295A3131292929313931294229314A31294A29
315A31316329316B31296B29317331317331317B31297B31318439318431318C31298431
319439318C31318C31298429318C31318429318431297B29317B31317329316B31294A29
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294229294A29295A29296B31317329297329317B29298431318429297B292984
31318C31318C31318C31318C31318431318429297B29297B29297331296B292963292963
29294A312942000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000293131314231314A31315231295A31316331
316B31316B31296B31317331317B31317B31297B31318431318431318431318C31318C31
318431318431298431318431317B31317B31297331316B31316331315231293929313931
313931314231294229314231294A31314A31294A29315231315231315A31295A29315A31
295A29316331296329316331296331316B31296329316B31296B29316B31296B29297329
2973293173312973293173313131313931313929314231314231314A31314A29314A3131
5231315231315229315A31315A3131633131632931633131633131633131632931633131
6331316331316331316331316331316331315A29316331316331316331315A2931523131
5231315231314A2931393131315A29315A31295229315231295229315231294A29314A29
314A29314A31294229314231314229314231293929313931313929313931293129313131
3131293131312929293131313929294A3131312929313131313931314A31314A29315231
315A31316331316B31317331317331317B31317B31318431318431318C31318431318C31
318C31319431318C31318C31318C31318431318429317B31317331317331316329000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029293931315229295231316B29296B31317329297331318429298431318429297B
31318429298431319429318C31318C29297B31318429297B31317B29296B31316329295A
31295A292939000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000313131293929314A31294A29315A31295A29
316B31296B29317331297331317B31297B31318431297B31318431298429318C31298431
318C31298429318431297B29318431297329317331296B29316B31295229315231294229
314231294229314A31294A29315231294A29315231295229315A31295229316331295A29
316331295A29316B31296329316B31296B29317331296B29317331296B29317331297329
317B31297329317B31297329317B31297329317B31312929393131392929423131392931
4231314229294A31314A29315231315229295A31315229315A31315229295A3131522931
5A31315229295A31315229315A31315229295A3131522929523131522929523131422931
39313129294A313142295A29315A31295229315A31295229315231294A29315231294A29
315231294A29314A31294229314A31294229314231293929314231293929313931293929
313931293129313131292929313131292929313131293129314231294229315231295229
316331296329317331316B29317B31297329318431317B31318431298431318C31318431
318C31298C29318C31318429318C31298429318431317B29317B31296B29315A31000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029293929295229295A29296329297329297331317B29297B313184292984313184
29297B31318429318431318C29318431318429317B31317B29297329297329296331295A
292952292939000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000313931294231314A31315231315A31
316331316B31316B31317331297331317B31317B31318431318431318431318431318C31
318431318431318431318431317B31317B31317331317331296B31316331295A31315231
294A31314A31294A31315231295229315231295229315A31295A29315A31296331316331
296329316331296B29316B31296B29316B31296B29317331296B29317331297329317331
297329317331297329317331297329317331297329317B31297329317331297329313131
3129313931313931314231314231314A31314A31314A31314A29314A31314A31314A3131
4A31314A31314A31314A31314A29314A31314231314A313142313131317331296B31316B
29295A31314A312942315A31295A29315A31295229315231295229315231295229315231
294A29314A31314A29314A31294229314231314229314231314229314231313929313931
293129313131313129313131313129313131313931313931314229314A31315231315A31
316331316B31316B31317331317331317B31317B31318431318431318431318431318C31
318C31318C31318C31318C31318431318431318431317B31317329316B31314A31000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029293931294A29295231316329296B31317329297331318429318431318C292984
31318429297B31318429298431318C29298431318429297B31317B29297331296B29295A
31315A29294A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000293129314231294A29315231295A29
316B31296B31317331296B29317B31297B31318431297B31318431298431318C31298429
318C31298431318431297B29318431297B29317B31297329317331296329316331295A29
315A31295229315A31295229315A31295A29315A31295A29316331296329316331296329
316B31296B29316B31296B29317331296B29317331296B29317331296B29317331296B29
317B31297329317B31297329317B31297329317B31297329317331297329317331296B29
317331296B29316B31292952313163312931393131392929423131392931393131312929
39313131293131399429318C31319429318C31318C29297B31317B292973313173292963
31316329294A313139295229315A31295229315A31295229315A31295229315231294A29
315231294A29314A31294229314A31294229314A31294229314231293929314231293929
313931293129313931292929313131293929314231294229314A31314A29315A31295A29
316B31316B29317331297329317B31317B31318431297B31318C31318431318C31298429
318C31318C31318C31298429318C31318429318431297329317331316329000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029293929294A29295A29295A31296B29297329297B29317B31318429318431318C
29318431318429298429298429318431318C29298431317B29297B31317329296B292963
29295A312952292931000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000313131314231314A31295231315A31
316331316B31296B31317331317331317B31297B31318431318431318431298431318431
318431318C31298431318431318431318431297B31317331317331316B31296331316331
315A31315A31295A29315A31315A31316331295A29316331316331316B31296329316B31
316B31316B31296B29316B31296B31317331296B29317331297329317331297329317331
297329317331297329297329297329317331297329297329296B29317331296B29296B29
296B29316B3129293931295231296331317329297331297B31297B31318431318431318C
31318C31319431319431319431318C31318C29298431318431317B31317329296B313163
31295A31314A292931315A31295A29315A31295229315231315229315231295229315231
315229315231294A29314A31314A31314A31294A29314A31314231314231293929313931
313931313931293129313131313931314231314229314A31314A31315231315A29316331
316B31316B31317329317331317B31318431317B31318431318431318C31318431318C31
318C31318C31318431318431318431318431317B29317331316B31315231000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031314A29295A31296329296331317329297331317B29297B31318C29318C
31318C29298431318C29298431318429298431318429297B31317B292973313173292963
313163292952313142000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000292931292929313931294229315231295229
316331296329316B31296B31317331297331318431297B31318431297B31318431298431
318C31298429318C31298429318431297B29318431297329317331296B29316B31296329
316B31295A29316331295A29316331295A29316331296329316B31296329316B31296B29
316B31296B29317331296B29317331296B29317331296B29317331296B29317331297329
317331297329317B31297329317331296B29317331296B29316B31296329316B31295A29
315A3129522931314A29295231316B29296B31317B29297B31318429298431318C29318C
31319429319431399C29318C31319429298431318429297B31317B29296B313173292963
31315A29294A313139295229315A31295229315231295229315231295229315231294A29
315231294A29315231294A29315231294A29314A31294229314A31294229314231293929
314231293929313931293929314231294229314A31294A29315231295229316331296329
316B31296B29317331297329317B31297B29318431318431318431298429318C31318429
318C31298429318C31318429318431297B29317B31296B29316331000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000029294A29295A29296329296B29296B29297B29297B31318429318431318C
29318C31318C29318C31318C29298429298429298431318429297B29297329297329296B
29296329295229294A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000292929292931292931292939292931313931314A31315231
295A31316331316B31316B31297331317331317B31317B31297B31318431318431318431
318431318431318431318431298431318431317B31317B31297331317331317331316B31
296B31316331296331316331296331316331296331316B31296B29316B31296B31316B31
296B29316B31296B29317331296B29317331297329317331297329317331297329317331
297329317331297329317331296B29316B31296B29316331296329295A29295229294A31
00000000000029294A31315A29296B31317331297331317B29298431318431318C31318C
31319431319431319431319431318C31318C31318431318431297331317331296B31316B
29295231314A312931315A31295229315231295229315231295229315231315229315231
295229315231314A29315231294A29315231314A29314A31294229314231314229314231
314229314231314231314231314229314A31314A31315231315231315A31316331316B31
316B31317331317331317B31317B31318431318431318431318431318431318431318C31
318431318431318431318431317B31317B31317331316B31315229000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031294A29295231316329296331297329297331318429297B313184293184
31399429318C31319429298431318C29298431318429297B31318429297331297329296B
31316329295231294A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000029293129212931293929293929294229294231294A293129314A31294A29
315A31295A31316B31296329317331297331317B31297329318431297B31318431297B29
318431298431318C31298429318431297B31318431297B29317B31297329317331296B29
317331296B29316B31296329316B31296329316B31296B29317331296B29317331296B29
317331296B29317331296B29317331297329317331296B29317331296B29317331296B29
317331296B29317331296B29316B31296329316331295229315231000000000000000000
00000000000031315229295A31316B29297331317B29297B31318429298431318C29298C
31319431319431319C29298C31319429298431318429297B31317B29297331316B292963
31315A29294A315A31295229315A31295229315A31295229315231295229315231294A29
315231294A29315231294A29315231294A29314A31294229314A31294229314A31294229
314A31294229314A31294229314A31314A29315231295229316331295A29316331296329
317331317329317B31297329318431317B31318431297B29318C31318431318C31298429
318C31318429318431297B29318431317329317331295A29000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000029294229295229296331296B29296B29297329297B31318429298429318C
31318C31399429318C31318C29318C31318C29298429298429297B31297B29297329296B
29296B31295A29294A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
29292929293129293129293929293929294229294229294A29294A313131293929314A31
315231315A31296331316B31316B31317331297331317B31317B31318431297B31318431
318431318431298431318431318431318431297B31317B31317B31317B31297329317331
317331316B31296B29316B31316B31316B31296B29316B31316B31317331296B29317331
297331317331297329317331297331317331296B29317331296B29316B31296B29296B29
296B29316B31296B29296B31295A29315A31000000000000000000000000000000000000
00000000000029295231296331296B31317329297331317B31298431318C29298C31318C
31319431319C31319431319431318C31318C29298431317B31317331317329296B313163
31295231314A295229315229315229315231294A29315231315229315231295229315231
315231315231295229315231315231315231294A29314A31314A31314A31294A29314A31
314A31314A31294A29314A31314A31315231315229315A31316331316331316329316B31
316B31317331317331317B31317B31318431317B31318431318431318C31318431318431
318431318431317B29317B31317B31317331316B29315231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031314A29295231316329296331317329297331297B29297B31318C292984
31319429318C31399429318C31319429298431318C29298431318429297331317329296B
31316B29295A313152292942000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000312931292929
29293929293131294229294231294A29294A31295229295229295A292952313931294229
315231295229316331296329316B31296B29317331297329317B31297B31318431297B31
318431297B29318431298429318431297B29318431297B29317B31297B29317B31297329
317331296B29317331296B29317331296B29317331296B29317331296B29317331297329
317331296B29317331296B29317331296B29317331296B29316B31296B29316B31296329
316B31296329316331295229000000000000000000000000000000000000000000000000
00000029293931315A29296331317329297331317B29297B31318429298431318C29298C
31319429319431319429318C31318C29298431318429297B31317B29296B31316B29295A
313152292942315231294A29315231294A29315231294A29315231294A29315231294A29
315231295229315231295229315231294A29315231294A29315231294A29315231294A29
315231294A29315231294A29315231295229315A31295A29316331296329316B31296B29
317331297329317B31317B29318431297B29318431298429318C31298429318C31318429
318431297B29318431297329317B31296B29316331000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000029294229295229296329296B29296B31297329297B29298429298431318C
29318C31318C29319431319429318C31318C29318431318429298429297B292973312973
29296B292963292952292942000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000292931292931292939
29293929294229294229294A29294A29295229295229295A29295A29295A392929313931
294A31315231315A31316331296331316B31316B31317331317331317B31317B31318431
297B31318431318431318431318431318431318431318431297B31317B31317B31317B31
297331317331317331317331297329317331297331317331297331317331297331317331
297329317331296B29317331296B29316B31296B29316B31296B29316B31296329316331
296329315A31295229000000000000000000000000000000000000000000000000000000
00000031294229295A31316329296B31317331297B31318429298431318C31318C313194
31319431319431319431318C31318431318431297B31317B31297331317331296B313163
292952312939294A29314A31294A29314A31294A29315231295229315231295229315231
295229315231315229315231295229315231315229315231315229315231315231315231
314A29315231315231315231315231315A31315A31316331316331316B31316B31317331
317331317B31317B31317B31317B31318431318431318431318431318431318431318431
318431318431317B31317B31317331316B31315231000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029294A31315A29296331296B29296B31317B29297B313184292984
31318C29318C31319429318C31399429318C31318C29298431318429297B31297B29296B
31316B29296331295A292942000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000292929312939292939292942292942
31314A29294229294A29294A31295229295229295A292952313163292929392931293129
314231294A29315231295A29316B31296331317331296B29317B31297331317B31297B29
318431297B31318431297B29318431298429318431297B29318431297B29317B31297329
317B31297329317B31297329317B31297329317331296B29317331297329317331296B29
317331296B29317331296B29317331296B29316B31296329316331295A29315A31295229
315231292929000000000000000000000000000000000000000000000000000000000000
00000029294231316329296331317329297331317B29297B31318C29298431318C29298C
31319429319431319429298C31318C29298431318429297331317B29296B31316B29295A
313152294A29314A31294229314A31294A29315231294A29315231294A29315231294A29
315231294A29315231295229315A31295229315231294A29315231295229315231294A29
315231295229315A31295229315A31295A29316331296329316B31316329317331296B29
317B31317329317B31297B29318431317B29318431298429318C31318429318431297B29
318431317B29317B31296B29317331315A29000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029294A29295231296329296B29297329297331297B29297B292984
29298431318C29318C31319429318C31318C293184293184292984312984292973292973
29296B31296B29295A29294A292939000000000000000000000000000000000000000000
00000000000000000000000029292929293129293129293929293929294229294229294A
29294A29295229295229295229295229295A29295A29295A2929633129313929294A2929
313931314A31294A31315A31316331316B31296B31316B31317331317B31297B31317B31
317B31318431297B31318431318431318431298431318431317B31317B31297B31317B31
317B31317B31297329317B31317331317331297329317331317331317331297329317331
317331317331296B29316B31296B31316B31296329296331295A29315A31295229292942
292939292931292929000000000000000000000000000000000000000000000000000000
29293131314A29295A31296B31297331317B29297B31298431298431318C29298C313194
31319431319429318C31318C31318431318429297B31317B31297331316B29296331295A
31294A314A31294229314A31294229314A31294229314A31314A31315231294A29314A31
314A31315231295229315231315231315231295229315231315231315231295229315231
315231315A31315A29315A31315A31316331316329316331316B31316B31316B29317331
317331317B31317B29317B31317B31318431318429318431318431318431317B29318431
317B31317B31317329317331316B31315231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029293931295229295A31316B29296B31317329297331318429297B
31318429298431318C29298C31319429318C31318C29298431318429297B31317B292973
31317329296331316329295231294A000000000000000000000000000000000000000000
00000000000031293129293131293929293931294229293931294A29294231294A29294A
31295229295231295A29295A31296329295A31296329296331296B292929423131422929
313131293929314A31295229315A31296329316B31296B29317331297329317B31297B29
318431297B29318431297B29318431297B29318431297B29318431297B29318431297B29
317B31297329317B31297329317B31297329317B31297329317331297329317331296B29
317331296B29317331296B29316B31296329316331295A29315A31292942313142292939
313139292931313139292931000000000000000000000000000000000000000000000000
31314229295231316329296331317329297331318429297B31318C29298431319429298C
31319429298C31319429298431318C29297B31318429297331317329296B313163292952
31314A000000314231293929314231293929314231294229315231294A29315231294A29
315231294A29315231295229315231295229315A31295229315A31295229315A31295229
315A31295A29316331295A29316331295A29316B31296329316B31296B29317331297329
317B31297329317B31297B29318431297B29318431297B29318431297B29318431297B29
318431297329317331296B29315A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029293129294A29295A29296B29296B29297329297B29297B292984
29298429298429298431318C29318C31318C29318C31318C29298429318429297B29297B
29296B29296B29296329295A29294A292942000000000000000000000000000000292931
29293129293129293129293929293929294229294229294A29294A29294A292952292952
29295229295A29295A29295A29295A2929632929632929632929633129313929294A2931
522929313931294231315231295231316331316331316B31296B31317331317331317B31
297B31317B31317B31317B31297B31317B31317B31318431297B31317B31317B31317B31
297B31317B31297B31317B31297331317331297331317331297329317331297331317331
296B29316B31296329316331296329315A31295A2929294A29294A292942292942292942
292942292939292939292939292931292931292929000000000000000000000000392929
29294231315A29296331316B29297331317B31297B31318429298431318C31318C313194
31318C31319431318C31318C29298431318431297B31317B29297331316B312963313152
292939000000000000313131293129313931294229314231294A29314A31294A29315231
294A29315231315229315231295229315231315231315A31315229315A31315A31315A31
315A29315A31315A31316331316329316331316331316B31316B31317331317331317331
317331317B31317B31317B31317B31318431317B31318431317B31317B31317B31317B31
317B31317331316B31316331000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031315229295231296329296B31317329297331297B29297B
31318429298431318429298431318C29318C31318C29298431318C29297B313184292973
31317329296B31296B29295A31315A292942312939292929313131292929292931292931
31313929293131293929293931314229294229294A29294A31315229295231295A292952
31315A29295A29296329295A31316329296329296329296331296B312929392931422929
5A31315A2929313931294A29315231295A29316331296329317331296B31317331297329
317B31297329317B31297329318431297B29317B31297B29318431297B29317B31297329
317B31297B29317B31297329317B31297329317331296B29317331296B29317331296B29
316B31296329316331295A29316331294A2931295A29294A31315229294A31294A292942
31314A2929423131422929313131422929313129392929293131314A2929522929292931
31315229295A31316329296B31317B29297B31318429297B31318C29298C31318C29298C
31319429318C31318C29298431318429297B31317B29297331317329296331316329294A
313139000000000000000000000000000000000000293129314A31294229314A31294A29
315231294A29315231295229315A31295229315A31295229315A31295229315A31295A29
316331295A29316331295A29316331296329316B31296329317331296B29317331297329
317B31317329317B31297329318431297B29317B31297329317B31317B29317B31297329
317331296B29315A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000029294231295229295A29296B29297331297329297B292984
29298431298429298429298429298431318C29298C29318429318431318429297B29297B
29297331297329296B29296329295A312952292942292939292931292931292931292939
29293929294229294229294229294A29294A29294A29295229295229295A29295229295A
29295A29296329295A29296329296329296B29296329296B29296B312931312929422929
5229295A31315A2929313131314A31315231295231316331316331316B31296B31317331
317331317331297331317B31317B31317B31297B31317B31317B31317B31297B29317B31
317B31317B31297331317331317331317331297329317331317331317331296B29316B31
296331316331295A2929523129295A31295A29295229295229295231295229294A29294A
292942312942292939292939292939312939292931292931292931312931292929312942
29295231316329296331297331297331317B29297B31318431298431318C29298C31318C
31318C31319429298C31318431318431318429297B31297331296B31316B29295A31294A
000000000000000000000000000000000000000000000000293929314231294229314231
314A31314A31294A29315231315231315231295229315231315231315A31295A29315A31
315A31315A31295A29316331316331316B31316B29316B31316B31317331317329317331
317331317B31317329317B31317B31317B31317329317B31317B31317B31317329316B31
316331315A31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031293929295231316329296331297329297331317B29297B
31318429298431318C29298431318C29298431318C29298431318C29297B31318429297B
31317B29296B31317329296331316329295231314A292942312942292939312942292939
31294229294231314A29294A31295229294A31295229295231295A29295231316329295A
31296329295A31296329296331296B29296331296B29296B31296B292929393131392929
5231315229296331316329296B3131294229315231295229316331296329316B31296329
317331296B29317331297329317B31297329317B31297329317B31297329317B31297329
317B31297329317B31297329317331296B29317331296B29316B31296329316B31296329
316331295229314A3129295A31315A29295231315A29295231315A29295231315229294A
31314A29294231314A29294231314229293931313929293131313129292931313929294A
31315A29296331316B29297331317B29297B31318429298431318C29298431318C29298C
31318C29298C31318C29298431318429297B31317B29297331317329296331315A292942
000000000000000000000000000000000000000000000000000000293929314231293929
314231294229314A31294A29315231294A29315231295229315A31295229315A31295A29
316331295A29316331296329316B31296329316B31296B29317331296B29317331297329
317331297329317B31297329317B31297329317B31297329317B31296B29316B31296329
315231422929000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000029295229295A29296329296B29297329297329297B
29298429298429298431298429298429298C29298431318C29298429298429297B31297B
29297B29297329296B29296B29296329296329295229294A292942292942292942292942
29294229294A29294A29295229295229295229295229295A29295A29295A29295A292963
29296329296329296329296B29296329296B29296B29296B29296B29296B392929422931
4229295229315A29316331316329316B3131294231314A31295231315A31316331316331
296331316B31296B31317331297331317331317331317331297331317331297331317331
297331317331297331317331297329317331296B31316B31296329316331296331315A31
294A2929296329295A29295A29295A29295A29295A31295A292952292952292952292952
29294A29294A29294A312942292942292939292931312931292931312931292939313152
29295A31296329296B31317329297B31317B31298431318429298431318C31298C31318C
29298C31318C31318431318429298431317B31297B31317329296B313163312952313142
000000000000000000000000000000000000000000000000000000000000293929313931
294229314231314A29314A31294A29315231315229315231315229315A31315A31315A31
315A29316331316331316331316331316B31316B31316B31316B29317331317331317331
317331317331317331317331317329317331317331317331316B31316B31316331314A31
522929422929000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000029294231295A29295A31317329297331297B292973
31318429298431318C29298431318C29298431318C29298431318C29298431318429297B
31317B29297331297329296B31316B29296331295A29295231315229294A31294A29294A
31315229294A31295229295231315A29295231295A29295A31316329295A292963292963
31316B29296331296B29296331316B29296329296B29296331316B29296B392931392929
4A31314A29295A29315A29296B31316B2929733131293929314A31294A29315A31295A29
316331296329316B31296329317331296B29317331297329317331297329317331296B29
317331296B29317331296B29317331296329316B31295A29316331295229315231292963
31316B29296331296329295A31316329295A31295A29295231315A29295231315229294A
31315229294A31314A29294231314229293931313929293931313929293931314A292952
31316329296B31317329297331317B29297B31318429298431318C29298431318C292984
31318C29298C31318C29298431318429297B31317B29296B31316B29295A313152292931
000000000000000000000000000000000000000000000000000000000000000000000000
313931293929314231294229315231294A29315231295229315A31295229315A31295A29
316331295A29316331295A29316331296329316B31296329317331296B29317331296B29
317331297329317331296B29317331296B29317331296329316B31295A29314231632929
523131000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000029293929295229296329296B31297329297329297B
29297B31298429298429298429298431318C29298429298C29298431318429298429297B
29297B31297B29297329297329296B31296329295A29295A29295A31295229294A292952
29295229295A29295229295A29295A29295A29295A292963292963292963292963292963
29296329296B29296329296B29296B29296B29296B29296B29296B29296B312929422929
4229294A29315A29296329316331316B31316B2931733131293131314231294A29315231
315A31316331296331316331316B31316B31296B31317331316B31317331296B29316B31
316B31316B31296329316B31316331316331295A29315231294A3129296B292963292963
29296331296329295A29295A29295A31295A29295A29295A29295A31295A292952292952
29295231295229294A29294A29294A31314229294229293929294231314A29295231295A
31296331316B29297331297B31297B31318429298431298431298431318C29298431318C
31318C31318C29298431318431297B31317B29297331316B31296331315A292942392929
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000293129313931314231314A31294A29314A31315231315231295229315A31
315A31316331295A29316331316331316331296329316B31316B31316B31316B29317331
317331317331316B29316B31316B31316B31316329316331315A313142316B2929632929
522929000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000031315229295A31316B29296B31317B292973
31318429297B31318C29298431318C29298431318C29298431318C29298431318429297B
31318429297B31317B29297331317329296331316B29295A31316329295A31315A292952
31295A29295231315A29295A31296329295A31316329296331296B29296331296B292963
31296B29296331296B29296331296B29296B31296B29296B31296B29296B313131392929
4A31314A29295A31315A29296B31316329297331317329317B3131732929314231294229
315231295229316331295A29316331296329316B31296329316B31296329316B31296329
316B31296329316331295A29315A3129522931296329295A31296329295A31316B292963
31296329295A31316329295A31316329295A31316329295A31315A29295231315A292952
31315A29295231315229294A31314A29294231314A29294231314A29295231315A29295A
31316B29296B31317B29297331318429297B31318C29298431318C29298431318C292984
31318C29298431318429297B31318429297331317329296B313163292952313142522929
4A2929392929000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313931293929314A31294229314A31294A29315231295229
315A31295A29316331295A29316331295A29316B31296329316B31296329316B31296329
316B31296B29316B31296B29316B312963293163312952297B31316B29296B3131632929
523131000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000029293129295A29296329296B29297329297B
29297B29297B292984292984292984292984292984292984292984292984292984292984
29297B29297B29297B29297B29297329296B29296B29296B29296329296329295A292963
29295A29296329295A29296329296329296329296329296329296329296B29296329296B
29296B29296B29296B29296B29296B29296B29296B29296B29296329296B312929392931
4229294A29315229296329316329296B31316B31317331317331317B31317B29317B2931
7B2929314231294A31315A31295231315A31295231315A31295231315A31295A31315A31
295A29315A31294A3129294A29295229295A29295A29295A29295A29295A292963292963
29295A29295A29295A29296329295A29296329295A29295A29295A29295A29295A31295A
29295229295229295231295229294A31294A29294A31294A29295231295A29295A313163
29296B31317329297331317B29297B31318429298431318429298431318C31298431318C
29298431318431298431318429297B31317331296B31316B29295A31314A5A29295A2929
5229294A2929392929000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000314231294229314A31314A29315231
315229315A31315A31315A31315A29316331315A31316331315A29316331316331316331
316329316331316331316331315A29315A313152317B31317329297329296B2929633131
4A2929000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000029294231315A292963312973292973
31317B29297B31298429297B31318C29298431298429298431318C29298431318429297B
31318429297B31297B29297331317B29296B31297329296B31316B29296331296329295A
31316329296331296329296331316B29296331296B29296331316B29296331296B292963
31316B29296B29296B29296331316B29296331296B29296331316B292963312931392929
4A31314A29295229315229296B31316329296B31316B29297B39317B31317B3131732929
7B31317B29298431317329297B3131732929733131632929314231294229314A31294229
00000000000000000000000000000029294A29295229294A31315A29295A31296329295A
31316329295A31295A29295231316329295A31316329295A31315A29295A31315A292952
31315A29295231315229295231315229294A31315229295231315A29295A313163292963
31317329297331317B29297331318429298431318429298431318C29298431318C292984
31318C29298431318429297B31317B29297331316B29295A31315A2929396329295A2929
6329314A29294A2929422929000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000293129313931293929
314A31294A29315231295229315A31295229315A31295229315A31295A29315A31295A29
315A31294A29314A317B31318431397B29317B31317329297B31316B29296B31315A2929
523131000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000029295229295A292963292973
29297331297B29297B292984292984312984292984292984292984312984292984292984
29298431298429297B29297B29297331297329297329297329296B29296B292963292963
29296331296B29296329296B29296329296B29296329296B29296B29296B29296B29296B
29296B29296B29296B29296B29296B29296B292963292963292963292963292929392929
4229295229315229295A29316329316B31316B29317331317331317B31317B31317B3131
7B31317B31317B29297B29317B29317B29317329296B29315A2929000000000000000000
00000000000000000000000000000000000000000029294229294A29295A29295A29295A
29295A29295A29295229295A29295A31296329295A29295A29295A31295A29295A29295A
29295A31315A29295229295A29295231315A29295229295A29295A31316329296331296B
29297331317329297331297B31298431318429298431298431298431318C292984313184
31318431318429297B31317B31297331317329296331295A31294A632929632929632929
5A29295A2929522929522929522929422929000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000312931293929314A31314A31315231294A29315231315231314A312942297B3131
7B31318431398431318431317B31317B31317B29297B29297329297331316329295A2929
4A2929000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031293929295231316329296B
31317329297331317B29297B312984292984313184292984313184292984313184292984
31318429297B31318429297B31317B29297331317B29297331297329296B31316B292963
31316B29296B31316B29296331296B29296331316B29296B31296B29296B31316B29296B
31296B29296331296B29296331296B29296331316B29296329296329295A313131312929
4231314A29295A31315A29296331316329297331316B29297B31317331318439397B3131
8431317B29298431317B29297B31317329297331316B2929632931000000000000000000
000000000000000000000000000000000000000000000000000000292939313152292952
31295A29295231315A29295A31316329295A31316329295A31316329295A31316329295A
31316329295A31315A29295A31315A29295231315A29295A31316329296331317329296B
31317329297331317B29297B31318429298431318429298431318C29298431318C292984
31318429297B31318429297331317329296B31316329295231314A6329296B2931632929
6B29296329296329315A29295A29315229295229314A2929000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003129294A31315229296331316329296B31316B29297331317329297B31317B2931
8431397B31318439397B31318431317B29297B31317329297331316B29296B3131522929
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000029294A29295A292963
29296B29297329297329297B29297B292984292984292984292984292984292984292984
29298429298429297B29297B29297B29297B29297329297329297329297329296B29296B
29296B29296B29296B29296B29296B29296B29296B29296B29296B29296B29296B29296B
29296B29296B29296329296B29296329296329296329295A29295A292952292929392931
4229294A29315229295A29316329296B29316B29317331317331317B31317B3131843131
7B31317B31317B29317B29317B2929732931732929732931632929522929000000000000
0000000000000000000000000000000000000000003129294229294A2929292942292952
29294A29295229295229295A29295A29295A29295A31295A29295A29295A292963312963
29295A31295A29295A31295A29295A31295A29295A31296329296331296B29296B313173
29297331297B29297B313184292984313184312984313184292984313184292984313184
29297B31317B31297B31317329296B31316B29295A3131526B29296B2929632929632929
6329296329296329296329295A29295A29295229295229294A2929392929000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003929294A29295A29315A29296329316B29297331317329297B29317B29317B3131
7B31318431398431318431317B31317B31317B29297331316B29296B2931632929523131
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000029292931295229295A
31316B29296B31297329297331317B29297B31298429297B31318429298431298429297B
31318429297B31298429297B31318429297B31297B29297331317B29297331297329296B
31317329296B31297329296B31317329296B31297329296B31317329296B31296B292963
31316B29296329296B29296331316329295A29295A29295231315229294A000000312929
4231314A29295229315A29296331316329297331316B29297B31317331317B31317B3131
8439397B31317B31317B29298431317329297329316B29296B3131522929000000000000
0000000000000000000000000000000000004229314229295229315229295A293129294A
31315229294A31294A29294A31315A29295A31295A29295231316329295A31316329295A
31316329295A31316329295A31316329295A31316329296331316B29296B31317329296B
31317B29297331317B29297B31318429298431318429298431318C29298431318429297B
31318429297B31317B29296B31316B29296331315A2929426B31316329296B2929632929
6B31316329296B29295A29296331315A29295A2929522929523131422929422929000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003929295231315229296331316329296B31316B29297B31317329297B31317B2931
8431397B31318431397B29318431317B29297B31317329297331316329296331314A2929
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000292939292952
29295A29296B29296B29297329297331297B29297B292984292984312984292984292984
29298431298429297B29297B29297B31297B29297B29297B292973312973292973292973
29297329297329296B29296B29296B29297329296B29296B29296B29296B29296B29296B
29296B29296B29296329296329295A29295A292952292952000000000000000000312929
3929294A29315229295A29315A29316329316B29296B29317329317331317331317B3131
7B31317B31317B31317B31317B29317B29317329296B29296B29295A2931000000000000
0000000000000000000000003929294229294A29295229295A29295A29295A2929632929
63292929294229293929294A29295231295A29295229295A29295A31296329295A292963
29296331316329296329296329296331316329296331296B29296B31317329296B312973
31297331317B29297B31297B31297B31318429298431298431298431318429297B31317B
31297B31317B29297331296B31296331316329294A3129396B29296B29296B29296B2929
6B29296B29296329296329296329296329295A29295A29295229295229294A29294A2929
422929000000000000000000000000000000000000000000000000000000000000000000
3129294231315229295A29296329296B31316B29297329297329297B31317B29317B3131
7B31318431317B31317B31317B31317B31317329297329296B29296B31315A29294A2929
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000029294A
31295A29295A31316B29296B31317329297331317B29297B31318429297B31318429297B
31318429297B31318429297B31318429297B31317B29297331317B292973313173292973
31297329296B31317329296B31297329296B31317329296B31296B29296331316B292963
31296329295A31296329295231295A292952312939000000000000000000000000292929
3929314229295231315229295A31316329296B31316B29297331317329297B3131732931
8431317B31317B31317B29297B31317329297B31316B29296B31315A2929523131000000
0000000000003929313929294A29314A29295A29315229295A29315A2929632931632929
6B29316B292973293129294231315229295231315A29295A31316329295A31316329295A
31316329296331316B29296331316B29296331316B29296331316B29296B313173292973
31317B29297B31318429297B31318429297B31318429298431318429297B31318429297B
31317B29297331317329296B31316329295231314A6329296B29316329296B2931632929
6B29316329296329315A29296329315A29295A29315229295A29315229295A3131522929
5229314A2929422931000000000000000000000000000000000000000000000000000000
3931314229295231315A29296331316329297331316B29297331317329297B31317B2931
8431317B31318431317B29297B31317329297B31316B29296B31316329295A3131392929
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
29295229295A29296329297329297329297329297B29297B29297B29298429297B292984
29297B29297B29297B29297B29297B29297B29297B29297B292973292973292973292973
29297329297329297329297329296B29296B29296B29296B29296B29296B292963292963
29295A29295A29295A29295229294A000000000000000000000000000000000000312931
3129293929314A29295229315A29296329316329296B29316B29297329317329317B3131
7B31317B31317B31317B31317B29297B29317329297329316B29296B2931522929000000
3929294229294A29295229295229295A29295A29295A29295A29296329296329296B2929
6B29296B29296B29297329296B292929295229295229295A29295A29295A292963312963
29296329296329296331296B29296331296B29296B31316B29296B312973292973313173
29297B31317B29297B31317B29297B31318429297B31318429297B31317B29297B31317B
29297331317329296B31316B29295A3131426329296329296329296B29296B29296B2929
6329296329296329296329296329296329296329296329295A29295A29295A29295A2929
5229295229294A29294A2929422929000000000000000000000000000000000000312931
4229294A29295229295A29316329296B29316B29297331317329297B31317B31317B3131
7B31317B31317B31317B31317B2929733131732929733131632929632929522929000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029295231296329296331317329297331297329297331317B29297B31297B29297B
31318429297B31297B29297B31317B29297B31297B29297331317B29297331297B292973
31317329296B31297329296B31317329296B29296B29296331316B29296331296329295A
31315A29295229294A292939000000000000000000000000000000000000000000000000
3931313929294A29314A29295A31315A29296329316329297331316B2929733131732929
7B31317B29297B31317329297B31317329297B31316B29297331316329296329314A2929
4A31314A29295229315229295A31315A29296329315A29296B31316329296B29316B2929
7329316B29297329316B292973313173292931294A29294A31315229295231316329295A
31316329295A31316329296331316B29296331316B29296B313173292973313173292973
31317B29297331317B29297331318429297B31318429297B31317B29297331317B292973
31317329296B31316329295A31314A6329296329296329296B31316329296B2931632929
6B31316329296329295A29296331315A29296329315A29296331315A29295A2929522929
5A31315229295229314A2929523131422929422929392929000000000000000000392929
4A31314A29295A31315A29296B31316B29297331316B29297B31317329297B3131732929
7B31317B31317B31317329297B31317329297331316B29296B31315A2929523131000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000029294A29295A29296329296B29296B29297329297331297329297329297B
29297B29297B29297B29297B29297B31297B29297329297B29297329297B292973292973
29297329297329296B29296B29296B29296B29296329296329296329296329295A29295A
29294A292942000000000000000000000000000000000000000000000000000000000000
3129293929314229294A29315229295A29315A29296B29316B2931732931732929732931
7329317B31317B29297B31317B29317B31317329297329316B29296B29316329295A2929
4A29295229315229295A29295A29296329316329296329296B29296B29316B29296B2929
6B29297329296B292973292973292973292929294229294A29294A31295229295A29295A
29295A31316329296329296329296331316B29296B29296B292973313173292973312973
29297331317B29297331297B31297B31317B29297B31297B292973313173292973312973
31296B31316329295A3129425A29296329295A2929632929632929632929632929632929
6329296329295A29295A29296329296329296329296329295A29295A29295A29295A2929
5A29295A29295229295229294A29294A2929422929422929392929392931312929422929
4A29295231315A29296329296329296B31316B29297329317329297B31317329317B3131
7B31317B31317B29297B29317329297331316B29296B29296329295A3131422929000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031295229295A31316B29296B31317329296B313173292973
31297B29297331317B29297331317B29297331317B29297331297329297331317329296B
31317329296B31317329296B31296B29296331296329295A312963292952312952292942
000000000000000000000000000000000000000000000000000000000000000000000000
3129313129294231314229295A31315229296331316329296B31316B2929733131732929
7B31317329297B31317329297B31317329297B31317329297331316B29296B31315A2929
5A31315229295A31315A29296329315A29296B29316329296B29316B29297331316B2929
7329316B29297329317329297329316B29297329316B292931314A29294A31315A29295A
31316329295A31316329296331316B29296331316B29296B31317329296B313173292973
31317B29297331317B29297331317B29297331317B29297331317329296B313173292963
31316329295231314A5229295A29295229295A29315A29296329315A29296331315A2929
6329315A29296329315A29296329315A29296331315A29296329315A29296331315A2929
5A31315A29295A31315229295229314A29294A3131392929393131312929423131422929
5231315229295A31316329296B31316B29297331316B29297331317329297B3131732929
7B31317329297B31317329297331316B2929733131632929633131522929393131000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000029294A29295229295A29296329296B29296B29296B
29296B29297329297329297329297329297329297329297329296B29297329296B29296B
29296B29296B29296B29296329296329296329295A29295229294A292942000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
2929293129313929294A29315229295A29315A29296329316329296B2931732929732931
7329297B29317B29297B31317B29297B29317329297B29317329297329316B29296B2931
5A29295A29315A29296329316329296329296329296B29296B29296B29296B2929732929
6B29297329297329297329297329297329296B29297329296B29296B292929293929295A
29295A29295A29296331296329296331296B29296B31296B29296B31296B29296B313173
29297331297329297331317B29297331317329297331317329296B31296B292963313163
2929523129425229295229295229295229295229295A29295A29295A29295A2929632929
5A29295A29295A29295A29295A29296329295A29295A29295A29295A29295A29295A2929
5A29295A29295229295229294A29294A29294229294229293929293929294229294A2931
5229295A29295A29296329316329296B29316B29297331317329297B29317329297B3131
7329297B31317B29297331317329297329296B29296329315A2929522929000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000029294A31315A29295A312963292963
31316B29296B31297329296B31317329296B31297329296B31317329296B31296B292963
31316B29296329296329295A31315A29295229294A292942000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003129293929314229295231315229295A29315A29296B31316B29297329316B2929
7B31317329297B31317329297B31317329297B29317329297B31316B2929732931632929
6331316329296329315A29296B31316329296B29316B29297331316B29297329316B2929
7331316B29297329316B29297331316B29297329316B29297331316B29296B2931292939
31315229295231315A29295A31316329296331316B29296331316B29296B31316B29296B
31317329297331317B29297331317329296B31317329296B31316B292963313163292952
31314A4229294A29294A29295231315229295229295229295A31315229295A2929522929
5A31315229295A29315A29296331315A29295A29315A29296331315A29296329315A2929
5A31315229295A29315229295231314A29294A31314229294231314229294A31314A2929
5A31315A29296331316329296B31316B29297331316B29297B31317329297B3131732929
7B31317329297B31317329297331316B29296B31315A29295A3131422929000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000292942292952292952292963
29296329296329296329296B29296B29296B292963292963292963292963292963292963
29296329295A29295229294A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003129313929294229294A29295229315A29296329316329296B29316B2929732931
7329317329317329297B29317B29297B29317329297329297329297329316B29296B2929
6B29296B29316329296329296B29296B29316B29296B29296B29297329316B2929732929
7329297329316B29296B29296B29296B29296B29296B29296B29296B2929632929632929
63292931294A29294A29295229295A31316329296329296329296331316B29296B29296B
29296B31317329296B31297329296B31316B29296B31296B29296331315A292952312942
3129293929293929294229294229294A29294A2929522929522929522929522929522929
5229295A29295A29295A29295A29295A29295A29295A29295A29296329315A29295A2929
5A29295A29315229295229295229294A29314229294229294229294A29314A2929522929
5A29296331316329296329296B29296B31316B2929732931732931733131732929732931
7329297B31317329297329296B29296B31316329295A2929522929000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292942
31295A29295A31316329295A31295A29295231315A29295231295A29295231315A292952
31294A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000002929293931313929294A31314A29295A31315A29296331316329297331316B2929
7331317329297B31317329297B31317329297B31317329297B31317329297331316B2929
6B31316329296B31316329296B29316B29297331316B29297329316B29297329316B2929
7329316B29297329316B29297329316B29296B29316329296329315A29296329315A2929
5A293152292952293129294231315229295231316329295A31316329296331316B292963
31316329296331316B29296331316B29296331316B29295A31315A29294A000000000000
0000002929293129313129293929293929294A29314A29295229314A29295229314A2929
5229315229295A31315229295A29315229295A31315229295A31315A29296331315A2929
5A31315229295A31315229295231314A29294A31314A29295231314A2929523131522929
5A31315A29296331316329296B31316B29297331316B2929733131732929733131732929
7B31317329297331316B29296B31316329296331315229294A3131000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000002929293929314229294A29315229295A29315A29296329316B29296B2931
6B2929732931732929732931732929732931732929732931732929732931732929732931
6B29296B29316B29296B29316B29296B29296B29296B29316B29297329296B2929732929
6B29296B29296B29296B29296B29296B29296329296329295A29295A2929522929522929
52292942292900000000000000000000000029293931294A29295A31295A29295A31295A
29295A31295A29295A31315A29295231294A292942000000000000000000000000000000
0000000000000000002929293129293929294229294A29294A29294A29294A29294A2929
4A29294A29295229295229295229295229295229295A29295A29295A29295A29295A2929
5229295A29295229295229294A29295229294A29295229314A29295229295229295A2931
5A29296329316329296B29316B29296B29316B2929732931732929732931732929733131
7329297331316B29296B31316B29296329315A2929522931000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003129294231314229295229315229296331315A29296B2931632929
7331316B29297329316B29297B31317329297B31317329297B31317329297329316B2929
7331316B29297329316B29297331316B29297329316B29297331316B29297329316B2929
7331316B29296B29316329296B31316329296329315A29295A2931522929522929422129
39293100000000000000000000000000000000000000000000000000000031315229294A
313152292942313139000000000000000000000000000000000000000000000000000000
000000000000000000000000000000292929392929392929423131422929422931392929
4A31314A29295229294A29295231315229295A29315229295A31315A29295A2931522929
5A31315229295A31315229295A31315229295231314A29295231315229295A31315A2929
6331315A29296331316329296B31316B29297331316B29297331316B29297331316B2929
7331316B29296B31316329296B31315A2929523131422929000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002929293129294229314A29295229295A29295A2931632929632931
6B29296B29316B2929732931732929732931732929732931732929732931732929732929
7329297329316B29296B29296B29297329316B29296B29296B29297329316B29296B2929
6B29296B29316329296329296329296329315A29295A29294A2929422929392929000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000312929392929392929392929392929
4229294A29294A29294A29295229295229315229295229295A29295A29295229295A2929
5A29295A29315229295229295229295229315229295229295229295A31315A29295A2929
5A29296331316329296B29296B29296B31316B29296B29316B29297331316B29296B2929
6B29296B31316329296329295A29295A31314A2929000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003129313129294231314A29295231315229296331315A2929
6B31316329296B31316B29297331316B29297331317329297331316B29297331316B2929
7331316B29297331316B29297329316B29297331316B29297329316B29296B31316B2929
6B29316329296329315A29296329315A29295A29314A2929422931312929000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000312931292929392931392929
4229314229294A29314229294A29314A29295231314A29295229315229295A3131522929
5A31315229295A31315229295231315229295A31315229295A31315229296331315A2929
6331315A29296B31316329296B31316B29297331316B29297331316B29297331316B2929
6B31316329296B31315A29295A31314A2929423131000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002929293129313929294229314A29295229315A29295A2931
6329296B29316B29296B29316B2929732931732929732931732929732931732929732931
6B29297329316B29297329316B29296B29296B29296B29316B29296B29296B29296B2931
6329296329295A29295A29295229295229294A2929422929000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000312929312929
3929293929294229294229294229294A29294A2929522929522929522929522929522929
5229295229295229295229315229295229295229295A29295A29295A29295A29295A2931
5A29296329296329296B29316B29296B29316B29296B31316B29296B29316B29296B3131
6329296329316329295A31315229294A2929000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002929293929313929294A31314A29295A2931522929
6331316329296B29316329296B31316B29297329316B29297331316B29297329316B2929
7331316B29297329316B29297331316B29296B29316329296B31316329296B29315A2929
6331315A29295A29314A2929523131422929000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292929
3131313129293929313929294231314229294A29314A29295231314A29295229314A2929
5231315229295A29315229295A31315229295A31315229295A31315A29295A31315A2929
6331316329296B31316329296B31316329296B31316329296B31316B29296B3131632929
6B31316329295A3131522929523131392929000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002929312929293929294229294A2931522929522929
5A29296329316329296329316B29296B29316B29296B29296B29297329316B29296B2929
6B29296B29316B29296B29296B29296B29316B29296B29296329296329316329295A2929
5A29295A29314A29294A2929422929000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000003129293929293929294229294229294A29294A29295229314A2929522929
5229295229315229295229295229295A29315229295A29295A29295A29315A29295A2929
6329296331316329296329296329296B31316329296B29296B29296B3131632929632929
632929633131522929522929422929000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003131313129293929314229295231314A2929
5A31315A29296331316329296B31316329296B31316329296B31316329296B31316B2929
6B31316329296B31316329296B29316329296B31316329296329315A2929632931522929
5229314A29294A2931312929000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003129293929313929294231314229294A31314229294A31314A2929
5231314A29295231315229295231315229295A31315229295A31315229295A31315A2929
6331315A29296331315A29296331316329296B31316329296B31316329296331315A2929
6331315229295231314A2929000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003129313129293929314229294A2931
5229295A29315A29295A29316329296329316329296329316329296329316329296B2931
6329296B29316329296329316329296329296329296329315A29295A2929522929522929
4A2929422929000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003129293129293929293929294229294229294229294A2929
4A29294A29294A29295229295229295229295229295229315229295A29295A29295A2931
5A29295A29295A29296329315A29296329296329296329315A29295A29295A29295A2931
522929522929422929000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000312929392931422929
5231314A29295229315229295A31315A29296329315A29296331315A29296329315A2929
6B31316329296329315A29296331315A29295A29315229295A31314A29294A2931422929
392931000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000292929312931312929393131392929422931392929
4A31314229294A31314A29295231314A29295231315229295A31315229295A3131522929
5A31315A29295A31315A29296331315A29295A31315229295A31315229295A31314A2929
4A3131422929000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000292929312929
3929294A29314A29294A29295229295229315229295A29295A29295A29315A29295A2929
5A29295A29315A29295A29295A29295A29315229295229294A2929422931392929000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000292929312929392929392929392929
4229294229314229294A29294A29295229314A29295229295229295231315229295A2929
5A29295A29315229295A29295A29295A31315229295229295229295231314A2929422929
392929000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292929
3129313129294231314229294A31314A29295231315229295A29315229295A3131522929
5A31315229295A31315229295229314A29294A2931422929392931292929000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000292929393131312929
4231313929294231314229294A31314229294A31314A29295231314A2929523131522929
5A31315229295A31315229295231315229295231314A2929523131422929423131292929
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003129293129293929313929294229314229294A29314A2929522929522929522931
4A29294A29314A29294A2931422929422929392929000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000292929312929
3129293929293929294229294229294229294229294229294229294A29294A2929522931
4A29295229295229295229314A29294A29294A29294A2931422929392929000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000312931292929393131392929422931392929423131422929422931392929
423131392929422931392929000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000002929293129313129293931313929294231313929294231314229294A3131422929
4A31314A29294A31314229294A3131422929423131392929393131000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000312929
292929000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000292929312931392929392929392929392931392929392929
392929422931392929422929422929392931000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000313131292929313131312929
313131000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[18]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L7
00061ol050071`001PM]0003K75a009aX@071`001PMW000=L@1a0740L@1a0740L@2O00L700061f@0
00AaK75/476K00L700061f<001Ua0740L@1a0740L@1a0740L@1a0740L@1a04<000ea0740L@1a0740
L@1a04L01`L000H7H0000g5[L@02L@03K75a01Ia?`000faaL@0;L@03L75a04D01`L000H7G`008G40
L@1a0740L@1a0740L@1a0740L@1a0740L@1a0740L@0k000EL@1a075=LDeaCG5=LDeaCG5=L@1a04<0
1`L000H7GP000g5/L@0RLC@06W400g1a000o00L700061ed002Ua0740L@1a0740L@1a0740LDeaCG5=
L@1a0740L@1a0740L@1a0740L@0_000OL@1a0740L@1a075=LDeaCG5=LDeaCG5=LDeaCG40L@0o00L7
00061e`0;G4[01Ia00EfLGIaMP04L@03L00003d01`L000H7F`00<g40L@1a0740L@1a075=L@1aCG5=
LDeaCG5=LDeaCG5=LDeaCG40L@1a0740L@1a0740L@0O000YL@1a0740L@1aCG5=LDeaCG5=LDeaCG5=
MTefCG5=MTefCGI=MTeaCG40?@071`001PMJ0004L@1aL0]a00E`LG5aL003L@03L75a01ea6@000g5`
L@02L@05L75aLG0037405gI`MW5fL7IaMW5fLGIaMW5fLGIaMW1a03`01`L000H7F@00>g40L@1a0740
L@1aCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG40L@1a0740L@1a01<0031a0740
L@1a0740LDeaCG5=LDeaCG5=LDeaCGI=MTefCGI=MTefCGI=MTefCGI=LDdl00L700061eP000Ia075=
LDdmL@@05g405WIaMW5fLGIaMW1fLGI`MW5fL7IaMW45MP07LGIaMW1fL00k00L700061eL0085a0740
L@1aCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG40
L@1a0740L@1a0740L@1aCG5=LDeaCG5=LDeaCG5=LDeaCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=
MTefCG40>`071`001PMF0008L@1aL75=LG0;L@05L75aLG0047402gIaMW5fLGIaMW5f02=a01afLGIa
MW5fLGIaMW5fLGIaMW5fLGI`MW5fL7Ia3GH0171fL74j00L700061eL008=a0740LDeaCG5=LDeaCG5=
LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=
LDeaCG5=LDeaCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=L@0i00L7
00061eH000QaCG5=LDeaC@ea00=`LG403W408GIaMW5fLGIaMW5fLGIaMW5fLGIaMW5fLGIaMW5fLGIa
MP03L@0DMW5fLGIaMW5fLGIaMW5fLGIaMW4;MP05LGIfMW000gH00gEfMP0;MP05L7I`MW00>@071`00
1PME0026L@1a075=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=MTefCGI=MTefCGI=
MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=
MTefCGI=MTefCGI=MTefCGI=MTdh00L700061eH000QaCG5=LDefL0]a00E`LG5aL003L@03L7Ia00Ia
02YfLGIaMW5fLGIaMW5fLGIaMW5fLGIaMW5fLGIaMW5fLGIaMW5fLGIaMW45MP03LGIa01Ef00EeMWIf
M@0?MP06MGI`MW1f>0071`001PME0027LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=
MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGIBMU9fDWI=
MTefCGI=MTefCGIBMTefCGI=MTefCGI=MTefCGI=MTefCGI=MTea03L01`L000H7E006L@07MW5fLGIa
MP0CL@0HMW5fLGIaMW5fLGIaMW5fLGIaMW5fLGIaCgH00g40000e00L700061eD008MaCG5=LDeaCG5=
MTeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=
MTefCGI=MTefCGI=MTefDWIeMU9fDWIBMU9fDWIBMU9fDWI=MU9fDWI=MTefCGI=MTefCGI=MTefCGI=
MTefCGH0=`071`001PMD00Aa00]fLGIaMW5fLGIaMP0AL@0@MW5fLGIaMW5fLGIaMW5fLDmf00=eMWH0
1GH2LCH01`L000H7=@000g40L@0K002:L@1aCG5=MTefCGI=MTefCG5=LDeaCG5=LDeaCG5=LDeaCG5=
MTefCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fMGIeMU9fDWIB
MU9fDWIBMU9fDWIBMU9fDWI=MTefCGI=MTefCGI=MTefDWIBMU9fCGI==P071`001PLb000<LFaa6740
LAQa675/5@05L@0=MW5fLGIaMW5fLGIaMP0?L@0>MW5fLGIaMW5fLGIaMW4VMP03NgIf01if00EaMWIf
L@0=MP03LGH003D01`L000H7<@004G40L@1a0740L@1a0740L@1a014008YaCG5=LDefCGI=MTefCGI=
MTeaCG5=LDeaCG5=LDeaCG5=LDeaCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=MU9fDWIBMU9fDWIB
MU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWI=MTefCGI=MTefCGI=MTefCGIBMU9fDWI=
MTdf00L700061bh0011aK75/LFaa675/LAQaK74H0g400faaL@0?0007CG5aMW5fL@07MP06LGIaMW5f
3g402WIaMW5fLGIaMW4NMP0KNgIfMW]fNgIkMW]fNgIkMW]fNgIkMW]fNgIk011f00MaMW5fLGIe00=f
00=eMWH01gH00g40000c00L700061bd001Ua0740L@1a0740L@1a0740L@1a0740L@1a00d008]aCG5=
MTefCGI=MTefCGI=MTefCGI=LDea075=LDeaCG5=LDeaCGI=MTefCGI=MTefCGI=MTefCGI=MTefCGIB
MU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWI=MTefCGI=MTefCGI=
MTefCGIBMU9fDWIBMTea03D01`L000H7:`005FaaK75/LFaaK75/LAQaK75/LFaaK003L@04K75aL@/0
17400gIaMP0>MP05LGIaMW400`09L@0:MW5fLGIaMW5fLAYf01]kMW]fNgIkMW]fNgIkMW]fNgIkMW]f
NgIkMW/047H01g5fLGIaMW403gH00g40000c00L700061b/001ea0740LAQa674HLAQa674HLAQa0740
L@1a0740L@0;000HLDefCGI=MTefCGIBMU9fDWIBMTefCGI=2P00JG5=LDeaCG5=MTefCGI=MTefCGI=
MTefDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMTefCGH0
LDeaCGI=MTefCGIBMU9fDWIBMU9fDWI=MP0e00L700061bT000/HLAQa674HLAQa600;L@09674HLAQa
674H00Aa1P02L@0:MTefCWIBMU9fDP]f00==00002@07L@08MW5fLGIaMW4@MP0ONgIkMW]fNgIkMW]f
NgIkMW]fNgIkMW]fNgIkMW]fN`0<MP03LGIa00<000E`LG1fL@0:MP0;NgEkMGIeMWEfL740=0071`00
1PLY000UL@1a674HLAQa674HLAQa674HLAQa674HL@1a0740L@1a0740L@03000IL@1aCGI=MTefCGIB
MU9fDWIBMU9fDWI=MP0=0019LDeaCG5=LDefCGI=MTefCGI=MTefDWIBMU9fDWIBMU9fDWIBMU9fDWIB
MU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fCGI=MP05000JLDeaCGI=MTefDWIBMU9fDWIBMU9fDWI=
MTdd00L700061bT000/HLAQa674HLQQa600;L@09674HLFaa675/00Qa00h0LG5aMTefCWICMU9fD`]f
00==0000200:L@08MW5fLGIaMW4@MP0MNgIkMW]fNgIkMW]fNgIkMW]fNgIkMW]fNgIkMW/037H70003
LG1f00]f00]kMG]eNgEfMGI`MP0d00L700061bT0041a674HLAQa674HLAQa674HLAQa674HLAQa674H
LAQa0740L@1a0740L@1aCGI=MTefCGIBMU9fDWIBMU9fDWI=2000CG40L@1a0740LAQaCG5=LDefCGI=
MTefCGI=MU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWI=MTef00T0
01QaCGI=MTefDWIBMU9fDWIBMU9fDWI=MTdd00L700061bL00W402aQa678HLQQb678H00Ya00Yb678H
LAQa674H37402WI=MTifDgICMU8;MPD000IaK75/LF`<L@08MW5fLGIaMW4@MP0INgIkMW]fNgIkMW]f
NgIkMW]fNgIkMW]fN`0:MP03L@0000P000AaCGIa2WH02g]eNgEkMGIeMWEf03@01`L000H79`00@W4H
LAQa674HLAQa674HLAQa674HLAQa674HLAQa674HLAQa674HL@1a0740L@1aCGI=MU9fDWIBMU9fDWIB
MTefC@8004ma0740L@1a074HLAQa674HLAQa675=LDefCGI=MTefCGI=MTefDWIBMU9fDWIBMU9fDWIB
MU9fDWIBMU9fDWIBMU9fDWIBMU9fDWI=MTef00d001QaCGI=MTefDWIBMU9fDWIBMU9fDWI=MTdd00L7
00061bH002=aK74HLQQb678HLQQb679aLW5bLG9aLW5b678HLVab679/LP0<L@04K74HL@9f00=CMU<0
2WH0375/LFaa675/LFaaK0Ya00=bK7800g4017IaMW4BMP03NgIk00Ef00EkMWIfN`03MP03NgIk00Yf
3`00175`MW4:MP0;NgEkMG]eMU9fMGH0=0071`001PLW002>LAQa674HLAQa674HLAQa674HLAQa674H
LQQa678HLAQa674HLAQa674HLAQa0740L@1a0740L@1a0740L@1a0740LAQa674HLAQa674HLAQa674H
LAQb678HLQQbCG5=MTefCGI=MTefCGI=MU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMU9fDWIBMTefCA80
01QaCGI=MTefDWIBMU9fDWIBMU9fDWIBMTdd00L700061bD002hHLAQbK79/LVabK79/LVabK79/LVab
K79/LW5bLG9aLW5bLG9/LW5bK79/LVab47400faaL@06L@0TLW5bLG9aLW5bLG9aLVabLG9aLW5bLG9a
LW5bLG9aMW5fLGIa7GH0175fLGHF009a00=`MWH027H01G]fNgIk00If=0071`001PLU0028LAQa678H
LQQa674HLAQa678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLAQa674HLAQa674HLAQa674H
LAQa674HLAQb678HLQQb678HLQQb678HLQQb678HLQQb678HLTeaCGI=MTefCGI=MTefDWI=MTefCGI=
MTefCGI=MTefCAX001QaCGI=MTefCGIBMU9fDWIBMU9fDWIBMTdd00L700061bD005/HLQQbK79/LVab
K79aLVab679/LVabK79/LW5bLG9aLW5bLG9aLVabLG9/LW5bK79aLW5bLG9aLW5bLG9/LW5bK79aLW5b
LG9aLW5bLG9aLW5bK79aLVabLG9a00=b00iaLW5bLG9/LW5bK79aLP=a00AfLGIa2gH01W5fLGIaMQh0
0g400gI`MP05MP07NgIkMW]fN`06MS@01`L000H79@00Og4HLQQb678HLQQa674HLAQb678HLQQb678H
LQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678H
LQQb678HLQQb678HLQQb678HLQQb678HLQQb675=LDeaCG5=MTeaCG408`00675=MTefCGI=MU9fDWIB
MU9fDWIBMU9fCC@01`L000H79000774HLQQbKG9/LVabK79/LVabK79/LVabK79/LW49LP05KG9]LVd0
0g800febLP09LP07LG9aLW5bL@0OLP0GLG9bLW5bLG9/LVabK79/LVabLG9aLW401G801FabK75/02<0
0W401G1fLGIe00If00EkMW]fN`06MS@01`L000H79@00Og4HLQQb678HLQQa674HLAQb678HLQQb678H
LQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678H
LQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb67408`00675=MTefCGI=MU9fDWIB
MU9fDWIBMU9fCC@01`L000H790006W4HLQQbKG9;LVebK79/LVab679/LQQbK79/3G801FebLW9]00=b
00=]LW803g800febLP0CLP03KG9b00Mb01=aLW5bK79aLVabLG9/LW5bLG9a00Mb00I]LVabK74R00=a
00=fL7H01gH01G]fNgIk00If=0071`001PLS0022LAQb678HLQQb678HLQQb674HLQQb678HLQQb678H
LQQb678HLQQb678HLQQb678HLT]bBg9;LT]bBg8HLQQb678HLQQb679;LVebBg9;LT]b678HLQQb678H
LQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQb678HLQQa628001QaCG5=MTefCGIBMU9fDWIB
MU9fDWIBMTdd00L700061b<001/HLAQbBg9]LVebKG9]LVabK75/LQQbK79/LV`02g801FebLW9]03Mb
015aLW5bLG9/LW5bK79/LVabL@09LP07KG9/LVaaK00E0003Jg5[009a1`02L@03L7Ia00Qf00EkMW]f
N`06MS@01`L000H78`00Pg4HLQQb679;LT]bBg8HLQQb674HLAQa678HLQQb678HLQQbBg9;LQQb679;
LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bKG9]LT]bBg9;LT]bBg8HLT]b678HLQQb678HLQQb678H
LQQb678HLQQb678HLQQbBg9;LQQb678HLQQa01<000]aJg5[LF]aJg5[L@03000HL@1aCGI=MTefDWIB
MU=fDWIBMU9fDWI==0071`001PLS0003LG9/00=b00];LT]bBg9;LT]bB`05LP0?LG9aLW5b679;LT]b
KG9;00eb00U]LfecKG=]Lfd037801G=bLg9c00Ab00=]LW802G8036ebKG9/LVabK79aLPEa00ibLG9a
LT]bBg9;LT]bB`Eb00=aLW403P04L@04RW6:L@>:00U[RV^:JhY[RP000W400gIaMP07MP0;NgIkMW]e
NgEfMGH0=0071`001PLS001XLAQb678HLT]bBg9;LT]bBg8HLQQb674HLAQb678HLQQb678HLT]bBg9;
LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LVebBg9;LT]bBg9;LT]bBg9;LT]bBg9;LQQb678HLQQb678H
LAP4000GLAQb678HLT]bBg9;LT]bBg8HLQQb67403@00;75[LF]aJhY[RV^:JhY[RV^:JhY[L@1aCGI=
MTefDWIBMU9fDWIBMU9fDWI==0071`001PLR0003LFab009b00m]LT]bBg9;Ld]bBg9;LVd00g803f`0
001aLQQb679]LT]bK@0<LP0OLfecKG=]LfecKG=bLg9cLW=bLg9cLW=bLg9cLW=]L`0<LP0:KG9]LQQb
K78HLPH000=aLW400g802D]bBg9;LT]bB`05LP04LG9aL@X000][RV^:LHYaRW6:L@03RP09Q8Y[RV^:
JhY[009a00=fLGH01gH02g]fNgEkMG]eMWEf03@01`L000H78`005g4HLQQbBg9;LT]bBg9;LT]bBg8H
LQQa00@004THLQQb678HLQQb679;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]b
Bg9;LT]bBg9;LT]b678HLQQb674H00T001LHLQQbBg9;LT]bBg9;LT]bBg8HLQQa6008000`LF]aJg5[
RV^:JhY[RV^:JhY[RV^:JhY[L@1aCGI=MTefDWIBMU9fDWIBMU9fDWI==0071`001PLR009a1G802d]b
Bg=;Ld]cBg9;00=b00=/LW401@002AQb679;LVebB`0<LP0PLfecKG=]LfecKG=bLg9cLW=bLg9cLW=b
Lg9cLW=bLfd;LP07KG9/LVab600:009a0g802D]bBg=;Ld]cB`07LP9a1P001f]aJhY[RV/02hX02HB:
Q8Y[RV^:0002L@03MW5f00Mf00]kMW]fNgEkMGIeMP0d00L700061b8001PHLAQb679;LT]bBg9;LT]b
Bg8HLQQb67460016678HLQQb678HLT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]b
Bg9;LT]bBg9;LT]b678HLQQb675[L@X001LHLQQbBg9;LT]bBg9;LT]bBg8HLQQb6005000cJg5[LF^:
JhY[RV^:JhY[RV^:JhY[RV^:JhY[RP1aCGI=MTefDWIBMU9fDWIBMU9fDWI=03@01`L000H78P000g5/
LP02LP0@KG9;LT]cBg=;LT]bBg9]LP9a2@001W9/LQQbK@ab021cKG=]LfecKG=]Lg9cLW=bLg9cLW=b
Lg9cKG=bLfecK@]b00THLV^:JhY[RV/02002L@=b00U;LT]cBg=?Ldl01g83L@0<06]aJhY[RV^:JhY[
2hX02HB:Q8Z4RXB:J`02L@03MW5f00Mf00]kMW]eNgIkMGIeMP0d00L700061b8001LHLAQb679;LT]b
Bg9;LT]bBg8HLQQb600;0015678HLQQbBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]b
Bg9;LT]bBg9;LQQb678HLQR:JhY[RV^:Jg5[00P004ia679;LT]bBg9;LT]bBg9;LQQb678HL@1aJhY[
RV^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RV]a075=MTefCGIBMU9fDWIBMU9fDWIBMTdd00L700061b80
0W400g9]LP02LP0;Bg=;Ld]cBg9;LT/00g82L@`000<HLQP0378087=]LfecKG=]LfecLW=bLg9cLW=b
Lg9cKG=]LfecKG9]1g803W5bLHY[RV^:JhY[RV^:0W46000=LG9bLT]bBg=;Ld]cC`07LPAa00][RV^:
JhZ4RXB:Q00;RP09Q8Z4RXB:JhX0009a00=fLGH01gH02g]fNgIkMG]eMU9f03@01`L000H78P005aQb
678HLT]bBg9;LT]bBg9;LQQb678H00h004Mb678HLT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;
LT]bBg9;LT]bBg9;LQQb678HLF^:JhY[RV^:JhY[RV^:Jg5[L@03001>LAQb679;LT]bBg9;LT]bBg9;
LV]aJg5[RV^:JhY[RV^:JhZ4RV^:Q8Z4RXB:JhY[RV^:JhY[L@1aCGI=MTefDWIBMU9fDWIBMU9fCGI=
=0071`001PLR0006LAQbBg9;0g801D]cLW9?00=b00I;LT]b674A0003LVab009b00E]LW9bK@04LP0C
Lg9cLW=bLg9cLW=bLfecLW=bL`0:LP07LG9/RVb:K00=RP0=QHZ:RVb:LHY/RW5bK@03LP0;KG=bLg6:
K8YaRV`00hX00hF:RP0;RP05Q8Z:RX@00hX00hB:RP07RP=a00=fL7H01GH01g]fNgIkMW/01WHd00L7
00061b8001HHLQQb678HLT]bBg9;LT]bBg8HLQQa5@00TW8HLQQb679;LT]bBg9;LT]bBg9;LT]bBg9;
LT]bBg9;LT]b678HLQR:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RV^:Jg5[LF]aJg5[LF^:JhY[
RV^:JhY[RXB:Q8Z4RXB:Q8Z4RV^:JhY[RV^:JhY[RV^:JhY[LDefCGI=MTefDWIBMU9fDWIBMU9fDWI=
=0071`001PLR0006LAQbBg9;0W8017=]LVd3LP08Bg9]LQQb674F0003679]01eb00maLW5bLHY/RVb:
K8Z5RXD02hX00hF:Q@03RP03QHZ:00>:00MaRW6:LHYa00V:00>5RXX03hX01hF:RXZ4RXD00hX01W6:
RXYaRP9a00=`MWH01WH01g]fNgIkMW/01WHd00L700061b4001Ma678HLQQb679;LT]bBg9;LT]b678H
L@0I002>LAQb678HLT]bBg9;LQQbBg8HLQQb678HLT]b674HLF]aJg5[RV^:JhY[RV^:JhY[RV^:Q8Z4
RXB:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RV^:JhY[RV^:JhY[
RV^:Jg5=MTefCGI=MU9fDWIBMU9fDWIBMU9fCC@01`L000H78@001aQb679;LT/01G800febB`03LP05
Bg8HLQP06003L@07CWIfMW5bL@07LP0:KG9bLVabLG9aLP9a00E/LG5aK002L@08RVb:K8YaRV`=RP05
QHZ:RXD00hX00hF:RP0;RP05QHZ:RXD00hX00hF:RP0;RP05QHZ:RX@00hX02XB:RXYaRW6:LHX3L@Qf
00MkMW]fNgIk00If=0071`001PLQ000GLAQb678HLT]bBg9;LT]bBg9;LQQb67406@00SW40LDefCGIB
MU9fDW4HLAQfDWI=MTea0740L@1aJg5[LF]aJhY[RV^:JhY[RV^:JhZ4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4
RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Y[RV^:JhY[RV^:Jg5[L@1aCGI=MTefCGIB
MU9fDWIBMU9fDWI=MTdd00L700061b4000LHLQQbBg9;00Ab00acLW9]LVebKG8HLQPH0006LDiaLGIa
27H00g]fMP06MP05CW5/LF`027402XY/RVb:K8Y/RXD=RP03QHZ:01B:00fSRZ>:XhZSRZ>:XhZS00R:
00F5RXZ:Q@03RP04LHYaRPMa27H01G]fNgIk00Qf=0071`001PLQ000GLAQb678HLT]bBg9;LT]bBg9;
LQQb67406@00SW5=LDefCGIBMU9fDWIBMU9fDWI=MTea0740L@1a0740LF]aJg5[RV^:JhY[RV^:JhY[
RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Y[RV^:JhY[
LF]aJg5[L@1aCGI=MTefDWIBMU9fDWIBMU9fDWI=MTdd00L700061b4000LHLQQbBg9;00Ub00M]LT]b
678H01P00g400difL@0:MP03NgIf009f00EBMW5aK003L@05K75aLF`00g402Fb:K8Y/RVb:L@0;RP05
QHZ:RXD00hX00hF:RP02RP0IXhZSRZ>:XhZSRZ>:XhZSRZ>:XhZSRZ>:X`06RP05QHZ:RV`00hX00f^:
L@05L@03Jg5a009a27H01G]fNgIk00If00=aMP00<`071`001PLQ000GLAQb678HLT]bBg9;LT]bBg9;
LQQb67406@00SW5=LDefCGI=MU9fDWIBMU9fDWI=MTefJg5[LF]a0740L@1aJg5[LF^:JhY[RV^:JhY[
RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:JhY[RV^:JhY[
LF]aJg5[L@1aCGI=MTefDWIBMU9fDWIBMU9fDWI=MTdd00L700061b800W43LP0?KG9;LT]cBg=;LT]b
Bg9/009a6@03L@04MTifD`=f00=CNgH02gH016^:Jg450003LF]a00Ea00j:LHYaRV^:QHZ4RXF:Q0b:
022SQJ>:XhFSRZ>5XhZSRZ>:XhZSRZ>5XhBSQJ>4XhF:Q0F:00=aRW400g402F]aJhY[RV]aL00:MP0<
NgIkMW]BNgEfDWI`=@071`001PLR000F678HLQQb679;LT]bBg9;LQQb678HLAT001]aCG5=MTefCGIB
MU9fDWIBMU9fDWI=MTf:Jg401`00Jg40LF]aJg5[LF^:JhY[RV^:JhY[RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4
RXB:Q8Z4RXBSQ:>4RXB:Q8Z4RXB:Q8Z4RXB:JhY[RV]aJg40LF]aJg5[RV]a075=MTefCGIBMU9fDWIB
MU9fDWIBMTef03D01`L000H78P02L@Eb00a;LT]bBg9;LT]bBg82L@03K00001L00g401WI=MU=fD`9f
00=kMWH02WH00f^:J`0800Qa00b:LHY[RV^:QHZ4RXD:RP0RXhZSRZ>:XhZSRZ>:XhZSRZ>:XhZSRZ>4
XhBSQ:>4RXB:Q0>:1742009a00U[RV^:JhY=LDd02WH037]fNgIkMGIeMWEfCCD01`L000H78P005AQa
678HLT]bBg9;LT]bBg8HLQQa600J000KLDeaCGI=MTefDWIBMU9fDWIBMU9fCGI=RV]a00T004Aa0740
LF]aJg5[RV^:JhY[RV^:JhZ4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:JhY[
RV]aJ`@0025aJg5[RV^:Jg5=LDefCGIBMU9fDWIBMU9fDWIBMU9fCGH0=@071`001PLR0003LFab00Ab
00e;LT]bBg9;LT]bBg9/009a6@03L@0=MTifDWICMU=kMW]fN`08MP03JhY[00/01g402XY[RVb:JhZ5
RX@<RP0CXhFSRZ>5XhZSQJ>:XhZSRZ>:X`02RP0<QJ>4RXF:Q8Z5RV^:0W46009a00U[RV^:JhY=LG00
2WH037]fNgIkDWIeMU9fL3D01`L000H78P005QQa678HLT]bBg9;LT]bBg8HLQQb674I000KL@1aCGI=
MTefDWIBMU9fDWIBMU9fCGI=RV]a00d003ea075[LF]aJhY[RV^:JhY[RV^:Q8Z4RXB:Q8Z4RXB:Q8Z4
RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:JhY[RV]a00L0021a075[RV^:075=LDefCGIBMU9fDWIBMU9fDWIB
MTefCCH01`L000H78P02L@03LW5b009b00m?Ld]bBg9;LT]bBg9aLV`06@03L@0=MTefCWICMU=fDg]f
N`08MP03JhX000/027402Fb:JhY[RV^:K00<RP09XhZSQJ>5XhFS00>:00>SRZ<01XX038B:QHZ4RXB:
JhY[L@T000YaJhY[RV]aCG5=27H03G]fNgIkMG]eMU9fMGH0=P071`001PLR000F678HLQQbBg9;LT]b
Bg9;LQQb678HLA/001QaCGI=MTefCGIBMU9fDWIBMU9fCGI=RV/:000oL@1a075=L@1aJg5[LF^:JhY[
RV^:JhY[RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Y[RV^:JhY[RV]a00/001iaJhY[RP1aCGI=
MTefDWIBMU9fDWIBMU9fDWI=MTdf00L700061b400W400fabLP03LP0?Cg=?Ld]bBg9;LT]bK79a01X0
0W402WI>MTefDgIBMU<2MP03NgIf00If00==L@00200;L@09Jg5/LF^:K8Y[00f:00V5RXZ:QHZ5RXD0
2hX02XF:Q8Z5RV^:Jg4;000:LF^:JhX0LDefL0Qf00ekMW]fNgEkDWIeMTef03H01`L000H78@005g4H
LQQb679;LT]bBg9;LT]bBg8HLQQa01/001QaCGI=MTefCGIBMU9fDWIBMU9fDWI=LF/80011L@1aCG5=
LDeaCG5=L@1aJg5[LF^:JhY[RV^:JhY[RV^:Q8Z4RXB:Q8Z4RXB:Q8Z4RXB:Q8Y[RV^:JhY[L@1a0740
2`007G5[RV]a075=MTefCGIBMU9fDWIBMU9fDWIBMTef03L01`L000H78P001W4HLT]bB`9b00EcLW=b
L`04LP05KG9;LQP06`001DiaLGIa00Qf00=kMW/01WH00g5f000200Ma00]=LDefCGI=MTefL@03MPMa
00R:K8YaRW6:LAF:00AaRW6:2W4014eaCG450004LF^:K0Aa00=fLGH01GH01g]fNgIkMW/01WH00g40
000e00L700061b8001HHLAQb679;LT]bBg9;LT]bBg8HLQQa6`00JG5=LDefCGI=MU9fDWIBMU9fDWIB
MTefCG40L@1a075=LDeaCGI=MTefCGI=MTefCGI=MP1aJg5[LF^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[
RV^:JhY[LF]aCG5=LDeaCG40L@1a0740L@03000MLF]a075=LDefCGI=MU9fDWIBMU9fDWIBMU9fCG40
=`071`001PLR0006LAQbBg9;0g8014]cLW<4LP06Bg8HLQQa6P001DiaLGIa00Yf00AkMWIC1GH9L@0:
MW5fCGI=MTefCPEf00=BMWH00WH00g5/L@02L@0KK75aRVb:LHYaRW6:LHYaRW6:K8YaRVb:LHY/009a
00=fLGH00W400g1aL@03L@09CG5`LDeaL74000Qa00=fLGH01GH01g]fNgIkMW/01WHh00L700061b<0
01Ea678HLQQbBg9;LT]bBg9;LQQb67806`00R75=LDefCGI=MU9fDWIBMU9fDWIBMTefCG5=LDeaCG5=
LDefCGI=MTefCGI=MTefCGIBMU9fDWH0L@1a075[LF]aJg5[LF]aJg5[LF^:Jg5[LF]a675=LDeaCG5=
LDeaCG5=L@1a0740L@1a0740LDeaCG5=MTefCGIBMU9fDWIBMU9fDWIBMTdh00L700061b8000Ia678H
LT/=LP04678HLA/00W400gIaMP09MP03NgIf00If00=aMW40174037IaMTefCWI=MU9fDP]f00MBMU9f
DW5/019a00ebKG8HLFafLGIaMW5f00Ia00U`LG1aL75=LG001W400gIaMP09MP03NgIk00Mf00=a0000
=P071`001PLS000FLAQb678HLT]bBg9;LT]bBg8HLQQb61/008I=LDefCGI=MU9fDWIBMU9fDWIBMU9f
CGI=MTeaCG5=MTefCGI=MTefCGI=MTefCGIBMU9fDWI=MTefCGI=MQQb075[L@1a075;LT]bBg8HLQQb
675=LDeaCG5=LDeaCG5=LDea0740L@1a075=LDeaCGI=MTefDWIBMU9fDWIBMU9fDWI=MST01`L000H7
8P001W4HLQQbB`Mb00=;LW800g801D]b678H01X00W400gIaMP08MP05DgIfNe<01gH02g5fLGIaMW5f
LGI=00=f00=BMU<02GH03e9fDWIBMU9fDWHHLVabK@04LP06Lg9cLW=;0g803T]bK74HMW5fLGIaMW5f
17401DeaLG5=00=a00==LG4017400gIaMP07MP05NgIkMW/01WH00g40000g00L700061b<001Ma678H
LQQbBg9;LT]bBg9;LQQb678HL@0I0026L@1aCG5=MTefCGIBMU9fDWIBMU9fDWI=MTefCGI=MTefCGI=
MTefCGI=MTefCGI=MTefDWI=MTefCGI=MTea679;LT]bBg9;LT]bBg9;LQQb678HLDeaCG5=LDeaCG5=
LDeaCG5=L@1a0740LDeaCG5=MTefCGIBMU9fDWIBMU9fDWIBMTdj00L700061b<000DHLQQbB`04LP03
Lg9c009b00U;LW9bBg8HLQP060000g4HL@03L@03MW5f00Af00=CMWH01GH00e=fMP05MP0?LGIaMW5f
DWICMU9fDWIB00]f00IBMTefCG420003679;00Ab00McLW=bLg9c009b011]LVeb675/MW5fLGIaMW5f
0g4027I`LDeaL75`1W400gIaMP0;MP03NgIf00If>P071`001PLS000GLAQb678HLT]bBg9;LT]bBg8H
LQQb67405@00?G40L@1a675=LDefCGI=MTefDWIBMU9fDWIBMU9fCGI=MTefCGI=MTefCGI=MTefCGI=
MTefCGI=MTefCG401@00Ag4HLT]bBg9;LT]bBg9;LT]b678HLQQfCG5=LDeaCG5=LDeaCG5=LDeaCG5=
LDeaCG5=MTefCGIBMU9fDWIBMU9fDWIBMTef03/01`L000H78`000g5bL@03LP0;Bg=;LdmcCg=;Ld/0
1G82LA0001=/LAQaK75/LFab675=MTefCGIB00]f00]BMU9fDWIBMU9fDP0;MP09DWIBMTefCGI=00P0
0W43LP09Bg9;Ld]cCg=?00Mb17402defCGI=MTefCGI=00]a00U=MTefCGIBMU803GH00e9f000k00L7
00061b<001Qa678HLQQbBg9;LT]bBg9;LT]b678HLAP@000lL@1a674HLAQa6780LDeaCGI=MTefDWIB
MU9fDWIBMU9fDWI=MTefCGI=MTefCGI=MTefCGI=MTefCGI=2@00AaQb679;LT]bBg9;LT]bBg9;LQQb
6740LDeaCG5=LDeaCG5=LDeaCG5=LDeaCG5=MTefCGI=MTefDWIBMU9fDWIBMU9fCGI=03`01`L000H7
9002L@0@LVebBg9;LdmcCg=?Ld]bK@=b00=/L@003@000faa6003L@0?675/LQQb075=LDefCGI>00]f
00]BMU9fDWIBMU9fDP0;MP06CGI=MTef2`02L@=b00U;Ld]cBg=?Ld/01g84L@0<CGI=MTefCGI=MTef
0g4047IaLG5fLGI=MTefL7IBMWD=MP03L00003/01`L000H7900061Qa678HLT]bBg9;LT]bBg9;LQQb
678HL@/003ma074HLAQa674HLAQb6740L@1aCGI=MTefDWIBMU9fDWIBMU9fDWIBMTefCGI=MTefCGI=
MTefCGI=MTea07402@00Ag4HLQQbBg9;LT]bBg9;LT]bBg8HLQQaCG5=LDeaCG5=LDeaCG5=LDeaCG5=
LDefCGI=MTefCGI=MU9fDWIBMU9fDWIBMTea03d01`L000H790000g5/LP02LP0;Bg9;LT]cBg=?Ld/0
1G800g5bK008000JLAQa675/LFaaK79/LVabK780LF]aCG5=MTd;MP0;DWIBMU9fDWIBMU802gH02Def
674HLAQa60070003LG9a00=b00U;Ld]cBg=;Ld/01G80175bL@02L@0OCG5=MTefCGI=MTefLGIaMW5f
LGIaMTefCGIBMU9fDP0=MSh01`L000H79@00674HLQQbBg9;LT]bBg9;LT]bBg8HLQQa60@004Ua0740
L@1a674HLAQa678HLQQb6740LF]aCG5=MTefCGIBMU9fDWIBMU9fDWIBMU9fCGI=MTefCGI=MTefCG4H
LAQa6740L@1a00D001Ma678HLT]bBg9;LT]bBg9;LT]b678HL@03000[LDeaCG5=LDeaCG5=LDeaCGI=
MTefCGI=MTefCGI=MU9fDWIBMU9fDWI=MP0o00L700061bD000maLVabBg9;LT]bBg=?Ld/01G803feb
LG5/L@1a674HLAQa6002L@0DLVabLG8HLW5b679[LF^:Jg5=MTd;MP0;DWIBMU9fDWIBMU801gH03g5f
LG8HLQQa674HLAQaK003L@05Jg5aLW400g802D]cBg=;LdmcB`05LP=a1@007DeaCG5=MTefCGIaMW5f
LGIaMW5fCGI=MU9fDWIe00]f00=a0000?@071`001PLU0020LAQb679;LT]bBg9;LT]bBg9;LQQb678H
L@1a0740LAQa674HLAQa678HLQQb678HL@1aJhY[LDeaCGI=MTefCGIBMU9fDWIBMU9fCGI=MTefCGI=
MTefCG8HLAQa674HL@1a0740L@1a674HLQQbBg9;LT]bBg9;LT]b678HLQP8000VLDeaCG5=LDeaCG5=
MTefCGI=MTefCGI=MTefCGIBMU9fDWI=MTe000L700061bD000m/LW5bBg9;LT]cBg=;Ld/01g809G5b
LG4HLAQa674HLQQbK79aLVabK79]LVeb075[RV^:JhY=MW402GH02e9fDWIBMU9fDWIB00Ef00maMW5b
LG8HLQQb674HLAP02745LP09Bg=;Ld]cBg=;00Eb0W480007CG5=LDeaC@02L@0BMW5fLGIaMW5fCGIB
MU9fDWIB2GH2LD001`L000H79P00OaQb678HLT]bBg9;LT]bBg9;LQQb678HLAQa674HLAQa674HLQQb
678HLQQb678HLP1aJg5[RV^:CG5=MTefCGI=MTefCGI=MTefCGI=MTefCGI=LAQb678HLQQa674HLAQa
6740L@1a678HLQQbBg9;LT]bBg9;LT]b678HLAP02P00975=LDeaCG5=LDefCGI=MTefCGI=MTefCGI=
MTefCGI=MTeaJd001`L000H79P002W4HLVebBg9]LT/3LP0YBg9;LT]bBg9;LQQbK74HLQQb678HLQQb
678HLQQbBg9;LT]aJg5/RV`00hX01diaCGIaMTh03GH08g5fCW8HLQQb678HLQQbK78HLVaa674HLAQa
679/LT]bKG9]00Ab00=cLW800W801T]bK78HL@d00g402deaLGI=MW5fCGIa00]f00M`MW5aQ8Y[0440
1`L000H79`00Og4HLT]bBg9;LT]bBg9;LT]bBg8HLQQb678HLAQa678HLQQb678HLQQb678HLQQb075[
LF^:JhY[RV]aCG5=MTefCGI=MTefCGI=MTea678HLQQb678HLQQb674HLAQa674HLAQa678HLQQb679;
LT]bBg9;LT]b678HLQQaJg403`007G5=LDeaCGI=MTefCGI=MTefCGI=MTefCHY[RV]a04401`L000H7
9`00=AQbBg9;LT]bBg9;LT]bBg9;LT]bBg9;LQQb678HLQQb678HLQQbBg9;LT]bBg9;LFaaK8Y/00F:
00j5RXAaCWI>MW5fLGIaMP9a02<HLQQb678HLQQb678HLQQb679/LQQbK74HLAQb679;LT]bB`0;LP09
KG8HLF^:Jg5[00h001aaCG5`MW1fLGIaMW5fLGIaMW5fL7IaRXB:Q8Y[@@071`001PLX0016678HLT]b
Bg9;LT]bBg9;LT]b678HLQQb678HLQQb678HLQQb678HLQQb678HLP1aJg5[LF^:JhZ4RV^:Q8Y[RV^:
CG5=L@D003Ia678HLQQb678HLQQb678HLAQa674HLAQb678HLQQb679;LT]bBg9;LT]b678HLP2:JhY[
LF/@000IL@1aCG5=LDeaCGI=MTeaCHZ4RV^:JhY[L@1100L700061bT003<HLT]bKG9;LVebBg9;LT]b
Bg9;LT]bBg8HLVebBg9;LT]bBg9;LT]bBg8HLQQaJg5/RV`02HX01XB:QHZ5RPT0029a678HLQQb678H
LQQbK78HLW5b678HLQQb679;LT]bKG9]2G802aQbK8Y[RV^:JhY[018000=aCG400g400gIaL@06RP07
Q8Z4RV^:J`1100L700061bT0049a678HLT]bBg9;LT]bBg9;LT]b678HLQQb678HLQQb678HLQQb678H
LQQb6780LF]aJhY[RV^:JhZ4RXB:Q8Y[RV/<000eLAQb678HLQQb678HLQQb678HLQQb678HLQQbBg9;
LT]bBg9;LT]b678HLF^:JhY[RV]aJg404@00575[RV^:JhZ4RXB:Q8Z4RV^:JhY[@P071`001PLY000e
678HLT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg9;LT]bBg8HLQQb678HLFaaK8Y/RXD01hX01hF:
Q8Z5RV/02000875[LAQb678HLQQbK78HLVabK79/LQQb679;LT]bBg9;2g803aQbQ8Y[RV^:JhY[RV]a
J`0>0004LF^:K0>:00>5RXX01HX01XB:Q8Y[RT801`L000H7:@00@g4HLQQb679;LT]bBg9;LT]bBg9;
LT]b678HLQQb678HLQQb678HLQQb678HL@1aJg5[RV^:JhY[RXB:Q8Z4RV^:Jg402000>f]aJhXHLAQa
678HLAQb678HLQQb678HLQQb678HLT]bBg9;LT]bBg9;LQQb68Y[RV^:JhY[RV]aJg5[00h001AaJhY[
RXB:Q8Z4RXB:Q8Y[RV^:Jd801`L000H7:P00:W4HLT]bBg9]LT]bBg9;LT]bBg9;LVebBg9;LT]bBg9;
LT]bBg8HLQQb60<000U[LFb:K8Y/RXD01hX01HF:Q8Y/00H002AaJhY[RV^:Jg4HLAQb679/LQQbLG8H
LQQbBg9;LVebBg9]LVd7LP0DKG8HLHB:Q8Z4RV^:JhY[RV]aJg4:0003Jg5[00>:00>4RXX01hX01XB:
Q8Y[LD801`L000H7;0009QQb678HLT]bBg9;LT]bBg9;LQQb678HLQQb678HLQQb678HLQQa1@005W40
LF]aJhY[RV^:Q8Z4RXB:Q8Y[RV/20015LF]aJhY[RV^:JhY[RV^:678HLQQb678HLQQb678HLQQbBg9;
LT]bBg9;LQQb678HRV^:JhY[RV^:JhY[RV^:JhY[LF]a00L001EaJg5[RV^:JhZ4RXB:Q8Y[RV^:Jg40
@`071`001PL]0009678HLT]bBg9]00=b00m]LW9bKG9;LT]bBg9;LT/00g801VabLG8HL@L000IaJg5[
RV/;RP0;Q8Y[RV]aJhY[RV/00hX01HF:QHZ500::00Ya678HLQQb678H3G802d]bBg9;LT]b68Y[00>:
01B5RXZ:JhY[RV^:JhY[RV^:Jg5[L@<0174028Z4RXB:Q8Z42HX00g40001100L700061bh0024HLQQb
678HLT]bBg9;LT]b678HLQQb678HLQQb678HLAP02P00Mg5[LF^:JhY[RV^:Q8Z4RXB:JhY[RV]aJhY[
RV^:JhY[RV^:JhXHLAQa678HLQQb678HLQQb679;LT]bBg8HLQQb678HLF^:JhY[RV^:JhY[RV^:JhY[
RV^:Jg5[LF]a075[LF^:JhY[RV^:JhZ4RXB:JhY[RV]a04<01`L000H7<000178HLT/3LP03KG9b00Eb
00i;LT]bBg9;LQQbK79/L@/000IaJg5[RV/;RP0=Q8Z4RV^:JhY[RV^:Q@03RP0?QHZ5RXF:Q74HLQQb
678H00eb01m;LT]b678HLF^:K8Y/RW6:K8Y/RV^:JhY[RV^:JhY[00Qa00b:LHZ4RXB:Q8Z4RX@5RP03
LHYa04<01`L000H7<@006W4HLQQb678HLQQb678HLQQb678HLQQb674H3P00MW40LF^:JhY[RV^:Q8Z4
RXB:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RV^:078HLQQb678HLQQb678HLT]b678HLQQb675[LF]aJhY[
RV^:JhY[RV^:JhY[RV^:Jg5[LF]a075[LF^:JhY[RV^:JhY[RXB:JhY[RV]400L700061c<000<HLV`0
0g8046ebLW9]LT]b678HLQQb674@0005Jg5[RV/00hX00hF:RP05RP0;Q8Z4RV^:JhY[RV/00hX01HF:
RXZ500>:00V4RXB:678HLQP03G807d]b678HLF]aJhY[RW6:JhYaRVb:JhY[RV^:JhY[RV/0274038Ya
RXB:Q8Z4RXB:Q0F:0W5400L700061c@0018HLAQb678HLQQb678HLQQb674C001eL@1aJg5[RV^:JhY[
RXB:Q8Y[RV^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RV]a674HLQQb678HLQQb678HLQQb6740L@1aJg5[
LF]aJg5[RV^:JhY[RV^:JhY[LF]aJg5[LF^:JhY[RV^:JhY[RV^:JhY[RV]a04D01`L000H7=P003G5/
LW5bK79aLQQb67405`001W5[LF^:L@V:02>4RXB:Q8Y[RV^:Q8Z5RXF:QHZ5RXF:JhY[RV^:JhXHLW5b
L@03LP03LG9b00=b00<HLQP01000675[LFaaK75/LFb:LHY[RV^:JhY[RV^:J`Ia00j:LHYaRXB:Q8Z4
RXB:Q0>:00=aRW40A@071`001PMK000[LF]aJhY[RV^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[
RV^:Jg5[L@03000<LAQb678HLQQb674H2000<740LF]aJg5[LF]aJg5[RV^:JhY[RV]aJg5[LF]aJhY[
RV^:JhY[RV^:JhY[RV^:JdH01`L000H7F`001f]aJhYaRXD00hX03hF:RXZ4RXB:Q8Z4RXB:Q003RP05
QHZ:RXD00hX01V^:JhY[L@P01G4<0003LF]a009a019/LG5aJg5[RV^:JhY[RV^:JhX3L@0@RW6:LHYa
RXB:Q8Z4RXB:Q0>:0W5600L700061ed002EaJg5[RV^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[
LF]a01l002]a0740LF]aJg5[LF]aJg5[LF]aJhY[RV^:JhY[RV^:JhY[RV^:JhY[RV]a04L01`L000H7
G000975/LF^:JhY[RXB:Q8Z5RXF:QHZ5RXF:Q8Z4RXB:JhY[RV]aJb<001e[LF]aJg5[LF^:JhY[RV^:
K8Y[RV^:JhY[RV^:Q@03RP08QHZ:RVb:Jg5800L700061ed0025a075[RV^:JhY[RV^:JhY[RV^:JhY[
RV^:JhY[RV^:Jg409`009G40L@1aJg5[LF]aJg5[LF^:JhY[RV^:JhY[RV^:JhY[RV^:Jg40B@071`00
1PMM000Q675[LF^:JhY[RXB:Q8Z5RXB:QHZ4RXF:JhY[RV^:Jg5[02T001M[LF]aJg5/LF^:K8Y[RV^:
JhY[RV^:J`03RP09Q8Z:RV^:LHY[04T01`L000H7G`007G5[LF^:JhY[RV^:JhY[RV^:JhY[RV^:JhY[
LF]a02d0021aJg5[LF]aJg5[LF^:JhY[RV^:JhY[RV^:JhY[RV]aJdX01`L000H7GP007G5[LF]aJhY[
RV^:K8Y/RXF:QHY/RV^:JhY[RV]a02h0021[LF]aJg5[LF]aJhY[RV^:JhY[RV^:JhY/RV^:K8Y[LDX0
1`L000H7H@006G5[LF]aJhY[RV^:JhY[RV^:JhY[RV]aJg40;`007g40L@1aJg5[LF]aJg5[RV^:JhY[
RV^:Jg5[LF]aJg40B`071`001PMR000FLF]aJg5[RV^:K8Y[RVb:JhY/RV]aJc@001MaJg5[LFaaJhY[
RV^:JhY[RV^:LHY[RP02L@03Jg4004/01`L000H7H`004g40LF]aJg5[LF]aJg5[LF]aJg40=`006G40
LF]aJg5[LF]aJg5[RV^:Jg5[LF]aJg40C@071`001PMT000@LF]aJg5/LF^:K75/LFaaJcX001IaJg5[
LF]aJg5[LF^:JhY[RV]aK75[C`071`001PM]0003L@0004D000]a0740L@1aJg5[L@1C00L700061kL0
1G5G00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`001POo01@01`L000H7o`0D00L700061ol050071`001POo01@01`L000H7
o`0D00L700061ol050071`00o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00
\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "The color directives in ",
  StyleBox["SurfaceGraphics[]",
    FontFamily->"Courier"],
  " objects and ",
  StyleBox["ParametricPlot3D[]",
    FontFamily->"Courier"],
  " output are supported as well."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(\(csurf = 
      Plot3D[Evaluate[\({#, Hue[#]}\ &\)\ @\ Exp[\(-x\^2\) - y\^2]], {x, 
          \(-1.5\), 1.5}, {y, \(-1.5\), 1.5}, DisplayFunction -> Identity]; 
    \)\)], "Input",
  CellLabel->"In[19]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(MVClear[]; MVShow3D[csurf]; \)], "Input",
  CellLabel->"In[20]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
      GraphicsArray[{Graphics[MVGetRasterImage[], AspectRatio -> Automatic], 
          MVGetGraphics3D[]}]]; \)\)], "Input",
  CellLabel->"In[21]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .47619 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0113379 0.47619 [
[ 0 0 0 0 ]
[ 1 .47619 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .47619 L
0 .47619 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0113379 0.477324 0.464853 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000CE4A31CE4A39CE4231D64A39CE4A31D64A39CE4A31
D64A39CE4A31D64A39CE4A31D65239CE4A31D64A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39
CE4A39CE4A39CE4A31CE4A39CE4A39D64A39CE4A31CE4A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000CE4A31D64A39CE4A31CE4A39CE4A31D64A39CE4A31D64A39CE4A31D64A39CE4A31
D64A39CE4A31D64A39CE4A31D64A39CE4A31D64A39CE4A31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000CE4A39
CE4A39CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39
CE4A31CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39CE4A31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000CE5231CE4A39CE4A31
D64A39CE4A31CE4A39CE4231D64A39CE4A31CE4A39CE4A31D64A39CE4A31CE4A39CE4A31
D64A39CE4A31CE4A39CE4A31D64A39CE4A31D64A39CE4A31D64A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000CE5A39CE5A39CE5A39CE4A31CE4A39
CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39
CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000CE5231CE5A39CE5231CE5A39CE5231CE5A39CE4A31
CE4A39CE4A31CE4A39CE4A31CE4A39CE4A31D64A39CE4A31CE4A39CE4A31D64A39CE4A31
CE4A39CE4A31D64A39CE4A31D64A39CE4A31D64A39CE4A31D64A39CE4A31000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CE5A39CE5239CE5A39CE5A39CE5A39CE5A39CE5A39CE5A39CE4A39
CE4A31CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39
CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000CE5231CE5A39C65231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231
D64A39CE4A31CE4A39CE4A31D64A39CE4A31CE4A39CE4231D64A39CE4A31CE4A39CE4A31
D64A39CE4A31CE4A39CE4231D65239CE4A31CE4A39CE4A31D64A39CE4A31CE4A39000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39
CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39
CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000C65231CE5A39CE5231
CE5A39C65231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231
CE5A39CE4A31CE4A39CE4A31CE4A39CE4A31CE4A39CE4A31CE4A39CE4A31D64A39CE4A31
CE4A39CE4A31CE4A39CE4A31CE4A39CE4A31D64A39CE4A31CE4A39CE4A31D64A39CE4A31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000CE5A39CE5239CE5A39CE5A39CE5A39
CE5231CE5A39CE5A39CE5A39CE5A39CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39
CE5A39CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39
CE4A31CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000CE5231CE5A39C65231CE5A39CE5231CE5A39C65231
CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE4A39CE4231D64A39CE4A31CE4A39CE4A31D64A39CE4A31CE4A39CE4A31
D64A39CE4A31CE4A39CE4A31D64A39CE4A31CE4A39CE4A31D64A39CE4A31CE5A39CE5231
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39
CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39CE5A39CE5A39CE5A31CE5A39
CE5A39000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000C66331CE6B39C66B31CE5A39C65231CE5A39C65231CE5A39C65231CE5A39C65231
CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE5A39CE4A31CE4A39CE4A31CE4A39CE4A31CE4A39CE4A31CE4A39CE4A31
CE4A39CE4A31CE4A39CE4A31CE4A39CE4A31CE5A39CE5231CE5A39CE5231CE5A39CE5231
CE5A39000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000CE6B39
CE6B39CE6B39CE6B39CE6B39CE5239CE5A39CE5A39CE5A39CE5239CE5A39CE5A39CE5A39
CE5239CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5239CE5A39CE5A39CE5A39
CE5231CE5A39CE5A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39
CE4A31CE4A39CE4A39CE4A39CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39
CE5A31CE5A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000CE6B39C66331
CE6B39C66B31CE6B39C66331CE6B39C65231CE5A39C65231CE5A39CE5231CE5A39C65231
CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231
CE5A39CE5231CE5A39CE5231CE4A39CE4A31CE4A39CE4231D64A39CE4A31CE4A39CE4A31
CE4A39CE4A31CE5A39CE5231D65A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231
D65A39CE5231000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000CE6B39CE6B39C66B31CE6B39
CE6B39CE6B39CE6B31CE6B39CE6B39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE4A39CE4A39CE4A31CE4A39CE4A39CE4A39CE4A31CE4A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000CE6B39C66331CE6B39C66331CE6B39C66331
CE6B39C66331CE6B39C66B31CE6B39C66331CE5A39C65231CE5A39C65231CE5A39CE5231
CE5A39C65231CE5A39C65231CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231
CE5A39CE5231CE5A39CE5231CE4A39CE4A31CE4A39CE4A31CE4A39CE4A31CE5A39CE5231
CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE5A39000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000C66B39CE6B39CE6B39CE6B39C66B39CE6B39CE6B39CE6B39
CE6B39CE6B39CE6B39CE6B39CE6B31CE6B39CE5239CE5A39CE5239CE5A39CE5A39CE5A39
CE5231CE5A39CE5A39CE5A39CE5239CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39
CE5A39CE5A39CE5A39CE5A39CE5231CE4A39CE4A39CE4A39CE5A39CE5A39CE5A39CE5A39
CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39
CE5A31CE5A39CE5A39000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66B31CE6B39C66331
CE6B39C66B31CE6B39C66331CE6B39CE6B31CE6B39C65231CE5A39CE5231CE5A39C65231
CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231
CE5A39CE5231CE5A39C65231CE5A39CE4A31CE5A39C65231CE5A39CE5231CE5A39C65231
CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE5A39CE5231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
C66B39CE6B39C66B31CE6B39CE6B39CE6B39C66B31CE6B39CE6B39CE6B39C66B31CE6B39
CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE7339CE7339CE5231CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5A31CE5A39
CE5A39CE5A39CE5231CE5A39000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000CE6B39C66331
CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66331
CE6B39C66331CE6B39C66B31CE6B39C66331CE6B39CE6B31CE5A39C65231CE5A39CE5231
CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE7339CE6B31CE7339CE6B31CE5A39CE5231CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE5A39CE5231CE5A39000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000C67B39C67B39C66B39CE6B39
C66B39CE6B39C66B39CE6B39C66B39CE6B39C66B39CE6B39C66B39CE6B39CE6B39CE6B39
CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE5231CE5A39CE5A39CE5A39
CE5239CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5239CE5A39CE5A39CE5A39
CE6B31CE7339CE7339CE7339CE6B31CE7339CE7339CE5A39CE5A31CE5A39CE5A39CE5A39
CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39
CE5A31CE5A39CE5A39CE5A39CE5A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000CE8439C67B31CE8439C67B31CE6B39C66331
CE6B39C66331CE6B39C66331CE6B39C66B31CE6B39C66331CE6B39C66B31CE6B39C66331
CE6B39C66B31CE6B39C66331CE6B39CE6B31CE6B39C66331CE6B39CE5231CE5A39C65231
CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231CE7339C66B31
CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339C65231CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231
CE5A39CE5231CE5A39C65231CE5A39CE5231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000C68439C67B31C67B39C67B39CE8439C67B39CE6B39
C66B39CE6B39C66B31CE6B39CE6B39CE6B39C66B39CE6B39CE6B39CE6B39C66B31CE6B39
CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE7339CE6B31CE7339
CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31
CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66331
CE6B39C66331CE6B39C66B31CE6B39C66331CE6B39C66B31CE6B39CE6B31CE5A39CE5231
CE5A39C65231CE5A39CE5231CE5A39CE5231CE5A39CE6B31CE7339CE6B31CE7339CE6B31
CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE5A39CE5231CE5A39CE5231
CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39C65231CE5A39CE5231
CE5A39CE5231CE5A39CE5231CE5A39C65231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000C67B39C67B39C67B39C67B39C67B39C67B39C68439C67B39CE8439C67B39CE8439
C66B39CE6B39C66B39CE6B39C66B39CE6B39CE6B39CE6B39C66B39CE6B39CE6B39CE6B39
CE6B39CE6B39CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B39CE6B39CE5A39CE5A39
CE5231CE5A39CE5A39CE5A39CE5A39CE7339CE7339CE7339CE6B31CE7339CE6B39CE7339
CE6B39CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE7339CE5A39CE5A39CE5A39
CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A31CE5A39
CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A31000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000C67B31
CE8439C67B31C68439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C66331CE6B39C66331CE6B39C66B31CE6B39C66331CE6B39C66B31CE6B39C66331
CE6B39C66B31CE6B39C66331CE6B39CE6B31CE6B39C66331CE6B39CE6B31CE6B39C65231
CE5A39CE5231CE5A39C66B31CE7339CE6B31CE7339C66B31CE7339CE6B31CE7339C66B31
CE7339CE6B31CE7339C66B31CE7339CE6B31CE7339CE6B31CE7339CE5231CE5A39C65231
CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39C65231
CE5A39CE5231CE5A39C65231CE5A39CE6B31CE6B39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000C67B39C67B39C67B39
C67B39C67B39C67B39C67B39C67B39CE8439C67B39C67B39C67B39CE8439C67B39CE7B39
CE7B39CE6B39C66B39CE6B39CE6B39CE6B39C66B31CE6B39CE6B39CE6B39CE6B31CE6B39
CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE5A39
CE5A39CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339
CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE5231CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5231CE5A39
CE5A39CE5A39CE5231CE6B39CE6B39CE6B39CE6B31CE6B39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000C67B31C68439C67B31C68439C67B31
C68439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C67B31CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66331CE6B39C66B31
CE6B39C66331CE6B39C66B31CE6B39C66B31CE6B39CE6B31CE6B39C66B31CE6B39CE8C31
CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31
CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE5A39CE5231
CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231CE5A39CE5231
CE5A39C66B31CE6B39CE6B31CE6B39C66B31CE6B39C66B31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000C67B39C67B39C67B39C67B39C67B39C67B39C67B39
C67B39C67B39C67B39C67B39C67B39C67B39C67B39CE8439C67B39CE7B39C67B39CE8439
C67B39CE8439CE6B39CE6B39C66B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39
CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE9439CE8C39CE9439
CE6B39CE7339CE6B39CE7339CE6B31CE7339CE6B39CE7339CE6B31CE7339CE7339CE7339
CE6B31CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339CE7339CE5A39
CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE5A39CE5A31CE5A39CE5A39CE6B39
CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000BD8C31C69439C67B39C67B39BD7B31C68439C67B31C67B39C67B31
CE8439C67B39CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B39CE8439C67B31
CE8439C67B31CE8439C66331CE6B39C66B31CE6B39C66331CE6B39C66B31CE6B39C66331
CE6B39CE6B31CE6B39C66331CE6B39CE6B31CE6B39C66331CE9439CE8C31CE9439CE8C31
CE9439CE6B31CE7339C66B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339C66B31
CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339C66B31CE7339CE6B31CE7339CE5231
CE5A39CE5231CE5A39C65231CE5A39CE5231CE5A39CE5231CE5A39CE6B31CE6B39C66B31
CE7339CE6B31CE6B39C66B31CE7339CE6B31CE6B39C66B31CE7339000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000C69439BD8C39C69439C69439C67B39C67B39C67B39C67B39C67B39C67B39C67B39
C67B39C68439C67B39C67B39C67B39CE8439C67B39C67B39C67B39CE8439C67B39CE8439
CE7B39CE8439C67B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B31CE6B39
CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE8C31CE9439CE9439CE9439CE8C31CE9439
CE9439CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339
CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE5A39
CE5A39CE5A39CE5231CE5A39CE5A39CE5A39CE5A31CE6B39CE6B39CE6B39CE6B31CE6B39
CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
C69439BD8C31C69439BD8C31C69439BD8C31C68439C67B39C68439C67B31C68439C67B31
CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C67B31CE8439C67B31CE6B39C66331CE6B39C66B31CE6B39C66331CE6B39C66B31
CE6B39C66B31CE6B39CE6B31CE9439C68C31CE9439CE8C31CE9439CE8C31CE9439CE8C31
CE9439CE8C31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31
CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31
CE5A39CE5231CE5A39CE5231CE5A39CE6B31CE6B39CE6B31CE6B39C66B31CE6B39CE6B31
CE6B39C66B31CE6B39CE6B31CE6B39C66B31CE6B39C66B31CE6B39C66B31000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000BD9439C69439
BD8C39C69439C69439C69439C69439C69439C69439C67B39C67B39C67B39C67B39C67B39
C67B39C67B39C67B39C67B39C67B39C67B39C67B39CE8439C67B39CE8439C67B39CE8439
C67B39CE8439CE7B39CE8439CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39CE6B39
CE6B39CE6B39CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C39CE9439
CE8C31CE9439CE6B39CE7339CE6B39CE7339CE7339CE7339CE6B31CE7339CE7339CE7339
CE6B39CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE7331CE7339CE7339CE7339
CE5A31CE5A39CE5A31CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B31CE6B39
CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000C69439BD8C39C69439BD8C31
C69439BD8C39C69439BD8C31C69439C69439C69439C67B31CE8439C67B39C68439C67B31
CE8439C67B39CE8439C67B31CE8439C67B39CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C67B39CE8439C67B31CE8439C66B31CE6B39C66331CE6B39CE6B31CE6B39C66331
CE9439CE8C31CE9439C68C31CE9439CE8C31CE9439CE8C31CE9439CE8C31CE9439CE8C31
D69439CE8C31CE9439CE6B31D67339CE6B31CE7339CE6B31D67339CE6B31CE7339CE6B31
D67339CE6B31CE7339CE6B31D67339CE6B31CE7339CE6B31CE7339CE6B31CE7339CE6B31
CE7339CE6B31CE6B39CE6B31CE7339CE6B31CE6B39C66B31CE7339CE6B31CE6B39C66B31
CE7339CE6B31CE6B39C66B31CE7339CE6B31CE6B39C66B31CE7339C66B31CE6B39000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000BD8C39BD9439BD8C39C69439BD8C39C69439
C69439C69439C68C39C69439C69439C69439C69439C69439C67B39C67B39C67B39C67B39
C67B39C67B39C67B39C67B39C67B39CE8439C67B39C67B39C67B39CE8439C67B39CE7B39
CE7B39CE8439C67B39CE8439CE8439CE6B39CE6B39CE6B39CE6B39CE6B39CE8C39CE8C39
CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C31CE9439CE9439CE9439CE8C31CE9439
CE9439CE9439CE8C31CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339
CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339
CE9439CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39
CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39C66B31000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000C69439BD8C31C69439BD8C39C69439BD8C31C69439BD8C39
C69439BD8C31C69439BD8C39C69439BD8C31C69439C69439C68439C67B39CE8439C67B39
CE8439C67B31CE8439C67B39CE8439C67B31CE8439C67B39CE8439C67B31CE8439C67B31
CE8439C67B31CE8439C67B31CE8439C67B31CE6B39C66B31CE6B39C68C31CE9439CE8C31
CE9439C68C31CE9439CE8C31CE9439CE8C31CE9439CE8C31CE9439CE8C31CE9439CE8C31
CE9439CE8C31D69439CE8C31CE7339CE6B31D67339CE6B31CE7339CE6B31D67339CE6B31
CE7339CE6B31D67339CE6B31CE7339CE6B31D67339CE6B31CE7339CE6B31CE9439CE8C31
CE9439CE8C31CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39C66B31CE6B39C66B31
CE6B39C66B31CE6B39C66B31CE6B39C66B31CE6B39C66B31CE6B39C66B31CE6B39C66B31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000BD8C39BD9439BD8C39BD9439BD8C39C69439BD8C39C69439BD8C39C69439
BD8C39C69439C69439C69439C69439C69439C69439C69439C69439C67B39C67B39C67B39
C67B39C67B39C67B39C67B39C67B39C67B39C67B39CE8439C67B39CE8439C67B39CE8439
C67B39CE8439CE7B39CE8439CE7B39CE8439CE6B39CE9439CE8C39CE9439CE8C39CE9439
CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C39CE9439
CE8C39CE9439CE9439CE9439CE6B39CE7339CE7339CE7339CE6B39CE7339CE7339CE7339
CE6B31CE7339CE7339CE7339CE7331CE7339CE7339CE7339CE9431CE9439CE9439CE9439
CE9431CE9439CE6B39CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39CE6B39CE6B39
CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
BD9439BD8C39BD9439BD8C31C69439BD8C39C69439BD8C31C69439BD8C39C69439BD8C31
C69439BD8C39C69439BD8C31C69439C69439C69439C68C31CE9439C67B39C68439C67B31
CE8439C67B39CE8439C67B31CE8439C67B39CE8439C67B31CE8439C67B39CE8439C67B31
CE8439C67B39CE8439C67B31CE8439CEB539CEB539C68C31CE9439CE8C31CE9439C68C31
CE9439CE8C31CE9439CE8C31CE9439CE8C31CE9439CE8C31D69439CE8C31CE9439CE8C31
D69439CE8C31CE9439CE8C31D69439CE6B31CE7339CE6B31D67339CE6B31CE7339CE6B31
D67339CE6B31CE7339CE6B31D67339CE6B31CE9439CE8C31D69439CE8C31CE9439CE8C31
CE9439CE8C31CE9439C66B31CE7339CE6B31CE6B39C66B31CE7339CE6B31CE6B39C66B31
CE7339CE6B31CE6B39C66B31CE6B39C66B31CE6B39C66B31CE7339C66B31CE6B39C66B31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000BD9C39BD9C39
BD9C39BD9439BD8C39BD8C39BD8C39BD9439BD8C39BD8C39BD9439C69439BD8C39C69439
C69439C69439BD8C39C69439C69439C69439C69439C69439C69439C69439C67B39C67B39
C67B39CE8439C67B39C67B39C67B39CE8439C67B39C67B39CE7B39CE8439C67B39CE8439
CE7B39CE8439C67B39CEB539CEB539CEB539CEB539CE9439CE8C39CE9439CE8C39CE8C39
CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C31CE9439CE9439CE9439CE8C39CE9439
CE9439CE9439CE8C31CE9439CE9439CE7339CE6B31CE7339CE7339CE7339CE6B31CE7339
CE7339CE7339CE6B31CE7339CE7339CE9439CE8C31CE9439CE9439CE9439CE9431CE9439
CE9439CE9439CE8C31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39
CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39C66B31CE6B39CE6B39CE6B39C66B31CE6B39
CE6B31000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000BDA539B59C31BDA539BD9C39
BDA539BD8C39C69439BD8C39C69439BD8C31C69439BD8C39C69439BD8C39C69439BD8C39
C69439BD8C39C69439BD8C39C69439C68C39C69439C69439C69439C69439CE8439C67B39
CE8439C67B39CE8439C67B39CE8439C67B39CE8439C67B39CE8439C67B39CE8439C67B39
CE8439C6B531CEB539C6B531CEB539C6B531CEBD39CEB539CE9439CE8C31CE9439CE8C31
CE9439CE8C31CE9439CE8C31CE9439CE8C31CE9439CE8C31CE9439CE8C31D69439CE8C31
CE9439CE8C31D69439CE8C31D69439CE8C31D67339CE6B31D67339CE6B31D67339CE6B31
D67339CE6B31D67339CE8C31CE9439CE8C31D69439CE8C31CE9439CE8C31D69439CE8C31
CE9439CE8C31CE9439CE8C31CE6B39CE6B31CE6B39CE6B31CE6B39C66B31CE6B39C66B31
CE6B39C66B31CE6B39C66B31CE6B39C66B31CE6B39C66B31CE6B39C66B31CE6B39C66B31
CE6B39000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000B59C39BD9C39BD9C39BD9C39BD9C39BD9C39
BD9C39BDA539BD8C39BD9439BD8C39BD8C39BD8C39C69439BD8C39C69439BD9439C69439
BD8C39C69439C69439C69439C69439C69439C69439C69439C69439C69439C69439C67B39
C67B39C67B39C67B39CE8439C67B39CE7B39C67B39CE8439C67B39CE8439CE7B39CE8439
CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEB539CE8C39CE9439CE8C39CE9439
CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C39CE9439CE8C39CE9439CE9439CE9439
CE8C39CE9439CE9439CE9439CE9439CE9439CE7339CE7339CE6B31CE7339CE7339CE7339
CE7339CE9439CE9439CE9439CE9431CE9439CE9439CE9439CE9431CE9439CE9439CE9439
CE9431CE9439CE9431CE9439CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B31CE6B39
CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39CE6B31CE8439
C68431CE8439000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000BD9C39B59C39BD9C39B59C31BDA539B59C39BDA539B59C31
BDA539BD9C39BD9439BD8C31C69439BD8C39C69439BD8C39C69439BD8C39C69439BD8C31
C69439BD8C39C69439BD8C39C69439C69439C69439C68C39CE9439C69439C69439C67B39
CE8439C67B39CE8439C67B39CE8439C67B39CE8439C67B39CE8439C67B39CEB539C6AD31
CEB539C6B539CEB539C6B531CEBD39CEB539CEB539C6B531CEBD39CE8C39CE9439CE8C31
D69439CE8C39CE9439CE8C31D69439CE8C39CE9439CE8C31D69439CE8C31CE9439CE8C31
D69439CE8C31D69439CE8C31D69439CE8C31D69439CE6B31D67339CE6B31D67339CE8C31
D69439CE8C31CE9439CE8C31D69439CE8C31CE9439CE8C31D69439CE8C31CE9439CE8C31
D69439CE8C31CE9439CE8C31CE9439CE6B31CE6B39C66B31CE7339CE6B31CE6B39C66B31
CE7339CE6B31CE6B39C66B31CE7339C66B31CE6B39C66B31CE7339C66B31CE8439C67B31
CE8439C68431000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000B59C39B59C39B59C39BD9C39B59C39BD9C39BD9C39BD9C39BD9C39BD9C39
BD9C39BDA539BD9C39BD8C39BD8C39BD9439BD8C39BD9439BD8C39C69439BD8C39C69439
C69439C69439C69439C69439C69439C69439C69439C69439C69439C69439C69439C69439
C67B39CE8439C67B39C67B39C67B39CE8439C67B39CE7B39CEB539CEB539C6B539CEB539
CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEB539CE9439CE8C39CE9439
CE9439CE9439CE8C39CE9439CE9439CE9439CE8C39CE9439CE9439CE9439CE8C39CE9439
CE9439CE9439CE8C39CE9439CE9439CE9439CE8C31CE7339CE7339D69439CE9431CE9439
CE9439CE9439CE9431CE9439CE9439CE9439CE9431CE9439CE9439CE9439CE8C31CE9439
CE9439CE9439CE9431CE9439CE9439CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39
CE6B39CE6B39CE6B31CE6B39CE6B39CE6B39CE6B31CE6B39CE8439CE8439C68431CE8439
CE8439CE8439C68431000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000B59431BD9C39B59C39BD9C39B59C31BD9C39B59C39BDA539B59C39BDA539BD9C39
BDA539BD9C39BDA539BD8C39C69439BD8C39C69439BD8C39C69439BD8C39C69439BD8C39
C69439BD8C39C69439BD8C39C69439C68C39C69439C69439C69439C69439CE9439C69439
CE9439C67B39CE8439C67B39CE8439C67B39CEB539C6AD39CEB539C6AD39CEB539C6B539
CEB539C6B539CEB539CEB539CEB539CEB539CEBD39CEB539CEBD39CEB539D69439CE8C39
CE9439CE8C31D69439CE8C39D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31
D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31
D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31CE9439CE8C31CE9439CE8C31
CE9439CE8C31CE9439CE8C31CE9439CE6B31CE6B39CE6B31CE6B39C66B31CE6B39CE6B31
CE6B39C66B31CE6B39C66B31CE6B39C66B31CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C67B31CE8439000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000B59C39
B59C39B59C39B59C39B59C39B59C39BD9C39B59C39BD9C39BD9C39BD9C39BD9C39BD9C39
BD9C39BD9C39BD9C39BDA539BD8C39BD9439BD8C39C69439BD8C39C69439BD8C39C69439
C68C39C69439C69439C69439C69439C69439C69439C69439C69439C69439C69439C69439
C69439C67B39C67B39CE8439C67B39CEB539C6B539CEB539C6B539CEB539CEB539CEB539
CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEB539CE9439CE9439
CE8C39CE9439CE9439CE9439CE9439CE9439CE9439CE9439CE9439CE9439CE9439CE9439
CE9439CE9439CE9439CE9439CE9439CE9439CEC639CEC639CE9439CE9439CE9439CE9439
CE9431CE9439CE9439CE9439CE9431CE9439CE9439CE9439CE9431CE9439CE9439CE9439
CE9431CE9439CE9439CE9439CE9431CE9439CE6B31CE6B39CE6B31CE6B39CE6B31CE6B39
CE6B31CE6B39CE6B31CE6B39CE8431CE8439CE8431CE8439C68431CE8439C68431CE8439
C68431CE8439C68431000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000B59439B59C39B59431
BD9C39B59C39BD9C39B59431BD9C39B59C39BD9C39B59C31BDA539B59C39BDA539B59C39
BDA542BD9C39BDA539BD9C39C69442BD8C39C69439BD8C39C69442BD8C39C69439BD8C39
C69442BD8C39C69439BD8C39C69442C69439C69439C68C39CE9442C69439C69439C69439
CE9C42C69439CE8439C6AD39CEB542C6AD39CEB539C6AD39CEB542C6B539CEB539C6B539
CEB539CEB539CEB539C6B539CEBD42CEB539CEBD39CEB539D6BD39CEB539CEBD39CE8C31
D69439CE8C39D69439CE8C31D69439CE8C39D69439CE8C31D69439CE8C39D69439CE8C31
D69439CE8C39D69439CE8C31D6CE39CEC631D6C639CEC631D6CE39CE8C31D69439CE8C31
D69439CE8C31D69439CE8C31D69439CE8C31CE9439CE8C31D69439CE8C31CE9439CE8C31
D69439CE8C31CE9439CE8C31CE9439CE8C31CE6B39C66B31CE7339CE6B31CE6B39C66B31
CE7339CE6B31CE8439C67B31CE8439C68431CE8439C67B31CE8439C68431CE8439C67B31
CE8439C67B31CE8439C67B31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000B5A539B5A539B59C39B59C39B59C39
B59C39B59C39B59C39B59C39B59C39BD9C39B59C39BD9C39BD9C39BD9C39BD9C39BD9C39
BD9C39BDA539BD9C39BDA539BD9C39BD9439BD8C39BD8C39BD9439C69439BD8C39C69439
C69439C69439C68C39C69439C69439C69439C69439C69439C69439C69439C69439C69439
C69439B5C639C6B539CEB539C6B539CEB539C6B539CEB539CEB539CEB539CEB539CEB539
CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEBD39CEB539CE9439
CE9439CE9439CE9439CE9439CE9439CE9439CE8C39CE9439CE9439D69439CE9439CE9439
CE9439D69439CE9439CEC639CEC639D6C639CEC639CEC639CEC639D69439CE9431CE9439
CE9439D69439CE9431CE9439CE9439CE9439CE8C31CE9439CE9439CE9439CE9431CE9439
CE9439CE9439CE8C31CE9439CE9439CE9439CE9431CE6B39CE6B39CE6B39CE6B31CE6B39
CE6B39CE8439CE8431CE8439CE8439CE8439C68431CE8439CE8439CE8439C68431CE8439
C68431CE8439C68431C68439000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000AD9C31B5A539AD9C31B5A539AD9431B59C39B59439
B59C39B59431BD9C39B59C39BD9C39B59C39BDA539B59C39BD9C39B59C39BDA539B59C39
BDA539BD9C39C6A542BD9C39C6A539BD9C39C69442BD8C39C69442BD8C39C69442BD8C39
C69439BD8C39C69442BD8C39C69442C68C39C69442C69439C69439C69439CE9442ADBD39
B5C642B5C639B5C642C6AD39CEB542C6AD39CEB542C6B539CEB539C6B539CEB542C6B539
CEB539C6B539CEBD39CEB539CEBD39CEB539CEBD39CEB539CEBD39CEB539D6BD39CEB539
D69439CE8C39D69439CE8C39D69439CE8C31D69439CE8C39D69439CE8C39D69439CE8C39
D69439CEC631D6CE39CEC631D6C639CEC631D6CE39CEC631D6C639CEC631D69439CE8C31
D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31
CE9439CE8C31CE9439CE8C31CE9439CE8C31CE9439CE6B31CE6B39C66B31CE6B39C67B31
CE8439C67B31CE8439C68431CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C67B31CE8439C67B31CE8439000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000AD9C39B5A539AD9C39B5A539ADA539B5A539B5A539B5A539B59C39B59C39
B59C39B59C39B59C39B59C39B59C39BD9C39B59C39BD9C39B59C39BD9C39BD9C39BD9C39
BD9C39BD9C39BD9C39BDA539BD9C39BDA539BD8C39C69439BD8C39C69439BD9439C69439
BD8C39C69439C69439C69439C69439C69439C69439C69439C69439B5C639B5C639B5C639
B5C639B5C639B5C639CEB539C6B539CEB539CEB539CEB539CEB539CEB539CEB539CEB539
CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEBD39CEB539CEB539CEB539CEBD39
CE9439CE9439CE9439D69439CE9439CE9439CE9439CE9439CE9439D69439CE9439D6C639
CEC639CEC639CEC639D6C639CEC639D6C639CEC639D6C639CEC639CEC639CE9439D69439
CE9439CE9439CE9439CE9439CE9431CE9439CE9439CE9439CE9431CE9439CE9439CE9439
CE9431CE9439CE9431CE9439CE9431CE9439CE9439CE9439CE6B31CE8439CE8431CE8439
CE8431CE8439CE8431CE8439C68431CE8439C68431CE8439C68431CE8439C68431CE8439
C68431C68439C68431C68439C68431000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
B5A539AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C31B5A539B59C39B59C39AD9431
BD9C39B59C39B59C39B59431BD9C39B59C39BD9C39B59C39BDA542B59C39BDA539B59C39
BDA542BD9C39BDA539BD9C39C6A542BD9C39C6A539BD8C39C69442BD8C39C69439BD8C39
C69442BD8C39C69439BD8C39C69442C69439C69439C68C39B5C642ADBD39B5C642ADBD39
B5C642B5C639B5C642B5C639CEB542C6B539CEB542C6B539CEB542C6B539CEB539C6B539
CEBD42CEB539CEB539CEB539CEBD42CEB539CEBD39CEB539D6BD42CEB539D6BD39CEB539
D6BD42CE8C39D69439CE8C39D69439CE9439D69439CE8C31D69439CE9439D6C639CEC631
D6CE39CEC639D6C639CEC631D6CE39CEC631D6C639CEC631D6CE39CEC631D6C639CE8C31
D69439CE9431D69439CE8C31D69439CE9431D69439CE8C31D69439CE8C31CE9439CE8C31
D69439CE8C31CE9439CE8C31D69439CE8C31CE9439C6B531CE8439CE8431CE8439C67B31
CE8439C68431CE8439C67B31CE8439C68431CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C67B31CE8439C67B31CE8439000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000AD9C31AD9C39
AD9C39ADA539AD9C39AD9C39AD9C39B5A539AD9C39B5A539B5A539B5A539B5A539B59C39
B59C39B59C39B59C39B59C39B59C39BD9C39B59C39BD9C39BD9C39BD9C39BD9C39BD9C39
BD9C39BDA539BD9C39BD9C39BD9C39BDA539BD9C39BD9439BD8C39C69439BD8C39C69439
C69439C69439C68C39C69439C69439C69439ADBD39B5C639B5C639B5C642B5C639B5C639
B5C639B5C642B5C639B5C639CEB539CEB542C6B539CEB539CEB539CEB542CEB539CEB539
CEB539CEB539CEB539CEB539CEB539CEBD39CEB539CEB539CEB539CEBD39CEB539CEBD39
CEB539D6BD39CE9439CE9439CE9439D69439CE9439CE9439CEC639D6C639CEC639CEC639
CEC639D6C639CEC639CEC639CEC639D6C639CEC639CEC639CEC639D6C639CEC631CE9439
CE9439D69439CE9431CE9439CE9439D69439CE9431CE9439CE9439CE9439CE9431CE9439
CE9439CE9439CE9431CE9439CE9439CE9439CEB531CEBD39CEBD39CE8439CE8431CE8439
CE8439CE8439C68431CE8439CE8439CE8439C68431CE8439C68439CE8439C68431C68439
C68439CE8439C68431C68439C68431C68439000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000AD9C39A59C31ADA539AD9C31
ADA539AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C31B5A539B59C31
B59C39B59431BD9C39B59C39BD9C39B59C39BD9C39B59C39BD9C39B59C39BDA542B59C39
BDA539BD9C39BDA542BD9C39C6A542BD9C39C6A542BD9C39C69442BD8C39C69442BD8C39
C69442BD8C39C69442BD8C39B5C642ADBD39B5C642ADBD39B5C642ADBD39B5C642B5C639
B5C642B5C639B5CE42B5C639BDCE42C6B539CEB542C6B539CEB542C6B539CEBD42CEB539
CEBD42CEB539CEBD42CEB539CEBD42CEB539D6BD42CEB539D6BD39CEB539D6BD42CEB539
D6BD39CEB539D69439CE8C39D69439CE8C39D6CE39CEC639D6C639CEC639D6CE39CEC639
D6CE39CEC631D6CE39CEC639D6CE39CEC631D6CE39CEC631D6CE39CEC631D6CE39CEC631
D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31D69439CE8C31
D69439CE8C31D69439CE8C31CEBD39CEB531CEBD39CEB531CEBD39C67B31CE8439C68431
CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C67B31CE8439C67B31CE8439C67B31000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39
AD9C39AD9C39AD9C39ADA539AD9C39B5A539AD9C39B5A539B5A539B5A539B5A539B5A539
B59C39B59C39B59C39B59C39B59C39B59C39B59C39BD9C39B59C39BD9C39BD9C39BD9C39
BD9C39BD9C39BD9C39BDA539BD9C39BDA539BDA539C6A539BD8C39C69439BD8C39C69439
C68C39C69439ADBD39B5C639ADBD39B5C639B5C639B5C642B5C639B5C642B5C639B5C642
B5C639B5C642B5C639B5C642B5C639CEB542CEB539CEB542CEB539CEB542CEB539CEB542
CEB539CEB539CEB539CEB539CEB539CEB539CEB539CEBD39CEB539CEBD39CEB539CEBD39
CEB539CEBD39CEBD39D69439CE9439D6C639CEC639D6C639CEC639D6C639CEC639D6C639
CEC639D6C639CEC639D6C639CEC639D6C639CEC639D6C639CEC639D6C639CEC639D6C639
CE9439D69439CE9439D69439CE9439CE9439CE9439CE9439CE9431CE9439CE9439CE9439
CE9431CE9439CEBD39CEBD39CEB531CEBD39CEB531CEBD39CEB531CEBD39CE8431CE8439
CE8431CE8439CE8431CE8439C68431CE8439C68431CE8439C68431CE8439C68431CE8439
C68431C68439C68431C68439C68431C68439C68431000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000AD9C39A59C31AD9C39A59C31ADA539AD9C31AD9C39A59C31
B5A539AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C39B5A539AD9C31
B5A539B59439B59C39B59431BD9C39B59C39BD9C39B59C39BDA542B59C39BD9C39B59C39
BDA542BD9C39BDA539BD9C39C6A542BD9C39C6A542BD9C39C6A542BD8C39C69442BD8C39
B5C642ADBD39B5C642ADBD39B5C642ADBD39B5C642ADBD39B5C642B5C639B5C642ADBD39
B5CE42B5C639B5C642B5C639BDCE42B5C639CEB542C6B539CEBD42CEB539CEBD42C6B539
CEBD42CEB539CEBD42CEB539D6BD42CEB539CEBD42CEB539D6BD42CEB539D6BD39CEB539
D6BD42CEB539D6BD39CEC639D6CE42CEC639D6C639CEC639D6CE39CEC639D6C639CEC631
D6CE39CEC639D6C639CEC631D6CE39CEC639D6CE39CEC631D6CE39CEC631D6C639CEC631
D6CE39CE9431D69439CE8C31D69439CE9431D69439CE8C31D69439CE9431D69439CE8C31
D69439CEB531CEBD39CEB531D6BD39CEB531CEBD39C6B531CEBD39CEB531CE8439C67B31
CE8439C68431CE8439C67B31CE8439C67B31CE8439C67B31CE8439C68431CE8439C67B31
CE8439C67B31CE8439C67B31CE8439C67B31C68439000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000A59C31A59C39A59C39AD9C39A59C31AD9C39AD9C39AD9C39AD9C31AD9C39
AD9C39ADA539AD9C39AD9C39AD9C39B5A539AD9C39B5A539B5A539B5A539B5A539B5A539
B5A539B59C39B59C39B59C39B59C39BD9C39B59C39BD9C39B59C39BD9C39B59C39BD9C39
BD9C39BDA539BD9C39BD9C39BD9C39BDA539BD9C39BDA539BDA539C69442BD8C39ADBD39
ADBD39B5C642ADBD39B5BD39B5C639B5C642ADBD39B5C642B5C639B5C642B5C639B5C642
B5C642B5C642B5C639B5C642B5C642B5CE42CEB542CEB542CEB542CEB542CEB539CEB542
CEB542CEBD42CEB539CEB542CEB539CEBD42CEB539CEBD39CEB539D6BD42CEB539CEBD39
CEBD399CCE399CCE39D6C639CEC639D6C639CEC639CEC639CEC639D6C639CEC639D6C639
CEC639D6C639CEC639CEC639CEC639D6C639CEC639D6C639CEC639D6C639CEC631CEC639
CEC639D69439CE9431CE9439CE9439D69439CE9431CE9439CE9439D69439CE9431CEBD39
CEBD39CEBD39CEB531CEBD39CEBD39CEBD39CEB531CEBD39CEB539CEBD39CE8431CE8439
CE8439CE8439C68431CE8439CE8431CE8439C68431C68439C68439CE8439C68431C68439
C68431C68439C67B31C68439C68439C68439C69431C69C39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
A59C39A59431AD9C39A59431AD9C39A59431AD9C39A59C31AD9C39A59C31ADA539AD9C31
ADA539AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C31B5A539B59C39
B5A539B59C31B59C39B59439BD9C39B59C39BD9C39B59C39BD9C39B59C39BDA539B59C39
BDA542BD9C39BDA542BD9C39BDA542BD9C39C6A542BD9C39C6A5428CBD39B5C642ADBD39
B5C642ADBD39B5C642ADBD39B5C642ADBD39B5C642ADBD39B5C642ADBD39B5C642B5C639
B5C642B5C639BDCE42B5C639BDCE42B5C639BDCE42C6B539CEBD42CEB539CEBD42CEB539
CEBD42CEB539CEBD42CEB539D6BD42CEB539D6BD42CEB539D6BD42CEB539D6BD42CEB539
9CCE4294CE399CD64294CE39D6CE39CEC639D6CE39CEC639D6CE39CEC639D6CE39CEC639
D6CE39CEC631D6CE39CEC639D6CE39CEC631D6CE39CEC631D6CE39CEC631D6CE39CEC631
D6CE39CE9431D69439CE9431D69439CE8C31D69439CE9431D69439CE8C31D6BD39CEB531
D6BD39CEB531D6BD39CEB531CEBD39CEB531CEBD39CEB531CEBD39C6B531CEBD39C67B31
CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C67B31CE8439C67B31CE9C39C69431C69C39C69431000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000A59C39A59439A59C39
A59431A59C39A59C39A59C39A59C31A59C39A59C39AD9C39A59C31AD9C39AD9C39AD9C39
AD9C39AD9C39AD9C39ADA539AD9C39ADA539AD9C39B5A539AD9C39B5A539B5A539B5A539
B5A539B5A539B59C39B59C39B59C39B59C39B59C39BD9C39B59C39BD9C39BD9C39BD9C39
BD9C39BD9C39BD9C39BDA539BD9C39BDA539BD9C398CBD398CBD3994BD42ADBD39ADBD42
ADBD39ADBD42ADBD39B5C642ADBD39B5C642B5C642B5C642B5C639B5C642B5C642B5C642
B5C642B5C642B5C642B5C642B5C642B5C642B5C642CEB542CEB542CEB542CEB542CEBD42
CEB542CEB542CEB542CEBD42CEB542CEBD42CEB542CEBD42CEB539CEBD429CCE399CCE42
9CCE399CCE399CCE399CCE39CEC639D6C639CEC639D6C639CEC639D6C639CEC639D6C639
CEC639D6C639CEC639D6C639CEC639D6C639CEC639D6C639CEC639D6C639CEC639D6C639
CEC639D6C639CE9439D69439CE9431D69439CE9439D69439CEBD39CEBD39CEBD39CEBD39
CEB531CEBD39CEBD31CEBD39CEB531CEBD39CEBD39CEBD39CEB531CEBD39CEB531CE8439
CE8431CE8439CE8431CE8439C68431CE8439C68431CE8439C68431CE8439C68431CE8439
C68431C68439C68431C69C39C69431C69C39C69431C69C39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000009C9431A59C399C9431A59C399C9431
A59C39A59431A59C39A59431AD9C39A59C31AD9C39A59431AD9C39A59C31AD9C39A59C31
ADA539AD9C31ADA539AD9C31B5A539AD9C39B5A539AD9C31B5A539AD9C39B5A539AD9C31
B5A539B5A539B5A539B59431BD9C39B59C39BD9C39B59439BDA542B59C39BD9C39B59C39
BDA542BD9C39BDA539BD9C39C6A5428CBD3994BD428CB53994C6428CBD3994BD42ADBD39
B5C642ADBD39B5C642ADBD39B5C642ADBD39B5C642ADBD39B5C642B5C639B5C642ADC639
B5CE42B5C642B5CE42B5C639BDCE42B5C642BDCE42B5C642CEBD42CEB542CEBD42CEB539
D6BD42CEB542CEBD42CEB539D6BD42CEB539D6BD42CEB539D6BD4294CE399CCE4294CE39
9CD64294CE399CCE4294CE399CD642CEC639D6CE39CEC639D6CE42CEC639D6CE39CEC639
D6CE39CEC639D6CE39CEC631D6CE39CEC639D6CE39CEC631D6CE39CEC631D6CE39CEC631
D6CE39CEC631D69439CE8C31D69439CE9431D6BD39CEB531D6BD39CEB531D6BD39CEB531
D6BD39CEB531CEBD39CEB531D6BD39CEB531CEBD39C6B531CEBD39CEB531CEBD39C6B531
CE8439C68431CE8439C67B31CE8439C68431CE8439C67B31CE8439C67B31CE8439C67B31
CE8439C69431C69C39C69431CE9C39C69431C69C39BD9431C69C39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000A594399C9431A59439A59439A59C39A59431A59C39
A59C39A59C39A59431A59C39A59C39AD9C39A59C31A59C39A59C39AD9C39A59C31AD9C39
AD9C39AD9C39AD9C39AD9C39AD9C39B5A539AD9C39B5A539B5A539B5A539AD9C39B5A539
B5A539B5A539B5A539B59C39B59C39B59C39B59C39B59C39B59C39BD9C39B59C39BD9C39
BD9C39BD9C39BD9C398CBD398CBD398CBD428CBD398CBD428CBD3994BD428CBD39ADBD42
ADBD42B5C642ADBD42B5C642B5C642B5C642ADC642B5C642B5C642B5C642B5C642B5C642
B5C642B5C642B5C642B5C642B5C642BDCE42B5C642B5CE42CEB542CEBD42CEB542CEB542
CEB542CEBD42CEB542CEBD42CEBD42CEBD42CEB5429CCE429CCE429CCE429CCE399CCE42
9CCE399CCE429CCE399CCE399CCE39D6CE39CEC639D6C639D6C639D6CE39CEC639D6C639
D6C639D6CE39CEC639D6C639D6C639D6CE39CEC639D6C639D6C639D6CE39CEC639D6C639
CEC639D6CE39CEC631D69439CE9439D6BD39CEBD31CEBD39CEBD39D6BD39CEB531CEBD39
CEBD39CEBD39CEB531CEBD39CEBD39CEBD39CEB531CEBD39CEBD39CEBD39CEB531CEBD39
CE8439CE8439C68431CE8439CE8439CE8439C68431C68439C68431CE8439C68431C69C39
C69C31C69C39C69431C69C39C69431C69C39C69431C69439C69431000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000009C9431A59C399C9431A59C399C9431A59C399C9431A59C399C9431
A59C39A59431A59C39A59431AD9C39A59431AD9C39A59431AD9C39A59C31AD9C39A59C31
AD9C39A59C31ADA539AD9C31B5A539AD9C31B5A539AD9C39B5A539AD9C31B5A539AD9C39
B5A539B59C39B5A539B5A539B59C39B59439BD9C39B59C39BD9C39B59C39BDA542B59C39
BDA5398CB5398CBD428CB53994BD428CB53994BD428CBD3994BD428CBD3994C6428CBD39
B5C642ADBD39B5C642ADBD42B5C642ADBD39B5C642ADBD42B5C642ADC642B5C642B5C642
B5CE42B5C642BDCE42B5C642BDCE42B5C642BDCE4AB5C642BDCE42CEB542CEBD42CEB542
CEBD42CEB542D6BD42CEB542D6BD4294CE399CCE4294CE399CCE4294CE399CD64294CE39
9CD64294CE399CD64294CE399CD64294CE39D6CE42CEC639D6CE39CEC639D6CE39CEC639
D6CE39CEC639D6CE39CEC639D6CE39CEC631D6CE39CEC631D6CE39CEC631D6CE39CEC631
D6CE39CEC631D6CE39CEB531D6BD39CEB531D6BD39CEB531D6BD39CEB531D6BD39CEB531
D6BD39CEB531D6BD39CEB531CEBD39CEB531CEBD39CEB531CEBD39C6B531CEBD39C6B531
CEBD39C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE8439C67B31CE9C39C69431
CE9C39C69431CE9C39C69431C69C39C69431C69C39BD9431C69C39BD9431000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
9C8C319C8C399C8C319C94399C94319C94399C9439A594399C9431A59439A59431A59C39
A59431A59C39A59439A59C39A59C31A59C39A59C39AD9C39A59C31AD9C39AD9C39AD9C39
AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39B5A539AD9C39B5A539B5A539B5A539
B5A539B5A539B5A539B5A539B5A539B59C39B59C39B59C39B59C39BD9C39B59C398CBD39
8CB5398CBD398CBD398CBD428CBD398CBD428CBD3994BD428CBD4294BD428CBD4294BD42
ADBD42B5BD42ADBD42B5C642ADBD42B5C642B5C642B5C642B5C642B5C642B5C642B5C642
B5C642B5C642B5C642B5CE42B5C642B5CE42B5C642BDCE42B5CE42CEBD42CEB542CEBD42
CEB542CEBD42CEB542CEBD4294CE429CCE429CCE429CCE429CCE429CCE429CCE429CCE42
9CCE399CCE429CCE399CD6399CCE399CD639D6CE39D6CE39CEC639D6CE39D6C639D6CE39
CEC639D6CE39D6C639D6CE39CEC639D6CE39D6C639D6CE39CEC639D6CE39D6C639D6CE39
CEC639D6CE399CCE399CCE39CEBD39D6BD39CEBD39D6BD39CEBD31CEBD39CEBD39CEBD39
CEBD31CEBD39CEBD39CEBD39CEB531CEBD39CEB531CEBD39CEB531CEBD39CEB531CEBD39
CEB531CE8439CE8431CE8439C68431CE8439C68431CE8439C69431C69C39C69431C69C39
C69431C69C39C69431C69C39C69431C69C39C69431C69C39C69431C69439000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000009C8C39948431
9C8C399C8C319C8C399C8C31A59C399C9431A594399C9431A59C399C9431A59C399C9431
A59C39A59431A59C39A59431AD9C39A59431AD9C39A59431AD9C39A59C31AD9C39A59431
ADA539AD9C31AD9C39AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C39B5A539AD9C31
B5A539B59C39B5A539B59C31BDA539B59C39BD9C39B59439BD9C4284B5398CBD3984B539
8CBD428CB5398CBD428CB53994BD428CBD3994BD428CB53994C6428CBD4294BD428CBD39
94C642ADBD42B5C642ADBD42B5C64AADBD42B5C642ADBD42B5C64AB5C642B5C642B5C642
BDCE4AB5C642B5CE42B5C642BDCE4AB5C642BDCE42B5C642BDCE4AB5CE42CEBD4ACEB542
D6BD4ACEB5429CCE4294C6429CD64A94CE429CCE4294CE399CD64294CE399CCE4294CE39
9CD6429CCE399CD64294CE399CD6429CCE399CD639CEC639D6CE42CEC639D6CE39CEC639
D6CE39CEC639D6CE39CEC631D6CE39CEC639D6CE39CEC631D6CE39CEC639D6CE39CEC631
9CD6399CCE319CD63994CE31D6BD39CEB531D6BD39CEB531D6BD39CEB531D6BD39CEB531
D6BD39CEB531CEBD39CEB531D6BD39CEB531CEBD39C6B531CEBD39CEB531CEBD39C6B531
CEBD39C6B531CE8439C67B31CE8439C67B31CE9C39C69431CE9C39C69431CE9C39C69431
CE9C39C69431C69C39BD9431C69C39C69431C69C39BD9431C69C39BD9431C69C39000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000009C8C319C8C39948C319C8C39
9C8C399C8C399C8C319C8C399C94399C94399C94319C94399C9439A59C399C9431A59439
A59439A59C39A59431A59C39A59C39A59C39A59C31A59C39A59C39AD9C39A59C31AD9C39
AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39B5A539AD9C39B5A539
B5A539B5A539B5A539B5A539B5A539B5A539B59C398CB5398CB5398CB5398CB5398CB539
8CB5398CBD428CB5398CBD428CBD428CBD428CBD398CBD428CBD4294BD428CBD4294BD42
94BD42B5C642ADBD42B5C642B5C642B5C642B5C642B5C642B5C642B5C642B5C642B5C642
B5C642B5C642B5C642B5C642B5C642BDCE42B5C642B5CE42BDCE42BDCE4ACEB542CEBD42
CEBD429CCE4294CE429CCE429CCE429CCE4294CE429CCE429CCE429CCE429CCE399CCE42
9CCE429CD6429CCE399CD6399CCE399CD6399CCE39D6CE39D6CE39D6CE39D6C639D6CE39
D6CE39D6CE39CEC639D6CE39D6CE39D6CE39CEC639D6CE39D6CE39D6CE39CEC6399CCE39
9CCE399CD6399CCE399CCE399CCE39D6BD39CEBD31CEBD39CEBD39D6BD39CEBD31CEBD39
CEBD39CEBD39CEB531CEBD39CEBD39CEBD39CEB531CEBD39CEB539CEBD39CEB531CEB539
CEB539CEBD39C68431CE8439C69C31CE9C39C69431C69C39C69C39C69C39C69431C69C39
C69431C69C39C69431C69439C69439C69C39C69431C69439C69431C69C39BD9431000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000009484319C8C399484319C8C399484319C8C39948431
9C8C399484319C8C399C8C319C8C399C9431A594399C9431A59C399C9431A59C399C9431
A59C399C9431A59C39A59431A59C39A59431AD9C39A59431AD9C39A59431AD9C39A59C31
AD9C39A59C31ADA539AD9C31ADA539AD9C31B5A539AD9C31B5A539AD9C31B5A539AD9C39
B5A539AD9C31B5A539B59C39B5A53973B5398CB53984B5398CBD4284B5398CBD428CB539
8CBD428CB5398CBD428CB5398CBD428CB53994BD428CBD3994BD428CBD4294C6428CBD42
94C6428CBD42B5C64AADBD42B5C64AADBD42B5C64AADBD42B5C64AADC642B5C64AB5C642
B5CE4AB5C642BDCE4AB5C642BDCE4AB5C642BDCE4AB5C642BDCE4AB5C642BDCE4A94CE42
9CCE4A94CE429CCE4A94CE429CCE4294CE429CD64294CE429CD64294CE399CD64294CE39
9CD64294CE399CD6429CCE399CD64294CE399CD6429CCE39D6CE39CEC639D6CE42CEC639
D6CE39CEC639D6CE39CEC639D6CE39CEC639D6CE39CEC639D6CE3994CE319CD63994CE31
9CD63994CE319CD63994CE319CD639CEB531D6BD39CEB531D6BD39CEB531D6BD39CEB531
D6BD39CEB531CEBD39CEB531CEBD39CEB531CEBD39CEB531CEBD39C6B531CEBD39C6B531
CEBD39C6B531CEBD39C69431CE9C39C69431CE9C39C69431CE9C39C69431CE9C39C69431
C69C39C69431C69C39C69431C69C39BD9431C69C39BD9431C69C39BD9431C69C39000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000948C39948431948C39948C319C8C399484319C8C39948C319C8C39
9C8C319C8C399C8C319C8C399C8C319C8C399C94319C94399C94319C94399C9439A59C39
9C9431A59C39A59439A59C39A59431A59C39A59C39A59C39A59C31A59C39A59C39AD9C39
A59C39AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39ADA539AD9C39B5A539
AD9C39B5A539B5A5397BB53973B5397BB5397BB5398CB5398CB5398CB5398CB5398CBD42
8CB5398CBD428CBD428CBD428CBD428CBD428CBD428CBD428CBD4294BD428CBD4294BD42
8CBD4294BD42ADBD42B5C642B5BD42B5C642B5C642B5C64AB5C642B5C642B5C642B5C64A
B5C642B5C64AB5C642B5CE4AB5C642B5CE4AB5C642BDCE4AB5CE425ACE4A5ACE429CCE4A
94CE429CCE429CCE429CCE429CCE429CCE429CCE429CCE429CCE429CCE429CCE429CD642
9CCE399CD6429CCE399CD6429CCE399CD6399CD6399CD639D6CE39D6CE39D6CE39D6CE39
D6CE39D6CE39D6CE39D6CE39D6CE39D6CE39D6CE39D6CE399CCE399CD6399CCE399CD639
9CCE399CD6399CCE399CCE399CCE319CCE39CEBD39D6BD39CEBD31CEBD39CEBD39CEBD39
CEBD31CEBD39CEBD31CEBD39CEB531CEBD39CEBD39CEBD39CEB531CEBD39CEB531CEBD39
CEB531B5CE39B5CE31CE9C39C69431C69C39C69431C69C39C69431C69C39C69431C69C39
C69431C69C39C69431C69C39C69431C69439C69431C69C39BD9431C69439BD9431C69439
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000948431948C399484319C8C399484319C8C399484319C8C399484319C8C39948431
9C8C39948C319C8C399484319C8C399C8C319C94399C9431A59C399C9431A594399C9431
A59C399C9431A59C399C9431A59C39A59431A59C39A59431AD9C39A59431AD9C39A59431
AD9C39A59C31AD9C39A59C31ADA539AD9C31ADA539AD9C31B5A539AD9C39B5A539AD9C31
B5A53973AD397BB53973AD397BB53973B5397BB53984B5398CBD4284B5398CBD4284B539
8CBD428CB5398CBD428CB53994BD428CBD4294BD428CB54294BD4A8CBD4294BD428CBD42
94C64A8CBD4294C64AADBD42B5C64AADBD42B5C64AADBD42B5C64AB5C642B5C64AB5C642
BDCE4AB5C642B5CE4AB5C642BDCE4AB5C642BDCE4AB5C6425ACE4A5ACE425ACE4A52CE42
9CD64A94CE429CCE4A94CE429CD64A94CE429CD64294CE399CD64294CE429CD64294CE39
9CD6429CCE399CD64294CE399CD6429CCE399CD64294CE399CD642D6CE39D6CE39CEC639
D6CE42D6CE39D6CE39CEC639D6CE39CECE399CD63994CE319CD6399CCE399CD63994CE31
9CD6399CCE319CD63994CE319CD63994CE31D6BD39CEB531D6BD39CEB531D6BD39CEB531
D6BD39CEB531CEBD39CEB531CEBD39CEB531CEBD39C6B531CEBD39CEB531CEBD39C6B531
BDCE39B5C631B5CE39C69431CE9C39C69431CE9C39C69431CE9C39C69431C69C39C69431
C69C39C69431C69C39BD9431C69C39BD9431C69C39BD9431C69C39BD9431C69C39BD9431
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C7331947B39
947B31948C39948431948439948431948C39948431948C39948C399C8C399484319C8C39
9C8C399C8C39948C319C8C399C8C399C8C399C8C319C94399C94399C94399C94319C9439
9C9439A59C399C9431A59439A59439A59C39A59431A59C39A59C39A59C39A59C31A59C39
A59C39AD9C39A59C31AD9C39AD9C39AD9C39AD9C39AD9C39AD9C39ADA539AD9C3973AD39
73B5397BB53973AD3973B53973B5397BB53973B5397BB5398CB5398CB5428CB5398CB542
8CB5428CBD428CB5428CBD428CBD428CBD428CBD428CBD428CBD4294BD428CBD4294BD42
94BD4294C64A94BD42B5C64AB5C64AB5C64AB5BD42B5C64AB5C64AB5C64AB5C642B5C64A
B5C64AB5C64AB5C642B5C64AB5C64ABDCE4A5AC6425ACE4A5ACE4A5ACE4A5ACE425ACE4A
9CCE429CCE4A94CE429CCE429CCE429CD6429CCE429CCE429CCE429CD6429CCE429CD642
9CCE429CD6429CCE399CD6429CD6399CD6429CD6399CD6399CD639D6CE39D6CE39D6CE39
D6CE39D6CE39D6CE39D6CE39D6CE399CD6399CCE399CD6399CD6399CD6399CCE399CD639
9CCE399CD6399CCE399CCE399CCE399CCE399CCE31CEBD39CEBD39D6BD39CEBD31CEBD39
CEBD39CEBD39CEB531CEBD39CEBD39CEBD39CEB531CEBD39CEBD39CEBD39B5CE31B5CE39
B5CE39B5CE39B5CE31B5CE39C69C39CE9C39C69431C69C39C69431C69C39C69431C69439
C69431C69C39C69431C69439C69431C69C39BD9431C69439BD9431C69439BD9431BD9439
BD9431000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000947B398C7331947B398C7331
947B398C7331948C39948431948C399484319C8C399484319C8C399484319C8C39948431
9C8C399484319C8C399484319C8C399484319C8C399C94319C94399C9431A59C399C9431
A59C399C9431A59C399C9431A59C399C9431A59C39A59431A59C39A59431AD9C39A59431
AD9C39A59431AD9C39A59C31AD9C39A59C31ADA539AD9C31ADA53973AD317BB53973AD39
7BB53973AD397BB53973AD397BB53973AD397BB54273B5398CB54284B5398CBD428CB539
8CBD428CB5398CBD428CB54294BD428CB54294BD4A8CBD4294BD4A8CBD4294C64A8CBD42
94C64A8CBD4294C64A8CBD42B5C64AADBD42B5C64AADBD42B5C64AB5C642B5C64AB5C642
B5C64AB5C642BDCE4AB5C6425ACE4A52C6425ACE4A5AC6425ACE4A52CE425AD64A5ACE42
9CD64A94CE429CD64A94CE429CD64A94CE429CD64A94CE429CD64294CE429CD64294CE39
9CD64294CE399CD6429CCE399CD6429CCE399CD6429CCE399CD6429CCE39D6CE42D6CE39
D6CE39CECE39D6CE399CD6399CD6399CCE399CD6399CCE399CD63994CE319CD6399CCE31
9CD63994CE319CD63994CE319CD63994CE319CD639CEB531D6BD39CEB531D6BD39CEB531
D6BD39CEB531D6BD39CEB531CEBD39CEB531CEBD39CEB531CEBD39B5C631BDCE39B5C631
B5CE39B5C631B5CE39B5C631CE9C39C69431CE9C39C69431C69C39C69431C69C39C69431
C69C39BD9431C69C39BD9431C69C39BD9431C69C39BD9431C69C39BD9431C69C39BD9431
C6AD39000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008C7B398C7331947B398C7331947B398C7331947B39
8C7331947B39947B31948439948431948439948431948C39948431948C39948C319C8C39
9484319C8C39948C319C8C399C8C319C8C399C8C319C8C399C94319C94399C94319C9439
9C9431A594399C9439A59439A59431A59C39A59439A59C39A59431A59C39A59C39A59C39
A59C39AD9C39A59C39AD9C39AD9C39AD9C39AD9C3973AD3973AD3973AD3973AD3973B539
73AD3973B53973B5397BB53973B5397BB53973B5397BB5397BB5398CB5398CB5398CB542
8CB5398CBD428CB5428CBD428CBD428CBD428CBD428CBD428CBD4294BD428CBD4294BD4A
8CBD4294BD4A94BD4A94C64AB5BD4AB5C64AB5C64AB5C64AB5C64AB5C64AB5C64AB5C64A
B5C64AB5C64AB5C64A5ACE4A5AC64A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A
5ACE429CD64A9CCE429CD64A9CCE429CD6429CCE429CD6429CCE429CD6429CD6429CD642
9CD6429CD6429CD6429CD6429CD6399CD6429CD6399CD6399CD6399CD639D6CE39D6CE39
D6CE39D6CE399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CCE399CD639
9CCE399CD6399CCE399CD6399CCE399CCE399CCE399CCE39CEBD31CEBD39CEBD39CEBD39
CEBD31CEBD39CEBD39CEBD39CEB531CEBD39CEB531CEBD39B5CE31B5CE39B5CE31B5CE39
B5CE31B5CE39B5CE31B5CE39B5C631C69C39C69431C69C39C69431C69C39C69431C69C39
C69431C69439C69431C69C39BD9431C69439BD9431C69439BD9431BD9439BDA531BDAD39
BDA531BDA539000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008C7331947B398C7331947B398C7331947B398C7331947B398C7331
947B398C7331947B398C8431948C39948431948C399484319C8C399484319C8C39948431
9C8C399484319C8C399484319C8C399484319C8C399484319C8C399C94319C94399C9431
A59C399C9431A594399C9431A59C399C9431A59C399C9431A59C39A59431A59C39A59431
AD9C39A59C31AD9C39A59431AD9C3973AD3173AD396BAD317BB53973AD3973B53973AD31
7BB53973AD397BB53973AD397BB54273B5397BB53973AD397BB54284B5398CB54284B539
8CBD428CB5398CBD428CB53994BD428CB54294BD428CB54294BD4A8CBD4294BD4A8CBD42
94C64A8CBD4294C64A8CBD4294C64AADBD4AB5C64AADBD42B5C64AB5C64AB5C64AB5C642
BDCE4A52C64A5ACE4A52C6425ACE4A5AC64A5ACE4A52C6425AD64A5ACE425AD64A5ACE42
5AD64A94CE429CD64A94CE429CD64A9CCE429CD64A94CE429CD64A9CCE429CD64294CE39
9CD6429CD6429CD64294CE399CD6429CD6399CD64294CE39A5D6429CD6399CD639CECE39
D6D6429CD6399CD63994CE399CD6429CD6399CD63994CE319CD6399CCE399CD63994CE31
9CD6399CCE399CD63994CE319CD63994CE319CD63994CE31D6BD39CEB531D6BD39CEB531
D6BD39CEB531CEBD39CEB531D6BD39CEB531BDCE39B5C631BDCE39B5CE31B5CE39B5C631
BDCE39B5C631B5CE39ADC631B5CE39C69431C69C39C69431CE9C39C69431C69C39BD9431
C69C39BD9431C69C39BD9431C69C39BD9431C69C39BD9431C69C39BDA531C6AD39BDA531
C6AD39BDA531000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C73318C7B398C73318C73398C73318C7B398C73318C73318C7331947B398C73318C7B39
8C7331947B398C7331947B39948439948C39948431948439948431948C39948431948C39
948C399C8C399484319C8C399C8C399C8C39948C319C8C399C8C399C94399C94319C9439
9C94399C94399C94319C9439A59439A59C39A59431A59C39A59439A59C39A59431A59C39
A59C39AD9C3973AD3173AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD39
73AD3973B53973AD3973B53973B5397BB53973B5397BB5397BB5397BB5398CB5398CB539
8CB5428CBD428CB5428CBD428CBD428CBD428CBD428CBD428CBD4294BD4A8CBD4294BD4A
94BD4A94C64A8CBD4A94C64A94C64AB5C64AB5C64AB5C64AB5C64AB5C64AB5C64AB5C64A
5AC64A5ACE4A5AC64A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A
5ACE4A5AD64A9CCE429CD64A9CD6429CD64A9CD6429CD6429CD6429CD6429CD6429CD642
9CD6429CD6429CD6429CD6429CD6429CD6429CD6399CD6399CD6399CD6429CD6399CD639
9CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CCE399CD639
9CD6399CD6399CCE399CCE399CCE399CD63994CE319CCE399CCE39D6BD39CEBD31CEBD39
CEBD39CEBD39CEB531CEBD39CEBD39BDCE39B5CE31B5CE39B5CE39B5CE39B5CE31B5CE39
B5CE39B5CE39B5C631B5CE39B5CE31B5CE39C69431C69439C69439C69C39C69431C69439
C69431C69C39BD9431C69439C69431C69C39BD9431BDAD39BDA531BDAD39BDA531BDA531
BDA531BDAD39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C6B39846331
8C7B398473318C7B398C73318C7B398C73318C7B398C7331947B398C7331947B398C7331
947B398C7331947B398C7331947B398C8431948C398C8431948C39948431948C39948431
9C8C399484319C8C399484319C8C399484319C8C399484319C8C399484319C94399C9431
9C94399C9431A59C399C9431A59C399C9431A59C399C9431A59C399C9431A59C39A59431
73AD396BA53173AD396BA53173AD396BA53173AD3973AD3173AD3973AD397BB53973AD39
7BB53973AD397BB53973AD397BB54273AD397BB54273B5397BB54273B5398CB54284B539
8CBD428CB5398CBD428CB5428CBD428CB54294BD4A8CB54294BD4A8CBD4294C64A8CBD42
94C64A8CBD4294C64A8CBD4A94C64A8CBD4AB5C652B5BD4AB5C64AB5BD4A5AC64A52C64A
5ACE4A52C64A5ACE4A52C64A5ACE4A52C64A5ACE4A5ACE4A5AD64A5ACE425AD64A5ACE42
5AD64A5ACE429CD64A94CE429CD64A94CE429CD64A9CD6429CD64A94D6429CD64A9CD642
9CD6429CD6429CD6429CD6399CD6429CD6399CDE429CD6399CDE429CD63942DE4242D639
9CD6399CD6399CD6399CD6399CD6399CCE399CD6399CCE399CD63994CE319CD63994CE31
9CD63994CE319CD63994CE319CD63994CE319CD63994CE319CD639CEB531D6BD39CEB531
D6BD39CEB531D6BD39B5CE31BDCE39B5CE31BDCE39B5CE31BDCE39B5C631BDCE39B5C631
B5CE39B5C631B5CE39B5C631B5CE39ADC631C69C39C69431C69C39C69431C69C39BD9431
C69C39BD9431C69C39BD9431C69C39BD9431C6AD39BDA531C6AD39BDA531C6AD39BDA531
BDAD39BDA531BDAD39000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008C63398463318C63398C63318C6339
8C63318C73398C73318C73398C73318C73398C73318C73398C73318C73398C73318C7B39
8C73318C7B398C7331947B398C7331948439948431948439948431948439948431948C39
948431948C399484319C8C399484319C8C39948C319C8C399C8C319C8C399C94399C9439
9C94319C94399C94319C94399C9431A594399C9439A59C39A59431A59C396BA53973AD39
6BA53973AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD39
73AD3973AD3973AD397BB53973AD397BB53973B5397BB53973B5397BB5397BB5398CB542
8CB5398CB5428CB5428CBD428CB5428CBD428CBD428CBD428CBD4294BD4A8CBD4294BD4A
8CBD4A94BD4A94BD4A94C64A94BD4A94C64AB5C64AB5C64A52C64A5AC64A5AC64A5AC64A
5AC64A5ACE4A5AC64A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5AD64A
5ACE4A5AD64A5AD64A9CD64A9CD64A9CD64A9CD6429CD64A9CD6429CD6429CD6429CD642
9CD6429CD6429CD6429CD6429CD6429CD6429CD6429CD64242D63942D63942D63942D639
9CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD639
9CCE399CD6399CCE399CD6399CCE399CCE399CCE399CCE399CCE39CEBD39CEBD39CEBD39
CEBD31CEBD39B5CE31BDCE39B5CE31BDCE39B5CE31B5CE39B5CE31B5CE39B5CE31B5CE39
B5CE31B5CE39B5CE31B5CE39B5C631B5CE39B5C631C69C39C69431C69C39C69431C69C39
C69431C69439C69431C69439BDA531C6AD39BDA531BDAD39BDA531BDA539BDA531BDA539
BDA531BDA539BDA531000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008463318C63398463318C6B398463318C6339846331
8C6B398473318C73398473318C7B398C73318C7B398C7331947B398C73318C7B398C7331
947B398C7331947B398C7331947B398C73319484398C8431948C398C8431948C398C8431
948C39948431948C399484319C8C399484319C8C399484319C8C399484319C8C39948C31
9C94399C94319C94399C9431A59C399C9431A59C396BA53173AD396BA53173AD396BA531
73AD396BA53173AD396BA53173AD3973AD3173AD396BA53173AD3973AD3973AD3973AD39
7BB54273AD397BB53973AD397BB54273AD397BB54273AD397BB54273B5397BB54284B539
8CBD428CB5398CBD428CB53994BD428CB5428CBD428CB54294BD4A8CBD4294BD4A8CBD42
94C64A8CBD4294C64A8CBD4294C65294BD4A94C65252BD4A5AC65252C64A5AC64A52C64A
5ACE5252C64A5ACE4A52C64A5ACE525ACE4A5ACE4A52CE4A5AD64A5ACE4A5AD64A52CE42
5AD64A5AD64A5AD64A94D6429CDE4A9CD6429CD64A94D6429CDE4A9CD6429CD64A9CD642
9CDE4A9CD6429CDE429CD639A5DE429CD6429CDE4239D63942DE4242D63942D64239D639
A5DE429CD6399CD6399CCE399CD6399CD6399CD63994CE319CD6399CCE399CD63994CE31
9CD6399CCE399CD63994CE319CD63994CE319CD63994CE319CD63994CE31D6BD39CEB531
BDD639B5CE31BDCE39B5CE31BDCE39B5CE31BDCE39B5C631BDCE39B5CE31B5CE39B5C631
B5CE39B5C631B5CE39ADC631B5CE39ADC631B5CE39C69431C69C39C69431C69C39BD9431
C69C39BD9431C69C39BDA531C6AD39BDA531C6AD39BDA531C6AD39BDA531BDAD39B5A531
BDAD39BDA531BDA539B5A531000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008463318463318463318C63398463318463318463318C63398463318C6331
8463318C63398473318C73318C73318C73398C73318C73398C73318C73398C73318C7339
8C73318C7B398C73318C73398C7331947B398C7331948439948439948439948431948439
948431948C399484319484399484399C8C399484319C8C399C8C399C8C39948C319C8C39
9C94399C94399C94319C94399C94396BA5396BA5316BA5396BA53973A5396BA5316BA539
6BA53973AD396BA53173AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD39
73AD3973B53973AD3973B53973B5397BB53973AD397BB5397BB5397BB54273B5397BB539
8CB5428CB5428CB5428CB5428CB5428CBD428CB5428CBD428CBD4294BD4A8CBD4294BD4A
94BD4A94BD4A8CBD4A94BD4A94BD4A39C66339BD635AC64A5AC64A5AC6525AC64A5AC64A
5AC64A5ACE4A5AC64A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5AD64A5ACE4A5AD64A
5AD64A5AD64A5AD64A5AD64A9CD64A9CD64A9CD6429CD64A9CD6429CDE4A9CD6429CD642
9CD6429CDE429CD6429CD6429CD64242DE4242D64242D64242D63942D64242D63942D639
42D6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CCE399CD639
9CD6399CD6399CCE399CCE399CCE399CD6399CCE319CCE399CCE399CCE39CEBD31BDCE39
BDCE39BDCE39B5CE31BDCE39B5CE39BDCE39B5CE31B5CE39B5CE39B5CE39B5CE31B5CE39
B5CE39B5CE39B5C631B5CE39B5CE39B5CE39ADC631B5C639C69431C69C39C69431C69439
C69431C6AD39BDA531BDAD39BDA531C6AD39BDA531BDA531BDA531BDAD39BDA531BDA531
BDA531BDA539B5A531BDA531000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8463398463318C63398463318C63398463318C63398463318C63398463318C6B39846331
8C63398463318C6B398473318C73398473318C7B398473318C7B398473318C7B398C7331
8C7B398C7331947B398C7331947B398C7331947B398C84319484398C8431948C398C8431
948C398C8431948C39948431948C399484319C8C399484319C8C399484319C8C39948431
9C9439948C316BA5396B9C316BA5396B9C3173A5396BA53173A5396BA53173AD396BA531
73AD396BA53173AD396BA53173AD396BA53173AD3973A53173AD3973AD3973AD3973AD39
7BB53973AD397BB54273AD397BB54273AD397BB54273AD397BB54273AD397BB54273B539
8CB54284B5398CBD428CB5398CBD428CB54294BD4A8CB54294BD4A8CB54294BD4A8CBD42
94BD4A8CBD4294C64A39BD5A39C66339BD5A39C66339BD635AC65252C64A5ACE5252C64A
5ACE5252C64A5ACE5252C64A5ACE5252CE4A5AD6525ACE4A5AD64A5ACE4A5AD64A5ACE4A
5AD64A5AD64A5ADE4A5AD64A9CDE4A9CD6429CDE4A9CD6429CDE4A9CD6429CDE4A9CD642
9CDE4A9CD6429CDE4A9CD64242DE4242D63942DE4242D63942DE4242D63942DE4242D639
42D6429CD6399CD6399CD6399CD6399CCE399CD6399CCE399CD63994CE319CD6399CCE31
9CD63994CE319CD63994CE319CD63994CE319CD63994CE319CD63994CE315AD639B5CE31
BDCE39B5CE31BDCE39B5CE31BDCE39B5C631BDCE39B5C631BDCE39B5C631B5CE39B5C631
B5CE39ADC631B5CE39ADC631B5CE39ADC631B5CE39ADC631C69C39BD9431C69C39BD9431
C6AD39BDA531C6AD39BDA531C6AD39BDA531C6AD39BDA531BDAD39BDA531BDAD39B5A531
BDAD39B5A531BDAD39B5A531BDA539000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000845231845231846339
846331846331846331846339846331846331846331846339846331846339846331846339
8463318463398463318C63398473318C73398C73318C73398C73318C73398C73318C7339
8C73318C73398C73318C73398C73318C73398C73318C7B398C84319484398C8431948439
948431948439948431948439948431948C39948431948C399484319C8C39948C319C8C39
739C316BA5396BA5396BA5396BA5316BA5396BA5396BA5396BA5316BA5396BA53973A539
6BA53173A5396BA53973AD396BA53973AD3973AD3973AD3973AD3973AD3973AD3973AD39
73AD3973AD3973AD3973B53973AD397BB53973B5397BB54273B5397BB5427BB5397BB542
7BB5398CB5428CB5428CB5428CB5428CBD428CB5428CBD428CBD428CBD428CBD4294BD4A
8CBD4A39BD5A39BD5A39C66339BD6339C66339C66339C6635AC64A5AC6525AC64A5ACE52
5AC64A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5ACE4A5AD64A5ACE4A5AD64A5AD64A5AD64A
5AD64A5AD64A5AD64A5AD64A5AD64A9CDE4A9CD64A9CDE4A9CD6429CDE4A9CD6429CDE4A
9CD6429CDE429CD64242DE4242D64242DE4242D64242DE4242D63942D64242D63942D639
42D63942D6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD639
9CCE399CD6399CCE399CD6399CCE399CD6399CCE399CCE395ACE315ACE395ACE31BDCE39
B5CE31BDCE39B5CE31BDCE39B5CE31B5CE39B5CE31B5CE39B5CE31B5CE39B5CE31B5CE39
B5C631B5CE39B5C631B5CE39B5C631B5CE39ADC631B5CE39ADC631C69439C6AD31C6AD39
BDA531C6AD39BDA531C6AD39BDA531BDAD39BDA531BDAD39BDA531BDA539BDA531BDA539
BDA531BDA539B5A531BDA539B5A531000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007B5231845A398452318452397B5231
846339846331846339845A318C6B398463318463398463318C6339846331846339846331
8C6B398463318C63398463318C73398473318C73398473318C7B398473318C7339847331
8C7B398C73318C7B398C7331947B398C7331947B398C7331947B398C84319484398C8431
948C398C8431948C398C8431948C39948431948C399484319C8C39739C3173A539739C31
7BA539639C316BA539639C316BA5396B9C316BA5396B9C3173A5396BA53173A5396BA531
73AD396BA53173AD396BA53173AD396BA53173AD396BA53173AD3973AD3973AD3973AD39
7BB53973AD3973B53973AD397BB54273AD397BB54273AD397BB54273B5397BB54273AD39
7BBD428CB5398CBD428CB5398CBD428CB5428CBD428CB54294BD4A8CBD4294BD4A31BD5A
39C66339BD5A39C66331BD5A39C66339BD6339C66331BD6339C66B52C64A5ACE5252C64A
5ACE525AC64A5ACE5252C64A5AD6525ACE4A5AD64A52CE4A5AD6525ACE4A5AD64A5ACE4A
63D64A5AD64A5AD64A5AD64263DE4A9CD64A9CDE4A9CD642A5DE4A9CD6429CDE4A9CD642
A5DE4A42D64242DE4A42D64242DE4242D64242DE4239D63942DE4242D63942DE4239D639
42DE4242D6399CD6399CD631A5D6399CD6399CD6399CCE319CD6399CCE399CD63994CE31
9CD6399CCE319CD63994CE319CD63994CE319CD63952CE315AD6395ACE315ACE3952CE31
BDCE39B5CE31BDCE39B5C631BDCE39B5CE31BDCE39B5C631BDCE39B5C631B5CE39ADC631
B5CE39B5C631B5CE39ADC631B5CE39ADC631B5CE39ADC63194CE39BDA531C6AD39BDA531
C6AD39BDA531C6AD39BDA531C6AD39BDA531BDAD39BDA531BDAD39BDA531BDA539B5A531
BDAD39B5A531BDA539B5A531BDA539000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008452318452397B5231845231845231845239845231845231
845231846339846331846331846331846339846331846331846331846339846331846331
8463318463398463318463398463318C73398473318473318C73318C73398473318C7339
8C73318C73398C73318C73398C73318C7B398C73318C73398C73318C84398C84318C8439
8C84399484398C8431948439948431948C39739C31739C39739C39739C39739C31739C39
739C3973A5396B9C316BA5396BA5396BA5396BA5316BA5396BA5396BA5396BA5316BA539
6BA53973A5396BA53173A53973A53973AD396BA53173AD3973AD3973AD3973AD3973AD39
73AD3973AD3973AD3973AD3973AD397BB54273AD397BB5397BB5397BB54273B5397BB542
7BB5428CB5428CB5428CB5428CB5428CBD428CB5428CBD428CBD4294BD4A39BD5A39BD5A
39BD5A39BD6339BD5A39BD6339C66339C66339C66339C66339C6635AC6525AC64A5AC652
5ACE525ACE525ACE4A5ACE525ACE4A5ACE525ACE4A5ACE4A5AD64A5AD64A5AD64A5AD64A
5AD64A5AD64A5AD64A5AD64A5ADE4A5ADE4A9CD64A9CDE4A9CDE4A9CDE4A9CD6429CDE4A
42DE4242DE4A42D64242DE4242D64242DE4242D64242D64242D64242DE4242D63942D639
42D63942D63942D6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD639
9CD6399CD6399CCE399CD6399CCE399CD6395ACE315ACE395ACE395ACE395ACE315ACE39
BDCE39BDCE39B5CE31B5CE39B5CE39BDCE39B5CE31B5CE39B5CE39B5CE39B5C631B5CE39
B5CE39B5CE39B5C631B5CE39B5C631B5CE39ADC63194C63994C631C6AD39BDA531C6AD39
BDA531C6AD39BDA531BDA531BDA531BDAD39BDA531BDA539BDA531BDA539B5A531BDA531
BDA531BDA539B5A531B5A531B5A531B5A539000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008452397B52318452397B52318452397B52318452397B5231845A397B5231
8452397B52318463397B63318463397B6331846339846331846339846331846339846331
8463398463318C63398463318C63398463318C73398473318C73398473318C7339847331
8C73398473318C7B398473318C7B398C73318C7B398C73318C7B398C73319484398C8431
9484398C8431948C396B9431739C396B9431739C396B9C31739C39739C3173A539739C31
73A539739C316BA539639C316BA5396B9C316BA5396B9C316BA5396B9C3173A5396BA531
73AD396BA53173AD396BA53173AD396BA53173AD396BA53173AD396BA53173AD3973AD39
73AD3973AD397BB53973AD397BB54273AD397BB54273AD397BB54273AD397BB54273B539
7BB54273B5398CBD428CB5398CBD428CB54294BD428CB54239BD5A31B55A39BD6339BD5A
39BD6331BD5A39C66339BD5A39C66339BD6339C66B39C66339C66B52C64A5ACE5252C64A
5ACE5252C64A5ACE525ACE4A5ACE5252CE4A5AD6525ACE4A5AD6525ACE4A5AD6525AD64A
5AD64A5AD64A5ADE4A5AD64A63DE4A5AD64AA5DE4A9CD64A9CDE4A9CD642A5DE4A42D642
42DE4A42D64242DE4A42D64242DE4242D64242DE4242D63942DE4242D63942DE4242D639
42DE4239D63942DE399CD6399CD6399CD6399CD6399CD6399CD63994CE319CD6399CCE31
9CD63994CE319CD63994CE319CD6395ACE315AD6395ACE315AD6395ACE315AD6395ACE31
5ACE39B5C631BDCE39B5C631BDCE39B5C631BDCE39B5C631B5CE39B5C631B5CE39B5C631
B5CE39ADC631B5CE39ADC631B5CE39ADC63194CE398CC63194CE398CC631C6AD39BDA531
C6AD39BDA531C6AD39BDA531BDAD39BDA531BDAD39BDA531BDAD39B5A531BDAD39B5A531
BDA539B5A531BDA539B5A531BDA539B5AD31000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000845231
7B52318452317B52318452397B52318452317B52318452317B5231845231845231845239
7B5231845231845231846339846331846339846331846339846331846331846331846339
8463318463398463318463398463318463398463318473398473318473398473318C7339
8473318C73398C73318C73398C73318C73398C73318C73398C73318C73398C73318C8439
6B9431739C396B9C31739C396B9C31739C39739C31739C39739C31739C39739C31739C39
739C31739C39739C316B9C396B9C316BA5396BA5396BA5396BA5316BA5396BA5396BA539
6BA5396BA5396BA53973A5396BA53973A53973A53973AD3973AD3973AD3973AD3973AD39
73AD3973AD3973AD3973AD3973AD3973B53973AD397BB54273AD397BB5427BB5397BB542
7BB5397BB5428CB5428CB5428CB5428CB54239B55A39BD5A39BD5A39BD5A39BD5A39BD5A
39BD5A39BD6339BD5A39C66339BD6339C66339C66339C66339C66339C66B5AC6525ACE52
5AC6525ACE525ACE525ACE525ACE4A5ACE525ACE4A5AD6525ACE4A5AD6525AD64A5AD64A
5AD64A5AD64A5AD64A5ADE4A5ADE4A5ADE4A9CDE4A9CDE4A9CDE4A42DE4A42DE4A42DE4A
42DE4242DE4A42DE4242DE4242D64242DE4242D64242DE4242D64242D64242D63942D642
42D63942D63942D63942D6399CD6399CD6399CD6399CD6399CD6399CD6399CD6399CD639
9CD6399CD6399CD6395AD6395ACE315AD6395ACE395ACE395ACE315ACE395ACE395ACE39
5ACE31BDCE39B5CE31B5CE39B5CE31B5CE39B5CE31B5CE39B5C631B5CE39B5C631B5CE39
B5C631B5CE39B5C631B5CE39ADC63194C63994C63194C6398CC63194C639BDA531C6AD39
BDA531BDA539BDA531BDA539BDA531BDA539BDA531BDA539BDA531BDA539B5A531BDA539
B5A531B5A539B5A531B5A539B5AD31B5AD31B5AD31000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007B4231844A397B5231
8452397B52318452397B52318452397B52318452397B52318452397B52318452397B5231
845A397B52318452397B5A318463397B63318463397B5A318463398463318463397B5A31
846339846331846339845A318C6339846331846339846B318C73398473318C7339846B31
8C73398473318C73398473318C7B398473318C73398473318C7B397B9431849C396B9431
739C396B9431739C396B9431739C396B9431739C396B9431739C39739C31739C39739C31
7BA539739C3173A539639C316BA5396B9C316BA5396B9C3173A5396BA5316BA5396B9C31
73AD396BA53173A5396BA53173AD396BA53173AD396BA53173AD3973A53173AD396BA531
73AD3973AD3973AD3973AD397BB54273AD397BB54273AD397BB54273AD397BB54273AD39
7BB54273B5397BB54284B53939BD5A39B55239BD5A31B55239BD5A39BD5A39BD5A31BD5A
39C66339BD5A39C66331BD5A39C66339BD6339C66331BD6339CE6B39C6635ACE5252C652
5ACE525AC6525ACE5252C64A5AD6525ACE4A5AD65252CE4A5AD6525ACE4A5AD6525AD64A
63DE525AD64A5ADE525AD64A63DE525ADE4A5ADE4A9CDE4A4ADE4A42DE4A42DE4A42D642
4ADE4A42D64242DE4A42D64242DE4A42D64242DE4242D63942DE4242D63942DE4239D639
42DE4242D63942D63939D639A5DE399CD6399CD63994CE319CD6399CD6399CD63994CE31
9CD6399CCE315AD63952CE315AD6395ACE315AD63952CE315AD6395ACE315ACE3952CE31
5ACE395ACE31BDCE39B5C631BDCE39B5C631B5CE39B5C631BDCE39B5C631B5CE39ADC631
B5CE39ADC631B5CE398CC63194CE398CC63194CE398CC63194CE398CC63194C639BDA531
C6AD39BDA531BDAD39BDA531BDAD39BDA531BDA539B5A531BDAD39B5A531BDA539B5A531
BDA539B5A531B5AD39B5AD31BDB539B5AD31B5AD39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007B42317B4231844A397B42317B4A31
7B52318452397B52317B52317B52318452397B52317B52317B52318452397B52317B5231
7B52318452397B52318452317B63318463397B6331846331846331846339846331846331
846331846339846331846331846331846339846331846331847331847339847331847331
8473318C73398473318C73398C73318C73397B94317B94397B94318494397B9431849439
6B9431739C396B94316B9C39739C39739C396B9431739C39739C39739C39739C31739C39
739C39739C39739C31739C396B9C396BA5396B9C316BA5396BA5396BA5396BA5316BA539
6BA53973A5396BA5316BA5396BA53973AD396BA53173AD3973A53973AD3973A53973AD39
73AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD397BB54273AD397BB542
7BB5427BB54239B57339B55239B55239BD5A39B55239BD5A39BD5A39BD5A39BD5A39BD5A
39BD5A39C66339BD5A39C66339C66339C66339C66339C66339C66339C66B39C6635ACE52
5ACE525ACE525ACE525ACE525ACE525ACE525ACE525AD6525AD6525AD6525AD64A5AD652
5AD64A5ADE525AD64A5ADE4A5ADE4A5ADE4A39DE9C42DE4A42DE4A42DE4A42DE4A42DE4A
42DE4A42DE4A42D64242DE4242DE4242DE4242D64242DE4242D64242DE4242D63942D639
42D63942D63942D63942D63942D6399CD6399CD6399CD6399CD6399CD6399CD6399CD639
9CD6395AD6395ACE315AD6395ACE395AD6395ACE315ACE395ACE395ACE395ACE315ACE39
5ACE395ACE39B5CE31B5CE39B5CE39B5CE39B5C631B5CE39B5CE39B5CE39B5C631B5CE39
B5C631B5CE3994C63194C63994C63194CE398CC63194C6398CC63194C6398CC631BDA539
BDA531BDAD39BDA531BDA539BDA531BDA539B5A531BDA531BDA531BDA539B5A531B5A531
B5A531B5AD39B5AD31B5AD31B5AD31B5AD39B5AD31B5AD31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000844A397B4231844A397B4231844A397B4231844A397B4231
844A397B52318452397B52318452397B52318452397B52318452397B52318452397B5231
8452397B52318452397B52318452397B5A318463397B5A318463397B5A318463397B5A31
8463397B5A318463397B63318463397B6331846339846331846339846B31847339846B31
8C7339846B318C73397B94317B94397B94318494397B94318494397B9431849C397B9431
739C396B9431739C396B9431739C396B9431739C396B9431739C396B9431739C39739C31
739C39739C317BA539739C316BA539639C316BA5396B9C316BA5396B9C3173A5396B9C31
73A5396BA53173A5396BA53173AD396BA53173AD396BA53173AD396BA53173AD3973A531
73AD3973A53973AD3973AD397BAD3973AD397BB54273AD397BB54273AD397BB54273AD39
39B57331AD6B39B57331B57339BD5A31B55239BD5A39B55A39BD5A31B55A39BD6339BD5A
39C66339BD5A39C66339BD6339C66339BD6339C66B39C66339C66B39C66339CE6B52C652
5ACE5A52C6525ACE525ACE525ACE525ACE525AD6525ACE525AD6525ACE4A5AD6525AD64A
5ADE525AD64A63DE525AD64A63DE5239D69C42DEA539DE9C42DE4A42DE4242DE4A42DE42
42DE4A42D64242DE4A42D64242DE4A42D64242DE4242D64242DE4242D63942DE4242D639
42DE4242D63942DE4242D63942D6399CD6319CD6399CD6319CD6399CCE319CD6395ACE31
5AD6395ACE315AD6395ACE315AD6395ACE315AD6395ACE315AD63952CE315ACE3952CE31
5ACE3952C6315ACE39B5C631BDCE39B5C631B5CE39B5C631B5CE39B5C631B5CE39ADC631
B5CE398CC63194CE398CC63194CE398CC63194CE398CC63194C6398CC63194C6398CC631
BDAD39BDA531BDAD39BDA531BDAD39B5A531BDAD39B5A531BDA539B5A531BDA539B5A531
BDB539B5AD31B5B539B5AD31B5AD39ADAD31B5AD39ADAD31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B42317B4A317B42317B4A317B42317B4A317B42317B4A317B42317B4A31
7B42317B4A317B52317B52317B52317B52317B52317B52397B52317B52317B52317B5239
7B52317B52317B52317B52397B52317B52317B63318463397B63318463317B6331846339
7B6331846331846331846339846331846339846331846339846331846339847331847339
7B94317B94397B94317B94397B94317B94397B94317B94397B94317B94397B9431849439
7B94316B94396B94316B9C396B9431739C396B9431739C39739C31739C39739C31739C39
739C31739C39739C31739C39739C316B9C396B9C396BA5396B9C316BA5396BA5396BA539
6BA5316BA5396BA53973A5396BA53973A5396BA53973AD396BA53973AD3973A53973AD39
73AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD3973AD397BB53939AD7339B573
39B57339B57339B57339B57339B55239B55A39B55A39BD5A39BD5A39BD5A39BD5A39BD63
39BD5A39BD6339BD6339C66339C66339C66339C66339C66B39C66339C66B39C66B39CE6B
5AC6525ACE525ACE525ACE525ACE525ACE525ACE525AD6525AD6525AD6525AD6525AD652
5AD6525AD6525AD65239DEA539DEA539DEA539DE9C39DE9C42DE4A42DE4A42DE4A42DE4A
42DE4242DE4A42DE4242DE4242DE4242DE4242DE4242DE4242D64242DE4242D63942DE42
42D63942D63942D63942D63942D63942D6399CD6399CD6399CD6399CD6395AD6395AD639
5AD6395AD6395AD6395AD6395ACE395AD6395ACE395ACE395ACE315ACE395ACE395ACE39
5ACE315ACE395ACE39B5CE39B5CE31B5CE39B5CE31B5CE39B5C631B5CE39B5C63194CE39
94C63194C63994C63194C6398CC63194C6398CC63194C6398CC6318CC6398CC6318CC639
BDA531BDA539BDA531BDA539BDA531BDA539B5A531BDA539B5A531B5A539B5AD31B5AD39
B5AD31B5AD31B5AD31B5AD39B5AD31B5AD31B5AD31B5AD39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B4231
7B4A397B42317B4A397342317B4A397B42317B4A397B42317B4A397B42317B4A397B4231
844A397B42317B4A397B4A318452397B52317B52397B4A318452397B52317B52397B4A31
8452397B52317B52397B4A318452397B52318452397B5A318463397B5A318463397B5A31
8463397B63318463397B5A318463397B63318463397B5A318463398484317B9439738C31
7B9439738C317B9439738C317B94397B94317B94397B8C318494397B94317B94397B9431
849C397B94316B94396B9431739C396B9431739C396B9431739C396B9431739C396B9431
739C39739C31739C39739C317BA539739C316BA539639C316BA5396B9C316BA5396B9C31
73A5396BA53173A5396BA53173AD396BA53173AD396BA53173AD396BA53173AD396BA531
73AD3973A53973AD3973A5397BB54273AD397BAD3973AD3939B57331AD6B39B57331AD6B
39B57339B57339B57331B57339BD7B39B55239BD5A31B55239BD6339BD5A39BD6331BD5A
39C66339BD5A39C66331BD5A39C66B39C66339C66B39C66339CE6B39C66339CE6B39C663
5ACE5A5ACE525ACE5A52CE525AD65A5ACE525AD65252CE5263D65A5AD6525AD6525AD64A
63DE525AD64A39DEA539D69C42DEA539DE9C42DEA539DE9C42E7A542DE4242DE4A42DE42
4AE74A42DE4242DE4A42D64242DE4A42D64242DE4242D63942DE4242D63942DE4239D639
42DE4242D63942DE3939D63942DE3942D6399CD6399CD6319CD6395AD6315AD6395ACE31
5AD6395ACE315AD63952CE315AD6395ACE315AD63952CE315AD6395ACE315ACE3952CE31
5ACE3952CE315ACE3952C631BDCE39B5C631B5CE39B5C631B5CE39B5C63194CE398CC631
94CE398CC63194CE398CC63194CE398CC63194C6398CC63194CE398CC63194C6398CBD31
94C639BDA531BDA539B5A531BDAD39B5A531BDA539B5A531BDA539B5AD31B5AD39B5AD31
BDB539B5AD31B5AD39ADAD31B5AD39ADAD31B5AD39ADA531B5AD39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007B42317B39317B4231
7B42317B4A317B42317B42317B42317B4A397B42317B42317B42317B4A317B42317B4231
7B42317B4A397B42317B42317B52317B52397B52317B52317B52317B52397B52317B5231
7B52317B52397B52317B52317B52317B52397B52317B52317B63317B63397B5A317B6331
7B63318463397B63318463317B63318484397B84318484318484318484398484317B9439
7B8C317B9439738C317B94317B94317B94397B94317B94397B94317B94397B94317B9439
7B94398494396B94316B94396B94317394396B94316B9439739C39739C396B9431739C39
739C39739C39739C31739C39739C39739C39639C316B9C396B9C396BA5396B9C316BA539
6BA5396BA5396BA5316BA5396BA53973A5396BA53173A53973A53973AD3973A53973AD39
73AD3973AD3973AD3973AD3973AD3973AD3939AD6B39AD6B39AD6B39B57339AD6B39B573
39B57339B57339B57339B57339B57339B55A39B55239BD5A39BD5A39BD5A39BD5A39BD5A
39BD6339C66339BD6339C66339C66339C66339C66339C66B39C66B39CE6B39C66B39CE6B
39CE6B5ACE5A5ACE525ACE525ACE525AD6525ACE525AD6525AD6525AD6525AD6525AD652
5AD65239DEA539D6A539DEA539DEA539DEA539DE9C39DE9C39DE9C42DE4A42DE4242DE4A
42DE4242DE4A42DE4242DE4242DE4242DE4242DE4242DE4242DE4242DE4242D63942DE42
42D63942DE3942D63942D63942D63942D63942D6399CD6395AD6395AD6395AD6315AD639
5AD6395AD6395ACE395AD6395ACE395AD6395ACE315ACE395ACE395ACE395ACE315ACE39
5ACE395ACE395ACE315ACE39B5CE39B5CE39B5C631B5CE39B5C63194CE3994C63194C639
94C63994CE398CC63194C6398CC63194C6398CC6318CC6398CC6318CC6398CC6318CC639
8CC631BDA539B5A531BDA531BDA531BDA539B5A531B5AD31B5AD31B5AD39B5AD31B5AD31
B5AD31B5AD39B5AD31B5AD31B5AD31B5AD39ADAD31ADAD31ADAD31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007339317B42397339317B4239733931
7B4A397342317B4A397342317B4A397342317B4A397B42317B4A397342317B4A397B4231
7B4A397B42317B4A397B42317B4A39734A317B52397B4A317B52397B4A317B52397B5231
7B52397B4A317B52397B52317B52397B52318452397B52317B63397B5A318463397B5A31
8463397B5A318484397B7B318484397B7B318484397B7B318484397B7B318484397B8431
7B9439738C317B9439738C317B9439738C317B94397B8C317B94397B8C317B94397B8C31
8494397B94318494396B94316B94396B9431739C396B9431739C396B9431739C396B9431
739C396B9431739C39739431739C39739C317B9C39639C316BA539639C316BA5396B9C31
6BA5396B9C3173A5396BA53173A5396BA53173A5396BA53173AD396BA53173AD396BA531
73AD396BA53173AD3973A53939AD6B31AD6B39B57331AD6B39B57331AD6B39B57331AD6B
39B57331AD6B39B57331B57339B57331B57339BD5A31B55239BD5A31B55A39BD6339BD5A
39C66331BD5A39C66339BD6339C66B39BD6339C66B39C66339CE6B39C66339CE6B39C66B
39CE7352C6525ACE5A5ACE525AD65A5ACE525AD65A5ACE525AD65A5ACE525AD65A39D6A5
39DEAD39D6A542DEA539D69C42DEA539D69C42DEA539DE9C42E79C39DE944AE74A42DE42
42E74A42DE424AE74A42DE4242DE4242D63942DE4242D63942DE4242D63942DE4242D639
42DE4242D63942DE4242D63942DE3942D63139DE8C5AD6315AD6395ACE315AD6395ACE31
5AD6395ACE315AD6395ACE315AD63952CE315AD63952CE315AD63952CE315ACE3952CE31
5ACE3952C6315ACE3952C6315ACE39B5C631B5CE39B5C63194CE398CC63194CE398CC631
94CE398CC63194CE398CC63194C6398CC63194C6398CC63194C6398CC63194C6398CBD31
8CC6398CBD31BDAD39B5A531BDA539B5A531BDB539B5AD31BDB539B5AD31B5AD39B5AD31
B5AD39ADAD31B5AD39ADAD31B5AD39ADA531B5AD39ADA531B5AD39ADA531000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B39317B39317B39317B39317B39317B39317B3931
7B39317B42317B42317B42317B42317B42317B42317B4A317B42317B42317B42317B4A31
7B42317B42317B42317B4A317B42317B42317B52317B52317B52317B52317B52317B5239
7B52317B52317B52317B52317B52317B52317B52317B52397B52317B63317B5A317B8439
7B84317B84397B84317B84397B84317B84317B84318484397B84318484397B8431848439
738C31738C39738C317B94397B8C317B94397B94317B94397B8C317B94397B94317B9439
7B94317B94397B94316B94396B94316B94396B94316B94396B9431739C396B9431739C39
739431739C39739C31739C39739C31739C39739C396B9C396B9C316B9C396B9C396BA539
6B9C396BA5396BA5396BA5396BA5396BA5396BA53973A5396BA53973A53973A53973AD39
73A53939AD6B39AD6B39AD6B39AD6B39AD6B39AD6B39AD7339AD6B39AD7339AD6B39B573
39B57339B57339B57339B57339B57339B57339B55A39BD5A39B55A39BD5A39BD5A39BD63
39BD5A39BD6339BD6339C66339C66339C66339C66339C66B39C66B39C66B39C66B39CE6B
39CE6B5ACE5A5ACE5A5ACE5A5ACE525ACE5A5ACE525AD65A5AD6525AD65239D6AD39D6AD
39D6A539D6A539D6A539DEA539DEA539DEA539DE9C39DE9C39DE9C39DE9C42DE4242DE4A
42DE4242DE4242DE4242DE4242DE4242DE4242DE4242DE4242DE3942DE4242DE3942DE42
42D63942DE3942D63942DE3942D63939D68C39D6845AD6395AD6395AD6395AD6395AD639
5AD6395AD6395ACE395AD6395ACE395ACE395ACE395ACE395ACE395ACE395ACE395ACE39
5ACE315ACE395ACE315ACE395ACE31B5CE3994C63194CE3994C63194C63994C63194C639
94C63194C6398CC63194C6398CC6318CC6398CC6318CC6398CC6318CC6398CC6318CC639
8CC6318CC639B5A531BDA539B5AD31B5AD39B5AD31B5AD39B5AD31B5AD31B5AD31B5AD39
B5AD31B5AD39B5AD31B5AD39ADAD31ADAD31ADAD31ADAD39ADA531ADAD31000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B39317339317B42397339317B42397339317B42397339317B4239733931
7B42397339317B4A397342317B4A397342317B4A397342317B4A397342317B4A39734231
7B4A397342317B4A397342317B4A397342317B4A39734A317B52397B52317B5239734A31
7B52397B52317B5239734A317B52397B52317B5239734A317B6B397B63317B84397B7B31
8484397B7B317B84397B7B318484397B7B318484397B7B318484397B7B318484397B7B31
848439738C317B9439738C317B9439738C317B9439738C317B94397B8C317B9439738C31
7B94397B8C317B94397B8C316B94396B94316B94396B9431739C396B9431739C396B9431
739C396B9431739C396B9431739C39739C31739C397394316BA539639C316BA539639C31
6BA5396B9C316BA5396B9C3173A5396BA53173A5396BA53173AD396BA53173AD3931A563
39AD6B31A56B39AD6B31A56B39AD7331AD6B39AD7331AD6B39B57339AD6B39B57331AD6B
39B57339AD7339B57331AD7339BD7B39B57339BD5A31B55239BD5A39B55A39BD6331B55A
39C66339BD5A39C66331BD6339C66B39C66339C66B39C66339CE6B39C66B39CE6B39C66B
42CE7339CE6B5ACE5A52CE525AD65A5ACE525AD65A5ACE5242D6AD39D6A539D6AD39D6A5
42DEAD39D6A539DEA539D69C42DEA539DE9C39DEA539DE9C42E7A539DE9C42E74A42DE42
4AE74A42DE4242E74242DE424AE74242DE4242DE4242D6394ADE4242DE3942DE4242D639
42DE4242D63942DE3942D63142DE8C39D68439D68C39CE8463D6395AD6315AD6395ACE31
5AD6395ACE315AD63952CE315AD6395ACE315AD63952CE315AD6395ACE315ACE3952C631
5ACE3952CE315ACE3952C6315ACE3939C64294CE398CC63194CE3994C63194CE398CC631
94CE398CC63194C6398CC63194CE398CC63194C6398CBD3194C6398CC6318CC6398CBD31
8CC6398CBD318CC639B5AD31BDB539B5AD31B5AD39B5AD31B5B539B5AD31B5AD39ADA531
B5AD39ADAD31B5AD39ADA531B5AD39ADA531B5AD39ADA531B5AD39ADA531ADAD39000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7339317B39317339317B39317339317B39317339317B39317B39317B39317339317B3931
7B39317B39317339317B42317B42317B4A317342317B42317B42317B4A317342317B4231
7B42317B4A317342317B42317B42317B4A397342317B42317B52317B52397B52317B5231
7B52317B52397B52317B52317B52317B6B397B63317B6B317B6B317B6B397B63317B8431
7B7B317B84397B7B317B84317B84317B84397B7B317B84317B84317B84397B7B317B8431
7B8431848439738C31738C39738C317B9439738C317B8C397B8C317B94397B8C317B9439
7B94317B94397B8C317B94396B94396B94396B94316B94396B94397394396B9431739439
739439739C39739431739C39739C39739C39739C31739C396B9C396B9C396B9C316B9C39
6B9C396BA5396B9C316BA5396BA5396BA5396BA5316BA53939A56B39A56B31A56339A56B
39A56B39AD6B39A56B39AD6B39AD6B39AD6B39AD6B39AD6B39AD6B39B57339AD6B39B573
39B57339B57339B57339B57339B57339B57339B57339B55A39B55A39BD5A39BD5A39BD5A
39BD5A39BD6339BD6339C66339C66339C66339C66339C66B39C66B39CE6B39C66B39CE6B
39CE6B39CE735ACE5A5ACE5A5ACE5A5AD65A5ACE5A39D6AD39D6AD39D6AD39D6A539D6A5
39D6A539DEA539D6A539DEA539DEA539DEA539DE9C39DE9C39DE9C39E79C39DE9442E742
42DE4242E74242DE4242DE4242DE4242DE4242DE4242DE4242DE3942DE4242DE3942DE39
42DE3942DE3942D63939DE8C39D68C39D68C39D68439D6845AD6395AD6395AD6395AD639
5AD6395AD6395ACE315ACE395ACE395AD6395ACE315ACE395ACE395ACE395ACE315ACE39
5ACE395ACE395AC6315ACE3939CE4A39CE4A94C63194C63994C63994CE3994C63194C639
94C63194C6398CC6318CC6398CC63194C6398CC6318CC6398CC6318CC6398CBD318CC631
8CC6317BC6397BBD31B5AD31B5AD31B5AD39B5AD31B5AD31B5AD31B5AD39B5AD31B5AD31
B5AD31B5AD39ADAD31ADAD31ADAD31ADAD39ADA531ADAD31ADA531ADAD39ADA531000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7B42397339317B42397339317B42397339317B42397339317B42397339317B4239733931
7B42397339317B42397339317B42397342317B4A397342317B4A397342317B4A39734231
7B4A397342317B4A397342317B4A397342317B4A397342317B4A39734A317B5239734A31
7B52397363317B6B397363317B6B397363317B6B397363317B6B397363317B6B39736331
7B8439737B317B84397B7B317B84397B7B317B84397B7B317B84397B7B318484397B7B31
8484397B7B317B8C39738C317B8C39738C317B9439738C317B9439738C317B9439738C31
7B9439738C317B94397B8C317B94396B8C316B94396B94317394396B9431739C396B9431
739C396B9431739C396B9431739C396B9431739C39739431739C39639C316B9C39639C31
6BA539639C316BA5396B9C316BA539319C6339A56B31A56339A56B31A56339AD6B31A563
39AD6B31A56339AD6B31A56B39AD6B31AD6B39AD7331AD6B39B57331AD6B39B57331AD6B
39B57331AD6B39B57331B57339B57B31B57339BD7B39B55239BD5A31B55A39BD6339BD5A
39BD6339BD5A39C66339BD6339C66339BD6339C66B39C66339CE6B39C66339CE6B39C66B
39CE7339C66B39CE735ACE5A5AD65A39CEAD42D6B539CEAD39D6AD39CEA542D6AD39D6A5
42DEAD39D6A542DEA539D69C42DEA539D69C42DEA539DE9C42E79C39DE9442E79C42DE42
42E74A42DE424AE74242DE4242E74242DE394AE74242DE3942DE4242DE3942DE4242DE39
42DE4242D63942DE8C39D68439DE8C39D68439DE8C39D68439D68C5ACE315AD6395ACE31
5AD6395ACE315AD6395ACE315AD63952CE315AD63952CE315ACE3952CE315ACE3952C631
5ACE3952C6315ACE3939C64239CE4A31C64239CE4A8CC63194CE398CC63194CE398CC631
94C6398CC63194C6398CC63194C6398CBD3194C6398CBD318CC6398CBD318CC6398CBD31
7BC63973BD317BC63973BD31B5AD39B5AD31B5AD39B5AD31B5AD39ADAD31B5AD39ADAD31
B5AD39ADA531B5AD39ADA531B5AD39ADA531ADAD39ADA531ADAD39A5A531ADAD39A5A531
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007339317339317339317339317339317339317B39317339317339317339317B3931
7339317B39317339317B39317339317B39317342317B42317342317B42317342317B4231
7342317B42317342317B42317342317B42317342317B42317342317B52317363317B6B31
7363317B6B317B63317B6B317363317B6B317B63317B6B397363317B6B317B63317B6B39
7B7B317B7B317B7B317B84397B7B317B84397B7B317B84397B7B317B84397B7B317B8439
7B7B317B84397B7B31738C39738C31738C39738C317B8C39738C317B8C397B8C317B9439
7B8C317B94397B8C317B94397B8C317B94396B94316B94396B94316B94396B94316B9439
6B94317394396B9431739C39739431739C39739C31739C39739C316B9C396B9C396B9C39
6B9C316B9C396B9C3939A56B399C6339A56B39A56339A56B39A56339A56B39A56B39A56B
39A56B39AD6B39A56B39AD6B39AD6B39AD6B39AD6B39AD7339AD6B39AD7339AD7339B573
39AD7339B57339B57339B57339B57339B57339B57339B57B39B55A39BD5A39BD5A39BD5A
39BD5A39BD6339BD6339C66339BD6339C66339C66339C66B39C66B39C66B39C66B39CE6B
39CE6B39CE7339CE6B5ACE5A39CEAD39CEAD39CEAD39D6AD39D6AD39D6AD39D6A539D6AD
39D6A539D6A539D6A539DEA539DE9C39DEA539DE9C39DE9C39DE9C39DE9C39DE9C39E79C
42DE4242E74242DE4242E74242DE4242DE4242DE4242DE4242DE3942DE3942DE3942DE39
39DE8C39DE8C39DE8C39DE8C39D68439D68C39D68439D68C39D6845AD6395AD6395AD639
5ACE395AD6395ACE395ACE395ACE395ACE395ACE395ACE395ACE315ACE395ACE395ACE39
5ACE315ACE3939CE4A39CE4A39C64A39CE4A39C64A94C63994C63194C63994C63194C639
8CC63194C6398CC6318CC6398CC6318CC6398CC6318CC6398CC6318CC6398CC6317BC639
7BBD317BC6397BBD317BBD39B5AD31B5AD39B5AD31B5AD39B5AD31B5AD39ADAD31B5AD39
ADAD31ADAD31ADAD31ADAD39ADA531ADAD31ADA531ADAD39ADA531ADA531A5A531ADA531
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B39317339317B42397339317B39317339317B42397339317B4239733931
7B42397339317B39317339317B42397339317B42397342317B4A397342317B4A39734231
7B4A397342317B4A397342317B4A397342317B4A397352317B5A397352317B6B39736331
7B6B397363317B6B397363317B6B397363317B6B397363317B6B397363317B6B39736331
7B6B39737B317B8439737B317B8439737B317B8439737B317B84397B7B317B84397B7B31
8484397B7B317B84397B7B317B8C39738C317B8C39738C317B9439738C317B9439738C31
7B9439738C317B9439738C317B94397B8C316B9439638C316B94396B94316B94396B9431
739C396B9431739C396B9431739C396B9431739C396B9431739C397394316B9C39639431
39A56B319C6339A56B319C6339A56B319C6339A56B319C6339A56B31A56339A56B31A563
39AD6B31A56339AD6B31A56339AD7331AD6B39AD7331AD6B39B57331AD6B39B57331AD6B
39B57339AD7339B57331AD7339B57B39B57339B57B31B57339BD5A39B55A39BD5A31B55A
39BD6339BD5A39C66331BD5A39C66B39BD6339C66B39C66339CE6B39C66B39CE6B39C66B
42CE7339CE6B39C6F739C6EF42D6B539CEAD39D6AD39CEA542D6AD39D6A539D6AD39D6A5
42DEAD39D6A539DEA539D69C42DEA539DE9C42DEA539DE9C42E79C39DE9442E79C39DE94
4AE74A42DE4242E74242DE394AE74242DE3942E74242DE394AE74242DE3942DE4239DE8C
42DE8C39D68C39DE8C39D68442DE8C39D68439D68C39D68439D68C39CE845AD63952CE31
5AD6395ACE315AD63952CE315AD6395ACE315ACE3952CE315ACE395ACE315ACE3952C631
5ACE3939C64239CE4A31C64239CE4A39C64239CE4A8CC63194CE398CC63194C6398CC631
94CE398CC63194C6398CBD3194C6398CBD318CC6398CBD318CC6398CBD317BC63973BD31
7BC63973BD317BC63973BD317BC639B5AD31B5AD39ADA531B5AD39ADA531B5AD39ADA531
B5AD39ADA531ADAD39ADA531ADAD39ADA531ADAD39A5A531ADAD39A5A531ADAD39A5A531
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007339317339317339317339317339317339317B3931733931733931
7339317B39317339317339317339317B39317339317339317342317B4231734231734231
7342317B4A397342317342317342317B52317352317352317352317B5239735231736331
7363317B6B397363317363317363317B6B397363317363317363317B6B39736331736331
7363317B6B39737B317B7B317B7B317B7B39737B317B7B317B7B317B84397B7B317B7B31
7B7B317B84397B7B317B7B31738C31738C39738C31738C39738C317B8C39738C31738C39
7B8C317B94397B8C317B8C397B8C317B94397B8C316B94396B94396B94396B94316B9439
6B94397394396B94316B9439739439739C39739431739C39399C4A399C4A319C63399C63
399C63399C6B319C63399C63399C6339A56B319C6339A56B39A56B39A56B39A56339A56B
39A56B39AD6B31A56339AD6B39AD6B39AD6B39AD6B39AD6B39AD6B39AD7339AD6B39AD73
39B57339B57339B57339B57339B57339B57339B57339B57339B57B39BD5A39B55A39BD5A
39BD5A39BD6339BD5A39BD6339BD6339C66339C66339C66B39C66B39C66B39C66B39CE6B
39CE6B39C6F739C6F739C6F739CEAD39D6B539CEAD39D6AD39D6AD39D6AD39D6A539D6A5
39D6A539DEA539D6A539DEA539DEA539DEA539DE9C39DE9C39DE9C39E79C39DE9439E794
39E79442E74242DE4242E74242E74242E74242DE3942DE3942DE3942E74239DE8C39DE8C
39DE8C39DE8C39D68C39DE8C39D68C39D68C39D68439D68439D68439D6845ACE315ACE39
5ACE395AD6395ACE315ACE395ACE395ACE395ACE315ACE395ACE395ACE395ACE3139CE4A
39CE4A39CE4A39C64239C64A39C64A39CE4A39C64239C64A94C63194C6398CC6318CC639
8CC63994C6398CC6318CC6398CC6318CC6398CBD318CC6318CC6317BC6397BBD317BBD39
7BBD317BC63973BD3173BD3173BD31B5AD39ADAD31B5AD31ADAD31B5AD39ADA531ADAD31
ADAD31ADAD39ADA531ADAD31ADA531ADAD39A5A531ADA531ADA531ADA539A5A531A5A531
A5A531000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007339317339317B42397339317B42397339317B4239733931
7B42397339317B42397339317B42397339317B42397339317B42397342317B4A39734231
7B4A397342317B52397352317B52397352317B52397352317B52397352317B5239735231
7B6B397363317B6B397363317B6B397363317B6B397363317B6B397363317B6B39736331
7B6B397363317B6B39737B317B7B39737B317B8439737B317B8439737B317B8439737B31
7B8439737B317B84397B7B317B84396B8431738C39738C31738C39738C317B8C39738C31
7B8C39738C317B9439738C317B9439738C317B9439638C316B94396B8C316B94396B8C31
6B94396B94317394396B9431739C39319442399C4A31944A399C4A319442399C4A319463
399C6B319C6339A56B319C6339A56B319C6339A56B319C6339A56B319C6339A56B31A563
39AD6B31A56339AD6B31A56339AD6B31A56B39AD7331AD6B39AD7331AD6B39B57331AD6B
39B57331AD7339B57331AD7339B57B31B57339BD7B31B57339BD7B31B55A39BD6339B55A
39BD6331BD5A39C66339BD5A39C66339BD6339C66B39C66339C66B39C66339CE6B39BDEF
39C6F739BDEF42C6F739C6EF39D6B539CEAD42D6B539CEAD42D6AD39D6A542D6AD39D6A5
42DEA539D6A542DEA539D69C42DEA539DE9C42DEA539DE9C42E79C39DE9442E79C39DE94
42E79C42DE424AE74242DE3942E74242DE394AE74242DE3942E74239DE8C42E79439DE8C
39DE8C39D68442DE8C39D68439DE8C39D68439DE8C39D68439D68C39CE8439D68C5ACE31
5AD63952CE315AD6395ACE315ACE3952CE315ACE3952CE315ACE3952C63139CE4A39C642
39CE4A31C64239CE4A39C64239CE4A39C64239CE4A39C64294C6398CC63194C6398CC631
94C6398CC63194C6398CBD318CC6398CBD318CC6397BBD317BC63973BD317BC63973BD31
7BC63973BD317BC63973BD317BBD3973BD31B5AD39ADA531B5AD39ADA531B5AD39ADA531
ADAD39ADA531ADAD39A5A531ADAD39A5A531ADAD39A5A531ADAD39A5A531ADAD39A5A531
A5A539000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000733931733931733931733931733931733931
733931733931733931733931733931733931733931733931733931733931734231735231
735231735231735231735231735231735231735231735231735231735231735231735231
735231736331736331736331736331736331736331736339736331736331736331736331
736331736331736331737B39737B31737B31737B317B7B39737B317B7B397B7B317B7B39
7B7B317B7B397B7B317B7B397B7B31738C39738C31738C39738C31738C39738C31738C39
738C317B8C39738C317B8C39738C317B8C397B8C317B94396B8C316B94396B94316B9439
6B943139944A31944A39944A31944A39944A39944A399C4A31944A399C4A399C4A399C63
399C63399C63399C63399C6B319C6339A56B399C6339A56B39A56339A56B39A56B39A56B
39A56339A56B39A56B39AD6B39A56B39AD6B39AD6B39AD6B39AD6B39AD7339AD6B39B573
39AD7339B57339B57339B57339B57339B57339B57339B57B39B57339BD5A39B55A39BD5A
39BD5A39BD6339BD6339BD6339BD6339C66339C66339C66B39C66339C66B39BDEF39C6EF
39C6EF39C6F739C6EF39C6F739C6F739D6AD39D6AD39D6AD39D6AD39D6AD39D6A539D6AD
39D6A539DEA539D6A539DEA539DE9C39DE9C39DE9C39DE9C39DE9C39DE9C39DE9439E79C
39E79439E79442E74242E74242E73942E74242DE3942E73939DE8C39DE8C39DE8C39DE8C
39DE8C39DE8C39D68C39DE8C39D68439D68C39D68439D68C39D68439D68439CE845AD639
5ACE395ACE395ACE395ACE395ACE315ACE395ACE395ACE395ACE3139CE4A39CE4A39CE4A
39C64A39CE4A39C64A39CE4A39C64A39C64A39C64A39C64A39C64294C6398CC6318CC639
8CC6318CC6398CC6318CC6398CC6318CC6397BBD317BC6397BBD317BC6397BBD317BBD39
73BD317BBD3973BD317BBD3973BD3173BD39ADAD31ADAD39ADA531ADAD31ADA531ADAD39
ADA531ADAD31ADA531ADAD39A5A531ADA531A5A531ADA531A5A531A5A531A5A531A5A531
A5A531A5A531000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007339317B4239733931733931733931
7B42397339317339317339317B42397339317339317339317B4A39734231734A39734A31
7B5239735231735239734A317B5239735231735239734A317B5239735231735239734A31
7B5239735231736B397363317B6B39736331736B397363317B6B39736331736B39736331
7B6B397363317363397363317B7B39737B317B7B39737B317B7B39737B317B7B39737B31
7B8439737B317B7B39737B317B84397B7B31738C396B8431738C39738431738C39738431
7B8C39738C317B8C39738C317B9439738C317B9439738C316B9439318C4239944A318C42
39944A31944239944A318C42399C4A31944A39944A319442399C4A31944A399C4A319442
399C6B319C63399C6B319C6339A56B319C6339A56B319C6339A56B31A56339A56B31A563
39AD6B31A56339AD6B31A56339AD6B31A56B39AD6B31A56B39AD7331AD6B39AD7331AD6B
39B57331AD7339B57331AD7339B57B39B57339B57B31B57339BD7B39B57339BD5A31B55A
39BD6339BD5A39BD6331BD5A39C66339BD6339C66B31BD6339BDEF39BDEF39C6EF39BDEF
42C6F739C6EF39C6F739C6EF42CEF739CEAD39D6B539CEAD42D6AD39D6A539D6AD39D6A5
42DEAD39D6A542DEA539D69C42DEA539DE9C42DE9C39DE9442E79C39DE9442E79C39DE94
42E79C39E79442E74242DE394AE74242DE3942E74239DE8C42E79439DE8C39DE8C39DE8C
42DE8C39D68C39DE8C39D68442DE8C39D68439D68C39CE8439D68C39CE8439D68431CE84
5AD6395ACE315ACE3952CE315AD6395ACE315ACE3952C63139CE4A39C64A39CE4A31C642
39CE4A39C64239CE4A31C64239CE4A39C64239C64A31C64239CE4A8CC63194C6398CBD31
94C6398CC6318CC6398CBD3194C6397BBD317BC63973BD317BC63973BD317BC63973BD31
7BC63973BD317BBD3973BD317BBD3973BD3173BD39ADA531B5AD39ADA531ADAD39A5A531
ADAD39A5A531ADAD39A5A531ADAD39A5A531ADA539A5A531ADAD39A5A531A5A5399C9C31
A5A539A5A531000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000733931733931733931733931
733931733931733931733931733931733931734231734231734231734231734231734231
735231735231735231735231735231735231735231735231735231735231735231735231
735231735239735231736331736331736339736331736331736331736339736331736331
736331736339736331736331736331737B39737B31737B31737B317B7B39737B31737B31
737B317B7B39737B317B7B317B7B317B7B397B7B31738C31738C31738C39738431738C39
738C31738C39738C31738C39428C314A8C39428C314A8C394A8C3139944A318C4239944A
39944A39944A31944A39944A39944A39944A31944A39944A39944A399C4A31944A399C4A
399C63399C63319C63399C63399C6339A56B319C6339A56B39A56339A56B31A56339A56B
39A56B39A56B39A56339A56B39A56B39AD6B39A56B39AD6B39AD6B39AD7339AD6B39AD73
39AD7339B57339AD7339B57339B57339B57B39B57339B57B39B57B39BD7B39B55A39BD5A
39BD5A39BD6339BD5A39BD6339BD6339C66339B5E739BDEF39BDEF39BDEF39BDEF39C6EF
39C6EF39C6EF39C6EF39C6F739CEF739CEF739D6AD39D6AD39D6AD39D6AD39D6A539D6A5
39D6A539DEA539D6A539DEA539DEA539DEA539DE9C39DE9C39DE9C39E79C39DE9439E794
39E79442E79442E73942E74242E74242E79439DE8C39E78C39DE8C39DE8C39DE8C39DE8C
39DE8C39DE8C39D68C39DE8C39D68C39D68C39D68439D68439D68439D68439CE8439CE84
5ACE395ACE395ACE315ACE395ACE395ACE395ACE3139CE4A39CE4A39CE4A39C64A39CE4A
39C64A39CE4A39C64239C64A39C64A39C64A39C64239C64A39C64A39C64A8CC6318CC639
8CC6318CC6398CBD318CC6397BC6317BC6397BBD317BBD397BBD317BC63973BD317BBD31
73BD317BBD3973BD3173BD3173BD3173BD3973BD31ADAD31ADA531ADAD39ADA531ADA531
ADA531ADAD39A5A531A5A531A5A531ADA539A5A531A5A531A5A531A5A539A5A531A5A531
A5A531A5A5399CA531000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006B3931734239733931
733939733931734239734231734239734231734A39734231734A39734231734A39734231
734A39734A31735239735231735239734A317352397352317352396B4A31735239735231
7352396B4A317352397352317363396B6331736B39736331736339736331736B39736331
736339736331736B39736331736339736331737B39737331737B397373317B7B39737B31
7B7B39737B317B7B39737B317B7B39737B317B7B396B8431738C396B8431738C39428431
4A8C394284314A8C394284314A8C39428C314A8C39428C314A8C39318C4239944A318C42
39944A318C4239944A318C4239944A31944239944A319442399C4A319442399C4A31944A
399C6B319463399C6B319C6339A56B319C6339A56B319C6339A56B319C6339A56B31A563
39A56B31A56339AD6B31A56339AD6B31A56339AD6B31A56B39AD7331AD6B39B57331AD6B
39B57331AD6B39B57331AD7339B57B31AD7339B57B31B57339BD7B31B57339BD7B31B55A
39BD6331B55A39BD6339BD5A39C66339B5E739BDEF39BDE739BDEF39BDE739C6EF39BDEF
39C6EF39C6EF42CEF739C6EF42CEF739C6EF42D6B539D6AD42DEAD39D6A542DEAD39D6A5
42DEAD39D6A542DEA539DE9C42DEA539DE9C42E7A539DE9C42E79C39DE9442E79C39DE94
42E79439DE9442E79442E73942E79439DE8C42E79439DE8C42E79439DE8C42E78C39DE8C
42DE8C39D68442DE8C39D68439DE8C39D68439D68C39D68439D68C39CE8439D68C39CE84
39D68452CE315ACE3952CE315ACE3952C63139CE4A39C64239CE4A31C64239CE4A39C642
39CE4A31C64239CE4A39C64239CE4A31C64239C64A39C64239C64A31C64294C6398CBD31
8CC6398CBD318CC6397BBD317BC63973BD317BC63973BD317BC63973BD317BC63973BD31
7BBD3973BD317BBD3973BD3173BD3973B53173BD3973B531ADAD39A5A531ADAD39A5A531
ADAD39A5A531ADAD39A5A531ADA539A5A531ADA539A5A531A5A5399CA531A5A5399C9C31
A5A5399C9C31A5A539000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000733931733931
734231734231734231734231734231734231734231734231734231734231734231734231
734231734231735231735231735231735231735231735231735231735231735231735231
735231735231735231735231735231736331736331736331736331736331736331736339
736331736331736331736339736331736331737B31737B39737B31737B31737B31737B39
737B31737B39737B317B7B39737B316384395A8431638439428431428439428431428C39
428431428C39428C314A8C39428C314A8C39428C314A8C39428C314A8C39318C4A398C4A
318C4A39944A31944A39944A31944A39944A31944A39944A31944A39944A39944A399C4A
31944A399C63399C63399C63399C63399C63399C6339A56B399C6339A56B39A56339A56B
39A56339A56B39A56B39A56B39A56B39AD6B39A56B39AD6B39AD6B39AD6B39AD6B39AD73
39AD6B39AD7339AD7339B57339B57339B57339B57339B57B39B57339B57B39B57B39BD5A
39BD5A39BD6339BD6339B5E739B5E739BDE739BDE739BDEF39BDE739BDEF39BDEF39C6EF
39C6EF39C6EF39C6EF39CEF739C6EF39CEF739CEF739D6AD39D6AD39D6AD39D6AD39DEAD
39DEA539DEA539DEA539DEA539DE9C39DEA539DE9C39E79C39DE9C39E79C39E79439E79C
39E79439E79439E79439E79439E78C39E79439E78C39E78C39DE8C39DE8C39DE8C39DE8C
39DE8C39DE8C39D68C39DE8C39D68439D68C39D68439D68439D68439D68439CE8439D684
39CE845ACE395ACE395ACE395ACE3139CE4A39CE4A39CE4A39CE4A39CE4A39C64A39CE4A
39C64A39C64A39C64A39C64A39C64239C64A39C64A39C64A39C64239C64A39C6428CC639
8CC6317BC6397BBD317BC6397BBD317BC6397BBD317BBD3973BD317BBD3973BD317BBD39
73BD3173BD3973BD3173BD3973BD3173BD3173B53173BD39ADA531ADA531ADA531ADA539
A5A531A5A531A5A531A5A531A5A531A5A531A5A531A5A531A5A531A5A531A5A531A5A531
9CA531A5A5319C9431A59C31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006B4231
734A396B42317342316B4231734A397342317342316B4231734A397342317342316B4231
734A397342317342396B4A317352396B52317352316B4A317352396B52317352396B4A31
7352396B4A317352396B4A317352396B63317363396B63317363396B63317363396B6331
736B396B63317363396B63317363396B63317363396B7331737B39737331737B39737331
6384395A84316384395A8431638C395A84316384395A8431638C39428431428C39428431
4A8C394284314A8C394284314A8C394284314A8C394284314A8C39428C31398C4A318C42
39944A318C4239944A318C4239944A31944A39944A319442399C4A31944A399C4A319442
399C4A319463399C6B319463399C6B319C63399C6B319C6339A56B319C6339A56B319C63
39A56B31A56339A56B31A56339AD6B31A56339AD6B31A56B39AD7331AD6B39AD7331AD6B
39B57339AD6B39B57331AD6B39B57B31AD7339B57B31B57339BD7B39B57339BD7B31B573
39BD6339ADDE39B5E731B5DE39BDE739B5E739BDE739B5E739C6EF39BDE739C6EF39BDE7
42C6EF39C6EF39CEEF39C6EF42CEF739CEEF39CEF739D6AD42DEB539D6AD42DEAD39D6A5
42DEAD39DEA542DEA539DE9C42E7A539DE9C42E7A539DE9C42E79C39DE9442E79C39DE94
42E79C39E79442B5D639ADCE42E79439DE8C42E79439DE8C42E79439DE8C42DE8C39DE8C
42DE8C39D68C39DE8C39D68442DE8C39D68439D68C39CE8439D68C39CE8439D68431CE7B
39D68439CE845ACE3952C63139CE4A39CE4A39CE4A31C64239CE4A39C64239CE4A31C642
39CE4A39C64239CE4A31C64239CE4A39C64239C64A31C64239C64A39C64239C64A8CBD31
7BC6397BBD317BC63973BD317BC63973BD317BC63973BD317BC63973BD317BBD3973BD31
7BBD3973BD3173BD3973B5317BBD3973B53173BD396BB53173BD39A5A531ADA539A5A531
ADAD39A5A531A5A539A5A531A5A539A5A531A5A5399C9C31A5A5399C9C31A5A5399C9C31
A5A5399C9431A59C399C9431000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7342317342316B42317342317342317342317342317342317342317342316B4231734231
7342317342317342317342317352317352316B52317352317352317352316B5231735231
7352317352316B52317352317352317352396B63317363317363317363396B6331736331
7363317363396B63317363317363317363396B7B31737B315A84315A84395A84315A8431
5A84315A84395A84315A84315A84316384395A84315A84315A8431428439428431428431
428431428C39428431428C39428C314A8C39428431428C39428C394A8C39428C31398C4A
318C4A39944A318C4239944A39944A39944A31944A39944A39944A39944A31944A39944A
39944A399C4A319463399C63399C63399C6B319C63399C63399C6339A56B319C6339A56B
39A56B39A56B31A56339A56B39A56B39AD6B39A56B39AD6B39AD6B39AD6B39AD6B39AD6B
39AD7339B57339AD7339B57339B57339B57339B57339B57339B57339B57B39B57B39B57B
39B5DE39B5E739B5DE39B5E739B5E739BDE739BDE739BDE739BDE739C6EF39BDE739C6EF
39C6EF39C6EF39C6EF39CEEF39CEEF39CEF739CEEF39D6AD39D6AD42DEAD39D6AD39DEAD
39DEA539DEA539DEA539DEA539DEA542E7A539DE9C39E79C39E79C42E79C39E79439E794
39B5D642B5D639ADD639B5D639E78C39E79439DE8C39DE8C39DE8C39DE8C39DE8C39DE8C
39DE8C39DE8C39D68439D68C39D68C39D68C39D68439D68439D68439D68439CE8439CE84
39CE8439CE845ACE3139CE4A39CE4A39CE4A39C64A39CE4A39CE4A39CE4A39C64A39C64A
39C64A39C64A39C64239C64A39C64A39C64A39C64239C64A39C64A39C64A39BD4239C65A
7BC6397BC6397BBD317BBD397BBD317BBD3973BD317BBD3173BD317BBD3973BD3173BD39
73BD3173BD3973B53173BD3173BD3173BD3973B53173B53173B531ADA539A5A531A5A531
A5A531A5A539A5A531A5A531A5A531A5A5399CA531A5A531A5A531A5A5399C9C319C9C31
9C94319C9C319C94319C9431000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007342396B42317342396B4231734A396B42317342396B42317342396B4231
7342396B4231734A396B42317342396B4A317352396B4A317352396B4A317352396B4A31
7352396B4A317352396B4A317352396B4A317352396B63317363396B5A317363396B6331
7363396B7B317384396B7B317384396B7B317384396B7B317384395A7B315A84395A7B31
5A84395A7B316384395A7B316384395A7B316384395A84316384395A84314A8439428431
4A8C394284314A8C394284314A8C394284314A8C394284314A8C394284314A8C39318C42
398C4A318C4239944A318C4239944A318C4239944A31944239944A319442399C4A319442
399C4A319442399C6B319463399C6B319463399C6B319C6339A56B319C6339A56B319C63
39A56B319C6339A56B31A56339AD6B31A56339AD6B31A56B39AD6B31A56B39AD7331AD6B
39AD7331AD6B39B57331AD7339B57331AD7339B57B31B57339B57B31B573398CDE398CDE
39B5E731ADDE39B5E739B5DE39BDE739B5DE39BDE739BDE739C6EF39BDE739C6EF39BDE7
39C6EF39C6E742CEEF39C6EF42CEF739C6EF42CEF739CEEF42DEB539D6AD42DEAD39D6A5
42DEAD39DEA542DEAD39DEA542E7A539DE9C42E7A539DE9C42E79C39DE9442E79C39ADD6
42B5DE39ADD642B5D639ADCE42B5D639DE8C42E79439DE8C42E79439DE8C42DE8C39DE8C
39DE8C39D68439DE8C39D68439DE8C39D68439D68C39CE8439D68C39CE8439D68439CE84
39D68439C6BD39C6BD39CE4A39CE4A31C64239CE4A39C64239CE4A39C64239CE4A39C642
39CE4A31C64239CE4A39C64239C64A31C64239C64A39C64239C64A31BD4239C66331BD5A
39C66373BD317BC63973BD317BC63973BD317BBD3973BD317BBD3973BD317BBD3973B531
73BD3973B53173BD3973B53173BD396BB53173BD396BB53173BD396BB531ADA539A5A531
A5A539A5A531A5A5399C9C31A5A5399C9C31A5A5399C9C31A5A5399C9C31A59C399C9431
9C9C399C94319C9C399C94319C9C39000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007342316B42317342316B42317342316B42317342316B4231734231
6B42317342317342317342316B42317342316B52317352316B52317352316B5231735231
6B52317352316B52317352316B52317352316B52317352316B7B316B7B316B7B31737B31
6B7B31737B316B7B31737B396B7B31737B316B7B31737B316B7B315A84315A84315A8439
5A7B315A84315A84315A84395A84315A84395A84315A84395A84315A8439428431428439
428431428439428431428C39428431428C39428431428C39428C314A8C39428C314A8C39
318C42398C4A318C4A398C4A318C4A39944A39944A39944A31944A39944A31944A39944A
31944A39944A39944A399C63319C63399C63399C63399C6B399C63399C6B399C6339A56B
39A56339A56B39A56B39A56B39A56B39A56B39A56B39AD6B39A56B39AD6B39AD6B39AD6B
39AD6B39AD7339AD7339B57339AD7339B57339B573398CDE3984D6398CDE398CDE398CDE
398CDE39B5E739B5DE39B5E739B5E739BDE739BDE739BDE739BDE739C6E739C6E739C6EF
39C6EF39C6EF39C6EF39CEEF39CEEF39CEEF39CEEF39CEF739D6AD39DEAD39DEAD39DEAD
39DEA539DEA539DEA539DEA539DEA539E7A539E79C42E79C39E79C39E79C39ADDE39B5DE
39ADD639B5D639ADD639ADD639ADCE39E78C39DE8C39DE8C39DE8C39DE8C39DE8C39DE8C
39D68C39DE8C39D68439D68C39D68439D68439D68439D68439CE8439D68439CE8439CE84
39C6BD39C6BD39C6BD39CE4A39CE4A39CE4A39CE4A39CE4A39C64A39CE4A39C64A39CE4A
39C64239C64A39C64A39C64A39C64239C64A39C64A39C64A39C64239C65A39BD5A39C65A
39BD5A7BBD397BBD317BBD3973BD317BBD3973BD317BBD3973BD3173BD3973BD3173BD39
73B53173BD3973B53173BD3973B53173B53973B53173B5396BB53173B531A5A531A5A531
A5A531A5A531A5A531A5A5399CA531A5A5319C9C319CA5319C94319C94319C94319C9431
9C94319C94319C94319C94319C9431000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007342396B42317342316B4231734A396B42317342316B4231
734A396B42317342316B4231734A396B42317342396B4A317352396B4A317352396B4A31
7352396B4A317352396B4A317352396B63317363396B5A317363396B7B31737B396B7B31
7384396B7B31737B396B7B317384396B7B31737B396B7B31737B396B7B315A8439527B31
5A84395A7B315A84395A7B316384395A7B315A84395A7B316384395A7B31638439427B31
4A84394284314284394284314A8C394284314A8C394284314A8C394284314A8C39428431
4A8C39318C42398C4A318C4239944A318C4239944A318C4239944A31944A39944A319442
399C4A31944A399C4A319463399C6B319463399C6B319C6339A56B319C6339A56B319C63
39A56B319C6339A56B31A56339AD6B31A56339AD6B31A56339AD6B31A56B39AD6B31A56B
39AD7331AD6B39B57331AD6B39B5733984D6398CDE3184D6398CDE3984DE398CDE3184DE
398CE739ADDE39B5E731B5DE39BDE739B5DE39BDE739BDDE42C6EF39BDE739C6EF39BDE7
42CEEF39C6E739CEEF39C6E742CEF739CEEF42CEF739CEEF42D6F739D6AD42DEAD39D6A5
42E7AD39DEA542E7AD39DEA542E7A539DE9C42E7A539DE9C42E7A539ADDE42B5DE39ADD6
42B5DE39ADD642B5D639ADCE42B5D639ADCE42E79439DE8C42E79439DE8C39DE8C39D684
42DE8C39D68439DE8C39D68442DE8C39D68439D68C39CE8439D68C39CE8439D68431C6BD
39C6C639C6BD39C6BD31BDB539CE4A39C64A39CE4A31C64239CE4A39C64239CE4A31C642
39CE4A39C64239C64A31C64239C64A39C64239C64A31BD4239C66339BD5A39C66331BD5A
39C66339BD5A7BBD3973BD317BC63973BD317BBD3973B5317BBD3973B53173BD3973B531
73BD3973B53173BD396BB53173BD396BB53173B5396BB53173B5396BB53173B5399C9C31
A5A5399C9C31A5A5399C9C31A5A5399C9C31A5A5399C9431A59C399C94319C9C39949431
9C9C399C94319C94399494319C9C39949431000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007342316B42316B42316B42317342316B42316B4231
6B42317342316B42316B42316B42317342316B42316B42316B52317352316B4A316B5231
6B52317363316B63316B63316B63316B63316B63316B63316B63316B63396B7B316B7B31
6B7B316B7B396B7B316B7B316B7B316B7B396B7B316B7B316B7B316B7B396B7B315A7B31
5A7B315A84395A7B315A84315A84315A84395A7B315A84315A84315A84395A7B315A8431
428431428439428431428439428431428C394284314284394284314A8C39428431428C39
428C39398C4A318C42398C4A398C4A39944A318C4A39944A39944A39944A31944A39944A
39944A39944A31944A39944A399463399C63319C63399C63399C63399C6B319C63399C63
39A56339A56B31A56339A56B39A56B39A56B39A56339A56B39A56B39AD6B39AD6B39AD6B
39AD6B39AD7339AD6B3984D63984D6398CD63984D6398CD6398CDE398CDE398CDE398CDE
398CDE398CE739B5DE39B5E739B5E739BDE739BDE739BDE739BDE739C6E739BDE739C6E7
39C6E739C6EF39C6EF39CEEF39CEEF39CEEF39CEEF39CEEF39D6EF42DEAD39DEAD39DEAD
39DEAD42DEAD39DEA539DEA539DEA542E7A539E79C39E79C39ADDE42B5DE39ADDE39ADDE
39ADD639ADD639ADD639ADD639ADCE39ADCE39DE8C39DE8C39DE8C39DE8C39DE8C39DE8C
39D68C39DE8C39D68439D68439D68439D68C39CE8439D68439CE8439D68439C6BD39C6BD
39C6BD39C6BD39C6BD39C6BD39CE4A39CE4A39C64A39CE4A39C64A39CE4A39C64239C64A
39C64A39C64A39C64239C64A39C64A39C64A39C64239C65A39C65A39C66339BD5A39BD5A
39BD5A39BD5A73BD317BBD3973BD317BBD3973BD3173BD3173BD3173BD3973B53173BD31
73B53173BD3973B53173B53173B53173B5396BB53173B5316BB53173B5396BB531A5A531
9CA531A5A5399C9C319CA5319C9C319C9C399C94319C94319C94319C94319C94319C9431
9C94319C94319494319C94319C94319C9431000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007342396B42317342396B42317342396B4231
7342396B42317342396B42317342396B42317342396B42317342396B5A317363396B5A31
7363396B5A317363396B5A316B63396B5A317363396B5A316B63396B5A316B63396B7B31
6B7B396B73316B7B396B7B316B7B396B73316B7B396B73316B7B396B73316B7B39527B31
5A8439527B315A8439527B315A8439527B315A84395A7B315A84395A7B316384395A7B31
428439427B314284394284314A84394284314A8C394284314A8C394284314A8C39428431
4A8C39428431398C4A318C42398C4A318C4239944A318C4239944A318C4239944A319442
39944A319442399C4A31944A399C6B319463399C6B319463399C6B319C6339A56B319C63
39A56B319C6339A56B31A56339A56B31A56339AD6B31A56339AD6B31A56B39AD6B31A56B
39AD733184CE3984D63184CE398CD63184D6398CDE3184D6398CDE3184D6398CDE318CDE
398CE7318CDE39B5E739B5DE39BDE739B5DE39BDE739BDE739C6E739BDE742C6EF39C6E7
39C6EF39C6E742CEEF39C6E742CEEF39CEEF42D6F739CEEF42D6F739CEEF42DEAD39DEAD
42E7AD39DEA542E7AD39DEA542E7A539DE9C42B5E739ADDE42B5E739ADDE42B5DE39ADD6
42ADDE39ADD642ADD639ADCE42ADD639A5CE42ADCE39DE8C39DE8C39D68442DE8C39D684
39DE8C39D68439DE8C39D68439D68C39CE8439D68C39CE8439D68439C6BD39CEC639C6BD
39C6BD31BDB539C6BD39BDB539C6BD31C64239CE4A39C64239CE4A31C64239CE4A39C642
39C64A31C64239C64A39C64239C64A31BD4239C66331BD5A39C66331BD5A39C66331BD5A
39C66331BD5A39BD6373BD317BBD3973B5317BBD3973B53173BD3973B53173BD3973B531
73BD396BB53173BD396BB53173B5396BB53173B5396BB53173B5396BAD3173B5396BAD31
A5A5399C9C31A5A5399C9C319C9C399C94319C9C399C94319C9C399494319C9C39949431
9C94399494319C94399494319C94399494319C9439000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006B42316B42316B42316B42316B4231
6B42316B42316B42316B42316B42316B42316B4A316B4A316B4A316B4A316B63316B6331
6B63316B63316B63316B63316B63316B63316B63316B63316B5A316B63316B63316B7B31
6B7B316B7B316B7B316B7B316B7B316B7B316B7B316B7B396B7B316B7B316B7B316B7B39
527B315A7B315A7B315A7B395A7B315A7B395A7B315A84395A7B315A84395A7B315A8439
5A7B314284394284314284394284314284394284314284394284314284394284314A8C39
4284314A8C39318C4A398C4A318C42398C4A318C4A398C4A318C4A39944A31944A39944A
31944A39944A39944A39944A31944A399C63399463399C63319C63399C63399C63399C6B
399C6339A56B39A56339A56B39A56339A56B39A56B39A56B39A56B39AD6B39AD6B3984CE
3984CE3984CE3984CE3984D63984D63984D63984D6398CDE3984D6398CDE398CDE398CDE
398CDE398CE739B5E739B5E739B5E739BDE739BDE739BDE739BDE739C6E739C6E739C6EF
39C6E739C6EF39C6E739CEEF39CEEF39CEEF39CEEF39D6EF39D6EF39D6F739DEAD42DEAD
39DEAD39DEAD39DEA542E7A539E7A539B5E739ADE742ADE739ADDE39ADDE39ADDE39ADDE
39ADD639ADD639ADD639ADD639ADCE39ADCE39ADCE39DE8C39DE8C39DE8C39DE8C39DE8C
39D68439D68C39D68439D68C39D68439D68439CE8439D68439C6BD39C6BD39C6BD39C6BD
39C6BD39C6BD39C6BD39C6BD39BDB539CE4A39C64A39CE4A39C64A39C64A39C64A39C64A
39C64239C64A39C64A39C64A39C64239C66339C65A39C66339BD5A39BD5A39BD5A39BD5A
39BD5A39BD5A39BD5A73BD3973BD3173BD3973BD3173BD3973B53173BD3973B53173B539
73B53173B53973B53173B5396BB53173B5316BB53173B5396BB5316BB5316BB5316BB531
9C9C319CA5319C94319C94319C94319C94319C94319C94319C94319C94319C94319C9431
9494319C94319494319C9431949431949431949431000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000007342396B42316B42316B3931
7342396B42316B52316B4A317352396B4A316B52316B4A317352396B4A316B52396B5A31
7363396B5A316B6331635A316B63396B5A316B6339635A316B63396B5A316B6339635A31
6B7B396B73316B7B396373316B7B396373316B7B396373316B7B396B73316B7B39637331
6B7B39527B315A7B39527B315A8439527B315A8439527B315A84395A7B315A8439527B31
5A8439427B31428439427B314A84394284314284394284314A8C394284314A8C39428431
4A8C394284314A8C39318442398C4A318C42398C4A318C4239944A318C4239944A318C42
39944A31944A39944A319442399C4A319463399C63319463399C6B319C63399C6B319C63
39A56B319C6339A56B319C6339A56B31A56339A56B31A5633984CE317BC63984CE317BC6
3984D63984CE3984D63184CE398CD63184D6398CDE3184D6398CDE3984DE398CDE3184DE
398CE7398CDE398CE731ADDE39BDE739B5E739BDE739B5E739C6EF39BDE739C6EF39BDE7
42CEEF39C6E739CEEF39C6E742CEEF39CEEF42D6EF39CEEF42D6F739D6EF42DEAD39DEAD
42E7AD39DEA542E7AD39DEA542B5EF39ADE742B5E739ADDE42B5E739ADDE42ADDE39A5D6
42ADDE39ADD642ADD639A5CE42ADD639A5CE39ADCE39D68442DE8C39D68C39DE8C39D684
42DE8C39D68439D68C39CE8439D68C39CE8439D68431C6BD39CEC639C6BD39C6BD31BDB5
39C6BD39C6B539C6BD31BDB539C6BD39BDB539CE4A31C64239CE4A39C64239C64A31C642
39C64A39C64239C64A31BD4239C66339BD5A39C66331BD5A39C66339BD5A39BD5A31BD5A
39C66339BD5A39BD5A31B55A7BBD3973B53173BD3973B53173BD3973B53173BD396BB531
73BD396BB53173B5396BAD3173B5396BB53173B5396BAD3173B5396BAD316BB5396BAD31
6BB5399C94319C9C399494319C9C399494319C94399494319C9C399494319C9439949431
9C94399494319C9431948C319C9439949431948431947B31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006B42316B42316B4A31
6B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A31
6B63316B63316B5A316B63316B63316B63316B5A316B63316B5A316B6331635A316B5A31
6B5A316B7B396373316B7B316B7B316B7B396373316B7B316B7B316B7B396373316B7B31
6B73315A7B39527B31527B315A7B315A7B39527B315A7B315A7B315A7B395A7B315A7B31
5A7B315A8439427B31428439428431428439428431428439428439428439428431428439
4284314A8C39428431398C4A398C4A398C4A318C42398C4A398C4A39944A318C4A39944A
39944A39944A31944A39944A39944A399463319463399C63399C63399C6B319C63399C63
399C6339A56B319C6339A56B39A56B397BC6397BC6397BC6397BC63984CE397BCE3984CE
3984CE3984D63984CE3984D63984D6398CD63984D6398CD6398CDE398CDE398CDE398CDE
398CDE398CE7398CDE39B5E739B5E739BDE739B5E739BDE739BDE739C6EF39BDE739C6E7
39C6EF39C6EF39C6E739CEEF39CEEF39CEEF39CEEF39D6EF39D6EF42D6F739D6EF39DEAD
39DEAD42E7AD39DEA539ADEF39ADE742ADE739ADE739ADE739ADDE39ADDE39ADDE39ADDE
39ADD639ADD639ADD639ADD639ADCE39ADCE39A5CE39A5CE39DE8C39DE8C39D68439D68C
39D68439D68C39D68439D68439D68439D68439C6BD39C6BD39C6BD39C6BD39C6BD39C6BD
39C6BD39C6BD39BDB539C6B539BDB539C6B539C64239C64A39C64A39C64A39C64239C64A
39C64A39C64A39C64239C65A39C65A39C66339BD5A39BD5A39BD5A39BD5A39BD5A39BD5A
39BD5A39BD5A39BD5A39BD5A73BD3173BD3973B53173B53173B53173BD3973B53173B531
73B53173B5396BB5316BB5316BB53173B5396BAD316BB5316BAD316BB5396BAD317BAD31
73AD319C94399C94319C94319C94319C94319494319C94319494319C9431949431949431
949431949431949431949431948431948431948431948431000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006B4A31
6B52396B4A316B52396B4A316B4A31634A316B52396B4A316B52396B4A316B52396B4A31
6B5239635A316B6339635A316B6339635A316B6339635A316B6339635A316B6339635A31
6B6339635A316B7B396373316B7B396373316B7B396373316B7B396373316B7B39637331
6B7B396373315A7B395273315A7B39527B315A7B39527B315A7B39527B315A8439527B31
5A8439527B31428439427B31428439427B31428439427B314A84394284314A8439428431
4A8C394284314A8C39428431398C4A318442398C4A318C42398C4A318C4239944A318C42
39944A318C4239944A31944A39944A319442399C6B319463399C6B319463399C6B319C63
399C6B319C63397BC6317BC6397BC6317BC63984CE317BC63984CE317BC63984CE317BCE
3984D63184CE3984D63184CE398CD63184D6398CDE3184D6398CDE3184D6398CDE318CDE
398CE7318CDE3994E7318CE739BDE739B5E739BDEF39B5E739BDEF39BDE739C6EF39BDE7
39C6EF39C6E742CEEF39C6E742CEEF39CEE742D6EF39CEEF42D6EF39D6EF42DEF739DEAD
42E7AD39ADEF42B5EF39ADE742B5EF39ADE742B5E739ADDE42ADE739A5DE42ADDE39A5D6
42ADDE39A5D642ADD639A5CE39ADD639A5CE42ADCE39A5C639DE8C39D68439DE8C39D684
39DE8C39D68439D68C39CE8439D68C39C6BD39CEC639C6BD39C6C631C6BD39C6BD39C6B5
39C6BD39BDB539C6BD39BDB539C6BD31BDB539C6B531C64239CE4A31C64239CE4A39C642
39C64A31BD4239C66331BD5A39C66331BD5A39C66331BD5A39C66331BD5A39BD6331BD5A
39BD5A31B55A39BD5A31B55A39BD5A73B53173BD3973B53173BD396BB53173BD396BB531
73B5396BAD3173B5396BAD3173B5396BAD3173B5396BAD316BB5396BAD317BB53973AD31
7BB53973AD319C94399494319C94399494319C94399494319C9439948C319C9439948C31
949439948C31948439947B319484398C7B319484398C7B31948431000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
6B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A31
6B4A316B4A316B5A316B63316B5A316B63316B5A316B6331635A316B5A31635A316B5A31
635A316B5A31635A316B7B316373316373316373316B73316373316B73316373316B7339
6373316B7331637331527B39527B31527B39527B315A7B39527B315A7B315A7B315A7B39
5A7B315A7B395A7B31427B39427B31427B39427B31428439428431428439428431428439
428431428439428431428439318442398C4A318C4A398C4A318C42398C4A318C4A398C4A
318C4A39944A39944A39944A31944A39944A399463399463399463399C633973BD397BC6
3173BD397BC6397BC6397BC6397BC6397BC6397BC6397BC6397BC63984CE397BCE3984CE
3984CE3984CE3984CE3984D63984D63984D63984D6398CDE3984D6398CDE398CDE398CDE
398CDE398CE7398CE7398CE739B5E739B5E739B5E739BDE739BDE739BDE739BDE739C6EF
39C6E739C6EF39C6EF39CEEF39CEEF39CEEF39CEEF39D6EF39D6EF39D6EF39D6EF42DEF7
39ADEF39ADEF39ADEF39ADEF39ADE739ADE739ADE739ADE739ADDE39ADDE39ADDE39ADDE
39A5D639ADD639A5D639ADD639A5CE39A5CE39A5CE39A5CE39A5C639DE8C39D68C39D68C
39D68439D68439D68439D68439C6BD39CEC639C6BD39C6BD39C6BD39C6BD39C6BD39C6BD
39C6B539C6BD39C6B539C6B539BDB539BDB539BDB539C64A39C64A39C64A39C64A39C64A
39C64239C66339C65A39C66339BD5A39C65A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A
39BD5A39BD5A39BD5A39BD5A39B55A73B53973B53173B53973B53173B53973B53173B539
6BB5316BB5316BB5316BB5396BAD316BB5316BAD316BB5396BAD317BAD3173AD3173AD31
73AD3173AD319494319C9431949431949431949431949431949431949431949431949431
948C319484319484319484319484319484319484319484318C7B31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000634A316B4A31634A316B5239634A316B4A31634A316B5239634A316B4A31634A31
6B5239634A316B4A31635A316B6339635A316B6331635A316B6339635A316B6339635A31
6B6339635A316B63396373316B7B396373316B7B396373316B7B396373316B7339637331
6B7B396373316B73396373315A7B395273315A7B395273315A7B39527B315A7B39527331
5A7B39527B315A7B39527B31428439427B31428439427B314A8439427B31428439427B31
4A84394284314A84394284314A8C39318442398C4A318442398C4A318C42398C4A318C42
39944A318C4239944A318C4239944A31944A39944A3173BD397BC63173BD397BC63173BD
397BC63173BD397BC63173BD397BC6317BC6397BC6317BC63984CE317BC63984CE317BC6
3984CE3184CE3984D63184CE398CD63984D6398CDE3184D6398CDE3984DE398CDE3184DE
398CE7398CDE398CE7318CDE39B5EF39B5E739BDEF31B5E739BDEF39BDE739C6EF39BDE7
42C6EF39C6E739CEEF39C6E742CEEF39CEEF42D6EF39CEEF42D6F739D6EF42D6F7394AEF
42B5F739ADEF42ADEF39ADE742ADEF39ADE742ADE739A5DE42ADE739A5DE42ADDE39A5D6
42ADDE39A5D639ADD639A5CE42ADD639A5CE39A5CE39A5C642ADCE39D68439DE8C39D684
39DE8C39D68439D68C31C6BD39CEC639C6BD39CEC631C6BD39CEC639C6BD39C6BD31BDB5
39C6BD39BDB539C6BD31BDB539C6BD39BDB539C6B531BDAD39CE4A39C64239C64A31BD42
39C66339BD5A39C66331BD5A39C66339BD5A39C66331BD5A39C66339BD5A39BD5A31B55A
39BD5A31B55A39BD5A31B55239BD5A31B55273B5396BB53173B5396BB53173B5396BAD31
73B5396BAD3173B5396BAD3173B5396BAD316BB53963AD317BB53973AD317BAD3973A531
7BAD3973AD3173AD39948C319C9439948C319494398C8C31949439948C319494318C7B31
9484398C7B319484318C7B319484398C7B319484318C7B319484398C7B31000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000634A316B4A316B4A316B4A31634A316B4A316B4A316B4A31634A316B4A31
6B4A316B4A31634A316B4A31635A316B6331635A31635A31635A316B6331635A31635A31
635A31635A31635A31635A31637331637339637331637331637331637339637331637331
637331637339637331637331527331527B39527331527B31527B315A7B39527331527B31
527B315A7B39527B315A7B31427B31427B39427B31427B39427B31428439427B31428439
42843942843942843142843942843939844A318442398C4A398C4A398C4A318C42398C4A
398CB5398CB5318CB5398CBD398CBD398CBD318CBD3973BD3973BD3973BD3173BD397BBD
397BBD397BC6317BC6397BC6397BC6397BC6317BC6397BC6397BC63984CE397BC63984CE
3984CE3984CE3984CE3984D63984D6398CD63984D6398CD6398CDE398CDE398CDE398CDE
398CDE398CE7398CE7398CE7398CE739B5E739B5E739BDE739BDE739BDEF39BDE739C6EF
39C6EF39C6EF39C6EF39CEEF39CEEF39CEEF39CEEF39CEEF39D6EF424AF7394AEF394AEF
3952EF42ADEF39ADE739ADEF39ADE742ADE739ADE739ADE739ADDE39ADDE39A5DE39ADDE
39A5D639ADD639A5D639A5D639A5D639A5D639A5CE39A5CE39A5CE39A5CE39D68439D684
39D68439D68C39C6C639CEC639C6BD39CEC639C6BD39C6BD39C6BD39C6BD39C6BD39C6BD
39C6B539C6BD39BDB539BDB539BDB539C6B539BDB539BDB539C64A39C64A39C64239C663
39C65A39C66339BD5A39C65A39BD5A39C65A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A
39BD5A39BD5A39B55A39B55A39B55A39B55A39B55273B5316BB53173B5396BB5316BB531
6BB5316BB5396BAD316BAD316BAD316BAD396BAD3173AD3173AD3173AD3973AD3173AD31
73AD3173AD3173A531949431949431949431948C319494319494319484318C7B31948431
9484319484318C7B319484318C84319484318C7B318C84318C8431948431000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000634A316B4A31634A316B4A39634A316B4A31634A316B4A39634A31
6B4A39634A316B4A39634A316B4A39635A316B6339635A316B6339635A316B6339635A31
636339635A31636339635A31635A39637331637339637331637339637331637339637331
637339637331637339637331637339527331527B395273315A7B395273315A7B39527331
5A7B395273315A7B395273315A7B39397B31427B39427B31428439427B31428439427B31
39847B317B7339847B31847339847B3184733984B5317BAD3984B53184AD398CB53184B5
398CB53184B5398CBD318CB5398CBD318CB53994BD318CBD3973BD3173BD397BC63173BD
397BC63173BD397BC63173BD397BC6317BC63984CE317BC63984CE317BC63984CE317BC6
3984CE317BCE3984D63184CE398CD63184D6398CDE3184D6398CDE3184D6398CDE318CDE
398CE7318CDE3994E7318CE73994E731B5E739BDEF39B5E739BDEF39BDE739C6EF39BDE7
39C6EF39C6E742CEEF39C6EF42CEEF39C6EF42D6F739CEEF424AF7394AEF4252F7394AEF
4252EF39ADE742B5EF39ADE742ADEF39ADE742ADE739A5DE42ADE739A5DE42ADDE39A5D6
42ADDE39A5D642ADD639A5CE39A5D639A5CE39A5CE39A5CE39A5CE399CC639DE8C39D684
39D68C39C6BD39CEC639C6BD39CEC639C6BD39CEC639C6BD39C6BD39C6B539C6BD39BDB5
39C6BD31BDB539C6BD39BDB539C6B531BDAD39C6B539BDAD39BDB531C64239C66331BD5A
39C66331BD5A39C66331BD5A39C66331BD5A39C66331BD5A39BD5A31B55A39BD5A31B55A
39BD5A31B55239BD5A31B55239BD5A31B55239B55A6BAD3173B5396BAD3173B5396BAD31
73B5396BAD316BB5396BAD316BAD3973AD317BAD3973AD317BAD3973A53173AD3973A531
73AD396BA53173AD396BA5319494398C8C319494398C8C319484398C7B319484398C7B31
9484398C7B319484398C7B319484318C7B319484398C7B319484318C7B31000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000634A316B4A31634A316B4A31634A31634A31634A316B4A31
634A31634A31634A31634A31634A31634A31635A31635A31635A31635A31635A31635A31
635A31635A31635A31635A31635A31635A31527331527331527331527331527331527331
527331527331527331527339527331527331317331397339317331397339397331397B39
317331397B39397331397B39397B31317B73317B73397B73317B73397B73317B73398473
317B7339847B31847339847B31847339847B317BAD3984B53184AD3984B53184B53984B5
3184B5398CB5318CB5398CBD318CB5398CBD398CBD398CBD3173BD3973BD3973BD397BC6
3973BD397BC6397BC6397BC6397BC6397BC6397BC6397BC6397BC6397BCE397BC63984CE
3984CE3984CE3984CE3984D63984D63984D63984D6398CDE3984D6398CDE398CDE398CDE
398CDE398CE7398CE7398CE7398CE7398CE739B5E739BDEF39BDE739BDEF39BDEF39C6EF
39C6EF39C6EF39C6EF39CEEF39C6EF39CEEF394AEF394AEF394AEF394AEF394AEF424AEF
394AEF39ADEF39ADE739ADE739ADE739ADE739ADE739ADE739ADDE39ADDE39A5DE39ADDE
39A5D639A5D639A5D639A5D639A5CE39A5CE39A5CE39A5CE39A5C639A5C639A5C639D68C
39CEC639CEC639C6C639CEC639C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39C6BD
39C6B539C6B539BDB539C6B539BDB539BDB539BDB539BDB539BDAD39C66339C65A39C663
39BD5A39C65A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A
39B55A39B55A39B55A39B55A39B55A39B55A39B55A39B55A6BAD316BB5316BAD316BB539
6BAD316BAD316BAD316BAD3973AD3173AD3173AD3173AD3173AD3173AD3173AD3173AD31
73A53173AD3173A53173AD316BA5319494318C7B319484318C7B319484318C7B31948431
8C7B318C84318C7B318C84318C7B318C84318C7B318C84318C7B318C84318C7B31000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000634A316B4A316342316B4A39634A316B4A31634A31
6B4A39634A316B4A316342316B4A39634A316B4A31635A316B6339635A31637331636B31
637339636B316373395A6B316373395A6B31636B394A7331527B394A73315273394A7331
527B394A73315273394A7331527B39527331397339317331397B39317331397B39317331
397B39317331397B39317331397B39317331397B7331736B397B7B317B73398473317B73
39847B317B7339847B317B7339847B3184733984AD317BAD3984B53184AD3984B53184AD
398CB53184B5398CB53184B5398CBD318CB5398CBD318CB53973BD3173BD397BBD3173BD
397BC63173BD397BC63173BD397BC6317BC6397BC6317BC63984CE317BC63984CE317BC6
3984CE3184CE3984CE3184CE398CD63184D6398CD63184D6398CDE3984DE398CDE3184DE
398CE7398CDE398CE7318CDE3994EF398CE739BDEF31B5E739BDEF39BDE739BDEF39BDE7
42C6EF39C6EF39C6EF39C6EF42CEF73942EF424AEF3942E7424AEF394AEF424AEF394AE7
4252EF394AE742B5EF39ADE742B5EF39ADE742ADE739A5DE42ADE739A5DE39ADDE39A5D6
42ADDE39A5D639ADD639A5CE42ADD639A5CE39A5CE399CC642A5CE399CC639A5C631C6C6
39CEC639C6BD39CEC631C6BD39CEC639C6BD39C6BD31C6BD39C6BD39C6B539C6BD31BDB5
39C6BD39BDB539C6B531BDB539C6B539BDAD39BDB531B5AD3994B53994AD39C66331BD5A
39C66339BD5A39C66331BD5A39C66339BD5A39BD5A31B55A39BD5A31B55A39BD5A31B552
39BD5A31B55A39B55A31B55239B55A31B55239B55A31AD5273B5396BAD316BB5396BAD31
6BB5396BAD316BAD3973A5317BAD3973AD3173AD3973A53173AD3973A53173AD396BA531
73AD396BA53173AD396BA53173AD398C7B319484318C7B319484398C7B319484318C7B31
9484398C7B318C84318C7B319484398C7B318C84318C7B318C84398C6B318C6B31000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000634A31634A31634A31634A31634A31634A31
634A31634A31634A31634A31634A31635A31635231637331636B31637331636B31636B31
636B31637331636B31636B31636B31636B315A6B31636B315273315273394A7331527331
5273315273394A73315273315273315273394A7331397331317331397339317331397331
397331397B39317331397331397331397B39317373317B73317B73397B73317B73317B73
317B7339847B317B7331847331847339847B3184733984AD3184AD3984B53184AD3984B5
3984B5398CB53184B5398CB5398CB5398CBD318CB5398CBD398CBD3973BD3173BD3973BD
3973BD397BC63173BD397BC6397BC6397BC6397BC6397BC6397BC63984CE317BC63984CE
3984CE3984CE3984CE3984CE3984CE3984D63984D63984D63984D6398CDE398CDE398CDE
398CDE398CE7398CDE398CE7398CE73994EF398CE739BDEF39BDEF39BDEF39BDEF39BDEF
39C6EF39C6EF39C6EF3942EF3942EF394AEF3942EF394AEF394AEF424AEF394AE7394AEF
394AE74252EF39ADE739ADE739ADE742ADE739ADDE39ADDE39ADDE39ADDE39A5D639ADD6
39A5D639ADD639A5D639A5D639A5CE39A5CE39A5CE39A5CE39A5C639A5CE3939C63939C6
39CEC639CEC639C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39C6B539C6B5
39C6B539C6B539BDB539BDB539BDB539BDB539BDAD3994AD3994AD3994AD39BD5A39C65A
39BD5A39C65A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39B55A39B55A
39B55A39B55A39B55239B55A39B55A39B55A31AD5239AD5239AD526BB5396BAD316BAD31
6BAD316BAD3973AD3173AD3173AD3173AD3973A53173AD3173AD3173AD3173A53173A531
73A53173A5316BA53184A53184A53184A5318C7B318C7B318C7B318C84318C7B318C7B31
8C7B318C84318C7B318C7B318C7B318C84318C7B318C6B318C6B318C6B318C6B318C6B31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000634A31634A31634A316B4A39634A31
635A31635231635A39635231635A39635231635A396352316373395A6B316373395A6B31
6373395A6B316373395A6B31636B395A6B31636B395A6B315273394A73315273394A7331
5273394A6B315273394A6B315273394A6B315273394A6B31397339317331397339317331
397B39317331397B39317331397B39317331397B3931736B397B73317373397B73317B73
397B7B317B7339847B317B7339847B317B7339847B317BAD3984B5317BAD3984B53184AD
398CB53184B5398CBD3184B5398CBD3184B5398CBD318CB53994BD3173BD397BBD3173BD
397BC63173BD397BC63173BD397BC6317BC6397BC6317BC63984CE317BC63984CE317BC6
3984CE317BC63984CE3184CE3984D63184CE398CD63184D6398CDE3184D6398CDE3184DE
398CE7318CDE3994E7318CE73994E7318CE73994EF39B5E739BDEF31B5E739BDEF39BDE7
39C6EF39BDEF4242EF3942EF4242EF3942E7424AEF3942E7424AEF3942E7424AEF394AE7
424AEF394AE74252EF39ADE742B5E739ADDE42ADE739ADDE42ADDE39A5D642ADDE39A5D6
42ADDE39A5D642ADD639A5CE39A5D639A5CE39A5CE39A5C639A5CE3939C63942C63939C6
39CEC631C6BD39CEC639C6BD39CEC639C6BD39CEBD39C6BD39C6BD31BDB539C6BD39BDB5
39C6BD31BDB539C6B539BDB539BDB531BDAD3994B5318CAD3994AD318CAD3994AD31BD5A
39C66331BD5A39C66331BD5A39BD5A31B55A39BD5A31B55A39BD5A31B55239BD5A31B552
39BD5A31B55239B55A31B55239B55A31AD5239B55A31AD5239B55A31AD526BB5396BAD31
7BAD3973A5317BAD3973A53173AD3973A53173AD3973A53173AD396BA53173AD396BA531
73AD396BA53184A5397BA53184A5397BA5318C84398C7B318C84398C7B318C84398C7B31
8C8431847B318C8439847B318C8431846B318C7339846B318C6B31846B318C6B31846B31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000635231635231635231635231
635231635231635231635231635231635231635231635231635231636B31636B31636B31
5A6B31636B315A6B31636B315A6B315A6B315A6B315A6B315A6B315273314A7331527331
4A73315273314A73315273314A73315273314A7331527339317331397331317331397339
317331397339397331397339317331397339397331397339317373397B73317B73397B73
317B73397B73317B73397B73317B7339847B31847339847B317BAD3984AD3184AD3984B5
3184B53984B53184B5398CB53184B5398CB5318CB5398CBD318CBD398CBD3973BD3973BD
3173BD3973BD3973BD397BC6397BC6397BC6397BC6397BC6397BC6397BC6397BC63984CE
397BC63984CE3984CE3984CE3984CE3984D63984D63984D63984D6398CDE398CDE398CDE
398CDE398CE7398CE7398CE7398CE73994E7398CE739BDEF39B5EF39BDEF39BDEF39BDEF
3939EF3942EF3942EF3942EF3942EF3942EF3942E7394AEF3942E7394AEF394AE7394AE7
394AE7394AE7394AE739ADE739ADE739ADE739ADDE39ADDE39ADDE39ADDE39ADD639ADD6
39A5D639ADD639A5D639A5D639A5CE39A5CE39A5CE39A5CE3942C63942C63939C63942C6
3939C639CEC639C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39C6B539C6BD39C6B539C6B5
39BDB539BDB539BDB539BDB539BDAD3994B53994AD3994AD3994AD3994AD398CAD39C65A
39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39BD5A39B55A39B55A39B55A39B55A
39B55A39B55A39B55A39B55A39AD5239B55A39AD5239AD5A39AD5239AD526BAD3173AD39
73AD3173AD3173AD3173AD3173A53173AD3173A53173AD316BA53173A5316BA53173A531
6BA53184A53184A53184A5317BA53184A5317BA5318C7B318C7B318C7B318C7B318C7B31
847B318C7B318C7B318C6B31846B318C6B318C6B318C6B31846B318C6B318C6B318C6B31
8C6B31000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006352316352315A5231
635A396352316352315A5231635A395A52316352315A5231635A395A5231636B315A6B31
6373395A6B31636B315A6331636B395A6B31636B395A6331636B395A63315273394A6B31
5273394A6B315273394A6B315273394A6B315273394A6B31527339316B31397339316B31
397339317331397339317331397B39317331397339317331397B7331736B397B73317373
397B7B317B73397B73317B7339847B317B7339847B317B733984AD317BAD3984B5317BAD
3984B53184B5398CB53184B5398CBD3184B5398CBD3184B53994BD318CB53973BD3173BD
397BC63173BD397BC63173BD397BC6317BC6397BC6317BC63984CE317BC63984CE317BC6
3984CE317BC63984CE317BCE3984D63184CE3984D63184D6398CDE3184D6398CDE3184D6
398CE7398CDE398CE7318CDE3994EF398CE73994EF318CE739BDEF39B5EF39BDEF3939E7
4242EF3939E73942EF3939E74242EF3942E7394AEF3942E7424AEF3942E7424AE73942E7
424AEF394AE74252E7394ADE42B5E739ADDE39ADE739A5DE42ADDE39ADD639ADDE39A5D6
42ADD639A5CE39ADD639A5CE42ADD639A5CE39A5CE3939C64242CE3939C63942C63939BD
3942C639C6BD39CEC631C6BD39CEC639C6BD39C6BD31BDB539C6BD39BDB539C6BD31BDB5
39C6BD39BDB539BDB531BDAD3994B53994AD3994B5318CAD3994AD398CA53994AD318CA5
39C66339BD5A39BD5A31B55A39BD5A31B55A39BD5A31B55239BD5A31B55A39B55A31B552
39B55A31B55239B55A31AD5239B55A31AD5239B55A31AD5239B55A31AD5239AD4273A531
7BAD3973A53173AD396BA53173AD396BA53173A5396BA53173AD396BA53173A5397B9C31
84A5397BA53184A5317B9C3184A5397B9C3184A531847B318C8439847B318C7B31847B31
8C8439847B318C6B318463318C6B39846B318C6B318463318C6B39846B318C6B31846331
8C6B39000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000635231
6352316352315A52316352316352316352315A52316352316352316352315A5231636B31
5A6B31636B315A6B315A6B315A6B315A6B315A6B315A6B315A6B315A6B315A63314A7331
4A73314A73394A6B314A73314A6B314A73394A6B314A6B314A6B31527339316B31397331
317331397339317331397331397331397339317331397331397331397373317373317B73
317B73397B73317B73317B73317B7339847B317B73398473318473397BAD317BAD3984AD
3184B53984B53184B53984B53984B5398CB5318CB5398CBD398CBD398CBD318CBD3973BD
3973BD397BBD3173BD397BC6397BC6397BC6317BC6397BC6397BC6397BC6397BC6397BCE
397BCE3984CE397BCE3984CE3984CE3984D63984CE3984D63984D6398CDE3984D6398CDE
398CDE398CE7398CDE398CE7398CE73994E7398CE73994EF39B5EF39BDEF3939EF3939EF
3939EF3942EF3939E73942EF3942E73942EF3942E73942E73942E7394AE73942E7394AE7
394AE7424AE7394AE7394AE739ADE739ADE739ADDE39ADDE39ADDE39ADDE39ADD639ADD6
39ADD639ADD639A5CE39A5CE39A5CE39A5CE3942C63942CE3942C63942C63939C63939C6
3939C63939C639C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39C6BD39BDB539BDB5
39BDB539BDB539BDB53994B53994AD3994B53994AD3994AD3994AD3994AD398CA5398CA5
398CA539BD5A39BD5A39BD5A39BD5A39BD5A39B55A39B55A39B55A39B55A39B55239B55A
39B55A39B55A31AD5239B55A39AD5239B55A31AD5239AD5239AD5239AD4231A54239AD42
73A53173AD3973A53173A53173A53173A5396BA5316BA5316BA5316BA5317BA5317BA531
7BA53184A5317B9C317BA5317BA5317BA5317B9C317BA531847B318C7B31847B31847B31
846B318C6B31846B31846B31846B318C6B31846B31846B31846B318C6B31846B31846B31
846B318C6B31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
6352315A52316352395A52316352315A52316352395A52316352395A52316352395A5231
636B395A6B31636B395A63315A6B395A63315A6B395A63315A6B395263315A6B394A6B31
4A73394A6B315273394A6B315273394A6B315273394A6B315273394A6B31397339316B31
397339316B31397339316B31397339316B3139733931733139733931736B397B7331736B
397B73317373397B7B317B73397B7B317B7339847B317B7339847B317BA53984AD317BAD
3984B5317BAD3984B53184B5398CB53184B5398CBD3184B5398CBD318CB53994BD3173BD
3973BD3173BD397BC63173BD397BC63173BD397BC6317BC6397BC6317BC63984CE317BC6
3984CE317BC63984CE317BCE3984D63184CE398CD63184D6398CD63184D6398CDE3184D6
398CDE3184DE398CE7318CDE3994E7318CE73994EF318CE77339EF3931E73939EF3939E7
3942EF3939E73942EF3939E73942EF3942E74242EF3942E7394AE73942E7424AE73942E7
424AE7394ADE424AE7394ADE42B5E739ADDE42B5E739ADDE42ADDE39A5D642ADDE39A5D6
39ADD639A5CE39ADD639A5CE39ADCE3942C63942CE3939C63942CE3939C63942C63939BD
3942C63939BD39CEC639C6BD39C6BD31C6B539C6BD39BDB539C6BD31BDB539C6BD39BDB5
39C6B531BDAD3994B53194AD3994B5318CAD3994B5318CAD3994AD318CA53994AD318CA5
3994A531B55A39BD5A31B55A39BD5A31B55239BD5A31B55239BD5A31B55239B55A31B552
39B55A31AD5239B55A31AD5239B55A31AD5239AD5A31AD5239AD4231A53939AD4231A542
73AD396BA53173AD396BA53173A5396BA53173A5396B9C316BA5397B9C3184A5397B9C31
84A5397B9C3184A5397B9C317BA5397B9C317BA5397B9C317BA531847B318C6B39846B31
8C6B318463318C6B39846B318C6B318463318C6B31846B318C6B318463318C6B31846B31
8C6B31846331000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000006352315A52316352315A52315A52315A52315A52315A52315A52315A52315A5231
5A52315A6B315A6B315A6B315A6B315A6B315A63315A6B315A63315A6B315A63315A6331
4A6B314A6B314A6B314A6B314A6B314A6B314A6B314A6B314A6B314A6B314A6B31396B39
316B31396B31316B3139733931733139733939733139733931733139733931736B397373
317373397B73317B73397B73317B73397B73317B73397B73317B7339847B317BAD3984AD
317BAD3984B53184B53984B53184B5398CB53184B5398CBD318CB5398CBD398CBD3973BD
3173BD3973BD3973BD397BC63173BD397BC6397BC6397BC6397BC6397BC6397BC63984CE
397BC63984CE3984CE3984CE3984CE3984CE3984CE3984D63984D63984D63984D6398CDE
398CDE398CDE398CDE398CE7398CE7398CE76B39E76B39EF6B39EF3939EF3939EF3939EF
3939E73939E73939E73942E73942E73942E73942E73942E73942E73942E73942E7394AE7
394ADE394AE7394ADE394AE7394ADE39ADDE39ADDE39ADDE39ADD639ADD639ADD639ADD6
39A5D639ADD639A5CE3942CE3942CE3942CE3942C63942CE3942C63942C63939C63939C6
3939BD3939C63939BD39C6BD39C6BD39C6BD39C6BD39C6BD39BDB539C6B539BDB539BDB5
39BDB53994B53994B53994B53994AD3994AD3994AD3994AD398CA53994AD398CA5398CA5
398CA5398CA539BD5A39BD5A39B55A39B55A39B55A39B55A39B55A39B55A39B55A39B55A
39AD5239B55A39AD5239AD5A39AD5239AD5239AD5239AD4231A54239AD4239A54239A542
31A54273A5316BA53173A5316BA5316BA5316BA53184A5317B9C317BA5317B9C317BA531
7B9C317BA5317B9C317BA5317B9C317B9C317B9C317B9C31848C31846B31846B31846B31
846B31846B31846B31846B31846B31846B31846B31846B31846B31846B31846B31846B31
846B31845A31845A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000006352315A4A316352395A52316352315A4A316352395A52315A52315A4A31
6352395A52315A6B315A63315A6B395A63315A6B315263315A6B395263315A6B39526331
5A6B394A6B314A6B39426B314A73394A6B314A6B394A6B314A73394A6B314A6B394A6B31
397339316B31397339316B31397339316B31397339316B3139733931733139737331736B
397B7331736B397B7331736B397B7B317B73397B7B317B7339847B317B73397BAD317BA5
3984B5317BAD3984B53184AD398CB53184B5398CB53184B5398CBD318CB5398CBD318CB5
3973BD3173BD3973BD3173BD397BC63173BD397BC63173BD397BC6317BC6397BCE317BC6
3984CE317BC63984CE317BC63984CE3184CE3984D63184CE398CD63984D6398CDE3184D6
398CDE3184DE398CE7318CDE6B39E76B31E76B39EF6B31E77339EF3939EF3939EF3139E7
3942EF3939E73942E73939E73942E73939E73942E73942DE424AE73942DE394AE73942DE
424AE73942DE394AE7394ADE4252E739ADDE39B5DE39ADD642ADDE39ADD639ADD639A5CE
42ADD639A5CE394AD63942CE424ACE3942C63942CE3939C64242CE3939C63942C63939BD
3942C63939BD3939BD31C6BD39C6BD39C6BD39C6BD31BDB539C6BD39BDB539C6B531BDB5
39C6B53994AD3994B5318CAD3994B5398CAD3994AD318CA53994AD398CA53994AD318CA5
3994A5318C9C39BD5A31B55239BD5A31B55A39B55A31B55239B55A31B55239B55A31AD52
39B55A31AD5239AD5A31AD5239AD5A31A55239AD4231A53939AD4231A54239AD4231A539
39A54231A53973A5396B9C3173A5396B9C3184A5397B9C3184A5397B9C317BA5317B9C31
7BA5397B9C317BA531739C317BA539739C317B9C317B8C31849439848C31846B31846331
846B39846331846B317B6331846B39846331846B31846331846B39846331846B31845231
8C5A39845231845A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005A52315A52315A52315A52315A52315A52315A52315A52315A5231
5A52315A52315A52315A6B315A6B315A6B315A63315A63315A63315A6B315263315A6331
5263315A63314A6B314A6B314A6B314A6B394A6B314A6B314A6B314A6B394A6B314A6B31
316B31396B39316B31396B31316B31397339316B31397331397331397339316B31317373
317373397373317373317B73317B73397B73317B73397B73317B73397B7B317B73397BAD
397BAD3984AD317BAD3984B53184B53984B53184B5398CB5398CB5398CBD318CB5398CBD
3973BD3973BD3173BD3973BD3973BD397BC63173BD397BC6397BC6397BC6317BC6397BC6
397BC63984CE397BC63984CE3984CE3984CE3984CE3984D63984D6398CD63984D6398CDE
398CDE398CDE6331DE6B39DE6B39E76B39E76B39E76B39E76B39E76B39EF3939E73939EF
3939E73939EF3939E73942E73942E73942E73942DE3942E73942E73942E73942DE394ADE
394ADE394ADE394ADE394ADE394ADE424ADE39ADDE39ADDE39ADD639ADD639ADD639ADD6
39ADD6394AD63942CE3942CE3942CE3942CE3942C63942C63942C63942C63939C63939C6
3939BD3939C63939BD3939BD39C6BD39C6BD39BDB539C6BD39BDB539C6B539BDB539BDB5
3994B53994B53994AD3994AD3994AD3994AD398CAD3994AD398CA53994AD398CA5398CA5
398CA5398CA5398C9C39B55A39B55A39B55A39B55239B55A39B55A39B55A31AD5239AD52
39AD5239AD5A31AD5239AD5239AD5239AD4231A54239A54239A54239A54231A54239A542
39A54239A5426B9C316BA5316B9C317BA5397B9C317B9C317B9C317BA5317B9C317B9C31
7B9C317B9C31739C317B9C31739C31849431848C31848C31848C31849431846331846B31
846B31846B31846331846B31846B31846B31846331846B31846B31845A31845231845A31
845A31845A31845A31845A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005A52315A4A315A52395A4A315A52315A4A315A52395A4A31
5A52395A4A315A52395A4A315A6B315263315A6B395263315A6B395263315A6B39526331
5A63395263315A63394263314A6B394263314A6B39426B314A6B394263314A6B39426331
4A6B39316331396B39316B31396B39316B31397339316B31397339316B31397339316B6B
39737331736B397B7331736B397B73317373397B7B317B73397B7B317B7339847B317BA5
397BAD317BAD3984B5317BAD3984B53184AD398CB53184B5398CBD3184B5398CBD318CB5
398CBD3173B53973BD3173BD397BC63173BD397BC63173BD397BC6317BC6397BCE317BC6
3984CE317BC63984CE317BC63984CE317BCE3984D63184CE398CD63184D6398CDE6331D6
6B39DE6331D66B39E76331DE6B39E76331DE6B39E76B31E76B39EF6B31E73939EF3939E7
3939EF3939E73942EF3939E73942E73939DE3942E73942DE3942E73942DE424AE73942DE
394AE73942DE424ADE394ADE424ADE394ADE42B5DE39ADD639ADDE39ADD639ADD639A5CE
424AD63942CE424AD63942CE3942CE3942C63942CE3939C63942CE3939C63942C63939BD
3942C63939BD3939C63939BD39C6BD31BDB539C6BD39BDB539C6BD31BDB539C6B53194B5
3994B53194AD3994B5318CAD3994AD318CA53994AD318CA53994AD318CA53994A5318CA5
398CA5318C9C398CA531849C39BD5A31B55239B55A31AD5239B55A31AD5239B55A31AD52
39AD5A31AD5239AD5A31A55239AD4231A53939AD4231A53939AD4231A53939A542319C39
39A542319C3939A5427B9C317BA5397B9C317BA5397B9C317BA539739C317B9C39739C31
7B9C39739C317B9C39739C318494317B8C318494397B8C318494317B8C318494397B6331
846B317B6331846B317B6331846B317B6331846B317B5231845A317B5231845A317B5231
845A317B5231845A31845231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005A52315A52315A52315A52315A52315A52315A5231
5A52315A52315A5A315A5A315A5A31526B31526B31526B31526B31526B31526B31526B31
4A6B31526B314A6B31396B31396B31396B314A6B314A6B314A6B314A6B314A6B314A6B31
4A6B31396B31316B31396B39316B31396B31316B31396B39316B31397339396B31397339
316B6B31737331736B397373317373397B73317B73397B73317B73397B73317B73397B7B
317BAD397BAD317BAD3984AD3184AD3984B53184B53984B53184B5398CB5398CB5398CBD
318CBD3973BD3973BD3973BD3973BD3973C63973BD397BC6317BC6397BC6397BC6397BC6
397BC6397BCE397BCE3984CE3984CE3984CE3984CE3984D63984D66339D66331D66B39DE
6331D66B39DE6B39DE6B39DE6B39DE6B39E76B39E76B39E76B39E76B39EF6B39EF3939EF
3939E73939EF3939E73942E73939E73942E73942E73942E73942DE3942E73942DE3942DE
3942DE394ADE394ADE394ADE394ADE394ADE394ADE39ADDE39ADD639ADD639ADD6394AD6
3942CE394ACE3942CE3942CE3942CE3942CE3942C63942C63942C63942C63939C63942C6
3939BD3939BD3939BD3939BD3939BD39C6BD39BDB539C6BD39BDB539BDB53994B53994B5
3994AD3994B53994AD3994AD398CAD3994AD398CA53994AD398CA5398CA5398CA5398CA5
398C9C398C9C398C9C398C9C39B55A39B55A39B55239B55A39AD5239AD5239AD5239AD52
39AD5239AD5239AD5239AD4231A54239A54231A54239A54231A54239A54239A54239A542
319C42399C424A9C314A9C317B9C317B9C317B9C317B9C31739C317B9C31739C317B9C31
739C31739C317B8C31848C317B8C31848C317B8C31848C317B8C31848C317B8C31846B31
7B6331846B317B6331846B317B5231845A31845A31845A317B5231845A31845A31845A31
845231845A31845A31845A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005A5A315252315A5A39525A315A5A31525231
5A5A39525A315A5A315252315A5A39525A315273314A6B315273394A6B31526B314A6331
526B394A6B31526B394A6331396B39396331396B39316331396B39396331396B39396331
396B39396331396B39316331396B39316B31396B39316B31397339316B31397339316B31
397339316B6B397373316B6B397B7331736B397B7331736B397B7B317B73397B7B317B73
397BAD317BA5397BAD317BAD3984B5317BAD3984B53184AD398CBD3184B5398CBD3184B5
398CBD318CB53973BD3173BD397BC63173BD397BC63173BD397BC6317BC6397BC6317BC6
3984CE317BC63984CE317BC63984CE3184CE3984CE6331CE6B39D66331D66B39D66331D6
6B39DE6331D66B39DE6331DE6B39E76B31DE6B39E76331E77339EF6B39E76B39EF3131E7
3942EF3939E73942EF3139E73942EF3939E73942E73939DE4242E73942DE3942E73942DE
424AE73942DE394ADE3942D6424ADE394AD6394ADE39ADD642B5DE39ADD6394AD63942CE
424AD63942CE394ACE3942C6424ACE3942C63942CE3939C64242CE3939C63942C63939BD
3942C63939BD3939BD3131B53939BD39BDB539C6BD31BDB539C6BD3994B53994B5318CAD
3994B53994AD3994AD318CAD3994AD398CA53994AD318CA53994AD318CA5398CA5318C9C
398CA5318C9C398C9C318494398C9C31AD5239B55A31AD5239B55A31AD5239AD5A31A552
39AD5A31A55239AD4231A53939AD4231A54239A542319C3939A542319C3939A542319C39
39A542429C314A9C39429C314A9C39739C317B9C317394317B9C397394317B9C31739431
7B9C397B8C31848C317B8C318494397B8C31848C317B8C318494397B8C31848C317B8C31
846B397B6331845A317B5231845A397B5231845A317B5231845A317B5231845A317B5231
845A317B5231845A317B5229845A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005A5A315A5A315A5A31525A315A5A31
5A5A315A5A31525A315A5A31525A315A5A31525A31526B31526B31526B314A6B31526B31
4A6B31526B314A63314A6B314A6331396B31396331396331396331396B39396331396331
396331396B39396331316B6B316B6B316B6B316B6B316B6B316B31396B39316B31396B31
396B3139736B316B6B31736B317373397373317373317373317B73397B73317B73397B73
317B73397BAD317BAD397BAD397BAD3984B53184AD3984B53984B5398CB53184B5398CB5
398CBD398CBD3173BD3973BD3973BD3973BD3173BD397BC6397BC6397BC6317BC6397BC6
397BC6397BCE317BC63984CE3984CE6339CE6331CE6339CE6331D66339D66331D66339D6
6339D66B39DE6331DE6B39DE6B39DE6B39E76B39DE6B39E76B39E76B39EF6B39E73939EF
3939EF3939EF3939E73939E73939E73942E73939E73942E73942E73942E73942DE3942DE
3942DE394ADE3942DE394ADE394ADE394ADE394AD639ADD639ADD6394AD6394AD6394AD6
394ACE394ACE3942CE3942CE3942CE3942CE3942C63942C63942C63942C63939BD3939C6
3939BD3939BD3939BD3939BD3939BD3939BD39BDB539BDB53994B53994B53994B53994B5
3994AD3994AD398CAD3994AD398CAD3994AD398CA5398CA5398CA5398CA5398C9C398CA5
398C9C398C9C39849C398C9C398494398C9C31AD5239AD5239AD5239AD5A31AD5239AD52
39AD5239AD4231A54239A54239A54239A54231A54239A54239A54239A542319C39399C42
4A9C314A9C31429C314A9C31429C317B9C31739431739C31739C31739C317394317B8C31
7B8C317B8C317B8C317B8C317B8C317B8C317B8C317B8C317B8C317B8C317B8C317B8C31
7B7331845A317B52317B5A317B5231845A317B52317B5A317B5A31845A317B5231845A31
7B5A31845A317B5231845A31845A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005A5A315252315A5A39525231
5A5A315252315A5A395252315A5A315252315A5A39525231526B314A6331526B394A6331
526B394A6331526B394A63314A6B394A6331396B39396331396B39316331396B39396331
396B39396331396B39396331396B6B316363396B6B316363396B6B316363396B6B316B6B
396B6B316B31397373316B6B397373316B6B39737331736B397B73317373397B7B317B73
397B7B3173A5397BAD317BAD3984AD317BAD3984B5317BAD3984B53184B5398CBD3184B5
398CBD318CB53973BD3173BD3973BD3173BD397BC63173BD397BC63173BD397BC6317BC6
6339C65A31C66339CE5A31C66339CE5A31C66339CE6331CE6339D66331CE6B39D66331D6
6B39DE6331D66B39DE6331DE6B39E76331DE6B39E76B31DE6B39E76B31E77339EF6B39E7
3939EF3939E73942EF3939E73942EF3939E73942E73939DE3942E73942DE394AE73942DE
394ADE3942DE424ADE3942D6394ADE3942D6424ADE394AD6424AD6394ACE424AD63942CE
394AD63942CE424ACE3942CE394ACE3942C63942CE3939C63942C63939BD3942C63939BD
3942C63939BD3939BD3939BD3939BD3931B539C6BD3194B53994B53194AD3994B53194AD
3994B5318CAD3994AD318CA53994AD318CA53994AD318CA5398CA5318C9C398CA5318C9C
398CA531849C398C9C318494398C9C31849439849431AD5239AD5A31A55239AD5A31A552
39AD4231A53939AD4231A53939A542319C3939A542319C3939A542319C3939A542429C31
4A9C394294314A9C394294314A9C394294317B9C39739431739C39739431848C397B8431
7B8C317B8431848C397B84317B8C317B8431848C397B8C317B8C317B84317B73317B6B31
8473317B5231845A317B5231845A317B5231845A317B5231845A317B5231845A317B5231
845A317B5231845A317B5231844A317B4A31000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005A5A31525A315A5A31
525231525A31525A31525A31525231525A31525231525A31525231526B314A6B314A6B31
4A63314A6B314A63314A6B314A63314A63314A6331396331396331396331396331396331
39633139633139633139633131636331636B316363316B6B316363316B6B316B6B316B6B
316B6B316B6B315A94395A9C315A9C315A9C315A9C397373317373397373317373397B73
317B73397B73317BA5397BAD317BAD397BAD317BAD3984B53184AD3984B53184B5398CB5
3184B5398CBD398CBD3973BD3173BD3973BD3973BD397BC65A31BD5A31C65A31C65A39C6
5A31C66331C65A31C66339CE6331CE6339CE6331CE6339CE6331CE6339D66331D66339D6
6331D66B39DE6339DE6B39DE6339DE6B39DE6B39DE6B39E76B39E76B39E76B39E76B39EF
3939EF3939EF3939E73939EF3939E73942E73942E73942E73942E73942E73942DE3942DE
3942DE394ADE3942DE394ADE394AD6394ADE394AD6B54AD6394AD6394AD6394AD6394AD6
3942CE3942CE3942CE3942CE3942C63942CE3942C63942C63939C63942C63939C63939C6
3939BD3939BD3939BD3939BD3939B53939BD6B31B53994B53994B53994B53994AD3994B5
3994AD3994AD398CAD3994AD398CA5398CA5398CA5398CA5398CA5398CA5398C9C398C9C
398C9C398C9C39849C39849C39849439849439849439AD5239AD5239AD5239A55239AD42
31A54239A54239A54239A54231A54239A542319C42399C42319C42399C42429C314A9C31
429C314A9C31429C314A9C31429431429C314294317394317B8C317B8C317B8C317B8C31
7B8C317B8C317B8C317B8C317B8C317B8C317B8C317B8C317B8C317B73317B6B317B7331
7B6B317B73317B52317B5A317B52317B5A317B52317B5A317B52317B5A317B52317B5A31
7B52317B5A317B4A31844A317B4A31844A31000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005A5A31525231
5A5A39525231525A315252315A5A39525231525A315252315A5A39525231526B314A6331
526B394A63314A6B314263314A6B394A63314A6339426331396339316331396339316331
396B3939633139633931633139633931636339636B316363396B6B316363396B6B316363
396B6B316B6B396B6B315294395A9C315A9C395A9C315A9C3963A5315A9C3963A5315A9C
3963AD3163A5397BAD3173A5397BAD317BAD3984AD317BAD3984B53184AD3984B53184B5
398CBD3184B5398CBD5231B55A39BD5231BD5A39C65231BD5A39C65A31BD5A39C65A31BD
6339CE5A31C66339CE5A31C66339CE5A31C66339CE5A31CE6B39D66331CE6B39D66331CE
6B39DE6331D66B39DE6331D66B39DE6339DE6B39E76331DE6B39E76B39E76B39EF6B31E7
7339EF3939E73939EF3139E73942EF3939E73942E73139DE4242E73942DE3942E73942DE
424AE73942DE394ADE3942D6424ADEAD4AD6B54ADEAD42D6B54AD63942CE394AD63942CE
424AD63942CE3942CE3942C64242CE3942C63942CE3939C64242C63939BD3942C63939BD
3942C63939BD3939BD3131B53939BD6B31B56B39B53194AD3994B53994AD3994B5318CAD
3994B5398CAD3994AD318CA53994AD318CA5398CA5318C9C3994A5318C9C398CA531849C
398C9C318494398C9C318494398C9C318494398494317B8C39AD5A31A55239AD4231A539
39AD4231A54239A542319C3939A542319C39399C42319C39399C42429C314A9C39429431
4A9C394294314A9C314294314A9C394294314A94315A94317B8C397B84317B8C31738431
7B8C397B84317B8C317384317B8C397384317B8C31736B317B73397B6B317B7331736B31
7B73317B6B317B5A317B5229845A317B52317B5A317B5231845A317B52317B5A317B4229
844A317B4A31844A317B4229844A317B4A31844A31000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
525A31525A31525231525A31525231525A31525231525231525231525A315252314A6B31
4A63314A6B314A63314A63314A63314A63314A63314A6331396331396331396331396331
396331396339396331396331396331396339316363316363316363396B6B316363316B6B
316B6B396B6B316B6B315A94315A94395A9C315A9C315A9C315A9C3963A5315A9C3163A5
3163A53963A53131A53931AD3131AD3931AD3131AD3931AD3931AD3939B53131B53931B5
3931B53939BD3131B53931BD5A31BD5A39BD5A31BD5A31BD5A31BD5A39C65A31BD5A31C6
5A31C66339C65A31C66331CE6331CE6339CE6331CE6331CE6331CE6339D66331CE6339D6
6339D66B39D66331D66B39DE6B39DE6B39DE6B39DE6B39E76B39E76B39E76B39E76B39E7
6B39EF3939EF3939E73939EF3939E73942E73939E73942E73942E73942E73942DE3942DE
3942DE394ADE3942DE394ADEAD4ADEB54ADEAD42D6AD4AD6AD4AD6394AD63942CE3942CE
3942CE3942CE3942CE3942CE3942C63942CE3939C63942C63939C63942C63939BD3939BD
3939BD3939BD3931BD3939BD6B39B56B39BD6B31B56B39B53994B53994B53994AD3994AD
3994AD3994AD398CA5398CA5398CA5398CA5398CA5398CA5398CA5398CA5398C9C398C9C
39849C398C9C31849439849439849439849431848C39848C39848C39A54231A54239A542
39A54239A542319C42399C42399C42399C42319C39399C42429C314A9C39429431429C31
4294314A9C314294314294314294316394315A94316394316394317B8C317384317B8431
7B84317B8C317384317B8C317B84317B7331736B317B6B317B6B317B73317B6B317B6B31
7B6B317B73317B6B317B52317B52317B5A317B52317B52317B52317B4A317B4A317B4A31
7B4A317B4A317B4A317B4A317B4A31844A317B4A31000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000525231525A31525231525A314A5231525A314A5231525A314A5231525A394A5231
4A6B314A63314A6B394263314A63394263314A63394263314A6339315A31396339315A31
396339315A31396339315A31396339315A3139636331636339636B316363396B6B316363
396B6B316363396B6B316363395A9C315294395A9C315A9C395A9C315A9C3963A5315A9C
3963A5315AA53963AD3131A53939AD3131A53939AD3131AD3939B53131AD3939B53131B5
3939B53131B53939BD3131B55A39BD5231BD5A39C65231BD5A39C65A31BD5A39C65A31BD
6339C65A31C66339CE5A31C66339CE5A31C66339CE6331CE6339D66331CE6B39D66331CE
6B39D66331D66B39DE6331D66B39DE6331DE6B39E76339DE6B39E76331DE6B39E76B39E7
7339EF6B39E73942EF3939E73942EF3939E73942EF3939E73942E73942DE3942E73942DE
394AE73942DE394ADEAD42DEB54ADEAD42D6B54ADEAD42D6B54AD6AD42CE394AD63942CE
394AD63942CE3942CE3942C63942CE3939C63942CE3939C63942C63939BD3942C63939BD
3939BD3131BD3939BD6B31B57339BD6B31B56B39B56B31AD6B39B5318CAD3994B5318CAD
3994AD318CA53994AD318CA53994A5318C9C398CA5318C9C398CA531849C398C9C318494
398C9C318494398C9C318494398494317B8C398494317B8C39A58C319C8439A542319C39
39A542319C3939A542319C39399C423194394A9C394294314A9C394294314A9C39429431
4A9C394294314A94394294316394395A8C316394395A94316394317384317B8C31738431
7B8C317384317B8C39736B317B7331736B317B7331736B317B7331736B317B7331736B31
7B7331736B317B73317352317B5A317352317B4A317B42317B4A317B42317B4A317B4231
7B4A317B4231844A317B4231844A317B4229844A317B4231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000525231525A31525231525231525231525231525231525231525231525231
4A63314A63314A63314A63314A63314A63314263314A6331426331426331395A31396331
395A31396331396331396331395A31396331395A3131636331636331636331636331636B
31636331636B31636B316B6B315294315A94315A94395A9C315A9C315A9C315A9C395AA5
315A9C3963A53163A53931A53131A53931AD3131AD3931AD3131AD3931B53931B53931B5
3131B53931B53931B53931BD3931BD5A31BD5A31BD5A39BD5A31BD5A31C65A31C65A39C6
5A31C65A39C65A31C66339CE5A31C66331CE6331CE6339CE6331CE6339CE6331CE6339D6
6331D66339D66339D66B39DE6339DE6B39DE6B39DE6B39E76B39DE6B39E76B39E76B39E7
6B39E76B39EF3939EF3939EF3939E73942E73939E73942E73942E73942E73942E73942E7
3942DEAD42DEAD42DEAD42DEAD42DEAD42DEAD42D6AD4AD6AD42D6AD42D63942CE3942CE
3942CE3942CE3942CE3942CE3939C63942C63939C63939C63939C63939C63939BD3939BD
3939BD3939BD6B31BD6B39BD6B31B56B39B56B31B56B39B56B31AD3994AD3994AD3994AD
398CAD3994AD398CA5398CA5398CA5398CA5398C9C398CA5398C9C398C9C39849C398C9C
398494398494398494398494398494398494397B8C39A58C319C8C399C8C319C8439A542
319C42399C42319C42399C42319C394A9C314294314A9C31429431429431429431429431
4294314294314294316394315A94316394315A94316394315A94315A9431738431738431
738431736B31736B31736B31736B31736B31736B317B6B31736B317B6B31736B317B6B31
736B317B6B317B6B317B73317B4A317B4A317B4A317B4A317B4A317B4A317B4A317B4A31
7B4A317B4A317B4A317B4A317B4A317B4A317B4A317B4A31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A5231525A394A5231525A314A5231525A394A52315252314A5231
525A394A63314A6B314263314A63394263314A6331425A314A6339425A314A6339315A31
396339315A31396339315A31396339315A31396339315A31396363315A6339636B316363
396B6B316363396B6B316363396B6B315294395A94315294395A9C315A9C395A9C315A9C
3963A5315A9C3963A5315A9C3939AD3131A53939AD3131A53939B53131AD3939B53131AD
3939B53131B53939BD3131B53939BD5231B55A39BD5231BD5A39C65A31BD5A39C65A31BD
6339C65A31C66339C65A31C66339CE5A31C66339CE5A31C66339D66331CE6339D66331CE
6B39D66331D66B39DE6331D66B39DE6339DE6B39DE6331DE6B39E76B39DE6B39E76B39E7
7339EF6B39E73939EF3139E73942EF3939E73942EF3939E73942E73942E73942E73942DE
AD4AE7AD42DEAD42DEA542D6AD4ADEAD42D6AD42D6AD42D6B54AD6AD42CEAD42D63942CE
4242D63942CE3942CE3939C63942CE3939C63942C63939BD3942C63939BD3939C63131BD
3939C66B31BD6B39BD6B31B57339BD6B31B56B39B56331AD6B39B56331AD3994AD318CA5
3994AD318CA53994A5318CA53994A5318C9C398CA531849C398CA531849C398C9C318494
398C9C318494398494317B8C398494317B8C39A58C319C8439A58C319C84399C8C319484
39A542319C39399C423194394A9C394294314A9C394294314A9C39429431429439428C31
4A94395A8C316394315A8C316394395A8C316394315A8C316394395A8C31639431738431
7B7339736B31736B31736B317B7339736B317B6B31736B317B7331736B317B7331736B31
7B7331736B317B5A317352297B5A317342317B4A317342297B4A317B42317B4A31734229
7B4A317B42317B4A317342297B4A317B42317B4A317B4229844A31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000525A31525A314A5A31525A314A5A31525A314A5A314A5A31
4A5A31525A31426331426331426331426331426331426331426331426331425A31425A31
395A31396331315A31395A31395A31396339315A31395A31315A63316363315A63316363
31636339636B31636331636B31636B396B6B315294315A94315A94395A9C315A9C315A9C
315A9C3963A5315A9C3963A53131A53931A53131A53931AD3931AD3931AD3131AD3931B5
3931B53939B53131B53931B53931B55A39BD5231BD5A31BD5A31BD5A39C65A31BD5A31C6
5A31C65A39C65A31C65A31C65A31C66339CE5A31C66331CE6331CE6339CE6331CE6339D6
6339D66339D66331D66339D66339DE6B39DE6339DE6B39DE6B39DE6B39E76B39E76B39E7
6B39E76B39EF6B39E73939EF3939EF3942EF3939E73942E73942E73942E73942E7AD42E7
AD42DEAD42DEAD42DEAD42DEAD42DEAD42DEAD42D6AD42D6AD42D6AD42D6AD42CE3942CE
3942CE3942CE3939C63942CE3939C63942C63939C63939C63939C63939C63939BD3939BD
6B39BD6B39BD6B31B56B39B56B39B56B39B56B31AD6B39AD6B31AD6B39AD6331AD398CAD
398CA5398CA5398CA5398CA5398CA5398CA5398C9C398C9C398C9C398C9C318494398494
39849439849431848C39848C39848C39A58C319C8C399C8C399C84399C8C319484399484
399484399C423194394294314294314A9439429431429431429431429431428C31429431
5A8C315A94315A8C315A8C315A8C315A94315A8C315A8C315A8C315A9431738C31738C31
736B31736B31736B31736B31736B31736B31736B31736B31736B31736B31736B31735A31
735A317B5A317352317B5A317B5A317B4A317342317B4A317B4A317B4A317B4A317B4A31
7B4A317B4A317B42317B4A317B4A317B4A317B4A317B42317B4231000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A5A31525A314A5A31525A314A5231525A314A5A31
525A314A5231525A39426331426331395A31426339395A31426339395A31426339395A31
316342315A31396339315A31396339315A31396339315A31395A39315A5A396363315A63
39636B315A6339636B31636339636B316363395A94315294395A9C315294395A9C315A9C
395A9C315A9C3963A5315A9C3963A53131A53939AD3131A53939AD3131AD3939B53131AD
3939B53131AD3939B53131B53939BD3131B55A39BD5231BD5A39BD5231BD5A39C65A31BD
5A39C65A31C66339CE5A31C66339CE5A31C66339CE5A31C66339CE5A31CE6339D66331CE
6339D66331D66B39DE6331D66B39DE6331D66B39DE6331DE6B39E76339DE6B39E76B39E7
6B39EF6B39E77339EF3939E73942EF3939E73942EF3939E73942E7A539E7AD42E7A539DE
AD42E7A539DEAD42DEA539D6AD42DEA539D6AD42DEA539D6AD42D6A539CEAD42D6AD42CE
3942CE3939C63942CE3939C63942CE3939C63942C63939BD3939C63931BD3939C66B31BD
7339BD6B31B56B39BD6B31B56B39B56331AD6B39B56331AD6B39AD6331A56B39AD318CA5
3994AD318CA53994A5318C9C398CA531849C398CA531849C398C9C318494398C9C318494
398494317B8C398494317B8C39A58C319C84399C8C319484399C8C319484399C8431947B
399484318C7B4A9C394294314A94394294314A9439428C31429439428C314294395A8C31
6394395A8C316394395A8C315A94315A8C315A94395A8C31738C316B8431738C316B8431
738C316B6B31736B31736B31736B31736B31736B31736B31736B317352317B5A31735231
7B5A317352317B5A317352317B5A317352297B4A317342317B4A317342297B4A31734231
7B4A317342297B4A317342317B42317339297B42317B39317B42317B3929000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004A5A314A5A314A5A314A5A314A5A314A5A31
4A5A314A5A314A5A314A5A31426331426331426331426331395A31425A31395A31425A31
395A31315A42315A39315A42315A39395A31315A31395A31315A31395A31315A63315A63
315A6331636331636331636331636331636B316363395294315294315A94315A94395A9C
315A9C395A9C315A9C395AA5315AA53931A53131A53931AD3131AD3931AD3131AD3931B5
3131AD3931B53931B53931B53131B53931BD5231BD5A39BD5231BD5A31BD5A31BD5A39C6
5A31C65A31C65A31C65A39C65A31C66339CE5A31CE6339CE6331CE6339CE6331CE6339D6
6331CE6339D66339D66339D66339D66B39DE6339DE6B39DE6B39DE6B39E76B39E76B39E7
6B39E76B39E76B39E76B39EF3939EF3939EF3939E73942E7A539E7AD42E7A542E7AD42E7
A539DEAD42DEAD42DEAD42DEA539D6AD42D6AD42D6AD42D6AD42D6AD42D6AD42CEAD42CE
3942CE3942CE3939C63939CE3939C63939C63939C63939C63939BD3939BD6B31BD6B39BD
6B31BD6B39BD6B31B56B39B56B31B56B39B56B31AD6B39AD6331AD6B39AD6331A56B39AD
398CA5398CA5398CA5398CA5398C9C398C9C398C9C398C9C398494398494398494398494
398494398494397B8C39A58C319C8C399C8C399C84399C84319484399484319484399484
31947B39945A31945A429431429431429431428C31429431428C315A8C315A8C315A8C31
5A8C315A8C315A8C315A8C315A8C315A8C315A8C31738C316B8431738C31738C31738C31
738C31736B31736B31736B31736B31736B31736B31735A31735231735A31735231735A31
735231735A317352317B5A317352317B5A31734A317B4A31734A317B4A317B4A317B4A31
7B4A317B42317B39317B42317B39317B42317B39317B42317B39317B4231000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A5231525A394A5A314A5A314A5231
4A5A394A52314A5A314252314A5A39396331426331395A31426339395A31425A31395A31
426339395A31315A42315239395A42315A39315A42315239395A39315A31395A63315A5A
396363315A63396363315A6339636B31636339636B31528C395A94315294395A94315294
395A9C315A9C395A9C315A9C3963A5315A9C3939A53131A53939AD3131A53939AD3131AD
3939B53131AD3939B53131AD3939BD3131B53939BD5231B55A39BD5231BD5A39C65231BD
5A39C65A31C65A39C65A31C66339CE5A31C66339CE5A31C66339CE6331CE6339CE6331CE
6339D66331CE6339D66331D66B39DE6331D66B39DE6331DE6B39E76339DE6B39E76331DE
6B42E76B39E76B39EF6B39E73942EF3939E73942EFA539E7AD42EFA539E7AD42E7A539DE
AD42E7A539DEAD42DEA539D6AD42DEA539D6AD42D6A539CEAD42D6A539CEAD42D6A539CE
AD42CE3939C63942CE3939C63942CE3939C63939C63131BD3939C66B31BD6B39BD6B31B5
6B39BD6B31B56B39B56331AD6B39B56331AD6B39B56331AD6B39AD6331A56B39AD6331A5
3994AD318CA5398CA531849C398CA531849C398C9C318494398C9C318494398494317B8C
398494317B8C39848C319C84399C8C319C84399C8C319484399C8431947B399484318C7B
399C6331945A39945A318C5A4A9439428C31429431398C315A94395A8C315A8C31528C31
5A8C395A8C315A8C315284315A8C396B8431738C316B8431738C396B8431738C316B8431
738C396B8431736B316B6331736B316B5231735A316B5229735A31735231735A31735229
7B5A317352317B5A317352297B5A317352317B5A317342297B4A317342317B4231733929
7B42317339317B42317339297B42317339317B42317339297B42317B3931000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A5A314A5A314A5A314A5A31
4A5A314A5A314A52314A5A314A5A314A5A31395A31426331395A31425A31395A31395A31
395A31395A31395A31315A39315A39315A42315239315A42315A42315A42315A39315A63
315A63315A63315A63315A6331636339636B31636331636331528C395294315294315A94
315A9C395A9C315A9C395A9C315A9C3963A531319C3931A53931A53939AD3131AD3931AD
3931AD3939B53131AD3931B53931B53939BD3131B55231BD5231BD5A39BD5231BD5A31BD
5A31C65A39C65A31C65A31C65A31C66339CE5A31C66331CE6331CE6339CE6331CE6339CE
6331CE6339D66331D66339D66339D66B39DE6339DE6B39DE6B39DE6B39E76B39DE6B39E7
6B39E76B39E76B39E76B39EF6B39EFA539EFA539E7A539E7A539E7AD42E7A539DEA539DE
A539DEAD42DEA539DEA539DEA539D6AD42D6A539D6AD39D6A539D6AD42D6A539CEAD39CE
AD39CE3942CE3939C63939C63939C63939C63939C63939C66B39BD6B39BD6B31BD6B39BD
6B39B56B39BD6B31B56B39B56B39B56B39B56331AD6B39AD6331AD6B39AD6331A56339A5
6331A5398CA5398C9C398C9C398C9C398C9C31849439849439849439849431848C39848C
397B8C39848C319C8C399C8C399C84399C84319484399484319484399484318C7B39945A
31945A39945A318C5A398C5A318C5A398C5A428C315A8C315A8C315A8C315A8C315A8C31
5A8C315A8C315284315A8C316B84316B84316B84316B84316B8431738C316B84316B8431
6B8C31738C316B8431735231735231735A31735231735231735231735A31735231735231
735231735A317352317352317352317B5A317352317342317339317B42317339317B4231
7B39317B42317339317B42317B39317B42317339317B42317B39317B42317B3931000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A5A314A5231
4A5A314252314A5A314252314A5A314252314A5A39425231426331395A31426339395A31
425A39395231425A39395231315A42315239315A42315239395A42315239395A42315239
395A8431527B395A63315A5A396363315A6339636B315A6339528C31528C395A94315294
395A9C315294395A9C315A9C395AA5315A9C3963A531319C3939A53131A53939AD3131AD
3939AD3131AD3939B53131AD3939B53131B53939BD3131B55A39BD5231B55A39BD5231BD
5A39C65231BD5A39C65A31C66339CE5A31C66339CE5A31C66339CE5A31CE6339D66331CE
6339D66331CE6339D66331D66B39DE6331D66B39DE6331DE6B39DE6331DE6B39E76339DE
6B39E76B39E76B42EF6B39E7C639C6A539E7A542EFA539E7AD39E7A539DEAD42E7A539DE
AD39DEA539DEAD42DEA539D6AD42DEA539D6AD42D6A539CEAD42D6A539CEAD42D6A539CE
AD42CEA539C63942CE3939C63939C63131BD3939C66B31BD7339C66B31BD6B39BD6B31B5
6B39BD6331B56B39B56331AD6B39B56331AD6B39B56331AD6B39AD6331A56B39AD6331A5
6339A563319C398CA531849C398C9C318494398C9C318494398494317B8C398494317B8C
39848C319C84399C8C319484399C8C319484399C8431947B399484318C7B399463318C5A
39945A318C5A39945A318C52398C5A318452398C395284315A8C395284315A8C39528431
5A8C315284316B84316B84316B84316B84316B8C316B84316B8C316B8431738C316B8431
738C316B6331736B316B6331735A316B5231735A316B5231735A316B5231735A31735231
735A317352297B5A317352317B5A317342297B4A317342317B42317339297B4231733929
7B42317339297B42317339317B42317339297B42317339297B42317339297B4231000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A5A31
4A52314A5A314A52314A5A314252314A52314252314A5231425231395A31395A31395A31
395A31395A31395A31395A31395231315A39315239315A42315239315A42315239315A42
315239315284315284315A84315A84315A63315A63316363315A6331528C31528C395294
315294395A94315A94395A9C315A9C395A9C315A9C3931A53131A53931A53131A53931AD
3131AD3931AD3131AD3931B53131B53931B53931B53939BD5231B55231BD5231BD5A39BD
5A31BD5A39C65A31C65A39C65A31C65A31C65A31C66339CE5A31CE6339CE6331CE6339CE
6331CE6339D66331D66339D66339D66339DE6339DE6B39DE6339DE6B39DE6B39DE6B39E7
6B39E76B39E7BD39BDBD39C6BD39C6A539EFA539E7A539E7A539E7A539E7A539DEA539DE
A539DEA539DEA539DEA539DEA539D6A539D6A539D6A539D6A539CEA539CEA539CEAD39CE
A539CEAD39CE3939C63939C63939C63939C66B31C66B39C66B31BD6B39BD6B39BD6B39BD
6B31B56B39B56B31B56B39B56331AD6B39AD6331AD6B39AD6331A56339A56331A56339A5
63319C63399C63319C398C9C39849439849439849439849439848C398494397B8C39848C
319C8C399C8C319C84399C8431948439948431947B39947B318C7B39945A31945A39945A
318C5A398C5A318C5A398C5A398431398C31398431398C315284315A8C315284315A8C31
6384316B84316B84316B84316B84316B84316B84316B84316B84316B84316B84316B6B31
6B63316B6B316B6B31736B316B5231735231735231735231735231735231735231735A31
735231735231734231734A31734231734A31734231734A31733931734231733931734231
7339317342317339317B42317339317B42317339317B42317339317B42317339317B4231
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
4A5A314252314A5A314252314A5A394252314A5A314252314A5A39425231425A31395A31
425A39395231395A31315231395A39395231315A42315239395A42315239315A42315239
395A4231527B395A8431527B395A84315284395A8C315A6339636B315A6339528C31528C
395A94315294395A9C315294395A9C315A9C395AA5315A9C3939A53131A53939A53131A5
3939AD3131AD3939B53131AD3939B53131B53939B53131B53939BD5231B55A39BD5231BD
5A39C65231BD5A39C65231BD6339C65A31C66339CE5A31C66339CE5A31CE6339CE5A31CE
6339D66331CE6339D66331CE6B39DE6331D66B39DE6331D66B39DE6339DE6B39E76331DE
BD39C6B539BDBD39C6BD39BDC642C6BD39BDA539E7A539DEA542E7A539DEA539E7A539DE
AD39DEA539DEA539DEA531D6AD39DEA539D6A539D6A531CEAD39D6A539CEAD39CEA531C6
AD39CEA539C6AD39CE3131C63939C66B31BD6B39C66B31BD6B39C66B31BD6B39BD6331B5
6B39BD6331B56B39B56331AD6B39B56331AD6B39AD6331A56B39AD6331A56339A55A319C
6339A55A319C63399C5A3194398C9C318494398494317B8C398494317B8C39848C319C84
399C8C319484399C8431947B399C84318C7B39947B318C73399463318C5A39945A318C52
39945A318C52398C5A318431398C39318431398C31318431398C39318431528431637B31
6B8439637B316B8431637B316B84396384316B84316384316B8C316B63316B6B316B6329
736B316B6331736B316B6329736B316B5231735A316B5229735A316B5231735A316B5229
734A31734231734A317342297B4A31734231734A317342297B4A317339317B4231733929
7B42317339317B42317339297B42317339317B4231733929000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A5A31425A314A5A31425A314A5A31425231425A31425A31425A31425231395A31
395A31395A31395231395A31395231395A31315231315239315239315242315239315239
31523931524231527B31527B31527B315A84315284315A84315A84395A8C314A8C31528C
31528C395294315294315A94315A9C395A9C315A9C395A9C31319C3931A53131A53931A5
3931AD3939AD3131AD3931AD3931B53939B53131B53931B53931B55239BD5231BD5231BD
5231BD5A39C65A31BD5A31C65A31C65A39C65A31C65A31CE5A31CE6339CE5A31CE6331CE
6331CE6339D66331CE6339D66339D66339D66331D66339DE6339DE6B39DEB539B5B539BD
B539BDBD39BDB539BDBD39BDBD39BDBD39C6A539E7A539E7A539E7A539E7A539DEA539DE
A539DEA539DEA539D6A539D6A539D6A539D6A539D6A539D6A539CEA539CEA539CEA539CE
A539CEA539CEA531C63939C66B39C66B39C66B31BD6B39BD6B39BD6B39BD6B31B56B39B5
6B39B56B39B56331AD6B39AD6331AD6B39AD6331AD6339AD6331A56339A56331A56339A5
63319C63399C5A31945A31945A319439849431848C39848C397B8C39848C319C84399C8C
399C84399C8431948439948439947B39947B318C7B39945A318C5A39945A318C5A318C5A
318C5A398C5A318431398431398431398431318431398431398431528431528431638431
6384316B8431637B316B84316B84316B84316384316B63316B63316B6B316B63316B6331
6B63316B6B316B63316B6B316B6B31736B316B5231735231735231734A31734231734231
734231734A31734231734A31734231734A31734231734A31734231734A31733931733931
733931734231733931733931733931000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A5A314252314A5A314252314A5A314252314A5A31425231425A31425231
395A31315231395A39395231395A39315231395A39315231315242315239395242315239
31524231523939524231527B39528431527B395A8431527B395A84315284395A8C313184
39318C31528C395A94315294395A9C315294395A9C315A9C3963A531319C3939A53131A5
3939AD3131A53939AD3131AD3939B53131AD3939B53131B53939BD5231B55A39BD5231BD
5A39BD5231BD5A39C65231BD5A39C65A31C66339CE5A31C66339CE5A31C66339CE5A31CE
6339D65A31CE6339D66331CE6339D66331D66B39DEAD31B5B539BDAD31B5B539BDB531B5
BD39BDB531B5BD39BDB539B5BD39BDBD39BDC639C69C39DEA539E79C39DEA539E7A539DE
A539DEA531D6A539DEA539D6A539D6A531CEA539D6A531CEA539D6A531CEAD39CEA531C6
A539CEA531C6AD39CEBD319C7339C66B31BD6B39C66B31BD6B39BD6331B56B39BD6331B5
6B39B56331AD6B39B56331AD6B39AD6331A56B39AD6331A56339AD63319C6339A55A319C
63399C5A319463399C5A31946339945A318C398494317B8C39848C319C84399C8C319484
399C8431947B399484318C7B39947B318C7339945A318C5A39945A318C52398C5A318452
398C5A318431398C39318431398439318431398439317B31528431527B31528431528431
6B8431637B316B8431637B316B84316363316B6B316363316B6B316363316B6B316B6331
6B6B316B6331736B316B6331736B316B6329734A316B4231734A316B4229734A316B4231
734A316B4229734A316B4231734A316B4229734A31734229734A31734229734231733929
734231733929000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A5A31425A31425A31425231425A31425231425A31425231425A31
425231395A31395A31395A31315231395231315231395231315239315239315239315239
31523931524231523931524231527B31527B31527B315284315284315A84315A8439318C
31318C31318C31318C393194315294395A94315A94395A9C315A9C39319C31319C3931A5
3131A53931AD3131AD3931AD3131AD3931B53931B53931B53131B53931BD5231B55239BD
5231BD5A31BD5231BD5A39C65A31C65A39C65A31C65A39C65A31C65A39CE5A31CE6339CE
6331CE6339CE6331CE6339D66331D6AD39B5AD39B5AD39B5AD39B5B539B5B539B5B539B5
B539B5B539B5B539B5BD39BDBD39BDBD39BDBD39BDA539E7A539DEA539DEA539DEA539DE
A539DEA539DEA539D6A539D6A539D6A539D6A539CEA539D6A539CEA539CEA539CEA539CE
A531C6A539C6BD39A5C639A5BD319C6B39BD6B39BD6B39BD6B31BD6B39BD6B31B56B39B5
6331B56B39B56331AD6339AD6331AD6339AD6331A56339A56331A56339A563319C63399C
5A319C5A399C5A31945A39945A31945A3994397B8C397B8C319C8C399C8C319484399484
319484399484318C7B398C7B318C7339945A318C5A398C5A318C5A398C5A318452398452
318431398431398431398431318431398431527B31528431527B31528431528431528431
528431528431637B316363316363316363316363316B63316B63316B63316B63316B6B31
6B63316B6B316B63316B6B316B52316B52316B5231734A316B42317342316B4231734A31
6B4231734A31734231734A31734231734A31734231734A31734231734A31734231000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004252314A5A31425231425A31425231425A31395231
425A39395231395A31315231395A39315231395231315231395239314A4A315242314A39
395242314A39315242314A3939527B314A7339527B31527B395A8431527B395A84315284
39398C31318C39318C31318C39399431319439319C315294395A9C315A9C3939A531319C
3939A53131A53939AD3131A53939B53131AD3939B53131AD3939BD3131B55239BD5231B5
5A39BD5231BD5A39C65231BD5A39C65A31BD5A39C65A31C66339CE5A31C66339CE5A31C6
6339D65A31CE6339D6A531ADAD39B5A531ADAD39B5AD31ADB539B5AD31ADB539B5AD31AD
B539BDB539B5B539BDB531B5BD39BDB539B5BD39BDB531B5A539E79C39DEA539DE9C31D6
A539DE9C39D6A539D69C31CEA539D6A539CEA539D69C31CEA539D6A531CEA539CE9C31C6
A539CEBD31A5C639A5BD319CC639A56B31BD6B39BD6331B56B39BD6331B56B39B56331AD
6B39B56331AD6B39AD6331AD6B39AD6331A56339AD5A31A56339A55A319C63399C5A3194
63399C5A31945A39945A318C5A399452318C5A398C319C84399C8C319484399C84318C7B
399484318C7B398C7B31847339945A318C5A398C5A318452398C5A318452398431317B31
398439317B31398431317B313984394A7B315284314A7B315284394A7B315284314A7B31
528431637331637B316363296B6B316363316B63316363296B6B316363316B6B31636329
6B6B316B63316B52316B4A297352316B4A317352316B4A29734A316B4231734A316B4229
734A316B4231734A316B4229734A316B4231734A316B4229000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000425231425A31425231425A31425231425231
425231425231395231395A31395231395A31315231315231315231395231314A3131524A
31524A315239314A3931523931523931527B314A7331527B31527B31528431527B315284
315A8439318C31318C31318C31318C39319431319431319431319C39319C31319C39319C
3131A53931A53131A53931AD3931AD3939AD3131AD3931B53931B53939B53131B55231BD
5231BD5A39BD5231BD5A31BD5A31BD5A39C65A31C65A31C65A31C65A39CE5A31C66331CE
9C31ADA539ADA531ADA539ADA539ADAD39ADAD31ADAD39ADAD39ADAD39B5AD31ADB539B5
B539B5B539B5B539B5B539B5B539B5BD39BDB539B5BD39BDA539DEA539DE9C39DEA539DE
A539D6A539D69C39D6A539D6A539D6A539D6A539CEA539CEA539CEA539CEA531C6A539C6
BD39A5BD39A5BD319CBD39A5BD399CBD399C6B31BD6B39BD6B39B56B39B56331B56339B5
6331AD6339B56331AD6339AD6331AD6339AD6331A56339A56331A56339A55A319C5A319C
5A31945A39945A31945A31945A318C5A398C31318439318439318439948431947B39947B
318C7B398C7B318C73398C5A318C5A398C5A318452318452318452398431317B31398431
398431398431317B314A7B314A7B31527B314A7B31527B31527B315284314A7B31637331
637331637B31637331637B316363316B63316363316B63316B63316B63316B63316B5231
6B52316B52316B4A316B52316B52316B52316B4A316B52316B4231734A316B42316B4231
6B4231734A316B4231734231734231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000425231425A31425231425A31395231
425A31395231425A31395231395A31315231395A31315231395231314A31395239314A31
31524A314A4231524A314A39395242314A3939527B314A7339527B314A7B39528431527B
395A8431318439398C31318439318C31318C39399431319439399C31319439399C52319C
5A39A552319C6339A55A31A53939AD3131AD3939B53131AD3939B53131B53939BD5231B5
5239BD5231B55A39BD5231BD5A39C65231BD9439A58C319C9439A59431A59C39AD9C31A5
A539AD9C31A5A539ADA531A5AD39ADA531A5AD39ADA531ADAD39B5AD31ADB539B5AD31AD
B539B5AD31ADB539B5B531B5B539B5B531B5BD39BDB531B5BD39BD9C31DEA539DE9C31D6
A539DE9C31D6A539D69C31CEA539D69C31CEA539D69C31CEA539CE9C31C6C639ADBD31A5
BD39A5B5319CBD39A5B5319CBD39A5B5319CBD399C6331B56B39BD6331B56B39B56331AD
6B39B56331AD6B39AD6331A56339AD6331A56339A55A319C6339A55A319C63399C5A3194
5A399C5A31945A399452318C5A398C31318439398C31318439398431317B393984318C7B
398C7B31847339945A318C52398C5A31845239845A317B52398439317B31398439317B31
397B39317B31527B394A7B31527B314A7B315284314A7B315284315A7331637B315A7331
637B315A7331637B316373316B7B316363316B6B316363316B6331634A296B5231634A31
6B5231634A296B52316B4A316B52316B4A297352316B4A297352316B4229734A316B4229
734A316B4229000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000425231425231425231425231
425231425231395231425231395231395A31315231395231315231395231315231315231
314A31314A4A314A4A314A4A314A4A314A39314A39314A73314A7331527B31527B31527B
31527B31528431318439318431318431318C31318C39319431319439319431319452319C
52319C5A31A55A31A55A31A55A31A56331AD6331AD6339AD6331AD6B31B56B31B57331B5
7B319C84319C84319C8C399C8C319C8C319C8C319C9439A59431A59439A59431A59C39A5
9C31A59C39AD9C31ADA539ADA531ADA539ADA531ADAD39ADAD31ADAD39ADAD31ADAD39AD
AD31ADB539B5B539ADB539B5B539B5B539B5B539B5BD39B5BD39B5A539DE9C39DEA539DE
9C39D6A539D69C39D6A539D69C39CEA539CEA539CEA539CE9C31CEBD39A5BD31A5BD39A5
BD31A5BD39A5B5399CBD399CB5319CB5399CB5319C6B39BD6331B56B39B56331B56339B5
6331AD6339AD6331AD6339AD6331A56339A56331A56339A55A319C5A399C5A319C5A3994
5A31945A39945A318C5A398C31318439318439318439318431317B39317B39317B39317B
313173398C5A318C52398C5A318452398452317B52398431317B31397B31317B31397B31
4A7B314A7B314A7B314A7B314A7B314A7B314A7B315A73315A73316373315A7331637331
637331637331637331637B31637331637B31634A316B5231634A316B5231634A316B5231
6B4A316B52316B52316B52316B4A316B52316B52316B52316B52316B52316B4231000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000425231425A31395229
425A31395231425231395231425231395231395A31315231395A39315231395231314A31
395239314A31314A4A314A4239524A314A42314A4A314A42395273314A7339527B314A73
39528431527B39528431317B39398C31318439318C31318C393994313194393194313194
52399C52319C5A39A552319C6339AD5A31A56339AD6331A56B39B56331AD6B39B56B31B5
84399C7B319484399C84319C8C39A584319C8C39A58C319C9439A59431A59439A59431A5
9C39AD9C31A5A539AD9C31A5A539ADA531A5A539ADA531A5AD39ADA531ADAD39ADAD31AD
B539B5AD31ADB539B5AD31ADB539B5B531ADB539B5B531ADBD39B5B531B5A539DE9C31D6
A539DE9C31D6A539D69C31CEA539D69C31CEA539CE9C31C6BD39ADBD31A5BD39A5B5319C
BD39A5B5319CBD39A5B5319CBD399CB5319CB5399CAD31946B39B56331B56B39B56331AD
6B39B56331AD6339AD5A31A56339AD5A31A56339A55A319C6339A55A31945A399C523194
5A399452318C5A398C31318439398C31318439398431317B39398431317B39397B313173
397373316B6B396B7331636B396B6B317B31397B39317B31397B39317331397B314A7331
4A7B394A73314A7B314A7331527B394A7B315A73315A6B316373315A73316373315A7329
637B31637331637B316373296B5A31635A316B5231634A296B5231634A316B5231634A29
6B5231634A316B5231634A296B52316B4A316B52316B4A29000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000425231
425231425231395231425231395231425231395231395231315231395231315231315231
314A31315231314A31314A4A314A42314A4A314A42314A4A314A4A314273314A73314A73
31527B31527B31527B31317B31318439318431318431318C31318C393194313194313194
31319452319C52319C5A31A55A31A55A39A55A31A56331AD6331AD6B39B56331AD6B31B5
6B31B57B399C7B319C84319C84319C8C399C8C319C8C31A58C31A59439A59431A59431A5
9431A59C39AD9C31A59C31ADA531ADA539ADA531ADA539ADA531ADAD39ADA531ADAD39AD
AD31ADAD39ADAD31ADB539ADB539ADB539B5B531ADB539B5B539B5BD39B59C39D69C39D6
9C39D6A539D69C31CE9C39CE9C39CEA539CE9C31CEBD39A5BD39A5BD39A5B531A5B539A5
B5399CB539A5B5319CB5399CB5399CB5399CAD3194B53994AD39946339B56331AD6339AD
6331AD6339AD6331A56339A56331A56339A55A319C5A319C5A319C5A399C5A31945A3194
5A318C5A398C31318439318439318439398431317B39317B39317B39397B316B73396B73
316B73396B7331636B31636B317B52397B52317B4A317B52317352397B314A73314A7331
4A73314A7B314A73314A7B315A73315A73315A6B315A73315A73316373315A7331637331
637331637B31635A31635A31635A31635A31635A31634A31634A316B5231634A316B5231
634A316B5231634A316B52316B4A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
425A31395231425A31395231425231395231425231395231395231315231395231314A31
395231314A31395239314A31314A4A314242314A4A314242314A4A314242394273314273
39527B314A7339527B314A7B39318431317B39398C31318439398C31318C393994313194
39399C4A319452399C52319C5A39A55231A56339AD5A31A56339AD6331AD6B39B56331AD
6B39B573319484399C7B319484399C84319C8C39A584319C8C39A58C319C9439A59431A5
9C39A59431A59C39AD9C31A5A539AD9C31A5A539ADA531A5A539ADA531A5AD39ADA531A5
AD39ADAD31A5B539ADAD31ADB539ADAD31ADB539B5B531ADBD39B5B531ADBD39B59C31D6
A539D69C31CEA539D69C31CEA539CEB531A5BD39ADB531A5BD39A5B531A5BD39A5B5319C
B539A5AD319CB5399CAD319CB5399CAD3194B5399CAD3194AD39946331AD6B39B56331AD
6339AD5A31A56339AD5A31A56339A55A319C63399C5A31945A399C5231945A399452318C
5A398C31318439398C31318439398431317B39398431317B39397B316B73396B7331636B
396B7331636B396B6B317B4A397B5231734A397B5231734A397331397331397B31397331
397B314273315A7331526B315A7331526B315A73315A6B316373315A73296373315A5231
635A315A5229635A31635231635A316352296B5A31635A296B5231634A296B5231634A29
6B5231634A29000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000425231425231425231395231425231395231395231395231395231315231395231
315231315231314A31314A31314A31314A42314A42314A4A314242314A4A31426B314273
314273314A73314A7331527B31527B31317B31318439318431318431318C31318C393194
3131943931944A319452319C52319C5A31A55A31A55A31A55A31A56331AD6331AD6339B5
6331AD6B31B57B31947B399C7B319C84319C84319C8C399C8C319C8C39A58C31A59439A5
9431A59439A59431A59C39AD9C31A59C39AD9C31ADA539ADA531ADA539ADA531ADAD39AD
A531ADAD39ADAD31ADAD39ADAD31ADB539ADB531ADB539ADB531ADB539ADB539AD9C39D6
9C39D69C39D69C39CE9C39CEB531A5BD39ADB531A5B539A5B531A5B539A5B531A5B539A5
B5319CB5399CAD319CB5399CAD319CAD399CAD3194AD3994AD3194AD39946331AD6339AD
6331AD6339AD6331A56339A55A319C5A399C5A319C5A399C5A31945A399452318C5A398C
31318C39318439318439398431317B39317B31317B393173316B73396B73316B6B396B6B
31636B31636B317B52317B5231734A317352397331397331397331397331397331397331
4A73315273315273315A73315A6B315A73315A73315A73315A73315A5A315A5A31635A31
5A5A31635A31635A31635A31635A31635A31635A31635A31635A31634A31634A31000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000395229425A31395231425231394A29425231395231395231394A31
395A31315231395231314A31395239314A31314A4A314242314A4A314242314A4A313963
39427331426B394273314A7339527B31317B39318431317B39398C31318439318C31318C
3939943131944A39944A319452399C52319C5A39A55231A56339AD5A31A56339AD6331AD
6B39B56331AD7B399C73319484399C7B319C84399C84319C8C39A58C319C8C39A58C319C
9439A59431A59C39A59431A59C39AD9C31A5A539AD9C31A5A539ADA531A5A539ADA531A5
AD39ADA531A5AD39ADAD31A5B539ADAD31ADB539ADAD31A5B539ADB531ADB539ADB531AD
A539D69C31CE9C39D6B531A5BD39ADB531A5B539A5B531A5B539A5AD319CB539A5AD319C
B539A5AD319CB5399CAD3194B5399CAD3194AD3994A53194AD3994A5318CAD39945A31A5
6339AD5A31A56339A55A319C6339A55A319C5A399C5231945A399452318C5A398C313184
39398C31318439398431317B39398431317339397331636B396B7331636B396B6B316363
397B5231734A397B5231734A397352396B31397331316B313973393973315273314A7329
527B314A7331527B315273295A7B315A6B315A73315A5229635A315A5231635A315A5229
635A315A5231635A315A5229635A31635231635A316352296B5A31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000425231425231395231395231395231395231395231395231
395231395231315231315231314A31314A31314A31314A31314242314A42314242314242
31396B31426B31426B31427331427331527B31317B31317B31318439318431318431318C
31318C3931943131944A319452319C52399C52319C5A31A55A31A55A39AD5A31A56331AD
6331AD6B39B56331B57B31947B31947B399C7B319C84319C84319C8C39A58C319C8C31A5
8C31A59439A59431A59431A59C31A59C39AD9C31A59C39AD9C31ADA539ADA531A5A539AD
A531ADAD39ADA531ADAD39ADAD31ADAD39ADAD31ADAD39ADB539ADB539ADB531ADB539AD
9C39CE9C39D6B531A5B539ADB539A5B539A5B531A5B539A5B539A5B539A5AD319CAD399C
AD399CAD399CAD319CAD399CAD3194AD3994A53194A53994A53194AD3994A5318CA5398C
6331A56339A55A319C5A319C5A319C5A399C5A31945A319452318C5A398C313184393184
39318439398431317B39317B313173393973316B6B316B6B31636B396B6B316363317352
31734A397352316B4A396B31396B31397331396B313973314A73314A73314A73314A7331
527331527B31527331527B315A63315A63315A52315A52315A52315A5A315A52315A5A31
5A5231635A315A5231635A31635A31635A31000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000394A29425231395231425231395231425231394A31
395231314A31395A31315231395231314A31395231314A31314A31314242314A4A314242
314A4A31396339426B31396B394273314273394A7B31317339318431317B393984313184
39398C31318C39399431318C52399C4A319452399C52319C5A39A55231A56339AD5A31A5
6339AD6331AD6B39B57331947B399C73319484399C7B319484399C84319C8C39A58C319C
9439A58C319C9439A59431A59C39AD9431A59C39AD9C31A5A539AD9C31A5A539AD9C31A5
A539ADA531A5AD39ADA531A5AD39ADA531A5AD39ADAD31A5B539ADAD31A5B539ADB531AD
B53963B531A5B539ADB531A5B539ADAD31A5B539A5AD319CB539A5AD319CB539A5AD319C
AD399CA53194AD399CA53194AD399CA53194AD3994A5318CA53994A5318CA539949C318C
9C398C5A319C6339A55A31945A399C5231945A399452318C5A398C31318439398C313184
39398431317B39397B31317339397331636B396B7331636B39636B315A63397B5231734A
397352316B4A397339316B31397331316B314A73314A6B314A73314A6B315273314A7331
527B314A73315A6331525A315A6331525A295A63315A52295A5A315A5229635A315A5229
635A315A5229635A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000425231425231425231395231395231
394A31395231394A31394A31315231315231314A31314A31314A31314A31314242314242
31424231424231396331426B31396B31426B31427331427331317B31317B313184393184
31318431318C31318C3931944A31944A31944A319C52319C52319C5A31A55A31A55A31AD
5A31AD6331AD6331AD6339B57331947B31947B31947B399C7B319C84399C84319C8C39A5
8C319C8C39A58C31A59439A59431A59C39A59C31A59C39AD9C31A59C39AD9C31A5A539AD
A531A5A539ADA531ADAD39ADA531A5AD39ADAD31ADAD39ADAD31ADB539ADB53163B53963
B53163B53963B531A5B539A5AD31A5B539A5AD31A5AD39A5AD319CAD399CAD319CAD399C
AD319CAD399CA53194AD399CA53194A53994A53194A53994A5318CA5398C9C318C9C398C
9431849439845A319C5A399C5A31945A399452318C52398C31318C393184393184393984
31317B39317B313173393173316B6B396B6B31636B39636B315A6331734A31734A316B4A
396B31396B31396B31396B31426B314A6B314A6B314A73314A6B314A73314A7331527331
525A315A6331525A315A6331525A315A63315A63315A63315A52315A52315A52315A5A31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000425231395231425231394A29
425231394A31395231394A29395231315231395231314A31395239314A31314A31294231
314A4231424231424231395A39426B31396331426B31396B394A7331317339317B31317B
39398431318439318C31318C3939944231944A39944A319452399C52319C5A39A55231A5
6339AD5A31A56339AD6331AD7339947331947B399C73319484399C7B319C84399C84319C
8C39A58C319C9439A58C319C9439A59431A59C39AD9431A59C39AD9C31A5A539AD9C31A5
A539ADA531A5A539ADA531A5AD39ADA531A5AD39ADA531A5AD3963AD315AAD3963AD315A
B53963AD315AB539ADAD31A5B539A5AD31A5AD39A5A5319CAD39A5A5319CAD399CA53194
AD399CA53194A5399CA53194A5399CA53194A539949C318CA539949C318C9C398C943184
9C398C9431849439845231945A399452318C5A398C31318439398C31318439398431317B
39397B31317339397331636B396B73316363396363316B4A397352316B4A396B31316331
396B39316B314A6B31426B314A7331426B314A73314A6B29527331525A31526331525A29
5A6331525A315A6331525A295A6331525A315A6331525A295A6331000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000425231425231425231
425231425231394A31395231394A31395231394A31315231315231315231314A31314A31
31423131424231424231424231424231396331396331426B31426B31427331317331317B
39317B39318431318431318C31318C4A31944A31944A31944A319C52399C52319C5A31A5
5A31A55A39AD5A31AD6331AD6331AD7339947331947B31947B31947B399C7B319C84319C
84319C8C39A58C319C8C31A59431A59439A59431A59C31A59C31A59C39AD9C31A59C39AD
9C31A5A539ADA531A5A539ADA531ADAD39ADA5315AAD3963AD3163AD3963AD315AAD3963
AD3963AD3963AD315AAD39A5AD39A5AD39A5AD319CAD399CA5399CAD399CA5319CA5399C
A5319CA5399CA53194A53994A53194A539949C318C9C398C9C318C9C398C943184943184
9431849439848C317B8C317B52318C52398C52318C39318439318439398431317B39317B
31317339397331636B31636B316363396363316B4A316B4A316B4A316B39316339316B39
426B31426B31426B314A6B314A6B314A73314A6B31525A31525A31525A31525A31525A31
525A31526331525A31525A31525A315A6331525A31000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000425229
425231394A29425231394A31425231394A29395231394A31395A31315231395231314A31
314A31314231314A3131424231424231394239396331396339426B31396B394273313173
39317B31317B39398439318439398C31318C4A399442318C4A399C4A319452399C52319C
5A39A55231A56339AD5A31A56339AD6331AD7339947331947B399C73319484399C7B319C
84399C84319C8C39A58C319C9439A58C31A59439A59431A59C39AD9431A59C39AD9C31A5
A539AD9C31A5A539AD9C31A5A53963A5315AA53963A5315AAD3963A5315AAD3963A5315A
AD3963A5315AAD3963A5315AAD39A5A5319CAD39A5A5319CAD399CA5319CA5399CA53194
A5399C9C3194A5399C9C3194A539949C318C9C399494318C9C398C94318494398C8C3184
8C398484317B8C398484317384397B52318439398C31317B39398431317339397B313173
39397331636B39636B315A63395A63316B4A396B5A316331316B39316331426B31396331
426B31396331426B31426B315263314A5A29525A314A5A31525A314A5A29526331525A29
526331525A295A6331525A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
425231425231425231425231395231395231394A31395231394A31394A31315231315231
314A31314A31314A31314A3131423931424231394231424231396331396B31396B31316B
31317331317331317B39318439318439318C31318C4A31944A31944A31944A319C52319C
52319C5A31A55A31A55A31AD5A31AD6331AD6B31947339947331947B31947B31947B399C
7B319C84399C84319C8C39A58C319C8C39A59431A59439A59431A59C39A59C31A59C39AD
9C31A59C39AD9C315AA539639C315AA53963A5315AA53963A5315AA53963A5315AA53963
A5315AA5395AA5315AA5395AA5319CA539A5A5319CA5399CA5319CA5399CA5319CA5399C
9C3194A539949C31949C39949C318C9C398C94318C94398C9431849439848C31848C3984
84317B84397B84317B84397B6B31846B31846B318463397B63317B39317331317339316B
31636B393163313163316B5A31635A316339316339316339396331396331396331426B31
426B31426B314A63314A63314A63315263314A6331526331525A31525A31525A31525A31
525A31000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000425231394A29425231394A31424A31394A29424A31394A31394A31315229
395A31315231395231314A31314A31314231314A3929393939424231395A31426B313963
39396B31316B39317331317339398439318439398C4231734A397B4231944A399C4A3194
5239A552319C5A39A55231A56339AD5A31AD6339AD6B318C73399C7331947B399C733194
84399C7B319C84399C84319C8C39A58C319C9439A58C31A59439AD9431A59C39AD9431A5
9C39639C315A9C39639C315AA539639C315A9C39639C315AA539639C315AA539639C315A
A539639C315AA539639C315AA53963A5319CA5399C9C319CA5399C9C3194A5399C9C3194
A5399C9C31949C399494318C9C399494318C94398C8C318494398C8C31848C398484317B
8439847B317384397B6B31846B398C63317B63397B5A317363397B5A316B39396B313163
39396B31315A396B5A315A52396339315A31396331396329426B31396331426B31425A29
4A63314A5A314A63314A5A295263314A63315263314A5A295263314A5A29000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000424A31424A31425231424A31424A31424A31424A31394A31394A31
394A31395A31315231395231314A31314A31314A31314A3131423931423931395A31395A
31396331316B31316B31317331317339317B39317B3931844231734A317B4A317B52317B
52318452399C52319C5231A55A31A55A39AD5A31AD6B31946B31947339947331947B3194
7B319C7B399C7B319C84319C84319C8C39A58C319C8C31A59431A59439A594315A94315A
94315A9C396394315A9C395A9C315A9C39639C315A9C315A9C315A9C39639C315A9C395A
9C315AA5395A9C315A9C395AA5315AA5395A9C319CA5399C9C319C9C399C9C31949C3994
9C31949C399494318C94398C94318C94398C8C31848C31848C31848C398484317B84317B
7B317B7B397B6B31846B318463317B63397B5A31735A317352316B39316B313163313163
31635A39635A315A31396331396331396331396331396331426B314A63314A5A314A5A31
4A5A314A63314A5A314A63314A63315263314A6331000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000424A294A5231424A29425231394A29425231394A29
424A31394A29424A31395231395A31315229395231314A31314A31294231314A39313931
31395A31396339316B31316B39317331317339397B31317B3939844231734A397B423173
5239844A317B5A39845A318463398C5A318C6339AD5A31AD6B39946B318C733994733194
7B399C73315284395A7B315A84395A84315A8C396384315A8C39638C315A94396394315A
94396394315A9C396394315A9C396394315A9C396394315A9C396394315A9C396394315A
9C395A94315A9C395A9C315AA5395A9C315AA5395A9C31949C399C9C31949C399C943194
9C399494318C9439948C318C94398C8C31848C398C84317B8439847B317B84397B7B3173
7B397B6331846B39845A317B63397B5A31735A397352316B39396B313163393963315A52
396339315A31396331395A313963313963294A5A31425A294A6331425A294A63314A5A31
4A63314A5A295263314A5A29000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000424A31425231424A31424A31424A31424A31
424A31424A31424A31424A31395A31395A31395231395231315231315231314A31314A31
31423131423931395A31316B31316B31317331317339317B31317B393184423173423173
4231734A317B4A317B5231845231845A31845A318C63318C6331526B315A6B315A73395A
73315A73315A73315A7B395A7B315A84395A84315A84395A84315A8C395A8C315A94395A
8C315A94395A94315A94395A94315A94395A94315A94395A94315A94395A94315A9C395A
94315A9C395A94315A9C395A9C315A9C395A9C315A9C399C9C31949C3994943194943994
94318C94398C8C318C8C398C8C31848439848431848439847B317B7B397B7B31736B398C
6331846331845A317B5A317B5A317352317331316B39316331315A316352315A31395A31
395A31396331396331426331425A31426331425A314A5A31425A314A5A314A5A314A6331
4A5A31000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A5231424A314A5231424A29
4A5231424A31425231394A29425231394A31395A31315229395A31315231395231314A29
395231314A31314A3129422939395A31316339317331316B39397B31317B39398439316B
42397B4231734A397B4A317B5239845231845A398C5A318463398C6331526B395A633152
73395A6B315273395A7331527B395A7B315A7B395A7B315284395A84315A8C395A84315A
9439638C315A94395A8C315A9439638C315A94395A8C315A9439638C315A94395A8C315A
94395A94315A94395A9431529C395A94315A9C395A9431529C399C9431949439948C318C
9439948C318C8C398C8431848C398484317B8439847B317B7B397B7331736B398C63317B
6339845A317B5A397B52316B523973313163393963313152395A31315A31396331315A29
426331425A31426331425A294A6331425A314A6331425A294A6331425A29000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A4A314242314A4A31
424A314A4A31424231424A31424231424A31424231424231425231425231395231395231
395231395231314A31314A31314A31314A3131315A31316B31317339317331317339317B
39316B4231734231734A317B4A317B52398452317B5A31845A318463398C633152633152
6B31526B395A6B315273315A73315A73395A73315A7B315A7B315A84395A84315A84315A
84315A8C395A8C315A8C315A8C315A8C395A8C315A8C315A8C315A94395A8C315A94315A
94315A94395A94315A94395A94315A94395A94315A94315A94315A9439948C318C8C398C
8C318C8C398C8431848431847B31847B397B7B317B7B317B73317363398C6331845A317B
5A317B5A397352316B4A316B31316339315A315A52395A31396331425A31425A31425A31
425A31425A31425A31425A31425A314A6331425A31000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000424229
4A4A314242294A4A314242294A4A314242294A4A31424229424A31395229425A31395229
425A31395229425231314A29395231314A29314A3131315239316331316B393973313173
39396B39316B4239734231734A397B4231735239844A317B5A398452318463398C5A3152
63395A6331526B395A6B315273395A6B315273395A7331527B395A73315284395A7B315A
84395A84315A8C395A84315A8C395A84315A8C395A84315A8C395A8431528C395A8C3152
8C395A8C315294395A8C315294395A8C315294395A8C31528C395A8C31528C398C843184
84398C7B31848439847B317B7B39847331737B397B73317363398C5A317B633984523173
5239734A316B39396B31315A393152395229425A31395229425A31395A29425A31425A29
426331425A294A6331000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
4A4A314A4A314A4A314A4A314A4A314A4A31424A314A4A31424231424A31424231425A31
425231425A31425231425231395231395231395231395231314A3131315A31315A313173
31317339316B39316B4231734231734231734231734A317B4A317B5231845231845A3152
5A31526331526331526339526331526B315A6B315273395A73315273395A73315A7B395A
7B315A84395A84315A84395A84315A84395A84315A84395A84315A8C395A84315A8C395A
8C315A8C395A8C315A8C395A8C31528C395A8431528C395A843152843952843152843952
7B31847B39847B317B7B397B73317B73397B6B31736339845A317B5A317B523173523173
4A316B39316331315A425A31395A31425A31425A31425A31425A31425A31425A31425A31
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A4A314A42294A4A314A42294A4A314242294A4A314242294A4A31424229
4A4A31425229425A31395229425A31395231425A31395229425231395229395231313152
39396331317339396B39316B42397339316B4239734231734A397B4A317B52398452317B
5A39525A31526339525A315263395A6331526B395A6B315273395A6B315273395A733152
7B395A7B315284395A7B315284395A7B315284395A7B315284395A84315284395A843152
8C395A84315284395A84315284395A7B315284395A7B315284395A7B31527B395273314A
7B395273317B7B397B6B317373397B5A318463398452317B5A397B4A317352396B423163
393963395231425A31395229425A31425A31425A31425A29000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A4A314A4A314A4A314A42294A42314A42314A42314A42314A4231
4A42314A4231423929424A31424A314A4A31424A31424A31424A31424A31394A31394A31
394A3131315A31316339316339316B39316B39316B4231734231734A397B4A317B4A317B
5231845A39525231525A31525A31526339526331526331526B31526B395A6B3152733152
73315A7B395A7331527B315A7B315A7B395A7B31527B315A7B315A84395A7B315284315A
84315284395A7B31527B31527B31527B39527331527B315273315273395273314A733152
7331527339526B314A6B31735A31845A398452317B5231734A316B4A396B313163425A31
425A31425A31425A31425A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A39295242314A42295242314A39295242314A4229
4A42314A39294A42314239294A5231424A294A5231424A294A5231424A294A5231424A29
424A31394A29424A3131315239397331316339396B39316B4239734231734A397B423173
4A397B4A317B5239525231525A39525A31526339525A31526339526331526B395A6B3152
73395A6B315273395A7331527B395A7331527B395A7B31527B395A7B31527B395A733152
7B395A7331527B395A7331527339527331527339526B314A7339526B314A7339526B314A
6B395263314A6B397363316B5A396B5231635239634A315A4A395A395229425A31425A29
425A31000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004A42315242314A42315242314A4231524231
4A42314A42314A42314A42314A42314A42314A4A314A52314A4A314A52314A4A314A5231
424A31424A31424A31424A31424A3131315A31315A31316339316342316B423173423173
4231734A317B4A317B5231525231525A31525A31525A39525A3152633152633152633952
6B31526B39526B315273395A73315273315A7331527B395A733152733952733152733952
7331527331526B31526B39526B31526B39526B31526B395263314A63314A63314A63394A
6331735A316B52316B5239634A315A4A315A425231425A31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A42295242314A42295242314A3929
5242314A42295242314A39295242314A42295242314A39295252314A4A294A4231424229
4A4A314242294A42314242294A42314242294A423131315239395A31315A393163393163
4239734231734A397B4A314A5239524A314A52395252314A5A39525A31526339525A3152
6B395A6331526B395A63315273395A6B315273395A6B315273395A6B31526B395263314A
6B39526331526B395263314A6B395263314A6339525A314A6339525A314A5A3973523163
52396B4A31634A3963393152000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000524231524229524231
5242315242314A42295242315239315239314A39295239314A39315242314A42294A4231
4A42314A42314A42294A42314A42314A42314A42294A42314A423131315A31315231315A
31315A39316339316B4231734A314A4A39524A314A4A315252315252395252314A523152
5A31525A39525A31525A31526331526339526331526331526331526B3952633152633152
6331526339525A314A5A314A5A314A5A394A5A314A5A314A52316B52396B4A31634A3163
42315A423152000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000523931523929
5239315239295242315239295239314A39295239314A39295239314A39295242314A4229
524A314A4229524A314A4229524A314A4229524A314A4229524A314A4229524A314A4A29
39395A31315239396331315A39396B39316B4A395242314A4A395242314A4A395242314A
4A39524A314A5239524A314A52395252314A5A395252314A5A39525A314A6339525A314A
5A395252314A5A395252314A5A394A52314A52396B4A31634A396342315A42395A393152
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
523929523931523931523931523929523931523931523931523929523931523931523931
5242295242315242315242315242295242315242315242314A42295242314A3931523931
4A39314A42314A42314A423131315A31315A31316339316B39314A42314A42314A42314A
42314A42314A42314A4A314A42314A4A314A4A314A4A394A4A314A52314A52314A52394A
52314A52314A52314A52394A4A316B4A316342315A42315A393152000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005239295239315239295A42315239295239315239295A4231523929523931523929
5A4231523929524A31524229524A315239295239314A39295242314A39295239314A3929
5242314A39295242314A4229524A314A422952423131315239396339316339394A393142
42394A39314242394A39314242394A42314242394A4231424A394A42314A4A394A423142
4A394A4A314A393152000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005239315239315A3931523929523931523931523931523929523931
523131523931523129523931523931523931523929523931523931523931523929523931
5239315239315239295242315242315242314A42295242315242315242314A4229313142
39314239394A31315A313152000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005A42315239295A42315231295A39315231295A3931523129
5A39315231295A39315231295A39315239295A42315239295A42315239295A4231523929
5239315239295242315239295239314A4229524A31524229524231000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005A39315231295A39315231295A31315231295A3931
5231295A39315231315A3931523129523131523929523931523929523931523931523931
523929523931523929523931000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005231295A39315231295A3931523129
5A39315231295A39315231295A39315231295A39315239295A42315239295A3931523929
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005A31315A39315231295A3131
5A31295A39315231295A31315A31315A3931523129000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005231295A3931523129
5A39315231295A3931000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0113379 0.97619 0.464853 MathSubStart
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
1.56e-16 1.04948 -0.0840974 1.04948 [
[ 0 0 0 0 ]
[ 1 .7782 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.03114 .256 m
0 .45252 L
s
0 .45252 m
.39761 .7782 L
s
.39761 .7782 m
.40383 .60035 L
s
.40383 .60035 m
.03114 .256 L
s
.69069 0 m
.97431 .43047 L
s
.97431 .43047 m
1 .61842 L
s
1 .61842 m
.70263 .20566 L
s
.70263 .20566 m
.69069 0 L
s
.03114 .256 m
0 .45252 L
s
0 .45252 m
.70263 .20566 L
s
.70263 .20566 m
.69069 0 L
s
.69069 0 m
.03114 .256 L
s
.40383 .60035 m
.97431 .43047 L
s
.97431 .43047 m
1 .61842 L
s
1 .61842 m
.39761 .7782 L
s
.39761 .7782 m
.40383 .60035 L
s
0 0 m
1 0 L
1 .7782 L
0 .7782 L
closepath
clip
newpath
.5 Mabswid
.44426 .58656 m .40864 .59572 L .38732 .57745 L p .626 .101 0 r
F P
0 g
s
.48039 .57795 m .44426 .58656 L .42321 .56925 L p .6 .176 0 r
F P
0 g
s
.38732 .57745 m .42321 .56925 L .44426 .58656 L p .626 .101 0 r
F P
0 g
s
.42321 .56925 m .38732 .57745 L .36542 .55939 L p .595 .174 0 r
F P
0 g
s
.51703 .56985 m .48039 .57795 L .45962 .56219 L p .565 .281 0 r
F P
0 g
s
.42321 .56925 m .45962 .56219 L .48039 .57795 L p .6 .176 0 r
F P
0 g
s
.45962 .56219 m .42321 .56925 L .40156 .55275 L p .538 .302 0 r
F P
0 g
s
.36542 .55939 m .40156 .55275 L .42321 .56925 L p .595 .174 0 r
F P
0 g
s
.55424 .56197 m .51703 .56985 L .4966 .55618 L p .524 .411 0 r
F P
0 g
s
.45962 .56219 m .4966 .55618 L .51703 .56985 L p .565 .281 0 r
F P
0 g
s
.40156 .55275 m .36542 .55939 L .34291 .54151 L p .547 .271 0 r
F P
0 g
s
.4966 .55618 m .45962 .56219 L .43822 .5481 L p .445 .476 0 r
F P
0 g
s
.40156 .55275 m .43822 .5481 L .45962 .56219 L p .538 .302 0 r
F P
0 g
s
.59201 .55377 m .55424 .56197 L .5342 .55069 L p .494 .55 0 r
F P
0 g
s
.4966 .55618 m .5342 .55069 L .55424 .56197 L p .524 .411 0 r
F P
0 g
s
.43822 .5481 m .40156 .55275 L .37925 .53696 L p .431 .462 0 r
F P
0 g
s
.34291 .54151 m .37925 .53696 L .40156 .55275 L p .547 .271 0 r
F P
0 g
s
.5342 .55069 m .4966 .55618 L .47552 .54532 L p .323 .665 0 r
F P
0 g
s
.43822 .5481 m .47552 .54532 L .4966 .55618 L p .445 .476 0 r
F P
0 g
s
.63035 .5446 m .59201 .55377 L .57244 .54475 L p .499 .675 0 r
F P
0 g
s
.5342 .55069 m .57244 .54475 L .59201 .55377 L p .494 .55 0 r
F P
0 g
s
.37925 .53696 m .34291 .54151 L .31977 .52347 L p .486 .381 0 r
F P
0 g
s
.47552 .54532 m .43822 .5481 L .41613 .53555 L p .227 .666 0 r
F P
0 g
s
.37925 .53696 m .41613 .53555 L .43822 .5481 L p .431 .462 0 r
F P
0 g
s
.6692 .53391 m .63035 .5446 L .6113 .53716 L p .548 .762 0 r
F P
0 g
s
.57244 .54475 m .6113 .53716 L .63035 .5446 L p .499 .675 0 r
F P
0 g
s
.57244 .54475 m .5342 .55069 L .51351 .54353 L p .196 .734 0 r
F P
0 g
s
.47552 .54532 m .51351 .54353 L .5342 .55069 L p .323 .665 0 r
F P
0 g
s
.41613 .53555 m .37925 .53696 L .35627 .52134 L p .291 .61 0 r
F P
0 g
s
.31977 .52347 m .35627 .52134 L .37925 .53696 L p .486 .381 0 r
F P
0 g
s
.70852 .52148 m .6692 .53391 L .6507 .52689 L p .617 .777 0 r
F P
0 g
s
.6113 .53716 m .6507 .52689 L .6692 .53391 L p .548 .762 0 r
F P
0 g
s
.51351 .54353 m .47552 .54532 L .4537 .5371 L p 0 .566 0 r
F P
0 g
s
.41613 .53555 m .4537 .5371 L .47552 .54532 L p .227 .666 0 r
F P
0 g
s
.35627 .52134 m .31977 .52347 L .296 .50475 L p .43 .48 0 r
F P
0 g
s
.6113 .53716 m .57244 .54475 L .55226 .5411 L p .145 .746 0 r
F P
0 g
s
.51351 .54353 m .55226 .5411 L p .53323 .54394 L .196 .734 0 r
F P
0 g
s
.55226 .5411 m .51351 .54353 L .49207 .54029 L p .029 0 0 r
F P
0 g
s
.49207 .54029 m .53134 .54263 L .55226 .5411 L p .029 0 0 r
F P
0 g
s
.55226 .5411 m .53134 .54263 L p .57178 .5382 L 0 0 0 r
F P
0 g
s
.59172 .53601 m .55226 .5411 L p .57178 .5382 L 0 0 0 r
F P
0 g
s
.55226 .5411 m .57244 .54475 L p .53323 .54394 L .196 .734 0 r
F P
0 g
s
.4537 .5371 m .41613 .53555 L .39331 .52366 L p 0 .538 0 r
F P
0 g
s
.35627 .52134 m .39331 .52366 L .41613 .53555 L p .291 .61 0 r
F P
0 g
s
.74828 .50749 m .70852 .52148 L .69055 .51354 L p .675 .706 0 r
F P
0 g
s
.6507 .52689 m .69055 .51354 L .70852 .52148 L p .617 .777 0 r
F P
0 g
s
.6507 .52689 m .6113 .53716 L .59172 .53601 L p .174 .82 0 r
F P
0 g
s
.59172 .53601 m .6113 .53716 L p .5769 .53946 L .145 .746 0 r
F P
0 g
s
.63177 .52652 m .6507 .52689 L p .61772 .53199 L .174 .82 0 r
F P
0 g
s
.59172 .53601 m .63177 .52652 L p .61772 .53199 L .174 .82 0 r
F P
0 g
s
.63177 .52652 m .59172 .53601 L p .60694 .53247 L 0 0 0 r
F P
0 g
s
.61224 .53261 m .63177 .52652 L p .61316 .53097 L 0 0 0 r
F P
0 g
s
.57146 .54097 m .61224 .53261 L p .61316 .53097 L 0 0 0 r
F P
0 g
s
.61224 .53261 m .59191 .54252 L p .61129 .5343 L 0 .5 .66 r
F P
0 g
s
.63177 .52652 m .61224 .53261 L p .64618 .5209 L 0 .827 .259 r
F P
0 g
s
.65342 .51642 m .67223 .51191 L p .65253 .51871 L 0 .827 .259 r
F P
0 g
s
.61224 .53261 m .65342 .51642 L p .65253 .51871 L 0 .827 .259 r
F P
0 g
s
.65342 .51642 m .61224 .53261 L p .61129 .5343 L 0 .5 .66 r
F P
0 g
s
.65342 .51642 m .63389 .52452 L p .65435 .51406 L 0 .908 .841 r
F P
0 g
s
.69478 .49338 m .65342 .51642 L p .65435 .51406 L 0 .908 .841 r
F P
0 g
s
.67223 .51191 m .63177 .52652 L p .64618 .5209 L 0 .827 .259 r
F P
0 g
s
.59172 .53601 m .57146 .54097 L p .60694 .53247 L 0 0 0 r
F P
0 g
s
.55226 .5411 m .59172 .53601 L p .5769 .53946 L .145 .746 0 r
F P
0 g
s
.49207 .54029 m .51351 .54353 L p .4713 .53899 L 0 .566 0 r
F P
0 g
s
.49207 .54029 m .46979 .5392 L p .51342 .54163 L 0 0 0 r
F P
0 g
s
.53134 .54263 m .49207 .54029 L p .51342 .54163 L 0 0 0 r
F P
0 g
s
.4537 .5371 m .49207 .54029 L p .4713 .53899 L 0 .566 0 r
F P
0 g
s
.39331 .52366 m .35627 .52134 L .33263 .5049 L p .153 .628 0 r
F P
0 g
s
.296 .50475 m .33263 .5049 L .35627 .52134 L p .43 .48 0 r
F P
0 g
s
.78848 .49248 m .74828 .50749 L .73077 .49745 L p .706 .575 0 r
F P
0 g
s
.69055 .51354 m .73077 .49745 L .74828 .50749 L p .675 .706 0 r
F P
0 g
s
.33263 .5049 m .296 .50475 L .27163 .48464 L p .406 .551 0 r
F P
0 g
s
.49207 .54029 m .4537 .5371 L .43109 .53021 L p .013 0 0 r
F P
0 g
s
.39331 .52366 m .43109 .53021 L .4537 .5371 L p 0 .538 0 r
F P
0 g
s
.69055 .51354 m .6507 .52689 L .63177 .52652 L p .262 .907 0 r
F P
0 g
s
.82916 .47708 m .78848 .49248 L .77135 .47961 L p .715 .425 0 r
F P
0 g
s
.73077 .49745 m .77135 .47961 L .78848 .49248 L p .706 .575 0 r
F P
0 g
s
.43109 .53021 m .39331 .52366 L .36978 .51077 L p 0 .41 0 r
F P
0 g
s
.33263 .5049 m .36978 .51077 L .39331 .52366 L p .153 .628 0 r
F P
0 g
s
.73077 .49745 m .69055 .51354 L .67223 .51191 L p .419 .943 0 r
F P
0 g
s
.63177 .52652 m .67223 .51191 L .69055 .51354 L p .262 .907 0 r
F P
0 g
s
.87037 .46184 m .82916 .47708 L .81235 .4612 L p .71 .287 0 r
F P
0 g
s
.77135 .47961 m .81235 .4612 L .82916 .47708 L p .715 .425 0 r
F P
0 g
s
.36978 .51077 m .33263 .5049 L .30838 .48639 L p .097 .585 0 r
F P
0 g
s
.27163 .48464 m .30838 .48639 L .33263 .5049 L p .406 .551 0 r
F P
0 g
s
.30838 .48639 m .27163 .48464 L .2467 .46256 L p .42 .585 0 r
F P
0 g
s
.43109 .53021 m .46979 .5392 L .49207 .54029 L p .013 0 0 r
F P
0 g
s
.53134 .54263 m .57146 .54097 L .59172 .53601 L p 0 0 0 r
F P
0 g
s
.77135 .47961 m .73077 .49745 L .71298 .49281 L p .628 .923 0 r
F P
0 g
s
.67223 .51191 m .71298 .49281 L .73077 .49745 L p .419 .943 0 r
F P
0 g
s
.91217 .44706 m .87037 .46184 L .85385 .44324 L p .699 .178 0 r
F P
0 g
s
.81235 .4612 m .85385 .44324 L .87037 .46184 L p .71 .287 0 r
F P
0 g
s
.46979 .5392 m .43109 .53021 L .4077 .52214 L p 0 0 0 r
F P
0 g
s
.4077 .52214 m .38364 .50969 L p .42058 .52555 L 0 0 0 r
F P
0 g
s
.4077 .52214 m .43109 .53021 L p .39521 .51884 L 0 .41 0 r
F P
0 g
s
.4077 .52214 m .44664 .53675 L .46979 .5392 L p 0 0 0 r
F P
0 g
s
.44664 .53675 m .4077 .52214 L p .42058 .52555 L 0 0 0 r
F P
0 g
s
.36978 .51077 m .4077 .52214 L p .39521 .51884 L 0 .41 0 r
F P
0 g
s
.81235 .4612 m .77135 .47961 L .75397 .47087 L p .749 .768 0 r
F P
0 g
s
.71298 .49281 m .75397 .47087 L .77135 .47961 L p .628 .923 0 r
F P
0 g
s
.4077 .52214 m .36978 .51077 L .34563 .49482 L p 0 .383 .102 r
F P
0 g
s
.30838 .48639 m .34563 .49482 L .36978 .51077 L p .097 .585 0 r
F P
0 g
s
.95461 .43281 m .91217 .44706 L .89595 .42627 L p .687 .102 0 r
F P
0 g
s
.85385 .44324 m .89595 .42627 L .91217 .44706 L p .699 .178 0 r
F P
0 g
s
.34563 .49482 m .30838 .48639 L .28362 .46475 L p .111 .577 0 r
F P
0 g
s
.2467 .46256 m .28362 .46475 L .30838 .48639 L p .42 .585 0 r
F P
0 g
s
.57146 .54097 m .53134 .54263 L .50955 .54712 L p 0 0 0 r
F P
0 g
s
.46979 .5392 m .50955 .54712 L .53134 .54263 L p 0 0 0 r
F P
0 g
s
.85385 .44324 m .81235 .4612 L .79528 .44813 L p .741 .511 0 r
F P
0 g
s
.75397 .47087 m .79528 .44813 L .81235 .4612 L p .749 .768 0 r
F P
0 g
s
.28362 .46475 m .2467 .46256 L .22124 .43824 L p .461 .575 0 r
F P
0 g
s
.71298 .49281 m .67223 .51191 L .65342 .51642 L p 0 .96 .122 r
F P
0 g
s
.89595 .42627 m .93873 .41039 L .95461 .43281 L p .687 .102 0 r
F P
0 g
s
.89595 .42627 m .85385 .44324 L .83704 .42626 L p .723 .313 0 r
F P
0 g
s
.79528 .44813 m .83704 .42626 L .85385 .44324 L p .741 .511 0 r
F P
0 g
s
.50955 .54712 m .46979 .5392 L .44664 .53675 L p 0 0 0 r
F P
0 g
s
.75397 .47087 m .71298 .49281 L .69478 .49338 L p .163 .967 0 r
F P
0 g
s
.65342 .51642 m .69478 .49338 L .71298 .49281 L p 0 .96 .122 r
F P
0 g
s
.61224 .53261 m .57146 .54097 L .55034 .54951 L p 0 0 0 r
F P
0 g
s
.50955 .54712 m .55034 .54951 L .57146 .54097 L p 0 0 0 r
F P
0 g
s
.34563 .49482 m .38364 .50969 L .4077 .52214 L p 0 .383 .102 r
F P
0 g
s
.93873 .41039 m .89595 .42627 L .87941 .40618 L p .703 .177 0 r
F P
0 g
s
.83704 .42626 m .87941 .40618 L .89595 .42627 L p .723 .313 0 r
F P
0 g
s
.79528 .44813 m .75397 .47087 L .73624 .46608 L p .503 .921 0 r
F P
0 g
s
.69478 .49338 m .73624 .46608 L .75397 .47087 L p .163 .967 0 r
F P
0 g
s
.38364 .50969 m .34563 .49482 L .32099 .474 L p 0 .435 .288 r
F P
0 g
s
.28362 .46475 m .32099 .474 L .34563 .49482 L p .111 .577 0 r
F P
0 g
s
.32099 .474 m .28362 .46475 L .2584 .43951 L p .165 .585 0 r
F P
0 g
s
.22124 .43824 m .2584 .43951 L .28362 .46475 L p .461 .575 0 r
F P
0 g
s
.2584 .43951 m .22124 .43824 L .19524 .41186 L p .508 .523 0 r
F P
0 g
s
.83704 .42626 m .79528 .44813 L .77787 .43768 L p .779 .829 0 r
F P
0 g
s
.73624 .46608 m .77787 .43768 L .79528 .44813 L p .503 .921 0 r
F P
0 g
s
.55034 .54951 m .59191 .54252 L .61224 .53261 L p 0 0 0 r
F P
0 g
s
.87941 .40618 m .92249 .38807 L .93873 .41039 L p .703 .177 0 r
F P
0 g
s
.55034 .54951 m .50955 .54712 L .48682 .55011 L p .002 0 0 r
F P
0 g
s
.44664 .53675 m .48682 .55011 L .50955 .54712 L p 0 0 0 r
F P
0 g
s
.87941 .40618 m .83704 .42626 L .81989 .41076 L p .752 .503 0 r
F P
0 g
s
.77787 .43768 m .81989 .41076 L .83704 .42626 L p .779 .829 0 r
F P
0 g
s
.59191 .54252 m .63389 .52452 L .65342 .51642 L p 0 .5 .66 r
F P
0 g
s
.2584 .43951 m .29598 .44752 L .32099 .474 L p .165 .585 0 r
F P
0 g
s
.29598 .44752 m .2584 .43951 L .23274 .41098 L p .269 .596 0 r
F P
0 g
s
.19524 .41186 m .23274 .41098 L .2584 .43951 L p .508 .523 0 r
F P
0 g
s
.92249 .38807 m .87941 .40618 L .86249 .38671 L p .723 .282 0 r
F P
0 g
s
.81989 .41076 m .86249 .38671 L .87941 .40618 L p .752 .503 0 r
F P
0 g
s
.38364 .50969 m .42275 .52842 L .44664 .53675 L p 0 0 0 r
F P
0 g
s
.48682 .55011 m .44664 .53675 L .42275 .52842 L p .002 0 0 r
F P
0 g
s
.32099 .474 m .3591 .49006 L .38364 .50969 L p 0 .435 .288 r
F P
0 g
s
.3591 .49006 m .32099 .474 L .29598 .44752 L p 0 .494 .305 r
F P
0 g
s
.23274 .41098 m .19524 .41186 L .16865 .38391 L p .55 .44 0 r
F P
0 g
s
.42275 .52842 m .38364 .50969 L .3591 .49006 L p 0 .217 .762 r
F P
0 g
s
.73624 .46608 m .69478 .49338 L .67595 .49691 L p 0 .985 .497 r
F P
0 g
s
.63389 .52452 m .67595 .49691 L .69478 .49338 L p 0 .908 .841 r
F P
0 g
s
.77787 .43768 m .73624 .46608 L .71794 .46342 L p .039 .938 0 r
F P
0 g
s
.67595 .49691 m .71794 .46342 L .73624 .46608 L p 0 .985 .497 r
F P
0 g
s
.59191 .54252 m .55034 .54951 L .52823 .55648 L p .025 0 0 r
F P
0 g
s
.48682 .55011 m .52823 .55648 L .55034 .54951 L p .002 0 0 r
F P
0 g
s
.86249 .38671 m .90586 .36576 L .92249 .38807 L p .723 .282 0 r
F P
0 g
s
.81989 .41076 m .77787 .43768 L .75995 .42853 L p .519 .886 0 r
F P
0 g
s
.71794 .46342 m .75995 .42853 L .77787 .43768 L p .039 .938 0 r
F P
0 g
s
.86249 .38671 m .81989 .41076 L .80225 .39584 L p .782 .732 0 r
F P
0 g
s
.75995 .42853 m .80225 .39584 L .81989 .41076 L p .519 .886 0 r
F P
0 g
s
.23274 .41098 m .27065 .41594 L .29598 .44752 L p .269 .596 0 r
F P
0 g
s
.27065 .41594 m .23274 .41098 L .20659 .38012 L p .427 .61 0 r
F P
0 g
s
.16865 .38391 m .20659 .38012 L .23274 .41098 L p .55 .44 0 r
F P
0 g
s
.29598 .44752 m .33425 .46204 L .3591 .49006 L p 0 .494 .305 r
F P
0 g
s
.33425 .46204 m .29598 .44752 L .27065 .41594 L p 0 .529 .122 r
F P
0 g
s
.20659 .38012 m .16865 .38391 L .1414 .35503 L p .585 .342 0 r
F P
0 g
s
.52823 .55648 m .57057 .55083 L .59191 .54252 L p .025 0 0 r
F P
0 g
s
.63389 .52452 m .59191 .54252 L .57057 .55083 L p .149 0 .744 r
F P
0 g
s
.90586 .36576 m .86249 .38671 L .84511 .36728 L p .745 .409 0 r
F P
0 g
s
.80225 .39584 m .84511 .36728 L .86249 .38671 L p .782 .732 0 r
F P
0 g
s
.42275 .52842 m .46324 .54581 L .48682 .55011 L p .002 0 0 r
F P
0 g
s
.52823 .55648 m .48682 .55011 L .46324 .54581 L p .197 0 0 r
F P
0 g
s
.3591 .49006 m .39833 .51024 L .42275 .52842 L p 0 .217 .762 r
F P
0 g
s
.39833 .51024 m .3591 .49006 L .33425 .46204 L p 0 .288 .878 r
F P
0 g
s
.46324 .54581 m .42275 .52842 L .39833 .51024 L p 0 0 .777 r
F P
0 g
s
.84511 .36728 m .88876 .34315 L .90586 .36576 L p .745 .409 0 r
F P
0 g
s
.67595 .49691 m .63389 .52452 L .61338 .53102 L p .003 0 .865 r
F P
0 g
s
.57057 .55083 m .61338 .53102 L .63389 .52452 L p .149 0 .744 r
F P
0 g
s
.71794 .46342 m .67595 .49691 L .65619 .49889 L p 0 .636 .872 r
F P
0 g
s
.61338 .53102 m .65619 .49889 L .67595 .49691 L p .003 0 .865 r
F P
0 g
s
.20659 .38012 m .24495 .38091 L .27065 .41594 L p .427 .61 0 r
F P
0 g
s
.24495 .38091 m .20659 .38012 L .17984 .34812 L p .55 .549 0 r
F P
0 g
s
.1414 .35503 m .17984 .34812 L .20659 .38012 L p .585 .342 0 r
F P
0 g
s
.27065 .41594 m .30921 .42652 L .33425 .46204 L p 0 .529 .122 r
F P
0 g
s
.30921 .42652 m .27065 .41594 L .24495 .38091 L p .097 .556 0 r
F P
0 g
s
.75995 .42853 m .71794 .46342 L .6988 .4593 L p 0 .925 .482 r
F P
0 g
s
.65619 .49889 m .6988 .4593 L .71794 .46342 L p 0 .636 .872 r
F P
0 g
s
.80225 .39584 m .75995 .42853 L .74128 .41806 L p .181 .88 0 r
F P
0 g
s
.6988 .4593 m .74128 .41806 L .75995 .42853 L p 0 .925 .482 r
F P
0 g
s
.17984 .34812 m .1414 .35503 L .11339 .32575 L p .612 .244 0 r
F P
0 g
s
.84511 .36728 m .80225 .39584 L .78394 .37976 L p .692 .842 0 r
F P
0 g
s
.74128 .41806 m .78394 .37976 L .80225 .39584 L p .181 .88 0 r
F P
0 g
s
.46324 .54581 m .50513 .55511 L .52823 .55648 L p .197 0 0 r
F P
0 g
s
.57057 .55083 m .52823 .55648 L .50513 .55511 L p .166 0 .61 r
F P
0 g
s
.33425 .46204 m .37362 .48048 L .39833 .51024 L p 0 .288 .878 r
F P
0 g
s
.37362 .48048 m .33425 .46204 L .30921 .42652 L p 0 .459 .866 r
F P
0 g
s
.88876 .34315 m .84511 .36728 L .82713 .34684 L p .762 .538 0 r
F P
0 g
s
.78394 .37976 m .82713 .34684 L .84511 .36728 L p .692 .842 0 r
F P
0 g
s
.39833 .51024 m .43901 .52905 L .46324 .54581 L p 0 0 .777 r
F P
0 g
s
.43901 .52905 m .39833 .51024 L .37362 .48048 L p .074 0 .89 r
F P
0 g
s
.50513 .55511 m .46324 .54581 L .43901 .52905 L p .1 0 .691 r
F P
0 g
s
.82713 .34684 m .87112 .31963 L .88876 .34315 L p .762 .538 0 r
F P
0 g
s
.17984 .34812 m .21873 .34446 L .24495 .38091 L p .55 .549 0 r
F P
0 g
s
.21873 .34446 m .17984 .34812 L .15235 .316 L p .589 .398 0 r
F P
0 g
s
.11339 .32575 m .15235 .316 L .17984 .34812 L p .612 .244 0 r
F P
0 g
s
.24495 .38091 m .28392 .38607 L .30921 .42652 L p .097 .556 0 r
F P
0 g
s
.28392 .38607 m .24495 .38091 L .21873 .34446 L p .331 .584 0 r
F P
0 g
s
.61338 .53102 m .57057 .55083 L .54814 .55037 L p .625 0 .71 r
F P
0 g
s
.50513 .55511 m .54814 .55037 L .57057 .55083 L p .166 0 .61 r
F P
0 g
s
.15235 .316 m .11339 .32575 L .08454 .29637 L p .632 .159 0 r
F P
0 g
s
.30921 .42652 m .3488 .4404 L .37362 .48048 L p 0 .459 .866 r
F P
0 g
s
.3488 .4404 m .30921 .42652 L .28392 .38607 L p 0 .517 .468 r
F P
0 g
s
.65619 .49889 m .61338 .53102 L .59172 .52909 L p .652 0 .899 r
F P
0 g
s
.54814 .55037 m .59172 .52909 L .61338 .53102 L p .625 0 .71 r
F P
0 g
s
.78394 .37976 m .74128 .41806 L .72165 .40297 L p 0 .855 .115 r
F P
0 g
s
.74128 .41806 m .6988 .4593 L .67859 .4491 L p 0 .832 .839 r
F P
0 g
s
.67859 .4491 m .72165 .40297 L .74128 .41806 L p 0 .832 .839 r
F P
0 g
s
.82713 .34684 m .78394 .37976 L .76479 .36038 L p .518 .829 0 r
F P
0 g
s
.72165 .40297 m .76479 .36038 L .78394 .37976 L p 0 .855 .115 r
F P
0 g
s
.6988 .4593 m .65619 .49889 L .6353 .49341 L p .004 0 .855 r
F P
0 g
s
.6353 .49341 m .67859 .4491 L .6988 .4593 L p .004 0 .855 r
F P
0 g
s
.59172 .52909 m .6353 .49341 L .65619 .49889 L p .652 0 .899 r
F P
0 g
s
.37362 .48048 m .41444 .49755 L .43901 .52905 L p .074 0 .89 r
F P
0 g
s
.41444 .49755 m .37362 .48048 L .3488 .4404 L p .002 0 .849 r
F P
0 g
s
.87112 .31963 m .82713 .34684 L .80842 .3241 L p .772 .64 0 r
F P
0 g
s
.76479 .36038 m .80842 .3241 L .82713 .34684 L p .518 .829 0 r
F P
0 g
s
.21873 .34446 m .25822 .34386 L .28392 .38607 L p .331 .584 0 r
F P
0 g
s
.25822 .34386 m .21873 .34446 L .19181 .3083 L p .57 .588 0 r
F P
0 g
s
.15235 .316 m .19181 .3083 L .21873 .34446 L p .589 .398 0 r
F P
0 g
s
.19181 .3083 m .15235 .316 L .12401 .28429 L p .618 .264 0 r
F P
0 g
s
.08454 .29637 m .12401 .28429 L .15235 .316 L p .632 .159 0 r
F P
0 g
s
.54814 .55037 m .50513 .55511 L .48124 .53932 L p .48 0 .266 r
F P
0 g
s
.43901 .52905 m .48124 .53932 L .50513 .55511 L p .1 0 .691 r
F P
0 g
s
.28392 .38607 m .32384 .39373 L .3488 .4404 L p 0 .517 .468 r
F P
0 g
s
.32384 .39373 m .28392 .38607 L .25822 .34386 L p .024 .548 0 r
F P
0 g
s
.80842 .3241 m .85284 .29448 L .87112 .31963 L p .772 .64 0 r
F P
0 g
s
.48124 .53932 m .43901 .52905 L .41444 .49755 L p .377 0 .692 r
F P
0 g
s
.12401 .28429 m .08454 .29637 L .05478 .26692 L p .645 .095 0 r
F P
0 g
s
.3488 .4404 m .38976 .45301 L .41444 .49755 L p .002 0 .849 r
F P
0 g
s
.38976 .45301 m .3488 .4404 L .32384 .39373 L p 0 .332 .814 r
F P
0 g
s
.19181 .3083 m .23186 .30252 L .25822 .34386 L p .57 .588 0 r
F P
0 g
s
.23186 .30252 m .19181 .3083 L .16402 .27335 L p .605 .396 0 r
F P
0 g
s
.12401 .28429 m .16402 .27335 L .19181 .3083 L p .618 .264 0 r
F P
0 g
s
.59172 .52909 m .54814 .55037 L .52474 .53473 L p .726 0 .251 r
F P
0 g
s
.48124 .53932 m .52474 .53473 L .54814 .55037 L p .48 0 .266 r
F P
0 g
s
.25822 .34386 m .29857 .34495 L .32384 .39373 L p .024 .548 0 r
F P
0 g
s
.29857 .34495 m .25822 .34386 L .23186 .30252 L p .357 .585 0 r
F P
0 g
s
.80842 .3241 m .76479 .36038 L .74468 .33581 L p .413 .807 0 r
F P
0 g
s
.76479 .36038 m .72165 .40297 L .70093 .3804 L p 0 .818 .278 r
F P
0 g
s
.70093 .3804 m .74468 .33581 L .76479 .36038 L p 0 .818 .278 r
F P
0 g
s
.16402 .27335 m .12401 .28429 L .09473 .25309 L p .637 .159 0 r
F P
0 g
s
.05478 .26692 m .09473 .25309 L .12401 .28429 L p .645 .095 0 r
F P
0 g
s
.72165 .40297 m .67859 .4491 L .6572 .42876 L p 0 .541 .816 r
F P
0 g
s
.6572 .42876 m .70093 .3804 L .72165 .40297 L p 0 .541 .816 r
F P
0 g
s
.85284 .29448 m .80842 .3241 L .78888 .29793 L p .773 .69 0 r
F P
0 g
s
.74468 .33581 m .78888 .29793 L .80842 .3241 L p .413 .807 0 r
F P
0 g
s
.41444 .49755 m .45686 .50642 L .48124 .53932 L p .377 0 .692 r
F P
0 g
s
.45686 .50642 m .41444 .49755 L .38976 .45301 L p .381 0 .835 r
F P
0 g
s
.32384 .39373 m .36501 .40025 L .38976 .45301 L p 0 .332 .814 r
F P
0 g
s
.36501 .40025 m .32384 .39373 L .29857 .34495 L p 0 .531 .447 r
F P
0 g
s
.67859 .4491 m .6353 .49341 L .61321 .47523 L p .326 0 .828 r
F P
0 g
s
.61321 .47523 m .6572 .42876 L .67859 .4491 L p .326 0 .828 r
F P
0 g
s
.6353 .49341 m .59172 .52909 L .56893 .5126 L p .825 0 .659 r
F P
0 g
s
.52474 .53473 m .56893 .5126 L .59172 .52909 L p .726 0 .251 r
F P
0 g
s
.52474 .53473 m .48124 .53932 L .45686 .50642 L p .573 0 .239 r
F P
0 g
s
.56893 .5126 m .61321 .47523 L .6353 .49341 L p .825 0 .659 r
F P
0 g
s
.78888 .29793 m .83384 .26708 L .85284 .29448 L p .773 .69 0 r
F P
0 g
s
.23186 .30252 m .27271 .29773 L .29857 .34495 L p .357 .585 0 r
F P
0 g
s
.27271 .29773 m .23186 .30252 L .20465 .26346 L p .597 .544 0 r
F P
0 g
s
.16402 .27335 m .20465 .26346 L .23186 .30252 L p .605 .396 0 r
F P
0 g
s
.20465 .26346 m .16402 .27335 L .13526 .23978 L p .629 .242 0 r
F P
0 g
s
.09473 .25309 m .13526 .23978 L .16402 .27335 L p .637 .159 0 r
F P
0 g
s
.29857 .34495 m .34002 .34508 L .36501 .40025 L p 0 .531 .447 r
F P
0 g
s
.34002 .34508 m .29857 .34495 L .27271 .29773 L p .122 .565 0 r
F P
0 g
s
.38976 .45301 m .43227 .45841 L .45686 .50642 L p .381 0 .835 r
F P
0 g
s
.43227 .45841 m .38976 .45301 L .36501 .40025 L p .002 0 .795 r
F P
0 g
s
.56893 .5126 m .52474 .53473 L .50064 .50103 L p .707 0 .228 r
F P
0 g
s
.45686 .50642 m .50064 .50103 L .52474 .53473 L p .573 0 .239 r
F P
0 g
s
.78888 .29793 m .74468 .33581 L .72361 .30526 L p .405 .781 0 r
F P
0 g
s
.74468 .33581 m .70093 .3804 L .67915 .34907 L p 0 .779 .274 r
F P
0 g
s
.67915 .34907 m .72361 .30526 L .74468 .33581 L p 0 .779 .274 r
F P
0 g
s
.83384 .26708 m .78888 .29793 L .76849 .26783 L p .765 .676 0 r
F P
0 g
s
.72361 .30526 m .76849 .26783 L .78888 .29793 L p .405 .781 0 r
F P
0 g
s
.50064 .50103 m .45686 .50642 L .43227 .45841 L p .589 0 .62 r
F P
0 g
s
.70093 .3804 m .6572 .42876 L .63472 .39647 L p 0 .504 .798 r
F P
0 g
s
.63472 .39647 m .67915 .34907 L .70093 .3804 L p 0 .504 .798 r
F P
0 g
s
.36501 .40025 m .40759 .40098 L .43227 .45841 L p .002 0 .795 r
F P
0 g
s
.40759 .40098 m .36501 .40025 L .34002 .34508 L p 0 .503 .784 r
F P
0 g
s
.20465 .26346 m .24602 .25405 L .27271 .29773 L p .597 .544 0 r
F P
0 g
s
.24602 .25405 m .20465 .26346 L .17643 .22692 L p .624 .337 0 r
F P
0 g
s
.13526 .23978 m .17643 .22692 L .20465 .26346 L p .629 .242 0 r
F P
0 g
s
.27271 .29773 m .31453 .29219 L .34002 .34508 L p .122 .565 0 r
F P
0 g
s
.31453 .29219 m .27271 .29773 L .24602 .25405 L p .515 .606 0 r
F P
0 g
s
.6572 .42876 m .61321 .47523 L .59007 .44203 L p .311 0 .8 r
F P
0 g
s
.59007 .44203 m .63472 .39647 L .6572 .42876 L p .311 0 .8 r
F P
0 g
s
.76849 .26783 m .81411 .23715 L .83384 .26708 L p .765 .676 0 r
F P
0 g
s
.61321 .47523 m .56893 .5126 L .54524 .47884 L p .782 0 .622 r
F P
0 g
s
.50064 .50103 m .54524 .47884 L .56893 .5126 L p .707 0 .228 r
F P
0 g
s
.54524 .47884 m .59007 .44203 L .61321 .47523 L p .782 0 .622 r
F P
0 g
s
.34002 .34508 m .38273 .34094 L .40759 .40098 L p 0 .503 .784 r
F P
0 g
s
.38273 .34094 m .34002 .34508 L .31453 .29219 L p 0 .561 .108 r
F P
0 g
s
.43227 .45841 m .47616 .45141 L .50064 .50103 L p .589 0 .62 r
F P
0 g
s
.47616 .45141 m .43227 .45841 L .40759 .40098 L p .227 0 .78 r
F P
0 g
s
.54524 .47884 m .50064 .50103 L .47616 .45141 L p .677 0 .605 r
F P
0 g
s
.24602 .25405 m .28826 .24405 L .31453 .29219 L p .515 .606 0 r
F P
0 g
s
.28826 .24405 m .24602 .25405 L .21831 .2142 L p .624 .431 0 r
F P
0 g
s
.17643 .22692 m .21831 .2142 L .24602 .25405 L p .624 .337 0 r
F P
0 g
s
.76849 .26783 m .72361 .30526 L .70169 .26928 L p .494 .757 0 r
F P
0 g
s
.72361 .30526 m .67915 .34907 L .65648 .30991 L p 0 .745 .11 r
F P
0 g
s
.65648 .30991 m .70169 .26928 L .72361 .30526 L p 0 .745 .11 r
F P
0 g
s
.40759 .40098 m .4515 .39188 L .47616 .45141 L p .227 0 .78 r
F P
0 g
s
.4515 .39188 m .40759 .40098 L .38273 .34094 L p 0 .362 .772 r
F P
0 g
s
.31453 .29219 m .35745 .28373 L .38273 .34094 L p 0 .561 .108 r
F P
0 g
s
.35745 .28373 m .31453 .29219 L .28826 .24405 L p .388 .598 0 r
F P
0 g
s
.67915 .34907 m .63472 .39647 L .61137 .35357 L p 0 .683 .789 r
F P
0 g
s
.61137 .35357 m .65648 .30991 L .67915 .34907 L p 0 .683 .789 r
F P
0 g
s
.81411 .23715 m .76849 .26783 L .74732 .23414 L p .75 .606 0 r
F P
0 g
s
.70169 .26928 m .74732 .23414 L .76849 .26783 L p .494 .757 0 r
F P
0 g
s
.63472 .39647 m .59007 .44203 L .56618 .39552 L p .004 0 .78 r
F P
0 g
s
.56618 .39552 m .61137 .35357 L .63472 .39647 L p .004 0 .78 r
F P
0 g
s
.59007 .44203 m .54524 .47884 L .52098 .42982 L p .585 0 .784 r
F P
0 g
s
.47616 .45141 m .52098 .42982 L .54524 .47884 L p .677 0 .605 r
F P
0 g
s
.52098 .42982 m .56618 .39552 L .59007 .44203 L p .585 0 .784 r
F P
0 g
s
.74732 .23414 m .79365 .20483 L .81411 .23715 L p .75 .606 0 r
F P
0 g
s
.52098 .42982 m .47616 .45141 L .4515 .39188 L p .25 0 .771 r
F P
0 g
s
.38273 .34094 m .42666 .32968 L .4515 .39188 L p 0 .362 .772 r
F P
0 g
s
.42666 .32968 m .38273 .34094 L .35745 .28373 L p 0 .571 .264 r
F P
0 g
s
.28826 .24405 m .33146 .23219 L .35745 .28373 L p .388 .598 0 r
F P
0 g
s
.33146 .23219 m .28826 .24405 L .261 .20102 L p .629 .509 0 r
F P
0 g
s
.21831 .2142 m .261 .20102 L .28826 .24405 L p .624 .431 0 r
F P
0 g
s
.4515 .39188 m .49639 .37128 L .52098 .42982 L p .25 0 .771 r
F P
0 g
s
.49639 .37128 m .4515 .39188 L .42666 .32968 L p 0 .38 .767 r
F P
0 g
s
.56618 .39552 m .52098 .42982 L .49639 .37128 L p .003 0 .769 r
F P
0 g
s
.35745 .28373 m .40145 .27051 L .42666 .32968 L p 0 .571 .264 r
F P
0 g
s
.40145 .27051 m .35745 .28373 L .33146 .23219 L p .319 .602 0 r
F P
0 g
s
.70169 .26928 m .65648 .30991 L .63312 .26558 L p .164 .721 0 r
F P
0 g
s
.65648 .30991 m .61137 .35357 L .58742 .30384 L p 0 .695 .439 r
F P
0 g
s
.58742 .30384 m .63312 .26558 L .65648 .30991 L p 0 .695 .439 r
F P
0 g
s
.74732 .23414 m .70169 .26928 L .67905 .22959 L p .648 .739 0 r
F P
0 g
s
.63312 .26558 m .67905 .22959 L .70169 .26928 L p .164 .721 0 r
F P
0 g
s
.61137 .35357 m .56618 .39552 L .54181 .34063 L p 0 .434 .774 r
F P
0 g
s
.54181 .34063 m .58742 .30384 L .61137 .35357 L p 0 .434 .774 r
F P
0 g
s
.49639 .37128 m .54181 .34063 L .56618 .39552 L p .003 0 .769 r
F P
0 g
s
.42666 .32968 m .47156 .31011 L .49639 .37128 L p 0 .38 .767 r
F P
0 g
s
.47156 .31011 m .42666 .32968 L .40145 .27051 L p 0 .592 .263 r
F P
0 g
s
.79365 .20483 m .74732 .23414 L .72544 .19785 L p .733 .498 0 r
F P
0 g
s
.67905 .22959 m .72544 .19785 L .74732 .23414 L p .648 .739 0 r
F P
0 g
s
.54181 .34063 m .49639 .37128 L .47156 .31011 L p 0 .578 .77 r
F P
0 g
s
.72544 .19785 m .77253 .17066 L .79365 .20483 L p .733 .498 0 r
F P
0 g
s
.33146 .23219 m .37564 .21735 L .40145 .27051 L p .319 .602 0 r
F P
0 g
s
.37564 .21735 m .33146 .23219 L .30456 .18668 L p .639 .558 0 r
F P
0 g
s
.261 .20102 m .30456 .18668 L .33146 .23219 L p .629 .509 0 r
F P
0 g
s
.58742 .30384 m .54181 .34063 L .5171 .28326 L p 0 .648 .435 r
F P
0 g
s
.47156 .31011 m .5171 .28326 L .54181 .34063 L p 0 .578 .77 r
F P
0 g
s
.40145 .27051 m .44639 .25176 L .47156 .31011 L p 0 .592 .263 r
F P
0 g
s
.44639 .25176 m .40145 .27051 L .37564 .21735 L p .329 .617 0 r
F P
0 g
s
.5171 .28326 m .47156 .31011 L .44639 .25176 L p 0 .619 .107 r
F P
0 g
s
.63312 .26558 m .58742 .30384 L .56303 .2519 L p .034 .68 0 r
F P
0 g
s
.5171 .28326 m .56303 .2519 L .58742 .30384 L p 0 .648 .435 r
F P
0 g
s
.67905 .22959 m .63312 .26558 L .60925 .21934 L p .469 .708 0 r
F P
0 g
s
.56303 .2519 m .60925 .21934 L .63312 .26558 L p .034 .68 0 r
F P
0 g
s
.72544 .19785 m .67905 .22959 L .65584 .18829 L p .73 .634 0 r
F P
0 g
s
.60925 .21934 m .65584 .18829 L .67905 .22959 L p .469 .708 0 r
F P
0 g
s
.56303 .2519 m .5171 .28326 L .49206 .22816 L p .149 .649 0 r
F P
0 g
s
.44639 .25176 m .49206 .22816 L .5171 .28326 L p 0 .619 .107 r
F P
0 g
s
.49206 .22816 m .44639 .25176 L .42071 .1991 L p .425 .638 0 r
F P
0 g
s
.37564 .21735 m .42071 .1991 L .44639 .25176 L p .329 .617 0 r
F P
0 g
s
.77253 .17066 m .72544 .19785 L .70295 .16019 L p .714 .376 0 r
F P
0 g
s
.65584 .18829 m .70295 .16019 L .72544 .19785 L p .73 .634 0 r
F P
0 g
s
.42071 .1991 m .37564 .21735 L .349 .17058 L p .652 .566 0 r
F P
0 g
s
.30456 .18668 m .349 .17058 L .37564 .21735 L p .639 .558 0 r
F P
0 g
s
.60925 .21934 m .56303 .2519 L .53826 .20147 L p .452 .679 0 r
F P
0 g
s
.49206 .22816 m .53826 .20147 L .56303 .2519 L p .149 .649 0 r
F P
0 g
s
.65584 .18829 m .60925 .21934 L .58493 .17386 L p .714 .671 0 r
F P
0 g
s
.53826 .20147 m .58493 .17386 L .60925 .21934 L p .452 .679 0 r
F P
0 g
s
.70295 .16019 m .75078 .13528 L .77253 .17066 L p .714 .376 0 r
F P
0 g
s
.53826 .20147 m .49206 .22816 L .46654 .17781 L p .588 .663 0 r
F P
0 g
s
.42071 .1991 m .46654 .17781 L .49206 .22816 L p .425 .638 0 r
F P
0 g
s
.46654 .17781 m .42071 .1991 L .39429 .15247 L p .665 .529 0 r
F P
0 g
s
.349 .17058 m .39429 .15247 L .42071 .1991 L p .652 .566 0 r
F P
0 g
s
.70295 .16019 m .65584 .18829 L .6321 .14711 L p .709 .44 0 r
F P
0 g
s
.58493 .17386 m .6321 .14711 L .65584 .18829 L p .714 .671 0 r
F P
0 g
s
.58493 .17386 m .53826 .20147 L .51302 .15455 L p .693 .599 0 r
F P
0 g
s
.46654 .17781 m .51302 .15455 L .53826 .20147 L p .588 .663 0 r
F P
0 g
s
.75078 .13528 m .70295 .16019 L .67988 .12218 L p .697 .261 0 r
F P
0 g
s
.6321 .14711 m .67988 .12218 L .70295 .16019 L p .709 .44 0 r
F P
0 g
s
.51302 .15455 m .46654 .17781 L .44037 .13256 L p .675 .455 0 r
F P
0 g
s
.39429 .15247 m .44037 .13256 L .46654 .17781 L p .665 .529 0 r
F P
0 g
s
.6321 .14711 m .58493 .17386 L .56011 .13059 L p .696 .431 0 r
F P
0 g
s
.51302 .15455 m .56011 .13059 L .58493 .17386 L p .693 .599 0 r
F P
0 g
s
.67988 .12218 m .72841 .09924 L .75078 .13528 L p .697 .261 0 r
F P
0 g
s
.67988 .12218 m .6321 .14711 L .60782 .10701 L p .691 .282 0 r
F P
0 g
s
.56011 .13059 m .60782 .10701 L .6321 .14711 L p .696 .431 0 r
F P
0 g
s
.56011 .13059 m .51302 .15455 L .48718 .11142 L p .681 .357 0 r
F P
0 g
s
.44037 .13256 m .48718 .11142 L .51302 .15455 L p .675 .455 0 r
F P
0 g
s
.72841 .09924 m .67988 .12218 L .65623 .08438 L p .684 .167 0 r
F P
0 g
s
.60782 .10701 m .65623 .08438 L .67988 .12218 L p .691 .282 0 r
F P
0 g
s
.60782 .10701 m .56011 .13059 L .53469 .08975 L p .682 .255 0 r
F P
0 g
s
.48718 .11142 m .53469 .08975 L .56011 .13059 L p .681 .357 0 r
F P
0 g
s
.65623 .08438 m .70542 .06282 L .72841 .09924 L p .684 .167 0 r
F P
0 g
s
.65623 .08438 m .60782 .10701 L .58293 .06812 L p .679 .165 0 r
F P
0 g
s
.53469 .08975 m .58293 .06812 L .60782 .10701 L p .682 .255 0 r
F P
0 g
s
.70542 .06282 m .65623 .08438 L .63193 .04687 L p .674 .097 0 r
F P
0 g
s
.58293 .06812 m .63193 .04687 L .65623 .08438 L p .679 .165 0 r
F P
0 g
s
.63193 .04687 m .68175 .02605 L .70542 .06282 L p .674 .097 0 r
F P
0 g
s
.25 Mabswid
.69069 0 m
.97431 .43047 L
s
.97431 .43047 m
1 .61842 L
s
1 .61842 m
.70263 .20566 L
s
.70263 .20566 m
.69069 0 L
s
.03114 .256 m
0 .45252 L
s
0 .45252 m
.70263 .20566 L
s
.70263 .20566 m
.69069 0 L
s
.69069 0 m
.03114 .256 L
s
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[21]:=",
  ImageSize->{417.312, 198.688},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006Q0000aPP10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l7XPL00?l7XPL00?l7XPL00?l7XPL00?l7
XPL00?l7XPL00?l7XPL00?l7XPL00?l7XPL00?l7XPL000h7M@002XVRXj:jX[ZR^X@e0=l7000>1g@0
00^jXkZS^Z>jXkZS^P0e0=l7000>1g4000m[Q;ZR^Z:SXZ>RXj:jX[X0=03O1`003PM_000BJknS^Z>j
XkZS^Z>jXkZS^Z=`<`3O1`003PM^000CXZRRZ::SXZ>RXj:SXZ>RXj:SJ`0c0=l7000>1f/001I[Xj>o
XknS_j>jXkZS^Z>jXkZS^Z>j<`3O1`003PMY000IRJ:XXZRRZ::SXZRRXj:SX[ZRXj:SXZ>RR@0b0=l7
000>1fP001ZoXknS_j>oXknS_j>jXkZS^Z>jXkZS^Z>oJc80g`L000h7I@007F^9Z::XXZ>RXj:SXZRR
Z::SXZ>RXj:SXZ>RZ::X0380g`L000h7H`0086^oZ;nS_j>oXknS_j>oXknS_j>jXkZS^Z>jXknS_j>o
<@3O1`003PMR000RJjRRZ::XXZRRXj:SXZ>RZ::XXZ>RXj:SXZ>RZ::XXZRRRC00g`L000h7H00098ZX
_jRoZ;nX_j>oXknS_j>oXknS_j>jXknS_j>oXknS_j>oJc00g`L000h7G@02J`0UZ::XXZRRZ::XXZRR
Xj:SXZ>RZ::XXZRRXj:SXZ>RZ::XXZRRX`0`0=l7000>1e`002V:XknS_jRoXknS_jRoXknS_j>oXknS
_j>oZ;nS_j>oXknS_j>oXknSRP0_0=l7000>1eX002]`Z::XXZRRZ::XXZRRZ::XXZRRZ::XXZ>RZ::X
XZRRZ::XXZRRZ::XXZR902l0g`L000h7F000;HZ9_jRoZ;nX_jRoZ;nX_jRoZ;nS_j>oXknX_jRoZ;nX
_j>oXknS_j>oXknS_`0_09P70P151`003PMG000_Z:NXYjRWZ::XXZRRZ::XXZRRZ::XXZRRZ::XXZRR
Z::XXZRRZ::XXZRRZ::XXZP0;P2E1`<000<700L0A0L000h7E002L00a_jRoZ;nX_jRoZ;nX_jRoZ;nX
_jRoXknS_jRoZ;nX_jRoZ;nS_j>oXknS_j>oXknXL00]09<70P041`80A0L000h7DP00=F^?YjRWZ:NX
YjRWZ:NXXZRRZ::XXZRRZ:NXXZRRZ::XXZRRZ::XXZRRZ::XXZRRZ::XXZR902d0T0L300H700<01`00
@`L000h7D000=hZX_jRoZ;nX_jRoZ;nX_jRoZ;nX_jRoZ;nXa:RoZ;nX_jRoZ;nX_jRoZ;nS_j>oXknX
_jRoZ;l0;@2>1`802@L00`0700131`003PM?000iZ:NXYjRWZ:NXYjRWZ:NXYjRRZ::XXZRRZ:NXYjRW
Z::XXZRRZ::XXZRRZ::SXZRRZ::XXZRRZ::902`0R`L300P70P001@L01`L00487000>1dd003]`a:S4
Z<BX_jRoZ;nX_jRoZ;nX_jRoZ;nXa:S4Z<BXa:RoZ;nX_jRoZ;nX_jRoXknX_jRoZ;nX_jRoR@0/08T7
0P091`8000BR00L00`L00`071`0o1`003PM;000nSjNXYjRWZ:NXYjRWZ:NXYjRWZ:NXYjRRZ:NXYjRW
Z:NXYknW_j:XXZRRZ::oXZRWZ::XXZRRZ::XXZRRZ6/[08H70`081`<000B4WHBM0P031`0300L703l7
000>1dT0042X[JS4Z<BXa:S4Z<BX_jRoZ;nX_jRoZ;nXa:S4Z<BXa:S4Z<BX_jRoZ;nX_jRoZ<BXa:Ro
Z;nX_jRoZ;nX_jRo:`231`<02@L20009WJ:MXYfRWJ8000@700<01`L0?PL000h7A`00@hVWZ:NXYjRW
Z:NXYjRWZ:NXYjRWZ:NXYjRW[JNXYjRWZ:NoYjRWZ:NXX[nR_j:oYjRWZ:NXXZRRZ::XXZRRZ::XXV/0
:P211`802PL2000:WHBMQ9f4WHBMQ08010L00`071`0m1`003PM40016L8j]Z:fXa:S4Z<BXa:S4Z<BX
_jRoZ;nX_jS4[LB]a:S4Z<BXa:S4Z<BXa:RoZ;nX_jS4Z<BXa:S4Z;nX_jRoZ;nX_jRoRBX0OPL300X7
0P0039d0WJ:MXYfRWJ:MXP8010L00`071`0m1`003PM30017[JN]YjfWZ:NXYjRWZ:NXYjRWZ:NXYjRW
Z:N][:f/[JbXYknW_jNoYjRW_jNoX[nW_jNoYjRWZ:NXYjRRZ::XXZRRZ::XYjP0:P1l1`802PL3000?
XXBRQ024WHBMQ9f4WHBM008010L00`071`0l1`003PM20019[JR]Z:fX[JS4Z:fXa:S4Z<BXa:S4Z;nX
a:g4[LB]a:g4Z<BXa:S4Z<BXa:S4Z;nXa:S4Z<BXa:S4Z<BX_jRoZ;nX_jRoZ;nXRP0Y07T70`0:1`80
01FRWJ:MXYfRW@2MXYfRWJ:MXYd0W@0010L00`071`0k1`003PLo001<RJN]YjfW[JN]YjfWZ:NXYjRW
Z:NXYjRWZ:N][:f/[Jb][:f/[JNoYknW_jNoYknWa:c4YknW_jNXYjRWZ:NXXZRRZ::XYjRWZ:NXRBT0
M`L200/70P004hBRQ::4XXBRQ::408BMQ9f4WH@00P000j801`031`0300L703/7000>1ch004f]Z:fX
[JR]Z:fX[JR]Z:fXa:S4Z<BXa:S4[LB]a:g4[LB]a:g4[LBXa:S4Z<BXa:S8[LR]a:S4Z<BXa:S4Z<BX
_jRoZ;nX_jRoZ;nX_`0Y07@70`0;1`8001:R0::MXYfRWJ:MXYfR0::MXYd30005W@2MXP0010L00`07
1`0j1`003PLk001ARHj]YjfW[JN]YjfW[JN]YjfWZ:NXYjRWZ:N][:f/[Jb][:f/[Jb][<B/_jNoYknW
a:c4[<B/a:NoYknW_jNoYjRWZ:NXYjRWZ:NXYjRWZ:NX02P0LPL200/70`004XVRRJ80XXBRQ::4XXBR
Q:80W@8000R4XXBR0::400@700<01`L0>PL000h7>000EG2>[JR]Z:fX[JR]Z:fX[JR]Z<BX[JS4Z<BX
a:fb[K:]/Zg4[LB]a:g4[LB]a:g4Z<BXa:g8[LR]b:g4Z<BXa:S4Z<BXa:S4Z;nX_jRoZ;nX_jRoZ700
9`1_1`<02`L200NR01T0WJ:MXYfRWJ:MXYfR0::MXYfRW@2MXYd000@700<01`L0>@L000h7=`00EZf>
[Hj]YjfW[JN]YjfW[JN]YjfW[JNXYjRW[I>][:f/[Jb][:f/[Jb][<B/a:c4YknWa:c4[<B/a:c4[;nW
_jNoYknW_jN][:fWZ:NXYjRWZ:NXYjQ[9`1]1`8030L2000CXXVRRJ:9XXVRR@24XXBRQ::4XP02000>
XP24XXBRQ:80XXBRQ0041`0300L703P7000>1cD001Y`[JR]Z:fX[JR]Z:fX[JR]Z:fX[JS4Z<BXa0>]
03^b[K:]/Zfb[LB]a:g4[LB]a:g4[LB]b:g8[LR]b:g8[LR]a:S4Z<BXa:g4[LB]a:RoZ;nX_jRoZ;nX
_`0W06X70`0<1`8000>R0:802J801`2RWJ:MXYd00P04XP0<09fRWJ:M09fRWJ8010L00`071`0h1`00
3PLc000RShj]SZf>[Hj]YjfW[JN]YjfW[JN]YjfW[JN][:f/[Jb][0>]03J/[Jf][:f/[Jc4[<B/[I?4
[<B/a:c4[<B/a:c4YknW_jO4[<B/[Jb]YjRWZ:NXYjRWZ:NXYhTV06P70P0<1`<001>WRJN908VRRJ:9
XXVRRJ80XXBR00801Z803@2RQ::4XP2RQ::4XP0010L00`071`0g1`003PL`000NRXn]Sjf?[Hn]Z:fX
[JR]Z:fX[JR]Z:fX[JR]Z:fD1:d0?;:]/Zfb[K:]a:g4[LB]a:g4[K:D/Zg8[LR]b:g8[LR]b:g4Z<BX
a:g4[LB]a:g4Z;nX_jRoZ;nXa:S4L2H0I@L300`70P002JNRYj:WXZNR0009XP0409d000VR00P0XYfR
W@2MXP8000>R00L010L00`071`0f1`003PL^000QL8n>[Hj]SZf>[Hj]SZfW[JN]YjfW[JN]YjfW[HjD
TiBC00>]00>/[Jd00jd00jb][@02[@05a:f][L@00Zd0:Y>bTk:/a:c4[<B/a:c4[<B/_jO4[<B/[Jc4
[:f/[JNXYjRWZ:NXYjfWZ2H0H`L200d70P005XVWRJN9YhVWRJL0XXVRRJ:9XXVRR@0<XP040::4XP<0
1:800`071`021`0300L703H7000>1b/006A`LHn]Sjf?[Hn]SjfX[JR]Z:fX[JR]Z:fX[JR]Z:fD[IB]
U;:]/Zfb[K:]/Zfb[K:]a:g4[K:D/YBbU;:D/Zg8[LR]b:g8[LR]b:g4[LB]a:g4[LB]a:foZ;nX_jS4
Z<BXa:S49@1P1`<03@L2000>Y`2WXZNRYj:WXZNRY`06XP8000>W0:802j800`2M000200JR00<01`L0
0`L00`071`0e1`003PLZ000VSZR>Z8jXSZf>[Hj]SZf>[Hj]YjfW[Hj]YjfW[I>DTiBCU9>DTi@2[@03
U:f]00R]03?4[JfD[I>]Tk:C/Y?4[<B/a:c4[<B//Y>][<B/a:b][<B/[Jb][:RW[JN]YjfW[JN]YhT0
901N1`803@L3000DRJN9Y`2WRJN9YhVWRJN908VRRJ820005XZNRY`003:80102RXP07XP0300L700<7
00<01`L0=0L000h79`00JG1aSjf?[Hn]Sjf?[Hn]SjfX[JR]Z:fX[JR]Z:fX[IB]U:fD[IB]U;:D/Zfb
[K:]/Zfb[K:]b:fbU;:D/YBbU;:D/Zg8[LR]b:g8[K:D/YC4[LB]a:g4[LB]a:g4Z<BXa:S4Z<BXa:S4
L00T05/70`0=1`801jL03P2RYj:WXZNRYj:WXP2R0P07Y`030::R00JR0P001JL0XZ8000RR00<01`L0
0PL00`071`0d1`003PLU000TL8j?SZR>Z8jXSZR>Z8j]SZf>[Hj]SZf>[Hj]Yjf>U9>DTiBC29@:[@0e
U7^IU:fD/YBbTk:C/Y>b[<B/a:bb[;:C/Y>bTlB/a:b][:f/[Jb][:fW[JN]YjfW[JN]YjP0901H1`<0
3PL2000RYhVWRJN9YhVWR@29YhVWRJN9YhVWR@2RYj:WXZNRYj:W00FR0P001j:WXZL0XP002:800`07
1`031`0300L703<7000>1b<001f?Z8nXSjf?[Hn]Sjf?[Hn]Sjf?[Hn]SjfX[JR]Z006U01;[IB]U:fD
[IB]U;:]/Zfb[K:]/ZfINiVD/YBbU;:D/YBbU;:Db:g8[LRD/YBbU;:D/Zg4[LB]a:g4[LR]/Zg4Z<BX
a:S4Z<BXa:R:02<0EPL200l70P000jL0Y`0:Y`080::WXZNRYj820003[02W00VW00<0XZ800Z800`2W
Y`04Y`030:8000VR00<01`L00`L00`071`0b1`003PLP000YJhn9ShV?SXn>ShjXSZR>Z8j]SZR>[Hj]
SZf>[Hj]SYAjU7ZDNYBCU9<029@7[@:D03MkU7^IU;:D/YBbTk:D/Y>bTlB//Y>bTk:C/Y>bTjf/a:c4
[:f/[Jb][:fW[JN]YjRW[JNXYjf>02<0D`L300h70`005JN>Yhh0RJN9YhVWRJN9Y`2WRJN9Y`02000A
YjbW[02WXZNRYj:WXZNR0:800P002:NRYj:WXZNR0P06XP8000>W00L010L00`071`0b1`003PLM000O
L8n?Z8nXSjR?Z8nXSjf?[Hn]Sjf?[Hn]Sjf?[Hn]S`0<U019[IB]U:fD[IBbU;:]/ZfbNiUkVG^INiVD
/YBbU;:D/YBbU;:]/YBbU;:D/YBbU;:Db:g4[LB]/Zfb[K:]/ZS4Z<BXa:S4Z:fXa00S05470P0?1`80
1jL00`2WY`09Y`0<0:NRY`2/YjbW[:L02jL00`2WY`08Y`801:820005XZNRY`001@L00`071`0a1`00
3PLK000^L8V?RHn9ShV?RHn9Shj?SZR>Z8jXSZR>[Hj]SZf>U7FDMIAeU7FDNYAjU7ZDT`ZD0Zd00iAk
N`03N`0gVG^INk:D/YBbU;:C/Y>bTh2C/Y>bTk:C/Y>bTk:/a:c4[:f/[Jb][:f/[JN]YjfW[JN]YjfW
[@0R04h70`0?1`8001N>YhjWSZN>YhjW0:N9YhVWRJN9YhT0R@02000@[:N/YjbW[:L0XZNRYj:WXP80
00j/0:NRYj:WXZNRYj:W00>R00P0Yj:WXZNR00H700<01`L0<0L000h76002L00NZ8ZXSjR?Z8nXSjR?
Z8nXSjf?[Hn]Sjf?[Hn]Sjf?49@0AJfD[IB]U:fD/YBbU9UkVG^INiUkVG^IU;:D/YBbU;:D/W^0Nk:D
/YBbU;:D/YBbU;:]b:fb[K:]/Zfb[K:]a:S4Z<B]a002[@04Z:fXL240C0L201070P000j`0Y`0:Y`03
0:NW00NW0P002jN/YjbW[:N/Yj`000FW0P04[0030:NW00RW10002JNRYj:WXZNR00051`0300L70307
000>1aD00V/0<8n9ShV?RHn9ShV?RHn9ShV?SXn>Shj?SZR>Z8j?SYAeU7FDMIAeU7FDMIAjU7ZDNP^D
045TO7^DNiAkU7^INiUkVIBbU;:D/YBINh1jP7ZbTk:C/Y>bTk:C/Y>D[:f/[Jb][:f/[Jb]YjfW[JN]
YjfW[JN]SP0Q04T70`0@1`8001N>YhjW0:N>YhjWSZN>Yhh0RJN9YhVWR@02000aT`2/YjbW[:N/YjbW
[02WXZL0[:N/YjbW[02WXZNRYj:WXZNR0::WXZNRYj:WXZNR00051`0300L702l7000>1a@002J?RXn:
ShZ?RXn:Z8nXSjR?Z8nXSjR?Z8nXSjf?[Hn]SiB?U8nDSa2D04:]U:fD[IAlNh5kVG^INiUkVG^INiUk
VIBbU;:DVG^0Nh1kP9BbU;:D/YBbU;:DVG^b[K:]/Zfb[K:]/ZfbZ<B]a:P3[@04Z:fX[B40A`L20147
0P001jb>[8j/SP002jL0102WYjL3000ATjbC[02/YjbW[:N/YjbW0:L00P08[0030:NW00JW00l0[02W
XZNRYj:WXZNRY`001PL00`071`0^1`003PLA000fL6^:RHZ9ShV?RHn9ShV?RHn9ShV?RHn9Shj?SXn>
ShjDMIAeU7FDMIAeU7FDMIAeU7FDNYAj1i@0AW]lI7aTO7^DNiAkVG^INiUkVG^bU;:DVG^0Nh1jP7ZI
Tk:C/Y>bTk:CVGZDTjf/[Jb][:f/[Jb]TjfW[JN]YjfW[JN]Yg0P04@70`0A1`8001J>YhjWSZN>Yhh0
SZN>YhjWSZN>Y`2W0P009HjCSY>>Thh0YjbW[:N/YjbW[:L0YjbW[:N/YjbW[:L0XZNRYj800P004:bW
002WXZNRYj:WXZNRY`051`0300L702h7000>1`h0019aL8n:ShZ?RXn:ShZ?RXn:ShX4S`0HZ8nXSjR?
Z8nDSiB?U8nDSiB?U8nDSiB?3Y@0@JfDU7]lNh5kPG^1NiYkVG^INiUkVG^IU;:DP7^0Nh1kP7^0NiVD
/YBbU;:DVG^INiV]/Zfb[K:]/ZfbU;:]a:S400:]00NX[JR]Z:f>0200@PL201470`003J`0[8j/SZb>
[8j/SP002JL2000B[9>/TjbC[9>/T`2W[:N/YjbW0P000i<0[00:[0040:NWY`8001:W[:N/0:`0Yj:W
XZNRYj80Y`061`0300L702d7000>1`d003f:RHZ9RXV:RHZ9RXV?RHn9ShV?RHn9ShV?RHn9ShjDMIAe
U7FDMIAeU7FDMIAeU7FDMIAeU7FDNYAeU7ZD009k04QTO6AlI7aTO6BDNiAkVG^INiUkVG^INh1kP7^0
Nh1jP7Z0N[:C/Y>bTh1jP7ZDNYB/[Jb][:fC[I>DTiBW[JN]YjfW[JN][:dP03l70`0@1`<001NWSZN>
Y`2WSZN>YhjWSZL0YhjWSZN>Y`02000BM@2>ThjCSY>>ThjC0:bW[:N/0P008WFCMI<0[:N/YjbW[:N/
Yj`0Y`2/YjbW[:L0Yj`0Yj:WXZL200>W00<01`L010L00`071`0/1`003PL:0003LG2:00>:01:?RXn:
ShZ?RXn:ShZ?RXn:ShX2S`0HZ8n?SiB?U8nDSiB?U8nDSiB?U8nDSiB?39@0@9ElO7]lNgakPG^1Nh5k
VW^JNiYkVG^INiYlPG^0Nh1kP7^0Nh1kP9BbU;:DVG^INiUkVIBb[K:]/ZfbU;:D/Y@5[@03Z:fX00F]
7`0m1`803`L4000D[8j/SZb>[8j/0:b>[8j/SZb>[003Y`<001BCSY>>09>/TjbC[9>/Tj`0[:N/00JC
00<0[:`02Z`03`2/YjbW[:N/0:bW0::WXP0200JW00<01`L00`L00`071`0/1`003PL7002?L6^:RHZ9
RXV:RHZ9RXV:RHZ9ShV?RHn9ShV?RHn>Shj?SXn>U7FDMIAeU7FDMIAeU7FDMIAeU7FDMIAeU7FDNgaT
O6AlI7aTO6AlI7aTO7^DNiUkVG^INh5UPVF0Nh1kP7^0NX1jP7ZITiUjVGZINYUjU7ZDTjf/[I>DTiBC
U9>DTjfW[JN]YjfW[JN]Yg007P0j1`<03@L4000ISZL0YhjWSZN>YhjWSP2>YhjWSZN>Yhh0SP02000D
TgFCMI=eT`2CSY>>ThjCSY>>0:L2000STgFCMI=eTgD0YjbW[:N/YjbW09<0YjbW[:N/Y`2W[:L0XP00
2:L00`071`041`0300L702/7000>1`@000=aL8X02hX00hn:S`03RP06ShZ?RXn:1Xl06YB?U8nDSiB?
U8nDSiB?U8nDSiB?U8nDSiB?1I@00g]lO005O00HNgakO7^1Nh5kPGb1NiYkVW^JO89lPWb218009G^0
Nh1kP7^0NiUkVG^INiUkVG^INiVD/YBbU;:D/YBbU;:D[JP02Jd00hh0000L03P70P0=1`<001^WSZN>
YhjW0:b>[8j/SZb>[8h0SZb>[8j/SZ`00P005XjCSY>>ThjCSY<0[9>/TjbC[9>/T`0;T`030:b/00B/
0P03T`0=0:N/YjbW[02/YjbW000:Y`0300L700@700<01`L0:PL000h70`00ThZ4RXB:RHZ9RXV:RHZ9
RXV:RHZ9RXV:RHn>Shj?SXn>SgF?SXn>Shj?SXmeSgF?MIAeU7FDMIAeU7FDMIAeNe=lDgaCO6AlI7aT
O6AlI7aTO6AlI9EkU7^IO89UPVF2IH5kP7^0Nh1kP7Z0NX5jVGZINYUjU7ZDNYAjU9>DTiBCU9>DTiBC
U:N]YjfW[JN]Yjf>[@0N03D70`0<1`<001^W0:N>YhjWSZN>Y`2WSZN>YhjWSZL0YhjWSZL00`00:WX0
TgFCMI=eTgFCMI<0ThjCSY>>T`1j09=eTgFCMI=eTgFC0:bW[:N/00FC00H0[:N/Yj`20005Yj`0[000
2jL00`071`031`0300L702X7000>1`03LHZ:016:00J?RXn:ShX<S`0JU8nDSiB?U8nDSiB?U8nDSiB?
U8nDSiB?U7H=O00GNgakPG^1Nh5lPG^JNiYmPWb2O89lPX401H008G^0Nh1kPGb1NiUkVG^INiUkVG^I
Nk:D/YBbU;:D/YBbU00<[@03S`0001/0<`L200`70`006jN>YhjW0:N>YhjWSZN>08j/SZb>[8j/SZ`0
[003000FTgFCMI<0ThjCSY>>ThjCSP2C[9>/T`8000AjDWX02i<00`2/[00200NC00<0[0000P002:`0
002/Tj`02:L00`2W00061`0300L702T7000>1`1mJhY[RV^:JhY[RV^:L8Y`RW2:L8Y`Sg2?L8m`Sg2?
SXn>Shj?SXmeSgF?SXn>SgF?MIAeSgFDMIAeU7FDMWaCO5=lDgaCO5=lI7aTO6AlI7aTO6AlI7aTO7^2
IH9UPVF2IH9UPVB0Nh1kP7Z0I85UPFF1NYUjVGZINYAjNf@00W/069BCU9>DTiBCU9>DTjfW[JN]SZf>
[Hj]SQd0<0L300/70`007:N9YhVWRJN9Y`2WSZN>YhjWSP2>YhjWSZN>Yhh2000GNWEjMGYeNWD0MI=e
TgFCMI=eTgD0SY<00P00559jDWYB07FCMI=eTgFCMI=e0:L029<30004[:N/Y`8000F/TjbC0007Y`05
0:N/Y`001PL00`071`0X1`003PL00`2:RP0ARP04ShZ?RP>?00>:Shl02Xl06YB?U8nDSiAfU8nDSiB?
U8nDMWafO7MlMWak2g`067]lNh5kO7^1O85mPgf2OH9mPWb2O8:1PPJ001akPWb2O85lVG^INiUkVG^I
Nh1kVIBbU;:D/YBb0i@00iVD[@0:[Ad0;@L300X710007:N908jWSZN>YhjWSP2>YhjWSZN>Y`2/SZb>
[8h3000RM@1eTgFCMI=eTgD0SY>>ThjCSY>>T`2/07YBNU9jDWYB00^C00<0Ti<019<300:C00<0Yj`0
0P0029>/0:bC[9<01:L20006YjbW[:L01@L00`071`0X1`003PL2002FJhY[RV^:JhY[RW2:L8Y`RW2:
L8m`Sg2?L8m`Sg2?MHmeShj?SXmeSgF?SXmeSgF?MHmeU7IkDgaCO5=lDgaCO5=lDgaCO6AlI7aTO6Al
I7aTO6AlIGiWPVF2IH9UPVF2IH9UPG^0NX1TPFF1IH5UPG^INYAjU7YkI7]TNfBDTiBCU9>DTiAjU7ZD
TjfW[Hj]SZf>[Hj]700[1`802PL3000MRJ:9XXVRR@29YhVWRJN9YhT0SZN>YhjWSZL0Yhh00`006GEj
MGYe07EjMGYeNWEj09=eTgFCMI=eTgD00P005WYBNU9jDWYBNP2CMI=eTgFCMI<0F@03T`<01I<03P2/
0:bC[9>/0:bC[9<00ZL02@2/YjbW[:N/00061`0300L702L7000>1`@02XX048n:RXZ?RXn:ShZ?RXn:
ShX:S`0JU8nDSiB?U8nDMYAfNgIkMW]fO7IlMWagO7H=O00GNgakPG^2OX=nPgf2OH9mPWb2O8:1PX40
0h007X62O89lPWb2O85kVG^INh1kP7^0Nh1k/YBbU;:DV@JD1Zd01:R]Sjd2Sa/0:0L300T70`0078T0
RJN9YhVWRJN908jWSZN>YhjW0:N>YhjWSZL4000JMI=eTgFCMI<0TgFCMI=eTgFC09>>ThjCSY<2000=
DP1BNU9jDWYBNU9j0006T`800eT300VC01T0[9>/TjbC09>/Tj`0Y`2/YjbW[:N/Yj`000D700<01`L0
9`L000h71@00U8Y[RV^:JhY`RW2:L8Y`RW2:L8m`Sg2?L8m`SgF?SXmeShj?MHmeSgF?MGIeMU=fDgIC
MU=kDgaCO5=lDgaCO5=lDgaTO6AlIGaTO6AlI7iVOVInIgiUPVF2IH9UPVF2IH5jP6F1IH5UPFF1IH5U
U7ZDNW]TNfAkI7]TNgZDTiBCU9>DTiAjU7Z]SZf>[Hj]SZf>[HhK02H70P0:1`8001^4XXBRQ029XXVR
RJ:9XXT0RJN9YhVWRJL0Yhh0100079=e07EjMGYeNWEjMGX0NWEjMGYeNWD0MI=eTgD2000HFE9IDP1B
NU9jDWYBNU80MI=eTgD0DUUB0`09T`0L05T0TjbC[9>/T`2C[9>/0:bW[:N/YjbW[:N/00D700<01`L0
9PL000h71P0:RP05ShZ?RXl00hX028n:ShZ?RXn:2Hl06GIkMW]fNgIkMW]fNgIkMW]fO7IlMWafO7H0
3G`00g]lO004OP0?Pgj3OX=mPWf2O8:1PX6200>101n2PH:1PWb2O85lPGbINh1kP7^0Nh1kP7^0U9VD
VIBI00ND00>]Z:d00Zd01Xn]Sjf?[A/08`L300T70`006J:9XXVRRJ:908VWRJN9YhVW0:N>YhjWSZL0
1000<WFCMI=eTgD0MI=eTgFCMI=e07FCMI=eTgFCM@2>Thh0DVABI59T07YBNU9jDWYBNU800Y<20004
F@0000=I00<0Ti<019<2000LFGYI09>/TjbC[9<0T`2C0:N/YjbW[:N/Y`2/00H700<01`L09@L000h7
1`00KHY`RW2:L8Y`RW2:L8Y`RW2:L8m`Sg2?L8m`SgF?MHmeSgEfMGIeMWEfMGICMU=fDg]CMU=lDgaC
O5=lDgaCO5=lI7aTO6AlIGeVOVInIWiVOVInIX9UPVF2IH9UPVEmIX5UPFF1IH5UPFF1IH400W/096Ak
I7]TNfAkI7]TU9>DTiAjU7ZDNYAjU8j]SZf>[Hj]SZf>L1X08@L200P710006::408BRQ::4XXBRQ029
XXVRRJ:9XP2WR@@0026CSP1eTgFCMI=eT`1jMGYeNWEjMGYe07EjMGYeNWEj09<00P00559IDUUBFE80
DWYBNU9jDWYBNP2C10001e9IDUUBF@001I<03P1jFGYINP2/TjbC[9>/0P03T`0=0:N/YjbW[:L0YjbW
00051`0300L702D7000>1`P000=aRXX03hX00hn:S`06S`06MXmfSgJ?0gH05W]fMWIkMW]fNgIkMW]f
NgIlMgafO7H<O0In02j3OX=nPgj3OH9mPWf2PH:1OWenOH:1PWb2O89lPGb1O89kP7^0Nh1kP7^0Nh5k
0Y@00iVDU008U00;Z:f?[Hn]Sjf?[Hl06P0N1`<01`L3000HQ::4XXBRQ029XXVRRJ:9XXT0RJN9YhVW
10008hjCSY>>Thh0MI=eTgFCMI<0TgFCMI=eTgFC09=eTgFCMI=e008001ATDVABI59TDVAB059jDWYB
NU9jDP<02UT00`2CT`02000INUUjFGYINP2/TjbC[02H09RCV9<0YjbW[0020006[<>/`j`01@L00`07
1`0T1`003PL9002BLG2:L8Y`RW2:L8Y`RW2:L8m`SgF?MHmeSgF?MHmeSgF?MGIeMWEfMGIeMU=fDgIC
MU=kDgaCO5=lDgaCO5=lDgaTOEEiEGUVOVInIWiVOVInIh=UPVF2IH9UOFInEGiUPFF1IH5UPFF1IGeV
O6AkI7]TNfAkI7]TO7^DNYAjU7ZDNYAjU7ZDSZf>[HjXSZf>Z70I01d70P061`<001R408BRQ::4XXBR
Q024XXBRQ::4XX@0RJ83000TYhh0SY>>ThjCSY<0TgFCMI=eTgD0MGYeNWEjMGYeNP1jMGYe0`0055<0
DUUBFE9IDUUBF@1jDWYBNU9j0P007E<0DUUBFE9IDUUBFE80T`1INUUjFGYINUUj0:bC00800i<04@2C
Ti<0[:N/0:bW[:N/Yj`000H700<01`L08`L000h72P000g6:RP07RP08ShZ?RXn:ShX7S`08MXmfSgJ?
MXl3MP0HNgIkMW]fNgIkMW]fNgIlMWafO7IlMgag0W`027egNGQnN7ii17h028=nPgj3OX=n0h802862
OGimOWen0Wd06gb2O89lPGb1O8=mPWb0Nh1kP7^0Nh5kPG]lN`0<U00;[Hn]Sjf?[Hn]Sjd06@0M1`03
00L000870`006HBRQ::408BRQ::4XXBRQ029XXVRRJ:908T00P009:b>[8j/SP2>ThjCSY>>T`2CMI=e
TgFCM@1eTgFCMI=eTgD0M@<001ACI5=TD`1BI59TDVABI580DWYBNP@000ADIE@02eT09P1INUUjFGYI
NUUjF@2C09>HTiRCV02HTiP0[033[<>/`jc3[<<01@L00`071`0S1`003PL;009`08n:L8Y`RW2:L8m`
Sg2?L8meSgF?MHmeSgF?MHmeSgF?MGIeMWEfMGICMU=fDgICMU=gDgaCO5=lDgaCOEEhEGUENEEiEGiV
OVInIWiVOVJ3IX9UPVInEGiEOUEmEH5UPFF1IH5UOFImIWeTNfAkI7]TNfAlI7aTNgZDNYAjU7ZDNYAj
U8jXSZR>Z8j?SZR>L00H01d700<01`L00P005XBMQ9f4WHBMQ024XXBRQ::4XX@0Q:84000VRJL0YhjW
SZN>Yhh0SY>>ThjCSY<0TgFCMI=eT`1jMGYeNWEjMGX200QC00/0FE9IDUUBFE9IDP04000BNP1UDfEC
I@1IDUUBFE9IDUUB0P003GYINUUjFGYINUUjF@001i<00`2CT`02T`0=0:bW[:N/YjbW[:N/00051`03
00L70287000>1``02XX01Xn:ShZ?RPN?00QfSgJ?MXmfS`Uf01akMW]fNgIkMWafO7IlMWagOGQiN7Uh
NGQnN7ii17h02H=nPgj3OX=nP`02PP0UOGinOWenOGimOWb2O89lPGb2OH9mPWf2Nh1kP7^0Nh5kO7]l
N`0=U00;Sjf?[Hn]Sjf?[Hl0600M1`0I00L71`070::4XXBRQ::408BRQ::4XX@0Q002000WYhVWRJN9
Y`2/SZb>[8j/09>>ThjCSY>>07FCMI=eTgFC09=eTgFC00<001=T06ACI5=TDfACI01TDVABI59T00<0
00ajDP00IEAUE6EDI@08F@0>06EU07YINUUjFGYINUT2000IV9>HTiRCV9<0T`1j0:c3[<>/`jc3[02/
00061`0300L70247000>1`h008m[RW2:L8Y`RW2?L8m`Sg2?L8m`Sg2?L8meMWEfMHmeMWEfMGIeMWEf
DgICMU=fDgaCMeAgEGQEN5EhEGUENEEiEGUEOVInIWiVOVInIX=WPeEnEGiEOUEnEGiEOFF1IH5UOFIm
IWeVOFIlI7]TO6AlI7aTO6AlNiAjU7ZDNYAjU7FDMJR>Shj?SXn>Shj?L00G01`700<01`L00`L03P07
09f4WHBMQ9f408BR1000:8VR0:N9YhVWRJN9Y`2WSZN>YhjW09>>ThjCSY>>07FCMI=eTgD0MGX300EC
00<0De<01U<01@1BFE9I00<002EIDP1B05=UDfECIE=UD`1BFE9IDUUB06AUI01INUUjFGYINUT0009I
00<0Ti<01Y<03@1jTgX0YjbW[:N/Y`000j`00`071`031`0300L70247000>1`l000=aRW400hX02G6:
LHn:Sg6?L@03S`03MXmf00>?00IfSgJ?MXl9MP09NgIkMW]fNgIi00=h011iN7UhNGQiN7ihNGQnNGii
17h028=nPgj3OX=i1Gh04WenOGimOWf1O89lPgf2OH9mPP=m00f0Nh5kO7^1NgakO7]l00fD00^]Sjf?
[HnXSjR?Z00G01`700<01`L010L02`070::4XXBRQ024008002JRRJ:9XXVR0:N9YhVWRJN9Y`2/SZb>
[8h0SY>>ThjCSY<0TgFCM@H0019TDfACI5=TD`1CI5=TDfACI5<40005I59TDV@00P0035AUE6EDIEAU
E6ED00EI00<0IFD00fD04`1INUUjFGYI05UTF@2CV9>HTiP00`002IQjV7X0[<>/``020006[<>/`j`0
1@L00`071`0P1`003PL@000QL8Y`RW2:L8Y`Sg2?L8m`Sg2?L8m`Sg1fL7I`MWEfDWIe00=f06UCMeAg
E7QEN5EhEGQ@N5EhEGQEN5EiEGUENEEiEGUVOVInIWiVOVMiDGAEOUEnEGiEOUEnEGeEPFEmIWeVOFIm
IWeVOFAkI7aTO6AlI7aTO6AkNYAjU7FDMIAeU7FDSXn>Shj?SXn>Shh05`0L1`0300L700D700D01`2M
Q003000RQ::408VRRJ:9XXVR0:N9YhVWRJN908jWSZN>Yhh0SY>>T`P015800`1CD`07D`0305=C00IC
0P00:EUBFE9IDUT0E@1UDfECIE=UDfECI@1IDUUB06AUI6ETIF@0FGYINUT000EI00<0Ti<00Y<03@2C
NY<0TgZC0:bW[0002:`00`071`031`0300L701l7000>1a40011`RW6:LHYaRW6?LHmaSg6?0gH00hmf
S`07MPQg17P00gehO@03N00BNGQiN7UhNGQiN7UhOWQnNGii17h018=nPgh4M003NGin009n019mOWQn
OGihOWf3OH=mPWf2OH84O@0=Nh5kO7]lNgakO7]lN`0=U00;SjR?Z8nXSjR?Z8l05P0L1`0300L700H7
02801`00Q::4XXBRQ029XXVRRJ:9XP2WRJN9YhVWR@2>[8j/1`0077YBNU80DVABI59TDV@0I5=TDfAC
I5=T06ACI5<30007D`1BI59TDP02000@IeEW06EDIEAUE6EDIEAU009I00<0IFD01VD0:01jFGYI05UT
FFAII5T0TiP0V7ZHNYP0V7ZH0<<0`jc3[<>/`jc3[0041`0300L701l7000>1a8008e`RW2:L8Y`RW2?
L8m`Sg2?L7I`MW1fCGI=MTefDWMCMe=gDgMDN5EhEGQEN5EhD7Q@N5EhEGQEN5EiEGUENEEiEGiVOVIn
D7A@M51dLgiEOUEnEGiEOEEmEG=VOFImIWeVOFImIWeDO6AlI7aTO6AlI7]TNe>DMIAeU7FDMIAeU8j?
SXn>Shj?SXn9Sf/05@0L1`0300L700L70P005hBRQ::4XXBRQ029XXVRRJ:9XP2WRJN900P000B>MHh0
25800`1BDP05DP0305=C00MC00<0D`000P05D`03059I00801ED07@1CIE=UDfECIE=UD`1B06AUI6ET
IFAUI6D0NUT000QI00l0T`2CNY=jTgZC09=jT`002:`20003[00700@700<01`L07PL000h74P003g1a
RW6?LHmaSgJ?MXmfS`0?MPMg1GP00gehO@03N00BNGQiN7UhNGQiN7UhOWUnNGii0gh8M00FNGinOWen
OGihOWQdLgimPgf2OH9mPP9m015hOGR1NgakO7]lNgakO7]lN`07U00>SiB?U8nXSjR?Z8nXSjP2SaD0
70L00`071`081`8001:4XXBRQ::4XX@0RJ:9XXVR0:84000USZb>[02CMI=eTgFC07YBNU9jDWX0I59T
DVABI580DfACI5=TD`03000PDfACI5=TDf@0I01WEFMEIeEWE@1DIEAUE6EDIEAUE00;I@0Q05T0FFAI
I5UTFFAII02HNYQjV7ZHNP1j09<0[<>/`jc3008000FW`jO300051`0300L701d7000>1aD008^:L8Y`
RW2?L8m`Sg2?L7I`MW1fL7I=MTefCGI>Me=gDgMCMeAhEGQEN5EhD7Q@N5EhEGQEN5EiEGUENEEiEGUV
NDadC7A@M51dD7AEOUEnEGiEOUEcC7=@OFImIWeVOFImEGQDN5AlI7aTO6AlI7]TNe=kMIAeU7FDMIB>
U8jDSXn>ShV?RHn9ShU[01@06`L00`071`0:1`8000b4XXBRQ::4XX@0RJ830003SXT000R>00P0SWF>
MHie00QB00<0DU801E800`1CD`02D`<00U<00`1CD`06D`802ED06P1UDfECIE=UDfD0E@1UI6ETIFAU
I6ETIF@02UT2000>NY=jTgZCNY=j07ZCNP04[0090:bW[:N/Yj`000@700<01`L07@L000h75@003V^?
LHmaSg6?MXmfSg6?3GH8M`Ah00=mN7d00gP04gUhNGQiN7UhNGQiN7ihOWUnNHd02g@3OP0<OGimOWed
LgAcM7f20gd04h9mOGQmN7ehOG]lNgakO7]lNg`017/2U00>SiB?U8nDSiB?U8nXSjP3S`04Z8nXL1@0
6`L00`071`0;1`8000V4XXBRQ::408@00P00:JN9YhVWR@2>[8j/SZb>[02CMI=eTgD0DWYBNU9jDP1B
I59TDVABI01T008000]CI5=TD`1CI5=TD`03000ID01EIeEWEFMEIeEW06EDIEAUE6D0EUEF0009I@0F
06L0I5UTFFAII5UT06QT07ZHNYQjV08000NC09>HT`2/008000W3Yl>W`jO3Y`001@L00`071`0L1`00
3PLG0029RW2:L8Y`Sg2?L8m`Sg1fL7I`MW1fCGI=MTegCWM>Me=gE7QEN5EhEGQEN51hEGQEN5EhEGUE
NEEiEGUcSFidC7A<M51dD7A@M5EnEGiEOU1cD7=<LdacIWeVOFImEGQDN5AhE7aTO6AlI7]TNe=kDg]B
NgFDMIB>U8jDSYB>ShV?RHn9ShV?RHl0500K1`0300L700d700<0Q:80100038BR08j9SXV>RHj900N>
00T0MHieSWF>M@001e800`1BDP04DP<01e<00`1CD`030006D5E@EE002ED01P1CIE=UD`80019ED5E@
06AUI6ETIFAU06IEIP07F@0306AT009T01/0NY=jT`2CNY<0TgZCNP2W[:N/YjbW[:N/Y`001@L00`07
1`0K1`003PLH000>Sg6?LHYaSgJ?LHmaSg4<MPMg1GP00gehO@03N00<NGQiN7UhNGQiN7Uc1Hd:M00B
LgAmOWenLgAcM7=dLgAcN7f20Wd047QmN7ehOGMmMgakO7]lNg`4N`0>MW]fU8nDSiB?U8nDSi@6S`06
RXn:ShZ?4`0K1`0300L700h70P00:j:9XXVRRJ80YhVWRJN9YhT0SZb>[8j/SP1eTgFCMI<0NU9jDWYB
NP1TDV@00`002V@0I5=TDfACI5<3000PD5I@EU1FD5H0IeEWEFMEIeEW06EDIE@0EEIEEUEFE@06I@0]
06MVIfH0FFAII5UT06QTJ6AX09QjV02HTiRCV02H09P0`jO3Yl>W`jO30<<000D700<01`L06`L000h7
6@00R8Y`RW2:L8m`MW1fL7I=MW1fL7I=MTefCGI>MdigDgMCMeAhEGQEN5EhD7Q@N5EhEGQEN5EiKXe^
SFj=KXe<M51dD7A@M51dD7AEOU1cD7=@Le1cC7=<OFImEGQEN5AhE7QDN5AlI7]CNe=kDg]CNe=kMIB>
U8jDMHn>Shj?SXn9ShV?RHn9ShTC01/700<01`L03`L2000@XXBRQ::4XP2>RHj9SXV>00N>00T0SWF>
MHieSP001e8300EB00<0De<00U<3000=D01@EE1ED5E@EE1E0009E@0J06EC051ED5E@EE1E06ETIFAU
06IEIUEVE@04F@0306AT00ET01D0T`2CNY=jTgZC09RCV02/YjbW[:L00P03Y`0300L700<700<01`L0
6PL000h76P001hmaSg6?LHl017H01g5fLGIaMW401gH7M`Ih011mN7QhNGQiN7UhNG>=Lhec1Xd9M00>
LgAcOW=dLgAcM7=dLg@2L`0BN7ehOGQmN7egOGMmMgegO7]l0W/057IkMW]fNgIkMYB?U8nDSiB?U8nD
0Xl02HZ?RXn:ShZ?J`0B01/700<01`L040L2000SXXVRRJ:9XP2WRJN9YhVW0:b>[8j/SZ`0TgFCMI=e
059jDWX010002E9TDVABI580D`03000VD5I@EU00D5I@EU1FD5I@05EWEFMEIeEWE@1D05EFEEIEEUEF
E@04I@0X06MVIfIWIVMV05UT06QTJ6AXI6QTJ02HTiRCV9>H09P0V9>H0<>W``8000NW`jO3Yl<000D7
00<01`L06@L000h76`00QgI`LG1aL7I`MW1fL7I=MTefCGI>MdigCWM?MdmgCWM?N5AhEGQEN5EhD7Qc
N6j<KXa^SFj=KXe^SFj=KWA<M51dD7A@M51dLgA@M51cD7=<LdacC7QEN5EhE7QDN5AhE7QDMe=kDg]B
Ne=kDg]eNgEkMHmeShj?SXn>ShV?RHZ9RXV:RHZ9RP0B01X700<01`L04PL2000<XXBRQ::4XP2>RHj9
8`00551ED5E@EE1E05E@EE1ED5E@EE0025D2000KD5E@EE1ED5E@EE00I6D0IUEVEFIEIUEV05T000UT
01<0J02CNY=jT`2HT`2CV9>H0:`000VW00<01`L00PL00`071`0I1`003PLL0008Sg6?LHmaSg43MP03
LGIa00=f00=aMg4017L01WQbN79hLPUh015cSG><LhecSG>=Lhf<SG>=M003S@Qd00YcSJFVLgAcM7=d
0g<04gAcLhamN7ehOGQmMgegOGMhMgP017/02WIkMW]fNgIkMW/3S`03U8nD00J?00B9ShZ?1HX00g40
000?01X700<01`L04`L20004XXVRR@8000>R00000P004Hj/SP1eTgFCMI=e059IDUUB00<01U@0IP1F
EEIEEUEF05I@EU1FD5I@051FD5I@EU1FD5H0IeEWEFMEI`0405IEEUEFEEIEEUEF06D0IfIWIVMVIfIW
IVL0J6AXI6QTJ6AX06QTJ02HTiP0V9>HT`2CV9>H0<>W`jO3Yl>W00:W00<01`L00PL00`071`0H1`00
3PLM0026LG1aL75`MW1fL7I`MTefCGI=MdigCWM?N4mhCgQ?Ld]cBg=<Ldb<KXa^S6j<KXa^SFj=KXe^
SFj=KWA<M4adD7A@M:FVS7A@M51cC7=<LdacKXa^N5AhE7QDN5AhE7QDMeAgDg]BNe=kMG]eNgEfMHme
Shj?SXm`ShV?RHY`RXV:RHZ9RV/A01X700<01`L050L500>900<0SXh01Hh00`1eM@05M@0305800080
0U802@1DDeACE5=D0007E@0D05E@EE1ED5E@051ED5E@EE1ED005E@800d`06P1ED5E@EE1ED5E@E@1V
EFIEIUEVEFIEIUD026@0501XI6QTJ02C09RCV9>HT`2C0:`01ZL20005XZNRY`0010L00`071`0G1`00
3PLN009a01b?LGIaMW5fLGIaMW5fLGIaMg5gLWMbN79hLWQb27<03XacSG>=LhecSG>=Lhec1Hd7M00?
YZFVYJJ<M7=dLgAcM7=d00B<01icOGQmMgegOGMmMgQgOGMmMW]fNgIkMW]fU7JDMY@:S`03RHZ900N:
00=[00003P0J1`0300L701D701H0RJN9YhVWR@2>[8j/SZb>07FCMI=e0`001U9IDUUB00MD01l0EEIE
EUEFE@1@EU1FD5I@EP1FD5I@EU1FD5I@05EW0080011<14`4C01EEUEFEEIEEUEF0P0096IWIVMVIfIW
IVMV05D0I6QTJ6AX06QTJ6AXI6P0V9>HTiRCV08000N//J`0Yl>W00801jL01@071`L001T7000>1al0
08EfL7I`MW1fL7I`MW1fCGI=MTigCWM>MdmgCgQ?N4]cBg=<LdacKXa^S6j<KXa^S6j=KXe^SFj=KXe<
M4adD:J<YXbVS:I@LdacC7=<Lfj<KXa]S79hE7QDN5AgE7MDMeAlE7MCMWEfMGIeSgF?MHmeSg2?L8m`
Sg2?L8n9RXV:RHZ9RXV:01006PL00`071`0F1`0308V900F900<0SXh01Hh00`1e0002009e00<0DU80
15802@1CE5=DDeAC0007E@0F051ED5E@EE1E05E@EE1ED5E@E@1E00Q<00X0D5E@EE1ED5D00T`02`1E
IUEVEFIEIUD000=E00<0I6@00V@0401XI6QTJ6AX0080V9>HTiP3000B/Jba[02W0::WXZNRYj:WXZL0
0`L00`071`0F1`003PLP009a01YfLGIaMW5fLGIaMW5fLGIaMg5gLGQbN79hLPYc00j<LhecSG>=Lhec
SG>=L`F=0W@03XfVYZJUYZFVYJIcM7=d0g<5S00NLhahOGMmMgQgN7MmMgagO7MlMW]fNgJDMYAfU8nD
1Xl01XZ?RXn:S`B:00>9RXT00XX00g40000=01X700<01`L05`L03`29YhVWRJN908j/SZb>[002000=
MI=eTgFC05UBFE9I0007E00K05IEEUEFEEH0EU1FD5I@EU00D5I@EU1FD5I@00800444C0A<14`4C0@0
EUEFEEIEEP04C0A<06IWIVMVIfH0EEIEEP1XI6P0J6AXI6QTJ01XFFP0V9>H0;6//@2a[;6/0008Y`80
00>R00L00PL00`071`0F1`003PLQ0024LG1fL75`MTefCGI=MTeaCGI>MdigCW9?N4mhCg=;Ld]cC7=<
LficKXa^S6j<KXe^SFj=KXe^SFj=QjJ7YXNVQjJ7YXN<C7=<Lfj<KXa^S6f<KGQDN5AgE7MDMeAlE7aD
Me=gDgIeSgF?MHmeSgF?MHm`RW2:L8Y`RW2:L8Z4RV^:JhY[3`0J1`0300L701P700<0RHT01HT00`2>
SP030003SP1e00Ee00<0DU8015802@1DDeACE5=D0007E@0@05E@EE1ED5E@051ED5E@E@8000=<04`0
2D`01`1ED5E@E@001D`0201VEFIEIUD01UD06`1T06QTJ6AXI6P00UT2F@80V02a[;6//@2a0002[005
0::WXZL00`04XP0500L71`005`L000h78`006g5fLGIaMW5fLGIaMW5fLGMbMg5gLWQbN79hLP08L`0>
S7>=LhecSG>=LhecSG<5S@0@YJJVYZFVYJJUYZFVYGAcM0J<01]cS7><LWQgN7MhMgQgO7MlMgafO7Il
MYAfSgH00Xl00iB?S`04S`06RXn:ShZ?1XX00hV:R@03RPl06@L00`071`0J1`0508VWRJL03`002I=e
059IDUUB0007E00E05EFEEIEEUD0D5I@EU1FD5H0EU1F00<0041_KVm^04`4C0A<14`4C01EEUEF00A<
14`4C0@0IfIWIP1EEUEFEEIEEP1XI6QTJ6AX06QIJ5UXFFP0/Jba[;6/0P001/>/`j`0Y`802:801@07
1`L001H7000>1b@007i`LG1aL75=MTeaCG5=LDebCgM>MdmbCgQ?N4]cBg=<LdacKW=^S6j<KXa^SFj=
KXe^SHNUX:F7YXNVQjJ7YXNUQjE^Lfj<KXa^S6j<KHa]S6ehE7MDMeAlE7MDMe=gDgIBMU9fMHmeSgF?
L8m`Sg2:L8Y`RW2:L8Y`RV^:JhX2Ja006@L00`071`0K1`H000V>09=eTgFCM@00158600=B00P0DeAC
E5=D00ME00/0EE1ED5E@EE00D0020008KTa^C6i<KP0:C004051E00Q<00D0EFIE0009E@8001=XI6QT
J002F@9I0UT2FGl0/Jba008000>/0:`01:`00`2RXP09XP0400L701H7000>1b@001a`LGIaMW5fLGIa
MW5fLGIaMg9gLWQbN79hLWQb1W<03hacSG>=LhecSG>=Lhf<YP05Y@0?YXfVS:J<YXbVS:J<YHb=00B<
01IcS7><LXabS79hMgegO7MlMgafO7Ik1GH00hmfS`02S`09RXn:ShZ?RHZ900F:00J9RXZ:Jg4C01T7
00<01`L06`L3000DYhjWSZL0ThjCSY>>059jDWYBNP02E0@01e@03`1FEEIEEUEF05I@EU1FD003001>
KVm^Kfi_KVm^04`4C0A<14`4C0@0EP04C0A<14`4C0A<06H0EEIEEUEFEEIE05EW06QTJ01XFFQIJ5T0
0Wl2O`2a0<>/`j`0[<<0_ZL01j820003XP07008700<01`L04`L000h79`00Mg5`LG1aCG5=LDeaCG5=
LTmbCg9;LTmcCg=;Ld]cC7=<LficKXa^S6j=KXb7YJ2UX:FPYJ2VQjJ7YXNVQjF7YInUKXa^S6j<KHa]
S6f<KH]]MeAlE7MCMe=gDgIBMU9fMGIeMW2?L8m`Sg2?L8Y`RW2:L8Y`RP1`01H06@L00`071`0H1`<0
0PL03`29SXV>RHh0TgFCMI=e0006DP0505ACE5<01P000eAC0007E@03051E00@000i<KP1^C6i<KTa^
C6i<00U<0P0<C00305EE00ME01H0EEIEEP1X009I0UT2F@1IOeUoFGl01J`300BW00<0XZ800Z820005
WJ:MXP000`L00`071`0B1`003PLX000>MW5fLGIaMW5fLGIaMW42LP0:Mg9cLW=bLg9cLPIc00Z<Lhac
SG>=S:F<2JD03XbVS:J<YXbVS:JTYJBU0X`03G><LhacS7:<LXabS7800WL01GagO7Ig00Mf00n?MXmf
Sg6?RXn:ShZ?RXl01HXJ01T700<01`L05PL200H701D0SZN>YhjW09>>ThjCSP1BNU9jDP001E@01@1@
EU1F00d002Q_KVm^Kfi_06m^Kfi_KVm^K`04C0A<14`4C02704`4C0A<14`4C0A<0P007EIEEUEFEEIE
05EWEFMEI`1XFFQIJ5T00Wl2O`9o00:H00@0`jc30`002;jW_ZNnY`2R0P002::MXYfRWJ800PL00`07
1`0B1`003PLY001_LG1aCG5=LDeaCG5>LDibCg9?LT]bBg9]LfecKG=^S6f<KHb6S8JUQZF6YHJUQZFP
YJ2UQjJ7YXNVQjFOYInUWha^S6f<KHa]S6f<KG9;LT]bCgMCMU=fDWIBMWEfMHm`Sg2?L8m`RW2:L8Y`
RV]`01`06@L00`071`0C1`<02@L03P29SXV>RHh0TgFCMI<01U803P1CE5=DD`1@DE1AD5401d`0501^
C6i<KTa^C01<KTa^C6i<KT`01T`200>700<0C4`01d`0101^KP06E@0L05EFEEIEEUD0IP1I0UT0FGmI
OeT0TiRCV02/00>W00<0YjL01:L02`2MXYfRWJ:MXYd000<700<01`L04@L000h7:P02L@05MW5fLGH0
1g42LP0<Mg9bLW=bLg:<Lhac0h`02h^<RjF;YHbUS:F<00RU01VVS:J<YXbUY:FTYJBURhacS7><LXab
S79c00Eb0gL7MP0ASgJ?MXmfSg6?L8Y`RW6:Jg407`0H1`0300L701470`0=1``000RC07YBNU9j00ID
05<0D5I@EU1F00A<14`4C0@0Kfi_KVm^Kfh0KVm^Kfi_KVm^04`4C0@0XHNQQj4014`4C0A<14`0KVm^
K`1FEEIE05EWEFMEIeD0IPIV05T00Wl2O`0200JH00T0_ZNnYkh0_ZL00P003JL0XYfRWJ:MXYfRW@00
0`L00`071`0@1`003PL[001VMTeaCG5=LDeaCG5>LDibCW9;LT]bBg=]S6f<KHa]S8J<QXb6YHJUQZF6
YHJUQZFPYJ2VQjJ7YInUWjFOYInTKHa]S6f<KG=]Rd]bBg9>MdigDgI>MTefL7I`Sg2?L8Y`Sg2:L8Y`
8`0H1`0300L700h70`0A1`0308j>00B>00@0NWEj1@03DP0>05ACE5=D055@DE1AD007C00D04a^C6i<
KTa^06i<KTa^C6i<KP02C0801hL00`1<C004C00306i^00=^0180EEEE05EFEEIEEUD0IPIV1VH2000?
OeUo09RCV9>HT`2//J`000>W0`04XP060::MXYfR0`000id01`021`0300L70107000>1b`000MaL7I`
MW5f00Ea27805W=bLg:<LXacS8^<RjF;YH^US:F<YH`7Y@0FS:J<YJBUY:FTYJBUY:F;S7:<LXabL`Eb
0gL6MP0>MGIeSgJ?LHmaSg6?L74U01P700<01`L02`L301D700<0SXh018h0201jMGYe059C1006E01<
05I@EU1F00A<14`4C0@0Kfi_KVm^Kfi_06m^Kfi_KVm^04`0Qj67XHNQQj6704`4C0A<06i_KVm^Kfh0
E@1EIeEWEFME06H6IPIV009Y00<0O`001IP20007[0>/0j`0Y`030008XZNRYj:W0:830006XYfRWJ80
0`L00`071`0?1`003PL^001LCG5=LDeaCG5=LDia679>LT]bBg9;Lff<KHa]S8J<QXb6S8JUQZF6YHJU
QZF6YInUQjFOYInUQZF6YHJTQZA]S6f;KG9;LT]bCWM>MdifCGI`MW1fL7I`Sg2:L8X2L2P060L00`07
1`091`806@L00`2>SP04SP0607YeNWD01584000;DeAC051AD55@D@001d`04P1^C6i<KTa^C01<KTa^
C6i<KP802XL00`1<C002C00306i^00M^01P0EEIEEUEFE@1V1VH6IP1UJFEY09RCV9<20009/Jba[;6/
/J`000>R00<0XZ801:80302MQ9f4WHBMQ9d01`<040L000h7<0000gIaL@09L@Ib01IcLW=bS7><Lhb;
S8^URjF;YHbURjF<1JD05:>TY:FTYJBUY:F;YH^URjAbS7:<1G800gMbM`08MP08LHmaSg6?LHl[01P7
00<01`L01PL301`700<0SXh018h03@1jMGYe059CDU<0EUD01@007E1FD5I@04`4C0A<14`0KVm^Kfi_
KVl0Kfi_KVm^008001^P0:67XHNQQj67XHNQ00A<06i_KVm^Kfi_KVl00P0035EWEFME06H6IPIV00AY
0`802P2H0:`3[0>/0j`200:W00H0XZNR0:83000=W@2MXYfRWJ:MXYfR00021`0300L700h7000>1c40
05EaL75=LDeaCG5>LAQbCW9;LT]bKG=]Lff<KH]]S8J<QXb6YHJUQZF6YHJTQJB5Y8JUQZF6YHJUQZB6
RhJ;KG9;LT]bCWM>MTifCGI`MW1fL7I`MW1f02h060L00`071`031`<080L:0003NWD000AB00H0EU1F
D5H30005DE1AD0001T`03P1^C6i<KTa^C6h0KTa^0P001HJPQZ0000Z700<0C0002Vh04P1_KP1EEUD0
IPIV1VH0IFUU00D200L0[;6//Jba00801JL00`2R0002000;WJ:MXYd0Q9f4WH@00`041`0300L700h7
000>1c8000AaL75`0W400gIaL@03L@Ib02McLW=bS7><Rhb;S8^<RjF;YHbURjB:Y:>TRZB:YJBURjF;
YH^URj@018/6LP0?LGIaMW5fLGI`MW6?L75=03005`L010071`L302@700T0SY>>Thh0DWX00P003WD0
DU=BD`1FEEIEEP040`0055H014`4C0A<101_KVm^Kfi_KP1^0P00:j2OX9nPW`27XHNQQj67XHNQQ`1^
Kfi_KVm^Kfi_08Q^R6h0E@1V1VH6IP000fT00`020P030P>H0P000j`00008Y`0<09fR0::MXYfRW@2M
0`071`0300L700h7000>1c<004e`CG5=LDeaCG5=L@1b678HLT]bKG=]S6f<KHa]S8J<QXb6YHJTQJB5
Y8FTQJB6YHJUQZF6Y8J;QX^6Rff;K8/HLTifCGI=MW1fL7I`MP0d01L7100X1`0;07F>MHie059eDWD0
1002DP0:05I@EU1F00A<10801d`0201<KTa^C6i<0`002XJPQZ26X8JPQP08Q`030:0000U^00L0Kfi_
KVm^008000T6IPIV06EYI@001080202HTiRCV9<01ZL300:R00/0WJ80XYfRWJ:M000:1`0300L700h7
000>1c@000=aL7400W40171fLGH2L@03CW5a00Ib00N<LXabS8^<00>;01NTRjB:Y8ZTRZB:Y8ZTRZB;
YH^URjF;Y@06R`0@LX]bLW5fLGI`MW5fL7I=L3D05`L202/700`0SY>>Thh0DWYBNP02D`@000YFEEIE
04`4C0A<1@001D`406m^00D002Z6X02PWj2OX9nPWj00XHNQQj67X@2hWkP0Kfi_KVm^K`28KXQ^R6h0
C0@20003IP1Y009Y00D00P820006V0060;6//@2W0`06XP0309d000@03PL00`071`0=1`003PLg0011
LG1aL75=LDeaCG4HLQQbBg9/LVf;KH]/Rfb;QH^5RhF;QH^5RhF;QH^5Y8J;QX^6RhJ;K8]]Rfab675=
MTefL7H00W0j01T700<01`L0:@L0301eSWF>M@1BME9e00AC0`003E1FD01<14`408Q^R6h01@004hJ7
QXN6QhJ70:26X8JPQZ26X0001HL01`2PWj2OX0001Fh03@1_KVm^Kfh0C0A<14`00`000fD00P020P05
09>HTiP00P06[0030::R00RR0`0A1`0300L700d7000>1cP027400gIaL@02L@Mb01V;LX^:RhZ;RX^:
RhZTRZB:Y8ZTRZB:Y8^U00F;00><LX/00W80275fLGIaMW5f?@0J1`0300L702T72@000e9j0004D`18
054005H014`4C0@0R6j8KXQ^08JPQZ26X8JPQP2OX9nPWj2OX9l0Qj67X@2hWkRO^9nh06m^K`28KXQ^
R6h0C0A<14aWEFME1@000iP0V0020005/Jba[;400P02Y`030::R00>R0`041`0300L700h700<01`L0
3@L000h7>@00>G5`LG1aL75=LDeaCG4HLQQbK79]LVb:K8Y/RVb;QH^5RhF;QH^5RhF:QH^5RhJ;K8]]
LVabCWI=MP02L4006`L00`071`0Y1`0807F>MHieSU8400AC02L0D540EP04C0A<1028KXQ^R6h0QXN6
QhJ7QXL0X8JPQZ26X8JPQP000XL07@2PWj2OX9nPWj00KP1_KVm^Kfh0C0A<101WEFME00AY00Eo0Wl2
0005[08000NWXZNRY`2R00<020L00`071`0=1`0300L700d7000>1c/000M`LG1aL75`009a00IfL75a
LW46LP0CRg6;LH^:RhZ;RX^:RhZ;RX^:R`03RP03Rg:;009b00IaMW5a075301`70P0Z1`0:08jCSY>>
T`1jDP<00e<00`1AD@02000^C0A<14`0KXQ^R6j80:26X8JPQZ26X02PWj2OX9nPWj00X@2hWkRO^9nh
WkRO^08000m^R6j8KP1<14`406MEI`000fT02P1o0Wl2/IRa0;4200RW00<01`L02@L00`071`0=1`03
00L700d7000>1cd002maL75`LF]aL75`LDea678HLVabK75/RVb;K8Y/RXF;QHY/RVb:K8Y/RVb;K79=
MP1801h700<01`L0:0L0201eSWF>MHh015830005D`1@DE000P007PA<14`0KXQ^R6j808N6QhJ7QXN6
08JPQZ26X8JPQP8001ZPWj2OX9nPWj2O09nhW`1^Kfh0C0A<101WE@800VT04P02O`80Tk6C/I>a0:NR
Yj:WXP<03PL00`071`0<1`0300L700d7000>1ch000AaL75`0g400g1aL004L@03LW5b009b01AaRW6;
LH^:RhZ;RX^:RhZ;LHYaR`9b00<0L@00BP0O1`0300L702P72P0017YB05<300=A10006@@0R6j8KXQ^
R02PQZ26X8JPQP2OX9nPWj000P0079h0WkRO^9nhWkRO0:2hX;RP06h0C0A<101WE@02J@0307l20080
00FaV;6H/@020005YjbW0:L00`0B1`0300L700/700<01`L03@L000h7?`009G5[LG1aJg5[LG1aL75`
LFab679/LFaaK75/RVb:K8Y/RVabK7800Va>020700<01`L0:@L0202>MHieSWEB0P001E80CU=>0080
00EAD01<1003000CR6j8KXQ^08J7QXN6QhJ70:26X003000;WYfNW@2OX9nPWj000P001inhWkRO^9l0
0`001d`406ME06T01@0019>aTk4200NW00<01`L04`L00`071`0<1`0300L700`7000>1d4000e`RW2:
L8YaRW6:LHY`00Aa00=bLG401g401hYaRW5bLG80D@0Q1`0300L702X700T0MI=eTgFC07X01002D`@0
00@4K`A_0`003XQ^R6h0QZ26X8JPQP2O0P08WP0B09nhWkP0^:2hX;RP^02QR:681P0010H2J084V080
00F/YjbW[00301P700<01`L02`L00`071`0<1`003PM3000KLF]aJg5[LG1aJg5[LF]aL75/LAQa66`0
K01/05H08PL202/71P000hh00002DP06C@00D`1@0P001T`4C01^Q`D000B7QXN61@008YfNWIjMWYfN
WIh0X9nP0;RO^9nhWkP0XHNQQj5<14`6IP020P0;OiQoV7l0TjbC0:L00P0L1`0300L700X700<01`L0
30L000h7A0004HY`RW2:LHY`RW2:L8Y[L@1`05l090L00`071`0Y1`0308j>00B>00L0NP1jDP1=0080
00aAD000K`@0KXM^Qfh50005QIj5WP002ih02@2h0;RP^:2hX002000;XHRQR01<100606P00P02V080
1j`00`071`0M1`0300L700T700<01`L030L000h7A@002W5[LF^:Jg5[LF]U02D700<01`L0:@L02`1e
SWF>MHh0NU9j00<000EBC@1<D002000XC01^Qfj7KP24QHB5Q8F4Q@2NWIjMWYfNWIjMWP2hWkRO^9l0
Qj67X@8000E<06H00P020009V000TjbC[9>/00<08@L00`071`091`0300L700`7000>1dH000F:L8Y`
L@1Y02H700<01`L0:@L00`2>SP04SP0607YBNU9j0P0014eC0542000A101^Qfj7KP25WXFNQIj5WP00
1ih200>N00d0X;RP08RQR:4014`600@00Wl01P2aTk40[0<09@L00`071`081`0300L700`7000>1k@0
9`L202X74P000e00C002000CKXM^Qfh0Q8F4QHB5Q8D0WYfNW@0200JN00T0W`27XHNQ00@00P000e@0
NP020008[9>/TjbC[00Y1`0300L700L700<01`L030L000h7]00Y1`0300L702T70140SXV>RHj907Ej
MGYeNP1IDP03000BC00006j7KP25WXFNQIj5WP2N0P09WPL000=E07l00`0019>aTk4302`700<01`L0
20L00`071`0;1`003PNd02X700<01`L0:@L2000=RHj9SXT0MGYeNP1jM@030003DP00009<1@0018B5
Q8D500RN0P001:27X8L20005EEUoF@001I<2030700<01`L01`L00`071`0;1`003PNd02/700<01`L0
:@L60003R@0000801Hh20008NP1B101^C6h500J500<0WYh00Yh20003Qj2700<000ME002aTk6C00<0
=@L00`071`061`0300L700/7000>1k@0;0L202/700<0RHT018T2000<L8i`SW2>L01BNU9j0`03C003
08F500J500H0WP00X8L20008KP2CSY>>Thh303T700<01`L01@L00`071`0;1`003PNd02h700<01`L0
:@L02@29XXVRRJ:9XP0500:>1003SP8000=<KP0018D200:500<0Q`000`000i<00002T`<0?@L00`07
1`051`0300L700/7000>1k@0;`L00`071`0Y1`0308V900<000BRQ:801hT20007MHieSWF>M@050003
Bfh0008028h2044700<01`L010L00`071`0;1`003PNd030700<01`L0:@L2000@XXBRQ::4XP2RRJ:9
XXVRR@@000FRRJ:90008SP@0B0L00`071`031`0300L700/7000>1k@0<@L00`071`0Y1`8000RRQ::4
XXBR00:90`0018BRQ007R@0308V>00<0CPL00`071`041`0300L700X7000>1k@0<PL202X70P001j:4
XX@0Q:800P0038BRQ::4XX@0RJ:9XP8000>90000DPL00`071`031`0300L700X7000>1k@0=0L00`07
1`0X1`D000f4WHBMQ024XXBRQ::400@0F0L00`071`021`0300L700X7000>1k@0=@L00`071`0X1`80
00^RQ::4XXBRQ024XP0305h700D01`L7000<1`003PNd03H700<01`L0:0L20008WHBMQ9f4WH@20647
00D01`L7000<1`003PNd03L70P0Y1`8000>RQ0000P1V1`0400L700`7000>1k@0>@L00`071`0W1`<0
JPL00`07000<1`003PNd03X700<01`L09PL00`071`1Z1`0400L700/7000>1k@0>`L00`071`0U1`03
00L706/700<01`002`L000h7]00l1`809@L00`071`1/1`802`L000h7]00n1`0300L7028700<01`L0
J`L300/7000>1k@0?`L00`071`0Q1`0300L706L7100>1`003PNd040700<01`L080L00`071`1T1`<0
4PL000h7]0111`0300L701l700<01`L0H0L401D7000>1k@0@PL201l700<01`L0G0L401T7000>1k@0
A0L00`071`0L1`0300L705P7100M1`003PNd04D700<01`L06`L00`071`1E1`<08@L000h7]0161`03
00L701X700<01`L0D@L402@7000>1k@0A`L201X700<01`L0C@L402P7000>1k@0B@L00`071`0G1`03
00L704T7100/1`003PNd04X700<01`L05PL00`071`151`@0<0L000h7]01;1`0300L701D700<01`L0
@PL303@7000>1k@0C0L00`071`0D1`0300L703h7100g1`003PNd04d70P0D1`0300L703X7100k1`00
3PNd04l700<01`L040L00`071`0g1`@0?`L000h7]01@1`0300L700l700<01`L0=0L304<7000>1k@0
D@L00`071`0>1`0300L7030710161`003PNd05870P0>1`0300L702`7101:1`003PNd05@700<01`L0
2`L00`071`0X1`@0CPL000h7]01E1`0300L700X700<01`L09@L30587000>1k@0EPL00`071`091`03
00L70247101E1`003PNd05L70P091`0300L701d7101I1`003PNd05T700<01`L01PL00`071`0I1`@0
G@L000h7]01J1`0300L700D700<01`L05@L40647000>1k@0F`L00`071`041`0300L701870`1U1`00
3PNd05`700<01`L00`L00`071`0>1`@0J0L000h7]01M1`800`L00`071`0:1`@0K0L000h7]01O1`04
00L700P7101`1`003PNd060700<01`001@L307@7000>1k@0H@L200031`000080M`L000h7]01R1`80
N`L000h7]03O1`003PNd0=l7000>1k@0g`L000h7]03O1`003PNd0=l7000>1k@0g`L000h7]03O1`00
3PNd0=l7000>1k@0g`L000h7]03O1`003PNd0=l7000>1k@0g`L000h7]03O1`003PNd0=l7000>1k@0
g`L000h7]03O1`00o`NR1`00o`NR1`00o`NR1`00o`NR1`00o`NR1`00o`NR1`00o`NR1`00o`NR1`00
o`NR1`00\
\>"],
  ImageRangeCache->{{{0, 416.312}, {197.688, 0}} -> {-0.0529655, -0.0238146, 
  0.00505854, 0.00505854}, {{10.4375, 198.688}, {192.938, 
  4.6875}} -> {-24.9669, -15.4499, 1.67333, 1.67333}, {{217.562, 405.812}, {
  192.938, 4.6875}} -> {-1.10124, -0.0495867, 0.00506169, 0.00506169}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(colorknot = 
      ParametricPlot3D[
        Evaluate[
          Append[FrenetFrame[{4\ Cos[t\  + \ N[Pi]]\ /\ 3\  + \ 2\ Cos[3\ t], 
                \n\ \ \ \ \ \ \ 4\ Sin[t]\ /\ 3\  + \ 2\ Sin[3\ t], \n
                \ \ \ \ \ \ \ Sin[4\ t]\  + \ Sin[2\ t]\ /\ 2}, t, phi, 
              0.25], SurfaceColor[Hue[t/\((2*Pi)\)]]]], {t, 0, 2  Pi}, {phi, 
          0, 2  Pi}, PlotPoints -> {100, 18}]\)], "Input",
  CellLabel->"In[126]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .88237 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-2.48307e-17 1.05048 -0.0338809 1.05048 [
[ 0 0 0 0 ]
[ 1 .88237 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.05024 .25302 m
0 .57014 L
s
0 .57014 m
.39857 .88237 L
s
.39857 .88237 m
.40859 .59552 L
s
.40859 .59552 m
.05024 .25302 L
s
.68441 0 m
.95857 .42636 L
s
.95857 .42636 m
1 .72958 L
s
1 .72958 m
.70366 .33203 L
s
.70366 .33203 m
.68441 0 L
s
.05024 .25302 m
0 .57014 L
s
0 .57014 m
.70366 .33203 L
s
.70366 .33203 m
.68441 0 L
s
.68441 0 m
.05024 .25302 L
s
.40859 .59552 m
.95857 .42636 L
s
.95857 .42636 m
1 .72958 L
s
1 .72958 m
.39857 .88237 L
s
.39857 .88237 m
.40859 .59552 L
s
0 0 m
1 0 L
1 .88237 L
0 .88237 L
closepath
clip
newpath
.5 Mabswid
.36208 .64126 m .35363 .6457 L .37779 .68384 L p .706 .527 0 r
F P
0 g
s
.37779 .68384 m .38552 .67825 L .36208 .64126 L p .706 .527 0 r
F P
0 g
s
.37019 .68965 m .37779 .68384 L .35363 .6457 L p .762 .54 0 r
F P
0 g
s
.38552 .67825 m .37779 .68384 L .40606 .71979 L p .696 .465 0 r
F P
0 g
s
.37779 .68384 m .37019 .68965 L .39898 .72613 L p .746 .47 0 r
F P
0 g
s
.39898 .72613 m .40606 .71979 L .37779 .68384 L p .746 .47 0 r
F P
0 g
s
.36976 .63814 m .36208 .64126 L .38552 .67825 L p .634 .506 0 r
F P
0 g
s
.38552 .67825 m .39236 .67359 L .36976 .63814 L p .634 .506 0 r
F P
0 g
s
.39236 .67359 m .38552 .67825 L .41303 .71316 L p .632 .455 0 r
F P
0 g
s
.40606 .71979 m .41303 .71316 L .38552 .67825 L p .696 .465 0 r
F P
0 g
s
.41303 .71316 m .40606 .71979 L .43763 .752 L p .683 .406 0 r
F P
0 g
s
.40606 .71979 m .39898 .72613 L .4311 .75876 L p .727 .403 0 r
F P
0 g
s
.4311 .75876 m .43763 .752 L .40606 .71979 L p .727 .403 0 r
F P
0 g
s
.41303 .71316 m .41898 .7071 L .39236 .67359 L p .632 .455 0 r
F P
0 g
s
.41898 .7071 m .41303 .71316 L .44378 .74446 L p .627 .406 0 r
F P
0 g
s
.43763 .752 m .44378 .74446 L .41303 .71316 L p .683 .406 0 r
F P
0 g
s
.35363 .6457 m .34552 .65086 L .37019 .68965 L p .762 .54 0 r
F P
0 g
s
.3637 .69489 m .37019 .68965 L .34552 .65086 L p .816 .548 0 r
F P
0 g
s
.37019 .68965 m .3637 .69489 L .39273 .73133 L p .796 .47 0 r
F P
0 g
s
.39273 .73133 m .39898 .72613 L .37019 .68965 L p .796 .47 0 r
F P
0 g
s
.65031 .4848 m .65029 .49444 L .69682 .49253 L p 0 .631 .466 r
F P
0 g
s
.69682 .49253 m .6969 .48295 L .65031 .4848 L p 0 .631 .466 r
F P
0 g
s
.69728 .50197 m .69682 .49253 L .65029 .49444 L p 0 .57 .421 r
F P
0 g
s
.6969 .48295 m .69682 .49253 L .74304 .49271 L p 0 .592 .423 r
F P
0 g
s
.69682 .49253 m .69728 .50197 L .74261 .50216 L p 0 .536 .383 r
F P
0 g
s
.74261 .50216 m .74304 .49271 L .69682 .49253 L p 0 .536 .383 r
F P
0 g
s
.65162 .47622 m .65031 .4848 L .6969 .48295 L p 0 .691 .512 r
F P
0 g
s
.6969 .48295 m .69753 .47449 L .65162 .47622 L p 0 .691 .512 r
F P
0 g
s
.69753 .47449 m .6969 .48295 L .74316 .48313 L p 0 .648 .462 r
F P
0 g
s
.74304 .49271 m .74316 .48313 L .6969 .48295 L p 0 .592 .423 r
F P
0 g
s
.39898 .72613 m .39273 .73133 L .42504 .76384 L p .772 .395 0 r
F P
0 g
s
.42504 .76384 m .4311 .75876 L .39898 .72613 L p .772 .395 0 r
F P
0 g
s
.37565 .63675 m .36976 .63814 L .39236 .67359 L p .507 .462 0 r
F P
0 g
s
.39236 .67359 m .39741 .6705 L .37565 .63675 L p .507 .462 0 r
F P
0 g
s
.39741 .6705 m .39236 .67359 L .41898 .7071 L p .524 .431 0 r
F P
0 g
s
.65029 .49444 m .65156 .50386 L .69728 .50197 L p 0 .57 .421 r
F P
0 g
s
.69822 .51004 m .69728 .50197 L .65156 .50386 L p 0 .492 .362 r
F P
0 g
s
.69728 .50197 m .69822 .51004 L .74192 .51025 L p 0 .464 .332 r
F P
0 g
s
.74192 .51025 m .74261 .50216 L .69728 .50197 L p 0 .464 .332 r
F P
0 g
s
.74316 .48313 m .74304 .49271 L .78729 .49498 L p 0 .547 .376 r
F P
0 g
s
.74304 .49271 m .74261 .50216 L .78593 .50443 L p 0 .492 .34 r
F P
0 g
s
.78593 .50443 m .78729 .49498 L .74304 .49271 L p 0 .492 .34 r
F P
0 g
s
.34354 .60362 m .33443 .60679 L .35363 .6457 L p .714 .593 0 r
F P
0 g
s
.35363 .6457 m .36208 .64126 L .34354 .60362 L p .714 .593 0 r
F P
0 g
s
.34552 .65086 m .35363 .6457 L .33443 .60679 L p .776 .616 0 r
F P
0 g
s
.41898 .7071 m .42312 .70241 L .39741 .6705 L p .524 .431 0 r
F P
0 g
s
.42312 .70241 m .41898 .7071 L .44875 .73712 L p .537 .397 0 r
F P
0 g
s
.44378 .74446 m .44875 .73712 L .41898 .7071 L p .627 .406 0 r
F P
0 g
s
.35199 .60214 m .34354 .60362 L .36208 .64126 L p .633 .559 0 r
F P
0 g
s
.36208 .64126 m .36976 .63814 L .35199 .60214 L p .633 .559 0 r
F P
0 g
s
.74316 .48313 m .74295 .47467 L .69753 .47449 L p 0 .648 .462 r
F P
0 g
s
.74295 .47467 m .74316 .48313 L .78746 .48533 L p 0 .602 .412 r
F P
0 g
s
.78729 .49498 m .78746 .48533 L .74316 .48313 L p 0 .547 .376 r
F P
0 g
s
.74261 .50216 m .74192 .51025 L .78356 .51243 L p 0 .422 .293 r
F P
0 g
s
.78356 .51243 m .78593 .50443 L .74261 .50216 L p 0 .422 .293 r
F P
0 g
s
.33443 .60679 m .32587 .61125 L .34552 .65086 L p .776 .616 0 r
F P
0 g
s
.33881 .65608 m .34552 .65086 L .32587 .61125 L p .836 .633 0 r
F P
0 g
s
.34552 .65086 m .33881 .65608 L .3637 .69489 L p .816 .548 0 r
F P
0 g
s
.35921 .69889 m .3637 .69489 L .33881 .65608 L p .88 .548 0 r
F P
0 g
s
.3637 .69489 m .35921 .69889 L .38813 .73471 L p .854 .458 0 r
F P
0 g
s
.38813 .73471 m .39273 .73133 L .3637 .69489 L p .854 .458 0 r
F P
0 g
s
.65406 .46982 m .65162 .47622 L .69753 .47449 L p 0 .763 .567 r
F P
0 g
s
.69753 .47449 m .69861 .46827 L .65406 .46982 L p 0 .763 .567 r
F P
0 g
s
.69861 .46827 m .69753 .47449 L .74295 .47467 L p 0 .717 .511 r
F P
0 g
s
.65156 .50386 m .65397 .51184 L .69822 .51004 L p 0 .492 .362 r
F P
0 g
s
.69954 .51567 m .69822 .51004 L .65397 .51184 L p 0 .356 .26 r
F P
0 g
s
.69822 .51004 m .69954 .51567 L .74106 .51588 L p 0 .337 .242 r
F P
0 g
s
.74106 .51588 m .74192 .51025 L .69822 .51004 L p 0 .337 .242 r
F P
0 g
s
.44875 .73712 m .44378 .74446 L .47692 .77048 L p .617 .356 0 r
F P
0 g
s
.44378 .74446 m .43763 .752 L .47168 .7788 L p .663 .348 0 r
F P
0 g
s
.47168 .7788 m .47692 .77048 L .44378 .74446 L p .663 .348 0 r
F P
0 g
s
.35868 .60256 m .35199 .60214 L .36976 .63814 L p .491 .492 0 r
F P
0 g
s
.36976 .63814 m .37565 .63675 L .35868 .60256 L p .491 .492 0 r
F P
0 g
s
.43763 .752 m .4311 .75876 L .46572 .78589 L p .701 .337 0 r
F P
0 g
s
.46572 .78589 m .47168 .7788 L .43763 .752 L p .701 .337 0 r
F P
0 g
s
.39273 .73133 m .38813 .73471 L .42028 .76656 L p .823 .368 0 r
F P
0 g
s
.42028 .76656 m .42504 .76384 L .39273 .73133 L p .823 .368 0 r
F P
0 g
s
.74295 .47467 m .74244 .46846 L .69861 .46827 L p 0 .717 .511 r
F P
0 g
s
.74244 .46846 m .74295 .47467 L .78641 .47675 L p 0 .669 .456 r
F P
0 g
s
.78746 .48533 m .78641 .47675 L .74295 .47467 L p 0 .602 .412 r
F P
0 g
s
.37565 .63675 m .39741 .6705 L p .38048 .64002 L .156 .306 0 r
F P
0 g
s
.39741 .6705 m .39998 .66938 L .37897 .63729 L p .156 .306 0 r
F P
0 g
s
.39998 .66938 m .39741 .6705 L .42312 .70241 L p .232 .335 0 r
F P
0 g
s
.3997 .67041 m .37923 .63971 L p .39837 .66708 L .776 .293 0 r
F P
0 g
s
.39998 .66938 m .39995 .66947 L p .39837 .66708 L .776 .293 0 r
F P
0 g
s
.39995 .66947 m .3997 .67041 L p .39837 .66708 L .776 .293 0 r
F P
0 g
s
.37897 .63729 m .37865 .63724 L p .38048 .64002 L .156 .306 0 r
F P
0 g
s
.37865 .63724 m .37565 .63675 L p .38048 .64002 L .156 .306 0 r
F P
0 g
s
.74192 .51025 m .74106 .51588 L .78047 .51792 L p 0 .298 .21 r
F P
0 g
s
.78047 .51792 m .78356 .51243 L .74192 .51025 L p 0 .298 .21 r
F P
0 g
s
.60453 .4885 m .60466 .49825 L .65029 .49444 L p 0 .664 .506 r
F P
0 g
s
.65029 .49444 m .65031 .4848 L .60453 .4885 L p 0 .664 .506 r
F P
0 g
s
.65156 .50386 m .65029 .49444 L .60466 .49825 L p 0 .596 .452 r
F P
0 g
s
.45187 .73096 m .45273 .72679 L .42489 .69972 L p .311 .35 0 r
F P
0 g
s
.42489 .69972 m .45273 .72679 L p .44202 .71657 L 0 .03 0 r
F P
0 g
s
.42312 .70241 m .42489 .69972 L p .41312 .68813 L .232 .335 0 r
F P
0 g
s
.42489 .69972 m .42312 .70241 L .45187 .73096 L p .311 .35 0 r
F P
0 g
s
.42403 .69941 m .42489 .69972 L p .44202 .71657 L 0 .03 0 r
F P
0 g
s
.42489 .69972 m .39998 .66938 L p .41312 .68813 L .232 .335 0 r
F P
0 g
s
.44875 .73712 m .45187 .73096 L .42312 .70241 L p .537 .397 0 r
F P
0 g
s
.32587 .61125 m .31898 .61641 L .33881 .65608 L p .836 .633 0 r
F P
0 g
s
.3344 .66067 m .33881 .65608 L .31898 .61641 L p .907 .645 0 r
F P
0 g
s
.33881 .65608 m .3344 .66067 L .35921 .69889 L p .88 .548 0 r
F P
0 g
s
.45187 .73096 m .44875 .73712 L .48074 .76201 L p .544 .361 0 r
F P
0 g
s
.47692 .77048 m .48074 .76201 L .44875 .73712 L p .617 .356 0 r
F P
0 g
s
.6063 .47967 m .60453 .4885 L .65031 .4848 L p 0 .73 .559 r
F P
0 g
s
.65031 .4848 m .65162 .47622 L .6063 .47967 L p 0 .73 .559 r
F P
0 g
s
.4311 .75876 m .42504 .76384 L .45981 .7908 L p .74 .318 0 r
F P
0 g
s
.45981 .7908 m .46572 .78589 L .4311 .75876 L p .74 .318 0 r
F P
0 g
s
.60466 .49825 m .60669 .50764 L .65156 .50386 L p 0 .596 .452 r
F P
0 g
s
.65397 .51184 m .65156 .50386 L .60669 .50764 L p 0 .508 .382 r
F P
0 g
s
.35732 .70108 m .35921 .69889 L .3344 .66067 L p .93 .489 0 r
F P
0 g
s
.35921 .69889 m .35732 .70108 L .38582 .73578 L p .882 .372 0 r
F P
0 g
s
.38813 .73471 m .35921 .69889 L p .383 .73188 L .882 .372 0 r
F P
0 g
s
.38613 .73564 m .38813 .73471 L p .383 .73188 L .882 .372 0 r
F P
0 g
s
.35832 .70117 m .38614 .73439 L p .36024 .70447 L 0 .128 0 r
F P
0 g
s
.35732 .70108 m .35744 .70109 L p .36024 .70447 L 0 .128 0 r
F P
0 g
s
.35744 .70109 m .35832 .70117 L p .36024 .70447 L 0 .128 0 r
F P
0 g
s
.38582 .73578 m .38613 .73564 L p .383 .73188 L .882 .372 0 r
F P
0 g
s
.65731 .46648 m .65406 .46982 L .69861 .46827 L p 0 .846 .632 r
F P
0 g
s
.69861 .46827 m .70001 .46514 L .65731 .46648 L p 0 .846 .632 r
F P
0 g
s
.70001 .46514 m .69861 .46827 L .74244 .46846 L p 0 .803 .571 r
F P
0 g
s
.35868 .60256 m .37565 .63675 L p .36333 .60642 L .107 .277 0 r
F P
0 g
s
.37565 .63675 m .37897 .63729 L .36269 .60485 L p .107 .277 0 r
F P
0 g
s
.37923 .63971 m .36347 .60871 L p .37826 .63598 L .822 .409 0 r
F P
0 g
s
.37897 .63729 m .37898 .63741 L p .37826 .63598 L .822 .409 0 r
F P
0 g
s
.37898 .63741 m .37923 .63971 L p .37826 .63598 L .822 .409 0 r
F P
0 g
s
.36269 .60485 m .36248 .60472 L p .36333 .60642 L .107 .277 0 r
F P
0 g
s
.36248 .60472 m .35868 .60256 L p .36333 .60642 L .107 .277 0 r
F P
0 g
s
.65397 .51184 m .65719 .51729 L .69954 .51567 L p 0 .356 .26 r
F P
0 g
s
.70105 .51807 m .69954 .51567 L .65719 .51729 L p 0 0 0 r
F P
0 g
s
.69954 .51567 m .70105 .51807 L .74015 .5183 L p 0 0 0 r
F P
0 g
s
.74015 .5183 m .74106 .51588 L .69954 .51567 L p 0 0 0 r
F P
0 g
s
.33079 .56669 m .32116 .56859 L .33443 .60679 L p .718 .665 0 r
F P
0 g
s
.33443 .60679 m .34354 .60362 L .33079 .56669 L p .718 .665 0 r
F P
0 g
s
.32587 .61125 m .33443 .60679 L .32116 .56859 L p .789 .701 0 r
F P
0 g
s
.60975 .47294 m .6063 .47967 L .65162 .47622 L p 0 .806 .622 r
F P
0 g
s
.65162 .47622 m .65406 .46982 L .60975 .47294 L p 0 .806 .622 r
F P
0 g
s
.74244 .46846 m .74169 .46534 L .70001 .46514 L p 0 .803 .571 r
F P
0 g
s
.74169 .46534 m .74244 .46846 L .78427 .47037 L p 0 .754 .509 r
F P
0 g
s
.78641 .47675 m .78427 .47037 L .74244 .46846 L p 0 .669 .456 r
F P
0 g
s
.78746 .48533 m .78729 .49498 L .82774 .49911 L p 0 .492 .326 r
F P
0 g
s
.78729 .49498 m .78593 .50443 L .82548 .50854 L p 0 .436 .291 r
F P
0 g
s
.82548 .50854 m .82774 .49911 L .78729 .49498 L p 0 .436 .291 r
F P
0 g
s
.33988 .56688 m .33079 .56669 L .34354 .60362 L p .629 .614 0 r
F P
0 g
s
.34354 .60362 m .35199 .60214 L .33988 .56688 L p .629 .614 0 r
F P
0 g
s
.38813 .73471 m .38582 .73578 L .41745 .76653 L p .805 .239 0 r
F P
0 g
s
.42028 .76656 m .38813 .73471 L p .41402 .76282 L .805 .239 0 r
F P
0 g
s
.38614 .73439 m .41696 .76372 L p .38929 .73889 L 0 .227 0 r
F P
0 g
s
.38582 .73578 m .38587 .73558 L p .38929 .73889 L 0 .227 0 r
F P
0 g
s
.38587 .73558 m .38614 .73439 L p .38929 .73889 L 0 .227 0 r
F P
0 g
s
.41745 .76653 m .41787 .76654 L p .41402 .76282 L .805 .239 0 r
F P
0 g
s
.41787 .76654 m .42028 .76656 L p .41402 .76282 L .805 .239 0 r
F P
0 g
s
.45273 .72679 m .45187 .73096 L .48264 .75452 L p .379 .35 0 r
F P
0 g
s
.48074 .76201 m .48264 .75452 L .45187 .73096 L p .544 .361 0 r
F P
0 g
s
.60669 .50764 m .61034 .51544 L .65397 .51184 L p 0 .508 .382 r
F P
0 g
s
.65719 .51729 m .65397 .51184 L .61034 .51544 L p 0 .354 .26 r
F P
0 g
s
.74106 .51588 m .74015 .5183 L .77709 .52015 L p 0 0 0 r
F P
0 g
s
.78047 .51792 m .74106 .51588 L p .77623 .52005 L 0 0 0 r
F P
0 g
s
.7393 .51714 m .77385 .51878 L p .74088 .51831 L 0 .717 .482 r
F P
0 g
s
.74015 .5183 m .74013 .51828 L p .74088 .51831 L 0 .717 .482 r
F P
0 g
s
.74013 .51828 m .7393 .51714 L p .74088 .51831 L 0 .717 .482 r
F P
0 g
s
.77709 .52015 m .77717 .5201 L p .77623 .52005 L 0 0 0 r
F P
0 g
s
.77717 .5201 m .78047 .51792 L p .77623 .52005 L 0 0 0 r
F P
0 g
s
.78641 .47675 m .78746 .48533 L .82802 .48935 L p 0 .549 .36 r
F P
0 g
s
.82774 .49911 m .82802 .48935 L .78746 .48533 L p 0 .492 .326 r
F P
0 g
s
.42504 .76384 m .42028 .76656 L .45474 .79287 L p .777 .271 0 r
F P
0 g
s
.45474 .79287 m .45981 .7908 L .42504 .76384 L p .777 .271 0 r
F P
0 g
s
.78593 .50443 m .78356 .51243 L .82153 .51637 L p 0 .362 .246 r
F P
0 g
s
.82153 .51637 m .82548 .50854 L .78593 .50443 L p 0 .362 .246 r
F P
0 g
s
.32116 .56859 m .31225 .57233 L .32587 .61125 L p .789 .701 0 r
F P
0 g
s
.31898 .61641 m .32587 .61125 L .31225 .57233 L p .858 .732 0 r
F P
0 g
s
.37923 .63971 m .37897 .63729 L .39998 .66938 L p .776 .293 0 r
F P
0 g
s
.3997 .67041 m .39998 .66938 L .42489 .69972 L p .659 .141 0 r
F P
0 g
s
.31898 .61641 m .31468 .62159 L .3344 .66067 L p .907 .645 0 r
F P
0 g
s
.33291 .664 m .3344 .66067 L .31468 .62159 L p .964 .6 0 r
F P
0 g
s
.3344 .66067 m .33291 .664 L .35732 .70108 L p .93 .489 0 r
F P
0 g
s
.42489 .69972 m .42403 .69941 L .3997 .67041 L p .659 .141 0 r
F P
0 g
s
.34723 .56916 m .33988 .56688 L .35199 .60214 L p .479 .522 0 r
F P
0 g
s
.35199 .60214 m .35868 .60256 L .34723 .56916 L p .479 .522 0 r
F P
0 g
s
.78427 .47037 m .78641 .47675 L .82628 .48052 L p 0 .617 .4 r
F P
0 g
s
.82802 .48935 m .82628 .48052 L .78641 .47675 L p 0 .549 .36 r
F P
0 g
s
.48074 .76201 m .47692 .77048 L .51143 .78946 L p .598 .307 0 r
F P
0 g
s
.47692 .77048 m .47168 .7788 L .50727 .79841 L p .634 .29 0 r
F P
0 g
s
.50727 .79841 m .51143 .78946 L .47692 .77048 L p .634 .29 0 r
F P
0 g
s
.61445 .46922 m .60975 .47294 L .65406 .46982 L p 0 .883 .689 r
F P
0 g
s
.65406 .46982 m .65731 .46648 L .61445 .46922 L p 0 .883 .689 r
F P
0 g
s
.78356 .51243 m .78047 .51792 L .81641 .52157 L p 0 .236 .167 r
F P
0 g
s
.81641 .52157 m .82153 .51637 L .78356 .51243 L p 0 .236 .167 r
F P
0 g
s
.66095 .46668 m .65731 .46648 L .70001 .46514 L p 0 0 0 r
F P
0 g
s
.70001 .46514 m .70155 .46556 L .66095 .46668 L p 0 0 0 r
F P
0 g
s
.70155 .46556 m .70001 .46514 L .74169 .46534 L p 0 0 0 r
F P
0 g
s
.47168 .7788 m .46572 .78589 L .50192 .80575 L p .663 .269 0 r
F P
0 g
s
.50192 .80575 m .50727 .79841 L .47168 .7788 L p .663 .269 0 r
F P
0 g
s
.35832 .70117 m .35732 .70108 L .33291 .664 L p 0 .017 0 r
F P
0 g
s
.38614 .73439 m .38582 .73578 L .35732 .70108 L p 0 .128 0 r
F P
0 g
s
.31225 .57233 m .30524 .57744 L .31898 .61641 L p .858 .732 0 r
F P
0 g
s
.31468 .62159 m .31898 .61641 L .30524 .57744 L p .937 .757 0 r
F P
0 g
s
.36347 .60871 m .36269 .60485 L .37897 .63729 L p .822 .409 0 r
F P
0 g
s
.48264 .75452 m .48074 .76201 L .51387 .78008 L p .545 .323 0 r
F P
0 g
s
.51143 .78946 m .51387 .78008 L .48074 .76201 L p .598 .307 0 r
F P
0 g
s
.61034 .51544 m .61514 .5206 L .65719 .51729 L p 0 .354 .26 r
F P
0 g
s
.66081 .51949 m .65719 .51729 L .61514 .5206 L p 0 0 0 r
F P
0 g
s
.65719 .51729 m .66081 .51949 L .70105 .51807 L p 0 0 0 r
F P
0 g
s
.70256 .51689 m .70105 .51807 L .66081 .51949 L p 0 .788 .589 r
F P
0 g
s
.70105 .51807 m .70256 .51689 L .7393 .51714 L p 0 .759 .539 r
F P
0 g
s
.7393 .51714 m .74015 .5183 L .70105 .51807 L p 0 .759 .539 r
F P
0 g
s
.74169 .46534 m .74079 .46578 L .70155 .46556 L p 0 0 0 r
F P
0 g
s
.78427 .47037 m .78133 .46706 L .74169 .46534 L p 0 .754 .509 r
F P
0 g
s
.74079 .46578 m .74169 .46534 L .78133 .46706 L p 0 0 0 r
F P
0 g
s
.45118 .72519 m .45273 .72679 L .48235 .749 L p 0 .188 0 r
F P
0 g
s
.48264 .75452 m .48235 .749 L .45273 .72679 L p .379 .35 0 r
F P
0 g
s
.4799 .7462 m .45118 .72519 L p .47164 .74082 L 0 .188 0 r
F P
0 g
s
.48235 .749 m .4799 .7462 L p .47164 .74082 L 0 .188 0 r
F P
0 g
s
.4799 .7462 m .48235 .749 L .51252 .76484 L p .061 .296 0 r
F P
0 g
s
.5089 .76103 m .4799 .7462 L p .50172 .75867 L .061 .296 0 r
F P
0 g
s
.51252 .76484 m .5089 .76103 L p .50172 .75867 L .061 .296 0 r
F P
0 g
s
.53688 .76857 m .5089 .76103 L p .53134 .76918 L .29 .347 0 r
F P
0 g
s
.54196 .77304 m .53688 .76857 L p .53134 .76918 L .29 .347 0 r
F P
0 g
s
.53688 .76857 m .56278 .76804 L p .54388 .76779 L .201 .364 0 r
F P
0 g
s
.5089 .76103 m .53688 .76857 L p .51784 .76397 L .282 0 0 r
F P
0 g
s
.51425 .77149 m .51252 .76484 L .48235 .749 L p .433 .338 0 r
F P
0 g
s
.51252 .76484 m .54196 .77304 L p .5111 .76183 L .29 .347 0 r
F P
0 g
s
.50918 .76133 m .51252 .76484 L p .5111 .76183 L .29 .347 0 r
F P
0 g
s
.53477 .76806 m .53087 .76763 L .52703 .76619 L .282 0 0 r
F
.53688 .76857 m .53645 .7685 L p .53477 .76806 L F P
0 g
s
.53645 .7685 m .53087 .76763 L p .53477 .76806 L .282 0 0 r
F P
0 g
s
.53087 .76763 m .53688 .76857 L p .54388 .76779 L .201 .364 0 r
F P
0 g
s
.53688 .76857 m .54196 .77304 L .56944 .77264 L p .462 .35 0 r
F P
0 g
s
.56944 .77264 m .56278 .76804 L .53688 .76857 L p .462 .35 0 r
F P
0 g
s
.55572 .76697 m .53087 .76763 L p .55336 .76792 L .201 .364 0 r
F P
0 g
s
.56278 .76804 m .55572 .76697 L p .55336 .76792 L .201 .364 0 r
F P
0 g
s
.58574 .75885 m .57771 .7581 L .55572 .76697 L p .558 .391 0 r
F P
0 g
s
.55572 .76697 m .57771 .7581 L p .5584 .7659 L .269 .343 0 r
F P
0 g
s
.56278 .76804 m .58574 .75885 L p .55595 .76691 L .558 .391 0 r
F P
0 g
s
.55572 .76697 m .55578 .76698 L p .55595 .76691 L .558 .391 0 r
F P
0 g
s
.55578 .76698 m .56278 .76804 L p .55595 .76691 L .558 .391 0 r
F P
0 g
s
.54921 .76961 m .55572 .76697 L p .5584 .7659 L .269 .343 0 r
F P
0 g
s
.53087 .76763 m .55572 .76697 L p .53517 .76922 L .286 0 0 r
F P
0 g
s
.54543 .78042 m .54196 .77304 L .51252 .76484 L p .478 .319 0 r
F P
0 g
s
.54921 .76961 m .52474 .77037 L p .54649 .76798 L .286 0 0 r
F P
0 g
s
.55572 .76697 m .54921 .76961 L p .54649 .76798 L .286 0 0 r
F P
0 g
s
.57085 .76085 m .54921 .76961 L p .56915 .76156 L .269 .343 0 r
F P
0 g
s
.57749 .75819 m .57085 .76085 L .56915 .76156 L .269 .343 0 r
F
.57085 .76085 m .57771 .7581 L .5961 .74039 L p .72 .388 0 r
F P
0 g
s
.58903 .74342 m .57085 .76085 L p .58838 .74665 L .72 .388 0 r
F P
0 g
s
.5961 .74039 m .58903 .74342 L p .58838 .74665 L .72 .388 0 r
F P
0 g
s
.58903 .74342 m .60271 .7168 L p .58973 .74001 L .764 .304 0 r
F P
0 g
s
.57085 .76085 m .58903 .74342 L p .57263 .7601 L 0 0 0 r
F P
0 g
s
.57765 .75813 m .57085 .76085 L p .57749 .75819 L .269 .343 0 r
F P
0 g
s
.59388 .76296 m .58574 .75885 L .56278 .76804 L p .576 .323 0 r
F P
0 g
s
.58462 .74914 m .56611 .76674 L p .58137 .75121 L 0 0 0 r
F P
0 g
s
.58903 .74342 m .58462 .74914 L p .58137 .75121 L 0 0 0 r
F P
0 g
s
.58462 .74914 m .58903 .74342 L p .58973 .74001 L .764 .304 0 r
F P
0 g
s
.58903 .74342 m .5961 .74039 L .60981 .71337 L p .864 .318 0 r
F P
0 g
s
.60981 .71337 m .60271 .7168 L .58903 .74342 L p .864 .318 0 r
F P
0 g
s
.59865 .72233 m .58462 .74914 L p .59649 .72791 L .764 .304 0 r
F P
0 g
s
.60271 .7168 m .59865 .72233 L p .59649 .72791 L .764 .304 0 r
F P
0 g
s
.59865 .72233 m .60271 .7168 L .61061 .68112 L p .938 .255 0 r
F P
0 g
s
.60696 .68643 m .59865 .72233 L p .60636 .69576 L .938 .255 0 r
F P
0 g
s
.61061 .68112 m .60696 .68643 L p .60636 .69576 L .938 .255 0 r
F P
0 g
s
.6085 .64242 m .60696 .68643 L p .60991 .65545 L .987 .187 0 r
F P
0 g
s
.61162 .63737 m .6085 .64242 L p .60991 .65545 L .987 .187 0 r
F P
0 g
s
.60256 .59196 m .6085 .64242 L p .60635 .60829 L .998 .124 0 r
F P
0 g
s
.60503 .58729 m .60256 .59196 L p .60635 .60829 L .998 .124 0 r
F P
0 g
s
.61514 .5206 m .61034 .51544 L .5683 .52053 L p 0 .329 .238 r
F P
0 g
s
.5683 .52053 m .57441 .52526 L .61514 .5206 L p 0 .329 .238 r
F P
0 g
s
.62047 .52242 m .61514 .5206 L .57441 .52526 L p 0 0 0 r
F P
0 g
s
.57441 .52526 m .58108 .52652 L .62047 .52242 L p 0 0 0 r
F P
0 g
s
.61514 .5206 m .62047 .52242 L .66081 .51949 L p 0 0 0 r
F P
0 g
s
.6256 .5206 m .62047 .52242 L .58108 .52652 L p 0 .833 .682 r
F P
0 g
s
.63259 .5074 m .63347 .4978 L .67001 .49614 L p 0 .7 .533 r
F P
0 g
s
.58108 .52652 m .5874 .52411 L .6256 .5206 L p 0 .833 .682 r
F P
0 g
s
.62984 .51537 m .6256 .5206 L .5874 .52411 L p 0 .899 .716 r
F P
0 g
s
.5874 .52411 m .59252 .51833 L .62984 .51537 L p 0 .899 .716 r
F P
0 g
s
.61034 .51544 m .60669 .50764 L .56354 .51299 L p 0 .51 .387 r
F P
0 g
s
.56354 .51299 m .5683 .52053 L .61034 .51544 L p 0 .51 .387 r
F P
0 g
s
.63259 .5074 m .62984 .51537 L .59252 .51833 L p 0 .812 .639 r
F P
0 g
s
.59252 .51833 m .59572 .50994 L .63259 .5074 L p 0 .812 .639 r
F P
0 g
s
.60669 .50764 m .60466 .49825 L .56078 .50366 L p 0 .613 .473 r
F P
0 g
s
.56078 .50366 m .56354 .51299 L .60669 .50764 L p 0 .613 .473 r
F P
0 g
s
.63347 .4978 m .63259 .5074 L .59572 .50994 L p 0 .733 .57 r
F P
0 g
s
.58498 .53313 m .58885 .54021 L .63521 .50047 L p .259 0 .607 r
F P
0 g
s
.62984 .51537 m .63259 .5074 L .66929 .50558 L p 0 .773 .591 r
F P
0 g
s
.67001 .49614 m .66929 .50558 L .63259 .5074 L p 0 .7 .533 r
F P
0 g
s
.58885 .54021 m .59068 .54507 L .63764 .50635 L p .202 0 .457 r
F P
0 g
s
.63764 .50635 m .63521 .50047 L .58885 .54021 L p .202 0 .457 r
F P
0 g
s
.66435 .51809 m .66081 .51949 L .62047 .52242 L p 0 .809 .635 r
F P
0 g
s
.62047 .52242 m .6256 .5206 L .66435 .51809 L p 0 .809 .635 r
F P
0 g
s
.6256 .5206 m .62984 .51537 L .66731 .51324 L p 0 .86 .663 r
F P
0 g
s
.66929 .50558 m .66731 .51324 L .62984 .51537 L p 0 .773 .591 r
F P
0 g
s
.59068 .54507 m .59018 .54704 L .63746 .50978 L p .042 0 .047 r
F P
0 g
s
.63746 .50978 m .63764 .50635 L .59068 .54507 L p .042 0 .047 r
F P
0 g
s
.53853 .57576 m .53982 .57946 L .59068 .54507 L p .216 0 .352 r
F P
0 g
s
.59068 .54507 m .58885 .54021 L .53853 .57576 L p .216 0 .352 r
F P
0 g
s
.59018 .54704 m .59068 .54507 L .53982 .57946 L p 0 0 0 r
F P
0 g
s
.59018 .54704 m .5874 .54582 L .63466 .51028 L p .306 0 .799 r
F P
0 g
s
.63466 .51028 m .63746 .50978 L .59018 .54704 L p .306 0 .799 r
F P
0 g
s
.53544 .56932 m .53853 .57576 L .58885 .54021 L p .294 0 .552 r
F P
0 g
s
.58885 .54021 m .58498 .53313 L .53544 .56932 L p .294 0 .552 r
F P
0 g
s
.5874 .54582 m .58269 .54153 L .6296 .50773 L p .364 0 .906 r
F P
0 g
s
.6296 .50773 m .63466 .51028 L .5874 .54582 L p .364 0 .906 r
F P
0 g
s
.53097 .56102 m .53544 .56932 L .58498 .53313 L p .33 0 .651 r
F P
0 g
s
.58108 .52652 m .57441 .52526 L .53598 .53073 L p 0 0 0 r
F P
0 g
s
.53598 .53073 m .54372 .53126 L .58108 .52652 L p 0 0 0 r
F P
0 g
s
.5874 .52411 m .58108 .52652 L .54372 .53126 L p 0 .863 .736 r
F P
0 g
s
.52874 .52659 m .53598 .53073 L .57441 .52526 L p 0 .277 .185 r
F P
0 g
s
.54372 .53126 m .5509 .52808 L .5874 .52411 L p 0 .863 .736 r
F P
0 g
s
.57441 .52526 m .5683 .52053 L .52874 .52659 L p 0 .277 .185 r
F P
0 g
s
.59252 .51833 m .5874 .52411 L .5509 .52808 L p 0 .931 .763 r
F P
0 g
s
.58498 .53313 m .57958 .52478 L .53097 .56102 L p .33 0 .651 r
F P
0 g
s
.56035 .49375 m .56078 .50366 L .60466 .49825 L p 0 .691 .539 r
F P
0 g
s
.60466 .49825 m .60453 .4885 L .56035 .49375 L p 0 .691 .539 r
F P
0 g
s
.63234 .48788 m .63347 .4978 L .59655 .50009 L p 0 .657 .505 r
F P
0 g
s
.59572 .50994 m .59655 .50009 L .63347 .4978 L p 0 .733 .57 r
F P
0 g
s
.5683 .52053 m .56354 .51299 L .52295 .51943 L p 0 .496 .372 r
F P
0 g
s
.52295 .51943 m .52874 .52659 L .5683 .52053 L p 0 .496 .372 r
F P
0 g
s
.5509 .52808 m .55656 .5216 L .59252 .51833 L p 0 .931 .763 r
F P
0 g
s
.59572 .50994 m .59252 .51833 L .55656 .5216 L p 0 .846 .679 r
F P
0 g
s
.57958 .52478 m .58498 .53313 L .63052 .49294 L p .289 0 .689 r
F P
0 g
s
.63347 .4978 m .63234 .48788 L .66935 .48625 L p 0 .633 .479 r
F P
0 g
s
.66935 .48625 m .67001 .49614 L .63347 .4978 L p 0 .633 .479 r
F P
0 g
s
.63521 .50047 m .63052 .49294 L .58498 .53313 L p .259 0 .607 r
F P
0 g
s
.58269 .54153 m .57669 .53473 L .62295 .50245 L p .344 0 .841 r
F P
0 g
s
.62295 .50245 m .6296 .50773 L .58269 .54153 L p .344 0 .841 r
F P
0 g
s
.58889 .53727 m .60256 .59196 L p .59526 .55605 L .986 .067 0 r
F P
0 g
s
.59061 .53314 m .58889 .53727 L p .59526 .55605 L .986 .067 0 r
F P
0 g
s
.56232 .48459 m .56035 .49375 L .60453 .4885 L p 0 .764 .602 r
F P
0 g
s
.60453 .4885 m .6063 .47967 L .56232 .48459 L p 0 .764 .602 r
F P
0 g
s
.56644 .47737 m .56232 .48459 L .6063 .47967 L p 0 .845 .674 r
F P
0 g
s
.6063 .47967 m .60975 .47294 L .56644 .47737 L p 0 .845 .674 r
F P
0 g
s
.57217 .47309 m .56644 .47737 L .60975 .47294 L p 0 .913 .742 r
F P
0 g
s
.52485 .48272 m .52039 .49054 L .56232 .48459 L p 0 .878 .719 r
F P
0 g
s
.56232 .48459 m .56644 .47737 L .52485 .48272 L p 0 .878 .719 r
F P
0 g
s
.58533 .47525 m .59096 .48146 L .55326 .48418 L p 0 .395 .276 r
F P
0 g
s
.56035 .49375 m .56232 .48459 L .52039 .49054 L p 0 .794 .638 r
F P
0 g
s
.55813 .49258 m .55326 .48418 L .59096 .48146 L p 0 .564 .424 r
F P
0 g
s
.59096 .48146 m .59489 .49014 L .55813 .49258 L p 0 .564 .424 r
F P
0 g
s
.56354 .51299 m .56078 .50366 L .51937 .5102 L p 0 .62 .481 r
F P
0 g
s
.51937 .5102 m .52295 .51943 L .56354 .51299 L p 0 .62 .481 r
F P
0 g
s
.55656 .5216 m .5599 .51268 L .59572 .50994 L p 0 .846 .679 r
F P
0 g
s
.59655 .50009 m .59572 .50994 L .5599 .51268 L p 0 .76 .599 r
F P
0 g
s
.56078 .50366 m .56035 .49375 L .51848 .50012 L p 0 .711 .562 r
F P
0 g
s
.52039 .49054 m .51848 .50012 L .56035 .49375 L p 0 .794 .638 r
F P
0 g
s
.56045 .50254 m .55813 .49258 L .59489 .49014 L p 0 .672 .52 r
F P
0 g
s
.59489 .49014 m .59655 .50009 L .56045 .50254 L p 0 .672 .52 r
F P
0 g
s
.51848 .50012 m .51937 .5102 L .56078 .50366 L p 0 .711 .562 r
F P
0 g
s
.5599 .51268 m .56045 .50254 L .59655 .50009 L p 0 .76 .599 r
F P
0 g
s
.62497 .47244 m .62937 .47902 L .59096 .48146 L p 0 .429 .314 r
F P
0 g
s
.59489 .49014 m .59096 .48146 L .62937 .47902 L p 0 .567 .429 r
F P
0 g
s
.62937 .47902 m .63234 .48788 L .59489 .49014 L p 0 .567 .429 r
F P
0 g
s
.59655 .50009 m .59489 .49014 L .63234 .48788 L p 0 .657 .505 r
F P
0 g
s
.56714 .50867 m .57336 .51625 L .61708 .47701 L p .332 0 .809 r
F P
0 g
s
.57336 .51625 m .57958 .52478 L .6242 .48476 L p .312 0 .751 r
F P
0 g
s
.6242 .48476 m .61708 .47701 L .57336 .51625 L p .312 0 .751 r
F P
0 g
s
.63052 .49294 m .6242 .48476 L .57958 .52478 L p .289 0 .689 r
F P
0 g
s
.60975 .47294 m .61445 .46922 L .57217 .47309 L p 0 .913 .742 r
F P
0 g
s
.57877 .47234 m .57217 .47309 L .61445 .46922 L p 0 0 0 r
F P
0 g
s
.58533 .47525 m .57877 .47234 L .61976 .46903 L p 0 .136 .073 r
F P
0 g
s
.61976 .46903 m .62497 .47244 L .58533 .47525 L p 0 .136 .073 r
F P
0 g
s
.59096 .48146 m .58533 .47525 L .62497 .47244 L p 0 .429 .314 r
F P
0 g
s
.56173 .50302 m .56714 .50867 L .6101 .4707 L p .353 0 .873 r
F P
0 g
s
.61708 .47701 m .6101 .4707 L .56714 .50867 L p .332 0 .809 r
F P
0 g
s
.61445 .46922 m .61976 .46903 L .57877 .47234 L p 0 0 0 r
F P
0 g
s
.55813 .49258 m .56045 .50254 L .52635 .50456 L p 0 .675 .515 r
F P
0 g
s
.55786 .50008 m .56173 .50302 L .60416 .46665 L p .361 0 .919 r
F P
0 g
s
.6101 .4707 m .60416 .46665 L .56173 .50302 L p .353 0 .873 r
F P
0 g
s
.55605 .50026 m .55786 .50008 L .60006 .46544 L p 0 0 0 r
F P
0 g
s
.60006 .46544 m .55786 .50008 L p .58215 .48254 L .361 0 .919 r
F P
0 g
s
.60416 .46665 m .60006 .46544 L p .58215 .48254 L .361 0 .919 r
F P
0 g
s
.63776 .42936 m .60006 .46544 L p .62296 .44548 L .31 0 .874 r
F P
0 g
s
.64205 .42883 m .63776 .42936 L p .62296 .44548 L .31 0 .874 r
F P
0 g
s
.66999 .39341 m .63776 .42936 L p .6587 .40761 L .255 0 .791 r
F P
0 g
s
.67457 .39113 m .66999 .39341 L p .6587 .40761 L .255 0 .791 r
F P
0 g
s
.69591 .35908 m .66999 .39341 L p .68821 .37079 L .205 0 .703 r
F P
0 g
s
.70089 .35505 m .69591 .35908 L p .68821 .37079 L .205 0 .703 r
F P
0 g
s
.71486 .3278 m .69591 .35908 L p .71063 .33669 L .164 0 .651 r
F P
0 g
s
.72031 .32199 m .71486 .3278 L p .71063 .33669 L .164 0 .651 r
F P
0 g
s
.7264 .30079 m .71486 .3278 L p .72544 .30678 L .125 0 .646 r
F P
0 g
s
.73225 .29325 m .7264 .30079 L p .72544 .30678 L .125 0 .646 r
F P
0 g
s
.7264 .30079 m .73027 .27905 L p .72781 .29201 L .064 0 .14 r
F P
0 g
s
.71486 .3278 m .7264 .30079 L p .72096 .31496 L 0 0 .048 r
F P
0 g
s
.71486 .3278 m .72031 .32199 L .73225 .29325 L p .125 0 .646 r
F P
0 g
s
.69591 .35908 m .70089 .35505 L .72031 .32199 L p .164 0 .651 r
F P
0 g
s
.72863 .31889 m .72031 .32199 L .70089 .35505 L p .174 0 .906 r
F P
0 g
s
.72441 .30996 m .71304 .33557 L p .72209 .312 L 0 0 .048 r
F P
0 g
s
.7264 .30079 m .72441 .30996 L p .72209 .312 L 0 0 .048 r
F P
0 g
s
.72441 .30996 m .7264 .30079 L p .72781 .29201 L .064 0 .14 r
F P
0 g
s
.74114 .28834 m .73225 .29325 L .72031 .32199 L p .125 0 .902 r
F P
0 g
s
.7264 .30079 m .73225 .29325 L .73629 .26999 L p .081 0 .676 r
F P
0 g
s
.73629 .26999 m .73027 .27905 L .7264 .30079 L p .081 0 .676 r
F P
0 g
s
.72825 .28946 m .72441 .30996 L p .72835 .28915 L .064 0 .14 r
F P
0 g
s
.73027 .27905 m .72825 .28946 L p .72835 .28915 L .064 0 .14 r
F P
0 g
s
.72825 .28946 m .7245 .2747 L p .72813 .28984 L .022 0 0 r
F P
0 g
s
.72441 .30996 m .72825 .28946 L p .72724 .30747 L 0 0 .441 r
F P
0 g
s
.72031 .32199 m .72863 .31889 L .74114 .28834 L p .125 0 .902 r
F P
0 g
s
.73052 .29984 m .72657 .31954 L p .72774 .29857 L 0 0 .441 r
F P
0 g
s
.72825 .28946 m .73052 .29984 L p .72774 .29857 L 0 0 .441 r
F P
0 g
s
.73052 .29984 m .72825 .28946 L p .72813 .28984 L .022 0 0 r
F P
0 g
s
.72825 .28946 m .73027 .27905 L .72633 .26333 L p .028 0 .377 r
F P
0 g
s
.72633 .26333 m .7245 .2747 L .72825 .28946 L p .028 0 .377 r
F P
0 g
s
.72697 .28575 m .73052 .29984 L p .72638 .28255 L .022 0 0 r
F P
0 g
s
.7245 .2747 m .72697 .28575 L p .72638 .28255 L .022 0 0 r
F P
0 g
s
.72697 .28575 m .71588 .27775 L p .72762 .28929 L 0 .009 0 r
F P
0 g
s
.73052 .29984 m .72697 .28575 L p .73358 .30124 L 0 0 .607 r
F P
0 g
s
.72863 .31889 m .73871 .3189 L .75188 .28671 L p .109 0 .935 r
F P
0 g
s
.75188 .28671 m .74114 .28834 L .72863 .31889 L p .109 0 .935 r
F P
0 g
s
.73342 .29497 m .73679 .30878 L p .73006 .293 L 0 0 .607 r
F P
0 g
s
.72697 .28575 m .73342 .29497 L p .73006 .293 L 0 0 .607 r
F P
0 g
s
.72762 .28929 m .73291 .29448 L .72697 .28575 L 0 .009 0 r
F
.73342 .29497 m .73322 .29469 L p .73291 .29448 L F P
0 g
s
.73322 .29469 m .72697 .28575 L p .73291 .29448 L 0 .009 0 r
F P
0 g
s
.72244 .28729 m .73342 .29497 L p .72158 .28335 L 0 .009 0 r
F P
0 g
s
.71588 .27775 m .72244 .28729 L p .72158 .28335 L 0 .009 0 r
F P
0 g
s
.72244 .28729 m .70399 .28614 L p .7233 .29127 L 0 .014 0 r
F P
0 g
s
.73195 .29356 m .74298 .30113 L p .72974 .29221 L 0 .005 .601 r
F P
0 g
s
.72305 .28771 m .73195 .29356 L .72974 .29221 L 0 .005 .601 r
F
.73342 .29497 m .72244 .28729 L p .73678 .29695 L F P
0 g
s
.73871 .3189 m .74919 .32201 L .76305 .28855 L p .094 0 .914 r
F P
0 g
s
.76305 .28855 m .75188 .28671 L .73871 .3189 L p .094 0 .914 r
F P
0 g
s
.75439 .30338 m .74298 .30113 L .73195 .29356 L p 0 .018 .885 r
F P
0 g
s
.73195 .29356 m .71328 .29276 L p .73397 .2948 L 0 0 0 r
F P
0 g
s
.74315 .29571 m .75439 .30338 L p .74161 .29779 L 0 .018 .885 r
F P
0 g
s
.73195 .29356 m .74315 .29571 L p .74161 .29779 L 0 .018 .885 r
F P
0 g
s
.73397 .2948 m .74131 .29553 L .73195 .29356 L 0 0 0 r
F
.74315 .29571 m .74235 .29555 L p .74131 .29553 L F P
0 g
s
.74235 .29555 m .73195 .29356 L p .74131 .29553 L 0 0 0 r
F P
0 g
s
.74298 .30113 m .73342 .29497 L p .73678 .29695 L 0 .005 .601 r
F P
0 g
s
.74623 .31506 m .73679 .30878 L .73342 .29497 L p .002 0 .872 r
F P
0 g
s
.73342 .29497 m .74298 .30113 L .74623 .31506 L p .002 0 .872 r
F P
0 g
s
.72401 .29522 m .74315 .29571 L p .72645 .29406 L 0 0 0 r
F P
0 g
s
.71328 .29276 m .72401 .29522 L p .72645 .29406 L 0 0 0 r
F P
0 g
s
.72401 .29522 m .69768 .30193 L p .72203 .29617 L 0 0 0 r
F P
0 g
s
.73476 .29316 m .75453 .29341 L p .74322 .29408 L 0 .069 .836 r
F P
0 g
s
.72603 .2951 m .73476 .29316 L .74322 .29408 L 0 .069 .836 r
F
.72401 .29522 m .72481 .29507 L p .72603 .2951 L F P
0 g
s
.72481 .29507 m .73476 .29316 L p .72603 .2951 L 0 .069 .836 r
F P
0 g
s
.74315 .29571 m .72401 .29522 L p .7444 .29401 L 0 .069 .836 r
F P
0 g
s
.73476 .29316 m .72401 .29522 L p .72203 .29617 L 0 0 0 r
F P
0 g
s
.72244 .28729 m .72273 .28749 L p .72305 .28771 L 0 .005 .601 r
F P
0 g
s
.72273 .28749 m .73195 .29356 L p .72305 .28771 L 0 .005 .601 r
F P
0 g
s
.73195 .29356 m .72244 .28729 L p .7233 .29127 L 0 .014 0 r
F P
0 g
s
.73225 .29325 m .74114 .28834 L .74548 .2635 L p .075 0 .898 r
F P
0 g
s
.74548 .2635 m .73629 .26999 L .73225 .29325 L p .075 0 .898 r
F P
0 g
s
.70399 .28614 m .71328 .29276 L .73195 .29356 L p 0 .014 0 r
F P
0 g
s
.72401 .29522 m .71328 .29276 L .68762 .29878 L p 0 0 0 r
F P
0 g
s
.68762 .29878 m .69768 .30193 L .72401 .29522 L p 0 0 0 r
F P
0 g
s
.70754 .3006 m .73476 .29316 L p .72122 .29636 L 0 0 0 r
F P
0 g
s
.69768 .30193 m .70754 .3006 L p .72122 .29636 L 0 0 0 r
F P
0 g
s
.69768 .30193 m .66548 .3153 L p .70304 .30217 L 0 0 0 r
F P
0 g
s
.70754 .3006 m .70619 .30078 L p .70304 .30217 L 0 0 0 r
F P
0 g
s
.70619 .30078 m .69768 .30193 L p .70304 .30217 L 0 0 0 r
F P
0 g
s
.67419 .31511 m .70754 .3006 L p .69201 .30603 L 0 0 0 r
F P
0 g
s
.66548 .3153 m .67419 .31511 L p .69201 .30603 L 0 0 0 r
F P
0 g
s
.63661 .3355 m .67419 .31511 L p .65704 .32242 L 0 0 0 r
F P
0 g
s
.62932 .33424 m .63661 .3355 L p .65704 .32242 L 0 0 0 r
F P
0 g
s
.63661 .3355 m .59707 .36017 L p .62266 .34444 L 0 0 0 r
F P
0 g
s
.67419 .31511 m .63661 .3355 L p .66192 .32136 L 0 0 0 r
F P
0 g
s
.64199 .33254 m .63661 .3355 L p .62266 .34444 L 0 0 0 r
F P
0 g
s
.66548 .3153 m .62932 .33424 L p .66925 .31721 L 0 0 0 r
F P
0 g
s
.67419 .31511 m .67296 .31513 L p .66925 .31721 L 0 0 0 r
F P
0 g
s
.67296 .31513 m .66548 .3153 L p .66925 .31721 L 0 0 0 r
F P
0 g
s
.68121 .31058 m .67419 .31511 L p .66192 .32136 L 0 0 0 r
F P
0 g
s
.64199 .33254 m .68121 .31058 L p .6418 .3326 L 0 0 0 r
F P
0 g
s
.63661 .3355 m .63741 .33506 L p .6418 .3326 L 0 0 0 r
F P
0 g
s
.63741 .33506 m .64199 .33254 L p .6418 .3326 L 0 0 0 r
F P
0 g
s
.68121 .31058 m .71585 .29491 L p .67892 .31281 L 0 .186 .753 r
F P
0 g
s
.67419 .31511 m .6752 .31446 L p .67892 .31281 L 0 .186 .753 r
F P
0 g
s
.6752 .31446 m .68121 .31058 L p .67892 .31281 L 0 .186 .753 r
F P
0 g
s
.70754 .3006 m .67419 .31511 L p .69809 .30352 L 0 .186 .753 r
F P
0 g
s
.71585 .29491 m .70754 .3006 L p .69809 .30352 L 0 .186 .753 r
F P
0 g
s
.73476 .29316 m .70754 .3006 L p .72894 .29252 L 0 .143 .854 r
F P
0 g
s
.74114 .28834 m .75188 .28671 L .75662 .26043 L p .064 0 .96 r
F P
0 g
s
.75662 .26043 m .74548 .2635 L .74114 .28834 L p .064 0 .96 r
F P
0 g
s
.74408 .28681 m .73476 .29316 L p .72894 .29252 L 0 .143 .854 r
F P
0 g
s
.75453 .29341 m .74315 .29571 L p .7444 .29401 L 0 .069 .836 r
F P
0 g
s
.73679 .30878 m .73052 .29984 L p .73358 .30124 L 0 0 .607 r
F P
0 g
s
.72657 .31954 m .72441 .30996 L p .72724 .30747 L 0 0 .441 r
F P
0 g
s
.71304 .33557 m .71486 .3278 L p .72096 .31496 L 0 0 .048 r
F P
0 g
s
.69591 .35908 m .71486 .3278 L p .70686 .34242 L 0 0 .124 r
F P
0 g
s
.66999 .39341 m .67457 .39113 L .70089 .35505 L p .205 0 .703 r
F P
0 g
s
.70854 .35382 m .70089 .35505 L .67457 .39113 L p .222 0 .908 r
F P
0 g
s
.69428 .3654 m .69591 .35908 L p .70686 .34242 L 0 0 .124 r
F P
0 g
s
.66999 .39341 m .69591 .35908 L p .68588 .37339 L 0 0 .109 r
F P
0 g
s
.63776 .42936 m .64205 .42883 L .67457 .39113 L p .255 0 .791 r
F P
0 g
s
.68157 .39172 m .67457 .39113 L .64205 .42883 L p .268 0 .905 r
F P
0 g
s
.66847 .39827 m .66999 .39341 L p .68588 .37339 L 0 0 .109 r
F P
0 g
s
.63776 .42936 m .66999 .39341 L p .65858 .40668 L 0 0 .014 r
F P
0 g
s
.63621 .43274 m .63776 .42936 L p .65858 .40668 L 0 0 .014 r
F P
0 g
s
.60006 .46544 m .63776 .42936 L p .62555 .44112 L 0 0 0 r
F P
0 g
s
.60006 .46544 m .60416 .46665 L .64205 .42883 L p .31 0 .874 r
F P
0 g
s
.59839 .46726 m .60006 .46544 L p .62555 .44112 L 0 0 0 r
F P
0 g
s
.52631 .51499 m .52635 .50456 L .56045 .50254 L p 0 .779 .614 r
F P
0 g
s
.52874 .52659 m .52295 .51943 L .48577 .52626 L p 0 .459 .327 r
F P
0 g
s
.48577 .52626 m .49259 .53292 L .52874 .52659 L p 0 .459 .327 r
F P
0 g
s
.53598 .53073 m .52874 .52659 L .49259 .53292 L p 0 .187 .086 r
F P
0 g
s
.52313 .5245 m .52631 .51499 L .5599 .51268 L p 0 .873 .708 r
F P
0 g
s
.56045 .50254 m .5599 .51268 L .52631 .51499 L p 0 .779 .614 r
F P
0 g
s
.49259 .53292 m .50085 .5363 L .53598 .53073 L p 0 .187 .086 r
F P
0 g
s
.50085 .5363 m .49259 .53292 L .46082 .53859 L p 0 .045 0 r
F P
0 g
s
.46256 .55533 m .41394 .57423 L p .44485 .56442 L .448 0 .966 r
F P
0 g
s
.46407 .55833 m .46256 .55533 L p .44485 .56442 L .448 0 .966 r
F P
0 g
s
.52313 .5245 m .51723 .53181 L .50287 .5331 L p 0 .971 .824 r
F P
0 g
s
.54372 .53126 m .53598 .53073 L .50085 .5363 L p 0 0 0 r
F P
0 g
s
.50085 .5363 m .50945 .53592 L .54372 .53126 L p 0 0 0 r
F P
0 g
s
.46082 .53859 m .47001 .54106 L .49226 .53762 L p 0 .045 0 r
F P
0 g
s
.49226 .53762 m .50085 .5363 L p .49337 .53673 L 0 .045 0 r
F P
0 g
s
.49337 .53673 m .46082 .53859 L .49226 .53762 L 0 .045 0 r
F
.48029 .54599 m .46256 .55533 L p .47101 .55129 L .385 0 .924 r
F P
0 g
s
.50945 .53592 m .50085 .5363 L .49226 .53762 L p 0 0 0 r
F P
0 g
s
.49226 .53762 m .47001 .54106 L p .49245 .53814 L 0 0 0 r
F P
0 g
s
.50945 .53592 m .49226 .53762 L .49245 .53814 L 0 0 0 r
F
.47001 .54106 m .47932 .53964 L .49366 .53787 L p F P
0 g
s
.49366 .53787 m .50945 .53592 L p .49245 .53814 L 0 0 0 r
F P
0 g
s
.49245 .53814 m .47001 .54106 L .49366 .53787 L 0 0 0 r
F
.51723 .53181 m .50945 .53592 L .47932 .53964 L p 0 .935 .848 r
F P
0 g
s
.47932 .53964 m .48746 .53449 L .51723 .53181 L p 0 .935 .848 r
F P
0 g
s
.51037 .53015 m .48029 .54599 L p .49592 .53894 L .385 0 .924 r
F P
0 g
s
.51216 .53162 m .51037 .53015 L p .49592 .53894 L .385 0 .924 r
F P
0 g
s
.52313 .5245 m .50287 .5331 L .51187 .52797 L 0 .971 .824 r
F
.52313 .5245 m .49199 .53408 L .49265 .53304 L F
.49259 .53292 m .48577 .52626 L .45296 .53258 L p 0 .389 .234 r
F P
0 g
s
.45296 .53258 m .46082 .53859 L .49259 .53292 L p 0 .389 .234 r
F P
0 g
s
.48746 .53449 m .49334 .52629 L .52313 .5245 L p 0 .971 .824 r
F P
0 g
s
.52295 .51943 m .51937 .5102 L .48129 .51722 L p 0 .614 .468 r
F P
0 g
s
.48129 .51722 m .48577 .52626 L .52295 .51943 L p 0 .614 .468 r
F P
0 g
s
.52631 .51499 m .52313 .5245 L .49334 .52629 L p 0 .891 .721 r
F P
0 g
s
.51037 .53015 m .51216 .53162 L .55786 .50008 L p .283 0 .76 r
F P
0 g
s
.54138 .51051 m .53284 .51648 L .51037 .53015 L .116 0 .101 r
F
.55605 .50026 m .59839 .46726 L p .57718 .48569 L .139 0 .287 r
F P
0 g
s
.55786 .50008 m .55605 .50026 L p .53897 .51204 L .283 0 .76 r
F P
0 g
s
.51723 .53181 m .52313 .5245 L .55656 .5216 L p 0 .954 .799 r
F P
0 g
s
.5599 .51268 m .55656 .5216 L .52313 .5245 L p 0 .873 .708 r
F P
0 g
s
.56173 .50302 m .55786 .50008 L .51216 .53162 L p .406 0 .922 r
F P
0 g
s
.55659 .50358 m .55605 .50026 L p .57718 .48569 L .139 0 .287 r
F P
0 g
s
.55605 .50026 m .51037 .53015 L p .53897 .51204 L .283 0 .76 r
F P
0 g
s
.51037 .53015 m .55605 .50026 L p .54138 .51051 L .116 0 .101 r
F P
0 g
s
.51055 .53206 m .51037 .53015 L p .53284 .51648 L .116 0 .101 r
F P
0 g
s
.49289 .53936 m .49722 .53708 L p .49295 .53953 L .045 0 0 r
F P
0 g
s
.49722 .53708 m .51037 .53015 L p .4977 .53697 L .045 0 0 r
F P
0 g
s
.4977 .53697 m .49295 .53953 L .49722 .53708 L .045 0 0 r
F
.51216 .53162 m .48029 .54599 L .47101 .55129 L .385 0 .924 r
F
.48508 .54347 m .49289 .53936 L p .48402 .54433 L .045 0 0 r
F P
0 g
s
.49295 .53953 m .48402 .54433 L .49289 .53936 L .045 0 0 r
F
.46256 .55533 m .46407 .55833 L .51216 .53162 L p .385 0 .924 r
F P
0 g
s
.48029 .54599 m .48508 .54347 L p .46967 .55205 L .045 0 0 r
F P
0 g
s
.48402 .54433 m .46967 .55205 L .48508 .54347 L .045 0 0 r
F
.46256 .55533 m .48029 .54599 L p .47514 .54886 L F P
0 g
s
.46262 .55584 m .46256 .55533 L p .47514 .54886 L .045 0 0 r
F P
0 g
s
.46967 .55205 m .46262 .55584 L .48029 .54599 L .045 0 0 r
F
.50287 .5331 m .49199 .53408 L p .51187 .52797 L 0 .971 .824 r
F P
0 g
s
.49199 .53408 m .48746 .53449 L p .49265 .53304 L 0 .971 .824 r
F P
0 g
s
.5509 .52808 m .54372 .53126 L .50945 .53592 L p 0 .9 .794 r
F P
0 g
s
.50945 .53592 m .51723 .53181 L .5509 .52808 L p 0 .9 .794 r
F P
0 g
s
.51566 .53622 m .51216 .53162 L .48346 .54756 L p .445 0 .904 r
F P
0 g
s
.48346 .54756 m .46407 .55833 L p .48132 .55093 L .445 0 .904 r
F P
0 g
s
.51566 .53622 m .48346 .54756 L .48132 .55093 L .445 0 .904 r
F
.60006 .46544 m .59839 .46726 L .55605 .50026 L p 0 0 0 r
F P
0 g
s
.55656 .5216 m .5509 .52808 L .51723 .53181 L p 0 .954 .799 r
F P
0 g
s
.51216 .53162 m .51566 .53622 L .56173 .50302 L p .406 0 .922 r
F P
0 g
s
.56714 .50867 m .56173 .50302 L .51566 .53622 L p .393 0 .848 r
F P
0 g
s
.55942 .50962 m .55659 .50358 L .59939 .4719 L p .235 0 .538 r
F P
0 g
s
.59839 .46726 m .59939 .4719 L .55659 .50358 L p .139 0 .287 r
F P
0 g
s
.59939 .4719 m .60296 .47874 L .55942 .50962 L p .235 0 .538 r
F P
0 g
s
.51566 .53622 m .5204 .54331 L .56714 .50867 L p .393 0 .848 r
F P
0 g
s
.57336 .51625 m .56714 .50867 L .5204 .54331 L p .374 0 .781 r
F P
0 g
s
.56418 .51758 m .55942 .50962 L .60296 .47874 L p .276 0 .649 r
F P
0 g
s
.5204 .54331 m .52573 .55196 L .57336 .51625 L p .374 0 .781 r
F P
0 g
s
.57958 .52478 m .57336 .51625 L .52573 .55196 L p .354 0 .72 r
F P
0 g
s
.57022 .52636 m .56418 .51758 L .60863 .48688 L p .302 0 .72 r
F P
0 g
s
.60296 .47874 m .6075 .48526 L p .56418 .51758 L .276 0 .649 r
F P
0 g
s
.6075 .48526 m .60863 .48688 L .60675 .48817 L p .276 0 .649 r
F P
0 g
s
.60675 .48817 m .56418 .51758 L p .6075 .48526 L .276 0 .649 r
F P
0 g
s
.53129 .47772 m .52485 .48272 L .56644 .47737 L p 0 .937 .79 r
F P
0 g
s
.56644 .47737 m .57217 .47309 L .53129 .47772 L p 0 .937 .79 r
F P
0 g
s
.57877 .47234 m .58533 .47525 L .54652 .47847 L p 0 .044 0 r
F P
0 g
s
.55326 .48418 m .54652 .47847 L .58533 .47525 L p 0 .395 .276 r
F P
0 g
s
.52573 .55196 m .53097 .56102 L .57958 .52478 L p .354 0 .72 r
F P
0 g
s
.57669 .53473 m .57022 .52636 L .61561 .49517 L p .323 0 .78 r
F P
0 g
s
.61561 .49517 m .62295 .50245 L .57669 .53473 L p .323 0 .78 r
F P
0 g
s
.60863 .48688 m .61561 .49517 L .60338 .50357 L p .302 0 .72 r
F P
0 g
s
.60338 .50357 m .57022 .52636 L p .59604 .49982 L .302 0 .72 r
F P
0 g
s
.60863 .48688 m .60338 .50357 L .59604 .49982 L .302 0 .72 r
F
.61976 .46903 m .61445 .46922 L .65731 .46648 L p 0 0 0 r
F P
0 g
s
.65731 .46648 m .66095 .46668 L .61976 .46903 L p 0 0 0 r
F P
0 g
s
.62497 .47244 m .61976 .46903 L .66095 .46668 L p 0 .188 .126 r
F P
0 g
s
.6101 .4707 m .61708 .47701 L .65621 .43607 L p .29 0 .836 r
F P
0 g
s
.60416 .46665 m .6101 .4707 L .64849 .43118 L p .312 0 .893 r
F P
0 g
s
.65621 .43607 m .64849 .43118 L .6101 .4707 L p .29 0 .836 r
F P
0 g
s
.64849 .43118 m .64205 .42883 L .60416 .46665 L p .312 0 .893 r
F P
0 g
s
.59939 .4719 m .59839 .46726 L .63621 .43274 L p .134 0 .404 r
F P
0 g
s
.63776 .42936 m .63621 .43274 L .59839 .46726 L p 0 0 0 r
F P
0 g
s
.60296 .47874 m .59939 .4719 L .63766 .43854 L p .202 0 .595 r
F P
0 g
s
.63621 .43274 m .63766 .43854 L .59939 .4719 L p .134 0 .404 r
F P
0 g
s
.63766 .43854 m .64193 .44598 L .61519 .46846 L p .202 0 .595 r
F P
0 g
s
.61519 .46846 m .60296 .47874 L p .61357 .46645 L .202 0 .595 r
F P
0 g
s
.63766 .43854 m .61519 .46846 L .61357 .46645 L .202 0 .595 r
F
.66731 .51324 m .66435 .51809 L .6256 .5206 L p 0 .86 .663 r
F P
0 g
s
.66448 .47044 m .66095 .46668 L .70155 .46556 L p 0 .207 .147 r
F P
0 g
s
.70155 .46556 m .70301 .4695 L .66448 .47044 L p 0 .207 .147 r
F P
0 g
s
.66095 .46668 m .66448 .47044 L .62497 .47244 L p 0 .188 .126 r
F P
0 g
s
.66742 .47727 m .66448 .47044 L .70301 .4695 L p 0 .432 .316 r
F P
0 g
s
.62937 .47902 m .62497 .47244 L .66448 .47044 L p 0 .44 .325 r
F P
0 g
s
.66448 .47044 m .66742 .47727 L .62937 .47902 L p 0 .44 .325 r
F P
0 g
s
.63234 .48788 m .62937 .47902 L .66742 .47727 L p 0 .556 .417 r
F P
0 g
s
.66742 .47727 m .66935 .48625 L .63234 .48788 L p 0 .556 .417 r
F P
0 g
s
.63052 .49294 m .63521 .50047 L .67684 .45838 L p .222 0 .652 r
F P
0 g
s
.63521 .50047 m .63764 .50635 L .67984 .4651 L p .177 0 .53 r
F P
0 g
s
.67984 .4651 m .67684 .45838 L .63521 .50047 L p .177 0 .53 r
F P
0 g
s
.70301 .4695 m .70419 .47647 L .66742 .47727 L p 0 .432 .316 r
F P
0 g
s
.66935 .48625 m .66742 .47727 L .70419 .47647 L p 0 .535 .393 r
F P
0 g
s
.63764 .50635 m .63746 .50978 L .67996 .46987 L p .07 0 .229 r
F P
0 g
s
.67996 .46987 m .67984 .4651 L .63764 .50635 L p .07 0 .229 r
F P
0 g
s
.63746 .50978 m .63466 .51028 L .67715 .47204 L p .218 0 .593 r
F P
0 g
s
.67715 .47204 m .67996 .46987 L .63746 .50978 L p .218 0 .593 r
F P
0 g
s
.63466 .51028 m .6296 .50773 L .67176 .47126 L p .321 0 .909 r
F P
0 g
s
.67176 .47126 m .67715 .47204 L .63466 .51028 L p .321 0 .909 r
F P
0 g
s
.6296 .50773 m .62714 .50578 L p .66449 .46763 L .302 0 .866 r
F P
0 g
s
.62714 .50578 m .62295 .50245 L .6263 .49964 L p .302 0 .866 r
F P
0 g
s
.6263 .49964 m .66449 .46763 L p .62714 .50578 L .302 0 .866 r
F P
0 g
s
.61708 .47701 m .6242 .48476 L .66419 .44285 L p .269 0 .781 r
F P
0 g
s
.6242 .48476 m .63052 .49294 L .67138 .45062 L p .248 0 .724 r
F P
0 g
s
.67138 .45062 m .66419 .44285 L .6242 .48476 L p .248 0 .724 r
F P
0 g
s
.66419 .44285 m .65621 .43607 L .61708 .47701 L p .269 0 .781 r
F P
0 g
s
.67684 .45838 m .67138 .45062 L .63052 .49294 L p .222 0 .652 r
F P
0 g
s
.66449 .46763 m .67176 .47126 L .6296 .50773 L p .302 0 .866 r
F P
0 g
s
.60863 .48688 m .60296 .47874 L .61214 .47102 L p .236 0 .689 r
F P
0 g
s
.61214 .47102 m .64193 .44598 L p .61667 .477 L .236 0 .689 r
F P
0 g
s
.61667 .477 m .60863 .48688 L .61214 .47102 L .236 0 .689 r
F
.62295 .50245 m .61561 .49517 L .65635 .46161 L p .28 0 .808 r
F P
0 g
s
.65635 .46161 m .66449 .46763 L .62295 .50245 L p .28 0 .808 r
F P
0 g
s
.61561 .49517 m .60863 .48688 L .64846 .45404 L p .26 0 .753 r
F P
0 g
s
.64193 .44598 m .64846 .45404 L .61807 .4791 L p .236 0 .689 r
F P
0 g
s
.61807 .4791 m .60863 .48688 L p .61667 .477 L .236 0 .689 r
F P
0 g
s
.64193 .44598 m .61807 .4791 L .61667 .477 L .236 0 .689 r
F
.64846 .45404 m .65635 .46161 L .61561 .49517 L p .26 0 .753 r
F P
0 g
s
.62185 .46719 m .63838 .50961 L p .62026 .4556 L .172 .009 0 r
F P
0 g
s
.63838 .50961 m .64137 .51728 L p .63913 .51074 L .172 .009 0 r
F P
0 g
s
.62026 .4556 m .63838 .50961 L .63913 .51074 L .172 .009 0 r
F
.62187 .45587 m .62026 .4556 L p .63658 .50329 L F P
0 g
s
.63717 .50407 m .63658 .50329 L .64137 .51728 L .172 .009 0 r
F
.62187 .45587 m .63658 .50329 L .63717 .50407 L F
.62026 .4556 m .61712 .45507 L .62185 .46719 L p F P
0 g
s
.64137 .51728 m .64559 .51956 L .63637 .49479 L p .172 .009 0 r
F P
0 g
s
.63637 .49479 m .62187 .45587 L p .63283 .49039 L .172 .009 0 r
F P
0 g
s
.63283 .49039 m .64137 .51728 L .63637 .49479 L .172 .009 0 r
F
.6095 .45613 m .60002 .45889 L .62271 .51345 L p .613 .018 0 r
F P
0 g
s
.62271 .51345 m .62446 .51765 L p .62354 .51388 L .613 .018 0 r
F P
0 g
s
.6095 .45613 m .62271 .51345 L .62354 .51388 L .613 .018 0 r
F
.59592 .52851 m .5964 .52825 L p .58056 .46788 L .854 .021 0 r
F P
0 g
s
.5964 .52825 m .60409 .52401 L .60001 .51428 L p .854 .021 0 r
F P
0 g
s
.60001 .51428 m .58056 .46788 L p .5964 .52825 L .854 .021 0 r
F P
0 g
s
.56768 .4809 m .58889 .53727 L p .57656 .50088 L .946 .019 0 r
F P
0 g
s
.56862 .47747 m .56768 .4809 L p .57656 .50088 L .946 .019 0 r
F P
0 g
s
.53885 .47623 m .53129 .47772 L .57217 .47309 L p 0 .651 .592 r
F P
0 g
s
.57217 .47309 m .57877 .47234 L .53885 .47623 L p 0 .651 .592 r
F P
0 g
s
.54652 .47847 m .53885 .47623 L .57877 .47234 L p 0 .044 0 r
F P
0 g
s
.56768 .4809 m .56862 .47747 L .53975 .42301 L p .86 0 0 r
F P
0 g
s
.53975 .42301 m .53962 .42433 L p .56768 .4809 L .86 0 0 r
F P
0 g
s
.50801 .3686 m .50503 .3725 L .53975 .42301 L p .933 0 .037 r
F P
0 g
s
.5395 .42557 m .53975 .42301 L .50503 .3725 L p .723 0 0 r
F P
0 g
s
.50503 .3725 m .50801 .3686 L .46822 .32503 L p .903 0 .046 r
F P
0 g
s
.46822 .32503 m .46582 .32853 L .50503 .3725 L p .903 0 .046 r
F P
0 g
s
.50528 .37412 m .50503 .3725 L .46582 .32853 L p .553 0 0 r
F P
0 g
s
.46629 .32921 m .50528 .37412 L p .47972 .34459 L .553 0 0 r
F P
0 g
s
.46582 .32853 m .46629 .32921 L p .47972 .34459 L .553 0 0 r
F P
0 g
s
.46582 .32853 m .46822 .32503 L .42576 .29015 L p .85 0 .042 r
F P
0 g
s
.4238 .29329 m .46582 .32853 L p .4379 .30178 L .85 0 .042 r
F P
0 g
s
.42576 .29015 m .4238 .29329 L p .4379 .30178 L .85 0 .042 r
F P
0 g
s
.3825 .26544 m .38089 .26834 L .4238 .29329 L p .77 0 .025 r
F P
0 g
s
.4238 .29329 m .38089 .26834 L p .4102 .28511 L 0 0 .205 r
F P
0 g
s
.4238 .29329 m .42576 .29015 L .3825 .26544 L p .77 0 .025 r
F P
0 g
s
.42421 .29313 m .4238 .29329 L p .4102 .28511 L 0 0 .205 r
F P
0 g
s
.46629 .32921 m .46582 .32853 L .4238 .29329 L p 0 0 .131 r
F P
0 g
s
.46629 .32921 m .42421 .29313 L p .45447 .31566 L .319 0 .191 r
F P
0 g
s
.46961 .32694 m .46629 .32921 L p .45447 .31566 L .319 0 .191 r
F P
0 g
s
.50528 .37412 m .46629 .32921 L p .49411 .35803 L .227 0 .13 r
F P
0 g
s
.50503 .3725 m .50528 .37412 L p .51768 .39199 L .723 0 0 r
F P
0 g
s
.50503 .3725 m .53858 .42419 L .53861 .4242 L .723 0 0 r
F
.50875 .3732 m .50528 .37412 L p .49411 .35803 L .227 0 .13 r
F P
0 g
s
.50528 .37412 m .53858 .42419 L p .51768 .39199 L .723 0 0 r
F P
0 g
s
.53858 .42419 m .5395 .42557 L p .53861 .4242 L .723 0 0 r
F P
0 g
s
.49779 .41968 m .50409 .42053 L .53531 .42299 L p 0 .566 0 r
F P
0 g
s
.5395 .42557 m .50528 .37412 L p .52955 .40774 L .112 0 .073 r
F P
0 g
s
.53962 .42433 m .5395 .42557 L p .54147 .42909 L .86 0 0 r
F P
0 g
s
.54201 .43049 m .54672 .43973 L p .53962 .42433 L .86 0 0 r
F P
0 g
s
.54672 .43973 m .56768 .4809 L p .54547 .43611 L .86 0 0 r
F P
0 g
s
.53962 .42433 m .54672 .43973 L .54547 .43611 L .86 0 0 r
F
.55901 .47608 m .55914 .47575 L p .51271 .46522 L .262 .319 0 r
F P
0 g
s
.5395 .42557 m .54201 .43049 L p .54147 .42909 L .86 0 0 r
F P
0 g
s
.56768 .4809 m .56425 .47416 L p .55951 .4625 L 0 0 .022 r
F P
0 g
s
.56425 .47416 m .5395 .42557 L p .55921 .46194 L 0 0 .022 r
F P
0 g
s
.55921 .46194 m .55951 .4625 L .56425 .47416 L 0 0 .022 r
F
.54345 .41873 m .53975 .42301 L .56862 .47747 L p .945 0 .015 r
F P
0 g
s
.59061 .53314 m .59592 .52851 L .57312 .47293 L p .944 .022 0 r
F P
0 g
s
.57312 .47293 m .56862 .47747 L .59061 .53314 L p .944 .022 0 r
F P
0 g
s
.58889 .53727 m .59061 .53314 L .56862 .47747 L p .946 .019 0 r
F P
0 g
s
.56862 .47747 m .57312 .47293 L .54345 .41873 L p .945 0 .015 r
F P
0 g
s
.57668 .4827 m .57048 .48273 L .55276 .44647 L p .41 0 .025 r
F P
0 g
s
.55276 .44647 m .54277 .42605 L p .55394 .44471 L .41 0 .025 r
F P
0 g
s
.55394 .44471 m .57668 .4827 L .55276 .44647 L .41 0 .025 r
F
.57312 .47293 m .58056 .46788 L .55006 .41332 L p .862 0 .02 r
F P
0 g
s
.55869 .40749 m .55006 .41332 L .58056 .46788 L p .789 0 .022 r
F P
0 g
s
.55006 .41332 m .54345 .41873 L .57312 .47293 L p .862 0 .02 r
F P
0 g
s
.5395 .42557 m .54277 .42605 L .54738 .43548 L p 0 0 .022 r
F P
0 g
s
.5395 .42557 m .54738 .43548 L .54694 .4373 L 0 0 .022 r
F
.5395 .42557 m .55276 .44647 L .5537 .45177 L F
.54277 .42605 m .54917 .42435 L .57668 .4827 L p .41 0 .025 r
F P
0 g
s
.50409 .42053 m .5096 .42533 L .54937 .42743 L p .184 .692 0 r
F P
0 g
s
.58056 .46788 m .57312 .47293 L .59592 .52851 L p .854 .021 0 r
F P
0 g
s
.58546 .48079 m .57668 .4827 L .54917 .42435 L p .582 0 .024 r
F P
0 g
s
.54917 .42435 m .55782 .42068 L .58546 .48079 L p .582 0 .024 r
F P
0 g
s
.54937 .42743 m .56035 .42801 L p .54958 .42658 L .184 .692 0 r
F P
0 g
s
.53975 .42301 m .54345 .41873 L .50801 .3686 L p .933 0 .037 r
F P
0 g
s
.54277 .42605 m .5395 .42557 L p .52955 .40774 L .112 0 .073 r
F P
0 g
s
.50409 .42053 m .54937 .42743 L .54947 .42703 L .184 .692 0 r
F
.54947 .42656 m .54947 .42703 L .54958 .42658 L F
.54345 .41873 m .55006 .41332 L .51377 .36296 L p .864 0 .055 r
F P
0 g
s
.51377 .36296 m .50801 .3686 L .54345 .41873 L p .864 0 .055 r
F P
0 g
s
.54917 .42435 m .54277 .42605 L .50875 .3732 L p .466 0 .076 r
F P
0 g
s
.50528 .37412 m .50875 .3732 L .54277 .42605 L p .112 0 .073 r
F P
0 g
s
.50409 .42053 m .54947 .42703 L .54947 .42656 L .184 .692 0 r
F
.54738 .43548 m .55276 .44647 L p .54694 .4373 L 0 0 .022 r
F P
0 g
s
.55276 .44647 m .57048 .48273 L p .5537 .45177 L 0 0 .022 r
F P
0 g
s
.56035 .42801 m .54937 .42743 L p .56077 .43452 L .236 .614 0 r
F P
0 g
s
.54937 .42743 m .5096 .42533 L p .5503 .43264 L .236 .614 0 r
F P
0 g
s
.5503 .43264 m .56077 .43452 L .54937 .42743 L .236 .614 0 r
F
.56354 .43502 m .56229 .43227 L p .56077 .43452 L F P
0 g
s
.56229 .43227 m .56035 .42801 L p .56077 .43452 L .236 .614 0 r
F P
0 g
s
.5096 .42533 m .51359 .43341 L .5479 .43451 L p .236 .614 0 r
F P
0 g
s
.5479 .43451 m .56354 .43502 L p .54687 .43203 L .236 .614 0 r
F P
0 g
s
.5096 .42533 m .5479 .43451 L .54687 .43203 L .236 .614 0 r
F
.57048 .48273 m .56768 .4809 L p .55951 .4625 L 0 0 .022 r
F P
0 g
s
.55914 .47575 m .56264 .46665 L .55593 .46645 L p .262 .319 0 r
F P
0 g
s
.55593 .46645 m .51271 .46522 L p .55914 .47575 L .262 .319 0 r
F P
0 g
s
.55006 .41332 m .55869 .40749 L .52154 .35632 L p .795 0 .063 r
F P
0 g
s
.52154 .35632 m .51377 .36296 L .55006 .41332 L p .795 0 .063 r
F P
0 g
s
.55782 .42068 m .54917 .42435 L .51501 .36984 L p .61 0 .073 r
F P
0 g
s
.50875 .3732 m .51501 .36984 L .54917 .42435 L p .466 0 .076 r
F P
0 g
s
.53531 .42299 m .55593 .42461 L p .53487 .42283 L 0 .566 0 r
F P
0 g
s
.49779 .41968 m .53531 .42299 L .53487 .42283 L 0 .566 0 r
F
.51525 .45476 m .51271 .46522 L .55593 .46645 L p .264 .41 0 r
F P
0 g
s
.55593 .46645 m .56264 .46665 L p .55529 .4648 L .264 .41 0 r
F P
0 g
s
.51525 .45476 m .55593 .46645 L .55529 .4648 L .264 .41 0 r
F
.58889 .53727 m .56768 .4809 L p .58293 .51969 L 0 .001 0 r
F P
0 g
s
.59105 .54036 m .58889 .53727 L p .58293 .51969 L 0 .001 0 r
F P
0 g
s
.60256 .59196 m .58889 .53727 L p .59896 .57648 L 0 0 0 r
F P
0 g
s
.60256 .59196 m .60503 .58729 L .59061 .53314 L p .986 .067 0 r
F P
0 g
s
.60402 .59616 m .60256 .59196 L p .59896 .57648 L 0 0 0 r
F P
0 g
s
.6085 .64242 m .60256 .59196 L p .60716 .62995 L 0 0 0 r
F P
0 g
s
.6085 .64242 m .61162 .63737 L .60503 .58729 L p .998 .124 0 r
F P
0 g
s
.60929 .6476 m .6085 .64242 L p .60716 .62995 L 0 0 0 r
F P
0 g
s
.60696 .68643 m .6085 .64242 L p .60754 .67747 L 0 0 0 r
F P
0 g
s
.60696 .68643 m .61061 .68112 L .61162 .63737 L p .987 .187 0 r
F P
0 g
s
.60713 .69251 m .60696 .68643 L p .60754 .67747 L 0 0 0 r
F P
0 g
s
.59865 .72233 m .60696 .68643 L p .60071 .71697 L 0 0 0 r
F P
0 g
s
.5982 .72924 m .59865 .72233 L p .60071 .71697 L 0 0 0 r
F P
0 g
s
.58462 .74914 m .59865 .72233 L p .58773 .74716 L 0 0 0 r
F P
0 g
s
.57771 .7581 m .58574 .75885 L .60486 .74043 L p .713 .338 0 r
F P
0 g
s
.60486 .74043 m .5961 .74039 L .57771 .7581 L p .713 .338 0 r
F P
0 g
s
.58349 .75681 m .58462 .74914 L p .58773 .74716 L 0 0 0 r
F P
0 g
s
.56611 .76674 m .57085 .76085 L p .57263 .7601 L 0 0 0 r
F P
0 g
s
.54921 .76961 m .57085 .76085 L p .55199 .77127 L .173 0 0 r
F P
0 g
s
.54196 .77304 m .54543 .78042 L .57482 .78014 L p .517 .294 0 r
F P
0 g
s
.57482 .78014 m .56944 .77264 L .54196 .77304 L p .517 .294 0 r
F P
0 g
s
.54412 .77562 m .54921 .76961 L p .55199 .77127 L .173 0 0 r
F P
0 g
s
.57771 .7581 m .57765 .75813 L p .57749 .75819 L .269 .343 0 r
F P
0 g
s
.52474 .77037 m .53087 .76763 L p .53517 .76922 L .286 0 0 r
F P
0 g
s
.53087 .76763 m .50384 .7606 L p .52703 .76619 L .282 0 0 r
F P
0 g
s
.5089 .76103 m .50918 .76133 L p .5111 .76183 L .29 .347 0 r
F P
0 g
s
.50384 .7606 m .5089 .76103 L p .51784 .76397 L .282 0 0 r
F P
0 g
s
.4799 .7462 m .5089 .76103 L p .49013 .75286 L .602 .026 0 r
F P
0 g
s
.48235 .749 m .48264 .75452 L .51425 .77149 L p .433 .338 0 r
F P
0 g
s
.4756 .74653 m .4799 .7462 L p .49013 .75286 L .602 .026 0 r
F P
0 g
s
.45118 .72519 m .4799 .7462 L p .4618 .73518 L .758 .19 0 r
F P
0 g
s
.45273 .72679 m .45118 .72519 L p .44257 .71709 L 0 .03 0 r
F P
0 g
s
.4474 .72641 m .45118 .72519 L p .4618 .73518 L .758 .19 0 r
F P
0 g
s
.45118 .72519 m .42403 .69941 L p .44257 .71709 L 0 .03 0 r
F P
0 g
s
.41696 .76372 m .41745 .76653 L .38582 .73578 L p 0 .227 0 r
F P
0 g
s
.42028 .76656 m .41745 .76653 L .45121 .79178 L p .662 .077 0 r
F P
0 g
s
.45474 .79287 m .42028 .76656 L p .44774 .78895 L .662 .077 0 r
F P
0 g
s
.41696 .76372 m .4497 .78765 L p .42087 .76878 L .128 .293 0 r
F P
0 g
s
.41745 .76653 m .41738 .76615 L p .42087 .76878 L .128 .293 0 r
F P
0 g
s
.41738 .76615 m .41696 .76372 L p .42087 .76878 L .128 .293 0 r
F P
0 g
s
.45121 .79178 m .45171 .79194 L p .44774 .78895 L .662 .077 0 r
F P
0 g
s
.45171 .79194 m .45474 .79287 L p .44774 .78895 L .662 .077 0 r
F P
0 g
s
.77385 .51878 m .77709 .52015 L .74015 .5183 L p 0 .717 .482 r
F P
0 g
s
.34723 .56916 m .35868 .60256 L p .35192 .57357 L .106 .267 0 r
F P
0 g
s
.35868 .60256 m .36269 .60485 L .35184 .57323 L p .106 .267 0 r
F P
0 g
s
.36347 .60871 m .35308 .57856 L p .36259 .60457 L .786 .469 0 r
F P
0 g
s
.36269 .60485 m .3627 .60489 L p .36259 .60457 L .786 .469 0 r
F P
0 g
s
.3627 .60489 m .36347 .60871 L p .36259 .60457 L .786 .469 0 r
F P
0 g
s
.35184 .57323 m .35179 .57318 L p .35192 .57357 L .106 .267 0 r
F P
0 g
s
.35179 .57318 m .34723 .56916 L p .35192 .57357 L .106 .267 0 r
F P
0 g
s
.46572 .78589 m .45981 .7908 L .49609 .81049 L p .689 .235 0 r
F P
0 g
s
.49609 .81049 m .50192 .80575 L .46572 .78589 L p .689 .235 0 r
F P
0 g
s
.31468 .62159 m .31358 .6261 L .33291 .664 L p .964 .6 0 r
F P
0 g
s
.33456 .66562 m .33291 .664 L .31358 .6261 L p 0 0 0 r
F P
0 g
s
.33291 .664 m .33456 .66562 L .35832 .70117 L p 0 .017 0 r
F P
0 g
s
.37636 .6437 m .37923 .63971 L .3997 .67041 L p .917 .531 0 r
F P
0 g
s
.3997 .67041 m .39657 .67349 L .37636 .6437 L p .917 .531 0 r
F P
0 g
s
.39657 .67349 m .3997 .67041 L .42403 .69941 L p .882 .427 0 r
F P
0 g
s
.42403 .69941 m .42062 .70157 L .39657 .67349 L p .882 .427 0 r
F P
0 g
s
.42062 .70157 m .42403 .69941 L .45118 .72519 L p .836 .317 0 r
F P
0 g
s
.78427 .47037 m .82274 .47381 L p .78354 .46742 L 0 .7 .446 r
F P
0 g
s
.78151 .46726 m .78427 .47037 L p .78354 .46742 L 0 .7 .446 r
F P
0 g
s
.82274 .47381 m .81786 .47014 L .78133 .46706 L p 0 .7 .446 r
F P
0 g
s
.81229 .47001 m .77797 .46731 L p .81585 .46999 L 0 0 0 r
F P
0 g
s
.81786 .47014 m .81755 .47013 L p .81585 .46999 L 0 0 0 r
F P
0 g
s
.81755 .47013 m .81229 .47001 L p .81585 .46999 L 0 0 0 r
F P
0 g
s
.78133 .46706 m .78151 .46726 L p .78354 .46742 L 0 .7 .446 r
F P
0 g
s
.82628 .48052 m .82274 .47381 L .78427 .47037 L p 0 .617 .4 r
F P
0 g
s
.51387 .78008 m .51425 .77149 L .48264 .75452 L p .545 .323 0 r
F P
0 g
s
.78047 .51792 m .77709 .52015 L .8108 .52342 L p 0 0 0 r
F P
0 g
s
.81641 .52157 m .78047 .51792 L p .80916 .52313 L 0 0 0 r
F P
0 g
s
.77385 .51878 m .80544 .52164 L p .7785 .52023 L 0 .661 .417 r
F P
0 g
s
.77709 .52015 m .77691 .52008 L p .7785 .52023 L 0 .661 .417 r
F P
0 g
s
.77691 .52008 m .77385 .51878 L p .7785 .52023 L 0 .661 .417 r
F P
0 g
s
.8108 .52342 m .81113 .52331 L p .80916 .52313 L 0 0 0 r
F P
0 g
s
.81113 .52331 m .81641 .52157 L p .80916 .52313 L 0 0 0 r
F P
0 g
s
.30524 .57744 m .30107 .58324 L .31468 .62159 L p .937 .757 0 r
F P
0 g
s
.31468 .62159 m .30107 .58324 L p .31344 .62562 L .984 .708 0 r
F P
0 g
s
.31359 .62605 m .31468 .62159 L p .31344 .62562 L .984 .708 0 r
F P
0 g
s
.30107 .58324 m .30034 .58896 L .31358 .6261 L p .984 .708 0 r
F P
0 g
s
.30316 .59383 m .31585 .62932 L p .30051 .5894 L 0 0 0 r
F P
0 g
s
.30034 .58896 m .30037 .58901 L p .30051 .5894 L 0 0 0 r
F P
0 g
s
.30037 .58901 m .30316 .59383 L p .30051 .5894 L 0 0 0 r
F P
0 g
s
.31358 .6261 m .31359 .62605 L p .31344 .62562 L .984 .708 0 r
F P
0 g
s
.70301 .4695 m .70155 .46556 L .74079 .46578 L p 0 .197 .142 r
F P
0 g
s
.36211 .6991 m .35832 .70117 L .33456 .66562 L p .367 .403 0 r
F P
0 g
s
.35832 .70117 m .36211 .6991 L .38907 .73069 L p .404 .393 0 r
F P
0 g
s
.38907 .73069 m .38614 .73439 L .35832 .70117 L p .404 .393 0 r
F P
0 g
s
.36087 .61363 m .36347 .60871 L .37923 .63971 L p .949 .638 0 r
F P
0 g
s
.37923 .63971 m .37636 .6437 L .36087 .61363 L p .949 .638 0 r
F P
0 g
s
.66081 .51949 m .66435 .51809 L .70256 .51689 L p 0 .788 .589 r
F P
0 g
s
.45981 .7908 m .45474 .79287 L .49054 .81198 L p .695 .154 0 r
F P
0 g
s
.49054 .81198 m .49609 .81049 L .45981 .7908 L p .695 .154 0 r
F P
0 g
s
.45118 .72519 m .4474 .72641 L .42062 .70157 L p .836 .317 0 r
F P
0 g
s
.74079 .46578 m .73988 .46974 L .70301 .4695 L p 0 .197 .142 r
F P
0 g
s
.73988 .46974 m .74079 .46578 L .77797 .46731 L p 0 .158 .118 r
F P
0 g
s
.78133 .46706 m .77797 .46731 L .74079 .46578 L p 0 0 0 r
F P
0 g
s
.70385 .51225 m .70256 .51689 L .66435 .51809 L p 0 .814 .605 r
F P
0 g
s
.70256 .51689 m .70385 .51225 L .73864 .51251 L p 0 .765 .545 r
F P
0 g
s
.73864 .51251 m .7393 .51714 L .70256 .51689 L p 0 .765 .545 r
F P
0 g
s
.35308 .57856 m .35184 .57323 L .36269 .60485 L p .786 .469 0 r
F P
0 g
s
.38614 .73439 m .38907 .73069 L .41891 .75849 L p .439 .376 0 r
F P
0 g
s
.41891 .75849 m .41696 .76372 L .38614 .73439 L p .439 .376 0 r
F P
0 g
s
.4497 .78765 m .45121 .79178 L .41745 .76653 L p .128 .293 0 r
F P
0 g
s
.7393 .51714 m .73864 .51251 L .77121 .51396 L p 0 .714 .486 r
F P
0 g
s
.77121 .51396 m .77385 .51878 L .7393 .51714 L p 0 .714 .486 r
F P
0 g
s
.3247 .53183 m .31475 .53256 L .32116 .56859 L p .675 .702 0 r
F P
0 g
s
.32116 .56859 m .33079 .56669 L .3247 .53183 L p .675 .702 0 r
F P
0 g
s
.31225 .57233 m .32116 .56859 L .31475 .53256 L p .751 .755 0 r
F P
0 g
s
.3342 .5336 m .3247 .53183 L .33079 .56669 L p .584 .634 0 r
F P
0 g
s
.33079 .56669 m .33988 .56688 L .3342 .5336 L p .584 .634 0 r
F P
0 g
s
.37072 .64873 m .37636 .6437 L .39657 .67349 L p .848 .548 0 r
F P
0 g
s
.39657 .67349 m .39099 .6782 L .37072 .64873 L p .848 .548 0 r
F P
0 g
s
.39099 .6782 m .39657 .67349 L .42062 .70157 L p .824 .463 0 r
F P
0 g
s
.77797 .46731 m .78133 .46706 L .81786 .47014 L p 0 0 0 r
F P
0 g
s
.31358 .6261 m .31585 .62932 L .33456 .66562 L p 0 0 0 r
F P
0 g
s
.33916 .66528 m .33456 .66562 L .31585 .62932 L p .337 .411 0 r
F P
0 g
s
.33456 .66562 m .33916 .66528 L .36211 .6991 L p .367 .403 0 r
F P
0 g
s
.42062 .70157 m .41511 .7059 L .39099 .6782 L p .824 .463 0 r
F P
0 g
s
.41511 .7059 m .42062 .70157 L .4474 .72641 L p .796 .38 0 r
F P
0 g
s
.80544 .52164 m .8108 .52342 L .77709 .52015 L p 0 .661 .417 r
F P
0 g
s
.31475 .53256 m .30565 .53569 L .31225 .57233 L p .751 .755 0 r
F P
0 g
s
.30524 .57744 m .31225 .57233 L .30565 .53569 L p .827 .803 0 r
F P
0 g
s
.34198 .53762 m .3342 .5336 L .33988 .56688 L p .443 .523 0 r
F P
0 g
s
.33988 .56688 m .34723 .56916 L .34198 .53762 L p .443 .523 0 r
F P
0 g
s
.31585 .62932 m .31358 .6261 L .30034 .58896 L p 0 0 0 r
F P
0 g
s
.82802 .48935 m .82774 .49911 L .86287 .5047 L p 0 .423 .274 r
F P
0 g
s
.82774 .49911 m .82548 .50854 L .85975 .51408 L p 0 .361 .239 r
F P
0 g
s
.85975 .51408 m .86287 .5047 L .82774 .49911 L p 0 .361 .239 r
F P
0 g
s
.3682 .69514 m .36211 .6991 L .33916 .66528 L p .577 .486 0 r
F P
0 g
s
.36211 .6991 m .3682 .69514 L .39423 .72517 L p .582 .443 0 r
F P
0 g
s
.39423 .72517 m .38907 .73069 L .36211 .6991 L p .582 .443 0 r
F P
0 g
s
.70419 .47647 m .70301 .4695 L .73988 .46974 L p 0 .407 .291 r
F P
0 g
s
.51387 .78008 m .51143 .78946 L .54606 .79968 L p .57 .258 0 r
F P
0 g
s
.51143 .78946 m .50727 .79841 L .54313 .80903 L p .589 .23 0 r
F P
0 g
s
.54313 .80903 m .54606 .79968 L .51143 .78946 L p .589 .23 0 r
F P
0 g
s
.82628 .48052 m .82802 .48935 L .86338 .49478 L p 0 .484 .307 r
F P
0 g
s
.86287 .5047 m .86338 .49478 L .82802 .48935 L p 0 .423 .274 r
F P
0 g
s
.35521 .61896 m .36087 .61363 L .37636 .6437 L p .873 .642 0 r
F P
0 g
s
.37636 .6437 m .37072 .64873 L .35521 .61896 L p .873 .642 0 r
F P
0 g
s
.82548 .50854 m .82153 .51637 L .85442 .52167 L p 0 .281 .194 r
F P
0 g
s
.85442 .52167 m .85975 .51408 L .82548 .50854 L p 0 .281 .194 r
F P
0 g
s
.45474 .79287 m .45121 .79178 L .48604 .80998 L p .404 0 0 r
F P
0 g
s
.49054 .81198 m .45474 .79287 L p .48309 .80837 L .404 0 0 r
F P
0 g
s
.4497 .78765 m .48321 .80474 L p .45405 .79293 L .271 .326 0 r
F P
0 g
s
.45121 .79178 m .45104 .79133 L p .45405 .79293 L .271 .326 0 r
F P
0 g
s
.45104 .79133 m .4497 .78765 L p .45405 .79293 L .271 .326 0 r
F P
0 g
s
.48604 .80998 m .48656 .81021 L p .48309 .80837 L .404 0 0 r
F P
0 g
s
.48656 .81021 m .49054 .81198 L p .48309 .80837 L .404 0 0 r
F P
0 g
s
.51425 .77149 m .51387 .78008 L .54686 .78971 L p .54 .286 0 r
F P
0 g
s
.54606 .79968 m .54686 .78971 L .51387 .78008 L p .57 .258 0 r
F P
0 g
s
.66435 .51809 m .66731 .51324 L .70385 .51225 L p 0 .814 .605 r
F P
0 g
s
.70475 .50475 m .70385 .51225 L .66731 .51324 L p 0 .729 .539 r
F P
0 g
s
.70385 .51225 m .70475 .50475 L .73826 .50501 L p 0 .683 .487 r
F P
0 g
s
.73826 .50501 m .73864 .51251 L .70385 .51225 L p 0 .683 .487 r
F P
0 g
s
.73988 .46974 m .73908 .47672 L .70419 .47647 L p 0 .407 .291 r
F P
0 g
s
.73908 .47672 m .73988 .46974 L .77464 .47111 L p 0 .364 .255 r
F P
0 g
s
.77797 .46731 m .77464 .47111 L .73988 .46974 L p 0 .158 .118 r
F P
0 g
s
.38907 .73069 m .39423 .72517 L .42305 .75153 L p .583 .399 0 r
F P
0 g
s
.42305 .75153 m .41891 .75849 L .38907 .73069 L p .583 .399 0 r
F P
0 g
s
.4474 .72641 m .44189 .73032 L .41511 .7059 L p .796 .38 0 r
F P
0 g
s
.44189 .73032 m .4474 .72641 L .4756 .74653 L p .757 .293 0 r
F P
0 g
s
.4799 .7462 m .4756 .74653 L .4474 .72641 L p .758 .19 0 r
F P
0 g
s
.50727 .79841 m .50192 .80575 L .53845 .81651 L p .601 .195 0 r
F P
0 g
s
.53845 .81651 m .54313 .80903 L .50727 .79841 L p .601 .195 0 r
F P
0 g
s
.36308 .65412 m .37072 .64873 L .39099 .6782 L p .789 .543 0 r
F P
0 g
s
.39099 .6782 m .38372 .68391 L .36308 .65412 L p .789 .543 0 r
F P
0 g
s
.38372 .68391 m .39099 .6782 L .41511 .7059 L p .77 .468 0 r
F P
0 g
s
.35074 .58445 m .35308 .57856 L .36347 .60871 L p .982 .757 0 r
F P
0 g
s
.36347 .60871 m .36087 .61363 L .35074 .58445 L p .982 .757 0 r
F P
0 g
s
.73864 .51251 m .73826 .50501 L .76953 .50632 L p 0 .633 .434 r
F P
0 g
s
.76953 .50632 m .77121 .51396 L .73864 .51251 L p 0 .633 .434 r
F P
0 g
s
.41696 .76372 m .41891 .75849 L .45044 .78102 L p .468 .354 0 r
F P
0 g
s
.45044 .78102 m .4497 .78765 L .41696 .76372 L p .468 .354 0 r
F P
0 g
s
.30565 .53569 m .29861 .54082 L .30524 .57744 L p .827 .803 0 r
F P
0 g
s
.30107 .58324 m .30524 .57744 L .29861 .54082 L p .911 .842 0 r
F P
0 g
s
.41511 .7059 m .40822 .71185 L .38372 .68391 L p .77 .468 0 r
F P
0 g
s
.40822 .71185 m .41511 .7059 L .44189 .73032 L p .748 .396 0 r
F P
0 g
s
.347 .54339 m .34198 .53762 L .34723 .56916 L p .144 .273 0 r
F P
0 g
s
.34723 .56916 m .35184 .57323 L .347 .54339 L p .144 .273 0 r
F P
0 g
s
.31585 .62932 m .32123 .6308 L .33916 .66528 L p .337 .411 0 r
F P
0 g
s
.34612 .66301 m .33916 .66528 L .32123 .6308 L p .57 .531 0 r
F P
0 g
s
.33916 .66528 m .34612 .66301 L .3682 .69514 L p .577 .486 0 r
F P
0 g
s
.82274 .47381 m .82628 .48052 L .8612 .48562 L p 0 .556 .344 r
F P
0 g
s
.86338 .49478 m .8612 .48562 L .82628 .48052 L p 0 .484 .307 r
F P
0 g
s
.77464 .47111 m .77797 .46731 L .81229 .47001 L p 0 .087 .077 r
F P
0 g
s
.82153 .51637 m .81641 .52157 L .84758 .52644 L p 0 .147 .115 r
F P
0 g
s
.84758 .52644 m .85442 .52167 L .82153 .51637 L p 0 .147 .115 r
F P
0 g
s
.51252 .76484 m .51425 .77149 L .54543 .78042 L p .478 .319 0 r
F P
0 g
s
.54686 .78971 m .54543 .78042 L .51425 .77149 L p .54 .286 0 r
F P
0 g
s
.37575 .68984 m .3682 .69514 L .34612 .66301 L p .672 .517 0 r
F P
0 g
s
.3682 .69514 m .37575 .68984 L .40093 .71858 L p .665 .459 0 r
F P
0 g
s
.40093 .71858 m .39423 .72517 L .3682 .69514 L p .665 .459 0 r
F P
0 g
s
.77385 .51878 m .77121 .51396 L .80108 .51644 L p 0 .658 .427 r
F P
0 g
s
.80108 .51644 m .80544 .52164 L .77385 .51878 L p 0 .658 .427 r
F P
0 g
s
.70419 .47647 m .70493 .48552 L .66935 .48625 L p 0 .535 .393 r
F P
0 g
s
.70493 .48552 m .70419 .47647 L .73908 .47672 L p 0 .502 .359 r
F P
0 g
s
.35448 .65911 m .36308 .65412 L .38372 .68391 L p .735 .533 0 r
F P
0 g
s
.38372 .68391 m .37575 .68984 L .35448 .65911 L p .735 .533 0 r
F P
0 g
s
.37575 .68984 m .38372 .68391 L .40822 .71185 L p .721 .466 0 r
F P
0 g
s
.39423 .72517 m .40093 .71858 L .42881 .74381 L p .654 .404 0 r
F P
0 g
s
.42881 .74381 m .42305 .75153 L .39423 .72517 L p .654 .404 0 r
F P
0 g
s
.50192 .80575 m .49609 .81049 L .53264 .82112 L p .601 .136 0 r
F P
0 g
s
.53264 .82112 m .53845 .81651 L .50192 .80575 L p .601 .136 0 r
F P
0 g
s
.32123 .6308 m .31585 .62932 L .30316 .59383 L p .323 .423 0 r
F P
0 g
s
.34726 .62396 m .35521 .61896 L .37072 .64873 L p .807 .626 0 r
F P
0 g
s
.37072 .64873 m .36308 .65412 L .34726 .62396 L p .807 .626 0 r
F P
0 g
s
.66731 .51324 m .66929 .50558 L .70475 .50475 L p 0 .729 .539 r
F P
0 g
s
.70513 .49541 m .70475 .50475 L .66929 .50558 L p 0 .663 .489 r
F P
0 g
s
.70475 .50475 m .70513 .49541 L .7382 .49567 L p 0 .621 .443 r
F P
0 g
s
.7382 .49567 m .73826 .50501 L .70475 .50475 L p 0 .621 .443 r
F P
0 g
s
.34612 .66301 m .35448 .65911 L .37575 .68984 L p .672 .517 0 r
F P
0 g
s
.40822 .71185 m .40093 .71858 L .37575 .68984 L p .721 .466 0 r
F P
0 g
s
.40093 .71858 m .40822 .71185 L .4354 .73639 L p .704 .402 0 r
F P
0 g
s
.44189 .73032 m .4354 .73639 L .40822 .71185 L p .748 .396 0 r
F P
0 g
s
.73908 .47672 m .73849 .48578 L .70493 .48552 L p 0 .502 .359 r
F P
0 g
s
.73849 .48578 m .73908 .47672 L .77179 .47798 L p 0 .457 .317 r
F P
0 g
s
.77464 .47111 m .77179 .47798 L .73908 .47672 L p 0 .364 .255 r
F P
0 g
s
.73826 .50501 m .7382 .49567 L .76905 .49689 L p 0 .573 .394 r
F P
0 g
s
.76905 .49689 m .76953 .50632 L .73826 .50501 L p 0 .573 .394 r
F P
0 g
s
.4354 .73639 m .42881 .74381 L .40093 .71858 L p .704 .402 0 r
F P
0 g
s
.4756 .74653 m .47 .74998 L .44189 .73032 L p .757 .293 0 r
F P
0 g
s
.4354 .73639 m .44189 .73032 L .47 .74998 L p .718 .325 0 r
F P
0 g
s
.32123 .6308 m .32899 .63032 L .34612 .66301 L p .57 .531 0 r
F P
0 g
s
.35448 .65911 m .34612 .66301 L .32899 .63032 L p .676 .578 0 r
F P
0 g
s
.48321 .80474 m .48604 .80998 L .45121 .79178 L p .271 .326 0 r
F P
0 g
s
.82274 .47381 m .85661 .47845 L p .821 .47081 L 0 .636 .38 r
F P
0 g
s
.81834 .4705 m .82274 .47381 L p .821 .47081 L 0 .636 .38 r
F P
0 g
s
.85661 .47845 m .85022 .47425 L .81786 .47014 L p 0 .636 .38 r
F P
0 g
s
.84286 .47359 m .81229 .47001 L p .84726 .47392 L 0 0 0 r
F P
0 g
s
.85022 .47425 m .84955 .47419 L p .84726 .47392 L 0 0 0 r
F P
0 g
s
.84955 .47419 m .84286 .47359 L p .84726 .47392 L 0 0 0 r
F P
0 g
s
.81786 .47014 m .81834 .4705 L p .821 .47081 L 0 .636 .38 r
F P
0 g
s
.8612 .48562 m .85661 .47845 L .82274 .47381 L p 0 .556 .344 r
F P
0 g
s
.67001 .49614 m .66935 .48625 L .70493 .48552 L p 0 .603 .444 r
F P
0 g
s
.70493 .48552 m .70513 .49541 L .67001 .49614 L p 0 .603 .444 r
F P
0 g
s
.70513 .49541 m .70493 .48552 L .73849 .48578 L p 0 .565 .403 r
F P
0 g
s
.29861 .54082 m .29458 .54726 L .30107 .58324 L p .911 .842 0 r
F P
0 g
s
.30034 .58896 m .30107 .58324 L .29458 .54726 L p .892 .739 0 r
F P
0 g
s
.41891 .75849 m .42305 .75153 L .45336 .77278 L p .579 .355 0 r
F P
0 g
s
.45336 .77278 m .45044 .78102 L .41891 .75849 L p .579 .355 0 r
F P
0 g
s
.81641 .52157 m .8108 .52342 L .84015 .52776 L p 0 0 0 r
F P
0 g
s
.84758 .52644 m .81641 .52157 L p .83818 .52725 L 0 0 0 r
F P
0 g
s
.80544 .52164 m .83311 .52541 L p .8125 .52358 L 0 .592 .349 r
F P
0 g
s
.8108 .52342 m .81031 .52326 L p .8125 .52358 L 0 .592 .349 r
F P
0 g
s
.81031 .52326 m .80544 .52164 L p .8125 .52358 L 0 .592 .349 r
F P
0 g
s
.84015 .52776 m .84087 .52763 L p .83818 .52725 L 0 0 0 r
F P
0 g
s
.84087 .52763 m .84758 .52644 L p .83818 .52725 L 0 0 0 r
F P
0 g
s
.66929 .50558 m .67001 .49614 L .70513 .49541 L p 0 .663 .489 r
F P
0 g
s
.73849 .48578 m .7382 .49567 L .70513 .49541 L p 0 .565 .403 r
F P
0 g
s
.34857 .55015 m .347 .54339 L .34916 .5567 L p 0 0 0 r
F P
0 g
s
.347 .54339 m .34857 .55015 L .35048 .5241 L p 0 0 0 r
F P
0 g
s
.34857 .55015 m .34646 .55698 L .34777 .5411 L p .793 .839 0 r
F P
0 g
s
.35048 .5241 m .34873 .51609 L .34796 .52817 L p 0 0 0 r
F P
0 g
s
.34796 .52817 m .347 .54339 L p .35048 .5241 L 0 0 0 r
F P
0 g
s
.34853 .53182 m .35048 .5241 L .34857 .55015 L p .793 .839 0 r
F P
0 g
s
.35048 .5241 m .34853 .53182 L .35391 .51855 L p 0 0 0 r
F P
0 g
s
.35391 .51855 m .35758 .5095 L p .35048 .5241 L 0 0 0 r
F P
0 g
s
.34777 .5411 m .34853 .53182 L p .34857 .55015 L .793 .839 0 r
F P
0 g
s
.34916 .5567 m .35184 .57323 L p .34857 .55015 L 0 0 0 r
F P
0 g
s
.35184 .57323 m .35308 .57856 L .34857 .55015 L p 0 0 0 r
F P
0 g
s
.34511 .59008 m .35074 .58445 L .36087 .61363 L p .901 .75 0 r
F P
0 g
s
.36087 .61363 m .35521 .61896 L .34511 .59008 L p .901 .75 0 r
F P
0 g
s
.33809 .62793 m .34726 .62396 L .36308 .65412 L p .746 .605 0 r
F P
0 g
s
.36308 .65412 m .35448 .65911 L .33809 .62793 L p .746 .605 0 r
F P
0 g
s
.7382 .49567 m .73849 .48578 L .76984 .48698 L p 0 .519 .358 r
F P
0 g
s
.77179 .47798 m .76984 .48698 L .73849 .48578 L p 0 .457 .317 r
F P
0 g
s
.77179 .47798 m .77464 .47111 L .80678 .47348 L p 0 .298 .209 r
F P
0 g
s
.81229 .47001 m .80678 .47348 L .77464 .47111 L p 0 .087 .077 r
F P
0 g
s
.76984 .48698 m .76905 .49689 L .7382 .49567 L p 0 .519 .358 r
F P
0 g
s
.32899 .63032 m .33809 .62793 L .35448 .65911 L p .676 .578 0 r
F P
0 g
s
.47 .74998 m .4756 .74653 L .50384 .7606 L p .692 .194 0 r
F P
0 g
s
.5089 .76103 m .50384 .7606 L .4756 .74653 L p .602 .026 0 r
F P
0 g
s
.77121 .51396 m .76953 .50632 L .7983 .50851 L p 0 .579 .38 r
F P
0 g
s
.7983 .50851 m .80108 .51644 L .77121 .51396 L p 0 .579 .38 r
F P
0 g
s
.49609 .81049 m .49054 .81198 L .52646 .82221 L p .522 0 0 r
F P
0 g
s
.53264 .82112 m .49609 .81049 L p .5244 .82141 L .522 0 0 r
F P
0 g
s
.48604 .80998 m .52075 .81959 L p .49249 .81247 L 0 .291 0 r
F P
0 g
s
.49054 .81198 m .49021 .81183 L p .49249 .81247 L 0 .291 0 r
F P
0 g
s
.49021 .81183 m .48604 .80998 L p .49249 .81247 L 0 .291 0 r
F P
0 g
s
.52646 .82221 m .52693 .82212 L p .5244 .82141 L .522 0 0 r
F P
0 g
s
.52693 .82212 m .53264 .82112 L p .5244 .82141 L .522 0 0 r
F P
0 g
s
.42881 .74381 m .4354 .73639 L .46387 .7561 L p .68 .34 0 r
F P
0 g
s
.47 .74998 m .46387 .7561 L .4354 .73639 L p .718 .325 0 r
F P
0 g
s
.30316 .59383 m .3092 .59716 L .32123 .6308 L p .323 .423 0 r
F P
0 g
s
.32899 .63032 m .32123 .6308 L .3092 .59716 L p .561 .576 0 r
F P
0 g
s
.42305 .75153 m .42881 .74381 L .45805 .76407 L p .638 .349 0 r
F P
0 g
s
.45805 .76407 m .45336 .77278 L .42305 .75153 L p .638 .349 0 r
F P
0 g
s
.46387 .7561 m .45805 .76407 L .42881 .74381 L p .68 .34 0 r
F P
0 g
s
.81229 .47001 m .81786 .47014 L .85022 .47425 L p 0 0 0 r
F P
0 g
s
.83311 .52541 m .84015 .52776 L .8108 .52342 L p 0 .592 .349 r
F P
0 g
s
.33694 .5947 m .34511 .59008 L .35521 .61896 L p .825 .72 0 r
F P
0 g
s
.35521 .61896 m .34726 .62396 L .33694 .5947 L p .825 .72 0 r
F P
0 g
s
.4497 .78765 m .45044 .78102 L .48245 .79694 L p .489 .327 0 r
F P
0 g
s
.48245 .79694 m .48321 .80474 L .4497 .78765 L p .489 .327 0 r
F P
0 g
s
.76984 .48698 m .77179 .47798 L .80207 .48011 L p 0 .396 .269 r
F P
0 g
s
.80678 .47348 m .80207 .48011 L .77179 .47798 L p 0 .298 .209 r
F P
0 g
s
.29458 .54726 m .29414 .55418 L .30034 .58896 L p .892 .739 0 r
F P
0 g
s
.30316 .59383 m .30034 .58896 L .29414 .55418 L p 0 0 0 r
F P
0 g
s
.76953 .50632 m .76905 .49689 L .79752 .49892 L p 0 .517 .343 r
F P
0 g
s
.79752 .49892 m .7983 .50851 L .76953 .50632 L p 0 .517 .343 r
F P
0 g
s
.34646 .55698 m .34857 .55015 L .35308 .57856 L p .939 .835 0 r
F P
0 g
s
.35308 .57856 m .35074 .58445 L .34646 .55698 L p .939 .835 0 r
F P
0 g
s
.50384 .7606 m .49803 .76363 L .47 .74998 L p .692 .194 0 r
F P
0 g
s
.46387 .7561 m .47 .74998 L .49803 .76363 L p .672 .249 0 r
F P
0 g
s
.3092 .59716 m .31765 .5985 L .32899 .63032 L p .561 .576 0 r
F P
0 g
s
.33809 .62793 m .32899 .63032 L .31765 .5985 L p .677 .644 0 r
F P
0 g
s
.76905 .49689 m .76984 .48698 L .79884 .48899 L p 0 .46 .309 r
F P
0 g
s
.80207 .48011 m .79884 .48899 L .76984 .48698 L p 0 .396 .269 r
F P
0 g
s
.3356 .50337 m .32599 .50033 L .3247 .53183 L p .535 .653 0 r
F P
0 g
s
.3247 .53183 m .3342 .5336 L .3356 .50337 L p .535 .653 0 r
F P
0 g
s
.31475 .53256 m .3247 .53183 L .32599 .50033 L p .624 .741 0 r
F P
0 g
s
.79884 .48899 m .79752 .49892 L .76905 .49689 L p 0 .46 .309 r
F P
0 g
s
.32734 .59764 m .33694 .5947 L .34726 .62396 L p .755 .686 0 r
F P
0 g
s
.34726 .62396 m .33809 .62793 L .32734 .59764 L p .755 .686 0 r
F P
0 g
s
.52075 .81959 m .52646 .82221 L .49054 .81198 L p 0 .291 0 r
F P
0 g
s
.32599 .50033 m .31597 .50015 L .31475 .53256 L p .624 .741 0 r
F P
0 g
s
.30565 .53569 m .31475 .53256 L .31597 .50015 L p .705 .818 0 r
F P
0 g
s
.31765 .5985 m .32734 .59764 L .33809 .62793 L p .677 .644 0 r
F P
0 g
s
.34353 .50886 m .3356 .50337 L .3342 .5336 L p .413 .529 0 r
F P
0 g
s
.3342 .5336 m .34198 .53762 L .34353 .50886 L p .413 .529 0 r
F P
0 g
s
.54543 .78042 m .54686 .78971 L .57822 .78954 L p .53 .251 0 r
F P
0 g
s
.54686 .78971 m .54606 .79968 L .57918 .7996 L p .53 .212 0 r
F P
0 g
s
.57918 .7996 m .57822 .78954 L .54686 .78971 L p .53 .212 0 r
F P
0 g
s
.80678 .47348 m .81229 .47001 L .84286 .47359 L p 0 0 .027 r
F P
0 g
s
.45044 .78102 m .45336 .77278 L .48388 .78764 L p .568 .312 0 r
F P
0 g
s
.48388 .78764 m .48245 .79694 L .45044 .78102 L p .568 .312 0 r
F P
0 g
s
.54606 .79968 m .54313 .80903 L .57758 .809 L p .521 .169 0 r
F P
0 g
s
.57758 .809 m .57918 .7996 L .54606 .79968 L p .521 .169 0 r
F P
0 g
s
.80544 .52164 m .80108 .51644 L .82743 .51969 L p 0 .59 .366 r
F P
0 g
s
.82743 .51969 m .83311 .52541 L .80544 .52164 L p 0 .59 .366 r
F P
0 g
s
.45805 .76407 m .46387 .7561 L .49225 .76975 L p .645 .276 0 r
F P
0 g
s
.49803 .76363 m .49225 .76975 L .46387 .7561 L p .672 .249 0 r
F P
0 g
s
.31597 .50015 m .30687 .50287 L .30565 .53569 L p .705 .818 0 r
F P
0 g
s
.29861 .54082 m .30565 .53569 L .30687 .50287 L p .789 .891 0 r
F P
0 g
s
.57822 .78954 m .57482 .78014 L .54543 .78042 L p .53 .251 0 r
F P
0 g
s
.29414 .55418 m .29738 .56063 L .30316 .59383 L p 0 0 0 r
F P
0 g
s
.3092 .59716 m .30316 .59383 L .29738 .56063 L p .309 .419 0 r
F P
0 g
s
.3409 .56297 m .34646 .55698 L .35074 .58445 L p .874 .833 0 r
F P
0 g
s
.35074 .58445 m .34511 .59008 L .3409 .56297 L p .874 .833 0 r
F P
0 g
s
.49803 .76363 m .50384 .7606 L .53087 .76763 L p .567 .066 0 r
F P
0 g
s
.34873 .51609 m .34353 .50886 L .34198 .53762 L p .202 .306 0 r
F P
0 g
s
.34198 .53762 m .347 .54339 L .34873 .51609 L p .202 .306 0 r
F P
0 g
s
.54313 .80903 m .53845 .81651 L .57362 .81647 L p .497 .111 0 r
F P
0 g
s
.57362 .81647 m .57758 .809 L .54313 .80903 L p .497 .111 0 r
F P
0 g
s
.45336 .77278 m .45805 .76407 L .48731 .77814 L p .612 .295 0 r
F P
0 g
s
.48731 .77814 m .48388 .78764 L .45336 .77278 L p .612 .295 0 r
F P
0 g
s
.49225 .76975 m .48731 .77814 L .45805 .76407 L p .645 .276 0 r
F P
0 g
s
.48604 .80998 m .48321 .80474 L .51629 .8136 L p .39 .332 0 r
F P
0 g
s
.51629 .8136 m .52075 .81959 L .48604 .80998 L p .39 .332 0 r
F P
0 g
s
.80207 .48011 m .80678 .47348 L .83554 .4766 L p 0 .205 .154 r
F P
0 g
s
.84286 .47359 m .83554 .4766 L .80678 .47348 L p 0 0 .027 r
F P
0 g
s
.80108 .51644 m .7983 .50851 L .8239 .51135 L p 0 .512 .326 r
F P
0 g
s
.8239 .51135 m .82743 .51969 L .80108 .51644 L p 0 .512 .326 r
F P
0 g
s
.30687 .50287 m .2999 .50812 L .29861 .54082 L p .789 .891 0 r
F P
0 g
s
.29458 .54726 m .29861 .54082 L .2999 .50812 L p .855 .928 0 r
F P
0 g
s
.86338 .49478 m .86287 .5047 L .89153 .51115 L p 0 .328 .218 r
F P
0 g
s
.86287 .5047 m .85975 .51408 L .88758 .52046 L p 0 .257 .182 r
F P
0 g
s
.88758 .52046 m .89153 .51115 L .86287 .5047 L p 0 .257 .182 r
F P
0 g
s
.29738 .56063 m .30388 .56573 L .3092 .59716 L p .309 .419 0 r
F P
0 g
s
.31765 .5985 m .3092 .59716 L .30388 .56573 L p .518 .587 0 r
F P
0 g
s
.8612 .48562 m .86338 .49478 L .8924 .50106 L p 0 .397 .251 r
F P
0 g
s
.89153 .51115 m .8924 .50106 L .86338 .49478 L p 0 .328 .218 r
F P
0 g
s
.53845 .81651 m .53264 .82112 L .5678 .82101 L p .424 .003 0 r
F P
0 g
s
.5678 .82101 m .57362 .81647 L .53845 .81651 L p .424 .003 0 r
F P
0 g
s
.85975 .51408 m .85442 .52167 L .88104 .52773 L p 0 .168 .136 r
F P
0 g
s
.88104 .52773 m .88758 .52046 L .85975 .51408 L p 0 .168 .136 r
F P
0 g
s
.33265 .56727 m .3409 .56297 L .34511 .59008 L p .791 .785 0 r
F P
0 g
s
.34511 .59008 m .33694 .5947 L .33265 .56727 L p .791 .785 0 r
F P
0 g
s
.53087 .76763 m .52474 .77037 L .49803 .76363 L p .567 .066 0 r
F P
0 g
s
.49225 .76975 m .49803 .76363 L .52474 .77037 L p .596 .162 0 r
F P
0 g
s
.79884 .48899 m .80207 .48011 L .82926 .48288 L p 0 .312 .215 r
F P
0 g
s
.83554 .4766 m .82926 .48288 L .80207 .48011 L p 0 .205 .154 r
F P
0 g
s
.85661 .47845 m .8612 .48562 L .89004 .49152 L p 0 .475 .286 r
F P
0 g
s
.8924 .50106 m .89004 .49152 L .8612 .48562 L p 0 .397 .251 r
F P
0 g
s
.85442 .52167 m .84758 .52644 L .87281 .53199 L p 0 .029 .059 r
F P
0 g
s
.87281 .53199 m .88104 .52773 L .85442 .52167 L p 0 .029 .059 r
F P
0 g
s
.7983 .50851 m .79752 .49892 L .823 .50154 L p 0 .446 .29 r
F P
0 g
s
.823 .50154 m .8239 .51135 L .7983 .50851 L p 0 .446 .29 r
F P
0 g
s
.30388 .56573 m .31276 .56876 L .31765 .5985 L p .518 .587 0 r
F P
0 g
s
.32734 .59764 m .31765 .5985 L .31276 .56876 L p .631 .673 0 r
F P
0 g
s
.32282 .5693 m .33265 .56727 L .33694 .5947 L p .715 .734 0 r
F P
0 g
s
.33694 .5947 m .32734 .59764 L .32282 .5693 L p .715 .734 0 r
F P
0 g
s
.48321 .80474 m .48245 .79694 L .51371 .80504 L p .504 .299 0 r
F P
0 g
s
.51371 .80504 m .51629 .8136 L .48321 .80474 L p .504 .299 0 r
F P
0 g
s
.2999 .50812 m .29603 .51524 L .29458 .54726 L p .855 .928 0 r
F P
0 g
s
.29414 .55418 m .29458 .54726 L .29603 .51524 L p .64 .626 0 r
F P
0 g
s
.79752 .49892 m .79884 .48899 L .82488 .49159 L p 0 .384 .256 r
F P
0 g
s
.82926 .48288 m .82488 .49159 L .79884 .48899 L p 0 .312 .215 r
F P
0 g
s
.82488 .49159 m .823 .50154 L .79752 .49892 L p 0 .384 .256 r
F P
0 g
s
.31276 .56876 m .32282 .5693 L .32734 .59764 L p .631 .673 0 r
F P
0 g
s
.53264 .82112 m .52646 .82221 L .5609 .82197 L p .132 0 0 r
F P
0 g
s
.5678 .82101 m .53264 .82112 L p .5598 .82194 L .132 0 0 r
F P
0 g
s
.52075 .81959 m .55384 .81919 L p .52746 .8221 L .342 .371 0 r
F P
0 g
s
.52646 .82221 m .52623 .8221 L p .52746 .8221 L .342 .371 0 r
F P
0 g
s
.52623 .8221 m .52075 .81959 L p .52746 .8221 L .342 .371 0 r
F P
0 g
s
.5609 .82197 m .56118 .82193 L p .5598 .82194 L .132 0 0 r
F P
0 g
s
.56118 .82193 m .5678 .82101 L p .5598 .82194 L .132 0 0 r
F P
0 g
s
.85661 .47845 m .88478 .48381 L p .85377 .47523 L 0 .556 .311 r
F P
0 g
s
.85105 .4748 m .85661 .47845 L p .85377 .47523 L 0 .556 .311 r
F P
0 g
s
.88478 .48381 m .8773 .47897 L .85022 .47425 L p 0 .556 .311 r
F P
0 g
s
.86861 .47768 m .84286 .47359 L p .87385 .47843 L 0 0 0 r
F P
0 g
s
.8773 .47897 m .87624 .47881 L p .87385 .47843 L 0 0 0 r
F P
0 g
s
.87624 .47881 m .86861 .47768 L p .87385 .47843 L 0 0 0 r
F P
0 g
s
.85022 .47425 m .85105 .4748 L p .85377 .47523 L 0 .556 .311 r
F P
0 g
s
.89004 .49152 m .88478 .48381 L .85661 .47845 L p 0 .475 .286 r
F P
0 g
s
.84758 .52644 m .84015 .52776 L .86397 .53265 L p 0 0 0 r
F P
0 g
s
.86397 .53265 m .87281 .53199 L .84758 .52644 L p 0 0 0 r
F P
0 g
s
.48731 .77814 m .49225 .76975 L .51931 .77646 L p .59 .209 0 r
F P
0 g
s
.52474 .77037 m .51931 .77646 L .49225 .76975 L p .596 .162 0 r
F P
0 g
s
.48245 .79694 m .48388 .78764 L .51338 .79507 L p .55 .269 0 r
F P
0 g
s
.51338 .79507 m .51371 .80504 L .48245 .79694 L p .55 .269 0 r
F P
0 g
s
.48388 .78764 m .48731 .77814 L .51533 .78509 L p .574 .241 0 r
F P
0 g
s
.51931 .77646 m .51533 .78509 L .48731 .77814 L p .59 .209 0 r
F P
0 g
s
.51533 .78509 m .51338 .79507 L .48388 .78764 L p .574 .241 0 r
F P
0 g
s
.29603 .51524 m .29579 .52326 L .29414 .55418 L p .64 .626 0 r
F P
0 g
s
.29738 .56063 m .29414 .55418 L .29579 .52326 L p 0 .069 0 r
F P
0 g
s
.84286 .47359 m .85022 .47425 L .8773 .47897 L p 0 0 0 r
F P
0 g
s
.84015 .52776 m .83311 .52541 L .85571 .52962 L p 0 .51 .279 r
F P
0 g
s
.85571 .52962 m .86397 .53265 L .84015 .52776 L p 0 .51 .279 r
F P
0 g
s
.34311 .53821 m .34853 .53182 L .34646 .55698 L p .831 .932 0 r
F P
0 g
s
.34646 .55698 m .3409 .56297 L .34311 .53821 L p .831 .932 0 r
F P
0 g
s
.55384 .81919 m .5609 .82197 L .52646 .82221 L p .342 .371 0 r
F P
0 g
s
.57482 .78014 m .57822 .78954 L .6063 .77868 L p .518 .218 0 r
F P
0 g
s
.57822 .78954 m .57918 .7996 L .60894 .78822 L p .479 .169 0 r
F P
0 g
s
.60894 .78822 m .6063 .77868 L .57822 .78954 L p .479 .169 0 r
F P
0 g
s
.56944 .77264 m .57482 .78014 L .60104 .76987 L p .551 .266 0 r
F P
0 g
s
.6063 .77868 m .60104 .76987 L .57482 .78014 L p .518 .218 0 r
F P
0 g
s
.83554 .4766 m .84286 .47359 L .86861 .47768 L p 0 0 0 r
F P
0 g
s
.83311 .52541 m .82743 .51969 L .84915 .52328 L p 0 .501 .301 r
F P
0 g
s
.84915 .52328 m .85571 .52962 L .83311 .52541 L p 0 .501 .301 r
F P
0 g
s
.57918 .7996 m .57758 .809 L .60862 .79722 L p .421 .108 0 r
F P
0 g
s
.60862 .79722 m .60894 .78822 L .57918 .7996 L p .421 .108 0 r
F P
0 g
s
.29579 .52326 m .29924 .5311 L .29738 .56063 L p 0 .069 0 r
F P
0 g
s
.30388 .56573 m .29738 .56063 L .29924 .5311 L p .308 .428 0 r
F P
0 g
s
.51931 .77646 m .52474 .77037 L .54921 .76961 L p .456 .054 0 r
F P
0 g
s
.56278 .76804 m .56944 .77264 L .59388 .76296 L p .576 .323 0 r
F P
0 g
s
.60104 .76987 m .59388 .76296 L .56944 .77264 L p .551 .266 0 r
F P
0 g
s
.33494 .54238 m .34311 .53821 L .3409 .56297 L p .749 .866 0 r
F P
0 g
s
.3409 .56297 m .33265 .56727 L .33494 .54238 L p .749 .866 0 r
F P
0 g
s
.34454 .47708 m .33508 .47321 L .32599 .50033 L p .482 .671 0 r
F P
0 g
s
.32599 .50033 m .3356 .50337 L .34454 .47708 L p .482 .671 0 r
F P
0 g
s
.31597 .50015 m .32599 .50033 L .33508 .47321 L p .564 .782 0 r
F P
0 g
s
.35238 .48362 m .34454 .47708 L .3356 .50337 L p .387 .539 0 r
F P
0 g
s
.3356 .50337 m .34353 .50886 L .35238 .48362 L p .387 .539 0 r
F P
0 g
s
.52075 .81959 m .51629 .8136 L .5476 .81303 L p .485 .32 0 r
F P
0 g
s
.5476 .81303 m .55384 .81919 L .52075 .81959 L p .485 .32 0 r
F P
0 g
s
.33508 .47321 m .32524 .47252 L .31597 .50015 L p .564 .782 0 r
F P
0 g
s
.30687 .50287 m .31597 .50015 L .32524 .47252 L p .643 .89 0 r
F P
0 g
s
.57758 .809 m .57362 .81647 L .60534 .80449 L p .312 .01 0 r
F P
0 g
s
.60534 .80449 m .60862 .79722 L .57758 .809 L p .312 .01 0 r
F P
0 g
s
.35756 .49195 m .35238 .48362 L .34353 .50886 L p .252 .35 0 r
F P
0 g
s
.34353 .50886 m .34873 .51609 L .35756 .49195 L p .252 .35 0 r
F P
0 g
s
.82926 .48288 m .83554 .4766 L .85987 .48011 L p 0 .074 .093 r
F P
0 g
s
.86861 .47768 m .85987 .48011 L .83554 .4766 L p 0 0 0 r
F P
0 g
s
.82743 .51969 m .8239 .51135 L .8452 .51448 L p 0 .42 .268 r
F P
0 g
s
.8452 .51448 m .84915 .52328 L .82743 .51969 L p 0 .42 .268 r
F P
0 g
s
.29924 .5311 m .30595 .53769 L .30388 .56573 L p .308 .428 0 r
F P
0 g
s
.31276 .56876 m .30388 .56573 L .30595 .53769 L p .473 .598 0 r
F P
0 g
s
.32511 .54374 m .33494 .54238 L .33265 .56727 L p .665 .788 0 r
F P
0 g
s
.33265 .56727 m .32282 .5693 L .32511 .54374 L p .665 .788 0 r
F P
0 g
s
.51533 .78509 m .51931 .77646 L .54412 .77562 L p .503 .137 0 r
F P
0 g
s
.54921 .76961 m .54412 .77562 L .51931 .77646 L p .456 .054 0 r
F P
0 g
s
.30595 .53769 m .31498 .54211 L .31276 .56876 L p .473 .598 0 r
F P
0 g
s
.32282 .5693 m .31276 .56876 L .31498 .54211 L p .579 .704 0 r
F P
0 g
s
.31498 .54211 m .32511 .54374 L .32282 .5693 L p .579 .704 0 r
F P
0 g
s
.32524 .47252 m .31635 .47511 L .30687 .50287 L p .643 .89 0 r
F P
0 g
s
.2999 .50812 m .30687 .50287 L .31635 .47511 L p .713 .983 0 r
F P
0 g
s
.82488 .49159 m .82926 .48288 L .85228 .48597 L p 0 .194 .156 r
F P
0 g
s
.85987 .48011 m .85228 .48597 L .82926 .48288 L p 0 .074 .093 r
F P
0 g
s
.8239 .51135 m .823 .50154 L .8444 .50441 L p 0 .347 .233 r
F P
0 g
s
.8444 .50441 m .8452 .51448 L .8239 .51135 L p 0 .347 .233 r
F P
0 g
s
.35938 .50098 m .35756 .49195 L .34873 .51609 L p .016 .02 0 r
F P
0 g
s
.34873 .51609 m .35048 .5241 L .35938 .50098 L p .016 .02 0 r
F P
0 g
s
.51629 .8136 m .51371 .80504 L .54304 .80432 L p .517 .27 0 r
F P
0 g
s
.54304 .80432 m .5476 .81303 L .51629 .8136 L p .517 .27 0 r
F P
0 g
s
.47974 .507 m .48129 .51722 L .51937 .5102 L p 0 .723 .571 r
F P
0 g
s
.51937 .5102 m .51848 .50012 L .47974 .507 L p 0 .723 .571 r
F P
0 g
s
.48133 .49697 m .47974 .507 L .51848 .50012 L p 0 .816 .663 r
F P
0 g
s
.51848 .50012 m .52039 .49054 L .48133 .49697 L p 0 .816 .663 r
F P
0 g
s
.57362 .81647 m .5678 .82101 L .59953 .80903 L p .039 0 0 r
F P
0 g
s
.60534 .80449 m .57362 .81647 L p .59932 .80909 L .039 0 0 r
F P
0 g
s
.5609 .82197 m .59195 .8102 L p .56797 .82093 L .429 .395 0 r
F P
0 g
s
.5678 .82101 m .56775 .82101 L p .56797 .82093 L .429 .395 0 r
F P
0 g
s
.56775 .82101 m .5609 .82197 L p .56797 .82093 L .429 .395 0 r
F P
0 g
s
.59953 .80903 m .59957 .809 L p .59932 .80909 L .039 0 0 r
F P
0 g
s
.59957 .809 m .60534 .80449 L p .59932 .80909 L .039 0 0 r
F P
0 g
s
.823 .50154 m .82488 .49159 L .84688 .49446 L p 0 .276 .198 r
F P
0 g
s
.85228 .48597 m .84688 .49446 L .82488 .49159 L p 0 .194 .156 r
F P
0 g
s
.84688 .49446 m .8444 .50441 L .823 .50154 L p 0 .276 .198 r
F P
0 g
s
.51338 .79507 m .51533 .78509 L .54118 .78423 L p .519 .187 0 r
F P
0 g
s
.54412 .77562 m .54118 .78423 L .51533 .78509 L p .503 .137 0 r
F P
0 g
s
.48584 .48845 m .48133 .49697 L .52039 .49054 L p 0 .903 .755 r
F P
0 g
s
.52039 .49054 m .52485 .48272 L .48584 .48845 L p 0 .903 .755 r
F P
0 g
s
.51371 .80504 m .51338 .79507 L .5408 .79425 L p .524 .228 0 r
F P
0 g
s
.5408 .79425 m .54304 .80432 L .51371 .80504 L p .524 .228 0 r
F P
0 g
s
.54118 .78423 m .5408 .79425 L .51338 .79507 L p .519 .187 0 r
F P
0 g
s
.31635 .47511 m .30959 .48067 L .2999 .50812 L p .713 .983 0 r
F P
0 g
s
.29603 .51524 m .2999 .50812 L .30959 .48067 L p .652 .889 0 r
F P
0 g
s
.35575 .51041 m .35048 .5241 L p .35758 .5095 L 0 0 0 r
F P
0 g
s
.35238 .51637 m .35758 .5095 L .34853 .53182 L p .695 .979 0 r
F P
0 g
s
.35758 .5095 m .35238 .51637 L .35966 .50848 L p 0 0 0 r
F P
0 g
s
.35966 .50848 m .3693 .49801 L p .35758 .5095 L 0 0 0 r
F P
0 g
s
.35758 .5095 m .35938 .50098 L .35575 .51041 L p 0 0 0 r
F P
0 g
s
.4927 .48258 m .48584 .48845 L .52485 .48272 L p 0 .957 .832 r
F P
0 g
s
.52485 .48272 m .53129 .47772 L .4927 .48258 L p 0 .957 .832 r
F P
0 g
s
.59195 .8102 m .59953 .80903 L .5678 .82101 L p .429 .395 0 r
F P
0 g
s
.8924 .50106 m .89153 .51115 L .91269 .51774 L p 0 .183 .158 r
F P
0 g
s
.89153 .51115 m .88758 .52046 L .90793 .52695 L p 0 .101 .121 r
F P
0 g
s
.90793 .52695 m .91269 .51774 L .89153 .51115 L p 0 .101 .121 r
F P
0 g
s
.88758 .52046 m .88104 .52773 L .90038 .53386 L p 0 .005 .074 r
F P
0 g
s
.90038 .53386 m .90793 .52695 L .88758 .52046 L p 0 .005 .074 r
F P
0 g
s
.30959 .48067 m .3059 .48845 L .29603 .51524 L p .652 .889 0 r
F P
0 g
s
.29579 .52326 m .29603 .51524 L .3059 .48845 L p 0 0 0 r
F P
0 g
s
.89004 .49152 m .8924 .50106 L .914 .50748 L p 0 .265 .191 r
F P
0 g
s
.91269 .51774 m .914 .50748 L .8924 .50106 L p 0 .183 .158 r
F P
0 g
s
.34853 .53182 m .34311 .53821 L .35238 .51637 L p .695 .979 0 r
F P
0 g
s
.50099 .48018 m .4927 .48258 L .53129 .47772 L p 0 .742 .703 r
F P
0 g
s
.53129 .47772 m .53885 .47623 L .50099 .48018 L p 0 .742 .703 r
F P
0 g
s
.88104 .52773 m .87281 .53199 L .89104 .53754 L p 0 0 .004 r
F P
0 g
s
.89104 .53754 m .90038 .53386 L .88104 .52773 L p 0 0 .004 r
F P
0 g
s
.88478 .48381 m .89004 .49152 L .91171 .49754 L p 0 .357 .222 r
F P
0 g
s
.914 .50748 m .91171 .49754 L .89004 .49152 L p 0 .265 .191 r
F P
0 g
s
.59388 .76296 m .60104 .76987 L .62273 .74925 L p .579 .234 0 r
F P
0 g
s
.60104 .76987 m .6063 .77868 L .62949 .75685 L p .51 .188 0 r
F P
0 g
s
.62949 .75685 m .62273 .74925 L .60104 .76987 L p .51 .188 0 r
F P
0 g
s
.5609 .82197 m .55384 .81919 L .58363 .8078 L p .567 .359 0 r
F P
0 g
s
.58363 .8078 m .59195 .8102 L .5609 .82197 L p .567 .359 0 r
F P
0 g
s
.58574 .75885 m .59388 .76296 L .61414 .74353 L p .645 .281 0 r
F P
0 g
s
.62273 .74925 m .61414 .74353 L .59388 .76296 L p .579 .234 0 r
F P
0 g
s
.87281 .53199 m .86397 .53265 L .88116 .53751 L p 0 0 0 r
F P
0 g
s
.88116 .53751 m .89104 .53754 L .87281 .53199 L p 0 0 0 r
F P
0 g
s
.6063 .77868 m .60894 .78822 L .63353 .76534 L p .423 .134 0 r
F P
0 g
s
.63353 .76534 m .62949 .75685 L .6063 .77868 L p .423 .134 0 r
F P
0 g
s
.88478 .48381 m .90609 .48925 L p .88058 .48014 L 0 .45 .238 r
F P
0 g
s
.87842 .47969 m .88478 .48381 L p .88058 .48014 L 0 .45 .238 r
F P
0 g
s
.90609 .48925 m .89792 .48373 L .8773 .47897 L p 0 .45 .238 r
F P
0 g
s
.88827 .48175 m .86861 .47768 L p .89462 .48305 L 0 0 0 r
F P
0 g
s
.89792 .48373 m .89656 .48345 L p .89462 .48305 L 0 0 0 r
F P
0 g
s
.89656 .48345 m .88827 .48175 L p .89462 .48305 L 0 0 0 r
F P
0 g
s
.8773 .47897 m .87842 .47969 L p .88058 .48014 L 0 .45 .238 r
F P
0 g
s
.91171 .49754 m .90609 .48925 L .88478 .48381 L p 0 .357 .222 r
F P
0 g
s
.3059 .48845 m .3058 .49743 L .29579 .52326 L p 0 0 0 r
F P
0 g
s
.29924 .5311 m .29579 .52326 L .3058 .49743 L p .12 .184 0 r
F P
0 g
s
.34446 .52063 m .35238 .51637 L .34311 .53821 L p .678 .955 0 r
F P
0 g
s
.34311 .53821 m .33494 .54238 L .34446 .52063 L p .678 .955 0 r
F P
0 g
s
.54118 .78423 m .54412 .77562 L .56611 .76674 L p .36 .057 0 r
F P
0 g
s
.57085 .76085 m .56611 .76674 L .54412 .77562 L p .173 0 0 r
F P
0 g
s
.5096 .48158 m .50099 .48018 L .53885 .47623 L p 0 0 0 r
F P
0 g
s
.53885 .47623 m .54652 .47847 L .5096 .48158 L p 0 0 0 r
F P
0 g
s
.61414 .74353 m .60486 .74043 L .58574 .75885 L p .645 .281 0 r
F P
0 g
s
.86397 .53265 m .85571 .52962 L .87208 .53374 L p 0 .411 .208 r
F P
0 g
s
.87208 .53374 m .88116 .53751 L .86397 .53265 L p 0 .411 .208 r
F P
0 g
s
.60894 .78822 m .60862 .79722 L .63428 .77358 L p .285 .053 0 r
F P
0 g
s
.63428 .77358 m .63353 .76534 L .60894 .78822 L p .285 .053 0 r
F P
0 g
s
.86861 .47768 m .8773 .47897 L .89792 .48373 L p 0 0 0 r
F P
0 g
s
.55384 .81919 m .5476 .81303 L .57571 .80215 L p .558 .293 0 r
F P
0 g
s
.57571 .80215 m .58363 .8078 L .55384 .81919 L p .558 .293 0 r
F P
0 g
s
.3058 .49743 m .30933 .50637 L .29924 .5311 L p .12 .184 0 r
F P
0 g
s
.30595 .53769 m .29924 .5311 L .30933 .50637 L p .311 .447 0 r
F P
0 g
s
.33487 .5217 m .34446 .52063 L .33494 .54238 L p .6 .846 0 r
F P
0 g
s
.33494 .54238 m .32511 .54374 L .33487 .5217 L p .6 .846 0 r
F P
0 g
s
.5408 .79425 m .54118 .78423 L .56414 .775 L p .443 .137 0 r
F P
0 g
s
.56611 .76674 m .56414 .775 L .54118 .78423 L p .36 .057 0 r
F P
0 g
s
.85571 .52962 m .84915 .52328 L .86503 .52674 L p 0 .371 .229 r
F P
0 g
s
.86503 .52674 m .87208 .53374 L .85571 .52962 L p 0 .371 .229 r
F P
0 g
s
.85987 .48011 m .86861 .47768 L .88827 .48175 L p 0 0 0 r
F P
0 g
s
.51737 .48663 m .5096 .48158 L .54652 .47847 L p 0 .328 .195 r
F P
0 g
s
.54652 .47847 m .55326 .48418 L .51737 .48663 L p 0 .328 .195 r
F P
0 g
s
.60862 .79722 m .60534 .80449 L .63162 .78047 L p 0 0 0 r
F P
0 g
s
.63162 .78047 m .63428 .77358 L .60862 .79722 L p 0 0 0 r
F P
0 g
s
.5476 .81303 m .54304 .80432 L .56928 .79402 L p .529 .24 0 r
F P
0 g
s
.56928 .79402 m .57571 .80215 L .5476 .81303 L p .529 .24 0 r
F P
0 g
s
.30933 .50637 m .31601 .51405 L .30595 .53769 L p .311 .447 0 r
F P
0 g
s
.31498 .54211 m .30595 .53769 L .31601 .51405 L p .428 .609 0 r
F P
0 g
s
.32493 .5194 m .33487 .5217 L .32511 .54374 L p .518 .733 0 r
F P
0 g
s
.32511 .54374 m .31498 .54211 L .32493 .5194 L p .518 .733 0 r
F P
0 g
s
.54304 .80432 m .5408 .79425 L .56525 .78453 L p .492 .192 0 r
F P
0 g
s
.56414 .775 m .56525 .78453 L .5408 .79425 L p .443 .137 0 r
F P
0 g
s
.31601 .51405 m .32493 .5194 L .31498 .54211 L p .428 .609 0 r
F P
0 g
s
.84915 .52328 m .8452 .51448 L .86096 .51746 L p 0 .281 .204 r
F P
0 g
s
.86096 .51746 m .86503 .52674 L .84915 .52328 L p 0 .281 .204 r
F P
0 g
s
.56525 .78453 m .56928 .79402 L .54304 .80432 L p .492 .192 0 r
F P
0 g
s
.85228 .48597 m .85987 .48011 L .87847 .48357 L p 0 0 .029 r
F P
0 g
s
.88827 .48175 m .87847 .48357 L .85987 .48011 L p 0 0 0 r
F P
0 g
s
.36887 .46249 m .36128 .45539 L .34454 .47708 L p .363 .554 0 r
F P
0 g
s
.34454 .47708 m .35238 .48362 L .36887 .46249 L p .363 .554 0 r
F P
0 g
s
.33508 .47321 m .34454 .47708 L .36128 .45539 L p .428 .688 0 r
F P
0 g
s
.37393 .47154 m .36887 .46249 L .35238 .48362 L p .285 .398 0 r
F P
0 g
s
.35238 .48362 m .35756 .49195 L .37393 .47154 L p .285 .398 0 r
F P
0 g
s
.8452 .51448 m .8444 .50441 L .86043 .50714 L p 0 .194 .171 r
F P
0 g
s
.86043 .50714 m .86096 .51746 L .8452 .51448 L p 0 .194 .171 r
F P
0 g
s
.36128 .45539 m .35217 .45118 L .33508 .47321 L p .428 .688 0 r
F P
0 g
s
.32524 .47252 m .33508 .47321 L .35217 .45118 L p .49 .822 0 r
F P
0 g
s
.84688 .49446 m .85228 .48597 L .86983 .48896 L p 0 .015 .091 r
F P
0 g
s
.87847 .48357 m .86983 .48896 L .85228 .48597 L p 0 0 .029 r
F P
0 g
s
.52322 .49465 m .51737 .48663 L .55326 .48418 L p 0 .542 .39 r
F P
0 g
s
.55326 .48418 m .55813 .49258 L .52322 .49465 L p 0 .542 .39 r
F P
0 g
s
.37578 .48133 m .37393 .47154 L .35756 .49195 L p .174 .182 0 r
F P
0 g
s
.35756 .49195 m .35938 .50098 L .37578 .48133 L p .174 .182 0 r
F P
0 g
s
.8444 .50441 m .84688 .49446 L .86353 .49719 L p 0 .109 .135 r
F P
0 g
s
.86353 .49719 m .86043 .50714 L .8444 .50441 L p 0 .109 .135 r
F P
0 g
s
.86983 .48896 m .86353 .49719 L .84688 .49446 L p 0 .015 .091 r
F P
0 g
s
.60534 .80449 m .59953 .80903 L .62587 .78507 L p .541 .345 0 r
F P
0 g
s
.62587 .78507 m .63162 .78047 L .60534 .80449 L p .541 .345 0 r
F P
0 g
s
.35217 .45118 m .34273 .45043 L .32524 .47252 L p .49 .822 0 r
F P
0 g
s
.31635 .47511 m .32524 .47252 L .34273 .45043 L p .542 .952 0 r
F P
0 g
s
.52635 .50456 m .52322 .49465 L .55813 .49258 L p 0 .675 .515 r
F P
0 g
s
.37417 .49056 m .37578 .48133 L .35938 .50098 L p 0 0 0 r
F P
0 g
s
.35938 .50098 m .35758 .5095 L .37417 .49056 L p 0 0 0 r
F P
0 g
s
.34273 .45043 m .33423 .45326 L .31635 .47511 L p .542 .952 0 r
F P
0 g
s
.30959 .48067 m .31635 .47511 L .33423 .45326 L p .505 .95 0 r
F P
0 g
s
.56414 .775 m .56611 .76674 L .58462 .74914 L p .14 0 0 r
F P
0 g
s
.59953 .80903 m .59195 .8102 L .6178 .78672 L p .728 .372 0 r
F P
0 g
s
.6178 .78672 m .62587 .78507 L .59953 .80903 L p .728 .372 0 r
F P
0 g
s
.36717 .49855 m .35758 .5095 L p .3693 .49801 L 0 0 0 r
F P
0 g
s
.3618 .50266 m .3693 .49801 L .35238 .51637 L p .517 .979 0 r
F P
0 g
s
.3693 .49801 m .3618 .50266 L .37143 .49747 L p 0 0 0 r
F P
0 g
s
.37143 .49747 m .38731 .48893 L p .3693 .49801 L 0 0 0 r
F P
0 g
s
.3693 .49801 m .37417 .49056 L .36717 .49855 L p 0 0 0 r
F P
0 g
s
.60486 .74043 m .61414 .74353 L .62902 .71425 L p .686 .235 0 r
F P
0 g
s
.61414 .74353 m .62273 .74925 L .63858 .71842 L p .602 .2 0 r
F P
0 g
s
.63858 .71842 m .62902 .71425 L .61414 .74353 L p .602 .2 0 r
F P
0 g
s
.62273 .74925 m .62949 .75685 L .64643 .72445 L p .508 .161 0 r
F P
0 g
s
.64643 .72445 m .63858 .71842 L .62273 .74925 L p .508 .161 0 r
F P
0 g
s
.5961 .74039 m .60486 .74043 L .61899 .71249 L p .774 .274 0 r
F P
0 g
s
.62902 .71425 m .61899 .71249 L .60486 .74043 L p .686 .235 0 r
F P
0 g
s
.62949 .75685 m .63353 .76534 L .6515 .73156 L p .376 .108 0 r
F P
0 g
s
.6515 .73156 m .64643 .72445 L .62949 .75685 L p .376 .108 0 r
F P
0 g
s
.33423 .45326 m .32783 .45932 L .30959 .48067 L p .505 .95 0 r
F P
0 g
s
.3059 .48845 m .30959 .48067 L .32783 .45932 L p 0 0 0 r
F P
0 g
s
.56525 .78453 m .56414 .775 L .58349 .75681 L p .35 .093 0 r
F P
0 g
s
.58462 .74914 m .58349 .75681 L .56414 .775 L p .14 0 0 r
F P
0 g
s
.59195 .8102 m .58363 .8078 L .60849 .78518 L p .677 .31 0 r
F P
0 g
s
.60849 .78518 m .6178 .78672 L .59195 .8102 L p .677 .31 0 r
F P
0 g
s
.35238 .51637 m .34446 .52063 L .3618 .50266 L p .517 .979 0 r
F P
0 g
s
.44613 .50307 m .4451 .51355 L .47974 .507 L p 0 .831 .671 r
F P
0 g
s
.47974 .507 m .48133 .49697 L .44613 .50307 L p 0 .831 .671 r
F P
0 g
s
.48129 .51722 m .47974 .507 L .4451 .51355 L p 0 .723 .557 r
F P
0 g
s
.4451 .51355 m .44747 .52384 L .48129 .51722 L p 0 .723 .557 r
F P
0 g
s
.48577 .52626 m .48129 .51722 L .44747 .52384 L p 0 .588 .422 r
F P
0 g
s
.61899 .71249 m .60981 .71337 L .5961 .74039 L p .774 .274 0 r
F P
0 g
s
.63353 .76534 m .63428 .77358 L .65311 .7388 L p .131 .013 0 r
F P
0 g
s
.65311 .7388 m .6515 .73156 L .63353 .76534 L p .131 .013 0 r
F P
0 g
s
.45046 .4938 m .44613 .50307 L .48133 .49697 L p 0 .922 .778 r
F P
0 g
s
.48133 .49697 m .48584 .48845 L .45046 .4938 L p 0 .922 .778 r
F P
0 g
s
.32783 .45932 m .3244 .4678 L .3059 .48845 L p 0 0 0 r
F P
0 g
s
.3058 .49743 m .3059 .48845 L .3244 .4678 L p .053 0 0 r
F P
0 g
s
.44747 .52384 m .45296 .53258 L .48577 .52626 L p 0 .588 .422 r
F P
0 g
s
.56928 .79402 m .56525 .78453 L .58581 .76539 L p .462 .161 0 r
F P
0 g
s
.58349 .75681 m .58581 .76539 L .56525 .78453 L p .35 .093 0 r
F P
0 g
s
.58363 .8078 m .57571 .80215 L .59923 .78063 L p .609 .257 0 r
F P
0 g
s
.59923 .78063 m .60849 .78518 L .58363 .8078 L p .609 .257 0 r
F P
0 g
s
.35266 .50384 m .3618 .50266 L .34446 .52063 L p .507 .898 0 r
F P
0 g
s
.34446 .52063 m .33487 .5217 L .35266 .50384 L p .507 .898 0 r
F P
0 g
s
.45749 .48698 m .45046 .4938 L .48584 .48845 L p 0 .974 .867 r
F P
0 g
s
.48584 .48845 m .4927 .48258 L .45749 .48698 L p 0 .974 .867 r
F P
0 g
s
.57571 .80215 m .56928 .79402 L .59131 .77372 L p .541 .211 0 r
F P
0 g
s
.58581 .76539 m .59131 .77372 L .56928 .79402 L p .462 .161 0 r
F P
0 g
s
.59131 .77372 m .59923 .78063 L .57571 .80215 L p .541 .211 0 r
F P
0 g
s
.914 .50748 m .91269 .51774 L .92528 .52361 L p 0 0 .089 r
F P
0 g
s
.91269 .51774 m .90793 .52695 L .91983 .53271 L p 0 0 .052 r
F P
0 g
s
.91983 .53271 m .92528 .52361 L .91269 .51774 L p 0 0 .052 r
F P
0 g
s
.90793 .52695 m .90038 .53386 L .91147 .53926 L p 0 0 .009 r
F P
0 g
s
.91147 .53926 m .91983 .53271 L .90793 .52695 L p 0 0 .009 r
F P
0 g
s
.3244 .4678 m .32445 .47757 L .3058 .49743 L p .053 0 0 r
F P
0 g
s
.30933 .50637 m .3058 .49743 L .32445 .47757 L p .213 .281 0 r
F P
0 g
s
.91171 .49754 m .914 .50748 L .92709 .51319 L p 0 .04 .122 r
F P
0 g
s
.92528 .52361 m .92709 .51319 L .914 .50748 L p 0 0 .089 r
F P
0 g
s
.90038 .53386 m .89104 .53754 L .90133 .54239 L p 0 0 0 r
F P
0 g
s
.90133 .54239 m .91147 .53926 L .90038 .53386 L p 0 0 0 r
F P
0 g
s
.63428 .77358 m .63162 .78047 L .651 .74521 L p 0 0 0 r
F P
0 g
s
.651 .74521 m .65311 .7388 L .63428 .77358 L p 0 0 0 r
F P
0 g
s
.34313 .50139 m .35266 .50384 L .33487 .5217 L p .449 .759 0 r
F P
0 g
s
.33487 .5217 m .32493 .5194 L .34313 .50139 L p .449 .759 0 r
F P
0 g
s
.90609 .48925 m .91171 .49754 L .92501 .50287 L p 0 .165 .149 r
F P
0 g
s
.92709 .51319 m .92501 .50287 L .91171 .49754 L p 0 .04 .122 r
F P
0 g
s
.89104 .53754 m .88116 .53751 L .89078 .54168 L p 0 0 0 r
F P
0 g
s
.89078 .54168 m .90133 .54239 L .89104 .53754 L p 0 0 0 r
F P
0 g
s
.46631 .48354 m .45749 .48698 L .4927 .48258 L p 0 .828 .81 r
F P
0 g
s
.4927 .48258 m .50099 .48018 L .46631 .48354 L p 0 .828 .81 r
F P
0 g
s
.32445 .47757 m .32798 .48729 L .30933 .50637 L p .213 .281 0 r
F P
0 g
s
.31601 .51405 m .30933 .50637 L .32798 .48729 L p .311 .472 0 r
F P
0 g
s
.90609 .48925 m .91931 .49402 L p .90029 .48487 L 0 .302 .162 r
F P
0 g
s
.8991 .48453 m .90609 .48925 L p .90029 .48487 L 0 .302 .162 r
F P
0 g
s
.91931 .49402 m .91076 .48786 L .89792 .48373 L p 0 .302 .162 r
F P
0 g
s
.90053 .48521 m .88827 .48175 L p .90829 .48719 L 0 0 0 r
F P
0 g
s
.91076 .48786 m .90936 .4875 L p .90829 .48719 L 0 0 0 r
F P
0 g
s
.90936 .4875 m .90053 .48521 L p .90829 .48719 L 0 0 0 r
F P
0 g
s
.89792 .48373 m .8991 .48453 L p .90029 .48487 L 0 .302 .162 r
F P
0 g
s
.92501 .50287 m .91931 .49402 L .90609 .48925 L p 0 .165 .149 r
F P
0 g
s
.88116 .53751 m .87208 .53374 L .88124 .53721 L p 0 .292 .139 r
F P
0 g
s
.874 .52959 m .88124 .53721 L .87208 .53374 L p 0 .176 .148 r
F P
0 g
s
.87523 .53136 m .87928 .53597 L p .88124 .53721 L 0 .082 0 r
F P
0 g
s
.88291 .53826 m .88116 .53751 L p .88124 .53721 L 0 .292 .139 r
F P
0 g
s
.87928 .53597 m .88238 .53949 L .88124 .53721 L p 0 .082 0 r
F P
0 g
s
.88124 .53721 m .89078 .54168 L .88291 .53826 L p 0 .292 .139 r
F P
0 g
s
.3345 .49561 m .34313 .50139 L .32493 .5194 L p .385 .621 0 r
F P
0 g
s
.32493 .5194 m .31601 .51405 L .3345 .49561 L p .385 .621 0 r
F P
0 g
s
.32798 .48729 m .3345 .49561 L .31601 .51405 L p .311 .472 0 r
F P
0 g
s
.88827 .48175 m .89792 .48373 L .91076 .48786 L p 0 0 0 r
F P
0 g
s
.87208 .53374 m .86503 .52674 L .874 .52959 L p 0 .176 .148 r
F P
0 g
s
.87847 .48357 m .88827 .48175 L .90053 .48521 L p 0 0 0 r
F P
0 g
s
.86503 .52674 m .86096 .51746 L .87003 .51985 L p 0 .045 .129 r
F P
0 g
s
.87003 .51985 m .874 .52959 L .86503 .52674 L p 0 .045 .129 r
F P
0 g
s
.47574 .48397 m .46631 .48354 L .50099 .48018 L p 0 0 0 r
F P
0 g
s
.50099 .48018 m .5096 .48158 L .47574 .48397 L p 0 0 0 r
F P
0 g
s
.63162 .78047 m .62587 .78507 L .64542 .74989 L p .862 .332 0 r
F P
0 g
s
.64542 .74989 m .651 .74521 L .63162 .78047 L p .862 .332 0 r
F P
0 g
s
.86983 .48896 m .87847 .48357 L .88999 .48645 L p 0 0 0 r
F P
0 g
s
.90053 .48521 m .88999 .48645 L .87847 .48357 L p 0 0 0 r
F P
0 g
s
.86096 .51746 m .86043 .50714 L .86987 .5093 L p 0 0 .098 r
F P
0 g
s
.86987 .5093 m .87003 .51985 L .86096 .51746 L p 0 0 .098 r
F P
0 g
s
.86353 .49719 m .86983 .48896 L .88057 .49139 L p 0 0 .018 r
F P
0 g
s
.88999 .48645 m .88057 .49139 L .86983 .48896 L p 0 0 0 r
F P
0 g
s
.39803 .45527 m .39313 .4459 L .36887 .46249 L p .299 .443 0 r
F P
0 g
s
.36887 .46249 m .37393 .47154 L .39803 .45527 L p .299 .443 0 r
F P
0 g
s
.36128 .45539 m .36887 .46249 L .39313 .4459 L p .338 .574 0 r
F P
0 g
s
.86043 .50714 m .86353 .49719 L .87355 .49936 L p 0 0 .061 r
F P
0 g
s
.87355 .49936 m .86987 .5093 L .86043 .50714 L p 0 0 .061 r
F P
0 g
s
.88057 .49139 m .87355 .49936 L .86353 .49719 L p 0 0 .018 r
F P
0 g
s
.39995 .46558 m .39803 .45527 L .37393 .47154 L p .25 .293 0 r
F P
0 g
s
.37393 .47154 m .37578 .48133 L .39995 .46558 L p .25 .293 0 r
F P
0 g
s
.61899 .71249 m .62902 .71425 L .63725 .67569 L p .711 .19 0 r
F P
0 g
s
.62902 .71425 m .63858 .71842 L .64741 .67815 L p .622 .165 0 r
F P
0 g
s
.64741 .67815 m .63725 .67569 L .62902 .71425 L p .622 .165 0 r
F P
0 g
s
.39313 .4459 m .38588 .43871 L .36128 .45539 L p .338 .574 0 r
F P
0 g
s
.35217 .45118 m .36128 .45539 L .38588 .43871 L p .372 .707 0 r
F P
0 g
s
.63858 .71842 m .64643 .72445 L .65599 .68245 L p .516 .136 0 r
F P
0 g
s
.65599 .68245 m .64741 .67815 L .63858 .71842 L p .516 .136 0 r
F P
0 g
s
.60981 .71337 m .61899 .71249 L .62684 .67538 L p .801 .215 0 r
F P
0 g
s
.63725 .67569 m .62684 .67538 L .61899 .71249 L p .711 .19 0 r
F P
0 g
s
.48449 .48821 m .47574 .48397 L .5096 .48158 L p 0 .211 .049 r
F P
0 g
s
.5096 .48158 m .51737 .48663 L .48449 .48821 L p 0 .211 .049 r
F P
0 g
s
.39863 .47547 m .39995 .46558 L .37578 .48133 L p .176 .087 0 r
F P
0 g
s
.37578 .48133 m .37417 .49056 L .39863 .47547 L p .176 .087 0 r
F P
0 g
s
.58581 .76539 m .58349 .75681 L .5982 .72924 L p .258 .063 0 r
F P
0 g
s
.59865 .72233 m .5982 .72924 L .58349 .75681 L p 0 0 0 r
F P
0 g
s
.38588 .43871 m .37725 .43467 L .35217 .45118 L p .372 .707 0 r
F P
0 g
s
.34273 .45043 m .35217 .45118 L .37725 .43467 L p .395 .849 0 r
F P
0 g
s
.62587 .78507 m .6178 .78672 L .6371 .75221 L p .822 .299 0 r
F P
0 g
s
.6371 .75221 m .64542 .74989 L .62587 .78507 L p .822 .299 0 r
F P
0 g
s
.64643 .72445 m .6515 .73156 L .66184 .68803 L p .351 .09 0 r
F P
0 g
s
.66184 .68803 m .65599 .68245 L .64643 .72445 L p .351 .09 0 r
F P
0 g
s
.60271 .7168 m .60981 .71337 L .61755 .67727 L p .903 .243 0 r
F P
0 g
s
.62684 .67538 m .61755 .67727 L .60981 .71337 L p .801 .215 0 r
F P
0 g
s
.49137 .49573 m .48449 .48821 L .51737 .48663 L p 0 .491 .311 r
F P
0 g
s
.51737 .48663 m .52322 .49465 L .49137 .49573 L p 0 .491 .311 r
F P
0 g
s
.39423 .48362 m .39863 .47547 L .37417 .49056 L p .04 0 0 r
F P
0 g
s
.37417 .49056 m .3693 .49801 L .39423 .48362 L p .04 0 0 r
F P
0 g
s
.37725 .43467 m .36837 .43433 L .34273 .45043 L p .395 .849 0 r
F P
0 g
s
.33423 .45326 m .34273 .45043 L .36837 .43433 L p .351 .91 0 r
F P
0 g
s
.59131 .77372 m .58581 .76539 L .60147 .73661 L p .442 .135 0 r
F P
0 g
s
.5982 .72924 m .60147 .73661 L .58581 .76539 L p .258 .063 0 r
F P
0 g
s
.49544 .50549 m .49137 .49573 L .52322 .49465 L p 0 .662 .479 r
F P
0 g
s
.52322 .49465 m .52635 .50456 L .49544 .50549 L p 0 .662 .479 r
F P
0 g
s
.49334 .52629 m .49613 .51617 L .52631 .51499 L p 0 .891 .721 r
F P
0 g
s
.52635 .50456 m .52631 .51499 L .49613 .51617 L p 0 .787 .607 r
F P
0 g
s
.6178 .78672 m .60849 .78518 L .62718 .7518 L p .729 .255 0 r
F P
0 g
s
.62718 .7518 m .6371 .75221 L .6178 .78672 L p .729 .255 0 r
F P
0 g
s
.49613 .51617 m .49544 .50549 L .52635 .50456 L p 0 .787 .607 r
F P
0 g
s
.6515 .73156 m .65311 .7388 L .66416 .69416 L p .005 0 0 r
F P
0 g
s
.66416 .69416 m .66184 .68803 L .6515 .73156 L p .005 0 0 r
F P
0 g
s
.38731 .48893 m .39423 .48362 L .38226 .49053 L p 0 0 0 r
F P
0 g
s
.39423 .48362 m .38731 .48893 L .42082 .47966 L p .092 0 0 r
F P
0 g
s
.38731 .48893 m .37879 .49067 L .39181 .48751 L p .128 .726 0 r
F P
0 g
s
.41308 .48234 m .42082 .47966 L .40034 .48533 L p .128 .726 0 r
F P
0 g
s
.40034 .48533 m .38731 .48893 L p .41308 .48234 L .128 .726 0 r
F P
0 g
s
.42082 .47966 m .427 .47341 L .39423 .48362 L p .092 0 0 r
F P
0 g
s
.42082 .47966 m .41308 .48234 L .45477 .47863 L p .097 0 0 r
F P
0 g
s
.41308 .48234 m .40481 .48107 L .42305 .48001 L p .136 .727 0 r
F P
0 g
s
.45477 .47863 m .46159 .4747 L .43725 .47766 L p .097 0 0 r
F P
0 g
s
.43725 .47766 m .42082 .47966 L p .45477 .47863 L .097 0 0 r
F P
0 g
s
.44737 .47859 m .45477 .47863 L .41308 .48234 L p .136 .727 0 r
F P
0 g
s
.45477 .47863 m .44737 .47859 L .46872 .47941 L p .031 0 0 r
F P
0 g
s
.46872 .47941 m .49613 .48046 L p .45477 .47863 L .031 0 0 r
F P
0 g
s
.42305 .48001 m .44737 .47859 L p .41308 .48234 L .136 .727 0 r
F P
0 g
s
.39181 .48751 m .41308 .48234 L p .38731 .48893 L .128 .726 0 r
F P
0 g
s
.38226 .49053 m .3693 .49801 L p .38731 .48893 L 0 0 0 r
F P
0 g
s
.61755 .67727 m .61061 .68112 L .60271 .7168 L p .903 .243 0 r
F P
0 g
s
.59923 .78063 m .59131 .77372 L .60804 .74342 L p .554 .181 0 r
F P
0 g
s
.60147 .73661 m .60804 .74342 L .59131 .77372 L p .442 .135 0 r
F P
0 g
s
.60849 .78518 m .59923 .78063 L .61702 .74873 L p .642 .218 0 r
F P
0 g
s
.61702 .74873 m .62718 .7518 L .60849 .78518 L p .642 .218 0 r
F P
0 g
s
.36837 .43433 m .36045 .43778 L .33423 .45326 L p .351 .91 0 r
F P
0 g
s
.32783 .45932 m .33423 .45326 L .36045 .43778 L p 0 0 0 r
F P
0 g
s
.60804 .74342 m .61702 .74873 L .59923 .78063 L p .554 .181 0 r
F P
0 g
s
.37879 .49067 m .38731 .48893 L .3618 .50266 L p .357 .896 0 r
F P
0 g
s
.3618 .50266 m .35266 .50384 L .37879 .49067 L p .357 .896 0 r
F P
0 g
s
.65311 .7388 m .651 .74521 L .6626 .70001 L p .706 .196 0 r
F P
0 g
s
.6626 .70001 m .66416 .69416 L .65311 .7388 L p .706 .196 0 r
F P
0 g
s
.41929 .59517 m .4223 .60525 L .47503 .58222 L p .461 0 .726 r
F P
0 g
s
.47503 .58222 m .47075 .57274 L .41929 .59517 L p .461 0 .726 r
F P
0 g
s
.47921 .59161 m .47503 .58222 L .4223 .60525 L p .444 0 .654 r
F P
0 g
s
.47075 .57274 m .47503 .58222 L .52573 .55196 L p .417 0 .753 r
F P
0 g
s
.47503 .58222 m .47921 .59161 L .53097 .56102 L p .397 0 .687 r
F P
0 g
s
.53097 .56102 m .52573 .55196 L .47503 .58222 L p .397 0 .687 r
F P
0 g
s
.36045 .43778 m .35456 .44457 L .32783 .45932 L p 0 0 0 r
F P
0 g
s
.3244 .4678 m .32783 .45932 L .35456 .44457 L p .081 0 0 r
F P
0 g
s
.52573 .55196 m .5204 .54331 L .47075 .57274 L p .417 0 .753 r
F P
0 g
s
.41668 .58586 m .41929 .59517 L .47075 .57274 L p .475 0 .796 r
F P
0 g
s
.47075 .57274 m .46693 .56441 L .41668 .58586 L p .475 0 .796 r
F P
0 g
s
.46693 .56441 m .47075 .57274 L .5204 .54331 L p .433 0 .821 r
F P
0 g
s
.4223 .60525 m .42532 .61479 L .47921 .59161 L p .444 0 .654 r
F P
0 g
s
.48274 .59968 m .47921 .59161 L .42532 .61479 L p .417 0 .562 r
F P
0 g
s
.47921 .59161 m .48274 .59968 L .53544 .56932 L p .372 0 .608 r
F P
0 g
s
.53544 .56932 m .53097 .56102 L .47921 .59161 L p .372 0 .608 r
F P
0 g
s
.36735 .60911 m .36884 .61957 L .4223 .60525 L p .51 0 .702 r
F P
0 g
s
.4223 .60525 m .41929 .59517 L .36735 .60911 L p .51 0 .702 r
F P
0 g
s
.42532 .61479 m .4223 .60525 L .36884 .61957 L p .495 0 .623 r
F P
0 g
s
.5204 .54331 m .51566 .53622 L .46693 .56441 L p .433 0 .821 r
F P
0 g
s
.36981 .48856 m .37879 .49067 L .35266 .50384 L p .369 .78 0 r
F P
0 g
s
.35266 .50384 m .34313 .50139 L .36981 .48856 L p .369 .78 0 r
F P
0 g
s
.36629 .59912 m .36735 .60911 L .41929 .59517 L p .519 0 .775 r
F P
0 g
s
.41929 .59517 m .41668 .58586 L .36629 .59912 L p .519 0 .775 r
F P
0 g
s
.62684 .67538 m .63725 .67569 L .63783 .62913 L p .73 .146 0 r
F P
0 g
s
.63725 .67569 m .64741 .67815 L .6483 .62986 L p .642 .13 0 r
F P
0 g
s
.6483 .62986 m .63783 .62913 L .63725 .67569 L p .642 .13 0 r
F P
0 g
s
.35456 .44457 m .35153 .4538 L .3244 .4678 L p .081 0 0 r
F P
0 g
s
.32445 .47757 m .3244 .4678 L .35153 .4538 L p .195 .172 0 r
F P
0 g
s
.61755 .67727 m .62684 .67538 L .62734 .63027 L p .817 .161 0 r
F P
0 g
s
.63783 .62913 m .62734 .63027 L .62684 .67538 L p .73 .146 0 r
F P
0 g
s
.41481 .57855 m .41668 .58586 L .46693 .56441 L p .482 0 .88 r
F P
0 g
s
.46693 .56441 m .46407 .55833 L .41481 .57855 L p .482 0 .88 r
F P
0 g
s
.46407 .55833 m .46693 .56441 L .51566 .53622 L p .445 0 .904 r
F P
0 g
s
.64741 .67815 m .65599 .68245 L .65737 .63238 L p .532 .109 0 r
F P
0 g
s
.65737 .63238 m .6483 .62986 L .64741 .67815 L p .532 .109 0 r
F P
0 g
s
.36884 .61957 m .37057 .62912 L .42532 .61479 L p .495 0 .623 r
F P
0 g
s
.42795 .62254 m .42532 .61479 L .37057 .62912 L p .469 0 .514 r
F P
0 g
s
.42532 .61479 m .42795 .62254 L .48274 .59968 L p .417 0 .562 r
F P
0 g
s
.53853 .57576 m .53544 .56932 L .48274 .59968 L p .327 0 .484 r
F P
0 g
s
.36581 .59091 m .36629 .59912 L .41668 .58586 L p .52 0 .858 r
F P
0 g
s
.41668 .58586 m .41481 .57855 L .36581 .59091 L p .52 0 .858 r
F P
0 g
s
.651 .74521 m .64542 .74989 L .65732 .70479 L p .95 .258 0 r
F P
0 g
s
.65732 .70479 m .6626 .70001 L .651 .74521 L p .95 .258 0 r
F P
0 g
s
.36161 .48287 m .36981 .48856 L .34313 .50139 L p .342 .637 0 r
F P
0 g
s
.34313 .50139 m .3345 .49561 L .36161 .48287 L p .342 .637 0 r
F P
0 g
s
.48515 .60536 m .48274 .59968 L .42795 .62254 L p .362 0 .401 r
F P
0 g
s
.48274 .59968 m .48515 .60536 L .53853 .57576 L p .327 0 .484 r
F P
0 g
s
.60147 .73661 m .5982 .72924 L .60713 .69251 L p .193 .048 0 r
F P
0 g
s
.60696 .68643 m .60713 .69251 L .5982 .72924 L p 0 0 0 r
F P
0 g
s
.35153 .4538 m .35178 .46421 L .32445 .47757 L p .195 .172 0 r
F P
0 g
s
.32798 .48729 m .32445 .47757 L .35178 .46421 L p .26 .357 0 r
F P
0 g
s
.61061 .68112 m .61755 .67727 L .6182 .63314 L p .918 .178 0 r
F P
0 g
s
.62734 .63027 m .6182 .63314 L .61755 .67727 L p .817 .161 0 r
F P
0 g
s
.3173 .60319 m .31645 .61349 L .36735 .60911 L p .565 0 .758 r
F P
0 g
s
.36735 .60911 m .36629 .59912 L .3173 .60319 L p .565 0 .758 r
F P
0 g
s
.36884 .61957 m .36735 .60911 L .31645 .61349 L p .563 0 .683 r
F P
0 g
s
.35531 .47437 m .36161 .48287 L .3345 .49561 L p .306 .502 0 r
F P
0 g
s
.3345 .49561 m .32798 .48729 L .35531 .47437 L p .306 .502 0 r
F P
0 g
s
.35178 .46421 m .35531 .47437 L .32798 .48729 L p .26 .357 0 r
F P
0 g
s
.31645 .61349 m .31615 .62402 L .36884 .61957 L p .563 0 .683 r
F P
0 g
s
.37057 .62912 m .36884 .61957 L .31615 .62402 L p .555 0 .596 r
F P
0 g
s
.65599 .68245 m .66184 .68803 L .66383 .63637 L p .354 .076 0 r
F P
0 g
s
.66383 .63637 m .65737 .63238 L .65599 .68245 L p .354 .076 0 r
F P
0 g
s
.31859 .59448 m .3173 .60319 L .36629 .59912 L p .557 0 .839 r
F P
0 g
s
.36629 .59912 m .36581 .59091 L .31859 .59448 L p .557 0 .839 r
F P
0 g
s
.41394 .57423 m .41481 .57855 L .46407 .55833 L p .448 0 .966 r
F P
0 g
s
.37057 .62912 m .37232 .63651 L .42795 .62254 L p .469 0 .514 r
F P
0 g
s
.42984 .62746 m .42795 .62254 L .37232 .63651 L p .402 0 .303 r
F P
0 g
s
.42795 .62254 m .42984 .62746 L .48515 .60536 L p .362 0 .401 r
F P
0 g
s
.91147 .53926 m .90133 .54239 L .90285 .54583 L p 0 0 0 r
F P
0 g
s
.90133 .54239 m .89078 .54168 L .89202 .54455 L p 0 0 0 r
F P
0 g
s
.89202 .54455 m .90285 .54583 L .90133 .54239 L p 0 0 0 r
F P
0 g
s
.89078 .54168 m .88124 .53721 L .88238 .53949 L p 0 0 0 r
F P
0 g
s
.88238 .53949 m .89202 .54455 L .89078 .54168 L p 0 0 0 r
F P
0 g
s
.91983 .53271 m .91147 .53926 L .91342 .54316 L p 0 0 0 r
F P
0 g
s
.90285 .54583 m .91342 .54316 L .91147 .53926 L p 0 0 0 r
F P
0 g
s
.88124 .53721 m .874 .52959 L .87523 .53136 L p 0 .082 0 r
F P
0 g
s
.36597 .58557 m .36581 .59091 L .41481 .57855 L p .489 0 .96 r
F P
0 g
s
.41481 .57855 m .41394 .57423 L .36597 .58557 L p .489 0 .96 r
F P
0 g
s
.92528 .52361 m .91983 .53271 L .92229 .53689 L p 0 0 0 r
F P
0 g
s
.91342 .54316 m .92229 .53689 L .91983 .53271 L p 0 0 0 r
F P
0 g
s
.874 .52959 m .87003 .51985 L .87152 .52124 L p 0 0 .042 r
F P
0 g
s
.87152 .52124 m .87523 .53136 L .874 .52959 L p 0 0 .042 r
F P
0 g
s
.53982 .57946 m .53853 .57576 L .48515 .60536 L p .212 0 .199 r
F P
0 g
s
.64542 .74989 m .6371 .75221 L .64902 .70782 L p .854 .23 0 r
F P
0 g
s
.64902 .70782 m .65732 .70479 L .64542 .74989 L p .854 .23 0 r
F P
0 g
s
.92709 .51319 m .92528 .52361 L .92825 .52786 L p 0 0 .008 r
F P
0 g
s
.92229 .53689 m .92825 .52786 L .92528 .52361 L p 0 0 0 r
F P
0 g
s
.87003 .51985 m .86987 .5093 L .87174 .51049 L p 0 0 .012 r
F P
0 g
s
.87174 .51049 m .87152 .52124 L .87003 .51985 L p 0 0 .012 r
F P
0 g
s
.31615 .62402 m .31645 .63342 L .37057 .62912 L p .555 0 .596 r
F P
0 g
s
.37232 .63651 m .37057 .62912 L .31645 .63342 L p .531 0 .467 r
F P
0 g
s
.60804 .74342 m .60147 .73661 L .61114 .69854 L p .436 .114 0 r
F P
0 g
s
.60713 .69251 m .61114 .69854 L .60147 .73661 L p .193 .048 0 r
F P
0 g
s
.92501 .50287 m .92709 .51319 L .9304 .51717 L p 0 0 .039 r
F P
0 g
s
.93035 .5169 m .92501 .50287 L .9304 .51717 L 0 0 .039 r
F
.93049 .51729 m .92871 .50662 L .92501 .50287 L p F P
0 g
s
.92311 .51891 m .92168 .50796 L .92842 .50667 L p 0 .696 .02 r
F P
0 g
s
.92842 .50667 m .92871 .50662 L p .9285 .50709 L 0 .696 .02 r
F P
0 g
s
.92842 .50667 m .9285 .50709 L .92311 .51891 L 0 .696 .02 r
F
.9304 .51717 m .93049 .51729 L p .93035 .5169 L 0 0 .039 r
F P
0 g
s
.92825 .52786 m .93049 .51729 L .92709 .51319 L p 0 0 .008 r
F P
0 g
s
.86987 .5093 m .87355 .49936 L .87587 .50055 L p 0 0 0 r
F P
0 g
s
.87587 .50055 m .87174 .51049 L .86987 .5093 L p 0 0 0 r
F P
0 g
s
.41597 .50779 m .41564 .51866 L .4451 .51355 L p 0 .836 .653 r
F P
0 g
s
.4451 .51355 m .44613 .50307 L .41597 .50779 L p 0 .836 .653 r
F P
0 g
s
.44747 .52384 m .4451 .51355 L .41564 .51866 L p 0 .707 .506 r
F P
0 g
s
.91931 .49402 m .91997 .49505 L p .92018 .49519 L 0 0 .067 r
F P
0 g
s
.92871 .50662 m .92314 .4973 L .91931 .49402 L p 0 0 .067 r
F P
0 g
s
.91647 .4983 m .90819 .49124 L .9139 .49066 L p 0 .096 0 r
F P
0 g
s
.91931 .49402 m .9229 .4971 L p .91192 .48874 L 0 .107 .086 r
F P
0 g
s
.92245 .49677 m .91192 .48874 L .9229 .4971 L 0 .107 .086 r
F
.91076 .48786 m .91176 .48858 L p .91192 .48874 L F P
0 g
s
.91176 .48858 m .91931 .49402 L p .91192 .48874 L 0 .107 .086 r
F P
0 g
s
.91455 .4906 m .91076 .48786 L p .92198 .49642 L 0 .107 .086 r
F P
0 g
s
.9139 .49066 m .91455 .4906 L p .91471 .49124 L 0 .096 0 r
F P
0 g
s
.91647 .4983 m .9139 .49066 L .91471 .49124 L 0 .096 0 r
F
.9229 .4971 m .92314 .4973 L p .92245 .49677 L 0 .107 .086 r
F P
0 g
s
.9041 .48741 m .90053 .48521 L p .91324 .4901 L 0 0 0 r
F P
0 g
s
.91455 .4906 m .91339 .49025 L p .91324 .4901 L 0 0 0 r
F P
0 g
s
.91339 .49025 m .9041 .48741 L p .91324 .4901 L 0 0 0 r
F P
0 g
s
.92314 .4973 m .92218 .49655 L p .92198 .49642 L 0 .107 .086 r
F P
0 g
s
.92218 .49655 m .91455 .4906 L p .92198 .49642 L 0 .107 .086 r
F P
0 g
s
.92501 .50287 m .92857 .50648 L p .92018 .49519 L 0 0 .067 r
F P
0 g
s
.92839 .50619 m .92018 .49519 L .92857 .50648 L 0 0 .067 r
F
.91997 .49505 m .92501 .50287 L p .92018 .49519 L F P
0 g
s
.92168 .50796 m .91647 .4983 L .92277 .49736 L p 0 .429 0 r
F P
0 g
s
.92277 .49736 m .92314 .4973 L p .92307 .49782 L 0 .429 0 r
F P
0 g
s
.92168 .50796 m .92277 .49736 L .92307 .49782 L 0 .429 0 r
F
.92857 .50648 m .92871 .50662 L p .92839 .50619 L 0 0 .067 r
F P
0 g
s
.87355 .49936 m .88057 .49139 L .88337 .49276 L p 0 0 0 r
F P
0 g
s
.88337 .49276 m .87587 .50055 L .87355 .49936 L p 0 0 0 r
F P
0 g
s
.48611 .60786 m .48515 .60536 L .42984 .62746 L p .186 0 0 r
F P
0 g
s
.48515 .60536 m .48611 .60786 L .53982 .57946 L p .212 0 .199 r
F P
0 g
s
.6182 .63314 m .61162 .63737 L .61061 .68112 L p .918 .178 0 r
F P
0 g
s
.41564 .51866 m .41896 .52895 L .44747 .52384 L p 0 .707 .506 r
F P
0 g
s
.45296 .53258 m .44747 .52384 L .41896 .52895 L p 0 .532 .323 r
F P
0 g
s
.41991 .4978 m .41597 .50779 L .44613 .50307 L p 0 .935 .783 r
F P
0 g
s
.44613 .50307 m .45046 .4938 L .41991 .4978 L p 0 .935 .783 r
F P
0 g
s
.88057 .49139 m .88999 .48645 L .89321 .48817 L p 0 0 0 r
F P
0 g
s
.89321 .48817 m .88337 .49276 L .88057 .49139 L p 0 0 0 r
F P
0 g
s
.43177 .454 m .42975 .44342 L .39803 .45527 L p .28 .375 0 r
F P
0 g
s
.39803 .45527 m .39995 .46558 L .43177 .454 L p .28 .375 0 r
F P
0 g
s
.39313 .4459 m .39803 .45527 L .42975 .44342 L p .298 .487 0 r
F P
0 g
s
.90053 .48521 m .91076 .48786 L .91455 .4906 L p 0 0 0 r
F P
0 g
s
.88999 .48645 m .90053 .48521 L .9041 .48741 L p 0 0 0 r
F P
0 g
s
.9041 .48741 m .89321 .48817 L .88999 .48645 L p 0 0 0 r
F P
0 g
s
.42975 .44342 m .42501 .43411 L .39313 .4459 L p .298 .487 0 r
F P
0 g
s
.38588 .43871 m .39313 .4459 L .42501 .43411 L p .31 .6 0 r
F P
0 g
s
.43081 .46446 m .43177 .454 L .39995 .46558 L p .253 .241 0 r
F P
0 g
s
.39995 .46558 m .39863 .47547 L .43081 .46446 L p .253 .241 0 r
F P
0 g
s
.32016 .58851 m .31859 .59448 L .36581 .59091 L p .519 0 .942 r
F P
0 g
s
.36581 .59091 m .36597 .58557 L .32016 .58851 L p .519 0 .942 r
F P
0 g
s
.41896 .52895 m .4255 .53729 L .45296 .53258 L p 0 .532 .323 r
F P
0 g
s
.46082 .53859 m .45296 .53258 L .4255 .53729 L p 0 .268 .066 r
F P
0 g
s
.66184 .68803 m .66416 .69416 L .6668 .6413 L p 0 0 0 r
F P
0 g
s
.6668 .6413 m .66383 .63637 L .66184 .68803 L p 0 0 0 r
F P
0 g
s
.42694 .49003 m .41991 .4978 L .45046 .4938 L p 0 .986 .89 r
F P
0 g
s
.45046 .4938 m .45749 .48698 L .42694 .49003 L p 0 .986 .89 r
F P
0 g
s
.27129 .59755 m .26826 .60766 L .31645 .61349 L p .614 0 .745 r
F P
0 g
s
.31645 .61349 m .3173 .60319 L .27129 .59755 L p .614 0 .745 r
F P
0 g
s
.31615 .62402 m .31645 .61349 L .26826 .60766 L p .622 0 .669 r
F P
0 g
s
.6371 .75221 m .62718 .7518 L .63882 .70868 L p .756 .203 0 r
F P
0 g
s
.63882 .70868 m .64902 .70782 L .6371 .75221 L p .756 .203 0 r
F P
0 g
s
.61702 .74873 m .60804 .74342 L .61848 .70369 L p .57 .151 0 r
F P
0 g
s
.61114 .69854 m .61848 .70369 L .60804 .74342 L p .436 .114 0 r
F P
0 g
s
.42501 .43411 m .41818 .4273 L .38588 .43871 L p .31 .6 0 r
F P
0 g
s
.37725 .43467 m .38588 .43871 L .41818 .4273 L p .311 .728 0 r
F P
0 g
s
.27465 .58887 m .27129 .59755 L .3173 .60319 L p .594 0 .824 r
F P
0 g
s
.3173 .60319 m .31859 .59448 L .27465 .58887 L p .594 0 .824 r
F P
0 g
s
.26826 .60766 m .26595 .61789 L .31615 .62402 L p .622 0 .669 r
F P
0 g
s
.31645 .63342 m .31615 .62402 L .26595 .61789 L p .623 0 .576 r
F P
0 g
s
.427 .47341 m .43081 .46446 L .39863 .47547 L p .203 .047 0 r
F P
0 g
s
.39863 .47547 m .39423 .48362 L .427 .47341 L p .203 .047 0 r
F P
0 g
s
.4255 .53729 m .43438 .54256 L .46082 .53859 L p 0 .268 .066 r
F P
0 g
s
.47001 .54106 m .46082 .53859 L .43438 .54256 L p 0 0 0 r
F P
0 g
s
.37232 .63651 m .37384 .64075 L .42984 .62746 L p .402 0 .303 r
F P
0 g
s
.43073 .62885 m .42984 .62746 L .37384 .64075 L p .134 0 0 r
F P
0 g
s
.42984 .62746 m .43073 .62885 L .48611 .60786 L p .186 0 0 r
F P
0 g
s
.31645 .63342 m .3173 .64044 L .37232 .63651 L p .531 0 .467 r
F P
0 g
s
.37384 .64075 m .37232 .63651 L .3173 .64044 L p .453 0 .193 r
F P
0 g
s
.43613 .48552 m .42694 .49003 L .45749 .48698 L p 0 .895 .897 r
F P
0 g
s
.45749 .48698 m .46631 .48354 L .43613 .48552 L p 0 .895 .897 r
F P
0 g
s
.4142 .5735 m .41394 .57423 L .43258 .56698 L p 0 0 0 r
F P
0 g
s
.36676 .58384 m .36597 .58557 L .384 .58131 L p .209 0 .823 r
F P
0 g
s
.384 .58131 m .41394 .57423 L p .36676 .58384 L .209 0 .823 r
F P
0 g
s
.32179 .58611 m .32016 .58851 L .3372 .58742 L p .296 0 .96 r
F P
0 g
s
.3372 .58742 m .36597 .58557 L p .32179 .58611 L .296 0 .96 r
F P
0 g
s
.32016 .58851 m .32179 .58611 L .28063 .5801 L p .343 0 .996 r
F P
0 g
s
.32179 .58611 m .32328 .58764 L .30378 .58459 L p .316 0 0 r
F P
0 g
s
.28063 .5801 m .27791 .58278 L .29378 .58493 L p .343 0 .996 r
F P
0 g
s
.29378 .58493 m .32016 .58851 L p .28063 .5801 L .343 0 .996 r
F P
0 g
s
.27791 .58278 m .28063 .5801 L .24429 .56604 L p .367 0 .98 r
F P
0 g
s
.28063 .5801 m .28243 .58124 L .26517 .57442 L p .308 0 0 r
F P
0 g
s
.26517 .57442 m .24648 .56704 L p .28063 .5801 L .308 0 0 r
F P
0 g
s
.30378 .58459 m .28243 .58124 L p .32179 .58611 L .316 0 0 r
F P
0 g
s
.36597 .58557 m .36676 .58384 L .32179 .58611 L p .296 0 .96 r
F P
0 g
s
.36676 .58384 m .36808 .58601 L .34671 .58679 L p .312 0 0 r
F P
0 g
s
.34671 .58679 m .32328 .58764 L p .36676 .58384 L .312 0 0 r
F P
0 g
s
.41394 .57423 m .4142 .5735 L .36676 .58384 L p .209 0 .823 r
F P
0 g
s
.4142 .5735 m .41556 .57652 L .39296 .58103 L p .306 0 .048 r
F P
0 g
s
.39296 .58103 m .36808 .58601 L p .4142 .5735 L .306 0 .048 r
F P
0 g
s
.43258 .56698 m .46256 .55533 L p .4142 .5735 L 0 0 0 r
F P
0 g
s
.46256 .55533 m .46262 .55584 L .4142 .5735 L p 0 0 0 r
F P
0 g
s
.62718 .7518 m .61702 .74873 L .62815 .70723 L p .667 .178 0 r
F P
0 g
s
.62815 .70723 m .63882 .70868 L .62718 .7518 L p .667 .178 0 r
F P
0 g
s
.61848 .70369 m .62815 .70723 L .61702 .74873 L p .57 .151 0 r
F P
0 g
s
.53913 .5799 m .53982 .57946 L .48611 .60786 L p 0 0 0 r
F P
0 g
s
.53982 .57946 m .53913 .5799 L .59018 .54704 L p 0 0 0 r
F P
0 g
s
.48547 .60682 m .48611 .60786 L .43073 .62885 L p .319 0 .872 r
F P
0 g
s
.48611 .60786 m .48547 .60682 L .53913 .5799 L p 0 0 0 r
F P
0 g
s
.41818 .4273 m .41016 .42392 L .37725 .43467 L p .311 .728 0 r
F P
0 g
s
.36837 .43433 m .37725 .43467 L .41016 .42392 L p .262 .832 0 r
F P
0 g
s
.27791 .58278 m .27465 .58887 L .31859 .59448 L p .543 0 .92 r
F P
0 g
s
.31859 .59448 m .32016 .58851 L .27791 .58278 L p .543 0 .92 r
F P
0 g
s
.26595 .61789 m .26468 .62689 L .31645 .63342 L p .623 0 .576 r
F P
0 g
s
.3173 .64044 m .31645 .63342 L .26468 .62689 L p .609 0 .426 r
F P
0 g
s
.43438 .54256 m .44443 .54404 L .47001 .54106 L p 0 0 0 r
F P
0 g
s
.47932 .53964 m .47001 .54106 L .44443 .54404 L p 0 .679 .729 r
F P
0 g
s
.66416 .69416 m .6626 .70001 L .66584 .64653 L p .828 .153 0 r
F P
0 g
s
.66584 .64653 m .6668 .6413 L .66416 .69416 L p .828 .153 0 r
F P
0 g
s
.42694 .49003 m .43613 .48552 L .41185 .48514 L p 0 .883 .904 r
F P
0 g
s
.43305 .486 m .43243 .48585 L p .44624 .48489 L 0 0 0 r
F P
0 g
s
.44624 .48489 m .43613 .48552 L .45462 .48431 L p 0 0 0 r
F P
0 g
s
.42254 .48351 m .41185 .48514 L .43613 .48552 L p 0 .635 .795 r
F P
0 g
s
.43613 .48552 m .44624 .48489 L .42254 .48351 L p 0 .635 .795 r
F P
0 g
s
.43243 .48585 m .42254 .48351 L .44624 .48489 L p 0 0 0 r
F P
0 g
s
.45462 .48431 m .46631 .48354 L p .44624 .48489 L 0 0 0 r
F P
0 g
s
.46631 .48354 m .47574 .48397 L .44624 .48489 L p 0 0 0 r
F P
0 g
s
.51037 .53015 m .51055 .53206 L .46262 .55584 L p .045 0 0 r
F P
0 g
s
.62734 .63027 m .63783 .62913 L .63014 .5765 L p .747 .103 0 r
F P
0 g
s
.63783 .62913 m .6483 .62986 L .64068 .57554 L p .662 .094 0 r
F P
0 g
s
.64068 .57554 m .63014 .5765 L .63783 .62913 L p .662 .094 0 r
F P
0 g
s
.6182 .63314 m .62734 .63027 L .61981 .57902 L p .831 .112 0 r
F P
0 g
s
.63014 .5765 m .61981 .57902 L .62734 .63027 L p .747 .103 0 r
F P
0 g
s
.6483 .62986 m .65737 .63238 L .65006 .57628 L p .556 .081 0 r
F P
0 g
s
.65006 .57628 m .64068 .57554 L .6483 .62986 L p .556 .081 0 r
F P
0 g
s
.44443 .54404 m .45428 .5415 L .47932 .53964 L p 0 .679 .729 r
F P
0 g
s
.48746 .53449 m .47932 .53964 L .45428 .5415 L p 0 .964 .891 r
F P
0 g
s
.22984 .58221 m .22453 .59154 L .26826 .60766 L p .664 0 .738 r
F P
0 g
s
.26826 .60766 m .27129 .59755 L .22984 .58221 L p .664 0 .738 r
F P
0 g
s
.26595 .61789 m .26826 .60766 L .22453 .59154 L p .686 0 .663 r
F P
0 g
s
.45592 .48825 m .44624 .48489 L .47574 .48397 L p 0 .028 0 r
F P
0 g
s
.47574 .48397 m .48449 .48821 L .45592 .48825 L p 0 .028 0 r
F P
0 g
s
.3173 .64044 m .3186 .64411 L .37384 .64075 L p .453 0 .193 r
F P
0 g
s
.37495 .64122 m .37384 .64075 L .3186 .64411 L p .064 0 0 r
F P
0 g
s
.37384 .64075 m .37495 .64122 L .43073 .62885 L p .134 0 0 r
F P
0 g
s
.23537 .5742 m .22984 .58221 L .27129 .59755 L p .631 0 .811 r
F P
0 g
s
.27129 .59755 m .27465 .58887 L .23537 .5742 L p .631 0 .811 r
F P
0 g
s
.43048 .62649 m .43073 .62885 L .37495 .64122 L p .404 0 .975 r
F P
0 g
s
.43073 .62885 m .43048 .62649 L .48547 .60682 L p .319 0 .872 r
F P
0 g
s
.41016 .42392 m .402 .42446 L .36837 .43433 L p .262 .832 0 r
F P
0 g
s
.36045 .43778 m .36837 .43433 L .402 .42446 L p 0 0 0 r
F P
0 g
s
.41556 .57652 m .4142 .5735 L .46262 .55584 L p .297 0 .223 r
F P
0 g
s
.46262 .55584 m .46426 .55986 L .41556 .57652 L p .297 0 .223 r
F P
0 g
s
.46426 .55986 m .46262 .55584 L .51055 .53206 L p .283 0 .36 r
F P
0 g
s
.36808 .58601 m .36676 .58384 L .4142 .5735 L p .306 0 .048 r
F P
0 g
s
.61162 .63737 m .6182 .63314 L .61105 .58278 L p .928 .121 0 r
F P
0 g
s
.61981 .57902 m .61105 .58278 L .6182 .63314 L p .831 .112 0 r
F P
0 g
s
.61114 .69854 m .60713 .69251 L .60929 .6476 L p .176 .043 0 r
F P
0 g
s
.6085 .64242 m .60929 .6476 L .60713 .69251 L p 0 0 0 r
F P
0 g
s
.22453 .59154 m .22015 .60097 L .26595 .61789 L p .686 0 .663 r
F P
0 g
s
.26468 .62689 m .26595 .61789 L .22015 .60097 L p .701 0 .565 r
F P
0 g
s
.53651 .57698 m .53913 .5799 L .48547 .60682 L p .438 0 .942 r
F P
0 g
s
.53913 .5799 m .53651 .57698 L .5874 .54582 L p .384 0 .915 r
F P
0 g
s
.5874 .54582 m .59018 .54704 L .53913 .5799 L p .384 0 .915 r
F P
0 g
s
.26468 .62689 m .26462 .63347 L .3173 .64044 L p .609 0 .426 r
F P
0 g
s
.3186 .64411 m .3173 .64044 L .26462 .63347 L p .519 0 .075 r
F P
0 g
s
.45428 .5415 m .46261 .53528 L .48746 .53449 L p 0 .964 .891 r
F P
0 g
s
.49334 .52629 m .48746 .53449 L .46261 .53528 L p 0 .982 .832 r
F P
0 g
s
.46385 .49516 m .45592 .48825 L .48449 .48821 L p 0 .394 .16 r
F P
0 g
s
.48449 .48821 m .49137 .49573 L .46385 .49516 L p 0 .394 .16 r
F P
0 g
s
.65737 .63238 m .66383 .63637 L .65704 .57863 L p .382 .061 0 r
F P
0 g
s
.65704 .57863 m .65006 .57628 L .65737 .63238 L p .382 .061 0 r
F P
0 g
s
.6626 .70001 m .65732 .70479 L .66103 .65134 L p .973 .187 0 r
F P
0 g
s
.66103 .65134 m .66584 .64653 L .6626 .70001 L p .973 .187 0 r
F P
0 g
s
.4833 .60232 m .48547 .60682 L .43048 .62649 L p .479 0 .931 r
F P
0 g
s
.48547 .60682 m .4833 .60232 L .53651 .57698 L p .438 0 .942 r
F P
0 g
s
.24042 .56855 m .23537 .5742 L .27465 .58887 L p .564 0 .896 r
F P
0 g
s
.27465 .58887 m .27791 .58278 L .24042 .56855 L p .564 0 .896 r
F P
0 g
s
.55605 .50026 m .55659 .50358 L .51055 .53206 L p .116 0 .101 r
F P
0 g
s
.65621 .43607 m .66419 .44285 L .69886 .40072 L p .226 0 .811 r
F P
0 g
s
.66419 .44285 m .67138 .45062 L .70688 .40793 L p .206 0 .757 r
F P
0 g
s
.70688 .40793 m .69886 .40072 L .66419 .44285 L p .206 0 .757 r
F P
0 g
s
.32328 .58764 m .32179 .58611 L .36676 .58384 L p .312 0 0 r
F P
0 g
s
.51055 .53206 m .51271 .53712 L .46426 .55986 L p .283 0 .36 r
F P
0 g
s
.51271 .53712 m .51055 .53206 L .55659 .50358 L p .262 0 .462 r
F P
0 g
s
.46261 .53528 m .46827 .52622 L .49334 .52629 L p 0 .982 .832 r
F P
0 g
s
.49613 .51617 m .49334 .52629 L .46827 .52622 L p 0 .901 .71 r
F P
0 g
s
.67138 .45062 m .67684 .45838 L .71303 .41575 L p .183 0 .691 r
F P
0 g
s
.71303 .41575 m .70688 .40793 L .67138 .45062 L p .183 0 .691 r
F P
0 g
s
.46893 .50468 m .46385 .49516 L .49137 .49573 L p 0 .623 .393 r
F P
0 g
s
.49137 .49573 m .49544 .50549 L .46893 .50468 L p 0 .623 .393 r
F P
0 g
s
.64849 .43118 m .65621 .43607 L .69004 .39507 L p .246 0 .861 r
F P
0 g
s
.69886 .40072 m .69004 .39507 L .65621 .43607 L p .226 0 .811 r
F P
0 g
s
.402 .42446 m .39481 .42887 L .36045 .43778 L p 0 0 0 r
F P
0 g
s
.35456 .44457 m .36045 .43778 L .39481 .42887 L p .126 0 0 r
F P
0 g
s
.40481 .48107 m .41308 .48234 L .37879 .49067 L p .271 .784 0 r
F P
0 g
s
.37879 .49067 m .36981 .48856 L .40481 .48107 L p .271 .784 0 r
F P
0 g
s
.22015 .60097 m .21728 .60925 L .26468 .62689 L p .701 0 .565 r
F P
0 g
s
.26462 .63347 m .26468 .62689 L .21728 .60925 L p .701 0 .396 r
F P
0 g
s
.46827 .52622 m .47048 .51553 L .49613 .51617 L p 0 .901 .71 r
F P
0 g
s
.49544 .50549 m .49613 .51617 L .47048 .51553 L p 0 .781 .566 r
F P
0 g
s
.47048 .51553 m .46893 .50468 L .49544 .50549 L p 0 .781 .566 r
F P
0 g
s
.61105 .58278 m .60503 .58729 L .61162 .63737 L p .928 .121 0 r
F P
0 g
s
.61848 .70369 m .61114 .69854 L .61394 .65223 L p .449 .094 0 r
F P
0 g
s
.60929 .6476 m .61394 .65223 L .61114 .69854 L p .176 .043 0 r
F P
0 g
s
.3186 .64411 m .32018 .64392 L .37495 .64122 L p .064 0 0 r
F P
0 g
s
.37547 .63781 m .37495 .64122 L .32018 .64392 L p .449 0 .994 r
F P
0 g
s
.37495 .64122 m .37547 .63781 L .43048 .62649 L p .404 0 .975 r
F P
0 g
s
.42912 .62065 m .43048 .62649 L .37547 .63781 L p .515 0 .911 r
F P
0 g
s
.43048 .62649 m .42912 .62065 L .4833 .60232 L p .479 0 .931 r
F P
0 g
s
.41785 .5829 m .41556 .57652 L .46426 .55986 L p .395 0 .487 r
F P
0 g
s
.46426 .55986 m .46728 .56685 L .41785 .5829 L p .395 0 .487 r
F P
0 g
s
.46728 .56685 m .46426 .55986 L .51271 .53712 L p .353 0 .549 r
F P
0 g
s
.36975 .59182 m .36808 .58601 L .41556 .57652 L p .444 0 .418 r
F P
0 g
s
.41556 .57652 m .41785 .5829 L .36975 .59182 L p .444 0 .418 r
F P
0 g
s
.53231 .57105 m .53651 .57698 L .4833 .60232 L p .443 0 .862 r
F P
0 g
s
.53651 .57698 m .53231 .57105 L .58269 .54153 L p .405 0 .887 r
F P
0 g
s
.58269 .54153 m .5874 .54582 L .53651 .57698 L p .405 0 .887 r
F P
0 g
s
.67684 .45838 m .67984 .4651 L .71651 .42315 L p .146 0 .583 r
F P
0 g
s
.71651 .42315 m .71303 .41575 L .67684 .45838 L p .146 0 .583 r
F P
0 g
s
.66383 .63637 m .6668 .6413 L .66065 .58229 L p 0 .012 0 r
F P
0 g
s
.66065 .58229 m .65704 .57863 L .66383 .63637 L p 0 .012 0 r
F P
0 g
s
.65732 .70479 m .64902 .70782 L .653 .65507 L p .869 .17 0 r
F P
0 g
s
.653 .65507 m .66103 .65134 L .65732 .70479 L p .869 .17 0 r
F P
0 g
s
.24429 .56604 m .24042 .56855 L .27791 .58278 L p .367 0 .98 r
F P
0 g
s
.64205 .42883 m .64849 .43118 L .68157 .39172 L p .268 0 .905 r
F P
0 g
s
.69004 .39507 m .68157 .39172 L .64849 .43118 L p .246 0 .861 r
F P
0 g
s
.28243 .58124 m .28063 .5801 L .32179 .58611 L p .316 0 0 r
F P
0 g
s
.26462 .63347 m .26579 .63672 L .3186 .64411 L p .519 0 .075 r
F P
0 g
s
.32018 .64392 m .3186 .64411 L .26579 .63672 L p .002 0 .788 r
F P
0 g
s
.20195 .55098 m .19442 .55772 L .22984 .58221 L p .667 0 .798 r
F P
0 g
s
.22984 .58221 m .23537 .5742 L .20195 .55098 L p .667 0 .798 r
F P
0 g
s
.22453 .59154 m .22984 .58221 L .19442 .55772 L p .716 0 .735 r
F P
0 g
s
.19442 .55772 m .18702 .56567 L .22453 .59154 L p .716 0 .735 r
F P
0 g
s
.22015 .60097 m .22453 .59154 L .18702 .56567 L p .753 0 .665 r
F P
0 g
s
.39481 .42887 m .38959 .43659 L .35456 .44457 L p .126 0 0 r
F P
0 g
s
.35153 .4538 m .35456 .44457 L .38959 .43659 L p .213 .123 0 r
F P
0 g
s
.39714 .47597 m .40481 .48107 L .36981 .48856 L p .298 .658 0 r
F P
0 g
s
.36981 .48856 m .36161 .48287 L .39714 .47597 L p .298 .658 0 r
F P
0 g
s
.47988 .59495 m .4833 .60232 L .42912 .62065 L p .483 0 .837 r
F P
0 g
s
.4833 .60232 m .47988 .59495 L .53231 .57105 L p .443 0 .862 r
F P
0 g
s
.51271 .53712 m .51656 .54468 L .46728 .56685 L p .353 0 .549 r
F P
0 g
s
.55659 .50358 m .55942 .50962 L .51271 .53712 L p .262 0 .462 r
F P
0 g
s
.51656 .54468 m .51271 .53712 L .55942 .50962 L p .314 0 .603 r
F P
0 g
s
.32441 .59294 m .32328 .58764 L .36808 .58601 L p .505 0 .346 r
F P
0 g
s
.36808 .58601 m .36975 .59182 L .32441 .59294 L p .505 0 .346 r
F P
0 g
s
.62815 .70723 m .61848 .70369 L .62184 .65565 L p .589 .12 0 r
F P
0 g
s
.61394 .65223 m .62184 .65565 L .61848 .70369 L p .449 .094 0 r
F P
0 g
s
.20862 .54634 m .20195 .55098 L .23537 .5742 L p .585 0 .868 r
F P
0 g
s
.23537 .5742 m .24042 .56855 L .20862 .54634 L p .585 0 .868 r
F P
0 g
s
.18702 .56567 m .18072 .57378 L .22015 .60097 L p .753 0 .665 r
F P
0 g
s
.21728 .60925 m .22015 .60097 L .18072 .57378 L p .785 0 .568 r
F P
0 g
s
.21728 .60925 m .2163 .61528 L .26462 .63347 L p .701 0 .396 r
F P
0 g
s
.26579 .63672 m .26462 .63347 L .2163 .61528 L p .598 0 0 r
F P
0 g
s
.38959 .43659 m .38707 .44657 L .35153 .4538 L p .213 .123 0 r
F P
0 g
s
.35178 .46421 m .35153 .4538 L .38707 .44657 L p .254 .294 0 r
F P
0 g
s
.64902 .70782 m .63882 .70868 L .64283 .65718 L p .774 .154 0 r
F P
0 g
s
.64283 .65718 m .653 .65507 L .64902 .70782 L p .774 .154 0 r
F P
0 g
s
.39112 .46773 m .39714 .47597 L .36161 .48287 L p .293 .536 0 r
F P
0 g
s
.36161 .48287 m .35531 .47437 L .39112 .46773 L p .293 .536 0 r
F P
0 g
s
.42077 .59182 m .41785 .5829 L .46728 .56685 L p .433 0 .609 r
F P
0 g
s
.46728 .56685 m .47127 .57589 L .42077 .59182 L p .433 0 .609 r
F P
0 g
s
.47127 .57589 m .46728 .56685 L .51656 .54468 L p .387 0 .648 r
F P
0 g
s
.61981 .57902 m .63014 .5765 L .61404 .52023 L p .763 .061 0 r
F P
0 g
s
.63014 .5765 m .64068 .57554 L .62446 .51765 L p .683 .057 0 r
F P
0 g
s
.62446 .51765 m .61404 .52023 L .63014 .5765 L p .683 .057 0 r
F P
0 g
s
.37547 .63781 m .37534 .63094 L .42912 .62065 L p .515 0 .911 r
F P
0 g
s
.42682 .6121 m .42912 .62065 L .37534 .63094 L p .523 0 .816 r
F P
0 g
s
.42912 .62065 m .42682 .6121 L .47988 .59495 L p .483 0 .837 r
F P
0 g
s
.5271 .56291 m .53231 .57105 L .47988 .59495 L p .427 0 .786 r
F P
0 g
s
.53231 .57105 m .5271 .56291 L .57669 .53473 L p .386 0 .814 r
F P
0 g
s
.57669 .53473 m .58269 .54153 L .53231 .57105 L p .386 0 .814 r
F P
0 g
s
.37155 .60052 m .36975 .59182 L .41785 .5829 L p .486 0 .57 r
F P
0 g
s
.41785 .5829 m .42077 .59182 L .37155 .60052 L p .486 0 .57 r
F P
0 g
s
.67984 .4651 m .67996 .46987 L .71682 .42915 L p .066 0 .336 r
F P
0 g
s
.71682 .42915 m .71651 .42315 L .67984 .4651 L p .066 0 .336 r
F P
0 g
s
.61105 .58278 m .61981 .57902 L .60409 .52401 L p .843 .065 0 r
F P
0 g
s
.61404 .52023 m .60409 .52401 L .61981 .57902 L p .763 .061 0 r
F P
0 g
s
.32018 .64392 m .32183 .63982 L .37547 .63781 L p .449 0 .994 r
F P
0 g
s
.37534 .63094 m .37547 .63781 L .32183 .63982 L p .547 0 .892 r
F P
0 g
s
.47567 .58571 m .47988 .59495 L .42682 .6121 L p .471 0 .76 r
F P
0 g
s
.47988 .59495 m .47567 .58571 L .5271 .56291 L p .427 0 .786 r
F P
0 g
s
.38707 .44657 m .3876 .45745 L .35178 .46421 L p .254 .294 0 r
F P
0 g
s
.35531 .47437 m .35178 .46421 L .3876 .45745 L p .278 .42 0 r
F P
0 g
s
.63882 .70868 m .62815 .70723 L .63194 .65739 L p .687 .138 0 r
F P
0 g
s
.62184 .65565 m .63194 .65739 L .62815 .70723 L p .589 .12 0 r
F P
0 g
s
.3876 .45745 m .39112 .46773 L .35531 .47437 L p .278 .42 0 r
F P
0 g
s
.63194 .65739 m .64283 .65718 L .63882 .70868 L p .687 .138 0 r
F P
0 g
s
.64068 .57554 m .65006 .57628 L .63399 .51662 L p .585 .051 0 r
F P
0 g
s
.63399 .51662 m .62446 .51765 L .64068 .57554 L p .585 .051 0 r
F P
0 g
s
.51656 .54468 m .52159 .55369 L .47127 .57589 L p .387 0 .648 r
F P
0 g
s
.52159 .55369 m .51656 .54468 L .56418 .51758 L p .344 0 .685 r
F P
0 g
s
.55942 .50962 m .56418 .51758 L .51656 .54468 L p .314 0 .603 r
F P
0 g
s
.6668 .6413 m .66584 .64653 L .66038 .58678 L p .837 .097 0 r
F P
0 g
s
.66038 .58678 m .66065 .58229 L .6668 .6413 L p .837 .097 0 r
F P
0 g
s
.87523 .53136 m .87152 .52124 L .86489 .52124 L p 0 .6 .014 r
F P
0 g
s
.87152 .52124 m .87174 .51049 L .86546 .51032 L p 0 .817 .036 r
F P
0 g
s
.86546 .51032 m .86489 .52124 L .87152 .52124 L p 0 .817 .036 r
F P
0 g
s
.24042 .56855 m .24429 .56604 L .21355 .54442 L p .384 0 .937 r
F P
0 g
s
.24648 .56704 m .24665 .57146 L .23373 .56227 L p .785 0 .161 r
F P
0 g
s
.24648 .56704 m .24429 .56604 L .25888 .57169 L p .308 0 0 r
F P
0 g
s
.21355 .54442 m .20862 .54634 L .24042 .56855 L p .384 0 .937 r
F P
0 g
s
.21605 .54551 m .21355 .54442 L .22671 .55368 L p 0 0 .674 r
F P
0 g
s
.22671 .55368 m .24429 .56604 L p .21605 .54551 L 0 0 .674 r
F P
0 g
s
.20862 .54634 m .21355 .54442 L .18909 .51592 L p .4 0 .883 r
F P
0 g
s
.18909 .51592 m .21355 .54442 L p .19201 .51757 L 0 0 .695 r
F P
0 g
s
.19177 .51727 m .19174 .51726 L p .19201 .51757 L 0 0 .695 r
F P
0 g
s
.19174 .51726 m .18909 .51592 L p .19201 .51757 L 0 0 .695 r
F P
0 g
s
.21355 .54442 m .21605 .54551 L .19177 .51727 L p 0 0 .695 r
F P
0 g
s
.21577 .5495 m .20508 .5371 L p .21594 .54542 L .893 0 .141 r
F P
0 g
s
.21605 .54551 m .21605 .54555 L p .21594 .54542 L .893 0 .141 r
F P
0 g
s
.21605 .54555 m .21577 .5495 L p .21594 .54542 L .893 0 .141 r
F P
0 g
s
.18909 .51592 m .18339 .51694 L .20862 .54634 L p .4 0 .883 r
F P
0 g
s
.20508 .5371 m .19104 .52083 L p .21605 .54551 L .893 0 .141 r
F P
0 g
s
.24429 .56604 m .24648 .56704 L .21605 .54551 L p 0 0 .674 r
F P
0 g
s
.23373 .56227 m .21577 .5495 L p .24648 .56704 L .785 0 .161 r
F P
0 g
s
.25888 .57169 m .28063 .5801 L p .24648 .56704 L .308 0 0 r
F P
0 g
s
.87174 .51049 m .87587 .50055 L .86983 .50034 L p 0 .947 .054 r
F P
0 g
s
.86983 .50034 m .86546 .51032 L .87174 .51049 L p 0 .947 .054 r
F P
0 g
s
.88238 .53949 m .87523 .53136 L .8682 .53164 L p 0 .313 0 r
F P
0 g
s
.86489 .52124 m .8682 .53164 L .87523 .53136 L p 0 .6 .014 r
F P
0 g
s
.28307 .58609 m .28243 .58124 L .32328 .58764 L p .582 0 .274 r
F P
0 g
s
.32328 .58764 m .32441 .59294 L .28307 .58609 L p .582 0 .274 r
F P
0 g
s
.26579 .63672 m .26804 .63617 L .32018 .64392 L p .002 0 .788 r
F P
0 g
s
.32183 .63982 m .32018 .64392 L .26804 .63617 L p .474 0 .98 r
F P
0 g
s
.4239 .60202 m .42077 .59182 L .47127 .57589 L p .455 0 .69 r
F P
0 g
s
.47127 .57589 m .47567 .58571 L .4239 .60202 L p .455 0 .69 r
F P
0 g
s
.47567 .58571 m .47127 .57589 L .52159 .55369 L p .409 0 .719 r
F P
0 g
s
.87587 .50055 m .88337 .49276 L .87743 .49262 L p 0 .97 .067 r
F P
0 g
s
.87743 .49262 m .86983 .50034 L .87587 .50055 L p 0 .97 .067 r
F P
0 g
s
.89202 .54455 m .88238 .53949 L .87496 .54013 L p 0 0 0 r
F P
0 g
s
.8682 .53164 m .87496 .54013 L .88238 .53949 L p 0 .313 0 r
F P
0 g
s
.42682 .6121 m .4239 .60202 L .47567 .58571 L p .471 0 .76 r
F P
0 g
s
.52159 .55369 m .5271 .56291 L .47567 .58571 L p .409 0 .719 r
F P
0 g
s
.18072 .57378 m .17634 .58101 L .21728 .60925 L p .785 0 .568 r
F P
0 g
s
.2163 .61528 m .21728 .60925 L .17634 .58101 L p .806 0 .382 r
F P
0 g
s
.47007 .45759 m .47064 .44671 L .43177 .454 L p .277 .343 0 r
F P
0 g
s
.43177 .454 m .43081 .46446 L .47007 .45759 L p .277 .343 0 r
F P
0 g
s
.42975 .44342 m .43177 .454 L .47064 .44671 L p .284 .439 0 r
F P
0 g
s
.5271 .56291 m .52159 .55369 L .57022 .52636 L p .366 0 .75 r
F P
0 g
s
.57022 .52636 m .57669 .53473 L .5271 .56291 L p .366 0 .75 r
F P
0 g
s
.47064 .44671 m .46856 .43613 L .42975 .44342 L p .284 .439 0 r
F P
0 g
s
.42501 .43411 m .42975 .44342 L .46856 .43613 L p .286 .53 0 r
F P
0 g
s
.32503 .60133 m .32441 .59294 L .36975 .59182 L p .548 0 .535 r
F P
0 g
s
.36975 .59182 m .37155 .60052 L .32503 .60133 L p .548 0 .535 r
F P
0 g
s
.56418 .51758 m .57022 .52636 L .52159 .55369 L p .344 0 .685 r
F P
0 g
s
.88337 .49276 m .89321 .48817 L .88726 .48821 L p 0 .859 .071 r
F P
0 g
s
.88726 .48821 m .87743 .49262 L .88337 .49276 L p 0 .859 .071 r
F P
0 g
s
.60503 .58729 m .61105 .58278 L .59592 .52851 L p .937 .069 0 r
F P
0 g
s
.60409 .52401 m .59592 .52851 L .61105 .58278 L p .843 .065 0 r
F P
0 g
s
.37534 .63094 m .37455 .62154 L .42682 .6121 L p .523 0 .816 r
F P
0 g
s
.4239 .60202 m .42682 .6121 L .37455 .62154 L p .517 0 .738 r
F P
0 g
s
.89202 .54455 m .88427 .54557 L p .90171 .54581 L 0 0 0 r
F P
0 g
s
.90208 .54574 m .89202 .54455 L p .90171 .54581 L 0 0 0 r
F P
0 g
s
.88427 .54557 m .89489 .54722 L .90285 .54583 L p 0 0 0 r
F P
0 g
s
.9054 .54482 m .91342 .54316 L p .89613 .54694 L 0 .697 .063 r
F P
0 g
s
.89489 .54722 m .89568 .54704 L p .89613 .54694 L 0 .697 .063 r
F P
0 g
s
.89568 .54704 m .9054 .54482 L p .89613 .54694 L 0 .697 .063 r
F P
0 g
s
.90285 .54583 m .90208 .54574 L p .90171 .54581 L 0 0 0 r
F P
0 g
s
.87496 .54013 m .88427 .54557 L .89202 .54455 L p 0 0 0 r
F P
0 g
s
.37323 .61091 m .37155 .60052 L .42077 .59182 L p .506 0 .663 r
F P
0 g
s
.42077 .59182 m .4239 .60202 L .37323 .61091 L p .506 0 .663 r
F P
0 g
s
.46692 .46735 m .47007 .45759 L .43081 .46446 L p .261 .219 0 r
F P
0 g
s
.43081 .46446 m .427 .47341 L .46692 .46735 L p .261 .219 0 r
F P
0 g
s
.17547 .52016 m .16637 .52514 L .19442 .55772 L p .703 0 .787 r
F P
0 g
s
.19442 .55772 m .20195 .55098 L .17547 .52016 L p .703 0 .787 r
F P
0 g
s
.18702 .56567 m .19442 .55772 L .16637 .52514 L p .766 0 .736 r
F P
0 g
s
.65006 .57628 m .65704 .57863 L .64137 .51728 L p .426 .041 0 r
F P
0 g
s
.64137 .51728 m .63399 .51662 L .65006 .57628 L p .426 .041 0 r
F P
0 g
s
.89321 .48817 m .9041 .48741 L .89799 .48772 L p 0 .612 .065 r
F P
0 g
s
.89799 .48772 m .88726 .48821 L .89321 .48817 L p 0 .612 .065 r
F P
0 g
s
.37455 .62154 m .37323 .61091 L .4239 .60202 L p .517 0 .738 r
F P
0 g
s
.91342 .54316 m .90285 .54583 L .89489 .54722 L p 0 .697 .063 r
F P
0 g
s
.2163 .61528 m .21737 .61824 L .26579 .63672 L p .598 0 0 r
F P
0 g
s
.26804 .63617 m .26579 .63672 L .21737 .61824 L p .055 0 .89 r
F P
0 g
s
.61394 .65223 m .60929 .6476 L .60402 .59616 L p .207 .04 0 r
F P
0 g
s
.60256 .59196 m .60402 .59616 L .60929 .6476 L p 0 0 0 r
F P
0 g
s
.46856 .43613 m .46409 .42725 L .42501 .43411 L p .286 .53 0 r
F P
0 g
s
.41818 .4273 m .42501 .43411 L .46409 .42725 L p .278 .631 0 r
F P
0 g
s
.16637 .52514 m .15729 .53124 L .18702 .56567 L p .766 0 .736 r
F P
0 g
s
.18072 .57378 m .18702 .56567 L .15729 .53124 L p .819 0 .676 r
F P
0 g
s
.32183 .63982 m .32332 .63235 L .37534 .63094 L p .547 0 .892 r
F P
0 g
s
.37455 .62154 m .37534 .63094 L .32332 .63235 L p .565 0 .798 r
F P
0 g
s
.9041 .48741 m .91455 .4906 L .90819 .49124 L p 0 0 0 r
F P
0 g
s
.90819 .49124 m .89799 .48772 L .9041 .48741 L p 0 0 0 r
F P
0 g
s
.92229 .53689 m .91342 .54316 L .9054 .54482 L p 0 .896 .065 r
F P
0 g
s
.18339 .51694 m .17547 .52016 L .20195 .55098 L p .607 0 .839 r
F P
0 g
s
.20195 .55098 m .20862 .54634 L .18339 .51694 L p .607 0 .839 r
F P
0 g
s
.71389 .43293 m .71682 .42915 L .67996 .46987 L p 0 0 0 r
F P
0 g
s
.67996 .46987 m .67715 .47204 L .71389 .43293 L p 0 0 0 r
F P
0 g
s
.69004 .39507 m .69886 .40072 L .72748 .35993 L p .182 0 .841 r
F P
0 g
s
.69886 .40072 m .70688 .40793 L .73628 .36645 L p .164 0 .791 r
F P
0 g
s
.73628 .36645 m .72748 .35993 L .69886 .40072 L p .164 0 .791 r
F P
0 g
s
.32504 .61166 m .32503 .60133 L .37155 .60052 L p .562 0 .64 r
F P
0 g
s
.37155 .60052 m .37323 .61091 L .32504 .61166 L p .562 0 .64 r
F P
0 g
s
.46159 .4747 m .46692 .46735 L .427 .47341 L p .221 .015 0 r
F P
0 g
s
.427 .47341 m .42082 .47966 L .46159 .4747 L p .221 .015 0 r
F P
0 g
s
.91455 .4906 m .92314 .4973 L .91647 .4983 L p 0 .096 0 r
F P
0 g
s
.66584 .64653 m .66103 .65134 L .6562 .59151 L p .983 .126 0 r
F P
0 g
s
.6562 .59151 m .66038 .58678 L .66584 .64653 L p .983 .126 0 r
F P
0 g
s
.92825 .52786 m .92229 .53689 L .91436 .53869 L p 0 .952 .057 r
F P
0 g
s
.9054 .54482 m .91436 .53869 L .92229 .53689 L p 0 .896 .065 r
F P
0 g
s
.70688 .40793 m .71303 .41575 L .74304 .37423 L p .142 0 .727 r
F P
0 g
s
.74304 .37423 m .73628 .36645 L .70688 .40793 L p .142 0 .727 r
F P
0 g
s
.24665 .57146 m .24648 .56704 L .28243 .58124 L p .677 0 .21 r
F P
0 g
s
.28243 .58124 m .28307 .58609 L .24665 .57146 L p .677 0 .21 r
F P
0 g
s
.28244 .59402 m .28307 .58609 L .32441 .59294 L p .621 0 .506 r
F P
0 g
s
.32441 .59294 m .32503 .60133 L .28244 .59402 L p .621 0 .506 r
F P
0 g
s
.26804 .63617 m .27109 .63184 L .32183 .63982 L p .474 0 .98 r
F P
0 g
s
.32332 .63235 m .32183 .63982 L .27109 .63184 L p .576 0 .873 r
F P
0 g
s
.15729 .53124 m .1494 .53766 L .18072 .57378 L p .819 0 .676 r
F P
0 g
s
.17634 .58101 m .18072 .57378 L .1494 .53766 L p .871 0 .584 r
F P
0 g
s
.59592 .52851 m .59061 .53314 L .60503 .58729 L p .937 .069 0 r
F P
0 g
s
.92314 .4973 m .92871 .50662 L .92168 .50796 L p 0 .429 0 r
F P
0 g
s
.32332 .63235 m .32445 .62251 L .37455 .62154 L p .565 0 .798 r
F P
0 g
s
.37323 .61091 m .37455 .62154 L .32445 .62251 L p .567 0 .72 r
F P
0 g
s
.93049 .51729 m .92825 .52786 L .92055 .52964 L p 0 .877 .041 r
F P
0 g
s
.91436 .53869 m .92055 .52964 L .92825 .52786 L p 0 .952 .057 r
F P
0 g
s
.46409 .42725 m .45781 .42127 L .41818 .4273 L p .278 .631 0 r
F P
0 g
s
.41016 .42392 m .41818 .4273 L .45781 .42127 L p .237 .742 0 r
F P
0 g
s
.68157 .39172 m .69004 .39507 L .71781 .35553 L p .201 0 .885 r
F P
0 g
s
.72748 .35993 m .71781 .35553 L .69004 .39507 L p .182 0 .841 r
F P
0 g
s
.17634 .58101 m .1745 .58637 L .2163 .61528 L p .806 0 .382 r
F P
0 g
s
.21737 .61824 m .2163 .61528 L .1745 .58637 L p .679 0 0 r
F P
0 g
s
.92871 .50662 m .93049 .51729 L .92311 .51891 L p 0 .696 .02 r
F P
0 g
s
.92055 .52964 m .92311 .51891 L .93049 .51729 L p 0 .877 .041 r
F P
0 g
s
.32445 .62251 m .32504 .61166 L .37323 .61091 L p .567 0 .72 r
F P
0 g
s
.65704 .57863 m .66065 .58229 L .64559 .51956 L p .064 .017 0 r
F P
0 g
s
.64559 .51956 m .64137 .51728 L .65704 .57863 L p .064 .017 0 r
F P
0 g
s
.71303 .41575 m .71651 .42315 L .74687 .38224 L p .11 0 .624 r
F P
0 g
s
.74687 .38224 m .74304 .37423 L .71303 .41575 L p .11 0 .624 r
F P
0 g
s
.62184 .65565 m .61394 .65223 L .60926 .59932 L p .476 .072 0 r
F P
0 g
s
.60402 .59616 m .60926 .59932 L .61394 .65223 L p .207 .04 0 r
F P
0 g
s
.21737 .61824 m .22037 .61768 L .26804 .63617 L p .055 0 .89 r
F P
0 g
s
.27109 .63184 m .26804 .63617 L .22037 .61768 L p .49 0 .951 r
F P
0 g
s
.67457 .39113 m .68157 .39172 L .70854 .35382 L p .222 0 .908 r
F P
0 g
s
.71781 .35553 m .70854 .35382 L .68157 .39172 L p .201 0 .885 r
F P
0 g
s
.28062 .60396 m .28244 .59402 L .32503 .60133 L p .626 0 .624 r
F P
0 g
s
.32503 .60133 m .32504 .61166 L .28062 .60396 L p .626 0 .624 r
F P
0 g
s
.27109 .63184 m .27451 .62429 L .32332 .63235 L p .576 0 .873 r
F P
0 g
s
.32445 .62251 m .32332 .63235 L .27451 .62429 L p .608 0 .785 r
F P
0 g
s
.60409 .52401 m .61404 .52023 L .58993 .46299 L p .777 .02 0 r
F P
0 g
s
.61404 .52023 m .62446 .51765 L .60002 .45889 L p .702 .019 0 r
F P
0 g
s
.60002 .45889 m .58993 .46299 L .61404 .52023 L p .702 .019 0 r
F P
0 g
s
.67715 .47204 m .67176 .47126 L .70808 .43395 L p .274 0 .893 r
F P
0 g
s
.70808 .43395 m .71389 .43293 L .67715 .47204 L p .274 0 .893 r
F P
0 g
s
.66103 .65134 m .653 .65507 L .64866 .59582 L p .881 .117 0 r
F P
0 g
s
.64866 .59582 m .6562 .59151 L .66103 .65134 L p .881 .117 0 r
F P
0 g
s
.45781 .42127 m .45054 .419 L .41016 .42392 L p .237 .742 0 r
F P
0 g
s
.402 .42446 m .41016 .42392 L .45054 .419 L p .075 .678 0 r
F P
0 g
s
.63766 .43854 m .63621 .43274 L .66847 .39827 L p .112 0 .478 r
F P
0 g
s
.66999 .39341 m .66847 .39827 L .63621 .43274 L p 0 0 .014 r
F P
0 g
s
.1494 .53766 m .14376 .54355 L .17634 .58101 L p .871 0 .584 r
F P
0 g
s
.1745 .58637 m .17634 .58101 L .14376 .54355 L p .916 0 .391 r
F P
0 g
s
.15695 .48313 m .1468 .48609 L .16637 .52514 L p .697 0 .73 r
F P
0 g
s
.16637 .52514 m .17547 .52016 L .15695 .48313 L p .697 0 .73 r
F P
0 g
s
.15729 .53124 m .16637 .52514 L .1468 .48609 L p .768 0 .695 r
F P
0 g
s
.58993 .46299 m .58056 .46788 L .60409 .52401 L p .777 .02 0 r
F P
0 g
s
.62446 .51765 m .63399 .51662 L .6095 .45613 L p .613 .018 0 r
F P
0 g
s
.24476 .57872 m .24665 .57146 L .28307 .58609 L p .707 0 .488 r
F P
0 g
s
.28307 .58609 m .28244 .59402 L .24476 .57872 L p .707 0 .488 r
F P
0 g
s
.1468 .48609 m .13654 .49007 L .15729 .53124 L p .768 0 .695 r
F P
0 g
s
.1494 .53766 m .15729 .53124 L .13654 .49007 L p .831 0 .651 r
F P
0 g
s
.16563 .48157 m .15695 .48313 L .17547 .52016 L p .596 0 .76 r
F P
0 g
s
.17547 .52016 m .18339 .51694 L .16563 .48157 L p .596 0 .76 r
F P
0 g
s
.21577 .5495 m .21605 .54551 L .24648 .56704 L p .785 0 .161 r
F P
0 g
s
.27784 .61454 m .28062 .60396 L .32504 .61166 L p .621 0 .708 r
F P
0 g
s
.32504 .61166 m .32445 .62251 L .27784 .61454 L p .621 0 .708 r
F P
0 g
s
.27451 .62429 m .27784 .61454 L .32445 .62251 L p .608 0 .785 r
F P
0 g
s
.1745 .58637 m .17546 .58913 L .21737 .61824 L p .679 0 0 r
F P
0 g
s
.22037 .61768 m .21737 .61824 L .17546 .58913 L p .097 0 .901 r
F P
0 g
s
.71651 .42315 m .71682 .42915 L .74723 .38943 L p .045 0 .394 r
F P
0 g
s
.74723 .38943 m .74687 .38224 L .71651 .42315 L p .045 0 .394 r
F P
0 g
s
.63194 .65739 m .62184 .65565 L .6176 .60099 L p .611 .088 0 r
F P
0 g
s
.60926 .59932 m .6176 .60099 L .62184 .65565 L p .476 .072 0 r
F P
0 g
s
.66065 .58229 m .66038 .58678 L .64606 .52318 L p .765 .045 0 r
F P
0 g
s
.64606 .52318 m .64559 .51956 L .66065 .58229 L p .765 .045 0 r
F P
0 g
s
.653 .65507 m .64283 .65718 L .63877 .59911 L p .789 .108 0 r
F P
0 g
s
.63877 .59911 m .64866 .59582 L .653 .65507 L p .789 .108 0 r
F P
0 g
s
.22037 .61768 m .22491 .61363 L .27109 .63184 L p .49 0 .951 r
F P
0 g
s
.27451 .62429 m .27109 .63184 L .22491 .61363 L p .604 0 .855 r
F P
0 g
s
.39205 .50992 m .3925 .52109 L .41564 .51866 L p 0 .777 .559 r
F P
0 g
s
.41564 .51866 m .41597 .50779 L .39205 .50992 L p 0 .777 .559 r
F P
0 g
s
.41896 .52895 m .41564 .51866 L .3925 .52109 L p 0 .621 .367 r
F P
0 g
s
.39551 .4993 m .39205 .50992 L .41597 .50779 L p 0 .883 .718 r
F P
0 g
s
.41597 .50779 m .41991 .4978 L .39551 .4993 L p 0 .883 .718 r
F P
0 g
s
.63399 .51662 m .64137 .51728 L .61712 .45507 L p .475 .015 0 r
F P
0 g
s
.61712 .45507 m .6095 .45613 L .63399 .51662 L p .475 .015 0 r
F P
0 g
s
.45054 .419 m .44326 .42079 L .402 .42446 L p .075 .678 0 r
F P
0 g
s
.39481 .42887 m .402 .42446 L .44326 .42079 L p .144 0 0 r
F P
0 g
s
.3925 .52109 m .39679 .53131 L .41896 .52895 L p 0 .621 .367 r
F P
0 g
s
.4255 .53729 m .41896 .52895 L .39679 .53131 L p 0 .398 .124 r
F P
0 g
s
.13654 .49007 m .12749 .49456 L .1494 .53766 L p .831 0 .651 r
F P
0 g
s
.14376 .54355 m .1494 .53766 L .12749 .49456 L p .897 0 .577 r
F P
0 g
s
.64283 .65718 m .63194 .65739 L .62792 .60092 L p .705 .098 0 r
F P
0 g
s
.6176 .60099 m .62792 .60092 L .63194 .65739 L p .611 .088 0 r
F P
0 g
s
.17172 .48163 m .16563 .48157 L .18339 .51694 L p .394 0 .776 r
F P
0 g
s
.18339 .51694 m .18909 .51592 L .17172 .48163 L p .394 0 .776 r
F P
0 g
s
.40242 .49065 m .39551 .4993 L .41991 .4978 L p 0 .935 .846 r
F P
0 g
s
.41991 .4978 m .42694 .49003 L .40242 .49065 L p 0 .935 .846 r
F P
0 g
s
.67176 .47126 m .66449 .46763 L .70016 .43203 L p .258 0 .887 r
F P
0 g
s
.70016 .43203 m .70808 .43395 L .67176 .47126 L p .258 0 .887 r
F P
0 g
s
.62792 .60092 m .63877 .59911 L .64283 .65718 L p .705 .098 0 r
F P
0 g
s
.64193 .44598 m .63766 .43854 L .67026 .40509 L p .166 0 .641 r
F P
0 g
s
.66847 .39827 m .67026 .40509 L .63766 .43854 L p .112 0 .478 r
F P
0 g
s
.14376 .54355 m .14113 .54814 L .1745 .58637 L p .916 0 .391 r
F P
0 g
s
.17546 .58913 m .1745 .58637 L .14113 .54814 L p .741 0 0 r
F P
0 g
s
.24105 .58786 m .24476 .57872 L .28244 .59402 L p .697 0 .616 r
F P
0 g
s
.28244 .59402 m .28062 .60396 L .24105 .58786 L p .697 0 .616 r
F P
0 g
s
.39679 .53131 m .40437 .53921 L .4255 .53729 L p 0 .398 .124 r
F P
0 g
s
.43438 .54256 m .4255 .53729 L .40437 .53921 L p 0 .07 0 r
F P
0 g
s
.41185 .48514 m .40242 .49065 L .42694 .49003 L p 0 .883 .904 r
F P
0 g
s
.60926 .59932 m .60402 .59616 L .59105 .54036 L p .27 .031 0 r
F P
0 g
s
.58889 .53727 m .59105 .54036 L .60402 .59616 L p 0 0 0 r
F P
0 g
s
.71682 .42915 m .71389 .43293 L .74404 .39481 L p 0 0 0 r
F P
0 g
s
.74404 .39481 m .74723 .38943 L .71682 .42915 L p 0 0 0 r
F P
0 g
s
.44039 .47454 m .44737 .47859 L .40481 .48107 L p .246 .682 0 r
F P
0 g
s
.40481 .48107 m .39714 .47597 L .44039 .47454 L p .246 .682 0 r
F P
0 g
s
.22491 .61363 m .23038 .60662 L .27451 .62429 L p .604 0 .855 r
F P
0 g
s
.27784 .61454 m .27451 .62429 L .23038 .60662 L p .651 0 .775 r
F P
0 g
s
.21271 .55587 m .21577 .5495 L .24665 .57146 L p .803 0 .485 r
F P
0 g
s
.24665 .57146 m .24476 .57872 L .21271 .55587 L p .803 0 .485 r
F P
0 g
s
.40437 .53921 m .41421 .54372 L .43438 .54256 L p 0 .07 0 r
F P
0 g
s
.44443 .54404 m .43438 .54256 L .41421 .54372 L p 0 0 0 r
F P
0 g
s
.17546 .58913 m .17913 .58888 L .22037 .61768 L p .097 0 .901 r
F P
0 g
s
.22491 .61363 m .22037 .61768 L .17913 .58888 L p .504 0 .912 r
F P
0 g
s
.14714 .44183 m .13649 .44267 L .1468 .48609 L p .691 0 .675 r
F P
0 g
s
.1468 .48609 m .15695 .48313 L .14714 .44183 L p .691 0 .675 r
F P
0 g
s
.13654 .49007 m .1468 .48609 L .13649 .44267 L p .766 0 .656 r
F P
0 g
s
.23602 .59762 m .24105 .58786 L .28062 .60396 L p .678 0 .701 r
F P
0 g
s
.28062 .60396 m .27784 .61454 L .23602 .59762 L p .678 0 .701 r
F P
0 g
s
.44326 .42079 m .43696 .42643 L .39481 .42887 L p .144 0 0 r
F P
0 g
s
.38959 .43659 m .39481 .42887 L .43696 .42643 L p .23 .097 0 r
F P
0 g
s
.19104 .52083 m .19177 .51727 L .21605 .54551 L p .893 0 .141 r
F P
0 g
s
.15611 .442 m .14714 .44183 L .15695 .48313 L p .586 0 .686 r
F P
0 g
s
.15695 .48313 m .16563 .48157 L .15611 .442 L p .586 0 .686 r
F P
0 g
s
.66038 .58678 m .6562 .59151 L .64265 .52765 L p .987 .07 0 r
F P
0 g
s
.64265 .52765 m .64606 .52318 L .66038 .58678 L p .987 .07 0 r
F P
0 g
s
.23038 .60662 m .23602 .59762 L .27784 .61454 L p .651 0 .775 r
F P
0 g
s
.66449 .46763 m .65635 .46161 L .69121 .42742 L p .237 0 .836 r
F P
0 g
s
.69121 .42742 m .70016 .43203 L .66449 .46763 L p .237 0 .836 r
F P
0 g
s
.67026 .40509 m .67514 .41294 L .64193 .44598 L p .166 0 .641 r
F P
0 g
s
.64846 .45404 m .64193 .44598 L .67514 .41294 L p .195 0 .725 r
F P
0 g
s
.12749 .49456 m .12085 .49895 L .14376 .54355 L p .897 0 .577 r
F P
0 g
s
.14113 .54814 m .14376 .54355 L .12085 .49895 L p .962 0 .4 r
F P
0 g
s
.13649 .44267 m .12556 .44441 L .13654 .49007 L p .766 0 .656 r
F P
0 g
s
.12749 .49456 m .13654 .49007 L .12556 .44441 L p .835 0 .627 r
F P
0 g
s
.17172 .48163 m .18909 .51592 L p .17507 .48488 L 0 0 .613 r
F P
0 g
s
.18909 .51592 m .19177 .51727 L .17437 .48332 L p 0 0 .613 r
F P
0 g
s
.19104 .52083 m .17318 .48643 L p .1909 .51583 L .932 0 .151 r
F P
0 g
s
.19177 .51727 m .19173 .51746 L p .1909 .51583 L .932 0 .151 r
F P
0 g
s
.19173 .51746 m .19104 .52083 L p .1909 .51583 L .932 0 .151 r
F P
0 g
s
.17437 .48332 m .17423 .48323 L p .17507 .48488 L 0 0 .613 r
F P
0 g
s
.17423 .48323 m .17172 .48163 L p .17507 .48488 L 0 0 .613 r
F P
0 g
s
.41421 .54372 m .425 .54422 L .44443 .54404 L p 0 0 0 r
F P
0 g
s
.45428 .5415 m .44443 .54404 L .425 .54422 L p 0 .751 .822 r
F P
0 g
s
.58056 .46788 m .58993 .46299 L .55869 .40749 L p .789 0 .022 r
F P
0 g
s
.58993 .46299 m .60002 .45889 L .56823 .402 L p .719 0 .023 r
F P
0 g
s
.56823 .402 m .55869 .40749 L .58993 .46299 L p .719 0 .023 r
F P
0 g
s
.44624 .48489 m .45592 .48825 L .43305 .486 L p 0 0 0 r
F P
0 g
s
.43478 .467 m .44039 .47454 L .39714 .47597 L p .272 .572 0 r
F P
0 g
s
.39714 .47597 m .39112 .46773 L .43478 .467 L p .272 .572 0 r
F P
0 g
s
.60002 .45889 m .6095 .45613 L .57741 .39758 L p .639 0 .024 r
F P
0 g
s
.57741 .39758 m .56823 .402 L .60002 .45889 L p .639 0 .024 r
F P
0 g
s
.65635 .46161 m .64846 .45404 L .68246 .42075 L p .217 0 .784 r
F P
0 g
s
.68246 .42075 m .69121 .42742 L .65635 .46161 L p .217 0 .784 r
F P
0 g
s
.67514 .41294 m .68246 .42075 L .64846 .45404 L p .195 0 .725 r
F P
0 g
s
.8682 .53164 m .86489 .52124 L .84999 .51958 L p 0 .667 .013 r
F P
0 g
s
.86489 .52124 m .86546 .51032 L .85086 .50852 L p 0 .822 .019 r
F P
0 g
s
.85086 .50852 m .84999 .51958 L .86489 .52124 L p 0 .822 .019 r
F P
0 g
s
.86546 .51032 m .86983 .50034 L .85521 .49842 L p 0 .926 .023 r
F P
0 g
s
.85521 .49842 m .85086 .50852 L .86546 .51032 L p 0 .926 .023 r
F P
0 g
s
.14113 .54814 m .14192 .55077 L .17546 .58913 L p .741 0 0 r
F P
0 g
s
.17913 .58888 m .17546 .58913 L .14192 .55077 L p .131 0 .857 r
F P
0 g
s
.43696 .42643 m .43251 .43518 L .38959 .43659 L p .23 .097 0 r
F P
0 g
s
.38707 .44657 m .38959 .43659 L .43251 .43518 L p .26 .267 0 r
F P
0 g
s
.16221 .44318 m .15611 .442 L .16563 .48157 L p .388 0 .676 r
F P
0 g
s
.16563 .48157 m .17172 .48163 L .16221 .44318 L p .388 0 .676 r
F P
0 g
s
.425 .54422 m .43527 .54063 L .45428 .5415 L p 0 .751 .822 r
F P
0 g
s
.46261 .53528 m .45428 .5415 L .43527 .54063 L p 0 .923 .864 r
F P
0 g
s
.87496 .54013 m .8682 .53164 L .85272 .53012 L p 0 .434 .005 r
F P
0 g
s
.84999 .51958 m .85272 .53012 L .8682 .53164 L p 0 .667 .013 r
F P
0 g
s
.86983 .50034 m .87743 .49262 L .86248 .49062 L p 0 .978 .027 r
F P
0 g
s
.86248 .49062 m .85521 .49842 L .86983 .50034 L p 0 .978 .027 r
F P
0 g
s
.20725 .56378 m .21271 .55587 L .24476 .57872 L p .772 0 .619 r
F P
0 g
s
.24476 .57872 m .24105 .58786 L .20725 .56378 L p .772 0 .619 r
F P
0 g
s
.44196 .49227 m .43305 .486 L .45592 .48825 L p 0 .209 0 r
F P
0 g
s
.45592 .48825 m .46385 .49516 L .44196 .49227 L p 0 .209 0 r
F P
0 g
s
.17913 .58888 m .18503 .58562 L .22491 .61363 L p .504 0 .912 r
F P
0 g
s
.23038 .60662 m .22491 .61363 L .18503 .58562 L p .632 0 .835 r
F P
0 g
s
.6176 .60099 m .60926 .59932 L .59683 .54197 L p .511 .046 0 r
F P
0 g
s
.59105 .54036 m .59683 .54197 L .60926 .59932 L p .27 .031 0 r
F P
0 g
s
.12556 .44441 m .11576 .44684 L .12749 .49456 L p .835 0 .627 r
F P
0 g
s
.12085 .49895 m .12749 .49456 L .11576 .44684 L p .911 0 .574 r
F P
0 g
s
.71389 .43293 m .70808 .43395 L .73769 .39764 L p .226 0 .863 r
F P
0 g
s
.73769 .39764 m .74404 .39481 L .71389 .43293 L p .226 0 .863 r
F P
0 g
s
.51525 .45476 m .51556 .44367 L .47064 .44671 L p .276 .418 0 r
F P
0 g
s
.47064 .44671 m .47007 .45759 L .51525 .45476 L p .276 .418 0 r
F P
0 g
s
.46856 .43613 m .47064 .44671 L .51556 .44367 L p .274 .493 0 r
F P
0 g
s
.51271 .46522 m .51525 .45476 L .47007 .45759 L p .273 .33 0 r
F P
0 g
s
.47007 .45759 m .46692 .46735 L .51271 .46522 L p .273 .33 0 r
F P
0 g
s
.43134 .45698 m .43478 .467 L .39112 .46773 L p .277 .475 0 r
F P
0 g
s
.39112 .46773 m .3876 .45745 L .43134 .45698 L p .277 .475 0 r
F P
0 g
s
.71781 .35553 m .72748 .35993 L .74919 .32201 L p .138 0 .875 r
F P
0 g
s
.72748 .35993 m .73628 .36645 L .75869 .32779 L p .121 0 .828 r
F P
0 g
s
.75869 .32779 m .74919 .32201 L .72748 .35993 L p .121 0 .828 r
F P
0 g
s
.43527 .54063 m .44365 .53343 L .46261 .53528 L p 0 .923 .864 r
F P
0 g
s
.46827 .52622 m .46261 .53528 L .44365 .53343 L p 0 .926 .77 r
F P
0 g
s
.43251 .43518 m .43054 .44586 L .38707 .44657 L p .26 .267 0 r
F P
0 g
s
.3876 .45745 m .38707 .44657 L .43054 .44586 L p .273 .38 0 r
F P
0 g
s
.6095 .45613 m .61712 .45507 L .58504 .39481 L p .52 0 .025 r
F P
0 g
s
.58504 .39481 m .57741 .39758 L .6095 .45613 L p .52 0 .025 r
F P
0 g
s
.88427 .54557 m .87496 .54013 L .8587 .53876 L p 0 .074 0 r
F P
0 g
s
.85272 .53012 m .8587 .53876 L .87496 .54013 L p 0 .434 .005 r
F P
0 g
s
.51556 .44367 m .51359 .43341 L .46856 .43613 L p .274 .493 0 r
F P
0 g
s
.46409 .42725 m .46856 .43613 L .51359 .43341 L p .265 .572 0 r
F P
0 g
s
.6562 .59151 m .64866 .59582 L .6358 .53238 L p .892 .068 0 r
F P
0 g
s
.6358 .53238 m .64265 .52765 L .6562 .59151 L p .892 .068 0 r
F P
0 g
s
.73628 .36645 m .74304 .37423 L .76597 .33549 L p .101 0 .763 r
F P
0 g
s
.76597 .33549 m .75869 .32779 L .73628 .36645 L p .101 0 .763 r
F P
0 g
s
.87743 .49262 m .88726 .48821 L .87171 .48616 L p 0 .932 .029 r
F P
0 g
s
.87171 .48616 m .86248 .49062 L .87743 .49262 L p 0 .932 .029 r
F P
0 g
s
.44806 .5015 m .44196 .49227 L .46385 .49516 L p 0 .504 .205 r
F P
0 g
s
.46385 .49516 m .46893 .50468 L .44806 .5015 L p 0 .504 .205 r
F P
0 g
s
.18697 .52614 m .19104 .52083 L .21577 .5495 L p .901 0 .5 r
F P
0 g
s
.21577 .5495 m .21271 .55587 L .18697 .52614 L p .901 0 .5 r
F P
0 g
s
.67026 .40509 m .66847 .39827 L .69428 .3654 L p .082 0 .526 r
F P
0 g
s
.69591 .35908 m .69428 .3654 L .66847 .39827 L p 0 0 .109 r
F P
0 g
s
.43054 .44586 m .43134 .45698 L .3876 .45745 L p .273 .38 0 r
F P
0 g
s
.62305 .45845 m .62187 .45587 L .64559 .51956 L p 0 0 0 r
F P
0 g
s
.64559 .51956 m .64606 .52318 L .62305 .45845 L p 0 0 0 r
F P
0 g
s
.70854 .35382 m .71781 .35553 L .73871 .3189 L p .155 0 .909 r
F P
0 g
s
.74919 .32201 m .73871 .3189 L .71781 .35553 L p .138 0 .875 r
F P
0 g
s
.44365 .53343 m .44896 .52358 L .46827 .52622 L p 0 .926 .77 r
F P
0 g
s
.47048 .51553 m .46827 .52622 L .44896 .52358 L p 0 .842 .621 r
F P
0 g
s
.12085 .49895 m .11754 .50267 L .14113 .54814 L p .962 0 .4 r
F P
0 g
s
.14192 .55077 m .14113 .54814 L .11754 .50267 L p .733 0 0 r
F P
0 g
s
.14643 .39864 m .13577 .39734 L .13649 .44267 L p .683 0 .62 r
F P
0 g
s
.13649 .44267 m .14714 .44183 L .14643 .39864 L p .683 0 .62 r
F P
0 g
s
.12556 .44441 m .13649 .44267 L .13577 .39734 L p .76 0 .616 r
F P
0 g
s
.45051 .51243 m .44806 .5015 L .46893 .50468 L p 0 .704 .437 r
F P
0 g
s
.46893 .50468 m .47048 .51553 L .45051 .51243 L p 0 .704 .437 r
F P
0 g
s
.17318 .48643 m .17437 .48332 L .19177 .51727 L p .932 0 .151 r
F P
0 g
s
.44896 .52358 m .45051 .51243 L .47048 .51553 L p 0 .842 .621 r
F P
0 g
s
.50826 .47368 m .51271 .46522 L .46692 .46735 L p .261 .2 0 r
F P
0 g
s
.46692 .46735 m .46159 .4747 L .50826 .47368 L p .261 .2 0 r
F P
0 g
s
.20015 .57214 m .20725 .56378 L .24105 .58786 L p .737 0 .702 r
F P
0 g
s
.24105 .58786 m .23602 .59762 L .20015 .57214 L p .737 0 .702 r
F P
0 g
s
.18503 .58562 m .19238 .57977 L .23038 .60662 L p .632 0 .835 r
F P
0 g
s
.23602 .59762 m .23038 .60662 L .19238 .57977 L p .695 0 .768 r
F P
0 g
s
.15523 .40053 m .14643 .39864 L .14714 .44183 L p .578 0 .615 r
F P
0 g
s
.14714 .44183 m .15611 .442 L .15523 .40053 L p .578 0 .615 r
F P
0 g
s
.88427 .54557 m .86715 .54432 L p .89411 .54714 L 0 0 0 r
F P
0 g
s
.89457 .54717 m .88427 .54557 L p .89411 .54714 L 0 0 0 r
F P
0 g
s
.8587 .53876 m .86715 .54432 L .88427 .54557 L p 0 .074 0 r
F P
0 g
s
.86715 .54432 m .87696 .54605 L .89489 .54722 L p 0 0 0 r
F P
0 g
s
.8868 .54367 m .9054 .54482 L p .87783 .54601 L 0 .817 .027 r
F P
0 g
s
.87696 .54605 m .87727 .54597 L p .87783 .54601 L 0 .817 .027 r
F P
0 g
s
.87727 .54597 m .8868 .54367 L p .87783 .54601 L 0 .817 .027 r
F P
0 g
s
.89489 .54722 m .89457 .54717 L p .89411 .54714 L 0 0 0 r
F P
0 g
s
.74304 .37423 m .74687 .38224 L .77004 .34409 L p .072 0 .655 r
F P
0 g
s
.77004 .34409 m .76597 .33549 L .74304 .37423 L p .072 0 .655 r
F P
0 g
s
.88726 .48821 m .89799 .48772 L .88168 .48566 L p 0 .684 .025 r
F P
0 g
s
.88168 .48566 m .87171 .48616 L .88726 .48821 L p 0 .684 .025 r
F P
0 g
s
.13577 .39734 m .12464 .39681 L .12556 .44441 L p .76 0 .616 r
F P
0 g
s
.11576 .44684 m .12556 .44441 L .12464 .39681 L p .832 0 .603 r
F P
0 g
s
.16221 .44318 m .17172 .48163 L p .16517 .44813 L 0 0 .504 r
F P
0 g
s
.17172 .48163 m .17437 .48332 L .1646 .4452 L p 0 0 .504 r
F P
0 g
s
.17318 .48643 m .16292 .44782 L p .17347 .48053 L .951 0 .196 r
F P
0 g
s
.17437 .48332 m .17426 .48361 L p .17347 .48053 L .951 0 .196 r
F P
0 g
s
.17426 .48361 m .17318 .48643 L p .17347 .48053 L .951 0 .196 r
F P
0 g
s
.1646 .4452 m .16437 .44501 L p .16484 .44682 L 0 0 .504 r
F P
0 g
s
.16437 .44501 m .16221 .44318 L p .16517 .44813 L 0 0 .504 r
F P
0 g
s
.16101 .40276 m .16221 .44318 L p .16289 .40902 L 0 0 .386 r
F P
0 g
s
.16297 .40506 m .16272 .40476 L p .16282 .40737 L 0 0 .386 r
F P
0 g
s
.16272 .40476 m .16101 .40276 L p .16289 .40902 L 0 0 .386 r
F P
0 g
s
.15611 .442 m .16221 .44318 L .16101 .40276 L p .382 0 .582 r
F P
0 g
s
.15523 .40053 m .16101 .40276 L .16814 .3629 L p .377 0 .493 r
F P
0 g
s
.16814 .3629 m .16101 .40276 L p .16858 .36962 L 0 0 .264 r
F P
0 g
s
.16956 .36536 m .16935 .365 L p .16888 .36784 L 0 0 .264 r
F P
0 g
s
.16935 .365 m .16814 .3629 L p .16858 .36962 L 0 0 .264 r
F P
0 g
s
.16294 .35968 m .16814 .3629 L .18308 .32613 L p .38 0 .407 r
F P
0 g
s
.18308 .32613 m .16814 .3629 L p .18223 .33241 L 0 0 .148 r
F P
0 g
s
.18395 .32867 m .18383 .32832 L p .1829 .3307 L 0 0 .148 r
F P
0 g
s
.18383 .32832 m .18308 .32613 L p .18223 .33241 L 0 0 .148 r
F P
0 g
s
.18308 .32613 m .17867 .32203 L .16294 .35968 L p .38 0 .407 r
F P
0 g
s
.17867 .32203 m .18308 .32613 L .20491 .29479 L p .4 0 .328 r
F P
0 g
s
.20491 .29479 m .18308 .32613 L p .20324 .3 L 0 0 .048 r
F P
0 g
s
.20527 .29737 m .20523 .29708 L p .20418 .29863 L 0 0 .048 r
F P
0 g
s
.20523 .29708 m .20491 .29479 L p .20324 .3 L 0 0 .048 r
F P
0 g
s
.20491 .29479 m .20147 .28987 L .17867 .32203 L p .4 0 .328 r
F P
0 g
s
.20147 .28987 m .20491 .29479 L .23244 .27068 L p .444 0 .261 r
F P
0 g
s
.23244 .27068 m .20491 .29479 L p .23053 .2747 L 0 0 0 r
F P
0 g
s
.23235 .27309 m .23244 .27068 L p .23053 .2747 L 0 0 0 r
F P
0 g
s
.26459 .25494 m .23244 .27068 L p .26288 .25807 L .141 0 0 r
F P
0 g
s
.26406 .25758 m .26408 .25747 L p .26288 .25807 L .141 0 0 r
F P
0 g
s
.26408 .25747 m .26459 .25494 L p .26288 .25807 L .141 0 0 r
F P
0 g
s
.23244 .27068 m .23234 .27328 L .26406 .25758 L p .141 0 0 r
F P
0 g
s
.22986 .27245 m .24456 .26509 L p .23281 .27296 L .604 0 .461 r
F P
0 g
s
.23234 .27328 m .23224 .27325 L p .23281 .27296 L .604 0 .461 r
F P
0 g
s
.23224 .27325 m .22986 .27245 L p .23281 .27296 L .604 0 .461 r
F P
0 g
s
.24456 .26509 m .26215 .25628 L p .23234 .27328 L .604 0 .461 r
F P
0 g
s
.23234 .27328 m .23235 .27309 L p .23147 .27387 L 0 0 0 r
F P
0 g
s
.23053 .2747 m .23234 .27328 L .23147 .27387 L 0 0 0 r
F
.20491 .29479 m .20527 .29737 L .23234 .27328 L p F P
0 g
s
.20247 .2972 m .21616 .28483 L p .20604 .29648 L .753 0 .47 r
F P
0 g
s
.20527 .29737 m .20506 .29735 L p .20604 .29648 L .753 0 .47 r
F P
0 g
s
.20506 .29735 m .20247 .2972 L p .20604 .29648 L .753 0 .47 r
F P
0 g
s
.21616 .28483 m .22986 .27245 L p .20527 .29737 L .753 0 .47 r
F P
0 g
s
.20324 .3 m .20527 .29737 L .20418 .29863 L 0 0 .048 r
F
.18308 .32613 m .18395 .32867 L .20527 .29737 L p F P
0 g
s
.18112 .32928 m .19267 .31192 L p .18482 .32701 L .853 0 .451 r
F P
0 g
s
.18395 .32867 m .18364 .32874 L p .18482 .32701 L .853 0 .451 r
F P
0 g
s
.18364 .32874 m .18112 .32928 L p .18482 .32701 L .853 0 .451 r
F P
0 g
s
.19267 .31192 m .20247 .2972 L p .18395 .32867 L .853 0 .451 r
F P
0 g
s
.18223 .33241 m .18395 .32867 L .1829 .3307 L 0 0 .148 r
F
.16814 .3629 m .16956 .36536 L .18395 .32867 L p F P
0 g
s
.16697 .36672 m .17499 .34549 L p .17021 .36299 L .914 0 .407 r
F P
0 g
s
.16956 .36536 m .1692 .36555 L p .17021 .36299 L .914 0 .407 r
F P
0 g
s
.1692 .36555 m .16697 .36672 L p .17021 .36299 L .914 0 .407 r
F P
0 g
s
.17499 .34549 m .18112 .32928 L p .16956 .36536 L .914 0 .407 r
F P
0 g
s
.16858 .36962 m .16956 .36536 L .16888 .36784 L 0 0 .264 r
F
.16101 .40276 m .16297 .40506 L .16956 .36536 L p F P
0 g
s
.1608 .4071 m .16429 .38427 L p .16313 .40247 L .945 0 .341 r
F P
0 g
s
.16297 .40506 m .16266 .40536 L p .16313 .40247 L .945 0 .341 r
F P
0 g
s
.16266 .40536 m .1608 .4071 L p .16313 .40247 L .945 0 .341 r
F P
0 g
s
.16429 .38427 m .16697 .36672 L p .16297 .40506 L .945 0 .341 r
F P
0 g
s
.16289 .40902 m .16297 .40506 L .16282 .40737 L 0 0 .386 r
F
.16221 .44318 m .1646 .4452 L .16297 .40506 L p F P
0 g
s
.16292 .44782 m .16178 .42595 L p .16428 .44304 L .955 0 .265 r
F P
0 g
s
.1646 .4452 m .16438 .44554 L p .16428 .44304 L .955 0 .265 r
F P
0 g
s
.16438 .44554 m .16292 .44782 L p .16428 .44304 L .955 0 .265 r
F P
0 g
s
.16178 .42595 m .1608 .4071 L p .1646 .4452 L .955 0 .265 r
F P
0 g
s
.16517 .44813 m .1646 .4452 L .16484 .44682 L 0 0 .504 r
F
.19238 .57977 m .20015 .57214 L .23602 .59762 L p .695 0 .768 r
F P
0 g
s
.62792 .60092 m .6176 .60099 L .60548 .54185 L p .635 .054 0 r
F P
0 g
s
.59683 .54197 m .60548 .54185 L .6176 .60099 L p .511 .046 0 r
F P
0 g
s
.51359 .43341 m .5096 .42533 L .46409 .42725 L p .265 .572 0 r
F P
0 g
s
.45781 .42127 m .46409 .42725 L .5096 .42533 L p .237 .664 0 r
F P
0 g
s
.14192 .55077 m .14604 .55108 L .17913 .58888 L p .131 0 .857 r
F P
0 g
s
.18503 .58562 m .17913 .58888 L .14604 .55108 L p .52 0 .868 r
F P
0 g
s
.11576 .44684 m .1084 .44963 L .12085 .49895 L p .911 0 .574 r
F P
0 g
s
.11754 .50267 m .12085 .49895 L .1084 .44963 L p .989 0 .427 r
F P
0 g
s
.70089 .35505 m .70854 .35382 L .72863 .31889 L p .174 0 .906 r
F P
0 g
s
.73871 .3189 m .72863 .31889 L .70854 .35382 L p .155 0 .909 r
F P
0 g
s
.64866 .59582 m .63877 .59911 L .62643 .5367 L p .804 .063 0 r
F P
0 g
s
.62643 .5367 m .6358 .53238 L .64866 .59582 L p .804 .063 0 r
F P
0 g
s
.70808 .43395 m .70016 .43203 L .72904 .39751 L p .212 0 .904 r
F P
0 g
s
.72904 .39751 m .73769 .39764 L .70808 .43395 L p .212 0 .904 r
F P
0 g
s
.61712 .45507 m .62187 .45587 L .5901 .39406 L p .273 0 .024 r
F P
0 g
s
.5901 .39406 m .58504 .39481 L .61712 .45507 L p .273 0 .024 r
F P
0 g
s
.16101 .40276 m .15523 .40053 L .15611 .442 L p .382 0 .582 r
F P
0 g
s
.9054 .54482 m .89489 .54722 L .87696 .54605 L p 0 .817 .027 r
F P
0 g
s
.89799 .48772 m .90819 .49124 L .89105 .48923 L p 0 0 0 r
F P
0 g
s
.89105 .48923 m .88168 .48566 L .89799 .48772 L p 0 0 0 r
F P
0 g
s
.50248 .47901 m .50826 .47368 L .46159 .4747 L p .213 0 0 r
F P
0 g
s
.46159 .4747 m .45477 .47863 L .50248 .47901 L p .213 0 0 r
F P
0 g
s
.55869 .40749 m .56823 .402 L .53029 .34957 L p .731 0 .067 r
F P
0 g
s
.53029 .34957 m .52154 .35632 L .55869 .40749 L p .731 0 .067 r
F P
0 g
s
.67514 .41294 m .67026 .40509 L .69628 .37316 L p .127 0 .68 r
F P
0 g
s
.69428 .3654 m .69628 .37316 L .67026 .40509 L p .082 0 .526 r
F P
0 g
s
.63877 .59911 m .62792 .60092 L .61583 .54001 L p .724 .059 0 r
F P
0 g
s
.60548 .54185 m .61583 .54001 L .62792 .60092 L p .635 .054 0 r
F P
0 g
s
.56823 .402 m .57741 .39758 L .53889 .34358 L p .659 0 .071 r
F P
0 g
s
.53889 .34358 m .53029 .34957 L .56823 .402 L p .659 0 .071 r
F P
0 g
s
.18006 .53248 m .18697 .52614 L .21271 .55587 L p .848 0 .633 r
F P
0 g
s
.21271 .55587 m .20725 .56378 L .18006 .53248 L p .848 0 .633 r
F P
0 g
s
.61583 .54001 m .62643 .5367 L .63877 .59911 L p .724 .059 0 r
F P
0 g
s
.74687 .38224 m .74723 .38943 L .77034 .35246 L p .016 0 .413 r
F P
0 g
s
.77034 .35246 m .77004 .34409 L .74687 .38224 L p .016 0 .413 r
F P
0 g
s
.12464 .39681 m .11451 .39711 L .11576 .44684 L p .832 0 .603 r
F P
0 g
s
.1084 .44963 m .11576 .44684 L .11451 .39711 L p .912 0 .571 r
F P
0 g
s
.59683 .54197 m .59105 .54036 L .57048 .48273 L p .342 .012 0 r
F P
0 g
s
.56768 .4809 m .57048 .48273 L .59105 .54036 L p 0 .001 0 r
F P
0 g
s
.64606 .52318 m .64265 .52765 L .62047 .46247 L p .982 .021 0 r
F P
0 g
s
.62047 .46247 m .62305 .45845 L .64606 .52318 L p .982 .021 0 r
F P
0 g
s
.11754 .50267 m .11805 .50521 L .14192 .55077 L p .733 0 0 r
F P
0 g
s
.14604 .55108 m .14192 .55077 L .11805 .50521 L p .149 0 .742 r
F P
0 g
s
.5096 .42533 m .50409 .42053 L .45781 .42127 L p .237 .664 0 r
F P
0 g
s
.45054 .419 m .45781 .42127 L .50409 .42053 L p .133 .715 0 r
F P
0 g
s
.91436 .53869 m .9054 .54482 L .8868 .54367 L p 0 .975 .028 r
F P
0 g
s
.15469 .35614 m .14448 .35274 L .13577 .39734 L p .676 0 .566 r
F P
0 g
s
.13577 .39734 m .14643 .39864 L .15469 .35614 L p .676 0 .566 r
F P
0 g
s
.12464 .39681 m .13577 .39734 L .14448 .35274 L p .752 0 .574 r
F P
0 g
s
.90819 .49124 m .91647 .4983 L .89853 .49641 L p 0 .242 0 r
F P
0 g
s
.89853 .49641 m .89105 .48923 L .90819 .49124 L p 0 .242 0 r
F P
0 g
s
.16827 .49055 m .17318 .48643 L .19104 .52083 L p .937 0 .502 r
F P
0 g
s
.19104 .52083 m .18697 .52614 L .16827 .49055 L p .937 0 .502 r
F P
0 g
s
.14604 .55108 m .15298 .54898 L .18503 .58562 L p .52 0 .868 r
F P
0 g
s
.19238 .57977 m .18503 .58562 L .15298 .54898 L p .66 0 .815 r
F P
0 g
s
.70016 .43203 m .69121 .42742 L .71925 .3944 L p .192 0 .865 r
F P
0 g
s
.71925 .3944 m .72904 .39751 L .70016 .43203 L p .192 0 .865 r
F P
0 g
s
.16294 .35968 m .15469 .35614 L .14643 .39864 L p .572 0 .547 r
F P
0 g
s
.14643 .39864 m .15523 .40053 L .16294 .35968 L p .572 0 .547 r
F P
0 g
s
.57741 .39758 m .58504 .39481 L .54621 .33914 L p .557 0 .074 r
F P
0 g
s
.54621 .33914 m .53889 .34358 L .57741 .39758 L p .557 0 .074 r
F P
0 g
s
.16292 .44782 m .1646 .4452 L .17437 .48332 L p .951 0 .196 r
F P
0 g
s
.69628 .37316 m .70165 .38132 L .67514 .41294 L p .127 0 .68 r
F P
0 g
s
.68246 .42075 m .67514 .41294 L .70165 .38132 L p .153 0 .761 r
F P
0 g
s
.14448 .35274 m .13365 .34992 L .12464 .39681 L p .752 0 .574 r
F P
0 g
s
.11451 .39711 m .12464 .39681 L .13365 .34992 L p .822 0 .575 r
F P
0 g
s
.17127 .539 m .18006 .53248 L .20725 .56378 L p .795 0 .708 r
F P
0 g
s
.20725 .56378 m .20015 .57214 L .17127 .539 L p .795 0 .708 r
F P
0 g
s
.92055 .52964 m .91436 .53869 L .89533 .53747 L p 0 .97 .025 r
F P
0 g
s
.8868 .54367 m .89533 .53747 L .91436 .53869 L p 0 .975 .028 r
F P
0 g
s
.91647 .4983 m .92168 .50796 L .90309 .50624 L p 0 .553 .008 r
F P
0 g
s
.90309 .50624 m .89853 .49641 L .91647 .4983 L p 0 .553 .008 r
F P
0 g
s
.1084 .44963 m .10449 .45242 L .11754 .50267 L p .989 0 .427 r
F P
0 g
s
.11805 .50521 m .11754 .50267 L .10449 .45242 L p .716 0 0 r
F P
0 g
s
.49613 .48046 m .50248 .47901 L .45916 .47867 L p .031 0 0 r
F P
0 g
s
.50248 .47901 m .49613 .48046 L .5492 .48576 L p .16 0 0 r
F P
0 g
s
.49613 .48046 m .49006 .47779 L .52049 .48164 L p .085 .659 0 r
F P
0 g
s
.5492 .48576 m .55433 .48276 L .50745 .47937 L p .16 0 0 r
F P
0 g
s
.50745 .47937 m .50248 .47901 L p .5492 .48576 L .16 0 0 r
F P
0 g
s
.57048 .48273 m .57668 .4827 L .59683 .54197 L p .342 .012 0 r
F P
0 g
s
.60548 .54185 m .59683 .54197 L .57668 .4827 L p .548 .016 0 r
F P
0 g
s
.61583 .54001 m .60548 .54185 L .58546 .48079 L p .659 .018 0 r
F P
0 g
s
.57668 .4827 m .58546 .48079 L .60548 .54185 L p .548 .016 0 r
F P
0 g
s
.62643 .5367 m .61583 .54001 L .59562 .47726 L p .74 .02 0 r
F P
0 g
s
.58546 .48079 m .59562 .47726 L .61583 .54001 L p .659 .018 0 r
F P
0 g
s
.59562 .47726 m .60575 .47257 L .62643 .5367 L p .74 .02 0 r
F P
0 g
s
.55433 .48276 m .5492 .48576 L .60437 .49354 L p .208 .037 0 r
F P
0 g
s
.60098 .49397 m .60437 .49354 L .5492 .48576 L p 0 .517 0 r
F P
0 g
s
.5492 .48576 m .54432 .48465 L .57667 .48997 L p 0 .517 0 r
F P
0 g
s
.57667 .48997 m .60098 .49397 L p .5492 .48576 L 0 .517 0 r
F P
0 g
s
.52049 .48164 m .54432 .48465 L p .49613 .48046 L .085 .659 0 r
F P
0 g
s
.45916 .47867 m .45477 .47863 L p .49613 .48046 L .031 0 0 r
F P
0 g
s
.69121 .42742 m .68246 .42075 L .70967 .38875 L p .173 0 .817 r
F P
0 g
s
.70967 .38875 m .71925 .3944 L .69121 .42742 L p .173 0 .817 r
F P
0 g
s
.15298 .54898 m .1618 .54476 L .19238 .57977 L p .66 0 .815 r
F P
0 g
s
.20015 .57214 m .19238 .57977 L .1618 .54476 L p .738 0 .763 r
F P
0 g
s
.70165 .38132 m .70967 .38875 L .68246 .42075 L p .153 0 .761 r
F P
0 g
s
.59186 .39547 m .5901 .39406 L .62187 .45587 L p 0 0 .015 r
F P
0 g
s
.62187 .45587 m .62305 .45845 L .59186 .39547 L p 0 0 .015 r
F P
0 g
s
.74723 .38943 m .74404 .39481 L .7668 .35946 L p 0 0 0 r
F P
0 g
s
.7668 .35946 m .77034 .35246 L .74723 .38943 L p 0 0 0 r
F P
0 g
s
.92311 .51891 m .92055 .52964 L .90138 .52829 L p 0 .889 .021 r
F P
0 g
s
.89533 .53747 m .90138 .52829 L .92055 .52964 L p 0 .97 .025 r
F P
0 g
s
.92168 .50796 m .92311 .51891 L .90409 .51738 L p 0 .755 .015 r
F P
0 g
s
.90409 .51738 m .90309 .50624 L .92168 .50796 L p 0 .755 .015 r
F P
0 g
s
.1618 .54476 m .17127 .539 L .20015 .57214 L p .738 0 .763 r
F P
0 g
s
.16814 .3629 m .16294 .35968 L .15523 .40053 L p .377 0 .493 r
F P
0 g
s
.90138 .52829 m .90409 .51738 L .92311 .51891 L p 0 .889 .021 r
F P
0 g
s
.11451 .39711 m .1067 .39821 L .1084 .44963 L p .912 0 .571 r
F P
0 g
s
.10449 .45242 m .1084 .44963 L .1067 .39821 L p .995 0 .462 r
F P
0 g
s
.50409 .42053 m .49779 .41968 L .45054 .419 L p .133 .715 0 r
F P
0 g
s
.44326 .42079 m .45054 .419 L .49779 .41968 L p .11 0 0 r
F P
0 g
s
.64265 .52765 m .6358 .53238 L .61444 .46741 L p .901 .022 0 r
F P
0 g
s
.61444 .46741 m .62047 .46247 L .64265 .52765 L p .901 .022 0 r
F P
0 g
s
.51377 .36296 m .52154 .35632 L .47989 .31186 L p .796 0 .1 r
F P
0 g
s
.52154 .35632 m .53029 .34957 L .48761 .30393 L p .738 0 .11 r
F P
0 g
s
.48761 .30393 m .47989 .31186 L .52154 .35632 L p .738 0 .11 r
F P
0 g
s
.53029 .34957 m .53889 .34358 L .49528 .29646 L p .675 0 .117 r
F P
0 g
s
.49528 .29646 m .48761 .30393 L .53029 .34957 L p .675 0 .117 r
F P
0 g
s
.11805 .50521 m .12234 .5062 L .14604 .55108 L p .149 0 .742 r
F P
0 g
s
.15298 .54898 m .14604 .55108 L .12234 .5062 L p .508 0 .774 r
F P
0 g
s
.13365 .34992 m .1236 .34805 L .11451 .39711 L p .822 0 .575 r
F P
0 g
s
.1067 .39821 m .11451 .39711 L .1236 .34805 L p .9 0 .563 r
F P
0 g
s
.58504 .39481 m .5901 .39406 L .55126 .33684 L p .356 0 .076 r
F P
0 g
s
.55126 .33684 m .54621 .33914 L .58504 .39481 L p .356 0 .076 r
F P
0 g
s
.17134 .31691 m .16205 .31144 L .14448 .35274 L p .669 0 .511 r
F P
0 g
s
.14448 .35274 m .15469 .35614 L .17134 .31691 L p .669 0 .511 r
F P
0 g
s
.13365 .34992 m .14448 .35274 L .16205 .31144 L p .741 0 .53 r
F P
0 g
s
.16027 .49514 m .16827 .49055 L .18697 .52614 L p .866 0 .617 r
F P
0 g
s
.18697 .52614 m .18006 .53248 L .16027 .49514 L p .866 0 .617 r
F P
0 g
s
.17867 .32203 m .17134 .31691 L .15469 .35614 L p .569 0 .48 r
F P
0 g
s
.15469 .35614 m .16294 .35968 L .17867 .32203 L p .569 0 .48 r
F P
0 g
s
.16205 .31144 m .15203 .30635 L .13365 .34992 L p .741 0 .53 r
F P
0 g
s
.1236 .34805 m .13365 .34992 L .15203 .30635 L p .807 0 .542 r
F P
0 g
s
.49006 .47779 m .49613 .48046 L .44737 .47859 L p .178 .695 0 r
F P
0 g
s
.44737 .47859 m .44039 .47454 L .49006 .47779 L p .178 .695 0 r
F P
0 g
s
.15734 .45068 m .16292 .44782 L .17318 .48643 L p .958 0 .514 r
F P
0 g
s
.17318 .48643 m .16827 .49055 L .15734 .45068 L p .958 0 .514 r
F P
0 g
s
.50801 .3686 m .51377 .36296 L .47313 .31921 L p .857 0 .084 r
F P
0 g
s
.47989 .31186 m .47313 .31921 L .51377 .36296 L p .796 0 .1 r
F P
0 g
s
.53889 .34358 m .54621 .33914 L .5019 .29044 L p .587 0 .124 r
F P
0 g
s
.5019 .29044 m .49528 .29646 L .53889 .34358 L p .587 0 .124 r
F P
0 g
s
.1608 .4071 m .16297 .40506 L .1646 .4452 L p .955 0 .265 r
F P
0 g
s
.74404 .39481 m .73769 .39764 L .75986 .36413 L p .177 0 .832 r
F P
0 g
s
.75986 .36413 m .7668 .35946 L .74404 .39481 L p .177 0 .832 r
F P
0 g
s
.10449 .45242 m .10459 .45482 L .11805 .50521 L p .716 0 0 r
F P
0 g
s
.12234 .5062 m .11805 .50521 L .10459 .45482 L p .158 0 .627 r
F P
0 g
s
.6358 .53238 m .62643 .5367 L .60575 .47257 L p .816 .021 0 r
F P
0 g
s
.60575 .47257 m .61444 .46741 L .6358 .53238 L p .816 .021 0 r
F P
0 g
s
.49779 .41968 m .49154 .42295 L .44326 .42079 L p .11 0 0 r
F P
0 g
s
.43696 .42643 m .44326 .42079 L .49154 .42295 L p .23 .057 0 r
F P
0 g
s
.62305 .45845 m .62047 .46247 L .59005 .39887 L p .959 0 .009 r
F P
0 g
s
.59005 .39887 m .59186 .39547 L .62305 .45845 L p .959 0 .009 r
F P
0 g
s
.12234 .5062 m .12988 .50549 L .15298 .54898 L p .508 0 .774 r
F P
0 g
s
.1618 .54476 m .15298 .54898 L .12988 .50549 L p .651 0 .747 r
F P
0 g
s
.15026 .49955 m .16027 .49514 L .18006 .53248 L p .801 0 .674 r
F P
0 g
s
.18006 .53248 m .17127 .539 L .15026 .49955 L p .801 0 .674 r
F P
0 g
s
.85086 .50852 m .85521 .49842 L .83219 .49467 L p .027 .831 0 r
F P
0 g
s
.83219 .49467 m .82812 .50496 L .85086 .50852 L p .027 .831 0 r
F P
0 g
s
.84999 .51958 m .85086 .50852 L .82812 .50496 L p .028 .753 0 r
F P
0 g
s
.85272 .53012 m .84999 .51958 L .82697 .5161 L p .028 .645 0 r
F P
0 g
s
.82812 .50496 m .82697 .5161 L .84999 .51958 L p .028 .753 0 r
F P
0 g
s
.1067 .39821 m .10229 .39998 L .10449 .45242 L p .995 0 .462 r
F P
0 g
s
.10459 .45482 m .10449 .45242 L .10229 .39998 L p 0 0 .025 r
F P
0 g
s
.19543 .28325 m .1876 .27583 L .16205 .31144 L p .667 0 .457 r
F P
0 g
s
.16205 .31144 m .17134 .31691 L .19543 .28325 L p .667 0 .457 r
F P
0 g
s
.15203 .30635 m .16205 .31144 L .1876 .27583 L p .731 0 .484 r
F P
0 g
s
.69628 .37316 m .69428 .3654 L .71304 .33557 L p .046 0 .552 r
F P
0 g
s
.71486 .3278 m .71304 .33557 L .69428 .3654 L p 0 0 .124 r
F P
0 g
s
.1236 .34805 m .11567 .34739 L .1067 .39821 L p .9 0 .563 r
F P
0 g
s
.10229 .39998 m .1067 .39821 L .11567 .34739 L p .982 0 .495 r
F P
0 g
s
.47989 .31186 m .48761 .30393 L .44172 .26706 L p .74 0 .15 r
F P
0 g
s
.48761 .30393 m .49528 .29646 L .44805 .25823 L p .686 0 .163 r
F P
0 g
s
.44805 .25823 m .44172 .26706 L .48761 .30393 L p .686 0 .163 r
F P
0 g
s
.15203 .30635 m .14257 .3023 L .1236 .34805 L p .807 0 .542 r
F P
0 g
s
.11567 .34739 m .1236 .34805 L .14257 .3023 L p .879 0 .548 r
F P
0 g
s
.47313 .31921 m .47989 .31186 L .43536 .27609 L p .788 0 .134 r
F P
0 g
s
.44172 .26706 m .43536 .27609 L .47989 .31186 L p .74 0 .15 r
F P
0 g
s
.85521 .49842 m .86248 .49062 L .83866 .4866 L p .025 .888 0 r
F P
0 g
s
.83866 .4866 m .83219 .49467 L .85521 .49842 L p .025 .888 0 r
F P
0 g
s
.20147 .28987 m .19543 .28325 L .17134 .31691 L p .576 0 .415 r
F P
0 g
s
.17134 .31691 m .17867 .32203 L .20147 .28987 L p .576 0 .415 r
F P
0 g
s
.5901 .39406 m .59186 .39547 L .55336 .33703 L p 0 0 .061 r
F P
0 g
s
.55336 .33703 m .55126 .33684 L .5901 .39406 L p 0 0 .061 r
F P
0 g
s
.56264 .46665 m .56475 .4557 L .51525 .45476 L p .264 .41 0 r
F P
0 g
s
.51556 .44367 m .51525 .45476 L .56475 .4557 L p .261 .478 0 r
F P
0 g
s
.8587 .53876 m .85272 .53012 L .82891 .52663 L p .026 .469 0 r
F P
0 g
s
.82697 .5161 m .82891 .52663 L .85272 .53012 L p .028 .645 0 r
F P
0 g
s
.47313 .31921 m .46822 .32503 L .50801 .3686 L p .857 0 .084 r
F P
0 g
s
.1876 .27583 m .17899 .26855 L .15203 .30635 L p .731 0 .484 r
F P
0 g
s
.14257 .3023 m .15203 .30635 L .17899 .26855 L p .788 0 .506 r
F P
0 g
s
.12988 .50549 m .13964 .50317 L .1618 .54476 L p .651 0 .747 r
F P
0 g
s
.17127 .539 m .1618 .54476 L .13964 .50317 L p .735 0 .714 r
F P
0 g
s
.54621 .33914 m .55126 .33684 L .50658 .28666 L p .422 0 .131 r
F P
0 g
s
.50658 .28666 m .5019 .29044 L .54621 .33914 L p .422 0 .131 r
F P
0 g
s
.56475 .4557 m .56506 .44468 L .51556 .44367 L p .261 .478 0 r
F P
0 g
s
.51359 .43341 m .51556 .44367 L .56506 .44468 L p .254 .542 0 r
F P
0 g
s
.13964 .50317 m .15026 .49955 L .17127 .539 L p .735 0 .714 r
F P
0 g
s
.4851 .47133 m .49006 .47779 L .44039 .47454 L p .244 .611 0 r
F P
0 g
s
.44039 .47454 m .43478 .467 L .4851 .47133 L p .244 .611 0 r
F P
0 g
s
.49528 .29646 m .5019 .29044 L .45352 .25074 L p .612 0 .174 r
F P
0 g
s
.45352 .25074 m .44805 .25823 L .49528 .29646 L p .612 0 .174 r
F P
0 g
s
.51271 .46522 m .50826 .47368 L .55901 .47608 L p .262 .319 0 r
F P
0 g
s
.14859 .4534 m .15734 .45068 L .16827 .49055 L p .875 0 .603 r
F P
0 g
s
.16827 .49055 m .16027 .49514 L .14859 .4534 L p .875 0 .603 r
F P
0 g
s
.73769 .39764 m .72904 .39751 L .75045 .36581 L p .165 0 .919 r
F P
0 g
s
.75045 .36581 m .75986 .36413 L .73769 .39764 L p .165 0 .919 r
F P
0 g
s
.86248 .49062 m .87171 .48616 L .84668 .48182 L p .02 .902 0 r
F P
0 g
s
.84668 .48182 m .83866 .4866 L .86248 .49062 L p .02 .902 0 r
F P
0 g
s
.46822 .32503 m .47313 .31921 L .4298 .28414 L p .837 0 .108 r
F P
0 g
s
.43536 .27609 m .4298 .28414 L .47313 .31921 L p .788 0 .134 r
F P
0 g
s
.49154 .42295 m .48621 .42992 L .43696 .42643 L p .23 .057 0 r
F P
0 g
s
.43251 .43518 m .43696 .42643 L .48621 .42992 L p .261 .252 0 r
F P
0 g
s
.22585 .2571 m .22006 .24794 L .1876 .27583 L p .67 0 .404 r
F P
0 g
s
.1876 .27583 m .19543 .28325 L .22585 .2571 L p .67 0 .404 r
F P
0 g
s
.17899 .26855 m .1876 .27583 L .22006 .24794 L p .722 0 .438 r
F P
0 g
s
.10459 .45482 m .10875 .45652 L .12234 .5062 L p .158 0 .627 r
F P
0 g
s
.12988 .50549 m .12234 .5062 L .10875 .45652 L p .499 0 .686 r
F P
0 g
s
.56506 .44468 m .56354 .43502 L .51359 .43341 L p .254 .542 0 r
F P
0 g
s
.86715 .54432 m .8587 .53876 L .83368 .53517 L p .02 .133 0 r
F P
0 g
s
.82891 .52663 m .83368 .53517 L .8587 .53876 L p .026 .469 0 r
F P
0 g
s
.15474 .40861 m .1608 .4071 L .16292 .44782 L p .962 0 .527 r
F P
0 g
s
.16292 .44782 m .15734 .45068 L .15474 .40861 L p .962 0 .527 r
F P
0 g
s
.16697 .36672 m .16956 .36536 L .16297 .40506 L p .945 0 .341 r
F P
0 g
s
.43536 .27609 m .44172 .26706 L .39423 .24051 L p .735 0 .19 r
F P
0 g
s
.44172 .26706 m .44805 .25823 L .39871 .23048 L p .693 0 .208 r
F P
0 g
s
.39871 .23048 m .39423 .24051 L .44172 .26706 L p .693 0 .208 r
F P
0 g
s
.70165 .38132 m .69628 .37316 L .71514 .34425 L p .087 0 .715 r
F P
0 g
s
.71304 .33557 m .71514 .34425 L .69628 .37316 L p .046 0 .552 r
F P
0 g
s
.23016 .26503 m .22585 .2571 L .19543 .28325 L p .595 0 .356 r
F P
0 g
s
.19543 .28325 m .20147 .28987 L .23016 .26503 L p .595 0 .356 r
F P
0 g
s
.74919 .32201 m .75869 .32779 L .77315 .29362 L p .079 0 .871 r
F P
0 g
s
.77315 .29362 m .76305 .28855 L .74919 .32201 L p .079 0 .871 r
F P
0 g
s
.4298 .28414 m .43536 .27609 L .38965 .25052 L p .772 0 .166 r
F P
0 g
s
.39423 .24051 m .38965 .25052 L .43536 .27609 L p .735 0 .19 r
F P
0 g
s
.62047 .46247 m .61444 .46741 L .58488 .40382 L p .907 0 .018 r
F P
0 g
s
.58488 .40382 m .59005 .39887 L .62047 .46247 L p .907 0 .018 r
F P
0 g
s
.22006 .24794 m .21355 .23875 L .17899 .26855 L p .722 0 .438 r
F P
0 g
s
.17074 .26239 m .17899 .26855 L .21355 .23875 L p .766 0 .467 r
F P
0 g
s
.17899 .26855 m .17074 .26239 L .14257 .3023 L p .788 0 .506 r
F P
0 g
s
.13494 .29983 m .14257 .3023 L .17074 .26239 L p .849 0 .525 r
F P
0 g
s
.14257 .3023 m .13494 .29983 L .11567 .34739 L p .879 0 .548 r
F P
0 g
s
.48195 .46192 m .4851 .47133 L .43478 .467 L p .263 .525 0 r
F P
0 g
s
.43478 .467 m .43134 .45698 L .48195 .46192 L p .263 .525 0 r
F P
0 g
s
.75869 .32779 m .76597 .33549 L .78085 .30126 L p .06 0 .8 r
F P
0 g
s
.78085 .30126 m .77315 .29362 L .75869 .32779 L p .06 0 .8 r
F P
0 g
s
.11093 .34803 m .11567 .34739 L .13494 .29983 L p .952 0 .516 r
F P
0 g
s
.11567 .34739 m .11093 .34803 L .10229 .39998 L p .982 0 .495 r
F P
0 g
s
.55433 .48276 m .55901 .47608 L .50826 .47368 L p .247 .158 0 r
F P
0 g
s
.50826 .47368 m .50248 .47901 L .55433 .48276 L p .247 .158 0 r
F P
0 g
s
.10229 .39998 m .10191 .40218 L .10459 .45482 L p 0 0 .025 r
F P
0 g
s
.10875 .45652 m .10459 .45482 L .10191 .40218 L p .159 0 .516 r
F P
0 g
s
.44805 .25823 m .45352 .25074 L .40251 .22174 L p .635 0 .225 r
F P
0 g
s
.40251 .22174 m .39871 .23048 L .44805 .25823 L p .635 0 .225 r
F P
0 g
s
.26147 .23987 m .25826 .22947 L .22006 .24794 L p .68 0 .355 r
F P
0 g
s
.22006 .24794 m .22585 .2571 L .26147 .23987 L p .68 0 .355 r
F P
0 g
s
.21355 .23875 m .22006 .24794 L .25826 .22947 L p .714 0 .391 r
F P
0 g
s
.48621 .42992 m .48254 .43967 L .43251 .43518 L p .261 .252 0 r
F P
0 g
s
.43054 .44586 m .43251 .43518 L .48254 .43967 L p .269 .364 0 r
F P
0 g
s
.10191 .40218 m .10229 .39998 L .11093 .34803 L p .7 0 .09 r
F P
0 g
s
.72904 .39751 m .71925 .3944 L .73985 .36425 L p .146 0 .896 r
F P
0 g
s
.73985 .36425 m .75045 .36581 L .72904 .39751 L p .146 0 .896 r
F P
0 g
s
.5019 .29044 m .50658 .28666 L .45739 .2456 L p .477 0 .187 r
F P
0 g
s
.45739 .2456 m .45352 .25074 L .5019 .29044 L p .477 0 .187 r
F P
0 g
s
.87171 .48616 m .88168 .48566 L .85521 .481 L p .007 .734 0 r
F P
0 g
s
.85521 .481 m .84668 .48182 L .87171 .48616 L p .007 .734 0 r
F P
0 g
s
.13785 .45558 m .14859 .4534 L .16027 .49514 L p .802 0 .642 r
F P
0 g
s
.16027 .49514 m .15026 .49955 L .13785 .45558 L p .802 0 .642 r
F P
0 g
s
.38965 .25052 m .39423 .24051 L .34677 .22525 L p .726 0 .228 r
F P
0 g
s
.39423 .24051 m .39871 .23048 L .34886 .21437 L p .699 0 .253 r
F P
0 g
s
.34886 .21437 m .34677 .22525 L .39423 .24051 L p .699 0 .253 r
F P
0 g
s
.26368 .24875 m .26147 .23987 L .22585 .2571 L p .627 0 .305 r
F P
0 g
s
.22585 .2571 m .23016 .26503 L .26368 .24875 L p .627 0 .305 r
F P
0 g
s
.55126 .33684 m .55336 .33703 L .50869 .28566 L p 0 0 .12 r
F P
0 g
s
.50869 .28566 m .50658 .28666 L .55126 .33684 L p 0 0 .12 r
F P
0 g
s
.59186 .39547 m .59005 .39887 L .55217 .33971 L p .907 0 .012 r
F P
0 g
s
.55217 .33971 m .55336 .33703 L .59186 .39547 L p .907 0 .012 r
F P
0 g
s
.48105 .45086 m .48195 .46192 L .43134 .45698 L p .269 .447 0 r
F P
0 g
s
.43134 .45698 m .43054 .44586 L .48105 .45086 L p .269 .447 0 r
F P
0 g
s
.76597 .33549 m .77004 .34409 L .78511 .31046 L p .033 0 .671 r
F P
0 g
s
.78511 .31046 m .78085 .30126 L .76597 .33549 L p .033 0 .671 r
F P
0 g
s
.87696 .54605 m .86715 .54432 L .84067 .54055 L p .004 0 0 r
F P
0 g
s
.83368 .53517 m .84067 .54055 L .86715 .54432 L p .02 .133 0 r
F P
0 g
s
.71514 .34425 m .72088 .35267 L .70165 .38132 L p .087 0 .715 r
F P
0 g
s
.70967 .38875 m .70165 .38132 L .72088 .35267 L p .111 0 .799 r
F P
0 g
s
.23244 .27068 m .23016 .26503 L .20147 .28987 L p .444 0 .261 r
F P
0 g
s
.10875 .45652 m .11644 .45726 L .12988 .50549 L p .499 0 .686 r
F P
0 g
s
.13964 .50317 m .12988 .50549 L .11644 .45726 L p .642 0 .681 r
F P
0 g
s
.4298 .28414 m .42576 .29015 L .46822 .32503 L p .837 0 .108 r
F P
0 g
s
.42576 .29015 m .4298 .28414 L .38555 .25921 L p .803 0 .126 r
F P
0 g
s
.38965 .25052 m .38555 .25921 L .4298 .28414 L p .772 0 .166 r
F P
0 g
s
.38555 .25921 m .38965 .25052 L .34446 .23598 L p .745 0 .196 r
F P
0 g
s
.34677 .22525 m .34446 .23598 L .38965 .25052 L p .726 0 .228 r
F P
0 g
s
.48254 .43967 m .48105 .45086 L .43054 .44586 L p .269 .364 0 r
F P
0 g
s
.25826 .22947 m .25447 .21891 L .21355 .23875 L p .714 0 .391 r
F P
0 g
s
.20718 .23073 m .21355 .23875 L .25447 .21891 L p .741 0 .424 r
F P
0 g
s
.21355 .23875 m .20718 .23073 L .17074 .26239 L p .766 0 .467 r
F P
0 g
s
.30133 .23258 m .30096 .22165 L .25826 .22947 L p .695 0 .31 r
F P
0 g
s
.25826 .22947 m .26147 .23987 L .30133 .23258 L p .695 0 .31 r
F P
0 g
s
.25447 .21891 m .25826 .22947 L .30096 .22165 L p .708 0 .345 r
F P
0 g
s
.34446 .23598 m .34677 .22525 L .30096 .22165 L p .712 0 .268 r
F P
0 g
s
.34677 .22525 m .34886 .21437 L .30017 .21052 L p .703 0 .299 r
F P
0 g
s
.30017 .21052 m .30096 .22165 L .34677 .22525 L p .703 0 .299 r
F P
0 g
s
.71925 .3944 m .70967 .38875 L .72952 .35966 L p .129 0 .854 r
F P
0 g
s
.72952 .35966 m .73985 .36425 L .71925 .3944 L p .129 0 .854 r
F P
0 g
s
.39871 .23048 m .40251 .22174 L .35043 .20474 L p .658 0 .277 r
F P
0 g
s
.35043 .20474 m .34886 .21437 L .39871 .23048 L p .658 0 .277 r
F P
0 g
s
.16393 .25814 m .17074 .26239 L .20718 .23073 L p .809 0 .496 r
F P
0 g
s
.17074 .26239 m .16393 .25814 L .13494 .29983 L p .849 0 .525 r
F P
0 g
s
.30096 .22165 m .30133 .23258 L .34446 .23598 L p .712 0 .268 r
F P
0 g
s
.30096 .22165 m .30017 .21052 L .25447 .21891 L p .708 0 .345 r
F P
0 g
s
.30124 .24187 m .30133 .23258 L .26147 .23987 L p .668 0 .263 r
F P
0 g
s
.26147 .23987 m .26368 .24875 L .30124 .24187 L p .668 0 .263 r
F P
0 g
s
.72088 .35267 m .72952 .35966 L .70967 .38875 L p .111 0 .799 r
F P
0 g
s
.34446 .23598 m .3422 .24515 L .38555 .25921 L p .745 0 .196 r
F P
0 g
s
.3422 .24515 m .34446 .23598 L .30133 .23258 L p .71 0 .227 r
F P
0 g
s
.61444 .46741 m .60575 .47257 L .59086 .43995 L p .826 0 .021 r
F P
0 g
s
.59562 .47726 m .58546 .48079 L .55782 .42068 L p .679 0 .023 r
F P
0 g
s
.60575 .47257 m .59562 .47726 L .56756 .41554 L p .754 0 .022 r
F P
0 g
s
.56756 .41554 m .57702 .40965 L .60575 .47257 L p .754 0 .022 r
F P
0 g
s
.55782 .42068 m .56756 .41554 L .59562 .47726 L p .679 0 .023 r
F P
0 g
s
.61168 .48125 m .61463 .47107 L .57489 .46769 L p .246 .402 0 r
F P
0 g
s
.57489 .46769 m .56264 .46665 L p .57361 .46991 L .246 .402 0 r
F P
0 g
s
.57361 .46991 m .61168 .48125 L .57489 .46769 L .246 .402 0 r
F
.59086 .43995 m .58439 .42579 L p .59253 .43706 L .826 0 .021 r
F P
0 g
s
.59253 .43706 m .61444 .46741 L .59086 .43995 L .826 0 .021 r
F
.58439 .42579 m .58061 .4175 L p .58485 .42376 L F P
0 g
s
.58485 .42376 m .61444 .46741 L .58439 .42579 L .826 0 .021 r
F
.56475 .4557 m .56264 .46665 L .57489 .46769 L p .244 .473 0 r
F P
0 g
s
.57489 .46769 m .61463 .47107 L p .57677 .4594 L .244 .473 0 r
F P
0 g
s
.56475 .4557 m .57489 .46769 L .57677 .4594 L .244 .473 0 r
F
.58079 .41547 m .61444 .46741 L .58061 .4175 L .826 0 .021 r
F
.58488 .40382 m .61444 .46741 L p .58079 .41547 L F P
0 g
s
.57702 .40965 m .58095 .40674 L p .58079 .41547 L .826 0 .021 r
F P
0 g
s
.58095 .40674 m .58488 .40382 L p .58079 .41547 L .826 0 .021 r
F P
0 g
s
.61463 .47107 m .61658 .4599 L .58005 .45694 L p .244 .473 0 r
F P
0 g
s
.58005 .45694 m .57702 .45669 L p .58003 .45785 L .244 .473 0 r
F P
0 g
s
.58003 .45785 m .61463 .47107 L .58005 .45694 L .244 .473 0 r
F
.57702 .45669 m .56475 .4557 L p .57677 .4594 L F P
0 g
s
.57677 .4594 m .61463 .47107 L .57702 .45669 L .244 .473 0 r
F
.58061 .4175 m .57702 .40965 L p .58079 .41547 L .826 0 .021 r
F P
0 g
s
.12661 .45693 m .13785 .45558 L .15026 .49955 L p .73 0 .666 r
F P
0 g
s
.15026 .49955 m .13964 .50317 L .12661 .45693 L p .73 0 .666 r
F P
0 g
s
.30133 .23258 m .30124 .24187 L .3422 .24515 L p .71 0 .227 r
F P
0 g
s
.25058 .20956 m .25447 .21891 L .30017 .21052 L p .712 0 .377 r
F P
0 g
s
.25447 .21891 m .25058 .20956 L .20718 .23073 L p .741 0 .424 r
F P
0 g
s
.11644 .45726 m .12661 .45693 L .13964 .50317 L p .642 0 .681 r
F P
0 g
s
.14557 .40938 m .15474 .40861 L .15734 .45068 L p .874 0 .588 r
F P
0 g
s
.15734 .45068 m .14859 .4534 L .14557 .40938 L p .874 0 .588 r
F P
0 g
s
.18112 .32928 m .18395 .32867 L .16956 .36536 L p .914 0 .407 r
F P
0 g
s
.34886 .21437 m .35043 .20474 L .29907 .20063 L p .684 0 .328 r
F P
0 g
s
.29907 .20063 m .30017 .21052 L .34886 .21437 L p .684 0 .328 r
F P
0 g
s
.45352 .25074 m .45739 .2456 L .40511 .21545 L p .528 0 .245 r
F P
0 g
s
.40511 .21545 m .40251 .22174 L .45352 .25074 L p .528 0 .245 r
F P
0 g
s
.26459 .25494 m .26368 .24875 L .23016 .26503 L p .513 0 .21 r
F P
0 g
s
.23016 .26503 m .23244 .27068 L .26459 .25494 L p .513 0 .21 r
F P
0 g
s
.88168 .48566 m .89105 .48923 L .86308 .48431 L p .013 0 0 r
F P
0 g
s
.86308 .48431 m .85521 .481 L .88168 .48566 L p .013 0 0 r
F P
0 g
s
.30017 .21052 m .29907 .20063 L .25058 .20956 L p .712 0 .377 r
F P
0 g
s
.13016 .29928 m .13494 .29983 L .16393 .25814 L p .902 0 .523 r
F P
0 g
s
.13494 .29983 m .13016 .29928 L .11093 .34803 L p .952 0 .516 r
F P
0 g
s
.16066 .36679 m .16697 .36672 L .1608 .4071 L p .95 0 .537 r
F P
0 g
s
.1608 .4071 m .15474 .40861 L .16066 .36679 L p .95 0 .537 r
F P
0 g
s
.38555 .25921 m .3825 .26544 L .42576 .29015 L p .803 0 .126 r
F P
0 g
s
.3825 .26544 m .38555 .25921 L .3422 .24515 L p .751 0 .141 r
F P
0 g
s
.77004 .34409 m .77034 .35246 L .78533 .31999 L p 0 0 .375 r
F P
0 g
s
.78533 .31999 m .78511 .31046 L .77004 .34409 L p 0 0 .375 r
F P
0 g
s
.37854 .49707 m .37559 .50818 L .39205 .50992 L p 0 .819 .62 r
F P
0 g
s
.39205 .50992 m .39551 .4993 L .37854 .49707 L p 0 .819 .62 r
F P
0 g
s
.3925 .52109 m .39205 .50992 L .37559 .50818 L p 0 .691 .405 r
F P
0 g
s
.84067 .54055 m .84898 .54203 L .87696 .54605 L p .004 0 0 r
F P
0 g
s
.8868 .54367 m .87696 .54605 L .84898 .54203 L p .014 .852 0 r
F P
0 g
s
.38526 .48772 m .37854 .49707 L .39551 .4993 L p 0 .876 .787 r
F P
0 g
s
.39551 .4993 m .40242 .49065 L .38526 .48772 L p 0 .876 .787 r
F P
0 g
s
.10191 .40218 m .10567 .40451 L .10875 .45652 L p .159 0 .516 r
F P
0 g
s
.11644 .45726 m .10875 .45652 L .10567 .40451 L p .492 0 .603 r
F P
0 g
s
.37559 .50818 m .37679 .51957 L .3925 .52109 L p 0 .691 .405 r
F P
0 g
s
.39679 .53131 m .3925 .52109 L .37679 .51957 L p 0 .488 .138 r
F P
0 g
s
.20178 .22493 m .20718 .23073 L .25058 .20956 L p .759 0 .459 r
F P
0 g
s
.20718 .23073 m .20178 .22493 L .16393 .25814 L p .809 0 .496 r
F P
0 g
s
.3007 .24831 m .30124 .24187 L .26368 .24875 L p .598 0 .176 r
F P
0 g
s
.26368 .24875 m .26459 .25494 L .3007 .24831 L p .598 0 .176 r
F P
0 g
s
.11007 .34991 m .11093 .34803 L .13016 .29928 L p .68 0 .197 r
F P
0 g
s
.11093 .34803 m .11007 .34991 L .10191 .40218 L p .7 0 .09 r
F P
0 g
s
.50658 .28666 m .50869 .28566 L .45915 .24352 L p .104 0 .186 r
F P
0 g
s
.45915 .24352 m .45739 .2456 L .50658 .28666 L p .104 0 .186 r
F P
0 g
s
.39483 .48138 m .38526 .48772 L .40242 .49065 L p 0 .849 .892 r
F P
0 g
s
.40242 .49065 m .41185 .48514 L .39483 .48138 L p 0 .849 .892 r
F P
0 g
s
.3422 .24515 m .3403 .25156 L .3825 .26544 L p .751 0 .141 r
F P
0 g
s
.3403 .25156 m .3422 .24515 L .30124 .24187 L p .682 0 .156 r
F P
0 g
s
.37679 .51957 m .382 .52969 L .39679 .53131 L p 0 .488 .138 r
F P
0 g
s
.40437 .53921 m .39679 .53131 L .382 .52969 L p 0 .199 0 r
F P
0 g
s
.40251 .22174 m .40511 .21545 L .35127 .19766 L p .58 0 .304 r
F P
0 g
s
.35127 .19766 m .35043 .20474 L .40251 .22174 L p .58 0 .304 r
F P
0 g
s
.30124 .24187 m .3007 .24831 L .3403 .25156 L p .682 0 .156 r
F P
0 g
s
.10567 .40451 m .10191 .40218 L .11007 .34991 L p .156 0 .408 r
F P
0 g
s
.56035 .42801 m .55593 .42461 L .50409 .42053 L p .184 .692 0 r
F P
0 g
s
.59005 .39887 m .58488 .40382 L .54781 .34455 L p .904 0 .047 r
F P
0 g
s
.54781 .34455 m .55217 .33971 L .59005 .39887 L p .904 0 .047 r
F P
0 g
s
.40598 .47892 m .39483 .48138 L .41185 .48514 L p 0 .702 .893 r
F P
0 g
s
.41185 .48514 m .42254 .48351 L .40598 .47892 L p 0 .702 .893 r
F P
0 g
s
.24712 .20267 m .25058 .20956 L .29907 .20063 L p .7 0 .414 r
F P
0 g
s
.25058 .20956 m .24712 .20267 L .20178 .22493 L p .759 0 .459 r
F P
0 g
s
.382 .52969 m .39052 .53719 L .40437 .53921 L p 0 .199 0 r
F P
0 g
s
.41421 .54372 m .40437 .53921 L .39052 .53719 L p 0 0 0 r
F P
0 g
s
.55336 .33703 m .55217 .33971 L .50789 .28761 L p .82 0 0 r
F P
0 g
s
.50789 .28761 m .50869 .28566 L .55336 .33703 L p .82 0 0 r
F P
0 g
s
.35043 .20474 m .35127 .19766 L .29781 .19329 L p .639 0 .361 r
F P
0 g
s
.29781 .19329 m .29907 .20063 L .35043 .20474 L p .639 0 .361 r
F P
0 g
s
.15948 .25641 m .16393 .25814 L .20178 .22493 L p .831 0 .515 r
F P
0 g
s
.16393 .25814 m .15948 .25641 L .13016 .29928 L p .902 0 .523 r
F P
0 g
s
.89105 .48923 m .89853 .49641 L .86923 .49131 L p .024 .31 0 r
F P
0 g
s
.86923 .49131 m .86308 .48431 L .89105 .48923 L p .024 .31 0 r
F P
0 g
s
.29907 .20063 m .29781 .19329 L .24712 .20267 L p .7 0 .414 r
F P
0 g
s
.13455 .40928 m .14557 .40938 L .14859 .4534 L p .798 0 .609 r
F P
0 g
s
.14859 .4534 m .13785 .45558 L .13455 .40928 L p .798 0 .609 r
F P
0 g
s
.20247 .2972 m .20527 .29737 L .18395 .32867 L p .853 0 .451 r
F P
0 g
s
.78146 .32857 m .78533 .31999 L .77034 .35246 L p 0 0 0 r
F P
0 g
s
.77034 .35246 m .7668 .35946 L .78146 .32857 L p 0 0 0 r
F P
0 g
s
.41346 .48099 m .42254 .48351 L p .4172 .48067 L 0 .411 .739 r
F P
0 g
s
.42254 .48351 m .43305 .486 L .4172 .48067 L p 0 .411 .739 r
F P
0 g
s
.42698 .48639 m .4172 .48067 L .42364 .48283 L p 0 0 0 r
F P
0 g
s
.42364 .48283 m .43305 .486 L p .42698 .48639 L 0 0 0 r
F P
0 g
s
.434 .49534 m .42698 .48639 L .44196 .49227 L p 0 .318 0 r
F P
0 g
s
.42698 .48639 m .434 .49534 L .4309 .49057 L p 0 0 .473 r
F P
0 g
s
.43305 .486 m .44196 .49227 L .42698 .48639 L p 0 0 0 r
F P
0 g
s
.4309 .49057 m .42771 .48566 L p .42698 .48639 L 0 0 .473 r
F P
0 g
s
.4172 .48067 m .42698 .48639 L .42237 .48022 L p 0 .267 .754 r
F P
0 g
s
.42237 .48022 m .41992 .47693 L p .4172 .48067 L 0 .267 .754 r
F P
0 g
s
.4172 .48067 m .40598 .47892 L .41346 .48099 L p 0 .411 .739 r
F P
0 g
s
.39052 .53719 m .4012 .54103 L .41421 .54372 L p 0 0 0 r
F P
0 g
s
.425 .54422 m .41421 .54372 L .4012 .54103 L p 0 .525 .77 r
F P
0 g
s
.89533 .53747 m .8868 .54367 L .85747 .53936 L p .023 .913 0 r
F P
0 g
s
.84898 .54203 m .85747 .53936 L .8868 .54367 L p .014 .852 0 r
F P
0 g
s
.10567 .40451 m .1131 .40667 L .11644 .45726 L p .492 0 .603 r
F P
0 g
s
.12661 .45693 m .11644 .45726 L .1131 .40667 L p .635 0 .618 r
F P
0 g
s
.38089 .26834 m .3825 .26544 L .36449 .25951 L p .657 0 0 r
F P
0 g
s
.38089 .26834 m .38097 .26749 L .39721 .27712 L p 0 0 .205 r
F P
0 g
s
.39721 .27712 m .42421 .29313 L p .38089 .26834 L 0 0 .205 r
F P
0 g
s
.36449 .25951 m .3403 .25156 L p .38089 .26834 L .657 0 0 r
F P
0 g
s
.45739 .2456 m .45915 .24352 L .40615 .21248 L p .229 0 .255 r
F P
0 g
s
.40615 .21248 m .40511 .21545 L .45739 .2456 L p .229 0 .255 r
F P
0 g
s
.54432 .48465 m .5492 .48576 L .49613 .48046 L p .085 .659 0 r
F P
0 g
s
.82812 .50496 m .83219 .49467 L .80115 .48914 L p .075 .748 0 r
F P
0 g
s
.80115 .48914 m .79757 .49967 L .82812 .50496 L p .075 .748 0 r
F P
0 g
s
.82697 .5161 m .82812 .50496 L .79757 .49967 L p .077 .685 0 r
F P
0 g
s
.17493 .32785 m .18112 .32928 L .16697 .36672 L p .925 0 .538 r
F P
0 g
s
.16697 .36672 m .16066 .36679 L .17493 .32785 L p .925 0 .538 r
F P
0 g
s
.15147 .36554 m .16066 .36679 L .15474 .40861 L p .863 0 .569 r
F P
0 g
s
.15474 .40861 m .14557 .40938 L .15147 .36554 L p .863 0 .569 r
F P
0 g
s
.82891 .52663 m .82697 .5161 L .79614 .51084 L p .077 .601 0 r
F P
0 g
s
.79757 .49967 m .79614 .51084 L .82697 .5161 L p .077 .685 0 r
F P
0 g
s
.4012 .54103 m .41261 .5407 L .425 .54422 L p 0 .525 .77 r
F P
0 g
s
.43527 .54063 m .425 .54422 L .41261 .5407 L p 0 .77 .868 r
F P
0 g
s
.12892 .30075 m .13016 .29928 L .15948 .25641 L p .622 0 .28 r
F P
0 g
s
.13016 .29928 m .12892 .30075 L .11007 .34991 L p .68 0 .197 r
F P
0 g
s
.12319 .40833 m .13455 .40928 L .13785 .45558 L p .724 0 .618 r
F P
0 g
s
.13785 .45558 m .12661 .45693 L .12319 .40833 L p .724 0 .618 r
F P
0 g
s
.3007 .24831 m .26459 .25494 L p .29951 .25103 L .326 0 0 r
F P
0 g
s
.26459 .25494 m .26406 .25758 L .29978 .251 L p .326 0 0 r
F P
0 g
s
.26215 .25628 m .29859 .24953 L p .2643 .25752 L .414 0 .423 r
F P
0 g
s
.26406 .25758 m .26405 .25757 L p .2643 .25752 L .414 0 .423 r
F P
0 g
s
.26405 .25757 m .26215 .25628 L p .2643 .25752 L .414 0 .423 r
F P
0 g
s
.29978 .251 m .29978 .25098 L p .29951 .25103 L .326 0 0 r
F P
0 g
s
.29978 .25098 m .3007 .24831 L p .29951 .25103 L .326 0 0 r
F P
0 g
s
.19809 .22216 m .20178 .22493 L .24712 .20267 L p .735 0 .491 r
F P
0 g
s
.20178 .22493 m .19809 .22216 L .15948 .25641 L p .831 0 .515 r
F P
0 g
s
.3403 .25156 m .33902 .25432 L .38089 .26834 L p .657 0 0 r
F P
0 g
s
.33902 .25432 m .3403 .25156 L .32355 .25018 L p .507 0 0 r
F P
0 g
s
.33902 .25432 m .33854 .25302 L .35444 .25844 L p 0 0 .284 r
F P
0 g
s
.35444 .25844 m .38097 .26749 L p .33902 .25432 L 0 0 .284 r
F P
0 g
s
.32355 .25018 m .3007 .24831 L p .33902 .25432 L .507 0 0 r
F P
0 g
s
.1131 .40667 m .12319 .40833 L .12661 .45693 L p .635 0 .618 r
F P
0 g
s
.83219 .49467 m .83866 .4866 L .80644 .48064 L p .07 .802 0 r
F P
0 g
s
.80644 .48064 m .80115 .48914 L .83219 .49467 L p .07 .802 0 r
F P
0 g
s
.89853 .49641 m .90309 .50624 L .87278 .5011 L p .027 .565 0 r
F P
0 g
s
.87278 .5011 m .86923 .49131 L .89853 .49641 L p .027 .565 0 r
F P
0 g
s
.3007 .24831 m .29978 .251 L .33902 .25432 L p .507 0 0 r
F P
0 g
s
.44196 .49227 m .44806 .5015 L .434 .49534 L p 0 .318 0 r
F P
0 g
s
.56506 .44468 m .56475 .4557 L .61658 .4599 L p .239 .53 0 r
F P
0 g
s
.41261 .5407 m .42319 .53623 L .43527 .54063 L p 0 .77 .868 r
F P
0 g
s
.44365 .53343 m .43527 .54063 L .42319 .53623 L p 0 .87 .82 r
F P
0 g
s
.90138 .52829 m .89533 .53747 L .865 .53287 L p .026 .871 0 r
F P
0 g
s
.85747 .53936 m .865 .53287 L .89533 .53747 L p .023 .913 0 r
F P
0 g
s
.58488 .40382 m .57702 .40965 L .54086 .35088 L p .83 0 .059 r
F P
0 g
s
.54086 .35088 m .54781 .34455 L .58488 .40382 L p .83 0 .059 r
F P
0 g
s
.40511 .21545 m .40615 .21248 L .35128 .19409 L p .352 0 .326 r
F P
0 g
s
.35128 .19409 m .35127 .19766 L .40511 .21545 L p .352 0 .326 r
F P
0 g
s
.11007 .34991 m .11326 .35279 L .10567 .40451 L p .156 0 .408 r
F P
0 g
s
.1131 .40667 m .10567 .40451 L .11326 .35279 L p .486 0 .524 r
F P
0 g
s
.55593 .42461 m .55084 .42533 L .49779 .41968 L p 0 .566 0 r
F P
0 g
s
.49154 .42295 m .49779 .41968 L .55084 .42533 L p .204 0 0 r
F P
0 g
s
.11326 .35279 m .11007 .34991 L .12892 .30075 L p .162 0 .306 r
F P
0 g
s
.56264 .46665 m .55901 .47608 L .61168 .48125 L p .246 .402 0 r
F P
0 g
s
.4238 .29329 m .42421 .29313 L .46629 .32921 L p 0 0 .131 r
F P
0 g
s
.50869 .28566 m .50789 .28761 L .45852 .24483 L p .7 0 0 r
F P
0 g
s
.45852 .24483 m .45915 .24352 L .50869 .28566 L p .7 0 0 r
F P
0 g
s
.83368 .53517 m .82891 .52663 L .79708 .5212 L p .074 .459 0 r
F P
0 g
s
.79614 .51084 m .79708 .5212 L .82891 .52663 L p .077 .601 0 r
F P
0 g
s
.22986 .27245 m .23234 .27328 L .20527 .29737 L p .753 0 .47 r
F P
0 g
s
.43731 .50629 m .434 .49534 L .44806 .5015 L p 0 .582 .222 r
F P
0 g
s
.44806 .5015 m .45051 .51243 L .43731 .50629 L p 0 .582 .222 r
F P
0 g
s
.24454 .19917 m .24712 .20267 L .29781 .19329 L p .615 0 .45 r
F P
0 g
s
.24712 .20267 m .24454 .19917 L .19809 .22216 L p .735 0 .491 r
F P
0 g
s
.42319 .53623 m .43151 .52822 L .44365 .53343 L p 0 .87 .82 r
F P
0 g
s
.44896 .52358 m .44365 .53343 L .43151 .52822 L p 0 .856 .68 r
F P
0 g
s
.7668 .35946 m .75986 .36413 L .774 .33502 L p .128 0 .813 r
F P
0 g
s
.774 .33502 m .78146 .32857 L .7668 .35946 L p .128 0 .813 r
F P
0 g
s
.61658 .4599 m .61727 .44918 L .56506 .44468 L p .239 .53 0 r
F P
0 g
s
.56354 .43502 m .56506 .44468 L .61727 .44918 L p .228 .588 0 r
F P
0 g
s
.43644 .51777 m .43731 .50629 L .45051 .51243 L p 0 .758 .479 r
F P
0 g
s
.45051 .51243 m .44896 .52358 L .43644 .51777 L p 0 .758 .479 r
F P
0 g
s
.43151 .52822 m .43644 .51777 L .44896 .52358 L p 0 .856 .68 r
F P
0 g
s
.35127 .19766 m .35128 .19409 L .29654 .18952 L p .483 0 .393 r
F P
0 g
s
.29654 .18952 m .29781 .19329 L .35127 .19766 L p .483 0 .393 r
F P
0 g
s
.71514 .34425 m .71304 .33557 L .72441 .30996 L p .008 0 .542 r
F P
0 g
s
.90309 .50624 m .90409 .51738 L .87323 .51232 L p .028 .707 0 r
F P
0 g
s
.87323 .51232 m .87278 .5011 L .90309 .50624 L p .028 .707 0 r
F P
0 g
s
.29781 .19329 m .29654 .18952 L .24454 .19917 L p .615 0 .45 r
F P
0 g
s
.51501 .36984 m .50875 .3732 L .46961 .32694 L p .511 0 .128 r
F P
0 g
s
.46629 .32921 m .46961 .32694 L .50875 .3732 L p .227 0 .13 r
F P
0 g
s
.55217 .33971 m .54781 .34455 L .50426 .29228 L p .888 0 .068 r
F P
0 g
s
.50426 .29228 m .50789 .28761 L .55217 .33971 L p .888 0 .068 r
F P
0 g
s
.90409 .51738 m .90138 .52829 L .87051 .52342 L p .027 .801 0 r
F P
0 g
s
.865 .53287 m .87051 .52342 L .90138 .52829 L p .026 .871 0 r
F P
0 g
s
.56756 .41554 m .55782 .42068 L .52323 .36448 L p .696 0 .069 r
F P
0 g
s
.51501 .36984 m .52323 .36448 L .55782 .42068 L p .61 0 .073 r
F P
0 g
s
.83866 .4866 m .84668 .48182 L .81275 .47529 L p .06 .84 0 r
F P
0 g
s
.81275 .47529 m .80644 .48064 L .83866 .4866 L p .06 .84 0 r
F P
0 g
s
.15803 .25746 m .15948 .25641 L .19809 .22216 L p .505 0 .326 r
F P
0 g
s
.15948 .25641 m .15803 .25746 L .12892 .30075 L p .622 0 .28 r
F P
0 g
s
.87051 .52342 m .87323 .51232 L .90409 .51738 L p .027 .801 0 r
F P
0 g
s
.14063 .36313 m .15147 .36554 L .14557 .40938 L p .789 0 .574 r
F P
0 g
s
.14557 .40938 m .13455 .40928 L .14063 .36313 L p .789 0 .574 r
F P
0 g
s
.57702 .40965 m .56756 .41554 L .53226 .35785 L p .763 0 .065 r
F P
0 g
s
.53226 .35785 m .54086 .35088 L .57702 .40965 L p .763 0 .065 r
F P
0 g
s
.19683 .29429 m .20247 .2972 L .18112 .32928 L p .885 0 .529 r
F P
0 g
s
.18112 .32928 m .17493 .32785 L .19683 .29429 L p .885 0 .529 r
F P
0 g
s
.54034 .47952 m .54432 .48465 L .49006 .47779 L p .207 .648 0 r
F P
0 g
s
.49006 .47779 m .4851 .47133 L .54034 .47952 L p .207 .648 0 r
F P
0 g
s
.6081 .48909 m .61168 .48125 L .55901 .47608 L p .242 .29 0 r
F P
0 g
s
.55901 .47608 m .55433 .48276 L .6081 .48909 L p .242 .29 0 r
F P
0 g
s
.26215 .25628 m .26406 .25758 L .23234 .27328 L p .604 0 .461 r
F P
0 g
s
.38097 .26749 m .38089 .26834 L .33902 .25432 L p 0 0 .284 r
F P
0 g
s
.52323 .36448 m .53226 .35785 L .56756 .41554 L p .696 0 .069 r
F P
0 g
s
.45915 .24352 m .45852 .24483 L .40548 .21328 L p .55 0 0 r
F P
0 g
s
.40548 .21328 m .40615 .21248 L .45915 .24352 L p .55 0 0 r
F P
0 g
s
.11326 .35279 m .1201 .35628 L .1131 .40667 L p .486 0 .524 r
F P
0 g
s
.12319 .40833 m .1131 .40667 L .1201 .35628 L p .628 0 .557 r
F P
0 g
s
.16618 .32455 m .17493 .32785 L .16066 .36679 L p .846 0 .545 r
F P
0 g
s
.16066 .36679 m .15147 .36554 L .16618 .32455 L p .846 0 .545 r
F P
0 g
s
.84067 .54055 m .83368 .53517 L .80025 .52936 L p .061 .146 0 r
F P
0 g
s
.79708 .5212 m .80025 .52936 L .83368 .53517 L p .074 .459 0 r
F P
0 g
s
.61727 .44918 m .6166 .44034 L .56354 .43502 L p .228 .588 0 r
F P
0 g
s
.56035 .42801 m .56354 .43502 L .6166 .44034 L p .203 .655 0 r
F P
0 g
s
.75986 .36413 m .75045 .36581 L .76395 .33844 L p .117 0 .933 r
F P
0 g
s
.76395 .33844 m .774 .33502 L .75986 .36413 L p .117 0 .933 r
F P
0 g
s
.29859 .24953 m .29978 .251 L .26406 .25758 L p .414 0 .423 r
F P
0 g
s
.13143 .30408 m .12892 .30075 L .15803 .25746 L p .192 0 .216 r
F P
0 g
s
.12892 .30075 m .13143 .30408 L .11326 .35279 L p .162 0 .306 r
F P
0 g
s
.72088 .35267 m .71514 .34425 L .72657 .31954 L p .046 0 .744 r
F P
0 g
s
.72441 .30996 m .72657 .31954 L .71514 .34425 L p .008 0 .542 r
F P
0 g
s
.19661 .22283 m .19809 .22216 L .24444 .19922 L p .324 0 .332 r
F P
0 g
s
.19661 .22283 m .24444 .19922 L .24444 .19922 L .324 0 .332 r
F
.24454 .19917 m .2432 .19959 L .19661 .22283 L p F P
0 g
s
.29654 .18952 m .29545 .18987 L .24379 .19948 L p 0 0 0 r
F P
0 g
s
.24379 .19948 m .2432 .19959 L p .24379 .19948 L 0 0 0 r
F P
0 g
s
.24379 .19948 m .29654 .18952 L .24379 .19948 L 0 0 0 r
F
.24444 .19922 m .24454 .19917 L p .24444 .19922 L .324 0 .332 r
F P
0 g
s
.19809 .22216 m .19661 .22283 L .15803 .25746 L p .505 0 .326 r
F P
0 g
s
.33854 .25302 m .33902 .25432 L .29978 .251 L p .202 0 .36 r
F P
0 g
s
.55084 .42533 m .54577 .43011 L .49154 .42295 L p .204 0 0 r
F P
0 g
s
.48621 .42992 m .49154 .42295 L .54577 .43011 L p .251 .229 0 r
F P
0 g
s
.12966 .35989 m .14063 .36313 L .13455 .40928 L p .716 0 .57 r
F P
0 g
s
.13455 .40928 m .12319 .40833 L .12966 .35989 L p .716 0 .57 r
F P
0 g
s
.29978 .251 m .29859 .24953 L .33854 .25302 L p .202 0 .36 r
F P
0 g
s
.1201 .35628 m .11326 .35279 L .13143 .30408 L p .487 0 .448 r
F P
0 g
s
.1201 .35628 m .12966 .35989 L .12319 .40833 L p .628 0 .557 r
F P
0 g
s
.40615 .21248 m .40548 .21328 L .35043 .19456 L p .367 0 0 r
F P
0 g
s
.35043 .19456 m .35128 .19409 L .40615 .21248 L p .367 0 0 r
F P
0 g
s
.84668 .48182 m .85521 .481 L .81924 .47385 L p .029 .759 0 r
F P
0 g
s
.81924 .47385 m .81275 .47529 L .84668 .48182 L p .029 .759 0 r
F P
0 g
s
.50789 .28761 m .50426 .29228 L .45555 .24938 L p .856 0 .08 r
F P
0 g
s
.45555 .24938 m .45852 .24483 L .50789 .28761 L p .856 0 .08 r
F P
0 g
s
.52323 .36448 m .51501 .36984 L .47534 .32199 L p .632 0 .121 r
F P
0 g
s
.46961 .32694 m .47534 .32199 L .51501 .36984 L p .511 0 .128 r
F P
0 g
s
.54781 .34455 m .54086 .35088 L .49828 .29906 L p .827 0 .093 r
F P
0 g
s
.49828 .29906 m .50426 .29228 L .54781 .34455 L p .827 0 .093 r
F P
0 g
s
.42421 .29313 m .42697 .28962 L .46961 .32694 L p .319 0 .191 r
F P
0 g
s
.42697 .28962 m .42421 .29313 L .38097 .26749 L p .402 0 .255 r
F P
0 g
s
.22529 .26824 m .22986 .27245 L .20247 .2972 L p .831 0 .51 r
F P
0 g
s
.20247 .2972 m .19683 .29429 L .22529 .26824 L p .831 0 .51 r
F P
0 g
s
.2432 .19959 m .24454 .19917 L .29654 .18952 L p 0 0 0 r
F P
0 g
s
.60437 .49354 m .6081 .48909 L .55433 .48276 L p .208 .037 0 r
F P
0 g
s
.75045 .36581 m .73985 .36425 L .75267 .33837 L p .101 0 .932 r
F P
0 g
s
.75267 .33837 m .76395 .33844 L .75045 .36581 L p .101 0 .932 r
F P
0 g
s
.35128 .19409 m .35043 .19456 L .29545 .18987 L p .148 0 0 r
F P
0 g
s
.29545 .18987 m .29654 .18952 L .35128 .19409 L p .148 0 0 r
F P
0 g
s
.53784 .47102 m .54034 .47952 L .4851 .47133 L p .241 .572 0 r
F P
0 g
s
.4851 .47133 m .48195 .46192 L .53784 .47102 L p .241 .572 0 r
F P
0 g
s
.72657 .31954 m .73261 .32821 L .72088 .35267 L p .046 0 .744 r
F P
0 g
s
.72952 .35966 m .72088 .35267 L .73261 .32821 L p .069 0 .841 r
F P
0 g
s
.47534 .32199 m .46961 .32694 L .42697 .28962 L p .55 0 .181 r
F P
0 g
s
.15606 .31982 m .16618 .32455 L .15147 .36554 L p .776 0 .536 r
F P
0 g
s
.15147 .36554 m .14063 .36313 L .15606 .31982 L p .776 0 .536 r
F P
0 g
s
.80025 .52936 m .80527 .53425 L .84067 .54055 L p .061 .146 0 r
F P
0 g
s
.84898 .54203 m .84067 .54055 L .80527 .53425 L p 0 .501 0 r
F P
0 g
s
.1598 .26118 m .15803 .25746 L .19661 .22283 L p .261 0 .144 r
F P
0 g
s
.15803 .25746 m .1598 .26118 L .13143 .30408 L p .192 0 .216 r
F P
0 g
s
.18911 .289 m .19683 .29429 L .17493 .32785 L p .822 0 .516 r
F P
0 g
s
.17493 .32785 m .16618 .32455 L .18911 .289 L p .822 0 .516 r
F P
0 g
s
.6166 .44034 m .61466 .43454 L .56035 .42801 L p .203 .655 0 r
F P
0 g
s
.55593 .42461 m .56035 .42801 L .61466 .43454 L p .117 .691 0 r
F P
0 g
s
.73985 .36425 m .72952 .35966 L .74172 .33479 L p .086 0 .898 r
F P
0 g
s
.74172 .33479 m .75267 .33837 L .73985 .36425 L p .086 0 .898 r
F P
0 g
s
.54577 .43011 m .54142 .43834 L .48621 .42992 L p .251 .229 0 r
F P
0 g
s
.48254 .43967 m .48621 .42992 L .54142 .43834 L p .259 .356 0 r
F P
0 g
s
.73261 .32821 m .74172 .33479 L .72952 .35966 L p .069 0 .841 r
F P
0 g
s
.25911 .25116 m .26215 .25628 L .22986 .27245 L p .759 0 .48 r
F P
0 g
s
.22986 .27245 m .22529 .26824 L .25911 .25116 L p .759 0 .48 r
F P
0 g
s
.13143 .30408 m .13739 .30882 L .1201 .35628 L p .487 0 .448 r
F P
0 g
s
.12966 .35989 m .1201 .35628 L .13739 .30882 L p .624 0 .496 r
F P
0 g
s
.38097 .26749 m .38274 .26295 L .42697 .28962 L p .402 0 .255 r
F P
0 g
s
.38274 .26295 m .38097 .26749 L .33854 .25302 L p .486 0 .319 r
F P
0 g
s
.53226 .35785 m .52323 .36448 L .48273 .31503 L p .707 0 .113 r
F P
0 g
s
.47534 .32199 m .48273 .31503 L .52323 .36448 L p .632 0 .121 r
F P
0 g
s
.54086 .35088 m .53226 .35785 L .49075 .30701 L p .767 0 .105 r
F P
0 g
s
.49075 .30701 m .49828 .29906 L .54086 .35088 L p .767 0 .105 r
F P
0 g
s
.79757 .49967 m .80115 .48914 L .7627 .48214 L p .116 .682 0 r
F P
0 g
s
.7627 .48214 m .75969 .49293 L .79757 .49967 L p .116 .682 0 r
F P
0 g
s
.79614 .51084 m .79757 .49967 L .75969 .49293 L p .119 .627 0 r
F P
0 g
s
.45852 .24483 m .45555 .24938 L .40316 .21778 L p .804 0 .083 r
F P
0 g
s
.40316 .21778 m .40548 .21328 L .45852 .24483 L p .804 0 .083 r
F P
0 g
s
.13739 .30882 m .13143 .30408 L .1598 .26118 L p .5 0 .377 r
F P
0 g
s
.53716 .46032 m .53784 .47102 L .48195 .46192 L p .254 .504 0 r
F P
0 g
s
.48195 .46192 m .48105 .45086 L .53716 .46032 L p .254 .504 0 r
F P
0 g
s
.79708 .5212 m .79614 .51084 L .75798 .50405 L p .12 .553 0 r
F P
0 g
s
.75969 .49293 m .75798 .50405 L .79614 .51084 L p .119 .627 0 r
F P
0 g
s
.48273 .31503 m .49075 .30701 L .53226 .35785 L p .707 0 .113 r
F P
0 g
s
.146 .31432 m .15606 .31982 L .14063 .36313 L p .708 0 .521 r
F P
0 g
s
.14063 .36313 m .12966 .35989 L .146 .31432 L p .708 0 .521 r
F P
0 g
s
.29729 .24406 m .29859 .24953 L .26215 .25628 L p .671 0 .437 r
F P
0 g
s
.26215 .25628 m .25911 .25116 L .29729 .24406 L p .671 0 .437 r
F P
0 g
s
.85521 .481 m .86308 .48431 L .82505 .47655 L p .037 0 0 r
F P
0 g
s
.82505 .47655 m .81924 .47385 L .85521 .481 L p .037 0 0 r
F P
0 g
s
.19759 .2269 m .19661 .22283 L .2432 .19959 L p .368 0 .099 r
F P
0 g
s
.19661 .22283 m .19759 .2269 L .1598 .26118 L p .261 0 .144 r
F P
0 g
s
.33854 .25302 m .33892 .24777 L .38274 .26295 L p .486 0 .319 r
F P
0 g
s
.33892 .24777 m .33854 .25302 L .29859 .24953 L p .577 0 .382 r
F P
0 g
s
.13739 .30882 m .146 .31432 L .12966 .35989 L p .624 0 .496 r
F P
0 g
s
.80115 .48914 m .80644 .48064 L .76662 .47308 L p .111 .732 0 r
F P
0 g
s
.76662 .47308 m .7627 .48214 L .80115 .48914 L p .111 .732 0 r
F P
0 g
s
.54142 .43834 m .53841 .4489 L .48254 .43967 L p .259 .356 0 r
F P
0 g
s
.48105 .45086 m .48254 .43967 L .53841 .4489 L p .259 .437 0 r
F P
0 g
s
.50426 .29228 m .49828 .29906 L .45063 .25658 L p .813 0 .122 r
F P
0 g
s
.45063 .25658 m .45555 .24938 L .50426 .29228 L p .813 0 .122 r
F P
0 g
s
.29859 .24953 m .29729 .24406 L .33892 .24777 L p .577 0 .382 r
F P
0 g
s
.66635 .47764 m .66844 .46646 L .61658 .4599 L p .218 .527 0 r
F P
0 g
s
.61658 .4599 m .61463 .47107 L .66635 .47764 L p .218 .527 0 r
F P
0 g
s
.61727 .44918 m .61658 .4599 L .66844 .46646 L p .211 .578 0 r
F P
0 g
s
.66392 .48832 m .66635 .47764 L .61463 .47107 L p .222 .467 0 r
F P
0 g
s
.61463 .47107 m .61168 .48125 L .66392 .48832 L p .222 .467 0 r
F P
0 g
s
.53841 .4489 m .53716 .46032 L .48105 .45086 L p .259 .437 0 r
F P
0 g
s
.75188 .28671 m .76305 .28855 L .76822 .2612 L p .052 0 .959 r
F P
0 g
s
.76305 .28855 m .77315 .29362 L .77874 .26571 L p .039 0 .919 r
F P
0 g
s
.77874 .26571 m .76822 .2612 L .76305 .28855 L p .039 0 .919 r
F P
0 g
s
.76822 .2612 m .75662 .26043 L .75188 .28671 L p .052 0 .959 r
F P
0 g
s
.40548 .21328 m .40316 .21778 L .34883 .19907 L p .726 0 .081 r
F P
0 g
s
.34883 .19907 m .35043 .19456 L .40548 .21328 L p .726 0 .081 r
F P
0 g
s
.21925 .2612 m .22529 .26824 L .19683 .29429 L p .791 0 .482 r
F P
0 g
s
.19683 .29429 m .18911 .289 L .21925 .2612 L p .791 0 .482 r
F P
0 g
s
.42697 .28962 m .43172 .28321 L .47534 .32199 L p .55 0 .181 r
F P
0 g
s
.43172 .28321 m .42697 .28962 L .38274 .26295 L p .585 0 .236 r
F P
0 g
s
.48273 .31503 m .47534 .32199 L .43172 .28321 L p .65 0 .169 r
F P
0 g
s
.24331 .20392 m .2432 .19959 L .29545 .18987 L p .497 0 .079 r
F P
0 g
s
.2432 .19959 m .24331 .20392 L .19759 .2269 L p .368 0 .099 r
F P
0 g
s
.77315 .29362 m .78085 .30126 L .78676 .27336 L p .023 0 .83 r
F P
0 g
s
.78676 .27336 m .77874 .26571 L .77315 .29362 L p .023 0 .83 r
F P
0 g
s
.18036 .28204 m .18911 .289 L .16618 .32455 L p .762 0 .496 r
F P
0 g
s
.16618 .32455 m .15606 .31982 L .18036 .28204 L p .762 0 .496 r
F P
0 g
s
.80025 .52936 m .79708 .5212 L .75779 .51406 L p .117 .426 0 r
F P
0 g
s
.75798 .50405 m .75779 .51406 L .79708 .5212 L p .12 .553 0 r
F P
0 g
s
.85747 .53936 m .84898 .54203 L .81149 .53514 L p .047 .832 0 r
F P
0 g
s
.80527 .53425 m .81149 .53514 L .84898 .54203 L p 0 .501 0 r
F P
0 g
s
.35043 .19456 m .34883 .19907 L .29467 .19434 L p .622 0 .077 r
F P
0 g
s
.29467 .19434 m .29545 .18987 L .35043 .19456 L p .622 0 .077 r
F P
0 g
s
.29545 .18987 m .29467 .19434 L .24331 .20392 L p .497 0 .079 r
F P
0 g
s
.1646 .26709 m .1598 .26118 L .19759 .2269 L p .53 0 .314 r
F P
0 g
s
.1598 .26118 m .1646 .26709 L .13739 .30882 L p .5 0 .377 r
F P
0 g
s
.66844 .46646 m .6699 .45623 L .61727 .44918 L p .211 .578 0 r
F P
0 g
s
.6166 .44034 m .61727 .44918 L .6699 .45623 L p .199 .633 0 r
F P
0 g
s
.66145 .49709 m .66392 .48832 L .61168 .48125 L p .222 .379 0 r
F P
0 g
s
.61168 .48125 m .6081 .48909 L .66145 .49709 L p .222 .379 0 r
F P
0 g
s
.146 .31432 m .13739 .30882 L .1646 .26709 L p .626 0 .437 r
F P
0 g
s
.61466 .43454 m .61168 .4326 L .55593 .42461 L p .117 .691 0 r
F P
0 g
s
.55084 .42533 m .55593 .42461 L .61168 .4326 L p .12 0 0 r
F P
0 g
s
.80644 .48064 m .81275 .47529 L .77093 .46697 L p .098 .779 0 r
F P
0 g
s
.77093 .46697 m .76662 .47308 L .80644 .48064 L p .098 .779 0 r
F P
0 g
s
.78085 .30126 m .78511 .31046 L .79119 .28314 L p 0 0 .64 r
F P
0 g
s
.79119 .28314 m .78676 .27336 L .78085 .30126 L p 0 0 .64 r
F P
0 g
s
.49828 .29906 m .49075 .30701 L .44443 .26546 L p .764 0 .143 r
F P
0 g
s
.44443 .26546 m .45063 .25658 L .49828 .29906 L p .764 0 .143 r
F P
0 g
s
.25532 .24288 m .25911 .25116 L .22529 .26824 L p .753 0 .443 r
F P
0 g
s
.22529 .26824 m .21925 .2612 L .25532 .24288 L p .753 0 .443 r
F P
0 g
s
.45555 .24938 m .45063 .25658 L .3995 .22541 L p .789 0 .148 r
F P
0 g
s
.3995 .22541 m .40316 .21778 L .45555 .24938 L p .789 0 .148 r
F P
0 g
s
.38274 .26295 m .38598 .2553 L .43172 .28321 L p .585 0 .236 r
F P
0 g
s
.38598 .2553 m .38274 .26295 L .33892 .24777 L p .622 0 .291 r
F P
0 g
s
.17179 .27438 m .18036 .28204 L .15606 .31982 L p .701 0 .471 r
F P
0 g
s
.15606 .31982 m .146 .31432 L .17179 .27438 L p .701 0 .471 r
F P
0 g
s
.49075 .30701 m .48273 .31503 L .43782 .27477 L p .713 0 .157 r
F P
0 g
s
.43172 .28321 m .43782 .27477 L .48273 .31503 L p .65 0 .169 r
F P
0 g
s
.86308 .48431 m .86923 .49131 L .82936 .48309 L p .07 .321 0 r
F P
0 g
s
.82936 .48309 m .82505 .47655 L .86308 .48431 L p .07 .321 0 r
F P
0 g
s
.1646 .26709 m .17179 .27438 L .146 .31432 L p .626 0 .437 r
F P
0 g
s
.43782 .27477 m .44443 .26546 L .49075 .30701 L p .713 0 .157 r
F P
0 g
s
.20091 .23384 m .19759 .2269 L .24331 .20392 L p .579 0 .263 r
F P
0 g
s
.19759 .2269 m .20091 .23384 L .1646 .26709 L p .53 0 .314 r
F P
0 g
s
.29606 .23528 m .29729 .24406 L .25911 .25116 L p .709 0 .397 r
F P
0 g
s
.25911 .25116 m .25532 .24288 L .29606 .23528 L p .709 0 .397 r
F P
0 g
s
.21256 .25228 m .21925 .2612 L .18911 .289 L p .746 0 .453 r
F P
0 g
s
.18911 .289 m .18036 .28204 L .21256 .25228 L p .746 0 .453 r
F P
0 g
s
.33892 .24777 m .34014 .23926 L .38598 .2553 L p .622 0 .291 r
F P
0 g
s
.34014 .23926 m .33892 .24777 L .29729 .24406 L p .664 0 .345 r
F P
0 g
s
.43782 .27477 m .43172 .28321 L .38598 .2553 L p .665 0 .217 r
F P
0 g
s
.80527 .53425 m .80025 .52936 L .75915 .52162 L p .096 .116 0 r
F P
0 g
s
.75779 .51406 m .75915 .52162 L .80025 .52936 L p .117 .426 0 r
F P
0 g
s
.29729 .24406 m .29606 .23528 L .34014 .23926 L p .664 0 .345 r
F P
0 g
s
.75969 .49293 m .7627 .48214 L .71789 .47426 L p .154 .627 0 r
F P
0 g
s
.71789 .47426 m .71542 .48529 L .75969 .49293 L p .154 .627 0 r
F P
0 g
s
.75798 .50405 m .75969 .49293 L .71542 .48529 L p .157 .574 0 r
F P
0 g
s
.78511 .31046 m .78533 .31999 L .79142 .29374 L p 0 0 .224 r
F P
0 g
s
.79142 .29374 m .79119 .28314 L .78511 .31046 L p 0 0 .224 r
F P
0 g
s
.40316 .21778 m .3995 .22541 L .34669 .20702 L p .751 0 .171 r
F P
0 g
s
.34669 .20702 m .34883 .19907 L .40316 .21778 L p .751 0 .171 r
F P
0 g
s
.865 .53287 m .85747 .53936 L .81806 .53189 L p .066 .831 0 r
F P
0 g
s
.81149 .53514 m .81806 .53189 L .85747 .53936 L p .047 .832 0 r
F P
0 g
s
.17179 .27438 m .1646 .26709 L .20091 .23384 L p .637 0 .382 r
F P
0 g
s
.6699 .45623 m .67054 .4483 L .6166 .44034 L p .199 .633 0 r
F P
0 g
s
.61466 .43454 m .6166 .44034 L .67054 .4483 L p .166 .692 0 r
F P
0 g
s
.65928 .5028 m .66145 .49709 L .6081 .48909 L p .208 .199 0 r
F P
0 g
s
.6081 .48909 m .60437 .49354 L .65928 .5028 L p .208 .199 0 r
F P
0 g
s
.59839 .49026 m .60098 .49397 L .54432 .48465 L p .162 .679 0 r
F P
0 g
s
.54432 .48465 m .54034 .47952 L .59839 .49026 L p .162 .679 0 r
F P
0 g
s
.24487 .21162 m .24331 .20392 L .29467 .19434 L p .639 0 .224 r
F P
0 g
s
.24331 .20392 m .24487 .21162 L .20091 .23384 L p .579 0 .263 r
F P
0 g
s
.75779 .51406 m .75798 .50405 L .71338 .49627 L p .159 .503 0 r
F P
0 g
s
.71542 .48529 m .71338 .49627 L .75798 .50405 L p .157 .574 0 r
F P
0 g
s
.66844 .46646 m .66635 .47764 L .71542 .48529 L p .188 .576 0 r
F P
0 g
s
.71338 .49627 m .71542 .48529 L .66635 .47764 L p .191 .522 0 r
F P
0 g
s
.66635 .47764 m .66392 .48832 L .71338 .49627 L p .191 .522 0 r
F P
0 g
s
.45063 .25658 m .44443 .26546 L .39499 .23513 L p .754 0 .178 r
F P
0 g
s
.39499 .23513 m .3995 .22541 L .45063 .25658 L p .754 0 .178 r
F P
0 g
s
.7627 .48214 m .76662 .47308 L .72047 .46461 L p .148 .675 0 r
F P
0 g
s
.72047 .46461 m .71789 .47426 L .7627 .48214 L p .148 .675 0 r
F P
0 g
s
.34883 .19907 m .34669 .20702 L .29432 .20237 L p .7 0 .195 r
F P
0 g
s
.29432 .20237 m .29467 .19434 L .34883 .19907 L p .7 0 .195 r
F P
0 g
s
.29467 .19434 m .29432 .20237 L .24487 .21162 L p .639 0 .224 r
F P
0 g
s
.71542 .48529 m .71789 .47426 L .66844 .46646 L p .188 .576 0 r
F P
0 g
s
.6699 .45623 m .66844 .46646 L .71789 .47426 L p .181 .625 0 r
F P
0 g
s
.20615 .24271 m .21256 .25228 L .18036 .28204 L p .698 0 .422 r
F P
0 g
s
.18036 .28204 m .17179 .27438 L .20615 .24271 L p .698 0 .422 r
F P
0 g
s
.81275 .47529 m .81924 .47385 L .77508 .46465 L p .059 .76 0 r
F P
0 g
s
.77508 .46465 m .77093 .46697 L .81275 .47529 L p .059 .76 0 r
F P
0 g
s
.25133 .23256 m .25532 .24288 L .21925 .2612 L p .729 0 .408 r
F P
0 g
s
.21925 .2612 m .21256 .25228 L .25133 .23256 L p .729 0 .408 r
F P
0 g
s
.20091 .23384 m .20615 .24271 L .17179 .27438 L p .637 0 .382 r
F P
0 g
s
.61168 .4326 m .60805 .43485 L .55084 .42533 L p .12 0 0 r
F P
0 g
s
.54577 .43011 m .55084 .42533 L .60805 .43485 L p .225 .162 0 r
F P
0 g
s
.86923 .49131 m .87278 .5011 L .83158 .49259 L p .076 .538 0 r
F P
0 g
s
.83158 .49259 m .82936 .48309 L .86923 .49131 L p .076 .538 0 r
F P
0 g
s
.44443 .26546 m .43782 .27477 L .39027 .24558 L p .714 0 .199 r
F P
0 g
s
.38598 .2553 m .39027 .24558 L .43782 .27477 L p .665 0 .217 r
F P
0 g
s
.39027 .24558 m .38598 .2553 L .34014 .23926 L p .679 0 .265 r
F P
0 g
s
.71204 .50575 m .71338 .49627 L .66392 .48832 L p .193 .447 0 r
F P
0 g
s
.66392 .48832 m .66145 .49709 L .71204 .50575 L p .193 .447 0 r
F P
0 g
s
.20615 .24271 m .20091 .23384 L .24487 .21162 L p .657 0 .332 r
F P
0 g
s
.39027 .24558 m .39499 .23513 L .44443 .26546 L p .714 0 .199 r
F P
0 g
s
.78533 .31999 m .78146 .32857 L .7874 .30372 L p .073 0 .405 r
F P
0 g
s
.7874 .30372 m .79142 .29374 L .78533 .31999 L p .073 0 .405 r
F P
0 g
s
.29507 .22438 m .29606 .23528 L .25532 .24288 L p .712 0 .362 r
F P
0 g
s
.25532 .24288 m .25133 .23256 L .29507 .22438 L p .712 0 .362 r
F P
0 g
s
.87051 .52342 m .865 .53287 L .8241 .52488 L p .073 .781 0 r
F P
0 g
s
.81806 .53189 m .8241 .52488 L .865 .53287 L p .066 .831 0 r
F P
0 g
s
.3995 .22541 m .39499 .23513 L .34431 .21736 L p .736 0 .213 r
F P
0 g
s
.34431 .21736 m .34669 .20702 L .3995 .22541 L p .736 0 .213 r
F P
0 g
s
.75915 .52162 m .75779 .51406 L .71204 .50575 L p .155 .375 0 r
F P
0 g
s
.71338 .49627 m .71204 .50575 L .75779 .51406 L p .159 .503 0 r
F P
0 g
s
.34014 .23926 m .34202 .22863 L .39027 .24558 L p .679 0 .265 r
F P
0 g
s
.34202 .22863 m .34014 .23926 L .29606 .23528 L p .695 0 .314 r
F P
0 g
s
.71789 .47426 m .72047 .46461 L .6699 .45623 L p .181 .625 0 r
F P
0 g
s
.29606 .23528 m .29507 .22438 L .34202 .22863 L p .695 0 .314 r
F P
0 g
s
.67054 .4483 m .6699 .45623 L .72047 .46461 L p .167 .677 0 r
F P
0 g
s
.24767 .22163 m .25133 .23256 L .21256 .25228 L p .699 0 .374 r
F P
0 g
s
.21256 .25228 m .20615 .24271 L .24767 .22163 L p .699 0 .374 r
F P
0 g
s
.24767 .22163 m .24487 .21162 L .29432 .20237 L p .684 0 .288 r
F P
0 g
s
.24487 .21162 m .24767 .22163 L .20615 .24271 L p .657 0 .332 r
F P
0 g
s
.76662 .47308 m .77093 .46697 L .72282 .45761 L p .134 .725 0 r
F P
0 g
s
.72282 .45761 m .72047 .46461 L .76662 .47308 L p .134 .725 0 r
F P
0 g
s
.75915 .52162 m .76191 .52571 L .80527 .53425 L p .096 .116 0 r
F P
0 g
s
.81149 .53514 m .80527 .53425 L .76191 .52571 L p .001 .562 0 r
F P
0 g
s
.39499 .23513 m .39027 .24558 L .34202 .22863 L p .712 0 .241 r
F P
0 g
s
.87278 .5011 m .87323 .51232 L .83137 .50376 L p .077 .648 0 r
F P
0 g
s
.83137 .50376 m .83158 .49259 L .87278 .5011 L p .077 .648 0 r
F P
0 g
s
.34669 .20702 m .34431 .21736 L .29446 .21287 L p .712 0 .249 r
F P
0 g
s
.29446 .21287 m .29432 .20237 L .34669 .20702 L p .712 0 .249 r
F P
0 g
s
.29432 .20237 m .29446 .21287 L .24767 .22163 L p .684 0 .288 r
F P
0 g
s
.67054 .4483 m .67028 .44374 L .61466 .43454 L p .166 .692 0 r
F P
0 g
s
.61168 .4326 m .61466 .43454 L .67028 .44374 L p .042 .629 0 r
F P
0 g
s
.65768 .50464 m .65928 .5028 L .60437 .49354 L p .104 0 0 r
F P
0 g
s
.60437 .49354 m .60098 .49397 L .65768 .50464 L p .104 0 0 r
F P
0 g
s
.34202 .22863 m .34431 .21736 L .39499 .23513 L p .712 0 .241 r
F P
0 g
s
.59698 .48287 m .59839 .49026 L .54034 .47952 L p .213 .616 0 r
F P
0 g
s
.54034 .47952 m .53784 .47102 L .59698 .48287 L p .213 .616 0 r
F P
0 g
s
.87323 .51232 m .87051 .52342 L .82875 .51505 L p .076 .722 0 r
F P
0 g
s
.8241 .52488 m .82875 .51505 L .87051 .52342 L p .073 .781 0 r
F P
0 g
s
.78146 .32857 m .774 .33502 L .77965 .31172 L p .078 0 .808 r
F P
0 g
s
.77965 .31172 m .7874 .30372 L .78146 .32857 L p .078 0 .808 r
F P
0 g
s
.29446 .21287 m .29507 .22438 L .25133 .23256 L p .703 0 .328 r
F P
0 g
s
.25133 .23256 m .24767 .22163 L .29446 .21287 L p .703 0 .328 r
F P
0 g
s
.71158 .51249 m .71204 .50575 L .66145 .49709 L p .186 .304 0 r
F P
0 g
s
.66145 .49709 m .65928 .5028 L .71158 .51249 L p .186 .304 0 r
F P
0 g
s
.82875 .51505 m .83137 .50376 L .87323 .51232 L p .076 .722 0 r
F P
0 g
s
.34431 .21736 m .34202 .22863 L .29507 .22438 L p .708 0 .284 r
F P
0 g
s
.73261 .32821 m .72657 .31954 L .73052 .29984 L p .009 0 .746 r
F P
0 g
s
.29507 .22438 m .29446 .21287 L .34431 .21736 L p .708 0 .284 r
F P
0 g
s
.60805 .43485 m .60428 .44101 L .54577 .43011 L p .225 .162 0 r
F P
0 g
s
.54142 .43834 m .54577 .43011 L .60428 .44101 L p .242 .34 0 r
F P
0 g
s
.72047 .46461 m .72282 .45761 L .67054 .4483 L p .167 .677 0 r
F P
0 g
s
.81924 .47385 m .82505 .47655 L .77849 .46648 L p .053 0 0 r
F P
0 g
s
.77849 .46648 m .77508 .46465 L .81924 .47385 L p .053 0 0 r
F P
0 g
s
.71204 .50575 m .71158 .51249 L .75915 .52162 L p .155 .375 0 r
F P
0 g
s
.774 .33502 m .76395 .33844 L .76923 .31665 L p .069 0 .943 r
F P
0 g
s
.76923 .31665 m .77965 .31172 L .774 .33502 L p .069 0 .943 r
F P
0 g
s
.76191 .52571 m .75915 .52162 L .71158 .51249 L p .123 .04 0 r
F P
0 g
s
.67028 .44374 m .67054 .4483 L .72282 .45761 L p .129 .724 0 r
F P
0 g
s
.73052 .29984 m .73679 .30878 L .73261 .32821 L p .009 0 .746 r
F P
0 g
s
.74172 .33479 m .73261 .32821 L .73679 .30878 L p .03 0 .883 r
F P
0 g
s
.77093 .46697 m .77508 .46465 L .72463 .45423 L p .092 .747 0 r
F P
0 g
s
.72463 .45423 m .72282 .45761 L .77093 .46697 L p .092 .747 0 r
F P
0 g
s
.59694 .47278 m .59698 .48287 L .53784 .47102 L p .231 .554 0 r
F P
0 g
s
.53784 .47102 m .53716 .46032 L .59694 .47278 L p .231 .554 0 r
F P
0 g
s
.76395 .33844 m .75267 .33837 L .75755 .31781 L p .058 0 .97 r
F P
0 g
s
.75755 .31781 m .76923 .31665 L .76395 .33844 L p .058 0 .97 r
F P
0 g
s
.60428 .44101 m .60086 .45028 L .54142 .43834 L p .242 .34 0 r
F P
0 g
s
.53841 .4489 m .54142 .43834 L .60086 .45028 L p .243 .432 0 r
F P
0 g
s
.81806 .53189 m .81149 .53514 L .76571 .52574 L p .083 .791 0 r
F P
0 g
s
.76191 .52571 m .76571 .52574 L .81149 .53514 L p .001 .562 0 r
F P
0 g
s
.75267 .33837 m .74172 .33479 L .74623 .31506 L p .045 0 .948 r
F P
0 g
s
.73679 .30878 m .74623 .31506 L .74172 .33479 L p .03 0 .883 r
F P
0 g
s
.74623 .31506 m .75755 .31781 L .75267 .33837 L p .045 0 .948 r
F P
0 g
s
.67028 .44374 m .66912 .44321 L .61168 .4326 L p .042 .629 0 r
F P
0 g
s
.60805 .43485 m .61168 .4326 L .66912 .44321 L p .167 0 0 r
F P
0 g
s
.6569 .50231 m .65768 .50464 L .60098 .49397 L p .112 .69 0 r
F P
0 g
s
.60098 .49397 m .59839 .49026 L .6569 .50231 L p .112 .69 0 r
F P
0 g
s
.71207 .51556 m .71158 .51249 L .65928 .5028 L p .131 0 0 r
F P
0 g
s
.65928 .5028 m .65768 .50464 L .71207 .51556 L p .131 0 0 r
F P
0 g
s
.5983 .46139 m .59694 .47278 L .53716 .46032 L p .239 .496 0 r
F P
0 g
s
.53716 .46032 m .53841 .4489 L .5983 .46139 L p .239 .496 0 r
F P
0 g
s
.72282 .45761 m .72463 .45423 L .67028 .44374 L p .129 .724 0 r
F P
0 g
s
.60086 .45028 m .5983 .46139 L .53841 .4489 L p .243 .432 0 r
F P
0 g
s
.71158 .51249 m .71207 .51556 L .76191 .52571 L p .123 .04 0 r
F P
0 g
s
.82505 .47655 m .82936 .48309 L .7807 .47226 L p .11 .292 0 r
F P
0 g
s
.7807 .47226 m .77849 .46648 L .82505 .47655 L p .11 .292 0 r
F P
0 g
s
.38624 .47331 m .37663 .48017 L .38526 .48772 L p 0 .802 .866 r
F P
0 g
s
.38526 .48772 m .39483 .48138 L .38624 .47331 L p 0 .802 .866 r
F P
0 g
s
.37854 .49707 m .38526 .48772 L .37663 .48017 L p 0 .805 .702 r
F P
0 g
s
.37663 .48017 m .37015 .48998 L .37854 .49707 L p 0 .805 .702 r
F P
0 g
s
.37559 .50818 m .37854 .49707 L .37015 .48998 L p 0 .721 .459 r
F P
0 g
s
.39769 .47031 m .38624 .47331 L .39483 .48138 L p 0 .713 .941 r
F P
0 g
s
.39483 .48138 m .40598 .47892 L .39769 .47031 L p 0 .713 .941 r
F P
0 g
s
.37015 .48998 m .36769 .50141 L .37559 .50818 L p 0 .721 .459 r
F P
0 g
s
.37679 .51957 m .37559 .50818 L .36769 .50141 L p 0 .541 .142 r
F P
0 g
s
.40944 .47158 m .39769 .47031 L .40598 .47892 L p 0 .534 .911 r
F P
0 g
s
.40598 .47892 m .4172 .48067 L .40944 .47158 L p 0 .534 .911 r
F P
0 g
s
.36769 .50141 m .36958 .51294 L .37679 .51957 L p 0 .541 .142 r
F P
0 g
s
.382 .52969 m .37679 .51957 L .36958 .51294 L p 0 .269 0 r
F P
0 g
s
.76571 .52574 m .76191 .52571 L .71207 .51556 L p .024 .625 0 r
F P
0 g
s
.41992 .47693 m .40944 .47158 L .4172 .48067 L p 0 .267 .754 r
F P
0 g
s
.66912 .44321 m .67028 .44374 L .72463 .45423 L p 0 .503 0 r
F P
0 g
s
.36958 .51294 m .37559 .52299 L .382 .52969 L p 0 .269 0 r
F P
0 g
s
.39052 .53719 m .382 .52969 L .37559 .52299 L p 0 .058 .555 r
F P
0 g
s
.77508 .46465 m .77849 .46648 L .72565 .45496 L p .051 0 0 r
F P
0 g
s
.72565 .45496 m .72463 .45423 L .77508 .46465 L p .051 0 0 r
F P
0 g
s
.42771 .48566 m .41992 .47693 L .42698 .48639 L p 0 0 .473 r
F P
0 g
s
.8241 .52488 m .81806 .53189 L .77005 .52164 L p .106 .76 0 r
F P
0 g
s
.76571 .52574 m .77005 .52164 L .81806 .53189 L p .083 .791 0 r
F P
0 g
s
.37559 .52299 m .3849 .53021 L .39052 .53719 L p 0 .058 .555 r
F P
0 g
s
.4012 .54103 m .39052 .53719 L .3849 .53021 L p 0 .371 .796 r
F P
0 g
s
.43153 .50828 m .43176 .4966 L .43731 .50629 L p 0 .615 .243 r
F P
0 g
s
.43176 .4966 m .43153 .50828 L .43287 .50261 L p 0 0 .089 r
F P
0 g
s
.43055 .49003 m .434 .49534 L p .43176 .4966 L 0 .371 0 r
F P
0 g
s
.434 .49534 m .43731 .50629 L .43176 .4966 L p 0 .371 0 r
F P
0 g
s
.43287 .50261 m .43472 .49482 L p .43176 .4966 L 0 0 .089 r
F P
0 g
s
.43176 .4966 m .42771 .48566 L .43055 .49003 L p 0 .371 0 r
F P
0 g
s
.3849 .53021 m .39625 .5336 L .4012 .54103 L p 0 .371 .796 r
F P
0 g
s
.41261 .5407 m .4012 .54103 L .39625 .5336 L p 0 .61 .903 r
F P
0 g
s
.66912 .44321 m .66723 .44683 L .60805 .43485 L p .167 0 0 r
F P
0 g
s
.60428 .44101 m .60805 .43485 L .66723 .44683 L p .215 .29 0 r
F P
0 g
s
.65704 .49607 m .6569 .50231 L .59839 .49026 L p .183 .66 0 r
F P
0 g
s
.59839 .49026 m .59698 .48287 L .65704 .49607 L p .183 .66 0 r
F P
0 g
s
.43731 .50629 m .43644 .51777 L .43153 .50828 L p 0 .615 .243 r
F P
0 g
s
.82936 .48309 m .83158 .49259 L .78138 .48126 L p .12 .496 0 r
F P
0 g
s
.78138 .48126 m .7807 .47226 L .82936 .48309 L p .12 .496 0 r
F P
0 g
s
.39625 .5336 m .4081 .5327 L .41261 .5407 L p 0 .61 .903 r
F P
0 g
s
.42319 .53623 m .41261 .5407 L .4081 .5327 L p 0 .755 .883 r
F P
0 g
s
.71346 .51449 m .71207 .51556 L .65768 .50464 L p .063 .672 0 r
F P
0 g
s
.65768 .50464 m .6569 .50231 L .71346 .51449 L p .063 .672 0 r
F P
0 g
s
.42702 .51912 m .43153 .50828 L .43644 .51777 L p 0 .759 .542 r
F P
0 g
s
.43644 .51777 m .43151 .52822 L .42702 .51912 L p 0 .759 .542 r
F P
0 g
s
.4081 .5327 m .41884 .52764 L .42319 .53623 L p 0 .755 .883 r
F P
0 g
s
.43151 .52822 m .42319 .53623 L .41884 .52764 L p 0 .804 .759 r
F P
0 g
s
.72463 .45423 m .72565 .45496 L .66912 .44321 L p 0 .503 0 r
F P
0 g
s
.41884 .52764 m .42702 .51912 L .43151 .52822 L p 0 .804 .759 r
F P
0 g
s
.71207 .51556 m .71346 .51449 L .76571 .52574 L p .024 .625 0 r
F P
0 g
s
.82875 .51505 m .8241 .52488 L .77433 .51394 L p .114 .709 0 r
F P
0 g
s
.77005 .52164 m .77433 .51394 L .8241 .52488 L p .106 .76 0 r
F P
0 g
s
.77005 .52164 m .76571 .52574 L .71346 .51449 L p .117 .746 0 r
F P
0 g
s
.74548 .2635 m .75662 .26043 L .75248 .24141 L p .02 0 .978 r
F P
0 g
s
.75662 .26043 m .76822 .2612 L .7642 .24146 L p .015 0 .995 r
F P
0 g
s
.7642 .24146 m .75248 .24141 L .75662 .26043 L p .015 0 .995 r
F P
0 g
s
.73629 .26999 m .74548 .2635 L .74132 .24549 L p .025 0 .896 r
F P
0 g
s
.75248 .24141 m .74132 .24549 L .74548 .2635 L p .02 0 .978 r
F P
0 g
s
.66723 .44683 m .66912 .44321 L .72565 .45496 L p .167 .127 0 r
F P
0 g
s
.83158 .49259 m .83137 .50376 L .78042 .49223 L p .12 .593 0 r
F P
0 g
s
.78042 .49223 m .78138 .48126 L .83158 .49259 L p .12 .593 0 r
F P
0 g
s
.76822 .2612 m .77874 .26571 L .77489 .24564 L p .008 0 .944 r
F P
0 g
s
.77489 .24564 m .7642 .24146 L .76822 .2612 L p .008 0 .944 r
F P
0 g
s
.77849 .46648 m .7807 .47226 L .72572 .45977 L p .144 .23 0 r
F P
0 g
s
.72572 .45977 m .72565 .45496 L .77849 .46648 L p .144 .23 0 r
F P
0 g
s
.73027 .27905 m .73629 .26999 L .7322 .25314 L p .028 0 .718 r
F P
0 g
s
.74132 .24549 m .7322 .25314 L .73629 .26999 L p .025 0 .896 r
F P
0 g
s
.66723 .44683 m .66484 .45416 L .60428 .44101 L p .215 .29 0 r
F P
0 g
s
.60086 .45028 m .60428 .44101 L .66484 .45416 L p .22 .42 0 r
F P
0 g
s
.6581 .48675 m .65704 .49607 L .59698 .48287 L p .204 .602 0 r
F P
0 g
s
.59698 .48287 m .59694 .47278 L .6581 .48675 L p .204 .602 0 r
F P
0 g
s
.83137 .50376 m .82875 .51505 L .77796 .50366 L p .118 .656 0 r
F P
0 g
s
.77433 .51394 m .77796 .50366 L .82875 .51505 L p .114 .709 0 r
F P
0 g
s
.77796 .50366 m .78042 .49223 L .83137 .50376 L p .118 .656 0 r
F P
0 g
s
.77874 .26571 m .78676 .27336 L .78312 .25341 L p 0 0 .786 r
F P
0 g
s
.78312 .25341 m .77489 .24564 L .77874 .26571 L p 0 0 .786 r
F P
0 g
s
.7322 .25314 m .72633 .26333 L .73027 .27905 L p .028 0 .718 r
F P
0 g
s
.71557 .50938 m .71346 .51449 L .6569 .50231 L p .15 .703 0 r
F P
0 g
s
.6569 .50231 m .65704 .49607 L .71557 .50938 L p .15 .703 0 r
F P
0 g
s
.72565 .45496 m .72572 .45977 L .66723 .44683 L p .167 .127 0 r
F P
0 g
s
.78676 .27336 m .79119 .28314 L .78774 .26373 L p 0 0 .442 r
F P
0 g
s
.78774 .26373 m .78312 .25341 L .78676 .27336 L p 0 0 .442 r
F P
0 g
s
.71346 .51449 m .71557 .50938 L .77005 .52164 L p .117 .746 0 r
F P
0 g
s
.66484 .45416 m .66229 .46421 L .60086 .45028 L p .22 .42 0 r
F P
0 g
s
.5983 .46139 m .60086 .45028 L .66229 .46421 L p .219 .493 0 r
F P
0 g
s
.65993 .47561 m .6581 .48675 L .59694 .47278 L p .213 .549 0 r
F P
0 g
s
.59694 .47278 m .5983 .46139 L .65993 .47561 L p .213 .549 0 r
F P
0 g
s
.77433 .51394 m .77005 .52164 L .71557 .50938 L p .141 .7 0 r
F P
0 g
s
.66229 .46421 m .65993 .47561 L .5983 .46139 L p .219 .493 0 r
F P
0 g
s
.66484 .45416 m .66723 .44683 L .72572 .45977 L p .19 .378 0 r
F P
0 g
s
.79119 .28314 m .79142 .29374 L .78813 .27521 L p .024 0 .079 r
F P
0 g
s
.78813 .27521 m .78774 .26373 L .79119 .28314 L p .024 0 .079 r
F P
0 g
s
.7807 .47226 m .78138 .48126 L .72483 .46804 L p .157 .443 0 r
F P
0 g
s
.72483 .46804 m .72572 .45977 L .7807 .47226 L p .157 .443 0 r
F P
0 g
s
.71811 .50087 m .71557 .50938 L .65704 .49607 L p .174 .649 0 r
F P
0 g
s
.65704 .49607 m .6581 .48675 L .71811 .50087 L p .174 .649 0 r
F P
0 g
s
.79142 .29374 m .7874 .30372 L .78421 .28627 L p .028 0 .542 r
F P
0 g
s
.78421 .28627 m .78813 .27521 L .79142 .29374 L p .028 0 .542 r
F P
0 g
s
.72572 .45977 m .72483 .46804 L .66484 .45416 L p .19 .378 0 r
F P
0 g
s
.77796 .50366 m .77433 .51394 L .71811 .50087 L p .151 .651 0 r
F P
0 g
s
.71557 .50938 m .71811 .50087 L .77433 .51394 L p .141 .7 0 r
F P
0 g
s
.7874 .30372 m .77965 .31172 L .77652 .29541 L p .026 0 .814 r
F P
0 g
s
.77652 .29541 m .78421 .28627 L .7874 .30372 L p .026 0 .814 r
F P
0 g
s
.78138 .48126 m .78042 .49223 L .72309 .47865 L p .158 .539 0 r
F P
0 g
s
.72309 .47865 m .72483 .46804 L .78138 .48126 L p .158 .539 0 r
F P
0 g
s
.66229 .46421 m .66484 .45416 L .72483 .46804 L p .192 .483 0 r
F P
0 g
s
.75755 .31781 m .74623 .31506 L .74298 .30113 L p .011 0 .979 r
F P
0 g
s
.77965 .31172 m .76923 .31665 L .7661 .30138 L p .022 0 .947 r
F P
0 g
s
.7661 .30138 m .77652 .29541 L .77965 .31172 L p .022 0 .947 r
F P
0 g
s
.76923 .31665 m .75755 .31781 L .75439 .30338 L p .017 0 .997 r
F P
0 g
s
.74298 .30113 m .75439 .30338 L .75755 .31781 L p .011 0 .979 r
F P
0 g
s
.75439 .30338 m .7661 .30138 L .76923 .31665 L p .017 0 .997 r
F P
0 g
s
.72074 .49013 m .71811 .50087 L .6581 .48675 L p .184 .599 0 r
F P
0 g
s
.6581 .48675 m .65993 .47561 L .72074 .49013 L p .184 .599 0 r
F P
0 g
s
.78042 .49223 m .77796 .50366 L .72074 .49013 L p .156 .6 0 r
F P
0 g
s
.71811 .50087 m .72074 .49013 L .77796 .50366 L p .151 .651 0 r
F P
0 g
s
.72074 .49013 m .72309 .47865 L .78042 .49223 L p .156 .6 0 r
F P
0 g
s
.72483 .46804 m .72309 .47865 L .66229 .46421 L p .192 .483 0 r
F P
0 g
s
.65993 .47561 m .66229 .46421 L .72309 .47865 L p .189 .547 0 r
F P
0 g
s
.72309 .47865 m .72074 .49013 L .65993 .47561 L p .189 .547 0 r
F P
0 g
s
.72633 .26333 m .7322 .25314 L .71996 .2434 L p 0 .029 .757 r
F P
0 g
s
.7322 .25314 m .74132 .24549 L .72865 .23515 L p 0 .03 .897 r
F P
0 g
s
.72865 .23515 m .71996 .2434 L .7322 .25314 L p 0 .03 .897 r
F P
0 g
s
.7245 .2747 m .72633 .26333 L .71455 .25424 L p 0 .026 .548 r
F P
0 g
s
.71996 .2434 m .71455 .25424 L .72633 .26333 L p 0 .029 .757 r
F P
0 g
s
.74132 .24549 m .75248 .24141 L .73945 .2306 L p 0 .028 .975 r
F P
0 g
s
.73945 .2306 m .72865 .23515 L .74132 .24549 L p 0 .028 .975 r
F P
0 g
s
.4104 .45831 m .39839 .45716 L .39769 .47031 L p 0 .572 .986 r
F P
0 g
s
.39769 .47031 m .40944 .47158 L .4104 .45831 L p 0 .572 .986 r
F P
0 g
s
.38624 .47331 m .39769 .47031 L .39839 .45716 L p 0 .691 .96 r
F P
0 g
s
.42131 .46357 m .4104 .45831 L .40944 .47158 L p 0 .401 .919 r
F P
0 g
s
.40944 .47158 m .41992 .47693 L .42131 .46357 L p 0 .401 .919 r
F P
0 g
s
.72697 .28575 m .7245 .2747 L .71313 .26622 L p 0 .02 .244 r
F P
0 g
s
.71455 .25424 m .71313 .26622 L .7245 .2747 L p 0 .026 .548 r
F P
0 g
s
.39839 .45716 m .38688 .46028 L .38624 .47331 L p 0 .691 .96 r
F P
0 g
s
.37663 .48017 m .38624 .47331 L .38688 .46028 L p 0 .745 .831 r
F P
0 g
s
.42966 .47222 m .42131 .46357 L .41992 .47693 L p 0 .179 .753 r
F P
0 g
s
.41992 .47693 m .42771 .48566 L .42966 .47222 L p 0 .179 .753 r
F P
0 g
s
.75248 .24141 m .7642 .24146 L .75095 .23039 L p 0 .023 .969 r
F P
0 g
s
.75095 .23039 m .73945 .2306 L .75248 .24141 L p 0 .023 .969 r
F P
0 g
s
.38688 .46028 m .37742 .46725 L .37663 .48017 L p 0 .745 .831 r
F P
0 g
s
.37015 .48998 m .37663 .48017 L .37742 .46725 L p 0 .707 .575 r
F P
0 g
s
.43435 .48313 m .42966 .47222 L .42771 .48566 L p 0 0 .474 r
F P
0 g
s
.42771 .48566 m .43176 .4966 L .43435 .48313 L p 0 0 .474 r
F P
0 g
s
.71313 .26622 m .71588 .27775 L .72697 .28575 L p 0 .02 .244 r
F P
0 g
s
.37742 .46725 m .3713 .47717 L .37015 .48998 L p 0 .707 .575 r
F P
0 g
s
.36769 .50141 m .37015 .48998 L .3713 .47717 L p 0 .549 .198 r
F P
0 g
s
.43472 .49482 m .43435 .48313 L .43176 .4966 L p 0 0 .089 r
F P
0 g
s
.7642 .24146 m .77489 .24564 L .76157 .23457 L p 0 .013 .816 r
F P
0 g
s
.76157 .23457 m .75095 .23039 L .7642 .24146 L p 0 .013 .816 r
F P
0 g
s
.3713 .47717 m .36936 .4887 L .36769 .50141 L p 0 .549 .198 r
F P
0 g
s
.36958 .51294 m .36769 .50141 L .36936 .4887 L p 0 0 .225 r
F P
0 g
s
.43071 .50574 m .43472 .49482 L .43153 .50828 L p 0 .607 .33 r
F P
0 g
s
.43153 .50828 m .42702 .51912 L .43071 .50574 L p 0 .607 .33 r
F P
0 g
s
.36936 .4887 m .37187 .50027 L .36958 .51294 L p 0 0 .225 r
F P
0 g
s
.37559 .52299 m .36958 .51294 L .37187 .50027 L p 0 .008 .584 r
F P
0 g
s
.42285 .51441 m .43071 .50574 L .42702 .51912 L p 0 .721 .673 r
F P
0 g
s
.42702 .51912 m .41884 .52764 L .42285 .51441 L p 0 .721 .673 r
F P
0 g
s
.77489 .24564 m .78312 .25341 L .76987 .2426 L p 0 .001 0 r
F P
0 g
s
.76987 .2426 m .76157 .23457 L .77489 .24564 L p 0 .001 0 r
F P
0 g
s
.37187 .50027 m .37852 .51031 L .37559 .52299 L p 0 .008 .584 r
F P
0 g
s
.3849 .53021 m .37559 .52299 L .37852 .51031 L p 0 .27 .824 r
F P
0 g
s
.41219 .51964 m .42285 .51441 L .41884 .52764 L p 0 .719 .884 r
F P
0 g
s
.41884 .52764 m .4081 .5327 L .41219 .51964 L p 0 .719 .884 r
F P
0 g
s
.37852 .51031 m .3884 .51745 L .3849 .53021 L p 0 .27 .824 r
F P
0 g
s
.39625 .5336 m .3849 .53021 L .3884 .51745 L p 0 .478 .949 r
F P
0 g
s
.40016 .52071 m .41219 .51964 L .4081 .5327 L p 0 .63 .97 r
F P
0 g
s
.4081 .5327 m .39625 .5336 L .40016 .52071 L p 0 .63 .97 r
F P
0 g
s
.3884 .51745 m .40016 .52071 L .39625 .5336 L p 0 .478 .949 r
F P
0 g
s
.78312 .25341 m .78774 .26373 L .7747 .25339 L p 0 .014 0 r
F P
0 g
s
.7747 .25339 m .76987 .2426 L .78312 .25341 L p 0 .014 0 r
F P
0 g
s
.7661 .30138 m .75439 .30338 L .74315 .29571 L p 0 .026 .975 r
F P
0 g
s
.78774 .26373 m .78813 .27521 L .77539 .26548 L p 0 .023 .373 r
F P
0 g
s
.77539 .26548 m .7747 .25339 L .78774 .26373 L p 0 .023 .373 r
F P
0 g
s
.77652 .29541 m .7661 .30138 L .75453 .29341 L p 0 .029 .939 r
F P
0 g
s
.74315 .29571 m .75453 .29341 L .7661 .30138 L p 0 .026 .975 r
F P
0 g
s
.78813 .27521 m .78421 .28627 L .77185 .2772 L p 0 .028 .644 r
F P
0 g
s
.77185 .2772 m .77539 .26548 L .78813 .27521 L p 0 .028 .644 r
F P
0 g
s
.78421 .28627 m .77652 .29541 L .76456 .28695 L p 0 .03 .826 r
F P
0 g
s
.75453 .29341 m .76456 .28695 L .77652 .29541 L p 0 .029 .939 r
F P
0 g
s
.76456 .28695 m .77185 .2772 L .78421 .28627 L p 0 .03 .826 r
F P
0 g
s
.71588 .27775 m .71313 .26622 L .69429 .26441 L p 0 .064 .466 r
F P
0 g
s
.71313 .26622 m .71455 .25424 L .69514 .2522 L p 0 .078 .645 r
F P
0 g
s
.69514 .2522 m .69429 .26441 L .71313 .26622 L p 0 .078 .645 r
F P
0 g
s
.71455 .25424 m .71996 .2434 L .69984 .24121 L p 0 .086 .789 r
F P
0 g
s
.69984 .24121 m .69514 .2522 L .71455 .25424 L p 0 .086 .789 r
F P
0 g
s
.72244 .28729 m .71588 .27775 L .69737 .27625 L p 0 .045 .227 r
F P
0 g
s
.69429 .26441 m .69737 .27625 L .71588 .27775 L p 0 .064 .466 r
F P
0 g
s
.71996 .2434 m .72865 .23515 L .70778 .23293 L p 0 .089 .897 r
F P
0 g
s
.70778 .23293 m .69984 .24121 L .71996 .2434 L p 0 .089 .897 r
F P
0 g
s
.69737 .27625 m .70399 .28614 L .72244 .28729 L p 0 .045 .227 r
F P
0 g
s
.72865 .23515 m .73945 .2306 L .7179 .22849 L p 0 .083 .929 r
F P
0 g
s
.7179 .22849 m .70778 .23293 L .72865 .23515 L p 0 .083 .929 r
F P
0 g
s
.73945 .2306 m .75095 .23039 L .72887 .22851 L p 0 .056 .774 r
F P
0 g
s
.72887 .22851 m .7179 .22849 L .73945 .2306 L p 0 .056 .774 r
F P
0 g
s
.4397 .45517 m .43103 .44643 L .42131 .46357 L p 0 .295 .849 r
F P
0 g
s
.42131 .46357 m .42966 .47222 L .4397 .45517 L p 0 .295 .849 r
F P
0 g
s
.4104 .45831 m .42131 .46357 L .43103 .44643 L p 0 .437 .953 r
F P
0 g
s
.43103 .44643 m .42004 .44092 L .4104 .45831 L p 0 .437 .953 r
F P
0 g
s
.39839 .45716 m .4104 .45831 L .42004 .44092 L p 0 .557 .998 r
F P
0 g
s
.44491 .466 m .4397 .45517 L .42966 .47222 L p 0 .117 .672 r
F P
0 g
s
.42966 .47222 m .43435 .48313 L .44491 .466 L p 0 .117 .672 r
F P
0 g
s
.75095 .23039 m .76157 .23457 L .7392 .23304 L p 0 0 0 r
F P
0 g
s
.7392 .23304 m .72887 .22851 L .75095 .23039 L p 0 0 0 r
F P
0 g
s
.42004 .44092 m .40819 .43939 L .39839 .45716 L p 0 .557 .998 r
F P
0 g
s
.38688 .46028 m .39839 .45716 L .40819 .43939 L p 0 .65 .966 r
F P
0 g
s
.44596 .47749 m .44491 .466 L .43435 .48313 L p 0 0 .38 r
F P
0 g
s
.43435 .48313 m .43472 .49482 L .44596 .47749 L p 0 0 .38 r
F P
0 g
s
.76456 .28695 m .75453 .29341 L .73476 .29316 L p 0 .088 .912 r
F P
0 g
s
.40819 .43939 m .39705 .44207 L .38688 .46028 L p 0 .65 .966 r
F P
0 g
s
.37742 .46725 m .38688 .46028 L .39705 .44207 L p 0 .681 .79 r
F P
0 g
s
.44267 .48809 m .44596 .47749 L .43472 .49482 L p 0 0 0 r
F P
0 g
s
.43415 .49636 m .43071 .50574 L .44267 .48809 L p 0 0 0 r
F P
0 g
s
.43547 .49638 m .44267 .48809 L .43071 .50574 L p 0 .618 .561 r
F P
0 g
s
.44267 .48809 m .43547 .49638 L .44785 .48306 L p 0 0 0 r
F P
0 g
s
.44785 .48306 m .45643 .47382 L p .44267 .48809 L 0 0 0 r
F P
0 g
s
.43472 .49482 m .43415 .49636 L p .44267 .48809 L 0 0 0 r
F P
0 g
s
.76157 .23457 m .76987 .2426 L .74745 .24147 L p 0 .028 .017 r
F P
0 g
s
.74745 .24147 m .7392 .23304 L .76157 .23457 L p 0 .028 .017 r
F P
0 g
s
.39705 .44207 m .38812 .44861 L .37742 .46725 L p 0 .681 .79 r
F P
0 g
s
.3713 .47717 m .37742 .46725 L .38812 .44861 L p 0 .566 .393 r
F P
0 g
s
.77185 .2772 m .76456 .28695 L .74408 .28681 L p 0 .09 .841 r
F P
0 g
s
.73476 .29316 m .74408 .28681 L .76456 .28695 L p 0 .088 .912 r
F P
0 g
s
.43071 .50574 m .42285 .51441 L .43547 .49638 L p 0 .618 .561 r
F P
0 g
s
.76987 .2426 m .7747 .25339 L .75249 .25267 L p 0 .055 .322 r
F P
0 g
s
.75249 .25267 m .74745 .24147 L .76987 .2426 L p 0 .055 .322 r
F P
0 g
s
.77539 .26548 m .77185 .2772 L .75067 .27701 L p 0 .083 .71 r
F P
0 g
s
.74408 .28681 m .75067 .27701 L .77185 .2772 L p 0 .09 .841 r
F P
0 g
s
.38812 .44861 m .38264 .45817 L .3713 .47717 L p 0 .566 .393 r
F P
0 g
s
.36936 .4887 m .3713 .47717 L .38264 .45817 L p 0 0 .117 r
F P
0 g
s
.7747 .25339 m .77539 .26548 L .75361 .26509 L p 0 .072 .542 r
F P
0 g
s
.75361 .26509 m .75249 .25267 L .7747 .25339 L p 0 .072 .542 r
F P
0 g
s
.75067 .27701 m .75361 .26509 L .77539 .26548 L p 0 .083 .71 r
F P
0 g
s
.4253 .50123 m .43547 .49638 L .42285 .51441 L p 0 .672 .88 r
F P
0 g
s
.42285 .51441 m .41219 .51964 L .4253 .50123 L p 0 .672 .88 r
F P
0 g
s
.38264 .45817 m .38135 .46946 L .36936 .4887 L p 0 0 .117 r
F P
0 g
s
.37187 .50027 m .36936 .4887 L .38135 .46946 L p 0 0 .514 r
F P
0 g
s
.41352 .50193 m .4253 .50123 L .41219 .51964 L p 0 .608 .992 r
F P
0 g
s
.41219 .51964 m .40016 .52071 L .41352 .50193 L p 0 .608 .992 r
F P
0 g
s
.38135 .46946 m .38448 .48095 L .37187 .50027 L p 0 0 .514 r
F P
0 g
s
.37852 .51031 m .37187 .50027 L .38448 .48095 L p 0 .194 .761 r
F P
0 g
s
.40175 .49838 m .41352 .50193 L .40016 .52071 L p 0 .498 .986 r
F P
0 g
s
.40016 .52071 m .3884 .51745 L .40175 .49838 L p 0 .498 .986 r
F P
0 g
s
.38448 .48095 m .39161 .49105 L .37852 .51031 L p 0 .194 .761 r
F P
0 g
s
.3884 .51745 m .37852 .51031 L .39161 .49105 L p 0 .362 .908 r
F P
0 g
s
.39161 .49105 m .40175 .49838 L .3884 .51745 L p 0 .362 .908 r
F P
0 g
s
.69737 .27625 m .69429 .26441 L .66855 .26942 L p 0 .108 .58 r
F P
0 g
s
.69429 .26441 m .69514 .2522 L .66868 .25733 L p 0 .127 .705 r
F P
0 g
s
.66868 .25733 m .66855 .26942 L .69429 .26441 L p 0 .127 .705 r
F P
0 g
s
.70399 .28614 m .69737 .27625 L .67204 .28137 L p 0 .084 .429 r
F P
0 g
s
.66855 .26942 m .67204 .28137 L .69737 .27625 L p 0 .108 .58 r
F P
0 g
s
.69514 .2522 m .69984 .24121 L .67244 .24669 L p 0 .142 .817 r
F P
0 g
s
.67244 .24669 m .66868 .25733 L .69514 .2522 L p 0 .142 .817 r
F P
0 g
s
.67204 .28137 m .67868 .29161 L .70399 .28614 L p 0 .084 .429 r
F P
0 g
s
.71328 .29276 m .70399 .28614 L .67868 .29161 L p 0 .05 .22 r
F P
0 g
s
.69984 .24121 m .70778 .23293 L .67936 .23894 L p 0 .15 .895 r
F P
0 g
s
.67936 .23894 m .67244 .24669 L .69984 .24121 L p 0 .15 .895 r
F P
0 g
s
.67868 .29161 m .68762 .29878 L .71328 .29276 L p 0 .05 .22 r
F P
0 g
s
.70778 .23293 m .7179 .22849 L .68852 .23514 L p 0 .128 .812 r
F P
0 g
s
.67936 .23894 m .70778 .23293 L p .69032 .23493 L 0 .128 .812 r
F P
0 g
s
.72887 .22851 m .69872 .23585 L p .71613 .22917 L 0 0 0 r
F P
0 g
s
.7179 .22849 m .71917 .22849 L p .71613 .22917 L 0 0 0 r
F P
0 g
s
.71917 .22849 m .72887 .22851 L p .71613 .22917 L 0 0 0 r
F P
0 g
s
.68852 .23514 m .6875 .23556 L p .69032 .23493 L 0 .128 .812 r
F P
0 g
s
.6875 .23556 m .67936 .23894 L p .69032 .23493 L 0 .128 .812 r
F P
0 g
s
.69872 .23585 m .68852 .23514 L .7179 .22849 L p 0 0 0 r
F P
0 g
s
.457 .43511 m .44829 .4261 L .43103 .44643 L p 0 .326 .848 r
F P
0 g
s
.43103 .44643 m .4397 .45517 L .457 .43511 L p 0 .326 .848 r
F P
0 g
s
.42004 .44092 m .43103 .44643 L .44829 .4261 L p 0 .423 .929 r
F P
0 g
s
.46269 .4458 m .457 .43511 L .4397 .45517 L p 0 .211 .727 r
F P
0 g
s
.4397 .45517 m .44491 .466 L .46269 .4458 L p 0 .211 .727 r
F P
0 g
s
.44829 .4261 m .43768 .41994 L .42004 .44092 L p 0 .423 .929 r
F P
0 g
s
.40819 .43939 m .42004 .44092 L .43768 .41994 L p 0 .515 .978 r
F P
0 g
s
.46459 .45677 m .46269 .4458 L .44491 .466 L p 0 .052 .529 r
F P
0 g
s
.44491 .466 m .44596 .47749 L .46459 .45677 L p 0 .052 .529 r
F P
0 g
s
.43768 .41994 m .42659 .41747 L .40819 .43939 L p 0 .515 .978 r
F P
0 g
s
.39705 .44207 m .40819 .43939 L .42659 .41747 L p 0 .599 .963 r
F P
0 g
s
.72887 .22851 m .7392 .23304 L .70856 .24103 L p 0 .021 .021 r
F P
0 g
s
.70856 .24103 m .69872 .23585 L .72887 .22851 L p 0 .021 .021 r
F P
0 g
s
.75067 .27701 m .74408 .28681 L .71585 .29491 L p 0 .149 .855 r
F P
0 g
s
.70754 .3006 m .71585 .29491 L .74408 .28681 L p 0 .143 .854 r
F P
0 g
s
.46242 .46655 m .46459 .45677 L .44596 .47749 L p 0 0 .162 r
F P
0 g
s
.44596 .47749 m .44267 .48809 L .46242 .46655 L p 0 0 .162 r
F P
0 g
s
.42659 .41747 m .41646 .41903 L .39705 .44207 L p 0 .599 .963 r
F P
0 g
s
.38812 .44861 m .39705 .44207 L .41646 .41903 L p 0 .61 .746 r
F P
0 g
s
.7392 .23304 m .74745 .24147 L .71668 .24998 L p 0 .067 .304 r
F P
0 g
s
.71668 .24998 m .70856 .24103 L .7392 .23304 L p 0 .067 .304 r
F P
0 g
s
.75361 .26509 m .75067 .27701 L .72147 .28561 L p 0 .137 .757 r
F P
0 g
s
.71585 .29491 m .72147 .28561 L .75067 .27701 L p 0 .149 .855 r
F P
0 g
s
.45643 .47382 m .46242 .46655 L .44267 .48809 L p 0 0 0 r
F P
0 g
s
.74745 .24147 m .75249 .25267 L .72194 .26148 L p 0 .097 .492 r
F P
0 g
s
.72194 .26148 m .71668 .24998 L .74745 .24147 L p 0 .097 .492 r
F P
0 g
s
.75249 .25267 m .75361 .26509 L .72361 .27393 L p 0 .119 .635 r
F P
0 g
s
.72147 .28561 m .72361 .27393 L .75361 .26509 L p 0 .137 .757 r
F P
0 g
s
.67868 .29161 m .67204 .28137 L .64105 .29281 L p 0 .123 .539 r
F P
0 g
s
.67204 .28137 m .66855 .26942 L .63704 .28098 L p 0 .15 .649 r
F P
0 g
s
.63704 .28098 m .64105 .29281 L .67204 .28137 L p 0 .15 .649 r
F P
0 g
s
.72361 .27393 m .72194 .26148 L .75249 .25267 L p 0 .119 .635 r
F P
0 g
s
.41646 .41903 m .40868 .42446 L .38812 .44861 L p 0 .61 .746 r
F P
0 g
s
.38264 .45817 m .38812 .44861 L .40868 .42446 L p 0 0 0 r
F P
0 g
s
.66855 .26942 m .66868 .25733 L .63625 .26939 L p 0 .174 .748 r
F P
0 g
s
.63625 .26939 m .63704 .28098 L .66855 .26942 L p 0 .174 .748 r
F P
0 g
s
.64105 .29281 m .64775 .30334 L .67868 .29161 L p 0 .123 .539 r
F P
0 g
s
.68762 .29878 m .67868 .29161 L .64775 .30334 L p 0 .087 .395 r
F P
0 g
s
.4474 .47759 m .45643 .47382 L .43547 .49638 L p 0 .62 .876 r
F P
0 g
s
.43547 .49638 m .4253 .50123 L .4474 .47759 L p 0 .62 .876 r
F P
0 g
s
.66868 .25733 m .67244 .24669 L .6388 .25957 L p 0 .198 .842 r
F P
0 g
s
.6388 .25957 m .63625 .26939 L .66868 .25733 L p 0 .198 .842 r
F P
0 g
s
.69768 .30193 m .68762 .29878 L .65626 .31118 L p 0 .032 .166 r
F P
0 g
s
.64775 .30334 m .65626 .31118 L .68762 .29878 L p 0 .087 .395 r
F P
0 g
s
.40868 .42446 m .40432 .43306 L .38264 .45817 L p 0 0 0 r
F P
0 g
s
.38135 .46946 m .38264 .45817 L .40432 .43306 L p 0 0 .343 r
F P
0 g
s
.43653 .47729 m .4474 .47759 L .4253 .50123 L p 0 .564 .986 r
F P
0 g
s
.4253 .50123 m .41352 .50193 L .43653 .47729 L p 0 .564 .986 r
F P
0 g
s
.40432 .43306 m .40402 .44368 L .38135 .46946 L p 0 0 .343 r
F P
0 g
s
.38448 .48095 m .38135 .46946 L .40402 .44368 L p 0 .125 .633 r
F P
0 g
s
.67244 .24669 m .67936 .23894 L .6444 .25285 L p 0 .21 .882 r
F P
0 g
s
.6444 .25285 m .6388 .25957 L .67244 .24669 L p 0 .21 .882 r
F P
0 g
s
.4253 .47294 m .43653 .47729 L .41352 .50193 L p 0 .472 .964 r
F P
0 g
s
.41352 .50193 m .40175 .49838 L .4253 .47294 L p 0 .472 .964 r
F P
0 g
s
.65626 .31118 m .66548 .3153 L .69768 .30193 L p 0 .032 .166 r
F P
0 g
s
.40402 .44368 m .40783 .45488 L .38448 .48095 L p 0 .125 .633 r
F P
0 g
s
.39161 .49105 m .38448 .48095 L .40783 .45488 L p 0 .266 .794 r
F P
0 g
s
.41527 .46511 m .4253 .47294 L .40175 .49838 L p 0 .375 .897 r
F P
0 g
s
.40175 .49838 m .39161 .49105 L .41527 .46511 L p 0 .375 .897 r
F P
0 g
s
.40783 .45488 m .41527 .46511 L .39161 .49105 L p 0 .266 .794 r
F P
0 g
s
.67936 .23894 m .68852 .23514 L .65232 .25017 L p 0 0 0 r
F P
0 g
s
.65232 .25017 m .6444 .25285 L .67936 .23894 L p 0 0 0 r
F P
0 g
s
.44829 .4261 m .457 .43511 L .48041 .41288 L p 0 .32 .817 r
F P
0 g
s
.48657 .42333 m .48041 .41288 L .457 .43511 L p 0 .24 .724 r
F P
0 g
s
.457 .43511 m .46269 .4458 L .48657 .42333 L p 0 .24 .724 r
F P
0 g
s
.48041 .41288 m .47192 .40345 L .44829 .4261 L p 0 .32 .817 r
F P
0 g
s
.43768 .41994 m .44829 .4261 L .47192 .40345 L p 0 .391 .888 r
F P
0 g
s
.48958 .43344 m .48657 .42333 L .46269 .4458 L p 0 .129 .578 r
F P
0 g
s
.46269 .4458 m .46459 .45677 L .48958 .43344 L p 0 .129 .578 r
F P
0 g
s
.47192 .40345 m .46221 .39627 L .43768 .41994 L p 0 .391 .888 r
F P
0 g
s
.42659 .41747 m .43768 .41994 L .46221 .39627 L p 0 .465 .943 r
F P
0 g
s
.68852 .23514 m .69872 .23585 L .66151 .25196 L p 0 0 0 r
F P
0 g
s
.66151 .25196 m .65232 .25017 L .68852 .23514 L p 0 0 0 r
F P
0 g
s
.48903 .44187 m .48958 .43344 L .46459 .45677 L p 0 0 .291 r
F P
0 g
s
.46459 .45677 m .46242 .46655 L .48903 .44187 L p 0 0 .291 r
F P
0 g
s
.64775 .30334 m .64105 .29281 L .6062 .30978 L p 0 .161 .609 r
F P
0 g
s
.64105 .29281 m .63704 .28098 L .60145 .29836 L p 0 .192 .698 r
F P
0 g
s
.60145 .29836 m .6062 .30978 L .64105 .29281 L p 0 .192 .698 r
F P
0 g
s
.72147 .28561 m .71585 .29491 L .68121 .31058 L p 0 .21 .867 r
F P
0 g
s
.46221 .39627 m .45254 .39229 L .42659 .41747 L p 0 .465 .943 r
F P
0 g
s
.41646 .41903 m .42659 .41747 L .45254 .39229 L p 0 .543 .953 r
F P
0 g
s
.6062 .30978 m .61304 .32045 L .64775 .30334 L p 0 .161 .609 r
F P
0 g
s
.65626 .31118 m .64775 .30334 L .61304 .32045 L p 0 .124 .499 r
F P
0 g
s
.63704 .28098 m .63625 .26939 L .59942 .28766 L p 0 .222 .782 r
F P
0 g
s
.59942 .28766 m .60145 .29836 L .63704 .28098 L p 0 .222 .782 r
F P
0 g
s
.66548 .3153 m .65626 .31118 L .6211 .32898 L p 0 .068 .329 r
F P
0 g
s
.61304 .32045 m .6211 .32898 L .65626 .31118 L p 0 .124 .499 r
F P
0 g
s
.69872 .23585 m .70856 .24103 L .67072 .258 L p 0 .059 .265 r
F P
0 g
s
.67072 .258 m .66151 .25196 L .69872 .23585 L p 0 .059 .265 r
F P
0 g
s
.48494 .44747 m .48903 .44187 L .46242 .46655 L p 0 0 0 r
F P
0 g
s
.46242 .46655 m .45643 .47382 L .48494 .44747 L p 0 0 0 r
F P
0 g
s
.72361 .27393 m .72147 .28561 L .68557 .30228 L p 0 .189 .792 r
F P
0 g
s
.68121 .31058 m .68557 .30228 L .72147 .28561 L p 0 .21 .867 r
F P
0 g
s
.45254 .39229 m .44419 .39207 L .41646 .41903 L p 0 .543 .953 r
F P
0 g
s
.40868 .42446 m .41646 .41903 L .44419 .39207 L p 0 .539 .71 r
F P
0 g
s
.63625 .26939 m .6388 .25957 L .60042 .27912 L p 0 .255 .864 r
F P
0 g
s
.60042 .27912 m .59942 .28766 L .63625 .26939 L p 0 .255 .864 r
F P
0 g
s
.70856 .24103 m .71668 .24998 L .67866 .2675 L p 0 .106 .457 r
F P
0 g
s
.67866 .2675 m .67072 .258 L .70856 .24103 L p 0 .106 .457 r
F P
0 g
s
.72194 .26148 m .72361 .27393 L .68663 .2913 L p 0 .164 .694 r
F P
0 g
s
.68557 .30228 m .68663 .2913 L .72361 .27393 L p 0 .189 .792 r
F P
0 g
s
.6211 .32898 m .62932 .33424 L .66548 .3153 L p 0 .068 .329 r
F P
0 g
s
.47782 .44945 m .48494 .44747 L .45643 .47382 L p 0 .565 .874 r
F P
0 g
s
.45643 .47382 m .4474 .47759 L .47782 .44945 L p 0 .565 .874 r
F P
0 g
s
.71668 .24998 m .72194 .26148 L .68422 .27914 L p 0 .137 .587 r
F P
0 g
s
.68422 .27914 m .67866 .2675 L .71668 .24998 L p 0 .137 .587 r
F P
0 g
s
.68663 .2913 m .68422 .27914 L .72194 .26148 L p 0 .164 .694 r
F P
0 g
s
.44419 .39207 m .4383 .39569 L .40868 .42446 L p 0 .539 .71 r
F P
0 g
s
.40432 .43306 m .40868 .42446 L .4383 .39569 L p 0 0 0 r
F P
0 g
s
.6388 .25957 m .6444 .25285 L .60434 .27389 L p 0 .267 .852 r
F P
0 g
s
.60042 .27912 m .6388 .25957 L p .60822 .27228 L 0 .267 .852 r
F P
0 g
s
.65232 .25017 m .61069 .27274 L p .64055 .25512 L 0 0 0 r
F P
0 g
s
.6444 .25285 m .64558 .25245 L p .64055 .25512 L 0 0 0 r
F P
0 g
s
.64558 .25245 m .65232 .25017 L p .64055 .25512 L 0 0 0 r
F P
0 g
s
.60434 .27389 m .60378 .27464 L p .60822 .27228 L 0 .267 .852 r
F P
0 g
s
.60378 .27464 m .60042 .27912 L p .60822 .27228 L 0 .267 .852 r
F P
0 g
s
.47192 .40345 m .48041 .41288 L .50887 .38925 L p 0 .297 .78 r
F P
0 g
s
.51551 .39936 m .50887 .38925 L .48041 .41288 L p 0 .236 .699 r
F P
0 g
s
.48041 .41288 m .48657 .42333 L .51551 .39936 L p 0 .236 .699 r
F P
0 g
s
.50887 .38925 m .5008 .37934 L .47192 .40345 L p 0 .297 .78 r
F P
0 g
s
.46221 .39627 m .47192 .40345 L .5008 .37934 L p 0 .353 .846 r
F P
0 g
s
.51986 .40836 m .51551 .39936 L .48657 .42333 L p 0 .153 .577 r
F P
0 g
s
.48657 .42333 m .48958 .43344 L .51986 .40836 L p 0 .153 .577 r
F P
0 g
s
.5008 .37934 m .49236 .37092 L .46221 .39627 L p 0 .353 .846 r
F P
0 g
s
.45254 .39229 m .46221 .39627 L .49236 .37092 L p 0 .414 .907 r
F P
0 g
s
.61304 .32045 m .6062 .30978 L .56969 .33098 L p 0 .198 .661 r
F P
0 g
s
.6062 .30978 m .60145 .29836 L .56393 .32029 L p 0 .233 .738 r
F P
0 g
s
.56393 .32029 m .56969 .33098 L .6062 .30978 L p 0 .233 .738 r
F P
0 g
s
.56969 .33098 m .57683 .34159 L .61304 .32045 L p 0 .198 .661 r
F P
0 g
s
.6211 .32898 m .61304 .32045 L .57683 .34159 L p 0 .159 .57 r
F P
0 g
s
.46861 .4475 m .47782 .44945 L .4474 .47759 L p 0 .509 .964 r
F P
0 g
s
.4474 .47759 m .43653 .47729 L .46861 .4475 L p 0 .509 .964 r
F P
0 g
s
.4383 .39569 m .4357 .40271 L .40432 .43306 L p 0 0 0 r
F P
0 g
s
.40402 .44368 m .40432 .43306 L .4357 .40271 L p 0 .034 .445 r
F P
0 g
s
.60145 .29836 m .59942 .28766 L .5603 .3109 L p 0 .27 .813 r
F P
0 g
s
.5603 .3109 m .56393 .32029 L .60145 .29836 L p 0 .27 .813 r
F P
0 g
s
.61069 .27274 m .60434 .27389 L .6444 .25285 L p 0 0 0 r
F P
0 g
s
.52132 .41504 m .51986 .40836 L .48958 .43344 L p 0 .005 .336 r
F P
0 g
s
.48958 .43344 m .48903 .44187 L .52132 .41504 L p 0 .005 .336 r
F P
0 g
s
.62932 .33424 m .6211 .32898 L .58442 .35074 L p 0 .102 .433 r
F P
0 g
s
.57683 .34159 m .58442 .35074 L .6211 .32898 L p 0 .159 .57 r
F P
0 g
s
.49236 .37092 m .48465 .36508 L .45254 .39229 L p 0 .414 .907 r
F P
0 g
s
.44419 .39207 m .45254 .39229 L .48465 .36508 L p 0 .486 .938 r
F P
0 g
s
.45856 .44184 m .46861 .4475 L .43653 .47729 L p 0 .43 .924 r
F P
0 g
s
.43653 .47729 m .4253 .47294 L .45856 .44184 L p 0 .43 .924 r
F P
0 g
s
.4357 .40271 m .43679 .41219 L .40402 .44368 L p 0 .034 .445 r
F P
0 g
s
.40783 .45488 m .40402 .44368 L .43679 .41219 L p 0 .184 .657 r
F P
0 g
s
.68557 .30228 m .68121 .31058 L .64199 .33254 L p 0 .27 .875 r
F P
0 g
s
.59942 .28766 m .60042 .27912 L .5593 .30406 L p 0 .312 .884 r
F P
0 g
s
.5593 .30406 m .5603 .3109 L .59942 .28766 L p 0 .312 .884 r
F P
0 g
s
.5008 .37934 m .50887 .38925 L .54145 .36508 L p 0 .268 .743 r
F P
0 g
s
.54857 .37475 m .54145 .36508 L .50887 .38925 L p 0 .218 .665 r
F P
0 g
s
.50887 .38925 m .51551 .39936 L .54857 .37475 L p 0 .218 .665 r
F P
0 g
s
.54145 .36508 m .53388 .35474 L .5008 .37934 L p 0 .268 .743 r
F P
0 g
s
.49236 .37092 m .5008 .37934 L .53388 .35474 L p 0 .313 .809 r
F P
0 g
s
.44906 .43322 m .45856 .44184 L .4253 .47294 L p 0 .356 .858 r
F P
0 g
s
.4253 .47294 m .41527 .46511 L .44906 .43322 L p 0 .356 .858 r
F P
0 g
s
.43679 .41219 m .44145 .42286 L .40783 .45488 L p 0 .184 .657 r
F P
0 g
s
.41527 .46511 m .40783 .45488 L .44145 .42286 L p 0 .28 .776 r
F P
0 g
s
.63661 .3355 m .62932 .33424 L .59148 .35723 L p 0 0 .158 r
F P
0 g
s
.58442 .35074 m .59148 .35723 L .62932 .33424 L p 0 .102 .433 r
F P
0 g
s
.52132 .41504 m .48903 .44187 L p .51899 .41901 L 0 0 0 r
F P
0 g
s
.48903 .44187 m .48494 .44747 L .51967 .41849 L p 0 0 0 r
F P
0 g
s
.47782 .44945 m .51507 .4182 L p .48563 .44679 L 0 .508 .873 r
F P
0 g
s
.48494 .44747 m .48476 .44752 L p .48563 .44679 L 0 .508 .873 r
F P
0 g
s
.48476 .44752 m .47782 .44945 L p .48563 .44679 L 0 .508 .873 r
F P
0 g
s
.51967 .41849 m .51971 .41841 L p .51899 .41901 L 0 0 0 r
F P
0 g
s
.51971 .41841 m .52132 .41504 L p .51899 .41901 L 0 0 0 r
F P
0 g
s
.44145 .42286 m .44906 .43322 L .41527 .46511 L p 0 .28 .776 r
F P
0 g
s
.57683 .34159 m .56969 .33098 L .53388 .35474 L p 0 .234 .704 r
F P
0 g
s
.53388 .35474 m .54145 .36508 L .57683 .34159 L p 0 .234 .704 r
F P
0 g
s
.58442 .35074 m .57683 .34159 L .54145 .36508 L p 0 .191 .623 r
F P
0 g
s
.65232 .25017 m .66151 .25196 L .61864 .27586 L p 0 .016 .152 r
F P
0 g
s
.61864 .27586 m .61069 .27274 L .65232 .25017 L p 0 .016 .152 r
F P
0 g
s
.56969 .33098 m .56393 .32029 L .52684 .34507 L p 0 .274 .773 r
F P
0 g
s
.52684 .34507 m .53388 .35474 L .56969 .33098 L p 0 .274 .773 r
F P
0 g
s
.55432 .3825 m .54857 .37475 L .51551 .39936 L p 0 .149 .55 r
F P
0 g
s
.51551 .39936 m .51986 .40836 L .55432 .3825 L p 0 .149 .55 r
F P
0 g
s
.53388 .35474 m .52684 .34507 L .49236 .37092 L p 0 .313 .809 r
F P
0 g
s
.48465 .36508 m .49236 .37092 L .52684 .34507 L p 0 .365 .873 r
F P
0 g
s
.48465 .36508 m .47868 .36264 L .44419 .39207 L p 0 .486 .938 r
F P
0 g
s
.4383 .39569 m .44419 .39207 L .47868 .36264 L p 0 .472 .693 r
F P
0 g
s
.54145 .36508 m .54857 .37475 L .58442 .35074 L p 0 .191 .623 r
F P
0 g
s
.59148 .35723 m .58442 .35074 L .54857 .37475 L p 0 .13 .502 r
F P
0 g
s
.56393 .32029 m .5603 .3109 L .52125 .33735 L p 0 .317 .842 r
F P
0 g
s
.52125 .33735 m .52684 .34507 L .56393 .32029 L p 0 .317 .842 r
F P
0 g
s
.68663 .2913 m .68557 .30228 L .64471 .32571 L p 0 .241 .823 r
F P
0 g
s
.64199 .33254 m .64471 .32571 L .68557 .30228 L p 0 .27 .875 r
F P
0 g
s
.60042 .27912 m .60434 .27389 L .56112 .30073 L p 0 .319 .803 r
F P
0 g
s
.5593 .30406 m .60042 .27912 L p .56509 .29855 L 0 .319 .803 r
F P
0 g
s
.61069 .27274 m .56553 .30142 L p .60033 .27673 L 0 0 0 r
F P
0 g
s
.60434 .27389 m .60516 .27374 L p .60033 .27673 L 0 0 0 r
F P
0 g
s
.60516 .27374 m .61069 .27274 L p .60033 .27673 L 0 0 0 r
F P
0 g
s
.56112 .30073 m .56089 .30115 L p .56509 .29855 L 0 .319 .803 r
F P
0 g
s
.56089 .30115 m .5593 .30406 L p .56509 .29855 L 0 .319 .803 r
F P
0 g
s
.54857 .37475 m .55432 .3825 L .59148 .35723 L p 0 .13 .502 r
F P
0 g
s
.55792 .38729 m .55432 .3825 L .51986 .40836 L p 0 .025 .322 r
F P
0 g
s
.51986 .40836 m .52132 .41504 L .55792 .38729 L p 0 .025 .322 r
F P
0 g
s
.66151 .25196 m .67072 .258 L .6271 .28289 L p 0 .097 .408 r
F P
0 g
s
.6271 .28289 m .61864 .27586 L .66151 .25196 L p 0 .097 .408 r
F P
0 g
s
.52684 .34507 m .52125 .33735 L .48465 .36508 L p 0 .365 .873 r
F P
0 g
s
.59148 .35723 m .59707 .36017 L .63661 .3355 L p 0 0 .158 r
F P
0 g
s
.51507 .4182 m .51967 .41849 L .48494 .44747 L p 0 .508 .873 r
F P
0 g
s
.47868 .36264 m .48465 .36508 L .52125 .33735 L p 0 .428 .92 r
F P
0 g
s
.59707 .36017 m .59148 .35723 L .55432 .3825 L p 0 .017 .264 r
F P
0 g
s
.5603 .3109 m .5593 .30406 L .51787 .3326 L p 0 .37 .903 r
F P
0 g
s
.51787 .3326 m .52125 .33735 L .5603 .3109 L p 0 .37 .903 r
F P
0 g
s
.68422 .27914 m .68663 .2913 L .64435 .31589 L p 0 .209 .738 r
F P
0 g
s
.64471 .32571 m .64435 .31589 L .68663 .2913 L p 0 .241 .823 r
F P
0 g
s
.47868 .36264 m .47529 .36397 L .4383 .39569 L p 0 .472 .693 r
F P
0 g
s
.4357 .40271 m .4383 .39569 L .47529 .36397 L p 0 0 .06 r
F P
0 g
s
.67072 .258 m .67866 .2675 L .63489 .29287 L p 0 .143 .549 r
F P
0 g
s
.63489 .29287 m .6271 .28289 L .67072 .258 L p 0 .143 .549 r
F P
0 g
s
.55432 .3825 m .55792 .38729 L .59707 .36017 L p 0 .017 .264 r
F P
0 g
s
.67866 .2675 m .68422 .27914 L .64092 .30441 L p 0 .177 .65 r
F P
0 g
s
.64435 .31589 m .64092 .30441 L .68422 .27914 L p 0 .209 .738 r
F P
0 g
s
.52125 .33735 m .51787 .3326 L .47868 .36264 L p 0 .428 .92 r
F P
0 g
s
.64092 .30441 m .63489 .29287 L .67866 .2675 L p 0 .177 .65 r
F P
0 g
s
.56553 .30142 m .56112 .30073 L .60434 .27389 L p 0 0 0 r
F P
0 g
s
.55792 .38729 m .52132 .41504 L p .55696 .38978 L 0 0 0 r
F P
0 g
s
.52132 .41504 m .51967 .41849 L .55887 .38844 L p 0 0 0 r
F P
0 g
s
.51507 .4182 m .55699 .38573 L p .52167 .41673 L 0 .45 .874 r
F P
0 g
s
.51967 .41849 m .51939 .41847 L p .52167 .41673 L 0 .45 .874 r
F P
0 g
s
.51939 .41847 m .51507 .4182 L p .52167 .41673 L 0 .45 .874 r
F P
0 g
s
.55887 .38844 m .55881 .38837 L p .55696 .38978 L 0 0 0 r
F P
0 g
s
.55881 .38837 m .55792 .38729 L p .55696 .38978 L 0 0 0 r
F P
0 g
s
.50812 .41414 m .51507 .4182 L .47782 .44945 L p 0 .454 .935 r
F P
0 g
s
.47782 .44945 m .46861 .4475 L .50812 .41414 L p 0 .454 .935 r
F P
0 g
s
.59707 .36017 m .60042 .35913 L .64199 .33254 L p 0 0 0 r
F P
0 g
s
.64471 .32571 m .64199 .33254 L .60042 .35913 L p 0 .331 .877 r
F P
0 g
s
.47529 .36397 m .47868 .36264 L .51787 .3326 L p 0 .416 .707 r
F P
0 g
s
.60042 .35913 m .59707 .36017 L .55792 .38729 L p 0 0 0 r
F P
0 g
s
.47529 .36397 m .47497 .36894 L .4357 .40271 L p 0 0 .06 r
F P
0 g
s
.43679 .41219 m .4357 .40271 L .47497 .36894 L p 0 .083 .468 r
F P
0 g
s
.5593 .30406 m .56112 .30073 L .51717 .33151 L p 0 .367 .749 r
F P
0 g
s
.51787 .3326 m .5593 .30406 L p .52049 .32935 L 0 .367 .749 r
F P
0 g
s
.56553 .30142 m .51931 .33429 L p .55771 .30347 L 0 0 0 r
F P
0 g
s
.56112 .30073 m .56155 .3008 L p .55771 .30347 L 0 0 0 r
F P
0 g
s
.56155 .3008 m .56553 .30142 L p .55771 .30347 L 0 0 0 r
F P
0 g
s
.51717 .33151 m .51724 .33161 L p .52049 .32935 L 0 .367 .749 r
F P
0 g
s
.51724 .33161 m .51787 .3326 L p .52049 .32935 L 0 .367 .749 r
F P
0 g
s
.55792 .38729 m .55887 .38844 L .60042 .35913 L p 0 0 0 r
F P
0 g
s
.49975 .40684 m .50812 .41414 L .46861 .4475 L p 0 .385 .882 r
F P
0 g
s
.46861 .4475 m .45856 .44184 L .49975 .40684 L p 0 .385 .882 r
F P
0 g
s
.61069 .27274 m .61864 .27586 L .57198 .30608 L p 0 .051 .298 r
F P
0 g
s
.57198 .30608 m .56553 .30142 L .61069 .27274 L p 0 .051 .298 r
F P
0 g
s
.51787 .3326 m .51717 .33151 L .47529 .36397 L p 0 .416 .707 r
F P
0 g
s
.55699 .38573 m .55887 .38844 L .51967 .41849 L p 0 .45 .874 r
F P
0 g
s
.47497 .36894 m .47781 .37692 L .43679 .41219 L p 0 .083 .468 r
F P
0 g
s
.44145 .42286 m .43679 .41219 L .47781 .37692 L p 0 .196 .643 r
F P
0 g
s
.64435 .31589 m .64471 .32571 L .60105 .35419 L p 0 .294 .851 r
F P
0 g
s
.60042 .35913 m .60105 .35419 L .64471 .32571 L p 0 .331 .877 r
F P
0 g
s
.47497 .36894 m .47529 .36397 L .51717 .33151 L p 0 0 .055 r
F P
0 g
s
.60105 .35419 m .60042 .35913 L .55887 .38844 L p 0 .391 .876 r
F P
0 g
s
.49113 .39729 m .49975 .40684 L .45856 .44184 L p 0 .326 .817 r
F P
0 g
s
.45856 .44184 m .44906 .43322 L .49113 .39729 L p 0 .326 .817 r
F P
0 g
s
.51931 .33429 m .51717 .33151 L .56112 .30073 L p 0 0 0 r
F P
0 g
s
.47781 .37692 m .48345 .38683 L .44145 .42286 L p 0 .196 .643 r
F P
0 g
s
.44906 .43322 m .44145 .42286 L .48345 .38683 L p 0 .267 .743 r
F P
0 g
s
.48345 .38683 m .49113 .39729 L .44906 .43322 L p 0 .267 .743 r
F P
0 g
s
.55887 .38844 m .55699 .38573 L .60105 .35419 L p 0 .391 .876 r
F P
0 g
s
.61864 .27586 m .6271 .28289 L .57959 .31411 L p 0 .131 .5 r
F P
0 g
s
.57959 .31411 m .57198 .30608 L .61864 .27586 L p 0 .131 .5 r
F P
0 g
s
.64092 .30441 m .64435 .31589 L .59883 .34597 L p 0 .253 .774 r
F P
0 g
s
.60105 .35419 m .59883 .34597 L .64435 .31589 L p 0 .294 .851 r
F P
0 g
s
.51717 .33151 m .51931 .33429 L .47497 .36894 L p 0 0 .055 r
F P
0 g
s
.5525 .37948 m .55699 .38573 L .51507 .4182 L p 0 .399 .906 r
F P
0 g
s
.51507 .4182 m .50812 .41414 L .5525 .37948 L p 0 .399 .906 r
F P
0 g
s
.6271 .28289 m .63489 .29287 L .5873 .32443 L p 0 .179 .614 r
F P
0 g
s
.5873 .32443 m .57959 .31411 L .6271 .28289 L p 0 .179 .614 r
F P
0 g
s
.47781 .37692 m .47497 .36894 L .51931 .33429 L p 0 .091 .446 r
F P
0 g
s
.63489 .29287 m .64092 .30441 L .59403 .33557 L p 0 .217 .698 r
F P
0 g
s
.59883 .34597 m .59403 .33557 L .64092 .30441 L p 0 .253 .774 r
F P
0 g
s
.59883 .34597 m .60105 .35419 L .55699 .38573 L p 0 .346 .878 r
F P
0 g
s
.59403 .33557 m .5873 .32443 L .63489 .29287 L p 0 .217 .698 r
F P
0 g
s
.56553 .30142 m .57198 .30608 L .52402 .3406 L p 0 .078 .391 r
F P
0 g
s
.52402 .3406 m .51931 .33429 L .56553 .30142 L p 0 .078 .391 r
F P
0 g
s
.55699 .38573 m .5525 .37948 L .59883 .34597 L p 0 .346 .878 r
F P
0 g
s
.54599 .37052 m .5525 .37948 L .50812 .41414 L p 0 .34 .843 r
F P
0 g
s
.50812 .41414 m .49975 .40684 L .54599 .37052 L p 0 .34 .843 r
F P
0 g
s
.51931 .33429 m .52402 .3406 L .47781 .37692 L p 0 .091 .446 r
F P
0 g
s
.48345 .38683 m .47781 .37692 L .52402 .3406 L p 0 .185 .61 r
F P
0 g
s
.59403 .33557 m .59883 .34597 L .5525 .37948 L p 0 .297 .809 r
F P
0 g
s
.57198 .30608 m .57959 .31411 L .53067 .34961 L p 0 .162 .564 r
F P
0 g
s
.53067 .34961 m .52402 .3406 L .57198 .30608 L p 0 .162 .564 r
F P
0 g
s
.53836 .36007 m .54599 .37052 L .49975 .40684 L p 0 .291 .777 r
F P
0 g
s
.49975 .40684 m .49113 .39729 L .53836 .36007 L p 0 .291 .777 r
F P
0 g
s
.52402 .3406 m .53067 .34961 L .48345 .38683 L p 0 .185 .61 r
F P
0 g
s
.49113 .39729 m .48345 .38683 L .53067 .34961 L p 0 .243 .705 r
F P
0 g
s
.5525 .37948 m .54599 .37052 L .59403 .33557 L p 0 .297 .809 r
F P
0 g
s
.53067 .34961 m .53836 .36007 L .49113 .39729 L p 0 .243 .705 r
F P
0 g
s
.5873 .32443 m .59403 .33557 L .54599 .37052 L p 0 .255 .739 r
F P
0 g
s
.57959 .31411 m .5873 .32443 L .53836 .36007 L p 0 .213 .663 r
F P
0 g
s
.53836 .36007 m .53067 .34961 L .57959 .31411 L p 0 .213 .663 r
F P
0 g
s
.54599 .37052 m .53836 .36007 L .5873 .32443 L p 0 .255 .739 r
F P
0 g
s
.25 Mabswid
.68441 0 m
.95857 .42636 L
s
.95857 .42636 m
1 .72958 L
s
1 .72958 m
.70366 .33203 L
s
.70366 .33203 m
.68441 0 L
s
.05024 .25302 m
0 .57014 L
s
0 .57014 m
.70366 .33203 L
s
.70366 .33203 m
.68441 0 L
s
.68441 0 m
.05024 .25302 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[126]:=",
  ImageSize->{288, 254.062},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000oPP10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/<70P1K1`00`0L300031`0705X7002n1`80
10L205X7002k1`<01PL00`07001I1`00^@L200T700<01`00F@L00;H70`0;1`0400L705P7002d1`80
3PL01@071`L005L7002a1`<040L01@071`L005L7002_1`804`L00`071`021`0300L705@7002/1`<0
5@L00`071`021`0300L705@7002Z1`8060L00`071`031`0300L705<7002W1`<06`L00`071`031`03
00L70587002U1`807PL00`071`031`0300L70587002R1`<080L00`071`041`0300L70547002P1`80
8`L00`071`051`0300L70507002M1`<09@L00`071`051`0300L70507002K1`80:0L00`071`061`03
00L704l7002H1`<0:PL00`071`071`0300L704h7002F1`80;@L00`071`071`0300L704h7002C1`<0
;`L00`071`081`0300L704d7002A1`80<PL00`071`091`0300L704`7002>1`<0=0L00`071`091`03
00L704`7002<1`80=`L00`071`0:1`0300L704/700291`<0>@L00`071`0:1`0300L704/700271`80
?0L00`071`0;1`0300L704X700241`<0?PL00`071`0<1`0300L704T700221`80@@L00`071`0<1`03
00L704T7001o1`<0@`L00`071`0=1`0300L704P7001m1`80APL00`071`0>1`0300L704L7001j1`<0
B0L00`071`0>1`0300L704L7001h1`80C0L00`071`0>1`0300L704H7001e1`<0CPL00`071`0?1`03
00L704D7001c1`80D@L00`071`0?1`0300L704D7001`1`<0D`L00`071`0@1`0300L704@7001^1`80
EPL00`071`0A1`0300L704<7001[1`<0F0L00`071`0A1`0300L704<7001Y1`80F`L00`071`0B1`03
00L70487001W1`80G@L00`071`0B1`0300L70487001T1`<0G`L00`071`0C1`0300L70447001R1`80
HPL00`071`0D1`0300L70407001O1`<0I0L00`071`0D1`0300L70407001M1`80I`L00`071`0E1`03
00L703l7001J1`<0J@L00`071`0F1`0300L703h7001H1`80K0L00`071`0F1`0300L703h7001E1`<0
KPL00`071`0G1`0300L703d7001C1`80L@L00`071`0H1`0300L703`7001@1`<0L`L00`071`0H1`03
00L703`7001>1`80MPL00`071`0I1`0300L703/7001;1`<0N0L00`071`0J1`0300L703X700191`80
O0L00`071`0I1`0300L703X700161`<0OPL00`071`0J1`0300L703T700141`80P@L00`071`0J1`03
00L703T700111`<0P`L00`071`0K1`0300L703P7000o1`80QPL00`071`0L1`0300L703L7000l1`<0
R0L00`071`0L1`0300L703L7000j1`805PL<06T700<01`L07@L00`071`0f1`00=`L301<7300019f4
WH@:060700<01`L07PL00`071`0e1`00=@L201472P000ij50004WPP000F4WX@0W@0405d700<01`L0
7PL00`071`0e1`00<PL300l72@001Ij4WXBN00L000RNWIjMWYfNW@H00Yd405X700<01`L07`L00`07
1`0d1`00<0L201071P001Ij5WXFN00D00Yh00`2NWP02WPL019h00`2M]P030003WYd000<0E`L00`07
1`0P1`0300L703<7000]1`<040L40004Q@24WPD001>NQIj5WXFN09jMWYfNWIjMWYfN00H000BMWYfN
0`000id0000305@700<01`L080L00`071`0c1`00:`L201071@000ij5WP0300ZN1`0:WP0309jN00<0
00BM]Yff0`000ih00003054700<01`L08@L00`071`0b1`00:0L301070P000h@00003Q@8000RN09j5
WXFNQ@D00Yh00`2NWP02WPL000VMWP2NWIjMWYd00`0019fNWIh30003W@000080C`L00`071`0R1`03
00L70347000V1`804@L20006JhD0WXFN0P001YjOWYl0WP@000RNWijOWYnN00^N1@07WP<000BM]Yff
0P0019d0001>1`0300L7028700<01`L0<@L002<70`0A1`8000B4J`000XD20005WXFNQIh00`0;WP03
09jN00^N00<0WYd00`001ifNWIjMWYd00`000id0W@020004W@0004`700<01`L08`L00`071`0`1`00
8@L201<700<0J`000P002HFNQ@2NWijOWP02000;W`2OWYnNWijOWYl01003WP@01Yh00`2NWP03WP<0
1ih20005]Yff0;H0101:1`0300L702<700<01`L0<0L001h70`0D1`0406/000>50P0018FNQIh20005
WYnNW`000Yh60004WijO00NN1`0029jMWYfNWIjM0`001YfNWIh0WP8000FMWP00W@0304T700@01`L7
2P0I1`0300L702l7000L1`805PL01P1[08D0Q@8000BOWYnN0P001InNWijO00@001VNWijOWYnNWijO
09nNWijOWYnNWijOWP2N00<01ih40008]YffW@2M]Yd20004]P0004L74`0E1`0300L702h7000I1`<0
5`L00`1[0002Q@8000B5WXFN0P001IjOWYnN008000^O09nNWijOWYnNW`0400:N1005WP0608FNQIj5
0`001ifNWIjMWYd00P0019jMWYd20005WYfN09d010121`l000<H00001P0C1`0300L702h7000G1`80
6@L02P1[08F609n509h20005WijOWYl00P04W`0309nN00H000ZOWYl0WijOWYnN1P08WP<019h0202N
0;JM]Yff0P001Iff002M00<0?PL60003600H00H000E>64hHCP0801<700<01`L0;@L001@70`0J1`<0
0XD0102OQIl2000<WYnNWijO09nNWijO10001YjOWYnNW`@000f5WhFOQ@25WhFOQ@2500<000N5WXFN
QIj500@000>NWIh00P002IfNWIjM09fNW@020004W@0003T71`03600301P000801aP:01<700<01`L0
;0L001870P0L1`8000j6QHH0WhFO09l0WijOWP8019l20006]`2OWYnN1003W`0709n5WhFOQ@060004
WXFNQ@<01ih00`2NWP03WP0609ffWKJM0P001KJM]P2f00<0=PL600DH0`001Dm;01Q>00D000Q?B`0H
BaQ;60<04PL00`071`0/1`003`L301d70P03Q@0?09n5W`2OWYl0Wih0WYnN00<000>NWih010001Yn5
WhFOQ@<001>5QXF6Q@25QXF6Q@2508FNQIj5008000bNQIj5WP2NWIjMWYd20005WYfNWIh00P001IfN
002M0080<PL700@H0`001D/0BaQ;00<024/00`0H600360<04PL00`071`0[1`003PL201h70P0028J5
QP2OQIl019l200>O0P0019ngWkL20003W`2O00>O10091`H000B6QHJ50P000ij5WP030003WP2N00:N
0P04WP8000^M]YffW@2M09d0W@0202l70`000aP00003000:600004m;Cd/0B`8000U@04m;Cd]?Bdl0
300B1`0300L702X7000>1`0300L001`70P03Q@0C09n5W`2OWYnN09h0WYl0WijOWP0200>O100E1`03
68F500:50P002hFNQIh0WXBNQ9j4008000ZNWIjM09fNWIjM0P00106M09d202`710001QP001Q;6080
00I;64/HBaP30005Bdm;C`000T/40003C4/000I;00<0B`004PL00`071`0Z1`003PL010071`0K1`09
08J5QP2OQIl000BO00T0WinO09ngWkL00P000ingW`0201/70`000hJ500020004Q@2500JN00<0WYh0
0Yh2000:]Yff0;H0]Yd0W@80:@L5000=601?Bdm;04/0Bdm;C`020007B`1;D4]@B`03000ABe1;D4]@
B`1;D4]@B`0HB`004PL00`071`0Y1`003PL01@071`L001T701L0QP2509n5W`2OWYnN09jOWP2O09jO
0003W`8080L40005Q@24QH@00P001Ij4WXBN008000nMWYfN09h0WYfN006M09d00`0V1`<000LH0000
BaQ;008000E;Cd/060020007Cd]?04m;C`020009C01;C4]<Bda;00D00T/01@1;64/0014700<01`L0
:@L000d700<01`L00`L00`071`0E1`8000N5QP2OQIl000>O0P03W`0:09ngW`2O]ing02H70P02Q@03
WXFN00801Ih200>N0P002IffWKH0]Yd0W@0302<710001QP004]?B`8000E@Be1;D0020005Be1;D4/0
1000251<D01@Be1;0`001D`0C51<00@000=;60004PL00`071`0X1`003@L00`071`041`0300L701@7
0140QHJ5W`2O09nNW`2OWYnN0003W`0309l000:O00<01`L0:0L06XE/08B5Q024WXBNQ9h0WP2NW@2M
WYfN006M0P000`40000P1`D000`H04/HBaP0Bdm;Cd/20006D01?Bdm;0P00251<D4a@C01;0`001Da@
C50000E<00<0C0000P0C1`0300L702L7000=1`0300L700D700<01`L04PL20006QXD0QYl029l01`2O
]il0WkL00P0]1`0606b509j50P05WP0@09jMWYd0WKJM]P2fWKH0]P807@L80003Cd]?008000a;D4]@
B`1;D4]@05020007C01<D4a@C003000BC55<DDaA051<D4a@04a;C4/04`L00`071`0W1`003@L00`07
1`061`0300L701070P001XJ508FOQ@8000NOWYnNWih000>O00@0WkL0<@L01P25K024Q@8000j4WX@0
Q9jMWYd0WIjMWP8000BM0@010P0K1`<00QP2000364/H008000E?Bdm;C`020005Be1;D4/00P001e1<
D01@C5000P001d`0C51<D4`02P000da;C00201<700<01`L09PL000d700<01`L01`L00`071`0?1`80
00J5QP2OQP02W`800il0202O]il0WkL0<`L2000JK000WXD0WP24WX@0WIjM09ffWKJM09ff0;H201P7
1002600;04m;04/0Be1;D4/00P003U1;D4]@B`1@04a@C01<0P001E5@D@1A00@000e<D@1AC55<D@04
04a;008050L00`071`0U1`003@L00`071`081`0300L700d70P001XF609n500>O00@0W`2N19l0102O
]`0f1`0306b5008000^408BNQ9j409fNW@020009WYfNW@2M0@0100805PL3008H0P000dm;C`03000;
Cd]?B`1;D4]@Be000P001Ta@051<D08000A<D4a@0`003e5<DDaA00A<14`400A<100401@700<01`L0
9@L000d700<01`L02@L00`071`0;1`8000N5QP2OQYl000:O00`0WkNO09ngW`2O]`0h1`8001Z5K024
WXBNQ9j409fNWIh0]YffW@2M]P2f01@70`02608000Y;Cd]?051;D01@0P001D]@Be1;00801E001`1@
001@DE000P001E40DDaA00T000P4C0A<101<01D700<01`L090L000`700<01`L02`L00`071`0:1`80
00J6Q@25W`02W`0309nO00:O0P001Il0WkL003/701P0Q@1/Q8D0WXBNQ02MWYfN09jMWYd0W@430187
0`02600504m;Cd/00P00151;D4/2000=D4]@Be1;04a@C51<D0020003C50000<000A<D@1A0`001e5<
14`4C0000`@01`1<04`0B`005PL00`071`0S1`0030L00`071`0<1`0300L700P70P001XJ508JO00>O
00L0]ing0;NO008000>gW`L0?0L06@25K8F4WP2NQ9j409fNWIh0]YffW@2M00400P0?1`<00QP2000A
Cd]?B`1;D4]@B`1@04]@Be000P05D08000=A05400P000e1AD0030005C54000@01@05100604`404`0
5PL00`071`0S1`0030L00`071`0=1`0300L700H70`001XD0QIn500:O00/0WinO0;NO]`2g00101`0C
K6]/08D0WXBNQ02MWYfN09jMWP020003W@2M00803@L2008H0P0014]?Bdl20004Be1;D08000a<D01@
Be00D4a@C5020008C540D01A0543000;DD`0601<1004C0000P@02P0400@014`0B`0G1`0300L70287
000<1`0300L700h700<01`L01@L20006QHH0WhH00Yl02@2gWkNO09ng0002]`0300L703l700F5K8D0
WP02000AQ9j409fNWIh0]P2f0;JM00400P0:1`@000hH04m;Cd]?051;D4]@00E@00<0D50015020003
DE1A008000=@DE000P000e4000021`0604/004`41`02100304`400<060L00`071`0Q1`0030L00`07
1`0?1`0300L700<70`001HH0WhD000>O00/0WinO09ng0;NO00131`0@06]/08F408BNQ02MWYfN00>M
00D00Id0W@0200P70P003T/H001?Bdm;C`1@Be1;0P001E1<D4a@008000m<D4a@C01<DDaA0540DD`0
0`051`0301Q<008000E<10@40002100600@4101<0P0I1`0300L70247000<1`0300L7010700<01`L0
0PL02026QP2OQYl00Yl02`2O]il0WkNO0;L004D701P0K8D0WX@0Q9j409fNW@2M]Yff0;JM004200D7
0`0024lH04/0Bdm;0P000e1;D00200I@00<0D@000e020003D55@008000=A05400P021`0300L700<7
00@0Bda<0`02100400@400<4100J1`0300L70207000;1`0300L7018700<01`L00P001XJ5W`2O00>O
00/0WinO0;NO09ng00161`8000e/08F408BNQ02M09d000BM00D00Id0W@0200<70P005AQ;001;Cd/0
B`1;D4]@051<D4a@C00200=@00H0D01AC5420003D540008010L00`071`031`800T`010A<00021005
00@400@00P000d`0000J1`0300L70207000;1`0300L701<700P01`00QP26W`<000VO0;NO]`2gW`00
0[L00`071`171`0I08D0WX@0Q9j409fNW@2M]Yff0;JM00401`03000864l0D4]@Be020003Be1;0080
1U001@1@DE1A008000M@DE1A055@00801`L00`071`021`0304/0009<0`@20003100400<000A_06l0
70L00`071`0O1`002`L00`071`0D1`8000J5QP2OQP02W`800il01P2O]`2g04/700L0K025Q0240080
19d0202MWId0W@6M10005D/HB`1?Bdm;C`1@B`1;04a@C51<D00200A@00P0D@1AC540D@802@L00`07
1`021`0304/0009<00<4C0@00`@20004100410<07@L00`071`0N1`002`L00`071`0D1`800XH01026
W`02W`0809l0W`2O]`02]`0300L704X700@0Q@2N0P003XBN09jMWYd0WKJM09ff0`0024lH001@Be1;
0P03D0801U001@1AD55@008000QAD540DE1A00/700<01`L00PL01Da;04`0008400D010@400021003
00A_00<07@L00`071`0N1`002`L00`071`0D1`8000F508JO0003W`0309nO00D000>O00L0C@L02@1/
08D0WHBM0003W@0309fM00:M10005T/H04]?Bdm;04]@Be00D01@C51<D004D005051AD5400P000e00
000<1`0800L71`1;04`2000314`0008400T010@400@0C0007`L00`071`0M1`002`L00`071`0C1`<0
00F609n60002W`0:0;RO^02hW`2g05070P003XB509j4WP2NWIh0]Yff0P002QQ?001;D4]@B`04D005
051@D0000e00101AD5420006D55@D@1F0P0>1`0700L71`1<0002C08000D400@40002100606l404`0
80L00`071`0L1`002`L00`071`0C1`8000J6Q@26W`02W`0:09nO0;NO09ng050700X0W@24Q@2MQ9d0
0id4000864/0Cd]?Bdl20005Be1;D00015000`1<D004D005055@DE00100?1`800PL300A<00H40000
1002100300@400<080L00`071`0L1`002PL00`071`0C1`<000F608JO0003W`090;RO09nh0;P00587
00`00@24Q@2NQ9h0WYd2000:C`1?051;D4]@00A@0P05D08000Q@DE1A055@D@@03PL300031`070080
00Y;04a<06m<00A_2@0R1`0300L701/7000:1`0300L701<700P0K02609n600:O00X0WinO0;NO09l0
D`L02@2M08B509f4W@030008BaP004m;Cd/20005D4]@B`001U000`1@0003D006055@D@1A0P0<1`<0
1@L3000AB`1<C01<04`4C01<14`0C0@00P0S1`0300L701X7000:1`0300L701<70P02QP0408JO00:O
00P0WkP0^9l0]`80E0L01`0108@0Q9h00`002Dm;04/0Be1;0005D003051@00=@0P00255@D@1AD55@
100;1`<020L2000AC01<C01<K`1_1004K`@016l00P0T1`0300L701X7000:1`0300L701870`001HD0
QYl000>O00T0Wil0WkL0]`00EPL0102M08@3000>64/H04]@B`1;04]@Be0200E@00<0D5000U020006
D55@05002PL300X70P004T/004`0C0A<04`400A<1004C0809@L00`071`0I1`002PL00`071`0B1`08
06`0QP2OQP02W`070;RO^02hW`0305L700<00@000P0024]?B`1;D4]@0P04D003051@00A@00d0D55@
D@1F055@EU4000P70`0=1`8001A;04a<06m<00A_1004K`1_101<02H700<01`L060L000X700<01`L0
4PL026d008D0QYl00Yl02@2OW`2gW`2O001I1`<000lHBaP0Be1;D01@051;D0001E0200=@0P001UI@
001A00L70`0?1`8000=;04`00`0034`400A<04`4C01<10809`L00`071`0H1`002@L00`071`0C1`03
BhJ600@00il02@2hW`2O^02h001H1`<000M;C`00Be1;00800e000`1@0005D009055@DE00DEIA00<0
1PL3014701L06000C6i<04`0C004K`1_1004K`1_000X1`0300L701L700091`0300L701<700I]QXJ6
WhH400>O00<0]il00`1F1`<000i;64/0D4]@B`1<D4a@C0801E00101@D5020003D5H000<010L301@7
0P0024/0C4`0C01<0`001d`0C0@014`00P0Y1`0300L701L700091`0300L7018700X0BhJ6X8JP0:2O
1`000kL01`1E1`<000U?Bdl0D4]@B`001E000`1@0005D00:DE1A05IAEP1F00<70`0F1`8000a<06i<
04a_06l0Kd`70003C00702T700<01`L05PL000T700<01`L04PL200>600>OQP000Yl01P2OW`2O]`@0
E0L30007BaQ;051;D0020004C51<D0800e000`1@D004D00305I@008000=F00L00`0H1`@00d`03P1<
C4`0Kda_06m<04a_0P0/1`0300L701D700091`0300L701870P003XH0X8JP0:2O09nh0;RO0`000`L0
1`1A1`8000U;04]?051;D0001E001P1AD55@D@8000Y@DE1A05IA055F0`0K1`<0015^C6h0Kda_06l0
K`1_1004K`0202d700<01`L05@L000T700<01`L04PL20005QP26W`000il01P2OW`2O]`<00PL00`07
1`1>1`8000<HBaP00P002U1;001@C51<D005D004055@00=@00=FD0000`0M1`8000=;04`00T`0401<
C01_C6l0Kd`0C6l0K`0^1`0300L701@700091`0300L701870P003XH0X8H0Wj00^9l0WkRO0P041`03
00L704/70`000d]?B`020003D4/000E@0P002E5@DE1A05I@EP020003DEH000D070L2000EB`1<04`0
C6m<04a_C004K`1_06l0030700<01`L04`L000P700<01`L04`L01aP0QP26W`000Yl02@2OW`2gW`2g
00051`0300L704T70`003QQ;601;04/0C51<D4`01E020005DE1A05H00`0015I@09d301X70P001T/0
Bfi;00=<00H0C6m<04`20004K`1_034700<01`L04`L000P700<01`L04`L04Vd0QP2PQP2OX02hW`2O
^02h00H700<01`L0A`L30006Bdm;04m@0P04D08000Y@DE1AD01@EU1F1@001UIA0Id00@806PL2000?
K@1^C6i<04`0C6m<00A_008000<40000<`L00`071`0B1`0020L00`071`0B1`0801P0QP26W`02W`06
09nO0;NO0`071`80APL3000=64/H04]@0500D4a@0006D003051A00<000AFD01@0`001Id1006M0080
60L20008B`1^Bfh0KT`2000;Kda_06m<04`0C000=0L00`071`0A1`0020L00`071`0B1`800XH0302P
QP2OX02hW`2O^0<02@L00`071`121`<000M;Cd/0Ce0000=@00T0D01AD55@DE001@003U5FD@1E09ff
W@2M0@010`0F1`80019]06i<KP1_C6l0K`1_06l400@303D700<01`L04@L000P700<01`L04PL20006
QP2OQYl00Yl01P2OW`2gW`<02PL00`071`101`<000LHB`00Be0000A@0P05D0<000EAD01@EP020003
D00000>M00H00Id1004201H700h0B`1^Bfh0KTa^06m<K`T0=PL00`071`0@1`0020L00`071`0B1`80
00j608JP0:2OX02hW`2O^08030L00`071`0n1`<000M;C`1@Ce0000A@00D0D5H0D@030009EU1FD01A
EP1F008000^MWP2fWKH00Id00@0301@701T0KD/0C6h0Kda_06m<Kd`0KVl0K`1_00L003H700<01`L0
3`L000L700<01`L04`L40003WhH000:O00T0Wil0WkL0]`003@L00`071`0l1`<000LHB`1@Be0000A@
00<0D5400`02D00?DE1AD01@EP1F09f409d000BM00D00Id0W@0201<70P005D/0Bfh0KTa^06m<Kd`0
C6l0K`1_00031`0300L703<700<01`L03`L000L700<01`L04`L00d^7QP040008X9l0WkP0^9l300h7
00<01`L0>PL40006C`1@Ce001504000;EU0005I@EU1FDEH00`0039j4WP2NW@2M]Yd0W@800P420147
10004da^06m<K`1_C6m<06i_06l0K`0010L00`071`0d1`0300L700h700071`0300L701<700I]QXN6
X8H:0003]`0700h700<01`L0>0L300076000D4]@0003D0<000iAD55@051F055@EU1FD08000J409f4
W@02W@0309fM008000BM0@010P0@1`<001A^06h0KTa^06m<Kd`0C6l0K`1_00H700<01`L0<`L00`07
1`0>1`001`L00`071`0C1`0BBhN6X8JP0:2O09nh0;RO0;L040L00`071`0f1`<000M;001@Ce00009@
0`002E1FD5I@055FD@03000FDEIE00L0Q8D0WX@0WIh0]P01006M00810P0>1`<0011^Bfh0Kda_06m<
Kd`0KVl00Vl200P700<01`L0<`L00`071`0=1`001`L00`071`0C1`0BK@270:2609nP0:2O09ng0;L0
4@L00`071`0d1`<000HH04/0Be04000>D01@DE1AD01@EU00D5H400870P001f/0Q9f409d00P03W@05
006M09d00P0>1`80019^Bfh0KP1^06m<Kd`0C6l0Kd`200X700<01`L0<`L00`071`0<1`001`L00`07
1`0C1`80012608JP0:2O09nh0;RO0;L04PL00`071`0b1`<000E;04a@C00300=@00/0D5I@EU00DEIA
0002EP0305H000L000B408BN0P002Id0W@6M0Id100020@8030L20004KT]^00=^0P001da_C01^K`00
0Vl00`1^000:1`0300L703<700<01`L030L000L700<01`L04`L2000@Q`2PQP2OX02PW`2O]`2g01<7
00<01`L0;0L70003601;00@015001@1@DE1A008000a@EU00D5H0EP1[06/>00>M00D1W@400@0200/7
0P004fi;KP1^C6i<04`0C01<K`1_C6l00P0<1`0300L703<700<01`L02`L000H700<01`L050L2000?
QP26X02PW`2O^02hW`2g00804`L00`071`0X1`T000I;0000C50200=@00/0D5I@EP1FDEIA0002EP03
05H000L016/01@1`Jg1[00H000<1W@000P4200T70P0016i;KP04KP0406i_KP8000=_06l00Vl200h7
00<01`L0<@L300`700061`0300L701D7011;Q`2PQP2OX02PW`2O]`2g0P0D1`0300L702@710001F/0
001[00@000E;D4]@0002D00J0500D55@D@1FD5I@051F05H0L6]`Jg1[L6/80005L01`07002@071`<0
00m^Bfh0KTa^C01<KT`0C6l00`000fl0000@1`0300L702d70`021`0300L700X700061`0300L701D7
00m]08L0X8H0Wj00^9l0WkP00`0E1`0300L7020710000f/0J`02000?J`0004/0C51<D01AD55@0080
00UFD5H0EU5FD@000UH00`1F000800A`00@0L7E`1P001W1[L6]`J`P000P70000KT]^00A^00T0KVm^
06i_KP000Vl3010700<01`L0:PL300D700<01`L02PL000H700<01`L05@L03d/0QP26X02PW`2OX02g
W`0301H700<01`L07@L30003J`1[00X000E;D4]@0004D00I051F055@EU1FD01@EU1F07E`MG1eL7E`
M@060008CG1=L4e`CG050006L01`071=200056h0KTa^C01<KT`0C6m<06i_06l04PL00`071`0U1`@0
20L00`071`0:1`001PL00`071`0E1`0?K@270:2609nP0;RO09nh00<05`L00`071`0J1``000=`0000
0P003Ta@C500DE1AD01@EU1F0P000e5FD@05EP8000e`MG1eL7E`MG1eL7D0009e10002GE`MG1eL7E`
M@0500E`2P0026i_KP1^Kfh00Vl00`1_000D1`0300L702470`0<1`0300L700X700061`0300L701D7
0P003HH0QZ00X9l0Wj00]il00`0H1`0300L701L72`000f/000020007601;D4]@0004D00B051FD5H0
EU00D5I@EU1F07E`20002gE`M@1eCGE=MDee00D000U`CG1=L4d0CGD010001Te`CG1=L0X000M<06i_
06l001D700<01`L07PL3010700<01`L02@L000D700<01`L05PL2000=Q`2PQP2OX02hW`2O^00301T7
00<01`L050L40003J`00008000=[L6/01@004AQ;051<D01AD55@051FD5H000=F1`0;M@H02WD40003
L01e00Ie1@05L0`05`L00`071`0J1`<04`L00`071`091`001@L00`071`0F1`8000V608JP0:2O09l0
1`0J1`0300L701873@03L00801PH04`0D004D00:051FD5H0EU1F009F00<0E@001@003WE`MG1eL7E`
MG1e07EB10002WE=MDeeCGE=MDd4000:MDeeCGE=MDeeC@@000I`C@1=MDd;01@700<01`L05@L401H7
00<01`L02@L000D700<01`L05PL40003X00000@000BO09nh0P000kH01`0I1`0300L701472007L00B
01Q;051;D01AD55@051FD5H00eH0201FEP1EIgD017D800=e00<0MGD017D500Ie00<0MGD01@09M@@0
00Q`MG1eL7E`M@/03`L00`071`0A1`<06PL00`071`091`001@L00`071`0F1`03BhN6008000bP0:2O
09nPW`2O0;L201/70P0?1`H000=`07001W02008H00@0C4]<0P03D00:051FD5H0EU1F009F00H0EEI`
M@0<M@@000]eDWEBME9eDWEBM@04000:C@1BME9eDWEBM@@000U=MDeeCGD0ME8010002GE=MDeeCGE=
M@050003L00000<02PL00`071`0>1`<07@L00`071`091`001@L00`071`0F1`800XL00j270002X007
0;RO^02hW`0301d700<01`L02@L70004Jg1[009`1@004WD0B`1@Be00DE00D01@EU1F00=F00H0EUH0
EFL400me00<0MGX01009M@@02WD50003M@1e00Ee1@002GE`MG1eL01BM@040003CG0000@01PL00`07
1`091`@080L00`071`091`001@L00`071`0G1`0?KXH0QZ2609nP0:2O09nh008000<100L070L00`07
1`051`<00PL50003L00000<000EeL7E`M@030004C4]<00A@0P0025I@EP1FD5H00UH00`1EEP04M@04
07EBM@D01gD01`1eDWEBME801@00259eDWEB059e1@002WEBME9eDWEBME840009ME9eDWEBME9e00@0
00U=04eeCGE=MDd010000g00L00800L70`0T1`0300L700T700051`0300L701L700M;08NPQj0000:P
00P0WkRO09l0]`807PL010071`L300@70`000g1[00020003L7D000Ee0140601@Be00DE1AD01AEP1F
0003EP0605IFEUEW0`03M@0807YeNWEjMGX5009e00`0MGYeNWEjMGYeNWD40009M@1jMGYeNWEj00@0
1gD00`1eNP0400Ue1@00259eDWEBME9e1@000g1=000=02@700<01`L02@L000@700<01`L060L2000>
Q`2PQP2OX02PWj00^9l301l70`061`H000E`MG1eL0090003Bd`000A@00D0D5I@EP030003EP1F009F
0P00159jDWX4000<DWEBME9eDWEBME9e0`003gEBME9eDWEBME9eDWEBM@02000;DWEBME9eDWEBME80
1@002E80DWEBME9eDP050009DWEBME80DWEB00H000IeCGE=MDd400==400O1`0300L700P700041`03
00L701P70P001XL0Qj2700:P00H0^9l0WkP301`70`000`L01`051`D000EeL7E`M@020005M@1eNWD0
0`003D/0055@DE00DEIAEP000eH60003MGX000L000=eNWD020001WYe07EjM@H000YjMGYeNWEj07Ye
10002WYeNWEjMGYeNWD4000:NWEjMGYeNWEjM@@000]e07YBNU9jDWYBNP050005ME80DWD01@000g1=
L00@01X700<01`L020L000@700<01`L060L0401;08H0QZ00X9l0Wj2O09l20003]P0701L70`051`03
00L700871P000g1eL00200Ae00<0MGD00P02B`03051@009@00H0D5I@EP03EP0405IFEP8000AINUUj
3`000gYBNP040008DWYBNU9jDWX50009ME80DWYBNU9j00D000UeDWEBME80DWX010002U9eDWEBME9e
DWD4000<DWEBME9eDWEBME9e1@05DPD00dd?01H700<01`L020L000@700<01`L06@L20005Q`2PQ`00
0Z00202hWkP0^02g0P0E1`<02@L010071`L4009`0`05M@H00T/02P1@DE00DEIAEP03EP0406MFIah0
00=INP0010001GYeNWEj00<000]jMGYeNWEjMGYeNP040009M@1BNU9jDWYB00D000UeNWEjMGX0NU80
10003GYBNU9jDWYBNU80DWX00`001WEBM@1eDPD000A=MDee3`0A1`0300L700P700041`0300L701T7
0P003XH0QZ00X9nP0:2O09nh0`0B1`<03@L00`0700030004L4e`00Ee0P000gEjM@03000;B`1<0500
D5I@EP000eH0101FEUHK0003E000008000IB001jDWX50005NU80DWX010002E9jDWYBNU9jDP02000<
NU9jDWYBNU9jDWYB10002GD0NU9jDWYBNP040007DWEBME9e0005DP@00e80101BDU860003C@1=00`0
3`L00`071`081`0010L00`071`0I1`0804/0Q`27X002X00409nhW`@000>f00L03PL201471003L080
17D01P2CMI=eT`8000E;051AD0020004EU5F00=F00@0IeIW0`04C@0804eCCE==D`09D`0=06ECIE=U
DfECI@1UE002DP@000Yj0000NUUjFGYI1@0027EjMGYe059j1@002GYBNU9jDWYBNP02000;DWYBNU9j
DWYBNU80100037X0NU9jDWYBNU9jDP<000YB059jDWYBNU9j1@02C@T03PL00`071`081`000`L00`07
1`0J1`0=06h0Q`2PQP2OX02PW`020003^02g008030L301@70P00171=L002M@0707Ee07YeNP030006
D`1;051<0U0015H0EP03EP0605IFEP1W30000deC0005Da0000@H06][0h@015800002F@08059jDWYB
NU840009NP1jDWYBNU9j00@000UBNU9jDWX0NU8010002GYBNU9jDWYBNP0500QB00<0DP001007DP03
059B00D000==04d0100=1`0300L700P700031`0300L701X700P0B`2708NP00>P0P001:00WkP20003
0@0700L70`0G1`05071`L0000gD30006TgD009Qj10001e5@D@1FDEH00`001EH0IeIW00D01E<01@1U
DfEC00P016D00`1UI@07I@<000M]06/0Q9d000:M00<0WId00P0015UjFGX4000?FGX0059jDWYBNU9j
DWYB00D000QB059jDWYBNPD000UjDWYBNP1jDWX01@002WYBNU80DP1BNU860008NU80DUUBFE8900`7
00<01`L020L000<700<01`L06`L20007QhH0QZ00X0020007Wj2O09l0W`0200D70`0I1`8000==L000
0gD0102CMI<30004TgX0NP8000U@C500EU1FD0000UH50005I`1CDU<010006U9C06ECIE=UD`1DIEAU
E6EDIE@0E6EDIEAU0`0016e;K@02J`>400<0W@000id01@01W@2M00@000II0000NUT4000:NU9jDWYB
NU9jDP8000]jDWYBNU9jDWYBNP04000:DP1BFE9IDUUBF@D000]B05UBFE9IDUUBF@0500AB00<0DU80
0U800`1B000=1`0300L700L700031`0300L701/72@002:00^:2h0;RO0P000kH01`0301`71003M@08
09=eT`2HTiP3000;V000D55@055FD@000eH0101FIeH40003I5<000=U50000fD000050003KVd000<0
00F409f40002W@0709fM09d100020@<000YU0000FGX0NUUj1@002GYBNU9jDP1INP040009DWYBNU9j
DWYB008000ejDWYBNU9jDWYBNU9j00D000UjDWYBNU9jDP000UT60007DUUBF@1B000<1`0300L700L7
00031`0300L701/700I^C027X8L5X0080:2OX02h0;L301l700<0C@000gD0302CMI<0TgZC09P0V080
00Q;04`0D5I@00=F00@0EVMF0`002e80001B06AUI6D000MU00d0IVEVIFIUIVD0IFIU00<000Y/04/0
KFi]06/00X@0102MQ002W@06006M0@011@05I@03DUUB00<000=INP000`002E80FGYINUUjF@040007
NU9jDWYB0002F@D000QIDUUBFE9I00=I1P000e9I0007F@T02`L00`071`071`000`L00`071`0L1`06
KXN8Qj271:00102P^:0500030@0701d701<0L01e07D0TgFC09RCV02H09P0008700Q<B`00EU5F009F
00H0EVMF06L20004I01YI@D00VD04P1ZIFYUJVD0IFYUJVEZIFX0JP@00Vd20003KFh000<000A[08@0
0Yd06P2MWId00Id00@0106EZIFYUJP1VIFIU05Uj10003UT007YINUUjFGYINUUj10002U80FGYINUUj
FGX5000<NP1j07YINUUjFGYI1@06F@08009I0UT0F@0;1`0300L700L700031`0300L701`700P0KXN7
0:2700>P1@000il0W`0201h70`000g1e0002000:T`2CNY<0V02H00<700L0Be1A05H000=F00<0IeH0
10000fD00003I@0406EZI@X000=V00002@0026e/001^06h00Vh20008J`24Q024W@02W@0609d1006M
2@002fD0001UJVEZI@1B008015T50007FGYINUUjF@0200eI0P000eT0F@09F@0305UI00D000=I05T0
0eT300/700<01`L01`L0008700<01`L07PL01d`0QhP0X8L010001[RP^02hW`8000>f00L070L00`1`
0002M@03ThjC00<000>H09P00`041`8000=AD5400`000eH000030005I`00IFX00P002FEY06YUJVEZ
0006JP0306YZ00EZ0`02K@8000=]Qfd00P03KP8000E[08BM0002W@0909fM006M0@0100P000=UJP00
3`001EUjFGYI00D000IINP1oFGl5000:NUUjFGYINP1oF@@000ijFGYINUT0F@9I0UT2F@X02PL00`07
1`071`000PL00`071`0N1`0404b7Q`@00j00202PW`2O^02g0P0L1`<000MeL01eTgD000:C10000iP0
1`041`8000=AD5H00UH20005EVL0IfH00P02I@0G06D006YUJVEZ06YVJVIZIP1VJVIZIVX010003V`0
06i]KVd0KHL0KP1^0P001f/0Q024W@000id01`01W@2M00400`0H1`D00UT200UI1@000eT0F@03F@03
05UI00@000=I05T025T0201I0UT200802`L00`071`071`000PL00`071`0O1`03KP00008000>7X000
0j001P2hX02O^0@06`L3009e00@0Thh00i<0102C0;4200H700=<D0000eH01P1FIeH0I`8000H6001V
JVH>0003JP0000D00Vd01@1]Qff7008000=^Qfh00P0028Q^002409d00id01`2MW@2M0@000P4201l7
0`000gmIO`050007OeUoFGmIO`02000;FGmIOeT0FGmIOeT00P000`9I00020004F@9I00D200<00P00
2`L00`071`071`000PL00`071`0M1`<000M^08L0X8L000>P00D0Wj00^0020004W`0001X700T0C@1e
09=eT`000Y<01P2C09Ra00P700Y;001F05IWEP1V0P02IP0F06EZIFX0JVIZIVX01VX6JPIZ06X6JP<0
00M/K@00KFi]008000>7KHL00P000fh0KP02KP<00X@0102MQ002W@04006M0@D08PL00e800004F@D0
0UT01@1I0UT200H00eT01P1I0UT2F@`030L00`071`061`000PL00`071`0J1`<00`L20006Q`27X8L0
0j00202P^02hW`2g0`0I1`08C@1eM@2>T`03T`040;40/@802@L00`1@EP060007I`1V06H6IP02000?
1P00JPH01VX6JPIZ06X600@001=]0027KHM]06j7KXL0Qfj7KP1^00<000A[08@00Yd01P2MWId00@80
00@100409`L00eT00002000?FGl005UoFGmIOeUo07mI00<000YI0UT2F@9I0UT01@8200d700<01`L0
1PL0008700<01`L05`L300L700M<08L0Qj0000>P00T0X000WkRO09l00`0H1`0807D0M@2CM@02T`05
0;6C09P00P0:1`0D0500EFME06IWIP1V1VH0IPIV1VH200031P0000T000M]001]KVe^008000=]Qfd0
0P001HM^Qfj7008000a^001[08@0WHBM09d30006W@010040:`L300050UT2F@801`001P9I05T2F@@0
00<2F@001@8200h700<01`L01PL000031`0701H70`0:1`8000J80:27X002X08000^hX02O^9l0]`01
000G1`03CGEe00<000JC09>C09<30003/@0000X70`000eIWEP02000<1VH6IP1Z1VX6JPH01@H00`06
JP030007K000Q`27K@020003Qfj7008000E^Qfj7KP020004R6h0KP<000B4W@2M0`001id0^Id00KT0
0P0`1`03DP00008000@2F@9I0`080PT03`L00`071`061`0000<700L04`L300h700]^08L0X8L0X02P
0003X00909nhW`2O09d001H70P00172CSY<70006/@0300L030L40009IfIW06MVIfIW00<000<60000
1`001fd006f6KHH00P000ff7K@02000;Qfj7KXM^06j8KXP00`02J`@019d01P2MWId0W@80=PL50005
F@9I0UT010050P<040L00`071`061`0000<700L040L301470P0018L0Qj02000?X:6P0:2hX02O^9l0
]`01008050L2000<MHjCSZbC[02a[02/0`021`0300L700d700MEEP06IPIV00X70`0026`008M]Qff7
0P002Vj7KP1^Qfj7KXL20003KXQ^008000>800000V/3Q003WH@000>M00D0WKT0^@0303X73`0B1`03
00L700H700000`L01`0=1`<05@L50003X@2P00>P00X0X:2P09nhW`2O0P000ih01`0B1`8000=`08h0
19<01@2C[02a00<00`L00`071`0K1`<000M]001]QVf600<000>708L00P001Fj7KXM^008000>8KXP0
0P003Fh0000706/0Q024WH@00id0202MW@2M0040>`L00`071`0M1`0300L700H700000`L01`0:1`<0
60L2000BR8L0X:40XJ2QX02P^:00WkRO0P000kL0]P0201470P002WD0Thh0Tj`0/J`200030`0000@7
00<01`L06@L20007K000Qff7K@020004Qff7K@8000F7KXM^Q`020003KXQ^008000>8K`000P021`03
06/000:400>MQ9d00id0202M^@2i0;T0?0L00`071`0L1`0300L700H700000`L01`071`<070L20006
R8L0Qj401:00202PX:00WkRO0P000il0000201070P001G00SY<000:C00H0T`2/0:`200D700<01`L0
5PL20007K@00KHI]QP020004KHM]Q`8000i^06j7KXL0R6j8KP1^R0<010L00`1[J`03Q003WH@000:M
00H0WId0W@4203`700<01`L07@L00`071`051`0000<700L010L302070P006HR70:2Q0:6PX@2hX;RP
09l0WkP0]il0W@003`L2000;M@2C0:bC0:ba0;4010061`0300L701870`001V`008M]Q`<000B7KHM]
0P04Q`0706h008Q^R0020003K`0000801@L20007J`2408BM0002W@0609fiW@010`0l1`0300L701`7
00<01`L01@L0000400L71`<090L20006R8L0Qj400j000`2PX002X08000>PWj000P0019l0000?1`05
M@00SP000Y<00`2/T`030003[00000L700<01`L03PL30007J`0006d0K@020005Qff7KHL00P001Vj7
KXM^Q`8000=^R6h00P001HQ^06h000P700P0J`2409f400>M00D0WId00@0303d700<01`L06`L00`07
1`051`00100X1`8001N8Q`2PX@2QX:40^:2h0:00^9nhW`2O]`0400`70WD30004[9<0[08000H300<0
0`071`0300L700X70`0026`0001]QVf60P001Ff7KHM]00801HL30003R028008000=_R0000`091`03
06/000:400@0WH@00Yd0202iW@010040?PL00`071`0J1`0300L700D7000202/70P001XR708NQ00>P
00<0X0000j020003Wj2O008000>O000010091`04072>SPD000R/0:`0[03300P700<01`L01PL30008
J`0006f6KHH30005Q`27KHL00P001Fj7KXM^008000B7KXM^0P001XQ^002800/70`02Q00408BM00:M
00H0WId00Id203h700<01`L06PL00`071`051`000PL00`071`0Y1`8000V8Q`2PX@2QX:400P05X005
09nhWkP00P0019l009h300L700<0M@000Y<00jbC0002[00500>/0:`00P091`0300L700870`0026`0
001]QVf60`001HN6QhJ700<018L30004QhR7R08000F706l0K`0200d700P0J`2409f400:M00H0^Ifi
0;T203l700<01`L06@L00`071`051`000`L00`071`0Y1`8000R8Q`27X@2P00:P00<0X:000j001@2O
X9nP00800Yl30004WP0000@700P0L02>08j/00:/00D0[:`00`0200T71@0026/0001]QVf60`001XM]
Qff7K@<000A^Qfj70P001Fj7KXM^008000>8KXP00P000fh01`0>1`0806/0Q024W@03W@0509fM09d0
0`0o1`0300L701P700<01`L01@L000@700<01`L0:@L20004R8L0X08000BQX:6P0P04X08000BhW`2O
0P03W`800Yh400081`00Thj/T`02[00500>/0:`00`041`@000=[00000`02QP0308H0008000F7QXN6
Q`0400:700<0Q`000P0018R7R8L30003QhR7008000>8000040L02@5[08@0WHBM0002W@0509fi0;T0
0`101`0300L701P700<01`L010L000D70P0Z1`8000J80027XHL5X0030:2P00>P0P0019nPWj0300:O
0`03Q@H000>/0:`00Z`00`2/000=0004K8I/QPD000F7KHM]Q`030006Qfj7KXM^0P001HM^Qfj70080
00A^R6j80P000fl0000B1`<00X@0102MQ002W@0809fM09d00@101`0300L701P700<01`L010L000L7
00<01`L0:0L200>800>7XHL00P001:6PX@05X08000JOX9nPWj0200:O1`000ij5000@00:61`001XN6
Q`27QP@01XL3000:Q`27R8N808R7R0<000>800005@L200:400@0Q9d00Yd01P2M^@2i0@80@@L00`07
1`0G1`0300L700@700081`0300L702T700Q<R8N80:67X@800Z0200>P00X0Q`00X9nPWj2O0`001XJO
QYn6WaD02HH40007KXM^Qfj7KP0300F700H0KP1^R6h20004R6l001L700P0J`2409f400:M00@0WIfM
10121`0300L701H700<01`L010L000T70P0Z1`041028R08000Z7XHL0X:6P0:001:030005X9nPWj00
100078JOQYn6WhJO0:26X8JPQZ26X8JPQP26QhJ7QXL500:700<0QhL00hL30007QhR7R8N8Q`020005
R8N8Q`000XP201T70`001H@0Q9d000>M00D0^@000@0204<700<01`L05@L00`071`041`002`L00`07
1`0X1`<000J8Q`27XHL200:P0P001j27X8NPQj000P001inPW`26X8H01P000in6000<QP0308H000@0
00Z7QXN6QhJ7QP1^0`07Q`<000N706j808Q_00<06PL02P1[002408BM09d30003W@0100<0@PL00`07
1`0E1`0300L700@7000<1`0300L702T700A_08R80P001j67X@2Q0:400`06X0<000Z6X8JPQZ26X8JP
4@0;Q`@000N7R8N8QhR700<000B7R8N80P03R08070L01@1[RHB900D000JR0;VM004304<700<01`L0
50L00`071`041`003@L00`071`0Y1`8000f8QhP0XHNQ0:27X8NP008000J7X8NPQj04000NQZ26X8JP
QZ2608J7QXN6QhJ7QXN6Q`27QXN6QhJ71@02Q`0308N700>70`0018N8QhP30003R00000807PL00`1[
Q002Q003WHBM00>M00P0WId00@2i04@700<01`L04`L00`071`041`003PL202X70`02R00;0:40XHL0
Qj67XHL00P04X00408NPQ`L001:6X02PQZ26X8JPQZ26X8JP08L500R700<0QhP00P0018N8QhP30003
R00000<08@L2000@Q8V4XXBRWJ80^Id00@0104D700<01`L04`L00`071`031`0040L00`071`0Y1`0;
C027R02QQj67XHL00P000j27X004000;Qj27X8NPQj27X8L04P0;Q`@000B7R8N80`001HR7001_0080
8`L200:400@0Q9d00Yd0202MW@010;T0A@L00`071`0C1`0300L700<7000A1`0300L702T70P002XP0
R027XHNQQj420004Qj67X@D001RPQj27X8NPQj00X8NPQj27X8NPQj27X007Q`D000J8Q`27R8L200:8
0P000hP0000V1`0B06^4R@2RQ02MXP2iW@010040APL00`071`0B1`0300L700<7000B1`80:PL200B8
0P001HNQQj6700<000J708NPQj070003X8L000f700<0QhL01@001HR7R8N800D000>8QhP00P000fh0
1`0X1`800X@01024W@02W@0809fM0040^@171`0300L7014700<01`L00`L001@700<01`L0:@L00`@0
R003R0<001:7XHNQQ`27XHNQQj67XHNQQj430003X00000d000F8QhR7R00500>800L0R8P0KXP002/7
0100J`290::409fR0;VM006i0P181`0300L7010700<01`L00`L001D700<01`L0:`L00`@0R003R080
00JQQj40R8L50008Qj27X8NPQ`0=Q`0508N8QhP00`001HR7R8N800@000=^K`00;PL30005Q024W@00
0Yd01P2MWId0W@80B0L00`071`0@1`0300L700<7000F1`0300L702d700<400000XP30008XHNQQj67
XHL50009Q`27XHNQQj6700P000>808P018P50004R00003470`003HV408@0WJ:M09fi0;T00`181`03
00L700l700<01`L00`L001L70P0`1`<000n808N8QhR7R8N8QhR7R8L00P02Q`D000b7R8N8QhR708N8
QhP30003R6l00080=@L300:400@0WH@00Yd01@2MW@2M00<0B@L00`071`0>1`0300L700<7000I1`03
00L7030700<0R8P00hP7000;Qj6708RQR:68XHP020001HP0R6n800800PL00`071`0f1`800129Q024
XP2RW@2M^@2i0;T0B@L00`071`0?1`0300L70087000J1`0300L703@71004R0@000>8Q`0010000fn8
K`050003K`0700H700<01`L0=PL0201[08@0WH@00Yd0202MW@2M0040BPL00`071`0>1`0300L70087
000K1`80>`L400>81@001jbW0:O30<<00`071`0300L703H70180J`2408BR0::M09fi0;T0^@1;1`03
00L700d700<01`L00PL001d700<01`L0?@L03gD0SP2WSP2W[:L0Yj`0_P0300D700<01`L0=`L30005
Q02MQ0000Yd01@2MW@2M00<0C0L00`071`0<1`0300L70087000N1`0300L703`700`0SP2W0:bW0:N/
0<<20004Yl<000D700<01`L0=`L00f/000020008Q:80XYd0WKT30003^@0704/700<01`L030L00`07
1`021`007`L00`071`0k1`<000N>08jW0:bW00800ZL00`2W000200@700<01`L0=`L0101[Jh@400:M
00P0W@00W@0104d700<01`L02`L00`071`021`0080L203`700L0M@2>0:N/008000NW0:O3Y`2W00<0
0`L00`071`0g1`07J`29Q::4XP050006^@2i0;T0CPL00`071`0:1`0300L70087000R1`0300L703T7
0`000hh0Y`0200>W00H0YjL0_ZL200<700<01`L0=`L00`1[J`04Q00;XXBR0::M09d00@00C`L00`07
1`091`0300L70087000S1`0300L703T71@02Y`090:N/Y`2W``3300<00PL00`071`0g1`0BJ`29Q::4
XYfRW@2M^@2i0;T0C`L00`071`0:1`0300L70047000T1`80>@L02@1`SXh0SZN>0002Y`060:NWY`2n
0`000`L01`0h1`0306][00B400^RQ:80XYd0W@01001@1`0300L700T700<01`L00@L002H700<01`L0
=`L01@1`Yhh000:W00h0[:N/0<>W0:L0_P0703T7019[08V408BRWJ:M09fi0;T0^@1A1`0300L700P7
00<01`L00@L002L700<01`L0=`L0201`SXh0SZL00jL01@2WY`2W00@0>@L01@1[Jh@000:400P0Q:80
XYd0W@80D`L00`071`071`0300L70047000X1`0300L703H700D0SP2W0003Y`070:bW0:O3Y`020003
_P0003T70P003HV408BR0::M09fi0;T00P1C1`0300L700L700<01`L00@L002T70P0g1`0808h0SP2W
SP03Y`070:L00;h0_P0203P70P000f/0Q002Q00:08BR0:80^@2i05@700<01`L01PL00`071`011`00
:`L00`071`0e1`8000>>Y`000ZL00`2W[0020006Ykh0_P2n0P0g1`8000n90::409fR0;VM09d00@00
E@L00`071`051`0300L70047000/1`0300L703@70`001Hh0SZN>008000ZW0:NRY`2n0;h0=`L2000?
J`24Q02RQ02MXP2i0;T005H700<01`L010L00`071`011`00;@L00`071`0d1`060700SP2W0P03Y`09
0;jW_P2n0;h003H70P003hT0XX@0WJ80^Id0W@01001F1`0300L700D70@011`47000^1`80=0L3000B
SP00RJL0Yj:W0:NRY`2n0;h0=0L3000?J`24Q02RQ02MXP2i0;T005L700<01`L010L100470@L00307
00<01`L0<PL01P2>0:N>00:W00`0YjNW0;jW_P2n0;h2038700<0J`000P003J:409fR0;VM09d00@00
F0L00`071`031`400@L11`00<@L00`071`0b1`800129SXT0RJL0Yj:W0:NRY`2n0P000j801`0`1`80
00>4J`000`001j:409fR0;T00`1I1`0300L700870@011`47000b1`80<PL01`1`08jWSP000ZL03P2W
YjL0_ZNn0;jR0:80;`L20006JhB9Q::410000kT000020003^@0705P700<01`L00PL100470@L003@7
00<01`L0<0L05`1`08V>R@29Y`2WXZL0Y`2W0:NR0:8002h700/0J`24JhV4R@2RQ0020003XP2i0080
F`L01@071`L00087000e1`0300L7030700L0L02>Yhh000:W00h0YjNW0:NRY`2nXP2R02d70100J`24
08BRQ02MXP2iXP2R0P1L1`0400L70087000f1`0300L7030701L0L029SXT0RJL0Y`2W0:NRY`2WXP2R
000/1`0B06/0Q029Q024XP2RW@2M0;T0G@L00`0700021`00=`L2034700P0L02>Yhh0Y`<000bRY`2W
XZL0_Z80XP0Z1`<000f9Q024XP2RW@2R^@2i0080GPL010071`011`00>@L00`071`0_1`0F0700RHj9
08VWR@2RY`2WXZL0Yj80XP80:0L3000=Jh@0RH@0Q:80XYd0W@0205l700<01`000@L003X700<01`L0
;`L04`1`08h0R@29YhT0XZL0Yj:W0;h00`000kT01`0V1`0@06/0R@2RQ:80XP2iXP2R^@80H0L20047
000k1`80<0L04@1`002>R@29YhT0XZL0Yj:W00800Z800`2R000U1`@000a[08B90::409fR0;T305l7
0`011`00?@L00`071`0^1`8001F>0:N908VWR@2RY`2W0:NR0::i0;T00P0S1`<000f4R@2RQ02MXP2i
XP2R0080G0L400@7000n1`0300L702h70P002hh0SXT0RJN90::W00800j800`2RXP020003W@070207
0P003f/0Q6/0Q8T0XP2RW@2M^@0205P710081`00?`L00`071`0^1`8000b>0:N908VWR@2R0:820009
Yj800;VR04d001l70`003f^90024XX@0W@2R^@2iXP0205@7100<1`00@0L202l70P0028h0SXT0RJN9
0P000j:9XP0200:R00D0XZ80W@0201d70P0016/0JhT40009Q:80XYd0WKT00547100@1`00@PL00`07
1`0]1`8001>>0:N908T0RJL0Yj:WXP2RY`2i00<000>M00006PL20008Jg00RHBRQ:830005^J80X[T0
0P1=1`@050L004<700<01`L0;@L20004SP2>R@8000RWRJL0XXVRR@8000FR002MXP020003Q00001L7
0`002V/0RF/0Q8V408@704X7100H1`00A0L202d70`002Hh0Y`2W0:N9Y`020004XZNRY`8000FR0000
XP020003W@0701@700l0J`1`J`2408BRQ029XP000Z80102R^@161`@070L004H700<01`L0:PL01@07
002>008000f9SP00RJN908VRR@29008000>RQ0001P0B1`0?06/006/0JhT0RH@0Q:8000:R00@0X[T0
@`L4020700171`0300L702T700<01`L00`0048T0RJN908VWR@2R08VRRJ8200>R1`0<1`<000A[08U[
0P002HT0XXBR0::90002XP040;T003l7100T1`00B0L00`071`0X1`0400L71`<000E`002>R@02000:
YhT00::9XXVRR@8000BR0::420031`L000A[08A[0P002HU[08B90::40002XP040;VR03`7100X1`00
B@L202P700<01`L00PL03G00002>0:N9Y`2W0:800P001XVRRJ:9XP<000FRQ::4XP0<000AJ`00RF/0
Q02408BR0::9XP000Z80102R000h1`@0;0L004/700<01`L09@L00`071`041`040700SP8000N9SP00
RJ:9008000NRRJ:908B900@000V4XP24JhA[Q6/00P004XA[0024J`1[RF/0Q8V408BR00:R00@0X[T0
=@L40307001<1`0300L702@700<01`L01@L02@1`08j908T0R@020003XXVR008000V908VRRJ:9XXT0
2@000f^4J`02000?RF/0Q8V408T0Q029XXT000:R00@0^@00<@L403@7001=1`0300L702<700<01`L0
1PL40003L00000>90P000hVRR@03000>Q8V4RHB9Q024RHB9Q8T4000;J`1[06^9J`24RH@00P000j:4
0002XP040;T002h7100h1`00CPL202<700<01`L020L01@1=002>008000>9XXT00P001:80XXT40005
XXT0Q8T01P0018B9Q8T2000=Q8V408VRR@29XP2R0002XP030:8002/7100l1`00D0L00`071`0P1`03
00L700T700@0C@000XT200>90`001HV4RHB900D000N9Q8V406^900<000b908U[08B9Q029XXT702P7
10101`00D@L00`071`0O1`0300L700X70P002gD008V>0::908VR00D000VR08V4RHB9Q8T00`05R@80
00R9XXT0RJ:900:R00<0Y`009PL404@7001B1`807`L00`071`0;1`800W020004R@0000N92P001XB9
Q8V4R@8000f4RH@0RJ:908VR0:L002<710181`00E0L00`071`0L1`0300L700d70P001G1e08j900<0
00N9XXVRRJ8000N900@0RHV90`02R@0708T0RJ:90002XP030:L00207101<1`00E@L00`071`0K1`03
00L700l70P0017000002R@D000V908B9Q8V4RH@010001HV4RHB9008000V9XXT0RJ80Y`007@L40507
001F1`0300L701X700<01`L04@L20004L00000F92P06R@8000FRR@290002XP030:L001X7101D1`00
E`L201X700<01`L04`L20003RG0000800hT00`29R@05R@0508B9Q8T00P04R@<000BR00005`L405P7
001I1`0300L701L700<01`L05@L40006L00008T01hT01`29002RRJ800`000jN9000201D7101L1`00
FPL00`071`0F1`0300L701T73`02R@<000>>R@000P0D1`@0H0L005/70P0F1`0300L701d73@000jN9
00030187101T1`00G@L00`071`0C1`0300L702073@0A1`@0J0L005h700<01`L04PL00`071`0j1`@0
K0L005l700<01`L04@L00`071`0f1`@0L0L006070P0A1`0300L70387101d1`00HPL00`071`0=1`03
00L702l7101h1`00H`L00`071`0<1`0300L702/7101l1`00I0L200`700<01`L09`L40807001V1`03
00L700T700<01`L08`L408@7001W1`0300L700P700<01`L07`L408P7001X1`0300L700L700<01`L0
6`L408`7001Y1`801`L00`071`0G1`@0T0L006/700<01`L010L00`071`0C1`@0U0L006`700<01`L0
0`L00`071`0?1`@0V0L006d70P031`0300L700/7102L1`00K`L010071`091`@0X0L0070700<01`00
1@L40:@7001a1`8000<700000P2X1`00LPL20:`70000\
\>"],
  ImageRangeCache->{{{0, 287}, {253.062, 0}} -> {-0.000338972, 0.0322486, 
  0.00331925, 0.00331925}},
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \("-MeshGraphics3D-"\)], "Output",
  CellLabel->"Out[126]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[BoxData[
    \(MVClear[]; MVShow3D[colorknot]; \)], "Input",
  CellLabel->"In[127]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[36]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F2F30402F2F502F304F2F2F4F2F304F2F2F4F2F30402F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F2F2F3F2F2F4F2F30
502F2F602F30602F306F2F3F6F2F30702F3F6F2F30702F3F6F2F306F2F306F2F306F2F30602F30\
602F305F2F30602F305F2F2F5F2F30502F2F502F304F2F2F4F2F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F2F30402F2F4F2F30502F30602F3F602F306F2F3F6F2F3F
702F40702F3F7F2F407F2F40802F40802F40802F407F2F3F802F407F2F3F802F407F2F3F802F40\
7F2F3F7F2F3F7F2F3F7F2F3F7F2F3F7F2F3F7F2F307F303F7F2F30702F30602F30
5F2F30000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F2F2F\
4F2F304F2F305F2F305F2F30602F3F602F3F6F2F3F702F3F7F2F40702F40802F40
802F408F2F4F8F2F4F8F2F4F8F2F40902F4F8F2F40902F4F8F2F40902F4F8F2F408F2F408F2F40\
8F2F408F2F408F2F408F2F408F2F40802F3F8F2F40802F3F8F2F3F802F30802F3F
7F2F30702F30602F2F5F2F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F2F303030302F2F303030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000402F305F2F30602F3F\
602F3F6F2F406F2F3F702F40702F407F2F4F7F2F40802F4F802F4F8F2F508F2F4F
902F50902F4F902F50902F509F2F50902F4F9F2F50902F4F9F2F4F902F4F9F2F50902F4F9F2F4F\
902F4F9F2F4F9F2F4F9F2F4F902F4F9F2F4F902F4F9F2F4F902F40902F408F2F3F
8F2F3F802F3F802F3F7F2F30702F30602F2F5F2F30000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F30303030303F3030302F303F30303030303F30303030303F3030303030
3030303030303F303030303030303030302F3F3030303030303030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F202F502F305F2F3F6F2F406F2040\
702F407F2F40802F4F802F4F802F4F802F4F8F2F508F2F4F902F50902F50902F50
902F509F2F509F2F509F2F5F9F2F4F9F2F509F2F50A02F509F2F4FA02F509F2F4FA02F509F2F4F\
A02F509F2F4FA02F509F2F4FA02F4F9F2F4FA02F509F2F4FA02F4F9F2F409F2F40
902F3F902F408F2F3F8F2F3F802F30802F3F7F2F30702F30602F2F5F2F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000303030302F2F\
303030302F2F30303030303030303F30303030303F303030303030303030303F30
303030303030303030303030303030303030303030303030302F30303030302F2F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F2F30502F30602F3F6F2F40702F4F7F2F4F802F50\
802F4F8F2F508F2F4F902F508F2F50902F50902F509F2F5F9F2F509F2F5F9F2F5F
A02F5FA02F5FA02F60A02F5FAF2F5FA02F5FAF2F5FA02F50AF305FA02F50AF2F5FA02F50AF305F\
A02F50AF2F5FA02F50AF305FA02F50AF2F50A02F50AF3050A02F4FA02F4FA02F4F
A0304FA02F409F2F4F9F2F409F3040902F3F902F408F2F3F8F303F7F2F30702F306F2F30603030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003030303F30303030303F30303030303F303F\
3030303F3F3F30303F3F3F4030303F3F3F3F30303F3F3F3F303F3F3F3F3F303F3F
3F3F3F303F3F3F3F3F30303F3F3F3F30303F3F3F3F3030303F3F3F3030303030303030303F3030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F2F502F305F2F3F602F406F2F407F2F4F7F2F4F802F508F2F50\
8F2F508F2F50902F50902F509F2F5F902F509F2F5F9F2F5FA02F5F9F2F5FA02F60
A02F5FAF2F60A02F60AF2F60A02F5FAF2F5FA02F5FAF2F5FAF2F5FAF2F5FAF2F5FAF2F5FAF2F5F\
AF2F5FAF2F50AF2F5FAF2F50AF2F5FAF2F50AF2F5FAF2F4FAF2F50A02F4FAF2F4F
A02F4FAF2F4FA02F40A02F4F9F2F40A02F40902F409F2F40902F3F902F3F8F2F3F8F2F3F7F2F30\
702F306F2F2F602F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003030303F3030302F303030303030303F303030303030303F30303F\
3F3F3F30303F303F4030303F3F3F3F30303F303F40303F3F303F40303F3F303F40
303F3F303F40303F3F303F4030303F303F4030303F303F3F303030303F3F30303030303030302F\
303030302F2F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F2F304F2F305F2F3F602F3F702F4F702F4F802F50802F508F2F5F8F2F5F902F60\
902F5F9F2F5F9F2F5FA02F5F9F2F5FA02F60A02F60A02F60A02F60AF306FA02F60
AF2F6FA02F60AF306FAF2F6FAF2F60AF2F60AF3060AF2F60AF2F60AF2F60B03060AF2F60B02F60\
AF2F5FB03060AF2F5FB02F60AF2F5FB03060B02F5FB02F50B02F50B02F50AF2F50
B02F50AF2F4FAF2F50AF2F4FAF2F4FA02F4FAF2F4FA02F40A02F4F9F2F409F2F40902F3F902F40\
8F2F3F8F2F3F7F2F30702F303F2F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F30303030303F30303030303F303F30303F3F3F3F30303F3F3F40303F3F3F3F40\
303F403F3F40303F403F3F40303F403F3F4F303F403F3F4F303F403F404F303F40
3F404F303F403F404F303F403F3F40303F403F3F40303F3F303F4030303F3F3F3F30303F303F3F\
3030303F303030302F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
30202F4F2F30502F3F602F40602040702F4F7F2F4F802F50802F508F2F5F8F2F5F9F2F609F2F60\
9F2F609F2F5FA02F609F2F60A02F60A02F60A02F6FA02F60AF2F6FA02F6FAF2F6F
A02F6FAF2F6FAF2F6FAF2F70AF2F60AF2F6FAF2F60AF2F6FAF2F60AF2F60AF2F60B02F60AF2F60\
AF2F60AF2F60B02F60AF2F60B02F60AF2F60B02F60AF2F50B02F5FB02F50B02F50
B02F50B02F50AF2F4FB02F50AF2F4FAF2F4FAF2F4FAF2F4FA02F40AF2F4FA02F40A02F409F2F40\
9F2F40902F3F902F3F802F307F2F306F2F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F3030303030303030303F30303030303F30303F303F4030303F303F40303F4030404F303F40\
303F4F303F4F303F4F303F4F303F4F303F4F304050303F4F30404F303F4F30404F
303F4F30404F303F4F30404F303F4F303F4F303F40303F4F303040303F4030303F303F4030303F\
303F3F303030303030302F2F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F2F30402F30
5F2F3F5F2F406F2F4F702F4F7F2F50802F508F2F5F8F2F5F902F60902F609F2F609F2F60A0306F\
A02F60A02F6FA02F60AF306FA02F6FAF306FAF2F6FAF3070AF2F6FAF3070AF2F70
B03070AF2F70B0307FAF2F70B03070AF2F6FB03070AF2F6FB03070AF2F6FB0306FAF2F6FB0306F\
AF2F6FB0306FAF2F6FB0306FAF2F6FB0306FAF2F5FB03060B02F5FB0305FB02F5F
BF305FB02F50BF2F5FB02F50BF3050B02F50B02F50AF2F4FB03050AF2F4FAF2F4FAF2F4FAF304F\
A02F4FA02F409F2F3F9F3040902F3F8F2F3F702F306F3030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F303F303030
3F3F3F30303F3F3F3F30303F3F3F40303F403F3F4F303F403F404F303F4F3F405030404F3F4050\
303F503F4050303F503F405F3040503F405F3040503F405F3040503F4050304050
3F40503040503F40503040503F4050303F503F4050303F4F3F4050303F4F3F3F4F303F403F3F40\
303F3F303F3F30303F3F303F303030000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000402F304F2F305F2F40
602F40702F4F702F4F7F2F50802F508F2F5F8F2F5F902F60902F609F2F6F9F2F60A02F6FA02F6F\
A02F70A02F6FAF2F6FA02F6FAF3070A02F6FAF2F70AF2F70AF3070AF2F70AF2F70
AF2F70AF3070A02F70AF2F70A02F6FAF2F6FA02F6FAF2F6FAF2F6FAF306FAF2F6FAF2F6FAF2F6F\
AF306FAF2F6FAF2F6FAF2F6FB03070AF2F6FB02F60AF2F60B03060AF2F5FB02F5F
B02F5FB02F5FB02F50B02F5FB02F50BF2F50B02F50B02F50B02F4FB02F50AF2F4FB02F4FAF2F4F\
AF2F4FAF2F40AF2F409F2F3F9F2F40902F3F902F3F802F30702F30502F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003030303030303F303F30303030303F
30303F3F3F40303040303F4F303F4F3F3F4F303F4F304050303F4F3F4050304050304050304050\
3F4F5F303F5F30405F30405F3F405F30405F30405F30405F3F405F30405030405F
30405030405F30405030405F30405030405F303F50303F50303F4F303F50303F4F303F4F303F40\
303F40303F3F30303F303030303030302F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F2F3F502F3F602F4F6F2F4F
7F2F507F2F50802F5F802F5F902F5F902F5F9F2F609F2F60A02F6F9F2F6FA03070A02F70AF307F\
A02F70AF3070AF2F70AF307FAF2F70AF307FAF2F70AF3070AF2F70AF3070AF2F70
AF2F70A02F70AF3070A02F70AF2F6FA02F60AF306FA02F6FAF2F6FA02F6FAF306FAF2F6FAF306F\
AF2F6FAF306FAF2F6FAF306FAF2F6FAF306FAF2F60B03060AF2F60B0306FB02F60
B03060B02F5FBF3060B02F5FBF305FB02F5FBF305FBF2F50BF2F5FB02F50BF3050B02F50B02F50\
B02F40B0304FAF2F40AF2F40A02F40A030409F2F3F902F3F8F2F308030306F2F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003030303F303F30303F3F3F3F30303F3F3F40303040
3F3F4F303F4F3F4050303F503F40503040503F405F3040503F405F30405F3F4F5F30405F3F4F60\
3040603F4F603040603F40603040603F4F603040603F406030405F3F4F6030405F
3F4F6030405F3F4F6030405F3F4F6030405F3F405F303F5F3F405F303F503F4050303F4F303F4F\
303F403F404F303040303F3F303030303030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F2F3F50203F6F2F4F702F4F7F2F50
7F2050802F5F8F2F5F902F60902F5F9F2F609F2F60A02F6F9F2F6FA02F70A02F70AF2F7FA02F7F\
AF2F7FA02F70AF2F70A02F70AF2F70A02F70AF2F70A02F6FAF2F70A02F6FAF2F70
A02F6FA02F70A02F6FA02F709F2F60A02F60A02F60A02F6FA02F60A02F6FA02F60A02F6FA02F60\
A02F6FA02F60A02F6FA02F60A02F6FA02F60AF2F60A02F5FAF2F60AF2F60B02F60
AF2F60B02F60B02F60B02F60B02F5FB02F5FB02F5FBF2F5FB02F50BF2F5FBF2F50BF2F50B02F4F\
BF2F4FB02F4FB02F4FAF2F40AF2F40AF2F40A02F409F2F3F9F2F3F902F308F2F3F
7F2F2F6F2F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000030303030303030303F30303F3F303F30303F303F40303F403F3F4F
303F4F303F50303F503F405F303F5030405F30405F3F406030405F304F603040603F4F60304060\
304F6F304F60304F6F30406030406F304060304F6F304060304060304060304F60
304060304060304060304F6030405F304060303F5F30405F303F50303F5F303F50303F50303F4F\
303F4F303F4F303F4F30303F30303F303030303030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F2F3F502F406F2F50702F50802F5F802F5F
8F2F608F2F5F902F60902F609F306F9F2F6FA03070A02F70AF307FA02F7FAF307FAF2F7FAF3080\
AF2F7FAF307FA02F70AF3070AF2F70AF2F70A02F6FAF3070A02F6FAF2F70A02F6F
A03070A02F6FA02F709F2F6FA030709F2F609F2F609F2F60A0306F9F2F60A02F6F9F2F60A0306F\
9F2F60A02F6F9F2F60A0306FA02F60A02F6FA02F5FA03060A02F60AF3060AF2F60
AF306FAF2F60B0306FAF2F60B0306FB02F60B03060B02F5FBF3060B02F5FBF305FBF2F5FBF3050\
BF2F4FBF3050B02F4FBF304FB02F4FB02F4FAF2F40AF3040A02F40A02F409F2F3F
90303F802F307F2F30502F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030303F303F30303F3F3F4030303F3F3F40303F403F3F4F303F503F4050303F50
3F405F30405F3F406030405F3F4F603040603F4F603040603F4F6F304F6F3F4F6F304F6F3F5070\
304F6F3F5070304F6F3F4F70304F6F3F4F7030406F3F4F70304F6F3F4F6F30406F
3F4F6F304F6F3F4F6F3040603F4F6F3040603F4060303F603F4060303F5F3F405F303F503F405F\
303F5030405030304F3F3F4F303040303F3F303030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F2F4F2F3F502F3F6F2F4F702F507F2F5F802F5F8F2F60
902F609F2F60902F609F2F6F9F2F6FA02F70A02F70AF2F7FA02F7FAF307FA02F7FAF2F7FA02F7F\
AF2F7FA02F70AF2F70A02F6FAF2F70A02F6FA02F6FA02F6FA02F6F9F2F6F9F2F6F
9F2F6F9F2F6F902F6F902F6F902F6F902F60902F5F902F60902F60902F60902F60902F60902F60\
9F2F60902F609F2F60902F609F2F60902F609F2F5F9F2F5FA02F609F2F5FA02F60
A02F5FAF2F60A02F60AF2F60AF2F60B0306FAF2F60B02F60B02F60B03060B02F5FBF2F50B02F50\
BF2F50BF2F4FBF2F4FBF2F4FBF2F4FB02F40B02F4FAF2F40B02F40A02F40A02F40
9F2F3F9F2F3F8F2F30802F30702F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000303030303030\
3F303F30303F303F40303F403F3F4F303F4F303F50303F503F405F303F5F304060
3040603F4060304060304F6F3040603F4F6F30406F304F6F304F6F3F4F70304F6F304F70304F6F\
3F5070304F703050703040703F4F70304070304F703040703F4F7030406F304F70
30406F3F4F6F30406F304F6F30406F304F6F303F60304060303F60304060303F5F30405F303F5F\
30405F303F50303F5030304F303F4F30303F30303F302F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F2F4F2F3F502F406F2F50702F50802F5F802F5F902F6F902F6F
9F2F6F9F2F6FA03070A02F70A0307FA02F70AF307FAF2F7FAF307FAF2F7FAF307FAF2F7FAF2F7F\
A02F70AF2F7FA02F70A02F70A02F6FA02F709F2F6F9F2F6F9F2F6F9F2F6F902F6F
9F2F6F902F6F902F6F8F2F6F8F2F6F8F2F5F8F2F608F2F608F2F608F2F5F8F2F608F2F608F2F60\
8F2F60902F608F2F60902F608F2F60902F60902F5F902F5F902F5F9F2F609F2F5F
A02F60A02F5FA03060A02F60AF3060AF2F60B0306FAF2F60B0306FB02F60B0306FB02F5FBF305F\
B02F50BF3050BF2F50C03050BF2F4FBF3050BF2F4FBF2F4FB02F40B0304FAF2F40
AF2F40A02F3FA03040902F3F8F2F3F7F2F30703030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003030303F303F30303F3F3F40\
3030403F3F4F303F4F3F4050303F503F405F30405F3F40603040603F4F6F304060
3F4F6F30406F3F4F70304F6F3F4F70304F703F5070304F703F507F304F703F507F304F703F507F\
304F703F507F30507F3F507F304F7F3F4F7F304F703F4F7F304F703F4F70304F70
3F4F70304F6F3F4F70304F6F3F4F70304F6F3F407030406F3F406F3040603F406F304060304060\
303F5F3F4060303F5F303F5030304F303F4F30304030303F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F2F4F2F3F5020406F2F4F702F507F2F5F80205F8F2F60902F6F9F2F70
9F2F70A02F70A02F70A02F7FA02F7FAF2F7FA02F7FAF2F7FA02F70AF2F7FA02F70AF2F7FA02F70\
A02F70A02F70A02F709F2F6F9F2F6F902F6F9F2F6F902F60902F6F8F2F60902F6F
8F2F608F2F6F802F60802F6F7F2F607F2F5F7F2F5F802F5F7F2F5F802F5F7F2F5F802F607F2F5F\
802F607F2F5F802F607F2F5F802F60802F5F802F5F802F508F2F5F8F2F5F902F5F
902F5F9F2F5F9F2F5FA02F609F2F5FA02F60A02F5FAF2F60A02F60AF2F60AF2F60B0305FB02F5F\
B02F5FB02F50BF3050BF2F50C02F50BF2F4FBF2F4FB02F4FBF2F4FB02F40B02F4F
AF2F40AF2F40A02F3FA02F409F2F3F902F3F802F307F2F30502F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000303030302F3030303F30303F3F3F40303040\
303F4F303F4F3F4050303F5030405F3040603F4F6F304060304F6F30406F3F4F70
30406F304F70304F703F4F70304F70304F7F304F703F507F304F7030507F304F7F3F507F304F7F\
30507F304F7F3F507F304F7F304F7F304F7F304F7F304070304F7F304F70304F70
304070304F70304F70304F70304070304F7030406F304070303F6F30406F303F60304060303F5F\
30406030305F303F5F303050303F5030304F303F40302F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F304F2F3F502F406F2F50702F50802F5F802F60902F6F902F6F9F2F709F2F70
A0307FA02F7FAF307FAF2F7FAF3080AF2F7FAF307FAF2F7FAF307FAF2F7FAF2F7FA02F70AF2F7F\
A02F70A02F709F2F709F2F70902F6F902F6F902F6F902F6F8F2F6F8F2F6F802F60
802F6F7F2F607F2F60702F5F702F606F2F5F6F2F5F6F2F50702F5F6F2F5F702F5F6F2F5F70305F\
702F5F702F5F702F5F70305F702F5F7F2F5F7F2F5080305F802F5F8F2F5F8F2F5F
90305F902F5F902F5F902F5F9F30609F2F5FA03060A02F5FA03060A02F60AF305FAF2F5FB03060\
B02F5FBF305FBF2F5FBF305FBF2F50C03050C02F4FC03050BF2F4FBF304FBF2F4F
BF304FB02F40B02F4FAF2F40AF3040A02F3F9F2F3F902F3F8F3030702F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003030303F303030303F3F3F40303F403F3F4F303F4F3F4050\
303F503F405F30405F3F4F60304F6F3F4F6F304F6F3F5070304F703F4F7F304F70
3F507F304F7F3F507F304F7F3F508030507F3F5080304F7F3F508030507F3F508030507F3F5080\
3050803F508030507F3F5080304F803F4F80304F7F3F4F80304F7F3F4F7F304F7F
3F4F7F304F7F3F4F7F304F703F4F7F304F703F407030406F3F407030406F3F406F303F603F406F\
304060303F60303F5F3F3F5F303F50303F5030304F303040303030000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F2F\
4F2F3F502F406F2F50702F507F2F5F802F5F8F2F6F902F6F9F2F709F2F70A02F7F
A02F7FAF308FA02F7FAF2F7FA02F7FAF2F7FA02F70AF2F7FA02F70AF2F7FA02F70A02F709F2F70\
9F2F70902F6F902F708F2F6F8F2F6F802F60802F607F2F607F2F60702F5F702F60
6F2F5F6F2F5F602F5F602F5F5F2F50502F504F2F4F502F4F502F4F5F2F50502F4F5F2F505F2F50\
5F2F505F2F505F2F505F2F50602F505F2F50602F506F2F50702F50702F507F2F5F
7F2F50802F5F802F508F2F5F8F2F5F902F5F902F5F9F2F5F9F2F5FA02F609F2F50A02F5FA02F5F\
AF305FAF2F5FB02F5FB02F5FBF305FBF2F50C03050C02F50C03050BF2F4FBF2F4F
BF2F4FBF2F4FB02F40B02F4FAF2F40AF2F40A02F3FA02F3F902F308F2F30702F2F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003030303F303F30303F303F4F30304F3F3F4F303F4F304050303F5F\
3F4060304060304F6F304F6F3F4F70304F6F3F5070304F703F507F30507F3F4F7F
304F7F3F5080304F7F3F5080304F803F5080304F7F3050803050803F5080305080305080305080\
3F50803050803050803050803F5080304F7F304F80304F7F3F4F80304F7F304F7F
304F7F3F4F7F304070304F7F304F7030407F303F7030407030406F304070303F6F30406F303F6F\
30406F303060303F6030305F303F5F303050303F5030304030303F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F304F2F3F\
502F406F2F50702F507F2F5F802F60902F6F902F6F9F2F709F2F70A0307FA02F80
AF308FAF2F80AF3080AF2F7FAF2F7FAF2F70AF2F7FAF2F70AF2F7FA02F70A02F7F9F2F709F2F70\
902F70902F708F2F6F8F2F6F7F2F607F2F60702F5F702F5F6F2F5F6F2F5F602F50
5F2F504F2F4F402F4F3F2F40000000000000000000000000000000000000000000000000302F40\
302F403F2F403F2F40402F4F402F4F4F2F4F4F2F4F5F2F505F2F4F6F2F506F2F50
702F50702F507F2F5F7F2F50802F5F802F508F305F8F2F5F902F5F902F509F305F9F2F50A0305F\
A02F5FAF305FAF2F5FB03060B02F5FBF305FBF2F50C0305FC02F50C03050C02F4F
C03050BF2F4FBF304FBF2F4FBF2F4FB02F40B0304FA02F3FA02F3F9F2F308F303F7F2F2F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F30303030303F3F403030403F3F4F303F4F3F3F50303F503F40603040603F4F6F\
30406F3F4F70304F703F5070304F703F507F3F507F3F508030507F3F5F80305080
3F508F3050803F508F3050803F508F3050803F508F3050803F508F3050803F5F8F30508F3F508F\
3050803F5F8F3050803F508F3050803F508F304F803F4F80304F803F5080304F7F
3F4F7F304F7F3F4F7F304F7F3F4F7F3040703F407F3040703040703040703F4070304070304070\
303F6F3F3F6F303F60303F6030305F303F5F30305030304F302F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000302F304F2F3F502040\
6F2F50702F507F2F5F80205F8F2F6F902F6F9F2F709F2F70A02F7FA02F7FAF308F
A02F80AF2F80A02F80AF2F7FA02F70AF2F7FA02F70A02F709F2F70A02F709F2F709F2F70902F6F\
902F6F802F6F802F6F702F60702F60602F5F5F2F504F2F4F4F2F4F3F2F40000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3F402F404F2F4F502F4F
5F2F4F602F506F2F50702F50702F507F2F507F2F50802F5F802F508F2F50902F509F2F509F2F50\
A02F5FA02F50AF2F5FAF2F5FB02F5FB02F5FBF305FB02F50BF2F5FBF2F50C03050
C02F4FC02F50BF2F4FC02F4FBF2F40BF2F4FB02F3FB02F40A02F3FA02F3F9F2F30902F307F2F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F3030303030303F3F3F4030304F303F50303F503F405F303F5F3040603040603F4F6F30406F\
304F70304F703F507F304F7F3F507F30507F3F50803050803F5F8F3050803F5F8F
304F8030508F30508F3F508F30508F30508F30508F3F508F30508F30508F30508F3F508F305080\
30508F3050803F508F30508030508F304F80304F80304F80304F80304F7F304F7F
30407F304F7F304F7F304F7F30407030407F30407030407F303F70304070303F6F304070303F6F\
303F6F30306F303F6F303060303F6030305F30305F302F4030303F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F304F2F40502F406F2F50\
702F5F7F2F60802F60902F6F902F6F9F2F709F2F70A03080A02F80AF308FAF2F80
AF308FAF2F80AF2F80AF2F7FAF2F7FA02F70A02F709F2F70A02F709F2F709F2F70902F6F8F2F6F\
802F6F7F2F6F702F606F2F605F2F504F2F4F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F3F
40304F4F2F4F502F505F2F4F6030506F2F50702F5F702F507F305F7F2F508F2F508F2F509F3050\
9F2F50A0305FA02F50AF305FAF2F5FB03060B02F5FBF3060BF2F5FBF305FBF2F50
C0305FC02F50CF3050C02F4FC03050BF2F40BF3040B02F40B03040AF2F3FAF2F3F9F2F3F90303F\
802F305F2F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F3030303030
3F3F3F3030403F3F4F303F503F405F303F5F3F40603040603F4F6F30406F3F4F70304F703F507F\
3F4F7F3F50803F507F3F50803F50803F5F8F3F508F3F5F90305F8F3F5F90305F8F
3F5F9030508F3F5F9030508F3F5F903050903F5F9030508F3F5F9030508F3F5F8F30508F3F5F8F\
30508F3F5F8F30508F3F5F8F30508F3F508F304F803F5080304F803F4F80304F7F
3F4F80304F7F3F4F80304F7F3F407F30407F3F407F3040703F40703040703F4070303F703F3F70\
303F6F303F6F303F6F3F3F6F303F60303F6030305030304F30303F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000402F40502F406F2F50702F50\
7F2F60802F608F2F6F902F6F9F2F709F2F70A02F7FA02F80AF308FA02F80AF2F80
AF2F80AF2F80A02F7FAF2F7FA02F7FA02F709F2F6F9F2F70902F6F902F708F2F6F8F2F6F7F2F60\
7F2F606F2F5F5F2F504F2F4F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F3F3F2F40402F4F502F4F502F4F602F50602F50702F50702F4F802F50802F4F\
902F50902F509F2F509F2F50A02F5FA02F50AF2F5FAF2F5FB0305FB02F5FBF305F
BF2F50C0305FC02F50CF3050C02F40C0304FBF2F40BF2F40B02F40B02F40AF2F3FA02F3F9F2F3F\
902F3F802F2F6F2F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F3030303030303040
3030403F3F50303F50303F5F303F5F3F4060304060304F6F3040703F4F7F304F703F507F304F7F\
3F50803F50803F50803F50803F5F8F30508F3F5F90305F8F3F5F90305F8F3F5F90
305F903F5F903050903F5F903050903F5F90305090305F9030508F3F5F9030508F305F8F30508F\
3F5F9030508F30508F30508F3F508F304F80304F80304F803F4F80304F7F304F80
304F7F304F7F30407F304F7F303F7030407F303F70304070303F7030407F303F70303F70303F70\
303F7030306F303F6F30306F303F6F30306030305F30304F303040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F2F3F4F2F406F2F5F702F5F7F2F60\
802F608F2F6F902F6F9F2F709F2F70A03080A02F80AF308FAF2F80AF3080AF2F80
AF2F80AF2F7FAF2F80A02F7FA02F7F9F2F7F9F2F70902F6F902F6F802F60802F6F7F2F60702F60\
5F2F50502F50000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F3F402F4F402F4F502F4F5F2F4F6F3050702F4F7F2F50\
802F4F8F3050902F509F2F509F2F50A0305FA02F5FAF305FAF2F5FB03060B02F5F
BF305FBF2F50C0305FC02F50CF304FC02F4FC0304FC02F40BF3040B02F40BF3040AF2F3FAF2F3F\
A02F3F9F303F802F30702F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003030303F303030303F3F3F4F303F4F
3F3F5F303F5F3F40603040603F406F30406F3F4F70304F703F507F304F7F3F50803F50803F508F\
3F50803F5F8F3F508F3F5F903F5F903F5F90305F903F5F90305F903F609F305F90
3F609F305F903F5F9F3050903F5F9F3050903F5F9F3050903F5F903050903F5F903050903F5F90\
30508F3F5F8F30508F3F5F8F304F803F508F304F803F4F80304F803F4F80304F7F
3F4F7F30407F3F4F7F30407F30407F303F703F407F30407F30407F303F703F3F7F303F70303F7F\
303F70303F70303F70303F7030306F30306F30305F303050303040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3040203F5F2F506F2F5F7F2F607F2060\
8F2F6F902F6F9F2F709F2F70A02F7FA02F80AF308FA02F80AF2F80AF2F80AF2F80
A02F7FAF2F7FA02F7FA02F7F9F2F709F2F7F902F708F2F6F802F607F2F60702F5F602F504F204F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F30402F40502F40602F4F6F2F4F\
702F4F7F2F4F802F508F2F4F902F50902F509F2F5F9F2F50A02F5FAF2F5FB02F5F
AF2F5FB02F5FB02F50BF3050BF2F4FC0304FC02F40C0304FBF2F40BF2F40B02F40B02F40AF2F3F\
AF2F3FA02F3F9F2F3F802F307F2F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F3030303F30303F303F4F303F503F405F
303F5F3040603040603F4F70304070304F7F304F7F3F5080304F7F3F50803F50803F508F3F508F\
3F508F3F508F3F5F903050903F5F90305F903F5F9F305F90305F9F305F903F609F
305F9F30609F30509F3F5F9F305090305F903050903F5F90305090305F903050903F5F9030508F\
30508F30508F30508F305080304F8F304F80304F8030407F304F7F30407F304F7F
30407030407F303F7030407F303F70303F70303F70304070303F70303F70303070303F70303070\
303F70303070303F70302F6F30306F30306030305F2F2F50303040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F2F3F5F2F4F6F2F5F7F2F607F2F608F2F70\
902F6F9F2F709F2F7FA03080A02F80AF308FAF2F80AF308FAF2F80AF2F80AF2F7F
AF2F80A02F7FA02F7F9F2F7F9F2F7F902F708F2F70802F6F7F2F60602F505F2F50000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F2F404F2F405F304F\
602F4F702F50702F4F8030508F2F50902F50902F509F305F9F2F50A0305FA02F5F
B03060B02F5FB03050BF2F50C03050C0304FCF304FCF2F4FCF304FC02F40C0304FBF2F40BF3040\
B02F40AF3040A02F3FA0303F8F2F307F2F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F303F3030303F3F403030403F3F50303F5F3F4060304060
3F4F6F3040703F4F7F304F7F3F5080304F803F508F3F50803F508F3F508F3F5F903F5090405F90\
3F5F903F5F9F3F5F903F609F3F5F9F3F609F305F9F3F60A030609F3F60A030609F
3F60A030609F3F5F9F305F9F3F5F9F305F9F3F5F9F3050903F5F9F305F903F5F9030508F3F5F90\
30508F3F5F8F30508F3F5F8F304F803F4F80304F7F3F4F7F304F7F3F4F7F30407F
3F4F7F303F703F4070303F703F4070303F7030407F3030703F3F7F303F70303F7F3030703F3F7F\
303F70303F7F30307030307F30306F30306F30305F30305F302F40000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003020304F2F405F2F50702F607F2F60802F6F8F2F70\
9F2F7F9F2F70A02F7FA02F80AF308FA02F80AF2F80AF2F80AF2F80AF2F7FAF2F7F
A02F7FA02F7F9F2F709F2F7F902F708F2F707F2F607F2F605F2F50000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F2F40\
4F2F4F5F2F406F2F4F702F4F7F2F507F2F4F8F2F508F2F509F2F509F2F50A02F5F
A02F50AF305FAF2F50B03050B02F50BF3050BF2F4FCF304FCF2F4FCF304FC02F40C02F40BF2F40\
BF2F40AF2F3FAF2F3FA02F3FA02F3F8F2F30802F30000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003030303030303F3F4030304F303F50303F5F3F4060304060304F6F
304F703F507F304F7F304F80304F803F508F30508F3F508F3F508F3F5F903F50903F5F9F3F5F9F\
3F5F9F3F5F9F3F5F9F305F9F3F609F305F9F3F60A030609F3F60A0305F9F3F60A0
30609F3F60A030509F305F9F30509F3F5F9F305090305F9F3050903F5F9030508F30508F30508F\
3F508F30508F305080304F803F4F8030407F304F7F30407F304070303F6F304070
303F6F303F6F30306030306F30306F303F7030306F303F70303070303F7F30307F303F7F30307F\
303F7F30307030307F303070303070302F6030306030305030303F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F4F2F405F2F4F6F2F5F702F60802F6F8F2F70902F7F\
9F2F7FA0308FA02F80AF308FAF2F80AF308FAF2F80AF2F80AF2F7FAF2F80A02F7F
A02F7F9F2F70A02F7F902F708F2F70802F6F7F2F60602F50000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F3F40304F502F40602F4F6F2F4F7F30507F2F4F802F508F2F50903050902F50
9F2F5FA02F4FAF3050AF2F50B03050B02F50C03050C0304FCF304FCF304FCF304FC02F40C0304F\
BF2F40BF3040B02F40B02F40A02F3FA0303F902F30802F30502F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030303F303F30303F3F3F4F303F503F405F3040603F406F30406F3F4F70304F7F
3F50803050803F5F8F3F508F3F50903F508F3F5F903F5090405F9F3F5F9F405F9F3F5F9F3F60A0\
3F5F9F3F60A03F609F3F60A030609F3F60A03060A03F60A03060A03F60A030609F
3F60A030609F3F60A0305F9F3F5F9F305F9F3F5F9F3050903F5F9030508F3F5F9030508F3F508F\
3050803F5080304F803F4F8030407F3F4F7F30406F3F406F000000000000000000
30304F302F5F30306F30306F303070303070303F7F30307F303F7F30307F303F8030307F303080\
30307F30307F30307030307F30307030306F30306030305F302F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F2F3F4F2F40602F5F6F205F7F2F6F802F6F902F70902F7F\
9F2F80A02F80AF3090A02F80AF2F80AF2F80AF2F80AF2F7FAF2F7FA02F7FA02F7F
9F2F709F2F7F902F708F2F6F802060702F605F2F50000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000302F40402F40502F4F602F4F6F2F4F702F4F7F2F50802F4F902F50
902F4F9F2F4F9F2F4FAF2F50AF2F50B02F50B02F50BF3050BF2F4FC0304FCF2F4FCF304FC02F40\
C02F40BF2F40BF2F40B02F3FB02F3FA02F3FA02F3F902F30802F305F2F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000302F30\
30303F3030403F3F50303F5030405F303F603F406F30406F304F70304F7F3F5080
30508030508F30508F3F5F903F50903F50903F5090405F9F3F509F3F5FA03F5F9F3F60A0305F9F\
3F60A0305FA03F60A0305FA03F60A03060A03F60A03060A03060A030609F3F60A0
305F9F3060A030609F3F5FA030509F305F9F3050903F5F9030508F30509030508F30508F304F80\
305080304F7F304F7F30406F30406F000000000000000000000000000000000000
2F2F4F302F60302F603030702F2F70303070302F7030307F302F7F30307F30307F303080302F7F\
30308030307F30307F2F2F7030307030306F303060302F50000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F3F4F2F40602F506F2F5F7F2F6F802F6F902F70902F7F9F3080\
A02F80AF3090AF2F8FAF3090AF2F8FAF2F80AF2F7FAF2F80A02F7FAF2F7FA02F7F
A02F7F902F70902F70802F607F2F60602F50000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F2F404F2F4F5F2F4F6F3050702F4F7F2F507F2F4F
8F304F902F4F9F30509F2F4FAF3050AF2F50B0305FB02F50C03050C03050CF3050CF304FCF304F\
C0304FC0304FBF2F40BF3040B02F40B03040AF2F3FAF303F9F2F308F30306F2F2F
0000000000000000000000000000000000000000000000000000000000003F30303030303F3F40\
303F4F3F3F50303F5F3F406F30406F3F4F70304F703F507F304F803F508F30508F
3F5F903F5F903F5F9F3F5F9040609F3F5F9F405FA03F5FA04060AF3F5FA03F60A03F60A03F60AF\
3F60A03F60AF3060A03F6FAF3060AF3F60AF3060A03F60AF3060A03F60A03060A0
3F60A03060A03F60A0305F9F3F5F9F3050903F5F903050903F5F9030508F3F508F304F803F5080\
304F7F3F406F000000000000000000000000000000000000000000000000000000
303050302F60303070302F7030307F30307F30308030307F303080303080303080303080303080\
30308030308030307F30307F303070303070302F5F3F3050000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F2F3F502F4F602F5F702F607F2F608F2F708F2F709F2F7F9F2F80\
A0308FA02F8FAF2F8FAF2F8FAF2F8FAF2F80AF2F7FA02F7FA02F7F9F2F70A02F7F
902F70902F6F802F60702F605F2F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F303F2F404F2F405F2F4F602F4F702F4F
7F2F408F2F4F902F4F9F2F4F9F2F4FAF2F50AF2F50B02F50B02F50BF3050BF2F4FC0304FCF304F\
CF304FC02F40C0304FBF2F40BF2F40B02F3FB02F40AF2F3FAF2F3F902F308F2F30
702F2F0000000000000000000000000000000000000000000000003F3030302F3030304030304F\
3F3F5F303F5F30406F30406F3F4F70304F70304F80304F803F508F30508F3F5F90
305F903F5F9F3F5F9F3F609F3F609F4060A03F5FA04060AF3F5FA03F60AF3F5FA03F60AF3060A0\
3F60AF3060A03F60AF3060AF3F60AF3060A03F60A03060A03F60A03060A03060A0
3060A03F60A0305F9F305F9F3050903F5F9030508F30508F3050803F5080304F7F304F7F304060\
000000000000000000000000000000000000000000000000000000000000302F3F
2F2F5030306F2F2F6F30307F302F7F303080302F7F303080302F8030308F302F80303080303080\
303080302F80303080302F7F30307F302F6F30305F302F4F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000302F3F4F2F4F602F50702F607F2F608F2F708F2F709F307F9F2F80A0308F\
A02F8FAF3090AF2F8FAF2F8FAF2F80AF2F8FAF2F80AF2F7FA02F70A02F7F902F70
902F6F802F607F2F605F2F4F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F3F3F2F40503040602F40
702F4F7F2F408F304F902F4F9F2F4F9F2F4FAF3050AF2F50B03050B02F50C03050C03050CF3050\
CF304FCF304FC0304FC0304FBF2F40C0304FBF2F40B02F40AF2F3FAF303F9F2F3F
8F2F306F2F2F0000000000000000000000000000003030303F303F30303F3F3F4F303F503F4060\
303F603F40703040703F4F7F304F7F3F508030508F3F5F903050903F5F9F3F5F9F
3F609F3F5F9F4060A03F60A04060AF3F60AF4060AF3F60AF3F60AF3F60AF3F60AF3F60AF3F6FAF\
3060AF3F6FB03060AF3F60AF3060A03F60AF3060A03F60AF3060A03F60AF3060A0
3F60A0305F9F3F609F3050903F5F9030508F3F508F3050803F5080304F7F3F406F000000000000\
000000000000000000000000000000000000000000000000000000000000302F4F
30305F302F6F30307F302F7F303080302F8030308F302F8030308F30308F30308F30308F30308F\
302F803F308F302F8030307F302F703F306F302F5F303040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F30402F4050204F6F2F5F702F60802F6F8F206F902F7F902F7FA0308FA02F8F\
AF2F90AF2F8FAF2F8FAF2F80AF2F8FAF2F80AF2F80A02F7FA02F7F902F6F902F6F
802060702F5F5F2F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F3F4F2F40
5F2F40702F407F2F40802F4F8F2F409F2F4F9F2F4FA02F50A02F4FB02F50B02F50BF3050BF2F4F\
C03050CF304FCF304FC02F40C0304FBF2F40C02F40B02F3FB02F40AF2F3FAF2F3F
9F2F308F2F305F2F2F000000000000000000302F3030303F30303F303F4F303F5F3F4060303F60\
3040703040703F4F7F304F7F30508F30508F3F5F903050903F5F9F305F9F3F60A0
3F5F9F3F60A03F60A04060AF3F60AF3F6FAF3F6FAF3F60AF3F60AF3F60AF3060AF3F6FB03060AF\
3F60B03060AF3F60AF3060A03060AF3060A03F60AF3060A03060A030609F3F609F
305F9F30609F305F903F509030508F305080304F7F304F7F304060000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000303050
2F2F5F302F70302F703030802F2F7F303080302F8030308F302F8F30308F302F8F30308F302F80\
30308F302F80303080302F70303070302F6030305F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000402F40502F4F6F2F5F702F607F2F6F802F6F902F7F902F7FA03080A02F8FAF3090\
AF2F8FAF308FAF2F8FB0308FAF2F80AF2F80A02F80A02F809F2F7F902F70802F6F
7F2F605F2F4F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F3F
4F30405F2F40702F4F7F2F408F304F8F2F4F9F304F9F2F4FA03050AF2F50B03050B02F50BF3050\
C03050CF3050CF304FD03050CF304FCF304FC02F40C0304FBF2F40B0303FAF2F3F
AF303F9F2F308F3030502F2F3F303F3030303F303F3030403F3F5F303F603F406F30406F3F4F7F\
304F7F3F4F80304F803F50903050903F5F9F305F9F3F5FA03F5F9F3F60A03F60A0
4060AF3F60AF406FB03F60AF406FB03F6FB03F6FB03F60B03F6FB03F60B03F6FB03060B03F6FB0\
3060AF3F6FB03060AF3F6FAF3060AF3F60AF3060AF3F60AF30609F3F60A0305F9F
3F609F305F903F5F903050803F5080304F7F3F4060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000303030302F5F
30306F302F7030307F302F8030308F302F8F303090302F8F303090302F8F303090302F8F3F308F\
302F8F30308F302F7F3F307F302F7030306F302F50000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000302F3F4020405F2F506F2F5F7F2F6F802F6F8F2F70902F7F9F2F809F2F80A03090A02F8F\
AF308FAF2F8FAF2F8FAF2F80AF2F80A02F7FA02F809F2F7F902F708F2F6F7F2F60
602F50000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F3F4F2F405F2F406F2F40702F40802F4F8F2F40902F4F9F2F4FA02F50A02F4FB02F50B02F50\
BF3050BF2F4FC03050CF304FCF304FC02F4FC0304FC02F40C03040B02F3FB02F3F
AF2F30AF2F3F902F303030303030303F304030304F303F5F303F603F406F304070304F7F304F7F\
3F4F80304F803050903050903F5F9F30509F3F5FA0305FA03F60AF3F5FA03F60AF
3F60AF406FB03F60AF3F6FB03F6FB03F6FB03F6FB03F70B03060B03F6FBF3060B03F6FB03060AF\
3F6FB03060AF3F60AF3060AF3F60AF3060A03060A0305F9F3F609F305F9030609F
305F903F5F8F304F80304F7F30405F000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F4F302F60
302F6F30307F302F7F30308F302F8F30308F302F8F303090302F8F303090302F90303090302F8F\
30308F302F80303080302F70303070302F6030304F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F2F405F2F506F2F5F702F607F2F6F8F2F70902F7F9F2F809F2F80A03090AF2F90AF3090\
AF2F8FB0308FAF2F80AF2F8FA02F80AF2F80A02F7F9F2F7F8F2F6F8F2F6F6F2F5F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F30403040502F406F2F4F702F4080304F8F2F4F902F4F9F2F4FA03050A02F50B03050\
B02F50BF3050BF3050C03050CF304FD03F50CF304FCF3040C02F3FC03040BF2F3F
B02F3FAF2F3F3F303F30303F3F3F4F303F503F406030406F3F40703040703F4F80304F803F508F\
30508F3F5F9F30509F3F5F9F305F9F3F60A03F5FA03F60AF3F60AF406FB03F60B0
406FB03F6FB0406FB03F6FB03F70B03F6FB03F70BF3F70B03F6FBF3060B03F6FBF3060B03F6FB0\
3060AF3F6FB03060AF3F60AF3060A03F60A030609F3F609F305F9F3F609F305F8F
3F5F8F30507F3F4060000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030305F302F6F
30307F302F7F30308F302F8F303090302F8F303090302F90303090302F9030309F302F90303090\
302F8F30308F302F8030307F302F703F3060000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3020304F2F4F5F2F50702F60702060802F708F2F70902F7F9F2F80A02F8FA02F8FAF3090AF2F8F\
AF2F8FAF2F80AF2F8FA02F80A02F80A02F7F9F2F7F90206F8F2F6F702F5F602F4F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000402F40502F3F602F40702F40802F4F802F40902F4F902F4FA02F50A02F4F\
AF2F50B02F50BF3050BF2F4FC03050C0304FCF3040C02F40C03040C02F3FC02F40
302F3030303F30303F303F4F303F50303F60303F6F3F407030407F304F80304F803F508F304F8F\
30509F30509F3F5FA0305F9F3F5FA0305FA03F60AF3F60AF3F60B03F60B03F6FB0
3F6FB03F6FB03F6FB03F70BF306FB03F70BF3070BF3F70BF306FB03F6FB03060B03F6FB03060AF\
3060B03060AF3F60AF3060A03060A0305F9F3F609F305F90305F903050803F5080
304060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F402F2F5F303070
2F2F70302F80302F8030308F302F8F302F90302F9030309F302F90302F90302F90303090302F8F\
302F8F302F80303080302F703F2F6F302F4F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F2F40502F506F2F60702F60802F708F2F70902F7F9F2F80A0308FA0308FAF3090AF2F8FB0308F\
AF2F8FB02F8FAF2F80AF2F80A02F7FA02F80902F70902F70802F606F2F50000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F3040502F4060304F6F2F4080304F8F2F4F90304F902F4FA03050\
A02F50AF3050B02F50BF3050BF3050C0304FCF3040D03F4FCF3040CF3040303030
3F303F3030403F3F50303F5F3F406F3040703F4F7F304F7F3F508F304F8F3F50903050903F5FA0\
305F9F3F5FA0305FA03F60AF3F60AF3F60B03F60B0406FBF3F6FB0406FBF3F6FB0
4070BF3F6FBF3F70BF3F70BF3F70BF3F70BF3F70BF3070B03F70BF306FB03F6FB03060B03F6FB0\
3060AF3F60AF3060A03F60A030609F3F609F305F903F5F8F3050803F507F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F5030306F302F70
303080302F80303090302F903F309F302F9F3F309F302F9F30309F302F9F30309F302F903F3090\
302F8F3F308F302F7F3F307F302F60000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F2F
402F405F2F5F6F2F607F2F70802F70902F7F902F7F9F2F80A02F8FAF3090A02F8FAF2F8FAF2F8F\
B02F8FAF2F80AF2F80A02F7FA02F809F2F70902F70802F6F702F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F2F3F4F2F3F602F406F2F407F2F4F802F40902F4F902F4F\
9F2F50A02F4FAF2F50B02F50BF304FBF2F40C0304FCF3040CF3040302F3030303F
3030403F3F50303F5F30406F3040703F4F7F304F80304F8F304F8F3F509030509F3F5FA0305FA0\
3F5FAF305FA03F60AF3060AF3F60B03F60B03F6FBF3F6FB03F6FBF3F6FB03F70BF
3F6FBF3F70BF3070BF3F70BF3070BF3F70BF306FB03F70BF306FB03F70BF3060B03060B03060AF\
3F60AF305FA03060A0305F9F3F609F305F8F305F8030507F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3F30305F2F2F6F302F7F
302F8030308F302F8F303090302F903F309F302F9F30309F302F9F3F309F302F903F3090302F90\
3F3090302F803F307F302F6F3F3050000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F3F
4F2F50602F5F702F6F7F2F708F2F7F902F809F308F9F2F8FAF3090AF2F8FAF3090AF2F8FB02F8F\
AF2F80AF2F80AF2F80AF2F809F2F7F9F2F708F2F6F7F2F60602F4F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003030404F2F40602F406F2F407F304F802F4F902F4F\
902F4FA03050A02F50AF304FB02F4FBF304FBF304FCF304F30303F3F304030304F
3F3F5F303F603F407030407F3F4F80304F803F50903050903F509F30509F3F5FA0305FA03F60AF\
305FAF3F60B03F60B03F6FBF3F60BF406FBF3F6FBF3F6FBF3F6FBF3F70C03F70BF
3F70C03070BF3F7FC03070BF3F70BF3070BF3F70BF3070B03F70BF306FB03F6FB03060AF3F60AF\
3060A03F60A0305F9F3F609F305F8F3F5F8030507F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000303050302F603F307F302F7F
30308F302F8F3F309F302F903F309F302F9F3F30A0302F9F3F309F302F9F3F309F302F903F3090\
302F8F3F3080302F703F306F302F4F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F2F40
4F20506F2F60702F6F802F708F207F902F809F2F8FA03090A02F90AF2F8FAF2F8FAF2F8FAF2F80\
AF2F80A02F7FAF2F80A0207F9F2F708F2F6F8F2F6F6F2050000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3F402F3F5F2F40602F40702F4F7F2F40\
8F2F4F902F4F9F2F4FA02F40AF2F4FB02F4F303030302F3030304030304F303F5F
303F6030407030407F304F80304F803050903050903F5F9F30509F305FA0305FA03F60AF305FAF\
3F60B03060B03F6FBF3F60BF3F6FBF3F6FBF3F6FBF3F6FBF3F70BF3F70BF3F70C0
3070BF3F70C03070BF3F70BF306FB03F70BF306FB03F70BF306FB0306FB03060AF3F60AF305FA0\
3060A0305F9F3F5F9030508030507F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F402F2F5F302F6F302F7F303080
302F8F302F90302F9030309F302F9F302FA0302F9F3F30A0302F9F3F2F9F302F903F3090302F8F\
3F2F8F302F7F3F3070302F50000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F30402F4F
5F2F5F6F2F607F2F70802F70902F80902F8FA03090A02F90AF309FAF2F90B03090AF2F80B0308F\
AF2F80AF2F80A02F7FA02F7F902F708F2F6F702F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000303030402F3F5F3040602F4070304F\
7F2F4F903040902F40A0304FA02F4F3F303F30303F3F3F4F303F503F3F6030406F
3F4F7F304F7F3F4F80304F8F3F5F90305F9F3F5FA0305FA03F60AF3F5FAF3F60B03F60B03F60BF\
3F60BF3F6FC03F6FBF406FC03F6FBF3F70C03F6FBF3F70C03F70C03F7FC03070C0
3F7FC03070BF3F70C03070BF3F70BF3070BF3F70BF306FB03F70B0306FAF3F60AF3060A03F60A0\
305F9F3F5F8F30507F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F503F306F302F7F3F3080302F8F
3F3090302F9F3F309F302F9F3F30A0302FA03F30A0302F9F3F30A03F2F9F3F309F302F903F3090\
302F803F307F302F6F3F3050000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030203F4F2F50
5F2F5F702F6F7F2F708F2F7F902F809F308F9F2F8FAF309FAF2F90AF3090AF2F8FB02F90AF2F8F\
AF2F80A02F7FA02F7F902F70902F6F7F2F606F2F4F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F2F3F502F40602F40\
702F407F2F3F8F2F40902F4030303F302F3F30304F3030503F3F60303F6F30407F
30407F3F4F8F304F8F30509030509F3F5FA0305FA03F60AF3060AF3F6FB0305FB03F60BF3F60BF\
3F6FC03F60BF3F6FC03F6FC03F70C0306FBF3F70C03070C03F7FC03070C03F70C0
3070BF3F70BF306FBF3F70BF306FBF3F70BF306FB0306FB0306FAF3F6FAF30609F30609F305F8F\
3F5F8030507F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F40302F60302F6F3F3080302F803F2F90
302F903F309F302F9F3F2FA0302FA03F30A0302F9F3F30A0302F9F3F309F302F903F3090302F8F\
3F3080302F70302F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F2F4F502F50
6F2F6F702F6F802F7F8F2F7F902F8F9F2F8FA0309FAF309FAF309FAF2F90B03090AF2F8FAF2F8F\
A02F80AF2F8F9F2F7F902F708F2F60702F5F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F2F3F502F3F\
602F3F703040802F403F303F30303F3F3F50303F5F3F406F3040703F4F7F304F80
3F508F30508F3F5F9F305F9F3F5FA0305FAF3F60B03060B03F6FBF306FBF3F6FC03F60C03F6FC0\
3F6FC03F70C03F6FC03F70C03F70C03F70CF3F70C03F7FCF3070C03F7FC03070C0
3F70C0306FBF3F70BF3070BF3F70BF306FB03F70B0306FAF3F6FAF3060A03F6F9F305F8F3F5080\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030305F302F6F3F307F302F803F30903F2F90
3F309F3F2F9F3F30A0302FA03F30AF302FA03F30A0302FA03F30A0302F9F3F309F302F903F308F\
302F7F3F3070302F50000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F303F2F4F5F2F5F
6F20607F2F70802F70902F80902F809F2F90A02F90AF309FAF2F90AF2F90AF2F8FAF2F8FA02F80\
AF2F80A02F809F2F7F8F206F802F60602F4F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F2F30\
502F3F302F3030303F303040303F50303F5F30406F30407030407F304F8030508F
304F9030509F30509F3F5FAF305FAF3060B03060B03F6FBF3060BF3F6FC03F6FC03F6FC0306FC0\
3F6FC0306FC03F70CF3070C03F70CF3070C03F7FCF3070C03F70C03070BF3F70C0
306FBF3070BF306FBF3F70BF306FAF306FAF306FA03F6FA030609F306090305080000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000302F40302F5F302F70302F7F3F2F8F302F903F309F
302F9F3F2FA03F2FA03F30AF302FA03F2FA0302FA03F30A0302F9F3F2F9F302F903F3090302F80\
302F7F302F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F3F4F2F50602F60
702F6F7F2F708F2F7F902F809F308FA02F90AF309FAF2F90B0309FAF2F90B02F90AF2F8FAF2F8F\
A02F80A02F80902F70902F707F2F5F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030303F30403030403F3F5F303F603F407030407F3F4F80304F8F3F5090305090
3F5FA0305FA03F60AF305FAF3F60B03060B03F6FBF3F6FBF3F70C03F70C03F70CF3F70C03F70CF\
3F70CF3F70CF3F70CF3F7FCF3F70CF3F7FCF3070C03F7FCF3070C03F70C03070BF
3F70C03070BF3F70BF306FB03F6FB0306FA03F6FA030609F3F6090305F7F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000030303F302F5F30306F302F7F3F308F3F2F903F309F3F2F9F
3F30A03F2FA03F30AF3F2FAF4030AF3F2FA04030AF302FA03F30A03F2F9F3F309F302F8F3F3080\
3F2F703F304F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F2F4F4F20506F2F60
702F6F802F7F8F2F7F902F809F2F8FA03090A02F90AF309FAF2F90B02F90AF2F8FAF2F8FA02F80\
AF2F809F2F7F902F708F2F6F702F50000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303030\
30304030304F303F5F303F603F407030407F304F80304F8F3F5090305090305FA0
305FA03F60AF305FAF3060B03060B03F6FBF306FBF3F6FC03F6FC03F70CF3070C03F7FCF307FCF\
3F70CF3070CF3F7FCF3070CF3F7FCF3070C03F70C03070C03F70C0306FBF3070BF
306FBF3F70BF306FAF306FAF3060A03F6FA0305F90305F80000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000302F4F303060302F703F2F80302F8F3F3090302F9F3F2FA0
3F2FA03F30AF3F2FA03F2FAF3F2FA03F30AF3F2FA03F2FA03F2F9F3F309F3F2F8F3F2F803F2F70\
3F305F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F30402F4F5F2F5F6F2F6F
7F2F70802F7F902F80902F8FA03090A02F90AF309FAF2F90B03090AF2F8FB02F90AF2F8FAF2F8F\
A02F809F2F7F902F70802F60602F4F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000030303F3F3040\
30304F3F3F60303F6F3F407F30407F3F4F8F304F8F3F509030509F3F5FA0305FAF
3F60AF305FB03F60BF3060BF3F6FC0306FC03F70CF3F70CF3F70CF3F70CF3F7FCF3F7FCF3F7FCF\
3070CF3F7FD0307FCF3F7FCF3070CF3F7FCF3070C03F70C0306FBF3F70C0306FBF
3F6FB0306FAF3F6FAF306FA03F6FA0305F8F3F5F80000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000302F40303060302F6F3F30803F2F8F3F30903F2F9F3F30A03F2FA0
3F30AF3F2FAF4030AF3F2FAF4030AF3F2FA04030AF3F2FA03F309F3F2F903F308F3F2F7F3F3070\
302F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F203F4F2F505F2F60702F6F
7F20708F2F7F8F2F809F2F8F9F2F8FA0309FAF2F90AF3090AF2F8FB02F90AF2F8FAF2F8FA02080\
A02F80902F708F2F6F702050000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000302F3F30304F303050\
303F60303F6F30407F30407F304F8F304F8F30509F30509F305FA0305FAF3060B0
305FB03060BF3060BF3F6FC0306FC03F6FCF306FC03F70CF3070C03F7FCF3070CF3F7FD0307FCF\
3080D03070CF3F7FCF3070C03070CF3070C03F70C0306FBF3070C0306FB03F6FB0
3060AF306FAF3060A03F609F305F80305F7F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F30302F5F302F603F2F7F302F803F2F90302F903F2FA03F2FA03F2FAF
3F2FA03F2FAF3F2FAF3F30AF3F2FA03F2FAF3F2FA03F309F3F2F903F2F903F2F803F3070302F50\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F2F4F502F5F6F2F6F702F6F
802F7F8F2F7F902F8F9F2F8FA0309FAF309FAF309FAF2F90B03090B02F8FB02F8FAF2F80AF2F80\
9F2F7F9F2F708F2F6F702F50000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030303F3F304F3030503F406F\
303F703F40803040803F4F8F304F903F509F30509F3F5FAF305FAF3F60B03060B0
3F60BF3060BF3F6FC0306FC03F70CF3F70CF3F70CF3F70CF3F7FD03F7FCF3F7FD0307FCF3F80D0\
3080D03F80D03070CF3F7FCF3070C03F70C0306FC03F70C0306FBF3F6FB0306FAF
3F6FAF3060A03F6090305F80000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000303050302F603F307F3F2F803F308F3F2F903F30A03F2FA03F30AF3F2FAF
4030B03F2FAF4030AF3F2FAF4030AF3F2FA04030A03F2F9F40309F3F2F8F3F307F3F2F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F303F2F4F5F2F5F602F607F2F70
802F7F902F80902F809F3090A02F90AF309FAF2F90AF2F90AF2F8FB02F8FAF2F80AF2F80A02F80\
9F2F7F8F2F6F802F60602F40000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303030303F30304F303050303F6F303F70\
304080304080304F8F304F903F509F3050A0305FAF305FAF3F60B0305FB03060BF
3060BF3F6FC0306FC03F70CF306FCF3F70CF3070CF3F7FCF3070CF3F7FD0307FCF3F80D0307FCF\
3F80D0307FCF3070CF3070C03F70C0306FC03070BF306FB03F6FB03060A0306FA0
30609F3F608F30507F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F40302F5F3F2F703F2F7F3F308F3F2F90402F9F3F2FA03F30AF3F2FA03F2FAF
3F2FAF4030AF3F2FAF3F2FAF3F2FAF4030A03F2F9F3F2F9F3F2F8F3F3080302F70302F50000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F3F4F2F505F2F60702F707F2F70
8F2F7F8F2F809F308F9F2F90AF309FAF2F90B0309FAF2F90B02F90AF2F8FAF2F8FAF2F80A02F80\
902F70902F6F702F50000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F303F30303F3F304F30305F3F406F30407F3F4080\
30408F3F4F90304F903F50A03050A03F5FAF305FB03F60B03060BF3F6FC03060C0
3F6FCF306FCF3F70D03F70CF3F70D03070CF3F7FD0307FD03F7FD0307FD03F80D03080CF3F80D0\
307FCF3F80CF307FC03F70CF306FC03F70C0306FB03F6FB03060A03F6FAF306090
3F5F80000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303030\
302F503F306F3F2F7F3F308F3F2F8F40309F402FA03F30AF3F2FA04030AF3F2FAF
4030B03F2FAF4030B03F2FAF4030AF3F2FA040309F3F2F903F308F3F2F7F3F3060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F404F2050602F6F702F6F802F7F
8F207F902F809F2F8FA03090A02F90AF309FAF2F90B02F90AF2F8FAF2F8FAF2F80AF2F809F207F\
902F708F2F606F2F50000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000303030302F3F30304F30305F303F6F304070304F80304F8F\
304F90304F903050A03050A0305FAF305FB03060BF305FBF3060C03060C03F6FCF
306FCF3F6FCF306FCF3F70D03070CF3F70D03070CF3F7FD0307FD03080D0307FCF3F80CF307FCF\
307FCF307FC03F7FC03070BF306FBF306FB03F6FB03060A03060A0305F8F305F7F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F4F\
302F60302F703F2F803F2F8F3F2F903F2F9F402FA03F2FA0402FAF3F2FAF402FB0
3F2FAF3F2FAF3F2FAF4030AF3F2FA03F2F9F3F2F903F2F90302F7F3F2F70302F4F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F4F502F60602F6F702F7F802F7F
902F8F902F8FA03090A03090AF309FAF3090B03090AF2F8FB02F90AF2F8FAF2F8FA02F80A02F7F\
902F6F802F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F303F3030403F3050303F5F3F407030407F3F4F80304F8F3F509F\
30509F3F5FA03050AF3F5FB0305FB03F60BF3060BF3F6FC03060C03F6FCF306FCF
3F70D03F70D03F70D03070D03F7FD0307FD03F7FD0307FD03F80D0307FD03F80D0307FCF3F80CF\
307FCF3F7FCF3070BF3F7FBF306FB03F6FB03060A03F609F305F80000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F3F3F3060\
3F2F703F30803F2F8F40309F402F9F4030A0402FA04030AF402FAF4030B03F2FB0
4030B03F2FAF4030B03F2FA04030A03F2F9F40309F3F2F803F307F302F5F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F2F4F4F2F5F6F2F70702F70802F80
8F2F8F9F2F909F2F90A030A0AF2F9FAF309FAF2F8FB02F90AF2F8FAF2F8FA02F80A02F7F9F2F70\
902F6F702F50000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000030303F30303F30305030305F303F6F30407F304F80304F8F30509030509F\
3F5FAF3050AF305FB0305FB03F60BF3060BF3060C03060C03F6FCF306FCF3F70D0
306FCF3F70D03070CF3F70D03070D03F7FD0307FD0307FD0307FCF3F80D0307FCF307FCF307FC0\
3F7FCF3070BF3070BF3070B03F70AF3060A0306090305F7F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F2F302F5F302F6F\
3F2F7F3F2F803F2F903F2F9F4030A03F2FA0402FAF402FAF4030B0402FAF402FB0
402FAF4030B03F2FAF3F2FA03F2F9F3F309F3F2F8F3F2F80302F6F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F2F4F502F60602F6F7F2F7F802F80
902F8F902F90A0309FA030A0AF30A0AF2F9FB0309FB02F90B02F90AF2F8FAF2F80A02F7F9F2F70\
8F2F60702F50000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F303F30304030305030305F3F407030407F3F4F80304F8F3F509F30509F3F5FAF\
305FAF3F5FBF305FBF3F60BF3060C03F6FCF306FCF3F6FD0306FD03F70D03070D0
3F70D03070D03F7FD03070D03F7FDF307FD03F80D0307FD03F7FD0307FCF3F7FCF307FC03F7FC0\
3070BF3F70BF3070B03F70AF306F9F3F608F30507F602F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000030304F302F603F307F\
3F2F804030903F2F9F4030A0402FA04030AF402FAF4030B0402FB04F30B0402FB0
4030B0402FAF4F30AF402FA040309F402F9040308F3F2F703F304F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F2F5050205F6F2F70702F7F802F80
8F2F8F9F2F909F2F9FAF30A0A02F9FAF2F9FAF2F90B02F9FAF2F8FAF2F8FA02F80A02F80902070\
802F60602F4F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F30304030305030305F303F70303F7F304F80304F8F30509030509F305FAF305FAF\
3060B03060B03060C03060C03060CF3060CF3F6FD0306FCF306FD0306FD03F70D0
3070D03F70D03070D03F7FDF307FD0307FD0307FCF3F7FCF307FCF307FCF3070C03F7FC03070BF\
3070B0306FAF3070AF306090306080A02F309F2F30802F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F40302F5F3F2F703F2F80\
3F2F8F3F2F90402FA03F2FA0402FAF402FAF402FB0402FB0402FB0402FAF402FB0
402FAF402FAF402FA0402FA03F2F90402F8F3F2F7F3F2F5F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F303F2F505F2F6F6F2F707F2F7F802F80
9030909F2F90A0309FA030A0AF30A0AF2F9FB0309FAF2F90B02F90AF2F8FAF2F8F9F2F7F9F2F70\
802F60602F4F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F303F3030403F3F503030603F407030407F3F4F8F304F8F3F509F30509F3F5FAF305FAF3F60BF\
3060BF3F6FC0306FC03F6FCF306FCF3F6FD0306FD03F70DF3070D03F70DF3070D0
3F7FDF3070D03F7FDF307FD03F7FD0307FD03F7FD0307FCF3F7FCF307FCF3F7FC03070BF3F70BF\
3070AF3F70AF3060903F6080B02F30B0303FA02F309F3030702F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000303040302F5F3F306F3F2F7F40308F\
3F2F904030A0402FA04030AF402FAF4030B0402FB04F30B0402FB04F30B0402FAF
4F30B0402FA04030A0402F9F4030903F2F7F3F306F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2030402F505F2F606F2F70702F7F8F2F80
902F8F9F30909F2F9FAF30A0AF2F9FAF309FAF2F90B02F90AF2F8FAF2F8FA02F809F2F7F902F6F\
802F60000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000030303F
303040303050303060303F7030407F304F8F30408F30509F30509F305FAF305FAF3060B03060BF\
3F6FC03060C0306FCF306FCF3F6FD0306FD03070D0306FD03F70DF3070D03F70DF
3070D03F7FDF3070D0307FD0307FCF3F7FD03070CF307FCF3070C03F70C03070B03070B0306FAF\
3F6FA030608F305F7FBF2F30BF2F3FB02F30AF2F309F2F30902F305F2F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F30302F503F2F6F3F2F703F2F8F3F2F8F\
402F9F3F2FA0402FAF402FAF4030B0402FB0402FB0402FB04030B0402FAF402FB0
402FA04030A0402F9F402F903F2F803F3070000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F3F402F50602F6F6F2F707F2F808F2F80
902F909F2F90A030A0AF30A0AF30A0AF2F9FB0309FAF2F90B02F90AF2F8FAF2F8F9F2F7F902F70\
7F2F5F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F303F303040
303F5F3030603F407030407F3F4F8F304F8F3F509F3050A03F5FAF305FAF3F60BF3060BF3F6FC0\
306FC03F70CF3070D03F70DF306FD03F70DF3070D03F70DF3070D03F7FDF3070DF
3F7FDF307FD03F7FD0307FD03F7FD0307FCF3F7FCF3070C03F70C03070B03F70B0306FAF3F6FA0\
306080CF3F40C0303FC0303FBF2F30BF303FB02F30AF30309F2F30803030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F503F306F3F2F7040308F402F8F40309F\
402FA04030AF402FAF4030B0402FB04F30BF402FB04F30B0402FB04F30B0402FAF
4030A0402F9F40309F3F2F8040307F302F50000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F3F4F2F5F5F206F702F707F2F7F8F2F8F
902F8F9F2F90A02F9FAF30A0AF2F9FAF2F9FAF2F90B02F90AF2F8FAF2F8FA02F809F2F7F8F206F\
7F2F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000030303F30304030305F
303060303F70303F7F30408F30408F304F9F30509F305FAF305FAF3060B03060B03060C03060C0\
306FCF306FCF3070D03070D03070DF306FD03070DF3070D03070DF3070DF307FDF
3070D0307FD03070CF307FD03070CF307FCF3070C03070BF306FB03070B0306FA0306F9F305F7F\
C03F40C0303FCF303FC02F30C02F3FBF2F30BF2F3FAF2F30A02F30902F2F702F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F403F2F603F2F70402F80402F8F402F9F402F9F\
402FA0402FAF402FB0402FAF402FB0402FB0402FB0402FAF402FB0402FAF402FAF
402F9F402F9F3F2F8F402F7F302F5F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F4F4F2F5F602F70702F70802F808F2F8F
9F30909F2F90A030A0AF30A0B030A0AF2F9FB02F9FAF2F90B02F90AF2F8FAF2F809F2F70902F70\
702F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F303F30304F3F3F5F303060
3F407F3040803F4F8F304F903F509F3050A03F5FAF305FAF3F60BF3060BF3F6FC0306FC03F70CF\
306FD03F70D03070D03F7FDF307FDF3F7FDF3070DF3F7FDF3070DF3F7FDF3070D0
3F7FDF307FD03F7FD03070CF3F7FCF3070C03F70BF3070B03F70B0306FA03F6F90305F7FC03F40\
C03F40CF3F40C0303FCF3040C02F3FC0303FBF2F30BF303FAF2F30A030308F2F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F403F30603F2F70403080402F8F4F309F402F9F4F30AF\
4F2FAF4030B0402FB04F30BF402FB04F30B0402FB04F30B0402FAF4F30AF402FA0
4030A0402F904030803F2F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030204F4F2F5F602F6F702F7F7F2F7F8F2F8F
902F8F9F309FA02F9FAF30A0AF2F9FB0309FAF2F90B02F90AF2F8FAF2F8FA02F7F9F2F708F2F6F\
702F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000030303F30304F303F5F303060303F70
303F7F30408F30408F304F9F3050A0305FAF305FAF305FB03060BF3060C03060C0306FCF306FCF\
3070D03070D03070DF3070D03F7FDF307FD0307FDF3070DF3F7FDF3070D0307FD0
3070CF3F7FD03070CF3070CF3070C03F70BF306FB03070AF306FA03F608F305F70B03040B03040\
C03F40C03040C03040C0303FC0303FBF2F30BF2F3FB02F30B02F3FA02F309F2F30
7F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F303F2F5F3F2F6F402F80402F8F402F90402F9F4F2FA04F2FAF\
4F2FAF4F2FAF402FB0402FB04F30B0402FB0402FB0402FAF4030AF402FA0402FA0
402F9040308F3F2F703F2F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F2F504F2F606F2F70702F7F802F808F2F8F
9F30909F2F90AF30A0AF30A0AF30A0AF2F9FB02F9FAF2F90AF2F8FAF2F80A02F80902F70902F6F\
6F2F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F303F303040303F5F3030603F407F304080
304F8F304F903F509F3050A03F5FAF305FAF3F60BF3060BF3F60C03060C03F6FCF306FCF3F70D0\
3070D03F7FDF3070D03F7FDF307FDF3F80DF307FDF3F80DF3070D03F7FDF3070D0
3F7FD03070CF3F7FCF3070C03F70BF306FB03F70AF306F9F3F60809F2F3FAF3040AF2F3FB03040\
BF3040C03F4FC03040CF3040C0303FCF303FC02F3FBF303FB02F30B0303FA02F30
9030306F2F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F2F3F2F503F2F6F40307F402F8F403090402F9F4F30A04F2FAF4F30AF\
4F2FAF5030BF4F2FB04F30BF402FB04F30BF402FB04F30B0402FA04030A0402F9F
40308F3F2F7F3F305F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F4F502F6060206F702F7F802F80902F8F
902F8F9F309FA02F9FAF30A0AF2F9FAF2F9FAF2F90B02F90AF2F8FAF2F80A02F7F9F2F70802060\
6F2F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000303030302F4030305F303060303F7F303F7F30408F
30408F304F9F3050A0305FAF3050AF305FB0305FB03060C03060C0306FCF306FCF3070D0306FD0\
3070D03070D0307FDF3070D0307FDF307FDF307FDF307FD0307FD03070CF307FD0
3070CF3070C03070BF3070BF306FAF306FAF30609030607F802F30902F3F9F2F3FAF3040AF2F3F\
B03040BF3040C03F40C0303FC0303FC02F3FC0303FBF2F30BF2F3FB02F30AF2F30
9F2F308F2F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000302F50302F60402F7F402F80402F90402F904F2FA0402FA04F2FAF4F2FAF\
4F2FB04F2FB04F2FBF4F2FB04F2FB0402FB0402FB0402FA0402FA0402F9F402F90
3F2F7F3F2F60000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F2F402F50502F606F2F7F702F7F802F8F8F2F8F
9F309F9F309FAF30A0AF30A0B030A0AF2F9FB02F9FAF2F90B02F8FAF2F80A02F80902F6F8F2F60\
602F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F303F3030403030503030603F407F3040803F4F8F304F90
3F509F3050A03F5FAF305FB03F60BF3060BF3F60C03060C03F6FCF306FD03F70D03070D03F70DF\
3070D03F7FDF307FDF3F7FDF307FDF3F80DF307FD03F80D0307FD03F80D03070CF
3F7FCF3070BF3F70BF306FAF3F70AF30608F3F5F7F6F2F3F80303F8F2F3F9F303FA02F3FAF3040\
B03040BF3040BF3040C03F4FC03040CF3040C02F3FC0303FBF2F3FBF303FB02F30
AF303F902F307F3030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030304F3F2F6040307F402F80403090402F904F30A04F2FA04F30AF4F2FAF5030BF\
4F2FB05030BF4F2FB05030BF4F2FB05030B04F2FAF4F30AF402FA04030903F2F80
403070000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000302F303F20505F2F6F6F2F707F2F80802F80902F8F
902F90A0309FA02F9FAF30A0AF2F9FB02F9FAF2F90B02F90AF2F8FAF2F809F2F7F902F70802F5F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000303030302F3F303050303060303F70303F7F30408F304090304F9F
3050A0305FAF3050AF305FB0305FBF3060C03060C0306FCF306FCF3070D0306FD03070D03070D0\
307FDF3070D0307FDF307FDF307FDF307FD0307FD0307FCF3F80D0307FCF307FC0
307FBF3F70B0306FAF306FA0305F803F2F304F2F30602F3F702F3F802F3F8F2F3F9F2F3FA02F3F\
AF3040B0303FBF3040BF3040C03F40C0303FC0303FC02F3FC0303FBF2F30B02F3F
AF2F30A02F308F2F2F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000302F40302F5F3F2F703F2F80402F8F402F904F2FA04F2FA04F2FAF4F2FAF4F2FB04F2FB0\
4F2FBF4F2FB04F2FB04F2FB04F2FB04F2FAF4F2FAF4F2F9F4F2F9F402F80402F7F
302F4F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F30402F5F5F2F6F6F2F7F702F7F8F2F8F902F8F
9F309F9F2F9FAF30A0AF30A0B0309FAF2F90B02F9FAF2F8FAF2F8FA02F80A02F7F902F6F802F5F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030303030303F30305030305F3F3F70303F7F304F8F304F903F50A03050A0
305FAF305FB03F60BF305FBF3F60C03060C03F6FCF306FCF3F70D0306FD03F70DF3070D03F7FDF\
3070D03F7FDF307FDF3F7FDF307FD03F80D0307FCF3F7FD0307FCF3F7FC0307FBF
3F7FB03070AF3F709F305F8000000000000040303F502F306F303F7F2F3F8F303F902F3FA0303F\
A02F3FB03040B03040BF3F4FC03040C03F40C0303FC0303FC02F3FC0303FBF2F30
B0303FA02F2F9F30307F2F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F403F2F5F403070402F7F402F8F402F904F30A04F2FA04F30AF4F2FAF5030B04F2FB05030BF\
4F2FB05030BF4F2FB05030B04F2FAF4F30AF4F2FA04F309F402F8F4030803F2F50
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F303F2F5F5F2F706F207F7F2F80802F80902F90
902F90A0309FA030A0AF30A0AF2F9FB02F9FAF2F90B02F90AF2080AF2F809F2F70902F6F7F2050\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F3F30304F30305F303F70303F7F30408F30408F30509F304FA03050AF
3050AF305FB0305FBF3060C03060C0306FCF306FCF306FD0306FCF3070D03070D03070DF3070D0\
307FDF307FD0307FDF307FD0307FD0307FCF307FCF307FC0307FC03070B03070B0
3070A0306F9030607F0000000000000000002F2F30402F3F5F2F30702F3F7F2F3F8F2F3F902F3F\
A02F3FA02F3FB03040B03040BF3F40BF3040C03040C02F3FC0303FBF2F30BF3030
B02F2FB02F309F2F2F902F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F3F
302F50402F6F402F7F402F8F402F8F4F2F9F402FA04F2FAF4F2FAF4F2FB04F2FB04F2FBF4F2FB0\
4F2FB04F2FB04F2FB04F2FAF4F2FAF4F2FA04F2F9F402F8F402F7F3F2F5F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000402F5F502F6F6F2F807F2F808F2F90902F90
9F30A0A030A0AF30AFAF30A0B030AFB02FA0B02F9FAF2F90AF2F90AF2F8FA02F80902F70802F5F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30303F30304F30305F3F3F70303F7F30408F30408F3F509F3050A03F5FB03050B0
3F5FBF305FBF3F60C03060C03F6FCF306FCF3F70D0306FD03F70DF3070D03F7FDF3070D03F7FDF\
307FD03F7FDF307FD03F7FD0307FCF3F7FCF307FC03F7FC03070BF3F7FB03070AF
3F6F9030607F000000000000000000000000000000302F3050303F602F3F7F303F802F3F90303F\
9F2F3FAF3040AF303FB03040BF3040C03F4FC03040C03040C0303FC03030BF2F30
BF3030B02F30AF3030902F2F7F3030000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030303F3F2F50
40306F402F7F4F308F402F904F309F4F2FA04F30AF4F2FAF5030B04F2FB05030BF4F2FB05030BF\
4F2FB05030BF4F2FAF5030AF4F2FA04F30A0402F8F4030803F2F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F2050502F6F6F2F7F7F2F80802F8F902F9F
902F9FA030A0A030AFAF30AFAF2FA0B02FA0AF2F9FB02F90AF2F8FAF2F8F9F2F7F902F707F2F5F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F30\
30304F302F5030306F303F7030408030408F304F9F3050A0305FAF305FAF305FBF
305FBF3060C03060C0306FCF306FCF306FD0306FCF3070D03070D03070DF3070D0307FDF3070D0\
307FD03070D0307FD0307FCF307FCF3070C0307FBF3070B03070B0306FA0306F90
305F700000000000000000000000000000000000000000003F2F30502F3F602F307F2F3F802F3F\
902F3F9F2F3FAF303FAF303FB03040BF3040C03F40BF303FC0303FC02F30C03030
BF2F30BF3030AF2F2FA02F308F2F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F3F302F50402F6F
402F7F4F2F8F402F8F4F2F9F4F2FA04F2FAF4F2FAF4F2FB04F2FB0502FBF4F2FB04F2FBF4F2FB0\
502FBF4F2FAF4F2FAF4F2FA04F2FA0402F8F402F803F2F6F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F2F5F502F6F6F2F7F702F80802F908F2F90
9F309F9F2FA0AF30AFAF30AFB030A0AF2F9FB02FA0AF2F90AF2F90AF2F8FA02F80902F70802F60\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000303030303040\
3030503F3F6F303F7030408030408F3F4F9F30509F305FAF305FAF3F60BF3060BF
3060C03060C03F6FCF306FCF306FD0306FCF3F70D03070D03F70DF3070D03F7FDF307FD03F7FD0\
307FD03F7FD0307FCF3F7FCF307FC03F7FC03070B03F70B0306FA03F6F8F000000
000000000000000000000000000000000000000000000000302F30402F305F303F702F3F80303F\
8F2F3F9F303FA02F3FAF3040AF3040BF3F40BF303FC03F3FC03030C0303FC02F30
C03030B02F30B03030A02F309F3030702F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F303F2F5040306F402F7F
4F2F8F4F2F8F50309F4F2FA05030AF4F2FAF5030B04F2FB05030BF4F2FB05030BF4F2FB05030BF\
4F2FB05030B04F2FA04F30A0402F9040308F3F2F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F50502F6F602070702F80802F8F902F90
902F9F9F30A0A030A0AF30AFAF2FA0B02FA0AF2F9FB02F9FAF208FAF2F8F9F2F7F9F2F7080205F\
602F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000302F30303040302F50\
303060303070303F80303F80304F90304F9F3050AF3050AF305FB0305FB03060C0
3060C03060CF3060CF306FD0306FCF3070D0306FD03070D03070D03070DF3070D0307FD03070CF\
307FD03070CF307FCF3070C03070BF3070B03070AF306F9F306080000000000000
0000000000000000000000000000000000000000000000000000003F2F304F2F30602F3F702F30\
802F3F8F2F3F9F2F3FA02F3FAF303FB03030BF3F3FBF303FC0303FBF3030C03030
BF2F2FBF2F30B02F2FAF2F309F2F2F8F2F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F30302F4F402F60402F704F2F80
402F8F4F2F9F4F2FA0502FAF4F2FAF502FB04F2FB0502FBF4F2FB04F2FBF4F2FB0502FBF4F2FAF\
4F2FAF4F2FA04F2FA0402F90402F8F3F2F703F2F50000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F504F2F606F2F7F702F80802F8F8F2F90
9F309F9F30A0AF30AFAF30AFB030AFAF2FA0B02FA0AF2F90B02F90AF2F8FA02F80902F708F2F6F\
6F2F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003030403030503F3060\
303070303F7F3040803F4F90304F9F3050AF3050AF3F5FB0305FB03F60C03060C0
3F70CF306FCF3F6FD0306FCF3F70D03070D03F70DF3070D03F7FDF3070D03F7FD03070CF3F7FD0\
307FCF3F7FCF3070C03F7FBF3070B03F70AF306F9F3F6F804F3F2F4F3F3040302F
4F3F3040302F4F303040302F4F303040302F403030402F2F403030402F2F4030303F2F2F6F303F\
7F2F3F8F303F902F30A0303FAF3030B0303FB0303FC03F40C0303FC0303FC03030
C03030BF2F30BF3030B02F30AF3030902F307F3030000000000000000000000000000000000000\
000000000000000000000000000000000000302F4F40306F402F7F4F30804F2F8F
50309F502FA05030AF502FAF5F30B0502FB05F30BF4F2FB05030BF4F2FBF5030BF4F2FB05030B0\
4F2FAF4F30A0402F9F4F3090402F7F3F305F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030204F4F2F60602F70702F807F2F8F8F2F90
902F9F9F30A0A02FA0AF30AFAF2FA0B030A0AF2F9FB02F9FAF2F90AF2F8FA02F809F2F708F2F60\
6F2F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030303F302F4F30306030306F\
303F7F303F8030408F304F903050A03050A0305FB0305FB03060BF3060BF306FCF
306FCF3070CF3070CF306FD0306FD03070D03070D03070D03070D0307FD03070CF307FCF3070CF\
307FCF3070BF3070BF3070AF3070AF306F9030607F4F402F504F304F402F4F4F30
4F402F504F304F402F5040304F402F504F304F402F5040304F402F5040304F3F2F4F40304F3F2F\
4F3F30403F2F4F3F3040302F403030402F2F4030303F2F2F3F3030BF3030C03030
C02F30C03030BF2F2FB02F30AF2F2FA02F308F2F2F000000000000000000000000000000000000\
000000000000000000000000000000302F4F402F60402F704F2F804F2F8F4F2F9F
4F2F9F502FAF502FAF502FB0502FB05F2FBF502FB0502FBF4F2FB0502FBF4F2FB04F2FB04F2FAF\
4F2FA0402F9F402F8F402F7F3F2F60000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F4F402F60602F7F6F2F80802F8F8F2F90
902F9F9F2F9FAF30AFAF30AFB030AFAF2FA0B030A0B02F9FB02F90AF2F8FAF2F8F9F2F70902F6F\
6F2F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000030303030304F30306030306F303F7F\
303F803F4090304F903050A03050A03F5FB0305FB03060BF3060BF3F6FCF306FCF
3070CF3070CF3F70D03070D03070D03070D03F7FDF3070D0307FD03070CF3F7FD03070CF3F7FCF\
3070BF3F7FBF3070AF3070AF306F903F607F50502F505F3050502F505F3050502F
505F3050502F505F3050502F50503050502F5F5F3050502F5F503050502F5F503050502F505030\
4F502F5050304F4F2F504F304F4F2F504F304F402F4F40304F3F2F4F40304F3F2F
4F3F3040302F403F3040302F403030A02F2F9F3030702F2F000000000000000000000000000000\
000000000000000000000000302F4F403060402F704F2F804F2F8F50309F502FA0
502FAF502FAF5F30B0502FB05F30BF502FB05F30BF502FBF5F30BF502FB05030B04F2FAF4F30AF\
402F9F4F3090402F803F306F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F3F402F605F20706F2F807F2F808F2F90
902F909F30A0A02FA0AF30AFAF2FA0AF2FA0AF2F9FB02F9FAF2F8FAF2F8FA02F809F2F7F8F2060\
702F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F3F30305F30306F303F70303F7F\
30408F304090304FA0304FA03050AF305FAF3060BF305FBF3060C0306FC0306FCF
306FCF3070CF3070CF307FD03070D03070D03070D03070D03070CF3070CF3070CF307FC03070BF\
3070B0306FAF3070AF30608F306070505F2F506030505F2F506030505F2F5F6030
505F2F5F6030505F2F5F6030505F2F5F6030505F2F5F6030505F2F5F6030505F2F5F6030505F2F\
506030505F2F505F30505F2F505F30505F2F505F3050502F5050304F502F505030
4F4F2F504F304F4F2F504F304F402F4F40304F402F4F4030403F2F403F303F302F3F3030000000\
000000000000000000302F403F2F60402F704F2F80402F8F4F2F9F4F2F9F502FAF
502FAF502FB0502FB05F2FBF502FB0502FBF502FB05F2FBF502FB0502FB0502FAF502FAF402F9F\
402F90402F803F2F6F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F3F402F5F5F2F706F2F807F2F8F802F90
90309F9F2F9FA030AFAF30AFB030AFAF2FA0B030A0B02F9FB02F90AF2F8FAF2F8F9F2F70902F6F\
702F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303F30305030306F303F70303F7F3F408F\
304090304F9F3050A03F5FAF305FAF3060BF3060BF3F6FC0306FC0306FCF306FCF
3F70D03070CF3F7FD0307FD03F7FD0307FD03F7FD03070CF3F7FD03070CF3F7FC03070BF3F70BF\
3070AF3F70A030608F5F6F3050602F5F6F3050602F5F6F305F6F2F5F6F305F6F2F
5F6F305F6F2F5F6F305F6F2F5F70305F6F305F70305F6F2F6070305F6F30607030506F2F5F7030\
5F6F305F6F30506F2F5F6F30506F305F6F3050602F5F6F305060305F6F3050602F
5F60305060305F6030505F2F5F5F30505F30505F3050502F5050304F50304F50304F4F2F4F4F30\
4F40304F4030403F2F403F303F30304F30804F2F8F50309F502FA05030AF502FAF
5F30B0502FB05F30BF502FBF5F30BF502FBF5F30BF502FB05F30B0502FAF5030AF4F2F9F50309F\
4F2F80403070000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2030402F60502F706F2F80702F808F2F90
902F909F30A09F2FA0AF30AFAF30A0B030A0AF2F9FB02F9FAF2F90AF2F8FA02F809F2F7F902F6F\
7F2F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F30303040302F5F30307030307F30408030408F\
304F9F304F9F3050AF3050AF305FB0305FB03060C03060C0306FCF306FC03070CF
3070CF3070D03070CF307FD0307FCF307FD0307FCF307FCF307FC03070C03070BF3070B0306FAF\
306FA03060805F6F30506F2F5F7030506F2F5F70305F6F2F5F70305F6F2F5F7030
5F702F5F70305F702F5F70305F702F6070305F702F6070305F702F607F305F702F5F7F305F702F\
5F7F305F702F5F70305F702F5F70305F702F5F70305F702F5F70305F6F2F5F7030
5F6F2F5F70305F6F2F5F6F305F6F2F5F6F30506F2F5F6F3050602F5060304F5F2F505F304F5F2F\
505F304F502F4F50304F4F2F4F4F304F402F4F4030403F2F403F303F302F403030
502FB05F2FBF502FB05F2FBF502FBF5F2FBF502FB0502FB0502FAF502FAF4F2F9F4F2F90402F80\
402F70000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F303F2F5F5F2F706F2F7F7F2F8F802F90
902F9F902F9FA030A0AF30AFAF30AFAF2FA0B030A0B02F9FB02F90AF2F8FAF2F8F9F2F7F902F6F\
7F2F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F2F302F2F30304030305030306F30307030408030408F304F9F\
304F9F3050AF3050AF305FB0305FB03F60C03060C0306FC0306FC03F70CF3070CF
3070D03070CF3F7FD0307FCF307FCF307FCF3F7FCF307FC0307FC0307FBF3F7FB03070AF3070A0\
306F805F70305F702F5F703050702F5F7F305F702F5F7F305F702F5F7F305F7F2F
5F7F305F7F2F607F305F7F2F607F305F7F2F607F305F7F2F607F305F7F2F6080305F7F2F5F8030\
5F7F2F5F80305F7F2F5F80305F7F2F6080305F7F2F6080305F7F2F6080305F7F30
607F305F7F2F607F305F7F30607F305F702F607F305070305F703050702F5F70305070305F7030\
506F2F5F6F3050603050603050602F506030505F30505F304F502F5050304F4F30
4F4F3040402F4F4030403F2F403F303F2F2F5F30B0502FAF5030AF4F2FA050309F4F2F8F402F70\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F2F5F50206F6F2F7F702F80802F90
8F2F909F2F9F9F2FA0AF30AFAF2FA0AF30A0AF2F9FB02F9FAF2F90AF2F8FA02F80A02F7F90206F\
7F2F50000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000302F2F302F2F30303F302F4F30306030306F303F7F303F8F304F90304F9F\
3050A03050A0305FB0305FB03060BF3060BF306FC03060C0306FCF306FC03070CF
3070CF3070CF3070CF307FCF3070C0307FCF307FC0307FC03070B0307FB03070A030709F306080\
5F70305F702F5F7F3050702F5F7F30507F2F5F7F30507F2F5F7F305F7F2F5F8030
5F7F2F5F80305F7F2F6080305F7F2F6080305F802F6080305F802F6080305F802F5F80305F802F\
5F80305F802F5F80305F802F5F80305F802F6080305F802F6080305F802F608030
5F802F6080305F802F6080305F802F6080305F7F2F5F8030507F2F5F7F30507F2F5F7F3050702F\
5F7F3050702F5F7F3050702F5F7030506F2F507030506F2F506F3050602F506030
4F5F2F505F304F5F2F505F3040502F4F503040402F4040303F3F2F403030402F80402F7F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F2F5F502F70602F7F7F2F8F802F8F
902F9F902F9FA030A0A030A0B030AFAF30A0B030A0B02F9FB02F9FAF2F8FAF2F8FA02F7F9F2F70\
802F5F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F3F30303F2F2F30303F30304030305F30306F303F7F304080304F90304F9F3F50A0\
3050A0305FB0305FB03F60BF3060BF306FC03060C03F6FCF306FCF3070CF3070CF
3F7FD03070CF307FCF3070C03F7FCF307FC03F7FC03070B03F7FB03070AF3F709F306F805F7F30\
5F7F2F5F7F305F7F2F5F80305F7F2F5F80305F7F2F5F80305F80305F80305F802F
6080305F8030608F305F802F608F30608F30608F30608F2F608F30608F305F8F305F8F2F609030\
5F8F30608F305F8F2F6090305F8F30608F305F8F2F609030608F30609030608F2F
609030608F30609030608F2F609030608F306090305F8F2F5F8F305F8F305F8F305F802F5F8F30\
5F80305F80305F802F5F80305F80305F80305F7F2F5F80305F7F305F7F3050702F
5F70305070305F7030506F2F506F304F60305060304F5F2F4F5F304F50304F5030404F2F4F4030\
403F30403F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F2F5F4F2F6F6F2F7F702F80802F90
8F2F909F309F9F2F9FA030AFAF30A0B030A0AF2F9FB02F9FAF2F90AF2F90AF2F8FA02F80902F70\
902F6F6F2F40000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F2F3F302F30302F3F302F302F3F303050302F60303070303F7F30408F304090304FA0304FA0\
3050AF3050AF305FBF305FBF3060C03060BF306FC0306FC03070CF306FCF3070CF
3070CF307FCF3070C0307FC03070C0307FC03070B03070B03070A030709F3060805F7F305F7F2F\
5F80305F7F2F5F8030507F2F5F8030507F2F5F80305F802F5F80305F802F5F8030
5F802F5F8F305F802F5F8F305F8F2F608F305F8F2F608F305F8F2F608F305F8F2F5F90305F8F2F\
5F90305F8F2F5F90305F902F6090305F8F2F6090305F902F6090305F8F2F609030
60902F6090305F902F60903060902F6090305F902F5F90305F902F5F90305F8F2F5F90305F8F2F\
5F90305F8F2F5F8F305F8F2F5F8F305F802F5F8F305F802F5F8F305F802F5F8030
5F802F5F80305F7F2F5F7F30507F2F507F304F702F5070304F6F2F506F304F602F4F60304F5F2F\
4F603040502F4F503040402F404030403F2F403030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F504F2F6F602F7F702F8F7F2F8F
8F2F9F902F9FA030A0A030A0AF30AFAF2FA0B030AFB02FA0B02FA0AF2F90AF2F90A02F8F9F2F7F\
902F6F7F2F50000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F30302F2F
3F302F3F302F403F3030303F30304F302F5F30306F30307030408030409030509F3050A03050AF\
3050AF305FBF305FBF3060BF3060BF3F6FC0306FC03070CF306FCF3F70CF3070CF
3070CF3070C03F7FCF3070C0307FC03070B03F7FB03070A030709F3060805F7F30507F2F5F8030\
507F2F5F80305F802F5F803050802F5F803050802F5F8F3050802F5F8F305F8F2F
5F8F305F8F2F608F305F8F2F608F305F8F2F6090305F8F2F6090305F8F2F5F90305F90305F9030\
5F902F6090305F90306090305F902F609F305F9030609F305F902F609F30609F30
609F3060902F609F30609F30609F3060902F609F305F9F305F9F305F902F609F305F9F30609F30\
5F902F609F305F9030609F305F902F609F305F90306090305F8F2F6090305F8F30
6090305F8F2F608F305F8F305F8F3050802F5F8F30508030508030507F2F507F30507030507030\
4F6F2F5070304F6F30506F304F602F5060304F5F304F5030404F2F4F40303F3F2F
403030000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F4F40206F602F806F2F8F7F2F90
80209F902FA09F2FA0A030B0AF30AFAF30AFAF2FAFB02FAFAF2F9FB02F9FAF2F90AF2F8F9F207F\
902F707F2F50000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F2F2F2F2F30302F
3F302F403F2F3F3F2F3030402F2F5030306030306F303F80303F8F304F90304F9F3050A03050AF\
305FB0305FB03060BF3060BF3060C03060C0306FC0306FC03070CF3070C03070C0
3070C03070C03070BF307FBF3070B03070AF3070A030709F30607F507F30507F2F507F30507F2F\
508030507F2F5F8030507F2F50803050802F50803050802F5F8F3050802F5F8F30
50802F5F8F305F8F2F5F8F305F8F2F5F90305F8F2F6090305F8F2F60903050902F5F903050902F\
5F90305F902F5F9F305F902F5F9F305F902F609F305F902F609F305F9F2F609F30
5F9F2F609F30609F2F60A030609F2F609F305F9F2F5FA0305F9F2F5F9F305F9F2F5F9F305F9F2F\
5F9F305F9F2F609F305F9F2F609F305F9F2F609F305F9F2F609F305F902F609F30
5F902F6090305F902F5F9030508F2F509030508F2F5F9030508F2F508F3050802F508F3050802F\
508030507F2F507F304F702F5070304F702F5070304F6F2F4F6F304F602F4F6030
40502F4050303F402F4040303F302F3F3030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F30402F6F5F2F7F6F2F8F7F2F90
8F2F9F902FA09F30AFA030AFAF3FB0AF30B0B030AFB02FA0B02FAFB02F9FB02F9FAF2F8FA02F80\
9F2F708F2F606F2F3F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F30302F2F3F30303F302F
40403040402F404F30302F4F30306030306F303F7F303F80304F90304F903050A03050A03F5FB0\
3060B03060BF3060BF3F6FC0306FC0306FC0306FC03F70CF3070C03070C03070C0
3F7FC03070C0307FBF3070B03F7FB03070A030709F306080508030507F2F508030507F2F5F8030\
50802F5F803050802F508F3050802F508F3050802F5F8F30508F305F8F30508F2F
5F90305F8F305F90305F8F2F5F90305F90305F90305F902F6090305F90305F9F305F902F5F9F30\
5F9F305F9F305F9F2F609F305F9F30609F305F9F2F60A0305F9F3060A0305F9F30
60A03060A03060A03060A03060A03060A0305FA0305FA03060AF305FA03060A0305FA03060A030\
5FA03060A0305FA03060A03060A03060A0305FA03060A03060A03060A03060A02F
60A03060A03060A0305F9F2F5FA0305F9F305F9F3050902F5F9F305090305F903050902F5F9030\
5090305F9030508F2F5F8F30508030508030507F2F508030507F30507F3050702F
4F70304F6F304F6F3040602F4F603040503040503040402F4040303F3F303F3030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F5F502F70602F80702F90
802F90902FA0902FA09F30AFA030AFAF30B0AF2FAFB030AFB02FA0B02FA0AF2F90AF2F90A02F80\
9F2F7F8F2F60702F4F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F2F2F2F2F30302F30302F3F3F2F
3F402F404F2F2F2F3F30305F303060303070303F7F30408F304090304F9F304F9F305FAF305FAF\
3060BF3060B03060BF3060BF306FC0306FC0306FC0306FC03070C0306FBF3070C0
3070BF3070BF3070B03070AF306FA0306F9F30607F4F7F304F7F2F507F304F7F2F507F304F7F2F\
5080304F7F2F508030507F2F5080304F802F5080304F802F508F3050802F508F30
508F2F508F30508F2F508F30508F2F5F9030508F2F5F903050902F5F903050902F50903050902F\
5F9F3050902F5F9F305F9F2F5F9F305F9F2F5F9F305F9F2F5FA0305F9F2F5FA030
5F9F2F60A0305F9F2F60A0305FA02F60A0305FA02F5FA0305FA0305FAF305FA0305FAF305FA030\
60AF305FA03060AF305FA03060AF305FA02F60AF305FA03060AF305FA02F60AF30
60A03060AF3060A02F5FAF305FA02F5FA03050A02F5FA0305F9F2F5FA030509F2F5F9F30509F2F\
5F9F3050902F5F9F3050902F5F9030508F2F5F9030508F2F5F8F3050802F508030
4F802F4F80304F702F4F7F304F702F4F7030406F2F4F6F3040602F405F3040502F404F303F402F\
4040303F302F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F50402F6F602F80702F8F
802F9F8F2F9F9030A09F2FAFA030B0AF30B0B030B0AF2FAFB02FAFB02FA0B02F9FAF2F90AF2F8F\
9F2F7F9F2F707F2F50000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F2F30302F30302F4040303F402F
404F30404F2F30304F30305F30306F30307030408F30408F304F90304F9F305FAF305FAF3060B0\
3060B0306FBF306FBF306FC0306FC03070C0306FC03070C0306FBF3070C03070BF
3070BF3070B03F70AF3070A0306F9F30607F4F7F304F7F2F4F7F304F7F2F4F7F304F7F2F4F7F30\
4F7F2F5080304F7F2F5080304F7F2F4F80304F802F5080304F802F508F304F802F
508F30508F2F508F30508F2F508F30508F2F5F9030508F2F5F903050902F509030509030509030\
50902F5F9F30509F305F9F30509F2F5F9F305F9F305FA0305F9F2F60A0305FA030
60A0305FA02F60A0305FA03060A0305FA02F5FAF305FA0305FAF305FA03060AF305FAF3060AF30\
5FAF3060AF3F5FAF3060B0305FAF3060B03F60AF3060AF3060AF3060B03F60AF30
60AF3060AF3060AF305FAF305FAF305FAF305FAF305FAF305FAF305FAF2F60AF305FA0305FA030\
5FA02F5FA0305FA0305FA0305F9F2F5FA0305F9F305F9F3050902F5F9F305F9030
5090304F8F2F508F304F8F304F80304F802F4F80304F7F2F4F7F304F702F4F70304F6F2F4F6F30\
405F2F4F5F3040502F404F303F402F3F3F3030302F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2030402F6F502F7F6F2F8F
70208F802F9F8F2FA09F30AF9F2FAFAF30B0AF30AFB030AFAF2FA0B02FA0AF2F9FAF2F90A0208F\
A02F80902F6F8F2F60602030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F2F30302F30302F3F3F2F3F402F404F2F
404F2F3F6050302F503030602F306F303F7F303F80304090304F903050A03050A0305FAF305FAF\
3060B03060B0306FBF306FBF3070C0306FBF306FC0306FBF3070BF3070BF3070BF
306FAF3070AF306FA0306F9030607F40702F40702F40703040702F40703040702F4F7F3040702F\
4F7F3040702F4F7F30407F2F407F30407F2F4F8030407F2F4F80304F802F4F8030
4F802F4F80304F802F508F304F802F508F304F8F2F508F304F8F2F5090304F8F2F5090304F902F\
50903050902F509F3050902F509F30509F2F5F9F30509F2F5F9F30509F2F5FA030
5F9F2F5FA0305FA02F5FA0305FA02F5FA03050A02F5FAF3050A02F5FAF305FAF305FAF305FAF30\
5FAF305FAF305FAF305FAF305FB0305FAF3060B03F5FAF3060B0305FAF3060B03F
5FAF3060B03060AF305FB03050AF305FB0305FAF305FB0305FAF2F5FAF305FAF2F60AF305FAF2F\
60AF305FAF2F60AF305FA02F5FAF305FA02F5FA0305F9F2F5FA0305F9F2F5FA030
4F9F2F509F304F902F509F304F902F4F90304F8F2F4F8F304F802F4F80304F7F2F4F7F3040702F\
4F703040702F4F703040602F4060303F5F2F4050303F4F2F3F40303F3F2F3F3F30
2F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F2F60502F7F602F80
7F2F90802F9F902FA0902FA0A030AFA030B0AF30B0AF30AFB030AFB02FA0B02FA0AF2F90B02F90\
A02F809F2F7F8F2F60702F4F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F2F30302F30302F3F40303F402F404F30404F2F
4F5F3030303030305F30306F303F7F303F8030408F30409030509F3050A0305FAF305FAF3060B0\
3060B0306FBF306FBF3F70C03070BF3070C03070BF3F7FC03070BF3070BF306FAF
3F70AF306FA0306F90306F7F4070303F6F2F4070303F6F2F40703040702F40703040702F407030\
40702F40703040702F40703040702F40703040702F407F30407F2F4F7F30407F2F
4F8030407F304F80304F802F4F80304F80304F8F304F802F508F304F8F304F8F304F8F2F509030\
4F90305090304F902F509F30509030509F3050902F5F9F30509F305F9F30509F2F
5FA0305FA0305FA0305FA0305FAF3050A0305FAF3050A0305FAF305FAF305FAF305FAF305FB030\
5FAF305FB0305FAF3060B03F5FB03060B03F5FB03060B03F60B03060B03F60B030
60BF3F60B03F5FBF3F5FB0305FBF3F5FB0305FB03F5FB03060BF3F5FB03060B0305FB03060B030\
5FB03060B0305FB03060B03060B03060B03060AF2F60B03060AF3060AF3050A02F
5FAF3050A03050A030509F2F50A030509F30509F304F9F2F509F304F90305090304F8F2F508F30\
4F80304F80304F7F2F4F7F304F7F30407030406F2F4070304060304060303F502F
4050303F4F2F3F40303F3F2F3F3030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F50402F6F602F80
6F2F8F802F9F8F2F9F902FA09F2FA0A030B0AF30AFAF30AFAF2FAFB02FAFB02F9FB02F9FAF2F8F\
AF2F8F9F2F7F902F6F7F2F4F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F2F30302F3F3F2F3F402F404F2F3F4F2F40502F
40502F3030402F2F5F30307030307F30408030408F304F90304F9F3050A03050A0305FAF305FAF\
3060B03060B0306FBF306FBF3070BF306FB03070BF3070B0307FB03070AF3070AF
30709F30709F306F8030605F305F2F3F6030305F2F30603030602F3F603030602F3F603030602F\
3F6F3030602F3F6F303F602F3F6F3030602F3F6F303F6F2F3F70303F6F2F3F7030
3F702F407F303F702F407F30407F2F407F30407F2F408030407F2F4F803040802F40803040802F\
4F8F3040802F4F8F304F8F2F4F90304F8F2F5090304F902F509F304F902F509F30
509F2F509F30509F2F509F30509F2F50A030509F2F50A03050A02F50AF3050A02F5FAF3050AF2F\
5FAF3050AF2F5FAF305FAF305FB0305FAF305FB0305FB03060B0305FB03060B030
5FB03060BF3F50B0305FBF3F5FBF305FBF3F5FB0305FBF3F5FB0305FBF3F5FB0305FBF305FB030\
60BF305FB03060B0305FB03060B0305FB02F60B03060B02F60B03060B02F5FB030
50AF2F5FB03050AF2F5FAF3050A02F50AF3050A02F50A030509F2F50A0304F9F2F509F304F902F\
509F304F902F4F90304F8F2F4F8F3040802F408030407F2F407F303F702F407030
3F6F2F406F303F5F2F3F5F303F502F3F4F303F402F3F3F3030302F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3F402F6F5F2F7F
6F2F8F7F2F908F2FA0902FA09F30AFA030AFAF3FB0AF30AFB030AFB02FA0B02FA0B02F9FB02F90\
AF2F8FA02F80902F6F8F2F5F602F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F2F302F2F3F40303F402F404F2F404F2F405F30405F2F
4F5F2F302F4F303060303070303F80303F80304F90304F903050A03050A0305FAF305FAF3060B0\
3060B0306FBF306FBF3070BF306FB03070BF3070B03070B03070AF3070AF3070A0
30709F306F80306F702F4F2F000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F4F2F3050302F502F305F30305F2F
30603030602F3F6F303F6F2F3F6F303F6F2F3F70303F702F4070303F702F407F303F7F2F408030\
407F2F4F803040802F4F8F3040802F4F8F304F8F304F90304F8F2F5090304F9030
509F304F902F509F30509F30509F304F9F2F50A03050A03050A03050A02F50AF3050A03050AF30\
50AF2F5FAF3050AF305FB03050AF305FB0305FB0305FB0305FB0305FB0305FB030
60B0305FB0305FBF3F50BF305FBF3F50BF305FBF3F5FBF3F5FC03F5FBF3F5FC03F5FBF3F5FC03F\
5FBF3060C03F5FBF3060BF3F5FBF3060BF3F60BF3060BF3060BF3060BF305FBF30
5FBF305FB02F5FBF305FB0305FB03050AF2F5FB03050AF305FAF3050A02F5FAF3050A03050A030\
509F2F50A030509F2F509F304F902F4F9F3040902F4F8F3040802F4F8F3040802F
407F30407F2F407F3040702F4070303F602F4060303F5F2F3F5030304F2F3F4F30303F2F3F3F30\
302F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F60502F7F
602080702F90802F90902FA0902FA09F30AFA030AFAF30B0AF2FAFB02FAFB02FA0B02F9FAF2090\
AF2F8FA02F809F2F708F2060702F40000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F2F2F202F30302F3F3F2F3F4F2F3F4F2F40502F40502F405F2F
2F2F3030304F303060303F7F2F3F7F30408F304090304F9F2F4F9F3050A0305FA03060AF305FAF\
3060B03060B0306FBF306FB03070B0306FB03070B03070AF3070AF30709F30709F
306F80306F70000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F2F2F2F2F2F2F2F2F2F2F2F2F2F2F302F2F2F2F2F302F2F302F2F3F2F
2F302F2F3F302F3F2F2F40302F40302F4F302F402F3050302F5F2F30603030602F306F3030602F\
3F6F303F6F2F3F70303F702F407F303F7F2F40803040802F40803040802F4F8F30
408F2F4F8F304F8F2F4F90304F902F4F9F304F902F4F9F304F9F2F4FA0304F9F2F50A0304FA02F\
50A03050A02F50AF3050A02F50AF3050AF2F5FB03050AF2F5FB03050B02F5FB030
50B02F5FB03050B03050BF3050B03050BF3050BF305FBF3F50BF305FC03F50BF305FC03F50BF30\
5FC03F5FC0305FC03F5FC0305FC03F5FC03060C03F5FBF3060C0305FBF305FC030
50BF305FBF3050BF305FBF3050B02F5FBF3050B02F5FB03050AF2F50B03050AF2F50AF3050A02F\
50AF3050A02F50A0304FA02F50A030409F2F4F9F3040902F409F3040902F409030
408F2F408F3040802F4080303F7F2F407F303F702F4070303F6F2F3F6F30305F2F3F5F3030502F\
3F4F2F30402F3F403030302F30302F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F5F4F2F6F
602F8F702F8F802F9F8F2F9F9F30AF9F30AFA030B0AF30B0B030B0B02FAFB030AFB02F9FB02F9F\
AF2F90AF2F8F9F2F7F9F2F6F7F2F50000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F2F3030303F3F2F404F2F404F2F405F30405F2F405F30405F2F
30304030305030306F30307030408030408F304F9F304F9F305FA0305FA0305FAF305FAF3060B0\
3060B0306FB0306FB03070B0306FB03070B03070AF3070AF3070A030709F306F8F
306F7F2F302F3030302F302F3030302F302F3030302F302F303F302F302F3030302F302F303F30\
2F3F2F303F302F3F2F3040302F3F303040302F4030304F302F4F30304F3F2F4F30
30503F2F503F2F50402F503F305F402F5F402F60402F5F4030604F2F604F2F604F2F604F306F50\
2F6F502F6F502F402F3060303060303F6F303F6F2F3F70303F7030407F303F7F2F
40803040803040803040802F4F8F304F8F304F903040902F4F9F304F90304F9F304F9F2F50A030\
4F9F3050A0304FA03050AF3050A03050AF3050AF305FAF3050AF305FB03050B030
5FB0305FB03050BF3050B0305FBF3050BF305FBF3050BF305FC03F50BF305FC03F50C0305FC03F\
5FC03F5FCF3F5FC03F5FCF3F5FCF3F5FCF3F5FC03F60CF3F5FC03F60CF3F50C030
5FCF3F50C0305FC03F50C0305FC03F50BF305FC03050BF305FBF3050BF305FBF3050B02F5FBF30\
50B03050B03050AF2F50B03050AF304FAF3040A02F4FA03040A0304FA030409F2F
4F9F304090304F9030408F2F4F9030408F2F408030407F2F4080303F7F2F3F70303F6F2F3F6F30\
3F602F3F6030305F2F3F5030304F2F304030303F2F303F3030302F303030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2040402F6F
5F2F7F6F2F8F702F908F2F9F902FA09F30AF9F2FAFAF30B0AF30AFB030AFAF2FA0B02FA0AF2F9F\
B02F90AF2F8FA02F80902F70902F6F6F2F40000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F202F2F2F2F30302F3F402F3F4F2F40502F40502F405F2F405F2F40602F
2F2F3F3030502F306F303F7F303F80304F90304F9030509F30509F305FA0305FA03060AF3060AF\
306FB03060AF306FB0306FAF3070B0306FAF3070A0306F9F30709F306F8F306F7F
2F3F2F3040302F3F2F303F302F3F2F3040302F3F2F2F40302F3F2F3040302F402F2F40302F4030\
2F4F302F4F302F4F302F4F302F503F2F4F302F503F2F503F2F5F402F503F2F5F40
2F5F402F604F2F5F402F604F2F60402F6F4F2F604F2F6F4F2F6F4F2F6F502F6F502F70502F7050\
2F705F2F70502F7F5F2F7F5F2F7F5F2F7F5F302F30302F2F4F30302F5F2F306030
30602F3F6F303F6F2F3F7F303F7F2F4080303F7F2F40803040802F408F30408F2F40903040902F\
4F9F3040902F4F9F304F9F2F4FA0304F9F2F50A0304FA02F50AF304FA02F50AF30
50AF2F50AF304FAF2F50B0304FB02F50B0304FB02F50BF3050BF3050BF3050BF3050BF3050BF30\
5FC03F50C0305FC03F50C0305FCF3F50C0305FCF3F5FCF3F5FCF3F50CF3F50CF3F
50CF3050CF3F50C03050CF3F50C03050C03F50C03050C03050BF3050C03050BF3050BF3050BF2F\
5FBF3050B02F50BF3050B02F50B03040AF2F4FB03040AF2F4FAF3040A02F4FA030
40A02F4FA030409F2F409F3040902F409F3040902F409030408F2F3F8F303F802F3F7F303F7F2F\
3F7030306F2F3F6F3030602F3F6030305F2F3F5F30304F2F304F302F402F303F2F
2F302F3030302F2F2F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F2F60
502F7F602F807F2F90802F9F902FA0902FA0A030AFA030AFAF30B0AF30AFB030AFB02FA0B02FA0\
AF2F9FB02F9FAF2F8FA02F80902F6F802F5F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F2F30302F3F402F3F402F40503040502F405F2F405F2F4F603040602F
303040302F5F303F6F30307030408030408F30509F30509F305FA0305FA03060AF3060AF3060AF\
3060AF306FB0306FAF3070AF306FAF3070AF306F9F30709F306F8F306F8030605F
304F302F402F304F302F4F2F304F302F4F303050302F4F3030503F2F4F3030503F2F503030503F\
2F503F305F3F2F5F3F305F402F5F3F3060402F5F402F60402F6040306F4F2F604F
2F6F4F2F6F4F306F502F6F502F70502F6F5030705F2F70502F7F5F2F705F307F5F2F7F5F2F8060\
2F7F5F2F80602F80602F80602F80603F2F3F402F50303030402F2F5F3030702F2F
8030308F2F2F9F3030A03030305F30305F2F3F6030306F303F70303F7F2F4080303F8030408F30\
408F2F409030408F304F903040902F4F9F304F9F304FA0304F9F2F50A0304FA030
50A0304FA02F4FAF304FAF3050AF304FAF2F50B0304FB03050B03050B03050BF3050BF3050BF30\
50BF305FC03050BF305FC03050C0305FC03F5FC0305FCF3F5FC03050CF3F50CF3F
50CF3F50CF3050D03F50CF3F50CF3F50CF3050CF3F50C03050CF3F50C03050C03F50C03050C030\
50BF3050C03050BF3050BF3050BF2F4FBF304FB02F4FB0304FB02F4FB0304FAF2F
4FAF3040A02F4FAF3040A02F4FA030409F2F4FA030409F2F409F3040902F4090303F8F2F3F8F30\
3F802F3F80303F7F2F3F7F303F702F3F70303F6F2F306030305F2F305F3030502F
304F3030402F3040302F3F2F303030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F5F
40206F602F806F2F8F802F9F8F209F902FA09F2FAFA030B0AF30AFAF30AFAF2FA0B02FAFB02FA0\
B02FA0AF2F90AF2F90A020809F2F708F2F607F2F4F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F2F2F303F2F30402F3F4F2F3F502F405F2F405F2F4F602F40602F406F2F
40602F30305F2F306F303F7F2F3F7F304F8F304F9030509F2F509F305FA0305FA03060AF2F60A0\
3060AF3060AF306FAF306FA0306FA0306F9F306F9F2F6F8F306F8030606F305030
2F4F2F3050302F4F303050302F4F302F50302F5030305F3F2F50302F5F3F2F503F2F5F3F2F5F3F\
2F60402F60402F60402F60402F604F2F604F2F6F4F2F6F4F2F6F4F2F6F4F2F7050
2F6F502F70502F70502F7F5F2F70502F7F5F2F7F5F2F7F5F2F7F5F2F80602F7F5F2F80602F8060\
2F80602F80602F8F6F2F80603F2F40402F5F4F2F6F4F2F70302F2F4F2F2F6F3030
702F2F8F302F902F2F9F3030A02F2FAF3030B03030BF3030B0302FBF3030305F2F30603030602F\
3F70303F702F3F80303F7F2F4080303F802F408F30408F2F40903040902F4F9F30
4F9F2F4FA030409F2F4FA03040A02F4FAF304FA02F4FAF304FAF2F4FB0304FAF2F50B0304FB02F\
50B03050B02F50BF3050BF2F50BF3050BF3050C03050BF3050C0304FC03050C030
4FC03050CF304FC03050CF3F4FC03050CF3050CF3050CF304FC03050CF3050C03050C0304FC030\
50C03050C02F50C0304FBF2F50BF3040BF2F4FBF3040B02F4FBF3040B02F4FB030
40AF2F4FB03040AF2F4FAF3040A02F40A03040A02F40A030409F2F409F303F902F3F9F303F902F\
3F90303F8F2F3F8F3030802F3F802F3F7F2F3F7F30306F2F306F2F30602F306030
2F5F2F305F2F2F502F304F2F2F402F303F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F40
402F705F2F7F6F2F907F2F908F2FA0902FA09F30AFA030AFAF3FB0AF30BFB030B0B02FAFB030AF\
B02FA0B02FA0AF2F90AF2F90A02F7F9F2F70802F5F6F2F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030302F303F2F3F4F3040502F405F2F405F2F4F60304F602F4F6F30406F2F
4F703030305F303F6F303F70304080304F8F30509030509F305FA0305FA03060AF3060A0306FAF\
3060AF306FAF306FA0306FA0306F9F306F9F306F8F306F80306F7030605F2F5030
305F3F305F30305F3F2F5F3F305F3F2F5F3F30603F2F5F3F3060402F60403060402F6040306F4F\
2F6F4F306F4F2F6F4F3070502F6F503070502F70503070502F7050307F5F2F7050
307F5F2F7F5F307F602F7F5F3080602F7F602F80602F8060308F602F80602F8F6F2F8060308F6F\
2F8F6F2F8F6F302F2F40304F4F2F5F502F70502F7F5F308F302F2F4F3030602F2F
7F303080302F9030309F302FA03030AF3030B03F3FB03030BF3F30BF3030BF3F30B0302FB03030\
AF302FA0303030502F3F6030306F303F70303F7F2F4080303F8030408F3040802F
40903040903040903040902F4FA03040A0304FA0304FA02F4FAF304FAF304FAF304FAF3050B030\
4FB03050B0304FB03050BF3050BF3050BF3050BF3050C03050BF3050C0304FBF30
50C0304FC03050C0304FC03050CF304FC03050CF304FC03050CF3F50CF3050CF3050CF3050CF30\
50C03050CF3050C03050C0304FC0304FC0304FBF2F4FC0304FBF304FBF3040BF2F
4FBF304FB0304FBF3040B02F4FB03040AF304FAF3040AF2F4FAF3040A02F40A0303F9F2F409F30\
3F9F2F3F9F303F902F3F90303F8F2F3F8F303F802F3F8030307F2F307F30306F2F
306F3030602F306030305F2F3050302F4F2F304030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F2F60502F7F602F80702F90802F9F902FA0902FAF9F30B0A030B0AF30BFAF2FB0B030B0B02FAF\
B02FAFAF2F9FB02F9FAF2F90AF2F8F9F2F70902F6F7F2F4F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F2F2F302F3F402F3F4F2F40502F405F2F40602F40602F4F6F2F406F2F4F6F2F
406F2F3030602F306F303F7F304080304F8F2F4F8F30509F30509F3060A0305FA03060A03060A0\
306FA03060A0306FA0306F9F306F9F306F8F306F80306F7F306F6F2F5F3F305F3F
2F5F3F305F3F2F5F3F30603F2F5F3F3060402F603F3060402F6040306F4F2F60402F6F4F2F6F4F\
3070502F6F4F2F70502F70502F705F2F70502F7F5F2F705F2F7F5F2F7F5F2F7F5F
2F7F5F2F80602F7F5F2F80602F80602F806F2F80602F8F602F80602F8F6F2F8F602F8F6F2F8F6F\
2F9070302F30402F504F2F60502F70502F7F5F2F8F5F2F90602F9F3F2F2F5F3030
6F2F2F7F30308F2F2F9F30309F302FAF3030AF3030B03030B03030BF3030B0302FBF3030B0302F\
B03030A02F2F9F30307F2F2F402F6F2F9F702FA07F2F9F70305F3030602F3F7030
3F7F2F3F80303F7F2F3F80303F8F2F4090303F8F2F409030409F2F4FA030409F2F4FA03040A02F\
4FAF304FA02F4FAF304FAF2F4FB0304FAF2F50B0304FB02F4FBF3040B02F4FBF30
4FBF2F4FBF304FBF2F4FC0304FBF2F4FC0304FBF2F4FC0304FC02F50C0304FC02F50C0304FC030\
50CF304FC02F50CF304FC02F4FCF3040C02F4FC03040C02F4FC03040BF2F4FBF30
40BF2F4FBF3040B02F4FBF3040B02F4FB03040B02F4FB03040AF2F4FAF303FA02F3FA0303FA02F\
3FA0303F9F2F3F9F303F902F3F90303F8F2F3F8F303F802F308030307F2F307F2F
30702F3070302F6F2F30602F2F5F2F305F302F4F2F30402F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F604F2F70602F90702F907F2FA08F2FA09030B09F2FB0A030BFAF30BFB030BFAF30B0B030B0\
B02FAFB02FAFB02F9FB02F9FAF2F8FA02F809F2F708F2F606F2F3F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F2F303F2F3F4F2F40502F405F2F4F603040602F4F6F2F4F6F2F4F70304F6F2F
4F702F302F50303F6F303F70304080304080305090305090305F9F305F9F3060A03060A0306FA0\
306FA03070A0306F9F30709F306F9030708F306F7F306F7030605F30604F305F3F
3060402F603F3060402F60403060402F6040306F4F2F6F4F306F4F2F6F4F3070502F6F50307050\
2F7050307F5F2F7F50307F5F2F7F5F3080602F7F5F3080602F7F5F3080602F8060
2F80602F8060308F6F2F80602F8F6F2F8F6F308F6F2F8F6F2F8F6F2F8F6F3090702F8F6F302F30\
3F2F404F305F4F2F60502F70502F805F308F5F2F90602F9F602FA03F30304F2F2F
6F3030702F2F8F303090302F9F3030A03030B03030B03030BF3F30B03030BF3F30B0302FBF3030\
B0302FAF30309F302F8F30302F9F702F9F702F9F702F9F702F9F702F9F702F9F6F
2F9F702F9F6F3050303060303F7030307F2F3F80303F8030408F303F8F2F409030409030409F30\
409F2F4FA03040A0304FA0304FA02F4FAF304FAF304FAF3040AF2F4FB03040B030
4FBF304FB02F4FBF304FBF304FBF304FBF2F4FC0304FBF304FC0304FBF2F50C0304FC03050C030\
4FC02F50CF304FC03050CF3040C02F4FCF3040C0304FC03040C02F4FC03040C030
4FC03040BF2F4FBF3040BF2F4FBF3040B02F4FBF3040B02F4FB03040B02F40B0303FAF2F3FAF30\
3FA02F40A0303F9F2F3F9F303F9F2F3F9F303F902F3F90303F8F2F308F3030802F
308030307F2F307F3030702F306F302F602F305F302F4F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F20303F2F6F502F7F6F2F90702090802FA08F2FAF902FB09F2FB0A030B0AF30BFB030BFAF2FAF\
B02FAFB02FA0B02FA0AF2F90AF2F90A02F809F2F7F9020607F2F50000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F2F303F2F3F4F2F3F502F405F2F40602F4F6F2F4F6F2F4F6F2F4F6F2F4F702F
4F6F2F4F702F2F306F303F7F2F3F7F304F802F4F8F3050902F5090305F9F305F9F30609F2F609F\
306F9F306F9F306F9F2F6F8F306F8F306F80306F7F2F60603060502F5F40306040
2F60403060402F6040306F4F2F60402F6F4F2F6F4F3070502F6F4F2F70502F70503070502F7050\
2F7F5F2F7F5F2F7F5F2F7F5F2F80602F7F5F2F80602F7F602F80602F80602F8060
2F80602F8F6F2F80602F8F6F2F806F2F8F6F2F8F6F2F8F6F2F8F6F2F8F6F2F8F6F3F2F3F402F4F\
4F2F60502F6F5F2F7F502F805F2F8F5F2F90602F9F5F2FA0602FAF2F2F2F4F3030
5F2F2F70302F7F2F2F903030902F2FA03030A03030B03030B03030B03030B0302FBF3030B0302F\
B0302FAF2F2FA03030902F2F70302F2F8F6F2F906F208F6F2F906F2F8F602F906F
208F602F8F6F2F8F602F8F6F208F602F8F6F2F8F6030503030602F3F7030307F2F3F80303F802F\
3F8F303F8F2F4090303F902F409F30409F2F4FA03040A02F40A03040A02F40AF30
40AF2F40AF3040AF2F4FB03040B02F4FB03040B02F4FBF3040BF2F4FBF304FBF2F4FBF304FBF2F\
4FC0304FBF2F4FC03040BF2F40C03040C02F40C03040BF2F40C03040C02F40C030
40BF2F40C03040BF2F40BF3040B02F40BF3040B02F40BF3040B02F3FB0303FAF2F3FB0303FAF2F\
3FAF303FA02F3FA0303FA02F3FA0303F9F2F3F9F3030902F3F902F308F2F308F30
30802F30802F307F2F307F302F6F2F306F2F2F602F305F2F2F4F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000302F5F4F2F7F5F2F8F702F9F7F2F9F8F2FAF902FAF9F30B0A030B0AF30BFAF30BFB030BF\
B02FB0B030AFB02FA0B02FA0AF2F90AF2F90A02F809F2F70902F60702F40000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F302F3F402F3F4F2F4F5F3040602F4F6F2F4F6F2F5070304F702F50702F4F702F
5070304F702F303060303F7030408030408030508F305090305F9F305F9030609F30609F306F9F\
306F9F306F9F306F90307090306F80306F80306F70306F60306050306F4F2F604F
306F4F306F4F306F4F2F6F4F3070502F6F503070502F7050307F5F2F7050307F5F2F705F307F60\
2F7F5F3080602F80603080602F806030806F2F8060308F6F2F8060308F6F2F8F6F
308F6F2F8F6F308F6F2F8F6F3090702F8F702F90702F8F6F309070302F30402F4F4F2F5F50306F\
502F705F2F805F2F8F6030905F2F90602FA0602FA06F30B0602FB06F30BF3F2F2F
5F30306F302F7F30308F302F9F30309F3030AF3030AF3030B03F30B03030BF3F30B0302FBF3F30\
B0302FB03030A0302F9F303080302F2F806F2F80602F806F2F7F5F2F80602F7F5F
2F7F602F7F5F2F7F602F7F5F2F7F602F705F2F7F5F2F70502F7F5F2F70502F705F2F7050305030\
306F303F7F303F7F2F3F80303F8030408F303F8F2F4090303F9F3040A030409F2F
40A03040A03040AF3040AF2F4FAF3040AF304FB03040B02F4FB0304FB0304FBF304FBF2F4FBF30\
4FBF304FC0304FBF2F4FC03040BF3040C03040BF2F4FC03040BF3040C03040BF2F
4FC03040BF304FC03040BF2F4FC03040BF304FBF3040BF2F4FBF303FB03040B0303FB02F40B030\
3FAF2F3FAF303FA02F3FAF303FA02F3FA0303F9F2F3FA0303F9F2F309F30308F2F
308F3030802F308030307F2F307F3030702F306F302F5F2F305F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F6F502F7F602F8F702F9F7F2F9F8F2FAF902FAF9F30B0A02FB0AF30BFAF30B0\
B030B0AF2FAFB02FAFB02FA0B02F9FAF2F90AF2F8F9F2F7F9F2F6F802F50000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F202F\
30302F30402F40502F405F2F4F602F4F602F4F6F2F4F6F2F50702F4F702F50702F
4F702F4F702F2F2F4F303F702F407F304F802F4F8030509030508F305F902F5F90306090306090\
306F902F608F306F8F306F80306F802F6070306F6F2F6060306F5F2F604F306F50
2F6F4F3070502F6F4F3070502F6F5030705F2F7050307F5F2F705F307F5F2F705F2F7F602F7F5F\
3080602F80602F80602F80602F8F6F2F80602F8F6F2F806F2F8F6F2F80602F8F6F
2F8F6F2F8F702F8F6F2F8F702F8F6F2F8F702F8F6F2F8F6F3F2F3F4F2F504F2F5F502F6F502F7F\
602F805F2F8F602F90602F9F6F2FA0602FA0602FAF602FB06F30BF602FBF302F30
4F2F2F6F3030702F2F803030902F2F9F3030A02F2FAF3030AF3030B03030B0302FBF3030B0302F\
B03030AF2F2FAF30309F2F2F8F30302060502F60502F5F4F2F5F4F2050402F5040
2F4F402F4040203F302F3F30000000000000000000000000000000000000000000000000000000\
00000000000000000000000030503030602F307F30307F2F3F803030802F3F8F30
3F8F2F4090303F902F40A0303F9F2F40A03040A02F40AF3040A02F40AF3040AF2F40B03040B02F\
4FB03040B02F4FBF3040BF2F40BF3040B02F40BF3040BF2F40BF3040B02F40BF30
40BF2F40BF3040B02F40BF3040BF2F40BF3040B02F40BF3040B02F3FBF303FB02F3FB0303FAF2F\
3FB0303FAF2F3FAF303FA02F3FA0303FA02F3FA0303F9F2F3F9F3030902F309030
308F2F308F3030802F30802F307F2F307F302F6F2F30602F2F5F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F4F402F705F2F8F6F2F907F2FA0802FA0902FAF902FAFA030B0A030B0AF30BF\
AF30B0B030B0B02FAFB02FAFB02F9FB02F9FAF2F90AF2F809F2F70902F6F702F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F2F\
2F302F3F4F2F40502F4F602F4F602F506F2F50702F50702F4F702F507F3050702F
507F2F4F702F50702F303060303F7030407F30508F30508F305F90305F8F306090306090306090\
30608F306F90306F8F306F80306070307070306F6F306F602F605F30705F306F50
306F5F2F705030705F2F705F30705F2F705F307F5F2F7F5F307F602F7F5F3080602F7F60308060\
2F8060308F6F2F8060308F6F2F806F308F702F806F308F702F806F308F6F2F8F6F
2F8F702F8F6F308F702F8F6F2F8F702F8F6F302F30402F4F4F2F5F502F605F307F5F2F80602F8F\
602F8F6F309F602F9F6F2FA06F2FAF6F30B0602FB06F30BF602FBF6F30C02F2F2F
4030305F2F2F70303080302F90303090302FA03030AF3030B03030B03030BF3F30BF302FBF3030\
B0302FB03030AF302FA03030902F2F7F3030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000305F30306F30
3F7F30307F2F3F80303F80303F8F303F902F409F303F9F3040A030409F2F40A03040A03040AF30\
40AF2F4FAF303FAF3040B0303FB02F40B03040B02F40BF3040B02F40BF3040B02F
40BF3040B02F40BF3040B02F40BF3040B02F40BF3040B02F3FBF303FB02F3FBF303FB02F3FB030\
3FAF2F3FB0303FAF2F3FAF303FA02F3FAF303FA02F3FA030309F2F309F3030902F
309030308F2F308F3030802F3080302F7F2F3070302F6F2F306030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000302F6F4F207F602F8F6F2F907F2FA08020A0902FAF9F2FAFA030B0A030B0\
AF30BFAF2FB0B030B0B02FA0B02FA0AF2F9FAF2F90A0208FA02F7F902F6F8F2F60
702040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F2F\
303F2F3F4F2F405F2F405F2F4F6F2F4F6F2F50702F50702F507F2F50702F507F2F
50702F50702F506F2F3030502F30603040702F4F7F30508F2F5080305F8F2F5F8F3060902F5F8F\
30608F2F6080306F802F6F7F306F7F2F6F6F306F6F2F6060306F602F6F5F306F5F
2F6F5030705F2F705F30705F2F705F2F705F2F7F5F307F602F7F5F2F7F602F7F60307F602F7F60\
2F806F2F80602F806F2F806F2F806F2F806F2F8F702F806F2F806F2F806F2F8F6F
2F806F2F8F6F2F806F2F8F702F806F3F2F3F402F4F502F60502F6F5F2F7F5F2F80602F8F602F90\
602F9F602F9F6F2FAF6F2FAF6F2FB06F2FB06F2FBF602FBF6F2FC0602FBF6F2FC0
302F2F50302F6F2F2F7F3030802F2F9030309F2F2FA03030AF3030B03030B0302FBF3030B0302F\
B0302FB0302FAF30309F2F2F9F302F7F2F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000305F30306F2F307F30307F2F3F803030802F3F8F303F902F409F303F9F2F\
409F303F9F2F3FA0303FA02F3FA0303FA02F40AF303FAF2F40AF303FAF2F40B030
3FAF2F40B0303FB02F40B0303FB02F40B0303FB02F40B03030AF2F3FB0303FAF2F3FB03030AF2F\
3FAF303FAF2F3FAF3030A02F3FAF303FA02F3FA030309F2F309F2F309F2F309F30
2F8F2F308F2F308F2F308F302F7F2F307F2F2F702F306F2F2F5F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F3F402F70502F806F2F90702F9F802FAF8F2FAF9F30B09F2FB0AF30BF\
AF30BFB030BFAF2FB0B030B0B02FAFB02FA0AF2F9FB02F90AF2F80A02F80902F70
902F60702F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F302F\
303F2F40502F405F2F4F60304F6F2F50702F50702F5F7F30507F2F5F7F2F507F2F
5F7F30507F2F5F702F4F602F303060303F6F304F7F304F80305F8F305F8030608F30608F306F8F\
306080306F80306F80306F80306F70306F70306F6F30706F306F603070602F6F5F
3070602F7060307F602F7060307F602F7060307F602F7F6030806F2F7F6030806F2F7F6F30806F\
2F806F308F6F2F806F308F702F806F308F702F806F308F702F806F30806F2F806F
3080702F806F2F8070302F304F304F4F2F5F502F6F502F70603080602F8F602F90602F906F30A0\
6F2FA06F2FAF6F2FAF7030BF702FBF7030C0702FC06F30C06F2FC06F30C0602FBF
3030304F2F2F6F3030702F2F8F303090302F9F3030A03030AF3030AF3030B03030B03030BF3F30\
B0302FB03030AF302FAF30309F302F8F3030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003060303070303F7F30307F2F3F8030\
3F8F303F90303F902F3F9F303F9F303FA0303F9F2F40A0303FA03040A0303FA02F
40AF303FAF3040AF303FAF2F40B03040B03040B0303FAF2F3FB0303FAF2F3FB0303FAF2F3FAF30\
3FAF2F3FAF303FA02F3FAF303FA02F3FA0303FA02F3FA03030A02F309F3030902F
309F3030902F308F3030802F308030307F2F3070302F6F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F5F4F2F7F5F2F8F6F2F9F7F2F9F8F2FAF8F2FAF9F30B09F2FB0\
AF30BFAF30B0B030B0AF2FAFB02FAFB02FA0B02F9FAF2F9FAF2F90A02F8FA02F7F
902F6F8F2F606F2F3F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F2F\
30402F3F4F2F405F2F4F602F506F2F50702F5F7F2F507F2F5F7F2F5F7F2F5F7F2F
5F7F2F5F7F2F50702F50602F000000303F602F3F6F304F7F2F4F7F3050802F5080305F802F6080\
3060802F607F306F7F2F607F306F7F2F606F306F6F2F6F6F30706F2F6F60307060
2F70603070602F70603070602F7060307F6F2F7F602F7F6F2F7F60307F6F2F7F602F7F6F2F7F6F\
30806F2F7F6F2F806F2F7F6F2F80702F7F6F2F806F2F7F6F2F7F6F2F7F602F7F6F
2F7F602F7F6F3F2F3F4F2F504F2F5F5F2F705F2F7F5F2F80602F8F602F90602F9F6F2FA06F2FA0\
6F2FAF6F2FAF702FBF6F2FBF7030C06F2FC0702FC0602FBF6F30BF602FBF602FBF
602FB04030305F2F2F7030307F2F2F8F3030902F2FA03030A03030AF3030AF302FB03030B0302F\
B03030B0302FB03030A02F2FA030308F302F70302F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F3F2F\
306F30306F2F307F30307F2F3F803030802F3F8F3030902F3F9F303F902F3F9F30
3F9F2F3FA0303F9F2F3FA0303FA02F3FA03030A02F3FAF3030A02F3FAF3030A02F3FAF3030A02F\
3FAF3030A02F3FA03030A02F3FA030309F2F3FA030309F2F309F3030902F309030
308F2F3090302F8F2F308F2F30802F307F302F6F2F306F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F2F6F502F80602F90702F9F7F2FA0902FAF902FAF9F30B0\
A030B0AF3FBFAF30B0B030B0B02FAFBF2FAFB02FAFB02FA0AF2F9FAF2F9FA02F8F
A02F7F902F6F8F2F5F702F3F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F2F\
30402F40502F405F2F506F2F506F2F5F7F2F5F7F2F5F802F5F7F2F60802F5F802F
5F802F5F7F2F5F702F50602F000000000000303F6F304070304F7F2F4F7F305F80305F7F306080\
2F607F30608030607F306F7F2F6070307070306F6F30706F2F6F6F30706F2F7060
30706F2F7060307F6F2F706030706F2F7F60307F6F2F7F6F307F6F2F7F6F307F6F2F7F6F30806F\
2F7F6F3080702F7F6F307F6F2F7F6F307F6F2F70602F7F6F2F706030706F2F6F60
3F2F3F402F40502F5F502F6F5F2F7F5F2F7F60308F602F906F2F9F6F2F9F6F30AF6F2FAF702FB0\
6F2FB07030BF702FBF7030C0702FC07030C06F2FBF7030BF602FBF6F30BF602FB0
602FB0302F2F5030306F2F2F7F3030802F2F9030309F3030A03030AF3030B03030B03030BF3F30\
B0302FB03030B0302FAF30309F302F9F3F3070302F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F302F306F3030702F307F30307F2F3F803030802F
3F8F3030902F3F9F303F902F3F9F30309F2F3F9F30309F2F3F9F30309F2F3FA030309F2F3FA030\
309F2F3FA03030A02F3FA030309F2F3FA030309F2F309F3030902F309030308F2F
308F30308F2F308F3030802F3080302F702F3070302F602F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F50402F705F2F8F6F20907F2F9F802FA0902FAF902FAF\
A030B0A030B0AF30BFAF2FB0B030B0B02FB0B02FB0AF2FA0B02FA0AF2F9FAF2F90
A020809F2F7F902F6F8F2F5F6F2030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000202020\
30402F3F502F4F602F4F602F50702F50702F5F7F2F5F7F2F5F802F5F802F60802F
5F7F2F5F7F2F5F702F5F6F2F0000000000000000000000002F406F304F702F5070305F7F2F5070\
305F7F2F5F7F30607F2F6070306F702F606F306F702F6F6F306F6F2F6F6F30706F
2F6F6030706F2F706F30706F2F70602F706F2F706030706F2F70602F706F2F706F2F7F6F2F7060\
2F706F2F70602F706F2F6F602F6F602F605F2F605F2F5F502F5F503F2F304F2F4F
4F2F50502F60502F6F5F2F7F5F2F80602F8F602F906F2F9F602F9F6F2FAF6F2FAF702FB06F2FB0\
702FBF6F2FBF7030C06F2FBF702FC06F2FBF6F30BF6F2FB06F2FB05F2FAF602FAF
5F2FA0302F2F402F2F6030306F2F2F80302F8F2F2F9F30309F2F2FAF3030AF302FB03030B0302F\
B03030B0302FB03F30AF302FA03F2F90302F7F3030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000306F30
2F6F2F307030307F2F30803030802F308F30308F2F309030308F2F309030308F2F30903030902F\
30903030902F3F903030902F30902F30902F309F302F902F30902F308F2F308F30
2F802F30802F2F802F3080302F702F306F2F2F602F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F2F6F4F2F7F6F2F90702F9F802FA08F2FAF9030B0\
9F2FB0A030BFAF30BFB03FC0B030BFB030BFB02FB0BF2FB0B02FAFB02FA0AF2F9F
B02F9FA02F8FA02F7F902F6F8F2F5F6F2F3F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
303F2F40502F4F602F506F2F50702F5F7F2F5F7F2F60803060802F60802F60802F
6F802F607F2F607F2F5F6F2F00000000000000000000000000000000000030406F2F5070305F7F\
305F7030607F2F607030607F306070306F702F606F306F70306F6F30706F2F6F6F
30706F2F6F6F306F6F2F6F6F30706F2F706F30706F2F6F6030706F2F6F60306F6F2F6F60306F60\
2F60603060602F5F5F0000000000000000000000003F303F4F2F40502F5F502F60
603070602F7F602F80602F8F6F3090602F9F6F2FA06F2FA07030AF6F2FAF7030B0702FBF7030C0\
702FC07030C06F2FC07030C06F2FBF7030BF6F2FB06F30BF602FAF6030AF5F2FA0
5F30A0502F904030305F2F2F7030307F302F8F303090302FA03030A03030AF3030AF3030B04030\
B03F2FB03F30B03F2FB03F30A03F2F9F3F3080302F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000306F30306F2F30703030702F307F30307F2F30803030802F3F8F30\
308F2F3F8F30308F2F3F8F30308F2F308F30308F2F308F3030802F308F3030802F
30803030802F3080302F702F3070302F6F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F40402F70502F806F2F90702F9F802FA08F2FAF\
9F30B09F2FB0A030C0AF30C0B030C0AF2FBFB030BFB02FB0B02FB0B02FA0B02FA0
AF2F9FAF2F90A02F809F2F7F902F6F802F506F2F30000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
303F2F3F4F2F4F602F506F2F5F702F5F7F2F60802F60802F6F802F60802F6F8F2F
60802F6F802F60702F5F6F2F0000000000000000000000000000000000000000000000002F4060\
304F6F2F4F6F305F6F2F5F6F305F702F5F6F30606F2F606F30606F2F606030606F
2F6060306F6F2F60602F6F602F60603060602F5F5F2F5F602F5050000000000000000000000000\
000000000000000000000000000000302F304F2F404F2F50502F605F2F6F602F7F
602F7F6F2F8F602F8F602F90602F9F6F2FA06F2FA06F2FAF6F2FB0702FBF6F2FBF702FC06F2FBF\
7030C06F2FBF7030C06F2FBF6F30BF6F2FB06F2FB0602FAF602FAF5F2FA05F2F9F
4F2F8F4F2F80302F2F503030602F2F7F3030802F2F9030309F302FA03F30A03F2FAF3F30AF3F2F\
B03F30B03F2FB03F30AF3F2FAF3F309F302F8F302F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000304F302F602F306F30306F2F\
30703030702F307F302F7F2F3080302F7F2F307F302F7F2F307F302F702F307F2F
2F702F3070302F6F2F3060302F5F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F604F2F7F5F2F8F702F9F7F2FA08F2FAF\
902FB09F30BF9F30BFAF30C0AF30C0B030C0B030BFB030BFB02FB0BF2FB0B02FAF
B02FA0AF2F9FAF2F90A02F809F2F7F902F6F802F506F2F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F3F2F40502F4F602F506F2F5F702F60802F60802F6F8F2F6F802F6F8F2F6F8F2F
6F8F2F6F802F6F7F2F5F6F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000304F602F50603050602F505F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F2F30402F40502F50502F605F2F705F2F70602F80602F80
6F2F906F2F9070309F702FA06F2FAF6F2FAF7030B0702FB07030BF702FBF7030C0702FC07030C0\
6F2FBF7030C06F2FBF6F30BF6F2FB06F30B0602FAF602FAF5F2F9F5F3090502F80
4F2F7F302F4F3030304F302F60303070302F803F308F302F9F3F309F3F2FAF3F30AF3F30B03F30\
B03F2FB04030B03F2FB03F30A0302FA03F308F302F703030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000305F302F5F2F30603030602F306F302F602F306F302F6F2F
306F302F5F2F305F30000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F3F3F206F502F80602F90702FA07F20A0\
8F2FB08F2FB09F30BF9F2FBFA030C0AF30C0B030C0AF2FBFB030BFB02FB0B02FB0
AF2FA0B02FA0AF2F90AF2F90A020809F2F70902F607F2F506F203F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F3F2F3F4F2F4F602F506F2F5F702F5F7F2F60802F60802F6F8F2F6F8F2F708F2F
6F802F6F802F607F2F606F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F2F402F404F2F4F502F605F2F6F602F70602F7F602F80602F8F6F2F90
6F2F90702FA0702FA0702FAF6F2FAF702FB06F2FB0702FBF6F2FBF702FBF6F2FBF7030C06F2FBF\
6F30BF6F2FB06F2FB0602FAF602FAF602FA0602FA0502F90502F8F4F2F7F402F60
0000000000002F2F2F4030305F302F70302F7F302F8F3F3090302F9F3F30A03F2FAF3F30AF3F2F\
B03F30B03F2FB03F30AF302FAF3F2F9F302F903F3070302F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F5F402F7F5F2F906F2F9F7F2FA0\
7F2FAF8F2FB0902FB09F30BFA030BFAF30C0AF30C0B030C0B030BFB030BFB02FB0
BF2FB0B02FA0B02FA0AF2F90B02F90A02F809F2F7F8F2F6F8F2F60702F406F2F30000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F3F2F40502F4F5F2F5F702F5F702F60802F60802F6F8F2F6F8F2F708F2F6F8F2F
708F2F6F802F6F802F60702F5F5F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F30\
402F3F502F50502F5F5F30705F2F70602F80602F806F30906F2F90702F9F702F9F
7030AF702FAF7F30AF702FAF7F30BF702FBF7030BF702FBF7030C07030C07030C06F2FBF7030C0\
6F2FBF6F30B0602FAF6F30AF602FA05F2FA0502F9050308F4F2F703F2F50000000
0000000000000000003F2F2F5F30306F302F7F3F308F302F903F309F3F30AF3F30AF3F30B03F30\
B03F30B04030B03F2FB03F30AF3F2FA03F3090302F803F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F60402F7F5F2F90602F9F\
702FA07F2FAF8F2FB0902FB09F30BF9F2FBFAF30C0AF30C0B030C0AF2FBFB030BF
B02FB0B02FB0AF2FA0B02FA0AF2F90AF2F909F2F809F2F7F8F2F6F8F2F607F2F4F6F2F3F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F3F2F3F4F2F4F602F506F2F5F7F2F607F2F6F802F6F8F2F708F2F6F8F2F70902F
708F2F708F2F6F7F2F6F702F5F5F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F2F2F402F3F4F2F4F\
5F2F605F2F60602F70602F7F6F2F8F602F8F6F2F906F2F90702FA0702FA0702FAF
702FAF7F2FB0702FB07F2FBF7F2FBF7030BF6F2FBF7030C06F2FBF7030C06F2FBF6F30BF6F2FB0\
6F2FB0602FAF602FAF5F2F9F5F2F9F502F8F4F2F803F2F60000000000000000000
0000000000000000003F302F50302F6F303070302F8F3F308F302F9F3F309F302FAF3F30AF3F2F\
B03F30AF3F2FB03F30AF3F2FAF3F30A0302F9F3F2F80302F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000302F60402F7F5F2F90\
6F2F9F7F2FAF7F2FAF8F2FB0902FB09F30BFA030BFAF30C0AF30C0B030C0B030BF
B030BFB02FB0BF2FB0B02FA0B02FA0A02F90AF2F909F2F8F9F2F80902F70902F6F802F5F7F2F4F\
602F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2F2F404F2F4F5F2F506F2F5F702F60802F6F802F708F2F6F8F2F70902F70902F
70902F7F8F2F7F802F6F7F2F60602F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F2F402F3F4F2F40502F505F2F5F5F2F6F\
5F2F70602F7F602F806F2F8F6F2F90702F9F702F9F702FA0702FA07F30AF702FAF
7F2FB07F2FB07F30BF7F2FBF7F30BF702FBF7030C06F30C07030C06F2FBF6F30BF6F2FB06F2FB0\
602FAF6030AF5F2F9F5F2F904F2F804F307F000000000000000000000000000000
000000000000000000302F2F4F303060302F703F3080302F903F3090302FA03F30A03F30AF3F30\
AF3F2FB04030B03F2FB03F30AF3F2FAF3F309F302F8F3F3060302F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F6040207F\
5F2F906F2F9F702FA07F20AF8F2FB0902FB09F30BF9F2FBFAF30C0AF30C0B030C0
AF2FB0B02FB0B02FB0B02FAFAF2FA0AF2FA0A02F9FA02F909F2F8F9F2F80902F7F902F708F2060\
7F2F506F2F3F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030402F4F5F2F506F2F5F7F2F607F2F6F8F2F6F8F2F708F2F708F2F7F902F
7F8F2F7F8F2F70802F707F2F606020000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000402F304F2F3F502F4F502F505F2F605F2F6F602F705F2F70\
602F806F2F8F6F2F906F2F90702F9F702F9F702FAF702FAF7F2FB0702FB07F2FBF
702FBF7F30BF702FBF7F30C0702FBF7F30C0702FBF6F30BF6F2FB06F2FB0602FAF602FAF5F2FA0\
5F2FA0502F90502F8F4F2F7F3F2F60000000000000000000000000000000000000
000000000000000000302F30402F2F60302F6F302F803F308F302F903F309F302FA03F30A0302F\
AF3F30AF3F2FB03F30AF3F2FAF3F2FA0302F9F3F308F302F70302F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F6F\
402F805F2F906F2F9F7F2FAF802FAF8F2FB0902FB09F30BFA030BFAF30C0AF30C0
B030C0B030BFB030BFB02FB0B02FB0AF2FAFAF2FA0A02F9FAF2F9FA02F90A02F8F9F2F7F9F2F7F\
8F2F6F8F2F60702F406F2F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030402F4F5F2F5F6F2F60702F6F802F6F802F7F902F7F902F7F902F7F902F
80902F7F902F7F8F2F70802F70702F3F302F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000402F30402F3F50304F502F50602F60602F6F6F3070602F706F2F806F2F806F308F\
6F2F90702F9F702F9F7030A0702FA07F2FAF702FAF7F30B07F2FB07F30BF7F30BF
7F30C07F30C07F30C07030C07F30C07030BF7030BF6F2FB06F30BF602FB0602FAF5F2FA05F309F\
502F90502F8F402F70000000000000000000000000000000000000000000000000
0000000000000000000000004030305F302F6F30307F302F8F3F30903F2F9F3F30A03F30AF3F30\
AF3F30B03F30AF3F2FB04030AF3F2FAF3F309F302F903F307F302F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F6F402F7F5F2F906F2F9F7F2FAF7F2FAF8F2FB0902FB09F30BF9F2FBFAF30C0
AF30C0B030C0AF2FB0AF30BFAF2FB0B02FB0AF2FAFAF2FA0A02F9FAF2F9FA02F90A02F909F2F80\
9F2F7F8F2F6F8F2F607F2F506F2F40602F303F2F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F302F40502F50602F60702F607F2F70802F708F2F7F902F7F902F80902F
7F902F80902F7F8F2F7F802F70702F60602F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
402F30502F40502F4F5F2F5F5F2F60602F70602F706F2F7F6F2F80702F8F702F8F702F90702F90\
702F9F702F9F702FA0702FAF7F2FB0702FB07F2FBF7F2FBF7F30C07F2FBF7F30C0
7030C07F30C0702FBF7030C0702FBF7030BF6F2FB06F2FB0602FAF602FAF5F2F9F5F2F90502F8F\
4F2F7F3F2F5F000000000000000000000000000000000000000000000000000000
000000000000000000000000302F2F50303060302F7F303080302F903F3090302FA03F30A03F2F\
AF3F30AF3F2FB03F30AF3F2FAF3F30A0302FA03F3090302F803030000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F704F2F80602F906F2F9F7F2FAF802FAF902FBF902FB09F30BFA030BF
AF30C0AF30C0AF30C0AF30BFAF30BFAF2FB0B030B0AF2FAFAF2FAFAF2FA0AF2FA0A02F9FA02F90\
9F2F8F9F2F80902F70902F6F802F607F2F5F702F4F6F2F405F2F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F302F3F4F2F50602F5F6F2F6F802F6F802F7F8F2F7F902F809F2F80902F
809F2F80902F80902F7F802F7F802F60602F000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F2F304F2F40502F4F
5F2F5F5F2F60602F6F602F706F2F7F6F2F7F702F8F702F8F702F90702F907F2F9F7F2F9F7F2FA0\
702FA07F30AF7F2FAF7F30B07F2FB07F30BF7F30BF7F30C07F30C07F30C07F30C0
7F30C0702FBF7030C0702FBF7030BF6F2FB06F30B06F2FAF602FA05F2F9F5F30904F2F80402F70\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003030304F302F6F303070302F803F308F302F9F3F309F3F30AF3F30\
A03F2FB03F30AF3F2FB03F30AF3F2FAF3F309F3F2F9F3F3080302F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000302F7040207F5F2F906F2F9F7F2FAF7F2FAF8F2FB0902FB09F30BF
9F2FBFA030C0A030C0AF30C0A02FBFAF30BFAF2FB0AF2FB0AF2FAFAF2FAFAF2FA0AF2FA0A02F90\
A02F90A02F8F9F2F80902F708F2F70802F6F802F6F7F2F5F702F506F2F40602F3F
4F202F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F402F4F5F205F6F2F607F2F70802F708F2F7F902F7F902F809F2F
80902080902F80902F808F2F707F20706F2F403F2F000000000000000000000000000000000000\
0000000000000000003F2F2F402F304F2F3F502F40502F4F5F2F505F2F5F602F6F
602F6F6F2F7F6F2F7F702F806F2F80702F90702F907F2F9F702F9F7F2FA07F2FA07F2FAF7F2FA0\
802FAF702FAF7F30B0702FB07F30BF7F30BF7F30C07030C07F30C07030C07030C0
702FBF702FBF6F2FB06F2FB06F2FAF6F2FAF602FA0602F9F502F90502F8F402F703F2F60000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F2F4F302F5F302F7030307F302F8F3F2F90302F9F3F30A0302F\
AF3F30AF3F2FB03F30AF3F2FAF3F2FA0302FA03F3090302F8F302F6F302F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F2F704F2F80602F9F6F2F9F7F2FAF802FAF902FBF902FBF
9F30BF9F30BFA030C0A030C0AF3FC0AF30BFAF30BFAF2FB0B030BFAF2FB0B02FB0AF2FAFAF2FA0\
AF2F9FAF2F9FA02F909F2F8F902F80902F808F2F7F8F2F708F2F6F8F2F6F7F2F5F
7F2F506F2F40602F3F402F2F000000000000000000000000000000000000000000000000000000\
000000000000303F2F4F5F2F5F6F2F607F2F6F802F7F8F2F7F902F809F2F80902F
8F9F2F809F2F8F9F2F808F2F808F2F7F702F6F602F000000000000000000000000000000402F30\
4F2F3F502F40502F405F2F505F2F50602F5F602F606F306F602F6F6F2F706F2F7F
703080702F8F702F90702F907F309F7F2F9F7F2FA07F2FA08030AF7F2FAF802FB0802FAF8030B0\
802FB07F30BF7F30BF7F30C07F30C07F3FCF7F30C07F30C07030C07030C0702FBF
7030BF6F2FB06F2FB0602FAF6F30AF5F2F9F5F2F9F502F8F503080402F6F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000040302F6030306F302F803F308F302F903F309F3F2FA03F30\
A03F30AF3F30AF3F2FB04030AF3F2FAF3F30A03F2F9F3F308F302F703030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F2F704F2F80602F9F6F2F9F7F2FAF802FAF8F2FB0
8F2FB09030BF902FBF9F30C09F30BFA03FC0A030BFAF30BFAF2FBFAF30BFAF2FB0B02FB0AF2FAF\
AF2FAFA02F9FAF2F9F9F2F909F2F909F2F8F9F2F8F902F80902F7F8F2F708F2F70
802F6F802F60702F50702F4F602F3F602F3F5F2F30502F304F2F2F402F30000000000000000000\
0000000000002F302F40502F506F2F5F702F6F802F708F2F7F902F7F902F809F2F
809F2F8F9F2F80902F8F902F808F2F80802F6F602F3F2F30402F304F2F3F502F405F2F4F5F2F4F\
602F5F602F5F602F60602F606F2F6F6F2F6F702F7F702F7F702F806F2F80702F8F
702F8F7F2F90702F907F2F9F7F2F9F7F2FA07F2FA0802FAF7F2FAF8030B0802FB08030BF802FBF\
8030BF7F30BF7F30C07F30C07F30CF7030C07030C0702FBF7030BF6F2FB06F2FB0
6F2FAF6F2FAF602FA0602F9F5F2F905F2F904F2F7F402F6F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F3030502F2F6F303070302F8F3F308F302F9F3F309F302F\
A03F30A03F2FAF3F30AF3F2FAF3F30A0302FA03F309F302F903F3070302F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F2F704F2F80602F9F6F2F9F702FAF7F2FAF
802FB08F2FB09030BF902FBF9F30C09F30BFA03FC0A030C0AF30C0AF30BFAF30BFAF2FB0B02FB0\
AF2FAFB02FAFA02FA0A02FA09F2F9FA02F9F9F2F909F2F909F2F8F9F2F8F902F80
902F7F8F2F708F2F70802F6F7F2F60702F5F702F5F702F50702F506F2F4F6F2F4F602F40602F40\
5F2F3F5F2F3F4F2F30402F303F2F2F402F30402F304F2F304F2F30502F3F502F3F
502F3F502F3F5F2F405F2F405F2F405F2F40602F4F5F2F4F602F50602F506F2F606F2F606F2F6F\
6F2F6F702F7F702F7F702F80702F807F2F8F7F2F8F7F2F8F7F2F8F7F2F90702F90
7F2F9F7F2F9F802FA07F2FA0802FAF7F2FAF8030B0802FB08030BF802FBF8030C08030BF8030C0\
8030C0803FCF7F30C07F30C07030BF7030C0702FBF7030BF6F2FB06F30B0602FAF
602FAF5F2F9F5F309F502F90502F80402F6F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F2F50303060302F7F3F3080302F903F30903F2FA03F30\
A03F2FAF3F30AF3F2FAF3F30AF3F2FAF3F30A03F2F9F3F3090302F803F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F2F704020805F2F90602F9F702FA0
702FA07F2FB0802FB08F2FBF8F2FB09030BF9F30BFA030C0A030C0A030C0A030BFAF30BFAF2FB0\
AF2FB0AF2FAFA02FAFA02FA0A02FA0A02FA0A02FA09F2F9F9F2F9F9F2F909F2F90
902F8F902F808F2F7F902F7F802F70802F70802F6F802F707F2F6F7F2F6F7F2F607F2F60702F5F\
702F5F6F2F506F2F50602F4F602F4F602F4F602F50602F4F6F2F50602F506F2F5F
602F506F2F5F6F2F5F6F2F5F6F2F5F6F2F606F2F606F2F606F2F6F702F70702F70702F7F702F7F\
7F2F80702F807F2F8F7F2F8F7F2F907F2F90802F907F2F90802F9F7F2F9F7F2FA0
7F2FA0802FAF7F2FAF8030B0802FB08030B0802FB08030BF802FBF8030C08030C08030CF7F30C0\
8030C07F30C07F30C0702FBF7030BF6F2FB06F2FB0602FAF6F2FAF602FA0602FA0
5F2F90502F904F2F80402F70000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F2F4F2F2F6030306F302F80302F8F302F903F309F302F\
A03F30A03F2FAF3F30AF3F2FAF3F2FA03F2FA03F309F302F903F2F7F302F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F70402F7F502F90602F9F
6F2FA0702FA07F2FAF802FB08F30BF8F2FBF9030BF9030BFA030C0A030C0AF3FC0A030BFAF30C0\
AF30BFA030BFA02FB0AF30B0A02FB0AF2FAFA02FAFA030AFA02FA0A02FA09F2F9F
A02F9F9F2F909F2F90902F8F902F8F8F2F808F2F808F2F7F8F2F80802F7F8F2F7F802F70802F7F\
802F70802F707F2F6F802F6F702F6F7F2F6F702F6F7F2F707F2F6F7F2F707F2F70
7F2F7F7F2F707F2F7F7F2F707F2F7F7F2F7F7F2F7F702F7F7F30807F2F807F2F8F7F2F8F803090\
7F2F90802F907F2F9080309F802F9F802FA0802FA08F30AF802FA08F2FAF7F2FAF
8030B0802FB08030BF8030B08F30BF8030BF8F30C08030C08F3FCF8030C08030CF7F30C08030C0\
7F30BF7F30BF7F2FBF7030BF6F2FB06F2FB0602FAF6F30AF602FA05F2F9F502F90
50308F402F70000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F4F30305F302F703F307F302F8F3F3090302F9F3F30\
A03F30A03F30A03F2FAF3F30AF3F2FAF3F30A03F2FA03F3090302F803F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F40302F704F2F80
502F90602F9F6F2FA0702FAF7F2FAF8F2FB08F2FB09030BF902FBF9F30C09F30BFA030C0A030BF\
AF30BF9F2FBFA030BFA02FB0A030BFA02FB0AF2FB0A02FAFAF2FAFA02FA0A02FA0
A02FA0A02FA09F2F9F9F2F9F902F90902F90902F90902F908F2F8F902F8F8F2F808F2F8F8F2F80\
8F2F80802F7F8F2F80802F7F802F7F7F2F7F802F7F7F2F7F802F807F2F7F802F80
7F2F80802F807F2F80802F80802F80802F8F7F2F807F2F8F7F2F8F802F907F2F90802F9F802F9F\
802F9F802F9F802FA0802FA08F2FAF802FA08F2FAF8F2FAF8F30B08F2FAF8030B0
802FB08F30BF8030BF8F30C08030C08F3FC08030C08030C07F30C08030C07F2FBF7F30BF702FBF\
7F30BF702FB0702FB06F2FAF6F2FAF602FA0602FA05F2F9F5F2F904F2F80402F70
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F2F2F5F302F6F302F7F3F3080302F903F3090302F\
A03F30A0302FAF3F30AF3F2FAF3F30A0302FA03F309F302F9F4030803F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F60
402F804F2F8F602F9F6F2FA0702FAF7F2FAF802FB08F2FB09030BF902FBF9F30C09F30C0A030C0\
9F30C0A03FC0A030BFA030BFA030BFAF30BFA02FB0AF30BFA02FB0AF30B0A02FAF
AF2FAFA02FA0A02FAFA02FA09F2FA0902F9F9F2FA0902F9F9F2F9F902F9F9F2F9F902F90902F9F\
902F90902F908F2F8F902F90802F8F8F2F90802F8F8F2F90802F8F8F2F90802F90
8F2F90802F908F2F9F8F2F908F2F9F802F908F2F9F802F9F802F9F802F9F8F2FA0802FA08F2FAF\
802FA08F30AF8F2FAF8F30B08F2FAF9030B08F2FB09030BF8F30B09030BF8030BF
8F30C08030C08F3FC08030C08030C08030C08030C07F30C07F30C07F2FBF7F30BF702FB0702FB0\
702FB07030B06F2FAF6F2FAF5F2F9F602F9F502F90502F80402F70302F4F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F303050302F6F303070302F803F308F302F9F3F30\
9F3F2FA03F30A03F2FAF3F30AF3F2FAF3F30A0402FA04030903F2F8F3F306F302F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F2F70402F805F2F9060209F6F2FA0702FA07F2FB0802FB08F2FBF8F2FB09030BF902FBF\
9030C09030C09F30C09F30BFA030C09F30BFA030BFA02FB0A030BFA02FB0AF30B0
A02FAFAF2FB0A02FAFAF2FAF9F2FA09F2FAF9F2FA09F2FA0902FA09F2FA0902F9F9F2FA0902F9F\
9F2F9F902F9F902F9F902F9F8F2F9F8F2F9F8F2F9F8F2F9F8F2F9F8F2F9F902FA0
8F2F9F8F2FA08F2F9F902FA08F2F9F902FA08F2FA0902FA0802FA08F2FAF8F2FA08F2FAF8F2FAF\
8F2FB08F2FAF9030B08F2FB09030BF8F2FB09030BF8F30BF9030C09030C08F30C0
8030BF8030C08030BF8030C07F2FBF7F30BF7F2FBF7F30BF702FB0702FB0702FB0702FB06F2FAF\
6F2FAF602FA0602FA05F2F905F2F904F2F80402F7F2F2F4F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F2F5030305F302F703F2F7F302F8F3F3090302F\
9F3F309F302FA04030A0402FAF402FA0402FA040309F3F2F903F2F803F2F6F3F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F5F3F2F7F502F8F5F2F9F6F2FA06F2FA07F2FAF802FB08F2FBF8F2FB08F30BF\
8F30BF9030C09030BF9F3FC09F30C0A03FC09F30C0A030C0A030BFA030BFA030BF
AF30BFA02FB0AF30B0A02FB0A030B0A02FB0A030B09F2FAFA030B09F2FAFA030AF9F2FAFA030AF\
9F2FAF9F30AF9F2FA09F30AF902FA0902FA0902FA09030AF902FA09030AF902FA0
9F30AF902FAF9030AF902FAF9F30AF902FAF9F30AF902FAF9030B08F2FB09030B0902FB09030BF\
9030BF9030BF9030BF9030C09030C09030C08F30C0903FC08F30C08F30C08F30BF
8030C07F30BF8030C07F2FBF7F30BF702FBF7F30BF702FB07030B06F2FAF6F2FAF602FA06F30A0\
602F9F602F9F502F8F5030804F2F70302F50000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F30304F302F603F30703F2F803F308F3F2F904030\
9F402FA04030A0402FAF4030AF402FAF4030A0402FA04030903F2F8F3F306F302F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F704F2F80502F8F602F9F6F2FA0702FAF7F2FAF802FB0802FB0\
8F2FBF802FB08F30BF8F2FBF9030C09030BF9F30C09F30C09F30C09F30BFA030C0
9F30BFA030BFA02FB09F30BF9F2FB0A030BF9F2FB0A02FB0A02FB0A030B0A02FB0A02FB0A02FB0\
A02FB09F2FAFA02FAF9F2FAF9F2FAF902FAF9F2FAF902FAF9F30B0902FAF9F2FB0
902FAF9F30B0902FAF9F30B09F2FB09F30B0902FB09F30B0902FB09030BF902FB09030BF9030BF\
9030C08F30BF9030C08F30BF8F30C08F30BF8F30C08030BF8F30C0802FBF8030BF
7F2FBF7F30BF702FB0702FBF702FB0702FB06F2FAF6F2FAF602FA0602FA05F2F9F5F2F9F502F8F\
502F8F402F7F402F70000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F4F302F5F302F703F307F3F2F8F3F308F3F2F\
9F40309F3F2FA04030A0402FAF4030A03F2FA040309F3F2F904030803F2F703F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F50402F7F4F2F8F602F9F602F9F702FAF702FAF7F2FAF\
7F2FAF802FBF802FB08F30BF8F2FBF9030BF8F30BF9030C09030BF9F30C09F30C0
A030C09F30BFA030C09F30BF9F30BF9F30BFA030BF9F30BFA030BF9F2FBFA030BFA02FBFA030BF\
A02FB0A030BFA02FB0AF30BF9F2FB0A030BF9F2FB0A030BF9F2FB0A030BF9F2FBF
A030BF9F2FBFA030BF9F30BF9F30BF9F30BFA030BF9F30BF9030BF9030BF9030C08F30BF9030C0\
8F30BF8F30C08F30C08F30C0802FBF8F30C0802FBF8030BF802FBF8030BF7F2FBF
7F2FBF702FB07030B06F2FAF6F2FAF6F2FAF6F2FAF602FA0602F9F5F2F90502F8F4F2F804F2F7F\
2F2F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003030304F302F603F306F3F2F803F30803F2F904030\
903F2FA04030A0402FA04030A0402FAF4030A0402FA04030903F2F8F403070302F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F7040207F502F905F2F9F602FA0602F9F\
6F2FA0702FA07F2FAF7F2FAF802FB0802FB08F2FBF802FB08F30BF8F2FBF9030C0
9030BF9F30C09030C09030C09030BF9030C09030BF9F30C0902FBF9F30BF9F2FBF9F30BF9F2FBF\
9F30BF9F2FBFA030BF9F2FB09F30BF902FB09F30BF902FB09F30BF902FBF9F30BF
902FBF9F30BF902FBF9F30BF902FBF9030BF902FBF9030BF8F2FBF8F30BF8F2FBF8F30C0802FBF\
8030C0802FBF8030BF7F2FBF802FBF7F2FBF7F2FBF7F2FB07F2FB0702FB07F2FB0
702FAF6F2FAF602FAF6F2FAF602FA0602F9F5F2F90502F904F2F804F2F7F3F2F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F4F30305F302F6F3F2F7F3F2F803F308F3F2F\
90402F9F3F2FA04030A03F2FA0402FA0402FA040309F3F2F903F2F8F3F2F703F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F4F3F2F70402F804F2F8F5F2F9F\
5F2F9F602FA06F2FA0702FAF702FAF7F2FB07F2FB08030BF802FB08F30BF8F2FBF
9030C09030BF8F30C08F30BF9030C09030C09030C09030C0903FC09030C09F30C09030C09F30C0\
9030BF9F30C09F30BF9F30C09030BF9030C09030BF9030C09030BF9030C09030BF
9030C09030BF9030C08F30BF9030C08F30BF9030C08F2FBF8F30C08030BF8030C0802FBF8030C0\
7F2FBF8030BF7F2FBF7F30BF7F2FB07F2FB0702FB07F30B0702FAF702FAF6F2FAF
7030AF602FA0602FA05F2F9F5F3090502F8F4F2F803F2F6F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000030303040302F5F3F306F3F2F7F3F30803F2F904030\
90402F9F40309F402FA04030A0402FA04030A0402FA04030903F2F8F4030703F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000302F5F3F2F70\
402F804F2F8F5F2F905F2F9F602FA0602FA0702FAF702FAF7F2FB07F2FB0802FB0
802FB08F30BF802FB08030BF802FBF8F30BF802FBF8F30BF8F2FBF8F30C08F2FBF8F30C08F30BF\
9030C08F30BF9030C09030BF8F30C08F2FBF8F30C08F30BF8F30C08F2FBF8F30C0
8F2FBF8F30C0802FBF8F30C0802FBF8F30C0802FBF8F30BF7F2FBF8030BF7F2FB07F2FBF7F2FB0\
7F2FB0702FB0702FB0702FAF702FB06F2FAF6F2FAF6F2FA06F2FAF602FA0602FA0
5F2F905F2F904F2F8F4F2F80402F70302F5F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F40302F5F302F6F3F30703F2F803F308F3F2F\
904030903F2F9F4030A0402FA04030A03F2FA0402F9F3F2F9040308F3F2F703F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F6F3F2F7F4F2F804F2F8F5F2F9F602F9F6F2FA06F2FA0702FAF702FAF
7F2FB07F2FB07F2FBF7F2FB0802FBF7F2FB08030BF802FB08030BF802FB08F30BF802FBF8F30BF\
802FBF8F30C08F30BF8F30C0802FBF8F30C08030BF8030BF802FBF8030BF802FBF
8030BF802FBF8030BF802FBF8030BF7F2FB08030BF7F2FB08030BF702FB07F30B0702FB0702FB0\
6F2FAF702FB06F2FAF6F2FAF602FA06F2FA0602F9F602F9F5F2F905F2F90502F8F
502F804F2F7F3F2F60000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000040302F5F3F3060302F7F3F30803F2F8F4030\
903F2F9F40309F402FA04030A0402FA04030A0402FA04030903F2F8F4030703F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000302F6F3F2F704F2F80502F8F5F2F9F5F2F9F6F2FA0
6F2FA06F2FAF6F2FA0702FAF6F2FAF702FAF702FAF7F2FB0702FAF7F2FB07F2FB07F2FB07F2FB0\
802FB07F2FB0802FBF802FB07F2FBF7F2FB07F2FBF7F2FB07F2FB07F2FB07F2FB0
702FB07F2FB0702FB07F2FB0702FAF7F2FB0702FAF7F2FB0702FAF702FAF6F2FAF6F2FAF602FA0\
602FA0602F9F602F9F5F2F905F2F90502F8F502F8F4F2F80402F7F3F2F60302F50
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000040303050302F603F2F703F2F803F30803F2F\
90402F903F2F9F40309F3F2FA0402F9F3F2FA040309F3F2F903F2F8F3F2F7F3F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F2F703F2F7F4F2F8F502F90
602F9F5F2F9F602F9F602F9F602FA0602FA06F2FA06F2FA0702FAF6F2FAF702FAF702FAF7F2FB0\
702FAF7F2FB07F2FB07F30B0702FB07F2FB0702FAF7F30B0702FAF702FB0702FAF
7030B0702FAF702FAF6F2FAF7030AF6F2FAF702FAF6F2FA06F30AF602FA0602FA05F2F9F5F309F\
502F90502F904F2F804F3080402F703F2F6F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F2F2F5F3F30603F2F703F307F3F2F8F4030\
903F2F9F40309F402FA04030A0402FA040309F402FA04030903F2F8F4030703F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F70
3F2F7F402F80402F804F2F8F4F2F8F502F90502F905F2F9F5F2F90602F9F5F2F9F602F9F602F9F\
6F2FA0602F9F6F2FA06F2FA06F2FA0602FA0602FA0602FA0602FA0602F9F602FA0
602F9F602F9F5F2F9F602F9F5F2F90602F9F5F2F905F2F905F2F90502F904F2F8F4F2F8F402F7F\
3F2F70302F5F503F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F302F50302F603F306F302F7F3F2F803F2F\
904030903F2F9F40309F3F2FA040309F3F2F9F402F9F3F2F9040308F3F2F7F3F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F6F3F2F703F2F70402F7F402F7F4F2F8F4F2F8F502F90\
502F905F2F905F2F905F2F9F502F905F2F9F502F905F2F90502F905F2F90502F8F
502F8F4F2F8F502F8F4F2F804F2F80402F7F4F2F7F402F70402F70302F6F302F60A08F2F9F8F2F\
907F2F8F702F705F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F2F50303060302F703F307F3F2F8F4030\
8F3F2F9F40309F402F9F40309F402FA040309F402F9F4030903F2F8F40307F3F2F
4F3030000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F6F302F6F3F2F703F2F703F2F703F2F6F3F2F70302F6F302F6F2F2F602F302F
4050205F6F2F6F702F7F802F808F2F90902F90902FA09F2FA09F2FA09F2FA0902FA0902F9F8020\
9F802F8F7F2F806F2F605020000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003030304F302F60302F6F302F7F3F30803F2F\
8F3F2F903F2F9F40309F3F2FA0402F9F3F2F9F4030903F2F903F2F8F3F2F7F3F30
50302F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F402F505F2F6F702F707F2F808F308F902F9F9F309F902FA09F30A09F2FAF9F2FA0902FA0902F\
9F8F2F9F802F8F702F7F602F4F3F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F2F5030305F302F703F307F3F2F804030\
8F3F2F904030903F2F9F40309F402FA040309F402F9F4030903F2F8F40307F3F2F
603F30000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F502F5F6020707F2F7F802F8F802F8F8F2F9F902F9F902FA09F2FA0902FA0902FA0902F\
A08F2F908020907F2F806F2F705F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F2F4F302F5F3F306F302F7F3F2F803F2F\
8F4030903F2F9040309F3F2F9F40309F3F2F9F402F903F2F904030803F2F7F3F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000303F2F505F2F606F2F7F7F2F7F7F2F8F8F30908F2F9F9030A0902FAF9F30A09F2FA0902F\
A08F2FA08F2F9F802F907F2F7F6F2F705F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F2F4F30305F302F6F3F307F3F2F803F30\
8F3F2F904030903F2F9F40309F402F9F40309F402F9F4030903F2F8F40307F402F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F402F5F5F2F606F207F7F2F807F2F908F30908F2FA0902FA0902FAF9F2FA0902F\
A0902FA08F2F9F8F2F907F208F702F7F602F503F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030302F4F302F5F3F2F60302F7F3F307F3F2F\
8F3F2F8F3F2F904030903F2F9F402F9F3F2F9F4030903F2F90402F80402F7F4030
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F302F4F502F60602F70702F807F2F80802F9090309F902FA09030A0902FB09F30\
AF902FAF902FA08F2FA08F2F907F2F8F6F2F705F2F4F302F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F2F503F305F3F2F6F4030703F2F804030\
80402F904030903F2F9F40309F4F2F9F4F309F402F9F4F3090402F8F4030703F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000303F2F4F502F606F2F70702F80802F8F802F9F90309F8F2FA0902FAF902F\
AF9F2FA0902FA0902F9F8F2F9F7F2F8F7020806F2F70502F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F302F4F302F5F3F306F3F2F7F402F7F402F\
8F40308F402F904F2F90402F9F4F3090402F9F4F2F90402F90403080402F703F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000040402F505F2F6F702F7F7F2F8080308F8F2F9F90309F902FAF9F30\
AF9F2FB09F2FAF902FAF8F2F9F802F9F7F2F90702F80602F6F502F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F2F2F503F305F3F2F70403070402F804030\
80402F904F3090402F9F4F3090402F9F4F3090402F904F308F402F8040306F3F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F502F5F602F70702F7F7F2F8F802F908F2F9F90309F902F\
AF9F2FAF902FAF902FA0802FA0802F9F7F2F9F7F2F8F6F207F602F6F4F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F302F4F302F5F3F2F6F3F2F7F40307F3F2F\
8F402F8F402F904F3090402F9F4F2F90402F904F308F402F8F402F7F402F6F3F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000303F2F505F2F60602F7F7F2F807F2F8F8F30908F30A09030\
A08F30AF9030AF902FAF9030A08F2FA08F2F9F7F2F9F7F2F8F6F2F7F602F5F402F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F302F503F30603F2F704030703F2F804030\
80402F904F308F402F9F4F30904F2F9F4F3090402F904F308F402F804030603F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F40205F5F2F606F2F7F7F2F807F2F908F30908030\
A08F30A08F2FAF902FAF8F2FAF8F2FA0802FA0802F9F7F2F90702F8F6F20705F2F
4F302F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F40302F50302F603F306F3F2F7F402F7F402F\
8F40308F402F904F2F90402F904F3090402F904F2F8F402F8040307F3F2F603F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000404F2F60602F6F6F2F807F2F807F2F908030\
9F8030A08F30A08F2FAF902FAF8F2FAF8F2FA0802FA0802F9F7F2F90702F806F2F
705F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000030303040302F5F3F30603F2F70403070402F804030\
80402F8F4F308F402F904F3090402F904F3090402F8F4F3080402F7F40305F3F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000303F2F4F502060602F706F2F807F2F807F2F\
9080309F802FA08F30A08F2FAF8F2FAF8F2FAF8F2FA08020A0802F90702F90702F
80602070502F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F40302F50302F603F2F6F3F2F7F40307F3F2F\
8F402F8F402F904F308F402F904F2F8F402F904F3080402F80402F703F2F5F3F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F402F505F2F6F602F706F2F807F30\
8F7F2F9080309F8030A08F30A08F2FAF902FAF8F2FAF8F2FA0802FA0802F90702F
90702F805F2F70502F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F3040302F5F3F30603F2F70403070402F804030\
80402F8F4F308F402F904F3090402F904F308F402F8F4F307F402F70403040302F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000040402F5050206F602F70702F\
807F2F8F7F2F9080309F802FA08F30A0802FAF8F2FAF802FAF8F2FA07F2F9F7F2F
906F2F906F2F7F5F206F4F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F4F302F50302F603F306F3F2F7F402F7F402F\
8F403080402F8F402F8F402F904F308F402F8F402F80402F7F40306F3F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4F2F5F5F2F6F6F2F\
7F702F8F7F308F7F2F9F8F309F802FA08F30A08F2FAF8F2FA0802FAF802FA07F2F
A07F2F906F2F906F2F805F2F6F4F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003030304F302F5F3F30603F2F7040307F402F804030\
80402F8F4F308F402F904F308F402F904F308F402F8040307F402F6F4030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F302F4F502F5F5F20\
6F6F2F7F702F8F7F2F8F7F2F9F80309F802FA08F2FA0802FAF802FA07F2FA07F2F
9F70209F702F906F2F8F602F7F5F206F4F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F2F302F2F4F302F50302F603F2F6F3F2F7F402F7F3F2F\
80402F80402F8F4F308F402F8F402F8F402F8F402F7F402F7F402F603F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F3F2F50502F\
60602F706F2F7F702F8F7F3090802F9F8F309F802FA08030A0802FAF802FA07F2F
AF7F2FA07F2F9F702F906F2F906F2F805F2F6F4F2F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F2F4030304F302F5F3F30603F2F7040307F402F804030\
80402F8F4F308F402F904F308F402F8F4F3080402F80403070402F603F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303F2F\
505F2F60602F706F2F7F702F8F7F2F907F2F9F80309F7F2FA08030A07F2FAF802F
A07F2FA07F2F9F702F9F702F906F2F8F602F7F5F2F6F4F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F3040302F503F2F5F3F2F6F40306F3F2F7F402F7F402F\
80403080402F8F402F8F402F8F4F3080402F80402F7F402F7040305F3F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F402F5F5F2F60602F7F702F80702F8F7F30907F2F9F80309F7F2FA08030A07F2F
AF802FA07F2FA07F2F9F702F9F702F906F2F8F602F805F2F6F4F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F2F4F3F30503F2F6040306F402F704F307F402F804030\
80402F8F4F308F402F8F4F308F402F8F4F3080402F7F40306F402F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F402F5F5F2F6060207F6F2F806F2F8F702F8F702F9F7F2F9F7F2FA0802F
9F7F2FA0802FA07F2FA07F2F9F70209F702F906F2F8F602F7F50206F4F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F2F3F302F4F302F5F3F2F603F2F6F402F70402F7F4F2F7F402F\
804F2F804F2F8F4F2F804F2F8F4F2F804F2F804F2F70402F60402F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000404F2F605F2F6F602F7F6F2F80702F8F7F30907F2F9F80309F7F2F
A08030A07F2FAF802FA07F2FA07F2F9F702F9F702F90602F8F602F805F2F704F2F5F3F2F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F303040302F503F305F3F2F6F40306F402F7F4F307F402F804F30\
804F2F8F4F30804F2F8F5030804F2F804F307F4F2F704F305F3F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F3F2F404F2F5F5F2F605F2F7F6F2F806F2F8F702F8F702F9F7F2F
9F7F2FA07F2F9F7F2FA07F2FA07F2FA07F2F9F6F2F9F6F2F90602F8F602F80502F704F2F5F3F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F2F4F302F50302F603F2F603F2F70403070402F7F4F2F7F4F2F\
804F30804F2F8F4F2F804F2F804F307F4F2F7F4F2F6F402F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F3F2F404F2F605F2F6F602F7F6F2F806F2F8F7F2F90702F
9F7F2F9F7F2FA08030A07F2FA07F2FA0702FA0702F9F6F2F9F6F2F90602F90602F805F2F7F502F\
60402F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F2F4030304F302F5F3F30603F2F6F40306F402F7F4F307F4F2F804F30\
804F2F8F4F30804F2F804F307F4F2F7F4F3070402F604030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F3F2F404F2F5F5F2F605F207F6F2F806F2F8F702F
8F702F9F7F2F9F7F2F9F7F2F9F702FA0702F9F702FA0702F9F6F209F6F2F90602F8F602F805020\
7F502F6F402F2F2F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000302F2F402F2F503F2F5F3F2F60402F6F3F2F70402F70402F804F2F7F402F\
804F2F804F2F804F2F7F4F2F7F4F2F70402F6F402F503F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F3F2F404F2F605F2F6F602F7F6F2F80702F
8F7F2F90702F9F7F2F9F702FA07F2F9F702FA0702F9F702FA0702F9F6F2F9F6F2F90602F90602F\
805F2F80502F6F402F5F3F2F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003030303F2F2F4F3F30503F2F603F30603F2F70403070402F7F4F307F4F2F804F30\
804F2F804F307F4F2F804F307F4F2F704F3060402F4F3F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F3F2F404F2F5F5F2F605F2F7F6F2F
806F2F8F702F8F6F2F90702F90702F9F702F9F702F9F702F9F702FA0702F9F6F2F9F6F2F90602F\
90602F80502F80502F70402F603F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F2F40302F4F302F5F3F2F5F3F2F6F402F6F402F70403070402F7F4F2F7F4F2F\
804F307F4F2F7F4F2F7F4F2F7F4F306F402F5F402F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000303F2F404F2F605F2F6F602F
7F6F2F806F2F8F702F8F6F2F90702F90702F9F702F9F702F9F702F9F6F2FA0702F9F6F2F9F6F2F\
90602F90602F8F502F80502F7F4F2F704F2F5F302F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F2F2F4030304F302F503F305F3F2F6040306F402F70403070402F7F4F307F4F2F7F4F30\
7F4F2F804F307F4F2F7F4F3070402F6F4F30503F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F3F2F404F2F5F5F2F
605F2070602F7F602F806F2F806F2F90702F90702F9F702F906F2F9F702F9F6F2F9F702F906F20\
9F6F2F905F2F905F2F8F502080502F7F4F2F704F2F603F2050302F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F2F40302F40302F503F2F503F2F5F3F2F603F2F6F402F6F402F70402F7F402F7F4F2F7F402F\
7F4F2F7F4F2F7F4F2F70402F6F402F60402F4F3F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000303F2F404F2F
605F2F6F5F2F70602F7F602F8F6F2F8F6F2F90702F906F2F9F702F9F702F9F702F9F6F2F9F702F\
9F602F9F602F905F2F90602F8F5F2F8F5F2F7F4F2F7F4F2F70402F603F2F4F2F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F2F
40303040302F503F30503F2F5F403060402F6F40306F402F704F3070402F7F4F307F4F2F7F4F30\
7F4F2F7F4F30704F2F704F306F402F5F4030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000303F2F
40402F5F502F605F2F70602F7F602F806F2F806F2F8F6F2F8F6F2F9F702F906F2F9F702F90602F\
9F6F2F90602F90602F905F2F905F2F8F502F8F502F80502F7F4F2F70402F6F402F
50302F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F2F3F2F2F
402F2F4F302F50302F5F3F2F5F3F2F6F402F6F402F704F2F70402F7F4F3070402F7F4F2F7F4F2F\
7F4F3070402F704F2F6F402F60402F4F3F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
30302F3F402F5F502F605F2F6F602F70602F806F2F806F2F8F6F2F8F6F2F90702F906F2F9F6F2F\
90602F9F6F2F90602F9F602F905F2F90602F8F5F2F8F5F2F80502F80502F70402F
70402F6F3F2F603F2F50302F000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F2F2F40303040302F
4F302F50302F5F40305F3F2F6040306F402F6F4F306F4F2F704F3070402F7F4F307F4F2F7F4F30\
704F2F7F4F3070402F6F40305F402F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F3F402F50502F5F50206F5F2F70602F806F2F80602F8F6F2F8F6F2F906F2F90602F\
906F2F90602F90602F90602F90602F905F2F905F2F8F502F8F502F804F2F804F2F
7F402F70402F6F3F2F6F3F2F5F302F5F302F402F2F000000000000000000000000000000000000\
000000000000000000000000000000302F2F3F2F2F3F2F2F40302F402F2F4F302F
4F302F5F3F2F5F3F2F60402F60402F6F402F6F402F704F2F704F2F704F2F704F2F7F502F704F2F\
704F2F6F4F2F6F4F2F5F402F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000303F2F4F4F2F5F502F6F5F2F70602F7F6F2F806F2F8F6F2F8F602F906F2F\
90602F906F2F90602F906F2F90602F90602F8F5F2F90602F8F502F8F502F804F2F
80502F7F4F2F7F4F2F70402F70402F6F3F2F603F2F5F302F5F302F50302F50302F4F2F2F4F3030\
402F2F302F2F402F2F4F3030402F2F40302F402F2F4F30304F302F503F3050302F
5F3F305F3F2F60403060402F6F4F306F402F704F30704F2F7F5030704F2F7F5030704F2F705030\
704F2F704F3060402F504030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000303F2F4F4F2F50502F605F2F6F5F2F7F602F7F602F80602F80602F\
8F602F8F602F90602F8F602F90602F8F5F2F90602F8F5F2F8F5F2F8F502F8F502F
804F2F80502F7F4F2F7F4F2F70402F70402F6F3F2F6F3F2F60302F603F2F5F302F5F302F502F2F\
50302F4F2F2F4F302F4F2F2F4F302F4F2F2F4F302F4F302F503F2F50302F503F2F
5F3F2F60402F5F3F2F6F402F6F402F6F4F2F6F402F704F2F704F2F704F2F704F2F704F2F6F4F2F\
6F4F2F5F402F5F402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F302F404F2F50502F605F2F6F5F2F70602F7F602F80602F\
80602F8F602F8F602F8F602F8F602F90602F8F602F905F2F8F5F2F8F5F2F80502F
8F502F80502F80502F7F4F2F7F4F2F70402F70402F6F3F2F6F402F603F2F603F2F60302F5F3F2F\
5F302F5F3F3050302F50302F50302F503F3050302F5F3F2F503F2F5F40305F3F2F
60402F603F2F6F40306F402F6F4F306F402F704F30704F2F704F30704F2F7050306F4F2F6F4F30\
604F2F5F4F3030302F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F40402F4F4F2060502F60502F6F5F2F705F2F\
7F602F7F602F80602F805F2F8F602F8F602F8F5F2F8F5F2F8F5F2F80502F8F5F2F
80502F80502F7F4F2F80502F7F402F7F4F2F70402F70402F6F3F2F6F402F603F2F6F3F2F60302F\
603F2F5F302F5F3F2F50302F5F302F50302F5F3F2F503F2F5F3F2F5F3F2F60402F
5F3F2F60402F60402F6F4F2F60402F6F4F2F6F4F2F704F2F6F4F2F704F2F6F4F2F6F4F2F60402F\
5F402F403F2F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F402F4F4F2F5F502F60502F6F5F2F\
6F5F2F7F602F7F5F2F80602F80602F8F602F805F2F8F5F2F805F2F8F5F2F805F2F
8F5F3080502F80502F80502F8050307F4F2F7F4F2F70402F704F306F402F6F402F6F3F2F6F4030\
603F2F60402F603F2F6040305F3F2F603F305F3F2F60403060402F60403060402F
6F4F3060402F6F4F306F402F6F4F306F4F2F704F306F4F2F7050306F4F2F6F4F30604F2F5F4F30\
4F402F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000303F2F3F402F4F4F2F5F502F\
605F2F6F502F705F2F705F2F7F602F7F5F2F805F2F7F5F2F805F2F7F502F805F2F
80502F805F2F80502F80502F7F4F2F7F4F2F704F2F7F4F2F70402F704F2F6F402F6F402F6F3F2F\
6F402F603F2F60402F603F2F603F2F5F3F2F60402F60402F60402F60402F60402F
60402F6F4F2F60402F6F4F2F6F402F6F4F2F6F402F6F4F2F6F4F2F6F4F2F5F402F5F4F2F4F402F\
303F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F2F3F402F4F4F2F\
504F2F605F2F60502F705F2F705F2F7F5F2F7F5F2F805F2F7F5F2F805F2F7F5F2F
805F2F7F502F805F2F7F502F7F502F7F4F2F7F502F704F2F704F2F704F2F704F306F402F6F402F\
60402F6F403060402F6F402F60402F6F403060402F60402F60402F6F4F30604F2F
6F4F2F604F2F6F50306F4F2F6F4F306F4F2F6F4F306F4F2F604F305F402F604F3050402F403F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F2F20\
3F402F4F402F5F502F60502F6F5F2F6F502F705F2F70502F705F2F70502F7F5F2F
7F502F7F5F2F7F502F7F502F704F2F704F2F704F2F704F2F704F2F704F2F6F4F2F704F2F60402F\
6F402F60402F6F4F2F60402F60402F60402F60402F60402F604F2F60402F604F2F
60402F6F4F2F604F2F6F4F2F604F2F6F4F2F60402F604F2F5F402F5F402F50402F40402F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F3F2F40402F50502F5F502F60502F60502F6F5F2F6F502F705F2F70502F
7F5F3070502F7F5F2F7F502F7F503070502F7F502F704F2F705030704F2F704F2F6F402F704F30\
6F4F2F6F4F2F60402F6F4F30604F2F6F4F3060402F6F4F30604F2F6F4F30604F2F
6F4F30604F2F6F5030604F2F6F5030604F2F6050305F4F2F5F4F3050402F404030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F302F3F3F2F4F4F2F4F4F2F5F4F2F5F4F2F60502F60502F6F502F
6F502F705F2F6F502F70502F704F2F70502F6F4F2F70502F6F4F2F70502F6F4F2F6F4F2F6F402F\
6F4F2F60402F6F4F2F604F2F6F4F2F604F2F6F4F2F60402F604F2F60402F604F2F
604F2F604F2F5F4F2F604F2F5F4F2F5F4F2F504F2F504F2F40402F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F302F40402F40402F4F4F2F504F2F5F502F60502F
6F502F60502F6F5F2F6F4F2F6F502F6F4F2F6F502F6F4F2F6F502F6F4F2F6F50306F4F2F6F4F2F\
604F2F6F4F30604F2F6F4F2F604F2F6F5030604F2F604F2F604F2F604F30604F2F
604F2F5F4F2F604F305F4F2F5F4F2F50402F504F303F3F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000303F2F303F2F40402F4F402F504F2F
504F2F5F502F60502F60502F604F2F60502F604F2F6F502F604F2F6F502F604F2F6F4F2F604F2F\
604F2F604F2F604F2F5F402F604F2F5F4F2F604F2F5F402F604F2F5F402F5F4F2F
50402F5F4F2F50402F504F2F4F402F40402F303F2F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F302F3F3F2F3F402F
4F4F2F4F4F2F50502F504F2F5F4F305F4F2F5F4F2F5F4F2F6050305F4F2F60502F604F2F604F30\
5F4F2F604F2F5F4F2F604F305F4F2F60502F5F4F2F6050305F4F2F5F4F3050402F
5F4F3050402F4F4F3040402F3F4030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F302F3F402F40402F4F4F2F4F402F504F2F4F402F504F2F504F2F5F4F2F504F2F5F4F2F5F402F\
5F4F2F50402F5F4F2F504F2F5F4F2F50402F504F2F504F2F504F2F50402F504F2F
4F402F40402F3F3F2F303F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000303F2F3F402F3F3F2F40402F40402F4F4F2F40402F504F2F4F4F2F504F2F\
4F402F504F304F402F4F4F2F4F402F4F4F304F402F4F4F2F40402F404F303F402F
3F402F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F302F303F2F2F3F2F30402F303F2F3F402F3F402F\
40402F3F402F40402F3F402F3F402F3F3F2F3F3F2F303F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F302F303F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[36]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061d@000MaK75/LFaa0<L020L000H7?`005g40
LF^:JhY[RV^:JhY[RV^:JhY[LF]a0;`020L000H7>P007g5[LFb:K8Z5RXF:QHZ5RXF:QHZ5RXF:QHZ5
RXF:K8X0^@081`001PLg000TLF^:JhY[RV^:K8Z5RXF:QHZ5RXF:QHZ5RXF:Q8Z4RXB:Q8Y[K0000g40
L@1800P700061cD002U/RVb:QHZ5RXF:QJB5Y8FTQJ>5XhFSQJ>5XhFSQJ>5XhFSQJ>5RXF:K01N00=a
00=/LG402g5500P700061c<002iaJhY/RVb:QHZ5RXF:QHZ5XhFSQJ>5XhFSQJ>5XhFSQJ>5XhFSQ8Z4
RXB:JhY[F0005G40L@1a0740L@1a0740L@1a0740L@1400P700061c4002i/RVb:QHZ5RXFTQJB5Y8FT
QJB5Y9jTWZBSY9jTXjB5Y8FTQJB5XhFSQJ>:XhFS0XX00hF:L@1B01]a@@081`001PL_000gLF^:K8Z5
RXF:QHZ5Y8FTQJB5Y8FTQJB5Y8FTQJB5Y8FTQJB5Y8FSQJ>5XhFSQHZ5RXB:Q8Y[RP1=000OL@1a0740
L@1a0740L@1a674HLAQa674HL@1a0740L@1000P700061bh003YaK8Z5RXF;QJB5Y8FTQJB5Y8FTQJB5
Y9jTXjBNY:>TWZBNY9jTWZBNY9jTWZ>NXijSQJ>5XhFSQHY[B@0RL@03L00003`020L000H7:`00@75[
RVb:QH^5RhFTQJB5Y8FTQJB5Y8FTWZB5Y8FTQJB5Y8FTQJBNY9jTWZBNY9jTWZ>NXhFSQJ>5XhFSQHZ4
RV]2000YL@1a0740LAQa675>LDiaCW5>LDiaCW5>LDiaCW4HLAQa674HL@1a0740?0081`001PLY000A
K75/RXF:QJB5Y8FTQJB6Y9l00j@01InTY:BO00>T00^6Y8JTWjBOY9nTW`03Y00JWZBSY9jTWZBNY9jT
WZBNXijSWZ>5XhF:QHXn00]a01MbLG9aLW5bLG9aLW5bLG9aLW5bLG9aLP0:LC/020L000H7:000AV]a
K8Z5RXF;QJB5Y8FTQJB5Y8JTQZB5Y8FTQJB5Y8FTQJB5Y8FTQJB5Y8FTWZBNY9jTWZBNY9jTWZ>NXijS
QJ>5XhB:Q8Xk000]L@1a074HLAQa675>LTibCW9>LTibCW9>LTibCW9>LTibCW8HLDia675>L@1a03`0
20L000H79`00B6b:QHZ5RhFTQJB5Y8JTQZB6Y9nTWjBOY9nTWjB6Y8FTQJB5Y8FTQJB6Y8JTWZBNY9jT
WZBNY9jTWZBNY9jSWZ>NXhFSQJ>5RSP02G402W9aLW5bLG9aLW49LP0<LG9aLW5bLG9aLW5b2G4j00P7
00061bH004][LFb:QH^5RhFTQJB5Y8JTQZB6Y8FTQJB5Y8FTQJB5Y8FTQJB5Y8FTQJB5Y8FTQJB5Y8FT
QJBNY9jTWZBNXijSWZ>NXhFSQJ>4RV/0=000<g40LAQa674HLAQbCW9>LTibCW9>LTibCW9>LTibCW9>
LTibCW9>LQQbCW5>LDia6740L@0j00P700061b@004iaK8Z5RhF;QJB5Y8JTQZB6Y8JTQZBOY9nTQZB6
Y8JTQZB6Y8FTQJB5Y8FTQJB5Y8FTQJB5Y8FTWZBNY9jTWZBNXijSWZ>NXijSQJ>5RX@a00Ua00QbLG9a
LW5bLAMb00QaLW5bLG9aLPEa>P081`001PLT001@Jg5/RXF;QH^5Y8FTQJB6Y8JTQZB5Y8FTQJB5Y8F;
QH^5RhF;QH^5RhF;QH^5RhF;QH^5Y8FTQJB5Y8FTWZBNY9jSWZ>NXijSWZ>4XhB:Q8X_000eLAQa675>
LDibCW9>LTibCW9>LTibCg9?LTmbCg9?LTibCW9>LTibCW9>LTibCW9>LQQa6740>P081`001PLT001B
RXF;QH^5Y8JTQZB6Y8JTQZB6Y8JTQZB6Y8JTQZB6Y8J;QH^5RhF;QH^5RhF;QH^5RhFTQJB5Y8FTQJB5
Y9jTWZBSY9jSWZ>NXijSQJ>5XhF:JbX000=aK7401G400g9aLP0;LP03Mg9b01Ib00IaLW5bLG83LCX0
20L000H78`00E75/RXF;QH^5Y8JTQZB6Y8JTQZB5Y8FTQH^5RhF;QH^5Rfb;K8]/Rfb;K8]/Rfb;QHZ5
RhF;QH^5Y8FTQJB5Y8FTWZBNXijSWZ>NXijSQJ>4XhB:JbP003Ua074HLDiaCW9>LTibCW9>LTmbCg9?
LTmbCg9?LTmbCg9?LTmbCg9?LTibCW9>LTibCW8HLQQa6740>P081`001PLQ000;Jg5/RhF;QZB6Y8H0
0j@0?8JTQZB6Y8JTQZB6Y8J;QH^5RhF;QH]/Rfb:K8]/Rfb;K8]/Rfb;QH^5RhFTQJB5Y8FTQJB5Y9jT
XjBS^`:S00^NXijSWZ>5XhFSQ00U00Ma00=bLG802G803gMbMg9gLWMbMg9gLWMbM`0DLP04LG9aLP9a
>@081`001PLQ000TLFb:QH^5RhFTQZB6Y8JTQJB5Y8FTQJB5RhF;QH]/Rfb:K74H2000;74HLFaaK75/
RVb:K8^5RhF;QH^5Y8FTQJBNY9jSWZ>NXijSWZ>NXhBSQ8Z48P00?G40LAQa674HLTibCW9>LTmbCgM?
LTmbCg9?LTmbCg9?LTmbCg9?LTmbCg9?LTmbCW9>LTibCW8HLQQb6740>P081`001PLN000OLFb:QH^5
Y8JTQZB6YInUWjBOY8JTQZB6Y8J;QH]/L@0J000JK75/RW6;QH^:RhFTRZB5Y9jTWZBNY9jkXkX6X`04
WZ>5RQd01740179aLW44LP0MMg9gLWMbN79hLWQbN79hLWQbN79hLWQbN79hLWP05780175bLG4i00P7
00061ah001e/LFb:QH^5Y8FTQZB6Y8JTQZB5Y8F;QH^5RhF:K00N000SLAQaK75/RVb:QHZ5XhFTQJB5
Y9jTWZ>NXijSWZ>NXhB:Q8X06`00@W4HLAQa679>LTibCg9?LTmbCgM?MdmgCgM?LTmbCg9?LTmgCgM?
LTmbCg9?LTmbCg9?LTmbCW8HLQQb678HLQQb63T020L000H77@0076aaQH^5RhJTQZB6YInTWjBOY8JT
QZB6RhJ;QH/S0004LFaaK0>:01:5Y8FTQJB5Y9jTWZBS^j>jXkX4X`04QJ>5RQP01G400g9aLP07LP0M
N79hLWQbN79hLWQbN79hLWQbN79hLWQbN79hLWP05W82LCT020L000H77@006g5/RhF;QH^5Y8JTQZB6
Y8JTQZB6RhF;QH]/L@0W000OLFaaK8Z5RXF:QJB5Y8FTWZ>NXijSWZ>NXijSQ:>4RP0G0014L@1aCW9>
LTibCg9?LTmbCg9?MdmgCgQ?MdmhCg=?LTmbCg9?MdmbCg9?LTmbCg9?LTmbCg8HLTib678HLQQb678H
LQPi00P700061a`001YaK8^5RhJTQZB6YInUWjBOY8JTQZB6RhF;K2/001I/LG6;QH^5Y8FTQJB5Y:>T
XkZS^Z>j0Z<01YjSQJ>5RQ<017400g9aLP07LP0OMg9hLWQbN79hLWQbN79hLWQbN79hLWQbN79hLWQb
N00HLP9a>@081`001PLK000ILFb:K8^5RhJTQZB6Y8JTQZB6Y8F;QH^5RP0_000MLFb:K8Z5RXFSQJB5
XijSWZ>NXijSWZ>4XhBSQ8X04@00Ag40LAQa679>LTmbCg9?LTmbCgQ?N4mhCgQ?N4mhCgQ?LdmcCg9?
LTmbCg9?LTmbCg9>LTib678HLQQb679;LT]b678HLQQb03X020L000H76P006G5/RXF;QZB6Y8^UWjFO
Y9nTQZB6Y8J;QH/0<@002W5/RXF;QH^5Y8D2X`0BY:>TXkZS^Z>jWZ>NXhFSQHY[3@05L@03LW5b00=b
02=gLWQbN79hLWQbN79hLWQbN79hLWQbN79hLWQbN79hLWQbM`05LP<000=aLW803g82LCT020L000H7
6`005g5/RhF;QH^6Y8JTQZB6Y8JTQJB5Rfb:03<001iaK75/RVb:QHZ5XhFSWZ>NXijSWZ>NXifSQ:>4
RV/<000@LAQa678HLTibCg9?LTmbC`9b01mcCgQ?N4mhCgQ?N4mhCgQ?N4mcCg9?LTmbCg9?LTib00D0
01=a678HLQQbBg9;LT]bBg8HLQQb03X020L000H76P00275/RhJ;QZB;0Z@03JFOY9nTQZB6Y8JTQH/0
=P002FaaK8^5RhFSRP02X`0AY:>TXk^S^Z>jWZ>NXijSQJ<02P03L@03LW5b00=b02EgLWQbN79hLWQb
N7=hLWQbN79hLgQbN79hLWQbN79hLWQbN79g009b20000g5bLP0?LP03K00003P020L000H76P005Vb;
K8^5RhJTQZB6Y8JTQZB5Y8F;QHXi000KLFb:K8Z5RXFSQJ>NXijSWZ>NXijSWJ>4XhB:00T002ia678H
LTibCW9?LTmgCgQ?N79hLWQ?N4mhCgQ?N4mhCgQ?N4mhCg9?LTmbCg9>2P004g4HLT]bBg9;LT]bBg9;
LT]b6780>P081`001PLI000FK8^5RhJTRjB;YJBUWjFOY8JTQZB5Rc/000AaK8Y/0XX05J>5Xj>TXjBS
^j>jXkZN^YjSWZ>5RP0500Aa00=bLG800g808gMbN79hLWQbN7=hLgQcN7=hLgQcN79hLWQcN79hLWQb
N79g009b2`02L@Qb00YcLW=]LfebKG9/>P081`001PLH000FK75/RhF;QZB6Y8JTWjB6Y8JTQH^5RSl0
04UaK8Z5RXFSQJ>5XijSWZ>NXijSWZ>MXhB:Jg40LAQb679?LTmbCg9?N4mhCgQ?N4mhD7Q@N51hD7Q@
N4mhCgQ?N4mhCgM?LTmb00d001Ea678HLT]bBg9;LT]bBg9;LT]b6740>P081`001PLG000FK75/RhJT
QZB6YHJUWjBOY8JTQZB5Rd4000]aK8Z5RXFSQJB5Y003X`05^Z>jXkX00Z<00ijSQ@04L@03LW5b00Eb
00YhLWQbN7=hLgQc1GP00g=hN003N00:LgQcN79hLWQbN09b00=a00003@003QQbKG9]LfecKG=bLg9c
1W800g40000h00P700061aL001EaK8^5RhJTQZB6Y8JTQZB6Y8J;QH/0@`00@g5/RVb:QHZ5XhFSWZ>N
XijSWZ>NXhAa074HLTibCW9?LTmcCgQ?N4mhD7Q@N51hD7Q@N51hD7Q@N4mhCgQ?N4mgCg804@00578H
LT]bBg9;LT]bBg9;LT]bBg8H>`081`001PLF000FLFb;QH^6Y8JUQZFOYInTQZB6Y8J;K4@000]aK8Z5
RXFSQJBNY003X`06^Z>jXkY/1746LP0:Lg9hLWQbN7=hL`eh00]cN7=hLWQbN79gL@0B000@LAQbBg9]
LfecKG=]Lg9cK@=b00=]L@00>P081`001PLF000E675/RhF;QZB6Y8JTQZB6Y8J;QH]/04H003eaK8Y/
RXF:QJ>5XijSWZ>NXiia674HLTibCg9?LTmhCgQ?N51hLgQ@N51hD7Q@N51hD7Q@N4mhCgQ?Mdmb01@0
01@HLQQbBg9;Ld]cBg=;LT]bBg9;LS`020L000H75P00575/RhJTQZB6YHJUWjFOY8JTQX^5B0002g5/
RXF:QJ>5XijT00>S00JjXg5aLW44LP0:Lg9hLWQcN7=hL`mh00YcN7=hLWQbN79g5@003g9;LVecKG=]
LfecLW=bL`04LP03L@0003X020L000H75@005G5/Rfb;QZB6Y8JTQZB6Y8JTQH^5RP19000iLFb:K8Z5
RXFSQJ>NXiia678HLTmbCg9?LdmhCgQ@N51hLgQ@N51hD7Q@N51hD7Q@N4mhCgQ?Mdmb01H001@HLT]b
Bg=;Ld]cBg=;LT]bBg8HLC`020L000H75@0056abQH^6Y8JUQZFOYInTQZB6Y8J;B`002g5/RVb:QJ>5
XijS00Ea178027QbN79hLWQc27P00gUhN@08N007LgQbN79gLP0G000AK79]LVecKG=]LfecKG=bLfd0
0g8m00P700061aD001=aK8^5RhJTQZF6Y8JTQZB6RhF;04d003=aJhY/RXF:QJ<0LAQa679?LTmgCgQ?
N4mhD7Q@N7=hD7Q@N51hD7Q@N51hD7Q?N4mhCg806@00575;LT]bBg=;Ld]cBg=;LT]bBg8H?@081`00
1PLD000DLFb;QX^6YHJUWjFOYInTQZB6RhE?0006K8Z5RXFS0g403W9aLW9cLWQbN79hLgQc17P01gUh
N7QiN7T02WP0179hLWLI000ALFabKG=]LfecKG=]Lg9cLW<00W800f`0000l00P700061a<001Aa68]/
RhF;QZB6Y8JTQZB6Y8F;K58002Ya074HLAQbCW9?LTmcCgQ@N51hD7Q@N51hD7Q@N51hD7Q@N51hCgQ?
N4lL000DLAQbBg9;Ld]cBg=;Ld]cBg9;LQPo00P700061a<001=/LVb;QZB6YJBUY:FOYInTQZB505<0
0g400g9aLP03LP06N79hLgQc17P02gUhNGQiN7UhNGQi00Qh00AgN79g6`00275/LVecLW=b1g<01Vec
KG9]LCl020L000H74`004g5/Rfb;QX^6Y8JTWjB6Y8JTQHX0D`00:74HLQQbCg9?LTmhCgQ@N51hD7Q@
N5EhD7Q@N51hD7Q@N51hCgQ?MdlM000D678HLT]cBg=;Ld]cBg=;Ld]b679000P700061a8001AaK8^6
RhJTQZFTYJBUWjFOY8J;K5800W44LP08Lg9hLWQcN7<2N00?NGQiN7UhNGQiN7UhNGQi00Ih00=gN780
7@001fabKG9]Lg802G<01FecKG9/040020L000H74`004W5/RhF;QZB6Y8JTWjB6Y8J;K58002Qa678H
LTmbCg=?N4mhD7Q@N51hD7QEN5EiD7Q@N51hD7Q@N4mhCgM?7P00578HLT]bBg=;Ld]cBg=;Ld]bBg8H
@@081`001PLB000CLFb;QX^6YHJUY:FTYInTWjB6R`1A009a178027=bN79hLgQc0WP04GUhNGQiN7Uh
NGQiN7UhNGQi00Ah00=bM`007P001W9/LVecK@Uc00EbLfebK01200P700061a4001Aa68]/RhJ;QZB6
Y8JTWjB6Y8J;K50002Ma674HLTmbCg=?N4mhD7Q@N51hD7QEN5EhEGUEN51hD7Q@N4mhCg807`005G4H
LT]bBg=;Ld]cKG=;LfecBg8HL@1200P700061a4001=/LVb;QZB6YHJUY:FOYInTQZB604l00g44LP0K
Lg9hLWQcN7=iN7UhNGQiN7UhNGQiN7UhNGQi009h00=bN7808000275/LVecKG=]2G<0179cKG9300P7
00061a4001=aK8]/RhJ;QZB6Y9nTQZB6Y8F;04l002Ia678HLTmbCg=?N51hD7Q@N51hD7Q@N5EiEGQE
N5EhD7Q@N4mhCb4001DHLQQbBg9;LfecKW=;LfecBg9;LQP0@`081`001PLA000CK8]]RhJUQZF;YJBU
WjFOY8JTQ@1=00=a17806G=bN7=hLgQcNG=iN7UhNGQiN7UhNGQiN7T017P00g9hLP0Q0009K79]LVec
KG=]00Mc00EbLg9bK01400P700061a4001=aK8]/RhJTQZF6Y9nTQZB6Y8F:04d002Ia678HLTmbCgQ?
N51hD7Q@N51hD7Q@NEEiEGQEN5EhD7Q@N4mgCb8001Ab679;LT]cBg=^LficKW=;Ld]bBdD020L000H7
4@004fb;QX^6YHJUY:FOYInUQZB6Rf`0BP02L@03LW5b00=b01MhLWQcN7=iLgUhNGQiN7UhNGQiN7Uh
N@06N004LZ>5RR4001EbKG9]LfecKW=^LficKW=]LfecKG80A@081`001PLA000CLFb;QX^6Y8JUQZF6
Y8JTQZB5RP1:000Z674HLTmbCg=?N4mhD7Q@N51iD7U@NE1iEGUENEEhEGQ@N4mgCj>MXhB:7`005G4H
LVebKG=]LficKW=]LfecKG9;LP1600P700061a0001=aK8^6RhJURjFTYInUWjF6Y8J;04X000=aLW40
0W801W=bN79hL`9h01=iN7UhNGQiN7UhNGQiN7UhNGQi009h00UgN7:SWZ>NXhD07P00275/LVecKG=^
0g<02VicKW=]LfecKG9700P700061a40019aK8^6RhJTQZF6YHJTQZB6RhE:000[LAQbCg9?LdmhCgQ@
N51iD7U@NE1iEGUENEEiEGQ@N51hCgNNXifSWJ>4X`0N000E679]LVecKG=^LficKW=]LfecBg8H04L0
20L000H740004g5/RhJ;QZF6YJBUWjFOYHJTQX/0B003L@9b01UcLW=bN7=hLgUhNGQiN7UhNGQiN7Uh
NGQi00Ah0WL02Z>jXkZNXijSQHXL000EK79]LfecKG=^LficKW=^LfecKG9/04P020L000H740004aQb
K8^6RhJTQZF6YHJTQZB5Rf`0B000;W4HLTmbCg=?N4mhD7Q@NE1iEGUENEEiEGUENE1hD7Q@N4mgWZ>N
XijSWJ>MXh@K000E678HLVecKG=^LficKW=]LfecKG8H04T020L000H740004g5/RhJ;QZF;YJBUWjFO
YHJTQX/0AP03L@9b01UcLW=bN7=hLgUcNGQiN7UhNGQiN7UhNGQi00Ah00=gN7800Z<00kZS^P02X`05
WZ>NXh@06@002aQbKG9]LfecKW=^00=c00Q^LfecKG=]LTT020L000H740004aQbK8^6RhJTQZF6YHJT
QZB5Rf`0AP00<W4HLTmbCg=?N4mhD7Q@N51iEGUENEEiEGUENE1hD7Q@N4mgQ:>NXijSWZ>MXifSQ8Y[
60005G4HLVebKG=]LficKW=^LfecKG9;LP1:00P700061a0001=aK8^6RhJURjFTYInUWjB6Y8F:04@0
0g44LP0GLg9hLgQcNG=iN7UhNGQiN7UhNGQiN7T017P01G9gQJ>500BS00>jXkX00Z<019jSQJ<G000E
LVabKG=]LficKW=^LficKW=]Lfeb04/020L000H73`00575/LVf;QX^6YHJUQZF6Y8JTQH]/A000=G4H
LTmbCg9?N51hD7Q@N51iD7UENEEiEGUENEEhD7Q?N4mgJhZ4XhFSWZ>NXijSWJ>MXhB:01D001Ea679]
LVecKG=^LficKW=^LfebBg80C0081`001PL?000C679]RhJUQZF6YJBUWjFOY8JTQ@12009a01ebLG9b
Lg9hLWQcN7=iN7UhNGQiN7UhNGQiN7UhN@02N09g0P000g6:RP02RP03XhFS00>S00BjXkZN0j<2RQ80
00M/LW9bKG=b009c00^<KXa^S6j<KW=]L`02LTd020L000H73`004g4HRff;QX^6YHJUWjBOY8JTQHX0
@`008G4HLT]bCg=?Le1hD7Q@N51hD7U@NEEiEGUEN5EhD7Q?M`03000DLF^:JhZ4XhBSWZ>NXifSWJ>M
Xh@B000ELFabKG9]LfecKW=^LficKG=]Lfeb04h020L000H740004W9]RhJUQZF;YJBUWjFOYHJTQD80
0W42LP0ILg9hLWQcN7=iN7UhNGQiN7UhNGQiN7UhN@02N09g1@02L@::00>SQJ<00j<00kZS^P06X`03
RP0000d000=aLW800W800g=bL`03L`0;S6j<KXa^S6ecKG80C`081`001PL@000B679]RhJ<QZF6YInU
WjB6Y8F;@@008G4HLQQbCg=?N4mhD7Q@N51hD7U@NEEiD7UEN51hD7Q?M`07000CLF^:Q8Z4XhBSWZ>M
XifSWJ>4RP0?000ELFabKG9]LfecKW=^LficKG=]Lfeb050020L000H74000279]RhJ<QZF60jD01inU
WjF6Y8D0@002L@Ab01McLWQcN7=iN7UhNGQiN7UhNGQiN7UhN@02N003MgP000L000M/LG6:QJ>500BS
00>jXkX01:<2RPd000IaLW9]Lg86L`0:S6j<KXa]LfecKE0020L000H740004aQbK8^6RhJUQZFOYInT
QZB5Rf`0?P008G4HLAQbBg9?LdmhD7Q@N51hD7U@NE1iEGU@N51hD7Q?M`0;000BLF^:Q:>4XifSWJ>M
XifSQ:>430005g4HLVebKG=]LficKW=^LfecKG=]LQQa050020L000H74000279/RhJ<QZF;0jD029nU
WjF6Y8J:?P000g5bL@02LP0ELg9hLgQcNG=iN7UhNGQiN7UhNGQi00Ah0WL00g1aL@0:L@04RXFSQ@BS
00>jXkX01Z<00hX00009000:K79bLg:<LhacS0=c00V<KXa^LfecKG80D@081`001PL@000C679/RhJ;
QZF6YInUWjB6Y8F;K00n000oLAQbBg9?LdmhD7Q@N51hD7U@NE1iEGQEN51hCgQ?Mg1aL75`LG1aL75`
LG1aJg5[LF]aJg40LIfSWJ>MXhB:00X001HHLVabKG=]LficKW=^LficKG=]LVebDP081`001PL@0008
LFb;QXb6YHH3Y@08Y:FOYInTQXXl00=a0W805G=bN79hLgQcNGQiN7UhNGQiN7UhN@04N09g01U`Sg2?
LHmaSg6?LHmaSg6?L8m`RW2:L75`00Ia00>SRXX020000fabK@02LP0AS6j<LhacS7><Lha^LfecKG80
D`081`001PLA000BLVb;QX^6YHJUQZFOY8JTQX]/?000AG4HLQQbCg=?LdmhD7Q@N51hEGU@NE1hD7QE
N4mhCgM`MW1fL7I`MW1fL7I`MW1fL7I`MW1fL7I`LG1aL75`LG1aL75[L@04000F679/LVecKG=^Lfic
KW=^LfecKG9]LU@020L000H74000275/RhJ<QZF60jD02:BUWjFOY8J:>`02L@Ab01=cLWQcN7=iN7Uh
NGQiN7UhNGQi00Ah02UbSg2?L8meSgJ?MXmfSgJ?MXmfSgJ?MHmfSg2?LHmaSg6?LHmaSg5fL003L@0F
L75aLVf<LhacS7><LhacS7><KHa]LUD020L000H74@004W9/RhJ;QXb6YHJUWjB6Y8J;QCP0061a074H
LQQbCg=?Le1hD7Q@N51hD7UEN5EhEGQEN4mgL8m`MWF?MHmeSgF?MHmeSgF?MHmeSgF?MHmeSgF?MHme
SgF?L8m`Sg1fL7I`MW1fL7I`LG1aL75[L@2<KG=]LVaG00P700061a4000M/RhJ;QZF600>U00ROYInU
QZB6RcL000AaJg5/17801W=bN7=hL`9h00]iN7UhNGQiN7UhN@02N013MgQfSgJ?MXmfSgJ?MHmeSgF?
MHmeSgF?MXmfSgJ?MXmfSgF?MHmeSgF?MHmfSgJ?MXmfSgJ?L8m`Sg2?L7I`LG1bK@1H00P700061a40
019bK8^6RhJ<QZF6YInUQZB6Y8Df001TL@1a678HLTmbCg=@N51hD7Q@N51hD7Q@N5EhD7Q?N4n?MHme
SgF?MHmeSgF?MHmeSgF?MHmeSgF?MHmeSgF?MHmeSgF?MHmeSgF?MHmeSgF?MHmeSg2?L8m`MW1fL7I`
LG1aJeD020L000H74@004fabQX^6YH^US:FOYInUWjB6Y8D0=0001W5`LFabK09b00QcLW=bN7=hL`9h
00UiN7UhNGQiN7T00WP0B7MhLXmfSgJ?MXmfSgJ?MIAeU7FDMIAeU7FDMYAfU7JDMYAfU7FDMIAeU7FD
MIAfU7JDMYAfSgJ?MHmeSgF?MHmeSgEfLGIaMPEaD0081`001PLA000CLFb;KH^6S8JUQZFOYHJTQZB5
RP0c001_L@1a074HLQQbCg=?Le1hD7Q@N51hD7Q@N51hD7Q?N4n?MGIeSgEfMGIeMWF?MHmeSgF?MHme
SgF?MHmeSgF?MHmeU7FDMIAeU7FDMIAeSgF?MHmeSgF?MHmeSgF?MHmeMWEfL7I`MW1fL7I`L@1a04`0
20L000H74@004fabKHb6S8JUS:FPYInUWjF6Y8D0<P001W5`LG1a60Ab00mcLWQcN7=hLgUhNGQiN7T0
17P0BWMhLXmeSgJ?MXmfSgJ?MXmeU7FDMIAeU7FDMIAfU7JDMYAfU7JDMIAeU7FDMIAeU7JDMYAfU7JD
MYAeU7FDMIAeU7FDMHmfSgJ?1WH0271fL7I`LG1aA`081`001PLA000CLFb;KH^6S8JUQjFOYInUQZB5
R`0a001iL@1aL74HLQQbBg9?LdmhD7Q@N51hD7Q@N51hD7Q?N4mfMGIeMWEfMGIeMWEfMHmeSgF?MHme
SgFDMIAeU7FDMIAeU7FDMIAeU7FDMIAeU7FDMIAeU7FDMIAeU7FDMIAeU7FDMGIeMWEfMGIeMWEfL7I`
MW1fL75`L@1400P700061a80019bKH^6S8JUS:FPYInUWjF6Y8D`0008LG1aL74HLW42LP08Lg9cLWQc
N7<2N003NGQi00Ih00QgN79fMHmeS`=f00>?MWH00gH0CHmeU7FDMIAeU7FDMIAfU7JDMYAfU7JDMIAe
U7FDMIAfU7^DNiAkU7^DNiAeU7JDMIAeU7FDMIAfU7JDMYAfU7JDMHmeMWEfMGIeMW5f00Aa?`081`00
1PLB000B679]RhJ<QZF7YHJUWjF6Y8JT;`00QG40LG1aL74HLQQbCg9?N4mhD7Q@N51hD7Q@N51hCgQ?
MWEfMGIeMWEfMGIeMWEfMGIeMWEfMHmeSgEfMGIeNgEkMIAeU7FDMIAeU7FDMIAeU7FDMIAjU7ZDMIAe
U7FDMIAeU7FDMIAeU7FDMIAeU7FDMGIeMWEfMGIeMW1fL7I`LG1a0740>P081`001PLB0008LFf;QXb6
YH`3Y@08WjFOYHJTQXX]000:LG1aL75`LFabC`9b00IcLWQbN7<;N009MgQbMWEfMGIe00]f04YeMWFD
MIAeU7FDMG]fU7JDMYAfU7JDMIAeU7FDMIAeU7^DNiAkU7^DNiAkU7^DNiAkU7^DNYAkU7^DNiAkU7JD
MIAeU7FDMG]eN`Qf00IaMW5aL74g00P700061a<0019aK8]]S8J<QjF6YInUWjB6Y8D/002AL@1aCG5=
LG1b679?LTmcCg=?N51hD7Q@N51hCgQ?MdifCGI=MTefCGI=MTefCGI=MTefMGIeMWEfMGIeMWEfMGIe
MWEfMG]eNgEkMG]eNgEkMIAeU7FDMIAeU7ZDNYAjNgZDNYAjU7ZDNYAjU7ZDNYAjU7YkNW]eNgEkMG]e
NgEkMGIeMWEfMGI`MW1fCG5=L@0`00P700061a<000LHLVf<QZF600FU00NOYInUQZA/02X000YaJg5`
LG1fL7Ia17801W=bN7=hL`Uh0gL00g40000<0005CGI=MW402gH0BGEfMG]eNgIkMG]fNgJDMW]fU7JD
MYAjU7^DNiAkU7^DNiAkU7^DNiAkVG^INiUkU7^DNiAkU7^DNiAkU7^DMIAfU7FDMIAeNgD02GH01g5f
L75`LF/0:P081`001PLD000B679]RhJ<QZF7YInUWjB6Y8F::@008740L@1aCG5`MP1bBg9?LTmcCgQ?
N51hD7Q@N4mhCgQ>3000LG40L@1a0740L@1aCG5=LDefCGI=MTefCGIBMWEfMGIeMWEfMGIeNgEkMG]e
NgEkMIAeNgYkNW]jNgYkNW]jU7ZDNYAjU7ZDNYAjNgYkNW]jNgYkMG]eNgEkMG]eNgEfMGIeMWEfMGI=
MTefCG5=L@1a02H020L000H75000279]S8J<QZF<1JD01InUWjB502X000UaL75`MW1fL7400g8027=b
Lg9hLgQc1gP00gMhM`0;L@09CG5=LDiaCW5>009a019fCWIaMTifLGI>Me=gCGIaMWD8MP0gNgIkMW]e
U7JDMYAfU7^DNiAkU7^DNiAkU7^INiUkVG^INiUkVG^INiUkVG^DNiAkU7ZDNiAjU002N`0:MG]fNgIk
MW]fN`9f00UeMWEfL7I`MW00174R00P700061a@001<HLVf;KHb6YHNUWjFOYInTQH]/02P00:Ea075=
LG1fL7H0LT]bCg9?LdmhCgQ?N4mhCgQ?N4mgCG40LDeaCG5=LDeaCG5=LDeaCGI=MTifCWI>MTifCWI>
MdigDgMCL@1aCGI=MU9fMGIeMWEfMG]eNgEkMG]eNgEkMG]jNgYkNW]jNgZDNYAjU7ZDNW]jNgYkNW]j
NgYkNW]jU7YkNW]eNgEkMG]eNgEfMGIeMU9fDWI=MTefCG5=L@1a07407P081`001PLE0007K8]]S8JU
S005Y@06WjFOY8J;:0002W1aL7I`MW1fL743LP08Lg9hLWQcN7<5N004MgQgM`]a00afCWI>MTifCWI>
MU<2MP0BMe=gMWMCMgIgDgMCMe=aK75[0XX01Z>:Xg5fL@Af00akMG]fNgEkMW]fNgH8N`0PU7^DNiUk
VG^INiUkVG^INiUkVG^INiUkVG^INiAjNgX7N`0BMW]fNgEkMG]eMWEfMGIaMW5f1G4L00P700061aD0
01=bK8]]S8JUQjF6YInUWjB6Y8F:02L00:MaCG5=MW1fL7I`LT]bCg9?LdmhCgQ?N4mhCgQ?MdiaCG5=
LDeaCG5=LDefCWI>MTifCWI>MTifDgICMe=gDgMCMe=gDg5/LVaaJhZ4RXBSQ:>MXiefCGI=MU9fDWIe
MWEkMG]eNgEkMG]eNgYkNW]jNgYkNW]jNgYkNW]jNgYkNW]jNgYkNW]jNgYkNW]jNgEkMG]eNe9kDWIB
MU9fDWI=MTefCG5=L@0L00P700061aD000LHLVf<QZF<00NU00JOYHJTQHXV000:L75`MW1fL7IeMP=b
00IcLWQbN7<3N004MgQgN0Ag02]>MW5fLGIaMW5fLGICMU=gDgMCMe=gMWMfMgIgDgMfMeAlE7a[LFb;
KHa[00B:00>SRZ<01:<01J:SXZ=`00=f00=eMWH00WH01W]fNgIkMPAk00VDNiAkU7^DNi@00g/04X1k
P7^0Nh1kVG^INiUkVG^0NP]k00YeNgEkMG]eNgEk1WH0175fLG4K00P700061aH001<HLVf;QXb7YHNU
X:FOYHJTQH^502H00:UaCG5`MW1fL7I`LT]bCg9?LdmhCgQ?N4mhCgM?MdefCGI=MTifCWI>MTifCWI>
MdigCWMCMe=gDgMCMe=gDgMCM`1aK79]Rff<JhZ4RXBSQ:>MXifSWJ>4XhAbDgaCMTefDWIBMWEfMG]e
NgEkMG]eNgEkNW]jNgYkNW]jNgYkNW]jNgYkNW]jNgYkNW]jNgYkNW]eNe9kDW]BNe9fDWIBMU9fCGI=
MTea01X020L000H75P00279]S8JUS:F<1JD01j2UWjB6Y8D090002g5`MW1fL7IeMWEf00Eb00]hLWQb
N7MhMgQgN002M`07LGIaMW5fL@06MP0<Me=gDgMCMe=gDgMC0gL04UAlMgaDO7MaK79/Rff<KHa]L@>:
00>SRZ<01:<03j:SXZ>9XeAlE7aDO5AlL@06MP08NgIkMG]fNgD=N`0>NYUkVGZINiUjP7^0Nh08N`0<
NW]jNgEkMG]eNgIk27H00dd0000H00P700061aL001=bKH^6S8NUQjFPYInUQZB6Y8F:02D00:YaCGI`
MW1fL7I`LT]bCg9?LTmhCgQ?N4mgCgM>MTifCWI>MTifCWI>MTigDgMCMe=gDgMCMe=gDgMCMe=gDgM/
LVb;KH]]S6f<JhY[RXB:Q:>MXifSWJ>4XhB:DgMCMe=gDgMCMe=gCGIBMU9fDWIBNgEkMG]eNgEkNW]j
NgYkNW]jNgYkNW]jNgYkNW]jNgYkNW]jNe9kDW]BNe9kDWIBMU9fDWIBMTefCAT020L000H76000279]
S8JUQjF70jD02:2UWjFOY8J;8`005g5`MW1fMGIeMWEf679?LTmhLWQ?N79h00Ig0gH0:7MCMe=gDgMC
Me=gMWMCMeAgE7MDMeAlE7aDO5AlK79/Rff<KHb6S6h2S008LF^:Q8Z4Xh@9X`0:RWM>MTifCW5>L@`0
00eaMWIfMG]eNgEkMG]e00ek00UjNgYkNW]jNgX02G/03WIkMG]eNgEkMGIeMW1f6@081`001PLH000D
679]S8J<QZF7YHNUWjFOY8JTQHXQ001NL@1aL7I`MW1fMGIeMQQbCg9?MdmgCgM?MdmgCWM>MdifCWM>
MdigDgMCMe=gDgMCMe=gDgMCMe=gDgMCLFabK8]]Rff<KHa^S6j<075[RXB:Q:>4XifSWJ>MXhB:Q1P0
03IfCGIBMU9fDW]BNgEkMG]eNgYkNW]jNgYkNW]jNgYkFG]INeUkFG]BNe9kDW]BMU9fDWIBMTdI00P7
00061aT001A]S8J<QZF7YHbUX:FOYInUQZB5RR4001IaL7IeMWF?MHmeSgEbCW9?N4mhCgQ?1gL0=GIg
MWMCMe=gDgMCMe=gE7MDMeAgE7MDMeAlE7aDLFb;KH]]S6f<Qhb7S8N<Qhe[RXB:Q:>400FS00JRXj>S
RHXM0004MWEfM@9f00QkMW]fNgIkMP=k00UjNgYkNW]jNgX02g/037EkMG]BNgEkDWIeMQT020L000H7
6@005G5]Rff<QZF7YHNUX:FOYInTQX^5RP0O001NL@1fL7I`MWEfMGIeSg1b679?MdmgCgM?MdmgCWM>
MdigCWM>MdigDgMCMe=gDgMCMe=gDgMCMe=g075/LVb;KH]]S6f<KXa^S6j<KW5[RV^:Q:>4XhBSWJ>M
XhBSQ28002]fDWIBMU9fDW]BNe9kDW]BNe9kNW]INeUkDW]BNe9kDW]BNe9kDWIBMU9f01X020L000H7
6P00279]S8JUQjF70jD02Z2UWjFOYHJTQHXO000FLG1fMHmeSgF?MHmeS`1bCg9?MdmgC`Ug00]fMe=g
DgMCMe=gD`03M`0GE7MDMeAgE7MCMfabK8]]Rff<KHb7S8L00X`00hf7S002S009KW5[RXBSQ:>:00NS
00>:XhT090000g5fMP05MP0:NgEkMG]eNgEkM@Mk00mfNgIkMG]eNgEkDW]eMWD06P081`001PLJ000E
679]S8J<QZF7YJ2UWjFOYHJTQX^501h000iaCGI`MWF?MHmeSgF?L08004mbCW9?MdmgCgM?MdigCWM>
MdigCWMCMe=gDgMCMe=gDgMCMe=gCW5/LVb;KH]]S6f<KXa^S6j<KXa^S6j<075[RXB:Q:>4XifSWJ>4
XhB:02T002=aCGIBMU9fDW]BNe9kDW]BNe9kDW]BNe9kDWIBMU9fDWI=MP0J00P700061aX000YbKHb6
S8JUQjF<0jD02J2UWjFOY8JTQ@0N000=LG1fMHmeSgF?MHmeS`040007Cg9?MdmgLP09M`0QE7MCMe=g
DgMDMeAgE7MCMdigCW5/Rff;KHa]S6f<Qhb700><00n7S8b<Qha]LF^:Q8Z4Xh@01j<00hVSRP0^000M
MU9fMGIeNgEkMW]fNgIkMW]eNe9kMG]BMU9fDWH06`081`001PLK000FLVf;KHb6S8NUX:FPYJ2UWjFO
Y8J;Jah000afL7IeSgF?MHmeSg06000KLTigCgM?MdigCWM>MdigCWM>MdigCWM>Mdig00D002]aK8]]
Rff;KHa]S6f<KXa^S6j<KXa^S6f<KG5[RXB:Q:>4Xj:SWJ>MXhB:03<001QfCGIBMU9fDWIBMU9fDWIB
MU9fDWI=MTdK00P700061a/000XHLVf<QXb7YHNV1:D02J2UWjFOY8J;Q@0L000=LG2?MHmeShj?SXme
S`08000CCW9bMg9gLWMbMg9gCWM>MdigCP08000@LFabK8]]S6f<KHb6S8N<Q`N<00f7S6j<KG9[RV^:
Q:>900VS00>90000=@000g5fMP07MP09DWIBMU9fDWI=01`020L000H770005aQbKHb6S8NUQjFPYJ2U
X:FOYHJTQHY[01`000afL8meSgF?MHmeSg0?0004CW9>LPl0031aK79/Rff;KHa]S6f<KXa^S6j<KXa^
S6j<KHa]LQQaJhY[RXBSQ:>MXj:SWJ>4RV/m000:CGI=MTefCGI=MQh020L000H77@0036abKHb6S8NU
S:JUYP:U00^PYInUWjB6Y8F:Q@0H000>LG2?MHmeU8jDSYB>Sg0M0009K8]/Rff;KHb600Z<00b=S8f7
S6j<KHa]S6d4009a0XX00j>:X`04X`06XZ>RXhV:I@081`001PLN000I679]S6f<QjF7YHNUQjFPYHJT
QZB6RhF:J`0H000=MW2?MHmeSgF?MHm`S`0K000PLF^;K8]]Rff;KHa]S8N<Qha^S6j<KXa^S6f<KHa]
LQP6000@LF^:Q8Z4XhBSQ:>4XhBSQ6D020L000H77`0036abKHb7YHNUS:JUYP:U00bPYInUWjF6Y8J;
QHXE000>LG2?MHn>U8jDSYB>Sg0H000@LFaaK8]]Rff<QXb6S8J<Q`J<00>=S8`00X`01Vj<KHa]LPL0
0W44RP0;XhZSXZ>RXj:SRJ<0I@081`001PLP000I679]S8J<QjF7YHNUQjFPYHJUQZB6RhF;Q@0F000=
MW2?MHmeSgF?SXmeS`0F000RJg5/LFb;K8]]Rff<KHa]S6j<Qhb7S8N<KXa^S6f<KG9]LPT0011aJhZ4
RXBSQ:>4XhBSQ:>4I@081`001PLQ000<KG9]S8NUQjF<YZFV0ZD03:2UWjFOYHJTQZB5RQ@000i`SgF?
SYB>U8jDSXn>LA8000E/LG6;K003R`0;QX^6S8J<QXb7S8L00X`00hf<S@06S005KXa]S6d02`000g6:
RP02RP03XhZS00:S00JRXj:SRJ=U00P700061b8001`HLVf<QXb7YHNUQjF7YJ2UQZF6Y8JTQX^5RV]a
4@003W5`SgF?MHn>Shj?SXm`4@008faaK8]/Rff;KH]]S6f<KHa^S8N<Qhb7S6j<KXa^S6f<KG8H00`0
011aJhZ4RXBSQ:>4XhBSQ:>4I@081`001PLS000<KHb6S8NUQjF<YZFV1:D03J2UWjF6YHJTQH^5RV`0
3`003W2?MHn>U8jDSYB>Shj?3P004faaK8]]RhJ;QXb6S8J<QXb7S8L00X`00hf<S@08S004KHabLPh0
0W43RP0;XhZSRJ>RXj:SRJ<0I@081`001PLT000OKH]]S8J<QjF7YHNUQjF7YHJUQZB6Y8F;QH^5RVb:
J`0<000?LG2?MHmeShjDSXn>Sg1a00T002MaJg5/RVb;KH]]Rff;QXb6S8J<KHa^S8N<KXa^S6j<KXa]
S6ebKG803`004G5[RV^:Q:>4XhBSQ:>4XhB:06@020L000H790002g9]S8J<QjF7YHbV00JU012PYInU
QZF6Y8JTQZB5RhEa2P003g1fMHn>U8jDSZf>U8j?R@05000AK8Y/Rfb;K8]]RhJ<QXb;S8H00h`00hN<
Q`02S007SHb=S8f<S@02S006Qhb<S6f<0W8A009a0XX03:>9XhZSXZ>RXhVSRV@020L000H79@009W9]
S6f<QXb7S8NUQjF7YHNUX:F6YHJTQZB6RhF;QH]/RV^:Jg5[1000?75`MW2?MHn>U8j?SXn>S`1aJg5/
RVb;K8]/Rff;KH^6RhJ<QXb6S8N<Qhb7S8N<KXa^S6j<KHa]RfebK1@000maJhZ4XhBSQ:>4XhBSQ8X0
I0081`001PLV000=LVf<QXb7S8NUS:F<YP04Y@0kX:F6YHJUQZF6YHJTQX^5RhF;QHZ5RVb:K75[LFaa
K8Y/RVb:K8Y/RVb;QH^6RhJ;QXb6S8J<QXb6S8L028`03hf<SHb=Qhb7S6j<LXa]LP0E009a0hX02hVS
RJ>RXj:SRJ>:06@020L000H7:000GAQbKHa]S8J<Qhb7YHNUQjF7YHNUQZF6YHJUQZB6RhJ;QX^6RhF;
QH^5RhF;QH^5RhJ;QH^6Rff;QX^6RhJ;QXb6S8J<QXb7S8N<Qhb7S8N<Qhb7S6j<KXa]S6ebK00J000?
LF^:Q8Z4XhBSQ:>4XhB:06@020L000H7:@00CaQbKHa]S8NUQjF7YXbVS:F<YHbUX:FPYHJUQZF6YHJU
QZF6YHJTQX^6RhJ;QX^6RhJ;QXb6S8J<QXb6S8J<Qhb7YHbUS:F<YXbVS8f<SHL00h`028N<Qha]S6eb
70000f^:R@02RP03XhZS00BS00>:XhT0I0081`001PL[001HLVebKHb6S8N<QjF7YHNUQjF7YHNUQjF6
YHJUQZF6YHJUQXb6RhJ;QX^6RhJ;QXb6S8J<QXb6S8J<QXb7S8N<Qhb7S8N<Qhb7S8N<Qha^S6j<KHa]
LVeb61`0011aJhY[RXBSQ:>4XhBSRHY[H`081`001PL]000CKG9]S8J<QjF7YHNVS:F<YHbUS003Y@0n
X:F6YHJUQZF6YHJUQZF6YHJUQZF6YHJUQZF6YHJUQZF7YHbUS:F<YHbVS8f<SHN<Qhb7S8N<Qha]S6eb
KG4N0003JhY`00::00^SRZ>:Xj:SRZ>9RP1S00P700061bh0054HLVf<KHb7S8N<Qhb7YHNUQjF7YHNU
QjF7YHJUQZF6YHJUQXb6S8J<QXb6S8JUQZF6S8N<Qhb7S8N<Qhb7S8N<Qhb7S6j<KXa]S6f<KG9]LQP0
8000475[RV^:RJ>9XhVSRJ>9RV]S00P700061bl003Y]LVf<QXb7S8NUS:F<YHbVS:F<YHbUQjF7YHNU
QjF7YHNUQjF7YHbUQjF7YHNUQjF7YHNUS:F<YHbU1X`03HN<Qhb7S6f<KHa]LVd09000275`RXZSRZ>:
0j<018ZSRHYS00P700061c0004XHLVf<KHa]S8J<Qhb7S8N<QjF7YHNUQjF7YHNUQjF7YHNUQjF7YHNU
QjF7YHNUQhb7S8N<Qhb7S8N<Qhb7S6j<KHa]S6f<KG9]LBD0011aJhY[RXV:RJ>9XhVSRHY[H`081`00
1PLb000<LVf<KHa]S8N<Qhb70X`09:F<YHbVS:J<YXNUQjF7YHbUS:F<YHbUS:F7YHNUQjF7YHNVQ`><
0167S8N<Qhb7S6j<KHa]S6ebK00X000?LG2:RZ>9XhZSRJ>:XhV:06<020L000H7<`00@W4HLVf<KHa]
S8N<Qhb7S8N<Qhb7S8N<QjF7YHNUQhb7S8N<Qhb7S8N<Qhb7S8N<Qhb7S8N<Qhb7S6f<KHa]Rfeb62X0
011aJhY[RXV:RJ>9XhVSRHY[H`081`001PLf000<LVebKHa]S8N<Qhb70h`0;XNUQjF7YXNVS:J<YHbU
S:F<YHbUQjJ7YXNUQhb7S8N<Qhb7S8N<Qha]S6ebKG8]000?LG2:RZ>:XhZSRZ>9XhV:06<020L000H7
>`00<g8HLVf<KHa]S6f<KHb7S8N<Qhb7S8N<Qhb7S8N<KXb7S8N<Qha]S6f<KHa]RfebKG8HL@0b000?
Jg5[RXB:Q:>4XhBSQ8Z406<020L000H7?P00779]Lff<KHa]S8J<Qhb7S8N<Qhb7S8N<Qhb7S8L3S00<
Qhb<S6f<LXa]LVeb=`02L@0;RXVSRJ>9XhVSRJ<00XYS00P700061d4002EbBg9]LVebKH]]S6f<KHa]
S6f<KHa]S6f<KH]]Rff;KG9]LQQa03/000iaJhZ4RXBSQ:>4XhB:Q6<020L000H7AP00879]LVebKG=]
S6f<KHa]S6f<KHa]LVebKG9]LVb]SZR9>P02L@0;RXVSRJ>9XhVSRJ<00XX00g40001P00P700061dh0
01THLQQb678HLQQaL8n>Shj?SZf>Z8jXSXm`03X000maJhZ4RXBSQ:>4XhB:Q740HP081`001PMH000@
LHj?Sjf?[Hn]Yjf>Z8j?L3P00W402hZ9RXVSRJ>9XhVS00::00=a0000H0081`001PMI000?SgF?SXn>
Z8jXSZR>ShV?03T000iaJhZ4RXB:Q:>4XhB:Q6<020L000H7F@000g2?SP02S`0;[Hn]Z:f>[HjXSXl0
=`02L@0;RXV:RJ>9XhVSRJ<00XYS00P700061eT0015aL8n>Shj?SZR>Z8jXSXn9L@0g000>LF^:Q8Z4
RXBSQ:>4RXUS00P700061eT000E=MW2?SP02S`0;[Hn]Z:f>[HjXRG40=@02L@0;RXV:RJ>9XhVSRJ<0
0XYS00P700061e/0011aL8n>ShjXSZR>Z8jXRHm`=P003W5[RXV:RHZ9XhVSRHY[H`081`001PML000A
MW2?SZR>[Hj]YjR>Z8jXRHX0<`000g5`RP02RP0:RJ>:XhVSRZ>9RV<020L000H7G@004GI`Shj?SZR>
Z8jXSZR>ShV:03<000iaJhZ9RXV:RHZ9RXV:Jf<020L000H7G@004G2?MHn>[Hn]YjfWZ8jXSZR90380
00AaL8Z90XX02J>9XhZSRJ>9RP1S00P700061eh0011`Sg2?SXn>Z8jXSZR>Z8V?<@00475[RV^:RHZ9
RXV:RHZ9RV]S00P700061eh0015aL8n>ShjXSjfW[JNXSZR>S`0a0003LG2:00::00J9XhZSRJ<2RP03
RHX0068020L000H7G`004G5`Sg2?SZR>Z8jXSZR>ShV:02l0011aJhY[RXV:RHZ9RXV:RHY[H`081`00
1PMP000ALG2?SZR?Z8nXYjRWZ8jXRHX0;`00175`RXT3RP07RJ>:XhV:R@1T00P700061f40015aL8n>
ShjXSZR>Z8jXRHn9RP0]000?LF^:JhZ9RXV:RHZ9RXV:06@020L000H7H@004TefL8n>Z8nXSZRWZ8jX
SZR9RR/000QaJg5`RXV:R@::00JSRJ>9RW1T00P700061f<0015aL8n>ShjXSZR>Z8jXRHn9RP0[000?
LF^:JhZ9RXV:RHZ9RXV:06@020L000H7I0004Xm`ShjXSZR>Z8jXSZR>Z8V?RBP00W403HYaRXV:RJ>9
XhVSRHX0I@081`001PMT000CCHm`ShV?RJR>Z8jXSZR9ShV:L00V000@L@1aJhY`RXV:RHZ9RXV:L6D0
20L000H7I@00572?RHn>Z8jXSZRWZ8jXSZR9ShV:9003L@>:00Z9RXVSRJR9RXV:I@081`001PMU000D
LG2?L8n9ShVXSZR>Z8jXRHn9RW0T000@LF]aJhY`RXV:RHZ9RXV:L6D020L000H7IP00575`ShV?SZR>
Z8nXSZR>Z8jXRHm`8P04L@>:00Z9RXVSRHZ9RXV:I@081`001PMW000ELG2?L8n9ShjXSZR9Z8VXRHn9
RW1a0240011aJhY`RW2:RHZ9RXV:L8Y[I@081`001PMX000ELG2?RHn>Z8jXSZR>Z8jXSZR9ShV:01h0
0g400hYaRP02RP0:RHZ9RXV:RHY`LFD020L000H7J@005G5`Sg2?RHn9ShVXRJR9Z8V?RHY`RP0M000A
LF]aJhY`RW2:RHZ9RXV:L8X0IP081`001PMZ000FLG2?RHn>Z8jXSZR>Z8jXRJR9ShV:L1X017400hYa
RP02RP09RHZ9RXV:RHY`06H020L000H7J`005g5`Sg2?RHn9ShV?RJR9Z8V?RHZ9RV]a01L001=aJg5[
LF^:L8Y`RXV:RHY`RW1a06H020L000H7K0005g5`ShV?SZR>Z8jXSZR9Z8VXRJR9RXV:01D017400hYa
RP04RP08RHZ9RXV:RHYW00P700061fh001U=RW2?L8n9ShV?RHn9ShV?RHn9RXV:JhY[010001=aJg5[
RV^:L8Y`RW2:RHZ9RW2:06P020L000H7KP00775`Sg2?RHn9Z8jXSZR9Z8VXRJR9RXV:RHY`RV/:000E
LF]aL75`RW2:L8Z9RXV:RHZ9RXV:06T020L000H7L@00>75`Sg2?RHn9ShV?RHn9ShV:RHZ9RW2:JhY[
RV]aJg40LF]aJg5[LF^:JhY`RW2:L8Y`RW2:L8Y`J@081`001PMa000gL75`ShV?RHn9Z8VXRJR9Z8VX
RHn9RXV:RHY`RW2:L75`LG1aL75`RW2:L8Z9RXV:RHZ9RXV:L01Z00P700061g<003EaL8m`ShV?RHn9
ShV?RHn9RXV:RHY`RV^:JhY[RV]aJg5[LF]aJhY[RW2:L8Y`RW2:L8Y`L@1Z00P700061g@003=aL8m`
ShV?RHn9Z8VXRHn9ShV?RHZ9RXV:L8Y`RW2:L8Y`RW2:L8Y`RW2:RHZ9RXV:L740J`081`001PMe000a
LG2:L8n9ShV?RHn9ShV?RHZ9RXV:L8Y`RW2:L8Y`RV^:L8Y`RW2:L8Y`RW2:L8Y`L@1/00P700061gH0
031aL75`ShV?RHn9ShV?RHn9ShV:RHZ9RXV:RHY`RW2:L8Y`RW2:L8Z9RXV:RHY`LG1/00P700061gL0
02iaCG5`RW2?RHn9ShV?RHn9RXV:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L75[K@081`001PMi000[
L75`ShV?RHn9ShV?RHn9RXV:RHZ9RXV:RHY`RW2:L8Y`RW2:L8Y`RW2:L01^00P700061g/002QaL8Y`
RW2:L8m`ShV:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`RW2:L8Y`K`081`001PMm000?CG5`LG2?RHn9RXV:
RHZ900>:011`RW2:L8m`RW2:L8YaRW2:0W5a00P700061gl0025a075`LG2:L8Y`RW2:L8Y`RW2:L8Y`
RW2:L75`LG1aL740LP081`001PN1000JCG5`LG2:LHY`RW6:L8Y`RW2:L8m`Sg2:LHX3LG@020L000H7
P`006G5=LG1aL75`LG1aL8Y`RW2:L75`LG1aL740MP081`001PN6000DLG1aL75`LG2:L8Y`LG1aL75`
LG1h00P700061hT000ea075=LDeaCG5=L@1a07`020L000H7T@000dea001n00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Textures", "Subsection",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  "Textures can be given as an option to the",
  StyleBox[" MVShow3D[]",
    FontFamily->"Courier"],
  " command. Possible sources of textures are external PNG files, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " graphics objects, and ",
  StyleBox["RasterArray[]",
    FontFamily->"Courier"],
  " objects."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  "MathGL3d provides several texture mapping transformations. The simplest \
one is the ",
  StyleBox["MVPlaneMapping ",
    FontFamily->"Courier"],
  StyleBox["transformation. For planar mapping it is assumed that the ",
    FontFamily->"Times New Roman"],
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " coordinates of the",
  StyleBox[" polygons are in  ",
    FontFamily->"Times New Roman"],
  Cell[BoxData[
      \(TraditionalForm
      \`x\  \[Element] \ \([0, 1]\)\  \[And] \ 
        y\  \[Element] \ \([0, 1]\)\)]],
  ", and ",
  Cell[BoxData[
      \(TraditionalForm\`\((x, y)\)\)]],
  " will be used to calculate the texture coordinates ",
  Cell[BoxData[
      \(TraditionalForm\`\((x, y)\) \[RightTeeArrow] \((x\_t, y\_t)\)\)]],
  ". One translation, one rotation and one scaling can be used, to bring the \
",
  Cell[BoxData[
      \(TraditionalForm\`x\ y\)]],
  " coordinates into that range. These transformations are passed to MathGL3d \
by the ",
  StyleBox["MVTranslateTextureTarget",
    FontFamily->"Courier"],
  ", ",
  StyleBox["MVRotateTextureTarget",
    FontFamily->"Courier"],
  " and ",
  StyleBox["MVScaleTextureTarget",
    FontFamily->"Courier"],
  " options."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell["Textures for Unconnected Polygons", "Subsubsection",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(Needs["\<Graphics`Shapes`\>"]\)], "Input",
  CellLabel->"In[26]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; \n
    MVShow3D[\n\t\ \ 
      Graphics3D[\n\t\t\ \ 
        Polygon[{{0, 0, 0}, {1, 0, 0}, {1, 1, 0}, {0, 1, 0}}], 
        ViewPoint -> {0.401, \ \(-0.741\), \ 1.229}], \n\t\ \ 
      MVTexture -> "\<`ChessBoard`\>", \n\t
      MVTextureMapType\  \[Rule] \ MVPlaneMapping, \n\ \ 
      MVTranslateTextureTarget\  \[Rule] {0, 0, 0}\n]\)], "Input",
  CellLabel->"In[2]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVScene3D[] \[DoubleLeftArrow] MVTexture2D[1], 
      MVPlaneMapping, {256, 256, MVRGBTexture}}\)], "Output",
  CellLabel->"Out[2]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVPasteGraphics[]\)], "Input",
  CellLabel->"In[3]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000BFCFDFBFCFE0BFCFDFBFCFDFBF7F80BF0000BF0000BF7F8F\
BFCFDFBFCFDFBFCFDFC0D0E0BFCFDFBFCFDFBFCFDFC0D0E0BFCFDFBFD0DFBFCFDF
606F6F5F606F000000000000000000000000000000000000000000000000000000000000000000\
2F00009F00002000002F00000000000000000000000000000000009F0000900000
C00000BF0000500000200000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000BFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBF2F2FBF2F30BF0000BF7F80BF7F80\
BFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDF
BFCFDF5F606F000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F00009000009F00002000000000000000000000000000009F0000
900000BF0000BF0000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000BFCFDFBFCFE0BFCFDFBFCFDFBFCFDFBFCFE0BFCFDFB0C0D0BF0000BF0000BF0000BFCFDF\
BFCFDFBFD0E0BFCFDFBFCFDFBFCFDFBFD0E0BFCFDFBFCFDFBFCFDFBFD0E0BFCFDF
BFD0DFBFCFDF606F6F000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000009000002F00007000000000000000000000000F0F0F
C00000BF0000C00000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
B0CFDFBFCFDFBF0000BF505FB05050BFCFDFBFCFDFBFCFDFB0CFDFBF2F2FBF0000BF0000BF7F80\
BFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDF
BFCFDFBFCFDFBFCFDF5F606F000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002000009F00009000002F00000000002F0000
BF0000BF0000BF0000700000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000BFCFDFBFCFDF
BF0000BF0000BF0000BFCFDFBFCFE0BFCFDFBFCFDFBFCFDFBF7F8FBF7F80BF0000BF0000BF7F8F\
BFCFDFBFCFDFBFCFDFBFD0E0BFCFDFBFCFDFBFCFDFBFD0E0BFCFDFBFCFDFBFCFDF
BFD0E0BFCFDFBFCFDFBFCFDF5F6F6F000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005000004F00009F00004F0000
500000BF0000C000004F0000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B0CFDFBFCFDFB00000BF0000
B00000BF505FAFC0D0BFCFDFBFCFDFBFCFDFB0CFDFBFCFDFBFCFDFBF2F30B02F2FBF0000BF0000\
BFCFDFB0CFDFBFCFDFBFCFDFBFCFDFB0CFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDF
BFCFDFBFCFDFBFCFDFBFCFDFBFCFDF5F606F000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000004F00009F0000
700000BF0000BF0000BF00004F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000BFCFE0BFCFDFBFCFDFBF0000BF0000BF0000
BF0000BF0000BFA0B0BFCFDFBFCFDFBFCFDFBFCFE0BFCFDFBFCFDFBFA0AFBF0000BF0000BF0000\
BF2F2FBFCFE0BFCFDFBFCFDFBFCFDFBFCFE0BFCFDFBFCFDFBFCFDFBFD0DFBFCFDF
BFCFDFBFCFDFBFD0DFBFCFDFBFCFDF5F606F5F6F6F000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F4F0000
BF0000BF0000C00000BF00002F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000BFCFDFB0C0DFBF0000B00000BF2F2FB00000BF0000
B00000BF0000B00000BF7F80B0C0DFBFCFDFB0CFDFBFCFDFB0C0DFBFA0B0B07F80BF0000B00000\
BF0000BF2F2FBFCFDFB0CFDFBFCFDFBFCFDFBFCFDFB0CFDFBFCFDFBFCFDFBFCFDF
B0CFDFBFCFDFBFCFDFBFCFDFB0CFDFBFCFDF5F606F5F606F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F0000BF0000BF0000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000B0CFDFBFCFE0BFCFDFBF2F30B0CFDFBFCFE0BFA0B0AFBFD0B00000
BF0000BF0000BF0000BF7F80BF7F8FBFCFDFBFCFDFBFCFDFBFCFE0BFCFDFBF7F80BF7F80BF0000\
BF0000BF0000BF2F2FBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDF
BFCFDFBFCFDFBFCFDFBFCFDFBFD0DFBFCFDFBFCFDF5F606F5F6F6F000000000000000000000000\
0000002F0000000000000000000000000000000000000000000000000000000000
0000000000009F00004F0000500000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
BFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDF
B00000BF2F30B00000BF0000B05050BF505FB0C0DFBFCFDFB0CFDFBFCFDFB0C0DFBF7F80B07F80\
BF0000B00000BF2F2FB02F2FBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDF
BFCFDFBFCFDFB0CFDFBFCFDFBFCFDFBFCFDFB0CFDFBFCFDF5F606F000000000000000000000000\
2F00004F0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F10BFCFDF\
B0C0DFBFCFE0B0CFDFBFCFDFB0CFDFBFCFE0B0CFDFBFCFDFB0CFDFBFCFE0BFCFDF
BFCFDFB02F2FBF0000BF0000BF0000B02F2FBF2F30BFCFDFBFCFDFB0CFDFBFCFDFBFCFDFBF7F80\
BF7F80BF0000BF0000BF2F30BF2F2FBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDF
BFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDF5F606F000000000000000000\
6F00009F0000900000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
BFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDF
B0C0DFBFCFDFB05050BF0000B00000BF0000B02F2FBFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB0C0DF\
BF7F80B00000BF0000B00000BF2F2FB00000B0C0D0B0C0DFBFCFDFB0CFDFBFCFDF
B0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDF5F606F000000000000000000\
000000BF0000900000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000606F705F606F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000BF0000B00000000000\
0F0F10BFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDF
BFCFDFB0CFDFBFCFDFB0A0AFBF0000B00000BF0000B00000BFCFDFBFA0AFBFCFDFB0CFDFBFCFDF\
BFCFDFBF7F80B00000BF0000BF0000BF0000B07F80BF7F8FBFCFDFBFCFDFBFCFDF
BFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDFBFCFDF5F606F000000000000\
0F0F0FBF0000900000BF0000000000000000000000000000000000000000000000
000000000000000000000000000000606F70606F70BFCFDFBFD0DFBFCFDF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000B00000BF0000000000000000000000\
0000000F0F0FB0CFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDF
B0C0DFBFCFDFB0C0DFBFCFDFB09FAFBF0000B00000BF0000B00000BFA0B0B09FAFBFCFDFB0CFDF\
BFCFDFB0C0DFBFA0B0B00000BF0000B00000BF0000B05050BF505FAFC0D0BFCFDF
B0CFDFBFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFDFBFCFDF506060000000000000\
4F0000B00000BF00006F00000F0F0F000000000000000000000000000000000000
000000000000000000606F705F6F6FBFCFDFB0CFD0BFCFDFBFCFD0BFCFDF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000BF0000B00000BF00006F0000700000000000000000\
0000000000000F0F10BFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB0C0DFBFCFDFB0CFDF
BFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBF0000B00000BF0000B00000BF7F8FB0CFDFBFCFDF\
B0CFDFBFCFDFBFCFDFBFCFDFB00000BF0000BF0000BF0000B00000BF0000BF7F80
BFCFDFB0CFDFBFCFDFBFCFDFBFCFDFB0CFDFBFCFDFBFCFDFBFCFDFBFCFDF5F606F5F606F000000\
200000BF0000700000000000000000000000000000000000000000000000000000
000000606F70606F70BFCFDFBFCFDFBFCFDFBFD0DFBFCFDFBFD0DFBFCFDF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000B00000B00000B00000B00000B00000B00000200000000000000000\
0000000000000000000F0F0FB0CFDFB0C0DFBFCFDFB0C0DFB0CFDFB0C0DFBFCFDF
B0C0DFB0CFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBF2F30B00000BF0000B00000BF505FB0C0DF\
BFCFDFB0C0DFBFCFDFB0C0D0BFCFDFB07F80BF0000B00000BF0000B00000BF2F30
B0C0D0BF7F80B0CFDFBFCFDFB0C0D0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFDF5F606F6F0000\
BF0000000000000000000000000000000000000000000000000000000000000000
5F606FBFCFDFB0CFD0BFCFDFB0CFD0BFCFDFBFCFD0BFCFDFB0CFD0BFCFDFBFCFD0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000B00000BF00006F0000700000B00000BF0000B00000BF00004F0000000000\
0000000000000000001010100F0F10BFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB0C0DF
BFCFDFB0CFDFBFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB0CFDFBF2F30B00000BF0000B00000BF2F30\
B0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBF2F30B00000BF0000B00000
BF0000BF0000B0C0D0B0CFDFBFCFDFBFCFDFBFCFDFB0CFDFBFCFDFBFCFDFBFA0B06F505F700000\
6F0000000000000000000000000000000000000000000000000000606F705F6F70
BFD0DFBFCFDFBFCFDFBFCFDFBFD0DFBFCFDFBFCFDFBFCFDFBFD0DFBFCFDFBFCFDF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000B000000000000000002F00000000000F0F0F8F0000B00000B00000B000004F0000\
0000000000000000000000001010100F0F0FB0CFDFB0C0DFBFCFDFB0C0DFB0CFDF
B0C0DFBFCFDFB0C0DFB0CFDFB0C0DFBFCFDFB0C0D0B0CFDFB0C0DFBF505FB00000B00000B00000\
BF2F30B0C0D0BFCFDFB0C0DFBFCFDFB0C0D0BFCFDFB0C0DFBFCFDFB0202FBF0000
B00000BF0000B04F50BF0000B07F80BFA0B0B0C0D0BFCFDFB0CFDFBFA0B0B0202FBF2F2F5F606F\
0000000000000000000000000000000000000000005F606F606F70B0CFD0BFCFDF
B0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000B00000
2F0000000000000000000000000000000000000000000000000000900000BF0000B00000BF0000\
6F00000000000000000000000000001010100F0F10BFCFDFB0C0DFBFCFDFB0CFDF
BFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB07F80BFA0B0B00000BF0000\
B00000BF0000B0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFA0B0B07F80
BF0000B0505FBF0000B00000BF0000B00000BF7F8FB02F2FBF2F30B05050BF2F30BFCFDF5F606F\
000000000000000000000000000000606F70606F70BFCFDFB0CFDFBFCFDFBFCFDF
BFCFDFB0CFDFBFCFDFBFCFDFBFCFDFBFCFDFBFD0DFBFCFDFBFCFDFBFCFDFBFD0DFBFCFDF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000B00000B000002F0000
000000000000000000000000000000000000000000000000000000900000400000B00000B00000\
B000008F0000000000000000000000000000101010B0C0D0B0CFDFB0C0DFB0CFDF
B0C0D0B0CFDFB0C0DFB0CFDFB0C0D0B0CFDFB0C0DFBFCFDFB0C0D0B0CFDFB09FAFBFA0B0B00000\
B00000B00000BF0000B07080B0CFDFB0C0DFBFCFDFB0C0D0B0CFDFB0C0DFBFCFDF
B0C0D0BFCFDFB05050BF505FB00000BFA0AFB02F2FBFA0B0B07080BFCFDFB0CFD0BFCFDF506060\
5F606F0000000000005F606F606F70B0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDF
B0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B00000B00000900000200000
0000000000000000000000000000000000000000000000000000000000004F00004F0000BF0000\
B00000B00000B000000F0F0F000000000000000000101010B0C0DFBFCFDFB0C0DF
BFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB0C0DFBFCFDFB0CFDFBFCFDFB09FAFBFCFDF\
B00000BF0000B00000BF7F8FB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDF
BFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDF\
5F606F606F705F6F70BFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDF
BFCFDFBFCFDFBFCFDFB0CFDFBFCFDFBFCFDFBFCFDFB0CFD0BFD0DFBFCFDFBFCFDFB0CFD0BFD0DF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000B00000B00000B00000B00000B000004F0000
000000000000000000000000000000000000000000000000000000000000000000200000200000\
B00000B00000B00000B000002F0000000000000000000000101010B0C0D0B0CFDF
B0C0DFB0CFDFB0C0D0B0CFDFB0C0DFB0CFDFB0C0D0B0CFDFB0C0DFB0CFDFB0C0D0B0CFDFB0C0DF\
B02F30B00000B00000B04F50BF505FB0C0D0B0CFDFB0C0DFBFCFDFB0C0D0B0CFDF
B0C0D0BFCFDFB0C0D0B0CFDFB0C0D0BFCFDFB0C0D0B0CFDFB0CFD0BFCFDFB0C0D0BFCFDF50606F\
5F606FB0C0D0BFCFDFB0CFD0BFCFDFB0C0D0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDF
B0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000B00000B000004F0000B00000B00000B00000B00000B00000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
B00000BF0000B00000B00000B000004F0000000000000000000000101010B0C0DF
B0CFDFB0C0DFBFCFDFB0C0DFB0CFDFB0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDFB0CFDFBFCFDF\
B02F2FBF505FB00000BF0000B02F2FBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFAFC0D0
BFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDF5F606F000000\
000000606F70B0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFDF
BFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFBFCFDFBFCFDF\
B0CFD0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004000004F00000000000000000000002F0000AF0000B00000B00000B00000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000B00000900000B00000B000004000004F0000000000000000000000101010
B0C0D0B0CFDFB0C0DFB0CFDFB0C0D0B0CFDFB0C0DFB0CFDFB0C0D0B0CFDFB0C0DFB0CFDFB0C0D0\
B0CFDFB07080B00000B00000B00000A0BFCFAFBFCFB0C0D0B02F2FB04F50B0C0D0
B0C0D0B0CFDFB0C0D0BFCFDFB0C0D0B0CFDFB0C0D0BFCFDFB0C0D05F606F50606F000000000000\
0000005F606FBFCFDFB0C0D0B0CFDFB0CFD0BFCFDFB0C0D0BFCFDFB0CFD0BFCFDF
B0C0D0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0\
BFCFDF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F0000B00000B00000B00000
4F0000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000008F0000700000B00000B000006F00007000000000000000000F0F0F
101010B0C0DFB0CFDFB0C0DFBFCFDFB0C0DFB0CFDFB0C0DFBFCFDFB0C0DFB0CFDFB0C0DFBFCFDF\
B0CFDFB0A0B0B00000BF0000B00000BF0000B05050BF2F30B07F80BF7F80B0C0DF
BFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDF5F606F5F606F000000000000000000000000\
000000606F705F6F6FBFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0
BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDF\
B0CFDFBFCFDF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004F0000AF0000B00000
B000002F0000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F0000B00000B00000B000008F0000900000000000000000
0F0F0F101010B0C0D0B0CFDFB0C0DFB0CFDFB0C0D0B0CFDFB0C0D0B0A0B0B0C0D0B0CFDFB0C0D0\
B0CFDFB0C0D0B0A0AFB00000B00000B00000B00000B07080B0CFDFB0C0D0B0CFDF
B0C0D0B0CFDFB0C0D0B0CFDFB0C0D05F606F505F60000000000000000000000000000000000000\
000000000000606F70B0C0D0BFCFDFB0C0D0B0CFDFB0CFD0BFCFDFB0C0D0B0CFDF
B0CFD0BFCFDFB0C0D0B0CFDFB0CFD0BFCFDFB0C0D0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0\
BFCFDFB0CFD0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000B000002F0000000000000000\
000000000000000000000000000000000000000000000000000000900000B00000
B00000B00000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F0000B00000B00000B00000B00000000000000000
0000000F0F0FB0CFDFB0C0DFB0CFDFB0C0DFB0CFDFB0C0DFAFBFCFB00000BFCFDFB0C0DFB0CFDF\
B0C0DFBFCFDFB07F80B00000B00000BFA0B0B0A0AFB0CFDFB0C0DFBFCFDFB0CFDF
B0CFDFB0C0DFBFCFDF5F606F5F606F000000000000000000000000000000000000000000000000\
0000000000005F6F70BFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFDF
BFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDF\
B0CFD0BFCFDF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000B0C0DFAFC0DFB00000AF0000000000000000000000000000\
000000000000000000000000000000000000000000000000000000AF0000B00000
AF0000B00000AF0000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000200000B00000B00000B00000B000002F0000
0000000000000F0F0FB0CFDFB0C0D0B0C0DFB0C0D0B0CFDFB0C0D0B0505FB04F50B0CFDFB0C0D0\
B0CFDFB02F2FB00000B09FAFB0A0AFB0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDF
B0C0D05F606F506060000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F606FB0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDF
B0C0D0B0CFDFB0CFD0BFCFDFB0C0D0B0CFDFB0CFD0BFCFDFB0C0D0B0CFDFB0CFD0BFCFDFB0CFD0\
B0CFDFB0CFD0BFCFDF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000AFC0DFB0C0DFB00000B00000AF2F2F6F505F000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F0000B00000
B00000B00000900000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000F0FB00000B00000B00000B00000
4F0000000000101010B0C0DFB0CFDFB0C0DFB0CFDFB0C0DFB0A0B0B00000B07F80B0C0DFB07F80\
B07F80B07F80B07080B0CFDFB0C0DFB0CFDFB0C0DFBFCFDFB0C0DFB0CFDF50606F
5F606F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000606F70B0CFDFBFCFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0
BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDF\
B0CFD0BFCFDFB0CFD0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000B0C0DFAFBFD0B00000AF0000B00000AF0000B02F2FAF202F5F606F000000000000000000\
0000000000000000000000000000000000000000000000000000000000004F0000
AF0000B00000AF00006F0000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000900000AF0000B00000
4000000000000F0F0F0F1010B0C0D0B0CFDFB0C0D0B0C0DFB00000B00000B07080B07F80B0C0D0\
B0A0B0B0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDF505F60000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000606F70B0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDF
B0C0D0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDFB0C0D0B0CFDFB0CFD0\
BFCFDFB0C0D0B0CFDFB0CFD0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
B0C0DFAFC0DFB00000AF0000B0C0DFAF7080B0C0DFAF7080B00000AF00005F606F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
900000AF0000B00000B000004F0000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006F0000B00000B00000
B000008F00000000000F0F10B0CFDFB0C0DFB0CFDFB09FAFB00000B00000B00000B09FAFB0CFDF\
B0C0D0B0CFDFB0C0DFB0CFDFB0C0D0B0CFDFB0C0DF5F606F50606F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F606FBFCFDFB0CFD0B0CFDFB0CFD0BFCFDFB0CFD0
BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDF\
B0CFD0BFCFDFB0CFD0BFCFDF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000AFC0DFB0C0DF
AF0000B00000AFC0DFB0C0DFAFBFD0B0C0DFAFC0DFB0C0DFA0BFCFB0505FAF4F506F505F000000\
0000000000000000000000000000000000000000000000000000008F00008F0000
200000B00000AF0000B00000AF0000200000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004F0000400000B00000
AF0000B00000000000101010AFC0D0B0C0DFB0C0D0B02F2FAF9FAFB07F80AFBFCFB0CFDFAFC0D0\
B0C0DFB0C0D0B0CFDFAFC0D0B0CFDF505F605F606F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F6F70B0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDF
B0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDFB0C0D0\
B0CFDFB0CFD0B0CFDFB0C0D0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B0C0DFAFC0DFB00000AF0000
B00000AF0000B0C0DFAFC0DFB0C0DFAFC0DFB0C0DFAFC0DFB0C0DFAFC0DFA0BFCFA0B0CF902F30\
4000000000000000000000000000000000000000009000008F0000900000200000
2F0000B00000B00000AF0000B00000B00000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000004F0000B00000
B00000B000000F0F0F400000101010B02F2FB02F30B09FAFB0CFDFB0C0DFB0CFDFB0C0D0B0CFDF\
B0C0DFB0CFDFB0C0D05F606F50606F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F70B0C0D0BFCFDFB0CFD0B0CFDFB0CFD0
BFCFDFB0CFD0B0CFDFB0CFD0BFCFDFB0CFD0B0CFDFB0CFD0BFCFDFB0CFD0BFCFDFB0CFD0BFCFDF\
B0CFD0BFCFDFB0CFD0BFCFDFB0CFD0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000AFBFD0AFC0DFAF0000B00000AF0000AF0000
AF0000B02F30AF202FAFC0DFAFC0DFB0C0DFAFBFD0AFC0DFAFC0DFB0C0DFAFBFD0B0C0DF8F202F\
6F505F8F00000000004000000000008F00008F00008F0000200000200000000000
000000200000AF0000B00000AF00008F0000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000200000B00000
AF0000B00000AF0000B000008F0000B0A0AFAF9FAFB0C0DFAFC0D0B0CFDFAFC0D0B0C0DFB0C0D0\
B0CFDF505F605F606F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F606FB0CFDFB0C0D0B0CFDFB0C0D0B0CFDF
B0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFD0B0C0D0B0CFDFB0C0D0B0CFD0B0C0D0\
B0CFDFB0C0D0B0CFD0B0CFD0B0CFDF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000B0C0DFAFC0DFB0C0DFAF0000B00000A0B0CFB02F30AF0000
B00000AF0000B02F30AF2F2FB0C0DFAFC0DFB0C0DFAFC0DFB0C0DFAFC0DFB0C0DFAFC0DFB0C0DF\
6F4F505F606F6F0000B000008F00009000002000002F0000000000000000000000
0000000000002F0000AF0000B00000AF00006F0000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000200000
B00000B00000B000002000002F00000F0F0FB0CFDFB0C0DFB0CFDFB0C0D0B0CFDFB0C0DF5F606F\
50606F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F6F6FB0CFDFB0C0D0B0CFDFB0CFD0
B0CFDFB0C0D0BFCFDFB0CFD0B0CFDFB0CFD0BFCFDFB0CFD0B0CFDFB0CFD0BFCFDFB0CFD0B0CFDF\
B0CFD0BFCFDFB0CFD0B0CFDFB0CFD0BFCFDF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000AFC0DFAFBFD0AF9FB0AF0000AFC0DFAFBFD0AFC0DFAFBFDFAF9FB0
AF0000AF0000AF0000B02F2FAFBFD0AFC0DFAFBFD0B0C0DFAFBFD0AFC0DFAFC0D0B0C0DFAFBFD0\
AFC0DF505F6F50606F2000008F00006F00006F0000000000000000000000000000
0000000000000000006F0000AF0000B00000AF00004F0000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000008F0000
AF00002000002000000000000000000F1010AFC0D0B0C0DFAFC0D0B0C0DF505F605F606F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F6F70B0C0D0B0CFD0B0C0D0B0CFDF
B0C0D0B0CFD0B0C0D0B0CFDFB0C0D0B0CFD0B0C0D0B0CFDFB0C0D0B0CFD0B0C0D0B0CFDFB0C0D0\
B0CFD0B0C0D0B0CFDFB0C0D0B0CFD0B0CFD0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
AFC0DFB0C0DFAFBFDFB0C0DFAFC0DFB0C0DFAFC0DFB0C0DFAFC0DFB0C0DFAFC0DF
A0BFCFAF0000B00000AF0000B00000AFC0DFB0C0DFAFC0DFB0C0DFAFC0DFB0C0DFAFC0DFB0C0DF\
AFC0DFB0C0DF505F6F0000000000002F00006F0000900000200000000000000000
000000000000000000000000900000AF0000B00000200000000000000000000000000000000000\
2000006F00000000004F0000000000000000000F0F0000008F0000900000200000
2F00000000000000000000000000000F0F10B0CFDFAFC0D05F606F50606F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F606F606F70B0CFD0B0CFDFB0C0D0
B0CFDFB0CFD0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDFB0CFD0B0CFDFB0CFD0B0CFDFB0CFD0BFCFDF\
B0CFD0B0CFDFB0CFD0BFCFDFB0CFD0B0CFDF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000AFC0DFAFBFDF\
AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DF
AFBFD0AFC0DFAF0000AF0000AF0000AF00009FB0C0A0B0CFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0\
B0C0DFAFBFD0AFC0DF505F6F0000000000000000008F00006F0000600000000000
000000000000000000000000AF0000B00000AF0000B00000000000000000000000000000000000\
000000200000B00000AF00004F0000600000B000008F0000200000200000000000
0000000000000000000000000000000000000F0F0F50606F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F606FB0CFDFAFC0D0B0CFD0
B0C0D0B0CFDFB0C0D0B0CFD0B0C0D0B0CFDFB0C0D0B0CFD0B0C0D0B0CFDFB0C0D0B0CFD0B0C0D0\
B0CFDFB0C0D0B0CFD0B0C0D0B0CFD0AFBFCFAFC0CF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000AFBFDFB0C0DFAF0000AF0000\
A0B0CFB0C0DFAFC0DFAFC0DFAFBFDFB0C0DFAFC0DFAFC0DFAFBFDFB0C0DFAFC0DF
B0C0DFAFC0DFB0C0DFAF2F2FB00000AF0000B00000AF9FAFB09FB0AFC0DFB0C0DFAFC0DFB0C0DF\
AFC0DFB0C0DFAFC0DFB0C0DF505F6F0000000000000000004000009000008F0000
4F00000000000000000000002F0000AF0000B000008F0000000000000000000000000000000000\
0000000000006F0000B00000AF0000B000004000002F0000000000000000000000
0000000000000000000000000000005F606F606F70000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000606F70B0C0D0B0CFDFB0CFD0
B0CFDFB0C0D0B0CFDFB0CFD0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDF\
B0CFD0B0CFDFB0CFD0B0CFDFAFC0CF0F0F0F0F0F0F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000AFBFD0AFC0DFAFBFD0AFBFD0AF0000\
AF0000AF0000A0B0CFAF4F50AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DF
AFBFD0AFC0DFAFBFD0AFC0DFAF202FAF0000AF0000AF0000AF7080AFC0DFAFBFD0AFC0DFAFBFD0\
AFC0DFAFBFD0AFC0DFAFBFD0AFC0DF505F6F0000000000000000000000004F0000
4000008F0000000000000000400000AF0000AF00008F0000400000000000000000000000000000\
0000000000002000000000008F00008F0000000000000000000000000000000000
0000000000000000005F6F705F606FB0C0DFA0B0C0A0BFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F6F70B0C0D0B0CFDF
AFC0D0B0CFD0B0C0D0B0CFDFAFC0D0B0CFD0B0C0D0B0CFDFAFC0D0B0CFD0B0C0D0B0CFD0B0C0D0\
B0CFD0B0C0D0AFC0CFAFBFCF0F0F0F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000AFC0DFAFBFDFAFC0DFAFBFDFAFC0DFAFBFDFAF7080\
AF0000AF0000AF0000AF505FAF0000A0B0CFAFBFDFB0C0DFAFC0DFAFC0DFAFBFDF
B0C0DFAFC0DFAFC0DFAFBFDFB0C0DFAF7080AF0000AF0000B00000AF7080B0C0DFAFC0DFB0C0DF\
AFC0DFB0C0DFAFC0DFB0C0DFAFC0DFB0C0DF505F6F000000000000000000000000
0000004000004F00000000004F0000AF0000B00000AF00004F0000000000000000000000000000\
0000000000000000000000004000000F0F0F000000000000000000000000000000
0000005F606F5F6F70AFC0D0B0CFDFAFC0D0B0C0DFA0B0CF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F606FB0CFDFB0C0D0
B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDFB0C0D0B0CFDFB0CFD0B0CFDFB0C0D0B0CFDF\
AFC0CFAFC0D0000F0F0F0F0F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000AFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0\
AF7080AF0000AF0000AF0000AF2F2FAF202FAF4F5F9FB0CFAFC0DFAFBFD0AFC0DF
AFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAF7080AF0000AF0000AF0000AF4F50AFC0DFAFBFD0\
AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0505F6F505F60000000000000000000
0000000F0F0F8F00004F00008F00008F0000AF0000AF0000200000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F6070
5F606FB0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0A0BFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F6F705F606FB0CFD0
AFC0D0B0CFD0AFC0D0B0CFD0B0C0D0B0CFD0AFC0D0B0CFD0B0C0D0B0CFD0AFC0D0AFC0CFAFBFCF\
0F0F0F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000AFC0DFAFBFDFAF0000AF202FAFC0DFAFBFDFAFC0DFAFBFDFAFC0DFAFBFDFAFC0DF\
AFBFDFAF2F30AF0000AF0000AF0000AF9FB0AF2F2FAF2F30AF4F50AFC0DFA0B0CF
AFC0DFAFBFDFB0C0DFAFC0DFAFC0DFAFBFDFB07F80AF9FAFAF0000AF0000B00000AF2F2FAFC0DF\
AFC0D0B0C0DFAFC0DFAFC0DFAFC0D0B0C0DFAFC0DF5F606F000000000000000000
0000000000000000000F0F0F6F0000AF0000B00000AF0000B00000000F0F000000000000000000\
0000000000000000000000000000000000000000000000005F606F5F6F70AFC0D0
B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0CFDFAFC0D0A0BFCF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000606F70B0C0D0
B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFAFC0CFB0C0D00F0F0F0F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000A0BFD0A0B0CFAF0000A00000AFBFDFAFBFD0AFC0DFA0BFD0AFBFDFAFBFD0AFC0DFA0BFD0\
AFC0DFAFBFD0AF2F2FAF0000AF0000AF0000A0BFD0AF7080AFC0DFAF202FA0B0CF
AFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AF9FAFAF90AFAF0000AF0000AF0000AF0000\
AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DF505F60000000000000000000
000000000000000000000000200000200000600000AF00008F0000000000000000000000000000\
0000000000000000000000000000000000005F60705F606FAFC0DFAFC0D0B0C0DF
AFC0D0AFC0DFAFC0D0B0C0DFAFC0D0AFC0DFAFC0D0B0C0DF9FB0C0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000005F606FB0CFD0
AFC0D0B0CFD0AFC0D0B0CFD0AFC0D0B0CFD0AFC0D0AFC0CFAFBFCF0F0F0F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000AF0000AF0000AF0000AF9FB0AFBFDFAFC0DFAFBFDFAFC0DFAFBFDFAFC0DFAFBFDFAFC0DF\
AFBFDFAFC0DFAFBFDFAF0000AF0000AF0000AF0000AFC0DFAFBFDFAFC0DFAF7080
AF505FA0B0CFA0B0CFAFBFD0AFC0DFAFC0DFAFC0DFAFBFD0AFC0DFAFC0DFAF0000AF0000B00000\
AF0000AF9FB0AFBFD0B0C0DFAFC0DFAFC0DFAFC0D0B0C0DF505F6F000000000000
0000000000000000000000000000000000000F0F0FAF00004F0000000000000000000000000000\
0000000000000000000000000000005F6F70AFC0D0B0C0DFAFC0D0B0C0DFAFC0D0
B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0CFDFA0B0CF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F606F
B0CFDFB0C0D0B0CFDFB0C0D0B0CFDFAFC0CFAFC0D0000F0F0F0F0F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000AF0000
A00000200000A00000400000A00000AFC0DFA0BFD0AFBFDFAFBFD0AFC0DFA0BFD0AFBFDFAFBFD0\
AFC0DFA0BFD0AF9FAFAF0000AF0000A00000AF2F2FAF4F50AFC0DFA0BFD0AFC0DF
AFBFD0AF7080AF202FAF2F2FAF90AFAFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AF0000AF0000\
AF0000AF90AFAFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD050606F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F606F5F6070AFBFD0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DF
AFC0D0AFC0DFAFC0D0AFC0DFAFC0D0B0C0DFAFC0D0AFC0D0AFC0D0A0BFCF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F6F6F
AFC0D0B0CFD0AFC0D0B0CFD0AFBFCF0F0F0F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000400000200000AF0000AF9FB0AF9FAFAFC0DFAFBFD0AFC0DFAFBFDFAFC0DF\
AFBFD0AFC0DFAFBFDFAF7080AF0000AF0000AF0000AF505FAF7080AFC0DFAFBFDF
AFC0DFAFBFD0AFC0DFAF2F2FAF9FB0AF202FAFC0DFAFBFDFAFC0DFAFBFD0AFC0DFAF0000AF0000\
AF0000AF0000AF7080AFC0DFAFBFD0AFC0DFAFC0DFAFC0DFAFBFD05F606F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F6F705F606FB0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0
B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFA0B0C0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F606F
5F6F70AFC0D0AFC0D0AFBFCF0F0F0F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000AF0000A000004000000000000F0F0F
0000000000000000000000000000000F0F0FA04F50AF7080A0707FAFBFDFA0BFD0AFC0DFA0BFD0\
AFBFDFA0BFD0AFC0DFA0BFD0AF4F5FAF0000AF0000A00000AF7080AFBFD0AFC0DF
A0BFD0AFC0DFAFBFD0AFC0DFA0707FAF2F2FAF202FAF9FAFA0BFD0AFC0DFAFBFD0AFC0DFAF202F\
AF0000AF0000AF505FAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DF505F60000000
00000000000000000000000000000000000000000000000000000000000000000000000050606F\
5F6070AFC0D0AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFBFD0AFC0D0AFC0D0AFC0DF
AFBFD0AFC0D0AFC0D0AFC0DFAFC0D0AFC0D0AFC0D0AFC0DFAFC0D0AFC0D09FB0C0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F606FAFC0CF0000000F0F0F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000AF0000A00000AF0000AF0000AF0000600000
AF00002000000000000000000000000F0F0F0F0F0FAF90AFAF505FAFBFDFAFC0DFAFBFD0AFC0DF\
AFBFDFAFC0DFAFBFD0AFC0DFAFBFDFAF2F30AF0000AF0000AF0000A0B0CFAFBFD0
AFC0DFAFBFDFAFC0DFAFBFD0AFC0DFAFBFDFAF7080AF202FAF9FB0AF9FAFAFC0DFAFBFD0AF505F\
AF0000AF0000AF202FAF2F30AFC0D0AFC0DFAFBFD0AFC0DFAFC0D050606F505F6F
0000000000000000000000000000000000000000000000000000000000005F60705F606FB0C0DF\
AFC0D0AFC0DFAFC0D0B0C0DFAFC0D0AFC0DFAFC0D0B0C0DFAFC0D0AFC0DFAFC0D0
B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0A0BFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F000F0F
6F505F0F0F0F000000000000000000000000000000000000000000000000000000000000000000\
209F00000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000A00000AF0000000000000000000000400000000000600000
8000008F00006000000000000000000000000000000F0F0FA0B0CFAF2F2FA090AFAF9FAFA0BFD0\
AFC0DFA0BFD0AFBFDFA0BFD0AFC0DFA0BFD0AF0000A00000AF0000A00000AFBFDF
A0BFD0AFC0DFA0BFD0AFC0DFAFBFD0AFC0DFA0BFD0AFC0DFAF7080AF2F2FA090AFAFC0DFAF7080\
AF7080A00000AF0000AF202FAFC0DFA0BFD0AFC0DFAFBFD0AF505FAFBFD0505F6F
00000000000000000000000000000000000000000000000050606F5F6070AFBFD0AFC0DFAFBFD0\
AFC0DFAFBFD0AFC0D0AFBFD0AFC0DFAFBFD0AFC0D0AFC0D0AFC0DFAFBFD0AFC0D0
AFC0D0AFC0DFAFBFD0AFC0D0AFC0D0AFC0DFAFBFD0AFC0D0AFC0D0AFC0DFAFC0D0A0B0CF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0FAF4F50B05050
AF4F506F505F00000000000000000000000000000000000000000000000000000020CF0020C000\
20A000209F00000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000AF0000A00000000000000000000000000000000000000000000000000000
00000000000F0F0F0F0000000000000000000000000F0F0FA0BFD0A0B0CFAF505FAF7080AF7080\
A0BFD0AFC0DFAFBFDFAFC0DFAFBFD0AFC0DFAF0000AF0000AF0000AF0000AF0000
AFC0DFAFBFD0AFC0DFAFBFDFAFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAF202FAF2F30AF2F2FAF9FAF\
AF7080AF0000AF0000AF0000AF7080AF7080AF2F2FAF2F30AF7080AFC0DF505F6F
0000000000000000000000000000000000005F6F705F606FAFC0DFAFC0D0AFC0DFAFC0D0AFC0DF\
AFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0B0C0DFAFC0D0
AFC0DFAFC0D0B0C0DFAFC0D0AFC0DFAFC0D0B0C0DFAFC0D0AFC0DFAFC0D0B0C0DFA0B0CF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F0F0F0FB0505FAF4F50B00000AF0000
B0CFD05060600000000000000000000000000000000000000000000000002FCF0020C0002FCF00\
20C0002FA000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000A00000\
AF0000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000A0B0D0A0B0CFAF9FAFA04F50\
AFBFDFA0B0CFAFBFDFA0BFD0AFBFDFA0BFD09FB0CFA00000AF0000A00000AF2F2F
A0BFD0AFBFDFA0BFD0AFC0DFA0BFD0AFBFDFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AF2F2FA0202F\
AF9FAFAF0000AF0000A00000AF0000AF0000AF9FAFA0707FAFC0DFAFBFD0AFC0DF
505F600000000000000000000000005F6070AFBFD0AFC0DFAFBFD0AFC0D0AFBFD0AFC0DFAFBFD0\
AFC0D0AFBFD0AFC0DFAFBFD0AFC0D0AFBFD0AFC0DFAFBFD0AFC0D0AFBFD0AFC0DF
AFBFD0AFC0D0AFC0D0AFC0DFAFBFD0AFC0D0AFC0D0AFC0DFAFBFD0AFC0D0AFC0D0A0B0CF9FB0C0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0FAF4F50B0505FAF0000AF0000AF0000B07F80
AFC0D050606F505F6000000000000000000000000000000000000000000000000020CF0020C000\
20C000209F00000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000A00000AF0000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0FA0BFD0A0B0CFAFC0DF\
AF4F5FA0BFD0A07080AFC0DFAFBFDFAFC0DFA0BFD0AF7080AF0000AF0000A00000
AF505FAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAF7080AF2F30\
AF2F2FAF0000AF0000AF0000AF9FAFAF9FAFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0
50606F00000000000050606F5F6F70AFC0D0AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFBFD0AFC0DF\
AFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0
AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0B0C0DFAFC0D0AFC0DFAFC0D0A0BFCF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000F0FB0505FAF5050B00000AF0000B00000AF0000B0505FAF4F50
B0CFDFAFC0D05F606F00000000000000000000000000000000000000000000000020C0002FCF00\
20C0002FCF00209F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000AF0000A00000400000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000A0B0D0A0BFD0\
AFBFDFA06F7FAF7080A04F50AFBFDFA0BFD0AFBFDFA0BFD0AF4F5FA00000AF0000
A07080AFC0DFA0BFD0AFBFDFA0BFD0AFC0DFA0BFD0AFBFDFA0BFD0AFC0DFA0BFD0AFBFDFA0BFD0\
AF9FAFA0202FAF7080A07080AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0D0
505F60505F6F50606FAFC0D0AFBFD0AFC0DFA0BFD0AFC0D0AFBFD0AFC0DFA0BFD0AFC0D0AFBFD0\
AFC0DFA0BFD0AFC0D0AFBFD0AFC0DFAFBFD0AFC0D0AFBFD0AFC0DFAFBFD0AFC0D0
AFBFD0AFC0D0AFBFD0AFC0D0AFC0D0AFC0D0AFBFD0AFC0D0AFC0D0AFC0D0AFBFD0AFC0D09FB0C0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F000000AF5050AF0000AF0000AF0000AF0000AF0000B00000AF0000A0BFC0
AFC0D0B0C0D0505F6000000000000000000000000000000000000000000000000000000020C000\
20CF0020C00020A000000000000000000000000000000000000000000000000000
004F00005000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000A00000AF0000A00000AF0000A00000200000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
0F0F0F0000000000000000000000000000000000000000000000000000000000000F0F0FA0BFD0\
A0BFD0AFC0DFA0BFD0AF505FA0B0CFA0BFD0AFBFD0AFC0DFA0202FAF0000AF0000
AF0000A0B0CFAFC0DFAFBFD0AFC0DFA0BFD0AFC0DFAFBFD0AFC0DFA0BFD0AFC0DFAFBFD0AFC0DF\
AFBFD0AF9FB0AF9FAFAFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DF505F6F
50606F50606FAFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFBFD0AFC0DFAFC0D0AFC0DF\
AFBFD0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0
AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DF\
9FB0C0000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F0F0F0F\
AF9FAFB0505FAF2F2FB00000AF0000B00000AF0000B00000AF0000B0505FAFC0D0
B0CFD0AFC0D0B0CFD05060600000000000000000000000000000000000000000000000002FCF00\
20C0002FCF00209F002FA000000000000000000000000000000000005000005000
2FCF0020C0002FCF00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000A00000A00000600000A00000A00000A00000A00000A00000\
400000000000000000000000000000000000000000000000000000000000400000
4000004000000000000000000000000000000000000000000000000000000000000F0F0F000000\
A0B0D0A0BFD0AFBFDFA0BFD0AF7080A07080AFBFDFA0BFD0AF0000A00000AF0000
A00000AFBFDFA0BFD0AFBFDFA0BFD0AFBFDFA0BFD0AFC0DFA0BFD0AFBFDFA0BFD0AFC0DFA0BFD0\
AFBFD0A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0AFC0DFA0BFD0505F6F505F60000000
0000005F6070A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0AFC0DFA0BFD0AFC0D0AFBFD0AFC0DFA0BFD0\
AFC0D0AFBFD0AFC0D0A0BFD0AFC0D0AFBFD0AFC0D0A0BFD0AFC0D0AFBFD0AFC0D0
A0BFD0AFC0D0AFBFD0AFC0D0AFBFD0AFC0D0AFBFD0AFC0D0AFBFD0AFC0D0AFC0D0AFC0D0AFBFD0\
A0B0CF000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F000000AFBFCFA0BFCF\
A0BFC0AF707FAFC0D0AFC0D0AF0000AF0000AF0000AF0000AF00009FB0C0A0BFC0
AFC0D0AFC0D0AFC0D050606F00000000000000000000000000000000000000000000000020C000\
20C00020C00020CF00209F00000000000000000000004F0000500020C00020CF00
20C00020C00020C000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006000006F0000A000000F0F0F00000F0F0F0FA00000800000A00000AF0000\
A00000600000600000000000000000000000000000000000000000000000000000
4F00004000004000000000000000000000000000000000000000000000000000000000000F0F0F\
0F0F0FA0BFD0A0BFD0AFC0DFA0B0CFAF505FA04F50A0BFD0A09FAFAF0000A00000
AF0000A0BFD0AFC0DFA0BFD0AFC0DFAFBFD0AFC0DFA0BFD0AFC0DFAFBFD0AFC0DFA0BFD0AFC0DF\
AFBFD0AFC0DFA0BFD0AFC0DFAFBFD0AFC0DF505F6050606F000000000000000000
0000000000005F6070AFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFBFD0AFC0DF\
AFC0D0AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFC0D0
AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DF\
A0B0C0A0BFCF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000F0F0F0F0FA0BFCFAFBFCFAFC0D0B0C0DF\
AFC0D0AFC0D0AFC0D0B07F80AF0000AF0000AF0000B00000AF0000B0505FAFC0D0
B0CFD0AFC0D0B0CFD0AFC0D05F606F000000000000000000000000000000000000000000000000\
20C0002FCF0020C0002FA0000000000000000050002FCF0020C0002FCF0020C000
2FCF0020C0002FA000209F00000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000200000000000000000000000000000000000000000000000000000800000A00000\
A00000A00000800000000000000000000000000000000000000000000000000000
000000600000600000000000000000000000000000000000000000000000000000000000000000\
0F0F0F9FB0CFA0BFDFA0BFD0AFBFDFA0BFD0AF7080A07080AF7080A00000AF0000
A00000AF4F5FA0BFD0AFBFDFA0BFD0AFBFDFA0BFD0AFBFD0A0BFD0AFC0DFA0BFD0AFBFD0A0BFD0\
AFC0DFA0BFD0AFBFD0A0BFD0505F6F4F5F60000000000000000000000000000000
00000000000050606FAFC0D0A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0\
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0
AFBFD0AFC0D0A0BFD0AFC0D0AFBFD0AFC0D0A0BFD0AFC0D0AFBFD0AFC0D0A0BFD0AFC0D0AFBFD0\
AFC0D09FB0C0000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F000000A0BFCFA0BFCFAFC0D0AFC0D0AFC0D0AFC0D0\
AFC0D0AFC0D0AFC0D0AF202FAF2F2FAF0000AF0000AF0000AF0000AF9FA0AFC0D0
AFC0D0AFC0D0AFC0D0AFC0D0505F60000000000000000000000000000000000000000000000000\
20C00020C00020C00020C00020A000004F0020C00020C00020CF0020C00020C000
209F0020A000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
A00000A00000000000000000000000000000000000000000000000000000000000000000600000\
A00000AF0000A00000AF00000000000F0F0F000000000000000000000000000000
000000000000AF0000800000000000000000000000000000000000000000000000000000000000\
00000F0F0F0FA0B0CFAFBFDFA0BFD0AFC0DF9FAFCFAF505FA0202FAF2F30A00000
AF0000A07080AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DF\
A0BFD0AFC0DFAFBFD050606F000000000000000000000000000000000000000000
0000000000005F607050606FAFC0DFA0BFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DF\
AFBFD0AFC0DFAFC0D0AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFBFD0AFC0DFAFC0D0
AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFC0D0AFC0DFAFC0D0AFC0D0AFC0D0AFC0DFAFC0D0AFC0D0\
AFC0D0A0BFCF000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0F0F0FA0BFCFAFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0\
AFC0D0B0C0D0AFC0D0AF9FAFAF0000B00000AF0000AF0000AF0000B02F30AFC0D0
AFC0D0AFC0D0B0CFD0AFC0D050606F000000000000000000000000000000000000000000000000\
0000002FCF0020C0002FCF0020C0002FCF0020C0002FCF0020C0002FA000209F00
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000A00000A00000
400000000000000000000000000000000000000000000000000000000000000000000000000000\
400000A00000A00000A00000A00000000000200000000000000000000000000000
000000400000A00000A00000800000000000000000000000000000000000000000000000000000\
000000000000A0B0D0A0BFD0A0BFDFA0BFD0AFBFDFA0BFD0A04F5FA00000AF0000
A00000A09FAFA090AFAFBFDFA0BFD0AFBFD0A0BFD0AFBFDFA0BFD0AFBFD0A0BFD0AFBFDFA0BFD0\
AFBFD0505F60505F6F000000000000000000000000000000000000000000000000
0000000000000000005F6070A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0\
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0
A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0AFBFD0AFC0D0A0BFD0\
AFC0D0AFBFD0A0B0CF000000000000000000000000000000000000000000000000
0000000F0F0F000000AF505FA0BFCFAFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0\
AFC0D0AFC0D0AFC0D0AF4F50AF0000AF0000AF0000AF0000AF0000AF9FA0AFC0D0
AFC0D0AFC0D0AFC0D0AFC0D0AFC0D050606F000000000000000000000000000000000000000000\
00000020C00020C00020C00020C00020C00020C000209F00209F00000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000505F6F4F5F6FA00000A00000A00000A00000
400000000000000000000000000000000000000000000000000000000000000000000000000000\
200000200000A00000AF0000A00000A00000400000000000000000000000000000
000000A00000AF0000A00000AF0000A00000200000000000000000000000000000000000000000\
0000000F0F0F9FB0CFA0BFD0A0BFD0AFBFDFA0BFD0AFC0DFA07080AF0000A00000
AF0000A00000AFBFDFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD050606F\
505F6F000000000000000000000000000000000000000000000000000000000000
00000000000000000050606FAFC0DFAFBFD0AFC0DFA0BFD0AFC0DFAFBFD0AFC0DFA0BFD0AFC0DF\
AFBFD0AFC0DFA0BFD0AFC0DFAFBFD0AFC0DFAFBFD0AFC0DFAFC0D0AFC0D0AFBFD0
AFC0DFAFC0D0AFC0D0AFBFD0AFC0DFAFC0D0AFC0D0AFC0D0AFC0DFAFC0D0AFC0D0AFC0D0AFC0DF\
AFC0D0AFC0D09FB0C0000000000000000000000000000000000000000000000000
0F0F0FAF4F50AF505FAF4F50AF7F80AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0\
AFC0D0AFC0D0AFC0D0AF0000AF0000AF0000AF0000AF0000AF0000AFC0D0AFC0D0
B0CFD0AFC0D0AFC0D0AFC0D0B0CFD0506060000000000000000000000000000000000000000000\
00000020CF0020C0002FCF0020C0002FCF00209F00000000000000000000000000
000000000000002F00002F00000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000A0B0D0A0BFDFA0B0D0A00000A00000A00000A00000A00000
800000000000000000000000000000000000000000000000000000000000000000000000000000\
000000200000000000A00000A00000A00000A00000400000000000000000A00000
A00000200000200000400000400000A00000A00000200000000000000000000000000000000000\
0F0F0F0000000F0F0F9FB0CFA0BFDFA0BFD0A0BFD0A0BFD0AFBFDFA04F50A00000
A00000AF7080A0BFD0A0BFD0A0BFD0AFBFDFA0BFD0A0BFD0A0BFD0AFBFDF4F5F60505F6F000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000050606FAFC0D0A0BFD0AFBFD0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0\
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0
A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0\
AFC0D0A0BFD09FB0C09FB0C00000000000000000000000000000000000000F0F0F
AF4F50AF505FA00000AFC0D0AFC0D0AFC0D0A0BFD0AFC0D0AFC0D0AFC0D0AFBFD0AFC0D0AFC0D0\
AFC0D0AFC0D0AFC0D0AF4F50AF0000AF0000AF0000AF0000AF2F2FAF707FAFC0D0
AFC0D0AFC0D0AFC0CFAFC0D0AFC0D0AFC0D0505F60000000000000000000000000000000000000\
00000000000020C00020C00020A000209F00000000000000000000000000000000
002000002F0020C00020C00020C000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000A0BFDFA0BFDFA00000A00000A06F80A0000080202FA00000A00000
A00000A00000000000200000000000000000000000000000000000000000000000000000000000\
000000000000000000000000800000AF0000A00000A00000A00000AF0000000000
000000000000000000000000000000000000600000400000000000000000000000000000000000\
6000000000000F0F0FA0B0D0A0BFD0AFBFDFA0BFD0AFBFDFA0202FAF0000A090AF
AF7080A0BFD0AFC0DFA0BFD0AFBFDFA0BFD0AFC0DFA0BFD0505F6F505F60000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005F6070A0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DF\
A0BFD0AFC0DFAFBFD0AFC0D0A0BFD0AFC0DFAFBFD0AFC0D0A0BFD0AFC0DFAFC0D0
AFC0D0A0BFD0AFC0DFAFC0D0AFC0D0AFBFD0AFC0DFAFC0D0AFC0D0AFBFD0AFC0DFAFC0D0AFC0D0\
AFC0D0AFC0DFAFC0D0A0B0CF0000000000000000000F0F0F00000FAF505FAF4F50
AF0000AF0000AF7F80AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0\
AFC0D0AFC0D0AFC0D0AF0000AF0000AF0000AF0000AF0000AF0000AFC0D0AFC0D0
AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D05F606F000000000000000000000000000000000000\
000000000000209F0020A000000000000000000000000000000000002F00002F00
20CF0020C0002FCF0020C0002FCF00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
A0BFDFA0B0D0A070809F6F7FA0BFDF9FAFCFA0BFDF9FB0D0604F5F604F50A00000
9F0000A00000A00000A000001F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000800000A00000A00000A00000800000000000
000000000000000000000000000000000000200000400000000000000000000000000000400000\
400000000000000000000000A0B0CFA0BFD0A04F5FA0202FA09FAFA090AFA0BFDF
A0BFD0A0BFD0A0BFD0A0BFDFA0BFD0A0BFD04F5F60505F6F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000050606F5F606FA0BFD0AFC0D0A0BFD0AFBFD0A0BFD0AFC0D0A0BFD0\
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0
A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0\
AFC0D0A0BFD0AFC0D09FB0C00000000000000F0F0FA09FA0AF5050A00000AF0000
A00000AF0000A00000AF7080A0BFD0AFC0D0AFC0D0AFC0D0A0BFD0AFC0D0AFC0D0AFC0D0A0BFD0\
AFC0D0AFC0D0A0BFCFA00000AF0000AF0000AF0000A00000AF2F2FAFC0D0AFC0D0
AFC0CFAFC0D0AFC0D0AFC0D0AFC0CFAFC0D0505F6050606F000000000000000000000000000000\
000000000000000000000000000000000000000000000000002F0020C00020C000
20C00020C00020C00020C00020C000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000A0B0D0A0BFDF\
A0BFD0A0BFDFA0B0D0A0BFDFA0BFD0A0BFDFA0B0D0A0BFDFA0BFD0A0BFDF604F50
A00000A00000A00000A00000A00000400000000000000000000000000000000000000000000000\
000000000000000000000000000000400000400000A00000A00000A00000A00000
000000000000000000000000000000000000000000000000000000000000000000000000AF0000\
A000004000000000000F0F0F9FB0CFA0BFDFA07080AF7080A0BFD0AFBFDFA0BFD0
AFC0DFA0BFD0AFBFDFA0BFD050606F505F6F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000050606FAFC0D0A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0AFC0DF\
A0BFD0AFC0D0A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0
AFC0DFAFBFD0AFC0D0A0BFD0AFC0DFAFC0D0AFC0D0A0BFD0AFC0DFAFC0D0AFC0D0A0BFD0AFC0D0\
AFC0D0AFC0D0A0BFD0A0BFCF9FAFBFA0BFCFA0BFCFAF9FAFAF4F50AF9FAFAF9FAF
AF9FAFAF0000AF0000AF4F50AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0\
AFC0D0AFC0D0AFC0D0AF0000AF0000AF0000AF0000AF0000AF0000AFC0D0AFC0D0
AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0505F60000000000000000000000000000000\
000000000000000000000000000000000000002F00002F0020C00020CF0020C000
2FCF0020C00020CF0020C0002FCF0020C000AFCFD0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000A0BFDF9FB0D0A000009F0000\
A090AF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDFA0B0D0A0BFDF9FB0D0A0BFDF
4F5F608000007F0000A00000A000004000003F0000000000000000000000000000000000000000\
000000000000000000000000000000000000400000200000A00000A00000A00000
200000000000000000000000000000000000000000000000000000000000000000000000A00000\
A00000200000000000000000A0B0D09FB0CFA0BFD0A0BFD0A0BFDFA0BFD0A0BFD0
A0BFD0A0BFD0A0BFD0505F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F6070A0BFD0A0BFD0A0BFD0AFC0D0A0BFD0AFBFD0A0BFD0\
AFC0D0A0BFD0AFBFD0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0
A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0\
AFC0D09FB0CFA0BFCF00000010101FA0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0
A0BFD0AF505FA00000AF0000A00000AF7080A0BFD0AFC0D0A0C0D0AFC0D0A0BFCFAFC0D0A0C0D0\
AFC0D0A0BFCFA0BFCFAF0000AF0000A00000AF0000AF0000AF0000A0C0CFAFC0D0
AFC0D0AFC0D0A0C0CFAFC0D0AFC0D0AFC0D0A0C0CF506060000000000000000000000000000000\
00000000000000000000000000000000000000500020C00020C00020C00020C000
20C00020C000209F2F20C00020C00020C000AFC0CF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000A0B0D0A0BFDFA0B0D0A00000A00000A0BFDF\
A0B0D0A0BFDFA0B0D0A0BFDFA0B0D0A0BFDFA0B0D0A0BFDFA0B0D0A0BFDFA0BFD0
A0BFDF4F5F6F800000400000A00000A00000800000800000000000000000000000000000000000\
000000000000000000000000000000000000000000200000A00000A00000A00000
A000004000000000000000000000000000000000000000000000000000000000000F0F0FA00000\
A000000F0F0F00000000000000000FA0BFD0A0BFD0A0BFDFA0BFD0AFBFDFA0BFD0
A0BFDF4F5F60505F6F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F6070A0BFD0AFC0D0A0BFD0AFC0DFA0BFD0AFC0D0\
A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0AFC0DFA0BFD0
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0AFC0D0A0BFCF\
A0BFCF0F0F0F0F0F0F0000000F0F0F101010AFC0D0AFC0D0A0BFD0AFC0D0AFC0D0
AFC0D0A0C0D0AFC0D0AF9FAFAF0000AF2F2FAFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0\
AFC0D0AFC0D0AF707FAF2F2FAF0000AF0000AF0000AF0000AF707FAFC0D0AFC0D0
AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D050606F000000000000000000000000\
00000000000000000000000000000000000020C00020C00020C0002FCF0020C000
5F6F6F5F606F4FCF3020C00020CF0020C0002FCF00AFC0D0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000009FB0D0A0BFDF9F0000A000009F0000A070809FB0D0\
A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF
9FB0D0A0BFDF4F5F604000009F0000A000009F00008000007F0000000000000000000000000000\
0000000000000000000000000000000000000000001F0000000000800000A00000
A00000400000000000000000000000000000000000000000000000000000200000800000A00000\
2000002000000000000000000F0F0F9FB0CFA0BFD0A0BFD0A0BFD0A0BFD0505F6F
4F5F60000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000050606FAFBFD0A0BFD0A0BFD0A0BFD0AFC0D0A0BFD0\
A0BFD0A0BFD0AFC0D0A0BFD0A0BFD0A0BFD0AFC0D0A0BFD0AFBFD0A0BFD0AFC0D0
A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0A0BFCF000000\
0F0F0F0000000000000000000000000F0F0FAFC0D0A0BFD0AFC0D0A0BFD0AFC0D0
A0BFCFAFC0D0A0BFD0AFC0D0A0202FAF0000A00000AF5050A0BFCFAFC0D0A0BFD0AFC0D0A0BFCF\
AFC0D0A0C0D0AFC0D0A0202FAF0000A00000AF0000A00000AF0000A0C0D0AFC0D0
A0BFCFAFC0D0AFC0D0AFC0D0A0C0CFAFC0D0AFC0CFAFC0D0505F60000000000000000000000000\
000000000000000000000000000000002000207F0020C00020C00020C0004F7F50
AFC0CFAFC0D040C02F4FC02F20C00020C00020C000AFC0D0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000A0BFDFA0B0D0A000009F0000A00000A00000A02F309F202FA0BFDF\
A0B0D0A0BFDF9FB0D0A0BFDFA0B0D0A0BFDFA0B0D0A0BFDFA0B0D0A0BFDFA0B0D0
A0BFDFA0B0D0505F6F4F5F60400000A00000A00000A00000A00000000000200000000000000000\
000000000000000000000000000000000000000000000000000000800000A00000
A00000600000800000000000000000000000000000000000000000000000800000A00000200000\
0000000000000000000000000000000F0F0F9FB0CFA0BFDF505F6F505F6F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F607050606FAFC0DFA0BFD0AFC0D0A0BFD0AFC0DF\
A0BFD0AFC0D0A0BFD0AFC0DFA0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFCFA0BFCF00000F000000\
000000000000000000000000000000101010A0BFD0AFC0D0AFC0D0AFC0D0A0BFD0
AFC0D0AFC0D0AFC0D0A0C0D0AFC0D0A0BFCFAF0000A00000A0BFCFAFC0D0AFC0D0A0C0D0AFC0D0\
AFC0D0AFC0D0AF707FAF0000AF0000AF0000AF0000AF0000AF707FAFC0D0AFC0D0
AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D050606F505F60000000000000000000\
00000000000000000000000000000000000020C0002FCF0020C0006FC080AFC0D0
AFCFD0AFC0D0AFC0D040C02F2FCF0020C0006FCF80AFC0CF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009FB0D0A0B0DF9F0000A000009F6F7FA000009F0000A000009F0000A0202F9F202F\
A0BFDF9FB0D0A0B0DF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFD0
9FB0D0A0BFDF9FB0D0505F6F4F50602000009F0000A000009F0000A000001F0000200000000000\
0000000000000000000000000000000000000000000000000000005F0000A00000
A00000A000009F0000000000000000000000000000000000800000800000200000200000000000\
000000000000000000000000000000000000505F6F4F5F60000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000050606FA0BFD0A0BFD0A0BFD0A0BFD0A0BFD0\
A0BFD0A0BFD0A0BFD0A0BFD0AFC0D0A0BFD0A0BFD0A0BFD0AFC0D0A0BFD0A0BFD0
A0BFD0AFC0D0A0BFD0A0C0D0A0BFD0AFC0D0A0BFD0A0BFCF9FB0CF0F0F0F000000000000000000\
0000000000000000000000000000000000000F0F10A0BFD0AFC0D0A0BFCFAFC0D0
A0BFD0AFC0D0A0BFCFAFC0D0A0BFD0AFC0D0A04F50AF0000A00000AF2F2FA0BFCFAFC0D0A0BFD0\
AFC0D0A0BFCFAF2F2FA0202FAF0000A00000AF0000A00000A0B0C0A0BFCFAFC0D0
A0C0CFAFC0D0A0BFCFAFC0D0A0C0CFAFC0D0A0BFCFAFC0D0A0C0CF50606F000000000000000000\
00000000000000000000000000000000200020C00020C00020C00020C000A0BFC0
AFC0CFAFC0D0A0C0CF40C02F60C07F6FC080AFC0CFAFC0D0AFC0CF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
A0BFDF9FB0D0A070809F0000A0000090AFCFA070809FB0D0A04F5F9F0000A00000A00000A00000\
9F0000A0BFDF9FAFCFA0BFDF9FB0D0A0BFDFA0B0D0A0BFDF9FB0D0A0BFDFA0B0D0
A0BFDF9FB0D0A0BFDFA0B0D0505F6F4F5F60800000A00000A00000A00000A00000400000000000\
000000000000000000000000000000000000000000000000000000000000400000
A00000A00000A00000200000000000000000800000800000200000200000000000000000000000\
00000000000000000000000050606F5F60709FB0CF0F0F0F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000050606FAFC0D0A0BFD0AFC0D0A0BFD0AFC0D0\
A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D09FB0CFA0BFCF0F0F0F0F0F0F000000000000000000000000\
0000000000000000000000000000000000000F0F10AFC0D0A0BFD0AFC0D0A0C0D0
AFC0D0A0BFD0AFC0D0A0C0D0AFC0D0A0BFD0AFC0D0AFC0D0AF2F2FA00000AF9FAFAFC0D0AFC0D0\
A0C0D0AFC0D0AF9FA0AF0000A00000AF0000AF0000AF505FA04F50AFC0D0AFC0D0
AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0AFC0D0505F60000000000000000000\
00000000000000000000000050606F4FA03040C02F20C00020C0004FCF30AFC0D0
AFC0D0AFC0CFAFCFD0AFC0D0AFC0D0AFC0CFAFCFD0AFC0D0AFC0D0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000A0B0DF
9FB0D0A0B0DF9FB0D0A0B0DF9FB0D0A0B0DF9FB0D0A0BFDF9FB0D0A090AF9F90AFA000009F0000\
A000009F00009FB0D09F8FAFA0B0D09FB0D0A0BFDF9FB0D0A0B0D09FB0D0A0BFDF
9FB0D0A0BFD09FB0D0A0BFDF9F202F802F2F7F202F2000007F0000A000009F0000A000005F0000\
000000000000000000000000000000000000000000000000000000000000400000
9F0000A000009F0000A000009F0000A00000200000200000000000000000000000000000000000\
000000000000506070505F6FA0BFD0A0BFD09FB0CF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000505F6FA0BFD0A0BFD0A0BFD0A0BFD0\
A0BFD0A0BFD0A0BFD0A0BFD0A0BFD0A0BFD0A0C0D0A0BFD0A0BFD0A0BFD0A0C0D0
A0BFD0A0BFD0A0BFD0A0BFCF9FB0CF0F0F0F000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0FAFC0D0A0BFD0AFC0D0
A0BFCFAFC0D0A0BFD0AFC0D0A0BFCFAFC0D0A0BFD0AFC0D0A0707FAF0000A00000AF00009FB0C0\
AFC0D0A0BFCFAF2F2FA00000AF0000A00000AF0000A000009FB0C0A0C0CFAFC0D0
A0BFCFAFC0D0A0C0CFAFC0D0A0BFCFAFC0D0A0C0CFAFC0D0A0BFCFAFC0D0505F60000000000000\
0000000000005F606F50606FAFC0D0A0C0CF20C00020C00020C00060C050AFC0D0
A0C0CFAFC0D0AFC0CFAFC0D0A0C0CFAFC0D0AFC0CFAFC0D0A0C0CFAFC0D0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000009FB0D0A0BFDF9FB0D0
A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A090AF9F90AFA00000\
9F0000A000009F0000A070809F6F80A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0
A0BFDFA0B0D0A0BFDF9F202FA02F30A090AF604F5F200000000000400000A000009F0000800000\
000000000000000000000000000000000000000000000000000000000000000000
400000A00000A00000A00000200000200000000000000000000000000000000000000000000000\
50606F506070A0BFD0A0BFDFA0BFD0A0BFDF9FB0CF0F0F0F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F6070A0BFD0A0C0D0A0BFD0AFC0D0\
A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0
AFC0D09FB0CFA0BFCF00000F0F0F0F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000101010A0BFD0AFC0D0A0BFD0
AFC0D0A0C0D0AFC0D0A0BFD0AFC0D0A0C0D0AFC0D0A0BFD0AFC0D0A0C0D0AF505FA00000AF7F80\
A0C0D0AFC0D0A09FA0AF0000AF0000AF0000A00000AF505FAFC0D0AFC0D0A0C0D0
AFC0D0AFC0D0AFC0D0A0C0D0AFC0D0AFC0D0AFC0D0A0C0D0AFC0D0AFC0D050606F000000000000\
5F606F5F6F6FAFC0CFAFC0D0AFC0D0AFC0D020C0002FC00020C00020C00060C07F
AFC0D0AFC0D0AFC0D0AFC0CFAFCFD0AFC0D0AFC0D0AFC0CFAFCFD0AFC0D0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000009FB0DF9FB0D0A000009F8FAF9FB0DF
9FB0D0A0B0DF9FB0D09FB0DF9FB0D0A0B0DF9FB0D09FB0D09FB0D0A0B0DF9FB0D0A0B0D09FB0D0\
A02F2F9F0000A000009F0000A04F5F9FB0D0A0B0D09FB0D0A0BFDF9FB0D0A0B0D0
9FB0D0A02F2F9F202FA090AF9F90AFA090AF9FB0D04F5F6F0000002000009F0000A000007F0000\
A00000000000000000000000000000000000000000000000000000000000000000
9F0000A000001F0000200000000000000000000000000000000000000000000000000000505F6F\
A0BFD09FB0D0A0BFD0A0B0D0A0BFD09FB0D0A0BFD09FB0CF0F0F0F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000050606FA0BFD0A0BFD0A0BFD0\
A0BFD0A0BFD0A0BFD0A0BFD0A0BFD0A0BFD0A0BFD0A0BFD0A0BFD0A0BFD0A0BFCF
9FB0CF0F0F0F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0FAFC0D0A0BFCFA0C0D0
A0BFD0AFC0D0A0BFCFA0C0D0A0BFCFAFC0D0A0BFCFAFC0D0A0BFCFAFC0D0A090A0AF2F2FA0202F\
AF0000A09FA0AF5050A00000AF0000A00000AF0000A00000AF9FAFA0BFCFAFC0D0
A0BFCFAFC0D0A0BFCFAFC0D0A0C0CFAFC0D0A0BFCFAFC0D0A0C0CFAFC0D04F5F6050606050606F\
AFC0D0A0BFCFAFC0D0A0C0CFAFC0D0A0C0CF8FC0AF20C00020C00020C00060C07F
A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0AFC0CFAFC0D0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000009FB0D0A0BFDF9F0000A000009F0000A0BFDF9FB0D0
A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF\
9FB0D0A02F309F0000A000009F0000A02F309FB0D0A0BFDF9FB0D0A0BFDF9F202F
A02F309F90AFA090AF9F90AFA0BFDF9FB0D0A0BFDF4F5F6F000000200000A00000A00000A00000\
9F0000200000000000000000000000000000000000000000000000000000A00000
2000002000000000000000000000000000000000000000000000000000005F607050606FA0BFDF\
A0BFD0A0BFDFA0BFD0A0BFD0A0BFD0A0BFDFA0BFD09FB0CF00000F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000050606FAFC0D0A0BFD0A0C0D0\
A0BFD0AFC0D0A0BFD0A0C0D0A0BFD0AFC0D0A0BFD0A0C0D0A0BFD0A0BFCF0F0F0F
0F0F0F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0FAFC0D0A0BFD0
AFC0D0A0BFD0AFC0D0A0C0D0AFC0D0A0BFD0AFC0D0A0C0D0AFC0D0A0BFD0AFC0D0A0C0D0AF7080\
A00000AF505FA00000AF0000A00000AF0000A00000AF2F2FA0C0D0AFC0D0A0C0D0
AFC0D0A0C0D0AFC0D0AFC0D0AFC0D0A0C0CFAFC0D0AFC0D0AFC0D0505F6050606F50606FAFC0D0\
A0C0CFAFC0D0AFC0D0AFC0D0A0C0CFAFC0D08FC0A020C00020C0002FC00020C000
20C000A0BFC0AFC0D0AFC0D0AFC0D0AFC0CFAFC0D0AFC0CF6FC08060C07F2FCF00000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000009FB0DF9FB0D09FB0DF9F00009F00009F6F7F9FB0D09FB0D0A0B0DF
9FB0D09FB0D09FB0D0A0B0DF9FB0D09FB0D09FB0D0A0B0DF9FB0D09FB0D09FB0D0A0B0DF9FB0D0\
9FB0D09FB0D0A04F5F9F00009F00009F0000A000009FB0D0A0202F9F202FA02F2F
9F8FAFA090AF9FB0D0A0BFDF9FB0D0A0B0D09FB0D0505F6F4F50600000000000008000009F0000\
A000003F00004000000000000000000000000000000000002000007F0000200000
000000000000000000000000000000000000000000000000505F6F50606F9FB0D0A0BFD09FB0D0\
A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FAFCF9FB0CF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000050606FA0BFD0A0BFD0A0BFD0\
A0BFD0A0BFD0A0BFD0A0BFCFA0BFD0A0BFD0A0BFD09FB0C09FB0CF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F10A0BFCFAFC0D0
A0BFCFA0C0D0A0BFCFAFC0D0A0BFCFA0C0D0A0BFCFAFC0D0A0BFCFA0C0D0A0BFCFAFC0D09FB0C0\
A04F50A00000AF0000A00000AF0000A00000AF0000A09FA0AFC0D0A0BFCFAFC0D0
A0BFCFAFC0D0A0BFCFAFC0D0A0BFCFAFC0D0A0BFCFAFC0D04F5F600000000000005F606FA0BFCF\
AFC0D0A0C0CFAFC0D0A0BFCFAFC0D0A0C0CF8FC0AF20BF0020C00020C00020C000
20C00060C05080C07FAFC0D040C0509FB0C040C02F6FC05020C00020C00020C000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000A0B0DF9FB0D0A000009F0000A000009F0000A0BFDF9FB0D0A0B0DF9FB0D0
A0BFDF9FB0D0A0B0DF9FB0D0A0BFDF9FB0D0A0B0DF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF\
9FB0D0A0BFDF9FB0D0A070809F0000A000009F0000A000009F202FA090AF9F90AF
A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0505F6F0000000000007F0000A00000\
9F0000A000007F00000000000000000000004000004000000000000F0F0F000000
000000000000000000000000000000000000506070505F6FA0BFDFA0BFD0A0BFD09FB0D0A0BFDF\
A0BFD0A0BFD0A0BFD0A0BFDFA0BFD0A0BFD0A0BFD0A0BFDF9FB0CF0F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F6070A0BFD0A0C0D0\
A0BFD0A0BFD0A0BFD0AFC0D0A0BFD0A0BFCF9FB0CF0F0F0F0F0F0F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000101010A0BFD0
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0C0D0AFC0D0A0BFD0AFC0D0\
A0C0D0AF9FAFA0202FAF0000A00000AF0000A00000AF2F30A0C0D0AFC0D0A0C0CF
AFC0D0A0C0D0AFC0D0A0C0CFAFC0D0A0C0D050606F505F600000000000000000000000005F6F70\
AFC0D0AFC0D0A0C0CFAFC0D0AFC0D0AFC0D0A0C0CFA0BFCF20C00020C00020C000
2FC00020C00020C00020C0002FC00020C00020C00020C0002FC00020C00020C00020C000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F9F00009F00009F00009F00009F00009F00009F202F9FB0DF9FB0D09FB0D0
9FB0D09FB0DF9FB0D09FB0D09FB0D09FB0DF9FB0D09FB0D09FB0D09FB0DF9FB0D09FB0D09FB0D0\
9FB0DF9FB0D09FB0D09F8FAF90AFCF9F00009F00009F0000A000009F8FAF9FB0D0
9FB0D0A0BFD09FB0D09FB0D09FB0D0A0BFD09FB0D0A0B0D04F50600000000000000000007F0000\
A000009F0000A000003F00008000007F0000200000200000000000000000000000
000000000000000000000000505F6F5060709FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0\
A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D09FB0CF0000000F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000505F6FA0BFD0A0BFD0\
A0BFD0A0BFCFA0BFD09FB0CF9FB0CF0000000F0F0F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0FA0C0D0
A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFAFC0D0A0BFCFA0C0D0A0BFCF\
AFC0D0A0BFCFA0707FA00000AF0000A00000A00000A0707FAFC0D0A0BFCFA0C0D0
A0BFCFAFC0D0A0BFCFAFC0D0505F6050606000000000000000000000000000000000000050606F\
AFC0D0A0BFCFAFC0D0A0C0CFAFC0D0A0BFCFAFC0D0A0C0CF6FC05020BF0020C000
20C00020C00020C00020C00020C00020C00020C00020C00020C00020C00040C02F40C02F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000009F0000\
9F00002000009F6F809F00009F6F80A000009F00009F00009F0000A0000090AFCF
A0B0DF9FB0D0A0BFDF9FB0D0A0B0DF9FB0D0A0BFDF9FB0D0A0B0DF9FB0D0A0BFDF9FB0D0A0B0DF\
9FB0D0A0BFDF9FB0D0A0BFDF90AFC0A0BFDF9F0000A000009F0000A070809FB0D0
A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF4F5F60000000000000400000\
9F0000A000009F0000A00000200000200000000000000000000000000000000000
000000000000506070505F6FA0BFD09FB0D0A0BFDFA0BFD0A0BFD09FB0D0A0BFDFA0BFD0A0BFD0\
9FB0D0A0BFDFA0BFD0A0BFD09FB0D0A0BFDFA0BFD0A0BFD09FBFD09FB0D0000F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000050606FA0BFD0\
A0BFD0A0BFCF9FB0CF0F0F0F00000F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001010100F0F0F
AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0\
A0BFCFAFC0D0A0C0D0A0BFCFA04F50AF0000A00000AFC0D0A0BFCFAFC0D0A0C0D0
AFC0D0A0BFCF50606F505F600000000000000000000000000000000000000000000000005F606F\
50606FAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CF8FC0AF40C02F
20C00020C0002FC00020C00020C00020C0002FC00020C0006FC05040C02FAFC0D0AFC0CF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004000003F0000\
0000000000000F0F0F90AFCF90AFCF9F8FAF9F00009F00009F00009F00009F0000
9F8FAF9FB0D09FB0D09FB0DF9FB0D09FB0D09FB0D09FB0DF9FB0D09FB0D09FB0D09FB0D09FB0D0\
9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09F2F2F9F00009F00009F00009F4F5F
9FB0D09FB0D09FB0D0A0B0D09FB0D09FB0D09FB0D0A0BFD09FB0D04F5F6F0000000F0F0F000000\
9F00009F00002000001F0000000000000000000000000000000000000000000000
505F6F50606F9FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0\
A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0CF0F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000050606F9FB0C0\
9FB0CF0000000F0F0F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F1010
A0BFCFA0BFD0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCF\
A0C0D0A0BFCFA0C0D0A0BFCFA09FAFA04F50AF5050A0BFCFA0C0D0A0BFCFAFC0D0
4F5F60505F60000000000000000000000000000000000000000000000000000000000000000000\
5F606FA0BFCFAFC0D0A0BFCFAFC0D0A0C0CFAFC0D0A0BFCFAFC0D0A0C0CF8FC0AF
80BF7F80C07F80C0A020C02F60BF7F20C02F80C0A06FC07FA0C0CFAFC0D0A0C0CFAFC0D0A0C0CF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000009F00009F0000000000000000\
0000000000000000000F0F0F90AFCF9FB0D09FB0D0A000009F00009F00009F0000
A000009F6F809FB0DF9FB0D0A0B0DF9FB0D09FB0DF9FB0D0A0BFDF9FB0D09FB0DF9FB0D0A0BFDF\
9FB0D0A0B0DF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A070809F0000A000009F202F
A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFDF4F5F6F505F6F7F0000A00000\
200000200000000000000000000000000000000000000000000000000000505F6F
A0BFDF9FB0D0A0BFD09FB0D0A0BFDF9FB0D0A0BFD09FB0D0A0BFDF9FB0D0A0BFD09FB0D0A0BFDF\
9FBFD0A0BFD09FB0D0A0BFD0A0BFD0A0BFD09FB0D0A0BFD0A0BFD09FB0CF9FB0CF
0F0F0F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F50606F505F6F\
0F0F0F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
AFC0D0A0BFD0A0C0D0A0BFCFAFC0D0A0BFD0AFC0D0A0BFCFAFC0D0A0BFD0AFC0D0A0BFCFAFC0D0\
A0BFD0AFC0D0A0BFCFAFC0D0A0BFD0AFC0D0A0BFCFAFC0D0A0C0D050606F505F60
000000000000000000000000000000000000000000000000000000000000000000000000000000\
50606FAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CF
AFC0D0A0C0CFAFC0D0A0BFC0AFC0D0A0C0CFAFC0D0A0C0CFAFC0D0AFC0CFAFC0D0A0C0CF6FC05F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000009F00009F00002000001F0000000000000000\
0000000000000000000000000F0F0F90AFCF90AFCF9FB0D09F2F2F9F40509F0000
9F00009F00009F40509FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D0\
9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09F6F7F9F00009F00009F202F
9F00009FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D07F202F200000200000\
0000000000000000000000000000000000000000000000005060704F5F6F9FBFD0
9FB0D0A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0\
A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D09FB0CF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F9F90A09FB0CF4F5F60\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0FA0C0D0A0BFCFA0BFD0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCF\
A0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D04F5F60506060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000050606FA0C0D0A0BFCFAFC0D0A0BFCFA0C0D0A0BFCFAFC0D0A0BFCFAFC0D0
A0C0CFAFC0D0A0BFCFAFC0D0A0C0CFAFC0D0A0BFCFAFC0CFA0C0CFAFC0D040BF2F40C02F20C000\
20C000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000009F00009F00009F00009F0000000000000000000000000000\
0000000000000000000000000000000F0F0F90AFCF9FB0D09FB0D09FB0DF9F6F80
9F00009F00009F00009F202F9FB0DF9FB0D0A0B0DF9FB0D09FB0DF9FB0D0A0BFDF9FB0D09FB0DF\
9FB0D0A0BFDF9FB0D09FB0DF9FB0D0A0BFDF9FB0D09F90AF9F8FAFA000009F0000
A000009F90AFA0BFDF9FB0D0A0BFDF9FB0D0A0BFDF9FB0D0A0BFD09FB0D0604F5F000000000000\
000000000000000000000000000000000000505F6F5060709FB0D0A0BFDF9FB0D0
A0BFD09FB0D0A0BFDF9FB0D0A0BFD09FB0D0A0BFDF9FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD0\
9FB0D0A0BFD09FBFD0A0BFD09FB0D0A0BFD09FBFD0A0BFD09FB0D0A0BFD09FB0CF
0F0F0F00000F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000F0F0F0F0FA04F509FBFCFA0BFD0505F6F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
101010A0BFCFAFC0D0A0BFD0A0C0D0A0BFCFAFC0D0A0BFD0A0C0D0A0BFCFAFC0D0A0BFD0A0C0D0\
A0BFCFAFC0D0A0BFD0AFC0D0A0BFCF50606F505F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F6F6FA0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CF
AFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CF40C02F40C02F20C00020C00020C000\
20C000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000009F00009000009F00009F00009F00009000009F0000000000000000000000\
0000000000000000000000000000000000000F0F0F00000090AFCF9FB0D09F90AF
9F8FAF9F00009F00009F00009F00009FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D0\
9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09F90AF9F00009F0000
9F00009F6F809F4F509FB0D09FB0D09FB0D09F6F7F9FB0D09FB0D09FB0D09FB0D04F5F6F000000\
00000000000000000000000050606F505F6F9FBFD09FB0D09FB0D09FB0D0A0BFD0
9FB0D09FB0D09FB0D0A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D0\
9FBFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD0
90AFCF0F0F0F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F000000A04F509F4F50A000009FBFCFA0BFD09FBFCF\
505F60000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F0F0F10A0BFCFA0C0D0A0BFCFA0BFD0A0BFCFA0C0D0A0BFCFA0BFD0A0BFCFA0C0D0A0BFCF\
A0C0D0A0BFCFA0C0D0A0BFCF505F604F5F60000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000050606F5F606FA0BFCFA0C0D0A0BFCFAFC0D0A0BFCFA0C0CFA0BFCFAFC0D0
A0BFCFA0C0CFA0BFCFAFC0D0A0BFCFA0C0CF40C02F40C02F20BF0020C00020C00020C00020BF00\
40C02F20C000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009F00009F00000F0F0F5F00009F00009F00009F00005F0000400000000000000000\
0000000000000000000000000000000000000000000F0F0F000F0F9FAFD09FB0D0
9F90AF9F90AF9F00009F00009F00009F00009FB0D09FB0D09FB0DF9FB0D09FB0DF9FB0D09FB0DF\
9FB0D09FB0DF9FB0D0A0BFDF9FB0D09FB0DF9FB0D0A0BFDF9FB0D09F00009F0000
A000009F00009F00009F8FAFA02F309F202F90AFCF90AFC0A0BFDF9FB0D0A0BFD04F5F60000000\
000000000000505F6F5060709FB0D0A0BFD09FB0D0A0BFDF9FB0D0A0BFD09FB0D0
A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD0\
9FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FBFD0A0BFD09FB0D0
9FB0CF000F0F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000F0F0F0FA04F50A04F5FA00000A00000A000009FB0C0A0BFD0A0BFD0\
4F5F60000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0FA0C0D0A0BFCFA0C0D0A0BFD0A0C0D0A0BFCFA0C0D0A0BFD0A0C0D0A0BFCFAFC0D0\
A0BFCFA0C0D04F5F6050606F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000050606FAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CF
AFC0D0A0C0CFAFC0D0A0C0CF4FC02F40C02F20C00020C00020C00020C00020C00020C0006FC080\
40C02F8FC07F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F00000000000000000000000000005F00005F00009F00009F0000900000200000000000\
0000000000000000000000000000000000000000000000000F0F0F90AFCF90AFCF
9FB0D09FB0D090AFD09F00009F00009F00009F00009F6F809FB0D09FB0D09FB0D09FB0D09FB0D0\
9FB0D09FB0D09FB0D09FB0D09FB0CF90AFCF9FB0D09FB0D09FB0D09FB0D09F202F
9F00009F00009F00009F00009F8FAF9F90AF9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D04F5060\
5060704F5F6F9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D0
9FB0D09FBFD09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D0A0BFD09FB0D09FBFD09FB0D0\
A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D09FBFD0
90AFCF9FB0CF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F000000A04F509F4F50A000009F0000A000009F0000A000009FB0CFA0BFD09FBFCF\
505F6F4F5F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F1010A0BFCFA0BFD0A0BFCFA0C0D0A0BFCFA0BFD0A0BFCFA0C0D0A0BFCFA0BFD0A0BFCF\
505F604F5F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000050606FA0BFCFA0C0D0A0BFCFA0C0CFA0BFCFA0C0D0A0BFCFA0C0CF
A0BFCFAFC0D040BF2F40C02F20BF0020C00020BF0020C00020C00020C00060BF7F60C07FA0C0CF\
AFC0D0A0BFCF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007F00007F00009F00009F00009F0000000000\
0000000000000000000000000000000000000000000000000000000F0F0F90AFCF
9FAFD09FB0D09FB0DF9FB0D09F2F309F00009F00009F6F809FB0DF9FB0D09FB0DF9FB0D09FB0D0\
9FB0D09FB0DF9FB0D09FB0D09F4F509F4F5F9F90AF9FB0D09FB0D09F90AF9F0000
9F00009F0000A090AF9F90AF9FB0D09FB0D0A0BFDF9FB0D09FB0D09FB0D0A0BFDF4F5F604F5F6F\
9FB0D0A0BFDF9FB0D09FBFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0
A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD0\
9FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0
A0BFD090B0CF0F0F0F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000F0F
0F0F0F9F4F50A0505FA00000A000009F0000A00000A00000A000009F4F50A0BFD0A0BFD0A0BFD0\
9FBFD0505F6F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000101010A0BFD0A0C0D0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCFA0C0D0505F6050606F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F606FA0BFCFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CF
AFC0D040C02F20C00020C00020C00020C00020C00020C0006FC07F60C07FAFC0D0A0C0CFAFC0D0\
A0C0CFAFC0D09FB0BF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000009000009F00009000005F00005F0000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
90AFCF90AFCF9FB0D09FB0D09040509F00009F00009F4F5F9040509FB0D09FB0D09FB0D090B0D0\
9FB0D09FB0D09FB0D09FB0D090AFC09F00009FB0D09FB0D09FB0D09F00009F0000
9F8FA09F90AF9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D04F50600000004F5F6F\
9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D0
9FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D09FBFD09FB0D0\
9FBFD09FB0D09FBFD09FB0CF9FBFD09FB0D0A0BFD09FB0CF9FBFD09FB0D0A0BFD0
9FB0CF9FB0CF0000000F0F0F000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F000000A04F50
9F4F50A000009F0000A000009F0000A000009F0000A000009F0000A000009FBFCFA0BFD09FB0CF\
A0BFD04F5F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0FA0BFD09FBFCFA0BFD0A0BFCFA0BFD0A0BFCF505F604F5F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000050606FA0C0CFA0BFCFA0C0D0A0BFCFA0C0CFA0BFCFA0C0D0
A0BFCF20C00020BF0020C00020BF0020C00060BF7F60C07FA0BFCFA0C0CFA0BFCFAFC0CFA0BFCF\
A0C0CFA0C0CF40C02F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000009F00009000007F0000000000000000000000
0000000000000000000000000000000000000000000000002000009F00009F00009F0000400000\
3F0000000000000000000000000000000000000000000000000000000000000000
0F0F0F90AFCF9FB0DF9FB0D09F70809F6F809F00009F00009F2F309FB0D09FB0D09FB0D09FB0DF\
9FB0D09FB0D09FB0D09FB0DF9F202F9F00009FB0D09F90AF9F202F9F90AF9F8FAF
9FBFDF9FB0D09FB0D09FB0D09FBFDF9FB0D09FB0D09FB0D0505F6F4F5F60000000000000000000\
505F6F9FBFD09FB0D0A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D09FBFD09FB0D0
A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD0\
9FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0
A0BFD09FB0D09FB0CF00000F000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F9F4F50A04F5F9F0000
A000009F0000A000009F0000A000009F0000A000009F0000A00000A0202FA0BFD09FBFD0A0BFD0\
A0BFD0A0BFD04F5F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000001010100F0F0FA0C0D0A0BFCFA0C0D04F5F6050606F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000050606FA0C0D0A0BFCFAFC0D0A0C0CFA0C0D0A0BFCF
AFC0D0A0C0CF20C00020C0006FC08060C07FAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0D0\
A0C0CFA0BFCF20C00020C000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000009F0000900000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004000007000009F0000900000\
9F0000000000000000000000000000000000000000000000000000000000000000
00000090AFCF90AFD09FB0D090B0D09F6F809000009F00009F000090AFCF90AFCF9FB0D09FB0D0\
9FB0D090B0D09FB0D09FB0D09F4F5F9000009F202F9FB0D09F90AF90B0D09FB0D0
9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D04F50604F5F6F000000000000000000000000000000\
50606F9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D0
9FB0D09FB0D09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D09FBFD09FB0CF9FB0D09FB0D0\
9FBFD09FB0CF9FB0D09FB0D09FBFD09FB0CF9FBFD09FB0D09FBFD09FB0CF9FBFD0
9FB0D09FBFD090AFC00F0F0F000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F9F4F50A000009F0000A00000
9F0000A000009F0000A000009F0000A000009F0000A000009F0000A090AF9FB0CFA0BFD09FBFCF\
A0BFD09FB0CF4F5F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F10A0BFCF505F604F5F60000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000050606FA0BFCFA0C0CFA0BFCFA0C0D0A0BFCFA0C0CF
A0BFCFA0C0CF60BF7F60C07FA0BFCFA0C0CFA0BFCFA0C0CFA0BFCFA0C0CFA0BFCFA0C0CFA0BFCF\
A0C0CF40BF2F20C00020BF00000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000009000009F00005F0000600000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000009F00009F00009F0000\
9000000F0F0F000000000000000000000000000000000000000000000000000000
00000000000F9FAFD09FB0D09FB0D09F8FAF9F00009F00009F00009F00009F90AF9FB0D09FB0D0\
9FB0D09FB0DF9FB0D090AFCF9F00009F00009F4F509FB0D09FB0D09FB0DF9FB0D0
9FB0D09FB0D09FB0D09FB0D04F5F6F4F5060000000000000000000000000000000000000000000\
0000005060709FB0D09FB0D09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D0
9FBFD09FB0D0A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D09FBFD0\
9FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0
A0BFD09FB0D09FB0CF90B0CF0F0F0F000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F00000FA04F5F9F4F50A000009F0000A000009F0000
A000009F0000A000009F4F50A000009F0000A000009F0000A000009F202FA0BFD09FBFD0A0BFD0\
9FBFCFA0BFD04F5F60505F6F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F505F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000050606FAFC0D0A0C0CFA0C0D0A0BFCFAFC0D0A0C0CF
A0C0D0A0BFCFAFC0D0A0C0CFA0C0D0A0BFCFAFC0D0A0C0CFA0C0D0A0C0CFAFC0D0A0C0CF60C07F\
60C07F20C00020C00020C000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
9F00009000005F00005F00007F00003F0000400000000000000000000000000000
0000000000000000000000000000000000000000000000007000007F00007F00009F0000900000\
9F00007F0000000000000000000000000000000000000000000000000000000000
0000000F0F0F90AFCF9FB0D090AFD09FB0D090B0D09F2F2F9000009F0000906F7F9FB0D090AFD0\
9FB0D090B0D09F2F2F9000009F90AF9F405090AFCF90AFCF9FB0D09FB0D09FB0D0
90B0D09FB0D04F50604F5F6F000000000000000000000000000000000000000000000000000000\
0000004F5F6F9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D0
9FB0CF9FB0D09FB0D09FB0D09FB0CF9FB0D09FB0D09FB0D09FB0CF9FB0D09FB0D09FBFD09FB0CF\
9FB0D09FB0D09FBFD09FB0CF9FB0D09FB0D09FBFD09FB0CF9FB0D09FB0D09FBFD0
9FB0CF9FBFD09FB0CF9FB0CF000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F9F4F509F4F509F0000A000009F00009F00009F0000A00000
9F00009F707F9F6F7FA04F509F00009F00009F0000A000009F0000A0707F9FB0CFA0BFD09FB0CF\
A0BFD09FBFCFA0BFD04F5F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F000000000000000000000000000000000000000000000000000000000000
00000000000050606F50606F90AFBF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000050606FA0C0CFA0BFCFA0C0CFA0BFCFA0C0CF
A0BFCFA0C0CFA0BFCFA0C0CFA0BFCFA0C0CFA0BFCFA0C0CFA0BFCFA0C0CF60BF7F60C07F20BF00\
20C00020BF0020C00020BF00209F00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000009000009F0000\
000F0F0F0F0F00000F0000000000004000003F00007F0000000F0F400000000000
0000000000000000000000000000000000000000007F00007F00002000002000003F00009F0000\
9F00009F00007F0000000000000000000000000000000000000000000000000000
000000000F0F0F0F0F90AFCF9FB0DF9FB0D09FB0D09F202F9F00009F00009F00009F202F9FB0DF\
9F202F9F2F2F9F8FAF9F90AF9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D0
505F6F4F5F60000000000000000000000000000000000000000000000000000000000000000000\
0000005060704F5F6F9FBFD09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D0
9FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D0A0BFD0\
9FB0D09FBFD09FB0D0A0BFD09FB0D09FBFD09FB0D0A0BFD09FB0D09FBFD09FB0D0
A0BFD09FB0D0A0BFD090AFCF0F0F0F000F0F0000000000000000000000000000000000000F0F0F\
000F0FA04F5F9F4F50A000009F0000A000009F0000A000009F0000A000009F6F7F
A070809FBFD0A0BFD090B0C0A000009F0000A000009F0000A000009FBFD0A0BFD09FBFCFA0BFD0\
9FBFD0A0BFD09FBFCF505F6F000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000F000000A000006000009F0000600000\
8000000F0F0F000000200000000000000000000000000000000000000000000000
5F606F50606FA0BFD0A0BFCF9FB0C0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F606FA0BFCFA0C0D0A0BFCFA0C0D0A0BFCF
A0C0D0A0C0CFA0C0D0A0BFCFAFC0D0A0C0CFA0C0D0A0BFCF60C08060C07F20C00020C00020C000\
20C00020C000209F00209F00000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000009000009000003F0000\
2000000000000000000000000000000000000000000000004000007000007F0000
3F00000000000000000000000000007F00007000002000001F00002000000000003F00005F0000\
9F00009000003F00003F0000000000000000000000000000000000000000000000
00000000000000000090AFCF90AFD09FB0D090AFD09FB0D0906F7F9F00009000009F000090202F\
9F90AF908FA09FB0D090B0D09FB0D090AFD09FB0D090B0D09FB0D04F50604F5F6F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000505F6F9FB0D09FB0D090B0CF9FB0D09FB0D09FB0D09FB0CF9FB0D0
9FB0D09FB0D09FB0CF9FB0D09FB0D09FB0D09FB0CF9FB0D09FB0D09FB0D09FB0CF9FB0D09FB0D0\
9FB0D09FB0CF9FB0D09FB0D09FBFD09FB0CF9FB0D09FB0CF9FBFD09FB0CF9FB0D0
9FB0CF9FBFD09FB0CF9FBFD090AFC00F0F0F0000000000000000000000000000000F0F0F9F4F50\
9F4F509F00009F00009F00009F00009F00009F00009F00009F70809F6F7F9FBFD0
9FB0CFA0BFD09FB0CF9F707F9F0000A000009F00009F00009F0000A070809FB0CF9FBFD09FB0CF\
A0BFD09FB0CF9FBFD09FB0CF4F5F60000000000000000000000000000000000000
0000000000000000000000000000002000001F0000A000009F00004000005F00004000009F0000\
A000009F0000A000003F0000000000000000000000000000000000505F6F50606F
9FBFCFA0BFD09FBFCFA0BFCF90AFBF9FB0C0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000050606FA0BFCFA0C0CFA0BFCFA0C0CF
A0BFCFA0C0CFA0BFCFA0C0CFA0BFCFA0C0CF60BF7F60C07F20BF0020C00020BF0020C00020BF00\
20C000209F00000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000009FB0DF4F506F9F00009000009F00009000005F0000\
0000000000000000000000000000000000000000000000000000000000003F0000
4000007000000F0F0F7F00007F00001F00002000002000000000000000000000000000005F0000\
7F00009F0000900000200000000000000000000000000000000000000000000000
0000000000000F0F0F90AFCF90AFCF90B0D09FB0DF9FB0D09F4F5F9000009F00009F8FAF9F90AF\
90B0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D04F5060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000505F6F9FB0D09FB0D09FBFD09FB0D09FB0D09FB0D09FBFD09FB0D0
9FB0D09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD0\
9FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D0A0BFD09FB0D0
9FBFD09FB0D0A0BFD09FB0D09FB0CF00000F0000000000000F0F0F00000FA04F5F9F4F509F0000\
9F0000A000009F0000A000009F0000A000009F0000A070809FB0CFA0BFD09FBFD0
A0BFD09FB0CFA0BFD09F202FA000009F0000A000009F0000A0000090B0C0A0BFD09FBFD0A0BFD0\
9FBFCFA0BFD09FBFD0A0BFD04F5F60000000000000000000000000000000000000
0000000000000000000000000000009F0000A00000600000200000000000000000000000000000\
200000A000009F000080000000000000000000000050606F50606FA0BFD0A0BFCF
A0BFD09FBFCFA0C0D0A0BFCFA0BFD090AFBF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000050606FA0C0D0A0BFCFA0C0D0A0BFCF
A0C0D0A0BFCFA0C0D0A0C0CF60C07F60BF7F20C00020C00020C00020BF0020C00020C000209F00\
209F00000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000090AFD090AFD09000009000009000009000009000009F00003F0000\
3F0000000000000000000000000000000000000000000000000000000000000000
0000000000003F00002000009000003F0000000000000000000000000000000000000000000000\
9F00009000009F0000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F90AFCF9FB0D090AFD09F202F9000009F8FAF908FA09FB0D090AFD0\
9FB0D090AFD09FB0D090AFD09FB0D090B0D04F5F6F4F5060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004F5F6F9FB0D09FB0D09FB0D090B0CF9FB0D09FB0D09FB0D0
90B0CF9FB0D09FB0D09FB0D090B0CF9FB0D09FB0D09FB0D09FB0CF9FB0D09FB0CF9FB0D09FB0CF\
9FB0D09FB0CF9FB0D09FB0CF9FB0D09FB0CF9FBFD09FB0CF9FB0D09FB0CF9FBFD0
9FB0CF9FB0D09FB0CF9FBFD09FB0CF90B0CF0000000F0F0F90AFC09F707F9F00009F00009F0000\
9F00009F00009F00009F00009F00009F6F7F9F70809FB0CF9FBFD09FB0CF9FBFD0
9FB0CF9FBFD09FB0CF9F90AF9F00009F00009F00009F00009F00009FBFD09FB0CFA0BFD09FB0CF\
9FBFD09FB0CFA0BFD09FB0CF4F5F604F5F60000000000000000000000000000000
0000000000000000000F0F0F7F0000A000009F0000000000000000000000000000000000000000\
0000009F0000A000009F0000800000505F6F50606F9FBFCFA0BFCF9FBFCFA0BFD0
9FBFCFA0BFCF9FBFCFA0BFD09FBFCF90B0C0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000050606F506060A0C0CFA0BFCFA0C0CF
9FBFCFA0C0CFA0BFCFA0C0CF90AFBF20C00020BF0020C00020BF0020C000209F00209F00000000\
000000000000002F00000000207000004F00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000009FB0DF90AFD09F00009000009F8FAF90202F9F6F809000009F00009000009F0000\
3F0000400000000000000000000000000000000000000000000000000000000000
0000000000000000001F00005F00005F0000400000000000000000000000000000000000000000\
9000009F00009000009F00000000000000000000000000002000007F00007F0000
7F0000000F0F0F0F0F00000F90AFCF90B0D09F2F2F908FAF9F90AF9FB0D09FB0D090B0D09FB0D0\
9FB0D09FB0D090B0D09FB0D04F50604F5F6F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005060709FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D0
9FBFD09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD0\
9FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D0
9FBFD09FB0D09FBFD09FB0D09FBFD090AFCF9FB0CF90AFCFA0BFD09FB0D090AFC09F0000A00000\
9F00009F00009F0000A070809F6F7F9FBFD09FB0CFA0BFD09FB0D09FBFD09FB0CF
A0BFD09FBFD0A0BFD09F4F50A000009F0000A000009F0000A0000090B0CFA0BFD09FB0CFA0BFD0\
9FBFCFA0BFD09FBFCFA0BFD09FBFCF505F6F000000000000000000000000000000
000000000000000000400000A000009F0000A00000400000000000000000000000000000000000\
0000004000009F0000A0000080202FA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCF
A0BFD09FBFCFA0BFD09FBFCFA0C0D0A0BFCF9FB0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F606FA0BFCFA0C0CFA0BFCF
A0C0D0A0BFCF9FB0C090B0BF00000020BF0020C00020BF00209F00209F00000000000000000000\
000000002F0020BF0020C00020C00020C00020BF00000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD090AFD0904050904F5F900000908FAF908FAF90AFD090607F90AFD0900000900000900000\
9000009000007F0000000000000000000000000000000000000000000000000000
0000000000000000000000000000002000001F00005F00001F00000000000000000000001F0000\
3F00009000009000007000000000000000000000000000000000000000005F0000
9000007F00007000003F00009000009F0000908FA09FB0D090AFD09FB0D090AFD09FB0D090AFD0\
9FB0D090AFD04F5F6F4F5060000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004F5F6F505F6F90B0CF9FB0D090B0D09FB0D090B0CF9FB0D0
90B0D09FB0D090B0CF9FB0D09FB0CF9FB0D090B0CF9FB0D09FB0CF9FB0D090B0CF9FB0D09FB0CF\
9FB0D090B0CF9FB0D09FB0CF9FB0D09FB0CF9FB0D09FB0CF9FB0D09FB0CF9FB0D0
9FB0CF9FBFD09FB0CF90B0CF90AFC00F0F0F90AFC090B0CF9FB0CF9FBFD09FB0CF9F4F509F0000\
9F00009F00009F2F2F9FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD0
9FB0CF9FBFD09FB0CF9F00009F00009F00009F00009F00009F00009FBFD09FB0CF9FBFD09FB0CF\
9FBFD09FB0CF9FBFD09FB0CF9FBFD04F5060000000000000000000000000000000
0000000000000000000000002000004000007F00000F0F0F000000000000000000000000000000\
8000009F0000A000009F0000A09FA09FBFCFA0BFCF9FBFCFA0BFD09FBFCFA0BFCF
9FBFCFA0BFD09FBFCFA0BFCF9FBFCFA0BFD090AFBF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000050606FA0C0CF9FBFCFA0C0CF
A0BFCF9FB0C000000000000000000020C0001F9000209F00000000000000000000000000000000\
00000020BF0020C00020BF0020C00020BF0020C000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90B0DF90AFD09FB0DF90AFD090B0DF90AFD09FB0DF90AFD090B0DF90AFD09FB0DF7F202F7F2F2F\
9000009F00009000007F0000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002000007F00001F0000200000000000000000\
3F00009F00009000005F0000000000000000000000000000000000000000200000
7F00007000009F0000900000200000908FAF9FB0D090B0D09FB0D090B0D09FB0D090B0D09FB0D0\
4F50604F5F6F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F5F6F9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D0
9FB0D09FB0D09FBFD09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D0\
9FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D09FBFD09FB0D0
9FBFD090AFC09FB0CF000F0F0F0F0F0000000F0F0F90AFCF9FBFD09FB0CF9FBFD09FB0D09FBFD0\
9F202F9F00009F202F9F90AF9FB0CFA0BFD09FB0D09FBFD09FB0CFA0BFD09FB0D0
9FBFD09FB0CFA0BFD09F6F7F9F00009F0000A000009F00009F4F509FB0CFA0BFD09FBFCFA0BFD0\
9FB0CFA0BFD09FBFCFA0BFD09FBFCFA0BFD04F5F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000400000600000\
9F0000A000009F0000A000009F6F7FA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCF
A0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCF9FB0C090AFC0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000050606FA0C0D090B0BF
9FB0C0000000000000000000000000000000000000000000000000000000000000000000000000\
00000020C00020C00020C00020BF0020C000004F0020C000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD070202F\
7F202F9000009000009000007F0000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002000005F0000200000000000\
5F00009000009F0000900000200000000000000000000000000000000000000000
0000003F00009000003F000000000090AFCF8FAFCF9FB0D090AFD09FB0D090AFD04F5F6F405060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000505F6F90B0CF9FB0D090B0CF9FB0D090B0CF9FB0D0
90B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF\
9FB0D09FB0CF9FB0D090B0CF9FB0D09FB0CF9FB0D090B0CF9FB0D09FB0CF90B0CF
90AFC00F0F0F0000000000000000000000000000009FB0CF9FB0CF9FB0D09FB0CF9FBFD09FB0CF\
9F6F7F9F202F9F00009F00009F4F508FAFC09FBFD09FB0CF9FBFD09FB0CF9FBFD0
9FB0CF9FBFD09FB0CF9F00009F00009F00009F00009F00009F00009FBFD09FB0CF9FBFD09FB0CF\
9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF4F5F60000000000000000000000000
0000000000000000000000000000000000000000000000000000004000003F0000A000009F0000\
9F00009F0000A0000090B0C09FBFCF9FBFCFA0BFD09FB0CF9FBFCF9FBFCFA0BFD0
9FBFCFA0BFCF9FBFCFA0BFD09FBFCFA0BFCF9FBFCF90B0C0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000050606F90AFBF000000
000000000000000000000000000000000000000000000000000000000000004F00207000000000\
00000020BF0020C0001FBF00209F00000000000000002000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
9FB0DF90AFD090AFD090AFD09FB0DF90AFD090B0D090AFD09FB0DF90AFD090B0D090AFD09FB0DF\
70202F7F2F2F9000009F00009000009000001F0000200000000000000000000000
0000000000000000000000000000000000000000000000000000000000007F00005F0000000000\
2000007F00009000009F0000000F0F000000000000000000000000000000000000
0000000000002000001F0000000000000F0F90AFCF90AFD09FB0D04F50604F5F6F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000050606F90B0D09FB0D09FB0D09FB0D09FB0D0
9FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FB0D09FBFD09FB0D09FB0D09FB0D09FBFD0\
9FB0D09FB0D09FB0D09FBFD09FB0D09FBFD09FB0CF9FBFD090AFCF9FB0CF00000F
0F0F0F00000000000000000000000000000000000000000F9FB0CF9FB0D09FBFD09FB0CF9FBFD0\
9FB0D09FBFD09F6F7F9F00009F00009F70809FB0CF9FBFD09FB0CF9FBFD09FB0CF
9FBFD09FB0CF9FBFD09F0000A000009F00009F00009F0000A04F5F9FBFCF9FBFD09FB0CFA0BFD0\
9FBFCF9FBFD09FB0CFA0BFD09FBFCF9FBFD09FB0CF505F6F000000000000000000
0000000000000000000000000000000000000000003F00004000009F0000A000009F0000A00000\
9F0000A000009F6F7FA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCF
A0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD090AFBF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F1010104F5F60506060000000
00000000000000000000000000000000000000000000000000000000000020C00020BF00000000\
00000000000020BF0020C000209F00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0CF90AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD090AFD0\
90AFD070202F5F4F509000009000009000009000001F0000200000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000700000\
5F0000900000900000900000900000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F8FA0C04F50604F5060000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F5F6F9FB0D090B0CF9FB0D090B0CF9FB0D0
90B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF\
9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF90AFCF90AFC00F0F0F000000000000
0000000000000000000000000000000000000000000F0F0F90AFC09FBFD090B0CF9FB0D09FB0CF\
9FBFD09FB0CF90B0CF9F4F509F00009F00009F00009F6F7F9FBFD09FB0CF9FBFD0
9FB0CF9FBFD09FB0CF9F2F2F9F00009F00009F00009F00009F90A09FBFD09FB0CF9FBFD09FB0CF\
9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFCF4F5F60000000000000000000
0000000000000000000000000000000000001F00009F00009F00009F00009F00009F00009F0000\
9F2F2F9F4F509FBFCF9FBFCF9FBFD09FB0CF9FBFCF9FBFCFA0BFD09FB0CF9FBFCF
9FBFCFA0BFCF9FBFCF9FBFCF9FBFCFA0BFCF9FBFCF9FBFCF90AFBF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F10100F0F0FA0707F9F707FA02F2F4F5F60000000
000000000000000000000000000000000000000000000000000000209F001FBF0020C00020BF00\
002F0000000020C00020BF00209F001F9000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
906F8090607F90AFD090AFD090AFD090AFD090B0DF90AFD090AFD090AFD090B0DF90AFD090AFD0\
90AFD090B0DF90AFD05F4F5F9000009000009000009000003F0000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000400000\
7000009F00009000009F0000700000000000000000000000000000000000000000
00000000000000000000000000000000000F5060704F5060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005060704F5F6F9FB0D090B0D09FB0D09FB0D0
9FB0D090B0D09FB0D09FB0D09FB0D090B0D09FB0D09FB0D09FB0D09FB0CF9FBFD09FB0D09FB0D0\
9FB0CF9FBFD09FB0D09FB0D090AFC09FB0CF000F0F0F0F0F000000000000000000
000000000000000000000000000000000000000000000F0F0F0F0F90AFC09FBFD09FB0CF9FBFD0\
9FB0CF9FBFD09FB0CF9FBFD08FAFC09F4F5F9F00009F2F2F9F90A09FBFD09FB0CF
9FBFD09FB0CF9FBFD09F00009F00009F00009F00009F00009F2F2F9FB0CF9FBFD09FB0CF9FBFD0\
9FB0CFA0BFD09FBFCF9FBFD09FB0CFA0BFD09FBFCF4F5F6F000000000000000000
0000000000000000000000000000007F202FA000009F0000A000009F0000A000009F202FA02F2F\
9FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCF
A0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCF90B0C0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F0F10109F707FA0707FA00000A000009F000050606F000000
0000000000000000000000000000000000000000000000000F0F0F209F0020C00020BF00207000\
00000F00000000000020C00020BF00209F00000000000000004F00000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8F0000900000900000906F8090AFD08FA0CF90AFD090AFD090AFD090AFD090AFD090AFD090AFD0\
90AFD090AFD090AFD090AFD04F506F900000900000900000900000500000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F00007F00009000003F0000000000000000000000000000000000000000
0000000000000000000000004F5F6F4F5F6F8FA0CF0F0F0F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000505F6F90AFCF9FB0D090B0CF9FB0D0
90B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF\
9FB0D090B0CF9FB0D090AFC00F0F0F000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000090B0CF9FB0CF9FB0D090B0CF\
9FB0D09FB0CF9FBFD090B0CF9FB0D09F90A09F00009000009F00009F4F5090AFC0
90B0CF9FB0D09F90A09F00009F00009F00009F00009F00009F90A09FBFD09FB0CF9FBFD09FB0CF\
9FBFCF9FB0CF9FBFD09FB0CF9FBFCF9FB0CF9FBFD09FB0CF4F5F60000000000000
00000000000000000050606F4F5F6F9F00009F00009F00009F00009F00009F00009FB0CF9FB0CF\
9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FBFCF9FBFCF9FB0CF9FBFCF9FBFCF9FBFCF
9FB0CF9FBFCF9FBFCF9FBFCF9FB0CF9FBFCF9FBFCFA0BFCF9FBFCF90AFC0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F100F0F0FA0707F9F6F7FA000009F0000A000009F0000A000009F0000505F60
000000000000000000000000000000000000000000000000004F0020BF0020BF0020C000000000\
0000000000000000001FBF0020C00020900000000000000020700020BF00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
900000900000900000904050900000904050906F808FA0CF90AFD090AFD090AFD090AFD090B0DF\
90AFD090AFD090AFD090B0D04F50609000009000009000009000007F0000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F00009F00005F0000000000000000000000000000000000000000
000000000000505F704F5F6F9FB0D090AFD090AFCF00000F0F0F0F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F5F6F9FB0D090B0D09FB0D090B0D0
9FB0D090B0D09FB0D090B0CF9FB0D09FB0D09FB0D090B0CF9FB0D09FB0D09FB0D090B0CF9FB0D0\
9FB0D090B0CF90AFC00F0F0F000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F90AFC09FBFD09FB0CF9FBFD0\
9FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9F70809F00009F00009F6F7F
9FBFD09FB0CF9FBFD09F00009F00009F00009F00009F202F9FBFD09FB0CF9FBFD09FB0CF9FBFD0\
9FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD04F5F60000000000000
00000000000050606F9FB0CFA04F5F9F00009F00009F0000A000009F00009F4F509FBFCFA0BFD0\
9FBFCF9FBFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCF
A0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD090AFBF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F101010\
9F707FA0707F9F0000A000009F0000A000009F0000A00000A00000A000004F5F60
000000000000000000000000000000000000000000000000209F0020BF0020C00020BF000F0F0F\
00000000000000000020C00020BF0020C000209F0000500020BF0020C000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
9000009000008F000090000090607F8FA0CF809FC090405090607F906F80809FC090AFD090AFD0\
90AFD08FAFD090AFD090AFD090AFD0405060700000900000900000900000900000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4F5F6F4F5F6F90AFCF90B0D090AFD090B0D090AFCF90AFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F5F6F9FB0D090AFCF9FB0D0
90B0CF9FB0D090AFCF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090AFC0\
90AFCF0000000F0F0F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F90AFC09FB0D090B0CF\
9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D09FB0CF9F2F2F9000009F0000
9F202F9F707F8FA0BF9F00009F00009F00009000009F00009F6F7F9FBFD090B0CF9FBFCF9FB0CF\
9FBFD09FB0CF9FBFCF9FB0CF9FBFD09FB0CF9FBFCF9FB0CF9FBFD04F5060000000
4F5F6F50606F9FB0CF90B0CF9F00009F00009F00009F00009F00009F00009F90A09FBFCF9FB0CF\
9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF
9FBFCF9FBFCF9FB0CF9FBFCF9FBFCF9FBFCF9FB0CF9FBFCF9FBFCF90B0C0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F0F0F109F707F\
A000009F0000A000009F0000A000009F0000A000009F0000A02F2F9F0000604F50
5F4F5000000000000000000000000000000000000000000000200020C0001FBF0020BF00000000\
00000000000000000000000020C0001FBF0020C00020BF0020C0001FBF00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
900000900000900000900000900000809FC090AFD090AFD08FA0CF900000904F5F9040508FA0CF\
80A0C090B0D090AFD090AFD090AFD04F5F6F700000900000900000900000900000
200000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000505F704F5F6F
90B0D090AFD09FB0D090AFD090B0D090AFD09FB0D08FAFCF0F0F0F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000050606F90B0CF9FB0D090B0D0
9FB0D090B0CF9FB0D090B0D09FB0D090B0CF9FB0D090B0D09FB0D090B0CF90B0CF90AFCF0F0F0F\
00000F000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000F9FB0CF9FB0CF9FBFD0\
9FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FB0CF9F4F50
9F00009F202F9F70809F00009F00009F00009F00009F202F9FBFD09FB0CF9FBFD09FB0CF9FBFD0\
9FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF505F6F50606F
9FBFD09FB0CF9FBFD09FBFCF9F90A09F00009F00009F00009F00009F0000A000009F6F7F9FBFD0\
9FBFCFA0BFD09FBFCF9FBFD09FBFCFA0BFD09FBFCF9FBFD09FBFCFA0BFD09FBFCF
9FBFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0BFCF9FBFCF90B0C0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F10100F0F0FA09FA09F707FA00000\
9F0000A000009F0000A000009F202FA02F2F9FBFCFA09FA09FBFCFA0707F9F707F
605050000000000000000000000000000000000000000000002F0020BF0020C00020BF0020C000\
004F000F0F0F000000002F0020BF0020C00020BF0020C00020BF0020C000002F00
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8F607F9000008F00009000008F00009000008FAFD090AFD08FAFD090AFD08F40508FA0C08F0000\
904050809FC090AFD08FAFD090AFD090AFD04F506F3F0000900000900000900000
1F0000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F5F6F4F5F6F90AFD090B0D0
90AFCF90AFD090AFD090B0D090AFCF90B0D090AFD090B0D08FA0C00F0F0F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F5F6F4F5F6F90B0CF9FB0D0
90AFCF9FB0D090B0CF9FB0D090AFCF9FB0D090B0CF9FB0D08FAFC090AFCF0000000F0F0F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F00000090B0CF90B0CF\
9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0CF90B0CF90AFC0
9000009F4F509000009F00009000009F00009F00009F707F906F7F9FB0CF9FB0CF9FBFD090B0CF\
9FBFCF9FB0CF9FBFD090B0CF9FBFCF9FB0CF9FBFD04F50604F5F604F5F6F50606F
9FB0CF9FBFCF9FB0CF9FBFCF9F90A09F00009F00009F00009F00009F00009F00009F2F2F9FB0CF\
9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF
9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FBFCF9FBFCF9FB0CF9FBFCF90AFBF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F0F10109FBFCF9FBFCF90B0C09FB0C09F4F50\
9F00009F0000A000009F202F9FBFCF9FBFCFA0BFCF9FBFCFA0BFCF9FBFCFA0BFCF
4F5F6000000000000000000000000000000000000000000000200020BF0020BF0020C0001FBF00\
209F0020BF00004F001FBF0020BF0020BF0020C0001FBF00209F00004F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD090202F900000900000900000900000906F8090AFD090AFD090AFD090AFD090AFD0904F5F\
900000904F5F9040508FA0CF90AFD090AFD0405060400000900000900000900000
9000003F0000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000505F704F5F6F90B0D090AFD090B0D090AFD0
90B0D090AFD090B0D090AFD09FB0D090AFD090B0D090AFD090AFCF000F0F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004F5F6F9FB0D090B0CF
9FB0D090B0D09FB0D090B0CF9FB0D090B0D090AFCF90AFC00F0F0F000F0F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F90AFC09FBFD0\
9FB0CF9FBFD090B0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF
9FBFD09F90A09F00009F00009F00009F00009F00009F00009FB0CF9FB0CF9FBFD09FB0CF9FBFD0\
9FB0CF9FBFD09FB0CF9FBFD09FB0CF4F5F604F5F600000000000000000004F5F6F
9FBFD09FB0CF9FBFD09FB0CF9FBFD09FBFCF9F4F509F00009F00009F00009F00009F00009F0000\
9F6F7F9FBFD09FB0CF9FBFD09FBFCF9FBFD09FBFCFA0BFD09FBFCF9FBFCF9FBFCF
A0BFD09FBFCF9FBFCF9FBFCFA0BFD09FBFCF9FBFCF9FBFCFA0BFD09FBFCF9FBFCF90AFBF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000001010100F0F0FA0BFCF9FBFCFA0C0CF9FBFCFA0BFCF9FBFCF9FB0C0\
9F2F2FA000009F0000A000009F0000A0707F90AFBFA0C0CF9FBFCFA0C0CF9FBFCF
A0C0CF4F5F60000000000000000000000000000000000000000000002F0020C00020BF0020C000\
20BF0020C00020BF0020C00020BF0020C00020BF0020C000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FAFD090AFD08F00009000008F00009000008F0000906F808FAFD090AFD08FAFD090AFD0809FC0\
90AFD0809FC09000008F405090AFD08FAFD04F506F4050602000008F0000900000
9000003F0000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F5F6F90AFCF90AFD090AFD090B0D090AFCF90AFD0
90AFCF90B0D090AFCF90AFD090AFCF90B0D090AFCF90AFD08FA0C00F0F0F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000505F6F90AFCF90B0D0
90B0CF9FB0D090AFCF90B0D08FAFC090AFCF0000000F0F0F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000090AFCF90AFC0\
9FB0D090B0CF9FB0CF90B0CF9FB0D090B0CF9FB0CF90B0CF9FB0D090B0CF9FB0CF
90B0CF9FB0D090202F9F6F7F9000009F00009000009F6F7F90B0CF9FBFD090B0CF9FB0CF90B0CF\
9FBFCF90B0CF9FB0CF4F50604F5F6000000000000000000000000000000050606F
9FB0CF9FBFCF90B0CF9FBFCF9FB0CF9FBFCF9090A09F2F2F9F00009F00009F00009F00009F0000\
9F2F2F9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF
9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FBFCF9FBFCF9FB0CF90AFBF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F0F0F109F6F7F9FB0C09FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF\
A0BFCF9F6F7F9F00009F0000A000009F00009F2F2F9FBFCFA0707F9FBFCF9FBFCF
9FBFCF4F5F600000000000000000000000000000000000000000000000001F6F0020BF0020BF00\
20C0001FBF0020BF0020BF0020C0001FBF00000000000000000000000000004F00
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FAFD090AFD09000009000008F0000900000908FAF90AFD090AFD090AFD090AFD090AFD0\
90AFD090AFD0904050900000908FA0902F3090AFD04F506F1F0000900000900000
9000009000007F0000000000000000000000000000000000000000000000000000000000000000\
0000000000004F5F6F4F5F7090AFD090B0D090AFD090B0D090AFD090B0D090AFD0
90B0D090AFD090B0D090AFD090B0D090AFD090B0D090AFD090B0D08FA0CF0F0F0F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000050607090B0D0
9FB0D090B0CF9FB0D090AFC00F0F0F00000F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000F0F90B0CF\
90B0CF9FBFD09FB0CF9FB0D090B0CF9FBFD09FB0CF9FBFD090B0CF9FBFD09FB0CF
9FBFD090B0CF9FBFD09FB0CF9F00009F00009F00008FAFC090AFC09FB0CF9FBFD09FB0CF9FBFD0\
9FB0CF4F5F6F4F5F60000000000000000000000000000000000000000000000000
50606F9FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF9F90A09F4F509F00009F00009F0000\
9F00009F00009F202F9FBFCF9FB0CF9FBFD09FBFCF9FBFCF9FB0CF9FBFD09FBFCF
9FBFCF9FBFCF9FBFD09FBFCF9FBFCF9FBFCFA0BFD09FBFCF9FBFCF9FBFCFA0BFD090AFBF90B0C0\
000000000000000000000000000000000000000000000000000000000000000000
0F10100F0F0FA0707F9F707FA000009F202FA0BFCF9FBFCFA0BFCF9FBFCFA0BFCF9FBFCFA0BFCF\
9FBFCF9FBFCF9F4F50A000009F0000A000009F0000A000009F00009FBFC09FBFCF
A0BFCF4F5F60506060000000000000000000000000000000000000000000000000002F00005000\
20BF0020C00020BF00209F0000000000000000000000000020900000500020BF00
20C00020BF00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0D090AFD08FAFD0906F808F00009000008F000090AFD08FA0D090AFD08FAFD090AFD08FAFD0\
90AFD08FAFD090AFD08FAFCF9040508F0000908FAF4050600000008F0000900000
8F0000900000700000000000000000000000000000200000000000000000000000000000000000\
4F5F6F4F5F6F90AFD08FAFCF90AFD090AFCF90AFD08FAFCF90AFD090AFCF90AFD0
8FAFCF90AFD090AFCF90B0D090AFCF90AFD090AFCF90B0D090AFCF8FAFCF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000004F5F6F90B0D0
90AFCF90AFCF8FAFC00F0F0F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F8FAFC0\
9FB0CF90B0CF9FB0D090B0CF9FB0CF90B0CF9FB0D090B0CF9FB0CF90B0CF9FB0D0
90B0CF9FB0CF90B0CF9FB0D0906F7F90B0C09000009F6F7F90B0CF9FB0CF90B0CF9FBFCF90B0CF\
4F5F60000000000000000000000000000000000000000000000000000000000000
4F5F609FBFCF9FB0CF9FBFCF90B0CF9FBFCF9FB0CF9FBFCF90B0CF90B0C09F202F9F90A0900000\
9F00009F00009F0000906F7F90B0C09FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF
9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF8FAFBF\
0000000000000000000000000000000000000000000000000000000F0F0F0F1010
9F6F7F9F707F9F00009F00009F00009F707F9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF\
9FBFCF9FBFCF9F707F9F6F709F00009F00009F00009F00009F00009F4F50A00000
90B0C09FBFCF4F5F60000000000000000000000000000000000000000000000000000000000000\
004F000000000F0F0F000000000000000000209F001F900020BF0020BF0020C000
1FBF0020BF00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FAFD090AFD08F40509000009000009000008F202F90AFD090AFD090AFD08FAFD090AFD0\
90AFD090AFD08FAFD090AFD080A0C0904F5F8F202F9000005F4050000000700000
9000009000009000009000002000003F00002000000000000000000000000000004F5F6F505F70\
90AFD090B0D090AFD090B0D090AFD090B0D090AFD090B0D090AFD090B0D090AFD0
90B0D090AFD090B0D090AFD090B0D090AFD090B0D090AFD090B0D090AFD090AFCF00000F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000505F6F4F5F6F
90AFCF000F0F0F0F0F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000F0F0F0F\
90AFC09FB0D090B0CF9FBFD090B0CF9FB0D090B0CF9FBFD09FB0CF9FBFD090B0CF
9FBFD09FB0CF9FBFD090B0CF9FBFD09FB0CF9F707F90B0CF9FBFD09FB0CF9FBFD04F5F604F5F6F\
000000000000000000000000000000000000000000000000000000000000000000
50606F505F6F9FBFD09FB0CF9FBFD09FB0CF9FBFCF9FB0CF9FBFD09FB0CF9FBFCF9FB0CF9F707F\
9F202F9F00009F00009F00009F00009F00009F6F7F9FB0CF9FBFCF9FBFCF9FB0CF
9FBFD09FBFCF9FBFCF9FBFCF9FBFD09FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF90B0C0\
0000000000000000000000000000000000000000000000000F0F0F9F707F9F6F7F
A000009F00009F00009F0000A000009F00009FBFCF9FBFCFA0BFCF9FBFCFA0BFCF9FBFCFA0BFCF\
9FBFCFA0BFCF9FBFCF90B0C09F0000A000009F0000A000009F0000A000009F0000
A0707F9FBFCF505F60000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000004F0020700020BF0020C00020BF0020C00020BF00
20C00020BF00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0D090AFD08FA0D090AFD08F202F9000008F00009040508FAFD090AFD08FA0CF90AFD08FAFD0\
90AFD08FAFCF90AFD08FAFD090AFD08FAFCF8FA0C08F405070202F1F0000700000
8F00009000008F00009000007000000000000000000000000000004F5F6F4F506F90AFD08FAFCF\
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD090AFCF90AFD08FAFCF90AFD0
90AFCF90AFD08FAFCF90AFD090AFCF90B0D08FAFCF90AFD090AFCF90B0D08FA0C00F0F0F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F0000004F5F60
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
90B0CF90B0CF9FB0CF90B0CF9FB0D090B0CF9FB0CF90B0CF9FB0CF90B0CF9FB0CF
90B0CF9FB0CF90B0CF9FB0CF90B0CF9FB0CF90B0CF9FB0CF90B0CF4F5F604F5060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000050606F90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF9FB0CF\
9FBFCF90AFC09F90A09F00009F00009000009F00009F202F9F2F2F906F7F9FBFCF
90AFC09FBFCF904F509F90A09F00009F2F2F9F202F90B0C09FB0CF9FBFCF9FB0CF9FBFCF9FB0CF\
90AFBF0000000000000000000000000000000F0F100F0F0F9F707F9F00009F0000
9F00009F00009F00009F00009F00009F4F509FB0C09FBFCF9FBFCF9FBFCF9FBFC09FBFCF9FBFCF\
9FBFCF9FBFC09FBFCF9FBFCF9F2F2F9F20209F00009F00009F00009F00009F0000
9F4F509F00009FBFC04F5F60000000000000000000000000000000000000000000000000000000\
00000000000000000000000020BF001FBF0020BF0020BF0020BF001FBF00209F00
209000002F001F9000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FAFD090AFD08FAFD09000008F00009000008F0000906F808FAFD090AFD08FAFD090AFD0\
8FAFD090AFD090AFD090AFD08FAFD090AFD090AFD08FA0C08F00007F2F30700000
7000008F00009000009000000000000000000000004F5F6F4F5F6F90AFD090B0D090AFD090AFD0\
90AFD090B0D090AFD090AFD090AFD090B0D090AFD090B0D090AFD090B0D090AFD0
90B0D090AFD090B0D090AFD090B0D090AFD090B0D090AFD090B0D090AFCF90AFCF8FAFCF0F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002000000F0F0F8FAFC090AFCF4F5060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
90AFC090B0CF90B0CF9FB0D090B0CF9FBFD090B0CF9FB0D090B0CF9FBFD090B0CF
9FB0D090B0CF9FBFD090B0CF9FBFD090B0CF9FBFD04F5F604F5F60000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F5F6F9FBFD09FB0CF9FBFCF9FB0CF9FBFD09FB0CF9FBFCF9FB0CF9FBFD09FB0CF9FBFCF\
9FB0CF9FBFD09FB0CF9F707F9F4F509FBFD09F4F509F2F2F9F00009F00009F6F7F
9F00009F202F9F00009F00009F00009F00009F00009F707F9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF\
90AFBF0000000000000000000F0F0F1010109F707F9F707F9F00009F00009F0000
9F00009F00009F2F2F9F00009F00009F6F7FA0BFCF9FBFCF9FBFCF9FBFCFA0BFCF9FBFCF9FBFCF\
9FBFCFA0BFCF9FBFCF9FBFCF9FBFCFA0707F9F0000A000009F0000A000009F0000
A000009F0000A0C0CF4F5F60000000000000000000000000000000000000000000000000000000\
000000000000000000209F0020BF0020C00020BF0020C000004F00005000000000
000000000000002F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8F607F90AFD08FA0D090AFD0809FC09000008F00009000008F607F90AFD08FAFD090AFD08FA0CF\
90AFD08FAFD090AFD08FA0CF90AFD08FAFD090AFD08FAFCF90AFD0504050900000
8F00009000000000000000000000004F5F6F4F506F90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF\
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD0
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD090AFCF90AFD08FAFCF90AFD090AFCF8FAFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F00005F0000904F508FA0C090B0D090AFCF90B0D0
405060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F8FAFC09FB0CF90B0CF90B0CF90B0CF9FB0CF90B0CF9FB0CF90B0CF9FB0CF
90B0CF9FB0CF90B0CF9FB0CF90B0CF4F5F604F5060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004F5F6F9FBFCF90B0CF9FB0CF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF\
9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF90B0C090AFC09F2F2F9F00009F0000
9000009F00009F00009F00009000009F00009F00009F000090AFC09FBFCF9FB0CF9FBFCF90B0CF\
90AFBF8FAFBF0F10100F0F0F9FBFCF9F90A09F00009F00009F00009F00009F0000
9F20209F2F2F9F6F7F9F00009F00009F00009FBFCF9FBFCF9FB0C09FBFCF9FBFCF9FBFCF9FBFC0\
9FBFCF9FBFCF9FBFCF9FBFC09FBFCF9F6F7F9F00009F00009F00009F00009F0000
9F00009F00009FBFC09FBFCF4F5F5F000000000000000000000000000000000000000000000000\
000000000000002F001FBF0020BF0020BF00209F001F6F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0CF8F405090AFD08FAFD090AFD08F607F9000008F00009000008FAFD090AFD08FAFD090AFD0\
8FAFD090AFD08FAFD090AFD08FAFD090AFD08FAFD090AFD08FAFD04F506F500000
0000000000000000000000004F5F7090AFD090AFD08FAFD090AFD090AFD090AFD08FAFD090B0D0\
90AFD090AFD08FAFD090B0D090AFD090AFD090AFD090B0D090AFD090AFD090AFCF
90B0D090AFD090B0D090AFCF90B0D090AFD090B0D090AFCF90B0D090AFD090B0D08FA0C00F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F00005F00009000009000009000009F6F8090B0CF90B0D090AFCF
4F5F6F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000F0F90B0CF90B0CF9FB0D090B0CF9FB0D090B0CF9FBFD090B0CF9FB0D090B0CF
9FBFD090B0CF9FB0D04F50604F5F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000050606F90B0CF9FBFD09FB0CF9FBFCF90B0CF9FBFD09FB0CF9FBFCF9FB0CF9FBFD0\
9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9F00009F00009F707F9F4F50
9F4F509F00009F00009F00009F2F2F9F00009F707F9F6F7F9FBFCF9FBFCF9FBFCF9FB0CF9FBFCF\
90AFBF8FA0B09FBFCF9FBFCF9FBFCF9F707F9F4F509F00009F00009F00009F202F
9FBFCF9FBFCF9F707F9F00009F00009F4F509FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF\
9FBFCF9FBFCF9FBFCFA0BFCF9FBFCF9FBFCF9F90A0A000009F00009F00009F0000
A000009F00009F2F2F9FBFCF506060000000000000000000000000000000000000000000000000\
00000000500020BF0020C00020BF0020C000000000000000000000000000000000
000000000000000000407050000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0D090AFD08F607F9060808FA0D090AFD08F202F9000008F000090202F8FA0CF90AFD08FAFD0\
90AFD08FA0CF90AFD08FAFCF90AFD08FA0CF90AFD08FAFCF90AFD08FA0CF405060
0000000000004F506F4F5F6F8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF\
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD0
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FA0C0\
0F0F0F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000500000900000900000900000900000900000906F7F90B0D090AFCF90B0D0
90AFCF4F5F60000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000090AFC090B0CF9FB0CF90B0CF90B0CF90B0CF9FB0CF90B0CF90B0CF
90B0CF4F5F60405060000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004F5F6050606F90B0CF9FBFCF90B0CF9FB0CF90B0CF9FBFCF90B0CF9FB0CF90B0CF\
9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9F707F9000009F2F2F90B0CF9FBFCF
90B0CF9FBFCF90B0CF9F707F90B0CF9F90A090B0CF9FBFCF9FB0CF9FBFCF90B0CF90AFBF8FAFBF\
0000000F0F0F9FBFCF9FB0CF9FBFCF90B0C09F2F2F9F00009F00009000009FBFCF
9FB0CF9FBFCF9F00009F00009F00009F707F9FB0C09FBFCF9FBFCF9FBFCF9FB0C09FBFCF9FBFCF\
9FBFCF9FB0C09FBFCF9FBFCF9FBFCF9FBFC09F90A09F00009F00009F00009F0000
9F00009F00009F4F4F9FBFCF4F5F60000000000000000000000000000000000000000000000000\
00000000200020BF001FBF0020BF00002000000000000000000000000000000000
00000050606F4F5F6040BF2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FAFD0908FAF8F405090AFD08FAFD09000008F00009000008F0000904F5F8FAFD090AFD0\
8FAFD090AFD08FAFD090AFD08FAFD090AFD08FAFD090AFD08FAFD090AFD0405060
4F5F6F4F506F90AFD08FAFD090AFD08FAFD090AFD08FAFD090AFD08FAFD090AFD090AFD090AFD0\
8FAFD090AFD090AFD090AFD08FAFCF90B0D090AFD090AFD08FAFCF90B0D090AFD0
90AFD08FAFCF90B0D090AFD090B0D090AFCF90B0D090AFD090B0D090AFCF90B0D090AFD08FAFCF\
00000F000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F0000\
5F0000900000900000900000900000900000900000904F5090AFCF90B0D090B0CF
90B0D0405060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F90AFC09FB0D090B0CF9FB0D090B0CF9FB0D090B0CF9FB0D04F5F60
4F5F60000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F5F6F9FBFCF90B0CF9FBFD09FB0CF9FBFCF90B0CF9FBFCF9FB0CF9FBFCF\
90B0CF9FBFCF9FB0CF9FBFCF90B0CF9F90A09F00009F000090AFC09FBFCF9FB0CF
9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FBFCF9FBFCF8FAFBF90B0C0000000000000\
0000000000000F0F0F9FBFCF9FBFCF9FBFCF9F202F9F00009F00009F707F9FBFCF
9FBFCF90B0C09F00009F00009F00009F00009FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF\
9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9F90A09F9FA09F00009F00009F0000
9F00009F0000A0707F90B0C09FBFCF4F5F60000000000000000000000000000000000000000000\
000000209F0020BF0020BF0020BF0000500000000000000000000000000050606F
50606F9FBFC0A0C0CF20BF0020C000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0CF8FAFD08FA0D090AFD08F40508F6080809FC09000008F00008F00008F607F90AFD08FA0CF\
8FAFD08FA0CF90AFD08FA0CF90AFD08FAFCF90AFD08FA0CF90AFD04050604F506F
4F506F90AFD08FAFCF90AFD08FA0CF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF\
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD0
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FA0C0\
8FAFCF000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005000005F00008F0000\
9000009000009000008F000090000090000090000090000090202F90AFCF90B0D0
90AFCF4F5060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000090AFCF8FAFC090B0CF90B0CF90B0CF90B0CF90B0CF4F5060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000050606F90B0CF9FB0CF90B0CF9FBFCF90B0CF9FB0CF90B0CF9FBFCF90B0CF\
9FB0CF90B0CF9FBFCF90B0CF9FBFCF9090A09F000090202F9F0000904F509FBFCF
90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF8FAFBF000000000000000000000000\
0000000000000F101090B0C09FBFCF9FB0CF9FBFCF9020209F4F509FB0CF9FBFCF
90B0C09F4F509F00009F00009000009F00009FB0CF9FBFCF90B0C09FBFCF9FBFC09FBFCF9FB0C0\
9FBFCF9FBFC09FBFCF9FB0C09FBFCF9FBFC09FBFCF9F909F9F00009F00009F0000
9F00009F00009F00009FBFCF9FBFC04F5F60000000000000000000000000000000000000000000\
0000001FBF0020BF001FBF0020BF0000000000000000000050606F5060609FBFCF
9FBFC060BF7F3FBF2F20BF001FBF00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FAFD090AFD08FAFD090AFD08F607F9000008F607F9000008F00009000008F80A090AFD0\
8FAFD090AFD08FAFD090AFD08FAFD090AFD08FAFD090AFD0405060000000000000
0000004F5F6F90AFD08FAFD090AFD08FAFD090AFD08FAFCF90AFD08FAFD090AFD08FAFCF90AFD0\
8FAFD090AFD08FAFCF90AFD08FAFD090AFD08FAFCF90B0D090AFD090AFD08FAFCF
90B0D090AFD090AFD08FAFCF90B0D090AFD090AFD08FAFCF90B0D090AFD090B0D08FAFCF90B0D0\
8FA0C00F0F0F000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F00005F0000900000900000900000\
90000090000090000090000090202F90000090405090000090202F90B0D090AFCF
90B0D090B0CF4F5F6F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000F90B0CF90B0CF9FB0D090B0CF4F5F6F4F5F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000050606F90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF\
9FB0CF9FBFCF90B0CF9FBFCF9FB0CF9FBFCF90B0CF9F90A09F00009F2F2F904F50
9F707F9F6F7F9FBFCF90B0CF9FBFCF9FB0CF90AFBF8FAFBF000000000000000000000000000000\
0000000000000F0F0F1010109FBFCF9FBFCF9FB0CF9F00008FAFBF9FBFCF9FBFCF
9FBFCF9F707F9F00009F00009F00009F202F9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF\
9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9F9FA09F00009F00009F0000
9F00009F00009F4F509FBFCF9FBFCF9FBFCF4F5F60000000000000000000000000000000000000\
000000002F0020BF0020C00020BF000F0F0F50606050606F9FBFCF9FBFCF5FBF7F
60C07F20BF0020C00020BF0020C00020BF00000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0D090AFD08FA0CF8FAFD08FA0D090AFD08FA0CF8F40508F00009000008F00008F00008FA0CF\
90AFD08FA0CF8FAFD08FA0CF90AFD08FA0CF405060405060000000000000000000
0000004F5F6F8FA0CF8FAFD08FAFCF90AFD08FA0CF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF\
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD0
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF\
8FAFCF0000000F0F0F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005000005F00008F00009000008F0000900000900000\
9000008F0000906F7F90405090B0D08FA0C08FAFCF900000904F508F202F90B0CF
90AFCF90B0D0405060000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F8FAFC090B0CF4050604F5F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F5F609FB0CF90B0CF9FB0CF90B0CF9FBFCF90B0CF9FB0CF90B0CF\
9FBFCF90B0CF9FB0CF90B0CF9FBFCF90B0CF9FB0CF90B0CF9F90A090B0C090B0C0
90B0CF9FBFCF90B0C09FBFCF8FAFBF90AFBF000000000000000000000000000000000000000000\
0000000000000000000F0F0F9FBFCF90B0C09FBFCF906F7F9FBFCF90B0C09FBFCF
90B0C09F2F2F9000009F00009F00009F0000906F709FBFCF9FB0C09FBFCF90B0C09FBFCF9FB0C0\
9FBFCF90B0C09FBFCF9FBFC09FBFCF90B0C09FBFCF9FBFC09F4F509F00009F0000
9F00009F2F2F90B0BF9FBFCF9FBFC09FBFCF4F5F5F000000000000000000000000000000000000\
00000000200020BF001FBF0020BF001FBF0020BF005FBF7020BF2F3FBF4F20BF00
1FBF0020BF001FBF0020BF001FBF0040BF2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FA0D090AFD08FAFD090AFD08FA0D090AFD08FAFD0906F808F00009000008F000090202F\
8FAFD090AFD08FAFD090AFD04050604F506F000000000000000000000000000000
0000000000004F5F708FAFD090AFD08FAFCF90AFD08FAFD090AFD08FAFCF90AFD08FAFD090AFD0\
8FAFCF90AFD08FAFD090AFD08FAFCF90AFD08FAFD090AFD08FAFCF90AFD08FAFD0
90AFD08FAFCF90B0D08FAFCF90AFD08FAFCF90B0D090AFCF90AFD08FAFCF90B0D090AFCF90AFD0\
8FAFCF8FAFCF000F0F000000000000000000000000000000000000000000000000
0000000000000000000000005F0000500000900000900000900000900000900000900000900000\
900000906F8090AFCF90B0D090AFCF90B0D090AFCF90B0D080A0BF9090AF90AFCF
90B0D090AFCF90B0D04F5060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000050606F4F5060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000050606F4F5F6F9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF\
90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF9FB0CF
9FBFCF90B0CF90AFC08FAFBF000000000000000000000000000000000000000000000000000000\
0000000000000000000F10109FB0CF9FBFCF9FBFCF9FBFCF9FB0CF9FBFCF9FBFCF
9F707F9F00009F00009F00009F00009F00009F707F9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF\
9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9F202F9F00009F0000
9F00009F00009F9FA09FBFCF9FBFCF9FBFC0505F604F5F60000000000000000000000000000000\
000000000000000000409F2F20BF0020C00020BF0020BF0020BF0020C00020BF00
20BF0020BF0020C00040BF2F40BF2F9FBFC0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0CF8FAFD08FA0CF90AFD08FA0CF8FAFD08FA0CF90AFD08F80A08F00008F00009000008FA0CF\
8FAFD08FA0CF90AFD0405060405060000000000000000000000000000000000000
0000000000000000004F5F6F8FA0CF8FAFD08FAFCF90AFD08FA0CF8FAFD08FAFCF90AFD08FA0CF\
8FAFD08FAFCF90AFD08FAFCF8FAFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD0
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF\
90AFD08FA0C00F0F0F000000000000000000000000000000000000000000000000
0000000000005000005F00008F00009000008F00009000008F00009000008F00009000008F607F\
906F7F8FAFCF90B0D08FAFCF90AFCF8FAFCF8FAFCF8FAFCF90B0CF90AFCF90B0D0
8FAFCF90B0CF90AFCF4F5F60000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F5F6F4F5F6F8FAFC00F0F0F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F5F6F90B0CF9FB0CF90B0CF90B0CF90B0CF9FB0CF90B0CF\
9FB0CF90B0CF9FBFCF90B0C09FB0CF90B0CF9FBFCF90B0C09FB0CF90B0CF9FBFCF
8FA0BF8FAFBF000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F1090B0C09FBFCF90B0C09FBFCF90B0C09FBFCF
90AFBF9F00009000009F00009000009F00009000009F90A090B0C09FBFCF90B0C09FBFCF90B0C0\
9FBFCF90B0C09FBFCF90B0C09FBFCF9FBFC09FBFCF90B0C090AFBF9F00009F0000
9000009F90A09FBFC09FBFCF90B0C09FBFCF9FBFC04F5F60000000000000000000000000000000\
0000004F5F6050606F90B0BF40BF2F1FBF0020BF001FBF0020BF001FBF0020BF00
1FBF0040BF2F3FBF2F9FBFCF9FBFC09FBFCF9FBFC0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FAFD090AFD08FA0D090AFD08FAFD090AFD08FA0D090AFD08F00009000008FAFD090AFD0\
8FAFD04F506F405060000000000000000000000000000000000000000000000000
0000000000000000000000004F5F708FAFD090AFD08FAFCF90AFD08FAFD090AFD08FAFCF90AFD0\
8FAFD090AFD08FAFCF90AFD08FAFD090AFD08FAFCF90AFD08FAFCF90AFD08FAFCF
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90B0D08FAFCF90AFD08FAFCF90B0D0\
8FAFCF8FAFCF8FA0C00F0F0F000000000000000000000000000000000000000000
0000005000009000008F00009000009000009000008F0000900000900000906F7F906F7F90B0D0\
90AFCF90B0D090AFCF90B0D090AFCF908FAF908FA090B0D090AFCF90B0D090AFCF
90B0D090AFCF90B0D04050604F5F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000050606F\
4F5F6F90B0CF90B0CF90B0CF000F0F0F0F0F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F5F6F9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF\
90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0CF90AFBF8FAFBF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F9FBFCF90B0CF9FBFCF9FBFCF9FBFCF90B0CF
9F707F9F00009F00009F00009F00009F00009F00009F90A09FBFCF9FBFCF9FBFCF9FBFCF9FBFCF\
9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFC09FBFCF9F4F509F00009F0000
9F00009FBFCF9FBFCF9FBFC09FBFCF9FBFCF9FBFCF4F5F6000000000000000000000000050606F\
50606F9FBFCF9FBFC060BF7F5FBF7F20BF0020BF0020BF0020BF0020BF003FBF2F
40C02F9FBFC09FBFCF9FBFC09FC0CF20BF0020BF00000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0CF8FAFD08FA0CF8FAFD08FA0CF8FAFD08FA0CF8F8FAF8F607F8FAFD08FA0CF8FAFD08FA0CF\
40506F405060000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F5F6F8FA0CF8FAFD08FAFCF90AFD08FA0CF8FAFD08FAFCF\
90AFD08FA0CF8FAFD08FAFCF90AFD08FAFCF8FAFD08FAFCF90AFD08FAFCF8FAFD0
8FAFCF90AFD08FAFCF8FAFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF\
90AFD08FAFCF8FA0CF0000000000000000000000000000000000000000005F0000
5000009000008F00009000008F00009000008F00009000008F607F906F7F8FAFCF90AFCF8FAFCF\
90B0D08FAFCF90AFCF8FAFCF90B0D08FAFCF9000008F8FA090B0CF8FAFCF90B0CF
8FAFCF90B0CF8FAFCF90B0CF405060000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F5F604F5F6F90AFCF\
90B0CF90AFCF90B0CF90B0CF90AFC0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F5F6090B0CF90B0CF9FB0CF90B0C090B0CF90B0CF\
9FB0CF90B0C090B0CF90B0CF9FB0CF90B0C090B0CF8FAFBF8FAFBF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F10100F0F0F9FBFCF90B0C09FBFCF90B0C09FBFCF
904F509F4F509000009F00009000009F00009000009F90A090B0C09FBFCF90B0C09FBFCF90B0C0\
9FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09F00009000009F0000
90202F9FBFCF90B0C09FBFCF9FBFC09FBFCF90B0C09FBFCF4F5F600000004F5F6050606F9FBFC0\
9FBFCF5FBF705FBF7F1FBF0020BF001FBF0020BF001FBF0040BF2F3FBF209FBFCF
9FBFC09FBFCF9FBFC020BF001FBF0020BF001FBF0020BF00000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FA0D090AFD08FAFD090AFD08FA0D090AFD0809FC090AFD08FA0D090AFD04050604F506F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F506F90AFD08FAFD090AFD08FAFCF90AFD08FAFD090AFD0\
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90B0D08FAFCF90AFD0\
8FAFCF90B0D08FA0C00F0F0F00000F0000000000000000001F00005F00008F0000
9000008F00009000009000009000008F0000906F80906F7F90B0D08FAFCF90B0D090AFCF90B0D0\
8FAFCF90B0D090AFCF90B0D08FAFCF906F80900000908FA090AFCF90B0D090AFCF
90B0D090AFCF90B0D090AFCF4F5F60000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F5F6F90B0D090B0CF90B0CF\
90B0CF90B0D090B0CF90B0CF8FAFC00F0F0F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000050606F90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF\
90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF8FAFBF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000010101090B0CF9FBFCF90B0CF9FBFCF9FBFCF
9FBFCF9000009F00009F00009F00009000009F00009F00009F707F90B0C09FBFCF9FBFCF9FBFCF\
9FBFC09FBFCF9FBFCF9FBFCF9FBFC09FBFCF9FBFCF9F707F9F00009F00009F90A0
9FBFCF9FBFC09FBFCF9FBFCF9FBFCF9FBFC09FBFCF9FBFCF4F5F604F60609FBFCF9FBFC060BF7F\
5FBF7020BF0020BF0020BF0020BF0020BF003FBF2F40BF2F9FBFC09FBFCF9FBFC0
9FBFCF5FBF5040BF2F20BF0020BF0020BF0020C00020BF00000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0CF8FAFD08FA0CF8FAFD08FA0CF8FAFD08FA0CF8FAFD08FA0CF8FAFD0405060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F506F90AFD08FA0CF8FAFD08FAFCF90AFD08FA0CF\
8FAFD08FAFCF90AFD08FA0CF8FAFD08FAFCF90AFD08FA0CF8FAFD08FAFCF90AFD0
8FAFCF8FAFD08FAFCF90AFD08FAFCF8FAFD08FAFCF90AFD08FAFCF8FAFCF8FAFCF90AFD08FAFCF\
8FAFCF8FAFCF90AFD080A0C00F0F0F0000000F0F0F0000009040508F4050900000
8F00009000008F00009000008F607F906F7F8FAFCF90AFD08FAFCF90AFCF8FAFCF90AFD08FAFCF\
90AFCF8FAFCF90B0CF8FAFCF9000008F00009000008F0000908FA08FAFCF90B0CF
8FAFCF90AFCF8FAFCF90B0CF8FAFCF4F5060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000500000200000000000000000
0000000000000000000000000000000000000000004F5F6F4F5F6090B0CF90AFCF90B0CF90AFCF\
90B0CF90AFCF90B0CF90AFCF90B0CF8FAFC00F0F0F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F5F6050606F90B0C090B0CF90B0CF90B0CF90B0C0\
90B0CF90B0CF9FB0CF90B0C08FAFBF8FA0BF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F9FBFCF90B0C09FBFCF90B0C09FBFCF
90B0C09F00009000009F00009000009F00009000009F0000906F709FBFCF90B0C09FBFCF90B0C0\
9FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF906F709F000090202F9FBFCF
90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF4F5F604F5F604F5F6050606F8FAFBF5FBF7F1FB000\
20BF001FBF0020BF001FBF0020BF003FBF2F9FBFCF90BFC09FBFCF9FBFC09FBFCF
90BFC020BF001FBF0020BF001FBF0020BF001FBF007FBFA0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
90AFD08FA0D090AFD08FA0D090AFD08FA0D090AFD08FA0D04F506F405060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F506F90AFD08FAFCF90AFD08FAFCF90AFD0\
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD0\
8FAFCF90AFD08FAFCF8FAFCF000F0F0F0F0F8FA0C08FAFCF8FAFCF906F7F8F0000
9000008F000090202F8F202F90B0D08FAFCF90B0D08FAFCF90B0D08FAFCF90B0D08FAFCF90B0D0\
90AFCF90B0D08F00009000009000009000008F0000900000908FA090B0D08FAFCF
90B0D090AFCF90B0D08FAFCF90B0D0405060000000000000000000000000000000000000000000\
0000000000000000000000000000005F00007000009000005F0000000000000000
0000000000000000000000000000004F5F6F4F5F6F90AFCF90B0D090B0CF90B0CF90B0CF90B0D0\
90B0CF90B0CF90B0CF90B0CF90B0CF90AFC000000F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F5F6F9FBFCF90B0CF9FB0CF90B0CF9FBFCF\
90B0CF9FBFCF8FAFBF8FAFC0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F9FBFCF90B0CF9FBFCF90B0CF
9FBFCF90B0CF9F2F2F9000009F00009000009F00009000009F00009F4F509FBFCF90B0C09FBFCF\
9FBFCF9FBFCF90B0C09FBFCF9FBFCF9FBFCF9000009F00008FAFBF9FBFCF9FBFC0
9FBFCF9FBFCF9FBFCF9FBFC09FBFCF4F5F600000000000000000005060607FBFA020BF0020BF00\
20BF0020BF0020BF0040BF2F3FBF2F9FBFCF9FBFC09FBFCF9FBFC09FBFCF7FBF70
20BF2F20BF0020BF0020BF0020BF007FBF7F7FBFA09FBFC09FC0CF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8FA0CF8FAFD08FA0CF8FAFD08FA0CF8FAFD0404F60405060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004F506F8FAFD08FA0CF8FAFD08FA0CF\
8FAFD08FA0CF8FAFD08FAFCF8FAFD08FA0CF8FAFD08FAFCF90AFD08FA0CF8FAFD0
8FAFCF90AFD08FA0CF8FAFCF8FAFCF90AFD08FAFCF8FAFCF8FAFCF90AFD08FAFCF8FAFCF8FAFCF\
90AFD08FAFCF8FA0CF80A0C00F0F0F80A0C08FAFCF8FAFCF90AFD08F607F8F0000
8F000090000080A0C090AFCF8FAFCF90AFCF8FAFCF90AFCF8FAFCF90AFCF8FAFCF90AFCF8FAFCF\
90AFCF8F00009000008F00009000008F00009000008F0000908FA08FAFCF90AFCF
8FAFCF90B0CF8FAFCF90AFCF8FAFCF4F5060405060000000000000000000000000000000000000\
0000000000000000000000000000005000009000008F0000900000700000200000
0000000000000000004F5F6F4F5F6F90B0CF8FAFCF90B0CF90AFCF90B0CF8FAFCF90B0CF90AFCF\
90B0CF8FAFCF90B0CF90AFCF90B0CF8FAFC08FAFC0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F5F6F90B0CF90B0CF90B0C090B0CF90B0CF\
8FAFBF8FA0BF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F1090B0C09FBFCF90B0C09FBFCF
90B0C09FBFCF904F4F9F00009000009F00009000009F00009000009F2F2F90909F9FBFCF90B0C0\
9FBFCF90B0C09FBFCF90B0C09F707F904F4F9F707F906F709FBFCF90B0C09FBFCF
90B0C09FBFCF90B0C04F5F604F5F5F00000000000000000000000050606F7FB09F7FBFA01FBF00\
20BF003FB02040BF2F90BFC09FBFCF90B0C09FBFCF90BFC09FBFCF90BFC020BF00
1FBF0020BF001FBF0020BF2F5FBF4F9FBFCF90BFC09FBFC09FBFC0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008FA0D090AFD08FA0D08FAFD04050604F506F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F506F90AFD08FAFCF90AFD0\
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD0\
8FAFCF8FAFCF8FA0C00F0F0F00000F0F0F0F8FA0C090AFD08FAFCF90B0D08F202F
9000008F405090B0D08FAFCF90AFD08FAFCF90B0D08FAFCF90AFD08FAFCF90B0D08FAFCF900000\
8F00009000008F00009000008F00009000008F00009000008F00009090AF90AFCF
90B0D08FAFCF90B0D090AFCF90B0CF8FAFCF4F5F6F000000000000000000000000000000000000\
0000000000000000000000000000000000003F00009000009000009000003F0000
0000004F5F6F50606F90B0CF90B0CF90AFCF90B0CF90B0CF90B0CF90AFCF90AFCF90B0CF90B0CF\
90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF8FAFC00F0F0F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000050606F90B0CF9FBFCF8FAFBF8FAFBF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F101090B0C09FBFCF90B0CF
9FBFCF90B0C09F707F906F7F9F00009000009F00009000009F00009000009F000090202F9F707F\
90B0C090B0C0904F509F707F90202F9F2F2F9FBFC09FBFCF90BFC09FBFCF9FBFC0
9FBFCF4F5F604F5F6000000000000000000000000000000000000000000050606F7FBF9F40BF2F\
3FBF2F9FBFCF9FBFC09FBFCF9FBFC09FBFCF9FBFC09FBFCF9FBFC060BF5020BF00
20BF0020BF0020BF009FBFC09FBFCF9FBFC09FBFCF9FBFC09FBFCF9FBFC0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000405060405060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F5F6F8FA0CF8FAFD08FA0CF\
8FAFD08FA0CF8FAFD08FA0CF8FAFD08FAFCF8FAFD08FA0CF8FAFCF8FAFCF8FAFD0
8FA0CF8FAFCF8FAFCF8FAFD08FA0CF8FAFCF8FAFCF8FAFD08FAFCF8FAFCF8FAFCF90AFD080A0C0\
8FA0CF0000000F0F0F0000000000000000008FA0CF8FAFCF8FAFCF8FAFCF90202F
8F00008F00008FAFCF90AFCF8F80A08FA0C08FAFCF90AFCF8FAFCF8FAFCF8F00009000008F0000\
9000008F00009000008F00009000008F000090AFCF8F607F90202F8F8FA090AFCF
8FAFCF904F508F202F90B0CF8FAFCF90AFCF405060000000000000000000000000000000000000\
0000000000000000000000000000000000002000008F00009000008F0000900000
8F00008FA0BF8FAFCF906F7F8FA0C0904F5080A0BF9000008F607F90202F8FAFC08FAFC08FAFCF\
90B0CF8FAFCF90B0CF8FAFCF90B0CF90AFCF8FAFC0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004F5F608FAFBF8FA0BF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F90BFCF90B0C09FBFCF
90B0C090BFCF90B0C09FBFCF906F709F00009000009F00009000009F00009000009F0000900000\
9F4F509000009F707F9020209FBFCF90B0C09FBFCF90B0C09FBFCF90B0C04F5F60
4F505F0000000000000000000000000000000000000000000000000000004F5F607FBFA090B0C0\
9FBFCF90BFC09FBFCF90B0C09FBFCF90BFC09FBFCF90B0C09FBFC03FBF4F20BF00
1FB00020BF007FBF709FBFCF90BFC09FBFC090BFC09FBFCF90BFC09FBFC0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F5F6F8FAFCF90AFD0\
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD080A0C00F0F0F\
000F0F00000000000000000000000000000000000F8FAFCF8FAFCF90AFD08FAFCF
902F2F8F405090AFD08FAFCF902F2F8F6F7F908FA08F8FA09000008F00009000008F0000900000\
8F00009000008F00009000008FAFCF90B0CF8FAFCF90B0D08F6F7F906F7F8FAFCF
90B0D08F0000902F2F8FAFCF90B0D08FAFCF90B0CF405060000000000000000000000000000000\
0000000000000000000000000000000000000000007F000070202F9000008F0000
90000090202F904F50900000900000900000900000900000900000900000902F2F900000900000\
906F7F8FA0C090AFCF90B0CF90B0CF90B0CF90B0CF90AFCF000F0F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F1010104F5F60000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000010101090B0CF9FBFCF90B0C0
9FBFCF90B0CF9FBFCF90B0C09FBFCF90B0CF90B0C090202F9F2F2F9000009F00009000009F0000\
9000009F707F90B0C09FBFCF90BFC09FBFCF90B0C09FBFCF4F5F604F5F60000000
00000000000000000000000000000000000000000000000000000000000050606F4F5F609FBFCF\
9FBFC09FBFCF90BFC09FBFCF9FBFC09FBFCF90BFC09FBFCF5FBF7F20BF0020BF00
20BF007FBF9F7FBFA09FBFC09FBFCF9FBFC09FBFCF9FBFC09FBFCF9FBFC0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F5F6F8FA0CF\
8FAFD08FA0CF8FAFD08FA0CF8FAFD08FA0CF8FAFCF8FA0CF8FAFD08FA0CF8FAFCF
8FAFCF8FAFD08FA0CF8FAFCF8FAFCF8FAFD08FA0CF8FAFCF8FAFCF8FA0CF80A0C00F0F0F000000\
0000000000000000000000000000000000000000000000008FA0CF8FAFCF8FAFCF
8FAFCF908FA0809FBF8FAFCF8FAFCF90202F8F00008F00008F00009000008F00008F00008F0000\
9000008F00008F00008FAFCF90AFCF8FAFCF8FAFCF8FAFCF90AFCF8FAFCF90AFCF
8FA0C09000008F202F90202F8FAFCF90B0CF8FAFCF405060000000000000000000000000000000\
0000000000000000000000000000000000000000004F5F60906F7F8F0000900000
8F00009000008F0000908FA08F6F7F908FA08F00009000008F00009000008F00009000008F0000\
9000008F405090B0CF8FAFCF90B0CF8FAFCF90B0CF8FAFC00F0F0F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F90B0CF90B0C04F5F60000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F101090B0C090B0CF
90B0C09FBFCF90B0C090BFCF90B0C09FBFCF90B0C08FAFBF8FAFBF9F4F5090B0C09090A090B0C0\
8FAFBF90B0C09FBFCF90B0C09FBFCF90B0C04F5F604F5F5F000000000000000000
00000000000000000000000000000000000000000000000000000000000000000050606F90B0C0\
9FBFCF90B0C09FBFC090B0C09FBFCF90B0C09FBFC090BFC020BF001FB00020BF00
1FBF007FBF7F90B0C09FBFC090BFC09FBFCF90BFC09FBFC090BFC09FBFCF90BFC0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F5F6F\
8FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF8FA0CF00000F0F0F0F000000000000\
00000000000000000000000000000000000000000000000000000F8FAFCF8FAFCF
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF902F2F8F00009000008F00009000008F0000900000\
8F000090B0D08FAFCF90AFCF8FAFCF90B0D08FAFCF90B0CF8FAFCF90B0D08FAFCF
90B0CF8F00009000008F405090B0CF8FAFCF90B0D04050604F5F60000000000000000000000000\
0000000000000000000000000000004F5F6F4F5F6F90B0CF8F202F900000900000
9000008F0000900000906F7F90B0CF8FAFCF90B0CF90AFCF906F7F8F202F900000900000900000\
9000009000009000009090A090AFCF90B0CF90B0CF8FAFC08FAFC00F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F10100F0F0F90B0CF90B0CF90B0CF4050604F5F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F9FBFCF90B0CF
9FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF\
90B0C09FBFCF90B0C09FBFCF4F5F604F5F60000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F60609FBFCF\
90BFC09FBFCF9FBFC09FBFCF90BFC09FBFCF9FBFC090B0BF5FBF5020BF0020BF00
20BF0090BFC09FBFCF9FBFC09FBFCF90BFC09FBFCF9FBFC09FBFCF9FBFC09FBFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F5F6F8FA0CF8FAFD08FA0CF8FAFCF8FA0CF8FAFD08FA0CF8FAFCF8FA0CF8FAFD0
8FA0CF8FAFCF8FAFCF8FAFD08FA0CF8FAFCF80A0C08FA0CF000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000008FA0CF
8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8F202F8F00008F00008F00009000008FAFCF\
8FAFCF8FAFCF90AFCF8FAFCF8FAFCF8FAFCF90AFCF8FAFCF8FAFCF8FAFCF906F7F
8F202F8F00008F0000906F7F8FAFCF8FAFCF8FAFCF90AFCF405060000000000000000000000000\
0000000000000000004F5F604F5F6F8FAFCF90B0CF8F8FA09000008F0000906F7F
8F00009000008F000090202F8F202F8FA0BF8FAFCF90B0CF8FAFCF906F7F8F00009000008F0000\
9000008F00009000008F607090B0CF8FAFCF90B0CF8FAFC08FAFC0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F0F10108FB0C090B0CF90B0C090B0CF90B0C090B0CF405060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F90BFCF
90B0C090B0CF90B0C090BFCF90B0C090B0CF90B0C090BFCF90B0C090BFCF90B0C09FBFCF90B0C0\
90BFCF90B0C04F5F6040505F000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F5F60\
9FBFC090B0C09FBFCF90B0C09FBFC090B0C09FBFCF90B0C020BF2F1FBF0020BF00
1FB0205FBF5090BFC09FBFCF90B0C09FBFC090BFC09FBFC090B0C09FBFC090BFC09FBFC0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F5F6F8FAFD08FAFCF90AFD08FAFCF8FAFD08FAFCF90AFD08FAFCF8FAFD08FAFCF
90AFD08FAFCF90AFD08FAFCF8FAFCF8FA0C00F0F0F00000F000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F80A0C0
90AFD08FAFCF90AFD08FAFCF90AFD08FAFCF90AFCF8FAFCF902F2F8F00009000008FAFCF90B0D0\
8FAFCF90AFCF8FAFCF90B0D08FAFCF90AFCF8FAFCF90B0D08F8FA0906F7F8F0000
9000008F00009000008F0000906F7F8FAFCF90B0CF8FAFCF4F5F60000000000000000000000000\
0000004F5F6F4F5F6F90B0CF8FAFCF90B0CF8FAFCF9000008F0000904F508F8FA0
9000008F00009000008F00009000008F6F7F90B0CF8FAFCF90B0CF90AFCF8FAFC08F202F900000\
9000009000008F0000900000906F7F90B0CF8FAFCF90B0CF8FAFC00F0F0F00000F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
1010100F0F0F90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF4F5F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F101090B0C0
9FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF90B0C09FBFCF\
90B0C04F5F60000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000050606F\
90BFC09FBFCF90BFC09FBFCF90BFC09FBFCF90BFC09FBFCF20BF0020BF001FBF00
20BF005FBF709FBFCF90BFC09FBFCF9FBFC09FBFCF90BFC09FBFCF9FBFC09FBFCF90BFC0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000040506F8FAFCF8FA0CF8FAFD08FA0CF8FAFCF8FA0CF8FAFD08FA0CF8FAFCF
8FA0CF8FAFD08FA0CF8FA0CF80A0C00F0F0F000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
80A0C08FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8F202F8F202F8F80A08FAFCF8FAFCF\
8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8F8FA08F00008F00008F0000
8F00009000008F00008F00008F607F90AFCF8FAFCF8FAFCF8FAFCF4F50600000000000004F5F60\
4F5F6F8FAFCF8FAFCF8FAFCF90B0CF8F8FA090202F8F000090202F8F202F90AFCF
8FA0C090202F8F00009000008F000090202F8F20208FAFC08FAFCF90B0CF8FAFC0904F508F0000\
9000008F00009000008F0000906F7F8FAFC090B0CF8FAFCF90B0CF8FA0C00F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F0F0F10
90B0C090B0CF8FAFC090B0CF90B0C090B0CF8FAFC090B0CF90B0C090B0CF8FB0C04F5F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F0F0F10
90B0C090BFCF90B0C090B0CF90B0C090BFCF90B0C090B0CF90B0C090BFCF90B0C090B0CF40505F\
4F5F60000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F5F60\
50606F90B0C09FBFC090B0C09FBFCF90B0C09FBFC090B0C07FBFA01FB00020BF00
1FB0005FBF7F90B0C09FBFC090BFC09FBFC090B0C09FBFC090BFC09FBFC090B0C09FBFC0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F506F8FAFD08FA0CF90AFD08FAFCF8FAFD08FAFCF90AFD08FAFCF
8FAFD080A0C08FA0CF000F0F0F0F0F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F80A0C090AFD08FAFCF90AFCF8FAFCF90AFD08FAFCF90AFCF8F202F906F7F8FAFCF90AFCF\
8FAFCF90B0D08FAFCF90AFCF8FAFCF908FA08F8FA09000008F00009000008F0000
9000008F00009000008F8FA0908FA08FAFCF90B0CF8FAFCF90B0CF4050604F5F6F4F5F6F90B0CF\
8FAFCF90B0CF8FAFCF90B0CF8FAFCF9000008F00009000008F0000904F508FAFCF
90B0CF8F40509000008F00009000008F00009000008F6F7F90B0CF8FAFCF90B0CF8FAFCF906F7F\
8F00009000008F00009000008F6F7F8FAFC090B0CF90B0CF8FAFCF8FAFC0000F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F10100F0F0F90B0CF90B0CF
90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF90B0C090B0CF4F5F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
90BFCF90B0C09FBFCF90B0C090BFCF90B0C09FBFCF90B0C09FBFCF90B0C04F5F604F5F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F5F609FBFCF90BFC09FBFCF90BFC09FBFCF90BFC09FBFCF7FBF9F20BF0020BF00
20BF001FBF0040BF5090B0BF9FBFCF90BFC09FBFCF90BFC09FBFCF90BFC060BF7F5FBF7020BF00\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040506F8FAFCF8FA0CF8FAFD08FA0CF8FAFCF8FA0CF8FA0CF
80A0C00F0F0F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F80A0C08FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8F607F8FAFCF8FAFCF\
8FAFCF8FAFCF8FAFCF8FAFCF8F8FA08F00008F00008F00008F00008F00008F0000
8F00008F8FA08F8FA08FAFCF8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF4050604F5F608FAFCF8FAFCF\
8FAFCF8FAFCF90AFCF8FAFCF8FAFCF8F00009000008F00008F00008F607F90B0CF
8FAFC08FAFCF8F607F9000008F00009000008F0000904F508F404F8FAFC08FAFCF90B0CF8F6070\
9000008F00009000008F00009000008FAFCF90B0CF8FAFC090B0CF8FAFC08FAFC0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F0F0F108FAFC090B0CF8FAFC090B0CF
8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF90B0C04F5F6040505F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F1010
90B0C090B0CF90B0C090BFCF90B0C090B0CF90B0C090BFCF40505F4F5F60000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F606090B0C09FBFC090B0C090BFC090B0C09FBFC090B0C07FBF9F70B09F20BF00
1FB00020BF001FB0005FBF7F90B0C09FBFC090B0C08FB0BF90B0C05FBF505FBF7020BF001FB000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F506F8FAFD08FA0CF8FA0CF80A0C00F0F0F00000F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F80A0C090AFD08FAFCF90AFCF8FAFCF90AFD08FAFCF90AFCF8FAFCF90AFD0\
8FAFCF90AFCF8F8FA0908FA08F00009000008F00009000008F00009000008F0000
908FA08F8FA090AFCF8FAFCF90B0CF8FAFCF90AFCF4050604F5F600000004F5F6F4F5F6F90B0CF\
8FAFCF90B0CF8FAFCF90B0CF8FA0C09000008F00009000008F00009000008F8FA0
90B0CF8FAFCF8FAFC08F202F9000008F00009000008F00009000008F6F7F90B0CF8FAFCF904F50\
8F40509000008F00009000008F6F7F90B0CF8FAFCF90B0CF8FAFCF90B0CF8FAFC0
0F0F0F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F000050000090B0CF8FB0CF90B0CF90B0CF90B0CF90B0C0
90B0CF90B0CF90B0CF90B0C090B0CF90B0CF90B0CF90B0C090B0CF90B0CF90B0CF90B0C04F5F60\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
10101090B0C090BFCF90B0C09FBFCF90B0C04F5F604F5F60000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000050606F90B0C09FBFCF90BFC09FBFCF90B0C09FBFCF90BFC07FBFA01FBF00
20BF0020BF0020BF001FBF0020BF0020BF005FBF501FBF0040BF2F20BF0020BF001FBF0020BF00\
20BF00000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000040506080A0C00000000F0F0F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008FA0C08FAFCF8FAFCF8FAFCF8FAFCF8F202F8F4F5080A0C08FAFCF8FAFCF\
8F6F7F8F80A08F00008F00008F00008F00008F00008F00008F00008F80A08F8FA0
8FAFCF8FAFCF8FAFCF8FAFCF8FAFCF4050604050600000000000000000000000004F5F6F8FAFCF\
8FAFCF8FAFC08FAFCF8FAFCF8F4F508F00008F00008F00008F00008F00008F6F7F
8FAFCF90B0CF8FAFC08FAFCF8F40509000008F00008F00008F0000904F508F404F8FAFCF8FAFCF\
90202F8F00009000008F405090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC08FAFC0
0000000F0F0F000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F00008F0000904F508FAFC090B0CF8FAFC090B0CF8FAFC090B0CF
8FAFC090B0CF8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF405060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F90B0CF90B0C090B0C040505F4F5F60000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F5F6090BFC090B0C09FBFC090B0C090BFC090B0C09FBFC090B0C03FBF2F
1FB00020BF001FB00020BF001FB00020BF001FB00020BF001FB00020BF001FB00020BF0090BFC0\
9FBFC0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000F0F0F0F0F80A0C08FA0CF405060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000F0F8FA0CF8FAFCF90AFD08FAFCF8F202F8F0000904F508FAFCF8F6F7F\
8F607F9000008F00009000008F00009000008F00009000008F8FA0908FA08FAFCF
90AFCF8FAFCF90B0CF8FAFCF4F50604050600000000000000000000000000000000000004F5F6F\
8FAFCF90B0CF8FAFCF90B0CF8FAFCF9000008F00009000008F00009000008F6F7F
906F7F8FAFCF90B0CF8FAFCF906F7F8F00009000008F00009000008F00009000008F6F7F908FA0\
8F00009000008F0000904F508FAFCF90B0CF8FAFCF90B0CF8FAFCF90B0CF8FAFCF
8FAFC000000F000000000000000000000000000000000000000000000000000000000000000000\
5000005F00009000009000008FB0C090B0CF90B0CF90B0CF8FB0C090B0CF90B0CF
90B0CF8FB0C090B0CF90B0CF90B0CF8FB0C090B0CF90B0CF90B0CF90B0C090B0CF90B0CF90B0CF\
405060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F10100F0F0F4F5F604F5F60000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F60609FBFCF90B0C09FBFCF90BFC09FBFCF90B0C09FBFCF90BFC0
5FBF7F1FBF0020BF0020BF0020BF001FBF0020BF0020BF0020BF001FBF0020BF0090BFC08FB0BF\
8FAFB0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F809FC08FA0CF8FA0CF8FAFCF8FA0CF8FAFCF404F60000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008FA0C08FAFCF8FAFCF8FA0CF8F202F8F00008F00008F0000\
8F00008F00008F00008F00008F00008F00008F8FA08F80A08FAFCF8FAFCF8FAFCF
8FAFCF8FAFCF405060405060000000000000000000000000000000000000000000000000000000\
4F5F6F8FAFCF8FAFCF8FAFC08FAFCF8FAFCF8F00008F00008F00008F00008F0000
8F404F8FAFCF8FAFCF8FAFCF809FBF8FAFCF8F202F8F00008F00008F00008F0000904F508F0000\
8F00008F00009000008FAFC08FAFCF8FAFC090B0CF8FAFC08FB0CF8FAFC090B0CF
80A0BF0F0F0F0000000000000000000000000000000000000000000000000000005F0000500000\
9000008F00009000008F607090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF
8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FB0C090B0CF8FAFC0\
4F5F60000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F5F6F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000050606F90B0C090BFC090B0C090BFC090B0C090BFC090B0C090BFC0
8FAFBF7FBF701FB0005FBF501FB00020BF001FB00020BF001FB00090BFC090B0C08FAFBF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000F0F0F0F0F00000F8FA0CF\
80A0C08FAFD08FA0CF8FAFD08FA0CF8FAFD08FA0CF8FAFD0405060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000F0F8FA0C08FAFCF90AFCF8FAFCF8F4F508F0000900000\
8F00008F00008F00009000008F00008F8FA08FAFCF90AFCF8FAFCF8FAFCF8FAFCF
90AFCF405060000000000000000000000000000000000000000000000000000000000000000000\
0000004F5F6F8FAFCF90B0CF8FAFCF90B0CF8FAFCF902F2F8F00009000008F0000
9000008F0000906F7F8FAFCF90B0CF8FAFCF904F508F00009000008F00009000008F0000900000\
8F0000906F7F80A0BF90B0CF8FAFCF90B0CF8FAFCF90B0CF8FAFCF90B0CF8FAFCF
8FAFC08FAFC00F0F0F0000000000000000000000000000000000005000005F00008F0000900000\
8F00009000008F00009000008FB0CF90B0CF8FAFC090B0CF8FB0CF90B0CF8FB0C0
90B0CF90B0CF90B0CF8FB0C090B0CF90B0CF90B0CF8FB0C090B0CF90B0CF90B0CF8FB0C090B0CF\
4050604F5F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000004F5F60
50606F8FAFBF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F5F609FBFCF90BFC09FBFCF90B0C09FBFCF90BFC09FBFC090B0C0
9FBFCF90BFC09FBFC090B0C07FBF7F8FAFB05FBF5090BFC09FBFCF8FAFB08FB0BF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F00000080A0CF809FC08FAFCF8FA0CF\
8FAFCF80A0CF8FAFCF8FA0CF8FAFCF80A0CF8FAFCF8FA0CF8FAFCF404F60000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008FA0C08FAFCF8FAFCF8FA0CF8F6F7F8F0000\
8F00008F00008F00008F80A08F8FA08FAFCF8FAFCF8FAFCF8FAFCF8FAFCF405060
405060000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F5F6F8FAFCF8FAFCF8FAFC08FAFCF8FAFCF8F6F7F8F00008F0000
8F00008F00008F00008F202F8F60708F8FA08FAFC080A0BF8F00008F00008F00008F00008F0000\
8F00008FAFC08FAFCF8FAFC08FB0CF8FAFC08FAFCF8FAFC090B0CF8FAFC08FB0CF
8FAFC08FAFC00000000000000000000000000000003F00005000009000008F00008F00008F0000\
9000008F00009000008F000090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF
8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC0\
90B0CF40505F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004F606F4F5F6090B0C0
90B0C090B0CF80A0B0000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004F5F6090BFC090B0C090BFC090B0C090BFC090B0C090BFC0
90B0C090BFC090B0C090BFC090B0C090BFC090B0C08FAFBF8FAFB0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F00000F8FA0CF80A0C08FAFD08FA0CF8FAFD08FA0CF8FAFD0\
8FA0CF80A0C08FA0CF8FAFCF8FA0CF8FAFD08FA0CF8FAFCF8FA0CF4F506F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000F0F8FA0C08FAFCF90AFCF8FAFCF8F6F7F\
8F0000908FA08F8FA08FAFCF8FAFCF90AFCF8FAFCF8FAFCF8FAFCF4F5F60000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F5F6F8FAFCF8FAFCF90B0CF8FAFCF90AFCF80A0BF90B0CF8F4050
9000008F00009000008F00009000008F00009000008F0000906F7F8F00009000008F0000900000\
8F000090000080A0BF90B0CF8FAFCF90B0CF8FAFCF90B0CF8FAFCF90B0CF8FAFCF
90B0CF8FAFC00F0F0F000F0F0000000F0F0F0F10108F6F7F904F508F8F9F9090A08F202F900000\
8F00009000008F00009000008FAFC090B0CF8FB0CF90B0CF8FAFC090B0CF8FB0CF
90B0CF8FAFC090B0CF8FB0C090B0CF8FB0C090B0CF8FB0C090B0CF8FB0C090B0CF90B0C090B0CF\
8FB0C04F5F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F5F604F606F90B0C090BFCF90B0C0
90B0CF90B0C08FAFBF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000050606F90B0C09FBFCF90B0C090BFC090B0C09FBFCF90BFC0
9FBFC090B0C09FBFCF90BFC09FBFC08FAFB08FB0BF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F00000080A0CF809FC080A0C08FA0CF8FAFCF80A0CF8FA0CF8FA0CF8FAFCF1FA000\
1FA0001FA00020AF001FA00050AF506FA07F8FAFCF80A0CF8FAFCF8FA0CF405060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000080A0C08FAFCF8FAFCF8FA0CF\
8FAFCF8FAFCF8FAFCF8FA0CF8FAFCF8FAFCF8FAFCF405060405060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004050608FAFCF8FAFCF8FAFCF8FAFC08FAFCF8FAFC080A0BF
8F60708F00008F00008F00008F00008F00008F00008F00008F00008F6F7F8F202F8F00008F0000\
8F00008F00008F00008F20208FAFCF8FAFC08FB0CF8FAFC08FAFCF8FAFC08FB0CF
8FAFC08FAFCF80A0BF1010100F0F0F8FB0CF8FAFC090B0CF8FAFC08FB0CF8FAFC090202F8F0000\
8F00008F00009000008F00008F00008FAFC090B0CF8FAFC08FB0CF8FAFC090B0CF
8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC0\
90B0CF8FAFC04F5F60000000000000000000000000000000000000000000000000
0000000000001F00000F0F0F3F00003F00003F0000000000000000000000000000000000000000\
0000000000000000000000004F5F6F4F5F6090B0CF8FB0C090B0C08FB0C090B0CF
8FB0C090B0C080A0B08FAFBF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F606090B0C090BFC090B0C090BFC090B0C090BFC0
90B0C090BFC090B0C08FAFBF80AFB0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000000000F
0F0F0F809FC080A0CF8FA0CF8FAFD08FA0CF8FAFCF8FA0CF8FAFD01FA00020AF001FA00020AF00\
1FA00020AF001FA00020AF001FA00020AF001FA0005FAF7F8FA0CF8FAFCF8FA0CF
4F506F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F80A0C08FAFCF8FAFCF8FAFCF\
8FAFCF8FAFCF8FAFCF8FAFCF405060405060000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F5F6F8FAFCF8FAFCF90B0CF8FAFCF8FAFCF8FAFCF
90B0CF8FAFCF8FB0CF8F6F7F908FA08F202F8F6F7F8F0000906F7F8F6F7F908FA08F0000900000\
8F00009000008F00009000008F0000906F7F8FAFC090B0CF8FAFCF90B0CF8FAFC0
90B0CF8FAFCF80A0B080A0C090B0CF8FAFCF90B0CF8FAFC090B0CF8FAFCF90B0CF8F202F900000\
8F00009000008F00009000008F000090B0CF8FAFC090B0CF8FB0C090B0CF8FAFC0
90B0CF8FB0C090B0CF8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF8FB0C090B0CF8FB0C090B0CF\
8FB0C090B0CF405060000000000000000000000000000000000000000000000000
0000001F00009000008F0000900000900000700000000000000000000000000000000000000000\
0000000000004F5F6050606F90B0C090B0CF90B0C090B0CF90B0C090B0CF90B0C0
90BFCF90B0C090B0CF80A0B0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F5F6090BFC090B0C09FBFCF90B0C090BFC090B0C0
9FBFCF8FAFB08FAFBF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F809FC080A0C0
80A0CF8FA0CF80A0CF8FA0CF80A0CF8FA0CF1FA0001FA0001FA0001FAF001FA0001FA0001FA000\
3FAF2F1FA00050A0501FA0001FAF001FA0001FA0001FA00020AF0080A0CF8FAFCF
8FA0CF405060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F80A0C08FAFCF8FAFCF\
8FAFCF405060405060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004050608FAFCF8FAFC08FAFCF8FAFC08FAFCF
8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8F202F\
8F00008F00008F00008F00008F00008F60708FAFCF8FAFC08FB0CF8FAFC08FAFCF
80A0BF80A0BF0000000F0F0F80A0BF8FB0CF8FAFC08FAFCF8FAFC08FB0CF8FAFC08FB0CF8F202F\
8F00008F00008F00008F00009000008F00008FB0CF8FAFC090B0CF8FAFC08FB0CF
8FAFC090B0CF8FAFC08FB0CF8FAFC090B0CF8FAFC08FB0CF8FAFC090B0CF8FAFC090B0CF8FAFC0\
90B0CF8FAFC0405F6040505F000000000000000000000000000000000000000000
0000003F00008F00009000008F00009000008F0000900000700000000000000000000000000000\
4F5F6F4F5F6090B0C08FB0C090B0CF8FB0C090B0C08FB0C090B0C08FB0C090B0C0
8FB0C090B0C08FB0C08FAFBF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F606F4F5F6090BFC090B0C090BFC090B0C08FAFBF
8FAFB0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000809FC080A0CF8FA0CF8FAFD08FA0CF
8FAFCF8FA0CF8FAFD08FA0CF8FAFCF6FA07F20AF001FA00020AF001FA00050AF7F50A07F8FAFCF\
8FA0CF8FAFD08FA0CF8FAFCF8FA0CF8FA0CF1FA00020AF001FA00020AF006FA0A0
8FAFCF8FA0CF4F5060000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F80A0C0405060\
405060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F5F6F8FAFCF8FAFCF90B0CF8FAFCF
8FAFCF8FAFCF90B0CF8FAFCF8FAFCF8FAFCF90B0CF8FAFCF8FB0CF8FAFCF90B0CF8FAFCF8FB0CF\
8F6F7F9000008F00008F00008F00009000008F202F8FB0CF8FAFC08FA0BF80A0BF
000000000000000000000F0F8FAFC080A0C090B0CF8FAFCF90B0CF8FAFC090B0CF8FAFC0902F2F\
8F00009000008F00009000008F00009000008FAFC090B0CF8FAFC090B0CF8FB0C0
90B0CF8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF8FB0C090B0CF\
8FB0C090B0CF8FB0C04F5F60000000000000000000000000000000000000000000
0000007000009000008F00009000008F000090000090000090000000000F0000000000004F606F\
8FB0C090B0CF90B0C090B0CF8FB0C090B0CF90B0C090B0CF8FB0C090B0CF90B0C0
90B0CF90B0C090B0CF90B0C08FAFBF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000050606F90B0C090BFC090B0C08FAFBF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000080A0CF8FA0CF80A0CF8FA0CF
80A0CF8FA0CF80A0CF8FA0CF809FC01FA0001FA0001FA0001FA0008FA0CF80A0CF8FA0CF80A0CF\
8FA0CF80A0CF8FAFCF80A0CF8FA0CF80A0CF8FAFCF50A07F1FA0001FA0001FAF00
6FA0A08FAFCF80A0CF405060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F5F6F4050608FA0C0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F5F6F8FAFC08FAFCF8FAFC08FAFCF
8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC0\
8FAFCF80A0BF8F202F8F00008F6F7F8F60708FAFCF8FAFC080A0BF80A0BF000000
0000000000000000000000000000008FAFC08FAFC08FB0CF8FAFC08FAFCF8FAFC08FB0CF8FAFC0\
8F202F8F00008F00008F00008F00008F00008FB0CF8FAFC08FB0CF8FAFC08FB0CF
8FAFC08FB0CF8FAFC090B0CF8FAFC08FB0CF8FAFC090B0CF8FAFC08FB0C08FAFC090B0CF8FAFC0\
8FB0C08FAFC090B0CF40505F000000000000000000000000000000000000000000
0000000000001F00009000008F00009000008F00009000008F000090000070202F4F5F6F8FAFC0\
90B0C08FB0C090B0C08FAFC090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C0
8FB0C090B0C08FB0C090B0C080A0B0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F5F6090BFC080A0B08FAFB0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000008FA0CF8FA0CF80A0CF
8FAFD08FA0CF8FAFCF80A0CF3FAF2F1FA00020AF001FA0003FAF2F8FA0CF8FAFCF8FA0CF8FAFCF\
8FA0CF8FAFCF8FA0CF8FAFCF8FA0CF8FAFCF8FA0CF8FAFCF50A05020AF001FA000
20AF008FA0CF8FAFCF8FA0CF405060000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F5F6F40506F8FAFCF8FAFCF8FAFCF8FAFCF8FA0C0\
000F0F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004F5F6F8FAFCF8FAFCF8FAFCF
90B0CF8FAFCF8FAFCF8FAFCF90B0CF8FAFCF8FAFCF8FAFC090B0CF8FAFCF8FB0CF8FAFC090B0CF\
8FAFCF8FB0CF8FAFC090B0CF8FAFCF8FB0CF80A0BF80A0BF000000000000000000
000000000000000000000000000000000F0F8FAFC08FAFC090B0CF8FAFC090B0CF8FAFC090B0CF\
8F202F9000008F00009000008F00009000008F000090B0CF8FAFC090B0CF8FAFC0
90B0CF8FAFC090B0CF8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF8FB0C090B0CF8F6F7090B0CF\
8FB0C090B0CF8FAFC090B0CF405060000000000000000000000000000000000000
0000000000000000005000009000008F00009000008F00009000008F0000902F2F8FB0C090B0CF\
8FB0C090B0CF8FB0C090B0CF90B0C090B0CF8FB0C090B0CF90B0C090B0CF8FB0C0
90B0CF90B0C090B0CF8FB0C08FAFBF80A0BF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F10104F5F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000080A0CF8FA0CF
80A0CF8FA0CF80A0CF8FA0CF50A07F1FA0001FA0001FA000809FC08FA0CF80A0CF8FA0CF80A0CF\
8FA0CF80A0CF8FA0CF80A0CF8FA0CF80A0CF8FA0CF80A0CF8FAFCF1FA0001FA000
1FA00050AF5080A0CF8FA0CF80A0CF405060000000000000000000000000000000000000000000\
0000000000000000000000001F0000000000000000000000000000000000000000
0000000000000000004F5F6F4050604F5F6F80A0C08FAFCF8FA0CF8FAFCF80A0C08FAFCF8FA0CF\
80A0C0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004F5F6F8FAFC08FAFCF
8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC0\
8FAFCF8FAFC08FAFCF8FAFC08FAFCF809FBF000000000000000000000000000000
0000000000000000000000000000000000000000008FAFC08FAFC08FB0CF8FAFC08FAFCF8FAFC0\
8FB0CF8F20208F00008F00008F00008F00008F00008FAFC08FB0CF8FAFC08FB0CF
8FAFC08FB0CF8FAFC08FB0C08FAFC08FB0CF8FAFC08FB0C08FAFC08FB0CF8F404F8F6F7F8FAFC0\
90B0CF8FAFC08FB0C08FAFC04F5F60000000000000000000000000000000000000
0000000000000000000F0F0F0000002000007020209000008F00009000008F00009000008F2020\
904F508FAFC090B0C08FB0C090B0C08FAFC090B0C08FB0C090B0C08FB0C090B0C0
8FB0C090B0C08FB0C090B0C08FB0C08FAFBF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F0F10108FB0C04F5F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000008FAFD08FA0CF
8FA0CF80A0CF8FAFCF8FA0CF8FA0CF1FA00020AF001FA00050AF7F80A0CF8FAFCF8FA0CF8FAFCF\
80A0CF8FAFCF8FA0CF8FAFCF80A0CF8FAFCF8FA0CF8FAFCF80A0CF20AF001FA000
20AF001FA0008FAFCF8FA0CF8FAFCF8FA0CF405060000000000000000000000000000000000000\
0000000000000000000F0F0F8F00008F00008F0000000000000000000000000000
00000040506F4F5F6F8FAFCF8FAFCF8FA0CF8FAFCF8FAFCF8FAFCF8FA0CF8FAFCF8FAFCF8FAFCF\
8FAFCF8FA0C0000F0F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000004F5F6F8FAFCF
8FAFCF8FAFC08FB0CF8FAFCF8FAFCF8FAFC08FB0CF8FAFCF8FAFCF8FAFC090B0CF8FAFCF8FAFCF\
8FAFC090B0CF8FAFCF80A0BF80A0BF000000000000000000000000000000000000
000000000000000000000000000000000000000000000F0F8FAFC08FAFC090B0CF8FAFC08FB0CF\
8FAFC090B0CF8F202F8F00008F00009000008F00008F00008FAFC090B0CF8FAFC0
90B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF8FAFC090B0CF80A0BF8FAFBF8F0000906F7F\
8FAFC090B0CF8FB0C090B0CF4050604F5F60000000000000000000000000000000
0000000000000000000000004F5F6F4F5F6090B0CF8F4F50902F2F8F202F9000008F0000900000\
8F000090707F8F6F70906F7F8F6F7090B0CF8FB0C090B0CF8FB0C090B0CF8FB0C0
90B0CF8FB0C090B0CF90B0C090B0C080A0B0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000001010100F0F0F90BFC090B0C090BFC04F5F5F4F5F60000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000008FA0CF
80A0CF8FA0CF80A0CF8FA0CF6FA07F1FA0001FA0001FA0006FA0A08FA0CF80A0CF8FA0CF80A0CF\
8FA0CF80A0CF8FA0CF80A0CF8FA0CF80A0CF8FA0CF80A0CF6FA07F1FA0001FAF00
1FA0003FA05080A0CF8FAFCF80A0C08FA0CF80A0CF405060000000000000000000000000000000\
0000000000000000000000000000003000008F00008000005000004050604F5F6F
80A0C08FAFCF80A0CF8FAFCF80A0C08FAFCF80A0CF8FAFCF80A0C08FAFCF8FA0C08FAFCF80A0C0\
8FAFCF8FA0C080A0C0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000004F5F6F
8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC0\
8FAFCF809FBF80A0BF000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000008FAFC08FAFC08FB0CF8FAFC0\
8FAFCF8FAFC08FB0CF8F20208F00008F00008F00008F00008F00008FAFC08FB0CF
8FAFC08FB0C08FAFC08FB0CF8FAFC08FB0C08FAFC08FB0CF8FAFC08FB0C08F404F8F00008F6070\
8FB0C08FAFC08FB0CF8FAFC08FB0C040505F000000000000000000000000000000
0000000000004F5F604F5F6F8FAFC08FB0C08FAFC090B0C08FAFC08F8F9F8F6F709000008F0000\
8F202F8F00009000008F00009000008F6F709090A08FAFC08FAFBF8FB0C090B0C0
8FAFC090B0C08FB0C090B0C08FB0C090B0C080A0B0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F0F108FB0C090B0C08FB0C090B0C08FB0C090BFC040505F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8FAFCF80A0CF8FA0CF80A0CF8FAFCF30A02F20A0001FA00020AF008FA0CF8FA0CF80A0CF8FAFCF\
8FA0CF8FAFCF80A0CF8FAFCF8FA0CF8FAFCF80A0CF8FAFCF1FA00020AF001FA000
20AF008FA0CF8FAFCF80A0CF8FAFCF8FA0CF8FAFCF405060000000000000000000000000000000\
00000000000000000000000000000000000040506F8F00008F00008F00008F4050
8FAFCF8FA0CF8FAFCF8FA0CF8FAFCF8FAFCF8FAFCF8FA0CF8FAFCF8FAFCF8FAFCF8FA0CF8FAFCF\
8FAFCF8FAFCF8FA0CF8FA0C0000F0F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4F5F6F8FAFCF8FAFCF8FAFC08FAFCF8FAFCF8FAFCF8FAFC08FB0CF8FAFCF8FAFCF8FAFC080A0BF\
80A0BF000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F80AFC08FB0CF8FAFC090B0CF\
8FAFC08FB0CF8FAFC090B0CF8F202F8F00008F00009000008F00008F00008FAFC0
90B0CF8FAFC08FB0CF8FAFC090B0CF8FAFC08FB0CF8FAFC090B0CF8F40509000008F0000900000\
8F6F7F90B0CF8FAFC090B0CF8FB0C04F5F60000000000000000000000000000000
4F606F4F5F6090B0CF8FB0C090B0CF8FB0C090B0CF8FB0C090B0CF8FB0C0906F7F8F6F70900000\
8F0000906F7F8F202F9000008F00009000008F0000902F2F8F4F5090B0C08FB0C0
90B0CF8FB0C090B0C08FB0C090B0CF8FB0C08FAFBF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F10100F0F0F90BFC0\
90B0C090BFC090B0C090BFC090B0C090BFC090B0C04F5F60000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000008FA0CF80A0CF8FA0CF80A0CF8FA0CF30A02F1FA0001FA0001FA0006FA07F8FA0CF80A0CF\
8FA0CF80A0CF8FA0CF80A0CF8FA0CF6FA09F6FA07F1FA0001FA0001FA0001FA000
6FA07F8FA0CF80A0C08FA0CF80A0CF8FAFCF80A0C08FA0CF1F802F1F80001FA000000000000000\
0000000000000000000000004F5F6F4050608FAFCF8040508F00008000008F0000
80A0C08F607F8060708F00008000008F00008000008F00008000008F000080809F8FAFCF80A0C0\
8FAFCF80A0C08FAFCF80A0C080A0C0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F5F6F8FAFC08FAFCF8FAFC08FAFCF8FAFC08FAFCF8FAFC080A0BF7F9FB0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F80A0BF8FAFCF8FAFC0\
8FB0CF8FAFC08FAFC08FAFC08F202F8F00008F00008F00008F00008F00008F0000
8FAFC08FB0CF8FAFC08FB0C08FAFC08FB0CF8FAFC08FB0C08FAFC08F8FA08F00008F00008F0000\
80A0BF8FAFC08FB0C08FAFC08FB0C08FAFC04050600000000000000000004F5F60
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC090B0C08FAFC08F6F7F8F0000\
9000008FAFC08F8F9F8F8F9F906F7F8F00008F00008F00009000008F6F708F0000
80A0B090B0C08FAFC090B0C08FB0C090B0C08FAFC080AFBF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F0F10108FB0C090B0C08FB0C0\
90B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C04F5F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000008FA0CF80A0CF8FA0CF80A0CF8FAFCF50A07F20A0001FA00020AF001FA00050A07F\
50A05080A0C050A05050A07F1FA00020AF001FA00020AF001FA00020AF006FA07F
8FAFCF80A0CF8FAFCF8FA0CF8FAFCF80A0CF3FAF2F3FA02F20AF001FA00020AF001FA000000000\
0000004F5F6F40506F8FAFCF80A0CF8FAFCF8FA0CF8FAFCF7F9FBF8F00008F0000
8F00008F00008F6F7F8F202F8F6F7F8F00008F00008F00008F00008F00008F2F2F8F607F8FAFCF\
8FA0CF8FAFCF8FAFCF8FAFCF80A0C00F0F0F000F0F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004F5F6F8FAFCF8FAFCF8FAFC080A0BF80A0BF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F80A0BF8FB0CF\
8FAFC090B0CF8FAFC08FB0CF8FAFC0902F2F8F00008F00008F00009000008F0000
8FB0CF8FAFC090B0CF8FAFC08FB0CF8FAFC090B0CF8FAFC08F00008F00009000008F00008F8FA0\
8FAFC090B0CF8FAFC08FB0CF8FAFC090B0CF4050600000004F5F604F606F8FB0C0
90B0CF8FAFC090B0CF8FB0C090B0CF8FAFC090B0CF8FB0C090B0CF8FB0C090B0CF8F6F70900000\
8F00009000008F8F9F90B0C08FAFBF9000008F00009000008F00009000008F0000
904F508F6F7090B0CF8FB0C090B0C08FB0C090B0CF80A0B0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F90B0C08FB0C090BFC090B0C090BFC0\
8FB0C090BFC090B0C090BFC090B0C090BFC090B0C090BFC0405F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008FA0CF80A0CF8FA0CF80A0CF8FA0CF7F9FC01FA0001FA0001FA0001FA000\
1FA0001FA0001FA0001FA0001FA0001FA0001FA0001FA0001FA00030A02F8FA0CF
80A0C08FA0CF80A0CF3FA02F30A0201FA0001FA0001FA0001FA0001FA0001FA0003F802F308020\
8FA0CF80A0C08FAFCF80A0C08FA0CF80A0C08FAFCF80A0C0809FBF8000008F0000
8000008F0000800000809FBF80A0C08FAFCF809FC08F00008000008F00008000008F0000800000\
809FBF80A0C08FAFCF80A0C08FAFCF80A0BF0F0F0F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F5F6F7F9FB080A0BF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F80A0BF\
8FAFC08FAFC08FAFCF8FAFC08FAFC08FAFC08F202F8F00008F00008F00008F0000
8F00008FAFC08FAFC08FB0CF8FAFC08FAFC08F00008F00008F00008F00008F8F9F8F8FA08FAFC0\
8FB0C08FAFC08FB0C08FAFC08FB0C08FAFC0405F6040505F8FB0C08FAFC08FB0C0
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08F2020\
8F00008F00008F6F7F8FAFC08FB0C08FAFC08FAFBF8F00008F00008F0000900000
8F00008F00008F2020906F7F80A0B08FB0C08FB0C080AFBF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F100F0F0F90B0C08FB0C090B0C08FB0C090B0C08FB0BF\
90B0C08FB0C090B0C08FB0BF90B0C08FB0C090B0C08FB0BF4F5F6040505F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008FA0CF80A0CF8FA0CF80A0CF8FA0CF80A0CF80A0C01FA00020AF00\
1FA00020A0001FA00020AF001FA0003FA02F1FA0008FAFCF80A0CF8FAFCF80A0CF
3FAF2F30A02F20AF001FA00020AF001FA00020AF001FA00020AF003F802F3F802F4050608FAFCF\
8FA0CF8FAFCF80A0CF8FAFCF8FA0CF8FAFCF80A0CF8F6F7F8F00008F000080A0CF
8F8FA08F00008F00008000008F202F7F9FBF8FAFCF80A0CF8F6F7F8F00008F00008F00008F0000\
8F00008F6F7F8FA0C08FAFCF8FAFCF8FAFCF80A0C00F0F0F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F0F10104050604F5F60000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
80A0BF8FB0CF8FAFC08FB0CF8FAFC08FB0CF8FAFC08F2F2F8F00008F00008F0000
9000008F00008FB0CF8FAFC090B0CF8F00008F00008F00009090A08F8F9F8FB0CF8FAFC090B0CF\
8FAFC08FB0CF8FAFC090B0CF4050604F5F600000004F5F6F8FAFC08FB0CF8FAFC0
90B0CF8FB0C08FB0C08FAFC090B0CF8FB0C08FB0C08FAFC090B0CF8FB0C090B0C08FAFC0902F2F\
8F00009000008F0000906F7F8FB0C090B0C08FB0C0906F7F8F00009000008F0000
9000008F00009000008F0000902F2F8FB0C090B0C08FB0C08FAFBF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0F10108FB0C090B0C08FB0C090BFC08FB0C090B0C08FB0C090BFC0\
90B0C090B0C08FB0C090BFC090B0C090BFC08FB0C090BFC090B0C04F5F60000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008FA0CF80A0C080A0CF80A0CF8FA0CF80A0C08FA0CF80A0CF\
8FA0CF7F9FC08FA0CF7F90BF8FA0CF80A0C08FA0CF80A0CF8FA0CF30A0203FA02F
1FA0001FA0001FA0001FA0001FA0001FA0001FA0003F802F000000000000000000000000405060\
8FA0CF80A0C08FA0CF80A0C08FAFCF80A0C080A0C08000008F000080000080A0C0
80A0C08FAFCF80404F8F00008000008F00008060708FAFCF80A0C08FAFCF8020208F0000800000\
8F00008000008F202F80A0C08FAFCF80A0C08FAFCF80A0BF0F0F0F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F0F0F10
8F60708FAFCF80A0C08FAFCF40505F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F80A0BF8FAFC08FAFC08FAFCF8FAFC08FAFC08FAFC08F202F8F00008F0000
8F00008F00008F00008F00008F00008F00008F00008F8F9F8F8F9F8FB0C08FAFC08FB0C08FAFC0\
8FB0C08FAFC08FB0C040505F405F600000000000000000004F5F6F8FAFC08FB0C0
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC0\
8F00008F00008F00008F6F708FB0C08FAFC08FB0C08FAFC08FB0C08F6F708F202F
8F00008F00008F00008F4F508FB0C090B0C08FAFC08FB0C080A0B0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F10100F0F0F90B0C08FB0C090B0C08FB0BF90B0C08FB0C090B0C08FB0BF90B0C08FB0C0\
90B0C08FB0BF90B0C08FB0C090B0C08FB0BF90B0C08FB0C090B0C040505F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000008FA0CF80A0CF8FA0CF80A0CF8FA0CF80A0CF8FA0CF\
80A0CF8FAFCF80A0CF8FA0CF80A0CF8FAFCF30A02F3FA02F1FA00020AF001FA000
20A0001FA00020AF001F802F20AF0000400020AF001FA00020AF001FA000000000000000000000\
4050608FAFCF80A0CF8FAFCF80A0CF8FAFCF8FA0CF8F00008000008F00008F202F
8FAFCF80A0C08FAFCF8F80A08F00008000008F00008F00008F6F7F80A0C08FAFCF8F80A08F202F\
8000008F00008F00008F000080607F8FAFCF8FAFCF8FAFCF80A0C00F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F0F10108F60708F6F7F8F607F
8F00008F202F8FAFCF8FAFC08FAFCF405060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000F0F0F0F80A0BF8FB0CF8FAFC08FB0CF8FAFC08FB0CF8FAFC08F2F2F8F0000
8F00008F00008F00008F00008F00008F8F9F8F90A08FAFC08FB0CF8FAFC08FB0CF8FAFC08FB0CF\
8FAFC04F5F604050600000000000000000000000000000000000004F5F6F8FAFC0
8FB0C08FAFC090B0CF8FB0C08FB0C08FAFC090B0CF8FB0C08FB0C08FAFC090B0CF8FB0C080A0BF\
8F00009000008F00008F6F7F8F8F9F90B0CF8FB0C090B0C08FB0C090B0C08F6F70
9000008F6F70906F7F8FB0C090B0C08FB0C090B0C08FB0C090B0C080A0B0000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
0F10108FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090BFC08FB0C090B0C0\
8FB0C090BFC08FB0C090B0C08FB0C090BFC08FB0C090B0C08FB0C090BFC0405F5F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008FA0CF80A0C080A0CF80A0CF8FA0CF80A0C0\
80A0CF80A0C08FA0CF80A0C080A0CF6FA09F1FA0001FA0001FA0001FA0001FA000
1F80203F802F30802F0000000000000000000000000F0F0F1FA0001FA0001FA000002000000000\
0000004050608FA0CF80A0C08FA0CF80A0C08FAFCF80404F8F00008000008F0000
80202080A0C080A0C08FAFCF80A0C08F607F8000008F00008000008F4050809FBF8FAFCF80A0C0\
8F40508000008F00008000008F00008000008F607F809FBF8FAFCF80A0BF0F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F100F0F0F8F6F7F8060708F00008000008F0000
8000008F00008000008FAFCF80A0C08FAFCF40505F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000008FAFC08FAFC08FAFC08FAFC08FAFC08FAFC08FAFC08F20208F0000
8F00008F00008F00008F8FA08F8F9F8FAFC08FAFC08FB0C08FAFC08FAFC08FAFC08FB0C08FAFC0\
4050600000000000000000000000000000000000000000000000000000004F5F6F
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08F6070\
8F00008F00008F00008F00008FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C0
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC080A0B0000000000000000000\
0000000000000000000000000000000000000000000000000F0F100F0F0F90B0C0
8FB0BF8FB0C08FB0C090B0C08FB0BF8FB0C08FB0C090B0C08FB0BF90B0C08FB0C090B0C08FB0BF\
90B0C08FB0C090B0C08FB0BF90B0C08FB0C090B0C08FB0BF90B0C08FB0C04F5F60
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008FA0CF80A0CF8FA0CF80A0CF8FA0CF\
80A0CF8FA0CF80A0CF8FA0CF80A0CF8FA0CF6FA0A020AF001FA00020A0001FA000
0020000000000000000000000000000000000000000000000040001FA00020AF001FA000000000\
0000000000004050608FAFCF80A0CF8FAFCF80A0C08FAFCF80202F8F0000800000
8F00008000008F6F7F80A0C08FAFCF80A0CF8F8FA08000008F00008F00008F00008040508FAFCF\
8FA0CF8FAFCF7F9FBF8F00008F00008F00008000008F00008F607F8F6F7F80A0BF
0F0F0F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F10100F0F0F8F6F7F8F607F8F00008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008FAFCF8FAFC08FAFCF405060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000F8FAFC08FAFC08FB0CF8FAFC08FB0CF8FAFC08F00008F0000
8F8FA08F8F9F8FB0CF8FAFC08FB0CF8FAFC08FB0CF8FAFC08FB0CF8FAFC0405F6040505F000000\
000000000000000000000000000000000000000000000000000000000000000000
4F5F6F8FAFC08FB0C08FAFC090B0CF8FAFC08FB0C08FAFC090B0CF8FB0C08FB0C08FAFC090B0CF\
8F202F8F00008F00009000008FB0C08FB0C08FAFC090B0C08FB0C08FB0C08FB0C0
90B0C08FB0C08FB0C08FB0C090B0C08FB0C090B0C08FB0C090B0C080A0B080AFBF000000000000\
0000000000000000000000000000000000000F0F0F0F10108FB0C090B0C08FB0C0
90B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C0\
8FB0C090B0C08FB0C090BFC08FB0C090B0C08FB0C090BFC08FB0C090B0C040505F
4F5F60000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000080A0C08FA0CF80A0C080A0CF80A0C0\
8FA0CF80A0C080A0CF80A0C08FA0CF80A0C04050603060500020001FA0001FA000
0020000000000000000000000000000000000000000000000020001FA0001FA0001FA000002000\
0000000000000000004050608FA0CF80A0C08FA0CF80A0C08FA0CF8060708F0000
8000008F00008000008F202F7F9FBF8FAFCF80A0C08FA0CF8060708F00008000008F0000800000\
8F607F80A0C08FAFCF80A0C08FAFCF7F9FB08FAFCF809FBF8FAFCF80A0C08FAFCF
809FBF0F0F0F0000000000000000000000000000000000000000000000000000000F0F100F0F0F\
8F6F7F8060708F00008000008F00008000008F00008000008F00008000008F0000
8000008F00008000008F00008000008FAFCF80A0C08FAFC040505F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000080AFC080AFC08FAFC08FAFC08F8FA0808F9F8FAFC0
8FAFC08FAFC08FAFC08FAFC08FAFC08FAFC08FAFC08FAFC040505F405060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
405F604F5F6F8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC0\
8F6F7F8F00008F00008F00008FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C0
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C080A0B0000000000000\
0000000000000000000000000000000F0F0F8FB0C08FB0C08FB0C08FAFBF8FB0C0
8FB0C08FB0C08FB0BF8FB0C08FB0C090B0C08FB0BF8FB0C08FB0C090B0C08FB0BF8FB0C08FB0C0\
90B0C08FB0BF8FB0C08FB0BF90B0C08FB0BF90B0C08FB0BF90B0C08FB0BF90B0C0
40505F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000080A0CF8FA0CF80A0CF8FA0CF\
80A0CF8FA0CF80A0CF405060404F600000000000000000001F800020A0001F8000
0000000000000000000000000000000000000000001F800020A0001FA00020AF001FA000004000\
0000000000000000000000004050608FAFCF80A0CF8FAFCF80A0C08FAFCF80A0CF
8F00008000008F00008000008F000080607F8FAFCF80A0CF8FAFCF80809F8F00008000008F0000\
8000008F000080A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF8FA0C0
8FAFCF80A0BF0F0F0F0000000000000000000000000000000000000F0F0F0F10108F607F8F6F7F\
8000008F00008F00008F00008000008F00008F00008F000080202F8F2F2F8FAFC0
8FAFCF8FAFC08FAFCF8FAFC08F202F8F00008F8FA08FAFC08FAFCF405060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000F8FAFC08FAFC08FB0CF8FAFC08FB0CF8FAFC0
8FB0CF8FAFC08FB0CF8FAFC08FB0C040505F4F5F60000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F5F608FB0CF8FAFC08FB0C08FAFC08FB0CF80AFBF8FB0C08FAFC08FB0C08FAFC08F2F2F\
8F202F8F00008F00008F00008F404F90B0C08FB0C08FB0C08FAFC090B0C08FB0C0
8FB0C08FAFC090B0C08FB0C08FB0C08FB0C090B0C08FB0C08FB0C08FB0C080AFBF000000000000\
0000000000000000000F10100F0F0F90B0C08FB0C090B0C08FB0C090B0C08FB0C0
90B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C0\
8FB0C090B0C08FB0C090B0C08FB0C090BFC08FB0C090B0C08FB0C090BFC08FB0C0
4F5F60000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000080A0C080A0CF809FC0\
80A0CF3F4F604050600000000000000000000000000000000020001FA0001FA000
1FA0000020000020000000001F60000040001FA0001FA0001FA0001FA0001F8000002000000000\
00000000000000000000000040506F80A0C08FA0CF80A0C080A0CF80A0C08FA0CF
80404F8F00008000008F00008000008F00008060708FAFCF80A0C08FA0CF80A0C08F607F800000\
8F00008000008F000080404F8FA0CF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF
80A0C08FAFCF809FBF0F0F0F0000000000000F0F0F0F0F108060708F607F8000008F0000800000\
8F00008000008F00008000008F000080202F8F202F80A0C08FAFC080A0C08FAFCF
80A0C08FAFC080A0C08FAFCF80A0C08F607F80000080A0C080A0C08FAFC040505F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000080AFC080AFC08FAFC08FAFC08FAFC0
80AFC08FAFC040505F405060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000405F608FB0C08FAFC08FB0C08FAFC08F202F8F404F8F202F80A0B08F202F8F2020\
8F00008F00008F00008F20208F6F708FAFC08FB0C08FAFC08FB0C08FAFC08FB0C0
8FAFC08FB0C08FAFC08FB0C08FAFBF8FB0C08FAFC08FB0C08FAFBF8FB0C08FAFC080A0B0000000\
0000000F0F0F0F0F108FAFBF8FB0C08FB0C08FB0C08FAFBF8FB0C08FB0C08FB0C0
8FAFBF8FB0C08FB0C08FB0C08FAFBF8FB0C08FB0C08FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF\
8FB0C08FB0BF90B0C08FB0BF8FB0C08FB0BF90B0C08FB0BF8FB0C08FB0BF90B0C0
8FB0BF405F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000404F60405060\
0000000000000000000000000000000000000000000000000000000040001FA000
20A0001FA00020A0001FA00020A0001FA00020A0001FA00020AF001FA000002000000000000000\
0000000000000000000000000000004F5F6F80A0C08FAFCF80A0CF8FAFCF80A0C0
8FAFCF80809F8F00008000008F000080000080A0C080A0C08FAFCF80A0C08FAFCF80809F8F0000\
8000008F00008000008F202F80809F8FAFCF80A0C08FAFCF80A0C08FAFCF80A0C0
8FAFCF80A0C08FAFCF7F90B010101F80A0C08FAFCF80A0C08FAFCF8FAFC08F4F5080202F8F0000\
8F00008F00008000008F2F2F8F202F8FAFCF80A0C08FAFCF8FAFC08FAFCF80AFC0
8FAFCF8FAFC08FAFCF80AFC08FAFCF8FAFC08FAFCF80AFC08FAFCF8FAFC08FAFCF40505F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000F8FAFC08FAFC08FAFC08FAFC0
4F5F60405060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F5F608FB0C08FAFC08FB0C08FAFC08F6F7F8F202F8F00008F00008F0000\
8F00008F00008F00008FB0C08FB0C08FB0C08FAFC08FB0C08FB0C08FB0C08FAFC0
8FB0C08FB0C08FB0C08FAFC090B0C08FB0C08FB0C08FB0C090B0C08FB0C08FB0C080A0B00F1010\
0F0F0F8FB0C08FB0C090B0C08FB0C08FB0C08FB0C090B0C08FB0C08FB0C08FB0C0
90B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C0\
8FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C0
90BFC0405F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000002000
1FA0001FA0001FA0001FA0001FA0001FA0001FA0001F80001F8000000000000000000000000000\
00000000000000000000000000000000000040506F80A0C08FA0CF80A0C080A0CF
80A0C08FA0CF80A0C08000008000008F000080607080A0CF80A0C08FA0CF80A0C08FA0CF80A0C0\
8F202F8000008F00008000008F00008020208FA0CF80A0C08FAFCF80A0C08FAFCF
7F9FB0809FBF7F9FB00F0F0F809FBF8FAFCF80A0C08FAFC080A0C08FAFCF80A0C080A0C080A0BF\
8F8FA08000008F00008000008F00008020208F405080A0C08FAFCF80A0C08FAFC0
80A0C08FAFCF80A0C08FAFC080A0C08FAFCF80A0C08FAFC080A0C08FAFC080A0C08FAFC040505F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000040506040505F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000405F608FB0C08FAFC08FB0C08FAFC08FB0C08FAFC080AFBF8FAFC0\
8F8F9F8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFBF8FB0C0
8FAFC08FB0C08FAFBF8FB0C08FAFC08FB0C08FAFBF8FB0C08FAFC08FB0C080A0B080A0B00F0F0F\
8FB0C08FAFBF8FB0C08FAFC08FB0C08FAFBF8FB0C08FAFC08FB0C08FAFBF8FB0C0
8FB0BF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF\
8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C0
8FB0BF4F5F6040505F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4F5F604F6060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000040000F0F0F0040000000000000000000000000000000001F8000004000\
1FA0000040000040000000000000000000000000004F5F6F80A0C08FAFCF80A0C0
8FAFCF80A0C08FAFCF80202F8F000080809F8FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF\
80A0C08F8FA08000008F00008000008F00008000008F6F7F80A0C0809FBF7F9FBF
0000000000000000000000000F0F0F80A0BF8FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF\
80A0C08FAFCF8040508F202F8000008F00008000008F6F7F80809F8FAFCF8FAFC0
8FAFCF80A0C08FAFCF8FAFC08FAFCF80AFC08FAFCF8FAFC08FAFCF80AFC08FAFCF8FAFC08FAFCF\
40505F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F5F6F4F5F608FAFC080A0BF0F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000405F608FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C0\
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FB0C0
8FB0C08FAFC08FB0C08FB0C08FB0C08FAFC08FB0C08FB0C08FB0C080A0B080AFBF0000000F1010\
8FB0C08FB0C08FB0C08FB0C08FB0C090B0C08FB0C08FB0C08FB0C090B0C08FB0C0
8FB0C08FB0C090B0C08FB0C08FB0C08FB0C090B0C08FB0C08FB0C08FB0C090B0C08FB0C090B0C0\
8FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C090B0C08FB0C0
90B0C08FB0BF4F5F60000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000004F606F4F5F60
90BFC08FB0BF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000001FA0001FA0001FA0001FA0001FA000\
1FA0001FA0001FA0001FA0001F600000000000000000000040506F80A0C08FA0CF
80A0C080A0CF80A0C08FA0CF80607080A0CF80A0C08FA0CF80A0C080A0CF80A0C08FA0CF80A0C0\
80A0CF80A0C08FAFCF8020208000008000008F4050802020000000000000000000
0000000000000000000000000000000F0F0F809FBF8FAFCF80A0C08FAFC080A0C08FAFCF80A0C0\
8FAFC080A0C08FAFCF80A0C080A0C080809F8F00008000008F00008000008F202F
80404F8FAFC080A0C08FAFC080A0C08FAFC080A0C08FAFC080A0C08FAFC080A0C08FAFC080A0C0\
405060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F5F6F4050608FAFC080AFC08FAFC080AFC08FAFC080A0BF0F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000405F608FB0C080AFC08FAFC08FAFC08FB0C080AFC0\
8FB0C08FAFC08FB0C08FAFBF8FB0C08FAFC08FB0C08FAFBF8FB0C08FAFC08FB0C0
8FAFBF8FB0C08FAFC08FB0C08FAFBF8FB0C08FAFC080A0B07FA0B0000000000000000000000000\
0F0F108FAFC08FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C0
8FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF\
8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C0
8FB0BF8FB0C040505F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000405F604F5F608FB0BF90B0C0
8F00008F0000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000020001FA00020A0001FA00020A0001FA00020A000\
00400020A0001F800020AF001FA00020A0000020000000000000004F5F6F80A0C0
8FAFCF80A0C08FA0CF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF\
80A0C08FAFCF80A0C0809FBF7F9FBF000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F80A0BF8FAFCF80A0C08FAFCF80A0C08FAFCF\
80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80AFC08F4F50800000
8F00008000008F6F7F80A0C08FAFCF80AFC08FAFCF80A0C08FAFCF8FAFC080A0C080AFC08FAFCF\
8FAFC04050600000000000000000002000001F00008F00008F0000700000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F5F604F5F6F8FAFC08FAFCF8FAFC08FAFC08FAFC08FAFCF8FAFC08FAFC080A0BF
0F0F0F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000405F608FB0C08FAFC08FB0C08FAFC08FB0C0\
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC0
8FB0C08FAFC08FB0C08FAFC08FB0C080A0B080A0BF000000000000000000000000000000000000\
0000000F10108FB0C08FB0C08FB0C08FB0C08FB0C08FB0C08FB0C08FB0C08FB0C0
8FB0C08FB0C08FB0C08FB0C090B0C08FB0C08FB0C08FB0C090B0C08FB0C08FB0C08FB0C090B0C0\
8FB0C08FB0C08FB0BF90B0C08FB0C08FB0C08FB0BF90B0C08FB0C090B0C08FB0BF
90B0C08FB0C090B0C040505F000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F60604F5F6090BFC08FB0BF9000008F0000
90BFC08F6F70900000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000040001FA0001FA0001FA0001FA0001FA0001FA000\
0000000000000000000040001FA0001FA0001FA00000200000000000000040506F
80A0C080A0CF80A0C080A0C080A0C080A0CF80A0C080A0C080A0C08FA0CF80A0C080A0C080A0C0\
8FA0CF7F90B07F9FBF000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F7F9FBF8FAFC080A0C08FAFC080A0C0\
8FAFC080A0C08FAFC080A0C08FAFC080A0C08FAFC080A0C08FAFC080A0C08F0000
8000008F000080404F8FAFC080A0C08FAFC080A0C08FAFC080A0C08F6F7F8000008F8F9F80A0C0\
8FAFC080A0C04050600000003F00008000008F00008000008F00008000008F0000
3000000000000000000000000000000000000000000000000000000000006F20206F202F800000\
8F202F8000008F6F708060708FAFC080AFC08FAFC080AFC08FAFC080AFC080A0BF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000405F608FB0C080AFBF8FAFC08FAFC0\
8FB0C080AFBF8FAFC08FAFC08FB0C080AFBF8FB0C08FAFC08FB0C080AFBF8FB0C0
8FAFC08FB0C07FA0B080A0B0000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F8FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C0
8FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF\
8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FB0BF8FB0C0
8FB0BF8FB0C08FB0BF405F5F000000000000000000000000000000000000000000000000000000\
0000000000000000004F5F604F60608FB0BF8FB0C08FB0BF90B0C08FB0BF8FB0C0
8FB0BF8FB0BF8F0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000020A0001FA0001FA0001F60000000001F800020A000\
1FA00000000000000000000000000F20A0001FA000208000000000000000000000
4F5F6F80A0C08FAFCF80A0C08FA0CF80A0C08FAFCF80A0C08FA0CF80A0C08FAFCF7F9FBF7F9FBF\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F809FBF8FAFCF80A0C08FAFCF\
80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF804050
8F00008000008F6F7F80AFC08FAFCF80A0C08FAFCF80AFC08FAFCF80A0C08F4F508000008F8FA0\
80AFC08FAFCF80AFC04050600000000000001F00008F00008000008F00008F0000
8F00008000002000000000000000000000000000008F00008F00008000008F00008F40508F0000\
8000008F00008F00008F00008000008F2F2F8FAFC08FAFC080AFC08FAFC08FAFC0
80AFBF00000F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000405F608FB0C08FAFC08FB0C0\
8FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC08FB0C08FAFC0
80A0BF80A0B0000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F10108FAFC08FB0C08FB0C08FB0C08FB0C08FB0C08FB0C0
8FB0C08FB0C08FB0C08FB0C08FB0C08FB0C08FB0C08FB0C08FB0C08FB0BF8FB0C08FB0C08FB0C0\
8FB0BF90B0C08FB0C08FB0C08FB0BF90B0C08FB0C08FB0C08FB0BF90B0C08FB0C0
8FB0C08FB0BF90B0C08FB0C04F5F60000000000000000000000000000000000000000000000000\
0000000000004F5F6090B0C08FB0BF90BFC08FB0BF90B0C08FB0BF90BFC08FB0BF
90B0C08F909F9070708F0000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001F80001FA0001F9F000000000000000000001F8000\
1FA0001FA0000000000000000000001FA0001FA0001FA000004000000000000000
00000040506F80A0C080A0CF80A0C080A0C080A0C080A0CF7F90B07F9FBF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F7F9FBF8FAFC080A0C0\
80A0C080A0C08FAFC080A0C080AFC080A0C08FAFC080A0C080AFC080A0C08F607F
8000008F000080404F8FAFC080A0C08FAFC080A0C08FAFC080A0C08FAFC080809F8F0000800000\
8FAFC080A0C08FAFC080A0C04050600000000000000000000F0F0F5000008F0000
8000008F00008000005000000000004F5F6050404F8F00008000008FAFC080AFC08FAFC080A0C0\
8FAFC080404F8F00008000008F00008000008F000080404F8FAFC080AFC08FAFC0
80AFC080A0BF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F5F6080AFC08FB0C080AFBF\
8FAFC08FAFC08FB0C080AFBF8FAFC08FAFC08FB0C080AFBF80A0B07FA0B0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F8FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C0
8FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF\
8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF8FB0C0
8FAFBF8FB0C08FB0BF8FB0C040505F000000000000000000000000000000000000000000000000\
4F5F604F5F608FB0C08F00008F909F8FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C0
8FB0BF8FB0C08F404F8F0000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000020001FA0001FA00020A000000000000000000000000000\
1F80001FA0001FA0000000000020001FA0001FA00020A000000F0F000000000000
0000000000004F5F6F80A0C08FAFCF80A0C07F9FBF7F9FBF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F809FBF8FAFCF\
80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80A0C08FAFCF80202F
8F00008000008F000080A0C08FAFCF80AFC08FAFC080A0C08FAFCF80A0BF8F6F7F8000008F0000\
8000008FAFC080A0C08FAFCF80AFC0405060000000000000000000000000300000
8F00008000008F00008000008F00008060708F00008000008FAFC080AFC08FAFC080AFC08FAFC0\
8FAFC08FAFC080AFC08F6F7F8F00008F00008000008F00008F00008FAFC080AFC0
8FAFC08FAFC080AFBF00000F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F5F6F8FAFC08FB0C0\
8FAFC08FB0C08FAFC08FB0C08FAFC080A0B080A0B0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F10108FAFC08FB0C08FB0C08FB0C08FAFBF
8FB0C08FB0C08FB0C08FB0BF8FB0C08FB0C08FB0C08FB0BF8FB0C08FB0C08FB0C08FB0BF8FB0C0\
8FB0C08FB0C08FB0BF8FB0C08FB0C08FB0C08FB0BF8FB0C08FB0C08FB0C08FB0BF
90B0C08FB0BF8FB0C08FB0BF4F5F600000000000000000000000000000000000004F5F604F6060\
8FB0BF8FB0C08F000090B0C08FB0BF90B0C08FB0BF90BFC08FB0BF90B0C08FB0BF
90BFC08FB0BF902F2F8F0000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000003F001FA0001FA0001F8000001F00000000000000\
1F80001F9F001FA0000000000000001FA0001FA0001F8000000000000000000000
00000000000040506040506F7F90B0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F7F9FBF\
8FAFC080A0C080A0C080A0C08FAFC080A0C080AFC080A0C08FAFC080A0C0804050
8000008F000080404F80AFC080A0C08FAFC080A0C080AFC080A0C08F0000800000800000800000\
8F00008000008FAFC080A0C08FAFC080A0C04050600000000000000000004F5F60
4050608F6F7F8000008F00008000008F00008000008F202F80A0C08FAFC080A0C08FAFC080AFC0\
8FAFC080A0C08FAFC080AFC080A0BF8000008F00008000008F00008020208FAFC0
80AFC08FAFC080AFC080A0BF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F5F6080AFBF\
8FB0C080AFBF80A0B07FA0B0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F8FB0C08FAFBF8FB0C080AFBF8FB0C0
8FAFBF8FB0C080AFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF\
8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C0
8FB0BF8FB0C08FAFBF8FB0C08FB0BF405F5F0000000000000000004F5F60405F608FB0C08FB0BF\
8F00008F00008FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C0
8FB0BF8FB0C080AFB08F00008F0000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000040001FA0001FA00020A0001FA0001FA000\
1FA00020A0001FA0001FA0000000000000001F6000000000000000000000000000
0F10100F0F0F8FA0CF405060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
7F9FBF8FAFCF80A0C08FAFC080A0C08FAFCF80A0C08FAFC080A0C08FAFCF806070
8F00008000008F202F80A0BF8FAFC080A0C08F00008000008F00008000008F00008000008F8FA0\
80809F8FAFCF80AFC08FAFC080A0C08FAFCF80AFC04050604050604F5F6F80AFC0
8FAFC080AFC08F8FA080202F8F00008000008F00008000008F202F80AFC08FAFC080AFC08FAFC0\
80AFC08FAFC080AFC08FAFC080AFC08FAFC08000008F00008000008F00008F8F9F
8FAFC080AFC08FAFC08FAFC080AFBF00000F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F5F60\
7FA0B080A0BF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F8FB0C08FAFBF8FB0C08FB0C0
8FB0C08FAFBF8FB0C08FB0C08FB0C08FAFBF8FB0C08FB0C08FB0C08FB0BF8FB0C08FB0C08FB0C0\
8FB0BF8FB0C08FB0C08FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF
8FB0C08FB0BF8FB0C08FB0BF8FB0C040505F0000004F5F604F60608FB0BF90B0C08F00008F0000\
8F202080AFB08FB0BF8FB0C08FB0BF90B0C08FB0BF8FB0C08FB0BF90BFC08FB0BF
8FB0C08FB0BF9070708F0000900000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000040001FA0001FA0001F9F00\
1FA0001FA0001FA0001F9F000020000000000000000000000F0F100F0F0F0F0F10
30A02080A0C080A0C080A0C03F4F5F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F7F9FBF80AFC080A0C080A0C080A0C080AFC080A0C080A0C080A0C08F202F
8000008000008000008F00008000008000008000008F000080000080000080809F8F8F9F80A0C0\
80AFC080A0C08FAFC080A0C080AFC040505F405060000000405F6080A0C08FAFC0
80A0C08FAFC080A0C08FAFC08060708F00008000008F00008000008F000080809F8FAFC080A0C0\
8FAFC080A0C08FAFC080A0C08FAFC080AFC08F6F708000008F00008000008F4050
80A0BF8FAFC080AFC08FAFC080AFBF80A0BF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F0F0F1080AFBF\
40505F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F8FB0C08FAFBF8FB0C0
80AFBF8FB0C08FAFBF8FB0C080AFBF8FB0C08FAFBF8FB0C080AFBF8FB0C08FAFBF8FB0C080AFBF\
8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FB0BF8FB0C0
8FAFBF8FB0C08FB0BF8FB0C08FAFBF405F5F40505F8FB0C08FAFBF8F909F8F00008F909F8F0000\
8F6F708F202080AFB08FB0BF8FB0BF8FB0BF8FB0C08FB0BF8FB0BF8FB0BF8FB0C0
8FB0BF8FB0BF8F20208F2F2F8F00008F0000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000F208000\
0040001F60000000000020000000000000000F0F0F0F101030A02F3FA02F1FA000
20A0001FA00080A0CF80A0C08FA0CF405060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000F0F80A0C080A0C08FAFCF80A0C08FAFC080A0C08FAFCF80A0C08FAFC0800000
8F00008000008F00008000008F00008000008F000080809F8F8FA080A0C08FAFC080A0C08FAFC0\
80A0C08FAFC040505F4050600000000000000000000000000000004F5F6F80A0C0
8FAFC080AFC08FAFC080A0C08FAFC080A0BF8F202F8000008F00008000008F00008060708FAFC0\
80AFC08FAFC080AFC08FAFC080AFC08FAFC080AFC08F00008000008F000080404F
8FAFC080AFC08FAFC080AFC08FAFC080AFC080AFBF00000F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F0F10108FAFC08FB0C080AFC08FB0C0\
8FAFC0405F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F8FB0C08FAFBF
8FB0C08FB0C08FB0C08FAFBF8FB0C08FB0C08FB0C08FAFBF8FB0C08FB0BF8FB0C08FB0BF8FB0C0\
8FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF
8FB0C08FB0BF8FB0C08FB0BF4F5F60000F0F4F60608F4F4F80AFB08F00008FB0C08FB0BF8FB0C0\
8FB0BF8F2F2F8F6F708FB0BF80AFB08FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF
902F2F8F202F8F909F8F20209000008F0000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F0F0F10309F2030A02F1FA0001FA0001F9F001FA000
1FA0001FA0004FA07080A0C080A0C080A0C03F4F5F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000080A0BF80A0C080AFC080A0C080A0C080A0C080AFC080809F800000
80000080000080000080000080809F808F9F80A0C080AFC080A0C08FAFC080A0C080AFC040505F\
405060000000000000000000000000000000000000000000000000000000405F60
80A0C08FAFC080A0C080AFC080A0C08FAFC080A0C0808F9F8000008F0000800000800000800000\
8F6F7080A0BF8FAFC080A0C08FAFC080A0BF8F202F8000008F00008000008F202F
8000008F405080202080A0BF80AFC08FAFC080A0BF80A0BF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F100F0F0F8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF\
8FAFC080AFBF40505F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F8FB0C0
8FAFBF8FB0C080AFBF8FB0C08FAFBF8FB0C080AFBF8FB0C08FAFBF8FB0C080AFBF8FB0C08FAFBF\
8FB0C080AFBF8FB0C08FAFBF8FB0C080AFBF8FB0C08FAFBF8FB0C080AFBF8FB0C0
8FAFBF405F5F40505F3F00003000008F00008F00008F00008F00008F00008F6F708FB0BF8FB0BF\
8FB0C08F6F7080AFBF8F6F708F6F7080AFB08FB0BF8FB0BF8FB0C08F20208F202F
8F8F908F909F8FB0BF8F00008F00008F0000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F0F101030A02F3FA02F1FA00020A0001FA0001FA0001FA00020A0001FA000
50A07F4FA07080A0CF80A0C080A0CF80A0C080A0CF405060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000F0F80A0C080A0C08FAFCF80A0C08FAFC080A0C080A0C0800000
8F000080809F8F8FA080A0C08FAFC080A0C08FAFC080A0C08FAFC080A0C040506040505F000000\
000000000000000000000000000000000000000000000000000000000000405060
8FAFC080A0C08FAFC080AFC08FAFC080A0C08FAFC080AFC08F4F508000008F00008000008F0000\
80404F8FAFC080AFC08FAFC080202F8F00008000008F00008060708F2F2F806070
8F00008000008F00008000008F00008060708FAFC080AFC080A0BF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F10100F0F0F8FAFC080AFC08FB0C08FAFC08FAFC080AFC08FB0C08FAFC08FAFC0\
80AFC08FB0C08FAFC0405F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C08FB0BF8FB0C0\
8FAFBF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF4F5F6040505F
2000003F00005F00003F00000000000000000F0F0F000000702F2F8F6F708F00008F202F8FB0C0\
8FB0BF8FB0C08FB0BF8F6F708F20208F70708FB0BF8F2F2F8F20208F909F8F909F
8FB0C08FB0BF80AFB080A0B08F00008F00008F2F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F100F0F0F\
30A02F309F201FA0001F9F001FA0001FA0001FA0001F9F0050A07F4FA07080A0C0
809FC080A0C080A0C080A0C0809FC080A0C080A0C080A0C03F4F5F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000080A0BF80A0C080A0C080A0C080A0C080A0C080809F
80809F80A0C080A0C080AFC080A0C080A0C080A0C04050603F4F5F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
40506080AFC080A0C08FAFC080A0BF80AFC080A0C08FAFC080A0BF8060708000008F0000800000\
8000008000007FA0B080A0BF80202F80404F8F202F80A0BF80AFC080A0C08FAFC0
7F9FB08F00008000008F00008000008F405080AFBF8FAFC080A0BF80A0BF000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F10
0F0F0F8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF\
8FAFC080AFBF8FAFC080AFBF40505F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F8FB0C080AFBF8FB0C080AFBF8FB0C080AFBF8FB0C080AFBF8FB0C080AFBF8FB0C080AFBF\
8FB0C08FAFBF8FB0C080AFBF8FB0C08FAFBF8FB0C040505F405F5F000000000000
8000008F00000000000000000000000000000000000000000000004F5F608FB0BF80AFBF808F90\
8F909F8FB0BF8FB0C08FAFBF8FB0BF80A0AF8F00008F404F8F909F8FB0BF8FB0C0
8FB0BF8FB0BF8FB0BF8F6F708F00008F00008F404F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F10100F0F0F80A0CF80A0C080A0CF\
6FA09F1FA0001FA00020A0001FA0001FA00030A02F80A0CF80A0C080A0CF80A0C0
80A0CF80A0C080A0CF80A0C080A0CF80A0C080A0C080A0C080A0CF405060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000F0F80A0BF80A0C08FAFC080A0C080AFC080A0C0
8FAFC080A0C08FAFC080A0C040506040505F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004050608FAFC080A0C08FAFC080AFC08FAFC080A0C08FAFC080AFC08FAFC080404F8F0000\
8000008F00008000008F4F5080AFC08FAFC080AFC08FAFC080AFC08FAFC080AFC0
8FAFC080AFC08F4F508000008F00008000008F4F5080AFC08FAFC080A0BF0F0F0F000000000000\
0000000000000000000000000000000000000000000000000F0F0F0F101080AFC0
8FAFC080AFC08FAFC080AFC08FAFC080AFC08FAFC080AFBF8FB0C080AFC08FAFC080AFBF8FB0C0\
8FAFC08FAFC080AFBF8FB0C08FAFC0405F60000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C08FB0BF8FB0C08FAFBF8FB0C0\
8FB0BF8FB0C08FAFBF8FB0C040505F405F600000000000000000000000006F0000
8F00003F00000000000000000000000000000000000000000000000000004F60608FB0BF8FB0C0\
8FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF80AFB08F202F8F909F8FB0BF
8FB0C08FB0BF8FB0C08F6F708F00008F00008F00008F8F90000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F100F0F0F80A0C0809FC080A0C07F9FC080A0C0809FC0\
80A0C0609F9F1FA0004F9F4F1FA000309F207F9FBF809FC080A0C07F9FC080A0C0
80A0C080A0C0809FC080A0C04FA07050A07F1F9F0080A0C080A0C080A0C03F4F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000080A0BF80A0C080A0C080A0C080A0C0
80A0C080A0C03F4F5F405060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000040506080AFC080A0C080AFC080A0BF80AFC080A0C080AFC080A0BF80AFC0806070\
8F00008000008000008000008F000080809F80AFC080A0BF8FAFC080A0BF80AFC0
80A0BF8FAFC080A0BF8060708000008F000080000080AFC080AFBF8FAFC07FA0B00F0F0F000000\
0000000000000000000000000000000F0F0F0F0F0F80AFBF8FAFC080AFBF8FAFC0
80AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF\
8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF40505F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F8FB0C080AFBF8FB0C080AFBF8FB0C080AFBF8FB0C080AFBF8FB0BF80AFBF\
8FB0C040505F40505F0000000000008000005000000000000000000000006F0000
8000007000000000000000000000000000000000000000000000000000000000004F5F608FAFBF\
8FB0C080AFBF8FB0BF8FB0BF8FB0BF80AFBF8FB0BF8FB0BF8FB0BF808F908F202F
7FA0AF8FB0BF80B0BF8FB0BF8F20208F00008F00008FB0BF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0F0F10809FC080A0CF80A0C080A0CF809FC080A0CF80A0C080A0CF\
80A0C080A0CF80A0C0809FC01FA00020A00080A0C080A0C080A0C080A0CF80A0C0
80A0C04FA07050A07F1FA0001FA0001FA00020A0001FA00080A0C080A0C080A0CF404F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000F0F80A0BF80A0C08FAFC040505F
405060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004050608FAFC080A0C08FAFC080AFC08FAFC080A0C08FAFC080AFC08FAFC0\
80A0C08F6F7F8000008F00008000008F000080202F80A0B080AFC08FAFC080AFC0
8FAFC080AFC08FAFC080AFC08F4F508000008F000080A0BF8FAFC080AFC08FAFC080A0BF0F0F0F\
0000000000000F0F0F0F101080AFC08FAFC080AFC08FAFC080AFBF8FAFC080AFC0
8FAFC080AFBF8FAFC080AFC08FAFC080AFBF8FAFC080AFC08FAFC080AFBF8FB0C080AFC08FAFC0\
80AFBF8FB0C080AFC08FAFC080AFBF8FB0C040505F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F8FB0C08FAFBF8FB0C08FAFBF8FB0C08FAFBF8FB0C040505F405F60\
40505F0000000000008F00008F00000F0F0F0000000000000000000000003F0000
8F00008F00000000000000000000000000000000000000000000000000000000000000004F5F60\
8FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF
8F909F8F202080AFB08FB0BF8FB0BF8F00008F0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007F9FC080A0C07F9FC080A0C07F9FC080A0C0809FC080A0C07F9FC0\
80A0C0809FC080A0C01F9F001FA000309F2080A0C07F9FC080A0C0809FC050A07F
4F9F701FA0001F9F001FA0001F9F001FA0001FA0003FA02F309F2080A0C080A0C080A0C03F4F5F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000405060405060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000040506080AFC080A0C080AFC080A0BF80AFC080A0BF80AFC080A0BF\
80AFC080A0BF808F9F8020208000008000008000008000008060707FA0BF8FAFC0
80A0BF80AFC080A0BF8FAFC08000008000008060708FAFC080A0BF80AFC080A0BF8FAFC07090AF\
10101080AFBF8FAFC080A0BF80AFC080AFBF8FAFC080A0BF80AFC080AFBF8FAFC0
80AFBF80AFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF\
8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC03F505F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F8FB0BF80AFBF8FB0C080AFBF40505F40505F000000000000\
8F00008000008F00008000008F00001F00000000000000000000000000001F0000
8000008F00006F0000000000000000000000000000000000000000000000000000000000000000\
4F5F608FAFBF8FB0BF80AFBF8FB0BF8FB0BF8FB0BF80AFBF8FB0BF8FB0BF8FB0BF
80AFBF8FB0BF8FB0BF8F2F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000080A0C080A0CF809FC080A0CF80A0C080A0C0809FC080A0CF\
80A0C080A0C0809FC050A07F1FA0001FA0001FA00020A00030A02F1FA0001FA000
20A0001FA0001FA0001FA0003FA02F30A02F80A0C080A0C080A0CF80A0C080A0C080A0C0405060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004050604F5F6F40506080AFC07F9FBF0F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004050608FAFC080A0C08FAFC080A0C08FAFC080A0C08FAFC0\
80AFC08FAFC080A0C08FAFC080A0BF8F00008000008F00008000008F0000800000
8F6F7F80202F8F6F7F80809F8F000080202F8FAFC080AFC08FAFC080AFC080A0B07F9FB0000000\
000F0F80A0BF80AFBF8FAFC080AFC08FAFC080AFBF8FAFC080AFC08FAFC080AFBF
8FAFC080AFC08FAFC080AFBF8FAFC080AFC08FAFC080AFBF8FAFC080AFC08FAFC080AFBF8FAFC0\
80AFC08FAFC080AFBF8FB0C080AFBF8FAFC080AFBF8FB0C040505F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F10108FAFBF405F6040505F0000000000000000006F0000000000\
1F00002000008000008F00008F00003F0000300000000000000000000000000000
3F00008F00008F00006F0000000000000000000000000000000000000000000000000000000000\
0000004F5F608FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF8FB0C08FB0BF
8FB0C0000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007F9FC080A0C07F9FC080A0C07F9FC080A0C07F9FC0\
80A0C07F9FC080A0C07F9FC050A07F1F9F001FA0001F9F001FA0001F9F001FA000
1F9F001FA000309F2030A02F809FC080A0C07F9FC080A0C0809FC080A0C07F9FC080A0C080A0C0\
405060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000050405040506080A0C080A0BF80A0C080A0C080A0C07F9FBF0F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000040506080AFC080A0BF80AFC080A0BF80AFC080A0BF\
80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC080404F806F7080000080202F
80000080405080202080AFC080A0BF80AFC080A0BF7F9FB07F9FB0000000000000000000000000\
00000000000080A0BF80AFBF8FAFC080A0BF80AFC080AFBF8FAFC080A0BF80AFC0
80AFBF8FAFC080A0BF80AFC080AFBF8FAFC080AFBF80AFC080AFBF8FAFC080AFBF80AFC080AFBF\
8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC03F505F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F5F60405F6080A0B0000000000000000000000000000000000000000000\
0000000000000000006F00008F0000800000500000000000000000000000000000
3000008F00008000006F0000000000000000000000000000000000000000000000000000000000\
0000000000004F5F6080AFBF8FB0BF80AFBF8FB0BF80AFBF8F909F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000809FC080A0C0809FC080A0CF80A0C080A0C0\
809FC080A0CF80A0C080A0C0809FC06FA0A030A02F1FA0001F9F0020A00030A02F
3FA02F80A0C080A0CF80A0C080A0C080A0C080A0C080A0C080A0C080A0C080A0C080A0C080A0C0\
80A0C0405060000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000006F202F\
6F202F8000008000008000008060707F9FBF80A0C080A0C080A0C080A0BF000F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000040506080AFC080A0C08FAFC080A0C080AFC0\
80A0C08FAFC080A0C08FAFC080A0C08FAFC080AFC08FAFC080A0BF8FAFC080AFC0
8FAFC080A0BF8FAFC080AFC07FA0B07F9FB0000000000000000000000000000000000000000000\
000000000000000F0F80A0BF80AFBF8FAFC080AFC08FAFC080AFBF8FAFC080AFC0
8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC0\
80AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FB0C040505F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000405F604F5F608FAFBF8FB0C08000008F4F50000000000000000000000000000000000000\
0000000000000000000000003000008F00008F0000700000000000000000000000
0000003000008F00008F0000700000000000000000000000000000000000000000000000000000\
0000000000004F5F608FB0C08FB0BF8FB0C08FB0BF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000007F9FC080A0C07F9FC080A0C07F9FC0\
80A0C07F9FC080A0C07F9FC080A0C07F9FC080A0C07F9FC080A0C07F9FC080A0C0
7F9FC080A0C07F9FC080A0C0609F9F80A0C04F9F707F9FBF7F9FC080A0C0809FC080A0C07F9FC0\
80A0C080A0C0405060000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000040506040506080202F800000\
80000080607080A0C080A0BF80607080404F80405080A0BF80A0C080A0C080A0BF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000040506080AFC080A0BF80AFC080A0BF\
80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC0
80A0BF80AFC07F9FB07F9FB0000000000000000000000000000000000000000000000000000000\
00000000000000000000000080A0BF80A0BF80AFC080A0BF80AFC080AFBF8FAFC0
80A0BF80AFC080AFBF8FAFC080A0BF80AFC080AFBF8FAFC080A0BF80AFC080AFBF8FAFC080AFBF\
80AFC080AFBF8FAFC080AFBF80AFC080AFBF8FAFC080AFBF80AFC080AFBF8FAFC0
3F505F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000405F604F5F60
80AFBF8FAFBF8000008F000080404F80AFBF80AFBF80A0B0000000000000000000000000000000\
0000000000000000000000000000001F00008F00008000008F00001F0000000000
0000000000003000008F00008000008F0000300000000000000000000000000000000000000000\
000000000000000000405F5F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000809FC080A0C0809FC080A0CF\
809FC080A0C0809FC080A0CF80A0C080A0C0809FC080A0C080A0C080A0C0809FC0
80A0C080A0C03FA02F4FA04F20A0001FA0001FA0001FA00020A0001FA0001FA00060A09F80A0C0\
80A0C080A0C080A0C0405060000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F202F50405080000080607F80202F800000\
80000080000080809F80A0C080A0C080A0C080A0C080AFC080A0C080A0C080A0C0
80A0BF000F0F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004F5F6F80A0C080AFC080A0BF8FAFC0\
80A0C080AFC080A0BF8FAFC080A0C080AFC080A0BF8FAFC080A0C080AFC07F9FB0
7FA0B0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000F0F80A0BF80AFBF8FAFC080AFBF8FAFC080AFBF
8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC0\
80AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF
8FAFC040505F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000405F604F5F6080AFBF8FB0C0800000
8F000080404F8F00008000008F0000808F9F8FB0C080AFBF80A0B0000000000000000000000000\
0000000000000000000000000000000000000000007000008F00008F0000300000
0000000000000000000000005000008F00008F00008000000F0F0F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007F9FC080A0C07F9FC0\
80A0C07F9FC080A0C07F9FC080A0C07F9FC080A0C07F9FC080A0C07F9FC04FA070
4F9F4F1FA0001F9F001FA0001F9F001FA0001F9F001FA0001F9F001FA0001F9F001FA0001F9F00\
50A0507F9FC080A0C07F9FC0405060000000000000000000000000000000000000
0000000000000000000000006020206F202F7F1F2080607080A0C080A0C07FA0BF80809F806070\
8000007F000080000080202080A0C07FA0BF80A0C080A0C080A0C07FA0BF80A0C0
80A0BF80A0BF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000405F6080A0BF80AFC080A0BF\
80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC07F9FB07F9FB0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000080A0BF80A0BF80AFC080A0BF80AFC0
80A0BF80AFC080A0BF80AFC080AFBF80AFC080A0BF80AFC080AFBF80AFC080A0BF80AFC080AFBF\
8FAFC080A0BF80AFC080AFBF8FAFC080AFBF80AFBF80AFBF8FAFC080AFBF80AFBF
80AFBF8FAFC03F505F000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F5F604050608FAFBF80AFBF8F8F9F80AFBF8FAFBF
80AFBF8FAFBF7FA0B08F202F8000008F000080606F8FAFBF80AFBF80A0B0000000000000000000\
0000000000000000000000000000000000000000000000006F00008000008F0000
6F00000000000000000000000000003F00006F00008F00008000006F0000300000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000809FC080A0C0\
7F9FC080A0C0809FC080A0C07F9FC080A0C0809FC080A0C07F9FC03FA02F1FA000
1FA0001F9F0020A0001FA0001FA0004F9F4F3FA02F7F90B03FA05060A07020A0001FA0001FA000\
1FA00020A0007F9FBF80A0C080A0C0405060000000000000000000000000000000
0000004050604F5F6F80607080000080000080202F7F9FBF80A0C080A0C080A0C080A0C080A0C0\
80809F80000080000080000080000080000080809F80A0C080A0C080A0C080A0C0
80A0C080A0C080A0BF000F0F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F5F6F80A0C080AFC0\
80A0BF8FAFC080A0C080AFC080A0BF7FA0B07F9FB0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F80A0BF8FAFC080AFBF8FAFC080AFBF
8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC0\
80AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF
8FAFC080AFBF8FAFC040505F000000000000000000000000000000000000000000000000000000\
0000004F5F60405F608FAFC080AFBF8F000080000080A0B080AFBF8FB0C080AFBF
8FB0C080AFBF8FB0C080AFBF8F6F708000008F000080202F8FB0C080AFBF80A0B0000000000000\
0000000000000000000000000000000000000000000000000000003F0000800000
8F0000800000000000000000000000000000000000000000700000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000007F9FC0\
80A0C07F9FC080A0C07F9FC080A0C07F9FC080A0C07F9FC080A0C01F9F001FA000
1F9F001FA000309F4F50A07F7F9FC080A0C07F9FC080A0C07F9FBF80A0C07F9FC07F9FBF309F20\
1FA0001F9F001FA000609F9F80A0C07F9FC0405060000000000000000000405060
40506080A0C07FA0C08060707F00008000007F000080809F7F9FBF80A0C080A0C080A0C07F9FBF\
80A0C080A0BF8060707F00008000008000008000007F000080000080809F80A0C0
7FA0BF80A0C080A0BF80A0BF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000405F6080A0BF\
80A0C080A0BF7F9FB07F9FB0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F7FA0B080AFC080A0BF80AFC0
80A0BF80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFBF80AFBF80AFC080A0BF\
80AFBF80AFBF80AFC080A0BF80AFBF80AFBF80AFC080AFBF80AFBF80AFBF80AFC0
80AFBF80AFBF80AFBF8FAFC03F505F0000000000000000000000000000000000004F5F60405060\
80AFBF80AFBF8F00008000008000008000008F000080606F80AFBF80AFBF8FAFBF
80AFBF80AFBF80AFBF8FAFBF80AFBF8F6F708000008F000080000080AFB080AFBF80A0B0000000\
0000000000000000000000000000000000000000000000000000000000001F0000
8000008F00008000003F0000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
809FC080A0C07F9FC080A0C0809FC080A0C07F9FC080A0C0309F2F1FA0001F9F00
20A00060A09F80A0C07F9FC080A0C080A0C080A0C07F9FC080A0C080A0C080A0C0809FC080A0C0\
4FA0701FA0001FA00020A0006FA09F80A0C080A0C040506040506080A0C080A0C0
80A0C080A0C080A0C07F90B080000080000080000080000080A0C080A0C080A0C080A0C080A0C0\
80A0C080A0C080A0C080607F80000080000080000080202F80202F800000800000
80607F80809F80A0C080A0C080A0BF000F0F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F5F6F\
7F9FB0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F7FA0BF80AFC080AFBF
8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC0\
80AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF
8FAFC080AFBF8FAFC080AFBF405F600000000000000000004F5F60405F604F5F6080AFBF8FB0C0\
80AFBF8F6F7080AFBF8FB0C0808F9F8F00008000008F000080AFBF8FAFC080AFBF
8FB0C080AFBF8FB0C080AFBF8FB0C080AFBF80AFBF8000008F00008000008F6F7080AFBF80A0B0\
000000000000000000000000000000000000000000000000000000000000000000
0000006F00008F0000800000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
80A0C07F9FC080A0C07F9FC080A0C07F9FC080A0C0309F4F1FA0001F9F001FA000
309F2080A0C07F9FBF80A0C07F9FC080A0C07F9FBF80A0C07F9FC080A0C07F9FBF80A0C07F9FC0\
80A0C04F9F4F1FA0001F9F001FA0003F4F5F0000000000004050607F9FBF80A0C0
7F9FC080A0C07F9FBF80A0C07F20208000007F00008000007F000080A0C07F9FBF80A0C07FA0BF\
80A0C07F9FBF80A0C07FA0BF80A0C07F00008000007F000080202F7FA0BF80202F
8000008000007F00008060707F90B0809FBF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F100F0F0F80A0C080A0BF\
405060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F7F9FB080AFBF
80A0BF80AFC080A0BF80AFBF80A0BF80AFC080A0BF80AFBF80A0BF80AFC080A0BF80AFBF80AFBF\
80AFC080A0BF80AFBF80AFBF80AFC080A0BF80AFBF80AFBF80AFC080A0BF80AFBF
80AFBF80AFC080AFBF80AFBF80AFBF40505F40505F405F6080AFBF80AFBF80AFBF80AFBF80AFBF\
8FAFBF80AFBF80AFBF80AFBF8F6F708000008000008060708FAFBF80AFBF80AFBF
80AFBF8FAFBF80AFBF80AFBF80AFBF8FAFBF80AFBF80AFBF8020208F0000800000804F4F80AFBF\
7FA0B0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000080A0C07F9FC080A0C07F9FC080A0C07F9FC06FA07F1F9F0020A0001F9F00
7F9FBF7F9FC080A0C0809FC080A0C07F9FC080A0C080A0C050A07F609F7080A0C080A0C080A0C0\
7F9FC020A0001FA0001FA0001F7F00000000000000000000000000405F6F80A0C0
80A0C080A0C080A0C080A0C080A0C080000080000080000080000080000080809F80A0C080A0C0\
80A0C080A0C080A0C080A0C080A0C080A0C080000080000080000080405080A0C0
80A0C080202F80000080000080607080607080A0BF000F0F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F0F101080A0BF80AFC080A0BF80AFC0\
80A0BF405060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F7FA0B0
80AFC080A0BF8FAFC080AFBF80AFC080AFBF8FAFC080AFBF80AFC080AFBF8FAFC080AFBF80AFC0\
80AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF8FAFC080AFBF
8FAFC080AFBF8FAFC040505F40505F0000004F5F608000008F000080AFBF8FAFC080AFBF8FAFC0\
80AFBF8FAFC080AFBF8FAFC080AFBF8F2F2F8000008F0000808F908FB0C080AFBF
8FAFC080AFBF8FB0C080AFBF8FAFBF80AFBF8FB0C080AFBF8F6F708000008F000080202F8FB0BF\
7FA0AF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000080A0C07F9FBF809FC07F9FC080A0C07F9FBF1FA0001F9F001FA000
4F9F7080A0C07F9FC080A0C07F9FBF80A0C07F9FC050A0701F9F001FA000309F2080A0C03F4F5F\
404F60003F001FA0001F9F001FA00000200000000000000000000040506080A0C0
7F9FBF80A0C07F9FBF80A0C07F9FBF8000007F00008000007F00008000007F607080A0C07F9FBF\
80A0C07FA0BF80A0C07F9FBF80A0C07FA0BF7F9FB07F00008000007F000080A0C0
7FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF7F9FBF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F0F0F0F80A0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF\
80A0C080A0BF40505F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
7F9FB080AFBF80A0BF80AFC080A0BF80AFBF80A0BF80AFC080A0BF80AFBF80A0BF80AFC080A0BF\
80AFBF80A0BF80AFC080A0BF80AFBF80AFBF80AFBF80A0BF80AFBF80AFBF80AFBF
3F505F40505F0000000000000000008000008000008000007F9FAF80AFBF80AFBF80AFBF80AFBF\
80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80000080000080202F80AFBF80AFBF
80AFBF80AFBF80AFBF8FAFBF80AFBF80AFBF80AFBF8FAFBF80AFBF806F708000008F0000800000\
80AFBF7F9FAF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000080A0C07F9FC080A0C07F9FC080A0C04F9F701FA0001F9F00
1FA0007F9FC080A0C07F9FC080A0C07F9FC080A0C07F9FC020A0001F9F001FA0001F7F00000000\
0000000020001F9F0020A0001FA0001F6000000000000000000000000000405060
80A0C080A0C080A0C07FA0C080A0C080A0C080607F7F00008000008000008000007F202F7F9FBF\
80A0C080A0C080A0C080A0C080A0C080A0C080607080000080000080809F80A0BF
80A0C080A0C080A0C080A0BF80A0C080A0C080A0C080A0BF80A0BF000F0F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F0F101080A0BF80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC0\
80A0BF80AFC080A0BF405060000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F7FA0B080AFC080A0BF8FAFC080AFBF80AFC080AFBF8FAFC080AFBF80AFC080AFBF8FAFC0\
80AFBF80AFC080AFBF8FAFC080AFBF80AFC080AFBF8FAFC080AFBF40505F40505F
0000000000008000008000000000000000000000003000008F000080A0B08FAFC080AFBF8FAFC0\
80AFBF8FAFC080AFBF8FAFC080AFBF8FAFBF8060708F00008000008F4F5080AFBF
8FB0C080AFBF8FAFBF80AFBF8FB0C080AFBF8FAFBF80AFBF8FB0C080AFBF80AFBF8000008F0000\
8000008F6F707FA0AF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000080A0C07F9FBF809FC07F9FC080A0C0309F201F9F00
1F9F001FA0007F90BF80A0C07F9FC080A0C03F4F5F404F5F0000001FA0001F9F001FA0001F7F00\
1FA0001F9F001FA0001F9F001FA0001F9F00000000000000000000000000000000
40506080A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F607080202F7F00008000007F0000\
8000007F9FB080A0C07FA0BF80A0C07F9FBF7F9FBF7F00008000007F9FBF80A0C0
7FA0BF80A0C07F9FBF80A0C07FA0BF80A0C07FA0BF80A0C07F9FB00F0F0F000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
0F0F0F80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80A0C080A0BF\
80A0C07FA0BF80A0C080A0BF40505F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F7F9FB080AFBF80A0BF80AFBF80A0BF80AFBF80A0BF80AFBF80A0BF80AFBF80A0BF\
80AFBF80A0BF80AFBF80A0BF80AFBF80A0BF40505F3F505F000000000000800000
80000080000080000000000000000000000000000000000000000050404F80202080AFBF80AFBF\
80AFBF80AFBF80AFBF80AFBF80AFBF80202080202F800000800000800000808F9F
80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80404F\
8F0000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000080A0C07F9FC080A0C07F9FC080A0C0309F2F
1FA0001F9F001FA000609F704050603F4F5F0000000000000000000000001FA0001F9F001FA000\
1F9F001FA0001F9F001FA0001F9F00002000000000000000000000000000000000
00000040506080A0C080A0C080A0C07FA0C080A0C080A0C080A0C07FA0BF80809F806070800000\
7F000080000080000080202F7F000080202F8060708060707FA0BF80A0C080A0C0
80A0C080A0BF80A0C080A0C080A0C080A0BF80A0C080A0C080A0C07F9FBF0F0F0F000000000000\
0000000000000000000000000000000000000000000F10100F0F0F80AFC080A0BF
80A0C080A0BF80AFC080A0BF80A0C080A0BF80AFC080A0BF80AFC080A0BF80AFC080A0BF80AFC0\
80A0BF80AFC080A0BF80AFC080A0BF405060000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F7FA0B080AFC080A0BF80AFC080AFBF80AFC080A0BF80AFC080AFBF80AFC0\
80AFBF8FAFC080AFBF40505F40505F0000000000001F00008000000000001F0000
6F00008000008F00006F00000000000000000000000000000000000000004F5F607F9FAF80202F\
8060708FAFC080202F8F202F808F908F8F9F80AFBF7FA0B08000008F000080202F
8FAFBF80AFBF8FB0C080AFBF8FAFBF80AFBF8FB0C080AFBF8FAFBF80AFBF8FB0C0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000080A0C07F9FBF7F9FC07F9FBF80A0C0
4F9F4F1F9F001F9F001FA0001F9F001FA0001F7F001F60000000000000000000001FA0001F9F00\
1FA0001F9F00002000001F00002000000000000000000000000000000000000000
00000000000040506080A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF\
80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07FA0BF80A0C0
7F9FBF80A0C07FA0BF80A0C07F9FBF80A0C07FA0BF80A0C07F9FBF80A0C07F9FB00F0F0F000000\
0000000000000000000000000F0F0F0F0F0F80A0C07FA0BF80A0C07FA0BF80A0C0
7FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF\
80A0BF80A0BF80A0C07FA0BF80A0BF3F4F5F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F7F9FB080AFBF80A0BF80AFBF80A0BF80AFBF80A0BF80AFBF80A0BF\
80AFBF3F505F40505F000000000000000000000000000000000000000000000000
0000001F0000800000800000800000000000000000000000000000000000000000405F6080A0BF\
808F9F80202080404F80AFBF80AFBF80AFBF80AFBF80AFBF806F70800000800000
80404F80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF80AFBF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000080A0C07F9FC04050603F4F5F
0000000020001F80001F9F001FA0001F9F001FA0001F9F000040000000000000000000001FA000\
0020000020000000000000000000000000000000000000001F80001F80001F9F00
00000000000000000040506080A0C07FA0C080A0C07F9FBF80A0C080A0C080A0C07FA0BF80A0C0\
80A0C080A0C07FA0BF80A0C080A0C080A0C07FA0BF80A0C080A0C080A0C07FA0BF
80A0C080A0C080A0C07FA0BF80A0C080A0BF80A0C07FA0BF80A0C080A0BF80A0C07F9FB00F0F0F\
0000000000000F0F0F0F101080A0BF80A0C080A0BF80A0C080A0BF80A0C080A0BF
80AFC080A0BF80A0C080A0BF80AFC080A0BF80A0C080A0BF80AFC080A0BF80A0C080A0BF80AFC0\
80A0BF80AFC080A0BF80AFC080A0BF80AFC03F505F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F7FA0B080AFC080A0BF80AFC080AFBF80AFC03F505F405060\
000000000000000000800000500000000000000000000000000000000000000000
0000000000000000006F00008F00008000001F00000000000000000000000000000000004F5F60\
80AFBF80AFBF8080908F2F2F7F9FAF80AFBF80AFBF8FAFC080AFBF80202F800000
8F0000808F9080AFBF80AFBF8FAFBF80AFBF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000404F60000000000000
0000000000000000000020001F5F001FA0001F9F001F9F001F7F000F0F0F000000000000000000\
0000000000000000000000000000001F7F001F80001F9F001FA0001F9F001FA000
1F9F0000000000000000000040506080A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF\
80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C0
7F9FBF80A0C07F9FBF80A0C07FA0BF80A0C07F9FBF80A0C07FA0BF80A0C07F9FBF80A0C07090AF\
1010107FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C0
7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF\
80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF3F4F5F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000080A0BF80A0BF80AFBF3F505F40505F000000000000000000\
0000000000001F0000800000800000800000500000000000000000000000000000
000000000000000000000000600000800000800000300000000000000000000000000000000000\
405F6080A0BF80AFBF80AFBF80AFBF802020806F7080AFBF80AFBF80AFBF800000
80000080000080AFBF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F80001F7F001FA0001F9F001FA0001F9F001FA0001F9F00
1FA0001F9F0000000000000000000040506080A0C07FA0C080A0C07F9FBF80A0C07FA0C080A0C0\
7F9FBF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C080A0BF
80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80A0C080A0BF80A0C07090B07F9FB0000000000000\
00000080A0BF80A0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80A0C080A0BF
80A0C080A0BF80A0C080A0BF80A0C080A0BF80AFC080A0BF80A0C080A0BF80AFC080A0BF80A0C0\
80A0BF80AFC080A0BF80A0C080A0BF80AFC080A0BF80AFC03F505F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000040506040506040505F0000000000006F0000000000000000\
0000000000000000003F0000800000800000800000500000000000000000000000
0000000000000000000000003F0000800000800000300000000000000000000000000000000000\
0000004F5F6080AFBF80AFBF80AFBF8FAFBF80AFBF80202F8060708FAFBF806070
800000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
1F80001F7F001FA0001F9F001FA0001F9F001FA0001F9F00004000003F001F8000
1F9F001FA0001F9F0000000000000000000040506080A0C07F9FBF80A0C07F9FBF80A0C07F9FBF\
80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C0
7F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07090B07F9FB0000000000000000000000000000000\
0000000000007F9FBF7F9FBF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C0
7FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF\
80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF3F4F5F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000405F6040505F80AFBF7F9FB08000007F00001F0000000000000000000000\
0000000000000000000000006F00008000008000001F00003000004F00001F0000
000000000000000000000000000000300000800000800000600000000000000000000000000000\
00000040505F80AFBF80A0BF80AFBF80AFBF80AFBF7FA0B080202F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000001F80001F7F001FA000\
1F9F001FA0001F9F001FA0001F9F001FA000004000004000000000000000000000
1F80001F9F001FA0001F9F0000000000000000000040506080A0C07FA0BF80A0C07F9FBF80A0C0\
7FA0BF80A0C07F9FBF80A0C07FA0BF80A0C07F9FBF80A0C07FA0BF80A0C07FA0BF
80A0C07FA0BF80A0C07FA0BF7F9FB07090B0000000000000000000000000000000000000000000\
00000000000000000080A0BF80A0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF
80A0C080A0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80AFC0\
80A0BF80A0C080A0BF80AFC080A0BF80A0C080A0BF80AFC080A0BF80A0C03F505F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4F5F6040506080AFBF80A0BF80AFC080000080AFBF8080903F0000800000000000000000000000\
000000000000000000000000000000800000800000800000000000000000300000
5000001F00000000000000000000000000001F0000800000800000800000000000000000000000\
00000000000040506080AFBF80AFBF80AFBF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000001F7F001F80001F9F001F9F001F9F00\
1FA0001F9F001F9F00003F00004000000000000000000000000000000000000000
1F7F001FA0001F9F001FA0000000000000000000004050607F9FBF80A0C07F9FBF7FA0C07F9FBF\
80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C07F9FBF80A0C0
7F9FBF7F9FB07090AF000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000007F9FBF7F9FBF80A0BF7FA0BF80A0C07F9FBF80A0BF
7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF\
80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF
3F4F5F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000405F6040505F80A0BF
80A0BF8000007F000080000080000080A0B07FA0BF80AFBF7F9FB01F00006000001F0000000000\
000000000000000000000000000000300000800000800000500000000000000000
0000003F00006000001F00000000000000000000000000008000008000008000001F0000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000040001F9F001FA0001F9F001FA0001F9F001FA000\
004000004000000000000000000000000000000000000000000000000000000000
0000001F7F001FA0001F9F001FA0000000000000000000004050607F9FBF80A0C07F9FBF80A0C0\
7F9FBF80A0C07FA0BF80A0C07F9FBF80A0C07FA0BF80A0C07F9FBF7F9FB07090B0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000007FA0BF7FA0BF80A0C07FA0BF80A0C080A0BF
80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80A0C080A0BF80A0C0\
7FA0BF80A0C080A0BF80A0C07FA0BF80AFC080A0BF80A0C07FA0BF80AFC080A0BF
80A0C03F4F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000405060405F6080A0BF80AFC0800000
80000080A0BF80AFC080A0BF806F7080404F80000080404F80AFBF7FA0B00F0F0F1F0000500000\
0000000000000000000000000000000000004F0000800000800000200000000000
0000000000000000003000006F00001F00000000000000000000006F00008000008000003F0000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000040001F9F001F9F00003F00004000003F00\
000000000000000000000000000000000000000000000000000000000000000000
0000000000001F7F001FA0001F9F001FA0000000000000000000004050607F9FBF80A0C07F9FBF\
7FA0C07F9FBF80A0C07F9FBF7FA0C07F9FBF80A0C07090AF7090B0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000F0F7F9FB080A0C07F9FBF80A0BF7FA0BF80A0C0
7F9FBF80A0BF7FA0BF80A0BF7F9FBF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF\
80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF
7FA0BF80A0BF3F4F5F000000000000000000000000000000000000000000000000000000000000\
00000000000000000040505F4050607FA0BF80A0BF7F00008000007FA0BF80AFBF
7FA0BF80A0BF7FA0BF80AFBF7FA0BF80A0BF7FA0BF7FA0B07F404F80AFBF7F9FB00F0F0F000000\
5000001F00000000000000000000000000000000007F0000800000800000000000
0000000000000000000000000000000000006000006F00000000000000004F0000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000002000004000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000001F7F001FA0001F9F001FA0000000000000000000004050607F9FBF80A0C0\
7F9FBF80A0C07F9FBF80A0C07090B07F9FB0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F7F9FB080A0C07FA0BF80A0C07FA0BF
80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C0\
80A0BF80A0C07FA0BF80A0C080A0BF80A0C07FA0BF80AFC080A0BF80A0C07FA0BF
80AFC080A0BF80A0BF3F4F5F000000000000000000000000000000000000000000000000000000\
405060405F6080A0BF80AFC0800000800000800000806F7080A0BF80AFBF80A0BF
80AFC080A0BF80AFBF80A0BF80AFC080A0BF80AFBF80A0BF80AFBF80A0BF80AFBF7FA0B00F0F0F\
0000000000004F00000F0F0F0000000000000000003F00008000008000006F0000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004050607F9FBF\
80A0C07090AF7090B0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000F0F7F9FB080A0BF7F9FBF80A0BF
7FA0BF80A0BF7F9FBF80A0BF7FA0BF80A0BF7F9FBF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF\
80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF
7FA0BF80A0BF7FA0BF40505F000000000000000000000000000000000000000000405060405060\
80A0BF7FA0BF8000007F000080404F7F00008000007F00008060707FA0BF80A0BF
7FA0BF80AFBF7FA0BF80A0BF7FA0BF7FA0B07FA0BF80A0BF7FA0BF80AFBF7FA0BF80A0BF7F9FB0\
0F0F0F0000000000003000003000000000000000000000003000007F0000800000
300000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F405060\
7090B0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F7F9FB080A0C07FA0BF
80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0C0\
7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C080A0BF
80A0BF7FA0BF80AFC080A0BF40505F000000000000000000405F6040506080AFC080A0BF80809F\
7FA0BF80AFC080A0BF80AFBF7F9FB080202F8000008000007F607080AFBF80A0BF
80AFBF7FA0BF80AFBF80A0BF80AFBF806070806F7080A0BF80AFBF80A0BF80AFBF80A0BF80AFBF\
7F9FB00F0F0F0000000000000000005000003000000000000000006F0000800000
800000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F0F0F0F7FA0C07F9FBF\
3F4F5F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000F0F7F9FB080A0BF
7F9FBF80A0BF7FA0BF80A0BF7F9FBF80A0BF7FA0BF80A0BF7F9FBF80A0BF7FA0BF80A0BF7F9FBF\
80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF
7FA0BF80A0BF7FA0BF80A0BF7FA0BF40505F40506080A0BF7FA0BF8000007F000080A0BF7FA0BF\
80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80202F7F00008000007F202080A0BF
7FA0BF80A0BF7FA0BF80A0BF7FA0BF8060707F404F80AFBF7FA0BF80A0BF7FA0BF80AFBF7FA0BF\
7FA0B00000000000000000000000000000000F0F0F4F0000000F0F000000800000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F100F0F0F80A0C07F9FBF80A0C07F9FBF80A0C0\
7F9FBF405060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F7F9FB0
80A0C07FA0BF80A0C07FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C0\
7FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF
80A0C07FA0BF40505F3F4F5F405060000000405F607F00008000007F00008000007FA0BF80AFBF\
80A0BF80A0BF7FA0BF80AFBF80A0BF80A0BF7FA0BF806F708000008000007F0000
80A0BF80A0BF80AFBF7FA0BF80AFBF8060708000007F202080AFBF80A0BF80AFBF7FA0BF80AFBF\
80A0BF7FA0B00000000000000000000000000000003F0000300000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0F0F0F7FA0C07F9FBF7F9FBF7F9FBF7FA0C07F9FBF7F9FBF7F9FBF\
7FA0C07F9FBF3F4F5F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000F0F
7F9FB080A0BF7F9FBF7FA0BF7F9FBF80A0BF7F9FBF7FA0BF7FA0BF80A0BF7F9FBF7FA0BF7FA0BF\
80A0BF7F9FBF80A0BF7FA0BF80A0BF7F9FBF80A0BF7FA0BF80A0BF7FA0BF80A0BF
3F4F5F40505F0000000000003000008000000000004F404F7F404F8000007F000080202F7FA0BF\
80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF7F9FB07F0000800000
7F00008060707FA0BF80A0BF7FA0BF8000007F00008000007F00007F9FB07FA0BF80A0BF7FA0BF\
80AFBF7FA0B07FA0B0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F0F10107F9FBF7FA0C07F9FBF80A0C07F9FBF7FA0C07F9FBF80A0C07F9FBF80A0C0\
7F9FBF80A0C07F9FBF40505F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F7F9FB080A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF\
7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF3F4F5F405060000000
0000007F00001F0000000000000000000000000000000000405F607F80908000007F0000800000\
709FAF80AFBF7FA0BF80A0BF7FA0BF80AFBF7FA0BF80A0BF7FA0BF80AFBF802020
8000007F000080000080000080AFBF7FA0BF80AFBF80A0BF80AFBF7F0000806F7080A0BF80AFBF\
7FA0BF80AFBF80A0BF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F0F0F10
7F9FBF7F9FBF7F9FBF7F9FC07F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7FA0BF7F9FBF\
7F9FBF7F9FBF7FA0BF7F9FBF3F4F5F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000F0F7F9FB080A0BF7F9FBF7FA0BF7F9FBF80A0BF7F9FBF7FA0BF7F9FBF80A0BF7F9FBF\
7FA0BF7FA0BF80A0BF7F9FBF7FA0BF3F4F5F40505F0000000000007F0000800000
7F00000000000000000000000000000000000000000000000000004050607FA0BF80202F7F0000\
8000007F606F80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF7FA0BF80A0BF
7F606F8000007F00008000007FA0B080A0BF7FA0BF80A0BF7FA0B080A0BF7F80908060707FA0B0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F0F10107F9FBF7FA0C07F9FBF
80A0C07F9FBF7FA0C07F9FBF80A0C07F9FBF7FA0C07F9FBF80A0C07F9FBF7FA0C07F9FBF80A0C0\
7F9FBF7FA0C07F9FBF80A0C03F4F5F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F7F9FB080A0C07FA0BF80A0BF7FA0BF80A0C07FA0BF80A0BF7FA0BF80A0C0\
7FA0BF80A0BF7FA0BF40505F3F4F5F0000000000008000007F00008000007F0000
6F0000000000000000000000000000000000000000000000000000000000405F607FA0BF806070\
7F00008000007F202080A0BF7FA0BF80A0BF7FA0BF80AFBF7FA0BF80A0BF7FA0BF
80AFBF7F80908000007F00008000007F809080AFBF7FA0BF80AFBF7FA0BF80AFBF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F10000F0F7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF\
7F9FBF7F9FBF7F9FBF7F9FBF7FA0BF3F4F5F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000F0F709FB07FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF\
40505F3F4F5F0000000000000000003000000000001F00000000004F00007F0000
7F00006000000000000000000000000000000000000000000000000000000000004050607FA0BF\
8060707F00008000007F000080A0BF7FA0B080A0BF7FA0BF80A0BF7FA0B080A0BF
7FA0BF80A0BF7FA0B08000007F00008000007F606F80A0BF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F10100F0F0F7F9FC07F9FBF7FA0C07F9FBF7F9FC07F9FBF7FA0C07F9FBF
7F9FC07F9FBF7FA0C07F9FBF7FA0C07F9FBF7FA0C07F9FBF7FA0C07F9FBF80A0C07F9FBF7FA0C0\
7F9FBF80A0C07F9FBF7FA0C07F9FBF80A0C03F4F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000F7F9FB07F9FBF80A0BF7FA0BF80A0BF7FA0BF40505F3F4F5F000000\
0000000000000000000000000000000000000000000000000000000000001F0000
8000007F0000800000000F0F000000000000000000000000000000000000000000000000405F60\
7FA0BF7F9FB07F00008000007F000080809F7FA0BF80A0BF7FA0BF80AFBF7FA0BF
80A0BF7FA0BF80AFBF7FA0BF80202F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF\
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7FA0BF3F4F5F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007F9FB07F9FBF7FA0BF3F4F5F3F4F5F0000000000007F0000\
7F00000000000000000000000000000000000000000000000000000000001F0000
7F00007F00007F00001F0000000000000000000000000000000000000000000000000000000000\
60202F7F20208080907F20207F00007F000080404F7FA0B07FA0BF7FA0BF80A0BF
7FA0B07FA0BF7FA0BF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000007F9FC07F9FBF7F9FC07F9FBF7F9FC07F9FBF7FA0C07F9FBF7F9FC07F9FBF
7FA0C07F9FBF7F9FC07F9FBF7FA0C07F9FBF7F9FC07F9FBF7FA0C07F9FBF7FA0C07F9FBF7FA0C0\
7F9FBF7FA0BF7F9FBF7FA0C07F9FBF7FA0BF7F9FBF7FA0C03F4F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000040506040505F0000000000007F00008000007F00003F0000\
000000000000000000000000000000000000000000000000000000000000000000
0000006000008000007F00005000000000000000000000000000000000006F00006000001F0000\
4F404F80A0BF7FA0BF8060707F00008000007F202080A0BF7FA0BF80A0BF7FA0BF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000007F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF\
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF3F4F5F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F505F4050607F9FBF7FA0BF7F00007F00003000007F00007F00007F0000600000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004F00007F00007F00007F00000000006000006000001F00001F0000000000000000\
0000003F505F7FA0BF7FA0BF8080907F00007F00007F0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000007F9FC07F9FBF7F9FC07F9FBF7F9FC07F9FBF7F9FC07F9FBF
7F9FBF7F9FBF7FA0C07F9FBF7F9FBF7F9FBF7FA0C07F9FBF7F9FBF7F9FBF7FA0C07F9FBF7F9FBF\
7F9FBF7FA0C07F9FBF7FA0BF7F9FBF7FA0C07F9FBF7FA0BF7F9FBF7FA0C03F4F5F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000405060
40506040506080A0BF7F9FBF7F00007F9FBF80A0BF7F9FB00F0F0F0000001F00007F00007F0000\
7F00000F0F0F000000000000000000000000000000000000000000000000000000
0000000000000000001F00008000007F00008000003000001F0000000000000000000000000000\
00000000000040506080A0BF7FA0BF80A0BF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000007F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF
709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF7F9FBF\
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF
3F4F5F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004050603F505F7FA0BF
7F9FBF7F00007F00007F404F7090AF7FA0BF7F9FBF7FA0BF7090B0000F0F0000000000007F0000\
7F00007F0000300000000000000000000000000000000000000000000000000000
0000000000000000000000000000007F00007F00007F0000300000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000007F9FBF7F9FBF7F9FBF7F9FC07F9FBF7F9FBF7F9FBF
7F9FC07F9FBF7F9FBF7F9FBF7F9FC07F9FBF7F9FBF7F9FBF7FA0C07F9FBF7F9FBF7F9FBF7FA0C0\
7F9FBF7F9FBF7F9FBF7FA0C07F9FBF7F9FBF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF
7FA0BF3F4F5F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004050603F50607FA0BF7F9FBF7F00007F0000
8040507F00007F00007F000080A0BF7F9FBF7FA0BF7F9FBF80A0BF7F9FB00F0F0F000000000000\
4F00007F00007F0000300000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F00007F0000800000600000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000007090BF7F9FBF709FBF7F9FBF709FBF7F9FBF
7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF\
7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF
7F9FBF7F9FBF3F4F5F000000000000000000000000000000000000000000000000000000000000\
0000000000004050603F505F7FA0BF7F9FBF7F9FBF7F9FBF7FA0BF7F9FB07FA0BF
7F9FBF7F60707F00007F00007F5F6F7FA0BF7F9FB07FA0BF7F9FBF7FA0BF7090B0000F0F000000\
0000001F00007F00007F0000600000000000000000000000000000000000000000
0000000000000000000000000000000000000000003000007F00007F0000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000007F9FBF7F9FBF7F9FBF7F9FC07F9FBF
7F9FBF7F9FBF7F9FC07F9FBF7F9FBF7F9FBF7F9FC07F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF\
7F9FBF7FA0BF7F9FBF7F9FBF7F9FBF7FA0BF7F9FBF7F9FBF7F9FBF7FA0BF7F9FBF
7FA0BF7F9FBF40505F000000000000000000000000000000000000000000000000000000000000\
3F50604050607F9FBF7FA0BF7F20207FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF
7FA0BF7F9FBF7F202F7F00007F00007F9FBF7FA0BF7F9FBF7FA0BF7F9FBF80A0BF7F9FB00F0F0F\
0000000000001F00007F00007F000080000000000F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007090BF7F9FBF709FBF7F9FBF
709FBF7F9FBF709FBF7F9FBF709FBF7F9FBF709FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF\
7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF
709FBF7F9FBF7F9FBF3F4F5F0000000000000000000000000000000000003F505F405060709FB0\
7F9FBF7F00007F00007F9FB07F9FBF7F9FBF7FA0BF7F9FB07F9FBF7F9FBF7FA0BF
7F9FB07F9FBF7F9FBF7F00007F00007F404F7F9FBF7FA0BF7F9FB07FA0BF7F9FBF7FA0BF7090B0\
000F0F0000000000000000006000007F00007F0000300000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007F9FBF7F9FBF7F9FBF
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF\
7F9FBF7F9FBF7F9FBF7FA0BF7F9FBF7F9FBF7F9FBF7FA0BF7F9FBF7F9FBF7F9FBF
7FA0BF7F9FBF7F9FBF7F9FBF40505F0000000000003F50604050607F9FBF7FA0BF7F00007F0000\
7F404F7F00007F3F4F7FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF
7FA0BF7F9FBF7F60707F00007F00007F60707FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF7F9FB0\
00000F0000000000000000000000003000007F00007F00004F0000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000007090BF7F9FBF
709FBF7F9FBF7090BF7F9FBF709FBF7F9FBF709FBF7F9FBF709FBF7F9FBF709FBF7F9FBF709FBF\
7F9FBF709FBF7F9FBF709FBF7F9FBF709FBF7F9FBF709FBF7F9FBF709FBF7F9FBF
7F9FBF7F9FBF709FB07F9FBF7F9FBF3F4F5F3F4F5F7F9FBF7F9FBF7F00007000007F7F907090B0\
7F9FBF709FB07F60707F3F4F7F9FB0709FB07F9FBF7F9FBF7FA0BF709FB07F9FBF
7F9FBF7FA0BF7F9FB07F00007F00007F00007090AF7F9FBF7F9FB07FA0BF7F9FB07FA0BF7F9FB0\
7F9FB00000000000000000000000003000007F00007F0000600000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000007F9FBF
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF\
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7FA0BF7F9FBF
7F9FBF7F9FBF40505F3F4F5F0000000000007F00007F00007F00007F00007F40507F9FBF7FA0BF\
7F9FBF7FA0BF7F9FBF7FA0BF7F5F6F7F202F7090B07FA0BF7F9FBF7FA0BF7F9FBF
7FA0BF7F20207F202F7F60707F00007F00007F202F7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF7FA0BF\
7F9FBF7F9FB000000F00000000000000000000000F7F0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
7090BF7F9FBF709FBF7F9FBF7090BF7F9FBF709FBF7F9FBF7090BF7F9FBF709FBF7F9FBF709FBF\
7F9FBF709FBF7F9FBF709FB07F9FBF709FBF7F9FBF709FB07F9FBF709FBF3F4F5F
3F4F5F0000000000004F00007000001F00001F00004F404F7000007F00007F00007F6070709FB0\
7F9FBF7F9FBF7F9FBF709FB07F9FBF7F9FBF7F6070701F20709FB07F1F207F202F
707F907F80907F9FB07FA0BF703F4F7F00007F00007F6070709FB07F9FBF7F9FB07FA0BF709FB0\
7F9FBF7F9FB07F9FB0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000007F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF\
7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF7F9FBF40505F3F4F5F3F4F5F000000
0000000000000000000000000000000000000000000000004050607F20207F00007F00007F4050\
7F9FBF7F9FBF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7090AF7F00007F5F6F
7FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7F20207F00007F0000709FB07F9FBF7FA0BF7F9FBF7FA0BF\
7F9FBF7FA0BF7F9FBF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000007090B07F9FBF709FBF7F9FBF7090B07F9FBF709FBF7F9FBF7090B07F9FBF709FBF\
7F9FBF7090B07F9FBF709FBF7F9FBF3F4F5F3F4F5F0000001F0000700000000F0F
000000000000000000000000000000000000000000000000000000405060703F4F7F0000700000\
7F202F709FB07F9FBF709FB07F9FBF709FB07F9FBF709FB07F9FBF709FB07090AF
7F1F207F8090709FB07F9FBF7F9FB07090AF7000007F00007F00007FA0BF709FB07F9FBF7F9FB0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF709FBF7F9FBF7F9FBF7F9FBF\
709FBF7F9FBF3F4F5F3F4F5F0000000000000000000000000000004F00007F0000
7F00003000000000000000000000000000000000000000000000000000004050607F5F707F0000\
7F00007F00007F9FBF7F9FBF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF7FA0BF7F9FBF
7FA0BF7F7F907F202F7F7F907FA0BF7F9FBF7F60707F00007F00007F404F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007F9FBF7090B07F9FBF7090BF7F9FBF7090B07F9FBF709FBF7F9FBF3F4F5F\
3F4F5F0000000000005F00000000000000000000000000000000000000007F0000
7000007F0000700000300000000000000000000000000000000000000000000000405060707F90\
7F00007000007F0000707F907F9FBF709FB07F9FBF709FB07F9FBF709FB07F9FBF
709FB07F9FBF709FB07F8090701F207090AF709FB0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007F9FBF709FBF7F9FBF709FBF7F9FBF7F9FBF3F4F5F3F4F5F000000\
0000007F00007000000F0F0F000000000000000000000000000000000000700000
7F00007F00004F0000700000300000000000000000000000000000000000000000000000405060\
7090B07F00007F00007F00007F5F707F9FBF7F9FBF7FA0BF7F9FBF7F9FBF7F9FB0
7FA0BF7F9FBF7FA0BF7F9FB07FA0BF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007F9FBF7090B03F4F5F3F4F5F0000000000007F0000700000\
7F00005F0000000000000000000000000000000000000000000000000000300000
7000007F00007000003000001F00007F00002F00000000000000000000000000000000003F4F5F\
7F9FBF7000007F00007000007F404F709FB07F9FBF709FB07F9FBF709FB07F9FBF
709FB07F9FBF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000001F00000000000000001F0000\
3000007F0000300000000000000000000000000000000000000000000000000000
3000007000007F00007F00000000000000001F00004F00004F00001F0000000000000000000000\
3F505F7F9FBF7F20207F00007000007F00007F9FBF7F9FBF709FB07FA0BF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F00005F0000000000000000000000000000000000000000000000
0000006000007000007F00002F00000000000000000000001000004F00004000001F0000000000\
0000003F4F5F7F9FBF705F6F7F00007000007F0000707F90000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F0000300000000000000000000000000000000000
0000000000007F00007000007F00003000000000000000000000000000000000004F00007F0000\
1F00000000003F505F7F9FBF7F5F6F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F0000600000000000000000000000000000
0000000000002F00007F00007000007F0000000000000000000000000000000000000000000000\
4F0000700000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000300000300000000000000000
0000000000000000006000007F0000700000600000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000F0F5F00000F0F0F
0000000000000000000000007000007F0000700000300000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000001F0000
4F00000000000000000000003000007000007F0000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003000002F0000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000001F0000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[3]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7WP004jn`[k2YWJV`[k2_/:n`[kF_T7402P000f^4J`050005
WHBMWF/0BP081`001PNL0007]Jne[kFMX`02Z@0<[kF_]Jne[kF_]Jn@3@000if4J`030004WHBMWDX0
20L000H7VP002:n`[k2_/:n`0Yd03Zn`[k2_/:n`[kF_]Jn@3@000h@0Q00300>MBP081`001PNI000J
[kFMZJ>e[kF_XifY[kF_]Jne[kF_]Jne[i0=0005JhA[06/00id00h@0001700P700061iP000V_WIfM
[k2_/:l00ZT04IfY/:n`[k2_/:n`[kF_/:n@00h000A[WF][0Yd00f/0001700P700061iH000fe[ifM
ZJne[kF_]JnS00:M012_]Jne[kF_]Jne[kF_]Jn@3P000f^MQ003W@03J`0004H020L000H7T`001K2_
WIfS00BM00NY/:n`[k2Z00BM012`[k2_/:n`[k2_/:n`[i1a3`000f^MW@1900P700061i4000N_]JnS
[kFZ00BM0ZT01[F_]K2eZ@BM012e[kF_]Jne[kF_]Jne[i1g10000f/000080003Q6][04P020L000H7
S`002[2_/:n`[k2_/:l4W@:S00B`[k2_0ZT2W@0AXif`[k2_/:n`[k2_/:n`[i000`02JeD020L000H7
SP003KF_]Jne[kF_]Jne[j<00id01Z>N]Jne[`:Y0Yd04:>M]Jne[kF_]Jne[kF_]Jl30003Q9f405@0
20L000H7S`003K2_/:n`[k2_/:n`[j<00id01j>_/:n`[jT00id03j>M/:n`[k2_/:n`[k2_/0040003
WH@000d000=a0000@`081`001PN=000>W@00]Jne[kF_]Jne[kD2[`BM0jl00kF_Z@04W@0>ZJZ_]Jne
[kF_]Jne[i030003WHBM00/000B@MkF_A0081`001PN;00:M10002k2_/:n`[k2_/:n`00:_19d01:Z`
[k02[`BM00fS/:n`[k2_/:n`[k1a008000A[WIe[2@001W6@[k2_/4@020L000H7R@03W@:410003KF_
]Jne[kF_]Jne[kD019d01Zne[kF_]@FM00bY[kF_]Jne[kF_T7820003WH@000L000R@MkF_]Jne[d@0
20L000H7Q`06W@03J`0000<000f`[k2_/:n`[k2_/:n`00>M00RT[k2_/:n`Z@BM00f`ZK2_/:n`[k2_
T6^M00T000Z@[k2_/:n`[k2_@`081`001PN600:M0X@4W@@0011a0;F_]Jne[kF_]Jne[kFN0Yd02J>_
]Jne[kF_X`04W@06[kF_]Jne0Zl00hZ4Q007000<Mi2_]Jne[kF_]Jne@`081`001PN30003J`0000H0
00B4WIf41P003Jn`[k2_/:n`[k2_/:T019d01K2_/:n`00:_00>YWJ<019d01ZVMXif`L@D0011aT:n`
[k2_/:n`[kF_/:n`@P081`001PN100:M2@000ie[W@02W@D000je[kF_]Jne[kF_]JneZPBM01BY[kF_
]Jne[kFSZIf_WJnY]Jn@LP8001:@MkF_]Jne[kF_]Jne[kF_]Jm200P700061h000Yd00h@000080003
JifM00:M1@002k2_/:n`[k2_/:n`00>_0id0:JV`[k2_/:n`[k2_/:n`[k2_/:n@LI2_/:n`[k2_/:n`
[k2_/:n`[k2_044020L000H7OP05W@/000=[WId00Yd5000>]Jne[kF_]Jne[kF_]Ih2W@0YY:>e[kF_
]Jne[kF_]Jne[kF_T7:e[kF_]Jne[kF_]Jne[kF_]Jne[kD0@@081`001PMl00:M00=[WId00Yd<00FM
015[0000[k2_/:n`[k2_/:n`[`02X`:M016S[k2_/9J`[k2_/:n`[k2_T002000FLK2_/:n`[k2_/:n`
[k2_/:n`[k2_/40020L000H7NP02J`@00id=0004WHBMW@9[0P003g6_]Jne[kF_]Jne[kF_Z@03W@0?
/:neXkF_]Jne[kF_]Jn@00<001J@[kF_]Jne[kF_]Jne[kF_]Jne[kF_@0081`001PN00004JifMJ`d0
00J4JifMQ6/3000=[k2_/:n`[k2_/:n`ZP04W@0=XifY[k2_/:n`[k1aT004000GLI2_/:n`[k2_/:n`
[k2_/:n`[k2_/:l0?`081`001PN10004JifMJ`d000>4WId00Yd02X@0076_]Jne[kD3[`04]JneZPBM
00ZY[kF_]Jne[i1a1`005WNe[kF_]Jne[kF_]Jne[kF_]Jne[kDo00P700061gH000>MJ`002@000hBM
W@0>0003JifM00:M0`001K2_/:n`00:_00JM/:n`[jT2W@0:ZZn_/:n`[k1aT0T001J@[k2_/:n`[k2_
/:n`[k2_/:n`[k2_?`081`001PMb0004]JnMW@`019d>0005JifMWF/00`005[F_]Jne[jV_]JnSWJnZ
]Jne[kF_T74<000FT:ne[kF_]Jne[kF_]Jne[kF_]Jne[ch020L000H7L0001YJ`WIf4R`d00Yd00h@0
000>00>M00M[002_/:n`00:_00>M/:T00jT02:n`[k2_/76@3P005W6`[k2_/:n`[k2_/:n`[k2_/:n`
[k0n00P700061fh000NeWIf_ZJnY00:M00=b00002@001HBMWIe[00l000R4WIfM01R_]@:_0Yd02ZZe
[kF_]Jne[i0B000FL[F_]Jne[kF_]Jne[kF_]Jne[kF_]Cd020L000H7K00029J`WK2_/:n`0Zl2X`03
L@0000L000>4Jid00id?0004JifMW@8000n`[k2M[jV`U[2_/9J`LI004`005Y2_/:n`[k2_/:n`[k2_
/:n`[k2_/:lm00P700061fX000>e[id00Yd02jne[kF_]JneU[1[00H000BMQ9d01Id?00>M00D0J`2S
WP02[`08]Jne[kF_T78F000FT:ne[kF_]Jne[kF_]Jne[kF_]Jne[c`020L000H7J0001IJ`Q9f400:M
00nSU[2_/9J`[k2_Rh@0J`000h@016/00003W@03Q00000d01Id02hBY[jn`[k2_/76@01P001Ia/:n`
[k2_/:n`[k2_/:n`[k2_/:n`?0081`001PMV0003]Jne00:M01JFXifMWZ>_]Jne[kF_]HZ@Q9f4W@1[
1@03W@03Q00000`000E[WIfMJ`020007[kF_]Jn@LP0K000FMkF_]Jne[kF_]Jne[kF_]Jne[kF_]C/0
20L000H7I@002[2F[hB`U[2F[h@2W@0?/9J`[k2F/:n`LGL0Q6^400H000E[WIfMJ`0=0004Q9d0J`<0
00F`[k2_T00M000FT:n`[k2_/:n`[k2_/:n`[k2_/:n`[c/020L000H7I0002[F_]Jne[kF_]Jl4W@0:
]Jne[kF_]JneLP8000=[Q9d01P0018BMWF/50003Q01[00@000>4J`0010001AR_]G:@01h001Ib]Jne
[kF_]Jne[kF_]Jne[kF_]Jne>`081`001PMR000KU[2F/9J`U[2F/9J`UYf4WHB_U[2F/9J`U[1a00<0
0X@00f/0000200BM1P001if4Jf^MQ6/020000g40000O000F[k2_/:n`[k2_/:n`[k2_/:n`[k2F[cX0
20L000H7H0004K2_WIfe[kF_]Jne[kF_]JnS00>M00Z_Z[F_]Jne[kEb0`000f^MJ`0400>M1P0018BM
WId2J`L000>@M`008@005;F_]Jne[kF_]Jne[kF_]Jne[kF_?0081`001PMO000P/9J`UYf4[hZ`U[2F
/9J`U[2FXhBMQ:VF/9J`U[2F/7440009Jh@006^4WHA[00D000A[08B41`001G6@[jnF0240011a/:n`
[k2_/:n`[k2_/:n`0Zl00iH0000k00P700061e/000V_]Jne[kF_]Jl019d2X`0:/:n`[kF_]JneZ@>M
00ZS]Jn`[k2_/7:@10000hA[Q003W@d000V@L[F_]Jne[k008@003i1b]Jne[kF_]Jne[kF_/01100P7
00061eT002R`UYf4/9J`U[2F/8BMQ9fAXhBSUZnF/9J`UZV@WHBMQ;2_/9J`[k1a1P02Q003WHBM00/0
00^@U[2F/9J`U[2F[`0Q0009T:n`[k2_/:n`00>_@`081`001PMI000<UYf4/:neU[2_]IJ`19d02YJY
[j>F]Jne[kD2[`0<ZYfMWJn`[k2_/:n@1`02J`03WH@000P000ebT:ne[kF_]Jne[kFF024000Yb]Jne
[kF_]JneA@081`001PMI000YWHB_U[2F/9J`U[2F/8BMQ9fF/9JYUZnF/9J`U[2FWHBMTK2_/9J`[gL0
20000hA[0007000?LK2_/9J`U[2F/9J`U[2F024000Qa/:n`[k2F/4H020L000H7E`02W@0=Q6^4]IJ`
[kFF/:n`ZP04W@0>Y:n`[jVMXjZe[kF_]Jl2W@:_00J`[k2_/78@000AT7:`[kF_]Jne[kF_]Jne[k00
8@001Y2_]Jne[dP020L000H7FP02J`0WQ:n@/9J`U[2F/8^MQ9f:ZIJ`U[24[hB`U[2F/8BMQ:VF/9J`
U[1a00d001=aT:n`[k2_/9J`U[2F/9J`U[2F024000EaT9J`UP1900P700061e<00Yd00f/00005000;
RZV@/9JeU[2_/:<00id03ZV`[kF_]I2SWJnF]Jne0Yd02:>`[k2_/:n@2`005I1b]Jn`[kF_/:ne[kF_
]Jne[kF_]@0R0003L[D004X020L000H7DP0018BMQ9d3Q0D002F@Y9J`U[2F/9J`Q9f4WIJ`U[2F/9JY
Q:n@/9JMQ:>4/9J`U[1a00X001J@[k2F/:n`U[2_/9J`U[2F/9J`U[2F8P000h/000090003NP0003h0
20L000H7D002W@<000E[06^4J`05000:UZ>A[iJ`[k2_/0BM00f_]IJe[kFF]JVSTKFY00:M00NS[k2_
/:>`00P001UbT:n`[k2_/:n`[kF_]Jne[kF_]Jne[kFF01l000BSZJ>@200015UoDWXo00P700061dl0
00>M00003@009;2FY8^YU[2F/9JMQ9f4WIJ`U[2F/9J`Q:>4ZHBMQ:V;XhBYUPH001^@LK2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J`UZl07P001HZMQ;1a00L000EINUUjDP0o00P700061dd000>4W@0040002[2F
[hZ`[k2_/9H4W@0</9JeU[2F]IJeUZ>Z1Id01JZY[k1a00@001f@[kF_]Jn`[k2_/:n`[k2_]Jne[kF_
]Jne[kFF/00K0008Y:>MWJV_T7470004FGmINSl020L000H7BP000hBMJ`0E0013U[2;ZHZ`U[2FY8BM
U[2F/9J`U[2F/9J_Q:V;/9J`U[2FMg5gU[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`UP0H0004
XhBMQ0>M00BF/:mg200015UjFGX7009==@081`001PM800FM5P002IJeUZVF/:neQ002W@0;/:n`[k2_
/:n`[k000Zl0:jZe[kF_]JneLY2_/:n`[kF_/:ne[kF_]Jne[kF_]Jne[kF_]Jne[kF_]IH05@000jZY
W@05W@06XkF_]Jn@1`001GmIOe9j00@000E=DUUoF@0d00P700061dH000V4WF^MQ9f4WF/02002J`/0
04:F/9JYRk2FWHBMU[2F/9J`U[2F/9J`U[2F/9J`UWMa076`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`
U[2F/9J`UZlC000@[iJ_Rk2FWHBMQ:nF/9J`L@L000EINUUjDP030006CGYINUUj=0081`001PM400:4
00JM0000WH@3W@03J`0000H00V/:000M/9JeUZV:/96MQ;2_/:n`[k2_/:n`[kFF]JneLI0010008W:`
[k2_/:ne[kF_]Jne[kF_]Jne[kF_]Jne[kF_]JneU[0A0007/:ne[kF_Z@04W@07XkF_]JneLP070004
FGmINP8000Q=OeUoFGmBNS<020L000H7C00018BMQ8@90003Q00000P001^F/9J`UZV;ZHBMR[2F/9J`
U[2F/9J`U[2FMg401P008WNF/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9H?000DUZnF/9J`
U[24XhBMQ:nF/9J`UWL7000<NUUjFGY=NUUjFGYB=@081`001PM20003WH@000P000A[WHBM20000if4
0009000HU[2F]IJYQ:>4ZJn`[k2_/:n`[k2F]Jn@20008W:@[k2_/:n`[kF_]Jne[kF_]Jne[k2_]Jne
[kF_]Jne[k0=0009/:ne[kF_]Jne00:_19d01ije[kF_]G8020002GmIOeUoFGmBNP0g00P700061d00
00>4WF/02`001V^MQ9d0J`D000>MQ8@02@005k2F/9J`UZB4WI2ZU[2F/9J`U[2F/75g00X0029a/9J`
U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J_2P006HZ_U[2F/9J`U[2F/8ZMQ9f4[iJ`U[2F/740
1`001eUjFGYINU80>@081`001PLl0007T7:MQ9f4J`0;0006JhBMQ9e[1@04W@P001FF/9J`U[F@WHBM
[k2_/:n`[k2_/78030008Y2_/:n`[k2_]JneU[F_]Jne[k2_]Jn`[kF_]Jne[kF_]IH8000>Y:>TZKF_
]Jne[kF_]Jl5W@08[kF_]Jne[i070006OeUoFGmB1P000dd0000a00P700061cX000VF/9JMQ9f4WH@0
3@001HBMQ9e[008000>MQ6/00V/00hBM0007000CU[2F/9J`RYf;/9J`U[2F/9IgL@0>000RMiJ`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J_UPH001fSRYfF/9J`U[2F/9J`U[2:WHBMQ:VF/9J_U[2F
M`070004NUUjDP@000E=07YINP0b00P700061cT000bF/8BMT9f4WHBMQ6/<0006Q9f4WHBM1P02J`L0
016`U[2F/8BMTK2_/:n`[k1bT00@000RL[2_/:n`[kFF]JneU[F_]IJe[k2_]Jn`[kF_/:ne[kF_/0@0
016SY9fMZKF_]Jne[kF_]Jne[`05W@09[kF_]Jne[kEb00L000=BNP0010001TeIOeUoFC8020L000H7
=P004IJ`U[2F/9J`U[2FRXBMQ9e[00/00V/018BMQ9d;000@WF/00:nFZH^`U[2F/9IgLAD004Ea/9J`
U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J_UZn@[i2_Q9f:/9J`U[2F/9J`U[24WHBMQ;2F/9J`
U[2F/7402P002deINUUjFGYINUV`030020L000H7=0004[2FWHB_U[2F/9J`U[2FMhBMQ09[2`02J`04
Q9f4J`X000j4J`00U[2F/9J`U[2FT1L0022@U[2F]IJ`[k2_/:n`[k2_]IJeU[FF]IJeU[F_/:n`UP80
01B`[k2F]IJYQ9f4ZIJeU[2_]Jn`UPFM00ZF]JneU[F_]IJ@2P002deoFGmIOe9oFGn_030020L000H7
<P0059J`UYf4/9J`U[2F/9J`U[2FMf^M0h@<0004WHBMQ0X000>M00000P0029J`U[2F/75g6@007I2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J_00D002:`U[2F/9J`T9f4/9J`U[2F/9JYQ9f4WH^`U[2F
/9J`U[2F2P002gYINUV@LG]INUUj030020L000H7<@005YJ`Q9f4ZIJ`U[2F/9J`U[2F/9I[Q9d3Q0X0
00E[08B4W@0900:400Y[0000/9J`U[1b6`006g:`U[2F/9J`[k2F/:n`[kFF]IJeU[FF]IJeUP07000D
[k2F]IJeU[F4WHBYU[2_/:ne[j<4W@0:U[FF]JneU[F_]@T000a=DWmIMjneNX1IOeT`00P700061bl0
01f`UYf4WHBSQ;2F/9J`U[2F/9J`U[1aJhBMQ9d0J`0:0004Q9f4Q0L00X@00f/000030004[iIgLA`0
01Z@LK2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`UPP002>`U[2F/9J`U[2FWHB_U[2F/9JYQ9f4WH^`U[2F
/9J`U[2FM`090007NUVEU[2F/002NP03FID002l020L000H7;P007k24WI2MQ9f4XhB`U[2F/9J`U[2F
/9J@LF^4WHBM06/02@02Q003WH@000@00X@00`1[00050003LWL001h001R@U[2F/9J`U[2_/9J`[k2F
]IJeU[FF/9H:000C/9JeU[FF]IJeRYf4/:n`[k2MX`03W@0</9JeU[FF]IJeU[Eb20003DeIOeV`[kFF
P7bJ[kD0;`081`001PL/000KU[2;WI6YUZB4WHBMQ:nF/9J`U[2F/9J`UWMa00:400BMQ9e[2P000f^4
W@0300:400=[00001@000gMa[`0P000FLK2F/9J`U[2F/9J`U[2F/9J`U[2F[``002:F/9J`U[2F/9J`
UZ>@/9J`UZn4WHBMR[2F/9J`U[2F/9Ig1`003WMeNeUkU[2F/9J`U[2F;`081`001PL[000SU[2F/9J`
U[2F[i6MQ9fF[iJ`U[2F/9J`U[24RXA[Q9f4WF/02@0026^MQ9f4W@1[1P001G:@U[2F024001=b/9J`
U[2F/9J`U[2F]IJ`U[2F00l000^F]IJeU[FF]IJeT002W@03/:n`00FM00f`U[FF]IJeU[FF]Jn@00@0
015bT:neFGmk]IJe[kFF]Jne[`0^00P700061bT001n`U[2F/9J`U[2F/9JZT9f4WH^YU[2F/9J`U[24
Xi2;008000A[WHB42P0016^MQ6/80006LK2F/9J_8@004I2F/9J`U[2F/9J`U[2F/9J_014002:`U[2F
/9J`U[2F/8ZMT;2F[hBMQ9fF/9J`U[2F/9J`U[1a0P004i1a[iJ`UWYINWb`U[2F/9J`U[00;P081`00
1PLW000XU[24[iJ`U[2F/9J`U[2F/9JSQ9f:/9J`U[2F/8B_TJnFM`1[Q9f4W@T000>M06/0200027:`
U[2F/9J`8@003Y2F/9J`U[2F/9J`U[2F4`004IJeU[FF]IJeU[FF[ifSWJnS00BM02:_U[FF]IJeU[FF
]JmgLI2_/:ne[kFFOeVJU[FF]IJeU[F_;P081`001PLU000T/9JMQ9fF/9J`U[2F/9J`U[2F/9JMQ9f4
/9J`UZ>4ZY2`U[1a0P0019f4WH@80003W@0000L000Z@LK2F/9J`U[2F8@00376`U[2F/9J`U[2F/1D0
02V`U[2F/9J`U[2F/9J`Rif:WHBMQ9fF/9J`U[2F/9J`UWMaMiJ_U[2F[`02UP0=FGYINYJ`U[2F/7bE
F@0]00P700061b<002NFWHBMQ;2F/9J`U[2F/9J`U[2F/9J`UZV4WHBMQ:nA/9J`U[2F/7800P02Q003
WH@000800V/8000?LY2F/9J`U[2F/9J`U[2F024000Ub]IJ`U[FF/9H060007k2_/:n`[k2_/:neU[FF
[hBMQ9f4]Jne[kF_]JneLY000`005Y2_]IJe[kFFOeUoFGmIOeUoFGmIOeT/00P700061b<002NMQ9f4
WHB`U[2F/9J`U[2F/9J`U[2F/92FQ9f4WI2`U[2F/9J`UW800`02Q005W@24Q6/01`004GMa/9J`U[2F
/9J`U[2F/9J_024000MgU[2F/9J_01X001fF[iJ_UZnF[iJ_U[2F/9JYQ9f4WIJ_UZnF[iIgL@05000C
LK2F/9J`U[1kNUUjFGYINUUjF@02NP03N`0002X020L000H78@00:XBM09f@WHBMQ9fF/9J`U[2F/9J`
U[2F/9J`U[24WHBYU[2F/9J`U[2FM`8000I[WHBM06/6000DLY2F/9J`U[2F/9J`U[2F/9J`U[0Q0004
/9J`UQd001Z_/:n`[k2_/:neU[FF]IJ`RYf4/:ne[k1bT0L001J@L[FF]IJeU[FFP5UoFGmIOeVJN[F_
;0081`001PLP009[0P009i6FT9f4WHBMT;2F/9J`U[2F/9J`U[2F/9JSQ9f4Y9J`U[2F/9J`L@030003
Q6/000L001Ea/9J`U[2F/9J`U[2F/9J`U[2F/9H08@000iJ_000N000H[iJ_UZnF[iJ_U[2F/9J`UZn:
XiJ_UWMa2P002I2F/9J`U[2F[`03U@04UUVEF@:F00B`U[2F:`081`001PLN0003WH@000@002NF/9JM
Q9f4WI2`U[2F/9J`U[2F/9J`U[2FZHBMQ;2F/9J`U[2FMg800X@00f/00006000GL[2F/9J`U[2F/9J`
U[2F/9J`U[2F/9H080000i00000N000F[k2F/:n`[k2_]IJeU[FF]IJeU[EbT0`001Ib]IJeU[FF]Jne
[kF_/:ne[kF_]IJJ:`081`001PLL0003Q9d000L002RAUYJSRYf4WHZ`U[2F/9J`U[2F/9J`U[2;WHBS
Q;2F/9J`U[2FRP1[1`006GMa/9J`U[2F/9J`U[2F/9J`U[2F/9J`UZl07`000jYa000P000CUZnF[iJ_
UZnF/9J`U[2F/9J`L@0?000FLK2F/9J`UZnF[iJ_UZnF[iJ`NW]INRX020L000H76P0019f4WH@9000U
U[2F/92MQ9fF/9J`U[2F/9J`U[2F/9J_TIf4WI6`U[2F/9J`L@07000KLY2F/9J`U[2F/9J`U[2F/9J`
U[2F/9J`U[2F01h000>TUY008@004[2_/9J`[k2F]IJeU[FF]IJ@LA0001J@U[FF]IJe[kF_]Jne[kF_
P7YoFGmI:P081`001PLI0003WHBM00>42P02UP0RZY2MQ9fF/9J`U[2F/9J`U[2F/9JZQ9f4ZIJ`UZVF
/9J`L@@001egLK2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/00L0006RZ>4[iIg8@004:nF[iJ_UZnF
/9J`U[2FMg4A000GLI2F/9J`UZnF[iJ_UZnFNeUjFGYINeT0:@081`001PLH0005WHB4Q9d00X@00f/0
0008000S/9J_TIf4WIJ`U[2F/9J`U[2F/9J`UYf4WHB_QJ>F/9J`UWL00P007g:@U[2F/9J`U[2F/9J`
U[2F/9J`U[2F/9J`U[2F/9H06P002:B:WHBMU[1a8@003YJ`U[2F/9JeU[FF]G6@50005W:eU[FF]Jne
[kF_]GZ0FGmIOeVJN[<Y00P700061aL000N40000Q6^M00:42P00AI6FU[2AWHBMRk2F/9J`U[2F[i6`
U[2FXhBMQ:Z@/9J`U[2FLW6`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`UP0G000;RZ>4WHBM
Q:nFMg4080003:nF[iJ_UZnF/9IgLAH000]gU[2F[iJ_UZnFN`02NP09FGYINWbEU[2F02T020L000H7
7002Q003WHBM00X004BF/9J`QIf4ZIJ`U[2F/9J`RZBA/9J_Q9f4[iJ`U[2F/9IgL[2F/9J`U[2F/9J`
U[2F/9J`U[2F/9J`U[2F/9J`U[2F/1D000fTRYf4WHBMQ:BF/9J@024000V`U[2F/9JeLI006@005Y2F
]IJe[kF_P5UoFGmIVWbeU[FF]IHX00P700061a000h@:0005JhBMQ6/02`02UP0LZH^MQ:>F/9J`U[2F
XhB`T:>@ZYJ`U[2F/9IgL@<0029a/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J_4@004J>:
WHBMQ9f4WHBMQ;2F/9Ig024000F`U[1aM`0M000FMiJ_U[2F/9IjFIEl/9J_UZnF[iIjFBL020L000H7
3`000hBM000<0005JhBMQ9d02P0079J`UZV4WHB`U[2F/9J`RYf4/96`U[2F/9J`LGL5000RT9J`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`UQ0001:TQ9f4WHBMQ9f4WHB_U[2F/74P000469J@LAl0
01Ib/9JeU[FF]GbJU[FF/9J`U[2FNeUj9`081`001PL>00:400=[0000300019f4WH@;000ITK2@WHBM
Q:ZF/9J`TIf4XiJ`U[2F/9IgL@08000R/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J_UPd0
01F:XhBMQ9f4WHZMQ9f4WHB`UZnFMg4080000gL0000O000FMiJ_U[2F/9J`U[2F/9J_UZmlUEUjFBL0
20L000H7300019f4Q8@2J`/00h@019f4WH@:000G/9J`UZ>4WI2`U[24WI6TU[2F/9J`UWL02P008YJ`
U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[0;000GY8ZMQ9f4WHBYT:B4WHBMQ:VF/9J`UWL0
7P000g:@UP0R000FT9JeU[FF]IJeU[FF/9JEO7YINUUjDRH020L000H72P000hBM0003009[00>406/0
1`02Q00606/0WHBM2`005K2F/8BMQ9f4/8BZT;2F/9J`U[1aM`0;000RMiJ`U[2F/9J`U[2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J`T@T001V:XhBMQ9f4WHBYU[2FWHBMQ9fF/9J`UZma00l000B4WF^42@001I1a
/9J_028001Ea[iJ`U[2F/9J`U[1lUEUjFGYINU809`081`001PL90004Q9e[J`H000A[Q8A[0`02Q004
J`1[009[00>MQ6/02P004iJ`U[2@WHBMQ:nF/9J`U[2FMg803@008g:`U[2F/9J`U[2F/9J`U[2F/9J`
U[2F/9J`U[2F/9J`U[2F00L001^:WHBMQ9f4WI2YU[2FZHBMQ9f4ZIJ`U[2F/7402P0016/0WH@2J`05
WHBMQ6/0100027:@U[2F/9J`8P004W:eU[FF]IJeU[ElNUUjFGYINRT020L000H71@001K1aWHBM00:4
2@02J`05Q02406/01@02Q003W@0000T00YH03k2FY8BMT:ZF/9J`U[2FLP0?000SMiJ`U[2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J`UZl01@007JB:WHBMQ9f4WH^_U[2F/8BMQ9f4WIJ`UZnF/9Ig00X0
00B4WF][0`0016^4WH@30009LK2F/9J`U[2A0280015gU[2F/9J`UYElNUUjFGYBNP0Z00P700061`@0
00N`Q9f4WHBM009[2`000f/0J`070003WH@000X0012F/9JSQ:n@/9J`U[2F/79g4@008gNF/9J`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`0080022FZHBMQ9f4WHBYT;2F/9J`UZn4WHBMQ;2F/9J`
U[1aM`P000>4WH@01P003if4WHB@L[2F/9J`U[2F/00R000?L[FF]IJeU[FFOeUjFGYB00D000=BC@00
90081`001PL2000<UYN4ZXBYQ9f4W@1[3002J`H000>4WH@0100016^4Q8@200:F00bST:ZF/9J`UYNF
LW4C000VLK2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/96_TK02UP0MQ9f4WH^YUZnF/9J`
U[2:WHBMQ9fF/9J_UZnF[g401`001F^MQ9e[00D0011[WHB:UZnF/9J`U[2F/9J_8P0039J_U[2F[iH0
FGYBNPD000AjFGYI9@081`001PL03K2FY8ZMTK2;/8BMQ9d00X@;0003J`2400D000=[Q8@01P000f^M
Q002Q00;WI2`U[2F/9J`LWL05@00AGMb/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9H0U[2F
/9JTQ9f4/9J`U[2F/9J`UYf4WHBMQ;2F/9J`U[2FM`090003Jh@000@00164WHBMT;2F/9J`U[2F/9J`
UP0R0005]IJeU[000`000gmB00050005NUUjFGX09@081`001PL0396GTINAUi6GTINARPB43`0018@0
06/2Q003J`0000D000A[Q02A0YH01INFUg5b01T000MgU[2F/9J`00fF00N`U[2F/9J`00:F1P008YJ_
UZnFZHBMQ:>A/9J`UZnF[hBMQ9f4WIJ_U[2F/9J`UWL<000EJif4WHBMUZnF/9J_UZnF[iJ_UZnF0240
00=aUP002@002Tee001INUUj04dT00P700061`0C/9J`U[2F/9J`U[2FRXBMQ9d0J`0<00:400@0JhBM
2@0026/009J`UWMb70006i2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/008000RU[2F/9J`T9f4ZIJ`
U[2F/9JMQ9f4WHZ`U[2F/9J`U[2FT0T00V/05If4WHBMT;2F/9J`U[2F/9J`U[2F[`0Q0003Mg4000T0
00=oF@000P000eUoDP0V00P700061`0?TINAUi6GTINAUi6GTHYa00>400<0J`0030000hA[Q002Q0/0
00>FLG807P001W6`U[2F/0fF00>`U[000YH00i40000800BF01j`UYJ:WHBMRk2F[iJ_Q9f4WHBZUZnF
/9J`U[2F[g49000FQ9f4WHBSRZnF/9J_UZnF[iJ_UZnF[`:F7`000jV4M`09000:NUUjFDd0NUUjDRD0
20L000H701BYRk2F/9J`U[2F/9J`U[24WHBMJ`d000E[Q8BMQ00;0003LWL001h001N@L[2F/9J`U[2F
/9J`U[2F/9J`U[2F/00=000Q/9J`U[2F/9JTQ:>@/9J`UYf4WHBMQ;2F/9J`U[2F/9Ig00L001Z:Q9f4
WHBSU[2F/9J`U[2F/9J`U[2F/9J`UQ`000I>T:V4WG890004DWmIM@<000EoFGX0C@0S00P700061`>4
00f@TINAUi6GTINAUi5b00B400=[000030000f^4J`090003Mg6F024000EgTK2F/00=UP03/94000h0
00fF[iJ`U[2F/92MQ9f:00:F01>ZQ9f4WHBZUZnF[iJ`UZnF/9Ig00@001aaMhBMQ9f4[iJ_UZnF[iJ_
UZnF[iJ_UZnF[iJ_6`002:V;WHBMQ9ea1`0014ejFGX40006FGYB04ej8P081`001PL04if4WHZMT;2F
/9J`U[2F/76MQ9d00X@?0003WF/000L000B@U[2F8@004g:`U[2F/9J`U[2F/9J`U[2F/9H040008[2F
/9J`U[2F/9JYQ9fF/9JMQ9f4XiJ`U[2F/9J`U[2F/743000ML[2:WHBMQ9fF/9J`U[2F/9J`U[2F/9J`
U[2F/9H060002g6@ZHBMQ9f4WHAg00L000AjFGmI10001WmIOdejFB8020L000H718@03H^FRY2;UY6G
TINAUg401H@F0004LGNAU`:F00>A00007`00176`TK0<UQ<00YH03k2F/9J`U[2FXhBSRiJ4W@03Q00]
ZIJFU[2F[iJ`UZnF/01gLJnFWHBMQ9f4[iJ_UZnF[iJ_UZnF[iJ_UZnF[iJ_01P000b;WHBMQ9f4WHBM
LHX70003NUUj00D000=jFGX00UTR00P700061`0GWHBMQ9fF/96MRZBAUiJ`UWN4WHBMQ6/04`002I1b
/9J`U[2F/00Q000?T9J`U[2F/9J`U[2F/9J`01@004:F/9J`U[2F/9J`U[24Xi2MQ9f4XiJ`U[2F/9J`
U[2F/9J@U[2F/92MQ9f4WI2`U[2F/9J`U[2F/9J`U[2F/9J`U[0E000>[i2MQ9f4WHBSU[2@ZG460006
CEUoFGm=0P001deIOeUjFDd08@081`001PL00h^4Q002Q00>Ui6GTHZAQ8ZFTINALP03Q18000MaMi6G
TINA00:F00>GUYH080001G5gU[2A00NF00>A00005@003IJ`U[2F/9J`UZnFQ8X018@0<9f;T9J_UZnF
/9J_U[1aMg6_UZnFZXBMQ9f4XiJ_UZnF[iJ_UZnF[iJ_UZnF[iJ_TA@000n_UZn:WHBMQ:nF[iJ_UWL0
1`002GYINUUjFDeINP02F@03DTd0024020L000H701R`Q9f4WI2`U[2F/8ZMRZBF/75[Q9f4WF/>000>
T7:`U[2F/9J`U[2F/9HQ000:L[2F/9J`U[2F/1T001j`U[2F/9J`U[2F/9J_Q9f4WHBMU[2F/9J`U[2F
Mg42000RMiJ`U[2F/8ZMQ9f4WI2`U[2F/9J`U[2F/9J`U[2F/9J`UQ4001:F/9J`U[24WHBMQ:VF/9J`
UWL7000:OeUoFGmIOeUoFB@020L000H700>AUh@00h@03X^GTINAUY6FQ9NALW5[0h@00f/0000;000:
Mi6GTINAUi6GT@>F00>AUP0080000gNAU`04UP03T@0001T000ZA/9J_U[2F[iJ`0iH00hZ;Q002Q003
RiJF00:F00F_UZmaM`040003LJnF00:F01^_UY64WHBMQ:>F[iJ_UZnF[iJ_UZnF[iJ_UZl00YH?000D
ZIJ_UZnF[iJ_Rif4WHBSRjnF[g470009CGYINUUjFGYI00<000==00007`081`001PL05k2F/92MQ;2A
/9J`U[2F/8ZMTGL0WHBM00:40`000f/00003000CMg:`U[2F/9J`U[2F/9J`U[2F/00Q0004MiJ`UQl0
01NF/9J`U[2F/9J`U[2F/92`Q:VF/9J`L@09000RLK2F/9J`U[2F/92MQ9f4ZIJ`U[2F/9J`U[2F/9J`
U[2F[`T001Y>T:V4WHBYU[2F/9J`U[2@ZHBMQ9f:WIJ`L@T000==00000`001U9jFGmINQl020L000H7
00BAUi6:0h@02INAUi6GTINAUP02RP03Q8X000F400<0J`000`005W5gTINFUiJGUYNFUiJGUYNAUi6G
TIHP0003LGNA020029H03ZnF/9J`U[2FZIJ`U[1a2P008WMa/9J`UZnF/9J_Rj>4WHBMQ:VF[iJ`UZnF
/9J_UZnF[iH8000KZH^MQ9f4WHB_UZnF[iJ_UZnFWHBMQ9f4ZIIg00d00Td01WYINUUjFAl020L000H7
01Z`U[2FXhBTU[2A/9J`U[2F/96TQ6^4WHBMQ0@001QgL[2F/9J`U[2F/9J`U[2F/9J`U[2F/94Q0003
L@0001l001FF/9J`U[2F/9J`U[2F/9J`U[2FMg4030008Y2F/9J`U[2F/9J`U[2@WHBMQ:>@/9J`RZn4
XhB`U[2F/9H6000MZHBMQ9f4WHBTU[2F/9J`U[2F/8BSQ9f4WHZMUWL030002EUjFGYINU9=DP0N00P7
00061`04TINAU`>400b@TINAUi6GTINAUi43Q004JhB4Q08001]aMi6GTINAUi6GUYNFUiJGUYNFUi6G
TINAUY4070001F/0UY5b02402YH02K2F/9J`U[1aM`0>000PLK2F[iJ`UZnF[iJ_UZn;XiJSQ9f;WHBM
Q9f4WH^_UZl2UPD001jYQ9f4WHBSQ9f;[iJ_UZnF[iJ_UZn;WHBMQ9f4[g4;0004DWYINP9=0`000dd0
000L00P700061`0HZIJ`UYN4WH^`U[2A/9J`U[2F/9J;Q9f40P007GMb/9J`U[2F/9J`U[2F/9J`U[2F
/9J`U[2F/9J`01X000Q[Q8Z`U[2FMb40012`U[2F/9J`U[2F/9J`UWMa4@008W:`U[2F/9J`U[2F/9J`
U[2F/8BMQ9f4WHBMQ9fF/9J`UZl2000QUZn4WHBMQ:>@WHBMU[2F/9J`U[2F/9JYQ9f4WHBMU[1a00X0
00AjFGYB90081`001PL01I6:TINA00>400jAUi6GTINAUi6GTIMaJ`@0016AUi6GTINAUi6GUYNAUiJG
T@02UP0:UiJGTINAUi6GTAP000>4Jh@00X@01H^FUYIa02402IH01K2F/75g01<001mgU[2F[iJ`UZnF
[iJ_UZnF[hBYRZ>4WHBSQ:V;[iJ_00:F02BA[iJ_Rj>4WHB_UZV4WHZ_UZnF[iJ_UZnF[iJZQ9f4WHBS
UWL80005CGYINUT01`000g40000K00P700061`0F/9JYRk24WHBSTK2F/96`U[2F/9J`L@80021gL[2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`TAL000Z4WHBMQ:VF/9J`8P002iJ`U[2F/9J`UWMa01D0
021aT9J`U[2F/9J`U[2F/9J`T9fF/9J`UZVF[iJ`U[2F[`8002>F/9J`Q9f4/9J`Q9f4ZIJ`U[2F/9J`
U[2F/92MQ9f4WHZ`L@080005CEUjFDd01@000i1aN`0M00P700061`05TIN@RY4018@078^AUi6GTINA
Ui6GTG9aLY6GTINAUi6GTINAUi42UP06Ui6FUYNA0YH02INAUi6GTINAUP0E0003Q6^400B400>:UYH0
0YHR00RF00=g00005P00276_U[2F[iJ_0iH04ZnFUY2MQ:nF[iJ_UZnF[iJ_UPD002>F[iJSQ:VF[iJM
Q9f4[iJ_UZnF[iJ_UZnFZXBMQ9f4ZIJ_L@070005DWYINTd00`001W5gUZmINQ`020L000H703V`U[2F
Y96MQ9f;/96`TK2F/9J`UWMbMiJ`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`TK004P003V^4
Q9f4WHBMQ9f4/9J`8P0029J`U[2F/9Ig6@00792F/9J`U[2F/9J`U[2F[hBSR[2F/9J`U[2F/9H6000S
/9J`UZBF/9JTQ9f4WIJ`U[2F/9J`U[2F/9JMQ9f4WHB`UWL01`0017YINUT20005T76`UYD00WX00eT0
000J00P700061`07TINAUi64R`03Q00:T9NAUi6GTINALP<0011gTINAUi6GTINAUi6GTINA0YH049NA
UYJGTINAUi6GTINAUi4@0003Q6^400F400F:Q8Z4RP03UP03M`00020019H00g5g000J0005MiJ_UZl0
0iH01JnFUYJ_00:F00R@XhZYRjnF[`:F00>A00001P008jnF[i6_UZn;WHBMQ:nF[iJ_UZnF[iJ_UZn4
WHBMQ:>F[iIg00L000ijFGX0MiJ_O9EINUUjFA/020L000H7016`TK2F/9J`T9f4WHB`U[1aM`07000Q
MiJ`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`00`001I[Q8BMQ9f4WHBMU[2F/9J`TJnF/9Ig
8@000i1a000N000FLWNF/9J`U[2F/9J`U[2F/9J`U[2F[`/002A>U[2F/9J`T9f4WHBMT;2F/9J`U[2F
/9J`UZ>4WHBMT;2F/748000=NeUoFGYINUUjFG]j/00K00P700061`06TINAUi6F0Y43Q005UY6FTG80
2@0087NAUY6FTIJAUY6FTINAUi6GTINAUi6FTIJAUY6FTIJA2`000f^4Q006Q00>T8^GTIJAUY6FTINA
UW4O0003LGNA024000=aUYH00YH02JnF[iJ_UZnF[`04UP03T@0000d000>F[iH00iH6Q00>T:nF[iJ_
UZnF[iJ_TId2Q003T:nF00:F00>_00001000175gUW/6F@05NW^F[iH06`081`001PL03k2F/96`U[2F
WHB`U[1bM`0;000P/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/949000IJif4WHBMQ9f@ZIJ`
U[2F[i2`U[2F/9IgLP0M0004L[2F/240015gU[2F/9J`U[2F/9J`U[2F[`0A000R/9J`U[2@WHBMQ9f4
[iJ`U[2F/9J`U[2FY8BMQ;2F/9J`UP@0016@L[2FUGajFGYINgZ`U[2FNP0K00P700061`06TINAUi6F
0Y001Y6FTIIaLP`0021aUY6FTIJAUY6FTINAUi6GTINAUi6FTIJAUY6FTIJAUPL000>4Jh@01H@0192;
UY43UP09TINAQ92FTIJA00:F00=b00006@0017MaUY43UR4000=gUYH00iH01JnFUYJ_00:F00>A0000
4P001:nFUYH2RPF400b@[iJ_UZnF[iJ_UZl3Q005RYJFUZl00YH02W40LGNF[g^EFGX2F@06NW^F[iJ_
0eT00gX0000H00P700061`0</96`U[2FUiJ`UWMb3P0087NF/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F
/9J`U[2A1P02Q00JWHBMQ9f@ZIJ`U[2F/9J`UZV4[iJ`U[2F/74J0008MiJ`U[2F/9HQ000=L[2F/9J`
U[2F/9J`UP0D000/LIJ`U[2FWHBMQ9f4WIJ`U[2F/9J`U[2FWHB_U[2F/9J`UWMa/9JEO7YINUT2NP05
/9J`UYD00WX00eUjF@0J00P700061`0:TINAUi6FTIJALQ40021bTIJAUY6FTIJAUi6GTINAUi6GTIJA
UY6FTIJAUY6GT@@00XX4Q00<T8^FTIJAUY6FTIJA18@2T@03UY6F00:F00=a00002`02J`P000=aMiH0
0iH01I6FTIJA020000=gLIH01IH01:nFUY4F0003[iJF00:F1h@03JnF[iJ_UZnF[iJ_Q8X01iH037Ma
Mi6EFGYINUUjF@:F00>_UYH01ET00iH0000H00P700061`09/96`TK2F/79g01<001mgU[2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J`0080022A/9JYQ9f4XiJ`U[2F/9J`U[24WHBMQ9f@/9J`U[2FM`/0
0X@00ie[0005000=Mg:`U[2F/9J`U[2F/00Q0009MiJ`U[2F/9J`01T001n`U[2F/8BMQ9f4WHZ`U[2F
/9J`U[24[iJ`U[2F/9Ig008001IaUUUjFGYINgZ`U[2FUEUjFGYIUIJe6@081`001PL01i6GTINALW40
5P0097:AUY6FTIJAUi6GTINAUi6GTIJAUY6FTIJAUY6FT@2AUY6GR`>400^AUY6FTIJAUY6FT@06Q0:A
00NFTIJAUW5b00X000A[Q8B40V/20003LGNF00NF00JAUY6FTIHP0003LIJF00>F00>A00006@05UP03
RXB400>400^ST:nF[iJ_UZV:Z@06UP03LGL0008000UgUIIINUUkUZl019H5F@04U9J_UQT020L000H7
00H0TK2A/78H000LMiJ`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`T@<0026A/9J`Q9f:/9J`U[2F/9J`
Q9f4WHBMQ9f4[iJ`U[2F/7802`005f^MQ9e[Mg:`U[2F/9J`U[2F/9J`U[2F024000Ab/9J_7@007;2F
/9J`T:V4WHBMQ9f4ZIJ`RZV4XiJ`U[2FMg46000FT9EkN[2F/9J`U[1kNUUjFK2F/9J`UQP020L000H7
0P000g5b000I000KLIJAUY6FTINAUi6FTINAUY6FTIJAUY6FTIJA00@000FFTIJARP02Q0>A00BFTIJA
28@01iJ;RY2FTHX00YH00i5b000:0003JhB400:401:AUX^FRYJ4T8BFTIJAUY6FTIHQ0003Mi4001h0
1iH03i24Q8BMQ9f4XhBYQ:nF[`02UP03LG8000L000QaUYJ_UZnF[`>F00EkFEUIU@03UP03[iJ_01P0
20L000H77`0067:`U[2F/9J`U[2F/9J`U[2F/9J`U[2FU`L0026`U[2FXhZ`Q:V@[hBMQ9f4WHBMU[2F
/92YU[24/9J`UWL02`02Q00EWHBSRYf4WHBMQ9f4WI2`U[2F/9J`024000=g00007@006;2F/9J`U[2F
/9JSQ9f4WHBYU[2F/9IgL@X001J@LK2F/9J`U[2FUEUjFIJE/9J`U[2F60081`001PLQ000CMi6GTIJA
UY6FTIJAUY6GTINAUP0<0008TIJAUY6GTHX7Q003UY6F00NF00N4RYJAUg5g00P000AaMiJ:18@01X^F
TIJAT0J400J@UYJFTIHM00>F00=a00007P002JnF[iJ_UZnF[`09UP03Mg4000d000EgUYJF[`04UP0=
NgYINYJ_UZnF[iJ_UP0G00P700061b80015gTK2F/9J`U[2F/9J`TK2A/00>000OTK2F/9J`UZ>4WHBM
U[2F/9J`U[2F/92SQ:VF/9J`L@06000LMg:`UZn4WI2MQ:>4[iJ`UZV4WHBMQ:VF/9J`UQX000M>U[2F
/9J`0240016`U[2F/9J`U[2F/9J`U[1aM`0A000FMiJ`U[2F/9IkFGYIUIJ`U[2F/9J`UQH020L000H7
8P003g6GTIJAUY6FTIJAUY6FT@0@0009UY6FTINAUY6:00:400>AUi400iH01I6FUY6;00B400F@TIJA
M`040003LGNF00>F0X@00hZAQ003Q006RiJAUY6F1H@00i2FUP02UQT029HQ0008UZnF[iJ_UZl6UP03
L@00014000=aUYH019H02JmINUUjNjnF[`03UP03[iJF01H020L000H78`003G:`U[2F/9J`U[2F/940
4P007iNF/9J`U[24Xi2`U[2F/9J`UZn4WHBMQ9f;/9J`UWL00P00876`U[2F[hBMQ:>FXhBMQ:>4/9J`
UZB4WHBMT;2F/9J`5@0031RF/9J`U[2F/9J`LB0000hHU[2F/9J`U[2F/9IgLA@000MgL[2F/9J`00:F
00eINUVEU[2F/9J`U[2F01H020L000H790002i6FTIJAUY6FTIJA01@0016FTIJAUi6FQ92AUY6FTIJ@
T@06Q009T96AUY6FLGMa00FF18@00hZFRP05Q005RiJAUY4018@00i2AUP02UP03T@0001<02iH00gL0
000O00:F00F_UYJF[`02UP03LGL001H000=gUYH00iH02JnENUUjFG^F[`04UP03O9EI01D020L000H7
90002GNF/9J`U[2AU`0G000o/9J`U[2F/8^`U[2F/92MQ9f4WHBMT;2F/9J`UWMa/9J`U[2FWHBMQ:VF
/8^MQ9f4Y8Z`U[24WHBMQ;2F/9J`01<000n`U[2F/9J`U[2F/9J`LGL07`002TjF/9J`U[2F/74I0006
LK2F/9J`0YH03YEjFGYIUIJ`U[1kUEUj5@081`001PLU0005LY6FTIH06P002Y6GTIJAUi6FTIL2T@03
T8B400F400^@UY6FTIIaM`1gL@05UPF40Y400iJ4Q004Q004RiJARP>400>@UYH00iH?009[49H00g40
000O00>F00B_UWMa6`000g6FUP02UP09[iJFFGYINUUj009I0WX3F@03NP00018020L000H79P000gNA
000K000J/9J`U[24Y96`UZV4WHBMQ9f4ZY2`U[2FMg44000QMiJ`U[2:WHBMQ9f@/9J`RYf4WHBTR[24
WHBTU[2F/9J`00h001A[WHZ`U[2F/9J`U[2F/9J`U[2FMb4000F`U[1aM`0M000FMiJ`U[2F/9IkFGYI
NUUjFGYINUV`UQ@020L000H79@000iJALP0M000:UY6GTHZ4RY6@R`J400RAT9JAUY5bL@H000=gUYH0
0YH5Q005T96FTI001X@01H^4Q8B:00NF2`02J`:44YH00g80000O0003Mg4001l000mgUYJF[iJ_UYEI
NUUjFGX00eT3UQ@020L000H78`001Y6`U[2FLQd001F`U[2AXhBMQ9f4WHBMT:ZF/9J`LGL02@008;2F
/9J`Q9f4WHB`U[2A/8BMQ9f4Y8BMQ;2F/9J`U[2A2P02Q00FWHBYU[2F/9J`U[2F/9J`U[2F/9J`LB00
00=g000080004g:`U[2F/9J`UYEIUEUjFGZF/9H05P081`001PLN000=TIJAUY6FTIJAUY6FL@0M0005
TIJAUX/00h@02I2ATIJAUY5bL@0<0007Mi6FTIJAT005Q0:@00>FTH@018@02i6FTIJAUY6FTIJA00H0
00=[Q8@01H@019JAUY4=UP03TIIa01/000=aMiH00iHR0003LIJF00^F00>A00005`081`001PLK000@
TINF/9J`U[2A/96`TK2AMah000jA/9J`T9f@/9J`U[2FM`h0021b/9J`U[2A/8ZMQ9f4WHBMT9f4WHBM
TK2F/9J`U[2F/0D001fYRZn@WHBMQ9fF/9J`U[2F/9J`U[2F/9J`U[2FM`0J0007LK2F/9J`UP0R000=
T9J`U[2F/9J`U[2F[`0K00P700061aT001>FTIJAUY6FTEUBFE9kMiJAUY5b01h000bAUY6FTIJAUY6F
LG8@0008LIJAUY6FTIH8Q003RhZ400B400VFTIJAUY6FTIH00P06UPJ400>AUY403IH0196FTGL;0003
J`1[00T000=gLIH01YH00i40000P0003MiJF00JF00>AUP006`081`001PLG000FTINA/9J`UWYINUUj
FGYBNU:ETK2AMad000VGU[2F/9J`LGL05000@G:`U[2F/9J`UZV@Xi2MT:V@WHBMQ9f4ZIJ`U[2F/96G
U[2F/9J`UYf4WHBMQ;2F/9J`U[2F/9J`U[2F/9J`U[1a00P000I[Q9f4WH@7000:LI2F/9J`U[2F/0:F
8P002G6`U[2F/9J_UP0N00P700061a@001ZAUY6FTIJAFE9IDUUBFE9IDUUBFE9jTIJALQd000FFTIJA
LP0H000ALIJAUY6FTIJAUY6FTIJAUY401H@01hZFTIJAUY400P03UP04TIJFUPN400>FTIH039H01I6F
TGMa00P01h@30003Mg6F00ZF00>A000080000gMaUP02UP03TIH001l020L000H74@007Y6GTK2A/96`
TIEINUUjO9FF/9J`TK1BNU9jT;2AMad000>GLGL06P006W:`U[2F/9J`U[2F/9J`U[2@WHBMQ9f4/9J_
10008YJGU[2F/9JSQ9f4WHBMU[2F/9J`U[2F/9J`U[2F/9J`UWL700:400JMQ9f4WH@2000@MiJ`U[2F
/9J`U[2F/9J`UR8000B@U[2F8P081`001PLB000MTIJAUY6FTIIBFE9ITIJAUY6FTIJAUWMIDUUgUY40
6`000g5bT@0L000JMi6FTIJAUY6FTIJAUY6FTIJARXB@RYJAUY450008TIJFUY6FTHX5Q003UY6F00bF
00FAUY6FL@0900J400=[Mi400iH00i6FUP09UR8000=aUY408`081`001PLC0009TK2A/96`TGYI009j
01:`U[2F/9J`TK2A/7IjDWZA/74G0007LWNF/9J`T@0L000GMiJ`U[2F/9J`U[2F/9J`U[2F/9J`TIH0
20008IJ`U[2F/8BMQ9f4WHB`U[2F/9J`U[2F/9J`T;2F/9J`L@0800:401FMQ9f4XiJ`U[2F/9J`U[2F
/9J`U[000YHQ0003CP0002<020L000H750007I6FTIJAUU9IDYJAUY6FTIJAUY6FTIIBFE9kTIIa01<0
00]aLW6FTIJAUY6FT@0L000DMi6FTIJAUY6FTIJAUY6FTIJAUY4;0006TIJFUY6F1X@00i6FUP0:UP07
T96FTIJAM`0;0003RXB400>400>:TIH00YH00i6FUP07UP03T@0001l000>F00008`081`001PLD000N
/96`TK2ANUUjO9NF/9JGU[2A/96`TGYBNU:`U[1a2@000hBMQ005000<L[2F/9J`U[2F/9J`7@003gNF
/9J`U[2F/9J`U[2F/002UP03T@0000/001JF/9J`U[24WHBMQ9fF/9J`U[2F/9J`0YH028BYU[2F/75g
20005W9gUZB4WHBMQ:V@ZI2`U[2F/9J`U[02UQd000Ia0;2F/74T00P700061aD001jFTIJAUU9IDYJA
UY6FTIJAUY6FTIEBFE9kTIJAUW4:00:4015[LGNAUY6FTIJAUY6FTIJAUP0M000?Mi6FTIJAUY6FTIJA
UY6F010000JAUY6FTIH6Q0^F00R4RYJAUY6FL@H000=gLIH019H018Z4Q8X4Q003RY6A00VF7005UP03
LP00028020L000H75`007i6GTK1eNU9jT;2A/96GTIJ@NU9jDYFAUi6`TK1BME801@00179gUX/3Q00@
Uh^MQ9f4WHBMT;2A/9JGTA`000MgU[2F/9J`00:F00>A00004`007i6`U[2F/8BMQ9f4WIJ`U[2F/9J`
UZn4WI6`U[2F/7400P007GNF/9J`U[2F/9J`T9f4/92_T9f4WI2MTK2F/9J`00:F5`002k2F/9J`U[2F
/9Ig02<020L000H760007i6FTIIgNU9jDWafUWMIDUUBFE:ETIJAUY5kDWYBNU800P001g:AUY6FTIH0
1H@00i24Q003Q007RX^FTIJAUP0M0006Mi6FTIJA6@06UPF400RFTIJAUY6FT@>400>@UYH019H00g80
M`02UP05TIJFUY4019H01I24Q8BA00:F1H@2RPFF00>A00004`0<UP03L@00024020L000H76@009i6`
TK2ANU9jDWYBNU9jDWYBNi6GTIMeNU9jDWYBMWF`UYNAUiJGT@04Q00:UY6`TIf4WHBMQ0:F00>GUYH0
7@000gNAUP0K00:F03^`U[2F/8BMQ9f4/9J`U[24WHBMT;2F/9J`U[1aMiJ`U[2F/9J`U[2F/9JSQ9f@
/9J`UYf4WHBMQ:VA/002UQ<0010H0;2F/9J_U[2F[iJ`UWMa8P081`001PLJ000UTIJAUY6FDWYBNU9I
DUVAUY6FDW]BNU9jDWYBMW6FTIJAUY6FR`02Q009UY24Q8ZAUY6@00F400FFTIJAUP0L0003LGL001`0
1iH5Q003UY6F00>400>AT9H01IH01W9a076FT@>F00FAUYJFT@03UPB400B;UYJF1h@018ZFUYHC00nF
00=g000080081`001PLJ000I/96GTK2A/96`TIJAUi6GTG]eNU9jDWYBMP040007LK2AUiJGT@03Q00B
TINFWHBMRk2A/8BMQ9f4UiJG0YHI0004/96`LAh00YH06;2F/9J`Q9f4WHBMQ9f4[i2`U[2F/9IgL@<0
022F/9J`U[2F/9J`U[2F/8BMQ:VF/9J`T:>4WHBTU[2F/0l001A>0;2F/9J`UZnF/9J_U[2F[iJ`LB80
20L000H76`006YJAUY6FTIJAUY6FTG]BFE9IDUUBNTejDWYB0`001W6FTIJAUP>400B:TIJ@18@01I2A
UY2:00>400F@TIJAUP0E0008T8^@Q9JAUW4N00NF1X@01I2ATIJA00>F00=aLP00100017NAUY43UP03
TIJF00>F00JAQ8B4Ri44UP04RXB:T0JF3P0DUP03L@0001l020L000H770004k2AUi6`TINAUi6FDWYB
NU9fMGH01000159jDTd20006TINAUiJ:0h@03HZA/96YQ9f4Y96`TJ@018@018^FUYHB00:@0h@01Yf4
/96`LAd001FF/9J`U[2FWHBMQ:n@/9J`U[2F/740200087NF/9J`U[2F/9J`UZV4WHBMU[2F/9J`U[2F
/9J`U[2A2P006AP0/9J`U[2F/9J_U[2F[iJ`UZnF/9J`UWL08@081`001PLM000?UY6FTIJAUY6FTIIB
FE9I00L000A=DWYB0P001g:AUY6FTHX018@0196FTI44Q004RYJAUPF400>@RiH03P000h^@Q008Q004
UY6FLAd000>AUYH00iH038BAT9JAUY6FTIJALP/000QgTIJAUY6FT@BF00B:Q8B43iH00i40000701RF
00=aM`007`081`001PLM000@TK2AUi6`TINAUg5g07YBC@L000=jDWX00`001gNAUi6GTI000h@03J>A
/96`Rif4WHBYUYL01IH00iNFUP090003602@00V400RMQ9f4/96`LAd0012FUiJ`T;2F/9J`U[2F/75g
30007W5gU[2F/9J`U[2F/92MQ9fF/9J`U[2F/9J`U[2F/0:F20006iJ`U[2F/9J`U[2F/9J_U[2F[iJ_
UZnF/9J`L@0P00P700061ah000RAUY6FTIIaLP<00e860006DWYBNU9=0`001g:AUY6FTIH018@01H^F
TIJ@00B400ZFTIJAUY6FTIJA2@000h^@Q006Q008RXBFTIJARX@2T@03UW4001`000bAUYJFTIJAUY6F
LGL@0007LIJAUY6FT@03UP03Q8Z400:400B:UYJA0iH00i6FUP05UP03T@0000D06iH00gL0000N00P7
00061b0000=bL@001`002deBNU9jDWYBNUUj00L000QgTK2A/96`T0>400JAUiJGUY43Q009Xi2`TK2A
/96`00:F00H0UiJGU[02RP0FQ9f4XhB`TK2F/9J`UYNF/9JGU[2FMad000F`U[1aM`0G0011MiJ`U[2@
WHBMQ9f4/9J`U[2F/9J`U[2F/9J`U[2ACP2`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`UWL0
7`081`001PL[0007CE9IDUUBF@02DPT000MbTIJAUY6F00>400N@TIJAUY6F00F4016FTIJAUY6F09JA
UY6FTIJAUP05Q00@RY6FTIJAUY6FTIJAUY6FTAh000=b00006@000gNFUP02UP03TIJA00:A00BFTIJA
3IH02I6F09JAUY6FT@0=UP03TIJF00^F00=aLP004P000gMa000800P700061c0000==00000`0017E=
NTd4000JMi6`TK2ARXBAUYNFUiJGUYN4WHBMQ:VAUY4400:F01bGUYNFUiJ`RZ>4WHBYT;2FUiJ`UYNF
/9JGU[1a6P0017Ma/94L0010MiJ`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/96F04jF/9J`U[2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J_LA4000AbMiJ_2@081`001PLc0007FE9IDUUBF@02DP<0015bTIJA
UY6@TIJAUY6FTIJAUP03Q003RX@000L000bFTIJAUY6FTIJAUY04Q00=RY6FTIJAUY6FTIJALP0G0007
Mg6FTIJAUP0M000<MiJFTIJAUY6FTIJA2iH00i4000030005TIJAUY402iH01I6FUYJA00bF00=b0000
3@0017MaUYH2Q0T020L000H7<P002gYBNU9jDTeBNU9j00<0019gTK2AUiJGUYNFUiJGUYNAUY4<00:F
00fGUYNF/96`TK2A/9JT00:400b@UiJGUYNF/96GUWL30003J`2M00:43000275gU[2F/9J`0YHM000C
MiJ`U[2F/9J`U[2F/9J`U[2F/002UP03T@0000@002=>U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F
/9J_U[2FM`0=0008MiJ_Q9fFZH@800P700061c4000M=DUUBFE9I008000A=FE9I0`003g:AUY6FTIJA
UY6FTIJAUP0?000>UY6FTIJAUY6FTIJAUY42Q00?RY6FTIJAUX^4T9JALP1[00F400=[00001`0038Z4
RXB@RYJAUY6FTAh000YgTIJAUY6FTIJA1iH00i4000090003TIJA00fF00FAUYJFT@03UP03TIJF00FF
00=a00002@000g5gUP06UP03Q00000H020L000H7<@00259jDWD0ME9j0`000gYBM@030009Mi6GTINF
Ui6G00:F00>A00004002UP0KUiJ`TK2A/96`TK2:Q92GUYNFUiJGRXBF/9Ig00<000>MQ9d00X@5000@
WHB4Q9f4WHBMQ:>F/9J`UQh000n`U[2F/9J`U[2F/9J`TIH03@008TjF/9J`U[2F/9J`U[2F/9J`U[2F
/9J`U[2F[iJ`UZnF/749000<LK2F[iJ_UZnF[i2M1`081`001PLa009B00AI00000U830004DUUBC@<0
00EbTIJAUP03TAH001FFTIJAUY6FTIJAUY6@Q8ZAUY6FTIH00Y401XBFTIJALP@000I[Q8B4J`02L@:4
00FAUY6FRP05Q005RYJAUY407@00396FTIJAUY6FTIJAUQ8000>AUYH02IH01I6FTIJA00>F00>AUYH0
29H00g8000050005Mg6FQ9401iH00hZ4000600P700061c0000A=DWYB10001gEB001jDWX01@001INF
Ui6F01T00YH02K2A/96`TK2A/003Q005UiJGUYL00YH02924Q;2A/9Ig1005Q00DT8B4UYNF/9J`UZV4
WHBMQ9NF/9HL0007MiJ`U[2F/002UP03T@000180021>U[2F[iJ`UZnF/9J`U[2F/9J`U[2F/9J`UZnF
[iJ_L@D0011aMiJ`Q;2F[iJ_UZnF[hBM1`081`001PLa0004CE9IDP<000MBF@00DUUB00D000=aT@00
70004iJAUY6FTIJAUY6:Q8ZAUY6FTIH01H@01IJAUY5b00<000=bRh@01H@02IJAUY6FTIJAUP04Q005
TIJAUY4070001WNAUY6FTAP02iH02i6FTIJAUY6FTIJA00VF00Ea001gL@02UP:42iH2Q0H020L000H7
<P002DeBNU9jDWYBNP020003C@00008000=>07L07`002[2A/96`TK2A/8/2Q0:F00>GQ8@018@03Y2`
U[2A/9IgLGNFUi2:1H@04YNF/9J`U[2F/8BMQ:nFUiJ`UQ`000=gTIH06`009ZnF/9J_U[2F[iJ`U[2F
/9J`U[2F/9J_UZnF[iJ_UWMaMiJ`Q9f40YH03;2F[iJ_UZnF[i2MQ0H020L000H7=`000gE=00050008
Ne9IDUVAUW4M0009TIJAUY6FTIJA00N400V@TI6FTIJALW401@00276FTIJAUY6:18@0296FTIJAUY6F
0X@028ZFTIJAUY6F6004UP03L@0001X000>FTIH00YH00i6FUP0DUP05M`2:TH@019H00hB@UP06UP03
RXBA00:41P081`001PLl000<NgEjDWYBNU:GTIMa7@00296GUYNAUiJA18@02I2FTINA/96GL@09000:
TK2AUi6`TIJ4W@:400B@UYJG0YH3Q004XhBTQ0:F00>`TIH04`000aP0/006UP03L@0001X000V`U[2F
/9J`UZl00iH01JnFUYJ_00>F00J`U[2FMg42J`0FWHBMQ:VF[iJYUZV@[iJ`Q:>@[iJMQ0H020L000H7
>@0047]BFE9IDUUBFGMlTIJAUW4M0009TIJAUY6FTHB@00:A00RFTIJAUY5bL@X000VFTIJAUY6FTHX0
0h@01HZAUY6:00>400>@RX@018@018ZFTIHA0004UY6FT@>F00FAUYJFL@0L0003UY6F00:F00>AUYH0
39H01GMaJ`2400@000A[T8B:19H2RP05UXZ4TI000iH00i64Q00500P700061cH001@H07]BNU9jDWYg
O96GTINAUi6GLAd000nAUi6GTIN@Ui6GTINAMg403@002W6GTK2AUi6`TI04Q005TIJ4RX@00YH02k2A
WHBMQ:BF/96F00/000TH0;2F/9J`U[0029H00g40000J0005/9J`U[001iH01jnFUYJ_LGL00P000hBM
0005000DLJnFUY2_UZnFWHZ_U[2F[iJYQ:<500P700061cD001JFTIIgFE9IDW^AUY6FTIJAUY6FTIIa
7@002i6FTIJAUY6FTG9a014000YaUY6FTIJAUY6F18@02XZFTIJAUY6FTHX3Q004RY6FT@/000^AUY6F
TIJAUY6FT@03UP05TIJFUW40700019JAUY4:UP03LW40008000=[Q0001`000g6FUP07UP03Q96F00>F
00B:Q8B@10081`001PL`000L602GTINAUi6GDW]BNi6GTINAUi6GMiEBUi6GLAd000RAUi6GTIMaMa@0
00YaUi6GTINAUi6@1H@03i6FUYJGTK2AT8BMQ9NF/00900:F00V`U[2F/9J`U[002YH00g40000J0003
/9J`00NF00EgL@00Q00300>420000g6_UP02UP0>[iJ_UZn@Xi6_UZn4WIH400P700061c0001fAUY6F
TIJAUY6FTEUBUY6FTIJAO59IDWYBUY6FL@0M0004TIJALQP000]aUY6FTIJAUY6FR`04Q009RY6FTIJA
UY6:00:400BAUY6F1@003YJAUY6FTIJAUY6FTIJA0iH01I6FUYJA00:F00=b00006`000iJAUP03UP03
LW5b00800X@500:42@000g6FUP08UP03THBF00:F00>400000`081`001PL_000OTINAUi6GTINAUi5j
DW^AUi6GMgaBNU9jDW]eUi6GL@0L0003Mg4001X000MaUi6GTINA00>F00>@RX@00h@01X^`TINA/0:4
00B;/96G0YH00`2FUP06UP09/9JGU[2F/9J`00^F7P001JnFUW5g00801H@026/0001[Q9f42@000g6_
UP02UP09[iJ_UZnF[iJS00L020L000H7<0007Y6FTIJAUY6FTIIgFE9IDW]BFE9IDW]BUY6FTIJALQT0
00EbLG:AUP0L000<LIJAUY6FTIJAUY6F1H@01924T902Q005UY6FTIH00P005i6FTIJAUY6FTIJAUY6F
TIJAUY6FTIJA00>F00>AUW407@000i5g00020004Q00000>41@0016^4Q6/90003LIJF00NF2P081`00
1PLa000NTINAUi6GTINAUgMjDWYBNU9jDW^AUi6GTINAUi5g5P00275gTIJAUi6F70001G6GTINA00RF
00b:T8Z4RhBGTINAUY450003TIJF00>F00ZAUiJ`TINF/9JG2iH00k1a000I0003Mg6F00T00h@40004
Jif4Q0X000IaUYJ_UZl=00P700061c<001fFTIJAUY6FTIJAUY6FTIJAUY6FTGbAUY6FTIJALP0@0008
LG:4Q8ZFTI02RP04TIJAUQd0019bTIJAUY6FTIJAUY6FTIJAUY4<000MUY6FTIJAUY6FTIJAUY6FTIJA
UY6FTIJAUY6FTG804`0017MaUY42Q003RY6F00/00X@40004JhB4Q0T000=a00003`081`001PLd000M
Ui6GTINAUi6GTINAUi6GTG]BNU9jDWYBUY6GTGL03@001HYaQ8^:00:400V@Ui6GTINAUi407@003g:G
U[2AUi6`TINA/96FT@0@00:F00>`UYH02IH02K2F/9J`U[2F/005UP03M`0000h000=aUYH00X@018Z4
Q8@3UPX000B4WHA[10000f^MQ00J00P700061cD001fFTIJAUY6FTIJAUY5lMUUBFE9IDUUBFE9IMYJA
LP090007RV^5RYJAUP02T0>400RFTIJAUY6FTAd000aaUY6FTIJAUY6FTIHC000MUY6FTIJAUY6FTIJA
UY6FTIJAUY6FTIJAUY6FTG803@002GMaUY2FTIJAUP03Q003T96F00X000A[Q8B41@03Q1P020L000H7
=P007INAUi6GTINAUi5kDWYBNU9kMIIfUE9jDWZAUi5g00H000=bT8@00X@01YJAUi6GT@F400N@Ui6G
TINA01d000UbUi6`TINAUY405@000i6`UP0:UP09/9J`U[2F/9J`00NF00=g00001`001WMa/9JMQ0NF
00B@Q8B`0YH:0004JhB4Q0D000>400005`081`001PLg000MUY6FTIJAUY6FDUUBFGNFTIJAUY6FTG]B
FGNFTG800`00179aUXX3Q008T9JAUY6FTI06Q005T9JAUY407@001I6FTIJA01T001jAUY6FTIJAUY6F
TIJAUY6FTIJAUY6FTIJAUY6FTG850004Mg6FT@F400VAUY6FTIJAUXX00X@00iJAUP0;00>47P081`00
1PLh000SUi6GTINAUgEjDWYgUi6GTINAUi6GTGaBNY2GTGMaUi6GTIH018@0296GTINAUi6@0h@018Z4
Q8@2T@03Ui4001/000=gT@00700;UP09/9J`U[2F/9J`00RF00ma0000LGNF/9JYU[2@WH@02YH00if4
Z@02UP/00h@M00P700061cP001^AUY6FTIICFE9IDYJAUY6FTIJAUY6FTEUBFDh00P001g:AUY6FTHX0
18@0296FTIJAUY6F0h@00hZARP02Q003RY6A01/000>ALP007000:IJAUY6FTIJAUY6FTIJAUY6FTIJA
UY6FTIJAUY6FLG9aUY6FTIJAUY6@00:400ZAUY6FTIJAUY6F0h@00hZAUP0Z00P700061cT001VAUi6G
TIEBNU:FTINAUi6GMiFAUi6GDWYB00@000IgTINAUi45Q008T9NAUi6GTIL3Q003Ri6:00:400>;T940
6002UP03Ui5g01d02YH02K2FUiJ`U[2F/007UP0@Mg40LHBF/9J`U[2F/9JSQ0ZF00JYQ9f4[i4Z00P7
00061cT001VFTIJAUY5IDUUfUY6FTIIfFE9kTG9aFE9I00@000IaUY6FTIH5Q009T96FTIJAUY6F00>4
00RFTIJAUY6FTA@000VFTIJAUY6FTG807@0069JAUY6FTIJAUY6FTIJAUY6FTIJAUY5bL@<00h@02YJA
UY6FTIJAUY42Q00;UY6FTIJAUY6FTI000h@00iJ@000X00P700061cX001>GTINAUgMjDWZAUi6GTIMB
NU9e008000ABNU9e10001g6GTINAUh/018@028ZAUi6GTIN;0X@2T@08Ui6FTINAUY4@00RF00FGTINA
M`0M00RF00VGU[2FUiJ`UYL00YH0175g0002Q0<000i[TK2F/9J`U[2F[h^4RPRF00R`UYJ4WHBYTBP0
20L000H7>`004IJAUY6FDUUBFI6FTG9a059I00=B00AIDUUB1@00276FTIJAUY6@18@01Y6FTIJAUP:4
00ZAUY6FTIJAUY6F3P00496FTIJAUY6FTIJAUY6FTG8M000AUY6FTIJAUY6FTIJAUY6FTG800P04Q0D0
00]aQ9JAUY6FTIJ4RP02Q00>T9JAUY6FTIJAUY6FRX@Z00P700061cd000bAUY6FTG]BNU9jDWD30005
NU9j04d020007g6GTINAUi6GTINAUY6FTIJAUY6FTINAUi6GTINAUi401`005Y6FTIJAUi6GTINAUi6G
TIJAUY6FTG8N0005TIJAUY4019H00gMa00070004JhB4Q0D000=aUY000XX4UP04T8B4RPNF;`081`00
1PLn000<TIIaLP1=DUUBFE9=0`000eT00005009B029I001aUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJA
UY6FTIJA10006IJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTG807P001i6FTIJALW400`000hA[00070003
JhB400D000YaUY6@Q9JAUY6:0X@01I2FTIJA034020L000H7?`000g400002009=00AjDWYB20001gEB
NU9jDWX00P00?76GTINAUi6GTIJAUY6FTIJAUY6GTINAUi6GTINAU`2FTIJAUY6FTIJAUi6FTINAUY6G
TIJAUY6FTIJALQd000BFTIIa1P03Q003J`0000D000A[Q8A[10000g6FUP02UP04RY2FUP>400>F0000
<P081`001PM@00=B00IIDUUBFE83000LLIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIIhUP8001jAUY6F
TIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTG8L0005LG9a06/01P03Q0L000=[Q8@01P002W6FTIJAUXZF
RX@f00P700061dh000MeDWYBNU9j009=00=BNU800`00676GTINAUi6FTIJAUY6FTIJAUi6GTINAUPH0
01fAUY6FTIJAUY6FTINAUY6GTIJAUY6FTIJAUY6FT@0J0004Mg6FT@:41`03Q0=[1@000f^4Q0050003
MiJF00>F00>:0000=`081`001PM;009B00IIDUUBFE82C@<000=BFE800`005G6FTIJAUY6FTIJAUY6F
TIJAUY6FN009000MTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUTh05P002GMaUY6FQ9J@J`0800:4
0`000f/0000400:41@0017:AUY4k00P700061dT000MBME9jDWYB009=1P000gEBNP03000BMi6GTINA
UY6FTIJAUY6FTIJA3@007I6FTIJAUY6GTIJAUi6FTINAUY6FTIJAUY6FTIIa01<000=aUY4018@01IJA
UY6400L000=[Q8@00`03J`@00XA300P700061dL000Q=DUUBFE9IC@X000=IDUT00`003G:AUY6FTIJA
UY6FTIH00Y4@000MTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUTh040000g9aUP02Q008TIJAT8Z4
RYH20003J`0000@000=[Q6/01@0018@00003Q48020L000H7B0000deBNP03C@/000ABNU9j0`0037NA
Ui6FTIJAUY6FTA8001jFTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUW4<000AMg6FTHBAUY6FTIJA
UY6FRYH00P02J`D00X@70004Jh@0Jd<020L000H7B@02C@l000ABFE9I0`001G:AUY6F00>A5P007YJA
UY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FCPP000AbLIJA0h@03XZFTIJAUY6FTIJAUY6F0`000f/0
000200:400=[0000B`081`001PMS0003L@0001/001VAUi6FTIJAUY6FTIJAUi6GTIJAUi6FTIJA00>F
00Ia0000LGL3T@05Ui6FTHX00X@7UP:@0iH00i6FT@040003Q6/000>4C0081`001PMR0003TIH001d0
03:AUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FLG:AUXBFTIJAUY6FTHZ4RY6FTIJAUP::00JAUY6F
TIH80003Q00004/020L000H7G`001Y6FTIJAMah001fAUY6FTIJAUY6FTINAUY6GTIJAUi6FTIIaLW40
L@03Q00:TIJAUY6FTIJAT0:400>AUYH00YH00i24RP02UP04TIJAUPD000=[0000C@081`001PML000:
TIJAUY6FTIJALQd001RFTIJAUY6FTIJAUY6FTIJAUY6FTIJALW42000@Jh@0LHZ4RY6FTIJAUY6FT@>4
00B:UY6F18@01Y6FTIJAUU@020L000H7FP003IJAUY6FTIJAUY6GTGL07@0059JAUY6FTIJAUY6FTINA
UY6GTIIa0P000h@00004000;LI64Q96FTIJAUY400YH00hZ4Q002Q0FF00J@TIJAUY5D00P700061eP0
012AUY6FTIJAUY6FTIJAUY5b7@004IJAUY6FTIJAUY6FTIJAUTib00800X@8000>LHZ4Q8ZFTIJAUY6F
TI03Q005UY6FTIH00Y000i40001E00P700061e@001A>TINAUi6FTIJAUY6FTIJAUi6GLAh000jFTIJA
UY6FTINAUY5gL@8018@90003TI2400:400JFTIJAUY42UP05THB4Q94019II00P700061e<001JFTIJA
UY6FTIJAUY6FTIJAUY6FTII>7P002YJAUY6FTIJALTh700:400=[00001P000g:AT003Q009UY6FTIJA
UY6F00>400>@T@00FP081`001PM@000JUi6GTINAUi6GTIJAUY6FTIJAUY6GTINAUg4M0007TIJAUY6F
L@0:0004JhB4Q0P000=gTIH00h@2T@>F00FAUYJFQ01N00P700061dh001fAUY6FTIJAUY6FTIJAUY6F
TIJAUY6FTIJAUY6FCP0M0006TIJALTh00X@900>42@000hZ4T003Q007RY6FTIJAUP1P00P700061dl0
01fAUi6GTINAUi6GTIJAUY6FTIJAUY6FTINAUY6GL@0L0004Mg4000:400=[00002@03Q003J`0000<0
0V/0176FTI03Q003UY6F06<020L000H7D0007Y6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJALQD0
00QbLIJAQ96FT@<00X@=00>41`0017:AUY5W00P700061e4001fAUY6FTIJAUY6FTIJAUY6FTIJAUY6F
TIJAUY6FT@0C0003LGNA00:400J:TIJAUY4200>400=[00002P03Q003J`0006h020L000H7D@007YJA
UY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FCPl000IaLY6FQ8X3Q006UY6FTIJA0P000f^4Q00=00>4
00=[0000K@081`001PMB000NUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIIa2`002W5gTIJAUY6F
TI02Q007RiJAUY6FT@020004JhB4J``000=[Q8@0K`081`001PMC000MUY6FTIJAUY6FTIJAUY6FTIJA
UY6FTIJAUY6FTG802@003G9aUY6:TIJAUY6FTHX00X@01Y6FTIJAUP@00hAm00P700061e@001fFTIJA
UY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJALP060003LGNA00:400ZFTIJAUY6FTIJA0X@01hZAUY6FTIH0
100016^4Q6]k00P700061eD001fFTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJALP02000BCW:AUXB:
Q8ZAUY6FTIJAUY6@0X@01hZFTIJAUY401P02Q003J`0007P020L000H7EP0089JAUY6FTIJAUY6FTIJA
UY6FTIJAUY6FTIJAUY5bLIJA0X@2T@0<UX^:TIJAUY6FTIJA0h@01i6FTIJAUY40100016^4Q6]j00P7
00061eL001ZFTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY5bCP8018@028ZAUY6FTI240Y4019JAUX@2RP:4
00R:TIJAUY6FT@D000>40000N@081`001PMH000FUY6FTIJAUY6FTIJAUY6FTIJAUY6FL@8000E[Q6/0
L@03Q00<T96FTIJAUX^:THZ@0Y4039J:Q8B@TIJAUY6FTH0020L000H7F@0059IhUY6FN9JAUWRFTIJA
UY6FLG9>20002g:4Q8BFTIJAUY6F00:A00J4UY6FTIH3Q007UY6FTIJAUP2000P700061eX000fFTIJA
UY6FTIJAUW5b00D000=[Q6/02000376@Q8BAUY6FTIJAUP:A00F4TI6FR`02Q003RP0008<020L000H7
FP07T@03UY5b008000=[00001003Q0P0019aQ6^4MiIhUWRFN9JAUY25MiJ800P700061e/000VAUY6F
TIIaLP000X@700:400=[Q6/01`003i64Q8B;UY6FTIJAUY6FT@2:00P700061e`00Y4014ib0002Q003
J`0000H000=[Q8@00P000hA[0004000<UV^4JhYhUWRFN9JAS@081`001PMP0006J`0008A[2003Q0<0
0f/20003LIJ400>400BAUY6FS`081`001PMU009[1`0016^4Q6/4009[0P001djFLHA[Q7L0T@081`00
1PMW009[1P03Q003J`0000<000I[001aUX^D00P700061fT000>400000`0016^4Q8@6009[U`081`00
1PMZ009[1003Q003J`0009`020L000H7K0000f/0000200>4WP081`001PM^0003J`000:4020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7003o1b47003o1b47003o1b47003o1b47
003o1b47003o1b470000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output",
  CellLabel->"Out[3]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "For the cylindrical mapping with the ",
  StyleBox["MVCylinderMapping",
    FontFamily->"Courier"],
  " transformation it is assumed that the ",
  Cell[BoxData[
      \(TraditionalForm\`z\)]],
  " values of the polygons are in ",
  Cell[BoxData[
      \(TraditionalForm\`z\  \[Element] \ \([0, 1]\)\)]],
  ". When the linear transformations are applyed, the texture coordinates \
obtained by ",
  Cell[BoxData[
      \(TraditionalForm
      \`\(\((\[Phi]\/\(2\ \[Pi]\), z)\) \[RightTeeArrow] \ 
        \(\((x\_t, y\_t)\) . \)\ \)\)]],
  "The following set of transformations brings the default ",
  StyleBox["Cylinder[]",
    FontFamily->"Courier"],
  " into this range."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; \n
    MVShow3D[\n\t
      Graphics3D[\n\t\t\ \ Cylinder[], 
        ViewPoint -> {3.111, \ \(-0.319\), \ 1.281}\n\t1], \n\t
      MVTexture -> "\<`ChessBoard`\>", \n\t
      MVTextureMapType\  \[Rule] \ MVCylinderMapping, \n\t
      MVTranslateTextureTarget\  \[Rule] {0, 0, 1}, \n\t
      MVScaleTextureTarget\  \[Rule] \ {1, 1, 1/2}, \n\t
      MVScaleTexture -> {2, 1}\n]\)], "Input",
  CellLabel->"In[3]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVScene3D[] \[DoubleLeftArrow] MVTexture2D[1], 
      MVCylinderMapping, {256, 256, MVRGBTexture}}\)], "Output",
  CellLabel->"Out[3]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVPasteGraphics[]\)], "Input",
  CellLabel->"In[4]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007050606F5060703F40\
70303F70202F6F50607F2F2F705F607F3F406010107F00007F0000600000000000
0000000000000000000000003F00003000003F0000000000000000000000000000000000000000\
0000000000000000000000000000000000000000009000008F00009000008F0000
900000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000A070709F706FA07F709F706FA070709F706F
A07F709F706FA0706F9F706FA07F6F9F1F10A01F1F9F706FA07F6F9F706FA07F6F9F7060A07F6F\
9F7060A07F6F9F7060A07F6F9F7060A07F6F9F2F2FA0302F9F00009F70609F7060
A07F609F7060A07F609F7060A07F609F7060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F50606F4F6070101F6F20207050606F5060\
705060705060705F607050607F5F6F7050603F2F304F00000F000F6000001F0000
0000000000000000000000000000006F00006F00003F0000000000000000000000000000000000\
0000000000000000000000000000000000003F00008F00009000008F0000900000
700000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000009F706F9F70709F706FA070709F706F9F706F
9F706FA0706F9F706F9F706F9F706FA000009F00009F706F9F706FA0706F9F70609F706F9F7060\
A0706F9F70609F706F9F7060A0706F9F70609F1F109F1010A000009F403FA07060
9F7060A070609F7060A070609F7060A07F609F705F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F50606F4F606F00006F000070303F6F506070506F70506070506F\
7050607F5F6F7050607F5F6F705F607F5F6F3F2F300F0F0F0000001F0000000000
0000000000000000000000000000005000002000008000006F00006F00007000006F0000700000\
6F00007000007000007000007000007000007000009000008F0000900000700000
3F0000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000A07F709F706FA07F709F706FA07F709F706F
A07F709F706FA07F709F706FA07F6F9F2F2FA030309F4040A07F6F9F706FA07F6F9F706FA07F6F\
9F7060A07F6F9F7060A07F6F9F7060A07F6F9F0000A000009F0000A000009F7060
9F70609F7060A07F609F7060A07F609F7060A0302F9F0000A07F60000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
604F5F604F60604F5F6F4F60604F5F6F00006F00006F00006F4F606F50606F50607050606F5060\
705060705060705F6F7050607F5F6F705F603F2F30000000000000000000000000
0000000000000000000000000000000000000000005000004F00005000001F00001F00001F0000\
1F00001F00002000001F00001F00001F00007000008F00009000008F00003F0000
3F0000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000009F706FA070709F706FA070709F706FA07070
9F706FA0706F9F706FA0706F9F706FA04F409F403FA0302F9F706FA0706F9F706FA07F6F9F7060\
A0706F9F7060A07F6F9F7060A0706F9F5F50A000009F0000A000009F0000A04F40
9F403FA070609F7060A07F609F7060A070609F705FA0604F9F00009F70609F705F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000604F5F604F60604F5F
6F4F60604F60604F60604F5F6F101F6F00006F303F6F506070506F6F506070506F705060705F6F\
705060705F6F7050607F5F6F7F5F6F7F5F6F3F2F30000000000000000000000000
0000000000000000000000000000000000000000001F00007000006F0000000000000000000000\
0000000000000000000000000000000000008F00009000009000009000001F0000
200000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000A07F709F706FA07F709F706FA07F709F706F
A07F709F706FA07F709F706FA07F709F706F9F706F9F1F10A07F6F9F706FA07F6F9F706FA07F6F\
9F706FA07F6F9F7060A07F6F9F7060A04F409F0000A00000A00000A000009F0000
A000009F7060A07F609F7060A07F60A07060A07F609F7060A060509F5F4F9F00004F3F2F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F4F5F5F405F60202F5F1F20604F60
604F5F604F60600F106F10106000006F101F6F4F606F50606F50607050606F5060705060705060\
705F6F705060705F6F7050607F5F6F7050603F2F30000000000000000000000000
0000000000000000000000000000000000000000000000006F00007000001F0000000000000000\
0000000000000000000000002000001F00009000008F00009000008F0000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000009F706FA070709F706FA07F709F706FA07070
9F706FA07F709F706FA0706F9F706FA07F6F9F706FA0302F906F60A07F6F9F706FA0706F9F706F\
A07F6F9F7060A0706F9F7060A07F6F9F403FA000009F0000A000009F0000A00000
9F0000A04F409F7060A07F609F7060A07F609F705F9F70609F705F9F70609F10109F00009F0000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000604F605F0000604F60604F5F604F60604F60
604F60604F606F00006F00006F00006F303F6F50606F50607050606F5060705F6F705060705F6F\
7050607F5F6F705F6F7F5F6F705F6F7F5F6F3F2F30000000000000000000000000
0000000000000000000000000000000000000000000000002000001F0000700000000000000000\
0000000000000000000000003F00003F00008F00009000009000007F0000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000A07070A07F709F706FA07F70A0706F
A07F709F706FA07F70A0706FA07F709F706FA07F70A04F40A04F40906F60A07F6FA0706FA07F6F\
9F706FA07F6FA0706FA07F6F9F7060A01F1FA00000A000009F0000A00000A0302F
A0302F9F0000A07F6FA07060A07F609F7060A07F609F7060A07F609F705FA04F3F9F00009F0000\
9F00009F0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005040505F00005F1F205F4F5F5F405F604F60604F5F604F60
604F5F6010106000006F00006000006F50606F4F606F50606F4F607050606F506070506F705060\
70506F7050607F5F6F7050607F5F6F7F5F6F3F2F30000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F00002000001F0000\
0000000000000000000000003F00003F00009000008F00005F0000500000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000A07F709F706FA07F709F706FA07F70
9F706FA07F709F706FA07F709F706FA07F6F9F706F9F706F9F2F2FA04F409F706FA07F6F9F7060\
A07F6F9F706FA07F6F9F7060A07F6F9F1F10A000009F0000A000009F0000A04F40
9F403FA0302F9F7060A07F609F7060A07F609F7060A07F609F705F9F70609F705F9F00009F0000\
9F00009F00009F0000900000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000050405F5F00005F0000603F4F5F405F604F60604F5F604F60604F5F
6F50606000006F00006000006F101F6F50606F50606F506070506F6F506070506F705060705F6F\
70506F7F5F6F70506F7F5F6F7F5F6F7F5F6F3F2F30000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002000005000005F0000\
0000000000000000000000007000009000008F00009000003F00003F0000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000009F7070A07F70A07070A07F709F706F
A07F70A0706FA07F709F706FA07F70A0706FA07F70A0706FA01F1FA0302FA07F6FA0706FA07F6F\
A0706FA07F6FA0706FA07F6FA07F6FA00000A00000A00000A00000A00000906F5F
9F706090705FA00000A07060A07F6FA07F60A07F609F7060A07F609F7060A07F609F705FA00000\
9F00009F00009F00004000009F00009F0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
50405F5000005F00005F00005F00005F2F30604F605F405F604F60604F5F604F60
603F4F6F00006000006F10106F10106F50606F4F606F50606F506070506F6F506070506F705060\
705F6F7050607F5F6F70506F7F5F6F3F2F303F2F30000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007000006F0000\
2000000000000000000000007000008F00009000008F00002000001F0000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000A07F709F7070A07F709F706FA07F70
9F706FA07F709F706FA07F709F706FA07F709F706FA07F6F9F4040A01F1F9F7060A07F6F9F706F\
A07F6F9F7060A07F6F9F706FA060509F0000A000009F0000A000009F1010A07F6F
9F7060A07F6F9F2F20A000009F7060A07F609F7060A07F609F705FA07F609F705F9F70609F0000\
9F0000900000400000000000200000000000900000900000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000050405F\
5000005F101F5F00005F00005F000060303F604F5F604F60604F5F604F60602F3F
6F303F6000006F00006F10106F50606F506070506F6F506070506F6F5060705F6F705060705F6F\
7050607F5F6F705F6F7F5F6F7F5F6F40303F3F2F30000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F00003F0000\
3F00000000000000000F0F0F8F0000900000900000900000000000000000000000
0000000000000000002000002000005F00007F00007F00006000003F00004000005F0000000000\
000000000000000000000000000000000000A07070A07F70A07070A07F70A07F70
A07F70A0706FA07F70A07F6FA07F70A0706FA07F70A07F6F9F706F906F60A04F40A07F6FA07F6F\
A0706FA07F6FA07F6FA07F6FA04F40A00000A00000A00000A00000A04F40A07F60
A07F6FA07F609F7060A0302FA04F40A07F60A07F60A07F60A07F609F7060A07F609F705F801F1F\
9F00009F00003F00000000000000000000000000005F00009000009F0000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F3F5050405F500F10\
50304F503F505F00005000005F00005F405F604F605F405F604F60604F5F60303F
602F3F6F00006000006F101F604F606F50606F506070506F6F506070506F6F5060705F6F705060\
705F6F7050607F5F6F7050607F5F6F3F2F303F3030000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000006F0000\
7000000000002000001F00009000008F0000900000700000000000000000000000
0000000000001F00007F00007000002000000000000000000000004000003F0000600000000000\
000000000000000000000000000000000000A07F709F706FA07F709F7070A07F70
9F706FA07F709F706FA07F709F706FA07F70A0706FA07F709F706FA07F6FA0302FA07F6F9F706F\
A07F6FA0706FA07F6F9F7060A0302FA00000A000009F0000A00000A04F3FA07F6F
9F7060A07F6FA070609F70609F00009F7060906F5FA07F609F705FA07F609F705F9F7F607F1010\
9F00009F00005F00000000000000000000000000000000009000005F0000900000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F3F5050405F503F5050405F\
50405F5F405F5F0F105F00005F0000604F605F4F5F604F60604F60604F60602F3F
6F30406000006F00006F10106F506F6F506070506F6F5060705F6F705060705F6F7050607F5F6F\
705F6F7F5F6F705F6F7F5F6F7F5F6F40303F3F2F300000000000000000001F0000
2000001F00000000000000000000000000000000000000000000000000000000000000003F0000\
3F00003F00003F00005F00008F00009000009000005F0000000000000000000000
0F0F0F0000007F00000000000F0F0F7F0000000000000000400000000000000000200000000000\
000000000000000000000000000000000000A07070A07F70A07F70A07F70A07070
A07F70A07F70A07F70A07F6FA07F70A07F6FA07F70A07F6FA07F70A07F6FA01F1FA07F6FA07F70\
A07F6FA07F6FA07F6FA07F6FA01F10A00000A00000A00000A00000A06050A07F60
A07F6FA07F60A07F6FA07F60A0302FA0302FA0302FA07F60A07F609F7F60A07F609F705F801F1F\
9F00009F00007F00000000000000000000000000000000007F0000400000000000
000000700000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F3F504F3F50503F504F3F5050405F503F50\
50405F5040505F1F205F00005F10105F405F604F60604F5F604F60604F5F60303F
6000006F00006000006F10106F4F606F50606F4F606F50606F506070506F6F506070506F705060\
705F6F7050607F5F6F705F6F7F5F6F3F2F300000000000000000000000004F0000
3000006F0000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F6F00008F00008F00009000008F00009000003F0000000000000000000000
3F00003F00003F00000000000000007F00000000000F00005F00004000003F0000000000000000\
000000000000000000000000000000000000A07F70A07070A07F709F706FA07F70
A07070A07F709F706FA07F70A0706FA07F709F706FA07F70A0706FA07F709F4040A04F40A0706F\
A07F6F9F706F9F7060906F60A00000A00000A00000A00000A00000A07060A07F6F
A07F60A07F6FA07060A07F6F90705FA000009F000090705F9F7060A07F609F705F9F7F605F2F20\
9F00009F00007F00000000000000000000000000000000003F0000500000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F4F3F504F3F504F3F5050405F503F5050405F\
50405F5F405F5F304F5F00005F0000603F50604F60604F60604F606F5060602F3F
6F00006000006F00006F00006F30406F506070506F6F506070506F705060705F6F70506F705F6F\
70506F7F5F6F7F5F6F7F5F6F7F5F6F40303F0000000000000000000000006F0000
6F00006F0000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F8F00009000008F00009000009000003F0000000000000000000000
4000003F00004000000000000000007F00004000003F00002000003F0000400000200000000000\
000000000000000000000000000000000000A07F70A07F70A07F70A07F70A07F70
A07F70A07F70A07F70A07F70A07F70A07F6FA07F70A07F6FA07F70A07F6FA0605FA0302FA03030\
A07F6FA07F70A04F40A04F40A00000A00000A00000A00000A01F10A07F6FA07F6F
A07F6FA07F60A07F6FA07F60A07F6FA0302FA0302FA0302FA07F609F7F60A07F609F7F6060302F\
9F0000A000007F00000000000000000000000000000000000000007F0000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F3F504F3F504F3F504F3F5050405F503F50\
50405F50405F5F00005F00005F00005F3F4F604F605F405F604F60604F606F4050
6000006F00006000006F00006F1F206F50606F506070506F6F506070506F705060705F6F705060\
705F6F70506F7F5F6F705F6F7F5F6F3F2F30000000000000000000000000800000
6000006F0000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F00003F00009000008F00009000001F0000000000000000000000
1F00002000005F00000000000000007F00000000007F00005F00000F00007F00007F0000000000\
000000000000000000000000000000000000A07F70A07070A07F70A07070A07F70
A0706FA07F70A07F70A07F70A0706FA07F70A07F6FA07F70A0706FA07F70A07F6FA01F1FA01010\
A07F6FA07F6FA04F40A0403FA00000A00000A00000A00000A0302FA07F6FA07F6F
A07060A07F6FA07F60A07F6FA07F60A07F60A07F60A000009F7060A07F609F7060A07F604F3F2F\
9F00009F00009F00000000000000000000000000000000000000003F00000F0F00
0000000000000000000000000000000000000000005F0000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F00000000000F0F0F4F3F5050405F4F3F5050405F503F505F405F\
50405F5F405F5F00006000005F000060303F604F60604F60604F606F4F60604F60
6F00006F00006F00006F00007000006F4F607030406F5060705F6F70506F705F6F70506F7F5F6F\
705F6F7F5F6F705F6F7F5F707F5F6F40303F0000000000000000003F0000800000
800000800000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000700000900000900000000000000000000000000000
0000000000007F00001F00002000005F00002000003F00004000007F00004000003F0000000000\
000000000000000000000000000000000000A07F70A07F7FA07F70A07F70A07F70
A07F70A07F70A07F70A07F70A07F70A07F70A07F70A07F6FA07F70A07F6FA07F70A0302FA03030\
A04F40A07F70A01F10A01F1FA00000A00000A00000AF0000A04F40A07F6FA07F6F
AF7F6FA07F6FA07F6FA07F60A07F6FA07F60A07F6FA0302F9F7060A07F60A07F609F7F60503F30\
9F0000A000009F0000200000000000000000000000000000000000000000700000
0000000000000000000000000000000000000000000000005F0000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004000000000000000000000004F3F504F3F5050405F4F3F5050405F503F50\
5F405F50405F5F10105F00006000005F1F20604F60604F60604F60604F60604F60
601F2F6F00006000006F00006F00006F30406F506070506F6F5060705F6F705060705F6F70506F\
7F5F6F70506F7F5F6F7F5F6F7F5F6F3F2F300000000000000000007F0000800000
4F0000500000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F00003F00003F0000000000000000000000000000
0000000000000000007F00007F00002000003F00004000003F00000F0F0F000000000000000000\
000000000000000000000000000000000000A07F70A07F70A07F70A07070A07F70
A07F70A07F70A0706FA07F70A07F6FA07F70A0706FA07F70A07F6FA07F70A07F6FA04F40A04F40\
A03030A07F6FA00000A00000A00000A00000A00000A00000A04F40A07F60A07F6F
A07F6FA07F6FA07F60A07F6FA07F60A07F6FA07F609F7060A00000A07F609F705FA07F604F3F2F\
A000009F00009F00001F00000000000000000000000000000000000000003F0000
0000000000000000000000000000000000000000000000000000008F0000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F00004000000000000000000000000F0F0F4F3F5050405F4F3F5050405F50405F5F405F\
50405F5F4F605F1F205F00005F000060101F604F60604F60604F606F506F604F60
6F30406F10106F101F6F00006F00006F101070304070303F70506F70506F705F6F705F6F7F5F6F\
705F6F7F5F707F5F6F7F5F707F5F6F40303F0000001F00007F00008000004F0000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000200000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000A07F70A07F7FA07F70AF7F7FA07F70
A07F70A07F70AF7F70A07F70A07F70A07F70AF7F70A07F70A07F70A07F6FAF7F70A07F6FA07F70\
A000009F706FA00000A00000A00000AF0000A00000A000009F7060AF7F6FA07F6F
A07F6FA07F6FAF806FA07F6FA07F6FA07F60A0806FA07F60A0302F90705FA07F60A07F60503F30\
7F0000A000009F0000400000000000000000000000000000000000000000000000
4000000000000000000000000000000000000000000000000000000000008F0000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F00004000000F00000000000000000000000000004F3F504F3F50503F504F3F5050405F503F50\
5F405F50405F5F2F3F5F00006000005F0F10604F60604F60604F60604F606F5060
604F606F303F6F2F3F6F101F6F00006F00006F00007000006F202F70506F70506F705F6F70506F\
7F5F6F705F6F7F5F6F705F6F7F5F6F4F202F4F00007F0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000A07F7FA07F70A07F70A07F70A07F70
A07F70A07F70A07F70A07F70A07F6FA07F70A07F6FA07F70A07F6FA07F70A07F6FA07F70A07F6F\
A0302FA04F40A00000A00000A00000A00000A01F1FA01F10A07F6FA07F6FA07F6F
A07F60A07F6FA07F6FA07F6FA07F60A07F6FA07F60A07F6F90705FA0302F9F7F60A07F604F3F2F\
8000009F00009F00003F0000000000000000000000000000000000000000000000
0000007F00000000000000000000000000000000000000000000000000005000008F0000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F0000
4000004000001F00000000000000000000000F0F0F4F3F5050405F503F5050405F50405F5F405F\
50405F5F405F5F304F6000005F0000600000603F4F605060604F606F5060604F60
6F506F6F50606F101F6F40507000006F00007000006F0000700000703040704F5F70506F7F5F70\
705F6F7F5F707F5F6F7F4F5F7F4F5F6F1F1F4F00003F0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000A07F70AF7F7FA07F70A07F7FA07F70
AF7F7FA07F70AF7F70A07F70AF7F70A07F70AF7F70A07F70AF8070A07F70AF7F70A07F6FAF8070\
A04F40AF1F1FA00000AF0000A00000AF0000A0302FAF3030A07F6FAF7F6FA07F6F
AF806FA07F6FAF7F6FA07F6FAF806FA07F6FA0806FA07F60A0806FA00000A060509F7F60504030\
400000A000009F0000600000000000000000000000000000000000000000000000
0000003F00000F0F0F0000000000000000000000000000000000000000005F00008F00008F0000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F00000F0000000000
3F00004000001F00000000000000000000004F3F504F3F504F3F5050405F4F3F5050405F50405F\
5F405F50405F5F3F4F5F00006000005F0000602F3F604F60604F60604F606F4F60
604F606F506F6F4F606F506F6F303F70202F6F000070000070000070000070000070304070506F\
7F101F70303F7F101F7F10107F101F3F2F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004F3F3F503F3F4F3F3F503F404F3F3F503F3F4F3F3F503F404F3F3F503F3F503F3F\
504040503F3F50403F503F3F504040503F3F503F3F4F3F3FA07F7FA07F70A07F70
A07F70A07F70A07F70A07F70A07F70A07F70A07F6F9F706FA07F6FA07F70A07F6FA07F70A07F6F\
A0605FA00000A00000A00000A00000A00000A04F40A04F40A07F6FA07F6FA07F6F
A07F6FA07F6FA07F60A07F6FA07F60A07F6FA07F60A07F60A07F60A0302F9F1F10A07F604F3F2F\
4000009F00009F00007F00007F0000000000000000000000000000000000000000
0000002000007000007F00000000000000000000000000000000000000006F00008F0000500000\
800000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F0000000000000000000000
4000004000002F00000000000000000000004F3F5F4F3F5050405F503F5050405F50405F5F405F\
50405F5F4F605F405F6000005F0000600000602F3F604F60604F606F506F604F60
6F506F6F4F606F506F6F506F70506F6F506F70203070202F70202F7000007F1F1F7000007F3F40\
7030407F2F307F5F6F7F60703F2F3040303F000000000000000000000000000000
0000000000000000000000000000000000000000000000004F3F404F303F4F3F404F3F3F4F3F40\
4F3F3F4F3F404F3F3F4F3F404F3F3F503F404F3F3F503F404F3F3F503F404F3F3F
503F404F3F3F9F707F9F70709F707F9F70709F7F7F9F70709F7F7F9F7070A07F7F9F707FA07F7F\
9F7070A07F7FA07F7FA07F7FA07F70AF807F000000000000503F3F50403F503F3F
50403FA07F70AF807FA07F70AF7F70A07F70AF8070A07F709F706FA0605FAF4F4FA07F70AF7F70\
A0302FAF0000A00000AF0000A00000AF0000A00000AF6F5FA07F6FAF8070A07F6F
AF806FA07F6FAF806FA07F6FAF806FA07F60AF806FA07F60A0806F90705FA00000A07F60504030\
4000004000009F0000A000009F0000000000000000000000000000000000000000
0000000000009F00003F00007F00007000000000000000000000000000009000008F00008F0000\
1F0000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0000000000000000000000300000
3F00004000004000000000000000000F000F4F3F504F3F5F4F3F5050405F503F5050405F503F5F\
5F405F5F405F5F10105F000060000060000060202F604F60604F60604F606F506F
604F606F506F6F4F606F506F6F506070506F6F506F705F6F6F506F705F6F70506F7F4F5F70506F\
7F5F6F705F6F7F5F707F5F6F3F303F3F2F3000000000000000000040303F40303F
40303F4F303F40303F4F3F4040303F4F3F3F40303F4F3F408F60708F6F7F8F6F70906F7F8F6F70\
906F7F8F6F7090707F8F6F7090707F906F7090707F906F7090707F906F709F707F
906F709F707F9070709F707F9070709F707F9F70709F707F9F70709F707F9F7070A07F7F9F7070\
A07F7F9F7070A07F7FA07070A07F7FA07F70000000000000000000000000000000
000000503F3F4F3F3F503F3FA07F70A07F70A07F70AF7F70A07F6FA07F70A07F6FAF3030A0302F\
A00000A00000AF0000A00000A00000A00000AF00009F7060A07F6FA07F6FAF7F6F
A07F6FA07F6FA07F6FAF7F6FA07F60A07F6FA07F60A07F6FA07F60A07F60A0302FA000009F604F\
4F3F30200000A000009F00009F0000000000000000000000000000000000000000
0000000000001F00003F00001F00002000007000007000000000000000008F00008F0000300000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000001F0000
4000004000004000000F000000000000000F4F3F5F4F3F5050405F503F5F50405F50405F5F4060\
5F405F5F4F605F1F206000006000006000006010106F506F604F606F506F604F60
6F506F6F506F6F506F6F506F70506F6F506F705F6F70506F705F6F70506F7F5F6F705F6F7F5F70\
7F5F6F7F5F707F5F6F8060703F2F3040303F40303F40303F40303F806070806070
8F6F7F8060708F6F7F8F60708F6F7F8F6F708F6F7F8F6F70906F7F8F6F70906F7F8F6F7090707F\
906F7090707F906F7090707F906F7090707F906F709F707F90707F9F707F907070
9F707F9F707F9F707F9F70709F7F7F9F707F9F7F7F9F7070A07F7F9F707FA07F7F9F707FA07F7F\
A07F7FA07F7FA07F70A07F7FA07F7FAF7F7F000000000000000000000000000000
000000000000000000000000000000503F3F50403F503F3FAF8070A07F70AF8070A07F70AF8070\
A07F70AF8070A00000AF0000A00000AF0000A00000AF8070A07F6FAF8070A07F6F
AF8070A07F6FAF806FA07F6FAF806FA07F6FAF806FA07F60AF806FA07F60A07F6090705FA01F1F\
4F3F302000009F0000A000009F0000000000000000000000000000000000000000
0000000000000000007000000000000000002000001F00007F00007000007000008F0000200000\
0000000000008000000F0F0F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003000000000000000000000000000001F0000
3F00004000004000000000000000000F0F0F4F3F504F3F504F3F5050405F503F5050405F50405F\
5F405F5F405F5F2F3F5F00006000005F00006000005F405F6F506F604F606F506F
604F606F50606F4F606F506F6F506070506F6F506070506F70506F705F6F70506F705F6F705F6F\
7F5F70705F6F7F5F707F5F6F40303F3F2F3F8060707F5F70806070805F70806070
8060708F6F708060708F6F7F8F60708F6F7F8F60708F6F7F8F6F70906F7F8F6F70906F7F8F6F70\
906F7F8F6F70906F7F906F7090707F906F7090707F906F7090707F906F709F707F
9070709F707F9070709F707F9F70709F707F9F70709F707F9F70709F7F7F9F70709F7F7F9F7070\
A07F7F9F7070A07F7FA07F70A07F7FA07F70000000000000000000000000000000
0000000000000000000000000000000000000000000000004F3F3050403F4F3F30503F3FA07F6F\
AF7F70A07F6FAF7F70A07F6FAF8070A07F6FAF1F1FA07F6FAF8070A07F6FAF7F6F
A07F6FAF806FA07F6FA07F6FA07F6FAF806FA07F60A07F6FA07F60A07F6FA07F60A07F609F0000\
503F30000000A000009F00009F00001F0000000000000000000000000000000000
0000000000000000005F00001F00000000000000000000001F00003F00008F00008F0000000000\
0000000000000000007F00007F6040000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4000004000004F000000000000000000000050405F4F3F5050405F503F5F5F405F50405F5F405F\
5F405F5F4F605F3F4F6000006000006000006000006F202F604F606F101F6F202F
6F202F604F6070506F6F506F70506F6F506F705F6F70506F705F6F70506F7F5F70705F6F7F5F70\
7F5F6F40303F3F2F3000000040303F806070806070806070806070806070806070
8F6F7F8060708F6F7F8F6F708F6F7F8F6F70906F7F8F6F7F906F7F8F6F7090707F8F6F7F90707F\
906F7090707F906F7F90707F906F709F707F90707F9F707F90707F9F707F90707F
9F707F90707F9F7F7F9F707F9F707F9F707F9F7F7F9F707FA07F7F9F707FA07F7F9F7F7FA07F7F\
A07F7FA07F7FA07F7FA07F7FA07F7FAF807F000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000050403F\
503F3050403FA07F70AF8070A07F6FAF8070A07F6FAF8070A07F6FAF8070A07F6F
AF8070A07F6FAF8070A07F6FAF806FA07F6FAF806FA07F6FAF806FA07F60A0806FA07F60A0302F\
801F100000009F0000A000009F0000400000000000000000000000000000000000
0000000000000000002000007F00000000000000000000000000000000009000008F0000000000\
0000000000000000000F0F0F7F00007F6040000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000300000000000000000000000000000000000000000
3000004000002F00000000000000000F000F4F3F50503F5F4F3F5050405F503F5F5F3F4F50405F\
5F405F5F405F5F2F3F5F000060000060000060000060000060101F602F3F6F3040
602F3F6F506F6F50606F506F6F506070506F6F506F705F6F70506F705F6F70506F3F2F3F3F2F30\
00000000000000000000000040303F7F5F6F8060707F5F6F806070806070806070
8060708F6F708060708F6F7F8F60708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F70906F7F8F6F70\
906F7F8F6F7090707F906F7090707F906F7090707F906F7090707F9070709F707F
9070709F707F9070709F707F9070709F707F9F70709F7F7F9F70709F7F7F9F707FA07F7F9F7070\
A07F7FA07F7FA07F7FA07F70A07F7FA07F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000050403F4F3F30503F3F503F30AF8070A07F6FAF8070A07F6FAF8070
A07F6FAF806FA07F6FAF806FA07F6FA0806FA07F60AF806FA07F60A07F60A07F60A07F6090705F\
A00000000000A000009F00009F00003F0000000000000000000000000000000000
0000000000000000000000007000000000000000000000000000000000008F00008F0000000000\
0000000000000000000000007F00007F0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F0000300000000000000000000000000000000000000000
2F00004000003F000000000000000000000F503F5F4F3F5050405F50405F50405F5000005F4060\
5F405F5F4F605F0000600000600000600000603F506F405060303F6F506F6F4F60
6F506F6F506F6F506F6F506F705F6F70506F705F6F70506F3F303F3F2F303F2F3F000000000000\
00000000000000000000000040303F8060707F6070806070806070806070806070
8F6F7F8F60708F6F7F8F6F708F6F7F8F6F7F906F7F8F6F7F90707F8F6F7F90707F8F6F7F90707F\
906F7F90707F906F7F90707F90707F90707F90707F9F707F90707F9F707F90707F
9F707F9F707F9F707F9F707F9F7F7F9F707F9F7F7F9F707FA07F7F9F707FA07F7F9F707FA07F7F\
A07F7FA07F7FA07F7FAF807FA07F7FAF807F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000503F3050403F503F30AF8070A07F6F
AF8070AF7F6FAF8070A07F6FAF806FA07F6FAF806FA07F60AF806FA07F60A0806FA07F60A0806F\
801F102000007F0000A000009F0000600000000000000000000000000000000000
0000000000000000000000002000002000000000000000000000000000009000007000000F0F0F\
0000000000000000000F0F0F7F60407F00007F0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000300000200000000000000000000000000000000000000000
4000004000004000000000000000000F0F0F4F3F50503F5F4F3F5050405F501F2050405F50405F\
5F40605F0F105F3F4F5F3F4F603F50603F4F604F60604F606F506F604F606F506F
6F4F606F506F6F50606F506F6F506F3F2F30302F303F2F30000000000000000000000000000000\
00000000000000000000000040303F7F5F6F806070806070806070806070806070
8060708F6F7F8060708F6F7F8F6F708F6F7F8F6F708F6F7F8F6F70906F7F8F6F70906F7F8F6F70\
90707F8F6F7090707F906F7090707F906F7090707F90707F9F707F9070709F707F
90707F9F707F9070709F707F9F707F9F7F7F9F70709F7F7F9F707FA07F7F9F707FA07F7F9F7F7F\
A07F7FA07F70A07F7FA07F7FAF7F7FA07F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000050403F
503F3050403F4F3F30504030503F305040304F3F30A0806FA07F60A0806FA07F60A07F60A07F60\
60302F7F00006000009F0000A000007F0000000000000000000000000000000000
0000000000000000000000000000005F00000000000000001F00002000008F00008F0000000000\
0000000000000000000000007F60407F00007F0000705F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003000003F0000000000000000000000000000000000000000000000
4000004000004F00001000000000004F3F5050405F4F3F5F50405F503F5F5F000050405F5F2F3F\
5F2F3F604F605F4F60604F60604F606F506F604F606F506F604F606F506F6F506F
6F506F6F506F3F2F3F302F303F2F3F000000000000000000000000000000000000000000000000\
00000000000000000000000040303F8060708060708060708060708F6F7F806070
8F6F7F8F60708F6F7F8F6F7F8F6F7F8F6F7F90707F8F6F7F90707F8F6F7F90707F906F7F90707F\
906F7F90707F90707F90707F90707F9F707F90707F9F707F90707F9F707F9F707F
9F707F9F707F9F7F7F9F707F9F7F7F9F707FA07F7F9F707FA07F7F9F707FA07F7FA07F7FA07F7F\
A07F7FA07F7FA07F7FAF807FA07F7FAF807F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000504030A07F60AF806FA07F60A0806F\
4F3F30A00000800000A000009F0000800000000000000000000000000000000000
0000000000000000000000000000001F00002000000000000000003F00009000008F0000000000\
0000000000000000000000007F603F8000007F100F7F6040705F3F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000300000000000000000000000000000000000000000100000
3F00004000004000002000000000004F3F504F3F5050405F4F3F5050304F5020305F405F50304F\
5F40605F405F604F605F4060604F60604F6060506F604F606F506F604F60302F30
302030302F30000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000040303F7F5F708060707F5F708060708060708F6F7F
8060708F6F7F8F60708F6F7F8F6F708F6F7F8F6F7F906F7F8F6F70906F7F8F6F7F90707F8F6F70\
90707F906F7F90707F906F7090707F906F7F9F707F906F709F707F90707F9F707F
90707F9F707F9F707F9F7F7F9F707F9F707F9F707F9F7F7F9F707FA07F7F9F707FA07F7F9F707F\
A07F7FA07F7FA07F7FA07F7FAF7F7FA07F7F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005040304F3F30A0806FA07F60\
503F30600000A000009F0000A000009F0000000000000000000000000000000000
0000000000000000000000000000000000005F00000000000000003F00008F0000900000000000\
0000000000000000000000007F60407F00007F1F10705F3F700000705F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F0000000000000000000000000000000000000000000000
4000004000004F00002F00000F0F0F4F3F5050405F4F3F5F50304F5000005F3F4F50405F5F4F60\
5F4060604F605F4F60604F6F604F6060506F604F606F506F302F30302F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000040303F80607080607080607F8060708F6F7F806070
8F6F7F8F6F7F8F6F7F8F6F7F906F7F8F6F7F906F7F8F6F7F90707F906F7F90707F906F7F90707F\
906F7F90707F906F7F9F707F90707F9F707F90707F9F707F90707F9F707F9F707F
9F7F7F9F707F9F7F7F9F707FA07F7F9F707FA07F7F9F707FA07F809F7F7FA07F7FA07F7FA07F80\
A07F7FA07F7FA07F7FAF807FA07F7FAF807F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F3F30504030\
503F30000000000000A000009F0000A00000200000000000000000000000000000
0000000000000000000000000000000000007F00007000000000005F00009000008F0000000000\
0000000000000000000000007F6040801F107F00007F6040705F3F70603F6F5F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000300000000000000000000000000000000000000000000000
3000004000004000000000000000004F3F5F4F3F50503F5F500000500000503F5F5F405F50405F\
5F4F605F405F604F605F4F60604F6F2F2030302F30000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000040303F7F5F7080607080607080607F8060708F607F
8060708F6F7F8F60708F6F7F8F6F7F8F6F7F8F6F708F6F7F8F6F7F90707F8F6F7F90707F906F7F\
90707F906F7F90707F906F7F90707F906F7F90707F90707F9F707F90707F9F707F
90707F9F707F9F707F9F707F9F707F9F7F7F9F707F9F7F7F9F707FA07F7F9F707FA07F7FA07F7F\
A07F7FA07F7FA07F7FA07F7FAF807FA07F7F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504030503F300000000000000000009F0000800000000000000000000000000000
0000000000000000000000000000000000000000003F00001F00007000008F00005F0000000000\
0000000000000000000000007F60407F603F7F0000705F3F7F603F705F30705F3F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0000200000000000000000000000000000000000000000000000
2F00004000003F00000000000F0F0F4F3F5050101F500F10502F3F503F5F5F40605F405F5F4F60\
2F202F302F30302030302F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000040303F80607F80607080607F8060708F6F7F80607F
8F6F7F8F6F7F8F6F7F8F6F7F906F7F8F6F7F90707F8F6F7F90707F8F6F7F90707F906F7F90707F\
906F7F90707F90707F9F707F90707F9F708090707F9F707F90707F9F7F809F707F
9F7F7F9F707F9F7F809F707F9F7F7F9F707FA07F809F7F7FA07F7F9F7F7FA07F80A07F7FA07F7F\
A07F7FAF8080A07F7FAF807FA07F7FAF8080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504030A0806F504030504030000000000000000000400000200000000000000000
0000000000000000000000000000000000000000000000007F00009000009000008F0000000000\
000000000000000000000000000000806F407F00007F5030705F3F7F603F705F3F
705F3F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000300000000000000000000000000000000000000000000000000000
1F00004000004000000000000000004F0F104F304050405F503F5F50405F50405F5F40602F202F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F303F40303F7F607080607080607080607F8060708F6F7F
8060708F6F7F8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F90707F8F6F7F906F7F8F6F7F90707F8F6F7F\
90707F906F7F90707F906F7F90707F90707F9F707F90707F9F707F90707F9F707F
90707F9F707F9F707F9F7F7F9F707F9F7F7F9F707FA07F7F9F707FA07F7F9F7F7FA07F7FA07F7F\
A07F7FA07F7FAF807FA07F7FAF807FA07F7F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504030A07F60A080609F7F60504030503F300000000000000000001F0000200000
0000000000000000000000000000000000000000000000003F0000900000900000900000000000\
0000000000000000000000000000007F60407F0000703F207F603F705F3F705F3F
6F5F30000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000300000000000\
000000000000300000000000000000000000000000000000000000000000000000
1000004000004F00001F00004000004F304050405F503F5F50405F50405F5F40602F202F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040304040303F80607F8060708F6F7F8060708F6F7F80607F
8F6F7F8F6F7F8F6F7F8F6F7F9070808F6F7F90707F8F6F7F907080906F7F90707F906F7F907080\
90707F90707F90707F9F708090707F9F707F90707F9F7F809F707F9F7F809F707F
9F7F809F707F9F7F809F707FA07F809F7F7FA07F809F7F7FA07F80A07F7FA07F80A07F7FA08080\
A07F7FAF8080A07F7FAF8080AF807FAF8080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504030A0806FA07F60A08060A07F60A08060504030000000000000000000000000
0000000000000000000000000000000000000000000000000000007000009000008F0000000000\
000000000000000000000000000000806F407F00007F402F70603F7F603F705F3F
70603F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F30000000000000\
0000002F0000300000000000000000000000000000000000000000000000000000
4000004F0000400000100000000000503F5F4F3F5050405F503F5F2F2030000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040303F40303F7F5F7080607080607080607F8060708F6F7F
80607F8F6F7F8F6F7F8F6F7F8F6F7F906F7F8F6F7F906F7F8F6F7F90707F8F6F7F90707F906F7F\
90707F906F7F90707F90707F9F707F90707F9F707F90707F9F707F90707F9F707F
9F707F9F7F7F9F707F9F7F7F9F707FA07F7F9F707FA07F7F9F7F7FA07F809F7F7FA07F7FA07F7F\
A07F7FA07F7FAF7F7FA07F7FAF807FA07F7F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
50403FA07F60A08060A07F60A080609F7F5FA07F609F7F5F504030000000000000
0000000000000000000000000000000000000000000000000000003F00003F0000700000000000\
0000000000000000000000000F0F007F603F7F00007F00007F603F705F3F705F3F
6F5F30000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000300000000000000000\
0000001000002F0000000000000000000000000000000000000000000000000000
4F00004000001000000000000F0F0F4F3F5050405F50405F2F2030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004030407F607080607F806070806F7F8060708F6F7F8F6F7F
8F6F7F8F6F7F9070808F6F7F90707F8F6F7F907080906F7F907080906F7F90708090707F907080\
90707F9F708090707F9F708090707F9F70809F707F9F70809F707F9F7F809F707F
9F7F809F707FA07F809F7F7FA07F809F707FA07F809F7F7FA07F80A07F7FA07F80A07F7FA08080\
A07F7FAF8080A07F7FAF8080AF7F7FAF8080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F403FA07F60A08060A07F60A08060A07F60A080609F7F5FA08060504030
0000000000000000000000000000000000000000000000000000000000000000003F0000000000\
000000000000000000000000000000806F407F60407F000070603F7F603F705F3F
70603F600000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F30300F0F000000000000\
2F00000F00002000000000000000000000000000000000000000000000003F0000
0F00001000000000000000004F3F504F3F5F201F2F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000003F303F80607080607080607F80607080607F8060708F6F7F
80607F8F6F7F8F6F7F8F6F7F8F6F7F906F7F8F6F7F90707F8F6F7F90707F906F7F90707F906F7F\
90707F90707F90707F90707F90707F90707F9F708090707F9F707F9F707F9F7F80
9F707F9F7F7F9F707F9F7F809F707F9F7F7F9F707FA07F809F707FA07F80A07F7FA07F80A07F7F\
A07F7FA07F7FAF8080A07F7FAF807FAF7F7F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000504030A080609F7F60A080609F7F60A080609F7F5FA07F609F7F5F9F7F60
4F3F2F504030000000000000000000000000000000000000000000000000000000200000000000\
0000000000000F0F000000000000007F603F7F6F407F00007F603F705F3F70603F
6F5F306F5F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F30300000000000000000\
0000003F00000F00000000002000000000000000000000000000003F00000F0000
00000000000000000000000050405F201F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040304080607080607F8060708F6F7F80607F8F6F7F8F6F7F
8F6F7F8F6F7F906F808F6F7F9070808F6F7F9070808F6F7F90708090707F90708090707F9F7080\
90707F9F708090707F9F708090707F9F708090707F9F7F809F707F9F7F809F707F
9F7F809F707FA07F809F707FA07F809F7F7FA07F809F7F7FA07F80A07F7FA07F80A07F7FAF8080\
A07F7FAF8080A07F7FAF8080AF807FAF8080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F403FA07F60A0806FA07F60A080609F7F60A080609F7F60A080609F7F5F
A080609F7F5F5040304F3F2F000000000000000000000000000000000000000000000000000000\
0000000000005000000000000000007F6F407F60407F000070603F7F603F705F3F
70603F6F5F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003030302F2F2F0000000F0000\
0000003000000000000000003F00001000002000003F00001000000F0000000000
0000000000000000002F202F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040303F80607F7F607080607F8060708F6F7F8060708F6F7F
8F6F7F8F6F7F8F6F7F8F6F7F8F6F7F9070808F6F7F90707F8F6F7F9070808F6F7F90707F90707F\
90708090707F90707F90707F9F708090707F9F707F90707F9F7F8090707F9F7F80
9F707F9F7F809F707F9F7F809F707FA07F809F707FA07F809F7F7FA07F80A07F7FA07F80A07F7F\
AF8080A07F7FAF8080A07F7FAF8080AF7F7F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000504030A08060A07F60A080609F7F60A07F609F7F5FA07F609F7F5F9F7F60
9F7F5F9F7F5F9F7F5F9F7F5F4F3F2F504030000000000000000000000000000000000000000000\
0000000000000000000000000000007F603F7F6F407F00007F0000705F3F70603F
705F306F5F3F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F3F2F302F302F0F10300000000000\
0000003F00000000000000000F00003F00003F0000100000000000000000000000
0000000000002F2030403F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040304080607080607F80607F8F6F7F80607F8F6F7F8F6F7F
8F6F7F8F6F7F9070808F6F7F9070808F6F7F90708090707F90708090707F90708090707F907080\
90707F9F708090707F9F708090707F9F7F809F707F9F7F809F707F9F7F809F707F
9F7F809F707FA07F809F7F7FA07F809F7F7FA07F80A07F80A07F80A07F7FA08080A07F7FAF8080\
A07F7FAF8080AF807FAF8080AF807FAF8080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F403FA08060A0806FA07F60A08060A07F60A080609F7F60A080609F7F5F
A07F609F7F5FA07F609F7F5F9F7F5F9F7F5F5040304F3F2F5040304F3F2F000000000000000000\
0000000000000000008F0000000000806F407F60407F402F7F00007F604070603F
70603F6F5F30605030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F2F302F301F101F200000000000\
0000002F00000000000000000000003F00001F0000000000000000000000000000
201F2F20202F4F3F504F3F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040303F80607F80607080607F8060708F6F7F80607F8F6F7F
8F6F7F8F6F7F8F6F7F9070808F6F7F90707F8F6F7F9070808F6F7F907080906F7F907080906F7F\
90708090707F9F708090707F9F708090707F9F708090707F9F70809F707F9F7F80
9F707F9F7F809F707FA07F809F707FA07F809F7F7FA07F809F7F7FA07F80A07F7FA08080A07F7F\
A08080A07F7FAF8080A07F7FAF8080AF807F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000504030A0806FA07F60A08060A07F60A080609F7F5FA07F609F7F5FA07F60
9F7F5F9F7F5F9F7F5F9F7F5F907F5F9F7F5F907F509F7F5F907050907F509070504F3F2F000000\
0000000000000000006F00000F00007F6040806F407F4F307F000070603F70603F
705F30705F3F5F4F30000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000303030302F301F1F1F2F0000000000\
0F00002F00000000000000001F000000000F000000000000000000000000201F2F
4F3F5F4F3F504F3F5F403050000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000403F4080607080607F80607F8F6F7F806F7F8F6F7F8F6F7F
8F6F808F6F7F9070808F6F7F907080906F7F907080906F7F90708090707F90708090707F9F7080\
90707F9F708090707F9F7F809F70809F7F809F707F9F7F809F70809F7F809F707F
A07F809F7F80A07F809F7F7FA07F80A07F80A07F80A07F7FA08080A07F80A08080A07F7FAF8080\
A07F80AF8080AF807FAF8080AF8080AF8080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F403FA08060AF806FA08060A08060A07F60A08060A07F60A080609F7F5F
A080609F7F5F9F7F609F7F5F9F7F5F9F7F5F9F7F5F907F509F7F5F907F509F7F5090705050402F\
0000000000000000003F00006F00000F0F0F7F60407F60407F00007F604070603F
70603F705F30605030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F30303030000000100000000000\
300000100000000000000000000000000000000000000000000000201F2F403F50
4030504F3F5F403F504F3F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000040303F80607F80607080607F80607F8F6F7F80607F8F6F7F
8F6F7F8F6F808F6F7F9070808F6F7F9070808F6F7F907080906F7F907080906F7F90708090707F\
90708090707F9F708090707F9F708090707F9F70809F707F9F7F809F707F9F7F80
9F707F9F7F809F707F9F7F809F7F7FA07F809F7F7FA07F80A07F7FA07F80A07F7FA07F80A07F7F\
AF8080A07F7FAF8080AF807FAF8080AF807F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000504030A0806FA08060A08060A07F60A080609F7F60A080609F7F5FA07F60
9F7F5F9F7F609F7F5F9F7F5F9F7F5F9F7F5F9070509F7F5F9070509F7F509070509070508F704F\
4F3F2F000000000000000000800000300000806F407F603F7F000070603F70603F
705F3F705F3F600000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000303030302F300000001000001F0000\
300000000000000000000000000000000000000000201F2F403050403F50403F50
4F3F5F4F3F504F3F5F4F3F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004030408060708F6F7F80607F8F6F7F80607F8F6F808F6F7F
8F6F808F6F7F9070808F6F80907080906F7F90708090708090708090707F9F70809070809F7080\
90707F9F7F809070809F7F8090707F9F7F809F70809F7F809F7080A07F809F7F80
A07F809F7F80A07F809F7F80A07F809F7F80A07F80A07F80A07F80A07F80AF8080A07F80AF8080\
A07F80AF8080AF8080AF8080AF807FB08F80000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F4F3FA08060AF806FA08060A0806FA08060A08060A07F60A080609F7F60
A080609F7F5FA080609F7F5F9F7F5F9F7F5F9F7F5F9F7F5F9F7F5F9070509F7F5F907050907F50\
8F704F4F402F000000000000800000800000806F40806F407F00007F000070603F
70603F705F306F2F1F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F30302F30000000000000100000\
300000000000000000000000000000000000201F2F3F304F403050403050403F50
4030504F3F5F4F3F504F3F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000040304040303F80607F8060708F6F7F80607F8F6F7F806F7F8F6F7F
8F6F7F8F6F808F6F7F9070808F6F7F9070808F6F7F907080906F7F90708090707F90708090707F\
90708090707F9F708090707F9F708090707F9F7F8090707F9F7F809F707F9F7F80
9F707F9F7F809F7F7FA07F809F7F7FA07F809F7F80A07F809F7F7FA07F80A07F7FAF8080A07F7F\
AF8080A07F7FAF8080AF807FAF8080AF807F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000504030AF806FA07F60A08060A07F60A080609F7F60A080609F7F5FA08060
9F7F5F9F7F609F7F5F9F7F5F9F7F5F9F7F5F9F7F5F9F7F5F9070509F7F50907050907F508F704F\
9070508F704F4F3F2F0000000000001F0000801F107F603F7F10107000007F603F
705F30705F3F6050305F0000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000303030302F300000000000002F0000\
3000000000000000000000001F1F2F3F30503F304F403F50403050403F50403050
4F3F5F4F3F504F3F5F403F50000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000040303F4F3F4080607F8F6F7F80607F8F6F7F8F6F7F8F6F808F6F7F
9070808F6F7F9070808F6F7F9070809070809070809070809070809070809070809070809F7F80\
9070809F7F809070809F7F809F70809F7F809F70809F7F809F7F809F7F809F7F80
A07F809F7F80A07F809F7F80A07F80A07F80A07F80A07F80A08080A07F80AF8080A07F80AF8080\
AF8080AF8080AF8080AF8080AF8080B08080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F4F3FA08060AF806FA08060A0806FA08060A08060A08060A080609F7F60
A080609F7F5FA080609F7F5FA080609F7F5F9F7F5F9F7F5F9F7F5F907F509F7F50907050907F50\
90704F9070508F704F0000000000000F0F0F805030806F407F3F2F7F000070603F
7F603F705F3F6F5F3F5F4F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F3030303F0000000000000F0000\
2F0000000000000000302F4F3F30503F2F4F3F30503F3050403F50403050403F50
4030504F3F5F4F3F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000040304040303F80607F806070806F7F80607F8F6F7F80607F8F6F7F
8F6F7F9070808F6F7F9070808F6F7F9070808F6F7F90708090707F90708090707F90708090707F\
9F708090707F9F70809070809F7F8090707F9F7F809F70809F7F809F707F9F7F80
9F7080A07F809F707FA07F809F7F80A07F809F7F7FA07F80A07F80A08080A07F7FA08080A07F80\
AF8080A07F7FAF8080AF8080AF8080AF807F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000504030A0806FA08060A0806FA07F60A08060A07F60A080609F7F5FA08060
9F7F5FA080609F7F5F9F7F5F9F7F5F9F7F5F9F7F5F9F7F5F907F509F7F5F907050907F5090704F\
9070508F704F8F704F806F4F0000000000007F60407F50307F2F1F7000007F603F
70603F70603F6F5F305F4F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000303030302F30000000000000200000\
000000000000302F403F304F3F304F4030503F3050403F504030504F3F5F403F50
4F3F5F4F3F504F3F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000040303F403F4080607F8F6F7F80607F8F6F7F8F6F7F8F6F808F6F7F
9070808F6F7F9070809070809070809070809070809070809070809070809F70809070809F7080\
9070809F7F809F70809F7F809F70809F7F809F70809F7F809F7080A07F809F7F80
A07F809F7F80A07F80A07F80A07F80A07F80A08080A07F80A08080A07F80AF8080A08080AF8080\
AF8080AF8080AF8080AF8080AF8080B08F80000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F4F3FA0806FAF806FA08060AF806FA08060A08060A08060A08060A08060
A080609F7F5FA080609F7F5FA080609F7F5F9F805F9F7F5F9F7F5F907F509F7F5F907F50907F50\
90704F9070508F704F8F704F403F200000007F6040806F407F50307F00007F3F20
7F603F705F3F70603F5F4F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003030302F2F301F1F1F000000000000000000\
0000001F1F203F304F302F4F3F30503F304F4030503F30504030504030504F3F5F
4030504F3F5F403050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000040304080607080607F80607F8F6F7F80607F8F6F7F8F6F7F8F6F80
8F6F7F8F70808F6F7F9070808F6F7F90708090708090708090707F9070809070809F708090707F\
9F70809070809F7F8090707F9F7F809F70809F7F809F70809F7F809F70809F7F80
9F7080A07F809F7F80A07F809F7F80A07F80A07F80A07F80A07F80A08080A07F80AF8080A07F80\
AF8080AF8080AF8080AF807FAF8080AF8080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000504030AF806FA08060A0806FA08060A08060A07F60A08060A07F60A08060
9F7F5FA080609F7F5FA080609F7F5F9F7F5F9F7F5F9F7F5F907F509F7F5F9070509F7F5090704F\
9070508F704F8F704F806F4F8F704F403020806F407F60407F6F407F00007F1F10
70603F70603F6F5F306050304F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3030303F1F1F1F000000000000000000\
0000003F304F302F403F30503F30504030503F3050403F504030504F3F5F403F50
4F3F5F4F3F5F4F3F50000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000040303F806F7F80607F8F6F7F80607F8F6F808F6F7F8F6F808F6F7F
9070808F6F809070809070809070809070809070809070809F70809070809F70809070809F7F80\
9070809F7F809F70809F7F809F70809F7F809F70809F7F8090707F9F7F8090707F
9F7F809F707F9F7F809F707F9F7F809F7F7F9F7F809F7F7FA07F809F7F7FA07F80A07F7FA08080\
A07F7FA08080A07F7FAF8080A0807FAF80800000000F0F0F0F0000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F4F3FA0806FAF8F6FA08060AF806FA08060AF806FA08060A08060A08060
A08060A08060A080609F7F5FA080609F7F5F9F805F9F7F5F9F7F5F9F7F509F7F5F907F509F7F50\
9070509070508F704F8F704F8F6F4F8F704F3F3020806F407F6F407F6F407F0000
7F604070603F70603F60301F400000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F30302F301F1F1F000000000000000000\
1F1F20302F403F304F3F2F4F3F30503F304F403050403050403F504030504F3F5F
403F504F3F5F403050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000040304080607F806F7F80607F8F6F7F80607F8F6F808F6F7F8F6F80
8F6F7F9070808F6F7F8F6F7F8F6F7F8F70808F6F7F90707F8F6F7F9070808F6F7F90707F906F7F\
907080906F7F90707F90707F90708090707F0F000F0000000F0F0F0000000F000F
0000000F0F0F0000000F000F0000000F0F0F0000000F000F0000000F0F0F0000000F000F000000\
0F0F0F0000000F0F0F0000000F0F0F000000AF807FAF807FAF80800000000F0F00
0000000F0F0F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000504030AF806FA08060AF806FA08060A08060A08060A080609F7F60A08060
9F7F5FA080609F7F5F9F80609F7F5F9F805F9F7F5F9F7F5F9F7F509F7F5F907050907F50907050\
907F508F704F8F704F8F6F4F8F704F806F400F00003F3020806F407F603F7F5030
70100F70603F605030600000500000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3030303F000000000000000000000000\
302F403F304F3F304F3F30503F304F403050403050403F504030504F3F5F403F50
4F3F5F4F3F5F4F3F5F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000040304080607F7F6070806F7F80607F806F7F80607F8F6F808F6F7F
8F6F808F6F7F0F0F0F0000000F0F0F0000000F0F0F0000000F0F0F0000000F0F0F0000000F0F0F\
0000000F0F0F0000000F0F0F0000000F0F0F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08080B08F80AF8080AF8080AF807F
AF8080AF807F0F0F0F0F00000F0F0F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000050403FAF8F6FA08060AF8F6FA08060A0806FA08060A0806FA08060
A080609F8060A080609F805FA080609F7F5FA080609F7F5F9F7F5F9F7F509F7F5F907F509F7F50\
907050907F508F704F8F704F8F6F4F8F704F7F6F400000003F3020806F407F603F
7F503070100F70000060402F2F2F1F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003030302F2F30000000000000000000302F3F\
303040302F403F304F3F2F4F3F30503F30504030503F3050403F504030504F3F5F
403F504F3F5F403050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004030400000000F00000000000F0F0F0000000F00000000000F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F80AF8080B08F80AF8080B08F80
AF807FB08F80AF807FAF807FAF807F0F0F000000000F0F0F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F4F3FA08060AF806FA08060A0806FA08060A08060A08060A08060
9F7F5FA080609F7F5FA080609F7F5F9F805F9F7F5F9F805F907F509F7F5F907F509F7F50907050\
907F508F704F9070508F6F4F8F704F806F40806F400000000000007F603F7F6F40
70603F705030704F2F6F5F30504F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3030303F0000000000001F1F203F3040\
302F403F304F3F2F4F3F30503F3050403050403050403F5F4030504F3F5F403F50
4F3F5F4F3F5F4F3F5F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F0000000F0F0F\
802F30802F304F2F30000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08080B08F8FB08F80B08F80B08080
B08F80B08F80B08F80B08080B08F80AF807FAF807FAF807FAF807F0F00000F0F0F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000504030AF8F6FA0806FAF8F6FA08060AF8F6FA08060A08060A08060
A08060A08060A080609F7F5FA080609F7F5FA080609F7F5F9F805F9F7F5F9F7F5F907F509F7F50\
907F50907F508F704F9070508F704F8F704F7F6F400000000000000000007F603F
7F6F4070603F70603F6F5F305F1010000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F30302F30000000000000303040302F40\
3F3040302F403F304F3F2F4F3F30503F304F403050403050403F5F4030504F3F5F
4F3F504F3F5F403050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F000F7F202F7F2F307F0000\
802F307F203040303F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F80AF8080B08F80AF8080B08F80
AF8080B08F80AF807FB08F80AF807FB08F80AF807FB08F80AF807FAF807FAF8070AF807FAF807F\
0F0F0F0000000F0F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F4F3FA0806FAF8F6FA08060A0806FA08060A08060A08060A08060
A08060A080609F7F5FA080609F7F5FA080609F7F5F9F7F5F9F7F5F9F7F5F907F509F7F50907F50\
907F508F704F90704F8F704F8F704F806F40806F400000000000000000003F3020
70603F7F603F705F306F201F5F4F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3030303F0000003F3040302F403F3040\
302F403F304F3F304F3F30503F3050403F50403050403F5F4030504F3F5F4F3F5F
4F3F5F4F3F5F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F0000000F0F0F7F20307F2F307F20308000007F0000803F4F\
80607F806F7F40303F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F80B08F80B08F80B08F80B08F80
B08F80B08F80B08F80B08F80B08F80B08F80B08F80B08F80B08F80B08F7FB08F80B08F7FB08F80\
AF807FB08F7FAF807F0F0F0F0000000F0F0F0F0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F403FAF8F6FA0806FAF8F6FA08060AF8F6FA08060AF8F6FA08060
A08060A08060A08060A08060A080609F7F5FA080609F7F5F9F805F9F7F5F9F805F9F7F509F7F50\
907F50907F5090704F9070508F704F8F704F806F40000000000000000000000000
3F302070603F70603F6F5F30605030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F0000301F2010101F1F1F20302F3F303040302F40\
3F3040302F403F304F3F2F4F3F30503F3050403F50403050403F5F4030504F3F5F
4F3F504F3F50000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F000F70202F7F2F3070202F7F00007F00007F00007F00008000007F6070\
80607F80607F40303F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F80AF8080B08F80B08080B08F80
AF8080B08F80B08080B08F80AF807FB08F80B0807FB08F80AF807FB08F80B0807FB08F80AF807F\
B08F7FB08F7FB08F7FAF8070AF807FAF8070AF807F0000000F0F000000000F0F0F
0000000F0F000000000F0F0F0000000F0F000000000F0F0F000000000000000000000000000000\
0000000000005F4F3FA08060AF8F6FA08060AF8F6FA08060A08060A08060A08060
A08060A080609F8060A080609F7F5FA080609F7F5F9F805F9F7F5F9F7F5F907F509F7F5F907F50\
907F5090704F907F508F704F8F704F806F40806F40000000000000000000000000
0000003F301F302F1F3F301F5F4F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000300000302F301F1F20302F3F303040302F403F3040\
3030403F304F3F304F4030503F3050403F504030504F3F5F403F504F3F5F4F3F5F
4F405F4F3F50000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F7020307F2F307020307F00007F00007F00007F00008000007F00008000007F2F3080607F\
80607F40304040303F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F80B08F8FB08F80B08F80B08F80
B08F80B08F80B08F80B08F80B08F80B08F80B08F80B08F80B08F80B08F80B08F80B08F7FB08F80\
B08F7FB08F80B08F7FB08F80B08F7FB08F7FB08F7FB08F7FAF8070AF8F7FAF8070
B08F7FAF8070AF8F70AF8070AF8F70AF8070AF8F70AF806FAF8F70AF806F0F0F0F000000000000\
00000000000050403FAF8F6FAF806FAF8F6FA08060AF8F6FA08060AF8F6FA08060
A0806FA08060A08060A08060A080609F805FA080609F7F5FA0805F9F7F5F9F805F907F509F7F5F\
907F50907F5090704F907F508F704F8F704F806F400F0F0F000000000000000000
00000000000000000000000060201F4F0000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2020302F301F101F30303F302F3F303040302F40\
3F304F302F403F304F3F30504030503F30504030504030504F3F5F4030504F3F5F
4F3F5F4F3F5F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F000F6F506F704F5F
70202F703F4F7010107000007000007F00007000007F00007F00007F00007F000080607F7F607F\
80607F3F2F3F40303F000000000000000000000000000000000000000000000000
0000000000000000002070001F70002070001F70002070001F5F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F80B08F80B08F80AF8080B08F80
B08F80B08F80B08080B08F80B08F80AF807FAF807FAF807F408F1F2F8F008F804F208F1F208F00\
B08F80B0807FB08F7FB08F7FB08F7FB0807FB08F7FB08F7FB08F7FB08F7FB08F7F
B08F70B08F7FAF8F70B08F70AF8F70B08F70AF8F70B08F70AF8F70B08F70A0806F0F0F00000000\
0000000000005F4F3FAF806FAF8F6FA08060AF806FA08060AF806FA08060A08060
A08060A080609F805FA080609F805FA080609F7F5F9F805F9F7F5F9F805F907F509F7F5F907F50\
9F7F5090704F907F508F704F8F704F806F408F704F000000000000000000000000
000000000000000000000000600000400000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000030303F302F3030303F302F3F3F3040302F403F3040\
302F403F304F3F304F4030503F3050403F5F403050403F5F403F504F3F5F4F3F5F
4F3F5F4F3F50000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F0000006F50706F506F705F7F6F506F
703F4F705F707F5F7F7000007F00007F00007F00007F00007F00007F2030802F3080607F80607F\
80607F40304040303F000000000000000000000000000000000000000000000000
0000000000000000001F70002070001F7000207F00207000206000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F80B08F8FB08F80B08F8FB08F80
B08F80B08F80B08F80B08F80B08F804F8F1F2F8F20208F1F2F8F00208F002F8F00208F002F8F00\
208F002F8F00208F002F8F00B08F7FB08F80B08F7FB08F7FB08F7FB08F7FB08F7F
B08F7FB08F70B08F7FB08F70B08F7FB08F70B0907FB08F70B08F70AF8F70B08F70AF8F70AF8F6F\
0000000F0F0F5F4F3FAF8F6FAF806FAF8F6FAF806FAF8F6FA08060AF8F6FA08060
A08060A08060A08060A08060A080609F805FA080609F805F9F805F9F7F5F9F805F9F7F509F7F5F\
907F509F7F50907050907F508F704F8F704F80704F0F0F0F000000000000000000
0000000000000000000000006F5F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F30000000302F3030303F302F3F303040302F40\
3F304F302F4F3F30503F304F4030503F3050403F5F403050403F5F403F504F3F5F
4F3F5F4F3F5F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F000F0000006F506F60506F6F5F706F5070705F706F5070705F70
705070705F70705F707F00007000007F00007F00007F00007F0000705F707F607F80607F7F607F\
80607F3F303F403040000000000000000000000000000000000000000000000000
0000000000000000002070001F70002070001F7000207F001F5F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F8FB08080B08F80B08F80B08F80
B08080B08F80B08F80B08F806F8F4F208F00208F002F8F00208F00208F00208F002F8F00208F00\
208F00208F002F8F00208F00208F00208F002F8F00B08F7FB08F7FB08F7FB08F7F
B08F70B08F7FB08F70B08F7FAF8F70B08F70AF8F70B08F70AF8F70B08F70AF8F6FB08F70AF8F6F\
AF806FA080605F4F3FA08060AF8F6FA0806FAF8F6FA08060A0806FA08060A0806F
A08060A08060A08060A080609F805FA080609F7F5FA080609F7F5F9F805F9F7F509F7F5F907F50\
9F7F50907050907F508F704F8F704F8F704F8F704F000000000000000000000000
000000000000000000000000605030504F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001F1F1F00000030303F302F3F3F3040302F403F304F\
30304F3F304F3F304F4030503F3050403F5F4030504F3F5F403F5F4F3F5F4F3F5F
5040604F3F500000000000000000000000000000000000000F0F0F0000000F0F0F00000060506F\
604F6F60506F604F6F6F50706F50706F5F706F5070705F7F705070705F7F705F70
705F7F705F707F3F4F7000007F00007F00007F00007F0000802F307F607F80607F7F607F806F80\
80607F40304040303F000000000000000000000000000000000000000000000000
0000000000000000001F7000207F00207000207F001F7000206000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B08F80B08F8FB08F80B08F8FB08F80
B08F8FB08F80908F50B08F802F8F00208F002F8F00208F002F8F00208F002F8F00208F002F8F00\
208F002F8F00208F002F8F00208F002F8F00208F002F9000208F002F8F00208F00
2F9000208F002F8F006F8F30709030A0806FAF8070B08F70B09070B08F70B08F70AF8F6FB08F70\
AF8F6FAF8F705040300F0F0FA08060A0806FA0806FAF8F6FA0806FAF8F6FA08060
AF8F6FA08060A08060A08060A08060A08060A080609F805FA080609F7F5F9F805F9F7F509F805F\
907F509F7F5090704F907F508F704F8F704F80704F806F4F000000000000000000
0000000000000000000000000000005F4F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F3F30303F302F3F303040302F40\
3F304F302F4F3F30503F30504030503F3050403F504030504F3F5F4030504F3F5F
4F3F5F4F3F5F0000000000000000000000000000005F1F2F5F1F2F60202F5F1F2F603040604F6F\
605070604F6F6F50706F50706F50706F50706F50706F5070705F706F5070705F70
705F7070101F7000007F00007000007F00007F00007F00007F4F6080607F7F5F7F80607F7F607F\
80607F3F303F403040000000000000000000000000000000000000000000000000
0000000000000000002070001F7000207F001F70002070001F5F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B08F80B08F8FB08F80B08F80
B08F80B08F806F8F4F908F50208F002F8F00208F00208F00208F004F8F1F208F004F8F1F8F8F60\
B08F80208F1F4F8F306F8F4F2F8F00208F00208F00208F002F8F00208F00208F00
208F002F8F00208F00208F00208F002F8F00208F008F8F4F8F8F40B08F70AF8F6FB08F70AF8F6F\
AF8F70AF8F6F5040300000000F0F0F000000A08060A08060AF8F6FA08060A08060
A08060A08060A08060A080609F8060A080609F7F5FA0805F9F7F5F9F805F9F7F509F7F5F907F50\
9F7F50907050907F508F704F90704F8F704F8F704F7F6F40000000000000000000
0000000000000000000000000000005F5030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000030303F302F3F3F3040302F403F304F\
302F4F3F30503F30504030503F3050403F5F4030504F3F5F403F5F4F3F5F4F3F5F
50405F4F3F500000000000000000000000005F202F5F00005F00005F0000604F6F604F6F605070\
604F6F6F50706F50706F50706F5070705F7F6F5070705F7F6F5070705F7F705F70
704F607000007F00007000007F00007F00007F00007F101F80607F7F607F80608080607F806080\
80607F40304040303F000000000000000000000000000000000000000000000000
000000000000000000207000207F001F7000207F00207000206000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B08F8FB08F80BF8F8FB08F80
B08F8FB08F804F90206F8F502F8F00208F002F9000208F002F8F00AF807F4F9020AF807FB08F80\
B08F80BF9080B08F80BF8F806F8F4F7090506F8F4F708F306F8F302F9000208F00
2F9000208F002F9000208F002F9000208F002F9000208F002F90006F8F2FB09070AF8F70B08F70\
AF8F6FB08F705040300000000000000000000F00000F0F0FA08060AF8F6FA08060
AF8F6FA08060A08F6FA08060A08060A08060A080609F805FA080609F7F5F9F805F9F7F5F9F805F\
907F509F7F50907F50907F508F704F907F4F8F704F806F4F000000000000000000
0000000000000000000000000000005F4F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F20302F3F30303F302F3F3F3040302F40\
3F304F3F2F4F3F30503F3050403050403050403F5F4030504F3F5F4F3F5F4F3F5F
4F3F5F000000000000000000501F205F1F2F5000005F00005F2F3F5F2F405F4F6F604F6F604F6F\
6050706050706F50706F50706F50706F5070705F706F5070705F70705070705F7F
70202F7000007000007F00007000007F00007F101F7F4F607F5F7F80607F7F607F80607F7F607F\
80607F3F303F403040000000000000000000000000000000000000000000000000
0000000000000000002070001F70002070001F7000207F001F5F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B08F80B08F8FB08F80B08F8F
B08F80B08F804F8F1F4F8F1F208F00208F00208F002F8F006F8F4FB08F80AF807FB08F80B08F7F\
B08F80B08F7FB08F80B08F7FB08F80B08F7FB08F80B08F7FB08F7FB08F7FB08F7F
208F00208F00208F002F8F00208F00208F00208F002F9000208F00208F008F8F40B08F70AF8F6F\
AF8F70AF8F6F50403F0000000000000000000000000000000F0F00000000A08060
A08060A08060A08060A080609F805FA080609F805FA080609F7F5F9F805F9F7F5F9F805F907F50\
9F7F50907F50907F508F704F90704F8F704F8F704F7F6F40000000000000000000
0000000000000000000000000000005F5030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F1F303040302F3F3F3040302F403F304F\
3F304F3F30503F3050403F5F403050403F5F4030504F3F5F4F3F5F4F3F5F4F3F5F
5040600000000F000F50304F5F202F5030505F00005F0000603F505F4F6F604F6F604F6F6F5070\
6050706F50706F50706F5F7F6F5070705F7F6F5070705F7F705F70705F7F6F506F
7F00007000007F00007000007F00007F00007F50607F607F8060807F607F80608080607F806F80\
80607F403040000000000000000000000000000000000000000000000000000000
0000000000000000001F7000207F00207000207F001F7000206000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000BF908FB08F8FBF8F8FB08F80
BF908FB08F804F8F204F8F1F2F9000208F002F8F00208F0090906FB08F80BF9080B08F80BF9080\
B08F80BF9080B08F80BF9080B08F80BF9080B08F7FBF9080B08F7FBF9080B08F7F
BF907FAF80706F9030208F002F9000208F002F9000208F002F9000208F002F9000AF8F70B09070\
AF8F70B08F70504030000000000000000000000000000000000000000000000000
0F0F0F9F8060A08F6FA08060A08F60A08060A080609F805FA080609F805F9F805F9F7F509F805F\
9F7F509F7F50907F50907F508F704F907F4F8F704F80704F000000000000000000
000000000000000000000000000000000000500000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F20302F3F303040302F3F303040302F40\
3F304F3F2F4F3F30503F305040305F403050403F5F4030504F3F5F403F5F4F3F5F
4030500000004F3F5F50304F5040605F3F50501F205F00005F2F3F604F6F5F4F6F604F6F604F6F\
6F50706050706F50706F50706F50706F5070705F706F5070705F7F6F5070702030
7000007F00007000007F00007000007F101F7F5F7F7F607F7F607F80607F7F607F80607F80607F\
806F803F303F000000000000000000000000000000000000000000000000000000
0000000000000000002070001F7000207F001F7000207F001F5F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B08F80B08F8FB08F80B08F8F
B08F80B08F8F408F1F4F8F1F208F002F8F00208F00208F00B08F80B08F80B08F80B08F80B08F80\
BF8F80B08F7FB08F80B08F7FBF8F80B08F7FB08F80B08F7FB08F80B08F7FB08F7F
B08F7FB0907FB08F70908F4F8F8F402F9000208F00208F00208F002F9000208F00AF8F6FAF8F6F\
B08F70AF8F6FAF8F70504030000000000000000000000000000000000000000000
0000000F0F0F9F7F5F9F8060A08060A080609F805FA080609F805F9F805F9F7F509F805F9F7F50\
9F7F5F907F50907F5090704F907F508F704F8F704F806F400F0F00000000000000
0000000000000000000000000000000000001F0000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3F303040302F3F3F30403030403F304F\
3F304F3F30503F305040305F403050403F5F403F5F4F3F5F4F3F5F4F3F604F3F5F
0F0F0F4F3F5F5040605040605F40605040605F40605F00006000005F4060605070604F6F6F5070\
6050706F50706F50706F5F706F5070705F7F6F5070705F7F705F70705F70700000
7F00007000007F00007F00007F00007F3F4F8060807F607F8060807F607F806F8080607F806F80\
80607F403F40000000000000000000000000000000000000000000000000000000
000000000000000000207000207F001F7000207F00206000206000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
003F00003000003F00003000003F00003000003F00003000003F00003000003F00003000003F00\
003000002000001F00000000000000000000000000BF908FB08F80BF908FB08F8F
BF908FB08F807090506F8F502F9000208F002F9000208F00BF9080B08F80BF9080B08F80BF9080\
B08F80BF9080B08F80BF9080B08F80BF9080B08F80BF9080B08F7FBF9080B08F7F
BF907FB08F7FBF907FB08F7FB0907F6F8F4F70904F208F002F9000208F002F90006F8F406F904F\
AF8F70B09070AF8F6F5F4F3F000000000000000000000000000000000000000000
0000000000000000000F0000A080609F7F5F9F8060A0805FA080609F805FA0805F9F805F9F805F\
9F7F509F805F907F50907F508F704F907F508F704F8F704F000000000000000000
000000000000000000000000000000000000100000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303F302F3F303040302F403F3040302F40\
3F304F3F304F4030503F3050403050403050403F5F4030504F3F5F403F50000000
000000504060503F5F5040605040605F40605F40605F3F505F0000602F40604F6F605070604F6F\
6F50706050706F50706F50706F50706F5070705F7F6F50706F5070700000700000
7000007000007000007F101F70101F7F607F7F5F7F7F607F7F5F7F80607F7F607F80608080607F\
806F8040303F000000000000000000000000000000000000000000000000000000
000000000000000000207F001F7000207F001F70002060001F5F00003000003000003000002F00\
003000003000003000003000003000003000003000003000003000003000208000
207F00208000207F00208000207F00208000208000208000208000208000208000208F00208000\
208F00003000003F00000000000000000000000000B08F80B08F8FB08F80BF8F8F
B08F80B08F8F908F50908F50208F00208F00208F002F9000A0807FB08F80B08F80BF9080B08F80\
B08F80B08F80BF9080B08F7FB08F80B08F7FBF9080B08F7FB08F80B08F7FB0907F
B08F7FB0907FB08F7FB0907FB08F70B0907FAF8F702F9000208F00209000208F00209000408F10\
B08F70AF8F6FB08F70504030000000000000000000000000000000000000000000
0000000000000000000000000000000F0F000000000F0F00907F5F9F7F5F907F509F7F5F907F50\
907F50907F50907F5090704F907F508F704F8F704F8070400F0000000000000000
0000000000000000000000000000000000002F2010000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3F3F3040302F403F3040302F403F304F\
3F304F3F30503F3050403F5F403050403F5F40305F4F3F5F000000000000000000
0F0F0F503F605040605040605F40605F40605F4F6F5F406060101F601010605070604F6F6F5070\
6F50706F50706F5070705F7F6F5070705F7F6F5070705F7F70304F700000700000
7F00007000007F00007F3F4F7F3F4F7F607F7F607F7F607F80608080607F80608080607F806F80\
806080403040000000000000000000000000000000000000000000000000000000
0000000000000000001F7000207F00207F00207F00207000207F00207F00207F00207F00207F00\
207F00207F00207F00208000207F00208000207F00208000207F00208000207F00
2080002080002080002080002080002080002080002080002F8F002080002F8F002080002F8F00\
208F00003F00003F00000000000000000000000000BF908FB08F8FBF908FB08F80
BF908FB08F8FB08F80AF8F802F9000208F002F9000208F00909050B08F80BF9080B08F80BF9080\
B08F80BF9080B08F80BF9080B08F80BF9080B08F80BF9080B08F7FBF9080B08F7F
BF9080B08F7FBF907FB08F7FBF907FB08F7FB0907F6F8F404F90302090002F9000208F002F9000\
B08F70B09070AF8F6F5F4F3F000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F0000000F0F0F\
0000000F0F0F907F50907F50907F4F907F508F704F8F704F000000000000000000
0000000000000000000000000000000000002F2F1F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303F302F3F303040302F403F3040302F4F\
3F30503F2F4F3F30503F3050403F5F403050403050000000000000000000000000
000000504060503F5F5040605040605F40605040605F4F6F5F2F3F6000005F3F50605070604F6F\
6F50706050706F50706F50706F5F7F6F5070705F7F6F5070700000700000700000
7000007F00007000007F607F7F5F7F7F607F7F5F7F7F607F7F607F8060807F607F80608080607F\
806F803F303F000000000000000000000000000000000000000000000000000000
000000000000000000207F001F7000207F001F7000207F001F7000207F001F7000207F001F7F00\
207F001F7F00207F00207F00207F001F7F00207F00207F002080001F7F00208000
207F00208000207F00208000208000208000208000208000208000208F00208000208F00208000\
208F00003000003F00000000000000000000000000B08F8FBF908FB08F80BF8F8F
B08F80BF908FB08F80BF8F8F208F002F9000208F002F8F004F8F1FBF9080B08F80BF8F80B08F80\
BF9080B08F80BF9080B08F80BF9080B08F7FBF9080B08F7FBF9080B08F7FB0907F
B08F7FB0907FB08F7FB0907FB08F70B0907FB08F706F904F6F8F402F9000208F00209000208F00\
8F905FAF8F6FAF8F70504030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000907F50907F4F907F508F704F8F704F8F704F0F0F00000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000302F3F303040302F403F304F3030403F304F\
3F304F4030503F3050403F5F3F30500F0F0F000000000000000000000000000000
0F0F0F503F5F5040605040605F40605F40605F4F6F5F4F6F5F4F6F600000602030604F706F5070\
6050706F5F7F6F5070705F7F6F5070705F7F705F70703F4F7000007F0000700000
7F00007000007F3F4F7F5F7F7F607F7F607F8060807F607F8060807F607F806F80806080806F80\
806080403F40000000000000000000000000000000000000000000000000000000
000000000000000000207000207F001F7000207F00207F00207F00207F00207F00207F00207F00\
207F00208000207F00208000207F00208000207F00208000206000206F00206000
206F00206000206F00206F00206F00206F002F6F00206F00206F00206F002F7000206F002F7000\
206F00003F00003F00000000000000000000000000BF908FB08F8FBF908FB08F8F
BF908FB08F80BF908FB08F8F70903F208F002F9000208F002F90006F8F50709050B08F80BF9080\
B08F80BF9080B08F80BF9080B08F80BF9080B08F80BF9080B08F7FBF9080B08F7F
BF9080B08F7FBF907FB0907FBF907FB08F7FBF907F6F90306F9030208F002F90002090002F9000\
8F8F5FB09070AF8F705F4F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000907F50907F4F907F508F704F8F704F806F40000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303F302F3F303040302F403F304F302F4F\
3F30503F30504030503F3050000000000000000000000000000000000000000000
000000504060503F605040605040605F40605F40605F4F6F5F406F60202F600000603040604F6F\
6F50706F50706F507F6F50706F5F7F6F5070705F7F6F1010700000700000700000
7000007F0000705F707F607F7F5F7F7F607F7F607F8060807F607F80608080607F806F8080607F\
806F8040303F000000000000000000000000000000000000000000000000000000
0000000000000000002060001F5F002060001F5F002060001F5F002060001F60002060001F6000\
2060002060002060001F6000206000206000206F00000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B08F8FBF908FB08F8FBF908F
B08F80BF908FB08F80BF908FB08F802F9000208F002F9000208F002F9000208F0070903FB08F80\
BF9080B08F80BF9080B08F80BF9080B08F80BF9080B08F7FBF9080B08F7FBF9080
B08F7FB0907FB08F7FB0907FB08F70B0907FB08F702F901F208F10209000208F002F9000208F00\
8F905FAF8F6FB09070504030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000008F704F8F704F8F7F4F8F704080704F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001F1F20302F3F3F30403030403F304F302F403F3050\
3F30503F30500000000F0F0F000000000000000000000000000000000000000000
5040605040605040605040605F4F6F5F40605F4F6F5F4F6F604F6F603F5F6000006010106F5070\
6F50706F5F7F6F5070705F7F6F5070703F4F70304F700000700000700000700000
7F00007F00007F607F7F5F7F7F60807F607F8060807F607F80608080607F806F8080607F8F6F80\
806F80403040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000BF908FBF8F8FBF908FB08F8F
BF908FBF8F8FBF908FB08F80BF908F908F6F90906F208F002F9000208F002F9000208F00B08F80\
B08F7FBF9080B08F80BF9080BF9080BF9080B08F80BF9080B09080BF9080B08F7F
BF9080B0907FBF907FB0907FBF907FB0907F90904F8F904F2F90002090002F90002090002F9000\
8F905FB09070AF8F705F4F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000008F704F8F704F8F7F4F806F40000000000000
0000000000000000000000000000000000000000002F201F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001F101F30303F302F3F303040302F40303040302F4F\
3F3050000000000000000000000000000000000000000000000000000000000000
4F3F5F504060503F605040605040605F406F5F40605F4F6F5F4F6F604F6F602F4060000060405F\
6F50706F50706F50706F5070705F7F6F1010700000700000700000700000700000
70304F705F70705F7F7F607F7F5F7F7F60807F5F7F8060807F607F8060807F607F806080806080\
40304040303F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000B08F8FBF908FB08F8FBF908F
B08F8FBF908FB08F80BF908FB08F80BF908FB08F8090906F208F002F9000208F002F9000208F00\
2F90006F8F30BF9080B08F80BF9080B08F80BF9080B08F80BF9080B08F7FBF9080
B08F7FBF907FB08F7FB0907F8F8F60909060208F00209000208F002F9000208F00209000208F10\
B09070AF8F6FB09070504030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000008F704F8F704F80704F000000000000
000000000000000000000000000000000000000000504F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000201F20302F3F3F3040302F403F304F00000000000F\
00000000000000000000000000000000000000000000000000000000000000000F
5040605040605F40605040605F406F5F4060604F6F5F4F6F604F6F604F6F6050706000006F2030\
6F50706F5F7F6F5070705F7F6F40607000007000007000007000007F0000700000
705F70705F7F7F60807F607F7F60807F607F8060807F607F80608080607F806F808060808F6F80\
40303F403F40000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000001F00001F00000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000BF908FB08F8FBF908FBF908F
BF908FB08F8FBF908FBF908FBF908FB08F80BF908FBF908F2F9000208F002F90002090002F9000\
208F002F900070905090906F908F6F9F906FBF9080BF9080B09080BF9080B0907F
BF9080B0907F90905090904F2F90002090002F90002090002F90002090002F90002090006F9030\
B08F70B09070B08F705F4F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000008F7F4F806F40000000000000
000000000000000000000000000000000000000000504F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F3F303040302F3F303040302F40000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
4F3F5F5040605040605F40605040605F406F5F4060604F6F5F4F6F604F6F604F6F6F2030600000\
6050706F50706F5F7F6F507070101F6F00007000006F00007000007000007F2F30
705F7F7F5F7F705F7F7F607F7F5F7F7F607F7F607F8060807F607F80608080607F806F80806080\
40304040303F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000001F00001F00001F00001F00001F00001F00\
001F00001F00001F00001F00208000207F00208000000000000000000000000000
0000000F0F0F206000206F00204F00205000001F00001F00000000000000000000000000000000\
000000000000000000000000000000000000000000B08F8FBF908FB08F8FBF908F
B08F8FBF908FB08F8FBF908FB08F80BF908FB08F80BF908FB08F8070903F6F8F302F9000208F00\
2F9000208F002F9000208F002F9000208F002F9000908F6090904F908F4F709030
208F1F2F901F2090002F9000208F002090002090002F9000208F002090002090002F9000A0806F\
B09070AF8F70B09070504030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000080704F000000000000
000000000000000000000000000000000000000000504F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000030303F302F3F3F3040303040000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5040605040605F40605040605F4F6F5F406F604F6F5F4F6F605070604F6F60507060405F6F0000\
6F304F6F5F7F6F5070705F7F6F00007000007000007000007000007F00006F5070
7F60807F5F7F7F60807F5F7F8060807F607F8060807F607F806F80806080806F808060808F6F80\
403040403F40000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000003000207F00207F00207F00207F00207F00208000\
207F00208000207F00208000207F00208000207F00000000000000000000000000
000000001F002080002080002F8F00208000208F00208000003F00003F00000000000000000000\
000000000000000000000000000000000000000000BF9090BF908FBF908FBF8F8F
BF908FBF908FBF908FBF908FBF908FBF908FBF908FBF9080BF908FBF908FBF908FAF807F2F9000\
2090002F90002090002F90002090002F90002090002F90002090002F9000209000
2F90002090002F90002090002F90002090002F90002090002F90002090004F90306F9040B0907F\
B09070B09070B08F705F4F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000080704F0F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F3F303040302F40000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F000F
4F3F605040605040605F40605040605F4F6F5F4060604F6F5F4F6F605070604F6F6F5070601010\
6F101F6F405F6F40606F405F7000006F0000700000700000700000702030705F7F
705F7F7F607F705F7F7F607F7F5F7F7F60807F607F8060807F607F806F8080607F806F80806080\
40304040303F000000000000000000000000000000000000000000000000000000
000000000000000000000000001F00001F001F7F00207F001F7F00207F001F7F00207F00207F00\
2080001F7F00208000207F002080001F7F00208000000000000000000000000000
000000003F00208000208000208000208F00208000208F00208000208F00205000205000000000\
000000000000000000000000000000000000000000B08F8FBF908FB08F8FBF908F
B08F8FBF908FB08F8FBF908FB08F8FBF908FB08F80BF908FBF8F80BF908FB08F80BF908F6F8F50\
7090504F8F304F901F208F002F9000208F002F90002090002F9000208F002F9000
2090002F9000208F002F90002090002F90002090002090006F90306F9030B09070B09070B09070\
B09070AF8F70B09070504030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000030304030303F303040000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000000000F
5040605040605F406F5040605F4F6F5F4060604F6F5F4F6F604F70604F6F6F50706050706F304F\
6F0000702F306F20307000006F00007000007000007000007000007F4F6F705F7F
7F60807F5F7F7F60807F607F8060807F607F806080806080806F808060808F6F80806F808F6F80\
403040403F40000000000000000000000000000000000000000000000000000000
000000000000000000000000204F00204F00207F00207F00207F00207F00208000207F00208000\
207F00208000207F00208000207F00208000208000000000000000000000000000
000000204F002F8F00208000208F002080002F8F00208F002F8F00208F002F8F00208F00002000\
001F00000000000000000000000000000000000000BF9090BF908FBF9090BF908F
BF908FBF908FBF908FBF908FBF908FBF908FBF908FBF908FBF908FBF9080BF908FBF908FBF908F\
BF9080BF908FBF90802F90002090002F90002090002F90002090002F90004F901F
4F901F2090002F90002090002F901F20901F90904F90904FBF907FB09070BF907FB09070B0907F\
B09070B0907FB090705F4F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000008F704F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000302F3F303040000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
4F3F5F5040605040605F406F5040605F4F6F5F406F604F6F5F4F6F604F70604F706F5070604F6F\
6F20306F00007000006F00007000006F0000700000700000705F7F705F7F7F607F
705F7F7F607F7F5F7F7F60807F5F7F7F60807F607F8060807F607F806080806080806F80806080\
403040403040000000000000000000000000000000000000000000000000000000
000000000000000000001F001F7000207F001F7F00207F001F7F00207F001F7F002080001F7F00\
207F0020601F206F201F7F00208000207F0020800050404F50404F50404F5F404F
50404F40503F206F1F206F20208000208F00208000208F00208000208F00208F00208F00208F00\
208F005F4F4F604F4F5F4F40604F4F5F4F4F604F4F5F40405F4F405F4040BF908F
B08F8FBF908FBF908FBF908FB08F8FBF908FBF908FBF908FB08F80BF908FBF908090906F4F8F1F\
BF908FBF9080BF908FB08F80BF9080BF9080BF9080B08F80BF9080B09080BF9080
B0907FBF9080B0907F90904FB0907FBF907FB0907FBF907FB09070B0907FB09070B0907FB09070\
B09070B09070B09070504030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000806F40807040
00000000000000000000000000000000000000000000000040402F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000001F1F20303040000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000000000F
5040605040605F406F5040605F4F6F5F4F6F604F6F5F4F6F605070604F706F50706050706F5F7F\
6F40607000006F00007000007000007000007000007F1F1F705F7F7F6080705F7F
7F60807F607F7F60807F607F8060807F6080806080806080806F80806080806F80806F808F6F8F\
403040403F4000000000000000000000000000000050404F4F3F4F50404F4F3F4F
50404F4F3F4F50404F3F601F207F00207F00207F00207F00208000207F00404F3F50404F50404F\
A07F90408020407F1F208000208000208000208000AF8090A08090AF8090A0808F
AF8F90A08090AF8F90AF808F6F8F3F6F803F2F8F00208F002F8F00208F002F8F00208F002F9000\
208F00909070908F70BF9090BF908FBF9090BF908F0000000000000000005F4040
5F4F4F5F4F40BF908FBF908FBF908FBF908FBF908FBF908FBF908FBF908FBF908F90906F2F9000\
4F901F4F90204F901FBF908FBF9080BF908FBF9080BF908FBF9080BF9080BF9080
BF9080BF907FBF9080B0907FBF9080B0907FBF907FB0907FBF9F7FB0907FBF907FB09070BF907F\
B09070B09070B090705F4F3F50403F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F807040
80704F4F0000000000000000000000000000000000000000504F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000001F1F20302F3F000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F000F
503F605040605040605F406F5F40605F4F6F5F406F604F6F604F6F605070604F706F50706F5070\
6F5F7F6F10106F00006F00007000006F0000700000704F60705F7F705F7F7F5F7F
705F7F7F60807F5F7F7F60807F607F8060807F607F8060808060804030403F303F40304040303F\
40304040303F4F3F4F403F404F3F4F4F3F404F3F4F90708F90708F90708F907F8F
90708F9F7F9090708F407F201F7F00207F001F7F00207F005F7F50607F5F9F7F8F9F7F909F7F8F\
A080903F7F1F408020207F00208000207F00208000A0808FA08090A0808FA08090
A0808FAF8090A0808FAF8090AF808FAF8F9040801F4F8F20208F00208F00208F00208F00208F00\
2F90006F8F506F905FB08F8FBF9090B08F8FBF9090000000000000000000000000
0000000000000000005F4F405F40405F4F40BF8F8FBF908FBF908FBF908FBF908090906F209000\
2F90002090002F90004F901F4F901F4F901F4F901F4F901F4F901F4F901F4F901F
4F901F4F901F4F901F4F901F4F901F4F901FB0907FBF907FB0907FBF907FB09070B0907FB09070\
B09070B09070B09070AF90705F4F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000008F704F
8070408000004F00004F00000000000000000000000000004F402F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000001F1F1F0F0F0F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000504060
5040605040605F4F6F5F40605F4F6F5F406F604F70604F6F605070604F706F50706F50706F5F7F\
6F507F703F4F6F000070000070000070000070101F705F7F705F7F7F60807F5F7F
7F60807F5F7F7F60807F60808060807F6080806F80403040403040000000000000000000000000\
403F404F3F4F8F6F8F90708F8F6F8F90708F90708F90708F90708F9F7F9090708F
9F7F90907F8F9F7F90207F00207F00207F00208000207F00907F8F907080A080909F7F90A08090\
9F7F8F40802040801F208000208000208000208000AF8090A08090AF8090A08090
AF8090A08090AF8F90AF8090AF8F90AF80908F8F5F8F8F502F8F00208F002F9000208F002F9000\
208F0070903F6F8F3FBF9090B0908FBF9090BF908F000000000000000000000000
0000000000000000000000000000000000005F4F4F5F4F405F4F405F4040BF908F90906F2F9000\
2090002F90002090002F90002090002F90002090002F90002090002F9F00209000
2F90002090002F9F002090002F90002090002F9F002090004F9F1F4F901F4F9F1F4F901F4F901F\
B09070B0907FB09070B09070504030000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F8F704F
8F704F8000008000007F00004F00004000004F0000000000504F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000030303F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000504060
503F605040605040605F406F5F40605F4F6F5F4F6F604F70604F6F605070604F706F507F6F5070\
6F5F7F6F50707020306F0000700000700000703F4F705F7F705F7F705F7F7F6080
705F7F7F60807F5F7F7F60803F2F3F403040000000000000000000000000000000000000000000\
4F3F4F4030408F708F8F6F808F708F8F708F90708F8F708F90708F90708F9F7F90
90708F90708F1F7F00207F001F7F00207F001F7F00407F209F7F8F9F7F909F7F8FA080909F7F8F\
A08090407F1F4080201F7F00208000208000208000A0808FA08090A0808FAF8090
A0808FAF8090AF808FAF8F90AF808FAF8F90A0808FAF808F208F00208F00208F002F8F00208F00\
208F004F8F1F4F9020B08F8FBF9090BF908FBF9090000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F40404F5F3F4F701F\
2F90002090002F90002090002F90002090002F90002090002F90002090002F9000
2090002F90002090002F90002090002F90002090002F90002090002F90002090002F9000209000\
4F901FB09070B09070AF8F705F4F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000008F704F
80704080402F7F00008000007000007F0000700000000000504F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F503F60
5F406F5040605F406F5F40605F4F6F5F4F6F604F70604F6F6050706050706F50706F50706F5F7F\
6F507F705F7F6F4060700000702030702F30705F7F7F6080705F7F7F6080705F7F
7F60803F303F403040000000000000000000000000000000000000000000000000000000000000\
4F3F404F3F4F8F6F8F90708F8F708F90708F90708F907F9090708F9F7F9090708F
9F7F907F7F70207F00207F00208000207F002080005F7F3FA080909F7F90A080909F7F90A08090\
A0809080805F80805060805F60805F60805F60805FAF8090A08090AF8090A08090
AF8F90AF8090AF8F90AF8090B08F90AF8F90B08F90AF8F902F8F00208F002F8F00208F002F9000\
208F004F90204F8F1FBF9090BF9090BF9090BF908F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000002000\
4F701F507F202090002F9F002F90002F90002090002F9F002090002F9000209000
2F9F002090002F9F002090002F9F002090002F9F002090002F9F002090002F9F002090002F9F00\
2090002F90004F901FB0907F504F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F8F704F
8F7F4F801F108000007F00007F00007F0000700000000000302F1F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000504060
5040605F406F5040605F406F5F406F5F4F6F5F4F6F604F70604F6F605070604F706F507F6F5070\
6F5F7F6F5070705F7F6F2030705F7F705F7F705F7F705F7F7F60803F2F3F3F3040
3F2F3F000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F4030408F708F8F6F8F90708F8F708F90708F90708F907F9090708F907F90
907F8F7F7F701F7F00207F001F7F00207F001F7F00607F5F9F7F8FA080909F7F8FA080909F7F90\
A080909F7F8FA08090A08090A08090A0808FA08090A08090AF8090A0808FAF8090
A08090AF8F90A0808FAF8F90AF8090AF8F90AF808FB08F90208F00208F00208F00208F00208F00\
2F90006F8F3F70903FB08F8FBF9090B08F8FBF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000207000\
0020002090002F90002090002F90002090002F900020701F4F7F1F4F701F4F7F1F
4F701F4F7F1F4F701F2F7F1F20701F2F90002090002F90002090002F9F002090002F9000209000\
2F90002090002090004090102F7F1F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000080704F
807040801F107F00007F00007F00007F0000700000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000030303F000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F504060
50406F5040605F4F6F5F406F5F4F6F5F4F6F604F70604F706050706050706F507F6F50706F5F7F\
6F507F705F7F6F5F7F705F7F705F7F705F803F2F3F3F30403F2F3F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
403F404F3F4F8F6F8F90708F8F708F90709090708F907F9090708F9F7F90907F90
9F7F907F7F70208000207F00207F00207F002080005F7F5FA080909F7F90A08090A08090A08090\
A08090A08090A08090A08090A08090AF8090A08090AF8090A08090AF8F90AF8090
AF8F90AF8090AF8F90AF8F90B08F90AF8F90908F7F8F8F702F8F00208F002F9000208F002F9000\
208F0070905F6F905FBF9090BF9090BF9090BF9090000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000020002F9000\
2090002F9F002F90002F9F00207000002000002000000000000000000000000000
000000000000000000000000000000000000000000003F002F9F002090002F9F002090004F7F1F\
2090002F9F002090002F90002090002F7F00207000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000807040
8F7F4F801F0F8030207F6F3F7F1F10703F207F40206F0000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000030101F302F3F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000504060
503F605040605040605F4F6F5F406F5F4F6F5F4F6F604F70604F6F6050706050706F507F6F5070\
6F5F7F6F507F705F7F302F3F3F2F3F302F3F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F403F408F708F8F6F8F90708F8F708F90708F90708F907F9090708F9F7F90
90708F907F8F1F7F00207F001F7F00207F001F7F00408020907F8F9F7F8F9F7F90A080909F7F8F\
A08090A07F90A08090A07F90A08090A08090A08090A08090A08090A08090AF8090
A08090AF8F90AF8090AF8F90A0808FAF808F208F00208F00208F00208F00208F002F9000208F00\
209000908F70909070B08F8FBF9090BF908FBF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000003F00209000\
2F90002090002F7F00207000002000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000003F002090002F9F00207000\
4F7F1F4F701F2F9000209000209000209000209000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000080704F
807040801F108000008070407F6F3F7F6F3F7060307000000000004F4030000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003010101F1F20000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F504060
5F406F5040605F4F6F5F406F604F6F5F4F6F604F70604F706F50706050706F507F6F5070705F7F\
302F3F3F2F40302F3F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F4F3F4F8F6F8F90708F90708F90709090708F907F9090708F9F7F90907F90
9F7F909F7F90207F00207F00208000207F00208000207F00608040607F3FA080909F7F90A08090\
A08090A08090A08090AF8090A08090AF8090A08090AF8F90A08090AF8F90A08090
AF8F90AF8090AF8090A0808F4F8F20408F1F2F8F00208F002F9000208F002F9000208F002F9020\
20901FBF9090B09090BF9090BF9090BF9090BF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000003F002F9F00\
2F90002F9F002070002F7F00000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F00002F9F002090002F9F00\
205F000000004F701F4F7F1F2090002F9000209000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000807040
8F7F4F8070408030207F6F408070407F6F3F7F6F3F7060300000004F4030000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F0000302F301F101F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000504060
5040605F406F5040605F4F6F5F406F604F705F4F6F302F3F302030302F3F30203F302F3F302F3F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F4030408F708F8F6F8F90708F8F708F90708F90708F907F9090708F9F7F90
907F909F7F905F7F3F5F7F3F1F7F002080001F7F00208000207F002080007F7F50A080909F7F90\
A080909F7F90A08090A08090A08090A08090A08090A08090AF8090A08090AF8090
8080708F8F7F40801F4F8F20208F00208F00208F00208F00208F002F8F00208F00209000908F50\
90905FB08F8FBF9090BF9090BF9090BF908FBF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000003F00209000\
2F9F002090002F9000209000002000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002090002F9F00209000\
2F9F000000000000000000004F7F1F407010209000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000080704F
8070408F704F806F408070407F50307F6F3F70603070603F000000504F30000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030000030303F1F1F20000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F504060
5F406F50406030203F2F2030302F3F2F2030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F404F3F4F8F708F90708F8F708F907F9090708F907F9090708F9F7F909F7F90
9F7F909F7F9080807F7F7F70208000207F00208000207F00208000207F0020800040803F408040\
907F8FA0809FA08090A08090A08090AF8F9FA08090AF8F90A08090AF8F9F208020
208F002080002F8F00208F002F8F00208F002F9000208F002F9000208F004F90404F8F3FBF9090\
B09090BF9090BF9090BF9090BF9090C09F90BF9090000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000020002F9F00\
2090002F9F002F90002F9F00207000002000002000000000000000000000000000
000000000000000000000000000000000000000000000000000000003F00004000209F002F9F00\
2090000000000000000000005F4F3FB09F7F4F701F5F503F000000000000000000
0000000000000000000F00002000002000002F00000F0000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8F7F4F8070408F704F807040801F107F40207F6F3F706030400000504F30000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030000030303F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000503F60504060
2F1F30000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F4F3F4090708F8F6F8F90708F90708F907F9090708F907F9090708F9F7F90
907F909F7F909F7F909F7F903F7F1F207F00207F002080001F7F00208000207F002080001F7F00\
20800040801F60803F60803F60805F60805F60805F60805040803F208000208F00
208000208F00208000208F00208000208F00208F00208F00408F1F4F8F20AF8F8FB08F90B08F90\
BF9090B09090BF9090B09090BF9090BF9090BF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F5F002090002F9F002090002F9F002090002F9F002070002F7F00205F00000000
000000000000000000000000000000000000000000002000001F002F9F002090002F9F00209000\
2F9F00001F000000000000005F4F3FB09070B09070AF906F5F4F3F5F4F3F000000
000000000000000000000000AF0000AF0000A000006F0000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F00
8070408F704F8070408070407F100F7F1F0F70502F70603F600000504F30000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000300F10302F3F000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000F50406F2F2030
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F4F3F4F8F708F90709090708F907F9090708F9F7F90907F909F7F90907F90
9F80909F7F909F80909F7F90A0809F7F7F5F80805F207F00208000208000208000208000208000\
208000208000208000208F00208000208F00208000208F00208000208F00208000
2F8F00208F002F8F00208F002F8F00208F004F8F404F8F3FB08F90AF8F8FBF9090B08F90BF909F\
BF9090BF9090BF9090BF9090BF9090C09F90BF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F9F002090002F9F002F90002F9F002090002F9F002F90002F9F00209000
2F9F00003F00004000205F002F6000207F002F7F002090002F9F00209F002F9F002090002F9F00\
209F000020000000000000005F4F3FB09F7FB09070B09F70B09070B090705F4F3F
0000000000000000000000000000008F0000AF0000A00000AF0000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8F7F4F80704F8F704F8070408070407F00007F00007000006F3F20504F30000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F3030303F000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2030201F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F8F6F8F8F708F8F708F9070908F708F90709090708F9F7F9090708F9F7F90
9F7F909F7F909F7F909F7F909F7F90A080909F7F9080807F207F1F2080201F7F00208000208000\
208000208000208000208000208000208000208000208000208F00208000208F00
208000208F0040801F4F8F208F8F708F8F7FAF8F90B08F90B08F90B09090B08F90B09090B08F90\
BF9090B08F90BF9090BF9090BF9090BF9090BF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F9F002090002F9F002090002F9F002090002F9F002090002F9F00
2090002F9F002090002F9F002090002F9F002090002F9F002090002F9F002090002F9F00209000\
2F7F00001F00000000000000605040B09070B09070B09070B09070AF906FB09070
AF906F5F4F3F0000000000000000004F0000A00000AF0000A00000400000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0000
8F704F8F7F4F8070408070407F6F3F7F00007000007000006F5F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030303F30303F0000000000000000000000001000003F0000100000000000000000\
0000000000000000000000000000000000000000002F20304F3F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F9070908F708F90709090708F907F909070909F7F90907F909F7F909F7F90
9F7F909F7F90A0809F9F7F90A0809F9F7F90A0809F9F80909F809060805F60805F208020208000\
208000208000208000208000208000208F002080004F8F2060803F6F8F40608F5F
6F8F5FA07F8FB08F9FAF8F90B08F9FAF8F90B0909FB08F90B0909FB08F90BF909FB09090BF9090\
B09090BF909FBF9090BF9090BF9090C09F9FBF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000003F000040002090002F9F002F9F002F9F002090002F9F002F9F00
2F9F002090002F9F002F9F002F9F00209F002F9F00209F002F9F00209F002F9F00209F00004000\
001F000000000000000000005F4F3FBF9F7FB09070B09F70B09070B09F70B09070
B09070AF906FB090705F4F3F000000200000AF0000AF0000AF0000600000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8F7F4F8070408F7F4F8070408070407F00007F000070000070603F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F3030303F0000000000000000000F00003000001F00003F0000000000000000\
0000000000000000000000000000000000002F20304F3F604F3F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F8F6F8F90708F8F708F90709090708F907F9090708F9F7F90907F909F7F90
9F7F909F7F909F7F909F80909F7F909F80909F7F90A080909F7F90A080909F8090A0809F808070\
80807F60805F60803F60803F40802040801FAF8F9FA08090AF8F90A08090AF8F9F
AF8090AF8F90AF8F90B08F9FAF8F90B08F90B08F90B09090B08F90B09090B08F90BF9090B08F90\
BF9090BF9090BF9090BF9090BF9090BF9090C09F90000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F003F00004000003F00003F00207F002F7F00
207F002F7F00207F002F7F00205F000040002090002F9F002090002F7F00207F00002000000000\
0000000000000000000000005F503FB09070B09F70B09070B09070AF9070B09070
AF906FB09070AF906FAF9070AF8F6F5F4F3FAF0000AF0000A00000AF0000504030504F30504030\
0000000000000000000000000000000000000000000000000000000000000F0F00
8070408F7F4F8070408070407F6F3F7F1F107000007F00006F00005F302040402F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030303F302F3F000000000000000000300000000000000000400000200000100000\
0000000000000000000000002020304F40604F3F6050406F4F3F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F9070908F708F907F9090708F907F909070909F7F90907F909F7F909F7F90
9F809F9F7F90A0809F9F7F90A0809F9F8090A0809F9F8090A0809FA08090A0809FA08090A0809F\
A08090A0809FA08090AF8F9FA08090AF8F9FA08090AF8F9FAF8090AF8F9FAF8F90
B08F9FAF8F90B08F9FAF8F90B0909FB08F90B0909FB08F90BF909FB09090BF909FB09090BF909F\
BF9090BF909FBF9090C09F9FBF9090C09F9FBF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000002000002000000000000000000000000000\
000000000000000000000000000000605040B09070B09F7FB09070B09070B09070
B09070AF9070B09070AF906FB09070AF906FAF3F2FAF3F2FAF0000A00000AF5F3FA0503FA08F60\
A08F5F504F3050402F504F3050402F000000000000000000000000000000000000
8F7F4F8F704F8F7F4F8070408070407F1F0F7F5F3070603F7000005F00004F4F30000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F2F2F301F1F200000000000000000002F00000000000000000F00004000001F0000\
000000000000000000201F304F3F604F3F5F4F3F604F3F604F3F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F8F6F8F90708F8F708F90709090708F907F909070909F7F90907F909F7F90
9F7F909F809F9F7F909F80909F7F90A0809F9F7F90A080909F8090A0809F9F7F90A08090A08090\
A0809FA08090A0809FA08090AF8F9FA08090AF8F9FA08090AF8F9FA08090AF8F9F
AF8F90AF8F9FAF8F90B08F90AF8F90B08F9FAF8F90B09090B08F90B0909FB08F90BF9090B09090\
BF9090B09090BF9090BF9090BF9F90BF9090C09F90000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F4F3FB09070B09070B09070AF9070B09070
AF906FB09070AF906FAF906FAF906FAF906F9F8060AF5F40A00000AF0000A03020A03F2FA08F5F\
A08F609F805FA08F5F9F805F9F8F5F000000000000000000000000000000000000
8070408F7F4F80100F8070407F6F4080402F703F207F6F3F7060306050304F402F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000302F3030303F1F1F1F0000000000001F00000000000000000000000000003F0000000000\
0000002F2030403F5F4F3F604F3F605040604F3F605040604F3F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F90709090708F90709090708F9F7F90907F909F7F90907F909F7F9F9F7F90
9F809F9F7F90A0809F9F8090A0809F9F7F90A0809FA08090A0809FA08090A0809FA08090A0809F\
A08090AF8F9FA08090AF8F9F40801F4F8F204080206F8F4080805F8F8F60AF8F90
B08F9FAF8F90B08F9FB08F90B0909FB08F90B0909FB08F90BF909FB08F90BF909FBF9090BF909F\
BF9090BF9F9FBF9090C09F9FBF9090C09F9FC09F90000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000002000002000003F00004000207F002F9F002F9F00
2F9F00209F002F9F002F9F002F9F00209F002F8000003F000040000000000F0F0F000000000000\
000000000000000000000000000000605040B09070B09F7FB09070B09F70B09070
B09070AF8F6FB09070AF906FB09070A08060AF5F40AF5040AF0000A00000AF0000A00000A08F60\
A08F5FA08F60A08F5FA08F5F9F80509F8F5F000000000000000000000000000000
8F7F4F8F704F8F7F4F80301F8070407F6F3F706F3F706F3F706F3F604F2F0F0F0F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F302F3F1F1F200000000000001000000000000000000000001F0000400000000000\
201F3040305F4F3F60403F5F4F3F604F3F605040604F3F604F3F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F8F708F9070908F708F907090907090907F909070909F7F90907F909F7F90
9F7F909F809F3F7F1F4080203F7F1F4080203F7F1F4080203F7F1F40802040801F40802040801F\
40802040801F408020208000208000208000208F00208000208F002080004F8F20
40801FAF8090A08090B08F9FAF8F90B08F9FB08F90B0909FB08F90B0909FB09090BF909FB09090\
BF9090BF9090BF909FBF9090BF9F90BF9090C09F9F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000205F002F9F00003F002F9F002090002F9F002090002F9F00209F002F9F00
2090002F9F00209F002F9F002090002F9F00209F002F9F00209F002F9F00209F00004000000000\
0000000000000000000000000000005F4F3FB09F7FB09070B09070B09070B09070
AF906FB03F30AF3F2FB05F40AF1F10AF0000AF0000AF0000A00000AF0000A00000A00000A03020\
A0503FA05030A06F4F9080509F8F5F9F805050402F000000000000000000000000
8070408F7F4F8070408000007F6F3F7F70407F6F3F7F6F3F706030600000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F30303F000000000000000000000000000000000000300000400000300F10403F60\
403F5F4F3F60403F5F4F3F604F3F605040604F3F6050406F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F9070908F708F907F90907090907F909070909F7F909F7F909F7F9F9F7F90
A0809F9F7F90208000207F00208000208000208000208000208000208000208000208000208F00\
208000208F00208000208F00208000208F002080002F8F00208F00208F00208F00
2F8F00408F204F8F20AF8F90B0909FB08F90B0909FB08F90BF909FB09090BF909FB09090BF909F\
BF9090BF9F9FBF9090C09F9FBF9090C09F9FBF9090000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
003F002F9F002F9F002F9F002F9F002F9F00209F002F9F002F9F002F9F00209F00
2F9F002F9F002F9F00209F002F9F002F9F002F9F00209F002F9F002F9F002F9F00209F00004000\
000000000000000000000000000000605040B09070BF9F7FB09070B09F70B09070
B07F60B00000B00000AF0000B00000AF0000AF0000AF3F2FAF3F2FAF0000AF0000A00000AF0000\
A00000A00000A00000A000009F1F109F8F5F9F8050504F30000000000000000000
0F0F0F8F7F4F8F7F4F8000008000007F6F407F6F40706F3F7F6F3F6F0000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F302F3F0000000000000000000000001F00003F00003F00003F0000402F4F40305F\
403F5F40305F4F3F604F3F5F4F3F604F3F605040604F3F60000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F4F8F708F90709090708F907F9090708F907F90907F909F7F90907F909F7F90
9F7F909F809F1F7F00208000207F002080001F7F00208000208000208000208000208000208000\
208000208000208000208000208F00208000208F00208000208F00208000208F00
208F00208F00208F008F8F7F8F8F7FB0909FB08F90B0909FB08F90BF909FB08F90BF909FBF9090\
BF909FBF9090BF909FBF9090C09F9FBF9090C09F90000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000002000000000\
2F7F002090002F9F00209F002F9F002090002F9F00209F002F9F00209F002F9F00
209F002F9F00209F002F9F00209F002F9F00209F002F9F00209F002F9F00209F002F9F00209F00\
0040000000000000000000000000005F4F3FB09F7FB09070B09F70B09070B09070
B03F2FB00000AF0000B00000AF0000AF3F2FAF3F2FAF906FAF7050AF0000A00000A00000A01F10\
A0503FA05030A01F109F0000A000009F80509F805090805050402F000000000000
0000008F7F4F80704080301F800000801F107F3F207F503070502F6F00002F2010000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000302F3F3030400000000000000000001F00003F00003F00003F000040000040305F403F60\
40305F4F3F604F3F604F3F604F3F6050406F503F60504060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F3F4F\
4F3F4F907090907090907F909070909F7F90907F909F7F909F7F909F809F9F7F90
A0809F9F7F90208000208000208000208000208000208000208000208000208F00208000208F00\
2080006F8F6060805F6F8F6060805F4F8F20408F20208F00208F002F8F00208F00
2F8F00208F002F90006F8F5F6F905FB08F90BF909FB0909FBF909FB09090BF909FBF909FBF909F\
BF9090C09F9FBF9090C09F9FBF9090C09F9FC09F90000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000003F00002000\
2F9F002F9F002F9F002F9F00003F002F9F002F5F002F6000003F000040002F7F00
2F9F00209F002F9F002F9F000020000020000040000040002F9F00209F002FA0002F9F002F9F00\
209F00000000000000000000000000605040B09F70BF9F7FB09070BF9F7FB09070
B0201FB00000B00000B00000B00000AF0000AF906FAF906FAF906FAF0000AF0000A00000AF0000\
9F805F9F805FA08F5FA0503F9F0000A000009F80509F805090805050402F000000
0F0F0F8F7F4F8F7F4F804F2F8000007F00008000007F00007F00006F5F305F5030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F302F3F0000000000000000003000003F00003F00003F00003F3050403F5F40305F\
403F5F403F5F4F3F60403F5F4F3F604F3F605040604F3F5F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000403F4F\
4F3F4F8F708F90709090708F907F909070909F7F90907F909F7F909F7F909F809F
9F7F909F809F5F7F5F60805F5F7F5F60805F60805F60805F5F805F60805F60805F60805F60805F\
A0809FA08090AF8F9FA08090AF8F9FA080908F8F5F80805F208F00208F00208F00
208F00208F00208F006F905F6F8F5FB0909FB08F90B0909FB08F90BF909FB09090BF909FB09090\
BF909FBF9090BF9F9FBF9090C09F9FBF9090C09F9F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000004000003F00\
2F9F00209F002F9F00209F000000000000000000000000000000000000002F7F00
209F002F9F00209F002F9F000000000000000000000000000000000F0F00207F002F9F00209F00\
2F9F00205F000000000000000000005F4F3FBF9F7FB09070B09F70B09070B09F70
AF3F2FB00000AF0000B00000AF1F10AF201FAF906FAF906FA08F60AF906FA03F2FAF0000A00000\
A00000A0503FA08F5F9F805FA08F5F9F00009F6F40908050908050907F4F4F402F
0000008F7F4F8F70408F4F2F8000008000007F00007F00007000006F60305F5030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F3030400000000000003000003F00003F00003F10103F3050403F5F40305F403F60\
403F5F4F3F604F3F604F40604F3F605040604F3F60504060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F404F\
4F3F4F907090907090907F909070909F7F90907F909F7F9F9F7F909F809F9F7F90
A0809F9F809FA0809F9F809FA0809FA0809FA0809FA0809FA08F9FA0809FA0809FA0809FAF8F9F\
A0809FAF8F9FA0809FAF8F9FAF8F9FAF8F9FAF8F9FAF8F9F408F204F8F20208F00
2F9000208F002F90006F8F5F709060B0909FBF909FB0909FBF909FBF909FBF909FBF9090BF9F9F\
BF909FC09F9FBF9090C09F9FC09F9FC09F9FC09F90000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000003F00004000\
2F9F002F9F002F7F002F8000000000000000000000000000000000000000207F00
2F80002F9F002F9F00209F00000000000000000000000000000000000000004000209F002FA000\
209F002F9F00000000000000000000605040B09F70BF9F7FB09F70B09F7FB09070
B0604FB00000B00000AF0000B0201FAF1F1FB0906FAF906FAF906FAF9060AF5F40A00000AF0000\
A00000A00000A08F5FA08F60A08F5FA000009F00009F8F5F908050908050907F4F
0F0F0F8F7F4F8F7F4F7F6F408F00008000008000007F00007F0000700000605F30000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F3030303F10101F0000000000003F00003000003F304F3F304F3F305F3F305040305F40305F\
403F6040305F4F3F604F3F604F40604F3F605040604F3F5F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F3F4F\
4F3F4F8F708F907090907090907F909070909F7F90907F909F7F9F9F7F909F809F
9F7F90A0809F9F7F90A0809F9F8090A0809F9F8090A0809FA08090A0809FA08090A0809FA0809F\
AF8F9FA08090AF8F9FA0809FAF8F9FA08090AF8F9FAF809F4F8F20408F1F208F00
208F00208F00208F00A08090A0808FB0909FB08F90B0909FB09090BF909FB09090BF909FBF9090\
BF9F9FBF9090BF9F9FBF9090C09F9FBF9090C09F9F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000002000001F00\
2F9F00209F002F9F00207F000000000000000000000000000000000000002F7F00
207F002F9F00209F002F9F000000000000000000000000000000000000000000002F9F00209F00\
2F9F00209F000000000000000000005F4F3FBF9F7FB09070B09F7FB09070B09F70
B0705FB00000AF0000B00000AF0000B00000A08F60AF906FAF8F60AF906FA0704FAF0000A00000\
A00000A00000A0503F9F8F5FA08F5F9F30209F0000904F309F8050907F4F908050
8F704F8F7F4F8F70408F7F4F807040805F3F7F50307F5F307000007000005F5030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
30303F302F3F1F1F200000003F30403000003F2F403F304F3F30503F3050403F5F40305F403F60\
403F5F4F3F604F3F604F40604F3F6050406F504060504060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F3F4F\
4F3F4F907F90907090907F909070909F7F9F907F909F7F9F9F7F909F809F9F7F9F
A0809F9F809FA0809FA0809FA0809FA0809FA0809FA0809FA0809FA0809FAF8F9FA0809FAF8F9F\
A0809FAF8F9FAF809FAF8F9FAF809FAF8F9F9F8090A08090208F002F8F00208F00
2F9000408F3F4F9040B08F9FB0909FB08F9FBF909FB0909FBF909FB0909FBF9F9FBF909FBF9F9F\
BF909FC09F9FBF9F9FC09F9FBF9F90C09F9FC09F9F000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F00000F0F0F\
2F9F002FA0002F9F002F9F000020000020000000000000000000000000002F7F00
2F8000209F002FA0002F9F000000000000000000000000000000000000000F0F0F0F00002FA000\
209F002FA000002000000000000000605040B09F7FBF9F7FB09F70BF9F7FB09F70
B09F70B00000B00000B00000B00000AF0000B05F40AF906FAF906FAF9060AF906FAF0000AF0000\
A00000AF0000A01F10A02010A08F5FA08F5F9F00009F00009080509F8050907F50
9080500000008F7F4F8F7F4F8F7F4F8070408070407F6F3F7F6F3F70502F604F2F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F3030303F1F101F1F1F2030000030000030203F3F30503F30503F305F3F305F403F5F40305F\
403F60403F5F4F3F604F3F604F3F604F3F6050406F4F3F5F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F3F4F\
4F3F4F8F7090907F90907090907F90907F909F7F9F907F909F7F9F9F7F909F809F
9F7F90A0809F9F809FA0809F9F8090A0809FA0809FA0809FA08090A0809FA0809FA08F9FA08090\
AF8F9FA0809FAF8F9FA0809FAF8F9FA08090408F2040801F208F00208F004F8F20
408F1FAF8F90AF8F90B0909FB08F90B0909FB08F9FBF909FB09090BF909FBF909FBF909FBF9090\
BF909FBF9090C09F9FBF9090C09F9FC09F90C09F9F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F9F00209F002F9F00209F002F9F00209F00004000003F000F0F0F0000002F8000
207F002F9F00209F002F9F00000000000000000000000000000000000000003F00004000209F00\
2F9F00209F000020000000000000005F4F3FBF9F7FB09070B09F7FB09070B09F70
B09070B00000B00000B00000AF0000B00000AF3F2FAF906FAF8F60AF906FA08F60AF3F2FA03020\
A00000A00000A000009F00009F805F9F80509F00009000009F8050908050908050
8F7F4F0F0000403F208F7F4F8070408070407F6F3F7F6F3F706030706F3F605F30000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
30303F30303F0000003030403F00003000003F20303F304F403F503F305F403F5F40305F4F3F60\
403F604F3F604F3F605040604F3F6050406F504060504060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F4050\
4F3F4F907F909070909F7F9F907F909F7F9F907F909F809F9F7F9F9F809F9F7F9F
A0809F9F809FA0809F9F809FA0809FA0809FA0809FA0809FAF8F9F60805F608F6060805F6F8F60\
60805F6F8F6060805F4F8F20408F20208F00208F002F8F00208F002F8F006F8F3F
6F90406F8F5F6F9060A08090AF8F90B0909FBF909FB0909FBF909FBF909FBF9F9FBF909FC09F9F\
BF909FC09F9FBF909FC09F9FC09F9FC09F9FC09F9F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0040002F9F002F9F002FA0002F9F002FA0002F9F002FA0002F9F002FA0002F9F00
2FA0002F9F002FA000209F000000000000000F0F0F0000002F60002F60002FA000209F002FA000\
2F9F002FA000002000000000000000605040B09F7FBF9F7FB09F70BF9F7FB09F70
BF9F7FB05F40B00000B00000B00000AF0000B0201FAF906FB0906FAF906FAF906FA0503FAF5F40\
A00000AF0000A00000A00000A0503FA08F5F9F00009F00009F1F109F8050907F4F
908050000000000000000000403F20403F20403F207F6F3F7F6F3F706F3F6F4F2F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F3F30303F1F101F3030403000003F0000300F0F3F30503F305040305F3F305F403F5F40305F\
4F3F60403F5F4F3F604F3F605040604F3F6050406F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F3F4F\
4F3F4F8F7090907F909070909F7F909070909F7F9F907F909F809F9F7F909F809F
9F7F9FA0809F5F7F5F60805F60805F60806060805F60805F60805F208F00208000208000208000\
208F00208000208F00208000208F00208000208F00208F00208F00208F00208F00
208F00209000208F0090907F8F8F7FAF8F90A0808FAF8F90AF8F8FAF8F90AF8F8FAF8F90AF8F8F\
B08F90AF8F8FB09090B08F8FB09090B08F8FB090900F0F0F100F0F0F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F207F002F7F00209F002F9F00209F002F9F00209F002F9F00209F002F9F00
209F002FA000209F002F9F00000000000000001F00002000209F002FA000209F002FA000209F00\
2FA000207F000000000000000000005F4F3FBF9F7FB09F70BF9F7FB09070B09F70
B09070B09F70B05F40B00000AF0000B00000AF0000AF00009F8060AF906FA08F60AF9060A0704F\
AF0000A00000A00000A00000A03F2F9F80509F1F109F00009F000090804F908050
8F7F4F0F0F000000000000000000000000000000000000003F301F3F301F6000002F2F1F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
30303F302F3F3F30403030403F00003000003F304F3F305040305F3F305F403F6040305F4F3F60\
403F604F3F604F3F604F40604F3F6050406F50406F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F3F4F\
4F3F4F907F908F6F8F8F708F8F708F9070908F708F907F9090708F907F90907F90
907F90907F90208000208000208000208000208F00208000208F00208000208F00208000208F00\
2080002F8F00208F00208F00208F002F8F00208F00208F00208F002F9000208F00
2F9000208F002F9000001F00002000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C09F9FC09F9FCFA09FC09F9F
100F0F0F0F0F101010000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F0F00002F80002F7F002FA0002F9F002FA0002F9F002FA0002F9F00
2FA0002F9F002FA0002F9F002FA0000000000000000000002FA000209F002FA0002F9F002FA000\
207F00004000000000000000000000605040BF9F7FBF9F7FB09F70BF9F7FB09F70
BF9F7FB09F70B07F60B00000B00000AF0000B00000AF0000AF5F40AF906FAF906FAF9060AF9060\
A00000AF0000A00000A00000A000009F8F5F9F8F5F9F00009F00009F8050908050
9080508F7F4F0000000000000000000000000000000000000000000000007050305F5030000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F30000000302F3F303040300F0F3F0000300F0F3F30503F305040305F40305F403F6040305F\
4F3F60403F604F3F604F3F605040604F3F604F3F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0F000F0F0F0F0F000F0F0F0F0F000F0F0F0F0F000F0F0F0F4F3F4F\
4F3F4F806F8F000000000000000000000000000000000000000000000000000000
0000000000002080002080001F8000208000208000208000208000208000208000208F00208000\
208F00208000208F00208000208F00208F00208F00208F00208F00208F00209000
208F00209000208F00002000001F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C09F90C09F9FC09F90C09F9F
C09F90C09F9FC09F90100F0F0F0F0F100F0F0F0F0F000000000000000000000000000000000000\
000000000000000000000000000000003F00004000205F002F9F00209F002FA000
209F002F9F00209F002FA000209F00000000000000000000205F002FA000209F002FA000207F00\
0F0F000000000000000000000000005F503F605040B09F70BF9F7FB09F70BF9F7F
B09070B09F70B09070B00000AF0000B00000AF0000B00000AF503FAF906FAF9060AF906FA08F60\
AF5F3FA00000A00000A00000A000009F5030A08F5F9F00009F0000908050908050
8F7F4F8F7F4F0000000000000000000000000000000000000000000000006F60305F5030000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
1F1F200000003030403030403F304F3000003F00003F3050403F5F40305F403F6040305F4F3F60\
403F604F3F604F3F6050406F4F3F6050406F4F3F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F0F0F0F0F0F0F803F508F405F8040508F405F8F40508F4F5F8F4050904F5F8F405F4F3F4F\
403F4F000000000000000000000000000000000000000000000000000000000000
000000208000208000208000208F00208000206F00206F00207000206F00206F00206F00207000\
206F00207000206F002F7000206F00207000206F002F70002070002F7000207000
2F70002070002F7000001F00002000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFA09FC09F9FCFA09FC09F9F
CFA09FC09F9FCFA09FC09F90CFA09FC09F90CFA09F0F0F0F1010100F0F0F000000000000000000\
000000000000000000000000000000000000000000000000000000000000002000
002000002000002000002000002000000000000000000000000000000000004000004000000000\
0000000000000000000000000000000000005F503FBFA07FBF9F7FBF9F7FB09F70
BF9F7FB09F70B09F70B00000B00000B00000B00000AF0000B05F40AF906FAF906FAF9060AF906F\
A07050AF0000A00000AF0000A00000A03F2F9F8F5FA000009F00009F6F40908050
9080508F7F4F0000000000000000000000000000000000000000000000000000005F100F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000030303F
0000001F1F20302F3F303040302F403F20303000003F30503F3050403F5F3F305F403F6040305F\
4F3F60403F604F3F604F3F6050406F4F3F604F3F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000F0F0F0F
7F3F50804050803F508000008000008F00008000008F00008F00008F00008F101F901F20403040\
4F3F4F000000000000000000000000000000000000000000000000000000000000
000000206F001F6000206F00206F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C09F9FCF9F9FC09F90C09F9F
C09F90CFA09FC09F90C09F9FC09F90CFA09FC09F90C09F90C09F90CFA09F0F0F0F100F0F0F0F0F\
100F0F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000605040B09F7FBF9F7FB09F70B09F7F
B09F70AF906FB01F10B00000B00000B00000AF0000AF0000AF5F3FAF906FAF8F60AF9060A08F60\
AF9060A08F5FA00000A00000A000009F0000A08F5F9F30209F0000900000908050
907F4F8F7F4F000000000000000000000000000000000000000000000000000000600000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000001F1F1F
0000001F1F203F30403030403F304F30304F3F00003F0000403F5F3F305F403F60403F604F3F60\
403F604F40604F3F6050406F4F3F6050406F4F3F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F0F0F0F7F3F5080405F803F50
8000008000008000008000008F00008000008F00008F00008F00008F00009000008F00004F3F4F\
403F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFA09FC09F9FCFA09FC09F9F
CFA09FC09F9FCFA09FC09F9FCFA09FC09F90CFA09FC09F90CFA09FC09F90CFA09FC09F90A0A07F\
509F1F50A0200F0F0F100F0F0F0F0F10100F0F0F0F100F0F0F0F0F10100F0F0F0F
100F0F0F0F0F10100F0F0F0F100F0F0F0F0F10100F0F0F0F100F0F0F0F0F10100F0F0F0F100F0F\
000000000000000000000000000000000000605040BFA07FB09F7FBFA07FB09F70
BF9F7FB09F70BF604FB00000B00000B00000B00000AF5F40B0906FAF906FAF906FAF9060AF906F\
A08F60AF9060A01F10A00000A00000A00000A08F5F9080509F00009F0000908050
9080508F7F4F000000000000000000000000000000000000000000000000000000600000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000001F1F20302F40303040302F4030304F3F00003F1F203F305040305F40305F403F6040305F\
4F3F604F3F604F3F604F3F6050406F503F60504060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000F0F0F0F7F60807F506F7F3F507F00007F0000800000
7F00008000008000008000008000008F00008000008F00008F00008F00008F0000900000403040\
4F3F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001F00\
001F00001F00002000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C09F9FC09F9FC09F9FCFA09F
C09F90C09F9FC09F90CFA09FC09F90C09F9FC09F90CFA09FC09F90CFA090C09F90CFA09F9F9F70\
2FA0002F9F002FA0004F9F1F50A0204F9F1F50A0204F9F1F50A01F4F9F1F50A020
4F9F1F50A01F4F9F1F50A01F4F9F1F50A01F4F9F1F50A01F4F9F1F4FA01F4F9F1F50A01F4F9F1F\
0F0F0F0F0F0F000000000000000000000000605040B09F70BF9F7FB09F70BF9F7F
B09F70B09F70B09F70B05F4FB07F5FB09F70AF906FB09F6FAF906FAF906FAF9060AF906FA08F60\
AF9060A08F60AF5F3FA00000A00000A00000A03F2F9F80509F00009000009F8050
907F4F8F7F4F000000000000000000000000000000000000000000000000000000605F30000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F3F30403030403F304F3F304F3F20303F0000403F5F40305F403F60403F5F4F3F60\
4F3F604F3F604F3F6050406F50406050406F4F3F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F0F0F0F705F807F608F7F608F7F608F7060808000007F00008000007F0000
800000800000801F2080101F8F1F208F00008F00008F00009000008F00009000008F00004F3F4F\
4F3F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000003F002050002050000000000000000000002F8F00\
208F00208F00208F002F8F00208F00205F00003F00003F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFA0A0C09F9FCFA09FC09F9F
CFA09FC09F9FCFA09FC09F9FCFA09FC09F9FCFA09FC09F90CFA09FC09F90CFA09FC09F90A0A07F\
2F9F002FA0002F9F002FA0002FA0002FA0002F9F002FA0002FA0002FA0002F9F00
2FA0002FA0002FA0002F9F002FA0002FA0002FA0002FA0002FA0002FA0002FA00020A0002FA000\
2FA00050A01F0F0F0F10100F00000000000060503FBFA07FBF9F7FBF9F7FB09F70
BF9F7FB09F70BF9F7FB09F70B09F70B09F70B09F70AF906FB09F70AF906FAF906FAF9060AF906F\
AF9060AF9060A0704FAF0000A00000A00000A00000A08F5F9F3F209F0000908050
9F805090804F0F0F0F000000000000000000000000000000000000000000000000302F1F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000303040302F40303040302F403F304F3F304F3F00003F0F10403F6040305F403F60403F60\
4F3F60403F5F4F3F604F3F6050406F4F3F60504060000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F00000F0F0F0F0F000F\
70608F705F807F608F705F807F608F7F60807F608F705F7F8000007F0000800000
80000080405080608F806F8F806F8F8F6F90806F8F8F5F708F5F708F1F208F0000900000502F3F\
5F303F000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F206F00206F00208000208F00208000000000000000000000208000\
208F00208000208F00208F00208F00208F00208F00208F002070002070000F0F0F
2070000F0F0F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C09F9FCFA09FC09F9FCFA09F
C09F9FCFA09FC09F90CFA09FC09F90CFA09FC09F90CFA09FC09F90CFA09FC09F90CFA0909F9F70\
2FA000209F002FA0002F9F002FA000209F002FA000209F002FA000209F002FA000
209F002FA000209F002FA000209F002FA000209F002FA000209F002FA000209F002FA00020A000\
2FA000209F004FA01FBF9F7F100F0F0F0F00605040BF9F7FBF9F7FB09F70BF9F7F
B09F70BF9F7FB09F70B09F70B09070B09F70AF906FB09F6FAF906FAF906FAF9060AF906FAF9060\
AF9060A08F60A09060A00000A00000A00000A000009F8F50908050900000908050
90804F908050000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030303F3F30403030403F304F3F304F3F3F503F102040000040305F4F3F60401F2F4F0000\
4F00004F3F604F3F6050406F503F6050406F503F60000000000000000000000000
0000000000000F0F0F0F000F0F0F0F0F000F0F0F0F0F0F0F703F506F304F703F50703F5070608F\
705F807F608F7F60807F608F7F608F7F608F7F608F806F8F7F101F800000800000
8000008000008F405F8040508F70908F6F8F8F70908F70908F70908F405F904F5F90405F5F3040\
5F303F000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F206F00208F00208000208F00208000208F00000000000000000000208F00\
208F002F8F00208F00208F00208F002F9000208F002F9000208F002F9000209000
2F9000003F00004000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFA0A0C09F9FCFA0A0C09F9F
CFA09FC09F9FCFA09FC09F9FCFA09FC09F9FCFA09FC0A09FCFA09FC09F90CFA09FCFA090BF9F8F\
7F9F5F7FA05F7FA03F2FA0002F9F002FA0002FA0002FA0002FA0002FA0002FA000
2FA0002FA0002FA0002FA0002FA0002FA0002FA0002FA0002FA0002FA0002FA0002FA0002FA000\
20A0002FA0002FA000C0A080BF9F7FC0A080605040B09070BF9F7FBFA07FBF9F7F
BF9F7FB09F70BF9F7FB09F70B09F70B09F70B09F70B0906FB09F70AF906FB09F6FAF906FAF906F\
AF9060AF9060A09060AF201FA01F10A00000A00000A0503F9F80509F00009F0000
9F805090804F0F0F0F0000000000000000000000000000000000000000000000000000005F5030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000303040302F403F304F302F403F304F3F30503F30503F000040305F400F104F0000400000\
4F00004F203F4F406F4F3F6050406F5040600000000000000F0F0F0000000F0F0F
00000060304F602F4060304F602F406F304F6F304F6F3F506F000070000070000070101F705F80\
70608F705F807F608F705F807F608F7F60807F608F7F6080803F507F0000800000
7F00008000008000008F00007F60808F6F908F6F8F8F70908F6F8F8F70908F7090907090403040\
4F3F4F000000000000000000000000000000000000000000000000000000000000
000000003F00003000208F00208000208F00208000208F00208000001F00000000000000208000\
208F00208F00208F00208F00208F00003F00209000208F00209000208F002F9000
208F002090002090000F0F0F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C09F9FCFA09FC09F9FCFA09F
C09F9FCFA09FC09F9FCFA09FC09F90CFA09FC09F90CFA09FC09F90CFA09FC09F90CFA09FC09F90\
C09F90BF9F8F50A020209F002FA0004F9F1F50A020BF9F8FBF9F8FC09F8F50A020
709F5070A050709F507FA050709F5070A05070A05070A050709F4F70A05070A04F70A050709F4F\
70A05020A0002FA000BF9F7FBFA07FBF9F7F5F503F000000AF9070AF9070BF9F7F
B09F70BF9F7FB09F70B09F70B09070B09F70AF906FB09F70AF906FAF906FAF906FAF906FA09060\
AF9060A08F60AF9060A08F5FA0503FA00000A000009F30209F8F5F9F6F409F0000
90804F9080500000000000000000000000000000000000000000000000000000000000005F5030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003030403F30403030403F304F3F304F3F30503F00004000004000004000004000004F0000\
4F00004F3F604F3F6050406F50406F50406000000F5F4F705F406F5F3F5F5F2F40
60304F6000006000006000006F00006F00006F00006F0000700000700000700000703F507F608F\
7060807F608F7060807F608F7F608F7F608F7F608F806F8F7F6080802F3F800000
8000008000008F000080101F8F1F208F6F908F70908F70909070908F70909070909070904F3F4F\
4F3F4F000000000000000000000000000000000000000000000000000000000000
000000206F00206F00208000208F00208000208F002050002050000F00000000000000002F8F00\
208F00207000206F000F0F0F003F00000000001F00002000208F002F9000209000
2F90002090002F9000001F00002000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA09FCFA0A0C09F9F
CFA0A0CFA09FCFA09FC09F9FCFA09FCFA09FCFA09FC09F9FCFA09FCFA09FCFA09FC09F90CFA09F\
50A03F50A0402FA0002FA0002FA000BF9F8FBF9080CFA090C0A090CFA090C0A08F
CFA090C0A08FC0A08FC0A08FC0A08FC0A08FC0A08FC0A080C0A08FC0A080C0A080C0A080C0A080\
C0A08070A05070A04FC0A080BFA07FC0A0805F4F3F000000000000000000AF9070
BFA07FB09F70BF9F7FB09F70BF9F7FB09F70B09F70B0906FB09F70AF906FB09F6FAF9060AF906F\
AF9060AF9060A08F60AF9060A0704FA00000A00000A000009F8F5F9F8F5F9F0000
9F50309080508F804F000000000000000000000000000000000000000000000000000000605F30\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
1F1F20303040302F4030304F30304F3F30503F1F303F0F103F1020400000400000400F10400000\
4F203F4F3F604F406F4F3F604F3F6000000050406F50406F5F4F705F2F405F0000
5F00006000006000006000006000006F00006F00006F00006F0000700000700000705F8F705F80\
7F608F705F807F608F7F60807F608F7F60807F608F7F608F806F8F7F3F50800000
8000008000008000008F00008000008F405F8F40508F70908F70909070908F7090907090403F4F\
4F3F4F000000000000000000000000000000000000000000000000000000000000
001F00208000208000208000208000208F00003000000000000000000000000000208F00208F00\
208F00206F00207000000000000000000000000000000000207000207000209000
2090002F9000209000003F00003F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA09FC09F9FCFA09F
C09F9FCFA09FC09F9FCFA09FC09F9FCFA09FC09F90CFA09FC09F90CFA09FC09F90CFA09FBF908F\
2FA000209F002FA0002F9F0050A020C09F90CFA090C0A08FCFA090C09F8FC0A08F
C0A08FC0A08FC09F8FC0A08FC0A08FC0A08FC09F80C0A08FC0A080C0A08FBFA080C0A080BFA080\
C0A080BFA07FC0A080BFA07FC0A080BF9F7F5F4F3F000000000000000000000000
000000AF9070B09F70B09F70B09F70B09F70AF906FB09F70AF906FB09F6FAF9060AF906FAF9060\
AF9060A08F60AF9060A08F5FA08F5FA00000A000009F0000A08F5F9F80509F3F20
901F109080508F7F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
1F1F203030403F304F30304F3F304F3F304F4000003F3050403F604010204F203F402F504F3F60\
4F3F6050406F4F3F604F3F600000000F0F0F50406F5F4F705F406F6020305F0000
6000006000006000006000006F00006F00006F101F6F0000700000702030702F3F705F807F608F\
705F807F608F7F608F7F608F7F608F806F8F7F608F806F8F7F608F7F608F800000
8000008000008F00008000008F00008F00008F00008F2F3F9070908F7090907F9F9070904F3F4F\
000000000000000000000000000000000000000000000000000000000000000000
002000208000208F00208000208F002080000F0F0F000000000000000000000000208F00208F00\
208F002F70002070000000000000000000000000000000002070002F7F00209000
2F90002090002F9F00003F00004000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA09FCFA0A0CFA09F
CFA0A0CFA09FCFA0A0CFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09F7FA05F\
2FA0002FA0002FA0002FA00050A01FCFA090C0A090CFA090C0A08FCFA090C0A08F
CFA090C0A08FCFA090C0A08FC0A08FC0A08FC0A08FC0A08FC0A08FC0A080C0A08FC0A080C0A080\
C0A080C0A080C0A080C0A080BFA07FC0A0805F4F3F000000000000000000000000
000000000000AF9070AF906FAF906FA08F6FAF906FA08F6FA0906FAF906FB09F6FAF906FAF906F\
AF9060AF906FAF9060AF9060A08F5FAF201FA00000A000009F3F20A08F5F9F8050
9F00008F7F4F908050000000000000000000000000000000000000000000000000000000000000\
2F201F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
1F1F1F303040302F403F304F30304F3F30503F0F10403F5F3F305F403F604030604F3F60403F60\
4F3F60403F5F4F3F600000000000000000005F406F50406F5F4F705F0F10600000
5F00006000006000006F101F60101F6F507F6F304F7000006F0000703F50705F8070608F705F80\
7F608F705F807F608F7060807F608F7F608F80608F7F608F806F8F7F608F800000
8000008000008000008F00008000008F00008F00008F0000806F8F8F70908F7090907F9F403F40\
000000000000000000000000000000000000000000000000000000000000000000
001F00208000208000208F00208000208F00000000000000000000000000000000208F00208F00\
208F00206F00207000000000000000000000000000000000209000208F002F9000
209000209000209000003F00003F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA09FC09F9FCFA09F
C09F9FCFA09FC09F9FCFA09FC09F9FCFA09FC09F9FCFA09FC09F90CFA09FC0A090CFA09F709F5F\
2FA0002FA0002FA000209F002FA0009FA050A0A05FC09F8FCFA090C0A08FCFA090
C0A08FC0A08FC0A08FC0A08FC0A08FC0A08FC0A080C0A08FC0A080C0A080C0A080C0A080BFA080\
C0A080BFA080C0A080BF9F7FBFA080BF9F7F5F5040000000000000000000000000
000000000000000000000000000000000000000000000000000000000000A08F60A08F609F805F\
A08F609F805F9F8F5F9F805F9F8F5F9F8050A03F2F9F0000A01F109F80509F8F50
8F7F4F9000008F7F4F000000000000000000000000000000000000000000000000000000000000\
504F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
201F203030403F304F30304F3F30503F30504000003F305F403F60403F604F3F60403F604F406F\
403F5F0000000000000000000000000F0F0F50406F5F4F705F4F705F101F5F0000
60304F604F7060507F60507F6F5F806F304F6F00006F0000700000705F8070608F705F807F608F\
70608F7F608F7F608F7F608F7F608F806F8F7F608F806F9080608F806F90803F50
8F00008000008F00008000008F00008F00008F00008F0000904F5F8F405F907F9F9070904F3F4F\
000000000000000000000000000000000000000000000000000000000000001F00
002000208000208F00208000208F00003F00003F00000000000000000000000000208F002F8F00\
208F00207000207000000000000000000000003F00003F002090002F9000209000
2F90002090002F9000001F00002000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA09FCFA0A0CFA09F
CFA0A0CFA09FCFA0A0CFA09FCFA0A0CFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09F7FA060\
2FA0002FA0002FA0002FA0002FA0002FA0002FA00050A02050A01F7FA05F70A050
7FA05F7FA0507FA05F70A0507FA05F70A0507FA050C0A080C0AF8FC0A080C0A08FC0A080C0AF8F\
C0A080C0A080BFA080C0A080BFA080C0A0805F4F3F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000A05F3F9F0000A08F5F9F8050
9F8F50905030908050000000000000000000000000000000000000000000000000000000000000\
504F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F3F303040302F403F304F30304F3F30503F2F4040203F40305F403F6040305F403F5F40305F\
0000000000000000000000000000000000005F406F50406F5F4F705F00005F2030
5F4F7060507F604F7060507F604F706F00006F00006F00006F2030705F8F705F8070608F705F80\
7F608F705F807F608F7F608F7F608F7F608F7F608F7F608F806F8F7F608F804050
8000008000008000008F00008000008F00008F00008F00008F2F3F90303F8F7090907F9F403F4F\
000000000000000000000000000000000000000000000000000000000000000000
000000208F00208000208F00208000206F00206F000F0F0F000000000000000000208F00208F00\
208F00206F00207000003F00205F00205000209000208F002F9000208F00209000
2090002F7F00207000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0C09F9FCFA09F
CFA09FCFA09FC09F9FCFA09FCFA09FCFA09FC09F9FCFA09FCFA09FCFA09FC09F90CFA09FA0A07F\
2FA000209F002FA0002FA0002FA00020A0002FA0002FA0002FA00020A0002FA000
2FA0002FA00020A0002FA00020A0002FA00020A0002FA00020A0007FA05070A04F70A05070A04F\
70A05070A04FC0A080BFA07FC0A080BF9F7F5F503F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000A000009F6F409F8F50
9080509F80508F7F4F0F0000000000000000000000000000000000000000000000000000000000\
505030000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3030403030403F304F3F304F3F30503F3050403F5F40305F403F60403060403F60000000000000\
0000000000000000000000000000000F0F0F50406F5F4F705F407060304F5F406F
60507F604F706F507F60507F6F20306F00007000006F0000703F50705F807F608F70608F7F608F\
70608F7F608F7F608F7F608F7F608F806F907F608F806F9080608F806F90804050
8000008000008F00008F00008F00008F00009000008F00009000008F0000907F9F90709F4F3F4F\
000000000000000000000000000000000000000000000000000000000000000000
000000205000208F00208F00208F00208F00208F00208F00003F00003F00002000208F00208F00\
208F002F9000208F002F9000208F002F90002090002F90002090002F9000209000
2F9000003F00004000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA09FCFA0A0CFA09F
CFA0A0CFA09FCFA0A0CFA09FCFA0A0CFA09FCFA0A0CFA09FCFA09FCFA09FCFA09FCFA09FCFA09F\
50A02050A0202FA0002FA0002FA0002FA0002FA0002FA0002FA0002FAF002FA000
2FA0002FA0002FAF002FA0002FA0002FA0002FAF002FA0002FA0002FA0002FAF002FA0002FAF00\
2FA0002FAF0070A04FC0A080BFA07FC0A0805F4F3F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000A05F3F9F8F50
9F8F509080509F8050000000000000000000000000000000000000000000000000000000000000\
2F2F1F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F3F303040302F403F304F30304F3F30503F3050403F5F40305F40305F000000000000000000\
0000000000000000000000000000000000005F407050406F5F4F705F4F70604F70
5F4F7060507F604F7F6F00006000006F00006F000070000060507F705F8F705F8070608F705F80\
7F608F70608F7F608F7F608F7F608F7F608F806F8F7F608F806F8F80608F805070
8000008000008000008F00008000008F00008F00009000008F0000900000907090907F9F403F4F\
000000000000000000000000000000000000000000000000000000000000000000
000000001F00208000208F00208000208F00208000208F00208000208F00208F00208F00208F00\
208F00208F00209000208F00209000208F00209000208F002090002090002F5F00
205000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0CFA09FCFA0A0
C0A09FCFA09FCFA09FCFA09FC0A09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA090\
C09F90C09F9050A0204FA01F2FA0002FA0002FA00020A0002FA0002FA0002FA000
20A0002FA0002FA0002FA00020A0002FA00020A0002FA00020A0002FA00020A0002FA00020A000\
2FA00020A0002FA000BFA07FC0A080BFA07FC0A0805F4F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000009F1F109F8F5F
9080509F805090804F0F0F00000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F30403030403F304F30304F3F30503F30503F30503F3050000000000000000000000000000000\
0000000000000000000000000000005F407050406F5F4F705F4F70604F705F4F70
60507F60507F6F101F6000006F00006F00007000006F000070608F705F8F70608F705F8F7F608F\
7F608F7F608F7F608F7F6F907F608F806F907F608F806F90806F8F806F90805070
8F00008000008F00008F00008F00008F00009000008F0000900000900000907F9F90709F4F404F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000003F00003000208F00208F00208F00208F002F8F00208F00208F00208F002F9000\
208F002F9000208F002F9000208F002F90002090002F9000003F00003F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0CFA0A0CFA09F
CFA0A0CFA09FCFA0A0CFA09FCFA0A0CFA09FCFA0A0CFA09FCFA0A0CFA09FCFA09FCFA09FCFA09F\
CFA09FCFA09FCFA090CFAF9FA0A05FA0A05F7FA03F7FAF4050A01F50A02050A01F
50AF2050A01F50AF2050A01F50AF2050A01F50AF204FA01F50AF2050A01F50AF204FA01F50AF20\
2FA0002FAF002FA0002FAF00C0A080C0A080BFA07F605040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000908050
9F8F5F9080509F8050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
302F40303040302F4030304F302F4F000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F50406F5F40705F406F5F4F705F4F70604F70
604F70604F7F6000006F00006F00006F00006F000070101F6F5F8070608F705F8070608F705F8F\
7F608F70608F7F608F7F608F7F608F7F608F806F8F7F608F806F9080608F805070
8000008F00008000008F00008F00008F00008F00008F00008F101F901F208F7090907F9F403F4F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000003000208F00208000208F00208F00208F00208F00208F00208F00\
208F00208F00208F00207000003F00003F00002000001F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0CFA09FCFA0A0
CFA09FCFA0A0CFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09F\
CFA09FCFA090CFA09FCFA090CFA09FC0A090CFA090C0A090CFA090C0A08FCFA090
C0A08FCFA090C0A08FC0A08FC0A08FCFA08FC0A08FC0A08FC0A080C0AF8FC0A080C0A080C0A080\
C0AF804FA01F50A01F20A000C0A080BFA07FC0A0805F4F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8F7F4F9F8F5090804F0F0F00000000000000000000000000000000000000000000000000000000\
000000400000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000001F1F20
3F30403030403F304F000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000F5F4F705F406F5F4F705F4F70604F7F604F70
60507F6000006F00006F00006F00006F000070000070203070608F705F8F7F608F70608F7F608F\
70608F7F60907F608F7F6F907F608F806F907F608F806F90806F8F806F90805070
8F00008000008F00008F00008F00008F00009000008F000090304090303F907F9F907F9F4F3F4F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F003F00207000206F00000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0CFAFA0CFA0A0
CFA0A0CFA09FCFAFA0CFA09FCFA0A0CFA09FD0AFA0CFA09FCFA0A0CFA09FD0AFA0CFA09FCFA09F\
CFA09FCFAF9FCFA09FCFA09FCFA090CFAF9FCFA090CFAF90CFA090CFAF90CFA090
CFAF90C0A08FCFAF90C0A08FCFAF90C0A08FCFAF8FC0A08FCFAF8FC0A08FCFAF8FC0A080C0AF8F\
C0A080C0AF8FC0A08050AF1FC0A080C0AF80BFA080605040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000008F7F4F9F8F508F7F4F000000000000000000000000000000000000000000000000000000\
000000504F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000001F1F20
302F40302F40000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F50406F5F4F705F406F5F4F705F4F70604F7F
5F2F406000006000006F00006000006F00006F304F703F506F5F8070608F705F8F7F608F705F8F\
7F608F7F608F7F608F7F608F7F608F7F608F806F907F608F806F90805070805F70
8000008F00008000008F00008F00008F00008F00009000008F405F904F5F9070904F3F4F403F4F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0CFA09FCFA0A0
CFA09FCFA0A0CFA09FCFA0A0CFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09FCFA09F\
CFA09FBF9F8F7FA05F7FA05F7FA03F7FA03FCFA090C0A090CFA090C0A090CFA090
C0A08FCFA090C0A08FCFAF90C0A08FC0A08FC0A08FCFAF8FC0A080C0A08FC0A080C0AF8FC0A080\
C0AF80C0A080C0AF80BFA080C0A080BFA07FC0A0805F4F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000008F7F4F90804F000000000000000000000000000000000000000000000000000000\
0000004F4F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000001F1F20
3F304F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F000F5F4F705F406F5F4F705F4F70604F7F604F70
6020306000006F00006F00006F00006F0000704F6F6F5F8070608F705F8F7F608F70608F7F608F\
7F608F7F608F7F608F806F907F608F806F907F608F806F90806F9080405F804050
8F00008000008F00008F00008F00008F00009000008F00008F70908F7090907F9F4F3F4F4F4050\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000003F00003F00205F002070002F7000\
2070002F7000207000003F00003F00003F000000000F0F0F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0CFA0A0CFA0A0
D0AFA0CFA0A0CFA0A0CFA09FD0AFA0CFA09FCFAFA0CFA09FCFA09FCFA09FA0AF80A0A07F50AF20\
50A0202FAF002FA0002FAF002FA0002FAF002FA0002FAF0050A02050AF2050A01F
50AF207FA03F7FAF5F70A050A0AF709FA070CFAF8FC0A08FCFAF8FC0A08FCFAF8FC0A080CFAF8F\
C0A080C0AF8FC0A080C0AF80C0A080C0AF80BFA07F605040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008F7F4F000000000000000000000000000000000000000000000000000000\
000000505030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000303040
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F50406F5F4F705F40705F4F705F4F70604F7F
6000006000006000006F00006F00006F00006F406F705F8F705F8070608F705F8F7F608F70608F\
7F608F70608F7F608F7F608F806F907F608F806F907F608F806F90802F3F801F20
8000008F00008000008F00008F00008F00008F2F3F90303F8F7090907F9F9070904F3F4F4F3F4F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000001F00208F00208F00208F00208F00208F00\
209000208F00209000208F00209000208F00209000209000003F00003F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0CFA09FCFA0A0
CFA09FCFA0A0CFA09FCFA0A0CFA09FCFA0A0C09F90C0A09F50A0202FA0002FA0002FA0002FA000\
2FA0002FA0002FA0002FA0002FAF0020A0002FA0002FA0002FAF0020A0002FA000
2FA0002FAF0020A0002FA0002FA0002FAF0020A01F2FAF1F70A0507FAF50BF9F80C0AF8FC0A080\
C0AF8FC0A080C0AF80C0A080C0AF80BFA07FC0A0805F4F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F000F0F000F0F0F000000000000000000000000000000000000000000\
000000504F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000302F3F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000F5F4F705F40705F4F705F4F7060507F600000
6000006000006F00006F00006F00006F0000704F6F705F8070608F705F8F7F608F70608F7F608F\
7F608F7F6F907F608F806F907F608F806F90806F90806F90806F908F1F20800000
8F00008000008F00008F00008F00008F1F1F8F70908F7090907F9F90709F9F7F9F4F3F4F4F404F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000002000206F002F7000208F00208F00208F002F9000208F002F9000\
208F002F90002090002F90002090002F90002090002F90002090002F9F00205F00
2F5F00000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0D0AFA0CFA0A0
D0AFA0CFA0A0D0AFA0CFA0A0D0AFA0A0A05F50AF4050A0402FAF002FA0002FAF002FA0002FAF00\
2FA0002FAF002FA0002FAF002FA0002FAF002FA0002FAF002FA0002FAF002FA000
2FAF002FA0002FAF002FA0002FAF002FA0002FAF002FA0002FAF002FA0002FAF202FAF1FA0AF6F\
C0A080CFAF8FC0A080C0AF80C0A080C0AF80C0A080605040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000090804F8F7F4F8F804F0F0F000F0F0F0F0F00000000000000000000\
000000302F20000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F50406F5F4F705F40705F4F705F4F70600000
6000006000006000006F00006F00006F00006F406F705F8F705F8070608F705F8F7F608F70608F\
7F608F7F608F7F60907F608F7F6F907F608F806F907F608F7F608F8000008F0000
8000008F00008000008F00008F00008F5F7F8F7090907F9F8F7090907F9F90709F4F3F4F403F4F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000003000003F00208F00208F00208F00208F00208F00208F00208F00209000208F00\
209000208F00209000208F002090002090002F90002090002090002090002F9F00
2090002F5F00205F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0CFA0A0CFA0A0
CFA09FCFA0A0CFA09FCFAFA0CFA09F2FA0202FA0002FAF002FA0002FA0002FA0002FAF002FA000\
2FA0002FA0002FAF002FA01F50A03F50A03F7FAF5F70A0507FAF5F7FA05F7FAF5F
70A0507FAF3F2FA0002FAF0020A0002FAF002FA0002FAF0020A0002FAF002FA0002FAF0020A000\
9FAF509FA04FC0AF80C0A080C0AF80BFA080C0AF805F4F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F008F7F4F90804F8F7F408F7F4F804020804F2F7F40200F0F0F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F000F5F4F705F4070604F7F5F4F7060507F60101F
6F00006000006F00006F00007000006F0000703F50705F807F608F70608F7F608F70608F7F6090\
7F608F7F6F907F608F806F907F608F806F90806F8F806F90802F3F8F0000800000
8F00008F00008F00008F5F7090607F8F7090907F9F8F709F907F9F907F9F907F9F4F3F4F4F4050\
000000000000000000000000000000000000000000000000000000000000000000
000000001F00208F00208F002F8F00208F00208F00208F002F7000205000205F00003F00003F00\
003F002F5F002070002F7F002090002F90002090002F9F002090002F9F00209000
2F9F00209F002F9F00002000002000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0D0AFA0CFA0A0
D0AFA0CFA0A0D0AFA0CFA0A0AFAF802FA0002FAF002FA0002FAF002FA0002FAF002FA00050AF20\
50A020A0AF7FA0A07FD0AF9FCFA09FCFAF9FCFA090CFAF9FCFA090CFAF90CFA090
CFAF90CFA090CFAF90C0A08F2FAF002FAF002FAF002FA0002FAF002FAF002FAF002FA0002FAF00\
2FAF002FAF00C0A080C0AF8FC0A080C0AF80C0A080605040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000009080508F804F8F804F8F7F408F00008000008000007F00007F4020\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F3F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F50406F5F4F705F4F70604F7F5F4F70602030
6000006F00006000006F00006F00007000006F304F705F8F705F8F7F608F705F8F7F608F70608F\
7F60907F608F7F60907F608F806F907F608F806F90803F50800000800000800000
80101F806F90806F8F8F70908F709090709F8F709090709F90709F907F9F90709F4F3F4F4F3F4F\
000000000000000000000000000000000000000000000000000000000000000000
000000207000208F00208F00208F00208F00208F00001F00000000000000000000000000000000\
000000000000000000000000002000001F00207F002070002F90002090002F9F00
2090002F9F00209000004000003F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000D0AFA0CFA0A0CFA0A0
CFA0A0D0AFA0CFA09FCFA0A07FA0602FAF002FA0002FA0002FA0002FAF002FA00050AF20C09F90\
CFA09FCFA09FCFAF9FCFA09FCFAF9FCFA090CFAF9FCFA090CFAF90C0A090CFAF90
CFA090CFAF90C0A08FCFAF90C0A08FBF9F804FA03050AF3F2FA0002FAF0020A0002FAF002FA000\
2FAF0020A00050AF1FC0A080C0AF80BFA080C0A0805F4F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008F7F4F90804F8F7F408F7F4F8070408000007F00007F0000700000\
0000000000004F4030000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000030303F1F1F20000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000F5F4F705F4F70604F7F5F4F7060507F60304F
6F00006000006F00006F00006F00006F0000702F3F705F8F70608F70608F7F60907F608F7F6090\
7F608F806F907F608F806F907F608F806F908000008000008000008F5F7F806F90
8F70908F6F908F709F8F709090709F8F7090907F9F90709F907F9F907F9F9F7FA04F3F4F4F4050\
000000000000000000000000000000000000000000000000000000000000001F00
002000208F00208F00208F00208F00003F00000000000000000000000000000000000000000000\
0000000000000000000000000000000000000020000020002090002F9F00209000
2F9F002090002F9F00003F00004000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0D0AFAFCFA0A0
D0AFA0CFA0A0D0AFA0CFA0A07FAF402FA0002FAF002FA0002FAF007FA05F2FAF00C0A09FD0AFA0\
CFA09FD0AFA0CFA09FD0AF9FCFA09FD0AF9FCFAF90CFAF9FCFA090CFAF9FCFAF90
CFAF90CFA090CFAF90CFAF90CFAF90C0A08FCFAF90CFAF8FA0AF5F9FAF502FAF002FAF002FAF00\
2FAF002FAF002FAF00C0AF8FC0A080C0AF80C0A080605040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000009080508F804F90804F8F7F4F8F7F4F7F703F8000007F00007F0000\
703F200000004F4F30000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003030401F101F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F5040705F4F705F4F70604F7F5F4F7060507F
6000006F00006000006F00006F00007000006F101F70608F705F8F70608F70608F7F608F70608F\
7F608F7F608F7F6F907F608F806F907F3F50800000800000801F20806F908F6F90
806F908F70908F6F908F709F8F70908F709F8F709F907F9F8F709F907F9F907F9F4F3F4F403F4F\
000000000000000000000000000000000000000000000000000000000000002000
001F00208F00208F00208F00208F00000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F7F002070002F9F00
2090002F9F00209F00004000003F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFAFA0CFA0A0D0AFA0
CFA0A0D0AFA0CFA0A0D0AFA050A01F2FAF002FA0002FAF002FA000A0AF80A0A07FD0AFA0CFA09F\
CFAF9FCFA09FCFAF9FCFA090CFAF9FCFA090CFAF9FCFA090CFAF90CFA090CFAF90
C0A090CFAF90CFA08FCFAF90C0A08FCFAF90C0A08FCFAF8FC0A08FCFAF8F70A0307FAF3F20A000\
2FAF002FAF002FAF009FA06FC0AF80C0A080C0AF805F4F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000090804F8F7F4F8F7F4F807040807F407F5F307F0000700000\
700000000000504F30000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000030304030303F1F1F20000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F5F40705F4F705F4F70604F7F5F4F7F60507F604F7F
6F305060101F6F00006F00007000006F0000700000704F6F7F609070608F7F60907F608F7F6F90\
7F608F7F6F907F608F802F3F8000008000007F6080806F90806F908F7090806F90
8F709F8F70908F709F8F7090907F9F8F709F907F9F90709F907F9F907F9F9F7FA04F3F4F4F4050\
000000000000000000000000000000000000000000000000000000000000001F00
002000208F00208F00208F002F90000F000F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000207F002F7F00209000
2F9F00209F002F9F00002000002000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFA0A0D0AFA0CFA0A0
D0AFAFCFA0A0D0AFA0CFA0A05FAF202FA0002FAF002FA0002FAF00CFAFA0D0AFA0CFA09FD0AFA0\
CFAF9FD0AF9FCFA09FD0AF9FCFAF9FD0AF9FCFA090D0AF9FCFAF90CFAF9FCFAF90
CFAF9FCFAF90CFAF90CFAF90CFAF90CFAF8FCFAF90C0AF8FCFAF90CFAF8FC0A08FBFA0802FAF00\
2FAF002FAF002FAF007FAF3FC0AF80C0AF80C0A080605040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F90804F90804F8F7F408F7F4F807F40807F407F00007F0000\
701F0F000000504F30000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F3F303040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F4F705040705F4F705F4F70604F7F5F4F7060507F
60507F60507F60101F6F00006F00007000006F0000700000703F507F608F705F8F70608F7F608F\
7F405F7F3F507F2F3F7F101F80405F7F608F806F90806F90806F90806F908F6F90
806F908F709F8F6F908F709F8F709090709F8F7090907F9F90709F907F9F907F9F4F3F4F4F3F4F\
000000000000000000000000000000000000000000000000000000000000000000
208F00208F00208F00208F00208F004070205040505F4F5F504F505F4F5F504F505F4F5F5F4F50\
5F4F5F5F4F505F4F5F5F4F505F4F5F5F4F504F5F404F5F402F9F002090002F9F00
2090002F7F20207F2060505F604F5060505F604F5060505F6050506F505F6050506F5050605050\
6F505F6050506F50506050506F5F5F6050506F5050CFA0A0605050604F4F605050
CFA0A0D0AFA0CFA0A0D0AFA050A0202FAF002FA0002FAF002FA000D0AFA0CFA09FD0AFA0CFA09F\
D0AF9FCFA09FCFAF9FCFA09FCFAF9FCFA090CFAF9FCFA090CFAF9FCFA090CFAF90
CFA090CFAF90C0A08FCFAF90CFA08FCFAF90C0A08FCFAF8FC0AF8FCFAF8FC0A08FCFAF8F4FAF1F\
50AF1F20A0002FAF0020AF00C0AF80C0A080C0AF805F503F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000009080508F7F4F8F7F4F807F40807F407F703F7F703F7F0000\
7F1F0F000000505030000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000030304030303F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F5040705F4F705F4F70604F7F5F4F7F60507F60507F
6F508060507F60507F6F101F7000006F00007000007000007F0000702F3F7F00007F2F3F7F2F3F\
7F101F801F207F101F806F907F6090806F90806F90806F90806F908F709F8F6F90
8F709F8F70908F709F8F709F90709F8F709F907F9F907F9F907F9F4F3F4F4F40504F3F4F4F4050\
4F405050405F504050504050504050504F5F5040505F4F5F5040505F4F5F504F50
407020A080A04F902F208F00209000208F006F906F6F8F60AF90AFAF8FA0B090AFAF90AFB090AF\
AF90AFB090AFB090AFB090AF8F9080909F8F4F90204F9F202090002F9F00209F00
2F9F00909F809F9F8FBF9FA0C09FAFBF9FA0C0A0AFC09FA0C0A0AFC09FA0CFA0AFC0A0A0CFA0AF\
C0A0A0CFA0AFCFA0A0CFAFAFCFA0A0D0AFAFCFA0A0D0AFAF000000000000000000
6F505060504F6F505060505080AF402FA00030AF002FAF002FAF00C0A09FCFA09FCFAF9FD0AFA0\
CFAF9FD0AFA0CFAF9FD0AF9FCFAF9FD0AF9FCFAF9FD0AF9FCFAF907FAF5F7FAF5F
7FAF5F7FAF5F7FAF5FA0AF5FA0AF5FCFAF8FCFAF90CFAF8FCFAF90C0AF8FCFAF90CFAF8F7FAF50\
70AF502FAF002FAF002FAF00BFA07FCFAF8FC0AF806050405F503F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F8F7F4F9080508F7F4F8F7F4F807F40807F407F703F7F703F\
7000000F0F0F2F2F1F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F3F303040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F4F705F40705F4F7F5F4F70604F7F604F7F60507F
60507F6F50806F50806F50806F2030700000704F6F7010207000007F0000703F507F1F207F101F\
7F60907F608F806F907F608F806F90806F90806F903F304040304F403040403F4F
403040403F4F40304F403F4F40304F4F3F4F403F4F4F3F4F0000000000000000005040504F3F50\
9F80A09F7F9F9F80A09F80A0A080A09F80A0A08FA0A080A0A08FAFA08FA0AF8FAF
608F60608F60608F40208F00208F00208F00208F00209000208F204F9040408F406F906F608F60\
6F90606F90606F906F4090204F90202090002F9F002090002F9F00209000709F40
6F9040709F60709F60B0909FAF8F9FC09FAFBF9FA0C0A0AFC09FA0C0A0AFC09FA0CFA0AFC09FA0\
CFA0AFC0A0A0CFA0AFCFA0A0CFA0A0CFA0A0D0AFAFCFA0A0000000000000000000
0000000000000000000000005080202FAF002FA0002FAF002FA000AFAF60A0A060D0AFA0CFA09F\
D0AF9FCFA09FD0AF9FCFA09FCFAF9FCFA09FCFAF9FCFA090CFAF9F2FAF002FAF00
2FA0002FAF002FAF0050AF204FA01FCFAF90CFAF8FCFAF90C0AF8FCFAF8FC0AF8FCFAF8F70AF30\
7FAF3F2FAF002FAF0020A0009FAF6FBFA07FC0AF80BFA080605040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008F804F8F804F8F804F807F40807F4080703F8070407F6F3F\
7F1F0F6F0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000303040303040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F5F40705F4F705F4F7060507F604F7F60507F60507F
6F50806F50806F5F806F5F80705F8F705F8F70608F705F8F70608F703F507F60907F608F7F6F90\
7F6090806F907F6090806F90806F9040304F3F3040000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F4050504050\
907F906080605F8060608060608060608F60608060608F6F608F606F8F6F208F00
2F9000208F00209000208F002F9000208F00209000208F002F90002090002F90002090002F9000\
2090002F90002090002F9F002090002F9F002090002F9F00209F002F9F00209F00
2F9F002F9F002F9F009F9F809FA08FC09FAFC0A0AFC09FA0CFA0AFC0A0AFCFA0AFC0A0A0CFA0AF\
CFA0A0CFA0AFCFA0A0D0AFAFCFA0A0D0AFAFCFA0A0D0AFAF000000000000000000
000000000000000000000000002F002FAF002FAF002FAF0030AF002F8F202F8F2060504F6F5050\
60504F6F505060504F6F505060504FD0AF9FCFAF9FD0AF9FCFAF902FAF002FAF00
2FAF002FAF002FAF0050AF1F50AF20CFAF90CFAF90CFAF8FCFAF90CFAF8FCFAF8FC0AF8F50AF20\
50AF1F2FAF002FAF002FAF002FAF00A0AF6FC0AF80C0AF805F5040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F8F804F90804F8F7F4F8F804F807F40807F407F703F807040\
7F5F2F700000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000302F3F303040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F4F705F40705F4F7F5F4F7060507F604F7F60507F
60507F6F50806F50806F5F806F5F80705F8F705F8F70608F70608F7F609070608F7F60907F608F\
7F6F907F608F40304F3F3040000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005040504F4050\
7F80801F7F002080002080002080001F8000208F00208000208F00208000208F00
208F00208F00208F00208F00208F00209000208F00209000208F00209000208F00209000209000\
2090002090002090002090002F90002090002F9F002090002F9F002090002F9F00
209F002F9F00209F009F9F8F9F9F80C0A0AFBF9FA0C0A0AFC09FA0C0A0AFC09FA0CFA0AFC0A0A0\
CFA0AFC0A0A0CFA0AFCFA0A0CFAFAFCFA0A0D0AFAFCFA0A0000000000000000000
0000000000000000000000000000002F8F002F80002FAF002FA0002FAF002FA0002F60002F6000\
00000000000000000000000000000000000060504F60504F60504F2FA0002FAF00
2FAF002FAF002FAF002F8F202F801F60504F6050406050406050406050404F6030506F3F2FAF00\
2FAF0020AF002FAF002FAF002FAF009FA06FC0AF80C0A080605040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000090804F8F804F8F804F8F7F408F7F4080703F8070407F703F\
7F703F6F60300F0F00404030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000030303F3030401F1F20000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F5F40705F4F7F5F4F7060507F604F7F60508060507F
6F50806F50806F5F8F6F5F80705F8F705F8F70608F70608F7F609070608F7F6F907F609040304F\
3F3040000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F4050504050\
7F7F80208000208000208000208000208F00208000208F00208F00208F00208F00
208F00208F002F9000208F00209000208F002F90002090002F9000208F002F90002090002F9000\
2090002F90002090002F9F002090002F9F002090002F9F002090002F9F00209F00
2F9F004F9F2050A02F9F9F8F9FA08FC09FAFC0A0AFC09FAFC0A0AFC0A0A0CFA0AFC0A0AFCFA0AF\
CFA0A0CFAFAFCFA0AFCFAFAFCFA0A0D0AFAFCFAFA0D0AFAF000000000000000000
000000000000000000000000000000002000002F002FAF002FAF002FAF0030AF002FAF002FAF00\
2F8000308F000020000000000000000000000000000000000000002FAF002FAF00
2FAF002FAF002FAF002F8F002F8F000000000000000000000000000000000000000040002FAF00\
2FAF002FB0002FAF002FAF002FAF00BFA080C0AF80C0AF805F503F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008F804F9080508F804F8F804F807F408F7F4080703F807040\
7F6F3F704020603F204F4F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000030303F302F3F1F1F20000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F4F705F4F70604F7F5F4F7060507F604F7F605080
60507F6F50806F50806F5F8F6F5F80705F8F705F8F70608F705F8F3F30403F30403F304F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005040504F3F50\
7F80801F80002080001F8000208000208000208F00208000408F20408F20408F20
408F20408F20408F20409020408F20409020408F204F9020408F204F9020408F204F9020408F20\
4F90204090204F90204090204F90204F90204F9F204F9020BF9FAFBF90AFBF9FAF
BF9FA0BF9FAFBF9FA0C09FAFBF9FA0C09FAFC09FA0C0A0AFC09FA0C0A0AFC0A0A0CFA0AFC0A0A0\
CFA0AFCFA0A0CFAFAFCFA0A0CFAFAFCFA0A0D0AFAFCFA0A0000000000000000000
0000000000000000000000000000000000000000002F6F002FA0002FAF002FAF002FAF002FA000\
2FAF002FAF002F6F000000000000000000000000000000000000002FAF002FAF00
2FAF002FAF002FAF002F8F002F80000000000000000000000000000020000020002FAF0020AF00\
2FAF002FAF002FAF0020AF002FAF005F5040C0AF80C0A080605040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000090804F8F804F90804F8F7F408F7F40807040807F407F703F\
7F703F704020600000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000030303F3030401F1F20000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F5F4F705F4F7F5F4F7060507F604F7F60508060507F
6F5F806F50806F5F8F302F403F30403F2F403F304F3F2F403F304F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F405050405F\
7F8080408020408020408F20408020408F20408020408F2FA08FAFAF8FAFA08FAF
AF90AFAF8FAFAF90AFAF8FAFAF90AFAF8FAFB090AFAF8FAFB090AFAF90AFB090AFAF90AFB090AF\
B090AFB090AFB090AFBF9FAFB090AFBF9FAFB090AFBF9FAFBF9FAFBF9FAFBF9FAF
C09FAFBF9FAFC0A0AFBF9FAFC0A0AFC09FAFC0A0AFC09FAFCFA0AFC0A0AFCFA0AFC0A0AFCFAFAF\
CFA0AFCFAFAFCFA0A0D0AFAFCFA0AFD0AFAFCFAFA0D0AFAF000000000000000000
00000000000000000000000000000000000000000000000030AF002FAF002FAF002FAF0030AF00\
2FAF002FAF000040000000000000000000000000000000000000002FAF002FAF00
2FB0002FAF002FAF002FAF002FB0002FAF002FAF002FAF002FB0002FAF002FB0002FAF002FB000\
2FAF002FB0002FAF002FB0000040000000005F50406050405F5040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000009080508F804F8F804F8F7F408F7F4080703F807040\
7F703F7F703F600000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000030303F302F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000F5F4F705F40705F4F7F2F203F302F3F2F203F302F40
302F3F302F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005040504F4050\
9F80A09F7FA09F80A09F80A0A08FAF9F80A0A08FAFA080A0AF8FAFA08FA0AF8FAF
A08FAFAF90AFA08FAFAF90AFAF8FAFAF90AFAF8FAFAF90AFAF8FAFB090AFAF8FAFB090AFAF90AF\
4F90202090002090002090002F9F002090002F9F002090002F9F004F90204F9F40
4F9F40709F6F709F60B090A0B0909FC0A0AFC09FA0C0A0AFC09FAFCFA0AFC0A0A0CFA0AFC0A0A0\
CFA0AFCFA0A0CFA0AFCFA0A0D0AFAFCFA0A0D0AFAFCFAFA0000000000000000000
0000000000000000000000000000000000000000000000000040002FAF002FAF002FAF002FAF00\
2FAF002FAF000020000020000000000000000000000000000000002FAF002FAF00
2FAF002FAF002FAF002FAF002FAF002FAF002FAF002FAF002FAF002FAF0020AF002FAF002FAF00\
2FAF0020AF002FAF002FAF00002000000000000000000000605040605040000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F008F804F90804F8F7F408F7F4F807040807F407F703F\
7F703F706F306F6030000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000030303F303040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F3F2F203F302F402F203F302F3F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000050405050405F\
9F80A0A080AF9F80A0A08FAFA080A0A08FAFA08FAFAF8FAFA08FAFAF90AFAF8FAF
AF90AFAF8FAFAF90AFAF8FAFAF90AFAF8FAFB090B0AF90AFB090AF8F90606F904F6F90402F9000\
2090002F9F002090002F9F002090002F9F00209F002F9F00209F002F9F002F9F00
2F9F00209F002FA0002F9F002FA000709F407FA04FBF9FA0C09FA0C0A0AFCFA0AFCFA0AFCFA0AF\
CFA0AFD0AFAFCFA0AFD0AFAFCFA0AFD0AFAFD0AFAFD0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000002000002F002F6F002F6F00\
2F8F00308F000F0F000F0F0F0000000000000000000000000000002FB0002FAF00
2FAF002FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB000\
2FAF002FB0002FAF002F6F00000000000000000000000000605F40C0AF8F605F40
6F5F40000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000908F508F804F8F804F8F7F408F7F40807040807040\
7F703F7F703F6F6030000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000303040302F3F000000000000000000
0000000F0000000000000000000000000000000000000000000000000000000000000000000000\
2F203F50406F50406F50406F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005040509F7FA0\
9F80A09F80A0A080AF9F80A0A080AFA080A0A08FAFA080A0A08FAFA08FAFAF8FAF
A08FAFAF8FAFAF8FAFAF90AFAF8FAFAF90AF8F8F60209020208F20209000209000209000209000\
2090002090002F9F002090002F9F002090002F9F002090002F9F00209F002F9F00
209F002F9F00209F002FA000209F002FA000209F002FA000209F00A0A08F9FA080CFA0AFC0A0A0\
CFA0AFCFA0AFCFAFAFCFA0A0D0AFAFCFA0A0D0AFAFCFAFA0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000002000002000
0020000020000020000020000020002FAF002FAF002FAF002FAF002FAF002FAF002FB00020AF00\
2FAF00004000004F00002000000000000000000000000000605F40C0AF80C0AF80
BFA07F605F40605040605F4060503F6050405F503F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F008F7F4F90804F8F7F408F7F4F807F40807F407F703F\
7F703F70502F6F603F2F201F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000030303F3F3040000000000000000000
1F00003F00001F00000000000000000000000000000000000000000000000000002F203F504070\
50406F5F407050406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F40509F80AF\
9F80A0A080AF9F80A0A08FAFA080AFA08FAFA08FAFAF8FAFA08FAFAF90AFAF8FAF
AF90B0AF8FAFAF90AFAF8FAFB090B06F90402F90002090002F90002090002F90002090002F9F00\
2090002F9F002090002F9F002090002F9F00209F002F9F00209F002F9F00209F00
2FA0002F9F002FA0002F9F002FA0002FA0002FA0002FA0002FA0002FA0002FA000A0A080A0AF8F\
CFA0AFCFAFAFCFA0AFD0AFAFCFAFAFD0AFAFD0AFAFD0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000002000002F00002000002F00002000002F00\
002000002F00000000000000000000000000000000000000605F40C0AF80C0AF80
C0AF80C0AF80C0AF80BFA07FC0AF80BFA07FBFAF7FBFA070BF604F60503F90201F90201F90201F\
90201F90201F8F201090201F000000000000000000000000000000000000000000
00000000000000000000000000000000000090804F8F804F90804F8F7F408F7F40807040807F40\
7F703F7F1F106F60305F5F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000303040303040000000000000000000
3F00003F00002F0000000000000000000000000000000000000000000000201F3050406F50406F\
50407050406F50406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005040509F7FA0\
9F80A09F80A0A080A09F80A0A08FAFA080A0A08FAFA08FAFAF8FAFA08FAFAF8FAF
AF8FAFAF90AFAF8FAF40904F208F00209000208F002090002090002090002090202090206F9040\
6F904F8F906090906F90908F909F8F909080909F8F909F8F9F9F8F6F9F60709F6F
709F4070A04F209F002FA000209F002FA000209F002FA0002FA0002FA00020A0002FA0002FA000\
C0A0A0C09FA0CFAFAFCFA0AFD0AFAFCFA0A0D0AFAFD0AFAF000000000000000000
000000000000000000000000004000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006F5F4FC0AF80C0AF80
C0AF80C0AF80BFA07FC0AF80BFA07FBFAF7FBFA070BF604FB00000BF0000B00000B00000B00000\
B0201FAF1F10B06040AF3F2FB0201F603F205F5030504F305F5030504F30000000
0000000000000000000000000000000F0F008F7F4F90804F8F7F4F8F804F807F40807F407F703F\
7F4F20702F10702F1F5F402F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000303040303040000000000000000000
3F00003F00004000000000000000000000000000000000002F203F4F3F6F50407050406F504070\
50406F5F4F7050406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005040509F80AF\
9F80A0A08FAFA080A0A08FAFA080A0A08FAFA08FAFAF8FAFA08FAFAF90B0AF8FAF
AF90B0608F6F6F906F2090002F90002090002F90002090002F90008F906FB090B0B090AFB09FB0\
B090AFBF9FB0B090AFBF9FB0B090AFBF9FB0BF9FAFBF9FB0BF9FAFC0A0B0BF9FAF
C0A0AFBF9FAFBF9FAF709F407FA04F2FA0002FA0002FA0002FA0002FA0002FAF002FA0002FAF00\
7FA04080AF4FCFAFAFD0AFAFCFAFAFD0AFAFD0AFAFD0AFAF000000000000000000
000000000000000000000000004F002FAF00308F002F8F002F8F002F8F00308F00000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000605F40CFB08FC0AF80
C0AF80C0AF80C0AF80C0AF7FC0AF80AF9F70B09F70BF0000BF0000BF0000BF0000B00000BF201F\
AF9F6FB09F6FB09F6FB09F6FAF9F60B09F6FAF0000AF603FA0905FA09060802010
00000000000000000000000000000000000090805090804F90804F8F7F408F804F807F40807F40\
7F00007F0000700000600000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F3F3030401F101F000000000000000000
3F00003F00003F00000000000000000000000000002F203F4F3F6F4F406F4F3F6F504070503F6F\
50407050407050406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005040509F7FA0\
9F80A09F80A0A080AF9F80A0A08FAFA080AFA08FAFA08FAFAF8FAFA08FAFAF90AF
9F809F209000208F00209000208F002090002090002090008F908FB090AFAF90AFB090B0B090AF\
B090AFB090AFBF9FAFB090AFBF9FAFB090AFBF9FAFBF9FAFBF9FAFBF9FAFC09FAF
BF9FAFC09FAFBF9FAFC0A0AFC09FAF70A06F70A0602FA00020A0002FA0002FA0002FA0002FA000\
2FAF002FA000CFA0A0C0A0A0D0AFAFCFAFAFD0AFAFD0AFAF000000000000000000
0000000000000000000000000040002FAF002FAF002FAF002FAF002FAF002FAF002FAF002FAF00\
2F8F002F8F002F8F002F8F002F8F002F8F002F8F002F8F002F8F002F8F002F8F00
2F8F002F8F002F8F002F8F002F8F002F8F002F8F002F8F002F8F002F8F002F8F00000000000000\
0000000000000000000000000000000000000000000000006F5F40C0AF80C0AF80
C0AF80C0AF80BFA07FC0AF80BFA07FBF4030BF4030BF0000B00000BF0000B00000BF0000B05F40\
B0A070B09F6FB0A06FAF9F60B09F6FAF9F60AF9F60AF0000AF2010AF1F10AF0000
801F100000000000000000000000000000008F804F90804F8F7F4F8F804F807F40807F407F4020\
7F00007F00007040205F5F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003030403030401F1F20000000000000000000
2F00003F0000400000000000000000201F302F203F4F3F6F4F406F4F3F6F50407050406F504070\
5040705F4F7050406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504050A080AF\
9F80A0A08FAF9F80A0A08FAFA08FAFA08FAFA08FAFAF90B0AF8FAFAF90B0AF8FAF
6F904F2090002090002090002F90002090002F9000AF90AFB090B0B090AFB090B0B090AFB09FB0\
B090AFBF9FB0B090AFBF9FB0BF9FAFBF9FB0BF9FAFBF9FB0BF9FAFC09FB0BF9FAF
C0A0B0C09FAFC0A0B0C09FAFC0A0B0C0A0AFCFA0AF50A02050AF2F2FA0002FAF002FA0002FAF00\
2FA0002FAF00A0AF60AFAF6FD0AFAFD0AFAFD0AFAFDFB0AF000000000000000000
000000000000000000000000004F002FAF002FB0002FAF0030B0002FAF002FB0002FAF0030B000\
2FAF002FB0002FAF0030B0002FAF002FB0002FAF0030B0002FAF002FB0002FAF00
2FB0002FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB000\
2F8F002F90002F8F002F9000000000000000000000000000605F40CFAF8FC0AF80
CFAF8FC0AF80C0AF80BFAF7FC0AF80BFA07FC0201FBF0000BF0000BF0000BF0000B00000BFA070\
B0805FB0A070B09F6FB0A06FB09F6FB09F6FAF9F60B09F6FAF7F4FAF0000A00000
AF0000801F100000000000000000000000000F0F0F90804F90804F8F804F8F804F806030800000\
7F00007F000070100F6F603F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000301F203030401F1F20000000000000000000
0F00004000003F0000201F30201F304F3F6F403F604F3F6F4F3F6F50406F4F3F6F50407050406F\
50407050407050406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005040509F80A0\
A080AF9F80A0A080AFA080A0A08FAFA080A0A08FAFA08FAFAF8FAFA08FAFAF90AF
208F20209000208F00209000208F002090008F8F60B090AFAF90AFB090B0AF90AFB090B0B090AF\
B09FB0B090AFBF9FAFB090AFBF9FB0B090AFBF9FAFBF9FAFBF9FB0BF9FAFC09FAF
BF9FAFC0A0AFBF9FAFC0A0AFC09FAFC0A0AFC0A0AFA0A06F9FA0602FA00020A0002FA0002FA000\
2FAF002FA0007FAF6F7FAF60D0AFAFCFAFAFD0AFAFD0AFAF000000000000000000
000000000000000000000000002000004F002FAF002FAF002FAF002FB0002FAF002FAF002FAF00\
2FB0002FAF002FAF002FAF002FB0002FAF002FAF002FAF002FB0002FAF002FB000
2FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB0002FAF002FB0002FAF00\
2FB0002FAF002FB0002FAF002FB0000000000000000000006F5F4FC0AF80C0AF80
C0AF80C0AF80BFA07FC0AF80BFA07FC0AF7FBF0000BF0000BF0000BF0000B00000BF0000B09F6F\
B0A070B09F6FB09F6FAF9F6FB09F6FAF9F60AF9F60AF9F60AF9F60A05F3FAF0000
A00000A00000504F2F00000000000000000000000090804F8F804F8F804F807F40800000800000\
8000007F00007F0000606030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030202F303040000000000000000000000000
20202F3F0000400000403F60403F6F403F604F406F4F3F6F4F406F4F3F6F50407050406F504070\
5040705F4F7F50406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504050A080AF\
9F80A0A08FAFA080AFA08FAFA08FAFAF8FAFA08FAFAF8FAFA08FAFAF90B0AF8FAF
209000208F002F90002090002F902F209020B090B0B090AFB090B0B090AFB09FB0B090AFB09FB0\
B090AFBF9FB0B09FAFBF9FB0BF9FAFBF9FB0BF9FAFBF9FB0BF9FAFC0A0B0BF9FAF
C0A0B0C09FAFC0A0B0C0A0AFC0A0B0C0A0AFCFA0B0BF9FA0BF9FA02FA0002FAF002FA0002FAF00\
2FA0002FAF0050AF205FAF20D0AFAFD0B0B0D0AFAFD0B0AF000000000000000000
000000000000000000000000000000000000004F00004000004F00004000004F00004000004F00\
004000004F00004000004F002FAF0030B0002FAF002FB0002FAF0030B0002FAF00
2FB0002FAF002FB0002FB0002FB0002FAF002FB0002FB0002FB0002FAF002FB0002FB0002FB000\
2FAF002FB0002FB0002FB0002FAF000000000000000000000000006F5F4FC0AF80
C0AF80C0AF80C0AF80C0AF80C0AF80BFA07FC00000BF0000BF0000BF0000BF0000B00000BFA070\
B0A06FBFA070B09F6FB0A06FAF9F6FB0A06FAF9F60B09F6FAF9F60AF9F60AF0000
AF0000A0905FA05F3F0000000000000000000F0F0F90804F90804F8F804F8F7F4F806030800000\
7F00008000007F00006F603F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030303F3030400000001F1F201F1F203F3050
3F30503F1F203F00004030504030604F3F6F403F6F4F3F6F4F3F6F4F406F4F3F6F50407050406F\
504070504070504070000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000050405F9F80A0\
9F80AF9F80A0A08FAF9F80A0A08FAFA08FAFAF8FAFA08FAFAF8FAF808F8F8F908F
208F00209000208F00209000608F406F906FAF90AFB090B0AF90AFB090B0B090AFB09FB0B090AF\
B09FB0B090AFBF9FB0B090AFBF9FB0BF9FAFBF9FB0BF9FAFBF9FAFBF9FAFC0A0B0
BF9FAFC0A0AFC09FAFC0A0B0C09FAFC0A0AFC0A0AFCFA0AFC0A0AF2FA0002FA0002FAF002FA000\
2FAF002FA00050AF2050A020D0AFAFD0AFAFD0AFAFD0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000004F00004000004F00004000004F00
004000004F00004000004F00004000004F00004000004F00004000004F00004000004F00004000\
2F8F002F8F002FB0002FAF002FB0002FAF00000000000000000000605F40C0AF8F
C0AF80C0AF80C0AF80C0AF80BFA07FC0AF7FBF4030BF0000BF0000BF0000B00000BF0000B00000\
B0A070B09F6FB0A06FAF9F6FB09F6FAF9F60B09F6FAF9F60AF9F60AF9060AF0000
A00000A0905FA05030A0905F00000000000000000090804F8F804F8F804F8F7F408F7F40805F30\
8060307F5F2F7F5F306F60302F2F1F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030304030304000000030304F3F304F3F3050
403F5F3F305F400000403F604F3F6F403F6F4F3F6F4F3F6F5040704F406F50407050406F5F4070\
5040705F4F7F50406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000050405F504050A080AF\
9F80AFA08FAFA080AFA08FAFA08FAFAF8FAFA08FAFAF90B0AF8FAF8F908F8F8F8F
2F90002090002F90002090006F906F6F906FB090B0AF90AFB09FB0B090B0B09FB0B090AFBF9FB0\
B090AFBF9FB0B09FAFBF9FB0BF9FAFBF9FB0BF9FAFC0A0B0BF9FAFC0A0B0BF9FAF
C0A0B0C0A0AFC0A0B0C0A0AFCFA0B0C0A0AFCFA0B0C0A0AFCFAFB02FA0002FAF002FA0002FAF00\
2FAF002FAF0050AF205FAF2FD0AFAFD0AFB0D0AFAFDFB0B0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F8F002F90002FB0002FB0002FB0002FB0000000000000000000006F5F4FC0AF80
CFB08FC0AF80C0AF80C0AF80C0AF80BFAF7FC06F4FBF0000C00000BF0000BF0000B00000BF0000\
B0805FBFA070B09F6FB0A070B09F6FB0A06FAF9F60B09F6FAF9F60AF9F60AF5F3F
AF0000A05F3FA0905F9F5030504F30504F2F0F0F0F8F804F908F508F804F8F804F807F408F7F40\
80703F80703F7F703F706F3F5F5030000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F3F3030401F1F203F304F30304F3F3050
3F3050403F5F3F0000400000403F604F3F6F403F604F3F6F4F3F6F5040704F3F6F504070504070\
5F4070504070504070000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F405050405F9F80A0\
A080AF9F80A0A08FAFA080AFA08FAFA08FAFAF8FAFA08FAFAF90B0808F8F8F908F
208F00209000208F002090004090404F904FAF90AFB090B0AF90AFB090B0B090AFB090B0B090AF\
BF9FB0B090AFBF9FB0B09FAFBF9FB0BF9FAFBF9FB0BF9FAFC09FB0BF9FAFC0A0B0
BF9FAFC0A0B0C09FAFC0A0B0C0A0AFCFA0B0C0A0AFCFA0AF9FA08F2FAF002FA0002FAF002FA000\
2FAF002FA00050AF2050AF20D0AFAFD0AFAFD0AFAFD0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F90002F8F002FB0002FB0002FB00020AF00000000000000000000605F40C0AF80
C0AF80C0AF80C0AF7FC0AF80BFAF7FC0AF80B09F70BF0000BF0000BF0000B00000BF0000B00000\
BF6040B09F6FB0A06FB09F6FB0A06FAF9F60B09F6FAF9F60AF9F60AF905FAF5F3F
A00000A02010A0905FA000009F8F509F905090804F4F402F403F204F40208F7F408F7F40807F40\
807F407F703F7F703F6F00005F5F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003030403030403F304F30304F3F3F503F3050
403F5F3F305F4000004000004F2040403F6F4F406F4F3F6F5040704F3F6F5040705040705F4070\
5040705F4F7F50406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F5F504050A080AF\
9F80AFA08FAFA08FAFA08FAFA08FAFAF8FB0A08FAFAF90B0AF8FAF8F90908F908F
2090002090002F90002090002F9000209000AF90AFB090B0B09FB0B090B0B09FB0B090B0BF9FB0\
B090AFBF9FB0BF9FB0BF9FB0BF9FAFBFA0B0BF9FB0C0A0B0BF9FAFC0A0B0C09FAF
C0A0B0C0A0AFC0A0B0C0A0AFCFA0B0C0A0AFCFAFB0CFA0AF7FAF6F2FA0002FAF002FAF002FAF00\
2FAF002FAF007FAF6F80AF6FD0AFAFDFB0B0D0AFAFDFB0B0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F8F002F90002FB0002FB0002FB0002FB0000000000000000000006F5F4FC0AF80
CFB08FC0AF80C0B080C0AF80C0AF80BFAF7FC0AF80BFAF7FBF6F4FBF0000BF0000BF0000BF0000\
B06040BFA070B0A06FB0A06FB09F6FB0A06FAF9F60B09F6FAF9F60B09F60AF7F4F
AF0000A00000A09F5FA03F20908F509F8F509F8F50000000000000000000000000403F20403F20\
3F3F1F807F407F703F7000005F5F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000030303F30303F1F1F20302F403F304F30304F3F3050
3F3050403F5F40000040203F4000004F20404F30504F3F6F4F3F6F5040704F3F6F504070504070\
5F4070504070504070000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F405050405F9F80A0\
A080AF9F80AFA08FAFA080AFA08FAFA08FAFAF8FB0A08FAFAF90B0AF8FAFAF90B0
208F002090002090002090002090002090006F906FB090B0AF90AFB090B0B090AFB09FB0B090AF\
BF9FB0B090AFBF9FB0B09FAFBF9FB0BF9FAFBF9FB0BF9FAFC09FB0BF9FAFC0A0B0
BF9FAFC0A0B0C0A0AFC0A0B0C0A0AFCFA0B0BF9FA0C0A0AF20A0002FAF002FA0002FAF002FA000\
2FAF002FAF00AFAF6FA0AF60D0AFAFD0AFAFD0B0B0D0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F90002F8F002FB0002FAF002FB0002FAF00000000000000000000605F40CFB08F
C0AF80C0AF80C0AF80C0AF80BFAF7FC0AF7FBFA07FC0AF7FB09F70BF201FBF0000BF0000B00000\
BF201FB0A06FB0A070B09F6FB0A06FAF9F60B09F6FAF9F60AF9F60AF9F60AF9F60
A00000A00000A0905FA0905F9F3F209F9050908F500F0F00000000000000000000000000000000\
0000000000003F3F1F706F30600000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000030303F1F1F201F1F203F304F30304F3F30503F3050
403F5F40305F400000403F604F00004F00004F101F4F1F2F502F4050406F5040705040705F4F70\
50407050407050406F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000050405F504050A08FAF\
A080AFA08FAFA080AFAF8FB0A08FAFAF8FB0A08FAFAF90B0AF8FB0AF90B0AF90B0
4F902F2090002F90002090002F90002090002F90008F906F909F6FB090B0B09FB0B090B0BF9FB0\
B09FB0BF9FB0BF9FB0BF9FB0BF9FB0BF9FB0BF9FAFC0A0B0BF9FB0C0A0B0C09FAF
C0A0B0C0A0B0C0A0B0C0A0AFC0A0AFC09FAF50AF2F50A0202FAF002FA0002FAF002FAF002FAF00\
2FAF002FAF00D0AFAFD0B0B0D0AFAFD0B0B0D0AFAFDFB0B0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F90002F90002FB0002FB0002FB0002FB0000000000000000000006F5F4FC0AF80
CFB08FC0AF80C0B080C0AF80C0AF80C0AF7FC0AF80BFA07FC0AF7FBF8060BF0000BF0000BF0000\
B00000AF9F6FAF9060B0A070B09F6FB0A06FAF9F60B0A06FAF9F60AF9F60AF9F60
AF0000A00000A00000A0905FA05F3F9F8F509F8F50000000000000000000000000000000000000\
0000000000000000003F3F20600000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003030400000001F1F203030403F304F30304F3F3050
3F3050403F5F3F00004030504000004F00004000004F00004F00005040704F3F6F504070504070\
5F4F7050406F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F405050405F9F80A0\
A08FAF9F80AFA08FAFA08FAFA08FAFA08FAFAF8FB0A08FAFAF90B0AF8FAFAF90B0
60906F2090002090002090002090002090002090002090006F90406F9F4FA08FA0AF8FA0B090AF\
BF9FB0B090AFBF9FB0BF9FAFBF9FB0BF9FAFBF9FB0BF9FAFC0A0B0BF9FAFC0A0B0
C09FAFC0A0B09F9F8F9FA08F4FA02050A02020A0002FA0002FA0002FAF002FA0002FAF002FAF00\
7FAF6F7FAF60D0AFB0D0AFAFD0AFB0D0AFAFD0B0AFD0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F8F002FB0002FB0002FB0002FAF00000000000000000000605F40CFB08F
C0AF80C0B080C0AF80C0AF80C0AF7FC0AF80BFA07FBFAF7FBFA07FBFAF7FBF6040BF604FB00000\
BF0000B03F2FB0402FB09F6FB0A06FAF9F60B09F6FAF9F60B09F60AF9F60AF9F60
AF0000AF0000A00000A0905F9F90509F905F908F5090804F000000000000000000000000000000\
0000000000000000000000006F00000F0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003030403F304F30304F3F3F503F3050
403F5F400F10400000403F604F00004F00004F00004F00005000004F20405040705040705F4F70\
5040705F4F70000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F5F50405FA08FAF\
9F80AFA08FAFA08FAFA08FB0A08FAFAF90B0AF8FAFAF90B0AF8FB0B090B0AF90B0
B090B08F906F9090702090002F90002090002F9F002090002F9F002090002F9F00209F004F9F2F\
4F9F20709F4F709F4F709F6F709F6F70A06F709F6F70A06F709F6F50A04F4FA04F
2FA02F2FA0202FA0002FA0002FA0002FA0002FAF002FA0002FAF002FA0002FAF002FAF0050AF2F\
C0A0AFCFAFAFD0AFAFD0B0B0D0AFAFDFB0B0D0AFAFDFB0B0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F90002FB0002FB0002FB0002FB0000000000000000000006F604FCFB08F
CFB08FC0AF80CFB08FC0AF80C0AF80C0AF7FC0AF80BFAF7FC0AF7FBFA07FBFAF7FBFA070BF604F\
B00000BF0000B00000B09F6FB09F6FB0A06FB09F6FB0A06FAF9F60B09F6FAF9F60
AF9F60A00000AF201FA0905FA0905F9F8F509F905F90804F000000000000000000000000000000\
0000000000000000000000006F0000200000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003030403030403F304F3F304F3F3F50
3F3050400000400000403F6F40203F4F00004F00004F00004F000050000050204050406F504070\
5F4F7050406F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504050504F5F9F80A0\
A08FAFA080AFA08FAFA080AFA08FAFA08FAFAF90B0A08FAFAF90B0AF8FB0AF90B0
AF90AFAF90B0AF90B06F904F2090002090002090002F9F00209000209F002090002F9F00209000\
2F9F00209F002F9F00209F002F9F00209F002FA000209F002FA000209F002FA000
209F002FA00020A0002FA00020A0002FA0002FA0002FAF002FA0002FAF002FA0202FAF20BF9FA0\
D0AFB0CFAFAFD0AFB0D0AFAFD0AFB0D0AFAFDFB0B0D0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F8F002F90002FB0002FB0002FB000000000000000000000605F40CFB08F
C0AF80C0B080C0AF80C0AF80BFAF7FC0AF80BFAF7FC0AF7FBFA070BFAF7FBFA070BFAF7FB0A070\
BF805FB0805FBF201FB01F10B0A06FB09F6FB0A06FAF9F60AF9F60AF9F60AF9F60
9F8F5FAF0000A01F10A000009F90509F905F9F8F50908F50000000000000000000000000000000\
0000000000000000000000006F60302F201F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003030403F304F3F304F3F3F503F3050
403F5F400000400000403F604F3F6F4F20404F00004F000050000050000050101F500F1F5F4F7F\
5F4F705F4F70000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F5F504050A08FAF\
A080AFA08FAFA08FAFAF8FB0A08FAFAF90B0AF8FAFAF90B0AF90B0B090B0AF90B0
B090B0AF90B0B090B0AF90B0909F702090002F9F002090002F9F002090002F9F00209F002F9F00\
209F002F9F00209F002FA0002F9F002FA000209F002FA0002FA0002FA0002FA000
2FA0002FA0002FA0002FA0002FAF002FA0002FAF002FA00050AF2F50AF20CFA0AFC0A0AFD0AFB0\
D0AFB0D0AFB0D0AFAFD0B0B0D0AFB0DFB0B0D0AFAFDFB0B0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F90002F8F002FBF002FB0002FB0000000000000000000006F5F4FC0B08F
CFB08FC0AF80CFB080C0AF80C0B080C0AF80C0AF80BFAF7FC0AF7FBFAF7FBFAF7FBFA070BFAF7F\
BFA070BFA070B0A06FBF6040B0201FB0201FA09060B0A06FAF9F60B09F60A0905F
AF603FAF0000AF201FA00000A07F4F9F90509F905090804F0F0F0F000000000000000000000000\
0000000000000000000000006F6F305F5F3F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000030304030304F3F30503F30503F305F
3F305F400000400000403F6F403F6F4F3F6F4F0F104F00004F0000501F3050406F504070504070\
5F4F7F50406F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F405050405F9F80AF\
A08FAF9F80AFA08FAFA08FAFAF8FB0A08FAFAF90B0AF8FAFAF90B0AF8FB0AF90B0
AF90B0B090B0AF90B0B090B0AF90B0A08FAFA08FA06F904F2090002F9F00209000209F00209F00\
2F9F00209F002F9F00209F002F9F00209F002FA000209F002FA000209F002FA000
20A0002FA00020A0002FA0002FA0202FAF209FA06FA0AF6FCFA0AFCFAFB0CFA0AFD0AFB0CFAFAF\
D0AFB0CFAFAFD0AFB0D0AFAFD0B0B0D0AFAFDFB0B0D0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F8F002F90002FB0002FB0002FB000000000000000000000605F40CFB08F
C0AF80CFB08FC0AF80C0AF80C0AF7FC0AF80BFAF7FC0AF7FBFA07FBFAF7FBFA070BFAF70BFA070\
BFA070B0A06FB0A070B0A06FB0A06FB09F6FB00000AF2010B0201FAF1F10AF0000
AF0000AF0000A03F20A07F4F9F1F10A0905F9F8F50908F50000000000000000000000000000000\
0000000000000000000000000000005F5030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003030403F304F30304F3F3F503F305F
4010104000004F3050403F6F4F3F6F4F3F6F502F404F2040502F405040705040705040705F4F7F\
5F40705F4F70000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F5F50405FA08FAF\
A080AFA08FB0A08FAFAF8FB0A08FAFAF90B0AF8FB0AF90B0AF90B0B090B0AF90B0
B090B0AF90B0B09FB0B090B0B09FB0B090B0B09FBFB090B0B09FB0B090B0BF9FBFB09FB0BF9FB0\
BF9FB0BFA0B0BF9FB0BFA0B0BF9FB0C0A0B0BF9FB0C0A0B0C0A0B0C0A0B0C0A0B0
C0A0B0C0A0B0CFAFB0C0A0B0CFAFB0CFA0B0CFAFB0CFAFB0CFAFB0CFAFB0D0AFB0CFAFB0D0AFB0\
D0AFB0D0B0B0D0AFB0D0B0B0D0AFB0DFB0B0D0B0B0DFB0B0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002FBF000000000000000000006F604FCFB08F
CFB08FC0AF80CFB08FC0AF80C0B080C0AF7FC0B080C0AF7FC0AF7FBFAF7FC0AF7FBFA070BFAF7F\
BFA070BFAF70B0A070BFA070B0A06FB0A06FB0A06FB06040AF5F3FB06040AF5F3F
AF603FAF0000AF0000A0905FA000009F9050A0905F90804F0F0F0F000000000000000000000000\
000000000000000000000000000000605F3F4F0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F1F2030304F30304F3F1F203F3050403F5F
3F0000400000403F604F3F6F403F6F4F3F6F4F3F6F5040704F3F6F5040705040705F4070504070\
5F4F7F504070000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504050504F5F9F80AF\
A08FAFA080AFA08FB0A08FAFAF8FB0A08FAFAF90B09F80A0A08FAF9F80A0A08FAF
A080A0A08FAFA08FA0A08FAFA08FA0A08FAFA08FA0AF8FAFA08FA0AF8FAFAF8FA0AF90AFAF8FA0\
AF90AFAF90A0B090AFAF90A0B090AFAF90A0B090AFB090A0B090AFB090A0BF9FAF
B090A0BF9FA0B090A0BF9FAFBF9FA0BF9FA0BF9FA0BF9FAFBF9FA0C0A0A0BF9FA0C0A0AFBF9FA0\
C0A0A0C09FA0C0A0A0C0A0A0000000000000000000000000100F0F0F0F0F10100F
0F0F0F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000605F40CFB08F
C0AF80CFB08FC0AF80C0B080C0AF7FC0AF80BFAF7FC0AF7FBFA07FBFAF7FBFA070BFAF7FB0A070\
BFA070B0A070BFA070B0A06FB0A06FB09F6FB0A06FAF9F60B09F6FAF9F60AF9F60
AF9F5FAF7F4FA01F10A09F5F9F0000A0905F9F8F509F8F50000000000000000000000000000000\
0000000000000000000000000000002F2F1F200000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000201F2F3030403F20303F0F103F10103F0F10
4000004000004F3F6F403F6F4F40704F3F6F4F40704F3F6F5040705040705040705040705F4F7F\
5F4F7F5F4F70000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000100F100F0F0F0F0F100F0F0F100F100F0F0F504F5F50405F9F80A0\
907FA09F80A09F80A09F80A09F80A0A08FAF9F80A0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFB0B0DFB0B0DFB0AF
DFB0B0100F0F1010100F0F0F101010000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006F604FCFB08F
CFB08FC0B080CFB08FC0AF80CFB080C0AF80C0AF80BFAF7FC0AF80BFAF7FC0AF7FBFA070BFAF7F\
BFA070BFAF70B0A070BFA070B0A06FB0A06FB09F6FB0A06FB09F60B0A06FAF9F60
B0A060AF9F60AF9F60A0905FA0201FA00000A0905F9F8F500F0F0F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F1F2030304F30304F3F00003F0000400000
3F0000401F2F4030604F3F6F403F6F4F3F6F4F3F6F5040704F3F6F504070504070504070504070\
5F4F7F504070000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F000F0F0F0F0F000F0F0F0F0F000F0F0F0F0F0F0F0F0F0F0F000F0F0F0F
0F0F0F0F0F0F0F000F907FA0907FA09F80AF907FA09F80AF9F80AF9F80AF4F3F504F4050000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFB0B0D0B0AFDFB0B0
DFB0AFDFB0B0D0B0AFDFB0B0DFB0AF10100F0F0F0F100F0F0F0F0F10100F0F0F0F100F0F0F0F0F\
10100F0F0F0F100F0F0F0F0F10100F0F0F0F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000605F406F5F4F
C0B080CFB08FC0AF80C0B080C0AF80C0B080BFAF7FC0AF7FBFAF7FC0AF7FBFA070BFAF7FBFA070\
BFAF70B0A06FBFA070B0A06FB0A070B09F6FB0A06FAF9F60B0A06FAF9F60AF9F60
AF9F60AF9F60A0905FA09F5FA0905FA000009F8F509F8F5090804F000000000000000000000000\
000000000000000000000000000000000000504F30000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F304F30304F3F30503F304F4000003F0000
40101F403F604F3F6F403F6F4F40704F3F6F5040704F40705040705040705F4F7F5040705F4F7F\
5F4F7F5F4F70000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F0F0F0F\
0F0F0F8F70A0806F9F8F70A08F709F8F70A08F70A0907FA08F70A0907FA0907FA0
907FA0907FA09F80AF907FA09F80AF907FA09F80AF9F80AF9F80AF9F80AF50405F4F4050000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFB0B0DFB0B0DFB0B0
DFB0B0DFB0AFDFB0B0DFB0AFDFB0B0DFB0AFDFB0B0DFB0AFDFB0AFDFB0AFDFBFAFDFB0AFDFB0AF\
D0B0A0DFBFAFDFB0A0DFB0AFD0B0A0DFBFAFD0B0A0DFBFA00F0F0F101010100F0F
10100F0F0F0F101010100F0F10100F0F0F0F101010100F0F10100F0F0F0F1010100F0F0F10100F\
0F0F0F1010100F0F0F10100F0000000000000000000000000000000000006F5F4F
CFB08FC0B080CFB08FC0B080CFB080C0AF80C0B080C0AF7FC0AF7FBFAF7FC0AF7FBFAF7FBFAF7F\
BFA070BFAF70BFA070BFA070B0A06FBFA070B0A06FB0A06FB09F60B0A06FAF9F60
B0A060AF9F60AF9F60A09F5FA09F5FA00000A000009F8F50908F50000000000000000000000000\
000000000000000000000000000000000000505030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003030403F304F30304F3F30503F2F40403F5F
3F305F403F60403F604F3F6F403F6F4F3F6F4F3F6F5040704F3F6F5040705040705F4F7F504070\
5F4F7F504070000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F0F000F806F9F806F9F\
806F9F806F9F8F709F806F9F8F70A08F709F8F70A08F709F8F7FA08F70A0907FA0
8F70A0907FA0907FA09F80AF907FA09F80AF9F7FA09F80AF9F80AF9F80AF4F4050504050000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFB0B0DFB0B0DFB0B0
DFB0AFDFB0B0DFB0AFDFB0B0DFB0AFDFB0B0DFB0AFDFB0AFD0B0AFDFB0AFDFB0AFDFB0AFD0B0A0\
DFB0AFD0B0A0DFB0AFD0B0A0DFB0A0D0B0A0DFB0A0D0B09FD0B0A0D0B09FD0B0A0
D0B09FD0B09FD0B09FD0BF9FD0B09FD0B09FD0B09FD0BF9FCFB090D0B09FD0B090D0BF90CFB090\
D0B090CFB090D0BF90CFB08F100F0F0F0F0F0000000000000000000000006F604F
C0B080CFB08FC0AF80CFB080C0AF80C0AF80C0AF7FC0AF80BFAF7FC0AF7FBFAF70BFAF7FBFA070\
BFAF70B0A070BFA070B0A06FB0A06FB0A06FB0A06FAF9F60B0A06FAF9F60B0A060
AF9F60AF9F60A09F5FAF9F5FA0905F9F90509F70409F5F3090804F000000000000000000000000\
000000000000000000000000000000000000505030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F304F30304F3F3F503F3050403F5F3F305F
403F60403F604F3F6F403F6F4F40704F3F6F5040704F40705040705040705F4F7F5040705F4F7F\
5040705F4F70000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F0F0F0F0F0F0F806F9F806F9F806F9F806F9F8070A0\
806F9F8F70A0806F9F8F70A08F709F8F70A08F70A0907FA08F70A0907FA0907FA0
907FAF907FA09F80AF907FA09F80AF9F80AF9F80AF9F80AFA08FAF9F80AF5040504F4050000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFB0B0DFB0B0DFB0B0
DFBFB0DFB0B0DFB0B0DFB0AFDFBFB0DFB0AFDFB0B0DFB0AFDFBFB0DFB0AFDFBFAFDFB0AFDFBFAF\
DFB0AFDFBFAFD0B0A0DFBFAFDFB0A0DFBFA0D0B0A0DFBFA0D0B0A0DFBFA0D0B09F
DFBFA0D0B09FDFBFA0D0B09FDFBFA0D0B09FD0BF9FD0B09FD0BF9FD0B090D0BF9FD0B090D0BF9F\
D0B090D0BF90CFB090D0BF90D0B090D0BF900F0F0F10100F0000000000006F5F4F
CFB08FCFB08FCFB08FC0B080CFB080C0AF80C0B080C0AF7FC0B080BFAF7FC0AF7FBFAF70C0AF7F\
BFA070BFAF70B0A070BFAF70B0A06FBFA070B0A06FB0A06FB0A06FB0A06FAF9F60
B0A060AF9F60AF9F60A09F5FAF9F60A0905FA0905F9F8F50908F50000000000000000000000000\
0000000000000000000000000000000000002F2F1F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003030403F304F30304F3F3F503F3050403F5F
40305F403F60403F604F3F6F4F3F6F4F40704F3F6F5040705040705040705040705F4F7F5F4070\
504F7050407000000000000000000000000000000000000000000000000F0F0F0F
00000F0F0F0F00000F0F0F0F0F000F7F6F9F7F60907F6F9F7F6F9F806F9F7F6F9F806F9F806F9F\
80709F806F9F8F709F806F9F8F70A08F709F8F70A08F70A0907FA08F70A0907FA0
907FA0907FAF907FA09F7FAF907FA09F80AF907FA09F80AF9F80AFA080AF4F3F50504050000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFB0B0DFB0AFDFB0B0
DFB0B0DFB0B0DFB0AFDFB0B0DFB0AFDFB0B0DFB0AFDFB0AFDFB0AFDFB0AFD0B0AFDFB0AFD0B0AF\
DFB0AFD0B0A0DFB0AFD0B0A0DFB0A0D0B0A0DFB0A0D0B0A0DFBFA0D0B09FD0B0A0
D0B09FD0BFA0D0B09FD0B09FD0B09FD0BF9FD0B090D0BF9FD0B090D0BF9FCFB090D0BF90D0B090\
D0BF90CFB090D0BF90CFB090D0BF90CFB08FCFB090CFB08F10100F0F0F0F6F5F4F
CFB08FCFB08FC0AF80CFB080C0AF80C0B080C0AF7FC0AF80BFAF7FC0AF7FBFAF70BFAF7FBFA070\
BFAF70B0A070BFA070B0A06FBFA070B0A06FB0A06FB09F6FB0A06FAF9F60B0A060
AF9F60AF9F60A09F5FAF9F5FA0905FA09F5F9F90509F905090804F0F0F00000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F304F3F304F3F3F503F3050403F5F403F5F
403F60403F604F3F6F4F3F6F4F40704F3F6F5040705040705040705040705F4F7F5F407F5F4F7F\
5040700000000F000F0F0F0F0F000F0F0F0F0F0F0F0F0F0F6F5F8F705F906F5F8F
706090705F907F60907060907F6F9F7F609F806F9F7F6F9F806F9F7F6F9F8070A0806F9F8070A0\
806F9F8F70A080709F8F70A08F709F8F7FA08F70A0907FA08F70A0907FAF907FA0
907FAF907FA09F80AF907FAF9F80AF9F80AF9F80AF9F80AFA08FAF9F80AF50405F4F4050000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFB0B0E0BFB0DFB0B0
DFBFB0DFB0B0E0BFB0DFB0B0DFBFB0DFB0AFE0BFB0DFB0AFDFBFB0DFB0AFDFBFB0DFB0AFDFBFAF\
DFB0AFDFBFAFDFB0A0DFBFAFD0B0A0DFBFAFD0B0A0DFBFA0D0B0A0DFBFA0D0B0A0
DFBFA0D0B09FDFBFA0D0B09FDFBF9FD0B09FDFBF9FD0B09FD0BF9FD0B090D0BF9FD0B090D0BF9F\
D0B090D0BF9FD0BF90D0BF90CFB090D0BF90CFB08FD0BF90CFB08FD0BF906F5F4F
BFAF80BFA07FBFAF80C0B080CFB080C0AF80C0B080C0AF7FC0B080BFAF7FC0AF7FBFAF7FBFAF7F\
BFA070BFAF7FBFA070BFAF70B0A06FBFAF70B0A06FB0A06FB0A06FB0A06FAF9F60
B0A060AF9F60AF9F60AF9F5FAF9F5FA0905FA0905F9F9050908F50000000000000000000000000\
0000000000000000000000000000000000000000004F4F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003030403F304F30304F3F30503F3050403F60
40305F403F60403F6F4F3F6F4F3F6F4F40704F3F6F5040704F3F6F50406F4F406F50407050406F\
50406F5F4F7F6050805F4F7F6050806050806F5F8F6050806F5F8F6F5F8F705F90
6F5F8F7060907060907F60907060907F6F9F7F609F806F9F7F609F806F9F7F6F9F806F9F806F9F\
806F9F806F9F8F70A0806F9F8F70A08F70A08F7FA08F70A0907FA08F70A0907FA0
907FA0907FAF907FA09F80AF907FA09F80AF9F80AF9F80AF9F80AFA080AF4F405050405F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFB0B0DFB0B0DFB0B0
DFB0AFDFB0B0DFB0B0DFB0B0DFB0AFDFB0B0DFB0AFDFB0B0DFB0AFDFB0AFDFB0AFDFBFAFD0B0A0\
DFB0AFD0B0A0DFBFAFD0B0A0DFB0A0D0B0A0DFBFA0D0B0A0DFBFA0D0B0A0DFBFA0
D0B09FD0BFA0D0B09FD0BF9FD0B09FD0BF9FD0B09FD0BF9FD0B090D0BF9FD0B090D0BF9FCFB090\
D0BF90D0B090D0BF90CFB08FD0BF90CFB08FD0BF90CFB08FCFB08FCFB08F605F40
000000000000000000000000000000BFA07FB0A070BFA07FB09F70B0A070B09F70B0A070AF9F6F\
BFAF70BFA070BFAF70B0A06FBFA070B0A06FB0A06FB09F60B0A06FAF9F60B0A06F
AF9F60AF9F60AF9F5FAF9F60A0905FA09F5F9F90509F9050908F500F0F00000000000000000000\
000000000000000000000000000000000000000000404030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000001F1F203F304F30304F3F3F503F3050403F5F40305F
403F60403F604F3F6F4F3F6F4F40704F3F705040704F3F6F000000000000000000000000000000\
0F000F6050805F4F7F6050806050806F508F6050806F5F8F6F5F8F705F906F5F8F
7060907060907F60907060907F6F9F7F6F9F806F9F7F6F9F806F9F806F9F806F9F806F9F8F70A0\
806F9F8F70A080709F8F70A08F70A08F7FA08F70A0907FA0907FA0907FAF907FA0
9F80AF907FA09F80AF907FAF9F80AF9F80AF9F80AF9F80AFA08FB09F80AF50405F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFBFB0DFB0B0
E0BFB0DFB0B0E0BFB0DFB0B0E0BFB0DFB0B0DFBFB0DFB0AFE0BFB0DFB0AFDFBFAFDFB0AFDFBFAF\
DFB0AFDFBFAFDFB0A0DFBFAFDFB0A0DFBFAFD0B0A0DFBFAFD0B0A0DFBFA0D0B0A0
DFBFA0D0B09FDFBFA0D0B09FDFBFA0D0BF9FD0BF9FD0B09FDFBF9FD0BF9FD0BF9FD0B090D0BF9F\
D0BF90D0BF90D0B090D0BF90D0BF90D0BF90CFB08FD0BF90CFB08FD0BF90605F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000B0A06FAF9F6FB09F6FAF9F6FAF9F6FAF9F60AF9F6FA09060B0A06FAF9F60
B0A06FAF9F60AF9F60AF9F5FAF9F60A0905FA0905F9F90509F9050000000000000000000000000\
000000000000000000000000000000000000000000504F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000001F1F203030403F304F3F30503F3F5F3F305F403F60
403F60403F6F403F6F4F3F6F4F3F6F4F4070403F60000000000000000000000000000000000000\
0F0F0F5F4F7F605080604F806050806050806F508F6F508F6F5F8F6F5F8F705F90
705F8F7060907060907F60907F60907F6F9F7F609F806F9F7F6F9F806F9F7F6F9F806F9F806F9F\
8F70A0806F9F8F70A08F709F8F70A08F709F8F70A08F70A0907FA08F70A0907FA0
907FA0907FAF907FA09F80AF907FAF9F80AF9F80AF9F80AF9F80AFA08FB04F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFB0B0DFB0B0
DFB0B0DFBFB0DFB0AFDFB0B0DFB0B0DFBFB0DFB0AFDFB0B0DFB0AFDFBFB0DFB0AFDFB0AFDFB0AF\
DFBFAFD0B0A0DFBFAFD0B0A0DFBFAFD0B0A0DFBFA0D0B0A0DFBFA0D0B09FDFBFA0
D0B09FDFBFA0D0B09FD0BF9FD0B09FD0BF9FD0B09FD0BF9FD0B09FD0BF9FD0B090D0BF9FD0B090\
D0BF90CFB090D0BF90CFB090D0BF90CFB08FD0BF90CFB08FD0BF90CFB08F605F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000A09060
A0905FA0905FAF9F5FAF9F60A0905FA09F5F9F90509F90509F8F500F0F00000000000000000000\
0000000000000000000000000000000000000000004F4F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000001F1F203F3F4F3F304F3F3F503F3050403F5F403F60
403F60403F604F40704F3F6F4F3F6F000000000000000000000000000000000000000000000000\
00000F6050806050806050806050806F5F8F6F508F6F5F8F6F5F8F706090705F90
7060907060907F609F7F60907F6F9F7F609F806F9F7F6F9F806F9F806F9F8070A0806F9F8F70A0\
806F9F8F70A08F70A08F70A08F70A0907FA08F70A0907FA08F7FA0907FAF907FA0
907FAF907FA09F80AF9F80AF9F80AF9F80AFA080B09F80AFA08FB09F80AF50405F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFDFB0B0
E0BFB0DFB0B0E0BFB0DFB0B0E0BFB0DFB0B0E0BFB0DFB0B0DFBFB0DFB0AFE0BFB0DFB0AFDFBFAF\
DFB0AFDFBFAFDFB0AFDFBFAFDFB0A0DFBFAFDFB0A0DFBFA0D0B0A0DFBFA0D0BFA0
DFBFA0D0B09FDFBFA0D0BF9FDFBFA0D0B09FDFBFA0D0BF9FD0BF9FD0BF9FDFBF9FD0BF90D0BF9F\
D0BF90D0BF9FD0BF90D0BF90D0BF90D0BF90D0BF90D0BF90CFB08FD0BF90605F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000A0905FAF9F60AF9F60A0905FA09F5F9F90509F905090804F000000000000000000\
00000000000000000000000000000000000000000050503F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000030304F30304F3F30503F30503F3F5F3F305F403F60
403060403F6F403F6F403F6F000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F7F605080604F806050806050806F5F8F6F508F6F5F8F6F5F8F706090
705F8F7060907060907F609F7F60907F6F9F7F6F9F806F9F7F6F9F806F9F806F9F8070A0806F9F\
8070A0806F9F8F70A080709F8F70A08F70A0907FA08F70A0907FA0907FA0907FAF
907FA0907FAF907FA09F80AF907FAF9F80AF9F80AF9F80AF9F80AFA08FB04F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFB0B0E0BFB0
DFB0B0DFB0B0DFB0B0E0BFB0DFB0AFDFBFB0DFB0AFDFBFB0DFB0AFDFBFAFDFB0AFDFBFAFDFB0AF\
DFBFAFDFB0AFDFBFAFD0B0A0DFBFAFD0B0A0DFBFAFD0B0A0DFBFA0D0B0A0DFBFA0
D0B09FDFBFA0D0B09FDFBFA0D0B09FD0BF9FD0B09FD0BF9FD0B090D0BF9FD0B090D0BF9FD0B090\
D0BF90D0B090D0BF90CFB090D0BF90CFB090D0BF90CFB08FD0BF90CFB08F605F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000A0905FA0905FA09F5FA09050A0905F9F8F50908F4F000000000000000000\
000000000000000000000000000000000000000000504F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003030403F304F3F304F403F5F3F305F403F60403F60
403F60403060000000000000000000000000000000000000000000000000000000000000000000\
0F000F605080604F806F508F6050806F5F8F6F508F6F5F8F6F5F8F706090705F90
7060907060907F609F7F60907F6F9F7F6F9F806F9F7F6F9F806FA0806F9F8070A0806F9F8F70A0\
8070A08F70A08F70A08F7FA08F70A0907FA08F70A0907FAF907FA0907FAF907FA0
9F80AF907FAF9F80AF9F80AF9F80AF9F80AF9F80B09F80AFA08FB0A08FAF50405F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFDFB0B0
E0BFBFDFB0B0E0BFB0DFB0B0E0BFB0DFB0B0E0BFB0DFB0AFE0BFB0DFB0AFDFBFB0DFB0AFE0BFB0\
DFBFAFDFBFAFDFB0AFDFBFAFDFBFAFDFBFAFDFB0A0DFBFAFDFBFA0DFBFA0D0BFA0
DFBFA0D0BFA0DFBFA0D0BF9FDFBFA0D0BF9FDFBF9FD0BF9FDFBF9FD0BF9FD0BF9FD0BF90DFBF9F\
D0BF90D0BF9FD0BF90D0BF9FD0BF90D0BF90D0BF90D0BF90CFBF8FD0BF90605F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000A0905FA09F5FA09F5F9F9050A0905F908F4F0F0F0F000000000000\
000000000000000000000000000000000000000000303020000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F304F30304F3F30503F30503F30503F305040305F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F7F6050806050806F508F6050806F5F8F6F5F8F705F8F6F5F8F706090
705F907F60907060907F609F7F609F7F6F9F7F609F806F9F7F6F9F806FA0806F9F806FA0806F9F\
8F70A0806F9F8F70A08F70A08F70A08F70A0907FA08F70A0907FAF8F7FA0907FAF
907FA09F80AF907FA09F80AF9F80AF9F80AF9F80AF9F80AF9F80AFA08FB04F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFB0B0DFBFB0
DFB0B0E0BFB0DFB0B0DFBFB0DFB0B0E0BFB0DFB0AFDFBFB0DFB0AFDFBFB0DFB0AFDFBFAFDFB0AF\
DFBFAFDFB0AFDFBFAFDFB0A0DFBFAFD0B0A0DFBFA0D0B0A0DFBFA0D0B0A0DFBFA0
D0B0A0DFBFA0D0B09FDFBFA0D0B09FDFBFA0D0B09FD0BF9FD0BF9FD0BF9FD0B090D0BF9FD0BF90\
D0BF9FD0B090D0BF90D0BF90D0BF90CFB08FD0BF90CFB08FD0BF90CFB08F605F40
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000009F905FA09050A0905F9F8F50908F50000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000030304F3F304F30304F3F3050000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000F60508060508060508F6050806F5F8F6F5F8F6F5F8F6F5F8F706090705F90
7060907060907F6F9F7F609F7F6F9F7F6F9F806FA0806F9F806FA0806F9F8070A0806F9F8F70A0\
8070A08F70A08F70A08F70A08F70A0907FAF8F7FA0907FAF907FA0907FAF907FAF
9F80AF907FAF9F80AF9F80AF9F80AF9F80AFA08FB09F80AFA08FB0A080AF50405F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFDFB0B0
E0BFBFDFB0B0E0BFBFDFBFB0E0BFB0DFB0B0E0BFB0DFBFB0E0BFB0DFB0AFE0BFB0DFBFAFDFBFB0\
DFB0AFDFBFAFDFBFAFDFBFAFDFBFA0DFBFAFDFBFA0DFBFAFDFBFA0DFBFAFDFBFA0
DFBFA0D0BFA0DFBFA0D0BFA0DFBFA0D0BF9FDFBFA0D0BF9FDFBF9FD0BF9FDFBF9FD0BF9FD0BF9F\
D0BF90DFBF9FD0BF90D0BF90D0BF90D0BF90D0BF90D0BF90CFBF8FD0BF90605F40
605F4F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000009F905F908F50908F50908F4F0F0F0F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000303040000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F806050806050806F508F6050806F5F8F6F508F6F5F8F6F5F8F706090
705F907060907060907F6F9F7F609F7F6F9F7F6F9F806F9F7F6F9F806FA0806F9F8070A0806F9F\
8F70A08070A08F70A08F70A08F70A08F70A0907FAF8F70A0907FAF907FA0907FAF
907FA09F80AF907FAF9F80AF9F80AF9F80AF9F80AFA08FB09F80AFA08FB04F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFB0B0E0BFBF
DFB0B0E0BFB0DFB0B0E0BFB0DFB0B0E0BFB0DFB0B0E0BFB0DFB0AFDFBFB0DFB0AFDFBFB0DFB0AF\
DFBFAFDFB0AFDFBFAFDFB0A0DFBFAFDFB0A0DFBFAFD0B0A0DFBFA0D0BFA0DFBFA0
D0B0A0DFBFA0D0BF9FDFBFA0D0B09FDFBF9FD0BF9FDFBF9FD0B09FD0BF9FD0BF9FD0BF9FD0BF90\
D0BF9FD0BF90D0BF9FD0BF90D0BF90D0BF90D0BF90CFB08FD0BF90CFB08F605F4F
605040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F000F0F0F0F0F0F0F0F00\
0000000000000000000000000000000000000000000000004F4F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000020202F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F000F6050806050806F508F60508F6F5F8F6F5F8F705F906F5F8F706090705F90
7F609F7060907F6F9F7F609F806FA07F6F9F806FA07F6F9F8070A0806F9F8070A0806F9F8F70A0\
8F70A08F70A08F70A08F7FAF8F70A0907FAF8F70A0907FAF907FAF907FAF907FAF
9F80AF907FAF9F80AF9F80AF9F80B09F80AFA08FB09F80AFA08FB0A08FB050405F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFDFB0B0
E0BFBFDFBFB0E0BFBFDFB0B0E0BFBFDFBFB0E0BFB0DFBFB0E0BFB0DFBFB0E0BFB0DFBFAFE0BFB0\
DFBFAFDFBFAFDFBFAFDFBFAFDFBFAFDFBFAFDFBFA0DFBFAFDFBFA0DFBFAFDFBFA0
DFBFA0DFBFA0DFBFA0D0BF9FDFBFA0D0BF9FDFBFA0D0BF9FDFBFA0D0BF9FDFBF9FD0BF9FDFBF9F\
D0BF9FD0BF9FD0BF90DFBF9FD0BF90D0BF90D0BF90D0BF90D0BF90D0BF90605F40
6F5F4F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000908F50908F4F908F4F\
8F804F0F0F0F0F0F000F0F0F0000000000000000000000004F4F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F1F20000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000F\
0F0F0F5F4F806050806050806F508F6050806F5F8F6F5F8F705F906F5F8F706090
7060907F609F7060907F6F9F7F609F806F9F7F6F9F806F9F806F9F8070A0806F9F8070A0806F9F\
8F70A08070A08F70A08F70A08F7FA08F70A0907FAF8F7FA0907FAF907FA0907FAF
907FAF9F80AF907FAF9F80AF9F80AF9F80B09F80AFA08FB09F80AF50405F4F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFB0B0E0BFBF
DFB0B0E0BFBFDFB0B0E0BFB0DFB0B0E0BFB0DFB0B0E0BFB0DFB0AFE0BFB0DFB0AFDFBFAFDFB0AF\
DFBFAFDFB0AFDFBFAFDFBFAFDFBFAFDFB0A0DFBFAFDFBFA0DFBFAFD0B0A0DFBFA0
D0BFA0DFBFA0D0BF9FDFBFA0D0BF9FDFBFA0D0BF9FDFBF9FD0BF9FDFBF9FD0BF9FD0BF9FD0BF90\
D0BF9FD0BF90D0BF90D0BF90D0BF90D0BF90D0BF90CFBF90D0BF90CFB08FD0BF90
605F40000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000090804F908F4F8F804F\
90804F8F7F408F8040807F3F807F40000000000000000000504F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003030401F1F20000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
00000F6050806050806F508F60508F6F5F8F6F5F8F705F906F5F8F706090706090
7F609F7060907F6F9F7F609F806FA07F6F9F806FA0806F9F8070A0806F9F8F70A08070A08F70A0\
8070A08F70A08F70A08F7FAF8F70A0907FAF907FA0907FAF907FAF9F80AF907FAF
9F80AF907FAF9F80B09F80AF9F80B09F80AFA08FB09F80B0A08FB04F4050504F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFE0BFB0
E0BFBFDFBFB0E0BFBFE0BFB0E0BFBFDFBFB0E0BFB0DFBFB0E0BFB0DFBFB0E0BFB0DFBFAFE0BFB0\
DFBFAFE0BFB0DFBFAFDFBFAFDFBFAFDFBFAFDFBFAFDFBFAFDFBFA0DFBFAFDFBFA0
DFBFA0DFBFA0DFBFA0DFBFA0DFBFA0D0BF9FDFBFA0D0BF9FDFBFA0D0BF9FDFC09FD0BF9FDFBF9F\
D0BF90DFC09FD0BF90D0BF9FD0BF90D0C09FD0BF90D0BF90D0BF90D0BF90CFBF8F
6F5F4F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F908F4F908F4F\
8F804F8F804F8F80408F8040807F3F807F3F7F703F000000504F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000303040000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
605080604F806050806050806F508F6F508F6F5F8F6F5F8F705F90705F90706090
7060907F609F7060907F6F9F7F609F806F9F7F6F9F806FA07F6F9F806FA0806F9F8F70A0806F9F\
8F70A08F70A08F70A08F70A08F7FA08F70A0907FAF8F7FA0907FAF907FAF9080AF
907FAF9F80AF9080AF9F80B09F80AF9F80B09F80AFA08FB09F80AF50405F4F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFB0B0E0BFBF
DFB0B0E0BFBFDFB0B0E0BFBFDFB0B0E0BFB0DFBFB0E0BFB0DFB0AFE0BFB0DFBFAFE0BFB0DFB0AF\
DFBFAFDFBFAFDFBFAFDFB0AFDFBFAFDFBFA0DFBFAFDFBFA0DFBFA0DFBFA0DFBFA0
D0BFA0DFBFA0D0BFA0DFBFA0D0BF9FDFBFA0D0BF9FDFBFA0D0BF9FDFBF9FD0BF9FDFBF9FD0BF90\
D0BF9FD0BF90D0BF9FD0BF90D0BF90D0BF90D0BF90CFBF90D0BF90CFBF8FD0BF90
605040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000908F508F804F\
90804F8F80408F8040807F3F807F3F7F703F7F703F000000505030000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000303040303040000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
604F806050806050806F5F8F6F508F6F5F8F6F5F8F706090705F90706090706090
7F609F7F609F7F6F9F7F609F806FA07F6F9F806FA0806F9F8070A0806FA08F70A0806FA08F70A0\
8F70A08F70A08F70A0907FAF8F70A0907FAF8F7FA0907FAF907FAF9080AF907FAF
9F80B09F80AF9F80B09F80AFA08FB09F80B0A08FB09F80B0A08FB050405F50405F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFE0BFB0
E0BFBFE0BFB0E0BFBFE0BFB0E0BFBFE0BFB0E0BFB0DFBFB0E0BFB0DFBFB0E0BFB0DFBFAFE0BFB0\
DFBFAFE0BFB0DFBFAFE0BFAFDFBFAFDFBFAFDFBFA0DFBFAFDFBFA0DFBFAFDFBFA0
DFBFAFDFBFA0DFBFA0DFBFA0DFC0A0DFBFA0DFBFA0D0BF9FDFC0A0D0BF9FDFBF9FD0BF9FDFC09F\
D0BF9FDFC09FD0BF90DFC09FD0BF90D0C09FD0BF90D0C090D0BF90D0BF90CFBF8F
6F5F4F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F908F4F908F50\
8F804F8F804F8F80408F8040807F3F807F3F7F703F0000002F2F1F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000303040303040000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000F\
605080604F806050806050806F5F8F60508F6F5F8F6F5F8F7060906F5F90706090
7060907F609F7060907F6F9F7F6F9F806FA07F6F9F806FA0806F9F8070A0806F9F8070A08070A0\
8F70A08070A08F70A08F70A0907FAF8F70A0907FAF907FA0907FAF907FAF9080AF
907FAF9F80AF907FAF9F80AF9F80AFA08FB09F80AFA08FB09F80B050405F4F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFB0B0E0BFBF
DFBFB0E0BFBFDFB0B0E0BFBFDFBFB0E0BFB0DFB0B0E0BFB0DFBFB0E0BFB0DFBFAFE0BFB0DFBFAF\
E0BFB0DFBFAFDFBFAFDFBFAFDFBFAFDFBFA0DFBFAFDFBFA0DFBFAFDFBFA0DFBFA0
DFBFA0DFBFA0D0BFA0DFBFA0D0BF9FDFBFA0D0BF9FDFBF9FD0BF9FDFBF9FD0BF9FDFBF9FD0BF9F\
DFBF9FD0BF90D0BF9FD0BF90D0BF9FD0BF90D0BF90D0BF90D0BF90CFBF8FD0BF90
605F40000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000908F4F908F4F\
908F4F8F80408F8040807F40807F407F703F7F703F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F3040303040000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
604F8060508F6050806F5F8F6F508F6F5F906F5F8F705F90705F9070609F706090
7F609F70609F7F6F9F7F6F9F806FA07F6F9F8070A0806FA08070A0806FA08F70A08070A08F70A0\
8F70A08F7FAF8F70A0907FAF8F70A0907FAF907FAF907FAF907FAF9F80AF907FAF
9F80B09F80AF9F80B09F80AFA08FB09F80AFA08FB0A08FB0A08FB04F4050504F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFE0BFBF
E0BFBFE0BFB0E0BFBFE0BFB0E0BFBFE0BFB0E0BFBFE0BFB0E0BFB0DFBFB0E0BFB0DFBFB0E0BFB0\
DFBFAFE0BFB0DFBFAFE0BFAFDFBFAFE0C0AFDFBFAFDFBFAFDFBFA0DFC0AFDFBFA0
DFBFAFDFBFA0DFC0A0DFBFA0DFC0A0DFBF9FDFC0A0DFBF9FDFC0A0D0BF9FDFC0A0D0BF9FDFC09F\
D0BF9FDFC09FD0BF9FDFC09FD0BF90DFC09FD0BF90D0C090D0BF90D0C090D0BF90
6F5F4F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F908F4F908F50\
8F804F908F4F8F80408F8040807F3F807F407F703F0F0F0F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030304030304F000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000F\
60508060508060508F6050806F5F8F6F508F6F5F906F5F8F705F90705F9070609F
7060907F609F7F609F7F6F9F7F6F9F806FA07F6F9F806FA0806F9F8070A0806FA08F70A0806FA0\
8F70A08F70A08F7FAF8F70A0907FAF8F70A0907FAF907FA0907FAF907FAF9F80AF
907FAF9F80AF9F80AF9F80B09F80AF9F80B09F80B0A08FB09F80B050405F4F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFB0E0BFBF
DFBFB0E0BFBFE0BFB0E0BFBFDFBFB0E0BFB0DFBFB0E0BFB0DFBFB0E0BFB0DFBFAFE0BFB0DFBFAF\
DFBFB0DFBFAFE0BFAFDFBFAFDFBFAFDFBFAFDFBFAFDFBFA0DFBFAFDFBFA0DFBFAF
DFBFA0DFBFA0DFBFA0DFBFA0D0BF9FDFBFA0D0BF9FDFBFA0D0BF9FDFBF9FD0BF9FDFC09FD0BF9F\
DFBF9FD0BF90DFC09FD0BF90D0BF90D0BF90D0C090D0BF90D0BF90D0BF90D0BF90
605F40000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000908F4F908F4F\
908F4F8F80408F8040807F3F807F407F7F3F7F7F3F706F300F0F00404030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030403030401F1F20000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
60508060508F6050806F5F8F6F5F8F6F5F906F5F8F706090705F9070609F706090
7F6F9F7F609F7F6F9F7F6F9F806FA0806FA08070A0806FA08F70A0806FA08F70A08070A08F70AF\
8F70A08F7FAF8F70A0907FAF8F7FAF907FAF907FAF9080AF907FAF9F80AF907FAF
9F80B09F80AF9F80B09F80AFA08FB09F80B0A08FB0A08FB0A08FBF50405F504F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFE0BFBF
E0BFBFE0BFBFE0BFBFE0BFB0E0BFBFE0BFB0E0BFBFE0BFB0E0C0B0E0BFB0E0BFB0DFBFB0E0C0B0\
DFBFAFE0BFB0DFBFAFE0C0B0DFBFAFE0BFAFDFBFAFE0C0AFDFBFAFDFC0AFDFBFA0
DFC0AFDFBFA0DFC0A0DFBFA0DFC0A0DFBFA0DFC0A0DFBF9FDFC0A0DFBF9FDFC0A0D0BF9FDFC09F\
D0BF9FDFC09FD0BF90DFC09FD0BF90DFC09FD0BF90DFC09FD0BF90D0BF90D0BF90
6F5F4F605F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000908F4F9F8F50\
908F4F908F4F8F80408F8040807F40807F407F703F7F703F605F304F4F3F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030403030401F1F20000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000F\
60508060508060508F60508F6F5F8F6F508F6F5F906F5F8F706090705F9070609F
7060907F6F9F7F609F7F6F9F7F6F9F806FA07F6F9F806FA0806FA08070A0806FA08F70A08070A0\
8F70A08F70A08F7FAF8F70A0907FAF8F70A0907FAF907FAF9080AF907FAF9F80AF
907FAF9F80B09F80AF9F80B09F80AFA08FB09F80AFA08FB0A08FB050405F4F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000DFBFB0E0BFBF
E0BFB0E0BFBFDFBFB0E0BFBFE0BFB0E0BFBFDFBFB0E0BFB0DFBFB0E0BFB0DFBFAFE0BFB0DFBFAF\
E0BFB0DFBFAFDFBFAFDFBFAFE0BFAFDFBFAFDFBFAFDFBFA0DFBFAFDFBFA0DFBFAF
DFBFA0DFBFA0DFBFA0DFBFA0DFBFA0DFC0A0D0BF9FDFBFA0D0BF9FDFC0A0D0BF9FDFBF9FD0BF9F\
DFC09FD0BF90DFC09FD0BF90DFC09FD0BF90D0BF90D0BF90D0BF90CFBF90D0BF90
605F40605F4F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000908F4F908F4F\
908F4F8F80408F8040807F408080407F7F3F7F7F3F70703060603F40402F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030403F304F1F1F20000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
6050806F508F60508F6F5F8F6F508F705F906F5F90706090705F907F609F706090
7F6F9F7F609F806FA07F6F9F806FA0806FA08070A0806FA08070A0806FA08F70AF8F70A08F70AF\
8F70A0907FAF8F70A0907FAF8F7FAF907FAF907FAF9080AF907FAF9F80B09080AF
9F80B09F80AF9F8FB09F80B0A08FB09F80B0A08FB0A08FB0A08FB04F405F504F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0C0BFE0BFBF
E0BFBFE0BFBFE0C0BFE0BFBFE0BFBFE0BFB0E0C0BFE0BFB0E0BFB0E0BFB0E0C0B0E0BFB0E0C0B0\
DFBFAFE0C0B0DFBFAFE0C0B0DFBFAFE0C0AFDFBFAFE0C0AFDFBFAFE0C0AFDFBFA0
DFC0AFDFBFA0DFC0AFDFBFA0DFC0A0DFBFA0DFC0A0DFBFA0DFC0A0DFBF9FDFC0A0DFBF9FDFC09F\
D0BF9FDFC09FD0BF9FDFC09FD0BF90DFC09FD0BF90D0C09FD0BF90D0C090D0BF90
6F5F4F605F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000908F4F9F8F50\
908F4F908F4F8F80408F8040807F40807F407F7F3F7F703F606030000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030403030401F1F20000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000F\
6050806050806F508F60508F6F5F8F6F5F8F705F906F5F8F7060907060907F609F
7060907F6F9F7F609F806FA07F6F9F806FA0806FA08070A0806FA08070A0806FA08F70A08070A0\
8F70A08F70A08F7FAF8F70A0907FAF8F7FAF907FAF907FAF907FAF907FAF9F80B0
907FAF9F80B09F80AF9F80B09F80AFA08FB09F80B0A08FB0A08FB050405F4F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFE0BFBF
E0BFB0E0BFBFE0BFB0E0BFBFE0BFB0E0BFBFE0BFB0E0BFB0DFBFB0E0BFB0DFBFB0E0BFB0DFBFAF\
E0BFB0DFBFAFE0BFB0DFBFAFDFBFAFDFBFAFE0C0AFDFBFA0DFBFAFDFBFA0DFC0AF
DFBFA0DFBFA0DFBFA0DFC0A0DFBFA0DFBFA0DFBF9FDFC0A0D0BF9FDFC0A0D0BF9FDFC09FD0BF9F\
DFC09FD0BF9FDFC09FD0BF90D0C09FD0BF90D0C09FD0BF90D0BF90D0BF90D0BF90
CFBF8F605F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F00908F4F\
908F4F8F80408F804F807F40808040807F3F807F3F7F70306F603F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030403F304F000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
6050806F508F60508F6F5F906F5F8F705F906F5F9070609070609070609F706090
7F6F9F7F609F7F6FA07F6F9F8070A0806FA08070A0806FA08F70A08070A08F70A08070A08F7FAF\
8F70A08F7FAF8F70A0907FAF907FAF907FAF907FAF9F80B0907FAF9F80B09080AF
9F80B09F80B09F8FB09F80B0A08FB0A08FB0A08FB0A08FB0A090BF50405F504F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFE0BFBF
E0C0BFE0BFBFE0C0BFE0BFBFE0C0BFE0BFB0E0C0BFE0BFB0E0C0BFE0BFB0E0C0B0E0BFB0E0C0B0\
E0BFB0E0C0B0DFBFAFE0C0B0DFBFAFE0C0AFDFBFAFE0C0AFDFBFAFE0C0AFDFBFA0
E0C0AFDFBFA0DFC0AFDFBFA0DFC0A0DFBFA0DFC0A0DFBFA0DFC0A0DFBF9FDFC0A0DFBF9FDFC0A0\
DFBF9FDFC09FD0BF9FDFC09FD0C09FDFC09FD0BF90DFC09FD0BF90D0C090D0BF90
D0C090605F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000009F9050\
908F4F908F4F8F80408F804F807F40807F407F7F3F807F3F6F6F30000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
303040303040000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000F\
60508F6050806F508F60508F6F5F906F5F8F705F906F5F90706090705F907F609F
70609F7F6F9F7F609F806FA07F6F9F806FA07F6F9F8070A0806FA08F70A0806FA08F70A08F70A0\
8F7FAF8F70A08F7FAF8F70A0907FAF8F7FAF907FAF907FAF9F80AF907FAF9F80B0
9080AF9F80B09F80AF9F80B09F80B0A08FB09F80B0A08FB0A08FB050405F4F4050000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0BFBFE0BFBF
E0BFBFE0BFBFE0BFB0E0BFBFE0BFB0E0BFBFE0BFB0E0BFB0E0BFB0E0BFB0DFBFB0E0BFB0DFBFAF\
E0C0B0DFBFAFE0BFB0DFBFAFE0C0AFDFBFAFDFBFAFDFBFAFE0C0AFDFBFA0DFC0AF
DFBFA0DFC0AFDFBFA0DFC0A0DFBFA0DFC0A0DFBF9FDFC0A0DFBF9FDFC0A0D0BF9FDFC09FD0BF9F\
DFC09FD0BF9FDFC09FD0BF90DFC09FD0BF90D0C09FD0BF90D0C090D0BF90D0BF90
D0BF8F605F4F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0090804F\
908F4F8F80408F804F8F80408F8040807F3F807F3F7F70306F6F3F505030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030403F304F000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F000F0F0F0F\
6050806F5F8F60508F6F5F906F5F8F705F906F5F907060907060907F609F70609F
7F6F9F7F609F806FA07F6FA08070A0806FA08070A0806FA08F70A08070A08F70AF8F70A08F7FAF\
8F70A0907FAF8F7FAF907FAF907FAF9080AF907FAF9F80B0907FAF9F80B09F80AF
9F80B09F80B0A08FB09F80B0A08FB09F8FB0A08FBFA08FB0A08FBF50405F504F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F5F60706060705F60706060\
705F607060607060607060607060607F60607060607F6060706060EFC0C0E0BFBF
E0C0BFE0BFBFEFC0BFE0BFBFE0C0BFE0BFB0E0C0BFE0BFB0E0C0BFE0BFB0E0C0B0E0BFB0E0C0B0\
E0BFB0E0C0B0E0BFAFE0C0B0DFBFAFE0C0B0DFBFAFE0C0AFDFBFAFE0C0AFDFBFAF
E0C0AFDFBFA0E0C0AFDFBFA0DFC0A0DFBFA0DFC0A0DFC0A0DFC0A0DFBF9FDFC0A0DFC09FDFC0A0\
DFBF9FDFC09FDFC09FDFC09FD0C09FDFC09FD0C090DFC09FD0BF90DFC09FD0BF90
D0C090605F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000908F50\
908F4F908F4F8F80408F804F807F408F8040807F3F807F3F6F6F305F5F3F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
303040303040000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F5F4F80\
60508F6050806F5F8F60508F6F5F906F5F8F7060906F5F907060907060907F609F
70609F7F6F9F7F6F9F806FA07F6F9F806FA0806FA08070A0806FA08F70A08070A08F70AF8F70A0\
8F70AF8F70A0907FAF8F70A0907FAF907FAF907FAF907FAF9080AF907FAF9F80B0
9080AF9F80B09F80B0A08FB09F80B0A08FB0A08FB0A08FBFA08FB050405F50405F000000000000\
000000504F5F5F50605F4F5F5F50605F4F605F50605F4F605F50605F4F60605060
5F4F606050605F50606050605F5060605060605060605060605060605060605060605F60605060\
6050606050606F5F606050606F5F606050606F5F606050606F5F606F5F606F5F60
6F5F606F5F606F5F606F5F606F5F60D0B0BFD0B0BFDFB0C0D0B0BFDFB0BFD0B0BFDFBFC0DFB0BF\
DFBFBFDFB0BFE0BFBFDFB0BFE0BFBFDFBFBFE0BFBFE0BFBFE0BFBF6F5F5F705F5F
6F5F5F705F5FE0BFBFE0C0BFE0BFB0E0BFBFE0BFB0E0C0BFDFBFB0E0BFB0E0BFB0E0C0B0DFBFAF\
E0C0B0DFBFAFE0C0B0DFBFAFE0C0AFDFBFAFE0C0AFDFBFAFDFC0AFDFBFA0E0C0AF
DFBFA0DFC0AFDFBFA0DFC0A0DFBFA0DFC0A0DFBFA0DFC0A0DFBF9FDFC0A0DFBF9FDFC0A0D0BF9F\
DFC09FD0BF9FDFC09FD0BF90DFC09FD0BF90DFC09FD0BF90D0C090D0BF90D0C090
CFBF8F605F4F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0090804F\
908F4F8F804F908F4F8F80408F8040807F3F807F3F7F703070703F5F5F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3030403F304F000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000F60508F\
60508F6F5F8F60508F6F5F906F5F8F705F906F5F9070609F7060907F609F70609F
7F6FA07F6F9F806FA07F6FA08070A0806FA08070A0806FA08F70AF8070A08F70AF8F70A08F7FAF\
8F70AF907FAF8F7FAF907FAF907FAF907FAF907FAF9F80B04F3F504F405F4F3F50
50405F4F405050405F4F405050405F4F405F50405F4F405F504F5F50405F504F5F504F5F5F5060\
AF90BFAF90BFAF90BFB09FC0B090BFB09FC0B09FBFBF9FC0B09FC0BFA0C0B09FC0
BFA0C0BF9FC0BFA0C0BF9FC0BFA0C0BFA0C0C0A0C0BFA0C0C0A0C0C0A0C0C0A0C0C0A0BFC0AFC0\
C0A0C0CFAFC0C0A0BFCFAFC0CFAFC0CFAFC0CFAFBFCFAFC0CFAFBFD0B0C0CFAFBF
D0B0C0D00000D00000D00000D00000D00000DF0000D00000DF7070DF6F70DFBFC0DFB0BFDFBFC0\
DFBFBFE0BFC0DFBFBFE0BFC0E0BFBFE0BFC0E0BFBFE0C0C0E0BFBF000000000000
00000000000070605F6F5F5F70605F6F5F5F70605F6F5F5F70605F6F5F5F70605F6F5F5070605F\
6F5F5070605F6F5F5070605F6F5F5070605F6F5F5070605F6F5F506F6050DFBFAF
E0C0AFDFC0A0E0C0AFDFBFA0E0C0AFDFC0A0DFC0A0DFBFA0DFC0A0DFC0A0DFC0A0DFC09FDFC0A0\
DFC09FDFC09FDFC09FDFC09FDFC09FDFC09FD0C090DFC09FD0C090DFC09FD0BF90
D0C090605F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000908F50\
908F4F908F4F8F80408F8F4F8F80408F8040807F3F807F3F70703060603F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303040\
30304F1F1F20000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F605080\
60508F6050806F5F8F6F508F6F5F906F5F8F705F90705F9070609F7060907F609F
7F609F7F6FA07F6F9F806FA07F6FA08070A0806FA08070A040304F403F5040304F403F50403F50\
403F50403F50403F50403F504F3F50403F504F3F50403F504000003F0000400000
3F00004000003F00004000003F00004000001F00000F0F0F0000005F4F60504F5FAF90BFAF90BF\
A08FB0AF7090AF7F9FAF0000B00000AF0000B00000B00000BF0000B00000BF0000
B00000BF0000B00000BF809FBF7F9FBFA0C0BF9FBFBFA0C0BFA0BFC0A0C0BFA0BFC0A0C0C0A0BF\
C08F9FC08090C00000C00000CF0000C00000CF0000CF0000CF0000CF0000CF0000
CF0000D00000CF0000D00000D08F90D0909FD0B0BFD0B0C0D0B0BFDFB0C0D0B0BFDFBFBFDFB0BF\
DFBFC0DFB0BFDFBFBFDFBFBFE0BFC0DFBFBFE0BFBFE0BFBFE0C0BF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006F60506F5F4F6F5F506F5F4F6F60506F5F4F6F5F506F5F4F6F60506F5F4F6F5F4F6F5F4F\
6F604F605F4F6F5F4F6F5F4F6F604F605F4FDFC09FD0BF90DFC09FD0BF90D0C090
D0BF906F5F4F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000908F4F\
908F508F804F908F4F8F80408F8040807F3F807F3F7F703F7F703F5F5F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303040\
3030401F1F2F000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F000F60508F\
60508F6F5F8F6F508F6F5F906F5F8F706090705F9070609F7060907F6F9F7F609F
7F6FA07F6F9F806FA0806FA08070A03F304F403F50000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F2000009F00007F00007F00005F0000
4000003F0000400000400000600000600000A00000A00000AF0000A00000AF0000AF3F4FAF0000\
AF0000B09FC0AF90BFB09FC0B07F9FB07F9FB00000BF0000B0202FAF90BFAF90B0
BFA0C0B09FC0BFA0C0BFA0C0BFA0C0BFA0C0C0A0C0BFA0C0C0A0C0BFA0C0C0AFC0C0A0C0C0AFC0\
C0A0C0CFAFC0C0AFC0CFAFC0CFAFBFCFAFC0C0A0BFCFAFBFCF0000D00000CF0000
D00000D00000D06F7FD06F70D0B0C0D0B0BFDFB0C0D0B0BFDFBFC0DFB0BFDFBFC0DFB0BFDFBFC0\
DFB0BFE0BFC0DFBFBFE0BFC0E0BFBFE0C0C0E0BFBFE0C0C0E0BFBF000000000000
000000000000000000000000EF0000E00000EF0000E00000E00000E00000E00000E00000E00000\
E00000E00000E00000E00000DF0000B00000B00000000000000000000000000000
0000000000000000000000000000000000000000002F00003000005F00005F0000000000000000\
0000000000000000000000000000000000000000006F604FD0C090DFC09FD0C090
D0C090605F406F604F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000908F50\
908F4F908F4F8F8F4F8F8F4F8F80408F8040807F3F807F3F7F70306F603F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303040\
3030401F1F20000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F604F80\
60508F60508F6F5F8F6F508F6F5F906F5F90706090705F9070609F70609F3F304F
3F304F3F304F3F304F403F50000000000000000000500000000000000000000000000000000000\
0000000000000F0F0F3F00003F00009000003F00003F0000000000000000000000
000000000000000000000000000000000000000000600000600000AF0000A00000AF0000AF0000\
AF0000AF5F70B05F70AF90BFB09FC0B09FBFB06070B05F70B0202FB09FC0BFA0C0
B09FBFBFA0C0BF9FC0BFA0C0BF9FBFBFA0C0BFA0C0C0A0C0BFA0BFC0A0C0C0A0BFC0AFC0C0A0BF\
C0AFC0C0A0BFCFAFC0C0A0BFCFAFC0CFAFBFCF8F9FCF8F90CF0000CF0000D00000
CF0000D00000C0A0B0C0AFB0D0AFBFD0B0C0D0B0BFDFB0C0D0B0BFDFB0C0DFB0BFDFBFC0DFB0BF\
DFBFC0DFB0BFE0BFC0DFBFBFE0BFBFE0BFBFE0BFC0E0BFBFE0BFBF000000000000
0000000000000000000000000000000000000000000000002F0000B00000B00000E00000DF0000\
B00000B000003000002F00003000000000000F0F0F000000000000000000000000
0000000000000000000000000000000000000000005F0000500000DF0000DF00000F0F0F000000\
0000000000000000000000000000000000000000000000006F5F4F605F4F6F604F
605F40605F4F605F40000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
908F4F908F4F908F4F8F80408F8040807F3F807F407F703F7F703F606030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F304F\
3030401F1F2F000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000F60508F\
60508F6F5F8F6F508F3F304F302F403F304F302F4F3F304F3F304F000000000000
000000000000000000000000800000800000500000000000000000000000000000000000000000\
00000F700000700000900000900000900000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000801F205F4F60AF5F70AF0000\
AF0000B04050AF3F4FB09FC0B09FC0B09FC0B09FC0BFA0C0B0202FAF90BFAF90B0
BFA0C0BF9FC0BFA0C0BFA0C0C0A0C0BFA0C0C0A0C0BFA0C0C0AFC0C0A0C0C0AFC0C0A0C0CFAFC0\
C0AFC0CFAFC0C0AFC0CFAFC0CFAFC0CFAFC0CF6070D06F7FCF0000D00000CF0000
D02F2FD0202FD0B0C0D0B0BFDFB0C0D0B0C0DFB0C0D0B0BFDFBFC0DFB0C0DFBFC0DFB0BFE0BFC0\
DFBFBFE0BFC0DFBFBFE0BFC0E0BFBFE0C0C0E0BFBFEFC0C0E0BFBF000000000000
0000000000000000000000000000000000000000000000000000005F00005F0000E00000E00000\
5F00005F0000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008000008F0000DF0000DF00005F00005F0000\
000000000000000000000000000000000000000000000000800000000000000000
0000006F604F6F604F6F604F6F604F6F604F6F604F6F5F406F604F6F5F406F604F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
908F4F908F4F8F8F4F908F4F8F80408F8040807F3F807F407F7F3F6F6F3F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303040\
30304F1F1F20000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F40302040302F4F302040\
302F40302F400000000000000000000000000000000000000000007000007F0000
7F00007F00007F0000800000000000000000000000000000000000000000000000000000000000\
7000008F00009000008F0000900000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F4F60A08FBFAF90BFAF3F4F\
AF3F4FAF1F20B0202FAF90BFB09FC0B090BFB09FC0B09FC0BF607FB05F70BF6070
B09FC0BFA0C0B09FC0BFA0C0BFA0C0BFA0C0BFA0C0C0A0C0BFA0C0C0A0C0BFA0BFC0AFC0C0A0C0\
C0AFC0C0A0BFCFAFC0C0AFC0CFAFC0CFAFBFCF4050CF404FD00000CF0000D00000
CF404FD04F50D0AFBFD0B0C0D0B0BFDFB0C0D0B0BFDFB0C0D0B0BFDFBFC0DFB0BFDFBFC0DFB0BF\
E0BFC0DFB0BFE0BFC0DFBFBFE0BFC0E0BFBFE0BFC0E0BFBFEFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000E00000E00000\
5F00005F0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000B00000AF0000DF0000DF0000DF0000DF0000\
0F0F0F000000000000000000000000000000000000000000000000D00000D00000
D00000D00000CF0000D00000CFBF8FD0BF8FCFB080CFBF8FCFB080CFBF80C0B080CFB080605F40\
6F6040605F3F605F40605F3F605F40605F3F605F3F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
9F9050908F4F908F4F8F80408F8040807F3F807F407F7F3F7F7F3F6F60302F2F1F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000030304F\
303040000000000000000000000000000000000000000000000000000000000000
0000000000002F20402F203F2F2F402F2040302F40302F405F4F805F4F8060508F5F4F7F000000\
0000000000000000000000000000000000005F00007000001F00006000007F0000
7F00007F00005000004F00000000000000000000000000000000000000000000003F00003F0000\
8F00009000009000007F00003F0000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000504F60AF90BFAF90BFAF90BF\
A08FB0B0202FA08FB0AF90B0B09FBFB09FC0B09FC0BFA0C0AF90BFBF202FBF202F
B09FBFBF9FC0BFA0C0BFA0C0C0A0C0BFA0C0C0A0C0C0A0C0C0AFC0C0A0C0C0AFC0C0A0C0CFAFC0\
C0AFC0CFAFC0CFAFC0CFAFC0CFAFC0CFB0C0CF202FD02F2FCF0000D00000D00000
D00000D06F70D0B0C0D0B0C0DFB0C0D0B0C0DFBFC0DFB0C0DFBFC0DFB0BFDFBFC0DFBFC0E0BFC0\
DFBFBFE0BFC0E0BFC0E0BFC0E0BFBFE0C0C0E0BFBFEFC0C0E0BFBF000000000000
0000000000000000000000000000000000000000000000000000000000000000008F00008F0000\
8F00008F0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000DF0000E00000DF0000E00000DF0000DF0000\
5F00005F00000000000000000000000000000000000000000000003000002F0000
2F00002F0000D00000D00000D00000CF0000D00000CF0000CF0000CF0000CF0000C00000CF0000\
C0B07FCFB080C0B07FC0B07FBFB070C0B07FBFAF70C0B070605F3F605F3F605F3F
605F3F5F5030605F3F5F5030605F3F0000000000000000000000000000000000000000000F0F0F\
908F4F9F9050908F4F908F4F8F80408F8040807F3F807F3F7F7F3F70703F5F5F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303040\
30304F000000000000000000000000000000000000000000000000000000000000
0000002F203F4F407050407F50407F5F4F7F50407F5F4F805F00006000005F000060203F000000\
0000000000000000000000000000000000000000000000000000001F00007F0000
7F00007F00001F00003F00000000000000000000000000000000000000000000007000006F0000\
8F00008F00009000003F0000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F4F60A08FBFAF90BFAF90BF\
AF90BFAF7090B07F9FAF7F90B09FC0B09FC0B09FC0B09FC0BF9FC0B05F70BF607F
B01F20B09FBFB09FBFBFA0C0BF9FC0BFA0C0BFA0C0C0A0C0BFA0C0C0A0C0C0A0C0C0AFC0C0A0C0\
C0AFC0C0AFC0CFAFC0C0AFBFC0A0BFC0A0BFCF0000CF0000CF0000CF0000D00000
CF0000D08F9FD0B0BFD0B0C0D0B0BFD0B0C0D0B0BFDFBFC0D0B0BFDFBFC0DFB0BFDFBFC0DFB0BF\
DFBFC0DFBFBFE0BFC0DFBFBFE0BFC0E0BFBFE0C0C0E0BFBFE0C0C0000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F0000\
B00000B00000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F0000DF0000DF0000DF0000DF0000DF0000DF0000\
DF0000DF00000F0F0F000000000000000000000000000000000000000000000000
0000000000007F4F3FD00000CF0000D00000CF0000CF0000CF6F4FCF7050C0906FC0B080BFAF7F\
BFB07FC00000C00000BF0000C00000BF0000C00000BF2010BF201FBFAF6FBFAF70
B0AF6FBFAF6FB0A06FB0AF6FB0A060000000000000000000000000000000000000000000000000\
908F4F908F4F908F4F8F80408F80408080408080407F7F3F7F7F3F6F6F305F5F3F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F304F\
30304F00000000000000000000000000000000000000000000000000000020203F
2F20404F40705000005000005F00005000005F00005F1F305F4F805F4F8060508F5F4F7F000000\
0000000000000000000000000000000000000000000000000000000000007F0000
8000007F00003F00000000000000000000000000000000000000001F00002000008F0000900000\
8F0000900000900000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000504F60AF90BFAF90BFB09FC0\
AF90BFB09FC0AF90BFB09FC0B09FC0B09FC0B09FC0BFA0CFB09FC0B09FC0BF202F
BF4050BF6070BF607FBFA0C0C0A0CFBFA0C0C0A0C0BFA0C0C0AFCFC0A0C0C0AFC0C0A0C0CFAFC0\
C0AFC0CFAFC0C0AFC0CFAFC0CF6F70CF6F7FCF0000D00000CF0000D00000D00000
D00000D0B0C0D0B0C0D0B0C0DFBFC0D0B0C0DFBFC0DFB0C0DFBFC0DFBFC0DFBFC0DFBFC0E0BFC0\
DFBFC0E0BFC0E0BFC0E0C0C0E0BFC0E0C0C0E0BFBFEFC0C0E0C0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
E00000E00000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F00005F0000E00000DF0000E00000DF0000E00000\
DF0000DF00005F00005F0000000000000000000000000000000000000000000000
00000000000070604FD00000D00000CF0000D00000CF4F30CFBF8FCFBF80CFBF80C0B080CFBF80\
C0B07FC0B080C00000C04F30C04F30C0B07FBFAF70C0B07FBFAF70BFB070BFAF6F
BFAF70BFAF6FBFAF6FB0A06FBF4F2FB0A0600000000000000000000000000000000000000F0F0F\
908F4F9F9050908F4F908F4F8F80408F8040807F3F8080407F7F3F70703F5F5F30
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000303040303040\
3F304F0000000000000000000000000000000000000000002F203F201F3F4F4070
4F000050407F4F3F6F501F305000005F00005000005F3F6F5F4F806050805F4F805F4F80000000\
000000000000000000000000000000000000000000000000000000000000600000
7F00008000004F00000000000000000000000000000000000000003F00003000008F00008F0000\
900000700000700000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F4F60A08FBFAF90BFAF90BF\
B09FC0AF90BFB09FC0AF90BFB09FC0B09FBFB09FC0B09FC0BFA0C0B09FC0AF90BF
AF90B0BF0000BF0000BF202FBF2020C0202FBF2020C0202FC02020C0202FC02020C0202FC02020\
CF202FC02020CF202FCF2020CF202FCF2020CF0000CF0000D00000CF0000D0202F
D02020D0B0C0D0B0BFD0B0C0D0B0C0DFB0C0D0B0BFDFB0C0DFB0C0DFBFC0DFB0BFDFBFC0DFBFBF\
E0BFC0DFBFBFE0BFC0E0BFBFE0C0C0E0BFBFE0C0C0E0BFBFEFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F00005F00005F00005F0000000000000000000000000000000000000000000000
0000000000000000000000000000000000005000005F0000DF0000DF0000DF0000DF0000DF0000\
5F0000800000DF0000DF00000F0F0F000000000000000000000000000000000000
0000000000006F604FD02F20CF201FD00000CF0000CF7050CF6F50CFBF8FC0B080CFBF80C0B07F\
CFB080C0B07FC06F4FC0201FC06F4FBFAF70C0B070BFAF70BFB070BFAF6FBFAF70
B0AF6FBFAF6FB0A060B0AF6FB0A060B000005F50305F50305050305F5030000000000000000000\
908F4F908F4F908F4F8F80408F804F808040808040807F3F807F3F70703060603F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003030403F304F\
30304F0000000000000000000000002020302020304F3F6F403F6F4F40704F3F70
50407F4F407050407F5040705F00005000005F00005F2F4F6050805F4F805F50805F4F80000000\
0000000000000000000000000000000000000000000000000000000000004F0000
8000007F00006F00000000000000000000000000000000000000003F00003F00008F0000900000\
8F00007F0000700000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000504F60AF90C0AF90BFB09FC0\
AF90BFB09FC0B09FC0B09FC0B09FC0BFA0C0B09FC0BFA0C0B09FC0BFA0CFBFA0C0
BFA0CFBF404FC04050BF4050C04050BF809FC08FA0C0809FC08FA0C0809FCF8FA0C0809FCF8FA0\
C0809FCF8FA0CF8F9FCF8FA0CF202FD02F2FCF0000D00000CF0000D00000D06F70
D06F7FD0B0C0DFB0C0D0B0C0DFBFC0D0B0C0DFBFC0DFB0C0DFBFC0DFB0C0E0BFC0DFBFC0E0BFC0\
DFBFC0E0C0C0E0BFC0E0C0C0E0BFC0EFC0C0E0BFC0EFC0C0E0C0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F000F0F0FB00000BF0000000000000000000000000000000000000000000000
000000000000000000000000000000000000B00000B00000E00000DF0000E00000DF0000E00000\
2F00003000008000008000005F00005F0000000000000000000000000000000000
000000000000706F50D02F1FD02F20CF0000D00000CF4F3FCF4F3FCFBF80CFBF8FCFBF80CFBF80\
C0B07FCFBF80C0B07FC0B07FC04F30C0B07FBFB070C0B070BFAF70C0B070BFAF70
BFAF70B0AF6FBFAF6FB0AF6FBFAF6FB0A060B00000AFA060AFA060A09F5F5F5030504F2F50502F\
4F40209F9050908F4F908F4F8F80408F80408080408080407F7F3F7F7F3F606030
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000303040303040\
1F1F2000000020202F1F1F2F403F5F403F5F403F604000004F00004F3F704F4070
4F3F7050407F50407050407F500F1F5F00005F00005F30505F4F806050805F4F7F605080000000\
000000000000000000000000000000000000000000000000000000000000300000
7F00008000006000000000000000000000000000000000000000003F00003000008F00008F0000\
900000700000700000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F4F60AF90BFAF90C0AF90BF\
AF90C0AF90BFB09FC0AF90BFB09FC0B09FC0BF9FC0B09FC0BFA0C0B09FC0BFA0C0
B09FC0AF9FBFAF90B0BF202FBF2020C0A0C0BFA0C0C0AFC0BFA0C0C0AFC0C0A0C0C0AFC0C0A0C0\
CFAFC0C0AFC0CFAFC0C0AFC0CF0000CF0000CF0000CF0000D00000CF0000D06F7F
D06F70D0B0C0D0B0C0DFB0C0D0B0C0DFB0C0D0B0C0DFBFC0DFB0BFDFBFC0DFB0C0E0BFC0DFBFBF\
E0BFC0DFBFC0E0BFC0E0BFBFE0BFC0E0BFBFEFC0C0E0BFBFEFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000008F0000800000300000000000000000000000000000000000000000
000000000000000000000000000000000000B00000B00000DF0000DF0000DF0000DF0000DF0000\
000000000000000000000000DF0000DF00000F0F0F000000000000000000000000
0000000000006F604FD0503FCF4F30D00000CF0000D02F20CF201FCFBF80CFB080CFBF80C0B07F\
CFB080C0B07FC0B07FC0B07FC06F4FBFAF70C0B07FBFAF70BFB070BFAF6FBFAF70
B0AF6FBFAF6FB0AF6FBFAF6FB0A060B0AF60AF9F5FAF0000A00000AF0000A00000A09F5F9F9050\
0F0F004F40204F402F4040204040208F80408F8040807F3F807F3F7F70306F603F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003030401F1F2F\
1F1F2020202F3F3050403F5F403F5F40000040203F4F0000403F6F4F40704F3F70
50407F50407F50407F50407F5F101F5F00005F00005F1F3060508F6050805F5080000000000000\
0000000000000000000000000000000000000000000000000000000000001F0000
8000007F00008000000000000000000000000000000000000000003F00003F00008F0000900000\
8F0000700000700000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F4F60AF90C0AF90BFB09FC0\
AF90C0B09FC0B09FC0B09FC0B09FC0BF9FC0B09FC0BFA0CFBFA0C0BFA0CFBFA0C0
BFA0CFBFA0C0C0A0CFBF202FC0202FC0809FC08FA0C0A0C0C0AFCFC0AFC0CFAFCFC0AFC0CFAFCF\
CFAFC0CFAFC0BFA0B0C0A0BFCF0000D00000CF0000D00000D00000D00000C0A0B0
CFAFBFD0B0C0DFB0C0D0B0C0DFBFC0DFB0C0DFBFC0DFB0C0DFBFC0DFBFC0E0BFC0DFBFC0E0BFC0\
E0BFC0E0C0C0E0BFC0E0C0C0E0BFC0EFC0C0E0BFC0EFC0C0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F00006000005F00005F0000000000000000000000000000000000
000000000000000000000000000000000000E00000DF0000E00000DF0000E00000B00000B00000\
0000000000000000000000002F00003000005F0000000000000000000000000000
00000000000070604F6F604FD0705FD00000D00000CF0000D02F20CFBF80CFBF8FCFB080CFBF80\
C0B080CFBF80C0B07FC0BF7FC0B07FC02F1FBF201FC0B07FBFAF70BFB070BFAF6F
BFB070BFAF6FBFAF6FB0AF6FBFAF6FB0AF60B0AF60AF0000AF0000A040209F9050A09050A05F3F\
9F00009F0000000000000000000000404020403F20404020403F1F807F3F6F6030
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000303040000000\
201F2F3000003F30503F305F403F603F305F4000004000004F3F704F3F704F4070
4F3F7050407F50407050407F502F4F5F00005F00005F101F5F4F8060508F5F4F7F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
7F00008000007F00001F00000000000000000000000000000000003F00003000008F00008F0000\
900000700000700000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F4F60AF90BFAF90C0AF90BF\
B09FC0AF90BFB09FC0B09FC0B09FC0B09FC0B09FC0B09FC0BFA0CFB09FC0BFA0C0
BFA0C0BFA0CFBFA0C0BF80A0BF809FC0202FBF2020C0AFC0C0A0C0C0AFC0C0A0C0C0AFC0C0AFC0\
CFAFC0C0AFC0CF6F7FCF6070CF0000CF0000CF0000CF0000D00000CF0000D0B0C0
D0B0C0D0B0C0D0B0C0D0B0C0D0B0C0DFBFC0D0B0C0DFBFC0DFB0C0DFBFC0DFB0C0DFBFC0DFBFC0\
E0BFC0DFBFC0E0BFC0E0BFC0E0C0C0E0BFBFE0C0C0E0BFC0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F0F00008F0000800000000000000000000000000000000000
0000000000000000000000002F0000300000DF0000E00000DF0000E00000DF00005F00005F0000\
0000000000000000000000000000000000008000002F0000000000000000000000
0000000000006F604F6F604FD07050D00000CF0000CF0000CF201FCFBF8FCFB080CFBF80C0B080\
CFBF80C0B07FC0B07FC0B07FC0B07FBF6F40C06F4FBF6F40C0B070BFAF6FBFAF70
BFAF6FBFAF6FB0AF6FBFAF6FB0A060B0AF6FAFA060AF0000AF0000AF0000A09F50A09F5F9F9050\
9F90503F00009000008F00009000008F00000000000000000000000000006F6F3F
2F2F1F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F0000\
3000003F3F503F3050403F5F403F5F403F604000004F00004F3F6F4F407F4F4070
50407F50407F504F7F503F605F3F6F5F00005F00005F000060508F605080605080000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6F00007F00008000001F00000000000000000000000000000000003F00003F00008F0000900000\
8F00009000009000000F0F0F000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F4F60504F60AF9FC0AF90BFB09FC0\
AF90C0B09FC0B09FC0B09FC0B09FC0BFA0CFB09FC0BFA0CFBF9FC0BFA0CFBFA0C0
BFA0CFBFA0C0C0A0CFBFA0C0C0AFCFBF202FC0202FB09FBFBFA0BFC0A0C0CFAFCFC0AFC0CFAFCF\
C0AFC0CFB0CFCF6F70CF6F7FCF0000D00000CF0000D00000CF202FD02F2FD0B0C0
D0B0C0D0B0C0DFBFCFD0B0C0DFBFC0DFB0C0DFBFCFDFBFC0DFBFC0DFBFC0E0BFC0DFBFC0E0BFC0\
E0BFC0E0C0C0E0BFC0E0C0C0E0BFC0EFC0C0E0C0C0EFC0C0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000B00000BF00000F0F0F0F0F0F000000000000000000
0000000000000000000000005F00005F0000E00000DF0000E00000DF0000E00000DF00005F0000\
0000000000000000000000000000000000002F0000800000000000000000000000
000000000000706F506F604FD09F70D00000D00000CF0000D00000CFBF80CFBF8FCFBF80CFBF80\
C0B080CFBF80C0B07FC0BF7FC0B07FBFB07FBFAF70C02F1FB0AF6FC0B070BFAF70
BFB070BFAF6FBFB070B0AF6FBFAF6FB0AF60B0AF6FAF2010AF0000A05F3FAFA05FA09F5FA09F5F\
908F4F0F0F0F3F00009000008F00003F00000000002000000000000000006F6F3F
5F5F3F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000300000300000\
3F304F3F30503F3F5F3F305F403F60403F6040000040203F4F3F704F3F7050407F
4F3F7050407F50407F50407F50407F5F00005F00005F00005F4F7F60508F5F4F7F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6000008000008000005000000000000000000000000000000000001F00001F00009000008F0000\
9000009000009000003F0000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000504F5F5F4F60AF90BFAF90C0AF90BF\
B09FC0AF90C0B09FC0B09FC0B09FC0B09FC0BFA0CFB09FC0BFA0CFBF9FC0BFA0CF
BF9FC0BFA0CFBFA0C0C0A0CFBFA0C0C080A0C0809FC04050C0404FC0AFC0C0A0C0CFAFCFC0AFC0\
CFAFC0CFAFC0CF202FCF2020CF0000CF0000D00000CF0000D04F50D0404FD0B0C0
D0B0C0D0B0C0D0B0C0DFBFC0D0B0C0DFBFC0DFB0C0DFBFC0DFB0C0DFBFC0DFBFC0E0BFC0DFBFC0\
E0BFC0E0BFC0E0C0C0E0BFC0E0C0C0E0BFC0EFC0C0E0BFC0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F00005F00005F00005F0000000000000000000000
0000000000000000000000008000008F0000DF0000E00000DF0000E00000DF0000E000002F0000\
000000000000000000000000000000000000000000DF00002F0000000000000000
0000000000000000006F604FC0B08FD00000CF0000D00000CF0000BFAF80BFAF7FCFBF80C0B080\
CFBF80C0B07FC0BF80C0B07FC0B07FC0B070C0B07FBF6F40BF201FBF2010BFB070
BFAF6FBFAF6FB0AF6FBFAF6FB0A060B0AF6FB0804FB00000AF00009F905FA09F5FA09F5F9F9050\
9090500000002000008F000070000000000000000000000000000000000070703F
5F5030000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000001F0000\
30304F3F3F503F3050403F60403F60403F604000004F305F4F3F704F407F4F3F70
50407F50407F504F7F50407F5F00005F00005F00005F000060407060508F605080000000000000\
0000000000000000000000000000000000000000000000000000000000004F0000
5000008000008000006F0000000000000000000000000000000000000000000000500000900000\
9000009000009000005F0000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F50605F4F60AF90C0AF90BFB09FC0\
B09FC0B09FC0B09FC0B09FCFB09FC0BFA0CFB09FC0BFA0CFBFA0C0BFA0CFBFA0C0
C0A0CFBFA0C0C0A0CFBFA0C0C0AFCFC0A0C0C0AFCFC0202FCF2F2FBF9FBFBFA0BFC0AFC0CFAFCF\
CFAFC0CFB0CFCF202FD02F2FCF0000D00000CF0000D00000D06F70D06F7FD0B0C0
DFBFCFD0B0C0DFBFCFD0B0C0DFBFCFDFB0C0DFBFC0DFBFC0E0BFCFDFBFC0E0BFC0DFBFC0E0C0CF\
E0BFC0E0C0C0E0BFC0EFC0C0E0C0C0EFC0C0E0C0C0EFC0C0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003000008F00008F0000000000000000000000
000000000000000000000000BF0000B00000E00000DF0000E00000DF0000E00000DF0000000000\
0000000000000000000000000000000000000000008F0000800000800000000000
0000000000000000006F604FD0C090D00000D00000CF0000D00000CF906FCF9F6FCFBF80CFBF80\
CFBF80CFBF80C0B07FCFBF80C0B07FC0B07FC0B070B0AF70BF0000C00000BF6F40
BF2F1FBF201FBF201FB02010BF2F1FB02010B06F40AF0000B00000AFA05FAFA05FA09F50A09F5F\
908F500F0F0F1F00009000008F0000000000000000000000000000000000000000
60603F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0000000000000000\
3F20303F30503F3F5F3F305F403F60403F60400000403F6F4F40704F3F704F407F
4F407050407F50407F504F7F500F1F5F00005F00005F00005F2F5060508F5F4F7F000000000000\
000000000000000000000000000000000000000000000000000000000000300000
7F00008000008000006F00000000000000000000000000000000000000000000003F00008F0000\
9000008F00009000009000000F0F0F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000504F605F4F60AF90BFAF90C0AF90BF\
B09FC0AF90C0B09FC0B09FC0B09FC0B09FC0BFA0CFB09FC0BFA0CFB09FC0BFA0CF
BFA0C0BFA0CFBFA0C0C0A0CFBFA0C0C0AFCFBFA0C0C0607FC06070C06F7FC06070CFAFCFC0AFC0\
C0AFC0C0A0BFCF0000CF0000CF0000CF0000D00000CF0000D090A0CF8F9FD0B0C0
D0B0C0DFB0C0D0B0C0DFB0C0D0B0C0DFBFC0DFB0C0DFBFC0DFBFC0E0BFC0DFBFC0E0BFC0DFBFC0\
E0C0C0E0BFC0E0C0C0E0BFC0EFC0C0E0BFC0EFC0C0E0C0C0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000BF0000B00000000000000000000000
000000000000000000300000DF0000E00000DF0000E00000DF0000E00000B00000B00000000000\
000000000000000000000000000000000000000000000000300000D000002F0000
0000000000000000006F604FD0BF8FD00000CF0000D00000CF0000CF7050CF6F4FCFBF80CFB080\
CFBF80C0B07FC0BF7FC0B07FC0B07FBFB070C0B07FBFB070C04F30BF402FBF6F40
BFAF6FBF8F5FB08F50BF8F50B08050BF8F50B02010B00000AF0000AFA060A09F5FA09F5F9F9050\
9F9050000000000000900000900000000000000000000000000000000000700000
2F2F1F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F0F103F3F5F3F3F5F403F60403F604F0000400F104F40704F3F7050407F4F407F
50407F50407F5F4F8050407F5F203F5F00006000005F000060203F60508F605080000000000000\
0000000000000000000000000000000000000000000000000000000000001F0000
800000800000800000800000000000000000000000000000000000000000000000000000700000\
7000009000009000009F00003F0000400000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F50605F4F60B09FC0AF90C0B09FC0\
AF9FC0B09FC0B09FC0B09FCFB07F9FBF80A0BF809FBFA0CFBFA0C0BFA0CFBFA0CF
C0A0CFBFA0C0C0AFCFC0A0CFC0AFCFC0A0C0C0AFCFC0AFC0CFAFCFC0202FCF2F2FC0A0C0C0AFC0\
CF8F9FCF8FA0CF0000D00000CF0000D00000D00000D00000D0B0C0D0B0CFD0B0C0
DFBFCFD0B0C0DFBFCFDFB0C0DFBFCFCFAFB0D0B0BFDF4F50E04F50DF202FE05050E04F50E0707F\
E07070E0C0CFE0BFC0EFC0C0E0C0C0EFC0CFEFC0C0EFC0C0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F00006000005F0000600000000000
0000000000000000005F00005F0000E00000E00000E00000E00000DF00008F0000800000000000\
0000000000000000000000000000000000000000000000000000008F0000800000
0000000000000000006F604FD0C090D0BF8FD02F20D00000D00000CF7050D07050CFBF80CFBF80\
CFBF80CFBF80C0B07FCFBF80C0B07FC0BF7FC0B07FC0B07FB0A06FB0AF70BF201F
BFB070BFAF6FBFB070BFAF6FBFAF6FB0AF60BF6F40B00000B00000AFA05FAFA060A09F5FA09F5F\
9F90500F0F0F0000007F00008F0000200000000000000000000000000000000000
1F0000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F20303F3050403F5F3F305F403F604000004F1F2F403F6F4F40704F3F7050407F
4F3F7050407F50407F5F4F80502F4F5F00005F00006000005F0F1F60508F5F4F80000000000000\
0000000000000000000000000000000000000000000000000000000000001F0000
7F00008000008000008000001F00000000000000000000000000000000000000000000003F0000\
3F00009000009000009000007F00007F0000000000000000000000000000000000
000000000000000000000000000000000000000000000000504F605F4F60AF90BFB09FC0AF90BF\
B09FC0B09FC0B09FC0B09FC0B04050B05F70BF607FB09FC0BFA0CFBFA0C0BFA0CF
BFA0C0BFA0CFBFA0C0C0AFCFBFA0C0C0AFCFC0A0C0C0AFCFC0A0C0C06F7FC06070CF6F7FC06070\
CF6F7FCF6070CF0000CF0000CF0000CF0000D0202FCF2020D0B0C0D0B0C0D0B0CF
D0B0C0D0B0C0D0B0C0DF90A0D0909FDF2F2FDF202FCFB0BFCFAFB0DFBFC0DFBFC0DFBFC0DF6F70\
E0707FE04F4FE05050E0BFC0E0C0C0E0C0C0EFC0C0E0C0C0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F00008F00008F0000000000
0000000000000000005F00005F0000E00000DF0000E00000DF0000E000005F00005F0000000000\
0000000000000000000000000000000000000000000000000000002F0000DF0000
2F00002F000000000070604FD0BF8FD0C090CF201FD00000CF0000CF0000CF201FCFBF80C0B080\
CFBF80C0B07FCFBF80C0B07FC0B07FC0B07FC0B07FBFB070C0B070BFAF70BF4F30
BF8F50BF8F5FBFAF6FBFAF6FB0AF60B0AF6FB0603FB00000AF0000A09F5FA09F5FA09F5F9F9050\
9F90500000000000003F00009000008F0000000000000000000000000000000000
000000505030000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000020202F\
3F0F103F3F5F3F305F40101F400F104F00004020404F40704F3F7050407F4F4070
50407F50407F5F4F8050407F5F30505F00006000005F000060000060508F605080000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8000008000008000008000003F0000000000000000000000000000000000000000000000000000\
0000007F00009000009F00009000009F00003F0000400000000000000000000000
0000000000000000000000000000000000000000000000005F506F5F4F60B09FC0AF90C0B09FC0\
B09FC0B09FC0B09FC0BFA0CFB0202FBF202FB0202FBFA0CFBFA0CFBFA0CFBFA0CF
C0AFCFBFA0CFC0AFCFBFA0CFC0AFCFC0AFCFC0AFCFC0AFC0CFAFCFBFA0C0C0AFC0C0202FCF2F2F\
CF4050CF4F50CF0000D00000CF0000D00000CF4F50D04F50D0B0C0D0B0CFD0B0C0
DFBFCFCFAFBFD0B0C0DF202FDF2F2FD0B0BFD0B0C0DF202FE02F2FDFBFC0E0BFCFE07070E0707F\
E07070E05050E04F4FEF7F7FE07070EFC0CFEFC0C0EFCFCFEFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000B00000BF0000000000
000000000000000000B00000BF0000E00000E00000E00000E00000E000005F00005F0000000000\
000000000000000000000000000000000000000000000000000000000000800000
8000008000000000006F604FDFC090D0C08FD0705FCF0000D00000CF0000D00000CFBF80CFBF8F\
CFBF80CFBF80C0B07FCFBF80C0B07FC0BF7FC0B07FC0B07FC0B070C0B070BF8F5F
C02F1FBF201FBFB070BFAF6FBFB06FB0AF6FBF4F2FB00000B0201FAFA05FAFA05FA09F5FA09F5F\
9F90509F9050000000000000900000900000000000000000000000000000000000
000000500000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000001F1F20\
3F2F403F0F10402F4F3F2F4F4030504000004F305F403F6F4F40704F3F7050407F
4F407F50407F50407F5F4F80503F605F00005F00006000005F000060508F5F4F80000000000000\
000000000000000000000000000000000000000000000000000000000000000000
600000800000800000800000300000000000000000000000000000000000000000000000000000\
0000001F00009000009000009F00009000009F00000000000F0F0F000000000000
000000000000000000000000000000000000000000000000504F605F5060AF90BFAF9FC0AF90C0\
B09FC0AF9FC0AF90BFAF90BFBF0000B01F20BF202FBFA0C0BFA0CFBFA0C0BFA0CF
BFA0C0C0A0CFBFA0C0C0A0CFC0A0C0C0AFCFC0A0C0C0AFCFC0AFC0CFAFCFC0AFC0CF6F7FCF6070\
CF0000CF0000CF0000CF0000D00000CF0000D06F7FD06F70D0B0CFD0B0C0D0B0CF
D0B0C0DF707FD06F70DF707FDF6F70DFBFCFDFB0C0DF2F2FDF202FDFB0C0D0B0BFE04F50E0BFC0\
E0C0CFD0B0B0D0B0BFE04F4FEF5050E0BFC0EFC0C0EFC0C0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000008F00008000005F0000
5F0000000000000000E00000E00000E00000DF0000E00000DF0000E000002F0000300000000000\
0000000000000000000000000000000000000000000000000000000000002F0000
D00000DF00002F00006F604FD0C08FD0C090CF7050D00000CF0000D00000CF0000CFBF80CFBF80\
CFBF80C0B07FCFBF80C0B07FC0BF7FC0B07FC0B07FBFB070C0B07FBFAF70BFB070
BF2010BF201FB08F50BFAF6FB0AF6FBFAF6FB00000B00000AF603FAFA060A09F5FA09F5FA09F50\
A09F50908F4F0000000000009000008F0000500000000000000000000000000000
0000002F2F20000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000020202F\
3F3050401F2F3F3F5F403F604000004F00004F3F6F4F40704F3F7050407F50407F
50407F50407F5F4F805F4F805F4F805F00006000005F000060407060406F60508F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5000008000008F0000800000500000000000000000000000000000000000000000000000000000\
0000000000003F00004000009000009F00009F00006000005F0000400000400000
0000000000000000000000000000000000000000000000005F50605F4F60B09FC0AF90C0B09FC0\
B09FC0B09FCFB05F7FBF607FB00000BF202FBF202FBFA0CFBFA0CFC0A0CFBFA0CF
C0AFCFBFA0CFC0AFCFC0A0CFC0AFCFC0A0CFCFAFCFC0AFCFCFAFCFC0AFCFCFB0CFC0A0C0C0AFC0\
CF0000D00000CF0000D00000CF0000D00000D06F7FD06F7FD0B0C0DFBFCFD0B0C0
DFBFCFD06F70DF707FDF6F7FDF707FDFBFC0E0BFCFDF202FE02F2FDF7070E0707FE04F50E05050\
E07070EFC0CFE0C0C0EF2F2FE02F2FEFC0CFEFC0C0EFC0CFEFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006000008F0000
8F00002F0000300000E00000E00000E00000E00000E00000B00000B00000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000800000B000006F604F706F4FD0C08FD07F5FD00000D00000CF0000D00000CF906FCF9F6F\
CFBF80CFBF80CFBF80CFBF80C0B07FCFBF7FC0B07FC0B07FBFB070C0B07FBFB070
C0B070BF8F5FBF4F30BFAF6FBFB06FB0AF6FAFA060B00000B0603FAFA05FAFA060A09F5FA09F5F\
9F90509F9050000000000000000000900000700000000000000000000000000000
700000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000001F1F20\
3F3F503F1020403F5F3F305F400000400F104F3F70403F6F4F40704F3F7050407F
4F407F50407F50407F5F4F80504F805F101F5F00006000005F2F505F50805F4F80000000000000\
0000000000000000000000000000001000001F0000000000000000000000000000
3000008000008000008F00006F0000000000000000000000000000000000000000000000000000\
0000000000000000000000005F00005F00009F00009F00009F00009F0000A00000
1F0000000000000000000000000000000000000000000000504F605F4F60AF90C0B09FC0AF90C0\
B07F9FB07F9FB00000B00000BF0000B00000BF0000B09FC0BFA0CFBFA0CFBFA0CF
BFA0C0C0A0CFBFA0CFC0AFCFBFA0C0C0AFCFC0A0C0C0AFCFC0AFC0CFAFCFC0AFC0CFAFCFC0AFC0\
CF6F7FCF6F70CF0000CF0000D00000CF0000D090A0CF8F9FD0B0CFD0B0C0D0BFCF
D0B0C0DF90A0D0909FDF4F50DF4F4FDFBFCFDFBFC0E02F2FDF2020E0BFCFDFBFC0E09FA0E0909F\
E0707FE09F9FEF9FA0E04F4FEF5050E0C0C0EFC0CFEFC0C0EFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000E00000
E000005F00005F0000E00000E00000E00000E00000E000008000008F0000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F0000D0000080503F7F4F3FD0C090D09F70D00000CF0000D00000CF0000CF9F6FCF906F\
CFBF80C0BF80CFBF80C0B07FC0BF7FC0B07FC0B07FBFB070C0B070BFB070C0B070
BFAF6FBFB070BF2010B0A06FAF9F60BFAF6FB0603FB00000AF0000AFA060A09F5FAF9F5FA09F50\
A09F50908F4F0F0F000000000000008F00008F0000000000000000000000000000
6F0000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F3F50\
3F3050401F2F3F3F5F403F604000004F1F2F403F6F4F40704F3F7050407F4F407F
504F8050407F5F4F80504F805F4F805F20306000005F000060305F5F4F8060508F000000000000\
0000000000000000000000001F00007F0000000000000000000000000000000000
2000008000008F00008000008F00001F0000200000000000000000000000000000000000000000\
0000000000000000000000000000000000005F0000A000005F0000A000009F0000
A000004000004000000000000000000000000000000000005F506F5F4F60B09FC0AF90C0B04050\
B00000B00000B00000BF0000B04050BF4050BF4050BFA0C0BFA0CFBFA0CFBFA0CF
C0AFCFC0A0CFC0AFCFC0A0CFC0AFCFC0AFCFCFAFCFC0AFCFCFAFCFCFAFCFCFB0CFCFAFCFCFB0CF\
CFAFCFD0B0CFCF202FD02F2FD00000D00000D0B0C0D0BFCFD0B0CFDFBFCFD0B0C0
DFBFCFDFB0CFDFBFCFDF202FDF2F2FDF90A0E09FA0DF4F50E05050E02F2FE02F2FE04F50E05050\
DFBFC0E0BFC0E02F2FEFA0A0EF9F9FEFC0CFEFC0C0EFCFCFEFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000008F0000
8F0000B00000BF0000E00000E00000E00000E00000E00000E000005F0000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000800000AF2F1FAF2F20D0C090D0C090D00000D00000CF0000D00000CF7050D07050\
CFBF80CFBF80C0BF7FCFBF80C0BF7FCFBF7FC0B07FC0BF7FC0B070C0B07FBFB070
C0B070BFB070BF8F5FBF402FBF4F30BFAF6FBF6F40B00000B00000AFA060AFA060A09F5FA09F5F\
A09F509F90500000000000000000009000008F0000200000000000000000000000
0000000000004F4F3F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000030304F\
3F3F503F0F10403F5F403F60400000402F504F3F704F3F704F407F4F3F7050407F
50407F504F8050407F5F4F805F4F805F30505F0000600000602030605080000000000000000000\
0000000000000000000000004F00005F0000000000000000000000000000000000
1F0000800000800000800000800000500000500000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F0000200000
7F0000A00000A00000800000800000400000200000200000603F4F6F3F50AF709FB03F50AF3F4F\
B0607FB05F70B09FC0AF90BFBFA0CFB09FC0BFA0CFBFA0CFBFA0CFBFA0CFBFA0CF
BFA0CFC0AFCFBFA0C0C0AFCFC0A0CFC0AFCFC0A0C0C0AFCFC0AFCFCFAFCFC0AFC0CFAFCFCFAFCF\
CFB0CFCFAFC0CF6F7FCF6F70D06F7FCF6F70D0B0CFD0B0C0D0B0CFD0B0C0D0B0CF
D0B0C0DFBFCFD0B0C0D0B0C0D0B0BFDF2F2FDF2020DF4F50DF4F4FE09FA0DF909FE0707FE07070\
E02F2FE02020E0707FE0C0C0EFC0CFE0C0C0EFC0CFEFC0C0EFCFCF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F0000E00000E00000E00000E00000E00000E00000E00000E00000300000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F00002F0000D00000D09F70D0BF8FD0C08FCF0000D00000CF0000CF0000CF704F\
CFBF80C0BF7FCFBF80C0B07FCFBF7FC0B07FC0B07FC0B070C0B07FBFB070C0B070
BFB070BFB070BFAF6FBF201FB02010AFA060B02010B00000AF2010AFA060A09F5FAF9F5FA09F50\
A09F5F908F4F0F0F000000000000005000008F0000300000000000000000000000
000000000000403F2F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F00000000000000003F3F50\
3F305040203F403F5F403F604000004F3F6F4F3F704F40704F3F7050407F50407F
504F8050407F5F4F805F4F805F4F805F2F506000006000006010205F4F80000000000000000000\
0000000000000000001F00007000004F0000000000000000000000000000000000
3F00008000008F00008000008F00008F00008F00007000007000007000005F00005F00000F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
0000002000002000004000004000006000008F00008F00008F202F8F1F2FB0202FB09FC0B09FCF\
B09FC0BFA0CFB09FC0BFA0CFB09FCFBFA0CFBFA0CFBFA0CFBFA0CFC0AFCFBFA0CF
C0AFCFBFA0CFC0AFCFC0AFCFC0AFCFC0AFCFCFAFCFC0AFCFCFAFCFC0AFCFCFB0CFCFAFCFCFB0CF\
CFAFCFD0B0CFCFB0CFD0B0CFCFB0CFD0B0CFD0B0CFD0B0CFD0B0CFDFBFCFD0B0CF
DFBFCFD0B0CFDFBFCFDFBFCFDFBFCFDFBFC0E0C0CFDF90A0E09FA0E0707FE0707FE0707FE0707F\
DFBFC0EFC0CFE0C0C0EFC0CFEFC0C0EFCFCFEFC0C0EFCFCFEFC0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
300000E00000EF0000E00000E00000E00000E00000E00000E000002F0000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000AF2F20D02F1FD0C090D0C08FD02F20CF0000D00000CF0000CF503F\
CFBF80CFBF80CFBF80CFBF80C0B07FCFBF80C0B07FC0BF7FC0B070C0BF7FBFB070
C0B070BFB070C0B070BFAF6FBF6F40B0402FBF0000B00000B06F40AFA060AFA060A09F5FA09F5F\
A09F50A09F5F0000000000000000003F00008F00003F0000000000000000000000
000000000000500000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000030304F\
3F3F5F3F305F40203F400000400F1F403F6F4F4070403F6F4F407F4F3F7F50407F
50407F504F8050407F5F4F805F4F805F406F5F00006000006000005F4F80000000000000000000\
0000000000001F00007000007000002F00000000000000006000006000006F0000
7F00008000008000001F00001F00001F00001F00003F00003000003F00003F00003F0000000000\
0F000F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000504F605F5060AF90C0B09FC0AF90C0\
B09FC0B09FC0B0A0CFB09FC0BFA0CFB09FCFBFA0CFBFA0CFBFA0CFBFA0CFC0AFCF
BFA0CFC0AFCFC0A0CFC0AFCFC0A0CFC0AFCFC0AFCFCFAFCFC0AFCFCFAFCFCFAFCFCFB0CFCFAFC0\
CFB0CFCFAFCFD0B0CFCFAFC0D0B0CFD0B0CFD0B0CFD0B0C0D0B0CFD0B0C0DFBFCF
D0B0C0DFBFCFDFB0C0DFBFCFDFB0C0DFBFCFDFBFC0E0BFCFDFBFC0E0BFCFE0BFC0E0C0CFE0BFC0\
E0C0CFE0C0C0EFC0CFE0C0C0EFC0CFEFC0C0EFC0CFEFC0C0EFCFCF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F00003000002F00003000005F0000000000000000000000
800000E00000E00000E00000E00000E00000E00000E00000E00000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007F4F3FD00000D09F70D09F70CF201FD00000CF0000D00000CF201F\
CFBF80CFBF80CFBF80C0B07FCFBF7FC0B07FC0BF7FC0B070C0B07FBFB070C0B070
BFAF70BFB070BFAF6FBFB070B0AF6FBF201FB00000B00000AF804FAFA060AFA05FAFA05F909050\
A09F5F9F90500F0F000000000000001F0000900000500000000000000000000000
0000000000004F0000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F00000000000000003F3F50\
3F3F5040203F403F5F400000400F104F40704F3F70501F304F2F4F50407F50407F
5F4F805040805F4F805F4F805F508F504F7F600000600000600000603050000000000000000000\
0000005F0000700000700000700000300000000000300000300000300000300000
2000001F0000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F506F5F4F60B09FCFAF9FC0B09FCF\
B09FC0B0A0CFB09FC0BFA0CFBFA0CFBFA0CFBFA0CFC0AFD0BFA0CFC0AFCFBFA0CF
C0AFD0C0A0CFC0AFCFC0AFCFCFAFCFC0AFCFCFAFCFC0AFCFCFB0CFCFAFCFCFB0CFCFAFCFD0B0CF\
CFB0CFD0B0CFCFB0CFD0B0CFD0B0CFD0B0CFD0B0CFDFBFCFD0B0CFDFBFCFD0B0CF
DFBFCFDFBFCFDFBFCFDFBFCFE0BFCFDFBFCFE0BFCFDFBFCFE0C0CFE0BFCFE0C0CFE0BFC0EFC0CF\
E0C0CFEFC0CFE0C0C0EFC0CFEFC0CFEFCFCFEFC0C0F0CFCFEFC0CF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006000005F0000EF0000E00000EF0000E00000EF00008F0000
EF0000E00000EF0000E00000EF0000E00000E00000B00000BF0000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000706F50D0503FD0503FD02F1FD0503FD00000D00000CF0000D02F20\
CFBF80CFBF80CFBF80CFBF80C0BF7FCFBF80C0B07FC0BF7FC0B07FC0BF7FBFB070
C0B070BFB070BFB070BFAF6FBFB070BF6F40BF0000B00000B00000AFA060AFA060A09F5FAF804F\
9F9050A09F500000000000000000000F0F0F8F0000700000000000000000000000
0000002F0000505030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0000000000201F2F3F3050\
3F3F5F3F305F40203F4000004F2040403F6F4F3F70403F6F5000004F204F50407F
50407F5F4F8050407F5F4F805F4F805F4F805F00006000006000006F0000500000500000500000\
5F00001000001F00002F0000700000100000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F4F605F506FAF90C0B09FC0B09FC0\
B09FCFB09FC0B09FCFB09FCFBFA0CFB09FCFBFA0CFBFA0CFBFA0CFBFA0CFC0A0CF
BFA0CFC0AFCFBFA0CFC0AFCFC0AFCFC0AFCFC0AFCFCFAFCFC0AFCFCFB0CFC0AFCFCFB0CFCFAFCF\
CFB0CFCFAFCFD0B0CFCFB0CFD0B0CFCFB0C0D0B0CFD0B0CFD0BFCFD0B0C0DFBFCF
D0B0CFDFBFCFD0B0C0DFBFCFDFBFCFDFBFCFDFBFC0E0BFCFDFBFC0E0C0CFE0BFC0E0C0CFE0BFC0\
EFC0CFE0BFC0EFC0CFE0C0C0EFC0CFEFC0C0EFC0CFEFC0C0F0CFCF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000EF0000E00000E00000E00000E00000E000008F0000800000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F604FCFBF8FC0B080D00000D0201FD00000CF0000CF0000CF0000\
CFBF80CFBF80CFBF80C0BF7FCFBF80C0B07FC0BF7FC0B07FC0BF7FBFB070C0B070
BFB070C0B070BFAF6FBFB06FBFAF6FBFB06FB0AF60B00000B00000B0AF60AFA05FAFA05FA09F5F\
A0201090904F9090500000000000000000009000008F0000000000000000000000
0000006F000050100F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000001F1F2F3F3F50\
3F305F403F60400F104000004020404F40704F3F704F1F304F0000500000501F30
500000500F1F5F10205F101F5F10205F101F6000006040706040706040701F00001000001F0000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F506F5F4F60B09FCFB09FC0B09FCF\
B09FC0BFA0CFB09FCFBFA0CFBFA0CFBFA0D0BFA0CFBFAFD0BFA0CFC0AFD0BFA0CF
C0AFD0C0A0CFC0AFD0C0AFCFCFAFCFC0AFCFCFB0D0CFAFCFCFB0CFCFAFCFCFB0D0CFAFCFD0B0CF\
CFB0CFD0B0CFD0B0CFD0B0CFD0B0CFD0BFCFD0B0CFDFBFCFD0B0CF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000706F50D0C090DFCF90D04F3FD00000D00000D00000CF0000D00000\
CF201FCF9F6FCFBF80CFBF80C0BF7FCFBF80C0BF7FC0BF7FC0B070C0BF7FC0B070
C0B070BFB070C0B070BFB06FBFB070BFAF6FBFB06FB00000B06F40AFA060B0AF60AFA05FAFA05F\
A07F40A0201090904F0000000000000000007000008F00001F0000000000000000
0000006F0000303020000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000201F2F3F3050\
3F3F5F3F3F5F401F2F400000403F6F403F6F4F40704F3F7050101F4F0000502F4F
503060503F6F503F605F3F6F5F3F6F5F406F5F4F8F60508F60508F60508F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F3F306F6F6F7070706F6F6F707070706F6F7070706F6F6F707070706F6F707070
706F6F707070706F6F707070706F6F707070706F6F707070706F6F707070706F70707070706F6F\
707070706F70707070706F6F707070706F70707070706F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F303F6F60706F6F706F6F706F6F706F6070\
6F6F706F60706F6F706F606F6F6F706F60706F6F706F606F6F6F706F60706F6F70
60606F6F60706060706F6F70000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F604FD0C090D0C08FCFBF8FD00000D00000CF0000D00000CF0000\
CF0000CF0000CF7050C09060CF9060C09060CFBF7FC0B070C0BF7FC0B070C0B070
BFB070C0B070BFB06FBFB070BFAF6FBFB06FB0AF6FBF4F2FB0804FB0AF60AFA05FAFA05FA09F5F\
A09F5F9F00009F20100000000000000000007000008F00003F0000000000000000
0000002F0000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000001F1F2F403F5F\
3F3F5F403F60401F2F4F0000403F6F4F40704F3F7050407F4F2F4F50101F50407F
5F4F80504F805F4F805F4F8060508F5F4F8F60508F60508F6F5F905F5080000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F3F3F\
3F3030707070706F6F70707070706F707070706F6F707070707070707070706F6F
707070707070707070706F707F7070707070707070706F707F70707070707F7070706F707F7070\
7070707F7070706F707F70707070707F7070706F707F7070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303F706F706F6F70706F706F6F70706F70\
6F6F706F6F706F6F70706F706F6F706F6F706F60706F6F706F60706F6F706F6070
6F6F706F60706F6F706F60706F6F706F60706F6F706060706F6F706060706F607060606F6F6070\
6060706F607060606F0F0F0F000000000000000000000000000000000000000000
000000000000000000000000706F506F604FDFCF90D0C08FD0C090D0C08FD0C08FCFBF8FD0C08F\
CFBF80CFC080CFBF80CF0000CF0000CF0000C06F4FCFBF7FC0B07FC0BF7FC0B070
C0BF7FBFB070C0B070BFB06FBFB070BFB06FBFB06FB0AF6FBFAF6FB0AF60B0AF60AFA05FAFA05F\
A09F5FA0603F9F00009F00000000000000005000009000008F0000000000000000
000000000000000000404030000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000201F2F3F3050\
403F5F3F305F403F60400F104F3F6F403F6F4F40704F3F7050407F4F407F504080
50407F5F4F805040805F4F805F4F8060508F5F4F8F60508F60508F60508F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6000006F00006000006F00006F00006F00006F6F606F6F606F6F606F6F6F6F6F606F6F6F6F6F60\
3F3F3F303030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303030\
7070706F6F6F707070706F6F7070706F6F6F707070706F6F707070706F6F707070
706F6F707070706F6F707070706F6F707070706F6F707070706F70707070706F6F707070706F70\
707070706F6F707070706F707F7070706F70707070706F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303F6F6F70706F706F60706F6F706F6F70\
6F6F706F60706F6F706F60706F6F706F60706F6F706F60706F6F706F606F6F6F70
6F60706F6F7060606F6F60706060706F607060606F6F607060606F6F607060606F60607060606F\
606070605F6F606070605F6F000000000000000000000000000000000000000000
00000000000000000030303F302F3030303F2F2F305F5F6F5F5F6FD0C08FCFBF8FD0C08FCFBF80\
CFC08FCFBF80CFBF80CFBF7FCFBF80C0B07FC0BF7FC0B07FC0BF7FC0B070C0B070
BFB070C0B070BFB06FBFB070BFAF6FBFB06FB0AF6FBFAF6FB0AF60B0AF60AFA05FAFA05FA09F5F\
A09F5F9F90509F3F201F00002000007000007000008F00008F0000000000000000
0000000000000000004F100F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003000000000000000003F30503F3F5F\
3F3F5F403F60403F604F2F40403F6F4F40704F3F7050407F4F407F504F8050407F
5F4F80504F805F4F8F5F4F8060508F5F4F8F60509060508F6F5F905F5080000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F00006F00006F00002F00001F00006060606F6F6F6F6F606F6F6F6F00007000006F0000700000\
6F00007000006F6F6F3F3F3F3F3F303F3F3F3F3F3F3F3F3F3F3F30000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F3F3F\
706F6F70707070706F707070706F6F707070707070707070706F6F707070707070
707070706F70707070707070707070706F707F7070707070707070706F707F70707070707F7070\
706F707F70707070707F7070706F707F70707070707F7070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303F706F706F6F70706F706F6F70706F70\
6F6F70706F706F6F706F6F706F6F70706F706F6F706F6F706F60706F6F706F6070
6F6F706F60706F6F706F60706F6F706060706F6F706F60706F60706060706F6F706060706F6070\
60606F6F60706060706060700F000F00000000000000000030303F3F303F30303F
30303F30303F606070605F6F604F5F5F00006000005F00006000005F00005F3F405F5F6F5F5F6F\
5F506F5F5F6F5F506F5F5F6F505060CFBF80C0BF7FCFBF7FC0B07FC0BF7FC0B070
C0BF7FBFB070C0B070BFB070BFB070BFAF6FBFB06FBFAF6FBFAF6FB0AF60B0AF60AFA05FAFA05F\
A09F5FA09F5F9F90500F0F0F0000002000001F00002000001F00007000006F0000
6F00007000000000004F0000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003000002F00002000003F3F503F3050\
403F5F3F3F5F403F60403F604F3F6F4F3F704F407F4F3F7050407F50407F504F80
50407F5F4F805F4F805F4F8F5F4F8060508F60508F60509060508F60508F1000006000004F0000\
000000000000000000000000000000000000000000000000000000000000000000
5000006F00001000000000000000006F6F606F6F606F6F6F6F6F606F6F6F6F6F606F00006F0000\
6F00006F60607000006F2F2F706F6F6F6F6F70706F6F6F6F706F6F6F6F6F70706F
6F6F6F706F6F6F6F6F70706F6F6F6F3F3F3F3F3F303F3F3F3F30303F3F3F3F3F3F3F3F3F303030\
3F3F3F3030303F3F3F6F6F6F707070706F6F7070706F6F6F707070706F6F707070
706F6F707070706F6F707070706F6F707070706F70707070706F6F707070706F70707070706F70\
707070706F707F7070706F70707070706F707F7070706F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003F303F6F6F706F6F706F6F706F6F706F6070\
6F6F706F6F706F6F706F60706F6F706F60706F6F706F60706F6F706F60706F6F70
60606F6F6F706F60706F6F7060606F6F607030303030303F302F3030303F30303030303F302F30\
30303F302F300000000000005F5F6F605F6F606070605F6F606070605F6F606070
5F5F6F6000005F0000605F705F5F6F605F6F505060603F405F00005F00005F40505F5F6F505060\
5F5F6F50506F5F5F6F5050605F506F5050605F506F50303F40101F000000000000
0000000000004F0000500000BFAF6FBFB06FBFAF6FBFB06FB0AF60B0AF60AFA05FAFA060A09F5F\
A09F5FA09F50A09F500000000000000000000000000000000000001F00001F0000
1F00007F000000000050503F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F00000F00001F00003F3050403F5F\
3F3F5F403F60403F604F406F4F3F6F4F40704F3F7050407F50407F50408050407F
5F4F805F4F805F4F8F30302F000000000000000000000000000000000000400000600000500000\
000000000000000000000000000000000000000000000000000000000000000000
6F00006F00001F00000000000F0F0F6060606F6F6F6F6F6070706F6F6F60706F6F6F2F2F700000\
6F2F2F70706F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F0000
7000007000007000006F0000700000700000700000700000700000700000700000700000403F3F\
0000000000000000000000003F30303F3F3F3030303F3F3F3F30303F3F3F3F3030
3F3F3F3F303F3F3F3F3F30303F3F3F707070707070706F707F70707070707F70707070707F7070\
7070707F70707070707F70707070707F70707070707F707F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030303F706F7F6F6F70706F706F6F70706F7F\
6F6F70706F7030303F3F3F3F30303F3F303F30303F3F3F3F30303F3F303F30303F
3F303F30303F000000000000000000000000000000000000000000000000500000000000000000\
000000000000000000000000602F2F6000005F5F6F602F30600000600000600000
602F305F5F606060705F5F6F6060705F5F6F605F705F5F6F6000005F00005F3F405F506F5F5F6F\
5F506F5F5F6F5F506F5F5F6F50506F5F506F5050605F101F30202F000000000000
0000000000005000004F0000000000000000000000000000200000000000B0AF60AFA05FAFA060\
A09F5FA09F5F9F9F500F0F0F000000000000000000000000000000000000000000
0000001F00006F0000505030000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000003F3F503F3050\
403F5F3F3F5F403F60403F6F4F4070403F704F407F5F5F4F5F60505F4F405F0000
5F00006010105F10104F10100000000000000000000000000000000000004F00006000002F0000\
0000000000000000000000000000000000000000000000000000000000006F0000
6F00005000000000000000000000006F6F606F6F606F6F606F6F606F6F6F6F6F606060606F0000\
701F1F6F6F606F6F6F6F6F6F70706F6F6F60706F6F6F6F6F70706F6F6F6F706F6F
6F6F6F7040406F101070706F6F6F6F70706F6F50507040406F00007000006F0000700000700000\
1F00001000001F0000400000000000000000000000000000000000000000300000
100000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000003F3F3F3F3030403F3F3F3F3F3F3F3F
3F3F3F3F3F3F3F3F3F3F3F3F3F3F3F3F3F3F3F303F3F3F3F3F3F3F3F3F3F3F303F3F3F3F3F3F3F\
3F3F3F3F303F3F3F3F3F3F3F3F3F3F3F303F3F3F3F3F3F3F3F3F3F3F303F3F3F3F
3F303F3F3F3F3F303F706F706F6F70706F706F6F703F3F3F0000001F00002F00002F00002F0000\
2F00002F00002F00001000000F0F0F000000000000000000000000000000000000
3F00006F00006000006F00006000006F0000600000500000100000000000000000000000000000\
000000000000000000000000605F6F605F6F606070604F506000005F000060101F
605F6F6060705F5F6F605F6F5F5F6F605F6F5F5F6F5F5F6F5F30405F00005F10105F5F6F5F506F\
5F5F6F5050605F506F5050605F506F50506050101F4F4F5F2F2F3F000000000000
0000000000004F00005000000000000000000000001000001F00000000000000004F0000400000\
100000A09F50A09F50908F4F000000000000000000000000000000000000000000
00000000000060000050503F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000020202F3F3F50403F5F\
3F3F5F403F60403F604F406F5000005F5F505F5F4F5F60505F5F50603F305F0000
6000005F4F4060504F3F2F201F00004F00005000004F00005000004F00006F00006000001F0000\
0000000000000000000000000000000000000000000000000000003000006F0000
6F00004000000000000000000F0F0F606060706F6F6F6F606F6F6F6F6F6070706F6F0000700000\
6F0000705F5F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F6F6F707070706F6F
6F6F6F6F2F2F704F40706F6F707070706F6F707070706F6F7040407000007000007000003F3F3F\
000000000000000000000000000000000000000000000000000000300000700000
4F00000000000000000000000000000000000000000000000F0F0F1F00001F00007000007F0000\
7000007F00007000007F00001F00000000000000000000007070707F707F707070
7F7070706F707F707F7070707F404F7000007F0000700000700000700000700000700000700000\
700000705F60705F5F70707F706F7070707F706F70705F60700000700000700000
7000006F0000700000700000705F6060606F7030306F00007000006F0000000000000000000000\
0000000000000000003000006F00007000005000000F0F0F000000000000000000
0000000000000000001000006F00006F00006F0000100000000000000000000000000000000000\
0000000000000000000000005F5F6F6F6070606070603F4060000060000060101F
606070605F6F606070605F6F6060705F5F6F6060705F5F6F604F5F5F00006000005F4F505F5F6F\
5F506F5F5F6F5F506F5F5F6F50506F5F3F4050202F5F506F2F2F30000000000000
0000000000003000005000001000000000000000003F00000000000000000000004F00004F0000\
1F00000000000000002F0000000000000000000000000000000000000000000000
0000000000000000002F2F1F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000001F1F2F3F3F5F3F305F\
403F60403F60505F4F5000005F1010505F4F5F5F505F5F4F5F5F505F2020600000
5F3F306060505F5F503030300000000000000000001F00001000002F0000600000600000100000\
0000000000000000000000000000000000000000000000000000003F00006F0000
6000001F00000000000000000000006F6F606F6F606F6F6F6F6F606F6F6F6F6F606F40406F0000\
6F00006F403F70706F6F6F60706F6F6F6F6F70706F6F6F6F706F6F6F6F6F70706F
6F403F7040406F6F6F70706F6F6F6F70706F6F6F6F7070706F40407000006F0000701F1F3F3F30\
000000000000000000000000000000000000000000000000300000700000700000
5F0000000000000000000000000000000000000000000000000000000000000000300000700000\
300000000000000000000000000000000000000000000000707070706F707F7070
706F70707070706F70707070706F70707070705F5F70000070000070000070000070404F706F70\
707070706F707070706F6F70706F70706F707070706F6F70706F706F6F70707070
6F101070404F6F6F70706F706F3F407000006F00005F1F1F000000000000000000000000000000\
0000000000000000000000000000005000006F00006F00002F0000000000000000
0000000000000000002F00006000006F0000500000000000000000000000000000000000000000\
0000000000000000000000005F5F6F605F6F60607060202F600000600000603F40
5F5F6F6060705F5F6F6060705F5F6F605F6F5F5F6F605F6F5F5F6F5F101F5F00005F3F405F1010\
5F101F5F10105F101F5010105F101F5010105F506F5050602F2F3F000000000000
0000000000004000005000000000000000000000002000000000000000000000003F0000400000\
2F00000000002F00000000000000000000003F3F5F3F3F503F3F5F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000020202F3F3050505F4F\
505040505F4F50504F5F00005F302F5F5F505F5F4F5F60505F5F50601F105F0000
60504F5F605060605F000000000000000000000000000000000000600000600000500000000000\
0000000000000000000000000000000000000000000000000000006F00006F0000
6F00001000000000000000000F0F0F6F60606F6F6F6F6F6070706F6F6F6F70706F6F5050700000\
6F00007000006F6F6F70706F6F6F6F70706F6F6F6F70706F706F6F70706F6F4040
704F40706F6F7070706F6F6F707070706F6F707070706F6F701F1F700000700000702F2F403F3F\
0000000000000000000000000000000000000000000F0F0F700000700000700000
7F00000000000000000000000000000000000000000000000000000000000000004000004F0000\
0000000000000000000000000000000000000000000000007070707F7070707070
7F707F7070707F707F706F707F707F70707070707F6F6F6F7F0000700000700000705F5F70707F\
706F7070707F706F7070707F706F7070707F706F7070707F706F7070707F6F606F
701F1F706F7070707F6F6F707030306F2F30706F7F3F3F3F000000000000000000000000000000\
0000000000000000000000000000003000006F0000700000500000000000000000
0000000000000000004000006F00006F00002F0000000000000000000000000000000000000000\
0000000000000000000000005F5F6F60607060606F600000600000600000604F5F
606070605F6F606070605F6F606070605F6F605F705F5F6F605F705F3F405F00005F00005F4F5F\
5F40505F4F5F5F506F5F5F6F5F506F5F4F5F50506F5F5F6F2F2F30000000000000
0000000000005000005000000000000000004000000000000000000000000000000000004F0000\
4000000F0F0F0000000000000000000000003F3F5040405F3F3F503F3F5F303050
3F000030304F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F50404F0000501010505040\
505F4F5050405050405000005F3F30505F4F5F5F505F5F4F5F4F405F00005F0000
5F5F5060605F5F5F50000000000000000000000000000000000000600000600000500000000000\
0000000000000000000000000000000000000000000000000F0F0F6000006F0000
6F00000000000000000000000000006F6F606F6F606F6F6F6F6F606F6F6F6F6F606F6F6F6F0000\
7000006F0000701F1F6F1010701F1F6F1010701F1F6F101070706F6F6F60701F1F
6F6F6070706F6F6F6F70706F6F6F6F7070706F6F6F70706F6F10107000006F00007040403F3F3F\
0000000000000000000000000000000000000000007000004F0000700000700000
7000003000000000000000000000000000000000000000000000000000000000005F00002F0000\
0000000000000000000000000000000000000000000000007F7070706F70707070
706F707F7070706F70707070706F70707070706F70707070702F2F700000700000702F30706F70\
707070706F70707070706F707070706F6F70707070706F7070707F6F606F701F1F
6F6F707070706F6F706F606F6F10106F6F706F6F703F3F3F000000000000000000000000000000\
0000000000000000000000000000000000005F00006F00006F00002F0000000000
0000000000000000005000006F00006F0000100000000000000000000000000000000000000000\
000000000000000000000000605F6F605F6F605F6F600000600000600000606070
605F6F6060705F5F6F605F705F5F6F605F705F5F6F5F5F6F5F5F6F5F4F5F5F00005F00005F4050\
5F5F6F50506F5F5F6F50506F5050605050605F506F5050602F2F3F000000000000
0000000000004F00005000000000000000003F0000000000000000000000000000000000400000\
4F00000F00001F00000000000000000000000000003F00003F3F5F3030503F3F50
30000030000030203F000000404F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F3F50403F500000501010\
50100F5F5F4F50302F5F0000504F3F5F60505F5F505F60505F3F306000005F1010
60605F5F60503030300000000000000000000000000000005000006000006F00002F0000000000\
0000000000000000000000000000000000000000000000002F00006F00006F0000
4000000000000000000000000F0F0F60606070706F6F6F60706F6F6F6F6070706F6F6F6F701F1F\
6F00007000006F4040705F5F6F5F50705F5F6F5F507040406F40407000006F6F60
7070706F6F6F70707070706F707070706F6F70707070706F700000700000700000705F5F3F3F3F\
0000000000000000000000000000000F0F0F7000004F00002F0000700000700000
7000002F00000000000000000000000000000000000000000000000000003000004F0000000000\
0000000000000000000000000000000000000000000000007070707F707F707070
7F707F7070707F707F7070707F707F706F707F707F70707070404F7000007F0000701F1F70707F\
706F7070707F706F7070707F706F7070707F706F7070707F706F7070404F70404F
70707F706F7070707F706F7070707F6F6F7070707F3F3F3F000000000000000000000000000000\
0000000000000000000000000000000000005000006F00006F0000300000000000
0000000000000000006F00006F00006F0000000000000000000000000000000000000000000000\
0000000000000000000000005F5F6F6F6070603F40600000600000601F1F606070
606070605F6F606070605F6F6060705F5F6F6060705F5F6F605F705F5F6F601F1F5F00005F3F40\
5F506F5F5F6F5F506F5F506F5010105F5F6F50506F5F506F2F2F30000000000000
0000000000005000005000000000000000002000000000000000000000000000000000004F0000\
4000003F00003000000000000000000000000000004000003F2F403F3F5F3F0F10
3F2F40302F403030500000004F503F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F50404F50405050404F4030\
50403F50403F5F00005000005F5F4F5F5F4F5F5F505F5F4F5F3F305F0000601010
5F5F5060605030302F000000000000000000000000000000600000600000600000000000000000\
0000000000000000000000000000000000000000000000002F00006F00006F0000
2F00000000000000000000000000006F6F606F6F606F6F6F6F6F606F6F6F6F6F60706F6F6F403F\
6F00006F000070302F6F6F6F70706F6F6F6F70706F6F10106F6F606F6F6F70706F
6F6F6F70706F6F6F6F7070706F6F6F70706F6F6F6F705F5F6F0000700000700000705F5F3F3F3F\
0000000000000000000000000000002F00004F0000000000000000700000700000
7000005F00000000000000000000000000000000000000000000000000005F00000F0F0F000000\
0000000000000000000000000000000000000000000000007F7070706F707F7070
706F70707070706F70707070706F70707070706F70707070706F70700000700000700000704040\
707070706F7070707F706F70707070706F7070707F6F6F7070404F6F2F2F70707F
6F6F70706F706F6F7070707F6F6F70706F706F6F703F3F3F000000000000000000000000000000\
0000000000000000000000000000000000005F00006F00006F00002F0000000000
0000000000001000006F0000600000500000000000000000000000000000000000000000000000\
000000000000000000000000605F6F60606F602F30600000600000603F40606070
5F5F6F606070605F6F6060705F5F6F605F705F5F6F605F705F5F6F5F5F6F5F202F5F00005F1010\
5F5F6F5F506F5F5F6F50303F5F304050506F5F506F5050602F2F3F000000000000
000000000000500000500000000000400000000000000000000000000000000000000000300000\
4000004000002F00000000000000000000000000003F00003F00003F2F403F3040
30304F30305030304F30304F4F4F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F40505040505040505F4F\
50504F5F5F4F50403F5F4F405F4F3F5F4F405F3F306060505F10106000005F3F30
60605F6060503030300000000000000000000000000000006000006000006F0000000000000000\
0000000000000000000000000000000000000000000000005000006F00006F0000
3000000000000000000000006F6F606F6F606F6F6F6F6F6070706F6F6F6F70706F6F6F6F6F6F60\
6F00007000006F6F6F70706F6F6F6F70706F6F5F50703030706F6F70706F6F6F6F
70707070706F707070706F6F70707070706F70707070404070000070000070000070706F403F3F\
0000000000000000001F00004000004F0000000000000000000000300000700000
7F00007000000000000000000000000000000000000000000000001F00004F0000000000000000\
0000000000000000000000000000000000000000000000007070707F707F707070
7F707F7070707F707F7070707F707F7070707F707F7070707F707F70404F7000007000007F1F1F\
705F5F705F60705F5F705F60705F5F705F60705F5F703030702F3070707F706F70
70707F706F7070707F6F6F7070707F706F7070707F3F3F3F000000000000000000000000000000\
0000000000000000000000000000000000005000007000006F0000300000000000
0000000000002F00006F00006F0000400000000000000000000000000000000000000000000000\
0000000000000000000000005F5F6F6F607060101F6F000060000060505F605F6F
606070606070606070605F6F606070605F6F6060705F5F6F6060705F5F6F604F5F5F0000600000\
5F5F6F5F5F6F5F30405F3F405F506F5F5F6F50506F5F5F6F2F2F30000000000000
000000000000500000500000100000400000000000000000000000000000000000000000000000\
4000004F00002F000000000000000000000000000040405F3F30403F3F5F3F3050
3F3F5030305030305030304F30304F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005050404F5040505040505040\
505F4F505040505F4F505F4F5F5F5050504F5F4F405F4F405F00005F00005F1010
5F5F5060605F30302F0000000000000000000000003F0000600000600000000000000000000000\
4000000000000000000000000000000000000000000000006F00006000006F0000
0000000000000000000000006060606F6F6F6F6F606F6F6F6F6F606F6F6F6F6F606F6F6F6F6F6F\
7000006F00007040406F6F6F70706F6F5050701F1F6F6F6F70706F6F6F6F70706F
6F6F6F70706F6F6F6F70706F6F6F6F7070706F6F6F702F2F7000007000006F10107070703F3F3F\
0000000000000000004000004F00001000000000000000000000002F0000700000
7000007000002F00000000000000000000000000000000000000004F0000000000000000000000\
0000000000000000000000000000000000000000000000007F7070706F707F7070
7070707F7070706F70707070706F7070707F706F70707070706F7070404F700000700000700000\
70707F706F70707070706F7070707F6F6F7070707070101F6F6F706F6F70707070
6F6F7070707F6F6F70706F706F6F70706F7F6F6F703F3F3F000000000000000000000000000000\
0000000000000000000000000000000000006F00006F00006F0000000000000000
0000000000005000006F00006F00002F0000000000000000000000000000000000000000000000\
000000000000000000000000605F6F605F6F6000006000006000005F5F6F606070
605F6F6060705F5F6F6060705F5F6F6060705F5F6F605F6F5F5F6F605F705050605F101F5F0000\
5F3F405F506F5F101F5050605F5F6F5050605F506F50506F2F2F3F000000000000
000000000000400000500000100000200000000000000000000000000000000000000000000000\
4F00004000004000003000001F00000000000000003F3F5F3F3F5F3F30503F3F50
3030503F305030304F30304F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F5040505F4F505040505F4F\
5050405F5F4F505F4F5F5F505F5F4F5F60505F5F505F60505F4F4060504F5F4F40
60504F5F2020303F300000002F00004F0000000000600000600000600000000000000000000000\
6000005000000000000000000000000000000000001F00006F00006F0000500000
0000000000000000000000006F6F606F6F6070706F6F6F6F70706F6F6F6070706F6F6F6F70706F\
6F2F2F7000006F404070706F606060703030705F5070706F6F6F6F70707070706F
707070706F6F70707070706F707070706F6F701F1F701F1F700000700000701F1F7070703F3F3F\
0000000000004000004F00001F0000000000000000000000000000000000700000
7000007000006000000000000000000000000000003000006000005F0000000000000000000000\
0000000000000000000000000000000000000000000000007070707F707F707070
7F707F7070707F707F7070707F707F7070707F707F7070707F707F7070707F0000700000700000\
70404070707F70707070707F706F7070707F701F1F705F60706F7070707F706F70
70707F706F7070707F706F7070707F6F6F7070707F3F3F3F000000000000000000000000000000\
0000000000000000000000000000004000006F00006F0000400000000000000000
0000000000006F00006F00006F0000100000000000000000000000000000000000000000000000\
0000000000000000000000005F5F6F6F506060000060000060101F606070606070
606070605F6F606070605F706060705F5F6F6060705F5F6F605F705F5F6F605F705F101F5F0000\
5F00005F1F1F50506F5F5F6F5F506F5F5F6F5F506F5F5F6F2F2F30000000000000
000000000000500000500000500000000000000000000000000000000000000000000000000000\
1F000020000000000000000000000000000000000040405F3F3F5F3F3F5F3F3F50
3F3F50303050000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000505040\
505F4F50504F5F5F4F50504F5F5F4F5F5F4F5F60505F5F505F60505F5F50606050
5F5F5060605F30302F000000100000100000100000600000600000600000600000500000000000\
6F00006000004000000000000000000000000000002F00006F00006F0000500000
0000000000000000000F0F0F6060606F6F6F6060606F1F1F6F50506F6F6F6F6F6070706F6F6F60\
7000006F0000701F1F6F6F6F702F2F6F505070706F6F6F6F70706F6F6F6F70706F
6F6F6F70706F6F6F6F7070706F6F6F7070706F00007000006F00007000007040407070703F3F3F\
0000001000007000001000000000000000000000000000000000000000005F0000
7000007000007000000000000000000000000000004F00001000001F0000000000000000000000\
0000000000000000000000000000000000000000000000007F70707070707F707F
706F707070707070707F707F706F70707070706F7070707F706F70707070702F2F700000700000\
701F1F706F7070707F706F70707070702F2F705F606F6F7070707F706F7070707F
6F6F7070707F6F6F7070707F6F6F70706F706F6F703F3F40000000000000000000000000000000\
0000000000000000000000002F00006F00006F0000500000000000000000000000
0000001F00006F00006F0000500000000000000000000000000000000000000000000000000000\
000000000000000000000000605F6F603F40600000600000602F30605F6F606070
605F6F606070605F6F6060705F5F6F602F2F5F202F605F705F5F6F5F5F6F5F5F6F5F3F405F0000\
5F00005F00005F5F6F50506F5F5F6F50506F5F5F6F50506050506F000000000000
0000002000005000005000003F0000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F3F503F3F5F3F3F503F3F5F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000505F4F5F5F505F5F4F5F60505F5F506060505F605060605F5F6050
60605F6060503030300000000000000000000000000000000000001000001F0000100000000000\
3F00006F00006000000000000000000000000000005F00006F00006F0000000000
0000000000000000000000006F6F606F1F106F00006F2F2F70706F6F6F6F70706F6F6F6F6F6F6F\
6F00007000006F00007030306F404070706F6F6F6F70707070706F7070706F6F6F
70707070706F707070706F6F707070705F5F7000007000007000007040407070703F3F3F403F3F\
1F00007000001F00000000000000000000000000000000000000000000002F0000
7F00007000007F00002F00000000000000003000002F0000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007070707F707F707070
7F707F7070707F707F7070707F707F7070707F707F7070707F707F7070707F4F4F7000007F0000\
701F1F70707F706F7070707F6F6F6F701F1F706F7070707F706F7070707F706F70
70707F706F7070707F706F7070707F706F7070707F3F3F3F000000000000000000000000000000\
0000000000000000000000005F00007000006F0000300000000000000000000000
0000002F00006F00006F00005F0000000000000000000000000000100000000000000000000000\
0000000000000000000000005F5F6F6F1F1F6000006F0000603F40606070606070
606070606070606070605F6F603F4F6000005F5F6F5F5F6F6060705F5F6F605F705F4F5F600000\
5F00005F2F305F506F5F5F705F506F5F5F6F50506F5F5F6F50506F3F0000400000
4F0000500000100000200000100000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000040405F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F5F505F5F506060505F5F50606050
5F5F5060605F000000000000000000000000000000000000000000000000000000000000000000\
0F0F000000001F00001000001F00001000001F00006F00006F0000600000500000
5000005F00006F00002F00006F10106F00006F10106F6F6F6F6F606F706F6F6F606F6F6F6F6F6F\
7000006F00007000006F101070706F6F6F6F70706F6F6F6F70706F6F6F6F70706F
6F6F6F7070706F6F6F7070706F6F6F7040406F0000700000700000705F5F6F6F6F403030700000\
1F00000000000000000000000000000000000000000000000000000000001F0000
7000007000007000005F00000000000000004000001F0000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007F707F706F707F7070
7070707F404F702F2F70303070101F7F30306F606F707070706F7070707F706F70700000700000\
700000705F5F706F706F6F70701F1F706F7070707F706F7070707F6F6F7070707F
706F70705F606F505F70707F6F6F7070707F6F6F703F3F3F000000000000000000000000000000\
0000000000000000006F00007000006F00001F0000000000000000000000000000
0000004000006F00006F00002F00000000000000000000000000006F0000100000000000000000\
00000000000000000000000060606F60000060000060000060505F605F6F606070
605F6F6060705F5F6F603F405F0000604F5F5F5F6F605F705F5F6F605F705F5F6F5F5F6F5F1010\
5F00005050605F5F6F5F506F5F5F6F50506F5F506F50506F5F506F2F2F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F6050
60605F5F6050000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000100000
1F00001000001F00001F00006F302F6F000070706F6F6F6F70706F6F6F6F70706F6F6F6F70706F\
6F40407000006F1F1F70706F6F6F6F70707070706F7070706F6F6F70707070706F
707070706F6F70707070706F707070702F2F70000070000070000070706F7070705F1F1F4F0000\
000000000000000000000000000000000000000000000000000000000000000000
7000007000007F00007000000000002F00004F0000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007070707F707F707070
7F3030704040707070706F707F707F6F6F6F7F1F1F705F607F707F7070707F707F701F1F700000\
7000007F4F4F70404F70404F70404070707F70707070707F706F7070707F707070
70707F70101F701F1F706F7070707F706F7070707F3F3F3F000000000000000000000000000000\
0000000000004000007000006F0000000000000000000000000000000000000000
0000006F00006F00006F00001F00000000000000000000000000003F00006F0000100000000000\
000000000000000000000000603F406F00006000006F00005F5F6F6F6070606070
6060705F5F6F600000600000601F1F605F6F606070605F6F6060705F5F6F6060705F5F6F603F40\
5F5F6F5F5F705F5F6F5F5F6F5F506F5F5F6F5F506F5F5F6F50506F30303F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F6060606060606F6F606F6F6F6F6F606F6F6F6F6F6F70706F6F6F6F\
705F506F2F2F70706F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F6F6F707070
6F6F6F70706F6F6F6F7070706F6F6F701F1F7000007000006F10107070707010105F1F1F000000\
000000000000000000000000000000000000000000000000000000000000000000
4000007000007000007000002F0000700000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007F70707070707F3030
6F6F6F7F707F702F2F7F30306F6F70701F1F7070707F404F70404070707F706F7070404F700000\
70000070000070404F70404070707F706F7070707F706F7070707F706F7070707F
6F6F70701F1F6F00006F6F706F6F70706F7F6F6F703F3F40000000000000000000000000000000\
1F00006F0000700000400000000000000000000000000000000000000000000000
1000006F00006F00006F00002F00000000000000000000000000002F00006000006F0000600000\
5000005000005000004F0000600000600000600000604F5F60505F604F5F60505F
604F5F604F5F603F4060101F5F3F406060705F5F6F6060705F5F6F605F705F5F6F5F5F705F5F6F\
5F5F6F5F5F6F5F5F6F50506F5F5F6F50506F5F5F6F50506F5F506F2F2F30000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F6F6F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F6F6F70706F\
6F6F6F70706F6F6F6F70707070706F7070706F6F6F70707070706F70707070706F
70707070706F70707070706F707070700000700000700000703030701F1F703030402F2F000000\
000000000000000000000000000000000000000000000000000000000000000000
3000007000007F0000700000600000400000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007F70707F707F702F30
7F606070404F7F5F60705F5F7F707F701F1F7F707F702F307F60607070707F707F705F5F7F0000\
700000704F4F7070707F707F70707070707F706F7070707F70707F70707F706F70
70707F700000700000700000705F60706F7070707F3F3F3F000000000000000000000000300000\
5F00001F00000000000000000000000000000000000F0F0F4000005F0000500000
7000006F00006F00006F00006F00002F00002F00001000000000001000006F00002F00001F0000\
1000000000000000000000005F5F6F6F60706060706F6070606070606070606070
606070606070606070605F70606070605F706060705F5F6F6060705F5F6F605F705F5F6F605F70\
5F5F6F5F5F705F5F6F5F5F705F5F6F5F5F6F5F506F5F5F6F50506F30303F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006060606F6F6F6F6F606F6F6F6F6F6070706F6F6F606F706F6F6F6F\
70706F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F6F6F7070706F6F6F70706F
6F6F6F705F5F6F5F507040407010107000006F00007000007010107000006F6F603F3F3F000000\
000000000000000000000000000000000000000000000000000000000000000000
1000007000007000007000007000001F0000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007F707F7070707F5F60
70101F707070702F2F7F1F1F70101F7F3030705F5F701F1F70707070707F706F7070707F700000\
701F1F706F7070707F706F7070707F706F7070707F706F7070707F706F7070707F
6F6F7070707F6F6F706F6F706F40407030306F10105F1F1F5F00005F00004000002F0000100000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060606F606070606070606070606070605F6F606070
606070606070605F6F606070605F6F6060705F5F6F605F705F5F6F605F705F5F6F5F5F705F5F6F\
5F5F705F506F5F5F6F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000070706F6F6F6F70706F6F6F6F70706F6F6F6F70706F\
6F6F6F70707070706F7070706F6F6F70707070706F70707070706F70707070706F
704F407000007000007000007000007000007000007000007000006F6F6F7070703F3F3F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F00007F00007000007F0000700000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007070707F707F7F7070
7F707F706F707F4F4F70404F7F4F4F705F5F7F707F7070707F707F7070707F707F7070707F707F\
7070707F707F70707F70707F7070707F707F70707F70707F706F7070707F70707F
70707F706F7070707F706F7F70707F706F7070707F3F3F3F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F5F6F6F6F706060706F60706060706F6070606070
606070606070606070606070606070605F70606070000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F6F6F70706F6F6F6F70706F6F6F6F707070
6F6F6F7070706F6F6F7070706F6F6F70707070706F7070706F6F6F70707070706F3F3F3F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F00007000007000007000007000007000007F0000700000300000000000000000000000\
0000000000000000000000000000000000000000000000007F707F7070707F707F
7070707F707F7070707F707F70707070707F7070707F707F706F7070707F70707070707F706F70\
70707F70707070707F706F7070707F706F7070707F706F7070707F706F7070707F
6F6F7070707F706F7070707F6F6F7070707F6F6F703F3F403F303F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007F70707F707F707070
7F707F7070707F707F7070707F707F70707F7F707F7070707F707F70707F7F707F7070707F707F\
70707F7F707F7070707F707F70707F70707F70707070707F70707F70707F707070
70707F70707F70707F706F7070707F706F7F70707F3F3F3F3F3F40000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[4]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7H@0028]aRV^;Q9250h@00f/000020003J`1[00`018@@000Q
RZV?ZHZYRZV:ZHBMRZV:ZHnYSjV?ZHnSQ:V:ZHZYRZV:05d020L000H7GP003i2:RXB@RY2:T8^@K6/0
Q00500:400=[000020001F^MQ9f4010002:YT:V@ZI2YT:V@WHBYT:V@ZI2YT:V@ZHBSRZV?ZHnYSjV?
G0081`001PMK000?Rg64JhYaRg6;RX^:T8Ya00P00V/2Q009JhA[Q6^4JhA[00B400=[00003P0098nY
SjV:ZHZYRZV4XhZYRZV?ZHnYSjV?WHBMRZV:ZHZYRZ>4ZEX020L000H7F00018]aRg43Q00;RY2:T8Z@
Ri2;T6`02P0048A[Q01[06/0J`1[Q9f4WF/@000UZI2YT:V@ZI2YT:>:Xi2YT:V@ZI2YT:V4WHBSRZV?
ZHnYSjR4Z@1I00P700061eD001F;LH]aRg64JhYaRg6;LI1aRhZ@RW402`000hA[000700B400=[0000
3P009XnYSjV:ZHZYRZV:ZHBYRZV?ZHnYSjV:WHBMQ9f:ZHZYRZV:Z8AaF0081`001PMC0009LH][RW6;
JhY[00::00b@RY2:T8Z@Ri2;T6`;00:41`001F^4WHBM014002NYT:V@ZI2YT:V@ZI2SSjV@ZI2YT:V@
XhBMQ9f4XhnYSjV?ZHnSQ9d0E`081`001PM@000JLHA[Rg6;LH]aQ6^4LH]aRg6;LH]aT76@RW4=0003
Q00000<000=[08@00X@00f/0000@000ZZHZYSjV?ZHnYSjV:XhZYRZV:ZHZYQ9f4WHZSQ:V?ZHnYSjV:
WHBMQ9f4D`081`001PM?000:LHA[RW6;LH]aT0>400j@RY2;T8^@Ri2;T8^@K0h000>400000P0018BM
Q9d2Ja8002b@ZI2YT:V@ZI2YT:V4Xi2YT:V@ZI2MQ9f4ZHnYQ:V@ZHnYSjV?WHBMQ6^4WE4020L000H7
CP007764JhA[Rg6;LH]/Q6^4LH]aRg6;LH]aT76@074>009[0`04Q003J`00014002fYRZV?ZHnYSjV?
ZHnSQ:V:ZHZYRZR4WHBMSjV?XhBYSjV?ZHZYQ9f4J`1[08@0D0081`001PM=000=LHA[Q6^:LH]aRfb:
Q002RP0>T8^@Ri2;T8^@Ri2;LF`>009[0`0018BMQ9d50003J`2400:40f/7000YT:V@ZI2YT:V@ZI2Y
T:V@ZI2YT:V:WHBMRZV@ZHnSRZV?ZHnYShZ4WF/010000if4001=00P700061d/00W407F]/Q6^4LH]a
Rg6:JhA[Rg6;LH]aRg6;LI1aT01a00l000B406/00h@00f/0000300:41002J`L002NYRZV?ZHnYSjV?
ZHnYSjV:ZHZYRZ>4WHBSSjV?ZHZMRY2?ZHZYQ9d00X@400>4C@081`001PM:000PLW6;LH][Q6^;LH]a
Rfb:Q8Z;T8^@Ri2;T8^@Ri2;LF`40003J`0000P00f/2Q003WHB400@000B400240P000f/00007000Y
T:V@ZI2YT:V@ZI2YT:V@ZI2YT:V4WHBMRZV@ZI2YRJ>9ZHnYShZ4WH@010001HA[002404/020L000H7
B0000faaK004L@0IQF^4LH]aRg6:JhA[Rg6;LH]aRg6;LI1aT005009[2`000f^4Q003Q0<000=[06/0
0P0018@008@2J`L002NYRZV?ZHnYSjV?ZHnYSj>:ZHZYRYf4WHBYSjV?ZHn@Q9f:ZHZYJid00X@50003
J`0004`020L000H7B000479aLW6;LH]aQ6^;LH]aT6`3Q00>RY2;T8^@Ri2;T8^@Rg4400:430001XBM
Q9f4J`800f/20004Q6][009[1`00:I2YT:V@ZI2YT:V@ZI2YRJ>@ZHZSQ9f4WI2YT:V@ZI2SRJ>?ZHn:
Q9f400D000>40000C0081`001PM80003K75/00Aa01V4JhA/Rg6;LHY[Q6^5LH]aRg6;LH]aT76@00D0
00>4J`002`000f/0Q002Q0D000F400240002Q00308@000H002VYRZV?ZHnYSjV?ZHnYSif4ZHnSRIf4
WHBYSjV?ZHnYSjV4ZHZYJif4W@050003J`0000H000=[0000@`081`001PM5000CJ`00LW5bLH]aRf^4
JhYaRg6;L@04Q00=RhZ@Ri2;T8^@Ri2;L@030003JhB400h000>4WH@01@001HA[08@0009[00>4J`00
1P006I2YT:V@ZI2YT:V@ZI2YRJ>:ZHBSQ9f4WI000jT03Y2YT:V@XhnYShZ4WHA[1@000h@0001;00P7
00061d<000=[00000P008FabK79aRg65JhA[Rg6;LH]/Q6^4JhY[Rg6@LI2:T8Z@J`0200:400=[0000
5`000f/0000;000VZI2YSjV?ZHnYSjV?ZHnYSif:WHBMQ9f4ZHnYSjV?ZHnYSj>?ZF/2Q003WF/000D0
00=[00001P000h@0001100P700061d800V/4000CLW6;LH]aRfb4JhEaRg6;RX^5RP05Q00;T8^@Ri2;
T8]aJh@0:P00:Y2YT:V@ZI2YT:V@ZI2YT:V9XhBMQ9f4Xi2YSjV@ZHnYSjV?ZHmaQ9f4J`L000>40000
1@02Q48020L000H7@@03J`@0021/LVabLH]aRV^4JhYaRg6;LHAaQ6^4JhA/Rg6@RY2:R`9[:`00:ZV@
ZHnYSjV?ZHnYSjV?ZHnSQ9f4WHBSQ:V?ZHnYSjV?ZHnYRZU`JhBMJ`l000=[Q8@0@@081`001PLo0003
J`00009[0`00575bLH]aRg6;LHA[Q6b;LH^:RhZ@0XX5Q007RX^:Q8Z4L@0L000@LHYaRW6:LHYaRW6:
LHYaRP9a02J@ZI2YT:V@ZI2YT:V@ZHZMQ9f4WHZST:V@ZHnYSjV?ZHnSSg5[W@>41P000f^4Q00500B4
@0081`001PM2009[0`008W9/LVabLH]aRf^4JhYaRg6;LH]aRg6:JhY[Q6^:Q92:LF/<000PLF]aJg5/
LFaaK75/LFaaJjV:ZHZYRZV:ZI2YT:V@ZI02000TLF]aSjV?ZHnYSjV:ZHnSQ9f4WHBMRZV?ZHnYSjV?
ZHnYQ:U`0V/00if4W@070004W@24J`<00hA100P700061d400f/3000?LG9aRg6;LH][Q6^4Jh]a00>;
00R:T8Z@Ri2;T0>;00>@Ri000W4300Ua022;ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@Z@D0
00V:LHZ@ZI2YT:T00Z<01HBMWIf400:M01BYT:V@ZI2YSjV?ZHnSQ:QaJhBMQ0P000=[06/00X@0102M
Q6]100P700061d800V/3001>LVabK79aRg65JhA[Q76;LH]aRg6;LI1aT76@LI2:T8YaJg5[LHZ@RY2:
T8Z@RY2:T8Z@Ri2;T8^@Ri2;ZH^YRZV;ZHZYRjV@ZI2YT:V@2@008V]aJjV?ZHnYSif4WHBMSjV?ZHnY
SjV?ZHnYSi24L@2MQ9d80008J`00J`24Jh@30003Q00003`020L000H7?0000f/00003009[0`003G5b
LH]aRg6;K8A[Q6/01H/03XZ@RY2:T8^@Ri2;T8^@0W40;8^@Ri2;ZH^YRjV;ZI2YT:V@ZI2YT:V@ZI2Y
T:V@ZI2YT:V@ZI2YT:V@ZI2Y30000g6:L@07Z@0FWJV@ZI2YT:V@ZHnYSjV?WG40Q9f4J`L000>40000
0P000f^MQ0040003Q8X003/020L000H7@P02J`<004ibK79/Rg6;LHY[Q6^4Jh][RV^;LH]aT76@LI1a
T6]a076:T8Z@RY2:T8Z@RY2;T8Z@Ri2;T8^@Ri2;T8^YRjV;ZH^YRjV;ZI2YT:V@ZI0?000FL75`ZHnY
SjV?ZHnYSjV?ZHnYSjV?Z@:400D0WHBMJ`080003Q00000800X@50003Q8X003X020L000H7>`000f/0
0004009[0`003g6;LH]aRW6;K8A[Q6^4J`04RP0:T8Z@Ri2;T8]aK0<002iaT8^@Ri2;ZH^YRjV@ZI2Y
T:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2Y4P03L@0GRZVYT:V@ZI2YT:V?ZHnYSjV408BMQ6/0
20000h@000020003WH@000@00X@k00P700061c/000=[00001@006f/0001bK79aLV^;LHA[Q6^:LHYa
Rg6;LI1aT002L@03074000@002iaRY2:T8Z@RY2:T8^@Ri2;T8^@Ri2;T8^@Ri2;ZH^YRjV;ZH^YRjV@
ZI2YT:V@5P004W5`ZHnYSjV?ZHnYSjV?ZHn400:400>MJ`001`000f/0000200:41@000hn4Q00j00P7
00061cT00V/700=[01L0LW5bLH][Rfb:LH]aT76@LI2;T6aaK00;000^LI2;T8^YRjV@ZI2YT:V@ZI2Y
T:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZB8000QaZHnYLIf4W@:42@001V/006^MQ0D000B:Q8B?
>@081`001PM30004J`00K0=a00=/LG400W401h]aRg6;LH/00W4?000^LHZ@RY2:T8^@Ri2;T8^@Ri2;
T8^@RY2;ZHZYRi2;ZI2YRjV@ZI2YRjV@ZI2YT2<000U[ZHmaJif4WH@02P000h@0J`02Q0D000F?Q8A[
S`0h00P700061cX000=[00001@03J`0B079aLW64LH]aRg6;LI1aT6aa4000;W6@Ri2;ZH^YT:V@ZI2Y
T:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:TT0003LHYa008000>MQ9d02P02Q003Jif400D0
0XX018B?RXlg00P700061d<000I[001/LF`2J`09LW5bLH]aR`1a01<002iaRY2:T8Z@Ri2;T8^@Ri2;
T8^@Ri2;T8^@RjV;T8^YRjV@ZI2YT:V@ZI2YT:V@9P000g5[000200:42`03Q0D000J?RHB9Sg0g00P7
00061cX000=[00001@02J`8000MbJhE/Rg6;009a00=/L@005@00;W6@Ri2;ZH^YT:V@ZI2YT:V@ZI2Y
T:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:TV0004LJUaRP<000=[000020000hBMQ0060006ShB?RXn:
=P081`001PLi0003J`0000L000A[001[0W400g9aLP02LAT002iaRY2:T8Z@Ri2;T8^@Ri2;T8^@Ri2;
T8^YRjV;T8^YT:V@ZI2YT:V@ZI2YT:V@9P001W6?ZHmaJ``00h@60006RXB9Sg2?=P081`001PLf0004
J`00J`P016/2L@03Rg6;009a6@02L@0ZT8^YRjV;ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@0jTV0007LJV?ZHnYL@0;0003Q9f400H000J?Q8n:ShXf00P700061cD000Ea0000J`08009[
0P0016aaLG8L000_Jg6:T8Z@RY2;T8^@Ri2;T8^@Ri2;T8^YRjV;ZH^YT:V@ZI2YT:V@ZI2YT:V@ZI00
9P00276?ZHnYSjU`2P000f/0Q0060006RHB9Sg2?=P081`001PLe0003J`0000X00V/20004LW6;61`0
02aaRi2;T8^YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT0>Y9`002HZ?ZHnYSjV?
RP0:0003J`0000@000N?RXn:ShZ403D020L000H7=@001F/0001[00L000E[0000L@02K1l002j@RY2:
T8Z@Ri2;T8^@Ri2;T8^@Ri2;T8^YRjV;ZI2YT:V@ZI2YT:V@ZI2YT:V@9`002g2YSjV?ZHnYSg5`00P0
00=[000010001hV?RHm`Sg00=@081`001PLd0003L@0000@00V/00`1[00040003L@0001l000AaRi2;
19008:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@1jTW000?RXnYSjV?ZHnYSjV?ZG2:00/0
0Xl01HB:ShZ?03D020L000H7<`000g40J`070003J`0000@000=/00007`005fb@Ri2;T8^@Ri2;T8^@
Ri2;T8^@Ri2;00:@01JYT:V@ZH^YT:V@ZI2YT:V@ZI2YT:V@9`004W2YSjV?ZHnYSjV?ZHnYSg5[L@H0
00Z408Z?Q8m`Sg2:=0081`001PLd0003L@1[00D000=[00000`000g40LP0Q0004LH^@R`B@02:YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@1JTW000DRY2YSjV?ZHnYSjV?ZHnYSjV?ZG04000:
Q02?ShB:ShZ?LC@020L000H7<`000g40000;009/00=aK00080005Fb@Ri2;T8^@Ri2;T8^@Ri2;T8^@
R`02T00HZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@9`00472YSjV?ZHnYSjV?ZHnYSjT4S`03L0000080
00V408n4Sg2?L8X0=0081`001PLc00066740J`1[1`02L@03LW5b024000AaRi2;19008:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@1jTW000FRY2YSjV?ZHnYSjV?ZHnYSjV?ZHnXL0<000V4Jhn4
RXn:Sf/0=0081`001PLc0003L@0000T000IaK75/LF`Q000CK92;T8^@Ri2;T8^@Ri2;T8^YR`06T00D
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI02ZBL0011`ZHnYSjV?ZHnYSjV?ZHnY1Xl00g000002Q007ShB4L8m`
R@0d00P700061c<000<HL@000P000f/00003009a00EbLG9aLP0P009a00>;T90019008:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@1jTW000HRZVYSjV?ZHnYSjV?ZHnYSjV?ZHnXSjQ`0P002F^4
RXB?RXmaQ00c00P700061c<000Ea0000J`03008H00QaK75/LFaaK20001AaK92;T8^@Ri2;T8^@Ri2;
T8^@R`B@01JYT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@0ZTW000@L:V?ZHnYSjV?ZHnYSjV?Z@J?00>:S`00
0P0028Z4Q72?L8m`<`081`001PLc00036740008000=[07400g401W9aLW5bLB400W400h^@T004T00P
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI07ZBL001N:T:V@ZHnYSjV?ZHnYSjV?ZHnYSjR?
Z002S`800XX018B?RXl2RS<020L000H7<`004G40001[074HLFaaK75/LFaa024001AaK92;T8^@Ri2;
T8^@Ri2;T8^YR`:@01RYT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI02ZBL0011`ZHnYSjV?ZHnYSjV?ZHnY
1Xl01hZ?RW40ShT00X@018m`Sg0c00P700061c8000=a67401005L@06LW5bLG9a8@000g6@R`04T00R
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT0NY9`005hZYZI2YSjV?ZHnYSjV?ZHnYSjV?
Z8nX00>?00AaShn40XX018n:Sf/b00P700061c<000=a00000`002g4HLAQaK75/LVab024001YaRi2;
T8^@Ri2;T8^@Ri2;T8^@Ri2;T8^@RaH000>YT:T090008W2YSjV?ZHnYSjV?ZHnYSjV?ZHZ?RXn:ShZ?
RP2:ShV4L8X2Jc8020L000H7<P02L@@00W402W9aLW5bLG9aLW4Q000:LI2;T8^@RjV@ZBH000>Y[ZT0
0jTR000LLJj?[XnYT:V?ZHnYSjV?ZHnYSjV?ZHnXSjR?L@:?00B:RHB:<`081`001PLc000AL@00074H
LAQa674HLFabK7808@000g40000]0003ZI2Y00BY00>@Z@007P008hZ?ZHnYSjV?ZHnYSjV?ZHnYSjR:
ShZ?RXn:S`2?RXm`RW1a038020L000H7<P02L@800g402fabLG9aLW5bLG9a020000>:K000;P000jV_
Z@02Z@03[ZVY00FY6`006g6^T:j?ZHnYSjV?ZHnYSjV?ZHnYSjV?ZHnXS`0200>?00>:Sf/0<P081`00
1PLc000AL@00674HLAQa675/LVabK7807@001HZ4Q8Aa02l000>YT:T02ZT00i2YT00H000CRXnYSjV?
ZHnYSjV?ZHnYSjV?Z002S`06RXn:ShZ?0`001Hn9Sf^:038020L000H7<P02L@03075a00=a00YbLG9a
LW5bLG9a6P000hZ5RP02Q003RY1a02l001>Y[ZV^ZJjY[ZV^ZJjY[ZV^ZJjY01D000Ea[Y2^S`03Z@0C
SjV?ZHnYSjV?ZHnYSjV?ZHnXS`030005L8n:Sg40<P081`001PLa000CJ`00074HLAQa674HLFabK79/
L@0G0003RXB:00F400>@Rg40;`000jV@Z@0<Z@08T:VYZHnYSjTA000CRXnYSjV?ZHnYSjV?ZHnYSjV?
Z@02S`06RXn:ShZ?1@000g00RP0b00P700061c4000=[L@001W402W9aLW5bLG9aLW4E0003QHZ400J4
00B:Rg5a;`006:V_ZJjY[ZV^ZJjY[ZV^ZJjY[ZV^ZJjY[P^Y1@001G6^ZJj?00>Y01>@ZHnYSjV?ZHnY
SjV?ZHnYSjR?00L00V/a00P700061c4001=a0000LAQa674HLAQaK79/LVab018000EaRf^4J`07Q004
Ri00L@X000I=DTeBCE8O00VY00N@ZI1eDWEB00fY00NXZJRYZ:V?00@001^:SjV?ZHnYSjV?ZHnYSjV?
ZHnYSjR:ShZ?RXl01`000hA[000`00P700061c402G402W9aLW5bLG9aLW4@0006T8^@RhZ;1X@018Z5
T8/2L@X000IeDWEBMDdO000VZJnY[ZV^ZJjYNe9kDWYBNU9jDWZY[ZV^ZJjY[ZV^ZJjY[ZV^ZJh20004
LJjY[PBY01>?ZHnYSjV?ZHnYSjV?ZHnYSjR?00L000>:0000<0081`001PLc000?L@1a674HLFaaK75/
LVab009a2@003X]aRg6;LH]bRg:@LXY[18@01hZ;T8^@K7402P001TeeCGE=MAl01ZT05XnYME9eDWEB
ME9eDWEBME9eDWEBME82M@:?01^YZ:VXZJRYSg40ZHnYSjV?ZHnYSjV?ZHnYSjP01Xl00hZ?00070003
L00002l020L000H7<`001FaaK75/009a00YbLG9aLW5bLG9a10004HY[RV^:LY2;T8^@Ri2;T8^@00J4
00F:T8^@R`02L@X000IeDWEBMDdP000X[ZV^ZJj?[U9jDWYBNgF^ZG]fU59jDWYBNU9jDWYBNU:]SjjY
[ZV^L@<001N?[XnYSjV?ZHnYSjV?ZHnYSjV?Z8nXS`080003S`0002l020L000H7<`003g40LAQa675/
LFaaK79/LP02L@<001B:JhA[Rg6;LH]aRg:;LX]bT764J`B400N@Ri2;T6aa00X000I=MDeeCGDP00FY
0WH01WEBME9eT0JY00>?MXl00WD0459eDWEBME9eDWFXZJRYZ744000>SjV?ZHnYSjV?ZHnYSjT8S`P0
00=`0000;`081`001PLb0006LFaaK75/0W402G9aLW5bLG9aLP02000DJhY[Q6b:LX]bT8^@Ri2;T8^@
RhX5Q0:;00B@Ri2;0W4:0006ME9eDWE=8000::jY[jV^MG]BNU9jSjjY[ZV^ZJjY[ZV^ZJiBNU9jDWYB
NU:]ZJjY[W46000DZHnYSjV?ZHnYSjV?ZHnYSjR?Z8l80003S`0002l020L000H7<`009G40LAQa675/
LFaaK79/LW40K8Y/Q6^:LH]aRg6;LX]bRg:@LX/00X@00f^4Q002R`06T8^@Ri1/2`001TeeCGE=MB00
1JT01gIeME9eDY003ZT2S`0<DWEBME9eDZVXZJQa200038nYSjV?ZHnYSjR?Z0J?2@000f/0000^00P7
00061c8000IaK75/LF`2L@0NLW5bLG9aLW5b08YaRf^4K8]bRg:@Ri2;T8^@Ri2518@02HZ;T8^@Ri2;
L@0;0006ME9eDWE=8000:JjY[jV_MG]BNU9jZJjY[ZV^ZJjY[ZV^ZJjY[ZV]SgYBNU9jZJjY[ZV:00P0
016?ZHnYSjV?ZHnYSjV?Z8nXS`0j00P700061c<002Aa074HLAQaK75/LVabK780LW6;LH][Q76;LH]a
Rg6;LX]bT745Q009QI2;T8^@Ri1/00/000I=MDeeCGD=000=C@1=04d0C@1=04d0C@0600FY00J?MWEB
ME8CZ@0;MXmBME:?MZVXZG002P002JV?ZHnYSjV?Z006ScX020L000H7<P001W5/LFaaK09a00MbLG9a
LW5b008001=aRg6;LH][RW:@LY2;T8^@Ri2;00B400^:Q92;T8^@Ri2;L@0;0005ME9eDWD03Td03E9e
DWEBME9eDWEBME800Td4000Y[ZV_ZJn?[U9jDWZ@[ZV^ZJjY[ZV^ZJjY[ZV^ZJjY[ZUjDWYBNjV^ZHX0
3P002hnYSjV?ZHnXSjR?03X020L000H7<`00374HLAQa675/LFabK0@001AbLH]aRg64Jh]aRg6;LX]b
Rg::J`>400^5RX^@Ri2;T8^@K00;000RCGEBMDeeDWEBME9eDWEBME9eDWEBME9eDWEBME9eDWD0C@@0
1ZT01Y1eDWEBSaBY0WH01e9eDZVXZG005005S`T000=`0000;P081`001PLb000<LFaaK75/LW5bLG9a
1@004W6;LH]aRfb4LI1bT8^@Ri2;T0F400^@Ri2;T8^@Ri2;L@0;000PME9eDWEBME9eDWEBME9eDWEB
ME9eDWEBME9eDWYBME82C@@002V_ZJnY[jV_DWYBNWF^ZJjY[ZV^ZJjY[ZV^ZJjY[ZV^ZIB?NU9jSjjY
RP0D0005SjR?Z8l0>P081`001PLc0009LAQa674HLFaa00L0019bLH]aRg6;JhYaRg6;LX]bT6`4Q00<
RX^@Ri2;T8^@Ri1/2`008TeeCGEBME9eDWEBME9eDWEBMDeeCGE=MDeeCGE=MDee04d400NY00J?DWEB
MGHCZ@09MHmBME:@Z:U`01D018l00hX0000g00P700061c801W400g800007000@Rg6;LX]bRg:;JhZ;
T8^@R`::1H@01h^@Ri2;T8/019000g40000_000Y[jV_ZJnY[ZV^T7YBNU:^ZJjY[ZV^ZJjY[ZV^ZJjY
[Xn]DWYBNXn^ZHX05`000jR?S`0i00P700061c<000Ia674HLAP9000SK79aLW6;LH]aRV^;LH]aT6^4
JhA[Q6b@Ri2;T8^@Ri2;T6/0<@0:Z@05SgEBME800WD<Z@0<T8meDWEBME:YSjU`60000hn:000h00P7
00061c80174<000@Rg6;LX]bRg:;LXB5T8^@RPJ400R@Ri2;T8^@R`B@00=a00005`000dd0000E000Y
[jV_ZJnY[jV^ZJmBNU9jDWZ?[Y2^ZJjY[ZV^SjeBNU9jDWYBU:V^ZHX06@000hl0000g00P700061c40
00Ea674HL@0<000SK79aLW6;LH]aRf^4LI1aQ6^4JhA[RW:@Ri2;T8^@Ri2;T6/060000gEBM@050004
MDe`C@d03:T2M@08DWEBME9eDWD2S`9e00mBME9eDWEBME9eSjV?ZG006@000hX0000g00P700061c40
174=000?Rg6;LX]bRg:;LX^4RX^@00J400N;T8^@Ri2;00J@00=a000030003DeeDWEBME9eDWEBME80
10001deBME9eDWD00Td9000Y[jV_ZJnY[jV_ZJjY[jUjDWYBNU9jDWYBNU9jDWYBNU9jDW^?[ZV^ZHX0
6@000hl0000g00P700061c4000=a67403P008fabLG9aRg6;LH]aQ6^;LHA[Q6^4Ji1bT8^@Ri2;T8^@
Ri1[00h01U801gEBME9eDWD01@002GEBME9eDWEBL00800jY01]fSgEeDWEBME9eDWEBME9eDWEBSjRY
Z:V?ZG00D`081`001PLa00=a3P0028]aRg:;LX]b0X/0192:Q8D6Q0:;00J@Ri2;T8/6T003L@0000X0
00m=L59eDWEBME9eDWEBME8010002g1BME9eDWYBNU9=00L002V_ZJnY[jV_ZJnY[jV^ZJnY[jUjDWYB
NU9kMGYBNU9kSjfY[ZV^ZJjYRP0J0003S`0003H020L000H7<@000g4H000>000SK79aLW6;LH]aRg6;
JhA[Q6^4Ji1bT7:@Ri2;T8^@Ri2;T6`03008DP07MDefDWEBM@04L@0DL7E=ME9eDWEBME9eLHYaRW6:
LHX;Z@03T7FY00bY00>?ZJT00jT01jRYZ:VXZG006P000hn:00060003L00002d020L000H7<002LA00
00R;LH]bRg:;LP:;00>@Rh/01H@02XZ;T8^@Ri2;T8/6T003L@0000<000>:LHX00g402gIBME9eDWEa
RY1f009e00YBME:ZT:ZYZZVZ0Xl03E9jDWYBNU:^T:nY[jT00P00:8YaRZV_ZJnY[jV_T7YeNgF_ZJnY
[jV^ZJjY[ZV^ZJjY[ZV^ZJjYRW4J00:?00=[00000`000hX0000]00P700061c4000=a00003P00:fab
LH]aRg6;LH]aRg:4JhA[Q76@LY1bT8^@Ri2;T6aaK75[LFaaK76;T8/00i000gEBDP02DP0DSgJYT:UB
MU9eDWF@ZI2YT:V@ZI03M@06DWEBMGJ@1:T60003LHYa00BY00B@DWEB3WD3Z@08Z:VXZJRYShXJ0003
RXn4009[0`000g00000]00P700061c0000=a00003P002G6;LH]bRg:;LP02R`03T8^@00F400R@Ri2;
T8^@R`:@0W44000GLI2ZT:Z@ZY2ZT:Z@ME9eDWF@ZI2ZT7H00WD06E9eDZZ@ZZVZZJZYZXn^DWYBNU:D
SjnY[jT02@008G6:LHZ@NU9jDWYBNU9jDWYBNU9jDWYBNWEkMG^Y[ZV^L@0J0003Z8n400:40f/00hX0
000]00P700061d4001UbLG9aRg6;LH]aRg6@LX][RV^@Ri2;T8]a00X000E/T8^@R`05T005SgEBME80
0WD01JV@ZI2Y00:?00=fT7H00Y003ZV@ZI2YT:V@ZI1eDWEB0gD4Z@l00gD0659eDWEBME9jDWYBNU9j
DWYBME9eDWFYLAX000>:Sh@01H@`00P700061bl000=/00003`005G6;LX]bRg:;LY1bT8^@RhZ;T8^@
R`03L@/000UaT:Z@ZY2ZT:X00Y001E9eDWEB00:@022ZT:Z@ZY2ZT:Z@ZZVZT:n@[jV_ZJmBNU9jDYB?
[jV_Z@h001ae07YBNU9jDWIeMWEfMGIBNU9jDWYBNU9jDW]B6P02S`03RXB400>4<0081`001PL_0003
L@0000l001EbLG9aRg6;LH]aRg6@LY1bT79a67403P001Fb@Ri2;00F@01Z?ME9eDWEfZI2YT:V@ZI2Y
T:V@ZI2YT:V@Z@:@00R?ME9eDWEfT0BY3@002DeBME9eDWD0C@0;0003NU9e009e00IBME9eDWDH0008
RXn4RXV4JhX`00P700061bl000=/00003`004W6;LH]bRg:;LY1bT8^@Rg5/LA40039aT:Z@ZY2ZT:Z@
ZU9eDWEBMY2ZT:Z@ZY2ZT:Z@ZY2ZT:nY[i2ZDWYBNU9jDZj@[jV_Z@d000I=NU9jDWD?000:NU9eMGIB
NU9jDQP00Xl00hZ4S`02S`03RP1a02h020L000H7;`000aP0000?000?LW5bLH]aRg6;LH]aT01a01@0
00E/T8^@R`06T00IME9eDWEBSgFYT:V@ZI2YT:V@ZI2YT:V@MP02M@06DWEBME9e1ZT=0006CE9eDWEB
3`001E9eDWD000=e00=BM@005`000hZ?RP02RP05ShV?0740;P081`001PL^0003J`00010000eaRg:;
LX]bLFaaK75/01H0039aT:Z@ZY2ZT:Z@ZWJ?DWEBME9eSjZ@ZY2ZT:Z@ZZVZT:ieNe9jDWYBNU:^SjnY
[jV_Z@d000I=NU9jDWX?0004NU9jDP<000=eMU806002S`08Z8n?RXn:074^00P700061bl000<H0000
3`001G9aLP1a01h000=/T8/01Y002ZV@T8meDWEBME82M@:@01>YT:V@ZI2YT7EBME9eDWEBME9f00RY
3@002DeBME9eDWD0C@0<0004CGYBM@<000>:ZGD01`000f/0J`0?0009ShZ?RXB9Sf]a02h020L000H7
;P000f/0000?0003Rg5a024000eaT:Z@ZY2ZT:Z@ZY2Z009e00MBME9eDWEB009f01Z?MY1fT7IeDWYB
NU9jDWYeNjV_ZJnY[jV_Z@l000U=NU9jDWYBMDd020001dd0NU9jDTd00P001G6^ZHYa00D00Yd2Q0h0
00F?Z8n?Q002R@03Q74002d020L000H7@002LB8000=/T8/029007ZV@ZXn@DWEBME9eDWEBME9eDWEB
ME9eDWEBMWFYT0VY40002e9eDWEBME9jDWX0009=00UeDWEBNU9eDWX00`001XZYZJRYL0D000>4WH@0
3P001Hn:ShZ?00:400=[L@00;@081`001PL^0003L@0000h000<HL@008P003jZ@ZY2ZT:Z@ZY2ZT:Z@
ZP02T00QDWIBME9eDWEBME9eDWEBNWEkT:jY[jV_ZJnY[jV_ZJnY014001IBNU9jDWYBNU9jDWYBNU9j
DWYBNU9=0P00276^ZJjY[ZV:0`0016^4WF/=0003SjR?00:?0X@00hl0000]00P700061c<000=[06/0
1`000g5/LP0T000DRi2;T8^@RjV@ZI2YT:V@ZI2YT:T2T00;MXmeMWFZT:V@ZI004:T00jh0000C000?
C@1eDWEBME9eCGEBME9=00D001::Z:jXZJRYZ:V?ZG2MQ9e`LG0:00>?00N:ShB4JhY`02d020L000H7
;P001G40001[00800V/40005679aRg409000<ZZ@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:n@[i2_ZJnY
[jV_ZJnY[jV_ZJnY[jW6ZLJY80000dd00007000FSjV^ZJjY[ZV^ZJjRXhBXRZj?RW2:L0D000>?Z8l0
0Xl01HV:Q6]a02d020L000H7;@000g4000060003J`00008000EbK79/LP0T000QRi2;T8^@RjV@ZI2Y
T:V@ZI2YT:V@ZY2ZT:Z@ZY2ZT:V@00jY00>^ZJh0:P005W2YZ:VXZJRYSjV?ZHZMQ:>?ZHnYSjT500:?
00R4RXn9RW2:L2d020L000H7;@000faa0006000:J`00LG9aRg6;LB@003:ZT:Z@ZY2ZT:Z@ZY2ZT:Z@
ZY2ZT:Z@[i2_MG]eU8n^ZJnY[jV_ZJnY[jV_ZJnYaZW6ZA@00dd02E9jDWYBNU9jDP02C@P000V?ZJjY
[ZV^ZJh00ZT03:>XQ9f4[Xn^Sjj?[P@000B?Z8n:18l00g00000/00P700061bd000=a00001P002V/0
079/LVabK78T000;Ri2;T8^@RjV@ZI002WD037IeMU9eDWEBME9eT0bY00FZZJjY[`0@000DCGX0NU9j
DWYBME9jDWEBNU9jDTd60007L:jXZJRYW@02X`0?WHBMQ9f4WHBSRJR?ZHma00<00hl01XB?RXm`Q2h0
20L000H7;@02L@D00V/02FaaLW5bLH]aR`0U000bZY2ZT:Z@ZY2ZT:YBME9eDWEBME9eDWYBNU9jDWYB
NU9kMJnY[jV_ZJnY[jV_ZLJYaZT?000FCGYBNU9jDWYBNU9jDWYBNU9jDWYBC@D000J?ZJjY[ZP6W@0<
XifMQ9f4WHBMQ:j?10001:R?Q8@3S`03Q00002`020L000H7;@000g400003000<J`1[K79/LVabK79/
9@001H^@Ri2;00:@00BYT:V@0U805GEBDU9eDWEBME9eDWEBME9eDWEBM@02T0NY00NZZJZY[jV^00d0
01U=07EBNU9jDWYBNU9jDWYBNU9jDWYBNU9=00@000I`[ZRYZ:<4W@0?Q:>?WHBMQ:>9WHBMSjQ`00<0
0Xl3Q003RW2402h020L000H7;@000faa000200A[00UbLG9aLW6;LH/09000<g6ZT:Z@ZY2ZT:Z@ZU9e
DWEBME9eDWEBUGJEMW]eNU9jDWYBUHn_ZJnY[jV_ZLJYaZW6Z@0=009=00mBNU9jCGY=MDeeDWYBNP00
0dd01E9jDWYB00@000F?ZJjY[P06W@0?ZJjMWHBMSjV?Z8BMSjU`008000>XShX018@00hZ?000/00P7
00061bd0019a0000J`1[075/LVabK79/LV`T000;LH^@Ri2;ZI2YT:X00Y003WJ@MY1fT7J@MZZ@ZY2Z
0Y0028meDWEBMGJ@1jT01jZY[jV_ZJl03@001Td0NU9jDPD000EBNU9jDP050005DWYBNTd00`001W2^
Z:VXX`BM00j4ZHnYQ9f4WHVYSjV4Z0:?00E`08n?R@03Q003Jhm`02d020L000H7;@02L@0306][009[
00]bLG9aLW5bLH]aR`0T000cLJZ@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2_T:nY[jV_MG]BNU:ESjnY[jV_
ZJnYaZW6ZLJY00d00Td0159jDWD50005ME9jDWX01@001DeBNU9j00<000J?ZJjY[Z<4W@03XjV^00:X
00j4WHBMSjj?WHBYSjP0Z0:?1H@00hl0000[00P700061bd0019a0000JaQa675/LVabK79/LV`T000O
LH^@Ri2;ZI2ZT:Z@ZI2YT:V@ZI2YT:Z@ZY2ZT:Z@MP02M@03DWF@00RY00N_ZJnY[jV_00d000I=07YB
NU850005DWEBNU801P0017YBNU830006L:jYZJRY1Id03JV?ZHZMQ9f4XhnYQ9d028l01HV:JhA`02d0
20L000H7;003L@9[00]/LW5bLG9aLW5bL@03LR@003=aZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2_ZJZYZZVZ
MGIBNWF_ZJnY[jV_ZJnY[jV_ZJnYaZT03`001U9jDWYBNP9=00H0ME9jDWX5009=00=jDWX00`001XZY
[ZV^Z@FM01BR[ZV^XZ>4WHBMSjV4ZHnYS`1`Z0J?;@081`001PL]000BL@1[074HLAQbK79/LW5bLG9a
90000g6@T005T00JZY2ZT:Z@ZY2ZT:YfT7J@MY1eMU9eDWEBSgH2T0:Y00fZZJZY[jVZZJZY[jV_00l0
00i=DWYBNU9jDWYBNU9jDP@000MeDWYBNU9=008000Ma[ZR^Z:jS00>M0124XXnYSj>:WHBMQ:>?WHBX
0Xl30006LF]aRHm`;@081`001PL/00=a0f/02W9aLW5bLG9aLW42LRD003=aZY2ZT:Z@ZY2ZT:YfT7J@
MY1fNU9eDWEBME9eDWYBNU9jDZj@[i2_ZJnY[jV_ZJnY[jT040003GEBNU9jDWYBNU9jDWX00`001deB
NU9jDWD00`0028ZY[ZV^ZJjS1Id01:V^Z:P2W@09Q9f9ZHBMSjV?00H000=aL8@0;@081`001PL]000A
LAQ[074HLQQbK79/LW5bLG809@001W6@T8^@R`J@01UeDWEBME9eDWEBME9eDWEBME9eDWEBM@1=00h0
00>Y[jT03`00359eDWYBNU9jDWYBNP<000IjDWYBMDd30008LJjX[ZR^Z:T2W@0@Q9f4ZHnYSif4WHBM
SjV4Z0>?1`000g2:000[00P700061b`000=a07400f/0379aLW5bLG9aLW5bLBD000=aT@002@00659e
DWEBME9eDWYBME9eDWYBNU9jDWYBC@l000W6ZLJYaZU[06/02`02C@08ME9jDWYBNU830004CGYBM@D0
00R?LJjY[ZV^Z@FM012Y[ZV^Xif4WHBXSif?ZHnX1`000hma000[00P700061bh0010HL@1[679/LVab
K79aLW5b70002XF;RX^:RhZ;RW4:000JDWEBME9eCGE=MDeeCGE=MDeeCGE=MDee04d>000:ZJnY[jV_
ZJnY[`l000E=04d0C@04009=1`001hZY[ZR^Z:h019d03XBYSjV?Z8BMQ9f4ZHBX0hl80003Q00002X0
20L000H7:`003g40075/LF]bLG9aLW5bL@03LP03L@0001P000b;RXB4WHBMQ9f4Xg4:0004MDeeDR@0
00g6ZLJYaZW6ZLJYaZW601h000Ia[ZV^ZJh6W@0@ZJjX[ZR^Q9f4WHnMQ:V?Z0P000=[0000:P081`00
1PL]008H00ma66/0LQQbK79/LW5bLG805`03RP^400=a0000<0004JV_ZJnY[jV_ZJnY[jV^T7Ye01X0
00F:ZJjX[P02Z0BM00j:ZHnYSjV?WHBMQ:V4W@>?20000h@0000Z00P700061bh000AaK75[0W401W9a
LW5bL@9b00>;L@004`000i2;R`07Q008WHBMQ9f4WG4F0003C@0001T002G6ZLJYaZW6ZLJYaZW6ZJiB
NU9kMG]eNgFDMIAeU7FDMIAeNgEk00H000Ma[ZV^ZJjY00:X01>^ZJjY[ZV^Z:jSWHBMRIf4ZHnX00P0
00=a0000:P081`001PL]000ALAQa6740LVabK79/LW5bLG804@001I2;T8^@00F400>5Q8D01h@00g40
000=009=00U`0000ME9eDWD00TdF000WZJnY[jV_ZJnY[jV_ZJj@NU9jDWYBNU9jDWYBNU9jDWYBNU9j
DWYe00@001b:ZJjX[ZR^Z:jX[XnYSjV?ZHnXQ9f4WHBMSjR?20000g00000Z00P700061bd01W4016]b
LG42J`05LH]aRg402`03RP:;00N@Ri6;TH^A00B400^:Ri2ZT:Z@Y8ZTL@0<0006ME9eDWEB0`002e9j
DWYBNU9jDWYB009=4000AlJYaZW6ZLJYaZW6ZLJY[hnESgYBNU9jDWYBNU9jDWYBNU9jDWYBNU;5ZLEa
[ZV^ZJjY[ZV^ZJjY[ZV^Z:j?XhBMQ:V4WHnX00T000>?0000:@081`001PL]0008LAQa674HJf`4J`04
LVabL@H001>:K8Y/RVb4JhA[T7:@LY2;T8^:00F400V@Ri2;T8^@Rg402`001deBME9eDWD00`003GEB
ME9e07EBME9eDWD04000;JV_ZJnY[jV^ZJjY[ZV_ZJjYNe9jMG^Y[WFDMYB?U7JDMYAfU7JDDWZY[ZV:
S`02Z@0F[ZV^Z:VXZJRYSjV?ZHnYSj>4WHnXQ0:?2@000g00000Y00P700061bd01G47J`0;LH]bR`2;
LX]/RV/028@2R`09T8^ARi6;TH^:00>400Z:Q:Z@ZY2ZT:Ya2P0027EBME9eDWE=0`001E9eDP1=0080
00M=DWYBNU9=010002^YaZW6ZLJYaZW6ZLJYaWJDDWYB[jW6ZLJYaZW6ZLJYaJW5ZLFYUHo5ZLEa0080
01Z^ZJjY[ZV^ZJjY[ZR^Z:j?[XnMQ:j?WHZYS`P000>?0000:@081`001PL]0007LAQa66`0K005J`0N
LVab079aRfb4JhA[Q6^4JhA[Q7:@Ri2;T8^@Ri251H@2RP06Ri2;T8]a2P001U9eDWEBC@@000ABMDee
1@001gE=ME9e04d03`00:jnY[jV_ZJnY[ZV^ZJnYNU9jDW^Y[ZV^ZJjY[ZV^ZJjY[ZV^ZJjY[ZV^ZHX0
100058n^Z:jXZJR^SjV?ZHnYSjV4WHnX0X@2Sc@020L000H7;005L@0:Jg9aLFabLG9aLP8000MbRg::
JhA[00J400f:QI2;T8^ARi6;TH^A00F400RMQ:>@ZY2ZL@X000EeDWEBM@050004NU9eDPD000EBME9j
DP02C@l002^YaZW6ZLJYaZW6ZLJYUE9jDWYeaZW6ZLJYaZW6ZLJYaZW6ZLFYaJW5ZLEa00D001NY[ZV^
ZJjY[ZV^Z:jX[XnSQ:>?ZHBXS`0d00P700061bd000ea674HJfabK79/LVab00<001Z;LH][Q6^4JhAb
RV^4K91bT8^@LY2;T8^@R`R400F@Ri2@L@0:0005DWEBME801@00159eCGD50007ME9eDWD0C@0?000[
[jV_ZJnY[jV_ZJjY[gIjDWYBNXn^ZJjY[ZV^ZJjY[ZV^ZJjY[ZV^ZJjYRP0<000<SjV?ZHnYSjV?XhBX
0Xl00hB?00080003L@0002P020L000H7;004L@06LV]bLG9a0g84000:LX]bRV^:LY2;RP>400n@Ri2;
TH^ARi6;TH^ARYd00h@029f4WHZTT:Ya2P0017EBME82C@@000AjDWEB0`02C@06DWYBNU9=4000:jW6
ZLJYaZW6ZLJYaZVEDWYBNU9jDYB?UHnESiF?UHo5[/JYaZW5ZLFYaG405P001Yf?ZHnXS`T000=a0000
:0081`001PL]000;LAQa675[LVabK7801@006h]aRf^:LH]bRf^4JhYbT7:@Ri2;T8^@Ri2;RP07Q005
RXB@T7402P001U9eDWE=M@@000iBMDee07E=ME9eDWEBMA4002^_ZJnY[jV_ZJnY[jV_T7YBNU9jDWYB
NU9jDWYBNU9jDWYfU7JDMZjY[ZV:01H000>4Z8l00hl90003L@0002P020L000H7;004L@06LW5bLG9a
1`002G:;LXYbT7:@Q@03Q00?RY2;TH^ARi6;TH^ARi6:00:400ZMQ9f4WHBMT:Ya2P001gEBNU9jDWX0
0Td03@1jDWYBNU9jDWYBNU800TdA000[ZLJYaZW6ZLJYaZW6ZLIeU59jDWYBNUUjDWYINU9jFGYBNU9j
DWZ?aJW5L@0G0005RZj?ZHl0=0081`001PL]0008LAQa675/LV`8000IRg6;LH]aRf^4JhA[Rg:@LY2;
T8^@Ri2;T002R`V40Y000g400009000EME9eDWEBME9eDWEBME9eDWEBME9`01<002^_ZJnY[jV_ZJnY
[jV_ZJjYNgEjDWYBNU9jDWYBNU9jDWYBNU9jDWZY[ZV^01L000>MShl00Xld00P700061b`017400g9a
LP090008Rg:;LX]bT785Q00KRi6;TH^ARi6;TH^AT96;WHBMQ9f4WHBMT:Ya00/00Td0459jDWYBNU9j
DWYBNU9jDWX2CA@002^YaZW6ZLJYaZW6ZLJYaZW6ZLJ?[XnDMIAjU7ZDNYAjNgZDMIABNU9jZLFY01P0
00BYSjV?=0081`001PL[008H00=a600030006W6;LH]bRg:4JhA[Q6^@LY1bT7:@Ri2;T8^@0X/7Q005
RXBAT7404@000gE=000O000[[jV_ZJnY[jW6ZJnY[jV_ZJnY[ZV^ZJjY[ZV^ZJjY[ZV^ZJjY[ZV^MJjY
[P0J0003SjR?00T000=`00009`081`001PL/0003LF`000d000b;LX]bRg::JhA[Q8X2R`0;TH^ARi6;
TH^AT9400Y002h^4Q9f4WHBMRZB@009a<`009ZW6ZLJYaZW6ZLJYaZW6ZJn?UHnDZLJYaZW6ZLJYaZW6
ZLF^aJk50Zh00jV^Z@0K0003SjP000P000=a00009`081`001PL[008H3`006W6;LH]bRf^4JhA[Rg:@
LY1bT8^@Ri2;T8^@0XX7Q005T8^AK7404@002Dd0MDeeCGD0C@0I000[[jW6ZJnYaZV_ZJnY[WEkDWYB
NU9jDW]eNhjEMZj?[ZV^ZJjY[ZV^ZJjY[P0L0003S`0000L000=`00009`081`001PL[0003L@0000h0
00R;LX]bRg:4J`>4012;TH^ARi6;TH^AT96@TI2:0h@029f4WHFST:Z@0W4@000=CE9jDWYBNU9jDWYB
NP02CA@0026YaZW6ZLJYaZW6MGYBNU9jDWYINU9jDWYBNU9jDWYBNhl00Zh01LF^aJk500:^00>Y0000
90000hX0000W00P700061b/000=a00003P006W6;LH]bQ6^4JhA[Rg:@LY1bT8^@Ri2;T8^@1h@01i2;
TH^ZK7403@02C@0AME9eDWEBME9eDWEBME9eDWD04`00:jnYaZW6ZLJ?NgIjDWYBNU9jDWYBNU9jDWYB
NU9jDWYBNU9jSjjY[ZV^ZJh07@03Sc0020L000H7?00028]bRg:;JhA[0h@03h^ARi6;TH^ARi6@TI2A
R`04Q008WHBYT:Z@ZY02L@/00Td05E9jDWYBNU9jDWYBNU9jDWYBNU9eC@0A000CZLJYaZW6ZG]BNU9j
FGYBNUUkMP02U00CUHnESiABNU9jDWYBNUUjU:f^a@03[Qd000J?Z8n:RHX]00P700061c`001YaRg:;
LXA[Q6^4JhYbT7:@LY2;T8^@Ri2;RPB400Z:T8^ARi6@TFaa2`002WEBME9eDWE=M@02C@0=ME9eDWEB
ME9jDWX0C@0?000[[jW6ZLJY[U9jDWYBNWEkT:jY[jV^ZJjY[ZV^ZGYBNU9jDWYBNU:^ZJjY[P0M00>?
18@00hT0000Y00P700061bX000=a00003`001X]bRg:;K0F400j:TH^ARi6;TH^AT96@R`>400Z:T:Z@
ZY2ZT:Z@0W4:0007ME9jDWYBC@090008CE9eDWYBNU82C@l000^^aZW6ZLJ?NU9jDP02NP0HaZW6[/J^
aZk6ZLJYaZW6ZJifU59jFGYI0WX01<F^[ZTM0004SjR?S`B40P000g40000V00P700061bT000=a0000
40005g6;LX]bRV^4JhA[RW:@LY2;T8^@Ri2;00>400j;T8^ARi6;TH^AT:Y/L@X000ABME9e3@002Dd0
ME9jDWX0C@0?000[[jW6ZLJYU59jDWZ?[jV_ZJnY[jV^ZJnY[ZV^ZJjY[ZV^DWYBNU9jZJjY[P0M00F?
0h@00hX0L@0X00P700061bT00W4A0006Rg:;LX]b1X@02i6;TH^ARi6@TI2;00>400bAT96@ZY2ZT:Z@
ZY02L@X000AjDWYB3`001E9eDWYB009=3`002jk6ZLJYaWEjDWYB00:^01C6[/J^aZk6[/JYaZW6ZLJY
aZW6Z@:D00AjFGYI0Zh00lF^000M0003Z8n?00:?0h@00`1a000W00P700061bT000=a00003`004h]a
Rg6;LH]bRf^4JhA[RX^@LY000XX03HB;Ri2;T8^@Ri2;T8/00Y001I6@TFaa00X000EBME9eL006L@19
Rg6;LH]aMW5jDWEBMW6:LHYaRW6:LH]aRW6@LHZYRW6_ZJnYNe9jDWZYaZW6ZJnY[jV_ZJjY[ZV^ZJjY
[ZV^ZJieNU9jDZjY[P0N00B?00J9ShB4074X00P700061bP00W4A000:LX]bRg:@LY1bRPF401J:Q8Z5
RXBARi6@TI2ZT:Z@ZY2ZT:Z@1740<h]aRg6;LH]aRg6ZMGYBNY2EZZnZ[jZ_ZZn@[gEkDWYB[Y2_ZZnZ
aZ[6Z/JYaZW6ZLJYaP020005RXn:T9@00gX08UW6[/J^aZk6[/J^aYBEU9FD[YC6[/J^aZjEU7YINZW6
[XlM0003Z8nX00B?00>40000:0081`001PLY0003L@0000l000j;LH]aRg:;LX]bRV^;J`:401B:Q8B;
T8^@Ri0HLFaaK75/LFaaK0<00W404I2ZT:Z@ZY2ZT:YfSe9eDWEB00Af01j@MY1eMU9jDWYBSgJEZJZY
[jV_ZJnY[jV_ZJnYaZT6000VMU9jDWZ?aZW6ZLJY[jV_ZGYBNU9jMGZY[ZV^ZJieU59jDZjY[W4M00B?
00F9ShV4J`0Y00P700061bP00W4A000>LX]bT7:@LY2;T8^ARi43R`06TI2ARi6@0W4>000eLW6@MY1f
UGJEMYEBNU9jDWYBNU9jDWYBNU9jDWYBNU9jDWYB[Y36Z/JZaZ[6ZLJYaZW6ZLH01`007GYINWYkLI1a
T76@LLJ^aUUjFGYIU7[6[/J^aZjD009j00=INUT00Zh00hl0000K0003Z8nX00B?00>9Q000:0081`00
1PLY0003L@0000l001J;LH]aRg:;LX]bT7:@LY2;T8^@Rg4H4@02L@0cDWEBME9eDWEBME9eDWEBME9e
DWEBME9eDWEBNU9jDWYBNY2^ZJnY[jV_ZJnY[jV_ZLJY00L000MBME9jDWE=00H001R:LGYBNU9jDWEa
RW6:L7IBNU9jDWZ?[W4M00B?00N9ShV?L01`02L020L000H79`02LA80019bRg:@LY1bT8^@Ri6;TH^A
T942LA<003EbLGEBME9jDWYBNU9jDWYBNU9jDWYBNU9jDWYBNU9jDWYBNgF_T<JZaZ[6Z/JYaZ[6ZLJ^
aP070009C@1jFGYINU9j00H000MINUUjFGYB00D000M=NUUjFGYI00:^00>?00006`001JR?Z8nX00>?
00>:L7409`081`001PLW0003L@0H014001:;LH]aRg:;LX]bT7:@LY0HLAPE009a03=BME9eDWEBMWEf
MGIeMWEfMGIeMWEfMGIeMWEkMJnY[jV_ZJnY[jV_ZJnY[jV_ZJnYaZT02@0024ejDWYBNU9e1@001gYB
NU9jDWD01@002WYBNU9jDWYa[W4M00B?00J:ShV?L8@X00P700061bL00W4B0009LX]bT7:@LY2;00Ia
6000=G9aMWEkMG]eNi6_TJnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZaZ[6Z/JZaZ[6ZLJZaZk600X0
0WX01EUjFGY=00D001IINUUjFGYINUUjFGYINUUjFGY=076?7P0018nXSjP4S`03J`0002H020L000H7
9`000g40000A0009Rg6;0740L@1a01h00W40<i2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZZVZZGIBME9jDWYBNWEk
MYFYZZV_ZJnY[jV_ZJnYaZW6Z@0:0007CE9jDWYBC@05000CNU9jDWYBNU9jDWYBNU9jDWYBC@020003
LHX001`00hl01XZ?RHm`SbP020L000H79`02LA0000@HLAQa9@00=H]aZY2_T:nA[i6_ZZnZ[jZ_ZZnZ
[XnDDWYBNU9jDWYBNU9jDWYBU8n_ZLJZaZ[6Z/JZaZo600/000E=DWEBNP06000BFGYINUUjFGYINUUj
FGYINUUe0`0018n^ShXK0004SjR?Z0B?00>:00009P081`001PLW0003L@0000l000=bL@009P00476Z
T:Z@ZY2ZT:Z@ZY2ZT:X2T00SDWIBME9eDWYBNU9jDWYBNU9jDWYBNU9jT:jY[jV_ZLJYaZT05P003dd0
C@1=04eBNU9jDWYBNP02C@@000Ya[ZV^LHYaRW6:5@05S`04RHm`SbP020L000H79`02L@<00f/60005
68]bRg80:000=H^@ZY2ZTJZA[jZ_ZZnZNe9jDWYBNU9kSjj@[i2^T:n@UHnDDWYBNU9jDWYBNZW6[lJ_
aZo600H000==00008P002hn^aJk5ZLFY[ZVX00FM0j<01If:LHma00L000>?Z8l00hl018V:Q8XW00P7
00061bL000=a60000`02J`@000MaK79aLW5b02P001MaZY2ZT:Z@ZY2ZTJYfT59eDWEBMHnZZ@06ZP0H
[jZ_ZZnY[gJDDWYBNU9jDYB?[jV_ZLJY1P001TeBME9eDQl000Ya[ZV^ZJjY[Xn^1Id02ZjX[ZR^Z9f?
ZH@600J?0h@00f/0000U00P700061bH00g4300=[0`00275bLW:;LX]b:000=H^@ZY6ZTJZA[jZ_ZWYB
NU9jDWZ@[jZ_ZZnZ[jZ_ZZnZ[j[6Z/JZUHmjDWYBNUUjZLJ_aZo600H001a=NUUjFGYINU9jDWYBNU9j
DWYBNU9jDWYBNU9j2@0028n^aJk5ZLFY0Z<4W@09Z:V^ZJjY[ZRS00:M00>400000`000hnXS`02S`05
RXB4RHl09`081`001PLV0003674000@00V/20009LFabK79aLW5b02P001MaZY2ZT:Z@ZY6ZTJYBME9e
DWF@ZZVZZ@04ZP0J[jZ_ZZnZ[jV_ZJnY[gEkDWYBNU:^SlJYaZh6000RCE9jDWYBNUUjFGYINUUjFGYB
NU9jDWYBNU9jFGYIME9eDP<000Ya[ZV^ZJjY[ZVR19d03JRYZ:jX[ZR^Sif4WH@01@04S`B400=`0000
9@081`001PLV00=a1002J`03679a009b00J;LX]bRg8X000_Ri2ZTJZAZY6_ZZnZNU9jDWZ?[jZ_ZZnZ
[jZ_ZZnZ[jZ_Z/JZaZ[6ZZj?NU9jFGX00YD01:o6[lH7000RCEUjFGYINUUjFGYINUUjFGYINUUjFGYI
NUUjFGYINUUjF@8000V?[/F^aJW5ZLD01Id03ZjY[ZV^ZJjX[XZMQ9e`10000jR?S`05Q003S`0002D0
20L000H79`000g400003009[00]bK79/LVabLG9aLP0X000GLJZ@ZY2ZT:Z@ZY6ZDWEBMU:ZZJZY[jT0
0ZX07:nZ[jZ_ZZnZ[jV_ZJnY[jV_DWYBNU9kMLJ_aZh800Y=01MjFGYINU9jDWYINU9jFGYINUUjFGYI
NP030008SjV^ZJjY[ZT5W@0>Z:jX[ZR^Z:j?[XBMSjP400B?18@00g00000U00P700061bH00W42009a
00AbJf]a1W8018]bRg8X000^Ri2ZTJZA[jZ_ZYJ@NU9jMYFZ[jZ_ZZnZ[jZ_ZZnZ[jZ_Z/JZaZ[6Z/JZ
NU9jF@9j00FD[lJ_aP0D011=00IBNUUjFGX20009LLJ^aJk5ZLFS00FM00jY[ZV^ZJjY[ZRMQ:j:[P<0
00>XSjP00Xl018Z?RXlW00P700061bH001@HL@1a675/LV]bK79/LW5bLH]aLRL000]bLJZ@ZY2ZT:ZA
ZP02T00;DWEBT7JZT:nZ[jT00ZX07:nZ[jZ_ZZnZ[jV_ZJnY[jV_DWYBNU9kMLJZaZlT0006ME9jFGYI
0P002HnY[ZV^ZJjYZ005W@0AZJR^Z:jX[Xn^RIf9ZHUaL0001Xl018V?L8XV00P700061bH01W401W9a
Jg5bL@9b00J;LX]bRg8W000_LH^@ZY6ZTJnZ[jZFT7YBNWIkZZnZ[jZ_ZZnZ[jZ_ZZnZ[j[6Z/JZaZ[6
T7YBNUT00WX01IB_aZo602@000IBNUUjFGX20009LLF^aJk5[/FY00FM00RS[ZV^ZJjY[P:X00Z4XhnM
Sjj?LG2X1Hl00hAa000U00P700061bH001@HLAQa675/LV]aK79/LVabLH]aLRL000]bLJZ@ZY2ZTJZA
ZP02T005DWEBME800ZX00jnZ[`03ZP0L[jZ_ZZnZ[jZ_ZJnY[jV_ZIEBNU9jDYF?aZo6[b@000IeDWYI
NUT2000:SjV^ZJjY[ZV^Z@BM0ZP03:jX[ZR^Sjj?WHB^Q0:?00NX0000LF]`00:?00=[S`009@081`00
1PLU009a00L0LG5bLG9a00A[00QaK8]bRg:;LRL003IaRi6_TJnA[i6_ZZnZNU9jDWYB[Y2_ZZnZ[jZ_
ZZnZaZ[6Z/JZaZ[6NYAINUUjFGZ_aZo6[lHT0006DWYINUUj0P0028[5[/F^aJk50Zh00jRMW@02W@09
[ZR^ZJjX[ZR^00:M00F4[XZ^S`080003LF/002D020L000H79@002W4061Qa674HLP05J`05LG9aLW40
:002L@0dZY2ZT:Z@ZY2ZT:YBME9eDWEBSgJZT:nY[jV_ZJnY[jV_ZZj@NgEjDWYBNU:ESlJYaZW6ZBD0
00EBNUUjDP020003Sjj^00:^00RY[ZV^ZJjSZ0>M00fSZ:j?[ZR^Sif4WHnX00:?2@000h@0000T00P7
00061bL00g400g9aLP07J`05LG:;LX/0:000=W6;TJnA[i6_ZZnZ[jZ^T7YBNU9jDWYBNWFDSiF@UI2E
T9ABNe9jDWYINUUjFIBZaZo6[lJ_aRD000EjFGYINP020008R/J^aJk5[/D2[P04ZJjYZ0>M00R^ZJjY
[ZV^Z0:M00B^Sjj?2@000hA[000T00P700061bL000Ia674HLAP2J`03LF][00=[0W400g9a000W009a
03BZT:Z@ZY2ZT:ZAZY6ZMWEBME9eDWEBNU9jDWYBNU9jDWYBNU9jDWYBNU9kZLJYaZW6ZLJY9@001E9e
FGYI008000>?[Zh00Zh06ZV^ZJjY[ZR^Z:VXXYf^Z:j?[XnYQ9f4Z8nX2@000hl0000T00P700061bL0
0g400g9aLP02J`03LW5[009[00F4JhEbR`0X000YLH^A[i6_TJnZ[jZ_ZZnZ[i1jDWYBNU9jDWYBNU9j
DWYBNU9jDWYINUT00WX02iBZaZo6[lJ_aZo602D000EjDWYINP020008R/J^aJk5[/D4[P0=ZJjY[ZVX
WJ>?[ZR^X`03W@03Sjj?00T00XlU00P700061bL000Ia674HLQP2J`03LVa[00=[00AbLG9b:002L@0:
ZY2ZT:Z@ZY2ZT@FZ02F@ZWIeDWEBNU9jDWYBNU9jDWYBNU9kDZj?[jZ_ZLJYaZW6ZLJY02D000EBMEUj
F@020003Sjj^00:^00nY[ZV^ZJjX[ZR^Z:jX[ZP00id028BMQ:>?WHnX2P000g00000S00P700061bL0
0g40179aLF/3LP08K75/Rg:;LX/X000fLH^A[i6_TJnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[j[6Z/JZaZ[6
[lJ_aZo6[lJ_aZo6[lJ_aZo6:@0077X008[6[/F^aJk5[/F^aJV^ZJjY[ZV^ZJRSZ:<2W@04[XB^S`X0
00>?00008`081`001PLW000@LAQ/6780JfabK79aLW5bL@9b:002L@0`ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@
ZY2ZT:Z@ZY2ZT:ZYZZVZZJZY[jV_ZJnY[jV_ZJnY<0000hn^[P04[P0HZJjY[ZR^Z:jX[ZR^Z:jX[Xn^
Sif?WHnX<0081`001PLV00=a00=[LF/00V/6LP05Rg:;LX/0:0002G6;T:Z@ZY2_T00]0004[lJ_JbP0
00R:aZk5[/F^a@>^01C5ZJjY[ZV^ZJjY[ZV^Z:jX[XBMSc0020L000H79`001W4HJ`1[009/00QbK79a
LW5bL@9b8@001i2AT96@ZY000W4d0007aZo6[lJ_aP0U0003Sg6^00J^01NY[ZR^Z:jX[ZR^Z:jX[Xn^
SjV?ZHBXS`0_00P700061bH00W40179aJf/2L@Ib00F;LX]bR`0F000DT:Z@ZY2ZT:Z@ZY6ZTJZAZY6Z
LG8d000H[lJ_aZo6[lJ_aZo6[lJ^aZk6[/J^aP1[0`003F/0001[0740L01a07001@002hn^aJk5[/F^
aJk500:^01:Y[ZV^ZJjY[ZV^Z:jX[XnMQ:h_00P700061bH001<HLAQaK79/LVabK79aLW6;LH]a0100
01RARi6;TH^ARi6;TI2AT96@ZY2ZT:Z@ZY02LC@002G6[lJ_aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^
aZk6[/J^aJk500>^00?5[P000P000g6^[P03[P0IZJjY[ZV^ZJjX[ZR^Z:jX[Xn^Sjj?[XnYS`0_00P7
00061bH01740179aLW44LP07Rg:;LX]bR`0=000MTH^AT96@TI2AT96@ZY2ZT:Z@ZY2ZT:Z@ZY6ZLH/0
=000=Zo6[lJ_aZo6[lJ_aZo6[lJ_aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZi`08n^aZk5[/F^a@J^
012Y[ZV^ZJjX[ZR^Sjj?[XnX;`081`001PLV000C674HLFabK79/LVabLG9aRg:;L@06000TT7:@LY1b
TH^ARi6;TH^ARi6;TI2AT96@TI2ZT:Z@ZY2ZT79a=0009lJ_aZo6[lJ^aZk6[/J^aZk6[/J^aZk6[/J^
aZk6[/J^aZk5[/F^a@05[P04aG6^Z@>^01VY[ZV^ZJjY[ZV^Z:jX[ZR^Sjj?[Xn^SjV?00X000=a0000
8P081`001PLV00Aa00AbLG9a1780<H]bRg:;LX]bRg:@Ri2;TH^ARi6;TH^ARi6@TI2AT96@ZY2ZT:Z@
ZY2ZT:ZAZY6ZLH/0=000;Zo6[lJ_aZo6[lJ_aZo6[lJ_aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZk6
[Xl50006[ZV^ZJjY1Jh03ZV^ZJjX[ZR^Z:j?[Xn^2P000g40000R00P700061bD00QP0374HLFabK79/
LVabK0H002V;LX]bRg:@LY1bT7:ARi6;TH^ARi6;TH^AT96@TI2AT:Z@ZY2ZT:Z@LP0f000X[lJ_aZo6
[lJ^aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aJk5[/F^a@>^00?5L@0030004JR^Z:j?[Xn^Sjj?[Xn^
SjV?00X000=a00008P081`001PLV009a00AbLG9a1G800g400005000YLX]bT8^ARi6;TH^ARi6;TI2A
T96@ZY2ZT:Z@ZY2ZT:Z@ZY6ZTJZA[g40=P00;LJ_aZo6[lJ_aZo6[lJ_aZo6[lJ^aZk6[/J^aZk6[/J^
aZk6[/J^aZk6[/J^S`0E0009[Xn^Z:j?[Xn^00X000=a00008P081`001PLU008H00Ua675/LVabLG80
2@00:H]bRg:;LY1bT7:@Ri6;TH^ARi6;TH^ARi6@TI2AT96@ZY2ZT:Z@ZY1b03H002f_aZo6[lJ_aZo6
[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZk5[/F^aJk5[/F^aG4060001JV?[Xn^00:?2@000g40000R00P7
00061bD00g40179aLW43LPX002UbRg:@Ri6;TH^ARi6;TI2AT96@TI2AT96@ZY2ZT:ZAZY2ZTJZAZY6_
L@0f000]aZo6[lJ_aZo6[lJ_aZo6[lJ_aZo6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZj?01T000J^Sjj?
[Xl90003L@00028020L000H79@002G4HLFabK79/LP0;000YRg:;LX]bT7:@LY2;TH^ARi6;TH^ARi6@
TI2AT96@TI2ZT:Z@ZY2ZT780=P00;Jo6[lJ_aZo6[/J^aZo6[/J^aZk6[/J^aZk6[/J^aZk6[/F^aJk5
[/F^aJk5L@0J0003ZHn^00:?2@000g40000R00P700061bD00g40179aLW4=000YLY2;T8^ARi6;TH^A
Ri6;TI2AT96@ZY2ZT:Z@ZY6ZTJZ@ZY6ZTJZA[g40=P00;LJ_aZo6[lJ_aZo6[lJ_aZo6[lJ^aZk6[/J^
aZk6[/J^aZk6[/J^aZk6[/J^S`0K0004[Xn^Sbh020L000H79@000g40000A000YRg:;LY1bT7:@LY2;
TH^ARi6;TH^ARi6;TH^ARi6@TI2AT:Z@ZY6ZTG40=P00:Zo6[lJ_aZo6[lJ_aZk6[/J^aZk6[/J^aZk6
[/J^aZk6[/J^aZk6[/F^a@:^00=aRP009`000g00000Q00P700061b@000=a00004P00:G:ALY6;TH^A
Ri6;TI2AT96@TI2AT:Z@ZY2ZT:ZAZY6ZTJZAZY6ZTJma03H02lH08jo6[lJ_aZo6[/J_aZk6[/J^aZk6
[/J^aZk6[/J^aZk6[Xn:01h000B^Sjf?1P000g40000Q00P700061cT002N;LX]bT7:@LY1bTH^ARi6;
TH^ARi6;TH^ARi6@TI2AT96@ZY6ZTJX00W4f000Z[lJ_aZo6[lJ_aZo6[/J^aZk6[/J^aZk6[/J^aZk6
[/J^aZk6[/F^aJk50jh00hl0000L00J?00F>S`00L00S00P700061b@000=a00004P00:G:ALY6;TH^A
Ri6;TI2AT96@TI2AT96@ZY2ZT:ZAZY6ZTJZAZY6ZTG9a03H02lH08jo6[lJ_aZo6[lJ_aZk6[/J^aZk6
[/Z^aZk:[/J^bZk6[/J:01l000J?[HnXSjP2S`0308X0028020L000H7>0008G:;LX]bT7:@LY1bTH^A
Ri6;TH^ARi6;TH^ARi6@TI2AT002T@05ZY6ZTJX00W4f000Z[lJ_aZo6[lJ_aZo6[/J^aZk6[/J^aZk6
[/J^aZk6[/J^aZk6[/J^aJk50jh00hX0000M00F?00F>ShT0L00S00P700061b<000=a00004P00:X]b
TH^ARi6;TH^ARi6@TI2AT96@TI2ZT:Z@ZY2ZTJZAZY6ZTJZAZY6;LCH02lH08jo6[lJ_aZo6[/J_aZo6
[/J^aZk6[/J^bZk:[/Z^bZk6[/J:01l000J?[XnXSjP2S`030740028020L000H78`000g40000B000X
LX]bT7:@LY1bT7:ARi6;TH^ARi6;TH^ARi6;TI2AT96@TI2ZTJZAZP9a=P00:Zo6[lJ_aZo6[lJ_aZk6
[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^a@>^00>?00007@07S`03R@0002<020L000H78`000g40
000B000ZRg:ARi6;TH^ARi6;TI2AT96@TI2AT:Z@ZY2ZTJZAZY6ZTJZAZY6_TG9a=P0;aP0S[lJ_aZo6
[lZ_aZo6[lJ^aZk6[/J^bZk:[/Z^bZk:[/J^aXX07`001Xn^Sjf?Z0:?9@081`001PLS0003L@000180
02QbRg:@LY1bT7:ALY6;TH^ARi6;TH^ARi6;TH^AT96@ZY2ZTJZAZY6Z0W4f000^[lJ_aZo6[lJ_aZo6
[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aJk5[/F^Sal01Hl01Xj?SXl0LB8020L000H78P02LA@0
01B@LY6;TH^ARi6;TH^AT96@TI2AT0:A01BZT:ZAZY6ZTJZAZY6ZTJZA[i6;LCH02lH09:o:[lJ_bZo6
[lZ_aZo6[/J^aZk6[/Z^bZk:[/Z^bZk6[/J:Sah000J?[Xn]SjP4S`03L@00020020L000H78P000aQa
000C000XLX]bT7:@LY1bTG:ARi6;TH^ARi6;TH^ARi6;TI2AT:Z@ZY6ZTJZAZP9a=P00:Zo6[lJ_aZo6
[lJ_aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^a@>^00>?L@007@07S`04SXm`LB8020L000H7
8P02LA@002Z@Ri6;TH^ARi6;TI2AT96@TI2AT:Z@ZY6ZTJZAZY6ZTJZAZY6_TJnARg4f00W600>_a/H0
1LH07Zo6a/J^bZk6[/Z^aZk:[/J^aZk6[/J^aZk6[/J^Sal000>]SjP01XlS00P700061b8000<HL@00
4`006g:;LY1bT7:@LY2;TH^ARi6;TH^ARi6;TI2AT006T@09ZY6ZTJZAZW5b03H002n_aZo6[lJ_aZo6
[lJ_aZo6[lJ^aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/F^aJk5L@0O00R?00=`00008@081`001PLR009a
5000:Y6;TH^ARi6;TH^AT96@TI2AT:Z@ZY2ZTJZAZY6ZTJZAZY6ZTJnA[i6;LCH04lH07:o6[/Z^bZk:
[/Z^bZk6[/J^aZk6[/J^bZk6[XlO0005Z8nXSjP018l00g40000P00P700061b8000<HL@004`006g:;
LY1bT7:@Ri6;TH^ARi6;TH^ARi6;TI2AT004T@0;ZY6ZTJZAZY6ZLG80:P00>g6@LI1aT76@RY2:T:o6
[lJ_aZo6[lJ_aZo6[lJ_aZo6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aJk6[/Ea01l028l00g2?000Q00P7
00061b800W4C000[LY6;TH^ARi6;TH^AT96@TI2AT:Z@ZY2ZTJZAZY6ZTJZAZY6_TJnA[i6;L@02000/
Rg6;LH]aRg6;LI1aT76@LI1aT76@LI2:T8Z@RY2:T8Z@RY2:aZo7[lJ_aZl8aP05T8Z@a/X00lH00lZ_
aP0:aP0J[/Z^bZk:[/Z^bZk6[/J^aZk6[/J^bZk:[XlO0005Z8nXSjP018l00g40000P00P700061b80
00<HL@004P0078]bRg:@LY1bT8^ARi6;TH^ARi6;TH^ARi6@TI02T@Aa00abLG9aLW5bLG9aRi44ZP0H
[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ0Zl05LJM0Id1W@6M`:?6[lJ_aZo6[lJ_aP03000/T76@LI1a
T76?LHmaSg6?LHmaSjk6[/J^aZk6[/J^aZk6[/J^aZk5[/J^aG4O00R?00=`S`008@081`001PLQ009a
50007W:ARi6;TH^ARi6;TI2AT96@TG5bLG9aLW5bLG9aLPQ[0`000h]a[`02ZP03ZIfM00RM02:ZZK2Z
/:[7Z/NY`Id1W@6M0Id1W@6M0JW6[lN_ajo6[lJ_1/HH000GT8Z@RY2:T8Z?RXn:ShZ?RXn^bZk:[Xl0
7`001Jf?Z8nX00B?00=a000080081`001PLQ0003LAPH01<0016;LX]bT7:@LY2;TH^ARi4HL@0;0003
JhB400:40f/02HA[WHBMQ9f5W@04ZP05ZIfMXi400ZX04ZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZP:M01D1WL2S
aZo6[lJ_aZo6[lJ_aZo6[lH01@003@6M0Id1W@6M0Id1W@400Yd;009[2@001Zk6[/EaSah028l00hV?
000Q00P700061b400W4D000:LY6;TH^ARi6;T@=a00Eb0000Q007009[00>MJf/02@0018A[WH@2W@0/
ZJB_ZZnTZIf`Z[2Z/:Z`Z/NZaj[7[lJ_aZo6ZLJM0Id1Z/J_ajo7[lN_aZl8aPX000>M0@400Yd3J``0
00>4J`402@001XmaSg6?LAl000B?[HnX18l00g40000P00P700061b4000=a61P04`002H]bRaQa674H
L@0500:400=[00001@04Q0d000B4RhZM0Z<01jZ_ZZnZXi400ZX04ZnZ[jZ_ZZnZ[jZ_ZZnZ[jZYY0:M
01FjWLJ_aZo6[lJ_aZo6[lJ_aZo6[lH02P001F/1WHA[00l00X@00if4J`07000=Q000076?LHmaSg6?
L@0H00N?00>>S`008@081`001PLQ009a2`002G4HLFaaLX]bR`060003Jh@000B400=[00001@02J`04
WHB4J`h003R;ZZnZXi6_Z[2Z/:ZSWK2Z/:Z`Z/N_ajo7[lN_ajo7WKZM0Ig0[lN_ajo7[lJ_ajo7a/O6
1lH7aP/000B4WHBM3`030@04W@5[Q0H00V/06`1[W@6M0Id1W@6M0Jk5[/F^aJj?LHmaSg6?L@070006
Sjj?[HnX18lQ00P700061b4000=a60002@02L@Ab00B4JhA[2P03Q0P000A[Q8B440007g6ZTJZ;ZY2_
ZZnZ[j>YZZnZ[jZ_ZZnZ[jZ_ZZnZ[jX01Id05:W6[lJ_aZo6[lJ_aZo6[lJ_aZo630000f^MW@0>0008
Jid1W@6M0Id:0003RYfM00>M0ZP00jjY[P07W@07[ZR^Z:jX[P0700N?00>>Sg008@081`001PLQ009a
200031QaLXA[Q6^:LX]bR`X00h@60006J`2MQ9f44000>8^Z[jZ`Z[2Z/:Z`Z[2MZJB`Z[2Z/:o7[lN_
ajo7ZL2M0Id1WLN_ajo7[lN_ajo7[lO61lH7aPO63@020@h000Y[Q041W@6M0F^4200048X1W@6M^Zk5
[/F^aIfjXlD8[P03XZh000D000J?[Xn]SjP4Sb4020L000H780000aQa60060006L@1bJg9/0f/01Fb;
LX]b00X00h@70003JhB400:44000376ZTJZA[i6_ZZnZ[`:Z0id03J>MXifSWJ>MXifSWJ<00id05`6M
^Zo6[lJ_aZo6[lJ_aZo6[lJ_aZo600d000B4JhA[3002J`040Id1W@:400<1W@001`000hnMXP02W@03
Xjj^00B^016XWJRX[Zf^Z:jX[ZR^WHY`RP0300N?00>>Sg008@081`001PLP00=a0`000g4HLP03LP0;
Rg:;JhA[RW:;LX/02P000f^4Q006009[00BMQ9f440003X^Z[jZ`Z[2Z/:Z`Z[2Z0Z<0::BSZZW1ZL6Y
`JW6ZLJM^Yd1W@6Yajo7[lN_ajo7a/N_1lH7aPO61lH?00:M3002W@8100BM0@1[0X@00f^40005000@
SkZM0Ifj[/J^aJk5[/FSa@Z^00bM[Xn?L8m`[Xn]SjP4Sb4020L000H780000aQa00020003LVab009[
00E/LVabL@02J`06Q6b;LX]b2P000f^4Q0070003JhB400:440007G6ZTJZA[i6_ZZnZ[jZ_TJZMXjZ_
ZZnZ[jZ_ZZnZ00BM01L1WL2_aZo6[lJ_aZo6[lJ_aZo6[lJ_aP0?00:400=[00002@02W@040Id1W@@0
00<1W@001@000hnRX`03W@R^0ZP05Jj][ZR^Z:jX[ZR^Q9f4[Xl0L75`L@02S`03SXm`024020L000H7
80002G40L@1bLF]/J`03LP0:Rg:;JhA[RW:@LP`00X@6009[00BMQ9f44000>8^Z[jZ`Z[2Z/:Z`Z[2Z
/:ZSWL6Yajo7[lN_aj[6W@6M0Id1[lN_ajo7[lN_alK7aPO61lH7aPO63`0016^4Jh@:00@10Yd5009[
1@004HZ?Z06M0Jk6[/F^aJk5[[ZM00V^00R]WHV^SjR4W@<000=aL7400XlQ00P700061b0000LH001[
678H009[00EbK79/LP02L@06JhA[Rg:;3002Q0L000=[Q8@00X@@000MLJZAZY6_ZZnZ[jZ_ZZnZ[jVZ
WJ>Z[jZ_ZZnZZJ<019d05`6MaZo6[lJ_aZo6[lJ_aZo6[lJ_aZo601400X@80008J`01W@6MQ6/60003
Q00000<000>?LKl00id9[P0CXjR][ZR^Z:jX[ZR^Q9f?[XnX0004Q0<00W0P00P700061b400V/0175b
LG82J`=b00Z;LX]bQ6^4LY1b3002Q003J`0000<00V/019f4WH@?000iLH^Z[jZ`Z[2Z/:Z`Z[2Z/:Z`
[j>M[j[7[lN_ajW0W@6M0Ifj[lN_ajo7[lN_alK7aPO61lH7aPO601400Yd80003Jh@100<100>MQ000
2`02S`0<Z@6M0Jk6[/F^aJk53Jh02if:[Xn^S`1[WHA[00<000>?L@007`081`001PLQ0006JaQa679/
0V/0179/LV`3LP06JhA[Rg:;30000f^4J`0700B43`005G9aZY6ZTJnZ[jZ_ZZnZ[jZ_ZZnYZP02X`06
ZZnZ[jZS0id06@6M^Z?6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lH04@0018A[Q6/600:400D1W@6M0@0>0003
LJjM00:M00>^ZJh01Zh02Zf^WJ:][ZR^Z:h2Z004Q:V?[P:?00@0JhB410000g2:000O00P700061b80
0W40179aLV/4LP0:Rg:;JhA[Q7:ALP/000A[Q8B41`02Q003WHB400h003UaRjZ_Z[2Z/:Z`Z[2Z/:Z`
Z[2_ajnjWJnZajo7WKZM0Id1ZL2_ajo7[lN_alK7aPO61lH7aPO61lH04P000f^4W@0600:M10400id0
000600:4100048n^0Id1ZLF^aJk5[/F^aJh2W@0>Z9fSWJ>MXjRMZ:j?[Xl20003WH@000@000=a0000
7P081`001PLS0003LVab009[00=aLW8017801hY[Q6^:LX/03003Q0P000B4JhB40V/<0007Rg6_TJnZ
[`02ZP0@ZJZZ[jZ_ZZnZ[jZ_ZZnMX`:Z00>YWId00Yd06@6MaZo6[lJ_aZ[6XkZM^ZC0[lJ_aZo6[lH0
4`000hA[Q0040006JhBM0Id10X@900:40`001g75[[ZMZ:<02Jh02:R^WJjX[ZR^0ZP01Yf^Sjj?Z080
0X@V00P700061b800W40179aJf/6LP08Rfb4JhA[TG8<00>42002J`04Q9f4Q0`002maRjZ`Z[2Z/:>Y
Y;2Z/:Z`Z/NZaj[7Z/2T`:W0W@6M0Ifj[lN_ajo6ZKZMaZo7[`02`0:j00K7aPO61lHD0003Jh@000<0
00R4J`6M0If4J`X001P106/0Sjk:W@6M^Zk5[/F^aJk5[/F^XjP3[P09[JjSWHn^Sjj?008000=[Q9d0
1@000hl0000M00P700061b4000=a07800f/2L@Ib00N:JhA[Q7:;00`00h@00f/0000700B40V/:000J
Rg6_TJnZ[jZSWJ>Z[jZ_ZZnZ[jZ_ZZnZ[id2X`BM0Z<05lJ_aZ[6WKZ_aYfj[lJT`:>jZL2_aZo601D0
00>M00000P03W@050Id1Jh@02P03Q0050775[/000id9[P0A[JjXXYf^Z:jXXif^Sjj?Z8l00P02Q0D0
00=[00007@081`001PLR0006LF]aLF]a1W8028]aQ6^4Ji5b3003Q003J`0000L000F4WHBMQ00;000`
LI2Z/:Z`ZZnMXif`Z[2Zaj[7Z/NZajo7[l2M0Id1W@6Y`:o7[l2Y`:W7[kZMajo70/H01jnj^/O61lH0
5@02Q003J`0000<100>M0Id00V/:000EJid108n^bZP1W@6^aJk5[/F^aJk500:^00>MXjP00jh029fX
Z:j?[Xn]0P000hBM00040003L@0001d020L000H78@00174HK6`2J`03LW5b00Eb00N;JhA[Rg6;00`0
00A[Q8A[2@02J`:400>MJh@00V/70008Rg6_TJnZZJ@2W@0@XjZ_ZZnZ[jZ_ZZnZ[jZ_ZPJM01W0Y<J_
aZ?0Y<2_aYfjXl2S`:o6WKZ_aZo601D00h@01P1[W@6M0@:M3P001XBMLHn^`0>M00>YZ:h01Zh04Jf^
[JjXXjR^Z:jR[Xn^SjR?00<000>400000P000h@0000N00P700061b8000EbJg9[K007LP08Rg::JhA/
T7860003J`0000<000A[Q9f42`0026^4Q9f4WHA[1P00176;Z[02ZPFM02R`Z[2Zaj[7Z/NZajo7[lNY
`9d1W@6YaZo7[lJY^j?7[kZMajo6`<K00[X010O61lHF008100>4J`4010400h@0000<000EJif?R/ZY
0Id1ZLF^aJk5[/F^aJk500:^00bM[ZR^XifX[ZR^Sjd30003W@00008000>400007P081`001PLQ0004
LAQ/K09[00=bLG801G801h][Q6^;LX/01P000h@0000400>400<0J`002P04Q003WHA[00D000F;LJnA
X`03W@>S00nZ[jZ_ZZnZ[jZ_ZZnZ[jX00Zl07J>M0Ig6[lJ_aZo6WKZYaZ>jWKZSaZnjZLJ_aZo601H0
0X@3W@050Id1W@403`02Q003Xjk500>M00>XXjh01Zh02:f^[Jj]ZJ:S0ZP01if^Sjj?[Hl00`000hA[
00040003L@0001`020L000H78@001G5bJg9[00Ib00V;LX]bRf^4K9001P000f^400040003JhB400:4
00=[000030000f/0Q003Q00lJ`1[LH^YY:>YZZnZ/:Z`Z[2Zaj[7Z/NZajo7[lN_ajW0ZL2_ajo7[lN_
ajnjWK^SaZW00K[01lH7aPO65`000f/10@060A0001=[WLF^bYd1WKn^aJk5[/F^aJk500>^00^SWJjM
XjR^Z:j?[@030003Q6/000@000=`000070081`001PLQ0008LAQaK6]/LW46LP06Rfb4JhEb1P000hA[
000500F400E[Q6^4J`0<00A[0X@06XZ4XjZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ1Zl06LJ_aZo6[lJ_
aZo6ZLJT`:C6[lJ_aZo6[lH05`002F^M0Id1W@6M0@0A0004Xig5[P>M00>R[Zh01Jh01Zf^[Jj][P:X
00VRWJ>^Sjj?[Hl010000f/000030003J`0001`020L000H78@000g5bL@02J`Ib00V;LX]bRf^4Jh/0
1@02Q004J`00J`B400A[06/01F/B000iLI2Z/:Z`Z[2Z/:Z`Z[2_aj[7Z/N_ajo7[lN_ajo7[lN_ajo7
[lN_ajo7[lN_alK7aPO61lH7aPO6010016/018@008@80A4001::0JW5W@6M^Zk5[/F^aJk5[/D5[P09
XifXZ:jX[Xn^00<00V/50003J`0001`020L000H78@000g4HL@02J`04LG9[L@Eb00J;LXA[Q6`40003
Q6^400<000=[06/07@006X]a[i6_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ1:l06<J_aZo6[lJ_aZo6[lJ_
aZo6[lJ_aZo6[`?640003V^4W@6M0H@1W@6M0Id10YdA0005Sj:jW@400id9[P07[Jj][ZR^X`02W@06
[Xn^SjR?10000h@000030003L@0001`020L000H78002L@04LW5[K09b00maJhA/Q6^:JhY[Rg:;06/0
:0008g:;Z[2Z/:Z`Z[2Z/:[7Z/N_ajo7[lN_ajo7[lN_ajo7[lN_04L001:?bZjjW@6M0Ig5[/F^aJk5
[/D7[P:X00J^Z:j?Xhl30003Q6/0008000>4L0007@081`001PLQ0008LAQbJg9/LV`2J`0:LFabK8]a
Rg:;LRl001eaT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@L@0P000BT76@LI1aT76@LI1aT76@LI1a
5P001hn^aJh1W@400id2Z003ZJR^00:^012][Zf^[Jj][ZR^Sjj?[XBR0`000h@0000R00P700061b00
0W400g9aJ`03LP0=Rfb:LX]bRg:<LY6;T00^0003LI2@01^@7`000g6@T008T003Ri2@00^@00EbT7:@
LP0;00:?012^bZk:[/Z^bJh1W@6XaJk52:h02:f^Z:j?Z8BM0P000hBM00040003L@0001/020L000H7
8@00275/LV]bK79/178028]bRg:;LX]b3@0036^4JhA[T76@LI1aT1D00W40791aT76@LI1aT76@LI1a
T76@LI1aT76@LI1aT74P000NT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1a2P001g40LG75[/D0
2Zh04Jf^[Jj][Zf^Z:j?[Xn^Sj<0009[0X@50003J`0001/020L000H77@001V/0079aLP9a0g803H]b
Rg:;LXabS7:ARi003@02Q003J`2@00>@1H@00i1aL@04L@d000=aT9006i0O0003LI2@01B@00R;T7:@
LY1bT0<01G4057:@LHA[Q6^:LY1bT7:@LLF^aJk52:h01jf^Z:j?[Xl00P000f/0J`04Q00306/001`0
20L000H77@0036/0JaQaK79/LVabK0Ab00V;LX]bRg:;LP000V/:000SJh@0076@LI1aT6^4JhA[T76@
LI1aT76@LI00L@1a0740L@000W406Y1aT76@LI1aT76@LI1aT76@LI1aT76@LI1a80004i1aT76@LI1a
T76@LI1aT76@LI000W401P1a0740L@<001Z@LI1aT76@JhAaT75bK8AaT76@LI1aRg6;J`@00V/02jj]
[ZR^Sjj?[Xn]00P000>407007@081`001PLM0007J`00LW5bL@04LP07Rg:;LX]bS0060003JhA[00X0
0X@200J@00>:Q8X0290<Q004L@0000]a3i0O0003LI2@00B@2W470003Q00000@001Q[Q7:4JhA/T7:@
LY1bT6^:LY1bT7:@LY02L@@00V/30008J`2^Z:j?[Xl80003JhB?01d020L000H780003aQaK79/LVab
LHmaQ6^4J`0600=[2@000hA[J`02000NLI1aT76@LHA[T76@LI1aT76@LHAaT76:LHA[Q6^40P000f/0
00040003J`0000l001aa0740L@1a0740L@1a0740L@1a0740L@1a07400W400i1aT0030005J`1[06/0
1`02J`06Q6^4JhA[20006I1aT6^4Ji1aT76@LI1aQ6^@LI1aRg6;LF/01@02J`H00V/0102?[Hl80003
Q70001`020L000H78000479aLW5bJi1aT76:JhAaS`05J`:42@0016^4Q6/200J@0h@:T003RY2@00>@
00F:Q8B4L@080003Jh@000L000=[08@018@400J@00>:Q8@01X@00i2:T004T0J400>:T8X00h@50004
JhB4J`H000E[Q8B4J`08000JLY1bQ6^:LY1bT7:@LY1[Q76@LY1bT7::Jg440003Jh@000D00V/30003
J`0000L000=a00006`081`001PLO008H00ibK79[Q76?LHm[Q6^?L@H000>4J`002@000hA[0002000N
LI1aT76@LHA[T76@LI1aT76@L91aT76@LHY[Q6]a1`02J`03Q00000T000=[00001@00891aT76@LI1a
T6^4JhAaT76@LI1aT76@LHAaT76:JhA[2@000f^4J`060003Q6][00T001V@LI1[Q6b@LI1aT76@LHA[
RV^4JhA[Q6^;00D00V/600=[00<0J`000P000g4H000R00P700061ah0015fJhYaSg64LHmaSg64Ji1a
T0050003Q6^400T00h@30007T8n@Si2?T003Q005RXB:Q8X00Y000hB@T005T0>400>:L@001@001HA[
Q8A[00T000>400001P0:T0>42Y000hZ@T002T004Q92@L@X000=[Q0001003Q0X001YbT7:4Ji1bT7:@
LY1bT764Jh]bT7:@LI1bL@@000E[Q000J`04009[1@0016]bLG82J`03K01a01`020L000H77P000g1a
J`02J`0;Sf]aLHmaRV^4LHl01@02J`03Q00000L00V/00hA[0002000LLI1aT76@LHA[RW6?LHmaRV^@
LI1aT76@LHA[Q09a1@001HA[Jf^400T00V/8000OT8^@LI1aT76@LHY[Q7:@LY1bT7:@LY1aT76@LI1a
T00;00=[10000f^4J`0:000IT76:JhAaT76@LI1aT76@JhA/Rg6;LF]aLP05009[0P000f/00002009[
00<0J`000P002F/HLP1a6740L@0L00P700061ah000mfLHmaRW64LHmaSg64Ji000W440003JhA[00X0
0X@03V/0002@Si2?T8n@RXB919000hZ?T008T0>400>@L@000`02J`800h@80003J`0000L02i02Q003
RY2@00J@00>:T9001Y000g4000080003Jh@000@000>4Jh@02P006W:@K8AaT7:@LY1bT7:@LXY[RW:@
LXYaT79a100016^406/500=[1002J`04K75aLP=a70081`001PLN00=a00efLHmaRW6:LHA[RW6?00D0
00>4Jh@02@02J`0QQ000091aT76@LI1aT6^@LI1aSf^@LI1aT76@LI1aQ6^4009a0`02J`@000>4J`00
1`000f/00007000<T8^@LI1aT76@LI1a0X@04I1aT76@LI1[RW6@LI1aT76@00/00f/3009[00>4J`00
2@006I1aQ6^;LI1aT76@LI1aT764Ji1aRVb;LG801@02J`0306/000D00f/30009LQQb674HLAQa01`0
20L000H77P003hmaSg6?LHmaSg6:JhA[T002L@<00V/01H@0001[00H00h@30007LI2?T8n@S`02T003
Q8Z@00:@00>4T90029001XZ4Q8B@L@800f/30003JhB400L000=[0000200;T003RXB400N@00>:T900
1i000g40000800:4100016^4Q6/:000JLY1[Q7:@LY1bT7:@LY1bT6^4LI1[T7:@LW440004Jh@0J`H0
16/20006LG9aLW5b0W4M00P700061ah0015aMW5fLHmaSg6?LHYaRW6:0002J`0406^4J`<000=[0000
10000hA[J`03000MT76@LI1aT76@JhYaT6^?LI1aT76@LI1aQ6^4Jh@00W400`1[J`060003Q6^400D0
00>4J`002@007i2;T76@LI1aT76@LHA[RW:@LY1bQ76@LI1aT76@LI002P000f^4J`040003Q6^400/0
01V@LHA[T76@LI1aT76@LI1aQ6^4Jg9aRg6;00D00f/=0006LQQb674H7`081`001PLQ000<LHmaSg6?
LI1aT76@0W402`1[08A[Q6^408A[00@000A[Q8A[0`000g6@L@02RP04T8n@S`:400>@Q9002Y04Q006
RY1aJhA[1P000f^4Q0040003J`1[00X03903Q0B@0XX:T003L@0000L00h@30004JhB4J`/001YbRV^:
LY1bT7:@LXY[T7:@LHA[Q7:@LY1bT0<00V/00hA[000<0004LG9aLR4020L000H790002W6?LHmaSg6?
LHl;0003Q000008000=[Q6/01000891[Q6^@LI1aT6^4JhYaT76@LI1aT76@LHA[Q76@075[2002J`05
Q00006/03@003I2;T76@LI1aT76@LHX00X@047:@LY1[T7:@LI1aT76@LI090003Q6^400D000>4Jh@0
2`006Y1[Q6b@LI1aT76:Ji1aT76;JhA[Rg6;LG9a16/@0003LP00028020L000H79`001g6@LI1aT740
300016/0J`03Q004JhA[J`>400B@Si2?0Y02Q003RY2@00V@00B:Q8B40Y000g64000900>40P02J``0
0i001XZ4RXB:R`B@0X@3T003Q92@00F@00>:T9000Y000g40000500>41@000f^4Q0050003J`0000@0
01]bQ6^@LY1bT7::Jh]bT7:@LXY[T7:@LY1bT1P0>0081`001PL[0003Sg6?01L001f:Ji1aT76@LI1a
Q76@LI1aT76@LI1aT6^4JhAaT002J`X000=[Q6/00P000f/0000;000?T8^:LI1aT764LI1aT6^400::
00iaRW:@LY1aT6^4LI1aT0L00V/00h@000040003JhA[00D000=[000010006HY[Q76@LI1aQ6^4LI1a
T76@K91aT76;LH/00W4h00P700061dD000EaT8n@S`0BT007RXB4Q924RP0;00>400<0Q0003@04T005
Q8Z@RY000XX2T0>40XX7T003RXB@00>@00=a00000`02Q003J`0000D00h@500>400=[Q6/00V/06XA[
T76@LH]aRW6@LY1bT7:@LY1bT7:@LY0H>0081`001PM5000MT76@LI1aT76@LI1aT76@LI1aT76?LHA[
Q6^4Ji00300016^4Jh@?0005T764LHX00h@06I1[T7:@JhAaT76@LI1aT7:@LY1aRV^4Jh@00V/K000E
T76@LI1aT76@LI1aT76@LI1aT76@03h020L000H7A`0AT003RXB400F40Y000g40000900F43P04T0>:
69000g40000L000=LY1bT8^@LY1bT7:@LP1600P700061e80011aT76@LI1aT76@LI1aT76@30001F^4
JhA[00>42`008I1aT8^@Ri1bT8^@LY1aT76@LI1aT76@LY1bT7:@LY0HL@1`00P700061h407i02LG00
20L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<0
20L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<0
20L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<0
20L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<0
20L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<0
20L000H7o`0C00P700061ol04`081`00o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00
\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output",
  CellLabel->"Out[4]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "For the spherical mapping with ",
  StyleBox["MVTextureMapType",
    FontFamily->"Courier"],
  " \[Rule] ",
  StyleBox["MVSphereMapping",
    FontFamily->"Courier"],
  " the texture coordinates are calculated from ",
  Cell[BoxData[
      \(TraditionalForm
      \`\((\[Phi], \[Theta])\) \[RightTeeArrow] \((x\_t, y\_t)\)\)]],
  ". The angles \[Theta],\[Phi] are obtained from a sphere centered at the \
origin of the coordinate system. The ",
  StyleBox["Sphere[]",
    FontFamily->"Courier"],
  " form the shapes package is in this position and the example needs no \
additional transformations."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; \n
    MVShow3D[\n\tGraphics3D[Sphere[]], \n\tMVTexture -> "\<`ChessBoard`\>", \n
      \tMVTextureMapType\  \[Rule] \ MVSphereMapping\n]\)], "Input",
  CellLabel->"In[5]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVScene3D[] \[DoubleLeftArrow] MVTexture2D[2], 
      MVSphereMapping, {256, 256, MVRGBTexture}}\)], "Output",
  CellLabel->"Out[5]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[6]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F2F2F2F2F
302F2F2F2F2F302F2F2F2F2F302F2F302F2F3010103010103000003000003000003000003F0000\
3000003F0000300F0F3F0F0F302F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000301010\
3F10103010103F10103010103F00003000003F00003F00003F00003F00003F0000
3F00003F00003F00003F00003F00004000003020203F2F2F3F2F2F402F2F3F2F2F402F2F3F2F2F\
402F2F3F2F2F402F2F402F2F402F2F3F2F2F402F2F402F2F402F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F2F3F202F3F2F2F3F0F0F\
3F0F0F400F0F3F00004000003F00004000003F00004000003F00004000003F0000
4000003F0F0F400F0F402F2F40302F402F2F403030402F2F40302F402F2F403030402F2F40302F\
402F2F4F3030402F2F4F302F402F2F4F3030402F2F4F302F402F2F4F3030402F2F
402F2F402F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F2F302F2F3F2F2F302F2F3F2F2F3F2F2F402F2F3F2F2F\
402F2F3F10104010103F10104000003F00004000003F0000400000400000400F0F
400F0F40302F402F2F40302F402F2F40302F402F2F40302F402F2F40302F402F2F4F302F402F2F\
4F302F402F2F4F302F402F2F4F302F402F2F4F302F402F2F4F302F402F2F4F302F
402F2F402F2F402F2F402F2F3F2F2F402F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F2F2F302F2F3F2F303F2F2F3F2F303F2F2F4030303F2F2F402F30402F2F403030\
402F2F403030402F2F403030402F2F403030402F2F4F3030402F2F4F3030402F2F
4F303040302F4F3030402F2F4F30304F302F4F30304F302F4F30304F302F4F30304F302F4F3030\
4F302F4F30304F302F5030304F302F4F30304F302F5030304F302F4F30304F2F2F
4F30304F2F2F4F302F402F2F4F3030402F2F40302F402F2F201F1F201010000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030202F302F2F
302F2F3F2F2F302F2F3F2F2F3F2F2F3F2F2F3F2F2F402F2F3F2F2F402F2F1F1010201F1F201010\
201F1F201010201F1F201010201F1F201010201F1F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F1F1F
4F302F4F302F4F2F2F4F302F4F2F2F4F302F402F2F4F302F402F2F201F1F1F1010000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F2F302F2F302F2F3F2F303F2F2F
3F2F303F2F2F4030303F2F2F402F30402F2F201F1F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000201F105030304F302F5030304F302F4F302F4F2F2F2F1F1F201F10000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F202F302F2F302F2F3F2F2F302F2F3F2F2F3F2F2F3F2F2F
3F2F2F3F2F2F3F2F2F201F1F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F1F1F4F2F2F4F302F201F102F1F1F201010\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F30\
302F2F302F2F302F2F3F2F30302F2F3F2F303F2F2F3F30303F2F2F402F301F1010
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000201F102F1F1F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F2F2F202F302F2F302F2F\
302F2F302F2F3F2F2F302F2F3F2F2F3F2F2F3F2F2F1F1010000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F1F1F2F1F10\
2F1F1F4F302F2F1F1F000000000000000000000000000000000000000000000000
0000000000000F2F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F2F302F2F302F2F302F30302F2F3F2F30\
302F2F3F2F303F2F2F3F2F301F1010000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000030201F\
2F1F1F503F3050302F503F3050302F503F302F1F1F2F1F1F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F2F2F2F302F2F302F2F302F2F302F2F3F2F2F302F2F\
1F101F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F1F10\
2F1F1F50302F5F3F2F50302F50302F50302F503F2F4F302F50302F4F302F2F1F1F
201F10000000000000000000000000102000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F30302F2F302F2F302F2F3F2F30302F2F201F1F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000030201F\
2F1F1F30201F503F2F5F3F305F3F2F5F3F3050302F5F3F3050302F503F3050302F
503F3050302F5030302F1F1F0000000000000000000F2F00000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000302F2F302F2F302F2F1F1010000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30201F2F1F105F3F305F3F2F5F3F3050302F5F3F2F50302F5F3F3050302F5F3F2F
50302F503F2F50302F50302F50302F50302F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F301F1010000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30201F30201F5F3F305F3F305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F305F3F2F
5F3F30503F2F5F3F30503F2F5F3F3050302F403F2F20300F000F00000000102F00000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000001F101F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30201F301F1F30201F5F3F2F5F3F305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F30
503F2F5F3F30503F2F5F3F3050302F5F3F2F50302F403F200F30001030000F2F000000000F2F00\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030201F30201F6040305F3F306040305F3F306040305F3F306040305F3F2F
6040305F3F305F3F305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F30103000103000103000103000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F1010000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030201F30201F30201F6040305F3F306040305F3F306040305F3F2F5F4030
5F3F2F5F40305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F30503F2F103F001030001030000F3000\
103000000F000F2000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F2F2030201F30202060403060403F604030604030604030604030604030
6040305F3F306040305F3F306040305F3F306040305F3F305F4030303F101F3F00103F00103F00\
20300F103000102F000000000F2F00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000001F1F1F302F2F302F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030201F30202030201F604030604030604030604030604030604030604030
5F3F306040305F3F306040305F3F2F6040305F3F306040305F3F2F5F3F30403F20103F00103000\
103F0030301F4F302F0F2F00103000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F30302F2F3F2F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F2F2030201F3F2F2060403F6F4F3F6040306F4F3F60403060403F604030
6F4F3F60403060403060403060403F6040306040305F40306040305F40306040305F3F2F1F4000\
103F00103F0020300F503F304F302F103000000F00102F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F1010302F2F302F2F3F2F30302F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030201F3F202030201F6F4F3F6040306F4F3F6040306F4F3F604030604030
60403060403F6040306040306040306040306040306040306040306040305F3F2F5F40304F3F20\
2F3F0F103F00103F0020300F503F2F4F302F2F1F1F0F2F000000000F2F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000302F2F3F2F30302F2F3F2F30302F2F3F2F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F2F203F20203F2F206F4F3F6F4F3F6F4F3F6F4F3F60403F6F4F3F6F403F
6F4F3F6040306F4F3F6040306F4F3F6040306F4F3F60403060403F604030604030604030604030\
4F3F201F4000103F00103F0050302F503F3050302F4F302F0F2F00103000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
1F1F1F302F2F302F2F302F2F3F2F303F2F2F3F2F30000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F2030201F6F4F3F6F4F3F6F4F3F6F403F6F4F3F6F403F6F4F3F
6040306F4F3F6040306F4F3F6040306F4F3F6040306F4F3F6040306040306040306040305F3F2F\
6040305F3F2F104000103F00103F00303F10503F2F4F302F4F302F40302F103000
000F00000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F30\
302F2F3F2F30302F2F3F30303F2F303F2F303F2F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F203F2F203F2F20704F406F4F3F7050406F4F3F6F4F3F6F4F3F
6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6040306F4F3F6040306F4F3F604030604F3F\
6040306040305F3F2F4F4020103F00103F00103F005F3F30503F2F503F301F300F
1030000F2F00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000101010302F2F302F2F\
302F30302F2F3F2F303F2F303F2F3030202F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F203F20203F2F206F4F3F6F4F406F4F3F704F406F4F3F6F4F3F
6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6040306F4F3F604F306F4F3F6040306F4F3F604030\
6040305F40306040305F3F2F104000103F00103F00403F1F503F2F50302F503F2F
4F302F1030000F2F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F2F302F30302F2F3F2F30\
302F2F3F30303F2F303F30303F2F303F2F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F203F2F203F2F207050406F4F407050406F4F3F7050406F4F3F
7050406F4F3F7050406F4F3F704F3F6F4F3F70503F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F2F4F10\
6040306F4F3F6040306040305F402F4F402F103F00104000103F005F3F30503F2F
503F304F302F303F20103000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F1F2F202F302F2F302F2F302F30302F2F\
3F2F303F2F303F2F303F2F303F2F303F2F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F0000000F0F0F
0000000F0F0F0000000F0F0F0000000F0F0F0000000F0F0F0000000F0F0F0000000F0F0F000000\
0F0F0F0F00000F0F0F0000000F0F0F0F00000F0F0F0000000F0F0F0F00000F0F0F
0000000F0F0F000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F203F20203F2F206F4F407050406F4F3F7050406F4F3F705040
6F4F3F704F3F6F4F3F7050406F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F604F30\
1F4F002F400F604F306040306040305F402F5F4030103F0F103F00103F00203F10
50302F503F2F4F302F50302F0F3000201F1F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000100000302F2F2F2F2F302F30302F2F3F2F30302F303F3030\
3F2F303F30303F2F3040303F3F2F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F0F00000F0F0F0F00000F0F0F0F00000F0F0F0F00000F0F0F\
0F00000F0F0F0F00000F0F0F0F00000F0F0F0F00000F0F0F0F00000F0F0F0F0000
0F0F0F6F4F4F70504F6F3F3F7040406F403F7030306F302F70303070302F70303070302F703030\
70302F70303070302F70303070302F70303070302F7F303070302F7F303070302F
7F3F300F0F000F0F0F0F00000F0F0F0F0F00000000000000000000000000000000000000000000\
0000000000003F2F20402F2F3F2F207050407050407050407050407050406F4F40
7050407050407050406F4F3F7050406F4F3F7050406F4F3F7050406F4F3F70503F6F4F3F70503F\
104F001F4F00104F002F4F106040306040305F40306040305F3F2F104000103F00
104000503F2F503F3050302F503F304F302F4F302F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000200F0F302F2F2F202F302F2F302F2F302F30302F2F3F2F303F2F30\
3F2F303F2F30402F3F3F2F30402F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F0000000F0F0F602F2F6F2F30602F2F6F2F30602F2F6F303060404F604F4F60404F\
6F4F50604F4F6F4F4F6F4F4F6F4F506F4F4F6F4F4F6F4F4F6F50506F4F4F6F4F4F
6F4F4F7050506F4F4F7040406F3F3F7040406F00007000006F00007000006F0000700000700000\
7000007000007000007000007000007000007000007000007F00007000007F0000
7000007F303070302F70303070302F7F30300000000F0F0F0F00000F0F0F000000000000000000\
0000000000003F2F203F2F203F2F20704F407050407050407050406F4F3F705040
6F4F3F7050406F4F3F7050406F4F3F7050406F4F3F70503F6F4F3F70503F6F4F3F70503F6F4F3F\
404F20104F001F4F001040002F4F0F6040306040305F402F604030303F1F104000
103F00103F004F3020503F2F50302F503F2F4F302F2F301F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F2F302F30302F2F302F30302F303F2F30302F303F30303F2F303F303F\
3F2F3040303F402F3F402F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F0F0000
602F30602F2F6F3030602F2F6F00006000006F00006F10106F10106F10106F4F506F4F4F6F4F50\
6F4F4F6F50506F4F4F6F50506F4F4F7050506F4F4F7050506F4F4F7050506F4F4F
7050506F4F4F7050506F4F4F70504F6F4F4F700000700000700000700000700000700000700000\
7000007F00007000007F00007000007F00007000007F00007000007F00007F0000
7F00007000007F00007000007F000070302F7F3F3070302F7F303070302F0F0F0F0F0F000F0F0F\
0F00000F0F0F000000402F2F3F2F207F504F705040705040705040705040705040
70504070504070504070504070504070503F70504070503F7050406F4F3F70504070503F705040\
6F4F3F40502F404F201F4F00104F001F4F002F400F604F306040306040305F402F
10400F103F00104000203F0F503F30503F2F503F304F302F503F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F2F302F2F2F2F2F302F30302F2F3F2F30302F303F2F303F2F303F303F3F2F30\
40303F3F2F30402F3F3F2F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F0000000F0F0F000000602F2F5F2F2F602F30
602F2F6000006000006F00006000006F00006000006F3030602F2F6F30306F4F4F6F4F506F4F4F\
6F4F506F4F4F6F50506F4F4F6F4F506F4F4F7050506F4F4F7050506F4F4F705050
6F4F4F7050506F4F4F7050506F4F4F70505070202070202070201F700000700000700000700000\
7000007000007F00007000007F00007000007F00007000007F00007000007F0000
7000007F00007000007F00007000007F00007000007F00007000007F000070302F7F303070302F\
70302F70302F0F0F0F3F2F203F2F206F4F4070504F705040705040705040705040
705040705040705040705040704F3F70504070503F7050406F4F3F7050406F4F3F7050406F4F3F\
70503F6F4F3F6F503F3F4F20404F20104F001F4F001040002F4F0F5F402F604030
5F3F2F5F402F103F00104000103F00303F10503F2F503F2F4F302F4F302F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F302F30302F2F302F30302F303F2F30302F303F303F3F2F303F303F3F2F3040303F\
402F3F40303F3F2F3040303F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F0000000F0F0F5F2F2F602F305F2F2F602F30600000600000600000
6F00006000006F00006000006F00006F00006F0000604F4F6F4F50604F4F6F50506F4F50705050\
6F4F507050506F4F507050506F4F507050506F4F507050507050507050506F5050
7050507050507050507050507050507050506F504F6F4F4F7050507000007F00007000007F0000\
7000007F00007000007F00007F00007F00007F10107F1F107F10107F10107F1010
7F00007F00007F00007F00007F00007F00007F00007000007F00007F00007F00007000007F0000\
7010107F3F3070302F3F2F2F3F2F203F2F206F4F407050406F5040705040705040
7F5F4F7050407050407050407F5F4070504070504070504070504070504070504070503F705040\
70503F7050406F4F3F70503F6F4F3F405020104F001F4F00104F001F4F002F400F
604F305F402F5F40304F3F201F4000103F00303F10503F2F503F30503F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F2F302F302F2F2F302F30302F2F302F30302F303F2F303F2F303F303F3F2F30402F3F3F2F30\
40303F3F2F30402F3F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F000F000000\
0F0F0F5F202F5F2F2F5F202F602F305F00006000005F0000600000600000600000
6000006F00006000006F00006000006F2020601F1F6F4F506F4F506F50506F4F4F6F4F506F4F50\
6F50506F4F4F6F50506F4F507050506F4F4F7050506F4F507050506F4F4F705050
70505070505070504F70505070505070505070504F7050507050507F101070100F7F1010700000\
7F00007000007F00007000007F000070100F7F10107F10107F10107F100F7F5F50
7F10107F10107F100F7F10107F00007F00007000007F00007000007F00007000007F0000700000\
7F000070100F7F10103F2F203F2F203F2F200000000000000000000000006F5040
6F4F406F50406F4F3F6F504070504070504070504070504070504070504070503F70504070503F\
7050406F4F3F70503F6F4F3F6F503F6F4F306F4F3F3F4F201F4F00104000104F00
20400F6040305F402F5F402F403F20104000103F00103F00303F1F503F2F4F302F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F2F
302F302F2F2F302F30302F303F2F30302F303F303F3F2F303F303F3F2F3040303F402F3F40303F\
402F3F40303F402F3F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F0F00000F0F0F0000005F2F305F2F2F5F2F30\
5F2F2F6000005F00006000005F00006000006000006000006000006F0000600000
6F00006000006F00006F00006F20206F2F306F30306F4F506F50506F4F507050506F4F50705050\
6F4F5070505F6F50507050506F505070505F70505070505070505070505F705050
7050507050507F5F5F7050507F5F507050507F5F5F7050507F5F507030307F3F3F7F30307F0000\
7000007F00007F00007F00007F00007F00007F30307F3F307F3030805F5F7F5F50
7F5F507F5F50805F507F5F507F00007F00007F00007F00007F00007F00007F00007F00007F0000\
7F00007F00007F00003F2F2F3F2F203F302F000000000000000000000000000000
0000000000000000006F50407050406F50407050406F50407F5F407050407F5F40705040705040\
70503F70504070503F7050406F4F3F70503F6F4F3F6F503F404F20405020104F00
1F4F001040002F4F105F402F6040305F402F3F4020103F00104000103F00303F204F302F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F2F302F30
2F2F2F302F30302F2F302F30302F303F2F30302F303F2F3F3F2F303F2F3F3F2F3040303F3F2F3F\
40303F3F2F3F40303F000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0000000F000F50202F5F2F2F50202F5F2F2F5F00005F00005F0000\
5F00005F00006000005F00006000005F00006000006000006000006000006F0000
6000006F00006000006F0000602F306F4F506F4F506F50506F4F506F50506F4F507050506F4F50\
7050506F4F507050506F4F507050506F50507050506F5050705050705050705050
705050705050705050705F507050507050507050507F5F50705050705050705050705050700000\
7F00007F00007F00007F00007F00007F00007F00007F00007F00007F30307F3F30
7F30307F5F507F5F507F5F507F5F4F7F5F507F5F4F7F4F407F00007F00007F00007F00007F0000\
7F00007F00007F00003F2F203F2F203F2F20000000000000000000000000000000
0000000000000000000000000000000000000000000000006F4F3F6F50406F4F3F6F5040705040\
7050404050207050406F4F3F70503F6F4F3F6F503F6F4F306F4F3F6F4F30404F20
3F4F1F104F001040001F4F0020400F5F402F5F402F5F4030203F0F103F00103F00303F1F4F302F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F302F2F2F
302F30302F30302F30302F303F303F3F2F303F303F3F2F3040303F3F2F3F40303F402F3F403040\
40303F40303F402F3F000000000000000000000000000000000000000000000000
0F0F0F0F00000F0F0F50202F5F2F3050202F5F2F305000005F00005F00005F00005F0000600000\
5F00006000005F0000600000600000601010600F106F10106010106F0000600000
6F00006F00006F00006F00006F505F6F4F5070505F6F4F5070505F6F505070505F6F505070505F\
70505070505F70505070505F70505070505F705050705F5F705050705F5F705050
7F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7F5F507F5F5F7F5F507F2F2F\
7F20207F2F207F00008000007F00007F00007F00008000007F00008000007F0000
8000007050507F5F507F5F508060507F5F50805F507F5F50804F407F4F407F10107F1010801F10\
7F00007F00007F000040302F3F2F203F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000705040\
6F503F1F5F00405020705F4070503F7050406F503F70503F6F4F3F6F503F6F4F30
6F503F3F4F201F4F00104F001F4F002F400F2F4F0F5F402F5F403020400F104000103F00203F1F\
4F302F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F2F2F2F2F302F30
2F2F2F302F30302F2F302F30302F303F2F303F2F303F2F3F3F2F3040303F3F2F3F40303F40303F\
4F3040402F3F40303F0000000000000000000F000F0000000F0F0F00000050303F
502F3F50303F50202F5000005000005F00005000005F00005F00005F00005F00005F00005F0000\
6000005F00006010105F0F1060101060404F602F30602F306F30306000006F0000
6000006F00006F1F206F20206F4F506F505F6F4F506F505F6F4F5070505F6F4F5070505F6F5050\
70505F6F505070505F70505070505F70505070505F705050705F5F70505070505F
7050507F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7F3030\
7F3F307F30307F00007F00007F00007F00007F00007F00008000007F0000800000
7F00008010107F10107F5F507F5F50805F507F5F507F5F507F5F507F5F507F302F7F3F307F3030\
7F00007F00007F00003F2F203F2F2F3F2F20000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F4F3F1F500040502040502F6F503F70503F6F4F3F6F503F6F4F306F503F
6F4F306F4F3F604F303F4F20104F001F4F001040002F400F5F402F5F4030503F2F20400F103F00\
503F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F2F2F2F302F30302F30
302F30302F303F3030302F303F303F3F2F3040303F3F2F3F40303F3F2F3F40304040303F4F3040\
40303F4F304040303F0000000000000F0F0F4F3F40503F4F503F40503F4F50202F
502F3050202F5F00005000005F00005000005F00005F00005F00005F00006000005F1010601010\
5F0F10601010604050604F50604F506F4F506000006F00006000006F00006F0000
6F00006F202070303F6F505070505F6F505070505F6F505F70505F6F505070505F70505F70505F\
705050705F5F70505F705F5F7050507F5F5F70505F7F5F5F7050507F5F5F70505F
7F5F5F7050507F5F5F7F5F5F7F5F5F7F5F507F5F5F7F5F5F7F5F5F7F5F507F5F5F7F5F5F705F50\
7050507F5F507F00008000007F00008000007F00008000007F0000800000800000
8000007F000080000080101080605F7F5F5080605F7F5F508060507F5F508060507F5F50803F30\
7F00008000007F00004F2F203F2F2040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F50001F5F001F50004F5F2F4050207050406F503F70503F6F503F
6F503F6F4F306F503F604F30404F203F4F1F1F4F001040001F4F0020400F6040305F402F5F4030\
103F004F3F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F302F302F2F2F302F30
302F2F302F30302F303F2F30302F303F2F303F2F3040303F3F2F3F40303F40303F40304040303F\
4F304040303F40303F0000004F3F4F4F3040503F4F4F3040503F4F503F40501F20
50101F501F205000005F00005000005F00005F00005F00005F00005F00005F0F106040505F404F\
604F505F4050604F50602F30602F30602F306F00006000006F00006000006F0000
6F00006F3030604F506F505F6F4F5070505F6F505070505F6F505070505F6F505070505F6F5050\
70505F705050705F5F70505070505F705050705F5F705050705F5F7050507F5F5F
7050507F5F5F7050507F5F5F7050507F5F5F7F5F507F5F5F7F5F507F5F5F7F5F507F5F5F7F5F50\
7F5F5F7F5F508000007F00008000007F00008000007F00008000007F0000800000
7F00008000007F00008000007F3F30803F307F3F3080605F7F5F508060507F5F508060507F5F50\
803F307F3030803F304F201F4F2F204F2020000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F5F001050001F5F001050001F50004050204050206F503F6F503F
6F4F306F503F6F4F306F4F3F604F30604F303F4F1F3F4F20104000104F001040002F400F5F3F2F\
5F402F103F004F302F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F2F302F302F2F30302F30302F30
3F3030302F303F0F103F0F0F3F20203F1F2040303F3F303F40304040303F40304040303F4F3F40\
4F304040303F4F3040503F4F4F3F4F503F4F4F3F4F503F4F503F4F50404F500F10
5F1010500F105F00005000005F00005F00005F00005F404F604F505F4050604F505F4050604F50\
604F50604F505F404F6F00006000006F00006000006F00006F00006F00006F0000
6F4F50604F506F4F506F505F70505F70505F70505F70505F705F5F70505F705F5F70505F705F5F\
70505F705F5F70505F7F5F5F70505F7F5F5F70505F7F5F5F705F5F7F5F5F705F5F
7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F605F7F5F5F7F605F7F5F5F80605F\
7F5F5F80605F7F00008000007F00008000007F0000800000800000800000800000
800000800000800000800000800000800000803F3F805F5080605F80605080605F7F5F5080605F\
805F508060507F5F50502F204F2F20502F20000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F4F001F5F001F5F001F5F001F50001F5F001F500040502F405020
70503F6F503F6F503F6F4F306F503F604F306F4F30604F30404F20104F001F4F001040002F4F0F\
5F402F5F4030103F004F3030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F202F2F2F302F2F2F302F302F2F2F302F30
30202F3F0F0F300F0F3F00003F1F203F20203F2F3F40303F40303F40304040303F40304040303F\
40303F0000004F3F4F4F3040503F4F4F3F40503F4F503F4F503F4F503F4F500F10
500F0F5F10105000005F00005F202F5F2F305F202F5F40505F404F604F505F404F604F505F4050\
604F505F404F6020206000006F00006000006F00006000006F00006000006F0000
6F4F506F505F6F4F5070505F6F505F70505F6F505F70505F70505F70505F70505F70505F70505F\
705F5F70505F705F5F70505F7F5F5F70505F705F5F70505F7F5F5F70505F7F5F5F
705F5F7F5F5F705F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F\
80605F7F5F5F8010107F10108010107F00008000007F00008000007F0000800000
8000008000007F00008000008000008000007F00007F5F507F5F507F60507F5F508060507F5F50\
8060507F5F508060503F2F2040302F3F2F20000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F40001F4F001050001F5F001050001F50001050001F5000
4050206F503F6F4F306F503F6F4F306F503F604F30604F30604F30604F303F401F104F00104000\
104000304010303F1F0F3000103000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F302F2F30302F30302F3030101F300000
3F00003000003F00003F00003F303F3F2F3F40304040303F40304040303F40304040303F000000\
0000000F0F0F4F3F4F503F4F4F3F4F50404F503F4F50404F503F4F5F4050500F10
5F10105000005F00005F00005F40505F404F604F505F4050604F505F4050604F50604F50604F50\
601F206F202F6000006F00006000006F00006F00006F00006F00006F00006F0000
70505F6F505F70505F6F505F70505F70505F705F6070505F705F5F70505F705F6070505F705F5F\
70505F7F5F60705F5F7F5F5F705F5F7F5F60705F5F7F5F5F705F5F7F5F607F5F5F
7F5F5F7F5F5F7F60607F5F5F7F5F5F7F5F5F8060607F5F5F80605F7F5F5F8060607F5F5F80605F\
7F5F5F806060803F3F803F3F7F3F30800000800000800000800000800000800000
800000800000800000800000800000800000800000802F20802F2F7F5F5080605F80605080605F\
80605080605F80605040302F3F2F2F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F4F001F4F001F5F001F50001F5F001F5000
1F500040502040502F6F503F6F503F6F4F306F503F604F306F5030604F306F4F303F4F1F3F4F20\
1040001F4F00304010103F000F3000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F2F2F2F2F302F302F2F2F301010300000300000
3000003F00003F00003F20203F2F3F40303F3F303F40304040303F40303F000000000000000000\
0F000F000000503F4F4F3F4F503F4F503F4F503F4F503F4F50404F500F105F1010
5000005F1F205F1F205F1F205F404F5F40505F4050604F505F4050604F505F4050602F30602F30\
6000006000006F00006000006F00006000006F00006000006F10106F10106F1010
6F4F5F6F505F6F505F70505F6F505F70505F70505F705F5F70505F705F5F70505F705F5F70505F\
705F5F70505F7F5F5F70505F7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F
705F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F605F7F5F5F7F605F7F5F5F80605F7F5F5F\
80605F7F5F5F803F3F7F3F30803F3F7F00008000007F0000800000800000800000
800000800000800000800000800000800000800000802F20802020802F2080605080605F7F5F50\
80605F80605080605F3F2F2040302F3F2F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000001040001F4F001050001F5F00
1050001F50001050004050203F4F206F503F6F4F306F503F604F30604F30604F30604F3060402F\
3F4F1F20400F104F00503F2F103F000F3000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000302F302F2F2F302F302F0F0F3000003000003F0000300000
3F00003F1F203F20203F303F40304040303F4030403F303F403040000000000000000000000000\
0000000F0F0F4F3F4F504050503F4F504050503F4F5F405050404F5F0000500000
5F2F30503F4F50404F503F4F604F505F4050604F505F4050604F5F604F50603030602F306F1010\
6000006F00006000006F00006F00006F00006F00006F00006F10106F10106F2020
7050606F505F7050606F505F705F6070505F705F6070505F705F60705F5F705F6070505F7F5F60\
705F5F7F5F60705F5F7F5F60705F5F7F5F60705F5F7F5F607F5F5F7F5F607F5F5F
7F60607F5F5F7F60607F5F5F8060607F5F5F8060607F5F5F8060607F5F5F8060607F5F5F806060\
80605F806060803F3F80403F803F3F800000800000800000800000800000800000
8F00008000008000008000008F000080000080000080000080000080000080605F7F5F5080605F\
80605F80605F80605040302F40302F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F4F001F4000
1F5F001F50001F5F001050001F50004050204050206F4F306F503F6F4F306F503F604F306F5030\
604F30604F30204F0F1F4F00503F2F103F000F3000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F2F2F2F302F00003000003000003000003000003F0000
301F203F303F3F2F3F3F303F3F303F4030403F2F3F000000000000000000000000000000000000\
0F0F0F000000503F4F503F4F504050503F4F504050503F4F5F405050202F5F2F30
50404F5F40505F404F5F40505F4050604F505F4050604F505F4050604F505F0F10601010600F10\
6000006000006F00006000006F00006000006F00006F00006F202F6F1F206F2020
6F505F70505F6F505F70505F6F505F705F6070505F705F6070505F705F6070505F705F6070505F\
7F5F6070505F7F5F60705F5F7F5F60705F5F7F5F60705F5F7F5F60705F5F7F5F60
7F5F5F7F60607F5F5F7F5F607F5F5F7F60607F5F5F7F60607F5F5F8060607F5F5F8060607F5F5F\
8060607F5F5F803F3F803F3F803F3F800000800000800000800000800000800000
8000008000008000008000008000008000008000008000008000008000007F5F507F60507F5F50\
80605F80605080605F3F302F40302F3F2F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
1040001F4F001050001F50001050001F5000104F004050206F4F306F503F604F30604F30604F30\
604F30604F2F604F304F401F1040004F3F2F1030000F2F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F2F302F302F0000300000300000300000300F0F3F2F303F202F
3F303F3F303F4030403F303F3F303F3F303F000000000000000000000000000000000000000000\
0000000F0F0F503F4F504050503F4F5F40505040505F40505040505F40505F4050
5F40505F4050604F5F5F4050604F5F5F4050604F5F5F4050604F505F40506000006000006F0000\
6000006F00006000006F00006F00006F00006F00006F00006F303070303F6F303F
7050606F505F705F6070505F705F6070505F705F60705F60705F60705F5F7F5F60705F607F5F60\
705F5F7F5F607F5F607F5F607F5F5F7F60607F5F607F60607F5F5F7F60607F5F60
7F60607F5F5F8060607F5F608060607F5F5F8060607F60608060607F605F806060806060806060\
80605F80606080504F80505080504F8000008000008000008000008F0000800000
8F00008000008F00008000008F00008000008F00008000008000008000008F403F803F3080403F\
80605F8F6F5F80605F40302F3F302F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000001F40001F4F001F50001F5F001050001F5F001F50004050204050206F503F6F4F306F503F\
604F306F5030604F30604F301040002F400F4F3F2F103000202F10000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F202F2F1F1F2F00003000002F0000301F20300F0F30303F302F303F303F
3F2F3F3F303F3F303F3F303F000000000000000000000000000000000000000000000000000000\
0F000F4F3F4F504050503F4F504050503F4F5F4050503F4F5F40505040505F4050
5F40505F40505F4050604F505F4050604F505F4050602F305F2F30602F30600000600000600000\
6000006000006F00006000006F00006F00006F00006F00006F303F6F303070303F
6F505F7050606F505F705F6070505F705F6070505F705F6070505F705F60705F5F7F5F60705F5F\
7F5F60705F5F7F5F60705F5F7F5F607F5F5F7F5F607F5F5F7F5F607F5F5F7F6060
7F5F5F7F60607F5F5F7F60607F5F5F7F60607F5F5F8060607F5F5F8060607F605F8060607F605F\
80606080605F805050804F4F80504F8000008000008000008000008000008F0000
8000008000008000008F0000800000800000800000800000800000800000802F20802F2F802020\
80605F80605F80605F3F2F2F40302F3F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0000000000000000000000000000
0000000000000000001F4F001040001F50001050001F50001050001F50003F4F20405020604F30\
604F30604F30604F30604F2F604F3010400050402F4F302F10300F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000301F202F000030101F301010301F20302F303F303F30303F3F303F3F303F
4030403F303F3F303F0000000000000000000000000000000000000000000000000000000F0F0F\
000000504050503F4F5F40505040505F40505040505F40505F40505F40505F4050
604F5F5F4050604F5F5F4050604F5F604F50604F5F602F306000006000006000006000006F0000\
6000006F00006000006F00006F00006F00006F00007000006F000070303F6F404F
705F60705060705F60705060705F60705F60705F60705F607F5F60705F607F5F60705F607F6060\
7F5F607F5F607F5F607F60607F5F607F60607F5F607F60607F5F607F60607F5F60
8060607F60608060607F60608060607F60608060607F6060806060806060806060806060806F60\
80606080606080504F8F505080504F8000008000008F00008000008F0000800000
8F00008000008F00008000008F00008000008F00008000008F00008000008F0000802F208F2F2F\
80605F8F6F5F80605F40303040302F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000002F00002F001F50001F4F001F50001F4F00002F00002000
0000000000000000000000000000001F40001F4F001050001F5F001F50001F5000105000405020\
2F500F6F503F604F306F5030604F301F4F0010400050403040302F303020000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F00002F000F302F302F2F30302F30302F30302F3F302F303F303F302F3F3F303F
302F3F000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F503F4F504050503F505F4050503F505F40505040505F40505040505F4050
5F40505F4F5F5F4050604F5F5F40505F4F505F4050600000600000600000600000600000600000\
6F00006000006F00006000006F00006F00006F00006F00007000006F404F70404F
6F505F705F606F505F705F60705060705F60705060705F60705F607F5F60705F607F5F60705F60\
7F5F60705F607F5F607F5F607F60607F5F607F60607F5F607F60607F5F607F6060
7F5F608060607F5F608060607F60608060607F60608060607F60608060607F6060806060806060\
80606080606080505080504F8050508000008000008000008F00008000008F0000
8000008F00008000008F00008000008F00008000008000008000008F0000800000800000800000\
8F6F5F80605F806F5F3F302F40302F3F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001F000010000000001F4F001F50001F5F001F60001F5F001F60001F5F001F5F00
0020000000000000000000000000000000000000001F40001040001F50001050001F5000105000\
1F50002F4F0F604F30604F30604F3020400F104F0020400F503F2F40302F3F2F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F0000\
2F0000302F302F2F30302F3F302F303F303F302F3F3F303F3F2F3F3F303F000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005040505040505F40505040505F405F5F40505F405F5F40505F4F5F5F4050
604F5F5F4050604F5F5F4F5F604F5F602F3060303F6000006000006000006F00006000006F0000\
6000006F00006F00006F00006F00007000006F00007000006F0000704F5070404F
705F60705060705F60705F60705F60705F607F5F6F705F607F5F60705F607F606F7F5F607F6060\
7F5F607F606F7F5F607F60607F5F607F606F7F60607F60607F5F608060607F6060
8060607F60608060607F60608060607F6060806060806060806060806060806F60806060806F60\
8060608F6F608050508F505080504F8F00008000008F00008000008F00008F0000
8F00008000008F00008000008F00008000008F00008000008F00008000008F00008000008F0000\
80605F8F6F60806F5F40303040302F403F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001000001F000010001F60001F50002060001F60001F60001F60001F60001F5F00
1F60001F3F001F3F000000000000000000000000000000000000001040001F5F001F50001F5000\
1050001F50001050001F50002F4F0F1F5000104F00104F0050402F503F3040302F
3F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F0000\
2F1F1F2F2F30302F302F2F30302F3F302F303F303F302F303F2F3F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F000F503F50504050503F505F40505040505F40505040505F40505F40505F4F5F
5F40505F4F5F5F4050604F5F5F2F30602F3F602F306000006000006000006000006F0000600000\
6F00006000006F00006000006F00006F00006F00006F00007000006F404F704F50
6F505F705F60705060705F60705F60705F60705F607F5F60705F607F5F60705F607F5F60705F60\
7F5F607F5F607F60607F5F607F60607F5F607F60607F5F607F60607F5F60806060
7F5F608060607F60608060607F60608060607F60608060607F6060806060806060806F60806060\
80606080606080505080504F8050508000008F00008000008F00008000008F0000
8000008F00008000008F00008000008F00008000008F00008000008F00008000008F0000800000\
8F6F5F80605F8F6F5F3F302F40302F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001F00001000001F001F60001F60001F60001F60001F60001F60001F5F001F6000
1F5F001F60001F5F000000000000000000000000000000000000000000000000001F4000105000\
1F50001050001F5000104F001F5000104F00104F0010400030401F503F2F4F3F2F
3F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F0000302F30\
2F2F30302F30302F3030303F302F3F3F303F302F30000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5040505040505F405F5F40505F405F5F40505F4F5F5F40505F4F5F5F4050
604F5F5F4F5F604F5F5F4F5F60202F60202060202F6000006F00006000006F00006000006F0000\
6F00006F00006F00006F00006F00007000006F00007000006F0000704F5070404F
704F50705F60705F60705F607F5F6F705F607F5F6F705F607F606F7F5F607F606F7F5F607F606F\
7F5F607F606F7F5F607F606F7F60607F606F7F606080606F7F606080606F7F6060
80606F80606080606F7F6060806F6F806060806F6F806060806F6F806060806F6F8060608F6F6F\
8060608F6F6F8050508F5F508050508F00008000008F00008F00008F00008F0000
8F00008F00008F00008F00008F00008F00008F00008000008F00008F00008F00008000008F0000\
806F5F8F6F60806F5F403F3040302F403030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001000001F00001000206F001F60001F6F001F60001F60001F60001F60001F6000
1F60001F60001F6000000000000000000000000000001000001F000000000000000000001F4F00\
1F40001F50001050001F50001050001F5000104F001F4F001040005F40304F3F2F
4F3F303F2F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F00002F2F2F2F2F2F\
302F302F2F30302F3F302F30302F3F302F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F503F505040505040505F405F5040505F405F5F40505F4F5F5F40505F4F5F
5F4050604F5F5F4F50604F5F5F1F206000006000006000006000006F00006000006F0000600000\
6F00006000006F00006F00006F00006F00006F00006F00007000006F404F704F50
70404F705F60705F60705F60705F607F5F6F705F607F5F60705F607F606F705F607F60607F5F60\
7F606F7F5F607F60607F5F607F606F7F5F607F60607F606080606F7F6060806060
7F606080606F7F60608060607F606080606F7F6060806060806060806F6F806060806F60806060\
806F6F80606080403F803F3F8F40408000008F00008000008F00008000008F0000
8F00008F00008000008F00008000008F00008000008F00008000008F00008010108F1F10801010\
8F6F6080605F8F6F6040302F4030303F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001F00001000001F001F60001F6F001F60001F60001F60001F60001F60001F6000
1F60001F60001F5F00000000000000000000000000000000105F00001000001000000000000000\
0000001040001F40001050001F5000104F00104F00104F00104F0030401F50402F
4F3F2F40302F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F302F0000302F30\
2F1F1F302F3F302F30302F3F000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005F405F5040505F405F5040505F4F5F5F405F5F4F5F5F405F604F5F5F4F5F
604F5F5F4F5F604F5F604F5F6000006000006F00006000006F00006000006F00006F00006F0000\
6F00006F00006F00006F00006F00007000006F00007000006F0000703F4070303F
703F40705F607F5F6F705F607F606F705F607F606F7F5F607F606F7F5F607F606F7F60607F606F\
7F5F607F606F7F60607F606F7F606080606F7F606080606F7F6060806F6F806060
80606F806060806F6F806060806F6F806060806F6F806060806F6F8060608F6F6F806F608F6F6F\
806F608F6F6F80403F8F404080403F8F00008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008F00008F00008F00008F00008F00008F00008F302F8F2F2F8F2F2F\
8F6F608F6F608F6F60403F3040302F403F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000001F000010001F6F001F6000206F001F60001F6F001F60001F6F00002F00
002F001F3F001F3F00001000001F000000000000000000001F5F001F5F001F5F00001000000000\
0000000000000000001F4F001050001F5000104F003F501F204F105F402F50402F
5040304F3F2F403030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F2F2F2F2F2F2F302F1010\
302F302F2F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5040505F405F5040505F405F5F40505F4F5F5F40505F4F5F5F405F604F5F
5F405F604F5F5F4F5F604F5F5F00006000006000006000006000006F00006000006F0000600000\
6F00006F00006F00006F00006F00006F00007000006F000070000070303F703F40
70303F705F60705F607F5F6F705F607F5F6F705F607F606F705F607F606F7F5F607F606F7F5F60\
7F606F7F60607F606F7F5F607F606F7F606080606F7F606080606F7F606080606F
7F606080606F806060806F6F7F6060806F6F806060806F6F806060806F6F8060608F6F6F806060\
8F6F6F806F608F2F2F802F2F8F2F2F8000008F00008000008F00008F00008F0000
8F00008F00008F00008F00008000008F00008F00008F00008000008F00008F403F8F2F2F802F20\
8F6F60806F5F8F6F6040302F40303040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F60001F6F001F60001F6F001F60001F60001F60000F0F00
0020000000000000000F0000000000001000000000000000105F001F5F00105F001F5F00001000\
0010000000000000000000001F40001040001F5000104F005F4F305F402F5F4F30
50402F503F2F40302F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F2F3000002F1010301F20\
302F30000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5F405F5040505F4F5F5F405F5F4F5F5F405F604F5F5F4F5F604F5F5F4F5F
604F60604F5F604F60604F5F6000006000006F00006000006F00006000006F00006F00006F0000\
6F00006F00006F00007000006F00007000006F0000700000700000703F4070303F
7F3F40705F607F606F705F607F606F7F5F6F7F606F7F5F6F7F606F7F606F7F606F7F606F7F606F\
7F606F7F606F7F606F80606F7F606F80606F7F606F806F6F80606F806F6F80606F
806F6F80606F806F6F80606F806F6F806F6F806F6F80606F8F6F6F806F6F8F6F6F806F6F8F6F6F\
806F6F8F6F6F8010108F1F1F8F1F108F00008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008F00008F00008F00008F00008F00008F00008F40408F403F8F4040\
8F6F608F70608F6F60403F30403030403F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000206F001F6F001F6F001F6000206F001F60001F6F00000000
0F0F0F0000000000000000000F0F0F0000000000000000001F60001F5F001F60001F5F001F5F00\
1F5F000000000000000000000000000000001040001F5000104F00604F305F402F
5F4F3050402F4F3F3040302F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F2F2F00002F0000302F302F2F30\
302F3F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5040505F405F5040505F4F5F5F40505F4F5F5F405F5F4F5F5F405F604F5F
5F4F5F604F5F5F4F5F604F5F600000601010600F106000006000006F00006000006F00006F0000\
6F00006F00006F00006F00007000006F00007000006F000070000070202F70202F
70202F705F6F705F607F5F6F705F607F606F705F607F606F7F5F6F7F606F7F5F607F606F7F606F\
7F606F7F60607F606F7F606F80606F7F606080606F7F606F80606F7F606080606F
80606F806F6F806060806F6F80606F806F6F806060806F6F80606F8F6F6F8060608F6F6F806F6F\
8F6F6F7F60608F00008000008F00008000008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008F00008F00008F00008F00008F00008F000080605F806F60806F5F\
8F7060806F608F6F6040302F403F3040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F3F001F40001F60001F6F001F60001F6F001F6000000000
0000000000000000000000000000000000000000000000001F5F001F6000105F001F5F00105F00\
1F5F001050001F5F000000000000000000000000001050001F5000304F1F5F4F2F
50402F50402F4F3F2F4F3F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F2F302F302F0F0F30202F302F30302F3F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005F4F5F5F405F5F4F5F5F405F5F4F5F5F4F5F604F5F5F4F5F604F605F4F5F
604F60604F5F605060604F5F60101F6010106F101F6000006F00006000006F00006F00006F0000\
6F00007000006F00007000006F0000700000700000700000700000702F2F701010
7F1F1F705F6F7F606F705F6F7F606F7F5F6F7F606F7F606F7F606F7F606F80606F7F606F7F606F\
7F606F806F6F7F606F80606F7F606F806F6F80606F806F6F80606F806F6F80606F
806F6F80606F806F6F806F6F806F6F806F6F8F6F6F806F6F8F6F6F806F6F8F6F6F806F6F806F60\
7F6060806F6F8F00008F00008F00008F00008F00008F00008F00009000008F0000
9000008F00009000008F00008F1F1F8F00009000008F2F2F8F302F8F2F2F8F706F8F6F608F6F60\
8F6F608F706F8F6F60403F304030304F3F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000001F001F4000206F001F6F001F6F001F6000206F00002F00
002F000000000000000000000000000000000000000000001F60001F5F001F60001F5F001F6000\
1F5F001F60001F5F001F5F00103000000000000000000000104F001F500050402F
5F4F3050402F504030503F2F4F3F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F2F2F202F30202F2F2F30302F3F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5040505F4F5F50405F5F4F5F5F405F5F4F5F5F405F5F4F5F5F4F5F604F60
5F4F5F604F60604F5F605060601F2060202F6020206F00006000006F00006000006F00006F0000\
6F00006F00007000006F00007000006F00007000006F000070000070101070101F
700000704F50704F507F4F5F705F6F7F606F7F606F7F606F7F5F6F7F606F7F606F7F606F7F606F\
7F606F7F606F80606F7F606F80606F7F606F806F6F7F606F806F6F80606F806F6F
80606F806F6F80606F806F6F80606F806F6F806F6F8F6F6F806F6F8F6F6F806F6F8F6F6F802F2F\
8F2F2F802F2F8F00008000008F00008F00008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008F10108F1F1F8F10108F6F60806F608F6F608F6F608F706F8F6F60\
8F70608F6F608F706040302F403F3040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000001000001F001F6F001F6F001F60001F6F001F50001F5000
002F00000000000000000000000000000000000000000000105F001F60001F5F001F6000105F00\
1F5F00105F001F5F00105F001F5F00103000000000000000104000104F004F4F2F
5F402F5F4F2F50402F50402F4F3F2F40302F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F2F2F2F302F30302F30302F3F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5F4F5F5F405F5F4F605F4F5F5F4F605F4F5F604F605F4F5F604F605F4F5F
605060604F60605060604F606F303F60303F6F303F6000006F00006F00006F00006F0000700000\
6F00007000006F0000700000700000700000700000700000700000700000700000
7F0000704F507F1F1F7010107F60707F606F7F60707F606F8060707F606F8060707F606F806F70\
7F606F806F707F606F806F7080606F806F7080606F806F70806F6F806F7080606F
806F70806F6F806F70806F6F8F6F70806F6F8F6F70806F6F8F70708040408F40408040408F0000\
8F00008F00008F00008F00008F00008F00008F00009000008F00009000008F0000
9000008F00009000008F0000905F5F8F5F50905F5F8F6F6F8F706F8F6F6090706F8F6F6090706F\
8F70608F706F8F6F604F3F304030304F3F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000206F001F6F00206F001F6F001F6F001F5000
2050000000000000000000000000000000000000000000001F60001F60001F60001F5F001F5000\
1F4F001F60001F5F001F60001F5F001F5F000020000000000000001F4000001000
5F4F305F4F2F5F4F3050402F5040304F3F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F202F2F2F302F2F30302F302F2030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F50405F5F4F5F5F405F5F4F5F5F405F5F4F5F5F4F5F604F605F4F5F604F60
604F5F605060604F5F60506060303F6F303F6050606F00006000006F00006000006F00006F0000\
6F00006F00007000006F00007000006F0000700000700000700000700000700000
7000007F101F7010107F101F7F3F407F3F407F3F407F606F7F606F7F60707F606F7F606F7F606F\
806F707F606F80606F7F606F806F707F606F806F6F80606F806F7080606F806F6F
80606F806F7080606F806F6F806F6F8F6F70806F6F8F6F6F806F6F8F4040803F408F4040800000\
8F00008000008F00008F00008F00008F00008F00008F00009000008F00008F0000
8F0000905F5F8F5F508F5F5F8F6F6F90706F8F5F508F706F8F6F608F706F8F6F608F706F8F6F60\
8F706F8F6F608F7060403030403F3040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000002F00002F001F6F001F6F001F60001F6F00
1F60000F0F000000000000000000000000000000000000001F60001F6000105F001F60001F4F00\
1F4F00104F00000000105F001F5F00105F001F5F00000000000000000000001000
00000050402F50402F50402F4F402F50402F40302F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F302F302F2F30302F30302F30000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005F4F5F5F4F5F5F4F605F4F5F604F605F4F5F604F605F4F5F605060604F60
605060604F606F50606050606F50606050606F506F6F303F6F30406F303F6F00006F00006F0000\
6F00007000006F00007000007000007000007000007000007000007F0000700000
7F00007000007F00007F00007F3F407F00008000007F2F2F802F307F2F2F7F606F7F606F7F606F\
7F606F806F7080606F806F7080606F806F70806F6F806F70806F6F806F70806F6F
806F70806F6F8F6F70806F6F8F6F708040408F404F8F40408F40408000008F00008F00008F0000\
8F00008F00008F00008F00008F0000901F1F8F1F1F901F1F8F1F108F706F8F6F6F
8F706F8F706F9070708F706F90706F8F706F90706F8F706F90706F8F706F90706F8F706F90706F\
8F706090706F8F70604F3F304030304F3F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000002F00206F001F6F00206F001F6F00
1F6F001F6000206F000000000000000000000000000000001F60001F60001F60001F60001F5000\
1F4F001F50000000000F0F0F0020001F60001F5F001F5F00000000000000000000
001000000000504F2F50402F5040304F402F504030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F2F2F2F302F302F2F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F50405F5F4F5F5F405F5F4F605F4F5F604F605F4F5F604F605F4F5F605060
5F4F5F605060604F606050606050606F50606050606F30406050606F50606F10106F101F6F1010\
6F00006F101070101F6F00007000007000007000007000007000007000007F0000
7000007F00007000007F00007F00007F00007F00007F2F307F202F7F2F2F7F606F7F606F7F606F\
806F707F606F802F307F3F40804040803F40804040803F40804040803F40802F30
802F2F802F2F8000008F00008000008F00008000008F00008000008F00008F00008F0000801010\
8F1F1F8F10108F5F5F8F5F508F5F5F8F6F6F9070708F6F6F9070708F706F907070
8F706F90706F8F706F90706F8F706F90706F8F706F90706F8F706F8F706F8F706F90706F8F7060\
8F706F8F706090706F403030403F30403030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000002F001F6F001F6F001F6F00
1F60001F6F001F6000002F00002F000000000000000000001F60001F60001F60001F6000104F00\
1F4F001F4F00000000000000002F000020001F5F00105F001F4F00000000000000
0000000000000000005F4F3050402F50402F4F3F2F4F3F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F302F2F2F302F302F2F30302F3F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5F4F605F4F5F604F605F4F60604F605F4F60605060604F60605060604F60
6050606050606F50606050606F506F6050606F5F6F6F50606F5F6F6F50606F506F6050606F5F6F\
6F101070101F6F5060701F1F701010701F1F7000007F0000702F2F7F0000700000
7F00007F00007F00007F10107F00007F00008000007F00008000007F00008000007F0000800000\
802F2F802F30802F2F80404F8040408040408040408F404F804040804040802F2F
8F3030802F2F8F00008000008F00008F00008F00008F00008F00008F00008F00008F00008F1F1F\
8F1F1F8F1F1F8F5F5F905F5F8F5F5F9070708F706F9070708F70709070708F706F
907F7090706F9070708F706F907F708F706F9070708F706F907F708F706F90706F8F706F90706F\
8F706F90706F8F706F4F3F3F403F304F3F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000003000002F001F6F001F6F00
206F001F6F001F6F001F6F001F6F00001000001F000000001F6F001F60001F60001F60001F5000\
1F4F001F50000000000000000000000000001F5F001F60001F5F000F0F00000000
0000000000000000000000005F4F3050402F5040304F3F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F2F302F302F2F302F2F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F50405F5F4F605F4F5F5F4F605F4F5F604F605F4F606050605F4F60605060
604F60605060604F606050606050606F506F6050606F506F6F50606F5F6F6050606F5F6F6F5060\
6F50606050606F50606F1010705F6F6F5F60702F2F70202F7F2F307000007F4F5F
704F507F1F1F7010107F101F7F000070606F705F607F3F407F3F408040407F3F408040407F3F40\
801F1F7F1010801F1F801010801F1F801010801F1F801010801F1F801010801F1F
8010108F1F1F803F408F40408040408F40408040408F40407F606F806F6F7F6060806F6F8F6F6F\
8F70708F6F6F8F70708F706F9070708F706F9070708F706F9070708F706F907070
8F706F9070708F706F9070708F706F9070708F706F90706F8F706F90706F8F706F90706F8F706F\
90706F8F706090706F4030304F3F30403030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F6F001F6F00
1F6F001F6F001F60001F6F001F60001F6F00001000001F001F60001F6F001F60001F60001F4F00\
1F5000104F00000000000000000000000000001000105F001F5F00104000000000
00000000000000000000000050402F5040304F402F4F402F4F3F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F2F302F302F2F30302F3F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005F4F605F4F60604F605F4F60605060604F60605060604F6060506F605060
60506F6050606F506F6050606F506F6050606F5F6F6F506F6F5F6F6F506F705F6F6F5F6F705F6F\
6F5F6F70606F705F6F705F6F6F5F6F706070705F6F7F6070705F6F7F505F704F5F
7F505F7F606F7F6F707F607070606F705F6F7F606F7F3F4080404F7F3F4080404F803F4080404F\
80101F801F1F80101F801F1F80101F8F1F1F80101F8F1F1F80101F8F1F1F801F1F
8F1F1F80101F8F404F8F40408F404F8F40408F4F4F8F6F708F70708F6F708F70708F70708F7070\
8F70709070708F70709070708F7070907070907070907070907070907F70907070
907F7090706F907F7090706F907F708F706F907F7090706F907F708F706F907F7090706F907F6F\
8F706F907F6F90706F4F3F30403F304F3F3F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F000000
1F6F001F6F00206F001F6F001F6F001F6F001F6F001F6F001F6F001F60001F6F001F60001F5000\
1F50001F5000000000000000000000000000000000001F00105F001F5F00000000
00000000000000000000000000000050402F504F304F402F4F4030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F2F2F2F302F202F302F302F2030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F5F5F4F605F4F5F5F4F605F4F606050605F4F60605060604F60605060
604F606050606050606F506F6050606F506F6050606F5F6F6F50606F5F6F6F506F705F6F6F5F6F\
705F6F6F5F6F705F6F6F5F6F705F6F705F6F706070705F6F706070705F6F7F6070
70606F7F60707F606F7F60707F606F7F60707F6070806F707F6070806F707F6070806F707F606F\
806F70806F70806F70806F6F806F70806F70806F70806F6F806F70806F708F6F70
806F6F8F6F70806F708F7070806F6F8F70708F6F708F70708F6F6F8F70708F6F708F70708F6F6F\
8F70708F70709070708F706F9070708F70709070708F706F907070907070907F70
8F706F9070708F706F9070708F706F9070708F706F9070708F706F90706F8F706F90706F8F706F\
90706F8F706F90706F4030304F3F30403F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000002F001F6F001F6F001F6F001F6F001F60001F6F001F60001F6F001F60001F6F00104F00\
1F50001F4F000000000000000000000000000000000000001F5F00105F00002000
0000000000000000000000000000000000004F402F50402F4F3F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F302F2F2F302F302F2F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5F4F605F4F606050605F4F606050605F4F6060506F60506060506F605060
6F506F6050606F506F6050606F5F6F6F506F6F5F6F6F506F705F6F6F5F6F705F6F6F5F6F705F70\
705F6F706070705F6F706070705F6F706070705F6F7F60707060707F6070706070
7F60707F60707F6F707F6070806F707F6070806F707F6070806F70806F70806F70806F70806F70\
806F70806F70806F708F6F70806F708F6F70806F708F7070806F708F7070806F70
8F70708F6F708F70708F6F708F70708F70708F70708F70709070708F70709070708F7070907070\
8F70709070708F7070907F70907070907070907070907F70907070907F70907070
907F70907070907F70907070907F70907070907F708F706F907F7090706F907F708F706F907F70\
90706F907F6F403F304F3F3F403F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000030001F5000206F001F6F001F6F001F6F001F6F001F6F001F6F001F60001F5000\
1F50000000000000000000000000000000000000000000001F4F001F5F00105F00
002F0000000000000000000000000000000000000050402F504030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F202F2F2F302F2F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F5F5F4F605F4F606050605F4F60605060604F6060506F604F6060506F
6050606F506F6050606F506F6050606F5F6F6050606F5F6F6F506F6F5F6F6F5F6F6F5F6F6F5F6F\
705F6F6F5F6F705F6F705F6F706070705F6F706070705F6F7F6070705F6F7F6070
7060707F607070606F7F60707F60707F6F707F60707F6F707F6070806F707F6070806F70806F70\
806F70806F70806F70806F70806F70806F70806F70806F708F7070806F708F6F70
806F708F7070806F708F70708F6F708F70708F6F708F70708F70708F70708F6F708F70708F7070\
9070708F70709070708F70709070708F7070907070907070907F70907070907F70
907070907F708F706F907F708F706F907F708F706F907F708F706F907F708F706F907F6F8F706F\
907F6F8F706F4F3F30403F304F3F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000001F50001F6F001F6F001F6F001F6F001F6F001F60001F6F001F6000\
1F6F00002000002F000000000000000000000000000000001F4F00105F001F5F00
1050000000000000000000000000000000000000000000004F402F4F402F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F302F2F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000006050605F4F6060506F5F4F6060506F60506060506F6050606F506F60506F
6F506F60506F6F5F6F6F506F6F5F6F6F506F6F5F6F6F5F6F6F5F6F6F5F6F705F706F5F6F705F70\
6F5F6F706070705F6F706070705F6F7F60707060707F60707060707F60707F6070
7F60707F60707F6F707F60707F6F707F6070806F7F7F6F70806F7F7F6F70806F7F806F70806F7F\
806F708F707F806F708F6F7F806F708F707F806F708F707F806F708F707F8F6F70
8F70708F6F708F707F8F70708F70708F707090707F8F70709070708F707090707F8F7070907070\
8F7070907F7F907070907F70907070907F7F907070907F709070709F7F7F907070
907F70907070907F70907070907F70907070907F70907070907F7090706F907F7090706F907F70\
90706F907F70403F304F3F3F403F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000001F00000000000000003000000000001F00000000000000000000
0000000000000000000000000000000000002070001F6F001F6F001F6F001F6F001F6F001F6F00\
1F60001F6F001F6000002F001F3F00000000002F000000001F4F001F60001F5F00
1F5F0000000000000000000000000000000000000000000050402F4F402F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F2F2F2F2F2F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F606050605F4F606050605F4F6060506F604F6060506F60506060506F
6050606F506F60506F6F5F6F60506F6F5F6F6F506F6F5F6F6F506F6F5F6F6F5F6F705F706F5F6F\
705F6F6F5F6F706070705F6F706070705F6F706070705F6F7F60707060707F6070
7060707F60707F60707F6F707F60707F6F707F6070806F707F6070806F707F6F70806F70806F70\
806F70806F70806F70806F708F6F70806F708F7070806F708F7070806F708F7070
806F708F707040303F403F3F403030403F3F40303F403F3F403030403F3F40303F4F3F3F403030\
4F3F3F403F3F4F3F3F4030304F3F3F403F3F4F3F3F403F304F3F3F403F3F4F3F3F
403F30907F70907070907F708F706F907F70907070907F708F706F907F708F706F907F708F706F\
907F708F706F4F3F3F403F304F3F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000001F00001000001F00002F000030001F7000001F001F5F00000000
001000001F000000000000000000000000000000000000001F6F001F6F001F6F001F6F001F6000\
1F6F001F60001F6F001060001F3F001F3F001F60000020001F6000105F001F5F00
105F000000000000000000000000000000000000000000000000004F402F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F2F2F30000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F0F0F0F5F4F6060506F60506060506F60506060506F60506F60506F60506F
6F5F6F60506F6F5F6F60506F6F5F6F6F5F6F6F5F6F6F5F6F705F706F5F6F705F706F5F6F706070\
705F6F706070705F6F7060707060707060707060707F60707060707F6070706070
7F6F7F7F60707F6F7F7F6070403F3F3F303F40303F3F303F403F3F3F303F403F3F3F303F403F3F\
40303F403F3F40303F403F3F40303F403F3F40303F403F3F40303F403F3F40303F
4F3F3F403F3F4F3F3F40303F4F3F3F403F3F4F3F3F403F3F4F3F3F403F3F4F3F3F403F3F4F3F3F\
403F3F4F3F3F403F3F4F3F3F4F3F3F4F3F3F4F3F3F4F3F3F4F3F3F4F3F3F4F3F3F
4F403F4F3F3F4F3F3F403F3F4F3F3F4F3F3F907F70907070907F70907F70907F7090706F907F70\
907F6F907F70403F304F3F3F403F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001F00\
001F000000001F5F00206000001F00207F001F7000207F001F70002070001F5F00
207000001F00001F00002F000000000000000000000000000000000000001F6F001F6F001F6F00\
1F6F001F6F001F60001F6F001F60001F6F001F60001F60001F60001F6000105F00
1F6000001000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F0000006050605F4F6060506F5F4F6060506F60506060506F60506060506F
60506F6F5F6F60506F6F5F6F60506F6F5F6F6F506F6F5F6F6F5F6F6F5F706F5F6F705F706F5F6F\
7060706F5F6F706070705F6F706070705F6F7060707060707F60707060707F6070
3F30303F303F3F30303F303F3F303F40303F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F3F3F403F304F3F3F403F304F3F3F403F304F3F3F403F304F3F3F8F706F\
907F708F706F4F3F3F403F304F3F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001F00\
001F00001000207F001F5F00207F001F70001F70001F70002070001F70001F7000
1F70001F70001F6F00002F00002F001F40000000000000000000000000000000001F6F001F6F00\
1F6F001F60001F6F001F60001F6F001060001F60001060001F6000105F001F5F00
105F00002F00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F202F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
00000F0F0F0F5F506060506F60506060506F60506F60506F60506F6F5F6F60506F
6F5F6F60506F6F5F706F5F6F6F5F706F5F6F6F5F706F5F6F705F706F5F6F7060706F5F70706070\
705F707060707060707060707060707F60707060703F303F3F303F3F303F3F303F
3F303F3F303F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004F3F3F403F304F403F4F3F3F4F3F3F\
403F304F403F403F304F3F3F403F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001F00\
001F00001F001F7F00207F001F7000207F001F7000207F001F7000207F001F7000
2070001F70002070001F70002070001F4000204F001F5F00000000000000000000000000000000\
1F6F001F5F001F50001F6F001F60001F6F001F60001F60001F60001F60001F5F00
1F6000002000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
20202F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000060506F5F4F6060506F60506060506F60506060506F60506F6F5F6F
60506F6F5F6F60506F6F5F6F60506F6F5F6F6F5F6F6F5F706F5F6F6F5F706F5F6F7060706F5F6F\
706070705F70706070302F303F303F3F2F303F303F302F303F303F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F3F403F304F3F3F403F304F403F4F3F3F4F403F4F3F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001000\
001F00001F00207F001F70001F7F001F7000207F001F70001F70001F7000207000
1F70001F70001F70001F70001F6F001F70001F6F001F5F001F50001F5F00000000000000000000\
0000000000001F50001F50001F6F001F60001F6F001060001F60001060001F6000
002000002000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F000F0F0F0F5F506060506F60506F60506F60506F605F6F60506F6F5F6F60506F
6F5F706F5F6F6F5F706F5F6F6F5F706F5F6F6F5F706F5F6F7060706F5F70706070302F303F303F\
3F303F3F303F3F2F3F3F303F3F303F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F403F4F3F3F50403F907F704F403F4F3F3F50403F4F403F4F403F4F3F3F
50403F4F403F000000000000000000000000000000000000000000000000000000000000001F00\
001F00207F001F7F00207F001F7F00207F001F7000207F001F7000207F001F7000
207F001F70002070001F70002070001F70001F70001F6F002070001F5F001F5F001F5F00000000\
000000000000000000000000002F00002F001F6000002F001F60001F6000104F00
001F00002000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000060506F5F506060506F5F506F60506F60506F605F6F60506F605F6F
60506F6F5F7060506F6F5F6F6F5F6F6F5F706F5F6F302F3F302F303F303F302F303F303F302F30\
3F303F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F3F3F4F403F4F3F3F907F708F7F6F907F70907F6F907F708F7F6F4F403F
4F3F3F4F403F4F3F304F403F4F3F304F403F4F3F30000000000000000000000000000000000000\
0000001F7000207F001F7000207F001F70001F7F001F7000207F001F70001F7000
1F70001F70001F70001F70001F70001F70001F6F001F70001F6F001F70001F6F001F5F001F5000\
1F5F00000000000000000000000000000000001F00002F001F6000103F00002F00
000000001000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000F0F0F0F5F506F60506F60506F605F6F60506F605F6F60506F6F5F70605F6F
6F5F70605F6F6F5F70302F3030303F302F303F303F302F3F3F303F302F30000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F4F403F80706F8F7F6F8F706F907F70907F6F908070907F6F
907F70907F6F908070907F6F4F403F4F403F50403F4F403F4F403F4F403050403F4F403F000000\
000000207F001F7F00207F001F7F00207F001F7F00207F000030000030001F7000
001F00001F002060001F5F002060001F70002070001F70001F70001F6F002070001F6F001F7000\
1F5F00003000002F00000000000000000000000000000000002F001F4000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000060506F5F506060506F60506F60506F60506F605F6F60506F302F3F
302F30302F3F302F30302F3F302F30302F3F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F3F3F4F403F4F3F3F8F706F80706F8F706F507F405F7F40507F405F7F40
507F40907F70907F6F907F708F7F6F907F6F907F6F907F6F8F7F6F4F403F4F3F304F403F4F3F30\
4F403F1F7F00207F001F70001F7F001F7000207F001F7000003000002F00003000
000000001F000000000000001F5F001F4F001F40001F70001F6F001F70001F6F001F70001F6F00\
1F70001F6F00002F00002F00002F00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F000F0F0F0F00000F60506F60506F302F3F302F30302F3F302F30302F3F302F30
30303F302F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F504040707F5F7F805F707F5F2080001F7F005F804F5F7F40
5F804F5F7F405F80405F7F405F80405F7F40908070907F6F908070907F6F90806F907F6F90806F\
4F403F3F601F3F6010207F001F7F00207F001F7F00207F00000000000000000000
000000000000000000000000000000000000204F00002F000030001F70001F70001F7000207000\
1F70001F70001F6F001F7000002F00003000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F0000000F0F0F2F2030302F302F2030302F30302F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F3F3F4F403F4F403F7F805F707F5F707F5F1F7F002080001F7F00207F00
1F7F002080001F7F005F7F40507F405F8040507F405F7F40507F405F8040507F3F907F6F907F6F\
907F6F8F70603F601F1F7F00207F001F70001F7F001F70000F0F0F000000000000
000000000000000000000000000000000000000000000000002F00003000002F001F70001F6F00\
1F70001F6F001F70001F6F001F70001F6F00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F302F3F302F3F302F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F504040707F5F7F80601F7F002080001F7F002080001F7F00
2080001F7F002080001F7F002080001F7F002080001F7F005F80405F7F405F80405F7F405F8040\
507F40907F6F8F7F603F7F1F1F7F00207F001F7F003F601F0000000F0F0F000000
000000000000000000000000000000000000000000000000000000000000003000002F001F4F00\
1F70002070001F70001F70001F6F001F70001F6F00000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F3F302F30302F3F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F403F5040404F3F3F7F805F707F5F2080001F7F002080001F7F00208000
1F7F002080001F7F002080001F7F002080001F7F002080001F7F002080001F7F00208000507F3F\
5F7F40507F3F5F7F403F7F103F7F1F1F7F00207F00305F103F6010305F10000000
0000000000000000000000000000000000000000000000000000000000000000000000001F4000\
1F40001F5F001F70001F6F001F70001F6F001F6F00002F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F2F30302F3F\
302F3F5F50605F4F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F7F80607F805F7F80601F7F002080001F7F002080001F7F00
2080001F7F002080001F7F002080001F7F002080001F7F002080001F7F002080001F7F00208000\
1F7F005F8040507F2F5F8030507F2F207F001F7F00207F001F7F004F403F4F3F30
4F403F000000000000000000000000000000000000000000000000000000000000000000000000\
0000001F5F001F5F002070001F70001F70001F6F00003000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F3F2F2030302F3F5F4F6F\
5F506F504F605F5060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F403F4F403F707F5F7F8060707F5F3F801F3F7F103F801F3F7F10208000
1F7F002080001F7F002080001F7F002080001F7F002080001F7F002080001F7F002080001F7F00\
2080001F7F00207F001F7F005F7F2F1F7F001F7F001F7F00207F0050703F507F3F
403F304F4030403F30000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F6F001F70001F6F001F70001F6F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F3F2F2F30302F3F5F506F5F506F\
5F506F5F506F5F4F605F506F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F9F807F9080709F807F3F801F40801F3F801F40801F3F801F
40801F3F801F40801F3F801F2080001F80002080002080002080001F80002080001F8000208000\
1F7F002080001F7F002080001F7F002080001F7F00207F001F7F005F7F40507F3F
8F7F608F7F604F403F403F304F4030000000000000000000000000000000000000000000000000\
000000000000000000001F00001F001F70001F70001F70001F4000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2030302F3F2F20305F506F5F4F605F506F5F4F6F\
5F506F504F605F506F5F4F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F403F504040907F709F807F907F709F807F907F70908070907F70908070
907F703F801F3F7F103F801F3F7F103F801F3F7F103F801F1F7F002080001F7F002080001F7F00\
2080001F7F002080001F7F00207F001F7F00207F001F7F001F7F001F7F00207F00
1F70108F7F608070608F7F60403F304F4030403F30000000000000000000000000000000000000\
000000000000000000001000001F001F5F001F70001F6F001F6F001F3F00000000
000000000000000000000000000000000000000000000000000000000000000000000000504F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F3F2F2F30302F3F504F605F506F5F4F6F5F506F5F506F605070\
5F506F6050705F4F605F506F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F9F807F90807F9F807F9080709F807F9080709F807F908070
9F807F9080709080709080709F80703F801F40801F3F801F40801F3F801F40801F3F8010208000\
1F80002080001F7F002080001F7F002080001F7F002080001F7F00207F001F7F00
207F1F1F7F103F7F2F8F7F608F7F60807F604F4030403F304F4030000000000000000000000000\
0000000000000000000000000000001F5F001F6F001F70001F6F001F6F00000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
504F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F20302F2F302F2030504F6F504F605F4F6F504F605F506F504F6F5F506F5F506F\
5F506F5F506F5F506F5F4F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5040404F403F5040409080709F807F907F709F807F9080709F807F907F70908070
908070908070907F70908070908070908070907F709080709080709080703F7F103F801F3F8010\
3F801F1F7F002080001F7F002080001F7F001F7F001F7F00207F001F7F001F7F00
1F7F00207F1F3070203F7F2F3070208F7F6080705F807F60403F304F4030403F2F000000000000\
000000000000000000000000001F000010001F70001F6F001F7000106F001F4000
000000000000000000000000000000000000000000000000000000000000000000000000504F30\
504F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F20302F2F3F2F2030504F6F504F605F506F504F6F5F506F5F4F6F5F506F5F506F5F5070\
5F506F605F705F506F5F506F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F403F504F405040409F807F90807F9F807F9F807F9F807F9080709F807F908070
9F807F9080709F807F9080709F807F9080709F807F9080709080709080709F8070908070908070\
3F801040801F3F801F3F801F1F7F002080001F7F002080001F7F002080001F7F00
207F001F7F00207F003F7F2F3F7F2F507F3F507F3F807F5F8F7F6080705F4F4030403F304F4030\
0000000000000000000000000000000000001F70001F70001F6F001F70001F4000
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
504F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2030
2020302F20304F4060504F60504F60504F6F504F605F506F504F605F506F5F4F6F5F506F5F506F\
5F506F5F506F5F506F5F4F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5040404F403F50404090807F9F807F90807F9F807F9080709F807F9080709F807F
90807090807F90807090807F90807090807090807090807090807090807090807090807090806F\
90807090806F3F801F3F7F103F801F3F7F102080001F7F001F7F001F7F00207F00
1F7F001F7F001F7F001F7F001F7000507F3F50703F507F3F80705F807F5F80705F807F5F403F2F\
403F30403F2F0000000000000000000000001F5F001F6F001F6F001F6F001F6F00
000000000000000000000000000000000000000000000000000000000000000000000000504F30\
504F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F20302F2F302F2030
504F60504F60504F6F504F605F506F504F6F5F506F504F6F5F506F5F506F5F506F5F506F605F70\
5F506F605F705F506F5F506F5F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504040504F405040409F8F809F807F9F807F9F807F9F8F7F9F807F9F807F908070
9F8F7F9080709F807F9080709F8F7F9080709F807F9080709F8F7F9080709F8070908070908070\
90807090807090806F9080703F801F3F801F3F7F103F801F1F7F002080001F7F00
2080001F7F00207F001F7F00207F001F7F00207F00507F3F507F3F507F3F807F6080705F807F60\
80705F404030403F2F4040300000001F60001F5F001F70001F6F001F7000002F00
001F000000000000000000000000000000000000000000000000000000000000000000005F5030\
504F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F20302020302F20304F405F4F4F60
4F4060504F604F4060504F60504F60504F6F504F605F506F504F6F5F506F504F6F5F506F5F506F\
6050705F506F5F506F5F506F5F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504F404F403F5040409F807F9F807F90807F9F807F90807F9F807F9080709F807F
9080709F807F90807090807F90807090807F908070908070908070908070908070908070908070\
9080708F806F9080708F806F9080708F7F6F90806F3F7F103F801F307F101F7F00
1F7F00207F001F7F001F7F001F7F001F7F001F70001F7F001F7000507F3F4F7030307F1080705F\
807F5F6070406F704F403F2F304F201F5F001F5F001F6F001F7000106F00002F00
002F00000000000000000000000000000000000000000000000000000000000000000000505030\
504F3040402F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002020302F20302020304F40604F4060504F604F4060
504F604F4F60504F6F504F60504F6F504F6F5F506F504F6F5F506F5F506F5F50705F506F605070\
5F506F605F706050705F506F5F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504040504F405040409F8F809F807F9F8F809F807F9F8F7F9F807F9F8F7F9F807F
9F8F7F90807F9F8F7F90807F9F8F7F9080709F8F7F9080709F8F7F9080709F8F7F908070908070\
9080709080709080709080708F806F9080708F806F90806F8F7F6F3F801F3F7F10
3F801F1F7F002080001F7F00207F001F7F00207F001F7F00207F001F7F00207F00307F103F7F1F\
3070106F7F4F6F704F3F7F103040201F70001F5F001F70001F6F001F7000002F00
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
504F2F4F4F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000202030201F2F20203040405F4F406040405F4F40604F4060504F60
4F4060504F604F4F60504F6F504F60504F6F504F6F5F506F504F6F5F506F5F506F5F50705F506F\
5F50705F5070605F705F506F5F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5040404F4040504F409F807F9F807F9F807F9F8F7F90807F9F807F90807F9F8F7F
9080709F807F9080709F8F7F90807090807F9080709F8F7F908070908070908070908070908070\
9080709080709080708F806F9080708F806F90806F8F7F6F8F806F8F7F6F8F806F
307F103F7F1F307F103F7F1F1F7F001F7F001F7F001F7F001F70001F7F001F70001F7F001F7000\
307F101F70001F7F003070103070101F70001F70001F6F001F7000002F00002F00
000000000000000000000000000000000000000000000000000000000000000000000000505030\
504F2F4F402F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001F0000000000000000000000000000000000000000\
0000002020302020304F406040405F4F40604F405F4F4F604F4060504F604F4060
504F6F504F60504F6F504F605F506F504F6F5F506F504F6F5F50705F506F5F50705F506F605F70\
5F5070605F705F50705F50705F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504040504F405040409F8F809F807F9F8F809F807F9F8F809F807F9F8F7F9F807F
9F8F7F9F807F9F8F7F90807F9F8F7F90807F9F8F7F9080709F8F7F9080709F8F7F9080709F8F7F\
90807090807090807090807090806F9080708F806F9080708F806F90806F8F7F6F
8F806F8F7F6F3F801F307F103F801F3F7F10207F001F7F00207F001F7F00207F001F7F00207F00\
1F7F001F7F001F7000207F001F70001F70001F70001F70001F7000001F00002F00
0000000000000000000000000000000000000000000000000000000000000000000000005F5F3F\
504F2F4F4F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000001000001F00002F0000000000000000000000000000000000000000\
000000201F2F40405F403F5F4F406040405F4F40604F40604F4F604F40604F4F60
4F4060504F6F4F4F60504F6F504F6050506F504F6F5F506F504F6F5F50705F506F5F50705F506F\
605F705F5070605F705F506F5F5070000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F4040\
504F404F403F9F8F809F807F9F8F8090807F9F8F7F9F807F9F8F7F90807F9F8F7F
90807F9F8F7F9080709F8F7F9080709F8F7F9080709F8F7F9080709F8F7F908070908070908070\
9080709080709080708F806F9080708F806F90806F8F806F90806F8F7F6F8F806F
8F7F608F806F807F608F7F6F307F103F7F1F307F103F7F101F7F001F7F001F70001F7F001F7000\
1F7F001F70001F7F001F70001F7F001F70001F70001F5F101F5F10001000001F00
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
50503040402F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000001F00001000003000001F0000000000000000000000000000000000\
20202F40405F40405F4F406040405F4F40604F40604F4F604F4060504F6F4F4F60
504F6F4F4F60504F6F504F6F50506F504F6F5F507050506F5F50705F506F5F5F705F5070605F70\
5F5070605F70605F705F50705F506F605F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F40\
504040504F409F807F9F8F809F807F9F8F809F807F9F8F809F807F9F8F809F807F
9F8F7F90807F9F8F7F908F7F9F8F7F90807F9F8F7F908F7F9F8F7F9080709F8F7F908070908F7F\
908070908F709080709080709080709080708F806F9080708F806F9080708F806F
8F806F8F7F6F8F806F8F7F608F806F807F603F801F307F103F7F1F307F10207F001F7F001F7F00\
1F7F00207F001F7F001F7F001F70001F7F004F702060704F1F5F10403F2F000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5F3F\
504F2F4F4F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F00003000003000003000003000001F00002F00000000000000000000001F1F2F\
202030403F5F40405F403F5F4F406040405F4F40604F40604F4F604F4060504F6F
4F4060504F6F4F4F60504F6F504F6050506F504F6F5F506F504F6F5F50705F506F5F50705F506F\
5F5F705F5070605F705F50705F50705F506F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F4040\
504F404F40409F8F8090807F9F8F8090807F9F8F8090807F9F8F7F90807F9F8F7F
90807F9F8F7F90807F9F8F7F9080709F8F7F9080709F8F7F908070908F7F908070908F7F908070\
9080709080709080708F806F9080708F806F9080707080407080408F806F8F806F
8F7F6F8F806F8F7F608F806F807F608F7F60807F608F7F60307F103F7F10307F103F7F101F7000\
1F7F001F70001F7F001F70001F70001F70004F7020406F307F70503F3F2F403F2F
000000000000000000000000000000000000000000000000000000000000000000000000505030\
5050304F402F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F00003000003000003F00003000003F00002F00003F0000000000000000201F2F202030\
403F5F40405F403F5F4F40604040604F40604F40604F4F604F4060504F6F4F4F60
504F6F504F6F504F6F504F6F5F506F504F6F5F507050506F5F50705F50705F50705F5070605F70\
5F5070605F705F5F70605F7F5F506F605F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F40\
504040504F409F807F9F8F809F8F7F9F8F809F807F9F8F809F8F7F9F8F8090807F
9F8F80908F7F9F8F7F90807F9F8F7F908F7F9F8F7F908F7F9F8F7F908F7F9F8F7F9080709F8F7F\
908070908F703F802F408F309080705F803050802F70804F70804020801F1F8010
708040708040708040507F2F8F806F8F7F608F806F807F608F806F807F603F7F1F307F103F7F1F\
307F101F7F001F7F00207F001F70001F7F001F70004F703F706F4F7F70503F3F2F
403F2F0000000000000000000000000000000000000000000000000000000000000000005F5F3F\
504F304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F000000000000000000
0000003000002F00003000003000003F00003000003F00003F00001F00002F0F1020202F3F3F50\
40405F403F5F40405F40405F4F406040405F4F40604F40604F4F604F40604F4F6F
4F4F60504F6F4F4F60504F6F504F6F50506F504F6F5F506F50506F5F50705F506F5F50705F5070\
605F705F5070605F70605F705F5F705F506F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504F404F4040\
504F4090807F9F8F8090807F9F8F8090807F9F8F8090807F9F8F8090807F9F8F7F
90807F9F8F7F90807F9F8F7F90807F9F8F7F908070908F7F908070908F7F8F7F6F908F7F5F8040\
5F8F4F5080403F80303F802F2080001F80005F802F1F80002080001F80101F801F
1F80002080006F7F3F50802F507F2F3F801F807F608F7F60807F608F7F60807F60807F60307F10\
3F7F103070101F7F001F70001F7F001F70001F70001F7000706F50706F50707050
3F3F2F3F3F2F0000000000000000000000000000000000000000000000000000000000005F5030\
50503040402F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F30000000000000000000
0000003000003000003000003F00003000003F00003F00003F00001F00003010103F3F5040405F\
403F5F40406040405F4F40604040604F4F604F40604F4F604F4060504F6F4F4F60
504F6F4F4F6F50506F504F6F50506F504F6F5F507050506F5F50705F50705F5F705F5070605F70\
5F5070605F7F605F70605F7F5F506F605F705F5F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504F40504F40\
4F40409F8F809F8F809F8F809F8F7F9F8F809F8F7F9F8F80908F7F9F8F80908F7F
9F8F80908F7F9F8F80908F7F9F8F7F908F7F9F8F7F908F7F9F8F7F908F7F8F80708F7F6F208F00\
5F80405F8F4F1F8000208F001F8000208F001F80002080001F80002080001F8000
2080001F80002080001F800050802F307F103F801F1F7F008F806F807F608F8060807F60807F60\
807F603F7F1F307F101F7F001F70001F7F001F70001F70006070407F7050706F50
7070503F3F2F3F3F2F000000000000000000000000000000000000000000000000605F3F5F5F3F\
504F304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F000000000000000000
1000003000003000003000003000003000003000003F00003000003F00003F0F10403F5F3F2F40\
403F5F403F5F404060403F5F4F40604040604F4F604F40604F4F604F4060504F6F
4F4F60504F6F504F6F50506F504F6F50506F504F6F5F507050506F5F50705F50705F5F705F5070\
5F5F705F5070605F7F5F5F70605F7F5F506F605F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504F404F4040\
504F40908F7F9F8F8090807F9F8F80908F7F9F8F8090807F9F8F80908F7F9F8F80
90807F908F7F908F7F9F8F7F90807F908F7F908F7F908F7F908070908F7F5F80308F80701F8000\
208F001F8000208F001F80002080001F80002080001F80001F80001F8000208000
1F80001F80001F80002080001F7F001F8000307F101F80001F7F001F7F00807F60807F60807F5F\
807F60807F5F807F603070101F7F001F70001F70005F7040607040706F50707050
706F4F706F503F30203F3F2F000000000000000000000000000000000000000000606040505030\
5050304F402F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F30000000000000000000
2F00003000003000003000003F00003000003F00003F00003F00003F00004010103F304040304F\
403F5F40406040203F4F40604040604F4F604F40604F4F6F4F4F60504F6F4F4F60
504F6F504F6F50506F504F6F5F507050506F5F507050506F5F5F705F50705F5F705F5070605F7F\
5F5F70605F7F5F5F70605F7F605F7F605F705F5070000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F4040504F4F\
9F8F809F8F809F8F809F8F809F8F809F8F80908F7F9F8F80908F7F9F8F80908F7F
9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F7F908F7F9F8F7F5F8F305F8F301F8000208F00\
1F8000208F001F8000208F001F8000208F001F8000208F001F80002080001F8000
2080001F80002080001F80002080001F80002080001F7F002080001F7F002080001F7F008F8060\
807F60807F60807F5F807F60307F101F7F001F7000607F4F6070407F7050706F50
7F7050706F4F706F503F3020000000000000000000000000000000000000000000605F3F5F5F3F\
5050304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F000000000000000000
1F00003000002F00003000003000003F00003000003F00003F00003F00003F00004000003F0F10\
401F2F401F20402F3F40203F4F4060402F3F4F2F3F4F2F3F4F2F404F40604F3F50
4F4F6F504F6F4F4F6F50506F504F6F505070504F6F5F507050506F5F50705F50705F5F705F5070\
605F705F5070605F7F5F5F70605F7F5F506F5F5F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4040504F404F4040\
9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80908F7F908F80
908F7F9F8F80908F7F908F7F908F7F908F7F908F7F908F7F908F7F208F1F1F8010208F001F8000\
208F001F8000208F001F8000208F001F80002080001F80002080001F80001F8000
1F80002080001F80001F80001F80001F80001F7F001F80001F7F001F80001F7F001F7F001F7F00\
1F7F00807F5F807F607F7F5F807F5F3070101F7000607040607040706F50707050
706F4F706F506F6F4F706F4F3030200000000000000000000000000000000000006060405F5030\
5050304F402F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3010101F000000000000
2000002F00002000003000003F00003000003F00003F00003F00003F00004000003F0000400000\
401F204000004000004F2F3F4000004F00004F00004F2F404F000050405F4F3F50
504F6F4F4F6050506F504F6F50507050506F5F50705050705F50705F50705F5F705F50705F5F7F\
5F5F70605F7F5F5F70605F7F605F7F605F705F5F70605F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F504F40504F4F\
9F8F809F8F809F8F809F8F80908F809F8F80908F809F8F80908F7F9F8F80908F7F
9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80908F7F908F7F1F8F00208F1F1F8F00208F00\
1F8000208F001F8000208F001F8000208F001F8000208F001F8000208F001F8000
2080001F80002080001F80002080001F80002080001F80002080001F7F002080001F7F001F8000\
1F7F00208000807F60807F607F7F5F807F5F30701060704F7070507F705F707050
707050706F4F7070506F6F4F706F4F3F3020000000000000000000000000000000605F3F5F5F3F\
5050304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F1F101F000000000000
0F00000000001F00003000000F00003000003000003F1F203000003F00003F00004000003F0000\
4000004000004000004000004F00004000004F00004F00004F00004F3F5050101F
4F0F104F4F6F501F2F505070504F6F50507050506F5F507050506F5F50705F50705F5F705F5070\
5F5F705F5F70605F7F5F5F70605F7F605F7F5F5F705F5070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4040504F404F4040\
9F8F80908F809F8F80908F7F9F8F80908F7F9F8F80908F7F908F80908F7F9F8F80
908F7F908F80908F7F908F80908F7F908F7F708F607F8F60908F7F208F001F8F00208F001F8000\
208F001F8000208F001F80002080001F8000208F001F80001F80001F8000708040
1F80001F801F1F801020801F1F7F001F80001F80001F80001F7F001F80001F7F001F80001F7F00\
1F7F001F7F00307F107F705F7F7F5F7F705F7F7F5F5F70407F7050707050707050
706F4F706F506F6F4F706F4F6F604F6F6F4F3030200000000000000000006060406060405F5030\
5050304F4F2F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3010101F000000000000
0F000000000000000F3000001000002F0F103F1F203F10203F10103F00004000003F0000400000\
4000004000004000004F00004000004F00004F00004F00004F00005000004F1010
50000050202F5020305030405F304F50507050506F504F6F5F5F705F50705F5F705F5070605F7F\
5F5F70605F7F5F5F7060607F605F7F60607F5F5F70605F7F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F4F4F409F908F\
9F8F809F8F80908F809F9080908F809F8F80908F809F9080908F809F8F80908F7F
9F9080908F7F9F8F80908F7F9F9080908F7F7F8F6F708F60208F001F8F00208F001F8F00208F00\
1F8F00208F001F8000208F00708040708F4F1F8000708F6070805F8F8F706F8040
708F408F806F20801F50802F50803050802F2080001F80002080001F80002080001F7F00208000\
1F7F001F8000307F103F7F2F7F7F5F7F7F5F7F705F7F7F5F7F70507F705F707050
707050706F507070506F6F4F706F506F6F4F3F3F2F3030200000000000006060405F5F3F5F5F3F\
5050304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F1F1F1F000000000000
0000000000000000000000001F1F202F0F1030304F3F30503F0F103F30403F20304000003F0000\
401010402030402F3F400F104F00004000004F00004F00004F00004F00004F0000
4F00005000004F00005030405000005000004F4F6050506F5050705F50705050705F5F705F5070\
5F5F7F5F5070605F7F5F5F70605F7F5F5F7F5F5F705F5F70605F70000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4040504F40908F80\
9F8F80908F809F8F80908F809F8F80908F7F908F80908F7F9F8F80908F7F908F80
908F7F908F80908F7F908F80908F7F908F80708F60708F601F8F00208F001F8F00208F001F8000\
208F001F8010208F1F1F8010708F4F6F80408F80708F807070805F8F806F8F8070
8F806F8F807080806F8F806F80806F6F80406F7F3F6F80401F7F001F80001F7F001F80001F7F00\
1F80001F7F001F7F00307F20307F2F7F705F7F7F5F7070507F705F707050707050
706F507070506F6F4F6F6F4F6F6F4F6F6F4F6060403F30203030200000005F5F3F5F5F3F5F5F3F\
10500F1F4F0F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3010101F000000000000
0000000000000000001F1F2F1F1F2F30304F3F3F503030503F304F3F3F50402F3F3F0000400000\
400F10402F3F4040604F101F4000004F30404F00004F00004F00005000004F0000
5000004F00005000005000005000005000005F203050202F5F5F705050705F5F7F5F50705F5F7F\
5F5F70605F7F5F5F7F605F7F605F7F60607F605F7F605F7F5F5F70000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F4F4F409F908F\
908F809F908F908F809F9080908F809F9080908F809F9080908F809F9080908F80
9F9080908F7F9F9080908F7F909080708F607F906F708F60208F001F8F00208F001F8F00208F00\
1F8F005F8F301F8F1F908F7F8F8070908F7F8F8070908F708F8070908F708F8070
8F80708F806F8F8F708F806F8F806F80806F8F806F6F80506F805F6F80507080501F8000208000\
1F7F002080001F7F001F7F004F7F3F4F7F3F7F705F7F7F5F7070507F705F707050
707050706F507070506F6F4F706F4F6F6F4F6F6F4F6F604F3F3F2F3030206060405F5F3F5F5F3F\
10500F20500F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F2F2F3F000000000000
0000000000001F1F201F1F2F30304F30305030304F3F3F503030503F3F5F3F2030402F3F3F0000\
40000040000040406040304F4F2F3F402F3F4F00004F00004F00004F0000500000
4F00005000004F000050000050000050000050202F5F20305030405F304F5030405F507050506F\
5F5F7F5F5F70605F7F5F5F70605F7F5F5F7F60607F5F50705F5F70000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4F40504F4F908F80\
9F8F80908F809F8F80908F80908F80908F809F8F80908F7F908F80908F7F909080
908F7F908F80908F7F909080908F7F708F60708F60708F601F8F00208F001F8F00208F001F8000\
208F0050802F5F8F308F80708F8F708F80708F8F708F806F8F80708F806F8F8070
80806F8F806F80806F8F806F80806F8F806F80806F8F806F807F606F80506F7F506F8050607F4F\
1F80001F7F001F7F001F7F001F7F004F70307070507070507F705F707050707050
706F507070506F6F4F706F506F6F4F6F6F4F6060406F6F4F6060403030205F5F3F5F5F3F505030\
1050000F4F00404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F302F2F30000000000000
0000000000001F1F2F30304F3030503030503F3F503030503F3F5F3F3F503F2F3F3F2030400000\
4000004000003F3F5F4F4F604040604F4F604F2F3F4F30404F00005000004F0000
5000005000005000005000005000005000005F00005000005F00005F101F5F101F5050705F5F70\
5F5F70605F7F5F5F7F60607F605F7F60607F605F7F605F7F5F5F70000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F4F4F409F908F\
908F809F908F908F809F908F908F809F9080908F809F9080908F809F9080908F80
9F9080908F80909080908F7F909080708F607F906F1F8F002090001F8F00208F001F8F00208F00\
1F8F005F8F4F508040908F7F8F8070908F708F8070908F708F80708F8F708F806F
8F8F708F806F8F807080806F8F807080806F8F806F80806F8F806F80806F8F806F6F80506F8050\
507F3F1F80001F7F001F80001F7F001F7F00706F507F7F5F70705F7F705F707050
7070507070507070506F6F4F7070506F6F4F6F6F4F6060406F6F4F3030206060405F5F3F5F5F3F\
1050001050001F400F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F2F2F3F000000000000
0000001F1F2F2F2F4F30304F30304F30305030304F3F3F503030503F3F5F3F3F503F203F3F0000\
4000003F000040305040304F4F4F604040604F4F6F4F2F3F4F4F6F4F00004F0000
4F00005000004F00005000005000005000005000005000005000005F101F5010105F304F5F3040\
5F3F4F5F5F705F50705F5F7F605F7F5F5F7F60607F5F5F70605F7F5F5F70000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4F404F4F4F908F80\
9F908F908F80909080908F80909080908F80908F80908F80909080908F7F908F80
908F7F909080908F7F908F80908F7F70906F708F60208F001F8F00208F001F8F00208F001F8F00\
5F8F4F5080408F8F7F8F80708F8F708F80708F8F708F80708F8F708F806F8F8070
8F806F8F807080806F8F806F80806F8F806F80806F80806F80806080806F807F6080806F807F60\
50803F4F7F3F1F7F001F7F001F7F001F70001F7F0070705F7F705F707050707050
706F507070506F6F4F6F6F506F6F4F6F6F4F6060406F6F4F6060400000005F5F3F5F5F3F20500F\
1050000F4F00104F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F302F2F30000000000000
0000002F2F4F30304F30304F30305030304F3F3F503030503F3F5F3F3F5040405F3F203F402F3F\
3F000040000040304F4F2F404040604F4F6F4F4F6F4F4F6F4F4F605030404F2F40
5000004F00005000005000005000005000005F00005000005F00005000005F00005F00005F3F4F\
5F00005F5F705F507060607F5F5F7F606080605F7F6060805F5F7060607F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F4F4F4F9F908F\
908F809F908F908F809F908F908F809F908F908F80909080908F809F9080908F80
909080908F80909080908F80909080908F7F9090801F8F002090001F8F002090001F8F00208F00\
508F4F5F8F4F8F8F70908F7F8F8070908F7F8F80708F8F708F80708F8F708F8070
8F8F708F806F8F8F708F806F8F807080806F8F806F80806F8F806F80806F8F806F80806080806F\
807F6050802F4F7F2F1F7F001F7F001F7F001F7F004F7F2F70705F7F7F5F707050
707050706F507070506F6F4F6F6F506F6F4F6F6F4F6F6F4F0000000000003030205F5F3F205F10\
105000105000104000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F2F2F30000000000000
0000002F304F2F2F4F30304F2F304F3030503030503F3F503030503F3F5F3F3F503F2F3F3F0F10\
4000003F000040000040203F4040604040604F4F6F4040604F4F6F4F2F404F3040
4F4F605000004F00005000004F00005000005000005000005000005F00005000005F0000500000\
5F00005F101F5F101F5F5F7F605F7F5F5F7F60607F5F5F7F5F5F7F5F5F70000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F4F40504F4F908F80\
90908F908F8090908F908F80909080908F80909080908F80909080908F80909080
908F7F909080908F7F9090808F8F7F9090808F8F7F2090001F8F00208F001F8F00208F001F8F00\
3F8F30508F408F8F7F8F80708F8F708F80708F8F708F80708F8F708F80708F8F70
80806F8F807080806F8F807080806F80806F80806F80806F807F6080806F80806080806F7F7F60\
8080607F7F604F7F2F1F7F001F7F001F7F001F7F0010700070705070705070705F
6F6F507070506F6F4F6F70506F6F4F6F6F4F606F4F6F6F4F0000000000002F2F1F5F5F3F105000\
1050000F4F00104F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F302F2F30000000000000
0000002F2F4F30304F2F304F3030503030503F3F503030503F3F5F3F3F5F3F3F5F3F3F5F40101F\
4000004000004000004F2F40402F3F4F4F6F4040604F4F6F4F4F6F4F4F6F4F4F6F
50506F4F30405030405000005000005000005000005000005F00005000005F00005F00005F1F1F\
5F101F5F101F5F101F60607F5F5F7F606080605F7F6060805F5F7060607F5F5F7F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F504F4F4F4F409F908F\
90908F90908F9090809F908F90908090908F908F809F908F909080909080908F80
909080909080909080908F809090809090803F901F3F8F1F2090001F8F002090001F8F00208F00\
3F8F303F8F308F8F70908F7F8F8F708F8F7F8F80708F8F7F8F80708F8F708F8070
8F8F708F80708F8F7080806F8F8F7080806F8F806F80806F8F806F80806F80806F80806080806F\
807F608080607F7F5F1F80001F7F001F7F001F7F001F7F001F70007F7F5F707050
70705F7070507070506F6F4F7070506F6F4F6F6F4F606F4F000000000000303020305F1F105F00\
1050001050000F4000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3F000000000000
00000030304F2F2F4F30304F30304F303050303050303F503030503F3F5F3F3F503F3F5F3F0F10\
4000003F0000400000402F3F402F404040604F4F6F404F6F4F4F6F4F4F6F4F4F6F
4F4F6F5030404F2F4050507050101050101F5000005000005000005F000050202F5F20305F101F\
5F101F5F5F705F5F7F5F5F7F60607F5F5F7F60607F605F7F6060805F5F70605F7F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F4F404F4F4F90908F\
90908F908F8090908F908F8090908F908F80909080908F80909080908F80909080
908F809090808F8F7F909080908F7F9090803F8F1F3F901F1F8F002090001F8F001F8F001F8F00\
208F008080708F8F7F8F80708F8F7F8F80708F8F708F80708F8F7080806F8F8070
80806F8F807080806F8F806F80806F8F806F80806F80806F80806F80806F807F6080806F7F7F60\
8080607F7F607F7F604F7F201F7F001070001F7F001F70001F7F005F7040707050
7070507070506F6F4F6F6F506F6F4F6F6F4F60604F000000000000000000000000305F1F105000\
1050002F4F1F3F4F20000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F301F1F20000000
1F1F202F2F4030305030304F3030503030503F3F503030503F3F5F3F3F503F3F5F3F3F5F40101F\
3F0000400000400000400000402F3F4F3040404F6F4F4F6F4F4F6F4F50704F4F6F
5050704F4F6F50507050507050101F5030405F00005000005F3F4F5030405F2F305F20305F5F7F\
5F5F7F605F7F5F5F7F60607F5F5F7F60608060607F60608060607F60607F5F5F7F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F50504F4F4F4F90908F\
90908F90908F90908F90908F90908090908F90908090908F90908090908F909080
909080909080909080909080909080908F805F9050508F4F2090001F8F002090001F8F00208F00\
1F8F008F8F7F8080708F8F7F8F8F708F8F7F8F8F708F8F708F80708F8F708F8070
8F8F7080806F8F8F7080806F8F8F7080806F8F8F7080806F80806F80806F80806F80806080806F\
7F7F608080607F7F604F7F2F4F7F201F7F001F7F001F7F001F7000607F4F707050
70705F6F70507070506F6F506F70506F6F4F6F6F50000000000000000000000000105F00105F00\
10500040502F3F4F20000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3010101F000000
1010202F304F2F2F4F3030502F304F3030503030503F3F503030503F3F5F3F3F5F3F3F5F3F3F5F\
4000003F0000400000400000402F3F402F3F4F4F6F4040604F4F6F4F4F6F4F4F6F
4F4F6F4F50704F4F6F5050704F2F4050304050507050304F50304050304F5050705F5F7F505070\
5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F6060805F5F7F6060805F5F7060607F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F4F4F4F4F4F90908F\
90908F90908F90908F908F8090908F90908090908F908F80909080908F80909080
8F8F809090808F8F809090808F8F7F909080508F4F5F90501F8F00208F001F8F00208F001F8F00\
1F8F00508F4F8F8F708F80708F8F7F8F80708F8F708080708F8F708080708F8F70
80806F8F807080806F8F807080806F80806F80806F80806F80806F80806F80806080806F7F7F60\
8080607F7F607F7F607F7F5F1F7F001F7F001F7F001070001F7F0010700070705F
6F70507070506F6F506F70506F6F4F6F6F4F606F4F000000000000000000000000105F00105000\
5050304F4F2F404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F3F000000
1F1F202F2F4030304F2F304F303050303050303F503030503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F00004000004000004000004000004F3F504040604F4F6F4F4F6F4F4F704F4F6F
5050704F4F6F5050704F50705050705050705050705030405F5F7F5050705F5F7F505F705F5F7F\
5F5F7F5F5F7F5F5F7F6060805F5F7F6060805F5F7F60608060608060607F5F5F7F
60607F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F504F4F4F4F909F90\
90908F90908F90908F90908F90908F90908F90908090908F90908090908F909080
90908F9090809090809090809090808F8F7F8F8F801F8F002090001F8F002090001F8F00209000\
1F8F005F8F4F508F4F8F8F7F8F8F708F8F7F8F80708F8F7F8F8F708F8F70808070
8F8F708080708F8F7080806F8F8F7080806F80807080806F80806F80806F80806F7F806080806F\
7F7F608080607F7F607F80604F7F3F1F7F001F7F001F7F001F7000607F4F5F7040
70705F7070507070506F6F506F70506F6F4F6F6F4F000000000000000000000000105F00105F00\
4F50304F5030404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F301F1F20
10102030304F2F2F4030304F2F304F303050303050303F503030503F3F5F303F5F3F3F5F303F50\
4000003F0000400000400000400000403050404F6F40406F4F4F6F404F6F4F4F6F
4F4F6F5050704F4F6F5050704F50705050705050705050705050705F5F7F5050705F5F7F505F70\
5F5F7F5F5F7F5F5F7F5F5F7F60607F5F5F7F6060805F5F7F60608060607F5F5F7F
5F5F7F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F504F4F4F4090908F90908F\
90908F8F908F90908F90908F90908F8F908090908F90908090908F8F8F80909080
8F90809090808F8F809090808F90809090808F8F7F2090001F8F002090001F8F001F8F001F8F00\
208F001F8F00708F4F8F8F708F8F7F8080708F8F708080708F8F708080708F8F70
8080708F8F7080806F80807080806F80807080806F80806F80806F80806F7F7F6080806F7F7F60\
8080607F7F607F7F607F7F5F4F7F3F1070001F7F001F70001F7F001070005F7040
6F70507070506F6F506F70506F6F4F6F6F50000000000000000000000000104F00105F00105000\
4F5030404F2F404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3030303F10101F
3030402F2F4030304F2F304F303050303050303F5F3030503F3F5F303F5F3F3F5F3F3F5F3F2F3F\
3F203F4000004000004000004000004F0000404F6F4F4F6F404F6F4F50704F4F6F
4F50704F4F6F5050704F5070505070505070505F7F5050705F5F7F5050705F5F7F5F5F7F5F5F7F\
5F5F7F5F607F5F5F7F6060805F5F7F60608060607F6060806060806060805F5F7F
60607F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F50504F90908F909F90\
90908F909F9090908F90908F90908F909F8F90908F90908F90908090908F909080
90908F90908090908F9090809090808F908090908070904F7090501F8F002090001F8F00209000\
1F8F00209000708F4F708F4F8F8F708F8F7F8F8F708F8F7F8080708F8F7F808070
8F8F708080708F8F708080708F8F7080806F8F8F7080806F80806F80806F80806F7F806080806F\
7F7F608080607F7F607F7F60307F101F7F001F7F001F7F001F70001F7F006F7050
70705F6F70507070506F70506F70506F6F4F000000000000000000000000001F00105F00405F20\
4F50304F5030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F302F2F3F
2F2F3F2F30402F2F4030304F2F304F303050303050303F5F3030503F3F5F303F5F3F3F5F3F3F5F\
3F10103F00004000003F0000400000400000404F6F4040604F4F6F404F6F4F4F70
4F4F6F4F50704F4F6F5050704F4F70505070505070505070505070505F7F5050705F5F7F505070\
5F5F7F5F5F7F5F5F7F5F5F7F5F607F5F5F7F6060805F5F7F6060806060805F607F
5F5F7060607F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F4F4F909F908F908F\
90908F8F908F90908F8F908F90908F8F908F90908F8F908090908F8F908090908F
8F90809090808F90809090808F8F809090808F8F807090506F8F4F1F90001F8F002090001F8F00\
1F8F001F8F00208F006F8F408F8F7F808F708F8F7F8080708F8F708080708F8F70
80806F808F7080806F808F7080806F80806F80806F80806F7F806F80806F7F7F6080806F7F7F60\
7F7F607F7F607F7F60707F5F307F102F70101F7F001070001F7F001070006F7050
6F6F507070506F70506F70506F6F4F6F6F50000000000000000000000000001000105F000F400F\
4F5030404F2F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F3F2F2F3F
1F1F202F2F4030304F2F304F303050303050303F5F3030503F3F5F303F5F3F3F5F3F3F5F3F4060\
3F0F104000003F0000400000400000402F40402F3F4F4F6F4F4F6F4F4F6F4F4F6F
5050704F4F705050704F50705050705050705050705050705F5F7F5050705F5F7F505F7F5F5F7F\
5F5F7F5F5F7F5F5F7F6060805F5F7F6060805F607F606080606080606080606080
60607F5F5F7F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F0F0F0F0F0F0F0F0F10\
0F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F
0F0F100F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F0F0F10\
0F0F0F0F0F0F0F0F0F0F10100F0F0F0F0F0F0F0F0F0F10100F0F0F0F0F0F0F0F0F
0F10100F0F0F0000000000000000000000000000000000004F504F4F4F4F4F504F8F908F909F90\
90908F909F908F908F909F9090908F909F8F8F908F909F8F90908F90908F8F9080
909F8F90908090908F8F908090908F8F90809090808F90805F903F5090302090001F8F00209000\
1F8F002090001F8F00208F00508F30508F308F8F708F8F7F8F8F708F8F70808070
8F8F708080708F8F7080806F8F8F7080806F808F7080806F80806F80806F80806F7F7F6080806F\
7F7F607F80607F7F60607F3F607F3F1F7F001F70001F7F001F70004F7F2F407020
70705F6F70507070506F70506F7050000000000000000000000000000000105F00105000104F10\
4F50304F5030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F302F2F3F
1010203030402F2F4030304F2F304F3030502F3050303F5F3030503F3F5F303F5F3F3F5F3F3F5F\
3F10103F00004000003F0000400000402F3F404F6F4040604F4F6F404F6F4F4F6F
4F4F6F4F50704F4F6F5050704F50705050704F5070505070505070505F7F5050705F5F7F505F7F\
5F5F7F505F7F5F5F7F5F5F7F5F60805F5F7F6060805F607F6060805F607F606080
5F5F7F60607F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0F0F0F0F0F0F00000F0F0F0F0F0F0F0F0F0F00000F0F0F0F0F0F0F\
0F0F0F00000F0F0F0F0F0F0F0F0F0F000F0F80809F80809080809F80809080809F
808090808F9F808090808F9F808F908F8F9F808F908F8F9F808F908F909F808F908F909F8F8F90\
8F909F8F8F908F909F8F909090909F8F909090909F8F9090909F9F0F0F0F0F0F0F
0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F4F4F4F4F504F4F4F4F9090908F908F\
90908F8F908F90908F8F908F90908F8F908090908F8F908090908F8F908090908F
8F908090908F8F90809090808F90809090808F8F7F8F9080508F305F90301F8F001F90001F8F00\
2090001F8F001F8F001F8F00508F307F7F6F8080707F7F6F8F8F70808070808F70
808070808F7080806F80807080806F808F7080806F80806F7F806F80806F7F7F607F806F7F7F60\
7F8060707F607F7F606F705F607F3F1070001F7F001070001F7F0010700070705F
6F70507070506F6F506F70506F6F4F000000000000000000000000000000103000105F00202F1F\
4F5030404F2F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3030303F10101F
0000002F2F4030304F2F304F3030502F3050303F5F3030503F3F5F303F5F3F3F5F3F3F5F3F4060\
3F203F4000003F000040000040000040202F40406F4F4F6F404F6F4F4F704F4F6F
4F50704F4F705050704F50705050704F5070505F7F505070505F7F5050705F5F7F505F7F5F5F7F\
5F5F7F5F60805F5F7F5F60805F5F7F6060805F6080606080606080606F80606080
60607F5F5F7F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F100F0F0F
0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F0F0F107F7F907F7F9F7F7F9080809F7F7F9F80809F\
7F7F9F80809F7F809F80809F7F809F80809F80809F80809F80809F808F9F80809F
808F9F80809F8F8F9F808F9F8F8F9F808F908F909F8F8F9F8F909F8F8F908F909F8F909F8F909F\
8F909090909F8F909F90909F8F9090909F9F8F909F909F9F8F9090909F9F8F9090
909F9F8F9090909F9F8F90900F0F0F0F0F0F0F10100F0F0F4F504F4F4F4F8F8F8F808F808F8F8F\
808F80909F908F908F909F8F8F908F909F8F8F908F909F8F8F908F909F8F8F9080
90908F8F9080909F8F8F90809090808F90809090808F90809090808F8F7F2090001F8F00209000\
1F8F002090001F8F00208F001F8F00208F007F806F8F8F7F808F708F8F7F808F70
8F8F708080708F8F70808070808F7080806F808F7080806F80806F7F806F80806F7F80607F806F\
7F7F607F80607F7F6070705F1F7F001F7F001F7F001F7F001F70001F7F006F7050
70705F6F70507070506F70506F7050000000000000000000000000000000105F00105000103000\
4F4F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F301F1F1F
0000002F30402F2F4030304F2F304F3030502F3050303F5F303050303F5F303F5F3F3F5F303F5F\
3F2F3F3F00004000003F0000400000401F2F404F6F404060404F6F404F6F4F4F70
4F4F6F4F50704F4F705050704F4F705050704F507050507F505070505F7F5050705F5F7F505070\
5F5F7F505F7F5F5F7F5F5F7F5F607F5F5F7F6060805F5F7F6060805F6080606080
60608060607F5F5F7F000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F000F0F0F0F0F00000F0F0F0F0F0F0F0F0F0F
00000F0F0F0F7070907F7F907070907F7F90707F907F7F9F7F7F907F7F9F7F7F907F809F7F7F90\
7F809F7F7F9080809F7F7F9080809F7F809080809F7F809080809F808090808F9F
808090808F9F808090808F9F808090808F9F808F908F8F9F808F908F8F9F808F908F909F808F90\
8F909F8F90908F909F8F90908F909F8F909090909F8F90908F909F8F9090909F9F
8F90908F90908F9090909F908F90908F90908F90904F4F4F404040404F4F808F80808F8F808F80\
808F80808F80808F80808F80808F80808F7F808F808F908090908F8F908090908F
8F90809090808F90809090808F90808F90808F90808F90808F8F7F8F90806F8F4F70904F1F8F00\
1F8F001F8F001F8F001F8F001F8F001F8F001F8F003080103F8F1F8080708F8F70
808070808F70808070808F7080806F80807080806F80806F7F806F80806F7F7F607F806F7F7F60\
7F7F60707F607F7F60707F5F1F7F001070001F7F001070001F70005F70306F7050
6F70506F70506F6F506F7050000000000000000000000000000000000000105F00105F00000000\
4F5030000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F3F10101F
0000001F1F2030304F2F304F303050303050303F5F3030503F3F5F303F5F3F3F5F303F5F3F4060\
3F00003F00003F000040000040000040202F4040604F4F6F404F6F4F4F6F4F4F6F
4F50704F4F705050704F507050507F50507050507F505070505F7F50507F5F5F7F505F7F5F5F7F\
5F5F7F5F5F7F5F5F7F5F60805F5F7F6060805F607F606080606080606080606080
606F8F5F607F6060800000000000000000000000000000000000000000000000000000000F0F0F\
0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F707090707F90707090
7F7F907070907F7F90707F907F7F9F7F7F907F7F9F7F7F907F809F7F7F9F7F809F7F7F9F80809F\
7F7F9F80809F7F7F9F80809F7F809F80809F7F809F80809F80809F808F9F80809F
808F9F80809F808F9F808F9F8F8F9F808F9F8F8F9F808F908F909F8F8F9F8F909F8F8F908F909F\
8F909F8F909F8F909090909F8F909F90909F8F9090909F9F8F9090909F9F8F9090
909F9F8F9090909F908F9090909F908F9090909F90404F404F4F4F000000000000000000000000\
0000000000000000008F8F80808F808F8F80808F808F8F80808F808F8F80808F7F
8F8F80808F7F90908F8F908090908F8F90809090808F90809090808F90808F90806F8F4F209000\
1F8F002090001F8F002090001F8F00208F001F8F003F8F1F308F1F508F3050802F
8F8F7F808F70808F70808070808F70808070808F707F806F80806F7F806F80806F7F7F607F806F\
7F7F607F8060707F601F7F001F7F001F7F001F70001F7F001F70005F7F3F5F7030
70705F6F70506F7050000000000000000000000000000000000000000000002F00105000000000\
202F1F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F301F1F1F
0000000000002F2F402F304F2F304F3030502F3050303F5F303050303F5F303F5F3F3F5F300F10\
3F10103F00003F00003F00004000003F4060404F6F404060404F6F404F6F4F4F6F
4F4F6F4F50704F4F6F4F50704F50705050704F5070505070505070505F7F505070505F7F505F7F\
5F5F7F505F7F5F5F7F5F5F7F5F60805F5F7F5F60805F607F6060805F6080606080
606080606F805F5F7F60607F0000000000000000000F0F0F00000F0F0F0F00000F0F0F0F00000F\
0F0F0F00000F0F0F0F6F708F7070906F708F70709070708F70709070708F707090
707090707F90707090707F907070907F7F90707F907F7F9F7F7F907F7F9F7F7F907F7F9F7F7F90\
7F809F7F7F907F809F7F7F9080809F7F7F9080809F7F809080809F7F809080809F
808090808F9F808090808F9F808090808F9F808F90808F9F808F908F8F9F808F908F909F808F90\
8F909F808F908F909F8F90908F909F8F90908F909F8F90908F909F8F90908F9090
8F90908F90908F90908F90908F90908F90908F908F404F4F404040000000000000000000000000\
00000000000000000000000000000000000000000000000000000080807F808F80
808F7F808F7F80807F808F7F8F90808F90808F8F808F90808F8F7F8F90808F8F7F808F7F7F8070\
1F90001F8F001F8F001F8F001F8F001F8F001F8F001F8F001F8F004F802F508F30
4F8040508F4F80806F808F7080806F808F707F806F80806F7F806F7F806F7F7F607F806F7F7F60\
7F80605F7F4F7F7F601F7F001F7F001070001F7F00107000407F3F4070306F705F
6F70506F70506F6F500000000000000000000000000000000000000000000F50001050000F4000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030304F2F304F3030502F3050303F5F303050303F5F303F5F3F3F5F300F103F101F\
3F00003F00003F00004000003F2F3F402F4040406F404F6F404F6F4F4F704F4F6F
4F50704F4F6F5050704F50705050704F5070505F7F505070505F7F5050705F5F7F505F7F5F5F7F\
505F7F5F60805F5F7F5F60805F5F7F6060805F60806060805F6080606F80606080
606F8F6060806060805F607F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F6F6F8F6F708F\
6F6F8F7070906F708F7070906F708F707090707090707090707090707F90707090
707F907070907F7F9F707F907F7F9F707F907F7F9F7F7F907F7F9F7F7F907F809F7F7F9F7F809F\
7F7F9F80809F7F809F80809F7F809F80809F7F809F80809F7F809F808F9F80809F
808F9F80809F808F9F808F9F808F9F808F9F8F8F9F808F9F8F8F9F808F908F909F8F8F9F8F909F\
8F8F908F909F8F909F8F909F8F90908F9F9F8F909F8F909F8F9090909F9F8F9090
8F9F9F8F9090909F9F8F9090909F908F90904F4F4F404F404F4F4F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000808F80808F7F808F7F80807F808F7F8F90808F90808F907F8F9080808070808F7F\
508F3050903F1F8F002090001F8F002090001F8F00208F001F8F00208F001F8F00
1F8F004F8040508F4F508040808F7080806F808F707F806F80806F7F806F80806F7F80607F806F\
7F7F606080501F7F001F7F001F7F001F7F001F7F001F7F002F70104F7F3F6F7050
6F705F6F7050000000000000000000000000000000000000000000000000105F00105000001000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000101F2030304F2F2F4F3030502F304F303F5F2F304F303050300F103F1010300000\
3F00003F00003F00003F0F104010103F4060404F6F404060404F6F404F6F4F4F70
404F6F4F50704F4F704F50704F50705050704F5070505F7F4F5070505F7F505070505F7F50507F\
5F5F7F505F7F5F5F80505F7F5F60805F5F7F5F60805F5F7F6060805F6080606080
5F6080606080505F705F5F7F505F700F0F0F606F806F6F8F606F806F6F8F6F6F8F6F708F6F6F8F\
6F708F6F6F8F6F70906F6F8F7070906F708F7070906F708F70709070708F707F90
707090707F90707090707F907070907F7F90707F907F7F9F707F907F7F9F7F7F907F809F7F7F90\
7F809F7F7F907F809F7F7F907F809F7F809080809F7F809080809F7F809080809F
7F809080809F808090808F9F808090808F9F808F90808F9F808F90808F9F808F908F909F808F90\
8F909F808F908F909F808F908F909F8F90908F909F8F90908F909F8F90908F909F
8F90908F90908F90908F90908F90908F9090404040404F4F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000808F7F80807F808F7F7F80708F9080808F7F8F9080808F7F\
8F907F508F301F90001F8F001F8F001F8F001F8F001F8F001F8F001F8F001F8F00
1F80001F8F001F8000508F404F80404F80407F806F80806F7F806F7F806F7F7F607F806F707F60\
4F7F3F407F3F1F7F00107F001F7F001070001F7F002F70102F70106F70506F705F
6F70506F7050000000000000000000000000000000000000000000000000103000105000000F00\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000001F1F202F2F403030502F3050301F2030000030101F3000003F10103000003F0000\
3F00003F00003F00004000003F4060404F6F404060404F6F404F6F4F4F6F404F6F
4F50704F4F704F50704F507050507F4F5070505F7F505070505F7F505F7F505F7F505F7F5F5F7F\
505F7F5F5F805F5F7F5F60805F5F805F60805F60806060805F60806060805F6080
5F60805F5F7F606080505F700F0F0F606F8F6F6F8F606F8F6F708F6F6F8F6F708F6F6F8F6F7090\
6F708F6F70906F708F7070906F708F7070906F708F707F90707090707F90707090
707F90707090707F90707F907F7F9F707F907F7F9F707F907F809F7F7F907F809F7F7F9F7F809F\
7F7F9F7F809F7F7F9F80809F7F809F80809F7F809F80809F7F809F80809F7F809F
808F9F80809F808F9F80809F808F9F808F9F808F9F808F9F8F909F808F9F8F909F808F9F8F909F\
808F9F8F909F8F90908F909F8F909F8F909F8F90908F9F9F8F90908F909F8F9090
8F9F9F8F90908F9F908F90908F9F90404F4F404F4F404F40000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000808F7F80807F808F7F7F80708F9080\
8F8F7F8F9080808F7F2090001F8F002090001F8F00208F001F8F001F8F001F8F00
208F001F8F001F8F001F8000208F0030802F308F307F806F6F805F7F806F7F806F307F107F7F60\
1F7F001F80001F7F001F7F001F7F001F7F001F7000307F1F2F7010707F5F6F7050
6F705F000000000000000000000000000000002F00104F00000000000000103F00105000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F304F2F2F4F3030502F0F0F300000300000300000300000300000300000\
3F00003000003F00003F40604040603F4060404F6F404060404F6F404F6F4F4F6F
404F6F4F50704F4F6F4F50704F507050507F4F507050507F505070505F7F505070505F7F505F7F\
5F5F7F505F7F5F5F7F5F5F7F5F60805F5F7F5F60805F6080606080505F705F5F7F
5F5F7F5F607F0000000F0F0F00000F606F8F606F806F6F8F606F8F6F6F8F606F8F6F6F8F6F6F8F\
6F708F6F6F8F6F708F6F6F8F7070906F708F7070906F708F7070906F708F707090
707090707F90707090707F907070907F7F90707F907F7F90707F907F7F9F707F907F7F9F7F7F90\
7F809F7F7F907F809F7F7F907F809F7F7F907F809F7F809080809F7F809080809F
7F8090808F9F7F8090808F9F808090808F9F808090808F9F808F90808F9F808F90808F9F808F90\
8F909F808F908F909F8090908F909F8090908F909F8F90908F909F8090908F9090
8F90908F90908F90908F90908F9090404F4F404040404F4F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007F807080807F7F8070\
808070808F7F808F7F808F7F8F8F7F808F701F8F001F8F001F8F001F8F001F8F00
1F80001F8F001F80001F8F001F80001F8F001F80101F801F60805030801F707F6030801F1F7F00\
1F8000107F001F7F00107F001F7F001070002F7F102F7010707F5F6F70506F705F
6F70500000000000000000000000000000000020001050000000000000000F4000001000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000001F1F2030304F2F304F303F50302030301010300F103F00003000003F0000\
300F103F101F3F40603F40603F4060404F6F40406F404F6F404F6F4F4F70404F6F
4F50704F4F6F4F50704F507050507F4F5070505F7F505070505F7F50507F505F7F505F7F5F5F7F\
505F7F5F60805F5F7F5F60805F5F7F5F6080505F7F5F5F7F505F7F5F607F000000
0000000000000000000000000F0F0F0F0F0F6F6F8F606F8F6F6F8F606F8F6F708F6F6F8F6F708F\
6F6F8F6F70906F708F7070906F708F7070906F70907070906F7090707F90707090
707F90707090707F90707F907F7F9F707F907F7F9F707F907F7F9F707F907F809F7F7F9F7F809F\
7F7F9F7F809F7F809F7F809F7F809F80809F7F809F80809F7F809F808F9F7F809F
808F9F7F809F808F9F80809F808F9F808F9F808F9F808F9F808F9F808F9F8F909F808F9F8F909F\
808F9F8F909F80909F8F909F8F90908F909F8F909F8F909F8F90908F909F8F9090
8F909F8F90908F9F9F8F90908F9090404F404F4F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
7F8070808F7F7F80707F8070808F7F8F907F808F7F808F7F1F8F00208F001F8F00
1F8F001F8F00208F001F8F001F8F001F80001F8F001F80001F80003080101F80001F80001F8000\
1F7F001F80001F7F001F7F001F7F001F7F002F7F10707F5F6F705F707F5F6F7050
000000000000000000000000000000002F001F6F00104F00000000000000000000001000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000001F1F2F2F2F4F30304F2F304F3020302F0F10300F10301020301F2F302F4F\
3F304F303F5F3F40603F3F603F40603F4060404F6F3F4060404F6F404F6F4F4F6F
404F6F4F4F704F4F704F50704F50704F50704F5070505F7F4F5070505F7F50507F505F7F505F7F\
505F7F505F7F5F5F80505F7F505F7F505F70505F7F505F70000000000000000000
00000000000000000000000000000F0F0F0F606F806F6F8F606F806F6F8F606F8F6F708F606F8F\
6F708F6F6F8F6F708F6F6F8F6F708F6F708F7070906F708F7070906F708F707F90
6F708F707F90707090707F90707090707F90707F907F7F9F707F907F7F9F707F907F809F707F90\
7F809F7F7F907F809F7F7F907F809F7F80907F809F7F80907F809F7F809080809F
7F809080809F7F8090808F9F7F8090808F9F808090808F9F808F90808F9F808F90808F9F808F90\
80909F808F908F909F808F908F909F8090908F909F8090908F909F8090908F909F
8090908F90908090908F9090809090404F4F404040000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000007F80707F80707F80707F8070808F7F808F70808F7F808F70808F7F
1F80001F8F001F80001F8F001F80001F80001F80001F80001F80001F80001F7F001F8000107F00\
1F7F00107F001F7F00107F001F7F00407030307F106F705F6F705F6F70506F705F
000000000000000000000000000000002F00106000106000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000030304F2F304F3030502F3050303F5F303050303F5F303F5F3F4060\
303F5F3F40603F3F603F40603F4060404F6F3F4060404F6F404F6F404F6F404F6F
4F50704F4F704F50704F507050507F4F5070505F7F4F5070505F7F505F7F505F7F505F7F5F5F80\
505F7F5F5F80505070505F7F505F70505F7F000000000000000000000000000000
0000000000000000000000000F0F0F0F0F0F0F0F0F606F8F6F708F606F8F6F708F606F8F6F7090\
6F6F8F6F70906F708F7070906F708F7070906F708F707F906F7090707F906F7090
707F90707090707F90707F907F7F9F707F907F7F9F707F907F809F707F907F809F7F7F907F809F\
7F7F9F7F809F7F7F9F7F809F7F809F7F809F7F809F80809F7F809F80809F7F809F
808F9F7F809F808F9F7F809F808F9F808F9F808F9F808F9F808F9F808F9F808F9F808F9F8F909F\
808F9F8F909F808F9F8F909F80909F8F909F8090908F909F8F90908F909F809090
8F909F8F90908F90908F90908F9090404F4F404F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000007F80707F80707F80707F80707F806F808F7F808F70
808F7F808F701F8F001F80001F8F001F80001F8F001F80001F80001F80001F80001F7F001F8000\
1F7F001F80001F7F001F7F005F7F3F407F3F6F705F707F5F6F705F6F7F5F000000
000000000000000000000000001F00106000106000106000105000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000101F2030304F2F304F3030502F3050303F5F303050303F5F303F5F\
3F4060303F5F3F40603F40603F40603F406040406F3F4060404F6F40406F404F6F
404F6F4F5070404F6F4F50704F50704F507F4F507050507F4F5070505F7F4F5070505F7F505F7F\
5050704F5070505F70505070000000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F00000F606F8F606F8F6F6F8F606F8F6F6F8F606F8F\
6F708F606F8F6F708F6F6F8F6F70906F708F6F70906F708F7070906F708F707090
6F7090707F906F7090707F90707090707F90707F90707F90707F907F7F9F707F907F7F9F707F90\
7F809F707F907F809F7F7F907F809F7F7F907F809F7F80907F809F7F80907F809F
7F809080809F7F809080809F7F8090808F9F7F8090808F9F808F90808F9F808F90808F9F808F90\
80909F808F9080909F808F908F909F808F908F909F8090908F909F808F908F9090
8090908F90908090908F9090404040404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007F806F7F8070707F6F7F806F
7F8070808F707F80701F8F001F80001F8F001F80001F80001F80001F8000107F001F80001F7F00\
1F7F00107F001F7F00107F005F7F3F5070306F7F5F6F705F6F7F5F6F7050000000
000000000000000000000000002F00106000106000106000104F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F304F303050303050303F50303050303F5F303F5F3F3F60\
303F5F3F40603F40603F40603F4060404F6F3F4060404F6F40406F404F6F404F6F
4F5070404F6F4F50704F50704F50704F5070505F7F4F507F505F7F50507F505F7F4F5070505070\
4F5070505F7F000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F0F0F0F606F8F6F6F8F606F8F6F708F606F8F6F708F\
6F6F8F6F70906F708F6F70906F708F7070906F70907070906F7090707F906F7090
707F90707090707F90707F90707F90707F907F7F9F707F907F7F9F707F907F809F7F7F9F7F809F\
7F7F9F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F809F808F9F7F809F
808F9F7F809F808F9F7F809F808F9F7F809F808F9F808F9F808F9F808F9F808F9F808F9F80909F\
808F9F8F909F808F9F8F909F80909F8F909F80909F8F909F8090908F909F809090
8F909F8090908F909F8F9090404F4F404040000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000007F80707F806F
7F8070707F6F808F707F80701F8F001F80001F8F001F80001F80001F80001F80001F7F001F8000\
2F7F101F8000407F3F707F606F7F5F707F606F705F6F7F5F6F705F000000000000
000000000000000000000000000000002F00106000106000104F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000001F1F2F2F304F3030502F3050303F50303050303F5F303F5F\
303F5F303F5F3F4060303F603F40603F406040406F3F4060404F6F404F6F404F6F
404F6F4F4F70404F6F4F50704F4F704F50704F507050507F404F6F4F50704F50704F5070000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000F0F0F0F606F80606F8F606F8F6F708F606F8F\
6F708F606F8F6F7090606F8F6F70906F708F6F70906F708F6F70906F708F707F90
6F708F707F906F7090707F906F7090707F90707F90707F9F707F90707F9F707F907F809F707F90\
7F809F707F907F809F707F907F809F7F80907F809F7F80907F809F7F80907F809F
7F80907F809F7F8090808F9F7F8090808F9F7F8090808F9F7F8090808F9F808F90808F9F808F90\
808F9F808F9080909F808F9080909F808F908F909F808F9080909F8090908F909F
808F908F90908090908F90903F4040404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000007F806F707F6F7F8F707F806F7F80701F80001F801F107F001F80001F7F001F801F107F10\
307F106F705F407F3F6F705F707F606F705F6F7F5F6F705F6F705F000000000000
000000000000000000000000000000002F00105F001060000F5F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001F1F2F2F304F303F50303050303F5F303F5F3F4060\
303F5F3F4060303F603F40603F40603F406F3F4060404F6F404F6F404F6F404F6F
4F5070404F704F50704F4F704F507F4F507050507F4F4F6F4F50704F5070000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F0F0F0F6F6F8F606F8F6F708F606F8F6F7090\
6F6F8F6F70906F708F6F70906F708F6F70906F708F707F906F7090707F906F7090
707F90707090707F90707F90707F9F707F90707F9F707F907F809F707F907F809F707F907F809F\
7F7F9F7F809F7F7F9F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F809F
808F9F7F809F808F9F7F809F808F9F7F809F808F9F7F8F9F808F9F808F9F808F9F808F9F80909F\
808F9F80909F808F9F8F909F80909F8F909F8090908F909F8090908F909F809090
8F909F8090908F9090809090404F4F404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000007F806F707F6F808F707F806F7F80701F801060804030801030801F707F60707F60\
6F705F707F60707F60707F606F7F5F707F606F705F6F7F5F000000000000000000
000000000000000000000000000000002F00106000105F00106000104F00000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000101F2030304F2F304F303F502F3050303F5F303F5F\
303F60303F5F3F4060303F603F40603F40603F406F3F4060404F6F3F406F404F6F
404F6F4F4F70404F6F4F50704F4F704F5070404F6F4F5070000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F00000F606F8F606F8F606F8F606F8F\
6F708F606F8F6F708F606F8F6F70906F708F6F70906F708F6F70906F708F6F7090
6F7090707F906F7090707F906F7090707F90707F90707F90707F90707F9F707F90707F9F707F90\
7F809F707F907F809F707F907F809F707F907F809F7F80907F809F7F80907F809F
7F80907F809F7F80907F809F7F8090808F9F7F8090808F9F7F8090808F9F7F8F90808F9F808F90\
808F9F808F9080909F808F9080909F808F9080909F808F9080909F808F90809090
808F90809090808F90809090404040404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000707F6F707F60707F6F70806F7F806F5F803F7F806F2F7F10707F60707F60\
6F70606F7F60707F606F7F5F707F606F705F6F7F5F000000000000000000000000
000000000000000000000000000000002F00105F00105F00105F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F2F303050303050303F5F303F5F303F5F\
303F5F3F4060303F603F40603F40603F4F6F3F4060404F6F3F406F404F6F404F6F
404F70404F6F4F50704F50704F5070404F6F4F5070000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F0F0F0F606F8F6F708F606F8F6F708F\
606F8F6F70906F708F6F70906F708F6F70906F70906F70906F7090707F906F7090
707F906F7090707F90707F90707F90707F90707F9F707F90707F9F707F907F809F707F9F7F809F\
707F9F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F8F9F7F809F
7F8F9F7F809F808F9F7F809F808F9F7F809F808F9F7F8F9F808F9F7F8F9F808F9F808F9F808F9F\
808F9F80909F808F9F80909F808F9F8F909F80909F80909F808F908F909F809090
8F909F8090908F909F809090404F4F404040000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000707F6F707F60707F6F70806F7F806F70806F7F806F707F6070806F\
707F60708060707F60707F606F7F5F707F606F705F000000000000000000000000
000000000000000000000000000000000000002F00105F00105F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F2F2F304F303050303050303F5F303F50\
303F5F303F5F3F4060303F5F3F40603F40603F406F3F4060404F6F3F406F404F6F
404F6F404F70404F6F4F5070404F6F404F6F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F00000F606F8F606F8F60708F606F8F\
6F708F606F8F6F7090606F8F6F70906F708F6F70906F708F6F70906F708F6F7F90
6F708F707F906F7090707F906F7090707F90707F90707F9F707F90707F9F707F9070809F707F90\
7F809F707F907F809F707F907F809F7080907F809F7080907F809F7F80907F809F
7F80907F809F7F80907F8F9F7F80907F8F9F7F8090808F9F7F8090808F9F7F8F90808F9F7F8F90\
808F9F7F8F9080909F808F9080909F808F9080909F808F90809090808F90809090
808F90809090808F908090903F4040404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F7F60707F60707F6070806F707F6070806F6F7F60\
707F606F7F60707F606F705F6F7F606F705F000000000000000000000000000000
000000000000000000000000000000000000002000105F000F5000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000001F202F303050303F50303F50303F5F\
303F5F3F4060303F5F3F40603F40603F406F3F4060404F6F3F406F404F6F404F6F
404F70404F6F4F5070404F6F404F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F0F0F0F60708F606F8F6F7090\
606F8F6F7090606F8F6F70906F708F6F70906F708F6F7F906F70906F7F906F7090
707F906F7090707F906F7090707F9F707F90707F9F707F9070809F707F9070809F707F907F809F\
707F9F7F809F707F9F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F809F
7F8F9F7F809F7F8F9F7F809F808F9F7F809F808F9F7F8F9F808F9F7F8F9F808F9F7F8F9F808F9F\
808F9F80909F808F9F80909F808F9F80909F808F9080909F80909080909F808F90
80909F809090809090809090404F4F40404F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F7F60707F6070806F70806F707F6070806F\
707F60707F606F7F60707F606F7F5F6F7F60000000000000000000000000000000
000000000000000000000000000000000000002F00105F00000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000001F1F203030502F3050303F50303F50\
303F5F303F5F304060303F603F40603040603F40603F4060404F6F3F406F404F6F
404F6F404F703F4060404F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000F0F0F0F606F8F606F8F606F8F\
6F708F606F8F6F708F606F8F6F7090606F8F6F70906F708F6F70906F708F6F7090
6F7090707F906F7090707F906F7090707F906F7090707F90707F90707F9F707F90707F9F707F90\
7F809F707F907F809F707F907F809F707F907F809F7080907F809F7080907F809F
7F80907F809F7F80907F809F7F80907F8F9F7F80907F8F9F7F8090808F9F7F8090808F9F7F8F90\
808F9F7F8F90808F9F7F8F9080909F7F8F9080909F808F9080909F7F8F90809090
808F90809090808F908090903F4040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006F70606F7F606F7F60707F606F7F60\
707F606F7F606F7F606F7F5F6F7F60000000000000000000000000000000000000
000000000000000000000000000000000000002000105F00000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000001F1F2F303F50303F50303F5F\
303F5F3F4060303F603F40603040603F406F3F40603F4F6F3F406F404F6F404F6F
404F703F4060404F6F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F0F0F0F606F8F6F708F\
606F8F6F709060708F6F709060708F6F70906F70906F70906F70906F7F906F7090
707F906F7090707F906F7F90707F906F7F90707F9F707F90707F9F707F9070809F707F9F70809F\
707F9F7F809F70809F7F809F70809F7F809F7F809F7F809F7F809F7F8F9F7F809F
7F8F9F7F809F7F8F9F7F809F7F8F9F7F809F808F9F7F8F9F808F9F7F8F9F808F9F7F8F9F808F9F\
7F8F9F80909F808F9F80909F808F9F80909F808F9080909F808F9080909F808F90
80909F808F9080909F40404F404F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F70606F7F60707F6070806F\
6F7F607080606F7F606F7F606F7F5F000000000000000000000000000000000000
000000000000000000000000000000000000002F00000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000001F1F2F303F50303050\
303F5F303F5F304060303F5F3F40603040603F406F3F40603F4F6F3F406F404F6F
3F40603F4060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F00000F606F8F606F8F\
60708F606F8F6F7090606F8F6F709060708F6F709060708F6F70906F708F6F7F90
6F708F6F7F906F7090707F906F7090707F906F7F90707F9F6F7F90707F9F707F9070809F707F90\
70809F707F907F809F707F907F809F7080907F809F7080907F809F7080907F809F
7080907F809F7F80907F8F9F7F80907F8F9F7F80907F8F9F7F80907F8F9F7F8F90808F9F7F8F90\
808F9F7F8F90808F9F7F8F90808F9F7F8F9080909F7F8F908090907F8F90809090
7F8F90809090808F90404F4F3F4040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F705F6F706060705F\
6F7F606F7F606F7F606F705F000000000000000000000000000000000000000000
0000000000000000000000000000000000000F5000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000001F202F303050303F5F\
303F5F3F3F5F303F5F3F40603040603F40603F40603F4F6F3F406F404F6F3F4060
404F6F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F0F0F0F606F8F6F7090\
606F8F6F7090606F8F6F709060708F6F709060708F6F7F906F70906F7F906F7090
707F906F7090707F906F7090707F9F6F7F90707F9F6F7F9070809F707F9070809F707F907F809F\
707F9F7F809F707F9F7F809F70809F7F809F70809F7F809F7F809F7F809F7F809F
7F8F9F7F809F7F8F9F7F809F7F8F9F7F8F9F7F8F9F7F8F9F808F9F7F8F9F808F9F7F8F9F808F9F\
7F8F9F808F9F7F8F9F80909F808F9F80909F7F8F9080909F808F9080909F808F90
80909F808F908090903F4040404F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000060705F6F7F60\
60705F6F7F606F7F606F7F60000000000000000000000000000000000000000000
000000000000000000000000000000000000105000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000001F1F2F303050\
303F5F303F5F303F5F303F5F3F4060303F603F40603F40603F4F6F30405F3F4060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000F0F0F0F606F8F\
60708F606F8F60708F606F8F6F7090606F8F6F709060708F6F709060708F6F7090
6F70906F7F906F70906F7F906F7090707F906F7090707F906F7F90707F9F6F7F90707F9F707F90\
70809F707F9070809F707F907F809F707F907F809F7080907F809F7080907F809F
7080907F809F7080907F809F7F80907F8F9F7F80907F8F9F7F80907F8F9F7F80907F8F9F7F8F90\
808F9F7F8F90808F9F7F8F90808F9F7F8F90808F9F7F8F9080909F7F8F90808F90
7F8F908090907F8F9040404F3F4040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
60705F606F5F6F7F60000000000000000000000000000000000000000000000000
000000000000000000000000000000002000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303F50\
303F503F3F5F303F5F3F4060303F603F406F3F40603F406F30405F3F4060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F0F0F0F60708F\
606F8F6F709060708F6F709060708F6F70906F70906F70906F70906F7F906F7090
6F7F906F7090707F906F7F90707F906F7F90707F9F6F7F90707F9F707F9070809F707F9F70809F\
707F907F809F70809F7F809F70809F7F809F70809F7F809F70809F7F809F70809F
7F809F70809F7F8F9F7F809F7F8F9F7F809F7F8F9F7F8F9F7F8F9F7F8F9F808F9F7F8F9F808F9F\
7F8F9080909F7F8F9F80909F7F8F9080909F7F8F9080909F7F8F9080909F7F8F90
80909F7F8F9080909F3F404F404F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F705F0F0F0F000000000000000000000000000000000000000000000000
0000000000000000000000000000000F5000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000001F1F2F\
303F50303F50303F5F303F5F304060303F603F4060303F5F3F4060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F00000F\
606F8F606F8F607090606F8F60709060708F6F709060708F6F709060708F6F7090
60708F6F7F906F70906F7F906F70906F7F906F7F90707F906F7F90707F906F7F90707F9F6F7F90\
707F9F707F9070809F707F9070809F707F9070809F707F9070809F7080907F809F
7080907F809F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F7F80907F8F9F7F8090\
7F8F9F7F8F907F8F9F7F8F907F8F9F7F8F90808F9F7F8F907F8F907F8F90808F90
7F8F90808F907F8F90404F4F3F4040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000060705F0F0F0F000000000000000000000000000000000000000000000000
0000000000000000000000000000000F4F00000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
1F1F2F303F5F303F5F3F4060303F603F4060303F5F3F4060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000F0F0F0F0F\
606F8F607090606F8F6F709060708F6F709060708F6F70906F70906F7F906F7090
6F7F906F70906F7F906F7090707F9F6F7F90707F906F7F90707F9F6F7F90707F9F6F7F9070809F\
707F9070809F707F9070809F70809F70809F7080907F809F70809F7F809F708090
7F8F9F70809F7F8F9F7080907F8F9F7F809F7F8F9F7F80907F8F9F7F8F9F7F8F9F7F8F907F8F9F\
7F8F9F7F8F9F7F8F9080909F7F8F90808F9F7F8F9080909F7F8F9080909F7F8F90
80909F7F8F90404F4F3F4040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
6F7F606F7F600F0F00000000000000000000000000000000000000000000000000
000000000000000000000000105000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
1F1F2F2F3050303F5F303F5F304060303F5F303F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000F\
0F0F0F606F8F60708F606F8F607090606F8F60709060708F6F709060708F6F7090
6070906F7F9060708F6F7F906F70906F7F906F70906F7F906F7F90707F906F7F90707F906F7F90\
70809F6F7F9070809F707F9070809F707F9070809F707F9070809F707F9070809F
7080907F809F7080907F809F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F7F8090\
7F8F9F7F80907F8F9F7F8F907F8F9F7F8F907F8F907F8F907F8F907F8F907F8F90
7F8F907F8F903F40403F404F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000006F7F60\
6F7F6060705F0F0F0F000F00000000000000000000000000000000000000000000
0000000000000000000000000F4F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000001F2030303F5F303F5F303F5F3F4060303F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000F0F60709060708F60709060708F6F709060708F6F709060708F6F7F906F7090
6F7F906F70906F7F906F70906F7F906F7090707F906F7F90707F906F7F90707F9F6F7F9070809F\
6F7F9070809F707F9070809F707F9070809F70809070809F7080907F809F708090
7F809F7080907F8F9F7080907F8F9F7080907F8F9F7F80907F8F9F7080907F8F9F7F8F907F8F9F\
7F8F907F8F9F7F8F907F8F9F7F8F907F8F9F7F8F907F8F9F7F8F9080909F7F8F90
7F8F903F4040404F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000006F7F60\
60705F6F7F6060705F0F0F0F000000000000000000000000000000000000000000
000000000000000000002000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F203030305F303F5F303F50303F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000F60708F606F8F60708F606F8F607090606F8F60709060708F6F7090
60708F6F709060708F6F7F9060708F6F7F906F70906F7F906F70906F7F906F7F90707F906F7F90\
707F906F7F9070809F6F7F9070809F707F9070809F6F7F9070809F707F9070809F
70809070809F70809070809F7080907080907080907F8F9F7080907F8F907080907F8F9F708090\
7F8F907F8F907F8F9F708F907F8F907F8F907F8F907F8F907F8F907F8F907F8F90
7F8F903F404F3F4040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006F7F6060705F\
60706060705F60705F000000000000000000000000000000000000000000000000
000000000000000000104F00000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F1F2F303F5F303F5F303F5F303F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F606F8F60709060708F60709060708F6F709060708F6F709060708F
6F7F906070906F7F906F70906F7F906F7F906F7F906F7F906F7F906F7F90707F906F7F9070809F\
6F7F9070809F6F7F9070809F707F9070809F707F9070809F70809070809F708090
70809F70809070809F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F\
7F8F907F8F9F7F8F907F8F9F7F8F907F8F9F7F8F907F8F9F7F8F907F8F907F8F90
404F4F3F404F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F7F606070606F7F60\
60705F60705F60705F0F0F0F0F0F0F000000000000000000000000000000000000
0000000000000000000F4000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F1F2F303F5F2F3050000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000F0F0F0F5F6F8F60708F5F6F8F60708F60708F60709060708F607090
60708F6F709060708F6F709060708F6F7F9060708F6F7F906F70906F7F906F70906F7F906F7F90\
707F906F7F90707F906F7F9070809F6F7F907080906F7F9070809F6F7F90708090
70809070809F70809070809070809070809F7080907080907080907F8F9F7080907F8F90708090\
7F8F907080907F8F90708F907F8F907080907F8F90708F907F8F90708F907F8F90
3F40403F404F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F7F6060705F607F6060705F\
60705F5F705F60705F5F6F500F0F0F000000000000000000000000000000000000
000000000000001F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000001F1F2F303F5F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F0F0F0F606F8F60709060708F60709060708F6F709060708F
6F709060708F6F7F906070906F7F906070906F7F906F7F906F7F906F7F906F7F906F7F906F7F90\
6F7F9070809F6F7F9070809F6F7F9070809F707F9070809F707F9070809F708090
70809F70809070809F70809070809F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F\
7080907F8F9F7F8F907F8F9F708F907F8F9F7F8F907F8F907F8F907F8F903F404F
3F404F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006F7F6F607F606F7F6060705F607F60\
60705F60705F5F705F5F705F000F00000000000000000000000000000000000000
000000000000104F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F1F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F00000F60708F5F6F8F60708F5F6F8F60708F60708F607090
60708F60709060708F6F709060708F6F7F9060708F6F7F9060708F6F7F906F7F906F7F906F7F90\
6F7F906F7F906F7F906F7F906F7F906F7F907080906F7F907080906F7F90708090
6F7F90708090708090708090708090708090708090708090708090708090708090708F90708090\
708F907080907F8F907080907F8F907080907F8F907080907F8F90708F903F404F
3F4040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F7F6F607060607F6060705F6070605F705F\
60705F5F6F505F705F5F6F500F0F0F000000000000000000000000000000000000
0000000F3F00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000F0F0F0F1F1F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F0F0F0F60708F606F8F60709060708F60709060708F
6F709060708F6F709060708F6F7F906070906F7F906070906F7F906F7F906F7F906F7F906F7F90\
6F7F906F7F906F7F9070809F6F7F9070809F6F7F9070809F6F7F9070809F6F7F90
70809F70809070809070809070809F708090708090708090708F9F708090708F907080907F8F9F\
7080907F8F907080907F8F9F708F907F8F90708F907F8F90708F903F404F3F4040
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000607F606F7F60607060607F6060705F607060\
5F705F60705F5F6F505F705F5F6F500F0F0F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000F0F2F30501F1F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000F0F0F0F5F6F8F60708F5F6F8F60708F5F6F8F60708F
60708F60709060708F60709060708F6F7F9060708F6F7F9060708F6F7F9060708F6F7F906F7F90\
6F7F906F7F906F7F906F7F906F7F906F7F906F7F906F7F907080906F7F90708090
6F7F907080906F7F907080906F80907080906F8090708090708090708090708090708090708090\
708F90708090708F907080907F8F90708090708F907080907F8F903040403F404F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000607F60607F60607060607F605F705F60705F5F705F\
5F705F5F6F505F6F50506F505F6F50000000000000000000000000000000000000
0F4000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000303F5F303F5F1F2030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000F0F0F0F0F5F6F8F60708F60708F60709060708F
60709060708F607F9060708F607F9060708F6F7F906070906F7F906070906F7F906F7F906F7F90\
6F7F906F7F906F7F906F7F906F7F9070809F6F7F907080906F7F9070809F6F8090
7080906F7F9070809F708090708090708090708090708090708090708090708F90708090708F90\
7080907F8F907080907F8F907080907F8F90708F907F8F907080903F404F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000607F606F7F6F607F60607F6060705F607F605F705F60705F\
5F6F505F705F5F6F505F6F50506F500F0F0F0F0F000000000000000000000F3F00
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000F2F3050303F5F2F30501F1F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000F60708F5F6F8F60708F5F6F8F60708F
5F6F8F60708F60708F60709060708F60708F60708F607F9060708F6F7F9060708F6F7F9060708F\
6F7F906F7F906F7F906F7F906F7F906F7F906F80906F7F906F7F906F7F90708090
6F7F906F80906F7F907080906F7F907080906F80907080906F8090708090708090708090708090\
708090708090708F9070808F708F90708090708F9070808F708F903F4040000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000607F60607F60607060607F605F705F60705F5F705F5F705F5F6F50\
5F6F5F506F505F6F5050604F506F5050604F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F00000F303F5F2F3050303F5F303F5F1F2030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F000F0F60708F5F708F60708F5F708F
60709060708F60709060708F607F9060708F607F9060708F6F7F906070906F7F90607F8F6F7F90\
6F7F906F7F906F7F906F80906F7F906F80906F7F907080906F7F907080906F7F90
7080906F80907080906F80907080906F80907080906F8090708090708090708090708090708F90\
708090708F90708090708F90708090708F907080907F8F90708F903F404F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000607F6F6F7F6F607F60607F60607060607F605F705F60705F5F6F5F5F705F\
5F6F505F6F50506F505F6F5050604F506F4F000000000000000000104F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F30502F30503030502F3050303F5F303F5F1F202F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F5F6F8F60708F5F6F8F60708F
5F6F8F60708F5F708F60708F60708F60709060708F60708F60708F607F9060708F607F9060708F\
6F7F9060708F6F7F90607F906F7F90607F906F7F906F7F906F80906F7F8F6F8090
6F7F906F80906F7F8F6F80906F7F907080906F7F8F7080906F80907080906F808F7080906F8090\
7080906F808F70809070808F708F9070808F708F9070808F3F404F303F40000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000607F60607F6F607060607F605F705F6070605F705F5F705F5F6F5F5F705F506F50\
5F6F50506F50506F5050604F50604F000000000000000000104000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000303F502F3050303F5F303F50303F5F303F5F
1F2030000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F0F0F0F5F6F8F60708F5F708F
60708F5F708F60709060708F60709060708F607F9060708F607F9060708F6F7F90607F8F6F7F90\
607F8F6F7F906F7F906F7F906F7F906F80906F7F906F80906F7F906F80906F7F90
6F80906F7F907080906F80907080906F80907080906F80907080906F8090708090708090708090\
708090708F90708090708F90708090708F90708090708F903040403F404F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006F7F6F607F60607F6F607060607F605F705F6070605F705F5F705F5F6F5F5F6F5F\
506F505F6F50506F50506F504F604F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F304F2F30502F3050303F5F2F3F50303F5F
303F5F1F2030000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000F5F708F5F6F8F60708F
5F6F8F60708F5F708F60708F5F708F60708F60708F60709060708F60708F60708F607F9060708F\
607F9060708F6F7F90607F8F6F7F90607F906F7F90607F8F6F7F906F7F906F8090
6F7F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F7080906F7F8F6F80906F808F7080906F808F\
7080906F808F7080906F808F7080906F808F708F906F808F3F4040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000607F6F607F60607F6F5F70606070605F705F5F70605F6F5F5F705F506F505F6F5F506F50\
506F50506050506F504F5F4F0F0F0F0000000000000F4F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F2F304F303F5F2F3050303F5F303F5F
304060303F5F1F2030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F0F0F0F5F708F5F6F8F
60708F5F708F60708F5F708F60709060708F60709060708F607F9060708F607F9060708F6F7F90\
607F8F6F7F90607F8F6F7F906F7F906F7F906F7F906F80906F7F906F80906F7F90
6F80906F7F906F80906F7F907080906F80907080906F80907080906F80907080906F8090708F90\
6F80907080906F808F708F90708090708F906F808F708F903F4040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6F806F607F6F607F6F607060607F605F70606070605F705F5F705F5F6F5F5F705F506F505F6F5F\
506F50506F504F604F4F604F000F00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F304F2F304F2F3F502F3050303F5F
2F3F5F303F5F303F5F1F2030000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F00000F5F708F
5F6F8F60708F5F6F8F60708F5F708F60708F5F708F60708F60708F607F8F60708F607F8F60708F\
607F9060708F607F90607F8F6F7F90607F8F6F7F90607F8F6F7F90607F8F6F7F90
6F7F8F6F8090607F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F808F\
7080906F808F7080906F808F7080906F808F7080906F808F3F404F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000607F6F
607F60607F6F5F7060607F605F70605F70605F6F5F5F705F506F5F5F705F506F50506F50506050\
506F504F604F4F604F405F40000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F30502F3050303F5F2F3F50
303F5F303F5F30406030405F1F2030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000F0F0F0F0F5F6F8F
5F708F5F6F8F60708F5F708F60708F5F708F60709060708F607F9060708F607F9060708F607F90\
60708F6F7F90607F8F6F7F90607F8F6F7F90607F906F7F90607F8F6F80906F7F90
6F80906F7F8F6F80906F7F906F80906F7F8F6F80906F80906F80906F808F7080906F8090708090\
6F808F708F906F808F7080906F808F708F906F808F708F90304040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006F7F6F607F6F
607F6F607060607F6F5F70606070605F705F5F70605F705F5F705F506F505F6F5F506F50506F50\
4F604F4F604F405F400F0F0F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F402F304F2F30502F3F5F
2F3050303F5F2F3F5F304060303F5F1F2030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000F0F0F0F
5F6F805F708F5F6F8F5F708F5F6F8F60708F5F708F60708F5F708F60708F5F708F607F8F5F708F\
607F8F60708F607F9060708F607F90607F8F6F7F90607F8F6F7F90607F8F6F7F90
607F8F6F7F90607F8F6F8090607F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F\
6F80906F808F6F80906F808F6F808F6F808F7080906F808F304040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000607F6F607F6F607F6F607F6F
5F7060607F605F70605F70605F705F5F705F506F5F5F705F506F50506F5F506F50506F504F604F\
4F604F405F40405F40000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F2F304F2F30502F3050
303F5F2F3F5F303F5F303F5F3040603040601F2030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000F0F
5F708F5F6F8F5F708F5F708F60708F5F708F60708F5F708F607F9060708F607F9060708F607F90\
60708F607F9060708F6F7F90607F8F6F7F90607F8F6F7F90607F8F6F7F90607F8F
6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F808F6F80906F808F708090\
6F808F6F80906F808F7080906F808F7080906F808F3F404F304040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000607F6F607F6F607F6F607F6F5F7060
607F6F5F7060607F605F70605F70605F705F5F705F506F5F5F705F506F50506F504F604F4F6050\
405F4F405F4F405040000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F304F2F2F4F2F3050
2F30502F3F5F2F3F50303F5F2F3F5F304060303F5F1F2030000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
00000F5F708F5F6F805F708F5F6F8F5F708F5F6F8F5F708F5F708F60708F5F708F60708F5F708F\
607F8F5F708F607F8F60708F607F9060708F607F90607F8F6F7F90607F8F607F90
607F8F6F7F90607F8F6F7F90607F8F6F8090607F8F6F80906F7F8F6F8090607F8F6F80906F808F\
6F80906F7F8F6F808F6F808F6F808F6F808F6F808F303F403F4040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000607F6F607F6F5F706F607F6F5F70605F7F60
5F70605F70605F705F5F7060506F5F5F705F506F50506F5F506F50506F504F604F4F604F405F40\
405F403F503F000F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030304F2F304F
303F502F3050303F5F2F3F5F30405F303F5F3040603040601F2030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F5F6F8F5F708F5F708F5F708F5F708F60708F5F708F60708F5F708F607F9060708F607F90\
60708F607F90607F8F607F90607F8F607F90607F8F607F90607F8F6F7F90607F8F
6F7F90607F8F30404F303F4030404F303F403F404F303F4030404F303F403F404F303F4030404F\
303F406F80906F808F6F80906F808F6F80906F808F3F404F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000607F6F607F6F607F6F607F6F5F7060607F6F5F7060
5F70605F70605F70605F705F5F7060506F5F5F705F506F5F506F5F4F60504F6050405F4F405F4F\
405040405F40000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F402F304F
2F304F2F30502F30502F3F5F2F3F5F303F5F2F3F5F3040603040601F2030000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F5F6F805F708F5F6F805F708F5F708F5F708F5F708F5F708F5F708F60708F5F708F\
60708F2F3F40303F402F3F40303F402F3F40303F402F3F40303F40303F40303F40
2F3F40303F40303F40303F40303F40303F40303F40304040303F40304040303F40304040303F40\
304040303F40304040303F40304040303F40304040303F403F404F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F706F607F6F5F706F607F6F5F70605F70605F70605F7060
50705F5F7060506F5F5F7060506F5F506F5F506F50506F504F604F4F604F405F4F405F4F3F5040\
3F5040000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F30402F2F40
30304F2F304F303F502F3F5030405F2F3F5F30406030405F3040603040601F2030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F5F6F8F5F708F5F708F5F708F5F708F303F4F2F3F40303F402F3F40303F4F\
2F3F40303F40000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F1000001F00002F00002F0000\
2F00002F00002F00001F00000000000F0F0F30404F6F808F3F404F3F404F304040
3F4F4F3F404F3F4F4F3040403F4F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000607F705F7F6F607F6F5F7F6F607F6F5F7060607F6F5F70605F70605F7060
5F706050705F5F7060506F5F5F705F506F5F506F5F4F60504F6050405F4F405F4F405F40405F40\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F40
2F2F402F304F2F304F2F3F502F3F502F3F5F2F3F5F3040602F405F3040603040601F20301F2030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000F303F4F2F30402F3F402F30402F3F402F3040000000000000000000000000\
0000000000000000005F00006000005F00006000005F00006000003F00003F0000
0000000000000000000000000000000F0F0F2000002F00006000006000004F00005000003F0000\
2F00003F00006F00006000006F00006000003F404F607F8F6F808F607F806F808F
6080806F808F607F806F808F3040403F404F3040403F4040000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
607F705F706F607F6F5F706F5F7F6F5F70605F70605F70605F7060506F5F5F7060
506F5F5F7060506F5F506F5F506F5F506F5F4F60504F6050405F4F405F4F3F50403F5040000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F30402F2F4030304F2F304F303F502F3F5030405F2F3F5F304060304060304060304060304F6F\
1F2030000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
303F4F2F3F40303F4F2F3F400000000000000000000000000000000000000000004F00005F0000\
5F00006000005F00006000005F00006000005F00003F0000000000000000000000
0000000000000000000000000F0F0F4F00006000006000006000002F0000000000000000000000\
0000000000000000004000006000006F000060303F607F8060707F6F6F706F0000
6F00006000006F00006F80806F808F607F806F808F6080803F404F3040403F404F304040000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000607F70\
5F7F6F607F705F706F5F7F6F5F70605F7F6F5F70605F70605070605F7060507060
5F7060506F5F507060506F5F506F5F4F60504F6F5040604F40604F405F40405F40000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F402F2F402F304F2F304F2F3F502F3F502F405F2F3F5F3040602F405F304060304060\
3040601F2030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F3F402F30402F3F4F
2F3040506F80506F805F6F8F4F607F000000000000000000000000000000000000000000000000\
0000001000004F00004000005F00005F0000600000100000100000000000000000
0000000000000000005000004F00006000006000002F0000200000000000000000000000000000\
0000000000000000000000006000006000006F00006000006F808F607F806F808F
60606F6F00006000006F0000600000600000606060608080607F80608080607F803F4040304040\
304040304040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000607F705F706F\
5F7F6F5F706F5F7F6F5F70605F70605070605F7060506F5F507060506F5F507060
506F5F506F5F4F6F5F506F5F4F60504F6050405F4F405F4F3F50403F5F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F30402F30403030502F3050303F502F3F5F3040602F3F5F3040602F405F304F60\
304060304F6F1F2030000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F30402F3F4F2F3040506F80506F80
506F8F506F805000005000005F0000502F30000000000000000000000000000000000000000000\
0000000000000000001F00005F00005F00005F0000600000200000000000000000
0000000000002F00005F0000600000600000500000000000000000000000000000000000000000\
000000000000000000501F20501F1F6F00006000006F5050604F506F808F607F80
6F50506000006F1F20607F80608080607F80608080607F806F8080607F80608080607F7F6F8080\
607F803F40403040403F4040304040000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000607F705F706F607F70\
5F706F5F7F6F5F70605F7F6F5F70605F70605070605F7060506F60507060506F5F
507060506F5F506F5F4F60504F6F5040604F40604F405F40405F4F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F402F304F2F304F2F3F502F3F50303F5F2F3F5F3040602F405F\
3040603040603040601F20301F2030000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002030402F30402030405060804F607F506F804F607F500000
500000500000500000506F80506F8050607F4F607F000000000000000000000000000000000000\
0000000000000000000000005F00005F00005F00005F00002F0000000000000000
0000000000004F00005F00005F0000600000200000000000000000000000000000000000000000\
00000000000030404F303F40604F5060000060000060303060808F607F80608080
604F4F601F1F607F80607F805F7F80607F805F7F7F607F80607F7F607F805F7F7F606060600000\
608080607F7F607F7F607F7F304040303F3F304040000000000000000000000000
0000000000000000000000000000000000000000000000005F7F6F5F7F705F706F5F7F6F5F706F\
5F706F5070605F70605070605F7060506F5F507060506F5F5070604F6F5F506F5F
4F6F50506F5F4F60504F6050405F4F40604F3F5F403F5F40000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F304F303F502F3050303F5F2F3F5030405F2F3F5F304060\
2F405F304F60304060304F6F3040601F2F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F30402F30402F30404F607F506F804F60805000004F0000500000500000
50000050000050303F502F30506F80506F80506F7F000000000000000000000000000000000000\
0000000000000000000000004000004F00005F00005F0000200000000000000000
0000001000006000005F0000600000000000000000000000000000000000000000000000000000\
000000000000304040607F8F607F80603F3F601F1F60808F607F806F808F607F80
601F205F707F608080607F80607F805F7F7F607F80607F7F607F805F7F7F608080607F7F600000\
600000600000600000608080607F7F608080304040304040000000000000000000
0000000000000000000000000000000000000000005F7F70607F705F706F5F7F705F706F5F7F6F\
5F70605F706F5070605F7060506F605F7060506F60507060506F5F506F5F4F6F5F
506F5F4F60504F6F50406050406050405F4F405F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F402F30502F304F2F3F502F3F50303F5F2F3F5F\
30405F2F405F3040602F4060304F603040601F2030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
20304020303F4F607F405F7F4F607F4F4F604F60804F00004F607F4F2F30500000
4F00005000004F000050506F4F607F506F80506F8050607F000000000000000000000000000000\
0000000000000000000000000000003000005F00005000005F0000300000000000
0000002000005F00005F0000400000000000000000000000000000000000000000000000000000\
00000000000030404F5F7F80607F80601F1F601F1F5F707F607F80607F80607F80
604F4F607F805F7F7F5F7F805F707F5F7F805F707F607F805F707F5F7F7F5F7F7F607F7F5F7F7F\
604F4F5F00006000005F00006000005F7F7F607F7F303F3F304040303F3F000000
0000000000000000000000000000000000005F706F5F7F705F706F5F706F50706F5F706F507060\
5F706F507060507060506F60507060506F5F506F604F6F5F506F5F4F60504F6F5F
4060504F605040604F40604F3F5F403F5F40000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F304F303F502F3050303F5F2F3F5030405F\
2F3F5F3040602F405F304F60304060304F603040603F4F6F1F2030000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F3040\
2030404F607F405F7F4F00004F00004F00004F607F5060804F607F506F804F607F
5000004F00005000004F0000504050506F80506F80506F80506F7F000000000000000000000000\
0000000000000000000000000000002000005000005F00005F00004F0000000000
0000000000005F00005F00004F0000000000000000000000000000000000000000000000000000\
000000000000607F80607F8F607F8060606F60606F60808F607F80607F805F7F7F
601F205F7F80607F805F707F5F7F805F7F7F5F7F805F707F607F805F7F7F5F7F7F5F7F7F5F7F7F\
5F0000600000600000606F606000006000005F7F70607F7F5F7F70304040303F3F
3040400000000000000000000000005F7F705F7F705F706F5F7F705F706F5F706F5070605F706F\
507060507060506F60507060506F60506F604F6F5F506F5F4F6F5F4F6F5F4F6050
4F6050406050406050405F4F405F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F304F2F30502F30502F3F502F3F50\
303F5F2F3F5F30405F2F405F3040602F4060304F60304060304F6F304F601F2F30
000000000000000000000000000000000000000000000000000000000000202F3F20303F202F3F\
405F7F405F704F00004000004F101F405F704F607F4F607F4F607F4F607F4F607F
4F607F4F1F1F4F00004F00004F00005040504F607F506F804F60704F607F000000000000000000\
0000000000000000000000000000001000001000005000005F0000400000000000
0000000000002000005F0000400000000000000000000000000000000000000000000000000000\
0000002F3F40607F805F7F80607F805F7F8060606F5F7F80607F805F7F80604F50
5F4F4F5F7F8050707F5F707F5F707F5F7F7F50707F5F7F7F5F707F5F7F7F5F5F5F5F60605F0000\
5F00005F7F7F607F7F5F7F70607F7F5F7F705F7F7F5F7F70607F7F5F70705F7F70
2F3F3F30403F0000000000005F706F5F7F7050706F5F706F50706F5F706F50706050706F506F60\
507060506F60506F604F6F5F506F604F605F4F6F5F4F60504F605F406050406050
405F4F4060503F5F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F304F303F502F3050303F50\
2F3F5030405F2F3F5F3040602F405F304F60304060304F60304F603F4F6F304F6F
1F2F30000000000000000000000000000000000000000000000000202F3F203040405F70405F7F\
405F704F00004000004F00004000004F3F4F4F607F4F607F4F607F4F60804F607F
4F60804F607F501F1F4F00005000004F2F30502F3F4F6F80506F804F607F000000000000000000\
0000000000000000000000000000001000001000005F00005000005F0000500000
0000000000002F00005F00004F0000000000000000000000000000000000000000000000000000\
00000030404F5F7F80607F805F7F80607F805F7F80607F805F7F80607F805F1F1F
5F707F5F707F5F7F805F707F5F7F7F50707F5F7F7F5F707F5F7F7F5F404F5F4F4F5F00005F0000\
5F7F7F607F7F5F7F7F607F7F5F7F70607F7F5F7F70607F7F5F7F70607F7F5F7F70
5F7F702F3F3F303F3F5F70705F7F705F706F5F7F7050706F5F706F5070605F706F507060507060\
506F60507060506F60506F604F6F5F506F5F4F605F4F605F4060504F6050405F50
406050405F4F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F304F2F30502F304F\
2F3F502F3F50303F5F2F3F5030405F2F405F3040602F4060304F60304060304F6F
304F603F4F6F1F20300000000000000000000000000000001F2F3F202F3F3F5070405F703F0000\
400000400000403F4F4000004000004000004F607F405F704F607F40607F4F607F
40607F4F607F4F607F4F404F4F00004F00004F101F506F804F607F506F804F6070000000000000\
0000000000000000000000000000000000001000005000005000005000005F0000
0000000000002000005F0000400000000000000000000000000000000000000000000000000000\
0000002F3F405F7F805F70805F7F805F7080607F805F707F5F7F805F707F5F4F4F
50707F5F707F50707F5F707F50707050707F5070705F707F502F2F5F30305000005F7F7F5F7070\
5F7F7F5F70705F7F7F5F70705F7F705F70705F7F705F70705F7F705F70705F7F70
0000000000002F3F302F3F3F50706F5F706F50706F5F706F506F6050706F506F605070604F6F60\
506F604F6F5F4F6F604F605F4F6F5F4060504F6050405F50406050405F4F405F4F
3F5F4F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F304F303F50\
2F3050303F502F3F5030405F2F3F5F30405F2F405F304F60304060304F60304F60
3F4F6F304F6F3F506F1F20300000000000000000001F2F3F20303F3F5070405F703F5070405F7F\
405F70405F7F405F70404F604000004F00004000004F607F40607F4F607F40607F
4F607F4F607F4F607F4F607F4F6F804F00004F00004F0000506F804F6F7F506F804F6070000000\
000000000000000000000000000000000000000000500000400000500000500000
1000000000000000005000005F0000000000000000000000000000000000000000000000000000\
0000005F7F805F7080607F805F7F80607F805F7F80607F805F7F805F707F5F707F
5F7F8050707F50707F50707F5F707F506F70507070501F1F5F00005000005F7F7F5F7070607F7F\
5F7F7F5F7F7F5F70705F7F7F5F70705F7F705F70705F7F705F7070000000000000
1F00000000000000002F3F305F707050706F5F706F50706050706F507060507060506F60507060\
4F6F5F4F6F604F605F4F6F5F4F60504F605F406050406050405F4F405F503F5F4F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F30502F304F2F3F502F3F50303F5F2F3F5030405F2F405F3040602F405F304F60
304060304F6F304F60304F6F304F6F1F2F30000000202F3F1F2F3F3F5070304F603F50703F5070\
405F703F5070405F70405F70403F4F40000040000040202F403F4F405F704F607F
405F7F4F607F40607F4F607F4F607F4F607F405F704F00004F00004F607F4F607F4F60704F6070\
000000000000000000000000000000000000000000000000400000500000500000
200000000000000000200000500000400000000000000000000000000000000000000000000000\
00000050707F5F7F805F70805F7F805F707F5F7F805F707F5F7F80500000501F1F
50101F501F1F4F6F7050707F50404050404F4F0000500000505F5F5F707F5070705F7F7F5F7070\
5F7F7F5070705F7F705070705F7F705070705F7070000000000000000000000000
5000000000000000002F3F302F303050706F506F6050706F506F605070604F6F604F6F604F6F5F\
4F6F5F40605F4F605F406050406050405F504060503F5F4F405F4F3F5040000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003030502F3050303F502F3F5030405F2F3F5F30405F2F405F304F60304060
304F60304F603F4F6F304F6F3F506F304F6F3F50701F2F303F50703F506F4000003F506F405F70\
3F5070405F70405F70405F70405F70405F7F4000004000004000004F5060405F7F
4F607F40607F4F607F4F607F4F607F4F607F4F6F7F4F607F4F00004F000050506F4F6F7F4F607F\
000000000000000000000000000000000000000000000000000000500000500000
5000000000000000001000005000005000000F0F0F000000000000000000000000000000000000\
2F3F405F7F805F70805F7F805F70805F7F805F70805F7F80502F305F3030505F60
505F604F101F501F1F501F1F501F1F4F00005000005070705F707F5070705F7F7F5F70705F7F7F\
5F70705F7F7F5F70705F7F705070705F7F70000000000000000000000000000000
5000005000000000000000002F3F302F303050706F506F6050706F4F6F60506F604F6F5F4F6F5F\
4F605F4F6F5F406050406050405F504060503F5F4F3F5F4F3F5040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F304F2F304F2F3F502F3050303F5F2F3F5030405F2F3F5F304060
2F405F304F60304060304F6F304F60304F6F1F20301F2F300000003F00003F00003F50703F506F\
3F50703F506F405F703F506F405F703F5F70405F703F5F7040000040000040607F
405F7040607F405F704F607F405F704F607F40607F4F607F40607F4F101F4F00004F404F4F607F\
4F6070000000000000000000000000000000000000000000000000300000500000
4F0000400000000000000000500000500000200000000000000000000000000000000000000000\
2F3F4050707F5F708050707F5F70805F707F5F7F8050707F501F1F506F7F50707F
4F6F704F5F604F101F5000004F00005000004F6F7050707F5070705F707F5070705F707F507070\
5F70705070705F707050706F000000000000000000000000000000000000400000
4F00005000004F00000000002030302F30304F6F60506F604F6F5F4F6F604F605F4F6F5F406050\
40605F4060504060503F5F4F405F503F5F4F3F5F4F1F201F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000030304F2F304F2F3F502F3F50303F5F2F3F5030405F2F405F
304060304060304F60304F60304F6F304F601F2F300000000000003F00003F00003F00003F5070\
3F506F3F5F703F506F405F703F5F70405F703F5F70405F70405F70400000400000
401F1F405F704F607F405F704F607F40607F4F607F40607F4F6F7F4F607F4F6F7F4F00004F0000\
4F607F4F6070000000000000000000000000000000000000000000000000200000
500000400000000000000000300000500000300000000000000000000000000000000000000000\
2F3F405F708050707F5F7F805F707F5F7F8050707F5F7080505F6050707F4F6F70
50707F4F00004F00004F000050404F506F7050707F5070705F7F7F5070705F7F7F5070705F7F7F\
5070705F7070000000000000000000000000000000000000000000500000200000
400000100000500000000000000000203030506F604F6F604F6F604F6F5F4F6F5F40605F4F605F\
406050406050405F50405F503F504F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F402F304F2F304F2F3F502F3F502F405F
2F3F5F30405F2F405F304F60304060304F601F20300000000000003F00001000003F00002F1010\
30101F304F603F506F3F506F3F5F703F506F405F703F5F70405F703F5F70405F70
40000040000040202F406070405F704F607F405F704F607F4060704F607F4060704F607F4F3F4F\
4F00004F101F406070000000000000000000000000000000000000000000100000
4F00004000000000000000002F00005000002F00000000000000000000000000000000002F303F\
50707F4F6F7F50707F50707F5F708050707F50707F505F6050707F4F6F704F6F70
4F00004F00004000004F404F4F6F705070704F6F705070705070705F7070507070507070507070\
0000000000000000000000000000000000003000002F0000200000000000000000
100000000F0F3F00000000000000002F30304F605F4F6F6040605F4F605F40605040605F405F50\
4060503F5F4F3F5F4F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F30402F304F2F304F303F502F3F50
30405F2F405F30406030405F304F603040601F2F30000000300000300000000000000000000000\
0000003F00002F101F3F30403F506F405F703F506F405F703F5F70405F70405F70
40101F40101F40000040000040607F405F704F607F4060704F607F4060704F6F7F4F607F4F6F7F\
4F3F4F4F00004F00004F6070000000000000000000000000000000000000100000
4F00003F00000000000000003000004F00003000000000000000000000000000000000002F3F40\
4F6F704F6F7F4F6F704F6F7F4F6F704F6F7F4F6F704F6F704F6F7050707F4F4040
4F00004000004F6F704F6F705070704F6F7050707F5070705F70705070705F7070000000000000\
000000000000000000000000000000400000400000000000000000000000000000
00000000000000000000000000000020302F4F6F604F605F4F605F40605F40605F406050406050\
3F5F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3F2F304F2F304F2F3F50
2F3F502F3F5F2F3F5F30405F2F405F30406010202F1F20300000003000003000002F0000000000\
0000000000000000000000003F101F3F202F3F30403F506F3F5F703F506F40101F
3F1010404F5F3F4F5F4000003F0000400000405F7040607F405F7040607F4060704F607F406070\
4F607F404F5F4F0000406070405F70000000000000000000000000000000000000
0F00004F00000000000000000000004F000040000000000000000000000000000000000020303F\
4060704060704F6F704060704060704060704F6F703F5F6F406F70403F404F0000
4010104F6F704F6F704F6F704F6F70507070506F70507070506F70507070000000000000000000\
0000000000003000002F0000000000000000000000000000000000000000000000
00000000000000000010000020302F20303040605F40605F405F5040605F3F5F503F5F50000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3F2F30402F304F
303F502F3F5030405F2F3F5F30405F2F405F304F601F202F0000003000002F00003000003F0000\
2F00002F00000000000000000000000000001F2F3F3F506F3F202F3F00003F1010
404F5F3F4F5F405F703F5F7040607040000040000040101F4F607F4060704F607F4060704F607F\
4060704F607F404F5F4F0000406070405F6F000000000000000000000000000000
0000001000004000003000000000000000002F0000400000000000000000000000000000000000\
0000004060704060704F6F70406070406F7040606F4F6F704060704F6F7040606F
4F6F704060704F6F704F6F70507070507070507070507070000000000000000000000000000000\
0000003000002F00000000000000000000000000000000000000000000003F0000
3000004F00004F000000000020303020302F40605F40605040605F3F5F50000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F402F304F2F304F2F3F502F3F5030405F2F405F1F20300000003000002F0000000000000000\
0F00003000003000003000000000000000000000000000000000001F2F303F5F6F
3F10103F20303F506F405F703F5F70405F703F5F704000004000004060703F5F70406070405F70\
4060703F5F70406070404F5F4000003F00003F5F6F000000000000000000000000
000000000000000000400000400000000000000000000000400000400000000000000000000000\
00000000000000000000000020303F20303040606F40606F20303F00000040606F
40606F406F7040606F4F6F704F6F705070700000000000000000000000000000000000002F0000\
0000000000000000000000000000000000001000004000004F00004000004F0000
1F0000000000000000000000202F2F40605F3F5F503F5F501F2F20000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030304F2F304F303F502F3F5030405F1F202F1F2030000000000000000000000000\
0000000000000000000000002F00003F0000300000100000000000000000000000
0000001F2F3F405F703F5F6F403F403F5F70406070405F70402F303F0000402F30405F70406070\
3F5F6F406070405F704060703F5F6F40505F4000004000003F5060000000000000
0000000000000000000000000000000F00004000003000000000000000000000002F0000400000\
4000002F00000000000000000000000000000000002F0000400000000000000000
000000000000000000000000000000000000000000000000000000000000200000200000000000\
0000000000000000000000004000004F00004000004F00001F0000200000000000
00000000000000000000000020302F3F5F5020302F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F30402F3F4F2F3F502F405F101F2F000000000000000000000000\
0000000000000000000000000000000000002F00003000003F00003000001F0000
0000000000000000000000003F506F3F5F703F5F6F40101F3F5F6F3F5F703F202F4000003F202F\
3F5F6F3F5F6F405F703F5F6F3F5F6F3F5F6F405F6F3F40503F00003F0000405F6F
3050600000000000000000000000000000000000000000000000001000003F00004000003F0000\
300000300000000000000000300000300000400000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F0000000000000000000000\
0000003F00004F00004000002F00002F00001F0000000000000000000000000000
00000040000000000020302F3F5F50000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F3F4F2F3F4F30405F10202F000000000000300000\
2F00002000000F00000000000000000000000000000000000000001F0000300000
3F00002F000000000000000000000000000020303F3F5F6F405F703F101F3F5F6F3F5F6F402F30\
3F00003F00003F5F6F405F6F3F5F6F3F5F6F3F5F6F405F6F3F5F6F3F5F6F3F5F6F
404F5F3F00004000003F5F6F3F5F60000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F00001F0000
0000000000000000000000000000000000000000003F00000F0F0F0000000000004000004F0000\
4000001000000F00002F00000000000000000000000000000000000000000F0000
40000020302F20302F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F304F2F3F502F3F501F202F000000\
0000000000001000002F00003000000F00001000001F00000F0000000000000000
0000000000000F00003F00003F00000000000000000000000000003F5F6F3050603F5F6F3F1010\
3F506F3050603F5F6F3F00003F00003050603F5F6F3F50603F5F6F3F50603F5F6F
3F50603F5F6F3F50603F5F6F3F5F603F5F6F3F40503F00003F00003F00003F00003F00003F0000\
3F00003F00003F00003F4F503F5F6F3F5F603F00003F0000300000000000000000
0000000000000000000000000000000000000000000000002F00004F0000400000000000000000\
2F00004000000000000000000000000000002030301F302F3F00003F00003F0000
1F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000303F502F3F501F202F\
0000000000000000000000000000000000003000003000002F0000000000000000
0000002F00001000000000000000000000003F00003F000000000000000000000000000020303F\
3050603F5F6F3F50603F101F3F50603F5F6F3F50603F202F3F00003F00003F202F
3F5F6F3F50603F5F6F3F5F603F5F6F3F5F603F5F6F3F5F603F5F6F3F5F603F5F6F3F5F603F5F6F\
3F5F603F5F6F3F5F603F5F6F3F00000000000000000000000000000000003F0000
4000003F00004000004000004000004000004F0000400000100000000000000000400000400000\
000000000000203030402F2F3F00004000003F00003F00003F00003F5F501F2F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F304F\
1F202F0000003000000000000000000000000000000000000F0000100000300000
3000000F00000000000000000000000000002F00000000000000003000003000003F0000000000\
0000000000000000001F2F303F50603050603F50603050603F10103050603F5060
3050603F5F603050603F5F603020203F202F30202F3F202F3020203F202F3050603F5F60305060\
3F5F6030505F3F5F603050602030300000000000000F00003F00000F0000000000
0000000000000000000000000000000000000000000000000000001000000000000000003F4F4F\
3F00003F00003F00003F20203F3F303050503F5F503000003F50501F2F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F3F4F3000002F0000300000000000000000000000000000000000
0000000000003000003000003F00000000000000000000000000000000002F00000F0000000000\
0000000000003000003F0000000000000000000000000000000000000000000000
20303F1F2F302030303050603F5F603050603F5F6030506020303F1F2F30000000000000000000\
0000003F00003F00003F5F603F202F3F00003F10103F5F60203030203030000000
0000000000001000000F00001000000F00001000000000002030303F00004000003F000040605F\
3F5F5F3F5F5F3F5F503F5F5F3050503F0000300000202F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000001F202F2F3F4F2F3F502F3F502F40502F3F50000000
0000002000001F00000000000000000000000000001000003000002F0000000000000000000000\
0000000000000000000000002F00000F0000000000000000000000000000000000
0000000000003000003000003000003F00000000000000000000000000000000000F0000000000\
0000000000002F000030505F3F5F6030505F3F5F6030505F3F2F2F3000003F0000
3F00003F10103F10103F10103F5F5F4060603F00003F3F3F3F5F5F3F5F5F3050503F5F5F305050\
3F3030304F4F30505030504F30504F1F2020000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000001F1F2F303F502F00003000002F1F20
3040502F4050304F5F2F4050304F5F000000000000000000200000000000000000000000000000\
1000003000003000003000002F0000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F00002F00003000001F0000000000000F0F\
0000000000003010103F4F5030505F3F5F603F5F5F3F5F603F5F5F3F5F603F5F5F
3F5F603F5F5F3F5F603F5F604060603F5F5F3F5F603F5F5F3F5F5F3010103F5F5F305050305050\
30504F305050304F4F1F2F2F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F304F2F0000
2F3F4F2F40502F3F503040502F405030405F2F405F304F5F2F4F5F304F5F2F4F5F304F5F000000\
0000000000000000000000000000001F00002000001F00002000001F0000200000
0000000000001F000020000000000000000000000000000000000030505F30505F30505F2F4F50\
30303F30000030000030505F30505F30505F3F101030404F3F5F5F30505F3F5F5F
30505F3F5F5F3F5F5F3F5F5F304F503F3030301010305050305050305050304F4F30504F2F4F4F\
304F4F102020000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F3F4F2F3F4F2F3F4F3040502F40503040502F4050304F5F2F2F303000002F0000300000\
2F4F5030505F2F4F5F305060304F5F30505F2F4F5F305060304F5F30505F304F5F
30505F304F5F30505F2F4F50304F5F30000030303F30505F30505F30505F30000030000030505F\
30505F30505F30505F3F3F3F3000003F00003000003F5F5F30505F3F5F5F30505F
3F5F6030505F3F5F5F3040403F10103050503050502F10103000002F00001F2F2F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F3F4F2F00002F00002F00002F00002F00002F2F302F2F30\
3010102F00002F4F5F2F4F5F304F5F2F4F50304F5F2F4F5F304F5F2F4F50304F5F
2F00003000002F00003000002F3F4030404F2F4F5030505F2F4F5030505F2F2F30300000000000\
30505F304F5030505F30505030505F30505030505F30505F30303F303030305050
304F503050502F4F4F304F4F2F404F2F40402F4040000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000101F20301F202F3F4F30404F\
2F3F4F3040502F4050304F5F2F4F50304F5F2F4F5F304F5F2F4F5030505F2F4F5F
304F5F2F4F5030303F2F2F303000002F202030505F000000000000000000000000300000300000\
1F0000000000000000000F0F304F5030505F304F50304F502F4F4F304F502F404F
304F4F2F4040304040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F303F2F00002F00002F00002F00002F101010202F000000
2000000000000000000000000000002F0000200000000000000000000000000000000000000000\
0000000000000000000000000000002F00001F2020000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[6]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061h<0015a0740L@1[06/0J`1[06/0L@1n00P700061gT000A[LF]a3F/03W5[LF]aJg5[LF]aJg5[
NP081`001PMg0003L@1[00Y[01L0Jf]aJg5[LF]aJg5[LF]aJg5[LF]aJ`1g00P700061g8000]aJg5[
LF]aJg5[L@07J`0KLF]aJg5[LF]aJg5`LG1aL75`LG1aJg5[LF]a07<020L000H7K`00<G40L@1a0740
LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg400V]`00P700061f`001=aJg5[
LF]aJg5[L@1a0740L@1a01T000]aL75[LF]aJg5[L@1_00P700061fT000Ua0740L@1a07400V/Y0005
LF]aJg400V]^00P700061fH000]aJg5[LF]aJg5[L@0_0005LF]a0740K@081`001PMQ0009L@1a0740
L@1a03T000=[06/00V/90003C@0005h020L000H7GP002G5[LFaaK75[L@0m0009L@2:L8Y`RP1a06D0
20L000H7G@001g40L@1a0740@@02J`09RV]aJg5[L@1[068020L000H7F`001VaaJg5/LD<000ea075`
RW2:L8YaRW6:009a0`000dd0001I00P700061e/000=a0000A`004700RV^:JhY[LF]aJg5[LF]M00P7
00061eT000=/L@00B@0046]aLHY`RW2:LHYaRW6:LHX2L003070005P020L000H7Y@004g00L6^:JhY[
RV]aJg5[LF]aJdd0FP081`001PNV000ELF^:LHYaRW6:LHYaRW6:LHYaL4e`05L020L000H7Y`005g00
RV^:JhY`RW2:JhY[RV]a04d0C@1=05@020L000H7E@000f/0001>000GLF]aLHYaRW6:LHYaRW6:LHYa
L4e`CG00E@081`001PMC0003L@00054001=`08Y`RW2:L8Y`RW2:L8Y`RW2:009=00@0L6]aDP081`00
1PM@0005LF]aK740D@000g5[L@04RP0<LHYaRW6:LHYaRW6:0W400de`C@02L003L@0004l020L000H7
C`001G40L@1a05<001Ea08Y`RW2:L8Y`RW2:L8Y`RW2:L7400Td01P1aJg40CDh020L000H7CP001W5[
LFaaK5<000=[LHX01HX0576:LHYaRW6:LHYaRW6:CG1=LG1a0W1=00P700061dd000Ma0740L@1a05<0
01ma075aRW6:L8Y`RW2:L8Y`RW2:L8Y`RW1=04e[L@1=04`020L000H7C0001g5[LFaaK740E0000f]a
J`0;RP03LHYa00>:00YaRW6:CG1=RW2:0W1;00P700061d/000Ma0740L@1a05D0025a075aRW6:L8Y`
RW2:L8Y`RW2:CHY`RW2:L4d0RV]aJg40BP081`001PM8000;L@1aJg5[LFaaK740E0000f]aJ`04RP05
ShZ:RXl028X044e`LHYaRW5`CG5`RW2:CG5700P700061dL000]a0740L@1a075[L@0b000FLHY[RW2:
JhY[RV^:JhY[RV^:Q8Y[RPd001AaJg5aRW6:LHYaRW2:L8Y`RW2:L0A=00f:L8Y[C@1=Jg5[LF]a04H0
20L000H7AP003740LF]aJg5/LFaaK1l000M[RV^:JhYa00b:00>@RXX00XXAQ005RXV:RHX020000f]a
J`02RP0;ShZ?RXn:ShZ?RXl018X2L006CG1=RW6:0W0024e`L8Y`RW1aA@081`001PM5000<L@1a0740
L@1a075[7@008F^:JhA[Q6^4JhYaRW6:LHYaRW6:LHYaRW6:LHA[Q6^4J`0;Q009JhA[Q6^:JhY[00D0
01I[LHZ:LHYaRW6:LHY`RW2:LHYaRW1a0dd2L00:RW1=04d0LF]aJdD020L000H7A000175/LF`7L@03
K74001D000M[RV^:JhA[00J418X04H^:T8Z@RY2:T8Z@RY2:T8Z@00>:2X@00hZ4RP0=Q00QRXUaJg6:
ShZ?RXn:ShZ?RXn:ShZ?RXn:ShYfCG1=L76:009a00E=L72:L01400P700061d<000ea0740L@1a075[
LF]a01@002Q[RV^:JhA[Q6^4JhA[Q6^4LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:0X@01f^4JhA[Q6/0
0h@00hZ4Q009Q006JhA[Q01a1000576:LHYaRW6:LHY`RW2:L8Y`RW2:0dd2L003RW1a009=00<0LF/0
@`081`001PM20004L@1aK0Ua00=/00003`003Fb:K8A[Q6^4JhA[Q6/018@2RP0GRhZ;RX^:T8Z@RY2:
T8Z@RY2:T8Z@RY000hX7Q0>:00F@RY2:T00;Q003LF]a00L0016?RXn:ShZ?RXn:ShZ?RXn:L@02L006
CG1=RW6:0W000deaL01200P700061d4000ma0740L@1a0740LF]aJg403@00;75[LF^4JhA[Q6^4JhA[
Q6^4JhA[RW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa0hX00g6:J`0:Q0::00J?RXn:ShX7Q003074000`0
00f:L8YaRW2:L8Y`RW2:009`0dd2L006RW1=071[@@081`001PM00004L@1aK0]a00=/0000200058Y/
RVb4JhA[Q6^4JhA[RV^:JhA[0h@078Z;RX^:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY02RP03Q8Z400R4
00N@RY2:T8Z@00::00>4RX@00X@00g5[L@0?0007ShYeL8n:S`05RP0<LDe`CG1=RTe`CG5`@0081`00
1PLo000AL@1a0740L@1a075[LF]aK7401`000g5[L@05J`0TQ6^4JhA[Q6^4LHY[RV^4JhE[RW6:LHYa
RW6;LH]aRg6;LI1a2hX00hB:Q00;Q006RXn:ShZ?0XX4Q0030740014000a`CG1aL8Y`RW2:L8X5C@05
RV]`0740@0081`001PLn0006L@1aK75/3743000ILHYaRV^:JhA[Q6^4JhA[RV^:LH]aRf^4J`02Q0::
026@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY002h@2RP05T8Z@RY000XX3Q003LF]a0180
00Q=MDefL8n:S`B:00Y`LDe`CHY`RTea?`081`001PLo000GL@1a0740L@1aJg5[LFaaK740LFaaK740
0V`5J`0VQ6^4JhAaRW6:JhY[Q6^4JhA[RW6:LHYaRW6:LH]aT76;LI1aT746RP08ShZ?RXn:T8X<Q007
RXB:RXn:S`02RP05Q8Y[LF/04P000e9=DP02C@08L8Y`RW2:L8X2L0==00E`JhX0L@0n00P700061ch0
00Ia675/LF`2J`Ya01Z:LHYaRW6:JhY[Q6^4JhYaRg6;LH]aQ6^4J`B402>;RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T00>Q008RY2:T8Z@RY02RP03JhX0014000Y=MDeeCGE`MXZ?1XX00de`
C@02L003RTea03d020L000H7?@006W40L@1a06/0K01aJg5/LFaaJg5/LFaaK75/16/0=XY[RW6:LHYa
RW6:JhA[Q6^4JhAaRW6;LH]aRg6;LI1aT76@LI1aT76@RY2:T8Z@RY2:T8Z@RPn400Z:ShZ?RXn:S`1a
5003C@0=DTe=CG5`RW2:L8Y`RP05C@03L01=03`020L000H7?0001G4HLFaa00A[2742000ILHYaRW6:
LH][RV^4LHYaRg6;LH]aRV^4J`06Q00ST8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY00
0hX<Q00;RXB@RY2:T8YaJg405@002G1=MDeeCGI`S`04RP05LHY`LDd00g000dd0000j00P700061c/0
00ma06/0J`1[06`0LF]a07401@07L@=[03MaK8YaRW6;LHY[Q6^4JhA[Q6^4JhYaRg6@LI1aT76@LI2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00::00>9RX@03X@028n:T8Z?Jg5[6002C@03DTeB009=00Q`LG2:
L8Y`RP9`00A=L@1=>P081`001PLj0003LAQ[00E[1G400f`00005000ELHYaRW6:LHY/RW6;LH]aRg6;
JhY[00R402F:Q92:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00>:3h@028Z@RY2:LG1a
6P0024e`CGE=MDef0XX0176:LHX2L@03CG5=03X020L000H7>@003G40J`1[06/0L@1a07402@08L@0i
RW6;LH]aRg6:LHA[Q6^4JhA[Q6^:JhYaT76@LI1aT76@LI2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00B:
3h@028Z9T8Z@Jg5[6`03C@0<DTeBCG5`RW2:L8Y`0Td00g40000h00P700061cX01F/6L@X001EaRW6:
LHYaRg6;LH]aRg6:K8Y[Q6/01h@3RP0ST8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY00
0hX4Q005WHB4Q9d01h@2RP05T8YaL7407@003W1=MDeeCG5`RW6:LHYa0g0i00P700061cT000][06/0
L@1a075/L@0;00Ia03^:LHYaRg6;LH][Q6^4JhA[Q6^4JhA[Q6^:LI1aT76@LI1aT76@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T004RPn400R:Q92:T6]aJa@01Dd600I=0g000hY`RP02C@03Jg4003L020L000H7
>0001V]/LFaaK0Ea30005G6;LH]aRg6;LH]aRg6;LHA[Q6^4J`08Q0::02J@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z?RPB400NMQ9f4WHBM00>400VMQ8B?T8maL7404@002Dd0L4eeCGE=
M@07000?L4eeCGE=L76:LG1=RW1a03L020L000H7=`002F/0L@1a0740L@0?00Aa03f:LHYaRW6;LH][
RV^4JhA[Q6^4JhA[Q6^4LHYaT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00B:4H@01Y2:
T6]aJa<00dd01e9=DTeB04d01@08C@07L4e=L75[L@0f00P700061cP000AaK75/0g400f`0000=000E
LH]aRg6;LH]aRg6;K8Y/Q6^4JhA[00R40XX09Y2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RXn:18@02if4WHBMQ9f4WHBM00:400J?T8maL74A000<C@1eCGE=MDeeCGE=1`003DeeCGE=L4e`
CG5`LF/0=P081`001PLg0007L@1a0740L@0A009a03n:LHYaRW6:LH]aRV^:JhA[Q6^4JhA[Q6^4JhAa
RW6@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY0018XAQ006T8Z@Jg5[4`002TeBCE9=DTeB
CE8800U=00>:Jg40=P081`001PLf0007L@1aK75/L@0B000ELH]aRg6;LH]aRg6;JhA[Q6^4JhA[00R4
0XX0892:T8Z@RY2:T8Z@Ri2:T8^@RY2;T8Z@RY2:T8Z@RY2:0i03RPB4012MQ9f4WHBMQ9f4WHB:Si2?
0g4A000<C@1eDWE=MDeeCGE=1P002Td0001`CGE=L4d2L003RW1a03D020L000H7=@000g40J`0G0011
Rg6;LH]aRg6;LH]aQ6^4JhA[Q6^4JhA[Q6^4JhA[RW6@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY000XX00hB:Q00>Q008RXB@RY1[LF/C0008CE9=DTeB04d70003DTeB00D00dd01W2:L75[LC@0
20L000H7=0001G5[LF]a01H000eaRg6;LH]aRg6;LH][01240XX0592;T8^@Ri2;T8^@Ri2;T8^@Ri2;
39006ZV@ZHBSQ9f4WHBMQ9f4WHBMQ9f4WHZSSjV?0g4C0006ME9eCGEB2@001DeeCGE=00D000I=L76:
LHX2LC<020L000H7<`001G40J`1a01L004:;LH]aRg6;LH]aRg64JhA[Q6^4JhA[Q6^4JhA[Q6^:LI2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8XBQ008T8Z@RY1[LF/D009=00=BCE802@001e9=
DTeBCE8010001de`L75`LF/0<`081`001PLb0006LFa[Jg5/5`003W6;LH]aRg6;LH]aRf^:3h@05XZ4
T8^@Ri2;T8^@Ri2;T8^@Ri2;T8/=T00IRY24WHBMQ9f4WHBMQ9f4WHBSRJ>?ZHnYS`03LA<00Td01GEB
ME9=00P000U=MDeeCGE=MDd00P00271=RW6:LHY`<P081`001PLc0003L@1a01T0022;LH]aRg6;LH]a
Rg65JhY[Q6^4JhA[Q6^4JhA[Q6^4L@::01n@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8B:0124
00^:T8Z@RY2:T6]aJ`0E0003CE9=00==1`002E9=DTeBCE9=DP02009=00M`RW1aL75[034020L000H7
<P00175/LF`I000<LH]aRg6;LH]aT76;0hX@Q00BRhB:Ri2;T8^@Ri2;T8^@Ri2;3902RP0JWHBMQ9f4
WHBMQ9f4WHBMRZV:ZI2YSjV?ZHl3LAD000IeDWEBMDd7000:CGE=MDeeCGE=M@800Td0176:LHX2LC40
20L000H7<@001G40L@1a01T001j;LH]aRg6;LH]aRg6:Jh][Q6^4JhA[Q6^4JhA[Q6/4Q00KRXF:RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00::38@04HZ?RY2:ShZ@RY2:T8Z@Jg5[01H00Td00e9=DP070003
DTeB00==00D0CE9=DP040006L75`LG1a<0081`001PL`0005LFaaK7406P003g6;LH]aRg6@LI2:T8Z@
R`03RQ2400f:Q8B:QHZ;T8^@Ri2;00Z@0XX07Z>:Q8BMQ9f4WHBSQ:>4ZI2YT:V@ZI2YT:V@ZHnYS`=a
5`001WEBME9eDPD000I=MDeeCGD30003MDd000@000=`RW400W4`00P700061c4000=a07406`007X]a
Rg6;LH]aRg6;LH]aRfb;LHA[Q6^4JhA[Q6^4J`R400F:Q92:T002RP08QHZ5RXF:QHX=Q00JT8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY1[LF/G0008CE9=DTeB04d30003DTeB00==0P0014d0DTd50004L75`LC00
20L000H7;`001FaaK75/01/001EaRg6;LI1aT76@RY2;T8^@Ri2;T8/00hX00hB:Q006Q003RXB400N4
00>:Q8X00hX00j>:RP02RP>4022MQ9f4WHBSQ:>:ZHZYT:V@ZI2YT:V@ZI2YT:V@ZI2YT0=a5`002DeB
ME9eDWEBC@020006CGE=MDee10000dee00050003LHYa030020L000H7;`000g40L@0M000ORg6;LH]a
Rg6;LH]aRg6;LH]aRg6;LH][T76:JhY[R`04Q003LI2500B:38@6RP0PT8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY1[LF/H0008DTeBCE9=DTd20003DTeB00==1@02C@D000AaL75[;`081`001PL^0004
L@1a61d002UaRg6;LH]aRg6;LI1aT76@RY2;T8^@Ri2;T8^@Ri2;T8^@Ri2;T8^@R`0<T00TZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@0g4K000=CE9eDWE=ME9eDW1=L0050003MDe=00H0
00=`L@00;P081`001PL_0003L@0001d005^;LH]aRg6;LH]aRg6@LI1aT76@LI1aT76@LI2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@Si2:T8n@RY2:T8Z@RY2:T8Z@Si2?LF]a01d0
0Td01gE=DTeBCE800Td60004CE9=C@H000=`0000;@081`001PL^0003LAP001h002UaRg6;LH]bRg:@
LI1aT76@Ri2;T8^@Ri2;T8^@Ri2;T8^@Ri2;T8^@R`0<T00SZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:T00g4O000:CGEBME9eDWEBC@D000A`CGE=1P000g5`000]00P700061dl0042;LH]a
Rg6;LH]aT76@LI1aT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@Ri2;T8^@Ri2;T8Z@RY2:T8Z@RY2:
0i0068n@T92?T8Z@RY2:T8Z@RY2:T8n@Sg5[LB4000aeCE9=DTeBCE80C@02C@03DTeB00H000=`L@00
;@081`001PL^0003L@0001h002MaRg6;LX]bT7:@LY1bT8^@Ri2;T8^@Ri2;T8^@Ri2;T8^@Ri2;T8/0
3Y000jUaL@0CL@0=ZI2YT:V@ZI2YT:V@Z@03LA@000==04d00dd20003C@0000@000mBME9eDWEBMDe`
CGE=MDd01`000g00000/00P700061bd000=a00007`00Fh]aRg6;LH]aRg6@LI1aT76@LI1aT76@LI1a
T8Z@RY2:T8Ya0740L@1a075[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[T8Z@Si2?
T8maJg404`000e9=0002DP04CGE=M@P000i=DTeBCE9=DTeBCE9=DSH020L000H7;@000aP0000O000O
LH]bRg:@LY1bT7:@LY2;T8^@Ri2;T8^@Ri2;T8^@R`05LBT027400jV@Z@03LA00015=04eBME9eDWEB
ME9eDTd0L004000<DWEBME9eDWE=MDee0Tde00P700061bd000=a00007`007h]aRg6;LH]aRg6@LI1a
T76@LI1aT76@LI2:T01a0740<`002G5[LF]aL75[L@0C00EB00]=MDeeCGE=MDe`C@04000;DTe=CE9=
DTeBCE80=P081`001PL/0003L@00020001MbRg:@LY1bT8^@LY1bT8^@Ri2;T8^@R`06LCT0274=000D
C@1=DWEBME9eDWEBME9eDWEBMDd5000:CGEBME9eCGD0CCD020L000H7C`006H]aRg6;LH]aRg6@LI1a
T76@LI00L@1a0740?P002g1aL91`LG1aJg5`00/01U803TeeCGE=MDeeCE9=DTeB1@001Td0DTeBCCL0
20L000H7C`002g:@LY1bT7:@LY2;00MaA@003HYaLI2YT:V@ZI2YT:T01G402XYa001eDWEBME82C@0=
M@1=CGE=ME9eDWEBM@02C@D00Tdh00P700061dl000N;LH]aRg6@009a00@0L@1aBP001V]aJi2:T0Af
01B?T8n@Si2?T6]aJg5BMDeBCGD0C@D000UBCDe=DTeBCE800Td00`1=000k00P700061e0000V@LW5/
LFaaK740C@0078YaRXn@SgEfSgJ?MXmfSi2YT:V@ZG5fL7EBME89009=00YBME9eDWEBM@1=>`081`00
1PMA0003L@1a058000I[LG2@MY06DPIf00YeT8n@RWEBMDeB2`002Td0DTeBCE9=DTdk00P700061dl0
00=/LF`0E00078YaRXn@DWEBME9eDWEBME9eMXmfSgJYSgIBME8<000:C@1`DWEBME9eDSX020L000H7
C`000g40001D0006L75[T7Ie3E82M@09SgF?DWEBMDee00`00dd01E9=DTeB00h000=`0000:@081`00
1PM<0004LFaaLEH001V:LI2?T59eDWEBME9eDWEBME9eDWEBSgF?009e00=BMG400W4<0006MDeeDWEB
3P000hl0000Y00P700061d/000Ea08]aR`1F0008L75fT7IeDWD;DP07ME9eDWEBS`03DP05MWEaL740
300014eBCE8>0003L00002T020L000H7B@0026aaK91bT76@E@007XYaZI2YMGIeMWEfMGIBME9eDWEB
ME9eDWEBME9eMP:?00BYLG5a2P001Dd0ME9`00d000>?0000:@081`001PM7000:L@1aLH]aRg6;LED0
011`LHnYSjV?T8n@DWEBME9e0e802GEBME9eDWEBM@04DP06Si2?LG1a2P0014eBCDd<0003L00002T0
20L000H7A@0036aaK8]bRg:@LY1aT5D001>:LJV@ZI2YT:V@ZI2YT7IeMWEf009e011BME9eDWEBME9e
DWIeZHn@0W490004ME9eDP`000>?L000:@081`001PM3000>L@1bLH]aRg6;LH]aRg5D00=a0i004:V@
ZHnYSjV?ZHnYSi2?T8l3M@EB00ieDWEBME9fMGJ?T71aL0H000ABCE9=2`000g1a000Y00P700061d40
010HLAR;LH]aT76@LY1bT7:;E003L@0TZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YMGIeMU9eDWEBME9eDXmf
0Xl00i2?L@02L@@000A=ME9e2`000hY`000Y00P700061cl001=a075aLW6;LH]aRg6;LI1aRg6;05<0
0g408i2YT:V@ZHnYSjV?ZHnYSjV?T8n@Si2?ME9eDWEBME9eDWEB009e00efSi2?T71aL01=DTeB00`0
00=`L@00:@081`001PLm000E674HLW5bLH]aRg6@LY1bT7:@LY1b05<00g406jV@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@MP02M@08DWEBME9eDWD2MP03MI2?00:?00MaL7E=ME9=00/000>?LG40:@081`00
1PLk000GL@1aLG9aLW5bLG9aRg6;LH]aRg6;LH/0D`03L@0@T:V@ZI2YT:V@ZI2YSjV?Z@:@01:?T8n@
Si2?T8meDWEBME9eDWD3DP0;ME9eMXmBLDeBCE8030000g1aL00Y00P700061cT001THLAQbLG9aLW5b
LH]aRg:@LY1bT7:@LY1b05<00g40::V@ZI2YT:j@ZI2^T:V@[Y2YT:V@ZI2YT:V@ZI2YMGIeMU9eDWEB
ME83M@08DWIeME9e04d;0003Sg1a02T020L000H7>@02L@0GLW5bLG9aLW5bLH]aRg6;LH]aT76;LH/0
D`03L@0@T:V@ZI2YT:V@ZI2YT:V?Z@B@012?T8n@Si2?T8n@SgEBME9e2U800deBC@0=0003L75`02T0
20L000H7<0000f/000060009LG9aLW5bLG9a009b00j;LI1bT7:@LY1bT7:@LU800g40=92YT:V@ZI2^
T:j@[Y2^T:j@ZI2YT:V@ZI2YT:V@ZHnYSjUeMWEfDWEBME9eDWEBME9f04d;0003Sg5a02T020L000H7
<@000f/00005000JLW5bLG9aLW5bLG9aLW6;LH]aT76@LI1aRg5A00=a016YT:V@ZI2YT:V@ZI2YT:V@
Z@04T00BSi2?T8n@Si2?T8n@Si2?ME9e2E800gI=L@0<0003L75`02T020L000H7;`04J`D00W401W9a
LW5bL@Ab00n;LI1bT7:@LY1bT7:@LY00D@03L@0XT:V@ZI2YT:j@[Y2^T:j@[Y2YT:V@ZI2YT:V@Si2Y
T:V?ZHnYSjUeMP9e00EBME9eDP02M@03T71a00/000B?LG5`:0081`001PL]000:J`1[06/0J`1[009a
01YbLG9aLW5bLG9aLW5bLG9aLW6;LH]aT76@LE000g401Y2YT:V@Z@b@00B?T7J?0WH0159eDXl5DP0C
SgEfDY2?T8n@DWE=DTeBCE9aS`02L@X000A`LG1a:0081`001PLX0003L@0000802F/07W9aLW5bLG9a
LW6;LH]bRg:;LY1bT7:@LY1bT7:@LTl002EaRW6^T:j@[Y2^T:j@[Y2^T:j@[Y2YT7YfU59jDWYBME9e
DWEB009e00UfMGF?ZHn@SgH00WD00e9eDP02S`04T8maL0L000EaSg5aL00X00P700061bd002M[06/0
J`1[06]/LFabK79aLW5bLG9aLW5bLG9aLW6;LH]aT76@LI00C`03L@06T:V@ZI2Y39002WF@DWEBME9e
DWD9DP0AME9BSi2?T8n@CE9=DW6?LHl00W470005Sg1aL740:0081`001PLX0004L@0000Y[0W40779/
LW5bLG9aRg6;LX]bT7:@LY1bT7:@LY1bT79?000bLHZ@[Y2^T:j@[Y2^T:j@[Y2^T:j@U59jDWYBNU9j
DWEBME9eDWEBME9eDWEBZHn@Si02M@04DXn?T0>?00=a000010001G6?LG5`02P020L000H7;@0026/0
J`1[06/016/01W5[LV]aJ`Aa015bLG9aLW5bLH]aRg6@LI1aT01?009a00NYT:V@ZI2Y00b@00YBME9e
DWEBME9e458038n@Si1=DW6?LHmaS`9a1@001Hm`LG1a02P020L000H7:0001W40001[00Y[00=aJg40
0f/05g5[Rg6;LH]bT7:@LY1bT7:@LY1bT7:@04h003QaRY2^T:j@[Y2^T:j@[Y2^T:j@[Y1kDWYBNU9j
DWYBME9eDWEBME9eDWEBME9eDWF?T8n@MI2?T0F?00=a00000P001G6?LG5`02P020L000H7<@0016/0
J`0;J`0EK6][LW5bLG9aLW6;LH]aT76@LY1a04h00W401jV@ZI2YT:T02Y0038n@DWEBME9eDWEBM@9B
00=fDU8035803GF:T8Z@RXmaSg6?LHl00W420006LHm`LG1a:0081`001PLX0003L@0000D000=/Jg40
3F/05HA[Q6^:LI1bRg:@LY1bT7:@LY1bT01>000OLJn@[Y2^T:j@[Y2^T:j@[Y2^T9EBNU9jDWYBNXnD
DP02T0:?00bYDXmfSe9eDWEBME82M@09MXn@Si2?T8n@00B?00I`002?LHl2L@03L00002H020L000H7
;`001aP0LAQ[67400f/00g5[J`08J`0BK6][LW5bLG9aRg:@LI1bT76@C@001g6@ZI2YT:T039005gIB
DWEBME9eDXmfT7J@Si2?T8n@MXmf00QB00efRY2:Sg6?LHmaSg6?008000=`Sg000Td00g40000V00P7
00061bP000=a00001@02L@04LW5bK09[00E/LG5[L@03J`0FQ6^4JhA[Q6b:LY1bT7:@LY1bT7:@LTd0
02Ia[i2_T:j@[Y2^T:j@[Y2^T9F@UE9jDWYBU5:^T:j@ZI2YT:V@Z@B@00R?T59eDWEBM@9f00J?T8n@
Si05S`08LG2?LHm=MG0X00P700061bT000=a00000`001g4HLAQa67400f/00g9/L@0;J`04LF]aK09a
0W801i1aT7:@LI00C@001G6@ZI2Y00>@00>YT9001Y000gJ@MP02DP04ME9eM@:?00f@Si2?T8n@Si2?
T8n@00:?00=fSgH01E80372?RXmaSg6?LHmaS`9a00=`Sg000Td00g40000V00P700061bP000=a6000
0`02L@05LW5bLG800W42J`=a0W800g5bJ`03J`0CQ6^4JhA[RV^:LHYbT7:@LY1bT01<000XLJj@[Y2^
T:j@[Y2^T:j@[Y2ESgYBNU9jMXn@ZI2YT:j@ZI2YT:V@Z@J@0WH0359eDWEBT8n@Si2?T0:?00IaS`2?
LGD3CBP020L000H7:@006G40001a674HLAQb6780L@1aK79aLW5bK7401F/048A[Q6^4JhY[Rg:@LY1b
T79<0003LI2@00N@00>YT9001Y000e9eDP02DP9f1I004Xn@Si2?T8n@Si2?T8n@Si1eMPAB00YaT76?
LHmaSg6?0`000hm=C@02CBP020L000H7:0000g4H00020008CW5>LW5bLG82L@=[0W44LP9a01EbJhA[
Q6^4JhA[Q6^4Ji1bT7:@LY00B`02L@0U[Y2^T:j@[Y2^T:j@[Y2^T:j@NU9jDWYfU92YT:V@[Y2^T:V@
Z@08T00>Si1eME9eDY2?T8n@Si02S`03MP00009a00AeCDe=:0081`001PLY000>L@00074HLAQa678H
LQP3J`08K75aLW5bLG82L@E[00n4JhA[Q6^4LY1bT7:@LY00BP02LA6@00=eDWD00e800gIBT004T00B
Si2?T8n@Si2?T8n@Si2?T8n@15802Tf@LI1aSg6?LHl30003MDe=009=:0081`001PLY0004600000=a
00MbLG9aLW5b00=[00=aK7801W805G6@JhY[Q6^4JhY[RW:@LY1bT7:@LP1:009a01R_T:j@[Y2^T:j@
[Y2^T:j@[WEkDWYBNU82T00=ZI2YT:j@ZI2YT:V@Z@06T00>Si2?MU9eDWEfT8n@Si02S`09L@00071e
CG5`02P020L000H7:@004740001a674HLAQb678HJ`02J`0MLFabLG9aLW5bLF]/Q6^:K8Y[T7:@LY1b
T7:@LY00BP02LA6@00>?MWD01586T00DSi2?T8n@Si2?T8n@Si2?T8n@MGH2DP0:CHmaT76?LHmaS`<0
00EBCDe`L@0X00P700061bT000lHL@1aCW5>LW5bLG9aLW400f/00faaLP06LP0DT6b:LXYaRW:@LY1b
T7:@LY1bT79:009a02F_T:n@[Y2^T:j@[Y2^T:j@[WJEDWYBNU:DT:V@ZI2^T:j@ZI2Y00R@00j?T8n@
DWEBME:@Si2?T0:?00If0000CGD2L@03L00002H020L000H7:@004740001a674HLAQb678HLP03J`0M
K79aLW5bLG9aLW5bK91aT76@LY1bT7:@LY1bT780B@02LA:@00=BME800e82MPB@01B?T8n@Si2?T8n@
Si2?T8n@Si2?MP9B00Y=Sg6@LHmaSg6?0`001E9=LG1a02P020L000H7:@000aQa0003L@09CW9aLW5b
LG9a00A[27805I1bT7:@LY1bT7:@LY1bT7:@LY1bT019000MMY2_T:n@[i2^T:j@[Y2^T:j@[Y1jDWYB
NU9jSjh00i001Jj@[Y2^00Z@00V?T8n@MWEBME800Xl01I2?T8n@00<000=`CGD00W400g00000V00P7
00061bT0011a000HLAQa678HLQQb674016/0779aLW5bLG9aLW5gLI1aT7:@LY1bT7:@LY1bT7990003
LI2@016@00=fSe8015800hmfT002T00DSi2?T8n@Si2?T8n@Si2?T8n@SgD2DP09CE9aT76?LHma00@0
00ABL75`:@081`001PLY0005674HLAP00W402G9aLW5bLG9aK003J`03LFab00=b01N@LWMbT7:@LY1b
T7:@LY1bT7:@LY1bT00S0005600001P03@001AP0001>00<000=>00002002L@0PT:n@[i2_T:n@[i2^
T:j@[Y2^T:ifNU9jDWYBNWJDT:h<T004Si2?T0:?00=fME800WD01WJ?T8n@S`@000EeCG5aMP0Y00P7
00061bT0019a0740LAQa674HLAQb6780J`02J`0<LW5bLG9aLW5bLG9a178037MbT7:@LY1bT7:@LRH0
00>AT9403I002I6@TI2AT96@UP09009a59001GIBDU9e00=B01QfSi2?T8n@Si2?T8n@Si2?T8n@MY1a
Sdd3DP06LHmaSg6?1@001580LG0Y00P700061bT000DHL@006002L@07LW5bLG9aLP02L@=[00=aLW80
0g805gMbMg9gLY1bT7:@LY1bT7:@LY1bT7:@01<000<H00001@00:I6@TI2AT96@TI2AT96@UY2AT9J@
UY2_T:n@[i2_T:n@[i2_T:n@[i2_00@00W408Y2_T:n@[i2_T:n@[Y2^T:j@[Y2^T7YBNU9jDWYBNY2E
T9D:T00ASi2?T8n@DWEBME:@Si2?T8l0100017E=L74Z00P700061bd000ia674HLAQa678HLP1[009[
00YbLG9aLW5bLG9a27802i1bT7:@LY1bT7:@01D02Y002I6@TI2AT96@T@0=T009TI2AT96@TI2A00R@
0g4ET003MXmB00QB01b@Si2?T8n@Si2?T8n@Si1fT7IBCE9=Sg6?LHma1@0014eB074Z00P700061bX0
00Aa00000W40279aLW5bLG9a16/2L@Ab01QgLWMbT7:@LY1bT7:@LY1bT7:@LY1bTG8A000eT96@TI2A
T96@TI2AT96@TI2AT96@TI2FT96@UY2FT9J@[i2_T:n@[i2_T:n@[i2_T:n@[i000W46000O[Y2^T:j@
UI2ET:j@[Y2^T:j?NU9jDWYBNU9kMXn@U@08T00;Si2?T8meDWEBMGH00Xl00i2?0005009=;0081`00
1PL^000K674HLAQa678HJ`1[06]>LW5bLG9aLW5bLG9a00Ab00igLY1bT7:@LY1bT7:@LP/000]bT7:@
LY1bT7N@M`08T009TI2AT96@TI2A00f@00VAT96@TI2AT9402902L@/03I000hmBDP06DP0AMWEfMY2?
T8n@Si2?T8n@MY000U8024eBMGIaSg6?1P03CB/020L000H7;P02L@07LW5bLG9aK004J`03K75b00Ab
01UgLWMbMg:@LY1bT7:@LY1bT7:@LY1bTG:@00L003fAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT9J@TI2FT9J@UY2FT:n@[i2_T:n@[i2_T:n@[i2_00=a3`000i2ET002T00@[Y2^T9EfU59jDWYB
NU9jDP9f00>?T90019002Hn@Si1BME9eDP02MP03T8n@00L000=eCDd0:`081`001PL_000HLAQa674H
J`1[06/0JdibLG9aLW5bLG9a27808Y1bT7:@LY1bT79gLWL0T7:@LY1bT7:@LY1bT7:@LY1gT7L8T009
TI2AT96@TI2A00f@00VAT96@TI2AT9401i02LA<029000gEBDP09DP=f00F@Si2?T002MP0;ME9BDTeB
CHmaSg4020000dd0000Z00P700061bh00W400g9[J`08J`03LG9a00Ab05agLWMbMg:@LY1bT7:@LY1b
T7:@LY1bT7:@096@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@UY2FT9J@
[i2_T:n@[i2_T:n@[i2_T0=a5P004i2ET:j@[Y1jDWYBNU9jDWYBME800WH3T00=MI1BME9eDWEBMXn@
S`080003L4d002/020L000H7;`005W4HJ`1[06/0J`1[CW9>LW5bLG9aLW48LP0<Mg:@LY1bT7:@LI1b
0`005G:@LY1bT7:@LY1bT7:@LY1bT7N@M`08T009TI2AT96@TI2A00f@00VAT96@TI2AT9401Y03LAL0
00>?T8l01I000hmBDP0:DP03MGIe00IB00I=MGJ?LHl50004C@00CBd020L000H7;`04L@03Jfa[009[
00A/LG9a1W805GMbMg9gLY1bT7:@LY1bT7:@LY1bT0060011T96@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TI2AT96@TI2AT9J@TI2AT9J@UY2FT:n@[i2_T:n@[i2_T:n@[i000W4K0003T9F@00:@00b^T7YB
NU9jDWYBME83M@0<DWEBME9eDWEBT8n@1@000deB000_00P700061bl001ma674HLQQb678HLTibLG9a
LW5bLG9aLW5gLGMbMg:@00=b00=gLP002P00=W:@LY1bT7:@LY1bT7N@Mi1bT7NAMi5gTGNAT96@TI2A
T96@TI2AT96@TI2AT96@TI2AT96@T@F@00>AT9000Y02LAl000R@Si2?T8n@S`iB00F?MI1aT0060003
CE9=02l020L000H7<000375>LTibLG9aLW5bL@Qb00egLWMbT7:@LY1bT7:@00d003nAMi5gTGNAT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2FT9J@UY2FT9J@UY2FT9J@[i2_T:l00W4S00B@
012ET7YBNU9eDWEBME9eDWEf0Xl00i2?00050003MDe`02l020L000H7<0007aQa674HLQQb679>LW5b
LG9aLW5bLG9aMg9gLWMaLW403`00>7:@LY1bT7:@LY1bT7:@Mi1bT7N@Mi5gTGNAT96@TI2AT96@TI2A
T96@TI2AT96@TI2AT96@TI2A0i00196@TI02LB@000J@Si1fT8l8DP08ME9fMY1fT7480003DTd002h0
20L000H7<000375>LTibLG9aLW5bL@Qb00UgLWMbT79gLWL04P00?WNAMi5gTI2AT96@TI2AT96@TI2A
T96@TI2AT96@TI2AT96@TI2AT9J@UY2FT9J@UY2FT9J@UY2FT:n@[i2_0W4W00:@016ET91BMU9eDWIe
T7J@Si2?T007009=00=B0000;@081`001PLa008H01Qa678HLQQbCW9>LW5bLG9aLW5bLG9aLW4D000e
T7:@LY1bT7:@LY1gT7N@LY1gTGNAMi5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT9401I00196@
T902LBP0016@MY2?T59eDY1fT7J@MY1fT0090003DTe=02h020L000H7<P002W5>LTibLG9aLW48LP05
Mg9gLWL05`00?I5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2FT9J@UY2FT9J@
UY2FT9J@UY2FT9H00W4Y0003T8n@00>@00YfSi2?T8n@Si2?2002C@03M@0002d020L000H7<`005G4H
LQQb679>LTibLG9aLW5bLG9aLP0H000fLY1bT7:@LY1bT7:@Mi1bT7N@Mi5gTGNAMi6@TI2AT96@TI2A
T96@TI2AT96@TI2AT96@TI2A0i00196@TI02LBX000n@MY1fT8n@MY1fT7J@MY002@02CBl020L000H7
<P002W5>LW5bLG9aLW48LP03Mg9b01T003fAMi5gTI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT9J@UY2FT9J@UY2FT9J@UY2FT9J@UY2F009a;0000hn@T003T006Si2?T8n@2`000e80000]00P7
00061c<001=>674HLQQbCW9>LW5bLG9aLW5b01/003>@LY1bT7:@LY1gT7N@LY1bTGNAMi5gTGNAT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT9401I00196@T902LBd000]fT7J@MY1fT7J@MP0:0003C@0002h0
20L000H7=0002W9>LW5bLG9aLW47LQ`003egTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TI2FT9J@UY2FT9J@UY2FT9J@UY2FT9Ia02l000V?T8n@Si2?T8l0?0081`001PLe008H00ea678H
LTibCW9>LTib01h0025bT7:@LY1bT7:@LY1gT7N@Mi1gT7N@Mi5gTGNAMi5gTGL039001i6@TI2AT940
1Y02LC0000MaT76@MY1a00`000==0000;P081`001PLf0008LW5bLG9aLW43LP03LG8001h003ZAT95g
TGNAMi6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT9J@UY2FT9J@UY2FT9J@UY2FT9J@0W4a0006
MY1fT8n@30000dd0000^00P700061cL000]a678HLQQbCW9>LP0Q000PT7:@LY1bT7:@LY1gT7N@Mi1g
T7N@Mi5gTGNAMi5gTGL<T009TI2AT96@TI2A00B@0W4c0003T76@03h020L000H7=`002W5bLG9aLW5b
LG8R000iMi5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@UY2FT9J@UY2FT9J@UY2F
T9J@009a=0000i00000;0003DP0002l020L000H7=`026007LQQb678HLP0T000OLY1bT7:@LY1bT7N@
Mi1gT7N@Mi1gTGNAMi5gTGNAM`0<T007TI2AT96@T@06T09a=0000g40000;0003C@0002l020L000H7
>002L@05LW5bLG809@00=i5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2FT9J@
UY2FT9J@UY2FT9H00W4d0003Si0000/000==0000<0081`001PLi0005LQQb67809P007g:@LY1bT7:@
LY1gT7N@Mi1gT7N@Mi5gTGNAMi5gTGL039001I6@TI2A00N@00=aCP00<P000gJ@L@0<0003C@000300
20L000H7>@001G5bLG9a02H003JAMi5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2FT9J@UY2FT9J@UY02LC@000B@MY1f2P000dd0000a00P700061cT000E>678HLP0W000TT7:@LY1b
T7:@LY1gT7N@Mi1gT7N@Mi5gTGNAMi5gT7N@Mi1g1I001GN@T95g00N@00=>L@00<P001I1aT76@00X0
00==0000<@081`001PLj0004LW5bLBL003AgTGNAMi5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TI2AT9J@UY2FT9J@UY2F0W4c0006T7J@MY1f2P000dd0000a00P700061c/000=b6000:0009G:@
LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gTGN@Mi5gT7N@Mi1gT7L00i002GN@Mi1gT7N@M`02T003LDh00340
00R@LI1aT76@LCd020L000H7?0000g40000X000aMi5gTGNAMi5gTGNAMi5gTI2AT96@TI2AT96@TI2A
T96@TI2AT96@TI2AT96@TI2AT002LC8000V@MY1fT7J@MY001`000dd0000c00P700061cd000<H0000
:0009G:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi5gT7NAMi1gTGL02i000dh0000a000:MY1fT76@
LHmaSc`020L000H7?0000g80000Y000aTGNAMi5gTGNAMi5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@L@0b000;MY1fT7J@MY1fT7H01@000dd0000d00P700061c`000<HLP00:P009Y1b
T7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7NAMi1gT7N@Mi1g2I0b000<MY1fT7J@LHmaSg5f?0081`00
1PLl0003LW5b02X002mgTGNAMi5gTGNAMi5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2A
T00a000>MY1fT7J@MY1fT7J?MXlk00P700061c`000@HLAQb:P00:g:@LY1bT7:@LY1bT7:@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7L00i0`000?MY1fT76@LI1aSg5fLGIa008000==0000=P081`00
1PLk0006CW9aLW5b:@00;Y5gTGNAMi5gTGNAMi5gTI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TG4_000?MY1fT7J@MY1fT7J@MXma008000=`0000=`081`001PLl0006674HLAQb:@00;I1bT7:@LY1b
T7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gL@0_000?T7J@LI1aT76@LGIaMW5f03`0
20L000H7?000275>LW5bLG8H9`00;GNAMi5gTGNAMi5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TI2AL@0^000=T7J@MY1fT7J@MY1fT002MSd020L000H7?@001g4HLAQbCW80:000;7:@LY1bT7:@
LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7Ma;@00491fT76@LI1aT75fLGIaMW4m00P7
00061ch000QbCW9>LW5b62L002]gTGNAMi5gTGNAMi5gTGNAMi5gTGNAT96@TI2AT96@TI2AT9J@TI2F
T9Ia02/00Y003gN@Mi1fT7J@MY1fT75fL@0n00P700061ch000PHLAQb679>LRP002YbT7:@LY1bT7:@
Mi1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mg4Z000BMY1fT7J@LI1aMg5gLGIaMW5f?P081`00
1PLo0009CW9>LTibLG8H02H002RAMi5gTGNAMi5gTGNAMi5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@
0W4Y000CMi1gT7N@MY1fT7J@MY1aMg5fL@0n00P700061cl000Ua674HLQQbCW809P00:W:@LY1bT7:@
LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi0HLBP001=fT7J@MY1aT75gLGIaMW5fLGI=03l0
20L000H7@0002W5>LTibLG9aLQPT000GTGNAMi5gTGNAMi5gTGNAMi5gTGNAT9402g401i6@TI2AT740
:00057N@Mi1gT7J@MY1fT7J@LGMaMW5f?`081`001PM0000:674HLAQb679>LRD002V@LY1bT7:@LY1b
T1Qa674HLAQa674HLAQa675>LDia674HLAQa674HL@0V000DLI1aT76@LGMaMg5gLGIaMW5fCW5000P7
00061d0000aa675>LW5bLG9aLQPT0005TGNAMi4017400dia000@0007J`1[06/0J`020003LI5a00Ia
7P005I1gT7N@MY1fT7J@MY1fT7IgLGIaMP1100P700061d4000ea674HLAQbCW9>LP1>028000HHLAQa
67460006JhA[Q6^41`000f/0Q002J`0C06/0Q6^4Jg5gT7N@Mi1gLDiaCP0J000ET76@LI1aT75gLGMa
Mg5gLGIaMTia048020L000H7@P002W4HLG5bLG9aLW42LP03L@0001d00g47009[00F4JhA[Q002J`L0
0V/00hA[J`05009[00B4LI1g18@5T0=a5P004I1gT7N@MY1fT7J@MY1fT7Ig009f00=aMP00@P081`00
1PM3000<LAQa675>LTibCW9>70001g4HLG9gLWL02@02J`03Q6^400H00f/00h@0J`07000DQ6^4Ji1g
T6^4JhA[T7N@Mg5>LDhB000ET76@LI1aT76@LGMaMg5gLGIaMW5f04@020L000H7A002L@0:LW5bLG9a
LW5b61T000Y>LDj@Mi1[Q6^:2@02J`03Q6^400D000B4JhA[200028Y[Q6^@LI2:0X@0191gT7L3T003
Mi1g00Aa3P004I1gT7N@MY1fT7J@MY1fT7Ig009f00=aMP00A0081`001PM6000<674HLQQb679>LTib
4P001QQa67MbM`E[00EaJgMbM`09009[00>4Jf/00`000hA[Q009000LLGN@JhAgT7N5LY1gT7N@Mi1g
T6^4JhAgT7MaCPT001IfT7J@MY1aMg5gLGMaMg5gLGIaMW5fAP081`001PM7000=679>LTibCW9aLW5b
600@0003LG9g00=b00][LW5[Q6^@LY1gT009009[00B4Jf/00V/00hA[0008000KLI1gRV^@Mi1aT7N@
Mi1gT7N@Mi1aQ6^4Ji1g009a1`005WN@Mi1gT7N@MY1fT7J@LGMaMg5fLGI700P700061dP000`HLAQb
678HLTibCW82CPd000=aLWL00f/01GMbMg9g00=[00EaLWMbM`09009[00F4J`00Q002J`T001n@Mi1a
T7N@MhEgT7:@Mi1bT7N@MhA[Q764Ji1gL@1a00@001IfT76@LGMaMg5gLGMaMg5gLGIaMW5fB0081`00
1PM9000;CW9>LTibLG9aLW400g800aP000070004674HM`=[00MgLWMbMg9a00=[00F;LY1bM`070003
J`24009[0`000hA[0007000QLGN@Mi1bT7N;LI1gT7N@Mi1gT764Ji1gT7N@Mi1gT4ia008001EgT7N@
Mi1gT7J@MWMfMgIgLGMaMg40BP081`001PM:000>674HLAQbCW9>LTibCW89009a00=gJf/00V/01fag
LWMbMg8016/0175bMg8900=[0P03J`P0025>T7N@Mi1gT6^@LY1bT7:@LY1aQ6^@Mi1gT7N@Mi1fT000
0W40591aT76@LGMaMg5gLGMaMg5fLGIaB`081`001PM;000;CW9>LW5bLG9aLW400g800aP000030004
675bM`9[00aaJf][LWMbMg9gLW43J`04T7:@LPP000>4Jh@00`000hA[0007000OLGN@Mi1gT7N:Mi1g
T7N@Mi1[RWN@Mi1gT7N@Mi1gT002000ECW5gT7N@MY1fMgIgMWMfMg5gLGIa04`020L000H7C0003QQa
675>LTibCW9>LTib1@00275>MdigLWMa0f/0279gLWMbMg9g0f/017MbMg4700=[0`000f^40007000M
Mi1gT7N@Mi1bT79gLY1aMf^4Ji1gT7N@Mi1fT7H01P004Y1aT75gLGMaMg5gLGIaMW5fCTd020L000H7
CP00379>LW5bLG9aLW5bL@9b00aa074HMg5gLWMbMg42J`9a00QgLWMbMg9gLP9[00=gLWL020000f^4
000200=[1`00791gT7N@Mi1gRV^:Mi1aRV^4LI1gT7N@Mi1gT7L3000EJ`00LDj@MY1fMgIgMWMaMg5g
LGIa04h020L000H7C`006W4HLDibCW9>LTibCW9>LP1bCV]>Mg5gLGMb0f/02G5gLWMbMg9gLP02J`9b
2@02J`<00V/60008LGN@Mi1gT7L2L@03Mf][00A[00]aT76@Mi1gT7IgMP04009[0P004740Mg5gLGMa
Mg5fLGIaMTi?00P700061dl0015>LDibLG9aLW5bLG9aLQQa0002J`09Mg9gLWMbMg9g009[00ebMg9g
LWMbMg9aJg5b00P00f/20003Jh@000D000Y>T7N@Mi1gT6^@0WL03V^4JhAgT7N@Mi1gT7N@1@02J`0B
Q6/0CW5fT7IgMWMaMg5fLGIaD0081`001PM@000<674HLAQaCW9>LTib0Th3000;J`1bCW9aMg5gLGL0
0V/03G5gLGMbMg9gLWM[Mg402002J`800f/50005LG9gMi000WL0179gLGL3J`9a00AgLI1a0gL00g40
00040005J`1[06/00`002GMaMg5gLGIaMP02LE8020L000H7D`0034iaCW9>LW5bLG8H009[0`002f/0
LG9gLWMbMg9a009[00EbMg9gLP03M`04LF][M`H00V/200=[10000g5gM`07M`03LF][00Eg00>@Mi00
1`02J`P000QaMWMaMg5gLEH020L000H7E@001g4HLAQbCW800Th30003J`0000<000U[075>Mdi[LG80
0f/02gMbMg9gLWMaJg5g00H000A[00000V/50009Mg5gLGMaMdig009[00UaMg5gLGMaMg4010000f/0
00090007L@1gLGI>MP1H00P700061eD000PHLDibLG9aLP9a00<0J`000f/4000367M[009[0W803gMb
Jf]aLWMbMg9gLF]bM`0500=[0`000f/000030008Mg9gLWMbMg86M`03T7N@00D000=[00001@04J`06
04iaLGMaFP081`001PMH0006674HLDib0`001F/0001[00L000Mg079aMg5g009[00YaMg5gCWMaJ`1g
1P02J`<00V/60003L@1g00=a00IgLGMaMg4=00=[1002L@03MTi>05X020L000H7F@03L@04LW5a60P0
0f/3000ALG9gLGMbMfa[K7MbMg9gLWL00V/00g400005009[0`03J`D00V/A00A[1P000degC@1L00P7
00061e/000Ea675>60090003J`1[00D001=bCWM[MdiaJgM>MdigCWM>J`1b00T000=[06/00`000f/0
J`0A00=[1P0016/007IN00P700061ed000A>LG5a0P02J`L00f/40006CWMbLG9g0V/02G9gLWMbMg9g
L@02J`03Mg4000l000=[00001@000f/00004J`0306/000D000=[LDd0H0081`001PMO0003LAQa00D0
00=[00001`000f/00002000RCW9>LTig06]>MdigCW9>LTigCW40J`1[06/0JdigCV/0J`T00V/2009[
10001G40J`1>068020L000H7H@000g5b0005009[0`000f/00002009[10000g5gL@02L@03Mg5a009[
015bMg9gLWMbMg9gLWMaMg5gJ`0400U[0P02J`8000=>LF/00f/2LF<020L000H7H`000g40J`070003
J`0000<000E[0000J`05000GLTibCV]>LTibCW40L@1a079>MdigCW400`000f/0000:000:CV/0J`1a
CWH0CV@020L000H7J0001g5>LTib06/01@03J`h016/90005Mg5gLGL00f/01W5[LG5gJ`9a00=gLGL0
0g400gIaL@1W00P700061fT00QP01f/0K4ibCW800`000f/000030003J`1[00d000=[00000`005f]>
LTigCWM>MdigCWM>Mdig075>LDia06X020L000H7K0003G40LTibCW9>LTibCW801@001F/0J`1[00<0
00=[00000`0017MaLW42J`04Mg5gJ`9a00EgLGMaM`02L@04JgMaM`AaJ`081`001PM^000d674HLDia
CW80J`1aCW9>LTibCW9>LTibCW80LDibCV]>LTib06/0JdibCWM>Mdi[CW40Jg0020L000H7LP001g40
J`1[07400V/09DibLG9aLW5b06/0JdiaCWM>MaP0LG9aMg5gLGMaMg5gCW5>LDh0LP081`001PMg000C
K1QaCW5>LDibCW5>LTia06/0LP090009LDiaCW5>LDia07H020L000H7O0001g40J`1[07401@000f/0
00090003J`0007/020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7003o1b47
003o1b47003o1b47003o1b47003o1b47003o1b470000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "The last possible mapping type for the scene is the ",
  StyleBox["MVTorusMapping.",
    FontFamily->"Courier"],
  " It uses a torus as reference surface and the texture coordinates are \
obtained by"
}], "Text",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(TraditionalForm
    \`\((x\_t, y\_t)\) \[RightTeeArrow] 
      \((\(tan\^\(-1\)\)(y/x) + \[Pi])\)/2  \[Pi], 
    \(sign(z)\)\ \(\(cos\^\(-1\)\)(z\/\@\(x\^2 + y\^2 + z\^2\))\)/2  
      \[Pi]\)], "DisplayFormula",
  FormatType->StandardForm,
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; \n
    MVShow3D[\n\tGraphics3D[Torus[1, 1]], \n\t
      MVTexture -> "\<`ChessBoard`\>", \n\t
      MVTextureMapType\  \[Rule] \ MVTorusMapping, 
      MVScaleTexture\  \[Rule] \ {3, 2}\n]\)], "Input",
  CellLabel->"In[7]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVScene3D[] \[DoubleLeftArrow] MVTexture2D[3], 
      MVTorusMapping, {256, 256, MVRGBTexture}}\)], "Output",
  CellLabel->"Out[7]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[8]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F2F2F2F2F
302F2F2F2F2F302F2F2F2F2F302F2F302F2F3010103010103000003000003000003000003F0000\
3000003F0000300F0F3F0F0F302F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000301010\
3F10103010103F10103010103F00003000003F00003F00003F00003F00003F0000
3F00003F00003F00003F00003F00004000003020203F2F2F3F2F2F402F2F3F2F2F402F2F3F2F2F\
402F2F3F2F2F402F2F402F2F402F2F3F2F2F402F2F402F2F402F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F2F3F202F3F2F2F3F0F0F\
3F0F0F400F0F3F00004000003F00004000003F00004000003F00004000003F0000
4000003F0F0F400F0F402F2F40302F402F2F403030402F2F40302F402F2F403030402F2F40302F\
402F2F4F3030402F2F4F302F402F2F4F3030402F2F4F302F402F2F4F3030402F2F
402F2F402F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000302F2F302F2F3F2F2F302F2F3F2F2F3F2F2F402F2F3F2F2F\
402F2F3F10104010103F10104000003F00004000003F0000400000400000400F0F
400F0F40302F402F2F40302F402F2F40302F402F2F40302F402F2F40302F402F2F4F302F402F2F\
4F302F402F2F4F302F402F2F4F302F402F2F4F302F402F2F4F302F402F2F4F302F
402F2F402F2F402F2F402F2F3F2F2F402F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F2F2F302F2F3F2F303F2F2F3F2F303F2F2F4030303F2F2F402F30402F2F403030\
402F2F403030402F2F403030402F2F403030402F2F4F3030402F2F4F3030402F2F
4F303040302F4F3030402F2F4F30304F302F4F30304F302F4F30304F302F4F30304F302F4F3030\
4F302F4F30304F302F5030304F302F4F30304F302F5030304F302F4F30304F2F2F
4F30304F2F2F4F302F402F2F4F3030402F2F40302F402F2F201F1F201010000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030202F302F2F
302F2F3F2F2F302F2F3F2F2F3F2F2F3F2F2F3F2F2F402F2F3F2F2F402F2F1F1010201F1F201010\
201F1F201010201F1F201010201F1F201010201F1F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F1F1F
4F302F4F302F4F2F2F4F302F4F2F2F4F302F402F2F4F302F402F2F201F1F1F1010000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F2F302F2F302F2F3F2F303F2F2F
3F2F303F2F2F4030303F2F2F402F30402F2F201F1F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000201F105030304F302F5030304F302F4F302F4F2F2F2F1F1F201F10000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F202F302F2F302F2F3F2F2F302F2F3F2F2F3F2F2F3F2F2F
3F2F2F3F2F2F3F2F2F201F1F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F1F1F4F2F2F4F302F201F102F1F1F201010\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F30\
302F2F302F2F302F2F3F2F30302F2F3F2F303F2F2F3F30303F2F2F402F301F1010
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000201F102F1F1F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F2F2F202F302F2F302F2F\
302F2F302F2F3F2F2F302F2F3F2F2F3F2F2F3F2F2F1F1010000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F1F1F2F1F10\
2F1F1F4F302F2F1F1F000000000000000000000000000000000000000000000000
0000000000000F2F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F2F302F2F302F2F302F30302F2F3F2F30\
302F2F3F2F303F2F2F3F2F301F1010000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000030201F\
2F1F1F503F3050302F503F3050302F503F302F1F1F2F1F1F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F2F2F2F302F2F302F2F302F2F302F2F3F2F2F302F2F\
1F101F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F1F10\
2F1F1F50302F5F3F2F50302F50302F50302F503F2F4F302F50302F4F302F2F1F1F
201F10000000000000000000000000102000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F30302F2F302F2F302F2F3F2F30302F2F201F1F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000030201F\
2F1F1F30201F503F2F5F3F305F3F2F5F3F3050302F5F3F3050302F503F3050302F
503F3050302F5030302F1F1F0000000000000000000F2F00000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000302F2F302F2F302F2F1F1010000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30201F2F1F105F3F305F3F2F5F3F3050302F5F3F2F50302F5F3F3050302F5F3F2F
50302F503F2F50302F50302F50302F50302F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F301F1010000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30201F30201F5F3F305F3F305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F305F3F2F
5F3F30503F2F5F3F30503F2F5F3F3050302F403F2F20300F000F00000000102F00000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000001F101F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
30201F301F1F30201F5F3F2F5F3F305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F30
503F2F5F3F30503F2F5F3F3050302F5F3F2F50302F403F200F30001030000F2F000000000F2F00\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030201F30201F6040305F3F306040305F3F306040305F3F306040305F3F2F
6040305F3F305F3F305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F30103000103000103000103000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F1010000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030201F30201F30201F6040305F3F306040305F3F306040305F3F2F5F4030
5F3F2F5F40305F3F2F5F3F305F3F2F5F3F305F3F2F5F3F30503F2F103F001030001030000F3000\
103000000F000F2000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F2F2030201F30202060403060403F604030604030604030604030604030
6040305F3F306040305F3F306040305F3F306040305F3F305F4030303F101F3F00103F00103F00\
20300F103000102F000000000F2F00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000001F1F1F302F2F302F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030201F30202030201F604030604030604030604030604030604030604030
5F3F306040305F3F306040305F3F2F6040305F3F306040305F3F2F5F3F30403F20103F00103000\
103F0030301F4F302F0F2F00103000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F30302F2F3F2F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F2F2030201F3F2F2060403F6F4F3F6040306F4F3F60403060403F604030
6F4F3F60403060403060403060403F6040306040305F40306040305F40306040305F3F2F1F4000\
103F00103F0020300F503F304F302F103000000F00102F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F1010302F2F302F2F3F2F30302F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000030201F3F202030201F6F4F3F6040306F4F3F6040306F4F3F604030604030
60403060403F6040306040306040306040306040306040306040306040305F3F2F5F40304F3F20\
2F3F0F103F00103F0020300F503F2F4F302F2F1F1F0F2F000000000F2F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000302F2F3F2F30302F2F3F2F30302F2F3F2F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F2F203F20203F2F206F4F3F6F4F3F6F4F3F6F4F3F60403F6F4F3F6F403F
6F4F3F6040306F4F3F6040306F4F3F6040306F4F3F60403060403F604030604030604030604030\
4F3F201F4000103F00103F0050302F503F3050302F4F302F0F2F00103000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
1F1F1F302F2F302F2F302F2F3F2F303F2F2F3F2F30000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F2030201F6F4F3F6F4F3F6F4F3F6F403F6F4F3F6F403F6F4F3F
6040306F4F3F6040306F4F3F6040306F4F3F6040306F4F3F6040306040306040306040305F3F2F\
6040305F3F2F104000103F00103F00303F10503F2F4F302F4F302F40302F103000
000F00000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000302F30\
302F2F3F2F30302F2F3F30303F2F303F2F303F2F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F203F2F203F2F20704F406F4F3F7050406F4F3F6F4F3F6F4F3F
6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6040306F4F3F6040306F4F3F604030604F3F\
6040306040305F3F2F4F4020103F00103F00103F005F3F30503F2F503F301F300F
1030000F2F00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000101010302F2F302F2F\
302F30302F2F3F2F303F2F303F2F3030202F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F203F20203F2F206F4F3F6F4F406F4F3F704F406F4F3F6F4F3F
6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6040306F4F3F604F306F4F3F6040306F4F3F604030\
6040305F40306040305F3F2F104000103F00103F00403F1F503F2F50302F503F2F
4F302F1030000F2F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F2F302F30302F2F3F2F30\
302F2F3F30303F2F303F30303F2F303F2F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F203F2F203F2F207050406F4F407050406F4F3F7050406F4F3F
7050406F4F3F7050406F4F3F704F3F6F4F3F70503F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F2F4F10\
6040306F4F3F6040306040305F402F4F402F103F00104000103F005F3F30503F2F
503F304F302F303F20103000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F1F2F202F302F2F302F2F302F30302F2F\
3F2F303F2F303F2F303F2F303F2F303F2F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F0000000F0F0F
0000000F0F0F0000000F0F0F0000000F0F0F0000000F0F0F0000000F0F0F0000000F0F0F000000\
0F0F0F0F00000F0F0F0000000F0F0F0F00000F0F0F0000000F0F0F0F00000F0F0F
0000000F0F0F000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F2F203F20203F2F206F4F407050406F4F3F7050406F4F3F705040
6F4F3F704F3F6F4F3F7050406F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F6F4F3F604F30\
1F4F002F400F604F306040306040305F402F5F4030103F0F103F00103F00203F10
50302F503F2F4F302F50302F0F3000201F1F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000100000302F2F2F2F2F302F30302F2F3F2F30302F303F3030\
3F2F303F30303F2F3040303F3F2F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F0F00000F0F0F0F00000F0F0F0F00000F0F0F0F00000F0F0F\
0F00000F0F0F0F00000F0F0F0F00000F0F0F0F00000F0F0F0F00000F0F0F0F0000
0F0F0F6F4F4F70504F6F3F3F7040406F403F7030306F302F70303070302F70303070302F703030\
70302F70303070302F70303070302F70303070302F7F303070302F7F303070302F
7F3F300F0F000F0F0F0F00000F0F0F0F0F00000000000000000000000000000000000000000000\
0000000000003F2F20402F2F3F2F207050407050407050407050407050406F4F40
7050407050407050406F4F3F7050406F4F3F7050406F4F3F7050406F4F3F70503F6F4F3F70503F\
104F001F4F00104F002F4F106040306040305F40306040305F3F2F104000103F00
104000503F2F503F3050302F503F304F302F4F302F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000200F0F302F2F2F202F302F2F302F2F302F30302F2F3F2F303F2F30\
3F2F303F2F30402F3F3F2F30402F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F0000000F0F0F602F2F6F2F30602F2F6F2F30602F2F6F303060404F604F4F60404F\
6F4F50604F4F6F4F4F6F4F4F6F4F506F4F4F6F4F4F6F4F4F6F50506F4F4F6F4F4F
6F4F4F7050506F4F4F7040406F3F3F7040406F00007000006F00007000006F0000700000700000\
7000007000007000007000007000007000007000007000007F00007000007F0000
7000007F303070302F70303070302F7F30300000000F0F0F0F00000F0F0F000000000000000000\
0000000000003F2F203F2F203F2F20704F407050407050407050406F4F3F705040
6F4F3F7050406F4F3F7050406F4F3F7050406F4F3F70503F6F4F3F70503F6F4F3F70503F6F4F3F\
404F20104F001F4F001040002F4F0F6040306040305F402F604030303F1F104000
103F00103F004F3020503F2F50302F503F2F4F302F2F301F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F2F302F30302F2F302F30302F303F2F30302F303F30303F2F303F303F\
3F2F3040303F402F3F402F3F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F0F0000
602F30602F2F6F3030602F2F6F00006000006F00006F10106F10106F10106F4F506F4F4F6F4F50\
6F4F4F6F50506F4F4F6F50506F4F4F7050506F4F4F7050506F4F4F7050506F4F4F
7050506F4F4F7050506F4F4F70504F6F4F4F700000700000700000700000700000700000700000\
7000007F00007000007F00007000007F00007000007F00007000007F00007F0000
7F00007000007F00007000007F000070302F7F3F3070302F7F303070302F0F0F0F0F0F000F0F0F\
0F00000F0F0F000000402F2F3F2F207F504F705040705040705040705040705040
70504070504070504070504070504070503F70504070503F7050406F4F3F70504070503F705040\
6F4F3F40502F404F201F4F00104F001F4F002F400F604F306040306040305F402F
10400F103F00104000203F0F503F30503F2F503F304F302F503F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F2F302F2F2F2F2F302F30302F2F3F2F30302F303F2F303F2F303F303F3F2F30\
40303F3F2F30402F3F3F2F30000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F0000000F0F0F000000602F2F5F2F2F602F30
602F2F6000006000006F00006000006F00006000006F3030602F2F6F30306F4F4F6F4F506F4F4F\
6F4F506F4F4F6F50506F4F4F6F4F506F4F4F7050506F4F4F7050506F4F4F705050
6F4F4F7050506F4F4F7050506F4F4F70505070202070202070201F700000700000700000700000\
7000007000007F00007000007F00007000007F00007000007F00007000007F0000
7000007F00007000007F00007000007F00007000007F00007000007F000070302F7F303070302F\
70302F70302F0F0F0F3F2F203F2F206F4F4070504F705040705040705040705040
705040705040705040705040704F3F70504070503F7050406F4F3F7050406F4F3F7050406F4F3F\
70503F6F4F3F6F503F3F4F20404F20104F001F4F001040002F4F0F5F402F604030
5F3F2F5F402F103F00104000103F00303F10503F2F503F2F4F302F4F302F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F302F30302F2F302F30302F303F2F30302F303F303F3F2F303F303F3F2F3040303F\
402F3F40303F3F2F3040303F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F0000000F0F0F5F2F2F602F305F2F2F602F30600000600000600000
6F00006000006F00006000006F00006F00006F0000604F4F6F4F50604F4F6F50506F4F50705050\
6F4F507050506F4F507050506F4F507050506F4F507050507050507050506F5050
7050507050507050507050507050507050506F504F6F4F4F7050507000007F00007000007F0000\
7000007F00007000007F00007F00007F00007F10107F1F107F10107F10107F1010
7F00007F00007F00007F00007F00007F00007F00007000007F00007F00007F00007000007F0000\
7010107F3F3070302F3F2F2F3F2F203F2F206F4F407050406F5040705040705040
7F5F4F7050407050407050407F5F4070504070504070504070504070504070504070503F705040\
70503F7050406F4F3F70503F6F4F3F405020104F001F4F00104F001F4F002F400F
604F305F402F5F40304F3F201F4000103F00303F10503F2F503F30503F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F2F302F302F2F2F302F30302F2F302F30302F303F2F303F2F303F303F3F2F30402F3F3F2F30\
40303F3F2F30402F3F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F000F000000\
0F0F0F5F202F5F2F2F5F202F602F305F00006000005F0000600000600000600000
6000006F00006000006F00006000006F2020601F1F6F4F506F4F506F50506F4F4F6F4F506F4F50\
6F50506F4F4F6F50506F4F507050506F4F4F7050506F4F507050506F4F4F705050
70505070505070504F70505070505070505070504F7050507050507F101070100F7F1010700000\
7F00007000007F00007000007F000070100F7F10107F10107F10107F100F7F5F50
7F10107F10107F100F7F10107F00007F00007000007F00007000007F00007000007F0000700000\
7F000070100F7F10103F2F203F2F203F2F200000000000000000000000006F5040
6F4F406F50406F4F3F6F504070504070504070504070504070504070504070503F70504070503F\
7050406F4F3F70503F6F4F3F6F503F6F4F306F4F3F3F4F201F4F00104000104F00
20400F6040305F402F5F402F403F20104000103F00103F00303F1F503F2F4F302F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F2F
302F302F2F2F302F30302F303F2F30302F303F303F3F2F303F303F3F2F3040303F402F3F40303F\
402F3F40303F402F3F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F0F00000F0F0F0000005F2F305F2F2F5F2F30\
5F2F2F6000005F00006000005F00006000006000006000006000006F0000600000
6F00006000006F00006F00006F20206F2F306F30306F4F506F50506F4F507050506F4F50705050\
6F4F5070505F6F50507050506F505070505F70505070505070505070505F705050
7050507050507F5F5F7050507F5F507050507F5F5F7050507F5F507030307F3F3F7F30307F0000\
7000007F00007F00007F00007F00007F00007F30307F3F307F3030805F5F7F5F50
7F5F507F5F50805F507F5F507F00007F00007F00007F00007F00007F00007F00007F00007F0000\
7F00007F00007F00003F2F2F3F2F203F302F000000000000000000000000000000
0000000000000000006F50407050406F50407050406F50407F5F407050407F5F40705040705040\
70503F70504070503F7050406F4F3F70503F6F4F3F6F503F404F20405020104F00
1F4F001040002F4F105F402F6040305F402F3F4020103F00104000103F00303F204F302F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F2F302F30
2F2F2F302F30302F2F302F30302F303F2F30302F303F2F3F3F2F303F2F3F3F2F3040303F3F2F3F\
40303F3F2F3F40303F000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0000000F000F50202F5F2F2F50202F5F2F2F5F00005F00005F0000\
5F00005F00006000005F00006000005F00006000006000006000006000006F0000
6000006F00006000006F0000602F306F4F506F4F506F50506F4F506F50506F4F507050506F4F50\
7050506F4F507050506F4F507050506F50507050506F5050705050705050705050
705050705050705050705F507050507050507050507F5F50705050705050705050705050700000\
7F00007F00007F00007F00007F00007F00007F00007F00007F00007F30307F3F30
7F30307F5F507F5F507F5F507F5F4F7F5F507F5F4F7F4F407F00007F00007F00007F00007F0000\
7F00007F00007F00003F2F203F2F203F2F20000000000000000000000000000000
0000000000000000000000000000000000000000000000006F4F3F6F50406F4F3F6F5040705040\
7050404050207050406F4F3F70503F6F4F3F6F503F6F4F306F4F3F6F4F30404F20
3F4F1F104F001040001F4F0020400F5F402F5F402F5F4030203F0F103F00103F00303F1F4F302F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000302F302F2F2F
302F30302F30302F30302F303F303F3F2F303F303F3F2F3040303F3F2F3F40303F402F3F403040\
40303F40303F402F3F000000000000000000000000000000000000000000000000
0F0F0F0F00000F0F0F50202F5F2F3050202F5F2F305000005F00005F00005F00005F0000600000\
5F00006000005F0000600000600000601010600F106F10106010106F0000600000
6F00006F00006F00006F00006F505F6F4F5070505F6F4F5070505F6F505070505F6F505070505F\
70505070505F70505070505F70505070505F705050705F5F705050705F5F705050
7F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7F5F507F5F5F7F5F507F2F2F\
7F20207F2F207F00008000007F00007F00007F00008000007F00008000007F0000
8000007050507F5F507F5F508060507F5F50805F507F5F50804F407F4F407F10107F1010801F10\
7F00007F00007F000040302F3F2F203F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000705040\
6F503F1F5F00405020705F4070503F7050406F503F70503F6F4F3F6F503F6F4F30
6F503F3F4F201F4F00104F001F4F002F400F2F4F0F5F402F5F403020400F104000103F00203F1F\
4F302F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F2F2F2F2F302F30
2F2F2F302F30302F2F302F30302F303F2F303F2F303F2F3F3F2F3040303F3F2F3F40303F40303F\
4F3040402F3F40303F0000000000000000000F000F0000000F0F0F00000050303F
502F3F50303F50202F5000005000005F00005000005F00005F00005F00005F00005F00005F0000\
6000005F00006010105F0F1060101060404F602F30602F306F30306000006F0000
6000006F00006F1F206F20206F4F506F505F6F4F506F505F6F4F5070505F6F4F5070505F6F5050\
70505F6F505070505F70505070505F70505070505F705050705F5F70505070505F
7050507F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7050507F5F5F7F3030\
7F3F307F30307F00007F00007F00007F00007F00007F00008000007F0000800000
7F00008010107F10107F5F507F5F50805F507F5F507F5F507F5F507F5F507F302F7F3F307F3030\
7F00007F00007F00003F2F203F2F2F3F2F20000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F4F3F1F500040502040502F6F503F70503F6F4F3F6F503F6F4F306F503F
6F4F306F4F3F604F303F4F20104F001F4F001040002F400F5F402F5F4030503F2F20400F103F00\
503F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F2F2F2F302F30302F30
302F30302F303F3030302F303F303F3F2F3040303F3F2F3F40303F3F2F3F40304040303F4F3040\
40303F4F304040303F0000000000000F0F0F4F3F40503F4F503F40503F4F50202F
502F3050202F5F00005000005F00005000005F00005F00005F00005F00006000005F1010601010\
5F0F10601010604050604F50604F506F4F506000006F00006000006F00006F0000
6F00006F202070303F6F505070505F6F505070505F6F505F70505F6F505070505F70505F70505F\
705050705F5F70505F705F5F7050507F5F5F70505F7F5F5F7050507F5F5F70505F
7F5F5F7050507F5F5F7F5F5F7F5F5F7F5F507F5F5F7F5F5F7F5F5F7F5F507F5F5F7F5F5F705F50\
7050507F5F507F00008000007F00008000007F00008000007F0000800000800000
8000007F000080000080101080605F7F5F5080605F7F5F508060507F5F508060507F5F50803F30\
7F00008000007F00004F2F203F2F2040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F50001F5F001F50004F5F2F4050207050406F503F70503F6F503F
6F503F6F4F306F503F604F30404F203F4F1F1F4F001040001F4F0020400F6040305F402F5F4030\
103F004F3F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F302F302F2F2F302F30
302F2F302F30302F303F2F30302F303F2F303F2F3040303F3F2F3F40303F40303F40304040303F\
4F304040303F40303F0000004F3F4F4F3040503F4F4F3040503F4F503F40501F20
50101F501F205000005F00005000005F00005F00005F00005F00005F00005F0F106040505F404F\
604F505F4050604F50602F30602F30602F306F00006000006F00006000006F0000
6F00006F3030604F506F505F6F4F5070505F6F505070505F6F505070505F6F505070505F6F5050\
70505F705050705F5F70505070505F705050705F5F705050705F5F7050507F5F5F
7050507F5F5F7050507F5F5F7050507F5F5F7F5F507F5F5F7F5F507F5F5F7F5F507F5F5F7F5F50\
7F5F5F7F5F508000007F00008000007F00008000007F00008000007F0000800000
7F00008000007F00008000007F3F30803F307F3F3080605F7F5F508060507F5F508060507F5F50\
803F307F3030803F304F201F4F2F204F2020000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F5F001050001F5F001050001F50004050204050206F503F6F503F
6F4F306F503F6F4F306F4F3F604F30604F303F4F1F3F4F20104000104F001040002F400F5F3F2F\
5F402F103F004F302F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F2F2F302F302F2F30302F30302F30
3F3030302F303F0F103F0F0F3F20203F1F2040303F3F303F40304040303F40304040303F4F3F40\
4F304040303F4F3040503F4F4F3F4F503F4F4F3F4F503F4F503F4F50404F500F10
5F1010500F105F00005000005F00005F00005F00005F404F604F505F4050604F505F4050604F50\
604F50604F505F404F6F00006000006F00006000006F00006F00006F00006F0000
6F4F50604F506F4F506F505F70505F70505F70505F70505F705F5F70505F705F5F70505F705F5F\
70505F705F5F70505F7F5F5F70505F7F5F5F70505F7F5F5F705F5F7F5F5F705F5F
7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F605F7F5F5F7F605F7F5F5F80605F\
7F5F5F80605F7F00008000007F00008000007F0000800000800000800000800000
800000800000800000800000800000800000803F3F805F5080605F80605080605F7F5F5080605F\
805F508060507F5F50502F204F2F20502F20000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F4F001F5F001F5F001F5F001F50001F5F001F500040502F405020
70503F6F503F6F503F6F4F306F503F604F306F4F30604F30404F20104F001F4F001040002F4F0F\
5F402F5F4030103F004F3030000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F202F2F2F302F2F2F302F302F2F2F302F30
30202F3F0F0F300F0F3F00003F1F203F20203F2F3F40303F40303F40304040303F40304040303F\
40303F0000004F3F4F4F3040503F4F4F3F40503F4F503F4F503F4F503F4F500F10
500F0F5F10105000005F00005F202F5F2F305F202F5F40505F404F604F505F404F604F505F4050\
604F505F404F6020206000006F00006000006F00006000006F00006000006F0000
6F4F506F505F6F4F5070505F6F505F70505F6F505F70505F70505F70505F70505F70505F70505F\
705F5F70505F705F5F70505F7F5F5F70505F705F5F70505F7F5F5F70505F7F5F5F
705F5F7F5F5F705F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F\
80605F7F5F5F8010107F10108010107F00008000007F00008000007F0000800000
8000008000007F00008000008000008000007F00007F5F507F5F507F60507F5F508060507F5F50\
8060507F5F508060503F2F2040302F3F2F20000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F40001F4F001050001F5F001050001F50001050001F5000
4050206F503F6F4F306F503F6F4F306F503F604F30604F30604F30604F303F401F104F00104000\
104000304010303F1F0F3000103000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302F302F2F30302F30302F3030101F300000
3F00003000003F00003F00003F303F3F2F3F40304040303F40304040303F40304040303F000000\
0000000F0F0F4F3F4F503F4F4F3F4F50404F503F4F50404F503F4F5F4050500F10
5F10105000005F00005F00005F40505F404F604F505F4050604F505F4050604F50604F50604F50\
601F206F202F6000006F00006000006F00006F00006F00006F00006F00006F0000
70505F6F505F70505F6F505F70505F70505F705F6070505F705F5F70505F705F6070505F705F5F\
70505F7F5F60705F5F7F5F5F705F5F7F5F60705F5F7F5F5F705F5F7F5F607F5F5F
7F5F5F7F5F5F7F60607F5F5F7F5F5F7F5F5F8060607F5F5F80605F7F5F5F8060607F5F5F80605F\
7F5F5F806060803F3F803F3F7F3F30800000800000800000800000800000800000
800000800000800000800000800000800000800000802F20802F2F7F5F5080605F80605080605F\
80605080605F80605040302F3F2F2F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F4F001F4F001F5F001F50001F5F001F5000
1F500040502040502F6F503F6F503F6F4F306F503F604F306F5030604F306F4F303F4F1F3F4F20\
1040001F4F00304010103F000F3000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F2F2F2F2F2F302F302F2F2F301010300000300000
3000003F00003F00003F20203F2F3F40303F3F303F40304040303F40303F000000000000000000\
0F000F000000503F4F4F3F4F503F4F503F4F503F4F503F4F50404F500F105F1010
5000005F1F205F1F205F1F205F404F5F40505F4050604F505F4050604F505F4050602F30602F30\
6000006000006F00006000006F00006000006F00006000006F10106F10106F1010
6F4F5F6F505F6F505F70505F6F505F70505F70505F705F5F70505F705F5F70505F705F5F70505F\
705F5F70505F7F5F5F70505F7F5F5F705F5F7F5F5F705F5F7F5F5F705F5F7F5F5F
705F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F605F7F5F5F7F605F7F5F5F80605F7F5F5F\
80605F7F5F5F803F3F7F3F30803F3F7F00008000007F0000800000800000800000
800000800000800000800000800000800000800000802F20802020802F2080605080605F7F5F50\
80605F80605080605F3F2F2040302F3F2F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000001040001F4F001050001F5F00
1050001F50001050004050203F4F206F503F6F4F306F503F604F30604F30604F30604F3060402F\
3F4F1F20400F104F00503F2F103F000F3000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000302F302F2F2F302F302F0F0F3000003000003F0000300000
3F00003F1F203F20203F303F40304040303F4030403F303F403040000000000000000000000000\
0000000F0F0F4F3F4F504050503F4F504050503F4F5F405050404F5F0000500000
5F2F30503F4F50404F503F4F604F505F4050604F505F4050604F5F604F50603030602F306F1010\
6000006F00006000006F00006F00006F00006F00006F00006F10106F10106F2020
7050606F505F7050606F505F705F6070505F705F6070505F705F60705F5F705F6070505F7F5F60\
705F5F7F5F60705F5F7F5F60705F5F7F5F60705F5F7F5F607F5F5F7F5F607F5F5F
7F60607F5F5F7F60607F5F5F8060607F5F5F8060607F5F5F8060607F5F5F8060607F5F5F806060\
80605F806060803F3F80403F803F3F800000800000800000800000800000800000
8F00008000008000008000008F000080000080000080000080000080000080605F7F5F5080605F\
80605F80605F80605040302F40302F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F4F001F4000
1F5F001F50001F5F001050001F50004050204050206F4F306F503F6F4F306F503F604F306F5030\
604F30604F30204F0F1F4F00503F2F103F000F3000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F2F2F2F2F302F00003000003000003000003000003F0000
301F203F303F3F2F3F3F303F3F303F4030403F2F3F000000000000000000000000000000000000\
0F0F0F000000503F4F503F4F504050503F4F504050503F4F5F405050202F5F2F30
50404F5F40505F404F5F40505F4050604F505F4050604F505F4050604F505F0F10601010600F10\
6000006000006F00006000006F00006000006F00006F00006F202F6F1F206F2020
6F505F70505F6F505F70505F6F505F705F6070505F705F6070505F705F6070505F705F6070505F\
7F5F6070505F7F5F60705F5F7F5F60705F5F7F5F60705F5F7F5F60705F5F7F5F60
7F5F5F7F60607F5F5F7F5F607F5F5F7F60607F5F5F7F60607F5F5F8060607F5F5F8060607F5F5F\
8060607F5F5F803F3F803F3F803F3F800000800000800000800000800000800000
8000008000008000008000008000008000008000008000008000008000007F5F507F60507F5F50\
80605F80605080605F3F302F40302F3F2F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
1040001F4F001050001F50001050001F5000104F004050206F4F306F503F604F30604F30604F30\
604F30604F2F604F304F401F1040004F3F2F1030000F2F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F2F2F302F302F0000300000300000300000300F0F3F2F303F202F
3F303F3F303F4030403F303F3F303F3F303F000000000000000000000000000000000000000000\
0000000F0F0F503F4F504050503F4F5F40505040505F40505040505F40505F4050
5F40505F4050604F5F5F4050604F5F5F4050604F5F5F4050604F505F40506000006000006F0000\
6000006F00006000006F00006F00006F00006F00006F00006F303070303F6F303F
7050606F505F705F6070505F705F6070505F705F60705F60705F60705F5F7F5F60705F607F5F60\
705F5F7F5F607F5F607F5F607F5F5F7F60607F5F607F60607F5F5F7F60607F5F60
7F60607F5F5F8060607F5F608060607F5F5F8060607F60608060607F605F806060806060806060\
80605F80606080504F80505080504F8000008000008000008000008F0000800000
8F00008000008F00008000008F00008000008F00008000008000008000008F403F803F3080403F\
80605F8F6F5F80605F40302F3F302F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000001F40001F4F001F50001F5F001050001F5F001F50004050204050206F503F6F4F306F503F\
604F306F5030604F30604F301040002F400F4F3F2F103000202F10000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F202F2F1F1F2F00003000002F0000301F20300F0F30303F302F303F303F
3F2F3F3F303F3F303F3F303F000000000000000000000000000000000000000000000000000000\
0F000F4F3F4F504050503F4F504050503F4F5F4050503F4F5F40505040505F4050
5F40505F40505F4050604F505F4050604F505F4050602F305F2F30602F30600000600000600000\
6000006000006F00006000006F00006F00006F00006F00006F303F6F303070303F
6F505F7050606F505F705F6070505F705F6070505F705F6070505F705F60705F5F7F5F60705F5F\
7F5F60705F5F7F5F60705F5F7F5F607F5F5F7F5F607F5F5F7F5F607F5F5F7F6060
7F5F5F7F60607F5F5F7F60607F5F5F7F60607F5F5F8060607F5F5F8060607F605F8060607F605F\
80606080605F805050804F4F80504F8000008000008000008000008000008F0000
8000008000008000008F0000800000800000800000800000800000800000802F20802F2F802020\
80605F80605F80605F3F2F2F40302F3F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0000000000000000000000000000
0000000000000000001F4F001040001F50001050001F50001050001F50003F4F20405020604F30\
604F30604F30604F30604F2F604F3010400050402F4F302F10300F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000301F202F000030101F301010301F20302F303F303F30303F3F303F3F303F
4030403F303F3F303F0000000000000000000000000000000000000000000000000000000F0F0F\
000000504050503F4F5F40505040505F40505040505F40505F40505F40505F4050
604F5F5F4050604F5F5F4050604F5F604F50604F5F602F306000006000006000006000006F0000\
6000006F00006000006F00006F00006F00006F00007000006F000070303F6F404F
705F60705060705F60705060705F60705F60705F60705F607F5F60705F607F5F60705F607F6060\
7F5F607F5F607F5F607F60607F5F607F60607F5F607F60607F5F607F60607F5F60
8060607F60608060607F60608060607F60608060607F6060806060806060806060806060806F60\
80606080606080504F8F505080504F8000008000008F00008000008F0000800000
8F00008000008F00008000008F00008000008F00008000008F00008000008F0000802F208F2F2F\
80605F8F6F5F80605F40303040302F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000002F00002F001F50001F4F001F50001F4F00002F00002000
0000000000000000000000000000001F40001F4F001050001F5F001F50001F5000105000405020\
2F500F6F503F604F306F5030604F301F4F0010400050403040302F303020000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F00002F000F302F302F2F30302F30302F30302F3F302F303F303F302F3F3F303F
302F3F000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F503F4F504050503F505F4050503F505F40505040505F40505040505F4050
5F40505F4F5F5F4050604F5F5F40505F4F505F4050600000600000600000600000600000600000\
6F00006000006F00006000006F00006F00006F00006F00007000006F404F70404F
6F505F705F606F505F705F60705060705F60705060705F60705F607F5F60705F607F5F60705F60\
7F5F60705F607F5F607F5F607F60607F5F607F60607F5F607F60607F5F607F6060
7F5F608060607F5F608060607F60608060607F60608060607F60608060607F6060806060806060\
80606080606080505080504F8050508000008000008000008F00008000008F0000
8000008F00008000008F00008000008F00008000008000008000008F0000800000800000800000\
8F6F5F80605F806F5F3F302F40302F3F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001F000010000000001F4F001F50001F5F001F60001F5F001F60001F5F001F5F00
0020000000000000000000000000000000000000001F40001040001F50001050001F5000105000\
1F50002F4F0F604F30604F30604F3020400F104F0020400F503F2F40302F3F2F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F0000\
2F0000302F302F2F30302F3F302F303F303F302F3F3F303F3F2F3F3F303F000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005040505040505F40505040505F405F5F40505F405F5F40505F4F5F5F4050
604F5F5F4050604F5F5F4F5F604F5F602F3060303F6000006000006000006F00006000006F0000\
6000006F00006F00006F00006F00007000006F00007000006F0000704F5070404F
705F60705060705F60705F60705F60705F607F5F6F705F607F5F60705F607F606F7F5F607F6060\
7F5F607F606F7F5F607F60607F5F607F606F7F60607F60607F5F608060607F6060
8060607F60608060607F60608060607F6060806060806060806060806060806F60806060806F60\
8060608F6F608050508F505080504F8F00008000008F00008000008F00008F0000
8F00008000008F00008000008F00008000008F00008000008F00008000008F00008000008F0000\
80605F8F6F60806F5F40303040302F403F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001000001F000010001F60001F50002060001F60001F60001F60001F60001F5F00
1F60001F3F001F3F000000000000000000000000000000000000001040001F5F001F50001F5000\
1050001F50001050001F50002F4F0F1F5000104F00104F0050402F503F3040302F
3F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F0000\
2F1F1F2F2F30302F302F2F30302F3F302F303F303F302F303F2F3F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F000F503F50504050503F505F40505040505F40505040505F40505F40505F4F5F
5F40505F4F5F5F4050604F5F5F2F30602F3F602F306000006000006000006000006F0000600000\
6F00006000006F00006000006F00006F00006F00006F00007000006F404F704F50
6F505F705F60705060705F60705F60705F60705F607F5F60705F607F5F60705F607F5F60705F60\
7F5F607F5F607F60607F5F607F60607F5F607F60607F5F607F60607F5F60806060
7F5F608060607F60608060607F60608060607F60608060607F6060806060806060806F60806060\
80606080606080505080504F8050508000008F00008000008F00008000008F0000
8000008F00008000008F00008000008F00008000008F00008000008F00008000008F0000800000\
8F6F5F80605F8F6F5F3F302F40302F40302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001F00001000001F001F60001F60001F60001F60001F60001F60001F5F001F6000
1F5F001F60001F5F000000000000000000000000000000000000000000000000001F4000105000\
1F50001050001F5000104F001F5000104F00104F0010400030401F503F2F4F3F2F
3F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F0000302F30\
2F2F30302F30302F3030303F302F3F3F303F302F30000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5040505040505F405F5F40505F405F5F40505F4F5F5F40505F4F5F5F4050
604F5F5F4F5F604F5F5F4F5F60202F60202060202F6000006F00006000006F00006000006F0000\
6F00006F00006F00006F00006F00007000006F00007000006F0000704F5070404F
704F50705F60705F60705F607F5F6F705F607F5F6F705F607F606F7F5F607F606F7F5F607F606F\
7F5F607F606F7F5F607F606F7F60607F606F7F606080606F7F606080606F7F6060
80606F80606080606F7F6060806F6F806060806F6F806060806F6F806060806F6F8060608F6F6F\
8060608F6F6F8050508F5F508050508F00008000008F00008F00008F00008F0000
8F00008F00008F00008F00008F00008F00008F00008000008F00008F00008F00008000008F0000\
806F5F8F6F60806F5F403F3040302F403030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001000001F00001000206F001F60001F6F001F60001F60001F60001F60001F6000
1F60001F60001F6000000000000000000000000000001000001F000000000000000000001F4F00\
1F40001F50001050001F50001050001F5000104F001F4F001040005F40304F3F2F
4F3F303F2F2F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F00002F2F2F2F2F2F\
302F302F2F30302F3F302F30302F3F302F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F503F505040505040505F405F5040505F405F5F40505F4F5F5F40505F4F5F
5F4050604F5F5F4F50604F5F5F1F206000006000006000006000006F00006000006F0000600000\
6F00006000006F00006F00006F00006F00006F00006F00007000006F404F704F50
70404F705F60705F60705F60705F607F5F6F705F607F5F60705F607F606F705F607F60607F5F60\
7F606F7F5F607F60607F5F607F606F7F5F607F60607F606080606F7F6060806060
7F606080606F7F60608060607F606080606F7F6060806060806060806F6F806060806F60806060\
806F6F80606080403F803F3F8F40408000008F00008000008F00008000008F0000
8F00008F00008000008F00008000008F00008000008F00008000008F00008010108F1F10801010\
8F6F6080605F8F6F6040302F4030303F302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
001F00001000001F001F60001F6F001F60001F60001F60001F60001F60001F6000
1F60001F60001F5F00000000000000000000000000000000105F00001000001000000000000000\
0000001040001F40001050001F5000104F00104F00104F00104F0030401F50402F
4F3F2F40302F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F2F302F0000302F30\
2F1F1F302F3F302F30302F3F000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005F405F5040505F405F5040505F4F5F5F405F5F4F5F5F405F604F5F5F4F5F
604F5F5F4F5F604F5F604F5F6000006000006F00006000006F00006000006F00006F00006F0000\
6F00006F00006F00006F00006F00007000006F00007000006F0000703F4070303F
703F40705F607F5F6F705F607F606F705F607F606F7F5F607F606F7F5F607F606F7F60607F606F\
7F5F607F606F7F60607F606F7F606080606F7F606080606F7F6060806F6F806060
80606F806060806F6F806060806F6F806060806F6F806060806F6F8060608F6F6F806F608F6F6F\
806F608F6F6F80403F8F404080403F8F00008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008F00008F00008F00008F00008F00008F00008F302F8F2F2F8F2F2F\
8F6F608F6F608F6F60403F3040302F403F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000001F000010001F6F001F6000206F001F60001F6F001F60001F6F00002F00
002F001F3F001F3F00001000001F000000000000000000001F5F001F5F001F5F00001000000000\
0000000000000000001F4F001050001F5000104F003F501F204F105F402F50402F
5040304F3F2F403030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F2F2F2F2F2F2F302F1010\
302F302F2F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5040505F405F5040505F405F5F40505F4F5F5F40505F4F5F5F405F604F5F
5F405F604F5F5F4F5F604F5F5F00006000006000006000006000006F00006000006F0000600000\
6F00006F00006F00006F00006F00006F00007000006F000070000070303F703F40
70303F705F60705F607F5F6F705F607F5F6F705F607F606F705F607F606F7F5F607F606F7F5F60\
7F606F7F60607F606F7F5F607F606F7F606080606F7F606080606F7F606080606F
7F606080606F806060806F6F7F6060806F6F806060806F6F806060806F6F8060608F6F6F806060\
8F6F6F806F608F2F2F802F2F8F2F2F8000008F00008000008F00008F00008F0000
8F00008F00008F00008F00008000008F00008F00008F00008000008F00008F403F8F2F2F802F20\
8F6F60806F5F8F6F6040302F40303040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F60001F6F001F60001F6F001F60001F60001F60000F0F00
0020000000000000000F0000000000001000000000000000105F001F5F00105F001F5F00001000\
0010000000000000000000001F40001040001F5000104F005F4F305F402F5F4F30
50402F503F2F40302F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F2F2F3000002F1010301F20\
302F30000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5F405F5040505F4F5F5F405F5F4F5F5F405F604F5F5F4F5F604F5F5F4F5F
604F60604F5F604F60604F5F6000006000006F00006000006F00006000006F00006F00006F0000\
6F00006F00006F00007000006F00007000006F0000700000700000703F4070303F
7F3F40705F607F606F705F607F606F7F5F6F7F606F7F5F6F7F606F7F606F7F606F7F606F7F606F\
7F606F7F606F7F606F80606F7F606F80606F7F606F806F6F80606F806F6F80606F
806F6F80606F806F6F80606F806F6F806F6F806F6F80606F8F6F6F806F6F8F6F6F806F6F8F6F6F\
806F6F8F6F6F8010108F1F1F8F1F108F00008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008F00008F00008F00008F00008F00008F00008F40408F403F8F4040\
8F6F608F70608F6F60403F30403030403F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000206F001F6F001F6F001F6000206F001F60001F6F00000000
0F0F0F0000000000000000000F0F0F0000000000000000001F60001F5F001F60001F5F001F5F00\
1F5F000000000000000000000000000000001040001F5000104F00604F305F402F
5F4F3050402F4F3F3040302F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F2F2F2F00002F0000302F302F2F30\
302F3F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5040505F405F5040505F4F5F5F40505F4F5F5F405F5F4F5F5F405F604F5F
5F4F5F604F5F5F4F5F604F5F600000601010600F106000006000006F00006000006F00006F0000\
6F00006F00006F00006F00007000006F00007000006F000070000070202F70202F
70202F705F6F705F607F5F6F705F607F606F705F607F606F7F5F6F7F606F7F5F607F606F7F606F\
7F606F7F60607F606F7F606F80606F7F606080606F7F606F80606F7F606080606F
80606F806F6F806060806F6F80606F806F6F806060806F6F80606F8F6F6F8060608F6F6F806F6F\
8F6F6F7F60608F00008000008F00008000008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008F00008F00008F00008F00008F00008F000080605F806F60806F5F\
8F7060806F608F6F6040302F403F3040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F3F001F40001F60001F6F001F60001F6F001F6000000000
0000000000000000000000000000000000000000000000001F5F001F6000105F001F5F00105F00\
1F5F001050001F5F000000000000000000000000001050001F5000304F1F5F4F2F
50402F50402F4F3F2F4F3F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F2F302F302F0F0F30202F302F30302F3F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005F4F5F5F405F5F4F5F5F405F5F4F5F5F4F5F604F5F5F4F5F604F605F4F5F
604F60604F5F605060604F5F60101F6010106F101F6000006F00006000006F00006F00006F0000\
6F00007000006F00007000006F0000700000700000700000700000702F2F701010
7F1F1F705F6F7F606F705F6F7F606F7F5F6F7F606F7F606F7F606F7F606F80606F7F606F7F606F\
7F606F806F6F7F606F80606F7F606F806F6F80606F806F6F80606F806F6F80606F
806F6F80606F806F6F806F6F806F6F806F6F8F6F6F806F6F8F6F6F806F6F8F6F6F806F6F806F60\
7F6060806F6F8F00008F00008F00008F00008F00008F00008F00009000008F0000
9000008F00009000008F00008F1F1F8F00009000008F2F2F8F302F8F2F2F8F706F8F6F608F6F60\
8F6F608F706F8F6F60403F304030304F3F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000001F001F4000206F001F6F001F6F001F6000206F00002F00
002F000000000000000000000000000000000000000000001F60001F5F001F60001F5F001F6000\
1F5F001F60001F5F001F5F00103000000000000000000000104F001F500050402F
5F4F3050402F504030503F2F4F3F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F2F2F202F30202F2F2F30302F3F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5040505F4F5F50405F5F4F5F5F405F5F4F5F5F405F5F4F5F5F4F5F604F60
5F4F5F604F60604F5F605060601F2060202F6020206F00006000006F00006000006F00006F0000\
6F00006F00007000006F00007000006F00007000006F000070000070101070101F
700000704F50704F507F4F5F705F6F7F606F7F606F7F606F7F5F6F7F606F7F606F7F606F7F606F\
7F606F7F606F80606F7F606F80606F7F606F806F6F7F606F806F6F80606F806F6F
80606F806F6F80606F806F6F80606F806F6F806F6F8F6F6F806F6F8F6F6F806F6F8F6F6F802F2F\
8F2F2F802F2F8F00008000008F00008F00008F00008F00008F00008F00008F0000
8F00008F00008F00008F00008F10108F1F1F8F10108F6F60806F608F6F608F6F608F706F8F6F60\
8F70608F6F608F706040302F403F3040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000001000001F001F6F001F6F001F60001F6F001F50001F5000
002F00000000000000000000000000000000000000000000105F001F60001F5F001F6000105F00\
1F5F00105F001F5F00105F001F5F00103000000000000000104000104F004F4F2F
5F402F5F4F2F50402F50402F4F3F2F40302F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F2F2F2F2F302F30302F30302F3F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5F4F5F5F405F5F4F605F4F5F5F4F605F4F5F604F605F4F5F604F605F4F5F
605060604F60605060604F606F303F60303F6F303F6000006F00006F00006F00006F0000700000\
6F00007000006F0000700000700000700000700000700000700000700000700000
7F0000704F507F1F1F7010107F60707F606F7F60707F606F8060707F606F8060707F606F806F70\
7F606F806F707F606F806F7080606F806F7080606F806F70806F6F806F7080606F
806F70806F6F806F70806F6F8F6F70806F6F8F6F70806F6F8F70708040408F40408040408F0000\
8F00008F00008F00008F00008F00008F00008F00009000008F00009000008F0000
9000008F00009000008F0000905F5F8F5F50905F5F8F6F6F8F706F8F6F6090706F8F6F6090706F\
8F70608F706F8F6F604F3F304030304F3F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000206F001F6F00206F001F6F001F6F001F5000
2050000000000000000000000000000000000000000000001F60001F60001F60001F5F001F5000\
1F4F001F60001F5F001F60001F5F001F5F000020000000000000001F4000001000
5F4F305F4F2F5F4F3050402F5040304F3F2F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F202F2F2F302F2F30302F302F2030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F50405F5F4F5F5F405F5F4F5F5F405F5F4F5F5F4F5F604F605F4F5F604F60
604F5F605060604F5F60506060303F6F303F6050606F00006000006F00006000006F00006F0000\
6F00006F00007000006F00007000006F0000700000700000700000700000700000
7000007F101F7010107F101F7F3F407F3F407F3F407F606F7F606F7F60707F606F7F606F7F606F\
806F707F606F80606F7F606F806F707F606F806F6F80606F806F7080606F806F6F
80606F806F7080606F806F6F806F6F8F6F70806F6F8F6F6F806F6F8F4040803F408F4040800000\
8F00008000008F00008F00008F00008F00008F00008F00009000008F00008F0000
8F0000905F5F8F5F508F5F5F8F6F6F90706F8F5F508F706F8F6F608F706F8F6F608F706F8F6F60\
8F706F8F6F608F7060403030403F3040302F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000002F00002F001F6F001F6F001F60001F6F00
1F60000F0F000000000000000000000000000000000000001F60001F6000105F001F60001F4F00\
1F4F00104F00000000105F001F5F00105F001F5F00000000000000000000001000
00000050402F50402F50402F4F402F50402F40302F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F302F302F2F30302F30302F30000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005F4F5F5F4F5F5F4F605F4F5F604F605F4F5F604F605F4F5F605060604F60
605060604F606F50606050606F50606050606F506F6F303F6F30406F303F6F00006F00006F0000\
6F00007000006F00007000007000007000007000007000007000007F0000700000
7F00007000007F00007F00007F3F407F00008000007F2F2F802F307F2F2F7F606F7F606F7F606F\
7F606F806F7080606F806F7080606F806F70806F6F806F70806F6F806F70806F6F
806F70806F6F8F6F70806F6F8F6F708040408F404F8F40408F40408000008F00008F00008F0000\
8F00008F00008F00008F00008F0000901F1F8F1F1F901F1F8F1F108F706F8F6F6F
8F706F8F706F9070708F706F90706F8F706F90706F8F706F90706F8F706F90706F8F706F90706F\
8F706090706F8F70604F3F304030304F3F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000002F00206F001F6F00206F001F6F00
1F6F001F6000206F000000000000000000000000000000001F60001F60001F60001F60001F5000\
1F4F001F50000000000F0F0F0020001F60001F5F001F5F00000000000000000000
001000000000504F2F50402F5040304F402F504030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F2F2F2F2F2F302F302F2F30000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F50405F5F4F5F5F405F5F4F605F4F5F604F605F4F5F604F605F4F5F605060
5F4F5F605060604F606050606050606F50606050606F30406050606F50606F10106F101F6F1010\
6F00006F101070101F6F00007000007000007000007000007000007000007F0000
7000007F00007000007F00007F00007F00007F00007F2F307F202F7F2F2F7F606F7F606F7F606F\
806F707F606F802F307F3F40804040803F40804040803F40804040803F40802F30
802F2F802F2F8000008F00008000008F00008000008F00008000008F00008F00008F0000801010\
8F1F1F8F10108F5F5F8F5F508F5F5F8F6F6F9070708F6F6F9070708F706F907070
8F706F90706F8F706F90706F8F706F90706F8F706F90706F8F706F8F706F8F706F90706F8F7060\
8F706F8F706090706F403030403F30403030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000002F001F6F001F6F001F6F00
1F60001F6F001F6000002F00002F000000000000000000001F60001F60001F60001F6000104F00\
1F4F001F4F00000000000000002F000020001F5F00105F001F4F00000000000000
0000000000000000005F4F3050402F50402F4F3F2F4F3F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F302F2F2F302F302F2F30302F3F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5F4F605F4F5F604F605F4F60604F605F4F60605060604F60605060604F60
6050606050606F50606050606F506F6050606F5F6F6F50606F5F6F6F50606F506F6050606F5F6F\
6F101070101F6F5060701F1F701010701F1F7000007F0000702F2F7F0000700000
7F00007F00007F00007F10107F00007F00008000007F00008000007F00008000007F0000800000\
802F2F802F30802F2F80404F8040408040408040408F404F804040804040802F2F
8F3030802F2F8F00008000008F00008F00008F00008F00008F00008F00008F00008F00008F1F1F\
8F1F1F8F1F1F8F5F5F905F5F8F5F5F9070708F706F9070708F70709070708F706F
907F7090706F9070708F706F907F708F706F9070708F706F907F708F706F90706F8F706F90706F\
8F706F90706F8F706F4F3F3F403F304F3F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000003000002F001F6F001F6F00
206F001F6F001F6F001F6F001F6F00001000001F000000001F6F001F60001F60001F60001F5000\
1F4F001F50000000000000000000000000001F5F001F60001F5F000F0F00000000
0000000000000000000000005F4F3050402F5040304F3F2F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F2F302F302F2F302F2F30000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F50405F5F4F605F4F5F5F4F605F4F5F604F605F4F606050605F4F60605060
604F60605060604F606050606050606F506F6050606F506F6F50606F5F6F6050606F5F6F6F5060\
6F50606050606F50606F1010705F6F6F5F60702F2F70202F7F2F307000007F4F5F
704F507F1F1F7010107F101F7F000070606F705F607F3F407F3F408040407F3F408040407F3F40\
801F1F7F1010801F1F801010801F1F801010801F1F801010801F1F801010801F1F
8010108F1F1F803F408F40408040408F40408040408F40407F606F806F6F7F6060806F6F8F6F6F\
8F70708F6F6F8F70708F706F9070708F706F9070708F706F9070708F706F907070
8F706F9070708F706F9070708F706F9070708F706F90706F8F706F90706F8F706F90706F8F706F\
90706F8F706090706F4030304F3F30403030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F6F001F6F00
1F6F001F6F001F60001F6F001F60001F6F00001000001F001F60001F6F001F60001F60001F4F00\
1F5000104F00000000000000000000000000001000105F001F5F00104000000000
00000000000000000000000050402F5040304F402F4F402F4F3F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F2F2F302F302F2F30302F3F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000005F4F605F4F60604F605F4F60605060604F60605060604F6060506F605060
60506F6050606F506F6050606F506F6050606F5F6F6F506F6F5F6F6F506F705F6F6F5F6F705F6F\
6F5F6F70606F705F6F705F6F6F5F6F706070705F6F7F6070705F6F7F505F704F5F
7F505F7F606F7F6F707F607070606F705F6F7F606F7F3F4080404F7F3F4080404F803F4080404F\
80101F801F1F80101F801F1F80101F8F1F1F80101F8F1F1F80101F8F1F1F801F1F
8F1F1F80101F8F404F8F40408F404F8F40408F4F4F8F6F708F70708F6F708F70708F70708F7070\
8F70709070708F70709070708F7070907070907070907070907070907F70907070
907F7090706F907F7090706F907F708F706F907F7090706F907F708F706F907F7090706F907F6F\
8F706F907F6F90706F4F3F30403F304F3F3F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F000000
1F6F001F6F00206F001F6F001F6F001F6F001F6F001F6F001F6F001F60001F6F001F60001F5000\
1F50001F5000000000000000000000000000000000001F00105F001F5F00000000
00000000000000000000000000000050402F504F304F402F4F4030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F2F2F2F302F202F302F302F2030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F5F5F4F605F4F5F5F4F605F4F606050605F4F60605060604F60605060
604F606050606050606F506F6050606F506F6050606F5F6F6F50606F5F6F6F506F705F6F6F5F6F\
705F6F6F5F6F705F6F6F5F6F705F6F705F6F706070705F6F706070705F6F7F6070
70606F7F60707F606F7F60707F606F7F60707F6070806F707F6070806F707F6070806F707F606F\
806F70806F70806F70806F6F806F70806F70806F70806F6F806F70806F708F6F70
806F6F8F6F70806F708F7070806F6F8F70708F6F708F70708F6F6F8F70708F6F708F70708F6F6F\
8F70708F70709070708F706F9070708F70709070708F706F907070907070907F70
8F706F9070708F706F9070708F706F9070708F706F9070708F706F90706F8F706F90706F8F706F\
90706F8F706F90706F4030304F3F30403F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000002F001F6F001F6F001F6F001F6F001F60001F6F001F60001F6F001F60001F6F00104F00\
1F50001F4F000000000000000000000000000000000000001F5F00105F00002000
0000000000000000000000000000000000004F402F50402F4F3F2F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F2F302F2F2F302F302F2F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F5F4F605F4F606050605F4F606050605F4F6060506F60506060506F605060
6F506F6050606F506F6050606F5F6F6F506F6F5F6F6F506F705F6F6F5F6F705F6F6F5F6F705F70\
705F6F706070705F6F706070705F6F706070705F6F7F60707060707F6070706070
7F60707F60707F6F707F6070806F707F6070806F707F6070806F70806F70806F70806F70806F70\
806F70806F70806F708F6F70806F708F6F70806F708F7070806F708F7070806F70
8F70708F6F708F70708F6F708F70708F70708F70708F70709070708F70709070708F7070907070\
8F70709070708F7070907F70907070907070907070907F70907070907F70907070
907F70907070907F70907070907F70907070907F708F706F907F7090706F907F708F706F907F70\
90706F907F6F403F304F3F3F403F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000030001F5000206F001F6F001F6F001F6F001F6F001F6F001F6F001F60001F5000\
1F50000000000000000000000000000000000000000000001F4F001F5F00105F00
002F0000000000000000000000000000000000000050402F504030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F202F2F2F302F2F30000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F5F5F4F605F4F606050605F4F60605060604F6060506F604F6060506F
6050606F506F6050606F506F6050606F5F6F6050606F5F6F6F506F6F5F6F6F5F6F6F5F6F6F5F6F\
705F6F6F5F6F705F6F705F6F706070705F6F706070705F6F7F6070705F6F7F6070
7060707F607070606F7F60707F60707F6F707F60707F6F707F6070806F707F6070806F70806F70\
806F70806F70806F70806F70806F70806F70806F70806F708F7070806F708F6F70
806F708F7070806F708F70708F6F708F70708F6F708F70708F70708F70708F6F708F70708F7070\
9070708F70709070708F70709070708F7070907070907070907F70907070907F70
907070907F708F706F907F708F706F907F708F706F907F708F706F907F708F706F907F6F8F706F\
907F6F8F706F4F3F30403F304F3F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000001F50001F6F001F6F001F6F001F6F001F6F001F60001F6F001F6000\
1F6F00002000002F000000000000000000000000000000001F4F00105F001F5F00
1050000000000000000000000000000000000000000000004F402F4F402F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000302F302F2F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0000006050605F4F6060506F5F4F6060506F60506060506F6050606F506F60506F
6F506F60506F6F5F6F6F506F6F5F6F6F506F6F5F6F6F5F6F6F5F6F6F5F6F705F706F5F6F705F70\
6F5F6F706070705F6F706070705F6F7F60707060707F60707060707F60707F6070
7F60707F60707F6F707F60707F6F707F6070806F7F7F6F70806F7F7F6F70806F7F806F70806F7F\
806F708F707F806F708F6F7F806F708F707F806F708F707F806F708F707F8F6F70
8F70708F6F708F707F8F70708F70708F707090707F8F70709070708F707090707F8F7070907070\
8F7070907F7F907070907F70907070907F7F907070907F709070709F7F7F907070
907F70907070907F70907070907F70907070907F70907070907F7090706F907F7090706F907F70\
90706F907F70403F304F3F3F403F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000001F00000000000000003000000000001F00000000000000000000
0000000000000000000000000000000000002070001F6F001F6F001F6F001F6F001F6F001F6F00\
1F60001F6F001F6000002F001F3F00000000002F000000001F4F001F60001F5F00
1F5F0000000000000000000000000000000000000000000050402F4F402F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F2F2F2F2F2F30000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F5F4F606050605F4F606050605F4F6060506F604F6060506F60506060506F
6050606F506F60506F6F5F6F60506F6F5F6F6F506F6F5F6F6F506F6F5F6F6F5F6F705F706F5F6F\
705F6F6F5F6F706070705F6F706070705F6F706070705F6F7F60707060707F6070
7060707F60707F60707F6F707F60707F6F707F6070806F707F6070806F707F6F70806F70806F70\
806F70806F70806F70806F708F6F70806F708F7070806F708F7070806F708F7070
806F708F707040303F403F3F403030403F3F40303F403F3F403030403F3F40303F4F3F3F403030\
4F3F3F403F3F4F3F3F4030304F3F3F403F3F4F3F3F403F304F3F3F403F3F4F3F3F
403F30907F70907070907F708F706F907F70907070907F708F706F907F708F706F907F708F706F\
907F708F706F4F3F3F403F304F3F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000001F00001000001F00002F000030001F7000001F001F5F00000000
001000001F000000000000000000000000000000000000001F6F001F6F001F6F001F6F001F6000\
1F6F001F60001F6F001060001F3F001F3F001F60000020001F6000105F001F5F00
105F000000000000000000000000000000000000000000000000004F402F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2F2F2F30000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
0F000F0F0F0F5F4F6060506F60506060506F60506060506F60506F60506F60506F
6F5F6F60506F6F5F6F60506F6F5F6F6F5F6F6F5F6F6F5F6F705F706F5F6F705F706F5F6F706070\
705F6F706070705F6F7060707060707060707060707F60707060707F6070706070
7F6F7F7F60707F6F7F7F6070403F3F3F303F40303F3F303F403F3F3F303F403F3F3F303F403F3F\
40303F403F3F40303F403F3F40303F403F3F40303F403F3F40303F403F3F40303F
4F3F3F403F3F4F3F3F40303F4F3F3F403F3F4F3F3F403F3F4F3F3F403F3F4F3F3F403F3F4F3F3F\
403F3F4F3F3F403F3F4F3F3F4F3F3F4F3F3F4F3F3F4F3F3F4F3F3F4F3F3F4F3F3F
4F403F4F3F3F4F3F3F403F3F4F3F3F4F3F3F907F70907070907F70907F70907F7090706F907F70\
907F6F907F70403F304F3F3F403F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001F00\
001F000000001F5F00206000001F00207F001F7000207F001F70002070001F5F00
207000001F00001F00002F000000000000000000000000000000000000001F6F001F6F001F6F00\
1F6F001F6F001F60001F6F001F60001F6F001F60001F60001F60001F6000105F00
1F6000001000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F0000006050605F4F6060506F5F4F6060506F60506060506F60506060506F
60506F6F5F6F60506F6F5F6F60506F6F5F6F6F506F6F5F6F6F5F6F6F5F706F5F6F705F706F5F6F\
7060706F5F6F706070705F6F706070705F6F7060707060707F60707060707F6070
3F30303F303F3F30303F303F3F303F40303F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F3F3F403F304F3F3F403F304F3F3F403F304F3F3F403F304F3F3F8F706F\
907F708F706F4F3F3F403F304F3F3F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001F00\
001F00001000207F001F5F00207F001F70001F70001F70002070001F70001F7000
1F70001F70001F6F00002F00002F001F40000000000000000000000000000000001F6F001F6F00\
1F6F001F60001F6F001F60001F6F001060001F60001060001F6000105F001F5F00
105F00002F00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F202F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000000F0F0F\
00000F0F0F0F5F506060506F60506060506F60506F60506F60506F6F5F6F60506F
6F5F6F60506F6F5F706F5F6F6F5F706F5F6F6F5F706F5F6F705F706F5F6F7060706F5F70706070\
705F707060707060707060707060707F60707060703F303F3F303F3F303F3F303F
3F303F3F303F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004F3F3F403F304F403F4F3F3F4F3F3F\
403F304F403F403F304F3F3F403F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001F00\
001F00001F001F7F00207F001F7000207F001F7000207F001F7000207F001F7000
2070001F70002070001F70002070001F4000204F001F5F00000000000000000000000000000000\
1F6F001F5F001F50001F6F001F60001F6F001F60001F60001F60001F60001F5F00
1F6000002000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
20202F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000060506F5F4F6060506F60506060506F60506060506F60506F6F5F6F
60506F6F5F6F60506F6F5F6F60506F6F5F6F6F5F6F6F5F706F5F6F6F5F706F5F6F7060706F5F6F\
706070705F70706070302F303F303F3F2F303F303F302F303F303F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F3F3F403F304F3F3F403F304F403F4F3F3F4F403F4F3F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000001000\
001F00001F00207F001F70001F7F001F7000207F001F70001F70001F7000207000
1F70001F70001F70001F70001F6F001F70001F6F001F5F001F50001F5F00000000000000000000\
0000000000001F50001F50001F6F001F60001F6F001060001F60001060001F6000
002000002000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F000F0F0F0F5F506060506F60506F60506F60506F605F6F60506F6F5F6F60506F
6F5F706F5F6F6F5F706F5F6F6F5F706F5F6F6F5F706F5F6F7060706F5F70706070302F303F303F\
3F303F3F303F3F2F3F3F303F3F303F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F403F4F3F3F50403F907F704F403F4F3F3F50403F4F403F4F403F4F3F3F
50403F4F403F000000000000000000000000000000000000000000000000000000000000001F00\
001F00207F001F7F00207F001F7F00207F001F7000207F001F7000207F001F7000
207F001F70002070001F70002070001F70001F70001F6F002070001F5F001F5F001F5F00000000\
000000000000000000000000002F00002F001F6000002F001F60001F6000104F00
001F00002000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000060506F5F506060506F5F506F60506F60506F605F6F60506F605F6F
60506F6F5F7060506F6F5F6F6F5F6F6F5F706F5F6F302F3F302F303F303F302F303F303F302F30\
3F303F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F3F3F4F403F4F3F3F907F708F7F6F907F70907F6F907F708F7F6F4F403F
4F3F3F4F403F4F3F304F403F4F3F304F403F4F3F30000000000000000000000000000000000000\
0000001F7000207F001F7000207F001F70001F7F001F7000207F001F70001F7000
1F70001F70001F70001F70001F70001F70001F6F001F70001F6F001F70001F6F001F5F001F5000\
1F5F00000000000000000000000000000000001F00002F001F6000103F00002F00
000000001000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000F0F0F0F5F506F60506F60506F605F6F60506F605F6F60506F6F5F70605F6F
6F5F70605F6F6F5F70302F3030303F302F303F303F302F3F3F303F302F30000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F4F403F80706F8F7F6F8F706F907F70907F6F908070907F6F
907F70907F6F908070907F6F4F403F4F403F50403F4F403F4F403F4F403050403F4F403F000000\
000000207F001F7F00207F001F7F00207F001F7F00207F000030000030001F7000
001F00001F002060001F5F002060001F70002070001F70001F70001F6F002070001F6F001F7000\
1F5F00003000002F00000000000000000000000000000000002F001F4000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000060506F5F506060506F60506F60506F60506F605F6F60506F302F3F
302F30302F3F302F30302F3F302F30302F3F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F3F3F4F403F4F3F3F8F706F80706F8F706F507F405F7F40507F405F7F40
507F40907F70907F6F907F708F7F6F907F6F907F6F907F6F8F7F6F4F403F4F3F304F403F4F3F30\
4F403F1F7F00207F001F70001F7F001F7000207F001F7000003000002F00003000
000000001F000000000000001F5F001F4F001F40001F70001F6F001F70001F6F001F70001F6F00\
1F70001F6F00002F00002F00002F00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F000F0F0F0F00000F60506F60506F302F3F302F30302F3F302F30302F3F302F30
30303F302F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F504040707F5F7F805F707F5F2080001F7F005F804F5F7F40
5F804F5F7F405F80405F7F405F80405F7F40908070907F6F908070907F6F90806F907F6F90806F\
4F403F3F601F3F6010207F001F7F00207F001F7F00207F00000000000000000000
000000000000000000000000000000000000204F00002F000030001F70001F70001F7000207000\
1F70001F70001F6F001F7000002F00003000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F0000000F0F0F2F2030302F302F2030302F30302F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F3F3F4F403F4F403F7F805F707F5F707F5F1F7F002080001F7F00207F00
1F7F002080001F7F005F7F40507F405F8040507F405F7F40507F405F8040507F3F907F6F907F6F\
907F6F8F70603F601F1F7F00207F001F70001F7F001F70000F0F0F000000000000
000000000000000000000000000000000000000000000000002F00003000002F001F70001F6F00\
1F70001F6F001F70001F6F001F70001F6F00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F302F3F302F3F302F30000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F504040707F5F7F80601F7F002080001F7F002080001F7F00
2080001F7F002080001F7F002080001F7F002080001F7F005F80405F7F405F80405F7F405F8040\
507F40907F6F8F7F603F7F1F1F7F00207F001F7F003F601F0000000F0F0F000000
000000000000000000000000000000000000000000000000000000000000003000002F001F4F00\
1F70002070001F70001F70001F6F001F70001F6F00000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
302F3F302F30302F3F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F403F5040404F3F3F7F805F707F5F2080001F7F002080001F7F00208000
1F7F002080001F7F002080001F7F002080001F7F002080001F7F002080001F7F00208000507F3F\
5F7F40507F3F5F7F403F7F103F7F1F1F7F00207F00305F103F6010305F10000000
0000000000000000000000000000000000000000000000000000000000000000000000001F4000\
1F40001F5F001F70001F6F001F70001F6F001F6F00002F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F2F30302F3F\
302F3F5F50605F4F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F7F80607F805F7F80601F7F002080001F7F002080001F7F00
2080001F7F002080001F7F002080001F7F002080001F7F002080001F7F002080001F7F00208000\
1F7F005F8040507F2F5F8030507F2F207F001F7F00207F001F7F004F403F4F3F30
4F403F000000000000000000000000000000000000000000000000000000000000000000000000\
0000001F5F001F5F002070001F70001F70001F6F00003000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000302F3F2F2030302F3F5F4F6F\
5F506F504F605F5060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F403F4F403F707F5F7F8060707F5F3F801F3F7F103F801F3F7F10208000
1F7F002080001F7F002080001F7F002080001F7F002080001F7F002080001F7F002080001F7F00\
2080001F7F00207F001F7F005F7F2F1F7F001F7F001F7F00207F0050703F507F3F
403F304F4030403F30000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F6F001F70001F6F001F70001F6F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F3F2F2F30302F3F5F506F5F506F\
5F506F5F506F5F4F605F506F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F9F807F9080709F807F3F801F40801F3F801F40801F3F801F
40801F3F801F40801F3F801F2080001F80002080002080002080001F80002080001F8000208000\
1F7F002080001F7F002080001F7F002080001F7F00207F001F7F005F7F40507F3F
8F7F608F7F604F403F403F304F4030000000000000000000000000000000000000000000000000\
000000000000000000001F00001F001F70001F70001F70001F4000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2030302F3F2F20305F506F5F4F605F506F5F4F6F\
5F506F504F605F506F5F4F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F403F504040907F709F807F907F709F807F907F70908070907F70908070
907F703F801F3F7F103F801F3F7F103F801F3F7F103F801F1F7F002080001F7F002080001F7F00\
2080001F7F002080001F7F00207F001F7F00207F001F7F001F7F001F7F00207F00
1F70108F7F608070608F7F60403F304F4030403F30000000000000000000000000000000000000\
000000000000000000001000001F001F5F001F70001F6F001F6F001F3F00000000
000000000000000000000000000000000000000000000000000000000000000000000000504F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000302F3F2F2F30302F3F504F605F506F5F4F6F5F506F5F506F605070\
5F506F6050705F4F605F506F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005040404F403F9F807F90807F9F807F9080709F807F9080709F807F908070
9F807F9080709080709080709F80703F801F40801F3F801F40801F3F801F40801F3F8010208000\
1F80002080001F7F002080001F7F002080001F7F002080001F7F00207F001F7F00
207F1F1F7F103F7F2F8F7F608F7F60807F604F4030403F304F4030000000000000000000000000\
0000000000000000000000000000001F5F001F6F001F70001F6F001F6F00000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
504F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F20302F2F302F2030504F6F504F605F4F6F504F605F506F504F6F5F506F5F506F\
5F506F5F506F5F506F5F4F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5040404F403F5040409080709F807F907F709F807F9080709F807F907F70908070
908070908070907F70908070908070908070907F709080709080709080703F7F103F801F3F8010\
3F801F1F7F002080001F7F002080001F7F001F7F001F7F00207F001F7F001F7F00
1F7F00207F1F3070203F7F2F3070208F7F6080705F807F60403F304F4030403F2F000000000000\
000000000000000000000000001F000010001F70001F6F001F7000106F001F4000
000000000000000000000000000000000000000000000000000000000000000000000000504F30\
504F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F20302F2F3F2F2030504F6F504F605F506F504F6F5F506F5F4F6F5F506F5F506F5F5070\
5F506F605F705F506F5F506F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F403F504F405040409F807F90807F9F807F9F807F9F807F9080709F807F908070
9F807F9080709F807F9080709F807F9080709F807F9080709080709080709F8070908070908070\
3F801040801F3F801F3F801F1F7F002080001F7F002080001F7F002080001F7F00
207F001F7F00207F003F7F2F3F7F2F507F3F507F3F807F5F8F7F6080705F4F4030403F304F4030\
0000000000000000000000000000000000001F70001F70001F6F001F70001F4000
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
504F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2030
2020302F20304F4060504F60504F60504F6F504F605F506F504F605F506F5F4F6F5F506F5F506F\
5F506F5F506F5F506F5F4F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5040404F403F50404090807F9F807F90807F9F807F9080709F807F9080709F807F
90807090807F90807090807F90807090807090807090807090807090807090807090807090806F\
90807090806F3F801F3F7F103F801F3F7F102080001F7F001F7F001F7F00207F00
1F7F001F7F001F7F001F7F001F7000507F3F50703F507F3F80705F807F5F80705F807F5F403F2F\
403F30403F2F0000000000000000000000001F5F001F6F001F6F001F6F001F6F00
000000000000000000000000000000000000000000000000000000000000000000000000504F30\
504F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F20302F2F302F2030
504F60504F60504F6F504F605F506F504F6F5F506F504F6F5F506F5F506F5F506F5F506F605F70\
5F506F605F705F506F5F506F5F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504040504F405040409F8F809F807F9F807F9F807F9F8F7F9F807F9F807F908070
9F8F7F9080709F807F9080709F8F7F9080709F807F9080709F8F7F9080709F8070908070908070\
90807090807090806F9080703F801F3F801F3F7F103F801F1F7F002080001F7F00
2080001F7F00207F001F7F00207F001F7F00207F00507F3F507F3F507F3F807F6080705F807F60\
80705F404030403F2F4040300000001F60001F5F001F70001F6F001F7000002F00
001F000000000000000000000000000000000000000000000000000000000000000000005F5030\
504F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F20302020302F20304F405F4F4F60
4F4060504F604F4060504F60504F60504F6F504F605F506F504F6F5F506F504F6F5F506F5F506F\
6050705F506F5F506F5F506F5F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504F404F403F5040409F807F9F807F90807F9F807F90807F9F807F9080709F807F
9080709F807F90807090807F90807090807F908070908070908070908070908070908070908070\
9080708F806F9080708F806F9080708F7F6F90806F3F7F103F801F307F101F7F00
1F7F00207F001F7F001F7F001F7F001F7F001F70001F7F001F7000507F3F4F7030307F1080705F\
807F5F6070406F704F403F2F304F201F5F001F5F001F6F001F7000106F00002F00
002F00000000000000000000000000000000000000000000000000000000000000000000505030\
504F3040402F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002020302F20302020304F40604F4060504F604F4060
504F604F4F60504F6F504F60504F6F504F6F5F506F504F6F5F506F5F506F5F50705F506F605070\
5F506F605F706050705F506F5F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504040504F405040409F8F809F807F9F8F809F807F9F8F7F9F807F9F8F7F9F807F
9F8F7F90807F9F8F7F90807F9F8F7F9080709F8F7F9080709F8F7F9080709F8F7F908070908070\
9080709080709080709080708F806F9080708F806F90806F8F7F6F3F801F3F7F10
3F801F1F7F002080001F7F00207F001F7F00207F001F7F00207F001F7F00207F00307F103F7F1F\
3070106F7F4F6F704F3F7F103040201F70001F5F001F70001F6F001F7000002F00
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
504F2F4F4F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000202030201F2F20203040405F4F406040405F4F40604F4060504F60
4F4060504F604F4F60504F6F504F60504F6F504F6F5F506F504F6F5F506F5F506F5F50705F506F\
5F50705F5070605F705F506F5F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5040404F4040504F409F807F9F807F9F807F9F8F7F90807F9F807F90807F9F8F7F
9080709F807F9080709F8F7F90807090807F9080709F8F7F908070908070908070908070908070\
9080709080709080708F806F9080708F806F90806F8F7F6F8F806F8F7F6F8F806F
307F103F7F1F307F103F7F1F1F7F001F7F001F7F001F7F001F70001F7F001F70001F7F001F7000\
307F101F70001F7F003070103070101F70001F70001F6F001F7000002F00002F00
000000000000000000000000000000000000000000000000000000000000000000000000505030\
504F2F4F402F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001F0000000000000000000000000000000000000000\
0000002020302020304F406040405F4F40604F405F4F4F604F4060504F604F4060
504F6F504F60504F6F504F605F506F504F6F5F506F504F6F5F50705F506F5F50705F506F605F70\
5F5070605F705F50705F50705F506F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504040504F405040409F8F809F807F9F8F809F807F9F8F809F807F9F8F7F9F807F
9F8F7F9F807F9F8F7F90807F9F8F7F90807F9F8F7F9080709F8F7F9080709F8F7F9080709F8F7F\
90807090807090807090807090806F9080708F806F9080708F806F90806F8F7F6F
8F806F8F7F6F3F801F307F103F801F3F7F10207F001F7F00207F001F7F00207F001F7F00207F00\
1F7F001F7F001F7000207F001F70001F70001F70001F70001F7000001F00002F00
0000000000000000000000000000000000000000000000000000000000000000000000005F5F3F\
504F2F4F4F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000001000001F00002F0000000000000000000000000000000000000000\
000000201F2F40405F403F5F4F406040405F4F40604F40604F4F604F40604F4F60
4F4060504F6F4F4F60504F6F504F6050506F504F6F5F506F504F6F5F50705F506F5F50705F506F\
605F705F5070605F705F506F5F5070000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F4040\
504F404F403F9F8F809F807F9F8F8090807F9F8F7F9F807F9F8F7F90807F9F8F7F
90807F9F8F7F9080709F8F7F9080709F8F7F9080709F8F7F9080709F8F7F908070908070908070\
9080709080709080708F806F9080708F806F90806F8F806F90806F8F7F6F8F806F
8F7F608F806F807F608F7F6F307F103F7F1F307F103F7F101F7F001F7F001F70001F7F001F7000\
1F7F001F70001F7F001F70001F7F001F70001F70001F5F101F5F10001000001F00
0000000000000000000000000000000000000000000000000000000000000000000000005F5030\
50503040402F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000001F00001000003000001F0000000000000000000000000000000000\
20202F40405F40405F4F406040405F4F40604F40604F4F604F4060504F6F4F4F60
504F6F4F4F60504F6F504F6F50506F504F6F5F507050506F5F50705F506F5F5F705F5070605F70\
5F5070605F70605F705F50705F506F605F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F40\
504040504F409F807F9F8F809F807F9F8F809F807F9F8F809F807F9F8F809F807F
9F8F7F90807F9F8F7F908F7F9F8F7F90807F9F8F7F908F7F9F8F7F9080709F8F7F908070908F7F\
908070908F709080709080709080709080708F806F9080708F806F9080708F806F
8F806F8F7F6F8F806F8F7F608F806F807F603F801F307F103F7F1F307F10207F001F7F001F7F00\
1F7F00207F001F7F001F7F001F70001F7F004F702060704F1F5F10403F2F000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5F3F\
504F2F4F4F30000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F00003000003000003000003000001F00002F00000000000000000000001F1F2F\
202030403F5F40405F403F5F4F406040405F4F40604F40604F4F604F4060504F6F
4F4060504F6F4F4F60504F6F504F6050506F504F6F5F506F504F6F5F50705F506F5F50705F506F\
5F5F705F5070605F705F50705F50705F506F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F4040\
504F404F40409F8F8090807F9F8F8090807F9F8F8090807F9F8F7F90807F9F8F7F
90807F9F8F7F90807F9F8F7F9080709F8F7F9080709F8F7F908070908F7F908070908F7F908070\
9080709080709080708F806F9080708F806F9080707080407080408F806F8F806F
8F7F6F8F806F8F7F608F806F807F608F7F60807F608F7F60307F103F7F10307F103F7F101F7000\
1F7F001F70001F7F001F70001F70001F70004F7020406F307F70503F3F2F403F2F
000000000000000000000000000000000000000000000000000000000000000000000000505030\
5050304F402F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F00003000003000003F00003000003F00002F00003F0000000000000000201F2F202030\
403F5F40405F403F5F4F40604040604F40604F40604F4F604F4060504F6F4F4F60
504F6F504F6F504F6F504F6F5F506F504F6F5F507050506F5F50705F50705F50705F5070605F70\
5F5070605F705F5F70605F7F5F506F605F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000504F40\
504040504F409F807F9F8F809F8F7F9F8F809F807F9F8F809F8F7F9F8F8090807F
9F8F80908F7F9F8F7F90807F9F8F7F908F7F9F8F7F908F7F9F8F7F908F7F9F8F7F9080709F8F7F\
908070908F703F802F408F309080705F803050802F70804F70804020801F1F8010
708040708040708040507F2F8F806F8F7F608F806F807F608F806F807F603F7F1F307F103F7F1F\
307F101F7F001F7F00207F001F70001F7F001F70004F703F706F4F7F70503F3F2F
403F2F0000000000000000000000000000000000000000000000000000000000000000005F5F3F\
504F304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F000000000000000000
0000003000002F00003000003000003F00003000003F00003F00001F00002F0F1020202F3F3F50\
40405F403F5F40405F40405F4F406040405F4F40604F40604F4F604F40604F4F6F
4F4F60504F6F4F4F60504F6F504F6F50506F504F6F5F506F50506F5F50705F506F5F50705F5070\
605F705F5070605F70605F705F5F705F506F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504F404F4040\
504F4090807F9F8F8090807F9F8F8090807F9F8F8090807F9F8F8090807F9F8F7F
90807F9F8F7F90807F9F8F7F90807F9F8F7F908070908F7F908070908F7F8F7F6F908F7F5F8040\
5F8F4F5080403F80303F802F2080001F80005F802F1F80002080001F80101F801F
1F80002080006F7F3F50802F507F2F3F801F807F608F7F60807F608F7F60807F60807F60307F10\
3F7F103070101F7F001F70001F7F001F70001F70001F7000706F50706F50707050
3F3F2F3F3F2F0000000000000000000000000000000000000000000000000000000000005F5030\
50503040402F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F30000000000000000000
0000003000003000003000003F00003000003F00003F00003F00001F00003010103F3F5040405F\
403F5F40406040405F4F40604040604F4F604F40604F4F604F4060504F6F4F4F60
504F6F4F4F6F50506F504F6F50506F504F6F5F507050506F5F50705F50705F5F705F5070605F70\
5F5070605F7F605F70605F7F5F506F605F705F5F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504F40504F40\
4F40409F8F809F8F809F8F809F8F7F9F8F809F8F7F9F8F80908F7F9F8F80908F7F
9F8F80908F7F9F8F80908F7F9F8F7F908F7F9F8F7F908F7F9F8F7F908F7F8F80708F7F6F208F00\
5F80405F8F4F1F8000208F001F8000208F001F80002080001F80002080001F8000
2080001F80002080001F800050802F307F103F801F1F7F008F806F807F608F8060807F60807F60\
807F603F7F1F307F101F7F001F70001F7F001F70001F70006070407F7050706F50
7070503F3F2F3F3F2F000000000000000000000000000000000000000000000000605F3F5F5F3F\
504F304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F000000000000000000
1000003000003000003000003000003000003000003F00003000003F00003F0F10403F5F3F2F40\
403F5F403F5F404060403F5F4F40604040604F4F604F40604F4F604F4060504F6F
4F4F60504F6F504F6F50506F504F6F50506F504F6F5F507050506F5F50705F50705F5F705F5070\
5F5F705F5070605F7F5F5F70605F7F5F506F605F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000504F404F4040\
504F40908F7F9F8F8090807F9F8F80908F7F9F8F8090807F9F8F80908F7F9F8F80
90807F908F7F908F7F9F8F7F90807F908F7F908F7F908F7F908070908F7F5F80308F80701F8000\
208F001F8000208F001F80002080001F80002080001F80001F80001F8000208000
1F80001F80001F80002080001F7F001F8000307F101F80001F7F001F7F00807F60807F60807F5F\
807F60807F5F807F603070101F7F001F70001F70005F7040607040706F50707050
706F4F706F503F30203F3F2F000000000000000000000000000000000000000000606040505030\
5050304F402F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F30000000000000000000
2F00003000003000003000003F00003000003F00003F00003F00003F00004010103F304040304F\
403F5F40406040203F4F40604040604F4F604F40604F4F6F4F4F60504F6F4F4F60
504F6F504F6F50506F504F6F5F507050506F5F507050506F5F5F705F50705F5F705F5070605F7F\
5F5F70605F7F5F5F70605F7F605F7F605F705F5070000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004F4040504F4F\
9F8F809F8F809F8F809F8F809F8F809F8F80908F7F9F8F80908F7F9F8F80908F7F
9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F7F908F7F9F8F7F5F8F305F8F301F8000208F00\
1F8000208F001F8000208F001F8000208F001F8000208F001F80002080001F8000
2080001F80002080001F80002080001F80002080001F7F002080001F7F002080001F7F008F8060\
807F60807F60807F5F807F60307F101F7F001F7000607F4F6070407F7050706F50
7F7050706F4F706F503F3020000000000000000000000000000000000000000000605F3F5F5F3F\
5050304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F000000000000000000
1F00003000002F00003000003000003F00003000003F00003F00003F00003F00004000003F0F10\
401F2F401F20402F3F40203F4F4060402F3F4F2F3F4F2F3F4F2F404F40604F3F50
4F4F6F504F6F4F4F6F50506F504F6F505070504F6F5F507050506F5F50705F50705F5F705F5070\
605F705F5070605F7F5F5F70605F7F5F506F5F5F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4040504F404F4040\
9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80908F7F908F80
908F7F9F8F80908F7F908F7F908F7F908F7F908F7F908F7F908F7F208F1F1F8010208F001F8000\
208F001F8000208F001F8000208F001F80002080001F80002080001F80001F8000
1F80002080001F80001F80001F80001F80001F7F001F80001F7F001F80001F7F001F7F001F7F00\
1F7F00807F5F807F607F7F5F807F5F3070101F7000607040607040706F50707050
706F4F706F506F6F4F706F4F3030200000000000000000000000000000000000006060405F5030\
5050304F402F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3010101F000000000000
2000002F00002000003000003F00003000003F00003F00003F00003F00004000003F0000400000\
401F204000004000004F2F3F4000004F00004F00004F2F404F000050405F4F3F50
504F6F4F4F6050506F504F6F50507050506F5F50705050705F50705F50705F5F705F50705F5F7F\
5F5F70605F7F5F5F70605F7F605F7F605F705F5F70605F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F504F40504F4F\
9F8F809F8F809F8F809F8F80908F809F8F80908F809F8F80908F7F9F8F80908F7F
9F8F80908F7F9F8F80908F7F9F8F80908F7F9F8F80908F7F908F7F1F8F00208F1F1F8F00208F00\
1F8000208F001F8000208F001F8000208F001F8000208F001F8000208F001F8000
2080001F80002080001F80002080001F80002080001F80002080001F7F002080001F7F001F8000\
1F7F00208000807F60807F607F7F5F807F5F30701060704F7070507F705F707050
707050706F4F7070506F6F4F706F4F3F3020000000000000000000000000000000605F3F5F5F3F\
5050304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F1F101F000000000000
0F00000000001F00003000000F00003000003000003F1F203000003F00003F00004000003F0000\
4000004000004000004000004F00004000004F00004F00004F00004F3F5050101F
4F0F104F4F6F501F2F505070504F6F50507050506F5F507050506F5F50705F50705F5F705F5070\
5F5F705F5F70605F7F5F5F70605F7F605F7F5F5F705F5070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4040504F404F4040\
9F8F80908F809F8F80908F7F9F8F80908F7F9F8F80908F7F908F80908F7F9F8F80
908F7F908F80908F7F908F80908F7F908F7F708F607F8F60908F7F208F001F8F00208F001F8000\
208F001F8000208F001F80002080001F8000208F001F80001F80001F8000708040
1F80001F801F1F801020801F1F7F001F80001F80001F80001F7F001F80001F7F001F80001F7F00\
1F7F001F7F00307F107F705F7F7F5F7F705F7F7F5F5F70407F7050707050707050
706F4F706F506F6F4F706F4F6F604F6F6F4F3030200000000000000000006060406060405F5030\
5050304F4F2F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3010101F000000000000
0F000000000000000F3000001000002F0F103F1F203F10203F10103F00004000003F0000400000\
4000004000004000004F00004000004F00004F00004F00004F00005000004F1010
50000050202F5020305030405F304F50507050506F504F6F5F5F705F50705F5F705F5070605F7F\
5F5F70605F7F5F5F7060607F605F7F60607F5F5F70605F7F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F4F4F409F908F\
9F8F809F8F80908F809F9080908F809F8F80908F809F9080908F809F8F80908F7F
9F9080908F7F9F8F80908F7F9F9080908F7F7F8F6F708F60208F001F8F00208F001F8F00208F00\
1F8F00208F001F8000208F00708040708F4F1F8000708F6070805F8F8F706F8040
708F408F806F20801F50802F50803050802F2080001F80002080001F80002080001F7F00208000\
1F7F001F8000307F103F7F2F7F7F5F7F7F5F7F705F7F7F5F7F70507F705F707050
707050706F507070506F6F4F706F506F6F4F3F3F2F3030200000000000006060405F5F3F5F5F3F\
5050304F4F3040402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F1F1F1F000000000000
0000000000000000000000001F1F202F0F1030304F3F30503F0F103F30403F20304000003F0000\
401010402030402F3F400F104F00004000004F00004F00004F00004F00004F0000
4F00005000004F00005030405000005000004F4F6050506F5050705F50705050705F5F705F5070\
5F5F7F5F5070605F7F5F5F70605F7F5F5F7F5F5F705F5F70605F70000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4040504F40908F80\
9F8F80908F809F8F80908F809F8F80908F7F908F80908F7F9F8F80908F7F908F80
908F7F908F80908F7F908F80908F7F908F80708F60708F601F8F00208F001F8F00208F001F8000\
208F001F8010208F1F1F8010708F4F6F80408F80708F807070805F8F806F8F8070
8F806F8F807080806F8F806F80806F6F80406F7F3F6F80401F7F001F80001F7F001F80001F7F00\
1F80001F7F001F7F00307F20307F2F7F705F7F7F5F7070507F705F707050707050
706F507070506F6F4F6F6F4F6F6F4F6F6F4F6060403F30203030200000005F5F3F5F5F3F5F5F3F\
10500F1F4F0F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3010101F000000000000
0000000000000000001F1F2F1F1F2F30304F3F3F503030503F304F3F3F50402F3F3F0000400000\
400F10402F3F4040604F101F4000004F30404F00004F00004F00005000004F0000
5000004F00005000005000005000005000005F203050202F5F5F705050705F5F7F5F50705F5F7F\
5F5F70605F7F5F5F7F605F7F605F7F60607F605F7F605F7F5F5F70000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F4F4F409F908F\
908F809F908F908F809F9080908F809F9080908F809F9080908F809F9080908F80
9F9080908F7F9F9080908F7F909080708F607F906F708F60208F001F8F00208F001F8F00208F00\
1F8F005F8F301F8F1F908F7F8F8070908F7F8F8070908F708F8070908F708F8070
8F80708F806F8F8F708F806F8F806F80806F8F806F6F80506F805F6F80507080501F8000208000\
1F7F002080001F7F001F7F004F7F3F4F7F3F7F705F7F7F5F7070507F705F707050
707050706F507070506F6F4F706F4F6F6F4F6F6F4F6F604F3F3F2F3030206060405F5F3F5F5F3F\
10500F20500F40402F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F2F2F3F000000000000
0000000000001F1F201F1F2F30304F30305030304F3F3F503030503F3F5F3F2030402F3F3F0000\
40000040000040406040304F4F2F3F402F3F4F00004F00004F00004F0000500000
4F00005000004F000050000050000050000050202F5F20305030405F304F5030405F507050506F\
5F5F7F5F5F70605F7F5F5F70605F7F5F5F7F60607F5F50705F5F70000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4F40504F4F908F80\
9F8F80908F809F8F80908F80908F80908F809F8F80908F7F908F80908F7F909080
908F7F908F80908F7F909080908F7F708F60708F60708F601F8F00208F001F8F00208F001F8000\
208F0050802F5F8F308F80708F8F708F80708F8F708F806F8F80708F806F8F8070
80806F8F806F80806F8F806F80806F8F806F80806F8F806F807F606F80506F7F506F8050607F4F\
1F80001F7F001F7F001F7F001F7F004F70307070507070507F705F707050707050
706F507070506F6F4F706F506F6F4F6F6F4F6060406F6F4F6060403030205F5F3F5F5F3F505030\
1050000F4F00404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F302F2F30000000000000
0000000000001F1F2F30304F3030503030503F3F503030503F3F5F3F3F503F2F3F3F2030400000\
4000004000003F3F5F4F4F604040604F4F604F2F3F4F30404F00005000004F0000
5000005000005000005000005000005000005F00005000005F00005F101F5F101F5050705F5F70\
5F5F70605F7F5F5F7F60607F605F7F60607F605F7F605F7F5F5F70000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F4F4F409F908F\
908F809F908F908F809F908F908F809F9080908F809F9080908F809F9080908F80
9F9080908F80909080908F7F909080708F607F906F1F8F002090001F8F00208F001F8F00208F00\
1F8F005F8F4F508040908F7F8F8070908F708F8070908F708F80708F8F708F806F
8F8F708F806F8F807080806F8F807080806F8F806F80806F8F806F80806F8F806F6F80506F8050\
507F3F1F80001F7F001F80001F7F001F7F00706F507F7F5F70705F7F705F707050
7070507070507070506F6F4F7070506F6F4F6F6F4F6060406F6F4F3030206060405F5F3F5F5F3F\
1050001050001F400F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F2F2F2F3F000000000000
0000001F1F2F2F2F4F30304F30304F30305030304F3F3F503030503F3F5F3F3F503F203F3F0000\
4000003F000040305040304F4F4F604040604F4F6F4F2F3F4F4F6F4F00004F0000
4F00005000004F00005000005000005000005000005000005000005F101F5010105F304F5F3040\
5F3F4F5F5F705F50705F5F7F605F7F5F5F7F60607F5F5F70605F7F5F5F70000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F4F404F4F4F908F80\
9F908F908F80909080908F80909080908F80908F80908F80909080908F7F908F80
908F7F909080908F7F908F80908F7F70906F708F60208F001F8F00208F001F8F00208F001F8F00\
5F8F4F5080408F8F7F8F80708F8F708F80708F8F708F80708F8F708F806F8F8070
8F806F8F807080806F8F806F80806F8F806F80806F80806F80806080806F807F6080806F807F60\
50803F4F7F3F1F7F001F7F001F7F001F70001F7F0070705F7F705F707050707050
706F507070506F6F4F6F6F506F6F4F6F6F4F6060406F6F4F6060400000005F5F3F5F5F3F20500F\
1050000F4F00104F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F302F2F30000000000000
0000002F2F4F30304F30304F30305030304F3F3F503030503F3F5F3F3F5040405F3F203F402F3F\
3F000040000040304F4F2F404040604F4F6F4F4F6F4F4F6F4F4F605030404F2F40
5000004F00005000005000005000005000005F00005000005F00005000005F00005F00005F3F4F\
5F00005F5F705F507060607F5F5F7F606080605F7F6060805F5F7060607F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000504F4F4F4F4F9F908F\
908F809F908F908F809F908F908F809F908F908F80909080908F809F9080908F80
909080908F80909080908F80909080908F7F9090801F8F002090001F8F002090001F8F00208F00\
508F4F5F8F4F8F8F70908F7F8F8070908F7F8F80708F8F708F80708F8F708F8070
8F8F708F806F8F8F708F806F8F807080806F8F806F80806F8F806F80806F8F806F80806080806F\
807F6050802F4F7F2F1F7F001F7F001F7F001F7F004F7F2F70705F7F7F5F707050
707050706F507070506F6F4F6F6F506F6F4F6F6F4F6F6F4F0000000000003030205F5F3F205F10\
105000105000104000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F202F2F2F30000000000000
0000002F304F2F2F4F30304F2F304F3030503030503F3F503030503F3F5F3F3F503F2F3F3F0F10\
4000003F000040000040203F4040604040604F4F6F4040604F4F6F4F2F404F3040
4F4F605000004F00005000004F00005000005000005000005000005F00005000005F0000500000\
5F00005F101F5F101F5F5F7F605F7F5F5F7F60607F5F5F7F5F5F7F5F5F70000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F4F40504F4F908F80\
90908F908F8090908F908F80909080908F80909080908F80909080908F80909080
908F7F909080908F7F9090808F8F7F9090808F8F7F2090001F8F00208F001F8F00208F001F8F00\
3F8F30508F408F8F7F8F80708F8F708F80708F8F708F80708F8F708F80708F8F70
80806F8F807080806F8F807080806F80806F80806F80806F807F6080806F80806080806F7F7F60\
8080607F7F604F7F2F1F7F001F7F001F7F001F7F0010700070705070705070705F
6F6F507070506F6F4F6F70506F6F4F6F6F4F606F4F6F6F4F0000000000002F2F1F5F5F3F105000\
1050000F4F00104F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F302F2F30000000000000
0000002F2F4F30304F2F304F3030503030503F3F503030503F3F5F3F3F5F3F3F5F3F3F5F40101F\
4000004000004000004F2F40402F3F4F4F6F4040604F4F6F4F4F6F4F4F6F4F4F6F
50506F4F30405030405000005000005000005000005000005F00005000005F00005F00005F1F1F\
5F101F5F101F5F101F60607F5F5F7F606080605F7F6060805F5F7060607F5F5F7F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F504F4F4F4F409F908F\
90908F90908F9090809F908F90908090908F908F809F908F909080909080908F80
909080909080909080908F809090809090803F901F3F8F1F2090001F8F002090001F8F00208F00\
3F8F303F8F308F8F70908F7F8F8F708F8F7F8F80708F8F7F8F80708F8F708F8070
8F8F708F80708F8F7080806F8F8F7080806F8F806F80806F8F806F80806F80806F80806080806F\
807F608080607F7F5F1F80001F7F001F7F001F7F001F7F001F70007F7F5F707050
70705F7070507070506F6F4F7070506F6F4F6F6F4F606F4F000000000000303020305F1F105F00\
1050001050000F4000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3F000000000000
00000030304F2F2F4F30304F30304F303050303050303F503030503F3F5F3F3F503F3F5F3F0F10\
4000003F0000400000402F3F402F404040604F4F6F404F6F4F4F6F4F4F6F4F4F6F
4F4F6F5030404F2F4050507050101050101F5000005000005000005F000050202F5F20305F101F\
5F101F5F5F705F5F7F5F5F7F60607F5F5F7F60607F605F7F6060805F5F70605F7F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F4F404F4F4F90908F\
90908F908F8090908F908F8090908F908F80909080908F80909080908F80909080
908F809090808F8F7F909080908F7F9090803F8F1F3F901F1F8F002090001F8F001F8F001F8F00\
208F008080708F8F7F8F80708F8F7F8F80708F8F708F80708F8F7080806F8F8070
80806F8F807080806F8F806F80806F8F806F80806F80806F80806F80806F807F6080806F7F7F60\
8080607F7F607F7F604F7F201F7F001070001F7F001F70001F7F005F7040707050
7070507070506F6F4F6F6F506F6F4F6F6F4F60604F000000000000000000000000305F1F105000\
1050002F4F1F3F4F20000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F301F1F20000000
1F1F202F2F4030305030304F3030503030503F3F503030503F3F5F3F3F503F3F5F3F3F5F40101F\
3F0000400000400000400000402F3F4F3040404F6F4F4F6F4F4F6F4F50704F4F6F
5050704F4F6F50507050507050101F5030405F00005000005F3F4F5030405F2F305F20305F5F7F\
5F5F7F605F7F5F5F7F60607F5F5F7F60608060607F60608060607F60607F5F5F7F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F50504F4F4F4F90908F\
90908F90908F90908F90908F90908090908F90908090908F90908090908F909080
909080909080909080909080909080908F805F9050508F4F2090001F8F002090001F8F00208F00\
1F8F008F8F7F8080708F8F7F8F8F708F8F7F8F8F708F8F708F80708F8F708F8070
8F8F7080806F8F8F7080806F8F8F7080806F8F8F7080806F80806F80806F80806F80806080806F\
7F7F608080607F7F604F7F2F4F7F201F7F001F7F001F7F001F7000607F4F707050
70705F6F70507070506F6F506F70506F6F4F6F6F50000000000000000000000000105F00105F00\
10500040502F3F4F20000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3010101F000000
1010202F304F2F2F4F3030502F304F3030503030503F3F503030503F3F5F3F3F5F3F3F5F3F3F5F\
4000003F0000400000400000402F3F402F3F4F4F6F4040604F4F6F4F4F6F4F4F6F
4F4F6F4F50704F4F6F5050704F2F4050304050507050304F50304050304F5050705F5F7F505070\
5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F5F5F7F6060805F5F7F6060805F5F7060607F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F4F4F4F4F4F90908F\
90908F90908F90908F908F8090908F90908090908F908F80909080908F80909080
8F8F809090808F8F809090808F8F7F909080508F4F5F90501F8F00208F001F8F00208F001F8F00\
1F8F00508F4F8F8F708F80708F8F7F8F80708F8F708080708F8F708080708F8F70
80806F8F807080806F8F807080806F80806F80806F80806F80806F80806F80806080806F7F7F60\
8080607F7F607F7F607F7F5F1F7F001F7F001F7F001070001F7F0010700070705F
6F70507070506F6F506F70506F6F4F6F6F4F606F4F000000000000000000000000105F00105000\
5050304F4F2F404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F3F000000
1F1F202F2F4030304F2F304F303050303050303F503030503F3F5F3F3F5F3F3F5F3F3F5F3F3F5F\
3F00004000004000004000004000004F3F504040604F4F6F4F4F6F4F4F704F4F6F
5050704F4F6F5050704F50705050705050705050705030405F5F7F5050705F5F7F505F705F5F7F\
5F5F7F5F5F7F5F5F7F6060805F5F7F6060805F5F7F60608060608060607F5F5F7F
60607F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F504F4F4F4F909F90\
90908F90908F90908F90908F90908F90908F90908090908F90908090908F909080
90908F9090809090809090809090808F8F7F8F8F801F8F002090001F8F002090001F8F00209000\
1F8F005F8F4F508F4F8F8F7F8F8F708F8F7F8F80708F8F7F8F8F708F8F70808070
8F8F708080708F8F7080806F8F8F7080806F80807080806F80806F80806F80806F7F806080806F\
7F7F608080607F7F607F80604F7F3F1F7F001F7F001F7F001F7000607F4F5F7040
70705F7070507070506F6F506F70506F6F4F6F6F4F000000000000000000000000105F00105F00\
4F50304F5030404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F301F1F20
10102030304F2F2F4030304F2F304F303050303050303F503030503F3F5F303F5F3F3F5F303F50\
4000003F0000400000400000400000403050404F6F40406F4F4F6F404F6F4F4F6F
4F4F6F5050704F4F6F5050704F50705050705050705050705050705F5F7F5050705F5F7F505F70\
5F5F7F5F5F7F5F5F7F5F5F7F60607F5F5F7F6060805F5F7F60608060607F5F5F7F
5F5F7F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F504F4F4F4090908F90908F\
90908F8F908F90908F90908F90908F8F908090908F90908090908F8F8F80909080
8F90809090808F8F809090808F90809090808F8F7F2090001F8F002090001F8F001F8F001F8F00\
208F001F8F00708F4F8F8F708F8F7F8080708F8F708080708F8F708080708F8F70
8080708F8F7080806F80807080806F80807080806F80806F80806F80806F7F7F6080806F7F7F60\
8080607F7F607F7F607F7F5F4F7F3F1070001F7F001F70001F7F001070005F7040
6F70507070506F6F506F70506F6F4F6F6F50000000000000000000000000104F00105F00105000\
4F5030404F2F404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3030303F10101F
3030402F2F4030304F2F304F303050303050303F5F3030503F3F5F303F5F3F3F5F3F3F5F3F2F3F\
3F203F4000004000004000004000004F0000404F6F4F4F6F404F6F4F50704F4F6F
4F50704F4F6F5050704F5070505070505070505F7F5050705F5F7F5050705F5F7F5F5F7F5F5F7F\
5F5F7F5F607F5F5F7F6060805F5F7F60608060607F6060806060806060805F5F7F
60607F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F50504F90908F909F90\
90908F909F9090908F90908F90908F909F8F90908F90908F90908090908F909080
90908F90908090908F9090809090808F908090908070904F7090501F8F002090001F8F00209000\
1F8F00209000708F4F708F4F8F8F708F8F7F8F8F708F8F7F8080708F8F7F808070
8F8F708080708F8F708080708F8F7080806F8F8F7080806F80806F80806F80806F7F806080806F\
7F7F608080607F7F607F7F60307F101F7F001F7F001F7F001F70001F7F006F7050
70705F6F70507070506F70506F70506F6F4F000000000000000000000000001F00105F00405F20\
4F50304F5030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F302F2F3F
2F2F3F2F30402F2F4030304F2F304F303050303050303F5F3030503F3F5F303F5F3F3F5F3F3F5F\
3F10103F00004000003F0000400000400000404F6F4040604F4F6F404F6F4F4F70
4F4F6F4F50704F4F6F5050704F4F70505070505070505070505070505F7F5050705F5F7F505070\
5F5F7F5F5F7F5F5F7F5F5F7F5F607F5F5F7F6060805F5F7F6060806060805F607F
5F5F7060607F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F4F4F4F4F4F909F908F908F\
90908F8F908F90908F8F908F90908F8F908F90908F8F908090908F8F908090908F
8F90809090808F90809090808F8F809090808F8F807090506F8F4F1F90001F8F002090001F8F00\
1F8F001F8F00208F006F8F408F8F7F808F708F8F7F8080708F8F708080708F8F70
80806F808F7080806F808F7080806F80806F80806F80806F7F806F80806F7F7F6080806F7F7F60\
7F7F607F7F607F7F60707F5F307F102F70101F7F001070001F7F001070006F7050
6F6F507070506F70506F70506F6F4F6F6F50000000000000000000000000001000105F000F400F\
4F5030404F2F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F3F2F2F3F
1F1F202F2F4030304F2F304F303050303050303F5F3030503F3F5F303F5F3F3F5F3F3F5F3F4060\
3F0F104000003F0000400000400000402F40402F3F4F4F6F4F4F6F4F4F6F4F4F6F
5050704F4F705050704F50705050705050705050705050705F5F7F5050705F5F7F505F7F5F5F7F\
5F5F7F5F5F7F5F5F7F6060805F5F7F6060805F607F606080606080606080606080
60607F5F5F7F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F0F0F0F0F0F0F0F0F10\
0F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F
0F0F100F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F0F0F10\
0F0F0F0F0F0F0F0F0F0F10100F0F0F0F0F0F0F0F0F0F10100F0F0F0F0F0F0F0F0F
0F10100F0F0F0000000000000000000000000000000000004F504F4F4F4F4F504F8F908F909F90\
90908F909F908F908F909F9090908F909F8F8F908F909F8F90908F90908F8F9080
909F8F90908090908F8F908090908F8F90809090808F90805F903F5090302090001F8F00209000\
1F8F002090001F8F00208F00508F30508F308F8F708F8F7F8F8F708F8F70808070
8F8F708080708F8F7080806F8F8F7080806F808F7080806F80806F80806F80806F7F7F6080806F\
7F7F607F80607F7F60607F3F607F3F1F7F001F70001F7F001F70004F7F2F407020
70705F6F70507070506F70506F7050000000000000000000000000000000105F00105000104F10\
4F50304F5030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F302F2F302F2F3F
1010203030402F2F4030304F2F304F3030502F3050303F5F3030503F3F5F303F5F3F3F5F3F3F5F\
3F10103F00004000003F0000400000402F3F404F6F4040604F4F6F404F6F4F4F6F
4F4F6F4F50704F4F6F5050704F50705050704F5070505070505070505F7F5050705F5F7F505F7F\
5F5F7F505F7F5F5F7F5F5F7F5F60805F5F7F6060805F607F6060805F607F606080
5F5F7F60607F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F0F0F0F0F0F0F00000F0F0F0F0F0F0F0F0F0F00000F0F0F0F0F0F0F\
0F0F0F00000F0F0F0F0F0F0F0F0F0F000F0F80809F80809080809F80809080809F
808090808F9F808090808F9F808F908F8F9F808F908F8F9F808F908F909F808F908F909F8F8F90\
8F909F8F8F908F909F8F909090909F8F909090909F8F9090909F9F0F0F0F0F0F0F
0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F4F4F4F4F504F4F4F4F9090908F908F\
90908F8F908F90908F8F908F90908F8F908090908F8F908090908F8F908090908F
8F908090908F8F90809090808F90809090808F8F7F8F9080508F305F90301F8F001F90001F8F00\
2090001F8F001F8F001F8F00508F307F7F6F8080707F7F6F8F8F70808070808F70
808070808F7080806F80807080806F808F7080806F80806F7F806F80806F7F7F607F806F7F7F60\
7F8060707F607F7F606F705F607F3F1070001F7F001070001F7F0010700070705F
6F70507070506F6F506F70506F6F4F000000000000000000000000000000103000105F00202F1F\
4F5030404F2F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3030303F10101F
0000002F2F4030304F2F304F3030502F3050303F5F3030503F3F5F303F5F3F3F5F3F3F5F3F4060\
3F203F4000003F000040000040000040202F40406F4F4F6F404F6F4F4F704F4F6F
4F50704F4F705050704F50705050704F5070505F7F505070505F7F5050705F5F7F505F7F5F5F7F\
5F5F7F5F60805F5F7F5F60805F5F7F6060805F6080606080606080606F80606080
60607F5F5F7F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F100F0F0F
0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F0F0F107F7F907F7F9F7F7F9080809F7F7F9F80809F\
7F7F9F80809F7F809F80809F7F809F80809F80809F80809F80809F808F9F80809F
808F9F80809F8F8F9F808F9F8F8F9F808F908F909F8F8F9F8F909F8F8F908F909F8F909F8F909F\
8F909090909F8F909F90909F8F9090909F9F8F909F909F9F8F9090909F9F8F9090
909F9F8F9090909F9F8F90900F0F0F0F0F0F0F10100F0F0F4F504F4F4F4F8F8F8F808F808F8F8F\
808F80909F908F908F909F8F8F908F909F8F8F908F909F8F8F908F909F8F8F9080
90908F8F9080909F8F8F90809090808F90809090808F90809090808F8F7F2090001F8F00209000\
1F8F002090001F8F00208F001F8F00208F007F806F8F8F7F808F708F8F7F808F70
8F8F708080708F8F70808070808F7080806F808F7080806F80806F7F806F80806F7F80607F806F\
7F7F607F80607F7F6070705F1F7F001F7F001F7F001F7F001F70001F7F006F7050
70705F6F70507070506F70506F7050000000000000000000000000000000105F00105000103000\
4F4F30000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F301F1F1F
0000002F30402F2F4030304F2F304F3030502F3050303F5F303050303F5F303F5F3F3F5F303F5F\
3F2F3F3F00004000003F0000400000401F2F404F6F404060404F6F404F6F4F4F70
4F4F6F4F50704F4F705050704F4F705050704F507050507F505070505F7F5050705F5F7F505070\
5F5F7F505F7F5F5F7F5F5F7F5F607F5F5F7F6060805F5F7F6060805F6080606080
60608060607F5F5F7F000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F000F0F0F0F0F00000F0F0F0F0F0F0F0F0F0F
00000F0F0F0F7070907F7F907070907F7F90707F907F7F9F7F7F907F7F9F7F7F907F809F7F7F90\
7F809F7F7F9080809F7F7F9080809F7F809080809F7F809080809F808090808F9F
808090808F9F808090808F9F808090808F9F808F908F8F9F808F908F8F9F808F908F909F808F90\
8F909F8F90908F909F8F90908F909F8F909090909F8F90908F909F8F9090909F9F
8F90908F90908F9090909F908F90908F90908F90904F4F4F404040404F4F808F80808F8F808F80\
808F80808F80808F80808F80808F80808F7F808F808F908090908F8F908090908F
8F90809090808F90809090808F90808F90808F90808F90808F8F7F8F90806F8F4F70904F1F8F00\
1F8F001F8F001F8F001F8F001F8F001F8F001F8F003080103F8F1F8080708F8F70
808070808F70808070808F7080806F80807080806F80806F7F806F80806F7F7F607F806F7F7F60\
7F7F60707F607F7F60707F5F1F7F001070001F7F001070001F70005F70306F7050
6F70506F70506F6F506F7050000000000000000000000000000000000000105F00105F00000000\
4F5030000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F3F10101F
0000001F1F2030304F2F304F303050303050303F5F3030503F3F5F303F5F3F3F5F303F5F3F4060\
3F00003F00003F000040000040000040202F4040604F4F6F404F6F4F4F6F4F4F6F
4F50704F4F705050704F507050507F50507050507F505070505F7F50507F5F5F7F505F7F5F5F7F\
5F5F7F5F5F7F5F5F7F5F60805F5F7F6060805F607F606080606080606080606080
606F8F5F607F6060800000000000000000000000000000000000000000000000000000000F0F0F\
0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F707090707F90707090
7F7F907070907F7F90707F907F7F9F7F7F907F7F9F7F7F907F809F7F7F9F7F809F7F7F9F80809F\
7F7F9F80809F7F7F9F80809F7F809F80809F7F809F80809F80809F808F9F80809F
808F9F80809F808F9F808F9F8F8F9F808F9F8F8F9F808F908F909F8F8F9F8F909F8F8F908F909F\
8F909F8F909F8F909090909F8F909F90909F8F9090909F9F8F9090909F9F8F9090
909F9F8F9090909F908F9090909F908F9090909F90404F404F4F4F000000000000000000000000\
0000000000000000008F8F80808F808F8F80808F808F8F80808F808F8F80808F7F
8F8F80808F7F90908F8F908090908F8F90809090808F90809090808F90808F90806F8F4F209000\
1F8F002090001F8F002090001F8F00208F001F8F003F8F1F308F1F508F3050802F
8F8F7F808F70808F70808070808F70808070808F707F806F80806F7F806F80806F7F7F607F806F\
7F7F607F8060707F601F7F001F7F001F7F001F70001F7F001F70005F7F3F5F7030
70705F6F70506F7050000000000000000000000000000000000000000000002F00105000000000\
202F1F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F301F1F1F
0000000000002F2F402F304F2F304F3030502F3050303F5F303050303F5F303F5F3F3F5F300F10\
3F10103F00003F00003F00004000003F4060404F6F404060404F6F404F6F4F4F6F
4F4F6F4F50704F4F6F4F50704F50705050704F5070505070505070505F7F505070505F7F505F7F\
5F5F7F505F7F5F5F7F5F5F7F5F60805F5F7F5F60805F607F6060805F6080606080
606080606F805F5F7F60607F0000000000000000000F0F0F00000F0F0F0F00000F0F0F0F00000F\
0F0F0F00000F0F0F0F6F708F7070906F708F70709070708F70709070708F707090
707090707F90707090707F907070907F7F90707F907F7F9F7F7F907F7F9F7F7F907F7F9F7F7F90\
7F809F7F7F907F809F7F7F9080809F7F7F9080809F7F809080809F7F809080809F
808090808F9F808090808F9F808090808F9F808F90808F9F808F908F8F9F808F908F909F808F90\
8F909F808F908F909F8F90908F909F8F90908F909F8F90908F909F8F90908F9090
8F90908F90908F90908F90908F90908F90908F908F404F4F404040000000000000000000000000\
00000000000000000000000000000000000000000000000000000080807F808F80
808F7F808F7F80807F808F7F8F90808F90808F8F808F90808F8F7F8F90808F8F7F808F7F7F8070\
1F90001F8F001F8F001F8F001F8F001F8F001F8F001F8F001F8F004F802F508F30
4F8040508F4F80806F808F7080806F808F707F806F80806F7F806F7F806F7F7F607F806F7F7F60\
7F80605F7F4F7F7F601F7F001F7F001070001F7F00107000407F3F4070306F705F
6F70506F70506F6F500000000000000000000000000000000000000000000F50001050000F4000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030304F2F304F3030502F3050303F5F303050303F5F303F5F3F3F5F300F103F101F\
3F00003F00003F00004000003F2F3F402F4040406F404F6F404F6F4F4F704F4F6F
4F50704F4F6F5050704F50705050704F5070505F7F505070505F7F5050705F5F7F505F7F5F5F7F\
505F7F5F60805F5F7F5F60805F5F7F6060805F60806060805F6080606F80606080
606F8F6060806060805F607F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F6F6F8F6F708F\
6F6F8F7070906F708F7070906F708F707090707090707090707090707F90707090
707F907070907F7F9F707F907F7F9F707F907F7F9F7F7F907F7F9F7F7F907F809F7F7F9F7F809F\
7F7F9F80809F7F809F80809F7F809F80809F7F809F80809F7F809F808F9F80809F
808F9F80809F808F9F808F9F808F9F808F9F8F8F9F808F9F8F8F9F808F908F909F8F8F9F8F909F\
8F8F908F909F8F909F8F909F8F90908F9F9F8F909F8F909F8F9090909F9F8F9090
8F9F9F8F9090909F9F8F9090909F908F90904F4F4F404F404F4F4F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000808F80808F7F808F7F80807F808F7F8F90808F90808F907F8F9080808070808F7F\
508F3050903F1F8F002090001F8F002090001F8F00208F001F8F00208F001F8F00
1F8F004F8040508F4F508040808F7080806F808F707F806F80806F7F806F80806F7F80607F806F\
7F7F606080501F7F001F7F001F7F001F7F001F7F001F7F002F70104F7F3F6F7050
6F705F6F7050000000000000000000000000000000000000000000000000105F00105000001000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000101F2030304F2F2F4F3030502F304F303F5F2F304F303050300F103F1010300000\
3F00003F00003F00003F0F104010103F4060404F6F404060404F6F404F6F4F4F70
404F6F4F50704F4F704F50704F50705050704F5070505F7F4F5070505F7F505070505F7F50507F\
5F5F7F505F7F5F5F80505F7F5F60805F5F7F5F60805F5F7F6060805F6080606080
5F6080606080505F705F5F7F505F700F0F0F606F806F6F8F606F806F6F8F6F6F8F6F708F6F6F8F\
6F708F6F6F8F6F70906F6F8F7070906F708F7070906F708F70709070708F707F90
707090707F90707090707F907070907F7F90707F907F7F9F707F907F7F9F7F7F907F809F7F7F90\
7F809F7F7F907F809F7F7F907F809F7F809080809F7F809080809F7F809080809F
7F809080809F808090808F9F808090808F9F808F90808F9F808F90808F9F808F908F909F808F90\
8F909F808F908F909F808F908F909F8F90908F909F8F90908F909F8F90908F909F
8F90908F90908F90908F90908F90908F9090404040404F4F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000808F7F80807F808F7F7F80708F9080808F7F8F9080808F7F\
8F907F508F301F90001F8F001F8F001F8F001F8F001F8F001F8F001F8F001F8F00
1F80001F8F001F8000508F404F80404F80407F806F80806F7F806F7F806F7F7F607F806F707F60\
4F7F3F407F3F1F7F00107F001F7F001070001F7F002F70102F70106F70506F705F
6F70506F7050000000000000000000000000000000000000000000000000103000105000000F00\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000001F1F202F2F403030502F3050301F2030000030101F3000003F10103000003F0000\
3F00003F00003F00004000003F4060404F6F404060404F6F404F6F4F4F6F404F6F
4F50704F4F704F50704F507050507F4F5070505F7F505070505F7F505F7F505F7F505F7F5F5F7F\
505F7F5F5F805F5F7F5F60805F5F805F60805F60806060805F60806060805F6080
5F60805F5F7F606080505F700F0F0F606F8F6F6F8F606F8F6F708F6F6F8F6F708F6F6F8F6F7090\
6F708F6F70906F708F7070906F708F7070906F708F707F90707090707F90707090
707F90707090707F90707F907F7F9F707F907F7F9F707F907F809F7F7F907F809F7F7F9F7F809F\
7F7F9F7F809F7F7F9F80809F7F809F80809F7F809F80809F7F809F80809F7F809F
808F9F80809F808F9F80809F808F9F808F9F808F9F808F9F8F909F808F9F8F909F808F9F8F909F\
808F9F8F909F8F90908F909F8F909F8F909F8F90908F9F9F8F90908F909F8F9090
8F9F9F8F90908F9F908F90908F9F90404F4F404F4F404F40000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000808F7F80807F808F7F7F80708F9080\
8F8F7F8F9080808F7F2090001F8F002090001F8F00208F001F8F001F8F001F8F00
208F001F8F001F8F001F8000208F0030802F308F307F806F6F805F7F806F7F806F307F107F7F60\
1F7F001F80001F7F001F7F001F7F001F7F001F7000307F1F2F7010707F5F6F7050
6F705F000000000000000000000000000000002F00104F00000000000000103F00105000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002F304F2F2F4F3030502F0F0F300000300000300000300000300000300000\
3F00003000003F00003F40604040603F4060404F6F404060404F6F404F6F4F4F6F
404F6F4F50704F4F6F4F50704F507050507F4F507050507F505070505F7F505070505F7F505F7F\
5F5F7F505F7F5F5F7F5F5F7F5F60805F5F7F5F60805F6080606080505F705F5F7F
5F5F7F5F607F0000000F0F0F00000F606F8F606F806F6F8F606F8F6F6F8F606F8F6F6F8F6F6F8F\
6F708F6F6F8F6F708F6F6F8F7070906F708F7070906F708F7070906F708F707090
707090707F90707090707F907070907F7F90707F907F7F90707F907F7F9F707F907F7F9F7F7F90\
7F809F7F7F907F809F7F7F907F809F7F7F907F809F7F809080809F7F809080809F
7F8090808F9F7F8090808F9F808090808F9F808090808F9F808F90808F9F808F90808F9F808F90\
8F909F808F908F909F8090908F909F8090908F909F8F90908F909F8090908F9090
8F90908F90908F90908F90908F9090404F4F404040404F4F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007F807080807F7F8070\
808070808F7F808F7F808F7F8F8F7F808F701F8F001F8F001F8F001F8F001F8F00
1F80001F8F001F80001F8F001F80001F8F001F80101F801F60805030801F707F6030801F1F7F00\
1F8000107F001F7F00107F001F7F001070002F7F102F7010707F5F6F70506F705F
6F70500000000000000000000000000000000020001050000000000000000F4000001000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000001F1F2030304F2F304F303F50302030301010300F103F00003000003F0000\
300F103F101F3F40603F40603F4060404F6F40406F404F6F404F6F4F4F70404F6F
4F50704F4F6F4F50704F507050507F4F5070505F7F505070505F7F50507F505F7F505F7F5F5F7F\
505F7F5F60805F5F7F5F60805F5F7F5F6080505F7F5F5F7F505F7F5F607F000000
0000000000000000000000000F0F0F0F0F0F6F6F8F606F8F6F6F8F606F8F6F708F6F6F8F6F708F\
6F6F8F6F70906F708F7070906F708F7070906F70907070906F7090707F90707090
707F90707090707F90707F907F7F9F707F907F7F9F707F907F7F9F707F907F809F7F7F9F7F809F\
7F7F9F7F809F7F809F7F809F7F809F80809F7F809F80809F7F809F808F9F7F809F
808F9F7F809F808F9F80809F808F9F808F9F808F9F808F9F808F9F808F9F8F909F808F9F8F909F\
808F9F8F909F80909F8F909F8F90908F909F8F909F8F909F8F90908F909F8F9090
8F909F8F90908F9F9F8F90908F9090404F404F4F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
7F8070808F7F7F80707F8070808F7F8F907F808F7F808F7F1F8F00208F001F8F00
1F8F001F8F00208F001F8F001F8F001F80001F8F001F80001F80003080101F80001F80001F8000\
1F7F001F80001F7F001F7F001F7F001F7F002F7F10707F5F6F705F707F5F6F7050
000000000000000000000000000000002F001F6F00104F00000000000000000000001000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000001F1F2F2F2F4F30304F2F304F3020302F0F10300F10301020301F2F302F4F\
3F304F303F5F3F40603F3F603F40603F4060404F6F3F4060404F6F404F6F4F4F6F
404F6F4F4F704F4F704F50704F50704F50704F5070505F7F4F5070505F7F50507F505F7F505F7F\
505F7F505F7F5F5F80505F7F505F7F505F70505F7F505F70000000000000000000
00000000000000000000000000000F0F0F0F606F806F6F8F606F806F6F8F606F8F6F708F606F8F\
6F708F6F6F8F6F708F6F6F8F6F708F6F708F7070906F708F7070906F708F707F90
6F708F707F90707090707F90707090707F90707F907F7F9F707F907F7F9F707F907F809F707F90\
7F809F7F7F907F809F7F7F907F809F7F80907F809F7F80907F809F7F809080809F
7F809080809F7F8090808F9F7F8090808F9F808090808F9F808F90808F9F808F90808F9F808F90\
80909F808F908F909F808F908F909F8090908F909F8090908F909F8090908F909F
8090908F90908090908F9090809090404F4F404040000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000007F80707F80707F80707F8070808F7F808F70808F7F808F70808F7F
1F80001F8F001F80001F8F001F80001F80001F80001F80001F80001F80001F7F001F8000107F00\
1F7F00107F001F7F00107F001F7F00407030307F106F705F6F705F6F70506F705F
000000000000000000000000000000002F00106000106000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000030304F2F304F3030502F3050303F5F303050303F5F303F5F3F4060\
303F5F3F40603F3F603F40603F4060404F6F3F4060404F6F404F6F404F6F404F6F
4F50704F4F704F50704F507050507F4F5070505F7F4F5070505F7F505F7F505F7F505F7F5F5F80\
505F7F5F5F80505070505F7F505F70505F7F000000000000000000000000000000
0000000000000000000000000F0F0F0F0F0F0F0F0F606F8F6F708F606F8F6F708F606F8F6F7090\
6F6F8F6F70906F708F7070906F708F7070906F708F707F906F7090707F906F7090
707F90707090707F90707F907F7F9F707F907F7F9F707F907F809F707F907F809F7F7F907F809F\
7F7F9F7F809F7F7F9F7F809F7F809F7F809F7F809F80809F7F809F80809F7F809F
808F9F7F809F808F9F7F809F808F9F808F9F808F9F808F9F808F9F808F9F808F9F808F9F8F909F\
808F9F8F909F808F9F8F909F80909F8F909F8090908F909F8F90908F909F809090
8F909F8F90908F90908F90908F9090404F4F404F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000007F80707F80707F80707F80707F806F808F7F808F70
808F7F808F701F8F001F80001F8F001F80001F8F001F80001F80001F80001F80001F7F001F8000\
1F7F001F80001F7F001F7F005F7F3F407F3F6F705F707F5F6F705F6F7F5F000000
000000000000000000000000001F00106000106000106000105000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000101F2030304F2F304F3030502F3050303F5F303050303F5F303F5F\
3F4060303F5F3F40603F40603F40603F406040406F3F4060404F6F40406F404F6F
404F6F4F5070404F6F4F50704F50704F507F4F507050507F4F5070505F7F4F5070505F7F505F7F\
5050704F5070505F70505070000000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F00000F606F8F606F8F6F6F8F606F8F6F6F8F606F8F\
6F708F606F8F6F708F6F6F8F6F70906F708F6F70906F708F7070906F708F707090
6F7090707F906F7090707F90707090707F90707F90707F90707F907F7F9F707F907F7F9F707F90\
7F809F707F907F809F7F7F907F809F7F7F907F809F7F80907F809F7F80907F809F
7F809080809F7F809080809F7F8090808F9F7F8090808F9F808F90808F9F808F90808F9F808F90\
80909F808F9080909F808F908F909F808F908F909F8090908F909F808F908F9090
8090908F90908090908F9090404040404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007F806F7F8070707F6F7F806F
7F8070808F707F80701F8F001F80001F8F001F80001F80001F80001F8000107F001F80001F7F00\
1F7F00107F001F7F00107F005F7F3F5070306F7F5F6F705F6F7F5F6F7050000000
000000000000000000000000002F00106000106000106000104F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F304F303050303050303F50303050303F5F303F5F3F3F60\
303F5F3F40603F40603F40603F4060404F6F3F4060404F6F40406F404F6F404F6F
4F5070404F6F4F50704F50704F50704F5070505F7F4F507F505F7F50507F505F7F4F5070505070\
4F5070505F7F000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F0F0F0F606F8F6F6F8F606F8F6F708F606F8F6F708F\
6F6F8F6F70906F708F6F70906F708F7070906F70907070906F7090707F906F7090
707F90707090707F90707F90707F90707F907F7F9F707F907F7F9F707F907F809F7F7F9F7F809F\
7F7F9F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F809F808F9F7F809F
808F9F7F809F808F9F7F809F808F9F7F809F808F9F808F9F808F9F808F9F808F9F808F9F80909F\
808F9F8F909F808F9F8F909F80909F8F909F80909F8F909F8090908F909F809090
8F909F8090908F909F8F9090404F4F404040000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000007F80707F806F
7F8070707F6F808F707F80701F8F001F80001F8F001F80001F80001F80001F80001F7F001F8000\
2F7F101F8000407F3F707F606F7F5F707F606F705F6F7F5F6F705F000000000000
000000000000000000000000000000002F00106000106000104F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000001F1F2F2F304F3030502F3050303F50303050303F5F303F5F\
303F5F303F5F3F4060303F603F40603F406040406F3F4060404F6F404F6F404F6F
404F6F4F4F70404F6F4F50704F4F704F50704F507050507F404F6F4F50704F50704F5070000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000F0F0F0F606F80606F8F606F8F6F708F606F8F\
6F708F606F8F6F7090606F8F6F70906F708F6F70906F708F6F70906F708F707F90
6F708F707F906F7090707F906F7090707F90707F90707F9F707F90707F9F707F907F809F707F90\
7F809F707F907F809F707F907F809F7F80907F809F7F80907F809F7F80907F809F
7F80907F809F7F8090808F9F7F8090808F9F7F8090808F9F7F8090808F9F808F90808F9F808F90\
808F9F808F9080909F808F9080909F808F908F909F808F9080909F8090908F909F
808F908F90908090908F90903F4040404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000007F806F707F6F7F8F707F806F7F80701F80001F801F107F001F80001F7F001F801F107F10\
307F106F705F407F3F6F705F707F606F705F6F7F5F6F705F6F705F000000000000
000000000000000000000000000000002F00105F001060000F5F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001F1F2F2F304F303F50303050303F5F303F5F3F4060\
303F5F3F4060303F603F40603F40603F406F3F4060404F6F404F6F404F6F404F6F
4F5070404F704F50704F4F704F507F4F507050507F4F4F6F4F50704F5070000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F0F0F0F6F6F8F606F8F6F708F606F8F6F7090\
6F6F8F6F70906F708F6F70906F708F6F70906F708F707F906F7090707F906F7090
707F90707090707F90707F90707F9F707F90707F9F707F907F809F707F907F809F707F907F809F\
7F7F9F7F809F7F7F9F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F809F
808F9F7F809F808F9F7F809F808F9F7F809F808F9F7F8F9F808F9F808F9F808F9F808F9F80909F\
808F9F80909F808F9F8F909F80909F8F909F8090908F909F8090908F909F809090
8F909F8090908F9090809090404F4F404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000007F806F707F6F808F707F806F7F80701F801060804030801030801F707F60707F60\
6F705F707F60707F60707F606F7F5F707F606F705F6F7F5F000000000000000000
000000000000000000000000000000002F00106000105F00106000104F00000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000101F2030304F2F304F303F502F3050303F5F303F5F\
303F60303F5F3F4060303F603F40603F40603F406F3F4060404F6F3F406F404F6F
404F6F4F4F70404F6F4F50704F4F704F5070404F6F4F5070000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F00000F606F8F606F8F606F8F606F8F\
6F708F606F8F6F708F606F8F6F70906F708F6F70906F708F6F70906F708F6F7090
6F7090707F906F7090707F906F7090707F90707F90707F90707F90707F9F707F90707F9F707F90\
7F809F707F907F809F707F907F809F707F907F809F7F80907F809F7F80907F809F
7F80907F809F7F80907F809F7F8090808F9F7F8090808F9F7F8090808F9F7F8F90808F9F808F90\
808F9F808F9080909F808F9080909F808F9080909F808F9080909F808F90809090
808F90809090808F90809090404040404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000707F6F707F60707F6F70806F7F806F5F803F7F806F2F7F10707F60707F60\
6F70606F7F60707F606F7F5F707F606F705F6F7F5F000000000000000000000000
000000000000000000000000000000002F00105F00105F00105F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F2F303050303050303F5F303F5F303F5F\
303F5F3F4060303F603F40603F40603F4F6F3F4060404F6F3F406F404F6F404F6F
404F70404F6F4F50704F50704F5070404F6F4F5070000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F0F0F0F606F8F6F708F606F8F6F708F\
606F8F6F70906F708F6F70906F708F6F70906F70906F70906F7090707F906F7090
707F906F7090707F90707F90707F90707F90707F9F707F90707F9F707F907F809F707F9F7F809F\
707F9F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F8F9F7F809F
7F8F9F7F809F808F9F7F809F808F9F7F809F808F9F7F8F9F808F9F7F8F9F808F9F808F9F808F9F\
808F9F80909F808F9F80909F808F9F8F909F80909F80909F808F908F909F809090
8F909F8090908F909F809090404F4F404040000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000707F6F707F60707F6F70806F7F806F70806F7F806F707F6070806F\
707F60708060707F60707F606F7F5F707F606F705F000000000000000000000000
000000000000000000000000000000000000002F00105F00105F00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F1F2F2F304F303050303050303F5F303F50\
303F5F303F5F3F4060303F5F3F40603F40603F406F3F4060404F6F3F406F404F6F
404F6F404F70404F6F4F5070404F6F404F6F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0F00000F606F8F606F8F60708F606F8F\
6F708F606F8F6F7090606F8F6F70906F708F6F70906F708F6F70906F708F6F7F90
6F708F707F906F7090707F906F7090707F90707F90707F9F707F90707F9F707F9070809F707F90\
7F809F707F907F809F707F907F809F7080907F809F7080907F809F7F80907F809F
7F80907F809F7F80907F8F9F7F80907F8F9F7F8090808F9F7F8090808F9F7F8F90808F9F7F8F90\
808F9F7F8F9080909F808F9080909F808F9080909F808F90809090808F90809090
808F90809090808F908090903F4040404F4F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F7F60707F60707F6070806F707F6070806F6F7F60\
707F606F7F60707F606F705F6F7F606F705F000000000000000000000000000000
000000000000000000000000000000000000002000105F000F5000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000001F202F303050303F50303F50303F5F\
303F5F3F4060303F5F3F40603F40603F406F3F4060404F6F3F406F404F6F404F6F
404F70404F6F4F5070404F6F404F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F0F0F0F60708F606F8F6F7090\
606F8F6F7090606F8F6F70906F708F6F70906F708F6F7F906F70906F7F906F7090
707F906F7090707F906F7090707F9F707F90707F9F707F9070809F707F9070809F707F907F809F\
707F9F7F809F707F9F7F809F7F809F7F809F7F809F7F809F7F809F7F809F7F809F
7F8F9F7F809F7F8F9F7F809F808F9F7F809F808F9F7F8F9F808F9F7F8F9F808F9F7F8F9F808F9F\
808F9F80909F808F9F80909F808F9F80909F808F9080909F80909080909F808F90
80909F809090809090809090404F4F40404F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F7F60707F6070806F70806F707F6070806F\
707F60707F606F7F60707F606F7F5F6F7F60000000000000000000000000000000
000000000000000000000000000000000000002F00105F00000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000001F1F203030502F3050303F50303F50\
303F5F303F5F304060303F603F40603040603F40603F4060404F6F3F406F404F6F
404F6F404F703F4060404F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000F0F0F0F606F8F606F8F606F8F\
6F708F606F8F6F708F606F8F6F7090606F8F6F70906F708F6F70906F708F6F7090
6F7090707F906F7090707F906F7090707F906F7090707F90707F90707F9F707F90707F9F707F90\
7F809F707F907F809F707F907F809F707F907F809F7080907F809F7080907F809F
7F80907F809F7F80907F809F7F80907F8F9F7F80907F8F9F7F8090808F9F7F8090808F9F7F8F90\
808F9F7F8F90808F9F7F8F9080909F7F8F9080909F808F9080909F7F8F90809090
808F90809090808F908090903F4040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006F70606F7F606F7F60707F606F7F60\
707F606F7F606F7F606F7F5F6F7F60000000000000000000000000000000000000
000000000000000000000000000000000000002000105F00000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000001F1F2F303F50303F50303F5F\
303F5F3F4060303F603F40603040603F406F3F40603F4F6F3F406F404F6F404F6F
404F703F4060404F6F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F0F0F0F606F8F6F708F\
606F8F6F709060708F6F709060708F6F70906F70906F70906F70906F7F906F7090
707F906F7090707F906F7F90707F906F7F90707F9F707F90707F9F707F9070809F707F9F70809F\
707F9F7F809F70809F7F809F70809F7F809F7F809F7F809F7F809F7F8F9F7F809F
7F8F9F7F809F7F8F9F7F809F7F8F9F7F809F808F9F7F8F9F808F9F7F8F9F808F9F7F8F9F808F9F\
7F8F9F80909F808F9F80909F808F9F80909F808F9080909F808F9080909F808F90
80909F808F9080909F40404F404F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F70606F7F60707F6070806F\
6F7F607080606F7F606F7F606F7F5F000000000000000000000000000000000000
000000000000000000000000000000000000002F00000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000001F1F2F303F50303050\
303F5F303F5F304060303F5F3F40603040603F406F3F40603F4F6F3F406F404F6F
3F40603F4060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F00000F606F8F606F8F\
60708F606F8F6F7090606F8F6F709060708F6F709060708F6F70906F708F6F7F90
6F708F6F7F906F7090707F906F7090707F906F7F90707F9F6F7F90707F9F707F9070809F707F90\
70809F707F907F809F707F907F809F7080907F809F7080907F809F7080907F809F
7080907F809F7F80907F8F9F7F80907F8F9F7F80907F8F9F7F80907F8F9F7F8F90808F9F7F8F90\
808F9F7F8F90808F9F7F8F90808F9F7F8F9080909F7F8F908090907F8F90809090
7F8F90809090808F90404F4F3F4040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F705F6F706060705F\
6F7F606F7F606F7F606F705F000000000000000000000000000000000000000000
0000000000000000000000000000000000000F5000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000001F202F303050303F5F\
303F5F3F3F5F303F5F3F40603040603F40603F40603F4F6F3F406F404F6F3F4060
404F6F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000F0F0F0F0F0F606F8F6F7090\
606F8F6F7090606F8F6F709060708F6F709060708F6F7F906F70906F7F906F7090
707F906F7090707F906F7090707F9F6F7F90707F9F6F7F9070809F707F9070809F707F907F809F\
707F9F7F809F707F9F7F809F70809F7F809F70809F7F809F7F809F7F809F7F809F
7F8F9F7F809F7F8F9F7F809F7F8F9F7F8F9F7F8F9F7F8F9F808F9F7F8F9F808F9F7F8F9F808F9F\
7F8F9F808F9F7F8F9F80909F808F9F80909F7F8F9080909F808F9080909F808F90
80909F808F908090903F4040404F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000060705F6F7F60\
60705F6F7F606F7F606F7F60000000000000000000000000000000000000000000
000000000000000000000000000000000000105000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000001F1F2F303050\
303F5F303F5F303F5F303F5F3F4060303F603F40603F40603F4F6F30405F3F4060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000F0F0F0F606F8F\
60708F606F8F60708F606F8F6F7090606F8F6F709060708F6F709060708F6F7090
6F70906F7F906F70906F7F906F7090707F906F7090707F906F7F90707F9F6F7F90707F9F707F90\
70809F707F9070809F707F907F809F707F907F809F7080907F809F7080907F809F
7080907F809F7080907F809F7F80907F8F9F7F80907F8F9F7F80907F8F9F7F80907F8F9F7F8F90\
808F9F7F8F90808F9F7F8F90808F9F7F8F90808F9F7F8F9080909F7F8F90808F90
7F8F908090907F8F9040404F3F4040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
60705F606F5F6F7F60000000000000000000000000000000000000000000000000
000000000000000000000000000000002000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000303F50\
303F503F3F5F303F5F3F4060303F603F406F3F40603F406F30405F3F4060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F0F0F0F60708F\
606F8F6F709060708F6F709060708F6F70906F70906F70906F70906F7F906F7090
6F7F906F7090707F906F7F90707F906F7F90707F9F6F7F90707F9F707F9070809F707F9F70809F\
707F907F809F70809F7F809F70809F7F809F70809F7F809F70809F7F809F70809F
7F809F70809F7F8F9F7F809F7F8F9F7F809F7F8F9F7F8F9F7F8F9F7F8F9F808F9F7F8F9F808F9F\
7F8F9080909F7F8F9F80909F7F8F9080909F7F8F9080909F7F8F9080909F7F8F90
80909F7F8F9080909F3F404F404F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F705F0F0F0F000000000000000000000000000000000000000000000000
0000000000000000000000000000000F5000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000001F1F2F\
303F50303F50303F5F303F5F304060303F603F4060303F5F3F4060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F0F00000F\
606F8F606F8F607090606F8F60709060708F6F709060708F6F709060708F6F7090
60708F6F7F906F70906F7F906F70906F7F906F7F90707F906F7F90707F906F7F90707F9F6F7F90\
707F9F707F9070809F707F9070809F707F9070809F707F9070809F7080907F809F
7080907F809F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F7F80907F8F9F7F8090\
7F8F9F7F8F907F8F9F7F8F907F8F9F7F8F90808F9F7F8F907F8F907F8F90808F90
7F8F90808F907F8F90404F4F3F4040000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000060705F0F0F0F000000000000000000000000000000000000000000000000
0000000000000000000000000000000F4F00000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
1F1F2F303F5F303F5F3F4060303F603F4060303F5F3F4060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000F0F0F0F0F\
606F8F607090606F8F6F709060708F6F709060708F6F70906F70906F7F906F7090
6F7F906F70906F7F906F7090707F9F6F7F90707F906F7F90707F9F6F7F90707F9F6F7F9070809F\
707F9070809F707F9070809F70809F70809F7080907F809F70809F7F809F708090
7F8F9F70809F7F8F9F7080907F8F9F7F809F7F8F9F7F80907F8F9F7F8F9F7F8F9F7F8F907F8F9F\
7F8F9F7F8F9F7F8F9080909F7F8F90808F9F7F8F9080909F7F8F9080909F7F8F90
80909F7F8F90404F4F3F4040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
6F7F606F7F600F0F00000000000000000000000000000000000000000000000000
000000000000000000000000105000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
1F1F2F2F3050303F5F303F5F304060303F5F303F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000F\
0F0F0F606F8F60708F606F8F607090606F8F60709060708F6F709060708F6F7090
6070906F7F9060708F6F7F906F70906F7F906F70906F7F906F7F90707F906F7F90707F906F7F90\
70809F6F7F9070809F707F9070809F707F9070809F707F9070809F707F9070809F
7080907F809F7080907F809F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F7F8090\
7F8F9F7F80907F8F9F7F8F907F8F9F7F8F907F8F907F8F907F8F907F8F907F8F90
7F8F907F8F903F40403F404F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000006F7F60\
6F7F6060705F0F0F0F000F00000000000000000000000000000000000000000000
0000000000000000000000000F4F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000001F2030303F5F303F5F303F5F3F4060303F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000F0F60709060708F60709060708F6F709060708F6F709060708F6F7F906F7090
6F7F906F70906F7F906F70906F7F906F7090707F906F7F90707F906F7F90707F9F6F7F9070809F\
6F7F9070809F707F9070809F707F9070809F70809070809F7080907F809F708090
7F809F7080907F8F9F7080907F8F9F7080907F8F9F7F80907F8F9F7080907F8F9F7F8F907F8F9F\
7F8F907F8F9F7F8F907F8F9F7F8F907F8F9F7F8F907F8F9F7F8F9080909F7F8F90
7F8F903F4040404F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000006F7F60\
60705F6F7F6060705F0F0F0F000000000000000000000000000000000000000000
000000000000000000002000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F203030305F303F5F303F50303F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0F0F0F00000F60708F606F8F60708F606F8F607090606F8F60709060708F6F7090
60708F6F709060708F6F7F9060708F6F7F906F70906F7F906F70906F7F906F7F90707F906F7F90\
707F906F7F9070809F6F7F9070809F707F9070809F6F7F9070809F707F9070809F
70809070809F70809070809F7080907080907080907F8F9F7080907F8F907080907F8F9F708090\
7F8F907F8F907F8F9F708F907F8F907F8F907F8F907F8F907F8F907F8F907F8F90
7F8F903F404F3F4040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006F7F6060705F\
60706060705F60705F000000000000000000000000000000000000000000000000
000000000000000000104F00000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F1F2F303F5F303F5F303F5F303F5F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F606F8F60709060708F60709060708F6F709060708F6F709060708F
6F7F906070906F7F906F70906F7F906F7F906F7F906F7F906F7F906F7F90707F906F7F9070809F\
6F7F9070809F6F7F9070809F707F9070809F707F9070809F70809070809F708090
70809F70809070809F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F\
7F8F907F8F9F7F8F907F8F9F7F8F907F8F9F7F8F907F8F9F7F8F907F8F907F8F90
404F4F3F404F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F7F606070606F7F60\
60705F60705F60705F0F0F0F0F0F0F000000000000000000000000000000000000
0000000000000000000F4000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001F1F2F303F5F2F3050000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000F0F0F0F5F6F8F60708F5F6F8F60708F60708F60709060708F607090
60708F6F709060708F6F709060708F6F7F9060708F6F7F906F70906F7F906F70906F7F906F7F90\
707F906F7F90707F906F7F9070809F6F7F907080906F7F9070809F6F7F90708090
70809070809F70809070809070809070809F7080907080907080907F8F9F7080907F8F90708090\
7F8F907080907F8F90708F907F8F907080907F8F90708F907F8F90708F907F8F90
3F40403F404F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F7F6060705F607F6060705F\
60705F5F705F60705F5F6F500F0F0F000000000000000000000000000000000000
000000000000001F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000001F1F2F303F5F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F0F0F0F606F8F60709060708F60709060708F6F709060708F
6F709060708F6F7F906070906F7F906070906F7F906F7F906F7F906F7F906F7F906F7F906F7F90\
6F7F9070809F6F7F9070809F6F7F9070809F707F9070809F707F9070809F708090
70809F70809070809F70809070809F7080907F8F9F7080907F8F9F7080907F8F9F7080907F8F9F\
7080907F8F9F7F8F907F8F9F708F907F8F9F7F8F907F8F907F8F907F8F903F404F
3F404F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006F7F6F607F606F7F6060705F607F60\
60705F60705F5F705F5F705F000F00000000000000000000000000000000000000
000000000000104F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F1F2F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F00000F60708F5F6F8F60708F5F6F8F60708F60708F607090
60708F60709060708F6F709060708F6F7F9060708F6F7F9060708F6F7F906F7F906F7F906F7F90\
6F7F906F7F906F7F906F7F906F7F906F7F907080906F7F907080906F7F90708090
6F7F90708090708090708090708090708090708090708090708090708090708090708F90708090\
708F907080907F8F907080907F8F907080907F8F907080907F8F90708F903F404F
3F4040000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F7F6F607060607F6060705F6070605F705F\
60705F5F6F505F705F5F6F500F0F0F000000000000000000000000000000000000
0000000F3F00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000F0F0F0F1F1F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F0F0F0F60708F606F8F60709060708F60709060708F
6F709060708F6F709060708F6F7F906070906F7F906070906F7F906F7F906F7F906F7F906F7F90\
6F7F906F7F906F7F9070809F6F7F9070809F6F7F9070809F6F7F9070809F6F7F90
70809F70809070809070809070809F708090708090708090708F9F708090708F907080907F8F9F\
7080907F8F907080907F8F9F708F907F8F90708F907F8F90708F903F404F3F4040
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000607F606F7F60607060607F6060705F607060\
5F705F60705F5F6F505F705F5F6F500F0F0F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000F0F2F30501F1F2F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000F0F0F0F5F6F8F60708F5F6F8F60708F5F6F8F60708F
60708F60709060708F60709060708F6F7F9060708F6F7F9060708F6F7F9060708F6F7F906F7F90\
6F7F906F7F906F7F906F7F906F7F906F7F906F7F906F7F907080906F7F90708090
6F7F907080906F7F907080906F80907080906F8090708090708090708090708090708090708090\
708F90708090708F907080907F8F90708090708F907080907F8F903040403F404F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000607F60607F60607060607F605F705F60705F5F705F\
5F705F5F6F505F6F50506F505F6F50000000000000000000000000000000000000
0F4000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000303F5F303F5F1F2030000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000F0F0F0F0F5F6F8F60708F60708F60709060708F
60709060708F607F9060708F607F9060708F6F7F906070906F7F906070906F7F906F7F906F7F90\
6F7F906F7F906F7F906F7F906F7F9070809F6F7F907080906F7F9070809F6F8090
7080906F7F9070809F708090708090708090708090708090708090708090708F90708090708F90\
7080907F8F907080907F8F907080907F8F90708F907F8F907080903F404F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000607F606F7F6F607F60607F6060705F607F605F705F60705F\
5F6F505F705F5F6F505F6F50506F500F0F0F0F0F000000000000000000000F3F00
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000F2F3050303F5F2F30501F1F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000F60708F5F6F8F60708F5F6F8F60708F
5F6F8F60708F60708F60709060708F60708F60708F607F9060708F6F7F9060708F6F7F9060708F\
6F7F906F7F906F7F906F7F906F7F906F7F906F80906F7F906F7F906F7F90708090
6F7F906F80906F7F907080906F7F907080906F80907080906F8090708090708090708090708090\
708090708090708F9070808F708F90708090708F9070808F708F903F4040000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000607F60607F60607060607F605F705F60705F5F705F5F705F5F6F50\
5F6F5F506F505F6F5050604F506F5050604F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F00000F303F5F2F3050303F5F303F5F1F2030000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F000F0F60708F5F708F60708F5F708F
60709060708F60709060708F607F9060708F607F9060708F6F7F906070906F7F90607F8F6F7F90\
6F7F906F7F906F7F906F80906F7F906F80906F7F907080906F7F907080906F7F90
7080906F80907080906F80907080906F80907080906F8090708090708090708090708090708F90\
708090708F90708090708F90708090708F907080907F8F90708F903F404F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000607F6F6F7F6F607F60607F60607060607F605F705F60705F5F6F5F5F705F\
5F6F505F6F50506F505F6F5050604F506F4F000000000000000000104F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000002F30502F30503030502F3050303F5F303F5F1F202F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F5F6F8F60708F5F6F8F60708F
5F6F8F60708F5F708F60708F60708F60709060708F60708F60708F607F9060708F607F9060708F\
6F7F9060708F6F7F90607F906F7F90607F906F7F906F7F906F80906F7F8F6F8090
6F7F906F80906F7F8F6F80906F7F907080906F7F8F7080906F80907080906F808F7080906F8090\
7080906F808F70809070808F708F9070808F708F9070808F3F404F303F40000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000607F60607F6F607060607F605F705F6070605F705F5F705F5F6F5F5F705F506F50\
5F6F50506F50506F5050604F50604F000000000000000000104000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000303F502F3050303F5F303F50303F5F303F5F
1F2030000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000F0F0F0F0F0F5F6F8F60708F5F708F
60708F5F708F60709060708F60709060708F607F9060708F607F9060708F6F7F90607F8F6F7F90\
607F8F6F7F906F7F906F7F906F7F906F80906F7F906F80906F7F906F80906F7F90
6F80906F7F907080906F80907080906F80907080906F80907080906F8090708090708090708090\
708090708F90708090708F90708090708F90708090708F903040403F404F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006F7F6F607F60607F6F607060607F605F705F6070605F705F5F705F5F6F5F5F6F5F\
506F505F6F50506F50506F504F604F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F304F2F30502F3050303F5F2F3F50303F5F
303F5F1F2030000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000F5F708F5F6F8F60708F
5F6F8F60708F5F708F60708F5F708F60708F60708F60709060708F60708F60708F607F9060708F\
607F9060708F6F7F90607F8F6F7F90607F906F7F90607F8F6F7F906F7F906F8090
6F7F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F7080906F7F8F6F80906F808F7080906F808F\
7080906F808F7080906F808F7080906F808F708F906F808F3F4040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000607F6F607F60607F6F5F70606070605F705F5F70605F6F5F5F705F506F505F6F5F506F50\
506F50506050506F504F5F4F0F0F0F0000000000000F4F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F2F304F303F5F2F3050303F5F303F5F
304060303F5F1F2030000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F0F0F0F5F708F5F6F8F
60708F5F708F60708F5F708F60709060708F60709060708F607F9060708F607F9060708F6F7F90\
607F8F6F7F90607F8F6F7F906F7F906F7F906F7F906F80906F7F906F80906F7F90
6F80906F7F906F80906F7F907080906F80907080906F80907080906F80907080906F8090708F90\
6F80907080906F808F708F90708090708F906F808F708F903F4040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6F806F607F6F607F6F607060607F605F70606070605F705F5F705F5F6F5F5F705F506F505F6F5F\
506F50506F504F604F4F604F000F00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F304F2F304F2F3F502F3050303F5F
2F3F5F303F5F303F5F1F2030000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000F0F0F00000F5F708F
5F6F8F60708F5F6F8F60708F5F708F60708F5F708F60708F60708F607F8F60708F607F8F60708F\
607F9060708F607F90607F8F6F7F90607F8F6F7F90607F8F6F7F90607F8F6F7F90
6F7F8F6F8090607F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F808F\
7080906F808F7080906F808F7080906F808F7080906F808F3F404F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000607F6F
607F60607F6F5F7060607F605F70605F70605F6F5F5F705F506F5F5F705F506F50506F50506050\
506F504F604F4F604F405F40000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F30502F3050303F5F2F3F50
303F5F303F5F30406030405F1F2030000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000F0F0F0F0F5F6F8F
5F708F5F6F8F60708F5F708F60708F5F708F60709060708F607F9060708F607F9060708F607F90\
60708F6F7F90607F8F6F7F90607F8F6F7F90607F906F7F90607F8F6F80906F7F90
6F80906F7F8F6F80906F7F906F80906F7F8F6F80906F80906F80906F808F7080906F8090708090\
6F808F708F906F808F7080906F808F708F906F808F708F90304040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006F7F6F607F6F
607F6F607060607F6F5F70606070605F705F5F70605F705F5F705F506F505F6F5F506F50506F50\
4F604F4F604F405F400F0F0F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F402F304F2F30502F3F5F
2F3050303F5F2F3F5F304060303F5F1F2030000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000F0F0F0F
5F6F805F708F5F6F8F5F708F5F6F8F60708F5F708F60708F5F708F60708F5F708F607F8F5F708F\
607F8F60708F607F9060708F607F90607F8F6F7F90607F8F6F7F90607F8F6F7F90
607F8F6F7F90607F8F6F8090607F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F\
6F80906F808F6F80906F808F6F808F6F808F7080906F808F304040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000607F6F607F6F607F6F607F6F
5F7060607F605F70605F70605F705F5F705F506F5F5F705F506F50506F5F506F50506F504F604F\
4F604F405F40405F40000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F2F304F2F30502F3050
303F5F2F3F5F303F5F303F5F3040603040601F2030000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000F0F
5F708F5F6F8F5F708F5F708F60708F5F708F60708F5F708F607F9060708F607F9060708F607F90\
60708F607F9060708F6F7F90607F8F6F7F90607F8F6F7F90607F8F6F7F90607F8F
6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F7F8F6F80906F808F6F80906F808F708090\
6F808F6F80906F808F7080906F808F7080906F808F3F404F304040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000607F6F607F6F607F6F607F6F5F7060
607F6F5F7060607F605F70605F70605F705F5F705F506F5F5F705F506F50506F504F604F4F6050\
405F4F405F4F405040000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F304F2F2F4F2F3050
2F30502F3F5F2F3F50303F5F2F3F5F304060303F5F1F2030000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
00000F5F708F5F6F805F708F5F6F8F5F708F5F6F8F5F708F5F708F60708F5F708F60708F5F708F\
607F8F5F708F607F8F60708F607F9060708F607F90607F8F6F7F90607F8F607F90
607F8F6F7F90607F8F6F7F90607F8F6F8090607F8F6F80906F7F8F6F8090607F8F6F80906F808F\
6F80906F7F8F6F808F6F808F6F808F6F808F6F808F303F403F4040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000607F6F607F6F5F706F607F6F5F70605F7F60
5F70605F70605F705F5F7060506F5F5F705F506F50506F5F506F50506F504F604F4F604F405F40\
405F403F503F000F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030304F2F304F
303F502F3050303F5F2F3F5F30405F303F5F3040603040601F2030000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F5F6F8F5F708F5F708F5F708F5F708F60708F5F708F60708F5F708F607F9060708F607F90\
60708F607F90607F8F607F90607F8F607F90607F8F607F90607F8F6F7F90607F8F
6F7F90607F8F30404F303F4030404F303F403F404F303F4030404F303F403F404F303F4030404F\
303F406F80906F808F6F80906F808F6F80906F808F3F404F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000607F6F607F6F607F6F607F6F5F7060607F6F5F7060
5F70605F70605F70605F705F5F7060506F5F5F705F506F5F506F5F4F60504F6050405F4F405F4F\
405040405F40000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F2F402F304F
2F304F2F30502F30502F3F5F2F3F5F303F5F2F3F5F3040603040601F2030000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F5F6F805F708F5F6F805F708F5F708F5F708F5F708F5F708F5F708F60708F5F708F\
60708F2F3F40303F402F3F40303F402F3F40303F402F3F40303F40303F40303F40
2F3F40303F40303F40303F40303F40303F40303F40304040303F40304040303F40304040303F40\
304040303F40304040303F40304040303F40304040303F403F404F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F706F607F6F5F706F607F6F5F70605F70605F70605F7060
50705F5F7060506F5F5F7060506F5F506F5F506F50506F504F604F4F604F405F4F405F4F3F5040\
3F5040000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F30402F2F40
30304F2F304F303F502F3F5030405F2F3F5F30406030405F3040603040601F2030000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F5F6F8F5F708F5F708F5F708F5F708F303F4F2F3F40303F402F3F40303F4F\
2F3F40303F40000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F0F0F1000001F00002F00002F0000\
2F00002F00002F00001F00000000000F0F0F30404F6F808F3F404F3F404F304040
3F4F4F3F404F3F4F4F3040403F4F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000607F705F7F6F607F6F5F7F6F607F6F5F7060607F6F5F70605F70605F7060
5F706050705F5F7060506F5F5F705F506F5F506F5F4F60504F6050405F4F405F4F405F40405F40\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F40
2F2F402F304F2F304F2F3F502F3F502F3F5F2F3F5F3040602F405F3040603040601F20301F2030\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000F303F4F2F30402F3F402F30402F3F402F3040000000000000000000000000\
0000000000000000005F00006000005F00006000005F00006000003F00003F0000
0000000000000000000000000000000F0F0F2000002F00006000006000004F00005000003F0000\
2F00003F00006F00006000006F00006000003F404F607F8F6F808F607F806F808F
6080806F808F607F806F808F3040403F404F3040403F4040000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
607F705F706F607F6F5F706F5F7F6F5F70605F70605F70605F7060506F5F5F7060
506F5F5F7060506F5F506F5F506F5F506F5F4F60504F6050405F4F405F4F3F50403F5040000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F30402F2F4030304F2F304F303F502F3F5030405F2F3F5F304060304060304060304060304F6F\
1F2030000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
303F4F2F3F40303F4F2F3F400000000000000000000000000000000000000000004F00005F0000\
5F00006000005F00006000005F00006000005F00003F0000000000000000000000
0000000000000000000000000F0F0F4F00006000006000006000002F0000000000000000000000\
0000000000000000004000006000006F000060303F607F8060707F6F6F706F0000
6F00006000006F00006F80806F808F607F806F808F6080803F404F3040403F404F304040000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000607F70\
5F7F6F607F705F706F5F7F6F5F70605F7F6F5F70605F70605070605F7060507060
5F7060506F5F507060506F5F506F5F4F60504F6F5040604F40604F405F40405F40000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F402F2F402F304F2F304F2F3F502F3F502F405F2F3F5F3040602F405F304060304060\
3040601F2030000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F3F402F30402F3F4F
2F3040506F80506F805F6F8F4F607F000000000000000000000000000000000000000000000000\
0000001000004F00004000005F00005F0000600000100000100000000000000000
0000000000000000005000004F00006000006000002F0000200000000000000000000000000000\
0000000000000000000000006000006000006F00006000006F808F607F806F808F
60606F6F00006000006F0000600000600000606060608080607F80608080607F803F4040304040\
304040304040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000607F705F706F\
5F7F6F5F706F5F7F6F5F70605F70605070605F7060506F5F507060506F5F507060
506F5F506F5F4F6F5F506F5F4F60504F6050405F4F405F4F3F50403F5F40000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F30402F30403030502F3050303F502F3F5F3040602F3F5F3040602F405F304F60\
304060304F6F1F2030000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F30402F3F4F2F3040506F80506F80
506F8F506F805000005000005F0000502F30000000000000000000000000000000000000000000\
0000000000000000001F00005F00005F00005F0000600000200000000000000000
0000000000002F00005F0000600000600000500000000000000000000000000000000000000000\
000000000000000000501F20501F1F6F00006000006F5050604F506F808F607F80
6F50506000006F1F20607F80608080607F80608080607F806F8080607F80608080607F7F6F8080\
607F803F40403040403F4040304040000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000607F705F706F607F70\
5F706F5F7F6F5F70605F7F6F5F70605F70605070605F7060506F60507060506F5F
507060506F5F506F5F4F60504F6F5040604F40604F405F40405F4F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F402F304F2F304F2F3F502F3F50303F5F2F3F5F3040602F405F\
3040603040603040601F20301F2030000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002030402F30402030405060804F607F506F804F607F500000
500000500000500000506F80506F8050607F4F607F000000000000000000000000000000000000\
0000000000000000000000005F00005F00005F00005F00002F0000000000000000
0000000000004F00005F00005F0000600000200000000000000000000000000000000000000000\
00000000000030404F303F40604F5060000060000060303060808F607F80608080
604F4F601F1F607F80607F805F7F80607F805F7F7F607F80607F7F607F805F7F7F606060600000\
608080607F7F607F7F607F7F304040303F3F304040000000000000000000000000
0000000000000000000000000000000000000000000000005F7F6F5F7F705F706F5F7F6F5F706F\
5F706F5070605F70605070605F7060506F5F507060506F5F5070604F6F5F506F5F
4F6F50506F5F4F60504F6050405F4F40604F3F5F403F5F40000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F304F303F502F3050303F5F2F3F5030405F2F3F5F304060\
2F405F304F60304060304F6F3040601F2F30000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000002F30402F30402F30404F607F506F804F60805000004F0000500000500000
50000050000050303F502F30506F80506F80506F7F000000000000000000000000000000000000\
0000000000000000000000004000004F00005F00005F0000200000000000000000
0000001000006000005F0000600000000000000000000000000000000000000000000000000000\
000000000000304040607F8F607F80603F3F601F1F60808F607F806F808F607F80
601F205F707F608080607F80607F805F7F7F607F80607F7F607F805F7F7F608080607F7F600000\
600000600000600000608080607F7F608080304040304040000000000000000000
0000000000000000000000000000000000000000005F7F70607F705F706F5F7F705F706F5F7F6F\
5F70605F706F5070605F7060506F605F7060506F60507060506F5F506F5F4F6F5F
506F5F4F60504F6F50406050406050405F4F405F4F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F2F402F30502F304F2F3F502F3F50303F5F2F3F5F\
30405F2F405F3040602F4060304F603040601F2030000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
20304020303F4F607F405F7F4F607F4F4F604F60804F00004F607F4F2F30500000
4F00005000004F000050506F4F607F506F80506F8050607F000000000000000000000000000000\
0000000000000000000000000000003000005F00005000005F0000300000000000
0000002000005F00005F0000400000000000000000000000000000000000000000000000000000\
00000000000030404F5F7F80607F80601F1F601F1F5F707F607F80607F80607F80
604F4F607F805F7F7F5F7F805F707F5F7F805F707F607F805F707F5F7F7F5F7F7F607F7F5F7F7F\
604F4F5F00006000005F00006000005F7F7F607F7F303F3F304040303F3F000000
0000000000000000000000000000000000005F706F5F7F705F706F5F706F50706F5F706F507060\
5F706F507060507060506F60507060506F5F506F604F6F5F506F5F4F60504F6F5F
4060504F605040604F40604F3F5F403F5F40000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F304F303F502F3050303F5F2F3F5030405F\
2F3F5F3040602F405F304F60304060304F603040603F4F6F1F2030000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F3040\
2030404F607F405F7F4F00004F00004F00004F607F5060804F607F506F804F607F
5000004F00005000004F0000504050506F80506F80506F80506F7F000000000000000000000000\
0000000000000000000000000000002000005000005F00005F00004F0000000000
0000000000005F00005F00004F0000000000000000000000000000000000000000000000000000\
000000000000607F80607F8F607F8060606F60606F60808F607F80607F805F7F7F
601F205F7F80607F805F707F5F7F805F7F7F5F7F805F707F607F805F7F7F5F7F7F5F7F7F5F7F7F\
5F0000600000600000606F606000006000005F7F70607F7F5F7F70304040303F3F
3040400000000000000000000000005F7F705F7F705F706F5F7F705F706F5F706F5070605F706F\
507060507060506F60507060506F60506F604F6F5F506F5F4F6F5F4F6F5F4F6050
4F6050406050406050405F4F405F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F304F2F30502F30502F3F502F3F50\
303F5F2F3F5F30405F2F405F3040602F4060304F60304060304F6F304F601F2F30
000000000000000000000000000000000000000000000000000000000000202F3F20303F202F3F\
405F7F405F704F00004000004F101F405F704F607F4F607F4F607F4F607F4F607F
4F607F4F1F1F4F00004F00004F00005040504F607F506F804F60704F607F000000000000000000\
0000000000000000000000000000001000001000005000005F0000400000000000
0000000000002000005F0000400000000000000000000000000000000000000000000000000000\
0000002F3F40607F805F7F80607F805F7F8060606F5F7F80607F805F7F80604F50
5F4F4F5F7F8050707F5F707F5F707F5F7F7F50707F5F7F7F5F707F5F7F7F5F5F5F5F60605F0000\
5F00005F7F7F607F7F5F7F70607F7F5F7F705F7F7F5F7F70607F7F5F70705F7F70
2F3F3F30403F0000000000005F706F5F7F7050706F5F706F50706F5F706F50706050706F506F60\
507060506F60506F604F6F5F506F604F605F4F6F5F4F60504F605F406050406050
405F4F4060503F5F4F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F304F303F502F3050303F50\
2F3F5030405F2F3F5F3040602F405F304F60304060304F60304F603F4F6F304F6F
1F2F30000000000000000000000000000000000000000000000000202F3F203040405F70405F7F\
405F704F00004000004F00004000004F3F4F4F607F4F607F4F607F4F60804F607F
4F60804F607F501F1F4F00005000004F2F30502F3F4F6F80506F804F607F000000000000000000\
0000000000000000000000000000001000001000005F00005000005F0000500000
0000000000002F00005F00004F0000000000000000000000000000000000000000000000000000\
00000030404F5F7F80607F805F7F80607F805F7F80607F805F7F80607F805F1F1F
5F707F5F707F5F7F805F707F5F7F7F50707F5F7F7F5F707F5F7F7F5F404F5F4F4F5F00005F0000\
5F7F7F607F7F5F7F7F607F7F5F7F70607F7F5F7F70607F7F5F7F70607F7F5F7F70
5F7F702F3F3F303F3F5F70705F7F705F706F5F7F7050706F5F706F5070605F706F507060507060\
506F60507060506F60506F604F6F5F506F5F4F605F4F605F4060504F6050405F50
406050405F4F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000002F304F2F30502F304F\
2F3F502F3F50303F5F2F3F5030405F2F405F3040602F4060304F60304060304F6F
304F603F4F6F1F20300000000000000000000000000000001F2F3F202F3F3F5070405F703F0000\
400000400000403F4F4000004000004000004F607F405F704F607F40607F4F607F
40607F4F607F4F607F4F404F4F00004F00004F101F506F804F607F506F804F6070000000000000\
0000000000000000000000000000000000001000005000005000005000005F0000
0000000000002000005F0000400000000000000000000000000000000000000000000000000000\
0000002F3F405F7F805F70805F7F805F7080607F805F707F5F7F805F707F5F4F4F
50707F5F707F50707F5F707F50707050707F5070705F707F502F2F5F30305000005F7F7F5F7070\
5F7F7F5F70705F7F7F5F70705F7F705F70705F7F705F70705F7F705F70705F7F70
0000000000002F3F302F3F3F50706F5F706F50706F5F706F506F6050706F506F605070604F6F60\
506F604F6F5F4F6F604F605F4F6F5F4060504F6050405F50406050405F4F405F4F
3F5F4F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F304F303F50\
2F3050303F502F3F5030405F2F3F5F30405F2F405F304F60304060304F60304F60
3F4F6F304F6F3F506F1F20300000000000000000001F2F3F20303F3F5070405F703F5070405F7F\
405F70405F7F405F70404F604000004F00004000004F607F40607F4F607F40607F
4F607F4F607F4F607F4F607F4F6F804F00004F00004F0000506F804F6F7F506F804F6070000000\
000000000000000000000000000000000000000000500000400000500000500000
1000000000000000005000005F0000000000000000000000000000000000000000000000000000\
0000005F7F805F7080607F805F7F80607F805F7F80607F805F7F805F707F5F707F
5F7F8050707F50707F50707F5F707F506F70507070501F1F5F00005000005F7F7F5F7070607F7F\
5F7F7F5F7F7F5F70705F7F7F5F70705F7F705F70705F7F705F7070000000000000
1F00000000000000002F3F305F707050706F5F706F50706050706F507060507060506F60507060\
4F6F5F4F6F604F605F4F6F5F4F60504F605F406050406050405F4F405F503F5F4F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F30502F304F2F3F502F3F50303F5F2F3F5030405F2F405F3040602F405F304F60
304060304F6F304F60304F6F304F6F1F2F30000000202F3F1F2F3F3F5070304F603F50703F5070\
405F703F5070405F70405F70403F4F40000040000040202F403F4F405F704F607F
405F7F4F607F40607F4F607F4F607F4F607F405F704F00004F00004F607F4F607F4F60704F6070\
000000000000000000000000000000000000000000000000400000500000500000
200000000000000000200000500000400000000000000000000000000000000000000000000000\
00000050707F5F7F805F70805F7F805F707F5F7F805F707F5F7F80500000501F1F
50101F501F1F4F6F7050707F50404050404F4F0000500000505F5F5F707F5070705F7F7F5F7070\
5F7F7F5070705F7F705070705F7F705070705F7070000000000000000000000000
5000000000000000002F3F302F303050706F506F6050706F506F605070604F6F604F6F604F6F5F\
4F6F5F40605F4F605F406050406050405F504060503F5F4F405F4F3F5040000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003030502F3050303F502F3F5030405F2F3F5F30405F2F405F304F60304060
304F60304F603F4F6F304F6F3F506F304F6F3F50701F2F303F50703F506F4000003F506F405F70\
3F5070405F70405F70405F70405F70405F7F4000004000004000004F5060405F7F
4F607F40607F4F607F4F607F4F607F4F607F4F6F7F4F607F4F00004F000050506F4F6F7F4F607F\
000000000000000000000000000000000000000000000000000000500000500000
5000000000000000001000005000005000000F0F0F000000000000000000000000000000000000\
2F3F405F7F805F70805F7F805F70805F7F805F70805F7F80502F305F3030505F60
505F604F101F501F1F501F1F501F1F4F00005000005070705F707F5070705F7F7F5F70705F7F7F\
5F70705F7F7F5F70705F7F705070705F7F70000000000000000000000000000000
5000005000000000000000002F3F302F303050706F506F6050706F4F6F60506F604F6F5F4F6F5F\
4F605F4F6F5F406050406050405F504060503F5F4F3F5F4F3F5040000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F304F2F304F2F3F502F3050303F5F2F3F5030405F2F3F5F304060
2F405F304F60304060304F6F304F60304F6F1F20301F2F300000003F00003F00003F50703F506F\
3F50703F506F405F703F506F405F703F5F70405F703F5F7040000040000040607F
405F7040607F405F704F607F405F704F607F40607F4F607F40607F4F101F4F00004F404F4F607F\
4F6070000000000000000000000000000000000000000000000000300000500000
4F0000400000000000000000500000500000200000000000000000000000000000000000000000\
2F3F4050707F5F708050707F5F70805F707F5F7F8050707F501F1F506F7F50707F
4F6F704F5F604F101F5000004F00005000004F6F7050707F5070705F707F5070705F707F507070\
5F70705070705F707050706F000000000000000000000000000000000000400000
4F00005000004F00000000002030302F30304F6F60506F604F6F5F4F6F604F605F4F6F5F406050\
40605F4060504060503F5F4F405F503F5F4F3F5F4F1F201F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000030304F2F304F2F3F502F3F50303F5F2F3F5030405F2F405F
304060304060304F60304F60304F6F304F601F2F300000000000003F00003F00003F00003F5070\
3F506F3F5F703F506F405F703F5F70405F703F5F70405F70405F70400000400000
401F1F405F704F607F405F704F607F40607F4F607F40607F4F6F7F4F607F4F6F7F4F00004F0000\
4F607F4F6070000000000000000000000000000000000000000000000000200000
500000400000000000000000300000500000300000000000000000000000000000000000000000\
2F3F405F708050707F5F7F805F707F5F7F8050707F5F7080505F6050707F4F6F70
50707F4F00004F00004F000050404F506F7050707F5070705F7F7F5070705F7F7F5070705F7F7F\
5070705F7070000000000000000000000000000000000000000000500000200000
400000100000500000000000000000203030506F604F6F604F6F604F6F5F4F6F5F40605F4F605F\
406050406050405F50405F503F504F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F2F402F304F2F304F2F3F502F3F502F405F
2F3F5F30405F2F405F304F60304060304F601F20300000000000003F00001000003F00002F1010\
30101F304F603F506F3F506F3F5F703F506F405F703F5F70405F703F5F70405F70
40000040000040202F406070405F704F607F405F704F607F4060704F607F4060704F607F4F3F4F\
4F00004F101F406070000000000000000000000000000000000000000000100000
4F00004000000000000000002F00005000002F00000000000000000000000000000000002F303F\
50707F4F6F7F50707F50707F5F708050707F50707F505F6050707F4F6F704F6F70
4F00004F00004000004F404F4F6F705070704F6F705070705070705F7070507070507070507070\
0000000000000000000000000000000000003000002F0000200000000000000000
100000000F0F3F00000000000000002F30304F605F4F6F6040605F4F605F40605040605F405F50\
4060503F5F4F3F5F4F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F30402F304F2F304F303F502F3F50
30405F2F405F30406030405F304F603040601F2F30000000300000300000000000000000000000\
0000003F00002F101F3F30403F506F405F703F506F405F703F5F70405F70405F70
40101F40101F40000040000040607F405F704F607F4060704F607F4060704F6F7F4F607F4F6F7F\
4F3F4F4F00004F00004F6070000000000000000000000000000000000000100000
4F00003F00000000000000003000004F00003000000000000000000000000000000000002F3F40\
4F6F704F6F7F4F6F704F6F7F4F6F704F6F7F4F6F704F6F704F6F7050707F4F4040
4F00004000004F6F704F6F705070704F6F7050707F5070705F70705070705F7070000000000000\
000000000000000000000000000000400000400000000000000000000000000000
00000000000000000000000000000020302F4F6F604F605F4F605F40605F40605F406050406050\
3F5F4F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F3F2F304F2F304F2F3F50
2F3F502F3F5F2F3F5F30405F2F405F30406010202F1F20300000003000003000002F0000000000\
0000000000000000000000003F101F3F202F3F30403F506F3F5F703F506F40101F
3F1010404F5F3F4F5F4000003F0000400000405F7040607F405F7040607F4060704F607F406070\
4F607F404F5F4F0000406070405F70000000000000000000000000000000000000
0F00004F00000000000000000000004F000040000000000000000000000000000000000020303F\
4060704060704F6F704060704060704060704F6F703F5F6F406F70403F404F0000
4010104F6F704F6F704F6F704F6F70507070506F70507070506F70507070000000000000000000\
0000000000003000002F0000000000000000000000000000000000000000000000
00000000000000000010000020302F20303040605F40605F405F5040605F3F5F503F5F50000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F3F2F30402F304F
303F502F3F5030405F2F3F5F30405F2F405F304F601F202F0000003000002F00003000003F0000\
2F00002F00000000000000000000000000001F2F3F3F506F3F202F3F00003F1010
404F5F3F4F5F405F703F5F7040607040000040000040101F4F607F4060704F607F4060704F607F\
4060704F607F404F5F4F0000406070405F6F000000000000000000000000000000
0000001000004000003000000000000000002F0000400000000000000000000000000000000000\
0000004060704060704F6F70406070406F7040606F4F6F704060704F6F7040606F
4F6F704060704F6F704F6F70507070507070507070507070000000000000000000000000000000\
0000003000002F00000000000000000000000000000000000000000000003F0000
3000004F00004F000000000020303020302F40605F40605040605F3F5F50000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F2F402F304F2F304F2F3F502F3F5030405F2F405F1F20300000003000002F0000000000000000\
0F00003000003000003000000000000000000000000000000000001F2F303F5F6F
3F10103F20303F506F405F703F5F70405F703F5F704000004000004060703F5F70406070405F70\
4060703F5F70406070404F5F4000003F00003F5F6F000000000000000000000000
000000000000000000400000400000000000000000000000400000400000000000000000000000\
00000000000000000000000020303F20303040606F40606F20303F00000040606F
40606F406F7040606F4F6F704F6F705070700000000000000000000000000000000000002F0000\
0000000000000000000000000000000000001000004000004F00004000004F0000
1F0000000000000000000000202F2F40605F3F5F503F5F501F2F20000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030304F2F304F303F502F3F5030405F1F202F1F2030000000000000000000000000\
0000000000000000000000002F00003F0000300000100000000000000000000000
0000001F2F3F405F703F5F6F403F403F5F70406070405F70402F303F0000402F30405F70406070\
3F5F6F406070405F704060703F5F6F40505F4000004000003F5060000000000000
0000000000000000000000000000000F00004000003000000000000000000000002F0000400000\
4000002F00000000000000000000000000000000002F0000400000000000000000
000000000000000000000000000000000000000000000000000000000000200000200000000000\
0000000000000000000000004000004F00004000004F00001F0000200000000000
00000000000000000000000020302F3F5F5020302F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F30402F3F4F2F3F502F405F101F2F000000000000000000000000\
0000000000000000000000000000000000002F00003000003F00003000001F0000
0000000000000000000000003F506F3F5F703F5F6F40101F3F5F6F3F5F703F202F4000003F202F\
3F5F6F3F5F6F405F703F5F6F3F5F6F3F5F6F405F6F3F40503F00003F0000405F6F
3050600000000000000000000000000000000000000000000000001000003F00004000003F0000\
300000300000000000000000300000300000400000000000000000000000000000
0000000000000000000000000000000000000000000000000000002F0000000000000000000000\
0000003F00004F00004000002F00002F00001F0000000000000000000000000000
00000040000000000020302F3F5F50000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000002F3F4F2F3F4F30405F10202F000000000000300000\
2F00002000000F00000000000000000000000000000000000000001F0000300000
3F00002F000000000000000000000000000020303F3F5F6F405F703F101F3F5F6F3F5F6F402F30\
3F00003F00003F5F6F405F6F3F5F6F3F5F6F3F5F6F405F6F3F5F6F3F5F6F3F5F6F
404F5F3F00004000003F5F6F3F5F60000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F00001F0000
0000000000000000000000000000000000000000003F00000F0F0F0000000000004000004F0000\
4000001000000F00002F00000000000000000000000000000000000000000F0000
40000020302F20302F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F304F2F3F502F3F501F202F000000\
0000000000001000002F00003000000F00001000001F00000F0000000000000000
0000000000000F00003F00003F00000000000000000000000000003F5F6F3050603F5F6F3F1010\
3F506F3050603F5F6F3F00003F00003050603F5F6F3F50603F5F6F3F50603F5F6F
3F50603F5F6F3F50603F5F6F3F5F603F5F6F3F40503F00003F00003F00003F00003F00003F0000\
3F00003F00003F00003F4F503F5F6F3F5F603F00003F0000300000000000000000
0000000000000000000000000000000000000000000000002F00004F0000400000000000000000\
2F00004000000000000000000000000000002030301F302F3F00003F00003F0000
1F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000303F502F3F501F202F\
0000000000000000000000000000000000003000003000002F0000000000000000
0000002F00001000000000000000000000003F00003F000000000000000000000000000020303F\
3050603F5F6F3F50603F101F3F50603F5F6F3F50603F202F3F00003F00003F202F
3F5F6F3F50603F5F6F3F5F603F5F6F3F5F603F5F6F3F5F603F5F6F3F5F603F5F6F3F5F603F5F6F\
3F5F603F5F6F3F5F603F5F6F3F00000000000000000000000000000000003F0000
4000003F00004000004000004000004000004F0000400000100000000000000000400000400000\
000000000000203030402F2F3F00004000003F00003F00003F00003F5F501F2F2F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000002F304F\
1F202F0000003000000000000000000000000000000000000F0000100000300000
3000000F00000000000000000000000000002F00000000000000003000003000003F0000000000\
0000000000000000001F2F303F50603050603F50603050603F10103050603F5060
3050603F5F603050603F5F603020203F202F30202F3F202F3020203F202F3050603F5F60305060\
3F5F6030505F3F5F603050602030300000000000000F00003F00000F0000000000
0000000000000000000000000000000000000000000000000000001000000000000000003F4F4F\
3F00003F00003F00003F20203F3F303050503F5F503000003F50501F2F2F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000002F3F4F3000002F0000300000000000000000000000000000000000
0000000000003000003000003F00000000000000000000000000000000002F00000F0000000000\
0000000000003000003F0000000000000000000000000000000000000000000000
20303F1F2F302030303050603F5F603050603F5F6030506020303F1F2F30000000000000000000\
0000003F00003F00003F5F603F202F3F00003F10103F5F60203030203030000000
0000000000001000000F00001000000F00001000000000002030303F00004000003F000040605F\
3F5F5F3F5F5F3F5F503F5F5F3050503F0000300000202F2F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000001F202F2F3F4F2F3F502F3F502F40502F3F50000000
0000002000001F00000000000000000000000000001000003000002F0000000000000000000000\
0000000000000000000000002F00000F0000000000000000000000000000000000
0000000000003000003000003000003F00000000000000000000000000000000000F0000000000\
0000000000002F000030505F3F5F6030505F3F5F6030505F3F2F2F3000003F0000
3F00003F10103F10103F10103F5F5F4060603F00003F3F3F3F5F5F3F5F5F3050503F5F5F305050\
3F3030304F4F30505030504F30504F1F2020000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000001F1F2F303F502F00003000002F1F20
3040502F4050304F5F2F4050304F5F000000000000000000200000000000000000000000000000\
1000003000003000003000002F0000000000000000000000000000000000000000
0000000000000000000000000000000000000000002F00002F00003000001F0000000000000F0F\
0000000000003010103F4F5030505F3F5F603F5F5F3F5F603F5F5F3F5F603F5F5F
3F5F603F5F5F3F5F603F5F604060603F5F5F3F5F603F5F5F3F5F5F3010103F5F5F305050305050\
30504F305050304F4F1F2F2F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000002F304F2F0000
2F3F4F2F40502F3F503040502F405030405F2F405F304F5F2F4F5F304F5F2F4F5F304F5F000000\
0000000000000000000000000000001F00002000001F00002000001F0000200000
0000000000001F000020000000000000000000000000000000000030505F30505F30505F2F4F50\
30303F30000030000030505F30505F30505F3F101030404F3F5F5F30505F3F5F5F
30505F3F5F5F3F5F5F3F5F5F304F503F3030301010305050305050305050304F4F30504F2F4F4F\
304F4F102020000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F3F4F2F3F4F2F3F4F3040502F40503040502F4050304F5F2F2F303000002F0000300000\
2F4F5030505F2F4F5F305060304F5F30505F2F4F5F305060304F5F30505F304F5F
30505F304F5F30505F2F4F50304F5F30000030303F30505F30505F30505F30000030000030505F\
30505F30505F30505F3F3F3F3000003F00003000003F5F5F30505F3F5F5F30505F
3F5F6030505F3F5F5F3040403F10103050503050502F10103000002F00001F2F2F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002F3F4F2F00002F00002F00002F00002F00002F2F302F2F30\
3010102F00002F4F5F2F4F5F304F5F2F4F50304F5F2F4F5F304F5F2F4F50304F5F
2F00003000002F00003000002F3F4030404F2F4F5030505F2F4F5030505F2F2F30300000000000\
30505F304F5030505F30505030505F30505030505F30505F30303F303030305050
304F503050502F4F4F304F4F2F404F2F40402F4040000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000101F20301F202F3F4F30404F\
2F3F4F3040502F4050304F5F2F4F50304F5F2F4F5F304F5F2F4F5030505F2F4F5F
304F5F2F4F5030303F2F2F303000002F202030505F000000000000000000000000300000300000\
1F0000000000000000000F0F304F5030505F304F50304F502F4F4F304F502F404F
304F4F2F4040304040000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F303F2F00002F00002F00002F00002F101010202F000000
2000000000000000000000000000002F0000200000000000000000000000000000000000000000\
0000000000000000000000000000002F00001F2020000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[8]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061h<0015a0740L@1[06/0J`1[06/0L@1n00P700061gT000A[LF]a3F/03W5[LF]aJg5[LF]aJg5[
NP081`001PMg0003L@1[00Y[01L0Jf]aJg5[LF]aJg5[LF]aJg5[LF]aJ`1g00P700061g8000]aJg5[
LF]aJg5[L@07J`0KLF]aJg5[LF]aJg5`LG1aL75`LG1aJg5[LF]a07<020L000H7K`00<G40L@1a0740
LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg400V]`00P700061f`001=aJg5[
LF]aJg5[L@1a0740L@1a01T000]aL75[LF]aJg5[L@1_00P700061fT000Ua0740L@1a07400V/Y0005
LF]aJg400V]^00P700061fH000]aJg5[LF]aJg5[L@0_0005LF]a0740K@081`001PMQ0009L@1a0740
L@1a03T000=[06/00V/90003C@0005h020L000H7GP002G5[LFaaK75[L@0m0009L@2:L8Y`RP1a06D0
20L000H7G@001g40L@1a0740@@02J`09RV]aJg5[L@1[068020L000H7F`001VaaJg5/LD<000ea075`
RW2:L8YaRW6:009a0`000dd0001I00P700061e/000=a0000A`004700RV^:JhY[LF]aJg5[LF]M00P7
00061eT000=/L@00B@0046]aLHY`RW2:LHYaRW6:LHX2L003070005P020L000H7Y@004g00L6^:JhY[
RV]aJg5[LF]aJdd0FP081`001PNV000ELF^:LHYaRW6:LHYaRW6:LHYaL4e`05L020L000H7Y`005g00
RV^:JhY`RW2:JhY[RV]a04d0C@1=05@020L000H7E@000f/0001>000GLF]aLHYaRW6:LHYaRW6:LHYa
L4e`CG00E@081`001PMC0003L@00054001=`08Y`RW2:L8Y`RW2:L8Y`RW2:009=00@0L6]aDP081`00
1PM@0005LF]aK740D@000g5[L@04RP0<LHYaRW6:LHYaRW6:0W400de`C@02L003L@0004l020L000H7
C`001G40L@1a05<001Ea08Y`RW2:L8Y`RW2:L8Y`RW2:L7400Td01P1aJg40CDh020L000H7CP001W5[
LFaaK5<000=[LHX01HX0576:LHYaRW6:LHYaRW6:CG1=LG1a0W1=00P700061dd000Ma0740L@1a05<0
01ma075aRW6:L8Y`RW2:L8Y`RW2:L8Y`RW1=04e[L@1=04`020L000H7C0001g5[LFaaK740E0000f]a
J`0;RP03LHYa00>:00YaRW6:CG1=RW2:0W1;00P700061d/000Ma0740L@1a05D0025a075aRW6:L8Y`
RW2:L8Y`RW2:CHY`RW2:L4d0RV]aJg40BP081`001PM8000;L@1aJg5[LFaaK740E0000f]aJ`04RP05
ShZ:RXl028X044e`LHYaRW5`CG5`RW2:CG5700P700061dL000]a0740L@1a075[L@0b000FLHY[RW2:
JhY[RV^:JhY[RV^:Q8Y[RPd001AaJg5aRW6:LHYaRW2:L8Y`RW2:L0A=00f:L8Y[C@1=Jg5[LF]a04H0
20L000H7AP003740LF]aJg5/LFaaK1l000M[RV^:JhYa00b:00>@RXX00XXAQ005RXV:RHX020000f]a
J`02RP0;ShZ?RXn:ShZ?RXl018X2L006CG1=RW6:0W0024e`L8Y`RW1aA@081`001PM5000<L@1a0740
L@1a075[7@008F^:JhA[Q6^4JhYaRW6:LHYaRW6:LHYaRW6:LHA[Q6^4J`0;Q009JhA[Q6^:JhY[00D0
01I[LHZ:LHYaRW6:LHY`RW2:LHYaRW1a0dd2L00:RW1=04d0LF]aJdD020L000H7A000175/LF`7L@03
K74001D000M[RV^:JhA[00J418X04H^:T8Z@RY2:T8Z@RY2:T8Z@00>:2X@00hZ4RP0=Q00QRXUaJg6:
ShZ?RXn:ShZ?RXn:ShZ?RXn:ShYfCG1=L76:009a00E=L72:L01400P700061d<000ea0740L@1a075[
LF]a01@002Q[RV^:JhA[Q6^4JhA[Q6^4LHYaRW6:LHYaRW6:LHYaRW6:LHYaRW6:0X@01f^4JhA[Q6/0
0h@00hZ4Q009Q006JhA[Q01a1000576:LHYaRW6:LHY`RW2:L8Y`RW2:0dd2L003RW1a009=00<0LF/0
@`081`001PM20004L@1aK0Ua00=/00003`003Fb:K8A[Q6^4JhA[Q6/018@2RP0GRhZ;RX^:T8Z@RY2:
T8Z@RY2:T8Z@RY000hX7Q0>:00F@RY2:T00;Q003LF]a00L0016?RXn:ShZ?RXn:ShZ?RXn:L@02L006
CG1=RW6:0W000deaL01200P700061d4000ma0740L@1a0740LF]aJg403@00;75[LF^4JhA[Q6^4JhA[
Q6^4JhA[RW6:LHYaRW6:LHYaRW6:LHYaRW6:LHYa0hX00g6:J`0:Q0::00J?RXn:ShX7Q003074000`0
00f:L8YaRW2:L8Y`RW2:009`0dd2L006RW1=071[@@081`001PM00004L@1aK0]a00=/0000200058Y/
RVb4JhA[Q6^4JhA[RV^:JhA[0h@078Z;RX^:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY02RP03Q8Z400R4
00N@RY2:T8Z@00::00>4RX@00X@00g5[L@0?0007ShYeL8n:S`05RP0<LDe`CG1=RTe`CG5`@0081`00
1PLo000AL@1a0740L@1a075[LF]aK7401`000g5[L@05J`0TQ6^4JhA[Q6^4LHY[RV^4JhE[RW6:LHYa
RW6;LH]aRg6;LI1a2hX00hB:Q00;Q006RXn:ShZ?0XX4Q0030740014000a`CG1aL8Y`RW2:L8X5C@05
RV]`0740@0081`001PLn0006L@1aK75/3743000ILHYaRV^:JhA[Q6^4JhA[RV^:LH]aRf^4J`02Q0::
026@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY002h@2RP05T8Z@RY000XX3Q003LF]a0180
00Q=MDefL8n:S`B:00Y`LDe`CHY`RTea?`081`001PLo000GL@1a0740L@1aJg5[LFaaK740LFaaK740
0V`5J`0VQ6^4JhAaRW6:JhY[Q6^4JhA[RW6:LHYaRW6:LH]aT76;LI1aT746RP08ShZ?RXn:T8X<Q007
RXB:RXn:S`02RP05Q8Y[LF/04P000e9=DP02C@08L8Y`RW2:L8X2L0==00E`JhX0L@0n00P700061ch0
00Ia675/LF`2J`Ya01Z:LHYaRW6:JhY[Q6^4JhYaRg6;LH]aQ6^4J`B402>;RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T00>Q008RY2:T8Z@RY02RP03JhX0014000Y=MDeeCGE`MXZ?1XX00de`
C@02L003RTea03d020L000H7?@006W40L@1a06/0K01aJg5/LFaaJg5/LFaaK75/16/0=XY[RW6:LHYa
RW6:JhA[Q6^4JhAaRW6;LH]aRg6;LI1aT76@LI1aT76@RY2:T8Z@RY2:T8Z@RPn400Z:ShZ?RXn:S`1a
5003C@0=DTe=CG5`RW2:L8Y`RP05C@03L01=03`020L000H7?0001G4HLFaa00A[2742000ILHYaRW6:
LH][RV^4LHYaRg6;LH]aRV^4J`06Q00ST8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY00
0hX<Q00;RXB@RY2:T8YaJg405@002G1=MDeeCGI`S`04RP05LHY`LDd00g000dd0000j00P700061c/0
00ma06/0J`1[06`0LF]a07401@07L@=[03MaK8YaRW6;LHY[Q6^4JhA[Q6^4JhYaRg6@LI1aT76@LI2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00::00>9RX@03X@028n:T8Z?Jg5[6002C@03DTeB009=00Q`LG2:
L8Y`RP9`00A=L@1=>P081`001PLj0003LAQ[00E[1G400f`00005000ELHYaRW6:LHY/RW6;LH]aRg6;
JhY[00R402F:Q92:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00>:3h@028Z@RY2:LG1a
6P0024e`CGE=MDef0XX0176:LHX2L@03CG5=03X020L000H7>@003G40J`1[06/0L@1a07402@08L@0i
RW6;LH]aRg6:LHA[Q6^4JhA[Q6^:JhYaT76@LI1aT76@LI2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00B:
3h@028Z9T8Z@Jg5[6`03C@0<DTeBCG5`RW2:L8Y`0Td00g40000h00P700061cX01F/6L@X001EaRW6:
LHYaRg6;LH]aRg6:K8Y[Q6/01h@3RP0ST8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY00
0hX4Q005WHB4Q9d01h@2RP05T8YaL7407@003W1=MDeeCG5`RW6:LHYa0g0i00P700061cT000][06/0
L@1a075/L@0;00Ia03^:LHYaRg6;LH][Q6^4JhA[Q6^4JhA[Q6^:LI1aT76@LI1aT76@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T004RPn400R:Q92:T6]aJa@01Dd600I=0g000hY`RP02C@03Jg4003L020L000H7
>0001V]/LFaaK0Ea30005G6;LH]aRg6;LH]aRg6;LHA[Q6^4J`08Q0::02J@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z?RPB400NMQ9f4WHBM00>400VMQ8B?T8maL7404@002Dd0L4eeCGE=
M@07000?L4eeCGE=L76:LG1=RW1a03L020L000H7=`002F/0L@1a0740L@0?00Aa03f:LHYaRW6;LH][
RV^4JhA[Q6^4JhA[Q6^4LHYaT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00B:4H@01Y2:
T6]aJa<00dd01e9=DTeB04d01@08C@07L4e=L75[L@0f00P700061cP000AaK75/0g400f`0000=000E
LH]aRg6;LH]aRg6;K8Y/Q6^4JhA[00R40XX09Y2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RXn:18@02if4WHBMQ9f4WHBM00:400J?T8maL74A000<C@1eCGE=MDeeCGE=1`003DeeCGE=L4e`
CG5`LF/0=P081`001PLg0007L@1a0740L@0A009a03n:LHYaRW6:LH]aRV^:JhA[Q6^4JhA[Q6^4JhAa
RW6@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY0018XAQ006T8Z@Jg5[4`002TeBCE9=DTeB
CE8800U=00>:Jg40=P081`001PLf0007L@1aK75/L@0B000ELH]aRg6;LH]aRg6;JhA[Q6^4JhA[00R4
0XX0892:T8Z@RY2:T8Z@Ri2:T8^@RY2;T8Z@RY2:T8Z@RY2:0i03RPB4012MQ9f4WHBMQ9f4WHB:Si2?
0g4A000<C@1eDWE=MDeeCGE=1P002Td0001`CGE=L4d2L003RW1a03D020L000H7=@000g40J`0G0011
Rg6;LH]aRg6;LH]aQ6^4JhA[Q6^4JhA[Q6^4JhA[RW6@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY000XX00hB:Q00>Q008RXB@RY1[LF/C0008CE9=DTeB04d70003DTeB00D00dd01W2:L75[LC@0
20L000H7=0001G5[LF]a01H000eaRg6;LH]aRg6;LH][01240XX0592;T8^@Ri2;T8^@Ri2;T8^@Ri2;
39006ZV@ZHBSQ9f4WHBMQ9f4WHBMQ9f4WHZSSjV?0g4C0006ME9eCGEB2@001DeeCGE=00D000I=L76:
LHX2LC<020L000H7<`001G40J`1a01L004:;LH]aRg6;LH]aRg64JhA[Q6^4JhA[Q6^4JhA[Q6^:LI2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8XBQ008T8Z@RY1[LF/D009=00=BCE802@001e9=
DTeBCE8010001de`L75`LF/0<`081`001PLb0006LFa[Jg5/5`003W6;LH]aRg6;LH]aRf^:3h@05XZ4
T8^@Ri2;T8^@Ri2;T8^@Ri2;T8/=T00IRY24WHBMQ9f4WHBMQ9f4WHBSRJ>?ZHnYS`03LA<00Td01GEB
ME9=00P000U=MDeeCGE=MDd00P00271=RW6:LHY`<P081`001PLc0003L@1a01T0022;LH]aRg6;LH]a
Rg65JhY[Q6^4JhA[Q6^4JhA[Q6^4L@::01n@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8B:0124
00^:T8Z@RY2:T6]aJ`0E0003CE9=00==1`002E9=DTeBCE9=DP02009=00M`RW1aL75[034020L000H7
<P00175/LF`I000<LH]aRg6;LH]aT76;0hX@Q00BRhB:Ri2;T8^@Ri2;T8^@Ri2;3902RP0JWHBMQ9f4
WHBMQ9f4WHBMRZV:ZI2YSjV?ZHl3LAD000IeDWEBMDd7000:CGE=MDeeCGE=M@800Td0176:LHX2LC40
20L000H7<@001G40L@1a01T001j;LH]aRg6;LH]aRg6:Jh][Q6^4JhA[Q6^4JhA[Q6/4Q00KRXF:RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00::38@04HZ?RY2:ShZ@RY2:T8Z@Jg5[01H00Td00e9=DP070003
DTeB00==00D0CE9=DP040006L75`LG1a<0081`001PL`0005LFaaK7406P003g6;LH]aRg6@LI2:T8Z@
R`03RQ2400f:Q8B:QHZ;T8^@Ri2;00Z@0XX07Z>:Q8BMQ9f4WHBSQ:>4ZI2YT:V@ZI2YT:V@ZHnYS`=a
5`001WEBME9eDPD000I=MDeeCGD30003MDd000@000=`RW400W4`00P700061c4000=a07406`007X]a
Rg6;LH]aRg6;LH]aRfb;LHA[Q6^4JhA[Q6^4J`R400F:Q92:T002RP08QHZ5RXF:QHX=Q00JT8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY1[LF/G0008CE9=DTeB04d30003DTeB00==0P0014d0DTd50004L75`LC00
20L000H7;`001FaaK75/01/001EaRg6;LI1aT76@RY2;T8^@Ri2;T8/00hX00hB:Q006Q003RXB400N4
00>:Q8X00hX00j>:RP02RP>4022MQ9f4WHBSQ:>:ZHZYT:V@ZI2YT:V@ZI2YT:V@ZI2YT0=a5`002DeB
ME9eDWEBC@020006CGE=MDee10000dee00050003LHYa030020L000H7;`000g40L@0M000ORg6;LH]a
Rg6;LH]aRg6;LH]aRg6;LH][T76:JhY[R`04Q003LI2500B:38@6RP0PT8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@RY1[LF/H0008DTeBCE9=DTd20003DTeB00==1@02C@D000AaL75[;`081`001PL^0004
L@1a61d002UaRg6;LH]aRg6;LI1aT76@RY2;T8^@Ri2;T8^@Ri2;T8^@Ri2;T8^@R`0<T00TZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@0g4K000=CE9eDWE=ME9eDW1=L0050003MDe=00H0
00=`L@00;P081`001PL_0003L@0001d005^;LH]aRg6;LH]aRg6@LI1aT76@LI1aT76@LI2:T8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@Si2:T8n@RY2:T8Z@RY2:T8Z@Si2?LF]a01d0
0Td01gE=DTeBCE800Td60004CE9=C@H000=`0000;@081`001PL^0003LAP001h002UaRg6;LH]bRg:@
LI1aT76@Ri2;T8^@Ri2;T8^@Ri2;T8^@Ri2;T8^@R`0<T00SZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:T00g4O000:CGEBME9eDWEBC@D000A`CGE=1P000g5`000]00P700061dl0042;LH]a
Rg6;LH]aT76@LI1aT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:T8Z@Ri2;T8^@Ri2;T8Z@RY2:T8Z@RY2:
0i0068n@T92?T8Z@RY2:T8Z@RY2:T8n@Sg5[LB4000aeCE9=DTeBCE80C@02C@03DTeB00H000=`L@00
;@081`001PL^0003L@0001h002MaRg6;LX]bT7:@LY1bT8^@Ri2;T8^@Ri2;T8^@Ri2;T8^@Ri2;T8/0
3Y000jUaL@0CL@0=ZI2YT:V@ZI2YT:V@Z@03LA@000==04d00dd20003C@0000@000mBME9eDWEBMDe`
CGE=MDd01`000g00000/00P700061bd000=a00007`00Fh]aRg6;LH]aRg6@LI1aT76@LI1aT76@LI1a
T8Z@RY2:T8Ya0740L@1a075[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[LF]aJg5[T8Z@Si2?
T8maJg404`000e9=0002DP04CGE=M@P000i=DTeBCE9=DTeBCE9=DSH020L000H7;@000aP0000O000O
LH]bRg:@LY1bT7:@LY2;T8^@Ri2;T8^@Ri2;T8^@R`05LBT027400jV@Z@03LA00015=04eBME9eDWEB
ME9eDTd0L004000<DWEBME9eDWE=MDee0Tde00P700061bd000=a00007`007h]aRg6;LH]aRg6@LI1a
T76@LI1aT76@LI2:T01a0740<`002G5[LF]aL75[L@0C00EB00]=MDeeCGE=MDe`C@04000;DTe=CE9=
DTeBCE80=P081`001PL/0003L@00020001MbRg:@LY1bT8^@LY1bT8^@Ri2;T8^@R`06LCT0274=000D
C@1=DWEBME9eDWEBME9eDWEBMDd5000:CGEBME9eCGD0CCD020L000H7C`006H]aRg6;LH]aRg6@LI1a
T76@LI00L@1a0740?P002g1aL91`LG1aJg5`00/01U803TeeCGE=MDeeCE9=DTeB1@001Td0DTeBCCL0
20L000H7C`002g:@LY1bT7:@LY2;00MaA@003HYaLI2YT:V@ZI2YT:T01G402XYa001eDWEBME82C@0=
M@1=CGE=ME9eDWEBM@02C@D00Tdh00P700061dl000N;LH]aRg6@009a00@0L@1aBP001V]aJi2:T0Af
01B?T8n@Si2?T6]aJg5BMDeBCGD0C@D000UBCDe=DTeBCE800Td00`1=000k00P700061e0000V@LW5/
LFaaK740C@0078YaRXn@SgEfSgJ?MXmfSi2YT:V@ZG5fL7EBME89009=00YBME9eDWEBM@1=>`081`00
1PMA0003L@1a058000I[LG2@MY06DPIf00YeT8n@RWEBMDeB2`002Td0DTeBCE9=DTdk00P700061dl0
00=/LF`0E00078YaRXn@DWEBME9eDWEBME9eMXmfSgJYSgIBME8<000:C@1`DWEBME9eDSX020L000H7
C`000g40001D0006L75[T7Ie3E82M@09SgF?DWEBMDee00`00dd01E9=DTeB00h000=`0000:@081`00
1PM<0004LFaaLEH001V:LI2?T59eDWEBME9eDWEBME9eDWEBSgF?009e00=BMG400W4<0006MDeeDWEB
3P000hl0000Y00P700061d/000Ea08]aR`1F0008L75fT7IeDWD;DP07ME9eDWEBS`03DP05MWEaL740
300014eBCE8>0003L00002T020L000H7B@0026aaK91bT76@E@007XYaZI2YMGIeMWEfMGIBME9eDWEB
ME9eDWEBME9eMP:?00BYLG5a2P001Dd0ME9`00d000>?0000:@081`001PM7000:L@1aLH]aRg6;LED0
011`LHnYSjV?T8n@DWEBME9e0e802GEBME9eDWEBM@04DP06Si2?LG1a2P0014eBCDd<0003L00002T0
20L000H7A@0036aaK8]bRg:@LY1aT5D001>:LJV@ZI2YT:V@ZI2YT7IeMWEf009e011BME9eDWEBME9e
DWIeZHn@0W490004ME9eDP`000>?L000:@081`001PM3000>L@1bLH]aRg6;LH]aRg5D00=a0i004:V@
ZHnYSjV?ZHnYSi2?T8l3M@EB00ieDWEBME9fMGJ?T71aL0H000ABCE9=2`000g1a000Y00P700061d40
010HLAR;LH]aT76@LY1bT7:;E003L@0TZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YMGIeMU9eDWEBME9eDXmf
0Xl00i2?L@02L@@000A=ME9e2`000hY`000Y00P700061cl001=a075aLW6;LH]aRg6;LI1aRg6;05<0
0g408i2YT:V@ZHnYSjV?ZHnYSjV?T8n@Si2?ME9eDWEBME9eDWEB009e00efSi2?T71aL01=DTeB00`0
00=`L@00:@081`001PLm000E674HLW5bLH]aRg6@LY1bT7:@LY1b05<00g406jV@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@MP02M@08DWEBME9eDWD2MP03MI2?00:?00MaL7E=ME9=00/000>?LG40:@081`00
1PLk000GL@1aLG9aLW5bLG9aRg6;LH]aRg6;LH/0D`03L@0@T:V@ZI2YT:V@ZI2YSjV?Z@:@01:?T8n@
Si2?T8meDWEBME9eDWD3DP0;ME9eMXmBLDeBCE8030000g1aL00Y00P700061cT001THLAQbLG9aLW5b
LH]aRg:@LY1bT7:@LY1b05<00g40::V@ZI2YT:j@ZI2^T:V@[Y2YT:V@ZI2YT:V@ZI2YMGIeMU9eDWEB
ME83M@08DWIeME9e04d;0003Sg1a02T020L000H7>@02L@0GLW5bLG9aLW5bLH]aRg6;LH]aT76;LH/0
D`03L@0@T:V@ZI2YT:V@ZI2YT:V?Z@B@012?T8n@Si2?T8n@SgEBME9e2U800deBC@0=0003L75`02T0
20L000H7<0000f/000060009LG9aLW5bLG9a009b00j;LI1bT7:@LY1bT7:@LU800g40=92YT:V@ZI2^
T:j@[Y2^T:j@ZI2YT:V@ZI2YT:V@ZHnYSjUeMWEfDWEBME9eDWEBME9f04d;0003Sg5a02T020L000H7
<@000f/00005000JLW5bLG9aLW5bLG9aLW6;LH]aT76@LI1aRg5A00=a016YT:V@ZI2YT:V@ZI2YT:V@
Z@04T00BSi2?T8n@Si2?T8n@Si2?ME9e2E800gI=L@0<0003L75`02T020L000H7;`04J`D00W401W9a
LW5bL@Ab00n;LI1bT7:@LY1bT7:@LY00D@03L@0XT:V@ZI2YT:j@[Y2^T:j@[Y2YT:V@ZI2YT:V@Si2Y
T:V?ZHnYSjUeMP9e00EBME9eDP02M@03T71a00/000B?LG5`:0081`001PL]000:J`1[06/0J`1[009a
01YbLG9aLW5bLG9aLW5bLG9aLW6;LH]aT76@LE000g401Y2YT:V@Z@b@00B?T7J?0WH0159eDXl5DP0C
SgEfDY2?T8n@DWE=DTeBCE9aS`02L@X000A`LG1a:0081`001PLX0003L@0000802F/07W9aLW5bLG9a
LW6;LH]bRg:;LY1bT7:@LY1bT7:@LTl002EaRW6^T:j@[Y2^T:j@[Y2^T:j@[Y2YT7YfU59jDWYBME9e
DWEB009e00UfMGF?ZHn@SgH00WD00e9eDP02S`04T8maL0L000EaSg5aL00X00P700061bd002M[06/0
J`1[06]/LFabK79aLW5bLG9aLW5bLG9aLW6;LH]aT76@LI00C`03L@06T:V@ZI2Y39002WF@DWEBME9e
DWD9DP0AME9BSi2?T8n@CE9=DW6?LHl00W470005Sg1aL740:0081`001PLX0004L@0000Y[0W40779/
LW5bLG9aRg6;LX]bT7:@LY1bT7:@LY1bT79?000bLHZ@[Y2^T:j@[Y2^T:j@[Y2^T:j@U59jDWYBNU9j
DWEBME9eDWEBME9eDWEBZHn@Si02M@04DXn?T0>?00=a000010001G6?LG5`02P020L000H7;@0026/0
J`1[06/016/01W5[LV]aJ`Aa015bLG9aLW5bLH]aRg6@LI1aT01?009a00NYT:V@ZI2Y00b@00YBME9e
DWEBME9e458038n@Si1=DW6?LHmaS`9a1@001Hm`LG1a02P020L000H7:0001W40001[00Y[00=aJg40
0f/05g5[Rg6;LH]bT7:@LY1bT7:@LY1bT7:@04h003QaRY2^T:j@[Y2^T:j@[Y2^T:j@[Y1kDWYBNU9j
DWYBME9eDWEBME9eDWEBME9eDWF?T8n@MI2?T0F?00=a00000P001G6?LG5`02P020L000H7<@0016/0
J`0;J`0EK6][LW5bLG9aLW6;LH]aT76@LY1a04h00W401jV@ZI2YT:T02Y0038n@DWEBME9eDWEBM@9B
00=fDU8035803GF:T8Z@RXmaSg6?LHl00W420006LHm`LG1a:0081`001PLX0003L@0000D000=/Jg40
3F/05HA[Q6^:LI1bRg:@LY1bT7:@LY1bT01>000OLJn@[Y2^T:j@[Y2^T:j@[Y2^T9EBNU9jDWYBNXnD
DP02T0:?00bYDXmfSe9eDWEBME82M@09MXn@Si2?T8n@00B?00I`002?LHl2L@03L00002H020L000H7
;`001aP0LAQ[67400f/00g5[J`08J`0BK6][LW5bLG9aRg:@LI1bT76@C@001g6@ZI2YT:T039005gIB
DWEBME9eDXmfT7J@Si2?T8n@MXmf00QB00efRY2:Sg6?LHmaSg6?008000=`Sg000Td00g40000V00P7
00061bP000=a00001@02L@04LW5bK09[00E/LG5[L@03J`0FQ6^4JhA[Q6b:LY1bT7:@LY1bT7:@LTd0
02Ia[i2_T:j@[Y2^T:j@[Y2^T9F@UE9jDWYBU5:^T:j@ZI2YT:V@Z@B@00R?T59eDWEBM@9f00J?T8n@
Si05S`08LG2?LHm=MG0X00P700061bT000=a00000`001g4HLAQa67400f/00g9/L@0;J`04LF]aK09a
0W801i1aT7:@LI00C@001G6@ZI2Y00>@00>YT9001Y000gJ@MP02DP04ME9eM@:?00f@Si2?T8n@Si2?
T8n@00:?00=fSgH01E80372?RXmaSg6?LHmaS`9a00=`Sg000Td00g40000V00P700061bP000=a6000
0`02L@05LW5bLG800W42J`=a0W800g5bJ`03J`0CQ6^4JhA[RV^:LHYbT7:@LY1bT01<000XLJj@[Y2^
T:j@[Y2^T:j@[Y2ESgYBNU9jMXn@ZI2YT:j@ZI2YT:V@Z@J@0WH0359eDWEBT8n@Si2?T0:?00IaS`2?
LGD3CBP020L000H7:@006G40001a674HLAQb6780L@1aK79aLW5bK7401F/048A[Q6^4JhY[Rg:@LY1b
T79<0003LI2@00N@00>YT9001Y000e9eDP02DP9f1I004Xn@Si2?T8n@Si2?T8n@Si1eMPAB00YaT76?
LHmaSg6?0`000hm=C@02CBP020L000H7:0000g4H00020008CW5>LW5bLG82L@=[0W44LP9a01EbJhA[
Q6^4JhA[Q6^4Ji1bT7:@LY00B`02L@0U[Y2^T:j@[Y2^T:j@[Y2^T:j@NU9jDWYfU92YT:V@[Y2^T:V@
Z@08T00>Si1eME9eDY2?T8n@Si02S`03MP00009a00AeCDe=:0081`001PLY000>L@00074HLAQa678H
LQP3J`08K75aLW5bLG82L@E[00n4JhA[Q6^4LY1bT7:@LY00BP02LA6@00=eDWD00e800gIBT004T00B
Si2?T8n@Si2?T8n@Si2?T8n@15802Tf@LI1aSg6?LHl30003MDe=009=:0081`001PLY0004600000=a
00MbLG9aLW5b00=[00=aK7801W805G6@JhY[Q6^4JhY[RW:@LY1bT7:@LP1:009a01R_T:j@[Y2^T:j@
[Y2^T:j@[WEkDWYBNU82T00=ZI2YT:j@ZI2YT:V@Z@06T00>Si2?MU9eDWEfT8n@Si02S`09L@00071e
CG5`02P020L000H7:@004740001a674HLAQb678HJ`02J`0MLFabLG9aLW5bLF]/Q6^:K8Y[T7:@LY1b
T7:@LY00BP02LA6@00>?MWD01586T00DSi2?T8n@Si2?T8n@Si2?T8n@MGH2DP0:CHmaT76?LHmaS`<0
00EBCDe`L@0X00P700061bT000lHL@1aCW5>LW5bLG9aLW400f/00faaLP06LP0DT6b:LXYaRW:@LY1b
T7:@LY1bT79:009a02F_T:n@[Y2^T:j@[Y2^T:j@[WJEDWYBNU:DT:V@ZI2^T:j@ZI2Y00R@00j?T8n@
DWEBME:@Si2?T0:?00If0000CGD2L@03L00002H020L000H7:@004740001a674HLAQb678HLP03J`0M
K79aLW5bLG9aLW5bK91aT76@LY1bT7:@LY1bT780B@02LA:@00=BME800e82MPB@01B?T8n@Si2?T8n@
Si2?T8n@Si2?MP9B00Y=Sg6@LHmaSg6?0`001E9=LG1a02P020L000H7:@000aQa0003L@09CW9aLW5b
LG9a00A[27805I1bT7:@LY1bT7:@LY1bT7:@LY1bT019000MMY2_T:n@[i2^T:j@[Y2^T:j@[Y1jDWYB
NU9jSjh00i001Jj@[Y2^00Z@00V?T8n@MWEBME800Xl01I2?T8n@00<000=`CGD00W400g00000V00P7
00061bT0011a000HLAQa678HLQQb674016/0779aLW5bLG9aLW5gLI1aT7:@LY1bT7:@LY1bT7990003
LI2@016@00=fSe8015800hmfT002T00DSi2?T8n@Si2?T8n@Si2?T8n@SgD2DP09CE9aT76?LHma00@0
00ABL75`:@081`001PLY0005674HLAP00W402G9aLW5bLG9aK003J`03LFab00=b01N@LWMbT7:@LY1b
T7:@LY1bT7:@LY1bT00S0005600001P03@001AP0001>00<000=>00002002L@0PT:n@[i2_T:n@[i2^
T:j@[Y2^T:ifNU9jDWYBNWJDT:h<T004Si2?T0:?00=fME800WD01WJ?T8n@S`@000EeCG5aMP0Y00P7
00061bT0019a0740LAQa674HLAQb6780J`02J`0<LW5bLG9aLW5bLG9a178037MbT7:@LY1bT7:@LRH0
00>AT9403I002I6@TI2AT96@UP09009a59001GIBDU9e00=B01QfSi2?T8n@Si2?T8n@Si2?T8n@MY1a
Sdd3DP06LHmaSg6?1@001580LG0Y00P700061bT000DHL@006002L@07LW5bLG9aLP02L@=[00=aLW80
0g805gMbMg9gLY1bT7:@LY1bT7:@LY1bT7:@01<000<H00001@00:I6@TI2AT96@TI2AT96@UY2AT9J@
UY2_T:n@[i2_T:n@[i2_T:n@[i2_00@00W408Y2_T:n@[i2_T:n@[Y2^T:j@[Y2^T7YBNU9jDWYBNY2E
T9D:T00ASi2?T8n@DWEBME:@Si2?T8l0100017E=L74Z00P700061bd000ia674HLAQa678HLP1[009[
00YbLG9aLW5bLG9a27802i1bT7:@LY1bT7:@01D02Y002I6@TI2AT96@T@0=T009TI2AT96@TI2A00R@
0g4ET003MXmB00QB01b@Si2?T8n@Si2?T8n@Si1fT7IBCE9=Sg6?LHma1@0014eB074Z00P700061bX0
00Aa00000W40279aLW5bLG9a16/2L@Ab01QgLWMbT7:@LY1bT7:@LY1bT7:@LY1bTG8A000eT96@TI2A
T96@TI2AT96@TI2AT96@TI2FT96@UY2FT9J@[i2_T:n@[i2_T:n@[i2_T:n@[i000W46000O[Y2^T:j@
UI2ET:j@[Y2^T:j?NU9jDWYBNU9kMXn@U@08T00;Si2?T8meDWEBMGH00Xl00i2?0005009=;0081`00
1PL^000K674HLAQa678HJ`1[06]>LW5bLG9aLW5bLG9a00Ab00igLY1bT7:@LY1bT7:@LP/000]bT7:@
LY1bT7N@M`08T009TI2AT96@TI2A00f@00VAT96@TI2AT9402902L@/03I000hmBDP06DP0AMWEfMY2?
T8n@Si2?T8n@MY000U8024eBMGIaSg6?1P03CB/020L000H7;P02L@07LW5bLG9aK004J`03K75b00Ab
01UgLWMbMg:@LY1bT7:@LY1bT7:@LY1bTG:@00L003fAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT9J@TI2FT9J@UY2FT:n@[i2_T:n@[i2_T:n@[i2_00=a3`000i2ET002T00@[Y2^T9EfU59jDWYB
NU9jDP9f00>?T90019002Hn@Si1BME9eDP02MP03T8n@00L000=eCDd0:`081`001PL_000HLAQa674H
J`1[06/0JdibLG9aLW5bLG9a27808Y1bT7:@LY1bT79gLWL0T7:@LY1bT7:@LY1bT7:@LY1gT7L8T009
TI2AT96@TI2A00f@00VAT96@TI2AT9401i02LA<029000gEBDP09DP=f00F@Si2?T002MP0;ME9BDTeB
CHmaSg4020000dd0000Z00P700061bh00W400g9[J`08J`03LG9a00Ab05agLWMbMg:@LY1bT7:@LY1b
T7:@LY1bT7:@096@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@UY2FT9J@
[i2_T:n@[i2_T:n@[i2_T0=a5P004i2ET:j@[Y1jDWYBNU9jDWYBME800WH3T00=MI1BME9eDWEBMXn@
S`080003L4d002/020L000H7;`005W4HJ`1[06/0J`1[CW9>LW5bLG9aLW48LP0<Mg:@LY1bT7:@LI1b
0`005G:@LY1bT7:@LY1bT7:@LY1bT7N@M`08T009TI2AT96@TI2A00f@00VAT96@TI2AT9401Y03LAL0
00>?T8l01I000hmBDP0:DP03MGIe00IB00I=MGJ?LHl50004C@00CBd020L000H7;`04L@03Jfa[009[
00A/LG9a1W805GMbMg9gLY1bT7:@LY1bT7:@LY1bT0060011T96@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TI2AT96@TI2AT9J@TI2AT9J@UY2FT:n@[i2_T:n@[i2_T:n@[i000W4K0003T9F@00:@00b^T7YB
NU9jDWYBME83M@0<DWEBME9eDWEBT8n@1@000deB000_00P700061bl001ma674HLQQb678HLTibLG9a
LW5bLG9aLW5gLGMbMg:@00=b00=gLP002P00=W:@LY1bT7:@LY1bT7N@Mi1bT7NAMi5gTGNAT96@TI2A
T96@TI2AT96@TI2AT96@TI2AT96@T@F@00>AT9000Y02LAl000R@Si2?T8n@S`iB00F?MI1aT0060003
CE9=02l020L000H7<000375>LTibLG9aLW5bL@Qb00egLWMbT7:@LY1bT7:@00d003nAMi5gTGNAT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2FT9J@UY2FT9J@UY2FT9J@[i2_T:l00W4S00B@
012ET7YBNU9eDWEBME9eDWEf0Xl00i2?00050003MDe`02l020L000H7<0007aQa674HLQQb679>LW5b
LG9aLW5bLG9aMg9gLWMaLW403`00>7:@LY1bT7:@LY1bT7:@Mi1bT7N@Mi5gTGNAT96@TI2AT96@TI2A
T96@TI2AT96@TI2AT96@TI2A0i00196@TI02LB@000J@Si1fT8l8DP08ME9fMY1fT7480003DTd002h0
20L000H7<000375>LTibLG9aLW5bL@Qb00UgLWMbT79gLWL04P00?WNAMi5gTI2AT96@TI2AT96@TI2A
T96@TI2AT96@TI2AT96@TI2AT9J@UY2FT9J@UY2FT9J@UY2FT:n@[i2_0W4W00:@016ET91BMU9eDWIe
T7J@Si2?T007009=00=B0000;@081`001PLa008H01Qa678HLQQbCW9>LW5bLG9aLW5bLG9aLW4D000e
T7:@LY1bT7:@LY1gT7N@LY1gTGNAMi5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT9401I00196@
T902LBP0016@MY2?T59eDY1fT7J@MY1fT0090003DTe=02h020L000H7<P002W5>LTibLG9aLW48LP05
Mg9gLWL05`00?I5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2FT9J@UY2FT9J@
UY2FT9J@UY2FT9H00W4Y0003T8n@00>@00YfSi2?T8n@Si2?2002C@03M@0002d020L000H7<`005G4H
LQQb679>LTibLG9aLW5bLG9aLP0H000fLY1bT7:@LY1bT7:@Mi1bT7N@Mi5gTGNAMi6@TI2AT96@TI2A
T96@TI2AT96@TI2AT96@TI2A0i00196@TI02LBX000n@MY1fT8n@MY1fT7J@MY002@02CBl020L000H7
<P002W5>LW5bLG9aLW48LP03Mg9b01T003fAMi5gTI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2AT9J@UY2FT9J@UY2FT9J@UY2FT9J@UY2F009a;0000hn@T003T006Si2?T8n@2`000e80000]00P7
00061c<001=>674HLQQbCW9>LW5bLG9aLW5b01/003>@LY1bT7:@LY1gT7N@LY1bTGNAMi5gTGNAT96@
TI2AT96@TI2AT96@TI2AT96@TI2AT9401I00196@T902LBd000]fT7J@MY1fT7J@MP0:0003C@0002h0
20L000H7=0002W9>LW5bLG9aLW47LQ`003egTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TI2FT9J@UY2FT9J@UY2FT9J@UY2FT9Ia02l000V?T8n@Si2?T8l0?0081`001PLe008H00ea678H
LTibCW9>LTib01h0025bT7:@LY1bT7:@LY1gT7N@Mi1gT7N@Mi5gTGNAMi5gTGL039001i6@TI2AT940
1Y02LC0000MaT76@MY1a00`000==0000;P081`001PLf0008LW5bLG9aLW43LP03LG8001h003ZAT95g
TGNAMi6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT9J@UY2FT9J@UY2FT9J@UY2FT9J@0W4a0006
MY1fT8n@30000dd0000^00P700061cL000]a678HLQQbCW9>LP0Q000PT7:@LY1bT7:@LY1gT7N@Mi1g
T7N@Mi5gTGNAMi5gTGL<T009TI2AT96@TI2A00B@0W4c0003T76@03h020L000H7=`002W5bLG9aLW5b
LG8R000iMi5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@UY2FT9J@UY2FT9J@UY2F
T9J@009a=0000i00000;0003DP0002l020L000H7=`026007LQQb678HLP0T000OLY1bT7:@LY1bT7N@
Mi1gT7N@Mi1gTGNAMi5gTGNAM`0<T007TI2AT96@T@06T09a=0000g40000;0003C@0002l020L000H7
>002L@05LW5bLG809@00=i5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2FT9J@
UY2FT9J@UY2FT9H00W4d0003Si0000/000==0000<0081`001PLi0005LQQb67809P007g:@LY1bT7:@
LY1gT7N@Mi1gT7N@Mi5gTGNAMi5gTGL039001I6@TI2A00N@00=aCP00<P000gJ@L@0<0003C@000300
20L000H7>@001G5bLG9a02H003JAMi5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TI2FT9J@UY2FT9J@UY02LC@000B@MY1f2P000dd0000a00P700061cT000E>678HLP0W000TT7:@LY1b
T7:@LY1gT7N@Mi1gT7N@Mi5gTGNAMi5gT7N@Mi1g1I001GN@T95g00N@00=>L@00<P001I1aT76@00X0
00==0000<@081`001PLj0004LW5bLBL003AgTGNAMi5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TI2AT9J@UY2FT9J@UY2F0W4c0006T7J@MY1f2P000dd0000a00P700061c/000=b6000:0009G:@
LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gTGN@Mi5gT7N@Mi1gT7L00i002GN@Mi1gT7N@M`02T003LDh00340
00R@LI1aT76@LCd020L000H7?0000g40000X000aMi5gTGNAMi5gTGNAMi5gTI2AT96@TI2AT96@TI2A
T96@TI2AT96@TI2AT96@TI2AT002LC8000V@MY1fT7J@MY001`000dd0000c00P700061cd000<H0000
:0009G:@LY1bT7:@LY1bT7:@Mi1gT7N@Mi1gT7N@Mi5gT7NAMi1gTGL02i000dh0000a000:MY1fT76@
LHmaSc`020L000H7?0000g80000Y000aTGNAMi5gTGNAMi5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@
TI2AT96@TI2AT96@L@0b000;MY1fT7J@MY1fT7H01@000dd0000d00P700061c`000<HLP00:P009Y1b
T7:@LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7NAMi1gT7N@Mi1g2I0b000<MY1fT7J@LHmaSg5f?0081`00
1PLl0003LW5b02X002mgTGNAMi5gTGNAMi5gTGNAT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2A
T00a000>MY1fT7J@MY1fT7J?MXlk00P700061c`000@HLAQb:P00:g:@LY1bT7:@LY1bT7:@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7L00i0`000?MY1fT76@LI1aSg5fLGIa008000==0000=P081`00
1PLk0006CW9aLW5b:@00;Y5gTGNAMi5gTGNAMi5gTI2AT96@TI2AT96@TI2AT96@TI2AT96@TI2AT96@
TG4_000?MY1fT7J@MY1fT7J@MXma008000=`0000=`081`001PLl0006674HLAQb:@00;I1bT7:@LY1b
T7:@LY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gL@0_000?T7J@LI1aT76@LGIaMW5f03`0
20L000H7?000275>LW5bLG8H9`00;GNAMi5gTGNAMi5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@TI2A
T96@TI2AL@0^000=T7J@MY1fT7J@MY1fT002MSd020L000H7?@001g4HLAQbCW80:000;7:@LY1bT7:@
LY1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7Ma;@00491fT76@LI1aT75fLGIaMW4m00P7
00061ch000QbCW9>LW5b62L002]gTGNAMi5gTGNAMi5gTGNAMi5gTGNAT96@TI2AT96@TI2AT9J@TI2F
T9Ia02/00Y003gN@Mi1fT7J@MY1fT75fL@0n00P700061ch000PHLAQb679>LRP002YbT7:@LY1bT7:@
Mi1bT7:@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mg4Z000BMY1fT7J@LI1aMg5gLGIaMW5f?P081`00
1PLo0009CW9>LTibLG8H02H002RAMi5gTGNAMi5gTGNAMi5gTGNAMi6@TI2AT96@TI2AT96@TI2AT96@
0W4Y000CMi1gT7N@MY1fT7J@MY1aMg5fL@0n00P700061cl000Ua674HLQQbCW809P00:W:@LY1bT7:@
LY1bT7:@LY1bT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi0HLBP001=fT7J@MY1aT75gLGIaMW5fLGI=03l0
20L000H7@0002W5>LTibLG9aLQPT000GTGNAMi5gTGNAMi5gTGNAMi5gTGNAT9402g401i6@TI2AT740
:00057N@Mi1gT7J@MY1fT7J@LGMaMW5f?`081`001PM0000:674HLAQb679>LRD002V@LY1bT7:@LY1b
T1Qa674HLAQa674HLAQa675>LDia674HLAQa674HL@0V000DLI1aT76@LGMaMg5gLGIaMW5fCW5000P7
00061d0000aa675>LW5bLG9aLQPT0005TGNAMi4017400dia000@0007J`1[06/0J`020003LI5a00Ia
7P005I1gT7N@MY1fT7J@MY1fT7IgLGIaMP1100P700061d4000ea674HLAQbCW9>LP1>028000HHLAQa
67460006JhA[Q6^41`000f/0Q002J`0C06/0Q6^4Jg5gT7N@Mi1gLDiaCP0J000ET76@LI1aT75gLGMa
Mg5gLGIaMTia048020L000H7@P002W4HLG5bLG9aLW42LP03L@0001d00g47009[00F4JhA[Q002J`L0
0V/00hA[J`05009[00B4LI1g18@5T0=a5P004I1gT7N@MY1fT7J@MY1fT7Ig009f00=aMP00@P081`00
1PM3000<LAQa675>LTibCW9>70001g4HLG9gLWL02@02J`03Q6^400H00f/00h@0J`07000DQ6^4Ji1g
T6^4JhA[T7N@Mg5>LDhB000ET76@LI1aT76@LGMaMg5gLGIaMW5f04@020L000H7A002L@0:LW5bLG9a
LW5b61T000Y>LDj@Mi1[Q6^:2@02J`03Q6^400D000B4JhA[200028Y[Q6^@LI2:0X@0191gT7L3T003
Mi1g00Aa3P004I1gT7N@MY1fT7J@MY1fT7Ig009f00=aMP00A0081`001PM6000<674HLQQb679>LTib
4P001QQa67MbM`E[00EaJgMbM`09009[00>4Jf/00`000hA[Q009000LLGN@JhAgT7N5LY1gT7N@Mi1g
T6^4JhAgT7MaCPT001IfT7J@MY1aMg5gLGMaMg5gLGIaMW5fAP081`001PM7000=679>LTibCW9aLW5b
600@0003LG9g00=b00][LW5[Q6^@LY1gT009009[00B4Jf/00V/00hA[0008000KLI1gRV^@Mi1aT7N@
Mi1gT7N@Mi1aQ6^4Ji1g009a1`005WN@Mi1gT7N@MY1fT7J@LGMaMg5fLGI700P700061dP000`HLAQb
678HLTibCW82CPd000=aLWL00f/01GMbMg9g00=[00EaLWMbM`09009[00F4J`00Q002J`T001n@Mi1a
T7N@MhEgT7:@Mi1bT7N@MhA[Q764Ji1gL@1a00@001IfT76@LGMaMg5gLGMaMg5gLGIaMW5fB0081`00
1PM9000;CW9>LTibLG9aLW400g800aP000070004674HM`=[00MgLWMbMg9a00=[00F;LY1bM`070003
J`24009[0`000hA[0007000QLGN@Mi1bT7N;LI1gT7N@Mi1gT764Ji1gT7N@Mi1gT4ia008001EgT7N@
Mi1gT7J@MWMfMgIgLGMaMg40BP081`001PM:000>674HLAQbCW9>LTibCW89009a00=gJf/00V/01fag
LWMbMg8016/0175bMg8900=[0P03J`P0025>T7N@Mi1gT6^@LY1bT7:@LY1aQ6^@Mi1gT7N@Mi1fT000
0W40591aT76@LGMaMg5gLGMaMg5fLGIaB`081`001PM;000;CW9>LW5bLG9aLW400g800aP000030004
675bM`9[00aaJf][LWMbMg9gLW43J`04T7:@LPP000>4Jh@00`000hA[0007000OLGN@Mi1gT7N:Mi1g
T7N@Mi1[RWN@Mi1gT7N@Mi1gT002000ECW5gT7N@MY1fMgIgMWMfMg5gLGIa04`020L000H7C0003QQa
675>LTibCW9>LTib1@00275>MdigLWMa0f/0279gLWMbMg9g0f/017MbMg4700=[0`000f^40007000M
Mi1gT7N@Mi1bT79gLY1aMf^4Ji1gT7N@Mi1fT7H01P004Y1aT75gLGMaMg5gLGIaMW5fCTd020L000H7
CP00379>LW5bLG9aLW5bL@9b00aa074HMg5gLWMbMg42J`9a00QgLWMbMg9gLP9[00=gLWL020000f^4
000200=[1`00791gT7N@Mi1gRV^:Mi1aRV^4LI1gT7N@Mi1gT7L3000EJ`00LDj@MY1fMgIgMWMaMg5g
LGIa04h020L000H7C`006W4HLDibCW9>LTibCW9>LP1bCV]>Mg5gLGMb0f/02G5gLWMbMg9gLP02J`9b
2@02J`<00V/60008LGN@Mi1gT7L2L@03Mf][00A[00]aT76@Mi1gT7IgMP04009[0P004740Mg5gLGMa
Mg5fLGIaMTi?00P700061dl0015>LDibLG9aLW5bLG9aLQQa0002J`09Mg9gLWMbMg9g009[00ebMg9g
LWMbMg9aJg5b00P00f/20003Jh@000D000Y>T7N@Mi1gT6^@0WL03V^4JhAgT7N@Mi1gT7N@1@02J`0B
Q6/0CW5fT7IgMWMaMg5fLGIaD0081`001PM@000<674HLAQaCW9>LTib0Th3000;J`1bCW9aMg5gLGL0
0V/03G5gLGMbMg9gLWM[Mg402002J`800f/50005LG9gMi000WL0179gLGL3J`9a00AgLI1a0gL00g40
00040005J`1[06/00`002GMaMg5gLGIaMP02LE8020L000H7D`0034iaCW9>LW5bLG8H009[0`002f/0
LG9gLWMbMg9a009[00EbMg9gLP03M`04LF][M`H00V/200=[10000g5gM`07M`03LF][00Eg00>@Mi00
1`02J`P000QaMWMaMg5gLEH020L000H7E@001g4HLAQbCW800Th30003J`0000<000U[075>Mdi[LG80
0f/02gMbMg9gLWMaJg5g00H000A[00000V/50009Mg5gLGMaMdig009[00UaMg5gLGMaMg4010000f/0
00090007L@1gLGI>MP1H00P700061eD000PHLDibLG9aLP9a00<0J`000f/4000367M[009[0W803gMb
Jf]aLWMbMg9gLF]bM`0500=[0`000f/000030008Mg9gLWMbMg86M`03T7N@00D000=[00001@04J`06
04iaLGMaFP081`001PMH0006674HLDib0`001F/0001[00L000Mg079aMg5g009[00YaMg5gCWMaJ`1g
1P02J`<00V/60003L@1g00=a00IgLGMaMg4=00=[1002L@03MTi>05X020L000H7F@03L@04LW5a60P0
0f/3000ALG9gLGMbMfa[K7MbMg9gLWL00V/00g400005009[0`03J`D00V/A00A[1P000degC@1L00P7
00061e/000Ea675>60090003J`1[00D001=bCWM[MdiaJgM>MdigCWM>J`1b00T000=[06/00`000f/0
J`0A00=[1P0016/007IN00P700061ed000A>LG5a0P02J`L00f/40006CWMbLG9g0V/02G9gLWMbMg9g
L@02J`03Mg4000l000=[00001@000f/00004J`0306/000D000=[LDd0H0081`001PMO0003LAQa00D0
00=[00001`000f/00002000RCW9>LTig06]>MdigCW9>LTigCW40J`1[06/0JdigCV/0J`T00V/2009[
10001G40J`1>068020L000H7H@000g5b0005009[0`000f/00002009[10000g5gL@02L@03Mg5a009[
015bMg9gLWMbMg9gLWMaMg5gJ`0400U[0P02J`8000=>LF/00f/2LF<020L000H7H`000g40J`070003
J`0000<000E[0000J`05000GLTibCV]>LTibCW40L@1a079>MdigCW400`000f/0000:000:CV/0J`1a
CWH0CV@020L000H7J0001g5>LTib06/01@03J`h016/90005Mg5gLGL00f/01W5[LG5gJ`9a00=gLGL0
0g400gIaL@1W00P700061fT00QP01f/0K4ibCW800`000f/000030003J`1[00d000=[00000`005f]>
LTigCWM>MdigCWM>Mdig075>LDia06X020L000H7K0003G40LTibCW9>LTibCW801@001F/0J`1[00<0
00=[00000`0017MaLW42J`04Mg5gJ`9a00EgLGMaM`02L@04JgMaM`AaJ`081`001PM^000d674HLDia
CW80J`1aCW9>LTibCW9>LTibCW80LDibCV]>LTib06/0JdibCWM>Mdi[CW40Jg0020L000H7LP001g40
J`1[07400V/09DibLG9aLW5b06/0JdiaCWM>MaP0LG9aMg5gLGMaMg5gCW5>LDh0LP081`001PMg000C
K1QaCW5>LDibCW5>LTia06/0LP090009LDiaCW5>LDia07H020L000H7O0001g40J`1[07401@000f/0
00090003J`0007/020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7003o1b47
003o1b47003o1b47003o1b47003o1b47003o1b470000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Textures on ",
  StyleBox["MeshGraphics3D[]",
    FontFamily->"Courier"],
  " Objects"
}], "Subsubsection",
  CellTags->"MathLinkMode"],

Cell[TextData[{
  StyleBox["MeshGraphics3D[]",
    FontFamily->"Courier"],
  " and ",
  StyleBox["SurfaceGraphics[] ",
    FontFamily->"Courier"],
  "objects offer a natural mapping by the two parameters used to create the \
surface. This natural mapping is activated with the option ",
  StyleBox["MVTextureMapType\[Rule]MVMeshUVMapping",
    FontFamily->"Courier"],
  " or for the transposed bitmap image with  ",
  StyleBox["MVTextureMapType\[Rule]MVMeshVUMapping. ",
    FontFamily->"Courier"],
  StyleBox["The ",
    FontFamily->"Times New Roman"],
  StyleBox["MVSphereMapping",
    FontFamily->"Courier"],
  StyleBox[", ",
    FontFamily->"Times New Roman"],
  StyleBox["MVCylinderMapping",
    FontFamily->"Courier"],
  StyleBox[" and the ",
    FontFamily->"Times New Roman"],
  StyleBox["MVTorusMapping",
    FontFamily->"Courier"],
  StyleBox[" may not work correctly with ",
    FontFamily->"Times New Roman"],
  StyleBox["MeshGraphics3D[]",
    FontFamily->"Courier"],
  StyleBox[" objects.",
    FontFamily->"Times New Roman"]
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(blankknot = 
      ParametricPlot3D[\ 
        Evaluate[
          FrenetFrame[{Cos[t] + 2\ Cos[3\ t], Sin[t] + 2\ Sin[3\ t], 
              Sin[2\ t]}, t, phi, 0.5]], {t, 0, 2  Pi}, {phi, 0, 2  Pi}, 
        PlotPoints -> {100, 18}, DisplayFunction -> Identity]\)], "Input",
  CellLabel->"In[140]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \("-MeshGraphics3D-"\)], "Output",
  CellLabel->"Out[140]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; 
    MVShow3D[blankknot, MVTexture -> "\<`IslamPattern2`\>", 
      MVTextureMapType -> MVMeshUVMapping, MVScaleTexture -> {10, 2}]\)], 
  "Input",
  CellLabel->"In[16]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVMesh3D[2] \[DoubleLeftArrow] MVTexture2D[4], 
      MVMeshUVMapping, {256, 256, MVRGBTexture}}\)], "Output",
  CellLabel->"Out[16]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[17]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000006060005F5F006060005F5F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000030402F3F4F2F
3F4020000000005F5F006060005F5F00000000605F00606000605F00606000605F00606000605F\
00606000605F006060000000006060005F5F006060005F5F006060005F5F006060
000000500010000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000005F5F006060\
005F5F500010000000006060005F5F006060005F5F006060006060006060006060
006F600060600000006000106F0010006F6F006F6F006F6F006F6F006F6F006F6F006F60006F6F\
006F60006F6F006F60006F6F006F60006F60006060006F60006F60006F60006060
006F60006060006F6060000F6F0010006060006F6000605F00606000605F006060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000005F5F006060005F5F006060005F60\
0060605F000F60001060000F00000000000000606F006060006F6F006060000000
6000106F00106F000F6F00106F00107000106F00100000006F0010700010000000000000006F6F\
007070006F6F00706F006F6F00706F006F6F00706F006F6F00706F006F60006F6F
006F60006F6F006F60006F6F006F60006F6F006F600000006F000F6F00106F000F006F6000605F\
006060005F5F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000005F5F00606000606000606F006060006F6F00606F006F6F\
0000006F00106000106F00106000106F00106F00106F0010006F6F0000006F0010
7000106F001070001F700010000000007070007F7F00707F007F7F007070007F7F007070007F7F\
007070007F7F007070007F70007070000000000000700010700010700010700010
70001070001070001070001070001070001070001070001000000000706F0000006F0010700010\
6F00107000100000006000105F000F006060000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000005F60005F5F006060005F6000606F00606F006F6F0000006000106000106F0010000000\
6F00106F00106F00106F0010000000006F7000707F00707000707F007070007F7F
00707F007F7F000000000000007F7F007F80007F7F008080007F7F007F80007F7F007F80007F7F\
007F7F007F7F007F7F007F7F007F7F007F7F007F7F007F70007F7F007F70007F7F
007F70007F7F007F70007F7F0070700000007000107F00107000107000107000107000106F000F\
5F70306F00107000106F000F6F001000000000000060000F006060005F5F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000005F5F006060005F60
00606000606000606F00606F006F70006F6F006F70006F7000707F006F7000707F007070000000\
6F001000000000707F007F7F00707F007F80007F7F007F80007F7F007F80007F7F
008080007F800080800000007F001F7F0010000000008080008080008080008080008080008080\
00808000808000808000808000808000808000808000808000808000808000807F
00808000807F00808000807F0000000000007F001F7F00100000000000007F00107000107F0010\
7000107F0010201F00907F1F8F70103F2F0F0000005F70308F6F108F7010706F1F
6F001000605F006060000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000005F5F005F60005F60006060005F6000606F
00000060001000606F006F70006F7000707F00707000707F00707F007F7F00707F007F7F000000\
007F8000707F007F80007F7F007F80007F7F007F80007F80008080007F80008080
007F8000808F007F800000007F00107F001F7F00107F001000000000808F00808000808F008080\
00808F00808000808000808000808F0080800080800080800000007F0010800010
7F001080001F7F00108000107F001080001F7F00108000107F00108000107F00107F00107F0010\
000000907F10907F10907F10907F108F7010907F108F70109070108F70108F7010
8F6F108F7010806F108F6F10201F00006F60005F5F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000500010000000005F6000606F00606000606F5F0010600010006F70
006F70006F7000707F00707F00707F00707F007F80007F80007F8000000070001F700010000000\
007F8000808F007F8000808F007F8000808F00808000808F00808000808F00808F
00808F00808F008F8F00000080001F7F001080001F7F001080001F7F001080001F7F001080001F\
7F001080001F000000000000008F8F008F8F008F8F008F8F008F8F008F8F008F8F
008F8F008F80008F8F008F8F008F8F008F80008F8F008F8000000080001080001F80001080001F\
7F001080001F7F00108000100000009F801F907F109F7F1F907F10907F1F907F10
907F10907010907F1F8F70109070108F6F108F70108F6F106F001060000F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000050000F\
50001000000000606F0000005F00105F001060001000606F006F70006F7000707F
006F7F00707F00707F007F8000707F007F80007F80007F80700010700010700010000000007F80\
00808F007F8000808F007F8000808F00808F000000000000000000000000008F8F
000000008F8F00000080001F7F001080001F7F001080001F7F001080001F7F001080001F000000\
008F8F008F8F008F90008F8F008F8F008F8F008F8F008F8F008F8F008F8F008F8F
008F8F008F8F008F8F008F8F008F8F008F8F008F8F008F8F008F80008F8F008F80008F8F008080\
00000080001080001F7F001010100070802F8F80209F80109F801F9F7F109F801F
907F10907F10907F10907F108F7010907F108F70109070108F70108F70108F6F106F0010000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000500010500010000000\
5F00106000105F0010600010006070006F70006F7000707F00707F00707F00707F
007F80007080007F80007F80007F8F00000070001F7000107F001F7000107F001F00000000808F\
00808F00808F0000005F6F405F603F5F6F405F603F5F6F405F6F3F5F6F405F6F3F
5F6F405F6F3F000000008F9000909F008F9000909000000000000080001F80001F80001080001F\
80001F009090008F90009090008F90009090008F90009090009090009090008F90
009090009090009090008F8F00909000908F009090008F8F009090008F8F00908F008F8F00908F\
008F8F008F8F00000080001F80001080001F8000101F100F608F40A08F1F9F8010
A08F1F9F80109F801F9F80109F801F907F109F7F1F907F109F7F1F907010907F108F70108F7010\
606F20006060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000050000F50001050000F5F00105F0010\
60001000606F006F70006F7000707F006F7000707F00707F00708000707F007F80
00000070001F6F001070001070001070001F70001070001F7000107F001F70001000000050603F\
5F6F4050603F5F6F4050603F5F6F4050603F5F6F405F603F5F6F405F6F3F5F6F40
5F6F3F5F6F4000000000909F008F9000909F008F9000909F00909000909F008F9000909F009090\
00000080001080001F00000000909F009090009090009090009090009090009090
00909000909000909000909000909000909000909000909000908F009090008F8F00908F000000\
00000080001080001F80001080001F80001080001080001080001F800010800010
8000100000009F8010A08F1F9F8010A0801F9F80109F80109F7F109F8010907F10907F108F7010\
9070108F6F108F6F1000605F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F602050001050602F5F0010600010600010600010\
6000106F001F6F00106F001F6F00106F001F6F0010007F80007080007F8F007F80
007F8F007F8F00808F7000107F001F700010000000000000008F90008090008F900000005F6F40\
5F6F3F5F6F405F6F405F6F405F6F405F6F405F6F405F6F405F6F406070405F6F40
00000000909F00909F00909F009F9F00909F009FA000909F009F9F00909F009F9F00909F009F9F\
00909F009F9F00909F009F9F00000000000000909F009F9F00909F009F9F00909F
009F9F009090009F9F00909F009F9F009090009F9F009090009F9F009090009F900000008F001F\
8000108F001F8000108F001F8000108F001F8000108F001F8000108F001F800010
8F001F8000106F90408F8F20A08F1FA08F10A08F1FA08F10A08F1F9F80109F801F9F7F109F801F\
907F10907F1F9070109070108F6F106F001000605F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F5F2F1F1F0F6F5F105F0010706010600010600010600010600010\
6F001060001000708000707F007080007080007F80007080007F8F007F8000808F
007F8F00808F7000107F001F00808F008F90008090008F90008090008F90008F90000000000000\
0000005F6F405F6F405F6F405F6F405F6F405F70405F6F4000000000909F0090A0
00909F009FA000909F009FA000909F009FA0009F9F009FA000909F009FA0009F9F009FA000909F\
009FA0009F9F009FA000909F009F9F009F9F0000008000108F001F000000009F9F
00909F009F9F009F9F009F9F00909F009F9F009F9F009F9F009090009F9F0000008F001F800010\
8F001F8000108F001F8000108F001F8000108F001F8000108F001F8000108F001F
80001080001F000000AF901FA08F10A08F1F7F8F307F8F30A08F10A08F1FA08010A08F1F9F8010\
9F8010907F10907F108F70109070108F70108F7010000000006060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000605F106F60106F601070601F5F001060001F6000106F001F006F7F007080\
007080007F80007080007F8F007F80007F8F007F8F00808F007F8F00809000808F
7F001F70001F008F90008090008F9F008F9F008F9F008F9F00909F008F9F00909F008F9F0090A0\
00909F0090A000909F009FA000000060704F5F704060704F607040000000000000
009FAF009FA000A0AF009FA000A0AF009FA000A0AF009FA000A0AF009FA000A0AF009FA000A0AF\
009FA000A0AF009FA000A0AF009FA000A0A0009FA000A0A08F001F8F001F8F001F
90001F8F001F0000000000000000008F001F90001F000000000000000000009F9F009F9F009F9F\
009F9F009F9F009F90009F9F009F90009F90009090009F90009090009F9000908F
009F9000908F00908F0000008F001F8F00108F001F8000108F001F800010809030A08F10A08F1F\
A08010A08F1F9F80109F801F907F10907F10907010907F108F70108F701060000F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000605F10605F106F60106F60106000105F0010600010006F7F007080006F7F007080007080\
007F80007080007F8F007F80007F8F007F8F00808F007F8F008090000000000000
008090008F9F008090008F9F008F9F00909F008F9F00909F00909F0090A000909F0090A00090A0\
00000000000060704F5F704060704F5F704060704F607040607F4F607040607F4F
607040607F4F607F40607F4F607F4F00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF\
00A0AF00A0A000A0AF00A0AF00A0AF00A0A000A0AF00A0A000A0AF00000090001F
8F001F90001F8F001F90001F8F001F90001F00000000A0A0009FA000A0A0009F9F009F9F009F9F\
009F9F009F9F009F9F009F9F009F9F009090009F9F009090009F90009090009F90
009090009F9000908F0090900000008F001F8000108F001F8000108F001F8000102F200FA08F10\
A08F1FA08F10A08F1F9F80109F801F907F109F7F10907F10907F108F70105F7030
302F006F0010005F5F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000605F10
6F60106F5F1070601F7060107F6F1F000000000000006F7F007080007080007F8F007080007F8F\
007F8F00808F007F8F008090007F8F008090008090008F909F8090008F9F008F90
008F9F008F9F0090A000909F0090A00090A0009FA00090A0009FA00090A0009FAF000000607F4F\
60704F607F4F60704F607F4F60704F607F4F607F4F607F4F607F4F6F804F607F4F
6F804F607F4F6F8050607F4F6F8050607F4F00000000AFB000AFB000A0B000AFB000AFB000AFB0\
00A0AF0000006F7F4F00000000000000000000000000AFB000A0AF00AFB0000000
90001F90001F90002090001F90001F90001F90002090001F00000000A0A000AFAF00A0A000A0A0\
00A0A000A0A0009F9F00A0A0009F9F00A0A0009F9F00A09F009F9F00A09F009F90
009F9F009F90009F9F009F900000008F001090001F8F001F8F001F8F00108F001F8F00108F001F\
7F8F30AF901F908F20809030A08F10A08F1F9F80109F801F9F7F109F801F907F10
907F107000107000106F0010600010000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000605F10605F106F6010
6F5F10706010706010000000600010600010600010007080007080007F8F007080007F8F007F8F\
007F8F007F8F008090007F8F0080900080907F001F000000008F9F008F9F00909F
008F9F0090A00090A0009FA00090A0009FAF009FA0009FAF5F7040607F4F5F704F607F4F5F704F\
607F4F607F4F607F4F607F4F607F4F607F4F60804F607F4F6F804F607F4F6F8050
607F4F6F804F60804F6F805060804F6F80506F804F6F805000000000AFBF00AFB000AFBF000000\
0000006F804F6F80506F804F6F804F6F804F6F80506F804F6F804F6F804F6F8050
00000000AFB000AFB000AFB000AFB000000000000090001F90001F90001F00000000AFAF00A0A0\
00A0AF00A0A000A0A000A0A000A0A000A09F00A0A0009F9F00A09F009F9F00A09F
009F90009F9F009F90009F9F009F900000008F00108F001F8F00108F001F8F00108F001F800010\
0000000000008F001F0000008000108000106F8F409F8010A08F109F80109F8010
907F10907F109070107000106F000F700010006F5F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F5F10606010605F106F60106F6010
706F1F5F00106000106000106F001F6000106F001F6F0010007F8F007F8F008090007F8F008090\
008090008090008090008F9F70001F7F001F0000000090A000909F0090A00090A0
009FAF0090A0009FAF009FAF00A0AF009FAF000000607F4F607F50607F4F607F4F607F4F6F8050\
607F4F6F8050607F4F6F805060804F6F805060804F6F80506F80506F80506F8050
6F8F506F80506F8F506F8050708F506F80506F8F506F8050708F506F80500000006F8050708F50\
6F8050708F506F8050708F506F8050708F506F8050708F506F8050708F506F804F
708F506F805000000000B0BF00B0BF00B0BF00B0BF00AFB000B0BF00AFB000000090001F000000\
00000000AFAF00AFAF00AFAF00A0AF00AFAF00A0A000AFAF00A0A000A0A000A09F
00A0A000A09F00A09F009F9F00A09F009F9F0000008F001F90001F8F001F90001F8F0010000000\
00908F009F8F00908F8F001F8F00108F001F8000108F001F608F40A08F1F000000
608F409F80102F200F7F00102F200F000000700010700010700010000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F5F10605F10605F106F60106F6010302F0F
5F00106000106000106000106000106F001F6F00106F001F000000000000000000008090000000\
0000000000007F001F7000107F001F0000000090A00090A00090A00090A0000000
000000009FAF009FAF00A0B000A0AF000000607F4F607F4F607F4F608050607F4F608050607F4F\
6F8050607F4F6F805060804F6F805060804F6F80506F80506F8F506F80506F8F50
6F80506F8F506F80506F8F506F8F50708F506F80506F8F5000000000BFC000B0BF00BFC0000000\
708F506F8050708F506F8F50708F506F8050708F506F8F50708F506F8050708F50
6F8F50708F506F8050708F5000B0BF00BFC000B0BF00B0BF00B0BF00B0BF00AFB000B0B000AFB0\
00B0B000000000000000AFAF00AFB000AFAF00AFAF00AFAF00AFAF00A0A000AFA0
00A0A000A0A000A09F00A0A000A09F00A0A0009F9F0000008F001090001F000000000000009F90\
009F9000908F009F8F00908F8F001F8000108F00108000108F001F800010000000
0000008000107F00109F80102020007F001070000F00000070000F700010005F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000605F106F601F6F601070601F7060105F6F30600010
60001F6000106F001F6F0010007F8F007F8F008090007F8F00809000809070001F70001F7F001F\
70001F7F001F7F001F7F001F000000009FAF0090A00000005F704F607F4F607F4F
607F4F607F4F607F5000A0B0608050607F4F6F80506080506F80506080506F80506F80506F8050\
6F80506F8F506F80506F8F506F8050708F5F6F8F50708F506F8F50708F5F6F8F50
708F5F6F8F5070905F708F5070905F00000000C0CF00BFCF00BFCF00BFC000C0CF00BFCF00BFCF\
708F5070905F708F5070905F708F5070905F708F5070905F708F5070905F708F50
70905F00000000000000BFC000BFCF00BFC000BFC000BFC000BFC000B0BF00BFC000B0BF00BFBF\
00B0BF00B0BF00B0B000B0BF0000009F002000000000B0B000AFAF00B0B000AFAF
00AFAF00AFAF00AFAF00AFA000AFAF00A0A000AFA000000000A0A000A09F00A09F009F9F00A09F\
009F9000A090009F90009F9000908F009F908F00108F001F8000108F001F8F0010
8F001F0000008F001F8000108000107F00108000107F00107F00100000000000006F0010000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F1F0F5F5F106F60106F5F10706010706010706F10000000000000
006F7F007080007080007F8F00708F007F8F007F8F008090008090008F9F008F9F7F001F700010\
7F001F7F001F7F001F0000005F704F5F7040607F4F5F704F607F4F5F7F4F607F4F
607F4F607F50607F4F00AFBF00AFBF0000000000006F80506080506F8F506F80506F8F506F8050\
6F8F506F8F506F8F506F80506F8F506F8F50708F5F6F8F50708F5F6F8F5070905F
6F8F5070905F708F5000C0D000BFCF00C0D000BFCF00C0D000BFCF00C0CF00BFCF00C0D000BFCF\
00C0CF00000000000000000000000000000070905F708F5070905F709050000000
00BFCF00C0CF00BFCF00C0CF00BFCF00C0CF00BFC000C0CF00BFC000BFC000BFC000BFC000B0BF\
00BFC000B0BF00BFBF00B0BF00B0BF00B0B000B0BF9F001F9F001F9F001F000000
00AFAF00B0B000AFAF00B0AF00AFAF0000000000009F001F00A0A000A0A000A09F00A09F009F9F\
00A09F009F9000A09F009F90009F90009F8F009F908F00108F001F8F00108F001F
8000108F00100000008F001F8000108000108000108000107F00107F0010700010007F6F000000\
006060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F5F10605F106F601F6F60105F00105F001060001F000000007080007080
007F8F00708F007F8F007F8F00809000809000809F00809F008FA00000007F001F7F001F80001F\
000000009FAF009FAF607F4F5F704F607F4F607F4F607F50607F4F608050607F4F
6F805060805000B0C000AFBF00B0C000B0C000B0C000B0C00000006F8F506F8F5F6F8F5070905F\
6F8F5070905F6F8F5070905F6F8F5F70905F6F8F5F70905F70905F70905F70905F
00000000C0D000C0D000C0D000CFDF00C0D000CFD000C0D000CFDF00C0D000CFD000C0D000CFDF\
00C0D000CFD000C0D000CFD000C0D000CFD000C0D000CFD0000000A00020A00020
AF0020A00020A00020A0002000000000000000C0D000C0CF00C0D000C0CF00C0CF00BFC000C0CF\
00BFC000C0C000BFC000C0C000BFC000BFC000BFBF00BFC000BFBF0000009F001F
A000209F001F9F00200000000000009F001F9F001F9F001F00000000AFA000AFA000A0A000AFA0\
00A09F00A09F00A09F00A09F009F9000A090009F90009F900000008F001F8F0010
8F001F8F001000000000908000908F0000008F001F80001080001F0080707F00107F0010007F6F\
007060007060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F5F105F5F106F60106F60105F00105F00106000105F001000708000708000708F
00708F007F8F007F8F008090008090008F9F00809F008FA070001F0000000090A00090AF0090AF\
009FAF000000607F4F5F704F607F505F7F4F607F50607F4F608050607F4F6F8050
60805000B0C000B0C000B0C000B0C000BFCF00B0C000BFCF00BFCF00BFCF00BFCF00000000BFCF\
0000006F8F5070905F6F905F70905F6F905070905F70905F00000000C0D000CFD0
00C0D000CFDF00C0D000CFDF00CFD000CFDF00C0D000CFDF00CFD000CFDF00C0D000CFDF00CFD0\
00CFDF00C0D000CFDF00CFD000CFDF00C0D000CFDFA00020AF0020A0001FA00020
A00020AF0020A0001FAF0020A00020AF0020A0001FA00020A0001F00000000C0CF00C0CF00C0CF\
00C0CF00C0C000C0CF00BFC000C0CF00BFC000C0C000BFC000C0C000BFBF00BFBF
000000A000209F001FA0001F9F001FA0001F9F001F9F001F9F001F00B0AF00AFA000AFAF00AFA0\
00AFA000A09F00A09F00A09F00A09F009F9000A09F009F9000A090009F8F8F001F
8F00108F001F8F0010009F8F00908000908F0090800000008000108000100000000000007F0010\
7F001000706000706000605F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000001F1F0F\
605F106F601F6F60106000105F001060001F600010000000007080007F8F007F8F
007F90007F9000809F008090008F9F008F9F0090A00000000090AF0090AF009FB0009FAF00A0B0\
009FB0607F50607F4F608050607F4F6F80506080506F80506080506F8F5F000000
00B0CF00B0C000BFCF00BFCF00BFCF00BFCF00C0D000BFD000C0D000BFD000C0D000C0D000C0D0\
00C0D000CFDF00C0D000CFDF00C0D000000070905F00000000000000CFDF00CFDF
00CFDF00CFDF00D0E000CFDF00D0E000CFDF00D0E000CFDF00D0E000CFDF00D0E000CFDF00D0E0\
00CFDF00D0E000CFDF00D0DF00CFDF000000AF0020AF0020AF0020AF0020AF0020
AF0020AF0020AF0020AF0020AF0020000000000000000000000000AF0020AF0020AF0020AF0020\
00000000CFD000CFCF00CFD000C0CF00CFCF00C0CF00C0CF00C0C000C0CF00C0C0
00C0C000BFC000C0C0A0001FA00020A0001FA000209F001FA0002000000000B0B000B0AF00B0AF\
00AFA000B0AF00AFA000AFA000A09F00AFA000A09F00A09F00A09000A09F009F90
0000008F001090001F000000009F8F00908F009F8F0090800090808000108F001F800010008F7F\
0000007F0010007F6F007F6F007060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000020200F605F10\
0000002F2F0F5F00105F001060001060001060001F007080007F8F007F8F007F90
007F9000809000809F008F9F008F9F0000000090A0009FAF0090AF009FB0009FB000A0B000A0B0\
00A0BF00A0B0000000607F4F60805060805000000000000000B0C000B0C000BFCF
00B0CF00BFCF00BFCF00C0D000BFCF00C0D000C0D000C0D000C0D000C0D000C0D000CFDF00C0D0\
00CFDF00C0D0000000000000709F5F709F5F7F9F60709F5F709F60709F5F7F9F60
709F5F7F9F6000D0DF00D0E000D0DF00D0E000D0DF00D0E000D0DF00D0E000D0DF00D0E000D0DF\
00D0E000D0DF00D0E0000000AF0020AF0020AF002000000000D0E000D0DF00D0E0
00D0DF00D0E000D0DFAF0020AF002000000000D0DF00D0DF00CFDF00D0DF00CFDF00D0DF00CFD0\
00CFDF00CFD000CFD000CFD000CFD0000000AF0020000000AF0020000000000000
00C0C000000000000000C0C000C0C0A00020A0001FA000209F001FA0001F9F001FA0001F000000\
00B0AF00AFAF00B0AF00AFA000AFA000A0A000AFA000A09F00A09F00A09F00A09F
009F9000A0908F001090001F000000009F8F009F8F009F8F0090800090800000008F0010800010\
008F7F0080707F0010000000007F6F007060007060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000605F10605F105F0010\
5F001000000000000060001F6000106F001F000000007F90007F8F008090008090
008F9F008F9F008FA00000000090AF0090AF009FAF009FAF00A0B000A0B000A0BF00A0BF00AFBF\
00AFBF00AFC000AFC000000000000000B0CF00B0C000BFCF00BFCF00BFD000BFCF
00C0D000C0D000C0D000C0D000CFDF00C0D000CFDF00C0DF00CFDF00CFDF00CFDF000000000000\
709F5F7F9F60709F5F7FA060709F607FA060709F607FA0607F9F607FA0607F9F60
7FA0607FA06000DFEF00D0E000DFEF00D0E000DFEF00D0E000DFEF00D0E000000000D0E0000000\
E0DFE000000000D0E000DFEF00DFE000DFEF00D0E000DFEF00DFE000DFE000D0E0
00DFEF00DFE000DFE000D0E0000000AF002000000000D0DF00DFE000D0DF00D0E000D0DF00D0E0\
00D0DF00D0DF00D0DF00D0DF00D0DF00D0DF00CFD000D0DF00CFD0000000AF0020
AF0020AF0020AF0020AF001FAF0020A00020AF0020000000000000A0001FA00020A0001FA00020\
9F001F00000000B0AF00B0B000B0AF00B0AF00AFA000AFA000AFA000AFA000A09F
00AF9F00A09F00A09F900010000000009F9000A090009F8F009F90009F8F009F8F0090808F001F\
80001000907F008F70008F7F7F0010007F6F007060007060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004F5F2F0F0F00000000500010\
5F0010006F7F00708000000060001F600010007F8F007F8F008090008090008F9F
008F9F00000070001F0090AF0090AF009FB0009FAF00A0B000A0B000AFBF00A0BF00AFC0000000\
9000208F001F90002090001F90002090001F90002000000000BFD000BFCF00C0D0
00C0D000C0DF00C0D000CFDF00C0DF00CFDF00CFDF00CFDF00CFDF00D0E000CFDF00D0E0000000\
7FA060709F5F7FA060709F607FA060709F5F7FA0607FA0607FA06070A0607FA060
7FA0607FA06000DFE000DFEF000000000000AF0020AF0020AF0020B00020AF0020000000000000\
00DFEF00DFE000DFEF00DFEF00DFEF00DFE000DFEF00DFE000DFEF00DFE000DFEF
00DFE000DFEF00DFE000DFE000DFE000DFE000000000DFE000D0E000DFE000D0DF00DFE000D0DF\
00DFE000D0DF00D0DF00D0DF00D0DF00D0DF00D0DF00D0DF00D0DF00CFD000D0DF
00CFD000D0D0000000AF0020AF001FAF0020A0001FAF002000000000C0C000BFBF00C0C000BFBF\
A000209F001FA0001F9F001F00000000B0AF00B0AF00AFA000B0AF00AFA000AFA0
00A09F00AF9F00A09F00A09F00A09000A09F009F9000A090009F8F009F8F009080009F8F009080\
8F001F80001000907F008F7F008F7F00806F000000007F60007F60006F5F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F00104F00104F00105F00105F00105F0010\
006F7F007080007080007F8F007F8F000000007F9000809F00809F008FA0000000
7F001F000000009FAF009FAF00A0B0009FB000A0BF00A0BF00AFC00000009000208F001F900020\
90001F9000209000209F00209000209F00209F00209F00209F0020A00020000000
00000000CFDF00CFE000CFDF00D0E000CFE000D0E000D0E000D0E000D0E000DFEF00D0E0000000\
7FA0607FA06F7FA0607FA0607FA0607FAF6F7FA0607FAF6F7FA06080AF6F7FA060
80AF6F00DFEFB0002FB00020B0002FAF0020B0002FB00020B0002FAF002000000000E0EF00E0F0\
00DFEF00E0F000E0EF00E0F000DFEF00E0F000E0EF00E0EF00DFEF00E0F000E0EF
00E0EF00DFEF00E0EF00E0EF00E0EF00DFEF00E0EF00000000E0EF00DFE000E0EF00DFE000DFE0\
00DFE000DFE000DFE000DFE000DFDF00DFE000DFDF00DFE000D0DF00DFE000D0DF
00DFDF00D0DF00DFDF000000B00020AF0020B00020AF0020AF0020AF001F00CFCF00C0C000C0C0\
00C0BF00C0C000BFBF00BFBF00BFB0A0002000000000BFB000B0AF00B0AF00AFA0
00B0AF00AFA000AFA000AF9F00AF9F00000000AF9F00A09000A090009F9000A090009F8F009F8F\
009080009F8F800010000000008F7F008F7F0080707F0010007F6F007F6F007060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F000F4F00104F000F5000105000105F0010006F7F\
007080007080007F8F00708F007F906F001000000000809F008FA070001F7F001F
000000009FAF009FAF00A0B0009FB000A0BF00A0BF0000008F001F900020000000000000000000\
90002090001F00000000000000C0D000C0D000C0DF00C0D000CFDF00C0DF00CFDF
00CFDF00CFE0A00020A00020A00020A00020A00020AF0020A00020AF0020A00020AF0020000000\
00DFEF00DFEF7FAF6F7FA0607FAF607FA0607FAF6F7FA06000000000000000E0F0
00DFEF00E0F0AF0020B0002FAF0020B00020B00020B0002FB00020B0002000000000E0F000E0EF\
00E0F000E0F000E0F000E0EF00E0F000E0EF00E0F000E0EF00E0F000E0EF00E0F0
00E0EF00E0F000E0EF00E0F000E0EF00E0EF00E0EF00E0EFB0002000000000DFE000E0EF00DFE0\
00E0E000DFE000E0E000DFE000DFE000DFE000DFE000DFDF00DFE000DFDF00DFE0
00D0DF00DFDF00D0DF000000AF002000000000D0D0000000000000AF0020AF001F00000000C0C0\
00CFC000C0C000C0C000BFBF00C0BF00BFB000BFBF00B0B000BFB000000000B0AF
00B0AF00B0AF00AFA000AFA000AF9F00000090001000A09F00A09000A090009F9000A090009F8F\
009F8F009080009F80800010009080008F7F00907F008070008070000000007F6F
007060007060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000005F604F00105000107060705000105F0010006F7F007080\
007080007F8F007F8F00809F00000070001F70001F7F001F7F001F7F001F0090AF
009FB0009FB000A0BF00A0BF00AFC00000008F002000AFC000B0CF00B0CF00BFD00000009F0020\
90002000000000C0D000C0DF00C0DF00CFDF00C0DF00CFE000CFDF00D0E000CFE0
00D0E000D0E000D0EF00D0E0000000000000AF002FAF0020AF002FAF0020AF002FAF0020000000\
00DFF000E0F000DFF000E0F000E0F000E0F000E0F000EFFF00E0F000EFFF00E0F0
00EFFF00E0F0B0002FB00020BF002FB0002000000000EFF0000000000000000000B00020BF002F\
00000000EFFF00EFF000EFFF00EFF000EFFF00EFF000EFFF00EFF000EFF000EFF0
00EFF000EFF000EFF000EFF000EFF000EFF000EFF0000000BF002FB00020BF002F00000000EFEF\
00E0EF00E0EF00E0E000E0EF00E0E000E0EF00DFE000E0EF00DFE000E0E000DFE0
00E0E000DFE000DFE000DFDF00000000DFDF00DFDF00D0D000DFDF00D0D000D0D0AF0020B00020\
00000000CFCF00CFC000CFCF00C0C000C0C000C0BF00C0BF00BFBF00C0BF00BFB0
0000009F001F9F001F00000000B0AF00AFA000000090001F00AF9F00A09F00A09F00A09000A090\
009F8F009F8F009F8F009F8F00000000000000907F009080008F7F008F7F00806F
008070007F6F007F6060000F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F00104F000F0060705000105F0010000000006F80007080\
007F8F007F8F00809000809F70001F70001070001F70001F7F001F0090AF009FB0
009FB000A0BF00000000000000AFBF00B0C000AFC000B0CF00B0CF00BFD000BFCF000000900020\
00C0DF00C0D000C0DF00C0DF00CFDF00CFDF00CFE000CFE000D0E000CFE000D0E0
00D0E000DFEF00D0E000DFEF000000AF002FAF0020AF0020AF0020B0002FAF0020B00020000000\
00E0F000E0F000E0F000E0F000EFFF00E0F000EFFF00E0F000EFFF00E0F000EFFF
00EFF000EFFF000000B0002F00000000EFFF00EFF000EFFF00EFFF00EFFF00EFF000EFFF00EFFF\
00EFFF00EFF000000000EFF000EFFF00EFF000EFFF00EFF000EFFF00EFF000EFFF
00EFF000EFFF00EFF000EFF000EFF000EFF000EFF000EFF0000000BF002FB00020BF0020B00020\
00EFF000E0EF00E0EF00E0EF00E0EF00E0E000E0EF00E0E000E0EF00DFE000E0E0
00DFE000E0E000DFE0000000B0002000DFE000DFDF00DFDF00DFDF00DFDF00D0D000DFD000D0D0\
000000AF001F00000000CFC000CFCF00C0C000CFC000C0BF00C0BF00C0BF00C0BF
00BFB000BFB000BFAF0000009F001F9F001F90001F9F001F90001F00000000A09000A09F00A090\
00A090009F8F009F8F009F8F009F8F00908000908000907F009080008F7F008F7F
008070000000007F60007F60007060000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004F00100060700060705F00105F0010007080007080007F8F\
007F9000809F00809F70001F70001F7F001F7F001F000000009FB0000000000000
8F002000000000AFC000AFC000B0CF00B0CF00BFD000BFCF00BFD000BFD000C0DF00000000CFDF\
00C0DF00CFE000CFDF00D0E000CFE000D0E000D0E000D0EF00D0E000DFEF00DFEF
00DFEF00DFEF00E0F0000000AF002F00000000E0FF00E0F000E0FF000000B0002FB00020000000\
00E0F000EFFF00EFFF00EFFF00EFF000EFFF00EFFF00EFFF00EFFF00F0FF00EFFF
00F0FF00EFFF00000000000000F0FF00EFFF00F0FF00EFFF00F0FF00EFFF00F0FF00EFFF00F0FF\
00EFFF00F0FF00EFFF00F0FF00EFFFBF002FBF0020BF002FBF002000F0FF00F0FF
00F0FF00F0FF00F0FF00F0FF00F0FF00EFF000F0FF00EFF000F0FF000000BF002FBF0020BF002F\
BF002000000000EFEF00EFF000EFEF00EFF000EFEF00EFEF00E0EF00EFEF00E0EF
00EFEF000000BF002FB0002000000000E0E000E0E000DFDF00E0E000DFDF00E0DF00DFDF00DFDF\
00DFD000DFDF00D0D000000000D0CF00D0CF00CFCF00CFCF00CFC000CFC000C0BF
00CFBF00C0BF00C0BF00BFB000C0B0000000A000209F001F9F001F90001F00000000A09F00AF9F\
00A09000A09000A09000A090009F8F009F8F009F8000000000908000908000907F
00907F008070008F70007F6F007F6F007060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F000F0060700060700000005F0010006F80007080007F8F007F8F\
00809000809F70001F70001F7F001F7F001F7F001F7F001F80001F80001F000000
00A0BF00AFC000AFC000B0CF00B0CF00BFCF00BFCF00C0D000BFD000C0DF00000000000000CFDF\
00CFE000CFE000D0E000CFE000D0EF00D0E000DFEF00D0EF00DFEF00DFEF00DFF0
00DFEF00E0F000000000000000E0F000E0F000E0F000EFFF00E0F000EFFF00E0F000EFFF000000\
00EFFF00EFF000EFFF00EFF000EFFF00EFF000EFFF00EFF000EFFF00EFF000F0FF
00EFF000000000000000F0FF00EFF000F0FF00EFFF00F0FF00EFF000F0FF00EFFF00F0FF00EFF0\
00F0FF00EFFF00F0FF00EFF000F0FF00F0FF00F0FF000000BF002FBF0020BF002F
BF0020BF002FBF002000000000EFF000F0FF00F0FF00F0FF00EFF0BF002FBF0020BF002FBF0020\
BF002FBF0020BF002F00EFEF00EFF000EFEF00EFF000EFEF00EFEF00EFEF000000
B00020BF0020B00020BF0020B0002000000000E0E000E0E000DFDF00E0E000DFDF00E0DF00DFDF\
00DFDF00DFD000DFDF00D0D000DFD000000000D0CF00CFCF00D0CF00CFC000CFC0
00CFBF00CFC000C0BF00C0BF00C0B000C0B000BFAFA0001F9F001F9F001F90001F9F001F000000\
00AF9F00A09000A090009F8F00A090009F8F009F8F009080000000009080009080
008F7F00907F008F70008F70000000007F6F007060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F0010006070006070006F7F5F0010007080007080007F8F007F9000809F\
00809F70001F70001F009FAF009FAF8000207F001F80002000000000AFC000AFC0
00B0CF00B0CF00BFCF00B0CF00BFD000BFD000C0DF00C0D000CFDF00CFDFA0002000CFE000D0EF\
00D0E000D0EF00D0E000DFEF00DFEF00DFF000DFEF00E0F000DFF000E0F000E0F0
00000000E0F000E0FF00E0F000EFFF00EFF000EFFF00EFF000EFFF00EFF000EFFF00EFF000F0FF\
00000000000000EFFF00F0FF00EFFF00F0FF00EFFF00F0FF00F0FF00F0FF00EFFF
00F0FFBF002000000000F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF\
00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00FFFF00F0FF00F0FFBF0020
C0002FBF0020BF002FBF0020C0002FBF002000000000F0FF00FFFF000000C0002FBF0020C0002F\
BF0020C0002FBF0020C0002F00000000F0F000EFF000F0F000EFF0000000BF0020
BF002FBF0020BF002FBF0020BF002F00000000EFEF00E0E000EFEF00E0E000E0E000E0DF00E0E0\
00E0DF00E0DF00DFDF00E0DF00DFD000DFDF00DFD0B0002000000000DFCF00D0CF
00D0CF00CFC000D0C000CFBF00CFC000C0BF00C0BF00BFB00000009F001F9F001F9F001F9F001F\
90001F90001F00000000AF9F00A09000A090009F8F00A090009F8F000000009080
009F8000907F009080008F7F008F7F00807000806F007060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000004F000F006070006070006F7F000000006F80007080007F8F007F8F00809F000000\
70001F008FA00090AF0090AF80001F7F001F80001F80001F00AFC000AFC000B0CF
00B0CF00BFCF00B0CF00BFD000BFD000C0DF00C0D000CFDF00CFDFA000209F002000D0E000D0E0\
00DFEF00D0EF00DFEF00DFEF00DFF000DFEF00E0F000DFF0000000AF002000E0F0
00E0F000EFFF00E0F000EFFF00EFF000EFFF00EFF000EFFF00EFF000EFFF00EFF000EFFF00EFF0\
000000B00020B0002FB0002000F0FF00EFF000F0FF00F0FF00F0FF00EFF0000000
B00020BF002F00000000F0FF00F0FF00F0FF00F0F000F0FF00F0FF00F0FF00F0F000F0FF00F0FF\
00F0FF00F0F000F0FF00F0FF00F0FF00F0F000F0FF00F0FF00F0FF00F0F000F0FF
00F0FF000000BF0020BF002FBF0020BF002FBF0020BF002FBF0020000000000000BF002FBF0020\
C0002FBF0020BF002FBF0020C0002FBF002000000000F0F0000000000000BF0020
BF0020BF002FBF0020BF0020BF002000000000EFE000EFEF00E0E000EFEF00E0E000E0E000E0DF\
00E0E000E0DF00E0DF00DFDF00E0DF00DFD000E0DF00DFD0000000AF001F000000
00D0CF00D0CF00D0C000D0C000CFBF00CFC000C0BF00C0BF00BFB000C0B00000009F001F9F001F\
9F001F000000000000900010BFAF9F00A09000A090009F8F00A08F009F8F000000
800010009F8000907F00907F008F7F008F7F00806F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F0010000000006070006F7F006F7F00000000708F007F90007F9000809F000000008FA0\
008FA0009FAF009FB000000080001F8F002080001F8F002000AFC000B0CF00B0CF
00BFD000BFD000C0DF00C0D000C0DF00C0DF00CFE000CFE0A00020A00020A0002F00D0EF00DFF0\
00DFEF00E0F000DFF000E0F000E0F0000000000000B0002FAF002000EFFF00E0F0
00EFFF00EFF000EFFF00EFFF00EFFF00EFF000F0FF00EFFF00F0FF00EFF000F0FF00F0FF00F0FF\
00F0FFBF002FB00020BF002FB0002000000000F0FF00F0FF00F0FF000000BF0020
BF002FBF002000000000000000000000000000FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF\
00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF
00FFFF00F0FF00FFFF000000C0002FBF0020C0002FBF0020C0002FBF0020C0002F000000000000\
000000C0002FBF0020C0002FBF0020C0002FBF0020000000BF0020C0002FBF0020
C0002FBF0020C0002FBF0020C0002FBF002000000000EFEF00F0EF00EFEF00EFEF00EFE000EFE0\
00E0E000EFE000E0E000EFE000E0DF00E0DF00E0DF00E0DF00DFD000E0DFB00020
B00020B0002000000000D0CF00D0CF00CFC00000007F9F507F9F5070904F7F9050708F4F00BFAF\
00B0AF00B0AF00B0A000B0A000AF9F00AF9F00A090000000000000000000009F8F
009F8F8F0010009F8F00908000908000907F00907F0080707F0010700010007060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F0010006070006F7F00607F006F80006F80007F8F007F8F00809F00809F008FA0008FA0\
0090AF009FAF00000000000000AFC000AFC000B0CF00000000B0CF00B0CF00BFD0
00BFD000C0D000C0D000CFDF00C0DF00CFE000CFE0A00020A00020A0002000000000DFF000DFEF\
00E0F000DFF000E0F0000000AF002FAF0020B0002F00000000EFF000E0F000EFFF
00EFF000EFFF00EFF000F0FF00EFF000F0FF00EFF000F0FF00EFF000F0FF00F0F000F0FF00F0F0\
00F0FF000000BF002FB00020BF002FB00020BF002FB00020BF002FB00020BF002F
B00020BF002FBF0020BF002FBF0020000000000000000000000000000000000000000000000000\
000000000000BF002F000000000000BF0020000000BF0020BF002FBF0020BF002F
BF0020BF002F000000000000000000000000BF0020C0002FBF0020BF002FBF002020200FEFF02F\
EFFF2FEFF02FD0FF4F000000C0002FBF0020D0FF4FEFF02FEFFF2F3F3F0FC0002F
BF0020C0002FBF0020C0002FBF0020BF002000000000F0F000EFEF00F0EF00EFEF00EFEF00EFE0\
00EFE000EFE000EFE000E0DF00EFE000E0DF00EFDF00E0DF00E0DF00E0D000E0DF
000000B00020B0001FB0002000D0CF00D0CF00D0C07F9F5070904F7F9F5070904F709050708F4F\
00000000B0AF00B0AF00AFA000B0A000AF9F00AF9F00A09000A0908F001090001F
8F00108F001F8F00108F001F00000000908000907F00907F0000007F001070000F007060000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00606F006070006F7F006F7F00708F000000007F90007F9000000000809F008FA00090A0009FB0\
009FB000000000000000AFC000AFC000B0CF00B0CF00000090002000000000BFD0
00C0DF00C0DF00CFE000CFDF00D0E000D0E0000000A00020AF002FA0002000000000DFF000E0F0\
000000AF002FAF0020B0002FAF0020B0002F00EFF000EFFF00EFF000EFFF00EFF0
00F0FF00EFFF00F0FF00EFF000F0FF00F0FF00F0FF00F0FF00FFFF000000BF002FB00020BF002F\
BF0020BF002FBF0020BF002FBF0020BF002FBF002000FFFF00FFFF000000000000
BF002FBF0020BF002F00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0FF00FFFF\
00FFFF00FFFF000000000000000000C0002FBF0020C0002FBF0020C0002FBF0020
C0002FBF0020C0002FBF0020C0002FBF0020404010B0F05FF0FF30CFFF405F60100000007F801F\
EFFF2FF0FF2FEFF02FF0FF30EFFF2FF0FF2FEFF02FF0FF30EFFF2FF0FF2FB0F05F
C0002FC00020C0002FBF0020C0002FBF002000000000F0F000F0F000F0F000F0F000EFEF00F0F0\
00EFEF00F0EF00EFE000EFEF00EFE000EFE000E0E000EFE000E0DF00EFDF00E0DF
00EFDF00E0DF000000B00020B0002000000000DFCF00D0C07FA05F7F9F507F9F5070904F7F9F50\
70904F70904F00000000BFAF00B0A000B0A000AF9F00AFA000AF9F00AF9F00A090
0000008F00108F001F8F00108F001F8F001000000000907F0000007F00107F0010700010007060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000005F60
4F001000607000607F006F805F0010007F906000106F001F00809F008FA0008FA00090AF009FB0\
00A0BF00A0BF00AFC000AFC000B0CF00B0CF00BFD000000090002090001F9F0020
00000000CFE000CFDF00D0E000D0E0000000A00020A00020A00020AF002FA0002000E0F0AF0020\
AF002FAF0020B0002FAF0020B0002F00EFF000EFFF00EFF000F0FF00EFF000F0FF
00EFF000F0FF00EFF000000000F0F000000000000000F0FF00F0F000FFFF00F0F000FFFF00F0FF\
00FFFF000000BF002FBF0020BF002F00000000FFFF00F0FF00FFFF00F0F000FFFF
00FFFFBF002FBF002000000000FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00F0FF\
00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF000000000000000000
BF0020C0002FBF0020C0002FBF0020C0002FBF0020C0002FB0F05FF0FF2FEFF02FEFFF2FEFF02F\
F0FF2FEFF02FEFFF2FEFF02FF0FF2FEFF02FEFFF2FEFF02FF0FF2FEFF02FEFFF2F
EFF02FF0FF2F3F3F0F3F400F3F3F0FC0002FBF002000000000F0F000F0F000F0F000F0F000F0EF\
00F0F000EFEF00F0EF00EFE000EFEF00EFE000EFE000EFE000EFE000E0DF00EFE0
00E0DF000000B0001FB00020B00020B00020B0001F0000000000007FA0507F9F507F9F5070904F\
7F9F5070904F70904F708F4F00BFAF00B0A000B0A000AF9F00AFA000A09F00AF9F
00A09000A0900000008F001F8F00108F001F80001080001000000000907F7F00107F0010007060\
007060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000006060005F6F
00607000607F006F8000000060001F6000106F001F00809F008FA00090A0009FB0009FB000A0BF\
00A0BF00AFC000AFC000B0CF00B0CF00BFD000BFD00000009000209F00209F0020
9F002000000000D0EF00D0EF00DFEFA00020AF002FAF0020AF002FAF0020000000AF0020B0002F\
AF0020B0002FB0002000000000EFF000F0FF00EFF000F0FF000000BF002FB00020
BF002FB00020BF002FB0002000FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFFBF0020BF002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFFBF0020C0002FBF0020C0002F3F400F4040103F3F0FF0FF30EFFF2FF0FF2F\
EFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFF02FF0FF30EFFF2F
F0FF2FEFF02FF0FF30EFFF2FF0FF2FEFF02F90FF80C00020C0002FBF0020C0002FBF0020C0002F\
BF0020C0002FBF0020C0002FBF0020C0002FBF0020BF0020BF0020000000000000
000000000000BF0020B00020BF0020B00020BF0020B000200000007FA05080A05F7F9F507F9F50\
7F9F507F9F5070904F7F9050708F4F00000000BFAF00BFAF00B0A000B0A000AF9F
00AF9F00A09000AF9F00A0900000008F00108F001F8F001000000000907F00907F0000007F0010\
007F60007060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000005F60006070
006070006F80006F8060001F60001000000000809F008FA0008FA00090AF009FB000A0BF00A0BF\
00AFC000AFC000B0CF00B0CF00BFD000BFD000C0DF0000009F00209F0020A00020
9F0020A00020000000000000A00020AF0020A00020AF002FAF0020A0E05F80E070B0002FAF0020\
B0002FAF0020000000B00020B0002FB00020B0002FB00020B0002FB00020BF002F
B0002000000000F0F000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFF00000000F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFFB00020BF002F00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000000000
BF0020BF002FBF0020C0002FBF0020C0002F8FF080F0FF2FEFF02FEFFF2FEFF02FF0FF2FEFF02F\
EFFF2FEFF02FF0FF2FEFF02FEFFF2FEFF02FF0FF2FEFF02FEFFF2FEFF02FF0FF2F
EFF02FEFFF2FEFF02FF0FF2FEFF02FEFFF2FEFF02F606F10BF0020C0002FBF0020C0002FBF0020\
C0002FBF0020C0002FBF0020BF0020BF0020BF0020BF002000000000000000EFEF
00EFE000EFE000E0DF00EFE0000000000000B00020B00020B0001F00000000DFCF80A05F7F9F50\
7F9F50709F507F9F5070904F70904F70904F00000000BFAF00BFAF00B0A000B0A0
00AF9F00AF9F00A09000AF9000A08F00A08F8F001000000000908000908000907F00907F008070\
7F0010007F60007060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000005F6000606F006070
006F80006F8060001F60001000000000809F008FA00090AF009FB0009FB000000000A0BF00AFC0\
00AFC000B0CF00B0CF00BFD000BFD000C0DF00C0DF0000009F0020A000209F0020
A0002FA00020AF002F303F0F1F1F0FAF0020AF002F000000DFEF2FDFEF2F8FEF7FAF002020200F\
DFEF2FAFF060B00020BF002FB00020BF002FB00020BF002FB0002000000000F0FF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
000000BF002FB0002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFFBF002FB00020BF002FB0002000000000FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF000000BF002FBF0020BF002FBF0020BF002FBF0020
BF002FBF0020BF002FBF0020C0002FBF0020C0002FBF0020C0002F90FF7FF0FF2FEFFF2FF0FF30\
EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2F
F0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2F000000C0002FC00020C0002FBF0020C0002F\
C00020C0002FBF0020C0002FBF0020C0002F00000000F0F000F0EF00F0EF00EFEF
00F0EF00EFE000EFE000EFE000EFE000E0DF000000B00020BF0020B0002000E0D000DFCF000000\
7FA0507FA0507F9F507F9F50709F507F9F5000000000C0B000BFAF00BFAF00B0A0
00BFA000B0A000B0A000AF9F00AF9F00A09000A09000000000A08F009F80009F8000907F00907F\
008F70008070007F6F007F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000005F5F00606F005F6F00607F
006F7F5F001F5F001000000000809F008FA0008FA0009FB0009FB080002000000000AFC000AFC0\
00B0CF00B0CF00BFD000BFD000C0DF00C0DF00CFE0000000A000209F0020A00020
A00020AF002F9FDF5FD0E02FD0DF2FD0E02FD0E02FDFEF2FD0E02FDFEF2FDFEF2FDFEF2FDFEF2F\
AFF060B00020B0002FB00020BF002FB0002000000000000000FFFF00F0F000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000\
BF002FB00020B0002F00000000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00000000FFFF00F0F000FFFF000000B0002FB00020B0002FB00020B0002FB00020000000000000\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F0000000000000BF002FBF0020BF002FBF0020BF002F1F1F00EFFF2FEFFF2F\
F0FF2FEFF02FEFFF2FEFF02FF0FF2FEFF02FEFFF2FEFF02FF0FF2FEFF02FEFFF2F
EFF02FF0FF2FEFF02FEFFF2FEFF02FF0FF2FEFF02FEFFF2FEFF02FF0FF2F000000C0002FBF0020\
C0002FBF0020C0002FBF0020C0002F00000000F0F000F0F000F0F000F0EF00F0EF
00EFEF00F0EF00EFE000EFE000EFE000EFE000E0DF000000B00020B0002000000000E0D000DFCF\
00DFCF7FA0507FA0507F9F507F9F5070904F709F5070904F00C0B000BFAF00BFAF
00B0A000BFA000AF9F00AF9F00AF9000AF9000A08F00000060703F009F8F00908000908000907F\
00907F008F70008070700010007060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000606F4F001000607F006F7F
00000000708F00809F4F6040506F4F0090AF009FB0009FB080002000000000AFC000AFC000B0CF\
00B0CF00BFD000BFD000C0DF00C0DF00CFE000CFE0000000A00020AF002FA00020
AF002F303F0FD0E02FD0E02FDFEF2FD0E02FDFEF2FDFEF2FDFEF2FDFEF2FE0F030DFF02FE0F02F\
3F3F0F3F40103F400F00000000000000000000FFFF00000000000000000000FFFF
00000000000000000000FFFF00FFFF00000000FFFF000000000000B00020BF002FB00020BF002F\
B00020BF002FB00020BF002F00000000000000000080C06F80BF6F80C06F80BF6F
80C06F80BF6F80C06F80BF6F00000000FFFF00FFFF000000B0002FB00020B0002FB00020000000\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000000000000000EFFF30CFFF40EFFF2F\
EFFF2FEFFF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2F
F0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2F\
EFFF2FC0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00F0F000FFFF00F0F0
00F0F000F0EF00F0EF00EFEF00F0EF00EFE000EFE000EFDF000000B00020BF002000000000E0D0\
00DFD000E0D000DFCF0000007FA0507FA0507F9F507F9F5000C0B000CFBF00C0B0
00C0B000BFAF00BFAF00B0A000B0A000AF9F00AF9F607F3F6F7F4060703F00A08F009F80009F80\
00907F00907F008F70000000700010000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000005F60005F6F000000500010006F80
007080007F904F6040506F4F4F6F40009FB0009FB08000207F001F00AFC000AFC000B0CF00B0CF\
00BFD000BFD000C0DF00C0DF00CFE000CFE000D0EF000000303F0F9FDF5FA0E05F
CFDF2FD0E02FD0E02FDFEF2FD0E02FDFEF2FDFEF2FDFF02FDFEF2FDFF02FDFF02FE0F02FDFF02F\
E0F02FE0F02F1F1F0FB00020BF002FB00020BF002FB00020BF002FB00020BF002F
B0002000000000FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000B0002FB00020\
B0002F00000000FFFF00F0F00000007FBF6F80BF6F7FBF6F80BF6F7FBF6F80BF6F
7FBF6F80BF6F7FBF6F80BF6F7FB06000F0FF00F0F000F0FF00F0F000FFFF00F0F000FFFF000000\
B0002F00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000BF002FB00020BF002FB00020BF002FB00020\
BF002FBF0020000000E0F02FEFFF2FEFFF2FEFFF2FEFF02FEFFF2FEFFF2FF0FF2F
EFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FEFFF2FF0FF2FCFF040\
5F5F10000000C0002FBF002000000000000000FFFF00F0F000FFFF00F0F000FFF0
00F0F000F0F000F0EF00F0EF00EFE000EFEF00EFE000EFE000E0DF000000B0002000000000E0D0\
00E0D000DFD000E0D000DFCF00DFCF00D0C000D0C000CFBF00CFBF70904F000000
70904F00000000000000BFAF00B09F00B09F00AF9F00AF9F607F3F607F3F60703F607F3F009080\
009F8000907F00907F008F707F0010700010700010000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000005F60500010500010006F80006F80
007F904F6040506F4F506F40000000009FB080002080001F00AFC000AFC000B0CF00B0CF00BFD0\
00C0D000C0DF00C0DF00CFE09F0020A0002FA00020000000CFDF2FD0E02FD0E02F
D0EF2FD0E02FDFEF2FD0EF2FDFEF2FDFEF2FE0F02FDFF02FE0F030E0F02FE0FF2FE0F02FE0FF30\
C0FF40BF002FB00020BF002FB00020BF002FB0002000000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000B0002F00000000FFFF\
00F0FF00FFFF00F0F000FFFF00F0F000F0FF00F0F000F0FF00F0F000F0FF000000
7FBF6F7FB06F7FBF6F7FB06000000000000000EFF000EFF000F0F000EFF000F0F000EFF000F0FF\
00F0F000F0FF00F0F000F0FFAF0020AF002F00F0F000FFFF00F0FF00FFFF00F0FF
00FFFF00FFFF00FFFF00FFFF00FFFF000000B0002FB00020BF002FB00020BF002FB00020BF002F\
BF0020BF002FBF0020BF002FEFFF2FEFFF2FEFFF2FEFFF30EFFF2FEFFF2FEFFF2F
F0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FBFFF60\
C00020C0002FBF0020C0002FC00020C0002FBF002000000000000000FFFF00FFFF
00FFFF00F0F000FFF000F0EF00F0F000F0EF00F0EF00EFE000F0EF00EFE0000000B00020000000\
00E0DF00E0DF00E0D000E0D000DFCF00DFCF00DFC000DFCF00D0C0000000709F50
7F9F5070904F70904F6F8F4F00000000B0A000B0A000AF9F000000607F3F6F7F4060703F607F3F\
009F80009F8000907F00907F008F707F0010700010007F6F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000040000F4F00104F0010006F7F006F8000708F
0000004F60404F604050704F0000007F00207F001F80002080001F00000000B0CF000000000000\
9F00209000209F00209F0020A00020A00020A00020A00020D0E02FCFDF2FD0E02F
D0E02FDFEF2FD0EF2FDFEF2FDFEF2FE0F02FDFEF2FE0F02FDFF02FE0FF2FDFF02FE0FF2FE0F02F\
CFFF4FAFF05F00000000000000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00000000FFFF00F0F000F0F000F0F0\
00F0F000EFF000EFF000EFF000000070B0607FB06F70B0607FB06F70AF6070B06F
70AF6070B06F70AF6070AF6070AF6070AF6F6FAF6070AF606FAF6070AF6F6FAF6000E0EF00E0EF\
00EFF000E0EF70B06F00000000EFF000EFEF00EFF0A00020AF002FA00020AF0020
A00020AF002F000000000000AF0020B0002FAF0020B0002FB00020B0002FB00020000000000000\
BF002FB000200000000000000000003F3F0F3F400F3F3F0F3F400FE0F02FEFFF2F
EFFF2FEFFF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FEFFF2F\
F0FF2F1F1F00C0002FBF0020C0002FBF0020C0002FBF0020C0002FBF002000FFFF
00FFFF00FFFF00F0F000F0F000F0EF00F0F000EFEF00F0EF00EFE000EFE000EFDF00EFE0EFE0DF\
00EFDF00E0D000E0DF00E0D000E0D000DFCF00DFCF00DFCF00DFCF00D0C07FA050
709F4F7F9F5070904F70904F6F8F4F708F4F00000000B0A000AF9F000000607F3F607F3F60703F\
607F3F009080009F8000907F00907F008F707F0010700010007F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005000104F001000607F006F8000708F007F90
4F60404F604050704F50704F0000007F001F80002080002000B0CF00B0CF00C0D000C0D09F0020\
9F00209F002F9F0020A0002FA00020AF002FA00020AF002FD0E02FDFEF2FD0E02F
DFEF2FDFEF2FDFF02FDFF02FE0F02FDFF02FE0FF30E0F02FE0FF2FE0FF2FE0FF30E0FF2FE0FF2F\
505F10BF002FBF0020BF002FB0002000000000FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00F0F000000000F0F000F0FF00F0F000F0F000EFF000F0F0\
00EFEF00EFF070AF6070B06F70AF6070B06F70AF6070AF6F6FAF6070AF6F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006FAF6070AF6F6FAF6070AF6F70AF6070AF6F70AF6070B06F000000
00EFF000E0EFAF002FA0002000F0F000F0F000F0FF00F0F000FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF000000BF002FB00020BF002FBF0020BF002F3F400F
EFFF2FEFFF2FEFFF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2F\
EFFF2FF0FF30EFFF2F20200FBF0020C0002FC00020C0002FBF0020C0002FC00020
C0002F00000000FFFF00FFF000FFFF00F0F000FFF000F0EF00F0EF00EFEF00F0EF00EFE000EFE0\
00000000EFE000E0DF00EFDF00E0D000E0D000E0D000E0D000DFCF00DFCF00D0C0
7FA050709F507F9F5070904F70904F708F4F70904F6F8F4000B0A000AF9F6F8040607F406F7F40\
607F3F607F3F000000009F8000907F00907F008F70000000700010007F6F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F001000607F00607F006F8F00708F000000
4F6040506F4F506F4F0000007F001F80002000AFC000B0CF00B0CF00BFD000BFD000C0DF00C0DF\
0000009F0020A00020A00020A00020A00020D0E02FD0E02FD0E02FD0E02FDFEF2F
D0EF2FDFF02FDFEF2FE0F02FDFF02FE0F02FE0F02FE0FF2FE0F02FE0FF2FE0F02FBF002FB00020\
BF002FB00020BF002FB00020B0002FB00020BF002F00000000000000FFFF00FFFF
00F0F000FFFF00F0F0000000AF0020AF002000000000000070B0607FB06F70AF6070B06F70AF60\
70AF60000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000DFEF00DFEF00E0F000E0EF00EFF0A0002000EFF000EFF000F0F000F0F000F0FF00F0F0\
00FFFF00FFFF00FFFF00F0F000FFFFB00020B0002FB00020B0002FB00020BF002F
B00020E0FF2FAFFF5F606F10707F10EFFF2FEFFF2FEFFF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02F\
C0002FBF0020C0002FBF0020C0002F00000000FFFF000000C0002FBF0020C0002F
BF0020C0002FBF002000000000F0F000FFFF00F0F000F0F000F0EF00F0EF00EFEF00F0EF00EFE0\
00EFE000000000EFE000E0DF00E0DF00E0D000E0D000DFCF00E0D000DFCF00DFCF
0000007FA050709F4F7F9F5070904F70904F6F8F4F708F4F6F80406F8F4000AF9F000000607F3F\
607F3F607F3F607F3F009F80009F8000907F00907F000000008F70000000007060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000005F6000607000607F006F8F00708F007F90000000
506F4F506F4F5F7050009FBF80002080001F00B0CF00B0CF00C0DF00C0DF00CFDF00CFDF00D0EF\
00D0EF00DFEF000000AF002F000000D0E02FD0E02FDFEF2FD0EF2FDFEF2FDFEF2F
E0F030DFF02FE0F02FE0F02FE0FF30E0FF2FE0FF2FE0FF2FE0FF30E0FF2FAFFF60B00020BF002F\
BF0020BF002FB00020BF002FB00020BF002F00000000000000FFFF00FFFF00F0F0
00F0FFAF0020AF002F00000000F0F07FB0607FB06F70AF607FB06F70AF60000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000DFEF00E0F000E0EF00EFF000E0F000000000EFF000F0F0\
00EFF000F0FF00F0F000FFFF00FFFF000000B00020B0002FB00020BF002FB00020
BF002F000000000000BF0020BF002FBF002020200FEFFF2FF0FF3090FF7F20200FEFFF2FF0FF30\
EFFF2FBFFF60BF0020C0002FC00020C0002FBF002000FFFF00FFFF00FFFF00FFFF
00FFFF000000C0002FBF0020C0002F00000000FFFF00F0F000FFFF00F0F000F0F000F0EF00F0EF\
00EFE0BF002000000000EFE000EFDF00EFDF00E0DF00EFDF00E0D000E0D000DFCF
00E0CF7FA0507FA050709F507F9F5070904F70904F708F4F70904F6F8F406F8F4F6F80406F8040\
607F406F7F40607F3F607F3F009F80009F8000000080001F009070008F70007F6F
007060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000005F6F006070006F80006F80007F90007F904F6F40
4F6F4050704F009FB000A0BF80001F00B0CF00B0CF00BFD000BFD000C0DF00C0DF00CFE000D0E0\
00D0EF00D0EF00DFF0A00020AF002F1F1F003F3F0FD0E02FDFEF2FDFEF2FDFF02F
DFF02FE0F02FDFF02FE0FF2FE0F02FE0FF2FE0F02FE0FF2FAFF05F3F400FB0002000FFFF00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00F0F000F0FF00F0F0AF0020
00EFF000EFF000EFEF00EFF000E0EF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000DFEF00DFEF00E0EF00E0EF\
A0002F00000000EFF000EFF000F0F000EFF0000000AF0020B0002FB00020000000
00000000FFFF00F0F0000000B00020BF002FB00020BF002F000000EFFF2F000000BF002FBF0020\
1F1F0F8FF07F7FFF90BF0020C0002FBF0020C0002FBF002000000000F0F000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF000000BF002000000000F0F000FFF000F0F000F0F000F0EF\
00F0EF00EFE0BF0020B0002000EFE000E0DF00EFDF00E0DF00E0DF00E0D000E0D0
00DFCF7FAF5F7FA0507FA050709F507F9F5070904F70904F6F8F4F708F4F6F80406F8F406F8040\
000000607F3F607F3F607F3F607F3F009F80000000800010800010008F70008F70
007F60000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000006F80006F80007F90007F90008FA0000000
000000009FB000A0BF00AFC08F002000B0CF00BFD000BFD000C0DF00CFE000D0EF00D0EF00DFEF\
00DFEFAF002FA00020AF002FAF0020B0002FDFEF2FDFEF2FDFEF2FE0F030DFF02F
E0FF2FE0F02FE0FF30E0FF2FE0FF2FE0FF2FB0FF60B00020BF002FB00020BF002F00000000FFFF\
00FFFF00FFFF00FFFF00FFFF00F0F000FFFF00F0F000F0FF00F0F000F0F000EFF0
00EFF000EFEF00DFE000C0C0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00DFEF00E0F000E0EFA0002FA00020A0002FA0002000000000F0F000FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFFB00020BF002FBF0020BF002FBF0020000000BF0020C0002F\
BF0020C0002FC00020C0002FBF0020C0002FC00020C0002FBF0020C0002F00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0F000FFF0\
00F0EF00F0F0000000BF002FBF002000F0EF00EFE000EFE000E0DF00EFDF00E0D0
00E0D000DFCF80AF5F7FA0507FA050709F507F9F5070904F70904F70904F70904F6F8F406F8F4F\
00000000AF9F607F406F8040607F3F607F3F0000008F001F80001000907F008F70
008070007060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000005F603F4F3000607F00708F0000004F60404F60400090AF
0090AF00A0BF00A0BF00AFC000000000BFD000BFD000C0DF00C0DF00CFE000D0E000DFEF000000\
A00020A00020AF002FA00020AF0020D0EF2FDFEF2FDFEF2FDFF02FDFF02FE0FF2F
DFF02FE0FF2FE0F02FE0FF2FE0F02FE0FF2FB00020BF002FB00020BF002FB0002000000000F0F0\
00FFFF00F0F000FFFF00F0F000F0F000F0F000000000EFF000EFF000EFEF00EFF0
00D0D000C0C0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000009F0020000000A0002F9F002000000000EFF000F0F0
00F0F000FFFF00FFFF00FFFF00F0F000FFFFB00020BF002FB00020BF002FB00020000000000000\
BF002FBF0020BF002FBF0020C0002F00FFFF000000BF0020C0002FBF0020C0002F
00000000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF000000BF0020\
BF0020BF0020000000BF0020BF0020B0002000EFE000EFDF00EFE000E0DF00E0DF
00E0D000E0D000DFCF7FAF5F7FA0507FA050709F507F9F5070904F70904F6F904F70904F6F8F40\
6F8F406F804000AF9F00A090607F3F00000000A08F80001080001080001000907F
008F7000806F007060000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000606F3F4F30006F8F405F3F4F60404F60400000000090AF
00A0BF00A0BF00AFC000AFC000BFD000BFD000C0DF00C0DF00D0E000D0EF00DFEFA00020AF002F\
A00020AF002FAF0020B0002F80EF7F3F4010B00020E0F030DFF02FE0FF2FE0F02F
E0FF30E0FF2FBF002FB00020BF002F000000000000B00020BF002FBF0020BF002FB0002000FFFF\
00F0FF00FFFF00F0F000F0FFAF0020AF002F00EFF000F0F000000000D0D0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000009F002F9F0020A0002F000000
00EFF000EFF000F0FF00F0F000FFFF00FFFF00FFFF000000BF002FB00020BF002F00000000FFFF\
00FFFF00FFFFBF0020C0002FBF0020C0002F00FFFF00FFFF000000C0002FBF0020
C0002FC0002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFFC0002FBF0020C0002FBF0020BF002FBF002000F0EF00EFE000EFE000E0DF
00EFDF00E0D000E0D07FA05F80AF5F7FA0507FA050709F507F9F50709F4F709F5070904F70904F\
6F8F406F8F4F00B09F00B09F00AF9000000000A08F00A08F8000108F001F00907F
00907F008F70007F6F006F60000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000005F6030402F00607F000000405F404F6040506F4F0090AF009FB0
009FBF00AFC000AFC000B0CF00BFD000C0DF00C0DF00CFE000CFE000D0EF9F0020A00020A00020\
AF002FA0002000EFF0AF0020B0002FAF00203F400FDFF02FE0FF2FB00020C0FF40
E0F02FBF002FB00020BF002FB0002000FFFF00F0F000FFFF00F0FF00FFFF00000000FFFF000000\
B0002FAF0020AF0020AF002000F0F000EFEF00E0E08F001F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000000DFEF
000000A0002F00E0EF00EFF000EFF000F0FF00F0F000FFFF000000B0002FB0002000000000FFFF\
00FFFF00F0F000FFFF000000BF002FBF0020BF002F00FFFF00FFFF00F0F0000000
BF0020C0002FBF002000000000FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F0000000000000BF002FBF0020BF002000000000F0EF00EFE0000000
00000000E0DF00E0D000E0D00000007FAF5F7FA0507FA050709F507F9F5070904F70904F6F904F\
70904F6F8F406F8F4000AF9F00AF9F00A09000A09000A08F000000800010000000
00907F00907F00806F007F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000005F603F4F30006080405F404060404F6F4F506F4F009FB0009FBF
00AFC000AFC000B0CF00000000C0DF00C0DF00CFE000CFE000D0EFA00020A0002F00000000E0F0\
00E0F000EFF0AF0020B0002FB00020B0002FE0F02F1F1F0FB00020BF002F505F10
000000B00020BF002FBF002000000000FFFF00FFFF00FFFF00FFFF00F0F000FFFF00F0F0000000\
AF0020B0002F000000AF002F00DFDF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000DFEF00E0F000E0F000000000EFF000F0FF00F0F000FFFF00FFFFB0002F00000000FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFFBF002FBF0020C0002F00FFFF00FFFF00FFFF
00FFFF000000C0002FC0002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF000000C0002FBF0020BF002FBF002000F0EF000000
80B06080AF5F00EFDF00E0D000E0D07FA05F80AF5F7FA0507FA050709F507FA050709F4F709F50\
70904F70904F6F8F4F6F8F4F00B09F00B09F00AF9000AF9000A08F8F001F800010
009F8000907F00907F00806F007F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000005F6F3040303F503F405F3F4F60404F6F40000000009FB000A0BF
00A0BF00B0CF00B0CF90002000BFD000CFDF00CFE000D0EF9F002000DFEF00DFEF00E0F000E0EF\
00EFF0AF0020B0002FAF0020B0002FB00020BF002FB00020B0002FB00020F0FFFF
B00020BF002FB00020BF002F00F0F000FFFF00F0FF00FFFF00F0F000F0FF00F0F0000000AF0020\
00F0F0A0002000E0E000D0D000CFC0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000DFEF00E0F000E0EF00EFF000EFF000F0F000F0F000000000F0F0\
00FFFF00FFFF00FFFF00F0F000FFFF00FFFF000000BF0020BF002F00FFFF00FFFF
00F0F000FFFF00FFFF00FFFFBF0020C0002F00000000FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFF00FFFF00F0F000FFFFBF0020BF002FBF0020BF002000000080B060
80AF5F80B0607FAF5F80AF6000000000E0D07FA0507FAF5F7FA0507FA050709F507F9F5070904F\
70904F6F904F70904F6F8F4000B0A000B09F00B09F00A09000A09000A08F8F001F
80001000907F009070008F706F000F007060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000005F6030402F0000003F503F4F60404F604050704F009FB000A0BF00A0BF
00B0CF00B0CF0000008F002000CFE000CFE000D0EF00D0EF00DFF000DFEF00E0F000E0F000EFF0\
000000B0002FB00020B0002F00F0FF000000B00020BF002FB0002000FFFF00FFFF
000000BF0020BF002FBF002000FFFF00FFFF00FFFF00F0F000FFFF00F0F0B0002F00F0F000F0F0\
A0002000E0E000D0CF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000DFF000DFEF00E0F000000000000000EFF0000000\
00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFBF002000000000FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00000000000000000000FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF000000BF0020C0002F00000000F0F080B060
8FB06080B06080B06080AF5F80B06080AF5F0000007FA05F80AF5F7FA0507FA05F709F507FA050\
709F4F709F5070904F70904F6F8F4F00000000B09F00B09F00AF9000AF90000000
8F001F00000000907F008F70008F70007F6F007060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000304F3000607F405F3F405F3F4F6F40506F4F00A0BF00A0BF00AFC0
00AFCF00BFD08F001F0000009000209F002000D0E000DFEF00DFEF00E0F000DFEF00E0F000E0F0\
B0002FAF002000000000F0F000FFFFB00020B0002FB0002000FFFF00F0F000FFFF
00F0FFBF002FB0002000FFFF00F0F000FFFF00F0F000FFFF00000000F0F000EFF000F0F000E0E0\
00DFDF8F001F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000DFEF00E0F000DFEF00E0F0A00020\
AF002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF000000BF002F
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00000000FFFF00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF000000C0002F00000000FFF000F0EF00F0EF
80B06080B06080AF5F80B0607FAF5F80AF607FAF5F0000007FA05F7FAF5F7FA0507FA050709F50\
7F9F5070904F70904F70904F70904F6F8F4000B0A000AF9F00AF9F00A08F00A08F
800010009F807F001000907F008F70008070007060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F0010005F70000000405F3F4F60404F6F4F50705000A0BF00AFC000AFCF
00BFD000BFD09000209000209F002000D0EF00DFEF00DFEF00E0F000E0F000EFF000EFF0000000\
AF002000000000F0FF00FFFF000000BF002FB0002000FFFF00FFFF00FFFF00FFFF
00FFFF000000BF002F00FFFF00FFFF00F0F000FFFF00F0F000F0FF00F0F000F0F000E0E000DFD0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000090002000DFEF00E0F0\
000000AF002F00000000F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
BF002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00000000000000FFFF00FFFF00FFFF00F0F0
00000080B0608FBF6F80B06080B06080AF5F80B0607FAF5F00E0D000000080AF5F7FA0507FA05F\
709F507FA050709F50709F5070904F70904F00BFA000BFA000AF9F00AF9F00A090
00A08F009F80009F8000000000907F008F70008070007060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000040001000607F3F503F405F404F604050704F50704F00A0BF00AFC000B0CF
00B0CF9000209000209F002000CFE000D0EF00D0EF00DFF000DFEF00E0F000E0F000EFF0AF0020\
00000000F0F000FFFF00F0F0000000B0002000FFFF00F0F000FFFF00F0FF00FFFF
00F0F000FFFF00000000FFFF00F0F0B0002F00F0F000F0F000EFF000000000DFDF00D0CF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000D0EF\
70AF6F00E0EFA0002FA0002000000000F0F000FFFF00F0F000FFFF00FFFF00FFFF
00F0F000000000FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF000000000000\
00FFFF00FFFF00FFFF00F0F000FFFF00000000FFFF00F0F000FFFF00FFFF00FFFF
00F0F000000080B06080B06080AF5F80B0607FAF5F80AF607FAF5F00E0D000DFCF00DFD0000000\
7FA050709F507F9F5070904F70904F6F904F00000000B0A000B0A000AF9F00AF90
00A08F8F001F009F80009F80000000009070008F6F700010006F60000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000607000607F405F404060404F6F4F50704F5F7F5000A0BF60805F608F50
00C0DF9000209F002000000000D0EF00D0EF00DFF000DFEF00E0F000E0F000EFF0AF0020000000\
00F0FF00FFFF00F0FF00FFFF000000BF002F00F0FF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFFBF002FB0002000000000F0F000F0FF00F0F0AF002F00DFD000D0CF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00B0BF70AF6F70AF607FB06F000000AF002FAF002000FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
C0002000000000FFFF00FFFF000000C0002FBF002000FFFF00FFFF00FFFF00FFFF
00FFFF00F0F08FBF6F80B0608FBF6F80B06080B06080AF5F80B06000000000E0DF00DFD000E0D0\
0000007FA05F7F9F507FA050709F50709F506F904F00C0AF00B0A000B0A000AF90
00AF9000A08F00A08F009F80009F8000000000907F008F6F700010600010000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000005F70006F80405F3F4F60404F6F4050704F50704F0000005F8050608F5F
608F5F90002090002000000000D0EF00DFEF00DFEF00E0F000E0F000EFF000000000000000F0F0\
00FFFF00F0F000FFFF00F0FFBF002F00F0F000FFFF00F0FF00FFFF00F0F000FFFF
00F0F0BF002FB0002000F0F000F0F000000000EFEFA0002000DFD000D0CF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000C0D06FA06070AF6F70AF6000EFF000EFF000000000F0F000FFFF
00FFFF00FFFF00F0F000000000000000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFFBF0020000000000000C0002FBF0020C0002F00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F000000080B06080B06080B06080B06080AF5F80B06000000000E0D000DFD0\
00DFD000D0CF00DFCF00D0C000000070904F70904F00BFAF00BFAF00B09F00B09F
00AF9000AF9000000000A08F009F80009F7F7F001000000000806F007F6F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
005F6000607F006F804060404F604050704F50704F0000005F8050608F5F608F5F
00000000CFE09F002F9F002000DFEF00DFEF00E0F000E0F000EFF000EFF000F0FF00F0FF00FFFF\
00F0FF00FFFF00FFFF00FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
BF002FB0002000000080BF6F80BF6F7FB060AF002F00DFDF00D0D0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006FA06070AF6F70AF6F00EFF000EFF000F0FF00F0FF
B0002FB0002000FFFF00FFFF000000BF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFFC00020C0002FBF0020C0002FC0002000000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00F0F000F0F080B0608FBF6F80B06080B06080AF5F80B06000E0DF00E0DF\
00DFD000E0D000DFCF00DFCF00D0C000D0C000000000C0B000BFAF00BFAF00B0A0
00B09F00AF9000AF9000A08F00A08F009F80009F807F0010000000000000007F6F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
005F6000607F000000405F3F4F6F404F6F4F5F7F5000A0BF5F805060805060905F
00C0DF00CFE000D0E000000000DFEF00E0F000E0EF00EFF000E0F0B0002F00F0F000F0FF00F0F0\
00FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF000000000000
00F0F00000007FBF6F80BF6F7FB06000EFEF9F002000DFD0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000B0BF70A06F6FA06070AF6F00E0F000EFF0
00EFF000F0FFB00020B0002FB00020BF002FB0002000000000F0F000FFFF00FFFF00FFFF00F0F0\
00FFFF000000C0002FBF0020C0002FBF0020C0002F00000000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF00F0F000F0F000F0EF80B06080B06080B06000000000EFE000E0D0\
00E0DF00DFD000DFD000D0CF00D0C000CFBF00CFBF00000000000000BFAF00BFA0
00B09F00B09F00AF90B0A090009F80009F8000907F00907F7F00107F00106F0010600010000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
005F6F006F800000004F60404F6F4F507050009FBF0000005F8050608F5F60905F
00CFE000CFE000D0EF00D0EF00000000E0F000EFF000EFF0B0002F00F0F000F0FF00F0FF00FFFF\
00F0FF00FFFF00F0FFBF002F00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF
00FFFF80BF6F80BF6F7FBF6F00F0EF00E0DF00DFD0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000CFDF70A06070AF6F00E0F0
00EFFF00EFF000F0FF00F0FF000000B00020BF002FB00020BF002FBF0020BF002F00000000FFFF\
000000C0002FBF0020C0002FC00020C0002FBF002000000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFF000FFF000F0EF00F0F000000000EFEF00EFE000EFE0\
00E0DF00E0DF00DFD000E0D000DFCF00DFCF00CFBF00CFBF70904F70904F6F8F4F
00000000B09F00B09F00AF9000AF9000A08F009F8000907F00907F7F00100000006F0010000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
005F6F006F80405F4040604050704F0090B000A0BF5F7F50608F5F608F5F6F905F
00CFDF00D0EF00D0E000DFEF000000AF002F000000AF002F00EFF000F0FF00F0F000F0FF00F0FF\
00FFFF00F0F0000000BF0020000000000000BF002F00F0F000FFFF00F0F000FFFF
80BF6F80BF6F7FB06000EFEF00E0DF00000080001F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000D0EF000000
00E0F000EFF000EFF000F0FF00F0F000FFFF00F0F0B0002FB00020BF002FB00020000000000000\
00FFFF000000000000BF0020C0002FBF0020C0002FBF002000FFFF00F0F000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000F0F0000000BF002000000000EFE000EFE0\
00EFE000E0D000E0D000DFCF00DFCF00D0C000D0C000CFBF709F506F904F70904F
6F8F4000BFA000AF9F00AF908F001000A08F009F80009F8000907F00907F70001000806F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000708F405F404F6F4F0090AF00A0BF00A0BF60805F608F5F6F906000C0DF
00CFE000D0E000DFEF00DFEF00E0F0AF0020B0002FAF002000F0FF00F0FF00FFFF00F0FF00FFFF\
00F0FF00FFFFBF0020BF002FBF0020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF
80BF6F00000000F0F000E0E000E0DF8F001F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000000BFCF
00000070AF6000EFF000EFF000F0FF00F0F000FFFF00FFFFBF002FB0002000000000FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF000000000000C0002F00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000C0002FBF0020BF002FBF002000EFEF\
00EFE000EFE000E0DF00E0DF00DFCF00DFD000D0CF00D0C0709F50709F5070904F
70904F6F8F4000BFA000B09F90001F00000000A08F009F80009F8000907F0000007F00106F0010\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
005F70006F80000000008FA0009FB0009FB000AFC0000000608F5F00000000CFE0
00CFE000D0EF00D0EF00E0F000DFEFAF0020AF0020B0002FB0002000000000F0FF000000B00020\
BF002FBF0020BF002FBF002000FFFF00F0F000FFFF00F0F000FFFF00F0F0000000
00000000F0F000E0E000E0DF9F001F00B0AF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8F001F00D0E06FA06070AF6F00000000EFF000F0F000FFFF00000000000000FFFF00FFFF00F0F0\
00FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0000000BF0020000000
00000000000000000000FFFF00F0F000FFFF000000C0002FBF0020BF002FBF0020BF002FB00020\
B0002000E0DF00EFDF00E0D000E0D000DFCF00DFCF00D0C000D0C0709F4F709F50
6F904F6F904F6F8F4000B0A000AF908F001F00A08F00A08F009F80009F7F00907000000070000F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
005F700000004060400090AF009FB000A0BF00AFC000B0D000000000C0DF00CFE0
00D0EF00D0EF00DFEF00DFEF00EFF0000000B0002FB00020BF002F00F0FF00FFFF00F0FF000000\
BF0020BF002FBF002000FFFF00FFFF00FFFF00FFFF00FFFF00000080BF6F80BF6F
00000000EFE000000070A05F00CFC07F001F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F0010000000005F5F006060006060006060000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000CFE06FA06070AF6F70AF6F00EFFF00F0F080BF6F80BF6F80C06F00000000FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002FBF0020
C0002FC00020C0002FBF0020C0002F000000000000000000000000000000C0002FBF0020BF002F\
BF0020BF002000E0DF00EFDF00E0D000E0D000DFCF00DFCF00D0C07FA050709F4F
709F506F904F70904F00000000B0A08F001F8F001F00A08F00A08F009F80009F800090707F0010\
807010000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F503F4060400090AF009FBF00A0BF00AFC000B0CF8F002000000000CFE0
00D0E000DFEF00DFEF00E0F000E0F000000000EFF000F0FF00F0FF00FFFF00F0F000FFFF00F0FF\
00FFFF00F0F000000000FFFF00FFFF00F0F000FFFF80BF6F80BF6F7FBF6080BF6F
00000000EFE070A05F70A05F00B0AF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000006060005F5F006060
006060006060000000600010006060006F6F006060006F6F006060006F6F006060006F6F006F60\
006F6F006060006F6F006F6000000060000F600010000000006F6000605F006060
005F5F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000C0DF6FA06070AF6F70AF6000000000EFF080BF6F7FBF6F80BF6F80BF6F\
00000000F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFFBF0020C0002F
BF0020C0002FBF002000000000FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0000000BF0020\
BF0020B00020B0002000E0DF00E0DF00DFD000DFD000DFCF00DFCF00D0C0709F50
70904F70904F6F8F406F8F4F00B09F90001F8F001000AF9000A08F00A08F009F7F00907F008F70\
700010000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F0010405F404060400090B0009FB000AFC000B0CF8F00208F00209F0020000000
00D0EF00DFEF000000AF0020AF002FAF002000000000F0FF00FFFF00F0FF00FFFF00FFFF00FFFF\
00F0FF8FC06F8FBF6F8FC06F00FFFF00FFFF80BF6F80BF6F80BF6F80C06F7FBF6F
00000070AF5F7FAF6000CFC0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000005F60006060000000006F6F006060006F6F600010
6F00106F00106F00106F001000707000706F0070700000000000006F0010000000000000007070\
00706F00707000706F00706F00706F00706F00706F00706F6F00107000106F0010
006F6F006F60006F6F000000600010006060006060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F905F6FA06070AF6F70AF6F7FB06F00EFF080BF6F80BF6F80C06F\
80BF6F8FC06F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002FC00020
C0002FBF0020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002F\
BF0020000000000000BF002000E0DF00E0DF00E0D000E0D000DFCF00DFCF00D0C0
7F9F5070904F709F506F904F6F904F90001F90001F8F001F00AF9000A08F00A08F009F7F00907F\
70001030300F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F00103F503F4F60400090AF00A0BF00A0C00000008F001F900020900020000000
000000000000A00020AF002FAF0020B0002FB0002000F0FF00F0F000FFFF00F0FF00FFFF000000\
8FBF6F80BF6F8FBF6F80BF6F00FFFF80BF6F80BF6F80BF6F80BF6F80BF6F80BF6F
00000070AF5F70A05F00BFB0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000003F402F\
404F2F404F2F4F502F404F2F4F502F4F4F2F006F6F0000006F00106F0010700010
00706F007070007070007F70007070007F70007F70007F70007F70007F70007F70007F70007F70\
007F70007F707000107000107000107000107000107000107000107000105F7F3F
8F70107000106F001000000060000F6F001060000F00000000605F006060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000608F5F6FA06070AF6F70AF607FB06F7FB06080BF6F7FB06F\
80BF6F80BF6F80BF6F00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000C0002F
000000000000BF002000000000F0F000FFFF00FFF000FFFF00F0F000FFFF00FFF000FFFF00F0F0\
BF002000F0EF00F0EF00EFE000000000000000E0DF00DFD000DFD000D0CF00DFCF
000000709F5070904F70904F6F8F400000008F001F90001F8F001000A08F009F80009F80009070\
0090708F7F10000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F0010405F40406040009FB000A0BF00AFC080001F90002090002000D0E000D0E0
00DFEF00DFEF00E0F0000000B0002FB00020B0002F00F0FF00FFFF00F0FF00FFFF00F0FF000000\
80BF6F8FBF6F80BF6F8FC06F80BF6F8FBF6F80BF6F8FC06F80BF6F80BF6F000000
7FB06070AF5F9F002000B0A0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000006060404F2F4F5030404F2F4F5030\
4F502F4F5030006F6F007070007070007070007070007F7F007F70007F7F007F70
007F7F007F7F00807F007F7F00807F00807F00807F00807F00807F00807F00807F00807F00807F\
00807F00807F00807000807F00807000807F00000000000010100060803F907F10
4F4010907F10907F108F7F10907F1F8F70108F7F106F001020200F6F0010006F6000605F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000608F5F6F9F6070AF6F70AF6F7FB06F7FB06F80BF6F\
80BF6F80C06F80BF6F8FC06F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000
00FFFF00FFFF00FFFF000000C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00000000000000F0EF00F0EF00EFE000EFE0000000B0002000000000DFD000DFCF
00DFCF709F50709F5070904F70904F6F8F4F00BFA08F001F90001F8F001F00A08F009F80009F80\
0090707F00107F6F10000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5000103F503F4F60400090AF00A0BF00A0C000000000BFD000000000CFE000D0EF
00D0EF00E0F000E0EF00EFF0AF0020000000B0002000000000F0FF00FFFF00F0F000FFFF80BF6F\
8FBF6F80BF6F8FBF6F00FFFF8FBF6F80BF6F80BF6F80BF6F80BF6F7FB06000EFE0
000000A0002000BFB000B09F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000006060005F6000606F00606F000000006F6F007070007070\
007070007070007F7F007070007F7F007F70007F7F007F7F7F001000000000807F
007F7F00808000807F00808000807F00808000807F0000000000007F00107F00107F00107F0010\
7F00107F00107F00107F00107F00107F00107F00107F00107F00107F00102F200F
9080109F801F907F109080108F7F10907F108F7F108F7F108F70108F7F108070108F701060000F\
600010000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006080506F9F5F70AF6F70AF6070AF6F7FAF60\
00F0FF7FB06080BF6F80BF6F80BF6F00F0F000FFFF00FFFF00000000000000FFFF
00FFFF00FFFF00F0F000FFFF00000000FFFF00F0F000FFFF00FFF000FFFF00F0F000FFFF00FFF0\
00FFFFBF002000F0F000EFEF00EFEF00E0E000EFDF00E0DFB00020AF001FAF0020
A0001F000000709F4F709F506F904F70904F00BFA00000008F0010000000000000000000000000\
0000007000108F7F10000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F0010405F404F6040009FB000000000AFCF00B0CF00C0DF00CFE000D0EF00D0EF
00DFEF00DFEF00EFF000000000F0FF00F0FF00FFFF00F0FF00000000000000FFFF00FFFF8FBF6F\
80BF6F8FC06F00FFFF00FFFF80BF6F8FC06F80BF6F80BF6F80BF6000F0EF00E0E0
00000000D0C000C0B080001F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000006060006060006F6F0000006F00106F00106F00106F0010000000000000007F7F\
007F7F008080007F7F008080007F7F00808000807F7F001F7F00107F001F000000
0000007F001080001F000000008F80008F80008F8F008F8000908F008F8000908F008F80009080\
008F80009080008F80008F8000000000000000000080001F000000A08F1F9F8F10
9F8F1F9F8F109F8F1F9F80109F801F9080109F801F90801090801F8F7F1090801F8F7F108F7F10\
8070106F0010006F60006060000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000C0DF70AF6F70AF6F7FB06F\
7FB06F00F0FF7FBF6F80BF6F80BF6F8FC06F000000BF002FBF0020C0002F00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFF000FFFF00F0F000F0F000EFEF00F0EF00EFE000EFE000E0DFB00020AF0020
AF0020A0001FAF002000CFBF0000006F904F00C0AF00BFA00000008F001F00AF9000A08F80001F\
7F00107F001F908F107F7010000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5000103F5F3F4F6F4F6F001F7F001F00AFC000BFD000BFD000000000CFE000DFEF
00DFEF00E0F000E0F0AF002F00EFF000F0FF00F0FF00FFFF00F0F0BF002FBF0020BF002F000000\
00000000000000FFFF00F0F08FBF6F80BF6F80BF6F80BF6000F0F000EFE000E0DF
A0001F00D0C000BFAF600010000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000006060
005F600000005F001060001060001000707000707000707F007070007F7F007F7F7F0010700010\
7F00107000100000000000000000000000007F001F7F00108000107F001080001F
008F80008F8F008F8000908F008F8000908F008F8F00908F008F8F00908F00908F00908F008F80\
00908F00908000908F008F8000908F80001080001F7F001080001080001080001F
7F00108000102F20009F8F1F9F8F109F8F1F9080109F8F1F9080109080109080109080108F7F10\
8F7F108F7010302F0F60000F006F6000605F000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000C0D070AF6F6FA060\
70AF6F70AF6000F0FF00F0F000000000FFFF000000B00020BF002FBF0020000000
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00000000FFFF00F0F000FFFF00FFF000FFFF00F0F0\
00FFFF00FFF000FFFF00F0F000F0EF00EFEF00EFE000E0DF00E0DF00E0D0000000
AF001FAF0020A0001FA0002000CFBF00CFB000C0AF00C0AF00B0A090001F00AF9000AF8F009F80\
8000107F00107F0010807F10000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
500010007F90008FA070002000000000B0CF00B0CF00C0DF00CFE000D0EF00D0EF
00E0F000E0F000EFF000EFF000F0FF00F0FF00FFFF00F0FF00FFFFBF0020BF002FBF002000FFFF\
00FFFF00FFFF00FFFF00FFFF80BF6F8FC06F80BF6F00FFFF00EFEF5F00105F0010
5F0010000000000000006060006F6F006060000000404F2F4F5030404F2F4F5030404F2F4F5030\
4F4F2F4F50304F4F2F4F50304F502F4F5030000000006F6F006F6F0000004F4F2F
4F50304F4F2F4F5030404F2F4F50304F4F2F4F5030404F2F4F5030404F2F4F5030404F2F4F5030\
0000000000000000000000000000000000000000000000005F001000606F00606F
006F70006F70000000007070007F7F00707F007F7F007F7F008080007F800080800000007F001F\
7F001080001F7F001080001F008F8F008F8F008F8F00908F00000080001F800010
00000000908F00909000908F009F9000908F009F9000908F009F90009F90009F9000908F009F90\
00908F009F900000008F001F8000108F001F8000108F001F8000108F001F800010
8F001F000000A0901FA09010A0901F9F8F1FA0901F9F8F109F8F1F9F8F109F8F1F9080109F8F1F\
90801090801F8F7F1010100F6F00106F0010006F5F000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000BFD070AF6F\
70AF6F7FB06F70AF6F00F0FF00F0FF00FFFF00FFFFBF002FBF0020BF002FBF0020
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFF000FFFF\
00FFFF00FFFF00FFF000FFFF00F0F000F0F000EFEF00F0EF00EFE000EFDF00E0D0
00E0D0AF0020AF0020A0001F00000000CFBF00CFB000C0AF00BFAF00000000B09F00A08F00A08F\
0000007F00107000105F8F3F706F10000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
50001000708F0000006F001F00A0BF00AFC000BFD000C0DF00000000CFE000DFEF
00DFEF00E0F000E0F000EFF000F0F000F0FF00F0F000F0FF00F0FF000000BF0020BF002F00F0FF\
00606000000000000000606000606F00606000606F006060006F6F006060006F6F
006060006F6F006060006F6F00606F0000000000004F50304F4F2F4F5030000000006F6F006F6F\
007070006F6F00706F006F6F007070006F6F00706F006F6F00706F006F6F006F6F
0000005050304F502F4F50304F502F5050304F4F2F4F50304F502F4F50304F4F2F4F50304F502F\
4F50304F4F2F4F502F4F4F2F4F5030404F2F4F4F2F000000006F6F006F6F006F70
006F7000707F007070007F7F007F7F007F80007F7F008080007F800080807F00107F0010000000\
80001F7F001080001F000000009090008F8F00909000908F00909000908F009090
009090000000009090009F90009F90009F90009F90009F90009F90009F90009F90009F90000000\
8F001F800010000000000000009F90009F8F009F90009F8F009F90009F8F009F8F
009F8F8F001F80001080001080001080001F3F300FA0901F9F8F10A0901F9F8F109F8F1F908F10\
9F8F1F9080109080108F7F108F7F100000006F001060000F600010000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000BFCF\
00D0EF6FA06070AF6F70AF6000EFFF00EFF000FFFF00FFFFBF002FB00020BF002F
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00000000FFFF00F0F000FFFF00FFF0\
00FFFF00F0F000FFFFBF002000FFF000F0EF00F0EF00EFE000EFE000E0DF00E0DF
00DFD000E0D0A0001FAF0020A0001F00000000C0B000C0B000BFA000BFA08F001000AF9000A08F\
00A0807F00107F0010700010807F10000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000007F906F001F70002000A0BF00B0CF00BFD000C0DF90002000D0EF00DFEF
00E0F000E0EF00000000EFF000F0FF00F0FF00FFFF5000100000005000105F001000606000606F\
00606F006F6F00606F0000006000100000000000000000000000000000006F0010
0000006F00107000106F00107000106F0010700010700010700010700010700010700010700010\
000000007F70007070007F70007070007F70007070007F70007070007F70007070
007F70007070007F70007070007F7000706F007F70505F30505F3050502F505F30505F30505F30\
50502F505F30505030505F3050502F505F304F502F5050304F502F5050304F502F
006F6F006060006F6F00606000808000808000808F0080807F001F00808F008F90008F8F000000\
800010000000009090009090009090009F9F009090009F9F009090009F9F009F90
009F9F00000090001F8F001F90001F009F9F00A09F00A09F0000008F001F00000000A09F00A09F\
00A09000A09F00A09000A09F00A09000A09F00A09000A09F009F9000A090009F90
0000008F00108F001F8F00108F001F8000108F001F2F2F0FA09F1FA09010A0901F9F901FA0901F\
9F8F109F901F9F8F10908F1F908010700010700010000000007060007060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00C0D000DFF070AF6F70AF6F00EFF000F0FF00F0FF00FFFFB00020BF002FBF0020
BF002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF\
00FFF000FFFF00FFF000000000000000FFFF00F0F000F0EF00EFE000EFE000E0DF
00E0DF00E0D0000000AF0020AF0020A0001F00D0BF00C0B000C0AF00BFA090001F00AF9000AF90\
00A080009F807F00107F0010000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
006070007F906F001F6F001F00A0BF00AFC000BFD000C0DF00000000D0E000DFEF
00DFEF0060600000005F001050000F5F001000606000606F5F000F60001000606F006F6F00606F\
006F6F006F6F006F70006F6F006F70006F6F007070006F6F007070000000700010
00000000707000707000707000707000707000707000707F00707000707F007070007F7F007070\
007F7F007F70000000000000000000007F70007F7F007070007F7F007F70007F7F
007070007F70007F70007F7F000000000000000000000000505F305F5F30505F305F5F30505F30\
5F5F3000000000000000706F007F7000706F007F7000706F000000000000000000
006F6F0000000000000000000000004F502F4F4F2F4F503000000000606F008F8F009090009090\
000000009090009F9F009090009F9F009F90009F9F009F9F009F9F009F9F00A09F
009F9F00A09F8F001F90001F8F001090001F8F001F90001F8F001000AFA000A09F00AFA000A09F\
00AF9F00A09F00AF9F00A09F00A09F00A09F00AF9F00A09000A09F00A09000A09F
0000008F001F8F00108F001F8000108F001F80001000000080001080001080001080001F7F0010\
A0901F70902F800010101000808F207000104F80406F000F6F0010007060007060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000C0D000DFF06FA06000E0F000EFF000F0FF00F0F000FFFFB00020BF002F
00F0F0BF002F00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0C0002FBF002000FFFF00F0F0\
00FFFF00FFF000FFFF00F0F0BF002F00FFF000FFF000F0EF00F0EF00EFE000EFE0
00E0DF00E0DF00DFD0AF0020000000A0002000000000CFBF00C0AF00BFAF00B09F00B09F00AF90\
00A08F009F80009F7F7000107000107F7010000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
005F6F007F906F001F70002000A0BF00B0CF00BFD000606F50001000606F006060
00606F5F00106000105F001060001000606F006F6F00606F006F70006F6F006F70006F6F007070\
006F70007070006F7000707F00707000707F00707070001F00000000707F007070
007F7F00707F007F7F00707F007F7F007F7F007F7F007F7F007F80007F7F007F80007F7F008080\
007F7F008080007F7F0080800000007F001F7F001080001F7F0010000000007F7F
008080007F7F00807F007F7F0000005F60305F603F5F5F305F603F5F60305F603F5F5F305F603F\
5F60305F603F5F60305F603F5F60305F6030007F7000807F007F70007F70007F70
007F70007F70007F7000706F007F7000706F00706F50502F505F304F502F5050304F502F505030\
4F502F000000009F9F00A0A000A09F00A0A000A09F00A0A000A0A000AFA000A0A0
00AFA000A0A000000090001F9F001F90001F90001F90001F9F001F90001F00000000AFA000B0A0\
00AFA000AFA000AFA000AFA000AFA000AFA000AF9F00AFA000AF9F00AF9F00A09F
90001F8F001F90001F00000000AF9F00A09000A090009F8F00A0908F00108F001F8000108F001F\
8000100000007F001080001F7F00100000007000107F00100000007000106F0010
007060000000000000000000000000000000000000000000000000000000000000000000000000\
00000000B0CF00C0DF00E0FF00000000EFFF00EFF000F0FF00F0FF00FFFFB00020
00FFFF00FFFFBF002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002F00000000FFFF\
00FFFF00FFFF00FFF000FFFFBF0020C0002F00FFF000FFF000F0EF00F0EF00EFE0
00EFE000E0DF00E0DFAF0020000000000000A0002000CFBF00CFB000C0AF00BFA000000000B09F\
00A08F00A08F009F7F000000700010908F10000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000070806F001F009FB0000000005F6000606F0000005F001000606000606F
00606000606F00000000000000606F006F7000606F006F70006F70007070006F70007070007070\
00707F00707000707F70001070001070001000000000707F007F7F00707F007F7F
007F7F007F80007F7F007F80007F7F007F80007F7F008080007F80008080007F80008080008080\
0080800080800080800080800080800080800000007F001080001F7F0010800010
7F001080001F7F00108000100000000000005F60305F603F5F60305F603F5F60305F603F5F6030\
5F603F5F60305F603F5F60305F603F5F60305F603000000000807F007F7000807F
007F70008070007F70007F70007F70007F70007F6F007F7000706F007F7000706F00000050502F\
505F304F502F50502F4F4F2F006F6F60000F0000006F7F406F804F6F7F406F804F
6F80406F804F00000000AFAF00000000B0AF0000000000000000000000000000009F001F00AFA0\
00B0AF00AFA000B0AF00AFA000B0A000AFA000B0A000AFA000B0A000AF9F000000
90001F00AFA000AF9F00AF9F00A09F00AF9F00A09000A09000A0900000008000108F001F800010\
000000009080009F807F00108000107F0010A08F7F7000107F001000000000806F
00000000706000605F000000000000000000000000000000000000000000000000000000000000\
00000000000000B0CF00CFDF00000070AF6000E0F000EFF000F0FF00F0F0BF002F
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF000000C0002FBF0020\
00FFFF00F0F000FFFF00FFF0000000BF002000000000F0F000FFF000F0EF00F0EF
00EFE000EFE000E0D0AF002000DFCF00DFCF00000000000000C0B000C0B000BFA000B0A000AF90\
00AF9000A080009F8000907F7F0010508F307F7010000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000606F00606000606F00606000606F00606F00606F00606F006F7000606F
0000006000106F0010000000006F70006F7000707F00707000707F00707F007F7F00707F000000\
7000107F001F7000107F001F700010000000007F80007F80007F80008080007F80
008080007F8000808F00808000808F00808000808F00808000808F008080008F8F00808F008F8F\
00000000000000000000000000808F00000000000000000000808F000000800010
80001F80001080001F800010000000008F80008F8F008F80008F8F008F800000005F6F3F606F3F\
5F6F3F606F3F5F6F3F606F3F000000606F3F5F603F00000000807F008F8000807F
008F7F00807F00807F00807F00807F00807000807F008070008070007F70008070000000000000\
0000007F0010000000000000000000000000006F6F006F6F0060600000006F804F
708F4F6F804F708F4F00B0AF00B0B000B0AF00BFB000B0AF00BFB000B0AF00BFB000B0AF00BFB0\
00B0AF0000009F001F00BFAF00B0AF00BFAF00B0AF00B0AF00B0A000000000B0A0
00B0A000B0A000B0A000AFA000B0A000AF9F00AF9F00AF9F00AF9F00A0900000008F001F8F001F\
00000000A090009F8F009F8F80001080001F7F001000907F7F00107F001F700010
008070007F6F007F6F006F5F000000000000000000000000000000000000000000000000000000\
00000000000000000000BFD000CFE070AF6F70AF6F00000000EFF000F0FF00F0FF
00000000FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF000000C0002F\
BF002000FFFF00FFFF00FFFF00FFF0C0002FBF002000000000FFF000FFF000F0EF
00F0EF00EFE000EFE000000000E0D000DFCF00DFCFA0001F00CFBF00C0B000C0AF00BFA000B09F\
00AF9000AF9000A080009F8000907090901F8F8010000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004F000F006060005F60\
00606000606000606F00606000606F00606F006F7000606F006F70006F70000000
6000106F00106F001000000000000000707F00707F00707F00707F00000070001070001F700010\
7000107000107F001F00000000808F007F8000808F007F8000808F00808000808F
00808000808F00000000000000000080001F7F001080001F7F0010000000008F8F008F90008F8F\
008F90008F8F008F90008F8F008F90008F8F008F90008F8F008F8F008F8F009090
008F8F0000008000108F001F000000008F8F008F8F00908F008F8F008F8F008F8F00908F008F80\
00908F008F8F00908F008F80008F8F008F80008F8F00000080001080001080001F
80001080001080001080001080001080001080001080001000807000807F00807000807F007F70\
0080700000000000000000007F00107000107F001000706F00706F006F6F006F6F
0060600000006F8F4F00BFB000B0B000BFB000B0B000BFB000BFB000BFB000BFB000BFB000BFB0\
00BFB000BFB00000009F001F9F001F9F001F9F001F90001F9F001F00000000BFAF
00B0A000B0AF00B0A000B0AF00AFA000B0A000AF9F00AFA000AF9F00AF9F00A0908F001F8F0010\
00A09000A08F00A090009F8F009F8F80001080001F00907F00907F008F7F7F0010
700010008070007F6F00706F006F5F000000000000000000000000000000000000000000000000\
00000000000000000000000000BFD000CFE070AF6F70AF607FB06F00EFF000F0FF
00000000FFFF00FFFF00FFFF00F0F000FFFFBF002000FFFF00F0F000FFFF00FFFF00FFFFBF0020\
C0002FBF002000FFFF00F0F000FFFFBF0020C0002FBF002000FFF000F0F000F0F0
00EFE000EFE000000000E0DF00DFD000DFD000D0CF00D0C000000000CFB000C0AF00BFAF000000\
00B09F00A08F00A08F009F7F00907F7000108F8F10000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004F001000606F00606000606F00606000606F\
00606F00606F00606F006F70006F70006F70006F7000707F006F706F00106F0010
70001F6F001070001F6F001070001F007080007F807000107F001F7000107F001F7000107F001F\
7F00107F001F00000000809000808F7F001F7F001080001F7F001080001F7F0010
80001F80001F80001F80001080001F000000009090008F90009090008F90009090008F9000909F\
00909000909F00909000909F00909000909000909000909F009090009090009090
00909000909000909000000000000000909000909000909000909000909000909000908F009F90\
00909000909000908F00909000908F00909000908F0000008F00108F001F800010
8F001F8000108F001F800010000000008F80008F80008F7F008F80008F7F008F8000807F008F7F\
00807F008F7F008070008F7F008070008070007F700000007F00107F0010700010
007F6F00706F00706F006F6F006F6F00606000C0BF00C0BF00C0BF00C0BF00C0BF00C0BF00C0BF\
00C0BF00C0BF00C0B000C0BF9F001FA000209F001FA000209F001F9F001F9F001F
00C0B000BFAF00BFB000BFAF00BFAF00B0AF00BFAF00B0A000B0A000AF9F00B0A000AF9F000000\
00AF9000AF9F00A09000A09000A08F00A08F80001080001F009080009080008F7F
00907F700010008F70007F6F007F6F006F60000000000000000000000000000000000000000000\
00000000000000000000000000000000C0DF00D0EF70AF6F70AF6F00000000EFF0
00000000F0FF00FFFF00FFFF00FFFF00FFFF00FFFFBF002000000000FFFF00FFFF00FFFF00FFFF\
C00020C0002FBF002000FFFF00FFFF000000BF0020C0002FBF002000FFFF00F0F0
00F0F000EFE000000000000000E0DF00DFD000DFCF00D0C000D0C000CFBF00CFB000BFAF00BFA0\
00B09F00AF9000A08F00A08F009F7F0000007000108F8F1F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000304F2F3F4F2F005F605F001000606000606F00606000606F00606F\
006F7000606F006F70006F70006F7F006F7000707F0000006F00106F00106F0010
6F00107000106F0010700010407F4F507F4F0F0F0070001F7000107F001F7000103F300F90801F\
0000007F00107F001F7F001080001F7F001080001F7F001080001F7F001080001F
7F001080001F000000008F90008F9000909F008F90009090008F9000909F008F9000909F009090\
00909F00909000909F00909000909F00909000909F00909000909F00909000909F
009090009F9F0090900000008F0010000000009090009F90009090009F9F009090009F90009090\
009F90009090009F90009090009F9000908F009F900000008F001F8F0010000000
00000000908F00908000908F0000008F001F800010000000008F80009080008F80008F80008F7F\
008F80008F7F008F7F00807F008F7F00807000807F008070008070000000000000
007F6F007F6F00000000000000706F00706F006F60006F6F60000F00CFC000C0BF00CFC000C0BF\
00000000000000000000000000CFBF000000A0001F9F001FA000209F001F000000
00000000000000000000000000BFAF00BFB000B0AF00BFAF00B0A000B0A000B0A000000000AF9F\
00B09F00AF9000AF9F00A09000A090009F8F00A08F800010009F8000907F009080
008F7F00907F700010008070007F6F6F00105F000F000000000000000000000000000000000000\
00000000000000000000000000000000000000C0DF6FA06070AF6F70AF6000EFFF
000000B0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFFBF0020BF002F00F0F000FFFF00FFF0\
000000BF0020C0002FBF002000FFFF00F0F0BF002FBF0020BF002F00F0F000FFF0
00F0EF00F0EFB00020B0002000E0D000E0D000DFCF00DFCF00D0C000000000C0B000C0AF00BFA0\
90001F00AF9000AF9000A080009F800000007000108F8010000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F4F303F4F2F5F00105F001060001000000000606F00606F006F70006070000000\
006F7000000000000000000000707F0070806F001070001F6F001070001F700010
70001F6F7F3090801F907F1F90801F90801F0000001010009F801F90801F9F8F1F9F801F1F100F\
7F001080001F7F001F80001F7F001080001F7F001F80001F80001080001F008F90
00909F008F9000909F00909F00909F00909F00909F00909F00909F00909F009F9F00909F009F9F\
00909F009FA000909F009F9F00909F009F9F00909F009F9F00909F009F9F009F9F
009F9F009F9F009F9F0000008F001F8F001F90001F000000009F9F009F9F009F9F009F9F009F9F\
009F90009F9F009F9F009F9F009F90009F9F000000000000009F90009F9F009F90
009F90009090009F9000908F009F9000908F009F9000908F00908F00000000908F00908000908F\
00908000908F009080009080008F7F009080008F7F008F7F00807F008F7F008070
00807000807000807000806F008070007F6F007F6F00000000706F006F606F00100060607F9F5F\
709F507F9F5F7F9F507F9F50709F507F9F5000000000CFC000CFBF00CFC000CFBF
00CFBF00C0BF00CFBF00C0BF00C0BF9F001FA0002000000000C0B000BFAF00BFAF90001F000000\
00B0A000B0A000B09F00B09F00AF9000AF9000A09000A090000000009F8F009F80
009F8000907F00907F008F70008F70007F6F6F0010000000000000000000000000000000000000\
00000000000000000000000000000000000000B0CF00CFE06FA0607FAF6F70AF6F
000000AF002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFFBF0020C0002FBF002000FFFF\
00FFFFC0002FC00020C0002FBF002000FFFFBF0020C0002FBF002000000000FFF0
00FFF0000000BF002FB0002000EFDF00E0D000E0D000DFCF00DFCF00D0C000D0BF00C0B000C0AF\
8F001F00B09F00AF9000AF8F009F80009F80700010700010000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F402F50001000000000606F00606000606F00606F0060705F00106000106000106F0010600010\
6F00106F00106F00106F00106F00108070108F7F1F8F70108F7F1F8F7F10907F1F
8F7F1090801F907F1F90801F907F1090801F90801F9F801F90801F9F8F1F9F801F504010708030\
3F300F101000608F4F3F300F10100000000000000000000000000000000000909F
008F9F00909F00000000909F00909F00909F00909F00909F00909F00909F00909F009FA000909F\
009FA000909F009FA000909F009FA0009F9F009FA000909F009F9F009F9F009FA0
009F9F009F9F009F9F0000008F00108F001F8F001F90001F000000009F9F009F9F009F9F009F9F\
009F9F009F9F00A09F009F9F0000008F001090001F000000009F9F009F90009F9F
009F90009F90009F90009F90009F90009F90009F90009F9000908F009F9000908F009F8F00908F\
8F001F00000000908F009080009080008F80009080008F7F008F7F008F7F008F7F
00807080001000000000807000806F008070007F6F00806F007F6F007F6F00706F000000000000\
6F00107F9F507FA05F709F507FA050709F507FA050709F5000D0C000CFC000D0C0
00CFBF00CFC000CFBF00CFC000C0BF00CFBF00C0B00000009F001F9F001F9F001F9F001F90001F\
00BFAF00B0A000BFA000B09F00B09F00AF9000AF9000A08F00A08F009F8F009F8F
00908000908000907F00907F00806F00806F007F60600010000000000000000000000000000000\
00000000000000000000000000000000000000000000B0CF0000006FA06F70AF6F
70AF60AF002FAF002000F0FF00F0F000FFFF00FFFF00FFFF00F0F000FFFFBF0020C0002FBF0020\
00FFFF00FFF0C0002FBF0020C0002F000000C0002FBF0020BF002FBF002000FFFF
00F0F0000000B00020B0002000E0DF00E0DF00DFD000DFD000D0C000D0C000CFBF00CFB000BFAF\
00BFA08F001F00B09F00A08F00A08F009F7F7F0010700010000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F4F2F500010
00606F00606F00606F00606F006F70006070006F700060706F001F6000106F001F6F00106F001F\
6F001070001F6F00107F7F2F8F7F1F907F1F8F7F1F90801F907F1F90801F907F1F
90801F90801F90801F90801F9F8F1F90801F9F8F1F90801F9F8F1F9F8F1F9F8F1F9F8F1FA09020\
9F8F1FA0901F9F8F1F7F904080001F80001F80001F80001F80001F80001F80001F
8F001F80001F8F001F80001F8F001F80001F8F001F80001F8F001F8F001F8F001F8F001F8F001F\
8F001F8F001F8F001F90001F009FA000A0A0009FA000A0A000000000A0A0009FA0
0000000000000000008F001F90001F8F001F90001F8F001F90001F90001F000000009F9F00A0A0\
00A09F00000090001F90001F90001F90001F90001F00000000A09F00A09F009F9F
00A09F009F9F00A09F009F9000A09F009F9000A09F009F9000A09F009F9000A090009F9000A090\
009F90009F90009F8F90001F8F0010009F8F009080009F8F009080009080008F80
009080008F7F8F0010800010000000008F70008F70008070008F7000806F008070007F6F007F6F\
700010000000600010006F6F7FA05F7FA05F7FA0507FA05F00D0CF00DFCF00D0CF
00DFCF00D0CF00D0CF00D0C000D0C000CFC000CFC000CFBF00CFBF000000A000209F001F9F0020\
9F001F9F001F90001F00000000B0A000B0A000AF9F00AF9F00A090000000000000
00A08F009F80009F8000907F00907F008F7000806F007F60006F60000000000000000000000000\
00000000000000000000000000000000000000000000000000BFD09F0020000000
7FAF6F00EFFFAF002FAF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFBF0020C0002F\
BF0020000000BF0020C0002FBF0020C0002F000000C0002FBF0020C0002F00FFF0
000000BF0020BF002FB0002000EFE000E0DF00E0DF00DFD000DFD000D0C000D0C09F001F00CFB0\
00BFAF90001F00B09F00AF9000A08F00A0807F00107F00108F8010000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000005F605F001000606000606F
00606F00607000606F006F70006070006F7F006F70006F7F006F7F6F00106F0010000000302F0F\
8F7F1F8070108F7F1F8F7F1F8F7F1F8F7F1090801F8F7F1F90801F8F7F1F90801F
90801F90801F90801F908F1F90801F9F8F1F90801F9F8F1F9F8F1F9F8F1F9F8F1F9F901F9F8F1F\
A0901F20200080001F80001F80001F7F001080001F80001F80001F80001080001F
80001F8F001F8000108F001F80001F8F001F80001F000000000000009FA0009FA0009FA0009FA0\
00A0AF009FA000A0A0009FA000A0AF009FA000A0A0009FA000A0A0009FA000A0A0
009FA090001F8F001F90001F8F001F90001F8F001F90001F90001F90001F8F001090001F000000\
90001F90001090001F90001F90001F90001090001F00000000A0A0009F9F00A09F
00A09F00A09F009F9F00A09F009F9F00A09F009F9000A09F009F9000A09F009F9000A090009F90\
00A090009F9000A090009F900000008F001090001F8F0010009F8F009080009080
009080009080008F7F8F0010800010800010008F70008F7F008070008F7000807000807000806F\
00806F007F607000106F00106F001000000000DFD000DFD000DFD000DFCF00DFD0
00DFCF00DFD000D0CF00DFCF00D0C000D0CF00CFC000D0C000CFBF00CFC000CFBF0000009F001F\
9F001F00C0B000C0B000BFAF00000000B0A000000000AF9F00000000A09080001F
009F8F00A08F009F80009080008F7F70001000806F6F0010007F60000000000000000000000000\
00000000000000000000000000000000000000000000000000B0CF00BFD09F0020
00DFF000E0FF00E0F0AF002FAF002000F0FF00F0F000FFFF00FFFF00FFFF00F0F000FFFFBF0020\
C0002FBF0020C0002FBF0020C0002FBF0020B0FF5FBF0020BF002FBF0020BF002F
000000BF002FB00020B0002000EFE000EFE000E0D000E0D000DFCF00DFCF00D0C000D0BF000000\
00C0AF0000008F001F00AF9000AF90009F807F00107000108F8F10000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000003F4F3000606000606F00606F006F70006070
006F70006070006F7F006F7F006F7F006F7F0070800070806F001F0000005F7F408F7F1F8F7F1F\
8F7F1F90801F8F7F1F90801F8F7F1F90801F90801F90801F90801F908F1F90801F
908F1F90801F9F8F20908F1F9F8F1F9F8F1F9F90209F8F1F9F90209F8F1FA090209F901FA09020\
A0901FA09020A0901FA09F207F903F2F2F0F80001F8F001F80001F000000000000
000000009FA0009FAF009FA0009FAF009FA000A0AF009FA000A0AF009FA000A0AF00A0AF00A0AF\
00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0A000A0AF00A0AF
00A0AF00A0A000AFAF00000090001F90001F90001F90001F90001F90001F00000000A0A000AFAF\
00A0A000000090001F9F001F90001F9F001F90001F9F001F90001F9F001F000000
00000000A09F00AFA000A09F00A0A000A09F00A0A000A09F00A09F00A09F00A09F00A09F00A09F\
00A09000A09F00A09000A09F009F9000A09000000090001F8F001090001F8F0010
000000009080009F8F0090808F001F8000108F001F800010000000008F7F008F7F008F70008F70\
00807000807000806F00806F000000700010006F6F000000B00020B00020B00020
B00020AF0020B0002000DFD000DFD0000000000000AF0020AF0020A0001FAF0020A0001FA00020\
00CFBF00CFBF00CFBF00C0B000C0AF00C0AF00BFA000BFA08F001F90001F8F001F
8F001F80001080001F00000000000070001070001070001000806F600010000000000000000000\
00000000000000000000000000000000000000000000000000000000000000C0DF
00000000E0F000EFFF000000B0002FB0002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000\
BF0020C0002FC000207FFF90BF0020C0002FB0FF5F90FF7FBF0020C0002F000000
C0002FBF0020BF002FB0002000000000EFE000EFDF00E0D000E0D000DFCF00DFCF00D0C0A00020\
00C0B000C0AF8F001F00000000000000AF8F7F00107F0010508F3F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000030402F00606000000000606F00606F006F70006070006F70
006070006F7F006F7F0000006000106F00106F00106F001F6F00106F001F6F00108F7F1F8F7F10\
8F7F1F8F7F1F90801F8F7F1F90801F8F801F90801F8F801F90801F90801F908F1F
90801F9F8F1F908F1F9F8F1F908F1F9F8F1F9F8F1F9F901F9F8F1FA0901F9F901FA090209F901F\
609050403F0F2F2F0F0F0F0000000080001F80001F80001F80001F009FA0009FAF
009FA0009FAF009FA000A0AF009FA000A0AF009FAF00A0AF009FAF00A0AF00A0AF00A0AF00A0AF\
00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00A0AF00AFAF
00A0AF00A0AF00A0AF00000090001F90001F90001F90001F00000000AFAF00A0AF00AFAF00A0A0\
00AFAF00A0A000AFAF90001F9F001F90001F9F001F00000000AFA000000000AFA0
00000090001F90001F00AFA000A09F00A0A000A09F00AFA000A09F00A09F00A09F00A09F00A09F\
00A09F00A09F00A09F00A09000A09F00A09000A090009F9000000090001090001F
8F001090001F8F00100000000090808F00108F00108F001F8000108F001000000000907F008F70\
008F7F008070008F700000007F0010000000007F6F00000000706F006F60000000
00E0D000E0DF00E0D000E0DF00DFD000E0D000DFCF00DFD000DFCF00DFCFA0001FAF0020A0001F\
A00020A0001F00000000C0B000C0B000C0B000C0B000BFA000BFA000B09F000000
8000108F001F00000000A08F0090807F001070001070001000806F00806F000000000000000000\
00000000000000000000000000000000000000000000000000000000000000C0DF
00CFE000E0FF00E0F000EFFFAF002000F0FF00F0F000FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
BF002FBF0020C0002F101000EFFF2F3F3F0F40400FEFF02F5F6010BF002090FF7F
000000BF002FBF0020BF002FB0002000EFE000E0DF00E0DF00DFD000DFCF00D0C000D0C000CFBF\
9F001F00C0AF90001F8F00108F001F80001000A080000000700010000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F4F2F00606000606F500010006070006070006F7F006F70006F7F000000
6F001F6000106F001F6000106F001F6F00106F001F6F001070001F302F0F8F7F1F8F7F1F90801F\
8F801F90801F8F801F908F1F90801F908F1F90801F908F20908F1F9F8F1F908F1F
9F90209F8F1F9F90209F8F1F9F90209F901FA090209F901FA09F20A0901FA09F20609050000000\
80001F80001F80001F8F001F80001F8F001F80001F8F001F8F001F8F001F80001F
8F002000000000A0B0009FAF00A0B000A0AF00A0B000A0AF00A0B000A0AF00AFB000A0AF00AFB0\
00A0B000AFB000A0AF00AFB000A0B000AFB000A0AF00AFB000AFAF00AFB000AFAF
00AFB000AFAF00AFB00000009F002000000000AFB000AFAF00AFB000AFAF00AFB000AFAF00AFB0\
00AFAF00AFB000AFAF00B0B00000009F001F90001F00000000AFAF00AFAF00AFA0
00AFAF00AFA000AFAF00AFA000AFAF00AFA00000000000009F001F90001F9F001F90001F000000\
00000000000000000000AFA000A09F00AF9F00A09000AF9F00A090000000900010
90001F90001F90001F8F001090001F8F00100000008F001090001F8F00108F001F8F0010000000\
00907F00907F00000080001F80001080001000806F00806F007F6F007F6F007060
00706F00E0DF00EFDF00E0DF00EFDF00E0D000E0DF00DFD000E0D000DFD000DFD0AF002000DFCF\
00DFCF00DFCF00D0C0A0002000CFBF00CFBF00C0B000C0B000BFAF00BFAF00B0A0
00B0A000000000AF9F00A08F00A08F00908000907F000000008F7000806F600010000000000000\
0000000000000000000000000000000000000000000000000000000000008F0020
00CFE09F002000E0FF00E0F000EFFF00EFFF00F0FF00F0FFBF002F00000000000000FFFF00FFFF\
00FFFFC0002FBF0020C0002FEFFF2FF0FF2FEFFF2FF0FF30EFFF2F000000CFFF40
EFFF2FBF0020BF002FBF002000000000EFE000EFE000E0DF00E0DF00DFD000DFCF00D0C000D0C0\
9F001F9F001F00000090001F8F001F000000009F80009F7F700010000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000304F2F\
006060005F6000000000606F0060700060700000005F0010600010600010600010
6000106F001F6000106F00106F00106F001F100F008F7F1F807F1F8F7F1F8F7F1F8F801F8F801F\
90801F8F801F90801F8F801F908F1F90801F908F1F908F1F9F8F1F908F1F9F8F1F
908F1F9F901F9F8F1F9F901F9F901FA090209F901FA0901F9F901FA09F20A0901FA09F20A09F1F\
00000080001F80001F80001F8F001F80001F8F001F80001F8F001F80001F8F001F
8F001F8F001F8F001F8F001F8F001F00000000A0AF00A0B000A0B000AFB000A0B000AFB000A0B0\
00AFB000A0B000AFB000AFB000AFB000AFB000AFB000AFB000AFB000AFB000AFB0
00AFB000AFB000AFAF00000000AFB000AFB000AFAF00AFB000AFAF00B0B000AFAF00AFB000AFAF\
00B0B000AFAF00B0B000AFAF00B0B00000009F001F00000000AFAF00AFAF00AFAF
00AFAF00AFAF00AFA000AFAF00AFA000AFAF00AFA000AFA000AFA000AFA000AFA0000000000000\
9F001F9000109F001F90001F9F001F90001000000090001000000000A09000A09F
00000090001F8F001090001F90001090001F8F00100000009F9F200000008F00108F00108F0010\
8F001F00907F0000008000108000107F001080001000806F00806F007F6F007F6F
00706000706F006F6000EFDF00E0DF00EFDF00E0D000E0DF00E0D000E0D000DFCF00000000DFCF\
00DFCF00D0CF00DFCF00D0C000D0C000CFBF00CFBF00C0B000C0B000BFAF6F8F4F
608F406F8F4F60804000AF9000A08F009F8F00907F00907F506F30000000007F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00BFD000CFE000DFF000E0FF00E0F0AF002F00EFF000F0FF00F0F0BF002FBF0020BF002FBF0020\
000000000000BF002FBF0020000000EFF02FEFFF2FEFFF2FEFFF2FEFF02FEFFF2F
EFFF2FBF002FBF0020BF0020B0002000F0EF00EFE000EFDF00E0D000E0D000DFCF00DFCF00D0C0\
00000090001F0000008F001F8F001F00A08F00A08F00907F0090706F000F000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000606000606F\
00606F5F001000606F5F0010006070006F7F006F7F007080006F80007080007080
00708F007080007F8F00708F0000006F001070001F20200F2F200F8F7F1F90801F8F801F908F1F\
90801F908F20908F1F908F20908F1F9F8F20908F1F9F9020908F1F9F90209F901F
9F90209F901FA090209F901FA09F209F901F2F2F0F0000001F1F0F80001F00000080001F8F001F\
0000000000000000008F001F0000000000008F001F8F00208F001F8F00208F001F
8F00208F001F9000208F001F9000208F001F9000208F001F90002000000000AFBF00AFBF00B0BF\
00AFBF00B0BF00AFBF00B0BF00AFBF00B0BF00AFBF00B0BF00AFB000B0BF00AFBF
0000000000009F002000000000B0BF00AFB000B0BF00B0B000B0BF00B0B000B0BF00B0B000B0BF\
00B0B000B0BF00B0B000B0B000B0B000B0B000B0B000000000AFAF00B0B000B0AF
00B0B000AFAF00B0B000AFAF00B0AF00AFAF00B0AF00AFAF00B0AF00AFA000B0AF00AFA000B0AF\
00AFA000B0AF00AFA000000090001F9F001F90001F9F001F90001F9F001F90001F
9F001F00000000000090001F90001F90001090001F90001FA0A02FB09F1FB0A01F9F9F2F90001F\
8F0010709F408090308F001F8000108F001F800010800010008F70008F7000806F
00806F007F60007F6F007060006F6F00EFDF00EFE000E0DF00EFDF00E0D000000000000000E0D0\
00DFD000E0D000DFCF00DFCF00D0C000DFCFA0001F00000000CFBF00CFBF000000
6F904F608F4F6F8F4F60804000AF9000A08F009F8F00907F50703F506F3000806F007060000000\
000000000000000000000000000000000000000000000000000000000000000000
00B0CF00C0DF00DFF000E0FF00EFFF00EFF000F0FF00F0FF00FFFF00FFFFBF002FBF0020BF002F\
BF0020C0002FEFFF2FB0FF603F400FF0FF30EFFF2FF0FF2FEFFF2FEFFF2FEFFF2F
EFFF2F90FF70C0002FBF0020BF002F00F0EF00F0EF00EFE000EFDF00E0D000E0D000DFCF00DFCF\
000000A0002090001F90001F8F001F00B09000A08F009F80008F70700010000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003F4F2F00606F50000F\
5F0010500010006F70006070006F7F006F7F007080006F8000708000708000708F
0070800000006F00106F001F6F00106F001F6F00106F803F8F7F1F8F801F8F801F908F1F8F801F\
908F1F908F1F908F1F908F1F908F1F908F1F9F9020908F1F9F901F9F901F9F9020
9F901F9F901F9F901FA09F209F901F80001F80001F80001F80001F80001F80001F8F001F80001F\
8F001F80001F00000000A0B000A0B000A0B000A0B000A0B000A0B000A0B000AFBF
00A0B000AFBF00A0B000AFBF00A0B000AFBF00AFB000AFBF00AFBF00AFBF00AFBF000000000000\
00000000000000B0C000AFBF00B0BF00B0BF00B0C000AFBF00B0BF00B0BF000000
90001F9F001F90001F00000000B0BF00B0BF00B0BF00B0BF00B0BF00B0BF00B0BF00B0BF00B0B0\
00BFBF00B0BF00BFBF00B0B000B0BF00B0B000B0BF00000000000000B0B000B0B0
00B0AF00B0B000B0AF00B0B000AFAF00B0AF00AFAF00B0AF00AFAF00B0AF00AFAF00B0AF00AFA0\
00B0AF00AFA000B0AF00AFA000AFA00000000000009000109F001F90001F9F001F
90001090001F900010302F0FB0A01FBFA01F9FA02F4F400F5F500FB0A01FB0A01FB0A01FB09F10\
B0A01FB09F1FB09F1FAF90108F00108000108F0010800010000000008F70008F70
00806F00806F00806F00806F007F6000706000EFDF00EFE000E0DF000000AF002000000000E0D0\
00E0D000DFD000E0D000DFCF00DFCF00D0C000D0C09F001FA0002000000000CFBF
6F904F6F904F608F40608F4000AF9000A090009F8F009F8000907F506F304F602F600010000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000BFD000CFE000DFF000E0F000EFFF00EFF000F0FF00F0F000FFFF00FFFFBF002FBF0020\
BF002FBF0020BF002FCFF040EFFF2FEFFF2FF0FF2FEFF02FEFFF2FEFFF2FEFFF2F
E0F02FEFFF2F000000BF002FBF002000F0F000F0E000F0E000E0DF00E0DF00DFD000DFD000D0C0\
0000009F001F9F001F0000008F001F00AF9000A08F009F7F0090706F0010000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F4F303F4F2F0000005000105F0010\
006070006F7F006F7F006F80006F8000708000708000708F00708F007F8F600010
6F001F6F00106F001F6F001070001F6F001F8F801F8F801F908F201010002F200F90801F908F20\
908F1F908F20908F1F9F9020908F1F9F902090901F9F90209F901F000000709040
A09F209F901FA09F209F9F1F609F5F80001F80001F80001F8F002080001F8F002080001F8F0020\
8F001F8F00208F001F00000000A0BF00AFBF00A0BF00AFBF00AFBF00AFBF00AFBF
00AFBF00AFBF00AFBF00AFBF00B0C000AFBF00B0C000AFBF00B0C000B0C000B0C000B0C000BFC0\
00B0C000B0C00000009F002090001F9F002090001F0000000000009F00209F001F
9F00209F001F9F002000000000BFC000B0C000BFC000B0BF00BFC000BFC000BFC000BFBF00BFC0\
00BFBF00BFC000BFBF00C0C000BFBF00BFBF00BFBF0000009F001F00000000B0B0
00BFBF00B0B000BFB000B0B000BFB000B0B000BFB000B0AF00BFB000B0AF00B0B000B0AF00B0B0\
00B0AF00B0AF00B0AF00B0AF00B0AF00B0AF00AFA000B0AF0000009F001F90001F
9F001F90001F9F001F90001F9F001F302F0FBFAF1FB0A01FBFAF1FB0A01FBFAF1FB0A01FBFAF1F\
B0A01FB0A01FB09F1FB0A01FB09F1F809F308F00108F001F80001000907F008F7F
00907F008F70008F7000806F00806F007F6F7000104F502F000000B00020B0002000000000EFE0\
00E0DF00E0DF00E0D000E0D000DFCF00DFCF00D0CF00DFCF000000A000209F001F
00CFBF6F904F6F904F608F40608F4F00AF9000AF90009F8F009F8000907F506F30600010007F60\
000000000000000000000000000000000000000000000000000000000000000000
00000000B0CF00C0DF00D0EF00E0FF00E0FF00EFFF00EFFF00F0FF00F0FF00FFFF00FFFF000000\
BF0020C0002FBF0020C0002FEFFF2FF0FF2FEFFF2FF0FF30EFFF2FEFFF2FEFFF2F
EFFF2FEFFF2FEFFF2FE0FF2F404010BF0020BF002FB0002000000000000000000000000000DFCF\
A0001F0000009F001F00C0AF8F001F00B09F00A08F00A08F00907F700010000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000005F603F4F2F3F4F2F00606F500010000000006070\
006F7F00607F006F80006F80007080006F8000708F6000106F001F6000106F001F
6F00106F001F6F00106F001F0000000000006F001070001F70001F10100F8F801F908F1F30300F\
00000070001F3F3F0F808F2F9F902090901F9F90207090402F2F0F7F001F80001F
7F001F80001F7F001F00000000000080001F80001F80001F80001F8F001F80001F8F001F80001F\
8F002080001F8F001F8F001F00000000A0BF00AFBF00AFBF00AFBF00AFBF00AFBF
00AFBF00B0C000AFBF00B0C000AFBF00B0C000AFC000B0C000B0C000B0C000B0C000B0CF00B0C0\
00BFCF00B0C000BFCF00B0C000BFCF0000009F002090001F9F002090001F9F0020
9F001F9F00209F001F00000000BFC000BFC000BFC000BFC000BFC000C0C000BFC000C0C000BFC0\
00C0C000BFC000C0C000BFC000C0C000BFBF0000009F001FA000209F001FA0001F
00BFBF00BFBF00BFB000BFBF00BFB000BFBF00B0B000BFB000B0B000BFB000B0AF00BFB000B0AF\
00BFB000B0AF00B0AF00B0AF00B0AF00B0A000B0AF00B0A000B0AF00AFA000B0A0
0000009F001F2F2F00302F0F2F2F00302F0F2F2F00BFAF1FB0A01FBFAF1FB0A01FBFAF1FB0A01F\
BFAF1FB0A01FB0A01FB0A01FB0A01FAF9F10B09F1FAF9F10AF9F1F50904F8F0010
000000000000008F70008F7000806F00806F007F60007F60000000000000B0002000000000EFDF\
00EFE000E0DF00E0DF00E0D000E0D000DFCF00DFCF000000A000209F001FA00020
9F001F00CFBF6F904F6F904F608F400000005F7F3F5F7F3F50703F009F804F60306F0010007060\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000B0CF00CFE000000000E0FF00E0F000000000EFF000F0FF00F0F000FFFF00FFFF\
000000BF0020BF002FBF002020200FEFF02FEFFF2FEFFF2FEFFF2FEFF02FEFFF2F
EFFF2FEFFF2FE0F02FEFFF2FE0FF2F3F400FB00020B00020B00020B00020AF001FAF0020000000\
00DFCF00CFBF9F001F00BFAF90001F00000000AF90009F80009F7F6F00106F0010
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003F4F3000606F006070500010006F7F00607F006F7F\
006F7F007080006F8000708F0070806F001F6000106F001F6000106F001F000000
000000007F900000006F001F70001F70001F70001F70001F10100F908F1F1F100F70001F7F001F\
70001F7F001F7F001F2F2F0F9F901F7F9F402F2F0F80001F7F001F80002080001F
80001F80001F80002000000000A0BF00A0B000A0BF00000000000080001F8F00208F001F8F0020\
8F001F9000208F001F8F002000000000B0C000AFC000B0C000AFC000B0C000B0C0
00B0C000B0C000B0CF00B0C000B0CF00B0C000BFCF00B0CF00BFCF00B0CF00BFCF00BFCF00BFD0\
00BFCF00BFD000BFCF0000009F00209F00209F00209F00209F001FA000209F0020
9F00209F001FA0002000000000C0CF00BFCF00C0CF00000070905F00000070905F709050709F5F\
7090507F9F5F00C0CF00CFCF00C0C000CFCF000000A00020A0001FA00020A0001F
A0002000000000C0C000C0BF00C0C000BFBF00C0BF00BFBF00C0BF00BFB000C0BF00BFB000BFBF\
00BFB000BFB000BFB0A0001F9F001FA000209F001F9F001F9F001F9F001F9F001F
9F001F9F001F9F001F00000090B040BFAF1FBFB020BFAF1FBFAF20BFAF1FBFAF20BFAF1FBFAF1F\
B0A01FBFAF1FB0A01FB0A01FB0A01FB0A01FB0A01FB0A01FAF9F1FB0A01FAF9F1F
8F001F8000108F001F8000108000107F00108000100000000000007000104F503000000000F0EF\
00EFE000EFE000EFDF000000000000AF002000DFD000DFD000DFCF00DFCFA00020
A0002000CFBF00CFBF00C0B000C0AF608F40608F405F80405F7F4050703F00000000806F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000909F2000C0DF00D0EF00E0F000EFFFAF002000EFFF00EFFF00FFFF00FFFF00FFFF\
00FFFF00FFFFBF0020C0002FBF0020404010EFFF2FF0FF2FEFFF2FF0FF30EFFF2F
EFFF2FEFFF2FEFFF2FEFFF2FEFFF2FBF0020BF002FB00020B00020B00020B00020AF002000E0D0\
00D0C000D0C000000000C0B000BFA08F001F00AF9000A08F00907F700010000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000030402F3F4F2F005F6F00607000000000607F00607F006F7F00607F\
006F80006F8000708F60001060001F60001000000000708F007F90007F9000809F
007F906F001F6F001070001F6F001070001F70001F00000000000070001F70001F7F001F70001F\
7F001F7F001F7F001F7F001F0000007F001F80001F7F001F80001F7F001F80001F
7F001F80001F00000000A0BF00A0B000A0BF00A0BF00AFBF00A0BF00AFBF00AFBF00AFC0000000\
8F00208F001F8F00208F001F90002000000000B0C000AFC000B0C000B0C000B0CF
00B0C000B0CF00B0C000BFCF00B0CF00BFCF00B0CF00BFD000BFCF00BFD000BFCF00BFD000BFCF\
00C0D00000009F00209F001F9F00209F001F9F00209F001F9F00209F001F000000
00000000C0D000C0CF00000070905070905F709050709F5F709050709F5F709050709F5F709050\
709F5F7090507F9F5F70905000CFCF00C0CF000000A0001FA00020A0001FA00020
A0001FA00020A0001FA000209F001FA00020000000000000CFBFBF000000000000000000000000\
A0001F9F001FA0001F9F001FA0001F9F001FA0001F9F001F9F001F9F001F9F001F
9F001F9F001F9F001F9F001F90001F9F001F4F4F0F7FB050BFAF1FBFAF1FBFAF1FBFAF1FB0AF1F\
BFAF1FB0AF1FBFAF1FB0A01FB0A01FB0A01FB0A01FAF9F10B0A01FAF9F10B0A01F
AF9F108F00108000108F00108000108000107F001000000000806F00806F007F60000000404F2F\
B00020AF0020B00020AF002000000000E0D000E0DF00DFD000DFD000DFCF00DFCF
9F001F00D0C000CFBF00CFBF00BFAF6F8F4F608F406080405F7F3F5F7F3F00907F008F7060000F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008F002000C0DF00DFF000E0F000EFFFAF002000EFFF00F0F000FFFF00F0F0\
00FFFF00FFFF00FFFFBF0020BF002F5F5F10EFFF2FEFF02FEFFF2FEFFF2FEFFF2F
EFF02FEFFF2FEFFF2FEFFF2FE0F02F3F400FBF0020BF002FB00020B00020AF002000000000DFCF\
00DFCF00D0C000D0BF00000000BFA08F001F00AF9000A080009F80700010000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000005F60000000006070006070006070006F7F00607F006F80006F7F007080\
006F80000000000000007F90007F8F007F90007F9000809F007F9000809F6F001F
70001F6F001F70001F70001F70001F000000008FA00000007F001F70001F7F001F7F001F7F0020\
7F001F7F001F000000009FB0009FB00000007F001F80002080001F80002080001F
80002080001F00000000A0BF00AFC000A0BF00AFC000AFBF00AFC000AFC000B0C000AFC000B0CF\
00B0C000B0CF00B0C00000008F002090002000000000BFCF00B0CF00BFCF00B0CF
00BFD000BFCF00BFD000BFCF00BFD000BFD000C0D000BFD000C0D000BFD000C0DF00C0D0000000\
CFC0DF0000009F002000000000000000C0DF00C0D000CFDF00C0D000CFDF00C0D0
00CFDF00C0D000CFDF0000007F9F60709F5F7F9F5F709F5F7F9F5F709F5F7F9F5F709F5F7FA05F\
709F5F7FA05F709F5F7FA05F709F5F7F9F5F000000000000A00020AF0020A0001F
AF0020A00020A00020A0001FA0002000000000CFCF00C0C000CFC000C0C000CFC000C0BF00CFC0\
00C0BF00CFBF00C0BF00C0BF00C0BF000000000000A000209F001FA0001F9F001F
A000209F001F9F001F9F001F9F001F9F001F9F001F9F001F30300FBFB01FBFB020BFAF1FBFB020\
BFAF1FBFB01FB0AF1FBFAF20B0AF1FBFAF1FB0A01FB0AF1FB0A01FB0A01FAFA01F
B0A01FAF9F1F403F0F1010008F001F8000108F0010008F70008F7F00807000806F007F6F700010\
4F502FB0002FB0002000000000EFE000EFE000E0DF00EFDF00E0D000E0D000DFCF
00000000D0C000D0C000CFBF00CFBF00C0AF6F904F608F406080405F7F3F50703F008F70000000\
00705F000000000000000000000000000000000000000000000000000000000000
00000000000000000080001F00000000CFE0A0002F00E0FFAF002F00EFFF00F0FF00F0FF00FFFF\
00FFFF00FFFF00FFFF000000EFFF2FEFFF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2F
F0FF30EFFF2FEFFF2FEFFF2FEFFF2FEFFF2FBF002FBF0020BF002FB0002000000000E0D000E0D0\
00DFCF00DFC000CFBF00000000BFAF00BFA000AF9000AF90009F8070001000806F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000005F6000606F005F6F0060700060705F001000607F006F80006F7F006F80006F80\
00000000708F00708F00708F007F90007F90007F90007F9000809F6F00106F001F
6F001F70001F6F0010000000008FA0008FA0008FA070001F70001F7F001F70001F7F001F7F001F\
0000000090AF009FB0009FB0009FB0009FB000000080001F80002080001F80001F
80001F80002000A0BF00AFBF00A0BF00AFC000AFBF00AFC000AFC000B0C000AFC000B0CF00B0C0\
00B0CF00B0C000B0CF00B0CF00B0CF00B0CF00BFCF00000000BFD000B0CF00BFD0
00BFCF00BFD000BFD000BFD000BFD000C0DF00BFD000C0DF00C0D000C0DF00000000000000C0DF\
00CFDF00C0DF00C0DF00C0DF00CFDF00C0DF00CFDF00C0DF00CFDF00C0DF00CFDF
00CFDF00CFDF00CFDF000000709F5F709F60709F5F709F5F709F5F7FA05F709F5F7FA05F709F5F\
7FA05F709F5F7FA05F709F5F7FA05F709F5F00000000D0D000D0D0000000000000
A0001FAF0020A0001FA00020A0001F00000000CFC000CFCF00CFC000CFCF00CFC000CFC000CFC0\
00CFC000C0BF00CFC000C0BF00CFBF00C0BF00C0BF00C0B000C0BF00BFB000C0B0
00000000BFB00000009F001F9F001F9F001F90001F1F1F00BFB01FBFB020BFB01FBFB01FBFB01F\
BFB01FB0AF1FBFAF1FB0AF1FBFAF1FB0A01FB0AF1FB0A01FB0AF1FAFA010B0A01F
AFA01FB0A01FAF9F10AF9F1FAF9F10809F3080001000907F008F70008F7F00806F008070007F6F\
70001000706000000000EFE000F0E000EFDF00EFE000E0DF00E0DF00DFD000DFD0
00DFCF00DFCF00CFC000CFBF00CFB000CFB0608F4F608F4F5F80405F804050703000907F00806F\
5F0010000000000000000000000000000000000000000000000000000000000000
000000000000000000909F1F8F001F00000000D0EF000000A00020AF002F00EFF000F0FF00F0F0\
00FFFF00F0F000FFFF000000C0002FB0F05FEFFF2FEFFF2FF0FF2FEFF02FEFFF2F
EFFF2FEFFF2FEFF02FEFFF2FE0F02FEFFF2FB00020BF0020B00020B0002000E0DF00E0DF00DFD0\
00DFCF00D0C000D0C090001F00C0AF00B09F00000000A08F00A080000000008F6F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
006060005F6000606F00607000607F0060705F001F5F0010006F80006F8000000000000000708F\
00708F007F90007F90007F90007F9000809F00809F6F001F6F001F70001F6F001F
000000008FA0008FA0008FA00090AF008FA000000070001F7F001F7F001F7F00207F001F009FB0\
009FB000A0BF009FB000A0BF009FB000A0BF00A0BF00000080001F8F002080001F
8F002000000000AFC000AFC000AFC000AFC000B0CF00AFC000B0CF00B0CF00B0CF00B0CF00BFCF\
00B0CF00BFD000B0CF00BFD000BFCF00BFD000BFD000BFD00000009F0020000000
00C0DF00BFD000C0DF00C0DF00C0DF0000000000009F00209F00209F002000CFE000CFE000CFE0\
00CFE000CFE000CFE000CFE000CFE000D0E000CFE000D0E000CFDF00D0E000D0E0
00D0E000D0DF00D0E000D0E000D0E0709F5F7FA0607FA0607FA06070A05F7FA0607FA05F7FA060\
7FA05F7FAF607FA05F7FA0607FA05F7FA06000000000DFDF00D0D000DFDF00D0D0
00DFD000D0D0000000A00020AF0020A0002000000000D0CF00D0CF00D0CF00D0CF00D0CF00D0CF\
00CFC000D0CF00CFC000D0C000CFC000CFC000CFBF00CFC000C0BF00CFBF00C0BF
00C0BF00C0B000C0BF00BFB000C0B0CFBFAF0000000000004F4F10A0B02FC0BF20BFB01FC0B020\
BFB01FBFB020BFB01FBFB020BFAF1FBFB020B0AF1FBFAF1FB0AF1FBFAF1FB0A01F
B0AF1FB0A01FB0A01FAFA01FB0A01F909F208F001F8F001000000000907F00907F008F70008F70\
00806F000000007F6000000000000000F0EF00EFE000EFE000EFDF00EFDF00E0D0
00E0D0A0001F00DFCF00D0C000D0C000CFB06F9F506F904F608F4F5F7F405F7F4000907F008F70\
007060000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008F0020000000000000A0002000EFFFAF002000000000F0FF00F0FF\
00F0FF000000BF0020C0002FBF0020C0002FB0FF60F0FF2FEFFF2FF0FF30EFFF2F
F0FF2FEFFF2FEFFF30EFFF2FEFFF2FE0FF2FEFFF2FBF0020BF002F00000000EFE000E0DF00E0D0\
00DFCF00DFCF00D0C000000000C0AF00BFA000B09F00AF90009F80009F7F008F70
6F0010000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000006060
4F000F4F0010005F6F0060700060700000005F00100000005F0010000000006F8000708F00708F\
007F90007090007F90007F9000809F007F9F6F001F6F00106F001F00809F008FA0
0080A0008FA0008FA00090AF008FA00090AF0000007F001F70001F7F001F000000009FB0009FB0\
009FBF009FB000A0BF009FB000A0BF00A0BF00A0BF00A0BF00000080001F8F0020
80001F00AFC000AFC000B0CF00AFC000B0CF00AFC000B0CF00B0CF00B0CF00B0CF00BFD000B0CF\
00BFD000BFD000BFD000BFD000BFD000BFD000C0DF00BFD00000009000209F0020
9000209F00209F00209F00209F00209F00209F00209F002000000000CFE000CFE000CFE000CFE0\
00CFE000CFE000D0E000CFE000D0E000CFE000D0E000CFE000D0E000D0E000D0E0
00D0E000D0E000D0E000DFE000D0E000DFE000D0E000DFE000000000000070A0607FAF6070A05F\
7FAF6000000000000000000000000000000000DFE000DFDF00DFDF00DFDF00DFDF
00D0D000DFDF00D0D000DFDF00D0D0AF0020A0002000000000D0CF00D0D000D0CF00D0CF00D0CF\
00D0CF00D0CF00D0CF00CFC000D0C000CFC000D0C000CFBF00CFC000CFBF00CFBF
00C0BF00C0BF00C0B000C0BF00BFB00000009F001F9F001F9F001F9F001F9F001F9F001F90001F\
000000BFB01FBFB020B0AF1FBFB01FB0AF1FBFB01FB0AF1FB0AF1FB0AF1FB0AF1F
B0A01FB0A01FAFA01FB0A01FAF9F10AFA01F8F00108F001F8000108F001000907F00907F008F70\
008F7000806F00806F007F60B00020AF0020B00020AF002000EFE000E0DF00E0DF
00DFD0A0002000D0CF00D0C000CFBF00CFBF00C0B06F904F608F405F8040507F3F009F80008F70\
007F6F000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000BFD000CFE0000000A00020AF002FAF0020B0002F00EFF0\
000000000000BF002FBF0020BF002FBF0020C0002F000000EFFF2FEFFF2FF0FF2F
EFF02FEFFF2FEFF02FEFFF2FE0F02FEFFF2FE0F02FEFFF2FB0002000F0EF00EFE000EFDF00DFD0\
00DFD000D0C000D0C000CFB000C0B000BFA000B09F00A08F00A08F00907F009070
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000060604F0010
50001000607000607F00607F0000005F00105F001F5F001000708F00708F00708F00708F007F90\
007F90007F9F007F9000809F00809F6F001F000000008FA0008FA0008FA0008FA0
0090AF008FAF0090AF0090AF0090B00090AF7F001F7F001F7F0020000000009FBF009FB000A0BF\
00A0BF00A0BF00A0BF00AFC000A0BF00AFC000A0BF00AFC000AFC000000080001F
00000000AFC000B0CF00B0CF00B0CF00B0CF00B0D000B0CF00BFD000BFD000BFD000BFD000C0DF\
00BFD000C0DF00BFD000C0DF00C0DF00C0DF00C0DF00CFDF00C0DF0000009F0020
9F00209F00209F00209F0020A0002F9F0020A0002000CFE000D0EF00CFE000D0EF00CFE000D0EF\
00D0E000D0EF00D0E000DFEF00D0E000DFEF00D0E000DFEF00D0E000DFEF00DFE0
00DFEF00DFE000DFEF00DFE000E0EF00DFE000E0EF00DFE00000000000007FAF607FAF607FAF60\
00E0E000E0EF00E0E000E0EF00E0E000E0E000E0E000E0E000E0E000E0E000DFDF
00E0E000DFDF00E0DF00DFDF00E0DF00DFDF00E0DF000000E0E0DF00000000DFD000DFD000DFD0\
00DFD000DFD000D0CF00DFD000D0CF00D0CF00D0C000D0CF00D0C000D0C000CFC0
00D0C000CFBF00CFC000CFBFA00020A0001FA000209F001FA000209F001FA0001F9F001FA00020\
9F001F1F1F0FBFB01FC0BF20BFB01FBFB020BFB01FBFB020B0AF1FBFB01FB0AF1F
BFAF1FB0AF1FB0AF1FB0A01FB0AF1FAFA01FB0A01F0000008F001F8F00108F001F80001000907F\
008F70008F7000806F00806F007F6F00F0EF00EFE000F0EF00EFE000EFE0000000
000000A00020A0002000D0CF00000000000000CFBF6F904F6F904F6080405F8040009F8000907F\
007F6F000000000000000000000000000000000000000000000000000000000000
0000000000000000000000009FA02000BFD000CFE000D0EFA0002F000000AF002FAF0020000000\
00F0FFBF002FBF0020C0002FBF0020C0002FBF0020404010EFFF2FF0FF2FEFFF2F
F0FF30EFFF2FEFFF2FEFFF2FEFFF2FEFFF2FEFFF2FAFFF5FBF002F00F0EF00F0E000EFDF00EFDF\
00DFD000DFCF00D0C000D0C000C0B000C0AF00B09F00000000A08F009F80009070
008F6F000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000005F6000606F
000000000000006070006F7F5000105F001F5F0010000000006F8F00708F00708F007F90007F90\
007F9F007F9000809F00809F00000000809F008FA00080A0008FA0008FA0008FAF
008FA00090AF008FAF0090AF0090AF00000070001F7F001F009FB0009FBF009FB000A0BF00A0BF\
00A0C000A0BF00A0C000A0BF00AFC000A0C000AFC000AFC000AFCF00AFC0000000
00AFCF00B0CF00B0CF00B0CF00B0CF00BFD000B0CF00BFD000BFD000BFD000BFD000C0DF00BFD0\
00C0DF00C0D000C0DF00C0DF0000000000009F00209F00209F00209F00209F0020
9F0020A000209F0020A000209F0020A0002000CFE000D0EF00D0E000D0EF00D0E000D0EF00D0E0\
00DFEF00D0E000DFEF00D0E000DFEF00D0E000DFEF00DFE000DFEF00DFE000DFEF
00DFE000E0EF00DFE00000007FA0607FAF607FAF607FAF607FAF607FAF607FAF607FAF607FAF60\
7FAF607FAF607FAF6000E0E000E0EF00E0E000E0E000E0E000E0E000E0E000E0E0
00E0DF00E0E000E0DF00E0E000E0DF00E0DF00DFDF00E0DF00DFDF00E0DF00000000000000DFD0\
00E0D000DFD000DFD000DFCF00DFD000D0CF00DFCF00D0CF00D0CF00D0C000D0C0
00D0C000D0C000CFBFA00020A0001FA000209F001FA0001F9F001FA0001F9F001FA0001F9F001F\
9F001F302F00C0BF20BFB01FBFBF20BFB01FBFB01FBFB01FBFB01FB0AF1FBFB01F
B0AF1FB0AF1FB0AF1FB0AF1FAFA01FB0AF1FAFA010AFA01F1010008F001F800010800010800010\
800010008F70008F7000806F00806F007F6000F0EF00EFE000EFE000EFDF00EFE0
A00020AF0020A0001F00000070A05070A0506F904F000000608F4F608F4F5F7F3F00A08F00907F\
00806F000000000000000000000000000000000000000000000000000000000000
00000000000000000000000080001000B0D000BFDF00D0EF9F002000E0FFA0002000000000EFFF\
00F0FF00F0F0000000BF0020BF002FBF0020BF002FB0FF5FEFFF2FEFF02FEFFF2F
EFF02FEFFF2FEFF02FEFFF2FE0F02FEFFF2FC0F040BF0020B00020BF002000EFE000EFDF00E0D0\
00E0D000DFCF00D0C000CFBF00CFB000BFAF00BFA000000000A08F009F7F00907F
008060000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000005F6000606F00606F
5000105000105F001F0000005F001F5F001060001F00708F00709000708F007F90007F90007F9F\
007F9F00809F0000000080A00080A0008FAF008FA0008FAF008FA00090AF0090AF
0090AF0090AF009FB00090B0009FB07F001F7F0020009FB000A0BF009FBF00A0C000A0BF00A0C0\
00A0C000AFC000AFC000AFC000AFC000B0CF00AFCF00B0CF00B0CF00B0D08F0020
00000000B0CF00BFD000BFD000BFD000BFD000C0DF00BFD000C0DF00C0DF00C0DF00C0DF00CFE0\
00C0DF00000000000000000000000000000000000000000000CFE000D0EF00CFE0
00D0EF00D0E000D0EF00D0EF000000709F607FA06F70A0607FA06F70A0607FA06F7FA0607FAF6F\
70A06000000000DFEF00DFEF00DFEF00E0F000DFEF00E0EF00DFEF00E0F0000000
7FAF6F7FAF6080AF6F7FAF607FB06F7FAF6080B06F7FAF6080B06F7FAF6080B06F7FAF6080B06F\
7FAF6080B06F7FAF6080B0607FAF6080B06F00E0E000EFEF00E0E000EFEF00E0E0
00EFE000E0E000EFE000E0E000EFE000E0DF00EFE000E0DF00EFE000E0DF00EFE0AF0020B00020\
00000000E0DF00E0D000E0DF00DFD000E0D000DFD000DFD000DFCF00DFD000D0CF
00DFCF00D0C0AF0020A0001FAF0020A0001FAF0020A0001FA00020A0001FA00020A0001FA00020\
9F001FA00020C0BF2090C04090BF4090BF405F5F10BFBF20BFB01FBFB020BFB01F
BFB01FB0AF1FBFB020B0AF1FB0AF1FB0AF1FB0AF1F8F00108F001F8F0010000000000000000000\
80001080001F7F0010009070008F6F7F0010007F6F00F0EF00EFE000F0EF00EFE0
00EFE0A00020AF002000DFD070A05F6F9F50709F5000C0B000C0B000BFA000B0A05F7F3F000000\
00806F000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000080001F00C0DF00CFE000DFF000E0FF00EFFFAF002000F0FF\
00F0FF00FFFF00FFFF00FFFF000000C0002FBF0020F0FF2FEFFF2FF0FF30EFFF2F
F0FF2FEFFF2FEFFF30EFFF2FEFFF2FE0FF2F000000BF0020BF002FB0002000000000EFE000EFDF\
00E0D000E0D000D0C000D0C000000000C0B000BFA000B09F00000000A08F00907F
008F6F000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000005F6000606F005F6F006070
500010500010500010006F80006F800000005F001000709000708F007F90007F90007F9F007F90\
00809F0000000080A000809F008FA0008FA0008FAF008FA00090AF008FAF0090B0
0090AF0090B00090B0009FB0000000009FBF009FB000A0BF009FBF00A0BF00A0BF00AFC000A0C0\
00AFC000AFC000AFCF00AFC000B0CF00AFCF00B0CF00B0CF00B0D0000000900020
8F001F00000000000000C0DF00BFD000C0DF0000000000009000209F002090002000000000C0DF\
00CFE000CFE000CFE000CFE000CFE000CFE000D0EF00CFE000D0EF00D0EF00D0EF
00D0EF00D0EF0000007FA06F70A06070A06070A0607FA06F70A0607FA06070A0607FAF6F70A060\
7FAF607FA0607FAF6F00000000E0EF00E0EF00E0F00000000000007FAF607FAF6F
7FAF607FAF607FAF6080B06F7FAF607FB0607FAF6080B06F7FAF6080B0607FAF6080B06F7FAF60\
80B0607FAF6080B0607FAF6080B0607FAF6080B0607FAF6000EFEF00E0E000EFEF
00E0E000EFE000E0E000EFE000E0E000EFE000E0E000EFE000E0DF00EFE000E0DF00EFE0000000\
B00020AF0020B0002000E0D000E0DF00E0D000E0DF00DFD000DFD000DFCF00DFD0
000000AF0020A0001FAF0020A0001FA0002000000000000000CFBF00CFC000CFBF00CFC000C0BF\
00CFBF00C0B0A0001F9F001F9F001F9F001F9F001F2F2F00BFBF1FBFB01FBFB01F
B0B01FBFB01FB0AF1FB0B01FB0AF1FB0AF1FB0AF1FB0AF1F8F00108F001F8F001000A08F009F80\
009F8000907F009F7F008F70008F70008F6F70001000706000EFEF00EFE000EFE0
00E0DFAF002000E0DF00E0DF70A05070A0506F9F5000CFBF00C0AF00BFAF00B09F5F80404F6F30\
00806F000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007F001F00BFD000CFE000D0EF00E0FF00E0FFAF002F00EFF0\
00F0FF00F0FF00FFFF00F0F000FFFF00FFFF00FFFF0000003F400FEFFF2FF0FF2F
EFF02FEFFF2FEFF02FEFFF2FE0F02FEFFF2FE0F02FBF002FB00020BF0020B0002000000000E0DF\
00E0D000DFCF00DFCF00D0BF00000000C0AF00BFA000AF9F80001F009F80009F7F
008060000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000606F00606F006070006070
5F001F500010006F80006F8000708F00708F007090600010000000007F90007F9F007F9F000000\
6F001F0080A00080A0008FAF008FA00090AF008FAF0090B00090AF0090B00090AF
009FB0009FB0009FBF009FB0000000009FBF00A0BF00A0BF00AFC000A0C000AFC000AFC000B0CF\
00AFCF00B0CF00B0CF00B0D000B0CF00BFD000B0D000BFD00000009000208F0020
9000209000209000209000209F00209000209F002090002000000000CFE000CFE000CFE000D0EF\
00D0E000D0EF00D0E000D0EF00D0EF00D0EF00D0EF00DFF000D0EF00DFF000D0EF
00DFF00000007FAF6F70A0607FAF6F7FA0607FAF6F7FA0607FAF6F7FAF607FAF6F7FAF607FAF6F\
7FAF607FAF6F7FAF6080B06F7FAF600000007FAF6080B06F7FAF6080B06F7FB060
80B06F7FB06080B06F7FB06080B06F80B06080B06F7FB06080B06F80B06080B06F80B06080B06F\
80B06080B06F80B06080B06F80B06080B06F80B06080B06F80B06000000000EFEF
00F0EF00EFE000EFEF00EFE000EFEF00EFE000EFE000EFE000EFE000EFE000EFE000EFE000EFE0\
000000B00020B00020B0002000000000000000E0DF00EFDF00E0D000E0DF000000
00E0D000DFD000DFD000DFCF00DFD000DFCF00DFCF00D0CF00D0CF00D0C000D0C000CFC000D0C0\
00CFBF00CFBF000000A000209F001FA000209F001F9F002030300FC0BF208FBF3F
30300F90001F70B050B0B01FBFB02070AF4F90001FA0AF208FAF3F8F00108F001F8F0010000000\
009F80009F80009F7F009F7F00907F00907F7F00107000104F5F2F00F0EF00EFE0
00000000EFE000EFE000E0D07FAF5F70A05070A05000CFBF00CFBF00BFAF00BFAF5F7F4050703F\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000070001F00000000C0DF00D0EFA0002000EFFF00EFFF000000\
00EFFF00F0FF00F0FF00FFFF00FFFF00FFFF00FFFFC0002FBF0020000000EFFF2F
F0FF30EFFF2FEFFF2FEFFF2FEFFF2FEFFF2FEFFF2F1F1F00BF002FBF0020B0002FAF002000EFE0\
00E0D000E0D000DFCF00DFC000000000CFB000BFA000BFA080001F00000000907F
008F6F000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000606F005F6F006070005F70500010
500010000000006F8000708F00708F00709000708F00000060001060001F60001060001F60001F\
0080A00080A0008FA0008FA0008FAF008FAF0090AF0090AF0090B00090AF009FB0
0090B0009FBF009FB0000000009FBF00A0C000A0BF00A0C000A0C000AFCF00AFC000B0CF00AFCF\
00B0D000B0CF00B0D00000000000008F001F9000208F00209000208F0020900020
9000209000209000209F002090002000000000000000CFE000CFE000D0EF00CFE000D0EF00D0E0\
00D0EF00D0EF00D0EF00D0EF00DFF000D0EF00DFF000DFEF00DFF000DFEF00DFF0
00DFF000000070A0607FAF6F70A0607FAF6F70AF607FAF6F7FAF607FAF6F7FAF607FAF6F7FAF60\
7FB06F7FAF607FB06F00000000EFF000EFF000EFF000000080B06F7FB06080B06F
7FB06080B06F7FB06080B06F7FB06080B06F7FB06080B06F7FB06080B06F7FB06080B06F80B060\
80B06F7FB06080B06F80B06080B06F7FB06080B06080B06080B0607FB06080B060
00EFEF00F0EF00EFE000EFEF00EFE000F0EF00000000000000000000EFE000EFE000EFE000EFE0\
00EFE0000000B00020B00020B00020AF0020B00020AF0020B0002000000000E0DF
00E0D000E0DF00DFD000E0D000DFD000DFD000D0CF00DFCF00D0CF00DFCF00D0C000D0CF00D0C0\
00D0C000CFBF00CFBFA0001FA000209F001F9F001F9F001F9F001F000000000000
90001F9F001F90001090001F70B04FBFB01F8F001090001F8F00100000008F00108F001F800010\
000000009F80009F80009F7F009F7F0090700090707000100000004F502FAF0020
00EFE000EFE000E0DF00E0DF00DFD000000000CFC000CFBF00C0B000C0B000B0A05F8040000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000080001F00C0DF00CFE000DFF000E0F000EFFFAF0020\
00F0FF00EFF000F0FF00F0FF00FFFF00F0F000FFFFBF0020C0002FBF0020EFFF2F
EFFF2FF0FF2FEFF02FEFFF2FE0F02FEFFF2FE0F02FE0FF2FB00020BF002FB00020B00020AF0020\
00E0DF00DFCF00DFCF00D0C000D0BF00C0B000C0AF00B09F00B0907F0010009F7F
008060000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000606F00606F00607000607000000000607F
5F001F006F8000708F00708F007F90007090007F9F00000060001F60001F6F001F60001F6F001F\
6F001F0000000000000090AF008FAF0090B00090B0009FB00090B0009FBF009FB0
009FBF7F001F7F002000000000A0C000A0C000AFCF00AFC000B0CF00000000B0D000B0CF000000\
8F00200000008F00209000208F00209000209000209000209000209F0020900020
9F00209000209F002F00000000D0EF00CFE000D0EF00D0EF00D0EF00D0EF00DFF000D0EF00DFF0\
00D0EF00DFF000DFF000DFF000DFF000E0F000DFF000E0F000DFF000E0FF00E0F0
00E0FF00E0F000E0FF0000007FB06F7FAF6F80B06F7FAF6F7FB06F7FAF6F80B06F7FB06F80B06F\
7FB06F00000000EFF000F0FF00EFF000F0FF00EFF000F0FF00EFF000000080B06F
80BF6F80B06F80BF6F80B06F80BF6F80B06F80BF6F80B06F80BF6F80B06F8FBF6F80B06F80BF6F\
80B06F8FBF6F80B06F80BF6F80B0608FBF6F80B06080BF6F80B0608FBF6F80B060
80BF6F00000000F0EF00F0EF00000000000080BF6F80B06080B06080B06080BF6F80B06080B060\
80B060000000000000B00020B00020B0002FB00020B0002000000000EFE000E0DF
00EFDF00E0DF00EFDF00E0DF00E0DF00DFD000E0DF00DFD000DFD000DFCF00DFD000DFCF00DFCF\
00D0C000D0CF00D0C0A00020A0001FA00020A0001FA000209F001F00000000C0AF
9F001F9F001F9F001F90001F9F001F90001FC0BFA090001F90001F8F00100000000000008F001F\
8000108F001F009F80009F80009F7F009F8000907F0090707000105060304F5F2F
00EFEF00EFE000EFE000E0DF00E0DF9F00209F002000CFC000CFBF00C0B000BFAF5F8040009F80\
00806F000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000008000208F002000CFE000D0EF00EFFF00EFFF00EFFF\
00EFFF00F0FF00F0FF00FFFF00FFFF00FFFF000000C0002FBF0020C0002FEFFF2F
F0FF2FEFFF2FF0FF30EFFF2FEFFF2FE0FF2F90FF7F8FFF70000000B00020BF002FB00020B00020\
00000000E0D000DFCF00DFCF00D0BF9F002000C0AF00BFA000AF907F001F00907F
00806F000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000005F6F006070005F7000000000607F000000
006F8000708F00708F007090007090007F90007F9000000060001060001F60001F6F001F60001F\
6F001F6F001F70001F00000000000000000070001F70001F70001F70001F7F0020
7F001F7F00207F001F8000207F001F80002080001F80002000000000B0D000B0CF00BFD000B0D0\
00BFDF00BFD000C0DF00BFDF00C0DF00C0DF00C0E000C0DF00CFE000C0E000CFE0
0000009F00209F00209F002000CFE000000000D0EF00000000D0EF00DFF000D0EF00DFF000DFEF\
00DFF000DFF000E0F000DFF000E0F000DFF000E0FF00E0F0000000000000000000
00000000000000000000000000000000000000EFF000EFFF00EFF000EFFF00EFF000F0FF00EFF0\
00F0FF00EFF000F0FF00EFF000F0FF00EFF000F0FF00EFF000F0FF00F0F000F0FF
7FB06080BF6F80B06F80BF6F80B06080BF6F80B06F80BF6F80B06080BF6F80B06080BF6F80B060\
80BF6F80B06080BF6F80B06080BF6F80B06080BF6F80B06080BF6F80B06080BF6F
80B06080BF6000000080BF6F80B06080BF6080B06080BF6080B06080B06080B06080B06080B060\
80B06080B06080B06000000000EFE0000000B00020B00020B00020B00020000000
00E0DF00E0DF00E0DF00E0DF00E0D000E0DF00E0D000E0DF00DFD000E0D000DFD000DFD000D0CF\
00DFCF00D0CF00D0CF000000A00020A0001FA000209F001F00000000C0B000C0B0
00C0AF9F001F9F001F9F001F90001F90001F90001F0000008F001090001F8F001000AF9000A08F\
0000008000108F001F000000009F80009F7F009F7F009070008F7000806F506030
4F502F00EFE000E0DF00E0DFA0001FA000209F001F9F002000000000C0B000B0A05F804000907F\
00806F000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000070001F00000000C0DF00D0EF00DFF000EFFF00E0F0\
00000000EFFF00F0FF00F0F000FFFF00FFFFC0002FBF0020BF002FBF0020D0FF4F
CFF0405F6010EFF02FEFFF2FE0F02FEFFF2FBF0020BF002F000000000000000000B00020AF0020\
AF0020A0002000DFD000D0C000D0C090001F00C0B000B0A000B09F7F0010009F7F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000606F00607000607000607F00607F006F80000000
00708F00708F007090007090007F9F007F9000809F007F9F6F001F60001F0000000080A0008FAF\
008FAF0090AF008FAF0090B00090B0009FBF0000000000007F001F7F00207F001F
8000207F001F80002080001F00000000B0CF00B0D000B0CF00BFD000BFD000BFDF00BFD000C0DF\
00C0DF00C0E000C0DF00CFE000C0E000CFE000CFE000CFEF0000009F002F9F0020
9F002F9F0020A0002F9F0020A0002F9F0020A0002FA00020A0002F00000000000000DFF000E0FF\
00E0F000E0FF00E0FF000000000000AF002FAF0020AF002FAF0020AF002FAF0020
B0002FAF0020B0002FAF0020B0002FB00020B0002F00000000F0FF00F0FF00F0FF00F0FF00F0FF\
00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00FFFF00F0FF00FFFF00F0FF
00FFFF00000080BF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F\
80BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F000000000000
00FFF000F0F000FFF000F0F000000080BF608FBF6F80B0608FBF6F80B0608FBF6F80B0608FBF6F\
80B06080BF6F80B0608FBF6F00000000F0EF00EFE000F0EF00EFE0000000B00020
B0002FB0002000EFE000E0DF00EFE000E0DF00E0DF00E0DF00E0DF00E0DF00E0DF00DFD000E0DF\
00DFD000DFD000DFCF000000A00020AF0020A0001F00000000CFBF00CFBF00CFBF
00CFBF00C0B09F00209F001F9F00209F001F9F001F00BFA000000090001F90001F8F001000AF90\
00AF9000AF900000008F001F80001000A080009F7F009F7F00907000000000806F
50603000000000000000E0DF00E0DF00DFD0A0002000D0C000CFC08F001F8F001F00AF90009F80\
00806F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000009FAF00B0D000000000CFE0A0002F00E0FF00EFFF\
00EFFF00F0FF00F0FF00FFFF00FFFF00FFFFBF0020C0002FBF0020C0002F000000
5F6010C00020F0FF2FEFFF2FB0FF60EFFF2FBF002FBF0020BF002F00FFF000FFF000F0EF00F0EF\
00000000000000000000DFCF00D0C09F002000C0AF00000000AF907F001F00907F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000606F00000000607F00607F006F80006080000000
006F8F00709000708F007F900000000000000000004F604000809F0080A00080A0008FAF008FA0\
0090AF008FAF0090B00090B0009FBF009FB0009FBF009FBF00A0C0000000800020
7F001F00000000AFCF00B0CF00B0CF00B0D000B0D000BFD000BFD000BFDF00BFD000C0DF00C0DF\
00C0E000C0DF00CFE000CFE000CFEF00CFE09F00209F00209F00209F00209F0020
9F0020A0002F9F002000000000DFF000DFF000DFF000DFF000DFF000E0FF00DFF000E0FF00E0F0\
00E0FF00E0F000E0FF00E0FF00EFFF00E0F000EFFF00EFFF00EFFF00EFF0000000
AF0020B0002FAF0020B0002FAF0020B0002FAF0020B0002FB0002000000000000000F0FF00F0FF\
00F0FF00F0F000F0FF00F0FF00F0FF00F0F000F0FF00F0FF00F0FF00F0F000F0FF
00F0FF00FFFF00F0F000FFFF00F0FF00000000000080BF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F\
8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F00000000F0F000FFFF00F0F000FFF0
00F0F000FFF000F0F000F0F00000008FBF6F80B0608FBF6F80B0608FBF6F80B06080BF6F80B060\
8FBF6F80B06080BF6080B06080B06000000000F0EF00EFE000EFEF00EFE000EFE0
00EFE000EFE000000000000000E0DF00EFE000E0DF00E0DF00E0DF00E0DF00E0DF00E0DF00E0D0\
00E0DF00DFD000DFD000DFCF000000A0001F00000000D0C000D0C000CFBF00CFBF
00CFBF00CFBF00C0B09F001F9F001F9F001F90001F00000000BFA00000008F001090001F8F0010\
00AF9000A08F00AF8F00A08F000000800010009F80009F7F009F7F008F70008F70
00806F505F2F00E0DF00E0DF00E0DF00E0DF9F001F00D0CF00CFBF00C0B080001F00AF9F009080\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000090A000AFCF00BFD000000000000000E0FF00E0FF\
00EFFF00000000F0FF00F0FF00FFFF00F0F0C0002FBF0020C0002FBF0020000000
BF0020C0002FEFF02FEFFF2FBF0020B0FF60BF0020BF0020B00020BF002F00F0EF00F0EF00EFE0\
00EFE000E0D000E0D0A0001FA0002090001F00C0B0608F4000B09F70001000907F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000606F00000000607F00607F006F80006F8000708F000000
007090007090007F9F405F404F6040405F404F604F4F60400000000080A0008FAF008FAF0090AF\
008FAF0090B00090B0009FBF009FBF00A0C000A0C000AFC000000060805F5F8050
60805F00000000B0D000B0D000BFD000BFD000BFDF00BFDF00C0DF00C0DF00C0E000C0E000CFE0\
00CFE000CFEF00CFEF9F002F9F002000000000000000000000D0EF00DFF000DFF0
000000A0002000000000DFF000E0FF00DFF000E0FF00E0F000E0FF00E0F000EFFF00E0FF00EFFF\
00E0FF00EFFF00EFFF00EFFF00EFFF00EFFF00EFFF00EFFF00EFFF00F0FF000000
B0002FAF002000F0FF00F0FF00F0FF00F0FF00F0FF00F0FF000000B00020BF002FB00020000000\
00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF
00FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000000000FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00F0F00000008FBF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F\
80BF608FBF6F80BF608FBF6F00000000000000F0EF00F0EF00EFEF00F0EF00EFEF
00F0EF00EFE000F0EF00EFE000EFE000000000000000EFE000EFE000E0DF00EFE000E0DF00E0DF\
00E0DF00E0DF00E0D000E0D000000000000000DFCF00DFCF00D0C000DFCF00D0C0
00D0C000CFBF00D0C000CFBFA000209F001FA0002000C0AF00C0AF00BFA000BFA090001F90001F\
8F001000B09000AF9000AF9000A08F00A08F000000009F8000907F009F7F000000
008F7000806F00EFE000E0DF00E0DF00E0D000DFD000D0C000CFBF00BFB08F001F00AF9F009F80\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000009FA000AFC000BFDF00C0DF00D0EF00E0F000EFFF\
00EFFF00000000F0FF00F0FF00F0FF00FFFFBF0020C0002F00000000FFFF000000
C0002FBF00201F1F0FEFFF2FC0002FBF0020000000BF0020BF002FB0002000FFF000F0EF00F0EF\
00EFE000EFDF00E0D000DFCF9F001F9F002000C0B0608F4F5F80407F001F00907F
00806F000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F0010005F6F00607000607F006080006080006F800000005F001F
00708F007090405F3F406040405F404F60404060404F604F000000008FAF008FAF0090AF008FAF\
0090B00090B0009FBF009FBF00A0C000A0C05F80505F7F505F805F5F8050608F5F
5F8050608F5F608F5F608F5F00000000BFDF00BFDF00C0E000C0DF00C0E000C0E000CFE000CFE0\
00CFEF00000000D0EF00D0EF00D0F000D0EF00DFF000D0EF00DFF000DFF000DFF0
00DFF000000000000000E0FF00DFF000E0FF00E0F000E0FF00E0F000EFFF00E0FF00EFFF00E0F0\
00EFFF00EFFF00EFFF00EFF000EFFF00EFFF00F0FF00EFF000F0FF00000000F0FF
00EFF000F0FF00F0FF00F0FF00F0F000F0FF00F0FF00F0FF00F0F000F0FF00F0FF00F0FF00F0F0\
00F0FF000000F0FFFF00F0F000F0FF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF
00F0F000FFFF00F0FF00000080BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F\
8FBF6F00000000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00F0F000FFFF000000\
8FBF6F80BF6000F0F000F0F000F0F000F0EF00F0F000EFEF00F0EF00EFEF00F0EF
00EFEF00F0EF00EFEF00F0EF00EFE000EFEF00EFE0000000B00020B0002000000000EFE000E0DF\
00E0DF00E0DF00E0DF00000000000000DFD000E0D000DFCF00DFCF00D0CF00DFCF
00D0C000D0C000D0BF00D0C000CFBFA0001F9F001F00000000BFAF00C0AF00BFA000BFA0000000\
90001F8F001000AF9000A08F00AF8F00A08F00A08F009F80009F7F00907F000000
008F6F008F6F00706000EFE000E0DF00E0DF00DFCF00D0CF00C0BF00C0B000B0A0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000909F00000000BFD000C0E09F0020A0002F00E0F0\
00EFFF00EFFFB0002F00F0F000F0FF00F0FFC0002F00000000000000FFFF00FFFF
BF0020C0002FBF00207FFF90BF0020BF002FBF0020BF002FB00020BF0020B0002000FFF000F0E0\
00EFE000E0DF00E0D000DFCF9F002090001F00C0B0608F405F804000A080700010
007F60000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000004F001050001000607F006080006080006F8F006F8F5F001F5F0010
007F900000004F6040405F404F604F4F60404F6F4F4F6040008FAF008FAF0090B00090B0009FBF\
009FB0009FBF009FBF00A0C00000005F805F5F8050608F5F5F805F608F5F608F5F
608F60608F5F60906000000000C0E000C0DF00C0E000C0E000CFEF00CFE000000090002000D0EF\
00D0EF00D0F000D0EF00DFF000DFF000DFF000DFF000E0F000DFF000E0F000DFF0
00E0FF00E0F0000000A0002000000000E0FF00EFFF00EFFF00EFFF00EFFF00EFFF00EFFF00F0FF\
00EFFF00F0FF00EFFF00F0FF00F0FF00F0FF00F0FF000000B0002000000000F0FF
00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF\
00F0FF00FFFF00F0FF000000BF0020BF002FBF002000000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF80BF6F8FC06F8FBF6F8FC06F80BF6F8FC06F8FBF6F8FC06F80BF6F8FC06F\
8FBF6F8FC06F80BF6F8FC06F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
C00020C0002FBF002000000000F0F000FFFF00F0F000FFF000F0F000F0F000F0EF
00F0F000F0EF00F0F000EFEF00F0F000EFEF00F0EF00EFEF00F0EF000000BF0020B00020B0002F\
00000000EFE0000000B0002000000000E0DF00E0D000E0DF00DFD000E0D000DFCF
00DFCF00DFCF00DFCF00D0C000D0C000D0BFA0002000000000CFB000C0AF00C0AF00BFA000BFA0\
00000090001F00AF9000B09F00AF9000AF9000A08F00A08F000000009F80000000
000000008F6F008F6F00E0E000E0E000000000DFD000D0C000CFBF00000000BFAF7F001F007F6F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000090A080001F00BFDF00C0DF9F0020A0002000EFFF\
00EFFF00EFFFB0002000F0FF00F0FF00000000000000FFFF00FFFF00FFFF00FFFF
C0002FC00020C0002F000000C0002FBF0020C0002F000000BF002FBF0020BF002F00F0EF00FFEF\
00EFE000EFDF00E0D000DFCF9F001F9F00206F904F608F4F5F804000A08F000000
00806F000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F000F4F001000000000608000607F006F80006F800000005000105F001F
000000405F40405F404060404060404F6F4F000000008FAF008FAF0090B00090AF009FB0009FB0\
00A0BF009FBF00A0C000A0C00000005F80505F805F5F8050608F5F5F8F5F608F5F
608F5F60906000000000C0E000C0DF00CFE00000009000209000209F002090002000D0EF00D0EF\
00D0F000D0EF00DFF000DFF000DFF000DFF000DFF000DFF000E0F000DFF000E0F0
00E0F000E0FF000000AF002FA0002000000000EFF000EFFF00EFFF00EFFF00EFF000EFFF00EFFF\
00F0FF00EFF000F0FF00EFFF000000AF0020B0002FB0002000000000F0F000F0FF
00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF\
00FFFF00F0F000FFFF00F0FF00FFFF000000BF002FBF0020BF002FBF0020BF002F
BF0020BF002F0000008FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80BF6F\
8FBF6F80BF6F8FBF6F80BF6F00000000F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF000000BF0020C0002FBF0020\
C0002FBF0020C0002FBF0020C0002FBF0020C0002F00000000000000F0F000F0F0
00F0F000F0F000F0EF00F0F000F0EF00F0F000EFEF00F0EF00EFEF00F0EF00EFE0B00020B00020\
B00020B00020B00020AF0020B0002000E0DF00E0DF00E0D000E0DF00DFD000E0D0
00DFCF00DFD000DFCF00DFCF00D0C000D0C0000000A0001F00CFB000CFB000C0AF00C0AF00BFA0\
00BFA000B09F90001F00AF9000B09000AF8F00AF8F00A08000A080000000000000
7F0010009070008F6F00000000E0DFA0002000DFCF00D0C000C0BF00C0B000B0A0009F8F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000008F9F00AFBF00B0D000C0DF000000A0002000E0F0\
00000000EFF0000000B00020BF002F00000000FFFF00FFFF00FFFF00F0F000FFFF
BF0020C0002FBF0020000000BF0020C0002FBF002000FFF0BF0020BF002FB0002000FFF000F0EF\
00F0E000E0DF00E0D000DFCF9F002090001F6F904F608F405F8040009F8000907F
007F60000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F00104F001000607F00607F006F80006F805F001F5000105F001F5F001F
000000405F404F604F4060404F6F4F008FA00090AF008FAF0090B00090B0000000000000000000\
00A0BF00AFC000AFC00000005F8050608F5F5F8F5F608F5F608F5F60906060905F
6F906060905F0000009000209000209000209F002F90002000000000D0EF00DFF000D0EF00DFF0\
00DFF000DFF000DFF000E0F000DFF000E0FF00E0F000E0FF00E0F000EFFF00E0F0
00EFFF000000AF002FAF0020AF002FAF0020AF002F00EFFF00F0FF00EFFF00F0FF00F0FF00F0FF\
00F0FF00F0FFB00020B0002FB00020B0002FB00020B0002F00000000FFFF00FFFF
00FFFF00F0FF00FFFF00FFFF00FFFF00F0FF00FFFF00FFFF00FFFF00F0FF00FFFF00FFFF00FFFF\
00F0FF00FFFF00FFFF00FFFF000000C0002FBF0020C0002FBF0020C0002FBF0020
C0002FBF002000000000FFFF00FFFF00FFFF0000008FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F\
8FBF6F8FC06F8FBF6F8FC06F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00000000FFFF000000000000000000000000000000000000000000\
000000C0002FC00020C0002FBF0020000000C00020C0002FBF0020C0002FBF0020
00000000000000FFFF00F0F000FFF000F0F000FFF000F0F000F0F000F0EF00F0F000EFEF00F0EF\
B00020BF002FB00020B00020B0002000000000E0DF00EFDF00E0DF00E0DF00E0D0
00E0D000DFD000E0D000DFCF00DFCF00D0C000DFCF00000000D0BF00CFB000CFBF00C0B000C0B0\
00BFAF00C0AF00BFA090001F00B09F00B09F00AF9000AF8F00A08000A0807F0010
7F0010009070009070008F6F0000009F002000DFD000D0C090002000BFAF00B0A0600010000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000909F00A0BF00BFDF00C0DF000000A00020000000\
7FAF6F00F0FF000000BF002FBF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
000000BF0020C0002F00FFFFC0002FBF0020C0002F00FFF000FFF0BF0020BF002F00000000F0EF\
00EFE000EFDF00E0D000DFCF9F001F9F00206F904F608F4F5F804000A08F00907F
00806F000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F001000607000607F00608000000050001F0000005F001F500010007F90
007F904060404060400000000000004F6F4F4F6F4F50704F4F6F4F50705050704F5F7F50507F50\
5F7F505F7F505F805F0000005F8F5F5F8F5F608F5F5F8F5F60905F00000000C0E0
00C0E09000209000209000209000209000209000209F002000000000D0F000D0EF00DFF000DFF0\
00DFF000DFF000E0F000DFF000E0F000E0F000E0F000E0F000EFF000E0F000EFFF
00EFF0000000A00020AF002FAF0020AF002FAF0020AF002F00000000F0FF00F0FF00F0FF00F0F0\
B0002FAF0020B0002FAF0020B0002FB00020B0002F00000000FFFF00F0FF00FFFF
00F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F0\
00FFFF00F0FF00FFFF00000000FFFF00FFFF00FFFF00F0F000FFFF000000C0002F
BF0020BF002F00000000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFF00FFFF00F0F0000000BF0020C0002FBF0020C0002FBF0020C0002F
BF0020C0002F00000000000000F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFF00FFFF000000C0002FBF002000000000F0F000FFFF00FFFF00FFFF
00F0F000FFFF00F0F000FFFF00F0F000000000000000000000000000F0F000F0EF00F0F000EFEF\
00F0EFB00020BF0020B00020B00020B00020B0002000000000EFDF00E0DF00E0DF
00E0D000E0D000DFD000E0D000DFCF00DFCF00D0C000DFC000D0BF00D0BF00CFBF00CFBF00C0AF\
00C0AF00BFAF00BFAF00B09F00000000B09F00B09000A08F00A08F009F80000000
7F00107F0010008F70008F709F00209F002000D0CF9F002000C0B000BFAF00AF9F00806F000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000A0BF00B0D000C0DF900020A0002070AF60\
7FB06F7FB06F00F0FFB00020BF002F00F0FF00FFFF00F0F000FFFF00FFFF00FFFF
BF0020C0002FBF002000FFFFBF0020BF002FBF002000FFF000F0F0000000B0002000000000F0EF\
00EFE000E0DF00E0D000DFCF9F00200000006F904F5F8F405F8040009F8000907F
007F60000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F5F104F001000000050001F000000006F8F006F8F0000005F001F007F90007F90
00809F00809F0080A00080A04F6F4F4F6F4F50705050704F507F50507F505F7F50507F505F805F\
5F80505F805F00000000BFD000B0D000000000BFDF00000000C0E000CFE000C0E0
90002F90002090002F9000209F002F00D0EF000000000000A0002F9F002000000000DFF000E0FF\
00E0F000E0FF00E0F000EFFF00E0F000EFFF00EFF000EFFF00EFF000EFFF00EFF0
00F0FF000000AF002FAF0020B0002FAF0020B0002FAF0020B0002F00000000FFFFAF0020B0002F\
B00020B0002FB00020B0002FB00020B0002FB0002000FFFF00FFFF00FFFF00FFFF
00FFFF000000000000000000BF002FBF0020BF002FBF0020BF002F00000000FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFFBF002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF000000C00020C0002FBF0020C0002FC00020C0002FBF0020
00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFFC0002F00000000FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0F000FFFFBF0020BF002FBF0020BF002F\
BF0020BF002F000000000000B00020000000B00020B0002FB00020B00020000000
00EFE000E0DF00E0DF00E0D000E0DF00DFD000E0D000DFCF00000000D0C000D0C000CFBF00CFBF\
00CFB000CFB000C0AF00C0AF00BFA0BFBFA000B09F00B09F00AF8F00A08F7F0010
80001F7F001000000070001000806F9F002000000090001F00C0BF00B0A070001F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000A0B000BFDF00C0DF9F00200000007FAF6F\
7FB06F80B06F7FB06FBF002FBF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
000000C0002000000000FFFF00FFFFBF0020BF002F00F0F000FFF000FFF0BF002FB0002000FFEF\
00EFE000EFDF00E0D000DFCF9F001F00D0BF60904F608F4F5F804000A08F00907F
00806F000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F60104F00104F00104F0010006F80006F8000708F500010000000007090007F9F
007F9F0080A00080A00000004F6F4F50704F4F704F507F505070505F7F50507F505F80505F8050\
5F805F5F805000B0D000B0D000BFDF00BFDF00C0DF00C0DF00CFE000C0E000CFEF
90002090002090002000000000D0EF00DFF000D0EF00DFF000DFF000DFF000DFF0000000000000\
A0002FA00020A0002FA0002000000000000000000000EFF000F0FF00EFF000F0FF
00F0FFAF002FAF0020AF002FAF0020B0002FAF0020B0002FAF002070FF907FF07F101F00B00020\
B0002FB00020B0002FB00020AFFF60BFF040B0002FB00020B0002FB00020B0002F
B00020BF002FB00020B0002FB00020BF002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00000000000000FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF000000C0002FBF002000000000000000FFFF00F0F0000000
BF0020C0002F00000000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0000000BF0020\
BF002FB00020BF002000000000F0EF00EFE000F0EF00EFE000EFE000EFE000EFE0
00000000EFE000000000E0DF00E0D000E0D000DFCF00DFCFA0001F00DFCF00D0C000D0C000CFBF\
00CFBF00C0B000C0B000C0AF00C0AF00BFA000000000AF9000AF9000A08F800010
009F80009F7F00000070001000806FCFDFD090001F00CFBF00BFAF00B0A0008F7F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000A0B000B0D000C0DF00000000DFEF70AF60\
7FB06F7FAF6F80B06F000000BF002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF
BF0020C0002F00F0F000FFFF00F0F0BF002FBF002000FFF000F0F000FFF0000000B0002000F0EF\
00F0E000E0DF00E0D000000000D0C000CFBF000000608F405F8040009F8000907F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F5F1050001F4F0010006F80006F8000708F00708F5F001F007090007F9F007F9F
0080A00080A00000004F6F4F50705050704F507F50507F505F8050507F505F805F5F805F608F5F\
5F805F00000000BFD000C0DF00BFDF00C0E000C0E000CFE000CFE000D0EF000000
9F002F00000000DFF000D0EF00DFF000DFF000E0F000DFF000E0F000E0F000E0FF00E0F000EFFF\
00E0F0AF002FA00020AF002FA00020AF002FAF0020AF002FAF0020B0002FAF0020
000000A0F05FA0FF603F400F3F4010AF0020000000DFF02FDFFF2FDFF02FDFFF30DFFF2FDFFF2F\
DFFF2FE0FF30DFFF2FE0FF2FDFFF2F70801FB00020B0002F000000B0002FB00020
B0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFFBF0020BF002FBF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00000000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00000000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002F00000000FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000\
BF0020C0002FBF0020BF002FBF002000F0F000F0EF00F0EF00EFEF00F0EF00EFE0
00EFEF00EFE000EFE000EFDF000000AF0020000000AF0020AF0020A0001F00DFCF00D0C000D0C0\
00D0BF00D0BF00CFB000CFB000C0AF00C0AF8F001F0000000000008F001F00A08F
00A08F009F7F009F7F700010008F700070609F002000C0BF00BFAF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000007070009FAF8F002000000000D0E000D0EF000000\
7FB06F80B06F7FB06F00F0FFB0002000000000FFFF00FFFF00FFFF00FFFF00FFFF
C0002F00000000FFFF00FFFF00FFFFBF0020C0002F00FFF000FFF000F0F000FFF0B0002000FFEF\
00EFE000EFDF00E0D000000000D0C000D0C0000000608F4F00AF9000A08F00907F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F00104F001000607F00607F006F8F006F8F000000007090007F9F007F9F0080A0
0080A0008FAF4F6F4F50704F4F704F507F50507F505F7F50507F505F805F5F80505F8F5F5F805F\
608F5F00BFD000C0DF00BFDF00C0E000C0E000CFE000CFE000CFEF00CFEFC0D0EF
00D0EF00D0EF00D0EF00DFF000DFEF00DFF000DFF000E0F000E0F000E0FF00E0F000EFFF00E0F0\
00EFFF00EFF0000000A00020AF002FAF0020AF002FAF0020AF002FAF0020AF002F
000000DFF02FD0F02FDFFF2FD0F02FDFFF2FD0F02FDFFF2FD0F02FDFFF2FDFF02FDFFF2FDFF02F\
DFFF2FDFFF2FDFFF2FDFF02FDFFF2FDFFF2FDFFF2FDFF02FA0FF603F3F0FB0002F
B00020B0002FB00020B0002FB00020B0002FB00020B0002FB00020B0002FB00020B0002FB00020\
B0002F00000000FFFF00F0F000000000000000FFFF000000000000B00020B0002F
B00020B0002FB00020BF002FB00020B0002FB00020BF002FB0002000FFFF00FFFF000000B00020\
BF002FB00020BF002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFFC0002F00000000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFFBF0020C0002F00000000FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF000000\
BF002F00FFFF00FFFF00F0F0000000BF002000000000F0EF00F0EF00EFEF00F0EF
00EFE000EFEF00EFE000EFE000E0DF00E0DFAF0020AF0020A0001FAF0020A0001FA00020000000\
00D0C000D0BF00D0BF00CFB000000090001F90001F8F00108F001F80001000AF90
00A08000A080009F7F7F0010008F6F00806F00CFC08F001F00000000A090000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F0010009FB000000090002000CFE000DFEF00DFEF\
7FB06F7FB06F80B06F00EFF000F0FFBF002000FFFF00F0F000FFFF00FFFF00FFFF
BF002000FFFF00FFFF00FFFF00F0F0BF002FBF002000FFF000F0F000FFF000F0EF00FFF000F0EF\
00EFE000E0DF00E0D000000000D0C000CFBF00C0B05F8F4000AF90009F8000907F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F001000607F00607F006F80006F8F007090007090007F9F007F9F0080A00080A0
008FAF00000050704F4F704F507F50507F505F8050507F505F8F5F5F805F608F5F5F8F5F609060\
00000000C0E000C0DF00CFE000C0E000CFEF00CFE000D0EF00D0EF000000000000
00DFF000DFF000DFF000DFF000E0FF00E0F000E0FF00E0F000EFFF00EFFF00EFFF00EFFF00F0FF\
00EFFF00F0FF00EFFF00F0FF00F0FF000000000000DFFF30D0F02FDFFF2FD0F02F
DFFF30D0F02FDFFF2FD0F02FDFFF30DFFF2FDFFF2FDFFF2FDFFF30DFFF2FDFFF2FDFFF2FDFFF30\
DFFF2FDFFF2FDFFF2FDFFF30DFFF2FDFFF2FDFFF2FDFFF30DFFF2FDFFF2FDFFF2F
DFFF30000000B0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF000000B00020B0002FB00020B0002F00000000FFFF00FFFF00FFFFB00020BF002F\
B00020BF002FB00020BF002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000C0002FBF0020000000\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002FC00020C0002FBF0020
00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000000000\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000BF002F00F0F000F0F000F0EF
00F0F000F0EF00F0EF00EFE000EFE000EFE000EFE0000000AF0020AF0020AF0020A00020AF0020\
A000200000000000000000009F001F9F001F90001F90001F8F001F00000000AF90
00AF9000A08000A08070001000000000806F90002000000000B0A0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000070001F009FB000000090002000CFE000D0EF00E0F0\
7FB06F80B06F7FB06F00F0FF00F0FF00000000F0FF00FFFF00FFFF00FFFF00FFFF
C0002F00FFFF00FFFF00FFFF00FFFF00000000000000F0F000FFF000F0F000FFF000F0EF00FFF0\
00EFE000EFDF00E0D000DFCF00D0C000D0BF00C0B000000000AF9000A08F00907F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000006070006080006080006F8F006F8F007F90007F900080A000809F008FAF
008FAF4F704F4F704F5070505070505F7F50507F505F805F5F8050608F5F5F8F5F608F5F000000\
00C0DF00C0DF00C0E000C0E000CFEF00CFE000D0EF00D0EF00000090002000DFF0
00DFEF00DFF000DFF000E0F000E0F000E0FF00E0F000EFFF00EFF000EFFF00EFF000EFFF00EFFF\
AF002FA00020AF002FAF0020AF002FAF0020AF002FAF00206F7F1FD0F02FD0FF2F
D0F02FDFFF2FD0F02FDFFF2FD0F02FDFFF2FD0F02FDFFF2FD0FF2FDFFF2FD0F02FDFFF2FD0FF2F\
DFFF2FD0F02FDFFF2FD0F02FDFFF2FD0F02FD0F02FAF0020AF002FA00020AF002F
00000000000000EFF000F0FF00EFF000F0FF00EFF000F0F000EFF000F0FF00EFF000EFF000EFF0\
00EFF000EFF000EFF000EFF000EFF000E0F0000000000000AF002FA00020AF002F
A00020AF002FAF002000F0FF00EFF000F0FF00EFF000F0FF00F0F000F0FF00F0F000F0FF00F0F0\
B0002FB0002000000000FFFF00FFFF00F0F000FFFF00000000FFFF00000000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0000000BF0020\
BF002FBF0020BF002F00FFFF00FFFF00F0F000FFFFBF0020C0002FBF0020C0002F
BF0020C0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF000000BF002F000000\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00F0F000FFFF00000000F0F0
00F0F000F0F000EFEF00F0EF00EFE000EFE000E0DF00EFDF00E0DF000000A0001F00000000DFCF\
00DFCF00D0C000DFC000D0BF00D0BF00CFB000000090001F90001F00B09F00B09F
00AF8F00AF8F009F80000000008F70008F70007F6000B0AF00A090000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F001000000080001F90002000C0DF00D0EF00DFEF\
7FB06F7FAF6F80B06F00EFFF00F0FF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00000000FFFF00F0F000FFFF00F0F0000000BF002000FFF000F0F000FFF000F0EF00F0F0000000\
00F0E000E0DF00000000DFCF00D0C000CFBF00C0B000B0A000AF90009F8000907F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F0010006F80006080006F8F006F8F000000007F900080A00080A0008FAF008FAF
0090B04F704F507F50507F505F7F50507F505F805F5F805F608F5F5F8F5F60906000000000C0E0\
00C0DF00CFE000CFE000CFEF00CFEF00D0F000D0EF0000009F002000000000DFF0
00E0F000E0F000E0FF00E0F000EFFF00EFFF00EFFF000000AF002FA00020AF002FAF0020AF002F\
AF0020AF002FAF0020AF002FAF002020200FD0F02FDFFF2FD0F02FDFFF30D0FF2F
DFFF2FD0F02FDFFF30D0FF2FDFFF2FD0FF2FDFFF30D0FF2FDFFF2FD0F02FDFFF30D0F02F000000\
AF0020AF002FAF002000000000EFF000F0FF00EFF000EFFF00EFF05F704F5F704F
60805000AFC000BFD000BFD07090606F905F70A06070A06000DFF000E0F000EFFF00EFFF00F0FF\
00EFFF00F0FF00000000F0FF00F0FF00FFFF00F0FF00FFFF00FFFFC0002FC00020
00FFFF00FFFF00FFFF000000C0002FBF0020000000C00020C0002F00000000FFFF0000007FB06F\
000000000000AF0020AF002F00F0F000F0FF00F0FF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF000000BF002FB00020BF002FB00020BF002F00000000FFFF00FFFF00FFFF\
000000BF002FBF0020BF002FBF0020C0002F000000000000BF0020C0002FC00020
C0002FBF0020C0002F00000000FFFF00FFFF00FFFF00FFFFC0002FBF0020C0002FBF0020000000\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
BF002F00F0F000FFF000F0EF00F0EF00EFE000000000000080B0607FAF5F7FB06000E0D000E0D0\
00DFCF00E0D000DFCF00DFCF00D0BF00D0BF00CFB000C0B000000000BFA000B09F
00B09F00AF8F00AF8F009F8000907F008F6F00806F00AF9F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000A0BF8F002000000000CFE000D0EF00E0F0\
00EFF080B06F00EFFF00F0FF00F0FF00FFFF00F0FF00000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFF0C0002F00FFF000FFF000F0F000FFF000F0EF00FFF0\
00EFE000EFE0A0002000DFCF00D0C000D0BF00C0B00000005F804000A08F00806F
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F00104F0010006F80006F8000708F500010007F9F00809F008FA0008FAF0090B0
0090B050705050704F507F50507F505F805F5F80505F8F5F5F8F5F60905F00000000C0E060905F\
609F606090606F9F6000000000D0EF00D0EF0000009000209F0020000000A0002F
00000000000000000000000000E0F000EFFF00EFFF00EFFF00EFF000F0FF00EFFF00F0FF00F0F0\
00F0FF00F0FF000000AF0020AF002FAF0020AF002FAF0020A0FF60D0F02FDFFF2F
D0F02FD0FF2FD0F02FD0FF2FA00020AF002FA00020AF002F00EFF000F0F000EFF000F0FF00EFF0\
0000000000000000000000005F001000708F007F9F00809F506F4F50704F5F7F50
5F7F5000B0CF00B0CF6F905F6F905F709F60709F607FA06F00DFEF00E0FF00EFFF00EFFF00EFF0\
00F0FF00EFFF00000000F0F000F0FF00F0FF00FFFF00000000000000FFFF00FFFF
00F0F000FFFF000000C0002FBF002000000000FFFFC0002FBF002000FFFF00FFF000FFFFE0F0F0\
00000000D0CF00AFAF00A0A000A0A080001F009F9FA0002000000000EFF000EFFF
00000000F0FF00F0F000F0FF00F0F000FFFF00FFFF000000B00020B0002FB00020BF002FB00020\
BF002FB00020000000B00020BF002FBF0020BF002F000000CFFF40CFFF40BF002F
BF0020C0002FBF0020C0002F00F0F0000000BF0020C0002FBF0020C0002FBF0020C0002F000000\
00FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF
00F0F0BF002F00000000F0F000F0EF00F0EF00000080B0607FAF5F7FB0607FAF5F7FB06070AF5F\
00E0D000DFCF00E0D000DFC000DFC000D0BF00D0BF00C0B06F904F608F40608F40
00B09000AF9000A08F000000009070000000008060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000090A000AFBF00B0CF00C0DF00CFDF00D0EF00DFEF\
00EFFF00000000F0FF00EFF000F0FF00F0FF00FFFFBF002000FFFF00FFFF000000
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00000000FFF000F0F000FFF000F0EF00FFF000F0EF\
00000000000000000000DFCF00D0C000CFBF00CFB05F8F405F8040009F80000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000060601F\
4F001050001F00000000708F00708F5F001F007F9F008FA04F6F4F4F704F4F704F
507F500000005F7F50507F505F805F5F805F608F5F5F8F5F00C0DF00C0DF6090606090606F9F60\
6F9F606F9F606F9F6070A06F0000009F002F9F0020A0002F9F0020A0002F000000
00EFFF00E0FF00EFFF00EFFF00F0FF00EFFF00F0FF00EFFF00F0FF00F0FF00F0FF00F0FF000000\
AF0020AF002FAF0020AF002FAF0020AF002F000000B0002FAF0020AF002FAF0020
000000AF0020AF002FA00020AF002F00EFF0000000000000000000000000000000000000000000\
00000000000000606F000000006F8F007F9000809F0090AF50704F5F7F505F7F50
00B0CF0000006F905F6F905F709F60709F607FA06F7FA06000EFFF00EFFF00EFFF00EFFF00F0FF\
00F0FFBF002F00F0FF00FFFF00F0FF00FFFF00000000FFFF00FFFF00FFFF00FFFF
00FFFF00FFFFC0002FBF002000FFFF00FFFFC0002FBF002000FFFF00FFFF00FFFF00F0F0AF002F\
00D0CF00BFBF00B0B090002000000090001F00A0A000A0A0009F9F009F9F009090
009090008F8F008F8F00808000EFFF00EFF000F0FF000000B0002F00000000FFFF000000BF002F\
B00020BF002FB00020BF002FAFFF60E0FF2FE0FF2FEFFF30E0FF2FEFFF2FE0FF2F
EFFF30EFFF2FD0FF4FEFFF2FF0FF303F400FC0002FBF0020C0002FBF0020C0002FBF0020000000\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF000000BF002FBF002000000000F0F000F0F07FB06080BF607FB06080B0607FAF5F7FB060\
7FAF5F00E0DF00E0D000E0D000DFCF00DFC000D0BF00D0BF6F904F6F904F608F40
608F4F00B09F00AF907F0010009F7F008F70008F6F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000009FAF00AFBF00BFD06F905F00CFE000D0EF00E0F0\
00000080B06F00000000F0FF00F0FF00F0FF00F0FFC0002F00000000FFFFBF0020
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000F0F000FFF000F0F000FFF000F0EF00FFF0\
AF0020AF0020A0002000DFD000D0C000D0BF00C0B060904F5F803F009F80000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5F10\
4F00104F0010000000606F8F50001F5000100000004060404F6F4F4F6F4F507050
00000000A0C000A0C000AFCF00AFCF00B0D000B0D000BFDF00BFDF60906060905F609F60609F60\
6F9F60609F606FA06000000000DFF00000009F002F9F002000000000E0F000E0FF
00E0F000EFFF00EFF000EFFF00EFFF00F0FF00EFF000F0FF00F0FFAF002FA00020AF002FAF0020\
AF002F00000000FFFF00F0FF000000A00020AF002F00F0F000F0FF00EFF000F0FF
00EFF0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000006070006F7F007090007F90008FA0506F4F5F70505F7F4F608050
0000006F8F5F6F905F709F60709F607FA06070A0607FAF6F00E0FF00EFFF00EFF000F0FF00EFFF\
000000BF002000F0FF000000C0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFFC0002F00000000FFFF00FFFF00FFFFBF002000FFFF00FFF000FFFF00F0F0AF00206FA050\
00C0C000BFBF00BFBF00B0B090001F8F001F00000000A0A08F001F800010009F9F
00909F00909F008F90008F90008F8F7F001F00808000808F00E0F000EFF000E0F000EFFF000000\
B0002FAF0020AFFF60DFFF2FE0FF2FDFF02FE0FF2FE0FF2FE0FF2FE0F02FE0FF2F
E0FF2FEFFF2FE0F02FEFFF2FE0FF2FEFFF2FAFF05FBF002FBF0020C0002FBF0020C0002F00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFF000FFFF00F0F0BF002FB00020BF002F00000080BF607FB06080B0607FB0607FB0607FAF5F\
7FB06000000000E0D000DFCF00DFCF00D0C000D0C000CFBF6F9F5060904F6F904F
608F4000000000AF90000000009F7F00907060000F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005000100090AF00AFC00000006F90606F9F5F00D0EF00DFEF\
7FB06F7FB06F80B06F00EFFF00F0FF00F0F000F0FF00F0FFC0002FBF0020C0002F
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00000000FFF000F0F000FFF000F0EF00F0EF00F0EF\
B00020A0002000000000DFCF00D0C000CFBF00C0B05F80405F7F3F00907F000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000606010\
00607F006F80006F8000708F5000105F001F0000004F6F4F4F6F4F507F50507F50
00000000A0C000AFCF00B0CF00B0D000B0D000BFDF00BFDF00C0E06090606F9F606F9F606F9F60\
6F9F6070A06F00000000DFF000DFF000E0FF000000A0002F00000000EFFF00EFFF
00EFFF00EFFF00F0FF00EFFF00F0FF00F0FF00F0FF000000AF002F00F0FF00FFFF00F0FF00FFFF\
00F0FF00FFFFAF002000000000F0F000F0FF00F0F000F0FF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000006070006F805F00100080A0008FA05F704F5F7F4F60805F000000
6F8F5F6F8F5F709F60709F6070A06070A0607FAF6F7FAF6F00EFFF00EFFF00F0FF00F0FF00F0FF\
BF0020C0002FBF0020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
C0002F00FFFF00FFFF00FFFF00FFFFBF002000FFFF00FFFF00FFFF00F0F0AF002F6FA05F00D0D0\
00C0CF00CFCF00C0C000BFC09F001F00000000B0B090001F8F001F90001F00A0A0
00A0A0009F9F009F9F00909F00909F008F900000007F001000000000808F00808F70001070001F\
00E0F0AF002FAF0020AF002FAF0020DFFF30DFFF2FE0FF2FDFFF2FE0FF30E0FF2F
E0FF2FE0FF2FEFFF30E0FF2FEFFF2FE0FF2FEFFF30EFFF2FEFFF2FEFFF2FF0FF303F400FC0002F\
BF0020C0002FBF0020C0002FBF0020C0002FBF002000000000000000FFFF000000
000000BF0020BF002FBF0020BF002FBF0020BF002F00FFF080BF6F80B06080BF6F7FB06080B060\
7FB06000000000E0DF00E0D000DFCF00DFCF00D0C000D0C0000000709F506F904F
6F904F608F4080001F000000009F80009070000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000007F8000A0B000AFC06F90606F905F70A06000D0EF7FAF6F\
7FAF6F80B06F80B06F00000000F0FF00FFFF00F0FF00FFFFBF0020C0002FC00020
00FFFF00FFFF00FFFF00FFFF00FFFF00FFF000000000FFF000FFF000F0F000FFF000F0EF00FFF0\
AF0020AF002000E0D000DFD000D0C000D0BF60904F608F40507F3F00907F000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000005F5F10\
00607F006080006F8F006F8F0000005F001F008FA0406F404F704F4F704F507F50
00000000AFCF00AFCF00B0D000B0D000BFD000BFDF00C0DF00C0DF000000609F606F9F60609F60\
6FA06000000000DFF000DFF000E0FF00E0F000E0FF00E0F000000000E0F000EFFF
00EFF000EFFF00EFFF00F0FF00EFF000000000F0FF00F0FF00F0F000FFFF00F0F000FFFF00F0F0\
00000000F0F000F0FF00EFF0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000606F006F7F006F8060001F00000050704F50704F5F7F5000AFC0608F5F
608F5F6F905F6F905F70A06070A0607FAF6F7FAF6F00000000EFF000EFFF00EFFF00F0FF00F0F0\
BF002FBF002000000000F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFFC0002F
00F0F000FFFF00FFFF00FFFF00000000FFFF00FFF000FFFF00F0F000EFE06FA050B0002000D0D0\
00D0D000CFCF00CFCF00000000BFC000BFBF00BFBF00000090001F00AFAF000000
00000000A0A0009FA0009FA000909F00909F00909080001F7F00107F001F7000107F001F700010\
70001F00000070001F6F00106F001000E0F000EFFFAF00208FF07FDFF02FE0FF2F
DFF02FE0FF2FDFFF2FE0FF2FE0F02FE0FF2FE0FF2FEFFF2FE0F02FEFFF2FE0FF2F90FF7FBF0020\
BF002FBF0020C0002FBF0020C0002FBF0020C0002FBF0020BF002F00FFF000FFFF
00F0F000FFFF00FFF000FFFF000000BF002FB0002000000000F0F000FFF07FB06080BF607FB060\
7FB06000000000000000000070AF5F00000000000000D0C000D0BF6F9F4F6F9F4F
60904F608F4F800010000000009F8000907F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000006F001F00000000B0CF608F5F6F90606F9F6000DFEF70AF60\
7FB06F7FB06F80B06F7FB06F00F0FF00F0FF00FFFF00F0F0000000BF0020C0002F
00F0F000FFFF00FFFF00FFFF00F0F000FFFFBF002000FFF000F0F000FFF000F0EF00FFF000F0EF\
AF0020A0002000E0D000D0CF00D0C000CFB060904F5F803F507F3F007060000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000060601F\
000000006F80006F8F007F905F001F0080A0008FA04F704F4F704F507F50000000
00AFCF00AFCF00B0D000B0D000BFDF00BFDF00C0DF00C0DF00CFE000CFE000000000D0EF00DFF0\
00DFF000DFF000DFF000E0FF00E0FF00EFFF00E0FF00EFFFA00020A0002F000000
00F0FF000000AF002F00000000FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00F0FFA0002000F0FF\
A00020A0002F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000607000607F00709060001F70001F0000005F7F5000000000B0CF608F5F
6F90606F905F709F6070A0607FAF6F7FAF6F80B06F00EFFF00F0FF00EFFF00000000F0FF000000\
BF0020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002F00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000EFE070A05F00E0EF00DFE0B00020\
00D0DF000000A0002000CFD000C0CF00C0CF00BFC000BFC09F001F00B0BF00AFB0
00AFB000A0AF0000008F001F8F001F000000009FA000909F80001F7F001F7F001F908F1F9F8F1F\
4F400F90801F90801F90801F8F7F1F70001F6F001070001FA00020AF002FAF0020
3F4010DFF02FE0FF2FDFFF2FE0FF30E0FF2FE0FF2FE0FF2FE0FF30E0FF2FEFFF2FE0FF2FEFFF30\
CFFF40C0002FBF0020C0002FBF002000000000000000FFFF00FFFF00FFFF00FFF0
00FFFF00FFF000FFFF00FFF000FFFF000000BF002F00000000FFFF00FFF000FFF000F0F000FFF0\
00F0EF0000007FAF5F7FB06070AF5F7FAF5F70A05070A05F6FA0500000006F9F4F
6F9F500000008F001F7F001000A08F00907F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005F001000000000000000B0CF6F90606F906070A0600000007FAF6F\
7FB06F80B06F80B06F80BF6F00F0FF00FFFF00F0FF00FFFF00FFFFC0002FBF0020
C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00000000F0F000FFF000F0F000FFF000F0EF000000\
AF0020AF002000DFD000DFCF00D0C0000000608F405F804050703F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F001000608000708F0070905F001F00809F008FAF4F6F4F5070504F704F00AFC0
00AFCF00B0D000B0CF00BFD000BFD000C0DF00C0DF0000008F002090002090002000000000D0EF\
00DFF000DFF000E0FF00E0F000E0FF00E0F000EFFF00EFF0A0002FA00020A0002F
A00020AF002F00F0F000F0FF00F0F000F0FF00F0F0AF002FA00020AF002FA0002000F0FF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004F0010006F8000708F6F001F6F001F009FB000A0BF00B0CF0000006F905F
6F905F709F60709F607FA0607FA0607FB06F7FAF6F00EFFF00000080B06F80B06F000000BF0020\
C0002FBF002000000000FFFF00FFFF00F0F000FFFF00FFFF00000000F0F000FFFF
00FFFF00FFFF00F0F000000000FFFF00FFFF00F0F000EFE06FA05F00EFEF00E0E0000000B00020\
B00020AF002000000000CFD000CFD000C0CF00C0CF00BFC000000000B0BF00B0BF
00AFB000AFB000A0AF00A0AF0000008F001F80001F80001F80001FA0901F9F8F1F9F901F9F8F1F\
9F8F1F90801F90801F8F801F90801F8F7F108F7F1F6F001070001F6F00106F0010
00E0F000EFFF000000DFF02FDFF02FDFFF2FDFF02FE0FF2FDFFF2FE0FF2FDFF02FE0FF2FE0FF2F\
E0FF2FB00020BF002FBF002000000000F0F000FFFF00FFF000FFFF00F0F000FFFF
00FFF000FFFF00F0F000FFFF00FFF000FFFF00000000000000FFF000FFF000F0F000FFF000F0EF\
00F0EF0000007FB06070AF5F7FAF5F70A05F70AF5F70A05070A0506F9F5000D0BF
00000000C0AF80001000AF9000A080000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000707F0090A0009FB000B0CF6F905F709F606F9F6070A0607FAF60\
80B06F7FB06F80B06F80B06F00F0FF00F0F000FFFF00F0FF00FFFFBF0020C0002F
BF0020C0002F00F0F000FFFF00FFFF00FFFFBF002000FFF000F0F000FFF000F0EF000000B00020\
AF002000000070AF5F00D0C000D0BF60904F5F8F40507F3F4F703F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F0010\
4F0010006F8000708F007F9F00809F008FAF008FAF009FB0009FBF700020000000
7F002000000000000000BFD0B0C0DF8F002090002F8F002090002F9000209F002F900020000000\
00DFF000E0FF00E0FF00EFFF00E0FF00EFFF00EFFF000000A00020AF002FA00020
AF002FAF002000FFFF00F0FF00F0FF00F0F000F0FF7FB06F7FBF6F6FA05F006060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00606F006F7F006F8F007F9F6F001F00000000A0BF00AFCF00B0CF6F905F6F905F
709F60709F607FA06F7FAF6F7FAF6F7FAF6F00000000000080B06F80B06F8FBF6F80BF6F00FFFF\
000000C0002F00000000FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF
00FFFF00FFFFC0002F000000BF002F00F0F000EFE000DFD0BF002F00000000EFEF00E0EF000000\
000000B0002F00DFE000DFE000D0DF00D0DF00CFD000CFD000000000C0CF00BFC0
00BFC000B0BF00B0BF00AFB000AFB000A0AF0000008F001F8F001F2F200FA09F209F901F9F9020\
9F8F1F9F9020908F1F908F1F90801F908F1F00000070001F6F0010007F8F007F8F
007F8F007080007F8F00708000EFFFAF0020B0002FBFF040E0FF2FDFFF2FE0FF30E0FF2FE0FF2F\
E0FF2FE0FF30BF0020BF002FBF0020BF002FBF002000000000FFFF00FFFF00FFFF
00FFFF00FFF000FFFF00FFFF00FFFF00FFF000FFFF00000000FFFF00FFF000FFFF00FFF000FFF0\
00F0EF0000007FB0607FB0607FAF5F7FB06070AF5F70AF5F70A05070A05F00D0BF
00CFBF00C0AF00000000AF9000AF8F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000080900090AF00AFBF0000006F90606F9F6000000070A0607FB06F\
7FB06F80BF6F80B06F80BF6F00000000FFFF0000008FBF6F000000000000000000
00000000000000FFFF00FFFF00FFFF00FFF0C0002F00FFF000FFF0000000000000000000000000\
00EFDF7FB06070A05F00DFCF00CFBF6F9F505F8040507F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00607000607F00708F0000000080A0008FA00090B00090B000000070001F7F0020
7F001F80002000000000C0DF00C0DF00CFE000CFE000D0EF90002000000000D0EF00DFF000DFF0\
00E0F000E0F000E0FF000000A0002FA00020A0002FA0002000000000F0F000F0FF
00F0F000F0FF00F0F000F0FF0000007FBF6F70B06F70AF600000004F5F2F006F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00607000607F00708F0000006F001F0090B000A0C000AFC000BFD0608F5F6F9F60
6F9F5F70A06070A0607FAF6F7FAF6F80B06F7FAF6F80B06F80B06F80BF6F80B06F00FFFF00F0F0\
00FFFF00FFFF00000000F0F000FFFF00FFFF00000000000000FFFF00FFFF00FFFF
00F0F000FFFFBF0020BF002F00F0F000EFEF00000000F0F000EFEF00EFF000E0EF00EFEF00E0E0\
00E0EF00000080AF607FA05F7FA06000D0DF00D0DF00CFD0A0002000C0CF00C0CF
00BFC000BFC000B0BF00B0C000AFB000AFBF00A0B08F001F90A02FAFA020A09F1FA09F209F901F\
9F901F9F8F1F9F901F908F1F908F1F90801F6F803F6F001070001F000000007F90
00708F007F8F00708000708F600010000000000000AF002F000000000000AF0020B0002FDFF02F\
3F400FB00020BF002FB00020000000B00020BF002FB00020BF002FBF002000FFFF
00F0F000FFFF00FFF000FFFF00F0F000FFFF00FFF000000000F0F000FFF000F0F000FFF000F0EF\
00FFF00000007FBF607FB05F7FB06070AF5F7FAF5F70A05070AF5F6FA05000D0C0
00CFB000C0B000BFA000B09F00A08F009F80000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000006F70008F9F009FAF00A0BF9000206F905F709F6000D0E07FAF6F7FAF60\
80B06F7FB06F80BF6F80B06F00000000F0FF8FBF6F80BF6F8FBF6F00FFFF00FFFF
00F0F000FFFF00FFFFC0002FBF0020BF002FBF0020BF002FB00020B0002000F0EF00F0EF00EFE0\
00EFDF70AF5F70A05F00D0BF6F9F50608F4F5F80404F7030406030000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
005F70006F80006F8F5F001F0080A00090AF0090B000A0BF70001F7F00207F001F
00B0D000BFD000C0DF00C0DF00CFE000CFE000D0EF00000000000000DFF000DFF000DFF000E0FF\
00E0F000EFFF00E0F000EFFF00EFFFAF002FA00020AF002FAF002000F0FF00F0FF
00F0FF00F0FF00F0FF0000007FBF6F00EFEF80001F009F8000907F00806F007F6F006F60006F60\
404F2F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
4F0010006F8000708F6F001F70001F009FBF00A0C000B0CF0000006F90606F9F60
70A06F70A0607FAF6F7FAF6F80B06F00EFFF80B06F80B06F8FBF6F80BF6F8FBF6F00F0FF00FFFF\
00FFFF00FFFF00FFFFC0002FC00020000000BF002000FFFF00FFFF00FFFF00FFFF
C0002FBF0020C0002F00FFFF00F0F07FAF5F70AF5F00F0F000F0F000EFF000F0F000EFEF00EFF0\
00E0EF80AF6080AF6080AF607FA06080AF6000D0E0000000AF002000CFDF00C0D0
00CFD000C0CF00C0CF0000009F002090001F90002090001F9000208F001FAFA020A0A01FAFA020\
A09F1FA09F209F901F9F90209F8F1F9F8F2020200F7F001F70001F70001F6F0010
000000007F90007F906000106F001F60001060001F00708F00708F00E0F000EFFF000000B0002F\
B00020B0002FB00020BF002FB0002000000000FFFF00FFFF00FFFF000000BF0020
BF002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000BF002F00FFF000FFF000F0F000FFF0\
00FFF00000007FBF6080BF6F7FB0607FB06070AF5F7FB06070AF5F70AF5F00D0C0
00D0BF00C0B000C0AF00B09F00AF90000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000070800090AF009FB000AFC08F002000CFE0709F6000DFEF7FAF6080B06F\
80B06F80BF6F80B06F8FBF6F80BF6F00000080BF6F8FC06F8FBF6F00FFFF00FFFF
00FFFF00FFFF00FFFF000000C0002FBF0020BF002FBF0020BF002F00F0F000FFF000F0EF00F0EF\
7FAF5F70AF5F6FA0500000006F9F4F60904F50803F507F3F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00607000000000708F500010008FA0008FAF009FB06F001F70002000A0C07F0020
00B0CF00BFD000BFD000C0DF00C0DF00CFE000CFE000000000D0EF00DFF000DFEF00E0F000E0F0\
00E0FF00E0F000EFFF00000000000000EFF000F0FF00F0F000F0FF000000AF002F
A00020AF002F00000000F0F070AF5F00CFBF80001080001F009080008F7F007F6F000000000000\
0000005F000F006060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000005F60\
00607F006F800000000080A070001F009FB000AFC000B0CF00BFD06F905F709F60
709F607FAF6F7FAF607FAF6F00E0FF0000007FB06F80B06F80B06F8FBF6F00000000FFFF00FFFF\
00FFFF00F0F000FFFFBF0020C0002FBF0020C0002F000000000000000000C0002F
BF0020BF002F00F0F000F0F07FAF5F70AF5F80BF6F8FBF6F00000000F0F000EFF000F0F000EFEF\
00EFF000E0EF80AF6080AF6080AF607FA0607FA060000000AF0020A00020AF0020
00CFD000CFD000C0CF00000000000000000090001F90002090001FB0AF20AFA01FAFAF20AFA01F\
AFA020A09F1FA09F209F901FA09F203F300F101000008F9F008FA000809F008F9F
00809000809F007F90000000007F90007F90000000007F9000708F007F905F001000E0F000EFF0\
000000AF002000FFFF00FFFFB0002FB00020B0002F00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F000FFFF00000000000000F0F0BF002FB0002000FFF000F0F000FFF000F0EF\
00FFF000F0EF7FBF607FB0607FBF6070B05F7FB06070AF5F70AF5F6FA05000D0C0
00CFBF00CFB000BFA000B0A000A08F6F0010000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000080900090AF00A0BF00AFC090002000C0DF00D0E000D0E07FAF6F7FAF60\
80B06F7FB06F80BF6F80B06F80BF6F80B06F8FBF6F8FBF6F8FBF6F00F0F000FFFF
00FFFF00FFFF00F0F000FFFFBF0020BF002FB00020BF002000F0EF00F0F000F0EF00F0EF00EFE0\
7FB06070A05070A0500000006F9F505F8F40507F3F00907F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000060805000105F001F0080A00090B000000000A0C000A0C000B0CF000000
00BFD000BFD000C0DF00C0DF00CFE000CFE000D0EF00D0EF00DFF000DFF000E0F000E0F000E0FF\
00E0F000EFFF00000000F0FF00EFFF00F0FF00F0FF00F0FF00F0FF00F0FFAF0020
AF002F00F0F000EFEF70AF6000D0CF8F001F8F001F00AF9F00A0900000000090800080706F0010\
00706F00706F60001060001000605F006060000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000006070\
00607F00708F007F9F008FAF70001F00AFC000AFCF00BFD000BFD0709F60709F60
7FA06F7FAF6F00000000EFFF00EFFF80B06F80BF6F80B06F8FBF6F00000000FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFFC0002FBF0020C0002FC0002000000000FFFF00FFFF00FFFF
00FFFF80BF6F8FBF6F00EFEF7FAF608FBF6F8FC06F8FBF6F8FBF6F00F0F000FFFF00F0F000F0F0\
00EFF000F0F000EFEF80B06F80AF6080AF6F00000000DFEF00DFE0B0002F00D0E0
00D0E000CFDF00CFDF00CFD000CFD000C0CF00C0CF00BFCF00000090001F90002070AF5FB0AF20\
AFA020AFAF20AFA020AFA020A09F1F8000207F001F80001F0090A00090AF008FA0
008FA0008F9F008FA000000000809F00809F00000000809F00809F00809F6F001F00809F008FA0\
00E0F000EFFFAF002000F0FF00F0FF00FFFF000000B0002F00000000FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFBF002FBF0020BF002FB0002000FFF0000000000000\
00F0EF00FFF000000080BF6F7FBF607FBF607FB0607FB06070AF5F00000000D0C0
00D0BF00CFB000C0B000BFA000A08F6F801F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000070706F001F009FB000A0BF00BFD090002000CFE000D0E000DFEF7FAF6080B06F\
80B06F80BF6F80B06F80BF6F80BF6F8FBF6F8FBF6F8FC06F8FBF6F00000000FFFF
00FFFF00FFFF00FFFF00FFFF000000BF0020BF002F00F0F000FFF000F0EF00FFF000F0EF00F0E0\
70AF5F70AF5F6FA0506F9F5060904F608F4F507F3F00907F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000607F00708F5000100080A0008FAF009FBF00A0BF00AFC000AFC000B0CF
00000000BFDF00C0DF00CFE000CFE000D0EF0000009F002F00DFEF00DFF000DFF000E0F000E0F0\
A0002F00000000F0FF00EFF000F0FF00EFF000F0FF00000000F0FF00F0F000F0FF
A00020A0002000E0E000DFD090001F90001F8F001F00B0A000AF90000000009080009080000000\
6F0010007F6F007F6F000000007060006F6000000050000F005F5F000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F0010\
0060805F001000809F008FAF00000000A0C000B0CF00B0D000C0DF6F905F70A060
70A06000000000E0FF00EFFF00EFF080B06F80B06F8FBF6F00000000FFFF00F0FF00FFFF00F0F0\
00FFFF00FFFF00FFFF000000C0002F00000000FFFF00F0F000FFFF00FFFF00FFFF
80BF6F8FBF6F0000007FB0606FA0508FBF6F8FBF6F8FBF6F80BF6F00FFFF00000000000080B060\
8FBF6F80B06080B06F80AF6000EFEF00E0EF00E0EF00DFE000DFEF00DFE0AF0020
00D0DF00D0E000CFDF00D0DF00CFD000CFD000C0D00000009F001F9F002090001F30300F2F2F0F\
900020AFAF20B0AF2080001F00000080001F80001F7F001F80001F0090AF0090AF
0090A00090AF008FA0008FA0008FA0008FA0000000008FA000809F0000000080A0008FA0008FA0\
000000608F5000000000E0F000EFFF00F0F000FFFF00F0F000FFFFB0002000FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000BF002FB00020B0002F80BF6080BF6F7FB060\
80BF6F7FBF600000007FB0607FBF607FB0607FB06070AF5F70AF5F00000000DFC0
00CFBF00CFB000000000BFA0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005F0010000000009FB000AFC08F00209F002000CFDF00D0EF00DFEF00E0F07FB060\
80B06F80B06F80BF6F80B06F00F0FF80BF6F8FBF6F80BF6F8FBF6F00000000FFFF
00F0F000FFFF00FFFF00FFFF00F0F000000000000000FFF000F0EF00F0EF00F0EF00F0EF000000\
70AF5F6FA0506FA05000000060904F5F8040507F3F008F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000607F006F805F001F000000000000009FB000AFC000AFC000B0CF00B0CF
80002000000000CFE000CFE000D0EF00D0EF9F002F90002000E0F000E0F000E0F000E0F0A0002F\
00000000F0FF00EFFF00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00F0F0
AF002F00EFE00000009F002000D0C000C0BF90001F00B0A000B0A080001F00A090009F8F009080\
008F7F008F7F008070008070007F6F007F6F006F6000706F600010600010005F5F
00606000000000000000000000000000000000000000000000000000000000000000000000607F\
000000007F900080A0009FB000A0BF00AFCF00B0CF00C0DF00C0DF709F60000000
00E0FF00E0FF00EFFF00EFFF00F0FF0000008FBF6F00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFFBF002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF8FBF6F
8FC06F80B0607FB06070AF5F00000000FFFF00FFFF00000000FFFF00FFFF8FBF6F8FBF6F8FC06F\
8FBF6F8FBF6F80B06F8FBF6F80B06000EFF000E0EF00EFF000E0EF00E0EF00DFE0
00DFEF00DFE000DFE000D0DF00D0E000CFDF00CFDFA00020A000209F002000C0D00000009F0020\
9000209000208F001F9000208F001F00AFBF00A0BF00000080001F800020009FB0
009FB0009FB00000000090AF0090B00090AF000000008FAF0090AF008FAF0090AF0000000090B0\
70001F009FB0608F5F00E0FFA0002000EFFF00EFF000FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000BF002000FFFF00FFF080BF6F80BF6F80BF6F\
80BF6F80BF6F7FBF6000FFF000F0EF0000007FB06000000070AF5F0000006FA050
00D0C000CFB08F001F00B09F6F001F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000606F0080900090AF00AFBF00B0CF00C0DF9F002000D0EF00D0EF00E0F000E0F080BF6F\
80B06F80BF6F80B06F00000000F0FF8FBF6F8FBF6F8FC06F8FBF6F00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFFBF002000FFF000F0F000FFF000F0EF00F0EF00F0EF000000\
70AF5F70A05F00D0C000000060904F5F8040000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000006F8050001000000060001F009FBF00A0BF00AFC000AFCF00BFD0
0000008F002000000000CFE000CFE000D0EF9000209F002000000000E0F000E0F0A0002FA00020\
00EFFF00EFF000F0FF00EFF0AF002F00EFF000F0FF00F0F000F0FF00EFF000EFEF
00EFE000EFE000DFD000D0CF00CFC000CFC000000000BFAF00B0A000000000000080001F009F8F\
009F8F008F7F000000008070008070007F6F007F6F00706F00706F006F60600010
5F000F006F60000000005F5F005F5F00000000000000000000000000000000000000000000607F\
5F001F007F90008FA00090B000A0C000AFC000BFD000BFD000000000CFE000DFF0
00DFF000EFFF00E0F000EFFF00EFFF00F0FFB0002000000000F0FF00FFFF00F0F000FFFF00FFFF\
C0002FBF0020C0002FBF002000FFFF00F0F000FFFF00FFFF00FFFF80BF6F8FBF6F
80B06080B06070AF5F00C0B0BF002000FFFF00F0F000FFFF00F0FF00FFFF0000008FBF6F8FBF6F\
8FBF6F80BF6F8FBF6F80B06F8FBF6F80B06000EFF000E0EF00EFF000E0EF00E0EF
00000000000000D0E000DFE000D0DF00D0E0000000A0002000CFDF00CFDF00C0D00000009F001F\
9F002000B0CF00BFCF90001F90002000000000AFC000AFBF00AFC000A0BF00A0BF
00000000A0BF7F001F009FB0009FB0009FB070001F7F001F0000000090B00090AF0090B00090B0\
009FBF009FB000AFC000000000D0EF00E0F0AF002FAF002000F0FF00F0F000FFFF
00FFFFB0002F00F0F000FFFF00000000000000F0F000FFFF00FFFF00FFFF00F0F080BF6F7FBF60\
80BF6F7FB0607FBF6F00F0EF00F0EF00EFE000F0E000EFDF00E0DF6FA0506FA050
6F9F5000CFBF80001F7F001F600010000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000707F6F0010009FB000A0BF00BFD000C0DF00000000CFE000DFEF00DFEF00EFF07FB06F\
80BF6F80B06F80BF6F00F0FF00FFFF80B06F8FBF6F8FBF6F00000000F0F000FFFF
00FFFF00FFFF00F0F000FFFF000000BF002FB0002000F0F000F0EF00F0EF00F0EF00F0EF000000\
70AF5F6FA05000D0C000C0B0608F4F507F3F009080000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000006F805F001F00809F6F001F009FB000A0C000AFC000B0CF00BFD0
00C0DF80002090002000CFE000D0EF9000209F002F9F00209F002F9F0020A0002FA00020A0002F\
00EFFF00F0FFA00020AF002F00F0FF00F0FF00F0FF00F0FF00000000F0F0000000
00EFEFA0002000E0DF00D0CF9F002000CFC000000000C0B000BFB08F001F0000008F001F00AF9F\
00A09000A0900090807F001F008F7F008F7F000000000000007F6F007F7000706F
00706F006F60007060006F60006F6000605F5F0010500010005F5F000000000000000000000000\
0000000080A0008FAF009FBF00A0C000B0CF00BFD06F90606F9F6000D0EF00DFF0
00E0FF00E0FF00EFFF00EFFF00F0FFBF0020BF002FBF0020C0002F000000000000C00020C0002F\
BF0020C0002FC00020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF8FC06F80BF6F
0000007FAF5F7FAF5F7F001F00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF8FBF6F8FBF6F\
8FBF6F8FBF6F8FBF6F8FBF6F80BF6F8FBF6F80B06F00F0F000EFF000EFF000E0EF
00E0F0AF0020B0002F00DFEF00DFEF00000000DFE000D0E000D0E000CFDF00CFDF00CFDFA00020\
9F002000C0D000BFD000BFD090002000000000B0CF00B0CF00AFC000B0CF00AFC0
00AFC080001F8F002000A0BF607F500000008000207F001F009FBF009FBF00A0BF009FBF009FBF\
00A0BF00000000AFC00000006F905F00DFEF70AF6F00EFFF00EFF000F0FF000000
B0002FB00020B0002FB00020B0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFF080BF6F\
7FBF6F80BF6F00F0EF00F0F000F0EF00F0EF00F0E000EFE070AF5F70AF5F6FA050
6FA0508F001F00000000A08F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006000107F001F009FB000B0CF00BFD000CFE000000000D0EF00DFEF00E0F000EFF080BF6F\
80B06F8FBF6F00000000FFFF00F0FF00FFFF8FBF6F8FC06F00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFFC00020C0002FBF0020BF002F00000000FFF000F0EF00F0EF00EFE0000000\
70A05F00000000CFBF0000005F8F405F804000907F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000708F00809F60001F6F001F009FBF00AFC000AFC000BFD0
00BFD08F00208F00209000200000009000209000209F002FA0DF40A0002F9F0020A0002F000000\
A0002FA00020AF002F00EFF000F0FF000000AF002F00F0F000F0F000EFEFAF0020
00E0E000000000DFD000DFD000D0CF00D0C000CFBF00CFBF00BFB000BFAF00B0A090001F8F0010\
00AF9F00A09000A090009F8F7F001F009080009080700010700010008070008070
6F0010007F7000706F00706F00706000706060000F006F6000605F404F2F0000000060604F0010\
007F9000809F0090B000000000AFC000B0CF6F905F6F905F709F6000000000E0F0
00E0F000EFFF00EFF000F0FF000000BF002FBF0020C0002F00000000FFFF00FFFF00FFFF00F0F0\
000000BF0020C0002FBF002000FFFF00FFFF00FFFF00F0F08FBF6F80BF6F00F0EF
7FAF5F7FAF5F00CFBF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF80BF6F\
8FBF6F80BF6F8FBF6F80B06F8FBF6F80B06F8FBF6F00EFF000F0F000EFF000EFF0
00E0EF00E0F0AF0020B00020AF002000DFEF00D0E000DFEF00D0E000D0E000CFDF00CFE09F0020\
00000000C0D000C0DF00BFD000C0D090001F00BFD000B0CF00B0CF00B0CF00B0CF
00AFC08F002080001F6080505F7F505F7F5000000000A0C000000000A0C000A0BF00A0C000A0BF\
00AFC000AFC000B0CF00B0CF00000000000000DFF070AF6000000000EFF000F0FF
00F0F000FFFF00FFFF00000000000000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
00000000F0EF00F0F000F0EF00F0EF00EFE000F0E000EFDF70AF5F6FA05F70A05F
00CFBF8F001F00B0A0000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006F001F0090AF00A0BF8F001F00C0DF00C0DF00D0E000D0E000E0EF00E0F000F0FF000000\
00000000F0FF00F0FF00F0F000FFFF00F0FF00FFFF00F0F0000000000000000000
00F0F000FFFFBF0020C0002FBF0020BF002FB00020BF002F00F0EF00F0EF00F0EF00EFE0000000\
00000000D0C000CFBF00C0AF5F8F40009F80000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000006F8000809F60001F009FB000A0BF70001F00000000B0CF
00C0DF00C0DF9000208F0020000000BFDF2F9F002FBFDF2FC0E02F303F0F9FEF60CFEF2FAF002F\
A00020000000000000AF002FAF002000F0FF00F0F000F0FF00F0F0000000AF0020
00000000E0DF00E0DF00000000DFD000D0CF00D0C000CFBF9F002000C0B000C0B090001F00BFAF\
00B0A000B0A000AF9F00AF9F00A09080001F7F001080001F7F00107F001F008F7F
7F001F70001000807F007F70008070007F70007F7000706F00000060001000000040502F50001F\
007F90008FAF0090B080002000AFCF00BFD060905F709F60709F607FA06F00E0F0
00EFFF00EFFF00EFFF00EFFFBF002FBF0020C0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF\
C0002000FFFF00FFFF00FFFF000000C0002FBF0020C0002F00FFFF00FFF000EFEF
80B06070AF5F00000000F0FF00FFFF00F0FFC0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
8FBF6F8FBF6F80BF6F8FBF6F80BF6F8FBF6F80B06F8FBF6F00F0FF00000080B06F
80B06F000000B0002FB00020B0002F00000000E0F000DFEF00DFEF00D0E000D0EF00D0E0A0002F\
00CFE000CFE000CFDF00CFDF00C0DF00C0DF00BFD000C0DF00BFD000BFD000B0CF
90002000B0CF00B0CF608050608F5F60805060805F00AFC000AFCF00AFC080002000AFC000AFCF\
00AFCF00B0CF80002000BFDF00BFDF00C0E000000000E0FFA00020AF002F000000
00F0FF00F0FF00FFFF000000B0002FB00020B0002FB00020B0002F00FFFF000000B00020B0002F\
B00020B0002F00000000F0F000F0EF00F0EF00EFE000F0E000000070AF5F000000
00CFBF00BFAF00B09F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000500010
6F001F70001F009FB080001F00BFD000C0DF00CFE000D0E000DFEF00E0EF00EFF000F0FF000000\
00F0FF00F0FF00F0FF00FFFF00F0FF00FFFF00FFFFC0002FC00020C0002FBF0020
C0002F00FFFF000000000000000000BF0020000000000000000000000000000000AF0020000000\
00DFCF00D0C000CFBF00C0B05F804000907F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000708F000000009FB0009FB000AFC0000000800020
8000200000008F0020900020900020BFDF2FBFDF20C0E02FC0E02FCFEF2FC0E02F1F200FCFEF2F\
AF002FA00020AF002FAF002000F0FF00F0F000F0FF00F0F0AF002FAF002000EFEF
00EFE000000000E0DF00E0D000DFCF00DFCF00D0C0A000209F001F00CFBF00C0B000000000BFAF\
00BFAF00B0A090001F8F001F00AFA000A09080001F8000100000007F00107F001F
7F00107F001F700010008F7F00807F00807F007F70000000000000007F7000706F6F00104F0010\
007F90008FA0009FBF7F001F0000000000006F905F6F9F6070A06000000000E0FF
00E0F000EFFF00EFF0BF002F000000000000BF002000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00000000F0F000FFFF00FFFF00FFFFBF0020C0002FBF002000FFFF00F0EF00EFE0
00E0DF00DFCF00EFF000000000000000F0FF00F0F0BF002FBF002000FFFF00F0F000FFFF00F0FF\
00FFFF80B06F8FBF6F80BF6F8FBF6F80B06F80BF6F00F0FF00F0FF80B06F80B06F
80B06F80B06F80AF6000EFF0000000B0002FAF002000E0F000DFEF00DFEF00D0EF00DFEF000000\
00D0EF00CFE000CFE000CFDF00CFE000C0DF00000000C0DF00C0DF00BFD0000000
00BFD000BFD000B0CF000000608F50608F5F00B0CF00B0CF00B0CF00B0CF00000080002080001F\
00B0D000B0CF8F002000000000C0DF00C0DF9000209F0020A0002FA00020A0002F
00000000F0FF00F0F000FFFF00FFFFB0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00F0F0\
B0002F00F0EF000000000000000000A0002000EFE000000000EFDF00DFCF00D0C0
00000000BFAF000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000002F2F0F
60803F7F001F7F001F8F002000BFD000CFE000CFE000D0EF00DFEF00E0F000000080BF6F80B06F\
00F0FF00F0F000F0FF00F0FF00FFFF00F0F000FFFFBF0020C0002FBF0020C0002F
00FFFF00FFFF00F0F000FFFFBF002000FFF000F0EF00F0F000F0EF000000AF0020AF0020000000\
00DFCF00CFBF00CFBF00BFA000A08F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000050001000809F0090B000A0BF00A0BF00B0CF7F0020
800020800020900020B0CF2FBFD02FBFD02FC0E02FC0DF2FC0E02FC0E02FCFEF30CFEF2F3F4010\
A00020AF002F00F0FF00F0FF00F0FF00FFFF00000000F0F000F0EFAF002F00EFEF
00F0EF000000AF0020A0002000E0D000DFD0A00020A0002000000000CFC000CFC000CFBF00CFBF\
00C0B000C0B000BFAF00000090001F90001F8F001F10100F000000AFAF1F000000
A0A01F2F2F0F80001F9F901F7F001F7F00107F001F700010008F8000807F008F7F00807F50001F\
007F900090AF009FBF8000200000006F90606F9060709F6070A06000E0FF00E0FF
00EFFF00EFFF00000000F0FF00F0FF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF000000C0002000000000F0F000F0F000EFE0
7FAF5F8F001F00F0FF00EFFF00F0FF00F0FF00F0FFB00020BF002FBF002000FFFF8FBF6F8FBF6F\
80BF6F8FBF6F00000000FFFF00F0FF00FFFF00F0FF00FFFF00F0FF8FBF6F80B06F
80BF6F80B06F8FBF6F00000000EFFF00EFF000EFFF00E0F000000000E0F000E0F000DFEFAF002F\
00D0EF00DFEF00D0EF00D0EF00CFE000D0E000CFE0A000209F00209F002000C0DF
00C0DF00BFDF00C0DF00BFD000C0DF00000000BFDF00BFD000BFD000B0D000BFDF0000008F0020\
00000000C0DF00BFDF00C0DF8F002090002F00CFE000000000DFF000E0FF00E0FF
00000000EFFF00F0FF00F0FF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
B0002000FFFF00F0F000F0F000F0EF00F0EFA00020A0002000000000DFCF00CFBF
00C0B000B0A0000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000005F001F20200F
7F001F7F001F8F002000B0CF00C0DF00CFE000D0EFA0002000000000E0F080BF6F80B06F80BF6F\
00F0FF00FFFF00F0FF00FFFF00F0FF00FFFFC00020C0002FBF002000000000FFFF
00FFFF00FFFF00FFFF00FFFF00000000F0F000FFF000F0EF00F0F000F0EFB0002FA00020000000\
00D0C000D0C000C0B000000000907F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004F001000708F000000009FB000A0BF00AFC000B0CF
0000008F00208F0020900020B0D020BFDF2FBFDF2FC0E02FC0E02FCFEF2FCFEF2FCFEF2FCFEF2F\
D0F02FA00020AF002FAF0020AF002FAF002000F0F000EFEF00F0F000EFEF00F0EF
00EFE000EFE0AF002000000000DFD000E0D000DFD000DFD000000000D0CF00D0C000000000CFBF\
00CFBF00C0B000C0B000BFAF00BFB090001F90001F8F001FB0B01FAFAF1FAFAF1F
AFA01FAFA01FA0A01FA0A01F80001080001F7F001080001F00908000908F008F80009080000000\
007F9F008FAF00A0BF7F001F8F0020608F5F6F9F606F9F6000000000DFF000E0FF
00000000000000EFF000F0FF00F0F000000000FFFF00FFFF00F0F000FFFF00FFFF00FFFF000000\
00FFFF00FFFF00FFFF00F0F000FFFFBF0020C0002F00000000F0F000EFE07FB060
70A05000B0A0A0002000EFFF00EFF000EFFF00EFFF00F0FF000000B0002FB0002000000080BF6F\
8FBF6F80BF6F8FBF6F80BF6F00FFFF00F0FF00FFFF00F0F000F0FF00F0FF000000
80B06F80B06F80B06F00000000EFF000EFFF00EFFF00EFFF00E0F000E0FFAF0020B0002FAF0020\
00000000DFEF00DFF000D0EF00D0EF000000A000209F0020A000209F002000CFE0
00C0DF00CFE000C0DF00C0E00000006F905F60905F6F906000BFDF00C0DF00BFDF6F906000BFDF\
00C0DF00BFDF00C0E000C0DF90002000000000D0EF00D0EF00000000E0F000E0FF
00E0F000EFFF00EFF000F0FF00F0F000FFFF00000000FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
00FFFF00F0F000F0F000F0EF00F0EF00EFE0A000209F002000E0D000D0C000CFBF
00BFAF000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000606F6F7F208F8F1F
90901F80001F80001F00BFD000C0DF00CFE000D0E0AF002000000000EFF080B06F80BF6F80B06F\
80BF6F00F0FF00FFFF00F0F000FFFF000000C0002FBF0020C0002F00FFFF00FFFF
00F0F000FFFF00FFFF00FFFFBF002000FFF000F0EF00F0F000F0EF00F0EF000000A0002000DFCF\
00D0C000C0B000BFAF000000008F70000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000006F7F007F90009FB000A0BF00AFC000B0CF
00C0D000C0DF30300FB0CF20BFD02FBFD02FC0DF2FC0DF2FCFE02FCFE02FCFEF2FCFEF2FD0F030\
303F0FAF002FAF002000000000F0FF00FFFF00F0F000F0F000F0F000F0F000F0EF
00F0EF00EFEF00EFE000E0DF00EFE000E0DF00E0DF0000007FAF5F7FA05F00DFCF00000000D0CF\
00CFC000CFC000CFBF00CFBF00C0BF00C0BF9F001F9F002030300FBFBF20B0B01F
BFB020B0B01FB0B020AFAF1FAFAF2090A02F1F1F0F000000000000009F90009F90000000006F8F\
007F9F0090AF7F001F8000200000006F90606F9F6070A06F00DFF0000000AF0020
00000000EFFF00F0FF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFFC0002000FFFFBF002000FFFF00F0EF80B0607FAF5F
9F002000AF9F00DFE000DFEF00EFF000EFFF00F0FF00EFFF00F0FF00000000000000F0FF00FFFF\
80BF6F8FBF6F80BF6F8FC06F00000000FFFF00F0FF00FFFF00F0FF00FFFF00F0FF
00F0FF000000BF002F00EFFF00F0FF00EFFF00F0FF00EFFF00EFFF00EFFF00EFFF00E0F0B0002F\
AF0020AF002F00DFF000E0F000DFF000DFF000D0EF00000000000000D0EF00CFE0
00D0EF00CFE000CFEF00CFE0709F606F90606F9F606F90606F9F6000C0E000CFE06F90606F9F60\
6F906000CFE000C0E000CFEF00000000000000D0EF00DFF000DFF0A0002F00E0FF
00EFFF00EFFFAF002F00000000FFFF00FFFF00FFFFAF002000FFFF00FFFF00FFFF00FFFF000000\
00FFFF00FFFF00F0F000F0F000EFEFA0002000E0DF00000000DFCF00D0C080001F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007070108F801F90901F
80002080001F00B0CF00BFD000CFE000D0E000DFEFAF0020B0002F80B06F80BF6F80B06F8FBF6F\
00000000FFFF00F0FF00FFFF000000C0002F00000000000000000000FFFF00FFFF
00FFFF00FFFF00FFFF00F0F000FFFF00F0F000FFF000F0EF00F0F000F0EFAF00209F002000DFCF\
00CFBF00CFB05F8040009F80000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000405F3F60001F009FB000A0BF00B0C0
0000008F00208F0020404F10B0CF20BFDF2FBFDF20C0E02FC0E02FCFEF2FCFE02FCFEF2FCFEF2F\
3F400FA0002000F0FF00F0F000FFFF00F0F000F0F000F0F000000000F0EF00F0EF
7FB06080BF6000000000EFE000E0DF00EFE000E0DF7FAF5F70A05F7FAF5F70A050AF0020A0001F\
00D0CF00CFC000D0C000CFBF0000009F001F9F001F9F001FC0C020BFBF1FBFC020
BFBF1FBFBF20B0B01FBFB020B0B01FB0B020AFAF1F8F001F8F001F8F001F8000105F6010500010\
0000000000007F001F80001F00BFD060905F6F9F60000000AF002FA00020AF002F
00EFF000F0FF00EFF000F0FF00F0FF00000000F0F000FFFF00FFFF00FFFF00F0F0C0002F00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00F0F000F0EF7FAF5F7FAF5F
00C0B080001F00C0D000CFDF00D0E000DFEF00DFEF00EFFF00EFFF00F0FF00000000F0FF00F0FF\
00F0FF00F0F000F0FF00F0FF00000000F0F000FFFF00F0FF00FFFF00F0F000F0FF
000000BF002FB00020B0002FB00020BF002F00000000EFFF000000000000AF0020B0002FAF0020\
00000000E0F000E0FF00E0F000E0F000DFF000DFF000DFF000DFF070A06070A060
709F6000D0EF00D0EF70A0606F9F60709F606F9F60709F606F9F606F9F6000CFE06F9F606F9F60\
6F9F6000CFE000CFEF00000000D0EF00D0EF00DFF000DFF000E0F000E0F0A0002F
A0002000EFFF00EFF0AF002FAF002000FFFF000000AF002FAF002000FFFF00F0F0AF002FAF0020\
00FFFF00F0F000F0F000EFEF00EFEF00E0DF00E0DF00DFD000DFCF80001F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000500010807F1F8F8F1F9F9F20
10100000000000B0D000C0E000CFE000D0EF00DFEFAF0020AF002080BF6F80B06F80BF6F80B06F\
00F0FF00F0F000FFFF00F0FF00000000F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F000000000F0F000F0F000F0EF00F0EF00F0EF00000000E0D09F002000D0C0\
00CFBF00BFAF507F3F008F70000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000004060406F001F00A0BF000000
8000208000208F00208F0020B0D02FB0D02FC0DF2FC0DF2FC0E02FC0E02FCFEF30CFEF2FAF002F\
A00020AF002FAF002000000000F0FF00FFFF00F0F000FFFFAF0020B0002F7FB060
80BF6F7FB06080BF6F00EFEF00EFEF00EFE000EFE07FAF5F7FAF607FAF5F7FAF60AF0020AF0020\
A0002000DFD0A00020A00020A0001FA00020A0001FA000209F001FCFCF20C0C020
C0CF20C0C020C0C020BFC020C0C020BFBF1FBFBF20BFB01FAFBF2F90001F90001F5F601050001F\
60001F0090B000A0BF80002000BFD000CFE000CFEF000000A00020AF002F00EFFF
00EFFF00EFFF00F0FF00F0FF00FFFFBF002000FFFF00FFFF00FFFF00FFFF00000000000000FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFF00000000000007FB060A00020
000000AFA01F80AF402F2F0F00C0D000CFDF00D0E000DFEFA0002F00E0F000000000EFFF00F0FF\
00EFFF00F0FF00F0FF00F0FFB00020BF002FB00020BF002FBF0020BF002F00FFFF
00FFFF00F0FF00FFFF00000000000000F0FF00F0FF00F0FF00F0FF00EFFF00F0FF000000B0002F\
AF0020B0002FAF002000EFFF00E0F000E0FF00E0F000E0F070A0607FAF6F70A06F
7FAF6F70A06000000070A06070A06F709F6070A06F709F6070A0606F9F6000000000D0EF70A060\
6F9F600000009F002000000000D0EF00DFF000DFF000E0FF00E0F000EFFF00E0FF
A0002FA00020000000AF0020AF002FAF0020B0002FAF0020AF002F00FFFFB0002FAF0020B0002F\
00FFFF00FFFF00000000000000EFE000EFE000E0DF00E0D080001F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F00107F701F8F801F9F9F20A09F20
90AF3F8F002000000000000000D0EF00D0E0AF002FAF0020B0002F80B06F8FBF6F80BF6F8FBF6F\
00F0FF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00F0F000FFF000F0EF00FFF000F0EF00F0EF00000000E0DFC0DFCF00D0C0\
00C0B05F8F4F009F80000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006F001F00A0BF
00AFC00000008000208F00200000001F1F0FBFD020C0DF2FC0DF20CFE02FCFEF2FCFEF2F4F500F\
AF002FA00020AF002FAF0020AF002FAF002000FFFF000000B0002000000080BF6F
7FB06080BF6000000000F0EF00EFE00000000000007FB0607FAF5F7FAF6000E0DF00E0DFAF001F\
00E0DF00DFD000DFD000DFCF00DFCF00D0CF000000A0001FB0D03FCFCF20CFCF20
C0CF1FCFCF20C0C020C0CF20C0C01FC0C020C0C01FC0C020302F009F001F90001F606010500010\
60001F0090AF00A0C080001F00BFD000C0DF00D0EF9F0020AF002F00000000EFFF
00EFF000F0FF00F0F000FFFF00F0F0C0002F00FFFF00FFFF00F0F000FFFFBF002000FFFF00F0F0\
00FFFF00FFFF00000000F0F000FFFF00FFFF00F0F0B00020000000000000A00020
00BFAFBFB020B0B01FBFB020B0B01FBFB020BFB01F00CFDF00CFDF00DFE09F0020A0002F00EFF0\
00EFFF00EFFF00F0FF00EFF0B0002FB00020000000B0002000F0FF00F0FF00FFFF
00F0F000FFFF00F0FF00FFFF00000000F0FF00F0FF00F0FF00EFF000F0FF00EFF000EFFF00EFF0\
B0002F00EFF000EFFF00000000000000E0F000E0F000DFF000E0F00000007FAF6F
70A06070AF6F00DFEF00DFF070A06070A06070A06070A06F6F9F6070A06000000000DFEF00D0EF\
00DFEF0000009F002F00000000DFF000DFF000E0F000E0F00000009F0020A0002F
00EFF0A0002FA00020D0F02FAF00203F400FD0F02FAF002F9FF05FAF002FAF0020AF002F00F0F0\
000000A0002000000000EFE000EFE000E0DF00E0DF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000060605F00107F7F1090901F9F9F1FAFAF20
8F001F9000209F0020A00020000000000000000000B0002F00EFF080BF6F80B06F8FBF6F000000\
000000BF0020C0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00000000FFF000F0EF00F0F000F0EF00F0EF00000000E0DF00DFCF00000000C0B0\
00C0B0507F3F600010000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006F001F009FB0
00AFC000B0C000C0D000C0DF9000209000209F0020BFDF2FA0002FC0E02F5F6010A00020AF002F\
00000000F0FF00F0F000FFFF00F0FF00FFFF00F0F000FFFFB0002000FFF000F0F0
00000000F0F080BF6F7FB06080BF6F7FB06080B0607FB06000F0E000EFE000EFE000E0DF000000\
00E0DF00E0DF00E0D000E0DF00DFD000E0D000DFCF000000AF0020AF00201F1F00
D0D02FCFD020CFD020CFCF20CFD020CFCF20CFCF20C0CF20CFCF2030300FA000205F60105F001F\
00809F009FB000A0BF80002000BFD000CFE000D0EFA0002FAF0020AF002F00EFF0
00F0FF00F0FF00F0FF00F0FF000000C0002000FFFF00FFFF00FFFFC00020C0002F00FFFF00FFFF\
00FFFFC0002F00FFFF00FFFF00FFFF00FFFF000000BF002F00EFE000EFDF00D0C0
00C0AFC0BF20C0C020C0BF20CFC020C0C020CFC020CFC020CFCF2F8F0020000000000000A0002F\
00000000EFFF00EFFFAF002F00000000F0FF00EFFF00F0FF00F0FF00000000F0FF
00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00F0FF00F0FF00F0FF00EFFF00F0FF\
00EFFF00F0FF00EFFF00EFFF00EFF000EFFF00E0F0AF002FAF0020AF002FAF0020
00E0FF00E0F000E0F000DFF000E0F00000007FAF6F70A0607FAF6F00000000DFF000DFEF00E0F0\
00DFF000DFF09F002000E0FF00000000E0FF00E0F000EFFF00E0FF00EFFFA00020
AF002FA00020D0F02FD0F02FD0FF30D0FF2FDFFF2FD0FF2FDFFF30D0FF2FD0FF2F000000AF002F\
AF002000F0F000EFEF00EFEF00E0DF00000000C0B0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000060600000007F7F1F908F209F9020A0A0208F001F
9000209F0020A0002000D0E000DFEF00DFEF00E0F0B0002000F0FF00F0FF8FBF6F00F0FFC0002F\
BF0020C0002F00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFFBF002F00F0F000FFF000F0EF00F0EF00F0EF00F0E000E0D000DFCF90001F00CFBF\
5F8F40000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
009FB000AFC000B0CF00C0D08F00209000208F002000DFEF9000209F002F000000A0002FA00020\
AF002F00EFF000F0FF00F0F000F0FF00F0F000FFFF00F0F000FFFF00F0F000F0F0
00F0F000FFF07FB06080BF6F7FB06080BF6F7FB06080B06000EFE000F0EF00EFE000EFE000EFDF\
00EFE000E0DF00EFDF00E0DF00E0DF00E0D000E0DFAF0020AF0020AF001FAF0020
D0D020D0DF20CFD020D0D020CFD020D0D020CFCF20000000A0001FA00020A0001F5F6010500010\
0080A00090AF00A0C000AFCF00BFDF00C0DF00D0EFA0002000E0FFAF002000EFFF
00EFF000F0FF00F0F000FFFF000000C0002F00F0F000FFFF000000C0002FBF002000FFFF00FFFF\
C0002F00F0F000FFFF00FFFF00FFFF00F0F0BF0020B0002000EFE000E0DF00D0C0
00BFA07F7F10CFC020CFCF20CFCF20D0CF20CFCF20D0D020CFCF20D0D020CFCF2000C0D000C0D0\
00D0E000D0E09F00209F002000000000EFF000EFFF00EFFF00F0FF00EFF000F0FF
00000000F0FF00EFF000F0FF00F0FF00F0FFB0002000000000F0FF00F0FF00F0F000F0FF00F0F0\
B0002F00EFF000F0FF00EFF000EFFF00EFF000EFFF00EFF000EFFFAF0020AF002F
AF002000E0F000E0F000E0F000E0F000E0F000DFF000E0F070AF6000E0F000DFF000E0F000DFF0\
00E0F000DFF000E0F000E0F000EFFF00E0F000EFFFA00020A0002FA00020A0002F
303F0FD0F02FCFF02FD0F02FD0F02FD0FF2FD0F02FD0FF2FD0F02FD0FF2FCFF02FD0FF2FA00020\
0000000000009F002000DFDF00DFD000BFB0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000005F605F00107070108F801F90901FA0A020A0A020900020
90002000000000CFE000D0EF00DFEF00E0F000000000000000F0F000FFFF00F0F000FFFFBF0020\
BF002F00F0F000FFFF00FFFF00FFFF00F0F000FFFFBF002000FFFF00F0F000FFFF
00F0F000000000000000F0F000F0EF00F0EF00EFEF00000000E0DF00E0D000D0C090001F00C0AF\
5080405F0010000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0090AF00A0BF00B0CF00BFCF0000008F002000000000D0E00000009F0020A0002F00EFF0000000\
A00020AF002F00F0F000F0FF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFF0
00FFFF80BF6F80BF6F80BF6F80BF6F80BF6080BF6F80BF6080BF6F00F0EF00F0EF00EFE000F0EF\
B0002000EFE000EFE000EFE000EFDF00EFE0000000B00020AF0020B00020000000
B00020AF0020B00020A0DF50BFDF3FAF0020505010D0D020AF0020AF001FAF0020AF00205F001F\
00809F009FBF00A0C000B0CF00BFD000CFE000D0EFA0002F00E0F000000000EFFF
00F0FF00F0FF00FFFF00FFFF000000BF0020C0002F00FFFFC0002FBF002000000000FFFFC0002F\
BF002000FFFF00FFFF00FFFF00FFFF00000000F0EF00000000EFE000EFDF00CFBF
00BFA07F8010D0D02FD0D020D0D02FD0D020DFDF2FD0D020DFDF2FD0DF208FDF70B00020B00020\
B0002000CFD000CFDF00D0E000000000E0F000000000EFFF00EFFF00F0FF00EFFF
00F0FF000000B0002F00000000F0FF00F0FF00F0FFB00020B0002F00F0FF00F0FF00F0FF00F0FF\
B0002000000000F0FF00F0FF00EFFF00F0FF00EFFF00F0FF00000000F0FF00EFFF
00EFFFAF002000EFFF00EFF000EFFF00E0F000EFFF00E0F07FB06F70AF6F7FB06F00000000EFFF\
00E0F000000000000000EFFF00EFFF00EFFF00EFFF00EFFF00EFFFAF002FA00020
AF002F303F0FD0FF30D0F02FD0FF2FD0F02FD0FF30D0F02FD0FF2FCFF02FD0FF30A00020AF002F\
9F002000EFE000DFDF00DFD0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000060605F00106000108F8F1F90901FA09F20A0A020B0B02F900020
00CFE000CFE000DFEF00DFEF00E0F000E0F0BF002F00F0F000FFFF00F0FF00FFFF000000C0002F\
BF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002F00FFFF00FFFF00FFFF
00FFFFBF002000FFF000F0F000F0F000F0EF00000000EFE000E0DF00DFCF0000008F001F608F4F\
009F80000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000090AF90A0B000B0CF00BFCF8F002000CFDF00D0E000D0E09F00209F002000E0F000E0F0\
00EFF000EFF0AF002F00EFF000F0FF00F0F000FFFFAF002000FFFF00F0F000FFFF
00F0F080BF6F7FBF6080BF6F7FB06080BF6F80BF6080BF6F80B06000F0F000F0EF00F0EF00EFE0\
000000B0002000F0EF00EFE000EFE000000000000000E0DF00EFE000E0DF00EFE0
000000B00020AF0020B00020000000B00020AF0020B0002000DFD0000000AF0020AF002000708F\
008FA00090B000AFC000000000BFDF00C0E000D0EF00000000E0F000E0F000EFFF
00EFF000F0FF00F0F000FFFFBF0020C0002FBF002000FFFFBF0020C0002F000000000000BF0020\
00000000FFFF00FFFF00F0F000FFFF00F0F000F0F000EFE000EFE000000000CFBF
8F001F808F10D0DF20DFDF2FD0DF20DFDF2FDFDF20DFE02F30300F1F1F0F00000000000000E0E0\
00EFEF00E0EF00EFEF00EFEF00C0D000000000D0E000D0E000DFEF00000000EFFF
00EFF000EFFF00EFFF000000AF0020AF002F00EFFF00F0FFAF0020B0002FAF002000F0FF00EFF0\
B0002FAF002000000000EFF000F0FF00EFFF00F0FF00000000000000EFF000F0FF
00EFF000EFFF00EFF000000000000000EFFF00EFF000000070AF6F7FB06F70AF6F7FB06F000000\
AF002FA0002000000000EFF000EFFF00EFFF00EFFF00EFF000EFFFA00020AF002F
A000201F200FD0F02FD0FF2FCFF02FD0F02FCFF02FD0F02FCFF02FCFF02F70F07F00F0F000E0E0\
00E0E000DFD000D0D0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005000105F001080801F908F1F9F9F20A09F1FAFAF202F2F0F00C0DF
00CFE000D0E000D0E000E0EF00E0EF00EFF000EFF000FFFF00F0F000FFFF00F0FFBF002F000000\
00000000F0FF00FFFF00F0F000FFFF00FFFF00FFFFBF002000000000FFFF00FFFF
000000BF002F00F0F000F0F000F0EF00000000000000EFE000DFD000D0CF8F001F8F001F507F3F\
600010000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000090A0009FB000B0CF00C0D000CFDF00CFDF00D0E000D0E09F002F00E0EF00EFF0\
00EFF000EFF000EFF000F0FF000000AF002FAF0020B0002F00FFFF00FFFF00FFF0
00FFFF00000080BF6F80BF6F80BF6F80BF6F80BF6F80BF6F8FBF6F00FFF000FFF000F0EF00F0F0\
000000BF002FB00020BF002F00000000F0EF00EFE000F0EF00EFE000EFE000EFE0
00EFE0B00020BF002000000000EFE0B00020B00020B0002000000000E0DF00E0DF00E0DF5F001F\
008FA0009FBF00A0C000000000BFDF00CFE000D0EF00E0FF00E0F000EFFF00EFF0
BF002F00F0FF00FFFF00FFFFC0002FC00020C0002F000000C0002FC00020000000BF0020C0002F\
00FFFF00FFFF00FFFF00000000FFFF00FFF000F0EF00F0EF00EFE000E0D000CFB0
90001F8F9010E0E02FDFE0203F3F10AFE05FE0EF2F8FE070BF002FBF0020BF002FBF0020000000\
00EFF000F0F000F0F000FFFF00F0F000FFFF00F0FF00CFDF00CFDF00D0E000D0E0
00DFEF000000A0002F00EFF000EFFFAF0020AF002FAF0020B0002FAF0020AF002FAF0020000000\
AF0020B0002FAF002000000000F0FF000000AF0020B0002F00000000F0FF00EFFF
00F0FF00EFFF00F0FF00EFFF00F0FFAF0020AF002FAF00200000007FB06F7FB06F7FB06F00EFFF\
00EFFFAF002F00000000F0FF00EFFF00F0FF00EFFF000000A00020AF002FAF0020
AF002F000000D0FF2F303F0FD0FF30CFF02FAF002FCFF02FAF002FA00020A0002F9F002000E0DF\
00D0CF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F00108F801F908F1F9F9F20A09F20AFAF202F2F0F9F002000CFE0
00D0EF00D0EF00DFEF00E0EF00EFF000000000F0FF00F0FF00FFFF00F0FF000000BF002000FFFF\
BF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002F00FFFF00FFFF00FFFF
C0002FBF002000FFF000F0EF00F0F0AF002000F0EF00E0DF00DFD000D0C08F002080001F00A08F\
5F0010000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000008FA0009FB07F002000BFCF00C0D000CFDF00D0E000D0E000000000E0EF\
00EFF000E0F000EFF000EFF000F0F0000000AF002FAF002000FFFF00F0F080BF6F
7FBF6F00000000000080BF6F80BF6080BF6F80BF6080BF6F00000000000080BF6080BF6F80BF60\
00FFF0B00020BF002000000000F0EF00EFEF00F0EF00EFEF00F0EF00EFE000EFEF
000000BF002000000000EFE000EFE000EFE0000000B00020B0002000EFE000E0DF00EFE0500010\
008FA0009FB000AFC080001F00C0DF00CFE000000000DFF000E0F000E0F000EFFF
B00020BF002F00000000FFFFBF0020C0002FBF0020B0FF60BF0020C0002FBF0020C0002FBF0020\
00FFFF000000C0002FBF002000FFFF00F0EF00F0EF00EFE000EFE000DFCF000000
8F001F909F1F70E080BF0020B00020BF0020BF002000F0F0000000BF002FBF0020BF002FBF0020\
00000000F0F000FFFF00F0F000FFFF00F0F000FFFF00F0FF00FFFF00000000FFFF
00CFD000D0DF00D0DF00DFE000DFEFA0002FA00020303F0FA00020AF002FCFEF2FD0EF2FD0EF2F\
D0EF2FD0EF2FA0EF60D0EF2FD0F02FAF0020AF002FAF0020AF002F00EFF000EFFF
00EFFF00F0FF00EFF000EFFF00EFFF000000A00020000000A00020AF002F00EFF000EFFF00EFFF\
00EFFF00EFF0AF002F00EFFF00F0FF00EFF000F0FF00EFFF00F0FF00F0F000F0FF
A00020AF002FA00020000000A00020A0002FA0002000000000EFEF00E0EF00D0D08F0020000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00606F006F7F6F001F8F8F1F9F9F20A09F1FAFAF20AFAF209000209F0020000000
00D0E000DFEF00DFEF00E0F000EFF000F0FF00F0F000FFFF00F0FF00FFFFBF002000FFFF00F0FF\
00000000F0F000FFFF00FFFF00FFFF00F0F0000000BF002000000000F0F0000000
BF002000000000F0EF00F0EFAF002000000000EFDF00E0D000D0CF00CFC080001F00B09F600010\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F6F4F4F704F7F002000BFCF00C0D000CFDF00D0E000D0E0000000\
00E0EF00EFF000EFF000EFF000EFF0AF002F00000000F0FF00000080BF6F80BF6F
80BF6F80BF6F00000000FFFF00FFFF00FFF000FFFF00FFF000FFFF80BF6F8FC06F80BF6F8FC06F\
80BF6F00FFFF000000BF002FBF002000FFF000F0EF00F0F000F0EF00F0F000F0EF
00000000000000F0EF00F0EF00F0EF00EFEF00F0EF00EFEFBF002F00000000F0EF00EFEF1F1F0F\
00000000A0BF00AFC08F002000BFDF00D0EF00D0EF00E0F000E0F000EFFF00EFFF
000000BF0020C0002FC00020000000BF0020C0002FEFFF2FC0002FEFFF2F1F1F0FC00020000000\
BF0020C0002FC0002000FFFF00F0F000FFF000F0EF00000000EFDF00E0D000CFBF
00C0AF00000000706FBF0020C0002FBF0020C0002F00FFF000FFFF00F0F000FFFF00FFFF000000\
BF002000000000FFFF00FFFF00F0FF00FFFF00FFFF000000BF002000FFFF00FFFF
00FFFF00F0FF00FFFF00F0FF9000209000209F002F9F00207FDF7090E05FCFE02FCFE02FD0EF2F\
CFEF2FD0F030D0EF2FD0F02FD0EF2FD0F030BFEF40BFF04F4F5F10AF002FAF0020
AF002FAF0020AF002FAF002000F0FF00EFFF00F0FF00EFFF00F0FF00EFFFAF002F00EFFF00F0FF\
00EFFF00F0FF00EFFF00000000EFFF00F0FF00F0FF00F0FF00F0F0000000A00020
00F0F000EFF000F0F000EFF0A0002F9F002000EFF000DFE000D0D000CFC0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000606000606F\
006F7F6F001F908F1F9F9F20A09F20AFAF20B0B0208FBF509F0020A0002FA00020
00DFEF00DFEF00EFF000EFF000000000F0FF00FFFF00F0FF00FFFF00000000FFFF00FFFF00FFFF\
00FFFF00000000FFFF00FFFF00FFFF00FFFFC00020C0002FBF002000FFFFBF0020
BF002F00F0F000FFF0B00020B0002F00EFE000EFE000DFD000D0C08F001F8F002000A08F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004F704F5F80507F001F80002000C0D000D0DF00D0DF\
9F002000DFEF00E0F000E0EF00EFF000EFF000F0F000EFF000F0F07FB06F80BF6F
7FBF6F80BF6F7FBF6F00FFFF00F0F000FFFF00FFF000FFFF00F0F080BF6F80BF6F80BF6F80BF60\
80BF6F00000000FFFF00F0F000FFF000000000F0F000F0EF00F0F000F0EF00F0F0
00000000F0F000F0EF00F0F000EFEF00F0EF00EFEF00F0EF00EFEF00000000EFEF00F0EF5F6010\
60001F009FB000AFC080001F00C0DF00CFE000DFF000DFF000E0F000E0F000EFFF
00EFF0BF002FBF0020C0002FEFF02F20200FB0FF5FF0FF2FEFF02FF0FF2F000000B0FF60BF0020\
C0002FBF002000FFFF00F0F000FFF000F0EF000000000000B00020A0001F00D0C0
00BFAF0000006F0010000000BF0020C0002F00F0F000FFFF00F0F000FFFF00F0F000FFFF00F0F0\
00FFFF00F0F0C0002F00000000FFFF000000C0002FBF002000FFFF00F0FF00FFFF
00F0F000FFFF00F0FF00000080BF6F8FBF6F8FBF6F8FBF6F00CFD09000209000201F1F0FBFD020\
C0DF2FC0DF2FCFE02FC0E02FCFEF2FCFEF2FD0EF2FCFEF2FD0EF2FA00020AF002F
A00020AF002FA0002000F0FF00EFF000EFFF00EFFF00F0FF00EFF000F0FF00EFFF00F0FF000000\
00F0FF00EFFF00F0FFA00020A0002FA00020A0002F00EFF000EFF000EFF000EFF0
00E0EF00EFF000E0EF00EFEF00E0E000DFDF00C0BF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000005F6000606F006F70\
6F001F508F4F7F90409F9F1FAFAF2010100010100F90001FA00020A00020AF0020
00DFEF00E0EF00E0EF00F0F000000000FFFF00F0FF00FFFF00000000FFFF00F0FF00FFFF00F0F0\
00000000000000000000F0F000FFFF00FFFFC0002FBF0020C0002F000000BF002F
B0002000F0F0B00020B0002000000000EFE000E0D000DFCF00000000000000AF9F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F805000BFCF00C0D000C0D000D0DF\
9000209F00209F0020A0002F00E0F000EFF000EFF000F0F000EFF000F0FF00F0F0
00000000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFF000FFFF00FFFF00FFFF80BF6F000000\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0F0C0002F00000000FFFF00F0F0
C0002F00F0F000FFF000F0F000F0F000F0EF00F0F000F0EF00F0F000000000F0F000F0EF00F0F0\
5F001F009FB000A0C08F002000C0DF9F002000DFF000E0F000E0F000EFFF00EFF0
00F0FFBF0020C0002FBF00205F6010EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FC0002F\
C0002000FFFF00FFFF00FFFF00F0F000F0F0B00020BF002FB0002000E0D000D0C0
00CFB000AF9000907F00F0F0C0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF000000C0002FC00020C0002FBF002000FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF0000008FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F000000\
8F00209000204F5010BFDF2FBFD02FC0DF2FBFDF2FC0E02FC0E02FCFE02FC0E02F
3F3F10A00020A0002FA00020AF002FA00020A0002F000000000000000000000000A00020A0002F\
A00020A0002FA0002000EFF000E0EF00EFF000E0EF00EFF000E0EF000000000000
00000000D0D0609050000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000006F70006F7F6F001F\
70001F80001FA09F1F8F0020B0AF209000209F002000CFE0000000AF002FAF0020
00000000E0F000EFF000000000FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF\
000000C0002FBF002000FFFF00FFFF000000BF0020C0002FC00020C0002FBF0020
BF002FB00020B0002FAF002000F0EF00EFDF00E0D000D0C000CFBF80001F00907F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000008FA000000000BFCF00C0D0\
00CFDF00CFDF00000000000000E0F000E0EF00EFF000EFF000EFF000EFF0AF002F
AF0020B0002FAF0020B0002F00F0F000FFFF00F0F000FFFF00FFF0BF002FB00020BF002FB00020\
00000000F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF000000BF0020BF002F
BF002000000000F0F000F0F000F0F000F0F000F0F000F0F000F0F0000000BF002000F0F0000000\
00708F0090AF00000080001F00C0DF90002000DFF000DFF000E0F000E0F000EFFF
00F0F0BF002FBF0020C0002F5F5F10F0FF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02F000000BF0020\
C0002F00F0F000FFFF00F0F000FFF000F0EFBF0020B0002000000000E0D000D0C0
00C0B000B09F7000106F0010BF002000FFFF00F0F000FFFF00F0F000FFFF00F0F000FFFF00F0F0\
00FFFFBF0020C0002FBF002000000000F0F08FBF6F8FBF6F8FBF6F80BF6F8FBF6F
8FBF6F0000000000008FBF6F8FBF6F8FBF6F80BF6F8FBF6F8FBF6F8FBF6F80BF6F00FFFF00FFFF\
00FFFF00F0F000FFFF00FFFFC0002FBF002000F0F08F0020B0D02F900020900020
0000000000009000209F00209000209F002000DFE000E0EF00DFE000E0EF00DFE000E0EF00DFE0\
00E0EF00DFE000E0EF00000000E0EF00DFE000DFE0609F5F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000006F705F00106F001F008090\
7F001F2F20008F001F00000090002090001F00CFE000CFE000D0EF00D0E000DFEF
00000000EFF000EFF0BF002F00F0FF00FFFFBF002000000000F0FF00FFFF00F0F000FFFF00FFFF\
000000BF0020C0002FBF002000FFFFBF0020C0002FBF00207FFF90BF0020BF0020
B00020B0002FAF002000F0EF00EFE000E0DF00000000CFBF00BFAF009F8F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000090A000B0CF\
00BFCF00CFDF8F0020900020900020A0002FA0002000000000E0EF00EFF000EFF0
00F0F000EFF0B0002F00F0F000FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFFBF0020000000\
00FFFF00000000000000000000000000FFFF00FFFF00000000FFFF00FFFF000000
C0002FBF0020000000000000000000000000C0002FBF0020C0002FBF0020C0002FBF0020C0002F\
5000100090AF00A0BF8000208F00209F002F00DFF000E0F000E0F000EFFF00EFF0
00F0FF000000C0002FC00020F0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30606F10404010\
00FFFF00FFFF00FFFF00FFFF00F0F0000000000000BF002F00EFE000EFDF00D0C0
00CFBF00BFA080001F008F7060001000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000\
00FFFF00FFFF00FFFF00FFFF00FFFF8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F
00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF8FBF6F8FC06F8FBF6F8FC06F\
8FBF6F00FFFFBF0020BF002FB00020BF002F00F0F000F0F000EFEF00F0F000EFEF
00EFEF00E0E09F002000D0D08F0020000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000006060500010006F70600010008090008F9F7F001F\
80001F8F00208F001F0000009F0020A0002000D0E000D0EF00D0EF00E0F000E0EF
00EFF0B00020BF002FBF002000FFFFBF0020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
BF0020C0002FC00020C0002F000000C0002FC000201F1F0F3F3F0FBF002FC0F040
B0002FAF002000000000EFE00000009F002000D0C000C0B000AF9F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
009FAF7F001F00C0D000C0D000CFDF00D0DF00000000DFE000E0EF00E0EF00EFF0
00E0EF00EFF000EFF000000000EFF000F0F000F0F000FFFF00F0F000FFFF00FFFF00000000F0F0\
00FFFF00FFFF00FFFF00F0F000FFFF00FFFF000000BF0020BF002FBF0020BF002F
00000000FFFF00FFFF00FFFF00F0F000FFFF00F0F000FFFF00F0F000FFFF00F0F000000000F0F0\
50001000809F00A0BF0000008F002090002000DFF000DFEF00E0F000E0F000EFFF
00EFF00000003F3F0FF0FF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FEFFF2FF0FF2F3F3F0F\
C0002FBF0020C0002FBF002000FFF000F0F000000000000000F0E000E0D000DFCF
00000000C0AF00AF90009F80007F6F006F60BF0020C0002FBF002000FFFF00F0F000FFFF00F0F0\
00FFFF00FFFF00FFFF00F0F000FFFF8FBF6F8FBF6F80BF6F8FBF6F00000000FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF8FBF6F80BF6F8FBF6F8FBF6F8FBF6F80B060\
00F0F0000000B0002F00000000EFF000EFEF00EFEF00E0EF00EFEF00E0E09F0020
00D0D000CFCF00BFBF00BFBF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000060605000107F6F7F007080008090008F9F7F001F7F001F\
00A0BF8F001F900020000000A00020A0002000D0EF00D0E000DFEF00E0EF00EFF0
00EFF0BF002FBF0020000000BF0020BF002F00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F0\
C0002FBF0020C0002FBF0020F0FF2FBF0020C0002FEFF02FBF0020E0F02FB0002F
AF0020AF002000EFE0A000209F001F00D0C000C0B000000000806F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000A0B000BFCF00C0CF00CFD000CFDF00D0E000000000E0EF00E0EF
00EFF000EFEF00EFF000EFF0AF002F00EFF000F0F000F0F000FFFF00F0FF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFBF0020C0002F000000
C0002FBF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF8FC06F8FBF6F8FC06F\
60601000809F009FB000AFCF8000209000209F0020000000AF0020B0002FB00020
BF002FBF0020404010EFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FC0002F\
C00020C0002FBF002000000000FFF000FFF000F0EF00FFEF00F0E000000000DFCF
00D0C000C0B000BFA0009F8F008F7F007F6F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF000000C0002FBF002000000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000080BF6F8FBF6F80BF6F8FBF6F00F0F000F0F0\
00EFF000F0F000000000EFF000EFEF00EFEF00E0E000DFE000D0D000CFCF00BFC0
00BFBF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000606F50001000707F007080008090008F9F80001F00000000AFBF\
8F001F90002000BFD0000000A0002000D0EF00D0EF00E0EF00E0EF00EFF000EFF0
00F0FFBF0020C0002FBF0020C0002FBF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
C00020C0002FBF002020200FEFFF2F404010EFFF2FEFFF2FE0F02F5F5F10B00020
000000AF0020AF00209F002000DFCF00CFBF00C0B0600010000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F001F00B0C000C0CF00C0D000CFDF00D0DF9F0020
00DFEF00E0EF00E0EF00EFF0000000AF002F00000000EFF000EFF000F0F000F0F0B0002FB00020\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F000000000000000FFFF00F0F000FFFF00FFFF00FFFF80BF6F8FBF6F80BF6F\
8FBF6F5F0010009FB000A0C080002000C0DF00000000DFEF00E0F0AF0020B0002F
B00020BF002FBF0020B0FF60EFFF2FF0FF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FBF0020\
C0002FBF002000000000F0F000FFFF00F0F000F0F000F0EF00F0EF00000000DFD0
00000000CFBF00BFAF00AF9000907F700010007060C0002FBF0020C0002FBF0020000000000000\
000000000000C0002FBF0020C0002FBF0020C0002FBF0020C0002F00000000FFFF
00FFFF00FFFF000000BF002FBF0020BF002F00000000F0F000EFF000F0F000EFEF00EFF000EFEF\
00EFEF00000000000000E0E000000000CFCF00CFCF00BFBF00BFB0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000030402F0000005F001000707F007080008090008F9F80001F009FB000AFBF00AFC0\
90002000000000C0DF000000AF002000D0E000DFEF00DFEF00EFF000EFEF00F0F0
BF002000000000F0F0C0002FBF0020C0002F00F0F000FFFF00F0FF00FFFF00F0F000FFFF00FFFF\
C0002FBF0020C0002FEFFF2FF0FF2FEFF02FEFFF2FE0F02FE0F02F303F0FDFF02F
AF0020AF00209F001F00DFCF00CFBF00C0B0009F8F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000080001F00C0D000CFD0
00D0DF9000209F002F00E0EF00E0EFA00020AF002FA0002000000000EFF0B0002FAF0020B0002F\
00000000FFFF00F0F000FFFF00FFFF00FFFFBF002000000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002FBF0020C0002FBF00200000008FBF6F8FC06F\
80BF6F5F001F0090AF00A0C000000000C0DF00000000DFF000DFF000EFFF000000
BF002FBF0020C0002FB0FF60F0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2F\
00000000FFFF00FFFF00FFFF00F0F000FFF000F0F000FFF000F0EF00F0E0AF0020
00000000CFBF00CFB000B09F00A08F008F7F70001000706000FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFFC0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00F0F0000000B00020B0002FB00020B0002F00EFEF00EFF000EFEF00EFEF00EFEF00EFEF\
9F00209F002090002000C0CF00BFBF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F4F2F006060404F305F0010000000007F8F008F9F0090A000000000A0B000AFC000AFC09F0020\
9F002000CFDF00CFE0000000AF002000E0EF00E0EF00EFF000EFF000F0FFBF0020
00FFFF00FFFFC0002FBF002000000000000000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
BF0020C0002F000000F0FF2FEFFF2FEFFF30E0F02FE0F02FDFF02FDFF02FAF0020
AF0020A0002000DFD000CFC000C0BF7F001F00806F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000007F8F00BFC0
80001F8F00208F00209000209000209F00209F0020A00020A00020A00020A00020AF002FAF0020\
AF0020AF002000F0F0000000B0002FB00020B0002F00000000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000BF002FBF0020BF002FBF002000FFFF00F0F0\
00FFFF000000008FAF009FBF80002000BFD090002000D0EF00E0F000E0F000EFFF
00EFF0BF002FBF0020F0FF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FEFFF2FF0FF2FBF0020\
C0002F00FFFF00FFFF00F0F000FFF000F0F000FFF000F0EF00F0EF00EFE0B00020
A0001F00D0C000C0B000BFAF00A090009F8F70001000807060000F00FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFFFFFFFF00F0F000FFFF00F0F000FFFF00F0F000F0F000F0F000F0F0
00EFEF00F0F000EFEF00EFF000EFEF00EFEF00E0EF000000A00020A000209F002000D0D000CFCF\
00C0C000B0BF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F4F2F
3F4F2F006070006F7F6F001F007F8F008F9F0090A0009FB000A0B000AFBF00AFC0000000000000\
00C0DF00CFE000D0EF00D0E000000000DFEF00E0F000E0EF00F0F000000000FFFF
00F0F000FFFFBF002000000000F0F000FFFF00FFFFC0002F000000000000000000C0002FBF0020\
F0FF2FCFFF40F0FF2FEFF02FEFFF2FE0F02FE0F02FDFF020DFF02F000000AF0020
A0002000DFD000D0C000CFBF00BFAF008F7F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000009FAF00BFCF80001F8F00209FCF3FBFD02FBFDF20C0E02F4F5010CFE02FCFE02F3F4010\
506010D0EF2F9FEF5FB0002FAF0020B0002FAF002000F0FF00F0F000FFFF00F0F0
00FFFF00FFFF00FFFF00FFFF00000000000000FFFF00FFFF00FFFF00FFFF000000BF002000FFFF\
00FFFF00FFFF00809F00A0BF00000000C0DF00000000DFF000DFF000E0F000EFF0
00F0FF00F0FF000000CFFF40F0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2F000000C0002F\
C00020C0002F00FFFF00FFFF00FFF000FFF000000000FFF000F0EF00FFEFB00020
B0002000000000D0C000C0B000B0A000A0907F001F00000000000000000000FFFF00FFFF00FFFF\
BF0020BF002F00F0F000FFFF00F0F000F0F000EFF000F0F0B0002000000000EFEF
00EFF000EFEF00EFEF00E0EF00EFEF00E0E09F002000000090002000C0C000BFBF00B0B0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000003F4F2F000000404F2F
006F7F00707F007F8F008090008FA00090AF00A0B000A0B000B0C000B0C000BFD09F002000CFE0\
00CFE000D0EF00D0EF00E0F000E0EF00EFF000EFF000F0FF00000000FFFF00FFFF
00FFFFBF002000000000FFFF00FFFF00FFFF00FFFF000000C0002FC00020C0002FBF0020C0002F\
EFFF2FF0FF2FEFFF2FF0FF30EFFF2FEFFF2FE0F02FE0F02FDFF02FAF002FA00020
00E0DF00D0CF00CFC000BFAF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007F001F80002060BF60AFC020AFCF20B0D02FB0D020C0DF2FC0DF20\
CFE02FC0E020CFE02FCFE02FD0EF2FCFE0205F6010AF0020AF002FAF0020B00020
AF0020B0002FB0002000F0F000F0F000FFFF00F0F000FFFF00F0F000FFFF00F0F000FFFF000000\
00FFFF00F0F060001F0090B000000000B0CF00CFE000D0EF00DFF000DFEF00EFF0
00EFF000F0FFBF0020BF002F5F5F10F0FF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FBF0020\
C0002FBF0020C0002F00FFFF00FFFF00F0F0BF002000F0EF00FFEF000000000000
AF002000000000000000000090001F8F001F009F8F009080700010007F70007060BF0020B00020\
B0002F00EFEF00EFF000EFEFB0002FAF0020AF002000E0EF00EFEF00E0E000E0EF
00E0E000E0E000D0D000D0D000C0CF00C0C000B0B0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000006060404F3000607000707F
0070800080907000100090A000000000A0B000A0B000AFC000B0C000BFD09F001F00CFDF00CFE0\
00D0EF00D0E000DFEF00DFEF00E0F000E0EF00EFF000000000FFFF00F0F000FFFF
00FFFFC0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFFBF0020C0002FBF0020C0002FBF0020\
F0FF2FEFF02FF0FF2FEFF02FEFFF2FDFF02FDFF02FDFEF2FD0F02F303F0F00E0DF
00DFCF00D0C000BFB07F001F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000002F300F9FBF20AFC02FAFC020B0CF2F\
B0CF20BFDF2FBFDF2FCFE02FC0E02FCFEF2FCFE02FB0EF4080E070AF002FAF0020
00000000EFEF00F0F000EFF000F0F000EFF000F0F000F0F000F0F000F0F000FFFF00F0F000FFFF\
000000BF002FBF0020009FB070001F00BFD000C0DF00D0EF00DFEF00E0F000E0F0
00F0FF000000C0002FBF002020200FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFFF2FC0002F\
BF0020C0002FC0002000000000FFF000FFFF00000000000000F0EF8FBF6F80BF60
00EFE000DFD000DFCF9F001F9F002000000000B09F009F8F7F001F008070008070007F6F5F0010\
AF0020AF002FAF0020AF002FA0002000EFEF00E0E0A000209F002000D0DF00CFCF
00CFCF00BFC000BFBF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000606000606F00606F000000006F7F007F80007F8F
00809000000080001F80001F00A0B000AFBF00B0C000B0CF00C0D000C0D000D0E000D0E000DFEF\
00DFEF00E0F000E0EF000000B00020BF002FBF002000FFFF00FFFF00FFFF00FFFF
C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000BF0020C0002FC000207F801F\
EFFF2FF0FF30EFFF2FEFFF2FE0F02FE0F02FDFF02FDFF02FD0EF2FAF002F00DFD0
00D0CF00C0B000BFAF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000202F0F\
A0BF20A0BF20AFC020AFC020B0D020BFD020C0DF2FC0DF204F5010A00020A00020
A00020A00020A00020AF0020A00020AF0020AF0020AF002FAF0020000000000000000000000000\
000000000000B0002F00809000000000B0CF00C0DF00CFE000DFEF00DFEF00EFF0
00EFF0B0002FBF0020BF002F90F07FEFFF2FEFFF2FF0FF2FEFF02FF0FF2FBF002020200F000000\
000000BF0020C0002FBF0020000000BF0020C0002FBF00208FBF6F80BF608FBF6F
00000000E0DF00DFCF00D0C000000000C0B000B0A000AF9F7F00100090800080700080706F000F\
00706050000F9F002090001F00CFD000C0CF00C0C000000080001F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000003F4F2F00606000606F5F0010600010000000007F8F00808F008F9F
008F9F80001F80001F00000000AFBF00B0C000B0CF00C0D000C0D000CFE000D0E000DFEF00DFEF\
00DFEF00E0EF00E0F0B00020BF0020BF002000FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0C0002F3F3F0FF0FF2FEFF02F\
F0FF2FEFFF2FEFFF2FE0F02FE0F02FDFF02FDFF02FD0EF20AF00209F00209F0020
00C0BF00BFB0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000007F00207F001FAFBFC080001F8F00208F0020000000900020
9F00209F00209F002000DFE000E0EF00E0EF00E0EF00E0EF00EFEF00E0EF00EFEF00E0EF00EFF0\
00EFEF00EFF000EFEFAF002F009FB000B0CF00BFD09F002F00D0EF00E0F000E0F0
00EFFFB00020BF002FBF0020FFFFFFBF00205F6010EFFF2F404010EFFF2FC0002FBF0020000000\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFBF0020C0002F8FBF6F8FBF6F80BF60
8FBF6F00EFDF00E0D000DFCF00D0CF709F5000C0B000B0A08F001F009F8F009080008F7F008F70\
6F0010007F6F000000006060000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000006F6F0000000000006000100000000000000000000000000090A0000000
8F00208F001F90002090001F00BFCF00BFCF00C0D000C0DF00000000D0EF00DFEF00DFEF00E0F0\
00E0EFB0002FB00020BF002FBF002000FFFF00FFFF00FFFF00FFFF00FFFF000000
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000F0FF2FEFFF2FF0FF30\
EFFF2FEFFF2FE0F02FE0FF2FDFF02FDFF02FD0EF2FAF002FA000209F00208F001F
000000600010000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000000B0BF
7F001F80002000BFC000C0CF00C0CF00CFD000000090002000000000DFE000DFE000E0E000DFE0\
00E0EF00DFE000E0EF00000000809F70001F00BFD000CFE000DFEF00DFEF00E0F0
AF0020B0002FB00020000000BF0020C0002FBF0020B0FF60BF0020B0FF60BF0020C0002FBF0020\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000000000000000FFF080BF608FBF6F
00000000F0E000E0D000E0D070A05070A0506F904F00BFAF00000000AF9F009F8F009080000000\
008F7F6F0010007F6F000000006F60000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000500010500010\
006F70006070006F7F000000007F80007F8F008090008F9F80001F80001F000000
00A0B000AFC000B0C0000000000000000000000000AF002000000000DFEF00DFEF000000B00020\
B0002FB00020BF0020BF002000FFFF00F0F000FFFF00FFFF00FFFFBF002000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF000000C0002FBF0020C0002FB0F05FF0FF2FEFFF2F\
EFFF2FE0F02FE0F02FDFF02FDFF02FD0EF203F3F0FA000209F002000000000C0B0
5F0010000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000AFBF00B0BF00B0BF00BFC000BFC000C0CF00BFC000C0CF\
00C0CF0000008F002090002090002070002000AFC000C0DF90002000DFF000E0F0
AF002F00EFF000FFFF00FFFFC0002FBF0020C0002FBF0020C0002FBF0020C0002FC00020C0002F\
00000000FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFF08FC06F8FBF60
00FFF000F0EF00F0E000E0DF7FAF5F7FA0507FA05F00CFBF00000000B0A000AF9F00A090009F8F\
7F0010008F7F7000100080706F0010007F6F0060605F0010000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000006060500010006F6F006F70006F70\
006F7F000000007F8000808F008090008F9F0090A000000080001F90002000AFBF
00B0C000B0C000C0D000C0D0000000A00020AF002FAF0020B0002FAF002000E0F000E0EF00EFF0\
00EFF000F0FF0000008FBF6F8FBF6F8FC06F00FFFF00FFFF00000000000000FFFF
00FFFF00FFFF00FFFF00FFFF000000BF0020C0002FC00020C0002F000000F0FF30EFFF2FEFFF2F\
E0F02FE0FF2FDFF02FDFF02FD0EF2FD0F02FA0002000000000CFC000C0BF600010
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000B0B000809F7000205F8050000000000000A0002F
00E0EF00EFF000F0F000FFFF000000BF002FBF002000FFFFBF0020C0002F000000000000BF0020\
C0002F00FFFF00FFFF00F0F000FFFF00FFF000000000F0F000FFF000FFF000FFF0
80BF608FBF6F80B06000000000E0D07FAF5F70A050709F509F001F00BFB000B0A000AF9F00A090\
009F8F7F0010008F7F7000107000106F0010000000006F60006F60006060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000606000606000606F000000006F70006F7000707F00707F\
000000007F8F008F90008F900090A00090A08F001F00A0B090001F00000000B0C0
00BFCF00C0D000C0D000CFDF00CFDFAF0020AF002000000000DFEF00E0EF00E0EF00EFF000EFF0\
00F0F00000008FBF6F80BF6F8FBF6F8FBF6F00FFFF000000C0002F00FFFF00FFFF
00F0F000FFFF00FFFFC0002FBF0020C0002FBF0020C0002FBF0020F0FF2FEFFF2FEFFF2FE0F02F\
E0F02FDFF02FDFF02FBFEF3F10100000000000DFD000CFC000C0B0600010000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F001F5F8F5F00C0DF9F002F9F0020
00E0F000EFF000F0FF00F0F000FFFFBF0020C0002F00FFFFC0002FBF0020C0002F00FFFF00FFFF\
C00020C0002F00FFFF00FFFF00FFFF00FFFF00FFF000FFFF00FFF000FFF000FFF0
8FC06F8FBF6F8FBF6F80B06080B0607FAF5F7FAF5F709F50A000209F001F00C0B000B0AF00B0A0\
00A09000A0907F00107F001F0F0F00908F1F7000100000006F00106F0010006F60
006F6F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000006060006F6F006F6F00000000000000707F00707F007F80700010008090\
008F9000909F0090A0009FAF00000000A0B000AFB000B0C000000000BFCF00BFCF
00CFDF00CFDF00D0E000D0E000DFEF00DFEF00E0EF00DFEF00E0F000E0F000EFF000EFF000F0FF\
80BF6F8FBF6F8FBF6F8FC06F8FBF6F000000BF0020C0002FC00020C0002F000000
00FFFFC00020C0002FBF0020C0002FC00020C0002F000000F0FF30EFFF2FF0FF2FE0F02FE0FF2F\
DFF02FDFF02F505F10AF002FA0002000DFD000CFC000C0BF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000A0BF608F5F8F00209F0020
00DFEF00EFF000EFF000F0FF00FFFFBF002FBF002000FFFF00FFFFC0002FBF002000FFFF00FFFF\
00FFFFBF002000FFFF00FFFF00FFFF00F0F0FFFFFF00FFF000FFFF00F0F000FFF0
80BF608FBF6F80BF6080BF6080B06000EFDF00DFD000DFCF00D0C0A0002000000000000000B0AF\
00B0A000A09000A09080001080001F908F10908F1F8F80106F80306F00106F0010
00706F00706F006F60006F60006060000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000050000F5F0010
006060006F6F00606F006F706000106F0010000000007F80007F8070001F700010008F90008F9F\
009FA000000000000000A0B000AFBF00AFBF00B0C000B0C000BFCF00BFCF00CFDF
00CFDF0000007FA0607FA0607FA06000000000DFEF00E0F000E0EF00EFF000EFF000F0F000F0F0\
8FBF6F80BF6F8FBF6F8FBF6F8FBF6F000000C0002FBF0020C0002F00F0F000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF000000000000C0002FBF0020F0FF2FE0F02FE0F02FDFF02F\
DFF02FB0EF3FAF0020A0002000DFD000CFC000C0B0008F80000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000009FB05F8050609060900020
A0002F00E0EF00EFF000F0F000FFFFBF002000000000FFFF00FFFFBF0020C0002F00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002000FFFF00FFF000FFFF000000
8FC06F8FBF6F8FC06F8FBF6F8FBF6F80B06000EFE000DFD000DFD0AF001F00D0CF00CFBF00C0BF\
90001F90001F8F001F8F001FA0A01FA09F1F9F901F9F901F908F10908F1F908010
90801F7000107000106F0010000000006F6F6000105F00100000003F4F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000006060000000006060006F6F00606F000000000000
007070006F700000006F001070001F6F001000808F7000107F001F7F001F0090A0009FA08F001F\
8F001F00AFBF00AFBF00B0C000B0C000BFCF00BFCF00C0D0A0002000D0E000D0E0
0000007FA06080AF6F7FAF6080AF6F80AF6000EFF000EFF000EFF000EFF000F0FF00F0FF8FC06F\
8FBF6F8FC06F8FBF6F8FC06F00FFFF000000C0002000000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFFBF0020C0002FC00020606F10EFF02FE0FF2FDFF02F000000\
D0EF2FAF002FA000209F002000CFC000C0BF00908F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000809F0000005F8F5000CFE0
90002000000000E0F000EFF000F0F000000000000000FFFF00F0F000FFFFBF002000FFFF00F0F0\
00FFFF00FFFF00FFFFF0F0F000FFFFBF0020C0002F00F0F000FFFF00FFF000FFFF
80BF608FBF6F8FBF608FBF6F80B06080B06000EFE000E0E000DFD000DFD000000000D0CF00CFBF\
00C0BF0000009F001F8F001F8F001F80A03FAFA01FA09F109F901F9F90109F901F
908F100000007000107000106F0010007F7000706F007F70006F6F00706F600010006F6F006060\
006060006060404F2F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000006060006060\
5F0010006060006F6F00606F006F6F006F6F006F70006F6F6F00106000106F0010
6F001070001070001070001F30300F1010007F001080001F7F001080001F80001F8F001F000000\
00AFBF00AFBF00B0C000B0C000BFCF00BFCF00C0D0000000AF0020000000000000
7FA06080AF607FA06080AF6F7FAF6000EFF000EFF000EFF000EFF000F0F000F0F000FFFF80BF6F\
8FBF6F8FBF6F00000000F0F000FFFFBF0020C0002F00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFFBF0020C0002FBF0020000000AFF05FBF0020DFF02FB00020AF0020\
AF0020A000209F002000CFC000C0B0009F8F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000090AF5F805060905F
00D0EF00000000E0F000EFF000F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF\
00FFFF00FFFF00FFFF00FFFFC00020C0002FBF002000FFFF00FFFF00FFFF00FFFF
8FC06F8FBF6F8FC06F8FBF6F8FC06F80BF6000000000EFE000EFE000E0DF00000000DFCF00DFCF\
00D0C000CFBF00C0B000000090001F4F4010AFAF1FAFAF20A0A01FA09F1FA0901F
A09F1F9F901F9F901F908F1000000000807F00807F007F7000807F007F70007F70007070000000\
00706F00706F006F6F00706F006F6F006F6F0000004F5030404F2F405030404F2F
000000006060006F6F006060006F6F006F6F006F6F006F6F600010600010000000000000000000\
000000000000000000000000000000007F7F00707F00000070001070001F700010
908F1F90801F9F8F1F9F8F1FA0902060904FAF9F2080001F8F00208F001F00000000AFBF00B0C0\
00B0C000BFCF00BFCF00C0D000C0D000CFDF00CFDFAF002FAF0020B0002F7FA060
80AF6F80AF6080AF6F80AF6000EFF000EFF08FB06F80B06F8FBF6F8FBF6F8FC06F8FBF6F00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFFC0002F00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFFBF0020C0002FC00020C0002FBF0020BF002F8FF070B0002FAF002000F0EF\
0000009F002000CFC000C0BF00A090000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000008FA0009FB05F8F5F
00CFE000D0EF00DFEF00E0F000EFF000F0FF00F0F000FFFF00FFFF00FFFF00F0F0C0002F00FFFF\
00FFFF00F0F000FFFF00FFFF000000BF0020C0002F00FFFF00FFFF80BF6F00FFFF
8FBF6F8FBF6F80BF608FBF6F8FBF6F8FBF6F00000000F0EF00EFE000EFE000000000E0DF00DFCF\
00DFCF00D0C000CFC000C0BF00000090001F70B050B0AF1FB0AF1FAFA01FAFA01F
A09F1FA09F1F9F9010A0901F7F00107F001F7F00107F00107000107F0010700010000000000000\
7000106F00107000106F0010000000000000505F304F50304F5F304F502F4F5F30
006F6F007070006F6F0000006F001000707000707000707000707000707000707000707F007070\
007F7F00000070001070001070001F70001000000090801F908F1F90801F9F8F1F
9F8F1FA0901FA0901FA09F1FA09F1FAFA0202F2F00302F0F8F001F00B0BF00B0BF00BFC000BFC0\
00C0CF00C0CF00CFD000CFDFAF0020AF0020AF0020AF002000000000DFEF80AF60
7FAF6080B06F00000080B06F80B0608FBF6F80B0608FBF6F8FBF6F8FBF6F80BF6F8FBF6F00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF000000BF0020C0002FBF0020000000BF0020BF0020B00020B00020AF002000000000EFE0\
00DFD08F001F00C0B000AF9F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000090AF00AFC0
00C0DF00CFE000DFEF00000000EFF000F0F000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF000000C0002FC000208FC06F8FBF6F8FC06F8FBF6F
0000008FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F00F0F000F0F000EFE000EFE0B0002000E0DF\
00DFD000DFD000D0CF00D0CFA0001FA000209F001F7FBF5FBFB01FBFB020B0AF1F
B0AF20AFA01FAFA020A09F1FAF9F1F7F903000000000908F00908F008F8000908F008F80008F80\
008080008F8000807F7F001F70001000808000807F008080007F7F00807F007F7F
70001070001070001F00000070001F000000008080007F7F008080007F8000808000808000808F\
0080800000000000000000005F8F409F901F9F8F1FA0901F9F901FA09F20A09F1F
AFA020AFA01FB0AF20B0A020B0AF20B0AF2070B06090001F9F00209F001FA000209F0020A00020\
A0002000000000D0DF00DFE000DFE000E0EFB00020B0002F00E0EF00EFF000E0F0
00EFF000EFF08FB06F80B06F8FBF6F80BF6F8FBF6F8FBF6F8FC06F8FBF6F8FC06F00000000FFFF\
00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
C0002FBF002000000000FFFF00FFFFBF0020BF002F000000000000AF0020AF002F00EFE000E0DF\
00CFC00000007F001F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000000090AF
5F805000CFE000D0E000DFEFA0002000EFF000000000FFFF00F0F000FFFF00FFFF00FFFFBF0020\
00FFFF00FFFF00FFFF00F0F0C0002F00000000FFFF00F0F08FBF6F8FBF6F8FBF6F
80BF6F00FFFF0000008FBF6F80BF608FBF6F8FBF6F00FFFF00F0F000F0EF00EFE0BF002000E0DF\
00E0DF00DFD000DFD0AF001FAF0020A0001FA000209F001FC0C020BFBF1FBFBF20
B0B01FBFB01FB0AF1FB0AF1FAFA01F8F001F8000108F001F80001000000000908F009F9000908F\
00908F008F8F00908F008F80000000008F80008F8F008080008F8F008080008F8F
000000000000008080008F8F008080008F8F00808F008F8F00808F7F001F7F00107F001F7F0010\
80001F7F001080001F7F0010A09F1FA0901FAF9F1FA09F1FAFA01FAFA01FB0AF20
AFA01FB0AF20B0AF1FBFB020BFB01F0000009F001F9F00209F001FA00020A0002000000000CFDF\
00D0E000D0E000DFEF00DFEF00E0EFB0002000E0F000E0EF00E0F000E0EF00EFF0
00EFF080B06F80B0608FBF6F80B0608FBF6F8FBF6F8FBF6F80BF6F8FBF6F8FBF6F00FFFF00F0F0\
00FFFF00FFFF00FFFF00F0F0C0002F00FFFF00FFFF00F0F000FFFFBF0020000000
00F0F000FFFF00FFFF00FFFFBF0020BF002000F0EF00F0EFAF0020AF002000EFE000E0DF00CFC0\
80001F7F001F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000005F8F5F00D0E000D0E000E0F0AF0020B0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00000000FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF8FBF6F8FC06F8FBF6F
8FC06F00FFFF00FFFF00FFFF00000000FFFF00FFFF8FBF6F8FC06F8FBF6F00F0F0BF0020BF002F\
000000B0002000E0DF00E0DF00DFD000DFD000D0CF00D0CFA0001FA0002030300F
C0C0201F1F00C0BF20A0B0309F001F90001F0000008F001F90001F8F001F8F001F8F001F8F001F\
009F90009F9F009F90009F9F00909080001F009090009F90009090009090009090
00000000909000909000909000909F00909000909F009090009F9F00909F009F9F00909F009FA0\
80001F80A040AF9F1FB0A020AFA01FB0AF20B0A01FB0AF20B0AF20BFB020BFAF20
BFB020BFB020C0BF20C0BF20CFC02FAFC03F00000000C0D000CFDF00CFDF00D0E000D0E000DFEF\
00DFEF00E0EF00DFEF00E0F000E0EF00EFF000E0EF00EFF000EFF000EFF000EFF0
8FBF6F80B06F8FBF6F8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F00000000FFFF00FFFF\
00FFFF00FFFF00FFFFC0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFFBF0020BF002F00F0EF00F0EFAF0020AF002F00EFE000DFD000CFC0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000507F505F8F5000D0E000D0E000E0EFA00020B0002F00FFFF00FFFF00F0F000FFFF00FFFF\
BF002F00000000000000000000FFFF00F0F000FFFF00FFFF00000080BF6F8FBF6F
8FBF6F00000000F0F000FFFF00FFF000FFFF00F0F08FBF6F8FBF6F8FBF6F80B060000000B00020\
BF002000E0E000EFE000E0DF00E0DF00DFD000DFD000DFCFAF0020A0001FAF0020
A0001FA000209F001F0000009F001F9F001F90001F90001F00AFAF00B0AF00AFA000AFA000A0A0\
00AFA000A09F0000008F001F8F001F8F00108F001F8F001F0000008000108F001F
000000009F9F009F9F009FA0009F9F00A0A00000000000008F001F8F001F8F001F8F001F8F001F\
90001F8F001F8FAF40B0AF1FBFB020B0AF1FBFB020BFB01FBFB020BFB01FC0BF20
C0BF20C0C020AFBF30A00020A0001FAF0020A0001F00000000D0DF00DFE000DFE000DFEF00DFEF\
00E0EF00DFEF00E0EF00E0EF00000000E0EF00EFF000EFF000F0F000EFF000F0F0
80B06F8FBF6F80BF6F8FBF6F8FBF6F8FBF6F80BF6F8FBF6F8FBF6F00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F0C0002FBF002000FFFFBF002000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFFB0002000F0F000EFEF00F0EF000000A0002000E0DF00DFD000C0BF7F001F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005F7F5060905F00D0E000DFEF00EFF0AF0020B0002FB00020000000000000BF002F\
BF0020C0002FBF002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF8FBF6F
8FC06F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF8FBF6F8FC06F8FBF6F8FC06F000000000000\
BF002000000000EFEF00EFEF00E0E000EFE000E0DF00E0DFAF0020B00020AF0020
00DFD000D0CFAF0020A0001FA00020000000A000209F001F9F002000BFB000BFB000B0B000B0B0\
00B0AF00B0B000AFAF00B0AF00AFAF90001F90001F00000000AFAF90001F90001F
00000000000000AFAF00AFAF00AFAF00AFAF00AFB000AFAF00AFB000000000B0B000000000B0B0\
0000009F00209F001F9F0020302F0FC0BF20C0BF20C0C020C0BF20CFC02FCFC020
CFCF2FCFCF20AF0020AF0020AF0020AF0020B0002FB0002000000000DFEF00E0F000E0EF00E0F0\
00E0EF00EFF0000000BF002F00EFF000F0F000EFF000F0FF00F0F000FFFF000000
8FBF6F8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F00000000FFFF00FFFF00000000FFFF\
00FFFF00FFFFC00020C0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFFB0002000F0F000F0EF00F0EF00EFE0AF002F00E0DF00D0CF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005F805060905F00D0E070A0607FB06F00EFF000FFFF00FFFF00FFFF00F0F0\
000000BF0020C0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF000000
00000000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF80BF6F8FBF6F8FBF6F8FBF6F00F0F0\
00F0F0BF002000F0F000EFEF00EFEF00E0E000EFE000E0DFB00020B00020000000
00DFD000DFD000D0D0AF0020A0001F00CFCF00CFC000CFC0A0001FA0002000000000C0BF00BFBF\
00BFBF00BFB000BFBF00B0B000BFB090001F00000000B0B000B0B000B0B09F001F
90001F00B0B000B0B000B0B000B0B000B0BF00B0B000BFBF00B0B000BFBF00B0BF00BFBF9F001F\
9F00209F001FA00020000000CFC0201F1F00A0002050500FCFCF2030300FAF0020
A00020AF002000D0DF00DFE000DFE0000000B00020B00020B0002000000000E0EF00EFF000E0EF\
00EFF0000000BF002F00000000F0F000EFF000F0FF00F0F000FFFF00F0F08FBF6F
80BF6F8FBF6F8FBF6F8FBF6F80BF6F8FBF6F8FBF6F0000000000008FBF6F8FBF6F8FBF6F000000\
000000BF0020C0002FBF002000FFFF00FFFF00FFFF00F0F000FFFF00F0F000FFFF
00F0F000F0F000EFEF00EFEF00EFE000000000DFD000CFC0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000008F9F5F805F609F5F00D0E070A0607FB06F00F0F000FFFF00FFFF00FFFF\
00FFFF00FFFF00000000000000FFFF00FFFF00FFFF00FFFF00FFFF000000BF0020
C0002FC00020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF8FBF6F8FC06F00FFFF00FFFF\
00FFFF00FFFF00F0F000F0F000F0EF00F0EF00EFE000EFE000000000000000E0DF
00E0E000E0DF00E0DF000000B0002000000000DFD000D0CF00D0D000D0CF00D0CFA0001F00CFCF\
00CFC000CFC000C0C000C0C000C0C000C0C000C0BF00C0C000C0BF00C0C000BFBF
A0002000BFBF00C0C000BFBF00C0C000C0C000C0C000C0C000C0CF00C0C0000000A0001FA00020\
A00020AF002000CFCF000000AF0020AF0020AF0020B00020AF0020B00020AF0020
B0002FB0002000E0EF00E0EF00E0EF00E0EF00E0F000E0EF00EFF000E0EF000000B00020000000\
000000BF002FBF002000F0FF00F0F08FBF6F8FBF6F00000000FFFF8FC06F8FBF6F
8FC06F8FBF6F8FC06F8FBF6F8FC06F00000000FFFF8FBF6F8FC06F8FBF6F8FC06F8FBF6F000000\
000000C0002FBF002000000000FFFF00FFFF00FFFF00FFFF00FFFF00000000F0F0
00F0F000EFEF00EFEF00EFE000EFE000D0CF00000000B0AF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000008F9F5F805060905F00000070A0607FB06F00000000FFFF00FFFF\
00FFFF00F0F0000000BF0020C0002FBF0020C0002F00FFFF000000000000C0002F
BF0020C0002FBF0020C0002FBF002000FFFFBF0020C0002FBF0020C0002F00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF00F0F000000000EFEF00F0EF00EFE000000000E0E000EFE0
00E0E000E0E000E0DF00E0DFB0002000000000DFD000DFDF00DFD000DFDF00D0D000DFD000D0CF\
AF002000CFCF00D0CF000000AF002000CFCF00CFCF00CFCF00CFCF00C0C000CFCF
00C0C000CFCF00C0C000CFCF00CFCF00CFCF00CFCF00CFD0000000AF0020A0001F00000000CFD0\
00D0DF00CFD000D0DFAF0020B00020AF002000000000DFE0000000B00020B00020
B0002000000000E0EF00E0EF00E0EF00EFF000E0EF00EFF000EFEF00EFF000EFEF000000BF0020\
BF002FBF002000F0FF80BF6F8FBF6F80BF6F8FBF6F8FBF6F8FBF6F00F0F000FFFF
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF80BF6F8FBF6F8FBF6F8FBF6F80BF6F00000000FFFF\
00FFFF000000C0002F00FFFF00FFFF00F0F000FFFF00F0F000000000F0EF00F0EF
00EFEF00EFEF00E0E000E0DF00CFC000C0BF00AFA0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000608F5F609F5F70A06070AF607FB06F00F0F000FFFF\
00FFFF00FFFF000000BF002FBF0020C0002F00000000FFFF00FFFF00FFFF00FFFF
00FFFFBF002000FFFF00FFFF00FFFF00FFFF00FFFF000000C0002FBF0020C0002FC0002000FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFFC0002F00F0F000F0F0BF002000F0F000EFEF
00EFEF00EFE000EFEF00E0E000EFE0B0002000E0E000E0DF00E0E000E0DF00E0E000DFDF00E0DF\
00DFDF000000B00020B00020AF002000DFDF00D0D000DFDF00D0D000DFDF00D0D0
00D0D0AF0020B0002000D0D000D0DF00D0D000000000D0DF00DFDF00D0DF00DFE000DFDF00DFE0\
00DFDF00E0E0000000BF0020B0002000E0EF00E0EF00E0EF00E0EF000000B00020
BF002F00000000EFF000EFEF00EFF000EFF000EFF000EFF000F0F000EFF000F0FFBF0020C0002F\
00000000FFFF8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F00000000FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF8FBF6F8FC06F8FBF6F8FC06F8FBF6F00000000FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0F000000000F0EF00F0F000EFEF
00EFEF00E0E000DFDF00CFC000BFB0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000608F5F0000006FA06070AF607FB06F00F0F0\
00FFFF00FFFFBF002F000000000000BF002000FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0C0002F00000000FFFFBF0020\
00000000FFFF00FFFF00F0F000FFFF00FFFFC0002FBF0020BF002000F0F000F0F0
00EFEF00EFEF00EFEF00EFEF00E0E000EFE000E0E000EFE000E0E000E0E000E0E000E0E000E0DF\
00E0E0B00020B00020B00020B000200000000000007FA05F80AF607FA05F000000
00DFDFB00020B00020B00020B0002000E0E000DFE000E0E000DFE000E0E000E0E000E0EF00DFE0\
00E0EF00E0E0BF0020B0002000E0EF00E0EF00EFEF00E0EF00EFEF00E0EF000000
B0002000EFF000EFEF00EFF000EFEF00EFF000EFF000F0F000EFF0000000BF002000000000F0F0\
00FFFF8FBF6F8FBF6F80BF6F8FBF6F8FBF6F8FBF6F00000000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF00F0F08FBF6F8FBF6F8FBF6F00000000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFF000000BF0020000000B0002000F0F000EFEF00F0EF00EFE000EFEF
00E0DF90002000C0BF00BFAF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000060905F00CFDF00000070AF6000F0F0\
00FFFF00000000FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFFBF0020C0002F00000000FFFF00FFFFC0002FC00020C0002FBF0020
00FFF000F0F000F0F000EFEF00F0F000EFEFBF002F00EFEF00EFEF00EFEF00EFEF00E0E0000000\
00000000EFE000E0E000EFEF00E0E080B06080AF6080B06080AF6080AF6080AF60
80B06000E0E0BF0020B00020BF002F00000000E0EF00E0E000EFEF00E0EF00EFEF00E0EF00EFEF\
00E0EF00EFEF00000000EFF000EFEF00EFF000EFEF00EFF000EFF000EFF000EFF0
00000000EFF000F0F000EFF000F0FF00F0F000F0FF00F0F000FFFF00F0FF00FFFF00FFFF00FFFF\
00FFFF0000008FBF6F8FC06F8FBF6F00000000FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFFBF0020BF002FB0002000000000EFEF00F0EF00EFEF0000009F0020
00000000C0BF00B0AF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F805000CFDF00D0E000DFEF\
AF002FAF002000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F0C0002F00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF00FFFFC0002FBF0020C0002FBF002000FFFF00F0F000FFFF
00F0F000FFFF000000BF0020BF0020BF002FBF0020BF0020BF0020BF0020B0002000EFEF00EFEF\
00EFEF00E0E000EFEF00E0E000EFEF80AF6080B06080AF6080B06080AF6080B060
80AF6000EFEF00E0E000EFEF00E0EF000000EFE0EF00EFEF00E0EF00EFEF00E0EF00EFEF00E0EF\
00EFF000000000EFF000EFEF00EFF000EFEF00EFF000EFF000F0F000EFF000F0F0
00000000000000F0F000F0F0000000C0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F000FFFFBF002000000000F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFF00F0F0C0002FBF0020C0002F00000000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFFBF0020BF0020B00020B0002F000000000000000000A0002090001F00CFC0
00B0AF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000A0B000CFDF00D0E0\
00DFEFAF002F00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF
00FFFF00FFFF00FFFF00000000000000FFFF00FFFF00FFFF00FFFF00FFFFC0002F00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF000000C00020C0002FBF002000FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF000000C0002FBF0020C0002F00000000F0F000F0F000F0F0\
00EFEF00F0F000EFEF00F0F000EFEF00F0F000EFEF00000080B0608FB06F000000
00EFEF00EFEF00EFF000EFEF00EFF000EFEF00F0F0000000BF002FBF002000000000EFF000F0F0\
BF002000F0F000EFF000F0F000EFF000F0FF00F0F000F0F000F0F000F0FF00F0F0
C0002FBF0020C0002FBF002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFFC0002FC00020C0002FBF0020C0002FC00020C0002F000000FFFFFFC00020
C0002FBF0020C0002FC00020C0002FBF0020C0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF\
00000000F0F000F0EF00F0F000EFEF00EFEF00EFE000DFDF609F5000C0BF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000C0D0\
00D0E000DFE0A0002FAF002000FFFF00F0FF00FFFF00F0F000FFFF00FFFFBF002F
00F0F000FFFF00FFFF00FFFFBF002000000000FFFF00FFFF00F0F000FFFFBF002000000000F0F0\
00FFFF00FFFF00FFFF00F0F000FFFF000000C0002F00F0F0000000BF0020000000
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00F0F000FFFF0000008FBF6F00000000F0F000F0F0\
00F0F000F0EF00F0F000EFEF00F0F000EFEF8FB06F80B0608FB06F00000000F0F0
00EFEF00F0F000EFEF00F0F000EFEF00F0F000EFF000F0F000EFEFBF0020BF0020BF002FBF0020\
BF002000000000000000000000000000F0F000F0FF000000C0002FBF0020C0002F
BF0020C0002F00000000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000C0002FBF0020\
C0002F000000C0002F00000000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF
00F0F0000000BF002000FFFF00F0F000FFFF00FFFF00FFFF000000C0002FBF0020BF002FB00020\
00F0F000EFEF00EFEF00E0E000EFE000DFDF6F9F5F00000050804F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
5F8F5000CFDF00D0E000E0F000EFF000FFFF00F0FF00FFFF00FFFF00FFFF00FFFF
C0002F00FFFF00FFFF00FFFFC0002FC00020C0002F00FFFF00FFFF000000C0002FBF002000FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF
00000000FFFF00FFFF00FFFF00FFFF00FFFF0000008FBF6F8FC06F8FBF6F8FBF6F8FBF6F000000\
00F0F000FFFF00F0F000FFFF8FBF6F8FBF6F80B06F8FBF6F8FBF6F8FBF6F80B06F
00000000F0F000F0F000F0F000F0FF00F0F000F0FF00F0F000FFFF00F0F0000000BF0020C0002F\
BF002000000000F0F000FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFFC00020C0002FBF0020C0002FC00020C0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFFC0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000BF002FB0002000F0F0\
00EFEF00EFEF00E0E000E0E06F9F5F60905050804F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000050704F00CFDF90002000E0EF00EFEF00F0FF00F0FF00FFFF00F0F000FFFF
00F0FFBF002FBF002000FFFF00FFFFC0002FBF0020C0002F00FFFFC0002FBF0020C0002F000000\
00FFFF00F0F000FFFF00FFFF00000000000000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFFBF002000FFFF00FFFF00FFFF00F0F08FBF6F8FBF6F8FBF6F80BF6F8FBF6F8FBF6F\
00FFFF00F0F00000008FBF6F8FBF6F80BF608FBF6F8FBF6F8FBF6F80B0608FBF6F
80BF6F8FBF6F00000000FFFF00F0F000FFFF0000008FBF6F80BF6F8FBF6F80BF608FBF6F00F0F0\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0C0002F
BF0020C0002FBF0020C0002F00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0000000
00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFFB0002000F0F000EFEFAF002000EFE0\
00EFEF00E0E06FA05F5F90505F8F4F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000808F00C0DF90002000DFEF00E0EF00F0FFB00020B0002F00FFFF
00FFFF00F0FFBF002FBF0020C0002FBF0020C0002FC00020C0002F000000C0002FC00020C0002F\
00FFFF00FFFF000000C0002FBF002000000000FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFFC00020C0002FBF00200000000000008FC06F8FBF6F8FC06F8FBF6F8FC06F\
0000008FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F
8FC06F8FBF6F8FC06F8FBF6F00FFFF00FFFF8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F\
00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000BF0020C0002FC00020
000000BF0020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
BF002000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFBF002000000000FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00F0F0BF002F00F0F000F0F000EFEF00EFEF000000000000\
00DFD0609F5F00B0B0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000008000208F002000DFE000E0EF00F0F000F0FFB0002F
B00020B0002F000000BF002FBF0020BF002FEFF02F40400FBF0020F0FF2F5F5F10C0002FBF0020\
C0002F000000C0002FBF0020C0002FBF002000FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFF00FFFF00FFFF000000C0002FBF0020C0002F00F0F000FFFF00FFFF00000000F0F0\
00FFFF00FFFF00000080BF6F8FBF6F8FBF6F8FBF6F80BF6F8FBF6F8FBF6F8FBF6F
80BF6F8FBF6F8FBF6F8FBF6F00F0F00000008FBF6F8FBF6F80BF6F8FBF6F8FBF6F8FBF6F000000\
00FFFF00FFFF00FFFF00F0F000FFFF00000000FFFF00F0F000FFFF00FFFF00FFFF
00F0F000FFFFBF002000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFFBF0020\
C0002F00FFFF00FFFF00F0F000FFFF00FFFF000000BF002000FFFF00FFFF00FFFF
00F0F000FFFF00F0F000FFFF00F0F000F0F000EFEF00F0EF00EFEF00EFEF00E0E09F002000CFC0\
00BFB0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000008F002000D0E000DFE000EFF000F0F0
00FFFFB00020B0002FB00020BF002FE0FF2FB0FF60EFFF2FF0FF30EFFF2FF0FF2FEFFF2FBFFF60\
EFFF2F404010BF0020C0002FC0002000000000FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFFC0002F00FFFF000000BF0020C0002F00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F
8FC06F8FBF6F8FC06F00FFFF00FFFF00FFFF0000008FBF6F8FC06F000000000000000000000000\
C00020C0002FBF002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000BF0020C0002F\
C0002000FFFF00FFFF00FFFF00FFFFC0002FBF002000000000FFFF00FFFF00FFFF
00FFFFBF002000000000F0F000F0F000EFEF00EFEF00EFE000EFEF00000090002000BFBF00B0AF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000C0DF00000000DFEF
00E0EF00F0FF00F0F0B0002FB00020BF002FE0F02FEFFF2FE0F02FEFFF2FEFF02FF0FF2FEFFF2F\
F0FF2FEFF02F7F7F1F3F3F0F40400FEFF02FC0002FBF0020C0002FBF0020C0002F
BF0020C0002F00000000FFFF00FFFF00FFFF00F0F000FFFF000000C0002F00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFF0000008FBF6F8FBF6F8FBF6F80BF6F8FBF6F
8FBF6F00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF000000C0002FBF0020\
C0002F00000000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF
00FFFF00FFFFBF0020C0002F00000000FFFF00F0F000FFFF00FFFFC0002FBF0020C0002FBF0020\
00000000F0F000FFFFBF0020C0002FBF002000000000FFFF00FFFF000000BF002F
00F0F000F0F000EFEF00EFEF00EFE000EFEF00E0E000DFD000000000BFB0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000000C0D0
00D0E0000000AF002FAF0020B0002F000000E0FF30E0F02FE0FF2FE0F02FEFFF30EFFF2FF0FF2F\
EFFF2FF0FF30EFFF2FF0FF2FEFFF2F000000C00020C0002FBF0020C0002FC00020
C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000C0002F00000000FFFF\
000000C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00FFFF000000C0002FC00020C0002FBF002000FFFF00FFFFC0002FBF0020C0002FC00020C0002F\
000000C0002FC00020C0002FBF002000FFFF000000BF002FB0002000F0F000EFEF
00F0F000EFEF00EFEF00E0E000E0E000D0CF00C0C000B0AF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000C0D000D0DF00E0EFA00020AF002FAF0020B0002FDFF02FE0F02FE0F02FEFFF2FE0F02F\
EFFF2FEFF02FF0FF2FEFF02FF0FF2FEFFF2FD0FF4F000000C0002FBF0020000000
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFFC0002F00F0F0\
00FFFF00FFFF00FFFF00F0F08FBF6F8FBF6F8FBF6F80BF6F8FBF6F8FBF6F8FBF6F
00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFF00FFFF00000000000000FFFF00FFFF00F0F000FFFF00FFFF00FFFF
00F0F0000000BF0020C0002FBF0020C0002F3F3F0F40400FBF0020C0002FBF0020F0FF2FB0F05F\
C0002FBF0020C0002F3F3F0FBF002FB00020BF002F00EFEF00F0F000EFEF000000
A0002000000000DFD000CFCF7F001F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000C0D000D0E000DFEF00EFF0AF0020CFFF40DFF02FE0FF2FE0F02FE0FF30\
E0F02FEFFF2FE0F02FEFFF30EFFF2FF0FF2FEFFF2FF0FF30B0FF60C0002FBF0020
00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFFC0002F\
C00020C0002F00000000FFFF8FBF6F8FC06F8FBF6F8FC06F8FBF6F8FC06F8FBF6F
8FC06F00FFFF00FFFF00FFFF00FFFFC00020C0002F00000000FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFFC00020C0002FBF0020C0002FC00020C0002F00FFFF
00FFFF000000C0002FBF0020C0002FEFFF2FF0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30\
EFFF2FEFFF2FBF0020BF002FB00020000000AF0020AF002FA00020A0002000E0E0
00D0D000C0BF7F001F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000C0D000CFDF00DFEFA00020AF002FAF0020DFF02FDFF02F\
E0F02FDFF02FE0F02FE0F02FEFFF2FE0F02FEFFF2FEFF02F000000BF0020C0002F
BF0020C0002FBF0020C0002FBF0020C0002FBF0020C0002F000000000000F0F0F0C0002FBF0020\
C0002FBF0020C0002F00000000FFFF0000008FBF6F8FBF6F8FBF6F80BF6F8FBF6F
000000C0002FBF0020C0002FBF002000000000F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF\
00FFFF00F0F000FFFF00FFFF00FFFFBF0020C0002FBF0020C0002FBF0020C0002F
000000F0FF2FEFF02FF0FF2FEFFF2FF0FF2FEFF02FF0FF2FEFFF2FEFFF2FE0F02FE0F02FE0F02F\
E0F02FDFEF2F000000AF0020AF0020A00020A0002000E0E000DFDF00CFC0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000C0CF900020900020A0002FA00020AF002F\
D0F02FDFFF30DFF02FE0FF2FE0F02FE0FF30E0F02FEFFF2FAFF05FB0FF603F400F
C0002FBF002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFFC0002FC0002000000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
000000BF0020C0002FC00020C0002FBF0020C0002F00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000C0002F00000020200F90FF7F
F0FF2FEFFF2FF0FF30EFFF2FF0FF2FEFFF2FF0FF30EFF02FEFFF2FE0F02FE0F02FDFF02FDFF02F\
D0EF2FBFEF40DFEFEF00EFEF00E0E000E0E000D0D000000000B0AF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000BFCF00CFDF00D0DF\
00E0EF000000AF002FAF0020B0002FDFF02F1F200FB00020505F103F3F0FBF002F
B00020BF002FBF002000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0\
00FFFF00FFFFC0002F00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF
000000C0002F00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF000000C0002FBF0020\
C0002FBF0020C0002FBF0020C0002FBF0020C0002FBF002090FF7FEFFF2FF0FF2F
EFF02FEFFF2FEFF02FEFFF2FE0F02FE0F02FDFF02FDFF02FD0EF20D0EF2FD0EF2FAF0020A00020\
A000209F002090002000C0C000BFBF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00BFCF00CFDF00D0E0000000A00020AF002F000000B0002FB00020B0002FB00020
00FFFF00F0FF000000B00020BF002FBF0020BF002F00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF\
00FFFF00FFFF00FFFF00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF
00000000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF\
000000C0002FBF0020C0002FC00020000000DFFF3FEFFF30EFF02FEFFF2FE0F02F
E0FF2FE0F02FE0F02FDFF02FDFF02FDFEF2FDFF02FD0EF2FD0EF2FCFEF2FA0002F00DFDF00D0D0\
00BFBF00B0B0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000BFC000CFD000D0DF9F002000EFF000F0FF00EFF0AF002F
AF0020B0002F00EFF000F0FF00F0F000F0FF00F0F000F0FF00F0F000FFFF00000000FFFF00F0F0\
00FFFF00F0F0000000BF0020C0002F00F0F000FFFF00FFFF00FFFF000000C0002F
00000000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00F0F0\
00FFFFBF0020BF002FBF0020BF002FB00020AFF05FDFF02FE0F02FDFEF2FDFF02F
D0EF2FD0EF2FD0EF20D0EF2FCFEF2F3F3F0FA00020A0002090002000C0C000B0B0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000BFC000C0D000CFD000DFE000DFE0
00EFF000EFF0000000AF002000F0FF00EFF000F0FF00F0F000F0FF00F0F000F0FF00F0F000F0FF\
00F0F000FFFF000000BF002FB00020BF002F00F0FFBF002FBF0020BF002FBF0020
BF002FBF0020000000BF002000000000F0F000000000F0F000000000F0F0000000B00020000000\
00000000000000F0F0000000AF0020B0002FAF0020AF002FB0EF40000000A00020
A0002FA00020A0002F0000000000008F001F00000070001F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
7F001F8F002000C0D000D0DF00D0E000E0EF00000000EFFF00EFF000EFF000EFF000F0FF00EFF0\
00F0F000000000000000EFF000F0F000EFF000F0F000EFF000F0F0AF0020B0002F
00EFF000F0F000EFF000F0F000EFF000F0F000EFF000F0F000EFF000F0F000EFF000F0F000EFEF\
000000AF0020AF002FA0002000000000E0EF00EFEF0000009F002000D0D000CFCF
00BFBF00B0B0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000B0BF00BFC000C0CF00D0DF9000209F00209F0020A0002F\
A0002000000000EFF000EFF000EFF000EFF000EFF000F0F000EFF000EFF000EFF0
00000000EFF000EFF000EFF000EFF000EFF000EFF000EFEF00EFF0000000A0002F00000000EFF0\
00E0EF00E0EF00DFDF00D0DF00C0CF00BFC000B0B0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000050804F5F8F500000000000005F8F50609050609050609F5F609F5F000000
00CFD000D0D000CFCF00CFCF00C0C000C0CF00BFBF00BFBF00B0BF00B0BF00000000AFB0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[17]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001PN;0003E4iD08@020L000H7
PP006G5`04iDCP1>E4iDCUA>E4h0CUA>E4iD08D0M`081`001PMh000[CUA[04iDCUA>E4iDCP1[Q4iD
CUA>E4iDCUA>E4iDCUA>E4iDJhA>E4iDCP1_00P700061gD000Q>E4iDCXA[Q08000m?E5ADJhF4QHB5
Q024Q@002e@03U=DDeACE5<0Q8ECE4iDJ`081`001PMb000KCUA>E4iDCP1[Q6^4JhA[E024JhA[04iD
CeA>00ED01UCE000JhA[Q6^4JhA[Q6/0CXA[Q6/0JhA>06L020L000H7KP0045A>E4iDCe@0QF^5Q8F4
Q@07E0805e@01@24QHB500:400=fQH@00X@20003Q4h006<020L000H7J`0045A>E4iDCUA>E4iDCUA>
Q00;E00408B401=D0P02Q08018@00f^>SP020006L8i`Q4iDH0081`001PMU0008J`1>E4iDJhD9E00K
08B5E5EDEEAEE5EDEEAEE024QHB5Q8F4QHB5008002EDIUAVE6IDIEAVE6EDI@25Q8F4QHB50:R>Z8jX
SZR>Z8jXSZR405X020L000H7H`0036/0E024JhA>E4iDC`ID00B4Jh@01U@50003E02400N400<0E5@0
55@02@24Q8@0MHn>Y`08SP04RHj9Q5P020L000H7H@001f^508E[QDl02U@06024QHB5Q01DE@2?LHma
SgJ?MXl0IUAVE0800365Q8F4IUAVE6IDIUAVE6IDIUAVE6IDIUAVE6H0QHB5Q72>Z8jXSZR>Z8jXSZR>
Z7ED05D020L000H7G`04J`0BQ4iDCUA>E4mDCeA[Q6^4JhA[0h@03P2?L8m`Sg2?L8m`Sg402E@01@24
Q01E011D0P0:Q0080:N>YhjWSZL2SP05RHj9STh0D`081`001PMM000=L8E`QF^5JhF4QHB5Q002E007
EEAEE8F4Q@02000NE5ED07J?MXmfSgJ?MXl0IUAVE6IDIUAVE6IDIUAV0P00;EAVE6IDIUAVE6IDIUAV
09j4WXBNQ9j4WXB5Q8F?[Hj]SZf>Z8jXSZR>Z8j5CP1@00P700061e/0011f08i[SV^4JhA[E4mDCeA?
15@2Q0MD0P00776?LHmaSg40E5EDEEAEE6IDEEAVE6IDIUAVE002Q008E5EDIUAEE6H3E0j400/0SZN>
ShjWSZN>Y`02SP05RHj9STh0C`081`001PMJ0008ShV?RHE[QH@4E00eEEAEE5EDEEB5E5EDIUAVE6ID
IUAVE6H0SgJ?MP1EIUEVEFIEIUEVEFIDIUAVE6IDIXBNQ9h00`000ij40002000RE6IDIUAVE6IDIUAV
E6ID08BNQ8F4U8j]SZf>Z8jXSZR>Q4d020L000H7F@0038i`SV^4JeA?E4mDC`MD0P03E008EEAEE5ED
EE@2000MSg6?LHmaSg6?MXmfSeAVE6IDIUAVE6IDIUAVE6H01h@03@1DIUAVE6IDIUAVE6H01e@00`24
Q003Q009L8jWSZN>YhjW00:>0W000f]D001900P700061eL000E`ShV?SP0200ID01iEE5EDEEAVT6ID
IUAVE6IDIUAVE@1fSgJ?MXmfSgH7S`0306IV00=V00=E08l01000=6IE08BNQ9j4WXBN06IDIUAVE6ID
IUAVE6IDIUAV09j4WXBNQ9j?[HjDSZf>Z8jXSXF4QF]800P700061eD000Z>L8i`SP24JhA?2U@09H@0
EEAEE5EDEEAVE8mfSgJ?MXmfSgJ?MXmfSgJ?MXmfSgH0EFH00P003WJ?MXmfSgJ?MP1EIUEV0P03Q00B
IUAVE6IDIUAVE6IDIEAUE6D01X@20007Q024Q8n>Y`04SP04JhA[E4H020L000H7DP003Hm`ShV?JhE[
QF^5Q8D00`005eD008B5Q01DIUAV06IEIUD0MXmfT7J@00N?01n@Si2?UHnESiF?U@1WIP2?T8nESi2?
UHnESiF?UFIW00MV0P05IP0;E6IDIUAVE6ID08@00P002VIDIUAVQ8F4WX@20009QHBX08F408AD04<0
20L000H7D@002Xi`SW2>L8A[Q6/6E003Q6^400>40280E5D0SgJ?MXmESgJ?MXmfSgJ?MXmfSgJ?MXmf
T7J@MY001VH02hmfSgJ?MXmfSgJ@00803FH00`240002IP0@E6IDIUAVE6H0IUAVE6EDI@=D1H@00`24
Q004Q08000>40000@0081`001PM@0006ShV?RHn>0P02E00FEEAEE5EDIUB5Q8F407J?MXmfSgJ@MP9V
0P00791fUHnESiF?T8nESiF?UHnES`IV1VH6IPIVIfH5000>SiF?06H6IPIV1VH6IVL9IP03WX@000EV
0P0079iDIUAVE6IDIUAVE9j4WXBN09j4QHB5Q8ED04i000P700061dl000R>L8i[Q6/0C`ID00AEE5D0
0h@03EAVMXmfSgJ?MXmfIUD016H0:01fT7J?MXmfT7J@MY1f06IWIVMVIfIWIVMVIfIWIVMVIfH0Q9j4
WX@20003IfIW00]V00D0Q9j4WP02000?Q9f405AVE6IDIUAUE6D000B400@0E5@00X@01UB4Q5=DCSl0
20L000H7CP001hn9Sf^5JhD00e@055EDEEAVE6J405AVE@1fSgJ?MXmf0Xl02VMVIfIWIVMVIfH2000c
UHnESiF?U@06IPIV1VH6IPIV1VH6IPIV1VH6IYjMWXBNQ9j4WXBNQ01V1VH6IVMVIfIW009V00T0WXBN
Q9j4WX@00fH04fEVE6IDIUAVE9j4WXAVE6ID08@00P001HF4E5=D03d020L000H7CP001g2>JhA[Q000
15@05EEDEEAE05EDIUAVMXmfSgJ?MXmf0005IP0=IfIWIVMVIfIWIVL0T002000LIVMV1VH6IPIV1VH6
IVMVIfIW09j4WXBNQ9j4WP@000B4WX@036H01XBNQ9j4WP9V00aUIVEVE6IDIUAVE002Q00506EDIE@0
0h@01U@0Q5ACE3`020L000H7C002L00606^5JhE[0U@04eEDEEAVE01DIUEVEFIEIP2?MY000P0046IW
IVMV1VH6IPIV1VH6IPH2000_SiF?UHnESiEV1VH6IPIV1VH6I`IW09fNW@1W1VH6IYjM06H6IPIV1VH6
IPIV09d00`0036H000IVWXBNQ9j400AV019UIVEVE6IDIXBN06IDIU@0Q8D2E005Q01DE5<0>P081`00
1PM;0009SW24J`1[Q6/000ED00eE05EDIUAVEFIEIUEV00801fH07VMVIfIWIVMVI`2ESiF?UHnESiF?
UHl6IPIV1VH6IP8002L700IV1VH6IPIV1VH6IP2406H6IPIV1VH6IPIVIfH0Q9j4WXBNQ9h00P001HBM
Q9d000=V00iUIUAVE6IDIX@0E6EDI@9D0X@3E003Q5=D03X020L000H7BP00?GH006^5E01[QEAEE5ED
IXAVEFIEIUEVEFL0WXBNQ9j406IWIPIV1VH6IPIV1VH6SiF?UHnESiF?UHnEI`H00P00;ifNWIjM06L6
I`IW1VL6I`IW1VL6I`H01VL6I`IW1VH6IPIV1VH6IPH0WYfNQ9h000=V00F4WXBN0003IP0BIFIDIUAV
E6IDIUAUE6F4QEAU0U@00`1DD`0h00P700061dP00f/018A[Q4l3E01V05ADEE@0Q6IDIUEVEFH0QHB5
Q8F4QHBNQ9j406IWIPIV1VH6IPIVUHnESiF?UHnESiEVWYfNQ9jMWX@0I`IW1VL6I`IV1VH6IPIV1VH6
00IV1VH6IPIV1VH6IPIV1VH609j4WXBN26H048@0IFIDIUAVE6H0IUAUE6D3E003Q01D009D00=CE4h0
=`081`001PM70006Jfa[QF^50e@04EEDQ@1EE8F405EVEFIE08BN00<001NN001V1VH6IPIV1XBNQ9j4
WXBNQ9iW1P02U@04U9FEU@8000]W1YfNWIjMWYfN000C1P03W@0600D600eW1VL6IPIV1VH0W@1V0080
00FN00IV1P06IP0D06IVIVEVI@24IUAVE6IDIEB5E6D2E00405ACE3D020L000H7AP02J`05CV][04l0
15@028A[Q6^4E6IE0P07IP1L08EVIfIWIVMV1VH6IPIV08BNQ9jMWYd6I`IW1VL6I`IW1VL0W@1W1VL6
I`IW1VL0I`IW1VL6I`IW1VL6I`IW09fNW@IV1VH6IPIV1VH6IPH0WVH6IPIV1VH0Q00:IP0309f400:4
00T0E6IDIEAUE6D01U@0101DDe@b00P700061dH000E/CeA[Q@03E008EEB5Q8F405D2000FWVIWIVMV
IfIWIP1V1VH6I`IW1VL6I`8600<0W@000`H00`2MWP0<1P803@H019jMWYd:1P0509fNW@002@H0102M
WP051P0>IPIV1VH0IPIV1VIZIVX3IP0A08BNQ01DIUAVE6IDIUAUE6D01U@a00P700061dD000I[CU@0
Q4l4E003Q6^400F400<0EFH01FH046MV001WIPIV1VH6IPIV1VL2000C1VL6I`IW1VL6I`IW1VL6I`IW
1P0200P600EW1VL6I`031P0709jMWYfN00051P0PWIjMWYfNIPIV1VH609jMWYfN00IV1VH6IPIV1VH6
06X:IP0=Q9f4W@1VE6EDIEAU0005E004D`1CE30020L000H7A0001Fa?E5B500=D021EE5F4IUF5Q8D0
IfIWIVMVIfH6IPJ41VL6I`IW1VL6I`d60P091P03WP06014600RNWIjMWYfN008600P0WKJM]Yff00<6
00P0WYfNWIjM00@6019Z1VX6IPIV1VH0IPIV1VIZIVX2IP0BQ9j4WXBN06IDIUAV06EDIEAU0e@`00P7
00061d@000E>E4l0C`04E00406]EE0B400=VEFH00fH096MVIfIWQ9iV1VH6IPIV1VL0Q0IW1VL6I`IW
1VL6I`2MWYd6I`<600AWWYd050H0;P2NWIjMWYfN09jMWYfNWIjM00H0WIjMWYfN00IV1VH6IPIV1VH6
IPH0W@1ZIVX6IP0409f4W@8000R4[eAUE6D0Q0ID<0081`001PM3000SJeADE01EE5ED05AVE@24WXBN
IVMVIfH6IPIW1XBNQ0IW1VL00PH20003WYd600d600FNWIjM00031P0309fN00@04`H01`2MWYfNWIh0
0`004;JM]YffW@2MWYfNWIjMWP091P0AJPIZ1VZNWIh01VH609F?UHl00fH01FEVIFID008000]VE9iD
IEAUE8F4E00]00P700061d<000A[CeA?15@015EDEE@20004IUEV009V01QWIVMVIfIWQ9j406L6I`H0
WXBN00IW1VL;1P0>09jMWYfNWIjMWYfNWIh9000:WP00W@2MWYfNW@@001NNWIjML;78/LP0WYg8/G2M
WYfNWIjM00021P0>IPIV1VH6IPIV1VH6IP03W@0@JVIZSiAfSgH0IFIDIUAVE0J400L0E5@0Q6]D02`0
20L000H7@P0085A?E5@0E5D0EEAVEFH0IfIWIP2406H6I`IW1P2NWIh00PH01IjMWYfN00d600X0WIjM
WYfNWIjM0PH0102MWYd;1P<000ffWKJMWYfNWIjMWW2c00;801<0S`?80lP3b0?80k>M]YfNWIh000d6
01iZ1VX609jM06H6SiF?UHnD06IUIVEVIFH0WXBNQ9h20004Q8F4E2/020L000H7@@0045A[E4n4E8A[
EEAEE6IEIUD2IP0JI`25Q9h0IfH609j4WXBNIij4WXBNI`IW1VL31P0400H000H600D0WYfN00051P03
WIh001060P007YjMWYfNWIjM/`?8/LRab;78/LRab;780g1=L9fN00<600eZ1VH6IPIV1VH6IPH000:M
01FNWId0U8nDMYAfSgIVIFIDIUAUE00018@01P1DQ8A>E2X020L000H7@0003UA?E4l0JhF4EEAVEFIE
0VH05fMVIfH0Q9j4WVL6IijMWYfN09jMWYd000@600H0WYfNWIh:1P03WIh000P600<01PH04`H019jM
WYd3L0090lP3b0?80lP300?801@3b<S80lRJ]YffWKJM]YffWKJMWP@0012MWYfNWIj?UHnESiF?UHl0
0VH046EVIFIUIP2NQ01DI@25De@Y00P700061cl0011DCUA?E6/0E5EDIUD0EFIE0VH02FMVI`2NQ9j4
WP02000AQ9h0b:bJQ72IWYfNWIjMWP0030H00`2MWP0;1P05WYfNW@001`H0D@2MWYfNWIjMWYfNWIjM
WY[80lP3b0?80lP3b;78/LRab;7809jMWYfNWIjMWYd0JPIV1VH6IPIV09fNIVX0U8nDSiB?06EVIFIU
IUAVE6EDI@04E003DeAC02L020L000H7?P00:eA?E4mDJ`1DIUAVEHD0IfIWIVMV1VL0Q9j4WYC8[<Ra
b;78/LRaWYfNWIh00P0>1P04WIjM00H600`01PH609jMWYfNW@0=1P8002>NWIjMWP?80lP3b0?80lP3
b0?80lP3b0?80`2MWYfNWIjM000:1P0>JP2M06X6IYF?UHnDSfX2IP0AIFIUIVEVE8mDIEAUE6EDQE<0
9P081`001PLn000?CV]?E01DLGIDIUF506IE009V01MWIVMVI`2NQ9j4b:c8[<R/b:c8/LQ=L0030005
1P0000H00P021P0;00H009fNWIjMWYd00`0029BEU9FDUI@00PH01@2NWIh001460`006k:ab;780lP3
b0?8/LRab;78/LRab0?80lRM00081P0BJPIV1VH6IYjM06H6IVX0U8nD16H02VEVIFIDIWJ?E6D4E003
08@002D020L000H7?@0014iD08D3E00[SgIVEHF4IfIWIVMV1VL6I`1`/iW8/LRab;78/LRab;7809jM
WYfNWIjM00081P0409jM008600<0UID01iD00i@61P041P0309d600h60200WYfNWIjMWP380lP3b0?8
0lP3b0?80lP3b0?8L02MWP8030H03`2N00IZ1VX6IPIVJXl0S`02000?IVEVIHmfSgIUE6EDIHB502D0
20L000H7?@0015A[Jdl2E008MW40EHF4IUD2IP0NIfIWIYj4WP2b[<R/b:c8[<Rab;78/LRMWYfNWIh0
2`H00`2N000:1P04UIBEU08000@6I`IW1PH00ij41P081P0U09jMWYfNWIjMWYg8/LRab;78/LRab;78
/LP3b;:NWIjMWYfN00051P0PJPIZ1VH609h01VH6IVYVJVH0MYAfSgH0IFH0SgJ?MFD4E003Q5@002<0
20L000H7?002J`0>K5ADE01aS`25Q8F406H2000GWXBNQ9j4W[78/LRab;78/LRab;780k<00P0>1P03
00H600D600<0UID00iD02iBEU9FDUIBEU9FD00<600>E00H00PH01IjMWXBN008000JMWYfNWIh20003
WIh000800g005`?80lP3b0?80lP3b0?80g2MWYfNWKJM00`601l7JPIZ1VX6IPIVUHnESiB?U01V08mf
SgIUE6EDIHAD02<020L000H7?002J`0;CeA?E75fL@24QH@00VH07VMVWXBNQ9j4WXC8[<R/b;78/LRa
b;78/LRMWYfN00X600<01PH00PH02fL6IiEkUG^EMYEf00h000bEMYEfUGJE00IVWX@<1P0N09fNWIi=
b;78/LRab;780lP3b0?809jMWYfNWIh01@H07VX6JPIV06H6IPIV1VIZIYAfU7J?MXmUIWJ?MXme00=D
00<0Je@08P081`001PLl0006K4mDE5D00WH09P24QFIWIVMV1VH0Q9j4W[78/LRab;78/LP3b0?80ijM
WYfNWIjM0P041P0309fN008000FDUIBEU00P00031VL6008600>41PH020H07ijMWYfNWLRbShk80lP3
b0?8WIjMWYd01[JM]YffW@002@H03`061VX6JPIV1P2ESiF?U003S`0=IP2?MXmfIEAU06D0E00Q00P7
00061c/000ADCUA?0U@01P1aMUF5Q09V021WIVMV1VH0Q02/b:c8[<Rab;78/LRab;6cWIjMWYfNW@80
0`H02HBN00JDUGJEMP0W0005IPIV1VL01`H01P2NWIjMWP80012MWYe`/LP0b0?8/YjMWYfN10H01@2N
WIh000D6029Z1Yd0IPIV1VH6IVZ?U7JDMXmfSgJ?MXmfSgEUE8ACE5=D80081`001PLk0003CeAD009D
019EMWIfIXAWIVMV1VH6I`IWWYd2L00>b;78/LRab0?80lRbL9d;1P03W@H60086<0031P03Q00600<6
00<0WYd00P021P0509fNWIh00P002YfN0;>JWYfNW@071P0309d000H6012MWPH6JPIZ1VJESiF?UHnD
18l02`2?MXmfIHB5E6EC020020L000H7?0000dmDE002E080021VEHEVIfIWIPIV1VJNQ9jMb:c8/LRa
b;78/K>MWYfN00/600=V1VH0=00020IW1XBNQ9h01`H049fNWIh0WYfNWIjMWYfNWIh?1P0J09fNJPIV
1VH6IYF?U7JDMXmfSgH0E8mfSgD2Q004E5=DCQl020L000H7>P008UAaEG5fL@1EIUEWIVMV1VL6IijM
WYfNVW6Mb;780lRMWYd20005WYfNWIh010H01YfN1PH01Sl000>NQ0001`H0102MWP031P0:WKJM]PH0
WIjM]P/600JfWKJMWYd31P0HJPIZUIBESiF?UHnDSiAUIVD0E8F4IEAU0U@M00P700061cT000YDCE@0
MW5fE6IE0VH05fMVIfIWIYj4WX@6Q9j4L;78WK:aWYfN00D600X01P24WX@6IPJ4@`001FH0Q0IW00@6
00<0W@0010H0102NWIh21P0509jMWP002PH20005WYfNJPH00P005FH6IP2?U7JDMXmfSgJ?E6IDI@24
0003E003D`0001/020L000H7>@001DiaCgIa009f00eVEFMV06H6I`IWWX@0008600fMWYfN0g2MWW00
WIh000L600D0WP2NIP1800<600<01PH00PH00ih01P051P03WYd600<600@0WKH02`H07P2fWIh609BE
JPIZUIBESiF?UHnDSiAUIVEVQ8EDI@=D70081`001PLi000RE4eaL7Ia05EVEFJ4IfIWIYiV1VL6Iij4
WYfNWIjM1ifNW@L600@0WX@60VI;00041VL6I`8600<01PH01@H00`2M1P051P03WYd000X601JMWYd0
U9FDUHl0IYB?U7JDMXmfSgIV0VD015AUQ8@2E003DhA>01/020L000H7>@004G00LGIaSeEVEFMVWVL6
I`IW00@600P0WIh609fNW@8600<0WIh01PH01Id61PIV04l000IW1PH01P071P03WP0600D60`091P0O
09d01YFDUIBEU9D0UHnESiF?UHnDS`1UIVD0Q01DI@02E1/020L000H7>@005g5?MW1fLFIEIUEW08F4
1VH6I`IW1Yd0008600>MWYd010H00ih61P031P0300H6008600>40000D0001@IV1XBN00L600<0W@H0
40H07P2N1PIZUIBEU9F?UHnESiAfU7J?MXmfIUAUE8ADQ09D00=C00006@081`001PLi0003C`1a009f
00Z@EFMVIfJNQ0IW1@H01`2N1PH0WYd01@H00id61P071P03IP00058000JNI`H0WP071P03W@H600H6
00<01PH01@H200<601d0UIFEU9FD1P2ESiF?UHnESiAUIVEVE6ED05AUD`0J00P700061cT000QDCGIa
MW5VE@9V00b5Q6MV1VL6I`IWWP021P0309d600D600@01PJM0`H00`06IP1G0006IYEVWX@01PH00`06
1P061P801@H00`061P031P0M09FDUIBES`IV1P2DMYAfSgIVIFIDIHAUE01CQ4h06@081`001PLi0003
CgIa009f00Z@EI1f1X@0I`IW10H00ih01P021P0309h600D600>MWP000`H00id6IP1I0007IYFDU@2N
W@051P0300H600L600NM00H60;JM00D619D039BE00IV1P2ESiF?U@9V00]UIVEVE6ED05B5J`0H00P7
00061cT0019DL7IaMW40MY1fQ@06IPIW1VL200@600>N1PH01@H029h61P06Q0IVG0001VMfUGH600D6
0P081P06W@00WYfN1PH04P2EU9F?U@06IPIVJVH0MVIUIP9U00QD05AUE01CE1P020L000H7>002E00<
MW6?MP1fUGH0IiiW4PH00ih0U@02U@03W@IV05l000>?UID00`H01YfN1PH0WPP600FM]YffW@081P>E
00ZDUFX6IPIV1VH616H01fEVIFIDIH@00P000e@0000E00P700061cP0019>075fLFH0SgJ@IVMV06L6
Iih61P03W@H6008600T01PH6UI@6IPH0I0001GJEI`IW008600l0WIjMWYd01P2MWYfNWIh02`H02@1V
1VH6IPIVJP02IP08MXmf06EVE6D3E003Q02401H020L000H7>0004EAaMWIVEI1fUHl6I`H0WP2N00H6
00D0WP00W@041P06UI@6JP24I@000fL01P061P03WYfN008000L6002NWIjM00X600<0W@0010H04fX6
IPIVUHnDSfIUIXAVE6EDQE@05P081`001PLi000@MW5EEFIfT7IWIPIV1XBNQ0H600BNWIjM1@H01I@6
IPJ406L000EV07H6I`041P03WP0600H60P0:1P0FWIjMWVH6IPIVJVIZMYAfSgIVIH@0I@9D00<0Q6/0
5@081`001PLh0004E01VE@9V01<0MP1W1VL6IijMWYd01P2MWYfN00H60P021P03Q6H006L000F41XnE
00031P0300H600P600<0W@000`031P0L09jMWYfNW@IZ1VX6IPJ?UHnDSfIUWUAUE6D0Q1D020L000H7
>0001DifE6IE009V00/0IVMV1VH609jMWP031P0309fN00D600<0U9D00P000gIVQ0180006Q01DCUA>
700016MfUGH21P03UI@000T600NNWIjMWYfN00@001JMWYfNW@IV1VIZIYAfU7J?06J4IEAU0U@00hB>
000C00P700061cT000=aIUD00VH01VN406L6I`8600<01PH01PH00`061P021PBE00@0SiEV>`002diD
CU@0QDiDCUA>00MD00L0Q6/0CUA>01P000D6SiFD0004U@0300H600D600JM]YffW@061P0I09jMWYd6
JPIV1VJESiB?U6FNIFIDIEAUQ00D00P700061cP0011[MUEVEFJ4QH@0IPJ4WX@01`H039FD1PJEU9FD
U@2EISP000eDCP1>E6^4JhA[E4iD008001A[04iDCUA>E4iDCXA[E4iD08A>E1<000Z@MYEf1YBEU9FD
1PH01IjMWYfN00P6012MWP00W@IV1VIZIYAfU7J?0X@01UAUE6EDQ1@020L000H7>0002FaaIUEW09j4
WP030004WYfNW@D600D0[YFE1P06U@0308mV03<000]aL75`LG1aE024Q@0>E00AQ8F4QHB5Q8n>QH@0
JhE[04h04P000iF?U@07U@D600H0]P00W@081P04W@H61P80019V1VH609F?U8l0Q9iDIEAUE:PC00P7
00061cP000E[MUEVE@02Q009QFH6IPH0WYfN00D600h0U9FDUIBEU9FD09BNIC0000YDL75`LG1DCUA>
0e@00e=DE00@E08000>?SW0018h01XV>Jg1>E14000Z@MYFDUIBEU9FD1@H01P061PH0WPL60P003PIZ
1VJN00IVJWJDMXmf0h@01UAUE5B4SQ8020L000H7>0002XEaIUEW06L01VL31P04W@2M00@600BE[YD6
1YD010H0IVH]0005E4iDC`002U@00hADE006E08001Z5Q8F4QHB5Q8F4QHA`SZR>Z8jXSZR>Shj4Ja00
00>@UID00YD00`JEU@02U@860P051P0300H600H600>N1PH00PH036ZNWIj408nESi@0WPD000>5SP00
4@081`001PLh000=JgI[Q5EWIP1V1VH6Q0051P03WIjM00800PH02YFDUI@6IPJ4IV/X0008E024JhA>
E4l3E003JhA[00@018@AE0N400=`SZL028h01HU`JeA>00d000EVUGJEMP021P0600H0WIh01PH00`06
1P081P09IPIV1P2MQ9f400=V00EUQ5AUE003Q003SP0000l020L000H7>0002XEDQ@1WIPIW1VL81P03
WYfN00@600>E/iD00PH018E[Q@03E003075a00Aa00IfL7I`MP02E003071a00Ya20001F]DE5@000QD
03B5Q8F4QEAVE6H0Q@1VE6IDIUAVE6IDIUAV09j4WXBNQ8F408j]SZf>[Hj]SZf>Z8i`Q8EC3@000fJE
U@02U@@600>NWIh050H026X6JYjMWP1Z0VH02@1VE6H0QHBDSP0@00P700061cP000e[06]VEFMV06H6
IPIW00D60180WYd6CP1>E4iDCUA>E4iDCU@2000WL75`04iDCUA>E4iDCUA>E71aL75`LG1aL75`LG1a
L75`L@1DCUA?00QD00>408@00X@00`1DE005E00;05ADIEAUE6EDI@000X@20007IEAUE6EDI@05Q007
L8jWSZN>Y`04SP0308A[00d000]V1WJEMPIW1YfNW@071P0300H600D601BM1VX6IPIV1VJMQ9d0IVEV
Q6EDI@:400=[SP003`081`001PLi0009Q8EEIfH6Q0IW00H600Z508E[E4mDC`1[1@003H@0Q8F4QHB5
Q8F4QH@045@03hmaSg6?L8maSg2?L8m`S`08E00iQ5ED08@0E6IDIUAVE6ID08BNQ6IDIP2NE6IDIUAV
E6IDIUAV09j4WXB5Q72>[Hj]SZf>[Hj5Q5AC00d000AV1YFE0`H019jMWYd71P0300H600@60P051P0@
JPH0WYfNIVYUWVEVE6F4QA4020L000H7>00034j4JfIEIfH0IPIVE0=[01YDCXA[E4iDCUA>E4iDCU@0
Q01DCUA>E4mDC`=D00]C0000DeA>E5=DD`02E0@000EfL7I`S`020005CUA>E4h00P000e@000020005
L75`04h00U@0301DIUAVE6IDIUAVE0N400iDIUAVE6IDIUAVE6ED00J400<0Q8@00X@02ZMeQ02?JgI[
E4h=0005IPIf1VL00PH019jM1Yd61P03WYd600D601NN1PIZ1VH6IPJ408AVIFIUIUAUE6E[Q00@00P7
00061cP000iDQ8EEIfIDJeA>E6^5J`iD00>505@045@01P25Q8F400AD00h0Sg6?LHmaSg6?LHmaS`]D
03b?LGIaSg00E6IDIUAVE6IUIP2NQ9j4WXBNIFIUIVEVIFIUIVEVQ9j406EVE6IDWXBNQ024QH@0Q024
QE<<0004IfH600@600BN1PJM1PH00kH01P041P03WKH600D600IZWP00IVX2IP0706IDIU@0Q00@00P7
00061cT001j4E@1>E024CUA>E01DCUA>E4iDCUA?E6^4J`1?E4lBE00308B400F40P00;8m`Sg2?L8m`
Sg2?L8l0E5=DDeACE5=DDeACE71fL75`E6/0MXmfSgH0E01U10005IeDIUAVE6IDIUAV08ADIUAVE6ED
I@03Q00305AU00>400NYQ01C04iD00/00VH00`1f1P021P03W@H600L600<0WYd010H00`2M1P021P08
JPIV1XAZIP02IP09IFIDIEAUE8B900l020L000H7>@0034iDCUA?E4mDC`1[00QD00H0QHB5Q00?E003
EE@0008001aD0000E024QHB506EDIEAU08mfSgJ?08ma05AU2e@20003Q@0000800U@00dj?S`02S`YV
00<0Q6H016H0601VIFIUIVEVIFIU08BN06IDIXB5Q8F4Q@AD2`02I`:E0`H00`061P021P0300H60086
00<0]Yd010H00kH01P051P0?06X6Q6YVJVEVIFIDIEBX00l020L000H7=@003f]>E4iDCUA>E4iDCP1[
Q002000;CeA?E024JhA[Q00025@300B400<0E5@02e@01024Q00>E00308B400R40e@00e=D00020009
Q6^4CUA>E02?00]V0180Q9f4WH@0IFIUIVEVIFIDIU@2Q005IEAUE6D00X@2E0:400ACE4iD2`0016MV
UI@21P0300H600<600>N1PH00PH00ifNW@031P03WIh600<600iV06H6IVX0IVEV06IDI@9D00>>0000
3@081`001PLb0008K4mDCeA?E4l6E006QHB5Q8F40U@0:8F4QHB5Q8D0EEB5Q8F4QHB5Q8F405AVE6ID
IUAVE6IDIUAVE6IDIU@2000QIUAVE6IDIUAVE6IDIU@0Q9j4QHB506EDIEAUE6EDIEAU00=D00<0Q8D0
1U@046IZIVYVJVIZIVZ4WXBNQ9h8IP0@IFIU06EVE6IDIXAUE6EDQ@AD2`001PIWU@0600H600<01PH0
0PH00iffW@021P0309fN00@60P001VX6IPIVJP9V00MUIUAVE8F>00h020L000H7;`005W5=E6]DCUA>
E4iDCUA>E024JhA[Q6/2MP0306^400:400<0SP002X@00`1DE009E009EEAEE5EDEEAE009D00<0Q000
1U@026EDIEAUE6D00X@01@1DE5@000:400<0E5@01e@01e=DDe@0E5<00P001EA>E4j400=V10001FH0
WHBM00D00fH036EVI@1DIUAVE6EDI@ED00F4DeA[Q00;0006IgJEI`2M1@H00ifN1P021P0409fNW@86
00>NW@H00PH026ZNW@IVJVH00VH026F4E6EDIF^>3P081`001PL/000ULF/0CUA?E6^5JhE[QHB5Q8n>
Z8jXSZR>Z8jXSZR>[Hj:Sg1fL0060015E6H0IUAVE6IDIUAVE6IDIUAVE6IDIUAVE024WXBN06IDIUAV
E6H0WX@0E6IDIUAVE6IDIUAVE6J405AUE6EDIEAUE8D000MD0P005XBESiF?UHnEIPIV1VH6IVYV08BN
Q9h2IP0<IFIUIVEVE6IDIEAU0U@00e>5000;0005T9FDWYd01PH00ijMWP021P07WIjM09fNW@021P09
09fNJPIV1VIZ009V00MUWVEVE8F400h020L000H7:P0044e[CUA>E4iDCXA[Q6^4Jh@;SP0<YhjWSZN>
YhjXSZN?5X@01fIDIU@0E6H00P07Q00405AV00F40180IUAVE6IDIUAVE6IDIEAUE6D2Q0MD0X@05`1C
E5=DDeACQ024CYF?U8mZIVYVJVIZ00AV00B4WHBM0X@01@1VIFID008000=UE6D00U@00e=DD`0;0007
IYh0UFNNQ0061P0GWYfN09jMWYfNWIh609fNW@IV1VIZIYd00VH018AVE6D2Q003SP0000/020L000H7
:00025A[E4mDCeA?15@0:XF406^XSZR>Z8jXSZR>Z8j]SZf>[Hj]SZf>[HB5Q8F4QHB5Q8F4WXBNQ080
04=VE6IDIUAVEFIDIUAVE9j4WXBNQ9j4WXBNQ9j4WXBNQ01DIUAVE6IDIUAVE6IDIUAVE6H0WXBNE6ED
IEBNQ8EDIEAU00AD01J4QH@0IPIV1VH6IPIV1VH6IVX0WXAZ0VH03P1V001DQEAVE6EDQEB52`0016MV
WVL21P03WYd600D600nNWIjMWYfNWIjMWP2NWIh00PH03VX6IPIV06H0Q6IUQHB]3@081`001PLV000@
CEA>E4iDCUA?E4mDCh@0S`n>00jWSZR>Z8jXSZR>Yhj?00>40`0066IDIUAVE6IDIUAVE6IDIUAVE6ID
IUAV00J400@0E6H028@04`1DIUAVE6IDIUAVE6EDIEAUE0000h@0101DIE@4Q00C05ADDeACE5<0Je@0
WYfNQ9iV1P020005Q9j4WH@01FH00fEVQ005Q08000A[Q6^430001fH0I`H0WYd01@H06P2NWIjJWYfJ
WIh0WYfNW@1V1VIZIVZ4IVF40P02Q003M@0000/020L000H79000175>04l7E00PJhF4QHB5Q:R>Z8jX
SZR>Z8jXSZR>[Hj]SZf>[Hj]SYD3000JQHB5Q6IEIUAVE6IEIUEVEFIEIUEVEFIEIUD3IP09E@24WXBN
06IE00AV00JNQ9j406D2000QWXAVE6IDIUAVE6IDIUAVE6IDIP2NQ9j405BNQ9j4WUAU00=D01T0Q01D
05<0JPIZ1VH6IPIVWYfNQ9h0JVIZ009V00]U08BN06IDQHB5E00<00031VL6008600>M1PH01@H07YfN
WG03L0>?WK<0WYfNW@IZ1VH6IVZ4JXBNQ6D0Q@d020L000H78`004W5>JdiDCeA?06^4JhA[Q6^400j>
00^WSZR>Z8j]SZef000;Q00RIUAVE6IDIUAVE6IEIUEVEFIEIUEVE6H0WP1VE6IDIUAVE@9V00`0Q8@0
IUAVE6IDIU@20003WHBM00@000IDIUAUE6D7Q00308B400:400@0E5@00h@055ACE4iDIPIV1VH6IPIV
WVIZIVZ40fH026EVIFH0IUAU0U@00`1DJ`0;0009Q6N41VL6I`JM00800`H069fNWLP3b;40//RMWYd0
IPIV1VIZIYf400:400=DIF/03@081`001PLQ0013L5A>E4mD08E[QF^5Q8F4Q@2XSZR>Z8jXSZR>Z8j]
SZf>[Hj]SZf>[Hj]SZf>08B5Q8F4QHB5Q9j4WX@0EFIEIUEVE@0<IP0306IV00]V00<0Q00016H09fEV
IFIUIVEVI@24WXBNQ9h0WP1V09j4WXBNQ02>08BNQ6D0QHB5Q003E00PDeAC1VX6JPIZ1P06IPIV1VIZ
IVYVU7JDMVIDIUAU05@<0003IPIW008600NN1PJMWYfN008001fMWP380lP3b0>NWIjM1VX6JPIV1P2N
09j4IUAUQ00<00P700061b0000e>E4j4JeA?E4mDCeA?009D00D0Jh@0L00=SP07YhjWSZf>L0020004
Q024Q08000>400001X@02XF4QHB5Q8D0IUD;IP0509h0IUD03FH04`1VIVIDIUAVE6IDIUAVIFIUIP00
1X@00id00005Q003[JN]00:400=fSh@018@02eACE5=DCUAV1VH6008000QV1VIZIVYVW@9V00`0SgJ?
MVEDIEAfDe@;009V00@6I`IW0`H06YfNWIjMb;9`0lRab;78VIjMWVX6IPIVJVH018@015AUDh@<00P7
00061al000E`E6^5J`09E00Y08F4QHB?SZR>[Hj]SZf>[Hj]SZf>[Hj]SXF4QHB5Q9j4WP1VEFIEIUD0
0fH00eEVE@06IP<026H0102NQ00?IP804FH03FD009j4WXBNQ72W[Hh00W003JfW[JN]Yjf4QHB5E6D0
15@06U=DJPIZWP06JPIZ1VH6IYh0JXnDMYAUIUAU0WH00f/0000:009W1`H04IfNWIjMb0?80lP3b0?8
09jM00<600eZ1VH609j408AVE6F400`020L000H77@00375=06^4CUA?E4mDC`9D00J4JhA[Q6/2SP05
S`1`SXl01hh01jL0ShjWSYD00h@01HF4QHB500:400l0IUEVEFIEIUEVEFIEIUD01VH00`24Q002Q080
00F4WHBN000?IP0309d000MV00=UIVD00fH06fEVIFIUIVD0Q9f4WHBM0:bW[:N/YjbW[:N/SP02Q0=D
01aCE5=DDhA`09d0IPIV1VIZIVX0WXAVMXmfSeAU0U@00g24D`0;0005IVMV1VL01@H04P2NWIjMb0?8
/LRab;78/G2MWP@000YVW@2MIHADIEB430081`001PLK0004LEA?Q@ED00MEE8E[QF^5008001]D08B5
Q8D0[@25Q8F4L8jD08F4QHB5Q01EIUD00P00?HF4WXBNQ9h0IfIWIVMVIfIWIVMVIfIWIVMVIfH0Q9j4
WXBNQ9j406IWIP2?08nESiEV1VH609j4WX@0IVX02fH099j4WXBNQ9j4WX@0U:f/[Jb][:fW[JN]YjfW
[JN]Q8F4QHB5Q08000=f00H00PH04fX009iV1VJNQ6YVJVJDMXmf05@03@000hh6I`091P05WKJML@<0
0lP05@?80lP3WYfNWIjM1VH606YUIUAUQ00<00P700061aX000m=E4h0CeA?E4mDJhA[04l00U@01eED
Q6^4Jh@00P000f^4Q004Q00308B400J4015EIUEVEFIEIUD0Q8F4Q@1VE@0?IP0:09j4WXBNQ9j4009V
00d0SgJ?MXmfSgJEMYEf009V00X0Q9j4WHBMQ9f40P004Jl0002MQ9f4WHBMQ9f4WHBM00>4015`U:bW
[:N/YjbW[:N/SZbW[006Q00@05AC072NQ9j406H6IVYVW@AV00AfSgJ?0U@00f/0000;0005Q6MW1X@0
1`H079fNL<Rab;78/LRab4fNWIj406IZIVX0IUAUE8@<00P700061aT000=>E4l01e@08@1EE5EDEEAE
Q8F4QH@0E024QHB5Q8D0IUD0Q8F4QHB5Q006IP0IIfIWIVMV08BN06MVIfIWIVMVIfIWIVMV1P02000J
Q0001VH6IPIV1VH0SiF?UHnESiF?UHnESiD2000CQ9jMWXBNQ01V1VH6IPIVJVIZIP02000KWXBNQ9j4
WXBNQ9i`[Jb][:f/[Jb]YjfW[JN]009`00B4QHAU0e@01HE`WYd00086015Z1VX600IVJVIZSiAfSe@0
D`0<0005Q@06Q9h01`H00`03b002b00B0lS8b0?80lRMWYd0JPIV1VH00VH2I@03Q5@000X020L000H7
60003DiDCUA[E4mDCe@0E4l015@02UE[Q6^4JeEDEF/5Q00806IEIUD0Q8D3Q006QEEVEFIE2fH03@1W
IVMVIfIWIVMVIfH00P009VMVIfIWIVMVIfIWIVL0UGJEMYF?UHnEMYEfUGH0IPH0WHBMQ9d046H01@1V
08BM00:400<0[:`00j`04:N/YjbW[:N/YjbW[8j/SX@2E00>DeACQ4h0IPIV1VH6IVX5IP07SgJ?L5AC
Q00<0007SX@01XBNI`051P0M09jbb0?8/LRab;78/IjMWVH6IVYVJXAV06ED05<02`081`001PLF0007
E4iDCeA[Q@02E08003]DEEAEE5EDQHB5Q01DIUEV08F4QHB5EFIEIUEVE@24WXBN06MVIfIWIVMVIfIW
IVMVIfIW09h01VH6IP02000mWXBNI`IW1VL6I`IW1VH6I`IW1VNESiF?UHnESiF?UHnE00IV1VH609jM
WP06IPIV1VH6IPIV1VH6IVYVJP03IP0PaP00L:f/a:b][:f/[Jb][:f/[JN]Yjf4WP1UE6ED05<200<6
00mZ1VZNIPIVJXnDMXmDIE<03@001X@000JM00<600H0WKJM]P<3b00@0lS8b0?8WIh01VX6IPIV009V
0VD00eAUQ00:00P700061aD000aDJf]>E02408@0E4l4E00=EEB4JhADEEAEE5ED0003Q00>05EVEFIE
IUEVE@24QH@<IP0VIfIWIVMV08BNQ9j4WXBNQ01VIfIWIPIV1VH6IPIV1VH6IPIV1VH20003UHnE00D0
01L6IPIV1VH6IPJ4WP1ZIVYVJVIZIVYVJP08IP0G09f4WHBMQ02/[JN/YjbW[:N/YjbW[8h00h@2E00>
DeACE5>NQ9iV1VH6Q6X4IP06MXmeIE=D3P001fIWQ9j4WVL00P005ijMWYd0/LP3b;78/LRab9d6IPIV
1VIZ009V0VD2E003D`0000T020L000H75000G5A[QDmDChE[QEAEE5EDEEAEE8D0IUAVE6IEIUEVQ8F4
05EVEFIEIeEWEFMV08@0IVMVIfIWIVMVIfIWIPIV1VH6IP24WXBNQ9j41VL6I`IW1VL6I`IW1VL6I`IW
1@H20003U9D600/603QZ1VX6JP3:06H6IPIV1VH6IPIV1VIZIVZ4WXBNQ9j4WXA`[<B/[Jb][:f/[Jb]
[:fW08BNQ6EDI@=D10H20003WX@6008000MVU7J?E6ED00d0016>IfNN09jM00JNWKJM]W380`03b006
0lP3b;:N0PH01VX6IPIV1P9V00D0IUAUE00:00P700061a@000]>E000CUA[Q01DC`06E00<05AEE5ED
EEAEE5D00X@046IEIUEVEFIEIUEVEFIE05D6IP08IfIWIVMVIfH20017WXBNQ9j4WXBNQ6MV1VH6IPIV
1VH6IPIV1VH6IPH0UHnESiF?UHnESiEV1VH6IPIV1VH6IPIV1VH6IP1V1VH6IPIVJVIZIVX00VH03XBM
Q9f4WHBMQ9d0[Jb]1Z`02:N/YjbW[8h018@055=DDeAC1VH6IYj4W@2DMXl0SgEU0U@=0008Q6IWQ0J4
06L21P0G09fNWK>ab;78/LRab;6NWIiV1VH6IVX00VH016D0E6D2E0X020L000H74`00?4iDChE[Q@25
JeEDEEAEE5EDEEAVE6IDIUEVEFIEQHAVEFIEIUEWIVMVIfIWIVN406IWIVMVIfH6IPIW1PH000eW1VL6
I`H0UI2ET9F@00:E00X0I`IW1VL6I`H02iD01YBEU9FDU@l602ZN00IZ1VH6IPIV1VH6WIjMWYfNQ9j4
WXBN[;:D[G2][:f/[Jb][:f/WX@30007Q8F4IEB5E0041P0<WYd6SiF?JVEVMP1D3P000hEV1P021P03
W@H600<600<0]P<00lP02P?8b<P3b9fNW@021P0;JPIV1P1VI@1DIE@02P081`001PLB0004CUA>E0=[
00AD08A?1U@02P1DEEAEE5EDEE@2E@0>05EVEFIEIUEVEFIEIUD2IP08I`25Q01VIfH2000QQHB506MV
IfIWIPIV1VH6IP2?UGJESiF?UHnESiD01VH6008003>?UHnEU9FDUIBEU9FDUHnESiF?UHl6IPIV1VH6
IPIV1VH609j4WVH6IPIV1VH0Q9f4WH@00P06IP06WHBMQ9d01j`00jN/Y`02Q00CIEAUE6ECE5>4IPIV
1X@6IYAfU002IP04IHm`E0h000N4IfH6IiiW00H601T0CLP3b;78/LRaWYfNW@1V1VIZIP1UIXAU009D
2P081`001PLB0006E4mDChE[0U@065EDQ@1EE5D0QEAVE6IEIUEVEFIEIP1VE@9V02EWIVMVIfIWIVMV
I`2NQ9j4WXBNQ9h01VL6I`IW1VL6I`IW1VL600bE00VDUI@0U9FDUI@02iD02YBEU9FDUIBEU00<1P04
09jMWP8002iZ1VX0JPIV1VH6IPIV1VH6IVX0WXBNQ9j/[G2NU:f/U8B]U9j405AUE6EDIHAf0PH02006
1PJDUHmZ0VH00gJ?000>0004Q@06Q08600<01PH010H00ifN0003b00E0lP3b02NWIjM1VX6IPH0IVF5
06ED00X020L000H740005DiDCU@0E6]DCeA?E@24JhA[Q6/0E006E@0jIUF4Q6IEIUEV06MV08B5Q8F4
QHB5Q8F4Q@06IPIV1VH6IPIV1VL6I`IW1VL6I`H0UHnESiF?UIBE00L601/0UIBEU9FDUIBEU9FDUIBE
U9FDUIBEU9D01VX00P001iFDUIBEU9D00P0099fNWIjM06H6IPIV1VH6IPIVJVIZIVZ4WHBMQ01UWHBM
Q9f4[`:40P03Q003IEAU009D00Y[MW06IPIV1XBN0VH016F?E5@?0006QFH6I`IW1PH06IjMW[78/LRa
b;6J09jMWYd0IVYVJVEVE8@00U@:00P700061a0000IDCe@0E002E00<EEAEE01[QF^5JhF40`009hB5
Q8F4QHB5Q8F4Q@1WIVMVIfIWIPIW1VL6I`24WVL0I`1W1VL6I`071PP04@H3U@07U9FEUIBEU00;U@1:
U9D0UIBEU9FDUIBEU9FD00H0WIjM06X6JPIZ1VX6IPIV1VH6IP24WXBNIVYVWXBNQ9j408BNIFH0QH@0
E6EDIEB?L0IZ1XBNQ9h2IP03MVED00l000>41VL00`H00`061P031P0;WIjMb;:?0lP3b9d00`0039fN
WIiV1VIZIFJ4I@/020L000H73`002UA>E4mDC`1?E4l2E008EEB4J`1DEE@4E@801X@06`1VIfIWIVMV
IfIWIVMVIfH0Q9j4WXBNQ9j4WP020004I`IW1P8000f4WXBNQ9jMWXBNWIh000l601D0UIBEU9FDUIBE
U9FDUIBEU9FDU@0010H09P2DUIBEU9FDUIBEU01Z1VX6WIjM1VH6IPIV1VH6IPIVJP2NQ9d016H01ij4
WXBMI@000X@015AUE002Q006E6EC05=f0P001FH6IYiV00:400=UE5@03`001VH0Ih@6I`H600bNWIh0
SYg8/K>aWYd31P0;JP0006IZI@1DQ5@02`081`001PL?0003C`1?00=D00D0EEAEE002000`MUAEE6ID
IUEVEFIEIUD0Q8D0IfIWIVMVIfH6IPIW1VL6Q9j4WXBNQ01W1VL6I`IW2`H02@2NWIjMWYfNW@020106
0P02U@03[YF^00JE00<01PH01@H01`2^U:jD[Y@00iD019BEU0061P80039Z1VX6JPIZ1VX6IPH0WP06
IVYVJVJNQ9j406FNQ9iUIVEV08EDIEAUE8mZ1VX6Q0IVQ@9U400015AW00031P0300H6008600fM]Yd0
WKH3b9fcWIjM00D600QZWXAZMVJ4I@/020L000H740001dmDCeA?04l00U@02WIaMW40E5EDEE@2E@0G
IUEVE@1fSgH0IVMVIfIWIVMVIfIWQ9h00`005PIW1P2NI`IW1VL6I`IW1VL6I`IW1VL21P0309d600D6
00D0WIjM000C1P8040H02P2DUIBEU9FDUI@200091VX6IPIV1VH6008000UV1VH6IPIV1VH00P003VYV
JVIZIVYVWXBNIFIU0h@2I@0>E6ED05AU05AC1VH6IPH2IP03Q6ID01000VH01VMV1VL600@600fN00H0
WYd0/IjMWYfN008600YZ1VH6IYeVSgJ40U@:00P700061`h000A/CeA?0U@0<P1[EEAfLGIaMP1VEFIE
IUEVEFMfT7J@MY1fU@1WIPIV1VL6I`IW1VL6I`IW1VL600IW40H00`061P0;1P0300L600T600/0UJjE
[YF^UJjE000C1P0309FE00l600<0WIh00PH00fX6JP02000=1VX6IPIV1VH6IYj40002IP0DI@24WVEV
IFIDIE@0E6EC1VX6IPH2IQ8000EDIfNNQ0031P04W@H61P800PH02YffWI^MWYfNWIh41P09JPJ4JWJ?
E8EC00X020L000H73P004F]DCeA?E6^4E01aMW5fLEED009E02iVEFIE07J?MY1fT7J@MVMVIfIWIP24
1VH6I`IW1VL6I`IW1X@0I`IW1VL6I`IW1@H00`2N000?1P0509jMWP001@H03YBEU9FDUIBEU9FDUI@0
4@H019jMWP051P0UJPIZ1VH6IPIV1P2NW@1V09d0IPIV1VIZIVYVJVJN06IUIVD0Q003I@04E6D0I@80
00]CE6H600IV06F4D`0@0009EFIWQ9iW1VNN008600<01PH00PH05ijMWP2NW@2MWYd6JPIV1VJMMXmf
I@1D00X020L000H73@000f]/C`03E00[08E[075fLGH0IUEVEFIEIUEW091fT7JEMYEf06H609j4WX@6
I`IW1VL6I`061P03WX@000X600D0WYfN000B1P0A09fNWIjMWP2^UJjE[YF^UJh019D00`061P0<1P0:
09fNWIjMWYfNW@802`H05YfNWIjMWYd6JPIV1VH6IPIV1P2NIVX2IP06IFJ4IVEV0VD20008Q6ED06ZN
IPH2IP03E00000l000aDIfH0Q0H01P2NW@041P0:WKJM09ff1YjMWP@600UZ1XBDMXmDIE<02P081`00
1PL=009[00=DCe@00f/028@0MW5fE5ED0UD3000YEFH0SgJ@MY1fT7J@08F4WX@0IPIV1VL6I`IW1VL6
I`IW08BNQ9iW1VL010H01ijMWYfNW@004@H02@2MWYfNWIjM00031P0;09BEU9FDUIBEU0002@H00`06
00060009WIjMWP2NWIjM00801`H05VX6JYjMWYfN00IV1VH6IPIVJVIZ06X4IP04IFJ4IP9U00YDIHB4
DeAC08@60VH00fF4000@000;EFIW09h0U@H0WIh01@H01P2MWPJNW@8600fNW@1Z1VH6IYefSgIU009D
2P081`001PL=000=JeA?06/0JhEDEG5f0002MP0LSgJ@MY1fT7H0MYEfUGH0I`J4WXBNQ9h01VL6I`/6
00L0WYfNWIh000@600NNWIjMWYd0014600<01PH00`H0102MWP0=1P0809jMWYfNWKH200/600@0WIh0
20H400@601FNWIjMWYd0JPIZ1VH6IPIV1VIZIVX016H03P1VIFID08B5E6F4WVJN0VH00e@0000@0004
IfJNQ0>E00<6WYd01@H01YffW@JMWP8600<0WP000`H02FX6Q9AfSeAUD`0:00P700061``000m`Jf]D
CeA[04mEE5ED07401WH05HmfSfIWIVMVIfIWQ8F41VL6I`IW1P020003Q9j400<000=W1VL00PH06Yj4
WYfNWIj1VP2NWIjM/k:NWIjMWYfNWIjM4PH200/600<0WYd00P021P03WYd001d601L0WIjMWP06IPIV
1VH0IP1V1VIZIYeVJP04IP0>IFIU05AUQ6ED06]D[h@2IP03E000014000YVI`06MYF?U@2N1@H00ijM
1P021P0AW@H61P2NJPIV1P1Z08mfIE@02`081`001PL<000TSf]/E5EDQEAEE5ED07J?MY1fT7J@MY00
1VH6I`IW1P2N00IW2@H02hBNQ9jMWYfNW@2I009`01>N0<Rab;;8/LRab;6?WIh0WYfN01L600>NWIh0
20H200T60P0>1P03]P0600/600H0WKJMWYd91P0F09h0WYd6IPIVJVIZIVZ408AVE6EDQ@9D00>4JP00
4`001UBN00IW00>E00<6W@0010H00id01P021P03WPH6008600bM1PH6JP1V07IVE6D;00P700061``0
0V/02dmDC`1?EEAEE5Ea00If04J?MY1fIfIWIVMVIfJeIPIV1VL6I`IW1VL6I`24WXBNQ9j4WP38/LRa
b;78/LRab;78/LRab;6c09jMWYfNWIjMWYfNWIh00PH2000=1P00WIjMWYfNWIjMWP021P0409jMWPh6
00>M00H02@H00ifN1P0;1P0309h6008601L0W@1Z1VX6IPIV1VJNQ9f4WP1ZIVYV0004Q00:E6EDIHAD
DfJ4IA<000b408EV1VJEU9EW1Yd41P03WPH60086016M1PH6JPIZ1VH606YVSeAUE00;00P700061``0
01i/CeADEEAEE5EDIP2?MY1fT7J@MYEf06L6I`IW1VL200031VL600d60P007LRab;78/LRbb;78//Ra
b;;8/LRab;78/LRab9d001@600H0WYfNW@031P05WYfNWIh03`H00`2M00071P04]YfN00T60P061P03
09h600P600H0WIj4WX@30006Q9j4WP1V0VD01EB505B401D000J408@6I`H3U@8600<01PH00PH00id6
1P021P0300H600H600QZ1VIZ06IDI@/020L000H73@0015A?E4l2E006EEAEE7Ia17H0?8mfT7H0IVMV
IfH6IP241VH6I`IW1VL6I`IW1XBNQ9j4WXB>/LRab;78/LRab;78/LRab;78/LR4WXBN00T600MW1VL6
I`IW008000FNQ9j4WP091P03WYd000@600<01P002`H019fNWIh41P05WIjMWYd020H00ih01P071P03
00H60086019Z1VH6IPJ406IZIVYVJVH0Q6H2I@07E6D0E5=DI@0D00:400R5IPIVUHnEI`H600<01PH0
0PH030061PIZ1P06IP1VJP9V00=DIE@02`081`001PL<000PK5ADE01DEEAVEFIfT7J@MY1fUGH0I`IW
1VL6I`2406L61P0O09fNWIjMWYfNWG2ab;78/LRab;78//Rab;;809jM00051P0:SgJ@IVN?UHnEI`D6
00<01PH010H03;JM1PH0WIh0]Yd01P<000>MWPH020H01`2MWYfNW@000PH01P2NWIjM]P8000JM]YfN
W@041P04WIjM00X600>M1PH00`H2000@U9EZ1VH6IPIVJVH0IFIUIP9U00ADIEAV5@0019h01VL21P03
U@H6008600<01PH03`H01X@6IVX0S`9D2`081`001PL<009[00M?E6]EE5ED009E01EfLGIfSgJ@MP1V
T7J@00IV08BN09h00`0010IW1VL61P0@08BNQ9j4/k78/LRaWXBNQ0D60`005HA?EEAfLHmfIWJ@MYF?
1VL6I`IW00041P800`H01`2MWP06WIh00`H02`1VIUAVQ6J406L000H601/0WYfNWIjM09fNWIh0/YfN
WIjM1P2NWIjMWP002`H079d01PIZ09BESiEf1VH6IVYVJWJ?MVEDI@1D05@E000:EFIWIPIV1P06I`86
00BN1PH02@H00fX6JP030006IVYVSgIU30081`001PL;0008L6a[05B5E6H3MP0FT02@MY1fUFH6MYF?
UHnE09j4WXBN00/60100WYfNWIh0WYfNW@2MWYd6200044l0E5EDIWJ@MVN?UHnET9D51P03WPH60086
00<01PH00`H01YfN1PJMWP@60fH02Ih0WUAVE6IDIP02E0<600h0W@06WYfNWIjbb0?80`?800Zbb<Q`
WIjMWYd02`H0102MWP021P0FUIBEU9FD1VX6IPIVU8nDMVIUQEAUE1D000UEIhl6I`H0U@000`H019d0
1Yh<1P09WIj41VIZMXmD00`020L000H72`0018i[J`02J`08075fLGH0IUD2IP0DIfIWMY1fUGJE00H0
WX@0I`IW1VL51P05Q9j4WX@00PH00`24WP021P03I`IW00`0019DCeADEG6?MXl0T7JESiF?1VL31P04
W@H0WPH600>M00H00PH00ih61P021P03Q6IV009V0X@01@1DQ8AE00AD01b4E5@6I`H0WXBc/LRab;78
/LRab;78/K>MWYfN3PH06YfNW@2DUIBESiD01VIZIVYfSgJ?06D0IE>450002f]E091f1VJEU9EW00<6
00>MWYd02@H036X6JYj406IZIXmeE0`020L000H72`001g1DE5A[Q@000WH0291f05EWIVMV0VL039F?
UHnET0IW1P2N00P600<0W@H010H00id01P031Q0000e?E6]EE8mfT02ESiF?00>E1@H00kJM]P0;1P03
WPH6008602>MUFH6IVN406JNQ9iDIUAVE6H0Q@1EE8D6WYfNWLRab;780`03b00A0lS8b0?8L;JM]Yff
WKJM00H00P001ijMWYfNW@H00YD04YBEU01Z1VH6IP2?U8nDQ01DIAD000IDEIF?UFL4U@0700H61YjM
]P0<1P09WIiZ1VJDMXmD00`020L000H72`009Xi?E4l0JeEaMW5f06IEIfIWIVL0UGJEMP1W1VL6I`1W
1VL6I`H01`H00`06I`0D000BE4n407IaSeF@MY1fUHnE00IW0`H00id01P061P05WPH61P0010H016JD
W@H2IP0<06IVIP24E01DIUAE0U@4Q00OJh@0Q6^4I`J4V[78/LRab;78/LRaVYfNWIjMWYfNW@061P09
09fN00JDUIBE00<000bE001VU7J?MX@0IE@E000;J`1fT7H6MYFDUI@00`H00`2NW@091P0<JPIZWX@6
IVYfSgED30081`001PL<0004K4mEJ`9E0gH03fIWIVMV1VH0Q9j406L6I`051P05Q9j4WYd010H01Ij4
WX@601P000EDEHB5E@02IP0608nESiF?0YD020H0UID0WIh01@H00`061P021P0300H6008600f?1PH0
WIh01VH6IVL000=V01AEIP2NQ8F4[Hj]SZR>Z8jXSXF4Q@8600l0b;78/LRab0?80ijMWP002`H200D6
00h0U9FDUIBESfX0JXAVE1D000MDIUFESiF?00FE10H019ffWKH81P0809jM09@6IY@2MPd020L000H7
2`02J`04E4mEE09E00AVJ`240P005FN4QHB5Q9j406L6I`IW1P2NQ9j4WP041P04U9EfE1P000m>E5B4
06IEIWJ@MYF?UHl00P001iFDUI@6W@0010H00`061P021P03WP2N008600iVWP06IP2M1VH6IVMV00AV
00AEIU@00X@0102]SZL5SP0306^4009D01=?E4l6Q9jbb;78/LRaWYfNWIh000T600<01PH00`H0202E
U9FDUGJE0VH016D0E6DE009D00eVMYD0UIBEU9FD00JE00D01PH400031YBE009V0WH>00P700061``0
01IDCeEDIUEV08F4QH@0IPIW1X@0I`IW0PH01P24WXBN00D600L0UIFE07IC01P00U@02P24IUEWIYF?
UHl8U@<600@01PH60P041P05WIh61P001PH03@2DUHl6IPJ41VIWIVL00VH05EFNSZf>[Hj]SZf>Z8n5
Q5EDEEAEJ`02000>WP00WLQ`WYfN09jMWYd81P0300H600<600l0U9FDUIBES`IVJVEVE6D04`02E005
IUFES`H01YD01006UJh51P06]YfNWIjM0`H016ZESfX2S`9f3@081`001PL<0005CUA?EE@00UD3Q006
IVMVIfH60P0036L6I`IW1VL6Q9j4WPD600>EIX@00U@01E=DCUA`01D000E[E6^4E@02IP0=07JEMYF?
UFNEU9FDU@041P04WIh0WP@600JMWPH6SiD21P0>I`IV1XnESiF?1P2NIVL2IP0@09j4QHB5Q:f>[8jW
SZN>L0:401=[05ADJhA[Q4mDI`2MWYfNWIh0008600<0WIh01PH00`2N1P031P0:09FDUIBEMYEVJP9V
0VDD0003CeEE009V00bEIYFDUIBEU9D0UI@51P0509jMWYd00`H026ZEMY@0U7Ef3P081`001PL<000D
E01EE6IEQHAWQ6MVIfH6I`1W1VL41P8010H0302MWYd0U6Z4QEAUE0<000=>00004@001DiD05F4009V
00UWIYF?UIBE1P000iD00jh01P031P04WYfNW@<000FMWPH6U002U@03[P0600@600^DUIBES`2MWVH6
IP03000JQ9j4[JN]Yjf>[Hi`06IDIUAEE01DE@1EE5D21P0709h61YfNW@071P8000<6WYd01@H03YBE
U9FDUHl6IVYUIUB54`02E@9V00<6I`H02ID00jh61P041P03WIjM00D600BDUHl00Xl00e@0000<00P7
00061`d000IDJeEE05D2IP0>06IWIVMV1VH6I`IW1VL81P05WX@6IVX00X@02eAU05ACQ4j4JeA>0100
00IDCeEDQ5D2IP0<IgJES`1W1YBEU9D01@H01IfNWIh000<600RDUFJEU9FDU@H600bDUHl0IPJM1VIW
IVL3IP0908B5Yjf>[Hj/00>4015EE5EDE@1EE01DEEAEE0IWWP031P0309d600L600BNWIh60P021P09
09BEU9Ef06IZ009V0VD00gD0000@000CE6]VEFMV1VJEU9FDUIBEU9FD00051P0309d600@600QZUGJD
MXmeE0h020L000H73@001DmEE6IE009V00aWIP1V1VL609iW1VL21P03WP0600<601P01PH6WX@6Q9j4
IVD0E6D0QE@0De@0Q4h=000<JeADIP1WIVMVUHnE0`H2U@03[P0600D600<0W@001@H02Jh0UHn^UJjE
1P0200BE10H09fL6IPIV1VH6IP24WXA`Q:f/Q@25Q8F4IUAVE6IDIP1VE01DIP2E00051P03W@H600D6
00<0WYd01ID0302EU9FDU@06IVX0IQ@000M[05EWQ0IW00861ID01@JEUJh000H600<01PH00`H0202D
UHl0MXmD3P081`001PL=0006E4l006IE0fH03H@0IVMVWX@6I`IWWP002@H02X@606YVQ6EVQ6D4E009
DeACE4j4JeA>00X000ID05EEIUD2IP09IgH0I`IW1P2E00L600>N00H01@H01IFDUGH0008601`01PJE
U9FDUIBEI`IW1VH6IPIV1VIWQ9j4IhB518@05eEV08AEIUD0E5ED05AEE01EQ5F@IiiW00X600T0W@JD
UIBEU9D00PH02P2EMP1fJVJ4IH@B0003CUAD009V00iWIPIW1YBEU9D01YBEU0L600bM1PH6JPH0UGIZ
08l@00P700061`h000AD08EE0VH02VMV08@0I`J4WP021P03WX@6008600>M1PH01`H01FH606IU0080
00B5E6H01e@01F^4DeA>00L000]?QEAVEFMVI`06I`051P03W@0600@600>NWIh01@H3U@03U6ZM00D6
00<0[YD01ID41P80011W1VH609iV1VJNQ6MVWX@016H03@1VQ6IEIXB506IEIUD00VH01@1W1YfN00<6
00>N1PH00P041PFE1@H01hnESfZ4QF/04P0025B4IUEWIPIW0PH4U@860YD00`061P041P0309h600@6
00L0U9EVJWIU00l020L000H73P001DmEJfIE009V011WIXF4IhBNQ9j4WXBNI`J410H02@0606JNIYiV
0002IP04Q024IP9D00>4E5@00P002eACE4iDCUA>Q6]D00D00UD2IP0=IgJEIPIW1VL6WIjMWP020005
WIjMWYd01@H01iBESiF41P001@H029BEU9FDUIBE0PH04fL6Q9iV1P06IPIVIhBNIVMVQ@000fH07EEV
Q8EES`25Q6IEIUEV06H0T6JEI`H0WYfNWIh000D600>DUI@00`H02FX6MYEfU8@0E00C0003JhAE009V
00UWIPIW1YBEU0000`H00iD61P041P04WIjM008600UZ1P2E06H0Se@03`081`001PL?000GE8EEIX@0
IPIVWXBbQ;:aL9W8WIh0WYd010H03@2M06X6IPIVJXAZIYh00VH02fEVE8F4QHB5E8F400ED00D0Q@1f
J`02E@07QFIWMYF?U@041P04WIjM00P600@0]Yff0`H00iD01P021P03W@H600<6012E[YF^UJjE[PH0
U@2MWYd00PH096L6IiiW1VH6IPIV1VIWQ6MVT7J@MVMVIfIWIVN4IfH600JM00<600`0WIjMWYd0WIjM
WP041P0309@0009V00=U0000400028F4QEF5IPIW10H00`061P041P06WKJMWYd60`000ih000020007
W@1V1VIZMP0@00P700061`l001ID06IEIP25Q024QJbb[;://P2bQ9j410H02ij41VH0IVYVJXBM009V
00<0IVD00X@01FIDQ8@000B415@20007DeA[JeAVQ0020008T7JE00IW1Yd20003WPH600P600RMWYd6
JPIV1P800PH00ifN1P031P06U9FDUIBE0PH069BEU9EW09fNI`IV1P06IVMVIfH0IVL0I`9V00<0SgH0
16H03HF4IfJ506MVQHBNQ00010H00id61P041P03WVX0008000N41VIZIP1U01@000b?Q8B4IfIWIPH0
UI@61P04WYfNW@@600VN1PIZ08BN06X00VH2IA0020L000H73`005F]EEFIEIhB5Q9j//[78/LRbb72N
W@031P0400H6WP8601`0WVX6Q9h01VIZIVYV08BNQ700[@2]0:f4QHB515@02F]EEHD0UHnES`031P04
00H600/600<0W@000`H00ih61P031P08WYfN1ZjE[P061P05[YFEU@0010H08P061YiW1VL6I`IWWX@6
IVMVIfH0IVMVI`2N00IV1XBNI`021P0300H600<600<01PH00`H00id61P031P04Q01VJP9V4P002f]`
Q8F4IfH6I`0600>E1PH00ifN00051P0300H6008600RNQ01V1VH0E10020L000H73`00:6]?05EVEFH0
QHB5[;://Zbb[;:aWXBNQ0IV1VH6IPJ41VH6IP1VJP03IP03IFJ400:400J/SZb>[8h4Q0AD00`0E6J4
QGJEMP1W1P031P0300H600@600<01PH00PH02ifN00IVUFFNI`IW008600P0WIh0UIBEU0H602jEU9D0
1VL6I`IWWX@0IPIV1P2NQ9iVIfIWIP1fT7IWIY1VIfIWIXD01VH6I`IW10H00`061P061P06JPIVWVIZ
0VHB0004CXn>SP:400YVIfH600JDUIBE10H00`2NW@061P09WPH6JPH0WVIZ009V00<0E0003`081`00
1PL@009D01=VEFMV1VI`[;:/b;78/LRbL9d000`601IZ09BEIP1V1VIZIVZ4WW2][:f/[Jb]0`003fID
05AEQ8D0UHnEI`2M000>1P03W@JM008600BEU6IW1PH200061YF^UK<01PH00`2M1P071P08WYfNI`IW
1VL2000A1VL6I`J?UHnES`J?UHl6I`H00P000fL6Q0031P03W@06008600FN1PH600051P05WVX6IXD0
4`001Xn>[HB5IP9W00<6Iih00iD00jh01P021P040;H000/600NNQ0IVJWIU014020L000H74@004gI[
IUEV08F4L:bb[;://ZbbCIh01@H06`1Z1YBE00IV1XnESiB4WFIZIP24WHB][:f/[@04[0B400=`J`00
0X@026MfU@2NQ9iW0`H00`061P031P03W@H600L600V?UHAWIPIV1VL00PH00`061P031P0300H600<6
00P0WIjMWYd0I`8002jNQ01W1VL6I`IW1WJEIPIfUGJEMYEfUGJEIVL01VH6I`J4WVL6Q9h608BN1Yj4
0`H01VH6IPIVQ1@00hh2000:IfIWIPJ4UIBEU0@600<01PH01PH020061PIZ06JN0VH00fEfD`0A00P7
00061a8001=fQ6H0QHBNQ;:/b;78//RMWYd000@600JMWYBEU9D31P0UUIBEU9jMWXBNQ9j4WXC8[<R/
/Zb][:f/WXB?JhEEQFH6I`24WP051P03WPH600860P071P8000bEQ:fDL6H6I`J41P051P06WIjMWYfN
0`H200H600BNWIjM1@H05HnEUID0UI2ET9F?UFNES`2406L6I`031P0<Q9h0WYfNWIjMWYfN0PH20086
00=ZQ@004`001V^XSZfDWP8000QW1YfNUJjE[P8600<01PH03@H01`06[`IVU5@04P081`001PLD000L
JfIE08B50;://Zbb[724WXBNQ0H0WP2EU9D01P8001:DUHl6IYeV1VIZIP24/Zb][:d5[004L8B4L09[
00YVQ6MV1XBN00IW0`H00id61P021P05WPH61P0010H019d009h6[00=IVMVWX@6I`IWWYd0W@061P03
00H600@600AWWVL60P00;VL6I`J?UGH6IiEfUGJEMPIV1P2N00IW1VNNQ0J4W[6NCLR4/hBN1P2406H6
IPHD0005CXB>[8h00X@00ij400020004IiFDU@8000>MWPH020H00`061P021P0400IV009V00=eQ000
4@081`001PLD000?QEEWIPIVWXBbQ<Q`WYd000H600FM1PH01P03U@03U9D600<601L01VX6IPIV09fN
0<R/b:c8[<R/L8B?J`02E@07QFH6IijMWP041P030;H6008600FfW@H6]P041P0>09h61VK4[<B/b:c8
[9h3008600>N00H00`H00`061P0A1P03WYfN00D600`0UIFE06L6I`J41P041P0>WYfN/LRbb;78//P0
WYd31P03JP1V01<000]>08j]SYj4WX@6I`041P05[PJfWKH02PH00id61P031P05JPJ4JWH050081`00
1PLF000=EFIVIhB5IXF408BNQ00;1P0[U9FDUI@6JPIV1VH6IPIV1XBNQ9j/b:bb[:d0WHBML6]DIUEW
IPJ41X@6I`021P0709d61P2MWP071P0KWYd6IVZ>[Jb][:f/[Jb]IPIVWX@0I`IW1VL000D600>N00H0
0`H0;9iW1VL6I`IW1XBNI`IW1VL6IiEW1VL6I`IW1VL6Q9j4L;6b/LRab;78/LR40P0019iV1VHC0003
CXA`00:>00V]Q8D0IfH6I`0010H00ifN1P041P03WPH600860P021P09JP1V1VJ4IGI[01<020L000H7
5P003VIEIfJN00H0WX@609jM2@H7U@@601JN1PIZ1P2NWIh0WYfc[Ii`b9fNWIi[0UD01fMV1VNN1P00
10H02@2NW@JMWP06W@041P0300H0008601=VSjc8[<R/b:c8UIjM1VH600H000@600<0W@000`H00ijM
1P021P03W@0600D600@01PJN1PH2U@0300H600801@H04YfNWG2bb;78//Rab8BNQ0IV1Q@000]>QF^]
SZf/WVL6I`031P07W@H61P2fW@051P03W@H6008600>N1PH010H01FH0Q9AD01@020L000H760002eAV
IVMV1VJNIPIW008600@0Q9jM10H01`2DUIBEU9D00PH08fX609jM06X6JPIV1VJN00JMWYd0IPIVQ5AV
06MV1VL6I`JM008600VNWIh0WYd0WIh00PH01P061PIZ1P9V00^>b:a`U<RDWYfNW@071P0XIVMV1P2N
I`J4WXBNQ9h0WYfN00H0WYd0I`IW1VL6Q9j409BEI`J400@60100WXBN0<P0b;6N/Ij4WVIZ5`003XB>
SXj]09iV1VH6IPH00`H0102N1Yh51P0=WPH61YjM1VX6Q0IVJP02Q003E8@001<020L000H76@0025B5
IPIV1VL01@H00`2MWP021P:E0P0019BEU9D20006U9FD1Yd01`H00`2M1P021P0709jM1VX6J`02IP04
QFH600@600nN00JM]YfcWKJM]Yd0WIh01PH03@2]VYjMWYd609jMWP0020H05P1V1VH6Iij4L8BN/LRa
b9W8/IjMWYd61P0508@0Q9h010H00id61P051P08Q9j408BNQ0021P03Q00001D000e?E8B]SZf/WX@0
I`IW00H600BN1PH01@H01Yd01P2M008600RM06X6IXEUQAD020L000H76P005G65IVMV1VH0IPIW1X@0
1P2DUIBE00051P08U9FDUI@0WIh61P8000UZ1VH6IYh01VH00P005fJ4IfH6I`IW1P2NW@2MW[6N/@2M
09fN00@600X01VIV05BMWYfN1@H00ih01P041P0309h600@600j5Q9j4VYBb[;:ab;78/@:b00aeQ9j4
WX@6I`IW1VL51P0300H6008600/0Q0IW1VNNQ0IV1P0G0004E4iDJ`:>00^]SZf4WX@6IPIW00031P03
00H6008600<01PH00PH01IjM1YfN008600FMWVH6IP02Q003I@0001@020L000H76`002GJ@Q8EV1VJN
I`061PFE1@H3U@03U9D6008600<01PH00`H00`061P051P0:00H6L8EVQFH6I`D600bM]P=`//P3b02c
WIh41P8000RMWVH0Q02M]PP600JM00H0WIh51P0F09F^UJiVWXBb[<Rab;78/LRaWXBNQ0P600P01PH6
WXBNQ0T600=V00005P003UA?E8BDSjf/L02NQ9iW0`H010061P041P<00`H02iff09jM1YfN00IV0080
00=V00005@081`001PLM000:SfIWIPJ4WX@6I`D60P081P03U00600D600FM00H6W@031P0AJPIZ1P06
JPI[IXAWQ0IW1VL00PH02ijMS[78/LRab;6N00D600FNWIiVJP02E0031Yh000P600D0WIjMWP051P0K
09BEU9FDUIBE08F4L:bb[;://Za`Q9j4WXBN00<000ZNQ9j4WVL6I`IW0`000fJ?000K0003CUA[00:4
00b/Q:f4WP2NQ01W1P081P0309fN008600l0WYfNWIjMWX@6IVYVQ5@05`081`001PLN0007IP1WIPIV
00061P04WIjMWPD600FNWIjM00061P0509jMWP001PH02P2M1P1E08EVWVL51P0;WKI`b0?80lP309d0
1@H019fN00H2IP03Q8FM00P600NNWIh01YF^00>E0P002:jE[YF^UJjE1@H01YfN1Yj/WP8000N4WXBN
I`IW00H600D01PIWU@0P000?E8F4EHA`Q024WVL6I`H0008600BN1PJM1@H03@2NWIh6WYfKWIjMWYd0
0`H2IP03E00001H020L000H780002eAVIVN4QHBN00IW00<600>M1PH01@H00id01P03008600H01PH0
WIh3000@WYfNWIjMWV]VQ8F41VL6I`8600VNWLRab;78/LP00W041P8000RM1VIZIHACQ0`600JDUIBE
U0081P09U9FDU@JNWIjM00<600IV1VJNIXD]0005E6^4E5D018@03P2NQ0IV1VH6WIjM1YfN1@H01YjM
WYd0W@8000ZN/Yj406H0Q6ID6@081`001PLR0006IX@6IPH03PH00`061P041P0609jMWYd02@H02006
QED0Q9iW1@H03W380lP3b0?80lQ`WYfN0PH2008600AVJVEU0U@00iffW@081P05[YFEU@001`H00jjE
[P02U@041P2M00D600>NIPH00VH]000>E6^@E6J4QEFNQ024WVL41P05WYd0WIh01PH04YfNWIh3WP>N
0ijMWPJNQ0H0E1T020L000H79003IP0:IfH0IPIW1VNNI`h600FNW@2MWP071P0LUIBEL6IEQHBN09jM
WYe`/LRab;78/LRaWYfN00<600AZ1P1Z0VH2E003D`0600P600@0WIh020H01@2DUIBE00<600L0I`IV
1VH6009V;`0025A[E4mEQ01E0X@03VL0WVH6IPIW1YfNWIjM1PH04IjMWP38L<RaSYd0Q9j4JVJ401X0
20L000H79@001hAWIPIV1X@00`H00`2N00041P03WYd600/60P051P09UJjE[UEWQ0H00086012NWIjM
/`?80lP3b0?8WIh01PH00`1VJP02IP07E8EC]YffW@030008WIjMWYfNW@021P0509jMWP001PH20003
1P0600=V;`000g00J`02E00<EHAVIVN406H0W@IW0`H01Yd01[JM]PH6012MWYg80lP3b738WIj41VIZ
6`081`001PLZ000EE6J4QHB5Q9j4WXBNQ9j41P2NWIh000P600D0WIjMWP031P9E01>5IXEV1VL6Iih3
b0?80lP3b0>N00H600AZ1VJN0VH2I@04E8ACQ0D600<71PH020H026L6IPH0WXBN0fHc0005CG5>E6/0
0U@2E@0:IUD006MV1VH0IP8600H01PH6WP031P<000nNWLRbb;78/LRa08BNIVX00VH00e@0000I00P7
00061bd0015EIhBb[;:aM[78L8nIWYfNW@071P8010H02`2N1PIDIP0600IW00<600D0b<S80`03b004
0`2M]P@600/01PH6W@1VJVEVQ003008600>MWPH01@H00id01P051P05Q0241VH0=0000g40L@03E004
EEAVE@9V00IWQ0IW1VL41P0600H61Yh010H04KJM]Yff0lP3b0?80lRMWVX6009V7@081`001PL_00:4
01>ETjf//Zbb[;:/b:b>Q9j4WXBN00T601T01XAE06IWIPIV1VNNWHh3b0?80lP3WYfN00<600>NJPH0
1@02Q00CIEB4DeBMWVH6IYj4WVH6IPIV1P04ISP000=`E4l00U@048AD05EVEFN4IfH6IPIV1VH51P03
W@H600<6012MWYfNWLRab;78/LRaL6IZ0VH00h@0000K00P700061c<0011=[Jbb[;:/b;78/K:JWYd0
2`H0202MIXAWIPIW0`H00ifN0003b0080lS8b9ffW@021P8000/6[Y@6IYj406EVQ003E005WIjMWX@0
0PH01Yj41VIWIST000=>E4l00U@01UED08B5E@9V00IWIPIW1VL21P0409fNW@<600>M1PH010H04P2M
]Yf?0lP3b0?8/LRM1VIZIQh020L000H7=`004g2C[I>b[;:/L8BNQ9j4WXBNQ9h01@004iiD06IWIPIV
1VNNWIZab0?80ih00P003YfNW@2MWYfEU9EVJVH00VH02UB4E5ACE6^NQ6L2IP0308@003X000EaCUA[
Q003E004EHB400AV00YWIPIV1VH6WIjM30H03Yh0b;78/LRab;78Q9j40VHO00P700061cX000Z5Q;24
WX@0Q9j430H01YiEIfJNI`<600^MWPNNL<Q`b9fN00051P0:WKJE[Y@6JPIVU@9V00R4IUAUQ5@0E3l0
00AD001[0`003E@0Q9j4WVIWIPIW1VL00PH019jMWYd41P0300H600H60100b0?80lP3b;78WIj4WP25
7`081`001PM10003IXB500=V01d0Q01V1VH6IPH0EF]WIPIV1XBN09jMW[:N/YjMWP051P8000d6U9EZ
1VJDMXmU05AU009D00A[E01D>`02J`04E4iD00=D0X@0101VEFH40009Q0IV09fNWIjM00@600>N1PH0
1@H04@2MW[;8/LRab;78[9j406J4020020L000H7AP04IP0<IfIWIPJ4WXB5IPJ40PH039h61YfNWIjM
WYffW@D600D01PJcU0031P0@U9F?JP1VIFJ4IHADQ5A>QCH000=>E5@00U@0201DEEAVE8F40VH026MV
1P2NWIjM1@H01@2^UK<600801@H04P2NWKH0b0?80lRab;7800IVQB4020L000H7D@0016IDQGH2000:
WVH61P2MWPJNW@8000>N1PH00`H00`061P021P08U9D01XnDMYd2IP0;E6F4E6^4J`1>E4h0<@0024iD
CP1>E4l00e@01EEDQ8@000=V019WIVN4WP06IPIW1P2EU9FD1Yd51P0=WIjMWYg8/LRab;78[002009V
00>4000080081`001PMC0005Q9EVWX@010H01YjM1YfNW@8600>f1PH01`H02JjEUIBEU9F4WP02IP0;
E6J4Q@2]Q024QE@0;P03E0800U@01hAEE6IDIP000VH0201VIfH6I`IW1`H07IF^UK>E09ffW@06]YfN
WKJMb0?80lRab72NQ0IV02@020L000H7E0001EF@Q9iV00<600JMWPH6WIh21P03W@H6008600<71PH0
0PH02IBEU9EV1VIZQ0020003IUAU00:40hh01hm[E4iDCU@09@02J`07E4iDCXA[0002E004JeADIP80
00=EIUD016H02VL0UHnE00IW1VL21P08UIBEU9D0WYd71P8000bMb;78/LR/WX@6IU@T00P700061e@0
00EE09EWWP031P800`H00id61P021P041`JNW@@600F^U:jDU@021P0CIPH01VH0Q9j4U:N]SZf>08B5
Q005E005JeA>E7004@000diDJ`07E006JhF4QHB50W001XB5Q8F400=V00EWIVMVWP020005SiF?UI@0
1PH2U@06[P061YjM1`H039ffWK>Mb9fNWIj41P9V9@081`001PME0006EGJ@IPIW1`H00`061P021P03
WYfN00@600bDUIBEU01V1VH0IVX2IP0<08A`[8j/SZb>Yhh00U@06E=DDe@0E4iDCUA>071aL01>E4iD
CUA>Q6/01`02E00308A[00:>00RWSZQf[HB500IV015WIVN4WXnESiF?1VNEU9FDU@061P03W@H600D6
00^MWYfNWIZMWVX0Q002IP03I@0002@020L000H7E@001EAVMPIW00P600>f1PH00PH00`2NW@021P08
U@J^U:jE[P031P0L00IV1VIZ09jD[Jb]SZf>[HB5Q8F4QH@0Q8F4Q@8000AaMW5f0e@00`24E007E00B
08B5Q8D0Z8j]SZf>[Hj]07240VH04fMV1VH6WIjM06NEU9D0UIB^U:h00iD61P0300H600<600d0WKJM
09fNWIh01VJN009V9`081`001PMF0005E6IV1P0030H03@2MWYBEU9FDUIBEU9D00PH02fH6W@IVJVJM
Q9fD00:/00N]SZf>[8h000QD0X@5E006Q6^408@01e@3000WShjWSZf>[Hj]YjfWUHBNQ9j4WP06IPIV
WVH6I`IWUIBEU9FDUI@000D600<01PH00`H00ijM00021P06WIh0WX@60VH00`24000W00P700061eL0
00MET6L6Q0H000D600JM1PH6W@021P09UJjE1P2^U:jE00@601fNJPIVWYfNQ9j/[Jb][:fWWXBNQ01D
IUAVE6D0I@03E00VIP00E6IDIUAVE8F4QHB5Q8F>[JN]YjfW[Jb][024WXBN00IW1VL21P03WPH600<6
00VEU:jD[YF^UJh01PH01Yd61PJM00@600ZMWPJNW@IZ1XB5:@081`001PMI00051VH6Q9h020H00`06
1P021P03UIBE00<601T01PJDUI@6W@2M1VH6IVYVWHB]0:fCWH@000F400=UE6D00U@00hADE002E00f
05ADE5EDEEAVE5EDQ8n/YjbW[:N]Yjf/[Jb]T`1VIfH6IPIV1VL6I`IW1VNEU9FDUIBEU9D01@H00id0
1P071P06WIiZ1X@60VH/00P700061eT000If1VL6Q9h51P<010H0102EU@051P06[YF^U02M0`H03VX6
IPJMWXBNQ024WXBN0VH04VEVIFH0WXBNQ9h0WP1VE6IDIP8001V4WXBNQ9jD[Jb][:f/a:c8[Ij4WX@0
IPIW00D600<01PH00`H029F^UJjE[YF^1PH019fN1Yh61P09W@H61P06IVZ402`020L000H7F@0028mf
1VH6Q9jM0P0019jMWP061P03U@H600@601RDUIBE09h01VH6IPIVWYd6IYj4WP2NQ9h6IP07IFJ4Q01V
Q0020007IUAVE6IDIP02000KIP24WHA`[:f/[Jb][9j4WXBNW@1W1VL6I`2M00D600T0UIBEU9FDUI@0
0PH01`061PJNW@001PH01id6JPIVWVH0;`081`001PMJ0005T7H6SiD01@H00ijM1P061P0300H600D6
00JE[YF^1Yh61P0=WIiV1VJNQ0IV1XBN0006IP06WP1VIYj42VH059j4WX@0[724SW2NQ9iV1VL0WIjM
10H00`2M00051P0<[YF^UJjE[YD0UJjE0P000ijMWP0;1P0306H002l020L000H7FP001UB@MPIfU@D6
0P051P0409jMWPH600>DU@H01@H00fX6IP02000;1VH609h01VH6IVX02fH00iiVIP07IP0GWHBNQ6H0
WXBNWIj4WYd6IPIV1VH609h00P0059fN1PJDU@JEU9FDUIBE00JDUIBE0P000ifN00041P0800H6IPIV
1VHa00P700061ed000=fUHl00YD31P0509jM]P0010H00ih61P031P0509fNWKH01@H01;H61Yd61P03
WPH600@601]V09fNW@IV1VH6IYjM1VH600IV1VH6IPH0WYd00`H0102NW@081P0:WKH01YFcUK>E00L6
00F^UJjE000>1P03IPIV038020L000H7GP001GH0MYFD008600BN002M3@H019d01P051P03WYfN0086
011V1VH6IPIV1VH6IPJMWYfN0P007hnE00JMWYfNIPIV1VH6IPIVWYd6IPIV1P2NI`IW1VL00`H02Yd0
1PJDUIBEU9D81P03U9FD00H600`0W@2M1VX6IPJ4IVDc00P700061el000If1P2E1P041P0300H600<6
00<01PH02@H00iff00021P03]Yff00H600>M1PH00PH200@600^EU9FDUI@6WIjM00081P0300H600D6
00<01PH02PH01@2E/iD000T600<01PH01PH019fNW@021P0509h0JVH0=0081`001PMP0005MVMV1X@0
2`H00id61P0:1P04WYfNW@@602P0WIjMWYfNWIiV1VH6IPJ?UHnEU9EV1VH600MV1VH6IPH01VH6IPIW
0PH2008600<0W@H020H00id01P071P04WIjM00H600BMWYfN0P02Q09V=P081`001PMQ0005E@IWWP00
1PH010061PH200D600>M1PH010H0102M]Yd61P0509fNW@002@H00`2D00071P0309fN008600>M1PH0
1`H01IfNWIh000P6016fWKJM]Yd01kJMWYffWIjM00041P0300H600@600>EIP00=`081`001PMS0004
IfJNQ0H600JN1PH6WP041P03WP0600@600D0WPH0W@081P0309@0008600QZ1VH6IYFDU@X600BMWYfN
10021P0609jMWYd01PH01`2NWIh0WP001`H00`2N1P041P0<WYfNW@IV1VH6IYEf>@081`001PMT0003
1VL600X600>M]Yd00PH00`2MWP051P0300H6008600<01PH00`H01P2^UJjE00@600BE[YF^0YD00`06
1P071P04WKJM00T600NfWIjM]Yd000D600>f1PH01PH00`061P051P0309jM00D600>?UGH0>P081`00
1PMT0005LFN41VH01@H02ifN1PJMWYd6WIjM00@60P061P03WPH6008600FDUIBEU0021P0;09BEU9FD
UIBEU0000`H01@2DUIBE00X600FMWYfNW@0A1P0300H600L600QVWVH6IYEfSc`020L000H7I@000e@6
Q0031P0>WIh61YfNWIjM]Yd0WKH21P040;JM00L600>fWIh00P004YF^UK<0/iFcUK>E/iF^UJjE[P86
00NE[YF^UK<000H600H0WKH0WYd91P03W@H600@600>N00H01PH019d61PH20003IYEV03h020L000H7
I`0018F41VH21P0BWYd0WIjMb72N0hjMWP2NWIjM20H0102MWYd61P0D09BEU9FDUIBEU9FD1P2EU9FD
U@051P0300H600@600>N1PH01PH00ifN1P041P0300H600P600AZ1VJN0VI000P700061fT000>41VL0
0PH01YjMWYg8/P?800P3b0=aWKJM00P600FM1P2NW@061P0:/iFcUK>E[YF^U@<600<0[YD00`0019ff
W@081P0300H600@600@0WKJM10H00kJM1P031P03WP0600@600@0WVIV@@081`001PM]000GI`2NQ9h0
b;78/LRab0?80lP0WYfNWIh02@H00`061P061P0300H600T60P000`H0W@091P0509jMWYd00PH04ijM
WYd0WIjMWPH0W@IV1VH6IPH00VH00fD0001300P700061fh001D6IPJ4WYfN/LP3b0?80lP3b;80WIh0
40H00iF^U@02U@03[PH600H600<01PH010H200H600D0WKJM]P02L00;WIh3/ifNWG2MWYd00PH01P24
06H6Q4P020L000H7L00056H6IPJ4/[78/LRab;780lP3/id02PH039fNW@06UIBEU9FDU@@600>MWP00
20H01IjMWYfN008601P0WYfN/LRab0?80lRaWYfN09j4WVH6IXA:00P700061g8001T6IPJ4WYg80lP3
b0?80lP0WYffWKJM]Yff00800187WYfNW@0609F^UJh0]YffW@0;1P0JWYfNWKH0b0?80lP3b0?80lP3
b02NWIj41VI>00P700061g@0019VQHBNQ<Rab;78/LRa/k:NW@0:1P03WIh000H600L0WIjMWYfN00/6
01P0WG2Jb;78/LRab;78/LRab:d7IPIV1VE?00P700061gL000YV1PH0WIjMb02N0W000ifNW@0;1P03
W@H600D600<0]PH01`H07IfNWIjMWYfNWIh3b0?80lP3b;78/LRMWXBNQ0IV058020L000H7N@02IP08
1P2NQ02MWYd21P0509fNWIh020H00`061P041P0300H600T601D0WIjMWP38/LRab;78/LR/b:c8Q0H0
0fIE00P700061g`000AWIPJ40`H00ifN1P071P0700H61P2NW@041P03]P0600/601>NWIjMW[;8/LRa
b;78/G24WXAV05T020L000H7O`02IP08IfH6I`241VL81P0C09fNW@JMWYfNW@2M00H01P06WP03000;
1P2NWIj408BNQ9h00P000h@0J`1K00P700061h<000N4WVH6I`H000H60P061P03WPH600X600D0WYfN
00021P0306H6009VH@081`001PN800=V00X6Q9j4WP06I`IW0`H02VL0I`IW1VL6IPH20004I`IV1PAV
IP081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol0
4`081`001POo01<020L000H7o`0C00P700061ol04`081`00o`LQ1`00o`LQ1`00o`LQ1`00o`LQ1`00
o`LQ1`00o`LQ1`00\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "A ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " graphics can serve as texture as well."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; 
    MVShow3D[blankknot, 
      MVTexture -> RasterArray[Map[GrayLevel, {{1, 0}, {0, 1}}, {2}]], 
      MVTextureMapType -> MVMeshUVMapping, MVScaleTexture -> {40, 6}]\)], 
  "Input",
  CellLabel->"In[141]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVMesh3D[4] \[DoubleLeftArrow] MVTexture2D[1], 
      MVMeshUVMapping, {2, 2, MVLuminantTexture}}\)], "Output",
  CellLabel->"Out[141]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[20]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006F5F606F60607060607060607F60607F60607F60607F60607F60607F6060706060\
7060607060607060607060606F60606F6060000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000605F600000000000000000000000000000000000000000007F6060806060
806060806060806060806060806060806F608F6F608F6F6F8F6F608F6F608F6F608F6F608F6F60\
8F6F608F6F608F6F608F6F608F6F608F60608F6060806060806060806060806060
7F60607F60607F60607F60600000000000000000000000006F6060000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000806F6F806F6F
806F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F706F8F706F8F706F8F706F8F706F8F6F6F8F6F6F\
8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F
8F6F608F6F608F6F600000000000000000000000000000000000000000000000000000006F6060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000008F6F6F8F6F70000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000090707090707090707090707090707090706F
90706F90706F90706F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
8F6F708F6F708F6F708F6F708F70708F70708F7070907070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009F7F709F7F709F7F709F7F709F7F709F7F70000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000806F70806F708F6F708F6F708F70708F70708F7070\
8F707F8F707F90707F90707F90707F907F7F907F7F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000A0807FA0807FA0807FA0807FA0807F9F7F7F9F7F7F9F7F709F7F709F7F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000605F60000000000000000000000000
000000000000000000806F70806F708F6F708F707F8F707F8F707F90707F90707F90707F907F7F\
907F80907F80907F80907F80907F809F7F809F7F80000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000A08080A08080A08080A08080A08080A08080A08080A0807FA0807FA0807F9F7F7F\
9F7F709F7F709F7F70000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F5F6070606070606000000000000000000000000080606F
806F70806F708F6F708F707F8F707F8F707F90707F907F80907F80907F80907F80907F80907F80\
907F809F7F809F7F809F7F809F80809F80809F8080000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000AF8F8FAF8F8FAF8F8FAF8F8FAF8F80AF8F80AF8F80AF8F80AF8F80A08080A08080\
A0807FA0807FA0807FA0807F9F7F709F7F709F7F70000000000000000000000000
000000000000000000000000000000000000000000000000706060000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
6F5F6070606070606070606070606F7F606F0000007F606F806F70806F70806F70
8F707F8F707F8F707F8F707F907F80907F80907F80907F80907F80907F80907F809F7F809F8080\
9F808F9F808F9F808F9F808F9F808F9F808F9F808F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000AF8F8FAF8F8FAF8F8FAF8F8FAF8F8FAF8F8FAF8F8FAF8F8FAF8F8FAF8F80AF8F80AF8F80\
AF8F80AF8F80AF8F80A08080A0807FA0807FA0807FA0807FA0807F9F7F709F7F70
000000000000000000000000000000000000000000000000000000806F60706060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006F5F60706060\
70606F70606F7F606F000000000000806F70806F70806F708F707F8F707F8F707F
8F707F8F7080907F80907F80907F80907F80907F8F907F8F9F808F9F808F9F808F9F808F9F808F\
9F808F9F808F9F808F9F808F9F808FA0808FA08F90000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000AF9090AF9090AF9090AF9090AF9090AF908FAF908FAF908FAF8F8FAF8F8FAF8F8FAF8F8F\
AF8F8FAF8F8FAF8F80AF8F80AF8F80AF8F80AF8F80AF8F80A0807FA0807FA0807F
A0807FA0807FA08070A07F709F7F700000000000000000000000009070609070608F6F607F6060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F5F606F606070606F7F606F\
000000000000000000000000000000806F7F8F707F8F707F8F707F8F70808F7F80
907F80907F80907F80907F8F907F8F9F808F9F808F9F808F9F808F9F808F9F808F9F808F9F8090\
9F8090A08F90A08F90A08F90A08F90A08F90000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000B09090B09090B09090B09090B09090B09090B09090B09090AF9090AF908FAF908FAF908F\
AF908FAF908FAF8F8FAF8F8FAF8F8FAF8F8FAF8F80AF8F80AF8F80AF8F80AF8F80
AF8F80AF8F80A0807FA0807FA0807FA0807FA08070A080709F7F709F7F6F9F7F6F90706F907060\
907060806F60000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F5F606F5F6070606F000000000000000000\
0000000000000000000000000000008F707F8F70808F70808F7F80907F80907F80
907F8F907F8F90808F9F808F9F808F9F808F9F80909F80909F80909F8F909F8F90A08F90A08F90\
A08F90A08F90A08F90A08F90A08F90A08F9F000000000000000000000000000000
000000000000AF909FAF909FAF909FAF909FAF909FAF909FAF909F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000B09F9FB09F9FB09F9FB09F9FB09F9FB09F90B09090B09090B09090B09090B09090B09090\
B09090B09090B0908FB0908FAF908FAF908FAF908FAF908FAF908FAF8F8FAF8F80
AF8F80AF8F80AF8F80AF8F80AF8F80AF8F80AF8F7FA0807F0000000000000000009F7F709F7F6F\
9F706F9070609070608F6F60706060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F5F6070606070606F000000000000000000000000\
0000000000000000000000008F70808F70808F7F80907F80907F8F907F8F907F8F
90808F90808F9F808F9F80909F80909F8F909F8F909F8F90A08F90A08F90A08F90A08F9FA08F9F\
A08F9FA08F9F000000000000000000000000AF909FAF909FAF909FAF909FAF909F
AF90A0AF9FA0AF9FA0B09FA0B09FA0B09FA0B09FA0B09FA0B09FA0B09FA0B09FA0B09FA0B09FA0\
B09FA0B09FA0B09F9FB09F9FB09F9FB09F9FB09F9FB09F9FB09F9FB09F9FB09F9F
B09F9FB09F9FB09F9FB09F9FB09F9FB09F9FB09F9FB09F9FB09F9FB09F9FB09F90B09F90B09F90\
B09090B09090B09090B09090B09090B09090B0908FB0908FB0908FB0908FB0908F
B0908FAF908FAF908FAF9080AF8F80AF8F80AF8F80AF8F80000000000000000000000000000000\
9F7F709F7F6F9F7F6F90706F9070609070607F6060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F5F606F606070606F000000000000000000000000000000000000\
0000000000000000000000008F7F80907F8F907F8F907F8F907F8F90808F90808F
9F80909F80909F80909F8F909F8F90A08F9F000000000000000000000000000000000000000000\
000000000000000000000000000000000000AF9FA0AF9FA0AF9FA0AF9FA0B09FA0
B09FA0B09FA0B09FA0B09FAFB0A0AFB0A0AFB0A0AFB0A0AFB0A0A0B0A0A0B0A0A0B0A0A0B0A0A0\
B0A0A0B0A0A0B0A0A0B0A0A0B0A0A0B0A0A0B0A0A0B0A0A0B09FA0BF9FA0BF9FA0
000000000000000000000000000000000000000000000000B09F9FB09F9FB09F9FB09F9FB09F9F\
B09F9FB09F9FB09F90B09F90B09F90B09F90B09090B09090B09090B09090B0908F
B0908FB0908FB0908FB0908FB0908FB0908FB0908FB09080000000000000000000000000000000\
0000000000000000009F7F6F9F7F6F90706F907060907060806F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006F5F606F606070606F000000000000000000000000000000000000000000000000\
000000000000000000000000907F8F907F8F90808F90808F9080909F80909F8090
9F8F909F8F90000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B09FAFB09FAFB09FAFB0A0AFB0A0AF
B0A0AFB0A0AFB0A0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AF\
BFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AF
000000000000000000000000000000000000000000000000000000000000000000000000BF9F9F\
BF9F9FBF9F9FBF9F9FBF9F9FB09F9FB09F9FB09F90B09F90B09F90B09F90B09F90
B09F90B09F90B09090B09090B0908FB0908FB0908F000000000000000000000000000000000000\
0000000000000000000000000000009F7F709F7F6F9F7F6F9070609070608F6F60
7F6060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
605F606F60606F606F000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000090808F9080909080909F80909F80909F8F90000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B0A0AFB0A0AFB0A0AFBFA0B0BFA0B0
BFA0B0BFA0B0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0\
BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFAFBFAFAFBFAFAFC0AFAF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000BF9F9FBF9F9FBF9F9FBF9F9FBF9F9FBF9F9FBF9F90
BF9F90BF9F90B09F90B09F90B09F90B09F90B09F90000000000000000000000000000000000000\
0000000000000000000000000000000000000000009F7F709F7F6F9F7F6F90706F
907060907060806F60000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000605F606F6060
6F606F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000009080909F8090000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0
BFAFB0BFAFBFBFAFBFC0AFBFC0AFBFC0AFBFC0AFBFC0AFBFC0AFBFC0AFBFC0AFBFC0AFBFC0AFBF\
C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000BFA09FBFA09FBFA09F
BF9F9FBF9F9FBF9F9FBF9F9FBF9F90BF9F90BF9F90000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000009F7F70
9F7F6F9F7F6F9070609070608F6F60706060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006F60606F606F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000009F8090000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000BFAFBFBFAFBFBFAFBFBFAFBFC0B0BF
C0B0BFC0B0BFC0B0BFC0B0BFC0B0C0C0B0C0C0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BF\
C0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000BFA09FBFA09FBFA09FBFA09FBF9F9F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009F80709F7F6F9F7F6F90706F907060907060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000908090\
9080909F809F9F8F9F9F8F9F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000C0B0BFC0B0C0C0B0C0C0B0C0C0B0C0
C0B0C0C0B0C0C0BFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0\
CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000009F7F6F9F7F6F9F7F6F9070607F6060000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F606F000000000000000000000000
00000000000000000000000000000000000000000000000000000090809090809090809F9F8F9F\
9F8F9F9F8F9F9F8FA0A090A0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000C0B0C0C0BFC0C0BFC0C0BFC0CFBFCF
CFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCF\
CFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000C0A0A0BFA09F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000009F80709F7F6F907060907060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F606F706070000000000000000000000000
00000000000000000000000000000000000000000090809090809090809F9F8F9F9F8F9F9F8FA0\
9F90A0A090A0A090A0A090A0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000CFBFCFCFBFCFCFBFCFCFBFCFCFBFCF
CFC0CFCFC0CFCFC0D0CFC0D0D0C0D0D0C0D0D0C0D0D0C0D0D0C0D0D0C0D0D0C0D0D0C0D0D0C0D0\
D0C0D0D0C0D0D0C0D0D0C0D0D0C0CFD0C0CFD0C0CFD0C0CFD0C0CFD0C0CF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000C0AFA0C0AFA0C0A0A0C0A09FBFA09F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000009F7F6F9F7F6F907060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F606F7060707F6070000000000000000000000000
0000000000000000000000000000008F809090809090809F9F8F9F9F8F9F9F90A0A090A0A090A0\
A090A0A09FAFA09FAFAF9FAF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000CFC0D0CFC0D0CFC0D0CFC0D0D0C0D0
D0C0D0D0C0D0D0C0D0D0C0D0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0\
D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000CFB0AFCFB0AFC0AFAFC0AFA0C0AFA0C0AFA0C0A0A0C0A09F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000009F80709F7F6F907F6F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F606F7060707F6F7F7F6F7F000000000000000000000000
0000000000000000008F7F90908090908090908F9F9F8F9F9F8FA09F90A0A090A0A090AFA09FAF\
AF9FAFAF9FAFAF9FB0AFA0B0B0A0B0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000D0C0D0D0C0D0D0C0D0D0CFD0D0CFDF
D0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFDFCFDFDFCFDFDFCFDFDFCFDF\
DFCFDFDFCFDFDFCFDFDFCFDFDFCFDFDFCFDFDFCFDFDFCFDFDFCFDFDFCFDF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000CFB0B0CFB0B0CFB0B0CFB0AFCFAFAFC0AFAFC0AFA0C0AFA0\
C0AFA0C0A09FC0A09F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000009F7F6F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000605F6F7060707060707F6F7F7F6F7F7F6F80000000000000000000000000
0000008F7F908F809090809090809F9F8F9F9F8FA09F90A0A090A0A090AFA09FAFAF9FAFAF9FB0\
AFA0B0AFA0B0B0A0B0B0A0BFB0AFBF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000D0CFDFD0CFDFD0CFDFD0CFDFD0CFDF
D0CFDFDFCFDFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DF\
DFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000D0BFBFD0BFBFCFBFB0CFB0B0CFB0B0CFB0AFCFB0AFC0AFAF\
C0AFA0C0AFA0C0AFA0C0A09FC0A09F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000A08070\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
605F6F7060707060707F6F7F7F6F7F7F6F80807080000000000000000000000000
8F7F9090809090809F908F9F9F8FA09F90A09F90A0A090AFA09FAFAF9FB0AFA0B0AFA0B0AFA0B0\
B0A0BFB0AFBFB0AFBFB0AFBFBFAFC0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000D0CFDFD0CFDF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000E0D0E0E0D0E0E0D0E0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000D0C0C0D0C0C0D0BFBFD0BFBFD0BFBFD0BFB0CFBFB0CFB0B0CFB0AF\
CFB0AFCFB0AFC0AFAFC0AFA0C0AFA0C0AFA0C0A09FC0A09F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000A08F7F\
A08070000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000605F6F\
706070706070706F7F7F6F7F7F70808070800000000000000000008F7F908F8090
90809F908F9F908F9F9F8FA09F90A0A090AFA09FAFA09FAFAF9FB0AFA0B0AFA0BFB0AFBFB0AFBF\
B0AFBFB0AFC0BFAFC0BFB0C0BFB0C0000000000000000000000000000000000000
000000000000000000000000000000000000000000D0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDF\
D0CFE0D0D0E0DFD0E0DFD0E0DFD0E0DFD0E0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0DFE0E0DFE0
E0DFE0E0DFE0E0DFE0E0DFE0E0DFE0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000DFCFCFDFC0C0DFC0C0D0C0C0D0C0BFD0BFBFD0BFBFD0BFBFCFBFB0\
CFB0B0CFB0B0CFB0AFCFB0AFCFB0AFC0AFA0C0AFA0C0AFA0C0AF9FC0A09F000000
000000000000000000000000000000000000000000000000000000000000000000000000AF907F\
AF8F7FA08070A08070000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000605F6F6F6070\
706070706F7F7F6F7F7F708080708F80708F0000008F7F908F7F908F809090809F
908F9F9F8FA09F90A09F90AFA090AFA09FAFAF9FB0AFA0B0AFA0BFB0A0BFB0AFBFB0AFC0BFAFC0\
BFB0C0BFB0C0BFB0CFBFB0CFC0BFCF000000000000000000000000000000000000
000000000000000000D0CFDFD0CFDFD0CFDFD0CFE0D0CFE0D0D0E0D0D0E0D0D0E0DFD0E0DFD0E0\
DFD0E0DFD0E0DFD0E0DFD0E0DFD0EFDFDFEF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000E0DFEFE0DFEF
E0DFEFE0DFEFE0DFEFE0DFEFE0DFEFE0DFE0E0DFE0E0DFE0E0DFE0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000DFCFD0DFCFCFDFCFCFDFCFCFDFC0C0DFC0C0D0C0C0D0C0BFD0BFBF\
D0BFBFD0BFB0CFBFB0CFB0B0CFB0AFCFB0AFCFB0AFCFAFA0C0AFA0C0AFA0C0AFA0
C0AF9FC0A09F000000000000000000000000000000000000000000000000000000000000B09080\
AF9080AF8F7FA08F7FA08070000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F5F6F6F6070706070\
706F7F7F6F7F7F708080708F807F8F0000008F7F908F809090809F908F9F908FA0
9F8FA09F90A0A090AFA09FAFA09FB0AFA0B0AFA0B0AFA0BFB0AFBFB0AFC0B0AFC0BFB0C0BFB0CF\
BFB0CFBFBFCFC0BFCFC0BFCFC0BFD0000000000000000000000000000000CFCFDF
CFCFDFD0CFDFD0CFDFD0CFE0D0D0E0D0D0E0D0D0E0DFD0E0DFD0E0DFD0E0DFD0EFDFD0EFDFDFEF\
DFDFEFDFDFEFDFDFEFDFDFEFE0DFEFE0DFEF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000EFE0EFEFE0EF
EFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFDFE0EFDFE0EFDFE0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000E0D0DFE0D0D0E0D0D0E0CFD0DFCFCFDFCFCFDFCFCFDFCFC0DFC0C0D0C0C0\
D0C0BFD0C0BFD0BFBFD0BFBFD0BFB0CFBFB0CFB0B0CFB0AFCFB0AFCFB0AFC0AFA0
C0AFA0C0AFA0C0AF9FC0AF9F000000000000000000000000000000000000000000B09F8FB09080\
B09080B09080AF907FAF8F7FA08F70A08070000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F5F6F6F6070706070706F7F\
7F6F7F7F708080708F807F8F00000000000000000090809F908F9F9F8FA09F90A0
9F90AFA09FAFA09FB0AF9FB0AFA0B0AFA0BFB0AFBFB0AFC0B0AFC0BFB0C0BFB0CFBFB0CFC0BFCF\
C0BFCFC0BFD0C0BFD0C0C0D0CFC0D0CFC0D0CFC0DFCFC0DFCFCFDFD0CFDFD0CFDF
D0CFE0D0D0E0D0D0E0D0D0E0DFD0E0DFD0EFDFD0EFDFDFEFDFDFEFDFDFEFDFDFEFDFDFEFE0DFEF\
E0DFEFE0DFEFE0DFF0E0E0F0E0E0F0E0E0F0E0E0F0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000EFEFF0EFEFF0
EFEFF0EFEFF0EFEFF0EFE0F0EFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EF\
EFE0EFEFE0EFEFE0E0EFE0E0EFE0E0000000000000000000000000000000000000
000000000000000000EFDFDFE0DFDFE0D0DFE0D0D0E0D0D0E0D0D0E0D0CFDFCFCFDFCFCFDFCFC0\
DFCFC0DFC0C0D0C0C0D0C0BFD0C0BFD0BFBFD0BFB0D0BFB0CFBFB0CFB0AFCFB0AF
CFB0AFCFB0A0C0AFA0C0AFA0C0AF9FC0AF9F000000000000000000000000000000B09F8FB09F8F\
B09F8FB09080B09080AF9080AF907FAF8F7FA080709F8070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000605F606F5F6F6F6070706070706F7F7F6F7F\
7F708080708F000000000000000000000000000000908FA09F90A09F90AFA090AF
A09FAFA09FB0AFA0B0AFA0BFAFA0BFB0AFBFB0AFC0BFB0C0BFB0CFBFB0CFBFBFCFC0BFD0C0BFD0\
C0C0D0C0C0D0CFC0D0000000000000000000CFCFDFD0CFDFD0CFE0D0D0E0D0D0E0
D0D0E0D0D0E0DFD0EFDFD0EFDFDFEFDFDFEFDFDFEFDFDFEFE0DFF0E0DFF0E0E0F0E0E0F0E0E0F0\
E0E0F0E0E0F0E0E0F0E0E0F0E0E0F0E0E0F0EFE0F0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000F0EFF0F0EFF0
F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0EFEFF0EFEFF0EFEFEFEFEFEFEFE0EFEFE0EF\
EFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0E0EFE0E0EFE0E0000000000000
000000000000000000EFDFDFEFDFDFEFDFDFEFDFDFEFDFDFE0DFD0E0D0D0E0D0D0E0D0D0E0D0CF\
DFCFCFDFCFCFDFCFC0DFCFC0DFC0C0DFC0BFD0C0BFD0C0BFD0BFB0D0BFB0D0BFB0
CFBFAFCFB0AFCFB0AFCFB0A0C0AFA0C0AFA0C0AF9F000000000000000000000000BF9F90BF9F8F\
B09F8FB09F8FB09F80B09080B09080AF907FAF8F7FA08F70A08070000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000605F606F5F6F6F6070706070706F7F7F6F807F7080\
7F708F0000000000000000000000000000000000009F90AF9F90AFA09FAFA09FB0
AFA0B0AFA0BFAFA0BFB0AFBFB0AFC0B0B0C0BFB0CFBFB0CFBFBFCFC0BFD0C0BFD0C0C0D0000000\
000000000000000000000000000000000000D0CFE0D0D0E0D0D0E0D0D0E0DFD0EF
DFD0EFDFDFEFDFDFEFDFDFEFDFDFF0E0DFF0E0E0F0E0E0F0E0E0F0E0E0F0E0E0F0E0E0F0E0E0FF\
E0E0FFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000F0F0FFF0F0FF
F0F0FFF0F0FFF0F0FFF0F0FFF0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0\
F0EFF0F0EFEFF0EFEFF0EFEFF0EFEFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EF
EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFDFE0EFDFDFEFDFDFEFDFDFEFDFDFEFDFDFE0DFD0E0D0D0\
E0D0D0E0D0CFE0D0CFDFCFCFDFCFC0DFCFC0DFCFC0DFC0BFDFC0BFD0C0BFD0C0B0
D0BFB0D0BFB0D0BFB0CFBFAFCFB0AFCFB0A0C0AFA0C0AF9FC0AF9F000000BFA090BFA090BFA090\
BF9F8FB09F8FB09F8FB09F8FB09080B09080AF9080AF907FAF8F7FA08F70A08070
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000605F606F5F6F6F6070706070706F7F706F807F6F807F708F\
000000000000000000000000000000000000000000000000A09FB0A09FB0AFA0BF
AFA0BFAFAFBFB0AFC0B0AFC0BFB0C0BFB0CFBFBFCFBFBFCFC0BFD0C0C0D0000000000000000000\
000000000000000000000000000000000000000000DFD0EFDFD0EFDFDFEFDFDFEF
DFDFEFDFDFF0E0DFF0E0E0F0E0E0F0E0E0F0E0E0FFE0E0FFE0E0FFE0EFFFEFEFFFEFEFFFEFEFFF\
EFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000F0F0FFF0F0FF
F0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0F0F0F0F0F0EFF0F0EFF0\
F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFEFF0EFEFF0EFEFF0EFEFF0EFEFF0EFEF
F0EFEFF0E0EF000000000000000000EFE0E0EFE0E0EFE0E0EFE0DFEFDFDFEFDFDFEFDFDFEFDFDF\
EFDFD0E0DFD0E0D0D0E0D0CFE0D0CFE0D0CFDFCFC0DFCFC0DFCFC0DFCFC0DFC0BF
DFC0BFD0C0BFD0C0B0D0BFB0CFBFAFCFB0AFCFB0AFCFB0A0C0AFA0C0AF9F000000BFA09FBFA090\
BFA090BFA090BF9F8FB09F8FB09F8FB09F80B09080B09080AF907FAF907FA08F7F
A08070000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000006F5F6F6F60706F6070706F7F706F807F6F807F708F000000\
000000000000000000000000000000000000000000000000AFA0B0AFA0BFAFAFBF
B0AFC0B0AFC0B0B0C0BFB0CFBFB0CFBFBFCFC0BFD0C0BFD0000000000000000000000000000000\
000000000000000000000000000000000000000000DFDFEFDFDFEFDFDFF0DFDFF0
E0E0F0E0E0F0E0E0F0E0E0FFE0E0FFE0E0FFE0EFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFF\
EFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFF0FF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000F0F0FFF0F0FF
F0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FF\
F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFEF
F0EFEFF0EFEF000000000000000000000000000000EFE0E0EFE0E0EFE0E0EFE0DFEFE0DFEFE0DF\
EFDFDFEFDFDFEFDFD0EFDFD0E0DFD0E0D0CFE0D0CFE0D0CFE0D0CFDFCFC0DFCFC0
DFCFC0DFCFBFDFC0BFD0C0BFD0C0B0D0BFB0CFBFAFCFB0AFCFB0A0C0B0A0000000000000C0AF9F\
BFA090BFA090BFA090BFA090BF9F8FB09F8FB09F8FB09F80B09080AF9080AF907F
AF8F7FA08F70A08070000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000605F6F6F60706F6070706F7F706F807F6F807F708F000000000000\
000000000000000000000000000000000000000000000000AFA0BFAFAFC0B0AFC0
B0B0C0B0B0CFBFB0CFBFBFCFBFBFD0C0BFD0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000DFDFF0E0E0F0E0E0F0E0E0F0
E0E0F0E0E0FFE0E0FFE0EFFFE0EFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFF\
EFEFFFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFF0F0FF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000F0F0FFF0F0FFF0F0FF
F0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FF\
FFF0FFFFF0FFFFF0FFF0F0FFF0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0
F0EFF0000000000000000000000000000000000000000000000000000000EFE0E0EFE0E0EFE0DF\
EFE0DFEFE0DFEFE0DFEFE0DFEFDFD0EFDFD0EFDFD0E0DFD0E0DFCFE0D0CFE0D0CF
E0D0C0DFD0C0DFCFC0DFCFBFDFC0BFD0C0B0D0BFB0D0BFAFCFBFAF000000000000000000000000\
000000BFA09FBFA090BFA090BFA090BF9F8FB09F8FB09F8FB09F80B09080B09080
AF907FAF907FAF8F7FA08070000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F60706F6070706F7F706F807F6F80000000000000000000000000\
000000000000000000000000000000000000000000000000B0AFC0B0AFC0B0B0CF
BFB0CFBFBFCFBFBFD0C0BFD0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000E0E0F0E0E0F0E0E0F0E0E0FF
E0EFFFE0EFFFE0EFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFF0FFEFF0FFEFF0FFEFF0FF\
EFF0FFEFF0FFEFF0FFEFF0FFF0F0FFF0F0FFF0F0FFF0F0FF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000F0F0FFF0F0FFF0F0FF
F0F0FFF0F0FFF0F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FF\
FFF0FFFFF0FFFFF0FFFFFFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0F0FFF0F0
F0F0F0000000000000000000000000000000000000000000000000000000000000000000F0E0E0\
EFE0E0EFE0DFEFE0DFEFE0DFEFE0DFEFE0DFEFE0D0EFDFD0EFDFD0EFDFD0E0DFCF
E0DFCFE0D0CFE0D0C0DFCFC0DFCFBFDFCFBFD0C0BFD0C0B0D0BFB0000000000000000000000000\
000000000000000000BFA09FBFA090BFA090BFA090BF9F8FB09F8FB09F8FB09F80
B09080AF9080AF907FAF8F7FA08070000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000605F600000006F6070706F7F706F807F6F80000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000B0B0CFBFB0CF
BFBFD0BFBFD0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000E0EFF0E0EFFFE0EFFF
EFEFFFEFEFFFEFEFFFEFEFFFEFEFFFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FF\
F0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FF000000000000000000
000000000000000000000000F0FFFF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000F0FFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFF0FF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000F0EFE0F0E0DFEFE0DFEFE0DFEFE0DFEFE0DFEFE0D0EFE0D0EFE0D0
EFDFD0EFDFCFE0DFCFE0D0CFE0D0C0DFCFC0DFCFBFDFC0BF000000000000000000000000000000\
000000000000000000000000BFAF9FBFA090BFA090BFA090BFA08FB09F8FB09F8F
B09F80B09F80B09080AF907FAF8F7FA08070000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F607F706F807F6F80000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000BFBFCFBFBFD0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000EFEFFFEFEFFF
EFEFFFEFEFFFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFF0F0FFF0F0FFF0F0FF\
F0F0FFF0F0FFF0FFFF000000000000000000000000000000F0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFFFFFFFFFFFFF000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000F0EFE0F0EFE0F0E0DFEFE0DFEFE0DFEFE0DFEFE0DF
EFE0D0EFE0D0EFDFD0E0DFCFE0D0CFE0D0C0DFD0C0DFCFBF000000000000000000000000000000\
000000000000000000000000000000000000BFAF9FBFA090BFA090BFA090BFA08F
B09F8FB09F8FB09F80B09080AF907FAF8F7FA08F70000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000706F7F706F80000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000BFBFD0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000EFEFFFEFF0FF
EFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFF0F0FF000000000000000000000000000000\
000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000F0EFE0F0EFDFF0E0DFEFE0DF
EFE0DFEFE0D0EFE0D0EFE0D0EFDFCFE0DFCFE0D0CF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000BFA090BFA090BFA090
BFA08FB09F8FB09F8FB09F80B09080AF907FAF907FA08F70000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
605F60000000000000000000706F80000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000BFBFD0C0C0D0C0C0DF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000EFF0FF
EFF0FFEFF0FFEFF0FFEFF0FF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000F0EFE0
EFE0DFEFE0DFEFE0DFEFE0D0EFE0D0EFDFD0E0DFCF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000BFAF9FBFA090
BFA090BFA090B0A08FB09F8FB09F80B09080AF907FAF907FA08F709F806F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000BFBFD0C0C0D0C0C0DFC0C0DFCFCFE0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000EFF0FF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000F0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
F0EFE0F0EFDFEFE0DFEFE0DFEFE0D0EFE0D0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
BFAF90BFA090BFA090B0A08FB09F8FB09F80B09080AF907FAF907FA08F709F80700000008F7060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000706F80000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000BFBFD0C0C0D0C0C0DFCFCFDFCFCFE0CFCFE0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000EFF0FFEFF0FFEFFFFF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000F0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000F0EFE0EFEFDFEFE0DFEFE0DF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000BFAF90BFA090BFA090B0A08FB09F8FB09F80B09F80AF907FAF907FA08F709F8070907F6F\
8F7060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000605F60000000
0000000000000000007F708F807F90000000000000000000000000000000000000000000000000\
000000000000000000000000BFBFD0C0C0DFC0C0DFCFCFE0CFCFE0CFD0E0D0D0EF
D0D0EF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000EFF0FFEFF0FFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000EFFFFF
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000EFEFDF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000BFAF9FBFA090BFA090B0A08FB09F8FB09F80AF9F80AF907FAF907FA08F70000000\
907F6F8F7060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000070708F7F7F9080809F000000000000000000000000000000000000000000000000\
000000000000BFBFD0BFBFD0C0C0DFC0C0DFCFCFE0CFCFE0CFD0EFD0D0EFD0DFEF
DFDFF0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000EFF0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000EFFFFF
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF\
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000FFFFFF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000EFEFE0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000BFA090BFA090B0A08FB09F8FB09F80AF9F80AF907FAF907F000000\
000000907F6F8F7060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000706F807F708F807F908F809F000000000000000000000000000000000000000000000000\
000000BFBFD0BFBFD0C0C0DFC0C0DFCFCFE0CFCFE0CFD0EFD0D0EFD0DFF0DFDFF0
DFDFF0DFE0F0000000000000000000000000000000000000000000000000000000000000EFF0FF\
EFF0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000EFFFFF
EFFFFFEFFFFFEFFFFFEFFFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFEFFFFFEFFFFFEFFFFFEFFFFF\
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000F0EFE0F0EFE0F0EFE0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000BFA090BFA090B0A08FB09F8FB09F80AF9F80000000000000\
000000000000907F6F907060000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
706F807F708F7F7F9080809F8F8FA09090AF000000000000000000000000000000000000000000\
BFBFD0BFBFD0C0C0DFC0C0DFCFCFE0CFCFE0D0D0EFD0D0EFD0DFF0DFDFF0DFE0F0
DFE0F0DFE0F0E0E0F0000000000000000000000000000000000000000000EFF0FFEFF0FFEFFFFF\
EFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000E0F0FFE0F0FFE0F0FFE0F0FFE0F0FFE0F0F0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000E0FFFFE0FFFFE0FFFFE0FFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000F0F0EFF0F0EFF0EFE0F0EFE0F0EFE0EFEFE0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000BFA090BFA090B0A08FB09F8FB09F80000000000000\
000000000000000000907F6F907F60000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
706F807F708F807F9F8F809F908FA09F90AF000000000000000000000000000000000000BFBFD0\
BFBFD0C0C0DFC0C0DFCFCFE0CFCFE0D0D0EFD0D0EFD0DFF0DFDFF0DFE0F0DFE0F0
DFE0F0E0EFF0E0EFF0E0EFF0000000000000000000000000EFF0FFEFF0FFEFFFFFEFFFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFEFFFFFEFFFFFEFFFFFEFFFFF
EFFFFFEFFFFFEFFFFF000000000000000000000000000000000000000000000000E0FFFFE0F0FF\
E0F0F0DFF0F0DFF0F0DFF0F0DFF0F0DFF0F0DFEFF0DFEFF0DFEFF0D0EFF0D0EFF0
D0EFF0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000DFF0F0DFF0F0DFF0F0
DFF0FFE0F0FF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000F0FFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000F0F0EFF0F0EFF0F0EFF0F0EFF0EFE0F0EFE0EFEFE0EFEFDF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000BFAF90BFA090B0A08FB09F8F000000000000\
0000000000000000000000009F806F907F60000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000706F80
7F708F807F908F809F8F8FA09090AF9F9FB0000000000000000000000000B0B0CFBFBFD0BFBFD0\
C0C0DFC0CFDFCFCFE0CFD0E0D0D0EFD0DFEFD0DFF0DFDFF0DFE0F0DFE0F0DFE0F0
E0EFF0E0EFF0E0EFF0E0EFFFEFF0FF000000EFF0FFEFF0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF
EFFFFFEFFFFFEFFFFFEFFFFF000000000000000000E0FFFFE0F0FFE0F0F0E0F0F0DFF0F0DFF0F0\
DFEFF0D0EFF0D0EFF0D0EFF0D0EFF0D0EFEFCFE0EFCFE0EFCFE0EFCFE0EFCFE0EF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000D0E0F0D0E0F0D0EFF0
D0EFF0D0EFF0DFEFF0DFF0F0DFF0F0E0F0FF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000FFFFF0FFF0F0F0F0F0F0F0EFF0F0EFF0F0EFF0EFE0F0EFE0EFEFE0EFEFE0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000BFAF90BFA090000000000000000000\
0000000000000000000000000000009F806F907F60000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006F6F8070708F
7F7F9080809F8F8FA09090AF9F9FB0A09FBFA0A0BF000000000000B0B0CFBFBFD0BFBFD0C0C0DF\
C0CFDFCFCFE0CFD0E0D0D0EFD0DFEFD0DFF0DFDFF0DFE0F0DFE0F0E0E0F0E0EFF0
E0EFF0E0EFF0E0F0FFEFF0FF000000EFF0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF
EFFFFFEFFFFFEFFFFF000000000000000000E0F0F0DFF0F0DFF0F0DFEFF0DFEFF0D0EFF0D0EFF0\
D0EFEFC0DFDF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
CFDFEFCFE0EFCFE0F0D0E0F0D0EFF0D0EFF0DFEFF0DFF0F0DFF0F0E0F0FF000000000000000000\
000000000000000000000000000000000000000000EFFFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0F0F0F0F0EFF0F0EFF0F0EFF0EFE0F0EFE0EFEFE0\
EFEFDF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000BFAF90000000000000000000\
0000000000000000000000000000000000000000008F7060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000706F807F708F
807F908F809F908FA09090AF9F9FB0A0A0BFAFAFC0AFB0CFB0B0CFBFBFD0BFC0D0C0C0DFC0CFE0\
CFCFE0CFD0EFD0D0EFD0DFEFD0DFF0DFDFF0DFE0F0DFE0F0E0E0F0E0EFF0E0EFF0
E0EFFFEFF0FF000000000000000000000000EFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF000000
000000000000000000000000000000000000000000D0EFF0D0EFF0D0EFEFB0C0C0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000CFDFEFCFE0EFCFE0F0D0E0F0D0EFF0D0EFF0DFEFF0DFEFF0DFF0F0\
E0F0FF000000000000000000000000000000000000EFFFFFEFFFFFEFFFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0F0F0F0F0F0F0F0F0F0EFF0F0EFF0F0EFF0EFE0\
F0EFE0EFEFE0EFEFDF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000BFAF90000000000000\
000000000000000000000000000000000000000000907F6F8F7060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006F6F8070708F7F7F90
80809F8F8FA09090AF9F9FB0A0A0BFA0AFC0AFAFC0B0B0CFB0BFD0BFBFD0C0C0DFC0CFE0CFCFE0\
CFD0EFD0D0EFD0DFEFD0DFF0DFDFF0DFE0F0DFE0F0E0E0F0E0EFF0E0EFF0E0EFFF
EFF0FF000000000000000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF000000000000000000
000000000000000000000000000000000000B0C0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000CFE0EFCFE0F0D0E0F0D0E0F0\
D0EFF0DFEFF0DFEFF0DFF0F0E0F0FF000000EFFFFFEFFFFFEFFFFFEFFFFFEFFFFF
EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0F0F0F0F0F0F0EFF0F0EFF0F0EF\
F0F0EFF0EFE0F0EFE0EFEFE0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000C0B09FBFAF9FBFAF90000000\
000000000000000000000000000000000000000000000000907F6F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000605F600000006F6F8070708F7F7F9080809F
8F8FA09090AF9F9FB09F9FBFA0A0BFAFAFC0B0B0CF000000000000BFC0DFC0CFDFCFCFE0CFD0EF\
D0D0EFD0DFEFD0DFF0DFDFF0DFE0F0DFE0F0E0EFF0E0EFF0E0EFF0E0EFFFEFF0FF
000000000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFEFFFFFEFFFFFEFFFFF000000000000000000000000000000
000000000000000000BFD0D0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
CFDFEFCFE0F0D0E0F0D0E0F0D0EFF0DFEFF0000000000000E0FFFFEFFFFFEFFFFF
EFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000FFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0F0F0F0F0F0F0EF\
F0F0EFF0F0EFF0EFE0F0EFE0EFEFE0EFEFDF000000000000000000000000000000
000000000000000000000000000000000000000000000000CFBFA0C0B0A0C0B09FBFAF9FBFAF90\
0000000000000000000000000000000000000000000000000000008F7F60000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000605F6F0000006F6F807F708F807F9080809F
8F8FA09090AF9F9FB0A0A0BFAFAFC0AFB0CF000000000000000000000000CFCFE0CFD0EFD0D0EF\
D0DFEFD0DFF0DFDFF0DFE0F0DFE0F0E0EFF0E0EFF0E0EFFFE0F0FF000000000000
000000000000000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFEFFFFF000000000000000000000000000000000000000000000000
000000BFCFCF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000CFDFEF000000000000000000000000000000E0F0FF
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000FFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0\
F0F0F0F0F0EFF0F0EFF0F0EFF0EFE0F0EFE0EFEFE0000000000000000000000000
000000000000000000000000000000000000000000000000CFBFAFCFBFA0C0B0A0C0B09FBFAF9F\
BFAF900000000000000000000000000000000000000000000000000000008F7060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000605F6F000000706F8F7F7F9080809F8F8FA0
9090AF9F9FB0A0A0BFA0A0C0AFAFC0000000000000000000000000000000000000D0D0EFD0DFEF\
D0DFF0DFDFF0DFE0F0DFE0F0E0EFF0E0EFF0E0EFFFE0F0FF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000F0FFFFF0FFFF\
EFFFFF000000000000000000000000000000000000000000000000000000B0C0C0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000E0FFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0\
FFF0F0F0F0F0F0F0F0F0F0EFF0F0EFF0EFE0F0EFE0EFEFE0000000000000000000
000000000000000000000000000000000000000000D0CFB0CFC0B0CFC0AFCFBFAFC0B0A0C0B09F\
BFAF9FBFAF90000000000000000000000000000000000000000000000000907F6F
8F7060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000605F606060706F608070708F807F908F8FA09090AF
909FB09F9FBFA0A0BFAFAFC0AFB0CF000000000000000000000000000000000000D0DFEFD0DFF0\
DFDFF0DFE0F0DFE0F0E0EFF0E0EFF0E0EFF0EFF0FF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000EFFFFF\
EFF0FF000000000000000000000000000000000000C0D0D0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000E0F0FFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFF0FFF0F0F0F0F0F0F0EFF0F0EFF0EFE0EFEFE0EFEFE0EFEFDF000000
000000000000000000000000000000000000000000D0CFBFD0CFB0CFC0B0CFC0AFCFBFAFC0BFA0\
C0B09FBFAF9FBFAF90000000000000000000000000000000000000000000000000
907F6F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000605F6F6060706F6F807F709080809F8F8FA09090AF
9F9FB0A0A0BFA0AFC0AFAFCF000000000000000000000000000000000000000000000000000000\
DFE0F0DFE0F0E0EFF0E0EFF0E0EFF0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000EFFFFFEFFFFFEFF0FF\
E0F0F0E0F0F0DFF0F0DFF0F0D0E0E0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000E0FFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFF000000000000000000000000
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFF0F0F0F0F0F0F0EFF0F0EFF0EFE0EFEFE0EFEFE0EFEFDF
000000000000000000000000000000000000DFD0C0DFD0BFD0CFBFD0CFB0CFC0B0CFC0AFCFBFAF\
C0BFA0C0B0A0BFB09FBFAF90000000000000000000000000000000000000000000
0000008F7F60000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000605F6F000000706F8F7F7F9080809F9090AF9F9FB0
9FA0BFA0A0C0AFAFC0000000000000000000000000000000000000000000000000000000000000\
000000DFEFF0E0EFF0E0EFF0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000EFFFFFEFFFFFEFF0FFE0F0F0\
E0F0F0DFF0F0DFEFEFCFDFDF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000E0F0FFEFFFFFEFFFFFEFFFFF\
EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000
000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0F0F0EFF0F0EFF0EFEFEFEFE0EFEFE0
EFEFDF000000000000000000000000E0DFCFDFDFCFDFD0C0DFD0BFD0CFBFD0CFB0CFC0B0CFC0AF\
CFBFAFC0BFA0C0B0A0BFB09FBFAF90000000000000000000000000000000000000
0000000000008F7060000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000605F60605F7000000070708F7F7F9F8F8FA09090AF9F9FBF
A0A0BFAFAFC0000000000000000000000000000000000000000000000000000000000000000000\
000000000000E0EFF0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000EFFFFFEFFFFFEFF0FFE0F0F0E0F0F0\
DFF0F0D0EFEF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000EFFFFF\
EFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFF000000000000000000000000000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0F0F0EFF0F0EFF0EFEFEFEFE0
EFEFE0EFEFDF000000000000E0E0D0E0DFD0E0DFCFE0DFCFDFD0C0DFD0C0D0CFBFD0CFB0CFC0B0\
CFC0AFCFBFAFC0BFA0C0B0A0BFB09FBFAF90000000000000000000000000000000
000000000000907F6F807060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000605F6F60607000000000000080809F8F8FA0909FB09FA0BF
A0AFC0AFAFCF000000000000000000000000000000000000000000000000000000000000000000\
000000000000E0F0FF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000EFFFFFEFFFFFEFFFFFEFF0FFE0F0F0E0F0F0DFF0F0\
D0E0E0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
E0F0FFEFFFFFEFFFFFEFFFFFEFFFFF000000000000000000000000000000000000
000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0F0EFF0F0EFF0EFEF
EFEFE0EFEFE0EFEFDF000000EFE0D0E0E0D0E0DFD0E0DFCFE0DFCFDFD0C0DFD0C0D0CFBFD0CFBF\
CFC0B0CFC0AFCFBFAFC0BFA0C0B0A0BFAF9FBFAF90000000000000000000000000
0000000000000000008F7F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000605F6F00000000000000000080809F9090AF9F9FB0A0A0BF
AFAFC0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000EFF0FFEFF0FF000000000000000000000000000000000000000000
000000000000000000000000000000EFFFFFEFFFFFEFFFFFEFF0FFE0F0F0E0F0F0DFEFEFCFDFDF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000BFC0D0000000000000000000\
000000000000E0FFFF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0F0F0F0F0EF
F0EFEFEFEFE0EFEFE0000000000000E0E0DFE0E0D0E0DFD0E0DFCFE0DFCFDFD0C0DFD0C0D0CFBF\
D0CFBFD0C0B0CFC0AFCFBFAFC0BFA0C0B09FBFAF9FBFAF90000000000000000000
0000000000000000000000008F7060000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000605F700000000000000000000000009090AF9F9FBFA0A0C0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000EFF0FFEFF0FFEFFFFFEFFFFFF0FFFFF0FFFF000000000000000000000000
000000000000000000F0FFFFEFFFFFEFFFFFEFFFFFEFF0FFE0F0F0E0F0F0D0EFE0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000DFEFF0E0F0FF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0F0
F0F0EFF0EFEF000000000000000000000000E0E0DFE0E0D0E0DFD0E0DFCFE0DFCFDFD0C0DFD0C0\
D0CFBFD0CFBFCFC0B0CFC0AFC0BFAFC0B0A0BFB09FBFAF90B0AF90000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000605F70000000000000000000000000909FB09FA0BFA0AFC0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
E0F0FFEFF0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000
000000000000F0FFFFEFFFFFEFFFFFEFF0FFE0F0F0E0F0F0E0F0F0D0E0E0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
CFDFF0D0E0F0D0E0F0DFEFF0DFF0F0E0F0FF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0
F0F0F0000000000000000000000000000000000000E0E0DFE0E0D0E0DFD0E0DFCFE0DFCFDFD0C0\
DFD0C0D0CFBFD0CFBFCFC0B0CFBFAFC0BFA0C0B0A0BFB09FBFAF90000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000605F60000000000000000000000000000000000000A0A0BFAFAFC0
000000000000000000000000000000000000000000000000000000000000000000000000E0F0FF\
EFF0FFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
000000F0FFFFEFFFFFEFFFFFEFFFFFEFF0F0E0F0F0DFEFEFD0E0E0CFDFD0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000CFDFF0D0E0F0D0EFF0DFEFF0E0F0FF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000EFE0DFE0E0D0E0DFD0E0DFCFE0DFCF\
DFD0C0DFD0C0D0CFBFD0CFB0CFC0B0CFBFAFC0BFA0C0B09FBFAF9FBFAF90000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000605F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000E0F0FFEFF0FF\
EFF0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
000000000000000000000000000000000000000000000000CFDFDFC0D0CF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000CFDFF0CFE0F0D0E0F0DFEFF0DFEFF0E0F0FF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000000000000000000000000000000000000000EFE0DFE0E0D0E0E0D0E0DFCF\
DFDFCFDFD0C0D0D0C0D0CFBFCFC0B0CFC0AFC0BFAFC0BFA0BFB09FBFAF9FB0AF90
00000000000000000000000000000000000090806F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000605F6F000000000000000000000000000000000000000000AFB0CF
000000000000000000000000000000000000000000000000000000000000000000EFF0FFEFF0FF\
EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000
000000000000000000000000000000000000000000D0DFDFC0D0CF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000CFDFF0CFE0F0D0E0F0DFEFF0DFF0FFE0F0FF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000
000000000000000000000000000000000000000000000000000000000000EFE0DFE0E0D0E0E0D0\
E0DFCFDFDFCFDFD0C0D0D0BFD0CFBFCFC0B0CFC0AFC0BFAFC0B0A0BFB09FBFAF90
B0AF900000000000000000000000000000009F8F7090806F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000605F6F000000000000000000000000000000000000AFAFC0B0B0CF
B0BFD0000000000000000000000000000000000000000000000000000000E0F0FFEFF0FFEFFFFF\
EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000
000000000000000000000000000000000000D0E0DFCFDFD0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFDFF0D0E0F0D0EFF0DFEFF0
E0F0FFE0F0FF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000EFE0DFE0E0D0\
E0DFD0E0DFCFDFDFCFDFD0C0D0CFBFD0CFB0CFC0B0CFBFAFC0BFA0C0B0A0BFB09F
BFAF90000000000000000000000000A0907FA090709F8F70000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000605F70000000000000000000000000000000000000AFAFCFB0BFD0
BFBFD0C0C0DF000000000000000000000000000000000000000000E0F0FFEFF0FFEFFFFFEFFFFF\
EFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000
000000000000000000000000000000000000CFDFD0A0AF9F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000C0D0EFCFE0F0D0E0F0
DFEFF0DFEFFFE0F0FF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000FFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000EFE0DF\
E0E0D0E0DFD0DFDFCFDFD0C0DFD0C0D0CFBFCFCFB0CFC0B0C0BFAFC0BFA0BFB09F
BFAF9FB0AF90000000000000000000A09F7FA0907FA090709F8F6F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000605F70000000000000000000000000000000000000AFB0CFB0BFD0
BFC0DFC0CFE0CFD0EF000000000000000000000000000000000000E0F0FFEFF0FFEFFFFFEFFFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000
000000000000000000000000000000CFDFD0BFC0B0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000CFDFF0
CFE0F0D0E0F0DFEFF0E0F0FF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000FFFFFF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
E0E0DFE0E0D0E0DFD0DFDFCFDFD0C0D0D0C0D0CFBFCFC0B0CFC0AFC0BFAFC0BFA0
BFB09FBFAF90B0AF90000000AFA08FAF9F80A09F7FA0907F9F8F70000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000A0AFC0AFB0CFB0BFD0
BFC0DFC0CFE0CFD0EFD0DFEFD0DFEF000000000000000000E0EFFFEFF0FFEFFFFFEFFFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000
000000000000000000000000D0DFDFCFD0CF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
C0DFF0CFE0F0D0E0F0DFEFF0E0F0FFE0F0FF000000000000000000000000000000000000000000\
000000000000000000000000FFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000E0E0D0E0E0D0E0DFCFDFDFCFDFD0C0D0D0BFD0CFBFCFC0B0CFC0AFC0BFA0
C0B0A0BFB09FBFAF90B0AF90B0A08FAF9F80AF9F80A0907FA09070908F6F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000AFAFCFB0B0D0BFC0DF
C0CFDFC0CFE0CFD0EFD0DFEFD0DFEFDFE0F0DFEFF0000000E0F0FFEFF0FFEFFFFFEFFFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000
000000000000000000DFE0DFD0DFD0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000007F6060806060806060806060806060806060806060806F60806F60000000000000000000\
0000000000000000000000000000000000000000000000007F6060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000C0D0E0CFE0F0D0E0F0DFEFF0DFEFFFE0F0FF000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000E0E0D0E0DFD0E0DFCFDFDFCFDFD0C0D0CFBFD0CFB0CFC0B0C0C0AF
C0BFA0BFB0A0BFB09F000000000000AFA08FAF9F80A09F80A0907F9F8F70000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000A0AFC0AFB0CFB0BFD0BFC0DF
C0CFE0CFD0EFCFD0EFD0DFEFD0E0F0DFE0F0E0EFF0000000EFF0FFEFF0FFEFFFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFFFFFFFF0FFFF000000000000000000000000000000
000000000000000000D0E0DFCFDFCF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007F606080606F806F6F806F6F
806F6F8F6F6F8F6F6F8F706F8F706F8F706F8F706F8F706F8F706F8F706F8F706F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000C0CFDFCFDFF0D0E0F0DFEFF0DFEFFFE0F0FF000000000000000000000000000000\
000000F0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000E0E0D0E0DFD0DFDFCFDFDFCFD0D0C0D0CFBFCFCFB0CFC0B0
C0BFAFC0BFA0BFB09F000000000000B0A08FAFA08FAF9F80A09F7FA0907090806F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000A0AFC0AFB0CFB0BFD0BFC0DF
C0CFE0CFD0EFCFDFEFD0DFEFDFE0F0DFEFF0E0EFF0000000000000000000000000000000F0FFFF\
F0FFFFF0FFFFF0FFFFFFFFFFFFFFFF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
7F60607F60607F606080606F806F6F806F6F806F6F8F6F6F8F6F6F8F70708F7070
907070907070907070907F709F7F709F7F709F7F709F7F709F7F709F7F709F7F709F7F709F7F70\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000806F60806F607F60607F6060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000BFC0D0CFDFF0D0E0F0D0EFF0DFEFF0E0F0FF000000000000000000F0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000E0E0DFE0E0D0E0DFD0DFDFCFDFD0C0D0D0C0D0CFBFCFCFB0
CFC0AFC0BFAF000000000000000000000000B0A08FAFA080A09F80A0907F9F8F70000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000A0AFC0AFB0CFBFC0DFC0CFE0
C0CFE0CFD0EFD0DFEFD0DFEFDFE0F0DFEFF0000000000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000
E0F0EFDFEFE0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006F60607F6060000000000000000000\
0000000000008F6F6F8F70708F70708F7070907070907070907F70907F709F7F7F
9F7F7F9F7F7F9F7F7F9F7F7F9F807F9F807F9F807F9F807F9F807F9F807F9F807F9F807F9F807F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000008F70608F7060806F607F6F60000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000BFBFD0CFDFF0D0E0F0D0E0F0DFEFF0E0F0FF000000EFFFFFEFFFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000E0E0DFE0E0D0E0DFCFDFDFCFDFD0C0D0D0BFD0CFBF
CFC0B0000000000000000000000000000000000000AFA08FAF9F80A0907F9F907090806F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000060607F000000000000000000000000A0AFC0B0B0D0BFC0DFC0CFE0
CFD0E0CFD0EFD0DFEFD0E0F0DFE0F0E0EFF0000000000000000000000000000000000000000000\
000000000000000000000000FFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFEFF0F0
E0EFEF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007F6060000000000000000000000000000000000000\
000000000000000000907F7F9F7F7F9F7F7F9F7F7F9F7F7F9F7F7F9F807F9F807F
9F807F9F807FA0807FA0807FA08080A08080A08080A08080A08080A08080A0807FA0807FA0807F\
A0807F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000907F6F907F6F9070608F7060806F60706060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000BFBFCF000000000000000000000000000000000000EFFFFF\
EFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000E0E0D0E0DFD0DFDFCFDFDFCFDFD0C0D0CFBF
000000000000000000000000000000000000000000B0A08FAF9F80A09F7FA0907F9F8F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F607F707F90000000000000000000A0AFC0B0BFD0BFC0DFC0CFE0
CFD0EFCFDFEFD0DFEFDFE0F0DFEFF0E0EFF0000000000000000000000000000000000000000000\
000000000000000000000000FFFFFFF0FFFFF0FFFFF0FFFFF0FFFFEFFFFFEFF0F0
E0EFE0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000706060000000000000000000000000000000000000000000000000000000\
0000000000000000000000009F807F9F807F9F80809F8080A08080A08080A08080
A08080A08F80A08F80000000000000000000000000000000000000000000000000000000000000\
000000000000A08F80A08F80A08F80A08F80A08F80A08F80A08F7FA08F7FA08F7F
A08F7FA08F7FA0807F000000000000000000000000000000907F6F907F6F907F6F907F6F8F7060\
806F607F6060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000B0B0C0000000000000000000000000000000000000\
EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000E0E0D0E0DFD0DFDFCFDFDFCFD0D0C0
000000000000000000000000000000000000000000000000AFA08FA09F80A0907F9F90708F806F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F6F7F707F90000000000000000000AFAFCFB0BFD0BFC0DFC0CFE0
CFD0EFD0DFEFD0DFEFDFE0F0DFEFF0000000000000000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFF0FFFFF0FFFFF0FFFFF0FFFFEFF0F0E0EFEF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
706060000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000A08080A08080A08F80A08F80000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000AF908FAF908FAF9080AF9080AF9080AF9080AF9080AF9080AF9080
AF9080A08F80A08F7FA08F7FA08F7FA08F7F9F80709F80709F80709F80709F8070907F6F907F6F\
907F6F8F7060806F607F6F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000E0E0D0E0DFD0DFDFCF000000
000000000000000000000000000000000000000000000000AFA08FAF9F80A09F7F9F90709F8F70\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F6F7F7F7F90808FA0000000000000AFB0CFB0BFD0BFC0DFC0CFE0
CFD0EFD0DFEFD0E0EFDFE0F0E0EFF0000000000000000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFF0FFFFF0FFFFF0FFFFF0FFFFEFF0F0E0EFE0
0000000000000000000000000000000000000000000000000000000000007F606F7F606F7F606F\
7F606F7F606F806F6F806F6F806F6F806F6F806F6F806F6F806F6F806F6F80606F
7F606F7F606F7F606F7F606F7F606F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000706060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000A08F8FA08F8FA08F8F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000B09F8FB09F8FAF908FAF908FAF908FAF908FAF908FAF908F
AF908FAF9080AF9080AF9080A09080A08F7FA08F7FA08F7FA08F7F000000000000000000000000\
907F6F907F6F8F7F6F8F7060806F607F6F60000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000EFF0FFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000E0E0D0000000000000
000000000000000000000000000000000000000000000000000000AFA08FA09F80A0907F9F8F70\
8F7F6F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F6F807F7F908F8FA0909FB0000000AFB0CFB0BFD0BFC0DFC0CFE0
CFD0EFD0DFEFD0E0F0DFE0F0E0EFF0000000000000000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFF0FFFF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000806F6F806F6F806F6F8F6F6F\
8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F
8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F806F6F806F6F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000A08F8FA08F8F\
A08F8FAF8F8FAF8F8FAF908FAF908F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000B09F90B09F90B09F90B09F90B09F90B09F90B09F8FB09F8F
B09F8FAF9F8FAF9F8FAF908FAF9080AF9080AF9080A09080A0907F000000000000000000000000\
000000000000907F6F907F6F8F7F6F8F7060807060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000BFC0DF000000000000\
000000000000000000000000EFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000E0E0DF000000000000000000
000000000000000000000000000000000000000000000000000000AFA08FA09F80A09F7F9F9070\
9F8F6F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F6F7F7F7F908F8FAF909FB0000000AFB0CFB0BFD0BFC0DFC0CFE0
CFD0EFD0DFEFD0E0F0DFE0F0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000907070907070907070907070907070\
9070709070709070709F70709F70709F70709F7070907070907070907070907070
90707090707090707090707090707090707090707090707090707090707090707090706F90706F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000A08F8FA08F8FA08F8FA08F8FAF8F90\
AF9090AF9090AF9090AF9090AF9090B09090000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000BFA09FBFA090BFA090BFA090B0A090B0A090B0A090B0A090
B0A090B09F90B09F8FB09F8FAF9F8FAF9F8FAF9F8FAF9080AF9080AF9080000000000000000000\
000000000000000000000000907F6F907F6F8F706F8F7060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000EFF0FFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000\
000000000000000000000000000000000000EFEFE0EFEFDFE0E0DF000000000000
000000000000000000000000000000000000000000000000000000000000AFA080A09F809F907F\
9F8F708F7F6F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000060607F7F7F908F90AF909FB0000000000000B0BFD0BFC0DFC0CFE0
CFD0EFD0DFEFDFE0F0DFE0F00000006F606070606F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000907070000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009F7F709F7F709F7F709F7F709F7F709F7F709F7F709F7F709F7F709F7F709F7F70\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000A08F8FAF8F90AF9090AF9090AF9090\
AF9090B09F9FB09F9FB09F9FB09F9FB09F9F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000BFA09FBFA09FBFA09FBFA09FBFA09FBFA09FBFA09F
BFA090B0A090B0A090B0A090B09F90B09F8FB09F8FAF9F8FAF9F8FAF9F8F000000000000000000\
0000000000000000000000000000000000000000008F7F6F8F7060000000000000
000000000000000000000000000000000000000000000000000000000000000000B0B0CF000000\
000000000000000000000000000000EFF0FFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000000000000000F0F0EFEFEFE0EFEFE0EFEFDFE0E0DF000000
000000000000000000000000000000000000000000000000000000000000AFA08FA09F80A0907F\
9F9070908F6F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006060707F7F908F90AF909FB0000000000000000000000000C0D0E0
6F606070606F70606F70606F70606F7F606F000000000000000000000000000000000000000000\
0000000000008F6F70907070907070907070907070907070907070907070907070
90707F90707F90707F90707F9F707F9F7F7F9F7F7F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
A0807FA0807FA0807FA0807FA07F70000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000B09F9FBF9F9FBFA0A0BFA0A0BFA0A0BFA0A0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000C0AFA0BFAFA0BFAFA0BFAF9FBFAF9FBFAF9FBFAF9F
BFAF9FBFAF9FBFA09FBFA090B0A090B0A090B0A090B0A090B09F8FB09F8F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000BFBFCF\
000000000000000000000000000000000000EFFFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000F0FFF0F0F0EFF0F0EFEFEFE0EFEFE0E0EFDFE0E0DF
000000000000000000000000000000000000000000000000000000000000000000AFA080A09F7F\
9F907F9F8F708F7F6F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000707F8F8F90AF9FA0BF0000006F60606F606F70606F70606F
7F606F7F606F7F606F80606F80606F0000000000000000000000008F6F708F6F708F6F708F6F70\
90707090707F90707F90707F90707F90707F90707F90707F907F7F9F7F7F9F7F7F
9F7F7F9F7F7F9F7F7F9F7F7F9F7F7F9F7F809F7F80000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
AF807FAF807FAF807FAF807FAF807FA0807FA0807FA0807FA0807FA0807FA08070
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000008F6F6F806F6FBFA0A0C0AFA0C0AFA0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000C0B0A0C0B0A0C0B0A0C0B0A0C0AFA0C0AFA0BFAFA0
BFAFA0BFAF9FBFAF9FBFAF9FBFAF9FBFAF9FB0A090B0A090B0A090B0A090000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
BFBFCF000000000000000000000000000000000000EFFFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFF0FFF0F0F0EFEFF0EFEFEFE0EFEFE0E0E0DF
000000000000000000000000000000000000000000000000000000000000000000AFA08FA09F80\
9F907F9F9070908F6F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F70806F60606F606070606F7F606F7F606F7F606F80606F
80606F806F6F806F6F0000008F6F708F6F708F6F708F6F708F6F708F707090707F90707F90707F\
90707F90707F907F7F907F809F7F809F7F809F7F809F7F809F7F809F7F809F7F80
9F7F809F7F809F80809F8080A08080A08080A08080000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
AF8F80AF8F80AF8F80AF8F80AF8F80AF8F80AF8F80AF8F80AF807FAF807FAF807F
AF807FAF807FAF807FA0807FA0807FA08070A08070000000000000000000000000000000000000\
00000000000000000090706F90706F8F6F6F8F6F6F806F6F000000000000000000
000000000000000000000000000000000000000000000000000000000000CFBFB0CFBFB0CFBFAF\
CFB0AFCFB0AFCFB0AFC0B0AF000000000000000000000000000000000000000000
000000000000C0B0A0BFAFA0BFAF9FBFAF9FBFAF9FBFAF9FB0AF90B0A090000000000000000000\
00000000000000000000000000000000000000000000000000000000000090806F
8F7F6F000000000000000000000000000000000000000000000000000000000000000000000000\
000000BFBFD0000000000000000000000000000000000000EFFFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0F0F0F0F0F0EFEFF0EFEFEFE0E0EFDF
E0E0DF000000000000000000000000000000000000000000000000000000000000000000A0A080\
A09F7F9F9070908F70807F60000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F60606F606070606F7F606F7F606F80606F80606F000000000000000000
0000000000000000008F6F708F6F708F707F8F707F8F707F90707F90707F90707F907F80907F80\
9F7F809F7F809F7F809F7F809F7F809F7F809F80809F808F9F808F9F808FA0808F
A0808FA0808FA0808FA0808FA0808FA0808FA0808F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
B08F8FB08F8FB08F8FB08F8FB08F8FAF8F8FAF8F80AF8F80AF8F80AF8F80AF8F80
AF8F80AF8F80AF8F80AF8F7FAF8F7FAF807FAF807FAF807FAF807FAF807FAF807FA08070A08070\
0000000000000000000000009F7F709F7F6F90706F90706F90706F8F6F6F8F6F6F
806F6F000000000000000000CFBFB0CFBFB0CFBFB0CFBFB0CFBFB0CFBFB0CFBFB0CFBFB0CFBFB0\
CFBFB0CFBFB0CFBFB0CFBFB0CFBFB0000000000000000000000000000000000000
000000000000000000000000000000000000000000BFAF9FBFAF9FBFAF9FB0AF9F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
9080708F7F6F000000000000000000000000000000000000000000000000000000000000000000\
000000000000BFC0D0000000000000000000000000000000000000EFFFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0F0F0F0F0F0EFEFEFE0EFEFE0
E0E0DF000000000000000000000000000000000000000000000000000000000000000000AFA08F\
A09F809F907F9F9070908F6F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F60606F60606F6060\
70606F7F606F7F606F7F606F000000000000000000000000000000000000000000
0000000000000000008F707F8F707F90707F907080907F80907F80907F80907F809F7F809F7F8F\
9F808F9F808F9F808F9F808FA0808FA0808FA0808FA0808FA0808FA0808FA08F8F
A08F90A08F90A08F90A08F90AF8F90AF8F90000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
B09090B09090B09090B0908FB0908FB0908FB0908FB0908FB0908FB0908FB0908F
B08F8FB08F80B08F80AF8F80AF8F80AF8F80AF8F80AF8F80AF8F80AF8F7FAF8F7FAF8F7FAF8F7F\
AF8F7FAF807FAF807FAF8070000000A08070A08070A07F709F7F6F9F7F6F90706F
90706F8F706F8F6F6F8F6F6F80606FD0C0BFD0C0BFD0C0BFD0C0BFD0C0BFD0C0BFD0C0BFD0C0BF\
D0C0BFD0C0BFD0C0BFD0C0BFCFC0B0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000BFAF9F000000\
000000000000000000000000000000000000000000000000000000000000000000
9F8F7090806F8F7F6F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000F0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0F0F0EFEFF0EFEFEFE0
E0EFDFE0E0DF000000000000000000000000000000000000000000000000000000000000000000\
A09F80A09F7F9F9070908F6F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006F60606F606070606070606F7F606F\
7F606F000000000000000000000000000000000000000000000000000000000000
000000000000907080907080907F80907F80907F80907F809F7F8F9F7F8F9F808F9F808F9F808F\
A0808FA08090A08090A08F90A08F90A08F90A08F90A08F90A08F90A08F90A08F90
AF8F90AF8F90AF8F90AF8F90AF8F90AF8F90000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
BF9F90BF9F90BF9F90BF9F90BF9F90BF9F90B09F90B09090B09090B0908FB0908F
B0908FB0908FB0908FB0908FB0908FB0908FB09080B09080B09080B08F80B08F80B08F80B08F80\
AF8F80AF8F7FAF8F7FAF8F7F000000000000000000000000A08070A08070A08070
A07F6F9F7F6F9F7F6F90706F90706F8F6F6F8F6F6F806F6FD0CFC0DFCFC0DFCFC0D0CFC0D0CFC0\
D0CFC0D0CFBFD0CFBFD0CFBFD0CFBF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000BFB0A0BFB09F\
BFAF9FBFAF9FB0AF90000000000000000000000000000000000000000000000000
9F8F7F9F8F7090806F8F7F6F000000000000000000000000000000000000000000000000000000\
000000000000000000B0B0CF000000000000000000000000000000000000000000
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFF000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0F0F0F0EFF0EFEFEFE0
EFEFE0E0E0DFE0E0D0000000000000000000000000000000000000000000000000000000000000\
AFA08FA09F809F907F908F70000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000070606070606F7F606F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000907F80907F8F907F8F9F7F8F9F808F9F808F9F808F9F80909F8090A08090A08090\
A08F90A08F90A08F90A08F90A08F90A08F90A08F90A08F90A08F90AF8F90AF8F9F
AF8F9FAF909FAF909FAF909FAF909FAF909F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
BF9F9FBF9F9FBF9F9FBF9F9FBF9F9FBF9F9FBF9F90BF9F90BF9F90BF9F90BF9F90
BF9F90BF9F90BF9F90BF9F90B09F8FB0908FB0908FB0908FB0908FB0908FB0908FB0908FB09080\
B09080B09080B09080B08F80000000000000000000000000000000000000000000
000000A08070A08070A07F6F9F7F6F9F7F6F9F706F90706F90706F8F6F6F80606FDFD0CFDFD0C0\
DFD0C0DFCFC0DFCFC0DFCFC0DFCFC0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000C0BFA0C0B0A0\
BFB0A0BFB0A0BFAF9FB0AF9FB0A090000000000000000000000000000000000000
A0907F9F8F7F908F7090806F8F7F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
EFF0FFF0FFFFF0FFFFF0FFFF000000000000000000000000000000000000000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0F0EFEFF0EF
EFEFE0E0EFDFE0E0DF000000000000000000000000000000000000000000000000000000000000\
000000A09F809F9F7F9F9070000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000070606F70606F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009F808F9F808F9F808F9F80909F80909F80909F80909F8090A08F90A08F90A08F90\
A08F90A08F9FA08F9FA08F9FA08F9FA08F9FAF8F9FAF8F9FAF909FAF909FAF909F
AF909FAF909FAF909FAF909FAF909FAF909F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
BF9F9FBF9F9FBFA09FBFA09FBF9F9FBF9F9FBF9F9FBF9F9FBF9F9FBF9F9FBF9F9F
BF9F90BF9F90BF9F90BF9F90BF9F90BF9F90BF9F90BF9F90BF9F90BF9F90BF9F8FBF9F8FBF9F8F\
BF9F8FB0908FB0908FB0908FB09080000000000000000000000000000000000000
000000000000000000000000000000A08070A0806FA07F6F9F7F6F9F7F6F90706F8F6F6F806F6F\
DFD0CFDFD0CFDFD0CFDFD0CFDFD0CF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000C0BFAFC0BFAF\
C0BFAFC0BFA0BFB0A0BFB09FB0AF9FB0AF90B0A090000000000000000000000000
A090809F907F9F8F70908F7090806F8F7F6F7F6F60000000000000000000000000000000000000\
000000000000000000000000000000000000C0CFE0D0E0FFD0E0FFDFEFFFE0EFFF
E0F0FF000000000000000000000000000000000000000000000000000000000000000000000000\
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0F0F0EFF0EF
EFEFE0E0EFE0E0E0DFE0E0D0000000000000000000000000000000000000000000000000000000\
000000A0A080A09F809F907F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000070606F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009F80909F80909F80909F80909F80909F80909F8090A08F90A08F90A08F9FA08F9FA08F9F\
A08F9FA08F9FA08F9FA0909FAF909FAF909FAF909FAF909FAF909FAF909FAF909F
AF909FAF909FAF90A0AF90A0AF90A0B090A0000000000000000000000000000000000000000000\
000000B09FA0B09FA0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
BFA0A0BFA0A0BFA0A0BFA0A0BFA09FBFA09FBFA09FBFA09FBFA09FBFA09FBFA09F
BFA09FBFA09FBFA09FBFA09FBFA09FBFA09FBF9F90BF9F90BF9F90BF9F90BF9F90BF9F90BF9F90\
BF9F90BF9F90BF9F90BF9F8FBF9F8F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000A0806FA0806F9F7F6F90706F\
90706F000000000000E0DFD0E0DFD0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000CFC0B0CFC0B0\
C0C0B0C0BFAFC0BFAFBFB0A0BFB0A0B0AF9FB0AF9FB0AF90AFA090AFA08F000000
A09F80A090809F907F9F8F70908F7090806F8F7F60000000000000000000000000000000000000\
000000000000000000000000000000000000000000C0D0EFD0E0FFD0E0FFDFEFFF
E0EFFFEFF0FF000000000000000000000000000000000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0FFF0F0F0EF
EFF0EFEFEFE0E0EFDFE0E0DF000000000000000000000000000000000000000000000000000000\
000000000000A09F80000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000070606F70606F7F606F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009F80909F80909F80909F80909F8F909F8F9FA08F9FA08F9FA08F9FA08F9FA08F9FA08F9F\
A0909FA0909FAF909FAF90A0000000000000000000000000000000000000000000
000000000000000000000000000000000000B09FA0B09FA0B09FA0B09FA0B09FA0B09FA0B09FA0\
B09FA0B0A0AFBFA0AFBFA0AFBFA0A0BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0
BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0BFA0A0000000000000\
C0A0A0C0A0A0C0A0A0C0A0A0C0A0A0C0A0A0C0A0A0C0A0A0C0A0A0C0A0A0C0A09F
C0A09FC0A09FC0A09FC0A09FC0A09FC0A09FC0A09FC0A09FBFA09FBFA09FBFA09FBFA090BFA090\
BFA090BFA090BFA090BFA090BF9F90000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000A0806F\
9F7F6F9F7F6F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000D0CFBFD0CFBFCFCFBF\
CFC0B0CFC0B0C0C0AFC0BFAFBFBFA0BFB0A0BFB09FB0AF9FB0AF90AFA090AFA08F
000000A09F80A090809F907F908F709080708F806F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFD0EFD0E0FFDFE0FF
DFEFFFE0F0FFEFF0FF000000000000000000000000000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0FFF0F0F0F0
EFF0EFEFEFE0E0EFE0E0E0DFDFE0D0000000000000000000000000000000000000000000000000\
000000000000A0A080000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000070606F
70606F7F60707F6F70000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009F80909F8F9F9F8F9F9F8F9FA08F9FA08F9FA08F9FA08F9F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000B09FAFB09FAFB09FAFB0A0AFB0A0AFB0A0AFBFA0AF\
BFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AF
BFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFC0A0AFC0A0AFC0A0AFC0A0AFC0A0AFC0AFAF\
000000000000000000000000000000000000000000000000C0A0A0C0A0A0C0A0A0
C0A0A0C0A0A0C0A0A0C0A0A0C0A09FC0A09FC0A09FC0A09FC0A09FC0A09FC0A09FC0A09FC0A09F\
C0A09FC0A09FC0A09FC0A090BFA090000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000A0806FA0806F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000D0D0C0D0D0C0D0CFBF\
D0CFBFCFCFBFCFC0B0C0C0B0C0BFAFC0BFAFBFB0A0BFB0A0B0B09FB0AF9FAFAF90
000000000000A09F809F907F9F907F908F7090806F8F806F000000000000000000000000000000\
000000000000000000000000000000000000000000000000C0CFE0CFDFF0D0E0FF
DFEFFFE0EFFFE0F0FFEFF0FF000000000000000000000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0FFF0F0FFF0
F0F0EFEFF0EFEFEFE0E0EFDFE0E0D0000000000000000000000000000000000000000000000000\
000000000000AFA08F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000070606F70606F7F6070
7F6F707F6F70000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009F8F9F9F8F9F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000B0A0AFB0A0AFB0A0AFB0A0AFBFA0AFBFA0AFBFA0AF\
BFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AFBFA0AF
BFAFAFBFAFAFC0AFAFC0AFAFC0AFAFC0AFAFC0AFAFC0AFAFC0AFAFC0AFAFC0AFAFC0AFAFC0AFAF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000C0AFA0C0AFA0C0AFA0C0AFA0C0AFA0C0A09FC0A09FC0A09FC0A09F\
C0A09FC0A09FC0A09FC0A09FC0A09F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000DFDFCFDFD0CFD0D0C0\
D0D0C0D0D0C0CFCFBFCFCFBFCFC0B0C0C0AFC0BFAFBFBFA0BFB0A0B0B09FB0AF9F
000000000000000000A09F809F907F908F7F908F708F806F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000CFD0EFD0E0FF
D0E0FFDFEFFFE0EFFFEFF0FF000000000000000000000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0FFF0F0FFF0
F0F0F0EFF0EFEFEFE0E0EFE0E0E0DFDFE0D0000000000000000000000000000000000000000000\
000000B0AF90AFA08F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F606F70606F7060707F6F707F6F7F
7F6F7F7F6F7F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000009F809F9F8F9F9F8F9F9F8F9F
9F8F9F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000B0A0AFB0A0B0BFA0B0BFA0B0BFA0B0BFA0B0BFA0B0\
BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0C0AFB0C0AFB0
C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFAFC0AFAFC0AFAFC0AFAF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000C0AFA0C0AFA0C0AFA0\
C0AFA0C0AF9FC0AF9FC0AF9FC0AF9F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000A08070A0806F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000DFDFCFDFDFCFDFDFCF\
DFDFCFD0D0C0D0D0C0D0CFBFCFCFBFCFC0B0C0C0B0C0C0AFBFBFAFBFBFA0BFB0A0
0000000000000000000000009F90809F907F908F70908F70000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000C0CFE0CFDFF0
D0E0FFDFEFFFE0EFFFE0F0FFEFF0FF000000000000000000000000000000000000000000000000\
000000000000000000000000FFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0FFF0F0FFF0
F0FFF0F0F0EFEFF0EFEFEFE0E0EFDFE0E0D0000000000000000000000000000000000000000000\
000000B0B09FAFA08F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F606F70606F70606F7060707F6F707F6F7F7F6F7F
806F7F000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000090809F9F809F9F809F9F8F9F9F8F9F9F8F9F9F8F9F9F8F9F9F8FA0
9F8FA0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000BFA0B0BFA0B0BFAFB0BFAFB0BFAFB0BFAFB0BFAFB0\
BFAFB0BFAFB0BFAFB0BFAFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0
C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0C0AFB0CFB0B0CFB0B0CFB0B0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000C0AFA0C0AFA0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000AF8F7FAF8F70A08F70A08070A0806F000000000000
000000000000EFEFE0EFEFDFEFEFDF000000000000000000000000000000E0E0D0DFDFD0DFDFD0\
DFDFCFDFDFCFDFDFCFD0D0C0D0D0C0CFCFBFCFCFBFC0C0B0C0C0B0C0BFAF000000
0000000000000000000000000000009F90809F907F908F70000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000CFD0EF
D0E0FFD0E0FFDFEFFFE0EFFFEFF0FF000000000000000000000000000000000000000000000000\
000000000000000000000000FFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0FFF0F0FFF0
F0FFF0F0FFF0EFF0EFEFEFE0E0EFE0E0E0DFDFE0D0000000000000000000000000000000000000\
BFBFA0B0B09FAFAF90A0A080000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F60606F606F6F606F70606F7060707F6F707F6F7F7F6F7F806F80806F80
807080000000000000000000000000000000000000000000000000000000000000908090908090\
90809F90809F9F809F9F8F9F9F8F9F9F8F9F9F8F9F9F8FA09F8FA09F8FA0A090A0
A090A0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000BFAFB0BFAFB0BFAFBFBFAFBFBFAFBFBFAFBFBFAFBF\
BFAFBFC0AFBFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BF
C0B0BFC0B0BFC0B0BFC0B0BFC0B0BFC0B0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0B0CFB0B0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000C0AFA0C0AFA0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000AF907FAF8F7FAF8F70A08F70A08070A0806F
000000000000000000000000EFEFE0EFEFE0E0EFDFE0EFDFE0E0DF000000000000E0E0D0E0E0D0\
DFE0D0DFDFCFDFDFCFDFDFCFD0D0C0D0D0C0D0CFBFCFCFBFCFCFB0C0C0B0000000
00000000000000000000000000000000000000000090907F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000C0CFDF
CFDFF0D0E0FFDFEFFFDFEFFFE0F0FFEFF0FF000000000000000000000000000000000000000000\
000000000000000000000000000000FFFFFFFFFFFFF0FFFFF0FFF0F0FFF0F0FFF0
F0FFF0F0FFF0F0F0EFEFF0EFEFEFE0E0EFDFE0E0D0000000000000000000000000000000C0C0AF\
BFBFA0B0B09FAFAF90A0A08F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000606060\
6F606F6F606F6F606F7060707060707F6F7F7F6F7F7F6F80806F80807080807080
807080000000000000000000000000000000000000000000907F9090809090809090809F90809F\
90809F9F8F9F9F8F9F9F8FA09F8FA09F8FA09F8FA09F8FA0A090A0A090A0A090A0
A090AF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000BFAFBFBFAFBFBFAFBFBFAFBFBFB0BFC0B0BFC0B0BF\
C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0B0BF
C0B0BFC0B0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BFCFB0BF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000CFAFA0CFAFA0C0AFA0C0AFA0C0AFA0C0AF9F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000B09080B0907FAF907FAF8F7FAF8F70A08F70
A08070A0806F000000000000000000EFEFE0EFEFE0EFEFE0E0EFDF000000000000000000000000\
000000DFE0D0DFDFD0DFDFCFD0DFCFD0D0C0D0D0C0D0D0C0CFCFBFCFCFBF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
CFD0E0D0E0FFD0E0FFDFEFFFE0EFFFEFF0FFEFF0FF000000000000000000000000000000000000\
000000000000000000000000000000000000FFFFFFFFFFFFF0FFFFF0FFF0F0FFF0
F0FFF0F0FFF0F0FFF0EFF0EFEFF0E0E0EFE0E0E0DFDFE0D0000000000000000000CFCFBFC0C0B0\
BFBFAFB0B09FAFAF90A0A08F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006060606F606F6F606F\
6F606F70606F7060707060707F6F7F7F6F7F7F6F80806F8080708080708F80708F
80708F0000000000000000000000008F7F908F7F9090809090809090809F90809F90809F908F9F\
9F8F9F9F8FA09F8FA09F8FA09F8FA09F90A09F90A0A090A0A090AFA090AFA090AF
A090AF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000BFB0BFBFB0C0BFB0C0C0B0C0C0B0C0C0B0C0C0B0C0\
C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0B0C0C0BFC0CFBFC0CFBFC0CFBFC0
CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0CFBFC0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000CFB0AFCFB0AFCFB0A0CFB0A0CFAFA0C0AFA0
C0AFA0C0AFA0C0AF9FC0AF9F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000B09F80B09F80B09080B0907FAF907FAF8F7F
AF8F70A08F70A08F70A0806F000000000000000000EFEFE0EFEFE0000000000000000000000000\
000000000000000000000000DFDFCFD0DFCFD0DFCFD0D0C0D0D0C0CFD0BF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000CFD0EFD0E0FFDFEFFFDFEFFFE0EFFFEFF0FF000000000000000000000000000000000000\
000000000000000000000000000000000000FFFFFFFFFFFFF0FFFFF0FFF0F0FFF0
F0FFF0F0FFF0F0FFF0F0F0EFEFF0EFEFEFE0E0EFDFE0E0DF000000DFDFCFD0D0C0CFD0BFC0CFB0\
BFBFAFB0B0A0AFAF90A0A08F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006060606F606F6F606F6F606F\
706070706070706F7F7F6F7F7F6F7F7F6F8080708080708080708F80708F80708F
8F7F8F8F7F908F7F908F7F908F7F908F809090809F90809F90809F90809F908F9F908FA09F8FA0\
9F8FA09F8FA09F90A09F90A09F90AFA090AFA090AFA090AFA090AFA090AFA09FAF
A09FAF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000C0B0C0C0B0C0C0B0C0C0B0C0C0B0CFC0BFCFC0BFCF\
C0BFCFC0BFCFC0BFCFC0BFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCF
CFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFCFBFCFD0C0CFD0C0C0D0C0C0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000CFB0AFCFB0AFCFB0AFCFB0AFCFB0A0CFB0A0
CFB0A0C0B0A0C0AFA0C0AFA0C0AFA0C0AF9FC0AF9F000000000000000000000000000000000000\
000000000000000000000000000000BFA08FBF9F8FB09F80B09F80B09080AF907F
AF907FAF907FAF8F70A08F70A0806F9F806F000000000000806F60000000000000000000000000\
000000000000000000000000000000000000D0DFCFD0DFCFD0D0C0000000000000
00000000000000000000000000000000000000000000000090907F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000C0CFE0D0DFF0DFE0FFDFEFFFE0EFFFEFF0FFEFF0FF000000000000000000000000000000\
000000000000000000000000000000000000000000FFFFFFFFFFFFF0FFFFF0FFF0
F0FFF0F0FFF0F0FFF0F0FFF0EFF0EFEFF0E0E0EFE0000000000000DFDFCFD0DFC0D0D0C0C0CFB0\
C0C0AFBFBFA0B0B09FAFAF8FA09F80000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F606F6F606F6F6070706070\
70607F7F6F7F7F6F7F7F6F807F6F807F708080708F80708F80708F000000000000
8F7F908F7F908F7F908F80908F809F90809F90809F90809F908F9F908FA0908FA09F8FA09F8FA0\
9F8FA09F90AF9F90AF9F90AFA090AFA090AFA090AFA09FAFA09FAFA09FB0A09FB0
A09FB0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000C0B0CFC0BFCFC0BFCFC0BFCFC0BFCFC0BFCFC0BFCF\
C0BFD0CFBFD0CFBFD0CFBFD0CFBFD0CFBFD0CFBFD0CFBFCFCFBFCFCFBFCFCFBFCF
CFC0CFCFC0CFCFC0CFCFC0CFCFC0CFD0C0CFD0C0CFD0C0CFD0C0CFD0C0CFD0C0CFD0C0CF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000CFBFB0CFB0AFCFB0AFCFB0AFCFB0AFCFB0AF
CFB0AFCFB0A0CFB0A0CFB0A0C0B0A0C0B0A0C0AFA0C0AFA0C0AF9FC0AF9FC0AF9F000000000000\
000000000000000000000000000000BFA08FBFA08FBFA08FBF9F8FB09F80B09F80
B09F80AF907FAF907FAF907FAF8F70A08F70A0806F9F806F0000008F7060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000009F9080000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000CFD0EFD0E0FFDFEFFFE0EFFFE0EFFFEFF0FF000000000000000000000000000000\
000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFF0FFF0
F0FFF0F0FFF0F0FFF0F0FFF0F0F0EFEFF0EF000000000000000000DFDFCFD0DFCFD0D0C0CFCFBF\
C0C0AFBFBFA0B0B09FAFAF90A0A080000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006F607070607070607070607F\
706F7F7F6F7F7F6F807F6F807F708080708F80708F000000000000000000000000
8F7F908F7F908F809F8F809F90809F90809F90809F908FA0908FA0908FA09F8FA09F8FA09F90AF\
9F90AF9F90AF9F90AFA090AFA090AFA09FB0A09FB0A09FB0A09FB0A09FB0AF9FB0
AF9FB0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000C0BFCFC0BFCFC0BFD0C0BFD0C0BFD0CFBFD0CFC0D0\
CFC0D0CFC0D0CFC0D0CFC0D0CFC0D0CFC0D0CFC0D0CFC0D0CFC0D0CFC0D0CFC0D0
CFC0D0D0C0D0D0C0D0D0C0D0D0C0D0D0C0D0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0D0CFD0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000D0BFB0CFBFB0CFBFB0CFBFB0CFBFAFCFB0AF
CFB0AFCFB0AFCFB0AFCFB0AFCFB0A0CFB0A0C0B0A0C0B0A0C0B0A0C0B0A0C0AF9FC0AF9FC0AF9F\
C0AF9F000000000000000000000000BFA090BFA090BFA08FBFA08FBFA08FB09F8F
B09F80B09F80B09F80AF907FAF907FAF8F70A08F70A080709F806F0000008F706F000000000000\
000000000000000000000000000000000000000000000000DFDFD0D0DFCF000000
0000000000000000000000000000000000000000000000009F9F8090907F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000C0CFE0D0DFF0DFE0FFDFEFFFE0EFFFEFF0FFEFF0FF000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFF0FFFF
F0FFF0F0FFF0F0FFF0F0FFF0000000000000000000000000000000DFDFD0D0DFCFD0D0C0CFCFBF\
C0C0B0BFBFA0B0B09FAFAF90A0A08F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006F607070607070607F706F7F706F7F\
7F6F807F6F807F6F807F7080000000000000000000000000000000000000000000
8F809F8F809F8F809F90809F90809F908FA0908FA0908FA0908FA09F8FA09F90AF9F90AF9F90AF\
9F90AF9F90AFA09FB0A09FB0A09FB0A09FB0A09FB0A09FB0A09FB0AF9FB0AFA0B0
AFA0BF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000C0BFD0C0BFD0C0C0D0CFC0D0CFC0D0CFC0DFCFC0DF\
CFC0DFCFC0DFCFC0DFCFC0DFCFC0DFCFC0DFCFC0DFCFC0DFD0C0DFD0CFDFD0CFDF
D0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0D0D0DFD0D0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000D0C0BFD0C0BFD0C0BFD0C0B0D0BFB0CFBFB0CFBFB0
CFBFB0CFBFAFCFBFAFCFB0AFCFB0AFCFB0AFCFB0AFCFB0A0C0B0A0C0B0A0C0B0A0C0B0A0C0B0A0\
C0B09FC0AF9FC0AF9FC0AF9F000000BFAF90BFA090BFA090BFA08FBFA08FBFA08F
BFA08FB09F8FB09F80B09F80B09F80AF907FAF907FA08F70A08F709F806F9F806F907F6F8F7060\
000000000000000000000000000000000000000000DFE0DFDFE0D0D0DFCFD0DFCF
000000000000000000000000000000000000000000AFAF909F9F8F90907F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000B0B0CF000000CFD0EFD0E0FFDFEFFFE0EFFFE0EFFFEFF0FFF0F0FF000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
F0FFF0F0FFF0000000000000000000000000000000000000000000000000D0DFCFD0D0C0CFD0BF\
C0C0B0BFBFAFB0B09FAFAF90A0A08F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006060600000000000006F607070607070607F706F7F706F7F7F6F80\
7F6F807F6F80000000000000000000000000000000000000000000000000000000
8F809F8F809F90809F908FA0908FA0908FA0908FA0908FA09F8FAF9F90AF9F90AF9F90AF9F90AF\
9F90B0A09FB0A09FB0A09FB0A09FB0A09FB0A09FB0AFA0BFAFA0BFAFA0BFAFA0BF
AFA0BF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000CFC0DFCFC0DFCFC0DFCFC0DFCFC0DF000000000000\
000000000000000000000000D0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDFD0CFDF
D0CFDFD0CFDFD0D0DFD0D0DFD0D0DFD0D0DFD0D0DFDFD0DFDFD0DFDFD0DFDFD0DFDFD0DF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000D0CFC0D0CFBFD0C0BFD0C0BFD0C0BFD0C0BFD0C0B0
D0C0B0CFBFB0CFBFB0CFBFB0CFBFAFCFBFAFCFBFAFCFB0AFCFB0AFCFB0A0C0B0A0C0B0A0C0B0A0\
C0B0A0C0B0A0C0B09FC0AF9FC0AF9F000000BFAF90BFAF90BFAF90BFA090BFA08F
BFA08FBFA08FB0A08FB09F80B09F80B09F80AF907FAF907FA08F7FA08F70A080700000009F7F6F\
90706F000000000000000000000000000000000000DFEFDFDFE0D0D0DFD0D0DFCF
D0D0CF000000000000000000000000000000000000AFAF9FA0A08F909F808F8F70000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009F9FAFBFBFD0000000CFDFF0DFE0FFDFEFFFE0EFFFEFF0FFEFF0FF000000000000000000\
000000000000000000000000000000000000000000000000000000FFFFFF000000
000000000000000000000000000000000000000000000000000000000000D0DFCFD0D0C0CFD0C0\
C0CFB0BFC0AFB0B0A0AFAF90A0A08F9F9F80000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000070607F70607F706F7F706F807F6F807F6F80\
000000000000000000000000000000000000000000000000000000000000000000
8F809F9080A0908FA0908FA0908FA0908FA0908FAF9F90AF9F90AF9F90AF9F90AF9F90B09F9FB0\
A09FB0A09FB0A09FB0A09FBFA0A0BFA0A0BFAFA0BFAFA0BFAFA0BFAFA0BFAFA0BF
AFA0BF000000000000000000000000000000000000000000000000000000000000000000000000\
000000BFBFCFBFBFD0BFBFD0BFBFD0C0BFD0C0BFD0C0BFD0C0BFD0C0BFD0C0BFD0
C0C0DFC0C0DFC0C0DFC0C0DFCFC0DFCFC0DF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000DFDFE0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000DFCFC0D0CFC0D0CFC0D0CFC0D0CFBFD0CFBFD0C0BF
D0C0BFD0C0BFD0C0B0CFC0B0CFC0B0CFBFB0CFBFB0CFBFAFCFBFAFCFBFAFCFBFAFCFB0AFCFB0A0\
C0B0A0C0B0A0C0B0A0C0B09FC0B09F000000000000000000000000BFAF90BFAF90
BFA090BFA08FBFA08FB0A08FB0A08FB09F80B09F80B09F80AF907FAF907FA08F70000000000000\
000000907F6F000000000000000000000000E0EFE0E0EFDFDFE0DFDFE0D0D0DFD0
D0DFCFCFD0CF000000000000000000000000000000AFB09FA0A0909F9F8090907F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000B0B0D0C0C0DF000000D0E0FFDFEFFFE0EFFFE0EFFFEFF0FFF0F0FF000000000000\
000000000000000000000000000000000000000000FFFFFFFFFFFF000000000000
000000000000000000000000000000000000000000000000000000000000DFDFCFD0DFCFCFD0C0\
CFCFBFBFC0AFB0BFA0AFB09FA0AF8F9FA080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000070607F706F7F706F807F6F80000000000000\
000000000000000000000000000000000000000000000000000000000000000000
908FA0908FA0908FA0908FA0908FAF9090AF9F90AF9F90AF9F90AF9F90B09F9FB09F9FB0A09FB0\
A09FB0A09FBFA0A0BFA0A0BFAFA0BFAFA0BFAFA0BFAFA0BFAFA0BFAFAFC0AFAFC0
AFAFC0000000000000000000000000000000000000B0B0CFBFB0CFBFB0CFBFB0CFBFBFD0BFBFD0\
BFBFD0BFBFD0C0BFD0C0BFD0C0BFD0C0BFD0C0C0DFC0C0DFC0C0DFC0C0DFC0C0DF
CFC0DFCFC0DFCFC0DFCFC0DFCFCFE0CFCFE0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000DFE0E0\
E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0DFE0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000DFD0CFDFD0C0DFD0C0DFD0C0D0CFC0D0CFC0D0CFBF
D0CFBFD0CFBFD0CFBFD0C0BFD0C0BFD0C0B0CFC0B0CFC0B0CFC0B0CFBFB0CFBFAFCFBFAFCFBFAF\
CFBFAFC0B0A0C0B0A0C0B0A0C0B0A0000000000000000000000000000000000000
BFAF90BFAF90BFA08FBFA08FBFA08FB0A08FB0A080B09F80B09F80AF907FAF907FA08F70000000\
000000000000000000000000000000E0F0EFE0F0E0E0EFE0DFEFDFDFE0DFDFE0D0
D0DFD0D0DFCFCFD0C0CFD0C0000000000000BFC0AFB0B0A0AFAF909FA08F90907F8F8F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000B0B0CFBFC0DF000000000000000000000000000000000000000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000000000000000000000000000000000000000000000DFDFD0D0DFCFD0D0C0\
CFCFBFC0C0B0BFBFA0B0B09FAFAF90A0A080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000706F7F706F80000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
908FA0908FA0908FAF908FAF9090AF9F90AF9F90AF9F90AF9F90B09F9FB09F9FB0A09FB0A09FBF\
A09FBFA0A0BFA0A0BFAFA0BFAFA0BFAFA0C0AFAFC0AFAFC0AFAFC0AFAFC0B0AFC0
000000B0AFCFB0B0CFB0B0CFB0B0CFB0B0CFBFB0CFBFB0CFBFBFD0BFBFD0BFBFD0BFBFD0C0BFD0\
C0BFD0C0C0DFC0C0DFC0C0DFC0C0DFC0C0DFC0C0DFCFC0DFCFC0DFCFCFDFCFCFE0
CFCFE0CFCFE0CFCFE0CFCFE0CFCFE0CFCFE0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000E0E0EF\
E0E0EFE0E0EFE0E0EFE0E0EFE0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0
E0E0E0E0E0E0E0E0DFE0E0DF000000000000000000000000000000000000000000000000000000\
000000000000000000000000DFDFCFDFDFCFDFD0CFDFD0CFDFD0C0DFD0C0DFD0C0
D0D0C0D0CFC0D0CFBFD0CFBFD0CFBFD0CFBFD0C0BFD0C0BFD0C0B0CFC0B0CFC0B0CFC0B0CFBFB0\
CFBFAFCFBFAFCFBFAFC0B0A0C0B0A0000000000000000000000000000000000000
000000000000000000BFAF90BFA08FBFA08FB0A08FB0A080B09F80B09F80AF9F7FAF907F000000\
000000000000000000000000E0F0EFE0F0EFE0F0EFE0EFE0E0EFE0DFEFDFDFE0DF
D0E0D0D0DFD0D0DFCFCFD0C0CFD0C0C0CFBFBFC0B0B0BFA0AFAF9FA0A090909F808F8F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000A0A0B0BFBFD0000000000000000000000000000000000000000000000000F0FFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000000000000000000000000000000000000000000000000000D0DFCFD0D0C0\
CFD0BFC0C0B0BFBFAFB0B09FAFAF90A0A08F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000706F7F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000908FAF9090AF9090AF9F90AF9F90AF9F90B09F9FB09F9FB09F9FB0A09FBFA09FBFA0A0BF\
A0A0BFA0A0BFAFA0C0AFA0C0000000000000000000000000000000000000000000
000000B0B0CFB0B0CFB0B0CFBFB0D0BFBFD0BFBFD0BFBFD0BFBFD0BFBFD0C0C0DFC0C0DFC0C0DF\
C0C0DFC0C0DFC0C0DFC0C0DFCFCFDFCFCFE0CFCFE0CFCFE0CFCFE0CFCFE0CFCFE0
CFCFE0CFCFE0D0D0E0D0D0EFD0D0EFD0D0EF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000E0E0EF\
E0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0E0E0E0E0E0E0E0E0E0E0
E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0E0DFE0E0DFE0E0DFE0E0DF000000000000\
000000000000000000000000DFDFD0DFDFD0DFDFCFDFDFCFDFDFCFDFD0CFDFD0CF
DFD0C0DFD0C0DFD0C0D0D0C0D0CFC0D0CFC0D0CFBFD0CFBFD0CFBFD0CFBFD0C0BFD0C0B0CFC0B0\
CFC0B0CFBFB0CFBFAFCFBFAFC0BFAF000000000000000000000000000000000000
000000000000000000000000000000BFAF90B0A08FB0A08FB0A080B09F80B09F80AF9F7F000000\
000000000000000000000000000000E0F0EFE0F0EFE0EFE0E0EFE0DFEFE0DFEFDF
DFE0DFD0E0D0D0DFCFCFD0CFCFD0C0C0CFBFBFC0B0BFBFAFAFB0A0A0AF909F9F8F8F8F70000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000090909FB0BFD0C0C0DF000000000000000000000000000000000000000000F0FFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000000000000000000000000000000000000000000000000000D0DFCFD0D0C0\
CFD0BFC0CFB0BFC0AFB0B0A0AFAF90A0A08F9F9F80000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009090AF9090AF9F90AF9F90B09F9FB09F9FB09F9FB09F9FBFA09FBFA0A0BFA0A0BFA0A0BF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000BFBFD0BFBFD0BFBFD0BFBFD0BFBFD0C0C0DFC0C0DFC0C0DFC0C0DFC0C0DFC0C0DF\
C0CFDFCFCFE0CFCFE0CFCFE0CFCFE0CFCFE0CFCFE0CFCFE0CFD0E0D0D0EFD0D0EF
D0D0EFD0D0EFD0D0EFD0D0EFD0D0EFD0D0EF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000E0EFEF\
EFEFEFEFEFEFEFEFEFEFEFEFEFEFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFE0
E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFDFE0EFDF\
E0EFDFE0E0DFE0E0DF000000E0E0D0E0E0D0E0E0D0DFDFD0DFDFD0DFDFD0DFDFCF
DFDFCFDFDFCFDFD0CFDFD0CFDFD0C0DFD0C0D0D0C0D0CFC0D0CFC0D0CFBFD0CFBFD0CFBFD0C0BF\
D0C0B0CFC0B0CFC0B0CFBFAFCFBFAF000000000000000000000000000000000000
000000000000000000000000000000000000000000B0A08FB0A08FB0A08FB09F80AF9F80AF9F7F\
000000000000000000000000000000000000E0F0EFE0EFEFE0EFE0DFEFE0DFEFE0
DFEFDFDFE0DFD0DFD0CFDFCF000000000000BFC0B0BFC0AFB0B0A0A0AF9F9F9F8F8F8F70000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000B0B0D0BFC0DF000000000000000000000000000000000000000000F0FFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000000000000000D0DFCFD0DFC0\
CFD0C0C0CFBFBFC0AFB0BFA0AFB09FA0A08F9F9F80000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000706F80000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009090B09F90B09F90B09F9FB09F9FB09F9FB09F9FBFA09FBF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000BFBFD0BFBFDFBFC0DFC0C0DFC0C0DFC0C0DFC0C0DFC0C0E0C0CFE0CFCFE0CFCFE0\
CFCFE0CFCFE0CFCFE0CFD0E0CFD0EFCFD0EFD0D0EFD0D0EFD0D0EFD0D0EFD0D0EF
D0D0EFD0DFEFD0DFF0D0DFF0DFDFF0DFDFF0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000EFEFF0\
EFEFF0EFEFF0EFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEF
EFEFEFEFEFEFEFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0E0EFE0\
E0EFE0E0EFE0E0EFDF000000000000000000000000E0E0DFE0E0D0E0E0D0E0DFD0
DFDFD0DFDFD0DFDFCFDFDFCFDFDFCFDFD0CFDFD0CFDFD0C0DFD0C0DFD0C0D0CFC0D0CFC0D0CFBF\
D0CFBFD0CFBFD0C0B0CFC0B0CFC0B0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000B0A08FB0A08FB0A08FB0A080AF9F80\
000000000000000000000000000000000000000000E0F0EFDFEFE0DFEFE0DFEFE0
DFEFE0DFE0DFD0E0D0000000000000000000BFCFBFBFC0B0B0BFAFAFB09F9F9F8F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000A0A0B0BFBFDF000000000000000000000000000000000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000000000000000000000D0DFCF\
CFD0C0CFCFBFC0C0B0B0BFA0AFB09FA0AF909FA080000000808060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006F607F706F80706F80000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009F90B09F9FB09F9FB09F9FB09F9FBF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000C0C0DFC0C0DFC0C0DFC0C0E0C0CFE0C0CFE0CFCFE0CFCFE0CFCFE0CFCFE0CFD0E0\
CFD0EFCFD0EFD0D0EFD0D0EFD0D0EFD0D0EFD0D0EFD0DFEFD0DFF0D0DFF0D0DFF0
DFDFF0DFDFF0DFDFF0DFDFF0DFDFF0DFE0F0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000EFEFF0EFF0F0\
EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0EFEFF0EFEFF0EFEFEFEFEFEFEFEFEFEF
EFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0\
EFEFE0EFEFE0EFEFE0000000000000000000000000000000000000000000000000
E0E0D0E0E0D0E0DFD0E0DFD0DFDFD0DFDFCFDFDFCFDFDFCFDFD0CFDFD0CFDFD0C0DFD0C0D0D0C0\
D0CFC0D0CFBFD0CFBFD0CFBFCFC0B0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000B0A08FB0A08FB0A080\
000000000000000000000000000000000000000000000000E0F0EFDFEFE0DFEFE0
DFEFE0DFEFDF000000000000000000000000000000BFC0B0B0BFAFAFB0A09FA08F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000090909FB0BFD0C0C0E0000000000000000000000000000000000000000000\
F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000000000000000000000D0DFCF\
CFD0C0CFD0BFC0C0B0BFBFAFB0B09FAFAF909FA080000000808060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006F607F706F80706F8070708F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009F9FB0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000C0CFE0C0CFE0C0CFE0CFCFE0CFCFE0CFCFE0CFD0EFCFD0EFCFD0EFCFD0EF\
D0D0EFD0D0EFD0DFEFD0DFF0D0DFF0D0DFF0D0DFF0D0DFF0DFDFF0DFDFF0DFDFF0
DFE0F0DFE0F0DFE0F0DFE0F0DFE0F0DFE0F0DFE0F0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000EFF0F0EFF0F0\
EFF0F0F0F0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0EFEFF0EFEFF0EF
EFF0EFEFF0EFEFF0EFEFF0EFEFF0EFEFF0EFEFF0EFEFF0EFEFF0EFEFF0EFEFF0EFEFEFEFEFEFE0\
EFEFE0EFEFE0EFEFE0000000000000000000000000000000000000000000000000
000000000000000000000000E0E0D0E0DFD0E0DFD0DFDFD0DFDFD0DFDFCFDFDFCFDFD0CFDFD0CF\
DFD0C0D0D0C0D0CFC0D0CFBFD0CFBF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000B0A08F\
000000000000000000000000000000000000000000000000000000E0F0EFDFEFE0
DFEFE0000000000000000000000000000000000000BFC0B0B0C0AFAFB0A09FA08F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000909090B0B0D0BFC0DF000000000000000000000000000000000000000000\
F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000000000000000000000D0DFCF\
CFD0C0CFD0BFC0CFB0BFC0AFB0B0A0AFAF909FA080000000808060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006F607F6F6F80706F80706F8F70708F000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000009090B09F9FB0
9F9FB0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CFCFE0CFCFEFCFD0EFCFD0EFCFD0EFCFD0EFCFD0EFD0D0EFD0DFEFD0DFF0\
D0DFF0D0DFF0D0DFF0D0DFF0DFDFF0DFE0F0DFE0F0DFE0F0DFE0FFDFE0FFDFE0FF
DFE0FFDFE0FFDFE0FFE0E0FFE0EFFFE0EFFFE0EFFF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000F0F0F0F0F0F0\
F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0
F0F0F0F0F0F0F0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFEFF0EFEFF0EFEFF0EFEFF0EF\
EFF0EFEFF0EFEFEFEF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000E0DFD0E0DFD0E0DFD0DFDFCFDFDFCF\
DFD0CFDFD0CFDFD0C0D0D0C0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000008F706F000000
000000000000000000000000000000000000000000000000B0BFAFAFB0A0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000A0AFBFBFBFDF000000000000000000000000000000000000000000\
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFF000000000000000000000000000000000000000000000000000000000000000000D0DFCF\
D0DFC0CFD0C0C0CFB0BFC0AFB0BFA0AFAF909FA080909070808060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6F607F6F6080706F80706F8F70708F7F708F000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000009090B09090B09F9FBF9F9FBF
9F9FBF9FA0BF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CFD0EFCFD0EFCFD0EFCFD0EFD0D0EFD0DFF0D0DFF0D0DFF0D0DFF0D0DFF0\
D0DFF0DFE0F0DFE0FFDFE0FFDFE0FFDFE0FFDFE0FFDFE0FFDFE0FFDFE0FFE0EFFF
E0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000F0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0F0F0F0F0F0F0F0F0
F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0EFF0F0EFF0F0EFF0F0EF\
F0F0EFF0F0EF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000DFDFD0\
DFDFCFDFDFCFDFD0CFDFD0CF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
B0A08FB0A08F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000B0BFAFAFB0A0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000A0A0B0B0BFD0C0CFE0000000000000000000000000000000000000\
000000F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFF000000000000000000000000000000000000000000000000000000000000000000000000\
D0DFCFCFD0C0C0CFBFBFC0AFB0BFA0AFAF909FA080909070808060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006060706F607F
6F607F6F6F80706F8F70708F70708F7F7090000000000000000000000000000000000000000000\
0000000000000000000000000000009090B09090B0909FB09F9FBF9F9FBF9FA0BF
A0A0C0A0A0C0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CFD0F0CFD0F0D0DFF0D0DFF0D0DFF0D0DFF0D0DFF0D0DFF0D0E0FFDFE0FF\
DFE0FFDFE0FFDFE0FFDFE0FFDFE0FFDFEFFFDFEFFFE0EFFFE0EFFFE0EFFFE0EFFF
E0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0F0FFE0F0FF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000F0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFF0
F0FFF0F0FFF0F0FFF0F0FFF0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0\
F0F0F0F0F0EF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000DFDFD0DFDFCFDFDFCF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
B0AF8FB0A08FB0A08FAFA0800000000000000000000000000000000000008F7F6F
000000000000000000000000000000000000000000000000B0BFAF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000A0A0B0B0B0D0BFC0DF000000000000000000000000000000000000\
000000F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFF000000000000000000000000000000000000000000000000000000000000000000000000\
D0DFCFCFD0C0CFCFBFBFC0AFB0BFA0AFAF900000008F9070807F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006060706F607F6F607F
6F6F80706F80706F8F70708F7070907F7F907F7F90000000000000000000000000000000000000\
0000000000000000008F90AF9090B09090B0909FB09F9FBF9FA0BF9FA0C0A0A0C0
A0AFC0A0AFCFAFAFCF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000D0DFF0D0DFF0D0DFF0D0E0FFD0E0FFDFE0FFDFE0FFDFE0FFDFE0FF\
DFE0FFDFEFFFDFEFFFDFEFFFDFEFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFF
E0EFFFE0F0FFE0F0FFE0F0FFE0F0FFEFF0FFEFF0FF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000F0FFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0F0F0F0F0F0\
F0F0F0F0F0F0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
B0AF90B0AF8FB0A08FAFA08FAFA08000000000000000000000000000000090806F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000B0B0D0BFC0DF000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000000000\
D0DFCFCFD0C0CFCFBFBFC0AFB0BFA00000000000008F90707F7F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000060606F00000060607F6F607F6F6F80
6F6F80706F8F70708F7070907F70907F7F907F7F9F000000000000000000000000000000000000\
8F8FAF8F8FAF8F90AF9090B0909FB09F9FBF9F9FBF9FA0BFA0A0C0A0AFC0A0AFCF
AFAFCFAFAFCFAFB0CF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000D0E0F0D0E0FFD0E0FFDFE0FFDFE0FFDFE0FFDFE0FFDFEFFFDFEFFF\
DFEFFFDFEFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFF000000000000000000
000000000000000000000000000000000000000000000000EFF0FFEFF0FFEFF0FFEFF0FFEFF0FF\
EFF0FFF0F0FFF0F0FFF0F0FFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000000000FFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFF0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000E0E0D0E0E0D0DFDFD0DFDFD0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
B0AF90B0AF90B0AF8FB0A08FAFA08FAFA080000000000000000000000000000000
90806F000000000000000000000000000000000000BFC0B0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000B0B0CFBFC0DF000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000000000\
D0DFCFCFD0C0C0CFBFBFC0AF0000000000000000008F90707F7F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000060606060606F6060706F607F6F60806F6F80
6F6F8F70708F70708F7070907F7F907F7F9F7F7F9F0000000000000000000000008F8FAF8F8FAF\
8F90AF8F90B09090B0909FB09F9FBF9FA0BF9FA0C0A0A0C0A0AFC0A0AFCFAFAFCF
AFB0CFAFB0D0AFB0D0B0BFD0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000DFE0FFDFE0FFDFE0FFDFEFFFDFEFFFDFEFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000EFFFFFEFFFFFEFFFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000E0E0DFE0E0DFE0E0D0E0E0D0DFE0D0DFDFD0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000BFB09F\
BFB09FB0AF90B0AF90B0AF90B0AF8FAFA08FAFA080000000000000000000000000
908F6F000000000000000000000000000000C0CFC0BFC0B0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000BFC0DFC0CFE0000000000000000000000000000000\
000000000000F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000000000\
D0DFCFCFD0C0C0CFBF0000000000000000000000008F90707F7F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000060606F60607060607F6F60806F6F806F6F8F
706F8F70708F7070907070907F7F907F7F9F7F809F000000000000808FA0808FAF8F8FAF8F90AF\
9090B0909FB0909FBF9F9FBF9FA0BF9FA0C0A0AFC0A0AFCFA0AFCFAFB0CFAFB0D0
AFB0D0B0BFD0B0BFDFB0BFDF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000DFE0FFDFE0FF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000EFFFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000EFEFE0E0E0DFE0E0DFE0E0DFE0E0DFE0E0D0E0E0D0DFE0D0DFDFD0
DFDFCF000000000000000000000000000000000000000000000000000000000000000000BFBFA0\
BFB09FBFB09FB0B090B0AF90B0AF90AFA08FAFA080AF9F80000000000000000000
9F8F708F7F6F000000000000D0DFD0CFDFCFC0CFC0BFC0B0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000C0CFE0000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000\
D0DFCFCFD0C00000000000000000000000000000008F90707F7F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060606060606F0000006F607F6F6F806F6F806F6F8F
70708F7070907070907F7F907F7F9F7F7F9F8080A08080A0808FA08F8FAF8F8FAF8F90B09090B0\
909FB0909FBF9FA0BF9FA0C0A0A0C0A0AFC0A0AFCFAFAFCFAFB0D0AFB0D0AFBFD0
B0BFDFB0BFDFB0C0DFBFC0DF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000D0E0F0D0E0F0D0E0F0D0E0F0
D0E0FFDFE0FFDFEFFFDFEFFFDFEFFFDFEFFF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000EFEFE0EFEFE0EFEFDFE0EFDFE0E0DFE0E0DFE0E0DFE0E0D0E0E0D0
DFE0D0DFE0D0DFDFCF000000000000000000000000000000000000000000000000000000C0BFAF\
BFBFA0BFBFA0BFB09FB0B09FB0AF90AFAF8FAFA08FAFA080A09F80000000000000
9F90709F8F6FD0E0E0D0E0DFCFDFD0CFDFCFC0CFC0BFC0B0000000000000707F60000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000CFD0EF000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000008F90707F7F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060606F0000000000006F60806F6F806F6F8F6F6F8F
70708F707090707F907F7F9F7F7F9F7F809F0000000000008F8FAF8F90AF8F90B09090B0909FBF\
909FBF9FA0BF9FA0C0A0AFC0A0AFCFA0AFCFAFB0CFAFB0D0AFB0D0B0BFD0B0BFDF
B0BFDFBFC0E0BFC0E0BFC0E0BFCFE0000000000000000000000000000000000000000000000000\
000000000000000000D0DFF0D0E0F0D0E0F0D0E0F0D0E0F0D0E0F0DFE0F0DFEFFF
DFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFF0FFE0F0FF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
FFFFFF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000EFEFEFEFEFE0EFEFE0EFEFE0EFEFE0EFEFDFE0EFDFE0E0DFE0E0DFE0E0DF
E0E0D0E0E0D0DFE0D0DFE0D0DFDFCF000000000000000000000000000000000000C0C0B0C0C0AF\
C0C0AFBFBFA0BFBFA0BFB09FB0B09FB0AF90AFAF8FAFA08FA0A080A09F7F000000
9F90709F907080706FD0E0DFCFDFD0CFDFCFC0CFC0B0C0B0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000CFD0EFD0E0F0000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000DFEFDFD0E0D0\
0000000000000000000000000000000000000000008F90707F7F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060606F00000060607F6F60806F6F8F6F6F8F70708F
707090707090707F907F7F9F7F809F0000000000000000008F90AF8F90B0909FB0909FBF9F9FBF\
9FA0C09FA0C0A0AFC0A0AFCFA0AFCFAFB0D0AFB0D0AFBFD0B0BFDFB0BFDFB0C0DF
BFC0E0BFC0E0BFCFE0BFCFEFC0CFEFC0CFEF000000000000000000000000000000000000CFDFF0\
CFDFF0D0DFF0D0E0F0D0E0F0D0E0F0D0E0F0D0E0FFD0EFFFDFEFFFDFEFFFDFEFFF
DFEFFFDFEFFFDFF0FFDFF0FFE0F0FFE0F0FFE0F0FF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000F0F0EFEFF0EFEFEFEFEFEFE0EFEFE0EFEFE0EFEFE0EFEFE0E0EFDFE0EFDF
E0E0DFE0E0DFE0E0D0DFE0D0DFE0D0DFDFCFDFDFCF000000000000000000000000CFCFB0CFC0B0\
C0C0AFC0C0AFBFBFA0BFBFA0BFB09FB0B090AFAF90AFA08FA0A080A09F80A09F7F
9F907F9F9070908F6FD0E0DFCFDFD0CFD0CFBFCFBFB0BFAF000000808F7F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000CFD0EFD0DFF0E0EFFFE0EFFFE0EFFF\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000EFFFF0EFF0EFE0EFE0DFEFDFD0E0D0\
0000000000000000000000000000000000000000008F90707F7F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000006F6F806F6F8F6F708F707090
707090707F907F7F9F000000000000000000000000000000000000909FBF909FBF9FA0BF9FA0C0\
9FA0C0A0AFCFA0AFCFA0B0CFAFB0D0AFB0D0AFBFD0B0BFDFB0C0DFB0C0E0BFC0E0
BFCFE0BFCFEFC0CFEFC0CFEFC0D0EFC0D0EF000000000000CFDFF0CFDFF0CFDFF0CFDFF0D0DFF0\
D0E0F0D0E0F0D0E0FFD0E0FFD0E0FFD0EFFFDFEFFFDFEFFFDFEFFFDFEFFFDFF0FF
DFF0FFDFF0FFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
F0F0F0F0F0F0F0F0EFF0F0EFEFF0EFEFF0EFEFEFEFEFEFE0EFEFE0EFEFE0EFEFE0
E0EFE0E0EFDFE0E0DFE0E0DFDFE0D0DFE0D0DFDFCFDFDFCF000000000000000000CFCFBFCFCFBF\
CFCFB0C0C0B0C0C0AFC0BFAFBFBFA0B0B09FB0AF90AFAF8FAFA08FA0A080A09F80
0000009F9070908F6FD0E0DFCFDFD0C0D0CFBFC0BF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000707070000000000000000000CFD0EFD0DFF0E0EFFFE0EFFFE0EFFF\
EFF0FFEFF0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000F0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0EFE0DFEFDFD0E0D0\
D0DFCF0000000000000000000000000000000000008F90707F7F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000006F6F806F6F8F6F708F707090
707F90707F9F0000000000000000000000000000000000000000000000009FA0C09FA0C09FAFC0\
A0AFCFA0AFCFAFB0D0AFB0D0AFBFD0B0BFDFB0BFDFB0C0DFBFC0E0BFC0E0BFCFE0
BFCFEFC0CFEFC0D0EFC0D0EF000000000000000000CFDFF0CFDFF0CFDFF0CFE0F0D0E0F0D0E0FF\
D0E0FFD0E0FFD0EFFFD0EFFFDFEFFFDFEFFFDFEFFFDFEFFFDFF0FFDFF0FFDFF0FF
E0F0FFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FFE0FFFFE0FFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000EFFFFFEFFFFFEFFFFF\
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000000000\
F0FFF0F0FFF0F0F0F0F0F0F0F0F0F0F0F0EFEFF0EFEFF0EFEFF0EFEFEFEFEFEFE0
EFEFE0E0EFE0E0EFDFE0E0DFE0E0DFDFE0D0DFE0D0DFDFCFDFDFCFD0DFCFD0D0C0D0D0C0CFD0BF\
CFCFBFCFCFB0C0C0B0C0C0AFBFBFA0BFBFA0B0B09FAFAF90AFAF8FA0A08FA09F80
000000000000908F708F806FCFDFD0C0CFC0B0C0B0000000909F8F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000CFD0E0D0DFF0DFEFFFE0EFFFE0EFFF\
EFF0FFEFF0FFF0F0FF000000000000000000000000000000000000000000000000
000000000000000000F0FFF0F0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0EFE0DFEFDFD0E0D0\
D0DFCF0000000000000000000000000000000000008F9070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000006F6F806F6F8F707090707090
707F900000000000000000000000000000000000000000000000000000009FA0C0A0AFCFA0AFCF\
A0B0CFAFB0D0AFB0D0AFBFD0B0BFDFB0C0DFB0C0E0BFC0E0BFCFE0BFCFE0BFCFEF
C0D0EFC0D0EF000000000000000000000000000000000000CFE0F0D0E0F0D0E0FFD0E0FFD0E0FF\
D0EFFFD0EFFFDFEFFFDFEFFFDFEFFFDFEFFFDFF0FFDFF0FFDFF0FFE0F0FFE0F0FF
E0F0FFE0F0FFE0F0FFE0F0FFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000EFFFFFEFFFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000EFFFFF
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFF0FFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000FFFFFF\
FFFFFFF0FFFFF0FFFFF0FFF0F0F0F0F0F0F0F0F0F0F0F0F0EFF0EFEFF0EFEFF0EF
EFF0EFEFEFE0E0EFE0E0EFE0E0EFDFE0E0DFDFE0D0DFE0D0DFE0D0DFDFCF000000000000D0D0C0\
CFD0BFCFCFBFCFCFB0C0C0B0C0C0AFBFBFA0B0B09FB0B09FAFAF90AFA08FA0A080
0000000000009090708F806FCFDFCFBFCFBFAFBFAFA0AF9F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000CFD0E0D0DFEFDFEFFFE0EFFFE0EFFF\
EFF0FFEFF0FFF0F0FF000000000000000000000000000000000000000000000000
000000000000000000F0FFFFF0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0EFE0DFEFDFD0E0D0\
D0DFCF0000000000000000000000000000000000008F9070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000006F6F8F6F708F707090707F90
000000000000000000000000000000000000000000000000000000000000000000A0AFCFA0B0D0\
AFB0D0AFBFD0AFBFDFB0BFDFB0C0DFB0C0E0BFC0E0BFCFE0BFCFEFC0CFEF000000
000000000000000000000000000000000000000000000000000000D0E0FFD0E0FFD0EFFFD0EFFF\
DFEFFFDFEFFFDFEFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFE0F0FFE0F0FFE0F0FF
E0F0FFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFF000000000000\
000000000000000000E0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFF
E0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000EFFFFFEFFFFFEFFFFF\
EFFFFFEFFFFF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000FFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFF0FFFFF0FFFFF0FFFFF0FFF0F0F0F0F0F0F0F0F0F0EFF0EF
EFF0EFEFF0EFEFEFE0E0EFE0E0EFE0E0EFDFE0E0DFDFE0D0DFE0D0000000000000000000000000\
000000CFD0C0CFD0BFCFCFB0C0C0AFBFC0AFBFBFA0B0B09FB0B090AFAF90A0A08F
0000000000000000008F8F6FC0D0CF0000000000009FAF9F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000CFD0E0D0DFEFDFE0F0E0EFFFE0EFFF\
EFF0FFEFF0FFF0F0FF000000000000000000000000000000000000000000000000
000000000000000000F0FFFFF0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0F0E0DFEFDFD0E0D0\
D0DFCF0000000000000000000000000000000000008F9070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006F708F000000000000
000000000000000000000000000000000000000000000000000000000000000000A0B0D0AFB0D0\
AFBFD0B0BFDFB0BFDFB0C0DFB0C0E0BFC0E0BFCFE0BFCFEF000000000000000000
000000000000000000000000000000000000000000000000000000000000D0EFFFDFEFFFDFEFFF\
DFEFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFE0F0FFE0F0FFE0FFFFE0FFFF
E0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFF000000000000000000E0FFFFE0FFFF\
E0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0F0FFE0F0FF
DFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFEFFF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000E0F0FFE0F0FFE0FFFF\
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF000000
000000000000000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000FFFFFF\
FFFFFFFFFFFFFFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFF0F0FFF0F0F0F0
EFF0F0EFF0EFEFF0EFEFEFE0E0EFE0E0EFE0E0EFDFE0E0DFDFE0DF000000000000000000000000\
000000000000CFD0C0CFCFBFC0CFB0C0C0AFBFBFA0B0BFA0B0B09FAFAF90AFAF8F
000000000000000000908F708F806F000000A0B0A0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000D0DFEFDFE0F0E0EFFFEFEFFF\
EFF0FFEFF0FFF0F0FF000000000000000000000000000000000000000000000000
000000000000000000F0FFFFF0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0F0E0DFEFDFD0E0D0\
D0DFCF000000000000000000000000000000000000808F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006F7090000000000000
000000000000000000000000000000000000000000000000000000000000000000000000AFBFDF\
B0BFDFB0C0DFB0C0E0B0C0E0BFCFE0BFCFE0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000DFEFFFDFF0FF\
DFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFF
E0FFFFE0FFFF000000000000000000000000000000000000000000000000000000000000000000\
E0FFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFEFF0D0EFF0000000
000000000000B0AFC0BFB0CFBFBFD0C0C0DFCFCFE0D0D0E0D0D0EFDFDFF0E0E0F0EFEFFFEFEFFF\
EFEFFFEFF0FF000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000DFEFF0\
DFEFF0DFEFF0DFEFFFE0F0FFE0F0FFE0F0FFE0F0FFE0FFFFEFFFFFEFFFFFEFFFFF
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFF000000000000F0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000FFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0F0F0EFF0F0EFF0EFEFF0EFEFEFEFE0EFE0E0EFE0E0EFDF000000000000000000000000000000\
000000000000000000CFD0BFCFCFBFC0CFB0BFC0AFBFBFA0B0BF9FB0B09F000000
0000000000000000000000008F806F00000090A090000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000D0DFEFDFE0F0E0EFFFEFEFFF\
EFF0FFEFF0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000F0FFFFF0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0F0E0DFEFDFD0E0D0\
D0DFCF000000000000000000000000000000000000707F60000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
B0C0DFB0C0E0BFC0E0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
DFF0FFDFF0FFDFF0FFDFF0FFE0F0FFE0FFFFE0FFFFE0FFFFE0FFFF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000706F8070708F807F908F809F908FAF000000
000000000000B0B0CFBFBFD0C0C0DFCFCFDFCFCFE0D0D0EFDFDFEFE0E0F0E0EFFFEFEFFFEFEFFF\
EFF0FF000000000000000000000000000000000000000000000000000000000000
000000000000000000FFFFFF000000000000000000000000000000000000000000000000000000\
E0EFEFCFDFD0000000000000000000000000000000DFE0F0DFEFF0DFEFF0DFEFFF
E0EFFFE0F0FFE0F0FFE0F0FFE0FFFFEFFFFFEFFFFFEFFFFF000000000000000000EFFFFFF0FFFF\
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000FFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFF0EFF0F0EFF0F0EFF0EFEFF0EFE0EFE0000000000000000000000000000000000000\
000000000000000000000000CFD0BFC0CFB0C0C0B0BFC0AFBFBFA0B0B09F000000
0000000000000000000000008F806F9FAF9F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000D0DFEFDFE0F0E0EFFFEFF0FF\
EFF0FFEFF0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0F0E0DFEFDFD0E0D0\
D0DFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000707090000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000DFF0FFDFF0FFE0FFFF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000706F807F709080809F8F8FA09F90AF000000
000000B0B0CFBFB0CFBFBFD0C0C0DFCFCFE0D0D0E0DFDFEFDFE0F0E0E0FFEFEFFFEFEFFFEFF0FF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000FFFFFF000000000000000000000000000000000000000000000000E0F0EF\
CFDFD0CFBFBFC0B0B0C0B0B0000000000000000000000000B09FA0AF9F9FAF9F9F
AF9090A09090A08F8FA08F8FDFE0F0000000000000000000000000000000000000000000000000\
000000000000000000EFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFF0EFF0F0EFF0EFEFF0EF000000000000000000000000000000000000\
000000000000000000000000000000CFD0BFC0CFB0BFC0AFBFC0AFB0BFA0000000
0000000000000000000000008F8F6F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000009F9FAF000000000000000000000000D0DFEFDFE0F0E0EFFFEFF0FF\
EFF0FFEFF0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFFFEFE0F0E0DFEFDFD0E0D0\
D0DFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000006F7090707F90000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
BFCFE0BFCFE0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
DFF0FFDFFFFFE0FFFF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F7F70708F807F908F80A09090AF9F9FB0A0A0BF
AFAFC0B0B0CFBFBFD0C0C0DFCFCFE0D0D0E0D0D0EFDFDFF0E0E0F0E0EFFFEFEFFFEFEFFFEFF0FF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000E0F0EFCFDFD0\
DFCFCFCFBFC0CFBFBFCFBFBFC0B0B0C0B0B0000000BFAFAFB0A0AFB0A0A0B09FA0
AF9F9FAF909FAF9090A09090A08F8FA08F909F808F000000000000000000000000000000000000\
000000000000000000000000000000000000000000EFFFFFEFFFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000F0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFF0EFFFF0000000000000000000000000000000000000000000\
000000000000000000000000000000000000C0D0BFC0CFB0BFC0AF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000A09FB0000000000000000000CFD0E0D0DFEFDFE0F0E0EFFFEFF0FF\
EFF0FFEFF0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFFFEFE0F0E0DFEFDFD0E0D0\
D0DFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F6F8F6F7090707F9F7F80A0000000
000000000000000000000000000000000000000000000000000000000000000000B0C0E0B0CFE0\
BFCFE0BFCFEFBFD0EF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000DFF0FFDFF0FFDFFFFF\
DFFFFFDFFFFFDFFFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFD0F0FF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F6F807F708F80809F908FA09F90B0A0A0BFAFAFC0
000000BFBFD0C0C0D0C0C0DFCFCFE0D0D0EFDFDFEFDFE0F0E0EFFFEFEFFFEFEFFFEFF0FF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFFF000000000000000000E0D0D0\
DFCFCFDFCFCFD0C0C0D0C0C0CFBFBFC0B0BFC0B0B0000000000000BFAFAFB0A0AF
B0A0A0B09FA0AF9F9FAF909FAF909FA09090A08F90000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
EFFFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C0CFBFC0CFB0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000908FA0A0A0B0B0B0CF000000000000CFD0EFD0DFEFDFE0F0E0EFFFEFF0FF\
EFF0FFEFF0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0F0E0DFEFDFD0E0D0\
D0DFCF0000000000000000000000000000007F7F6F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F708F707F90707F9F7F80A0808FAF
000000000000000000000000000000000000000000000000000000000000B0C0DFB0CFE0BFCFE0\
BFCFEFBFD0EFC0D0EFC0DFF0C0DFF0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000DFF0FFDFF0FFDFFFFFDFFFFFDFFFFF\
DFFFFFDFF0FFDFF0FFDFF0FFDFF0FFD0F0FF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000706F807F7F908F809F9090AF9F9FB0AFA0C0000000
000000000000C0C0DFCFCFE0D0D0EFD0DFEFDFDFF0E0E0FFE0EFFFEFEFFFEFEFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFFFF0FFFF000000000000EFDFE0E0D0DF\
E0D0DFDFCFD0DFCFD0D0C0C0D0C0C0CFBFC0000000000000000000000000000000
BFA0AFB0A0AFB0A0A0B09FA0AF9FA0AF9F9F000000000000000000000000000000000000000000\
000000000000000000907F80907F80000000000000000000000000000000000000
000000000000000000EFFFFF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000F0FFFFF0FFFF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000C0CFBF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000009090A0A0A0BFB0B0CF000000000000D0D0EFDFDFEFE0EFF0E0EFFFEFF0FF\
EFF0FFF0F0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0F0E0DFEFDFD0E0D0\
CFDFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000606F8F6F7090707F9F7F809F7F8FA08090AF
8F9FB0000000000000000000000000000000000000000000AFBFDFB0C0DFB0C0E0BFCFE0BFCFEF\
BFD0EFC0D0EFC0DFF0C0DFF0CFDFF0CFE0FF000000000000000000000000000000
000000000000000000000000000000DFF0FFDFF0FFDFF0FFDFFFFFDFFFFFDFF0FFDFF0FFDFF0FF\
DFF0FFDFF0FFDFF0FF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F607F706F8F807F9F8F8FA09F90B0A0A0BF000000000000
000000000000000000CFD0E0D0D0EFDFDFF0E0E0F0E0EFFFEFEFFFEFEFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000FFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFFFF0FFFF000000000000F0E0EFEFE0E0EFDFE0\
EFDFDFE0D0DFDFD0D0DFCFD0D0CFCFD0C0CF000000000000000000000000000000
000000000000000000B0A0AFB0A0A0AF9FA0000000000000000000000000000000000000000000\
00000000000000000090808F907F8F90808F8F7F80907F8FDFE0F0000000000000
E0F0FFE0F0FFEFFFFFEFFFFFEFFFFFEFFFFF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000F0FFFF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000C0D0BFC0CFB0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000009090A0A0A0BFB0B0D0C0C0DF000000D0D0EFDFE0F0E0EFF0EFF0FFEFF0FF\
EFF0FFF0F0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0F0E0DFEFDFD0E0D0\
CFDFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000060606F6060700000006F6F8F6F7090707F9F7F80A0808FAF8F90B0
8F9FB0000000000000000000000000000000000000AFBFDFB0C0DFB0C0E0BFCFE0BFCFEFBFD0EF\
C0D0EFC0DFF0C0DFF0CFDFF0CFE0FFCFE0FFD0E0FFD0EFFFD0EFFF000000000000
000000000000DFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FF\
DFF0FF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F6F8070708F80809F908FAF9F9FB0A0A0C0000000000000
000000000000000000000000000000DFE0F0E0EFFFE0EFFFEFEFFFEFF0FF000000000000000000\
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFFFF0FFFF000000000000F0EFEFEFE0EFF0E0EFEFE0E0\
EFE0EFEFDFE0E0D0DFE0D0DFDFCFD0DFCFD0000000000000000000000000000000
000000000000000000000000000000B0A0AFB09FAF000000000000000000000000000000000000\
0000000000009F8F909F809090808F90808F90808F907F8F907F8F8F7F8F8F7080
000000DFEFFFDFEFFFE0F0FFE0F0FFEFFFFFEFFFFFEFFFFFEFFFFF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000F0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFD0C0CFD0C0C0CFBF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000009090AFAFAFBFBFBFD0C0C0DFCFCFE0D0DFEFDFE0F0E0EFFFEFF0FFEFF0FF\
EFF0FFF0F0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0F0FFF0EFFFF0EFF0F0EFF0EFE0EFE0DFE0DFD0DFD0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060700000006F6F8F707F9070809F7F8FA08090AF8F90B0
909FBF90A0C0000000000000000000000000AFBFDFB0C0DFB0C0E0BFCFE0BFCFE0BFD0EFC0D0EF\
C0DFF0C0DFF0CFDFF0CFE0F0CFE0FFD0E0FFD0EFFFD0EFFFD0EFFFD0EFFFDFEFFF
DFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFCFE0DF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F6F807F7F908F80A09090AF9F9FBF000000000000000000
000000000000000000000000000000000000E0EFFFEFEFFFEFEFFF000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFF0FFFFF0FFFFF0FFFF000000000000000000000000F0EFF0F0E0EFF0E0EF\
EFE0E0EFE0EFEFDFE0E0DFE0E0D0DF000000000000000000000000000000000000
000000000000000000000000000000BFAFB0B0A0B0B0A0AFB09FAFAF9FA0AF9FA0000000000000\
0000000000009F8F9F9F8F909F8F909F8090908090908090907F8F8F7F8F000000
000000000000000000000000DFE0FFDFEFFFE0F0FFE0F0FFEFFFFFEFFFFFEFFFFFEFFFFF000000\
000000000000000000000000000000000000000000000000000000000000000000
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000\
000000000000000000000000000000000000D0DFCFCFDFCFCFD0C0C0D0BF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000009F9FAFAFAFC0BFBFD0C0C0E0CFD0E0000000000000E0EFFFEFF0FFEFF0FF\
EFF0FFF0F0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0EFFFF0EFFFF0EFF0EFE0F0EFDFEFE0D0E0DFD0DFCF\
00000000000000000000000090A08F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060700000006F6F8F707F907F80A0808FA08090AF8F9FB0
90A0BF9FA0C09FAFCF000000000000AFBFD0AFC0DFB0C0DFB0CFE0BFCFE0BFD0EFC0D0EFC0DFF0\
C0DFF0CFDFF0CFE0F0CFE0F0D0E0FFD0EFFFD0EFFFD0EFFFD0EFFF000000000000
000000000000000000000000000000000000000000000000000000D0E0E00000000000007F7060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000060607F706F8F7F7F908F8FA09090B0A0A0BF000000000000000000
000000000000000000000000000000000000000000EFEFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFF0FFFFF0FFFF000000000000000000000000000000000000000000EFEFEF\
EFE0EFEFE0EFEFE0EFEFDFE0E0DFE0000000000000000000000000000000000000
000000000000000000000000000000BFB0BFBFAFBFBFAFB0B0A0B0B0A0AFB0A0AFAF9FAFAF9FA0\
A090A00000000000009F8F9F9F8F9F9F8F9F9F8090908090908090908090000000
0000000000000000000000000000007F6F80DFE0FFDFEFFFDFEFFFE0F0FFE0F0FFEFFFFFEFFFFF\
EFFFFF000000000000000000000000000000000000000000000000000000F0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFEFFFF0000000000000000000000000\
000000000000000000000000000000D0E0D0D0E0D0CFDFCFCFD0C0C0D0BFBFCFB0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000908FA09F9FB0AFAFC0BFBFDFC0CFE0CFD0E0000000000000000000EFF0FFEFF0FF\
EFF0FFF0F0FFF0F0FFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000F0FFF0F0FFF0EFFFF0EFF0F0EFF0EFE0F0EFDFEFE0D0E0D0CFDFCF\
000000000000000000A0AF90909F80000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000060607060607F6F6F8F707F907F80A0808FAF8F90B08F9FBF
90A0BF9FAFC09FAFCFA0B0CF000000AFBFDFB0C0DFB0CFE0BFCFE0BFD0EFC0D0EFC0DFF0C0DFF0\
CFDFF0CFE0F0CFE0F0CFE0FFD0EFFFD0EFFFD0EFFFD0EFFF000000000000000000
000000000000000000000000000000000000000000D0EFEF0000000000000000000000008F7F6F\
80706F0000007F6F60000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F607F70708F80809F8F8FAF9F9FB0000000000000000000000000
000000000000000000000000000000000000000000EFEFFFEFF0FF000000000000000000000000\
000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFF0FFFFEFF0F0000000000000000000000000000000000000000000000000\
000000F0EFF0EFE0EFEFE0EFEFE0EF000000000000000000000000000000000000
000000000000000000000000CFBFC0C0B0C0C0B0BFBFB0BFBFAFBFBFAFB0B0A0B0B0A0B0B0A0AF\
AF9FAF000000000000000000000000000000000000000000000000908090908090
8F80908F7F908F7F90807F90807F8F000000000000000000D0E0FFDFEFFFDFEFFFE0F0FFE0F0FF\
EFFFFFEFFFFFEFFFFF000000000000000000000000F0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFF0EFFFF0EFFFF0000000000000000000\
000000000000000000000000DFEFDFD0EFDFD0E0D0CFDFCFCFDFC0C0D0BFBFCFB0
000000000000000000909F80000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009090AFA0A0BFB0B0CFC0C0DFCFCFE0D0D0EF000000000000000000000000000000\
EFF0FFF0F0FFF0F0FFFFFFFF000000000000000000000000000000000000000000
000000000000000000F0FFF0F0FFF0F0FFF0EFFFF0EFF0F0EFF0EFE0F0EFDFEFDFD0E0D0CFDFCF\
000000000000AFBFA09FAF908F9F80000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060607F6F6F8F707F907F80A08090AF8F9FB0909FBF
90A0C09FAFC0A0B0CFA0B0D0000000000000B0C0DFB0CFE0BFCFE0BFD0EFC0DFF0C0DFF0CFDFF0\
CFE0F0CFE0F0CFE0F0D0E0FFD0EFFFD0EFFF000000000000000000000000000000
000000000000000000000000000000000000D0E0E0C0CFC0000000AFAF90000000000000908070\
8F7F6F807F6F0000007F6F607F6F60000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F6F807F70908080A09090AF9F9FBF000000000000000000000000
000000000000000000000000000000000000000000EFF0FFF0F0FFF0F0FFF0F0FF000000000000\
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFF0FFFF000000000000000000000000000000000000000000000000000000\
000000000000000000EFE0EF000000000000000000000000000000000000000000
000000000000000000000000CFC0CFCFBFCFC0BFC0C0B0C0C0B0C0BFB0BFBFAFBFBFAFBFB0AFB0\
B0A0B00000000000000000000000000000000000000000000000009F8FA09F8F9F
908F9F90809F90809F8F809F8F8090000000000000000000000000808090808090000000000000\
E0F0FFE0F0FFEFFFFFEFFFFF000000EFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFEFFFF0EFFFF0EFFFF0EFFFF0000000000000\
000000000000000000E0F0E0DFEFE0DFEFDFD0E0D0CFDFCFCFDFC0C0D0BFBFCFBF
B0C0AF0000000000008F907F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6F606F0000009F9FAFAFAFBFB0B0CFC0C0DFCFCFE0D0D0EF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000F0FFF0F0FFF0EFFFF0EFF0F0EFF0EFEFF0EFE0F0E0DFEFDFD0E0D0CFDFCF\
C0D0BFBFCFB0AFB0A09FA08F8F907F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000707F907F80A08090AF8F9FB090A0BF
9FAFC09FAFCFA0B0CF000000000000000000000000000000BFD0EFC0D0EFC0DFF0C0DFF0CFDFF0\
CFE0F0CFE0F0D0E0F0D0EFFFD0EFFF000000000000000000000000000000000000
000000000000000000000000D0EFEFD0EFE0000000000000000000AFAF9F0000000000009F9080\
908F7F8F80708F7F6F80706F8070607F6F607F6F606F6060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000060607F6F6F8F7F7F908F8FA09090B0A0A0BF000000000000000000000000
000000000000000000000000000000000000EFEFFFEFF0FFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFF\
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000EFEFEF000000000000000000000000000000000000
000000000000000000000000D0CFD0D0C0D0CFC0CFCFBFCFC0BFCFC0BFC0C0B0C0BFB0C0BFAFBF\
000000000000000000000000000000000000000000000000000000A090AF9F90A0
9F8FA09F8FA0908FA0908FA090809F0000000000000000000000008F809F80809F8F8FA0000000\
000000000000000000000000000000000000000000EFFFFFEFFFFFEFFFFFEFFFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFEFFFF0EFFFF0EFFFF0EFFFF0EFFFF0000000\
000000000000E0FFEFE0F0EFDFF0E0DFEFDFD0E0D0CFE0D0CFDFCFC0D0C0BFCFBF
BFC0B00000009FAF90808F70000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
70707F908FA0A09FB0AFAFC0BFBFD0C0C0DFCFCFE0000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFF000000000000000000000000000000
000000000000000000F0FFF0F0FFF0EFFFF0EFF0F0EFF0EFE0F0EFE0EFE0D0E0DF000000CFD0C0\
C0CFBFBFC0B0A0B09F90A08F809070000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007F809F8090AF8F9FB090A0BF
9FAFC09FAFCF000000000000000000000000000000000000000000C0D0EFC0DFF0CFDFF0CFE0F0\
CFE0F0D0E0F0D0EFFF000000000000000000000000000000000000000000000000
000000000000000000DFEFEF000000000000000000000000000000B0B0A0B0AF9F000000A09F8F\
9F90809F8F7F9080708F806F8F7F6F00000080706F7F6F607F6F60706060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
60606060607F70708F807F9F8F8FAF9F9FB0000000000000000000000000000000
000000000000000000000000000000EFEFFFEFF0FFF0F0FFF0F0FFF0FFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000F0EFF0F0EFF0EFEFEFEFE0EFEFE0EF000000000000000000
000000000000000000DFD0DFDFCFDFD0CFDFD0CFD0D0C0D0CFC0D0CFBFCFC0BFCFC0BFCFC0B0C0\
000000000000000000000000000000000000000000000000000000A09FB0A09FAF
A090AF9F90AF9F90AF9F90AF0000000000000000000000000000008F8FA08F8FA08F8FAF9090AF\
000000000000000000000000000000000000000000000000000000EFFFFFEFFFFF
EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFEFFFF0EFFFF0EFFFF0EFFFF0EFFFF0\
000000E0FFF0E0F0EFE0F0EFDFF0EFDFEFE0D0E0DFCFE0D0CFDFCFC0D0C0BFCFBF
BFC0B0B0C0AF9FA08F7F8F6F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
807F8F9F90AFA0A0BFB0B0CFC0BFDFCFCFE0CFD0E0000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFF000000000000F0FFF0F0FFF0000000000000000000000000000000000000000000C0D0C0\
BFCFBFB0C0AFA0B09F909F80000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007F809F8090AF8F9FB090A0BF
9FAFC0000000000000000000000000000000000000000000000000000000000000000000000000\
CFE0F0D0EFF0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000BFBFAFB0B0A0B0AF9F000000\
A09F8FA09F809F9080908F7F90807000000000000000000080706F8070607F6F60
000000000000000000000000000000000000000000000000000000000000000000000000000000\
60606F6F608070709080809F9090AF9F9FBF000000000000000000000000000000
000000000000000000000000000000EFEFFFEFF0FFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFF000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFF0F0F0E0EFEFDFE0DF000000000000000000000000000000000000000000000000000000\
000000000000000000F0F0F0F0F0F0F0EFF0EFEFF0EFEFF0EFE0EFEFE0EF000000
000000000000000000DFD0E0DFD0E0DFD0DFD0CFDFD0CFDFD0CFDFD0C0D0CFC0D0CFC0D0C0BFCF\
000000000000000000000000000000000000000000000000000000AFA0BFAFA0B0
A09FB0A09FB0A09FB0A09FB00000000000000000000000000000009090AF9090B09090B09F9FB0\
000000AFAFC0000000000000000000000000000000000000000000000000EFFFFF
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFEFFFFFEFFFFFEFFFF0EFFFF0000000\
000000E0FFF0E0F0EFE0F0EFDFF0EFDFEFE0D0EFDFCFE0D0CFDFCFC0D0C0BFCFBF
BFCFB0AFBFA0909F80000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000006F6F6F
8F809F9F9FB0AFAFC0BFB0CFC0C0DFCFCFE0D0D0E0000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFF0FFFF000000000000000000000000000000000000000000000000000000C0D0C0\
BFCFBFB0BFAF9FAF908F9F80000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000808FAF8F9FB090A0BF
9FAFC0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000D0EFFF000000000000DFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FF
DFF0FF000000000000000000000000CFDFCF000000000000C0C0B0BFC0B0BFBFAFBFB0A0B0B09F\
AFAF9FAFA090A09F8FA090809F90800000000000000000000000000000008F706F
80706F000000000000000000000000000000000000000000000000000000000000000000000000\
60606F6F6F807F7F908F8FA09090B0000000000000000000000000000000000000
000000000000000000000000000000EFF0FFF0F0FFF0F0FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000
000000EFF0EFDFEFDFD0DFD0FFFFFF000000000000000000000000000000000000000000000000\
000000000000000000FFF0FFF0F0FFF0F0F0F0EFF0F0EFF0EFEFF0EFE0F0EFE0EF
E0E0EFE0E0EF000000E0DFE0DFDFE0DFD0E0DFD0E0DFD0E0D0CFDFD0CFDFD0CFDFCFC0DFCFC0D0\
000000000000000000000000000000000000000000000000B0AFC0B0AFC0B0AFBF
AFA0BFAFA0BFAFA0BFA0A0BF0000000000000000000000000000000000009F9FB09F9FB09F9FBF\
A0A0BFA0AFC0000000000000000000000000000000000000000000000000000000
000000EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFF0FFFFEFFFFFEFFFFF000000000000000000\
000000000000E0F0EFE0F0EFDFF0EFDFEFE0D0EFDFCFE0D0CFDFCFC0D0C0BFD0BF
000000A0B09F8F907F000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000007F707F
000000A09FB0B0AFC0BFBFD0C0C0E0CFCFE0D0D0EF000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFF0FFFF000000000000000000000000000000000000000000000000000000C0D0C0\
BFCFB0AFBFA09FAF908F907F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000008F9FB090A0BF
9FAFC0000000000000000000000000000000000000000000000000000000000000000000000000\
000000D0EFFFD0EFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FF
000000000000000000000000D0E0DF000000000000000000C0CFBFC0C0B0C0C0B0BFBFAFBFB0A0\
000000B0AF9FAFA090AFA090A09F8FA09080000000000000000000000000000000
00000000000000000000000000000000000000000000000000000070605F000000000000000000\
60606F0000007F7F9F8F8FAF9F9FB0000000000000000000000000000000000000
000000000000000000000000EFEFFFEFF0FFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000
EFF0F0E0EFE0DFE0DFBFCFBFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFF0F0FFF0F0FFF0F0FFF0EFF0EFEFF0EFEFF0
EFE0F0EFE0EFE0E0EF000000000000000000DFDFE0DFD0E0DFD0E0D0D0E0D0CFDFD0CFDFD0CFDF\
000000000000000000000000000000000000000000000000BFB0CFBFB0CFB0AFC0
B0AFC0B0AFC0AFAFC0AFA0C00000000000000000000000000000000000009FA0BF9FA0BFA0A0C0\
000000AFAFC0B0B0D0000000000000D0E0F0D0E0FF000000000000000000000000
000000000000EFFFFFEFFFFFEFFFFFEFFFFFEFFFFF000000000000000000000000000000000000\
000000000000E0F0EFDFF0EFDFF0EFDFF0E0D0EFDFCFE0D0CFDFCFC0D0C0000000
0000009FAF90000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000706F70807F8F
000000AFA0BFB0B0CFC0C0DFCFCFE0D0D0E0000000000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFF0FFF0000000000000000000000000000000000000000000000000000000C0CFBF\
BFC0B0AFB0A09FA08F808F70000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000000090A0BF
9FAFC0000000000000000000000000000000000000000000000000000000000000000000000000\
D0EFFFD0EFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFE0F0FFE0F0FF000000
000000000000000000DFEFE0D0E0DFD0DFD0000000000000CFD0C0CFCFBFC0C0BFC0C0B0000000\
000000000000BFB0A0B0AF9FAFA090AFA090000000000000000000000000000000
0000000000000000008F7F708F7F700000000000000000000000000000000000007F6F607F6060\
6060700000000000008F90AF9F9FBF000000000000000000000000000000000000
000000000000000000000000EFF0FFF0F0FFF0F0FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000
EFF0EFE0EFE0D0E0D0000000F0F0FFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000FFFFFFFFFFFFF0FFFFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFEFEFF0
EFEFF0EFEFF0EFE0F0000000000000000000000000000000000000000000DFD0E0D0D0E0D0CFE0\
000000000000000000000000000000000000000000000000BFBFD0BFBFCFBFB0CF
BFB0CFB0B0CFB0AFCFB0AFCFAFAFC0000000000000000000000000000000A0A0C0000000000000\
000000AFB0CFB0B0D0BFBFDFC0C0DF000000D0E0FFD0E0FFDFEFFFDFEFFFE0F0FF
E0FFFFE0FFFFE0FFFF000000000000000000000000000000000000000000000000000000000000\
000000000000000000DFF0EFDFF0EFDFEFE0D0EFE0D0E0DFCFDFD0000000000000
00000090A08F000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006060607F7080908F9F
000000000000BFBFD0C0C0E0CFCFE0D0D0EF000000000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000BFCFBF\
B0C0B0A0B09F8F9F80000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000D0EFFF\
D0EFFFDFEFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFE0F0FFE0F0FFE0F0FF000000
000000DFF0EFDFEFEFDFEFE0DFE0DFD0E0D0000000D0DFCFD0D0CFCFD0C0CFCFBFC0C0BF000000\
000000000000000000000000B0B0A0B0AF9F000000000000000000000000000000
0000000000000000009F8F7F90807F90807F9080708F7F700000000000000000000000008F706F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000EFEFFFEFF0FFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFF000000000000000000000000000000000000000000000000000000F0F0F0
E0EFE0DFE0DFC0D0C0EFF0FFF0FFFFF0F0FFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000FFFFFFF0FFFFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FF
F0F0FFEFEFFF000000000000000000000000000000000000000000000000000000000000000000\
D0D0E0000000000000000000000000000000000000000000C0C0D0C0BFD0BFBFD0
BFBFD0BFBFD0BFB0CFB0B0CFB0B0CF000000000000000000000000AFAFCFAFAFCF000000000000\
000000000000B0BFD0BFBFDFBFC0E0C0CFE0000000D0E0FFDFEFFFDFEFFFDFF0FF
E0F0FFE0FFFFE0FFFFE0FFFF000000000000000000000000000000000000000000000000000000\
000000000000000000DFF0EFDFF0E0DFEFE0D0EFE0D0E0DF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007070708F8090000000
000000000000BFBFD0CFCFE0D0D0E0D0DFEF000000000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000C0D0C0BFCFBF\
B0C0AF9FAF9080907F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009FB0CFA0BFCF000000000000000000000000000000000000000000000000000000D0EFFF\
D0EFFFDFEFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFE0F0FFE0F0FF000000E0F0F0
E0F0F0E0F0EFDFEFEFDFEFE0DFE0DFDFE0DFD0E0D0D0DFD0D0DFCFD0DFCFD0D0C0CFCFC0000000\
000000000000000000000000000000BFB0AF000000000000000000000000000000
000000000000000000A0908FA090809F8F809F8F809F8F7F90807F90807F907F709080708F7F70\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000EFEFFFF0F0FFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFF000000000000000000000000000000000000000000000000000000EFF0EF
E0EFE0DFE0DF000000EFF0FFE0F0FFEFF0FFEFF0FFF0FFFFF0F0FFFFFFFFF0FFFFFFFFFFFFFFFF\
000000000000F0FFFFF0FFFFF0FFFFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FFF0F0FF
F0F0FFF0F0FF000000000000000000000000000000000000000000000000000000000000000000\
DFD0EFD0D0EFD0D0E0D0D0E0D0CFE0CFCFE0CFCFE0CFCFDFCFC0DF000000000000
000000000000000000000000000000BFBFD0B0B0D0B0B0D0B0B0D0B0B0D0B0B0D0000000000000\
000000000000000000BFC0DFBFC0E0C0CFE0CFD0EFCFDFF0D0E0FFDFEFFFDFEFFF
DFF0FFE0F0FFE0FFFFE0FFFFE0FFFF000000000000000000000000000000000000000000000000\
000000000000000000000000DFF0EFDFEFE0D0EFE0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000006F6F6F807F8F908FA0000000
000000000000000000CFCFE0D0D0E0000000000000000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFF0FFFF000000000000000000000000000000000000000000000000000000C0D0C0BFCFB0\
AFB0A0909F80000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009FAFC0A0B0CFAFBFD0B0C0DF000000000000000000000000000000000000000000D0EFFF\
D0EFFFDFEFFFDFF0FFDFF0FFDFF0FFDFF0FFE0F0FFE0F0FFE0F0FFE0F0F0E0F0F0
E0F0EFE0F0EFE0F0EFE0EFE0DFEFE0000000DFE0DFDFE0D0DFDFD0D0DFD0D0DFCFD0D0C0000000\
000000000000000000000000000000000000C0BFAF000000000000000000000000
000000000000000000000000AF9F90A09F8FA0908FA0908FA090809F90809F8F809F8F809F8F80\
000000000000000000000000A0AFC0AFB0CF000000000000000000000000000000
000000000000EFEFFFEFF0FFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
000000000000000000000000000000000000000000000000000000F0F0F0EFF0EF
E0EFDFCFDFCFE0EFFFE0EFFFE0F0FFE0F0FFEFF0FFEFF0FFEFF0FFF0F0FFF0FFFFF0FFFFFFFFFF\
FFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0F0FFF0F0FFF0F0FF
F0F0FFF0F0FF000000000000000000000000000000000000000000000000000000000000000000\
DFDFF0DFDFEFDFDFEFD0D0EFD0D0EFD0D0EFD0D0E0CFCFE0CFCFE0000000000000
000000000000000000000000000000BFBFDFBFBFDFBFBFDFBFBFDFBFBFDFB0BFDF000000000000\
000000000000000000BFC0E0BFCFE0C0CFEFCFD0EF000000D0E0FFD0EFFFDFEFFF
DFF0FFDFF0FFE0FFFFE0FFFFE0FFFF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007F707F8F809F000000000000
000000000000000000000000D0D0EF000000000000000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFF0FFFF000000000000000000000000000000000000000000000000000000BFCFBFB0C0B0\
9FAF9080907F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009FAFC0A0B0CFAFBFD0B0C0DFB0CFE0BFD0E0C0D0EFC0DFEF000000000000D0E0F0D0EFFF\
D0EFFFDFEFFFDFF0FFDFF0FFDFF0FF000000000000000000000000E0F0F0E0F0F0
E0F0F0E0F0EFE0F0EFE0F0EFE0EFE0000000000000DFE0DFDFE0D0DFDFD0DFDFD0D0DFCF000000\
000000000000000000000000000000000000C0C0B0C0C0B0C0BFB0000000000000
000000000000000000000000B0AF9FB0A09FAFA090AFA090A09F8FAF9F90A0908FA09F8FA09080\
000000000000000000000000A0AFC0B0B0D0BFC0DF000000000000000000000000
000000000000EFEFFFEFF0FFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
000000000000000000000000000000000000000000000000000000F0F0EFEFEFE0
DFE0DFBFC0B0000000D0E0F0DFE0F0E0EFFFE0F0FFE0F0FFEFF0FFEFF0FFEFF0FFF0F0FFF0FFFF\
F0FFFFF0FFFF000000000000000000F0FFFFF0FFFFF0FFFFF0FFFFF0F0FFF0F0FF
F0F0FFF0F0FF000000000000000000000000000000000000000000000000000000000000000000\
E0E0F0DFDFF0DFDFF0DFDFF0DFDFEFD0DFEFD0D0EFD0D0EFD0D0EF000000000000
000000000000000000000000000000C0C0E0C0C0E0C0C0E0BFC0E0BFC0E0BFC0E0BFC0E0000000\
000000000000000000000000C0CFEFC0D0EFCFD0EF000000000000000000000000
DFEFFFDFF0FFE0F0FFE0FFFFE0FFFFE0FFFF000000000000000000000000000000000000000000\
000000000000DFF0EFDFF0EF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F6F70807F8F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFF0FFF0000000000000000000000000000000000000000000000000C0D0C0BFCFBFAFBFA0\
909F8F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009FAFC0A0B0CFAFBFD0B0C0DFB0CFDFBFD0E0BFD0E0C0DFEFCFDFEFCFE0F0000000000000\
000000000000000000000000000000000000000000000000E0F0FFE0F0F0E0F0F0
E0F0F0E0F0F0E0F0EFE0F0EF000000000000000000000000DFE0DFDFE0DFDFE0D0000000000000\
000000000000000000000000000000000000CFCFBFCFC0BFCFC0BFCFC0B0C0C0B0
C0BFB0000000000000000000BFB0A0BFB0A0BFAFA0B0AF9FBFAFA0B0A09FB0A09FAFA090B0A09F\
000000000000000000000000A0AFC0B0B0D0BFC0DFC0CFE0000000000000000000
000000000000EFEFFFEFF0FFF0F0FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
000000000000000000000000000000000000000000000000F0F0F0EFF0EFE0EFE0
D0E0D0BFC0AF000000000000000000000000DFEFFFE0EFFFE0EFFFE0F0FFE0F0FFEFF0FFEFF0FF\
EFF0FF000000000000000000000000000000000000000000F0FFFFF0FFFFF0F0FF
F0F0FFF0F0FF000000000000000000000000000000000000000000000000000000000000000000\
E0E0FFE0E0FFE0E0FFE0E0F0DFE0F0DFDFF0DFDFF0DFDFF0D0DFF0000000000000
000000000000000000000000000000000000C0CFE0C0CFE0C0CFE0C0CFE0C0CFE0C0CFE0000000\
000000000000000000000000000000000000CFDFF0000000000000000000000000
000000000000000000E0FFFFE0FFFFE0FFFFE0FFFF000000000000000000000000000000E0FFFF\
E0F0F0DFF0F0DFF0EFD0EFE0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006060607F7080908F9F000000000000000000
000000000000000000000000DFDFEFE0E0F0000000000000000000000000000000000000000000\
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
F0FFFF000000000000000000000000000000000000000000000000000000BFCFBFB0C0B09FAF9F\
80907F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009FAFC0A0BFCFB0C0DFB0CFDFBFCFE0BFD0E0C0DFEFC0DFEFCFE0F0000000000000\
000000000000000000000000000000000000000000000000000000E0F0FFE0F0F0
E0F0F0E0F0F0E0F0F0000000000000000000000000000000000000E0E0DFDFE0DF000000000000\
000000000000000000000000000000000000D0D0C0D0CFC0D0CFBFCFCFBFCFC0BF
CFC0BFCFC0BFCFC0B0CFC0B0C0BFB0C0BFB0C0BFAFC0B0AFBFB0AFBFB0A0BFB0A0BFAFA0B0AFA0\
0000000000000000009F9FBFA0AFC0B0B0D0BFC0DFC0CFE0D0D0F0DFE0FF000000
000000E0EFFFEFEFFFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000\
000000000000000000000000000000000000000000000000F0F0EFEFEFE0E0EFDF
000000B0BFAF000000000000000000000000000000000000DFEFF0DFEFFFE0EFFFE0F0FFE0F0FF\
EFF0FF000000000000000000000000000000000000000000000000000000000000
F0FFFFF0F0FF000000000000000000000000000000000000000000000000000000000000000000\
E0EFFFE0EFFFE0E0FFE0E0FFE0E0F0DFE0F0DFE0F0DFDFF0DFDFF0000000000000
000000000000000000000000000000000000CFD0EFCFD0EFCFD0EFCFD0EFC0D0EFC0D0EFC0D0EF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000E0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFF\
DFF0F0DFF0EFD0EFEFD0EFE0000000000000000000AFC0B0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000706F708F8090000000000000000000000000
000000000000000000000000DFDFEFE0E0F0EFEFF0000000000000000000000000000000000000\
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
F0FFFF000000000000000000000000000000000000000000000000000000BFCFBFAFBFAF90A08F\
7F8F70000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000007F7F90
0000000000009FAFBFA0B0CFAFBFD0B0CFDFBFCFE0BFD0E0C0D0EFC0DFEFCFE0F0000000000000\
000000000000000000000000000000000000000000000000000000E0FFFFE0F0F0
E0F0F0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000DFD0CFD0D0CFD0D0C0D0D0C0D0CFC0
D0CFC0D0CFBFD0CFBFCFCFBF000000000000000000CFC0B0CFBFB0C0BFB0CFBFB0C0BFAFC0BFB0\
0000000000000000009FA0BFA0AFCFB0BFD0BFC0DFC0CFE0D0DFF0DFE0FFDFE0FF
E0EFFF000000EFF0FFF0F0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000\
000000000000000000000000000000000000000000F0F0F0EFF0EF000000000000
000000B0BFA0000000000000C0B0B0BFB0B0000000000000000000000000D0E0F0DFEFFFE0EFFF\
E0F0FF000000000000000000000000000000000000000000000000000000000000
000000000000F0FFFF000000000000000000000000000000000000000000000000000000000000\
E0EFFFE0EFFFE0EFFFE0EFFFE0E0FFE0E0F0DFE0F0DFE0F0DFE0F0000000000000
000000000000000000000000000000000000000000CFD0EFCFD0EFCFD0EFCFD0EFCFD0EFCFD0EF\
CFD0EF000000000000000000000000000000D0E0FFD0E0FF000000000000000000
000000000000000000000000000000000000E0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFFE0FFFF\
DFF0F0DFF0EFD0EFEFD0EFE0000000000000AFC0B0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000007F7F80908FA0000000000000000000000000
000000000000000000000000E0E0F0E0EFF0EFF0FFF0F0FF000000000000000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000000000000000000000000000000000000000000000BFCFC0B0C0B0A0B09F80907F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
80809F00000090A0BF9FAFC0A0BFCFAFC0D0B0CFDFBFCFE0C0D0E0C0DFEFCFDFEF000000000000\
000000000000000000000000000000000000000000000000000000E0FFFFE0FFF0
000000000000000000000000000000000000000000000000000000E0EFE0E0EFE0000000000000\
000000000000000000000000000000000000DFDFD0DFDFD0DFDFCFDFD0CFDFD0CF
D0D0CFD0D0C0D0D0C0D0CFC0000000000000000000000000000000000000000000000000CFBFB0\
0000000000000000009FA0BFAFAFCFB0BFD0BFC0DFC0CFEFD0DFF0DFE0FFE0E0FF
000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000B0B0A0000000000000CFBFBFCFBFBFCFBFBFCFC0C0000000000000000000000000000000\
DFEFF0000000000000000000000000000000000000000000000000000000000000
000000000000F0FFFFF0FFFFF0FFFFEFFFFFEFF0FF000000000000000000000000000000000000\
EFEFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0E0F0DFE0F0DFE0F0DFE0F0000000
000000000000000000000000000000000000000000D0DFF0D0DFF0D0DFF0D0DFF0D0DFF0CFDFF0\
CFDFF0CFDFF0000000000000D0E0F0D0E0FFD0E0FFD0EFFFD0EFFFD0EFFF000000
000000000000000000000000000000000000E0FFFFE0FFFFE0FFFFDFFFFFE0FFFFE0FFFFDFFFFF\
DFF0F0D0F0EFD0EFEF000000000000AFC0BF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000008F8090000000000000000000000000000000
000000000000000000DFDFEFE0E0F0EFEFF0F0F0FFF0F0FFF0FFFFF0FFFF000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFF
000000000000000000000000000000000000000000000000000000BFCFBFB0C0AF90A08F7F8F70\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000008090A08F9FB0000000A0B0C0AFBFD0B0C0DFB0CFDFBFD0E0C0D0E0C0DFEFCFE0EF000000\
000000000000000000000000000000000000000000000000000000E0FFFF000000
000000000000000000000000000000000000000000000000000000EFF0EFEFF0EFEFEFE0E0EFE0\
000000000000000000000000000000E0E0DFE0E0DFE0E0DFE0E0D0DFDFD0DFDFD0
DFDFD0DFDFCFDFD0CFDFD0CF000000000000000000000000000000000000000000000000000000\
0000000000000000009FA0BFAFAFCFB0BFD0BFC0DFCFD0EFD0DFF0DFE0FFE0E0FF
000000000000000000000000000000000000000000000000000000000000FFFFFF000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFF0FFF0000000000000000000000000
000000B0B0A0000000000000000000D0CFCFDFCFCF000000000000000000000000000000000000\
000000CFDFEF000000000000000000000000000000000000000000000000000000
000000000000EFF0FFEFF0FFEFF0FFEFFFFFEFFFFFEFFFFFEFF0FFEFF0FFEFF0FF000000000000\
EFF0FFEFF0FFEFF0FFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFF000000
000000000000000000000000000000000000000000000000D0E0F0D0E0F0D0E0F0D0E0F0D0E0F0\
D0E0F0D0E0F0000000D0E0FFD0E0FFD0EFFFD0EFFFD0EFFFD0EFFFD0EFFFDFEFFF
DFF0FF000000000000000000000000000000E0FFFFE0FFFFDFFFFFDFFFFFDFFFFFDFFFFFDFF0FF\
D0F0EFCFEFE0CFE0DFC0DFD0AFC0BF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000009F8FA0000000000000000000000000000000
000000000000000000E0E0F0EFEFF0EFF0F0F0F0FFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0
000000000000000000000000000000000000000000000000BFD0C0B0C0B0A0B0A08F9080000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000808FA0000000000000000000AFC0D0B0CFDFBFCFDFBFD0E0C0DFEFCFDFEF000000\
000000000000000000000000000000000000E0F0FFE0FFFFE0FFFFEFFFFF000000
000000000000000000000000000000000000000000000000EFF0F0EFF0EFEFF0EFEFF0EFEFF0EF\
EFEFE0EFEFE0000000000000000000E0EFE0E0EFE0E0EFDFE0E0DFE0E0DFE0E0DF
E0E0D0E0E0D0E0DFD0E0DFD0000000000000000000000000000000000000000000000000000000\
0000000000000000009FA0BFAFAFCFB0BFD0BFC0E0CFD0EFD0DFF0DFE0FFE0E0FF
000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000
C0CFBFA0A090DFCFCFDFD0D0DFCFCF000000000000000000000000000000000000000000000000\
000000BFCFDFC0D0E0C0DFEFCFDFEF000000000000000000000000000000000000
000000000000E0F0FFE0F0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FFEFF0FF\
EFF0FF000000000000000000E0F0FFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFFE0EFFF
000000000000000000000000000000000000000000000000DFE0F0DFE0F0DFE0FF000000000000\
000000000000000000000000D0EFFFD0EFFFD0EFFFDFEFFFDFEFFFDFF0FFDFF0FF
DFF0FFDFF0FFDFFFFFE0FFFF000000DFFFFFDFFFFFDFFFFFDFFFFFDFFFFFDFF0FFDFF0F0000000\
000000C0E0DFC0DFD0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000A090AF000000000000000000000000000000
000000000000E0E0EFE0E0F0EFEFF0F0F0FFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFF000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000000000000000000000000000000000BFCFBFB0C0B090A090000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000808FA0909FB0000000000000000000000000000000000000C0DFEFCFE0EF\
000000000000000000DFEFF0DFF0F0DFF0FFE0F0FFE0F0FFE0FFFFEFFFFFEFFFFF
000000000000000000000000000000000000000000EFFFF0EFFFF0EFFFF0EFF0F0EFF0EFEFF0EF\
EFF0EFEFF0EFEFF0EFEFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0
EFEFDFE0E0DFE0E0DFE0E0DF000000000000000000000000000000000000000000000000000000\
0000000000000000009FA0BFAFAFCFB0BFD0BFC0E0CFD0EFD0DFF0DFE0FFE0EFFF
000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000
C0C0B0AFAF9FE0DFDFE0DFDFE0DFDF000000000000000000000000000000000000000000000000\
000000000000000000B0C0D0BFCFDFC0D0E0CFDFEFCFE0EF000000000000000000
000000000000000000E0F0FFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FFEFF0FFEFF0FF\
EFF0FF000000000000000000000000000000000000000000000000000000000000
E0EFFFE0EFFFE0EFFFE0EFFFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFF000000000000000000000000\
000000000000000000000000000000DFEFFFDFEFFFDFEFFFDFF0FFDFF0FFDFF0FF
DFF0FFDFF0FFE0FFFFDFFFFF000000000000000000000000000000000000000000000000000000\
C0DFDFBFD0D0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
606060000000000000000000A090AFAF9FB0000000000000000000000000000000
000000DFDFEFE0E0F0EFEFF0EFF0FFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFF000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFF000000
000000000000000000000000000000000000000000BFD0C0B0C0B0A0B0A0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000808FA0909FB0000000000000000000000000000000000000000000\
CFE0EFD0EFF0D0EFF0DFEFF0DFEFF0DFF0F0DFF0FFE0F0FFE0F0FFE0FFFFEFFFFF
EFFFFF000000000000000000000000000000EFFFF0EFFFF0EFFFF0F0FFF0F0FFF0F0FFF0EFF0F0\
EFF0F0EFF0EFEFF0EFEFF0EF000000000000EFF0EFEFEFEFEFEFE0EFEFE0EFEFE0
EFEFE0EFEFE0EFEFE0EFEFE0000000000000000000000000000000000000000000000000000000\
0000000000000000009FA0BFAFB0CFB0BFD0C0CFE0CFD0EFD0DFF0DFE0F0E0EFFF
000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000
C0C0B0000000000000EFE0E0E0DFDF000000000000000000000000000000000000000000000000\
000000000000000000000000000000FFF0F0B0C0D0BFCFDFC0D0E0CFDFEFCFDFEF
D0E0F0000000000000DFEFFFDFEFFFE0EFFFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FFE0F0FF\
E0F0FF000000000000000000000000000000000000000000000000000000000000
000000E0F0FFE0F0FFE0F0FFE0EFFFE0EFFFDFEFFFDFEFFFDFEFFFDFEFFF000000000000000000\
000000000000000000000000000000000000000000DFF0FFDFF0FFDFF0FFDFF0FF
DFF0FFDFFFFFDFF0FF000000000000000000000000000000000000000000000000000000C0DFD0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000606060\
000000000000000000A090AFAF9FB0B0A0BF000000000000000000000000000000
000000E0E0EFEFEFF0EFEFF0F0F0FFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000\
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFF0FFF0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000808F9F8F9FAF000000000000000000000000000000000000\
000000CFE0EFD0EFF0D0EFF0DFEFF0DFEFF0DFF0F0E0F0FFE0F0FFE0F0FFE0FFFF
EFFFFFEFFFFF000000000000000000EFFFFFEFFFFFEFFFFFEFFFFFF0FFF0F0FFF0F0FFF0F0FFF0\
F0FFF0F0FFF0F0F0F0000000000000000000000000000000EFF0EFEFF0EFEFEFEF
EFEFEFEFEFEFEFEFEF000000000000000000000000000000000000000000000000000000000000\
0000000000000000009FA0C0AFB0CFB0BFD0C0CFE0CFD0EFD0DFF0DFE0F0E0EFFF
000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000
C0C0B0000000000000EFE0E0F0EFEF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFB0CFD0BFCFDF
BFD0E0C0D0E0CFDFEFCFE0EFD0E0F0DFEFF0DFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFFE0F0FF\
E0F0FFE0F0FF000000000000000000000000000000000000000000000000000000
000000000000E0F0FFE0F0FFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFF000000000000\
000000000000000000000000000000000000000000000000000000DFF0FFDFF0FF
DFF0FF000000000000000000000000000000000000000000000000000000BFD0D0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000606060000000\
000000000000A090AFAF9FB0B0A0BFBFAFC0C0BFD0000000000000000000000000
E0E0EFE0E0F0EFEFF0F0F0FFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000F0FFF0EFF0F0
EFF0F0E0F0EFE0F0EFDFF0EFD0E0DFCFDFD0C0D0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000080809F8F9FAF000000000000000000000000000000\
000000000000CFE0EFD0EFF0D0EFF0DFEFF0DFEFF0DFF0F0E0F0FFE0F0FFE0F0FF
EFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFF0F0FFF0000000000000000000000000000000000000000000000000
000000F0F0EFF0F0EF000000000000000000000000000000000000000000000000000000000000\
0000000000000000009FA0C0AFB0CFB0BFDFC0CFE0CFD0EFD0DFF0DFE0F0E0EFFF
000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0000000000000000000000000000000
000000000000000000F0F0F0F0F0F0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000B0CFD0BFCFDF000000000000000000000000D0E0F0D0EFF0DFEFFFDFEFFFDFEFFF\
DFEFFFDFEFFFDFEFFF000000000000000000000000000000000000000000000000
000000000000E0F0FFDFF0FFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFF0FFDFF0FFDFF0FF000000\
000000000000000000000000000000000000000000000000000000000000DFF0F0
D0F0F0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F6060000000000000\
000000000000AF9FB0AFA0BFB0AFC0BFB0CFC0C0DF000000000000000000000000
E0E0F0EFEFF0EFF0F0F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000F0FFF0EFF0F0
EFF0EFE0F0EFE0F0EFDFEFE0D0E0D0C0DFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008F90AF000000000000000000000000\
000000000000000000CFE0EFD0E0F0D0EFF0DFEFF0DFEFF0DFF0F0E0F0F0E0F0FF
E0F0FFEFFFFF000000000000000000EFFFFFEFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFF000000000000000000000000000000000000000000000000000000
000000000000F0F0F0F0F0F0000000000000000000000000000000000000000000000000000000\
000000707F8F0000009FA0C0AFB0CFB0BFDFC0CFE0CFD0EFD0DFF0DFE0F0000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000D0DFCF
000000000000000000FFFFFFFFFFFF000000000000000000000000000000000000000000000000\
000000000000000000000000000000FFFFFF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000D0E0F0D0EFF0000000000000000000000000000000000000000000000000
000000000000000000DFF0FFDFEFFFDFEFFFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FFDFF0FF\
DFF0FF000000000000000000000000000000000000DFF0F0D0EFF0D0EFF0D0EFF0
D0EFF0D0EFF0D0EFEFCFEFEFC0E0E0BFD0D0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000A09FB0AFA0BFB0AFC0BFB0CFC0C0D0CFCFE0D0D0E0000000000000E0E0EF
EFEFF0EFEFF0F0F0FFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000000000000000000000000000F0FFF0F0F0F0EFF0F0
E0F0EFE0F0EFE0F0EFD0EFDFCFDFD0C0D0C0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000CFE0EFD0E0F0D0EFF0DFEFF0DFEFF0DFF0F0000000
000000000000000000000000000000000000000000EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF\
F0FFFF000000000000000000000000000000000000000000000000000000000000
000000000000F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0000000000000000000000000\
00000070708F0000009FA0C0AFB0CFB0BFDFC0CFE0CFD0EFD0DFF0DFE0F0000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000DFDFCF
000000000000000000000000FFF0F0000000000000000000000000000000000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFF000000000000\
000000000000000000000000CFDFEFCFE0EFD0E0F0D0E0F0D0E0F0000000000000
000000000000000000000000DFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFFDFEFFF\
DFEFF0D0EFF0000000D0EFF0D0EFF0D0EFF0D0EFF0CFEFEFCFEFEFCFE0EFCFE0EF
CFE0EFC0DFDF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006F606F000000000000000000000000\
A09FB0AFA0BFB0AFC0BFB0CFC0BFD0CFCFDFD0D0E0DFD0EFDFDFEF000000EFE0F0
EFEFF0F0F0F0F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000000000000000000000000000F0FFF0EFF0F0EFF0F0
E0F0EFE0F0EFDFEFE0D0E0DFC0DFCF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000009FAFBF\
000000000000000000000000000000CFDFEFD0E0F0000000000000000000000000
000000000000000000000000000000000000000000000000EFFFFFEFFFFFF0FFFFF0FFFFF0FFFF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000FFFFF0FFFFF0F0FFF0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0FFF0F0F0F0F0FFFFF0\
F0F0F06F6F7F808FA09FA0BFAFB0CFB0BFDFC0CFE0CFDFF0D0DFF0DFE0F0000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000EFF0EFEFEFE0DFE0D0
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFF000000000000000000000000BFD0DFC0D0E0C0DFE0C0DFE0C0DFE0
CFDFEFCFDFEFCFDFEFCFE0EFCFE0EFCFE0EF000000000000000000000000000000000000000000\
000000000000000000000000000000000000CFE0EFC0DFE0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F606F706F70000000000000000000A09FAF\
AFA0BFB0AFC0BFB0CFC0BFD0CFC0DFD0D0E0DFD0EFDFDFEFE0DFEF000000EFEFF0
F0F0F0F0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000000000000000F0FFF0F0F0F0EFF0F0E0F0EF
E0F0EFDFF0EFD0EFDFCFDFD0BFD0C0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000A0B0C0AFC0D0B0C0D0BFCFDFC0D0E0C0DFE0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000F0FFFFF0FFFF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFFFFFF0F0F0FFFFF0F0F0F0\
FFFFF06060707F809F000000A0AFCFB0BFD0C0CFE0CFDFF0D0DFF0DFE0F0000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000F0F0EFEFF0EFEFF0E0DFE0D0
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000D0E0E0000000000000000000000000C0DFE0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000070606F7F6F7F807F80000000000000A09FAFAFA0BF\
B0AFC0BFB0CFC0BFD0CFC0DFD0D0E0DFD0EFDFDFEFE0DFEF000000000000000000
F0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000000000000000F0F0F0EFF0F0EFF0EFE0F0EF
E0F0EFDFEFE0CFE0D0C0D0CF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000909FAFA0BFCFAFC0D0B0CFD0BFCFDFC0D0E0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000EFFFFFF0FFFF\
000000000000000000000000000000000000000000000000000000000000000000
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
0000000000007F7F90000000000000000000BFCFE0CFD0EFD0DFF0DFE0F0000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0FFF0F0F0EFEFF0EFEFF0E0DFE0D0
D0D0C0000000000000A0A08F908F70000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000FFFFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
CFDFDFC0D0D0B0CFCF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007060707F6F7F807F8F908090000000A09FAFAFA0B0B0AFC0\
BFB0CFC0BFD0CFC0DFD0CFE0DFD0EFDFDFEFE0DFEF000000000000000000000000
F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000000000F0FFF0EFF0F0EFF0F0E0F0EFE0F0EF
DFEFE0D0E0DFC0DFCF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000009FA0B0AFBFCFAFC0D0B0CFD0BFD0DFC0D0E0000000
000000000000000000000000000000000000000000000000000000E0F0FFEFFFFFEFFFFFEFFFFF\
F0FFFFF0FFFF000000000000000000000000000000000000000000000000000000
F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0\
000000000000707F8F000000000000000000000000000000D0DFF0DFE0F0000000
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000FFFFF0F0FFF0F0F0F0F0F0EFEFF0E0E0E0DF
D0D0C0000000000000B0AF9F9F9080000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFF0FFFF000000000000000000\
000000000000000000000000000000000000000000000000CFDFDFC0D0D0BFCFCF
AFC0C0A0BFB0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007060707F707F8F7F8F908F9FA090A0AF9FB0AFA0B0B0AFBFBFB0CF\
C0BFD0CFC0DFD0CFE0DFD0EFDFDFEFE0DFEF000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000000000000000000000EFF0F0EFF0F0E0F0EFE0F0EFDFF0EF
D0E0DFCFDFD0BFD0C0000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000009FAFBFAFBFCFAFC0D0B0CFD0BFD0DF
C0D0E0000000000000000000000000000000000000DFEFF0E0EFF0E0F0F0E0F0F0E0F0F0EFF0FF\
EFFFFFEFFFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000F0FFFF
F0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
0000000000007070808F90AF000000000000000000000000000000000000E0EFF0
000000000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF000000\
000000000000000000000000000000FFFFF0F0FFF0F0FFF0F0F0EFEFF0EFE0E0DF
DFDFCF000000000000BFB0A0A09F8F908F70000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFFFF0F0F0F0F0F0EFF0F0EFF0F0000000000000\
000000000000000000000000000000CFDFDFBFCFD0B0C0CFAFBFBF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006F606F7F6F708070808F7F8F908F9FA090A0000000000000B0AFC0BFB0C0C0BFD0\
CFC0DFD0CFE0DFD0EFDFDFEFE0DFEFE0E0EF000000000000000000000000000000
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000000000000000F0F0F0EFF0F0EFF0EFE0F0EFDFF0EFDFEFE0
CFDFD0C0D0C0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000A0BFCFAFC0CF
B0C0D0BFCFDFC0D0DF000000000000D0E0EFD0E0EFD0EFF0DFEFF0DFEFF0DFEFF0E0F0F0E0F0F0\
E0F0F0EFF0F0EFF0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF
F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000\
0000000000006F6F7F808FA0000000000000000000000000000000000000DFE0F0
E0EFFFEFF0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000FFFFFFF0FFF0F0FFF0F0F0EFF0F0EFE0EFDF
DFE0D0000000CFCFBFC0BFAFAFAF909F9080908F70807F6FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFF0FFFFF0FFFFF0F0F0F0F0F0EFF0F0EFF0F0EFF0F0E0EFF0E0EFF0E0EFEF000000000000\
000000000000C0D0DFBFCFCFB0C0C0AFBFBF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000606060
00000070606F7F6F7F8070809080909F8F9FA090AF000000000000000000BFB0CFC0BFD0CFC0DF\
D0CFE0DFD0EFDFDFEFE0DFEFE0E0EF000000000000000000000000000000000000
000000000000000000000000FFFFFFFFFFFF000000000000000000000000000000000000000000\
000000000000000000000000F0FFF0EFF0F0EFF0F0E0F0EFE0F0EFDFEFE0CFE0DF
C0D0CF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000CFDFE0D0E0EFD0E0EFD0E0EFDFEFF0DFEFF0DFEFF0\
DFEFF0E0F0F0E0F0F0E0F0F0E0F0F0EFF0FFEFF0FFEFFFFF000000000000000000
000000000000F0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000\
0000000000000000007F809F000000000000000000000000000000000000DFE0F0
E0EFF0EFF0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000FFFFFFF0FFF0F0FFF0F0F0EFF0F0EFEFF0E0
E0E0DFDFDFCFD0D0C0CFC0B0B0B09FAFA08F9F908000000080706FFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFF000000000000000000000000000000000000FFFFFFF0FFFFF0F0FFF0F0F0
F0F0F0EFF0F0EFF0F0EFF0F0E0EFF0E0EFEFE0EFEFDFEFEFDFEFEFDFEFEFD0E0E0CFDFDFC0D0D0\
BFCFCFB0C0C0A0BFBF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000006F606F70606F
7F6070806F7F8F70809080909F8F9FAF9FAF000000000000000000000000000000CFC0DFD0CFE0\
DFD0EFDFDFEFE0DFEFE0E0EF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000EFF0F0EFF0F0E0F0EFE0F0EFDFF0EFD0E0DFC0DFCF
BFCFC0000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000CFDFE0D0E0EFD0E0EFD0EFEF\
DFEFEFDFEFF0DFEFF0DFEFF0E0EFF0E0F0F0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000F0FFFFF0FFFF000000000000\
0000000000000000007F7F90000000000000000000000000000000000000000000
E0EFF0EFF0FFF0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000FFFFFFF0FFF0F0FFF0F0F0EFF0F0EFEFF0EF
E0EFDF000000D0D0C0CFCFBFBFBFAFB0AF9FA09F8F9F907F8F807080706F000000000000000000\
000000000000000000000000000000000000000000EFF0F0EFF0F0EFF0F0E0F0F0
E0EFF0E0EFEFE0EFEFDFEFEFDFEFEFDFEFEFD0E0E0CFDFDF000000000000000000000000A0B0B0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000006F606070606F7F60707F6F7F
8070808F7F8F908090A090A0AF9FAF000000000000000000000000000000000000D0CFE0DFD0EF\
DFDFEFE0DFEFE0E0EF000000000000000000000000000000000000000000000000
000000000000000000000000000000FFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000F0F0F0EFF0F0E0F0EFE0F0EFDFF0EFD0EFE0CFDFD0BFD0C0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000CFDFE0\
CFE0EFD0E0EFD0E0EFD0E0EFDFEFEF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000F0FFFFF0FFFFF0FFFF\
F0FFFFF0FFFFF0FFFFFFFFFF000000000000000000000000000000000000000000
E0EFF0EFF0FFF0F0FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000FFFFFFF0FFF0F0FFF0F0FFF0F0FFEFF0F0EF
000000000000DFDFCFD0D0C0CFC0B0BFBFA0B0AF9FA09F8F0000008F80708F7F6F000000000000\
000000000000000000000000000000000000000000E0EFEFE0EFEFDFEFEFDFEFEF
DFEFEFDFE0EFD0E0E0000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006060606F606F0000000000008070808F7F8F
9080909F8F9FA090A0000000000000000000000000000000000000000000000000DFD0EFDFDFEF\
E0DFEFE0E0EF000000000000000000000000000000000000000000000000000000
000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000F0FFF0EFF0F0EFF0F0E0F0EFE0F0EFDFEFE0CFE0DFC0D0CF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000C0D0DF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000EFF0F0EFF0FFEFF0FFEFF0F0\
F0FFFFEFF0F0F0FFFFF0FFFF000000000000000000000000000000000000000000
DFEFF0E0EFFFEFF0FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000000000FFFFF0F0FFF0F0FFF0F0FFF0F0FFEF
000000000000000000D0D0C0CFCFBFC0C0B0BFB0A0AFA090000000908F7F9080708F7F6F807060\
000000000000000000000000000000000000000000D0E0EFD0E0E0CFDFE0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060606F606F0000000000000000000000000000009F8F90
A090A0AF9FAF000000000000000000000000000000000000000000000000000000000000E0DFEF\
E0E0EF000000000000000000000000000000000000000000000000000000000000
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000EFF0F0EFF0F0E0F0EFE0F0EFDFEFEFD0E0DFC0DFD0BFCFC0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
A0B0BFA0BFC0AFBFCFB0C0CFBFCFD0BFD0DFC0D0DF000000000000000000000000
000000000000000000000000000000000000000000000000E0EFF0E0EFF0E0F0F0E0F0F0EFF0F0\
E0F0F0EFF0FFEFF0F0EFF0FFEFF0F0000000000000000000000000000000000000
DFE0F0E0EFF0EFF0FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000000000FFFFFFFFFFF0FFFFF0000000000000
000000000000000000DFDFCFD0D0C0CFCFBFC0BFAFBFB0A00000000000009F8F7F9080708F7F6F\
807060706060000000000000CFDFDFC0D0D0BFCFCF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F60606F606F000000000000000000000000000000000000000000A090A0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
000000F0F0F0EFF0F0E0F0EFE0F0EFDFEFEFD0EFE0CFDFD0BFD0C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000A0B0BFA0BFC0AFBFC0AFC0CFB0C0CFBFCFD0BFCFD0
C0D0DFC0D0DF000000000000000000000000000000D0E0EFDFE0EFDFEFEFDFEFEFDFEFEFDFEFEF\
DFEFEFDFEFEFE0EFF0DFEFEFE0EFF0000000000000000000000000000000000000
DFE0F0E0EFF0EFF0FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000000000FFFFFFFFFFFF000000000000000000
000000000000000000000000DFDFCFD0D0C0CFCFBF000000000000000000000000000000908F70\
908070000000000000706060A0B0BF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000006F6060\
70606F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000EFE0F0\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFF\
000000000000000000000000000000DFEFE0000000000000B0CFBFAFBFAF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000A0B0BF
A0B0C0AFBFC0AFBFCFB0C0CF000000000000000000000000CFDFE0CFDFE0CFDFE0D0E0EFD0E0EF\
D0E0EFD0E0EFD0E0EFDFEFEFD0E0EFDFEFEF000000000000000000000000000000
DFE0F0E0EFF0EFF0FFF0F0FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000DFDFD0DFD0CFD0D0C0000000000000000000000000000000000000\
9F8F7F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000070606F000000\
000000000000000000000000000000000000000000000000B0A0B0B0A0B0000000
000000000000000000000000000000000000000000000000000000000000000000EFEFF0EFEFF0\
F0EFF0000000000000000000000000000000000000000000000000000000000000
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFF000000\
000000000000000000000000000000000000000000B0CFBFAFBFAF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000DFEFF0E0EFF0EFF0FFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000000000000000FFFFFF000000000000000000000000
000000000000000000000000000000DFDFD0DFD0CF000000000000000000000000000000000000\
0000000000000000000000000000000000007F6F60706060000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007F6F70000000000000\
000000000000000000000000000000000000000000B0A0B0BFAFBFBFB0C0000000
000000000000000000000000000000000000000000000000000000000000EFEFF0EFEFF0F0F0F0\
F0F0FFFFFFFF000000000000000000000000000000000000000000000000FFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0000000\
000000000000000000000000000000000000000000AFBFB0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000A0B0BF000000000000000000000000
000000DFE0F0E0EFF0EFF0FFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000FFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000007F6F60000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000007F6F70806F7F000000000000000000\
000000000000000000000000000000B0A0B0BFAFBFBFAFBFC0B0C0C0BFCF000000
000000000000000000000000000000000000000000000000EFE0F0EFEFF0EFEFF0F0F0F0F0F0FF\
FFFFFFFFFFFF000000000000000000000000000000000000000000000000FFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0000000000000\
000000000000000000000000000000000000AFBFB0000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000008F90AF9FAFC0000000000000
000000D0E0F0DFEFF0E0F0F0F0F0FFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000AFA090AF9F8F0000000000000000000000000000000000008F7F6F80706F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007F6F6F7F6F70806F7080707F8F707F907F8090808F000000000000\
000000000000000000B0A0B0B0A0B0BFAFBFC0B0C0C0B0C0CFBFCFCFC0D0D0CFDF
000000000000000000000000000000000000000000EFE0F0EFEFF0F0EFF0F0F0F0F0F0FFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000FFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0000000000000000000\
000000000000000000000000000000AFBFB0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000909FB0A0B0CF000000
000000000000DFE0F0E0EFF0EFF0FFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFF000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000000000000000000000EFEFE0E0E0DF000000000000000000000000000000000000\
000000BFAF9FB0A09FAFA090AF9F8F000000000000000000000000000000907F70
8F7F6F8F706F80706F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000007F6F70806F70806F708F707F8F707F8F7F80907F8090808F9F8F90A08F9F000000000000\
000000B0A0B0BFAFB0BFAFBFC0B0BFC0B0C0CFBFCFCFC0D0D0C0D0D0CFDFDFD0E0
000000000000000000000000000000000000EFEFF0EFEFF0F0EFF0F0F0F0F0F0FFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0000000000000000000000000\
000000000000000000000000AFBFB0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000008F90AF9FAFC0B0BFD0
C0D0EFCFDFEF000000000000000000F0FFFFF0FFFF000000000000000000000000FFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000000000000000000000F0F0EFEFEFE0E0E0DF000000000000000000000000000000\
000000C0BFAFBFB0A0B0AF9FB0A09FAF9F90AF9F8F0000000000000000009F8F7F
90807F908070907F708F7F708F706F0000000000000000007F6060706060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000070606070606070606F706F6F7F6F6F7F6F6F806F70807070
8F707F8F707F8F7F7F907F80907F8090808F9F80909F8F90A0909FAF90A0AF9FA0000000B0A0B0\
BFAFB0BFAFBFC0B0C0C0B0C0CFBFCFCFBFCFD0C0D0D0CFDFDFD0E0DFD0E0E0DFEF
E0DFEF000000000000000000000000EFEFF0F0EFF0F0F0F0F0F0F0FFF0FFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000FFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0000000000000000000000000000000\
000000000000000000AFBFB0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000909FB0A0B0CF
BFCFE0C0D0EFD0DFEF000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000
000000000000000000000000FFFFF0F0F0F0F0EFEFEFEFE0EFE0DF000000000000000000000000\
000000CFC0B0C0BFB0C0B0AFBFB0A0B0AF9FB0A09FAF9F90000000000000A09080
A08F809F8F809F807F90807F908070907F70907F708F7F708F706F000000806F6F7F6F6F7F6F60\
7F6060706060706060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000706060706060706060\
7F6F6F7F6F6F7F6F6F7F6F6F0000000000000000000000000000008F707F907F80
907F8090808F9F808F9F8F90A08F90A08F9FA0909FAF90A0AF9FAFB0A0AF000000BFAFBFBFB0BF\
C0B0C0C0BFC0CFBFCFCFC0CFD0C0D0D0CFDFDFCFDFDFD0E0E0DFEFE0DFEFE0DFEF
EFE0EF000000000000EFEFF0F0EFF0F0EFF0F0F0F0F0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0000000000000000000000000000000000000\
000000000000AFBFB0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000008F90AF9FAFC0
B0BFD0C0D0E0CFDFEF000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000000000FFFFFFFFFFFFFFFFF0F0F0F0F0F0EFEFEFE0EFE0E0000000000000000000\
000000D0CFC0D0CFBFCFC0B0C0BFB0C0B0AFBFB0A0B0AF9FB0A09FAF9F90AF9F90
AF9F8FA0908FA09080A08F809F8F809F807F9F807F90807F90807F000000000000000000000000\
000000000000000000806F6F806F6F806F6F806F6F000000000000000000000000
000000000000000000000000000000000000000000000000000000806F6F806F6F000000000000\
0000000000000000000000000000000000000000000000000000009F808F9F808F
9F8F90A08F90A0909FAF909FAF9FA0AF9FA0B0A0AF000000000000000000000000C0B0C0CFBFC0\
CFBFCFCFC0D0D0C0D0D0CFDFDFCFDFDFD0E0E0DFE0E0DFEFE0DFEFEFE0EFEFE0F0
EFE0F0EFEFF0EFEFF0F0EFF0F0F0F0F0F0F0F0F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFF000000000000000000000000000000000000000000\
000000AFBFB0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000909FB0
AFB0CFBFCFE0C0D0EFD0DFEF000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0EFEFEFE0EFEFE0000000000000\
E0DFD0DFD0CFDFD0CFD0CFC0D0C0BFCFC0B0C0BFB0C0B0AFBFAFA0BFAFA0000000
000000000000AF9F90AF9F8FA0908FA0908FA08F80A08F809F8F80000000000000000000000000\
0000000000000000000000000000000000000000000000008F70708F70708F7070
8F70708F70708F70708F70708F70708F70708F70708F70708F707F000000000000000000000000\
000000000000000000000000000000000000000000000000000000A09090A0909F
AF909FAF9FA0B09FA0B0A0AF000000000000000000000000000000000000000000CFBFCFD0C0D0\
D0CFD0DFCFDFDFD0DFDFD0E0E0DFE0E0DFEFEFDFEFEFE0EFEFE0EFEFE0F0EFEFF0
000000000000F0EFF0F0F0F0F0F0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000FFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFF0FFFF000000000000000000000000000000000000000000000000\
AFBFB0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
9FA0BFB0BFD0C0D0E0CFDFEF000000000000000000000000000000000000000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0EFF0EFEFEFEFE0000000\
E0E0D0E0DFD0E0DFD0DFD0CFDFD0C0D0CFC0D0C0BFCFC0B0C0BFB0C0B0AF000000
000000000000000000000000B09F90AF9F90AF9F90AF908FA0908F000000000000000000000000\
0000000000000000000000000000000000000000000000009F807F90807F90807F
90807F90807F90807F9080809080809080809080809F80809F8080000000000000000000000000\
000000000000000000000000000000000000000000000000000000AF9FA0B0A0AF
B0A0AF000000000000000000000000000000000000000000000000000000000000000000DFCFDF\
DFD0DFE0D0E0E0DFEFEFDFEFEFE0EFEFE0EFEFE0EFEFE0F0EFE0F0000000000000
000000000000000000FFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000
FFFFFFFFFFFFFFFFFFF0FFFF000000000000000000000000000000000000000000000000AFBFB0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8F8FA0A0AFC0BFCFE0C0D0EF000000000000000000000000000000000000000000000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0F0EFF0EFEFEFEFE0\
EFE0DFE0E0DFE0DFD0E0DFD0E0DFD0DFD0CFDFD0C0D0CFC0D0C0BFCFC0B0000000
000000000000000000000000000000000000000000000000B09F9FAF9F90000000000000000000\
000000000000000000000000000000000000000000000000A08F8FA08F8FA08F8F
A08F8FA08F8FA08F8FA08F8FA08F8FA08F8FA08F8FA08F8FA08F90000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000E0D0E0\
E0DFEFEFDFEFEFE0EFEFE0EFEFE0EFEFE0F0EFE0F0EFEFF0000000000000000000
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFF000000000000000000000000000000000000000000000000000000000000
000000FFFFFFF0FFF0000000000000000000000000000000000000000000000000AFBFB0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000008F90A0A0B0C0BFCFE0CFD0EF000000000000000000000000000000000000000000000000\
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0F0F0EF000000\
000000000000EFE0DFEFE0DFE0E0DFE0DFD0E0DFD0DFD0CFDFD0C0D0CFC0000000
000000000000000000000000000000000000000000000000000000000000B0A0A0B0A09F000000\
000000000000000000000000000000000000000000000000AF9F90AF9090AF9090
AF9090AF9090AF9090AF9090AF909FAF9F9FAF9F9FAF9F9FAF9F9F000000000000000000000000\
000000000000000000BFA0AFBFAFAFBFAFB0BFAFB0BFAFB0C0B0BF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000EFE0EF\
EFE0EFEFE0EFEFE0F0EFE0F0EFEFF0EFEFF0000000000000000000000000000000
000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000EFF0F0000000000000000000000000000000000000000000A0BFAF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009090AFAFB0CFBFCFE0CFDFEF000000000000000000000000000000000000000000\
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0000000\
000000000000000000000000EFE0DFEFE0DFE0E0DFE0DFD0E0DFD0E0D0CF000000
000000000000000000000000000000000000000000000000000000000000C0B0AFBFAFAFBFAFAF\
BFAFA0BFAFA0BFA0A0BFA0A0BFA0A0B0A0A0000000000000B0A0A0B0A0A0B0A0A0
B0A0A0B0A0A0B0A0A0B0A0A0B0A0A0B0A0A0B0A0A0BFA0A0BFA0AF000000000000BFAFAFBFAFAF\
BFAFB0BFAFB0C0B0B0C0B0B0C0B0BFC0B0BFC0B0BFCFBFC0CFBFC0CFBFC0000000
000000000000000000000000000000000000000000000000000000000000000000000000EFE0EF\
EFE0F0EFE0F0EFEFF0000000000000000000000000000000000000000000000000
000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
F0F0F0EFF0F0E0F0EFE0F0EFDFEFEFDFEFE0CFE0DFBFD0CF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000909FB0AFBFD0BFD0E0CFDFEF000000000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000EFEFE0EFE0DFEFE0DFE0E0DFE0DFD0000000
000000000000000000000000000000000000000000000000000000CFBFBFCFBFB0CFBFB0C0BFB0\
C0B0B0C0B0B0C0B0AFC0B0AFC0B0AFC0B0AFC0B0AFC0AFAF000000000000000000
000000000000000000000000000000000000000000000000000000000000C0B0B0C0B0B0C0B0BF\
C0B0BFCFBFBFCFBFC0CFBFC0CFBFC0CFBFC0D0C0CFD0C0CFD0C0CFD0CFD0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
EFEFF0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000EFF0F0
EFF0F0E0F0EFE0EFEFDFEFEFD0EFE0C0DFD0B0CFC0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000009F9FB0B0BFD0C0D0E0CFDFEF000000000000000000000000000000\
000000000000000000FFFFFFFFFFFFFFFFFF000000000000000000000000000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000EFE0E0EFE0DF000000000000
000000000000000000000000000000000000000000000000000000D0CFC0D0CFC0D0C0C0D0C0BF\
CFC0BFCFBFBFCFBFBFCFBFBFCFBFBFCFBFBFCFBFBFCFBFBF000000000000000000
000000000000000000000000000000000000000000000000000000000000CFBFC0CFBFC0CFBFC0\
D0C0C0D0C0CFD0C0CFD0C0CFD0CFCFDFCFD0DFCFD0DFCFD0DFCFD0DFD0DF000000
000000000000000000000000000000000000000000000000000000000000000000000000F0EFF0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000FFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000EFF0F0EFF0F0
E0F0EFDFEFEFDFEFEFD0E0E0C0D0CFB0C0BF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000B0C0DFC0D0E0000000000000000000000000000000\
000000000000000000FFFFFF000000000000000000000000000000000000000000
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000E0D0D0DFD0CFDFD0CFDFCFCF\
DFCFCFD0CFC0D0CFC0D0CFC0D0C0C0D0C0C0D0C0C0D0C0C0000000000000000000
000000000000000000000000000000000000000000000000000000000000D0C0CFD0CFCFDFCFCF\
DFCFD0DFCFD0DFCFD0DFD0DFE0D0DFE0D0DFE0D0DFE0DFE0E0DFE0EFDFE0000000
000000000000000000000000000000000000000000000000000000000000F0EFF0F0EFF0F0F0F0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000FFFFFF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000EFF0F0E0F0F0E0F0EF
DFEFEFD0EFE0CFE0DFBFD0CF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000B0CFDFC0D0E0000000000000000000000000\
EFFFFFF0FFFFF0FFFFF0FFFFFFFFFF000000000000000000000000000000000000
000000000000000000000000000000000000000000FFFFFFFFFFFF000000000000000000000000\
000000000000000000000000000000000000000000000000000000F0EFEFF0EFEF
F0EFE0000000000000000000000000000000000000000000000000EFE0DFEFDFDFE0DFDFE0DFD0\
E0D0D0E0D0D0E0D0D0DFD0D0DFD0D0DFD0D0DFD0D0DFD0CF000000000000000000
000000000000000000000000000000000000000000000000000000000000DFD0D0E0D0DFE0D0DF\
E0D0DFE0DFDFE0DFE0E0DFE0EFDFE0EFDFE0EFE0EFEFE0EFEFE0EFEFE0EF000000
000000000000000000000000000000000000F0EFF0F0EFF0F0EFF0F0EFF0F0F0F0F0F0F0F0F0FF\
F0F0FF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000000000000000000000F0F0F0EFF0F0E0F0EFDFEFEFDFEFEF
D0EFE0C0DFD0B0CFC0000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000CFE0EFDFEFF0E0F0FF\
EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000000000
000000000000000000000000000000000000000000000000FFFFFF000000000000000000000000\
000000000000000000000000000000000000000000000000F0F0F0F0F0F0F0F0EF
F0EFEFF0EFEFF0EFEF000000000000000000000000000000000000EFE0E0EFE0E0EFE0E0EFE0E0\
EFE0DFEFE0DFEFDFDFEFDFDFEFDFDFEFDFDFE0DFDFE0DFDF000000000000000000
000000000000000000000000000000000000000000000000000000000000EFDFE0EFDFE0EFDFE0\
EFE0E0EFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFEFE0EFF0E0EFF0EFEFF0EFEF
000000000000000000F0EFF0F0EFF0F0EFF0F0F0F0F0F0F0F0F0F0F0F0F0F0F0FFFFF0FFFFFFFF\
FFFFFF000000000000000000000000000000000000000000000000000000000000
000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000EFF0F0EFF0F0E0F0EFDFEFEFDFEFEFD0EFE0
C0D0D0B0C0BF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000D0E0EFDFEFF0\
E0F0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000000000
000000000000000000000000000000000000000000000000FFFFFF000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFF0F0F0F0F0
F0F0F0F0F0F0F0F0EFF0EFEFF0EFEFF0EFEF000000000000000000F0EFE0EFEFE0EFE0E0EFE0E0\
EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0EFE0E0000000000000000000
000000000000000000000000000000000000000000000000000000000000EFE0EFEFE0EFEFE0EF\
F0E0EFF0E0EFF0E0EFF0EFEFF0EFEFF0EFEFF0EFF0F0EFF0F0EFF0F0EFF0F0EFF0
F0EFF0F0EFF0F0EFF0F0F0F0F0F0F0F0F0F0F0F0F0F0F0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000\
000000000000000000000000EFF0F0EFF0F0E0F0EFDFEFEFD0EFE0CFE0DFBFD0CF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000D0E0EF\
DFEFF0E0F0FFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000000000
000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFF0FFF0F0F0F0F0F0F0F0F0F0EFF0F0EFF0EFEFF0EFEFF0EFEF000000000000F0EFEFF0EFEF\
F0EFEFF0EFEFF0EFE0F0EFEFF0EFEFF0EFEFF0EFEFF0EFEF000000000000000000
000000000000000000000000000000000000000000000000000000000000F0EFEFF0EFEFF0EFEF\
F0EFEFF0EFEFF0EFF0F0EFF0F0EFF0F0EFF0000000000000000000000000000000
F0F0F0F0F0F0F0F0F0F0F0F0FFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFF000000000000000000000000000000000000000000000000FFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000\
000000000000EFF0F0EFF0F0E0F0EFDFEFEFDFEFEFD0EFE0CFDFDFB0CFC0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
CFDFEFDFEFF0E0F0F0EFFFFFEFFFFFF0FFFFF0FFFFF0FFFFF0FFFF000000000000
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000\
000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFF0F0FFF0F0F0F0F0F0F0F0F0F0F0F0F0F0000000000000000000000000\
000000000000000000000000F0EFEFF0EFEFF0EFEFF0EFEF000000000000000000
000000000000000000000000000000000000000000000000000000000000F0EFF0F0EFF0F0EFF0\
000000000000000000000000000000000000000000000000000000000000000000
FFF0F0FFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
EFF0F0EFF0F0E0F0EFE0EFEFDFEFEFD0EFE0CFE0DFBFD0CF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000CFDFEFD0E0EFE0F0F0EFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFF000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFF000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0F0FFF0F0000000000000000000000000\
000000000000000000000000000000000000000000000000F0EFF0F0EFF0F0EFF0
F0EFEFF0EFEFF0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0F0F0F0F0F0F0F0F0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFF000000000000\
E0F0F0E0EFEFDFEFEFD0EFEFD0E0E0BFD0D0AFC0BF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000C0DFE0D0E0EFDFEFF0EFFFFFEFFFFFEFFFFFF0FFFFF0FFFFF0FFFF
000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000F0F0F0F0F0F0F0F0F0
F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0F0F0000000000000000000\
000000000000D0EFE0C0DFD0B0CFC0000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000CFE0EFDFEFF0E0F0FFEFFFFFEFFFFFEFFFFFF0FFFF
F0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000FFF0F0FFF0F0FFF0F0
FFF0F0FFF0F0FFF0F0FFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFF000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFF0000000000000000000000000000000\
000000000000000000A0BFB0000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000CFDFEFD0EFF0E0F0FFEFFFFFEFFFFFEFFFFF
F0FFFF000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000FFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFF000000000000000000000000000000000000000000\
000000A0BFBF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000C0DFE0D0E0EFDFF0F0EFFFFF000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000FFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000\
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFF0FFFF000000000000000000000000000000000000000000000000000000\
9FB0AF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B0CFDF000000000000000000000000
000000000000000000000000000000000000F0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000
000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
F0FFFFF0F0F00000000000000000000000000000000000000000000000000000009FB0AF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000B0CFDF000000000000
000000000000000000000000000000000000000000000000F0FFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000FFFFFF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000FFFFFF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000FFFFFFFFFFFFF0FFFF000000
0000000000000000000000000000000000000000000000000000009FB0AF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000BFCFDF
000000000000000000000000000000000000000000000000000000000000F0FFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF000000000000000000\
000000000000000000000000000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000A0BFBF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
AFBFCFBFCFDF000000000000000000000000000000000000000000000000000000000000F0FFFF\
FFFFFFFFFFFF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000FFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000EFF0F0EFF0F0E0F0F0000000000000
0000000000000000000000009FB0B0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000AFBFCFBFCFDF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000000000\
000000000000000000F0F0F0EFF0F0E0F0F0E0EFEFDFEFEFDFEFEFD0EFEF000000
000000AFBFBF90AFAF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000AFBFCFBFCFDF000000000000000000000000000000000000000000\
EFFFFFEFFFFFF0FFFFF0FFFF000000000000000000000000000000000000000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000000000000000000000000000\
EFF0F0EFF0F0E0F0F0E0F0F0E0EFEFDFEFEFD0EFEFD0EFEFC0DFDFB0CFC0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000AFBFCFBFCFD0000000000000000000E0F0FFE0F0FF\
E0F0FFEFF0FFEFF0FFEFFFFFEFFFFFF0FFFFF0FFFF000000000000000000000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000FFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000000000EFF0F0EFF0F0EFF0F0E0F0F0\
E0EFF0DFEFEFDFEFEFDFEFEFD0EFEFCFE0E0BFD0D0AFC0BF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000C0D0E0CFE0EF\
DFEFF0E0F0FFE0F0FFE0F0FFE0F0FFEFF0FFEFF0FFEFFFFFF0FFFFF0FFFF000000
000000000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
FFFFFFFFFFFFFFFFFFFFFFFF000000000000EFF0F0EFF0F0E0F0F0E0F0F0E0EFF0DFEFEFDFEFEF\
D0EFEFD0EFEFC0DFDFBFCFCF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000C0D0E0CFE0EFDFEFF0DFF0FFE0F0FFE0F0FFE0F0FFEFF0FFEFF0FFEFF0FF
EFF0FFF0FFFFF0FFFFF0FFFFF0FFFFF0FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF0FFFFF0FFFF
F0F0FF000000000000000000000000E0F0F0E0EFF0DFEFEFDFEFEFD0EFEFD0EFEFCFE0E0BFD0D0\
AFC0C0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000C0D0E0CFE0EFDFEFF0DFF0FFE0F0FFE0F0FFE0F0FF
E0F0FFEFF0FFEFF0FF000000000000000000000000000000F0FFFFF0FFFFF0FFFFFFFFFFFFFFFF\
FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000FFFFFF\
FFFFFFFFFFFFFFFFFFF0FFFFF0FFFFF0F0FF000000000000000000000000000000
000000000000000000000000000000DFEFEFD0EFEFD0EFEFC0DFDFB0CFCF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C0D0DFCFDFE0D0E0EFDFEFF0
DFF0FFDFF0FFE0F0FF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000F0FFFFF0FFFFF0FFFF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000F0F0FF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000AFC0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000EFF0FFEFF0FFEFF0FFEFF0FFEFF0FF
EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0EFF0F0E0F0F0E0F0F0E0F0F0E0F0F0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000A0BFBF9FAFAF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000AFBFC0B0C0CF000000000000000000000000000000000000000000\
000000000000000000000000000000000000DFEFF0DFEFF0DFEFF0DFF0F0E0F0F0
E0F0F0E0EFF0DFEFF0DFEFF0DFEFF0DFEFF0DFEFF0DFEFF0DFEFF0DFEFF0DFEFF0DFEFF0000000\
000000000000000000000000000000000000000000A0B0BF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000A0B0BFA0B0BFAFBFC0AFBFC0\
B0C0CFB0C0CF000000000000000000C0D0DFC0D0DFC0DFE0CFDFE0CFE0EFD0E0EF
D0E0EFCFE0EFCFDFE0C0DFE0C0DFDFC0D0DFBFD0D0BFCFD0BFCFCFB0C0CFB0C0CF000000AFBFC0\
AFBFBFA0B0B0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[20]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7Q@00491aT76@RY2:T76@LI1aT75m00P700061gX000>@0000
1@006i2:T8Z@RY2?ZHnYSjV?ZHnYSjV:T8Z@RY2:T0040003RP0006h020L000H7PP006XZ@RY2:T8Z@
RY2:T8Z@RY2:T8Z@RY2:T8Z@20000g40001[00P700061gT000R@ZI2YT:V@ZAX000J@ZI2YT:Ua00P7
00061g<000Z@RY2:T8Z@Ri2;0i0K0009ZI2YT:V@ZHnY06h020L000H7J@000g40000400:@00jYT:V@
ZI2YT:V@ZI2YT1/000f@ZI2YT:V@ZI2YT:V@06X020L000H7IP000g6@L@040006T8Z@Ri2;29001JV@
ZI2Y01/0016YT:V@ZI2YT:V@ZI2YT:V?Z@0;0003T00005P020L000H7I0001Y2:T8^@00B@01:YT:V@
ZI2YT:V@ZI2ZT:Z@ZY0J000=[jV_ZJnY[jV_ZJjY[P02Z@06T:V@ZI2Y2@000i00001G00P700061f80
00AaT76@0P001hZ@RY2;T8/01Y002jV@ZI2YT:V@ZI2Y01X001Z@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2Y
T:V?Z@@000>:T8X0E`081`001PMP0004T8Z@R`@00Y005JV@ZI2YT:Z@ZY2ZT:Z@ZY2_T:n@[`0K000S
[jV_ZJnY[jV_ZJnY[jV_ZJnY[ZV^ZJj@ZI2YT:V@ZI2YSi00E@081`001PMN0003LI1a00L000>;T8/0
1Y003ZV@ZI2YT:V@ZI2YT:V@1P001jZ@ZY2ZT:X03P001ZV_ZJnY[PZY00f@ZI2YT:V@ZI2YT:V@0080
00NYSjV:T8Z@058020L000H7G@000hZ@R`07000DT:V@ZI2ZT:Z@ZY2_T:n@[i2_T:l4000hZJnY[jV_
ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[ZV^ZJh40007T:V@ZHnY
RP1@00P700061e/000>@LI002@08T004ZI2YT0`001Z_T:n@[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_Z@P0
00V_ZJnY[jV^ZJh03JT70005SjV:T8X0C`081`001PMH0003T8/000`000J@ZY2ZT:lB0005ZZnZ[jX0
5:l00lH0000>000=ZJnY[jV_ZJnY[jV_Z@0<0006ZI2YSjV?BP081`001PMF0003LI1a00h00Y0F000J
[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jXD0009[jV_ZJnY[ZV^00h000N?ZHn@RY1a04L020L000H7
E@000hZ@000>0003ZP0001H01Jl05<J_aZo7[lJ_aZo6[lJ_aZo6[lJ_6@001JV_ZJnY010000JYT:V@
ZHm600P700061f<00Y000jZ@000G01V_<@001:V?ZHY500P700061e<000>;00002P001jZ@ZY2_T:l0
60006Jo7[lN_ajo7[lN_ajo7[lN_ajo7[lN_ajl07@000jV_000C0003T:V?04@020L000H7DP000g6@
000900>@00JZT:Z@ZY0H00J_01>`[lN_ajo7[lN_ajo7[lJ_aZo601d000B_ZJnY50000jV:001200P7
00061e4000>;T8/020002jZ@ZY2_T:n@[i6_01P001V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_01`0
00O6[lJ^aZW6018000>@ZI00@P081`001PM@0004LI2;T0H00i002i6@ZY2ZT:n@[i6_01L001W7[lN_
ajo7[lN_ajo7[lN_ajo7[lN_ajo701`01:l01ZV_ZJnY[a4000>Y0000@0081`001PM?0003Ri2@00:@
1@003I2ZT:n@[i2_TJnZ[jX00ZlG000I[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`0L000<aZo6[lJ_
aZk6[/JY40000jT0000o00P700061dh000IaT8^@Ri04000@T96@ZY2ZT:nA[i6_ZZnZ[aD000?7[`00
5@000`N_1`0K0005aZo6[lH00Zl02:j_[ZnY[jV_3P000i2Y000o00P700061dd000=bT9000i03000<
ZY2ZT:n@[i6_ZZnZ0Zl01;2_/:l<000;1jl7[`N`1k07/0L06@001PO71lL7aaD0016_1jl7[lJ_aZo6
[lJ_aZk6Z@0<0004[ZVYT3h020L000H7B`00291aT7:@Ri000Y00396@TI2ZT:nA[i6_ZPJ_00>`0000
0`004[2_ajl7[`N_1jl7[`N_1jl7[aT000f_1jl7[`N_1jl7[`N_00d000j_1jo7[lJ_aZo6[lJ_aP:_
00N^[jV_ZJnY00H000>^ZJT00ZT00i2YS`0k00P700061dX000>@Ri000i003940002ZT:n@[i6_ZP:_
01j`[k2_/:o7[`N_1jl7[`N_1k07/0N`1k07/0O71lL41aP04PL8000I1lH7aPO61jl7[`N_bZo6[lJ_
aZo6[/J^aP040009ZJnY[ZV^ZJj@03X020L000H7B0001g6@LI1bT8/01@002I2ZT:Z@[i6_ZP06[`03
/:n`00<001N`[lN_1jl7[`N_1jl7/0N`1k07/0N`1`0H000Ea`O71lL7a`N`1jl7[`N_1jl7[`N_00D0
01:_1jl7[`N_1jo6[lJ_aZo6[lH2[`06[Zn^[jV_0`001JjY[ZV^00:Y00>@ZHl0>@081`001PM70003
LI2;00B@00>A00000`001ZnA[i6_ZP:_00N`[k2_ajl700H000V`1k07/0N`1k003PLH01d702K61lH7
aPO61lH7[`N_bZo:[lJ_aZo6[/H0[jV_ZJnY[jV^ZJj@ZCL020L000H7AP00276@LI1bT8^@1`001ZnA
[i6_ZPB_00>`[`0020005Zl7[`N`1k07/0N`1k07/0N`1k07/0LH00`700g71lL7a`O61lH7aPO600<0
01B_1jl7[`N_1jl7[lZ_aZo6[lJ_aP:_00b^[jj_0:nY[ZV^ZJh2Z@04T:V@ZCH020L000H7AP000i2;
T003T003T@0000H000ZZ[jn`[k2_/:o72`0010N`1k0B1aP06@L500<701G:1lH7aPO61lH7[`N_bZo:
[lJ_aZh00P003/JY[jV_ZJnY[ZV^ZJj@=0081`001PM50007T76@LY2;T0090003[i6_00>_00>`[`00
30005K07/0N`1k07/0N`1k07/0Ne1kD7]@021aH06@L8000E[`N_1jl7[`N_1jo:[lZ_aZo6[lJ^00D0
00JY[ZV^ZJh4Z@03T:V?03<020L000H7A@000h^@T002T0/000J_/:n`[lL?01L75P0I1`X00`L04<X7
bPO:1lH7aPO61jo:[lX7000<[jV_ZJnY[ZV^ZJj@<P081`001PM50004T7:@R`d00jlB000>1k07/0N`
1kD7]@Ne1kD81`H000<700003@0I1`d00107aPN_1jl7[`N_bZo:[lJ_20001jnY[ZV^ZJh00ZT00i2Y
S`0a00P700061dD000>;T9003P000k2_000C01071@0C1`<060L@008700c:1lX7bPO61lH7[lX:000;
[ZnY[jV_ZJjY[Y00<0081`001PM20004L@00T0h000>_/:l05P0010Ne1kD@01H72@0?1aD000S61jl7
[`N_bPd000N_ZJjY[ZV^00:Y00>?Z@00;@081`001PMC0005[`N_1k005@000`L0000B01D73@0:1aL0
0PL01<X7bPL@000<[ZnY[jV^ZJj@Z@2Y:`081`001PM40003LP0000/000J_/:n`[k0B0003]@Ne01D0
1PL00kD71`0<1a401PLI00041jl7[a4000bY[ZV^ZJj@ZHnYSi0Z00P700061d0000J@0000TI0;0008
[`N_1k07/0L>00L75@0D1c8000?:00004@003:j_ZJnY[ZV^T02@ZBT020L000H7A0000i2A0008000:
/:n`[k2_]Jl7/0`000<7]@L01PLE000D1kD7]@Ne1kD7]@Ne1kD7]@Ne1kDF00031`0001X000>_0000
4P001jjY[ZV^T:T00P000hn@000W00P700061d<000>@TI002@002Zl7[`N`1k07/0L:00`7500D1aH0
10LH00<750001JnY[jV^00@000>@Z@009P081`001PM20005LY2@TI001`003:n`[k2_1jl7/0N`1`L0
00<7]@L02`L?0005]@Ne1kD0400010Ne1kDF00L750001LX7aPO601@000B^ZJjY1@000hn@000U00P7
00061d800Y000jZ@[`06000;[k2_1k07/0N`1kD00`L4014720031`09]@Ne1kD7]@Ne01<000Be1kD7
4P091a801`LC0003[jV_00H000>@Z@0090081`001PM10004T8^@T0:A1002[`0D/:n`[`N_1k07/0N`
1k0700Ne1kD91`07]@Ne1kD7]@03000?]@Ne1kD7/0N`1jne[kF_01@000P7/0N`1kD7]@d030L?00<7
00K:1lX7[`LB0003ZJh000L000>?00008`081`001PM000:@00JAT:nA[i42000</:n`[`N`1k07/0Ne
1@L00`071`0B1`<00PL01[D7]@Ne1ah000Z`1kD7]@Ne1kD72@0>1``02`L00lX0000?0003[P0000L0
00>?00008P081`001PM00009LY2@TI2_TJnF00:_00j`[k2_1jl7/0N`1k07/0@02`L01KD7]@Ne00L0
00B`1jn_9@002KF_]K07/0Ne1`060003]@L700h72@081`06bPO61jl740000jT000060003T8X00280
20L000H7?0003900T92AT:nA[i6`[`8000Re[`N`1k07]@H720091`L000>e0000<P001[07]@Ne1`80
50L301<72`001LJ^aZj_00P000>?000080081`001PLl0003L@1b00:@00JAT96A[iH4000;/:l7[`N`
1k07/0L02`051`P000>_0000>0000kD000020007]@Ne1kD7]@0=1`803PL01/X7bPO61`X000J^[jj_
ZJh80003RP0001l020L000H7?0002Y00TI2AT:nA/:l600041k07]@D73@021`T000>_0000@@0W1`X0
00O:[lJ^aZj_00L000>YS`007`081`001PLk0004T76@LP:@0Y400jnA[`060008[`N`1k07/0L?0003
1`0000@000>_0000AP001PNe1kD7]@T70`0A1`04bPN_1`P00Zl01Zj_[ZnY[PL000>@00007P081`00
1PLk000:LY2@TI2_TJnA/0T010L?00@700>e1kD0CP0<1`D050L6000:[lZ_bZo6[/J^[`L000>Y0000
7@081`001PLk0009T02@T96@[i6_00/000<7/0003P001kD7]@Ne1jl0D@001KD7]@Ne00@7200?1`04
aPN_1`@000B_bZo60Zl01:j_[Zl2[PL000>@000070081`001PLj0008T780T96@[i42[`/000<70000
3@061eD01`L901<70P003/X7aPN_bZo:[lJ^aZj_1P02T1d020L000H7>P001G6@002@00:A00>_TJl0
2`000`L0000;00071kD7]@Ne1`1G00051kD7]@L0300<1`0>bPO61jl700N_1jo:[lH2[`05[Zn^[jT0
1`000hl0000K00P700061cX000B@00000Y400jnA/00<00872`071`03[`0005L000<700003@0A1`80
00h7bPO61jo:[lZ_aZk6[PL000>?00006P081`001PLj0003L@0000800Y400jl0000;00031kD70087
1P021`06]@Ne1kD7F`000`Ne000>00h700?600000P001`N_bZo:[lH00Zl01Jj_[ZjY02<020L000H7
>P000i0000020003TJnF00`020L300P7FP001[07]@Ne1`l030L6000=1lH7aPN_bZo:[lJ^[`0S00P7
00061cT000>@000040000kD7]@091`L000>e[`00G0001K2e/0Ne00l020L900071jl7[lZ_bP02[`05
[Zn^[jT08@081`001PLi0003LP0000D000>_0000200;1`L000<7[`00G`001@Ne1kD700h01PL;000>
1lX7aPN_bZo:[lJ^[jh50003T00001P020L000H7>@000i0000040003TJn_00P000@7]@Ne20L60003
1jl0068000F`1k07]@0>00<73@001`N_1jo:[lH00Zl01:j_[Zh40003T:T001T020L000H7>@000g80
00040004/:n`[`H030L700:_H`0010Ne1kD21`h00PL>000=1lX7[`N_bZo6[Zn^[`030003[Y2^01T0
20L000H7>@000i0000040004U[2_/0H000>e1kD020L700:_IP001K2e/0Ne00d000<700003P001@N_
1jo:00:_00^^[jj_ZJh0[Y2YS`0I00P700061cl000NF/:ne/0N`00<02`L70003[`L006L000>`1kD0
0`L:00<740004PO:1jl7[lZ_aZj_[ZnY[Y2^T1P020L000H7?`02[`0</:n`[`N`1`07]@Ne1PL70003
[`L003d000V:T8Z@RY2:T8X08`001Zne/0N`1`P010LA00051jo:[lH00Zl01:j_[Zl20004T:j@ZAP0
20L000H7?P002;2_/:l7/0Ne0PL00`071`081`L000<7[`00>@003I2;T92YT:V@ZI2YT:T090001Zl7
]@Ne1`D020L@000:1lH7[lZ_bZo6[P8000F_ZJj@Z@0G00P700061ch000ZA[jn`[kF_1k071@051cd0
01J@RY2:T8Z@RY2:T8Z@SjV?ZHnYSjV?3P00192:T8XA0005[kF`1k000`0;1`l000J_1jo:[lH2[`03
[P00008000B^T:V?5`081`001PLn0009/:ne/0N`1kD700P00`L600031`00030000>:T0000`00592Y
T:V@ZI2YT:V@ZI2YT:V@ZI2Y40000hnYS`0A0003[`Ne008700<01`L02`L?0007bPN_1jo:[`050005
ZJj@[Y005P081`001PLi0003T7L0008000VA/:n`[kF`1k002P051`03]@N_02d000=a00002P002JV@
ZI2YT:V@Z@0;000;ZI2YT:V@ZI2YT:T01@001Y2?T8Z@RPh000>_00000`000`Ne1`0:1`l000@7[lZ_
1`001I2^T:V?01D020L000H7>@000i2A00020008/:ne/0N`1kD:00P7:`000hX0000=0004T:V@[Q00
01RY[jV^ZJjY[ZV^T:j@[Y2YT:V@ZI2YSi0C00d73P000lX7[`080005[jV^T:h05@081`001PLi00:@
00^A002F/:n`[`N`1`0:00H700>e1`0020004I2:T8Z@RY2:T8Z@RY2:T8Z@00l000>@00003@000jV@
Z@0D000@ZJj@ZI2YT:V@ZI2YT:V@Z@<000J@Si2:T8XB00031kD700X73P000jl000080005[Y2YSi00
50081`001PLi000;T96@[`2`[kF`1k000PL900<73P04T00CZI2YT:V@ZI2YT:V@ZI2YT:V@Z@02T1D0
00J_T:nY[jTD000@[jV_ZJnY[jV_ZJnY[ZV^T0D000F@ZI2YS`0<0003]@0000@030L<0003bP0000T0
00FY[Y2^T00D00P700061cT00Y02T@0809J`[k2_1k0J000JT8Z@RY2:T8ZYRZV:T8Z@RY2:T8Z@RY2:
T8XB0008ZI2YT:V@ZI02ZA<0012Y[ZV^ZJjY[ZV^ZJj@[Y2Y1`000i2?T00C00`730000`N_000:0005
T:j@ZHl04`081`001PLi0004LY6A[`8000V_]K07/0L0T8/05P000jT0000>000:T:V@ZI2YT:V@ZA80
00Z@[jV_ZJnY[jV_50003jV_ZJnY[jV_ZJnY[jV_Z@0:0003Z@0000X000>_0000100<1`X00`L00lX0
00080005[i2^T:h04`081`001PLi00:@0Y440006]G6@LI2:20003i2:T8Z@RY2:T8^@RjV@Z@0K0005
ZI2YT:T04P001ZV_ZJnY[a<000n_ZJnY[jV_ZJnY[ZV^ZJh060000jl0000400/72@001PO:1jl7[`X0
00F^T:V?T00B00P700061cX00Y402Zl0T8^@Ri2;T8/4000DZI2YT:V@ZI2YT:V@ZI2YT:V@ZI0K00JY
00B@ZI2Y3P02T003ZLJ^01<000n^aZk6[Zn^[jj_[ZnY[jT06@000jl0000400/71P081`X000FY[Y2^
T00B00P700061cX001EbT76@RY2:T8Z@092:T8Z@RY2;T8/00Y003JV@ZI2YT:V@ZI2YT:T06`004JV@
ZI2YT:V@ZI2YT:V@ZHnY00P000F:T8Z@RP0<0006[jj_[Zn^20001jnY[jV_ZJh030000hn@000;0003
[`0000<000>e1`L02@L300L700Bd1jl72P001I2^Si2?014020L000H7>@001XZ@Ri2;T0D001ZYT:V@
ZI2YT:V@ZI2YT:V@ZY2ZT:Z@ZY2ZT1/000jY[jV_ZJnY[ZV^ZJjY[PRY10001i2YT:V@ZI000`003Jo6
[lJ_aZo6[lJ_aZl030000jn^[`0<0003T:T000/000>_00000`0<1`0300L700T700?:000020001Jn@
[Y2^014020L000H7=@001Y1aT76@RPT000>;T8/01Y004:V@ZI2YT:V@ZI2YT:V@ZI0L00fY02:@ZI2Y
T:V@ZI2YT:V?08nYSjV:T8Z@RY2_aZo6[lJ_aZo60Zl?0003[`0000T000>YSi004P0:1`8020L01LX7
[`N_00X000B^T:V?4@081`001PLb0003T8^@00l001VZT:Z@ZY2ZT:Z@[i2_T:n@[i2_ZJnY[jV_01`0
01ZY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV^ZJjY[PL0012YT:V@ZI2YT:V_1jl7[`N_3`001jk6[Zn^[jT0
1@001I2^T:V@014010L800`700?:000020000jj@[P0A00P700061c0000=aT0004@006I2YT:V@ZI2Y
T:V@ZY2ZT:Z@ZY2ZT:Z@ZY0070005jnY[jV_ZJnY[jV^ZJjY[ZV^ZJjY[ZV^00BY2P0038nYSjV?T8[:
[lZ_bPl000V_[Zn^[jj_ZJh00`001jV@ZHn@Si002`02/0031k0700`020L01K@7[`N_00T000>@[Y00
4@081`001PM3000JT:Z@ZY2ZT:n@[i2_T:nY[jV_ZJnY[jV_ZJl70003[jT0018001^Y[jV_ZJnY[jV_
ZJnY[jV_ZJnY[jV_ZJnY[jT03P001:V@ZI0200031lH000h000B_bZo60Zl03Jj_[ZnY092^T:j@ZHl0
30000`Ne1`021``02PL00lX700090003T000010020L000H7;@000i1a000C000>ZI2YT:Z@ZY2ZT:Z@
ZY0<000g[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_0:nY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV^ZJjY
[P0@0003SjV?014000>_aZl00jl02Zj_[ZnY[Y00T:T2T003Si0000/000J_1k07]@L;00T700B_1jl7
2@000jh0000@00P700061b/000B;T8^@50001i2_T:n@[i004`007:Z_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_
ZZnZaZ[6Z/H8000CZLJYaZW6ZLJYaZW6ZLJYaZW6Z@0B0003ZI0000l000d7[lZ_bZo:[lJ^[jj_0080
00J@[Y2^T:T;00041k07]@<72`0;1`03bP0000L000>Y000040081`001PLY0005T76@RY005@000jX0
000G000L[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_Z@h000f_ZJnY[jV_ZJnY[jV_02@000B_bZo:
1:l01Jj_[ZnY00<000F@ZI2@S`0<0006]K07/0Ne2`081`05]0N_1jl01`000jV^000A00P700061bL0
00>;T8/0190@0005ZY2_T:l06P001jZ_ZZnZ[jX02:l03LJ_aZo6[lJ_aZo6[lH050001jk6[/J^aZh0
5@000i2Y000<00091jl7[`N_bZo:00:_00>^[`000`00192^T:h<0003/0Ne00@72`0:1`03bPL000H0
00>_Z@004@081`001PLU0008T76@LI2;T8/<000:TI2ZT:Z@ZY2ZT1X001Z_ZZnZ[jZ_ZZnZ[jZ_ZZnZ
[jZ_ZZnZ[jZ_ZP:_6@000jV_000D0004ZHnYS`@000<7[`L010002:l7[lZ_bZo:1:l60003T:V?00d0
00J_1k07/0L;00T700>_1jl01P001:n^[Y0A00P700061b<000>:T8/01i09000=ZY2ZT:n@[i2_T:n@
[`0J00N_01G6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lH06`000lJ^000C0005[ZV^T:T010021`0?bPO:0<X7
]0N_1jl7[lZ_00P000>^00002`0010N`1kD31`/02`L50005bZj_[Zl04@081`001PLQ000<T76@LI2;
T8^@Ri2;1P04T00<TI2ZT:Z@ZY2ZT:Z@6P001:nZ[jXH[a/000FY[jV_Z@0A00:Y00F@ZHnYS`030004
[`N_1`@000J_1jo:[lX2[aL000Ne/0N`1kD700/01`L01;@7[`L300>_0Zh00i00000?00P700061ah0
00J@Ri2;T8/6T00ITI2ZT:Z@ZY2ZT:Z@[i2_T:nA[i6_TJnA[`0J000L[lN_ajo7[lN_ajo7[lN_ajo7
[lN_ajo7[`N_1a/000c6[lJ_aZk6[/J^aZh:000;[jV^ZJjY[ZV^T:T00P000hl000070003[`L001P0
00@7/0Ne0PL;00X70P001:l7[lX2[`03[Zn@010020L000H77P00376@LI1bT8^@Ri2;T0800i00596@
TI2AT96@ZY2ZT:Z@ZY2_TJnA6P08[`05/:n`[k002Jl01/J_aZo6[a/00Zl03Jj_[Zn^[jj_[ZnY[jT0
1`003JV^ZJjY[Y2YT:V?ZHl09P001KF`1k0700/02@L01[@00<Z_bP:_00B^[i2^40081`001PLN0004
T8^@R`B@00>AT0000`005jZ@ZY2ZT:n@[i2_TJnA[i6_TJnA[jZ_01X001b_ajo7[lN_1jl7[`N_1jl7
[`N_1jl7[`N_1jl76`004/J_aZo6[lJ_aZk6[/J^aZk6[P@000n_ZJnY[jV^ZJjY[Y2YT:T02@000`N_
00080003T00000d000>`1kD00`L;00L710001:l7[lX2[`03[Zn@010020L000H77@00291aT7:@Ri2;
1`005i2AT96@TI2AT:Z@ZY2_TJnA[i6_TJnA01X001b`[k2_/:n`[k2_/:n`[lN_ajo7[lN_ajo7[`N_
6P001LJ_aZo600:_026^[jj_[Zn^[jj_[ZnY[jT0ZJjY[ZV^ZJjY[Y2YSjV?T8X01`000jo:[`0700:@
3P002:l0[`N`1kD72P051`L000R_bZn_[Zn@[Q0020L000H770000i2;T005T0P001NZT:Z@[i2_TJnA
[i6_TJnA[i6_ZZnZ[`0J0005[`N_1jl01@004Zl7[`N_1jl7[`N_1jl7[`N_1aX002V_bZo:[lZ_aZo6
[lJ_aZo6[/J^aZk6[P2^[jV_ZJnY[ZV^ZJj@Z@2YT00600041jl7[`H000>_T:h03@02[`040;07]@87
2`000`L0000800081jl7[kB^[i0@00P700061a`000IbT7:@Ri0:000GT96@TI2AT:Z@ZY6_TJnA[i6_
TJnA[i4030003[2_/:n`[k2_/:n`[k2_=P002LJ_aZo6[lJ_aP04[`0:[Zn^[jj_[Zn^[`<000V^ZJjY
[ZV^ZJh00ZT01Hl0002?00@000H7[`N_bZl50003UJj@00h000J`[`2`1k021`P00PL;0008[lZ_[jj_
UJh@00P700061a`0190<000G[i2_TJnA[i6_TJnA[i6_ZZnZ[jZ`[k001@005K2_/:n`[k2_ajo7[`N_
1jl7[`N`1`0L00P74P005jl7[`N_bZo:[lZ_bZo:[lJ_aZo6[/J^00D000^^[jV_ZJnY[ZV^T00600<7
00Jd1jl7[`L200:_00>@[Y003@000jne000700T72P0020N_1jo:[Zn^40081`001PLL0003LY0000d0
01N@TI2AT:Z@ZY6_TJnA[i6_TJnA[i6_0006[`0D/:n`[k2_/:n`[k2_/:n`[k2_/:lL000>1jl7[`N_
1jl7[`N_1jl<000?bZo:[lZ_bZo6[lJ_aZo600B_00B^[jj_20002:V^ZJjY[Y2Y1@002`Nd1jl7[`N_
bZnd00:_0Zh2T0d00Zl700X72`000lZ_]002[`:^40081`001PLK0003T00000l000jA[i6_TJnA[i6_
TJnZ/0P001Z_/:n`[k2_/:l7[`N_1jl7/0N`1k07/0N`1a`04PL00lX7bP05000G[`N_1jl7[`N_1jl7
[lZ_bZo:[lJ_aZl02P001ZnY[jV^Z@H00`L02;D7]0N_1jo:0Zl00iF_T00=0003T;2_00H02PL;0004
[`N_bP:_0Zh?00P700061bd000^AT:ZA[i6_TJnA[`0<000I[k2_/:n`[k2_/:n`[k2_/:ne[`N_1jl7
[`0L000]1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`07[`N_1jo:[lZ_bZo:[lJ_aZo600:_00>^[`00
2`001JV^ZJj@00H000P7[`N_1jl7[`8000>_[Zl00Y0>0003[k0000D02`L:0003bZnd00:_00>^[i00
3`081`001PLJ0003T000010000NA[i6_TJnA010001V`[`N_1jl7/0N`1k07/0N`1k07/0N`1k0701`0
60L00lX00002000D1lX7[`N_1jl7[`N_1jo:[lZ_bZl=0004ZJnY[PL000H7]@Ne1k@30004[kB^[`l0
0Zl700X72`000`N_bP02[`:^00<0T00030081`001PLH0003LY1b018000>_000050006Zn`[k2_/:ne
[kF_1k07/0N`1k07/0N`1k076P006[D71`Ne1kD7]@Nd1k@7]0Nd1k@7]0N_1jl72`003@N_1jl7[lZ_
bZo6[lH040000jT0000600031jl700H00Zl00i00000=0003T;2_00H02PL;0009[kB_[jj_T02?00d0
20L000H75`0018^@T94?0003[i6_01L001H7/0N`1k07/0N`1k07/0Ne1kD7]@Ne10LJ01X73`002<H7
[`N_1jl76P000i0000070003[P0000l000>F/0001P0:1`X000D7[`N_]002[P:@3@081`001PLF0005
LY1bT8/03@03T@03[i6_01H001Z_]Jne[`N`1k07/0N`1k07/0N`1k07/0N`1aX050L01LX7bPO:01<0
00F_bZo:[`0B0003[ZT000l000>_00003`02[`03/00000@02PL;0005bZn_[Zl00Y000hl0000;00P7
00061a@000N@Ri2@TI2A00/000R_TJnA[i6`TAH00107/0N`1k07/0Ne1kD7]@Ne2PLJ01T75@000`N_
1`0B0004[ZnY[PD000>Y000060000i6`[`0600X72`001Zl7[kB^00:@3@081`001PLC0006T7:@LY1b
0i08000;T96AUY6_TJnA[i405P006@N`1k07/0N`1k07/0N`1k07/0N`1kD7]@L06P0I1bX000F^ZJj@
[P040003S`0001T000>_/0001P0:1`X000S:[jn^[`2@S`d020L000H74@000g80LP04T004TI2AT0H0
00f_TJnA[i6_TK2F/:n`01H000Ve1kD7]@Ne1kD01`L:01<71P0H1aP000@7bPN_3`001Zj_[ZnY[P@0
00>@000010000k@0000A0003/:l000H02PL:0004[`N_]0800Y0=00P700061a0000UaT76@LY1bT780
0Y000i4000020003T96@00:A00Z_TJnA[i6_U[2F5`001[07/0N`1a<000<7]@L05PL6018760001Jl7
[`N_00d000R^[jV^ZJj@[P<000>@00000`02[a@000>`00001@0:1`X000?:[jl00`000i2?000<00P7
00061a0000B@LY2;0Y001Y6@TI2AT080016FTJnA[i6_TK2F/9J`[k2_/00E00031kD001P06@L<00`7
5`021`07bPO:1k@7[`0:0008[jj_[Zn^[jT30008T:T00:l7[k@M00X72@000jl7000300:@3@081`00
1PL?000:T740LY1bT7:@LP:@00BAT96@194039JA[i6_TK2F/:n`[`l000V`1k07/0N`1k00600I1a<0
10LH000;1jl7[`N_1jl7[lX020001:j_[Zl2[P03T:j@008000FYS`N_]@03[`0307H0018000>_0000
1@0:1`l000>@S`0030081`001PL?0003LP0000:@00NAT96@TI2A008000nA[i6_TJnA/9J`[k2_/:l0
0[000kF`0009000<1kD7]@Ne1kD7]@Ne0PLG01P760000`L0000D00D700W:1lX7bPNd1jl01@002:o:
[jn^[jj_0Zh02Y00T:j@1jl7[k@E00031kD000D02PL70003]0L000D00Y0=00P700061a8000EbT7:@
LP02T0D001>AUY6_TJnF/9J`[k2_/:n`[kF_008001Fe/;F`1k07/0N`1k07/0Ne1kD7]@L0600G1aL0
2PL=00<700o:1k@7[`N_1jl7[`N_bZl00P05[`0<[Zn^[Y2^T02?T:ne0ZlC000:T000[`N`1k07]@/7
00X01`L7]@N_1jo:1@000i2?000<00P700061a800Y00196@TI07000@TK2F/9J`[k2_]K2e/;F`1`<0
0107/0Ne1kD7]@Ne1kD7]@Ne1`LH01H75`0=1`X030L05<X7bPNd1jl7[lZ_bZo:[Zn^[i2^0P001I2Y
[`N_01H000<7/0L01@L:00L700>d1jl01@000i00000<00P700061a8000EbT7:@M`08000?[i6_U[2F
/:n`[k2_/:ne00D001Ne/0N`1k07/0N`1kD7]@Ne1kD7]@Ne1`0E00031kD000/000P7]@Ne1kD7]@87
5`0A1`D02PL02[@7[`N_1jl7[`L20003[kB_00>_00F^[i2^T0020003T8nd00:_00>@00004`002:l7
/0N`1kD72P041`06]@N_1jo:1@000i00000<00P700061a8000BAT96@2P002k2F/:n`[kF`]K0700T0
00X7]@Ne1kD7]@Ne3PL401475P041a805@L00`071`0@1`03bPNd00@000T7[lZ_]:j_[Zl00`001Jj_
002_01D000<7/0L01@L:00L700>d1jl01@000i00000<00P700061a<000>@00002P002IJ`U[2_/:n`
[`0<000D1k07/0Ne1kD7]@Ne1kD7]@Ne1kD3000G]@Ne1kD7]@Ne1kD7]@Ne1kD7]@Ne1k00500030Ne
1kD7]@Ne1kD7]@X05PL00`071`0:1`08]0N_1jl7[`L50005]:n_[Zl00Zh02900002?T02_5`001`N`
1k07]@L02P041`06]@Nd1jo:1@000i00000<00P700061a<000>@00002`001Zn`[kF`]@l000H7]@Ne
1kD81`X000T7]@Ne1kD7]@L00P001jn`[`N`1k001@L900@71`000kD7]@0>1`805`L401071`000lZ_
bP02[`03[P0000<000>@09005`000k071`051`T01`L00k@7[`050003S`0000`020L000H78P000jn`
000C0008]@Ne1kD7]@LB0005LY2@TI000P02[`09/:n`[`N`1k0700`000<700001P000`N_0004000;
/0N`1kD7]@Ne1kD00`0D1`L02@L01;D7]0L90005]:n_[Zl01@02T1P000H7/0Ne1kD21`X00`L01[D7
]0N_bQ@020L000H74`000i00000S0003]@L001H00Y000i6@[`020007[k2_1k07]@041``000<70000
1P001:o6[lH40007ZJnY[i2_]@0:00l72`0:1`X000C:[kB^1@000i00000C0005T@000;001`L:00H7
00>d1jl050081`001PLB0003LY0000d000>_/0004`000kD7000F0004T7:@T0:A00f_TK2_/:l7/0N`
1k0700/01@L40003]<J_00B_00/0ZJnY[i2YT:V@Z@0<0003]@L700L73@071`03]@0000X000>d[P00
6`002Zl00:l7/0Ne1kD21`X00`L01[D7]0N_bQ@020L000H74@00192AT94;0004]K2e/14000^e1kD7
]@Ne1kD7]@0A0005T96@[i400Zl01P2_1k07/0D72`081`<000N_1jl7[lJ_008000R_ZJnY[jV_T0l0
1@LA00D73P000lX0000I0006[i6`00N`1`L:00H700>d1jl010000i00000=00P700061a4000F@Mi6@
T@090007/:n`[kF`]@0=000:1kD7]@Ne1kD7]A@00Y00196@[i420009/:l7[`N`1k0700`020L20007
[`N_ajo6[`040006[jV_ZJn@2@02T0P000>e00004P021`l000>_00006@002I2_[`2_1k07]@041`X0
0`L01[D7]0N_]1@020L000H73P002I00TI2AT9JA[`05000>/:ne/;F`1k07]@Ne1kD4000;1kD7]@Ne
1kD7]@L05`02T0:A00>_T@001@051`/020L200D700C61jl72@000jnZ0006000;T:Z@ZY2ZT:T01kD0
1PL?00@73@000`N_000I00:_00B`[`Ne1`L:00H700>d1`0050081`001PL@0005LY1gTI000Y4089H0
002`[k2_/:ne/;F`]K07/0N`1kD7]@Ne1kD7]@N_6@000g:@T002T@P000<7/0L02`081`@000L7aPN_
1jl700X000JZ[jV_T:l30003ZY2Y00B@1@001`N`1kD7]@L02`081`X000?:[k@06P00192_[k020003
/0Ne00@72P031`05]0N_1jl010000iD0000>00P700061a0001ZAT96@UY6_TK00/:ne/;F`1k07/0Ne
1kD7]@X000@7002?5`001W:AT:nA[`T000<700002P071`L010L00lL0000800>_00FZ[jZ_T0020006
[i2_T:Z@1P00192e1kD41`P02PL800051k@7[k@060000i2_[`02/0<01PL:00H700>d1`000P000jn@
000?00P700061`l000B@LY2@0Y4059JA[i6_U[2_/:ne/;F`]K07/0N`2@000jl000020003T8Z@01D0
00F@LY6@[`0:0003/0L000T020L80004a`N_1`T01Jl01:Z_ZJl800J@0`001`N`1kD7]@L0100<1`L0
00H7[`N_]:l30003T00001800Y42[`03/00000<010L900@700Fd1jl7[`020003[i2E010020L000H7
3`000g:AT002T@0C[i6`U[2F0:ne/;F`1k07]@Ne1`0:0004]@L0[`8000F@Z@2@S`0C0005T96@[i40
2P041`L020L;00031`0000P000B_ajo71:l00jX000060006[i2ZT:Z@10000i4000041`0300L700d7
1@0020Ne1k@7[kB_0P000jh0000@0003R`2A00:_00<7/0004@071`07]0N_]:j_T00@00P700061a40
0Y02T@04UY6_UP@000Ve[kF`]K2e/0L02@000`N_00020004T000Z@:@00F:T8Z@L@0@00:@0Y400jl0
000700031kD700@710081`d000<700001`000lN_[`04[`03ZP0000L000J@ZY2AT9440003T94000L0
00>e1kD01`L01;D7]@L300071k@7[`N_]002[`030:j?01800Y001JnA/:n`00P00`L7008700Je1k@7
[`04[`:@40081`001PLB00:A00B_TK2F1P001PN`1kD7]A0000V_[P2@[Y2YT0000Y000hn@000<00:@
0Y400jl0000700X70P071`d01@L500091jl7[`N_ajo700T000F_TJnA[`050003[i6_00P03PL03P07
1`Ne1k@7[kB_]92E4P001i2_TK2_1k0020081`800PL600051jndUJh04@081`001PLB0005T96AUY40
2`000k07000?0006[ZnY092Y0Y030003T8Z@00`000=aT9000Y490003/0L700P70P071`03[`0000T0
0PL01LL7/0N_00<000V_1jo7[lN_/:l02@001I6_TJnA00D000F@TI40T@070005]@Ne1kD01@L03;D7
002e1k@7[`N_]0:_0YDB0003LI6A00:_00>`[`L0200:1`P00Zl2[P03T00000l020L000H74`00196_
TK0?00071kD7]@Ne1`0300051`00[k@00Zl01Jj_T:j@00D00Y0;00:@00>_T@002@0:1`P000@7]0L7
2P091`0:00N`1k07[`N_1`P00Zl01:Z_ZZl60004TJnA/0T020L5008700Ve1k@7[k@0[i004P001900
[jl2/0031k0000L02PL800051jn_UJh04@081`001PLD0003TIJA00`000T7]@Ne1kD7]@L00`0010L0
0003[`07[Zl0[Y2^T00=0008L@00076AT:l900031kD700L72@000jl7[`031`P01PL01;07[`L30006
1jl7[lN_20001ZnZ[jZ_T@H000F_T@2A/0020003/0L000@000D7]@Ne1`0700071k@7]0N_]0020003
T00001000Y001P2A[jn`[`T02PL800:_00>E[Xl04@081`001PLQ0007]@Ne1kD7]@031`8000Ze1k@7
00N_1jo:10000jj_00060005[Y2YT:T010000i00000:00X72@021`03[`L700H70`0:1`X000>`0000
1@001jn`[k2_/:l010000k0000020007[k2`1`07]@051`/000@7]@Ne5P02T0<000B_1k072@091`P0
00F_]:n_T00B00P700061aL000>_UP001`005PN`1kD7]@Ne1`07]@N_1jl7[lZ_bZl50003[P0000D0
00B@ZI2Y19000hn@000;0003/0L700L72@001PN_0;D7]@H70P0:1`X000T7[`N_ajn`[k001P001Jn`
[k2_00D000b_/:ne/0N`1kD7]@L;00031jl701H000>@T@000`000jl7000800X720001;B_[i0C00P7
00061aL000>F/:l01`001KD7]@Ne00T700Re0;@7[`N_1`H000>_00001@002JV_T:n@[Y2^T0030003
U[0000D02`L801T72P002K07/0N`1k07/0060005/:n`[k001@002;F`1k00]@Ne10LT0003ZP0000@0
00>`00001`0:1`P000B_]9FE4`081`001PLG0007[iJ`[kF_]@020006/0N`1kD710001[D7]@Nd1`80
00F_1jl7[`0600>_1@002JnY[Y2^T:V@Z@0300:_00>`00000`000`Ne1`081`P000_:1jl0/0N`1kD7
]@041`<01`L:00091k07/0N_1jl700H000J_/:n`[k050003/:ne00<000H7]@Ne1kD800031k@001L0
00>@00003P0;1`L00Zl2UA@020L000H75`002IJ`[kF`]K07]@0900H710000k@7]0070006[lZ_bZo6
0P002Jj_[Zn^[jV_Z@030004U[2_1`@02`L700<700>d00000P000kD71`041`L010L:00@700Fe1k07
/0070005/0N`1k001`000k00000400@710031`03]@L001H00Y070003/0L000H02PL80004]:n_T1@0
20L000H76000296_[k2_]Jne2P001KD7]@Ne00D000>_1`001P000lZ_aP08[`06[Zn^[jV_0P001i6_
[k2_1k000P0:1`P000@7[`N^1P001@N`1kD700T00PL:00091k07/0N`1k0700L000Je[kF_]Jl@000:
1kD7]@Ne1k@7[aL000=aT0001P000`N`1`0600X72002[`:@50081`001PLE000;T000[iJe/;F`1kD0
2P031a0000R_1jl7[`N_bP<000K:[lJ_aZl20008[iJ`[`N`1kD;1`L00PL20003[`0000:_10000kD7
1`0;00031`0000L01PL00kD7]@0800061k07/0N`1@000kD7000700H700Be1kD70P000jl0000B00:@
1`031`D02PL80004[kBEUAD020L000H75P002Y00TJnF/:ne[kD:0003]@L000L000>_1`001P0020N_
bZo:[lZ_20001:l00942[`05/:l7/0L00`071``000>^00001:l400031`0000T00`L00kD700040009
1k07/0N`1k0700P000N`1k07/;F`008000D7/0N`1`0700091kD7]@Ne1k@7008000>F00004`000i00
00050003[`N`00<70`0:1`P000B_[YF?5@081`001PLG000:T:l0/:ne/0N`1`T000<700001`041`@0
00T7bPO:1jl7[`L02`02[`04/:l7/0872@0010L000051`@000J_00001jl600031`0000P02@L00`07
1`071`T000ne1kD7]@L01kD7]@Ne1kD01@021`09]@Ne1kD7]0N_01H000>_00001@081`0300L700P7
1`001;F_[i0F00P700061aP000V@002`[k2_]Jl01P0010Ne1kD800061k@7]0N_0`002Jl7[`N_1jl7
[`0;0008TJn_/:l7/0L900P71@001936[lH80004[kF_]@L000/7]@Ne1kD7]@Ne1`0300061k07/0N`
20000`N`0005000A1k07/0Ne1kD7]@2e1kD7]@L00P000jne000G0003T00000@000<7/0L01PL00`07
1`071`P00Zl00i00000E00P700061aT000>@000010001PNe0007]@H71P0A1`03bPL000X00Zl2/004
1k071`T020L50004[lH7aPX000Fe[`N`1`0500X72@041`05]@Ne1kD020001PNe1kD7]@<720000jne
000C0006L@00096_1@0:1`@01PL70003[kBE01P020L000H76P000i2F0006000;]K07/0Ne1kD7]@L0
0P000kD7]@071`D000D7]0N_1`0<0008TJn_/:l7/0L900P71`000jl7000<00<700n_]Jne[`N`1k07
/0Ne1kD02P0020Ne1k07/0N`2P000`Ne1`080003[`0001@000V@0000ZY6_Z[000`000`N`1`071`X0
00P7]@Nd1k@7[a/020L000H76`000i2_000600041kD7]A872@000`L0000:0006/:n`/0N`0PL900P7
1`021`d010L20003[kD000<000Ne1kD7]@Ne00T00PL01kD7]@Ne1kD02P000`Ne000M0003T0000080
00FZ[jn`[`0300X72P061`03]0L001X020L000H77@000i400004000:[`N`1k07]@Ne1`8000>e1kD0
1PL;00031`0000T000NA[jn`[`N`00T020L50003bP00008730000`L0000?0003/0L000T000Ze1k07
]@Ne1kD71P002KD7/0N`1jne[`0M0003TJnZ00:_00J`002_1k071`X00PL01kD7]0N_]Jl06`081`00
1PLU0006]@Ne1kD71`071`/01PL60007/:n`/0N`1`0900P71@001Jl0000700L01PL=00871@001K07
]@Ne00D001Be1kD7]@Ne1kD700Ne1kD7]@Ne1b0000VA[jn`[`N`1`002PL:00D700>e1jl070081`00
1PLQ0003[`00008000>_1`002P000`Ne1`021``02`L2T@:_0[000`N`000800P70`000`N_1`0400d7
3@051`@000Ze[kF_]Jne[kF_30000kF_000N0007L@00092_ZP02[`06/:l7[`2`20L:00<700Jd1k@7
[k@L00P700061b<000FF]Jl7/00=0087300<1`03T02F00:`00<7/0L02@081`800`L00k@0000200d7
3P000`L0000;0003]@00008000<70000:0001H^@002A00:_00J`[`N`1k0300L72`061`03]0L001`0
20L000H790001IJ`[k2_00d000>e1`002@0<1`030900008000B`[`N`2@0:1`04bPNd1`800Y000`07
1`0:1`d000<700002P000kF_[`0^0007LI2@092_ZP02[`05/:l7[`L00`071`/000L7]@Nd1jnd01h0
20L000H79@001Y6`[kF_1`T01@L900`700<0T00010000k07000800H70P051`05bP00[i000P0<1`P0
1@L:0004[kF_[bh000F;T92_ZP02[`07/:l7/0N`1`0500H72`041`03]0N_01h020L000H79`001ZnF
/:ne[`D000N`1kD7]@Ne00<71@0=1`03092@00H000<700001P031`D010L01;@70002[P03S`0700/7
10081`03]@L000D000B_/:n_;`00292:T92AT02Z0Zl01[2_1jl7[`L010L:008700Fd1jne[`0O00P7
00061bX000b`[kF_1`07]@Ne1kDG1`8000>@UP001@000kD71`061`P01@L01<X0[lH2[P:@6@L30004
[kF_[bl000>@09000i002ZnA002`[`N`1k021`T00PL:00@700>e1jl080081`001PLd0009[`N_1jl7
/0Ne00d00PL40003T00000D000>`1`L01@L800@700C:1`3:0Zl01JV^T92:00T000^e1kD7[`N_1jl7
[`030003UP00034000=aT8/019000jl000030005ajl7[`L030021`P00PL01K@7[kF_024020L000H7
=`001K07]@Ne00l01PL700P720061`030:l700:_00F^[`2YT00800<700>e1kD0>P00292;092ZT:n@
1`001;07a`L<00@71P041`03]@N_028020L000H7>@000kD0000>0003]@Ne00@71`000k071`051`T0
1@L200B_00F@092@S`070003[`N_03/000=aT0000`000i2ZT0090003[`L000/01`L400051kD7]0L0
0ZlR00P700061cP00Zl01;F_]Jl;00T71P081`T00`L400031jo:00:_0P00192YSi020003[`N_03`0
00>@R`001@000jl0000F00X700<01`L00`L00jne000R00P700061c/000B_UZnF0Zl00k2_]@04000:
[`N_1jl7[`N_1`H000>`1`L01@L900871P000lZ_[`050005T000LJl0?0000g6@000C0003/00000T0
2`L50004[`2_[b8020L000H7@@0019J`[kD4000:/0N`1kD7]@Ne1`D020LA0003[`N_00D000>@0000
?@000h/000060003[jX000X00`L900/71`02[b<020L000H7F@000`Ne1`051`P000<700001P000jo:
000:0003T74003H000>:00001`000jZ_00090003/0L700871`0<1`L000>_00008P081`001PMC0003
UP0000<020L600<75`000hl0000c0003T00000H00Zl00lN_000800H71`0;1bd020L000H7E0000i2_
000200031kD700D70`071`l000>@[P001@000hl0000^0004T8Z@R`:@1@000jZ_ZP02[`03/:l000H0
00>`1`L01@L500/720000iH0000S00P700061eD000FA/000]@081`801`L700031lX000D000B_ZJnY
1@000i2YT00Z00B@00FYT:V@[`0300B_00G7[`N_1`0600X70`0;1`P000>_000090081`001PME00:A
00>`[`000P021`@02@L70003bPN_00D000J^[jV_T:h30005T8n@Si000`000i1a000J0009LI1aT8Z@
RY2;00B@00RYT:Z@0:Z_ZPB_00G7[`N_1`040003/0L700T70P0:1`P000>F00009@081`001PMF0004
[iJ`/0T02PL300L70`004PN_bZo6[ZnY[jV_T:j@ZI2YT0H0190<0003T00000P000V@ZY2_T:nY[jX0
0`002Jo7[`N_1jl7a`0C1`<02@L70003[`0002L020L000H7E`02UP03]Jl000L02`L00`071`051`08
[`L01jo:[lH2[`03[ZnY008000N^T:V@ZI2Y00/000^@RY2:T8Z@RY2:T00;0006ZI2ZT:nY1P02[`0:
ajl7[`N_1jl7/0803@L500L71`000jl0000X00P700061eP000>_/0L02@0E1`0:bPO61jl7[lZ_aP@0
00FY[jV_T00;000;T:V@ZI2YT:V@ZI002`000jV_ZP0:00041jl7a`D71@0;1`P010L70003[`0002T0
20L000H7F0000i2`[`0901@700_61jl7[`N_bZo6[`080003[i0000T000^YT:V@ZI2YT:V@Z@0H0008
[`N_1jl7/0L600X72P021`L000>_0000:P081`001PMI0003U[F`00P02@L200X70P002<X7bPO61jl7
2P000jT00007000;ZJnY[jV_ZJnY[jT01P05[`d01`L800L73@000`L000040003[`0002/020L000H7
F@00196F/:l800H7100:1`@000H7[`N_1jl:0008[jV_ZJnY[jT2000@[jV_ZJnY[jV_ZJl0[jV_ZPN_
30001:l7/0L;00H730001`Ne1k@7[kD0<0081`001PMJ00:_00>e/0001`041`H020L600051lX7bPL0
2@002lJ_aZo6[lJ_aZo600`000^_aZo6[lN_ajl7[`0=00031`0000`00`L=00D700>e1jl0<0081`00
1PMK00:_00>e[`001P031`T01PL80003[`L000T000>_aZl02:l<00:_00W6[lJ_aZo7[`L030000lL0
000>0087300020Ne1jl7[kB_<@081`001PMM0003[`L000H000<700002P041aD000/7[`N_1jl7[`N_
1`0<000;[`N_1jl7aPO61lL02P031a0000<700002P031`04]@Nd]C<020L000H7GP000jne000300@7
30021``000?61`001`002jl7[`N_1jl7[`N_00`000/7[`N_1jl7[`N_1`0700031lL700@73@031`/0
00D7]0N_1`02[c@020L000H7HP000`Ne1`051`/000<700002@081`<02`L<01X72@071`P010L00kBe
000f00P700061f<000@7]@Ne10L900@72@071`04bPO61`8000W61lH7aPO61lH0300020O61lH7a`O7
100>1`L02PL500051kD7[`L00Zlh00P700061f@020L700H71`0<1`L010L<00872P0?1`@03@L200@7
00>e1k@0>P081`001PMT0005[`Ne1kD010L500P71P0<1`/000P7a`O61lL7a`@7300?1`<03PL20005
1jl7[`L00Zlj00P700061fD000>`1kD01PL300/70`0<1``030L=00h700<01`L03@L50003]@N_03`0
20L000H7I`001[F`1kD7]A4700<01`L02PL<00`73@0<1`<03PL70003UP0003/020L000H7J0000`Ne
1`031`803PL200/7300<1`d02@L700/71`000jl0000m00P700061fT000Be[`Ne1@0<1`D020L<00`7
3P061`T02@L80003[`0003h020L000H7J@000jl0000700X720051`d030L>00873P061`P000>F0000
@0081`001PM[0003/00000L01`L=00031`0000/030L=00031`0000l00PL90003[`00048020L000H7
K@000jl0000700D73P031`X030L800L76@000iH0001400P700061fl00ZlH00h7300?1``00PL01;D7
[`L200:FBP081`001PMa0003[kD000D010L;00h73@0@1`P01PL00kD7[`1>00P700061g<00Zl30005
]@Ne1kD00`L800h73@0@1`D000Ze1kD7[`N_1jneD0081`001PMh00031kD700P71@0>1`h03`L200@7
00Je1kD7[kEC00P700061gX000V_]K07]@Ne1kD04`L>00`710001[D7[`N_]@:_EP081`001PMm0003
/0Ne00H71@0;1`h01`L900041kD7[eX020L000H7P0001Zne[`Ne1`d00`L=00031`0000l000>_0000
FP081`001PNE00l73@000iJ_001N00P700061hL00Zl<000?1k07]@N`1k07/0N`1k0700P000>F0000
H`081`001PN<0005[iJ`[kD00`004@N_1k07]@N`1jl7[kF_]@2`00:_J`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[BoxData[
    \(translate2D[gr_, vec_] := \n\t\ \ 
      gr\  /. \ {
          h : \((Line | Polygon)\)[pnts_] :> Line[\(vec + #\ &\)\ /@\ pnts], 
          \n\t\t\t\ \ \ \ \ \ \ \ \ \ \ \ 
          Rectangle[p1_, p2_] :> \ Rectangle[vec + p1, vec + p2]}\)], "Input",\

  CellLabel->"In[29]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(tex2d = 
      Show[Graphics[
          \({#, translate2D[Reverse[#], {0, 1}]}\ &\)\ @
            \((\(Join[#, Reverse[translate2D[#, {1, 0}]]]\ &
                  \)\ @{{RGBColor[1, 0, 0], Rectangle[{0, 1/3}, {1, 2/3}]}, \n
                  \t\t{RGBColor[1, 1, 0], Rectangle[{1/3, 0}, {2/3, 1}]}})
              \)], AspectRatio -> Automatic, ImageSize -> {136, 136}, 
        Frame -> None, Background -> RGBColor[0, 0, 1]]\)], "Input",
  CellLabel->"In[30]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
%%ImageSize: 136 136 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Background color
0 0 1 r
MFill
% Scaling calculations
0.0238095 0.47619 0.0238095 0.47619 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
1 0 0 r
.02381 .18254 m
.02381 .34127 L
.5 .34127 L
.5 .18254 L
F
1 1 0 r
.18254 .02381 m
.18254 .5 L
.34127 .5 L
.34127 .02381 L
F
.65873 .02381 m
.65873 .5 L
.81746 .5 L
.81746 .02381 L
F
1 0 0 r
.5 .18254 m
.5 .34127 L
.97619 .34127 L
.97619 .18254 L
F
.5 .65873 m
.5 .81746 L
.97619 .81746 L
.97619 .65873 L
F
1 1 0 r
.65873 .5 m
.65873 .97619 L
.81746 .97619 L
.81746 .5 L
F
.18254 .5 m
.18254 .97619 L
.34127 .97619 L
.34127 .5 L
F
1 0 0 r
.02381 .65873 m
.02381 .81746 L
.5 .81746 L
.5 .65873 L
F
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[30]:=",
  ImageSize->{136, 136},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000280000R0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoIXP700271047002710470027104700271047
000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P4
5P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX4
5P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT4
0@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`00
60@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000311D1
5P=E0@@40@L000<45@4F0eD110@11`000`@E0AH3E@441047000311D15P=E0@@40@L000<45@4F0eD1
10@11`000`@E0AH3E@441047000311D15P=E0@@40@L000<45@4F0eD110@11`000`@E0AH3E@441047
000311D15P=E0@@40@L000<45@4F0eD110@11`000`@E0AH3E@441047000311D15P=E0@@40@L000<4
5@4F0eD110@11`000`@E0AH3E@441047000311D15P=E0@@40@L000<45@4F0eD110@11`000`@E0AH3
E@441047000311D15P=E0@@40@L000<45@4F0eD110@11`000`@E0AH3E@441047000311D15P=E0@@4
0@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`00
60@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3
:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H3
6@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047
000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P4
5P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX4
5P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT4
0@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`00
60@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3
:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L000<4E@4F0aD1
10@11`000`AE0AH35@441047000315D15P<E0@@40@L000<4E@4F0aD110@11`000`AE0AH35@441047
000315D15P<E0@@40@L000<4E@4F0aD110@11`000`AE0AH35@441047000315D15P<E0@@40@L000<4
E@4F0aD110@11`000`AE0AH35@441047000315D15P<E0@@40@L000<4E@4F0aD110@11`000`AE0AH3
5@441047000315D15P<E0@@40@L000<4E@4F0aD110@11`000`AE0AH35@441047000315D15P<E0@@4
0@L000<4E@4F0aD110@11`000`AE0AH35@441047000315D15P<E0@@40@L000<4E@4F0aD110@11`00
60@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3
:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H3
6@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047
000H11H3:P@F0aT40@L001P45P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P4
5P<Z11H36@@11`0060@F0bX45P<I1047000H11H3:P@F0aT40@L001P45P<Z11H36@@11`00Q`@11`00
Q`@11`00Q`@11`00\
\>"],
  ImageRangeCache->{{{0, 135}, {135, 0}} -> {-0.0500105, -0.0500105, 
  0.0155557, 0.0155557}},
  CellTags->"MathLinkMode"],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output",
  CellLabel->"Out[30]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " adds a small border of pixels around the graphics. This behavior is not \
very useful if the graphics will be used as texture. The option ",
  StyleBox["MVClipBorderColor",
    FontFamily->"Courier"],
  " for ",
  StyleBox["MVPutTexture[]",
    FontFamily->"Courier"],
  " and ",
  StyleBox["MVShow3D[]",
    FontFamily->"Courier"],
  " forces MathGL3d to remove this border. The default setting is ",
  StyleBox["True",
    FontFamily->"Courier"],
  " to remove the border, otherwise the border is preserved. The width of the \
border created by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " depends on the image size and a larger image size than the size of the \
texture is recommended. OpenGL allows only powers of 2 for the width and the \
height of the texture. MathGL3d will scale ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " images to the next lower power of 2 if the width or the height of the \
image are not powers of two. "
}], "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; 
    MVShow3D[blankknot, MVTexture -> tex2d, 
      MVTextureMapType -> MVMeshUVMapping, MVScaleTexture -> {40, 6}]\)], 
  "Input",
  CellLabel->"In[31]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVMesh3D[5] \[DoubleLeftArrow] MVTexture2D[2], 
      MVMeshUVMapping, {128, 128, MVRGBTexture}}\)], "Output",
  CellLabel->"Out[31]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVPasteGraphics[]\)], "Input",
  CellLabel->"In[24]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000005F00006000005F00006000005F0000607000007F20007F60007F6000\
705F007F6000705F007060006F5F006F60006F5F00000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000605F00605F00605F0020204000005F0000606F00007000007000007F0000
7F00007F00002F004000006000005F0000600000600000600000600000608060008F6F00806000\
8060008060008F6F008060008060008060008060008060008000007F00007F0000
70000000006000005F00006000005F00006000005F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006F00007F0000\
7F00007F40007F5F0000006000006000006000006000006F00006000006F000060
00006F00006000006F00006000006F00006F00006F00006F0000702F204F9070008F6F00907000\
8F6F008F70008F6F008F70008F6F008F6F008F6F00302F4F00006000006F000060
00006F0000600000600000600000602F00408F00008000008F000080000080000000005F000060\
00005F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000607000007F6000705F007F60007F6000\
8060007F60002F204F00006000006F00006000006F00006F00006F00006F00006F
00006F00007000006F00007000006F000070000070000070000070907000907000907000907000\
9070008F70009070008F70009070008F6F0090700000006F00007000006F00006F
00006F00006F00006F00006F00006000006F00006000006F00006000006F0000600000602F0040\
8F00007F0000000060705F00000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000005F7000000000607F60007F60007F60007F6000806F00806F00806F00\
806F008F6F0000007000007000007000007000007000007000007000007F000070
00007F00007000007F00007F00007F00007F00007F00007F302F5F9F7F009F7F009F7F009F7F00\
907F009F7F009F7F009F7F00907F009F7F00302F5000007F00007000007F000070
00007000007000007000007000007000007000007000006F00007000006F00007000006F00006F\
00006000006F2F20408F4F008F00008000007F6000706000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000060
6F000000006000006000006000006000006F7F6000806F00806000806F00806F008F7000806F00\
8F70002F204F00007F00007000007F00007000007F00007F00007F2F0050900000
9000009F00009000009F00009000009F00009F00009F00009F00009F00009F00009F00009F0000\
9F00009F0000A000009F00009F00009F00005F1F3F00007F00007F00007F00007F
00007F00007F00007F00007F00007000007F00007000007F00007000007000007000007000006F\
00007000006F302F4F8F6F009070008F6F00906F008F40008F00007F00007F6000
6F5F00000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000005F6F0000200040000060000060
00006F00006000006F00006F0000702F204F8F6F00806F008F70008F70009070008F7000907F00\
9070009000009000009000009000009F00009000009F00009000009F0000900000
9F00009F00009F00009F0000A000009F0000A000009F0000A000009F0000A000009F0000A00000\
A00000A000009F0000A00000A00000A00000A00000A00000A00000A00000A00000
A00000A00000A00000A00000A00000A000003F005F00007F00008000007F00007F00007F00007F\
0000700000709F7F009F7F009070009F7F009070009070009070009070008F6F00
904F008F000080000000005F000060000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F000000005F00006000006000006F00006000006F
00006F00007000006F0000700000708F70008F6F008F70008F7000907F008F7000907F00907000\
907F009000009000009000009F00009000009F00009000009F00009000009F0000
9F00009F00009F00009F00009F0000A000009F0000A000009F0000A000009F0000A00000A00000\
A00000A00000A00000A00000A00000A00000A00000A00000A00000A00000A00000
A00000A00000A00000A00000A00000A00000A00000A00000A00000AF0000A00000A00000300050\
000080302F50A080009F7F009F7F009F7F009F7F009070009F7F00907000907000
9070009070000000600000608F00008F00007F000000006000005F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F200000006000006000006000006F00006F00006F00006F000070000070
0000700000703000508F00008F2F008F7000907F00907F00907F00907F009F7F00907F009F8000\
907F009F00009F00009F00009F00009F00009F00009F00009F0000A000009F0000
A000009F00003F006030006000009000008F00009000008FA08F00A08F00AF8F00A08F00AF8F00\
A05F00AF6000A00000AF0000A00000AF0000A00000AF0000A00000AF0000A00000
AF0000A00000AF0000A00000AF0000A00000AF0000AF0000AF0000A00000AF0000AF0000AF0000\
A00000AF3000A08000AF8F00A08000A08000A08000A080009F7F00A080009F7F00
9F7F009F700000007000006F00006F00006000006F0000609000008F000080000000005F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000605F00\
6F5F0000006000006F00006F00006F00006F0000700000700000700000708F0000
8000008F00008F00008F00008F2000907F00907F00907F00907F00907F00907F009F8000907F00\
9F800030005F00008F00008F00008F00008F00008F00008F00008F00008F000090
00008F00009000008F00009000008F0000900000903F3060A08F00AF8F00A08F00AF8F00A08F00\
AF8F00A08F00AF8F00A08F00AF8F00302F6000009000008F00009000008F000090
00008F3F0060AF0000AF0000A00000AF0000AF0000AF0000AF0000AF0000AF0000AF0000A00000\
AF0000A02F00AF8F00A08000AF8F00A08000AF8F00A08000A08000A08000A08000
9F7F00302F5000007000007000006F00006F00006F00006F000060000060000060300040800000\
000060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006F60006F5F00706000\
20204F00007000006F00007000007000007F8000008F00008F00008F00008F0000
9000008F00009000008F0000902F00907F00908000907F009F80009F80009F80009F80009F8F00\
302F6000009000008F000090000090000090000090000090000090000090000090
00009F00009000009F00009000009F00009000009F302F60AF9000AF9000AF9000AF9000AF9000\
AF9000AF9000AF9000AF9000AF90003F306F00009000009F000090000090000090
00009000009000009000009000009000009000009000009000009000008F3F0060AF0000B00000\
AF0000AF9000AF8F00AF9000AF8F00AF8F00AF8F00AF8F00AF8F00AF8F00A08F00
AF8F00302F5000007F00007F00007F00007000007000007000007000006F00006F00006F00006F\
0000608F000000005F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000605F006F60006F5F00706000706000\
0000700000708000008000008000008000008F00008000008F00008F00008F0000
8F00009000008F000030005F907F009F8000907F009F80009F80009F80009F80009F80009F8000\
302F6000009000009000009000009000009000009000009000009F00009000009F
00009000009F00009000009F00009F00009F00009FAF9000AF9000AF9000AF9000AF9000AF9000\
AF9000AF9000AF9000AF9000AF900000009000009F00009000009F00009000009F
00009000009F00009000009000009000009000009000009000009000009000008F00009000008F\
000090AF8F00AF9000AF8F00AF9000AF8F00AF8F00AF8F00AF8F00AF8F00AF8F00
AF8F00AF3000A00000AF000030005000008000007F00007F00007000007F00007000007000006F\
00006F0000600000608F0000000060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F00006F5F007060007060007F6F007F60004F1F3F\
8000008F00008000008F00008000008F00008F00008F00008F000090000030005F
00008F00008F00008F00008F9F80009080009F8F009F80009F8F009F8F00A08F009F8F00A08F00\
302F6000009F00009F00009F00009F00009F00009F00009F00009F00009F00009F
00009F00009F0000A000009F0000A000009F0000A0AF9F00B09F00AF9000B09F00AF9F00B09F00\
AF9000B09F00B09F00B09F00AF90000000A000009F00009F00009F0000A000009F
00009F00009F00009F00009F00009F00009F00009F00009F00009F00009000009F000090000090\
3F3060B09F00B09000B09000AF9000B09000AF9000B09000AF8F00B09000AF8F00
B09000AF0000B00000AF0000AF0000AF0000AF0000AF00003F005F00007F00007F00007F000070\
00007000007000006F00006F0000609000002F0040000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006000000000606F5F007060007060007F60007F4000800000800000\
8000008000008F00008000008F00008F00008F000000008000008F00008F00008F
00008F00008F00008F0000909080009F8F009F8000A08F009F8F00A08F009F8F00A09000A08F00\
00009F00009F00009F00009F00009F00009F0000A000009F0000A000009F0000A0
00009F0000A00000A00000A00000A00000A00000A0B0A000B09F00B09F00B09F00B09F00B09F00\
B09F00B09F00B09F00B09F003F30700000A00000A000009F0000A000009F0000A0
00009F0000A000009F0000A000009F00009F00009F00009F00009F00009F00009F00009F000090\
3F306FB09000B09F00AF9000B09000B09000B09000AF9000B09000AF9000B09000
AF8F00B00000AF0000B00000AF0000AF0000AF0000AF0000AF0000AF0000AF00003F005F00007F\
00007F00007000007000006F00006F0000600000602F2040806F006F5F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000020004000006000006F00006F7060007060008000007F00008000008000008F0000\
8F00008F00008F000000008F00008000008F00008F00009000008F00009000008F
000090000090000090000090A08F009F8F00A08F00A08F00A09000A09000AF9000A09000AF9F00\
0000A00000A00000A00000A00000A00000A00000A00000AF0000A00000AF0000A0
0000AF0000AF0000AF0000AF0000AF0000AF0000AFB0A000BFA000B0A000BFA000B0A000BFA000\
B0A000BFA000B0A000BFA0003F30700000AF0000A00000AF0000A00000AF0000A0
0000AF0000A00000AF0000A00000AF0000A00000A00000A00000A00000A00000A000009F0000A0\
B09F00BF9F00B09F00BF9F00B09F00B09F00B09000BF9F00B09000B09F00B09000
3F306000008F3F0060B00000B00000B00000B00000AF0000B00000AF0000B00000AF0000B00000\
AF000000008000007F00007F00007000007000006F302F4F9070009070008F6F00
7F6000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
1F003F00006000006000006F00006F7F60007F00007F00007F0000800000800000800000800000\
00008F00008000008F00008F00008F00008F00009000008F000090000090000090
00009000009F00009000009F9F8F00A09000A08F00A09000A09000AF9F00A09000AF9F00AF9000\
0000AF0000A00000AF0000A00000AF0000A00000AF0000AF0000AF0000AF3F0070
3F00703F007FB00000BF0000B00000BF0000B00000BFAF00BFA000BFAF00BFA000BFAF00BFA000\
BFAF00BFA000BFAF00BFA000403F7F0000AF0000AF0000AF0000AF0000AF0000AF
0000AF0000AF0000AF0000AF0000AF0000AF0000A00000AF0000A00000AF0000A00000A00000A0\
BFA000B09F00BF9F00B09F00BF9F00B09F00BF9F00B09F00BF9F00B09000B09F00
00009000009000008F00009000008F0000903F005FB00000B00000B00000AF0000B00000AF0000\
B00000AF0000AF000000007F00007F0000700000709F7F009F7F00907000907000
9070009070007F6000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000200040000060
00006F00006000006F0000707F00007F00008000008000008000008000008F50008F700000008F\
00008F00009000008F00009000008F00009000009000009000009000009F00009F
00009F00009F0000A00000A0AF9000A09000AF9F00AF9F00AF9F00AF9F00B0A000AF6F00B06F00\
AF0000B00000B00000B00000B00000BF0000B00000BF0000B00000BF0000BF0000
BF0000BF0000BF0000BF0000BF0000BF0000C00000BFAF00C0B000BFAF00C0B000BFAF00C0AF00\
BFAF00C0B000BFAF00C0AF00BF3F00C00000C00000C00000BF0000C00000C00000
40007F3F007F40007F0000B00000B00000AF0000B00000B00000B00000AF0000AF0000AF403F70\
BFA000C0AF00BFA000BFA000BFA000BFA000BF9F00BFA000BF9F00BFA000BF9F00
00009F00009000009F00009000009F00009000009000009000009000008FB00000B00000B00000\
AF0000B00000AF0000B00000AF00003F005F00007FA08F00A08000A080009F7F00
9F7F009070009F7000907000806F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000006000006000006F
00006F0000707000007F00007F00008000007F0000800000804F008F70008F70008F7F0000008F\
00009000008F00009000009000009000009000009F00009000009F00009F0000A0
00009F0000A00000A00000AFA09000AF9F00AF6F00B00000AF0000B00000B00000B00000B00000\
B00000B00000BF0000B00000BF0000B00000BF0000B00000BF0000BF0000BF0000
BF0000BF0000BF0000C00000BF0000C00000C03F00C0B000BFAF00C0B000C0B000C0B000C0AF00\
C0B000C0AF00C0B000C0AF00C03F00C00000C00000C00000C00000C00000C00000
C00000C00000C00000C00000C00000C00000C00000C00000C0000040007F0000AF0000B03F3070\
C0AF00BFA000C0AF00BFA000BFA000BFA000BFA000BFA000BFA000BF9F0040306F
00009F00009F00009F00009F00009000009F00009000009F00009000009000008F000090B00000\
B00000B00000B00000AF0000B00000AF0000AF0000AF0000AF8F00A08000A08000
9F7F009F7F009F7F009F7F00907000907000802000000060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000006000006F00006F00006F
7F00007F00007F00007F00008000008000008050008070008F7F008F7F00907F008F7F00000090\
00009000009F00009000009F00009F00009F00009F0000A00000A00000A00000A0
0000AF0000AF3F0070AF0000B00000AF0000B00000B00000B00000B00000BF0000B00000BF0000\
BF0000BF0000BF0000BF0000BF0000BF0000BF0000C00000BF0000C00000BF0000
C00000C00000C00000C00000CF0000C00000CF3F00C0B000CFBF00C0B000CFBF00C0B000CFBF00\
C0B000CFBF00C0B000CFBF00C00000CF0000C00000CF0000CF0000CF0000C00000
CF0000CF0000CF0000C00000CF0000CF0000CF0000C00000CF0000C00000CF0000C00000CF0000\
C07000C0B000C0AF00C0B000C0AF00C0AF00C0AF00C0AF00C0A000C0AF003F306F
0000A00000A00000A000009F0000A000009F00009F00009F00009F00009000009F000090000090\
0000903F0060B00000B00000B00000B00000AF0000B00000AF0000AF0000AF8F00
AF8F00A08000A080009F7F009F80009F7F00302F4F000060900000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000006000006F0000707000007F0000
7000007F00007F00008000002F00508070008070008F7F008F7F008F7F008F7F00908000000090\
00009F00009F00009F00009F0000A000009F0000A00000A00000AF300070AF0000
AF0000AF0000AF0000B00000AF0000B00000B00000B00000B00000BF0000B00000BF0000BF0000\
BF0000BF0000BF0000BF0000C00000BF0000C00000C00000C00000C00000C00000
C00000CF0000C00000CF0000C0000040008F6F3F5FCFBF00CFBF00CFBF00C0BF00CFBF00CFBF00\
CFBF00CFBF00CFBF00CFBF00CF0000CF0000CF0000CF0000CF0000CF0000CF0000
CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000C00000CF0000C00000\
CF0000C00000CF0000C00000CF7F00C0AF00C0AF00C0AF00C0AF00C0AF000000AF
0000A00000AF0000A00000A00000A00000A000009F00009F00009F00009F00009000009F000090\
00009000008F00009000008FB00000B00000B00000AF0000B00000AF0000AF0000
AF0000AF6000A08000A08000A08000A0800030204F00006F00006000006000005F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F000000006000006F7000007000007F00007F0000
7F00007F00002F005F0000802F2F5F8070008F7F008F7F009080008F8000908F0090800000009F\
00009F0000A00000A00000A00000A03F0070A00000AF0000AF0000B00000AF0000
B00000AF0000B00000B00000BF0000B00000BF0000BF0000BF0000BF0000C00000BF0000C00000\
C00000C0000040008040008F4000800000CF0000CF0000CF0000CF0000CF0000CF
0000CF0000CF0000D00000CF0000D00000CF4F4090CFC000D0C000CFC000D0C000CFC000D0C000\
CFC000D0C000CFC000D0C0000000CF0000D040008F40008FCF0000D00000D00000
D00000CF0000D00000D00000D00000CF0000D00000D00000D00000CF0000D00000CF0000D00000\
CF0000D00000CF0000CF0000CF0000CF0000CF0000CF0000C07F00CFB0000000B0
0000B00000AF0000B00000AF0000AF0000A00000AF0000A00000A000009F0000A000009F00009F\
00009000009F00009000009000008F000090B00000B00000B00000B00000B00000
B00000AF0000B00000AF0000AF8F00A080003F305000007000007000006F00006F300040000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F5F00605F0000006F7000007000007F00007F00007F0000
00007F00008000008000008F2F205F8F7F008F7F009080008F8000908000908F009F8F0000009F\
0000A030006FA00000A00000AF0000A00000AF0000AF0000AF0000AF0000B00000
B00000B00000B00000BF0000B00000BF0000BF0000C07F00BFB000C0B000C0B000C0BF00C0B000\
0000CF0000CF0000CF0000CF0000CF0000CF0000D00000CF0000D00000CF0000D0
0000CF0000D00000D00000D00000D00000D0403F8FD0CF00CFC000D0C000D0C000D0CF00CFC000\
D0C000D0C000D0C000403F8F0000D00000D00000D00000D00000D00000D00000D0
0000CF0000D00000CF0000D00000CF0000D040008FD00000D00000D00000D00000D00000CF0000\
D00000CF0000D00000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000
C0000040007F0000B00000B00000AF0000AF0000AF0000AF0000A00000A00000A00000A000009F\
00009F00009F00009F000090000090000090BF9F00B09000B06F00B00000B00000
AF0000B00000AF0000B00000AF0000AF0000302F5000007F00007000007000006F00006F000060\
000060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F5F006F600000006F7000007000007F00007F0000000080000080
00008F00008F00008F00008F302F608F7F009080009080009F8F009F8F009F90009F9000A03000\
A00000AF0000A00000AF0000AF0000B00000AF0000B00000B00000B00000B00000
BF0000BF0000BF00003F0080C0B000BFB000C0BF00C0B000CFBF00C0BF00CFBF00C0BF00CFC000\
403F8F0000D00000CF0000D00000D00000D00000D00000D00000D00000D00000D0
0000DF0000D00000DF0000D00000DF0000DF4F4090D0CF00DFCF00D0CF00DFCF00D0CF00DFCF00\
D0CF00DFCF00D0CF004F40900000DF0000DF0000D00000DF0000D00000DF0000D0
0000DF0000D00000DF0000D00000DF0000D00000DF0000D00000D00000D00000D0D00000DF0000\
D00000D00000D00000D00000D00000D00000D00000D00000D00000D00000CF0000
D00000CF0000D00000CF0000D0000040007F0000B00000AF0000B00000AF0000AF0000A00000A0\
0000A00000A000009F00009F00009F403F6FBF9F00BFA000BF9F00BF9F00B09F00
BF0000B00000B00000B00000B00000AF0000B00000AF00003F005F00007F00007000006F00006F\
000060300040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000005F5F00605F003F1F307000007000007F000000007F00008000008000008F
00008000008F00008F0000902F2F609080009080009F8F009F8F00A090009F9000A09F00A03000\
AF0000A00000AF0000AF0000B00000AF0000B00000B00000BF0000B000003F0080
0000C00000C00000C00000CFC0B000C0BF00C0BF00C0BF00C0BF00CFC000C0BF00CFC000CFC000\
4040900000D00000D00000D00000DF0000D00000DF0000D00000DF0000D00000DF
0000DF0000DF0000DF0000DF0000DF0000DF404090DFD000D0CF00DFD000D0CF00DFD000D0CF00\
DFD000D0CF00DFD0004040900000DF0000DF0000DF0000DF0000DF0000DF0000DF
0000DF0000DF0000DF0000DF0000DF0000DF0000D00000DF0000D00000DF0000D0DFCF00D0CF00\
DFCF00D0C000DF0000D00000DF0000D00000D00000D00000D00000D00000D00000
D00000D00000CF0000D00000CF0000D00000CF0000CF000040007F0000B00000AF0000AF0000A0\
0000AF0000A00000A000009F00009FBFA000BFA000BF9F00BFA000B09F00BF9F00
B09F00BF9F00B00000B00000B00000B00000AF0000B00000AF0000AF000000007F000070000070\
00006F00006000006000005F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000600000\
605F006F60006F60007000007000007F000000008000008000008000008F00008F
00009000009000009F00009000009F908F009F90009F9000A09000A09000AF9F00A09F00AFA000\
AF0000B00000B00000B00000B00000BF00003F00800000C00000C00000CF0000C0
0000CF0000CF0000D00000CF404090C0BF00CFC000CFC000CFC000CFC000D0CF00CFC000D0CF00\
4040900000DF0000DF0000DF0000DF0000E00000DF0000E00000DF0000E00000DF
0000E00000DF0000E00000E00000E00000E04F4F9FDFD000DFD000DFD000E0D000DFD000DFD000\
DFD000E0D000DFD0004F4F9F0000E00000E00000E00000E00000E00000E00000E0
0000E00000DF0000E00000DF0000E00000DF0000E00000DF0000DF0000DF4F4F9FDFCF00E0D000\
DFCF00E0D000DFCF00DFCF00DFCF00DFCF00DF8F00DF0000DF0000DF0000DF0000
DF0000D00000DF0000D00000DF0000D00000DF0000D00000D00000CF0000D000000000B00000B0\
0000AF0000AF0000A00000AF0000A0C0AF00BFA000C0A000BFA000C0A000BFA000
BFA000BF9F00BFA000B09F00BF6F00B00000BF0000B00000B00000AF0000B00000AF000000007F\
00007000007000006F00006F2F0040000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000605F5F00\
6F5F006F5F007060007000007F000000007F00008000008000008F00008F000090
00009000009F00009F00009F00009F9F90009F9000A09F00A09000A09F00A09F00AFA000AFA000\
B00000B000003F00800000BF0000C00000C00000CF0000C00000CF0000CF0000CF
0000CF0000D00000CF0000D00000D0CFC000CFC000CFCF00CFC000D0CF00CFCF00D0CF00CFCF00\
D0CF000000DF0000E00000DF0000E00000DF0000E00000DF0000E00000E00000E0
0000E00000E00000E00000E00000E00000E040409FE0DF00DFD000E0DF00DFD000E0DF00DFD000\
E0DF00DFD000E0DF0040409F0000E00000E00000EF0000E00000E00000E00000E0
0000E00000E00000E00000E00000E00000E00000DF0000E00000DF0000E04F4090E0D000DFD000\
E0D000DFD000E0D000DFCF00E0D000DFCF00E0D000DFCF004F40900000CF4F0090
DF0000DF0000DF0000DF0000D00000DF0000D00000DF0000D00000D00000D00000D00000CF0000\
40007F0000AF0000AF0000AF403F70C0AF00C0AF00BFA000C0A000BFA000C0A000
BF9F00BFA000BF9F00BFA000B09F0000009000008FBF0000B00000B00000AF0000B00000AF0000\
00007F00007000007000006F00006F000060000060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000060201F406F6000\
6F60007060007060002F005F00008000008F00008F000090000090000090000090
00009F00009F0000A030006FA00000A09000A09F00A09F00AFA000AFA000B0A000AFA000B0AF00\
0000C00000C00000C00000CF0000CF0000CF0000CF0000D00000CF0000D00000D0
0000D00000D00000DF0000D00000DFCFCF00D0CF00D0CF00D0D000D0CF00D0D000D0CF00DFD000\
D0D0000000E00000E00000EF0000E00000EF0000E00000EF0000E00000EF0000E0
0000EF0000EF0000EF0000EF0000EF0000EF4F4FA0E0DF00E0E000E0DF00E0E000E0DF00E0E000\
E0DF00E0E000E0DF004F4FA00000EF0000EF0000EF0000EF0000EF0000EF0000E0
0000EF0000EF0000EF0000E00000EF0000E00000EF0000E00000EF0000E04F4F9FE0D000E0DF00\
E0D000E0DF00E0D000E0DF00E0D000E0D000E0D000E0D0000000DF0000DF0000D0
0000DF0000D00000D04F008FE00000DF0000DF0000DF0000DF0000D00000DF0000D00000D00000\
D00000D00000CF00000000B0CFB000CFB000C0AF00CFB000C0AF00C0AF00C0A000
C0AF00BFA000C0A000BFA000403F6F00009000009000008F000090B00000B00000B00000B00000\
AF0000AF000000007F00007F00007000006F000060300040000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000600000606F5F006F5F00\
7060007060007F6F0000007F00008F00008000008F00009000009F00009000009F
00009F3000709F00009F00009F0000A03000A09F00AFA000AFA000B0AF00AFA000B0AF00B0AF00\
0000CF0000C00000CF0000CF0000CF0000CF0000D00000D00000D00000D00000DF
0000D00000DF0000DF0000DF0000DFD0CF00D0CF00D0D000D0CF00DFD000D0D000DFD000D0D000\
DFDF000000E00000EF0000E00000EF0000EF0000EF0000EF0000EF0000EF0000EF
0000EF0000F00000EF0000F00000EF0000F07F2F6FE09F00E0DF00E0E000E0DF00E0E000E0DF00\
EFE000E0DF00E0E0004F4FA00000F00000EF0000EF0000EF0000EF0000EF0000EF
0000EF0000EF0000EF0000EF0000EF0000EF0000E00000EF0000E00000EFE0DF00E0DF00E0DF00\
EFDF00E0DF00E0DF00E0DF00E0DF00E0D000E0DF00E0D0000000DF0000DF0000DF
0000DF0000DF0000D00000DF0000D00000D040008FDF0000DF0000DF0000D00000DF0000D00000\
D00000D00000D00000CF0000D0BF00CFB000CFB000CFB000CFB000C0AF00C0AF00
C0AF00C0AF00BFA000C0A0003F306000009F00009000009000008F0000903F005FB00000B00000\
B00000AF0000AF000000007F00007F00006F00006F000060000060706000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000600000606F20006F6000706000\
7060007F6F007F6F0000008F00008F00009000009000009F00009F0000A030006F
9F00009F0000A00000A00000AF0000A03000AFA000AFA000B0AF00B0AF00BFB000BFB000BFBF00\
3F3F8F0000D00000CF0000D00000D00000DF0000D00000DF0000DF0000DF0000DF
0000E00000DF0000E00000E00000E040409FDFD000DFD000DFDF00DFD000DFDF00DFDF00E0DF00\
DFDF004F4FA00000EF4F00A04F00A04F00A0E00000E00000E00000E00000E00000
E00000E00000EF0000E00000EF0000E00000EF0000EF0000EF0000EF0000EF0000EF0000EF0000\
EF0000EF0000EF0000EF00004F00A05000AF0000F00000F00000F00000F00000F0
0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000EF0000EFEFE000EFE000EFE000\
EFDF00EFE000EFDF00EFE000E0DF00EFE000EFDF004F4F9F0000DF0000E00000DF
0000E00000DF0000E00000DF0000DF0000DF0000DF0000D00000D0DF0000E00000DF0000DF0000\
DF0000DF0000D00000D04000D0BF00D0C000CFBF00D0BF00CFB000CFBF00CFB000
CFB000C0AF00CFB000C0AF000000A000009F00009F00009000009F00009000009000008FBF0000\
B00000B00000B00000B00000AF000000007F00007000007000006F302F4F907000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000606F00006F1F00706000706000\
706F007F6F007F700000008F00009000009000009F00009F9000009000009F0000
9F0000A00000A00000AF0000A00000AF0000AFA000B0AF00B0AF00BFB000BFB000BFBF00BFB000\
C0BF000000D00000D00000D00000DF0000D00000DF0000DF0000DF0000DF0000E0
0000DF0000E00000E00000EF0000E04F40A0DFD000DFDF00DFDF00DFDF00DFDF00E0E000DFDF00\
E0E000E04F00E00000E00000E00000E00000E00000E00000E00000E00000EF0000
E00000EF0000E00000EF0000E00000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000\
EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000
EF0000EF0000EF0000EF00004F00A00000F00000EF0000F00000EF0000EFEFE000EFE000EFE000\
EFE000EFE000EFE000EFDF00EFE000EFDF00EFE0004F409F0000E00000E00000E0
0000DF0000E00000DF0000E00000DF0000DF0000DF0000DF0000D00000D00000D04F0090DF0000\
DF0000DF0000DF0000D0C000DFC000D0C000D0C000CFBF00D0BF00CFBF00CFBF00
CFB000CFB000C0B000403F700000A00000A000009F00009F00009000009000009000009000008F\
3F0060B00000B00000AF0000B00000AF000000007F000070000070302F4F9F7F00
8F70007F6F00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000606F00006F00007020007060007F6F00\
7F6F008070007F700000009000009000009F00009F9F00009F0000A000009F0000
A00000A00000AF0000AF00000000C00000C0B0B000B0AF00BFB000BFB000C0BF00BFBF00C0C000\
C0C0000000DF0000DF0000DF0000DF0000E00000DF0000E00000E00000E00000E0
0000EF0000EF0000EF0000EF4F00A0DF0000E0DF00DFDF00E0E000E0E000E0E000E0E000EFEF00\
E0E000EF4F00E00000EF0000E00000EF0000E00000EF0000EF0000EF0000EF0000
EF0000EF0000EF0000EF0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000\
EF0000F00000EF0000F00000EF0000F00000EF0000F00000F00000F00000EF0000
F00000F00000F00000F00000F00000F00000F00000EF0000F00000804F70F0EF00EFEF00F0EF00\
EFEF00F0EF00EFE000F0EF00EFE000F0EF00EFE0000000EF0000E00000EF0000E0
0000EF0000E00000E00000E00000E00000DF0000E00000DF0000DF0000DF0000DF0000D00000D0\
0000D0E00000DF4000E0D000DFCF00DFCF00D0C000DFCF00D0C000D0C000D0BF00
D0C000CFBF00D0BF000000AF0000AF0000A00000A000009F0000A000009F00009F000090000090\
00008F00008FB00000B00000B00000B00000AF0000B0000000007F3F30509F8000
9F8000907000907000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000606F00006F0000700000202050706F00706F00\
7F70007F7000807F002F2F6000009F2F006F9F00009F0000A000009F0000A00000
A00000AF00000000BF0000C00000C00000C0B0B000BFB000BFB000C0BF00BFBF00C0C000C0C000\
CFC000403F900000DF0000DF0000E00000E00000E00000E00000EF0000E00000EF
40009FDF0000DF0000DF0000DF0000E00000DFDF00E0E000E0E000E0E000E0E000E0E000E0E000\
EFEF00E04F00EF0000E00000EF0000E00000EF0000EF0000EF0000EF0000EF0000
EF0000EF0000EF0000EF0000EF0000F00000EF0000EF0000EF0000F00000EF0000F00000EF0000\
F00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000
F00000F00000F00000F00000F00000F00000F00000F00000F00000F05000EFEF00F0EF00EFEF00\
F0EF00EFEF00F0EF00EFEF00F0EF00EFE000F0EF000000EF0000EF0000E00000EF
0000E00000EF0000E00000E00000E00000E00000DF0000E00000DF0000DF0000DF0000DF0000D0\
0000DF0000D0E0D000E0D000E0D000DFCF00DFCF00DFCF00DFCF00D0C000DFC000
D0C000D0C000CFBF00D000000000AF0000AF0000A00000A00000A00000A000009F00009F000090\
00009000008F00008F00008FB00000AF0000B00000AF0000AF00005F2F30A08F00
9F80009F8000907000907000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000005F6F00006F000070000000007F2F2F5F7F6F007F7000\
7F70008F7F008F8000908F009000009F00009F0000A00000A00000AF0000AF0000
0000C00000BF0000C00000C00000CF0000CF0000D0BFBF00C0C000C0BF00CFC000CFC000CFCF00\
CFCF00D0D0000000E00000EF0000E00000EF0000EFDF0000DF0000DF0000DF0000
E00000DF0000E00000E00000E00000E00000E0EF00E0E000EFEF00E0E000EFEF00E0EF00EFEF00\
E0EF00EFEF00EF0000EF0000EF0000EF0000EF00005000AF4F00AF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF5050AFEFF000F0F000EFF000F0F000F0F000F0F000\
F0F000F0AF00F0A000F00000F00000F00000F00000F00000F00000F00000F00000
F00000F00000FF0000F00000F00000F00000FF0000F00000FF0000F0F000FFF000F0F000F0F000\
F0EF00F0F000F0EF00F0F000F0EF00F0F0004F4FA00000F00000EF0000F00000EF
0000EF0000EF0000EF0000EF0000EF0000E00000EF0000E00000E00000DF0000E00000DF0000DF\
0000DF0000DFE0DF00EFDF00E0D000E0DF00E0D000E0D000DFCF00E0D000DFCF00
DFCF00D0C000DF4000D00000D00000D000000000B00000AF0000AF0000A00000A000009F00009F\
00009000009F00009000009000008FBF0000B00000B00000AF0000B03000AF9000
AF8F00A08000A080009F7F009F7F008F0000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000006000006000006F000000007F000080706F007F70007F7000\
807F00807F00908F009000009F00009F0000A00000A00000AF00000000BF0000C0
0000BF0000C00000C00000CF0000CF0000D00000D0C0C000C0BF00CFC000C0C000CFCF00CFCF00\
D0D000D0CF000000EFD00000DF0000D00000DF0000DF0000DF0000DF0000E00000
DF0000E00000E00000E00000E00000E00000E00000EFEF00E0E000EFEF00E0EF00EFEF00E0EF00\
EFEF00EFEF000000FF0000F00000FF0000F00000FF0000F00000FF0000F00000FF
0000F00000FF0000FF0000FF0000F00000FF0000FFF0F000EFEF00F0F000EFF000F0F000EFF000\
F0F000F0F000F0F0000000F00000FF0000FF0000FF0000F00000FF0000FF0000FF
0000F05000AF4F00AFF00000F00000F00000F00000F00000F00000F0F000F0F000FFF000F0F000\
F0F000F0F000F0F000F0EF00F0F000F0EF00F00000F000005000A00000EF0000F0
0000EF0000F00000EF0000EF0000EF0000EF0000E00000EF0000E00000E00000DF0000E00000DF\
0000DF4F4090EFE000E0DF00EFDF00E0DF00EFDF00E0D000E0D000DFD000E0D000
DFCF00DFCF00DF0000DF0000D00000D00000D0000040007F0000AF0000AF0000A00000A000009F\
00009F00009000009F00009000009000008F00008FB00000B00000AF9000B09000
AF8F00AF9000A08F00A080009F7F00907F008F4F00000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000006F6000006F000000007F0000800000807F70007F70008F8000\
8F8000905F009000009F00009F0000A00000A000003F00800000BF0000C00000C0
0000CF0000CF0000CF0000CF0000D00000D00000DF0000DFCFCF00CFCF00D0CF00CFCF00D0D000\
D0D000DF0000D00000DF0000DF0000DF0000DF0000E00000DF0000E00000E00000
E00000E00000E00000E00000EF00004F00AF0000FFE0EF00EFF000EFEF00EFF000EFEF00EFF000\
EFEF00EFF0004F4FAF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FFF0F000F0FF00F0F000F0F000F0F000F0FF00\
F0F000F0FF00F0F0000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF5000AFFF5000F0F000FFFF00F0F000FFFF00\
F0F000FFFF00F0F000FFFF00F0F000FF5000F00000FF0000F00000FF0000F00000
FF00005000A00000F00000EF0000F00000EF0000EF0000EF0000EF0000E00000EF0000E00000E0\
0000E0F0E000EFE000EFE000EFE000EFE000EFDF00EFE000EFDF00EFDF00E0D000
E0DF00E04000E00000DF0000E00000DF0000DF0000D00000D000000000AF0000AF0000A00000AF\
0000A00000A000009F00009F00009000009000008F000090B03000B09F00B09000
B09F00AF9000AF9000A08F00A08F009F80009F7F00000060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000606000006F000000007F00008000008000008F2F205F807F00808000\
9000009000009F00009F0000A00000A000000000BF0000BF0000C00000C00000CF
0000CF0000D00000CF0000D00000D00000DF0000DF0000DFCFCF00D0D000CFCF00D09000D00000\
DF0000D00000DF0000DF0000E00000DF0000E00000DF0000E00000DF0000E00000
E000004F00AF0000F00000FF0000F00000FF0000F04F4FAFEFEF00EFF000EFEF00EFF000EFEF00\
EFF000EFEF00EFF0000000F00000FF0000F00000FF0000FF0000FF0000F00000FF
0000FF0000FF0000F00000FF0000FF0000FF0000F0F0F000F0F000F0FF00F0F000F0FF00F0F000\
F0FF00F0F000F0FF000000FF0000FF0000F00000FF0000FF0000FF0000F00000FF
0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF4F4FAFFFFF00F0F000FFF000F0F000\
FFFF00F0F000FFF000F0F000FFFF00F04F00FF0000F00000FF0000F00000FF0000
F00000FF0000F00000F000004F00A00000F00000EF0000EF0000EF0000EF0000E00000E00000E0\
504F9FEFE000EFE000EFE000EFE000EFDF00EFE000EFDF00EFE000E0DF00EFDF00
E0DF000000CFE00000E00000DF0000DF0000DF0000DF0000D00000D00000CF00000000AF0000A0\
0000A000009F00009F00009F00009F00009000009000008F3F3060B09F00B09F00
B09000B09F00AF9000AF9000AF8F00A08F009F7F0000006F000060000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006060006000006F000000007000008000008000008F00008F00009F8F8000900000\
9000009F00009F0000A00000A000000000BF0000C00000C00000C00000CF0000CF
0000D00000D00000DF0000DF0000DF0000DF0000E00000E040409FD08F00DF0000D00000DF0000\
DF0000E00000DF0000E00000DF0000E00000E00000E000004F00A00000FF0000F0
0000FF0000F00000FF0000FF0000FF0000F00000FF0000FFEFF000EFF000F0F000EFF000F0F000\
EFF000F0FF00EFF0000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FFF0FF00F0FF00F0F000FFFF00F0FF00FFFF00\
F0F000FFFF00F0FF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5050AFF0F000FFFF00FFF000FFFF00\
F0F000FFFF00FFF000FFFF00F0F000FF5000FF0000FF0000FF0000FF0000FF0000
FF0000FF0000FF0000FF0000FF0000F00000FF00000000F00000F00000EF0000EF0000EF0000EF\
4F4F9FF0EF00F0EF00F0EF00EFE000F0EF00EFE000F0E000EFE000F0E000EFE000
504F900000D00000D00000CFE00000E00000E00000DF0000DF0000D00000DF0000D000000000B0\
0000AF0000AF0000A00000A000009F00009F000090000090BFA000BFA000B09F00
B09F00B09F00B09F00AF9000B09000AF9000AF8F0000006F00006F000060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000060000060000000007000007F00008000008F00008F00009000009F900000900000\
9F00009F0000A00000A06F00AFA0000000BF0000C00000C00000CF0000CF0000D0
0000D00000DF0000DF0000E00000DF0000E00000E0D00000D00000D00000D00000DF0000DF0000\
DF0000DF0000E00000DF00004F00AF0000F00000FF0000F00000F00000F00000FF
0000F00000FF0000F00000FF0000F00000FF0000F00000FFEFF000EFF000EFF000F0FF00EFF000\
F0FF00EFF000F0FF004F4FA00000FF0000FF0000FF0000F00000FF0000FF0000FF
0000F00000FF0000FF0000FF0000F00000FF0000FFF0FF00F0F000F0FF00F0F000F0FF00F0F000\
F0FF00F0F000FFFF000000F00000FF0000FF0000FF0000F00000FF0000FF0000FF
0000F00000FF0000FF0000FF0000F00000FF0000FF0000FFF0F000FFFF00F0F000FFFF00F0F000\
FFFF00F0F000FFFF00F0F000FFFF000000FF5000AFF00000FF0000F00000FF0000
F00000FF0000FF0000FF0000F00000FF0000F00000FF0000F000005000A00000EF0000F00000EF\
F0EF00F0EF00F0EF00EFEF00F0EF00EFE000F0EF00EFE000EFE000EFE000EFE000
0000D00000DF0000D00000D00000CF0000CFE00000E00000DF0000DF0000D00000D00000D00000\
D000000000AF0000AF0000A00000A000009F00009F3F3060BFA000BFA000BFA000
B09F00B09F00B09F00B09F00AF9000AF9000AF8F0000007F00006F00006F000060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000605F0060000060000000007F00008000008F00008F00009F2F006F9000009000009F0000\
9F0000A00000A0A000AFAF00AFAF00B0B0000000CF0000D00000D00000D00000D0
0000DF0000DF0000E00000E00000EFD00000D00000D00000DF0000DF0000DF0000DF0000E00000\
E00000E00000E0E000E0EF004F4FA00000FF0000F00000FF0000F00000FF0000F0
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5050AFF0F000F0FF00F0F000F0FF00\
F0FF00F0FF00F0FF005050AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FFF0FF00F0FF00F0FF00F0FF00F0FF00FFFF00\
F0FF00FFFF00F0FF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF0000FF5000AF
FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000F000005050A0\
F0F000FFF000F0EF00F0F000F0EF00F0EF00F0EF00F0EF00F0EF00F0EF004F4F90
0000E00000DF0000DF0000D00000DF0000D00000D04F008FE00000DF0000DF0000DF0000DF0000\
D00000D000004000700000AF0000A00000A000009FC0B000BFAF00C0AF00BFA000
BFA000BFA000BFA000B09F00B09F00B09000B09000AF000000007F00007000006F000060000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F5F0060000060000000007F00007F00008F00008F0000908000009000009000009F00009F0000\
A00000A09F00AFA000AFAF00B0B000B0B0003F3F8F0000CF0000D00000D00000DF
0000DF0000E00000E0D00000D00000D00000D00000DF0000DF0000DF0000DF0000E00000DF0000\
E0EF00E0E000E0EF00E0EF00EFEF000000F00000FF0000F00000FF0000F00000FF
0000F00000FF0000F00000FF0000F00000FF0000F00000FF0000F0F0FF00EFF000F0FF00EFFF00\
F0FF00EFF000F0FF00EFAF005000AFEF0000F00000EF0000F00000EF0000F00000
EF0000F00000EF0000F00000EF0000F00000EF0000F05000EFFF00F0FF00EFF000F0FF00F0FF00\
F0FF00F0F000F0FF00F000005000AF0000F00000FF0000FF0000FF0000F00000FF
0000FF0000FF0000F00000FF0000FF0000FF0000F05050AFF0F000FFFF00F0F000FFFF00F0F000\
FFFF00F0F000FFFF00F0F0005050AF0000F00000FF0000FF0000FF0000F00000FF
0000FF0000FF0000F0FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000F00000\
FF0000F0F000F0F000F0EF00F0F000F0EF00F0EF00F0EF00F0EF00EFEF000000E0
0000DF0000E00000DF0000DF0000D00000D00000CF0000D00000CFE00000DF0000DF0000DF0000\
DF0000D00000D00000CF00000000AF0000A0403F70C0AF00C0B000BFA000BFAF00
BFA000BFA000B09F00BFA000B09F00B09F00AF2F00AF0000AF000000007F00006F00006F000060\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
60000060000000007F00007F00008000008F00009F8F00009000009000009F00009F0000A00000\
A0A000AFAF00AFAF00B0B000B0B000BFBF00BFBF000000DF0000DF0000E00000E0
0000E0CF0000D00000D00000DF0000D00000DF0000DF0000E00000DF0000E00000E0E000E0EF00\
E0EF00EFEF00E0EF00EFF000E0EF005050AF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF4F50AFF00000F00000F00000\
EF0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000
F00000EF0000F00000EF0000F00000EF0000F00000EF0000F0FF00EFFF00F0FF00EFFF00F0FF00\
EFFF00F0FF00F0FF00F00000F00000F00000F00000F00000F00000F00000F00000
F00000F00000FF00005000AF0000FF0000FF0000FF5050AFFFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000\
F00000FF0000F00000FFAF00F0F000FFF000F0F000F0F000F0EF005050A00000E0
0000E00000E00000E00000DF0000DF0000D00000DF0000D00000D00000CF4F008FE00000E00000\
DF0000DF0000D00000D00000D000004F007F403F70CFBF00C0B000CFB000C0AF00
C0AF00BFA000BFAF00BFA000BFA000B09F00B00000AF0000B00000AF000000007F00007000006F\
000060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000600000
6000006F600000007F0000800000800000902F00608F00008F00009F00009F000030007F30307F\
AFAF00AFAF00B0B000B0B000BFBF00BFBF00C0C0003F3F900000DF0000DF40009F
CF0000D00000D00000DF0000D00000DF0000DF0000DF00004000A00000F00000F04F4FAFE0EF00\
E0EF00E0EF00EFF000E0EF00EFF000EFF0000000FF0000F00000FF0000F00000FF
0000F00000FF0000FF0000FF0000F05000AFEF0000F00000EF0000F00000EF0000F00000EF0000\
EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000
EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EFF000EFFF00EFFF00EFFF00EFF000\
EFFF00EFFF00F0FF00EF4F00F00000EF0000F00000EF0000F00000EF0000F00000
EF0000F00000EF0000F00000F00000F00000F00000F0AF00F0F000FFFF00F0FF00FFFF00F0F000\
FFFF00F0F000FFFF004F4FA00000FF0000FF0000FF0000F00000FF0000FF0000FF
0000F00000FF0000FF0000FF0000F00000FF0000FF5000AFF00000FF0000FF0000FF0000F00000\
FF0000F00000FF0000F00000FF0000F00000FFF000F0EF00F0F0000000EF0000EF
0000E00000E00000E00000E00000DF0000DF0000DF0000DF0000D00000D00000CF0000CFDF0000\
E00000DF0000DF0000D00000D00000D00000D00000CFBF00CFBF00C0B000C0B000
C0AF00C0AF00BFA000BFAF00B0A000BFA00000008FB00000AF0000AF0000AF000000007F00006F\
00006F000060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000605F00600000
6F40006F60000000800000800000900000908F00008F00009F00009F00000000BF0000BF0000C0\
AFAF00B0B000B0B000BFBF00BFBF00C0C000C0C000CFCF000000E0D00000CF0000
D00000D00000DF0000DF0000DF0000DF00004F00AF0000F00000FF0000F00000FF0000F0EFF000\
EFF000EFF000EFF000EFFF00EFF000F0FF00EFF0000000FF0000FF0000FF0000FF
0000FF4F00AFF00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000\
EF0000EF0000EFAF00EFFF00EFFF005050AF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFEFFF00EFFF00EFFF00EFFF00EFFF00\
EFFF00EFFF00EFFF005050AF0000FF5000AFEF0000F00000EF0000F00000EF0000
F00000EF0000F00000EF0000F00000F00000F00000F00000F00000F00000F00000F0AF00FFFF00\
F0FF00FFFF00F0FF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000\
FF0000FF0000FF0000FF0000FF0000FF0000F00000FF00008030700000F00000EF
0000F00000EF0000EF0000E00000E00000E00000E00000DF0000DF0000D00000D00000D00000D0\
0000CFE00000DF0000DF0000DF0000DF0000D00000D00000CF0000CFBF00CFB000
CFBF00C0B000C0B000BFAF00BFAF003F306000009000008FB00000AF0000AF0000AF000000007F\
00007000006F000060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000005F0000600000
6060006F6F0000008000008F0000908F00008F00009000009F00000000BF0000BF0000C00000C0\
0000CFB0B000BFBF00BFBF00C0C000C0C000CFCF00CFCF00CF4000CF0000D00000
D00000DF0000D00000DF00000000F00000F00000F00000F00000F00000FF0000F00000FF4F4FA0\
EFF000EFF000EFF000EFF000F0FF00EFF000F0FF004F50AFF00000EF0000EF0000
EF0000F00000EF0000EF0000EF0000F00000EF0000EF0000EF0000EF00004F00A0EFFF00EFFF00\
EFFF00E0F000EFFF00EFFF00EFFF00E0F0000000FF0000FF0000FF0000F00000FF
0000FF0000FF0000F00000FF0000F00000FF0000F00000FF4F4FA0E0FF00E0F000E0FF00E0FF00\
EFFF00E0F000EFFF000000FF0000FF0000F00000FF0000FF0000FF0000F00000FF
0000FF0000FF0000F04F00AFEF0000EF0000EF0000F00000EF0000F00000EF0000F00000F00000\
F00000F00000F0AF000000FF0000FF0000F00000FF0000FF0000FF0000F00000FF
0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FFF0F000FFFF00FFA000\
FF0000F00000FF0000FF0000FF0000F00000FF0000F00000FF0000F00000FF0000
0000EF0000EF0000EF0000EF0000E00000E00000DF0000DF0000DF0000DF0000D00000D00000D0\
0000D00000CF4F008FDF0000DF0000D00000DF0000D00000D00000CF0000CF8000
CFBF00CFBF00C0AF00C0AF00BFAF0000009000008F00008F00008FB00000AF0000AF0000AF0000\
00007F00006F00006F000060000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000005F0000600000606000
6F6F006F6F002F2F600000908F00008F00009000009F00000000BF0000BF0000C00000C00000CF\
0000CF404090BFBF00C0C000C0C000CFCF00CFCF00D0D000D0D000DF0000D00000
DF0000DF00004F00AF0000F00000F00000F00000FF0000F00000FF0000F00000FF0000FF0000FF\
EFF000F0FF00EFFF00F0FF00EFFF00F0FF00EFFF00F0FF00EF5000F00000EF0000
F00000EF0000F00000EF0000F00000EF00000000FF0000FF0000FF0000FF0000FF0000FFEFFF00\
EFFF00EFFF00E0FF00E0FF00E0F000E0FF004F50AF0000FF0000F00000FF0000F0
0000FF0000F00000FF0000F00000F00000F00000FF0000F00000F0D0EF00DFF000DFEF00DFF000\
DFF000E0F000DFF0004F50AF0000F00000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FFEFAF00EF0000EF0000F00000EF0000F00000\
EF0000F00000F00000F00000F00000FF00005000AF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5050AFFFFF00FFFF00FFFF00\
FFFF00FFFF00FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000
FF00005000A00000F00000EF0000EF0000E00000EF0000E00000E00000DF0000DF0000DF0000DF\
0000D00000D00000CF0000CFDF0000E00000DF0000DF0000D00000D00000D00000
D00000CFBF00CFBF00C0B000403F6F00009F00009F00009000009000008FB00000AF0000AF0000\
AF00003F005000007000006F000060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000600000605F006F6000
6F60007070007F70008000008000009000009000000000B00000BF0000C00000C00000CF0000CF\
0000D00000D0C0C000C0C000CFCF00CFCF00D0D000CFD000D0DF00D0DF00DF0000
DF00000000F00000F00000F00000F00000F00000F00000FF0000F00000FF0000F00000FF0000F0\
0000FFEFF000F0FF00EFF000F0FF00EFFF00F0FF00EFF000EFFF00EF5000EF0000
EF00004F00AF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000F00000FF0000F0\
E0F000DFF000E0F000DFEF00DFEF00D0EF00DFEF00D0EF004000A0D00000D00000
D00000D00000D00000D00000CF0000D00000CF0000D00000CF0000CF0000CF0000D00000CF0000\
D00000D00000D0EF00D0EF000000F00000F00000F00000F00000F00000F00000F0
0000F00000F00000F00000FF0000F00000FF4F50AFEFFF00E0F000EFFF00EFFF00EFAF00EF0000\
EF0000EF0000F00000EF0000F00000EF0000F00000F00000F00000F000000000FF
0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0FFFF00FFFF00FFFF00F0F000\
FFFF00FFFF00FFFF00F0F000FFAF00FF0000FF0000F00000FF0000FF0000FF0000
F00000FF0000F000000000F00000EF0000EF0000E00000E00000E00000E00000DF0000DF0000D0\
0000D00000D00000D04F408FE0DF00DFD000E00000DF0000DF0000D00000D00000
CF0000D00000CFBF00CFBF0000009F00009F00009000009000008F00008F000080B00000AF0000\
AF0000A00000AF000000006F00006F000060000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006000006F60006F6000
7070007F7000807F00802F0090000090000030007F0000BF0000C00000C00000CF0000CF0000D0\
0000D00000DF0000DFCFCF00CFCF00D0D000D0D000DFDF00D0DF00DFE000DFE000
0000F00000F00000F00000F00000FF0000F00000FF0000F00000FF0000FF0000FF0000FF0000FF\
0000FFF00000EFFF00F0FF00EFFF00F0FF00EFFF00F0FF00EFFF00F0FF000000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000F00000FF0000F00000FF0000F00000F0\
0000F0DFEF00D0EF00D0EF00D0E000D0EF00CFE000D0EF00CFE000CF0000CF0000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000C00000CF0000CF0000CF0000CF0000D00000CF0000D00000D00000
4F00A00000F00000F00000F00000FF0000F04F50AFE0F000EFFF00E0FF00EFFF00E0FF00EFFF00\
EFFF00EFFF004F00AFF00000EF0000F00000EF0000F00000F00000F00000F00000
FF00005000AF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FFFF0000FF0000FF0000FF0000
FF0000FF0000FF0000F00000FF00000000EF0000EF0000EF0000EF0000E00000E00000DF0000E0\
0000DF0000DF0000D0EFE000E0E000EFE000E0DF00E09000DF0000DF0000D00000
DF0000D00000D000006F2F4F0000AF0000A00000A000009F00009F00009000009000008F3F0060\
AF0000AF0000AF0000AF000000007000006F000060000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F0000000070606000706F00
7070007F7F00807F008F8F009000009F00000000B00000BF0000C00000CF0000CF0000D00000D0\
0000DF0000DF40009FCF4000D0D000CFD000D0DF00D0DF00DFE000DFDF00DFE000
DFE0000000F00000F00000F00000F00000FF0000F00000FF0000F00000FF0000F00000FFEF0000\
F00000EF0000F00000EF5000F0FF00EFF000EFFF00EFFF00EFFF00EFF000EFFF00
0000FF0000FF0000F00000FF0000F00000F00000F00000F00000F04F00A0D00000D00000D00000\
D00000CF00000000D0000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
C00000CF0000C00000CF0000CF00004000A0404FA0DFEF00DFEF00DFF000DFF000E0FF00E0F000\
EFFF004F4FA00000FF0000FF0000FF0000F04F00AFEF0000EF0000EF0000F00000
EF0000F00000F00000F000000000FF0000FF0000F05050AFFFFF00FFFF00F0F000FFFF00FFFF00\
FFFF00F0F000FFFF00FFFF00FFFF004F4FA00000FF0000FF0000FFF00000FF0000
FF0000FF0000F00000FF0000F00000FF00004F00A00000EF0000EF0000EF0000E00000E00000DF\
0000E00000DF4F4F90E0E000EFE000E0DF00E0E000E0DF00E0DF00DF0000DF0000
D00000D00000CF0000D000004000700000AF00009F00009F00009000009000008F00008F00008F\
3F005FAF0000AF0000A00000AF000000006F00006F8F7000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000600000000070706F00706F00
7F7F00807F008F8F00908F009F00000000B00000C00000C00000CF0000CF0000D00000D00000DF\
0000DFCF0000CF0000D00000D0D000DFDF00D0DF00DFE000DFE000E0EF00E0EF00
E0EF00E0EF000000FF0000F00000FF0000FF0000FF0000FF0000FF0000FFF00000EF0000F00000\
EF0000F00000EF0000F00000EFFF00F0FF00EFFF00F0FF00EFFF00F0FF00EFFF00
EFFF00E0F0000000FF0000F00000F0DF0000DF0000D00000D00000D00000D00000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000C00000CFE000CFE000D0EF00D0E000DFEF00DFEF00DFF000\
DFF0004F50AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5000AFEF0000
F00000EF0000F00000F00000F00000F000005000AFF0FF00FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF005050AF0000FF0000FF0000FF0000FF0000FF5000AF
FF0000FF0000FF0000FF0000FF0000F00000FF0000F000000000F00000EF0000EF0000E00000E0\
0000E050509FEFEF00EFEF00EFE000EFE000E0E000EFE000E0DF00E0DF00DF9000
DF0000DF0000DF0000D00000D000004000700000AF0000A00000A000009F00009F000090000090\
00008F00008FAF0000AF0000AF0000AF00000000709F80008F7000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000060000000006F00007F6F6F007F7000
7F7F008080008F8F009090003030700000BF0000BF0000C00000CF0000D00000D00000DF0000DF\
CF0000CF0000D00000CF0000D00000D0DF00DFE000DFE000DFE000DFE000E0EF00
E0EF00E0F0004F4FA00000FF0000F00000FF0000FF5000AFEF0000EF0000EF0000F00000EF0000\
F00000EF00000000FF0000F00000FFEFFF00EFFF00EFF000EFFF00E0F000E0F000
DFEF00DFF000DF0000DF0000D00000D000000000D0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000CFDF00CFDF00CFE000CFE000\
D0EF000000F00000F00000F00000F00000F00000FF0000F00000FF0000FF0000FF
0000F0EF0000EF0000F00000EF0000F00000F05000F0FF00F0F000F0FF00F0FF00FFFF00F0F000\
FFFF00FFFF00FFFF00F0F0000000FF0000FF0000FF0000F00000FF0000FF0000FF
0000F0FF0000FF0000FF0000F00000FF0000F00000FF0000F00000F000000000EF0000EF0000E0\
0000E0EFEF00F0EF00EFE000EFEF00EFE000EFE000E0E000E0E000E0DF00E0DF00
DF9000DF0000D00000D00000CF0000CF0000CF00000000AF00009F00009F00009000009000008F\
00008F00008F00008FAF0000AF0000A00000AF00009F80009F80008F7000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F000000007000008000008F7F7F00
8080008F8F00909000909000A0A0000000BF0000C00000CF0000D00000D00000DF0000DFCF0000\
CF0000D00000D00000DF0000D000000000F0DFE000E0EF00E0EF00E0EF00E0EF00
EFF000EFF000EFFF004F50AF0000FF0000FFF00000EF0000F00000EF0000F00000EF0000F00000\
0000FF0000FF0000FF0000FF0000FF0000FF0000F0EFFF00E0F000E0F000DFA000
E00000DF0000DF0000CFE000BFD000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
CF4F00CF0000CF0000D000000000F00000F00000F00000FF0000F00000FF0000FF
0000FF0000FF0000FF4F00AFF00000EF0000F0FF00F0FF00F0FF00F0FF00F0FF00F0FF00FFFF00\
FFFF00FFFF00FFFF005050AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FFFF0000FF0000FF0000FF0000FF0000FF0000F00000FF00004F00A00000F0\
F0F000F0F000EFEF00F0F000EFEF00EFEF00EFE000EFEF00EFE000EFE000E0DF00
E0E0000000C0DF0000DF0000DF0000D00000D00000CF00000000AF0000A00000A000009F00009F\
00009000009000008F00008FAF0000AF0000A00000AF00009F8F00907F00000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F000000007000008000008F808000
8080009090009090009F9F00A0A000AFAF000000C00000CF0000D00000DF0000DFC00000CF0000\
D00000CF0000D00000D000000000F00000EF0000F0DFE000E0EF00E0EF00EFF000
E0F000EFFF00EFF000EFFF00EFF000F00000EF0000EF0000EF0000F00000EF00000000FF0000FF\
0000FF0000F00000FF0000F00000FF0000F00000F00000F0E00000DF0000DF0000
D00000C0DF000000BF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000CF0000CF0000CF00004000A00000F00000F00000F0
0000F00000FF0000F00000FF0000FF4F50AFEFF000EFFF00EFFF00F0FF00EFF000F0FF00F0FF00\
F0FF00F0F000FFFF000000FF0000FF0000F00000FF0000FF0000FF0000F00000FF
0000FF0000FF0000F00000FF5000AFFF0000F00000FF0000FF0000FF0000F00000FF0000F0F000\
F0F000F0EF00F0F000EFEF00F0EF00EFEF00EFEF00EFE000EFEF00E0E000E0E000
4F408F0000CF0000C0DF0000D00000D00000CF0000CF0000CF00000000AF0000A000009F00009F\
00009000008F00008F00008F00008FAF0000AF0000A00000AF00009F80008F7F00
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000060000000006F00007F000080800000800000
8F8F009090009F9F00A0A000AFAF00AFAF00B0B0000000D00000DF0000DFCF0000CF0000D00000\
D00000DF00004000A00000F00000F00000F00000F00000FFE0EF00EFF000EFF000
EFFF00EFF000F0FF00EFAF00F00000EF0000F00000EF0000F000000000FF0000FF0000FF0000FF\
0000FF0000FF0000F00000FF0000F00000FFDF0000DF0000DF00004F00A00000CF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CF0000CF0000D000000000F0
0000F00000F00000FF0000FF4F50AFEFFF00EFFF00EFFF00EFFF00EFFF00F0FF00F0FF00F0FF00\
F0FF00FFFF00F000000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000FF0000FF5000FFFF00\
F0F000FFFF00F0F000F0F000F0EF00F0F000EFEF00F0EF00EFE000EFEF004F4F90
0000D00000CF0000D00000C0DF0000D00000DF0000D00000D00000CF00000000AF0000A00000A0\
00009F00009F00009000009000008F00008FAF0000AF0000A00000AF0000908000
8F7000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F000000007000007F00008F7F00008F0000
8F8F009F9F009F9F00A0A000A0AF00B0B000B0B0003F40900000D040009FC00000CF0000CF0000\
D00000D000000000F00000EF0000F00000F00000FF0000F00000FFEFF000EFFF00
EFF000EFFF00EF0000F00000EF0000F00000EF0000F000000000F00000FF0000FF0000FF0000F0\
0000FF0000F00000F0DF0000E00000DF0000DF00000000DF0000C0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000C00000CF0000
CF00000000F00000F00000F0DFF000E0FF00E0FF00EFFF00E0F000EFFF00EFFF00EFFF00EFF000\
F0FF00EF0000F00000F00000F000000000FF0000FF0000F00000FF0000FF0000FF
0000F00000FF0000FF0000FF0000F00000FF0000FF0000FFF00000FF0000FF0000FFFF00F0F000\
FFFF00F0F000FFF000F0F000F0F000F0EF00F0F000EFEF00EFEF00EFE0000000DF
0000D00000D00000CF0000CF0000C0DF0000D00000D00000CF0000CF0000CF00000000AF0000A0\
00009F00009F00009000008F00008F00008FB09F00AF0000AF0000A00000A08F00
8F7F00802000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F5F0000006F00007F00008F2F00608000008F0000
30307F9F9F00A0A000A0AF00B0B000B0B000BFBF00BFC0000000E0C00000CF0000CF0000D00000\
D000000000F00000F00000F00000F00000FF0000F00000FF0000FF0000FFEFF000
F0AF00EF0000F00000EF0000F00000F00000F0AF00EFFF00F0FF000000FF0000FF0000F00000FF\
0000F04F00AFDF0000E00000DF00000000E00000CF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000CF0000D00000D00000DFEF00DFEF00E0F000E0F000EFFF00EFFF00EFFF00EFFF00F0FF00\
EF0000F00000EF0000F00000F00000F00000F000000000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5000AFFFFF00FFFF00FFFF00\
FFFF00FFFF00F0F000FFFF00F0F000F0F000F0F000F0F000EFEF000000E00000DF
0000DF0000D00000D00000CF0000D00000CFDF0000D00000DF0000D00000D00000CF00000000AF\
0000A00000A000009F00009F0000903F3F60B09F00B0A000AF0000AF0000A00000
9F8F008F7F00000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000605F0000006F00007F0000807F0000800000900000
0000B0A0A000A0A000AFAF00AFB000BFBF00BFBF00C0C000C00000CF0000CF0000D00000D00000\
0000F00000EF0000F00000F00000FF0000F00000FF0000F00000FF0000F0EF0000
EF0000F00000EF0000F00000EF0000F0FF00EFF000F0FF00EFF000F0FF00EFF000EFF0004F4FA0\
E00000DF0000DF00000000E00000DFB0CF00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CF0000CF0000D00000D0E000DFEF00DFEF00E0F000E0F000EFFF004F50AF\
0000FF0000F04F00AFEF0000F00000EF0000F00000F000005000AF0000F00000FF
0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FFFFFF00F0F000FFFF00FFFF00\
FFFF00F0F000FFFF00F0F000FFFF00F0F000F0F000F0EF005050A00000E00000E0
0000DF0000DF0000D00000D00000CF0000CF0000C0DF0000D00000D00000CF0000CF0000CF0000\
4000700000A00000A000009F00009F3F3060B0A000B09F00B09F00A00000AF0000
A000009F8000000060000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F5F0000007F0000800000907F00008F00000000AF
0000BF0000BFAFB000AFB000B0BF00BFBF00C0CF00C00000CF0000CF0000D00000D000000000F0\
0000F00000F00000F00000FF0000F00000FF0000FF0000FF0000FFF00000EF0000
F00000EF0000F00000F0AF00F0FF00F0FF00F0FF00EFFF00F0FF00EFF000EFFF00E0F000E0F000\
DFF000E0F000D0E000C0DF00000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000C00000CF0000CF0000D0A000D0EF00DFF0004F4FA00000FF\
0000FF0000FF0000FF0000FFEF0000F00000F00000F00000F00000FF00000000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0F0005050A00000EF0000EF0000E0
0000E00000DF0000DF0000D00000D00000CF0000D00000CF4F008FD00000DF0000D00000D00000\
CF00004000700000A00000A000009FBFAF00B0A000B0A000AF9F00B06F00AF0000
AF0000A05F0000006F000060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F5F0000007F00008F7000008000008F00000000B0
0000B00000C00000C0B0BF00B0BF00BFC000C00000CF0000CF0000D00000D000004F4FA00000EF\
0000F00000F00000F00000F00000FF0000F00000FF0000F0EF0000EF0000F00000
EF0000F000000000FF0000FF4F4FA0F0FF00EFF000EFFF00E0F000EFF000E0F000E0F000DFEF00\
DFEF00CFDF00C0D000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000C00000CF0000D000000000F00000F0\
0000F00000F00000FF0000F00000FF0000FFF00000EF0000F00000F00000F00000
F000000000FF0000FF0000FF0000F00000FF0000FFFFFF00F0F000FFFF00FFFF00FFFF00F0F000\
FFFF00FFFF00FFFF00F0F000FFFF00F0FF00FF5000F000005000A00000EF0000EF
0000E00000E00000DF0000DF0000D00000D00000CF0000CF0000C04F0080D00000D00000CF0000\
CF0000CF0000CF00000000A0C0B000BFAF00BFAF00B0A000B0A000AF9F00AF0000
A00000A0000000006F00006F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060005F5F0000008000008F80000080000030007F0000B0
0000C00000C00000CF0000D0BFC000C08000CF0000CF0000D00000D00000DF9F00D0E000E0EF00\
0000F00000FF0000F00000FF0000FF0000FF0000FFF00000EF0000F00000EF0000
F000000000FF0000FF0000FF0000FF0000FF0000FF0000F05050AFE0F000E0F000DFF000DFEF00\
0000D0000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000CF0000CF0000D00000\
0000F00000FF0000F00000FF0000FF0000FF0000FF5000AFEF0000F00000F00000
F00000F000005000AF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FFFF00FF5000FF0000FF0000F000005000A00000EF
0000EF0000E00000E00000DF0000DF0000D00000D00000CF0000D00000CF0000CFD00000DF0000\
D00000D00000CF0000CF4000C0BF00C0B000BFAF00BFAF00B0A000B0A000AF9F00
AF0000A00000A0000000006F00006F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F5F0060600000007F2F00607F00008F00000000AF0000BF
0000BF0000CF0000CF0000D00000DFC00000C00000CF0000CF0000D00000D0DF00DFE000DFE000\
E0EF000000F00000FF0000F00000FF0000F0EF0000EF0000F00000EF0000F00000
0000FF0000FF0000F00000FF0000F00000FF0000F0EF0000E00000E0000040009F0000DF0000CF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000BFC00000\
D000004000A00000F00000F00000FF0000F00000FF0000FF0000FFEF0000F00000
EF0000F00000F00000F000000000FF5050AFF0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00\
FFFF00F0F000FFFF00F0FF00FF5000F00000FF0000F00000FF0000F00000F00000
0000EF0000EF0000E00000E00000DF0000DF0000D00000D00000CF0000CF0000C00000C0D00000\
D00000CF0000CF4000CFBF00CFBF00C0B000C0B000BFAF00BFAF00B0A000B0A000
AF9F00AF0000A00000A0000000006F000060000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000605F006060007070007F00008000008F00000000B00000BF
0000C00000CF0000D00000DFC00000C00000CF0000CF0000D00000D0DF00DFE000DFE000E0EF00\
E0EF00EFF0000000FF0000FF0000FFF00000EF0000F00000EF0000F000000000FF
0000FF0000FF0000FF0000FF0000FF0000F0EF0000E00000E000000000E00000DF0000CF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000DFCF0000D000000000F00000F00000FF0000FF0000FF0000FF0000FF0000FF
F00000EF0000F00000F00000F00000F00000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF005050AF0000FFFF0000FF0000FF0000FF0000FF0000F00000
F000000000EF0000EF0000E00000E00000DF0000DF0000DF0000D00000D00000D00000CF0000CF\
D00000DF0000D0CF00D0CF00CFBF00CFBF00C0B000C0B000BFAF00BFAF00B0A000
B0A000AF0000AF0000A0000000007000006F800000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F5F006060006F6F007F00008000000000AF0000B00000C0
0000C00000CF0000D00000DFC00000CF0000CF0000D00000D0DF00DFE000DFE000E0EF00E0EF00\
E0F000E0F000EFFF00EFF000EF0000EF0000F00000EF0000F000000000FF0000FF
0000F00000FF0000F00000FF0000F0EF0000E00000E000000000E00000DF0000CF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000CF0000CF0000D000000000F00000F00000F00000FF0000FF0000FF
0000F0EF0000EF0000F00000EF0000F00000F00000FFAF00F0F000FFFF00FFFF00FFFF00F0F000\
FFFF00F0FF005050AF0000F00000FF0000FF5000AFF00000FF0000F00000FF0000
F00000F000000000EF0000EF0000E00000E00000DF0000DF0000D00000D00000CF0000CF0000C0\
0000C0D04000D0CF00CFC000CFC000C0BF00C0BF00BFB000BFB000B0AF00B0A000
AF9F00000080A00000A000009F000000006F000060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060006F6F007F00007F00008F00000000B00000BF0000C0
0000CF0000CF0000DFBF0000CF0000CF0000D000000000EF4F4FA0DFE000E0EF00E0EF00EFF000\
EFF000EFFF00EFF000F0FF00EFF000F0FF00F05000F000000000FF0000FF0000FF
0000FF0000FF0000FF0000F0EF0000E00000EF00000000EF0000E00000D0000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000D0CF0000D000004000A00000FF0000F00000FF0000FF
0000FF4F50AFF0FF00EF0000F00000F00000F00000F00000FF0000F0FF00FFFF00FFFF00FFFF00\
FFFF005050AF0000FF0000FF0000FF0000FF0000FF5000AFFF0000FF0000FF0000
FF0000F00000FF00000000EF0000EF0000E00000E00000DF0000E00000DF0000DF0000D00000D0\
4F408FDFDF00D0D000DFD000D0C000CFC000C0BF00C0BF00BFB000BFB000B0AF00
3F3F6000008F000080A00000A00000A0000000006F000060000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F5F006F6F007000008000000000AF0000B00000BF0000C0
0000CF0000D00000D0C00000C00000CF0000CF00000000EF0000EF0000F0DFEF00E0F000E0F000\
EFFF00EFF000EFFF00EFF000F0FF00EFF000F0FF00F0F0005050AF0000F00000FF
0000F00000FF0000F0EF0000E00000EF00000000EF0000E0CFDF00000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000C00000CF0000CF00000000F00000F00000F0
4F4FA0EFFF00EFF000EFFF00EFFF00F00000EF0000F00000F00000F00000F0A000FFFF00F0FF00\
5050AF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FFF00000FF0000
F00000FF0000F00000F000004F00A00000EF0000E00000E00000DF0000DF0000D00000D04F408F\
E0DF00DFD000DFD000D0CF00D0CF00CFC000CFC000C0BF00C0BF00BFAF00BFAF00
00008F00008F000080000080A00000A000009F000000006F00005F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060006F6F008000008000009090000000B00000C00000C0
0000D00000D0C00000C00000CF0000CF00000000F00000EF0000F00000F00000FF4F4FAFEFFF00\
EFF000F0FF00EFF000F0FF00F0F000F0FF00F0FF00F0FF00F0FF00FFFF00F0FF00
F0FF00EFFF00F00000EF0000EF0000E0EF00DFEF00CFDF000000AF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000B0B000CF0000CF0000D0EF00DFEF00
E0F000E0F000EFFF00EFFF00EFFF00EFFF00F00000F00000F00000F00000FF0000F000000000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000
FF0000FF0000FF0000FF0000FF00004F00A00000F00000EF0000E00000E00000E04F4F90EFE000\
E0DF00E0E000DFDF00DFD000D0CF00D0CF00CFC000CFC000C0BF00C0BF0000009F
00009F00008F00008F000080AF0000A00000A0000000006F00006F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000070704F007F00008F60009090009FA0000000BF0000CF
0000CF0000DFBF0000CF0000CF0000D000000000EF0000F00000EF0000F00000F00000FF0000F0\
4F50AFEFF000F0FF00EFF000F0FF00F0F000F0FF00F0F000F0FF00F0F000F0FF00
EFF000EF0000EF0000EF0000E0A000DFEF00CFDF00BF0000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000BFCF00CFDF00D0E000
D0E000DFEF00DFEF00E0F000E0F000EFFF00EFFF00F0FF00EF0000F00000F00000F00000F00000\
0000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF
F00000FF0000F00000FF0000F00000F00000F000000000EF0000E00000E04F4F90EFEF00E0E000\
E0E000E0DF00E0DF00DFD000DFD000D0CF00D0CF00CFC000CFBF006F3F4F0000A0
00009000009000008F00008F000080A00000A00000A0000000006F000060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000007F7000008000008F90009F9F009FA000AFAF000000CF
0000DFBF0000C00000C00000D000004000A00000F00000F00000F00000F00000FF0000FF0000FF\
EF0000F00000EF0000F05000F0FF00F0FF00F0FF00FFFF00F0FF00F0FF00EFFF00
F0AF00EF0000EF0000E00000E0EF00404F900000D00000A0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000C0DF00
D0E000D0E000DFEF00DFEF00E0F000E0F000EFFF00EFFF00F0FF00EF0000F00000F00000FF0000\
F000000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FFFF0000FF0000FF0000FF0000FF0000FF0000F000000000F04F4F9FF0EF00EFEF00EFEF00\
E0E000E0E000E0DF00E0DF00DFD000DFD000D0CF00D0CF00CF4000CF00000000A0
0000A000009000009000008F00008F000080AF0000A00000A0000000006F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F000000007F7000007F0000909000909F00A0AF00A0AF00B0B000
0000D0C00000C00000CF0000CF00000000EF0000EF0000F00000F00000FF0000F00000FF0000FF\
F00000EF0000F00000F000000000FF0000F0F0FF00F0F000F0FF00EFF000F0FF00
EF0000EF0000E000000000EF0000DF0000D0BF0000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
BFC000CFDF00CFDF00D0E000DFEF00E0F000E0F0004F50AF0000FF0000FFEF0000F00000F00000\
F00000F000000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF
0000FF0000FF4F00A0FF0000F00000FF0000F00000FF0000F00000F0F000EFEF00EFEF00EFE000\
EFEF00E0DF00E0E000DFDF00DFDF00D0D000D0D000CFCF00CF0000C00000C00000
00009F00009F00009000009000008F00008000007FA000009F00009F00002F2040000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F00000000807000008000009090009FA000A0AF00AFB000B0BF00
BFC000C00000CF0000CF00004F00A00000EF0000F00000F00000FF0000F00000FF0000FFF00000\
EF0000F00000F000000000FF0000FF0000FF0000FF0000FF0000FF0000FFEF0000
EF0000EF00000000F00000E00000DF0000CF0000AF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000007F60007F60002F2040000060000060000060000060000060000060000060000060\
0000600000600000602F20407F60007F60007F6000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F3F8FCFDF00CFE000D0EF00404FA00000FF0000FF0000FF0000FF0000FFEF0000F00000\
F00000F00000F000005000AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF5000AFFF0000FF0000FF0000FF0000FF0000F00000F0F000F0EF00F0EF00\
EFE000EFEF00E0E000E0E000DFDF00E0DF00DFD000D00000CF0000CF0000C00000
C0000000009F00009F000090000090000080000080A00000A000009F000000006F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000006F00007F700000800000909F009F9F00A0AF00AFB000B0BF00
BFC000C0CF00C00000CF00000000EF0000F00000F00000F00000F00000FF0000FF0000FFEF0000\
F00000F00000F000000000F00000FF0000FF0000FF0000F00000FFEF0000F00000
EF00004F00A00000E00000E00000D0BF0000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000005F7F60007F60007F6000
7F6000806F00806000806F00806F008F6F0000006F00006F00006000006F00006F00006F000060\
00006F00006000006F00006000006F806F008F6F00806000806F007F6000000060
00005F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000CFCF0000CF0000D000000000F00000FF0000F00000FF0000FF0000FF4F00A0\
F00000F00000F00000F00000F000000000FF0000FF0000F00000FF0000FF0000FF
0000F00000FF0000FF0000FF0000F0FF0000F00000FF0000F00000FF0000F00000F0F000EFEF00\
EFEF00E0E000EFE000E0DF00E0DF00DFDF004F4F8F0000BFD00000CF0000CF0000
C0000040007000009F00009F00008F00008F000080000080A00000A000009F0000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000006F0000807000008F00000000B0A0A000A0AF00B0BF00B0BF00
C0CF00C0CF00CFD000CFD0000000F00000F00000F00000F00000FF0000FF0000FFEF0000F00000\
F00000F000000000FF0000FF0000FF0000FF0000FF0000FF0000FFF00000EF0000
F000000000EF0000EF0000DF0000D0AFB000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000006000006000006000006F00006000006F00006F30204F806F00
8F70008F70009070008F7000907F00907000907F009000009F0000900000900000900000900000\
9000009000009000009000009000009000009000009000009000009000008F0000
8F0000000060000060000060000060000060000060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000CFCF0000CF0000D000000000F00000FF0000FF0000FF0000FF0000FF\
0000FFF00000F00000F00000F00000FF00000000FF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FFFFFF00FF0000FF0000FF0000FF0000FF0000F00000F0F000\
EFEF00EFEF00E0E000EFE000E0E0004F4F900000CF0000CF400080D00000CF0000
CF0000C000000000A000009F00009F00009000008F000080000080A00000A000008F7F00000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000006F0000807F00002F006F0000B00000BFAFAF00AFB000BFC000
C0CF00CFD000CFD000D0DF00D0DF00DFE0000000F00000F00000F00000FF0000F0EF0000EF0000\
F00000F000000000FF0000FF0000FF0000F00000FF0000F0F00000EF0000EF0000
0000F00000EF0000DF0000D0BF0000AFB000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000005F000060000060\
00006F00006000006F00006F00006F8000008F00008F00009000008F0000902F00
9070009F7F009070009F7F00907F009F7F00907F009F7F0000007000007F000070000070000070\
00007000007000007000007000007000006F000070907F009F7F00907F00907F00
9050009000008F00008F00008000008F000000006000006000005F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000C00000EFCF0000D000000000F00000FF0000F00000FF0000FF\
0000FF0000F0F00000F00000F00000F00000F000000000FF0000FF0000F00000FF
0000FF0000FF0000F05050AFF0FF00FFFF00F0F000FF0000F00000FF0000F00000F00000F0A000\
F0F000EFEF00EFEF00E0E0004F4F900000D00000CF0000CF0000C0D00000D00000
CF0000CF0000BF00000000A000009F00009000008F00008F000080A000009F00009F8F00000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000700000807F00000000AF0000B00000C00000C03F4090BFC000
C0CF00CFD000D0DF00D0DF00DFE000DFE000E0EF00E0F0000000FF0000FFF00000EF0000F00000\
F000000000FF0000FF0000FF0000FF0000FF0000FF0000FFF00000F00000EF0000
0000F00000E0DFE000D0DF00BFC000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007060002F204F00006000006F00006F8F0000\
8F00009000008F00009000009000009F00009000009F000030005000007F00007F
302F509F7F00A080009F8000A080009F8000A08000A080003F2F5F00007F00008000007F00007F\
00007F00007F00007F00007F00007000007F00007000007F9F8000A080009F8000
9F8000907F00302F4F00006F00006F3000409000008F00008F0000800000806F00705F00000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000C00000E0CF0000D000000000F00000FF0000F00000FF\
0000FF0000FF0000FFF00000F00000F00000F00000FF00000000FF0000FF0000FF
0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000FF0000F00000F00000\
F0F000F0F000EFEF004F4F9F0000DF0000DF0000D00000D00000C00000C0D00000
D00000CF0000CF0000C000000000A000009000009000008F00008F00007FA000009F8F00807000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000007020005F7F00000000A00000B00000BF0000CF0000CFBF4000
C0CF00CFD000CFD000D0E000D0E000DFEF00E0EF00EFF000E0F000EFF000EF0000F00000EF0000\
F000000000FF0000FF0000F00000FF0000FF0000FFF00000F00000EF00000000F0
E0EF00DFE000D0DF00CF00000000AF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000705F007F60007F6000806F008060008F00008F00009000008F0000\
90000090000000007F00007000007F00007F00007F00007F00007F00007F000080
00007FA080009F8000A080009F8000A08F00A08000A08F00A0800000008000007F00008000007F\
00008000007F00008000007F00007F00007F00007F00007F302F509F8000A08F00
9F80009F80009F800000007000006F00007000006F00006F00006000006F2F00409000007F6F00\
2F2040000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000DFC00000D000004000A00000FF0000F0\
0000FF0000FF0000FF0000F0F00000F00000F00000F00000F000000000FF0000FF
4F4FA0FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0A000FF0000F00000FF0000F00000\
F00000EFEF000000EF0000E00000DF0000D00000D00000CF0000CF0000C00000C0
CF0000CF0000C00000C000003F006F00009F000090000090000080000080A030009F9000908000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000707000007F00000000AF0000B00000C00000CF0000D0BF0000
CF0000CF4000D0DF00D0DF00DFEF00DFEF00E0F000EFF000EFFF00EF0000F00000F00000F00000\
F0FF00FFFF005050AF5050AF5050AFFFFF00F0FF00F0FF00F0FF00F0FF00E0F000
E0EF00DFE000DFE000C000000000AF000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000006000006F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000007060007F60007F6000806F00806F008F00008F000090000090000090000000007F000080\
00007F00008000007F000080000080000080000080000080000080000080000080
00008F000080AF8F00A08F00AF8F00A08F00AF9000A08F00AF9000AF8F0000008F00008000008F\
00008000008F00008000008F000080000080000080000080000080AF8F00A08F00
AF8F00A08F00A08F009F800000007F00007000007000007000007000006F00006F00006F00006F\
8F7000806F0000005F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000DFCF0000D00000D000000000FF\
0000F00000FF0000FF0000FF0000FFF00000F00000F00000F00000FF0000F0FF00
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000F00000FF0000\
F00000F000000000EF0000EF0000E00000E00000D00000DF0000CF0000CF0000C0
0000C0D00000CF0000C00000C000000000A000009F000090000090000080A09F009F90009F9000\
7F6F00000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006060006F00007F00000000AF0000B00000BF0000CF0000D0BF0000
C00000CF00000000EF404FA0DFE000E0EF00E0EF00EFF000EFF000F00000EF0000F00000F00000\
FFFF00F0F000FFFF00F0FF00FFFF00F0F000F0FF00F0FF00706000706000706000
7040007F00007000007F00007000007F00007000007F00007F00007F00007F00007F00007F0000\
8000007F0000800000800000800000800000800000802000806F00806000806F00
806000806F007F6000806F007F6000806F007F20002F004F00006000006F000060000060000060\
00000000000000000000000000000000000000000000000000000000005F000060
00006000006F7F60008000008000008F00008F00009000009070009F7F00907F0000008000007F\
00008000008000008000008000008F00008000008F00008000008F00008000008F
00008F00008FA08F00AF9000AF8F00AF9000AF8F00AF9000AF9000B06F0030005F3F006030005F\
3F006030005F3F005F30005F00008F00008000008F00008000008F302F5FAF9000
A08F00AF9000A08F00A08F0000007F00007F00007000007F00007000007000006F00007000006F\
907F008F70008F7000800000000060000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000D0C00000CF0000D00000\
0000FF0000F00000FF0000F00000FF0000F05000AFF05000F0FF00F0F000F0FF00
F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F00000FF0000F00000\
FF0000F00000F000000000EF0000E00000DF0000DF0000D00000D00000CF0000CF
0000C0400080CF0000CF0000C00000C0000000009F00009F00008F00008FA09F00A090009F8F00\
908000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F5F007000007F00000000B00000B00000C00000CF0000DFBF0000
CF0000CF00000000F00000EF0000F04F4FA0EFF000EFF000EFFF00EF0000F00000F00000FFAF00\
F0FF0000006F0000607000007000007F00007F00008000007F0000800000800000
8000008000008000008000008000008000008F00002F004F30004F00006F00007000006F000070\
00006F00007000006F00007000006F00007000006F9070008F6F009070008F6F00
9070008F6F009070008F6F009070008F6F009000008F00008F00008F00008F00008F00008F0000\
8F00008F00008F00008F00008000008F000080000000006F00006000006F00006F
8000008000008F00008F00009000009070009F7F009F7F009F80009F7F00A080009F800000008F\
00008000008F00008F00008F00008F00009000008F00009000008F00009000008F
00009000008F3F0060AF0000B00000AF0000B00000B00000B00000B00000B00000B00000B00000\
B00000B00000B00000B00000AF0000B00000AF0000B00000AF0000B00000AF9000
B09F00AF9000AF9F00AF9000AF900000008000008000007F00007F00007F00007F000070000070\
00006F9F8000907F00907F00000060800000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000404090C00000D00000\
D000000000FF0000F00000FF0000FF5050AFEFFF00F0FF00F0FF00F0FF00F0FF00
FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0FF00FFFF00FF0000FF0000\
F00000FF0000F000005000A00000E00000EF0000DF0000DF0000D00000D00000CF
0000CF0000C0D00000CF0000CF0000C00000C0000000009F000090AFA000AFA000A09F00A09000\
908F002F0040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F6F008000000000AF0000BF0000BF0000CF3F0090C00000
C00000CF00000000EF0000F00000EF0000F00000F0000060000060700000700000700000700000\
7F00007F00007F00007F0000806F008060008F6F008060008F6F008F6F008F6F00
8F6F0030204F00006F00007000006F00007000006F00007000006F00007000006F00007000006F\
0000700000700000700000700000700000700000709070009F70009070009F7F00
9070009F70009070009F70009070009F700030204F00007000006F00007000006F00007000006F\
00007000006F00007000006F30004F9000009000008F00008F00008F00008F0000
800000800000806000806000907F009F7F009F7F009F80009F8000A080009F8000A08F00302F5F\
00008F00008F00009000008F0000900000903F0060AF0000AF0000AF0000B00000
AF0000B00000B00000B00000B00000B00000B00000BF0000B00000B00000B00000BF0000B00000\
B00000B00000B00000B00000B00000B00000B00000B00000B00000B00000B09F00
AF9F00B09F00AF9000AF9F00AF9000AF0000A00000AF0000A00000A0000000007F00007F000070\
00007F9F80009F8000907F00907F000000600000607F0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000BFBF00C0D000\
CF0000D00000DFEF00DFEF00E0F000EFF000F0FF00EFF000F0FF00F0FF00F0FF00
F0F000F0FF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F0005050AF0000F00000FFF00000\
FF0000F00000FF0000F000000000EF0000E00000E00000DF0000DF0000D00000D0
0000CF0000C00000BFD00000CF0000CF0000BF00000000A03F3060B0AF00AF9F00A09F009F9000\
9F8F00000060000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000707000802F000000B00000BF0000C00000CFBF0000BF0000
CF00006F60002F004F6F00007000007000007000007000007F00007F000000006F00006F00006F\
00006F00006F8F6F008F6F008F6F008F6F008F6F009070008F6F009070008F6F00
90700000007000007000007000007F00007000007F00007000007F00007000007F00007000007F\
00007F00007F00007F00007F00007F00007F302F50A07F009F7F00A07F009F7F00
A07F009F7F00A07F009F7F00A07F009F7F00302F5000007000007F00007000007F00007000007F\
00007000007000007000007000007000007000007000007000007000007000006F
9000009000009000008F00009000008F00008F00008000008F6F00000060AF8F00A08F00AF9000\
000090000090AF0000B00000B00000B00000B00000B00000B00000BF0000B00000
BF0000B00000BF0000BF0000BF0000BF0000BF0000BF0000BF0000BF0000BF0000BF0000BF0000\
BF0000BF0000BF0000BF0000BF0000BF0000B00000BF0000B00000BF0000B03000
BFA000B0A000B0A000B09F00B09F00AF3000B00000AF0000AF0000AF0000AF0000A00000AF0000\
A000003F00509F8F009F8F00908000302F4F00006000006F000060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000BFC000\
C0D000CF0000D00000DFEF00DFEF00EFF000EFF000F0FF00F0FF00F0FF00F0FF00
FFFF00F0FF00FFFF00F0FF00FFFF00FFFF00FFFF00FFFF005050AF0000FF0000FF0000F05000AF\
FF0000FF0000F00000FF0000F000000000EF0000E00000E00000DF0000DF0000D0
0000D00000CF0000CF0000BFD00000CF0000C00000BF0000BFB000B0AF00AFA000A09F00A09F00\
9F90009F8F00000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000606F008080000000AF0000BF0000C00000D0605F00706000
701F0070000070000000006F00006000006F00006000006F00006F00006F00006F00006F00006F\
00007000006F8F6F008F6F009070008F6F00907000907000907000907000907000
2F205000007F00007000007F00007F00007F00007F00007F00007F00007F00007F00007F00007F\
00008000007F00008000007F00008000007F302F5F9F7F00A080009F7F00A08000
9F7F00A080009F7F00A080009F7F00A0800000007F00007F00007F00007F00007000007F00007F\
00007F00007000007F00007000007F00007000007F000070000070000070000070
00006FA07F009F70009F7F009070009F70009000009000008F00009000008F00008F0000800000\
8F0000000060000060B00000BF0000B00000BF0000B00000BF0000B00000BF0000
BF0000BF0000BF0000BF0000BF0000C07000BFA000BFAF00BFA000C0AF00BFA000BFAF0000009F\
0000A000009F0000A000009F0000A000009F00009F00009F00009F00009000009F
B0A000BFA000B0A000B0A000B09F00B0A000AF0000B00000AF0000AF0000AF0000AF0000A00000\
AF0000A00000A000009F00009F8F0090800000006F00006F00006F000060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
BFC000C00000CF0000D00000DFEF00DFEF00EFF000EFF000F0FF00EFF000F0FF00
F0FF00F0FF00F0F000FFFF00F0FF00FFFF004F4FA00000FF0000F00000FF0000F00000FF0000F0\
FF0000F00000FF0000F00000F00000EF00000000EF0000E00000E00000DF0000DF
0000D00000CF0000C00000C0CF0000CF0000C00000C00000BFB000B0AF00AFA000AFA000A09000\
9F9000908F00000060000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006F6F007F7F006F60006F60007060007060007F600000006F
00006F00006F00006F00006F00006F00006F00007000007000007000007000007000007000007F\
000070302F509070009070009070009F7F009070009F7F009070009F7F009F7F00
302F5F00007F0000800000800000800000803F005F30005F30005F30005FA00000A00000A00000\
A00000A00000A00000A00000A00000AF0000A00000AF0000A00000AF0000A05F00
AF5F00A08000AF8F00A08000AF8000A08000000080000080000080000080000080000080000080\
00007F00008000007F00008000007F00008000007F00007F00007F00007F00007F
00007FA07F00A08000A07F00A08000A07F00A080009F7F00A07F009F700000007000006F9F0000\
9000009000008F00008F000080000000006FBF0000C00000BF00004000703F0070
0000AF0000A00000AF0000A0403F70C0AF00C0B000C0AF00C0B000C0AF00C0B000C0AF000000AF\
0000A00000AF0000A00000A00000A00000A00000A00000A000009F0000A000009F
BFAF00BFAF00BFAF00BFA000BFAF00B0A000000090000090000090AF0000B00000AF0000AF0000\
AF0000AF0000A00000A000009F0000A00000302F4F00007000006F00006F000060
000000000000000000000000000000000000000000000000000000000000000000000000000000\
B0B000C0CF00CF0000D00000D00000DFEF00E0F000EFFF00EFFF00F0FF00F0FF00
F0FF00F0FF00FFFF00F0FF00FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000F0FF0000F00000FF0000F00000FF0000F000000000EF0000E00000E00000DF
0000DF0000D00000D00000C00000C0CF0000CF0000C00000BF7F00B0AF00B0AF00AFA000AF9F00\
9F90009F0000900000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006F00006F00007060007060007F60007F600080600000006F00006F
00006F00007000006F00007000006F00007000007000007F00007000007F00007000007F00007F\
00007F907000907F009070009F7F00907F009F7F00907F009F7F009F7F009F8000
9F00009F00009F0000A000009F0000A000009F0000A000009F0000A00000A00000A00000A00000\
A00000A00000AF0000A00000AF0000A00000AF0000A00000AF0000A00000AF0000
A00000AF0000A00000AF0000A00000AF0000AF0000AF0000A00000AF0000AF0000AF0000A00000\
AF0000AF00003F005F30005000008000008000008000007F00008000007F000080
00007FAF8000A08000AF8000A08000A08000A07F00A08000A07F00A08000A07F0000007000006F\
00007000006F00007000006F9000009000009000008F00008F00000000600000AF
0000AF0000AF0000AF0000AF0000AFCFB000C0B000CFB000C0B000CFB000C0B000CFB0000000AF\
0000AF0000AF0000AF0000A00000AF0000A00000AF0000A00000A00000A00000A0
BFAF00BFAF00BFAF00BFAF00BFA000BFAF0000009000009F00009000009000008F00009000008F\
00008FAF0000AF0000A00000A000009F0000A000009F000000007000006F00006F
0000607F0000000000000000000000000000000000000000000000000000000000000000000000\
0000000000CFC0CF00C00000D00000D00000DFEF00E0EF00EFF000EFFF00F0FF00
EFF000F05000F00000F00000F000005000AF0000FF0000FF0000F00000FF0000F00000FF0000F0\
0000FF0000F0FF0000F00000FF0000F00000F000004F009F0000EF0000E00000E0
0000D00000D00000CF0000CF0000C0D00000C00000C00000BF0000BFB000B0AF00AFAF00A09F00\
A09F009F00009F0000000060000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000606F0000\
6F00007000007060007F60007F60008060007F6000806F002F204F000070000070
00007F00007000007F00007000007F00007F00007F00007F9000009000009000009000009F0000\
902F009F7F00907F009F80009F7F009F80009F7F00A080009F8000A080009F2F00
A00000A00000A00000A00000A00000A00000A00000A00000AF0000A00000AF0000A00000AF0000\
AF0000AF0000AF0000AF0000AF0000AF0000AF0000B00000AF0000B00000AF0000
B00000AF0000B00000AF0000B00000AF0000B00000AF0000B00000AF0000B00000AF0000B00000\
AF0000B00000AF0000B00000AF0000B00000AF0000B00000AF0000B000003F005F
000080AF8F00AF8F00AF8000AF8F00AF8000AF8F00AF8000AF8F00AF8000AF800000007000007F\
00007000007000007000007000007000007000006F0000709000009000008F0000
8F00008060000000B00000B00000B0CFB000CFBF00CFB000CFBF00CFBF00CFBF00CFBF0040407F\
0000B00000B00000AF0000B00000AF0000AF0000AF0000AF0000AF0000AF0000A0
403F70C0B000C0B000BFB000C0B000BFAF0000009F00009F00009F00009000009F000090000090\
00008F00009000008FAF6F00A00000A00000A00000A000009F00009F000000006F
00006F807000804F00000000000000000000000000000000000000000000000000000000000000\
0000000000000000CF0000EFCF0000D00000D09F00E0F0004F4FAF5050AF0000FF
0000FF0000FFF00000F00000FF0000F000000000FF0000FF0000FF0000FF0000FF0000F00000FF\
0000FF0000FF0000F0FF0000F00000FF0000F00000F000000000EF0000EF0000E0
0000E00000D00000D00000CFDFDF00D0D000CF0000C00000C00000BF7F00B0B000AFA000AFA000\
30305FA0000090000000006F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006000006F00006F0000000060000060\
00006F7060007F60007F60008060007F6000806F00806F0000007F00007000007F
00007F3000508F00008F00008F00009000008F00009000009000009000009000009F0000900000\
9F2F009F7F009F80009F8000A080009F8000A080009F8000A08F00A08000A02F00
A00000A00000A00000A00000A00000AF0000A000003F00603000603F00603000603F0060000090\
00009000009000009000009000009F000090B09000AF9000B06000AF6000B06000
AF0000B00000AF0000B00000B00000B00000AF0000B00000B00000B00000B00000B00000B00000\
B00000B00000B00000B00000B00000B00000B00000B00000B00000AF0000B00000
AF0000B09000AF8F00B08F00AF8F00B08F00AF8F00AF8F00AF8F00AF8F00AF800000007F00007F\
00007F00007000007F00007000007F00007000007000006F00007000006F00006F
30004F902F008F6F008F6F00806000806F00CFBF00D0BF00CFBF00D0C000CFBF00D0C000CFBF00\
0000BF0000B00000B00000B00000B00000AF0000B00000AF0000B00000AF0000AF
403F70CFBF00C0B000C0B000C0B000C0B0003F306F0000A000009F00009F00009000009F000090\
00009000008F000090AF9F00AF9F00A09000A00000A00000A000009F00009F0000
90000000006F8070007F6F00000000000000000000000000000000000000000000000000000000\
0000000000000000000000D00000EFCF0000D000000000F00000FF0000F00000FF
0000FF0000FF4F00A0F00000F00000F00000F000000000FF0000FF0000FF0000F00000FF0000F0\
0000FF0000F00000FF0000F0FF0000F00000F00000F00000F000000000E00000E0
0000DF0000DF40408FDFDF00D0D000D0D000CF0000CF0000C00000BF0000B0AF00B0AF00AFA000\
00008000007F9F0000900000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F00006F000000006F00006000006F00006F00006F\
7F6000806F007F6000806F00806F008F7000804F008F00008F00008F00008F0000
9000008F00009000009000009000009000009F00009000009F00009F00009F00009F00009F0000\
9F8000A08F009F8000A08F00A08000A08F00A08F00A08F00A08F00AF8F00302F60
00009000009000009F00009000009F00009000009F00009000009F00009000009F00009F00009F\
00009000009F00009F00009F00009F00009FB09000B09F00B09000B09F00B09000
B09F00B09000B09F00B09000B09F00B0900000009F00009000009F00009000009F00009000009F\
3F006040006FB00000BF0000B00000BF0000B00000BF0000B00000BF0000B00000
BF0000B09000BF9F00B09000B09000B09000B09000B09000B09000AF8F00B090003F005F000080\
00008000008000007F00008000007F00007F00007F00007F000070000070000070
00007000006F3F2F4F9F70009070008F6F009070008F6F008F6F00D0C000DFCF00D0C000D0CF00\
0000BF0000C00000BF0000BF0000BF0000BF0000B00000BF0000B00000BF0000B0
40407FCFBF00CFC000CFBF00CFBF00C0B0004040700000A00000A000009F0000A000009F00009F\
00009000009F000090B0A000AF9F00AFA000AF9F00000080A00000A000009F0000
9F0000900000908000807000000060000000000000000000000000000000000000000000000000\
0000000000000000000000000000D00000F0CF0000DF00000000F00000FF0000FF
0000FF0000FF0000FFF00000F00000F00000FF0000F000000000FF0000FF0000FF0000FF0000FF\
0000F00000FF0000FF0000FF5000A0FF0000F00000FF0000F00000F000000000E0
0000E04F4F90E0E000DFDF00DFDF00D0D000D0D000CF0000C00000BF0000BFBF00B0AF00B0AF00\
0000800000809F00009F0000000060000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000060000000006000006000006F00006000006F00006F00007020204F\
806F007F6000806F008000008F00008000008F00008F00008F00008F0000900000
8F00009000009000009000009000009F000090000030006000008F00009000008F00009000008F\
A080009F8000A08F009F8000A08F00A08F00A08F00A08F00A08F00A08F0000009F
00009000009F00009000009F00009000009F00009F00009F00009000009F00009F00009F00009F\
00009F00009F00009F00009F00009F3F306FB09F00B09000B09F00B09000B09F00
B09000B09F00B09000B09F00B09000B09F0000009F00009F00009000009F00009F00009F000090\
00009F00009000009F00009000009F00009000009F0000900000903F0060BF0000
B00000BF9F00B09000BF9F00B09000B09F00B09000B09000B09000B09000B09000B03000B00000\
B00000B00000B00000AF00003F005F00007F00008000007F00007F00007000007F
00007000007000006FA080009F7F009F7F009F70009F70008F6F008F70008F000000006FD0CF00\
DF0000D00000DF0000D00000D00000D00000D00000D00000D00000CF0000D00000
CF0000D00000CF0000CF8000CFBF00CFBF00403F700000AF0000A00000AF0000A00000A000009F\
00009F00009F00009FB0A000B0A000AF9F00AFA00030305F00008000007FA00000
9F00009F00009000008F7F00807000000060000000000000000000000000000000000000000000\
0000000000000000000000000000CF0000DFCF0000CF0000D000000000F00000FF
0000F00000FF0000FF0000FFF00000F00000F00000FF0000F000000000FF0000FF0000FF0000F0\
0000FF0000F00000FF0000F00000FFF00000FF0000F00000F00000EF00000000EF
E0EF00E0EF00E0E000E0E000DFDF00DFDF00D0D000CF0000C00000C00000BF0000B0B00000008F\
00008F00008000007F900000900000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000060600000006F00006000006F00006F00006F00006F000070000070000070000070800000\
8000008F00008000008F00008F00009000008F0000900000900000900000900000
30006000008F00009000008F00009000008F0000900000900000900000900000900000903F306F\
A08F00A08F00A08F00A09000A08F00AF9000A08F00AF9000A08F00AF900000009F
00009F00009F00009F00009F0000A000009F0000A000009F0000A000009F0000A000009F0000A0\
00009F0000A000009F0000A00000A03F3070B09F00BF9F00B09F00BF9F00B09F00
BF9F00B09F00BF9F00B09F00BFA0003F306F0000A000009F0000A000009F0000A000009F0000A0\
00009F00009F00009F0000A000009F00009F00009F00009F00009F00009F000090
00009FBF9F00BF9F00BF9F00BFA000BF9F00BF9F00BF9F00BF9F00BF9F00BF9F00B03000BF0000\
B00000BF0000B00000BF0000B00000B00000B00000B0000000008000008000007F
00007F00007F00007F302F4FAF8F00A08000A08000A07F00A080009F7F009F7F00900000900000\
2F0040DF0000DF0000DF0000DF0000DF0000D00000DF0000D00000DF0000D00000
DF0000D00000D00000D00000D00000CF0000CF0000CF00000000B00000AF0000AF0000A00000AF\
0000A00000A000009FBFB000B0AF00B0AF00AFA0003F3F6000008F000080000080
0000809F00009F00009080008F8000000060000000000000000000000000000000000000000000\
0000000000000000000000000000000000D00000E0D00000D00000DF00000000F0
0000FF0000FF0000FF0000FF0000FFF00000F00000F00000FF0000F000000000FF0000FF0000FF\
0000FF0000FF0000F00000FF0000F00000FFF00000FF0000F00000F00000EFF000
EFF000E0EF00E0EF00DFE000E0E000DFDF00DFDF00CFCF00CF0000C00000C0000000009F00009F\
00008F00008F00007F9F0000900000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000600000
6F5F0000006F00006000006F00006F00007000006F0000702F004F8000007F0000800000800000\
8F00008000008F00008F00009000008F500000008F00008F00008F00008F00008F
00008F00009000008F00009000009000009000009000009000009000009F00009000009F302F60\
A08F00A08F00A08F00A08F00A09000A08F00AF9000A08F00AF9000A08F0000009F
00009F0000A000009F0000A000009F0000A000009F0000A000009F0000A000009F0000A00000A0\
0000A00000A00000A00000A00000A03F306FB09F00B09F00BF9F00B09F00BF9F00
B09F00BFA000B09F00BF9F00B09F003F307000009F0000A000009F0000A000009F0000A000009F\
0000A000009F0000A000009F0000A000009F0000A000009F00009F00009F00009F
00009FBFA000BF9F00BFA000BF9F00BFA000BF9F00BF9F00BF9F00BF9F00BF9F00BF3000BF0000\
BF0000B00000BF0000B00000BF0000B00000B00000B00000B00000B00000B00000
3F005000008000007F3F2F50AF8000AF8F00A08000AF8000A08000A08000A07F00A0800000006F\
30004F8F00008F0000DF0000E00000DF0000DF0000DF0000DF0000D00000DF0000
D00000DF0000D00000D00000D00000D00000CF0000CF0000CF0000CF0000C00000CF0000C00000\
0000AF0000A00000A0BFB000BFB000B0AF00B0AF00AFA00000008F000080000080
00008000007F900000908F008F80008F7F0000005F000000000000000000000000000000000000\
0000000000000000000000000000000000000000D00000E0D00000D000004F00AF
0000F00000FF0000F00000FF0000FF0000FFF00000F00000F00000FF00004F00A00000FF0000F0\
0000FF0000F00000FF0000F00000FF0000F0F00000F00000F0FF00F0F000F0F000
EFEF00EFEF00E0E000E0E000DFDF00DFDF00D0D000D0D000CF0000CF0000BF0000BF0000000090\
0000900000800000800000709F0000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006000006F60006F6000
00006F00006F0000700000700000707F00008000007F00008000008000008F00008000008F0000\
8F00009050008F7F00907F008F7F0090800000008F00009000008F000090000090
00009000009000009F00009000009F00009000009F00009F00009F00009F00009F00009F3F306F\
A08F00AF9000A09000AF9000A09000AF9000AF9000AF9000AF9000AF9F000000A0
0000A00000A00000A00000A00000A00000A00000AF0000A00000AF0000A00000AF0000A00000AF\
0000A00000AF0000A00000AF0000A0403F70B0A000BFA000B09F00BFA000BFA000
BFA000BFA000BFA000BFA000BFA0003F30700000AF0000A00000AF0000A00000AF0000A00000A0\
0000A00000A00000A00000A00000A00000A00000A00000A00000A00000A000009F
0000A0BFA000C0A000BFA000C0A000BFA000C0A000BFA000C0A000BF9F00C0A0003F306000009F\
00009000009F000090400060BF0000C00000BF0000BF0000B00000BF0000B00000
B00000B00000B00000B00000B09000AF8F00B09000AF8F00AF8F00AF8000AF8F00A08000000070\
00006F00006F00006F9000008F000000006FE00000E00000DF0000E00000DF0000
E00000DF0000E00000DF0000DF0000D00000DF0000D00000D00000CF0000D00000CF0000CF0000\
C00000CF0000C00000C0BF00BFB000BFB000B0AF003F3F6F00009000009000008F
00008F0000800000809F8F00908F008F800000006F000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000DF0000EFD00000D00000
0000FF0000F00000FF0000FF0000FF0000FF0000FFF00000FF0000F00000FF00000000FF0000FF\
0000FF0000FF0000FF0000FF0000F00000FFF0FF00FFFF00F0FF00FFFF00F0F000
F0F000EFEF00EFEF00E0E000E0E000DFDF00DFDF00D0D000D0D000CF0000C00000BF000000009F\
000090000090000080000080900000900000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000600000605F006F60006F600000006F
00006F2F00507F00007F00007F00008000007F00008000008000008F00008000008F70008F7000\
8F7F008F7000907F008F7F00907F00907F0000009000008F000090000090000090
00009000009000009000009F00009000009F00009F00009F00009F00009F00009F0000A0302F6F\
A09000A09000AF9000A09000AF9000A09000AF9F00A09000AF9F00AF90000000AF
0000A00000A00000A00000AF0000A00000AF0000A00000AF0000A00000AF0000A00000AF0000A0\
0000AF0000AF0000AF0000AF0000AF3F3070BFA000B0A000BFA000BFA000BFA000
B0A000BFA000BFA000BFA000BFA0003F30700000A00000AF0000A00000AF0000A00000AF0000A0\
0000AF0000A00000AF0000A00000A00000A00000AF0000A00000A00000A00000A0
00009FC0A000BFA000C0A000BFA000C0A000BFA000C0A000BFA000C0A000BFA000403F6F000090\
00009F00009000009F000090000090000090000090000090BF0000BF0000BF0000
B00000BF0000B00000B00000B00000B00000AF0000B06000AF8F00AF8F00AF8F00AF8F00302F4F\
00007000006F00006F00006000006F9000008F00008000000000D00000D00000D0
40408FE0DF00DFDF00E0DF00DF9000DF0000D00000DF0000D00000D00000D00000D00000CF0000\
CF0000CF0000CF0000C0BF00C0BF00BFB000BFB0003F306F00009F000090000090
00008000008000007F9F9000908F009080008F0000000060000000000000000000000000000000\
0000000000000000000000000000000000000000000000CF0000DF0000EFD00000
D000000000FF0000F00000FF0000F00000FF0000FF5000AFF00000F00000F00000FF00000000F0\
0000FF0000F00000FF0000F0FFFF00F0FF00FFFF00F0F000F0FF00F0FF00F0FF00
EFF000EFF000E0EF00E0EF00DFE000DFE000D0DF00D0DF00404080CF0000C00000C00000B00000\
00009F00008F00008F00007F9F00008F0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006000006F60006F60007060007060002F204F700000
7F00007F00008000007F00008000008000008F00002F005F00008F000080302F608F7000907F00\
8F7F00907F008F7F00908000908000302F6F00009000009F00009000009F000090
00009F00009F00009F00009F0000A000009F0000A000009F0000A00000A00000A00000A0AF9000\
A09000AF9F00A09000AF9F00AF9000AF9F00AF9000B09F00AF9F003F30700000AF
0000AF0000AF0000AF0000AF0000AF0000AF0000AF0000AF0000B00000AF0000AF0000AF0000B0\
0000AF0000B00000AF0000B00000AF403F7FBFA000BFAF00BFA000BFAF00BFA000
C0AF00BFA000C0AF00BFA000C0AF003F30700000AF0000AF0000B00000AF0000AF0000AF0000AF\
0000AF0000AF0000AF0000AF0000AF0000AF0000AF0000AF0000AF0000AF0000A0
0000AFC0AF00C0AF00C0A000C0AF00C0A000C0AF00C0A000C0AF00C0A000C0AF003F306F0000A0\
00009F00009F00009F00009F00009F00009F00009000009F000090000090000090
400060BF0000BF0000BF0000BF0000B00000BF0000B00000B00000B00000B00000AF8F00B09000\
00007F00007F00007000007000006F00006F00006F30004F8F00008F00000000D0
4F4F90E0E000E0E000E0DF00E0E000E0DF000000D00000CF0000CF40008FDF0000D00000D00000\
D00000D00000CF0000CFCF00C0C000C0BF00BFB00040007000009F00009F000090
00009000008F0000809F90009F9000908F009000008F0000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000D00000E0CF0000
D00000DF00000000FF0000FF0000FF0000FF0000FF0000FFFF0000F00000FF0000F000005000AF\
0000FF5050AFF0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00
F0FF00EFF000EFF000E0EF00E0EF00DFE000DFDF000000C00000C00000BFCF0000C00000BF0000\
00009F00009000008F00008000007F900000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006000000000606F60006F5F006F60006F60007060007000007F0000
7F00008000007F00002F005F00008000008000008000008F00008000008F8F70008F7F008F7F00\
907F008F7F00908000907F009080002F2F6000009F00009000009F00009F00009F
00009F00009F00009F0000A000009F0000A000009F0000A00000A00000A00000A00000A0A09000\
AF9F00A09000AF9F00AF6000AF6F00AF6000AF6F00AF0000B00000AF0000B00000
AF0000B00000AF0000B00000B00000B00000B00000B00000B00000BF0000B00000BF0000B00000\
BF0000B00000BF0000BF0000BF0000BFA000BFAF00BFA000BFAF00BFA000BFAF00
BFA000C0AF00BFA000C0AF00BFAF006F3F503F00700000B00000AF0000B00000AF0000B00000AF\
0000B00000AF0000AF0000AF0000AF0000AF0000AF0000AF0000AF0000A00000AF
0000A0C0AF00C0A000C0AF00C0AF00C0AF00C0A000C0AF00C0A000C0AF00C0A000403F6F00009F\
0000A000009F00009F00009F00009F00009F00009F00009000009F000090000090
00009000009000008F000090BF0000BF0000B00000BF0000B00000BF0000B00000B00000B00000\
B0000030005000007F00007000007000006F00006F00006F00006F0000608F0000
800000EFE000E0E000E0E000E0E000E0E0000000D00000D00000CF0000CF0000CF0000CF0000C0\
0000C0400080D04000CFCF00CFCF00C0BF00C0BF00BF0000BF0000B0000000009F
000090000090000080A09F009F90009F90002F004F8F0000000060000000000000000000000000\
000000000000000000000000000000000000000000000000B0B000BFC0000000EF
CF0000DF00004000A00000FF0000F00000FF0000F00000FF0000FFF00000F00000F00000F00000\
FFFF00F0F000FFFF00F0FF00FFFF00F0F000F0FF00F0FF00F0FF00F0F000F0FF00
F0F000F0F000EFEF00EFEF00E0E0004F4F900000CF0000CF0000C00000C0C00000C00000BF0000\
BF000000009000009000008000007F8F0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000006F00006020204F6F60007060007060007F6F007000007F00007F0000
00008000008000008000008000008F00008000008F00008F00008F00008F907F008F7F00908000\
907F009080009080009F800090800030306F00009F0000A000009F0000A000009F
0000A00000A00000A00000A00000A00000A00000AF0000A00000AF300070AF0000A00000AF0000\
AF0000AF0000AF0000AF0000AF0000B00000AF0000B00000AF0000B00000B00000
B00000B00000B00000B00000B00000B00000BF0000B00000BF0000B00000BF0000BF0000BF0000\
BF0000BF0000BF0000C00000BF0000C0B000BFAF00C0AF00BFAF00C0B000C0AF00
C0B000C0AF00C0B000C0AF00C0B000C03F00C00000C00000C00000C00000CF0000C00000CF0000\
C00000CF0000C00000CF0000C00000CF000040007F0000B00000AF0000B00000AF
0000AFC0AF00CFB000C0AF00CFB000C0AF00CFB000C0AF00CFAF00C0AF00CFAF00403F700000A0\
0000A00000A00000A00000A000009F0000A000009F0000A000009F00009F00009F
00009F00009000009F000090403F60BFA000C0A000BF6F00BF0000BF0000BF0000B00000BF0000\
B00000B00000B00000B0000000007F00007F00007000007000006F00006F00006F
9F7F008F00008F0000E0E000EFEF00E0E0000000DF0000D00000D00000D00000D00000CF0000CF\
0000CF0000CFD0D000D0D000CFCF00CFCF00C0C000C00000BF0000C00000BF0000
BF0000000090000090A09F00A09F009F900000007F30004F8F0000000000000000000000000000\
000000000000000000000000000000000000000000000000000000BFBF00C0CF00
CFDF00D04F00DF50004F4FAF4F50AF4F4FAF5050AF4F50AFF0FF00F00000FF0000F00000FF0000\
F0AF00FFFF00FFFF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00
F0FF00F0F000F0F000EFEF000000E00000DF0000D00000CF0000CF0000C00000BFC00000C00000\
BF000000009F00009000008F00007F900000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000060\
00006F00006000006F6F5F007060007060007F6F007060007F000000007F000080
00008000008000008000008F00008000008F00008F00009000008F0000908F7F009080008F7F00\
9080009080009080009080009F8F0000009F0000A000009F0000A00000A00000A0
0000A00000A030006FA000009F0000A00000A00000A00000A00000AF0000A00000AF0000A00000\
AF0000AF0000AF0000AF0000B00000AF0000B00000AF0000B00000B00000B00000
B00000B00000B00000BF0000B00000BF0000B00000BF0000B00000BF0000BF0000BF0000BF0000\
BF0000BF0000C00000BF0000C00000BFAF00C0B000BFAF00C0B000C0AF00C0B000
C0AF00C0B000C0AF00C0B000C0AF00C03F00C00000CF0000C00000CF0000C00000CF0000C00000\
CF0000C00000CF0000C00000CF0000CF0000CF0000C00000CF0000CF0000CF0000
C00000CF7F00C0AF00CFB000C0AF00CFB000C0AF00CFB000C0AF00CFAF00C0AF00403F700000A0\
0000AF0000A00000A00000A00000A00000A00000A000009F0000A000009F00009F
00009F00009F00009000009F3F3060C0A000BFA000BFA000BF9F00BFA000BF9F00BF0000B00000\
BF0000B00000B00000B00000B00000AF000000007F00007000007000006F000070
00006F9F7F00907000900000E0E0004F4F900000DF0000DF0000D00000D00000D00000D00000CF\
0000D00000CFDFDF00D0D000D0D000CFCF00CF4000C00000C00000BF0000BF0000
B00000B00000000090AFA0009F9F009F90000000700000708000002F2F40000000000000000000\
000000000000000000000000000000000000000000000000000000B0AF00C0C000
C0D000D0E000D0E000DFEF00DFEF00EFF000EFF000F0FF00F0F000F0FF00F00000F00000F00000\
FF0000F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000F0FF00F0FF00F0FF00
F0F000F0FF00EF4F000000EF0000DF0000DF0000D00000D00000C00000C00000BFCF0000C00000\
BF0000B0000000009000008000007F9000008F0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000006000006F\
00006F20004F6F60007060007060007F6F007F6F00000080000080000080000080
00008F00008F00008F00008F00009000008F000090000090000090000090908000908000908F00\
9080009F8F009080009F8F009F8F0030007030006FA000009F0000A000009F0000
A00000A00000A00000A00000AF0000A00000AF0000A00000AF0000AF0000AF0000AF0000B00000\
AF0000B00000AF0000B00000B00000B00000B00000B00000B00000B00000B00000
BF0000B00000BF0000B00000BF0000BF0000BF0000BF0000BF0000BF0000BF0000BF0000C00000\
BF0000C00000BF0000C00000C00000C0B000C0B000C0B000C0B000C0B000C0B000
CFBF00C0B000CFB000C0B000CFBF00C03F00CF0000C00000CF0000CF0000CF0000CF0000CF0000\
CF0000CF0000CF0000D00000CF0000D00000CF0000D00000CF0000CF0000CF0000
CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF7F00CFB000CFB000CFB000403F700000AF\
0000AF0000AF0000AF0000AF0000A00000AF0000A00000AF0000A00000A000009F
0000A000009F00009F00009F403F6FC0A000C0AF00BFA000C0AF00BFA000C0A000BFA000C0A000\
BFA000400060BF0000BF0000B00000B00000AF0000B0000000007F00007F000070
000070A08000A080009F7F009F7F008000000000E00000DF0000DF0000DF0000DF0000D00000DF\
0000D04F4F90DFDF00DFDF00D0D000D0D0000000BF0000C0CF0000CF0000C00000
C00000B00000B00000AFA000AFA0009F9F00000080000070000070800000000000000000000000\
000000000000000000000000000000000000000000000000000000000000B0BF00
C0CF00CFDF00D0E000DFE000E0EF00E0EF00EFF000EFF000F0FF00F0FF00F00000F00000FF0000\
F00000FF0000FFFF00FFFF00FFFF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00
FF5000F00000F00000EF00000000EF0000DF0000DF0000D00000D00000C00000C00000BFCF0000\
BF0000BF000000009000008F0000803000508F0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000006000006F20004F\
6F00006F1F007060007060007F6F007060007F6F0000008000008000008000008F
00008F00008F00008F00009000008F00009000009000009000009000009F8F8000908000908000\
908F009080009F8F00908F009F90009F00009F00009F0000A000009F0000A00000
9F0000A00000A00000A00000A00000AF0000A00000AF0000A00000AF0000AF0000AF0000AF0000\
B00000AF0000B00000B00000B00000B00000B00000B00000B00000B00000B00000
B00000BF0000B00000BF0000BF0000BF0000BF0000BF0000BF0000BF0000BF0000C00000BF0000\
C000003F008040008F3F008040008FC0B000C0BF00C0B000C0BF00C0B000C0BF00
C0B000CFBF00C0B000CFBF00C0B000CF3F00C00000CF0000CF0000CF0000CF0000CF0000CF0000\
CF0000CF0000CF0000CF0000D00000CF0000D00000CF0000D00000CF0000CF0000
CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF0000\
40007F0000AF0000AF0000AF0000AF0000A00000AF0000A00000AF0000A00000A0
0000A00000A000009F00009F3F306FC0AF00BFA000C0AF00BFA000C0AF00BFA000BFA000BFA000\
BFA00000008F00008F00008FBF0000B00000B00000B00000B00000AF00003F0050
0000703F3050A08000A080009F80009F80002F00400000600000DF0000E00000DF0000DF0000DF\
0000DFDFDF00DFE000D0DF00DFDF00D0D0000000C00000BF0000C00000B0CF0000
BF0000BF0000B00000B00000A0A000A0A00000008000007F0000708F0000706F00000000000000\
000000000000000000000000000000000000000000000000000000000000B0B000
BFBF00CFD000CFDF00D0E000DFE000E0EF00E0EF00EFF000F0FF00F0FF00F0F000F00000F00000\
FF0000F00000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF004F4FA05000AF
F00000F00000EF0000EF00000000E00000E00000DF0000D00000CF0000CF0000C00000BFC00000\
C00000B0000000009F00008F0000800000708F0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000006000006020004F6F0000700000\
7000007F6F007060007F6F007F6F007F70007F6F0000008F00008000008F00008F
00009000008F00009000009000009000009000009F00009F00009F2F2F6F908F00908F009F8F00\
908F009F90009F8F009F90009F8F00A000009F0000A000009F0000A00000A00000
A00000A00000AF0000A00000AF0000A00000AF0000AF0000AF0000AF0000B00000AF0000B00000\
AF0000B07000B07000B0AF00B0AF00BFAF00B0AF00BFAF00B0AF00BFB0000000C0
0000C00000BF0000C00000C00000C00000C00000C00000C00000CF0000C00000CF0000C00000CF\
0000CF0000CF0000CF0000CF0000CFCFBF00C0BF00CFBF00C0BF00CFBF00CFBF00
CFBF00CFBF00CFBF00CFBF00CFBF000000CF0000CF40008F40008F400080D00000CF0000D00000\
CF0000D00000D00000D00000D00000D00000D00000D00000D00000D00000D00000
D00000CF0000D00000CF0000D00000CF0000D00000CF0000D00000CF0000D00000CF0000D00000\
CF0000D00000CF0000CF0000CF0000CF00000000AF0000AF0000AF0000AF0000A0
0000AF0000A00000A00000A0403F70C0AF00C0B000C0AF00C0AF00C0AF00C0AF00C0AF00C0AF00\
BFA00000009000009000009000008F0000903F005FBF0000B00000B00000B00000
B00000AF0000AF9000A08F00A08F00A08000A080000000608F00000000E00000E00000E00000E0\
0000DFE0EF00DFE000DFE000DFDF004F4F8F0000CF0000CF0000C00000C00000BF
40007FBF0000BF0000B00000B00000A0A00000008F00008000007F00006F800000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000D0C0CF00D0DF00D0E000DFEF00DFEF00EFF000EFF000F0FF00F0FF00FFFF00F00000FF0000\
F00000FF0000FF0000FFFF00FFFF00FFFF005050AF0000FF0000F00000FF0000F0
F00000F00000F00000F00000EF00000000E00000E00000DF0000D00000CF0000CF0000C0CF0000\
C00000BF000000009F0000900000809F9F008F0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006060000000606F00006F00007000006F0000\
00007F7060007F6F00706F007F6F007F6F0080700000008000008F00008F000090
00008F00009000009000009F00009000009F00009F30006F8F0000902F00908000908F00908F00\
9F8F00908F009F90009F8F009F90009F0000A000009F0000A00000A00000A00000
A00000A00000A00000AF0000A00000AF0000AF00003F007F0000B00000BF0000BF0000BF3F307F\
B0AF00B0AF00B0AF00B0AF00B0AF00B0AF00BFB000B0AF00BFB000B0AF000000C0
0000C00000C00000C00000C00000C00000CF0000C00000CF0000C00000CF0000CF0000CF0000CF\
0000CF0000CF0000D00000CF0000D0C0BF00CFC000C0BF00CFBF00CFBF00CFC000
CFBF00CFC000CFBF00CFC000CFBF000000CF0000CF0000CF0000CF0000CF0000CF0000CF0000CF\
0000CF0000CF0000CF0000CF0000CF0000CF0000CF400080D00000D00000D00000
CF0000D00000D00000D00000CF0000D00000CF0000D00000CF0000D00000CF0000D00000CF0000\
D00000CF0000D00000CF0000CF0000CF0000CF0000CF0000CF00004000700000AF
0000A00000AF0000A00000A03F306FC0B000C0AF00C0B000C0AF00C0AF00C0AF00C0AF00BFA000\
C0AF0000009000009000009000009000008F00008F00008F3F005FB00000B00000
AF0000B00000AF9000AF9000A08F00A08F00A08000302F4F0000600000607F00004F009F0000DF\
E0EF00DFE000E0E000DFE000DFE0000000CF0000CF0000C00000CF0000C00000C0
0000BF0000B0BF0000BF0000AF0000AF000000008F00008F000070000070800000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00008F0000DF0000E0D0E000D0E000DFEF00E0EF00EFF000EFF000F0FF00F0FF00F0FF00F00000\
F00000F00000FF00004F4FA00000FF0000F00000FF0000F00000FF0000F00000F0
4F00A0F00000F00000F00000EF00000000EF0000DF0000DF0000D00000D00000C00000C00000B0\
C00000BF00003F006F00008F000080909000900000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006060000000606F00006F000070000000007000007F\
2020507F6F007F6F007F70007F6F008070007F7000000090000090000090000090
00009F00009000009F2F006F9000008F0000900000900000900000902F009F8F00908F009F9000\
9F8F009F90009F9000A09F009F9000A03000A00000A00000A00000A0000030007F
3F007F0000B00000BF0000B00000BF0000BF0000BF0000BF0000C00000BF0000C00000BF3F3F80\
B0AF00BFB000B0AF00BFB000BFB000BFB000BFB000BFB000BFB000BFB0000000C0
0000CF0000CF0000CF0000CF0000CF0000CF0000D00000CF0000D00000D00000D00000D00000D0\
0000D00000D00000D00000DF0000D0CFC000CFC000CFC000CFC000CFC000CFC000
D0C000CFC000D0C000CFC000D0CF000000D00000D00000D00000D00000D00000D00000D00000D0\
0000D00000D00000D00000D00000D00000D00000CF0000D00000CF0000CF0000CF
4F408FD0CF00D08F00D00000D00000D00000D00000D00000D00000D00000D00000D00000D00000\
D00000D00000D00000D00000D00000D00000CF0000D00000CF0000CF0000CF0000
CF0000CF00004000700000A0403F70C0B000CFB000C0AF00CFB000C0AF00C0B000C0AF00C0B000\
C0AF0000009F00009000009F00009000009000008F00009000008F00008F000080
BF0000B00000B00000AF9000B09000AF9000AF9000A08F0000007000006F00006F800000EF0000\
E00000E0EF00E0EF00E0EF004F4F900000D00000D00000D00000CF0000CF0000C0
0000C040407FC0C000BF0000BF0000AF0000AF000000008F000080000070808000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000D00000E00000EFDF0000DF00000000FF4F4FAF5050AF4F50AF5050AF4F50AF5000AF\
F00000FF0000F00000FF00000000FF0000FF0000FF0000FF0000F00000FF0000F0
0000FFF00000F00000F00000F00000EF00000000E00000DF0000DF0000D00000CF0000C00000BF\
C00000BF0000B00000000090A0A0009F9F009000008F0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000605F006060000000606F00006F000000007F00007000007F00007F\
2F205F706F007F6F007F6F008070007F70002F2F6000008F0000900000902F006F
8F00008F00008F00008F00008F00009000008F0000900000900000900000908F009F9000908F00\
9F90009F90009F90009F9000A09F003030700000B00000B00000B00000B00000BF
0000B00000BF0000B00000BF0000BF0000BF0000BF0000C00000BF0000C00000BF0000C03F3080\
B0B000B0AF00BFB000B0B000BFB000B0B000BFB000BFB000BFB000BFB0000000CF
0000CF0000D00000CF0000D00000CF0000D00000CF0000D00000D00000D00000D00000DF0000D0\
0000DF0000D00000DF0000DF0000DFCFC000CFC000CFC000CFCF00CFC000CFCF00
CFC000D0CF00CFC000D0CF00CFCF000000DF0000D00000DF0000D00000DF0000D00000DF0000D0\
0000DF0000D00000D00000D00000D00000D00000D00000D00000D00000CF0000D0
40408FDFD000D0CF00D0CF00D0CF00D0CF00D0CF00D0CF00D08F00D00000D00000D00000D00000\
D00000D00000D00000D00000D00000CF0000D00000CF0000D00000CF0000CF0000
CF0000CF0000CF0000CF0000C03F00CFB000C0B000CFB000C0B000CFB000C0AF00C0B000C0AF00\
C0B00000009F00009F00009F00009F00009000009000009000009000008F00008F
00008F00008FB00000B09F00AF9000B09F00AF9000AF900000007000007000006F00006F000060\
EF0000E00000E00000DF9F000000DF0000D00000D00000D00000D00000CF0000C0
0000BFCFCF00C0C000C0C000BF0000B00000AF0000A0000000008000007F8080007F0000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000A000000000D00000EFD00000DF0000DF00000000FF0000F00000FF0000F00000FF0000FF\
F00000F00000FF0000F000000000FF0000F00000FF0000F00000FF0000F00000F0
0000F00000F0F00000F00000EF0000EF00000000E00000E00000D00000D00000CF0000C00000BF\
C00000BF0000B00000303060A0A0009F9000909000800000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006060006060006000006F000020004F00007F00007F00008000007F000080\
706F007F70007F70008070007F7000807F002F2F600000908000008F00008F0000
8F00008F00009000008F00009000009000009000009000009F00009000009F90009F90009F9000\
9F9000A09F009F9F00A09F00A09F003F3F800000B00000BF0000B00000BF0000BF
0000BF0000BF0000C00000BF0000C00000BF0000C00000C00000C00000C00000CF0000C03F3F8F\
B0B000BFB000BFB000BFBF00BFB000C0BF00BFB000C0BF00BFBF00C0BF000000D0
0000D00000D00000DF0000D00000DF0000D00000DF0000DF0000DF0000DF0000DF0000DF0000E0\
0000DF0000E00000DF0000E00000E0D0CF00CFCF00D0CF00CFCF00D0D000D0CF00
D0D000D0CF00D0D000D0CF00D0D0000000DF0000E00000DF0000E00000DF0000E00000DF0000DF\
0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000D00000DF0000D0
DFD000DFD000DFD000DFD000DFD000D0D000DFD000D0CF00DFD000D0CF00DFD0000000CF0000CF\
0000C04F008F400080DF0000D00000DF0000D00000D00000D00000D00000D00000
D00000CF0000D00000CF0000CF4000CFBF00CFBF00CFB000CFBF00C0B000CFBF00C0B000CFB000\
C0B0000000A000009F0000A000009F00009F00009000009F00009000009000008F
00009000008F00008FB09F00B0A000B09F00B09F00AF9000AF900030004F00007000006F00006F\
000060EF0000E00000E00000E00000E000000000DF0000DF0000D00000D00000CF
D0D000CFCF00CFD000C0C00040407FB00000B00000A0000000008F00007F8F8F007F0000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000009F0000D00000E00000F0D00000DF00000000FF0000FF0000FF0000FF0000FF0000FF\
4F00AFFF0000F00000FF0000F000000000FF0000FF0000FF0000F00000FF0000F0
0000FF0000F0FF0000F00000F00000EF00005000A00000E00000DF0000D00000D00000C00000C0\
C00000C00000B03F00B0B000A0A000A0A0009090008F0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000606060006000006F000000007000007F00007F00007F00007F000080000080\
7F70007F6F007F70007F7000807F008070008000008000008F00008F00008F0000
8F00008F00008F00009000008F00009000009000009000009000003000709090009F90009F9000\
9F9F009F9000A09F009F9F00A09F0030307F0000BF0000BF0000BF0000BF0000BF
0000BF0000C00000BF0000C00000C00000C00000C00000CF0000C00000CF0000C00000CF3F3F8F\
BFB000BFB000BFBF00BFB000BFBF00BFBF00C0BF00BFBF00C0BF00BFBF000000DF
0000D00000DF0000D00000DF0000DF0000DF0000DF0000DF0000DF0000E00000DF0000E00000DF\
0000E00000E00000E00000E00000E0CFCF00D0D000CFCF00D0D000D0CF00D0D000
D0CF00D0D000D0D000D0D000D0D0000000E00000E00000E00000DF0000E00000E00000E00000DF\
0000E00000DF0000E00000DF0000E00000DF0000E00000DF0000DF0000DF0000DF
DFD000DFDF00DFD000DFDF00D0D000DFD000DFD000DFD000D0D000DFD00040408F0000CF0000CF\
0000CF0000C00000CF0000C00000C00000C04F0080D00000D00000D00000D00000
CF0000D00000CF0000D00000CFBF00CFBF00CFBF00CFBF00C0B000CFBF00C0B000CFBF00C0B000\
C0B0000000A00000A000009F00009F00009F00009F000090000090000090000090
00008F00008F00008FBFA000B09F00B09F00B09F00B09F00AF9000AF0000A0000000007000006F\
00006F000060EF0000E00000E00000E00000E00000DF00000000DF0000D00000CF
CFD000CFD000C0CF00C0CF000000B00000AFB00000AF00009F00009090008F80002F0040000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000CF0000DF0000E0D00000D00000DF00000000F00000FF0000F00000FF0000F0\
0000FFF00000FF0000F00000FF00000000FF0000FF0000F00000FF0000F00000FF
0000F00000F04F00A0F00000EF0000F00000EF00000000E00000DF0000D00000CF0000CF0000BF\
0000BFBF3F00BFBF00AFAF00A0AF009F9F009090008F0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000006060006000006F000000007000007F00007F00008000008000008000008000008F\
7F70008070007F70008050008000008000008000008F00008F00008F00008F0000
9000008F00009000009000009000003000700000AF0000AF0000B00000AF9F9F009F9000A09F00\
9F9F00A09F00A09F00A0A000A0A000AFA0000000BF0000C00000BF0000C00000C0
0000C00000C00000C00000C00000CF0000CF0000CF0000CF0000CF0000CF0000D00000CF404090\
BFBF00C0BF00BFBF00C0C000C0BF00C0C000C0BF00C0C000C0C000C0C0000000DF
0000E00000DF0000E00000DF0000E00000E00000E00000E00000E00000E00000E00000E00000EF\
0000E00000EF0000E00000EF0000E0D0D000D0D000DFD000D0D000DFDF00D0D000
DFDF00D0D000DFDF00D0D000DFDF000000E00000EF0000E00000EF0000E00000EF0000E00000EF\
0000E00000EF0000E00000EF0000E00000E00000E00000E00000E00000E00000DF
E0E000DFDF00E0E000DFDF00E0DF00DFDF00E0DF00DFDF00E0DF00DFDF004F4F900000D00000D0\
0000CF0000D00000CF0000D00000CF0000CF0000C00000CF0000C00000C0400080
DF0000D00000D00000D00000D0CF00CFC000D0C000CFC000D0C000CFBF00CFC000CFBF00CFBF00\
CF3F00CF0000C000000000A00000A00000A000009F00009F00009F00009F000090
00009F000090000090B0A000BFA000B0A000BFA000B09F00B0A000AF3000AF0000A000003F0050\
00007000007000006F302F4FE00000E00000E00000E00000DF00000000DFD0DF00
D0DF00CFD000CFD0000000BF0000BF0000B03F0070AF0000A00000909F00909000000060000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000AF00000000D00000E00000EFDF0000DF00004F00AF0000FF0000FF0000FF0000FF\
0000FF0000FFF00000FF0000FF0000FF00000000FF0000FF0000FF0000FF0000F0
0000FF0000F00000FFF00000F00000F00000EF00000000E00000E00000D00000D00000CF0000C0\
40407FC0C000B0BF00B0B000A0AF00A0A000909000900000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006000006000006F400000007000007F00007000007F00008000008000008000008F00008F\
7F70007F4F008000008000008000008000008F00008000008F00008F00008F0000
8F00009000000000A00000AF0000AF0000AF0000AF0000AF0000AF0000B09F90009F9F009F9F00\
A09F009F9F00A0A000A0A000AFA000A0A0000000C00000C00000C00000C00000C0
0000C00000CF0000C00000CF0000CF0000CF0000CF0000D00000CF0000D00000CF0000D00000D0\
C0C000BFBF00C0C000C0BF00C0C000C0C000C0C000C0C000CFCF00C0C0000000E0
0000DF0000E00000DF0000E00000E00000E00000E00000E00000E00000EF0000E00000EF0000E0\
0000EF0000E00000EF0000E00000EFD0D000DFDF00D0D000DFDF00D0D000DFDF00
DFDF00DFDF00DFDF00DFDF00DFDF000000EF0000E00000EF0000E00000EF0000E00000EF0000E0\
0000EF0000E00000EF0000E00000EF0000E00000E00000E00000E00000E00000E0
DFE000E0E000DFDF00E0E000DFDF00E0E000DFDF00E0E000DFDF00E0E00040408F0000D00000D0\
0000D00000D00000D00000CF0000D00000CF0000CF0000CF0000CF0000C00000C0
0000C00000C0400080D00000D0CF00D0CF00CFC000D0C000CFC000D0C000CFBF00CFC000CFBF00\
CF4000C00000CF0000C00000C000003F006F0000A000009F00009F00009F00009F
000090000090000090BFAF00B0A000BFA000B0A000B0A000B09F003F305FAF0000AF0000A00000\
A0000000006F00006F00006F907F00E00000E00000DF0000E00000D0E000DFE000
D0DF00D0DF004040800000C00000BF0000B00000AFB00000A000009F9F008F8F0000006F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009F9F000000D00000DF0000EFD00000DF0000DF00000000FF0000F00000FF0000F0\
0000FF0000F0FF0000F00000FF0000F000000000FF0000FF0000FF0000F00000FF
0000F00000FF0000F0F00000F00000F00000EF00000000E00000DF0000DF0000CF0000CFCFD000\
CFD000BFC000BFBF00AFB000AFAF009F9F009F9F008F0000800000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6000006000006F60006F600000007F00007F00008000008000008F00008000008F00008F7F0000\
7F00008000008000008000008000008F00008F00008F00008F00003000700000A0
0000AF0000AF0000AF0000AF0000B00000AF0000B00000B00000B00000B03F3F809F9F00A0A000\
A09F00A0A000A0A000AFAF00A0A000AFAF000000C00000CF0000C00000CF0000CF
0000CF0000CF0000D00000CF0000D00000D00000D00000D00000DF0000D00000DF0000DF0000DF\
C0C000C0CF00C0C000CFCF00C0C000CFCF00CFCF00CFCF00CFCF00CFCF006F406F
40009F40009F4F00A040009FD00000CF0000D00000D00000D00000D00000DF0000D00000DF0000\
D00000DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000E00000DF9000
E09F00DF9F00E0E000DFE000E0E0000000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF\
0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000E00000EF4F4F9F
E0EF00E0E000E0EF00E0E000E0EF00E0E000E0EF00E0E000E0E000E0E0000000E00000DF0000DF\
0000D00000DF0000D00000DF0000D00000D00000D00000D00000CF0000CF0000CF
0000CF0000C00000CF0000C0DFD000D0CF00DFD000D0CF00D0CF00D0C000D0CF00CFC000D0C000\
CF0000D00000CF0000CF0000C00000CF0000C00000C0000000009F0000A000009F
00009F00009000009FBFAF00BFAF00BFA000BFAF00B0A000B0A0003F305F0000803F005FAF0000\
A00000AF0000A00000000070907F000000EFE00000E00000DFEF00E0EF00DFE000
DFE000D0DF000000CF0000C00000C00000B00000B03F0070AFAF009F9F00909F0000006F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009090000000D00000DF0000E04F00AFDF0000E000000000FF0000FF0000FF0000FF\
0000FF0000FF5000AFFF0000F00000FF0000FF00000000FF0000FF0000FF0000F0
0000FF0000F00000FFF00000F00000F00000F000004F009F0000E00000DF4F4F90D0DF00D0DF00\
CFD000CFCF00BFBF00B0BF00AFAF00A0A0009F9F00000070800000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000060
600000606000605F006F600000007000007F00008000008F00008000008F7F00007F00007F0000\
8000007F00008000008000008000008000008F00002F006F0000A00000A00000AF
0000AF0000AF0000AF0000B00000AF0000B00000B00000B00000B00000BF30307FA0A0009F9F00\
A0A000A0A000AFAF00A0A000AFAF00AFAF003F3F8F0000CF0000CF0000CF0000D0
0000CF0000D00000D00000D00000D00000D00000D00000DF0000D00000DF3F009040009FC00000\
C04000C0C000CFCF00C0CF00CFCF00CFCF00CFD000CFCF00CFD000CFCF00D04000
CF0000D00000CF0000D00000D00000D00000D00000D00000D00000D00000D00000DF0000D00000\
DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000E00000
DF0000E00000DF0000E00000DF0000E00000E00000E00000E00000E00000E00000E00000E00000\
E000004F009F4F00A04F009F0000EF0000EF0000EF0000E00000EF0000E04F4F9F
E0E000E0EF00E0E000E0EF00E0E000E0EF00E0E000E0EF00E0E000E0EF000000DF0000E00000DF\
0000DF0000DF0000DF0000D00000DF0000D00000D00000D00000D00000CF0000D0
0000CF0000CF0000CF4F4F8FD0D000DFD000D0CF00DFD000D0CF00D0CF00D0CF00D0CF00CFC000\
D00000CF0000D00000CF0000CF0000C00000CF0000C00000C00000C000000000A0
00009F00009F000090BFAF00BFAF00BFAF00B0A000BFAF00B0A0003F305F00008000008000007F\
AF0000A00000A000009F0000000070907F000000EFDFEF00DFEF00DFEF00DFEF00
D0E0000000D00000CF0000CF0000BF0000BF0000AF0000AFAFAF00A0A0002F005000006F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000AFB0000000D00000E00000EFDF0000DF0000E000000000FF0000FF0000F0\
0000FF0000FF0000FFF00000FF0000F00000FF00000000F00000FF0000F00000FF
0000F00000F00000F00000F0EF0000F00000EF0000EF00004F4F90E0E000DFDF00DFDF00CFD000\
CFD000C0CF00BFC000B0B000AFAF00A0A00030305F000070800000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000600000
6F60006060006F60006F600000008000008000008F00008F7F00007F00007F00007F0000800000\
8000008000008000008F00008000008F8F008F80000000AF0000AF0000AF0000AF
0000B00000AF0000B00000B00000B00000B00000BF0000B00000BF0000BF3F3F80A0A000AFAF00\
A0A000AFAF00AFAF00B0B000AFB000B0B0003F3F8F0000D00000D00000D00000D0
0000DF0000D04000903F0090C00000BF0000C00000C00000C00000C00000CF0000C00000CF0000\
CF4000CFD000CFCF00D0D000CFD000D0D000CFD000D0D000D0D000D0DF00D0D000
D00000D00000D00000D00000DF0000D00000DF0000D00000DF0000DF0000DF0000DF0000E00000\
DF0000E00000DF0000E00000DF0000E00000DF0000E00000E00000E00000E00000
E00000E00000E00000E00000EF0000E00000EF0000E00000EF0000E00000EF0000E00000EF0000\
EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF00004F00A07F4F6F
EFEF00EFEF00EFEF00E0EF00EFEF00E0EF00EFEF00E0EF00EFEF00E0EF000000E00000E00000E0\
0000DF0000E00000DF0000E00000DF0000DF0000DF0000DF0000D00000DF0000D0
0000D00000D00000D040408FDFDF00DFD000DFDF00D0D000DFD000D0D000DFD000D0CF00DFD000\
0000BFD00000CF0000D00000CF0000CF0000CF0000CF0000C00000CF0000C00000
C000003F006F00009FBFAF00C0B000BFAF00BFAF00BFAF00BFAF003F306000008F000080000080\
000080000080AF0000AF0000A000009F8F00907F00E0F000DFEF00E0EF00DFEF00
DF0000D000000000D00000CF0000C00000BF0000B0B0BF00B0BF00A0AF0030005F00006F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000009FA000BFBF000000E00000EFD00000DF0000E000000000FF0000FF0000FF\
0000FF0000FF0000FFFF0000F00000FF0000F000000000FF0000FF0000FF0000F0
0000FF0000F00000FF0000F0F00000F00000F00000EF0000EFF000E0EF00DFE000D0DF00D0DF00\
CFD000CFCF00BFC000BFBF00AFB000A0AF0000008000007F800000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F004020204F
605F006F6000605F0020205000007F0000807000007F00007000007F00007F00007F00007F0000\
800000800000805F008080008F80008080008F8F000000A00000AF0000AF0000B0
0000AF0000B00000B00000B00000B00000BF0000B00000BF0000BF0000C00000BFAFAF00A0AF00\
AFAF00AFAF00AFB000AFB000B0B000B0B000B07F00B00000BF0000B00000BF0000
BF0000BF0000BF0000C00000BF0000C00000C00000C00000C00000CF0000C00000CF0000CF0000\
CF0000CFCF00CFD000CFD000D0D000CFD000D0DF00D0D000D0DF00D0D000D0DF00
D00000DF0000D00000DF0000D00000DF0000D00000DF0000DF0000DF0000DF0000DF0000DF0000\
E00000DF0000E00000E00000E00000E00000E00000E00000E00000E00000E00000
E00000EF0000E00000EF0000E00000EF0000E00000EF0000E00000EF0000E00000EF0000EF0000\
EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF4F00
EFEF00EFEF00E0EF00EFEF00EFEF00EFEF00E0EF00EFEF00E0EF004F4F9F0000E00000E00000E0\
0000E00000DF0000E00000DF0000E00000DF0000DF0000DF0000DF0000D00000DF
0000D00000DF0000D0E0DF00DFDF00DFDF00DFD000DFDF00DFD000DFD000D0D000DFD000404080\
0000C00000BF0000BF40007FD00000CF0000CF0000CF0000CF0000C00000C00000
C00000C00000BF0000C00000BFAF00BFB000BFAF00BFAF00B0AF003F3F6000008F00008F000080\
00008000007F00007FA00000A000009F00009F8000000060E0EF00DFEF00DF0000
DF0000DF0000D000000000CF0000C00000BFBFC000BFC000AFB000AFAF0090000000006F000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000030306FB0B000C0C000C0CF00D00000DF0000E000000000F00000FF0000FF\
0000FF0000F00000FF4F00AFFF0000F00000FF00005000AF0000FF0000F00000FF
0000F00000FF0000F05050A0F00000F00000EF0000EF0000E0EF00E0EF00DFE000DFDF00D0D000\
CFD000C0CF00C0C000B0BF00B0B00000008F000080000070800000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000605F00000070606000
6F60006060006F6F0000008000008F7000007F00007F00007F00007F00008000007F0000800000\
8050008F80008080008F8F008F80008F8F008F8F000000AF0000AF0000B00000B0
0000B00000B00000BF0000BF0000BF0000BF0000C00000C00000C00000C00000CF3F3F8FAFB000\
AFB000B0B000B07F00B00000B00000BF0000BF0000BF0000BF0000C00000BF0000
C00000BF0000C00000C00000C00000C00000CF0000CF0000CF0000CF0000CF0000CF0000D00000\
CF0000D0DF00D0D000D0DF00D0DF00DFDF00D0DF00DFDF00D0DF00DFE000DFDF00
DF0000DF0000DF0000DF0000DF0000DF0000E00000DF0000E00000DF0000E00000E00000E00000\
E00000E00000E00000E00000E00000EF0000E00000EF0000E00000EF0000EF0000
EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF0000F00000EF0000F00000\
EF0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000EFF000
F0F000EFEF00F0F000EFF000EFF000EFEF00F0F000EFEF00EFF000EF4F00EF00004F009F0000EF\
0000E00000EF0000E00000E00000E00000E00000E00000E00000DF0000E00000DF
0000DF0000DF0000DFE0E000E0E000DFDF00E0E000DFDF00E0DF00DFDF00E0DF00DFD0004F4F8F\
0000C00000CF0000C00000C00000BF4F0080CF0000D00000CF0000CF0000CF0000
CF0000C00000C00000C00000C00000BF0000C0B000BFAF00BFB0003F306000009000008F00008F\
00008000008F00008000008000007FAF0000A000009F800000006F0000EF0000E0
DF0000D00000DF00000000CF0000CF3F4080C0CF00BFC000B0BF00A0000090000000006F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000B0B000BFBF00CFD000D0DF00DF0000DF00004F00AF0000FF0000FF\
0000FF0000FF0000FF0000FFF00000FF0000FF0000FF00000000FF0000FF0000FF
5050AFF0FF00FFFF00F0FF00F0AF00EF0000F00000EF0000EF0000E0EF00E0EF00DFDF00D0DF00\
CFD000CFD000C0C000BFC000AF000000009000008000007F800000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000006F00006F6F6000
605F006F60006F60000000806F00007000007000007F00007F00007F00007F000000009F00009F\
2F2F6F8080008F8F008080008F8F008F8F009090002F2F700000B00000B00000B0
0000B00000BF0000BF0000C00000BF0000C00000C00000CF0000C00000CF0000CF3F008FAF0000\
B00000B00000B00000B00000BF0000BF0000BF0000BF0000C00000BF0000C00000
C00000C00000C00000CF0000C00000CF0000CF0000CF0000CF0000CF0000CF0000D00000CF0000\
D00000D0DF00D0DF00D0DF00D0DF00D0DF00DFE000D0DF00DFE000DFE000DFE000
0000F00000FF0000F00000FF0000F00000FF0000F00000FF0000F00000FF0000F00000FF0000F0\
0000FF0000F00000FF0000F00000FFE0EF00EFF000E0EF00EFF000EFF000EFF000
EFEF00EFF000EFF000EFF0007F306F4F00AF4F00A0EF0000EF0000EF0000EF0000F00000EF0000\
F00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000F0F000
EFF000F0F000EFF000F0F000EFEF00EFF000EFEF00F0F000EFEF00EF0000EF0000EF0000EF0000\
EF0000EF0000EF0000EF00004F009F0000E00000E00000E00000E00000DF0000E0
0000DF0000DF4F4F90E0E000E0E000E0E000E0E000E0E000DFDF00E0DF00DFDF00E0DF000000CF\
0000CF0000C00000CF0000C00000C00000BF0000BF0000B0CF0000CF0000CF0000
C00000CF0000C00000C00000C00000C00000BF0000BF0000BFAF003F3F6F00009000009000008F\
00008F00008000008000007F00007F00007FA000009F000000006F0000600000E0
D00000DF0000D00000D00000404080C0D000BFC000BFC0000000A0A000008F000000006F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000A03000BFBF00C0CF00D0DF00D00000DF0000DF0000EFF000EFEF00\
F0F0004F50AF5050AF4F4FA0FF5000F05000FFFF00F0F000FFFF00F0FF00FFFF00
F0F000F0FF00F0F000F0FF00EFF000F00000EF0000F00000E00000E0EF00DFE000DFE000D0DF00\
D0D000CFCF00C0CF00B00000B0000000009000008F000070800000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000020004F00006F000070606000
6F60006F6000706F006F00007000007000007F00007F000080000000009F00009F00009F0000A0\
0000A08F8F008080008F8F008F8F009090008F900030307F0000B00000BF0000BF
0000C00000BF0000C00000C00000CF0000C00000CF30008FB00000AF0000B00000B00000B00000\
B00000BF0000B00000BF0000BF0000C00000BF0000C00000C00000C00000C00000
CF0000C00000CF0000CF0000CF0000CF0000D00000CF00004F00A04000A00000F00000F00000FF\
0000F0DFE000D0DF00DFE000DFE000DFE000DFE000E0EF00DFE000E0EF00DFE000
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FFEFF000EFF000F0F000EFF000EFF000EFF000
F0FF00EFF000F0FF00EFF0005050AF0000FF0000FF0000FF0000FF0000FF0000FF0000F00000FF\
0000FF0000FF0000F00000FF0000F05000AFF00000F00000F00000F00000F0F000
F0FF00F0F000F0FF00F0F000F0FF00F0F000F0F000EFF000F0F000F00000F00000EF0000F00000\
EF0000F00000EF0000F00000EF0000F00000EF0000EF00000000E00000E00000E0
0000E00000E0EFEF00E0E000EFEF00E0E000E0E000E0E000E0E000E0E000E0E000DFDF000000D0\
0000D00000D00000CF0000CF0000C00000C00000C00000C00000BF0000BF40007F
D00000CF0000CF0000C00000CF0000C00000C00000BF0000C00000BF000000009F000090000090\
00008F00008F00008000008F000080000080A09000A000009F000000006F000060
0000E0D00000DF0000D00000D0DF00C0D000C0CF000000B00000AFA000009000007F8000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000B0B0BF00C0CF00CFD000DF0000DF0000E00000E0EF00EFF000\
EFF000F0FF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00
FFFF00F0FF00F0FF00F0FF00F0FF00F00000F00000EF0000EF0000E0EF00E0EF00DFE000DFE000\
CFD000404F800000B0C00000B000003F006F00008000007F7F0000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006060000000600000700000706F6000
6F60006F6F006F6F007000007000007F000070000000009000009000009F00009F0000A000009F\
0000A08080008F8F008F8F008F90008F8F009090009090000000BF0000BF0000BF
0000BF0000C00000C03F008FA00000AF0000AF0000AF0000AF0000B00000B00000B00000B00000\
BF0000B00000BF0000BF0000C00000BF0000C00000C00000C00000C00000CF0000
40009F0000F00000EF0000F00000EF0000F00000F00000F00000F00000F00000F00000FF0000F0\
0000FF404FA0DFE000DFE000DFEF00DFE000DFEF00DFE000E0EF00DFE000E0EF00
4F4FAF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FF4F4FA0EFF000EFF000F0F000EFF000F0F000
EFF000F0FF00EFF000F0F0004F4FAF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FF0000F00000FF0000F00000FF4F4FA0F0FF00
F0F000F0FF00F0F000F0F000F0F000F0FF00F0F000F0F000F04F00F00000EF0000F00000F00000\
F00000EF0000F00000EF0000F00000EF0000EF0000EF0000EF0000EF0000EF0000
4F009F0000E0E0E000EFEF00E0E000EFEF00E0E000E0E000E0E000E0E000E0E0004F4F900000D0\
0000DF0000D00000D00000CF0000CF0000C00000C00000C00000C00000BF0000BF
0000B00000BFCF0000CF0000C00000CF0000C00000C00000BF0000BF0000B00000000090000090\
00009000008F00008F00008000008000007FAF9F00A09000A0000090000000006F
0000E0DF0000D00000D00000CF0000CFD0000000BF0000B00000A0A000008F00007F8000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000A00000D0BFBF00CFD000D0DF00DF0000DF0000E0A000E0EF00\
F0F000F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF00
F0F000FFFF00F0F000F0FF00F0F000F00000EF0000EF0000EF0000E0A000DFE000DFE00040408F\
0000CF0000BF0000BFBF0000B0000030006000008F000070800000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000006000007000007000007F6F6000
6F6F006F6F0070700070200070000070000000009F00009F00009F00009F0000A00000A00000A0\
0000A08F8F008F8F009090008F9000909F00909F009FA0009F9F000000C00000C0
A00000A00000AF0000AF0000AF0000AF0000B00000B00000B00000B00000BF0000BF0000BF0000\
BF0000C00000BF0000C00000C00000CF9000C0CF00CFD0004040A00000F00000F0
0000F00000F00000F00000F00000F00000F00000FF0000F00000FF0000F00000FF0000FF0000FF\
0000FF0000FFDFE000DFEF00DFEF00E0EF00DFEF00E0EF00DFEF00E0F000E0EF00
E0F0000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF5050AFEFF000F0FF00EFF000F0FF00F0F000
F0FF00F0F000F0FF00F0F0005050AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00F0FF00
FFFF00F0FF00FFFF00F0FF00FFFF00F0F000FFFF00F0F000FF0000F00000FF0000F00000F00000\
F00000FF0000F00000F00000F00000F00000F00000F00000EF0000F00000EF0000
F00000EF0000F00000EF0000EFA000EFEF00EFEF00EFE000EFEF00E0E000EFEF000000DF0000DF\
0000DF0000DF0000D00000D00000D00000D00000CF0000CF0000C00000CF0000C0
0000C00000BFD0CF00CFCF00CF8F00CF0000CF0000C00000C00000BF0000C00000BF0000BF0000\
00009000009000008F00008F00008000008FAF9F00AF9F00A09000A00000900000
00006FD0E000DF0000D00000D000000000CF0000C00000B00000AFA00000900000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000AFAF00000000DF0000E0D0DF00DF0000E00000E00000EFF000\
EFF000F0FF00F0FF00FFFF00F0FF00FFFF00FFFF00FFFF00F0FF00FFFF00FFFF00
FFFF00F0FF00FFFF00F0FF00F0FF00EFAF00F00000EF0000EF0000E000004F509F0000DF0000D0\
0000CF0000C00000BFC00000B000003F006F00008000007F7F0000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F5F0000006F00006F00007F6020006F6F00
6F60006F6F006F6F0070700020006000009000009000009F00009F0000A00000A00000A00000A0\
0000AF8F8F009090008F9000909F00909F009FA0009F9F00A0A0009F0000A00000
A00000AF0000AF0000AF0000AF0000B00000B00000B00000B00000BF0000BF0000BF0000BF0000\
C00000BF8F00C0CF00C0CF00C0D000C0D000CFD000C0D000CFDF000000EF0000F0
0000F00000F00000F00000F00000F00000F00000F00000FF0000F00000FF0000F00000FF0000F0\
0000FF0000FF4F4FAFDFEF00E0EF00DFEF00E0EF00DFEF00E0EF00E0EF00E0F000
E0EF000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF4F4FAFF0FF00EFF000F0FF00F0F000F0FF00
EFF000F0FF00F0F000F0FF004F4FA00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FFF0FF00FFFF00
F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF000000F00000FF0000F00000FF0000F0\
5000AF4F00A0FF0000F00000F00000F00000F00000EF0000F00000EF0000F00000
EF0000F00000EF0000EF0000EF0000EF0000EF0000EF0000E09F00EFEF004F4F900000E00000DF\
0000DF0000DF0000DF0000D00000D00000D00000D00000CF0000CF0000C00000CF
0000C0404080D0CF00D0CF00CFCF00CFCF00CF8000CF0000C00000C00000BF0000BF0000B00000\
BF00003F006000009000008F00008F000080AFA000A09F00A09F00302F4F9F0000
900000DFEF00D00000D00000CF00000000CF0000BF0000B00000A0A000008F0000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000A0AF00000000D00000E00000EFDF0000DF0000E00000E0EF00\
EFF000EFF000FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF00FFFF00
F0F000FFFF00F0F000F0FF00EFF000F0FF00EF0000F00000EF0000EF00000000DF0000DF0000CF\
0000CF0000C00000BFBF0000B0000030006000008F0000707F0000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000060600000006F00007F6000006F00006F6F00
706F006F6F0070700070700000009000009000009F00009F0000A00000A00000AF0000A00000AF\
0000AF0000B09090009F9F00909F009FA0009FA000A0AF00A0AF00AF3F00A00000
AF0000AF0000B00000AF0000B00000B00000BF0000B00000BF0000BF000040009F0000E00000EF\
40409FC0D000C0D000CFD000CFD000CFDF00CFD000D0DF00CFDF00404FAF0000F0
0000FF0000F00000FF0000F00000FF0000F00000FF0000F00000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FFE0F000E0EF00E0F000E0EF00E0F000E0F000E0F000E0F000
EFF0004F50AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF5050AFF0F000F0FF00F0FF00F0FF00F0F000
FFFF00F0FF00FFFF00F0FF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00F0FF00
FFFF00FFFF00FFFF00F0FF00FFFF00FFFF00FFFF00F0FF000000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000F00000FF0000F0FF0000F00000FF0000F00000F00000F00000
F00000F00000F00000EF0000F00000EF0000F00000EF0000F00000EF0000EF0000EF00000000E0\
0000E00000E00000DF0000DF0000DF0000DF0000D00000D00000D00000D00000CF
0000CFD0D000DFD000D0D000D0D000D0CF00D0CF00CFC000CF0000C00000C00000BF0000C00000\
BF0000BF0000B0000000009000008F00008FAFA000AFA000A09F0000007F000070
9F00009000000000E0D00000D0000040008F0000BF0000B00000AF9F7000000070000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000A0AF00000000DF0000DF0000EF4000A0E00000E000005050AF\
EFF000F0F000F0F000FFFF00FFFF00FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00
FFFF00FFFF00FFFF00F0FF005050AF0000F0F00000EF0000F00000EF00000000E00000DF0000DF\
0000CF0000C00000BFC00000B000003F006000008000007F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060600000006000007000007F6F00006000006F6F00
6F6F007070006F700070700000009000009F00009F0000A00000A00000AF0000A00000AF0000AF\
0000B00000B030307F909F009FA0009FA000A0AF00A0AF00A0AF00A0AF00AF0000
AF0000B00000AF0000B00000B00000BF00003F00900000E00000E00000EF0000E00000EF0000EF\
0000EFC0D000CFD000CFD000CFDF00CFD000CFDF00CFDF00D0E000CFDF000000F0
0000F00000F00000F00000F00000F00000FF0000F00000FF0000F00000FF0000F00000FF0000FF\
0000FF0000F00000FFE0EF00E0F000E0EF00E0F000E0F000EFF000E0F000EFF000
E0F000EFFF000000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FF4F4FA0F0FF00F0F000F0FF00F0F000F0FF00
F0F000FFFF00F0F000F0FF004F50AF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF5050AFF0F000FFFF00
F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F0005050AF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FF0000F00000FF0000F00000FFF00000F00000
F00000F00000F00000F00000F00000F00000EF0000F00000EF0000F00000EF0000EF0000EF0000\
EF00004F00900000E00000DF0000DF0000D00000DF0000D00000D00000D00000D0
0000CFDFDF00D0D000DFD000D0D000D0D000CFCF00CFCF000000B00000B0C00000C00000BF0000\
BF0000B00000B00000B000003F006000008FAFA000A09F00AF9F0000007F000070
00006F9F00000000DFD00000CF000040008F0000BF0000B0AFB000A0AF00000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000A0AF00000000D00000E00000E00000F0DF0000E00000E00000\
0000FF0000F05050AF4F50AF5050AF4F4FA0FF5000F05000FF5000F04F005050AF
5050AF0000FF0000F00000F00000F00000F0EF0000F00000EF0000EF00004F009F0000DF0000D0\
0000CF0000C00000BFBF0000B00000300060000080000070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000605F000000700000706F00006000006F00006F6F00
707000707000707F0070700000009F00009F0000A00000A00000AF0000AF0000B00000B00000BF\
0000B03000809F00009FA0009FA000A0AF00A0AF00AFB000AFB000AFBF00AF3F00
B00000B0000040009F0000E00000E00000E00000EF0000E00000EF0000EF0000EF0000EF0000F0\
0000EF0000F0CFDF00CFDF00CFDF00D0E000D0DF00D0E000D0E000D0E000D0E000
0000FF0000F00000FF0000F00000FF0000F00000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF4F50AFE0A000EFAF00E00000EF0000E00000EF0000EF0000
EF0000EF0000EF0000EF0000EF0000EF0000F00000EF0000F00000EF0000F00000F00000F00000\
F00000F00000F00000F00000F00000FF0000F0FF00FFFF00F0FF00FFFF00F0FF00
FFFF00F0FF00FFFF00F0FF00FF5000F000005000AF5000AF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5050AFFFFF00FFFF00
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
FFFF00F0F000FF0000F00000F00000F00000F00000F00000F00000F00000F00000EF0000F00000\
EF0000EF0000EF0000EF00000000DF0000E00000DF0000DF0000D00000DF0000D0
4F4F90DFDF00E0E000DFDF00DFDF00D0D000D0D000D0D0000000BF0000B00000B0C00000C00000\
BF0000C00000BF0000BF0000B00000B00000AFA000AFA000A09F0000008000007F
00007000006F00006FD00000D000004000800000BFB0BF00AFB0008F0000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000A0A000000000DF0000DF0000EF0000EFE00000E00000EF0000\
0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF00000000FF
0000FF0000FF0000FF0000F00000FF0000F00000F0EF0000F00000EF00004F009F0000DF0000D0\
0000CF0000C00000BFBF0000AF00003F006F000080000070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060600000006F6000006000006F00006F00006F7000
6F6F007070007070007F7F0020206000009F00009F0000AF0000A00000B00000B00000BF900000\
9F00009F00009F00009F3000A0AF00A0AF00AFB000A0B000AFB000AFB000B0BF00
0000DF0000E00000E00000E00000E00000EF0000E00000EF0000E00000EF0000EF0000F00000EF\
0000F00000F00000F0CFDF00D0E000CFDF00D0E000D0E000D0E000D0E000DFEF00
404FA00000FF0000F00000FF0000F00000FF0000FF0000FF0000F04F00AFDF0000E00000DF0000\
E00000E00000E00000E00000E00000E00000EF0000E00000EF0000EF0000EF0000
EF0000EF0000EF0000EF0000EF0000EF0000EF0000F00000EF0000F00000F00000F00000F00000\
F00000F00000F00000F00000F00000F00000FFFF00F0F000FFFF00F0FF00FFFF00
F0F000FFFF00F0FF00FFFF00F00000FF0000F00000FF0000F00000FF0000F00000FF0000F00000\
FF0000F00000FF0000F00000FF00005000AF0000FF0000F05050AFF0FF00FFFF00
F0F000FFFF00F0FF00FFFF00F0F000FFFF00F0FF000000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF5050AF
F0F000FFFF00F0F000FFFF00F0F000F0FF00F00000F00000F00000F00000F00000F00000EF0000\
EF0000EF0000EF0000EF0000EF0000E000004F009F0000DF0000DF0000DF0000DF
DFDF00E0E000DFDF00DFE000DFDF00DFDF00D0D000404F800000BF0000BF0000B00000B00000AF\
C00000BF0000BF0000B00000B00000B00000B0AF00AFA000AFA00000007F00007F
00007000007000006FD00000CF0000C00000BFC000B0BF00A0AF00000070000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000908F00AF00000000D00000DF0000E00000F0DF0000E00000E00000\
0000FF0000FF0000FF0000F00000FF0000FF5000AFF00000FF0000F000005000AF
0000F00000FF0000F00000FF0000F00000F00000F0F00000EF0000EF00004F009F0000DF0000D0\
0000CF0000C00000BFBF0000B00000000090000080000070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060600000007F6000006F00006F000000008F20205F
7070007070007F7F00707F007F80000000A00000AF0000AF0000B00000B09000009000009F0000\
9F0000A000009F0000A00000A0B000AFB000AFB000B0BF00B0BF00B0C000B0BF00
40409F0000E00000EF0000E00000EF0000EF0000EF0000EF0000F00000EF0000F00000EF0000F0\
0000F00000F00000F04F4FAFD0E000D0E000D0E000DFEF00D0EF00DFEF00DFEF00
DFF0004F00AFDF0000DF0000E00000DF0000E00000DF0000E00000E00000E00000E00000EF0000\
E00000EF0000E00000EF0000E00000EF0000E00000EF0000EF0000EF0000EF0000
EF0000EF0000EF0000EF0000F00000EF0000F00000EF0000F00000EF0000F00000EF0000F00000\
F00000F00000F00000F00000F00000F00000F0FF00F0FF00F0FF00F0FF00F0FF00
FFFF00F0FF00FFFF00F0FF00FF0000F00000FF0000F00000FF0000FF0000FF0000F00000FF0000\
FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000
FF0000FFAF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0F0005050AF5000A0F00000F00000F00000\
F00000F00000EF0000F00000EF0000EF0000EF0000EF0000E000000000E04F4F90
E0EF00E0E000E0E000DFE000E0E000DFDF00DFDF000000C00000CF0000BF0000BF0000B00000B0\
0000AF40007FBF0000C00000BF0000BF3F00B0AF00B0AF00AFA00000008000007F
00007F000070000070CF0000CF0000C0D000BFCF00AFB00000009F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000909000A000000000DF0000DF0000EF0000EFDF0000DF0000EF0000\
0000FF0000FF0000FF0000FF0000FF0000FF5000AFFF0000FF0000FF00005000AF
0000FF0000FF0000FF0000F00000FF0000F00000F0EF0000F00000EF00004F009F0000DF0000D0\
0000CF0000C00000BFBF0000AF00000000900000808F9000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060600060000060000060000020005F00008F000090
6F7000707F007070007F7F007F80002F2F700000A00000AF0000B09000009000009F00009F0000\
A000009F0000A00000A00000AF3F00AFB000AFBF00AFBF00B0C000B0BF00BFC000
BFC00040409F0000E00000EF0000EF0000EF0000EF0000EF0000EF0000F00000EF0000F00000F0\
0000F00000F00000F00000F0D0EF00D0E000DFEF00D0EF00DFEF00D0EF00DFF000
DFEF00DF5000DF0000E00000DF0000E00000DF0000E00000E00000E00000E00000E00000E00000\
E00000E00000E00000E00000EF00007F3070EFFF00E0FF00EFFF00E0F000EFFF00
E0FF00EFFF00E0F0004F50AF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FF0000F0F0FF00EFFF00F0FF00EFF000F0FF00
EFFF00F0FF00EFF000F0FF004F50AF0000FF0000F05000AF4F00AFF00000F00000F00000F00000\
F00000F00000F00000F00000FF0000F00000FF0000F00000FF0000F00000FF0000
F00000FF0000F00000FF0000F00000FF0000F00000FF00005000AF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF4F4FA0FFFF00
F0FF00FFFF00F0F000FFFF00F0FF00FFFF00F0F000FFFF000000FF0000FF0000F00000F0F00000\
F00000EF0000F00000EF0000F00000EF0000EF0000EF0000EF0000E00000EFEF00
E0E000E0EF00E0E000E0E000DFE000E0E000DFDF000000CF0000C00000C00000BF0000BF0000B0\
0000B00000AF0000AFBF0000BF0000B0B000B0B000AFAF00AFAF00000080000080
00007F00007F00006F908F00C0D000C0D000B0BF000000A08F0000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008F8F00A000000000D00000DF0000E00000F0DF0000E00000E00000\
0000FF0000F00000FF0000FF0000FF0000F00000FFF00000FF0000F00000FF0000
0000FF0000FF0000F00000F00000F00000F00000EFEF0000EF0000EF000040009F0000DF0000D0\
0000CF0000C00000BFBF0000B0000000008F000080808F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060006000006000006000006F0000000080000090000090
707F00707F007F80007F8000808F00808F000000B08F00009000009000009F00009F0000A00000\
A00000AF0000AF00003F00900000D0B0BF00B0BF00B0C000B0C000BFCF00BFCF00
C0CF00BFCF00404FA00000EF0000F00000EF0000F00000F00000F00000F00000FF0000F00000FF\
0000F00000FF0000F0DF0000D00000DFEF00DFEF00DFF000DFEF00E0F000DFF000
E0F000DFF000E05000E00000E00000E00000E00000E00000E000004F00AF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FFE0FF00EFFF00E0FF00EFFF00E0FF00
EFFF00E0FF00EFFF00EFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF4F50AFEFFF00EFFF00F0FF00EFFF00
F0FF00EFFF00F0FF00EFFF005050AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FFF00000FF0000F00000FF0000F00000
FF0000F00000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000\
5000AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF\
0000F00000FFF00000F00000F00000F00000EF0000F00000EF0000EF5000EFEF00
EFEF00E0EF00EFEF00E0E000E0EF00DFE0000000DF0000D00000D00000CF0000CF0000C00000C0\
0000BF0000BF0000B00000B00000AFC0BF00B0B000B0B000AFAF00AF0000000080
00008000007F00007F908F00CFDF00BFCF000000B000009F000070000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000909000A000000000DF0000DF0000E00000EFDF0000E00000E00000\
0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000
0000FF0000FF0000FF0000F00000FF0000F00000F0EF0000EF0000EF0000E000000000DF0000D0\
0000CF0000C00000BFBF0000AF0000000090909F008F8F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F5F0060000060000060000000008000008F00008F000090
6F70007F7F007F8000808F00808F008F90008F00009000009000009F00009F0000A000009F0000\
A000000000D00000D00000D00000DF0000DFB0C000B0C000BFCF00BFCF00C0CF00
BFCF00C0D000C0D0000000F00000EF0000F00000EF0000F00000F00000F00000F04000AFD00000\
D00000D00000DF0000D00000DF0000DF4F00DFF000DFEF00DFF000DFF000E0F000
DFF000E0F000DFF0004F50AF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF0000FF0000FFE0F000E0FF00E0FF00E0FF00
E0F000E0FF00E0FF00EFFF004F4FA00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF0000FF4F50AFE0F000EFFF00E0FF00EFFF00
E0F000EFFF00EFFF00EFFF004F4FA00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF0000FF0000FFEFF000F0FF00F0FF00F0FF00
F0F000F0FF00F00000F00000F00000F00000F00000FF0000F00000FF0000F00000FF0000F00000\
FF0000F00000FF0000F000005000AF0000FF0000FF0000F05050AFF0FF00FFFF00
F0F000FFFF00F0FF00FFFF00F0F000FFFF005050AF0000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F05000A0F00000F00000EF0000EF0000EFEF00EFF000
E0EF00EFEF00E0EF00EFEF00E0E0004F4F900000D00000D00000CF0000CF0000CF0000CF0000C0\
0000C00000BF0000BF0000AFC0C000BFBF00BFBF00B0B000B0AF00AF0000A00000
00007F00007F000070908F00BF00000000BF0000A0000090000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008F8F00A00000B0B00040409F0000E00000EFDF0000E00000E00000\
0000FF0000FF0000FF0000F00000FF0000FF0000FFF00000FF0000F00000FF0000
0000F00000FF0000F00000FF0000F00000F00000F0F00000EF0000EF0000E000000000DF0000CF\
0000CF0000C00000BFB00000AF00009FA000909F00808F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000060000060000060000000008000008F00008F000090000090
00009F7F8000808F00808F008F60008F00009000009000009F00009F0000A00000A000000000D0\
0000D00000DF0000DF0000DF0000DF0000E00000E0BFCF00BFCF00C0D000C0CF00
C0D000C0D000CFDF00C0DF000000F00000F00000F04000A0D00000D00000D00000D00000DF0000\
D00000DF0000DF0000DF0000DF0000DF0000DF0000E0F000DFF000E0FF00DFF000
E0FF00DFF000E0FF00DFF000E0FF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF4F50AFE0FF00E0FF00E0FF00
E0FF00E0F000E0FF00E0F000E0FF00E05000E00000E00000E00000E00000E00000DF0000E00000\
E00000E00000DF0000E00000DF0000E00000DF0000E00000E00000E00000E00000
E00000E0F000E0FF00E0F0000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF4F50AFF0FF00EFFF00F0FF00EFFF00
F0FF00F0FF00F0FF00F0FF005050AF0000FF0000FF0000FF0000FFF00000FF0000F00000FF0000\
FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5000FFFF00FFFF00FFFF00
FFFF00FFFF00FFFF00FFFF00FFFF005050AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000F0F00000F0F000F0F000EFF000
F0F000EFF000EFF000E0EF00EFEF00E000004F009F0000D00000DF0000D00000D00000CF0000CF\
0000C00000C00000BF0000BFC0CF00C0C000BFBF00BFBF00B03F00B00000A00000
A0000000008000007F908F008F80000000B00000A0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000A0A06F00B0BF00BFC000CFD0000000EFDF0000E00000EF0000\
4F00AF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000
0000FF0000FF0000FF0000F00000FF0000F00000F0EF0000F00000EF0000E000000000DF0000D0\
0000CF0000C00000BFBF0000AF0000A07000909F008F8F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000060000060000060000000008000008000008F00008F00009F
00009F2F2F707F80008000008F00009000009000009000009F0000A000009F00000000D00000D0\
0000DF0000D00000DF0000DF0000E00000E00000E00000E0C0D000BFCF00C0D000
C0D000CFDF00C0DF00CFDF00CFDF00D00000CF0000D00000D00000D00000D00000D00000D00000\
DF0000D00000DF0000DF00000000FF0000F00000FF0000FF4F50AFDFF000E0F000
DFF000E0FF00DFF000E0FF00DFF000E0FF000000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FFDF0000E00000DF0000DF0000DF0000DF0000D00000DFF000
D0EF00DFEF00D0EF00D0EF00D0EF00D0EF00D0E000CFCF00CFCF00DFDF00DFDF00E0E000E0E000\
EFEF00EFEF000000FF0000FF0000FF0000F00000FF0000FF5000AFF00000FF0000
FF0000DFEF00D0E000DFEF00404FA00000F00000F04F00AFDF0000DF0000DF0000E00000DF0000\
E00000DF00004F00AF0000F00000FF0000F0EFFF00E0FF00EFFF00E0F000EFFF00
EFFF00EFFF00EFF0004F50AF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
5000AFF00000F00000F00000FF0000F00000FF0000F0FF00FFFF00F0F000FFFF00
F0FF00FFFF00F0F000FFFF00F0FF000000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FF0000F00000FF4F50A0F0FF00EFF000F0F000
EFF000F0F000EFEF00EFF000E00000E00000E00000E000000000D00000D00000CF0000D00000CF\
0000CF0000C00000C00000BFCFCF00C0C000BFC000B0BF00B00000AF0000AF0000
A00000A0000000007F9090002F2F400000AF000090000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000A0A00000B00000BFC000C0CF00D0DF004000A0E00000E00000\
4F00AF0000F00000FF0000FF0000FF0000F00000FFF00000FF0000F00000FF0000
0000FF0000FF0000F00000F00000F00000F00000EFEF0000EF0000EF0000DF00000000DF0000D0\
0000CF0000BFC0CF00B00000AF00009F7000909F00808F00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000600000600000606F0000008000008F00008F00009F00009F
0000A02F00708F00008F00009000009000009F00009F0000A000000000CF0000D00000D00000DF\
0000DF0000E00000E00000E00000E00000EF0000EF0000EF0000EFCFDF00C0DF00
CFDF00CFDF00CF0000CF0000D00000CF0000D00000D00000DF0000D00000DF0000D000004F00AF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFDFF000
E0FF00DFF000E0FF00E0FF00E0FF00DFFF00E0FF00DF0000E00000DF0000DF0000DF0000DF0000\
DF00000000FF0000F00000FF0000F00000FF0000807F00008000009000000000AF
0000B00000BF0000CF0000CF0000D00000D0CF4000CFCF00D0D000DFDF00E0E000E0E000EFEF00\
EFEF00F0F000EFF0005050AF0000FF0000FF0000FFFF0000F00000FF0000FF0000
FF00005000AF0000FF0000FF0000FF0000FF0000FF0000FF0000FFF00000FF0000F00000F0FF00\
EFFF00E0F000CFDF000000FF0000F0E0F000DFEF00E0F000E0F000EFFF00E0F000
EFFF00E0FF00EFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF4F00AFFF0000F0FF00FFFF00F0FF00FFFF00FFFF00
FFFF00FFFF00FFFF00FFFF00FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00F0FF00FFFF00F0FF00
F0FF00EFF000F0F000EFF000EF0000E00000EF0000E00000E000004F00900000DF0000D00000D0\
0000CF0000CF0000C0D0DF00CFCF00CFCF00C0C000BFC0003F006FB00000AF0000
AF0000A000003F005F90900030304F0000A0000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000AFA00000B00000BFC000CFD000D0D000DFE000E0EF00EF5000\
E000000000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000
0000FF0000FF0000FF0000F00000F00000F00000F0EF0000EF0000EF0000E000000000DF4F4F90\
CFDF00CFD000C0CF00BF0000AF0000A07000909F00000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F000060600060600000008F00008F0000900000900000A0
0000A08000008000008F00008F00009000009000009F00009FAF000000D00000D00000DF0000D0\
0000DF0000DF0000E00000E00000EF0000EF0000EF0000EF0000F00000F0CF9000
C00000CF0000CF0000D00000CF0000D00000D00000D00000D000000000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF
DFF000E0FF00DFA000DF0000DF0000DF0000D00000DFA000D0EF00D0EF00000000000000000000\
00000000000000000000000060600000007F7F00007F00008F00002F00700000B0
0000B00000C00000C00000D00000D0400090CF0000D04000D0D000DFE000E0E000EFEF00E0EF00\
EFEF00EFEF00F0F000EFF000F0F0000000FFFF0000F00000FF0000FF0000FF0000
0000F00000FF0000FF0000FF0000F00000FF0000FF0000FFF00000F00000F000000000FF0000F0\
0000F0CF0000C0AF00B0A0000000A000009F00009FAF0000AF0000A00000DFE000
D0E0004F00AFDF0000E00000DF0000E000000000F00000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FFF0FF00F0FF00F0F000F0FF00F0FF00FFFF00
F0F000FFFF00F0FF00FF5000F00000FF0000F00000FF00004F00A00000FF0000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF0000FFF0FF00F0F000F0FF00F0FF00F0FF00
F0F000F0FF00EFF000F00000EF0000EF0000E00000E00000E00000E00000DF00004F00900000D0\
0000D00000CF4F4F8FCFD000CFD000C0CF00C0CF000000AF0000A03F006FB00000
A00000A000009F000090900000006F90A000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000A0AF0000B00000BFC000C0CF00D0DF00DFE000E0EF00E0EF00\
EFF000EFEF000000FF0000F00000FF0000FF0000FFF00000FF0000F00000FF0000
0000F00000FF0000F00000FF0000F00000F00000EFEF0000EF0000EF0000E04F00DFEF00D0DF00\
CFDF00C0D000C0CF00B00000AF00009F7000909F00000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000600000606000606F0020205F00009000009000009F0000A0
8000008000008F00008F00009F00009F00009FAF009FAF00AFB000A0B0000000DF0000DF0000E0\
0000DF0000E00000E00000EF0000EF0000F00000EF0000F0CF0000CF0000CF0000
D00000CF0000D00000D00000D00000D00000DFEF00D0EF00DFF000404FAF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFDF0000DF0000
DF0000DF00004F00AF0000F00000FF000000000000000000000000000000000000000000000000\
0000000000000000006000006F60000000807F00008F00008F00000000B00000B0
0000C00000C00000D00000D00000DFCF0000D00000D00000DF4F00E0E000EFEF00EFEF00F0F000\
EFEF00F0F000F0F000F0FF00F0F000FF0000FF0000FF0000FF0000FF00000000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FFF00000FF0000F000000000FF0000FF0000F0\
D00000CFBF00C0B000C0B0000000AF0000AF0000A00000A000009F00009FAF0000
AF9000A08F00A08F009F80009F00009F00000000FF0000F0E00000E00000E000000000FF0000FF\
0000FF0000FF0000FF0000FFEFFF00F0FF00EFFF00F0FF00F0FF00F0FF00F0FF00
F0FF00F0FF00FF0000F00000FF0000FF0000FF0000FF0000FF0000FF0000FF00000000FF0000FF\
0000FF0000FF0000FF0000FF0000FFFFFF00F0FF00FFFF00F0FF00FFFF00F0FF00
FFFF00F0FF000000FF0000F05000AFEF0000EF0000E00000EF0000E00000E00000DF0000E00000\
0000D00000D0D0DF00D0DF00CFD000CFD000C0CF000000B00000AF0000A03F006F
B00000AF0000A00000909000000070000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007000000000B0AF00000000DFBFC000CFD000D0D000DFE000E0EF00EFF000\
EFEF00EFF000EFF000F0FF00FFFF005050AF5050AFFF0000FF0000FF0000FF0000
0000FF0000FF0000FF0000F00000FF4F50A05050A0EFF000EFFF00EFF000E0F000DFE000DFE000\
CFDF00CFD000C0CF00BF0000AF0000A070008F9000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000CF00005F0000606F00606000606F0000008F00009F00009F7F0000
8000008F00008F00009000009000009FAF009FAF00A0B000A0B000AFBF00AFBF003F40900000DF\
0000E00000E00000EF0000EF0000F03F00A0C00000C00000CF0000CF0000CF0000
CF0000D00000CF0000D0A000D0EF00DFEF00D0EF00DFF000D0EF00DFF000D0F0000000FF0000FF\
0000FF0000F00000FF0000F0DF0000D00000DF0000D00000DF00000000F00000F0
0000F0000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060006F6F007020008000008000000000AF0000B00000BF
0000C00000CF0000CF0000DFC00000CF0000D00000DF00004000A04F4FAFE0E000EFEF00EFEF00\
F0F000EFEF00F0F000F0F000FF0000F00000FF0000FF0000FF00000000F00000FF
0000FF0000FF0000F00000FF0000FF0000FFF00000F00000F000000000FF0000F00000EFCF0000\
D0C000CFBF00CFBF000000B00000B00000AF0000AF0000A00000A000009F3F306F
AF9000AF9000A08F00A030009F0000A000009F00009F00009000009000000000F00000F0DF0000\
E00000E00000E0F000E0FF00EFFF00E0F000EFFF00EFFF00EFFF00EFF000EFFF00
4F50AF0000FF0000F00000FF0000FFF00000F00000FF0000F00000FF0000F00000FF0000F00000\
FF00000000F00000FF0000FFFFFF00F0F000F0FF00F0FF00FFFF00F0F000F0FF00
F0FF000000FF0000F00000FF0000F00000F0EF0000EF0000E00000EF0000E00000E00000DF0000\
DF0000DF0000DFE000D0DF00D0DF00CFD0004040800000B00000B00000A00000A0
3F006FAF00009F0000900000000070000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000009F0000AFAF00000000D00000DF40409FD0DF00D0DF00E0EF00E0EF00\
EFF000EFEF00F0F000F0F000FFFF00F0F000FFFF00F00000FF0000F00000FF0000
F0FF00FFFF00F0F000F0FF00F0F000F0FF00EFF000EFF000EFF000EFFF00DFEF00DFEF00D0DF00\
CFDF00C0CF00C0CF00B00000AF00009FA00000007F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000600000600000606000606F006F6F002F2F6000009F2F0070800000
8F00008F00009000003000800000CF9FAF00A0B000A0B000AFBF00AFBF00B0C000B0C000BFCF00\
0000E00000EF0000EF0000F0C00000CF0000CF0000CF0000CF0000D00000CF0000
D000000000FF0000FF0000FFDFF000D0EF00DFF000DFF000DFF000DFF000E0FF00DFF000E0FF00\
4F50AFE05000DF0000DF0000D000000000FF0000F00000FF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006060006F6F007F70008000003000700000AF0000BF0000C0
0000CF0000CF0000DFC00000CF0000D00000DF0000DF00000000FF0000FF0000FF4F4FAFF0F000\
EFF000F0F000F0F000FF0000F00000FF0000FF0000FF0000FFFF005050AF0000FF
0000FF0000FF0000FF0000FF0000FFF00000FF0000F000000000FF0000FF0000F0D00000E0D000\
DFCF00DFCF00D0C0000000C00000BF0000BF0000B00000B00000AF0000AFB0A000
B0A000AF9F00AF9F0000009F3F006FA00000A000009F00009F00009F00009F0000900000900000\
900000900000DFE000DFEF00DF0000E00000E0A000EFFF00EFFF00EFFF000000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000F00000FF0000FF0000FF0000\
FF0000FF0000FF0000FFAF00F0FF00FFFF00F0FF00FFFF00F0FF00FFFF00F0FF00
0000FF0000F00000FF0000FF0000FF0000F00000FF0000F0EF0000E00000EF0000E00000E00000\
DF0000E00000DF9F00DFE000D0DF00D0DF000000C00000BF0000B00000B00000A0
0000A0AF0000A00000900000000070000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000009F0000B0AF00000000DF0000DF0000EF0000EFDFE000E0EF00EFF000\
EFEF00F0F000F0F000F0FF00F0F000FFFF00FFFF00FF0000FF0000FF0000FF0000
FFFF00FFFF00FFFF00F0FF00F0FF00F0F000F0FF00EFF000EFFF00E0F000E0F000DFEF00D0E000\
CFDF00CFD000C0CF00BF0000A0000000008F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006000005F000000007F606F006F70006F700000009F7F0000800000
8000008F00009000000000C00000C00000CFA0B000AFBF00AFBF00B0C000B0C000BFCF00B0CF00\
BFCF000000EFC00000C00000C00000C00000CF0000CF0000CF00000000F00000FF
0000F00000FF0000F00000FF0000F00000FFDFF000DFF000DFF000DFF000DFF000DFF000DFF000\
DFF000D0F000DFF000D0EF00D0F000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000006F606000706F007F70008F80000000A00000B00000BF0000CF
0000CF0000D03F0090CF0000CF0000D00000DF00000000FF0000F00000FF0000FF0000FF0000F0\
F0F000F0F000FFAF00F00000FF0000FF0000FF0000F0F000FFFF00FFFF00FFFF00
4F4FA00000FF0000FF0000FFF00000F00000F000000000FF0000F00000EFD00000EFDF00E0D000\
E0D000DFCF00DF0000CF00000000C00000BF0000BF0000B00000B03F3070BFAF00
B0A000B0A0003F306F0000A000009F00009F0000900000909F00009F00009F00009F0000900000\
9000009000009000008F00008F00008F00000000F00000F0DF0000DF00000000FF
0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FFF00000F00000\
FF0000F00000FF0000F00000FF0000F00000FFAF00F0F000F0FF00F0FF000000FF
0000F00000FF0000F00000FF0000F00000FF0000F00000FF0000F04F00A0E00000EF0000E00000\
E00000DF0000DF0000D09000D0DF000000C00000C00000BF0000BF0000B00000AF
0000A03F006FA000009F0000000070000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000A00000B0B000000000DF0000E00000E00000EF0000F0E0EF00E0EF00\
EFF000EFF000F0F000F0F000FFFF00F0FF00FFFF00F00000FF0000F00000FF0000
F0F000FFFF00F0F000FFFF00F0F000F0FF00EFF000EFFF00E0F000EFF000DFEF00DFEF00D0DF00\
CFDF00C0D000C0CF00AF0000A0000000007F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F000020005000007F6F6F006F70007080007F8000800000800000
8F00009000000000C00000C00000D00000D00000D03F3F90B0C000B0C000BFCF00BFCF00C0D000\
BFD000C0D000C00000CF0000CF0000CF0000CF00004F00AF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFDF0000DF0000DF0000DF00004F50AF\
D0F000DFF000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000006F6F6F00706F00807F008F800030307F0000BF0000C00000CF
0000D00000D0CF0000CF0000D00000DF00000000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FFFFFF00F00000FF0000FF0000FF0000FFFF00FFFF00FFFF00FFFF00FFFF00
FFFF00FFFF00FFFF00F05000FF0000F000000000FF0000FF0000EFD000000000EFEFDF00EFE000\
E0D000E0D000DF0000DF0000D00000D000000000C00000C00000BFC0B000BFAF00
BFAF00BFA0000000AF0000A00000AF0000A00000A000009F00009FA08F00A08F009F8F00A00000\
9F00009F0000900000900000900000907F008F7F00907F008F00000000FF0000F0
E00000E000000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
F0AF00FF0000F00000FF0000FF0000FF0000F00000FF0000F0FF000000FF0000F0
0000FF0000FF0000FF0000F00000FF0000F00000FF0000F00000FF0000F00000F04F00A0EF0000\
E00000E00000DF0000DF000040008F0000CF0000C00000C00000BF0000BF0000AF
0000AFAFB000A000009F0000900000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000007000000000A0AF0000B000000000DF0000E00000EF0000EF0000F0E00000EF0000\
EF4F00F0F000F0F000FFFF00F0FF00FFFF00FFFF00FF0000FF0000FF0000FF0000
FFFF00FFFF00FFFF00F0FF00F0FF00F0F000F0FF00EFF000EFFF00EFF000E0F000DFE000DFE000\
CFDF00CFD0000000B0B000009F0000000080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000007000007F000080707F00707F007F8F007F8F008F0000
8F00000000BF0000C00000CF0000CF0000D00000D00000DF0000DFB0CF00B0CF00BFCF00BFD000\
C0D000C0D000C0DF00C04000CF0000CF00000000FF0000F00000FF0000F00000FF
0000F00000FF0000F00000FF0000F04F00AFD00000DF0000D000004F00AF0000F00000FF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000706F60007F7000807F008F8F009090000000C00000C00000D0
0000D0C00000C00000D00000D000004F00AF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FFF00000FF0000FF0000FF0000F0A000FFFF00FFFF00FFFF00F0F000FFFF00
FFFF00FFFF00F0F000F0FF00F0FF00F0FF00EFF000E0EF00D000000000EFEFE000EFE000EFDF00\
EFDF00E04000E00000DF0000DF0000D00000D00000CF0000CF0000C0B000C0B000
BFAF00BFAF000000AF0000AF0000AF0000AF0000A00000A0A09000AF9000A08F00A09000000090\
00009000008F00009000008F300060907F00907F008F7F008F00008000008F0000
8000000000F00000F0DF0000DF00000000FF0000F00000FF0000FF0000FF4F4FA0EFFF00EFFF00\
F0FF00EFF000F0FF00F0AF00F00000F00000FF0000F00000FF0000F000000000FF
0000F00000FF0000F00000FF0000F00000FF0000F00000F00000F00000F00000F0EFFF00E0A000\
E00000DF0000DF0000D00000DF000040008F0000CF0000C00000C00000BF0000B0
0000A0AFB000A000009F00008F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000007F0000303070A00000BF00000000DF0000E00000E00000F04000A0E00000E00000\
EF00000000F00000FFF0F000FFFF00F0F000FFFF00FFAF00FF0000F00000FF0000
FFFF00FFFF00F0F000F0FF00F0F000F0FF00EFF000EFF000EFF000EFF000DFEF00DFE00040408F\
0000CF0000BF0000B0A0000000008F000070000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000007000007F00008F6F7F007F80007F8F008F90008F3000
3000800000BF0000CF0000CF0000D00000D00000DF0000DF0000E00000E0BFD000BFD000C0DF00\
C0D000CFDF00CFDF00CFE000CFE000D0EF00404FAF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF4F00AFDF0000DF0000DF00000000FF0000FF0000F08F00000000607F6F00\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000006F000070706F007F70008F8F00909000A09F00A0A0000000CF0000D0
C00000C00000D00000D00000DF00000000F00000FF0000FF0000FF0000FF0000FF0000FF0000FF\
F00000FF0000FF0000FF0000FF00005050AFFFFF00FFFF00FFFF00FFFF00FFFF00
FFFF00F0FF00FFFF00F0FF00F0FF00F0FF00EFF000D000000000F00000EFF0EF00EFE000EFE000\
EFDF000000EF0000E0E00000DF0000E00000DF0000D00000D00000D00000CF0000
CFBF000000BF0000BF0000B00000B00000AF0000B00000AFB09F00AF9000AF9F0000009F00009F\
00009F00009F0000900000909080009F800090800000009000008F00008F00008F
00008F8070008070002F205FDFE000DFE000E0F000E0F000EFFF00E0FF00EFFF00EFFF00EFFF00\
EFFF00F0FF00EFFF00F0FF00F0FF00F00000F00000FF0000F00000FF0000F00000
F000000000F00000FF0000FF0000FF0000F00000FF0000F00000FF4F50A0F0FF00EFFF00EFFF00\
E0A000E00000DF0000DF0000D00000DF000040008F0000CF0000C00000C00000B0
BFC000AFB000AFB0009F0000900000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006000008F8F009F9F00AF0000BF00000000E00000E00000EF0000EFE00000E00000EF0000\
EF00000000FF0000FF0000FF0000FF5050AF5050AFFF0000FF0000FF0000FF0000
FFFF00FFFF00FFFF00F0FF00F0FF00F0F000F0FF00EFF000EFFF00E04F00E000000000DF0000D0\
0000C00000BF0000AFA00000000080700000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000006F00007F000080000090708000808F008090008F9F00
0000BF0000C00000C00000CF0000CF0000D00000DF0000E00000DFBF0000BF0000C04F00C0D000\
C0DF00C0DF00CFE000CFDF00CFE000CFE000D0EF00D0EF00D0EF000000F00000FF
0000F00000FFDF0000DF0000D000004F00AF0000F00000FFCF000000008F9000008F00007F7000\
00006000005F000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000006F00007F6F20007F7F008F80009090009F9F00AFAF00B0B000404090
BF0000CF0000CF0000DF00000000F00000FF0000FF0000FF0000F00000FF0000FF0000FF4F00A0\
FF0000FF0000FF0000F000000000FF0000FF0000FF4F4FA0FFFF00FFFF00FFFF00
F0F000FFFF00F0FF00F0FF00F0F000EFF000DFE000CF00000000EFF0EF00EFEF00F0EF00EFE000\
EFE0000000E00000EF0000E00000E04F0090E00000D00000D00000D00000D00000
CF0000CF0000C000000000C00000BF0000BF0000B03F3F7FB0A000B0A000AF9F000000AF0000A0\
0000A000009F00009F0000909F8F0090800090800000008F00009000008F00008F
00008F80700080700000008F000080000080202050DFE000DFE000DFEF00E0F000E0FF00E0F000\
EFFF00EFFF00EFFF000000F00000FF0000FF0000FFEF0000F00000F00000FF0000
F00000F00000F000000000FF0000F00000FF0000F00000FFEFF000EFFF00EFF000EFFF00E0F000\
EFFF00E0A000E00000DF0000DF0000D00000D0000040008F0000CF0000BF0000BF
B0C000B0BF000000909F00008F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000007000008F8F009F9F00AF30003F00900000DF0000E00000E00000F0E00000EF0000E00000\
4F00AF0000FF0000FF0000F00000FF0000FF0000FFF00000FF0000FF0000FF0000
0000F00000FF0000F00000FF0000F00000F00000EFEF0000E00000E000000000DF0000D00000CF\
0000C00000B0B000009F0000000080000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000060600000007F00008F0000907F0000808F008F9F008F9F00
9FAF00303F800000CF0000CF0000D00000D00000DF0000DF40009FBF0000C00000C00000CF0000\
C00000CFE000CFE000D0EF00D0E000D0EF00D0EF00DFF000D0F000DFF000DFF000
DF0000DF0000DF0000DF0000DFF000DFF0004F009FCF00000000A00000909F9F009000008F0000\
807F0000006F0000607F000000005F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000007000007F7F00008080009090009F9F00AFAF00AFAF00BFBF00BF8000
CF0000CF0000D000000000F00000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000\
FF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000
FF5000FFFF00FFFF00F0FF00F0FF004F4F9FD000000000F00000FFF0F000F0F000F0EF00F0EF00\
0000EF0000EF0000EF0000EF0000E00000E00000E00000E0DF0000DF0000D00000
D00000CF0000CF0000C00000CF0000C00000400080BFAF00BFAF00B0A000B0A0000000AF0000AF\
0000AF0000AF0000A03F30709F8F009F8F0000009F00009F000090000090000090
8F80008F7F008F7F0000008F00008F00008F7F70007F6F0000008F0000F0DF0000DF0000E00000\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFF00000F00000
FF0000F00000F00000F000005000AF0000FF5050AFEFFF00F0FF00EFFF00EFFF00EFF000EFFF00\
E0FF000000F00000EFE00000DF0000DF0000D00000D000000000CF0000C0BFCF00
BFCF00B0BF0000009F9F00008F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000008F9090009F9F00B0B0000000DF0000E00000E00000EF0000F0EF0000E00000EF0000\
0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000
0000FF0000FF0000FF0000F00000FF0000F00000F0EF0000EF0000E000000000DF0000D00000CF\
0000C00000BFAF00009F000000007F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000007F0000802000607000008000008F900090A000
90A0009FAF009FAF000000D00000D00000DF0000DF3F009FBF0000C00000BF0000C00000C00000\
0000F00000F00000F00000F0D0EF00D0EF00DFEF00D0EF00DFF000D00000DF0000
DF0000DF0000DFF000DFF0000000EFD00000CF00000000BF0000A000009FA09F009F0000900000\
908000807F0000006F00006000006070600000005F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000600000706F00007000008F00008F8F009F9F00A0A000B0B000B0BF00C00000
C00000D00000D000000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0FF0000F00000\
FF0000F000000000FF0000FF0000FF0000F00000FF0000FF0000FF4F00A0FF0000
F00000FF00000000F00000FF0000EFDF0000F00000FF0000F00000FFF000F0F000F0F000F0EF00\
0000F00000EF0000EF0000EF0000EF0000E00000E0E0D000E0DF00DFD000DF0000
D00000D00000CF0000CF0000CF0000CF0000C00000C0B000BFAF00BFAF003F007F0000B00000B0\
0000B00000AF0000AFA09000A090009F90000000A000009F00009F00009F00009F
8F80009080002F2F600000900000900000907F7000807F0000008F2F00607F00000000F0DF0000\
E00000DF00000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF
EF0000F00000F00000F00000F00000F0FF00EFF000F0FF00EFF000EFFF00EFF000EFFF000000EF\
0000F00000EF0000EF0000E0DF0000D00000D00000CF0000CF0000C0D000C0CF00
BFC0003F407000009F300060000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6F000000009F909000A0A000B0B000C0C0000000DF0000EF0000EF0000F0E00000EF0000EF0000\
0000FF0000F00000FF0000FF0000FF0000F00000FFFF0000FF0000F00000FF0000
0000FF0000FF0000F00000F00000F00000F00000EFEF0000E00000E000000000D00000D00000C0\
0000BF0000B0AF000000008F00007F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000007F00008F7000007F000080000030307F90A000
9FAF009FAF00A0B000A0BF003F40900000DF0000E0BF0000BF0000BF0000C00000C000000000F0\
0000F00000FF0000F00000FF0000F00000FF0000FF0000FFDF0000DF0000DF0000
0000FF0000F00000FF0000EFDF0000D000000000CFBF00000000AF3F3F6FB0AF0000008FA00000\
908F00908F0000006F00006F0000608070007F6F007F0000700000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00006000007F6F00007F00000000A030307FA0A000AFAF00B0B000BF8000C00000
CF0000D00000DFE0000000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000\
FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000
FF00000000FF0000FF0000EF4F009FCF0000FF0000FF0000FF0000FF0000FF0000F0F0005050AF\
0000F00000F00000EF0000F00000EF0000EF0000E0E0E000E0DF00E0DF00DFD000
0000E00000DFDF0000D00000D00000CF0000CF0000C0BF00C0BF00C0B000C00000BF0000BF0000\
B00000B000003F007FB0A000AF9F00AF9F000000AF0000AF0000A00000AF0000A0
9F8F00908F0030306F00009F00009F00009F8F5F008000008F00008000008F00008080008F0000\
0000F04F00AFDF0000E000000000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FF0000FFF05000F0FF00F0FF00F0FF00F0FF00EFFF00F0FF00EFF000EFFF000000F00000F0\
0000EF0000F00000EF0000EF0000E0DF0000D00000D00000CFDF00CFDF00C0CF00
BFCF000000AF0000A0900000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
7F00000000A0A0A000AFAF00BFBF00C0C000CFD0000000EF0000F00000F0EF0000EF0000EF0000\
0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000
0000FF0000FF0000FF0000F00000F00000F05000A0EF0000EF0000DF00000000DF0000CF0000C0\
0000BF0000B0A0000000008F700000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006060000000807000007000008000000000B00000BF
90A0009FB000A0B000AFBF00AFBF00B0C000B00000BF0000BF0000C00000C000000000F00000F0\
0000F00000F00000FF0000F00000FF0000F04F00AFD00000DF0000DF00000000FF
0000F00000FF0000EFDF0000D000000000D0C0CF00C00000B00000B0B000B0AF0000009F00008F\
5F30409F8F009000008F000000007000006F8F7F007F6F0000006000005F7F0000
605F00605F00000000000000000000000000000000000000000000000000000000000000000000\
00006F6000007000007F00000000AF0000B0A0AF00AFAF00B0BF00BF0000CF0000
CF0000D00000DFE000E0EF00E0E0000000FF0000FF0000FF0000F0F00000F00000FF0000F00000\
FF00000000FF0000FF0000F00000FF0000FF0000FF0000F0FF0000FF0000FF0000
0000F00000FF0000EF0000E0D000000000FF5000AFFF0000F00000FF0000FF0000FF0000F00000\
0000F00000F00000F00000EF0000EF0000EF0000EFE0DF00E0E000E0DF00E0DF00
0000DF0000E00000DF0000DF0000D0400090CFC000CFC000C0BF00C0BF00C00000C00000BF0000\
BF0000B00000B00000AF0000B00000AF0000AF0000A00000A00000A00000A00000
9F8F009F90009000009F00009000009000008F00008F00008F00008F00008000008F8F000000A0\
0000AFBF0000DFE000D00000DF0000DF00000000FF0000F00000FF4F50AFEFFF00
EFF000EFFF00EFFF00F0FF00EFF000F0FF00EFF000F0FF00EFF0000000F00000F00000F00000EF\
0000F00000EF0000EF0000EF0000EF40009FDF0000CFE000CFDF00C0D000C0D000
0000B00000AF00009F8F9000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00008F0000AF0000B0B0B000BFBF00C0C000CFCF00D0DF000000EF0000F0E00000EF0000EF0000\
0000FF0000FF0000FF0000F00000FF0000FF5000AFF00000FF0000FF00005000AF
0000F00000FF0000F00000F00000EF0000F0EF0000EF0000E00000DF00000000D00000CF0000BF\
0000BF3F0070A0000000007F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F70007000007F00008000000000B00000BF
0000C0303F8FA0BF00AFBF00B0C000B00000BF0000BF0000C00000C000000000F00000F00000FF\
0000F00000FF0000F00000FF0000FF0000FFDF0000DF0000DF00000000FF0000FF
0000FF4F4FA0DF0000D000000000DFCFD000CFD000C00000C00000BFBF00BFBF0000009F00009F\
000090A09F009F30009F00009000009000000000708F80008F700000006F000060
0000607F00007F00006F6000000060000000000000000000000000000000000000000000000000\
00006F6F00007000000000A00000AF0000BF0000C0B0B000BFBF00C00000C00000
D00000D09000E0E000E0E000EFEF00EFEF00F0F0000000FF0000FFF00000FF0000FF0000FF0000\
0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF00000000FF
0000FF0000F00000EFD00000AFBF000000FF0000FFFFFF00FFAF00FF0000FF0000FF0000FF0000\
FF0000FF00004F00A00000FF0000F00000F0EFE000EFEF00EFE000EFE0000000E0
0000EF0000E00000E00000DF0000E00000DFD0CF00D0CF00D0CF00CF4000CF0000C00000CF0000\
C00000C00000BF0000BF0000B00000B00000AF0000B00000AF0000AF0000A00000
A09F00A09F00A000009F00009F00009000009F00009000009000009000009000008F8F000000B0\
0000B0A0A000BF0000DFE000DFE000DF0000DF0000E0AF00E0FF00EFFF00E0FF00
EFFF00EFFF00EFFF00EFFF00F0FF00EFFF00F0FF00EF0000F000004F00A00000FF0000F00000F0\
0000F00000F00000EF0000F00000EF0000EFD0EF00D0E000CFDF00CFDF000000BF
0000BF0000AF00009F808F00000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000006F0000
0000A00000AFAF0000B0B000C0C000C0CF00D0D000D0D000E0E0004F00A0EF0000EF0000F00000\
0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF00000000FF
0000FF0000FF0000FF0000F00000F00000F0EF0000E00000E000004F00900000D00000CF0000C0\
0000B0B000009F0000000080000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006060007000007F00008000000000B00000BF
0000C00000CF0000CFAFBF00AF8000B00000B00000BF0000BF0000C09000C0DF00404FA00000F0\
0000FF0000F00000FF0000F00000FFD00000DF0000DF00000000FF0000F00000FF
DFEF00DFEF00D00000DF00000000D0D0DF000000C00000C0C00000C0C0000000AF0000A00000A0\
0000A0AFA000AFA000A00000A000009F00009F000090000090800000006F00006F
00006F00006F000060806F007F00007F000070000000005F00005F000000000000000000000000\
0000706F00007F00000000A00000B00000BF0000CF0000CFBF0000C00000CF0000
D00000DFE000DFE000E0EF00E0EF00EFF000EFEF00F0F000F04F00FF0000F00000FF00000000FF\
0000FF0000F00000FF0000FF0000FF0000F0FF0000FF0000FF00004F00A00000FF
0000F00000EFDF0000CF00000000FF0000FF0000F0FFFF00F0F000FFFF00F00000FF0000F00000\
FF0000F00000FF0000F00000F000000000F0F0EF00EFEF00EFEF00E0E0000000EF
0000EF0000EF0000E00000E00000E00000E0D0CF00D0CF00D0CF000000DF0000D00000D00000D0\
400090C00000C00000BF0000BF0000B00000B00000B00000B00000AF0000AF3000
A0A000AFA000A00000A000009F0000A000009F00009F00003000700000B00000AF90900030307F\
0000BF0000BF0000C0BF00000000F00000F0DF0000DF0000E00000E0F000E0FF00
E0FF00EFFF00E0F0004F50AF0000FFEF0000EF0000EF0000EF0000EF00004F00A00000F00000F0\
0000F00000EF0000F00000EF0000EFDFEF00D0EF00CFE000CFE000C000000000C0
0000B00000AF909F00000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000007F0000
0000A00000B0AF0000BF0000C0C000CFCF00CFD000DFDF00DFE000EF0000EF0000EF0000EF0000\
0000FF0000F00000FF0000FF0000FF0000F0FF0000FF0000FF0000F000000000FF
0000FF0000FF0000F00000F00000EF0000F0E00000E00000DF00000000DF0000CF0000CF0000BF\
0000B0A00000900000808F00000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006F70007F00008000000000B00000BF
0000C00000C00000D00000D0B00000B00000BF0000BF0000C09000C0DF00CFE000CFE000D0EF00\
D0E000DFEF00404FAFDFF000DFEF00DFF000DFF000E0F000DFF000E0FF00DFF000
0000F0DF0000DF0000DF0000D0E00040408F0000D00000C0CF4000C0C000C000000000AF0000AF\
0000A0BFB000B0AF0000009F00008FAF0000A00000A000009F00009F0000000070
00007F00007000007000006F8F7F008070008070000000608000007F00007F000000005F6F6000\
0000707000007F00000000AF0000B00000C00000CF0000D0BF0000CF0000CF0000
DF9F00DFE000E0EF00E0EF00EFF000EFEF00F0F000F0F000FFFF00FFFF00FFFF00FF50000000FF\
0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF00000000FF0000FF
0000F00000E0DF0000B0C0000000FF0000FF5050AFFFFF00FFFF00FFFF00FFFF000000FFFF0000\
F00000FF0000F00000FF0000F00000FF0000F0F000F0F000EFEF00F0EF000000F0
0000F00000EF0000EF0000EF0000EF0000E0DFDF00DFD000DFD0000000DF0000E00000DF0000DF\
0000D00000DFC0BF00C0BF00C0BF000000D00000CF0000CF3F008040008FB0AF00
B0AF00AFA0003F00800000BF0000BF0000BFA0A000A09F000000BF0000BF0000BF30307F9FA000\
0000BF0000C0AFB000BF0000C0CF000000F00000FFDF0000DF0000E000000000FF
0000FF0000FF0000FF0000FF0000FF0000FF5000AFEF0000F00000EF0000F00000EF00000000FF\
0000F00000F00000EFE0F000DFF000DFF000D0EF00D0EF00CF0000CF0000BF0000
0000B00000A08F9F00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000702F00000090
0000AF0000BFB00000BF0000CFCF00CFCF00DFDF00DFDF00E0EF00E00000F00000EF0000803F70\
0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF00000000FF
0000FF0000F00000FF0000F00000F00000EFEF0000E00000E000000000DF0000D00000C00000BF\
0000B0AF0000909F00808F00000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000807000002F00700000B00000BF
0000BF0000CF0000CF0000D0AF0000B00000BF0000C00000BFD000C0DF00C0DF00CFE000CFE000\
D0EF00D0EF00DFF000D0EF00DFF000DFEF00DFF000DFF000E0F0000000F00000F0
0000EFDF0000DF0000DFEF00D0DF000000D00000CF0000CFCFD000CFCF00C0000040007F0000AF\
0000AFBFB000BFB00000009F00009F000090000090A00000A000009F00009F0000
9000009F000000007000007F8F7F008F7F008F7F00302F4F00006F00006F0000600000607F0000\
00007F6F00000000A00000AF0000BF0000C00000CF3F008FC00000C00000D00000
D0DF00DFE000E0E000EFEF00EFEF00EFF000EFF000F0FF00F0F000FFFF00FFFF00FFFF00F0F000\
FFFF00FFFF000000FF0000F00000FFFF0000FF0000F000000000FF0000FF0000F0
E0EF00DFEF00D0DF005000AF0000F00000FFF0FF00FFFF00F0F000FFFF00F0F0000000FF0000F0\
0000FF0000FFFF0000F00000F00000F04F00F0F000F0F000F0F000EFEF000000F0
0000EF0000F00000EF0000F00000EF0000EFDFDF00DFDF00D0D0000000E00000DF0000E00000DF\
0000DF0000DFCFC000C0BF00C0C0000000D00000D00000CF0000CF0000CF3F3F8F
B0AF00B0AF000000C00000C00000C00000C0A0A000A0A000A0A0000000C00000BF0000C09FA000\
0000C00000C00000CFB00000BF0000C0CF00D0DF000000F00000FFDF0000E00000
0000F00000FF0000FF0000FF0000F00000FF0000FF0000FFEF0000EF0000EF0000EF00000000F0\
0000F0E0F000E0F000DFEF00DFF000DFEF00DFEF00CF0000CF0000C00000BF0000
0000AFA0B000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000006060007F20000000A0
0000AFAF0000B000000000E00000DFD0D000D0DF00DFE000E0E000EF0000EF0000F00000EFF000\
F0F0004F4FAF0000FF0000F00000FF0000FFFF0000F00000FF0000FF00000000FF
0000F00000FF0000F00000F00000EF4F00A0E00000E00000DF00000000DF0000CF0000C00000B0\
0000AF9FAF008F9F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000707000008000000000B00000BF
0000C00000C00000D00000D0B00000B00000BF0000BF00000000EF0000EF0000F00000F0D0EF00\
D0EF00DFF000D0EF00DF5000DF0000DF00000000FF0000FF0000FF0000FF0000F0
4F00A0DF0000E0EF00DFEF000000E00000D00000D00000CFD0DF00CFD000CF0000CF0000CF0000\
40407FC0C000BFBF000000AF0000A00000A000009F00009FAFA000AF0000A00000
AF0000A00000A000009F000030005F90800090800090800000007F00007000007000006F000070\
00007F7000000000A00000B00000BF0000CF0000CFBF0000C00000CF0000D00000
0000FF0000FFE0EF00E0EF00F0F000F0F000F0F000F0F000FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000FF0000FFFF00FFFF00EFF000
EFF000DFE000C0CF00EF0000F00000805070F0FF00F0F000FFFF00F0FF005050AF0000FF0000FF\
0000FF0000FF0000FF0000FFF00000F0F000F0F000F0F000F0F000F05000EF0000
F00000EF00000000F00000F00000F04F4FA0E0E000DFDF00E0DF000000EF0000EF0000E00000E0\
0000E00000E0CFCF00CFCF00CFC0000000DF0000D00000DF0000D00000D03F3F8F
BFBF00B0B0000000CF0000CF0000CF0000CF3F3F8FAFAF00AFAF000000C00000CF0000C0AFAF00\
AFB0000000D0AF0000BFBF00BF00000000E00000EFD0E000D0E000DF0000DF0000
E000000000FF0000FF0000FF0000FF0000FF0000FF0000FF5000AFEF0000EF0000EF0000EF0000\
E0F000E0F000DFF000E0F000DFEF000000EF0000E04F009FCF0000C000000000B0
AFC000000090000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000706F002F2F600000AF
0000BFB00000C000000000DF0000E00000E04F4FA0DFE000EFA000EF0000F00000EF0000F0FF00\
F0F000FFFF00F0F0005050AF0000FF0000FFFF0000FF0000FF0000FF00000000FF
0000FF0000F00000F00000F00000F0EF0000EF0000E00000DF00000000D00000CF0000BF404080\
AFBF009FAF00808F00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000007F7F000080900090A000
0000BF0000CF0000CF0000D0AF0000BF0000B00000C000000000EF0000F00000EF0000F00000F0\
0000FF0000F0DF0000DF0000DF0000DF00000000FF0000F00000FF0000F00000F0
DFEF00E0F000DFEF00DF00000000DF0000DF0000D0D0DF00D0DF004F4F8FCF0000CF0000CF0000\
CF0000C0C000C0C0000000AF0000B00000AF0000AF0000A0BFAF00B0A0003F3F6F
AF0000AF0000A00000A00000A00000A090009F8F009F8F009F8F0000008000007F00007F000070\
6000007000000000A00000B00000C00000CF0000D0BF0000CF0000CF00000000F0
0000F00000FF0000F00000FF4F4FA0F0F000F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00\
FFFF00F0F000FFFF00FFFF00FF0000F00000FF0000FFFF00FFFF00F0F000EFF000
DFE000DFE00000009FE00000E00000EFF000EFF000F0F000F0F000F0FF000000F00000FF0000FF\
0000FF0000F00000FF0000FF0000FFF0F000F0F000F0F000F0F000EF4F00F00000
EF0000F00000EF0000EF0000E00000EFEF00E0E000E0E000DFDF000000EF0000EF0000EF0000E0\
0000EF0000E0D0D000CFCF00CFCF000000DF0000DF0000DF0000DF0000D00000D0
BFBF00BFBF003F3F8F0000D00000CF0000D00000CFB0B000AFB0000000D00000CF0000D00000CF\
AFB000AF0000B00000B00000BFC0000000DF0000EF0000EFD0E000D0E000DFEF00
DFEF00E0F000404FA00000FF0000FF0000FF4F4FA0E0FF00E0FF00EF0000E00000EF0000E00000\
E0FF00DFF000E0F0000000EF0000EF0000E00000E00000DFCF0000BF0000B0C000
A0B000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000006F6000807F0000009F0000AF
0000BFBF0000C000000000E00000E00000EF0000EF4F4FA0EF0000EF0000EF0000F00000F0F000\
F0FF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FF5000FF00004F00A00000FF
0000F00000FF0000F00000F00000EFEF0000E00000E00000400090404F90C0D000C0D000B0C000\
AFBF008F9F00000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000007F7000008090008F9F00
90AF009FAF000000D03F0090B00000B00000BF00003F009F0000EF0000EF0000F00000F00000FF\
0000F00000FFDF0000DF0000DF00004F00AF0000FF0000FF0000F0E0F000E0F000
E0F000E0EF00E00000E000000000E00000DFDFE000DFE000DFE0000000D04F0090D00000D00000\
CF0000D00000CFCF000000BF0000B00000B00000B00000B0BFBF00C0BF00BFB000
0000A000009F00009FAF0000B00000AF3000AF9F00A09F00A09F00A00000A000009F000030005F\
6000007F00000000A00000BF0000C00000D0B00000C00000C00000D000000000F0
0000FF0000F00000FF0000FF0000FF0000FFFF5000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FF0000FF0000FF0000FFFF00FFFF00F0FF00EFEF00
E000000000BF0000FF0000FF0000FFE0EF00EFF000EFF000EFF000EFF0000000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FFF0FF00F0F000F0FF00F0F0005000AFF00000
F00000EF0000F00000EF0000EF0000EF0000EF0000E00000E000000000F00000F00000F00000F0\
0000EF0000EFD0D000D0D000D0CF000000EF0000E00000E00000DF0000E00000DF
C0C000C0C000C0C0000000DF0000DF0000D00000DFB0BF00BFBF003F3F900000DF0000D00000DF\
B00000BF0000B00000BF0000BF0000C0CF00C0D0000000F00000F00000FFD00000
DFF000DFF000E0FF00E0FF00E0FF00E0FF00EFFF00E0FF00EFFF00E00000EF0000EF0000EF0000\
E000000000F00000EF0000EF0000E00000EF0000E00000DFC00000BF0000AFBF00
0000A0000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007F70008F80009F90000000B0
B00000BF00000000E00000E00000EF0000EF0000F00000F0F00000EF0000F00000F0A000F0FF00\
F0F000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00
FFFF00F0F000F0FF00EFF000EFFF00E00000EF0000DF0000DFEF00CFDF00CFDF00BFCF00B0C000\
9FAF00809000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000020005F7F00000000AF
8F9F009FAF00A0B000AFBF00AF0000B00000BF00000000E00000E00000EF0000EF0000F00000F0\
0000FF4000A0DF0000DF0000DF00000000F00000FF0000F0E0F000E0F000E0F000
4000A0E00000E00000E000000000E04F4F9FDFE000DFE0000000D00000D00000D00000D0D00000\
D00000D00000D00000CF00000000BF0000BF0000BF0000B0C0C000C0BF00C0BF00
0000AF0000AF0000A00000A000009F00009FB0A000B0A000AF9F00AF0000AF0000AF0000A00000\
6F00007000000000AF0000B00000C00000CFBF0000BF0000CF00000000F00000FF
0000F00000FF0000F00000FF0000F0F00000F00000FF0000F000005000AF5050AF5050AFF0F000\
FFFF00FFFF00FFFF00F00000FF0000FF0000FFFF00F0F000F0FF000000EFE00000
D000000000AF0000EF0000F00000F0E0EF00DFEF00E0F000E0EF00EFF000EF0000F000000000F0\
0000FF0000FF0000FF0000F00000FFF0F000F0FF00EFF000F0F0000000FF0000FF
0000F04F00AFEF0000EF0000EF0000EF0000EF0000EF0000E00000E00000E00000E00000DF0000\
4F00A00000EFDFDF00D0D000D0D00040409F0000EF0000E00000E00000E00000E0
C0C000C0CF00C0C0000000E00000DF0000DF0000DFBFC000BFBF00BFC0003F0090BF0000B00000\
BF0000B00000BF0000BF000040009F0000E0C0D000C0D0000000FF0000F00000FF
D00000DF0000DF00004F50AFDFF000E0FF00E0FF00E0FF004F4FA04F50AFE00000EF0000E00000\
E000000000F00000F00000EF0000EF0000E00000DFC00000BF0000AF00000000A0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000700000807F00908F009F90000000BF
B00000C000000000DF0000E00000E00000EF0000EFEF0000EF0000F00000EF0000F0F000F0F000\
F0FF00F0F000FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00
F0F000F0FF00EFF000EFFF00E0F000E00000E00000DF0000D0E000CFDF00C0D000BFCF00A0B000\
909F00000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000007F00000000A0
0000B00000BFA0B000A0BF00B0C000B0CF00BFD000BFD0000000EF0000EF0000F00000F00000FF\
0000F0DF0000DF0000DF0000DF0000E0F000E0F000E0FF00E0F000E0FF000000F0
0000F0E00000E00000E000004F00A0E0EF00E0EF000000DF0000E00000DF0000DF0000D00000D0\
DF0000DF0000D00000D00000D00000D000000000BF0000C0CFC000CFC000CFC000
0000BF0000B00000B00000AF0000B00000AFBFB000BFAF00BFB000B0AF00B00000B0000000006F\
6F00007F00002F2F700000BF0000C00000D0BF0000CF0000CF00000000FF0000F0
0000FF0000FF0000FF0000FF0000FFF00000FF0000FF0000FF00000000FF0000FF0000FF0000FF\
0000FF0000FF0000FFFF0000FF0000FF00000000FF0000FF0000F00000EFE00000
0000CF0000A0BF00000000E0CFDF00D0E000DFEF00DFEF00E0F000E04F00EF0000E00000EF0000\
EF00005000AF0000FF0000FF0000FFF0FF00F0FF00F0FF00F0F0000000FF0000FF
0000FF0000FF0000FF0000FF0000FFEF0000EF0000EF0000EF0000E00000EF0000E00000E00000\
E00000E00000DFDF00E0E000DFDF00DFDF00D00000DF00004000A04F00A040009F
D09000CF8F00CF8F000000E040009FC00000C00000C00000C0CF00C0C000C00000BF0000C00000\
BF0000BF0000BF00004000A00000E00000EF0000EFCFDF00CFE0000000FF0000FF
DF0000DF0000DF00004F00AF0000FF0000FF0000FF0000FF0000FF0000FFE00000E00000EF0000\
E000000000F00000EF0000EF0000E00000E0CFE000C00000B000000000A0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F00007F0000908F009F9000AFA000AF0000
BF00004000900000E00000E00000EF0000EF0000F0EF0000F00000EF0000F000000000FFFFFF00\
F0F000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00
FFFF00F0F000F0FF00EFF000EFAF00E00000E00000DF0000D0E000C0D000C0D000B0C000A0AF00\
00007F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000007F0000
0000AF0000BF0000BF303F8FAFBF00B0CF00B0CF00BFD000BFD000C0DF00C0DF00D0E000D0E000\
D0EF00D00000DF0000DF0000DF0000DFF000E0F000E0F0000000FF0000F00000F0
0000F0E00000E00000E00000E00000E0F000E0EF000000E00000DF0000DF0000DF0000DFDFE000\
E0E000DF9000DF0000D00000DF0000D00000D00000D04000D0CF00CFCF00D0CF00
0000BF0000BF0000B00000BF0000B00000B0C0BF00C0BF00C0B000C0BF000000A00000AF000060\
6F00007000008F8F000000BF0000CF0000CFBF0000C00000D000000000F00000FF
0000F00000FF0000F00000FF4F00A0FF0000F00000FF00005000AF0000FF0000F00000FF0000FF\
0000FF0000F0FF0000FF0000FF00000000F00000FF0000F00000EFE00000DF0000
0000BFB0B000B00000AF0000B00000C0D000C0D000D0E000D0E0000000FF0000F0E00000E00000\
EF0000E00000EF0000EF00005050AFEFF000F0FF00F0F000F0FF000000F00000FF
0000FF0000FF0000F00000FF0000FFEFF000EFEF00EFEF00E0EF00EF0000E00000E00000E00000\
E00000DF0000E04F00DFE000DFE000DFDF00DF0000D00000DF0000D00000D00000
D00000D00000CF0000CF0000CF0000CF0000C00000CF0000C0CF00C0CF00C0CF00C00000C00000\
C00000BF0000C0D000C0CF000000EF0000EF0000F00000F0D0E000CFE000D0EF00
D00000DF0000DF00004F00AF0000F00000FF0000F00000FF0000FF0000FF4000A0E00000E00000\
E000000000EF0000EF0000E0D0EF00CFE000CFE000BF00000000A0000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000700000000090908F00A09F00AF6F00BF0000
BF00000000E00000E00000EF0000EF0000F00000F0F00000EF0000F000000000F00000FF0000FF\
0000FFF0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00
F0F000F0FF00EFF000EFF000E00000E00000DF0000D0E000CFDF00C0D000B0C000AFBF008F0000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000AF0000AF0000BF0000C0AF0000AFC000BFCF00BFCF00C0D000C0D000CFE000CFE000D0EF00\
D0EF00DFA000DF0000DF0000DF00004F00AF0000FF0000FF0000F00000FF0000F0
0000F04F00A0EF0000E00000EF0000E000000000EF0000E00000E00000E00000E04F4F90E0EF00\
E0E000E0E0000000D0E00000DF0000E00000DF0000DFDF00D0D000DFD000D0D000
0000C00000C00000C00000BF0000C00000BFD0CF00CFC000CFC000CFC0000000BF0000B000006F\
6F00007F80008F90009FA000303F80B00000BF0000CF00004000A00000FF0000F0
0000FF0000F00000FF0000FFFF0000FF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF\
0000FF0000FFFF0000FF0000FF00000000FF0000FF0000F00000E0E00000DF0000
0000BFAFB00040007F0000B00000B00000B0BF0000BFCF00CFDF000000EF0000FF0000FF0000FF\
4F00AFE00000E00000EF0000E00000EF0000EFA000F0FF00EFF0000000FF0000FF
0000FF0000FF0000FF0000FF0000FFEFF000EFF000EFF000EFF0000000FF0000FF0000FF0000FF\
4F00AFE00000E00000E0EF00DFE000E0E000DF0000DF0000DF0000DF0000D00000
DF0000D00000D00000D00000D00000CF0000D00000CF0000CF0000C0D000CFD000CFD0004000A0\
0000EF0000EF0000EFCFDF00CFD0000000F00000F00000FF0000FF0000FF703070
DF0000D00000DF0000DFF0000000FF0000FF0000FF0000FF0000FF0000FFE00000E00000E00000\
DF0000DFF000D0EF00D0EF00CFE000CFE0000000D00000AF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F00002F00600000A0A09F00AFA000BF0000BF0000
CF8F000000E00000EF0000EF0000F00000F0F00000F00000F00000F000000000FF0000FF0000FF\
0000FF0000FF0000FF5050AFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0FF00
F0FF00EFF000F0FF00EFF000EF0000E00000DF0000D0E000CFDF00BFCF00BFCF009F000000007F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000AF0000B00000C00000C0AF0000AF0000BF00000000DF0000E00000EF0000F00000EF\
404FAF0000F0DF0000DF0000DF0000DF00000000FF0000F00000FF0000F00000F0
0000F0EFF000E00000EF0000E00000EF0000E000000000EF0000E00000EF0000E0E0EF00E0E000\
E0EF000000DF0000DF0000DF0000DFDF0000E00000DFDF00DFDF00DFD000DF4F00
D00000DF00004000800000CF0000C00000C0D0CF00D0CF00CFC000D0CF000000BF0000C000006F\
6F00007F7F008F90009F9F00AFAF00B0B000BF0000C000000000F00000F00000FF
0000F00000FF0000F00000FFF00000FF0000FF00005000AF0000F00000FF0000FF0000FF0000F0\
0000FFFF0000FF0000F000000000FF0000FF0000F00000EF0000E0E00000DF0000
BFC0000000700000BF0000C00000BF0000C00000C0D0C000B00000BF00000000E00000EF0000EF\
0000FF0000FF0000FFDF0000E00000E00000EF0000E00000EF00004F00AF0000FF
0000F00000FF0000FF0000FF0000F04F50AFEFF000EFF000E0EF000000FF0000F00000FF0000F0\
0000FF0000F00000FFDFE000E0EF00DFE0000000F00000F04F00A0DF0000DF0000
D00000D00000D00000D00000CF0000D000004000A00000F00000EFCFDF00CFD000CFDF000000EF\
0000EF0000EF0000F00000EFCFDF00CFDF000000FF0000F00000FF0000FFD00000
D00000DF0000DF0000DFFF00DFF000E0FF00DFF000E0FF00DFF000DFFF00DFF000E0FF00DFF000\
DFF000D0EF00D0EF00404F900000DF0000D0A0B000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000006000007F00000000900000AF30307FB00000BF0000C00000
CFC00040409F0000E00000EF0000EF0000F0EF0000F00000EF00005000AF0000FF0000FF0000F0\
0000FF0000FF0000FF4F00A0FF0000FF0000FF0000F04F00FFFF00F0F000F0FF00
EFF000EFF000EFF000EF0000E00000E00000D0E000CFDF000000C00000BF0000A0900000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000AF0000B00000C0A00000B00000B0000040009F0000E00000EF0000EF0000F0\
0000F00000F0D00000DF0000DF0000E000000000F00000FF0000FF0000FF0000F0
EFFF00EFF000EFFF00EF0000EF0000EF0000EF0000EF00000000F00000EFEFF000E0EF00EFF000\
E0EF000000E00000DF0000E00000DF0000E00000DFEFEF00E0E000E0E000DF4F00
E00000DF0000E00000DF0000DF0000DF0000DFD000D0D000DFD000D0D0000000CF0000C00000CF\
6F00008080008F9000A0A000AFAF00B0BF00BFC000CF4F000000F00000FF0000F0
0000FF0000FF0000FF4F00AFFF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF\
0000FFFF0000FF0000FF00000000FF0000FF0000F00000F0EF0000E00000D0DF00
C0CF000000700000CF0000CF0000CF0000CF0000D040408FDFD000DFCF00B00000B000000000E0\
0000E00000EF0000F00000FF0000FFE0F000E00000E00000E00000EF0000E00000
EF0000EF00000000FF0000FF0000FF4F50AFF0FF00EFF000EFFF000000FF0000FF0000FF0000FF\
0000FF0000FF0000F0E0EF00E0EF00E0EF00DFE0000000FF0000F00000FF0000F0
0000FF404FA0DFE000D0DF00DFE0000000F00000F00000F00000F00000F0D0E000CFDF00D0E000\
0000F00000F00000F00000FF0000F0D0EF00D0E000D0EF000000FFD00000D00000
DF0000DF0000DF00000000FFE0FF00DFFF00E0FF00DFFF00E0FF00DFFF00E0FF00DFF000E0FF00\
4000A00000EF0000E00000DF0000D09FB000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000006000007F000000008F0000AF0000B00000C0B00000C00000CFC000
D0D000D0D0004F4FA00000EF0000F00000F0F00000F00000F000000000FF0000FF0000FF0000FF\
0000FF0000FF0000FFFF0000FF0000FF0000FF00000000FF0000FF0000FF0000F0
0000F00000F00000F0E00000E00000DF00000000DF0000CF0000C00000B0A0000000007F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008F900030307F0000CFA00000B00000B000000000E00000E00000EF0000EF\
0000F00000F04000A0D00000DF0000DF00004F00AF0000F00000FFE0F000EFFF00
E0F000EFFF004F4FA00000F0EF0000EF0000EF0000EF0000EF0000EF0000EFF000EFF000EFEF00\
0000EF0000E00000E00000E00000E00000DF0000E0E0E000EFEF00E0E000E00000
E00000E00000DF0000E00000DF0000E00000DF0000DF0000DF0000DF000040008F0000CF0000CF\
7000002F2F6F9090009FA000AFB000B0BF00BFC000C0CF00D0DF004040A00000FF
0000F00000FF0000F0F00000F00000FF0000F000000000FF0000FF0000FF0000F00000FF0000FF\
FF0000F00000FF0000FF00000000FF0000F00000F00000EFEF0000DF0000D0DF00
0000AF00007F0000CF0000D00000D00000D00000D04F4090DFD000E0DF00DFD000E0DF000000DF\
B00000B000000000E00000E00000F0D0E000DFEF00DFEF00E0EF00DF0000E00000
E00000E00000E00000E00000E000004F00AFE0F000EFF000E0F000EFF0000000F00000FF0000F0\
0000FF0000F00000FFE0EF00E0EF00E0EF00E0EF000000F00000FF0000F00000FF
0000F00000F0DFE000DFE000D0E0000000F00000F00000F00000F00000F00000F0D0E000D0E000\
D0E0000000F00000FF0000FF0000FF0000F0D0EF00D0EF00DFEF00D0EF00DF0000
D00000DF00000000F00000FF0000F00000FF0000F00000FF0000F04F00AFDF0000DF0000D00000\
0000EF0000DF0000DF0000CF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000605F0070000000008F0000A00000AF0000BFB00000C00000C00000D0CF00
D0D000DFDF00DFDF004F4FA00000F0F00000EF0000F00000F000000000FF0000F00000FF0000FF\
0000FF0000F00000FFFF0000FF0000F00000FF00000000FF0000FF0000F00000F0
0000EF0000F0E00000E00000DF00004F009F0000CF0000C00000B00000AF8F0000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008F900090A000A0B000AF0000B000000000DF0000E00000E00000EF\
0000EF0000F00000F0DF0000DF0000E00000DFF000E0F000E0F000EFFF00E0FF00
EFFF000000F00000FF0000F00000FFEF0000F00000EF0000F00000EFF000F0FF00EFF0005050A0\
0000EF0000EF0000EF0000EF0000E00000EFEFEF00EFEF00EFEF00EFEF000000E0
0000E00000DFEF0000E00000EF0000E00000EF0000E00000E00000E00000E00000E00000E00000\
7000000000A0909000A0AF00AFB000BFC000BFCF00CFD000D0DF00DFE000E0E000
EFEF004F4FAF0000FFF00000FF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF\
FF0000FF0000FF00000000FF0000FF0000FFF0F000F0F000E00000E00000CFD000
0000AF9000000000DF0000DF0000DF0000DF0000E04F009FEF0000E00000EF0000EF0000EF0000\
EF0000F00000EF0000B00000BF0000C0DF00CFDF00D0E000D0E0000000FF0000FF
0000FF4F00AFE00000E00000E00000E00000EFF000E0F000EFF000E0F0000000FF0000FF0000FF\
0000FF0000FF0000FF0000FFE0F000EFF000E0F0000000FF0000FF0000FF0000FF
0000FF0000F04F50AFDFEF00DFEF00DFEF000000FF0000F00000FF0000F00000FF404FAFDFEF00\
D0EF00DFEF000000FF0000FF0000FFDF0000D00000DF0000D0EF00DFF000DFF000
DFF0000000FF0000FF0000FF0000FF0000F00000FF0000F0DF0000DF0000DF0000D000000000EF\
0000DF0000DFBFD000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
60600070000080000000009F0000AF0000BF3F0080BF0000C00000D0CF00D0D000
DFDF00DFDF00E0E000E0EF005050AFF00000F00000F000005000AF0000FF0000FF0000FF0000FF\
0000FF0000FF5000AFFF0000FF0000FF00005000AF0000FF0000F00000FF0000F0
0000F00000EFEF0000E00000DF00000000D00000CF0000BF0000B09F0000000080000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000A0909F00A0B000AF0000B00000B0CF00BFD000C0D000\
CFDF00404FA0404FA0D0E000DFEF00DFEF00DFF000DFEF00E0F000E0F000EFFF00
0000F00000FF0000F00000FF0000F00000F04F00A0EF0000EFF000F0FF00EFF000F0FF00EF0000\
0000F00000EF0000F00000EF0000EF0000EFEFF000EFEF00EFEF004F4F9F0000EF
0000E00000E00000E00000E00000E04F009FEF0000EF0000E00000EF0000E00000EF0000E00000\
70000000009F0000B09FA000AFB000B0BF00C0CF00C0D000D0DF00DFE000E0EF00
E0EF00EFF000EFF000F00000F00000FF00005000AF0000FF0000F00000FF0000FF0000FF0000F0\
FF0000FF0000FF5000F0F000FFFF00F0F000F0F000EFEF00EF0000DF00000000C0
0000AF9F0000E00000EF0000E00000EF0000EF0000EF0000EF0000EF0000EF0000F00000EF0000\
F00000F00000F00000F00000FF00004F00A0BFCF00BFCF00C0D000C0DF000000EF
0000EF0000FF0000FF0000FF0000F0DF0000DFEF00E0F000DFEF00E0F000DF0000E00000DF0000\
E00000E000000000FF0000F0E0F000E0EF00E0F000E0EF000000FF0000FF0000FF
0000F00000FF0000FFE0EF00DFEF00DFEF00404FAF0000FF0000F00000FF0000FF0000FF404FA0\
DFEF00D0EF00DFA000D00000DF0000D00000DF0000D00000DF0000DF0000DFF000
DFF000DFF000DFF000DFF000D0F0000000F00000F0DF0000D00000DF0000CF0000CFE000C0DF00\
BFDF00000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000605F00\
7000007F000000009F0000A00000B00000BFBF0000BF000040009F40409FDFD000
DFD000E0E000E0E000EFEF00EFF000F05000F00000F000000000F00000FF0000FF0000FF0000F0\
0000FF0000FFFF0000F00000FF0000FF00000000FF0000F00000FF0000F00000F0
0000EF4F00A0E00000E00000D000000000D00000C00000BF0000AF90000000006F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000A00000B0A0B000AF0000B00000B0CF00C0D000\
C0D000CFE000CFE000D0EF00D0EF00DFEF00DFEF00E05000E000000000FF0000F0
0000FF0000FF0000FF0000F00000FF0000FF5050AFEFFF00F0FF00EFFF00F0FF00F00000F00000\
F00000F000000000F00000F00000F0F0F000EFF000F0F000EFF0000000EF0000EF
0000EF0000EF0000EF0000E00000EF4F4F9FF0F000EFEF00F0F000EFEF0050009FEF0000F00000\
6F00000000A00000B00000C05F3F5FBFC000C0CF00CFDF00D0DF00DFE000E0E000
EFF000EFF000F0F000F00000FF0000FF0000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0F000F0F000EF00004F00900000BF
0000B0A00000F00000EF0000F00000F00000F00000F00000FF0000F00000FF0000F00000FFAF00\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5050AFBFCF00BF0000
0000E00000E00000EF0000EF0000FF0000FF0000FFDFEF00E0F000DFEF00E0F000DF0000E00000\
DF0000E00000DF0000E00000E00000E00000DF0000E00000E00000E000004F00AF
4F00AF0000FF0000FF0000FFE0F000DFEF00DFEF00704F704F00AFDF0000DF0000DF0000DF0000\
DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF00000000FF0000FF0000FF
0000FFDFF000DFF000DFF000DFF000D0F000DF0000D00000DF0000CF0000CFE0000000D0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000060700000\
800000908F000000AF0000B00000BFBF0000BF0000CF00000000E00000EF40409F
E0E000E0E000EFEF00EFF000F0FF00F0F000F0FF004F00AF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FFFF0000FF0000FF0000FF00000000FF0000FF0000F00000F00000F0
0000F0E00000E00000DF00000000DF0000CF0000C00000B0A0000000007F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000A00000B00000C0A00000B000000000DF\
40409F3F409FCFDF00CFE000404FA00000EFDF0000DF0000DF0000DF00000000FF
0000F00000FF0000F00000FF0000F0EFFF00EFF000EFFF00EFFF00F0FF00EF0000F00000EF0000\
F00000F00000F00000F000000000F0EFF000F0F000F0F000F0F0000000EF0000F0
0000EF0000F00000EF0000EF0000EFF0F000EFEF00F0F000EFEF005050A00000E00000EF0000E0\
6F000000009F0000BF0000C0AF0000B00000C0CF00CFD000D0DF00DFDF00E0EF00
E0EF00F0F000EFA000FF0000FF0000FF0000F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00\
FFFF00F0F000FFFF00FFFF00FFFF00F0F000F0F000EF0000EF00000000D00000C0
0000AFAF0000F00000FF0000F000005000AF0000F00000FFFFF000FFFF00F0F000FFFF00FFF000\
0000FF0000F00000FF0000F00000FF0000F00000FF0000FF0000FFF0F000FFFF00
FFF000FFFF00B00000BF00000000E00000EF0000EFD0E000D0E000DFEF00D0EF000000FF0000FF\
0000FF0000F0DF0000DF0000E00000DF0000DF0000DF0000DF0000DF0000DF0000
DF0000E00000DF0000DF0000DF0000DFEF00DFEF00DFEF00DF4F00DF0000DF0000DF0000DF0000\
DF0000D00000DF0000DF0000DF0000DFEF000000FF0000FF0000FF0000F00000F0
0000F00000F00000F00000F0D00000D00000CF0000CF00000000DF0000CF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000607000007F0000\
9080009F8F000000B00000BF3F0080BF0000C000000000E00000EF0000E00000EF
E0E000EFEF00EFEF00F0F000F0F000F0FF00F0F000F0FF004F4FAF0000FF0000F00000FF0000FF\
0000FFF00000FF0000FF0000FF00000000F00000FF0000F00000F00000EF0000F0
0000EFE00000DF0000DF00000000D00000CF0000B00000AF900000000070000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000AF3F008FAF0000B00000\
0000D00000E00000E00000EF0000EF0000F00000F0DF0000DF0000E00000DF0000
E000000000F00000FF0000FFEFFF00EFFF00F0FF00EFFF00F0FF000000F00000FF4F00AFF00000\
F00000F00000F00000F00000F00000FF0000F0AF00F0FF000000F00000FF0000F0
0000F00000F00000F00000EF0000F0F0F000F0F000F0F000F0F0000000EF0000F00000EF0000F0\
60000000009F0000B00000C0AF0000BF0000C00000404FA0D0DF00DFE000E0E000
EFEF00EFF000F00000FF0000FF0000FF0000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FFFF000000F00000EFEF0000E000000000DF0000C0
0000B0AF00000000700000F00000FF0000F00000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00
FFFF00FFF000FFFF005000AFFF0000FF0000BF0000BF0000C0DF00C0DF00CFE000CFDF000000F0\
0000F00000FF0000FF0000FF0000FF0000FFDFEF00DFF000DFEF00DFF0004F00AF
DF0000DF0000E00000DF0000DF0000DF0000DFF000DFEF00DFF000DFEF00DF0000DF00000000FF\
0000FF0000FF0000FF4F50AFDFF000DFF000DFF000DFF000404FA00000FF0000F0
0000F00000EFD00000CF0000D000000000E00000DF0000C0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000607000007F0000908000\
9F8F00AF9F000000BF0000C0BF0000CF0000CF00000000EF0000E00000EF0000EF
0000F0EFEF00F0F000F0F000F0FF00F0F000FFFF00F0FF00FFFF00F0FF000000FF0000FF0000FF\
0000FFFF0000FF0000FF0000FF00000000FF0000FF0000FF0000F00000F00000EF
E00000E00000E000000000DF0000D00000C00000BFA00000000080000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008080000000B0A00000\
AF00000000D00000DF0000DF0000EF0000EF0000F00000EFDF0000DF0000DF0000
DF0000E00000E0F000EFFF00E0F000EFFF00EFFF00EFFF000000F00000FF0000F00000FF0000F0\
0000FFF00000F00000F00000F00000F00000FF0000F00000F00000F000000000FF
0000F00000F00000F00000F0F0F000F0F000F0F000F0F000F0F0000000F00000EF0000F00000EF\
0000F00000900000B00000BFAF0000B00000C000000000EF0000F00000F00000FF
4F4FA0F0F000F00000FF0000F00000FF0000FFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000\
FFFF00FFFF005050AF0000F00000FF0000EF0000EFEF0000EF00000000D00000CF
C0C000B0000000007F0000FF0000F00000FF0000F00000FFF0F000FFFF00FFF000FFFF00F0F000\
0000FF0000F00000FF0000F00000FF0000FF0000FF0000F00000FFFFF000FFAF00
F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000B00000BF0000BF0000\
40009F0000E00000EF0000E00000EF0000EF0000F0D0E000D0EF00D0EF00DFEF00
0000F00000FF0000FF0000FF0000F00000FF0000FFDFF000D0EF00DFF000DFEF000000FF0000F0\
0000FF0000F00000F00000F00000F00000EFD0EF00CFE000D0EF00CFE000D0EF00
CFE000CFEF00C0E000BFD000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006000000000607000007F0000302F609F8F00\
A09F00AFA000B0AF00BF0000C00000CF00000000E00000E00000EF0000EF0000F0
0000EF0000F0F0F000F0FF00F0F000F0FF00F0F000FFFF00F0F000FFFF00F0F0005050AF0000F0\
FF0000FF0000FF0000F000005000AF0000FF0000FF0000F00000F00000EF0000EF
E00000E00000D000000000D00000C00000BF0000AF900000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000008F9000\
0000BFAF0000AF00000000DF0000DF0000E00000E00000F00000EF0000F0D00000
DF0000DF0000E00000E00000EFFF00EFF0005050AF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FFF0FF00F0AF00F00000FF0000F00000FF0000F00000FF0000F00000
FF0000F000005000AF0000F0FFFF00F0F000FFFF00F0F0000000FF0000F00000F00000F00000F0\
0000F07000000000AF0000C0AF0000BF0000C000000000F00000F00000F00000F0
0000FF0000FFFF0000FF0000FF00008F3F705050AF5050AF5050AF5050AFFFFF005050AFFF0000\
FF0000FF00000000FF0000FF0000F00000F00000EFF00000EF00004F009F0000CF
CFCF00B0000030005F00006F0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00\
0000FF0000FF0000FF0000FF0000FF5000AFFF0000FF0000FF0000FF0000FF0000
FF0000FF0000FF0000FF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF5050AF\
FFFF00FFFF00FFFF00BF0000BF0000C00000C000000000EFC0DF00CFDF00C0DF00
CFE0000000EF0000EF0000EF0000F00000EF0000F00000EF0000F0CFE000D0EF00CFE000D0EF00\
0000EF0000F00000EF0000F04000A0CF0000CF0000CF0000CF00000000DF0000C0
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000006000006F7F00007F00000000900000A0AF9F00\
AFA000B0AF00BFB000CF0000CF00000000E00000E00000EF0000EF0000F00000F0
0000FFF00000FF0000F0F000FFFF00F0FF00FFFF00F0F000FFFF00FFFF00FFFF00FFFF00FFFF00\
FF0000FF0000FF0000FF00000000FF0000FF0000F00000FF0000F00000F04F00A0
E00000DF00004F009F0000CF0000C00000B09F0000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000008F9000A0B000AFBF00B0C000B0CF00C0D000C0D000CFE000CFE000D0E000
D00000DF0000DF0000DF0000DF00004F00AF0000F00000FF0000F00000FF0000F00000FF0000FF\
0000FFEFF000F0FF00F0FF00F0FF00F0F000F00000F00000F00000F00000FF0000
F00000FF0000F00000FF0000F0F000FFFF00F0F000FFFF000000F00000FF0000F00000F00000F0\
0000F06F00000000AF0000BF3F008FB00000C000000000EF0000F00000F00000F0
0000F00000FFF00000FF0000FF00000000FF0000F00000FF0000FF0000FF0000F00000FFFF0000\
FF0000F000000000FF0000FF0000FF0000F00000EFEF0000EF0000E00000D0DF00
C0CF00404070A0000000007F7000000000FF0000F00000FF5050A0FFFF00F0F000FFFF00FFF000\
8F5070F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000
FF0000FF0000F0F0005050AF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
FFFF00F0F000F0F000EFF000EF0000EF0000EF0000E00000E00000DF00000000EF
0000EFBFD000BFD000BF0000BF0000C00000BF0000C00000C00000C00000BF0000C00000C00000\
C00000BF0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000006F600000006F7F00007F000000009000009F0000AFAF9F00\
B0AF00BFB000C0BF00CFC0000000E00000E00000EF0000EF0000F00000EF0000F0
F00000FF0000F00000F00000804F70FFFF00F0F000FFFF00F0F000FFFF00F0F000FFFF00FF0000\
FF0000F00000FF0000FFFF000000FF0000F00000F00000F00000F00000EFE00000
DF0000DF00000000D00000C00000B03F4070708000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000303F7FA0BF00B0C000B0C000BFD000BFD000CFDF00CFE000
D0EF00D0E000DF0000DF0000DF0000DF0000E00000E000000000FF0000F00000FF0000FF0000FF\
EFFF00F0FF00F0FF00F0FF00F0FF000000FF0000FF0000FF0000FF0000FFF00000
FF0000F00000FF0000FFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000FF0000FF0000FF0000\
0000F06F70000000A00000BF0000C0B00000BF00000000F00000F00000FF0000F0
0000FF0000F0FF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF5000AFFF0000\
FF0000FF00000000FF0000FF0000F00000FF0000F0F00000EF0000EFEF00DFDF00
CFCF000000AFB00000000080807F000000FF5000AFFF0000FF5000FFFF00FFFF00FFFF00FFFF00\
FF5000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF00000000FFFFFF00
FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
EFA000F00000EF0000EF0000E00000E000004F00A00000EF0000EF0000EF0000DF
0000D03F0080AF0000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F600020004F7F000080000000009000009F0000B00000B0B0AF00\
BFB000C0BF00CFC000D0D0000000E00000EF0000EF0000F00000EF0000F00000F0
FF0000F00000FF0000F000000000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000\
FF0000FF0000FF0000FFFF00FFFF00FFFF00F0F000EFF000E0F000EF5000E04F00
DF5000404F90CFDF00BFCF00B0C000800000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000030007FAF0000AF00000000D00000DF0000E0
0000E00000EF0000EF0000F0D00000DF0000DF0000E00000DF0000E000000000F00000FFE0F000\
EFFF00EFFF00F0FF00EFF0000000FF0000FF0000FF0000F00000FF0000FF0000FF
0000F00000FFF0FF00FFFF00F0F000FFFF00F0FF00FF0000F00000FF0000F00000FF0000F00000\
FF00006060007F80000000B00000C0AF0000BF00000000EF0000F00000F00000F0
0000F00000FFF00000FF0000F000000000FF0000FF0000FF0000F00000FF0000FF5000AFF00000\
FF0000FF00000000FF0000F00000FF0000F00000F0EF0000EFF000E0EF00DFDF00
0000BF0000B0B00000A000008F8000800000F00000FF0000F04F00FFFF00F0FF00FFFF00F0F000\
FF5000F00000FF00004F00A00000FF0000FF0000FF0000F00000FF0000FFFFFF00
F0F000FFFF00FFFF00FFFF000000F00000FF0000F00000FF0000F00000F04F00A0EF0000E00000\
E00000E00000E00000DFE000DFEF00D0E000D0E0000000D00000D00000C0AF0000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006F60007060007F000080000000009F0000A00000AF0000B00000C0BF3F00\
C0BF00CFC000D0CF00D0CF00DFD0000000EF0000EF0000EF0000F00000F0F00000
F00000FF0000F000000000FF0000F00000FF0000FF5050AFF0F000FFFF00FFFF00FF0000F00000\
FF0000FF0000FFFF00F0F000F0FF00EFF000EFF000E0F000E0F000DFEF00DFF000
CFE000C0DF00BFCF00B0C0008F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000090A00000AF00000000D0
0000DF0000DF0000E00000EF0000F00000EF4F00A0D00000DF0000DF0000E0F000E0EF00EFF000\
E0F000EFFF00EFF0005050AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
0000FFF0FF00FFFF00F0FF00FFFF00F0FF000000FF0000FFFF0000F00000FF0000F00000FF0000\
FF00006060007F7F000000B00000C0B00000BF00000000EF0000F00000F00000F0
0000FF0000F0F00000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000\
FF0000FF00000000FF0000FF0000FF0000F0F0F000F0FF00EFF000EFF0004F4F8F
0000CF0000BF0000B0AF00009F90008F8000800000FF0000FF5000FFFF00FFFF00FFFF00FFFF00\
5050AF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00
FFFF00FFFF00FFFF00F0F0000000FF0000F0F00000EF0000EF0000EF0000EF0000E00000E00000\
0000EFDFEF00D0E000D0E000C0D000BFCF00B0C000AFBF00000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000607060007060008070008F000000009F0000A00000B00000B00000C03F0080C00000\
CFC000D0CF00D0D000DFDF00DFDF004F4FA00000EF0000F00000F0F00000F00000
FF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FFFFFF00FF0000FF0000FF0000\
FF0000FFFF00FFFF00FFFF00F0F000F0FF00EFF000EFF000E0F000E0F000D0E000
CFE000BFD000B0CF00A07F002F2F4F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000009F0000
A00000B000000000D00000DF0000DF0000E00000EF0000EFD0E000DFEF00D0E000DFEF00DFEF00\
E0F000E04F00E000000000F00000FF0000F00000FF0000F00000FF0000F00000FF
0000FFF0FF00F0F000F0FF00F0FF005050AF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF803070707F008F90009FA000A00000BF00000000E00000F00000EF0000F0
0000F00000FFF00000FF0000F000000000FF0000F00000FF0000FF0000FF0000F00000FFFF0000\
FF0000F000000000FF0000FF0000FFF0F000F0FF00F0F000F0F000EFEF000000DF
0000CF0000C00000B0B00000A09F009F900000006F0000605050AFFFFF00F0F000FFFF00F0FF00\
5050AF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0F0FF00
F0F000F0F000EFF000EFF000E00000EF0000E00000E00000DF0000E000000000EF0000EF0000E0\
0000DF0000D0BFCF00AFBF00A0BF00000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000600000
0000607060007060008070008F7F0000009F0000A00000B00000B00000C03F0080C00000CF0000\
D0CF00D0CF00DFDF00DFDF00E0E000E0E0000000F00000F05000AFF00000FF0000
F00000FF00000000F00000FF0000FF0000FF0000F00000FF0000FF5000AFF00000FF0000FF0000\
FF0000F0F000FFFF00F0F000F0F000EFF000EFF000E0EF00E0F000D0EF00CFE000
BFD000B0CF00A0BF0000007F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000808F00AF0000AF0000B0C000B0CF00C0D000C0DF00CFE000D0E000D0EF00D0E000DFEF00\
DF0000E00000DF0000E00000E00000EF0000E000000000FF0000F00000FF0000FF
F0FF00F0FF00F0FF00F0FF00F0FF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FFFFFF007070008F900090A000AF0000B000000000EF0000EF0000F00000F0
0000FF0000F0F00000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000\
FF0000FF00005050AFFFFF00F0FF00FFFF00F0F000F0FF00EFF000F000000000DF
0000D00000C00000BFBFBF00B0AF0030305F00007F800000807000FFFF00FFFF00FFFF00FFFF00\
FFFF000000FF0000FF0000FF0000F00000FF0000F00000FF0000F0F00000EFF000
EFF000E0EF00EFF000E0EF00E0EF004F00A00000EF0000EF0000EF0000DF0000DF0000D00000CF\
AF0000A00000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000000006F00006F
2F00507F6F008F7F009080009F90000000AF0000B00000BF0000C0BF0000CF0000CF0000D00000\
0000E0E0DF00DFDF00E0E000E0E000EFEF00EFEF000000FFF00000FF0000FF0000
FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000\
5050AFFFFF00F0FF00F0FF00EFF000EFF000E0F000E0F000DFEF00D0E000C0D000
BFCF00AFBF008F0000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000009F0000A00000AF0000B0CF00B0CF00C0D0000000DF0000EF0000E0\
0000EF0000EF4F00A0D00000DF0000DF0000E00000DF0000E00000E00000EFA000
E0F000EFF000EFF000EFFF000000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
5050AFF0FF00F0FF00808F0090A000A00000B00000BF0000404FA00000EF0000F0
0000F00000FFEF0000FF0000F000000000FF0000FF0000FF0000F00000FF0000FF0000FFF00000\
FF0000FF0000FFFF00F0F000FFFF00F0F000F0FF00EF4F00F00000EF00004F009F
0000D00000CF0000BFCFC000B0AF003F0060000080000070807000807000F0FF00FFFF00F0F000\
F0F0000000F00000F0EF0000EF0000EF0000EF0000E00000E00000E00000E0EF00
DFE000DFEF00DFE000DFEF00D0E0000000DF0000D00000D00000C0AF0000A00000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000006000006F00006F7F0000
8000008F7F00908000A09000A090000000B00000BF0000C0BF0000CF0000CF0000D000000000E0\
0000EF4F409FE0E000E0E000EFEF00EFEF00F0F000F00000FF0000F00000FF0000
0000F00000FF0000FF0000FF0000F00000FF0000FF0000FFF00000FF0000FF0000FF00004F00A0\
0000FF0000F00000F00000EF4F50A0E0EF00E0F000DFEF00D0EF00C0DF00404080
0000B09F0000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000A00000AF00000000CF0000D00000DF\
0000DF0000E00000E00000EF0000EF0000EFD00000DF0000DF0000E00000DF0000
E00000E00000E00000E00000EF00004F00A00000FF0000F00000FF0000F00000FF0000FF0000FF\
F0FF00F0FF00F0FF00808F00909F00A07F00AF0000C00000CFD000D0E000DFE000
E0EF00EFF000F05000F00000FF00008050705050AFFFFF00FFFF00FFFF00FFFF00FFFF00FF0000\
FF0000FF0000FFFF00FFFF00FFFF00FFFF000000F00000F0F00000F00000EF0000
0000DF0000CFD0D000CFC000C0C000B0000000009000007F0000708F7F00802F00EF0000EF0000\
EF0000EF0000E00000E00000E00000E00000DF00004F00A00000EF0000EF404F9F
D0E000CFDF00C0D000B0CF00B0C000AF0000A00000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000060200000006F0000707F00008000008F0000
9000009F8F00A09000AF9F00B0A000B0AF000000CFC00000CF0000CF0000DF00000000E00000EF\
0000EF0000F0E0E000EFEF00EFEF00F0F000F00000FF0000FF0000FF00005000AF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF00000000FF\
0000FF0000FF0000F00000F00000EFE00000DF0000DF00000000D00000CF0000B0
0000AF00006F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000A00000\
AF00000000CF0000D00000D00000DF0000DF404F9FCFE000D0E000D0E000D0E000
D00000DF0000DF0000DF0000DF0000E00000DF0000E00000E00000EF0000E00000EF0000E0F000\
EFF000EFF000EFF000EFF000909F009FA000AF0000BF0000CFD000CFDF00DFE000
DFE000EFF000EFF000F0FF00F0FF00FFFF00F0F000FFFF00FFFF00FFFF00F0F000FFFF00FF0000\
FF0000F00000FFFF00FFFF00FFFF000000F00000F00000F0F00000F00000F00000
0000DF0000D0D0D000D0CF00CFC000C00000AF000000008F00007F9080008F7F00800000DF0000\
DF000040009F0000EF0000E00000EF0000E00000E00000DF0000DF0000CF3F408F
AF0000AF0000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000605F006F600000006F2F00507F00008F00008F0000000090
00009FA09F00AF9F00B0A000B0AF00C0B000C00000CF0000CF0000DF00000000E00000EF0000EF\
0000EF0000EF0000F0EFEF00F0F000F00000FF0000F00000FF0000F0A0000000FF
0000FF0000FF0000F00000FF0000FF0000FF0000F0FF0000FF0000FF0000F000000000FF0000FF\
0000FF0000F00000F00000EF4F00A0DF0000DF00004000900000CF0000BF0000AF
2F0050000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000009FB000AFBF00AFBF00B0C000B0C000BFCF00BFCF00C0DF00C0DF00
0000EF0000E00000EF0000EF0000EF0000EFDF0000DF0000DF0000DF0000E00000DFEF00E0EF00\
DFEF00E0F000E0EF00E000000000A0303F80AF0000BF0000C0D000D0DF00D0E000
E0EF00E0EF00F0FF00F0FF00FFFF00F0FF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FF0000\
FF0000FF0000FFFF000000FF0000FF0000FF0000F00000F0F00000FF0000F00000
0000EFE0E000E0E000D0D000404080C00000BF00003F006000008F9F8F009F8F008F00008F0000\
0000600000E00000D00000D00000CF0000CFAF0000AF00009F0000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006F60006F60007060007F600080000080000090000000009000009F0000A0
0000B03F307FBFAF00BFAF00C07F00C00000D00000D000004F009F0000EF0000EF0000EF0000F0\
0000EF0000F00000F00000FFF00000FF0000FF0000FF0000FF0000FFFF00FFFF00
5050AF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF00005000AF0000FF0000FF\
0000F00000F00000EF0000EFDF0000DF0000D000000000D00000BF0000B0900000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000A00000A00000AF0000
0000CF0000D00000D00000DF0000DF0000E00000E00000EF0000E00000EFD0E000D0E000D0E000\
D0E000D0E000DF4F00D000006000000000B0A00000B00000C0D000CFD000D0E000
DFE000E0EF00EFF000F0FF00F0F000F0FF00F0FF00FFFF00F0F000FFFF00FFFF00FFFF00F00000\
FF0000FF0000FF00000000F00000FF0000F00000FF0000F00000F0F00000F00000
EF0000EFEF00DFE000DFDF000000C00000BFC00000BF0000000090A09F009F900000007F8F0000\
8F0000000060706000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000200040\
00006F7060007F60007F00008000008F00009000000000900000A00000AF0000B0
0000B00000BFBFAF00C00000C00000D00000D000000000EF0000E00000EF0000EF0000F00000EF\
0000F00000F00000F0F00000FF0000F00000FF0000F00000FFFF00F0FF00FFFF00
F0F0005050AF0000FF0000FF0000F0FF0000FF0000FF0000F00000FF00000000FF0000FF0000F0\
0000F00000EF0000EFDF0000DF0000D000000000D00000C00000B00000A0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000A00000A00000AF0000AF00000000CFB0C000BFCF00B0CF00BFCF00BFCF00C0D000\
C0D0000000E00000DF0000E00000E00000AF0000BFB000000000E0404FA0D0DF00
DFEF00E0EF00EFF000F0F000FFFF00F0FF00FFFF00F0FF00FFFF00FFFF005050AF0000FFFF0000\
FF0000FF0000FF00000000FF0000FF0000FF0000F00000FF0000F0FF0000F00000
FF0000EFEF00EFEF000000D00000CF0000C00000BFC00000BF0000B0A000AFA00000008000007F\
9000009000008F0000807000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000006F00006F000070\
7F60008000008000008F000090000000009F00009F0000AF0000AF0000BF0000BF
0000C00000C0CF0000CF0000D00000D00000E0DF000000EF0000EF0000EF0000F00000EF0000F0\
0000F00000FFF00000FF0000FF0000FF0000FF0000FFFF00FFFF00FFFF00FFFF00
FFFF00FFFF00FFFF000000FFFF0000FF0000FF0000FF0000FF00000000FF0000FF0000F00000FF\
0000F00000F04F009FE00000DF00000000DF0000CF0000BF0000AF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000009F00009F0000\
A00000A00000AF00003000800000CF6000000000B0A00000B000000000E00000EF
0000EF0000F00000F0F00000F00000FF00000000F00000FF0000FF0000FF0000F00000FF0000FF\
FF0000F00000FF00000000FF0000FF0000F00000FF0000F00000F0F00000FF0000
F00000F00000EFEF000000DF0000D00000CF0000BFCF0000C00000BF0000AFA000000090000080\
00007F0000709000008F00008F0000000060700000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F000000006000006F00006F000070800000\
8F00008F0000900000905F00A08F000000A00000AF0000B00000BF0000BF0000C0
C00000CF0000CF0000D00000D00000E0DF00DFD000E0DF000000EF0000F00000EF0000F00000F0\
0000FFF00000FF0000F00000FF0000F000005050AFF0FF00FFFF00F0F000FFFF00
F0FF00FFFF00F0F000FFFF00FF5000FF0000F00000FF00000000FF0000FF0000F00000F00000EF\
0000F00000EFDF0000DF0000D000000000CF0000BF0000AF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000AF300080AF00000000EF0000EF
0000F00000F00000FFEF0000F00000F000005000AF0000FF0000FF0000FF0000FF0000FF0000FF\
FF0000FF0000FF0000FF00000000FF0000FF0000FF0000FF5050A0FFFF00F00000
FF0000F000000000EF0000E00000DF0000CF0000CF0000C0CF0000C00000C0000000009F000090\
00008000008000007F9F000090000090000000006F00006F000060000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006F00006F600000006F00006F0000702F00508F00008F0000900000\
9000009F8F009F8F00A09000AF9F000000B00000B00000BF0000C00000CFCF0000
D00000D00000DF0000DF0000E0DF00E0DF00EFE000E0E0000000F00000F00000F00000F00000FF\
F00000FF0000FF0000FF0000FF00000000FF0000FF5050AFFFFF00FFFF00FFFF00
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FF50000000FF0000FF0000F00000FF0000F00000F0\
0000EFE00000DF0000D000000000CF0000C00000AF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000007000000000B0A00000B000000000EF
0000EF0000F00000F04F00AFEF0000F00000F000000000FF0000F00000FF0000FF0000FF0000F0\
FF0000FF0000FF0000F000000000FF0000FF0000FFF0F000FFFF00F0FF00FF0000
F00000FF0000F000000000EF0000DF0000DF0000CF0000CFD0CF00CF0000C00000C0000000009F\
00009000008F00008F9F8F009F8F009000009000008F000000006F000060000060
700000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006F00006F6000706F007060002F204F0000708000008000008F00009000009F0000000090\
30306FA09000AF9F00AF9F00B0A0000000BF0000C00000C0CF0000CF0000D00000
D00000DF00000000E0E0DF00E0DF00E0E000E0E000EFE000EFE0000000F00000F00000FFF00000\
FF0000FF0000FF0000F000000000FF0000FF0000FF0000F05050AFF0FF00FFFF00
F0F000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00FFFF004F4FA00000F00000EF0000F00000EF\
DF0000D00000D000000000CF0000BF0000AF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000AF9F0000B000000000E0
0000EF0000EF0000F00000F0F00000F00000F000000000FF0000FF0000FF0000FF0000FF0000FF\
0000FFFF0000FF0000FF00000000FF0000FFFFFF00FFFF00FFFF00FFFF00F0FF00
FF0000FF0000FF0000F000000000EF0000DF0000DFDFDF00DFDF00D0D000D00000C00000C00000\
00009F00009F000090AF9F00A09000A0900000007F9F00009000009000008F0000
302F4F806F007F6F00700000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000060000060
00006F7F6F00806F00806F008F70008F00008F00008F00009F00009F000000009F00009F0000A0\
3F3070B0A000B0A000BFAF00BFAF00403F8F40008FD00000D00000DF0000DF0000
E000000000EF0000EF0000EFEFE000EFE000EFEF00EFEF00F0F0004F4FA05000AFFF0000FF0000\
FF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00F0F000F0FF00EFF000EFF0004F4F9FE00000\
DF0000D000000000CF3F4080303F70000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000007070000000B0A000003F009F
0000E00000EF0000EF0000F0E00000F00000F00000F000000000FF0000FF0000F00000FF0000FF\
0000FFF00000FF0000FF0000FFFF00F0F000FFFF00F0FF00FFFF00F0F0000000FF
5000A0FF0000F00000FF00000000EF0000E00000DFE0E000DFDF00DFDF00D0CF00D00000CF0000\
C00000BF000000009FB0A000AFA000A0900000008000007F30005F900000900000
900000907F008F700080700000006000006F0000602F0040000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000070000020004000006000006000006F00006F000070
0000708F70008070008F7F008F7F009000009000003000600000900000A00000A00000AF0000AF\
0000B0BFAF00BFB000C0B000CFBF00CFBF00D00000D00000DF0000DF00000000EF
0000EF0000EF0000EF0000EFEFE000EFEF00EFEF00F0EF00F0F000FFF000F00000FF0000FF0000\
FF00004F00A00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF
FF5000FFFF00F0F000FFFF00FFFF00FFFF00F0F000F0F000EFF000EFF000DFEF00DF0000D00000\
D00000BFD000B0C000A0B000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000008F9000A00000AF0000
0000EF0000EF0000F00000F00000FFEF0000F00000F000000000FF0000FF0000FF0000FF0000FF\
0000FFFF0000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF
0000FFFF0000FF0000FF0000FF00000000EFF0EF00EFE000EFE000E0DF000000D00000C0D00000\
CF0000CF0000C00000BFB000B0AF00B0AF0000009000008F00008000008F000080
9F00009080009F8000907F0090000000006F00007000006F00006F00006000006F7F6000706000\
706000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000700000706000\
7060007F6F007F6F002F204F00006F0000700000700000700000708F00008F0000
900000907F009F80009F80009F8F00302F6000009F0000A00000AF0000AF0000B00000B00000BF\
0000BFCFBF00C0BF00CFC000CFC000D0CF00DF0000DF00004F009F0000EF0000EF
0000F00000EF0000F00000F00000F0EFEF00F0F000F0F000FFFF00FFFF00FFFF00FF0000FF0000\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000
FF0000FF0000FFFF00FFFF00FFFF00F0F000F0FF00EFF000EFF000E0EF00E00000DF0000D00000\
BFD000B0CF00A0BF00000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000909F00AF3F00
0000DF0000EF0000EF0000F00000F0EF0000EF0000F000000000F00000FF0000FF0000FF0000F0\
5050AFFFFF00FFFF00F0F000FFFF00FFFF00FFFF00F0F0000000FF0000F00000FF
0000F00000FFFF0000FF0000F00000F00000F0EF00F0EF00E0E0004F4F900000D00000D00000C0\
D00000CF0000CF0000C0BF00C0B000BFAF0000009F00009000009000008F00008F
000080A090009F8F009F8F00902F009F00009000009000008F000030004F00006F8F7000806F00\
806F007F6F0000006F00006000006F00006000006F0000600000600000607F6000
7060007F60007060002F204F00006000006F00006000006F00006000006F00006F00006F7F6F00\
807000806F008F70002F204F2F00508F00008F00008F00009000009000009F0000
9F00009F8F00A08F00A09000A090003F30700000A00000B00000B00000BF0000BF0000C00000C0\
40008FCFBF00D0C000D0C000DFCF00DFCF00E000000000E00000EF0000EF0000F0
0000EF0000F00000EF0000F00000F05050AFF0F000FFFF00F0F000FFFF00FFFF00FFFF000000F0\
0000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FFFF0000FF0000
F00000FF00000000FF5050AFF0F000F0F000EFF000EFF000DFEF00DF0000D00000D00000BFD000\
B0C000A0B000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000008F90009FA000
B0C0003F409F404FA00000EF0000F04F00A0F00000F00000F00000F0FF00FFFF00F0FF00FFFF00\
FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF
0000FF0000FF5000AFFF0000FF0000FFFF00FFFF00F0F000F0F0000000E00000DF0000D00000D0\
0000CF0000CFD00000D0CF00CFC000CFBF003F00700000A000009F00009F000090
000090AF9F00AF9F00A09000A09000A00000A000009F00009F00009F00009F0000900000900000\
9000009050002F2F4F000070000070000070000070000070000070000070806F00
8F70008070008F700080700000007000007000007000007000007F00007000007F2F0050900000\
9000009000009000009F00009000009F00009F0000A000009F0000000090000090
0000A0AF9000B09F00B09F00BFA000B0A0000000B00000B00000C00000C00000CFCF0000D00000\
D00000DF4000DFCF00E0D000E0D000E0DF00E0DF000000F00000EF0000F00000EF
0000F00000F00000F00000F00000FF5000AFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF005050AF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF5000AFFF0000FF0000FF0000
FF00000000FF0000FF0000F00000FF0000F0EFF000E0EF00E00000DF0000D00000BFD000B0CF00\
A0BF00000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000909000
A0B000BFCF00C0D000CFDF00D0E000E0EF00EF0000F00000F00000F0FF00F0F000F0FF00F0FF00\
FFFF00F0F000FFFF00FFFF00FFFF00F00000FF00000000FF0000FF0000F00000FF
0000F00000FF0000F0FF0000FF5000FFFF00F0F000F0F000F0EF000000EF0000DF0000DF0000D0\
0000D00000CF4F408FDFD000D0CF00CFC000CF0000C000000000AF0000A00000A0
00009FB0A000AF9F00AF9F00AF900000009000008F00008F00008030005F9F00009F00009F0000\
9F00009000009F00009000009F0000900000900000900000900000900000907F00
907F00907F00907F00907F009000009000009000009000009000009000009000009F0000900000\
9F00009F00009F00009F0000A0000000009000009F00009F0000A000009F0000A0
0000AFBFA000B0A000BFAF00BFAF00C0B0000000BF0000C00000C0CF0000CF0000D00000D00000\
DF0000DF0000E0DF00E0DF00EFDF00E0DF00EFE0000000EF0000F00000EF0000F0
0000F00000F00000F00000FFF00000FF0000F00000FFFF00FFFF00FFFF00F0F000FFFF00FFFF00\
5050AF0000F00000FF0000FF0000FF0000F0FF0000FF0000FF0000F00000FF0000
0000FF0000FF0000F00000F00000F00000F00000EFDF0000D00000D000003F40800000BF0000AF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
909F00AFB000C0D000CFD000D0E000DFEF00EF0000EF0000F00000F00000F0FF00F0FF00FFFF00\
F0FF00FFFF00FFFF005000AFFF0000FF0000FF0000FF00000000FF0000FF0000FF
0000FF0000FF0000FF5050AFFFFF00FFFF00FFFF00FFFF00FF00000000EF0000EF0000E00000E0\
0000DF0000DFE0DF00E0DF00DFD000DFD000D00000D00000CF0000CF0000400070
0000AFBFAF00BFB000BFAF00B0A00000009F00009F00009000009000009000009000008FAF9000\
A08F00A06000A00000A000009F0000A000009F0000A000009F0000A000009F2F00
9F8F009F8000A08F009F8F00A000009F0000A000009F0000A000009F0000A00000A00000A09000\
A09000AF9000AF9000AF9F0000009F0000A00000A00000AF0000AF0000B00000B0
0000BFBFAF00C0B000C0B000CFBF00CFBF000000CFCF0000D00000D00000DF0000DF0000E00000\
E000000000EFE0DF00EFE000EFE000EFE000EFE000504FA00000F00000F00000F0
0000FF0000F0FF0000FF0000FF0000FF0000FF0000FF5000FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF005050AF0000FF0000FF0000FFFF0000FF0000FF0000FF0000FF00000000FF
0000FF0000F00000FF0000F00000F00000EFE00000D00000D000000000CF0000BF0000AF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000030007FAFBF00C0D000C0D000D0E000DFEF00EF0000EF0000F00000F0FF00F0FF00F0F000\
FFFF000000FF0000FF0000F0FF0000FF0000FF0000F000000000FF0000FF0000FF
0000F00000FF0000F0FFFF00F0F000FFFF00FFFF00FF5000F00000FF00000000EF0000EF0000E0\
0000E00000DFEFE000E0DF00E0DF00DFD0004F008FD00000D00000D00000D00000
CF0000CF0000C0B000C0B000BFAF000000A00000A00000A000009F00009F00009F00009FAF9F00\
AF9F00AF9F00AF9F0000009000009000009000009000008F00009000008F3F3060
A08F00AF9000A09000AF900000009000009000009000009F00009000009F00009F00009FAF9F00\
B09F00B09F00B0A000B0A0000000AF0000AF0000B00000B00000B00000B00000BF
0000BFCFBF00CFBF00CF8000CF0000D00000D00000DF0000DF0000DF0000E000004F00A00000EF\
0000EF0000EF4F4FA0EFE000EFE000EFE000F0EF00EFEF000000F00000F00000FF
4F00A0FF0000FF0000FF0000F00000FF0000FF00000000FFF0F000FFFF00FFFF00FFFF00F0F000\
FFFF00FFFF005050AF0000F0FF0000FF0000FF0000F00000FF00000000FF0000FF
0000F00000F00000EF0000F00000EFDF0000D00000D000000000CF0000BF0000AF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009F00000000DFC0D000CFDF00DFE000DFEF00EF0000EF0000F000000000FF0000FF\
0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF00000000FF0000FF
0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000FF0000FF00000000EF0000F0\
0000EFF0EF00EFE000EFEF00E0E0000000DF0000D00000D04F008FDF0000D00000
DF0000D00000D00000CF0000CF80000000B00000B00000AF0000AF0000AF0000AF0000A0BFAF00\
B0A000BFA000B0A0000000A000009F0000A000009F0000A000009F0000A000009F
B0A000B09F00B0A000B09F000000A00000A00000A00000A00000AF0000A00000AF0000AFBFAF00\
BFAF00BFAF00BFAF00C0B0000000B00000BF0000BF0000C00000C00000C00000C0
D00000D00000D00000D00000DF0000DF0000E00000DF0000EF00000000EF0000EF0000EF0000F0\
0000EF0000F00000EFF0EF00EFEF00F0EF00F0EF00F0F000504FA00000FFF00000
FF0000FF0000FF0000FF0000FF00000000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF00FFAF00FF0000FF0000FF0000FF00005000AF0000FF0000FF0000F0
0000F00000F00000F04F009FDF0000D000004000900000C00000BF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000009FA000000000DF0000E00000EF0000F04F00A0EF0000EF0000F000000000FF\
0000FF0000F00000FF0000FF0000FF4F00A0FF0000FF0000FF0000F000000000FF
0000FFFFFF00F0F000FFFF00FFFF00FFFF000000F0FF0000FF0000FF0000F00000FF00000000F0\
0000F0EFEF00F0EF00EFE000EFEF000000DF0000DF0000D00000D00000D00000D0
DF0000DF0000D00000D00000D00000D00000CF0000CF000040007F0000B00000B00000B0C0B000\
C0B000BFAF00C0B0000000AF0000AF0000A00000AF0000A00000AF0000A00000AF
BFA000BFAF00BFA000BFAF000000AF0000AF0000AF0000B00000AF0000B00000B00000B03F3F7F\
C0B000C0B000CFBF00C0B0000000C00000C00000C0400080D00000D00000D00000
D00000DF0000DF0000E00000E00000E00000E090000000EF0000EF0000EF0000EF0000EF0000EF\
0000F00000EF0000F0EFEF00F0F000F0EF00F0F000F0F000FF0000F00000FF0000
FF0000FF0000F000005000AF0000FF0000FF0000F00000FF0000FF0000FFF0F000FFFF00FFFF00\
FFFF00F00000FF0000FF0000FF0000F000000000FF0000FF0000FF0000F00000F0
0000EF0000EFDF0000DF0000D000000000D00000BF0000B0000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000A0AF00000000DF0000EF0000EF0000F0E00000F00000EF00005000AF\
0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF00008F3F70
FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FFFF0000FF0000FF0000FF0000FF0000\
F00000FFF000F0F000F0F000EFEF000000E00000E00000E00000DF0000DF0000DF
0000DFE0DF00E09000DF0000DF0000DF0000DF0000D00000D00000D00000D00000CF0000CFC000\
CFBF00CFBF00CFBF000000BF0000B00000BF0000B00000BF0000B00000BF0000B0
C0B000C0B000CFBF00C0B0000000BF0000BF0000BF0000BF0000BF0000BF0000C00000C0404080\
CFBF00D0C000D0C000D0C000D00000DF0000D00000DF0000DF0000E00000DF0000
E00000E00000EF0000EFDF00EFE000EFDF00EFE0000000EF0000F00000EF0000F00000F00000F0\
0000F00000F00000F00000FFF0F000FFFF00F00000FF0000FF0000FF0000FF0000
FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5050AFFFFF00FFFF00\
FF0000FF0000FF0000FF0000FF0000FFFF005050AF0000FF0000F00000F00000EF
0000EFDF0000DF0000CF00000000CF0000BF0000AF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000AFAF00000000DF0000EF0000EF0000F0E00000EF0000EF0000\
0000FF0000FF0000FF0000F00000FF0000FF0000FFF00000FF0000FF0000FF0000
F00000FFFF00FFFF00FFFF004F4FA00000FF0000FF0000FF0000F0FF0000FF0000FF0000F00000\
FF0000F00000FFAF00F0EF000000EF0000E00000EF0000E00000E00000DF0000E0
4F4090EFE000E0DF00EFDF00E0DF00E00000DF0000E00000DF0000DF0000DF0000DF0000D0C000\
D0CF00D0C000D0C000CF0000D00000CF0000D00000CF00004000804000800000C0
CFBF00CFC000CF8000D08000400080400080CF0000D00000CF0000D00000D00000D00000D00000\
DFCF00D0CF00DFCF00DFCF00DF4000DF0000E00000DF0000E00000E00000EF0000
0000E00000EFEFE000EFE000EFE000EFE000EFE000504FA00000EF0000F00000F00000F00000F0\
0000F00000F00000FF0000F0FF0000F00000FF0000F00000FF0000FF0000FF0000
0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF5000AFF00000\
FF0000FF0000FF0000F0A000FFFF00FFFF00F0FF00EFF0000000F00000EF4F00A0
D00000D000004000900000C00000B0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000B0B000000000DF0000EF0000EF0000F0E00000F00000\
EF00000000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000
FF0000FF0000FFFF005050AF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000\
FF0000FF0000FF0000FF00000000F00000F00000EF0000EF0000EF0000EF0000E0
F0EF00EFE000EFE000EFE000504F9F0000DF0000DF0000DF0000DFE00000E00000E00000E0DF00\
DFD000E0D000DFCF00E00000DF0000DF0000DF0000DF0000DF0000DF0000D00000
DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000DF0000E00000\
E0D000E0DF00E0D000EFDF00E0DF00EF0000E000000000EF0000EF0000EF0000EF
0000F00000EF504FA0EFE000F0EF00EFEF00F0EF00EFEF000000F00000F00000F00000F00000FF\
0000F05000AFF00000FF0000FF0000FF0000FF0000FF0000FF0000FFAF00FFFF00
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000\
FF0000FF0000FFFF00FFFF00FFFF00F0FF00EFF000EFF000E0EF00E0EF00D04F00
D000000000D00000C00000AF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000009000000000DF0000DF0000EF0000EF4F00A0E00000\
EF0000EF00000000FF0000FF0000FF0000F05050AFFFFF00FFFF00F00000FF0000
FF0000FF0000F000000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FFFFAF00\
FF0000F00000FF0000FF0000FF0000F000000000F00000EF0000F00000EF0000EF
EFEF00F0EF00EFE000EFE0000000DF0000E00000DF0000E00000DF0000E00000DF0000DFE0DF00\
EFDF00E0DF00E0DF00E00000E00000E00000E00000E00000E00000DF0000E00000
DF0000E00000DF0000E00000E00000E00000E00000E00000E00000E00000E00000E00000E00000\
EFE000EFDF00EFE000EFE000EFE0000000E00000EF0000EF0000EF0000EF0000EF
0000EF0000F00000EFF0EF00EFEF00F0EF00EFEF00F0F0000000F00000F00000F00000FFF00000\
FF0000F00000FF0000FF0000FF0000F00000FF0000FFAF00FFFF00F0F000FFFF00
5050AF0000FF0000F00000FF0000FF0000FF0000F00000FFFF0000FF0000F00000FF0000FF0000\
FF0000F0F000FFFF00F0F000F0FF00EFF000E0F000DFEF00DFEF00D0E000CFE000
BFCF00AFC0009FAF00000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000009F00000000DF0000E00000EF0000EFE00000\
EF0000F000004F00AF5050AFF0FF00FFFF00F0FF00FFFF00FFFF00FFFF00FF0000
FF0000FF0000FF0000FF00000000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00\
FFFF00FF0000FF0000FF0000FF0000FF0000FF0000FF00000000F00000F00000F0
F0F000F0EF00F0F000F0EF000000EF0000E00000EF0000E00000E00000E00000E00000E0EFE000\
EFE000EFE000EFE0000000E00000E00000E00000E00000E00000E00000E04F009F
EF9F00EF9000EFE000EFE000EFE0000000E00000EF0000E00000EF0000E00000EF0000EF0000EF\
EFE000F0EF00EFE000F0EF00EFE0000000F00000EF0000F00000EF0000F00000F0
0000F00000F00000F0F0EF00F0F000F0EF00FFF000F0F0005050AF0000F0FF0000FF0000FF0000\
FF0000FF0000FF0000FF0000FF00005050AFFFFF00FFFF00FFFF00FFFF00FFFF00
FFFF000000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000FF0000FF00000000FF\
0000FF0000FFF0F000F0FF00EFF000EFF000DFEF00DFEF00D0E000CFE000B0CF00
AFBF00000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008F8F003F40903F4090404F9F404F9F\
E05000EFF000EFFF00EFF000F0FF00F0FF00F0FF00F0F000FFFF00FFFF00FFFF00
F00000FF0000FF0000FF0000F000000000FF0000FF0000FF0000F00000FF5050AFFFFF00F0F000\
FFFF00FFFF000000FFF00000FF0000FF0000FF0000F00000FF0000F000008F5070
F0F000F0F000F0EF00F0F0000000EF0000EF0000EF0000EF0000E00000EF0000E0504F9FEFE000\
EFEF00EFE000F0EF000000E00000E00000E00000EF0000E00000E00000E00000EF
EFE000EFE000EFE000F0E000EFE0000000EF0000E00000EF0000E00000EF0000EF0000EF0000EF\
F0EF00EFE000F0EF00EFE000F0EF000000EF0000F00000EF0000F00000F00000F0
0000F00000F00000F0F0F000F0F000F0F000F0F000FFFF00F0F000FF0000FF0000FF0000F00000\
FF0000FF0000FF00000000F00000FF0000FFFFFF00F0F000FFFF00FFFF00FFFF00
F0F000FFFF000000FF0000FF4F00A0FF0000FF0000FF0000F00000FF00005000AF0000FF0000F0\
0000FF0000F00000F00000EFE0F000DFEF00DFEF00CFE000C0D000AFC000A0B000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000AF8000BFCF00CFDF00\
D0E000E0F000E0FF00EFFF00EFFF00F0FF00F0FF00FFFF00F0FF005050AF0000FF
0000FFFF0000FF0000FF0000FF0000FF00000000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00\
FFFF00FFFF000000FF0000FF5000AFFF0000FF0000FF0000FF0000FF0000FFFF00
FFFF00FFF000FFFF00F0F0000000F00000F00000F00000EF0000F00000EF0000F04F4FA0F0EF00\
F0EF00F0EF00EFEF000000EF0000EF0000EF0000EF0000EF0000EF0000EF0000EF
F0EF00F0EF00F0EF00EFE000F0EF000000EF0000EF0000EF0000F00000EF0000F00000EF0000F0\
F0EF00F0EF00F0EF00F0F000F0EF000000F00000F00000F00000F00000F00000F0
0000FF0000F00000FF805070FFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000FF0000FF0000\
5000AF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00
FFFF00FFFF005050AFFF0000FF0000FF0000FF0000FF0000FF00000000FF0000FF0000F00000FF\
0000F00000F00000EF0000EFDF0000DF4F00CFE000BFD000AFC0009FB000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000A00000B0CF00\
C0DF00CFDF00DFEF00E0F000EFFF00EFF000F0FF000000FF0000FF0000F00000FF
0000FF0000FF4F00A0FF0000FF0000FF0000F00000FF00000000FFFFFF00F0F000FFFF00FFFF00\
FFFF000000F00000FF0000FF0000FF0000F00000FFFF0000FF0000F00000FFFF00
FFFF00FFFF00F0F000FF5000F000005000AF0000F00000F00000F00000F00000EFF0F000F0EF00\
F0F000F0EF00F0EF000000EF0000F00000EF0000EF0000EF0000EF0000EF0000EF
F0EF00F0EF00EFEF00F0EF00F0EF000000F00000EF0000F00000EF0000F00000EF0000F00000EF\
F0F000F0EF00F0F000F0EF00F0F0004F4FA00000F00000F00000FF0000F00000FF
F00000FF0000F00000FF0000FFFF00FFFF00F0F000FFFF00FFFF00FF5000F000000000FF0000FF\
0000FF0000F00000FF0000FF0000FF0000F00000FF5050AFFFFF00F0F000FFFF00
FFFF00FFFF00F0F000FF5000FF0000FF0000F000000000FF0000FF0000FF0000F00000F00000EF\
0000F00000EFDF0000D00000D000000000CF0000BF0000A0000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000009F0000\
B0C000C0D000CFDF00DFEF00E0F000EF0000EF00000000FF0000FF0000FF0000FF
0000FF0000FF0000FF5000AFFF0000FF0000FF0000FF0000FFFF00FFFF00FFFF00FFFF00FFFF00\
FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00
FFFF00FFFF00FFFF00FF0000FF0000FF0000FF0000FF0000FF00005000A00000FFF0F000FFF000\
F0F000FFF000F0F0000000F00000F00000F00000F00000F00000F00000F00000F0
F0F000F0EF00FFF000F0F000F0F0000000F00000F00000F00000F00000F00000F00000F00000F0\
504FA0FFF000F0F000FFFF00F0F000FFFF005000AFFF0000F00000FF0000FF0000
FF0000FF0000FF0000FF0000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00
FFFF00FFFF00FFFF00FFFF00FF00000000FF0000FF0000FF0000FF0000F00000F00000EF0000EF\
DF0000DF0000D000000000D00000BF0000B0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00009F0000D00000E00000E00000F0DF0000EF0000E000000000FF0000FF0000FF
0000F00000FF0000FF0000FF0000F0FF0000FF0000FFFF00F0F000FFFF00FFFF00FFFF00F0F000\
0000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FFFFFF00FFFF00
F0F000FFFF00FFFF00FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000\
FF0000F00000FFAF000000F00000F00000F00000F00000F00000F00000F00000F0
F0F000F0F000F0F000FFF000F0F0000000F00000F00000FF0000F00000F00000F00000FF0000F0\
FF0000F00000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000
F00000FF0000FF0000FF0000804F70FFFF00FFFF00FFFF00F0F000FFFF000000FF0000FF0000F0\
0000FF0000FF0000FF0000F00000FF0000FF0000FF4F00A0FF5000FFFF00FFFF00
F0F000FFFF00FFFF00FFFF00F0F0000000FF0000F00000F00000EF0000F00000EFDF0000D00000\
D000000000D00000C00000AF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000CF0000DF0000E00000F0DF0000EF0000EF00005000AF0000FF
0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FF0000\
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF5050AFFFFF00FFFF00
FFFF00FFFF005050AF0000FFFF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000\
FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000F00000
FFFF00FFFF00FFFF00FFF000FFFF00FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000\
FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000
FF00000000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF\
0000FF0000FF0000FF0000FF0000FFFF0000FF0000FF0000FF0000FF5000FFFF00
FFFF00FFFF00FFFF00FFFF00FFFF00F0F0005050AF0000EF0000F04F009FDF0000D00000CF0000\
0000C00000B0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000009F00000000DF0000DF0000EFDF0000E00000E00000EF0000
0000FF0000FF0000F00000FF0000FFFFFF00F0F000FFFF00FFFF00FFFF00F0F000FF0000FF0000\
FF00004F00A00000FF0000FF0000FF0000F00000FF0000FFFFFF00F0F000FFFF00
FFFF00FFFF000000F00000FF0000FF0000FF0000F00000FF5000AFFF0000F00000FF0000FF0000\
FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000
FFFF00FFFF00F0F000FFFF00FFFF00FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000\
FF0000F00000FF0000FF0000FF0000F00000FF00005000AF0000FF0000F00000FF
0000FF0000FF0000F00000FF0000FFFFFF00F0F000FFFF00FFFF00FFFF00F0F0000000FF0000FF\
0000FF0000F00000FFFF0000FF0000F00000FF0000FF0000FF0000F00000FFFF00
FFFF00FFFF00F0F000F0F000EFF000EFF000E0EF004F4FA0D00000D00000C000000000CF0000B0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000DF0000DF0000EF0000EFE00000E00000
EF00004F00AF0000FFEFFF00F0FF00F0FF00FFFF00FFFF00FFFF00FFFF00FF5000FF0000FF0000\
FF0000FF0000FF0000FF00000000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00
FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00\
FFAF00FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000
FFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000\
FF0000FFAF00FFFF00FFFF00FFFF00FFFF005050AF0000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF5050AFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FF\
5000AFFF0000FF0000FF0000FF0000FF0000FF00005000AF0000FF0000FF0000FF
FFFF00EFF000F0F000EFF000E0F000DFEF00DF0000D00000CF0000BFCF003F40800000A0000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000D00000E00000E00000F0
DF0000E00000E00000EFFF00EFF000F0FF00F0F000F0FF00F0F000FFFF000000FF0000FFF00000\
FF0000FF0000FF0000F00000FF0000FF0000FF0000F0F000FFFF00FFFF00FFFF00
F0F0000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FFFFFF00FFFF00F0F000\
FFFF00FFFF000000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF
F0F000FFFF00FFFF00FFFF00F0F0000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
0000FFFFFF00FFFF00F0F000FFFF00FFFF00FFFF000000F00000FF0000FF0000FF
0000F00000FF0000FF0000FF0000F00000FFFFFF00FFFF00F0F000FFFF00FF0000FF0000F00000\
FF0000FF0000FF0000F00000FF00000000FF0000FF0000F00000FF0000F00000F0
0000EFE0F000DFEF00DFEF00D00000D00000CF0000BFD000AFBF009FB000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000B0CF00BFD000
CFE000D09F00E00000E00000EF0000EFF000F0FF00EFF0000000FF0000FF0000FF0000FF0000FF\
0000FFFF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFAF00FFFF00
FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00\
FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FFFFAF00FF0000FF0000FF0000FF0000FF0000FF0000\
FF0000FF00005000AF0000FF0000FF0000FF0000F00000FF0000F00000F00000EF
0000F0DF0000DF0000D00000CF0000B0CF00AFBF00000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
B0C000C0D000CFDF00DF0000E00000E00000E000000000FF0000FF0000FF0000F00000FF0000FF\
0000FF0000F00000FF0000FFFF0000F00000FF0000FF0000FF0000F00000FF0000
FF00005000AF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FFF0F000FFFF00FFFF00\
FFFF00F0F0000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF
FFFF00FFFF00F0F000FFFF00FFFF000000FF0000F00000FF0000FF0000FF0000F00000FF0000FF\
0000FF4F4FA0FFFF00FFFF00FFFF00F0F000FFFF000000FF0000FF0000F00000FF
0000FF0000FF0000F0FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000\
0000FF0000F00000FF0000F00000FF0000F00000F00000EF0000F00000EFDF0000
D00000D000000000CF0000C00000AF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000AFBF00BFCF00C0D000404FA0DF0000E00000E00000EF00000000FF0000FF0000FF0000FF\
0000FF0000FF0000FF0000FF0000FFFFFF00FFAF00FF0000FF0000FF0000FF0000
FF0000FF0000FF0000FF0000FF0000FF00000000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00\
FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF
FFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF\
0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF000000FF0000FF0000FFFF0000
FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFF00FFFF00FFFF00FFFF00\
FFFF000000FF0000F00000F00000F00000F00000EFDF0000D00000D00000C00000
0000CF0000B0000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000CF0000D00000E00000EFDF0000DF0000E00000E000000000FF0000F0\
0000FF0000FF0000FF0000F0F0FF00F0F000FFFF00F0F000FFFF00FFAF00FF0000
F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FFFF00FFFF00FFFF00F0F000\
FFFF005050AF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF
F0F000FFFF00FFFF00FFFF00F0F0000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000FFFFFF00F0F000FFFF00FFFF00FFFF00F04F00FF0000FF0000FF0000
F00000FF0000FF0000FF0000F000005000AF0000FF0000FFF0F000FFFF00F0F000F0FF00EFF000\
EFF0004F4FA00000F00000EF0000EFD00000D00000CF00000000D00000BF0000AF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000CF0000D00000DF0000EFD00000E00000E00000E00000\
E000000000FF0000FFF0FF00EFF000F0FF00F0F000FFFF00FFFF00FFFF000000FF
0000FF0000FF5000AFFF0000FF0000FF0000FF0000FF0000FF0000FFFF00FFFF00FFFF00FFFF00\
FFFF00FF0000FF0000FF0000FF0000FF0000FF00005000AF0000FF0000FF0000FF
FFFF00FFFF00FFFF00FFFF00FFAF005000AF5000AFFF0000FF0000FF0000FF0000FF0000FF0000\
FF0000FF0000FFFF00FFFF00FFFF00FFFF00FFFF00FFFF00FF0000FF0000FF0000
FF0000FF00000000FF0000FF0000FF0000FF0000FF0000F0F0FF00EFF000EFF000E0F000EFF000\
E0EF00E0EF00DF4F00DF0000D00000CF00000000CF0000BF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000CF0000D00000DF0000EFD00000\
E00000DF0000E0F000E0F000EFF000E0F000EFF000EFF000F0FF000000F00000FF
0000FF0000FF0000F00000FF0000FF0000FF0000F0FF0000FFFF00FFFF00F0F000FFFF00FFFF00\
FFFF00F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000
FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF0000\
FF0000F00000FFFF00FFFF00FFFF00F0F000FFFF00FFFF000000FF0000F00000FF
0000FF0000FF0000F00000FF0000F00000F00000F00000F04F4FA0E0F000DFEF00DFEF00D0E000\
D0EF00CFE000CFDF003F408F0000C00000AF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000000CF0000DF\
C0D000D0E000D0EF00E0F000DFF000E0F000E0F000EFF0000000FF0000FF0000FF
0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FFFFFF00FFFF00FFFF00FFFF00FFFF00\
805070FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000
FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000FF0000\
FF00000000FFFFFF00FFFF00FFFF00FFFF00FFFF00FFFF005050AF0000FF0000FF
0000FF0000F00000FF0000F00000F00000EF0000F04F00A0DF0000D04F00D0EF00CFE000C0D000\
AFC000A0BF00000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000AFBF00BFCF00BFD000CFE000D0EF00DFF000DF0000E00000E000000000FF
0000F00000FF0000FF0000FF0000F00000FF0000FF5050AFF0F000F0FF00F0F000FFFF00F0F000\
5050AF0000FF0000FF0000F00000FF0000FF0000FF0000F00000FF0000FF5050AF
F0F000FFFF00FFFF00FFFF00F0F0000000FF0000FF0000FF0000F00000FF0000FF0000FF0000F0\
0000FF0000F05050AFF0F000F0FF00EFF000F0F000EFF000EFF0000000F00000F0
0000EF0000F00000EF0000F0D00000DF0000D00000D000000000D00000CF303F7F9FAF00000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000AFBF000000D00000DF0000EF4000A0DF0000DF0000
E00000DF0000E00000E000000000FF0000FF0000FFEFF000EFFF00EFF000F0FF00EFF000F0FF00\
0000FF0000FF0000F00000FF0000FF0000FF0000FF0000FF0000FF0000FFF0FF00
FFFF00F0FF00FFFF00F0F000FFFF000000FF0000FF0000F00000FF0000FF0000FF0000F00000FF\
0000F00000FF0000F0EFF000E0F000EFF000E0EF00E0F000DFEF004F00A0DF0000
DF0000D00000D00000D00000CF00000000CF0000C00000AF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000C00000C00000D0
0000DF0000E0CF0000D00000DF0000DF0000DF0000E00000DF0000E0F000E0EF00E0F0004F4FA0\
0000FF0000F00000FF0000F00000FF0000F00000F00000F00000FF0000F0EFF000
EFF000EFF000E0F000EFF000E0F0000000F00000F00000F00000F00000F00000F00000F00000F0\
0000F00000EF0000F0DF0000DF0000D00000D00000D00000D00000CF0000C00000
0000D00000CF0000BF0000B0000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000C00000C00000D0BFCF00CFDF00CF0000DF0000D00000DF0000DF0000DF0000\
DF0000DF0000DF0000DF0000DF0000E00000DF0000E000004F00A00000F0E0EF00
E0F000DFEF00E0F000DFEF00E0F000DF0000E00000DF0000DF0000DF0000DF0000DF0000DF0000\
D00000D00000D00000CF0000C0D000BFCF00AFC000AFBF009FAF00000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000A0B000AFBF00AFBF000000CF0000CF\
0000D00000D00000DF0000DF0000E040009FCF0000CF0000D00000D00000D0EF00
CFE000D0E000CFE000CFE000C0DF000000E00000DF0000DF0000D00000D00000CF0000CF0000C0\
0000C00000B0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000B0A0B000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[24]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7QP001aQ;64/HBf/00XT01hj9SW2>L8h0O0081`001PMj0006
SW2>64/H1H@02aQ;64/HBaQ;RHj900>>00B9SXV>0h@01T/HBaQ;678020L000H7N0009f^4RHhHBaQ;
64/HBaQ;64/HBaQ;68j9SXV>RHj9SQQ;64/HBaQ;6004Q003BaQ;06`020L000H7L0001T^4BhV>R@B>
00>WBd/03T/02fbWSZN>YhjWSZM/00m;00@HBaRR0X@00hV>001T00P700061fd0019;64/HBaR>RHj9
SXV>64/HBaP3B`0368B400B400fMQ9f4WHBMQ9f4WHB500a;00hHBaQ;6769SXV>RHB9SV8020L000H7
J@001AR464/H00A;02iaSXj>YhjWSYf4WHBMQ9f4WHBMQ9f4WHBMQ9f4WHBMQ9f4WHBMQ9f4WHBMQ9e/
1d/03HjWSZN>YhjWRIf4BaP0GP081`001PMW000>Q1Q;64/HBaQ;64^9SXT5SP>403JMQ9f4WHBMQ9f4
WHBMQ9f4WHBMQ9f4WHBMQ9f4WHBMQ9f4WHBMQ9dHBaRWSZN>YhV>RHj9BaP3Q00364/005T020L000H7
I00018THBaP7B`0YK9f4YhjWSZN>YhjMQ9f4WHBMQ9f4WFa]Bda;C8j/Yjb9YhBMQ9f4WH@03Id038BR
SZb>YhjWSZN>Y`I;0X@00aP0001F00P700061f<000Z>64/HBaQ;64/H18@00hV>SP05SP03YaQ;00i;
00/HYhjWSZN>YhjW6005B`0R69f4WHBMQ9f4WHBMQ:N>YhjWSZN>YaQ;64/HBaQ;66b4BeH020L000H7
H@000hV>R@05B`0jQ9f4WHBMQ9f4YhjWSZN>Yfa<Bda;C4]<Bda;C4]<Bda;LZN/YjbW[:N/Yg9;C4]<
Bda;C4]<Bda;C0>M00^/YjbW[:N/YjbWLP0:B`0369dH05@020L000H7G`001Xi`SW2>60^400<HSXh0
0Xh01JN>Yhib00M;00=<Bd/00d/04Da;C8jWSZN>YhjWSZM;C4]<00i;00j>YhjWSZN>YhjWQ9f4K0A;
00PHBaQ;64^4Be8020L000H7G@000hB>R@03SP1=K9f4WHBMQ9f4WFa<Bdb>YhjWSZb>[6a<Bda;C4]<
Bda;C4]<Bda;[:N/YjbW[:N/Yda;C4]<Bda;C4]<Bda;C4]<LJbW[:N/YjbW[:L01id00fa;B`05B`03
69dH050020L000H7F`001H@HSW2>00:928@7B`:>03JWSZN>Yhi<Bda;C4]<Bda;C4]<Bda;C:N/YjbW
[:N/Yg9;C4]<Bda;C4]<Bda;C4]<Bg:W[8h3Y`04SZNWY`BM00J4WHBMQ6d2B`0864/HBaQ;RHi>00P7
00061eP000E/64];SP07Q09;01=<Bda;C4]<Bda;C8j/SZb>[8j/00Y<0Vd5W@04[:N/Y`F/00=bC4`0
2D`05d]<Bda;C:N/YjbW[:N/Yja;C4]<Bda/00RM14/01ZN>YhjWRDX020L000H7EP001AQ;64/H00J4
0XT:B`0<C4]<BjN>[8j/SZb94Id01JN/YjbW00>/00>WXYd019d00fe;K@02C00FBda<C4]<K:bW[:N/
YjbW[:M<Bda;C0A;29d02QQ;SZN>YhVWRHi800P700061eD000@HBd];1X@05HVWSZM;C4]<Bda;C4]<
Bda;C8j/Y`0BW@0O0Id1X/>/`jc3[<>/`j81W@6M0Id1W@6M0Id1WFe<LP04[00AYjbW[:MbBda;C4]<
Bda;C4/029d02jb>YhjWSZN>YhA;04D020L000H7E00011Q;64/5Q0:918h4B`0<C4]<Bda;C4]]Q9f4
5Yd00j:/[007[1:M00>W[:`00Z`03ZN/Yj`HC4]<Bda;C4]<14/00aRMW@03W@0;Q9f>YhjWSZN>LAP0
A@081`001PMC000364];00F400lHSXjWSZN>Yd]<Bda;C4/00T`00ffMW@0AW@0o0Id1W@6M0Ie^R/>/
`jc3[<>/`id1W@6M0Id1W@6M0Id1W@6M0Id1W@6W`jc3[4a;C4]<Bda;C4]<Bda;C4]<00NM00VWSZN>
Yfa;64/0@`081`001PMA0004Q1Q;J`B400A]Bg:91Xh024a;C4]<69f40id00hBMW@0:W@=]1d`02@A<
14ab[<>/``05[09<0Vd2W@090Id1W@6M0Id100^M00f/C4a;C4]<Bda;C4]<00M;1Id02XBMQ:N>LAQ;
66a200P700061e0000>>L4/01H@2B`0<C1RWSZN>Yhj/Bff43Yd0?Kj/`jc3C0A<14`4C0A<14`4C0A<
1783[0>/0j`3[7=<14`4C0A<14`4C0A]0Id1W@6M0Id1W@6M0Id1WFd01T`02T]<Bda;C4]<Bd`2Y`RM
00=/Bd/00d/00aP0000n00P700061dl000B>L4][0h@5B`0<LXj>SZN>YhjRQ9f42Id00d^/[006[01<
LTa<C0A<14`4C0A<14`4C7:/`jc3[<>/`g84C0A<14`4C0A<14`4C0A<0Id1W@6M0Id1W@6M0Id1W@5]
C4]<Bda;C4]<Bda/[:N/Y`FM00V4WHA;64/HBaP0?`081`001PM?0003L76400:41T/02T`HYhj/SZb>
[8T9W@1;K@A<1:c3[<>/0j`3LPA<14`4C0A<14`4C0A<1783[0>/0j`3[7=<14`4C0A<14`4C0A<14`4
[0>/0id1W@6M0Id1W@6M0Id1W@5]00E<00Y;C4^/YjbW[:N/1id5B`0364/003`020L000H7C@0018A`
SV/2Q0I;00]<Bdb>YhjWSZb>[005W@03KDa<00=<00<4C7801J`0@/>/LT`4C0A<14`4C0A<14`4C7>/
0j`3[0>/0g84C0A<14`4C0A<14`4C0A<`jc3[<>/`jbnW@6M0Id1W@6M0Id1W@=<00a;C4]<YjbW[:N/
Yj`2Y`JM0T/011Q;66`l00P700061d`000A;L8j90X@4B`0@C4]<Bda;[8j/SZb>[:NMK@9<0184C0A<
14`4C0A<0j`3[0>/0j`?100:LP>/0j`3``?3L``401e<14`4``?30l<3``?30da^W@6M0Id1W@6M0Id1
K@03C00=L/>/`jO3YjbW[:M<B`05W@I;00<H0000>0081`001PM:000664]`SW2>1D/03da;C4^MQ9f4
[8j/YjbW[005C00J14`4C0A<14`4[<>/`j`3[0=<14`4C0A<14`6100gRjL3[0>/0j`3LPA<14`4C0A<
14`4C0A<14`3[0>/0j`3[0>/14`4C0A<14a]W@6M0Id1W@6M``04[006YjbW[:Mb14/01QRMWIf4W@9;
00@HBaR>>0081`001PM9000564^4SXT00Xh02Ta;C4]<Bda]WH@2W@03XZN/00F/00X4C0A<14`4C0A<
10@03G>/0j`3[0>/L`A^KFh050400ff7100=108300S30l<3``?3L`H401e<14`4C06M0Id1WKV/0jc3
[<>/`jc3Bda;C4]<B`06W@=;00=/Y`00=P081`001PM90007Bf^4L8i`SP03B`0<C4^4Q9f4WHBMQ:bW
1J`0>T`4C0A<14`4C0A<14`4LP>/0j`3[0>R0Id1W@6M0Id1W@6M0Id1W@6M0Id1W@6M0Id1W@6M0Id1
WFh4100PC0>/0j`3[0>/0g84C0A<14`4C0A<14a]W@6M0Jc3[<<6[004K4a;C0E;00U]WIf4WHA;6740
0Xh00hT0000c00P700061dP000A;Q8B418h024a;C4^MQ9f40id00d`4[006[00614`4C0A<1`@00fd1
``060`03^@4102D100>;0`<01`<;100AC0A<14`1XP>/0j`3[<>/`j`00d`024]<Bda;C4]<1Id01T]a
YhjWSS@020L000H7A`004T][Q6]aL8j9SQQ<69f4WHBMQ0A<1j`08fd4C0A<14`4C0BM0Id1W@>/0j`3
[0>R0Id1W@6M0Id1W@6M00<100fM0@41W@6M0Id1W@6M00/102Bi0j`3[0>/0j`3C0A<14`4C0A<14`4
C0A<14`3[0>/`jc3[<<2[008WDa;C4]<Bd`5B`07WHBMQ8Z>Y`02SP03R@00034020L000H7AP001T^4
Q8A;60B>00J/Q9f4WH@2W@0614`4C0A<0Z`01L>/0j`300D400@1W@6M1@480`D100>7KP@01P@00g<3
0`040`03`kh100l12@<00g84100;100BC0A<1<<3``?30l<3[0>R0Id10T`02D]<Bda;C4]<B`03W@08
XZbWYhjWSYda00P700061dH000=[Q6/00T/00hj9SP02SP06Q9f4WHBM1d`00`B/[005[00F19d1W@6M
0Id1W@6M0Id1[0>/0j`3[0l410<01;430`<81003KFh100D11P<00l<3``020@D4029<14`4C0A<14`4
LP>/0j`3[0>/`j`1W@6M0Fe<Bda;C4]<14/02YfWYjN>YhjWSXT`00P700061dD000=;Q8@00d/3SP0N
YhBMQ9f4KD`4C0A<14`4C0B/0j`3[06M0Id1W@6M1P400fh40`060`03L`@400`420<>1003QkT300L3
00>n0@400`400hL4100910D300K30l<3`kT40@0GW@5<C4]<Bda;C4]<Bj:W[:N/Yjb>Yhh0<0081`00
1PM4000368A[00A;00QbSXj4WHBMQ0Q<00@4C0A<0j`03JL1W@6M0Id1W@6M0Fd01@@02783/@>a0j`3
3P@80`l400=c0`<01P<00kT10@060@0SKP@4C0A<14ac[0>/0j`3[0>/0d`1W@6M0Id1WDa;C4]<Bd`0
0T/00aR/Y`03Y`05SZN>YaP0;`081`001PM400:414/024a;YhBMQ9f40T`020A<14`4C0A<0`@02G;3
W@6M0Id1W@030@03K@@400L41`<>10P33`@00g<30`060`03_P4100T11@@00g830`070`04LPA<10@1
00>M0Id00d`04D]<Bda;[:N/YjbW[:N/Yd/H02h020L000H7@`000hi[B`03B`08C1R4Q9f4[8h2[09<
01<4C0A<14`4W@6M0Id1W@6M0jac00d400=c0`<010<00g84100;10P33`@90`@400=]0@402@400g83
0`030`0P``?30g84C0A<14`4K@6M0Id1WFe;C4]<YjbW[:N/Yj`2Y`9;00<HB`00:`081`001PM2000>
SV];Bda;C8BMQ9f4WJL3[005LPA<14`00P@02P6M0Id1W@6M0Id40`h41P<00l=^0@0;0@03_P<300D3
00<1KP@02`@00i430`060`03L`@400H42P480`0914`4C0A<14`4008100JM0Id1W@42C00:L/>W[:N/
YjbW[0:M14/[00P700061d8000=[Q4/00T/024b4Q8BMQ9f>1J`05D`4C0A<0Id1W@6M0Id1[0>/0k43
LP0=1003LP4100/100VM0Id1W@6M0k401@<:0@05K@@4178020<;10X100jn0`<3`g=<14`4C0A<109<
00Y]0Id1W@6M0Feb0Z`01jN/YjbW[:L00id2B`0364/002T020L000H7@@000f^4R@02B`08C1RMQ9f4
KG86[00<Ld`4K@6M0Id1W@5]0P@00g<30`040`T400>70@405P470`03^@4100/100?30`<01@<00g<4
100;1003Q`4100P10P<61007C0A<14`4C0030@0?W@6M0Ig3[<>/`jN/Yja;00>M14/Y00P700061d00
00A`Q72>0d/2Q003WHA<009<00>>[:`01:`02@BMWId1W@6MKP0510040k43/@<310@04Fd1W@6M0Id1
W@6M0Ig3/@=b00`400X3/@>a0k43LPA]0`401Id1W@6M00H100?30`<00P<@10X100j514`4C0A<14`4
C0A<109<00NM0Id1W@6M00F/00BW[1Q;0Yd01XBMBd/HBbP020L000H7@0002XA[SXi<Bif4WH@2C003
14`400>/00T3[0>R0Id1W@4020@00g830`040`03N04100X100=]0`<01@<<1003LP<300@32P@00fh1
0@080@03``@400h40P<00l<10@080@H401Q<14`4C0A<16d1W@6M0Ifn[<>/`ja<Bd`4W@A;9`081`00
1PLo0006Q6^>L8i;18@6C0J/00D3W@6MKP0:10L300Ri0@6M0Id1W@H400L3/@>a0k5c00`400N/0k43
/@=b00d400>W0Id020400fh4100:10D32@4036d4C0A<14`4C0A<10=<00L1W@6M0Id100>/00=bBd`0
0T/2W@06Q9dHBaQ;9P081`001PLo0003Jhj900:>0X@04if4C4`4C0A<1:`3[0>/0j`3W@402`@80`03
_P5^00X410<05[43/@=]0Id1W@6M0Id1W@6M0Id1/@<<1005L`<30l<02P4:10L300?30@401P471008
C0A<14ac``<20@0:W@6M0Jc3Bda;C09;1Id3BbD020L000H7?`001HA`SW2>00>400=]C4`00d`00`A<
1003[0050j`3[0<030@00`430`060`d400V/0j`3[;6//Id03P08W@03KP@400<400Uc/@>a0k43/Fh0
20400fd4100410X300<40@401P4064`4C0A<14`4C0B/0j`3W@6M0Id1K4a;C0A;00LHWHBMQ4/H02D0
20L000H7?P000hA;SP03SP03[8BM00=<00l4C0A<16fi[0>/0k43/@<02P@30@03_P<300D320@01VjM
0Id1C1d000VM0Id1WFib0k4010<00g<410021003KP4100D10`@00i430`070`03L`@400P100=]10@0
1@@00g830`030`0?W@6M0Id1KDa;C4]<Bda/00BM00=;SP008`081`001PLn0003Jd]`00B>00>MC4`0
0d`010A<14`2W@080j`3[0>/0j`810030Id100<11`<02[4410BM0Id1W@4V0009WK6/0j`3/@=b00H4
00A^W@6M10400fh30`070`03L`@4008400=]0@401P405`A<14`4LP>/0j`3[0>W0Id1W@5/C4]<00E;
00JMQ9f4YhTS00P700061cd000A[Bd]<0hh04:b>C4`4C0A<19d1W@6M1;460`03N0@400L11P@30`05
Y`6M0j`0<`001:81W@491003KP4100T300>A10@01`@70@03K@@300X300l4W@6M0Ie<Bda;C4]<Bd`0
0id00jb>000Q00P700061cd000B464];18h01jb>C4`4C0@00id03@6M14`4[0>/0k43/@<00P4019d1
0@4910050Id1[4`0>@03W@03K@@400D400Eb0k43/@050`/400=^0@4010420`0:``>/0j`3[0>/LP9<
00>M0Id00Yd00da;C006B`05Q9f4Yhh08@081`001PLl0004Q4];B`:400>/SZ`00j`020A<19d1W@5]
1@@50`03``4100<12@@01Id1WFi<03h000<1W@@010@00g<30`070`030@@400/410400kT30`080`0C
Ld`4C0BM0Id1WDa;C4]<Bda;C003W@03Yhh001l020L000H7?00016];Bd/2Q003SZb>00>/00=?16d0
19d00`A<10041003/@>a00H11`@019d1W@42C4400id3100:0k43/@>a0k43W@<13@@20@L300K30j`3
C0@4C003W@6M00:M00=<Bd`01D/00if4W@02SP03Q00001d020L000H7>`001G1;Bd`H00:400=bSZ`0
0j`01k440Id1W@4020@00`?30@030@03``<300@400I^W@6M14a600040Id1/@L31`4;1003KP<300T3
0P@04d`4C0A<19d1W@6MC4]<Bda;LZL00id00hjW000N00P700061c/000B>64];0h@00db/SP04[0BM
00<110@01`@00`6M0@020@H300NaLid1C0B/04T00id01P>/0k43LP8400I^W@6M0Fd:10/300Ib14`4
C0@4C003W@6M00:M00e]Bda;C1R/Yif4WHi;01h020L000H7>`001W1;Bdb4W@=<1:`01@6M0Id100T4
10400l<30`080`03/@<004`000H1W@6W0g8510D12@@;0`D401=<14`4C6jM0Id1WFe;C4^/YjbW00:M
00>WBd/07@081`001PLk0004ST];J`:40d`3[0BM00<1LP@01`@0106M0Id2100;L`<3/@>a0k43[;40
D@02W@P41@4610/300ViK@A<14`4C0@00d`00ffM0@04W@0;BjbW[:N/Q9f4BaP07@081`001PLj0004
SW1;B`:400a]C0A<1:bnW@6M0JL20`H41@461005L`<30k40E0000`6M0@0710D100A^10@42`<01;h1
0Fd5100>C0A<14`4W@6M0Ifi[<<2[005YjbMWH@00T/L00P700061cX000A`ST]]0X@00d]<C002C003
19fM00>M00@3[0>/10@01@6M0Id100H400BM0Ie^0TaE0004C9d1K@H41P400`Ab0`090`03^@4100<1
00L4C0A<14`400A<00BM0IfR1:`02JN/Yjb4WHA;600K00P700061cX000>>L8h00h@2C00:14`4W@6M
0Id3/@<30`@50@H40`400`A<101H0003C06M00L41P490`03T@@100D11@@034`4C0A<19d1[0>/``B/
00BWWIfM0T/K00P700061cX000=`SX@00X@5C0>M00P1C7>/0j`3/@@300>i0@@01@@01P6M0D`4C5d0
00A<WIe^10@00g83W@040@D300=b10@00`@60@0>C0A<14`4C0A<LZ`3[<<4[008Yg9;BhBMQ4/J00P7
00061cX00Xh2Q0A<00H4C06M0Id31003/@<300L300=c10@00P@30@8400>/0000GP000`6M10021003
LP<3008310401<<30g<710H1011]10@4C0A<Ll<3[0>/0jc30Z`01T]<Bd^4W@9;6@081`001PLj0005
L8j4Q8h014`01PBMWId4C0<400=b0k402P<01`6M0J`3[4`0H@002ZbM0j`3/@>a0k450@X41P4036d4
C0A<LZ`3[0>/``B/00=;C4/00T/019f4BaPI00P700061cX000I;RHB9[8h2C00514`1W@401`@00gP3
0`080`8100Bn0jbMH`001[6/0k43/@D3104:10L10P@00g<30`020`0A``>/0jc3RTa;C4]<Bif4BaP0
60081`001PLj0003BhB400:>00B/STa<0Yd00`5]10061003W@6i00L300Nn0Id3LPA;06D000V/0j`3
[0>a0k401@4:10H10104Ll<3[0>/0j`3[<>RWD]<1D/00if4B`0H00P700061cT000M[BhB4SZbC00:/
00@4W@6M20@30@841@<20@0414`4WFL000F/0j`3/@020`03N0@100@12P@00fd10@030@P300Z/0j`1
W@5;C4]<0T/018BMQ74G00P700061cT000B4BhB40Xh01:bC[:`2W@03KT`400D400<1W@401`@01@6M
14`4009<?@004Hj9LAQ;64/HBaQ;64/HSXV>01X000E;/J`3LP0410D100=]10@01`@00fd10@030@83
00S30j`3[0>/0@BM00=;C4/00T/019f4WAPG00P700061cT00T/2Q003SZbC00:/00>a[04020@30@H4
0`421003C9d003P000I;RHj9SXT2SP04Yd];60Q;0hh00hU;600G0004C06M0@D400=^0@400`4:10D1
1P<04<<3C06M0Ie]Bda;C4^MQ9dG00P700061cT00T/2Q003C8j/00F/1`@00id10@0710<100@4C0BC
=0004D/HBaQ;64^9SXV>RHj9SXBM00j400HHBaQ;64/B0004C9fM0@H41@4910030`4100<10P<02:`3
[0=bC6d10id00da;C003B`03WHB>01H020L000H7>@02B`04Q6e<C0>/00D3[0>/0`0510<11P@30@05
14`4WJ`0<0000aQ;B`04B`:400^MQ::>YhjWSZN>Y`0;B`0<YhjWRIf4WHBM64/H4@0014`4W@4610D1
1P@019430`<40@03``<3008300ab14`4W@6M0Ie<Bd`2B`03Q9f>01H020L000H7>@0014/HQ4/4C00:
XZba[0>/0k43/@@11@@20@06W@B/0ie<:`001Hj9SXV>00F400=;64/01T/028jWSZN>YhjW2d/03jN>
YhjW64/HBaQ;68B9L@0@0003C05]00H410421003L`<300@300>n0@400`401:`4C0@5C0BM00=]Bd/0
0T/00hBWSP0E00P700061cT000=;Q8@00T`020A<19d1XP>a10<40@040g>AL`L300>a0D`03@000aQ;
000H0003SXV>00:>00F4WHBMQ00=B`0>YjbW[:N/Yja;C4]<Bd`4B`06YjbW[8jW24/2SP03600000d0
00>M0Id01P@40@T31P44100AC0A<14`1W@6MC4]<Bjb>[8h050081`001PLi0004SV^4B`A<0id01W83
[0>a0`@11`<018i`SW0BQ007SXV>RHj9SP02R@0564/HBaP02@0011Q;68h5Q0>>3D/03ZN>YhjWSZNR
KAQ]66dH14/01W:>YhjWSPE;00PHBaR>RHj4B`d000@4WIfM1P@00fh30`0:0`H100A<14`414`00ffM
W@02W@05C4];SZL00XhD00P700061cT000=`Q8@00T`01PA<19d1W@8400Ac0`<30`400l<3600>Q09/
2T/06JN>YhjWSZN>YhBMQ9f4WHBMQ9f4WHA;64/00h@05Yf4YhjWSZN>C4]<Bda;C4]<Bda;C6`BW@06
[:N/YjbW24/01JN>YaR400d000AcW@6M10@00i430`0<0`D101F710A<14`4C0BM0Id1BdbW[8j/SV`0
4`081`001PLj0003L8A;00=<00A;WIfM10@01AQ;JhA[00B402f>RHj9SXV>RD/HBaQ;64/HBaQ;64/H
BaQ;RJN9YhV>RHj9LAQ;64/HBaQ;66`01h@02Xj9SXjWSZN>YaP6B`03KHBM01JM00ZW[:NWSZN4WHBM
0T/011Q;SZL2SP0364/000`00Z`2W@040j`3/@`300=c10@01@401@A<14`400A<0id024`H[8j/SZLH
4`081`001PLj0006SXA<C0A<0Yd00`69Q006Q0A;00Z>YhjWSZN>YhjW44/02fbWSZN>YhjWSZM/00m;
00J4WHBMQ9d2Q006ST^W[4]<7Id01j:W[:N/Yj8019d02XBMQ6f>[8ia64/=0006/J`1W@>a30<01I44
10B700D10`@014`4C0@2C003W@6M00>/00B>[8j/4`081`001PLj0003L8i;009<01X4L8i[Q1Q;64/H
BaQ;64/HSXV>RHj9SXV>611;00YbSZN>YhjWSZN>1d/031Q;64/HBaQ;64^>Y`:>00>9Q8@01H@00aQ;
W@0;W@0>XZbW[:N/Bda;C4]<Bd`3B`06[:N/YjbW0id038BMQ9f4WHBWST/HB`d000J/WIfM0j`:0`03
L`@400<41@401Yd4C0A<10=<0id2[006SZb>Yhi;4P081`001PLj00>>00B9SXV>3T/02XjWSZN>YhjW
SW84B`0AK6e/KHBMQ9f4WHBMQ9f4WHT01JLAB`0EYhjWSZN>YhjWBif4WHBMQ4^M0Ie]00A<019b`jc3
[<>/C4]<Bda;C4]<Bd`6[003Bda;00FM00F4WHBMK003B``000F/0id1W@090`030@@400H41@441008
C0A<14`1WJL2[005Yjb>WH@04P081`001PLi000>Q6^>L8j9SQQ;64/HBaP8B`0^SXVWSZN>YhjWQ9f4
WHBMQ9f4WHBMQ9f4WHBMQ9f4WHBMQ9f4WHBMQ9f4WHA]60M;00n>YhjWSZN>YhjW64/HBaP01H@01T]<
Bda;C0J/015;C4]<Bda;C4]<BjbW[:N/Y`07B`0;Q9f4WHBM64/HBh@02`001db/W@6M0k400`<01KT1
0@5]00P410401Fi<14`400A<0id02:bW[8j/Q9dH4@081`001PLf000:BhB4RHj9SXV>60U;01F4WHBM
Q::>YhjWSZN>YhBMQ9f4WH@09Id00d^/Y`08Y`U;00B4WHB>0d`01l>/`jc3[7802T`00g;3[003[00?
Bda;C4]<Bda;YhBMQ9f4009;00>>R@002P001D`4W@6W009c0`@40@T41045100>C0B/0id1YjbW[76M
Q4/A00P700061c<000j4Jh@HBaR>RHj9SXU;609;00=/Q8@02H@05:N>YhjWSZN>XXBMQ9f4WAQ]66dH
14/01da;YhjWRJL059d00hBMY`05Y`04SZN>Y`9;00lHBaQ;64/HBaQ/Q8j9SXT01:`00l>/C004C006
Bda;C4]b1J`01G9;C4]<00A;00J>[8jMQ9d2Q00368j900/000A<19d11@@00fh10@020@T4104020A<
14ab[0>/0id00jN/SP02B`03WH@0010020L000H7;`001F];64/H00=;00<HSXh018@0D9f4WHBMQ9f4
WHA]Bda;C4^WSZN>[8j/SZa;C4]<Bda;C4]<Bda;C4]<Bg:W[:N/YjbW[:N/Bda;C4]<Bda;C4]<Bda;
KIf/YjbW[:N/YjbW1Id00fa;B`06B`0VYhjWSZN>Yd/3W@6M0Id1W@6M0Id1WKj/`g9<Bda;C4]<BjbW
[:L2B`07WHBMQ:N>B`0:0004C0BM0@H4104910@100<40`<00P<01:`3W@42W@03[4a;009;00>MQ000
3`081`001PL]000:SQQ;64/HBaQ;60^400<HBd/024/0@G:>YhjWSZN>Yhi<Bda;C4]<Bda;C4]<Bda;
C1R/YjbW[:N/Yj`HC4]<Bda;C4]<Bda;C4]<BdbW[:N/YjbW[:N/00VM1D/01aRWSZN>Yhh00X@03fbM
0Id1W@6M0Id1W@6M0@03W@07C4]<Bda;C002[003Yja;00=;00F4WHj>600;0004C0BM0@H4104910<1
00Ja0j`3[0<2[0>M00=<Bd/00T/2Q10020L000H7:`000f^>B`05B`0368B400:405JMQ9f4Yd]<Bda;
C4]<Bda;C4]<Bda/[8j/SZbW[:N/Bda;C4]<Bda;C4]<Bda;C4]bYjbW[:N/YjbWLT]<Bda;C4]<Bda;
C4]<Bda;[:N/YjbW[:N/Y`^M00m/Bd]bYjb>YhjWST]/WH@00`4049d1W@6M0Id1W@6M0Id1W@42C003
Bjb/00:/00=<Bd/00T/01HB/SZLH00/000A<0Ie^1P@30@03KP@400H400<10`<01@<01:`3[043W@03
C4];009;00>400003P081`001PLY0007Q72>64/HB`09Q0B>24/0BDa;C4]<Bda;LXjWSZN>YhjWSTa;
C4]<Bda;C4]<Bda;C4]<K:bW[:N/YjbW[1Q<Bda;C4]<Bda;C4]<Bda;C:N/YjbW[:N/Yj`014/00fbM
W@08W@0;YhjWSZN>YaQ;64/00X@03QP1W@6M0Id1W@6M0Id11Yd4[0I;0hh00d/0000:0003C06M00L4
0`4810D300N/0j`3[0>/00>M1D/00h@0000>00P700061bL000A[SXV>0T/00aR4Q004Q00SWHBWSZN>
YhjWBda;C4]<Bda;C4]<Bda;C6f/YjbW[:N/Yj`00T`02d]<Bda;C4]<Bda;00A<011bYjbW[:N/YjbW
LT]<Bda;1D`06T]<C4a;C4_3Yl>W`jO3Yl>WC4]<Bda;C4]<2Id2Y`03[:Ma00E;01FMQ8A<1783[0>W
0Id1W@6M0Id1WL<00j`00da;C003B`05[8jWQ4/02`0210030Id400D4104510X300Ja0jabW@42W@05
Bda;Bh@040081`001PLU0006Q72>L8hH1h@01AQ;Bg:900F>00<HBd/014/044a;C4]<Bda;C4^WSZN>
Yhh2Y`0>[1Q<Bda;C4]<Bda;C4/5C00/KJbW[:N/YjbW[1Q<Bda;C4]<Bda;C4]<Bda;C:N/YjbW[:N/
Yja;C4]<Bd`5B`03KIfM00NM0ZL3B`0>64/HBaR4C7:/0j`3[0@2C004W@6M0@:M1:`00d]<B`03B`03
SZN>00:42`001D`4W@6M00D410400fd4LP090`05/@>/0j`00d`3W@03C4];009;40081`001PLT0006
BhV>RHj91H@3B`0MC4]<SZN>YhjWSW9;C4]<Bda;C4]<Bda;C4]<SZ`01JLDW@J/00K3Yl>/Rfd?C00J
Bl>W`jc3Yl>W`jM<Bda;C4]<Bda;C4]<Bd`9W@03KD];00A;00<HWH@01@<03D`4C0A<14a]XP>/`j`0
0Yd02Da;C4^/SZ`HW@0;00:/00@4W@5]1@@40@l300EcC0A<0@03W@04C4];Ba0020L000H78`001T/H
SW2>R@>42D/7SP0E64a;C4]<Bda;C4]<Bd`HWHBMQ9f401NM0Z`02:N/[:bW[:N/3Id00fe<C002C00M
YjbW[:N/YjbW[4]<Bda;C4]<Bda;C4]<Bg:W[:8029d3B`0464/HY`:400J/0j`4C0@5C0030jb/00:/
19d2B`05SZb>BaP03002[008/J:iLW=bL`<40@03``<300X300>/14`014`3W@A;40081`001PLQ0005
64/HSXT00hh00hA;B`04B`0NC4]<Bdb>YhjWSZN>C4]<Bda;C4^MQ9f4WHBMQ9f46Yd4[00R`jc3[<>/
^Id1W@6M0Id1W@6M0Id1W@6M_Zc3[<>/`jc3[0=<019;C4]<Bda;C4]<Bdaa`jN/Yj`8W@E;01>WSYd3
Ld`4C0A<14`4[0>/0j8100BM00A;[8j/0T/00aP000090007[0>/0k43/@040`@12`<01[T4C0A<109<
00BM0IfM0d/00id0000=00P700061b0000PHBaQ/L8j9SPe;1Xh05:LHKHBMQ9f4WHBMQ9f4WHBMQ9f4
6Id:[1RM01fW[:b/Bda;C4]<Bda;C4]<Bda;LZN/YjbW[:N/K006W@0EBaQ;SZN>Yh@4C0A<14`4LP>/
0ja<00JM00J>[8i;68@<00>/00@3[0>/0`<50@P300Ni0@6M14`400A<0id3B`03Q00000d020L000H7
7@000aQ;6003Q0F>01a<Bda;C4]<Bda;C4]bSZb>[8j/SYf4WHBMQ9f42id00jN/[005[00NC0A<14`4
C0A<14`4C0A<14c3[<>/`jc3[<>/14a^0Vd07Yd1W@6M0Id1W@6M0Id1W@6M0Id1W@6M0Id1W@6M0@I<
015;C7;3[<>/`jc3Yda;C4]<K005W@07Yjb>Yhi;Q0031008C0>a0jacC0@2C003KIfM00:M00F/Bd];
Q00<0005C0>/0k401@<50@<310@50@0714`4C0A<0@02W@04Bda;W@l020L000H77000371;JhA[Q4^>
L8j9SPU;00A<68B418h04:N>YhBMQ9f4WHBMQ9f4WD/4C00:LZN/YjbW[:N/Y`i<00<4C0@02Z`>C003
K@6M00:M00<1WId01id00`6M0@06W@06Bda;C4]b1:`01:N/Yj`7B`BM00nWSZN>YaQ;Q6i<0j`3[0<0
1d`3W@05Q4];Bh@0300024]<1:`3[0>a0`<40@03LP@400@400=^0@400P4014`4C0@3C0:M00A]Bd^4
3`081`001PLK0003L4^400:40T/00g:>SP03SP0IC4]<Bda;C8BMQ9f4XXj/SZb>[8jRQ9f4W@02K@Y<
00=b[:`01j`0ET`4C0A<14`4C0A<14`4C0A<0j`3[0>/0j`3[0A<14`4C0A<14`4C0A<14`4LP>W0Id1
W@6M0Id1W@6M0Id1W@6M0Id1KDab`jc3[<>/`ja<Bda;C4]<0T/3W@05[:N/Yj`00T/00h@10@030`09
LPA<14`4C7:/00BM00=<Bhh03@031008W@44TG>ALhL40@L41@4014`4C0@2C0>M00A<SYf43P081`00
1PLJ0004L4][Q0A;00HHSXV>RG83B`03KHB400N418h01:N>[1P@C003LZb/00N/24`02@A<14`4C0A<
100:[00VC0A<14`4C0A<14`4C0A<14ab[<>/`jc3[;jM0Id1W@6M0Id1W@45W@03XZb/00B/00JW[4]<
Bd`7B`03WJNW00:W00b>BaQ;606M0JL4C0@4C0>/0id018A;SX@<0006Q0A<0Id11@@40@P40`4019d4
C0@3C0>M00@H[8j43P081`001PLI0004L8B4K0E;1Hh05QQ<Q9f4WHBMQ9f4WHBMSZb>[8j/SW8;C006
14`4C0Ab1:`04<>/`jc3C0A<14`4C0A<14`3100=C0@40j`3[0>/0j`3[007100oC0A<14`4C0A<1:`3
[0>/0j`3[0>/14a^K@6M0Id1W@6M0Id1X/>/`jc3[<>/C4]<Bda;C4]<Bda;[:N/YjbW00=;00<H0@40
0`402D`4C0B/0j`3LP03W@03Bjb400`00T`210030Id400@400=^0@400P4710@10P@02d`4C0BM0J:/
SZb400h020L000H760000aR>J`07B`05RHj9SXT03H@02AR>SZN>[8j/B`0@C003LZb/00N/03L4C0A<
14`4C0A<14`4C0A<1:`3[0>/0j`3[0=<14`4C0A<14`4C0A<14`4C0>/0j`3[0>/0jab00M<00=]W@40
1Id9[006Bda;C4]<1T/01:N/Yj`2Y`0>Q4/HBaP1W@6M0D`4C;43[004C4^MQ0:>00=/00002`02C003
19d100H410471007KP410@A<1004C006WJb>[8j43P081`001PLH0003SV];00M;0hh3Q009WHBMQ9f4
WHA]00A<01V>[8j/SZb>[4`4C0A<14`4C0A<14`4C0Ab00:/00P3[0>/0j`3C10400X3[0>/0j`3[0>/
4@@40`0d``?30l<3LPA<14`4C0A<14a]W@6M0J`3[0>/`jc3XP6MC4]<Bda;C4]<BjbW[:N/Yif4K@=;
00=a0@400P401D`3[0>/00=<00E]WHj/600=0006W@@419e^1P@40@L40`421006C0A<17830Z`00i>/
Q00>00P700061aL000>4JhT024/2R@Z400E;C4]<B`02C008Yhj/SZb>[8h@C0031:b/00N/03d4C0A<
14`4C0A<14`4C0A<1:`3[0>/0j`3[0=<14`4C0A<14`4C0A<14`4C0>/0j`3[0>/0jabC0A<14`400M<
00=]WL<01j`00j:MW@02W@0464a;C0A;01JW[:N/Yjb4WHBM64/HSYd1W@>/0jab14`019f/ST/=0006
[4`4C06M1P@40@L40`400`A<1002C007[;6/[9>/SP02Q0d020L000H75P0018A[SXT6B`03C8B400>4
00JMQ9f4KD/9C00jLZb>[8j/SZa<14`4C0A<14`4C0A<14`4C0>/0j`3[0>/0h]^KFjM0Id1W@6M0Id1
W@6M0Id1W@6W``<34@@:0`8402=<14`4C0A<14`4C0A<1:`3[0>/0j`3W@6M0Id1W@5;C4]<B`05[00;
Yd]/WHBMQ4^>10400`<01;43C0@3C004KHj/B`d000J>14`4KP461004KP410@L40`4026d4C7>/0j`3
1:`00d^4000<00P700061aD000E;Jhi`SP05B`V400LHC4]<Bda;00I<00UbSZb>[8j/SW800d`02PA<
14`4C0A<14`2K@03WJ:/00>/00Fa[;6/^@03W@0m0Id1W@6M0Id1W@6M0Id1W@6M0Id1W@6M0Id1W@6M
0Fe^K@A<14`4C0>/0j`3[0>/0j`4C0A<14`4C0A<1004C00<LP>/`jc3[<>/0Id11Yd02T]<Bd^W[:N/
Yj`3B`0;WHBMQ4^>1:`3[0<01T`014^/64/>0006T`A<19d11P@30@P40`401W83[0>//@B/00<HBh@0
3@081`001PLE0005Jhi`SXT00T/00db4Q005Q004WHB/SPe<00=b[:`01J`02784C0A<14a^0id06P6M
0Id1WKV/0j`3[0>/0j`1W@6M0Id1W@6M0`401Id1W@6M01P100=^0`<020<2100SC0A<14`4C0A<14`4
C7>/0j`3[0>/0d`1W@6M0Id1W@6MKD/01Z`00da;B`02B`:M00>4[8h00`<019d1C0@3C004[9=]B`h0
0Z`210811P@30@P40`430`04[0>/0`>/0T/00h@0000;00P700061a<000R>68i`SW1;J`N400>9SXh0
0hh00d]<C00;C004BjbCYa>M02`1W@>/0j`3[0>/0id1W@6M0Id1W@6M0Id1W@6M0Id1W@6M0Id1W@6M
0Id1W@X10283[0>a0j`3[0>R0D`4C0A<14`4C0A<14`4[0>/0j`3[0=b1D`00ffMW@08W@03[:N/00Q;
00R4WAP4C06M0@9<00I?[:bMQ4/?00:/00BaW@5]1P@30@<400ec0`<3`id1W@>/0jba00>/00BMBd^4
3@081`001PLC009;00F9SXV>B`06Q003C4]b00>>00>/SW801d`01`A<14`4C6h02id05P6M0Id1W@6M
0Id1W@6/0j`3/@>a0k4@10T300>;KVd03`490`L100=]10@00`@00d`4LP050`08[0?314`4C0@4C008
W@6M0Id1W@42W@04[4a;C0I;00n4BaP4W@6M0G83[;64WD/03`001XV/[06M0@<300EhLg>i_P090`<1
0P<01K43[0>/00:M00=<Bh@03@081`001PLB000964/HSW2>L8A[00>400E;C4]<B`05SP03Bda<00M<
00=;WH@04Yd0106MKVd410090j`3[0>/0j`3010400Za0`<3/@>a0k5c30@00fd10@020@L300>a0`40
2@4049d4C0A<0j`3[0>/0j`3C0@8C003K@6M00NM2D/02JN4BaP4W@6M0`02[004C9f4SPl000Q<[:bM
0Id3/@l30P4029d3[0>//G9<0Yd2B`03Q00000/020L000H74@001G1;Bd^900>>0h@03Da;C4]<Bdb>
[8j/SZ`01D`00`B4W@0=W@050Id1WFh030@01K43/@>a00@300=c10@03@@00g<30`060`03L`@400h4
00=c0`<01P<00kT10@0;0@03K@@300L300Yb14`4C0A<14`414`01Yd1W@6M0@:M00A;C4]<14/02ZNM
Q4/40Id1W@<2C003Q9f>00l000M;1:`3W@6n01030`401;h3/G<3C0:M00=<Bh@03@081`001PLA000B
BaQ;Bhi`SW29Jda;C4]<Bda;18h00i>>[002C005Q9f4WH@02id01:N/[7<?100:0j`3[0>/0k43/@l4
00=b0`<01P<00g<4100>10T33@404Yd1W@6M`j`3[0>/14`4C0A<10M<00?3[:L01id6B`09[8jMQ4^/
0Id100=<00>MQ0003`0024b41:`1W@6a3`<01ZL1W@5b10A<0Yd2B`03Q00000/020L000H740001Hi;
Bd^900F>019<Bda;C4]<Bdb>[8j/SZb>WH@<W@080J`3[0>/0j`?1003L`<300L33`@00g<30`060`03
LP@400h42@<4109^40405[h310A<14`4C0A<14`4C7:/0j`3Y`45W@03KD]<009;00ZW[6bMQ0>M0Id4
0T`00hBM000?0005Bie<19d00P4A0`<100H4C0A<14`2W@03C4^400d020L000H740007G1;BhA[SW2>
L8i;C4]<Bda;C4]<SY>>[8j/Q9f400RM00a^C0Ab/Jba[;6//G8?1009/@>a0k43/@>a00l400=b0`<0
1`<@10T32@@80@0BW@6M0Id1W@5<14`4C0A<14`40T`01j`3[0>/`j`01id3B`0:[8i;69f419d1K@9<
00>W60003`0024b414a^W@5b3P<00g<40@020@03C0A<00=<0Yd2B`h020L000H73`00171;Bd/2Q0F>
00I<Bda;C4/4C007[8j/TjbC[006W@03K@@400D400R/0j`3[0>a0`l42@<?1003L`<300H300=c10@0
3@@00g<30`060`03T@@400/43P404fh414`4C0A<14`3[0>/0j`3C0401Yd016a<Yj`3B`05WD`1WFh0
0T`00i<0000?0003Bie<00840`421009LgQcLkVn^I5c00D40`401Vi<14`4C0:M0T/>00P700061`l0
012>64][Q6^>L8ieST]<Bda;14`02Ff>[8j/TjbCXP02W@0DK@A<14`4C0A<14ba[;6/0j`3[0<>100A
L/>W0Id1W@6M0Id1W@6M0Id020480`04^@5^K@`400=c0`<01P<?10832`404Id1W@A<14`4C7:/0j`3
[0>/009<29d00jb>B`02B`0864^M0Fe<[8@A0008C8@4C0BM0Id610<11`@30@03K@A<00=<0Yd2B`h0
20L000H73`000g1;J`03Q0F>00A<Bda;0d`01HBMQ::C00J/3P@02@>a0k43/@>aL`071004KYd1WA/1
20<=0@04Q`@4L`P33P@00i430`040``100a]14`4[0>/0j`3[784C0JM0Z`5B`030Id100:/00=;0000
3`000hjMC00210<11P@30@03Q`@400@40`401Vi<14`4C0:M0T/>00P700061`l000F>JhA[K@02B`08
L8i`SQQ<Bd`3Q008WHBMQ::C[9<4[006Ld`4C0A<20@080>/0j`3[0>/^Id1W@6M0Id1W@6M0Id1Q@>a
0k43/@=b3P@010>a0k440`05L`@4KFh05@400fh4100:10T30`@20@0@W@6M0Id1W@6/0j`3[0>/0`I<
00A;C9fM0Z`00hi;B`03B`05S[6/[4/04P0028j414`4W@6M1P@40@H400D1W@5]1004C0:M00=;SP00
3@081`001PL?0007JhA[Q4]<B`04SP08TdbMQ9f4WH@2W@0;KTb/[;6//J`3[7<02`@02P6M0k43/@>a
0k460@03KP@400H420<>1003L`<300D300=c10@02`@@0@03KP@400H420<710H100>i0`<01@<01T`4
C0A<10E<00C3[:bM14/2[003/Da<018000>>WD`00P@30@H4104610@100E<14`4C002W@03[8h000d0
20L000H73`000hA[Q004B`03L8j>00:>18@029f4WH@4C0A<1Z`00k6/1006100?K@6M0Id1XP>/0k43
/@=b00l400Ra0k43/@>aL`d400Uc/@>a0k43/G<03P@00k430`020`h100E^10@4L`070`03L`@400L4
00e^0@6M0k43[0>/0g8400Q<00M;[:b/Yif400=;00F>WDa<B`0B0008SXB/C0BM0Id610@11P@01@6M
0Id400A<0Yd2SPh020L000H73`005f^4Bda;C4]<SY>>WHBMQ9f4WD`4C0A<008400Na[0>/0j`300<4
019^W@6M0Id1W@6M0Id3/@>a0k420`h400=h0`<01@<00kh10@060@05W@6M0Id010430`h400=c0`<0
1@<00g<4100210X100>n0`<01@<00i44100;10030@<300D300T1K@A<14`4C0@00T`01:`3[:82W@07
Q4];[8i<B`0C0005BjN/0`@00`400fh4100310@11P@40@05C0A<14`00Yd00jb>000=00P700061`l0
00>4Jd/00d/014a;LXh4Q00CWHBMC0A<14`4C0A</Jba[;6//@03W@080Id1W@6M0Id41008Lk43/@>a
0k49100>W@6M0Id1W@>/0j`3[0<3[0040j`3[0H400M^0@6/0jac008400Q]0Id1W@6MKP<400P3/@>a
0k43LPP400=^0@4010480`h400=b0`<00`<02;41W@6M14`41D`3[0:M00J4WHA;SW42Ba<000U<Q9f/
0fd1WFh01@@40@H400D1W@6M1003C003[9fM00:>3P081`001PL?000CJhi;C4]<Bda;WHBMQ9f414`4
C008100>0j`3[06M0Id1W@6M0Fh91003/@>a00@30`4019d1W@441004BhB4W@=<00H4C0BR0j`60`04
T@@410@100>710@01@@30@8300F/10@3/@060`d400=^0@<01`<20@d41`<50@0;14`4C0A<0j`3[6d0
19d016b/LD`D0005C9fM0j`00P<00kh1100410@11P@40@05C7>/0j`00Yd00j`0000=00P700061`l0
00>>L4/00d/00da;Q005Q00<[4`4C0A<14`4C0A<0P@00jNMW@03W@030Id100h400Va0id1W@6W0j`0
1P000hi;Q002Q00364a<009<00h4KIfR0j`3[0>/0k4310D11`@30@8400nM[:M<Bdb4WH@3K@6M0Id0
2`@80`03^@41008100=^10@01`@70`8100bM0Id1WFe<14ab[;42[009BffMWHBMST^>01@000Y<Q9f/
0j`3[0>/10@40@H400L1W@6R0jba00:/0Yd00hh0000=00P700061`l000M`SQQ<Bda;00:400BMQ9fC
0j`00d`41007100<W@6M0Id1W@6a0k5c2`@0106M0Id210X000J4RD^4WH@2C00814`4C06M0J880`D1
1`@30@8400BM`jc30T`02T]<BdbW[8jWQ9d210<11@@80`P11`@70`8400=^0@400`4039d1C0B/0j`3
[4a;K@:M00>4[4/050001XA<WD`3[0P300>A0@400P451003N0<3008300Fa0j`3[002W@03[00000d0
20L000H73P001if>L8i;C4/01H@03Hj/TjbC[4l4C0A<14/01Yd02@6M0j`3/@>a0`0510050Id1W@40
0`@=0003L8i[00:41T`2W@080Ie^LP>/0j`20`D11`@30@8400A<`jb/0T`034]<Bda;YhjWQ9f4W@>4
00d4W@6M0k43/@>a0k5c00@4204310L31@@03@6M0Id1W@6M0j`3[7800T`01T]]Q9f4Ba@000];C8@4
LP>/0j`3[0040`@11`<01[43[0>//@:/0Yd00d/0000=00P700061`h000j4L8j>LT]]Q9f4WFf/T`B/
00>a[0@00P@029d1W@6M0Id10P@01K43/@>a00@300AhW@6M0`@A00>>00>4KD`00T`01`A<0Id1W@40
0`@017<30`<50@030i4400@40`4310050J`3[0@01T`04jN/Yda]WHBMQ9f4WHBM/@>M0Kh00P<910L1
00?30`<010<810@100JM0JL3[0<4C005Bif4WD/05002C004WD`410T3104:0`07/@>/0jbMQ00@00P7
00061`h000I[Bg2>L4`5Q09<00>/Tj`01J`00`BMW@04W@L400/3/@>a0k43/@>/0`0C0006Bg2>RHi;
14`01FfMWId100H400<3``400`440`<40`421007C06/0j`1W@05C0066:bWLT]<0d/00if4W@07Q084
00<110@02@@80@03``<300X400RM0Id1W@6W0`A<00I;C1RMQ4/D0005BdbM14`00P@010>/0k440`@1
1`<01[43[0>//@:/00>4W@003`081`001PL?009;0hh04i>4WD`4C0A<14ba[;6/0jbiW@402@@01FjM
0Ie^008460000d^>SP02SP09[4`4C0BM0Ie^00L410400l<30`0:0`06W@@30l>i0P401Yd1W@6M``:/
14`04D]<SZb>C4]<Bda/YhjWQ9f4008400<1W@@010@50`03``4100<12`@02P?30@6M0Ie^C0@2C005
Bjb4WH@04`001HAbQ9e<008400A^0@410P@30`04``410@X300>aLd`00T`00ie;000?00P700061`l0
0T/4SP03RD]<00=<00d4C0A</Jba[;6//Jac00L400<1W@400`@00h@HSP0G0003Bg2>00>>00N/C0BM
0Id100P410400g830`090`0@W@A<0j`3C0BM0Id1W@6M[0A<00Q;C4^/SZa;C0=;0hh4B`0;RHhH0j`3
/@>a0k401@<60@P40P<02;73W@6M0Ie]0d`00jbC[002Q18000=[SXh00Yd01`A<19d1W@4010@00g81
0@020@L300Ba^Id414`00hA;000?00P700061`l00d/2SP0:Thi<C0A<14`4C0:M00Zi[0>/0k43/@>a
10@00`6MKP021006WDb4WHi;5P001D^9SZN>00>/00AbWId120@016d10@431003L`<300H300>a0@@0
10<41009KP41W@6M0Id400=<011bYjbWC4]<Bdb>Yhi<Bda;0Xh01Ta;BaP3/@H310@60@D41@<01l<1
0Id1W@@00T`2[003Bif4018000J4SZbRKT`210<100=^10@00`@40@H40`400`A<1002C0:M00=;0000
3P081`001PL?00=;00J4SXjCSW83C00314a^00FM01>a[0>/0k43/@>M0Id3/FjMC4^/00:400B>BhA;
4`001T^4SXj/SP:/00>WW@402@@30@L400<10`<00P<0178110@20`0<``A<14`4C0A<0Id11Id03FfW
[:N/Bda;C1RWST`00d/3SP=;00J>RD/40Id810H100U^17>a0k43/@<00P@019d1W@42C0>/00=<Q8@0
4P002d^>SZa<14`4W@6M00H4104610050Id1C0@00d`00if4B`0@00P700061`l00T/04AR4Q:b>[9>/
C0A<KYd1W@6M00<401Ba0k43/@6M0Id3106MC4]<SYf4Y`9;00<HSQP040001T^4Q8B/SP>/00>M0Id0
20@30@L400=^0@400`@40@<31`@02`?30id1W@6M0Ig300:/00=]C4`00T`03Jb>[4]<Bdb>YaQ<Bd`0
0Xh00d]]10030@P41@460`D400ZM0Id1[0>/LT]]4P000hA<SP02[0040d`410<11P@40@H40`400`A<
1002C003WDa;010020L000H73`02B`>400Q;Thj/Tja<10FM20@00`6M0@03100?0Ie<WD`H[4^MST/H
BhV>00:43P002AR4Q4a;[8j/Y`02W@030`@400D4104710810P@00da^W@050@030`@400<400Y<1:`3
[0A<0Id10Yd3[0BM00m]Yjb>C4]<BjN>LT]<Bhh00h@01Xj416d1W@P400>i0`<00P<00k43/@051006
C0BM0Jba0Z`00d]<Q00B0004Q4^/SP:/00Ja10BM0Id610@11P@00`6M0@05C003WD^4010020L000H7
3`000g1;Q002Q09<00F/Tjb//@03W@030Id400H400A^0Id110@01Id4[06M00:/00E<BhZ4W@02B`:>
00DHBhB>L00:0003BhB4009<00>C[:`00Yd0106a0`<410@11`@30@<400>M16h01@4710<300Y<14`4
C6j/0jc32Yd03HBMSZb4WHBMQ9f4WHh00T`01Ifa0Id1008400=c0`<020<81005KYd3[0<00T`00d^/
000A0003Bda<00:/00@3[0@40`4610<100>710@010@20@03W@A<009<00=]Q4/04@081`001PL@0004
L8B4Q0=<00=bTj`01Id810030Id100<400EcW@B//@02W@:/00A;C4^R0h@02D^>SQQ;68A`B`080003
68A;00=<0Z`3W@050j`3[0<00P@40@H40`431006WJ`410?31@404Vh410@3[0>/14`4C0A<`jc3XP^M
00R4[8jMQ9f4W@>401F>C4b>WJ`3W@6W0k43/@>a0k43W@4020@014`3[;42C003Bdb>014000A[C4^M
0j`010>/KYd20@H40`4710030Ie]00A<0Yd00d/0000?00P700061a0000B>Q8B40T`010A<1:`4W@03
`jac00D400<1W@400`@02@>a0D`3C0BM0`02C009BdbWWHBMQ9f>00A;00J>Q8B4BaP40003BhA<00=<
00H4W@6M0K450`04^@410@H40`401Fh419d100<40`<70@0310<300831P@02:`3[0A<16d11id03J:>
[8BMQ9f4WFe<SW800d`01Id419d100D300=h10400`400fh4100510050k43W@@00T`00hh0000@0003
Q4]<00:M00B/0j`3104610<11`@20@03C0A<009<00>MQ9<04@081`001PLA0003SXB400E<0244WIfM
Yjba[0>/0k43/@>a0k43106M0g84C::/WD]<Bj`00T/01HBMQ9f400=;00DHSXV>6003Q00568hHQ4/0
14`3W@05Y`>/0j`01@<00kT4100410<10P@00d`1[0021003L`<3008300<40@4010430`0?``A<14`4
C0>/0d`4C0A<00>/0d`01FfW[8i]00=<00J/STa<LXh2C003[9f/008400>M0@@01P@00fjM0@020@@4
00Ba0k430id00da;T`0A0004RD]<C0:M0Z`00`>/0@020@03Q@@400@40`461004C06M10A<00>4Thh0
4@081`001PLA0003BhB4009<00@4C0A<0id41006/@>a_Yd11@@00fjM0`02100HC0B/0id1WG:/C4]<
BdbWWHBMQ9e/YhjW1d/2C00714bMW@6M100?0`<110<01:`10I040`03L`@400<400<10`<00P<00kh1
0@031004L`<3``D400R/0j`4C0A<L`:/00E<14`4LP02[00814b/[0BM[9d210040id10@L400A^0@41
0P<00k43/@0210050Ie<[4`04@0018hHC4`2W@<400=b``400P440`03T@@400<11P@30@04C0A<LP:/
00>>000040081`001PLB0004BhB>T`=<00@4WIfM1`@00id1W@051009[0>/14`4[0=b00BM016/Bda;
C4^/Yg:4WHBMSZN>Y`04B`03Jda<009<00@4WIfM10@00g<30`090`<110<01Ja<W@6a00@31`@40`0@
XP410Id1[0>/0d`4C0A<0`:/00E<14`4C002[003LTa<009<0Z`01D`4C:b400:M00V/C0A<0j`3/G<0
0`@03G>a0id1W@>a0d`4C0@00Yd2[18000B>C4]<0Yd024`4C0BM0Id11`<00kT1KP061006W@5]LZba
0Z`00hh0000A00P700061a8000Q;Q9>>[9<4K@:M00=^10@01@@00`6MKP0310@300`110Ba0k44K@6M
0J`5C005[<>/C4/00Yd02Z:W[8jMQ9e/Q4/2C00519d1W@401P@00kT30`070`<110<014`410@50`L4
10<00hL10@070@D400>/0j`00`@02D`4[0>/14`4C002[003Ld`400FM00H3[0@40K480`@11P@0106M
[4/B0003Yhi<00:M1@@30@03_P<300h300L1W@>/0jba00:C4`081`001PLC0007KHA<SZbC[002W@03
C0A<00@400=^W@400`@04P>a0fd1W@Ab0j`4C0A<0Id1W@A<0j`03D]<Bda;[:N/Q9f4WF/014`3W@H4
10400fecLP030`<10`<00d`1B`031005[0>a0k400P45108301Va0`@4K@6M0Id1W@6M0Ie^C0>/0g84
C0A<00>/00A<14`40j`6W@04KTba[08400d1W@5b0k43/G>M0Id1008400A<14`40Yd00d/0000@0004
Q8jWC0:M00P4C0A<19d1W@l300Ja0id1[;42[0:C50081`001PLD0009Q4]<C:bC/Jba00H400>M0Id0
1@<2108100=]0`<00`@03D`4C06M0Id1C0B/0j`01D`4[0:M00=;Q4/00T`010BM0Id610@11P@30@@4
00L1C9d4[0>a008300Fi0@41Q`0310@31P@90@0D[0>/0id1KFjW`jL4K@6M0J`3W@43W@05KP@410<0
0P@019d1WFh510@11@@00k6MB`0B00:400F/YifMKP0510<100<40`<03@<01/<1W@>/0`:/00=;0000
4P081`001PLE0005Q4]<C7800j`03;6//J`3[06M0K43/@@401D1W@6a0d`4C0A<0j`1W@6M0J;3[<<0
1D`4[005C4];Jhh00d`2W@030@@400<400E^0@41KP0510<110@02Yd1[9f4WJba[0<21007W@6M0Iec
/@030`H401:a0j`3W@6M0Ifi[0>/0Id1W@47W@:/19d01;6/14`21006/J`1W@5]1@@02fd1W@5<14`3
[;5;01<000A[BhjW0Yd00d`4C0031003W@4400<430<01[41W@>//@:/00>400004`081`001PLF00A<
00fa[;6/0k43/@>W0Ie^00H400A]0@411@@03G<30d`1W@6M0j`3[0@014`01:`3[0<2C005BhB>[780
0Yd00`5]100410@11P@30@@400@1WJa]0d`00ifa[0031003KP4100<100>n0`<01P@40`<401I^0@43
/@>M0Id1W@6M0Id1W@6M0ja^0`@00`>/10031005Rid1W@<01@@30@07W@>a0k43C00C0004Q6e;[0:M
00>W10@00`@30@H400>A0`<020<20@06/@>//HA;5@081`001PLG00=<00F4WId4C0031005LPBM0Id0
1P@05`>M0Id1C0A<1:`3[0A<19d1[0>/^Id100=<00C3[:b/0T`014][SXh2[0:M1P@30@03K@@400<4
0`431005C0BM0AP01D`01JbMW@A<00<400NM0Id1W@5]00D400Ab0k431P@00j`3[0021005KYd1W@40
0id036h41:baC0A<14ba[0<401BM0Id1/@>a0k43/@>a0j`3LPA<[18000A[Q4]<0id01:ab14`21003
0Ie^00H400A]0@6i10<01[43W@6M/@9<00=;Q0005@081`001PLH009<0id00fd4100410030Id100D4
0`<40@8410<01D`4C0A<00<300nR0Id1W@6/0j`3C0A<Q8h01:`00kT410031004KP410@H40`441081
00Z/Bd`4C0A<Lj`30Yd510030`4100@10`@017P30`<610<300>a10@00P@017>a0k451003[0<400@4
00T3/@>M0Id1W@@01P<00k43K@031003C:`0014000=[Q4/00T`2W@040j`3LP<40P4710<11`@30@05
C0A<WD/05P081`001PLI0003STbM00:M00=<14`00`@019d1WFh2100R/@>a0g81W@6M0Id3/@=<14`4
C0B/0j`1W@6M0Id1W@6MK@9<00=[SXh01:`00`=b100210@11P@30@<400E<0Id3B`02C00814`4LP>/
0j`2W@0B14`4[0>/0id1W@6M0Fd3/@>a1@@01:`3[0<510030j`300D400<3[0<01@@01P>/0id1W@L4
00NM0Id4C0A<01<000I`Q4]<Bd`2W@06[0>/0g840`4710<11P@0106M0Fd3C003Q00001D020L000H7
6P001:bCWId610040Id1/@D30`@40@<300=b10@00`@40`8400=<0@40204028B>[:ba[0>/0`<00`@1
0@020@H40P4310830P401JbMC0A<008400=^0@401@42W@040j`3/@8400=^0@400P440`H40`<51004
L`>a0`@400Ac0k430`@01P6M0K43/@L400NM0Id4C0B/01<000=`Q8@00d`01Fd1[0>/00<300Ac0@67
1P@016h10@46108100JM14a<Q4/G00P700061a/00T`03Jb4WJba[;6/0j`3XP401`@01G>a0k4300<1
0`@40`0?C0A<14`4LP>/0ja^W@5[009<00V;[:b/0j`3/@<00`4>0`060Fe<Q06M20481006Lj`4C0A<
0`@04:`3[0>M0Id1W@6M0Id1KFh3100@0j`3LFjM0Id1W@6M0Id1W@<400Za0k43/@6M0J`45@0014][
Q8h3C0:M00M<1783[0>a00<31`@40@D400BM0Id40T`00hA;000G00P700061a`00T`03@B4WDacLP=b
19d1W@401@@50`H100<40`<00P<610@300NA10@4Q4`400:M00>/0j`00`<01;h10@4=0`8100<4C9d0
0`400hL410040`T40`<2W@<400@3/@>a0`@06Yd10@6M0Id1W@6M0Id3/@>R0Id1W@6M0Id320@01Id1
W@5<01H000E;Q8B>[002C003KIfM00<41`<00g<4100310@11P@0206M0D`4C9e;60081`001PLN0007
Bff4WD`4C0031005W@6M0Id00P@01@>a0k43008400=]0@4010400l<310041003C0@300<300E<14`4
J`02C0>M00Eb0j`3[0040@/30P@01Yd1C4b4B`D410<910@300=]0@400Yd00k6//@07100?0k43K@6M
0Id1W@>/0j`100@400Eb0k43/@031007C06M0D`4C00F0006Bf^4SZN>0T`3W@05C0A<1:`01`<410@1
10@02D`1W@5<14bMB`0I00P700061al000F>C8BMC0051003W@6M00811@<610P110@50`D40T`3W@@4
00=c0@400P480`03L`@400840P4010A<WD/510@32@@00`<10@080@:M00=]10@010@00`>a0`071003
0k4300P400Ja0k43/@<2/AT00h@00fb/SP02[0:M00<110@010@80`03L`@100<100=^10@00`@20@05
W@A<C8@06P081`001PLQ0006STbMC0A<10@01`6M0Id3/G<01`@00`>n0@080@0310<300831P@2C0>M
1@@30@07QG=bLg83LP020@D40P401Fe<WFdH00@410<51003K@4100P11@@017830`<4W@0?C;6//D`4
C0A<14`4[;6/00@400I^WIfM14`L0009BaR4Bdb>[:N/00:M00=<14`00`@20@T310451004K@6MKP=<
6`081`001PLS000>SZb//J`3/@>a0Id1WFh810@320440`H40T`00ffMW@0510<11`@20@D40`401Jab
Q4^400<400Eb0`<3R`0;0@030i4400L410<50@8400^a[9fM0Id1W@6M0@02WB<000B>BhA;0T`4[003
0`@400@410400h/30`040`@100<310@00`@01`6M0D`4C7806`081`001PLT0003Bjb/00:/00Za[;6/
0id1W@6M1@@01;430`<510<11@<50@0414a<C0:M1@@30@H400=]0@401@@20@0:[0=;WD^>16h1^@<3
00>i0@401P400`@30`020`P400g30@6M0Ie^C0A<14^M02d000B>68B40d`00jN/[002[00414`4C084
104210@31@440`06/KVR^G:a0Z`00h@0000I00P700061bH000A]WIe<1@@01Yd1W@6M0@8410<910@3
1`400hi<C002W@D40`461003Q`4100D400D10`>/C002W@05SX@10KT00`<01;T10Fh610@31P@00fd1
0@020@070k43/@A<W@0_0004SXV4Q0A<00JR[0>/0j`410@110@30`@11P<01k43/@>//H@070081`00
1PLX000CBif414`4C0A<KYd1W@>/0k43LP0810@30P@60@03STa<00:M00=<10@00P@30@L40P4410<3
00ZaLTa<Q:N>Q06i0`<00g<4100710@30P@02`6M0Id1W@B/0jba00:/00>C0000;@001aR>L8j4C4/0
0T`026fM[:`3[0=b0P@50@D400<30@400P460`04/@>//@:/00>>L@006`081`001PL[0005Q9fM14`0
0`@01k43/@>a0`402@@30`03T@@400H400>>[8h00Yd510<11`@20@<41@<01T`4C9f>[09;00Ec0`<3
L`0910@310401Yd110A<10:/00>C0000;P0018@HSXT2SP0:C4]<C6fM0J`3[0<300<4Q`400`461003
KP4100811`<01[43[;6CBad020L000H7;`000if4W@03[00@14`4C6jM0Id1W@6M`k43/@T400Eb0hjC
Q002W@04Ld`410<11`@20@D300BR0Ie^0T`2[00564/HSXT00`<2100>W@6M0Id1W@>/0j`3C0@2C003
Q9d003<000@HBhB40hh01:M<C4`2W@090D`4LP>/0j`300<11`@40@0=K@@414ac[0>//G9<Q00O00P7
00061c<000F4WD`4C0051005W@6M0Id010400fh41005108300ZCSZNM0J`3/@>i0P400i1c0`040`81
10<210<100D4[0>/W@03B`03Yh@100H100^MKP@4/@>//JbMQ00f000?RD^4Q9f4[8j/Yja<0Id100@4
10<30@03Q`@400D4104610030Id400=<00=;00007@081`001PLg000LWH@4C0A<Lj`3[0>M0Id1W@6M
0Id1W@>a0k6CSP:M00Ba[0>/2P<20@<30`@30@04C0>/[0:M0T/2SP0=Q05]14`4C0A<14bMQ00i0004
L8hHK0:400E;C8j/Y`02[0:M00T1C0A<14`4[0<00`400l<4100610<11@@014a^WFd2C003Bf`001l0
20L000H7>@002I>/[;6//J`3[0051003W@6M008110<00`5;LP02W@03[0>a00/30P400`<4100310<1
0`<03W81WFe;[8jM60A<14`40Yd00h@0000h0004SXV>R@:400E<Bda<LP02[004Y`6M0@P410420`03
L`@400<40`400hL410041003W@5<009<00>M00007`081`001PLo000DWHBMC0A<14`4C0A<0j`3[;VM
Q4`2W@04/J`3[0X30`46108100>/0j`00T`2W@07BhjWBh@HSP0n000464]`SPB400=;C4`00T`02ZbM
WId1C0A<14`310@110<0178410@40@@400E<19d1W@03C28020L000H7A003W@09C;6//Jba[0=<0084
0T`019d4Lj`90`03L`@100811P@20@8300d4C0A<0If/Yd^4WHB>03/00d/01hV4Q9f4C4/014`01PBM
0Id1``L41@460`03104100811P@016d1W@@2C2<020L000H7C0003XBMQ9e;C6]<Q9e<14`40`471081
1P@30@030d`4009<0id01:a;BaP3Q00368@003D000A[BaQ;18@00hVWB`04C0BM00H1[0>/14`310D1
00=c0`<01@<00kT10@041006C0A<0Id10T`T00P700061e<000=<KId010@30@03KP@400@40`441004
LP<10@8400Q<14`4W@6MC0=;00>MQ9d00T/00aP0000_0003Q8i;009;00RMQ9f4[8j/Y`9<00P4C06M
0Id1``<310@50@842@<00kT4100410050Id1C0@09@081`001PMC0008Q4bMW@A<16d30@H40`4310<3
0`400d`4C002C004`ifMW@=;00>>Yh@00X@00d/HQ00/0006Jhi`SQQ;0h@01ie;LXj/SZ`00T`3W@08
0Id4[0>/0j`310D10`@00g830`070`08LPA<14`1W@42C003B`0002<020L000H7E0000db4W@0410<1
1P@30@841@<30@84015<0j`3W@5;C4^/SZa;WHBMK002SP03Q00002@000<HBd/018h02Yf4WHA<Bdab
[:L2[004Lfd1W@810`@40`03LXL100@11P@;0`05LP6M0D`00W8V00P700061e@000E`C8A^C00210@1
1@@30@H300/40@4114`4[0>/``03W@04C:N/SP9;00=]Q8@00Xh01HU;64/H01h000R464/HBaQ;R@>>
0X@01aQ;Bda;C:L01:`03@6M0D`4C0A<0j`3``<010400fd410061003^@<300@300Ja0j`1W@42[003
T`0002D020L000H7EP000jbR100310<11@@70`@410430`07Ld`4W@6M``02[00?Bda;C4^/SZb4WHBM
Q6a;00>>1T/02aR>RHj9LAQ;64/H009;18h03W4HWHBMQ9f4[8j/SW9;14`020A]`j`3[0<120@00i43
0`030`T410400`BA0`030`04W@6M/@:/:@081`001PMF0007SZb/Ld`4K@030@X31@@00fh10@040`03
C0A<009<00L1[:b/KD]<009;00Z>[8jWQ9f4WHBM0X@058Ua64/HBaQ;RHj9SXU;64/HBaQ/1h@02Yf4
Bd]<SZbW[:L4C0:M00X1X/>/0j`3C0A<1@@00fd30`030`03LP@400D4104510050id1WK400Z`Z00P7
00061eL000J>[:`3/@<30@P30P4610030KT300<300/4C0A<1783[0>M0@02C00[BjbW[:M<Bda;KHBM
Q9f4WHBMQ9f4YhjWSZN4WHBMQ9f4WHBMQ9f4C4]<B`02C00@[:N/[<=<14`1W@6M0L<3``831`@30@D3
00=c10@00`@40@L400<1WG800T`[00P700061eP000N>/J`3[06M00811@<00fd10@020@D400=b0`<0
0P<02`5<14`4C0>/0j`100>M00M<YjbW[4]<00A;01n>YhVMQ9f4WHBMQ:N>YhjMQ9f4WHBMSZN>Yhi<
Bda;00=<00>W[:`00Z`03DbM0Id1W@5<0j`3[0<01@@50@03^@<300<300Ab10@410461003C06M00=<
;0081`001PMI0005[0>/0k400`430`<4104510@300>n0@4010@20`0/``=]0Id1W@6/`ja<Bda;C4^/
YjbWC4]<Bda;LXj/SZa;C4]<Bda;[:N/Yj`7C00:`jbnW@6M0Id1K@<400=b0`<00P<31003K@4100@1
00<40`<010<00g<40@030@L400<1W@@00T`]00P700061eT000VMC;6/0j`1W@401P@40@<41@<40@0@
C0C30j`3C0A<KId1W@6MY`9<01e;C4]<YjbW[4]<Bda;C4^/YjbWC4]<Bda;C:N/Y`02[0I<00X1W@6M
0Id1W@5<1@@01/<3``=b10H10`@60`03``4100811P@01Fd1WFe<02l020L000H7F@000d^MC0031004
K@410@D400=^0@400P400`@30`030`03104100<100<40`<00P<03PA<14`4W@6M0Id1W@5]0T`01<>/
[:`7C004YjbW[0M<00ib[<>/`d`4C6fM0Id1W@<100>n10@01P@40`H100>710@00`@40`D11P@01Id1
W@A<030020L000H7FP002T^MC0A<19d1WFh510@100>;0`<00P<310D110<044`4C0A<1:bnW@6M0Id1
W@44[0M<1:`7C00CLZb/`j`1W@6M0Id1W@6/0j`3C00810831P471003LP<300D100T3L`@4C0BM0Id0
0d``00P700061e/000=<WD`00`@30@H410430`03LP@400841@400l<31005100JLP?30l<1W@6M0Id3
[0>/0Id1W@5]1:`3Ykh2K@0;W@6M0Id1[0>/0j801@4210D300><10@01P@60@X400=^0@400P401<<3
0`<21006KYd1K@A<<P081`001PML0008C9e<14`4W@4710@100<3LP@01P@50@<403M<14`3[0>/Ld`4
C0BM0J`3[<>M0Id1W@6M0Id1W@6M0Id1W@6M0j`3[0>M0D`4C0A<Lj`3[0>/00D400=]0@4010400l<3
100810D11@<01:`3XP42C003B`00034020L000H7GP000ie<1002108100=^T@<01@<40@L40P<70@84
10<710@31P@00fjn``030`L41@<810D300=c0@401P400i430`030`D41@4310@300Fa0k6a[00f00P7
00061el000R>Cg9bLk43/@L310451005L`<30`@01`440`0V14`4C0A<1:`3[0=<14`4C0A<0j`3``=<
14`4C0A<0j`3[0=<14`510H31P4310H30P@00fh10@020@03K@@400<400A<0j`30Z`00i<0000e00P7
00061f4000>W/J`020<00i440@040@@410<21003KP4100<110<710@31`@50`L41@<81003T0<300<3
0`400fh4100410D300=c0@400`481005WKVa/I<0>0081`001PMR0008Q:b/0k43/@<51003KP4100<1
00<40`<00P<610<10`<00kT1K@05100@``?30d`4C0A<14`3``?30`H400M<0l<3``=b00@410450`03
0@@400L41P<01;T10@441006C0A<0Id10T`j00P700061f<000>40j`00P<20@H400>70@400P450`P4
10<60@03Q`@300<31`@50`L400=b0`<00`<00fh10@050@D32P@70`030@@400D400BM0Id40T`k00P7
00061f@000M<14`4W@6M00L40P450`T410<;0@03``@400D41@<710h100>;0`<00P<:1003^@<300D3
10@014`1W@42C003B`0003/020L000H7I@001D`410BM008100>710@010@60`P400Ic0`<3L`@B0@D3
4P4410D31`@30@03^@<300D300Ic16d1W@42C3l020L000H7IP0028@4C0BM0Id110@60`8100=]10@0
10@40`H400=^0@402`450`h100=]10@01@@50`D41P460`05/@=b0Id00Ta100P700061fP000@4C0@4
0P400fh40`060`03^@4100<110@40`P40P<00l<10@070@D320400l<30`020`03L`@400H400>A0`<0
0`<00`A^0@050@<41P<01Id1[79;048020L000H7JP001d`4C0BM0K401P<210L110<910@320@50`P4
1P<910<320461004/@>/0@:M0Za400P700061f`000>a[<<00`430`D420400l<3100810@320@50`P4
1P<91003``4100D100=^10@01P@01P6M0Ifa[4H020L000H7KP0029>/[7>M0Id120@00`?30@080@<4
1@<810D32@@50`842P440`H400>M0Id00Ta;00P700061g0000H4C0@40Id20@D41@<00l<10@070@@3
00>A10@01P@50`T41@<00kT10@050@03Q`@400D300Mc10@40Id4009<C@081`001PMb0008C0A<19d1
W@4210L30P@00fh10@040@@31P400fd410040`03`fe^00L11P<40@H400/3/@>a0jbiW@6MC01@00P7
00061gD000A<14`40P470`P400<10`<00`<G0@H32P@02783/@>a0jac0TaA00P700061gP000U</J`3
/@>a0k402P@40`03R`4101@100<40`<010<71006C0A]0Jba0Z`00i<0001C00P700061gX00Z`2/@03
0k4100811`@00i430`020`03L`@400H400>A0`<00`<91003L`<300<31P@01id1W@5<LY<0F0081`00
1PMm000:[4`4C6jM0Id1W@<400>a0k400P<910H32P@02[43/@>aKYd1W@42W@9<00=;0000F@081`00
1PN1000;C0A<19d1W@6M0Id00P<00g<4100710H32P@02@6M0Id1W@5<1002C5l020L000H7Q@02C00S
1:baW@6M0Id1W@6M0Id1WFj/0j`3[0>M0Id1W@6M0Id1WK400Z`00i<0001S00P700061h/000NC[:`4
C0A<0084015^W@6M0k43/@>/14`4C0A<1002C6/020L000H7U`000d`0001h00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`00
1POo01<020L000H7o`0C00P7003o1b47003o1b47003o1b47003o1b47003o1b47003o1b470000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output",
  CellLabel->"Out[24]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[TextData[{
  "As long as one only works with ",
  StyleBox["MeshGraphics3D[]",
    FontFamily->"Courier"],
  " structures one can have more than one mesh in the scene and every mesh \
can have its individual texture."
}], "Text",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \(\(scurve = 
      FrenetFrame[{Cos[4*t]*Sin[t/2], Sin[4*t]*Sin[t/2], Cos[t/2]}, t, 
        t/2 + \[Phi], r*Sin[t]]; \)\)], "Input",
  CellLabel->"In[25]:=",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sband = 
      ParametricPlot3D[
        Evaluate[scurve\  /. \ \[Phi] -> 0], {r, \(-0.15\), 0.10}, {t, 0, 
          2  \[Pi]}, PlotPoints -> {6, 144}, 
        ViewPoint -> {0.204845, \(-1.424743\), 2.49310}]\)], "Input",
  CellLabel->"In[26]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.11857 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0460648 1.17806 -0.0594909 1.17806 [
[ 0 0 0 0 ]
[ 1 1.11857 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
.1516 .09701 m
0 .41812 L
s
0 .41812 m
.21058 1.11857 L
s
.21058 1.11857 m
.29073 .67238 L
s
.29073 .67238 m
.1516 .09701 L
s
.84177 0 m
.88762 .60026 L
s
.88762 .60026 m
1 1.03465 L
s
1 1.03465 m
.96125 .29266 L
s
.96125 .29266 m
.84177 0 L
s
.1516 .09701 m
0 .41812 L
s
0 .41812 m
.96125 .29266 L
s
.96125 .29266 m
.84177 0 L
s
.84177 0 m
.1516 .09701 L
s
.29073 .67238 m
.88762 .60026 L
s
.88762 .60026 m
1 1.03465 L
s
1 1.03465 m
.21058 1.11857 L
s
.21058 1.11857 m
.29073 .67238 L
s
0 0 m
1 0 L
1 1.11857 L
0 1.11857 L
closepath
clip
newpath
.5 Mabswid
.51912 .50956 m .5445 .51696 L .54119 .53125 L p .717 .578 .697 r
F P
0 g
s
.5445 .51696 m .57171 .52041 L .5706 .53532 L p .697 .557 .692 r
F P
0 g
s
.5706 .53532 m .54119 .53125 L .5445 .51696 L p .697 .557 .692 r
F P
0 g
s
.54119 .53125 m .51374 .52288 L .51912 .50956 L p .717 .578 .697 r
F P
0 g
s
.51374 .52288 m .54119 .53125 L .53788 .54556 L p .717 .578 .697 r
F P
0 g
s
.54119 .53125 m .5706 .53532 L .5695 .55025 L p .697 .557 .692 r
F P
0 g
s
.5695 .55025 m .53788 .54556 L .54119 .53125 L p .697 .557 .692 r
F P
0 g
s
.49624 .49868 m .51912 .50956 L .51374 .52288 L p .736 .599 .704 r
F P
0 g
s
.57171 .52041 m .59997 .51955 L .60113 .53472 L p .675 .538 .69 r
F P
0 g
s
.60113 .53472 m .5706 .53532 L .57171 .52041 L p .675 .538 .69 r
F P
0 g
s
.53788 .54556 m .50835 .53621 L .51374 .52288 L p .717 .578 .697 r
F P
0 g
s
.50835 .53621 m .53788 .54556 L .53456 .55987 L p .718 .578 .697 r
F P
0 g
s
.53788 .54556 m .5695 .55025 L .5684 .56519 L p .697 .557 .692 r
F P
0 g
s
.5684 .56519 m .53456 .55987 L .53788 .54556 L p .697 .557 .692 r
F P
0 g
s
.51374 .52288 m .48899 .5107 L .49624 .49868 L p .736 .599 .704 r
F P
0 g
s
.48899 .5107 m .51374 .52288 L .50835 .53621 L p .736 .599 .704 r
F P
0 g
s
.5706 .53532 m .60113 .53472 L .6023 .54991 L p .675 .538 .69 r
F P
0 g
s
.6023 .54991 m .5695 .55025 L .5706 .53532 L p .675 .538 .69 r
F P
0 g
s
.53456 .55987 m .50296 .54955 L .50835 .53621 L p .718 .578 .697 r
F P
0 g
s
.50296 .54955 m .53456 .55987 L .53124 .5742 L p .718 .578 .697 r
F P
0 g
s
.53456 .55987 m .5684 .56519 L .56729 .58015 L p .697 .557 .692 r
F P
0 g
s
.56729 .58015 m .53124 .5742 L .53456 .55987 L p .697 .557 .692 r
F P
0 g
s
.50835 .53621 m .48173 .52273 L .48899 .5107 L p .736 .599 .704 r
F P
0 g
s
.48173 .52273 m .50835 .53621 L .50296 .54955 L p .736 .599 .704 r
F P
0 g
s
.53124 .5742 m .49756 .5629 L .50296 .54955 L p .718 .578 .697 r
F P
0 g
s
.49756 .5629 m .53124 .5742 L .52792 .58853 L p .718 .578 .697 r
F P
0 g
s
.53124 .5742 m .56729 .58015 L .56619 .59512 L p .697 .557 .692 r
F P
0 g
s
.56619 .59512 m .52792 .58853 L .53124 .5742 L p .697 .557 .692 r
F P
0 g
s
.5695 .55025 m .6023 .54991 L .60347 .56511 L p .675 .538 .69 r
F P
0 g
s
.60347 .56511 m .5684 .56519 L .5695 .55025 L p .675 .538 .69 r
F P
0 g
s
.50296 .54955 m .47447 .53477 L .48173 .52273 L p .736 .599 .704 r
F P
0 g
s
.47447 .53477 m .50296 .54955 L .49756 .5629 L p .736 .599 .704 r
F P
0 g
s
.47642 .48487 m .49624 .49868 L .48899 .5107 L p .752 .62 .713 r
F P
0 g
s
.52792 .58853 m .49216 .57626 L .49756 .5629 L p .718 .578 .697 r
F P
0 g
s
.5684 .56519 m .60347 .56511 L .60464 .58034 L p .675 .538 .69 r
F P
0 g
s
.60464 .58034 m .56729 .58015 L .5684 .56519 L p .675 .538 .69 r
F P
0 g
s
.49756 .5629 m .46719 .54682 L .47447 .53477 L p .736 .599 .704 r
F P
0 g
s
.46719 .54682 m .49756 .5629 L .49216 .57626 L p .736 .599 .704 r
F P
0 g
s
.48899 .5107 m .46758 .49529 L .47642 .48487 L p .752 .62 .713 r
F P
0 g
s
.46758 .49529 m .48899 .5107 L .48173 .52273 L p .752 .62 .713 r
F P
0 g
s
.49216 .57626 m .45991 .55888 L .46719 .54682 L p .736 .599 .704 r
F P
0 g
s
.56729 .58015 m .60464 .58034 L .60581 .59558 L p .675 .538 .69 r
F P
0 g
s
.60581 .59558 m .56619 .59512 L .56729 .58015 L p .675 .538 .69 r
F P
0 g
s
.59997 .51955 m .62843 .51413 L .63185 .52919 L p .651 .52 .691 r
F P
0 g
s
.63185 .52919 m .60113 .53472 L .59997 .51955 L p .651 .52 .691 r
F P
0 g
s
.48173 .52273 m .45873 .50573 L .46758 .49529 L p .752 .62 .713 r
F P
0 g
s
.45873 .50573 m .48173 .52273 L .47447 .53477 L p .752 .62 .713 r
F P
0 g
s
.60113 .53472 m .63185 .52919 L .63528 .54427 L p .651 .52 .691 r
F P
0 g
s
.63528 .54427 m .6023 .54991 L .60113 .53472 L p .651 .52 .691 r
F P
0 g
s
.47447 .53477 m .44986 .51618 L .45873 .50573 L p .752 .62 .713 r
F P
0 g
s
.44986 .51618 m .47447 .53477 L .46719 .54682 L p .752 .62 .713 r
F P
0 g
s
.6023 .54991 m .63528 .54427 L .63871 .55938 L p .651 .52 .691 r
F P
0 g
s
.63871 .55938 m .60347 .56511 L .6023 .54991 L p .651 .52 .691 r
F P
0 g
s
.46719 .54682 m .44099 .52665 L .44986 .51618 L p .752 .62 .713 r
F P
0 g
s
.44099 .52665 m .46719 .54682 L .45991 .55888 L p .752 .62 .713 r
F P
0 g
s
.4601 .46875 m .47642 .48487 L .46758 .49529 L p .766 .642 .725 r
F P
0 g
s
.60347 .56511 m .63871 .55938 L .64215 .57451 L p .651 .52 .691 r
F P
0 g
s
.64215 .57451 m .60464 .58034 L .60347 .56511 L p .651 .52 .691 r
F P
0 g
s
.45991 .55888 m .43209 .53713 L .44099 .52665 L p .752 .62 .713 r
F P
0 g
s
.46758 .49529 m .45001 .47734 L .4601 .46875 L p .766 .642 .725 r
F P
0 g
s
.45001 .47734 m .46758 .49529 L .45873 .50573 L p .766 .642 .725 r
F P
0 g
s
.60464 .58034 m .64215 .57451 L .6456 .58967 L p .651 .52 .691 r
F P
0 g
s
.6456 .58967 m .60581 .59558 L .60464 .58034 L p .651 .52 .691 r
F P
0 g
s
.45873 .50573 m .4399 .48594 L .45001 .47734 L p .766 .642 .725 r
F P
0 g
s
.4399 .48594 m .45873 .50573 L .44986 .51618 L p .766 .642 .725 r
F P
0 g
s
.62843 .51413 m .65618 .50402 L .66177 .51861 L p .625 .503 .694 r
F P
0 g
s
.66177 .51861 m .63185 .52919 L .62843 .51413 L p .625 .503 .694 r
F P
0 g
s
.44986 .51618 m .42977 .49457 L .4399 .48594 L p .766 .642 .725 r
F P
0 g
s
.42977 .49457 m .44986 .51618 L .44099 .52665 L p .766 .642 .725 r
F P
0 g
s
.63185 .52919 m .66177 .51861 L .66737 .53323 L p .625 .503 .694 r
F P
0 g
s
.66737 .53323 m .63528 .54427 L .63185 .52919 L p .625 .503 .694 r
F P
0 g
s
.44099 .52665 m .41961 .50321 L .42977 .49457 L p .766 .642 .725 r
F P
0 g
s
.41961 .50321 m .44099 .52665 L .43209 .53713 L p .766 .642 .725 r
F P
0 g
s
.44759 .45097 m .4601 .46875 L .45001 .47734 L p .776 .663 .738 r
F P
0 g
s
.63528 .54427 m .66737 .53323 L .67299 .54788 L p .625 .503 .694 r
F P
0 g
s
.67299 .54788 m .63871 .55938 L .63528 .54427 L p .625 .503 .694 r
F P
0 g
s
.43209 .53713 m .40944 .51187 L .41961 .50321 L p .766 .642 .725 r
F P
0 g
s
.45001 .47734 m .43663 .45756 L .44759 .45097 L p .776 .663 .738 r
F P
0 g
s
.43663 .45756 m .45001 .47734 L .4399 .48594 L p .776 .663 .738 r
F P
0 g
s
.63871 .55938 m .67299 .54788 L .67862 .56257 L p .625 .504 .694 r
F P
0 g
s
.67862 .56257 m .64215 .57451 L .63871 .55938 L p .625 .504 .694 r
F P
0 g
s
.4399 .48594 m .42564 .46416 L .43663 .45756 L p .776 .663 .738 r
F P
0 g
s
.42564 .46416 m .4399 .48594 L .42977 .49457 L p .776 .663 .738 r
F P
0 g
s
.64215 .57451 m .67862 .56257 L .68427 .5773 L p .625 .504 .694 r
F P
0 g
s
.68427 .5773 m .6456 .58967 L .64215 .57451 L p .625 .504 .694 r
F P
0 g
s
.42977 .49457 m .41463 .47078 L .42564 .46416 L p .776 .663 .738 r
F P
0 g
s
.41463 .47078 m .42977 .49457 L .41961 .50321 L p .776 .663 .738 r
F P
0 g
s
.65618 .50402 m .68225 .48925 L .68985 .50302 L p .598 .489 .701 r
F P
0 g
s
.68985 .50302 m .66177 .51861 L .65618 .50402 L p .598 .489 .701 r
F P
0 g
s
.43904 .43226 m .44759 .45097 L .43663 .45756 L p .784 .683 .752 r
F P
0 g
s
.41961 .50321 m .40359 .47741 L .41463 .47078 L p .776 .663 .738 r
F P
0 g
s
.40359 .47741 m .41961 .50321 L .40944 .51187 L p .776 .663 .738 r
F P
0 g
s
.66177 .51861 m .68985 .50302 L .69748 .51683 L p .598 .489 .701 r
F P
0 g
s
.69748 .51683 m .66737 .53323 L .66177 .51861 L p .598 .489 .701 r
F P
0 g
s
.43663 .45756 m .42764 .43673 L .43904 .43226 L p .784 .683 .752 r
F P
0 g
s
.42764 .43673 m .43663 .45756 L .42564 .46416 L p .784 .683 .752 r
F P
0 g
s
.40944 .51187 m .39252 .48406 L .40359 .47741 L p .776 .663 .738 r
F P
0 g
s
.66737 .53323 m .69748 .51683 L .70513 .53069 L p .598 .489 .701 r
F P
0 g
s
.70513 .53069 m .67299 .54788 L .66737 .53323 L p .598 .489 .701 r
F P
0 g
s
.42564 .46416 m .4162 .44122 L .42764 .43673 L p .784 .683 .752 r
F P
0 g
s
.4162 .44122 m .42564 .46416 L .41463 .47078 L p .784 .683 .752 r
F P
0 g
s
.67299 .54788 m .70513 .53069 L .71281 .5446 L p .598 .489 .701 r
F P
0 g
s
.71281 .5446 m .67862 .56257 L .67299 .54788 L p .598 .489 .701 r
F P
0 g
s
.43445 .41329 m .43904 .43226 L .42764 .43673 L p .789 .701 .767 r
F P
0 g
s
.41463 .47078 m .40473 .44573 L .4162 .44122 L p .784 .683 .752 r
F P
0 g
s
.40473 .44573 m .41463 .47078 L .40359 .47741 L p .784 .683 .752 r
F P
0 g
s
.67862 .56257 m .71281 .5446 L .72051 .55855 L p .598 .489 .701 r
F P
0 g
s
.72051 .55855 m .68427 .5773 L .67862 .56257 L p .598 .489 .701 r
F P
0 g
s
.40359 .47741 m .39322 .45025 L .40473 .44573 L p .784 .683 .752 r
F P
0 g
s
.39322 .45025 m .40359 .47741 L .39252 .48406 L p .784 .683 .752 r
F P
0 g
s
.42764 .43673 m .42306 .41564 L .43445 .41329 L p .789 .701 .767 r
F P
0 g
s
.42306 .41564 m .42764 .43673 L .4162 .44122 L p .789 .701 .768 r
F P
0 g
s
.68225 .48925 m .70565 .46996 L .71504 .4826 L p .569 .477 .71 r
F P
0 g
s
.71504 .4826 m .68985 .50302 L .68225 .48925 L p .569 .477 .71 r
F P
0 g
s
.39252 .48406 m .38168 .45478 L .39322 .45025 L p .784 .683 .752 r
F P
0 g
s
.4162 .44122 m .41162 .418 L .42306 .41564 L p .789 .701 .768 r
F P
0 g
s
.41162 .418 m .4162 .44122 L .40473 .44573 L p .789 .701 .768 r
F P
0 g
s
.68985 .50302 m .71504 .4826 L .72447 .4953 L p .57 .477 .71 r
F P
0 g
s
.72447 .4953 m .69748 .51683 L .68985 .50302 L p .57 .477 .71 r
F P
0 g
s
.43371 .39475 m .43445 .41329 L .42306 .41564 L p .79 .717 .783 r
F P
0 g
s
.40473 .44573 m .40013 .42036 L .41162 .418 L p .789 .701 .768 r
F P
0 g
s
.40013 .42036 m .40473 .44573 L .39322 .45025 L p .789 .701 .768 r
F P
0 g
s
.69748 .51683 m .72447 .4953 L .73394 .50805 L p .57 .477 .71 r
F P
0 g
s
.73394 .50805 m .70513 .53069 L .69748 .51683 L p .57 .477 .71 r
F P
0 g
s
.42306 .41564 m .42275 .39505 L .43371 .39475 L p .79 .717 .783 r
F P
0 g
s
.42275 .39505 m .42306 .41564 L .41162 .418 L p .79 .717 .784 r
F P
0 g
s
.39322 .45025 m .3886 .42274 L .40013 .42036 L p .789 .701 .768 r
F P
0 g
s
.3886 .42274 m .39322 .45025 L .38168 .45478 L p .789 .701 .768 r
F P
0 g
s
.54431 .3627 m .53778 .36528 L .54397 .36213 L closepath p .696 .672 .823 r
F P
0 g
s
.54466 .36326 m .53778 .36528 L .54431 .3627 L closepath p .696 .672 .823 r
F P
0 g
s
.545 .36383 m .53778 .36528 L .54466 .36326 L closepath p .696 .672 .823 r
F P
0 g
s
.54534 .36439 m .53778 .36528 L .545 .36383 L closepath p .696 .672 .823 r
F P
0 g
s
.54568 .36495 m .53778 .36528 L .54534 .36439 L closepath p .696 .672 .823 r
F P
0 g
s
.55375 .36182 m .54568 .36495 L .54534 .36439 L p .692 .671 .824 r
F P
0 g
s
.55268 .36095 m .54534 .36439 L .545 .36383 L p .693 .671 .824 r
F P
0 g
s
.54534 .36439 m .55268 .36095 L .55375 .36182 L p .692 .671 .824 r
F P
0 g
s
.55161 .36007 m .545 .36383 L .54466 .36326 L p .694 .671 .824 r
F P
0 g
s
.545 .36383 m .55161 .36007 L .55268 .36095 L p .693 .671 .824 r
F P
0 g
s
.70513 .53069 m .73394 .50805 L .74345 .52086 L p .57 .477 .71 r
F P
0 g
s
.74345 .52086 m .71281 .5446 L .70513 .53069 L p .57 .477 .71 r
F P
0 g
s
.55054 .3592 m .54466 .36326 L .54431 .3627 L p .694 .672 .824 r
F P
0 g
s
.54466 .36326 m .55054 .3592 L .55161 .36007 L p .694 .671 .824 r
F P
0 g
s
.54947 .35832 m .54431 .3627 L .54397 .36213 L p .695 .672 .823 r
F P
0 g
s
.54431 .3627 m .54947 .35832 L .55054 .3592 L p .694 .672 .824 r
F P
0 g
s
.54397 .36213 m .5484 .35744 L .54947 .35832 L p .695 .672 .823 r
F P
0 g
s
.43651 .37728 m .43371 .39475 L .42275 .39505 L p .787 .73 .8 r
F P
0 g
s
.41162 .418 m .41175 .39534 L .42275 .39505 L p .79 .717 .784 r
F P
0 g
s
.41175 .39534 m .41162 .418 L .40013 .42036 L p .79 .717 .784 r
F P
0 g
s
.5608 .3558 m .55375 .36182 L .55268 .36095 L p .688 .667 .825 r
F P
0 g
s
.55881 .35498 m .55268 .36095 L .55161 .36007 L p .688 .667 .824 r
F P
0 g
s
.55268 .36095 m .55881 .35498 L .5608 .3558 L p .688 .667 .825 r
F P
0 g
s
.38168 .45478 m .37703 .42513 L .3886 .42274 L p .789 .701 .768 r
F P
0 g
s
.55683 .35417 m .55161 .36007 L .55054 .3592 L p .689 .667 .824 r
F P
0 g
s
.55161 .36007 m .55683 .35417 L .55881 .35498 L p .688 .667 .824 r
F P
0 g
s
.55484 .35336 m .55054 .3592 L .54947 .35832 L p .69 .667 .823 r
F P
0 g
s
.55054 .3592 m .55484 .35336 L .55683 .35417 L p .689 .667 .824 r
F P
0 g
s
.55285 .35255 m .54947 .35832 L .5484 .35744 L p .69 .667 .823 r
F P
0 g
s
.54947 .35832 m .55285 .35255 L .55484 .35336 L p .69 .667 .823 r
F P
0 g
s
.71281 .5446 m .74345 .52086 L .75301 .53373 L p .57 .477 .71 r
F P
0 g
s
.75301 .53373 m .72051 .55855 L .71281 .5446 L p .57 .477 .71 r
F P
0 g
s
.5484 .35744 m .55087 .35173 L .55285 .35255 L p .69 .667 .823 r
F P
0 g
s
.70565 .46996 m .72538 .44648 L .73627 .45772 L p .54 .468 .724 r
F P
0 g
s
.73627 .45772 m .71504 .4826 L .70565 .46996 L p .54 .468 .724 r
F P
0 g
s
.56568 .34722 m .5608 .3558 L .55881 .35498 L p .683 .666 .827 r
F P
0 g
s
.40013 .42036 m .4007 .39563 L .41175 .39534 L p .79 .717 .784 r
F P
0 g
s
.4007 .39563 m .40013 .42036 L .3886 .42274 L p .79 .717 .784 r
F P
0 g
s
.5628 .34687 m .55881 .35498 L .55683 .35417 L p .684 .666 .827 r
F P
0 g
s
.55881 .35498 m .5628 .34687 L .56568 .34722 L p .683 .666 .827 r
F P
0 g
s
.42275 .39505 m .42642 .37567 L .43651 .37728 L p .787 .73 .8 r
F P
0 g
s
.42642 .37567 m .42275 .39505 L .41175 .39534 L p .787 .73 .8 r
F P
0 g
s
.55992 .34651 m .55683 .35417 L .55484 .35336 L p .684 .666 .826 r
F P
0 g
s
.55683 .35417 m .55992 .34651 L .5628 .34687 L p .684 .666 .827 r
F P
0 g
s
.55705 .34615 m .55484 .35336 L .55285 .35255 L p .684 .666 .826 r
F P
0 g
s
.55484 .35336 m .55705 .34615 L .55992 .34651 L p .684 .666 .826 r
F P
0 g
s
.55417 .34579 m .55285 .35255 L .55087 .35173 L p .685 .665 .825 r
F P
0 g
s
.55285 .35255 m .55417 .34579 L .55705 .34615 L p .684 .666 .826 r
F P
0 g
s
.55087 .35173 m .55129 .34543 L .55417 .34579 L p .685 .665 .825 r
F P
0 g
s
.44246 .36143 m .43651 .37728 L .42642 .37567 L p .782 .74 .815 r
F P
0 g
s
.71504 .4826 m .73627 .45772 L .74722 .46903 L p .54 .468 .724 r
F P
0 g
s
.74722 .46903 m .72447 .4953 L .71504 .4826 L p .54 .468 .724 r
F P
0 g
s
.3886 .42274 m .3896 .39593 L .4007 .39563 L p .79 .717 .784 r
F P
0 g
s
.3896 .39593 m .3886 .42274 L .37703 .42513 L p .79 .717 .784 r
F P
0 g
s
.41175 .39534 m .41627 .37405 L .42642 .37567 L p .787 .73 .8 r
F P
0 g
s
.41627 .37405 m .41175 .39534 L .4007 .39563 L p .787 .73 .8 r
F P
0 g
s
.56746 .33677 m .56568 .34722 L .5628 .34687 L p .679 .668 .832 r
F P
0 g
s
.56389 .3372 m .5628 .34687 L .55992 .34651 L p .68 .668 .832 r
F P
0 g
s
.5628 .34687 m .56389 .3372 L .56746 .33677 L p .679 .668 .832 r
F P
0 g
s
.56032 .33763 m .55992 .34651 L .55705 .34615 L p .68 .668 .831 r
F P
0 g
s
.55992 .34651 m .56032 .33763 L .56389 .3372 L p .68 .668 .832 r
F P
0 g
s
.55676 .33806 m .55705 .34615 L .55417 .34579 L p .68 .667 .831 r
F P
0 g
s
.55705 .34615 m .55676 .33806 L .56032 .33763 L p .68 .668 .831 r
F P
0 g
s
.55319 .3385 m .55417 .34579 L .55129 .34543 L p .68 .667 .83 r
F P
0 g
s
.55417 .34579 m .55319 .3385 L .55676 .33806 L p .68 .667 .831 r
F P
0 g
s
.55129 .34543 m .54963 .33893 L .55319 .3385 L p .68 .667 .83 r
F P
0 g
s
.37703 .42513 m .37845 .39622 L .3896 .39593 L p .79 .717 .784 r
F P
0 g
s
.42642 .37567 m .43359 .35816 L .44246 .36143 L p .782 .74 .815 r
F P
0 g
s
.43359 .35816 m .42642 .37567 L .41627 .37405 L p .782 .74 .815 r
F P
0 g
s
.72447 .4953 m .74722 .46903 L .75824 .4804 L p .54 .468 .724 r
F P
0 g
s
.75824 .4804 m .73394 .50805 L .72447 .4953 L p .54 .468 .724 r
F P
0 g
s
.54963 .33893 m .54585 .3326 L .54978 .33114 L p .678 .671 .837 r
F P
0 g
s
.54978 .33114 m .55319 .3385 L .54963 .33893 L p .678 .671 .837 r
F P
0 g
s
.4007 .39563 m .40607 .37243 L .41627 .37405 L p .787 .73 .8 r
F P
0 g
s
.40607 .37243 m .4007 .39563 L .3896 .39593 L p .787 .731 .8 r
F P
0 g
s
.45104 .34766 m .44246 .36143 L .43359 .35816 L p .773 .747 .829 r
F P
0 g
s
.55319 .3385 m .54978 .33114 L .5537 .32968 L p .678 .672 .837 r
F P
0 g
s
.5537 .32968 m .55676 .33806 L .55319 .3385 L p .678 .672 .837 r
F P
0 g
s
.55676 .33806 m .5537 .32968 L .55763 .32822 L p .678 .672 .838 r
F P
0 g
s
.55763 .32822 m .56032 .33763 L .55676 .33806 L p .678 .672 .838 r
F P
0 g
s
.56032 .33763 m .55763 .32822 L .56157 .32676 L p .678 .673 .838 r
F P
0 g
s
.56157 .32676 m .56389 .3372 L .56032 .33763 L p .678 .673 .838 r
F P
0 g
s
.56389 .3372 m .56157 .32676 L .56551 .3253 L p .678 .673 .838 r
F P
0 g
s
.56551 .3253 m .56746 .33677 L .56389 .3372 L p .678 .673 .838 r
F P
0 g
s
.73394 .50805 m .75824 .4804 L .76931 .49183 L p .54 .468 .724 r
F P
0 g
s
.76931 .49183 m .74345 .52086 L .73394 .50805 L p .54 .468 .724 r
F P
0 g
s
.54585 .3326 m .53998 .32686 L .54387 .32426 L p .678 .679 .844 r
F P
0 g
s
.54387 .32426 m .54978 .33114 L .54585 .3326 L p .678 .679 .844 r
F P
0 g
s
.41627 .37405 m .42468 .35486 L .43359 .35816 L p .782 .74 .815 r
F P
0 g
s
.42468 .35486 m .41627 .37405 L .40607 .37243 L p .782 .74 .815 r
F P
0 g
s
.3896 .39593 m .39583 .37079 L .40607 .37243 L p .787 .731 .8 r
F P
0 g
s
.39583 .37079 m .3896 .39593 L .37845 .39622 L p .787 .731 .8 r
F P
0 g
s
.54978 .33114 m .54387 .32426 L .54776 .32165 L p .678 .68 .844 r
F P
0 g
s
.54776 .32165 m .5537 .32968 L .54978 .33114 L p .678 .68 .844 r
F P
0 g
s
.46164 .33632 m .45104 .34766 L .44369 .34303 L p .762 .749 .841 r
F P
0 g
s
.43359 .35816 m .44369 .34303 L .45104 .34766 L p .773 .747 .829 r
F P
0 g
s
.44369 .34303 m .43359 .35816 L .42468 .35486 L p .773 .747 .829 r
F P
0 g
s
.5537 .32968 m .54776 .32165 L .55166 .31903 L p .678 .68 .845 r
F P
0 g
s
.55166 .31903 m .55763 .32822 L .5537 .32968 L p .678 .68 .845 r
F P
0 g
s
.72538 .44648 m .74047 .41932 L .75253 .42893 L p .508 .463 .741 r
F P
0 g
s
.75253 .42893 m .73627 .45772 L .72538 .44648 L p .508 .463 .741 r
F P
0 g
s
.53998 .32686 m .53211 .32214 L .53554 .31839 L p .681 .689 .851 r
F P
0 g
s
.53554 .31839 m .54387 .32426 L .53998 .32686 L p .681 .689 .851 r
F P
0 g
s
.55763 .32822 m .55166 .31903 L .55557 .31642 L p .678 .68 .845 r
F P
0 g
s
.55557 .31642 m .56157 .32676 L .55763 .32822 L p .678 .68 .845 r
F P
0 g
s
.74345 .52086 m .76931 .49183 L .78045 .50332 L p .54 .468 .724 r
F P
0 g
s
.78045 .50332 m .75301 .53373 L .74345 .52086 L p .54 .468 .724 r
F P
0 g
s
.56157 .32676 m .55557 .31642 L .55949 .31379 L p .678 .68 .845 r
F P
0 g
s
.55949 .31379 m .56551 .3253 L .56157 .32676 L p .678 .68 .845 r
F P
0 g
s
.40607 .37243 m .41571 .35155 L .42468 .35486 L p .782 .74 .815 r
F P
0 g
s
.41571 .35155 m .40607 .37243 L .39583 .37079 L p .782 .741 .815 r
F P
0 g
s
.54387 .32426 m .53554 .31839 L .53898 .31462 L p .681 .689 .851 r
F P
0 g
s
.53898 .31462 m .54776 .32165 L .54387 .32426 L p .681 .689 .851 r
F P
0 g
s
.37845 .39622 m .38553 .36915 L .39583 .37079 L p .787 .731 .8 r
F P
0 g
s
.47362 .32761 m .46164 .33632 L .45603 .33069 L p .749 .748 .851 r
F P
0 g
s
.44369 .34303 m .45603 .33069 L .46164 .33632 L p .762 .749 .841 r
F P
0 g
s
.53211 .32214 m .52244 .31888 L .52501 .31408 L p .687 .701 .857 r
F P
0 g
s
.52501 .31408 m .53554 .31839 L .53211 .32214 L p .687 .701 .857 r
F P
0 g
s
.42468 .35486 m .43631 .33838 L .44369 .34303 L p .773 .747 .829 r
F P
0 g
s
.43631 .33838 m .42468 .35486 L .41571 .35155 L p .773 .747 .829 r
F P
0 g
s
.54776 .32165 m .53898 .31462 L .54243 .31084 L p .681 .69 .851 r
F P
0 g
s
.54243 .31084 m .55166 .31903 L .54776 .32165 L p .681 .69 .851 r
F P
0 g
s
.45603 .33069 m .44369 .34303 L .43631 .33838 L p .762 .749 .841 r
F P
0 g
s
.73627 .45772 m .75253 .42893 L .76467 .43861 L p .508 .463 .741 r
F P
0 g
s
.76467 .43861 m .74722 .46903 L .73627 .45772 L p .508 .463 .741 r
F P
0 g
s
.52244 .31888 m .51128 .31748 L .51262 .31184 L p .696 .713 .861 r
F P
0 g
s
.51262 .31184 m .52501 .31408 L .52244 .31888 L p .696 .713 .861 r
F P
0 g
s
.48631 .32162 m .47362 .32761 L .46986 .32139 L p .735 .743 .858 r
F P
0 g
s
.45603 .33069 m .46986 .32139 L .47362 .32761 L p .749 .748 .851 r
F P
0 g
s
.53554 .31839 m .52501 .31408 L .52759 .30926 L p .687 .701 .857 r
F P
0 g
s
.52759 .30926 m .53898 .31462 L .53554 .31839 L p .687 .701 .857 r
F P
0 g
s
.55166 .31903 m .54243 .31084 L .5459 .30706 L p .681 .69 .851 r
F P
0 g
s
.5459 .30706 m .55557 .31642 L .55166 .31903 L p .681 .69 .851 r
F P
0 g
s
.39583 .37079 m .4067 .34823 L .41571 .35155 L p .782 .741 .815 r
F P
0 g
s
.4067 .34823 m .39583 .37079 L .38553 .36915 L p .782 .741 .815 r
F P
0 g
s
.51128 .31748 m .49907 .31831 L .49888 .3121 L p .708 .724 .863 r
F P
0 g
s
.49888 .3121 m .51262 .31184 L .51128 .31748 L p .708 .724 .863 r
F P
0 g
s
.49907 .31831 m .48631 .32162 L .4844 .31522 L p .721 .735 .862 r
F P
0 g
s
.46986 .32139 m .4844 .31522 L .48631 .32162 L p .735 .743 .858 r
F P
0 g
s
.4844 .31522 m .49888 .3121 L .49907 .31831 L p .721 .735 .862 r
F P
0 g
s
.55557 .31642 m .5459 .30706 L .54937 .30326 L p .681 .69 .852 r
F P
0 g
s
.54937 .30326 m .55949 .31379 L .55557 .31642 L p .681 .69 .852 r
F P
0 g
s
.46986 .32139 m .45603 .33069 L .45039 .32504 L p .749 .748 .851 r
F P
0 g
s
.43631 .33838 m .45039 .32504 L .45603 .33069 L p .762 .749 .841 r
F P
0 g
s
.53898 .31462 m .52759 .30926 L .53018 .30442 L p .687 .701 .857 r
F P
0 g
s
.53018 .30442 m .54243 .31084 L .53898 .31462 L p .687 .701 .857 r
F P
0 g
s
.52501 .31408 m .51262 .31184 L .51397 .30617 L p .696 .713 .861 r
F P
0 g
s
.51397 .30617 m .52759 .30926 L .52501 .31408 L p .696 .713 .861 r
F P
0 g
s
.41571 .35155 m .42888 .33369 L .43631 .33838 L p .773 .747 .829 r
F P
0 g
s
.42888 .33369 m .41571 .35155 L .4067 .34823 L p .773 .747 .829 r
F P
0 g
s
.45039 .32504 m .43631 .33838 L .42888 .33369 L p .762 .749 .841 r
F P
0 g
s
.74722 .46903 m .76467 .43861 L .77689 .44836 L p .508 .463 .741 r
F P
0 g
s
.77689 .44836 m .75824 .4804 L .74722 .46903 L p .508 .463 .741 r
F P
0 g
s
.4844 .31522 m .46986 .32139 L .46608 .31514 L p .735 .743 .858 r
F P
0 g
s
.45039 .32504 m .46608 .31514 L .46986 .32139 L p .749 .748 .851 r
F P
0 g
s
.54243 .31084 m .53018 .30442 L .53278 .29957 L p .688 .701 .857 r
F P
0 g
s
.53278 .29957 m .5459 .30706 L .54243 .31084 L p .688 .701 .857 r
F P
0 g
s
.51262 .31184 m .49888 .3121 L .4987 .30587 L p .708 .724 .863 r
F P
0 g
s
.4987 .30587 m .51397 .30617 L .51262 .31184 L p .708 .724 .863 r
F P
0 g
s
.38553 .36915 m .39764 .34488 L .4067 .34823 L p .782 .741 .815 r
F P
0 g
s
.49888 .3121 m .4844 .31522 L .48247 .30878 L p .721 .735 .862 r
F P
0 g
s
.46608 .31514 m .48247 .30878 L .4844 .31522 L p .735 .743 .858 r
F P
0 g
s
.52759 .30926 m .51397 .30617 L .51533 .30047 L p .696 .713 .861 r
F P
0 g
s
.51533 .30047 m .53018 .30442 L .52759 .30926 L p .696 .713 .861 r
F P
0 g
s
.48247 .30878 m .4987 .30587 L .49888 .3121 L p .721 .735 .862 r
F P
0 g
s
.46608 .31514 m .45039 .32504 L .44472 .31935 L p .749 .748 .851 r
F P
0 g
s
.42888 .33369 m .44472 .31935 L .45039 .32504 L p .762 .749 .841 r
F P
0 g
s
.5459 .30706 m .53278 .29957 L .53539 .29469 L p .688 .701 .857 r
F P
0 g
s
.53539 .29469 m .54937 .30326 L .5459 .30706 L p .688 .701 .857 r
F P
0 g
s
.4067 .34823 m .42141 .32898 L .42888 .33369 L p .773 .747 .829 r
F P
0 g
s
.42141 .32898 m .4067 .34823 L .39764 .34488 L p .773 .747 .829 r
F P
0 g
s
.44472 .31935 m .42888 .33369 L .42141 .32898 L p .762 .749 .841 r
F P
0 g
s
.51397 .30617 m .4987 .30587 L .49852 .29961 L p .708 .724 .863 r
F P
0 g
s
.49852 .29961 m .51533 .30047 L .51397 .30617 L p .708 .724 .863 r
F P
0 g
s
.75824 .4804 m .77689 .44836 L .78919 .45817 L p .508 .463 .741 r
F P
0 g
s
.78919 .45817 m .76931 .49183 L .75824 .4804 L p .508 .463 .741 r
F P
0 g
s
.53018 .30442 m .51533 .30047 L .51669 .29475 L p .696 .713 .861 r
F P
0 g
s
.51669 .29475 m .53278 .29957 L .53018 .30442 L p .696 .713 .861 r
F P
0 g
s
.48247 .30878 m .46608 .31514 L .46227 .30885 L p .735 .743 .858 r
F P
0 g
s
.44472 .31935 m .46227 .30885 L .46608 .31514 L p .749 .748 .851 r
F P
0 g
s
.4987 .30587 m .48247 .30878 L .48053 .30231 L p .721 .735 .862 r
F P
0 g
s
.48053 .30231 m .49852 .29961 L .4987 .30587 L p .721 .735 .862 r
F P
0 g
s
.46227 .30885 m .48053 .30231 L .48247 .30878 L p .735 .743 .858 r
F P
0 g
s
.74047 .41932 m .75002 .38913 L .76286 .39695 L p .476 .461 .762 r
F P
0 g
s
.76286 .39695 m .75253 .42893 L .74047 .41932 L p .476 .461 .762 r
F P
0 g
s
.46227 .30885 m .44472 .31935 L .43902 .31363 L p .748 .748 .851 r
F P
0 g
s
.42141 .32898 m .43902 .31363 L .44472 .31935 L p .762 .749 .841 r
F P
0 g
s
.53278 .29957 m .51669 .29475 L .51806 .28901 L p .697 .713 .861 r
F P
0 g
s
.51806 .28901 m .53539 .29469 L .53278 .29957 L p .697 .713 .861 r
F P
0 g
s
.39764 .34488 m .4139 .32425 L .42141 .32898 L p .773 .747 .829 r
F P
0 g
s
.51533 .30047 m .49852 .29961 L .49833 .29332 L p .708 .724 .863 r
F P
0 g
s
.49833 .29332 m .51669 .29475 L .51533 .30047 L p .708 .724 .863 r
F P
0 g
s
.43902 .31363 m .42141 .32898 L .4139 .32425 L p .761 .749 .841 r
F P
0 g
s
.76931 .49183 m .78919 .45817 L .80158 .46805 L p .509 .463 .741 r
F P
0 g
s
.80158 .46805 m .78045 .50332 L .76931 .49183 L p .509 .463 .741 r
F P
0 g
s
.48053 .30231 m .46227 .30885 L .45844 .30253 L p .735 .743 .858 r
F P
0 g
s
.43902 .31363 m .45844 .30253 L .46227 .30885 L p .748 .748 .851 r
F P
0 g
s
.49852 .29961 m .48053 .30231 L .47858 .2958 L p .721 .735 .862 r
F P
0 g
s
.47858 .2958 m .49833 .29332 L .49852 .29961 L p .721 .735 .862 r
F P
0 g
s
.45844 .30253 m .47858 .2958 L .48053 .30231 L p .735 .743 .858 r
F P
0 g
s
.51669 .29475 m .49833 .29332 L .49815 .28699 L p .708 .724 .863 r
F P
0 g
s
.49815 .28699 m .51806 .28901 L .51669 .29475 L p .708 .724 .863 r
F P
0 g
s
.39052 .7229 m .44191 .74718 L .44002 .7531 L p .78 .49 .497 r
F P
0 g
s
.44191 .74718 m .4968 .76382 L .49551 .76939 L p .736 .443 .484 r
F P
0 g
s
.49551 .76939 m .44002 .7531 L .44191 .74718 L p .736 .443 .484 r
F P
0 g
s
.4968 .76382 m .5538 .77242 L .55301 .77757 L p .689 .395 .472 r
F P
0 g
s
.55301 .77757 m .49551 .76939 L .4968 .76382 L p .689 .395 .472 r
F P
0 g
s
.75253 .42893 m .76286 .39695 L .7758 .40483 L p .476 .461 .762 r
F P
0 g
s
.7758 .40483 m .76467 .43861 L .75253 .42893 L p .476 .461 .762 r
F P
0 g
s
.45844 .30253 m .43902 .31363 L .43328 .30788 L p .748 .748 .851 r
F P
0 g
s
.4139 .32425 m .43328 .30788 L .43902 .31363 L p .761 .749 .841 r
F P
0 g
s
.49833 .29332 m .47858 .2958 L .47662 .28925 L p .721 .735 .862 r
F P
0 g
s
.47662 .28925 m .49815 .28699 L .49833 .29332 L p .721 .735 .862 r
F P
0 g
s
.47858 .2958 m .45844 .30253 L .45459 .29617 L p .735 .743 .858 r
F P
0 g
s
.43328 .30788 m .45459 .29617 L .45844 .30253 L p .748 .748 .851 r
F P
0 g
s
.45459 .29617 m .47662 .28925 L .47858 .2958 L p .735 .743 .858 r
F P
0 g
s
.34396 .69154 m .39052 .7229 L .38795 .72906 L p .821 .54 .513 r
F P
0 g
s
.44002 .7531 m .38795 .72906 L .39052 .7229 L p .78 .49 .497 r
F P
0 g
s
.5538 .77242 m .61146 .77276 L .61108 .77742 L p .638 .348 .463 r
F P
0 g
s
.61108 .77742 m .55301 .77757 L .5538 .77242 L p .638 .348 .463 r
F P
0 g
s
.44002 .7531 m .49551 .76939 L .49421 .775 L p .736 .443 .484 r
F P
0 g
s
.49551 .76939 m .55301 .77757 L .55222 .78274 L p .689 .395 .472 r
F P
0 g
s
.55222 .78274 m .49421 .775 L .49551 .76939 L p .689 .395 .472 r
F P
0 g
s
.38795 .72906 m .44002 .7531 L .43812 .75906 L p .78 .49 .497 r
F P
0 g
s
.49421 .775 m .43812 .75906 L .44002 .7531 L p .736 .443 .484 r
F P
0 g
s
.76467 .43861 m .7758 .40483 L .78884 .41278 L p .476 .461 .762 r
F P
0 g
s
.78884 .41278 m .77689 .44836 L .76467 .43861 L p .476 .461 .762 r
F P
0 g
s
.55301 .77757 m .61108 .77742 L .61071 .78209 L p .638 .348 .463 r
F P
0 g
s
.61071 .78209 m .55222 .78274 L .55301 .77757 L p .638 .348 .463 r
F P
0 g
s
.43812 .75906 m .49421 .775 L .49291 .78063 L p .736 .443 .484 r
F P
0 g
s
.49421 .775 m .55222 .78274 L .55143 .78793 L p .689 .395 .472 r
F P
0 g
s
.55143 .78793 m .49291 .78063 L .49421 .775 L p .689 .395 .472 r
F P
0 g
s
.38795 .72906 m .34066 .69781 L .34396 .69154 L p .821 .54 .513 r
F P
0 g
s
.34066 .69781 m .38795 .72906 L .38537 .73526 L p .821 .54 .513 r
F P
0 g
s
.43812 .75906 m .38537 .73526 L .38795 .72906 L p .78 .49 .497 r
F P
0 g
s
.38537 .73526 m .43812 .75906 L .43621 .76505 L p .78 .49 .497 r
F P
0 g
s
.49291 .78063 m .43621 .76505 L .43812 .75906 L p .736 .443 .484 r
F P
0 g
s
.30348 .65384 m .34396 .69154 L .34066 .69781 L p .859 .591 .532 r
F P
0 g
s
.75002 .38913 m .75323 .35679 L .76642 .3627 L p .444 .466 .787 r
F P
0 g
s
.76642 .3627 m .76286 .39695 L .75002 .38913 L p .444 .466 .787 r
F P
0 g
s
.77689 .44836 m .78884 .41278 L .80198 .42079 L p .476 .461 .762 r
F P
0 g
s
.80198 .42079 m .78919 .45817 L .77689 .44836 L p .476 .461 .762 r
F P
0 g
s
.55222 .78274 m .61071 .78209 L .61033 .78678 L p .638 .348 .463 r
F P
0 g
s
.61033 .78678 m .55143 .78793 L .55222 .78274 L p .638 .348 .463 r
F P
0 g
s
.61146 .77276 m .66831 .76478 L .66825 .7689 L p .581 .3 .455 r
F P
0 g
s
.66825 .7689 m .61108 .77742 L .61146 .77276 L p .581 .3 .455 r
F P
0 g
s
.43621 .76505 m .49291 .78063 L .4916 .7863 L p .736 .443 .484 r
F P
0 g
s
.49291 .78063 m .55143 .78793 L .55064 .79315 L p .689 .395 .472 r
F P
0 g
s
.55064 .79315 m .4916 .7863 L .49291 .78063 L p .689 .395 .472 r
F P
0 g
s
.38537 .73526 m .33733 .70412 L .34066 .69781 L p .821 .54 .513 r
F P
0 g
s
.33733 .70412 m .38537 .73526 L .38277 .7415 L p .821 .54 .513 r
F P
0 g
s
.43621 .76505 m .38277 .7415 L .38537 .73526 L p .78 .49 .497 r
F P
0 g
s
.38277 .7415 m .43621 .76505 L .43429 .77107 L p .78 .49 .497 r
F P
0 g
s
.4916 .7863 m .43429 .77107 L .43621 .76505 L p .736 .443 .484 r
F P
0 g
s
.61108 .77742 m .66825 .7689 L .66818 .77303 L p .581 .3 .455 r
F P
0 g
s
.66818 .77303 m .61071 .78209 L .61108 .77742 L p .581 .3 .455 r
F P
0 g
s
.55143 .78793 m .61033 .78678 L .60995 .79149 L p .638 .348 .463 r
F P
0 g
s
.60995 .79149 m .55064 .79315 L .55143 .78793 L p .638 .348 .463 r
F P
0 g
s
.34066 .69781 m .29941 .66006 L .30348 .65384 L p .859 .591 .532 r
F P
0 g
s
.29941 .66006 m .34066 .69781 L .33733 .70412 L p .859 .591 .532 r
F P
0 g
s
.43429 .77107 m .4916 .7863 L .49028 .79199 L p .736 .443 .484 r
F P
0 g
s
.4916 .7863 m .55064 .79315 L .54984 .79839 L p .689 .395 .472 r
F P
0 g
s
.54984 .79839 m .49028 .79199 L .4916 .7863 L p .689 .395 .472 r
F P
0 g
s
.78919 .45817 m .80198 .42079 L .81524 .42886 L p .476 .461 .762 r
F P
0 g
s
.81524 .42886 m .80158 .46805 L .78919 .45817 L p .476 .461 .762 r
F P
0 g
s
.76286 .39695 m .76642 .3627 L .77973 .36868 L p .444 .466 .787 r
F P
0 g
s
.77973 .36868 m .7758 .40483 L .76286 .39695 L p .444 .466 .787 r
F P
0 g
s
.43429 .77107 m .38016 .74777 L .38277 .7415 L p .78 .49 .497 r
F P
0 g
s
.38016 .74777 m .43429 .77107 L .43236 .77713 L p .78 .49 .497 r
F P
0 g
s
.49028 .79199 m .43236 .77713 L .43429 .77107 L p .736 .443 .484 r
F P
0 g
s
.38277 .7415 m .33399 .71047 L .33733 .70412 L p .821 .54 .513 r
F P
0 g
s
.33399 .71047 m .38277 .7415 L .38016 .74777 L p .821 .54 .513 r
F P
0 g
s
.61071 .78209 m .66818 .77303 L .66812 .77717 L p .581 .3 .455 r
F P
0 g
s
.66812 .77717 m .61033 .78678 L .61071 .78209 L p .581 .3 .455 r
F P
0 g
s
.55064 .79315 m .60995 .79149 L .60957 .79622 L p .638 .348 .463 r
F P
0 g
s
.60957 .79622 m .54984 .79839 L .55064 .79315 L p .638 .348 .463 r
F P
0 g
s
.27023 .61072 m .30348 .65384 L .29941 .66006 L p .893 .645 .555 r
F P
0 g
s
.33733 .70412 m .29531 .66633 L .29941 .66006 L p .859 .591 .532 r
F P
0 g
s
.29531 .66633 m .33733 .70412 L .33399 .71047 L p .859 .591 .532 r
F P
0 g
s
.61033 .78678 m .66812 .77717 L .66806 .78133 L p .581 .3 .455 r
F P
0 g
s
.66806 .78133 m .60995 .79149 L .61033 .78678 L p .581 .3 .455 r
F P
0 g
s
.43236 .77713 m .37754 .75409 L .38016 .74777 L p .78 .49 .497 r
F P
0 g
s
.38016 .74777 m .33062 .71687 L .33399 .71047 L p .821 .54 .513 r
F P
0 g
s
.33062 .71687 m .38016 .74777 L .37754 .75409 L p .821 .54 .513 r
F P
0 g
s
.66831 .76478 m .72283 .74858 L .72299 .75213 L p .518 .25 .449 r
F P
0 g
s
.72299 .75213 m .66825 .7689 L .66831 .76478 L p .518 .25 .449 r
F P
0 g
s
.7758 .40483 m .77973 .36868 L .79317 .3747 L p .444 .466 .787 r
F P
0 g
s
.79317 .3747 m .78884 .41278 L .7758 .40483 L p .444 .466 .787 r
F P
0 g
s
.60995 .79149 m .66806 .78133 L .668 .78551 L p .581 .3 .455 r
F P
0 g
s
.668 .78551 m .60957 .79622 L .60995 .79149 L p .581 .3 .455 r
F P
0 g
s
.29941 .66006 m .26539 .61671 L .27023 .61072 L p .893 .645 .555 r
F P
0 g
s
.26539 .61671 m .29941 .66006 L .29531 .66633 L p .893 .645 .555 r
F P
0 g
s
.33399 .71047 m .29117 .67264 L .29531 .66633 L p .859 .591 .532 r
F P
0 g
s
.29117 .67264 m .33399 .71047 L .33062 .71687 L p .859 .591 .532 r
F P
0 g
s
.66825 .7689 m .72299 .75213 L .72315 .7557 L p .518 .25 .449 r
F P
0 g
s
.72315 .7557 m .66818 .77303 L .66825 .7689 L p .518 .25 .449 r
F P
0 g
s
.37754 .75409 m .32723 .7233 L .33062 .71687 L p .821 .54 .513 r
F P
0 g
s
.75323 .35679 m .74943 .32331 L .76254 .32729 L p .412 .476 .817 r
F P
0 g
s
.76254 .32729 m .76642 .3627 L .75323 .35679 L p .412 .476 .817 r
F P
0 g
s
.66818 .77303 m .72315 .7557 L .72331 .75927 L p .518 .25 .449 r
F P
0 g
s
.72331 .75927 m .66812 .77717 L .66818 .77303 L p .518 .25 .449 r
F P
0 g
s
.78884 .41278 m .79317 .3747 L .80673 .38079 L p .444 .466 .787 r
F P
0 g
s
.80673 .38079 m .80198 .42079 L .78884 .41278 L p .444 .466 .787 r
F P
0 g
s
.29531 .66633 m .26051 .62276 L .26539 .61671 L p .893 .645 .555 r
F P
0 g
s
.26051 .62276 m .29531 .66633 L .29117 .67264 L p .893 .645 .555 r
F P
0 g
s
.33062 .71687 m .28701 .67899 L .29117 .67264 L p .859 .591 .532 r
F P
0 g
s
.28701 .67899 m .33062 .71687 L .32723 .7233 L p .859 .591 .532 r
F P
0 g
s
.24521 .56326 m .27023 .61072 L .26539 .61671 L p .923 .7 .582 r
F P
0 g
s
.66812 .77717 m .72331 .75927 L .72347 .76286 L p .518 .25 .449 r
F P
0 g
s
.72347 .76286 m .66806 .78133 L .66812 .77717 L p .518 .25 .449 r
F P
0 g
s
.66806 .78133 m .72347 .76286 L .72363 .76646 L p .518 .25 .449 r
F P
0 g
s
.72363 .76646 m .668 .78551 L .66806 .78133 L p .518 .25 .449 r
F P
0 g
s
.76642 .3627 m .76254 .32729 L .77578 .33131 L p .412 .476 .817 r
F P
0 g
s
.77578 .33131 m .77973 .36868 L .76642 .3627 L p .412 .476 .817 r
F P
0 g
s
.32723 .7233 m .28282 .68539 L .28701 .67899 L p .859 .591 .532 r
F P
0 g
s
.29117 .67264 m .25559 .62885 L .26051 .62276 L p .893 .645 .555 r
F P
0 g
s
.25559 .62885 m .29117 .67264 L .28701 .67899 L p .893 .645 .555 r
F P
0 g
s
.80198 .42079 m .80673 .38079 L .82043 .38693 L p .444 .466 .787 r
F P
0 g
s
.82043 .38693 m .81524 .42886 L .80198 .42079 L p .444 .466 .787 r
F P
0 g
s
.26539 .61671 m .23964 .56884 L .24521 .56326 L p .923 .7 .582 r
F P
0 g
s
.23964 .56884 m .26539 .61671 L .26051 .62276 L p .923 .7 .582 r
F P
0 g
s
.72283 .74858 m .77349 .72445 L .77379 .72742 L p .445 .197 .442 r
F P
0 g
s
.77379 .72742 m .72299 .75213 L .72283 .74858 L p .445 .197 .442 r
F P
0 g
s
.72299 .75213 m .77379 .72742 L .77408 .73041 L p .445 .197 .442 r
F P
0 g
s
.77408 .73041 m .72315 .7557 L .72299 .75213 L p .445 .197 .442 r
F P
0 g
s
.28701 .67899 m .25063 .63499 L .25559 .62885 L p .893 .645 .555 r
F P
0 g
s
.25063 .63499 m .28701 .67899 L .28282 .68539 L p .893 .645 .555 r
F P
0 g
s
.77973 .36868 m .77578 .33131 L .78916 .33538 L p .412 .476 .817 r
F P
0 g
s
.78916 .33538 m .79317 .3747 L .77973 .36868 L p .412 .476 .817 r
F P
0 g
s
.26051 .62276 m .23402 .57447 L .23964 .56884 L p .923 .7 .582 r
F P
0 g
s
.23402 .57447 m .26051 .62276 L .25559 .62885 L p .923 .7 .582 r
F P
0 g
s
.72315 .7557 m .77408 .73041 L .77438 .7334 L p .445 .197 .442 r
F P
0 g
s
.77438 .7334 m .72331 .75927 L .72315 .7557 L p .445 .197 .442 r
F P
0 g
s
.22922 .51274 m .24521 .56326 L .23964 .56884 L p .947 .758 .614 r
F P
0 g
s
.74943 .32331 m .73818 .28988 L .75075 .29197 L p .381 .495 .849 r
F P
0 g
s
.75075 .29197 m .76254 .32729 L .74943 .32331 L p .381 .495 .849 r
F P
0 g
s
.72331 .75927 m .77438 .7334 L .77467 .7364 L p .445 .197 .442 r
F P
0 g
s
.77467 .7364 m .72347 .76286 L .72331 .75927 L p .445 .197 .442 r
F P
0 g
s
.28282 .68539 m .24563 .64117 L .25063 .63499 L p .893 .645 .555 r
F P
0 g
s
.72347 .76286 m .77467 .7364 L .77497 .73942 L p .445 .197 .442 r
F P
0 g
s
.77497 .73942 m .72363 .76646 L .72347 .76286 L p .445 .197 .442 r
F P
0 g
s
.25559 .62885 m .22835 .58015 L .23402 .57447 L p .923 .7 .582 r
F P
0 g
s
.22835 .58015 m .25559 .62885 L .25063 .63499 L p .923 .7 .582 r
F P
0 g
s
.79317 .3747 m .78916 .33538 L .80268 .33949 L p .412 .476 .817 r
F P
0 g
s
.80268 .33949 m .80673 .38079 L .79317 .3747 L p .412 .476 .817 r
F P
0 g
s
.23964 .56884 m .22301 .51772 L .22922 .51274 L p .947 .758 .614 r
F P
0 g
s
.22301 .51772 m .23964 .56884 L .23402 .57447 L p .947 .758 .614 r
F P
0 g
s
.76254 .32729 m .75075 .29197 L .76346 .29409 L p .381 .494 .849 r
F P
0 g
s
.76346 .29409 m .77578 .33131 L .76254 .32729 L p .381 .494 .849 r
F P
0 g
s
.25063 .63499 m .22264 .58588 L .22835 .58015 L p .923 .7 .582 r
F P
0 g
s
.22264 .58588 m .25063 .63499 L .24563 .64117 L p .923 .7 .582 r
F P
0 g
s
.77349 .72445 m .81877 .69286 L .81912 .69527 L p .359 .137 .436 r
F P
0 g
s
.81912 .69527 m .77379 .72742 L .77349 .72445 L p .359 .137 .436 r
F P
0 g
s
.80673 .38079 m .80268 .33949 L .81635 .34364 L p .412 .476 .817 r
F P
0 g
s
.81635 .34364 m .82043 .38693 L .80673 .38079 L p .412 .476 .817 r
F P
0 g
s
.23402 .57447 m .21676 .52274 L .22301 .51772 L p .947 .758 .614 r
F P
0 g
s
.21676 .52274 m .23402 .57447 L .22835 .58015 L p .947 .758 .614 r
F P
0 g
s
.22282 .4606 m .22922 .51274 L .22301 .51772 L p .963 .815 .65 r
F P
0 g
s
.77379 .72742 m .81912 .69527 L .81947 .69767 L p .359 .137 .436 r
F P
0 g
s
.81947 .69767 m .77408 .73041 L .77379 .72742 L p .359 .137 .436 r
F P
0 g
s
.77408 .73041 m .81947 .69767 L .81982 .70009 L p .359 .137 .436 r
F P
0 g
s
.81982 .70009 m .77438 .7334 L .77408 .73041 L p .359 .137 .436 r
F P
0 g
s
.77578 .33131 m .76346 .29409 L .77632 .29623 L p .381 .494 .849 r
F P
0 g
s
.77632 .29623 m .78916 .33538 L .77578 .33131 L p .381 .494 .849 r
F P
0 g
s
.24563 .64117 m .21688 .59166 L .22264 .58588 L p .923 .7 .582 r
F P
0 g
s
.73818 .28988 m .71928 .25781 L .73087 .25814 L p .354 .521 .884 r
F P
0 g
s
.73087 .25814 m .75075 .29197 L .73818 .28988 L p .354 .521 .884 r
F P
0 g
s
.77438 .7334 m .81982 .70009 L .82017 .70251 L p .359 .137 .436 r
F P
0 g
s
.82017 .70251 m .77467 .7364 L .77438 .7334 L p .359 .137 .436 r
F P
0 g
s
.22835 .58015 m .21044 .5278 L .21676 .52274 L p .947 .758 .614 r
F P
0 g
s
.21044 .5278 m .22835 .58015 L .22264 .58588 L p .947 .758 .614 r
F P
0 g
s
.77467 .7364 m .82017 .70251 L .82053 .70493 L p .359 .137 .436 r
F P
0 g
s
.82053 .70493 m .77497 .73942 L .77467 .7364 L p .359 .137 .436 r
F P
0 g
s
.22301 .51772 m .21614 .46477 L .22282 .4606 L p .963 .815 .65 r
F P
0 g
s
.21614 .46477 m .22301 .51772 L .21676 .52274 L p .963 .815 .65 r
F P
0 g
s
.78916 .33538 m .77632 .29623 L .78934 .2984 L p .381 .494 .849 r
F P
0 g
s
.78934 .2984 m .80268 .33949 L .78916 .33538 L p .381 .494 .849 r
F P
0 g
s
.22264 .58588 m .20406 .53292 L .21044 .5278 L p .947 .758 .614 r
F P
0 g
s
.20406 .53292 m .22264 .58588 L .21688 .59166 L p .947 .758 .614 r
F P
0 g
s
.75075 .29197 m .73087 .25814 L .74261 .25848 L p .354 .521 .884 r
F P
0 g
s
.74261 .25848 m .76346 .29409 L .75075 .29197 L p .354 .521 .884 r
F P
0 g
s
.21676 .52274 m .20939 .46899 L .21614 .46477 L p .963 .815 .65 r
F P
0 g
s
.20939 .46899 m .21676 .52274 L .21044 .5278 L p .963 .816 .65 r
F P
0 g
s
.2263 .40839 m .22282 .4606 L .21614 .46477 L p .967 .871 .69 r
F P
0 g
s
.21688 .59166 m .19762 .53808 L .20406 .53292 L p .947 .758 .614 r
F P
0 g
s
.81877 .69286 m .85714 .65453 L .85748 .65636 L p .254 .07 .427 r
F P
0 g
s
.85748 .65636 m .81912 .69527 L .81877 .69286 L p .254 .07 .427 r
F P
0 g
s
.80268 .33949 m .78934 .2984 L .80252 .3006 L p .381 .494 .849 r
F P
0 g
s
.80252 .3006 m .81635 .34364 L .80268 .33949 L p .381 .494 .849 r
F P
0 g
s
.71928 .25781 m .69282 .2285 L .70305 .22728 L p .334 .558 .918 r
F P
0 g
s
.70305 .22728 m .73087 .25814 L .71928 .25781 L p .334 .558 .918 r
F P
0 g
s
.81912 .69527 m .85748 .65636 L .85782 .65821 L p .254 .07 .427 r
F P
0 g
s
.85782 .65821 m .81947 .69767 L .81912 .69527 L p .254 .07 .427 r
F P
0 g
s
.21044 .5278 m .20258 .47324 L .20939 .46899 L p .963 .816 .65 r
F P
0 g
s
.20258 .47324 m .21044 .5278 L .20406 .53292 L p .963 .816 .65 r
F P
0 g
s
.76346 .29409 m .74261 .25848 L .75451 .25882 L p .354 .521 .884 r
F P
0 g
s
.75451 .25882 m .77632 .29623 L .76346 .29409 L p .354 .521 .884 r
F P
0 g
s
.21614 .46477 m .21935 .41159 L .2263 .40839 L p .967 .871 .69 r
F P
0 g
s
.21935 .41159 m .21614 .46477 L .20939 .46899 L p .967 .871 .691 r
F P
0 g
s
.81947 .69767 m .85782 .65821 L .85817 .66005 L p .254 .07 .427 r
F P
0 g
s
.85817 .66005 m .81982 .70009 L .81947 .69767 L p .254 .07 .427 r
F P
0 g
s
.81982 .70009 m .85817 .66005 L .85851 .6619 L p .254 .07 .427 r
F P
0 g
s
.85851 .6619 m .82017 .70251 L .81982 .70009 L p .254 .07 .427 r
F P
0 g
s
.82017 .70251 m .85851 .6619 L .85885 .66375 L p .254 .07 .427 r
F P
0 g
s
.85885 .66375 m .82053 .70493 L .82017 .70251 L p .254 .07 .427 r
F P
0 g
s
.20406 .53292 m .19569 .47754 L .20258 .47324 L p .963 .816 .65 r
F P
0 g
s
.19569 .47754 m .20406 .53292 L .19762 .53808 L p .963 .816 .65 r
F P
0 g
s
.23958 .35775 m .2263 .40839 L .21935 .41159 L p .955 .922 .735 r
F P
0 g
s
.73087 .25814 m .70305 .22728 L .71342 .22604 L p .334 .558 .918 r
F P
0 g
s
.71342 .22604 m .74261 .25848 L .73087 .25814 L p .334 .558 .918 r
F P
0 g
s
.20939 .46899 m .21233 .41482 L .21935 .41159 L p .967 .871 .691 r
F P
0 g
s
.21233 .41482 m .20939 .46899 L .20258 .47324 L p .967 .871 .691 r
F P
0 g
s
.77632 .29623 m .75451 .25882 L .76656 .25917 L p .354 .521 .884 r
F P
0 g
s
.76656 .25917 m .78934 .2984 L .77632 .29623 L p .354 .521 .884 r
F P
0 g
s
.69282 .2285 m .65926 .20337 L .66778 .20087 L p .323 .604 .948 r
F P
0 g
s
.66778 .20087 m .70305 .22728 L .69282 .2285 L p .323 .604 .948 r
F P
0 g
s
.19762 .53808 m .18873 .48189 L .19569 .47754 L p .963 .816 .65 r
F P
0 g
s
.21935 .41159 m .23264 .35982 L .23958 .35775 L p .955 .922 .735 r
F P
0 g
s
.23264 .35982 m .21935 .41159 L .21233 .41482 L p .955 .922 .735 r
F P
0 g
s
.20258 .47324 m .20523 .41808 L .21233 .41482 L p .967 .871 .691 r
F P
0 g
s
.20523 .41808 m .20258 .47324 L .19569 .47754 L p .967 .871 .691 r
F P
0 g
s
.74261 .25848 m .71342 .22604 L .72393 .22478 L p .334 .558 .918 r
F P
0 g
s
.72393 .22478 m .75451 .25882 L .74261 .25848 L p .334 .558 .918 r
F P
0 g
s
.78934 .2984 m .76656 .25917 L .77878 .25952 L p .354 .521 .884 r
F P
0 g
s
.77878 .25952 m .80252 .3006 L .78934 .2984 L p .354 .521 .884 r
F P
0 g
s
.26224 .31031 m .23958 .35775 L .23264 .35982 L p .924 .962 .781 r
F P
0 g
s
.85714 .65453 m .88715 .61037 L .88743 .61166 L p .123 0 .414 r
F P
0 g
s
.88743 .61166 m .85748 .65636 L .85714 .65453 L p .123 0 .414 r
F P
0 g
s
.85748 .65636 m .88743 .61166 L .88771 .61295 L p .123 0 .414 r
F P
0 g
s
.88771 .61295 m .85782 .65821 L .85748 .65636 L p .123 0 .414 r
F P
0 g
s
.19569 .47754 m .19805 .42138 L .20523 .41808 L p .967 .871 .691 r
F P
0 g
s
.19805 .42138 m .19569 .47754 L .18873 .48189 L p .967 .871 .691 r
F P
0 g
s
.70305 .22728 m .66778 .20087 L .67643 .19833 L p .323 .604 .948 r
F P
0 g
s
.67643 .19833 m .71342 .22604 L .70305 .22728 L p .323 .604 .948 r
F P
0 g
s
.21233 .41482 m .22561 .36192 L .23264 .35982 L p .955 .922 .735 r
F P
0 g
s
.22561 .36192 m .21233 .41482 L .20523 .41808 L p .955 .922 .735 r
F P
0 g
s
.85782 .65821 m .88771 .61295 L .88799 .61424 L p .123 0 .414 r
F P
0 g
s
.88799 .61424 m .85817 .66005 L .85782 .65821 L p .123 0 .414 r
F P
0 g
s
.85817 .66005 m .88799 .61424 L .88828 .61554 L p .123 0 .414 r
F P
0 g
s
.88828 .61554 m .85851 .6619 L .85817 .66005 L p .123 0 .414 r
F P
0 g
s
.65926 .20337 m .6194 .18377 L .62596 .18032 L p .327 .658 .972 r
F P
0 g
s
.62596 .18032 m .66778 .20087 L .65926 .20337 L p .327 .658 .972 r
F P
0 g
s
.85851 .6619 m .88828 .61554 L .88856 .61683 L p .123 0 .414 r
F P
0 g
s
.88856 .61683 m .85885 .66375 L .85851 .6619 L p .123 0 .414 r
F P
0 g
s
.75451 .25882 m .72393 .22478 L .73459 .22351 L p .334 .558 .918 r
F P
0 g
s
.73459 .22351 m .76656 .25917 L .75451 .25882 L p .334 .558 .918 r
F P
0 g
s
.23264 .35982 m .25562 .31118 L .26224 .31031 L p .924 .962 .781 r
F P
0 g
s
.25562 .31118 m .23264 .35982 L .22561 .36192 L p .924 .962 .781 r
F P
0 g
s
.18873 .48189 m .19079 .42472 L .19805 .42138 L p .967 .871 .691 r
F P
0 g
s
.20523 .41808 m .2185 .36405 L .22561 .36192 L p .955 .922 .735 r
F P
0 g
s
.2185 .36405 m .20523 .41808 L .19805 .42138 L p .955 .922 .735 r
F P
0 g
s
.29345 .26766 m .26224 .31031 L .25562 .31118 L p .872 .988 .827 r
F P
0 g
s
.71342 .22604 m .67643 .19833 L .6852 .19575 L p .323 .604 .948 r
F P
0 g
s
.6852 .19575 m .72393 .22478 L .71342 .22604 L p .323 .604 .948 r
F P
0 g
s
.76656 .25917 m .73459 .22351 L .74541 .22221 L p .334 .558 .918 r
F P
0 g
s
.74541 .22221 m .77878 .25952 L .76656 .25917 L p .334 .558 .918 r
F P
0 g
s
.66778 .20087 m .62596 .18032 L .63261 .17682 L p .327 .658 .972 r
F P
0 g
s
.63261 .17682 m .67643 .19833 L .66778 .20087 L p .327 .658 .972 r
F P
0 g
s
.6194 .18377 m .5744 .17092 L .57883 .16691 L p .349 .719 .986 r
F P
0 g
s
.57883 .16691 m .62596 .18032 L .6194 .18377 L p .349 .719 .986 r
F P
0 g
s
.22561 .36192 m .24891 .31206 L .25562 .31118 L p .924 .962 .781 r
F P
0 g
s
.24891 .31206 m .22561 .36192 L .2185 .36405 L p .924 .962 .781 r
F P
0 g
s
.19805 .42138 m .2113 .3662 L .2185 .36405 L p .955 .922 .735 r
F P
0 g
s
.2113 .3662 m .19805 .42138 L .19079 .42472 L p .955 .922 .735 r
F P
0 g
s
.25562 .31118 m .2875 .26729 L .29345 .26766 L p .872 .988 .827 r
F P
0 g
s
.2875 .26729 m .25562 .31118 L .24891 .31206 L p .872 .988 .827 r
F P
0 g
s
.33201 .23122 m .29345 .26766 L .2875 .26729 L p .801 .994 .869 r
F P
0 g
s
.72393 .22478 m .6852 .19575 L .69411 .19314 L p .323 .604 .948 r
F P
0 g
s
.69411 .19314 m .73459 .22351 L .72393 .22478 L p .323 .604 .948 r
F P
0 g
s
.88715 .61037 m .90744 .56158 L .90762 .56234 L p 0 0 .391 r
F P
0 g
s
.90762 .56234 m .88743 .61166 L .88715 .61037 L p 0 0 .391 r
F P
0 g
s
.2185 .36405 m .24211 .31295 L .24891 .31206 L p .924 .962 .781 r
F P
0 g
s
.24211 .31295 m .2185 .36405 L .2113 .3662 L p .924 .962 .781 r
F P
0 g
s
.5744 .17092 m .52576 .16583 L .52802 .16163 L p .391 .782 .989 r
F P
0 g
s
.52802 .16163 m .57883 .16691 L .5744 .17092 L p .391 .782 .989 r
F P
0 g
s
.88743 .61166 m .90762 .56234 L .90781 .5631 L p 0 0 .391 r
F P
0 g
s
.90781 .5631 m .88771 .61295 L .88743 .61166 L p 0 0 .391 r
F P
0 g
s
.19079 .42472 m .20402 .36838 L .2113 .3662 L p .955 .922 .735 r
F P
0 g
s
.67643 .19833 m .63261 .17682 L .63937 .17327 L p .327 .658 .972 r
F P
0 g
s
.63937 .17327 m .6852 .19575 L .67643 .19833 L p .327 .658 .972 r
F P
0 g
s
.88771 .61295 m .90781 .5631 L .90799 .56386 L p 0 0 .391 r
F P
0 g
s
.90799 .56386 m .88799 .61424 L .88771 .61295 L p 0 0 .391 r
F P
0 g
s
.62596 .18032 m .57883 .16691 L .58334 .16283 L p .349 .718 .986 r
F P
0 g
s
.58334 .16283 m .63261 .17682 L .62596 .18032 L p .349 .718 .986 r
F P
0 g
s
.88799 .61424 m .90799 .56386 L .90818 .56462 L p 0 0 .391 r
F P
0 g
s
.90818 .56462 m .88828 .61554 L .88799 .61424 L p 0 0 .391 r
F P
0 g
s
.37638 .20215 m .33201 .23122 L .32709 .22966 L p .715 .981 .907 r
F P
0 g
s
.2875 .26729 m .32709 .22966 L .33201 .23122 L p .801 .994 .869 r
F P
0 g
s
.88828 .61554 m .90818 .56462 L .90836 .56538 L p 0 0 .391 r
F P
0 g
s
.90836 .56538 m .88856 .61683 L .88828 .61554 L p 0 0 .391 r
F P
0 g
s
.24891 .31206 m .28146 .26692 L .2875 .26729 L p .872 .988 .827 r
F P
0 g
s
.28146 .26692 m .24891 .31206 L .24211 .31295 L p .872 .988 .827 r
F P
0 g
s
.32709 .22966 m .2875 .26729 L .28146 .26692 L p .801 .994 .869 r
F P
0 g
s
.52576 .16583 m .47524 .16918 L .47538 .16519 L p .453 .844 .981 r
F P
0 g
s
.47538 .16519 m .52802 .16163 L .52576 .16583 L p .453 .844 .981 r
F P
0 g
s
.73459 .22351 m .69411 .19314 L .70314 .19048 L p .323 .604 .948 r
F P
0 g
s
.70314 .19048 m .74541 .22221 L .73459 .22351 L p .323 .604 .948 r
F P
0 g
s
.2113 .3662 m .23523 .31385 L .24211 .31295 L p .924 .962 .781 r
F P
0 g
s
.23523 .31385 m .2113 .3662 L .20402 .36838 L p .924 .962 .781 r
F P
0 g
s
.42478 .18131 m .37638 .20215 L .37286 .19954 L p .622 .948 .939 r
F P
0 g
s
.32709 .22966 m .37286 .19954 L .37638 .20215 L p .715 .981 .907 r
F P
0 g
s
.47524 .16918 m .42478 .18131 L .42296 .17787 L p .532 .901 .964 r
F P
0 g
s
.42296 .17787 m .47538 .16519 L .47524 .16918 L p .532 .901 .964 r
F P
0 g
s
.37286 .19954 m .42296 .17787 L .42478 .18131 L p .622 .948 .939 r
F P
0 g
s
.57883 .16691 m .52802 .16163 L .53031 .15737 L p .391 .782 .989 r
F P
0 g
s
.53031 .15737 m .58334 .16283 L .57883 .16691 L p .391 .782 .989 r
F P
0 g
s
.6852 .19575 m .63937 .17327 L .64624 .16966 L p .327 .658 .972 r
F P
0 g
s
.64624 .16966 m .69411 .19314 L .6852 .19575 L p .327 .658 .972 r
F P
0 g
s
.63261 .17682 m .58334 .16283 L .58791 .15869 L p .349 .718 .986 r
F P
0 g
s
.58791 .15869 m .63937 .17327 L .63261 .17682 L p .349 .718 .986 r
F P
0 g
s
.24211 .31295 m .27535 .26654 L .28146 .26692 L p .872 .988 .827 r
F P
0 g
s
.27535 .26654 m .24211 .31295 L .23523 .31385 L p .872 .988 .827 r
F P
0 g
s
.37286 .19954 m .32709 .22966 L .32211 .22809 L p .715 .981 .907 r
F P
0 g
s
.28146 .26692 m .32211 .22809 L .32709 .22966 L p .801 .994 .869 r
F P
0 g
s
.32211 .22809 m .28146 .26692 L .27535 .26654 L p .801 .994 .869 r
F P
0 g
s
.20402 .36838 m .22826 .31476 L .23523 .31385 L p .924 .962 .781 r
F P
0 g
s
.52802 .16163 m .47538 .16519 L .47553 .16114 L p .453 .844 .981 r
F P
0 g
s
.47553 .16114 m .53031 .15737 L .52802 .16163 L p .453 .844 .981 r
F P
0 g
s
.42296 .17787 m .37286 .19954 L .36928 .1969 L p .622 .948 .939 r
F P
0 g
s
.32211 .22809 m .36928 .1969 L .37286 .19954 L p .715 .981 .907 r
F P
0 g
s
.47538 .16519 m .42296 .17787 L .42112 .17437 L p .532 .901 .964 r
F P
0 g
s
.42112 .17437 m .47553 .16114 L .47538 .16519 L p .532 .901 .964 r
F P
0 g
s
.36928 .1969 m .42112 .17437 L .42296 .17787 L p .622 .948 .939 r
F P
0 g
s
.23523 .31385 m .26915 .26616 L .27535 .26654 L p .872 .988 .827 r
F P
0 g
s
.26915 .26616 m .23523 .31385 L .22826 .31476 L p .872 .988 .827 r
F P
0 g
s
.69411 .19314 m .64624 .16966 L .65321 .166 L p .327 .658 .972 r
F P
0 g
s
.65321 .166 m .70314 .19048 L .69411 .19314 L p .327 .658 .972 r
F P
0 g
s
.58334 .16283 m .53031 .15737 L .53264 .15305 L p .391 .782 .989 r
F P
0 g
s
.53264 .15305 m .58791 .15869 L .58334 .16283 L p .391 .782 .989 r
F P
0 g
s
.39543 .87943 m .45328 .89708 L .45408 .89478 L p .474 0 0 r
F P
0 g
s
.45408 .89478 m .39527 .87689 L .39543 .87943 L p .474 0 0 r
F P
0 g
s
.51437 .90421 m .45408 .89478 L .45328 .89708 L p .374 0 0 r
F P
0 g
s
.36928 .1969 m .32211 .22809 L .31705 .22649 L p .715 .981 .907 r
F P
0 g
s
.27535 .26654 m .31705 .22649 L .32211 .22809 L p .801 .994 .869 r
F P
0 g
s
.63937 .17327 m .58791 .15869 L .59256 .15448 L p .349 .718 .986 r
F P
0 g
s
.59256 .15448 m .64624 .16966 L .63937 .17327 L p .349 .718 .986 r
F P
0 g
s
.31705 .22649 m .27535 .26654 L .26915 .26616 L p .801 .994 .869 r
F P
0 g
s
.45328 .89708 m .51246 .9064 L .51437 .90421 L p .374 0 0 r
F P
0 g
s
.57458 .90519 m .51437 .90421 L .51246 .9064 L p .251 0 0 r
F P
0 g
s
.34046 .85359 m .39543 .87943 L .39527 .87689 L p .553 0 0 r
F P
0 g
s
.39527 .87689 m .33948 .85072 L .34046 .85359 L p .553 0 0 r
F P
0 g
s
.90744 .56158 m .91684 .5096 L .91691 .50985 L p 0 0 .349 r
F P
0 g
s
.91691 .50985 m .90762 .56234 L .90744 .56158 L p 0 0 .349 r
F P
0 g
s
.90762 .56234 m .91691 .50985 L .91697 .5101 L p 0 0 .349 r
F P
0 g
s
.91697 .5101 m .90781 .5631 L .90762 .56234 L p 0 0 .349 r
F P
0 g
s
.90781 .5631 m .91697 .5101 L .91704 .51035 L p 0 0 .349 r
F P
0 g
s
.91704 .51035 m .90799 .56386 L .90781 .5631 L p 0 0 .349 r
F P
0 g
s
.90799 .56386 m .91704 .51035 L .9171 .5106 L p 0 0 .349 r
F P
0 g
s
.9171 .5106 m .90818 .56462 L .90799 .56386 L p 0 0 .349 r
F P
0 g
s
.90818 .56462 m .9171 .5106 L .91717 .51085 L p 0 0 .349 r
F P
0 g
s
.91717 .51085 m .90836 .56538 L .90818 .56462 L p 0 0 .349 r
F P
0 g
s
.53031 .15737 m .47553 .16114 L .47567 .15702 L p .453 .844 .981 r
F P
0 g
s
.47567 .15702 m .53264 .15305 L .53031 .15737 L p .453 .844 .981 r
F P
0 g
s
.42112 .17437 m .36928 .1969 L .36565 .19421 L p .622 .948 .939 r
F P
0 g
s
.31705 .22649 m .36565 .19421 L .36928 .1969 L p .715 .981 .907 r
F P
0 g
s
.51246 .9064 m .57145 .90741 L .57458 .90519 L p .251 0 0 r
F P
0 g
s
.63319 .89786 m .57458 .90519 L .57145 .90741 L p .103 0 0 r
F P
0 g
s
.28992 .81988 m .34046 .85359 L .33948 .85072 L p .609 .049 0 r
F P
0 g
s
.33948 .85072 m .28827 .81663 L .28992 .81988 L p .609 .049 0 r
F P
0 g
s
.22826 .31476 m .26286 .26577 L .26915 .26616 L p .872 .988 .827 r
F P
0 g
s
.47553 .16114 m .42112 .17437 L .41925 .17082 L p .532 .901 .964 r
F P
0 g
s
.41925 .17082 m .47567 .15702 L .47553 .16114 L p .532 .901 .964 r
F P
0 g
s
.36565 .19421 m .41925 .17082 L .42112 .17437 L p .622 .948 .939 r
F P
0 g
s
.58791 .15869 m .53264 .15305 L .53501 .14865 L p .391 .782 .989 r
F P
0 g
s
.53501 .14865 m .59256 .15448 L .58791 .15869 L p .391 .782 .989 r
F P
0 g
s
.36565 .19421 m .31705 .22649 L .31192 .22486 L p .715 .981 .907 r
F P
0 g
s
.26915 .26616 m .31192 .22486 L .31705 .22649 L p .801 .994 .869 r
F P
0 g
s
.3956 .88201 m .45247 .89943 L .45328 .89708 L p .474 0 0 r
F P
0 g
s
.45328 .89708 m .39543 .87943 L .3956 .88201 L p .474 0 0 r
F P
0 g
s
.51246 .9064 m .45328 .89708 L .45247 .89943 L p .374 0 0 r
F P
0 g
s
.31192 .22486 m .26915 .26616 L .26286 .26577 L p .801 .994 .87 r
F P
0 g
s
.64624 .16966 m .59256 .15448 L .59728 .15021 L p .349 .718 .986 r
F P
0 g
s
.59728 .15021 m .65321 .166 L .64624 .16966 L p .349 .718 .986 r
F P
0 g
s
.34146 .8565 m .3956 .88201 L .39543 .87943 L p .553 0 0 r
F P
0 g
s
.39543 .87943 m .34046 .85359 L .34146 .8565 L p .553 0 0 r
F P
0 g
s
.45247 .89943 m .51052 .90863 L .51246 .9064 L p .374 0 0 r
F P
0 g
s
.57145 .90741 m .51246 .9064 L .51052 .90863 L p .251 0 0 r
F P
0 g
s
.57145 .90741 m .62874 .9003 L .63319 .89786 L p .103 0 0 r
F P
0 g
s
.68868 .88256 m .63319 .89786 L .62874 .9003 L p 0 0 0 r
F P
0 g
s
.24532 .77881 m .28992 .81988 L .28827 .81663 L p .638 .116 0 r
F P
0 g
s
.28827 .81663 m .24317 .77514 L .24532 .77881 L p .638 .116 0 r
F P
0 g
s
.53264 .15305 m .47567 .15702 L .47582 .15283 L p .453 .844 .981 r
F P
0 g
s
.47582 .15283 m .53501 .14865 L .53264 .15305 L p .453 .844 .981 r
F P
0 g
s
.41925 .17082 m .36565 .19421 L .36197 .19149 L p .622 .948 .939 r
F P
0 g
s
.31192 .22486 m .36197 .19149 L .36565 .19421 L p .715 .981 .907 r
F P
0 g
s
.29158 .82318 m .34146 .8565 L .34046 .85359 L p .609 .049 0 r
F P
0 g
s
.34046 .85359 m .28992 .81988 L .29158 .82318 L p .609 .049 0 r
F P
0 g
s
.51052 .90863 m .56827 .90968 L .57145 .90741 L p .251 0 0 r
F P
0 g
s
.62874 .9003 m .57145 .90741 L .56827 .90968 L p .103 0 0 r
F P
0 g
s
.47567 .15702 m .41925 .17082 L .41734 .16722 L p .532 .901 .964 r
F P
0 g
s
.36197 .19149 m .41734 .16722 L .41925 .17082 L p .622 .948 .939 r
F P
0 g
s
.41734 .16722 m .47582 .15283 L .47567 .15702 L p .532 .901 .964 r
F P
0 g
s
.36197 .19149 m .31192 .22486 L .30671 .22322 L p .715 .981 .907 r
F P
0 g
s
.26286 .26577 m .30671 .22322 L .31192 .22486 L p .801 .994 .87 r
F P
0 g
s
.59256 .15448 m .53501 .14865 L .53742 .14418 L p .391 .782 .989 r
F P
0 g
s
.53742 .14418 m .59728 .15021 L .59256 .15448 L p .391 .782 .989 r
F P
0 g
s
.39577 .88463 m .45164 .90181 L .45247 .89943 L p .474 0 0 r
F P
0 g
s
.45247 .89943 m .3956 .88201 L .39577 .88463 L p .474 0 0 r
F P
0 g
s
.51052 .90863 m .45247 .89943 L .45164 .90181 L p .374 0 0 r
F P
0 g
s
.34247 .85945 m .39577 .88463 L .3956 .88201 L p .553 0 0 r
F P
0 g
s
.3956 .88201 m .34146 .8565 L .34247 .85945 L p .553 0 0 r
F P
0 g
s
.45164 .90181 m .50856 .91089 L .51052 .90863 L p .374 0 0 r
F P
0 g
s
.56827 .90968 m .51052 .90863 L .50856 .91089 L p .251 0 0 r
F P
0 g
s
.2475 .78253 m .29158 .82318 L .28992 .81988 L p .638 .116 0 r
F P
0 g
s
.28992 .81988 m .24532 .77881 L .2475 .78253 L p .638 .116 0 r
F P
0 g
s
.56827 .90968 m .62422 .90278 L .62874 .9003 L p .103 0 0 r
F P
0 g
s
.68284 .88541 m .62874 .9003 L .62422 .90278 L p 0 0 0 r
F P
0 g
s
.62874 .9003 m .68284 .88541 L .68868 .88256 L p 0 0 0 r
F P
0 g
s
.73957 .85977 m .68868 .88256 L .68284 .88541 L p 0 0 0 r
F P
0 g
s
.20818 .73112 m .24532 .77881 L .24317 .77514 L p .632 .169 0 r
F P
0 g
s
.24317 .77514 m .20567 .72702 L .20818 .73112 L p .632 .169 0 r
F P
0 g
s
.41734 .16722 m .36197 .19149 L .35823 .18872 L p .622 .948 .939 r
F P
0 g
s
.30671 .22322 m .35823 .18872 L .36197 .19149 L p .715 .981 .907 r
F P
0 g
s
.29327 .82652 m .34247 .85945 L .34146 .8565 L p .609 .049 0 r
F P
0 g
s
.34146 .8565 m .29158 .82318 L .29327 .82652 L p .609 .049 0 r
F P
0 g
s
.53501 .14865 m .47582 .15283 L .47597 .14858 L p .453 .844 .981 r
F P
0 g
s
.47597 .14858 m .53742 .14418 L .53501 .14865 L p .453 .844 .981 r
F P
0 g
s
.50856 .91089 m .56504 .91197 L .56827 .90968 L p .251 0 0 r
F P
0 g
s
.62422 .90278 m .56827 .90968 L .56504 .91197 L p .103 0 0 r
F P
0 g
s
.91684 .5096 m .91445 .45615 L .91438 .4559 L p 0 0 .278 r
F P
0 g
s
.91438 .4559 m .91691 .50985 L .91684 .5096 L p 0 0 .278 r
F P
0 g
s
.91691 .50985 m .91438 .4559 L .91432 .45566 L p 0 0 .278 r
F P
0 g
s
.91432 .45566 m .91697 .5101 L .91691 .50985 L p 0 0 .278 r
F P
0 g
s
.91697 .5101 m .91432 .45566 L .91425 .45541 L p 0 0 .278 r
F P
0 g
s
.91425 .45541 m .91704 .51035 L .91697 .5101 L p 0 0 .278 r
F P
0 g
s
.91704 .51035 m .91425 .45541 L .91419 .45517 L p 0 0 .278 r
F P
0 g
s
.91419 .45517 m .9171 .5106 L .91704 .51035 L p 0 0 .278 r
F P
0 g
s
.9171 .5106 m .91419 .45517 L .91412 .45492 L p 0 0 .278 r
F P
0 g
s
.91412 .45492 m .91717 .51085 L .9171 .5106 L p 0 0 .278 r
F P
0 g
s
.47582 .15283 m .41734 .16722 L .41541 .16355 L p .532 .901 .964 r
F P
0 g
s
.35823 .18872 m .41541 .16355 L .41734 .16722 L p .622 .948 .939 r
F P
0 g
s
.41541 .16355 m .47597 .14858 L .47582 .15283 L p .532 .901 .964 r
F P
0 g
s
.39594 .88729 m .4508 .90422 L .45164 .90181 L p .474 0 0 r
F P
0 g
s
.45164 .90181 m .39577 .88463 L .39594 .88729 L p .474 0 0 r
F P
0 g
s
.50856 .91089 m .45164 .90181 L .4508 .90422 L p .374 0 0 r
F P
0 g
s
.34349 .86244 m .39594 .88729 L .39577 .88463 L p .553 0 0 r
F P
0 g
s
.39577 .88463 m .34247 .85945 L .34349 .86244 L p .553 0 0 r
F P
0 g
s
.4508 .90422 m .50656 .91319 L .50856 .91089 L p .374 0 0 r
F P
0 g
s
.56504 .91197 m .50856 .91089 L .50656 .91319 L p .251 0 0 r
F P
0 g
s
.24971 .78631 m .29327 .82652 L .29158 .82318 L p .638 .116 0 r
F P
0 g
s
.29158 .82318 m .2475 .78253 L .24971 .78631 L p .638 .116 0 r
F P
0 g
s
.21071 .73527 m .2475 .78253 L .24532 .77881 L p .632 .169 0 r
F P
0 g
s
.24532 .77881 m .20818 .73112 L .21071 .73527 L p .632 .169 0 r
F P
0 g
s
.56504 .91197 m .61962 .90529 L .62422 .90278 L p .103 0 0 r
F P
0 g
s
.67692 .88831 m .62422 .90278 L .61962 .90529 L p 0 0 0 r
F P
0 g
s
.62422 .90278 m .67692 .88831 L .68284 .88541 L p 0 0 0 r
F P
0 g
s
.29498 .82991 m .34349 .86244 L .34247 .85945 L p .609 .049 0 r
F P
0 g
s
.34247 .85945 m .29327 .82652 L .29498 .82991 L p .609 .049 0 r
F P
0 g
s
.73233 .86326 m .68284 .88541 L .67692 .88831 L p 0 0 0 r
F P
0 g
s
.68284 .88541 m .73233 .86326 L .73957 .85977 L p 0 0 0 r
F P
0 g
s
.50656 .91319 m .56176 .91431 L .56504 .91197 L p .251 0 0 r
F P
0 g
s
.61962 .90529 m .56504 .91197 L .56176 .91431 L p .103 0 0 r
F P
0 g
s
.1799 .67777 m .20818 .73112 L .20567 .72702 L p 0 0 .374 r
F P
0 g
s
.20567 .72702 m .17719 .67326 L .1799 .67777 L p 0 0 .374 r
F P
0 g
s
.78444 .83017 m .73957 .85977 L .73233 .86326 L p .473 .871 .698 r
F P
0 g
s
.39611 .88999 m .44994 .90668 L .4508 .90422 L p .474 0 0 r
F P
0 g
s
.4508 .90422 m .39594 .88729 L .39611 .88999 L p .474 0 0 r
F P
0 g
s
.50656 .91319 m .4508 .90422 L .44994 .90668 L p .374 0 0 r
F P
0 g
s
.34453 .86548 m .39611 .88999 L .39594 .88729 L p .553 0 0 r
F P
0 g
s
.39594 .88729 m .34349 .86244 L .34453 .86548 L p .553 0 0 r
F P
0 g
s
.25195 .79013 m .29498 .82991 L .29327 .82652 L p .638 .116 0 r
F P
0 g
s
.29327 .82652 m .24971 .78631 L .25195 .79013 L p .638 .116 0 r
F P
0 g
s
.44994 .90668 m .50453 .91552 L .50656 .91319 L p .374 0 0 r
F P
0 g
s
.56176 .91431 m .50656 .91319 L .50453 .91552 L p .252 0 0 r
F P
0 g
s
.21328 .73947 m .24971 .78631 L .2475 .78253 L p .632 .169 0 r
F P
0 g
s
.2475 .78253 m .21071 .73527 L .21328 .73947 L p .632 .169 0 r
F P
0 g
s
.29672 .83335 m .34453 .86548 L .34349 .86244 L p .609 .049 0 r
F P
0 g
s
.34349 .86244 m .29498 .82991 L .29672 .83335 L p .609 .049 0 r
F P
0 g
s
.56176 .91431 m .61496 .90785 L .61962 .90529 L p .103 0 0 r
F P
0 g
s
.6709 .89125 m .61962 .90529 L .61496 .90785 L p 0 0 0 r
F P
0 g
s
.61962 .90529 m .6709 .89125 L .67692 .88831 L p 0 0 0 r
F P
0 g
s
.72499 .86681 m .67692 .88831 L .6709 .89125 L p 0 0 0 r
F P
0 g
s
.67692 .88831 m .72499 .86681 L .73233 .86326 L p 0 0 0 r
F P
0 g
s
.18265 .68233 m .21071 .73527 L .20818 .73112 L p 0 0 .374 r
F P
0 g
s
.20818 .73112 m .1799 .67777 L .18265 .68233 L p 0 0 .374 r
F P
0 g
s
.50453 .91552 m .55842 .91668 L .56176 .91431 L p .252 0 0 r
F P
0 g
s
.61496 .90785 m .56176 .91431 L .55842 .91668 L p .104 0 0 r
F P
0 g
s
.77583 .83455 m .73233 .86326 L .72499 .86681 L p .473 .871 .698 r
F P
0 g
s
.73233 .86326 m .77583 .83455 L .78444 .83017 L p .473 .871 .698 r
F P
0 g
s
.25421 .794 m .29672 .83335 L .29498 .82991 L p .638 .116 0 r
F P
0 g
s
.29498 .82991 m .25195 .79013 L .25421 .794 L p .638 .116 0 r
F P
0 g
s
.89872 .39951 m .91412 .45492 L .91419 .45517 L p .693 .413 0 r
F P
0 g
s
.21588 .74373 m .25195 .79013 L .24971 .78631 L p .632 .169 0 r
F P
0 g
s
.24971 .78631 m .21328 .73947 L .21588 .74373 L p .632 .169 0 r
F P
0 g
s
.89891 .40023 m .91419 .45517 L .91425 .45541 L p .693 .413 0 r
F P
0 g
s
.91419 .45517 m .89891 .40023 L .89872 .39951 L p .693 .413 0 r
F P
0 g
s
.8991 .40096 m .91425 .45541 L .91432 .45566 L p .693 .413 0 r
F P
0 g
s
.91425 .45541 m .8991 .40096 L .89891 .40023 L p .693 .413 0 r
F P
0 g
s
.8993 .40169 m .91432 .45566 L .91438 .4559 L p .693 .413 0 r
F P
0 g
s
.91432 .45566 m .8993 .40169 L .8991 .40096 L p .693 .413 0 r
F P
0 g
s
.89949 .40241 m .91438 .4559 L .91445 .45615 L p .693 .413 0 r
F P
0 g
s
.91438 .4559 m .89949 .40241 L .8993 .40169 L p .693 .413 0 r
F P
0 g
s
.91445 .45615 m .89968 .40314 L .89949 .40241 L p .693 .413 0 r
F P
0 g
s
.16179 .61997 m .1799 .67777 L .17719 .67326 L p 0 0 .39 r
F P
0 g
s
.17719 .67326 m .15899 .6151 L .16179 .61997 L p 0 0 .39 r
F P
0 g
s
.18543 .68694 m .21328 .73947 L .21071 .73527 L p 0 0 .374 r
F P
0 g
s
.21071 .73527 m .18265 .68233 L .18543 .68694 L p 0 0 .374 r
F P
0 g
s
.55842 .91668 m .61022 .91045 L .61496 .90785 L p .104 0 0 r
F P
0 g
s
.66479 .89424 m .61496 .90785 L .61022 .91045 L p 0 0 0 r
F P
0 g
s
.61496 .90785 m .66479 .89424 L .6709 .89125 L p 0 0 0 r
F P
0 g
s
.71753 .87041 m .6709 .89125 L .66479 .89424 L p 0 0 0 r
F P
0 g
s
.6709 .89125 m .71753 .87041 L .72499 .86681 L p 0 0 0 r
F P
0 g
s
.82193 .79462 m .78444 .83017 L .77583 .83455 L p .658 .959 .756 r
F P
0 g
s
.76709 .83899 m .72499 .86681 L .71753 .87041 L p .473 .871 .698 r
F P
0 g
s
.72499 .86681 m .76709 .83899 L .77583 .83455 L p .473 .871 .698 r
F P
0 g
s
.21852 .74804 m .25421 .794 L .25195 .79013 L p .632 .169 0 r
F P
0 g
s
.25195 .79013 m .21588 .74373 L .21852 .74804 L p .632 .169 0 r
F P
0 g
s
.16461 .62489 m .18265 .68233 L .1799 .67777 L p 0 0 .39 r
F P
0 g
s
.1799 .67777 m .16179 .61997 L .16461 .62489 L p 0 0 .39 r
F P
0 g
s
.18824 .69161 m .21588 .74373 L .21328 .73947 L p 0 0 .374 r
F P
0 g
s
.21328 .73947 m .18543 .68694 L .18824 .69161 L p 0 0 .374 r
F P
0 g
s
.81204 .80014 m .77583 .83455 L .76709 .83899 L p .658 .959 .756 r
F P
0 g
s
.77583 .83455 m .81204 .80014 L .82193 .79462 L p .658 .959 .756 r
F P
0 g
s
.61022 .91045 m .65858 .89728 L .66479 .89424 L p 0 0 0 r
F P
0 g
s
.70995 .87406 m .66479 .89424 L .65858 .89728 L p 0 0 0 r
F P
0 g
s
.66479 .89424 m .70995 .87406 L .71753 .87041 L p 0 0 0 r
F P
0 g
s
.16747 .62987 m .18543 .68694 L .18265 .68233 L p 0 0 .39 r
F P
0 g
s
.18265 .68233 m .16461 .62489 L .16747 .62987 L p 0 0 .39 r
F P
0 g
s
.75822 .84349 m .71753 .87041 L .70995 .87406 L p .473 .871 .698 r
F P
0 g
s
.71753 .87041 m .75822 .84349 L .76709 .83899 L p .473 .871 .698 r
F P
0 g
s
.15492 .55921 m .16179 .61997 L .15899 .6151 L p 0 0 .398 r
F P
0 g
s
.15899 .6151 m .15216 .55405 L .15492 .55921 L p 0 0 .398 r
F P
0 g
s
.19109 .69633 m .21852 .74804 L .21588 .74373 L p 0 0 .374 r
F P
0 g
s
.21588 .74373 m .18824 .69161 L .19109 .69633 L p 0 0 .374 r
F P
0 g
s
.87088 .34667 m .89872 .39951 L .89891 .40023 L p .828 .453 0 r
F P
0 g
s
.8508 .7542 m .82193 .79462 L .81204 .80014 L p .799 .997 .786 r
F P
0 g
s
.87119 .34787 m .89891 .40023 L .8991 .40096 L p .828 .453 0 r
F P
0 g
s
.89891 .40023 m .87119 .34787 L .87088 .34667 L p .828 .453 0 r
F P
0 g
s
.80201 .80572 m .76709 .83899 L .75822 .84349 L p .658 .959 .756 r
F P
0 g
s
.76709 .83899 m .80201 .80572 L .81204 .80014 L p .658 .959 .756 r
F P
0 g
s
.8715 .34907 m .8991 .40096 L .8993 .40169 L p .828 .453 0 r
F P
0 g
s
.8991 .40096 m .8715 .34907 L .87119 .34787 L p .828 .453 0 r
F P
0 g
s
.87181 .35027 m .8993 .40169 L .89949 .40241 L p .828 .453 0 r
F P
0 g
s
.8993 .40169 m .87181 .35027 L .8715 .34907 L p .828 .453 0 r
F P
0 g
s
.17035 .6349 m .18824 .69161 L .18543 .68694 L p 0 0 .39 r
F P
0 g
s
.18543 .68694 m .16747 .62987 L .17035 .6349 L p 0 0 .39 r
F P
0 g
s
.1577 .56442 m .16461 .62489 L .16179 .61997 L p 0 0 .398 r
F P
0 g
s
.16179 .61997 m .15492 .55921 L .1577 .56442 L p 0 0 .398 r
F P
0 g
s
.65858 .89728 m .70226 .87778 L .70995 .87406 L p 0 0 0 r
F P
0 g
s
.87213 .35148 m .89949 .40241 L .89968 .40314 L p .828 .453 0 r
F P
0 g
s
.89949 .40241 m .87213 .35148 L .87181 .35027 L p .828 .453 0 r
F P
0 g
s
.89968 .40314 m .87244 .35268 L .87213 .35148 L p .828 .453 0 r
F P
0 g
s
.74923 .84806 m .70995 .87406 L .70226 .87778 L p .473 .871 .698 r
F P
0 g
s
.70995 .87406 m .74923 .84806 L .75822 .84349 L p .473 .871 .698 r
F P
0 g
s
.8398 .7611 m .81204 .80014 L .80201 .80572 L p .799 .997 .786 r
F P
0 g
s
.81204 .80014 m .8398 .7611 L .8508 .7542 L p .799 .997 .786 r
F P
0 g
s
.16052 .56969 m .16747 .62987 L .16461 .62489 L p 0 0 .398 r
F P
0 g
s
.16461 .62489 m .1577 .56442 L .16052 .56969 L p 0 0 .398 r
F P
0 g
s
.17327 .63999 m .19109 .69633 L .18824 .69161 L p 0 0 .39 r
F P
0 g
s
.18824 .69161 m .17035 .6349 L .17327 .63999 L p 0 0 .39 r
F P
0 g
s
.79184 .81139 m .75822 .84349 L .74923 .84806 L p .658 .959 .756 r
F P
0 g
s
.75822 .84349 m .79184 .81139 L .80201 .80572 L p .658 .959 .756 r
F P
0 g
s
.16011 .49721 m .15492 .55921 L .15216 .55405 L p 0 0 .395 r
F P
0 g
s
.15216 .55405 m .15749 .49184 L .16011 .49721 L p 0 0 .395 r
F P
0 g
s
.70226 .87778 m .74009 .8527 L .74923 .84806 L p .473 .871 .698 r
F P
0 g
s
.16336 .57501 m .17035 .6349 L .16747 .62987 L p 0 0 .398 r
F P
0 g
s
.16747 .62987 m .16052 .56969 L .16336 .57501 L p 0 0 .398 r
F P
0 g
s
.82866 .76808 m .80201 .80572 L .79184 .81139 L p .799 .997 .786 r
F P
0 g
s
.80201 .80572 m .82866 .76808 L .8398 .7611 L p .799 .997 .786 r
F P
0 g
s
.16276 .50263 m .1577 .56442 L .15492 .55921 L p 0 0 .395 r
F P
0 g
s
.15492 .55921 m .16011 .49721 L .16276 .50263 L p 0 0 .395 r
F P
0 g
s
.78153 .81713 m .74923 .84806 L .74009 .8527 L p .658 .959 .756 r
F P
0 g
s
.74923 .84806 m .78153 .81713 L .79184 .81139 L p .658 .959 .756 r
F P
0 g
s
.87001 .71019 m .8508 .7542 L .8398 .7611 L p .888 .989 .789 r
F P
0 g
s
.83103 .29857 m .87088 .34667 L .87119 .34787 L p .875 .447 .039 r
F P
0 g
s
.83145 .30024 m .87119 .34787 L .8715 .34907 L p .875 .447 .039 r
F P
0 g
s
.87119 .34787 m .83145 .30024 L .83103 .29857 L p .875 .447 .039 r
F P
0 g
s
.83187 .30192 m .8715 .34907 L .87181 .35027 L p .875 .447 .039 r
F P
0 g
s
.8715 .34907 m .83187 .30192 L .83145 .30024 L p .875 .447 .039 r
F P
0 g
s
.16623 .58039 m .17327 .63999 L .17035 .6349 L p 0 0 .398 r
F P
0 g
s
.17035 .6349 m .16336 .57501 L .16623 .58039 L p 0 0 .398 r
F P
0 g
s
.83229 .30359 m .87181 .35027 L .87213 .35148 L p .875 .447 .039 r
F P
0 g
s
.87181 .35027 m .83229 .30359 L .83187 .30192 L p .875 .447 .039 r
F P
0 g
s
.16543 .5081 m .16052 .56969 L .1577 .56442 L p 0 0 .395 r
F P
0 g
s
.1577 .56442 m .16276 .50263 L .16543 .5081 L p 0 0 .395 r
F P
0 g
s
.83271 .30527 m .87213 .35148 L .87244 .35268 L p .875 .447 .039 r
F P
0 g
s
.87213 .35148 m .83271 .30527 L .83229 .30359 L p .875 .447 .039 r
F P
0 g
s
.87244 .35268 m .83313 .30696 L .83271 .30527 L p .875 .447 .039 r
F P
0 g
s
.81736 .77516 m .79184 .81139 L .78153 .81713 L p .799 .997 .786 r
F P
0 g
s
.79184 .81139 m .81736 .77516 L .82866 .76808 L p .799 .997 .786 r
F P
0 g
s
.1778 .43591 m .16011 .49721 L .15749 .49184 L p 0 0 .385 r
F P
0 g
s
.15749 .49184 m .17538 .43044 L .1778 .43591 L p 0 0 .385 r
F P
0 g
s
.85814 .7187 m .8398 .7611 L .82866 .76808 L p .888 .989 .789 r
F P
0 g
s
.8398 .7611 m .85814 .7187 L .87001 .71019 L p .888 .989 .789 r
F P
0 g
s
.74009 .8527 m .77107 .82296 L .78153 .81713 L p .658 .959 .756 r
F P
0 g
s
.16813 .51363 m .16336 .57501 L .16052 .56969 L p 0 0 .395 r
F P
0 g
s
.16052 .56969 m .16543 .5081 L .16813 .51363 L p 0 0 .395 r
F P
0 g
s
.18024 .44144 m .16276 .50263 L .16011 .49721 L p 0 0 .385 r
F P
0 g
s
.16011 .49721 m .1778 .43591 L .18024 .44144 L p 0 0 .385 r
F P
0 g
s
.80592 .78233 m .78153 .81713 L .77107 .82296 L p .799 .997 .786 r
F P
0 g
s
.78153 .81713 m .80592 .78233 L .81736 .77516 L p .799 .997 .786 r
F P
0 g
s
.84613 .72731 m .82866 .76808 L .81736 .77516 L p .888 .989 .789 r
F P
0 g
s
.82866 .76808 m .84613 .72731 L .85814 .7187 L p .888 .989 .789 r
F P
0 g
s
.17086 .51921 m .16623 .58039 L .16336 .57501 L p 0 0 .395 r
F P
0 g
s
.16336 .57501 m .16813 .51363 L .17086 .51921 L p 0 0 .395 r
F P
0 g
s
.18271 .44701 m .16543 .5081 L .16276 .50263 L p 0 0 .385 r
F P
0 g
s
.16276 .50263 m .18024 .44144 L .18271 .44701 L p 0 0 .385 r
F P
0 g
s
.78017 .25732 m .83103 .29857 L .83145 .30024 L p .861 .412 .115 r
F P
0 g
s
.87872 .66405 m .87001 .71019 L .85814 .7187 L p .931 .952 .776 r
F P
0 g
s
.78068 .25947 m .83145 .30024 L .83187 .30192 L p .861 .412 .115 r
F P
0 g
s
.83145 .30024 m .78068 .25947 L .78017 .25732 L p .861 .412 .115 r
F P
0 g
s
.7812 .26162 m .83187 .30192 L .83229 .30359 L p .861 .412 .115 r
F P
0 g
s
.83187 .30192 m .7812 .26162 L .78068 .25947 L p .861 .412 .115 r
F P
0 g
s
.20797 .37743 m .1778 .43591 L .17538 .43044 L p .07 0 .371 r
F P
0 g
s
.17538 .43044 m .20581 .37196 L .20797 .37743 L p .07 0 .371 r
F P
0 g
s
.78172 .26378 m .83229 .30359 L .83271 .30527 L p .861 .412 .115 r
F P
0 g
s
.83229 .30359 m .78172 .26378 L .7812 .26162 L p .861 .412 .115 r
F P
0 g
s
.77107 .82296 m .79433 .7896 L .80592 .78233 L p .799 .997 .786 r
F P
0 g
s
.83398 .73603 m .81736 .77516 L .80592 .78233 L p .888 .989 .789 r
F P
0 g
s
.81736 .77516 m .83398 .73603 L .84613 .72731 L p .888 .989 .789 r
F P
0 g
s
.78224 .26594 m .83271 .30527 L .83313 .30696 L p .861 .412 .115 r
F P
0 g
s
.83271 .30527 m .78224 .26594 L .78172 .26378 L p .861 .412 .115 r
F P
0 g
s
.18519 .45264 m .16813 .51363 L .16543 .5081 L p 0 0 .385 r
F P
0 g
s
.16543 .5081 m .18271 .44701 L .18519 .45264 L p 0 0 .385 r
F P
0 g
s
.83313 .30696 m .78276 .26811 L .78224 .26594 L p .861 .412 .115 r
F P
0 g
s
.86631 .67437 m .85814 .7187 L .84613 .72731 L p .931 .952 .776 r
F P
0 g
s
.85814 .7187 m .86631 .67437 L .87872 .66405 L p .931 .952 .776 r
F P
0 g
s
.21016 .38295 m .18024 .44144 L .1778 .43591 L p .07 0 .371 r
F P
0 g
s
.1778 .43591 m .20797 .37743 L .21016 .38295 L p .07 0 .371 r
F P
0 g
s
.1877 .45832 m .17086 .51921 L .16813 .51363 L p 0 0 .385 r
F P
0 g
s
.16813 .51363 m .18519 .45264 L .1877 .45832 L p 0 0 .385 r
F P
0 g
s
.82167 .74485 m .80592 .78233 L .79433 .7896 L p .888 .989 .789 r
F P
0 g
s
.80592 .78233 m .82167 .74485 L .83398 .73603 L p .888 .989 .789 r
F P
0 g
s
.21237 .38852 m .18271 .44701 L .18024 .44144 L p .07 0 .371 r
F P
0 g
s
.18024 .44144 m .21016 .38295 L .21237 .38852 L p .07 0 .371 r
F P
0 g
s
.85378 .68481 m .84613 .72731 L .83398 .73603 L p .931 .952 .776 r
F P
0 g
s
.84613 .72731 m .85378 .68481 L .86631 .67437 L p .931 .952 .776 r
F P
0 g
s
.71987 .22488 m .78017 .25732 L .78068 .25947 L p .816 .368 .171 r
F P
0 g
s
.25007 .32394 m .20797 .37743 L .20581 .37196 L p .19 0 .355 r
F P
0 g
s
.20581 .37196 m .24818 .31857 L .25007 .32394 L p .19 0 .355 r
F P
0 g
s
.72047 .22751 m .78068 .25947 L .7812 .26162 L p .816 .368 .171 r
F P
0 g
s
.78068 .25947 m .72047 .22751 L .71987 .22488 L p .816 .368 .171 r
F P
0 g
s
.72108 .23014 m .7812 .26162 L .78172 .26378 L p .816 .368 .171 r
F P
0 g
s
.7812 .26162 m .72108 .23014 L .72047 .22751 L p .816 .368 .171 r
F P
0 g
s
.21459 .39414 m .18519 .45264 L .18271 .44701 L p .07 0 .371 r
F P
0 g
s
.18271 .44701 m .21237 .38852 L .21459 .39414 L p .07 0 .371 r
F P
0 g
s
.79433 .7896 m .80922 .75379 L .82167 .74485 L p .888 .989 .789 r
F P
0 g
s
.87642 .61743 m .87872 .66405 L .86631 .67437 L p .942 .901 .756 r
F P
0 g
s
.7217 .23279 m .78172 .26378 L .78224 .26594 L p .816 .368 .171 r
F P
0 g
s
.78172 .26378 m .7217 .23279 L .72108 .23014 L p .816 .368 .171 r
F P
0 g
s
.84111 .69536 m .83398 .73603 L .82167 .74485 L p .931 .952 .776 r
F P
0 g
s
.83398 .73603 m .84111 .69536 L .85378 .68481 L p .931 .952 .776 r
F P
0 g
s
.25198 .32935 m .21016 .38295 L .20797 .37743 L p .19 0 .355 r
F P
0 g
s
.20797 .37743 m .25007 .32394 L .25198 .32935 L p .19 0 .355 r
F P
0 g
s
.72231 .23544 m .78224 .26594 L .78276 .26811 L p .816 .368 .171 r
F P
0 g
s
.78224 .26594 m .72231 .23544 L .7217 .23279 L p .816 .368 .171 r
F P
0 g
s
.78276 .26811 m .72293 .2381 L .72231 .23544 L p .816 .368 .171 r
F P
0 g
s
.21683 .3998 m .1877 .45832 L .18519 .45264 L p .07 0 .371 r
F P
0 g
s
.18519 .45264 m .21459 .39414 L .21683 .3998 L p .07 0 .371 r
F P
0 g
s
.8639 .62971 m .86631 .67437 L .85378 .68481 L p .942 .901 .756 r
F P
0 g
s
.86631 .67437 m .8639 .62971 L .87642 .61743 L p .942 .901 .756 r
F P
0 g
s
.25391 .3348 m .21237 .38852 L .21016 .38295 L p .19 0 .355 r
F P
0 g
s
.21016 .38295 m .25198 .32935 L .25391 .3348 L p .19 0 .355 r
F P
0 g
s
.30296 .27756 m .25007 .32394 L .24818 .31857 L p .296 .036 .339 r
F P
0 g
s
.24818 .31857 m .30134 .2724 L .30296 .27756 L p .296 .036 .339 r
F P
0 g
s
.82829 .70602 m .82167 .74485 L .80922 .75379 L p .931 .952 .776 r
F P
0 g
s
.82167 .74485 m .82829 .70602 L .84111 .69536 L p .931 .952 .776 r
F P
0 g
s
.65224 .2029 m .71987 .22488 L .72047 .22751 L p .758 .32 .212 r
F P
0 g
s
.65295 .20601 m .72047 .22751 L .72108 .23014 L p .758 .32 .212 r
F P
0 g
s
.72047 .22751 m .65295 .20601 L .65224 .2029 L p .758 .32 .212 r
F P
0 g
s
.25585 .34029 m .21459 .39414 L .21237 .38852 L p .19 0 .355 r
F P
0 g
s
.21237 .38852 m .25391 .3348 L .25585 .34029 L p .19 0 .355 r
F P
0 g
s
.85125 .6421 m .85378 .68481 L .84111 .69536 L p .942 .901 .756 r
F P
0 g
s
.85378 .68481 m .85125 .6421 L .8639 .62971 L p .942 .901 .756 r
F P
0 g
s
.65366 .20913 m .72108 .23014 L .7217 .23279 L p .758 .32 .212 r
F P
0 g
s
.72108 .23014 m .65366 .20913 L .65295 .20601 L p .758 .32 .212 r
F P
0 g
s
.3046 .28275 m .25198 .32935 L .25007 .32394 L p .296 .036 .339 r
F P
0 g
s
.25007 .32394 m .30296 .27756 L .3046 .28275 L p .296 .036 .339 r
F P
0 g
s
.65437 .21227 m .7217 .23279 L .72231 .23544 L p .758 .32 .212 r
F P
0 g
s
.7217 .23279 m .65437 .21227 L .65366 .20913 L p .758 .32 .212 r
F P
0 g
s
.80922 .75379 m .81534 .71681 L .82829 .70602 L p .931 .952 .776 r
F P
0 g
s
.2578 .34583 m .21683 .3998 L .21459 .39414 L p .19 0 .355 r
F P
0 g
s
.21459 .39414 m .25585 .34029 L .2578 .34583 L p .19 0 .355 r
F P
0 g
s
.65509 .21542 m .72231 .23544 L .72293 .2381 L p .758 .32 .211 r
F P
0 g
s
.72231 .23544 m .65509 .21542 L .65437 .21227 L p .758 .32 .212 r
F P
0 g
s
.30624 .28798 m .25391 .3348 L .25198 .32935 L p .297 .036 .339 r
F P
0 g
s
.25198 .32935 m .3046 .28275 L .30624 .28798 L p .297 .036 .339 r
F P
0 g
s
.36492 .24023 m .30296 .27756 L .30134 .2724 L p .39 .085 .322 r
F P
0 g
s
.30134 .2724 m .36355 .23537 L .36492 .24023 L p .39 .085 .322 r
F P
0 g
s
.57987 .19259 m .65224 .2029 L .65295 .20601 L p .693 .272 .242 r
F P
0 g
s
.72293 .2381 m .65581 .21858 L .65509 .21542 L p .758 .32 .211 r
F P
0 g
s
.83848 .65461 m .84111 .69536 L .82829 .70602 L p .942 .901 .756 r
F P
0 g
s
.84111 .69536 m .83848 .65461 L .85125 .6421 L p .942 .901 .756 r
F P
0 g
s
.863 .57211 m .87642 .61743 L .8639 .62971 L p .933 .846 .734 r
F P
0 g
s
.58069 .19618 m .65295 .20601 L .65366 .20913 L p .693 .272 .242 r
F P
0 g
s
.65295 .20601 m .58069 .19618 L .57987 .19259 L p .693 .272 .242 r
F P
0 g
s
.3079 .29325 m .25585 .34029 L .25391 .3348 L p .297 .036 .339 r
F P
0 g
s
.25391 .3348 m .30624 .28798 L .3079 .29325 L p .297 .036 .339 r
F P
0 g
s
.3663 .24512 m .3046 .28275 L .30296 .27756 L p .39 .085 .322 r
F P
0 g
s
.30296 .27756 m .36492 .24023 L .3663 .24512 L p .39 .085 .322 r
F P
0 g
s
.58152 .19979 m .65366 .20913 L .65437 .21227 L p .693 .272 .242 r
F P
0 g
s
.65366 .20913 m .58152 .19979 L .58069 .19618 L p .693 .272 .242 r
F P
0 g
s
.43369 .21356 m .36492 .24023 L .36355 .23537 L p .475 .132 .305 r
F P
0 g
s
.36355 .23537 m .43254 .20908 L .43369 .21356 L p .475 .132 .305 r
F P
0 g
s
.85083 .5864 m .8639 .62971 L .85125 .6421 L p .933 .846 .734 r
F P
0 g
s
.8639 .62971 m .85083 .5864 L .863 .57211 L p .933 .846 .734 r
F P
0 g
s
.50563 .19463 m .57987 .19259 L .58069 .19618 L p .625 .225 .266 r
F P
0 g
s
.82559 .66725 m .82829 .70602 L .81534 .71681 L p .942 .901 .756 r
F P
0 g
s
.82829 .70602 m .82559 .66725 L .83848 .65461 L p .942 .901 .756 r
F P
0 g
s
.58235 .20342 m .65437 .21227 L .65509 .21542 L p .693 .272 .242 r
F P
0 g
s
.65437 .21227 m .58235 .20342 L .58152 .19979 L p .693 .272 .242 r
F P
0 g
s
.36769 .25004 m .30624 .28798 L .3046 .28275 L p .39 .085 .322 r
F P
0 g
s
.3046 .28275 m .3663 .24512 L .36769 .25004 L p .39 .085 .322 r
F P
0 g
s
.30957 .29856 m .2578 .34583 L .25585 .34029 L p .297 .036 .339 r
F P
0 g
s
.25585 .34029 m .3079 .29325 L .30957 .29856 L p .297 .036 .339 r
F P
0 g
s
.5066 .19869 m .58069 .19618 L .58152 .19979 L p .625 .226 .266 r
F P
0 g
s
.58069 .19618 m .5066 .19869 L .50563 .19463 L p .625 .225 .266 r
F P
0 g
s
.5066 .19869 m .43369 .21356 L .43254 .20908 L p .552 .179 .287 r
F P
0 g
s
.43254 .20908 m .50563 .19463 L .5066 .19869 L p .552 .179 .287 r
F P
0 g
s
.58319 .20706 m .65509 .21542 L .65581 .21858 L p .693 .272 .242 r
F P
0 g
s
.65509 .21542 m .58319 .20706 L .58235 .20342 L p .693 .272 .242 r
F P
0 g
s
.43485 .21807 m .3663 .24512 L .36492 .24023 L p .475 .132 .305 r
F P
0 g
s
.36492 .24023 m .43369 .21356 L .43485 .21807 L p .475 .132 .305 r
F P
0 g
s
.36909 .25499 m .3079 .29325 L .30624 .28798 L p .391 .085 .322 r
F P
0 g
s
.30624 .28798 m .36769 .25004 L .36909 .25499 L p .391 .085 .322 r
F P
0 g
s
.83856 .60081 m .85125 .6421 L .83848 .65461 L p .933 .846 .734 r
F P
0 g
s
.85125 .6421 m .83856 .60081 L .85083 .5864 L p .933 .846 .734 r
F P
0 g
s
.65581 .21858 m .58403 .21072 L .58319 .20706 L p .693 .272 .242 r
F P
0 g
s
.50757 .20276 m .58152 .19979 L .58235 .20342 L p .625 .226 .266 r
F P
0 g
s
.58152 .19979 m .50757 .20276 L .5066 .19869 L p .625 .226 .266 r
F P
0 g
s
.43369 .21356 m .5066 .19869 L .50757 .20276 L p .553 .179 .287 r
F P
0 g
s
.50757 .20276 m .43485 .21807 L .43369 .21356 L p .553 .179 .287 r
F P
0 g
s
.81534 .71681 m .81258 .68001 L .82559 .66725 L p .942 .901 .756 r
F P
0 g
s
.43602 .2226 m .36769 .25004 L .3663 .24512 L p .475 .133 .305 r
F P
0 g
s
.3663 .24512 m .43485 .21807 L .43602 .2226 L p .475 .133 .305 r
F P
0 g
s
.50855 .20686 m .58235 .20342 L .58319 .20706 L p .625 .226 .266 r
F P
0 g
s
.58235 .20342 m .50855 .20686 L .50757 .20276 L p .625 .226 .266 r
F P
0 g
s
.3705 .25998 m .30957 .29856 L .3079 .29325 L p .391 .085 .322 r
F P
0 g
s
.3079 .29325 m .36909 .25499 L .3705 .25998 L p .391 .085 .322 r
F P
0 g
s
.43485 .21807 m .50757 .20276 L .50855 .20686 L p .553 .179 .287 r
F P
0 g
s
.50855 .20686 m .43602 .2226 L .43485 .21807 L p .553 .179 .287 r
F P
0 g
s
.43719 .22717 m .36909 .25499 L .36769 .25004 L p .475 .133 .305 r
F P
0 g
s
.36769 .25004 m .43602 .2226 L .43719 .22717 L p .475 .133 .305 r
F P
0 g
s
.82618 .61533 m .83848 .65461 L .82559 .66725 L p .933 .846 .734 r
F P
0 g
s
.83848 .65461 m .82618 .61533 L .83856 .60081 L p .933 .846 .734 r
F P
0 g
s
.50954 .21098 m .58319 .20706 L .58403 .21072 L p .625 .226 .266 r
F P
0 g
s
.58319 .20706 m .50954 .21098 L .50855 .20686 L p .625 .226 .266 r
F P
0 g
s
.43602 .2226 m .50855 .20686 L .50954 .21098 L p .553 .179 .287 r
F P
0 g
s
.41251 .82115 m .44002 .83349 L .43654 .8401 L p .541 .671 .922 r
F P
0 g
s
.43654 .8401 m .40672 .82716 L .41251 .82115 L p .541 .671 .922 r
F P
0 g
s
.4665 .84751 m .43654 .8401 L .44002 .83349 L p .562 .697 .931 r
F P
0 g
s
.83877 .5299 m .863 .57211 L .85083 .5864 L p .913 .793 .714 r
F P
0 g
s
.50954 .21098 m .43719 .22717 L .43602 .2226 L p .553 .179 .287 r
F P
0 g
s
.44002 .83349 m .46772 .84072 L .4665 .84751 L p .562 .697 .931 r
F P
0 g
s
.4953 .84966 m .4665 .84751 L .46772 .84072 L p .586 .721 .936 r
F P
0 g
s
.58403 .21072 m .51053 .21512 L .50954 .21098 L p .625 .226 .266 r
F P
0 g
s
.38648 .80363 m .41251 .82115 L .40672 .82716 L p .525 .642 .908 r
F P
0 g
s
.40672 .82716 m .37843 .80862 L .38648 .80363 L p .525 .642 .908 r
F P
0 g
s
.43837 .23176 m .3705 .25998 L .36909 .25499 L p .475 .133 .305 r
F P
0 g
s
.36909 .25499 m .43719 .22717 L .43837 .23176 L p .475 .133 .305 r
F P
0 g
s
.46772 .84072 m .49442 .84307 L .4953 .84966 L p .586 .721 .936 r
F P
0 g
s
.52172 .84701 m .4953 .84966 L .49442 .84307 L p .612 .74 .936 r
F P
0 g
s
.43719 .22717 m .50954 .21098 L .51053 .21512 L p .553 .179 .287 r
F P
0 g
s
.8137 .62998 m .82559 .66725 L .81258 .68001 L p .933 .846 .734 r
F P
0 g
s
.82559 .66725 m .8137 .62998 L .82618 .61533 L p .933 .846 .734 r
F P
0 g
s
.51053 .21512 m .43837 .23176 L .43719 .22717 L p .553 .179 .287 r
F P
0 g
s
.36323 .78108 m .38648 .80363 L .37843 .80862 L p .514 .614 .89 r
F P
0 g
s
.37843 .80862 m .35307 .78462 L .36323 .78108 L p .514 .614 .89 r
F P
0 g
s
.41835 .8151 m .44353 .82683 L .44002 .83349 L p .541 .67 .921 r
F P
0 g
s
.44002 .83349 m .41251 .82115 L .41835 .8151 L p .541 .67 .921 r
F P
0 g
s
.46772 .84072 m .44002 .83349 L .44353 .82683 L p .562 .697 .931 r
F P
0 g
s
.44353 .82683 m .46896 .83387 L .46772 .84072 L p .562 .697 .931 r
F P
0 g
s
.49442 .84307 m .46772 .84072 L .46896 .83387 L p .586 .721 .936 r
F P
0 g
s
.82748 .54617 m .85083 .5864 L .83856 .60081 L p .913 .793 .714 r
F P
0 g
s
.85083 .5864 m .82748 .54617 L .83877 .5299 L p .913 .793 .714 r
F P
0 g
s
.49442 .84307 m .519 .841 L .52172 .84701 L p .612 .74 .936 r
F P
0 g
s
.54471 .84023 m .52172 .84701 L .519 .841 L p .636 .754 .933 r
F P
0 g
s
.39459 .79861 m .41835 .8151 L .41251 .82115 L p .525 .642 .908 r
F P
0 g
s
.41251 .82115 m .38648 .80363 L .39459 .79861 L p .525 .642 .908 r
F P
0 g
s
.46896 .83387 m .49354 .83644 L .49442 .84307 L p .586 .721 .936 r
F P
0 g
s
.519 .841 m .49442 .84307 L .49354 .83644 L p .612 .74 .936 r
F P
0 g
s
.34407 .75384 m .36323 .78108 L .35307 .78462 L p .509 .587 .869 r
F P
0 g
s
.35307 .78462 m .33205 .75556 L .34407 .75384 L p .509 .587 .869 r
F P
0 g
s
.37347 .7775 m .39459 .79861 L .38648 .80363 L p .514 .614 .89 r
F P
0 g
s
.38648 .80363 m .36323 .78108 L .37347 .7775 L p .514 .614 .89 r
F P
0 g
s
.49354 .83644 m .51626 .83494 L .519 .841 L p .612 .74 .936 r
F P
0 g
s
.54048 .83508 m .519 .841 L .51626 .83494 L p .636 .754 .933 r
F P
0 g
s
.519 .841 m .54048 .83508 L .54471 .84023 L p .636 .754 .933 r
F P
0 g
s
.81258 .68001 m .80112 .64475 L .8137 .62998 L p .933 .846 .734 r
F P
0 g
s
.44706 .82012 m .4702 .82697 L .46896 .83387 L p .562 .697 .931 r
F P
0 g
s
.46896 .83387 m .44353 .82683 L .44706 .82012 L p .562 .697 .931 r
F P
0 g
s
.49354 .83644 m .46896 .83387 L .4702 .82697 L p .586 .72 .936 r
F P
0 g
s
.42423 .80899 m .44706 .82012 L .44353 .82683 L p .541 .67 .921 r
F P
0 g
s
.44353 .82683 m .41835 .8151 L .42423 .80899 L p .541 .67 .921 r
F P
0 g
s
.81611 .56255 m .83856 .60081 L .82618 .61533 L p .913 .793 .714 r
F P
0 g
s
.83856 .60081 m .81611 .56255 L .82748 .54617 L p .913 .793 .714 r
F P
0 g
s
.56339 .8301 m .54471 .84023 L .54048 .83508 L p .659 .762 .926 r
F P
0 g
s
.4702 .82697 m .49266 .82977 L .49354 .83644 L p .586 .72 .936 r
F P
0 g
s
.51626 .83494 m .49354 .83644 L .49266 .82977 L p .612 .74 .936 r
F P
0 g
s
.40278 .79354 m .42423 .80899 L .41835 .8151 L p .525 .642 .908 r
F P
0 g
s
.41835 .8151 m .39459 .79861 L .40278 .79354 L p .525 .642 .908 r
F P
0 g
s
.49266 .82977 m .5135 .82885 L .51626 .83494 L p .612 .74 .936 r
F P
0 g
s
.53623 .8299 m .51626 .83494 L .5135 .82885 L p .636 .754 .933 r
F P
0 g
s
.51626 .83494 m .53623 .8299 L .54048 .83508 L p .636 .754 .933 r
F P
0 g
s
.55807 .82606 m .54048 .83508 L .53623 .8299 L p .659 .762 .926 r
F P
0 g
s
.54048 .83508 m .55807 .82606 L .56339 .8301 L p .659 .762 .926 r
F P
0 g
s
.45063 .81335 m .47146 .82002 L .4702 .82697 L p .562 .697 .931 r
F P
0 g
s
.4702 .82697 m .44706 .82012 L .45063 .81335 L p .562 .697 .931 r
F P
0 g
s
.49266 .82977 m .4702 .82697 L .47146 .82002 L p .586 .72 .936 r
F P
0 g
s
.35618 .7521 m .37347 .7775 L .36323 .78108 L p .509 .587 .869 r
F P
0 g
s
.36323 .78108 m .34407 .75384 L .35618 .7521 L p .509 .587 .869 r
F P
0 g
s
.3838 .7739 m .40278 .79354 L .39459 .79861 L p .514 .614 .89 r
F P
0 g
s
.39459 .79861 m .37347 .7775 L .3838 .7739 L p .514 .614 .89 r
F P
0 g
s
.43017 .80283 m .45063 .81335 L .44706 .82012 L p .541 .67 .921 r
F P
0 g
s
.44706 .82012 m .42423 .80899 L .43017 .80283 L p .541 .67 .921 r
F P
0 g
s
.47146 .82002 m .49176 .82304 L .49266 .82977 L p .586 .72 .936 r
F P
0 g
s
.5135 .82885 m .49266 .82977 L .49176 .82304 L p .612 .74 .936 r
F P
0 g
s
.80466 .57904 m .82618 .61533 L .8137 .62998 L p .913 .793 .714 r
F P
0 g
s
.82618 .61533 m .80466 .57904 L .81611 .56255 L p .913 .793 .714 r
F P
0 g
s
.33022 .7225 m .34407 .75384 L .33205 .75556 L p .51 .563 .846 r
F P
0 g
s
.33205 .75556 m .31669 .72204 L .33022 .7225 L p .51 .563 .846 r
F P
0 g
s
.57711 .81759 m .56339 .8301 L .55807 .82606 L p .678 .765 .917 r
F P
0 g
s
.41103 .78843 m .43017 .80283 L .42423 .80899 L p .525 .642 .907 r
F P
0 g
s
.42423 .80899 m .40278 .79354 L .41103 .78843 L p .525 .642 .907 r
F P
0 g
s
.5135 .82885 m .53196 .8247 L .53623 .8299 L p .636 .754 .933 r
F P
0 g
s
.55273 .822 m .53623 .8299 L .53196 .8247 L p .659 .762 .926 r
F P
0 g
s
.53623 .8299 m .55273 .822 L .55807 .82606 L p .659 .762 .926 r
F P
0 g
s
.49176 .82304 m .51073 .82272 L .5135 .82885 L p .612 .74 .936 r
F P
0 g
s
.53196 .8247 m .5135 .82885 L .51073 .82272 L p .636 .754 .933 r
F P
0 g
s
.57117 .8148 m .55807 .82606 L .55273 .822 L p .678 .765 .917 r
F P
0 g
s
.55807 .82606 m .57117 .8148 L .57711 .81759 L p .678 .765 .917 r
F P
0 g
s
.80453 .4926 m .83877 .5299 L .82748 .54617 L p .887 .743 .698 r
F P
0 g
s
.45422 .80652 m .47272 .81302 L .47146 .82002 L p .562 .697 .931 r
F P
0 g
s
.47146 .82002 m .45063 .81335 L .45422 .80652 L p .562 .697 .931 r
F P
0 g
s
.49176 .82304 m .47146 .82002 L .47272 .81302 L p .586 .72 .936 r
F P
0 g
s
.36838 .75036 m .3838 .7739 L .37347 .7775 L p .509 .587 .869 r
F P
0 g
s
.37347 .7775 m .35618 .7521 L .36838 .75036 L p .509 .587 .869 r
F P
0 g
s
.47272 .81302 m .49086 .81627 L .49176 .82304 L p .586 .72 .936 r
F P
0 g
s
.51073 .82272 m .49176 .82304 L .49086 .81627 L p .612 .74 .936 r
F P
0 g
s
.43615 .79662 m .45422 .80652 L .45063 .81335 L p .541 .67 .921 r
F P
0 g
s
.45063 .81335 m .43017 .80283 L .43615 .79662 L p .541 .67 .921 r
F P
0 g
s
.39421 .77026 m .41103 .78843 L .40278 .79354 L p .514 .614 .89 r
F P
0 g
s
.40278 .79354 m .3838 .7739 L .39421 .77026 L p .514 .614 .89 r
F P
0 g
s
.51073 .82272 m .52766 .81948 L .53196 .8247 L p .636 .754 .933 r
F P
0 g
s
.54737 .81792 m .53196 .8247 L .52766 .81948 L p .659 .762 .926 r
F P
0 g
s
.53196 .8247 m .54737 .81792 L .55273 .822 L p .659 .762 .926 r
F P
0 g
s
.79313 .59564 m .8137 .62998 L .80112 .64475 L p .913 .793 .714 r
F P
0 g
s
.8137 .62998 m .79313 .59564 L .80466 .57904 L p .913 .793 .714 r
F P
0 g
s
.34385 .72295 m .35618 .7521 L .34407 .75384 L p .51 .563 .846 r
F P
0 g
s
.34407 .75384 m .33022 .7225 L .34385 .72295 L p .51 .563 .846 r
F P
0 g
s
.49086 .81627 m .50794 .81655 L .51073 .82272 L p .612 .74 .936 r
F P
0 g
s
.52766 .81948 m .51073 .82272 L .50794 .81655 L p .637 .754 .933 r
F P
0 g
s
.56522 .81199 m .55273 .822 L .54737 .81792 L p .678 .765 .917 r
F P
0 g
s
.55273 .822 m .56522 .81199 L .57117 .8148 L p .678 .765 .917 r
F P
0 g
s
.41935 .78328 m .43615 .79662 L .43017 .80283 L p .524 .642 .907 r
F P
0 g
s
.43017 .80283 m .41103 .78843 L .41935 .78328 L p .524 .642 .907 r
F P
0 g
s
.58548 .80376 m .57711 .81759 L .57117 .8148 L p .692 .764 .906 r
F P
0 g
s
.79467 .51069 m .82748 .54617 L .81611 .56255 L p .887 .743 .698 r
F P
0 g
s
.82748 .54617 m .79467 .51069 L .80453 .4926 L p .887 .743 .698 r
F P
0 g
s
.50794 .81655 m .52334 .81422 L .52766 .81948 L p .637 .754 .933 r
F P
0 g
s
.54198 .81383 m .52766 .81948 L .52334 .81422 L p .659 .762 .926 r
F P
0 g
s
.52766 .81948 m .54198 .81383 L .54737 .81792 L p .659 .762 .926 r
F P
0 g
s
.32283 .68786 m .33022 .7225 L .31669 .72204 L p .517 .543 .822 r
F P
0 g
s
.31669 .72204 m .30822 .68494 L .32283 .68786 L p .517 .543 .822 r
F P
0 g
s
.55926 .80918 m .54737 .81792 L .54198 .81383 L p .678 .765 .917 r
F P
0 g
s
.54737 .81792 m .55926 .80918 L .56522 .81199 L p .678 .765 .917 r
F P
0 g
s
.38068 .74859 m .39421 .77026 L .3838 .7739 L p .509 .587 .869 r
F P
0 g
s
.3838 .7739 m .36838 .75036 L .38068 .74859 L p .509 .587 .869 r
F P
0 g
s
.57944 .80225 m .57117 .8148 L .56522 .81199 L p .692 .764 .906 r
F P
0 g
s
.57117 .8148 m .57944 .80225 L .58548 .80376 L p .692 .764 .906 r
F P
0 g
s
.4047 .7666 m .41935 .78328 L .41103 .78843 L p .514 .614 .89 r
F P
0 g
s
.41103 .78843 m .39421 .77026 L .4047 .7666 L p .514 .614 .89 r
F P
0 g
s
.35757 .72341 m .36838 .75036 L .35618 .7521 L p .51 .563 .846 r
F P
0 g
s
.35618 .7521 m .34385 .72295 L .35757 .72341 L p .51 .563 .846 r
F P
0 g
s
.80112 .64475 m .78153 .61237 L .79313 .59564 L p .913 .793 .714 r
F P
0 g
s
.52334 .81422 m .53657 .80972 L .54198 .81383 L p .659 .762 .926 r
F P
0 g
s
.55327 .80636 m .54198 .81383 L .53657 .80972 L p .678 .765 .917 r
F P
0 g
s
.54198 .81383 m .55327 .80636 L .55926 .80918 L p .678 .765 .917 r
F P
0 g
s
.5734 .80073 m .56522 .81199 L .55926 .80918 L p .693 .764 .906 r
F P
0 g
s
.56522 .81199 m .5734 .80073 L .57944 .80225 L p .692 .764 .906 r
F P
0 g
s
.78475 .52888 m .81611 .56255 L .80466 .57904 L p .887 .744 .698 r
F P
0 g
s
.81611 .56255 m .78475 .52888 L .79467 .51069 L p .887 .743 .698 r
F P
0 g
s
.33753 .69079 m .34385 .72295 L .33022 .7225 L p .517 .543 .822 r
F P
0 g
s
.33022 .7225 m .32283 .68786 L .33753 .69079 L p .517 .543 .822 r
F P
0 g
s
.56734 .79921 m .55926 .80918 L .55327 .80636 L p .693 .764 .906 r
F P
0 g
s
.55926 .80918 m .56734 .79921 L .5734 .80073 L p .693 .764 .906 r
F P
0 g
s
.53657 .80972 m .54727 .80352 L .55327 .80636 L p .678 .765 .917 r
F P
0 g
s
.39307 .74682 m .4047 .7666 L .39421 .77026 L p .509 .587 .869 r
F P
0 g
s
.39421 .77026 m .38068 .74859 L .39307 .74682 L p .509 .587 .869 r
F P
0 g
s
.58845 .78973 m .58548 .80376 L .57944 .80225 L p .702 .758 .896 r
F P
0 g
s
.37139 .72387 m .38068 .74859 L .36838 .75036 L p .51 .563 .846 r
F P
0 g
s
.36838 .75036 m .35757 .72341 L .37139 .72387 L p .51 .563 .846 r
F P
0 g
s
.56127 .79769 m .55327 .80636 L .54727 .80352 L p .693 .764 .906 r
F P
0 g
s
.55327 .80636 m .56127 .79769 L .56734 .79921 L p .693 .764 .906 r
F P
0 g
s
.58281 .78942 m .57944 .80225 L .5734 .80073 L p .702 .758 .895 r
F P
0 g
s
.57944 .80225 m .58281 .78942 L .58845 .78973 L p .702 .758 .896 r
F P
0 g
s
.77478 .54717 m .80466 .57904 L .79313 .59564 L p .887 .744 .698 r
F P
0 g
s
.80466 .57904 m .77478 .54717 L .78475 .52888 L p .887 .744 .698 r
F P
0 g
s
.57718 .7891 m .5734 .80073 L .56734 .79921 L p .703 .758 .895 r
F P
0 g
s
.5734 .80073 m .57718 .7891 L .58281 .78942 L p .702 .758 .895 r
F P
0 g
s
.54727 .80352 m .55519 .79617 L .56127 .79769 L p .693 .764 .906 r
F P
0 g
s
.35232 .69374 m .35757 .72341 L .34385 .72295 L p .517 .543 .822 r
F P
0 g
s
.34385 .72295 m .33753 .69079 L .35232 .69374 L p .517 .543 .822 r
F P
0 g
s
.32285 .65095 m .32283 .68786 L .30822 .68494 L p .53 .528 .797 r
F P
0 g
s
.30822 .68494 m .30768 .64536 L .32285 .65095 L p .53 .528 .797 r
F P
0 g
s
.57154 .78879 m .56734 .79921 L .56127 .79769 L p .703 .758 .895 r
F P
0 g
s
.56734 .79921 m .57154 .78879 L .57718 .7891 L p .703 .758 .895 r
F P
0 g
s
.56589 .78848 m .56127 .79769 L .55519 .79617 L p .703 .758 .894 r
F P
0 g
s
.56127 .79769 m .56589 .78848 L .57154 .78879 L p .703 .758 .895 r
F P
0 g
s
.38531 .72434 m .39307 .74682 L .38068 .74859 L p .51 .563 .846 r
F P
0 g
s
.38068 .74859 m .37139 .72387 L .38531 .72434 L p .51 .563 .846 r
F P
0 g
s
.55519 .79617 m .56024 .78816 L .56589 .78848 L p .703 .758 .894 r
F P
0 g
s
.5863 .77664 m .58845 .78973 L .58281 .78942 L p .707 .75 .885 r
F P
0 g
s
.76475 .56557 m .79313 .59564 L .78153 .61237 L p .887 .744 .698 r
F P
0 g
s
.79313 .59564 m .76475 .56557 L .77478 .54717 L p .887 .744 .698 r
F P
0 g
s
.76159 .46183 m .80453 .4926 L .79467 .51069 L p .857 .699 .685 r
F P
0 g
s
.3381 .65657 m .33753 .69079 L .32283 .68786 L p .53 .528 .797 r
F P
0 g
s
.32283 .68786 m .32285 .65095 L .3381 .65657 L p .53 .528 .797 r
F P
0 g
s
.58153 .77732 m .58281 .78942 L .57718 .7891 L p .707 .75 .885 r
F P
0 g
s
.58281 .78942 m .58153 .77732 L .5863 .77664 L p .707 .75 .885 r
F P
0 g
s
.36721 .69671 m .37139 .72387 L .35757 .72341 L p .517 .543 .822 r
F P
0 g
s
.35757 .72341 m .35232 .69374 L .36721 .69671 L p .517 .543 .822 r
F P
0 g
s
.57675 .77799 m .57718 .7891 L .57154 .78879 L p .708 .749 .884 r
F P
0 g
s
.57718 .7891 m .57675 .77799 L .58153 .77732 L p .707 .75 .885 r
F P
0 g
s
.57197 .77867 m .57154 .78879 L .56589 .78848 L p .708 .749 .884 r
F P
0 g
s
.57154 .78879 m .57197 .77867 L .57675 .77799 L p .708 .749 .884 r
F P
0 g
s
.5672 .77935 m .56589 .78848 L .56024 .78816 L p .708 .749 .883 r
F P
0 g
s
.56589 .78848 m .5672 .77935 L .57197 .77867 L p .708 .749 .884 r
F P
0 g
s
.56024 .78816 m .56242 .78002 L .5672 .77935 L p .708 .749 .883 r
F P
0 g
s
.75368 .48147 m .79467 .51069 L .78475 .52888 L p .857 .699 .685 r
F P
0 g
s
.79467 .51069 m .75368 .48147 L .76159 .46183 L p .857 .699 .685 r
F P
0 g
s
.35344 .66222 m .35232 .69374 L .33753 .69079 L p .53 .528 .797 r
F P
0 g
s
.33753 .69079 m .3381 .65657 L .35344 .66222 L p .53 .528 .797 r
F P
0 g
s
.78153 .61237 m .75466 .58407 L .76475 .56557 L p .887 .744 .698 r
F P
0 g
s
.38219 .6997 m .38531 .72434 L .37139 .72387 L p .517 .543 .821 r
F P
0 g
s
.37139 .72387 m .36721 .69671 L .38219 .6997 L p .517 .543 .821 r
F P
0 g
s
.33102 .613 m .32285 .65095 L .30768 .64536 L p .547 .517 .774 r
F P
0 g
s
.30768 .64536 m .31586 .60461 L .33102 .613 L p .547 .517 .774 r
F P
0 g
s
.57975 .76556 m .5863 .77664 L .58153 .77732 L p .708 .74 .876 r
F P
0 g
s
.57616 .76689 m .58153 .77732 L .57675 .77799 L p .708 .74 .876 r
F P
0 g
s
.58153 .77732 m .57616 .76689 L .57975 .76556 L p .708 .74 .876 r
F P
0 g
s
.57258 .76823 m .57675 .77799 L .57197 .77867 L p .708 .739 .876 r
F P
0 g
s
.57675 .77799 m .57258 .76823 L .57616 .76689 L p .708 .74 .876 r
F P
0 g
s
.569 .76956 m .57197 .77867 L .5672 .77935 L p .708 .739 .875 r
F P
0 g
s
.57197 .77867 m .569 .76956 L .57258 .76823 L p .708 .739 .876 r
F P
0 g
s
.56542 .77089 m .5672 .77935 L .56242 .78002 L p .709 .738 .874 r
F P
0 g
s
.5672 .77935 m .56542 .77089 L .569 .76956 L p .708 .739 .875 r
F P
0 g
s
.56242 .78002 m .56184 .77222 L .56542 .77089 L p .709 .738 .874 r
F P
0 g
s
.74575 .5012 m .78475 .52888 L .77478 .54717 L p .857 .699 .685 r
F P
0 g
s
.78475 .52888 m .74575 .5012 L .75368 .48147 L p .857 .699 .685 r
F P
0 g
s
.36886 .6679 m .36721 .69671 L .35232 .69374 L p .53 .528 .797 r
F P
0 g
s
.35232 .69374 m .35344 .66222 L .36886 .6679 L p .53 .528 .797 r
F P
0 g
s
.56184 .77222 m .55866 .76521 L .5609 .76365 L p .705 .727 .868 r
F P
0 g
s
.5609 .76365 m .56542 .77089 L .56184 .77222 L p .705 .727 .868 r
F P
0 g
s
.34624 .62142 m .3381 .65657 L .32285 .65095 L p .547 .517 .774 r
F P
0 g
s
.32285 .65095 m .33102 .613 L .34624 .62142 L p .547 .517 .774 r
F P
0 g
s
.56542 .77089 m .5609 .76365 L .56314 .76209 L p .705 .728 .868 r
F P
0 g
s
.56314 .76209 m .569 .76956 L .56542 .77089 L p .705 .728 .868 r
F P
0 g
s
.569 .76956 m .56314 .76209 L .56539 .76052 L p .705 .729 .869 r
F P
0 g
s
.56539 .76052 m .57258 .76823 L .569 .76956 L p .705 .729 .869 r
F P
0 g
s
.57258 .76823 m .56539 .76052 L .56763 .75895 L p .705 .73 .87 r
F P
0 g
s
.56763 .75895 m .57616 .76689 L .57258 .76823 L p .705 .73 .87 r
F P
0 g
s
.57616 .76689 m .56763 .75895 L .56988 .75739 L p .705 .73 .87 r
F P
0 g
s
.56988 .75739 m .57975 .76556 L .57616 .76689 L p .705 .73 .87 r
F P
0 g
s
.73777 .52102 m .77478 .54717 L .76475 .56557 L p .857 .7 .685 r
F P
0 g
s
.77478 .54717 m .73777 .52102 L .74575 .5012 L p .857 .699 .685 r
F P
0 g
s
.55866 .76521 m .55307 .75944 L .55409 .75809 L p .698 .718 .865 r
F P
0 g
s
.55409 .75809 m .5609 .76365 L .55866 .76521 L p .698 .718 .865 r
F P
0 g
s
.38436 .67362 m .38219 .6997 L .36721 .69671 L p .53 .528 .797 r
F P
0 g
s
.36721 .69671 m .36886 .6679 L .38436 .67362 L p .53 .528 .797 r
F P
0 g
s
.5609 .76365 m .55409 .75809 L .55512 .75675 L p .699 .719 .865 r
F P
0 g
s
.55512 .75675 m .56314 .76209 L .5609 .76365 L p .699 .719 .865 r
F P
0 g
s
.36154 .62988 m .35344 .66222 L .3381 .65657 L p .547 .517 .774 r
F P
0 g
s
.3381 .65657 m .34624 .62142 L .36154 .62988 L p .547 .517 .774 r
F P
0 g
s
.56314 .76209 m .55512 .75675 L .55614 .7554 L p .699 .72 .866 r
F P
0 g
s
.55614 .7554 m .56539 .76052 L .56314 .76209 L p .699 .72 .866 r
F P
0 g
s
.56539 .76052 m .55614 .7554 L .55716 .75405 L p .699 .721 .866 r
F P
0 g
s
.55716 .75405 m .56763 .75895 L .56539 .76052 L p .699 .721 .866 r
F P
0 g
s
.56763 .75895 m .55716 .75405 L .55819 .7527 L p .699 .722 .867 r
F P
0 g
s
.55819 .7527 m .56988 .75739 L .56763 .75895 L p .699 .722 .867 r
F P
0 g
s
.72976 .54093 m .76475 .56557 L .75466 .58407 L p .857 .7 .685 r
F P
0 g
s
.76475 .56557 m .72976 .54093 L .73777 .52102 L p .857 .7 .685 r
F P
0 g
s
.55307 .75944 m .54534 .75541 L .54554 .75465 L p .691 .713 .866 r
F P
0 g
s
.54554 .75465 m .55409 .75809 L .55307 .75944 L p .691 .713 .866 r
F P
0 g
s
.55409 .75809 m .54554 .75465 L .54574 .75389 L p .691 .714 .866 r
F P
0 g
s
.54574 .75389 m .55512 .75675 L .55409 .75809 L p .691 .714 .866 r
F P
0 g
s
.7117 .43897 m .76159 .46183 L .75368 .48147 L p .826 .66 .677 r
F P
0 g
s
.55512 .75675 m .54574 .75389 L .54595 .75313 L p .691 .714 .866 r
F P
0 g
s
.54595 .75313 m .55614 .7554 L .55512 .75675 L p .691 .714 .866 r
F P
0 g
s
.34774 .57541 m .33102 .613 L .31586 .60461 L p .567 .511 .752 r
F P
0 g
s
.31586 .60461 m .33321 .56423 L .34774 .57541 L p .567 .511 .752 r
F P
0 g
s
.55614 .7554 m .54595 .75313 L .54615 .75238 L p .692 .715 .867 r
F P
0 g
s
.54615 .75238 m .55716 .75405 L .55614 .7554 L p .692 .715 .867 r
F P
0 g
s
.53594 .75369 m .54554 .75465 L .54534 .75541 L closepath p .685 .713 .87 r
F P
0 g
s
.37691 .63838 m .36886 .6679 L .35344 .66222 L p .547 .517 .774 r
F P
0 g
s
.35344 .66222 m .36154 .62988 L .37691 .63838 L p .547 .517 .774 r
F P
0 g
s
.53594 .75369 m .54574 .75389 L .54554 .75465 L closepath p .685 .713 .87 r
F P
0 g
s
.55716 .75405 m .54615 .75238 L .54635 .75162 L p .692 .715 .867 r
F P
0 g
s
.54635 .75162 m .55819 .7527 L .55716 .75405 L p .692 .715 .867 r
F P
0 g
s
.53594 .75369 m .54595 .75313 L .54574 .75389 L closepath p .685 .713 .87 r
F P
0 g
s
.53594 .75369 m .54615 .75238 L .54595 .75313 L closepath p .685 .713 .87 r
F P
0 g
s
.53594 .75369 m .54635 .75162 L .54615 .75238 L p .685 .713 .87 r
F P
0 g
s
.75466 .58407 m .72172 .56092 L .72976 .54093 L p .857 .7 .685 r
F P
0 g
s
.70623 .45977 m .75368 .48147 L .74575 .5012 L p .826 .66 .677 r
F P
0 g
s
.75368 .48147 m .70623 .45977 L .7117 .43897 L p .826 .66 .677 r
F P
0 g
s
.36233 .58664 m .34624 .62142 L .33102 .613 L p .567 .511 .752 r
F P
0 g
s
.33102 .613 m .34774 .57541 L .36233 .58664 L p .567 .511 .752 r
F P
0 g
s
.39234 .64693 m .38436 .67362 L .36886 .6679 L p .547 .517 .774 r
F P
0 g
s
.36886 .6679 m .37691 .63838 L .39234 .64693 L p .547 .517 .774 r
F P
0 g
s
.70073 .48065 m .74575 .5012 L .73777 .52102 L p .826 .66 .677 r
F P
0 g
s
.74575 .5012 m .70073 .48065 L .70623 .45977 L p .826 .66 .677 r
F P
0 g
s
.37698 .59792 m .36154 .62988 L .34624 .62142 L p .567 .511 .752 r
F P
0 g
s
.34624 .62142 m .36233 .58664 L .37698 .59792 L p .567 .511 .752 r
F P
0 g
s
.69522 .5016 m .73777 .52102 L .72976 .54093 L p .827 .661 .677 r
F P
0 g
s
.73777 .52102 m .69522 .5016 L .70073 .48065 L p .826 .66 .677 r
F P
0 g
s
.39168 .60923 m .37691 .63838 L .36154 .62988 L p .567 .511 .752 r
F P
0 g
s
.36154 .62988 m .37698 .59792 L .39168 .60923 L p .567 .511 .752 r
F P
0 g
s
.37308 .53973 m .34774 .57541 L .33321 .56423 L p .591 .51 .732 r
F P
0 g
s
.33321 .56423 m .35978 .52585 L .37308 .53973 L p .591 .51 .732 r
F P
0 g
s
.68968 .52262 m .72976 .54093 L .72172 .56092 L p .827 .661 .677 r
F P
0 g
s
.72976 .54093 m .68968 .52262 L .69522 .5016 L p .827 .661 .677 r
F P
0 g
s
.40643 .62059 m .39234 .64693 L .37691 .63838 L p .567 .511 .752 r
F P
0 g
s
.37691 .63838 m .39168 .60923 L .40643 .62059 L p .567 .511 .752 r
F P
0 g
s
.38642 .55365 m .36233 .58664 L .34774 .57541 L p .591 .51 .732 r
F P
0 g
s
.34774 .57541 m .37308 .53973 L .38642 .55365 L p .591 .51 .732 r
F P
0 g
s
.65701 .42501 m .7117 .43897 L .70623 .45977 L p .795 .626 .672 r
F P
0 g
s
.72172 .56092 m .68413 .54372 L .68968 .52262 L p .827 .661 .677 r
F P
0 g
s
.3998 .56761 m .37698 .59792 L .36233 .58664 L p .591 .51 .732 r
F P
0 g
s
.36233 .58664 m .38642 .55365 L .3998 .56761 L p .591 .51 .732 r
F P
0 g
s
.65433 .44649 m .70623 .45977 L .70073 .48065 L p .795 .626 .672 r
F P
0 g
s
.70623 .45977 m .65433 .44649 L .65701 .42501 L p .795 .626 .672 r
F P
0 g
s
.41322 .58161 m .39168 .60923 L .37698 .59792 L p .591 .51 .732 r
F P
0 g
s
.37698 .59792 m .3998 .56761 L .41322 .58161 L p .591 .51 .732 r
F P
0 g
s
.65164 .46803 m .70073 .48065 L .69522 .5016 L p .795 .626 .672 r
F P
0 g
s
.70073 .48065 m .65164 .46803 L .65433 .44649 L p .795 .626 .672 r
F P
0 g
s
.40665 .50754 m .37308 .53973 L .35978 .52585 L p .616 .514 .714 r
F P
0 g
s
.35978 .52585 m .39517 .49121 L .40665 .50754 L p .616 .514 .714 r
F P
0 g
s
.42668 .59566 m .40643 .62059 L .39168 .60923 L p .591 .51 .732 r
F P
0 g
s
.39168 .60923 m .41322 .58161 L .42668 .59566 L p .591 .51 .732 r
F P
0 g
s
.64894 .48963 m .69522 .5016 L .68968 .52262 L p .795 .626 .672 r
F P
0 g
s
.69522 .5016 m .64894 .48963 L .65164 .46803 L p .795 .626 .672 r
F P
0 g
s
.41816 .5239 m .38642 .55365 L .37308 .53973 L p .616 .514 .714 r
F P
0 g
s
.37308 .53973 m .40665 .50754 L .41816 .5239 L p .616 .514 .714 r
F P
0 g
s
.64623 .51128 m .68968 .52262 L .68413 .54372 L p .795 .626 .672 r
F P
0 g
s
.68968 .52262 m .64623 .51128 L .64894 .48963 L p .795 .626 .672 r
F P
0 g
s
.4297 .54031 m .3998 .56761 L .38642 .55365 L p .616 .514 .714 r
F P
0 g
s
.38642 .55365 m .41816 .5239 L .4297 .54031 L p .616 .514 .714 r
F P
0 g
s
.5999 .42052 m .65701 .42501 L .65433 .44649 L p .764 .597 .671 r
F P
0 g
s
.44127 .55675 m .41322 .58161 L .3998 .56761 L p .616 .514 .714 r
F P
0 g
s
.3998 .56761 m .4297 .54031 L .44127 .55675 L p .616 .514 .714 r
F P
0 g
s
.68413 .54372 m .64352 .533 L .64623 .51128 L p .795 .626 .672 r
F P
0 g
s
.60024 .44212 m .65433 .44649 L .65164 .46803 L p .764 .597 .671 r
F P
0 g
s
.65433 .44649 m .60024 .44212 L .5999 .42052 L p .764 .597 .671 r
F P
0 g
s
.44764 .48039 m .40665 .50754 L .39517 .49121 L p .644 .522 .7 r
F P
0 g
s
.39517 .49121 m .43848 .46197 L .44764 .48039 L p .644 .522 .7 r
F P
0 g
s
.45286 .57324 m .42668 .59566 L .41322 .58161 L p .616 .514 .714 r
F P
0 g
s
.41322 .58161 m .44127 .55675 L .45286 .57324 L p .616 .514 .714 r
F P
0 g
s
.60058 .46376 m .65164 .46803 L .64894 .48963 L p .764 .597 .671 r
F P
0 g
s
.65164 .46803 m .60058 .46376 L .60024 .44212 L p .764 .597 .671 r
F P
0 g
s
.45681 .49884 m .41816 .5239 L .40665 .50754 L p .644 .522 .7 r
F P
0 g
s
.40665 .50754 m .44764 .48039 L .45681 .49884 L p .644 .522 .7 r
F P
0 g
s
.60092 .48545 m .64894 .48963 L .64623 .51128 L p .764 .597 .671 r
F P
0 g
s
.64894 .48963 m .60092 .48545 L .60058 .46376 L p .764 .597 .671 r
F P
0 g
s
.466 .51733 m .4297 .54031 L .41816 .5239 L p .644 .522 .7 r
F P
0 g
s
.41816 .5239 m .45681 .49884 L .466 .51733 L p .644 .522 .7 r
F P
0 g
s
.54287 .42555 m .5999 .42052 L .60024 .44212 L p .732 .572 .674 r
F P
0 g
s
.60126 .50718 m .64623 .51128 L .64352 .533 L p .764 .597 .671 r
F P
0 g
s
.64623 .51128 m .60126 .50718 L .60092 .48545 L p .764 .597 .671 r
F P
0 g
s
.47521 .53586 m .44127 .55675 L .4297 .54031 L p .644 .522 .7 r
F P
0 g
s
.4297 .54031 m .466 .51733 L .47521 .53586 L p .644 .522 .7 r
F P
0 g
s
.49475 .4597 m .44764 .48039 L .43848 .46197 L p .672 .534 .688 r
F P
0 g
s
.43848 .46197 m .48833 .43967 L .49475 .4597 L p .672 .534 .688 r
F P
0 g
s
.5463 .44667 m .60024 .44212 L .60058 .46376 L p .733 .572 .674 r
F P
0 g
s
.60024 .44212 m .5463 .44667 L .54287 .42555 L p .732 .572 .674 r
F P
0 g
s
.64352 .533 m .60161 .52896 L .60126 .50718 L p .764 .597 .671 r
F P
0 g
s
.48444 .55442 m .45286 .57324 L .44127 .55675 L p .644 .522 .7 r
F P
0 g
s
.44127 .55675 m .47521 .53586 L .48444 .55442 L p .644 .522 .7 r
F P
0 g
s
.50119 .47978 m .45681 .49884 L .44764 .48039 L p .672 .534 .688 r
F P
0 g
s
.44764 .48039 m .49475 .4597 L .50119 .47978 L p .672 .534 .688 r
F P
0 g
s
.54974 .46782 m .60058 .46376 L .60092 .48545 L p .733 .572 .674 r
F P
0 g
s
.60058 .46376 m .54974 .46782 L .5463 .44667 L p .733 .572 .674 r
F P
0 g
s
.48833 .43967 m .54287 .42555 L .5463 .44667 L p .702 .551 .679 r
F P
0 g
s
.5463 .44667 m .49475 .4597 L .48833 .43967 L p .702 .551 .679 r
F P
0 g
s
.50764 .49989 m .466 .51733 L .45681 .49884 L p .672 .534 .688 r
F P
0 g
s
.45681 .49884 m .50119 .47978 L .50764 .49989 L p .672 .534 .688 r
F P
0 g
s
.55318 .48901 m .60092 .48545 L .60126 .50718 L p .733 .572 .674 r
F P
0 g
s
.60092 .48545 m .55318 .48901 L .54974 .46782 L p .733 .572 .674 r
F P
0 g
s
.49475 .4597 m .5463 .44667 L .54974 .46782 L p .702 .551 .679 r
F P
0 g
s
.54974 .46782 m .50119 .47978 L .49475 .4597 L p .702 .551 .679 r
F P
0 g
s
.5141 .52003 m .47521 .53586 L .466 .51733 L p .672 .534 .688 r
F P
0 g
s
.466 .51733 m .50764 .49989 L .5141 .52003 L p .672 .534 .688 r
F P
0 g
s
.55663 .51024 m .60126 .50718 L .60161 .52896 L p .733 .572 .674 r
F P
0 g
s
.60126 .50718 m .55663 .51024 L .55318 .48901 L p .733 .572 .674 r
F P
0 g
s
.50119 .47978 m .54974 .46782 L .55318 .48901 L p .702 .551 .679 r
F P
0 g
s
.55318 .48901 m .50764 .49989 L .50119 .47978 L p .702 .551 .679 r
F P
0 g
s
.52057 .54021 m .48444 .55442 L .47521 .53586 L p .673 .534 .688 r
F P
0 g
s
.47521 .53586 m .5141 .52003 L .52057 .54021 L p .673 .534 .688 r
F P
0 g
s
.60161 .52896 m .56009 .5315 L .55663 .51024 L p .733 .572 .674 r
F P
0 g
s
.50764 .49989 m .55318 .48901 L .55663 .51024 L p .702 .551 .679 r
F P
0 g
s
.55663 .51024 m .5141 .52003 L .50764 .49989 L p .702 .551 .679 r
F P
0 g
s
.5141 .52003 m .55663 .51024 L .56009 .5315 L p .702 .551 .679 r
F P
0 g
s
.56009 .5315 m .52057 .54021 L .5141 .52003 L p .702 .551 .679 r
F P
0 g
s
.25 Mabswid
.1516 .09701 m
0 .41812 L
s
0 .41812 m
.96125 .29266 L
s
.96125 .29266 m
.84177 0 L
s
.84177 0 m
.1516 .09701 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[26]:=",
  ImageSize->{257, 289.375},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004100018@P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l70PL00=@7100Y1`00c@L700<700<01`L0
9`L00<H71`0:1`80:0L00;l71`0A1`80:0L00;P71`0H1`0300L002L7002a1`L07`L00`07000W1`00
Z@L802H700<01`009`L00:871`0_1`0300L002H7002K1`L0=PL00`07000V1`00U0L703d700@01`L0
9@L008d71`141`0400L702D700261`L0B`L01@071`L002@7001o1`L0DPL01@071`L002@7001h1`L0
F@L01@071`L002@7001a1`L0H0L00`071`021`0300L70247001Z1`L0I`L00`071`021`0300L70247
001S1`L0KPL00`071`031`0300L70207001L1`L0M@L00`071`031`0300L70207001E1`L0O0L00`07
1`031`0300L70207001=1`P0P`L00`071`041`0300L701l700161`L0S0L00`071`031`0300L701l7
000o1`L0T`L00`071`041`0300L701h7000h1`L0VPL00`071`041`0300L701h7000a1`L0X@L00`07
1`051`0300L701d7000Z1`L0Z0L00`071`051`0300L701d7000V1`@0[`L00`071`051`0300L701d7
000V1`0300L70;0700<01`L01PL00`071`0L1`009@L20;8700<01`L01PL00`071`0L1`009@L00`07
002a1`0300L700L700<01`L06`L002@700@01`L0/@L00`071`071`0300L701/7000T1`0400L70;47
00<01`L01`L00`071`0K1`008`L01@071`L00;4700<01`L020L00`071`0J1`008`L00`071`021`03
00L70:h700<01`L020L00`071`0J1`008PL00`071`031`0300L70:l700<01`L020L00`071`0I1`00
8PL00`071`031`0300L70:l700<01`L020L00`071`0I1`008@L00`071`041`0300L70:l700<01`L0
20L00`071`0I1`008@L00`071`051`0300L70:h700<01`L02@L00`071`0H1`0080L00`071`061`03
00L705H72`1=1`0300L700T700<01`L060L0020700<01`L01PL00`071`1<1`X000InW7jLOY`40003
OP0000H0A@L00`071`0:1`0300L701L7000P1`0300L700H700<01`L0B0L40003W00000H000NGW9NL
UibG00<000=n00000`0017jLOY`5040700<01`L02PL00`071`0G1`007`L00`071`081`0300L704<7
10001INL002G00P000JLOYanW7h40003U`00008000AnUgjG1@000gh0000203`700<01`L02`L00`07
1`0F1`007`L00`071`081`0300L703l7100019`00002W08000>L00001P002YNLUibGW02LOY`40004
OYanW0T000>GOP000`0h1`0300L700/700<01`L05PL001h700<01`L02@L00`071`0l1`L000JG0000
Ui`20003U`0001<000AnUgjG1@02OP<000=n00000gh403@700<01`L02`L00`071`0F1`007PL00`07
1`091`0300L703X70P0019`009`700:L0P000i`0000@0003OYan00P000>G00000P001GjG001n0080
00BGOYMn0`0a1`0300L700`700<01`L05@L001d700<01`L02`L00`071`0f1`<000Be0;FF2P000i`0
000E1`P000=n00000P000gh00004OPH000AnN@00;`L00`071`0<1`0300L701D7000M1`0300L700/7
00<01`L0=0L20007]@00W;D0]@020005W00009`00`0P1`<000NGOP00UgjG00@000=i00000P000gii
OP0202h700<01`L030L00`071`0D1`0070L00`071`0<1`0300L703471P001kD009H0U[D01@0X1`80
0gh30005OP00NGh010000gUnN@0202`700<01`L030L00`071`0D1`0070L00`071`0<1`0300L702l7
0P0019`00;D50004]I`0W080;`L30004OP00N@8000=nN@000`000giiOP0202X700<01`L030L00`07
1`0D1`006`L00`071`0>1`0300L702`70P001;D009H20004]@00UP8000>e00L0<`L20005OWT007h0
0P0017Un07h20003NGii0080:0L00`071`0=1`0300L701<7000K1`0300L700h700<01`L0:PL20004
]@00]@8000>L00000P000kD0000R1``02PL20005NGh0OWT00P002Gii07T0NGiiOP0202H700<01`L0
3@L00`071`0C1`006PL00`071`0?1`0300L702T700D0]K@0]0020005]@000;D00`0H1ah01PL20009
OP1nNGh0OWUn00D000=iN0009@L00`071`0>1`0300L70187000J1`0300L700l700<01`L09`L20006
]@2e]@2e0`000i`0000F1bT00PL20005N@1iOWT00`000gT0N@020004OWT002<700<01`L03PL00`07
1`0B1`006@L00`071`0A1`0300L702D700P0]000]@2e]08000>e000050L40003Q@0000h000>S0000
7P001GQi07Uh008000Ii07UhN@0R1`0300L700h700<01`L04PL001T700<01`L04@L00`071`0T1`04
0;D0]@8000>e00000[D00`071`0?1`h000>50000:`0027Un001iOP1n0P000gUn000Q1`0300L700l7
00<01`L04@L001T700<01`L04@L00`071`0R1`8000Fd0;@0]0030003]00700h710000hX0000j0009
N7Uh07Qi07Uh008000=i00L07`L00`071`0?1`0300L70147000H1`0300L7018700<01`L08@L00`2e
00030005]@00]@003PL30005Q@00XhD010000j<0000C00h75`002WT0NGh0OWUn07h201h700<01`L0
40L00`071`0@1`0060L00`071`0C1`0300L701h70P001KD0]@2e00D030L30003K8D000<000>5RP00
20000hX0000202<74@002WQi07UhN@1iN00M1`0300L7010700<01`L040L001L700<01`L050L00`07
1`0M1`030;D000<000>e00000P0<1`8000>5RP001P000j>5000902d73P0017UnNGh401`700<01`L0
4@L00`071`0?1`005`L00`071`0D1`0300L701`700<0]0000P000k@0000200/70`000hE/00080003
QHZ500@0=PL>0005NGP0N00070L00`071`0@1`0300L700l7000F1`0300L701D700<01`L06`L01P2d
0;@0]0<02PL30003RXD000T000>:Q@000P0m1`/000IiUgT0N@0K1`0300L7010700<01`L03`L001H7
00<01`L05PL00`071`0I1`060;@0]02d0`091`8000>5K0002P000fb500131`0307T000P000EiN01h
000J1`0300L7014700<01`L03PL001D700<01`L05`L00`071`0H1`060;@0]02d0`081`<000>:0000
0`000hX000020004QHX004H700D0UgT0N@070004N@1i01T700<01`L04@L00`071`0>1`005@L00`07
1`0G1`0300L701L700H0]02d0;@300P700<0QF`010000hD000030004QF`004T700H0NGP0N7T70003
NGP001P700<01`L04PL00`071`0=1`0050L00`071`0H1`0300L701H71@001;@000071`8000>:0000
0P000hD000030004Q@0004`700H0UgT0NIL901P700<01`L04PL00`071`0=1`0050L00`071`0I1`03
00L701@70P000kD0]@0400H70P000f`000020003Q@00008000A/Q@00D0L0201iN01hN@1i1`0G1`03
00L7018700<01`L03@L001<700<01`L06PL00`071`0C1`8000P700L01`0700D700L0RP00K01/00@0
00B:Q@00D`L02GVGNIMiUgVGN@0701D700<01`L04`L00`071`0<1`004`L00`071`0J1`0300L70187
0P002;D0]@2e0;D010L20003Q@25008000=/00000P000hD0001F1a0050L00`071`0C1`0300L700`7
000C1`0300L701X700<01`L04@L200081`0700L01`041`0306`000<000J:0000RV`205T700/0TWP0
N980TWP0N00501<700<01`L050L00`071`0;1`004PL00`071`0L1`0300L700l71P001;D0]@031`@0
00=/06`010000hD0001K1`0<07Qi07Uh07Qi07Uh1@0B1`0300L701@700<01`L02`L0018700<01`L0
70L00`071`0>1`8000D700L01`030087100018X006`30003K8X005h700h0N980TWP0N980TWP0N0@0
40L00`071`0E1`0300L700X7000A1`0300L701d700<01`L03PL30006bP3:0<X00PL30004K8D0Q@<0
00B5K000G`L03`1iN01hN@1iN01hN@1iN00400l700<01`L05@L00`071`0:1`004@L00`071`0M1`03
00L700d70P002K@0]02d0;@01`020005RV`008X00`000hY/001R1`0>099h07RB099h07RB098500l7
00<01`L050L00`071`0:1`0040L00`071`0O1`0300L700/70P002K@0]02d0;@01`050006K00006b5
0P1S1`0A07Qi07Uh07Qi07Uh07Qi00L00`0>1`0300L701D700<01`L02@L0010700<01`L07`L00`07
1`0;1`<000Fd0;@0]0040003K@2:00@000=/RP00IPL04I9hTP2BN01hTWRBN01hTP0700<03@L00`07
1`0E1`0300L700T7000?1`0300L7020700<01`L02PL30005]02d0;@00P001F`0K01/00<000B5K000
90L=03H70P003WP0N7T0NGP0N7T0NGP00PL400/700<01`L05PL00`071`081`003`L00`071`0P1`03
00L700T70P003@L01`0700L0K@1]06d00P0016b:K00S1`<000B_0:n`10001Jn`[`2_0080=0L00`1h
00060008N99hTWRBN0031`<02PL00`071`0C1`@02PL000h700<01`L08@L00`071`081`0;0;@00<X0
bP3:06`00`0016`006`302870P001:n`[k030005/:l0[k00100019J`000c1`0807RB099h07P30005
N01hN@0010L300T700<01`L02`L800<700<01`L020L000h700<01`L08PL00`071`071`<000D700L0
1`040007K6d006a]000R1`8000>`0;000P001Jn`0;2_00L000F_0000[`02034700/0TWRB099h07P0
N00600@70`081`0300L700<7200;1`0300L700P7000=1`0300L702<700<01`L01PL30005bP3:0<X0
0P0016`0K6`2009/00<01`L07`L20003[k2_008000B`[k2_0`001jl00;2_09H00`001I40TK000307
0140N99h07RB099h07RB099h00061`801PL701<700<01`L020L000d700<01`L08`L00`071`051`07
00L00;@0]0030004K01/K@8000=/K@0080L20006/:l00;2_0`000jn`00020003/02`00P000B_/02`
0P0_1`0A07RB099h07RB099hTP2BN0001@L800031`0701L700<01`L020L000d700<01`L08`L00`07
1`051`L000E/06`0K0030003K000020700<0[`000P001jn`002_0:l00P000k00/006000<U[00/96`
0;00/940;PL03@2BN980TWP0N980TWP02@021`801@L00`071`0F1`0300L700P7000<1`0300L702D7
00<01`L00`L400041`00K0<000I/0000K00P1`8000B_/02`0`000k2_/0020003[k2_008000>`0;00
0`001K00002`00<000B`[`00;PL01`1hTWP0N9801`000i9h00081`8010L00`071`0F1`0300L700P7
000<1`0300L702D700<01`L00`L30004bP00K0800V`2009/00<01`L07PL00`2`[`020006/:l0[k2_
0P001K00002`00D00`L30004/000T@@000>`00L0:`L90009N980TWRB098000T70P031`0300L701H7
00<01`L020L000/700<01`L09PL00`071`021`0700L00<X060020007KF`006e/000P1`090;2_0:n`
[`2_00<000B_/02`0`0;1`8000R_0000[k2_02@720021`8000mhTWRB099hTP2BN01hTP002@L20087
00<01`L05PL00`071`081`002`L00`071`0V1`0300L700870`000k@06002009/0P02K00300L701h7
00`0/000[k2_0:n`0;030003/:l000l70P000k2A/00301`7200:1`0307<000<000^BN99h07RB099h
000:1`0500L71`005`L00`071`081`002PL00`071`0X1`0500L70<X00`0011P006d30003KF`00207
0100ajl0[`2_0:n`0;2_0:n`0P0B1`@000>_/00050L7018700H0TWRB09830008TWRBN01hTP0:1`80
0PL00`071`0F1`0300L700L7000:1`0300L702P700<01`L00`001/X0K01/608000=/00008@L03P2_
/02`[`2_/02`[`2_0P0E1`04TK2A/08030L801T700T0Li5c07>A0940100017RBN00;1`8000<700L0
5`L00`071`071`002@L00`071`0Y1`0400L01`<000A]000H0P000fd01`0Q1`080:o70<N_0:l20006
/:l0[k005PL600@7200R1`0<07RB099h07RBN01h0`000gP01`0:1`<060L00`071`071`002@L00`07
1`0Y1`0400L0]08000A/000H100R1`8000V`[`2_0:n`0;000P000jn`000H1`T0:PL03`2AL`1cT@2A
Li40TG<0L`0200`70P0H1`0300L700L700081`0300L702/700D0bP00K@02000566d0K@008`L03@37
002_a`370<N_0:l00P0C1`X0;`L04@1hTWP0N01hTWP0N980TWP000d70P0G1`0300L700L700081`03
00L702/700@0bP1/0P000aQ/000202<700D0[jl0[`020008/:l0[k00/00<1`P01PL30307015cTG>A
095c07>A095c07>A000=1`<05PL00`071`071`001`L00`071`0[1`0800L0KAP066d302@700L0[lL0
a`3700<000G70<N_00051`L03@L0102Z000`1`0B07RBN01hTP2BN980TP2BN98030L00`2W000F1`03
00L700L700071`0300L702/700@0]00H0P001F`001P002<700@0[jl00Zl01@2_[`2_00X050L30347
0`003i5cTG<0Li5c07<0Li5c000<1`0400L001D700<01`L01`L000H700<01`L0:`L0103:01P20005
K@00600090L01@37[`2_00X06@L3038700<0TI800P003GRB099hTP2BN01hTP0030L01007_P0E1`03
00L700L700061`0300L702/70P0021P006`H01P08`L800:_00H0[jl0[`0H1`@0<`L0102ALi43000;
TG>ALi5cT@2AL`0030L00`071`0201@700<01`L01`L000H700<01`L0:`L01P0HK@1]60<070L80003
1`0000D000R_0:o70<N_04l700L0TI:A096B008000UhTWRBN01hTP0030L01@071jL001@700<01`L0
1`L000D700<01`L0:`L00`0H000200056000K00050L800T700L0[jl0[`2_00D000>_00L0C`L02@2A
Li40TG<0L`030006Li5cTG<030L010071`L201<700<01`L01`L000D700<01`L0:`L01@1;000H0080
00=]00L03@L7010700`0ajl0[lL0ajl0[lL3050700/0TI:A096B09:ATP030005TWRBN00030L01@07
1`L001<700<01`L01`L000@700<01`L0:`L0201;000HK01/0P061`P05`L01P2_[`2_00>_00H0[jl0
[`1@1`0>095c07>AL`1cTG<0Li420003Li4000`700@01`L70P0B1`0300L700L700041`0300L702/7
00<060003@0O1`0?0<N_0:l0[lL0ajl0[lL005070100TI80TY6B09:A096BT@2A0`0<1`0300L70087
00<01`L040L00`071`071`000`L00`071`0[1`T000?600L090L00`2_0002[`040:n_00:_00@0[jl0
D0L04`2AL`1cT@2ALi40TG>A095cT@0030L00`071`021`804@L00`071`071`000`L00`071`0U1`L0
00=;00000P0011P0000U1`030:l000d0D0L4000?TY40TI:A096BT@2ATY4000`700<01`L00`L00`07
1`0?1`0300L700L700021`0300L701h720051`0504/001P00`000lH0000U1`8000F_0:n_0002[`04
0:n_00:_00<01`L0C@L04`1cTG<0Li40TG>A095cT@2AL`003@L00`071`031`8040L00`071`071`00
0PL00`071`0F1`P03@L00`0H00030003B`3:00809@L0402_a`37[`2_a`37[`2_a`1?1`04096AT@@0
00^ATY40TI80TY6B000=1`0300L700@700<01`L03PL00`071`071`0000<700L04@L701@700D06000
B`030003bP0002H700@0[jl00Zl0102_[`02[`040:n_04h701@0TG>A095cT@2AL`1cTG<0Li5c00d7
00<01`L010L2010700<01`L01PL000031`0700T7200K1`0304/000<000<H0<X00`0U1`0@0:o70<N_
0:o70<N_0:o704h700D0TI6A0003T@@000RBTI80TY6B00d700<01`L01@L00`071`0>1`0300L700H7
000010071`L8028700D0B`00600802D700@0[jl00Zl0102_[`02[`040:l004d701D0Li5c07>AL`1c
TG<0Li5c07>AL`003@L00`071`051`0300L700h700<01`L01PL000@0:PL00`0H00030004B`00aP<0
9@L0202_a`37[`2_1P000lL01`1<1`05096AT@000i40102ATI430006TP2BTI803@L00`071`051`03
00L700h700<01`L01PL0000300L702X700D06000B`030003bP3:00<09@L00`2_00050008ZP2Z[`2_
ZP0R1`l06`L0402A095c07>A095cT@2ALi420003Li4000h700<01`L01@L200l700<01`L01PL00003
1`0702X700<0B`000`001AP0aP36008000<700L08`L2000?a`37[`2_a`37[`2_ajl001P72@003jnZ
[jZ_ZZl0[jV_ZJnY[`09018700<0TI400P03T@05096AT@00194400h700<01`L01PL00`071`0=1`03
00L700H700000`L01`0Y1`0504/001P010000lX0bP0200031`0702<70140[jX0ZZl0[jZ_0:nZ0:Z_
000E1`<0026Z0:ZYZZVZZJZYZZVZZJZY0:VZZJZYZZVZZJZYZZVZ0:l0100=1`0F08bAS940TG>A095c
T@2ALi40TG>A00l700<01`L01@L00`071`0=1`0300L700H700000`L01`0Y1`H000I;00360<H20003
1`0702<700l0[lL0ajo70<N_0:o70<L00P0A1`@002R_ZZnZ0:Z_ZZnZ[jZ_ZZnZ[jZ_0:nY[jV_ZJnY
[jV_ZJnY0:[6Z/JZ0`0:1`06096AT@2A0`03T@03096A00:A00D0TI6A000?1`0300L700D700<01`L0
3@L00`071`061`0000<700L0:0L01@0H001;00<000G:0<X0bP0200031`0702<700`0[jZ_0:nZ0:Z_
0:l30003[`0700d70`00;i2ZT:Z@ZY2ZT02YZZVZZJZYZZVZZJZY0:VZZJZYZZVZZJZYZZT0ZJnY[jV_
ZJnY00@01@L05P2AS96<08bAS01cTG<0Li5cT@2AL`0@1`0300L700H700<01`L030L00`071`061`00
0PL00`071`0V1`0304/000<000<H0<H00P001<X0bP021`0300L7028700P0ajl0[lL0a`<000K60<J_
aP0:1`<001BZ[jZ_ZZnZ[jZ_ZZl0[jZ_ZZnZ[`d001F_ZJnY[jV_0<JZaZ[6Z/JZaZ[6Z/H00`021`05
096AT@000i400`2A0002T@05096AT@000i400`071`0>1`0300L700H700<01`L030L00`071`061`00
0PL00`071`0V1`8000=;000010001/D0a@35008700<01`L08PL01@2Z[`2_00<000V_0:nZ[`2_ZP00
20L2000>T02@ZY2ZT:Z@ZY2ZT:X9000=ZZVZZJZY0:VZZJZYZP09000?[jV_ZJnY[jV_ZJnY[`2_0080
00^<THbA096<T@2AS0020009TG>AL`1cTG<0010700<01`L01PL00`071`0<1`0300L700H700021`03
00L702H71@002T/0003:0<X0bP021`0300L7028700<0ajl00P003LH0aZo60<J_0:o6[`0010L300>Z
00X0[jZ_ZZnZ[jZ_0`007jl0[jZ_ZZnZ[jZ_ZZnZ[`2_ZJnY[jV_ZJnY[jV_0<H00`003/JZaZ[6Z/JZ
aP36Z/JZ0P000i40T@02T@03096A00:A00<0TI400Y40102AT@0A1`0300L700H700<01`L030L00`07
1`061`000`L00`071`0T1`0304/000@000W:00350<D0a@000PL00`071`0R1`@000b_ZZl0[jX0ZZnZ
0:X2000@1`000:Z@ZY2ZT:X0ZY2ZT0@002FZT:Z@0:VZZJZYZZVZZJZYZZT0ZJZYZZVZZJZYZZVZZ@2Y
[jV_00@000bY[jV_Z@2Y[jV_ZJl30008TH`0S96<08`20006T@2ALi404@L00`071`071`0300L700/7
00<01`L01PL000<700<01`L090L20003B`1;008000W6003:0<X0bP000`L00`071`0Q1`0>0<J_aP36
[`2_aP36[lH20003ZP0000ZZ00<0ZP000P00;JZ_ZZnZ[jZ_ZP2Z[jZ_ZZnZ[jZ_ZZl0[jV_ZJnY[jV_
ZJnY0:[6Z/JZaZ[6ZP03000<ZP2ZaZ[6Z/JZaZ[60P04T@04096AT@800i400`071`0?1`0300L700L7
00<01`L02`L00`071`061`000`L00`071`0T1`D000=;0<X00P001/D0a@3500<700<01`L08@L02`2Z
[jX0ZZnZ0:Z_00<000j_002@ZY2ZT:Z@ZY2ZT0<003>@ZY2ZT:Z@ZY2ZT:X0ZZVZZJZYZZVZZJZY0:VZ
ZJZYZZVZZJZYZP2_ZJnY[jV_ZJnY[jT00`002jV_ZJnY[jV_ZJnY008000bA096<TH`0S01cT@0A1`03
00L700L700<01`L02`L00`071`061`0010L00`071`0S1`L000K6003:0<X300<700<01`L08PL0302_
aZl0[lH0aZl0ZP<02ZX2000FZP2Z[jZ_ZZnZ[jZ_ZZnZ0:Z_ZZnZ[`d001J_ZJnY[jT0Z/JZaZ[6Z/JZ
aZ[6ZP2Z0P002/JZaZ[6Z/JZaZX300BA00<0TI400P0B1`0300L700L700<01`L02`L00`071`061`00
10L00`071`0R1`8000=;00000`001<X00<D30003a@07008700<01`L08PL0202_ZZl0[jZ_0P003JX0
T:Z@092ZT:Z@ZY000P004:Z@ZY00T:Z@ZY2ZT:Z@ZY08000=ZZVZZJZY0:VZZJZYZP09000?[jV_ZJnY
[jV_ZJl0[jV_008001BY[jV_ZJnY0:W6Z@2<T@2AS96<018700<01`L01`L00`071`0;1`0300L700H7
00041`0300L7028710001AP0003600@000C:0<X010L00`071`0R1`060<J_aP360P001jX0TJZAZP00
1:X300JZ00P0[jZ_ZZnZ[`@001jZ0:Z_ZZnZ[jZ_ZZnZ[`2_ZJnY[jV_ZJnY[jT0Z/H4000=Z/JZaZ[6
0<JZaZ[6ZP030009Z/JZaP36[lJ_008019400`071`0A1`0300L700L700<01`L02`L00`071`061`00
10L00`071`0R1`H000<70<X00P001/D0a@3500@700<01`L08PL0102Z[jX2000;[`00T:Z@ZY2Z0:X0
0P003I2ZT:Z@ZY2ZT:X0ZY0010009ZZ@ZY2Z0:ZYZZVZZJZYZZVZZ@2YZZVZZJZYZZVZZJX0[jV_ZJnY
10003:nY0:V_ZJnY[jV_Z@8000R_0<JYaZW6Z@8000BAS94050L00`071`061`0300L700/700<01`L0
1PL000D700<01`L08@L00`1;000300031`00008000C:0<[50P041`0300L702<70`002jX00:ZAZY6Z
TJZA00802jX4000^ZZnZ[jZ_ZZnZ[`2_ZZnZ[jZ_ZZnZ[`2_ZJnY[jV_ZJnY[jT0Z/JZaZ[6Z/JZaP@0
00ZZaZ[6Z/JZaZ[60P002jo6[lJ_aZo6094001D700<01`L01PL00`071`0<1`0300L700D700051`03
00L702070P000aP0000200031`36008000C5aP360P041`0300L702<700`0[jZ_0:Z@ZY2ZT:X3000:
ZY2ZT:Z@ZY2ZT08001FZ0:Z@ZY2ZT:Z@ZY2ZT02YZZVZZJX030005JVZZJZYZP2_ZJnY[jV_ZJnY[jT0
Z@020009[jV_ZJnY[jV_00<000S6ZLJYaZW601H700<01`L01PL00`071`0<1`0300L700D700051`03
00L7020710003d/01`L0a@350<D0a@3500031`0300L702@70`003:ZAZY6ZTJX0ZY6Z00NZ0P03ZP0<
0:Z_ZZnZ[jZ_ZZnZ20003:nZ[jZ_0:nY[jV_Z@P000o6Z/JZaZ[6Z/JZ0:[6Z/H00P004Z[6Z/JZaP36
[lH0aZo6[lJ_aP8050L00`071`061`0300L700`700<01`L01@L000H700<01`L07`L00`1;00030087
00<0a/D00P001/H0aP3600@700<01`L08`L04`2@ZY2ZT:Z@ZP2ZT:Z@092ZT:X00`003JZ@ZY2ZT02@
ZY2ZT:X010007900ZJZYZZVZZJZYZZT0ZJZYZZVZZJZYZZVZ0:l3000>[jV_ZJnY[`2_ZJnY[jT2000C
[jV_Z@2YaZW60<JYaZW6ZLJY000C1`0300L700H700<01`L030L00`071`051`001PL00`071`0O1`03
01P000<00`L02P3:00350<D0a@041`0300L7028700P0T@2AZY6ZT@8000RZTJZAZY40ZP802JX00`2_
ZP04000S[jZ_ZZl0[jZ_ZZnZ[jZ_ZZl0[jV_ZJnY[jV_ZJnY0:[6Z/H00P003Z[6Z/JZ0:[6Z/JZaZ[6
0P002:X0[lJ_aZo60P001Zo6[lJ_aP804PL00`071`061`0300L700`700<01`L01@L000H700<01`L0
7`L20004600000<700X0a@00aP360<H010L00`071`0Q1`0A096ZT@2@ZY00T:Z@ZY2ZT:X00P002i2Z
T:Z@ZY2ZT:Z@00<002V@ZY2ZT:Z@ZY2Z0:ZYZZVZZJZYZZT0ZJZYZZVZZJZYZZT0ZJnY[jV_Z@03000=
Z@2Y[jV_ZJnY[jV_Z@02000AaZW6ZLJYaZT0ZLJYaP36[P004@L00`071`061`0300L700`700<01`L0
1@L000L700<01`L07PL30003600700<700X0bP00a@350<D010L00`071`0P1`07096ZTJX0ZP020007
TJZAZY6ZT@020003ZP2Z00RZ0`00;ZnZ[jZ_ZZnZ[jZ_ZP2Z[jZ_ZZnZ[jZ_0:nY[jV_ZJnY[jV_0<JZ
aZ[6Z/JZaZX3000;Z/JZaZ[6Z/JZaZX00P004LJ_aZo6[lJ_0:o6[`2_aZl0010700<01`L01PL00`07
1`0<1`0300L700D700071`0300L701d71P041`080<D00<H0a/D200D700<01`L07PL0702AZY6ZTJX0
ZY2ZT:Z@ZY00T:Z@092ZT:Z@ZY02000EZY2Z0:Z@ZY2ZT:Z@ZY2Z0:ZYZZVZ00/001BZZJZYZZT0ZJnY
[jV_ZJnY[jT0Z@8000^_ZJnY[jV_Z@2YaP020007ZLJYaZW6Z@030006[/J^aZh03`L00`071`061`03
00L700`700<01`L01@L000L700<01`L07@L600@700P0bP00aLX0bP801@L00`071`0M1`0?096ZTJZA
ZP2Z0:ZAZY6Z008000JAZY6ZT@04ZP801JX02P2Z[jZ_ZZnZ[jX8000;[jZ_ZZl0[jV_ZJl020003Z[6
Z/JZaZ[6Z/H0aZ[60P007:[6Z/JZaP36[lJ_0:o6[lJ_aZo60<J_aZo6[`0>1`0300L700H700<01`L0
30L00`071`051`001`L00`071`0M1`D01@L020350<D0a@350P051`0300L701`701X0TJZAZY6Z0:ZA
ZP2ZT:X0ZY2ZT:Z@ZP2ZT08000^ZT:Z@ZY2ZT02@ZP06000KT:X0ZZVZZJZYZZVZZ@2YZZVZZJZYZZVZ
Z@2Y00@000b_ZJnY[`2_ZJnY[jT2000J[jV_Z@2YaZW6Z@2YaZW6ZLH0aP36[/J^aP0>1`0300L700H7
00<01`L030L00`071`051`0020L00`071`0L1`D01@L02`3:0<X0bP3:0<H000@700<01`L06`L03P2A
ZY6ZTJX0ZY6ZT@2A0P002JZAZY6ZTJZAZP0200VZ10009JnZ[jZ_ZZnZ0:Z_ZZnZ[jZ_ZZnZ0:V_ZJnY
[jV_ZJl0aZ[6Z/H010002ZX0Z/JZaZ[6Z/H2000IZP2_aZo6[lJ_0:o6[lJ_0:o60<J_aZo6000=1`03
00L700H700<01`L030L00`071`051`0020L00`071`0L1`D01@L00`35aP020006a@000:h01@L00`07
1`0I1`06096ZTJZA0P004:ZAZY6ZT@2@ZY2ZT:Z@ZY020009ZP2ZT:Z@ZY2Z008002f@092ZT:Z@ZY2Z
T:Z@0:VZZJZYZZVZZJX0ZZVZZJZYZZVZ0:nY[jV_ZJnY[jT00P002ZnY[jV_ZJnY[jT20008aZW6ZLJY
aZT2000>aZT0[/J^aP36[/J^aP0<1`0300L700H700<01`L030L00`071`051`0020L00`071`0L1`03
04/000801PL00`3500030004a/D000D700<01`L060L05`2AZY6ZT@2AZY6ZTJZA0940TJZAZY6Z0080
00BAZY401:X300BZ02`0ZZnZ[jZ_ZZnZ[`2_ZZnZ[jZ_ZZnZ0:V_ZJnY[jV_Z@2ZaZ[6Z/JZaZ[6ZP<0
00VZaZ[6Z/JZaZX00`005jo6[lJ_aZo60<H0aZo6[lH0aZo6[lH000`700<01`L010L00`071`0=1`03
00L700D700091`0300L701/700<060000P061`0;0<H00:k60<H0aP0010L00`071`0G1`0F0940TJZA
096ZTJZAZY40TJZA092ZT08000RZT:Z@ZY00T080012ZT:Z@ZY2Z0:Z@ZY2ZT:Z@4P005:ZYZZVZ0:nY
[jV_ZJnY[jV_0:nY0P002ZnY[jV_Z@2YaZT2000FaZW6ZLJYaP36[/J^aZh0[/J^aZk600/700<01`L0
10L00`071`0=1`0300L700D700091`0300L701/70P000aP000061`0;0<D0a@350<D0a@0010L00`07
1`0G1`0D0:ZA0940TJZAZY6ZT@2AZY6Z0:X20008TJZAZY6ZTJX200ZZ00<0[jX01@000jX0000>0003
ZP0000H000jZaZ[6Z/JZaP36Z/JZaP80026ZaZ[60<J_aZo60<J_aZo60<H0aZo6[lJ_0:o6[lH0bP00
2PL00`071`041`0300L700d700<01`L01@L000T700<01`L06PL600H70`001/H0aP36[P801@L00`07
1`0E1`03096A00:A00L0TJZAZY6Z008000nAZY6ZTJX0ZY2ZT:Z@ZY000P002jX0ZY2ZT:Z@ZY2Z00@0
01fY092ZT:Z@ZY2Z0:Z@ZY2ZT02@ZY2ZT:Z@ZP2ZT008000:[jT0ZJnY[jV_Z@8000V_0<JYaZW6ZLH0
0P005:W6Z@2^aP36[/J^aZh0[/H0aZl02PL00`071`041`0300L700d700<01`L01@L000T700<01`L0
6PL600H700<0aLH00P001<G60<H200D700<01`L050L0502ZTJZA0940TJZAZP2ZTJZAZY6Z0P00396Z
TJZAZY40TJZA00NZ0P003ZnZ002_0:nZ[jZ_ZZl02:X01`2AZY6ZTJX01000296ZTJZAZP2Z0`002LJZ
aZ[6Z/JZaP020009[lJ_aZo6[lJ_00<001:_aZo60<J_aZo6[`2_0<K:aP091`0300L700@700<01`L0
3@L00`071`051`002PL00`071`0I1`H01`L02`2^0:h0[P2^0:h000@700<01`L04`L00`2AT@03T@0F
096A096Z0:ZAZY6ZTJX0ZY40T:Z@ZP8000Z@ZY2ZT02@ZY2Z0P001ZVZZJZYZPD0012ZT:Z@092ZT:Z@
ZY2Z0:Z@10002ZX0ZY2ZT:Z@ZY03000<T9400:V_ZJnY[jV_0`007<JYaZW6ZLJYaP36[/J^aP36[/J^
aZh0[lJ_aP091`0300L700@700<01`L03@L00`071`051`002PL00`071`0I1`D020L02`360<H0aP36
0<H000@700<01`L04`L0902AZY6Z0:ZAZP2Z0:ZAZY6ZTJX0ZY6Z0:ZAZP2ZTJZAZY6Z00:Z10002jnZ
[jZ_ZZl0[jZ_010000^AZY6ZTJX0ZY6ZT@030007T@2AZY6ZT@02000VaZ[6Z/H0aZo60<J_aZo6[lH0
aP36[lJ_aP36[lJ_0<H0a/[6bP081`0300L700<700=;00L03PL00`071`051`002PL00`071`0I1`D0
20L01@2^aP36008000B^aP001@L00`071`0A1`03096A00:A00<0TI400i401`2ZTJZAZY400P001JZA
ZY6Z00<000VZT:Z@ZY2ZT:X00P02ZP030:X000<001nYZP2ZT:Z@ZY2ZT02@ZY2ZT:Z@ZP2ZT:Z@ZY2Z
T02@00<000V@ZY2ZT02@TI000P002i400:V_Z@2YaZW6008001ZYaZW6Z@2^aP36[/J^aP36[P2_aP36
[lJ_00P700<01`L00`L00`071`0?1`0300L700@7000;1`0300L701P71@091`0;0<D0a@350<D0a@00
10L00`071`0@1`0T096ZTJZA096ZTJX0ZP2ZTJZA096ZTJZAZY40TJZAZY6ZTJZA10002:l0[jZ_ZZnZ
1@001jZ_ZZnZ[`002:X01`2AZY6ZTJX010002Y6ZTJZAZY6Z0:X20005TJZAZY400`008Zo6[lJ_aZl0
[lJ_0:o6[lH0aZo6[lH0aP3:a/X0b/K:aP071`0300L700<700<01`L03`L00`071`041`002`L00`07
1`0H1`D02@L02`360<H0aP360<H000@700<01`L040L00`2AT@02T@03096A00:A00@0TI6A0P004Y40
TJZAZY6ZT@2A092ZT:Z@ZP801:X3000:ZZVZZJZYZP2ZT0@000jZT02@ZY2ZT:Z@ZP2ZT0@000^ZT02@
ZY2ZT:Z@ZP020008T96@TI2AT9430008aZW6ZLJYaZT2000HaP36[/J^aP36[/J^aP36[lJ_0:o6[lH0
1`L00`071`031`0300L700l700<01`L010L000/700<01`L060L500X70P002LG60035aP3600031`03
00L700l70240ZY6ZT@2AZY6ZT@2AZY6ZT@2AZY6ZTJZA096ZT@2AZY400`002:Z_ZZnZ0:Z_0P003:Z_
ZZl0[jZ_ZZnZ[`h000bAZY6ZTJZA096ZTJX30008ZY40TJZAZY4200<70P008Jo6[lJ_aZo60<J_aZo6
[`2_aZo60<X0b/K:aP36b/K:00061`0300L700<700<01`L03`L00`071`041`0030L00`071`0H1`@0
2PL02`2^0:h0[P2^0:T000@700<01`L03@L0102A0002T@03096A00:A00<0TI400i401`2A096ZTJX0
0P001Y6ZTJX0ZP800`L200>Z00D0ZZVZZ@03000PZ@2@ZY2ZT:Z@ZY00T:Z@ZY2ZT:X0ZY2ZT:Z@ZY2Z
0:X30009ZY2ZT:X0TI2A00801`L0202YaZW6ZLJY0P005lJ^aZk6[P2^aP36[`2_aZo6[`2_aZl000H7
00<01`L00PL2014700<01`L010L000`700<01`L060L400X700H0aP360<H20004a@3500<700<01`L0
3@L0102AZY42000JZY6ZT@2AZY6ZT@2AZY40TJX0ZY6ZTJZAZP071`<000R_ZZnZ[jZ_ZP@000R_ZZnZ
[jZ_00RZ00P0TJZAZY6ZT@@000VZTJZAZY6ZTJX00`0:1`8001o6[lJ_0:o60<J_aZo6[`2_0<K:aP36
b/K:0<[6b/H000D700<01`L00PL00`071`0@1`0300L700@7000<1`0300L701P7100;1`8000V^aP35
0<D0a@000`L00`071`0<1`03096A00:A00<0T@000Y400`2AT@04T@0?096AT@2Z0:ZAZY6ZTJX000X7
0P002jZYZZVZZJZY092Z00D000jZT02@ZY2ZT:Z@ZP2ZT0D000b@ZP2ZT:Z@ZY2ZT:X200h701d0aZW6
0<J^aP36[/J^aZh0[lJ_aP36[lJ_aP36[`0200H700<01`L00P0B1`0300L700@7000<1`0300L701P7
100;1`0<0<D0a@350<G60<H00PL00`071`0<1`0O0:ZAZY40TJX0ZP2ZTJZAZP2ZTJZAZP2ZTJZAZY6Z
000=1`<000f_ZZnZ[`2_ZZnZ[jZ_00h000jAZY6ZTJZAZP2ZTJZAZP<04@L3000;aZo6[lH0aZo6[lH0
0P003lK:a/X0b/K:aP360<K:00051`0400L701<700<01`L010L000d700<01`L05`L400/700d0aP00
Z@2Y0:T0ZLD0008700<01`L02PL00`2AT@02T@03096A00:A00<0TI400i400`2AT@03T@090940ZY6Z
TJX0014710008JT0T:Z@ZY2ZT:Z@092ZT:Z@ZY2Z0:Z@ZY2ZT:Z@ZY00T00401H701d0[/J^aZk60<J^
aP360<J_aZo60<J_aZl0[lZ_00051`0400L001<700<01`L010L000d700<01`L05`L400`700`0aP36
0<H0a/D0a@021`0300L700X700X0TJZAZP2ZTJZA0P004ZZAZP2ZTJZAZY40TJX0ZY6Z01H71P001JnZ
[jX000ZZ00P0ZY6ZTJZAZP@06`L05P2_aZo6[lH0aZl0a/X0b/K:aP36b/H20004b/K:00D700@01`00
4`L00`071`041`003@L00`071`0G1`@030L03`2Y0:W50<D0a@35Z@07000;1`03096A00:A00<0TI40
0Y400`2AT@020003T@2A00BA00<0TI400Y400`2Z000M1`X000R@ZY2ZT02@ZPD080L01`2^aZk6[/H0
0P004jo6[`2_aZo6[`2_0:o:0<Z_bP0010L00`07000D1`0300L700@7000>1`0300L701H71@0<1`0>
0<D0a@350<G60<H01`0;1`0M0:ZAZY40TJZAZP2ZTJZA096ZTJZA096ZTJZAZP00:0L802H701/0[lJ_
aZo60<[6b/H0a/[6bP3:0<[60<K:aP0010L00`1;000D1`0300L700@7000>1`0300L701L7100<1`0?
0<D0aJT0Z@2Y0:W500L000T700<0TI400Y400`2AT@02T@03096A00:A00<0T@000i400`2AT@04T@03
00L705H701X0[/J^aP360<J_aZl0[lJ_aP3:[lX0bZo:[`D000<H00L04`L00`071`041`003PL00`07
1`0G1`@03@L01@360<K5008000O6a@3500L000T70180TJZAZP2ZTJZA096ZTJX0ZY420008ZP2ZTJZA
ZP1J1`0?0:o60<[60<K:a/X0b/K:008000W6b/K:0<[6bP000`L00aP01`0D1`0300L700@7000?1`03
00L701H7100=1`0<0:W50<D0a@350<FY0P091`<000>A09400i400`2AT@02T@03096A00>A00<0TI40
0i400`071`1J1`0I0:h0[lJ_0:o6[lH0aP3:[`2_bZl0[lZ_00031`03B`0701@700<01`L010L000l7
00<01`L05PL400h700D0a@35aP020006aLH00<002@L00`2AZ`02000FTJZAZP2ZTJZA096ZTJX0ZP2Z
TJZA05/70`006LK:a/[60<K:aP360<K:0<[6b/H0a/[600L00`0E1`0300L700@7000?1`0300L701H7
100?1`0;0:T0Z@2Y0:T0ZLD00P071`03096A00:A0`000i40T@03T@03096A00>A00<0TI400P000i40
1`1B1`T00`L06@2_aZo60<J_aZl0[lZ_0:o:[lX0bZo:00L00`0E1`0300L700@7000?1`0300L701H7
100?1`050<K50<D00`001L00`<H000L700P0Zi6[T@2AZ`80012AZY6Z0:ZAZY40TJZAZY40C0L800`7
00T0b/K:a/X0b/H00P002l[6bP3:a/[60<K:00@000<700L050L00`071`041`0040L00`071`0F1`<0
40L03@350<D0a@35Z@2Ya@001PL01@2ATI4000BA0`000i40T@04T@03096A00:A00<01`L0@PL901D7
01<0aZo6[`2_0:o:0<Z_bP3:[lZ_008000C:00000PL00`071`0B1`0300L700@7000@1`0300L701H7
100?1`0=0<G60<H0aP360<K000061`0<0:^AZ`2[TJ^A096[0P002i6ZTJX0ZY6ZTJX003`7200O1`0C
0<[6b/H0a/[60<K:aP36b/K:0002bP800`L00`071`0C1`0300L700<7000@1`0300L701H7100@1`0=
0:T00<D0a@350<FY00051`05096AT@000i400`2AT@02T@800i400`2AT@02T@0300L703872@0W1`04
0:o6[`8000_:[lX0bZo:0<Z_bP020003[lX000@700<01`L04`L00`071`031`004@L00`071`0F1`<0
40L01P36a@30aP8000K0aP36`0031`0@096[TJ/0Zi6[0:^AZi40T@<000JAZY6ZT@0/1`P0<@L04P36
0<K:0<[6bP3:aP36bP3:00;:00<01`L00PL00`071`0C1`0300L700<7000A1`0300L701H70`0A1`80
00^Y0:T0Z@2Ya@3500031`03096A00:A00D0TI6A0003T@03096A00:A00<0TI400Y400`071`0R1`T0
>@L03@360<Z_0:o:[`2_bZl00P001/Z_0:o:00D700<01`L04PL00`071`031`004@L00`071`0G1`80
4PL03@36`0300<00`<H0aP000PL04P2AZi40TJ^AZ`2[TJ/0Zi6[T@<000>AZP0070L804<700d0b/K:
0<[6bP3:a/X000?:00@0b/X01@L00`071`0B1`0300L700<7000A1`0300L701L700=;00004PL03P35
0<FY0:T0ZLD0a@070`000i40T@02T@03096A00:A00D0TI6A0003T@8050L904/700/0[lZ_bP3:[`2_
bP020007[lZ_0:o:00051`0300L7018700<01`L00`L0018700<01`L05PL301<70100aP36`0300<00
`007096B0`004J^A096[TJ/0Zi6[0:^AZi4000`7201E1`080<K:aP36b/H20003bP3:00;:00@0b/X0
1@L00`071`0B1`0300L700<7000B1`0300L701L70P0C1`0<0:W50<FY0:T0Z@2Y0P03T@0309400080
0i400`2AT@02T@05096AT@0010L905d701<0bZo:0<Z_bP3:[`2_bZl0[lX000H700<01`L04@L00`07
1`031`004PL00`071`0G1`8050L01P30aP36`08000c6`02Y096BT@2ATY43000:T@2AZi6[0:^AZ`D0
IPL01`36b/H0a/X00P02bP050<[:bP000/X00`071`041`0300L7014700<01`L00`L001<700<01`L0
5PL301@700L0ZLD0a@35008000>Y00000i401@2ATI4000:A0`02T@03096A00:A00<01`L0J@L0102_
bZl2000<bP3:[lX0bZl0[lX01PL00`071`0A1`0300L700<7000C1`0300L701L70P0E1`0H0<360<H0
aP35Z@2ATY40TI:A096BT@2A0`001I6[TJ/006/700D0b/K:0002bP050<[:bP000/X0103:00061`03
00L7014700<01`L00`L001<700<01`L05`L201H700X0Z@2Ya@300<000i401@2ATI4000>A00@0TI6A
0`000i401`1[1`030:o:008000V_bP3:[lX0bZl00P000lX01`061`0300L7010700<01`L00`L001@7
00<01`L05PL301H700L0`000a@35008001:ATY40TI:A096BT@2ATY40TI8206d70P000lX0bP02bP04
0<[:00;:00<01`0020L00`071`0@1`0300L700<7000D1`0300L701L70P0F1`090<FY0:W00<0000>A
00D0TI6A0003T@05096AT@000i400`071`1[1`0:0<Z_0:o:[`2_bP@000>_00L01`L00`071`0@1`03
00L700<7000D1`0300L701L70P0G1`8000BY0:W50`004i:A096BT@2ATY40TI:A096BT@00K@L0103:
bP03bP090<[:0<X700L000P700<01`L040L00`071`031`0050L00`071`0H1`805`L02030Z@2Y`02A
1002T@05096AT@000i40102AT@1^1`090<Z_bP3:[`2_008000G:0<Z_00091`0300L700l700<01`L0
0`L001D700<01`L05`L201P700/0aJT0ZLD0TY40T@04000:TY40TI:A096B06h700D0b/[:0002bP80
00O:0<X0bPL000T700<01`L03`L00`071`031`005@L00`071`0H1`805`L30004Z@2Y00:A00H0TI6A
09480003T@0706d70100bZo:0<Z_0:l0[lX0bZl02@L00`071`0?1`0300L700<7000E1`0300L701P7
0P0G1`0500L70:T00P0049:ATP2BTI80TY40TI:A094206d70P03bP0;0<X000L01lX0bP002`L00`07
1`0>1`0300L700<7000F1`0300L701P700=;00L05@L020071`L0ZL000Y401@2ATI4000:A00D0TI6A
0002T@0300L706X70`000lZ_bP020009[`2_0:o:0<X000/700<01`L03`L00`071`021`005PL00`07
1`0H1`805@L00`071`031`0D0:T0TI80TY6B09:A096BT@2ATP1[1`@00lX02`3:0<X700O:0<X000/7
00<01`L03`L00`071`021`005PL00`071`0I1`03600701<700<01`L010L200:A00D0TI6A0003T@07
096AT@2A001Y1`H000k:[`00bP3:[`2_0:o:00/700<01`L03`L00`071`021`005`L00`071`0H1`80
50L00`071`051`0C096BT@2ATY40TI80TY6B09:A001W1`@000>;00000P002<X0bP3:1`07100<1`03
00L700h700<01`L00PL001L700<01`L06@L00d/01`0B1`0300L700H700@0TI400i40102AT@03T@04
096A06H700=]K0000`003`L0bZl0[lX0bP3:[`2d000=1`0300L700h700<01`L00PL001L700<01`L0
6@L201<700<01`L01PL@0004T@00068700@0Rff;0P031`080<X700O:0<X400031`0700`700<01`L0
3PL00`071`021`005`L00`071`0J1`036007010700<01`L01`L0402ATY40TI80TY40TI:A09430003
Z@0005h710000fd000041`0=0<Z_0:o:0<X0bP3:000>1`0300L700h700<01`L00PL001P700<01`L0
6@L2014700<01`L020L0402AU`2GTIL0Ui40TINA09430004ZL0005X71`061`070<X700L01`020004
]02d00l700<01`L03@L00`071`021`0060L00`071`0J1`03B`0700l700<01`L020L04`2BTI80TY40
TI80TY6B09:A0:T00P02Z@<0E@L800L700D0bZl0[`030004]02d010700<01`L03@L00`071`021`00
60L00`071`0K1`0300L700h700<01`L02@L0402GT@2AUi40TIL0Ui40TIL30006`000ZL2Y0P1A1`P0
20L03@07bP3:1`0700L01`0040L00`071`0=1`0300L70087000I1`0300L701X700<H00L03PL00`07
1`091`0C096BT@2ATP2BT@2ATP2BTI80Z@020003Z@0000>Y0P1=1`T02@L00`2_bP040005]000bP00
4@L00`071`0=1`0300L70087000I1`0300L701/70P0=1`0300L700/700l0TIL0Ui40TINA096G09L0
0P000jW0Z@050003ZL2Y00<0B0L900/700`01lX0]02d0;@0]00B1`0300L700`700<01`L00PL001T7
00<01`L06`L300`700<01`L02`L02@2BTI80TY40T@060004TY4000BY10001:T0ZJ<304872P0=1`@0
00?:0<X00P000k@01`0B1`0300L700`700<01`L00PL001T700<01`L070L00j801`0;1`0300L700`7
00<0U`001@002Y6G09NA096G0<030003`:W000D01:T201471`0T1``03`L02`0700L01`0700L001<7
00<01`L030L00`071`021`006PL00`071`0L1`802`L00`071`0<1`80012BT@2ATP2BT@2ATP2BTI80
0jT200>Y1`000j>Y000200l71@000k00000O1`d04@L01P2d0;@0]0@050L00`071`0<1`0300L70087
000J1`0300L701d70P0:1`0300L700d70100TIL0Ui40TINA096G09NA10001/2Y003000:Y0`001:T0
0002Z@<03PL50003[`0001T73`0B1`070;@0]0L01`020003]00701@700<01`L02`L00`071`021`00
6PL00`071`0N1`802@L00`071`0=1`0@09:ATP2BT@2ATY40T@2ATP<000>Y00000ZT30004ZJ>YX`80
00>YX`000`000jT0000300/70P000jl0[`0201<710000hD0000;01@700H0bP3:0<X20003]00701D7
00<01`L02`L00`071`021`006`L00`071`0N1`801`L00`071`0?1`0=09NA096GT@2AU`2GT@020003
Ui40008000JY00300:T400:Y0`03Z@D000>Y00001@051`03[`00008000>`00L02@L<0003R`0000T0
5@L01`0700Nd0;@00`0G1`0300L700/700<01`L00PL001/700<01`L07PL00h@01`061`0300L70107
00T0TY40TI80TY400`001i40TINA00L00P000jT000020003XjVS00@000VY0000Z@2YRZT03`000k00
00090003RhZ;00`000>500001P0H1`040;@0]0@000>d00L05PL00`071`0;1`0300L70087000K1`03
00L701l700>400L01@L00`071`0A1`0509L0Ui400`002I6G09NA096G00031`<00ZT400>Y2@002:VT
ZJBYY:VT1P000k00000:0003RjB;00P000FT0000R`0601T700@0]02d10000`L01`0H1`0300L700/7
00<01`L00@L001`700<01`L07`L200D700<01`L04@L00`2ATP02000=TIL0Ui6G09NA096G00051`@0
00>SZ@0010000jVS000<0005RZV:08X01P000k0000050007RhZ;RX^:R`080004Q@0001`700<0bP00
0P001K@0];@001T700<01`L02`L00`071`011`0070L00`071`0P1`8010L00`071`0B1`80012G09NA
U`2GTIL0Ui40TIL020L400>Y1P000j@0000=0004[`00/08000R;ZH^YR`2;Y0h07PL300031`070080
00>d00L06@L00`071`0;1`0300L70047000L1`0300L702470P021`0300L701@700h0TIL0Ui6G09NA
U`2GT@8000>G00L02PL40005X`2:ZHX020000jT0000A0003RhX000T08PL0102d0;@30003]00701/7
00<01`L02PL00`071`011`0070L00`071`0R1`8000<700L05PL02`2AU`2GTIL0Ui6G008000BB09:G
0P0=1`D000RTZJBYY:VTZ@L00[020003/00000H000>;Y000100V1`040;@0]0<000>d1`0070L00`07
1`0:1`0300L70047000M1`0300L702870`0H1`0>096G09NAU`2G09L0Ui420003Ui40014720001JV:
Z@2T008000BF002`2`0[1`8000>d0;@0100M1`0300L700X700<01`L00@L001d700<01`L08`L301P7
00H0TIL0Ui420003Ui:G008000FBU`00TP0201L71P001K00002`0080<`L01@07002d00@07PL00`07
1`0:1`0300L70047000M1`0300L702@70`0H1`03096G008000ZAU`00TINA096G0P00196G000G1`80
00>`0;000P0019H0000b1`040;@0]0<000>d00L07`L00`071`091`0300L70047000N1`0300L702<7
100H1`8000ZG002BUi80TYNB0P001YNB002GTP<04PL200:`0`000k00/00203<70P001PL0002d0247
00<01`L02@L00`071`011`007PL00`071`0R1`0300L700<060L20006Ui4009NA0`001Y4009NAU`<0
00>GT@000`0<1`8000Z`002F/02`0;2F0P0b1`030;@000<000>d00L08@L00`071`091`0300L70047
000N1`0300L7028700@01`L70`0I1`0409NBU`8000JBU`2GTYL20004TP2G/08000BG09N`30002YL0
0;00/02`0;03038700<0]0001@0S1`0300L700T700<01`L00@L001h700<01`L08PL00`071`021`@0
60L20003TINA008000JG002AUi430003TINA00@000ZGUP2F/9J`0;2F0`002IJ`U[00/02`UP0203<7
1P0V1`0300L700P700<01`L00@L001l700<01`L08@L00`071`031`@06@L30005TP00Ui800P0019L0
/9L30004U`2G/0X000>G/9L00`0019N`0;0303<71P0W1`0300L700P700<01`L00@L001l700<01`L0
8@L00`071`041`D06@L20006TIL0096G0`0019NAUi440006UiJG0;2F1@001IJ`UP2F008000F`0;2F
000b1`H0:0L00`071`081`0300L70047000O1`0300L7020700<01`L01PL501X70P000iNB00020003
Uk0000<000BG09N`1P001YN`U`2G/0@000BG002`0P0b1`H0:@L00`071`081`0300L70047000P1`03
00L701l700<01`L01`L501/70P000i6G00030003TINA010000>`U[000P000iH0000c1`030;@000<0
:`L00`071`071`0300L70047000P1`0300L701l700<01`L02@L501/70`000iN`U`060008/9N`U`2G
/9L60003/9N`0080<`L20004]00002d700<01`L01`L00`071`011`0080L00`071`0O1`0300L700X7
1@0M1`<000NAUi40UYNF00L000FF/9H0UP0303D71P0^1`0300L700L700<01`L00@L0024700<01`L0
7PL00`071`0<1`D07`L60007Uk2G09N`U`0503L71@0`1`0300L700L700<01`L00@L0024700<01`L0
7PL00`071`0=1`H08`L703/7100c1`0300L700L70@011`47000Q1`0300L701d700<01`L03`L70647
1@0d1`0300L700L70@011`47000Q1`0300L701d700<01`L04@L705h7100f1`0300L700L70@011`47
000R1`0300L701`700<01`L04PL805/7100h1`0300L700H70@011`47000R1`0300L701`700<01`L0
50L805L7100j1`0300L700H70@011`47000R1`0300L701`700<01`L05PL00id0000505@70`0l1`03
00L700H70@011`47000S1`0300L701X700<01`L06@L00h@0000505070`0n1`0300L700H70@011`47
000S1`0300L701X700<01`L06`L00`24000604/70`111`0300L700D70@011`47000S1`0300L701X7
00<01`L07@L;04@710131`0300L700D70@011`47000T1`0300L701T700<01`L07`L=03`71P151`03
00L700D70@011`47000T1`0300L701T700<01`L08@L20003J`0000X0=@L704L700<01`L01@L10047
0@L002@700<01`L060L00`071`0U1`>43@0]1`P0B`L00`071`041`400@L11`0090L00`071`0H1`03
00L702T700>400003P0S1`/0C@L00`071`041`400@L11`009@L00`071`0G1`0300L702`700>40000
400F1a00D0L00`071`041`400@L11`009@L00`071`0G1`0300L703073P000f^4J`0P05@700<01`L0
10L100470@L002D700<01`L05`L00`071`0d1`P000B4JhA[7@1I1`0300L700<70@011`47000V1`03
00L701H700<01`L0>0L50003Q6/001L0G`L00`071`031`400@L11`009PL00`071`0E1`0300L703h7
4`1V1`0300L700<70@011`47000V1`0300L701D700<01`L0]`L00`071`031`400@L11`009PL00`07
1`0E1`0300L70;P700<01`L00`L10047000W1`0300L701@700<01`L0^0L00`071`031`400@L002L7
00<01`L050L00`071`2h1`0300L700<70@011`009`L00`071`0C1`0300L70;T700<01`L00`L10047
000X1`0300L7018700<01`L0^PL00`071`021`400@L002P700<01`L04PL00`071`2j1`0300L70087
0@011`00:0L00`071`0B1`0300L70;X700<01`L00PL10047000Y1`0300L7014700<01`L0^`L01@07
1`L00047000Y1`0300L7014700<01`L0^`L01@071`L00047000Y1`0300L7010700<01`L0_0L01@07
1`L00047000Y1`0300L7010700<01`L0_0L01@071`L00047000Z1`0300L700l700<01`L0_@L01007
1`011`00:PL00`071`0?1`0300L70;d700@01`L00@L002X700<01`L03`L00`071`2m1`0400L70047
000[1`0300L700d700<01`L0_PL010071`011`00:`L00`071`0=1`0300L70;l700<01`000@L002/7
00<01`L03@L00`071`2o1`0300L00047000/1`0300L700`700<01`L0_`L00`0700011`00;0L00`07
1`0<1`0300L70;l700<01`000@L002`700<01`L030L00`071`301`400@L10000;0L00`071`0;1`03
00L70<470@011`40000]1`0300L700X700<01`L0`@L100470@0002d700<01`L02PL00`071`311`40
0@L10000;@L00`071`0:1`0300L70<870P0002h700<01`L02@L00`071`321`80000^1`0300L700P7
00<01`L0``L20000;PL00`071`081`0300L70<<70P0002h700<01`L020L00`071`341`40000_1`03
00L700L700<01`L0`0L50000;`L00`071`071`0300L70;H72P051`00;`L00`071`061`0300L70:d7
2P0?1`00<0L00`071`051`0300L70:@72@0I1`00<0L00`071`051`0300L709X72P0R1`00<0L00`07
1`051`0300L709072P0/1`00<@L00`071`041`0300L708L72@0f1`00<@L00`071`041`0300L707d7
2P0o1`00<@L00`071`031`0300L707@72P191`00<@L00`071`031`0300L706/72@1C1`00<PL00`07
1`021`0300L706472P1L1`00<PL00`071`021`0300L705L72P1V1`00<PL00`071`021`0300L704h7
2@1`1`00<`L010071`171`X0N@L003<700@01`L0?@L:08<7000c1`0400L703@72@2=1`00=0L00`07
000Z1`X0UPL003@700<01`0080L:0:07000d1`0300L001L72@2Z1`00=0L200h72P2c1`00=@L00`07
1`021`X0_@L003D71@371`00o`L21`00\
\>"],
  ImageRangeCache->{{{0, 256}, {288.375, 0}} -> {0.039098, 0.0471435, 
  0.00331587, 0.00331587}},
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \("-MeshGraphics3D-"\)], "Output",
  CellLabel->"Out[26]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; 
    MVShow3D[sband, MVTexture\  \[Rule] \ "\<`FishBlack`\>", 
      MVTextureMapType\  \[Rule] \ MVMeshUVMapping, 
      MVScaleTexture\  \[Rule] \ {1, 36}]; \n
    MVShow3D[MVRotate3D[sband, {0, 0, 1}, 2  \[Pi]/3], 
      MVTexture\  \[Rule] \ "\<`FishWhite`\>", 
      MVTextureMapType\  \[Rule] \ MVMeshUVMapping, 
      MVScaleTexture\  \[Rule] \ {1, 36}, MVGrayBackground -> True]\)], 
  "Input",
  CellLabel->"In[27]:=",
  CellTags->"MathLinkMode"],

Cell[BoxData[
    \({MVMesh3D[6] \[DoubleLeftArrow] MVTexture2D[8], 
      MVMeshUVMapping, {32, 128, MVRGBTexture}}\)], "Output",
  CellLabel->"Out[27]=",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[28]:=",
  CellTags->"MathLinkMode"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F30300000005F3030000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F30305F30306030305F3030CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
5030305F3F3F5F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF000000603030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0503030503030000000C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF50303050303F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000603F3F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF4F3030CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000
00000000000000000000000000000000000000000000000040505FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C00000005F3030CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF402F2FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000030404F4F5F6040505F4F5F6040505F4F5060\
40505F4F5F6040505F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF100F0F603030CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF4F3030CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000001010104F506040505F40505F40505F4F506040505F\
40506040505F000000000000000000000000000000000000000000000000000000
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F3030000000\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004F302F603F3060403F603F3060403F603F30604030603F3060403F\
603F30604030000000000000000000000000000000000000000000000000000000
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F2F2F5F3F30603F305F3F30603F30603F30603F305F3F30603F30603F30\
604030603F30603F30000000000000000000000000000000000000000000000000
0000000000004F2F20000000000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF405050000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F000000000000000000000000000000000000\
0000000000002F1F1F603F3F5F3F30603F3F5F3F30603F3F5F3030000000000000
000000000000000000402F2060403F603F3060403F603F3060403F603F3060403F603F3060403F\
402F200000000000000000000000000000000000000000001F100F604030604030
6F4030603F306F40306040306F4030603F306F40306040300F0F0F000000000000000000000000\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0302020C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
00000000000020202F0F0F0F1F2020404F504F5F5F3040404F5F5F4F505F4F5F5F
40505F4F505F000000000000000000000000000000000000000000000000000000000000000000\
3F2F205F3F3060201F5F0000600F005F3F303F2F200000000000000000000F0F0F
4F2F2F603F305F3F30603F305F3F30603F30603F30603F30603F30603F30603F30604030603F30\
0F00000000004F302F603F30604030603F30604030603F30604030603F306F4030
603F30604030603F306F4030603F30604030603F306F4030603F30604030603F306F40302F1F10\
000000000000000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000030202030201F000000000000\
0000000F0F0F0000000000000000000000005F3F3060403F603F30603F3F603F30
60403F603F3060403F603F3060403F603F3060403F603F3060403F603F3060403F603F306F403F\
603F306F403F603F306F403F603F306F4030603F306F403F6040306F4030603F30
6F403F6040306F40306040306F40306040306F40306040306F40306040306F40306040306F4030\
6040306F40304F2F200000000000000000000000006F4030000000000000000000
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F1F1F3F2F20000000000000000000000000000000000000000000\
000000000000100F0F5F3F30603F3F5F3F30603F305F3F30603F3F5F3F30603F30
603F30603F3F5F3F30603F30603F3060403F404F5F40505F40505F40505F000000000000000000\
00000010101010101F40505F40505F404F5F40505F40505F40505F404F5F40505F
40505F405060404F5F40505F603F306F4030603F306F4030603F306F4030603F306F4030604030\
6F4030603F2F6F40306040306F4030603F2F6F40300000000000000000000F0F00
50302F000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF403030CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000100F0F5030305F3F30
402F2F0000000000003F2F2000000000000000000000000000000000000000000030201F603F3F\
5F3F30603F3F5F3F30603F3F0000000000000000000000002F303040505F40505F
40505F40505F4F505F40505F40505F40505F4F505F40505F40505F0000001010100000003F4F50\
40505F4F506040505F40505F40505F4F506040505F40506040505F4F506040505F
40506040505F4F506040505F40506040505F0000000000000000000000000000006F40306F4030\
6040306F40306F40306F40306040306F40306F40300F0F0F0000000F0F0F0F0000
0F0F0F000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000000000000000\
0000000000002F1F1F5030305F3F3F3F20200000000000000000003F20205F3F3F
2010100000000000000000000000000000000000005F3F305F30305F3F3F5F3F30603F3F5F3F30\
00000000000040505F404F5040505F40505F40505F404F5040505F40505F40505F
404F5040505F404F5F40505F404F5040505F404F5F40505F404F5F40505F404F5F40505F404F5F\
40505F404F5F40505F404F5F40505F404F5F40505F404F5F40505F404F5F40505F
404F5F40505F404F5F405060404F5F40505F303F40000000000000000000000000000000404F5F\
1F20200000000000000000006F40306040306F403060402F6F403060402F1F1010
2F1F10100F0F000000100F0F000000604030000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF000000000000000000101010505F5F4F505F000000000000000000000000503F3F\
2F1F1F0000000000000F0F0F5F30305F3F3F301F1F000000000000000000000000
000000000000000000201F1F603F3F5F3F30603F3F5F3F304F505F40505F4F505F40505F4F505F\
40505F4F505F40505F4F505F40505F40505F40505F4F505F40505F40505F40505F
40505F40505F40505F40505F40505F40505F40505F40505F40505F40505F40505F404F5F40505F\
40505F40505F404F5F40505F40505F40505F404F5F40506040505F40505F40505F
40506040505F40506040505F4F506040505F40506040505F0F0F0F0000000000000000000F0F0F\
40505F4F50600000000000000000000000000000000000001F1F204F5060000000
6F40306F40306F40300F00000F0F0F0F000000000060402F000000000000100F0FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F40CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF3F2F2FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
0000000000000000000F0F0F0000000000000000000000000F000050303F000000000000000000\
1F1010503030100F0F000000000000000000000000000000000000301F1F5F3F3F
5F30305F3F3F5F30305F3F3F4050504F505F40505040505F40505040505F404F5040505F405050\
40505F404F5040505FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF404F5040505F404F5F40505F404F5040505F404F5F\
40505F404F5F40505F404F5F40505F404F5F40505F404F5F40505F404F5F40505F
404F5F40505F404F5F40505F0F10100F0F0F000000000F0F000000000000000000000000000000\
0F0F0F1F202F404F50000000000000000000000000000000000000202F302F303F
000000000000000000CFCFCF60402F0F0F00603F2F0F0000000000000000000000000000000000\
000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0303040
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
0000000000000000000000000000004F30305030300000000000000000000000002F1F1F302020\
0000000000000000000000000000004F30305F303F5F3F3F5F30305F3F3F40505F
4F505F40505F4F505F40505F4F505F40505F4F505FCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000CFCFCFCFCFCF\
CFCFCF40505F40505F40505F404F5F40505F40505F40505F404F5F40506040505F
40505F404F5F40506040505F405060404F5F4F506040505F3F404F0000000F1010000000000000\
0000000000002F303F303F40000000000000000000000000000000000000405060
2F2F30000000000000000000000000CFCFCFCFCFCF6F40301F100F2F1F100F0F0F0000006F4030\
0000000000000F0F00CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
0F0F0FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000
1F1010402F2F4F3030100F0F0000000000000000000000001F101050303F000000000000000000\
0000000000005F3F3F5030305F3F3F503030101F1F4050504F505F4050504F505F
4050504F505F405050CFCFCFC0C0C0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000CFCFCFC0C0C0CFCFCFC0C0C040505F3F4F5F40505F404F5F40505F
404F5F40505F404F5F40505F2F303F0F0F0F0000002F303F10101F0F0F0F0000000F0F0F000000\
0000000000000000002F3040404F5F0000000000000000000000000000004F5060
20202F000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF60402F6F4030000000\
000000000000100F0F0000005F402F000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C030304010101FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF00000000000000000000000050303F0F000F
00000000000000000000000000000000000050303F0000000000000000000000000000005F3F3F\
50303F5F3F3F4F505F0F0F0F1010104F505F40505F4F5F5F40505FCFCFCF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000303F3F303F402F2F300F0F0F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFCFCFCFCFCF40505F404F5F
40505F404F5F40505F404F5F405060303F4F2F2F301010100F0F0F0F10100F0F0F0000000F0F0F\
000000000000000000000000303F404F5060000000000000000000000000101010
4F5060101F1F000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
6F40306F403000000000000060402F0000003F2F1F000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF302F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF0000000000000000000F0F0F000000000000000000000000
00000000000000000050303F00000000000000000000000000000050303F50303050303F503030\
0F0F0F4050500F0F0F4050504F505FC0C0C0CFCFCFCFCFCF000000000000000000
0000000000000000000000000000000000000000000000000000000000000000000F0F10101010\
5F606F505F6050606F505F605F606F505F6050606F404F50303F401F20200F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
CFCFCF40505F3F4F5F40505F404F5F40505F202F3030404F404F5F40505F404F5F40505F000000\
0F0F0F000000000000000000000000404F5F303F40000000000000000000000000
1F1F204F5060000000000000000000000000000000000000000000000000C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF60402F6F4030000000000000000000000000000000000000
CFCFCFCFCFCFC0C0C0CFCFCF2F2F40000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000
50303F20101F0000000000000000000F000F5F3F3F50303F5F3F3F5030305F3F3F4F505F4F5F5F\
4F505F4F5F5F4F505FCFCFCFCFCFCF0000000000000000000F0F0F30303F505F60
2020201F1F200000000000000000000000000000000000000F0F0F4F505F5F606F5060605F606F\
505F605F606F5060605F606F505F605F606F505F605F606F505F605F606F505F60
5F606F505F601F2020000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000CFCFCFCFCFCF40505F404F5F405060303F400F0F0F0000000F0F0F00000F0F0F0F\
0000000F0F0F00000000000000000000000030404F4F5060000000000000000000
00000020202F4F5060000000000000000000000000000000000000000000000000000000CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F4F306F402F201F100000003F2F1F000000
6F4030000000CFCFCFCFCFCFCFCFCFCFCFCF30304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
00000000000000000000000000000000000000000000000050303F000000000000
0000000000000F00000F000F50303050303F50303050303F4F50504F505F4F50504F505F405050\
CFCFCFC0C0C0000000000000000000000000000000000000404F50505F5F505F60
505F5F505F60505F5F1F1F1F101010000000000000202F2F505F605F6060505F605F606F505F60\
506060505F605F606F505F60506060505F6050606F505F60506060505F6050606F
505F6050606F505F6000000000000000000000000000000000000000000020202F202F2F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000CFCFCFC0C0C040505F3F4F5F40505F404F5F40505F404F5F\
303F40303F401F20200000000000000000000000000F10104050601F1F20000000
0000000000002F2F304F5060000000000000000000000000000000000000000000000000000000\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C06F403060402F000000
000000100F0F60402F000000C0C0C0CFCFCFC0C0C0303040C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
0000000000000000000000002F1F2050303F000000000000000000000000100F0F
0F0F0F0000001F1010503030503F3F50303F4F5F5F4F505F4F5F5F4F505F4F5F5FCFCFCF000000\
0000000000000000000000000000000000000000002F303010101F303F3F000000
0F0F0F505F5F505F60505F5F505F605060605F606F505F605F606F5060605F606F505F605F606F\
505F605F606F505F605F606F505F605F606F505F605F606F505F605F606F505F60
404F504F5F5F5F606F0F0F0F000000000000000000000000000000000000000000505F60000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000CFCFCFCFCFCF404F5F40505F404F5F405060\
404F5F405060404F5F10101F0000000000000000000000000000003F4F5040505F
0F101000000000000040505F303F40000000000000000000000000000000000000000000000000\
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6F4F3060402F000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF302F40CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000000000\
0000004F2F303020200000000000000000000000000F00000F000F0000000F0000
100F0F50303F50303050303F4F50504F5F5F4F50504F505F4F5050CFCFCF000000000000000000\
000000000000000000000000000000000000101F1F50505F505F60505F5F505F60
4F5050505F604F5F5F404F4F505F605F606F505F605F6060505F605F606F505F605060602F3030\
2F2F302F2F30000000000000000000505F60505F60505F6050606F505F60000000
000000000000505F60505F601F20200000000000000000000000000000002F2F30505F6F505F60\
0F0F0F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000CFCFCFC0C0C040505F404F5F\
40505F404F5F40505F404F5F00000F0000000000000000000000000000001F1F1F
404F5F404F5F000000000000404F5F4F5060000000000000000000000000000000000000000000\
000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C06F403000000060402F000000CFCFCFCFCFCFCFCFCF100F10CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000002F1F1F4F2F30\
0000000000000000000000000000000F0F0F000000201F1F0F000F0F0F0F50303F
50303F4F505F505F5F4F505F4F5F5F4F50504F5F5FCFCFCF000000000000000000000000000000\
0000000000000000000000000000001F2020101F1F0F0F0F0000001F1F1F404F4F
505F60505F5F4F4F505060605F606F505F605F606F5060605F606F2F2F2F101010000000000000\
0F10101F20201F1F205F606F505F605F606F505F60404F50505F6030303F00505F
005F6000000030303F505F6050606F0000000000000000000000000000005F606F505F6050606F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000CFCFCF\
404F5F405060404F5F405060404F5F0F0F0F000000000000000000000000000000
2F303F40505F3F40500000000000002F2F304F5060101F1F000000000000000000000000000000\
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000006F4F30000000000000CFCFCFCFCFCF302F40000000\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF0000003020200F0F0F000000000000\
0000000000000F000F000000402F2F00000050303F00000050303F4F30304F5F5F
4F50504F5F5F4F50504F5F5FC0C0C0CFCFCF000000000000000000000000000000000000000000\
0000000000000000000000000000004F4F50505F60505F5F40404F3F4040505F60
4F505F5F6060505F605F6060505F60506060505F60101F1F0000000000003F404F4F505F505F60\
505F60505F605060600F0F0F404F501F20200F0F0F000000505F60000000006060
005F60000000505F60505F6F505F60000000000000000000000000000000505F60505F6F505F60\
404F4F000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000010101F4F50600000004F5060505F601F2020000000000000\
000000C0C0C040505F404F5F40505F404F5F404F5F000000000000000000000000
0000000000001F202F4050602F3040000000404F5F404F5F000000000000000000000000000000\
000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60402F000000000000000000C0C0C0CFCFCFC0C0C0\
002F40C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000004F2F301F10100000000000000000000F0F0F\
4F303F0F0F0F4F2F300F0F0F3F2F2F0F0F0F4F303050303F4F5F5F505F5F4F505F
505F5F4F505FCFCFCFCFCFCF000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F0000001010104F505F505F60505F5F
5F606F5060605F606F505F601F20200000000000003F40405F606F505F605F606F505F605F606F\
2020200000000000000000000F0F0F3F4040505F605F606F505F60000000000000
00000000000050606F505F6050606F00000000000000000000000000000050606F505F6050606F\
505F600F0F0F000000000000000000000000000000000000000000000000000000
0000002F30300000004F5F60000000101010505F6010101F1F1F1F505F60202020000000000000\
000000000000CFCFCF405060404F5F405060404F5F4F5060303F40000000000000
0000000000000000000000004F5060303F400000004050601F2020000000000000000000000000\
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F4F30000000000000000000CFCFCF\
CFCFCF30304F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF0000000000000000000000000000004F2F300F00004F2F30\
201F1F2F1F1F2F1F20402F2F50303F4F50504F5F5F4F50504F5F5F4F50504F5F5F
4F5050CFCFCF000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000003F3F40505F5F505F604F5F5F505F604F5F5F5F6060
505F605F6060505F601F1F2000000F2F3030505F60506060505F60506060505F60404F5000000F\
0000003F404F505F60505F60505F60505F60505F60505F60505F603F4040000000
000000505F60505F60505F604F505F000000000000000000000000000000505F60505F60505F60\
505F6F000000000000000000000000000000000000000000000000000000000000
0F0F0F4F505F0000004F505F1F20200000004F505F404F5F20202F505F604F5060000000000000\
000000000000000000000000CFCFCF404F5F40505F404F5F405060404F5F000000
000000000000000000000000000000000000000000000000404F5F000000000000000000000000\
000000000000000000000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000000000\
CFCFCFC0C0C030304F2F2F40CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF0000000000000000000000000000003F2F2F4F2F300F0F0F4F303F3F2F2F\
3F202F50303F4F303F0000004F5F5F505F5F4F5F5F505F5F4F505F505F5FCFCFCF
CFCFCF000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F202F2F4F505F505F5F5F606F506060
5F606F505F600000001F1F205F6060505F605F606F505F605F6060505F60505F5F101010404F50\
505F605F606F505F60506060505F6050606F505F60506060505F6050606F505F60
50606F505F6050606F505F601F1F200000000000000000000000001F1F2050606F505F60505F6F\
505F60000000000000000000000000000000000000000000000000000000000000
3F3F404F50601010103F404F505F60000000505F60505F600000004F5060505F60000000000000\
000000000000000000000000000000CFCFCFCFCFCF405060404F5F4F5060404F5F
1F202F000000000000000000000000000000000000000000101F1F404F5F10101F000000000000\
000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60402F000000\
000000000000CFCFCF30304F302F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF000000000000000000000000402F304F303F402F300F0000402F304F303F4F2F30\
4F303F0000000000004F50504F5F5F4F5050505F5F4F50504F5F5FC0C0C0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000303F3F5F6060505F605F6060
505F60000000505F60506060505F60506060505F605060603F404F0F0F0F000000506060505F60\
505F60505F60505F60505F60505F60505F60505F60505F60505F60505F60505F60
4F5F60505F60505F60505F60000000000000000000000000000000505F60505F60505F604F5F60\
1F2020000000000000000000000000000000000000000000000000000000000000
10101F505F604F5060505F604F505F505F604F5060505F604F505F505F604F506040404F000000\
000000000000000000000000000000000000CFCFCFC0C0C040505F404F5F405060
404F5F303040000000000000000000000000000000000000000000000000000000404F5F0F0F0F\
101F1F00000F000000000000000000000000000000000000000000C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
5F402F000000000000C0C0C030304F1F1F20000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
00000000000000000030202F4F303F4F2F3F4F303F4F2F304F303F4F303F4F303F4F303F505F5F\
505F5F404F4F0F0F0F505F5F4F5F5F505F5F4F505FCFCFCF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000010101F5F60605060605F6060505F60
0000004F50505F6060505F605F606F505F605F60604F505F0000003F404F506060505F605F606F\
505F60506060505F60506060505F60505F60505F6050606F505F60505F60505F60
50606F505F60505F60505F60000000000000000000000000000000505F60505F6F505F6050606F\
303F3F0000000000000000000000000000000000001F20201F1F201F1F1F0F0F10
50606F505F60505F6F505F605F606F505F6F50606F505F60505F6040404F404F50303F40000000\
000000000000000000000000000000000000000000000000CFCFCFCFCFCF404F5F
4F5060404F5F4F506010101F0000000000000000000000000000000000004F5060402F304F1F20\
4F50604F50602F3040000000000000000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF604F3000000030201F0000002F2F3F000000000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000
0000003F202F402F304F303F402F304F303F402F304F303F4F2F304F303F4F2F30505F5F4F5F5F\
505F5F4F5050505F5F4F5050505F5F4F5050CFCFCF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000005F6060505F605F60604F505F000000
505F605F6060505F60505F60505F60506060404F4F000000505F60505F60505F60505F60505F60\
505F60505F60505F60505F60505F604F5F60505F60505F60505F604F5F60505F60
505F60505F604F5F60505F60000000000000000000000000505F60505F604F5F60505F60505F60\
505F600000000000000000000000004F5F60505F60505F60505F6F505F60505F6F
505F60505F6F505F60505F6F505F60505F6F505F60505F6F505F60505F6F505F60505F6F505F60\
5F606F505F601F2020000000000000000000000000000000000000C0C0C0CFCFCF
404F5F405060404F5F4050602F303F0000000000000000000000002F303F404F5F4F0000400000\
4F505F404F5F4F5060000000000000000000000000000000000000000000000000
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF60402F0000001F100F000000000000CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000402F30
4F303F402F304F303F4F2F3F4F303F4F2F304F303F4F303F4F303F505F5F505F5F505F5F505F5F\
4F5F5F505F5F4F5F5F505F5FCFCFCF000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F606F5060605F6060505F603F40400F1010
5F6060505F605F6060505F60506060303F40000F0F4F505F506060505F60506060505F60506060\
505F60506060505F60505F60505F60506060505F60505F60505F60505F60505F60
505F60505F60505F60505F60000000000000000000000000505F60505F60505F6F505F60505F6F\
505F60505F6F4F505F505F6F505F60505F6F505F60505F6F505F6050606F505F60
505F6F505F6050606F505F60505F6F505F605F606F505F6F50606F505F605F606F505F6F5F606F\
505F6F5F606F505F6F5F606F0F0F0F000000000000000000000000000000000000
CFCFCF404F5F4F5060404F5F4F5060404F5F10101F0000000000000000000F0F0F4050604F5060\
40506040505F303040000000000000000000000000000000000000000000000000
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF503F2F603F20000000000000CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000000000000F0000402F304F303F
402F304F303F402F304F303F402F304F303F4F2F304F303F4F5050505F5F4F5F5F505F5F4F5050\
505F5F4F5050505F5FC0C0C0000000000000000000000000000000000000000000
000000505F5F5F60605F606F0F0F10303F3F0000000F0F0F000000000000000000000000000000\
0000000000000000000000001F1F1F505F605F6060505F602F2F2F0000005F6060
505F60505F60505F60506060505F60000000505F60505F60505F60505F60505F60CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000004F5F60505F604F5F60505F604F5F60\
505F604F5F60505F604F5F60505F60505F60505F6F4F505F404F50000F0F000000
0000000000000000000F0F0F0000000000000000000000003F3F40505F6F505F60505F6F505F60\
505F6F505F605F606F505F60505F6F505F6F1F2020000000000000000000000000
000000CFCFCF404F5F405060404F5F405060404F5F405060303040000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F303FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C060403000000060402F000000CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000004F303F402F304F303F402F30
4F303F402F3F4F303F402F304F303F4F303F000000404F40505F5F505F5F505F5F4F5F5F505F5F\
4F5F5F505F5FCFCFCF000000000000000000000000000000000000000000000000
0000000000005F60605F60605F606F5F60605F606F5F60605F606F3F4040000000000000000000\
0000000000000000000000005060605F6060505F605F60600F0F0F505F5F505F60
5F6060505F60506060505F603F404FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F6F\
505F60505F60505F60505F6F505F6020202F0000000F0F0F4F5060505F6F505F60
505F6F505F60505F6F505F6050606F505F60505F6F505F602F3030000000000000303F405F606F\
505F6F5F606F505F6F5F606F505F6F5F606F505F6F30303F000000000000000000
000000000000CFCFCF404F5F4F5060404F5F4F5060404F5F4F50604050604F5060404F503F4050\
30304000000000000000000000000010101F4F50603F404F4F50604F50604F5060
4F506F4F50604F5060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60402F604030000000
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF000000000000000000402F3F402F3040303F402F30402F3F
402F304F303F402F304F303F402F30000000000000505F5F4F5F5F505F5F4F5050505F5F4F5050\
CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000
0000000000000000001F1F1F5F60605F6060505F605F60605F60605F6060505F60000000000000\
0000000000000000000000005F6060505F605060602020201F1F1F505F60505F60
505F60506060C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF4F5F60505F604F5060505F60505F60505F6F4F5F60505F60
505F60505F6F4F5F60505F6F505F60505F6F505F60505F6F505F60505F6F303F40000000000000\
3F404F505F60505F6F505F605F5F6F505F60505F6F505F6F5F606F000000000000
000000000000000000CFCFCF404F5F4F5060404F5F405060404F5F4F5060404F5F4F5060404F5F\
4F5060404F5F4F50604050604F5060404F5F4F50604050604F50604050604F5060
4F50604F50604F50604F5060CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C060402F
000000CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF00000000000000000040303F402F3F40303F402F304F303F402F3F
4F303F402F304F303F502F20505F5F000000000000505F5F505F5F505F5F505F5F505F5FCFCFCF\
CFCFCF000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F60605F606F5F60605F60605F60605F606F5F60603F3F3F\
000000000000505F5F5F6060505F605F6060505F605F6060505F605F6060CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F6F505F60505F6F505F60505F6F505F60
505F6F505F60505F6F505F60505F6F505F6050606F505F60505F6F505F605F606F505F6F5F5F6F\
000000000000202F305F606F505F6F5F606F505F6F5F606F505F6F5F606F000000
000000000000000000000000CFCFCFCFCFCF4F5060404F5F4F5060404F5F4F50604050604F5060\
4050604F50604F50604F50604F50604F50604F50604F50604F50604F50604F5060
4F50604F50604F5F6F4F50604F506FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
604F30000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF000000000000000000402F3F402F3040303F402F30402F3F402F304F303F
402F304F303F5000005040401010100000003F3F3F505F5F4F5050505F5F4F5050CFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F5F6060505F605F6060505F605F6060505F60\
5F6060505F605F6060505F60506060505F605F6060505F5FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0505F604F5F60505F60
4F5060505F604F505F404F5030303F0F101020202F3F3F404F505F505F6F505F60505F6F505F60\
505F6F101F1F0000002F303F505F6F505F605F5F6F505F605F5F6F505F6F40404F
000000000000000000000000000000CFCFCFC0C0C04F5060404F5F4F5060404F5F4F5060404F5F\
4F5060404F5F4F5060404F5F4F50604050604F5060404F604F50604F50604F5060
4F50604F50604F50604F50604F5060CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C060402F000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
000000000000000000201F20402F3F40303F402F304F303F402F3F4F303F402F30
505F5F505F5F505F5F000F0F000000000000505F5F505F5F505F5F505F5FCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000001F20205F606F5F60605F6060505F605F6060\
5F60605F6060505F605F6060505F60CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F40
00000000000020202F2F2F3030303F1F1F20505F6F4F505F404F5030303F0000000000000F0F0F\
3F404F5F5F6F505F6F202020505F6F5F606F505F6F5F606F505F6F5F606F505F6F
000000000000000000000000000000000000CFCFCFCFCFCF4F50604050604F5060404F5F4F5060\
4F50604F50604050604F50604F50604F50604F50604F50604F50604F50604F5060
4F506F4F50604F50604F5060505F6F4F5060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF30304F604030201F10CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000\
000000000000000000402F30402F3F402F30402F3F402F3040303F402F30000000
000000000000000000000000000000505F5F4F5050505F5F4F5F50CFCFCF000000000000000000\
0000000000000000001010105F60605F60603F40404F4F4F5F60605F60605F6060
0F0F0F0000000000000000000000000000000000000000000000005F6060505F605F6060505F60\
5F6060505F5F5F6060CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF505F6F4F5060505F60505F60505F6F505F60505F6F505F60505F6F505F604040500F0F0F\
0000000F0F0F404F50505F60505F6F505F60505F6F505F6F5F5F6F505F605F5F6F
505F60000000000000000000000000000000000000CFCFCF404F5F4F5060404F5F4F5060404F5F\
4F50603F404F4F5060202F303F3F4F0000000000000F0F0F404F5F10101F303F40
000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF2F2F4060402F5F402FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000\
000000000000402F3F40303F402F304F303F402F3F4F303F000000202F2F202020
0000000000000000000000004F5050505F5F505F5F505F5FCFCFCF000000000000000000000000\
0000001010104F5050606F6F5F6060606F6F5F60605F60605F606060606F5F6060
5F60603F3F3F0000000000000000000000000000000000000000000000005F60605F60605F6060\
505F60CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF505F6F505F60505F6F505F60505F6F505F60505F6F505F605F5F6F505F6F505F6F\
505F601F2020000000101F1F4F505F5F606F505F6F5F606F505F6F5F606F505F6F
5F606F101010000000000000000000000000000000000000CFCFCF404F5F4F50604F50604F5060\
404F6030303F0F1010000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF30304F6040305F402FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F2F3F000000000000000000\
0000003F2F30402F3F402F30402F3F402F3040303F000000505F500F0F0F000000
0000000000000000002F3030505F50505F5F505F50CFCFCF000000000000000000000000000000\
2F30305F606060606F5F60605F6060404F4F0000000000000000005F60605F6060
5F60605F60605F60600F0F0F0000000000000000000000000000000000000000005F6060505F5F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF4F5060505F60505F60505F6F505F60505F6F505F60505F6F505F60\
505F6F505F60505F6F20202F000F0F505F6030303F101010505F6F505F6F5F606F
505F601F1F20000000000000000000000000000000000000000000CFCFCF404F5F4F5060404F5F\
4F5060404F5F4F5060404F5F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF2F2F400000005F402FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40303F3F2F3040303F000000000000\
00000040303F402F3040303F402F3F0000000000005F605F000000000000000000
0000000000001F1F1F505F5F505F5F505F5FCFCFCF0000000000000000000000000000004F4F4F\
5F6060606F6F5F6060606F6F5F60602F2F2F001F1F006060000000000000202020
5F606F5F60605F60605F60603F4040000000000000000000000000000000000000CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F6F505F60505F6F505F60505F6F505F605F5F6F\
505F6F5F5F6F505F605F606F505F6F00000000303F00606F000F0F505F60505F6F
3F404F000000000000000000000000000000000000000000000000000000CFCFCF404F5F4F5060\
4F50604F50604F50604F50604F50604F5060404F5F1F1F1F10101F000000000000
000000000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F0000005F402FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F2F3F3F2F30402F3F100F0F000000000000\
402F3F3F2F30402F3F402F30000000000000505F5F000000000000000000000000
000000000000505F50505F5F505F50CFCFCF0000000000000000000000001F202060606F5F6060\
60606F5F60605F60605F6060606060101010003F3F006060006060000000000000
5F60605F60605F5F605F60605F5F605F6060000000000000000000CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0505F60505F60505F6F505F60505F6F505F60\
505F6F505F60505F6F3F404F00000000404F005F6F000000000000505F605F5F6F
000000000000000000000000000000000000000000000000000000000000C0C0C04F5060404F5F\
4F5060404F5F4F5060404F5F4F5060404F5F4F50604F50604F50604F4F604F5060
20202F4F5060303040000000000000000000000000000000000000000000C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF0000000000000000005F402FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40303F3F2F3040303F3F2F3000000000000040303F\
402F3040303F402F3F0000002F2F2F5F60600000003F40401F1F1F000000000000
000000505F5F505F5F505F5FCFCFCF000000000000000000000000303F3F606F6F5F6060606F6F\
5F6060606F6F5F606060606F5F6060606F6F000000000000000000000000303F3F
5F60605F60605F60605F60605F60605F60605F6060000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F6F505F60505F6F505F60505F6F\
505F605F5F6F505F6F20202F00000000000000000000000F505F6F5F606F505F6F
000000000000000000000000000000000000000000000000000000000000000000CFCFCF4F5060\
4F4F604F50604F50604F50604F50604F50604F50604F50604F50604F50604F5060
4F50604F506050506F4F50604F50603F3F4F0F0F0F000000000000000000000000CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F0000000000005F402FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F2F3F3F2F303F2F3F3F2F30402F3F0000001010103F2F30\
402F3F402F30000000000000505F5F0000003F4040202020000000000000000000
505F50505F5F505F50CFCFCF00000000000000000000000030303060606F5F606060606F5F6060\
6060605F60606060605F60603F40403030305F60605F6060404F4F3030305F6060
5F5F605F60605F60605F6060505F605F6060C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000505F6F505F60505F6F505F60\
505F6F505F60505F6F505F60505F6F505F605F5F6F505F60505F6F505F6F000000
000000000000000000000000000000000000000000000000000000000000000000CFCFCFC0C0C0\
4F5060404F5F4F5060404F5F4F50604F4F604F5060404F5F4F50604F50604F5060
4F50604F50604F50604F50604F50604F50604F50600F0F0F000000000000000000CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF3F303F3F2F3040303F3F2F3040303F0000000000003F2F3040303F\
402F3F000000101F105F60600000003F40402F2F2F000000000000000000505F5F
5F5F5F505F5FCFCFCF0000000000000000000000002F2F2F606F6F606060606F6F5F6060606F6F\
5F6060606F6F5F6060606F6F000F0F2F30305F606060606F5F60605F60605F6060
5F60605F60605F60605F60605F6060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000002F2F30000000000000
0000000000000000000000000000000000001F1F1F000000000000000000000000000000CFCFCF\
CFCFCF4F50604F4F604F50604F50604F50604F50604F50604F50604F50604F5060
4F50604F50604F50604F506050506F4F506050506F4F5060000000000000000000000000CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F304F000000000000000000000000CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C03F2F303F2F303F2F3F3F2F303F2F3F3F2F300000003F2F30402F3F3F2F30\
0000000000005F5F5F000000404F4F2F3030000000000000000000505F50505F5F
505F50CFCFCF0000000000000000000000001F1F1F60606F5F6060606F6F5F60606060603F3F3F\
0000001F1F1F5F60602F30304040405F60605F60605F60605F60605F60605F6060
5F60605F60605F5F605F6060CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005050605F606F000000000000000000000000000000000000000000\
CFCFCFC0C0C04F50604F4F5F4F5060404F5F4F50604F50604F50604F4F604F5060
4F50604F50604F50604F50604F50604F50604F5060202F30000000000000000000000000CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF2F2F4F000000000000000000000000\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF0000003F2F303F303F3F2F3040303F3F2F3040303F3F2F3040303F402F3F000000\
1010105F6060000000000000505F5F0000000000000000003F3F3F5F605F505F5F
CFCFCFCFCFCF000000000000000000000000606F6F606060606F6F5F6060606F6F5F6060000000\
1F2020606F6F0000000F0F0F5F606060606F5F60606060605F60606060605F6060
5F60605F60605F6060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
CFCFCFCFCFCF4F50604F50604F50604F50604F50604F50604F50604F50604F5060
4F50604F506050506F4F50605050604F50604F50600F0F10000000000000000000000000000000\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F000000000000000000CFCFCF\
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C00000000000003F2F3F3F2F303F2F3F3F2F30402F3F3F2F30402F3F000000000000000000\
5F605F000000000000505F5F000000000000000000000000505F5F505F50CFCFCF
C0C0C00000000000000000000000000F10105F6060606F6F5F60606060605F60600000001F1F1F\
6060602F2F2F0000004040405F60605F60605F60605F60605F60605F60605F6060
5F5F605F6060C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000C0C0C0CFCFCF404F5F4F50604F4F604F50604F4F5F4F50604F50604F5060
4F50604F50604F50604F50604F50604F50603F3F4F000000000000000000000000000000000000\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C030304F000000000000000000\
CFCFCF503F2FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
0000000000002F202F3F2F3040303F3F2F3040303F3F2F3040303F0000000000000000005F6060\
000000000000505F5F1010100000000000000000005F605F505F5F5F5F5FCFCFCF
000000000000000000000000000000606060606F6F5F6060606F6F6060602F2F2F0F0F0F606F6F\
5F60600000000F0F0F606F6F5F60606060605F60606060605F60605F60605F6060
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F20204F3F30302F20000000000000000000000000000000000000000000000000000000\
000000CFCFCFCFCFCF4F50604F50604F50604F50604F50604F50604F50604F5060
5050604F50605050604F506050506F3F3F4F000000000000000000000000000000000000000000\
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F000000000000\
CFCFCFCFCFCF5F402FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000
0000000000000F0F0F3F2F3F3F2F303F2F3F3F2F303F2F3F0000000000000000005F605F000000\
0000001010104F504F000000000000000000505F50505F5F505F5FC0C0C0000000
0000000000000000000000003F3F3F606F6F5F60606060605F6060505F5F0000006060605F6060\
6060600000004F4F4F5F60606060605F60605F60605F60605F60605F5F60CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000001F1010
705F4F8F70608F6F5F8F6F60000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000004F50604F50604F50604F4F604F5060
4F50604F50604F50604F50604F50603F3F4F000000000000000000000000000000505060101010\
000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF2F2F4F10101F000000\
000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
000000000000302F303F2F303F303F3F2F3040303F0000000000001F1F1F5F5F5F000000000000\
0000005F605F0000000000000000000F0F0F505F5F5F5F5FCFCFCFCFCFCF000000
000000000000000000000000606060606060606F6F606060606F6F0000000000005F6060606F6F\
0F0F0F101F1F5F60606060605F60606060605F60606060605F6060CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000201F1F705F4F403F30000000
10100F5F403F9070608F6F607F6050000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000004F50605050604F5060
0F0F0F4F4F5F50506F4F50605050603F404F0000000000000000000000004F505F4F5060101010\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F000000\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000
0000000000003F2F303F2F3F3F2F303F2F3F0000000000000F0F0F5F605F000000000000000000\
5F5F5F0000000000000000003F4040505F5F5F5F5F505F50CFCFCF000000000000
0000000000000000000F0F0F5F60606060605F6060606F6F2F2F2F0000004F4F4F6060605F6060\
0000004F4F4F6060605F60605F60605F60605F60605F6060CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000705F508F6F5F7F5F50
201F100000008F6F5F9070608F6F5F705F502F201F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
4F50600F0F0F0000003F404F4F50600000000000000000000000001010104F50604F5060000000\
000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C030304F2F2F4F\
000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000
0000003F2F303F303F3F2F3040303F0000000000000000005F60600000000000000000005F6060\
0000000000000000003F3F3F505F5F5F605F505F5FCFCFCF000000000000000000
0000000000000000004F5050606F6F606060606F6F6060600000002F3030606F6F5F6060606F6F\
0000006060605F6060606F605F60606060605F6060CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000201F1F0F0F0F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000302020201F1F0F0F0F403F309070608F6F60
907060503F3F705F508F6F5F9070608F6F60907060000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000001F1F200000000F0F0F0000000000000000000000004F4F5F505F6F000000000000\
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F304F\
30304F1F1F2F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000000000000000000000000000
302F303F2F3F3F2F303F2F3FCFCFCF0000000F0F0F5F605F0000000000000000004040402F2F2F\
000000000000101010505F5F5F5F5F505F50CFCFCF000000000000000000000000
0000000000003F40405F5F5F606060606F6F5F6060606060000000505F5F5F60606060605F6060\
000000505F5F5F60605F60606060605F6060CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003F302F8060603F302F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000100F0F7F5F50201F1F000000000000705F4F8F6F60
8F6F5F50403F0000003F302F80605F9070608F6F5F705F4F2F201F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000F10105050601F1F20000000000000\
000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
30304F000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000001F1F1F302F30302F300F000F
3F303F3F2F303F303FCFCFCF0000001010101010100000000000000000001010105F5F5F000000\
0000000000005F5F5F5F605F505F5F5F605FCFCFCF0000000000000000001F1F1F
4F4F4F5F60604F5050606060606F6F606060606F6F0000001F20205F6060606F6F6060602F2F2F\
000000606F6F5F60606060605F6060606060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000201F1F100F0F000000\
0000000000000000000F0F0F000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000001F1010907070906F6F7F5F5F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000F0F0F8F6F605F4F400F0F0F0F0F0F4F3F30
9070608F6F602F201F000000100F0F8F6F5F9070608F6F60907060201F1F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000020202F4F506020202F000000000000\
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
2F304F30305020203F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000302F303F2F3F0000003F2F3F
302F303F2F3FC0C0C00000005F5F5F0F0F0F0000000000000000000000004F504F000000000000\
0000004040405F605F505F505F5F5FC0C0C00000000000000000001F1F1F5F6060
5F5F5F5F60604F5050606F6F5F60606060605F60600000005050506060605F60606060600F0F0F\
0000005F60606060605F60605F6060C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000001F1010000000000000604F504F3F404030302F20208F6F7080606F\
6F505F503F403F2F2F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000001F10108F6F6F90706F8F6F6F\
6F5050101010000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000050403F0000000000000F00009070608F6F5F80605F302F20503F3F
8F6F5F9070608F6F5F5F403F0F00008F6F608F6F5F9070608F6F5F907060000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005050600F0F0F000000000000\
000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF2F304F30304F2F2F40000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000000000302F303F1F1F0000003F303F3F2F30
3F303FCFCFCFCFCFCF5F605F0000000000000000000000000000005F5F5F0000000000001F1F1F\
0F0F0F404F40505F5F5F605FCFCFCF0000000000002F2F2F5F60606060603F3F3F
2020205F605F4F5050606060606F6F6060603F3F3F1F1F1F606F6F606060606F605F6060404040\
0000006060605F6060606F60CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000004F3F403F30302F20208F6F7080606F806F6F5F4F4F8F6F70806F6F8F6F70806F6F8F6F70\
8F6F6F8F6F708F6F6F8F6F7080606F000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008F6F6F8F6F6F907070\
90706F90706F50403F0F0F0F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000604F400F0F0F0000000000008F6F608F6F60907060000000
201F1F7F5F509070608F6F600F0F0F2F201F8F6F5F8F6F5F9070608F70606F5F4F201F1F000000\
0000000000000000000000000F0000302F205F4F3F0000000F0F00000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF30305000000030304F0F0F10CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF000000000000000000300000000000000000302F303F2F3F
CFCFCFCFCFCF5F5F5F0000000000000000000000000000005F5F5F0000000000000F0F0F000000\
0F0F0F505F505F5F5F505F50CFCFCF0000002020205F60604F5050000000000000
5050505F6060505F5F606F60606060606F601F1F1F505F5F5F60606060605F60606060604F5050\
0000005F6060606060CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F3030403F3F
4F3F3F806F7080606F806F6F80606F806F7080606F806F6F80606F8F6F7080606F8F6F6F806F6F\
8F6F7080606F8F6F6F806F6F8F6F7080606F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000604F4F90706F8F6F6F\
90706F8F6F6F90706F6F50500F0F0F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000005F4F408F6F608F6F5F3F2F20000000100F0F6F504F8F6F5F1F1010
0000000F0F0F8F6F5F9070607F60508F70608F6F5F9070608F6F5F9070608F6F5F907060604F40\
4F3F30000000100F0F6F50409070609070605F4F3F9070609F7F60100F0F1F1010
1F1F10000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C030304F2F304F202F3F000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF0000000000003F2F30302F30000000302F303F303FCFCFCF
CFCFCFCFCFCF0000000000000000000000000000005F605F0000000000000000000F0F0F0F0F0F\
0000005F605F5F5F5FCFCFCF000000000000000000404040000000505F5F5F6060
6060605F605F606F6F606060606F6F6060600F0F0F505F5F606F60606060606F6F6060605F6060\
000000606F605F6060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F20202F2020302F2F302F2F806F7080606F
806F7080606F806F7080606F806F7080606F8F6F70806F6F8F6F70806F6F8F6F70806F6F8F6F70\
806F6F8F6F708F6F6F8F6F708F6F6F8F6F708F6F6F5F4F4F000000000000000000
00000000000000000000000000000000000000000000000000000000000000000090706F907070\
906F6F90707090706F9070706F50500F0F0F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001010109070608F6F609070606F504F0000006F504F9070608F6F60
6F504F806F5F9070608F70609070608F6F609070608F70609070608F6F5F705F506F504F9F7F6F\
4F3F307F60509070609F7F609070609F7F6F9F7F609F7F609070609F7F60907F60
201F1F2F201F302F20000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000505060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF30304F0000002F304F30304FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0000000000000000000302F30000000302F303F2F3FC0C0C0CFCFCF
C0C0C00000000000000000000000000000004F4F4F000000000000000000000000505F5F0F0F0F\
101010505F5FCFCFCFC0C0C00000000000000000000000005F60605F605F1F1F1F
0000005F6060505050606F6F5F60606060601010106060605F60606060605F60606060605F6060\
0000005F6060CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001010101010101F101F1F1F1F8060707F606F80606F7F606F806F70
7F606F80607080606F806F7080606F806F7080606F806F7080606F806F7080606F8F6F7080606F\
8F6F70806F6F8F6F70806F6F8F6F708F6F6F8F6F708F6F6F6F505F000000000000
0000000000000000000000000000000000000000000000000000000000000000002F20208F6F6F\
90706F906F6F9070708F6F6F705050000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000009070608F6F60907060000000201F1F80605F907060
8F6F5F9070608F6F609070608F6F5F806F5F705F4F9070607F5F508F6F5F9070609F7F6F907060\
9F7F609070609F7F609070609F7F609070609F7F609070609F7F609070609F7F60
90705F9F7F6090706040302F40302F000000000000000000000000000000000000000000000000\
000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C030304F2F304F30304F2F2F4F0F0F0FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF0000000000000000000000000000001010103F303FCFCFCFCFCFCFCFCFCF
CFCFCF0000000000000000000000005F5F5F0000000000000000000F0F0F0F0F0F4040400F0F0F\
2F302F5F605FCFCFCFCFCFCF0000000000000000005050501010100F0F0F404F4F
6060605F60605F6060606060606F6F606060000000606060606F6F606060606F605F6060606F60\
404040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F0F100F0F806F707F606F8060707F606F806F7080606F806F7080606F
806F7080606F806F7080606F8F6F70806F6F8F6F7080606F8F6F70806F6F8F6F70806F6F8F6F70\
8F6F6F8F6F708F6F6F8F6F708F6F6F80606F8F6F6F8F70708F6F6F8060601F101F
0000000000000000000000000000000000000000000000000000000000000000000F0000907070\
90706F90707090706F90707090706F705F50000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000040302F9070608F6F60705F508F6F609070608F6F60
9070608F70609070608F6F5F7F605F8F6F5F7F60508F6F5F9F7F6F9F7F609F7F6F9070609F7F6F\
9F7F609F7F6F9070609F7F6F9F7F609F7F609070609F7F6F9F7F609F7F60907060
9F7F60907F609F7F609070609F7F605040305F403F5F4F3F000000000000000000000000000000\
000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF303F50000000303050202F40000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF0000000000000F0F0F0000000000003F2F3F302F30CFCFCFC0C0C0CFCFCF
000000000000000000000000505F50404F400000000000000000005F5F5F0000002F2F2F000000\
5F605FC0C0C0CFCFCF000000000000000000000000000000000000505F505F6060
1010105F60605F6060606F60606060606F600000002F30305F60606060605F60606060605F6060\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
7F60707F606F7F60707F606F8060707F606F8060707F606F8060707F606F806070
80606F806F7080606F806F7080606F806F7080606F806F7080606F8F6F7080606F8F6F70806F6F\
8F6F70806F6F80606F604F506F50507F60608F6F708F6F6F8F6F708F6F6F8F6F6F
0000000000000000000000000000000000000000000000000000000000000000000F0F0F8F6F6F\
9070708F6F6F90706F906F6F907070906F6F705F50000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000004030308F6F5F9070608F6F609070608F6F60806F5F
8F6F609070608F6F5F9F7F6F9070608F6F609070609F7F6F9070609F7F6F9070609F7F60907060\
9F7F6F9070609F7F609070609F7F609070609F7F609070609F7F609070609F7F60
9070609F7F6090705F907F6090705F9F7F6090705F5F4F3F40302F302F20000000000000000000\
000000000000000000000000000000000000000000000000CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF2F304F3030502F304F2F2F402F2F4FCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
0000000000000000000000000000003F303F302F30CFCFCFCFCFCFCFCFCFCFCFCF
0000000000000000002020205F5F5F0000000000000F0F0F0F0F0F4F504F0F0F0F1010105F605F\
5F5F5FCFCFCFCFCFCF0000000000000000000000002020205F60602F2F2F000000
5F60605F605F606060606060606F6F606060000000101010606F60606060606F6F606060606F60\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000002F202050404F\
7F606F8060707F606F8060707F606F806F707F606F806F707F606F806F7080606F
806F7080606F806F7080606F806F7080606F8F6F70806F6F8F6F70806F6F8F6F70806F708F6F70\
806F6F806F6F302F2F0000000000000F0F0F302F2F806F6F8F6F6F9070708F6F6F
8F707000000000000000000000000000000000000000000000000000000000000000000060504F\
90706F90707090706F90707090706F90707090706F705F50000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000002F202080605F907060806F5F8F7060806F5F
9070609070609F7F6F9F706F9F7F6F9F70609F7F6F9F7F609F7F6F9F70609F7F6F9F7F609F7F6F\
9F70609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F60907F609F7F609F7F60
9F7F60907F609F7F60907F609F7F609070609F7F60907F609F7F6090705F10100F000000000000\
000000000000000000000000000000000000000000000000000000CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF2F2F3F30304F202F3F2F2F4030304FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
000000000000100F10000000000000302F30CFCFCFC0C0C0CFCFCFC0C0C0000000
0000000000000000005F605F0000000000000000000F0F0F0000003F3F3F0F0F0F404F40505F50\
CFCFCFC0C0C00000000000000000000000000000000000000000000F0F0F606060
5F5F5F505F5F101010606F606060606060605F5F5F1010105F60606060605F6060606060C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007F60707F606F\
7F60707F606F8060707F606F8060707F606F8060707F606F8060707F606F806F70
7F606F806F7080606F806F7080606F806F7080606F806F7080606F806F70806F6F8F6F7080606F\
8F6F7080606F000000000000001F1F0000000000001F10108F6F6F8F6F6F8F6F70
8F6F6F8F6F700000000000000000000000000000000000000000000000000000000000002F201F\
907070906F6F907070906F6F90706F90706F90707090706F60504F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000201F1F8F6F609F7F6F8F6F5F9F7F6F
9070609F7F6F9070609F7F6F9070609F7F6F9070609F7F6F9070609F7F6F9070609F7F6F907060\
9F7F609070609F7F6F9070609F7F609070609F7F609070609F7F609070609F7F60
9070609F7F609070609F7F6090705F907F6090705F9F7F6090705F907F6090705F90705F000000\
000000000000000000000000000000000000000000000000000000000000000000
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C030304F2F304F3030502F2F4F000000C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000302F3F000000000000302F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
0000000000006060600000000000000000000F0F0F5F5F5F101010202F204F4F4F5F5F5F5F605F\
CFCFCFCFCFCF000000000000000000000000000000000000303030606060000000
000000000000101010606060606F6F606060606F60000000606060606060606F60606060CFCFCF\
CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F4F50705F6F806070\
7F606F8060707F606F806F707F606F806F707F606F806F70806070806F7080606F
806F70806070806F7080606F8F6F70806F708F6F70806F6F8F6F70806F708F6F70806F6F8F6F70\
8F6F708F6F70000000000F0F006F7000606F0000000000004F3F3F8F70708F6F6F
9070708F6F709070700000000000000000000000000000000000000000000000000000000F0000\
80606090707090706F90707090706F90707090706F9F7070604F4F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000604F409070609F7F6F9F706F9F7F6F9F706F
9F7F6F9F7F6F9F7F6F9F706F9F7F6F9F7F6F9F7F6F9F70609F7F6F9F7F609F7F6F9F7F609F7F6F\
9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F60907F60
9F7F60907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F5F705F4F\
5F4F3F000000000000000000000000000000000000000000000000000000000000
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000303F5030304F30304F000000CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000000000\
1F1F20000000000000302F303F2F3FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000
0000005F605F0000000000000000004F504F0000005F605F2020201F1F1F5F5F5F5F605FCFCFCF\
CFCFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000606F60606060606F605F6060000000505F50606F605F6060CFCFCFCFCFCF\
CFCFCF0000000000000000000000000000003F303F000000000000000000000000
000000000000000000000000000000000000000000000000000000100F0F7F5F6F7F60707F606F\
7F60707F606F7F60707F606F8060707F606F8060707F606F806F707F606F806070
80606F806F7080606F806F7080606F806F7080606F806F7080606F8F6F7080606F8F6F70806F6F\
8F6F7040303F000000006060006F700060600040400000001F10108F6F6F8F6F70
8F6F6F8F70708F6F6F2F2020000000000000000000000000000000000000000000000000000000\
4F3F3F8F6F6F90707090706F907070906F6F90707090706F90706F100F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000201F1F6F504F90706F9F7F6F90706F9F7F6F9070609F7F6F
90706F9F7F6F9070609F7F6F9070609F7F6F9070609F7F6F9070609F7F6F907060907F606F5040\
9F7F6F9070609F7F608F6F5F7F5F5040302F9F7F60907060806F5F9070609F7F60
9070609F7F609070609F7F609070609F7F6090705F907F6090705F9F7F6090705F907F6090705F\
907F6040302F201F1F100F0F000000000000000000000000000000000000000000
000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF3030502F304F0000002F304F000000C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000\
0000000000000000003F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
3F403F0000000000000000005050505F5F5F0F0F0F505F505F5F5F5F5F5F5F605F5F5F5FCFCFCF\
CFCFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000606060606F6F606060606F6F606060202020606060606F60CFCFCF000000\
0000000000000000000000000F0F0F2F20203F303F705F6F5F4F50705F6F5F4F50
000000201F20000000201F1F00000000000000000000000060505F7F606F7F60707F60707F6070\
7F606F806F707F60708060707F606F806F707F6070806F707F606F806F70806070
806F7080606F50404F5040405F4F5080606F8F6F70806F708F6F70806F6F8F6F70806F708F6F70\
806F701F101F000000004040006F6F0070700040400000000000009070708F6F70
9070708F6F6F9070702F2020000000000000000000000000000000000000000000000000000000\
00000090707090706F90707090706F9F707090706F90707090706F50403F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000302F2F705F509F7F6F9F706F9F7F6F9F7F6F9F7F6F9F706F9F7F6F9F7F6F
9F7F6F9F706F9F7F6F9F7F6F9F7F6F9070606F504F9F7F6F5F403F2F201F1F1010000000000000\
5040303F302F0F0F0F0000000000004F3F30302F201F1F10806F506F5F4F3F302F
9F7F60907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F609F7F60\
907F5F9F7F60907F5F10100F000000000000000000000000000000000000000000
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF30304F00000030304F303F5020203F000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000000000000000000F000F\
000000000000302F3FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000101010
1F1F1F0000000000000000005F5F5F5F605F1010105F605F5F5F5F5F605F5F5F5F5F605FC0C0C0\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000404040606060606060606060606F605F6060606060000000000000000000\
0000000000000000000000000000006F50606F5F6F6F50606F5F6F6F5F60705F6F
6F50605F4F506F5F605F4F5050404F1F1F1F0000001F1F1F3F2F307F60707F606F7F60707F606F\
7F60707F606F8060707F606F7F60707F606F8060707F606F8060707F606F40303F
4030304F3F404F3F40000000000000806F7080606F806F7080606F806F70806F6F50404F6F505F\
8F6F7010101000000000303F006F70006F6F003F3F0000000000000F0F0F8F7070
8F6F6F8F6F708F6F6F403030000000000000000000000000000000000000000000000000000000\
0000008F6F6F907070906F6F90707090706F90707090706F907070806060000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000007F5F5090706F9F7F6F9F706F9F7F6F90706F9F7F6F9F706F9F7F6F
9070609F7F6F9F706F9F7F6F705F500F0F0F605040302F20000000000000000000000000000000\
0000000000000000000F0F00000000000000000000000000000000000000000000
3F2F20403F2F907F609F7F6090705F9F7F60907F609F7F6090705F907F6090705F907F6090705F\
907F6090705F907F608F705F70604F000000000000000000000000000000000000
000000000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF30304F3030501F1F2F2F304F000000CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000302F3030303F\
0000002F2F30302F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000101F103F3F3F
0000000000000000005F605F6060605F605F5F605F5F5F5F60605F5F5F5F5F605FCFCFCFCFCFCF\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000001F1F1F606F6F606060606F60606060606F6F000000000000000000000000\
0000000000000000000000005F4F506F5F6F6F5060705F6F6F5F6F705F6F6F5F6F
705F6F6F5F6F705F6F6F5F6F70606F705F6F50404F7F606F7F60707F60707F60707F606F806070\
7F60708060707F606F806F707F6070806F707F606F3F2F30302F2F3F303040303F
0000000000000000000000002F202F2F20208F6F70806F708F6F70806F707F606F5040404F3F40\
806F6F8F6F70000000000F0F004F50006F70004F50000000000000100F0F8F6F70
9070708F6F709070708F6F6F000000000000000000000000000000000000000000000000000000\
00000000000090706F90707090707090707090706F9F707090706F9F7070302020
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2020806F5F9F706F9F7F709F7F6F9F7F6F9F706F9F7F6F9F7F6F9F7F6F9F706F
9F7F6F9F7F6F8F6F601F101000000040302F000000000000100F0F2F201F4F3F30000000100F0F\
2F201F6F504F806F5F9F7F6F2F201F6050406F504F000000000000000000000000
0000000000004F40305040305F4F3F907F609F7F60907F609F7F60907F609F7F60907F5F9F7F60\
907F609F7F60907F5F9F7F60907F5F3F302F1F1F10000000000000000000000000
000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303F50101020202F3F303050000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF0000000000001F1F20000000000000\
000000302F30CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000
0000000000005F5F5F5F605F5F5F5F5F605F5F5F5F5F605F5F5F5F5F605F5F5F50CFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000505050606F60606060606060000000000000000000000000000000\
0000000000000000000000000000000F0F0F0000001F101F6F5F6F60505F6F5F6F
6F5060705F6F6F5F60705F6F6F5F60705F6F50404F6F5F607F606F7F60707F606F7F60707F606F\
7F60707F606F7F60707F606F7F607050404F4F3F40302F2F000000000000000000
0000002F20202F1F202F20202F2020806F7080606F806F70806F6F0F0F0F0000000000004F3F3F\
8F6F70806F6F0F0F0F0000000000000000000000000000000000003020208F6F70
8F6F6F8F70708F6F6F8F7070000000000000000000000000000000000000000000000000000000\
000000000000302F2F90706F90707090706F90707090706F90707090706F302F20
0000000000000000000000000000000000000000000000000000000000000000000000005F4F40\
9F7F6F90706F9F7F6F90706F9F7F6F90706F9F7F6F9F706F9F7F6F90706F9F7F6F
604F402F20200000000000000000000000002F201F6F5040806F5F9F7F6F403F308F6F5F9F7F60\
9F7F6F9070609F7F6F9F7F609F7F6F9070609F7F607F60508F6F5F9070600F0F0F
1F1010201F1F000000000000000000605040604F3F9F7F60907F5F9F7F6090705F907F60907F5F\
907F6090705F907F60907F5F907F6090705F201F10000000000000000000000000
000000000000000000000000000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF30304F30305030304F303F50000000000000
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000\
302F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000
0000002020206060605F605F60605F5F605F60605F5F605F5F605F5F5F5FCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F5F605F000000000000000000000000000000000000000000\
000000000000000000201F20302F301F1F1F0000000F0F0F000000302F300F0F0F
705F6F6F5060705F6F705F6F705F6F5F4F507F60707F60707F60707F606F7F60707F60707F6070\
7F6070806F707F6070806F707F606F0000000000000000000000001F101F1F1F1F
2F20202F2020806F70806F70806F70806F705F4F5040303F302F2F1F101F201F203F30305F404F\
6F505F8F7070302F2F0000000000000000000000000000000000007F60608F6F70
9070708F6F709070708F6F708F6F6F000000000000000000000000000000000000000000000000\
00000000000000000090707090706F9F70709070709F707090706F9F7F70705F50
0F00000000000000000000000000000000000000000000000000000000000000002F201F9F7F70\
9F706F9F7F709F7F6F9F7F709F706F9F7F709F7F6F9F7F6F9F7F6F80605F3F302F
0000000000002F20200000001010108F6F5F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F\
9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F60
9F7F60201F10302F20302F20000000000000000000907F609F7F60907F609F7F60907F609F7F60\
907F5F9F7F60907F5F9F7F60907F5F9F7F608F705F2F2F20000000000000000000
000000000000000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303050303F50000000303F50000000
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000000000000F0F0F2F2F30000000101010\
302F3FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000
4040405F605F5F5F5F60605F5F5F5F5F605F5F5F5F5F605F5F5F5F5F605FC0C0C0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005F5F5F000000000000000000000000000000000000000000000000\
0000000000000000006F50606F5F6F6F50606F5F6F3F303F201F1F0000000F0F0F
000000201F206F5F6F705F6F5040506F5F60705F6F7F6070705F6F7F60707F606F7F60707F606F\
7F60707F606F3F3030201F200000000000000000000000000F0F0F101010806F70
7F606F806F7080606F806F7080606F80606F6F505F0000000000000000000F00008F6F70806F6F\
8F6F70806F70705F600000000000000000000000000000000F0F0F604F508F7070
8F6F6F8F70708F6F6F90707080606F000000000000000000000000000000000000000000000000\
00000000000000000090706F90707090706F90707090706F90707090706F9F7070
5040400000000000000000000000000000000000000000000000000F0F0050403F9F7F6F90706F\
9F7F6F90706F9F7F6F90706F9F7F6F9F706F9F7F6F90706F40302F000000000000
0F0F0F4F3F302F201F6F504F9070609F7F6F9F7F609F7F6F9070609F7F6F9F7F609F7F6F907060\
9F7F6F9F7F609F7F6F907F609F7F6F9F7F609F7F6F907F609F7F60907F609F7F60
907F609F7F60907F609F7F602F201F302F202F201F9F7F6090705F907F60907F5F9F7F6090705F\
907F60907F5F907F6090705F907F60907F5F907F608F705F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0303F5010101F30305030304F000000
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF00000000000030303F20202F00000000000030303F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000000F100F
6060605F605F6060605F605F6060605F605F60605F5F5F5F60605FCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F101F3F303F6F50606F5F6F6F5F6F705F6F6F5F6F705F6F0F000F
5F4F506F5F6F705F6F6F5F6F60505F7F60707F60707F606F7F60707F60707F60707F6070806070\
705F6F60505F3F2F30000000000000000000000000806F70806070806F70806070
806F70806F70806F708060702F2020000000000000000000403F3F705F608F6F70806F708F6F70\
8F6F708F6F708F6F708F70700F0F0F0000000000000000000F0F0F6F50508F6F70
9070708F6F709070708F6F70907070100F0F000000000000000000000000000000000000000000\
0000000000000000009F707090707090707090706F9F7F709070709F707090706F
9F7F703F302F00000000000000000000000000000000000000000080605F9F7F709F706F9F7F70\
9F7F6F9F7F709F706F9F7F709F7F6F9F7F709F7F6F60504F100F0F000000201F10
9F7F6F403F308F6F609F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F609F7F6F\
9F7F6F9F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F60
9F7F6F9F7F609F7F60907F609F7F609F7F60302F20907F609F7F60907F609F7F60907F609F7F60\
907F609F7F60907F5F9F7F60907F5F907F60907F5F4F3F30000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F303F50303050303F50303050
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000302F30000000000000000000302F30CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF0000000000002F2F2F60605F
5F605F60605F5F5F5F5F605F5F5F5F60605F5F5F5F5F605F5F5F5FCFCFCFC0C0C0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000302F301F101F5040506F5F6F6F50606F5F6F6F50606F5F6F
604F5F705F6F6F5F606F5F60705F6F7F6070705F6F7F60707F606F7F60707F606F7F60707F606F\
705F6F00000000000000000060505F60505F705F60705F6F806F707F606F806F70
806070705F604F3F403F2F30201F1F0000000F0F0F2F20203F30308F6F70806F6F8F6F70806F70\
8F6F70806F6F8F6F708F6F708F6F704F3F3F302F2F2F2020806F6F8F6F6F8F7070
8F6F709070708F6F6F8F70708F6F702F2020000000000000000000000000000000000000000000\
00000000000000000090706F90707090706F90707090706F90707090706F907070
90706F403030000000000000000000000000000000000000000000201F1F90706F9F7F6F90706F\
9F7F7090706F9F7F6F9F706F9F7F6F9F706F2F201F000000000000604F409F7F6F
9070609F7F6F9F706F9F7F6F9F7F6F9F7F6F9070609F7F6F9F7F609F7F6F907F609F7F6F9F7F60\
9F7F6F907F609F7F6F9F7F609F7F6F907F609F7F6F9F7F609F7F6F9070606F5040
907F609F7F606F5F409F7F60907F609F7F60907F609F7F60907F609F7F60907F5F907F60907F5F\
9F7F60907F5F907F603F2F208F705F6F5F40907F5F907F5F60503F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF30304F000000000000303050
000000CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000302F3F000000202020302F30CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000006060605F605F
6060605F605F6060605F605F60605F5F605F60605F5F605F60605FCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000101010303030
6F6F6F4F504F000000000000000000000000000000000000000000000000000000000000000000\
0000000000001F101F0000000000000F0F0F3F303F1F1F1F5F4F5F705F6F6F5F6F
705F6F6F5F6F6F5F605F4F507F60707060707F60707F60707F60707F60707F60707F6070201F1F\
0F0F0F2F20203F303050404F5F4050806F707F6070806F707F6070806F70806070
806F70705F600000000000000000000000007F606F806F708F6F70806F708F6F70806F708F6F70\
806F708F70708F6F708F70708F6F70806F6F8F6F708F70708F6F709070708F6F70
9070708F6F709070708F70709070703F3030000000000000000000000000000000000000000000\
0000000000000000009070709070709F7F709070709F707090706F9F7F70907070
9F7F7090706F2F202000000000000000000000000000000050403F90706F9F7F709F7F6F9F7F70\
9F706F9F7F709F7F6F9F7F709070604F3F30000000100F0F5F403F6F504F9F7F6F
9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F\
9F7F609F7F6F9F7F6F9F7F6F9F7F609F806F50403F1F1F1040302F100F0F000000
201F1F0F00000000005F4F3F2F2020907F60806F506050409F7F60907F609F7F60907F609F7F60\
907F609F7F60907F5F000000000000705F4F8F705F907F60705F4F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303F5030304F3F3F50303050
0F0F0FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF505F4F0000000F0F0F302F30000000000000302F30CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C000000000000060605F5F605F60605F
5F5F5F60605F5F5F5F60605F5F5F5F5F605F5F5F5F5F605FC0C0C0CFCFCF000000000000000000\
0000000000000000000000000000000000000000000000004F504F6060606F6F6F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000201F201F1F1F40303F0F0F0F0000000000000000000F0F0F504050705F6F
6F5060705F6F6F50607F6070705F6F7F6070705F6F7F6070705F6F7F607050404F403F401F101F\
0000000F0F0F7F60707F606F7F60707F6070806F707F606F806F707F6070302F2F
1F10100000000000001010102F20205F4F50705F60806F70806070806F70806F708F6F70806F70\
8F6F70806F708F6F70806F708F6F708F6F708F70708F6F708F6F708F6F708F7070
8F6F6F8F70708F6F709070708F6F6F907070000000000000000000000000000000000000000000\
00000000000000000090706F90707090706F90707090706F90707090706F9F7070
90706F9F707090706F90706F8F6F6090706F8F6F6090706F90706F9F7F7090706F9F7F7090706F\
9F7F6F9F706F9F7F706F504F10100F0000004F3F309070605F40409F7F6F9F7F6F
9F706F9F7F6F9F7F6F9F7F6F9F706F9F7F6F9F7F6F9F7F6F9F7F609F7F6F9F7F609F7F6F9F7F60\
9F7F6F9F7F607F60509070606050402F201F0000001F1010000000000000000000
00000000000000000000000000000010100F000000000000302F20201F10000000705F4F5F4F3F\
907F60907F5F00000000000000201F0000003F302F90705F907F5F0F0F0F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF30304F303F5010101F1F1F20
30304F000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF0000002F2F300F0F0F000000000000302F30CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF00000000000000000060605F6060605F605F
6060605F605F60605F5F605F6060605F605F60605F5F5F5FCFCFCFCFCFCF000000000000000000\
0000000000000000000000000000000F0F0F606F606F6F6F606F606F6F6F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000604F5F6F506F6F5F6F6F5F6F6050605040501F101F0000006F5F606F5F6F
705F6F604F5F6F5F6F705F707F60707F60707F60707F60707F60707F60705F4F50000000000000\
50404F8060707F6070806F707F6070806F707F6070806F70806070806F702F202F
0000000000000000001010108F6F70806F708F6F70806F708F6F70806F708F6F70806F708F7070\
8F6F708F6F708F6F708F70708F6F708F70708F6F709070708F6F708F70708F6F70
9070708F6F709070708F6F709070708F70705F404F000000000000000000000000000000000000\
0000000000000000000000009070709F70709070709F7F709070709F7F70907070
9F7F709070709F7F7090706F9F7F709F70709F7F7090706F9F7F709F7F6F9F7F709F706F9F7F70\
9F7F6F9F7F708F70603F302F0000003F302F8F6F6090706F9F7F6F9F7F709F7F6F
9F7F709F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F\
50403F1F1010705F500F0F0F000000000000302F2000000050403F7F60501F1010
3F302F705F4F000000302F2050403F00000010100F00000000000000000000000000000050403F\
907F609F7F60000000001F10007F60000F0F000000907F60907F5F201F1F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000CFCFCFCFCFCFCFCFCF3F3F503030500000000F0F0F
3F3F50000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C05F604F0000000000000000000000002F2F30302F3FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF0000000000000000006060605F605F60605F
5F605F60605F5F5F5F60605F5F5F5F60605F5F5F5F5F605FCFCFCFCFCFCF000000000000000000\
0000000000000000000000003F403F6060606F6F606060606F6F6F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F000F0F0F0F6F506F6050606F50606F5F6F6F506F6F5F6F6050606F5F6F
6F5F6F7F6070705F6F7F6070705F6F7F6070705F6F7F60707060700000000000000000003F2F30\
5040507F606F7F60707F60708060707F60708060707F60703F2F300F000F000000
0000004F3F40604F507F606F806070806F70806070806F70806F708F6F70806F708F6F70806F70\
8F6F70806F708F6F708F6F708F6F708F6F708F6F708F6F708F70708F6F708F7070
8F6F709070708F6F709070708F6F709070706F5050000000000000000000000000000000000000\
00000000000000000000000090707090706F90707090706F9F707090706F907070
90706F9F7F7090706F9F707090706F9F7F7090706F9F7F7090706F9F7F7090706F9F7F709F706F\
9F7F7090706F3F302F3F2F2F0000003F2F2F8F6F609F7F6F9F7F6F9F706F9F7F6F
9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F609F7F6F5F4F40\
00000010100F0000000000001F10105040303F2F20705F4F9F7F6F7F604F9F7F60
907F6070604F8F705F9F7F604F3F308F705F907F604F3F30605040806F50201F10403F2F7F604F\
907F60000000000000004030007F6000100F0000008F705F907F608F705F000000
00000000000000000000000000000000000000000020201F0F0000000000000000000000000000\
000000000000000000000000CFCFCFC0C0C0CFCFCF30304F303F5000000F10101F
30305010101FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF000000302F3000000000000000000030303FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000005F605F60606060605F
6060605F605F6060605F605F60605F5F605F60605FCFCFCFCFCFCFCFCFCF000000000000000000\
0000000000000000004F4F4F606F606F6F6F606F606F6F6F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000F0F0F0000002020201010105040506F5F6F6F5F6F705F6F6F5F6F
605060705F707F60707060707F60707060707F60704030402F20200000000000000F0F0F7F6070\
7F6070806F707F6070806F707F6070806F707F6070806F7060505F000000000000
1F1F1F4F3F40806F70806F70806F70806F708F6F70806F708F6F70806F708F6F70806F708F6F70\
806F708F70708F6F708F70708F6F708F70708F6F708F70708F6F709070708F6F70
9070708F6F709070708F70709070708F70708F6F6F000000000000000000000000000000000000\
0000000000000000000000009070709F7F709070709F7F709070709F7F70907070
9F7F709070709F7F709070709F7F7090706F9F7F709F70709F7F709F706F9F7F709F7F6F9F7F70\
9F7F6F9F7F703F2F2F0000007F605F9F7F709F7F6F9F7F709F7F6F9F7F709F7F6F
9F7F709F7F6F9F7F709F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F806F201F1F000000\
000000000000000000705F509F7F609F806F9F7F6F9F7F6F9F7F609F806F9F7F60
9F7F6F9F7F609F806F9F7F609F7F60907F609F806F9F7F609F7F60907F609F8060907F607F6F50\
907F5F10100F000000007F60007F5F004F3F000000504030907F5F907F60000000
0000000000000000000000000000000000000000003F3020000000000000000000000000000000\
000000000000000000000000000000CFCFCFCFCFCF3F3F5030305020202F303F50
3F3F50303050000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF505F4F000000000000202020000000302F3FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF00000000000000000000000040403F60605F
5F5F5F60605F5F605F60605F5F5F5F60605F5F5F5FCFCFCFC0C0C0000000000000000000000000\
0000000000005F5F5F6060606F6F6F6060606F6F6F000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000004030400F0F0F2F202F0000000F0F0F1F1F1F6050606F5F6F6F5F6F6F5F6F
705F6F706070705F6F7F6070705F6F7F6070706070403F400000000000000000006F5F607F606F\
7F60707F60707F60707F60707F60707F6070705F6F101010000000000000806F70
7F6070806F70806070806F70806070806F70806F70806F70806070806F70806F708F6F70806F70\
8F6F70806F708F6F70806F708F6F708F6F708F70708F6F708F70708F6F708F7070
8F6F708F70708F6F709070708F6F709070708F6F70100F0F000000000000000000000000000000\
00000000000000000000000090707090707090707090706F9070709070709F7F70
90706F9F707090706F9F7F7090706F9F7F7090706F9F7F7090706F9F7F7090706F9F7F7090706F\
9F7F709F706F9F7F707F60509F7F6F9F7F6F9F7F709F706F9F7F6F9F7F6F9F7F6F
9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F907060000000000000000000\
100F0F4F3F308F70609F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F
907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F5F\
907F60302F20000000005F40007F60003F3000000010100F907F5F8F7F5F504030
0000000000000000000000000000000000000000001F1F103F3020000000000000000000000000\
000000000000000000000000000000000000CFCFCF30304F303F503030503F3F50
30304F303F50000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF0000002F202F000000000000302F302F2F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000
60605F60605F60605F5F605F6060605F605F60605FCFCFCFCFCFCF000000000000000000000000\
0000002F2F2F101010000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F303F605060302F30504050101010000000705F6F6F5F6F6F50606F5060
7F60707060707F60707060707F60707F60702F2020000000000000201F205040507F60707F6F70\
7F60707F6F707F6070806F707F6070806F70705F600000001010104F3F4060505F
806F70806F70806F70806F708F6F7F806F708F6F70806F708F6F7F806F708F6F70806F708F707F\
8F6F708F70708F6F708F707F8F6F708F70708F6F7090707F8F6F709070708F6F70
90707F8F70709070708F7070907070907070907070201F1F000000000000000000000000000000\
0000000000000000000000009070709F7F709070709F7F709070709F7F70907070
9F7F709070709F7F709070709F7F709F70709F7F709F706F9F7F709F7F709F7F709F7F6F9F7F70\
9F7F6F9F7F709F7F6F9F7F709F7F6F9F7F709F7F6FA07F709F7F6F9F7F709F7F6F
A07F709F7F6F9F7F6F9F7F6FA07F709F7F6F9F7F6F9F7F6F6F504F100F0F0000000000000F0F0F\
9070609F7F6F9F7F6F9F806F9F7F6F9F806F9F7F609F806F9F7F609F806F9F7F60
9F806F9F7F609F806F9F7F609F806F9F7F609F8060907F609F80609F7F609F7F60907F609F8060\
907F603F3020000000000000005F40000F0F0000001F1F10907F5F907F60907F5F
000000000000000000000000000000000000000000807050907F5F2F201F000000000000000000\
0000000000000000000000000000000000000000003F3F50303F503F3F5030304F
3F3F50303F50000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF0000000000002F2F300000002F2F30CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000000000000000000000
1F1F105F605F5F5F5F60605F5F5F5F60605F5F5F5FCFCFCFCFCFCF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000006050606F5F6F6F506F6F5F6F6050605040506F506F6F5F6F605060706070
705F70706070705F6F706070705F707060704F404F0000000F0F0F7F60707060707F60707F6070\
7F60707F60707F60707F60707F607050404F00000000000060505F7F6070806F70
806070806F70806070806F70806070806F70806070806F70806F708F6F70806F708F6F70806F70\
8F6F70806F708F6F708F6F708F70708F6F708F6F708F6F708F70708F6F708F7070
8F6F709070708F6F709070708F6F709070708F6F70907070000000000000000000000000000000\
0000000000000000000000009070709070709070709070709F7F7090706F9F7070
9070709F7F7090706F9F7F7090706F9F7F7090706F9F7F7090706F9F7F7090706F9F7F709F706F\
9F7F709F706F9F7F709F7F6F9F7F709F706F9F7F709F7F6F9F7F709F7F6F9F7F6F
9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F8F6F5F0000000000000000003F302F8F6F5F\
9F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F907F609F7F6F
9F7F609F7F6F907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F60\
9F7F607F6050000000000000000000000000000000907F5F907F60907F5F907F5F
0000000000000000000000000000000000000000006F5F408F705F8F7F5F000000000000000000\
0000000000000000000000000000000000000000000000003F3F5030304F20203F
3030503F3F50000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF5F60501F101F000000000000101010CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000
0000001F1F1F60606060605F60605F5F605F60605FCFCFCFCFCFCF000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000001010105F4F5F6F5F6F6F506F6F5F6F6F5F6F6F5F6F6F5060706070705F70
7F6070705F707F60707060707F6070302F300000000000006F5F6F7F60707F60707F60707F6070\
7F6070806F707F6070806F707F60700000000000000F0F0F604F5F806F7F806F70
806F70806070806F7F806F70806F70806F708F6F7F806F708F6F70806F708F6F7F806F708F6F70\
806F708F707F8F6F708F70708F6F708F707F8F6F708F70708F6F7090707F8F7070
9070708F6F7090707F8F70709070708F707090707F907070403030000000000000000000000000\
0000000000000000000000009070709F7F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F7F70\
9F7F6F9F7F709F7F6F9F7F709F7F6FA07F709F7F6F9F7F709F7F6FA07F709F7F6F
9F7F709F7F6FA080709F7F6F9F7F6F9F7F6F604F400F0F0F000000000000806F5F9F7F6F9F806F\
9F7F6FA0806F9F7F6F9F806F9F7F6FA0806F9F7F6F9F806F9F7F609F806F9F7F60
9F806F9F7F609F806F9F7F609F806F9F7F609F806F9F7F609F8060907F609F8060907F609F8060\
907F609F80600F0F0F000000000000000000000000907F60907F5F907F60907F5F
1F1010000000000000000000000000000000000000403F2F907F5F8F7F5F1F1F10000000000000\
0000000000000000000000000000000000000000000000000000003F3F50000000
3F3F50303050000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0202020000000302F30000000302F30C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000000000
0000000000005F605F60605F5F5F5F60605FC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000F0F100F0F0F403F40302F306F506F6F5F6F6F506F605060705F70706070
705F6F706070705F707F60705F4F50302F300000002F2020705F707F60707F60707F60707F6070\
7F60707F60707F60705F4F50302F300F00004F3F40705F6F806F707F6070806F70
7F6070806F70806070806F70806070806F70806F70806F70806F70806F70806F708F6F70806F70\
8F6F70806F708F6F70806F708F6F708F6F708F70708F6F708F70708F6F708F7070
8F6F708F70708F6F709070708F6F709070708F7070907070504040000000000000000000000000\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0606070
9F7F6F9F7F6F9F7F6F9F7F6F9F7F6F6F5F4F0000000F0F0F1F1010705F509F7F6F9F7F6F9F7F6F\
9F7F6F9F7F6F9F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F9F7F609F7F6F
907F609F7F609F7F609F7F6F907F609F7F60907F609F7F60907F609F7F60907F609F7F60907F60\
9F7F60907F608F705F000000000000000000806F50907F5F907F60907F5F907F60
0000000000000000000000000000000000000000002F201F8F70508F7F5F8F7050000000000000\
000000000000000000000000000000000000000000000000000000000000000F0F
3030503F3F50000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF0000000000000000002020202F2F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000
0000000000004F4F405F605F60606060605FCFCFCFCFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000201F206050605F4F5F705F70706070705F70
7F60707060707F60707060701F101F0000000000007060707F60707F60707F60707F60707F6F7F\
7F60707F6F707F6070806F7F000000000000100F10806F7F806070806F70806070
806F7F806F70806F70806F708F6F7F806F708F6F7F806F708F6F7F806F708F6F7F806F708F707F\
8F6F708F707F8F6F708F707F8F6F708F707F8F6F7090707F8F6F7090707F8F6F70
90707F8F70709070708F707090707F907070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60607060606F
60607060606F0000000000000000000000000F000010100FA080709F7F6F9F806F9F7F6FA0806F\
9F7F6F9F806F9F7F6FA0806F9F7F6F9F806F9F7F6F9F806F9F7F6F9F806F9F7F60
9F806F9F7F609F806F9F7F609F806F9F7F609F806F9F7F609F806F9F7F609F8060907F609F8060\
907F609F8060907F603F3020000000100F0F7F6F509F8060907F5F907F60907F5F
0000000000000000000000000000000000000000008F7F5F907F5F8F7F5F907F5F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F3F503F3F50000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
2F2F2F0000002F2F300000002F2F30CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF00000000000000000060605F1F1F1F000000000000
000000000000303F3060605F5F605F60605FC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000001F101F504050706070705F6F706070
705F70706070705F706F5F6F100F10100F103F2F307F60707060707F60707060707F60707F6070\
7F60707F607040303F100F10000000202020806F707F6070806F707F6070806F70
806070806F70806070806F70806F70806F70806070806F70806F708F6F70806F708F6F70806F70\
8F6F70806F708F6F708F6F708F70708F6F708F70708F6F708F70708F6F708F7070
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF60606F606070
60606F6060705F5F6F0000000000000000000000000000000000000000000000009F7F6F9F7F6F\
9F7F6F9F7F609F7F6F9F7F6F9F806F9F7F609F7F6F9F7F609F806F9F7F609F7F6F
9F7F609F806F907F609F7F60907F609F8060907F609F7F60907F609F8060907F609F7F60907F60\
9F7F60907F5F9F7F60907F5F504F3F807050907F60907F5F907F60907F5F907F5F
0000000000000000000000000000000000000000008F7F5F8F7F5F8F7F5F8F70506F5F40000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F3F50000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
0000000000000000001F1F1F302F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF60606060605F606F6060605F60606060605F60606060605F
4F4F4F60605F3F403F60605F60605F5F605FCFCFCFCFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000403F40705F70705F70706070705F70
706070705F707F60700000000000000000007F60707F60707F60707F60707F607F7F60707F6070\
7F6070806F7F705F6F0000000000002F202F7F6070806F7F7F6070806F7F806F70
806F7F806F70806F7F806F70806F7F806F708F6F7F806F708F6F7F806F708F707F806F708F6F7F\
806F708F707F8F6F708F707F8F6F708F707FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000302F2FAF8F803F302F
0000000000006F60703F3F40000000000000000000000000000000000000000000000000000000\
0000000000000000000000009F7F6FA0806F9F7F6F9F806F9F7F609F806F9F7F6F
9F806F9F7F609F806F9F7F609F806F9F7F609F806F9F7F609F8060907F609F806F9F7F609F8060\
907F609F8060907F609F8060907F609F8060907F60908060907F5F90806070604F
00000000000000000000000000000000000000000060503F907F5F8F7F5F907F5F6F5F40000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000
0000002F2F30000000302F30C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF60605F60606060605F6060605F605F60605F60605F60605F
5F605F60605F5F605F60605F5F5F5F60605FC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000001010106F5F6F706070705F70706070
705F70705F704F3F4F201F201F1F1F5040507060707F60707060707F60707F60707F60707F6070\
7F60707F607050404F00000040303F705F6F806F707F6070806F707F6070806F70
7F6070806F70806070806F70806070806F70806F70806F70806F70806F70806F708F6F70806F70\
8F6F70806F708F707F0000007070706F6F70CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF907F809F808F9F7F80000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000000000000000090706F403030
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000CFCFCF9F7F609F7F6F9F7F609F806F
9F7F609F7F6F9F7F609F806F907F609F7F60907F609F8060907F609F7F60907F609F8060907F60\
9F7F60907F609F8060907F5F907F60907F5F907F60907F5F907F60907F5F8F705F
000000000000000000000000000000000000000000907F5F8F705F8F7F5F8F7F5F8F7F5F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000202030000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F305FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
0000000000001010102F2F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF606F6060605F60606060605F606F6060605F60606060605F
60606060605F60605F60605F606060CFCFCFCFCFCF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F1F1F6F5F70706070705F70706070705F70
7F60707060700000000000000000007060707F60707060707F607F7F60707F60707F60707F6F7F\
7F60707F6F7F00000000000040303F806F7F7F6070806F7F806070806F7F806070
806F7F806F70806F7F806F70806F7F806F708F6F7F806F708F6F7F806F708F6F7F806F708F707F\
0000000F0F0F6F6F7070707F6F7070707070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF9F808F907F809F808F9F80809F808F9F80809F808F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000A0807FAF8F7F
8F706F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000CFCFCFCFCFCFCFCFCF5F5F6F9F806F9F7F60
9F806F9F7F609F806F9F7F609F806F9F7F609F806F9F7F609F806F9F7F609F8060907F609F8060\
9F7F609F8060907F609F8060907F609F8060907F5F9F8060907F60908060907F5F
0000000000000000000000000000000000000000007F6F50907F5F8F7F5F907F5F8F7F5F0F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303050CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000
201F200000002F2F30CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF60605F60606060605F60605F60605F6060605F605F60605F
60605F60605F5F5F5F60605F5F605FCFCFCFC0C0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F101F705F70705F70706070705F70706070
705F70302F300000000F0F0F302F307F6070705F707F60707060707F60707060707F60707F6070\
7F607060505F1F1F1F3F303F7F60707F6070806F707F6070806F7F7F6070806F70
7F6070806F7F7F6070806F70806F70806F7F806F70806F70806F708F6F7F0000000000002F2F2F\
0000000000000000000000000000006F6F70CFCFCFCFCFCFCFCFCF000000000000
907F809F808F907F809F808F907F809F808F907F809F808F9F8080000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F50507F605F201F1F000000000000000000\
00000000000000000000000000000000000000000000000000000010100F7F6060
AF8F7F7F605F1F1010000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000C0C0C0CFCFCFCFCFCF5F5F6F5F50605F5F6F
5F5F609F806F907F609F806F9F7F609F806F907F609F8060907F609F8060907F609F7F60907F60\
9F8060907F609F7F60907F609F8060907F5F907F60907F5F907F60907F5F10100F
0000000000000000000000000000000000000000004F40308F7F5F8F7F5F8F70508F7F5F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
0000000F0F0F30303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF606F6060605F606F6060605F60606060605F60606060605F
60606060605F60606060605F60605FCFCFCFCFCFCF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F4F5F705F70706070705F70706070705F70
706070000000000000705F707F60707060707F607F7F60707F60707F60707F607F7F60707F607F\
7F60700F0F0F0000000F0F0F7F6070806F7F7F6070806F7F7F6070806F7F806F70
806F7F806F70806F7F806F70806F7F806F708F6F7FCFCFCFCFCFCFCFCFCF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
9F808F907F809F808F9080809F808F9080809F808F9F80809F808F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006F5050AF8F80AF8080000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
AF8F807F605F201F1F000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFCFCF5F5F6F605F6F5F5F6F
5F5F6F5F5F6F5F5F6F5F5F6F9F806F9F7F609F806F9F7F609F806F9F7F609F806F9F7F609F8060\
907F609F8060907F609F8060907F609F8060907F60908060907F5F9080602F201F
0000000000000000000000000000000000000000008F7F5F907F5F8F7F5F907F5F8F7F5F0F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF0000002F2F30
0000000000002F2F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF60605F60606060605F60606060605F60605F60605F60605F
5F605F60605F60605F60605F5F5F5FCFCFCFC0C0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000006F5F707060706F5F70706070705F70706070
6F506F302F301F101F4F404F705F707060707060707F60707060707F60707060707F6070706070\
706070302F300000006F5F607F60707F6070806F7F7F6070806F707F6070806F7F
7F6070806F70806070806F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000\
707070000000000000000000000000000000000000000000000000000000000000
504F4F90808F907F809F808F907F809F808F907F809F808F907F80000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000F0F0FAF8080AF8F80100F0F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
907F70AF8F806F504F101010000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
5F50605F5F6F5F5F6F5F5F6F5F50605F5F6F9F7F609F806F907F609F8060907F609F8060907F60\
9F8060907F609F8060907F5F907F60907F5F908060907F5F907F60907F5F403F2F
0000000000000000000000000000000000000000006F5F408F7F5F8F7F5F8F7F5F8F7F5F605F40\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000
0F0F0F2F2F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF60605F6F6F6060605F60606060605F606F6060605F60606060605F
60606060605F60605F60605F606060CFCFCFCFCFCF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000005F5060706070705F70706070705F70706070705F70
100F100000005040507060707F607F7060707F607F7060707F607F7F60707F607F7F60707F6F7F\
50404F000000000000806F7F7F6070806F7F7F6070806F7F806F70806F7F806070
806F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
70606F9080809F808F907F809F808F9F80809F808F9F80809F808F9F8080000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003F2F2F9F7F70AF80805F4F40000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F60504FAF8F80A0807F5F4F40000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005F5F6F605F6F5F5F6F5F5F6F5F5F6F605F6F9F7F609F806F907F609F806F\
9F7F609F8060907F609F8060907F609F8060907F609F8060907F609080605F4F3F
000000000000000000000000000000000000000000605F40907F5F8F7F5F907F5F8F7F5F605F40\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF0000002F2F30000000
201F20302F30C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C02F2F2F0000001F1F1F0000000000000000000000002F2F2F60605F
60605F60605F5F605F60605F60605FCFCFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000201F206F5F70705F706F5F707060706F5F70706070
4F3F4F000000100F10706070705F707F6070705F707F60707060707F60707060707F60707F6070\
5040500000006F50607F60707F6F7F7F60707F6F707F6070806F7F6F6F6F707070
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
80707F9F808F907F809F808F907F809F808F907F809F808F9F80809F808F000000000000000000\
0000000000000000000000000000000000000000000F0F0F0F0F0F0F0F0F0F0F0F
0F00000000000000000000000000000000000000000000000000000F0F0FA0807FAF8F80000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000806F60A08F7F0F0F00000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005F5F605F5F6F5F5F6F5F5F6F5F5F6F5F5F6F907F60\
9F8060907F609F8060907F609F8060907F5F907F60907F5F908060907F5F000000
000000000000000000000000000000000000000000302F208F7F5F907F5F8F7F508F7F5F807050\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000
0F0F0FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF0000000000000000000000000000000000000000000000000F0F0F
60606060605F60606060605F60605F000000000000000000000000000000000000000000000000\
0000000000001F101F000000000000201F20000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000403F40706070705F70706070705F7070607F605060
201F204030407060707060707F607F7060707F607F7060707F607F7060707F6F7F7F60706F5F6F\
201F200000007F6070806F7F7F6070806F7F7F6070806F7F706F70707070707070
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
8F7F7F907F809F808F90808F9F808F9080809F808F9F808F9F808F9F8080000000000000000000\
0000000000000000000000000000000F0F0F7F6060A08F8FA08080AF8F8FA08080
A080800000000000000000000000000000000000000000000000000F0000AF8F80AF8F804F403F\
000000000000000000000000000000000000000000000000000000000000000000
0000000000008F706FAF8F809F807F50403F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000605F6F5F5F6F60606F5F5F6F605F6F\
5F5F6F9F806F907F609F8060907F609F8060907F609F8060907F5F908060000000
0000000000000000000000000000000000000000008F7F5F907F5F8F7F5F907F5F8F7F5F8F7F5F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF0000002F2F30000000000000
2F2F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000
60605F60605F60605F60605F5F5F5F000000000000000000000000000000000000000000000000\
0000000000001F1F1F504F602F202F50405F6050705F506F3F30404F40500F0F0F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000705F706F5F707060706F5F70706070705F70706070
0000000F0F0F705F70706070705F707060707060707F60707060707F60707060707F607F706070\
0000001F10105F4F5F7F60707F60707F60707070706F6F6F7070706F6F6F707070
C0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F90808F907F809F808F907F809F808F9080809F808F907F809F808F000000000000000000\
0000000000000000000000000000000000007F606FA08080A08080A08080A08F8F
A08080806F6F1F10100000000000000000000000000000000000000000001F1010806F60AF8080\
907F70000000000000000000000000000000000000000000000000000000000000
000000000000000000403F30AF8F7FAF8F80705F50302F20000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005F5F6F5F5F6F5F5F6F\
5F5F6F5F5F6F000000907F60908060907F60908060907F5F907F60907F5F000000
000000000000000000000000000000000000000000907F5F8F7F5F8F7F5F8F7F5F8F7F5F403F2F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF000000\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000F0F0F2F2F30
30303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF00000000000000000000000000000000000050504F50504F000000
0F0F0F60605F60605F60605F000000000000000000000000000000000000000000000000000000\
0000000000000000002F203060506F605F7060506F605F70605070605F704F4050
0F0F0F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000201F206050607060706F5F7070607F705F70706070302F30
0F0F0F3F2F3F70607F7060707F607F7060707F607F7060707F607F7F60707F607F7F60704F404F\
0F0F0F5040507F6070806F7F707070707070707070707070707070707070707070
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
2F202090808F9F808F9080809F808F9F808F9F808F9F80809F808F9F808F000000000000000000\
000000000000000000000000000000605050A08080A08F8FA08080A08F8FA08080
AF8F8F806F6FA080804F3F3F000000000000000000000000000000000000000000000000A0807F\
AF8F80000000000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F5F4F40AF8F807F605F2F2020000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000060606F\
5F5F6F605F6F00000F000000907F609F8060907F609F8060907F60908060000000
0000000000000000000000000000000000000000008F7F5F9080608F7F5F907F5F8F7F5F807050\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002F201F000000605F40303020000000000000CFCFCFCFCFCFCFCFCFCFCFCF000000\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF2F2F300000002F2F30000000
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C00000000000000000000000001F1F1F60605F60606060605F0F0F0F
00000050504F60605F000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000005F506F60507060506F605F7060506F605F70
60506F605F704F405000000F000000000000000000000000000000000000000000000000000000\
000000000000000000000000705F706F5F70705F706F5F707060706F5F70100F10
000000706070705F70706070705F707F607F705F707F60707060707F607F7060707F6070504050\
0000004F3F4F7070706F6F707070706F6F6F7070706F6F707070706F6F6F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000090808F907F8090808F907F809F808F907F809F808F9F80809F808F000000000000000000\
000000000000000000000000000000403030A0808FA08080A08F8FA08080A08F8F
A08080AF8F8F9F807F4F3F3F0000000000000000000000000000000000000000000000009F7F70\
AF8F80806F60403030000000000000000000000000000000000000000000000000
0000000000000000000F0F0F50403FAF8F80A08F7F907F704F403F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000908060907F5F908060907F5F000000
000000000000000000000000000000000000000000907F5F8F7F5F907F5F8F7F5F8F7F5F2F201F\
000000000000000000000000000000000000000000000000000000000000000000
403F2F000000605F40201F103F3020000000706F4F3F3F2F000000CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000F0F0F2F2F30
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF0000000000000000000000005F5F5F6060606000006F605F1F1F1F
0000002F2F20000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000004F405060506F605F7060506F605F70605070
6F5F706050706F5F704F40504F4050000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F6F5F7070607F705F70706070705F7070607F504050
302F306050607F607F7060707F607F7060707F607F7060707F607F7060707F6F7F7F60705F4F50\
707070707070707070707070707070707070707070707070000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F303F9F808F90808F9F808F90808F9F808F9F808F9F808F9F808F9F808F000000000000\
000000000000000000000000000000403F3FA08080A08F8FA08080AF8F8FA08080
AF8F8FA08080AF8F8FAF8F801F1010000000000000000000000000000000000000000000100F0F\
60504FAF8F8FAF8F8060504F1F1010000000000000000000000000000000000000
0000000000000000000000000000002F201F6F5F50AF8F8050403F000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000CFCFCFCFCFCF907F609F8060000000
0000000000000000000000000000000000000000008F7F5F90805F8F7F5F90805F8F7F5F70604F\
000000000000000000000000000000000000000000000000000000000000000000
7F70500F0F0F0F0F0F4F40301F1F10605F3F4F4F302F2F1F000000CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000000000000000002F2F30CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C000000000000000000000000060606060605F600000605F50000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000302F3F5F506F60507060506F605F70
60506F605F7060506F6F5F7060506F100F10000000000000000000000000000000000000000000\
0000000000000000005F4F606F5F706F5F707060706F5F707060706F506F000000
5F4F5F706070705F7070607F705F707060707060707F607F7060707F60707060707070706F6F6F\
7070706F6F707070706F6F6F707070706F70000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000504F50907F809F808F907F809F808F9080809F808F9080809F808F907F80000000000000\
0000000000000000000000002020209F8080A08F8F9F8080A08F8FA08080A08F8F
A08080A08F8FA08080AF8F8F9F7F7F403F30000000000000000000000000000000000000000000\
00000060504FA0807F5F4F401F1010000000000000000000000000000000000000
000000000000000000000000000000000000000000AF9080A0807F6F5F50000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000403F309F8F70000000CFCFCFC0C0C0CFCFCF504F30000000
0000000000000000000000000000000000008F7F5F907F5F8F7F5F8F7F5F8F7F5F8F7F5F504030\
000000000000000000000000000000000000000000000000000000000000000000
605F3F706F4F0000007F704F5F5F3F706F4F706F4F706F4F000000000000C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F2F300F0F0F2F2F30302F30CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF00000000000000000000000060605F6F6F60602F2F50504F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000605F70605070605F70605070
6F5F706050706F5F706050706F5F70605F703F303F000000000000000000000000000000000000\
0000000000000000001F1F206F5F707060706F5F7070607F705F706F5F70000000
20202070607070607F7060707F607F7060707F607F7060707F607F000000707070707070707070\
707070707070707070707070000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000006F5F609F808F90808F9F808F90808F9F808F9F808F9F808F9F808F9F808F000000000000\
000000000000000000000000000000A08F8FA08080A08F8FA08080A08F8FA08080
AF8F8FA08F80AF8F8FA08F80A08080AF8F801F1F1F000000000000000000000000000000000000\
0000006F5F50A0807FAF908FAF8F803F302F000000000000000000000000000000
0000000000000000000000000000000000000F0F0F3F302FAF9080AF8F80000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000009F907F9F8F709F907F9F8F70CFCFCFCFCFCFCFCFCFCFCFCF
0000000000000000000000000000000000009080608F7F5F9080608F7F5F907F5F8F7F5F8F7F5F\
000000000000000000000000000000000000000000000000000000000000000000
504F3F7F6F4F0000006F60400F0F0F706F4F605F40706F4F000000000000CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C00000002020200F0F10C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C00000000000000000000000001F1F1F40403F60605F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000403F4F60506F605070
60506F605F7060506F605F706050706F5F7060506F000000000000000000000000000000000000\
0000000000000000003F303F6F5F706F5F70705F706F5F707060706F5F702F202F
705F7070607F705F70706070705F7070607F705F70000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000080707F907F8090808F907F809F808F907F809F808F9080809F808F60505F000000000000\
0000000000000000000000001010109F8080A0808FA08080A08F8FA08080A08F8F
A08080AF8F8FA08080AF8F8FA08080AF8F8F1F1010000000000000000000000000000000000000\
0000000000000F0F0F7F605FAF8F80AF8F809F7F7F10100F000000000000000000
000000000000000000000000000000000000000000000000100F0FAF90808F706F201F1F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000706F5F9F8F709F8F70908F709F8F70908F709F8F70C0C0C0CFCFCF
CFCFCF0000000000000000000000000000008F7F5F907F5F8F7F5F907F5F8F7F5F8F7F5F8F7F5F\
000000000000000000000000000000000000000000000000000000000000000000
0000007F705010100F7F6F4F000000706F4F000000706F4F000000000000000000CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000F0F0F2F2F30CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000605F70605070
605F706050706F5F70605F706F5F70605F706F5F706F5F7020202F000000000000000000000000\
0000000000000000000F0F0F6F5F7070607F6F5F7070607F6F5F7070607F705F70
5F5060705F7070607F70607070607F706070CFCFCF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000009F808F90808F9F808F90808F9F808F9F808F9F808F9F808F9F8F8F000000000000\
000000000000000000000000100F0FA08F8FA0808FA08F8FA08080A08F8FA08F8F
AF8F8FA08080AF8F8FA08F80AF8F8FAF8F80AF8F8F705F5F2F2020000000000000000000000000\
00000000000000000000000010100F907F709F7F701F1010000000000000000000
0000000000000000000000000000000000000000000000001010108F706FAF9080A08F800F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000706F5F9F907F9F8F709F907F9F8F709F9070908F709F907F908F70
CFCFCFCFCFCFCFCFCF0000000000000000009080608F7F5F90805F8F7F5F90805F8F7F5F8F7F5F\
000000000000000000000000000000000000000000000000000000000000000000
0000007F6F4F000000706F4F1F1010706F4F000000706F4F2F201F000000000000000000CFCFCF\
30305FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000302F30C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C000000000000000000000000000000000000060605F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F1F205F4F60
60506F605F70605070605F7060506F605F706050706F5F70605070403040000000000000000000\
000000000000605F6F60506F6F5F706F5F707060706F5F707060706F5F7070607F
6F5F70706070705F7070607FC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000060505F403F3F0F0F0F000000000000000000000000
0000000F0F0F907F8F9F808F907F809F808F90808F9F808F9080809F808F80707F000000000000\
0000000000000000000000004F40409F8080A08F8F9F8080A08F8FA08080A08F8F
A08080A08F8FA08080AF8F8FA08080AF8F8FA08F80AF8F8FAF80803F2F2F000000000000000000\
0000000000000000000000001010108F706FAF8F80A08F7F7F605F101010000000
000000000000000000000000000000000000000000000000000000000000201F1FAF9080807060\
6F5F502F201F0F0F0F000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000008F7F6F9F8F7F908F709F8F70908F709F8F70908F709F8F70908F709F8F70
000000CFCFCFC0C0C0CFCFCFC0C0C00000008F7F5F90805F8F7F5F8F7F5F8F7F5F8F7F5F8F7F50\
000000000000000000000000000000000000000000000000000000000000000000
00000070604F5F503F7F704F000000706F4F40402F706F4F5F503F000000000000000000C0C0C0\
000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000002F2F30303030CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF00000000000000000000000000000000000060605F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
3F303F605070605F706050706F5F70605F706F5F70605F706F5F706F5F706F5F702F2030000000\
0000006F5F7060506F70607F6F5F7070607F6F5F7070607F705F7070607F705F70
70607F706070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000001010105F505F60505F80707F2F2F2F000000000000000000
000000201F209F808F90808F9F808F90808F9F808F90808F9F808F9F808F8F707F000000000000\
0000000000000000000000008F707FA08F8FA0808FA08F8FA0808FA08F8FA08080
AF8F8FA08F8FAF8F8FA08F80AF8F8FAF8F80AF8F8FAF8F80AF8F8FAF8F80000000000000000000\
000000000000000000000000000000000000201F1FA0807FAF8F80AF90807F605F
5F504F00000000000000000000000000000000000000000000000000000060504F9F8070AF9080\
A08F7FAF9080000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000009F907F9F8F709F907F9F8F709F907F9F8F709F907F9F8F709F9070000000
000000000000000000CFCFCFCFCFCFCFCFCF9080608F7F5F9080608F7F5F90805F8F7F5F8F805F\
8070500000000000000000000000000000000F0F002F2F1F000000000000000000
0000007F6F4F7F7050706F4F706F4F706F4F6F60406F6F4F706F4F000000000000000000000000\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF0000002F2F302F2F30CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000605F7060506F605F706050706F5F706050706F5F706050706F5F706050706F5F701F101F\
605F6F60506F6F5F706F5F706F5F706F5F706F5F706F5F7070607F6F5F70706070
6F5F70CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000080707F90808F8F7F8F90808F2F20201F1F1F000000000000
000000000000907F8090808F90808F9F808F907F809F808F90808F9F808F4F3F40000000000000\
0000000000000000000000000000009F8080A08F8FA08080A08F8FA08080A08F8F
A08080A08F8FA08080AF8F8FA08F80AF8F8FA08080AF8F8FAF8F80AF8F8F4030300F0F0F000000\
000000000000000000000000000000000000000000000000AF8F80AF8F807F6F60
5F4F401F10100000000000000000000000000000000000000000000000000000000000006F5F50\
AF9080A08F7F907F6F6F5F50302F2F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
2F2020908F709F8F7F908F709F8F7F908F709F8F70908F709F8F70908F70000000
000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCF8F7F5F8F7F5F8F7F5F8F7F5F8F7F5F\
8F7F5F7F70500000000000000F0F004F4030706F4F807050807F507F70505F503F
4F4030605F40706F4F706F4F706F4F706F4F5050306F60406F6F4F000000000000000000000000\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000002F2F30302F30CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000202020000000000000000000
00000000000000000000000000000020202F0F0F0F1F1F202F2F300000000F0F0F0F0F0F000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000020202F605F706F5F706050706F5F70605F706F5F70605F706F5F7F6F5F706F5F70\
6F5F706F607F6F5F706F607F6F5F7070607F6F5F7070607F6F5F7070607F705F70
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F6070000000201F20000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000002F202F8F7F8F90808F907F8F90808F101010000000000000
0000000000003F303F90808F9F808F90808F9F808F9F808F9F808F9F808F9F808F000000000000\
000000000000000000000000302F2FA08F8FA0808FA08F8FA0808FA08F8FA08F8F
AF8F8FA08F80AF8F8FA08F8FAF8F8FA08F80AF8F8FAF8F80AF8F8FAF8F80AF8F8F8F706F100F0F\
0000000000000000000000000000000000000000000F0F0F7F605FAF9080AF8F80
AF9080AF8F80A08F7F5F4F403F302F00000000000000000000000000000000000030302F706050\
9F8F7FAF9080A08F7FA09080302F20000000000000000000000000000000000000
0000000000000000000000004F403F6F5F50000000000000000000000000000000000000000000\
3F3F309F907F9F8F709F907F9F8F709F907F9F8F709F907F908F70000000000000
000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF8F7F5F90805F8F7F5F8F7F5F\
8F7F5F8F7F5F8070504F403070604F8F7F5F807F50807F5F807F50807F5F807F50
807F507F7050807F50706F4F6F6040706F4F70704F706F4F706F4F000000000000000000000000\
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF00000F302F302F2F30CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000
00000000000000000000000000000020202F504F6F504F6F504F6F5F506F5F4F6F4F405F101010\
20202F000000000000000000000000000000000000000000000000000000000000
000000000000000000605F70605070605F706050706F5F706050706F5F70605F706F5F70605F70\
6F5F706F5F706F5F706F5F706F5F706F5F7070607F6F5F70705F70C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF20202F70607F7F6F80605F705F50601F1F1F4F4050\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000008070808F7F8F8F7F8F90808F8F7F8F90808F000000000000
00000000000050404F90808F907F8F90808F90808F9F808F9080809F808F908080000000000000\
0000000000000000002020209F80809F8080A08F8F9F8080A08F8FA08080A08F8F
A08080A08F8FA08080AF8F8FA08080AF8F8FA08F80AF8F8FAF8F80AF8F8FAF8F801F10100F0F0F\
0000000000000000000000000000000000000000000000000000004F3F3F7F6060
A08F7FAF8F80AF8F805F4F4F101010000000000000000000000000000000000000000000000000\
0000002F2020A09080A08F7FA09080908070706F5F000000000000000000000000
0000000000000000009080709F8F7F6F60500F0F0F000000000000000000000000000000000000\
9F8F7F9F8F709F907F908F709F8F70908F709F8F70908F709F8F70000000000000
000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF8F7F5F8F7F5F8F7F50\
8F7F5F807F508F7F5F807F508F7F5F807F50807F5F807050807F50807050807F50
7F7050807F507F70507F70507F704F605F40606040706F4F6F6F4050503F000000000000000000\
000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF302F302F2F30303030CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000
0000000000000000000F0F0F5F506F5F4F6F5F50705F4F6F5F50705F506F5F50705F506F5F5070\
5F4F6F605070403F4F000000000000000000000000000000000000000000000000
0000000000000000000000006F5F70605F706F5F70605F706F5F7F6F5F706F5F706F5F706F5F7F\
6F5F706F5F7F6F5F7070607F6F5F7070607F6F5F7070607FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000302F3F7F70807F708F80708F7F708F504F5F\
7F708040404F000000101010000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000060505F8F7F8F90808F8F7F8F907F80907F8F000000000000
0000000000006F5F6090808F9F808F90808F9F808F90808F9F808F9F808F706060000000000000\
0000000000000000002F202F9F808FA08F8FA0808FA08F8FA08F8FA08F8FA0808F
AF8F8FA08F8FAF8F8FA08F80AF8F8FAF8F8FAF8F8FAF8F80AF8F8FAF8F80AF8F8F1F1010100F0F\
1010100000000000000000000000000000000000000000000000000000007F605F
AF9080AF8F80AF9080AF8F80AF90809F807F5F4F4F302F20000000000000000000000000000000\
000000302F205F504F7F6F60A08F7FA09080504F403F3030000000000000000000
0000000000009F8F7F9F8F7FA0907F9F8F7F908F70000000000000000000000000000000706F5F\
9F8F7F9F907F9F8F709F907F9F8F709F907F908F709F907F908F70000000000000
000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF8F7F5F8F805F\
8F7F5F8F7F5F807F508F7F5F807F508F7F5F807F508F7F5F807F50807F50807050
807F507F7050807F507F70507F7F507F70507F7050605F3F706F4F1F1F10000000000000000000\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF0000002F2F2F2F2F30000000CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000
0000000000001F1F2050506F504F6F5F506F504F6F5F506F504F6F5F5070504F6F5F506F5F506F\
5F50705F506F5F50705F506F100F10000000000000000000000000000000000000
000000000000000000000000000000605F70605070605F706050706F5F706050706F5F70605F70\
6F5F706F5F706F5F706F5F706F5F7F6F5F706F5F70CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C00000000000000000000000000000000F0F105F50602F202F7F708F7F7080\
80708F7F7080706F7F7F70805F5060201F200000000F0F0F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000080707F907F808F202090202F8F708090808F8F7F80000000
00000000000000000090808F90808F9F808F90808F9F808F90808F9F808F302F2F000000000000\
000000000000000000202020A08F8F9F8080A08F8F9F8080A08F8FA08080A08F8F
A08080A08F8FA08080AF8F8FA08F80AF8F8FA08F80AF8F8FAF8F80AF8F8F604F4F0F0F0F0F0F0F\
0F0F0F0F0F00100F0F000000000000000000000000000000000000000000000000
000000201F1F4F403F6F5F509F807FAF9080A08F7F302F2F1F1F1F1F1010000000000000000000\
000000000000000000000000A09080A08F7FA0907FA08F7F9F8F7F000000000000
0000000000009F8F7F9F907F9F8F7F9F907F2F20200000000000000000000000000000006F6050\
9F907F908F709F8F7F908F709F907F908F709F8F70908F70000000000000000000
000000000000000000000000000000000000000000000000000000CFCFCFC0C0C0CFCFCF8F7F5F\
8F7F5F807F508F7F5F807F508F7F5F807050807F50807050807F507F7050807F50
7F7050807F507F704F7F70507F704F7F705070704F7F704F70704F4F4F30000000000000000000\
000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF0000003030302F2F30000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000
000000302F402F2F3F504F6F5F5070504F6F5F50705F506F5F5070403F505F5070504F6F5F506F\
5F506F6050705F50706050705F506F302F3F000000000000000000000000000000
0000000000000000000000000000006050706F5F7F605F706F5F70605F706F5F7F6F5F706F5F7F\
6F5F706F607F6F5F706F5F7F6F5F7070607FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF0000000000000000000000000000000000000000000000000000000000000F0F0F4F4050\
605F6F807F8F706F7F807F8F80708F807F8F70607F403F40504F5F202020000000
201F20000000000000201F200F0F0F0000000000002F202F000000000000302F30101010000000\
4F4050201F200000008070808F7F8F9000008F000090404F907F8F90808F000000
00000000000000000090808F9F808F90808F9F808F9F808F9F808F9F808F9F8F8F000000000000\
0000000000000000009F808F9F808FA08F8FA0808FA08F8FA0808FA08F8FA08F8F
A08F8FA08F8FAF8F8FA08F8FAF8F8FA08F80AF8F8FAF8F8FAF8F8F403030100F0F0F0F0F302F2F\
0F0F0F100F0F100F0F000000000000000000000000000000000000000000000000
000000000000000000000000000000AF8F80AF9080AF9080AF9080A08F80000000000000000000\
0000000000000000000000000000003F3030504F3F80706F7F6F5F60504F201F1F
0F0F00908070A090809F907FA0907F9F8F7F9F8F7F0000000000000000000000000000009F907F\
9F8F709F907F9F8F709F907F908F709F907F9F8F709F907F000000000000000000
000000000000000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCF\
8F7F5F8F805F8F7F5F8F7F5F807F508F7F5F807F50807F5F807F50807F5F807F50
807F507F7050807F507F70507F7F507F704F7F705070704F7F70506060403F3F2F000000000000\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C00000002F2F30302F30000000CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000
0000003F3F4F504F6F5F506F504F6F5F506F504F6F5F50704F4F60403F50000000000000101010\
302F3F5F506F5F50705F506F5F50705F506F000000000000000000000000000000
0000000000000000000000000000000000006050706F5F706050706F5F70605F706F5F70605F70\
6F5F706F5F706F5F7F6F5F70CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
1F1F1F0000007F6F803F304070607F4F40507F70807F708F80708F80708F706070
1F101F80708F6F60703F3F4080708F706F7F5F505F2F202F807F8F7F70804F404F8F7F8F807080\
504F50302F300000000000008F7F8F8F3F409000008F000090808F7F6F70000000
0000000000002F202F90808F907F8F90808F90808F9F808F90808F9F808F8F7F80000000000000\
0000000000002F20209F808FA08F8F9F808FA08F8F9F8080A08F8FA0808FA08F8F
A08080A08F8FA08F80AF8F8FA08080AF8F8FA08F80AF8F8F9070700F0F0F0F0F0F0F0F0F2F2020\
0F0F0F0F0F0F0F0F0F0F0F0F10100F000000000000000000000000000000000000
0000000000000000000000000000002F20205F4F4070605FA08F7FAF9080A08F7F9080707F6F60\
7F6F5F0000000000000000000000000000006F5F507F7060A0907F9F8F7F4F403F
0000009080709F8070A02F209F8F7F9F907F9F8F7F0000000000000000000000009F907F908F70\
9F8F7F908F709F907F908F709F8F70908F709F907F908F70000000000000000000
000000000000000000000000000000000000000000000000000000000000000000CFCFCFC0C0C0\
CFCFCF807F508F7F5F807F50807F5F807F50807F5F8070507F7050807050807F50
3030203030207F70507F70507F704F7F705070704F7F7050706F4F70704F5F5F3F303020000000\
000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF2F2F30302F302F2F30000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000000000101010
504F6F40405F5F506F504F6F5F5070504F6F5F5070504F6F5F5070000000000000000000000000\
0000005F50705F506F6050705F50706050703F3040000000000000000000000000
000000000000000000000000000000000000000000605F706F5F7F605F706F5F7F605F706F5F7F\
6F5F706F5F7F6F5F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F10000000202020000000302F3080708F807F8F6F6070
8F7F8F807F8F8F7F8F807F8F8F7F90807F8F8F7F8F807F8F8F7F908F7F8F8F7F8F8F7F8F706070\
8F7F8F8F7F8F403F40000000706F70903F3F8F0000900000906F7090808F000000
00000000000000000090808F9F809090808F9F808F90808F9F8F909F808F70606F000000000000\
0000000000002F20209F808F9F808FA08F8F9F808FA08F8FA08F8FA08F8FA08F8F
AF8F8FA08F8FAF8F8FA08F8FAF8F8FAF8F8FAF8F8FAF8F80907F7F0F0F0F0F0F0F302F20AF908F\
806F6F0F0F0F0F0F0F100F0F0F0F0F101010100F0F000000000000000000000000
0000000000000000000000000000000000000000008F7F6FAF9080A09080AF9080A08F8080706F\
70605F70605F70605F706F5F70605000000000000000000010100F403F304F403F
0000009F907FA04F409F0000A08F7F9F907F6F60500000000000000000000000009F8F7F9F907F\
9F8F7F9F907F9F8F709F907F9F8F709F907F908F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000CFCFCF\
CFCFCFCFCFCF807F5F8F805F807F508F7F5F807F5030302010100F0F0F00000000
0000000F0F0020201F403F2F3F30207F70507F705070704F7F705070704F70704F706F4F2F2F1F\
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF302F302F2F2F2F2F30CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C00000000000000000000000000000000000000000000000001F1F20
504F6F50506F504F6F504F6F504F6F5F506F504F6F5F506F504F6000000000101F004F6F000F10\
0000000000005F50705F506F5F50705F506F5F50705F4F6F000000000000000000
000000000000000000000000000000000000000000605F706050706F5F706050706F5F70605F70\
6F5F7F605F70CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F6F6070
1F1F1F7F708080708F5F505F7F70808F7F8F3F303F4F4F50706F7F8F7F8F2F2F2F403F406F606F\
101010201F201F1F1F000000403F408F7F8F900F0F8F0000906F708F7F8F000000
000000000000403F4090808F90808F9F808F90808F9F808F90808F9F808F90808F000000000000\
0000000000002F20209F808FA08F8F9F808FA08F8F9F808FA08F8F9F8080A08F8F
A0808FA08F8FA08080A08F8FA08F80AF8F8FA08F80AF8F8F4030300F0F0F101010AF8F8F907F70\
302F2F0F0F003F2F2F0F0F0F0F0F0F201F1F10100F000000000000000000000000
000000000000000000000000000000000000000000000000000000302F2F60504FA09080A08F7F\
A09080A08F7FA090806F605070605F0000000000000000002F20203F302F000000
000000A0907F9F00009F00009F403F9F907F8F806F0000000000000000003F302F9F907F9F8F70\
9F907F908F709F907F908F709F907F908F70908F70000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000C0C0C0CFCFCF807F508F7F5F807F5030302020201F3030200000004F4030
00000000000000000000000030302050503F403F2F7F704F70704F70704F706F4F70704F6F6F4F\
000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF0000002F2F30303030000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000303040504F6F
50506F504F6F5F5070504F6F5F5070504F6F5F507050506F5F5070000000004F6F00506F005070\
002F3F0000000F0F0F5F50705F50706050705F50706050700F0F0F000000000000
0000000000000000000000000000000000000000000000006F5F7F605F706F5F7F6F5F706F5F7F\
6F5F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F0F504F5F000000100F102F202F4F40500000000000001F1F1F000000000000\
0000000000000000000000000000009080908F7F8F90606F8F7080908090000000
000000000000605F5F90808F9F808F90808F9F809090808F9F808F90808F9F8F90000000000000\
000000403F40908080A08F8F9F808FA08F8FA08F8FA08F8FA0808FA08F8FA08F8F
A08F8FA08F8FAF8F8FA08F8FAF8F8FA08F8FAF908FAF8F8FAF8F8F403030A08080AF8F8F9F807F\
0F0F0F100F0F302F2F1F10100F0F0F100F0F0F0F0F2F2020000000000000000000
000000000000000000000000000000000000000000000000000000302F2F6F5F505F504F60504F\
504F404F403FA09080A09080000000000000000000000000000000000000000000
4F4F3F9F8F7FA000009F0000A0907F9F8F7F000000000000000000000000403F309F8F7F9F907F\
9F8F709F907F9F8F709F907F908F709F907F908F70000000000000000000000000
0000000000000000000000000F0F0F000000000000000000000000000000000000000000000000\
000000000000CFCFCFCFCFCF807F508F7F5F807F502F201F807F50807F5F4F4030
807F507F704F706F4F0F0F0F00000000000040403070704F7F705070704F70705070704F70704F\
6F6F4F000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF0000002F2F302F2F30000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0000000000000000000000000000000000000000000302F40504F6F
504F6F504F6F504F6F50506F504F6F50506F504F6F5F5070504F6F10101F000F10005070004F6F\
002F3F0000000000001F1F205F50705F506F6050705F506F6050701F101F000000
000000000000000000000000000000000000000000000000605070605F70605F706F5F70C0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000090808F8F7F8090808F504F500F0F0F
00000000000090808F90808F907F8F90808F90808F9F808F90808F9F808F90808F2F2F2F000000\
0000000000009F808F9F808F9F8F8F9F808FA08F8F9F808FA08F8FA0808FA08F8F
A08080A08F8FA08F8FAF8F8FA08F80AF8F8FA08F80AF8F8FA08F80403F3F9F807FAF8F8F10100F\
0F0F0F302F20AF908F2F20200F0F0F0F0F0F0F0F0F0F0F00201F1F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
504F40403F3F403F3030302F000000000000000000000000000000000000000000
4F403F9F7F6F9F00009F00009F8F7F807F6F0000000000000000000000009F8F709F907F908F70\
9F907F908F709F907F908F709F907F605F4F000000000000000000000000000000
000000000000000000706F5F0F0F000F0F0F000000000000000000000000000000000000000000\
000000000000000000C0C0C04F405F807F50807F5F807050807F50807050807F50
7F70507F7F507F70507F70506F60406F6040000000000000706F4F70704F706F4F70704F6F6F4F\
706F4F000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF0000002F2F30302F30000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF0000000000000000000000000000000000002F2F3F504F6F504F6F
505070504F6F50506F504F6F5F5070504F6F5F5070504F6F5F5070302F3F000000002F40005070\
003F50000F0F0000000F0F0F5F50706050705F5070605F705F5070605F702F2030
000000000000000000000000000000000000000000000000000000605F706F5F7FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000001F1F1F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000605F60000000000000\
0000000000000000000000000000000000000F0F0F90808F8F808F0F0F0F000000
0000001010107F6F7090808F9F809090808F9F809090808F9F8F909F808F9F808F0F0F0F000000\
00000080707F9F808FA08F909F808FA08F8F9F808FA08F8FA08F8FA08F8FA08F8F
AF8F8FA08F8FAF8F8FA08F8FAF8F8FAF8F8FAF8F8FAF8F8FAF908FAF8F8FAF908F403030100F0F\
6F5F5FAF908F806F6F100F0F0F0F0F0F0F0F0F0F0F100F0F201F1F101010000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
80706F9F907FA000009F7060A09080807F6F0000000000000000006F60509F907F9F8F7F9F907F\
9F8F7F9F907F908F709F907F908F702F2F20000000000000000000000000000000
0000003F3F30909070908F70909070908F700F0F0F000000000000000000000000000000000000\
0000000000000000000000001F1F204F405F807F50807F5F807F50807F5F7F7F50
807F507F70507F7F507F70507F7F506F60403F3F2F00000040402F70704F70704F706F4F70704F\
6F6F4F202010000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF000000302F302F2F2F000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C00000000000000000000000000000000000002F2F3F4F4F6F504F6F
504F6F504F6F504F6F504F6F504F6F50506F504F6F50506F504F6F5F50700F0F10000000000000\
00506F0000000000000000005F50705F506F5F50705F50706050705F506F605070
000000000000000000000000000000000000000000000000000000000000C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000000000504F601F101F
4F4050000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F50606050605F505F000000000000\
0000000000000000000000000000000000000000000000002F202F0F0F0F000000
000000807080907F8F90808F90808F90808F90808F90808F90808F9F808F90808F0F0F0F000000\
70606F9F808F9F8F8F9F808FA08F8F9F808FA08F8F9F808FA08F8F9F808FA08F8F
A08F8FA08F8FA08F80A08F8FA08F8FAF8F8FA08F80AF8F8FAF8F80AF8F8F9F807F0F0F0F403F30\
1F1010A08F8080706F0F0F0F0F0F0F1F10100F0F0F0F0F0F0F0F0F0F0F0F2F2020
101010000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F00A08F709F6F5F9F907F9F8F7F0000000000000000000000008F806F908F709F907F908F70\
9F907F908F709F907F908F7080806F000000000000000000000000000000000000
201F1F908F70908F6F908F70908F70908F708F806F000000000000000000000000000000000000\
201F1F000000000000000000000000403F504F405F807F50807F507F7050807F50
7F7050807F507F704F7F70507F704F7F705070704F0F0F0F0000000F0F0F706F4F70704F6F6F4F\
70704F6F6F40000000000000CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF0000002F2F3030303030302FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF00000000000000000000000000000000000020202F504F6F50506F504F6F
50506F504F6F505070504F6F505070504F6F5F5070504F6F5F5070504F6F5F5070000000000000\
000000000000000000000000403F506050705F50706050705F5070605F705F5070
4F405F000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF0000000000000000000000000000004F40507F708F706F8F
4F4F5F706F8040404F6F607F00000F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000001010100F0F0F0000007060706F60705F5060605F60807F8F8F80908F7F8F000000\
000000000000000000000000000000000000000000000000000000000000000000
000000201F1F90809090808F90809090808F9F809090808F9F809090808F9F8F90000000000000\
7060609F8F909F808FA08F8F9F808FA08F909F8F8FA08F8FA08F8FA08F8FA08F8F
A08F8FA08F8FAF8F8FA08F8FAF8F8FA08F8FAF908FAF8F8FAF908FAF8F8FA080800F0F0F4F3F3F\
9F807FAF8F8F0F0F0F0F0F0F0F0F0FA08F807F6F600F0F0F0F0F0F100F0F0F0F0F
1F1010000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F9F8F7FA090809F907F7F706000000000000000000030302F9F8F7F9F907F9F907F9F907F\
9F8F709F907F9F8F7F9F907F000000000000000000000000000000000000000000
908F70908F70909070908F70909070908F70908F700000000000000000000000000000005F5F4F\
8F8F6F0000000000000000000000004F405F3F3F4F000000807F50807F5F807F50
807F507F7050807F507F70507F7F5070704F7F7F5070704F3F3F2F0000004F4F3070704F70704F\
6F6F4F70704F000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF0000002F2F30C0C0C070705FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF0000000000000000000000000000000000001F1020504F6F4F4F6F504F6F
4F4F6F504F6F504F6F50506F504F6F504F6F504F6F5F5070504F6F4F405F504F6F5F50701F1F20\
000000000000000000000000504F6F5F506F5F50705F506F5F50705F5070605070
5F50705F506F000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C00000000000000000000000000000000000000000000F000F000000
6060707F708F706F80706F80706F802F2F305F5F6F2F2F30000000101010000000000000000000\
000000000000000000000000000000100F100000000000001010100000004F404F
2F2F302020201F101F807F8F80708F7F6F7F807F8F8F7F8F807F8F8F7F8F30303F302F30000000\
000000000000000000000000000000000000000000000000000000000000000000
0000008F808F90808F90808F90808F90808F90808F9F808F90808F9F808F90808F5F50505F4F50\
9F808F9F808F9F8F8F9F808F9F8F8F9F808FA08F8F9F808FA08F8FA0808FA08F8F
A0808FA08F8FA08F8FAF8F8FA08F80AF8F8FA08F8FAF8F8FA08F80AF8F8FAF8F80302F2FA08F80\
AF8F8F20201F0F0F0F1F1F1F806F609F807F100F0F0F0F000F0F0F0F0F0F0F0F0F
3F302F5F4F40000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F209F8F7F605F4F0000000000000000004F403F908F709F907F9F8F7F9F907F908F70\
9F907F908F709F907F100F0F0000000000000000000000000000000000006F6050
908F70908F70908F70908F708F8F6F908F708F8F6F0000000000000000000000000000005F5040\
6060500000000000000000000000000000001F1F200000000000007F7F50807F50
7F70507F7F507F70507F7F5070704F7F705070704F7F7050706F4F0F0F0F00000070704F6F6F40\
6F6F4F6F6F403F3F20000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF0000002F2F30CFCFCF706F50CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
00000000000000000000000000000000000010101F504F6F504F6F504F6F504F6F
505070504F6F505070504F6F5F5070504F6F5F5070504F6F504F6F0000004F405F504F6F5F5070\
0000000F0F0F00000020202F5F50705F50705F5070605F705F50706050705F5070
605F706050700F0F0F000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF000000000000000000000000000000000000000000000000000000000000
0000000F000F40404F5F506F7F708F7F708F7F708F7F708F706F80403F4F5F5F6F2F202F706F80\
100F105F5060101010000000403F4F2F2F30706F80605F6F80708F807F905F5060
807F90807F8F807F90807F8F8F7F90807F8F8F7F904F404F4F404F3F3F403F303F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
504F5090808F90809090808F9F809090808F9F809090808F9F8F9090808F9F8F9070606F9F8F90\
9F808F9F8F909F808FA08F909F8F8FA08F8F9F808FA08F90A08F8FA08F8FA08F8F
A08F8FA08F8FAF8F8FA08F8FAF908FA08F8FAF8F8FAF8F8FAF908FAF8F8FAF908FAF8F8FAF908F\
5F4F4F0F0F0F0F0F0F20201F806F60A080800F0F0F100F0F0F0F0F0F0F0F807060
AF908FAF90809F807F6050500F0F0F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2010100F0000000000000000006F60504F403F9F907F9F907F9F907F9F8F7F9F907F\
9F907F9F907F6F6050101010000000000000000000000000000000000000908F70
909070908F70909070908F70909070908F70908F70101010000000000000000000000000000000\
8F8F6F000000000000000000000000000000000000000000000000807F5F7F7F50
807F507F70507F7F507F70507F7F507F704F7F705070704F7F705070704F4F4F3000000070704F\
6F6F4F6F704F3F3F20000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF0F0F0F302F30C0C0C0707050CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
0000000000000000000000000000000F0F1010101F4F4F6F504F6F4F4F6F504F6F
4F4F6F504F6F4F4F6F504F6F3F304F505070504F6F50506F504F6F0000003030405F5070504F6F\
4F40603F3F50504F6F302F3F5F50705F506F5F50705F50705F50705F50705F5070
5F5070605F705F50702F2030000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000706F804040507F708F706F807F708F706F807F708F6F607F\
7F708F7F70807F708F7F708F5F5F6F7F708F80708F7F708F807F8F706F80807F8F
80708F807F8F6F6070807F8F80708F3F3F40504F504F4F50000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000009080908F7F8F90808F90808F90809090808F90808F90808F9F808F90808F70606F8F7F80\
9F8F8F9F808F9F8F8F9F808F9F8F8F9F808FA08F8F9F808FA08F8F9F808FA08F8F
A08F8FA08F8FA08F8FA08F8FA08F8FAF8F8FA08F80AF8F8FAF8F8FAF908FA08F80AF8F8FA08F80\
5F504F0F0F000F0F0FA08080AF908F1010100F0F0F0F0F0F0F0F0F7F6F60AF9080
A08F80AF908F908070302F2F000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000101010605F4F9F907F9F8F7F9F907F908F709F907F908F70\
9F907F908F706F6050000000000000000000000000000000000000404030908F70
908F70908F708F8F6F908F708F8F6F908F708F806F403F300000000000000000000000002F2F20\
8F8F6F0F0F0F000000000000000000000000000000000000000000000000807F50
7F70507F7F507F704F7F705070704F7F705070704F70704F70704F70704F6F6F4F0000006F6F4F\
6F704F6F6F406F6F4F000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF1010102F2F30CFCFCF707050CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000100F104F4F6F504F6F4F4F6F50506F504F6F
504F6F504F6F505070302F400000000000005F5070504F6F10101F0000001F1F205050705F5070\
50506F5F50705F50705F50705F50705F50705F50705F50705F5070605F705F5070
605F705F5070605F7F6050703F3040000000706F6F6F6F60CFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000F0F0F1F1F1F4F4F5F0000005F5F6F0F0F106F607F\
7F708F7F70807F708F7F708F7F708F706F7F7F708F807F903F3F406F6070000000
202020302F305F5F600000001F101F3F3040000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
6F5F6F90808F90809090808F90809090808F9F8F9090808F9F809090808F9F8F909F808F9F8F90\
9F808F9F8F909F808F9F8F909F808FA08F909F8F8FA08F8F9F8F8FA08F90A08F8F
A08F8FA08F8FAF8F8FA08F8FAF8F8FA08F8FAF908FAF8F8FAF908FAF8F8FAF908FAF8F8F8F706F\
0F0F0F100F0F80706F80706F2F2020100F0F0F0F0F2F2F200F0F0F80706FAF9080
AF908FAF9080AF908FAF9080A09080403F30000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002020207F70609F90809F907F9F907F9F8F7F9F907F9F907F9F907F\
908F7F9F907F0000000000000000000000000000000000000000009F907F908F70
909070908F70909070908F709090708F8F709090700000000000000000000000000000000F0F0F\
80806F00000000000000000000000000000000000000000020202F000000000000
807F507F70507F7F507F70507F7F5070704F7F7F5070704F70705070704F7070500000002F2F20\
6F6F4F6F704F6F6F40000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF0000002F2F30C0C0C070705FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
00000000000000000000000000000F00000F504F6F4F4F6F504F6F4F4F6F504F6F
4F4F6F504F6F504F6F302F400000001F1F20504F6F5050701F1F20000000504F6F5F5070504F6F\
5F5070504F6F5F5070504F6F5F50705F50705F50705F506F5F50705F5070605070
5F50706050705F5070605F705F5070706F6F6F6F60706F6FC0C0C0CFCFCFC0C0C0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000F5050600F0F0F5F50600000001010102F2F2F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000003F303F5F505F
8F7F8F90808F90808F90809090808F90808F90808F90809090808F9F808F90808F9F8F9090808F\
9F808F9F808F9F8F8F9F808F9F8F8F9F808FA08F8F9F808FA08F8F9F8F8FA08F8F
A08F8FA08F8FA08F8FA08F8FA08F8FAF8F8FA08F8FAF8F8FA08F80AF8F8FAF8F8FAF908FA08F80\
806F6FA08F809F807F0F0F0F0F0F0F0F0F0F0F0F0F2F2020AF908FA08F80AF908F
A08F80AF9080A09080AF908FA08F8010100F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F3F308F806F9F907F9F8F7F9F907F908F7F9F907F908F7F9F907F908F70\
9F907F000000000000000000000000000000000000000000908F70908F70909070
908F70908F70908F70908F708F8F6F908F708F8F6F1010100000000000000000000000004F4F3F\
8F8F6F0F0F00000000000000000000000000000000000000000000000000000000
7F70507F7F507F704F7F7F5070704F7F70506F6F4F70705070704F70704F6F6F4F70704F000000\
6F6F4F6F6F406F6F4F202010000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000504F6F4F4F6F504F6F4F4F6F504F6F4F4F6F
505070504F6F2F2F3F0000000000001F1F2050507040405F0000003F304F4F4060504F6F5F5070\
5050705F50705F50705F50705F50705F50705F50706050705F50706050705F5070
605F7F5F5070605F705F50700000006F6F60706F6F6F6F60706F6FCFCFCF000000000000000000\
0000000000000000000000000000000F0F0F0000000F0F0F000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000F0F0F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000201F20504F50
90809090808F90809090808F90809090808F90809090808F9F8F9090808F9F8F9090808F9F8F90\
9F808F9F8F909F808FA08F909F8F8FA08F909F8F8FA08F90A08F8FA08F8FA08F8F
A08F90A08F8FA08F8FA08F8FAF908FA08F8FAF908FA08F8FAF908FAF8F8FAF908FAF8F8FAF908F\
806F6FAF8F8F9F807F100F0F0F0F0F0F0F0F20201FAF908FAF9080AF908FAF9080
AF908FAF9080AF908FAF9080AF908FAF9080A08F809F807F8F7F70000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000005F504F908F7F9F90809F907F9F907F9F907F9F907F9F907F9F907F908F7F9F907F\
706F5F00000000000000000000000000000020201F908F70908F70909070908F70
90907F908F70909070908F70909070908F70908F7080806F0000000000000000000000001F1F1F\
8F8F6F4040300000000000000000000000000000000000002F2F207F7F5F000000
CFCFCF7F7F506F6F4F504F303F3F2F00000040402F5F5F3F70704F70704F70704F6F6F4F000000\
6F6F4F6F6F4F606F401F1F10000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF000000CFCFCFC0C0C000705FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000000000\
0000000000000000000000003030404F40604F4F6F4F4F6F504F6F4F4F6F504F6F
4F4F6F504F6F2020300000001F1F20505070504F6F504F6F0000002F2F3F504F6F505070504F6F\
5F5070504F6F5F507050506F5F5070504F6F5F50705F50705F50705F50705F5070
5F50706050705F50700000006F6F60706F6F6F6F60706F6F6F6F60000000000000000000000000\
0000000000000000000000000F0F0F00000F706F8030303F10101F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000005F505F8070808F7F8F908090
8F7F8F9080908F7F8F90809090808F90809090808F90809090808F9F809090808F9F808F90808F\
9F8F9090808F9F8F8F9F808F9F8F909F808F9F8F8F9F808FA08F8F9F808FA08F8F
A08F8FA08F8FA08F8FA08F8FA08F8FAF8F8FA08F8FAF8F8FA08F8FAF908FA08F80AF908FAF8F8F\
AF908F8F70702020200F0F0F0F0F0F6F5F5FAF908FA08F80AF908FA08F80AF908F
A09080AF908FA08F80AF9080A09080AF9080A08F8090807080706F7F70606F6050605F50504F40\
000000000000000000000000000000000000000000000000000000000000000000
0000000F0F0F504F409F907F9F8F7F9F907F908F7F9F907F908F7F9F907F908F709F907F908F70\
504F4000000000000000000000000020201F908F70908F70909070908F70908F70
908F709090708F8F70908F708F8F70908F708F8F6F3F3F2F0000000000000000000000006F6F50\
8F8F6F0000000000000000000000000000000000000000003F3F2F2F2F20000000
CFCFCF5F503F0000000000000000000000000000000000004F4F306F6F406F6F4F70704F1F1F10\
40402F606F406F6F4F606F40000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCF000000000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
0000000000000000000000001F1020504F6F4F4F6F504F6F4F4F6F504F6F4F4F6F
504F6F20202F0000000000001F1F20504F6F5050700F0F100000001F10205F50705050705F5070\
504F6F5F50705F50705F50705F50705F50705F50705F50705F5070605F705F5070
605F705F5070000000000000706F6F6F6F6070706F706F60000000000000000000000000000000\
0000000000000000000F0F0F0F0F0F70708F706F8F70708F5F506F706F8F30303F
1010102F2F30000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F1F1F3F303F5F50607F707F9080908F808F
9080908F808F90809090808F90809090808F9F8F9090808F9F8F9090808F9F8F909F808F9F8F90\
9F808F9F8F909F808F9F8F909F808FA08F909F8F8FA08F909F8F8FA08F90A08F8F
A08F8FA08F8FAF9090A08F8FAF908FA08F8FAF908FAF8F8FAF908FAF8F8FAF908FAF8F8FAF908F\
AF8F8FAF908FAF908F7F6F606F5F5F706060AF908FAF908FAF9080AF908FAF9080
AF908FAF9080AF908FAF9080AF908FA09080AF908FA09080AF9080A09080AF908F60504F504F40\
4F403F403F305F504F000000000000201F1F403F306F60508F80700F0F0F403F30
807F6F6F605FA090809F907F9F90809F907F9F90809F907F9F907F9F907F9F907F9F907F9F907F\
00000000000000000000000020201F908F70908F7090907F908F7090907F908F70
909070908F70909070908F709090708F8F70909070606050000000000000000000000000000000\
8F8F6F30302F00000000000000000000000000000000000010100F6F705F000000
CFCFCFCFCFCF50503F5F5F3F3030203F3F2040402F50503F0000000000003030206F6F4F20201F\
40402F6F704F606F406F6F4F000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF000000CFCFCFC0C0C06F6F50000000CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000\
0000000000000000000000000000004F4F6F4F4F6F4F406F4F4F6F4F4F6F504F6F
4F4F6F201F2F00000010101F4F4F6F504F6F504F6F1F1F20000000505070504F6F505070504F6F\
5F5070504F6F5F5070504F6F5F50705050705F50705F50705F50705F50705F5070
5F50700000000000005F50506F6F60706F6F6F6F60000000000000000000000000000000000000\
00000000000000000000000F706F8F6F6F80706F8F6F6F80706F8F706F8F70708F
706F80706F80504F6010101F1F1F1F00000020202F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001010108F80908F7F8F8F80908F7F8F908090
8F7F8F9080908F808F9080908F808F90809090808F90809090808F90809090808F9F8F9090808F\
9F8F909F808F9F8F909F808F9F8F8F9F808FA08F909F808FA08F8F9F8F8FA08F8F
9F8F8FA08F8FA08F8FA08F8FA08F8FA08F8FA08F8FAF908FA08F8FAF908FA08F8FAF908FA08F80\
AF908FA08F8FAF908F70605F6F5F5FA08F80AF908FA08F80AF908FA09080AF908F
A08F80AF908FA09080AF908FA09080AF9080A09080AF9080A09080A09080A09080A09080A08F80\
A09080A090806F605F7F7060A090809F907FA09080605F508F80709F907F9F9080
908F7F9F907F9F907F9F907F9F8F7F9F907F9F8F7F9F907F908F7F9F907F908F7F9F907F2F2020\
00000000000000000020201F2F2020908F7090907F908F70909070908F70909070
8F8F70908F708F8F70908F708F8F6F908F708F8F6F8F8F70000000000000000000000000000000\
8F8F6F00000000000000000000000000000000000000000060604F20201F000000
000000CFCFCFC0C0C07F7F5070704F7F705070704F50503F3F3F2F0000000F0F0070704F6F6F40\
000000606F406F6F4F606040000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCF7070507F705FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
00000000000000000000000030304F4F4F6F4F4F6F504F6F4F4F6F504F6F4F4F6F
3F3F500000000F0F100F0F10505070504F6F403F50000000302F3F3F3F50505070504F6F5F5070\
5050705F50705050705F50705F50705F50705F50705F50705F50705F50705F5070
605F7F0000006F60606F6F6070706F706F6F000000000000000000000000000000000000000000\
0000000000000000005F506F706F8F706F8F706F8F70708F706F8F70708F706F8F
7F708F706F8F7F708F706F8F7F708F70708F5F5F6F605F7030303F1F1F200F0F0F0000000F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000605F606F60708F80908F7F8F9080908F80909080908F808F
90809090809090809090808F90809090808F90809090808F9F8F9090808F9F8F9090808F9F8F90\
9F808F9F8F909F808F9F8F909F8F8FA08F909F8F8FA08F909F8F8FA08F90A08F8F
A08F90A08F8FA08F90A08F8FAF9090A08F8FAF908FA08F8FAF908FAF8F8FAF908FAF8F8FAF908F\
AF908FAF908FAF8F8FAF908FAF908FAF908FAF908FAF908FAF908FAF908FAF9080
AF908FAF9080AF908FA09080AF908FA09080AF908FA09080AF908FA09080A09080A09080A09080\
A09080A09080A09080A09080A09080A090809F9080A090809F9080A090809F907F
A090809F907F9F90809F907F9F90809F907F9F907F9F907F9F907F9F907F9F907F908F7F0F0F0F\
000000000000000000908F7090907090907F908F7090907F908F7090907F908F70
909070908F709090708F8F709090708F8F709090708F8F6F20201F0000000000000000000F0F0F\
8F8F6F30302F0000000000000000000000000000000000002F302060604F000000
000000000000CFCFCF7F70507F7F5070704F7F705070704F70705070704F0F0F0000000070704F\
20201F6F6F4F606F406F6F4F20201F000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C070705F706F50CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000000000\
00000000000000000000000020202F4F40604F4F6F4F4F6F4F4F6F4F4F6F4F4F6F
403F5F000000000000504F6F4F4F6F504F7040405F0000001F1F20505070504F6F505070504F6F\
5F5070504F6F5F5070504F705F5070504F6F5F50705F50705F50705F50705F5070
0000001010106F6F60706F6F6F6F606F6F8F6F6080000000000000000000000000000000000000\
000000000000000000000000706F8F6F6F80706F8F706F8F706F8F706F80706F8F
706F8F70708F706F8F70708F706F8F7F708F706F8F7F708F706F8F60607000000F1010100F000F\
1010100F0F0F10101F0F0F0F1F1F1F0F0F0F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000003F303F30303F4F404F7060708F7F8F8F80908F7F8F8F80908F7F8F8F8090
8F7F8F9080908F7F8F90809090808F90809090808F90809090808F9F809090808F9F809090808F\
9F8F9090808F9F8F909F808F9F8F909F808F9F8F8F9F8F8FA08F909F8F8FA08F8F
A08F8FA08F8FA08F8FA08F8FA08F8FAF8F8FA08F8FAF8F8FA08F8FAF908FA08F8FAF908FA08F8F\
AF908FA08F80AF908FA08F8FAF908FA08F80AF908FA09080AF908FA08F80AF908F
A09080AF908FA09080AF908FA09080AF908FA09080A09080A09080A09080A09080A09080A09080\
A090809F9080A090809F9080A090809F907F9F90809F907F9F90809F8F7F9F9080
9F907F9F90809F8F7F9F907F9F907F9F907F908F7F9F907F908F7F9F907F908F707F7F60000000\
0000002F2F2090907F908F7090907F908F70909070908F709090708F8F70908F70
8F8F709090708F8F6F908F708F8F70908F708F8F6F8F8F70000000000000000000000000000000\
8F8F6F0000000000000000000000000000000000000000007F7F607F8060000000
000000000000000000CFCFCF70704F7F705070704F70704F70704F70704F10100F0000006F6F40\
20201F606F40606F406060401F1F10000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCF70705F7F705FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000\
0000000000000000000000000000004F4F6F4F4F6F4F4F6F4F4F6F504F704F4F6F
00000F000000000000504F6F504F70504F6F0F0F0F000000101010504F6F5F5070504F705F5070\
504F705F50705050705F50705050705F50705F50705F50705F50705F507F000000
0F0F0F6F6F6F70706F706F6F6F6F8F6F60806F6F8F6F6F80000000000000000000000000000000\
00000000000000000000000040405070708F706F8F70708F706F8F70708F706F8F
70708F706F8F7F708F706F8F7F708F706F8F7F709070708F7F708F5F506F10101F0F0F0F0F0F0F\
0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F101010101010201F2010101020202F
00000000000000000F0000000F0F0F000000000000000000000000000000000000000000504F5F\
4F405040404F8F7F908F80908F7F8F8F80908F7F908F80908F7F8F9080908F8090
9080908F808F90809090809090809090808F908F909080909F8F9090808F9F8F9090808F9F8F90\
90808F9F8F909F8F8F9F8F909F8F8FA08F909F8F8FA08F909F8F8FA08F90706F70
706F70706F70706F70706F70706F7050505F303030000000000000000000000000000000000000\
000000000000000000AF908FAF908FAF908FAF908FAF908FAF9080AF908FAF908F
AF908FA09080AF908FAF9080AF908FA09080AF908FA09080AF908FA09080AF908FA09080A09080\
A09080A09080A09080A090809F9080A090809F9080A090809F907FA090809F907F
9F90809F907F9F90809F907F9F90809F907F9F90809F907F9F907F90907F9F907F0F0F0F1F1F1F\
6F6F509F907F90907F9F907F908F709F907F908F7090907F908F7090907F908F70
9090708F8F709090708F8F709090708F8F709090708F8F7030302000000000000000000020201F\
8080605050400000000000000000000000000000000000004F4F3F505F40000000
000000000000000000CFCFCFCFCFCF70704F7F7F5070704F7070506F704F70704F00000020201F\
6F6F406F704F606F40606F4F606040000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C07F705F706F50CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000\
0000000000000000000000002020304F406F4F4F6F4F406F4F4F6F4F4F6F4F4F6F
0F0F1000000040405F504F6F4F4F6F504F6F0F0F100000004F4060504F70504F6F505070504F6F\
505070504F6F5F5070504F6F5F50705050705F5070504F705F5070000000000000
50504F706F6F6F6F60706F6F6060806F6F806F60806F6F8F6F6080000000000000000000000000\
000000000000000000000000302F3F6F6F80706F8F6F6F80706F8F706F8F70708F
706F8F70708F706F8F70708F706F8F70708F706F8F7F708F706F8F4040500F0F0F0F0F0F201F20\
0F0F0F0F000F0F0F0F00000F0F0F0F0F000F0F0F0F00000F1F1F1F0F000F20202F
2F2F30605F701F1F1F504F5F20202F0000000F0F0F000000000000706F7F6F6070605F6F504F5F\
8F7F90807F8F8F8090807F8F8F80908F7F8F8F80908F7F8F8F80908F7F8F8F8090
8F7F8F9080908F808F9080908F808F90809090808F90809090808F90809090808F9F8F9090808F\
9F8F90706F707F707F706F707F6F70706F707F7070706F70706F70706F70706F70
706F70706F70706F70706F702F2F2F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F2FA09080AF908FA09080A09080A09080A09080A09080A09080A09080A090809F9080\
A090809F9080A090809F907FA090809F907FA090809F907F9F90809F907F9F9080
9F8F7F9F907F9F907F9F9080908F7F9F907F90907F9F907F908F7F9F907F706F5F10100F908F70\
90907F908F7090907F908F7090907F908F7090907F908F70909070908F70909070
8F8F70908F708F8F70908F708F8F6F8F8F708F8F6F8F8F7070705F000000000000000000000000\
8F8F6F10100F00000000000000000000000000000000000080806F7F8060000000
000000000000000000000000C0C0C070705070704F7070506F6F4F70704F6F6F4F6F704F000000\
6F6F4F606F40606F4F606040606F40000000000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCFCFCFCF000000000000CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000\
00000000000000000000000010101F4F4F6F4F4F6F504F6F4F4F6F504F6F2F2F40
0000003F3F5040405F4F4F6F50507030304F0000001F1F2F40405F504F6F505070504F6F5F5070\
504F705F5070504F705F50705050705F50705050705F5070000000000000000000
70706F706F6F70706F6060806F6F8F6F60806F6F8F6F60806F6F8F6F6F8F000000000000000000\
0000000000000000000000001F1F20706F8F6F6F8F70708F706F8F70708F706F8F
70708F706F8F70708F706F8F7F709070708F7F708F70708F7F709070708F0F0F0F0F0F0F100F10\
605F6F0F0F0F0F0F0F1010100F0F0F0F0F0F0F0F0F0F0F100F0F0F0F0F0F0F0F0F
0F0F0F807F8F807F90807F8F807F907F708F30303F807F8F807F8F706F7F8F7F90807F8F8F8090\
807F908F8090807F8F8F80908F7F908F80908F7F908F80908F80908F80908F8090
9080908F80909080908F808F90809090809090809090808F9F8F907F6F70303030706F6F7F707F\
7F70707F707F706F707F707F706F707F707F706F707F707F706F7070707F706F70
70707F706F70706F70706F70000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000000000020202F1F1F1F504F506F6070AF908FA09080AF908FA09080A09080A09080A0908F\
A09080A090809F9080A090809F9080A090809F9080A090809F90809F90809F907F
9F90809F907F9F90809F907F9F90809F907F9F907F9F907F9F907F9F907F70706090907F9F907F\
90907F9F907F908F709F907F90907090907F908F7090907F908F7090907F908F70
90907F908F709090708F8F709090708F8F709090708F8F6F3F3F300000000000000000000F0F0F\
60604F606F50000000000000000000000000000000000000606F504F4F3F0F0F00
000000000000000000000000CFCFCFCFCFCF70705070704F7070506F704F70704F6F6F4F000000\
303F206F6F4F606F40606F4F606F40000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCF707050000000C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000\
00000000000010101F30304F3030404040604F4F6F4F406F4F4F6F4F406F3F3F50
0000003F3F504F4F6F504F6F4F4F6F504F70000000101010504F6F505070504F6F505070504F6F\
505070504F6F505070504F705F5070504F705F50705050700000000000000F0F0F
6F6F60706F6F6F6F606F60806060806F6F8F6060806F6F8F6F60806F6F8F6F6080000000000000\
0000000000000000000000000000004F4F5F706F8F6F6F8F706F8F6F6F80706F8F
706F8F70708F706F8F70708F706F8F70708F706F8F70708F706F8F706F8F2F2F300F0F0F201F20\
6F6F7F00000F0F0F0F7F708F20202F00000F10101F0F0F0F10101000000F0F0F0F
20202F807F907F708F807F9080708F807F9080708F807F90807F8F807F90807F8F807F90807F8F\
8F7F90807F8F8F7F90807F8F8F8090807F8F8F80908F7F8F8F80908F7F8F8F8090
8F7F8F9080908F7F8F9080908F808F0000000000000F0F0F706F702020204F4F4F7F707F7F6F70\
7F707F706F707F707F706F707F707F706F70706F70706F701F101F706F70706F70
706F70706F706F6F70403F40000000000000000000000000000000000000000000000000000000\
2020200000002F2F2F000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006F606F6F60706F606F6F6F70A09080A09080A09080\
A090809F9080A090809F9080A090809F907F9F90809F907F9F90809F907F9F9080
9F907F9F9080908F7F9F907F9F907F9F907F908F7F9F907F90907F9F907F908F7F9F907F908F7F\
9F907F908F7090907F908F7090907F908F70909070908F709090708F8F70909070
8F8F709090708F8F6F908F708F8F6F908F708F8F6F8F8F7070705F000000000000000000000000\
8F8F6F2F2F2000000000000000000000000000000000000030302F707F5F000000
000000000000000000000000000000CFCFCF70704F70704F6F704F70704F6F6F406F704F6F6F40\
0000005F5F3F606F40606040606F40000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCFCFCFCF7F705F30302FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000\
1010203F3F504040604F406030304F4F4F6F4F4F6F4F4F6F4F4F6F504F70000000
1F1F204F4F6F504F704F4F6F504F703F3F50000000000000505070504F6F505070504F70505070\
504F705F50705050705F50705050705F50705F5070000000000000000000706F6F
70706F706F6F6F6F8F6060806F6F8F6060806F6F8F6F6F806F6F8F6F60806F6F8F6F6F8F000000\
0000000000000000000000000000000000006F6F8F706F8F6F6F8F70708F706F8F
70708F706F8F70708F706F8F70708F706F8F7F709070708F7F709070708F201F200F0F0F7F7090\
7F708F0F0F0F0F0F0F6F607F1F1F1F0F0F0F30303F7F708F5F506F0F0F0F0F0F0F
605F6F7F708F807F90807F90807F90807F8F807F90807F90807F90807F8F8F8090807F908F7F90\
807F908F80908F7F908F80908F7F908F80908F7F908F80908F7F909080908F8090
807F7F4040401010100000000000000000000000000F0F0F2F2F2F0F0F0F7F707F0F0F0F7F707F\
7F6F707F707F7F6F707F707F706F707F707F706F706060600000007F707F706F70
706F70706F7070707F0000000000000000000000000000000000000000003030303F303F706F70\
6F6F70706F706F6F70706F706F6F70706F706F6F70504F50403F40000000000000
00000000000000000000000000000000000000000010101030303F6F60700F0F0F000000000000\
000000A09F80A09080A090809F9080A090809F9080A090809F907FA090809F9080
9F90809F907F9F90809F907F9F90809F907F9F90809F907F9F907F90907F9F907F90907F9F907F\
90907F9F907F90907F90907F908F7090907F908F7090907F908F7090907F908F70
9090708F8F709090708F8F709090708F8F709090708F8F7070705F000000000000000000000000\
4040307F8060000000000000000000000000000000000000707F5F40403F0F0F0F
000000000000000000000000000000CFCFCFCFCFCF70704F70704F6F704F70704F6F6F4F6F704F\
00000050503F606F40606F4F5F6040000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCF7070507F705FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000000000003F3F50403F5F\
4040604040604040604040604040604F406F4F4F6F40406F4F4F6F4F4F6F10101F
0000004F4F6F4F4F6F504F6F4F4F6F504F6F00000040405F4F4F6F504F70504F6F505070504F6F\
505070504F6F505070504F6F505070504F705F5070000000000000000000706F6F
6F6F600000000000006F6F8F6060806F6F8F6060806F6F8F6060806F6F8F6F60806F6F8F6F6080\
0000000000000000000000000000000000000F0F106F6F80706F8F6F6F8F706F8F
6F6F8F706F8F706F8F70708F706F8F70708F706F8F70708F706F8F7F708F2F2F300F0F0F0F0F10\
7F708F0F0F1010101070708F605F700F000F0F0F0F7F708F3F30400F000F0F0F0F
201F20807F907F708F807F907F708F807F90807F8F807F90807F8F807F90807F8F807F90807F8F\
807F90807F8F8F8090807F8F8F7F90807F8F8F80908F7F8F000000000000000000
101010807F7F706F703F3F3F0000000000000000000F0F0F0000000F0F0F201F201F1F1F706F70\
7F707F706F707F707F706F707F707F706F707F707F000000000000706F70706F70
706F70706F705F505F0000006F6F704F404F0000000000000000000000006F6F70706F70505050\
6F606F0F0F0F20202000000020202F0000003F3F3F6F606F4F4F50605F606F6F70
6F606F1F101F0000000000000000000000000000000000000000000F0F0F000000000000000000\
000000000000000000C0C0C09F90809F90809F90809F907F9F90809F907F9F9080
9F907F9F90809F907F9F9080908F7F9F907F90907F9F907F908F7F9F907F908F7F9F907F908F70\
90907F908F7090907F908F7090907F908F7090907F8F8F709090708F8F70909070
8F8F70908F708F8F709090708F8F6F8F8F6F8F8F6F6F6F500F0F00000000000000000000000000\
70705F808F6F0000000000000000000000000000000000004040307F8060000000
000000000000000000000000000000CFCFCFC0C0C070704F6F6F4F6F704F6F6F4F6F704F606F40\
6F6F4F000000606F405F6040606F40000000000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF302F30CFCFCFCFCFCFCFCFCF7F705F7F705FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000001F1F20404060404060\
4040604040604040604F4F604040604040604F4F6F4F4F6F4F4F6F4F4F6F1F1F2F
0000004F4F6F504F704F4F6F5050702F2F3F0000000F0F10505070504F70505070504F6F505070\
504F70505070504F705F50705050705F5070000000000000000000000000706F6F
70706F0000000000000000006F6F8F6F60806F6F8F6F60806F6F8F6F6F8F6F6F8F6F6F806F6F8F\
00000000000000000000000000000000000000000060607F706F8F70708F706F8F
70708F706F8F70708F706F8F7F709070708F7F708F706F8F7F709070708F7F70903F304030303F\
7F708F0F0F0F0F0F0F7F7F907F708F0F0F0F10101F605F701010100F0F0F0F0F0F
605F6F7F7F90807F907F7F8F807F90807F90807F90807F90807F90807F90807F90807F908F8090\
807F908F8090807F908F80908F7F908F80900F0F0F807F7F807F7F000000000000
000000000000807F7F7F707F5F505F0000000000000000001010100F0F0F0F0F0F7F70707F707F\
7F707F7F707F706F707F707F706F707F707F706F7000000000000070707F706F70
70707F706F70000000000000706F7F70606F0F0F0F000000000000000000504F50000000000000\
000000000000000000000000000000000000000000000000000000000000504F50
403F4050505F60606F504F500F0F0F000000000000000000000000000000000000000000000000\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFA09F809F90809F90809F907F
9F90809F907F9F90809F907F9F90809F907F9F907F90907F9F908090907F9F907F90907F9F907F\
90907F9F907F9090709F907F90907090907F908F7090907F908F709090708F8F70
9090708F8F709090708F8F707F7F6030302F0F0F0F0F0F0F0F0F0F0F0F0F000000000000000000\
1F1F1F8F8F6F00000000000000000000000000000000000070705F5F5F4F20201F
000000000000000000000000000000000000CFCFCF70704F7070506F704F6F704F6F6F4F6F704F\
606F403F3F2F303F200F0F0F2F3020000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF101010CFCFCFC0C0C0CFCFCF7070507F705FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF00000000000000000F404060404060\
40406040405F4040603F3F5F4040604040604F4F6F4F406F4F4F6F3F3F50000000
2020304F4F6F4F4F6F4F4F6F4F4F6F4F40600000003F3F504F4F6F504F704F4F6F504F70504F6F\
505070504F6F505070504F70505070504F70000000000000000000000000706F6F
0000000000000000000000005F5F7F6F6F8F6060806F6F8F6060806F6F8F6060806F6F8F6F6080\
6F6F8F000000000000000000000000000000000000000000706F8F6F6F8F706F8F
6F6F8F70708F6F6F8F70708F706F8F70708F706F8F70708F706F8F7F708F706F8F7F708F70708F\
7F709010101F0F0F0F706F8060607F00000F0F0F0F706F804F4F5F00000F0F0F0F
1F1F20807F907F708F807F907F708F807F907F708F807F90807F8F807F90807F8F807F90807F8F\
8F7F90807F8F8F7F90000000000000000000000000000000807F7F7F707F000000
0000000000000000007F707F7F707F2F2F2F0000000000000000000F0F0F5F505F7F707F706F70\
7F707F706F707F707F706F707F707F706F70201F20000000000000706F70706F70
706F70706F70000000000000706F707000005F505F1010100F0F0F4F404F0000000000000F0F0F\
000000000000000000000000000000000000000000000000000000000000000000
00000000000000000060606F6F606F202020000000000000000000000000000000000000000000\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF9F907F9F9080
9F907F9F908090907F9F907F90907F9F907F908F7F9F907F90907F9F907F908F7F90907F908F7F\
90907F908F7090907F908F7090907F8F8F709090708F8F709090708F8F70909070
8F8F709090708F8F6F8F8F701F1F1F0F0F0F0F0F000F0F0F0F0F00000000000000000000000000\
50504080806F30302000000000000000000000000000000020201F7F8060000000
000000000000000000000000000000000000C0C0C0CFCFCF6F704F6F704F6F6F406F6F4F606F40\
6F6F4F606040000000000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCF7F705F000000CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF00000000000000000040406010101F\
0000000000000000004040604040604040604F4F6F4F4F6F4F4F6F4F4F6F000000
1F1F204F4F6F504F704F4F6F504F7030304F00000F1F1F2F5050704F4F6F505070504F70505070\
504F70505070504F70505070504F700000000000000000000000000F0F0F000000
0000000000000000000000000000000F0F0F6F6F8F6F60806F6F8F6F60806F6F8F6F6F8F6F6F8F\
6F6F8F6F6F8F00000000000000000000000000000000000020202F70708F706F8F
70708F706F8F707090706F8F70708F706F8F7F709070708F7F709070708F7F709070708F7F7090\
70708F706F8F2020207F70907F708F0F0F0F0F0F0F7F7F901F1F200F0F0F0F0F0F
5F506F7F708F807F90807F90807F90807F90807F90807F90807F90807F908F7F90807F908F7F90\
807F90000000000000000000000000000000000000000000000000807F7F6F6060
000000000000000000000000707070706F700000000000000000000000006F606F7F707F7F707F\
706F707F707F706F7F7F707F706F70706F700000000000000000007F707F706F70
70707F605F60000000000000706F70700000706F7F0000005F505F706F70706F70706F70706F70\
5F505F706F701F1F1F101010000000000000000000000000000000000000000000
000000000000000000000000000000101010706F70000000000000000000000000000000CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
9F90809F907F9F90809F907F9F908090907F9F908090907F9F907F90907F9F907F90907F9F907F\
90907F9F907F90907F90907F908F7090907F90907090907F8F8F7090907F8F8F70
9090708F8F709090703F3F2F0F0F0F0F0F0F2F2F20202F200F0F0F10100F000000000000000000\
0000008F8F6F808F6F0000000000000000000000000000004F503F808F6F0F0F0F
000000000000000000000000000000000000000000302F3F70704F6F6F4F6F704F6F6F4F6F704F\
606F406F6F4F003F2F000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF000000CFCFCFC0C0C0CFCFCFC0C0C07F705F000000CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000000000000000000000000F0F0F000000\
00000010101F2F2F4030304F4040603030504F4F6F40406F4F4F6F4F406F000000
0000004F4F6F4F4F6F4F4F704F4F6F4F4F6F0000004F4F6F4F4F6F504F704F4F6F5050704F4F6F\
504F70504F6F505070504F6F000000000000000000000000000000000000000000
0000000000000000000000000000000000001F1F206F6F8F6060806F6F8F6060806F6F8F606080\
6F6F8F6F6F806F6F8F0000000000000000000000000000001F1F206F6080706F8F
6F6F8F706F8F6F6F8F70708F6F6F8F70708F706F8F70708F706F8F70708F706F8F7F7090706F8F\
7F708F70708F5F506F70708F7F708F0F0F0F1F1F205F5F6F0F0F0F0F000F0F0F0F
7F708F7F7F907F708F807F907F708F807F907F7F8F807F907F708F807F90807F8F000000000000\
000000000000000000000000000000000000000000000000000000000000807F7F
2F2F2F0000000000000000000000001010106060600000000000000000000000007F707F706F70\
7F707F706F707F707F706F7070707F202020000000000000100F10706F70706F70
706F70403F400000000000006F6F70706F706F6F700000000000001010100000002020205F5F60\
3030306F606F3030306F60702F202F1F1F1F000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C09F908090907F9F907F90907F9F907F908F7F90907F90907F90907F908F70\
90907F908F7090907F908F7090907F908F7090907F8F8F709090708F8F70909070
8F8F708F8F708F8F70404F3F30302F8F8F708F8F6F6F705F000F000F0F0F000000000000000000\
0000008080607F80600000000000000000000000000000000000007F8060000000
0000000000000000000000000000000000000000000000006F6F4F6F704F6F6F406F704F606F40\
606F4020201F006F40505F3F000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF1F1F1FCFCFCFCFCFCFCFCFCFCFCFCF5F5040000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000000000001F1F2F3F3F50\
4040603F3F5F40406040406040406040406040406F4F4F6F4F4F6F4F4F6F0F0F10
0000004040604F4F704F4F6F504F704F4F6F0F0F0F0F0F10505070504F70505070504F6F505070\
504F70505070504F70505070CFCFCF000000000000000000000000000000000000
0000000000000000000000000000000000000000002F2F306F6F8F6F6F8F6F6F8F6F60806F6F8F\
6F6F8F6F6F8F6F6F8F0000000000000000000000000000000000006F6F8F6F6F8F
70708F706F8F70708F706F8F707090706F8F707090706F8F7F709070708F7F709070708F7F7090\
70708F7F709070708F7F7F907F708F2020200F0F0F7F709030303F0F0F0F1F1F1F
5F50607F7090807F907F7090807F90807F90807F90807F90807F90CFCFCFCFCFCF000000000000\
0000000000000000000000000000000000000000000000000000000000000F0F0F
7F707F0000000000000000000000000000005F5F5F7F707F0000000000000000000000007F707F\
7F707F7F707F706F707F707F706F7000000000000000000060606F70707F706F70
70707F706F70000000000000706F70706F70706F70000000000000000000000000000000000000\
000000000000000000201F20303030706F706F6F70000000000000000000000000
000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F907F9F908090907F9F908090907F9F907F90907F9F907F\
90907F90907F90907090907F90907090907F908F7090907F908F709090708F8F70
9090708F8F709090708F8F7080806F8F8F702F2F2010100F0F0F0F0F0F0F000000000000000000\
0000008F8F6F808F6F30302F0000000000000000000000002F30207F8060000000
000000000000000000000000000000000000000000000000CFCFCF6F704F6F704F606F406F704F\
606F400F0F0F006F40606F40000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF2F2F2FCFCFCFC0C0C0CFCFCFCFCFCF605F4020201FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C000000000000000000000000010101F3F3F60\
4040603F3F5F4040604040604040603030504F4F6F40406F4F4F6F40405F000000
2020304F4F6F4F406F4F4F6F4F4F6F4F4F700000003F3F504F4F6F504F704F4F6F504F704F4F6F\
5050704F4F6F505070CFCFCFCFCFCF000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004F4F606F6F8F6060806F6F8F606080\
6F6F8F6060806F6F8F6F6F8F0000000000000000000000000000006F6080706F8F
6F6F8F706F8F6F6F8F706F8F706F8F70708F706F8F70708F706F8F70708F706F8F70708F70708F\
7F709070708F7F709070708F7F709070708F5F5F6F4040501F1F2000000F504F5F
7F708F807F907F708F807F907F708F807F907F708FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
000000000000000000000000000000000000000000000000000000000000000000
505050403F3F000000000000000000000000100F107F707F4040400000000000000000002F2F2F\
7F707F706F707F707F706F707F707F00000000000000000070707F706F70706F70
706F70706F700000000000001F1F20706F700F000F000000000000000000000000000000000000\
0000000000000000000000000000000000003F303F6F60706F6F70000000000000
000000000000000000000000000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C09F907F90907F9F907F908F7F90907F90907F\
90907F908F7090907F908F7090907F8F8F709090708F8F709090708F8F70909070
8F8F709090708F8F6F8F8F7080806F101010000F000F0F0F0F0F001F1F10000000000000000000\
0000005F5F4F808F6F80806F0000000000000000000000000000007F8060000000
000000000000000000000000000000000000000000000000CFCFCF6F704F606F406F6F4F606F40\
606F4F0000000000005F6040000000000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF30202FCFCFCFCFCFCFCFCFCFCFCFCF7F705F7F705FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF00000000000000000000000030304F404060\
3030502F2F3F10101F0F0F0F404060404060404F6F4F4F6F40406F4F4F6F000000
0F0F0F4F4F6F4F4F704F4F6F4F4F704F4F6F1010101F1F20504F704F4F6F505070504F70505070\
504F70505070504F70CFCFCFCFCFCFCFCFCF000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005F5F7F6F6F8F6F6F8F6F6F8F\
6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F0000000000000000000000000000006F6F8F
70708F6F6F8F707090706F8F70708F706F8F70709070708F70709070708F7F709070708F7F7090\
70708F7F70907F708F7F70907F708F7F7F907F70907F7F907F708F7070805F5060
807F907F7090807F907F7F90807F90CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000
1010107F707F0000000000000000000000000000000000007F707F000000000000000000101010\
706F7F7F707F70707F7F707F706F70000000000000000000706F707F707F706F70
70707F60606F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000404040706F70100F10
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF90907F9F907F90907F90907F\
90907F90907F90907F90907F908F7090907F90907090907F8F8F7090907F8F8F70
9090708F8F7090907080806F3F3F301010102F2F202F2F201010100F0F0F000000000000000000\
000000707F60808F6F7F80600000000000000000000000003F3F306F705F000000
000000000000000000000000000000000000000000000000CFCFCF6F6F4F6F704F606F4F6F6F4F\
606F40000000000000606F40000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF2F0000CFCFCFC0C0C0CFCFCFC0C0C07F705F7F7050CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000\
0F0F1000000000000000000040405F40406040406040406F4F4F6F40406F1F1F20
0000004F4F6F4F4F6F4F4F6F4F406F4F4F6F20203F0000004F4F6F4F4F704F4F6F504F704F4F6F\
504F704F4F6FCFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006060806F6F8F606080\
6F6F8F6F60806F6F8F6F60806F6F8F4F4F600000000000000000000000006F6F8F
6F6F8F706F8F6F6F8F70708F6F6F8F70708F706F8F70708F706F8F70708F706F8F707090706F8F\
70708F70708F7F709070708F7F709070708F7F709070708F7F70907F708F7F7F90
7F708F7F7F907F708FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000000000
0000007F70704040400000000000000000000000000000003030307F707F000000000000000000\
101010706F707F707F706F70303030000000000000000000706F7F706F70706F7F
706F70505050000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000005F505F
6F606F000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C090907F908F70\
90907F908F7090907F8F8F7090907F8F8F7090907F8F8F709090708F8F70909070
8F8F708F8F708F8F701F1F1F3F3F2F5F5F4F70705F80806F000F0010100F0F0F0F000000000000\
000000000000808F6F7F80602F2F200000000000000000000000007F8060000F00
000000000000000000000000000000000000000000000000000000CFCFCF606F406F6F4F606F40\
606F40000000606F40000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F705F7F705F000000
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
0000003F405F40406040406040406040406040406F4F4F6F4F4F6F4F4F6F00000F
1F1F204F4F6F4F4F704F4F6F4F4F704F4F6F30304F0F0F0F2F2F3F4F4F70504F704F4F70505070\
504F706F6F60CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000F0F0F6060806F6F8F\
6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F0000000000000000000000001F1F20
7070906F6F8F70708F6F6F8F707090706F8F707090706F8F70709070708F70709070708F7F7090\
70708F7F709070708F7F70907F708F7F70907F708F7F7F907F70907F7F907F708F
807F907F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000
000000000000706F70000000000000000000000000000000000000706F702F202F000000000000\
0000007F707F706F705F5F5F000000000000000000000000706F7070707F706F70
70707F706F70000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF9F907F\
90907F90907F90907090907F90907090907F8F8F7090907F8F8F709090708F8F70
9090708F8F708F907020201F80806F808F6F6F705F20201F0F0F0F0F0F0F000000000000000000\
0000000000003F4030808F6F808F6F0000000000000000001F1F10808F6F000000
000000000000000000000000000000000000000000000000000000CFCFCF6F704F606F406F704F\
606F400F0F0F5F6040000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF7F705F6F604F000000
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF00000000000000000000000000000000000030304F\
4040603F3F5F4040603F40604040603F4060404F6F40406F4F4F6F40406F20202F
0000004F4F6F40406F4F4F6F4F4F6F4F4F704F4F6F00000040406F4F4F704F4F6F504F704F4F6F\
504F706F605F6F6060CFCFCFCFCFCFC0C0C0000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000606080\
6F6F8F6060806F6F8F6F608F6F6F8F6F6080606080000000000000000000000000
6F6F8F706F8F6F6F8F706F8F6F6F8F70708F6F6F8F70708F706F8F707090706F8F70708F706F8F\
707090706F8F7F709070708F7F709070708F7F70907F708F7F70907F708F7F7090
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000
0000000000007F707F1010100000000000000000000000000000004F4F4F4F404F000000000000\
0000000000000F0F0F0F000F00000000000000000000000070707F706F70706F70
706F70706F7F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
90907F8F8F7090907F8F8F7090907F8F8F709090708F8F709090708F8F708F9070
8F8F708F90708F8F6F6F6F5F70705F4F4F3F000F000F0F0F0F0F000F0F0F1F1F1F000000000000\
000000000000505040707F607F80600000000000000000004F503F7F80603F3F2F
000000000000000000000000000000000000000000000000000000000000606F40606F4F606F40\
606F405F60405F6F40000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCF7F705F807F5F7F705F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
303F502F30403F3F5F30304F3F3F5F40406F3F3F5F4F4F6F40406F4F4F6F4F4F6F
00000040405F4F4F704F4F6F4F4F704F4F6F4F4F7010101F0F0F0F4F4F6F5050704F4F70505070\
6F605F706F606F605FCFCFCFCFCFCFCFCFCF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
60608F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F50506F000000000000000000
706F8F6F6F8F7070906F6F8F7070906F6F8F707090706F8F707090706F8F70709070708F707090\
70708F7F709070708F7F709070708F7F7F907F70907F7F907F708FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF0000000000003030300000000000000000000000000000000000001F1F1F000000000000\
000000000000000000000000000000000000000000000000706F707F707F706F70
70707F6F606F000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF90907F90907F8F8F7090907F8F907090907F8F8F7090907F8F8F70
9090708F8F708F90708F8F7020201F0F0F0F0F0F0F0F0F0F2F2F200F0F0F000000000000000000\
000000000000000000808F6F80806F2F2F200000000000002F2F20808F6F20201F
000000000000000000000000000000000000000000000000000000000000CFCFCF606F40606F4F\
606F40606F4F5F6040000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0000000C0C0C0CFCFCFC0C0C0CFCFCF7F705F7F705F
000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000\
0F0F0F0000000000000F0F1030304F3F4060404F6F404060404F6F40406F2F2F3F
0000004F4F6F40406F4F4F6F40406F4F4F6F4F4F6F4F4F700000004040604F4F6F504F704F4F6F\
6F6F606F605F6F6F606F605FCFCFCF000000000000000000000000000000000000
0000000000000000000000000000000000002F2F3F606080000000000000000000000000000000\
0000006060806F6F8F6060806F6F8F6F6F8F6F6F8F6F6F8F3F3F4F000000000000
5F5F706F6F8F6F6F8F706F8F6F6F8F706F8F6F6F8F70708F6F6F8F70708F706F8F707090706F8F\
70709070708F7F709070708F7F709070708F7F709070708FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000706F7F706F7070707F
706F705F505F000000000000000000000000000000000000000000000000000000000000000000\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF8F8F7090907F8F8F7090907F8F8F708F90708F8F708F9070
8F8F708F8F706F6F5F0F0F0F000F000F0F0F606F5080806F0F0F000F0F0F20201F000000000000\
0000000000000000001F1F1F808F6F1F201F000000000000505F407F80600F0F0F
000000000000000000000000000000000000000000000000000000000000C0C0C0606F4F606F40\
606F405F6040000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFCFCFCF807F5F3F3F2F
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
0000000F0F0F1F1F2F303F503F40604040603F40604F4F6F40406F4F4F6F40406F
00000030304F4F4F704F4F6F4F4F704F4F6F4F4F7040406F0000004F4F70504F704F4F6F505070\
6F6060706F606F605F706F60CFCFCF000000000000000000000000000000000000
0000000000000000000000000000002F2F3F40405F6060806F6F8F60608F000000000000000000\
0000005F5F7F606F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F20202F000000
50506F6F6F8F70708F6F6F8F707090706F8F707090706F8F70709070708F70709070708F7F7090\
70708F7F709070708F7F70907070907F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000706F7070707F706F70
70707F404040000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF90907090907F8F8F7090907F8F90709090708F8F70
9090708F8F708F90700F0F0F10100F808F6F8F8F70808F6F1010100F0F0F101010000000000000\
000000000000000000303F2F7F806F808F6F0F0F0F0000000000007F806F505F40
000000000000000000000000000000000000000000000000000000000000CFCFCF606F40606F4F\
606F40606F400000000000003F402F000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF000000C0C0C0CFCFCFCFCFCFCFCFCF7F705F7F705F
7F705FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
0000002020304040603F3F604040603F3F6040406040406F4F4F6F40406F4F4F6F
10101F10102040406F4F4F6F40406F4F4F7040406F4F4F700F0F0F1F1F204F4F6F4F4F70000000\
504F4F6F605F6F6F606F605F6F6F60000000000000000000000000000000000000
0000000000000000000000000000000F0F0F60608060608F606080606F8F606080000000000000\
0000000000006F6F8F60608F6F6F8F6060806F6F8F6F6F8F6F6F8F6F6F80000000
10101F6F6F8F6F6F8F6F6F8F6F6F8F70708F6F6F8F70708F6F6F8F7070906F6F8F707090706F8F\
707090706F8F70709070708F7F7090C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C08F90708F8F708F90708F8F708F9070
8F8F708F9070606F50808F6F808F6F8F8F7010100F0F0F0F0F0F000F0F0F20201F000000000000\
0000000000000000000000006F705F7F80600000000000001010107080603F4030
000000000000000000000000000000000000000000000000000000000000000000CFCFCF606F40\
606F405F6040000000000000506040000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF605F30CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
807F5F000000CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
0000000000003F40604040603F40601010203030504F4F6F40406F4F4F6F40406F
0000002020304040604F4F6F4F4F704F4F6F4F4F704F4F6F4F4F70000000504F704F4F70000000\
000000100F0F6F6060706F60000000000000000000000000000000000000000000
00000000000000000000000000000000000060608060608F606F8F601F206F6F8F60608F000000\
0000000000003030406F6F8F606F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F905F5F70
0F0F106F6F8F7070906F6F8F7070906F6F8F707090706F8F707090706F8F70709070708F707090\
70708F7F70907070907F7090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F202F5F4F5F90907F8F8F7090907F8F8F70
8F90708F8F708F90707F80600F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F000000000000\
0000000000000000000000007F8060808F6F0000000000000000004F503F7F8060
000000000000000000000000000000000000000000000000000000000000000000CFCFCF606F4F\
5F604010100F000000000000505F3F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000CFCFCFC0C0C0CFCFCFC0C0C080705F
000000000000C0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000003F4060404F6F40406F4F4F6F
1F1F2F0F0F0F40406F4F4F6F40406F4F4F6F404F6F4F4F703F3F5F0000003030504F4F70000000\
3F30306F605F6F6F606F605F000000000000000000000000000000000000000000
0000000000000000000000000000000000000F101060608F606080600000600000606F8F606080\
0000000000000000006060806F6F8F60608F6F6F8F60608F6F6F8F6F6F8F6F6F8F
3F3F506F6F8F6F6F8F6F6F8F6F6F8F706F8F6F6F8F7070906F6F8F70708F6F6F8F707090706F8F\
707090706F8F707090C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0000000000000504F5F504F5F8F90708F8F708F9070
8F8F708F8F7070705F0F0F0F000F000F0F0F0F0F000F0F0F000F002F2F202F3020000000000000\
0000000000000000000000002F2F207F80606F705F000000000000505F4F7F8060
000000000000000000000000000000000000000000000000000000000000000000CFCFCF5F6040\
5F6F400000000000003F402F50603F000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFCFCFCF7F705F
807F5F706F50CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000002F2F4040406F4F4F6F40406F
3F3F5F00000030304F40406F4F4F704F4F6F4F4F704F4F6F4F4F7000000F30304F6F605F706F60\
6F6060706F606F605F000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000040405F606F8F6000006000006000006F6F8F\
60608F0F0F0F0000000F0F0F606F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F
6F6F9050506F6F6F906F6F8F7070906F6F8F7070906F6F8F707090706F8F707090706F8F707090\
70708FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF0000000000000000000000000000005F50605F4F608F90708F8F70
8F90708F8F708F90700F0F0F0F0F0F0F0F0F0F0F0F60604F6F6F50606F50808F6F1F1F10000000\
0000000000000000000000003F3F307F8F6F5F604F0000000000000F0F0F708060
2F3020000000000000000000000000000000000000000000000000000000000000CFCFCF606F4F\
5F60400000000000005F6040000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
7F705F80705F000000CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000404F6F40406F404F6F
40406F0000000F0F104F4F6F40406F4F4F7040406F4F4F704F4F6F4F4F700000006F6F606F605F\
706F606F605F6F6F60000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005F5F7F60608F602F30600000600F0F\
606F8F0000000000000000006F6F8F6060806F6F8F606F8F6F6F8F60608F6F6F8F
6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F706F8F6F6F8F706F8F6F6F8F7070906F6F8F707090706F8F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF0000000000000000000000000000000000000F0F0F5F50605F4F5F8F9070
8F8F708F90707F7F60404F3F5F5F4F808F70808F6F808F6F808F6F808F6F80806F707F60000000\
000000000000000000000000000000404F307F8060505F400000001010107F8060
10100F000000000000000000000000000000000000000000000000000000000000CFCFCF5F6040\
4F5030000000000000505F3F000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
00000010100F000000CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
00000000000000000000000000000000000000000010101F40406F4F4F6F40406F
4F4F6F00000F101010404F6F4F4F70404F6F4F4F704F4F6F4F4F704F4F6F3F3F506F6F60706F60\
6F6060706F606F6060000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000010101F00000F000000606F8F605070606080\
606080606F8F00000000000010101F6F6F8F6F6F8F6F6F8F6F6F8F6F6F906F6F8F
6F6F8F6F6F8F6F70906F6F8F7070906F6F8F7070906F6F8F7070906F6F8F707090706F8FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
5F505F0000000000000000000000000000000000000000003F303F5F50605F5060
8F90708F8F708F90705F5F4F8F9070808F6F8F9070808F6F808F70808F6F808F704F5040000000\
000000000000000000000000000000404F3F7F8060505F4F0000002F3020708060
000000000000000000000000000000000000000000000000000000000000000000CFCFCF606F40\
000000000000000000506040000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF605F30CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
7F705F80705F706F50CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000303F5040406F404F6F
40406F404F6F000000404F6F40406F4F4F6F40406F4F4F7040406F4F4F706F6060706F606F605F\
6F6F606F605F000000000000000000000000000000000000000000000000000000
0000000000000000000000000000001010105F5F7F3F3F505F5F7F000000000000303040606080\
6060801F1F200F0F0F0000002F2F3060608F6F6F8F60608F6F6F8F606F8F6F6F8F
606F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F6F6F8F7070906F6F8F706F8F6F6F8FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F4F5F\
5F50600000000000000000000000000000000000000000000000003F30405F5060
4F40508F8F70808F6F8F8F70808F6F808F70808F6F808F70808F6F808F6F808F6F808F6F000000\
000000000000000000000000000000000000303F2F708060000000000000707F60
404F30000000000000000000303F2F000000000000000000000000000000000000CFCFCF5F6040\
0000000000005060404F5F3F000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF80705F807F5F000000CFCFCFCFCFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000101F20404F6F40406F
4F4F7040406F10101F00000F4F4F70404F6F4F4F70404F6F4F4F704F4F6F706F606F6060706F60\
6F6F60706F60000000000000000000000000000000000000000000000000000000
00000000000000000000000000000030304F60608F5F5F7F101010202F30000000000000000000\
000F0F0F0F0F0000000000000000006F6F8F606F8F6F6F8F6F6F8F6F6F8F6F6F8F
6F6F906F6F8F6F6F906F6F8F7070906F6F8F7070906F6F8F7070906F6F8FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5060605060\
5F5060605060000000000000000000000000000000000000000000000000000000
0000000000008F9070808F6F8F9070808F6F808F70808F6F808F70808F6F808F6F5F604F000000\
0000000000000000000000000000000000002F30207F806F000000000000708060
30302F000000101F106F705F4F5F40000000000000000000000000000000404F3F607F5F2F3020\
00000000000050603F0F0F0F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF807F5F7F705F7F6F50CFCFCFCFCFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000303F5F40406F
40406F404F6F00000020203040406F4F4F7040406F4F4F6F40406F4F4F706F605F706F606F6060\
706F60000000000000000000000000000000000000000000000000000000000000
0000000000000000001F202F5F6080404F605F60803F405F000000000000000000000000000000\
000000000000000000000000000000000000606F8F60608F6F6F8F60608F6F6F8F
606F8F6F6F8F606F8F6F6F8F6F6F8F6F6F906F6F8F6F6F8F6F6F8F706F6F706F60CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6050605F505F5F50605F505F\
6050605F4F5F000000000000000000000000000000000000000000000000000000
000000000000000000808F70808F6F808F70808F6F808F6F808F6F808F6F7F806F808F6F000000\
0000000000000000000000000000000000002F302000000000000000000020201F
303F2F0000006F7F5F708060303F2F0000000000000000000000006070506070506F7F5F000000\
00000000000050603F000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F402FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF504F3F706F506F604FCFCFCFCFCFCF000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000001F1F2040406F
404F6F40406F404F6F0000004F4F7040406F4F4F70404F6F4F4F706F6F60706F606F6F60706F60\
6F6060000000000000000000000000000000000000000000000000000000000000
00000000000000000040405F60608F5F6080101F1F303040000000000000000000000000000000\
0000000000000000000000000000000000004F4F606F6F8F606F8F6F6F906F6F8F
6F6F8F6F6F8F6F6F906F6F8F6F6F906F6F8F7070906F6F8F706F6F706F6F7F6F6F706F6FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6050605F50606050605F5060605060\
5F50606050605F5060100F10000000000000000000000000000000000000000000
0000000000000000000000008F8F70808F6F808F70808F6F808F70808F6F808F6F7F8F6F000000\
000000000000000000000000000000000000000000000000000000000000101F10
000000707F60708060707F5F0000000000000000000000000F0F0F6F7F5F6F7F5F607F5F000000\
00000000000050603F000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF7F705F807F5F20201F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000404F6F
40406F40406F40406F0F0F101F1F2F404F6F40406F4F4F7040406F706F606F6060706F606F605F\
000000000000000000000000000000000000000000000000000000000000000000
0000002F2F3F5F60805050705F5F80404F60000000000000000000000000000000000000000000\
00000000000000000000000000000000000050507060608F606F8F60608F6F6F8F
60608F6F6F8F606F8F6F6F8F606F8F6F6F8F6F6F8F40403F3030302F2F2F5F5050706F6F706F60\
706F6FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF5F505F6050605F505F6050605F505F6050605F505F\
5F50605F505F6050605F4F5F1F1F1F000000000000000000000000000000000000
000000000000000000000000808F6F808F6F808F6F808F6F7F8F6F808F6F7F8F6F7F806F6F705F\
000000000000000000000000000000000000000000000000000000000000000000
5F6F4F70201F6F7F5F7080600000000000000000000000006070506F7F5F6070506F7F5F000000\
00000050603F50603F000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF605F3F000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF807F5F80705F0F0F0F000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000000003F3F5F
404F6F40406F404F6F40406F000000404F6F4F4F7040406F4F4F706F6F60706F606F6F60000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F4F506F60608F5F60801F202F000000000000000000404F60000000000000000000000000\
0000000000000000000000000000001F1F202F2F3F6F6F8F606F8F6F6F8F606F8F
6F6F906F6F8F6F6F906F6F8F6F6F906F6F8F0000000000000F0F0F0F0F0F2F2F2F706F6F7F6F6F\
706F6F706F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6050606050605F5060605F605F50606050605F5060605060\
5F50606050605F50606050605F50601F1F1F000000000000000000000000000000
000000000000000000CFCFCFCFCFCF808F6F808F70808F6F808F70808F6F808F6F7F8F6F6F705F\
000000000000000000000000000000000000000000000000000000000000303F2F
703F2F7000007080601010100000000000000000006F7F5F6F7F5F6F7F5F6F7F5F607050000000\
000000506040000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF7F705F0000000F0000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
40406F404F6F40406F404F6F000000404F6F40406F404F6FCFCFCFCFCFCF2F2F2F000000000000\
00000000000000000000000000000000000000000000000000000010101F4F5070
5F5F805F60805F5F804F506F00000000000000000040405F40405F60608F1F1F2F000000000000\
000000000000000000000000000000000000606F8F60608F6F6F8F60608F6F6F8F
606F8F6F6F8F60608F6F6F8F606F8F706F6F6F6060706F6F706F60706F6F706F60706F6F706F60\
706F6F706F60706F6FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C06050605F50606050605F505F5F505F000000000000000000201F205F5060\
6050605F505F5F50605F50606050605F4F5F000000000000000000000000000000
000000000000C0C0C0CFCFCFCFCFCFCFCFCF7F8F6F808F6F7F8F6F808F6F7F806F7F8F6F5F604F\
505F4F000000000000000000000000000000000000000000000000000000708060
6F00007040306F7F5F000000000000000000607F5F6F7F5F60705F6F7F5F607050607F5F000000\
506040505F3F000000000000CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F5F3FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF80705F7F705F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F0F40406F404F7040406F404F7040406F4F4F70404F6FCFCFCFCFCFCFCFCFCF000000000000\
0000000000000000000000000000000000001F1F205050702F2F3F5F60805F6080
5F608F303F4F0F0F0F00000000000000000000000040405F60608F202F302F2F3F000000000000\
000000000000000000000000000000404F5F5F5F7F606F8F60608F6F6F90606F8F
6F6F8F606F8F6F6F906F6F8F6F6F90706F6F7F706F706F6F706F6F706F6F7F6F6F706F6F706F6F\
706F6F7F6F6F706F6F706F6F706F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF605060605F60605060605F60605060605060000000000000000000000000000000100F10\
5F50606050605F50606050605F5060504F50504F5F000000000000000000000000
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808F6F808F6F7F8F6F808F6F7F8F6F7F8F6F\
4F503F000000000000000000000000000000000000000000000000000000707F5F
7000006F7F5F0000000000000000005F6F506F7F5F6F7F5F6F7F5F60705F6F7F5F607050000000\
50603F506040000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF0F0F00000000C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0807F5F80705F000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F3F5F40406F404F6F40406F404F6F40406F404F70C0C0C0CFCFCFC0C0C0000000000000\
0000000000000000000000000F0F0F2F2F3F5F5F80505F80505F7F3F405F5F6080
2F2F3F000000000000000000000000000000000000404F605F60802F2F40000000000000000000\
0000000000000000000000000000002F2F30606F8F60608F606F8F60608F6F6F8F
60608F6F6F8F606F8F6F6F8F706F60706F6F706F6F706F6F706F60706F6F706F6F706F6F706F60\
706F6F706F60706F6F706F60706F6F706F60706F6FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
5F505F6050606050606050605F505F605060000000000000000000000000000000000000000000\
0000000000000F0F0F302F30000000000000000000000000000000000000000000
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F806F7F8F6F7F806F7F8F6F7F8060\
404F3F3F3F30000000000000000000000000000000000000000000404F3F707F5F
6F7F5F0000000000000000006F7F5F6F7F5F607F5F6F7F5F607050607F5F607050000000C0C0C0\
506040000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000006F5F3FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF80705F807F605F5040000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
00000020203F404F6F40406F404F7040406F404F7040406FCFCFCFCFCFCFCFCFCFCFCFCF000000\
00000000000010101F3F3F50505F7F5F6080505F80505070303F500F0F0F000000
0000000000000000000000000000000000005F60805F608060608F303040000000000000000000\
000000000000000000000000000000606F8F60608F606F8F606F8F6F6F8F60608F
6F6F90606F8F6F6F906F6060706F60706F60706F6F706F6F7F706F706F6F706F6F706F6F60605F\
3F3F3F0000006F60607F6F6F706F6F706F6F706F6F7F6F6F706F6FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF605F60605060
605F601F1F1F5F5060605060605F603F303F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF808F6F7F8F6F808F6F7F806F7F8F6F\
7F806030302F0000000000000000000000000000000000000000007080606F7F5F
3F40300000000000002F30206F7F606F7F5F6F7F5F607F5F6F7F5F60705F000000000000CFCFCF\
50603F000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF6F5F3F000000000000CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF504F3F7F705F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F3F6040406F40406F404F6F40406F404F6FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
202030404F60505F80505F805F5F8040406020203000000F000000000000000000
0000000000000000000000000000000F10105F5F805F608F3F3F50000000000000000000000000\
000000000000000000000000000000505F7F606F8F60608F606F8F60608F606F8F
60608F6F6F8F6F60606F6F606F6060000000000000000000000000000000000000000000000000\
0000000000000000006F6060706F6F706F60706F6F706F60706F6F706060000000
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF605060605F60605060605F60
504F500F0F0F5F5060605060605060000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF7F8F6F7F80607F8F6F7F8060\
7F8F6F303F2F505F400000000000000000000000000000000000006F7F5F5F604F
0000000000003F40306F7F5F607F5F6F7F5F607050607F5F607050000000000000000000CFCFCF\
50603F000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F5F3F6F603FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF807F6080705F000000000000000000000000000000000000000000\
0000000F0F0F000000000000000000000000000000000000000000000000000000
00000000000040406F40406F404F6F40406F404F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
505F7F101010000000000000000000000000000000000000000000000000000000
00000000000000000000000020202F5F60805F608F5F60802F3040000000000000000000000000\
0000000000000000000000002F2F3F404F6060608F606F8F60608F606F90606F8F
6F6F8F6F60601F1F1F6F6060706F6F302F2F000000000000000000000000000000000000000000\
00000000000000000030302F706F6F7F6F6F706F6F706F6F706F6F7F6F6F000000
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F5F60605F60605F606050606F5F605F505F
000000605060605F60605060403F40000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F8F6F808F6F7F806F7F8F6F\
7F80607F8F6F6F7F5F1F1F10000000000000000000000000000000000000000000
0000000000006F80606F7F5F6F7F5F607F5F6F7F5F607F5F607F5F000000000000000000CFCFCF\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0201F1F000000000000000000000000000000000000000000\
0F0F103F3F602F30502F304F2F304F3F40603F406F1F2030000000000000000000
0000000000003F3F6040406F3F406F40406F40406FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F30405F60803F405F000000000000000000000000000000000000\
0000000000000000000000002F2F3F5F608F606F8F5F608F606F8F60608F606F8F
60608F2F2F2F605F5F0000006F60600F0F0F000000000000000000000000000000000000000000\
000000000000000000403F3F706F6F706F60706F6F706F60706F6F706F60000000
000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF0F0F0F605F60605060605F60605060605F60605060000000
605060605F60605060605F60000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F80607F8F6F708060\
7F806F7080607F806020201F4F50400000000000000000000000000000000F0F0F
1F201F6F7F5F6F7F5F6F7F5F60705F6F7F5F607050607F5F0000000000000000000000004F5F3F\
000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF2F201F2F201F6F5F3F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF807F5F0000000000000000000000000000001F1F2F303F5F\
2F2F4F3F406F3F40603F406F3F40603F406F3F406040406F3F3F5F1F1F20202F3F
1F20300000003F406F3F406F404F7040406F404F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000002020305F5F805F608F5F60802F2F3F000000000000000000000000000000000000\
0000000000000000003F3F50303040606F8F60608F606F8F60608F606F8F60608F
1F1F1F404040706F6F000000706F6F101010000000000000000000000000000000000000000000\
0000000000000000005F5F50706F6F706F6F706F6F7F6F6F706F6F000000000000
0000000000000000000000007F6F6F706F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
605F606F5F604F404F201F20605F606F5F606050606F5F6F605F60605F60000000
6F5F6F605060605F60000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F8F6F7F80607F8F6F\
7080607F8F6F7080607F8F6F5F604F0F0F0F00000000000000000010100F6F7F5F
6F80606F7F5F6F7F5F607F5F6F7F5F607F5F607F5F000000000000000000000000000000506040\
0000000F0F0FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F603F000000000000
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF00000000000000000000000000000010101F303F60\
3F4060303F603F40603F3F603F406F3F3F603F406F3F40603F406F3F40603F406F
3F406F3F406F3F406040406F3F406F40406F3F406FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000F10104F507F5F6080404F60000000000000000000000000000000000000000000000000\
0000000000000000003F3F5060608F5F608060608F5F608F606F8F5F608F606F8F
3030301F1F1F6F60600000003F3F30000000000000000000000000000000000000000000000000\
000000000000000000605F5F706F6F706F60706F6F706F6F706F6F000000000000
0000000000000000000F0F0F706F60706F6F706060706F6FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F5F60605F60\
6F5F605040500F0F0F605060605F60605060605F60605060605F60000000605F60
605060605F60000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C07F806F708060\
7F806F7080607080607080607080602F30200F0F0F3F40307080606F7F5F6F7F5F
6F7F5F6F7F5F607F5F6F7F5F60705F607F5F0000000000000000000000000000000000004F5F3F\
000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
605030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000000000000000000000000F1010\
303F5F3F406F3F40603F406F3F40603F406F3F406F3F406F3F406040406F3F406F
40406F3F406F404F6F3F406F404F6F3F406F404F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000002F30405F608F5F60802F3040000000000000000000000000000000000000000000000000\
000000000000000000404F605F608F606F8F60608F606F8F60608F606F8F1F1F1F
1F1F1F000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000706F6F706F6F7F6F6F706F6F706F6F000000000000000000
0000000000000000006F6060706F6F706F6F7F6F6F706F6F706F6F706F6F706F6FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF605F606F5F6F605F606F5F60\
605F60000000605F606F5F60605F606F5F6F605F606F5F60000000605F60605F60
605F60605060000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF50405F7F80607F8F6F\
7080607F8F6F7080607080607080605F6F507080607080606F7F5F7080606F7F5F
6F7F606F7F5F6F7F5F607F5F5F6F500000000000000000000000000000000000000000004F5F3F\
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6F5F3F10100F
000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000\
00000000000010101F303F603F406F3F40603F406F3F3F603F406F3F40603F406F
3F40603F406F3F406F40406F3F406F40406F3F406FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0F0F10505F7F505F805F60800F0F10000000000000000000000000000000000000000000000000\
000000000000505F7F5F60805F608F5F608F60608F5F608F60608F5F608F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000004F4040706F6F706F6F706F6F706F60000000000000000000
000000000000000000706F6F706060706F6F706060706F6F706060706F6F706060706F6F706060\
706F6FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF605F606F5F60605F606F5F60605F606F5F60605060\
0F0F0F605F606F5F60605060605F60605060605F60000000605F60605060605F60
6050605F505F5F505F403F400F0F0F000000000000000000000000000000000000000000000000\
000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0504050000000100F10708060\
7080607080607080607080607080606F7F607080606F7F5F6F7F606F7F5F6F7F5F
607F5F6F7F5F607F5F6F7F5F000000000000000000000000000000000000000000000000405F30\
3F5030C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000\
0000000000000000000000002F2F40303F5F3F40603F406F3F406F3F406F3F406F
40406F3F406F40406F3F406F404F6F3F406F404F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000000000000303040505F80
5F608F3F405F0F0F10000000000000000000000000000000000000000000000000000000000000\
202F300000005F5F80606F8F5F608F60608F5F608F606F8F60608F000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000005F5050706F6F706F6F706F6F706F6F000000000000000000
000000000000000000706F6F7F6F6F706F6F706F6F706F6F706F6F706F6F706F6F706F6F706F6F\
706F6F706F6F706060706F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF6F5F606F606F6F5F606F5F6F605F606F5F6F6F5F606F5F6F4F404F0F0F0F\
3030306F5F6F605F606F5F6F605F606F5F6F0000006F5F6F605F606F5F60605060
6050605F505F6050605F505F3F303F000000000000000000000000000000000000000000000000\
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF50405050405F2F202F0000000000007F8F6F\
70806070806F7080607080607080607080606F7F5F7080606F7F5F6F80606F7F5F
6F7F60607F5F4F5F400000000000000000000000000000000F0F0F0000000000004F5F3F4F5F3F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F603F
6F5F3F000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000F0F3F406F3F3F603F406F
3F40603F406F3F40603F406F3F406F40406F3F406FCFCFCFCFCFCFCFCFCFC0C0C0000000000000\
0000000000000000000000000000000000000000000F0F10404060505F804F5070
10101F00000F0000000000000000000000000000000000000000000000000000000000001F1F2F\
5F60805F60805F608F5F60805F608F5F60805F608F5F608F60608FC0C0C0000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000706F6F706F6F706F6F605F5F000000000000000000
0000000000000F0F0F706F6F706F6F706F6F706060706F6F706060706F6F706060706F6F706060\
706F6F706060706F6F706060706F6F706060706F6FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF605F606F5F60
6F5F606F5F60605F606F5F60605F606F5F60605F606F5F60605F606F5F605F505F1010102F202F\
6F5F606050606F5F60605F605F505F000000605F60605060605F606050605F4F5F
504F50100F10000000000000000000000000000000000000000000000000000000000000CFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF4F4050504050504050000000000000CFCFCFCFCFCF\
7080607080607080606F80607080606F7F5F6F80606F7F5F6F7F60607F5F6F7F5F
607F5F202F20000000000000000000000000101F103F4030000F000000000000004F5F3F405030\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F5F3F
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000\
0000000000000000000000000000000000000000000000000000000F0F0F3F406F
3F406F3F406F40406F3F406F40406F3F406F404F70CFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000202030404F6F506080505F80404F600F0F10
0000000000000000000000000000000000000000000000000000001F1F2F4F506F5F60805F608F\
5F60805F608F5F608F5F608F5F608F606F8F5F608F606F8FCFCFCFCFCFCFCFCFCF
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F706F6F7F6F6F706F6F202020000000000000000000
0000000000005F5050706F6F706F6F706F6F7F6F6F706F6F706F6F706F6F706F6F706F6F706F6F\
706F6F706F6F706F6F706F6F70606F706F6F706F6F706F6F706060706F6F70606F
706F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F606F6F5F606F606F6F5F606F606F6F5F60
6F606F6F5F606F606F6F5F606F5F6F6F5F606F606F6F5F606F5F6F4F404F0F0F0F403F406F5F6F\
605F606F5F6F605F60302F300000006F5F6F605F606F5F6F605F606050605F505F
6050605F4F5F6050601F1F1F000000000000000000000000000000000000000000CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF50405050405F504050403040000000CFCFCFCFCFCFCFCFCF\
70806070806F7080607080606F7F607080606F7F5F6F80606F7F5F6F80606F7F5F
4F5F40000000000000000000000000000000607F5F5F70500F0F0F0000000000004F5F3FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
0000006F5F3F000000CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000003F406F3F406F3F406F3F406F3F406FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000\
0000000000000F0F102F2F404F507F505F80505F804F5F7F303F50000000000000
0000000000000000000000000000000000000000000F0F10101F200F0F0F00000F5F608F505F80\
5F608F5F60805F608F5F60805F608F5F608F5F608F5F6080CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000706F6F706F6F706F6F000000000000000000000000
000000000000706F6F706F6F706060605F5F0F0F0F0000003F3F3F706F6F706060706F6F706060\
706F6F706060706F6F706060706F6F706060706F6F706060706F6F706060706F6F
6F6060706F6F706060706F6F6F606070606F4F4040CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
6F60606F5F606F606F6F5F606F60606F5F606F606F6F5F606F5F606F5F606F606F
605F606F5F606F5F606F5F6F605F606F5F60605F606F5F6050404F0000003030306F5F60605F60\
6F5F60302F2F0F0F0F0000006F5F60605F606F5F606050605F505F5F505F000000
0000000F0F0F101010000000000000000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF50405050405050405050405F000000CFCFCFC0C0C0CFCFCFC0C0C0\
7080607080607080606F7F5F6F80606F7F5F6F8060607F5F6F7F5F607F5F304030
0000000000000000000000002F302F6070505F70505F70500000000000000000004F5F3FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
706040000000000000CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000003F406F3F406F40406F3F406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
0000003F40604F5F7F505F80505F7F404F601F20300F0F0F000000000000000000
0000000000000000000000000000000000000F0F100F0F100F0F0F0F0F0F0F0F0F00000F5F608F\
5F608F5F608F5F608F5F608F5F608F5F608F5F608FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000706F6F7F6F6F706F6F000000000000000000000000
000000000000706F6F706F6F7F6F6F000000000000000000000000706F6F706F6F706F6F706F6F\
706F6F706F6F70606F706F6F706F6F706F6F70606F706F6F70606F706F6F70606F
706F6F70606F706F6F706060605F5F000000000000706060706F6F70606F706F6F706060706F6F\
70606F70606F6F6060706F6F70606070606FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF6F606070606F6F606070606F000000000000000000000000000000302F30\
6F60606F606F6F5F6070606F6F5F606F606F6F5F606F606F6F5F606F606F6F5F60
6F606F6F5F606F606F6F5F606F606F6F5F606F5F6F1010100F0F0F6F5F606F5F6F605F606F5F6F\
000000000000403F406F5F6F605F606F5F6F605F60605F6060505F6050605F505F
605F605F505F1F1F1F000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF5040505F4F5F50405050405F504050CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF7080606F80607080606F7F606F80606F7F5F6F80606F7F5F6F7F5F000000
00000000000000000020201F607F5F5F70505F705F5F7050000000000000000000405030CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF0F0F0F6F5F3FCFCFCFCFCFCFCFCFCFC0C0C0000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003F406F3F406FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000000000\
00000030304F303F501F1F200F0F0F000000000000000000000000000000000000
00000000000000000000000000000010101F000F0F0F0F0F00000F0F0F0F0000000F0F0F101F20\
5F608F505F805F608F5F60805F608F5F60805F608FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000000000000000000000000000\
000000000000000000000000000000706F6F50504F000000000000000000000000
000000000000706F6F706F6F706F6F000000000000006F6F0000002F2F2F706060706F6F70606F\
706F6F706060706F6F706060706F6F706060706F6F706060706F6F706060706F6F
7060606F5F606F5F600F000F000000100F0F6F606070606F6F606070606F6F606070606F6F6060\
70606F6F606070606F5F4F5070606F6F606070606F6F606070606F6F606070606F
6F606070606F6F5F606F606F6F606070606F6F5F600000000000000000000000000000000F0F0F\
6F606F6F5F606F60606F5F606F606F6F5F606F60606F5F606F606F6F5F606F5F60
6F5F606F606F605F606F5F606F5F601010100000002F2020605F606F5F6F605F604F404F000000\
000000605F606F5F60605F606F5F6F6050606050605F505F0F0F0F1F101F3F303F
4F404F1F1F1F000000000000000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF50405F50405000000050405050405F504050CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF6F7F607080606F7F5F6F80606F7F5F6F7F5F607F5F506F4F000000000000
0000001F201F5F6F5060705F5F70505F70505F6F50000000000000000000000000CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF0F0000000000000000CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003F406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000F0F10000F0F0F0F0F00000F101010000F0F0F0F0F0F10105F608F0F0F0F0F0F0F\
3F40605F608F5F608F5F608F5F608F5F608F706060CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
000000000000000000000000000000000000706F6F000000000000000000000000
0000000000007F6F6F706F6F403F3F000000006F6F006F6F000000706F6F605F5F706F6F706F6F\
706F6F706F6F706F6F706F6F70606F706F6F706F6F706F6F70606F706F6F3F3F3F
00000000000000000070606F706F6F70606F706F6F70606F706F6F706060706F6F70606F706F6F\
7060606F5F605F50500000006F6060706F6F70606F70606F6F606070606F6F6060
70606F6F606070606F6F606070606F6F606070606F000000000000000000000000000000000000\
6F5F6070606F6F60606F606F6F5F6070606F6F5F606F606F6F5F606F606F6F5F60
6F606F6F5F606F606F60505F00000000000060505F6F5F606F5F6F60505F000000201F206F5F6F\
605F606F5F6F6F5F606F5F6F605F60605F606050606050605F4F5F2F2F2F1F101F
0F0F0F000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
5F4F5F504050000000504050504F5FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF7080606F80607080606F7F5F6F80606F7F5F6F7F603F4F30000000000000
60705F607050607F5F5F705060705F5F70505F7050000000000000000000000000CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C070603F00000070603FC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000\
0000000000000F0F0F303F5F3F4060202F40101020000000000000000000000000
000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000101F2000000F0F0F0F0000000F0F0F000F0F5060800000000F0F0F505F800F1010000000\
3F4060505F805F608F505F805F608F5F60806F5F5F6F605FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000302F2F000000000000000000000000
0000000000001F1010706F6F1F1010000000006F6F006060000000706F6F3F3F3F000000000000\
0000000F0F0F000000202020000000403F3F1F1010000000101010000000202020
000000706F6F706060706F6F6F606070606F706060706F6F6F606070606F6F606070606F6F6060\
6F6060000000504F4F6F606070606F6F606070606F6F60606F6060605F606F5F60
6F5F6070606F6F606070606F6F5F606F606F6F6060000000000000000000000000000000000000\
6F606F6F5F606F606F6F5F606F606F3F303F0000000000006F606F6F5F606F606F
6F5F604F4040000000000000605F606F5F606F5F606F5F6F000000000000605F606F5F6F605F60\
6F5F60605F606F5F6F1010106050605F505F605F605F505F6050605F505F605060
000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F4F5F
50405000000000000050405FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF6F7F606F80606F7F5F6F7F60607F5F6F7F5F5F7050607F5F607F5F607F5F
5F705060705F5F70505F705F5F6F505F7050202F20000000000000101F10000000CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF0F0F0F000000000000CFCFCFCFCFCFCFCFCF000000000000000000000000000F0F\
10101F3F4060303F603F406F303F603F406F303F603F406F2F3050000000000000
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000101F203F4060
506080505F80404F6F00000F0F0F10505F800F0F0F1F1F205F608F0F0F0F0F0F0F505F800F0F0F\
000F0F5F608F5060805F608F5F608F5F608F605F50605F50706060706F60CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005F50507F6F6F6F60600000002F2F2F706F6F706F6F706F6F706F6F7F6F6F\
3F3030605F5F0F0F0F3030300000001F1F1F4F404F303030706F6F706F6F70606F
706F6F70606F706F6F70606F706F6F70606F706F6F70606F706F6F70606F706F6F70606F101010\
000000706F6F706060706F6F70606F70606F6F60601F1F1F000000000000000000
0000006F606F70606F6F606070606F6F606F70606F000000000000000000000000000000000000\
0F0F0F70606F6F5F6070606F4F404F0000000000000000002020206F606F403F40
0000001F10106F606F6F5F60202020201F200000000000002F2F2F6F5F606F606F6F5F606F606F\
6F5F606F5F6F000000000000000000000000000000000000000000000000000000
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F4F5F504050
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF6F7F5F6F80606F7F5F6F8060607F5F6F7F60607F5F607F5F60705F
607F5F5F705060705F5F70505F705F5F70500000000000003F4F3F000000CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF0000000F0F0050402F000000C0C0C0CFCFCF000000000000000000000000000000\
304060303F60304060303F60304060303F60304060303F603F406F303F603F406F
1F1F2F000000000000000000000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000\
0000000000000000000000000000000000000F0F101F20303F405F4F5F7F505F80
4F5F80505F804F5F8030304F00000F1F1F2F505F800F0F10303F50505F800000000F0F0F505F80\
0F0F0F0000005F608F505F805F608F0000000F0F0F000000706F606F605F706060
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004F4040706F6F706F6F70606F706F6F706F6F706F6F70606F706F6F706F6F\
706F6F202020706F6F70606F706F6F706060706F6F70606F706F6F706060706F6F
70606F706F6F6F6060706F6F706060706F6F6F606070606F706060605F5F0F000F0000006F5F60\
70606F6F606070606F6F606070606F60505F000000000000000000000000000000
0000000000006F606070606F6F606070606F6F5F60000000000000000000000000000000000000\
0000006F5F606F606F6F5F60000000000000005F600030300000006F5F60403F40
6F5F606F606F6F5F60000000403F403F3F3F605F606F5F6F6F5F606F5F6F605F606F5F6F605F60\
605F60000000000000000000000000000000000000000000000000000000CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF5040505F4F5F504050504050
000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C06F80606F7F5F6F7F60607F5F607F5F607F5F607F5F5F7050607F5F
5F705060705F5F70505F70505F6F500F100F0000000000000F100F000000C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF5F50300000006F5F3FCFCFCFCFCFCF000000000000000000000000000000\
303F603F4060303F60303F6000000020203F303F603F406F303F603F406F304060
3F406F304060000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000505F804F5F80505F804F5F80505F80505F80
505F80505F805060804F5F7F0F0F0F00000F3F405F505F800F0F0F202F3F5F608F0F0F0F101F20\
4F507F0F0F0F2F30405F608F506080706F60706060706F606F605F706F60706060
706F606F6060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000706F6F706F6F7F6F6F706F6F706F6F706F6F7F6F6F706F6F706F6F\
706F6F0000000000002F20201F1F1F706F6F706F6F6F5F60504F50706F6F706F6F
6F5F6070606F706F6F4F404F706F6F605F5F2F20203F3030000000201F1F70606F70606F706F6F\
70606F706F6F70606F706F6F201F1F000000000000000000000000000000000000
0000000000006F5F606F606F70606F6F606070606F100F0F000000000000000000000000000000\
00000070606F6F606F70606F000000000000006060005F6000000070606F6F5F60
6F606F6F5F606F606F6F5F606F606F6F5F606F606F6F5F606F606F6F5F606F606F6F5F60605F60\
000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F4F5F5F4F5F5040505F4F5F000000
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF6F7F606F8060607F5F6F8060607F5F607F5F607F5F607F5F60705F
607F5F5F70505F705F5F70505F70500000000000004F604F000000000000CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C00F0F0F1F100F000000000000CFCFCFC0C0C000000000000000000000000F\
303F60303F60304060202F40000000000F10000000000000304060303F603F406F
303F603F406F303F60303F600F0F1F000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
0000000000000000000000004F5F804F5F7F505F804F5F7F505F804F5F80505F80
4F5F7F505F804F5F80505F804F5F80202030202F3F5060804F507F1F1F20505F80404F6F10101F\
303F4F404F6F5F608F505F805F608F0000000F0F0F0000000F0F0F1F1F1F706F60
6F605F7060606F605FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000000000
000000000000000000000000000000706F6F706F6F706F6F70606F706F6F706F6F706F6F70606F\
706F6F70606F403F403F3030000000000000000000000000000000000000000000
0000000000000000000000000000002020200F0F0F5F50506F606070606F706060706F6F6F6060\
70606F6F606070606F201F20000000000000000000000000000000000000000000
0000000000003F303070606F6F606070606F6F60603F303F000000000000000000000000000000\
0000006F5F606F606F6F5F603F303F000000000F0F0000006F5F606F5F606F606F
6F5F606F606F6F5F606F606F6F5F606F5F6F6F5F606F606F605F606F5F6F605F600F0F0F000000\
000000000000000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C05F4F5F5040505F4F5F5040505F4F5F101010000000
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C06F8060607F5F6F7F5F607F5F607F5F60705F607F5F5F705060705F
5F70505F70505F70505F70503F4F3F000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF70603F60503F000000000000CFCFCFCFCFCF0000000000000000001F1F2F\
303F60304060303F603F4060000000004060003F6000203F000000303F5F303F60
3F406F3040603F406F3040603F406F3F406F202F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
0000000000000000000F0F0F4F5F80505F804F5F80505F804F5F80505F804F5F80
505F80505F80505F80505F8050608F505F80506080505F8050608F505F8050608F505F805F608F\
505F805F608F505F805F608F50608F0000000F0000706F607060606F6060706060
706F60706060706F606F6060706F60CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000
000000000000000000000000000000000000706F6F706F6F7F6F6F706F6F706F6F706F6F7F6F6F\
706F6F706F6F706F6F7F6F6F706F6F706F6F504F4F706F6F706F6F706F6F6F5F60
706F6F706F6F605F5F70606F706F6F70606F706F6F70606F706F6F70606F706F6F70606F706F6F\
6F6060101010000000000000000000000000000000000000000000000000000000
000000000000000000504F4F70606F6F606F70606F6F6060000000000000000000000000000000\
0000000000006F606070606F6F606F2F2F2F4F404F6F5F606F606F6F606F6F5F60
70606F6F5F6F6F606F6F5F606F606F6F5F6F6F606F6F5F606F606F101010000000000000000000\
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF5F4F5F5F4F5F5F4F5F5F40505F4F5F302F30000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF6F7F5F6F8060607F5F6F7F5F607F5F607F5F60705F607F5F5F7050
60705F5F70505F7050506F50000000000000506F50000000000000CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF2F201F70603F000000000000CFCFCF000000000000000000000000\
303F60303F5F303F60303F60303F60000000004060003F600000000000002F304F
303F603F406F303F603F406F303F603F406F303F60CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000\
00000000000000000000000F4F5F804F507F4F5F804F5F7F505F804F5F7F505F80
4F5F80505F804F5F80505F804F5F80505F804F5F80505F80505F80506080505F80505F80505F80\
50608F505F8050608F505F800000000000000000000000000F0F0F0000003F3030
6F6060706F606F605F7060606F605F706060C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000000000000000000000
0000000000000000000000000000000000000000001F1F1F303030706F6F70606F706F6F706F6F\
706F6F70606F706F6F70606F706F6F70606F706F6F70606F706F6F70606F706F6F
70606F706F6F706060706F6F70606F706F6F6F6060706F6F70606F706F6F6F60606F5F606F5F60\
1F1F1F000000000000000000000000000000000000000000000000000000000000
0000000000000000000F000F504F4F70606F6F606070606F000000000000000000000000000000\
00000000000000000060505F6F606F6F5F606F606F6F5F606F606F6F5F606F606F
6F5F606F606F6F5F606F606F6F5F603F303F302F30000000000000000000000000000000000000\
000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF5040505F4F5F5F40505F4F5F5040505F4F5F000000000000C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF607F5F607F5F60705F607F5F5F7050607F5F5F70505F7050
5F70505F7050506F503F4F3F000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF706040403F20706040000000000000CFCFCF000000000000000000\
202F40304060303F60304060303F6030406000000000203F000000000000000000
3F406F303F603F406F30406F3F406F30406F3F406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000505F804F5F80505F804F5F80505F804F5F80
505F804F5F80506080505F80505F804F5F8050608F505F8050608F505F8050608F505F8050608F\
505F805F608F505F805F608F403F3F0000000000000F0F0F504F4F706F60605F50
706F60706060706F60706060706F606F6060706F60706060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
000000000000000000000000000000000000000000000000000000000000100F0F706F6F706F6F\
706F6F7F6F6F706F6F706F6F706F6F7F6F6F706F6F706F6F706F6F706F6F706F6F
706F6F70606F706F6F706F6F706F6F70606F6F606070606F605F5F6F606070606F6F60606F6060\
6F5F604F404F000000000000000000000000000000000000000000000000000000
0000000000000000000000000F0F0F504F4F70606F6F606F3F303F000000000000000000000000\
00000000000000000000000000000070606F6F606F70606F6F5F6070606F6F606F
403F405F4F500F0F0F202020000000000000000000000000000000000000000000000000CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F4F5F\
5F4F5F5F4F5F5F4F5F5F4F5F5F4F5F5F4F5F000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF6F7F60607F5F607F5F607F5F607F5F5F7050607F5F5F7050
5F705F5070505F7050000000000000405F40000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C050402F705F3F70603F000000CFCFCF000000000000000000\
000000303F5F303F60303F5F303F60303F60304060000000000000000000000000
303F6030406F303F603F406F303F603F406F303F60CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
0000000000000000000000000000000000004F5F804F507F4F5F804F5F7F505F80
4F5F7F505F804F5F80505F804F5F80505F804F5F80505F804F5F80505F80505F805060804F5F80\
505F80505F8050608F505F80706F60403F3F0000000000000000006F6060706F60
6F605F7060606F6060706F606F605F7060606F605F7060606F605FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0000000000000000000000000000000000000000000000000000000000000000000403F3F\
403F3F706F6F706F6F70606F706F6F70606F706F6F70606F706F6F70606F706F6F
70606F706F6F504F4F6F60605F50506F60606F5F606F60606F5F60504F4F6F5F60504F4F000000\
2F2F2F6F5F604F4040000000000000000000000000000000000000000000000000
0000000000000000000000000000000F000F10101070606F6F6060000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F4F5F5F4F50\
5F4F5F5F4F505F4F5F504050000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF607F5F607F5F5F705F607F5F5F70505F705F5F70505F7050
506F505F7050506F500000002F3F2F0000000000001F201FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF70603F706040705F3F5F4F30000000000000000000000000\
101020304060303F60304060303F60304060303F60304060303F60303F5F202F4F
303F603040603F406F303F603F406F30406F3F406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000101F204F5F80505F804F5F80
505F804F5F80505F804F5F805060804F5F80505F804F5F8050608F505F8050608F505F8050608F\
505F8050608F505F80000000000000706F600000000000000000001F1F1F706060
706F60706060706F60706060706F60706060706F60706060706F60706060706F60CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000\
0000000000002F2F2F2F202F1F1F1F706F6F706F6F7F6F6F404040706F6F2F2F2F
0000000000006F606F6F606070606F6F6060504F506F606070606F1F1F1F000000605F5F70606F\
0000004F4040000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000302F302F2F2F000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F4F5F5F4F5F5F4F5F5F4F5F\
5F4F5F2F2020000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF607F60607F5F607F5F5F7050607F5F5F70505F705F5F7050
5F7050506F500000000000000F0F0F000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF705F3F70603F705F3F5F4F30000000000000000000\
0000002F3F5F303F60303F60303F602F3F60303F601F203F2F3050303F60304060
303F6030406F303F6030406F303F603F406F303F60CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000\
000000000000000000000000000000000000000000000000000000202F3F4F5F80
4F5F7F505F804F5F7F4F5F804F5F80505F804F5F80505F804F5F80505F804F5F80505F804F5F80\
50608F4F5F80505F80000000000000000000706060000000000000000000101010
6F6060706F606F605F7060606F6060706F606F605F7060606F60607060606F605F7060606F605F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000201F1F6F60606F6060403F3F403F406F5F606F60600000000F0F0F6F5F60\
60505F000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F4F5F5F4F500F0F0F302F305F4F5F101010\
000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF60705F607F5F5F70505F705F5F70505F705F506F505F7050
506F505070500000004F604F0000000000003F4F3FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F604070603F706040705F3F000000000000000000\
000000202F40303F60304060303F60304060303F6030406000000030406F303F60
30406F303F603F406F3040603F406F303F603F406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000000000000000000000000
1F202F4F5F80505F804F5F80505F804F5F805060804F5F80505F804F5F8050608F505F8050608F\
505F8050608F505F80CFCFCF0000000000000000006F5F5F000000000000000000
6F6060706060706F60706060706F60706060706F60706060706F60706060706F60706060706F60\
6F6060706F60706060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005F50506F60606F6060000000403F3F6F60606F6060000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F4F5F1F101F0F000F0F0F0F0000002F202F000000000000\
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF607F5F5F705F607F5F5F70505F705F5F70505F705F4F604F
3040300000000F0F0F000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0706040705F3F70603F705F3F000000000000\
0000001F1F30303F602F3F5F303F60303F60303F602F3F60303F60000000304060
303F60304060303F6030406F303F6030406F303F6030406FC0C0C0CFCFCFC0C0C0CFCFCF40507F\
4F5F7F4F507F4F5F80000000000000000000000000000000000000000000000000
000000101F204F507F4F5F804F5F7F4F5F804F5F80505F804F5F80505F804F5F80505F804F5F80\
505F804F5F8050608FC0C0C0CFCFCF0000000000000000004F4040000000000000
000000706F606F6060706F606F605F7060606F6060706F606F605F7060606F60607060606F605F\
7060606F605F7060606F605F706060C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005F50506F60606F6060403F3F0000001F1F1F000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C060505F0000005F4F5F4F3F400F0F0F0000000000000000003F303F000000\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF5F70505F705F5F70505F705F506F505F7050202F200F0F0F
0000000000003F4F3F000000000000101010C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF70603F7F604070603F706040705F3F000000\
000000000000303F5F304060303F60304060202F40000000303F60000000202F40
30406F303F6030406F303F603F406F3040603F406F303F60CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
4F507F4F5F804F507F4F5F804F5F7F4F5F80000000000000000000000000000000
000000000000000000101020505F804F5F80505F804F5F805060804F5F80505F804F5F8050608F\
505F8050608F4F5F80CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000
000000101010706F60706060706F60706060706F60706060706F60706060706F60706060706F60\
706060706F606F6060706F607060607060606F6060CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000
0000000000000000000000000000000000000000003F3030101010000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
60505F5F4F5F0F0F0F2F20200F0F0F0F000F0000000000001F10102F202F000000CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF607F5F5F705F5F705F5F70505F705F5070500F0F0F000F00
0000000F0F0F0000000000004F6F4F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF705F3F706040705F3F70603F000000\
000000000000101F202F3F5F303F602F3F5F303F6000000010102020304F10101F
303F60304060303F60304060303F6030406F303F6030406FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
4F5F7F40507F4F5F7F40507F4F5F8040507F4F5F804F507F4F5F80000000000000
0000000000000000000000001010204F5F804F5F7F4F5F804F5F80505F804F5F804F5F804F5F80\
505F804F5F80505F80CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000
000000000000302F2F706F606F6060706F606F60607060606F6060706F606F605F7060606F5F5F\
7060606F605F7060606F60607060606F605F7060606F605F706060C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F4F5F60505F
5F4F5F0F0F0F000000000000000000000000403040201F20000000000000C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF5F70505F705F5070505F7050506F505F7050405F400F0F0F
0000002F3F2F000000000000000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF70603F7F6040705F3F706040\
00000000000010101F303F60303F60304060303F601F203F101020304060000000
304060303F6030406F303F6030406F303F603F406F30406FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
40507F4F5F804F507F4F5F804F507F4F5F804F5F7F4F5F804F507F4F5F804F5F80
0F0F10000000000000000000000000202030505F804F5F80505F804F5F805060804F5F80505F80\
4F5F8050608F4F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
0000000000000000000F0F0F706F60706060706F60706060706F60000000000000000000000000\
0000000000000000000F0F0F6F6060706F607060607060606F6060706060706060
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60505F5F4F5F60505F1F1F1F
0000000000000000005F4F50000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF5F7F5F5F70505F705F5F70505F7050304F300F0F0F000000
3040300000000000004F604F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0706040705F3F706040705F3F\
0000000000000000002F3F5F303F602F3F5F303F60101F2F0000002F3F601F1F2F
000000303F60303F60304060303F60304060303F6030406FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF40507F4F5F7F40507F4F5F7F40507F4F5F8040507F4F5F804F507F4F5F80
40507F4F5F801F1F2F000000000000000000101F204F5F804F5F7F4F5F804F5F804F5F804F5F80\
4F5F804F5F80505F80C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0000000000000000000000000706060403F3F302F2F000000000000000000000000000000\
0000000000000000003030307060606F60607060606F605F7060606F605F706060
6F605F7060606F605FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000000000
00000000000000000000000000000F000000000000000000000000000000000000000000000000\
0000000000000000001F20200F10106F808F000000000000000000000000000000
000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C060505F5F4F5F60505F3F2F30000000000000000000
5F4F5F000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C00000005F70505F70505070505F70500000000F0F0F000000000000
1F201F000000405F40000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F604070603F7F6040\
0000000000000000000F0F10303F60303F602F3F60304060000000101020303F60
000000101020304060303F6030406F303F6030406F303F6030406FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF00000030405F4F5F804F507F4F5F8040507F4F5F804F5F7F4F5F804F507F
4F5F804F5F804F5F804F5F7F4F5F80202F40000000000000505F804F5F804F5F804F5F80505F80\
4F5F80505F804F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000000000000000\
000000101010504040706F60706060706060706060706F607060607060606F6060
7060607060607060606F6060706060706060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF000000000000000000000000000000202F30607F80708F9F4F5060000000
0000000000000000000000007F909F000000000000000000000000000000000000000000000000\
000000202F30000000708F90000F0F708F8F5F6F70000000000000000000000000
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF60505F60506060505F0000000000000000003F303F4F3F40000000
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5F70505F705F5F70505F7050101F103F4F3F000F00000000304F30
0000000000000F0F0F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0706040705F3F\
706040000000000000000000303F602F3F5F303F602F3F5F101F2F000F0F303F60
2F3F600000002F3F60303F60303F603040602F3F60304060303F60CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF0000000000000000000F0F0F10101F2F304F40507F4F5F8040507F4F5F80
40507F4F5F8040507F4F5F804F507F4F5F804F507F4F5F802F304F4F507F4F5F804F5F804F5F80\
4F5F804F5F804F5F80CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000000000000000000000000000\
4F40406F60607060606F60607060606F605F7060606F60607060606F605F706060
6F60607060606F605F7060606F605F7060606F5F5F7060606F605F706060C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000\
000000000000000000101F1F708F9F70809F708F9F708F9F708F9F70809F708F9F
000000000000000000000000708F90708F9F000000000000000000000000000000000000000000\
00000040505F0000006F808F0000006F808F70808F000000000000000000000000
000000000000000000000000000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C060505F60505F\
60505F604F5F00000000000000000000000060505F000000000000000000000000
5F4F5F50404FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C05F705F5070505F7050506F50506F504F604F0000000000002F3F2F
0000004F604F000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7060407F6040\
705F3F0000000000000000002F3F5F304060303F60303F60202F40000000303F60
3040600F0F10202F3F303F6030406F303F6030406F303F6030406FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF0000000000000000000000000000000000000000000000004F5F804F5F7F
4F5F804F507F4F5F804F5F804F5F804F5F804F5F804F5F804F5F804F5F804F5F804F5F804F5F80\
4F5F80505F8F4F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000\
706060706F60706060706F60706060706F60706060706060706060706F605F504F
6F5F5F2F20200F0F0F0000004F403F6F60607060607060607060606F60607060606F6060706060\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000\
000000000000708F9F708F9F708F9F708F9F708F9F708F9F708F9F7F8F9F000000
708F9F0000000000000000007F8F9F708F9F000000000000000000000000000000000000000000\
607F8010101F101F1F708F9050606F708F906F808F1F20200F0F102F30306F8080
10101000000000000000000000000000000000000000000060505F60505F6F5060302F30000000\
00000000000000000060505F1F10100000000000003F303F60505F000000000000
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5F70505F705F506F50304030000F000F0F0F000000405F40000000
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF705F3F\
7060400000000000000000000000002F3F5F303F602F3F5F303F60000000000000
2F3F5F303F600000003040602F3F60303F60303F6030406F2F3F6030406FC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000
0000000000003F4F6F4F5F8040507F4F5F8040507F4F5F8040507F4F5F804F5F804F5F804F5F80\
4F5F804F5F804F5F80C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000\
7060606F60607060606F60607060606F60607060606F6060000000000000000000
0000000F0F0F000000000000000000000000000000000000000000101010605F507060606F5F5F\
7060606F5F5F706060C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000006F7F906F809050606F\
4F5F606F809F708F9F6F809F708F9F3F4F5050606F70809F708F9F00101F000000
708F90000000000000000000708F90708F9F000000000000000000000000000000000000000000\
6F808F6F808F708F8F607F805F6F70708F90709090708F90708F90708F90709090
708F8F708F90303F40000000000000000000000000000000000000000000000000000000000000\
3F303060505F00000000000050404F5F4F5F000000000000000000C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C05F7050506F50507050000000101010000000000000405F40000000
405F40000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F6040\
7060407F6040000000000000000000000F0F2F3F5F3040602F3F60303F60000000
2F3050303F601F1F2F00000030406F303F6030406F303F6030406F303F60CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000
0000000000000000000000002F2F404F5F804F5F804F5F804F5F804F5F804F5F804F5F804F5F80\
4F5F804F5F804F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF706F6F\
706060706F60706060706F60706060706F607060600F0F0F000000000000000000
7060607060607050506F4F4F7060600F0000000000000000000000000000000000000000000000\
302F2F7060606F60607060606F60607060606F5F60706060CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000006F80900000006F8090\
70809F708F9F70809F0F0F0F0000002F3030708F9F1F20203F4F50004F50000F0F
7F8F9F3040400000000000007F909F708F9F708F9F000000000000000000000000000000000000\
6F7F806F7F80708F907F909F708F907F9090708F907F909F708F90709090708F90
7F9090708F90709090708F90101010000000000000000000000000000000000000403F40000000\
000000403F40000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5070505F7050405040203020000F000000004F604F000000000000
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
7F6040705F3F000000000000000000000000101F2F2F3F5F303F602F3F5F1F1F2F
202F40303F602F3F600F0F100F0F103040602F3F60303F602F3F6030406F000000CFCFCFCFCFCF\
CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000
2F3050000000000000000000000000000F0F3040604F5F8040507F4F5F8040507F4F5F804F5F80\
4F5F80405F804F5F804F5F80CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF6F60607060607060607060606F6060706060000000000000000000000000
100F0F7060606F1F1F7000006F2F2F605050000000000000000000000000000000000000000000\
000000000000000000000000504F4F6F5F5F7060606F5F5F7060606F5F5F6F6060
6F5F5F7060606F5F5FCFCFCFCFCFCFCFCFCF000000000000000000506070000000506070607F8F\
708F9F5F6F7F0000006F809F708F9F60708000000070809F708F9F000000000000
708F9F5F6F7F000000000000708F90708F9F0000000000000000000000000000001F20206F808F\
708F9F708F90708F90708F90709090708F90708F90708F90709090708F90708F90
708F90709090708F8F708F90708F8F3F4040000000000000000000000000000000000000000000\
000000000000000000000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0507050101F103040300F0F0F0000000000004F604F000000000000
000000CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF7F6040706040000000000000000000000000303F602F3F5F3040602F3F5F
0000002F3F60304060303F600000002F3F6030406F303F6030406F303F6030406F000000CFCFCF\
CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000202F3F40507F
4F5F8040507F4F5F80405F7F4F5F8000000000000010101F4F5F804F5F804F5F804F5F804F5F80\
4F5F804F5F804F5F804F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF706F60706060706F60000000000000000000000000000000
0000000F0F0F706F607060607060604F403F0F0F0F6F5F5F100F0F0000000000000000002F2020\
0000000000000000000000000000000000000000000000000F0F0F50404F6F6060
7060606F5F607060606F5F6070606000000000000000000000000040505F1F1F1F6F7F90708F9F\
6F809F3F404F70809F708F9F303F4040505F708F9F708F9F708F9F708F9F708F9F
708F9F20202F000000000000708F9F708F900000000000000000000000007F909F708F907F909F\
708F907F909F0000000000000000000000000000000F0F0F404F50304040708F90
709090708F90709090708F90709090708F8F000000000000000000000000000000000000000000\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF506F500F0F0F000F000F0F0F0000003F503F000000101F10000000
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF70603F7F604030201F000000000000000000000000303F602F3F5F303F60
000000000F0F2F3F60303F602F3F600000002F3F603040602F3F60303F602F3F60302F20000000\
CFCFCFC0C0C0CFCFCF0000000000000000000000000000000000000000004F5F7F
40507F405F7F401F204F000040000F4F5F8000000F0000000000002F304040507F4F5F8040507F\
4F5F80405F804F5F80405080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007060606F60602F2020000000\
4F40406F60603F3030000000000000000000000000000000000000000000000000
0000003F30306F5F5F5F4F4F0000000000000000000000000000000000002F3F4070809F6F8090\
0000006F809F708F9F2F303F5F6F706F809F708F9F6F8090708F9F708F9F708F9F
70809040505F000000000000708F90708F9F2F3F3F000000000000708F9F708F90708F9F2F303F\
0000003F404F6F7F80708F90708F90708F90708F906F8080101F1F0000004F5F5F
4F6060708F90708F8F708F906F8F8F708F90000000000000000000000000000000000000000000\
000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF506F500F0F0F0000000F0F0F00000000000020302040604F000000000000
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF7F60407F60400000000000000000000000002F3F5F303F602F3F5F
304060101F20202F402F3F603040602F3F600F0F102F3F6030406F303F6030406F8F7060907F60\
000000CFCFCFCFCFCF000000000000000000000000000000000000000000000000
4F5F7F40507F4F507F402F3F4F0F104050704F5F80000000000000000000000000405F804F5F80\
4F5F804F5F804F5F804F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000706060\
504F4F0000000000000000006F5F5F7060606F6060605050000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006F809F5F7080\
6F809F708F9F6F809F3F4F5070809F708F9F70809F708F9F708F9F708F9F708F9F
708F9F0F0F0F0000000000007F8F9F708F9F7080906F808F7F909F708F9070909F000000708F90\
708F9070909F708F907F909F708F90709090708F90709090708F90709090202F2F
709090708F90709090708F8F709090708F8F000000000000000000000000000000000000000000\
000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF5070504F6F4F1F201F0F0F0F0000001F201F000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C07F604070603F7F60400000000000000000002F3F5F202F40303F60
2F3F5F303F60000000303F602F3F60303F60202F4F1010202F3F603040602F3F60303F608F705F\
8F7F60000000CFCFCFCFCFCF000000000000000000000000000000000000000000
0000000000000000000000003F4F6F00000F00000000000000000000000000000010101F40507F\
4F5F8040507F4F5F80405F80CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000007060605F4F4F0000000000000000000000000000004F4040706060
504F4F3030300000000000000000000000000000000000000000000000000000006F809F1F2020\
70809F6F809070809F6F809F708F9F6F8090708F9F6F809F708F9F6F8090708F9F
708F90000000000000000000708F90708F9F708F90708F9F708F902F303F304040708F90708F90\
708F90708F90708F90708F90708F905F6F705060600000003040405F7070708F90
607F80708F906F8080506F6F6F8F8F0F0F0F000000000000000000000000000000000000000000\
000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
0000004F6F4F4F6F504F604F000000000000000000000000000000C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF7F60407F60400000000000001010202F305F303F5F2F3F5F
303F602F3F5F304060000000303F602F3F603040601F20301F20302F3F6030406F303F60000000\
8F7060907F60000000CFCFCF000000000000000000000000000000000000000000
00000000000000000040507010101F000000000000000000000000000000000000000000405070\
405F804F5F80405F804F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000\
0000000000000000000000000000007060605F5050000000000000000000000000
000000000000000000000000607F8F6F809F6F809F6F809F0000000000000000006F809F000000\
6F809F708F9F4F5F6F7060606F5F5F7060606F5F606F60606F5F5F7060606F5F60
6F5F606F5F5F6F5F606F5F606F5F60708F9070909F708F90708F9F708F907F909F708F9070909F\
708F9070909F000000101F1F506F6F708F90506F6F2F30300F0F0F5060601F2020
708F9000000000808F000000708F90000000000000000000000000000000202F2F000000000000\
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
4F6F50506F504F6F4F0000000000004F604F405F4F000000000000CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C07F604070603F00000000000020304F2F30502F3050
2F3050303F602F3F5F1F203F0F0F1F303F602F3F60303F601F203F1F20302F3F60304060100F0F\
2F2F208F705F907F60000000CFCFCF000000000000000000000000000000000000
00000000000000000040507F3F50700000000000000000000000000000000000000000001F2030\
405F8040507F4F5F8040507F4F5F80C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000\
0000000000000000000000000000000000000000000F0000706060403F3F1F1F1F
0000000000000000000000000000000000000000006F809F6080906F809F6F80906F809F6F8090\
00000000000000000050404F6F5F606F5F5F6F5F606F5F5F6F5F606F5F5F6F5F60
6F5F5F6F5F606F5F5F6F5F606F5F5F000000000000000000708F90708F90708F90708F90708F90\
000000304040708F906F8F90708F90708F90708F906F8F8F708F906F8F8F708F90
000000008F90000000708F903F4F4F0000000000000000000000006F8F8F000000000000000000\
000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000\
4F6F504F604F4F6F4F0000001F201F101F10000000000000CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F60407F60400000002F305F2F3F5F00000F
20304F2F3050303F602F3F5F00000F2F3050303F602F3F603040602F30500000002F3F6030406F\
000000302F208F7060907F60000000000000000000000000000000000000000000
000000000000000000202F40203040000000000000000000000000000000000000000000000000\
405F804F5F80405F804F5F80405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
000000000000000000000000000000000000000000000000000000000000303030
6F5F5F2F20200000000000000000000000000000000000006F809F6F809F6F809F403F3F403F3F\
000000000000000000000000000000000000000000504F4F6F5F606F60606F5F60
7060606F5F606F5F606F5F5F1F1F1F0000000000000000000000000000006F5F60708F901F2020\
708F90708F90708F90709090708F90708F90708F90709090708F90708F90000000
000000000000708F906F8F8F000000000000000000000000708F906F8F8F000000000000000000\
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000\
4F6F4F4F6F50000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C07F604000000010101F2F3050000000
1F20302F3F5F202F4F303F602F3F5F303F600F101F303F602F3F60303F602F3F600000002F3F60\
907F60000000201F1F8F705F907F60000000000000000000000000000000000000
00000000000000000040507F40507F00000F2F3F50405F7F101F20000000000000000000000000\
0000002F3F50405F8040507F4F5F80C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000
0000000000000000006F809F0000006F809F607F90000000000000000000000000000000000000\
605F5F6F5F5F3F30300000000000000000000000000000000000000000000F0F0F
0000000F0F0F00000000000000000000000000000000000000000000000060505F6F5F60CFCFCF\
708F906F8F90708F90708F90708F906F8F8F708F906F8F8F708F90607F80000000
5F7070708F906F8F8F0000000000000000000000006F8F8F6F8F8F0F0F0F000000000000000000\
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF0000000000000000000000004F604F\
4F6F5040604F000000405F403F5040000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF7F60400000002F304F0F0F10
303F5F2F305F303F5F2F3050303F602F3F5F2F30501F1F30303F602F3F60304060303F60000000\
2F3F60907F601F1F101010108F7060907F60000000000000000000000000000000
0000000000000000002F3F50405F7F000000000000303F50202030000000000000000000000000\
000000202F3F405F804F5F80405F7F4F5F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF0000006F809F60809F0F0F103F40506F809F6F809F000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF709090708F90708F90708F90709090708F90708F906F8F8F708F906F8F8F
708F90506F6F000000000000000000000000708F906F8F8F506F6F000000000000000000000000\
000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000503F400000004F6F50\
40604F000000000000101010000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000000000001F1F2F
2F30501F20300000002F3F5F2F3F5F303F602F3F5F1F20301F2040303F602F3F60303F602F3F60\
00000F8F7060907F600000008F7F608F705F907F60000000000000000000000000
000000000000000000101F2040507F0F0F10000000303F5F40507F000000000000000000000000\
000000000000405F8040507F405F8040507FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0000000607F906F809F607F9060809F607F90CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF6F8F8F708F906F8F8F708F906F8F8F708F906F8F8F708F90
0F0F0F0000000000000000000000006F8F8F6F8F8F0F0F0F000000000000000000000000000000\
000000000000000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C000000000000000000000000050304000000000000040604F\
000000000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
2F304F0F0F10101F202F305F303F5F2F3050303F602F3F5F1010202F3050303F602F3F60304060\
303F603040608F7060907F603F302F907F60101010000000000000000000000000
000000000000000000304060405F802F3F500000000000003F4F6F2F304F000000000000000000\
000000000000303F5F4F5F80405F804F5F80405F7FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF0000006F809F607F906F809F608090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF708F90708F906F8F8F708F906F8F8F708F90303F3F
0000000000000000004F6060708F906F8F8F506F6F000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF00000000000000000000000030202F4F30400000004F604F4F6F4F\
00000040604F000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000
0000000000000000002F3F5F202F402F3F5F2F3050303F602F3F5F0000002F3F5F303F602F3F60\
303F602F3F60303F608F7060907F6010100F907F60000000000000000000000000
000000000000000000202F3F40507F00000000000000000040507F405F7F000000000000000000\
00000000000000000040507F405F8040507F405F80CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF000000607F90608090607F9060809FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF6F8F8F708F906F8F8F708F90000000000000
0000000000006F808F6F8F8F6F808F6F8F8F000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C00000000000000000000000000000004F304000000000000040604F000000\
0000003F5F40000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
0000000000000000002F3F5F0000000F0F10303F5F2F305F303F602F3F5F00000F0F0F10303F60\
2F3F60304060303F60CFCFCF8F7060907F60000000907F60000000000000000000
0000000000000000002F3F50202F3F000000000000000000101F203F4F6F000000000000000000\
000000000000000000405F8040507F405F80405F7FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF0000006F809F607F900F0F0FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F8F708F90000000000000000000
0000006F8F8F6F8F8F6F8F8F000000000000000000000000000000000000000000000000000000\
0000000000004F60600000004F6060000000000000000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF000000000000000000000000000000000000503040503F400000004F6F502F3F2F000000\
0000003F5F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
0000000000000000000000000000002F30502F305F2F30502F305F303F602F3F5F2F3F5F000F0F\
303F602F3F60303F602F3F60CFCFCF8F7060907F6010100F907F60000000000000
000000000000000000202F3F101F2F00000000000000000000000040507F1F202F000000000000\
00000000000000000F40507F405F7F40507F405F8040507FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCF00000060809F607F90CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF0000000000000000006F8F8F
6F808F6F8F8F6F808F0000000000000000000000000000000000000000000000000000005F7F7F\
0000005F7F7F0000005F7F7F1F201F5F7F7F203030000000000000C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
0000000000000000000000000000000000003F2030503040000000000000405F4F000000000000\
3F5040C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF0000000000000000000000001F20302F304F0000000000002F3F5F303F602F3F60304060\
00000F303F602F3F60304060CFCFCFCFCFCF907060907F60302F20907F60000000
00000000000000000040507F101F200000000000000000000F101040507F000000000000000000\
0000000000003F4F70405F8040507F405F8040507F405F80CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF1F202F5F70806F809FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000006F8F8F6F8F8F
6F8F8F6F8F8F000000000000000000000000000000000000000000000000000000607F7F101F1F\
607F7F0000005F7F7F5070705F7F7F60807F000000000000000000CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
000000000000000000000000000000000000503F4F4F304000000040604F0000000000003F3020\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF0000000000000000000000000000000000000000000000000000000F0F10303F602F3F5F\
303F60000000303F602F3F60303F60C0C0C0CFCFCFCFCFCF907F600F0F0F000000
00000000000000000040507F1F2F3F00000000000000000000000030405F101F20000000000000\
0000000000001F202F40507F405F803F507F405F7F40507F405F80C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0000000607F90CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF6F8F8F6F808F6F8F8F
6F808F6080800000000000000000000F0F0F1F2020506F6F000000101F1F5F7F7F607F7F5F7F7F\
5F7F7F5F7F7F0F0F0F3F4F4F607F7F1F2020000000000000000000000000CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000
0000000000000000000000000000005030404F304050304040604F40604F0000003F0F0FCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000000000303F60\
2F3F603040600000002F30502F3F60CFCFCFCFCFCFCFCFCFCFCFCF907F600F0000
0000000000002F3F4F30405F000000000000000000000000000000202F403F4F6F000000000000\
000000000000000F0F0F0F0F00000F30406040507F405F8040507F405F80CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F8F8F6F8F8F
6F8F8F60808F6F8F8F6F808F6F8F8F6080806F8F8F6080806F8F8F6080806F8F8F608080607F7F\
5F7F7F6080805F7F7F607F7F1F2020000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000
0000000000000000000000004F3040503040503F4F00000040604F0000003F5F4FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0000000000000000000000000000000000000000000000000000000000000\
1F203F2F3F5F303F602F3F5F303F602F3F60CFCFCFC0C0C0CFCFCFC0C0C0907F60
70604F101F203F50700F0F0F0000000000000000000000000000000F0F103F50700F0F0F000000\
0000000000000F0F0F0000000F0F1000000F0F0F0F000000405F7F3F507FCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0404F50C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF60808F6F8F8F
6080806F8F8F6080806F8F8F60808060808F608080608F8F608080608080608080608080608080\
6080804F60604F60605F7F7F000000000000000000000000000000000000CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000
0000000000000000001F101F5030405030404F304040604F3F5F40101F10C0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000000000\
0000000000002F3F5F3040602F3F60303F602F3F60CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF6F5F4F3F50700F0F100000000000000000000000000000000F0F102F304F000000000000\
00000000000000000F0F0F0F00000F3040603F50700F0F0F00000F3F507040507F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404F5F5060705F708F607F90000000000000\
0000002F303F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60808F
6F8F8F2F3030101010000000000000000000000000000000101F1F608080608080608080608F80\
608080608080608080506F6F2F3030000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000
000000000000000000503040503F4F503040CFCFCF3F5F40000000CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000000000000000000000000000\
000000000000000000203050303F602F3F5F303F602F3F60CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF0000000000000000000000000000000000000000001F1F2F3F5070000000000000\
000000000000000F0F0F0F0F3F507000000F000F0F00000F20304F3F507F405F7F
3F507FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0507080000F0F\
000000507080000000000000000000000000CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C06F808F
6080806F8F8F608080608080608080608080607F805F70704050500F0F0F0F1010608080608080\
6080805F807F6080805F807F608080000000000000000000000000000000CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF000000000000000000000000000000000000000000000000000000000000
000000000000503040503040503040CFCFCF3F5F40000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
0000000000000000000000000000003040602F3F60304060CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF000000000000000000000000000000000000202F40405F7F000000000000\
00000000000000000F0F101000000F0F10103F5070405F7F3F507F405F8040507F
405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
5F7F90000000304050000000000000000000000000000000CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF608080
6F8F8F6080806F8F8F60808060808F608080608F8F608080608080507070202F2F1F1F1F5F7F7F\
5F80806080806080806080805F807F101F1F000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000100F105F30400000000000005F304F000000000000000000000000000000
0000005030405F3F4F503040CFCFCF3F5F403F5F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F3F5F2F3F5F303F60C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF0000000000000000000000000000003040603F5070000000000000\
00000000000F0F0F0F00000F101F203F507F3F50703F507F405F7F3F507F405F7F
3F507F405F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF5F7F8F0000000000000000005F7F8F000000000000000000C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C060808F
6080806080806080806080805F80806080806080806080805F80806080805F80800000000F1010\
6080805F7F7F6080805F7F7F5F80801F2F2F000000000000000000000000CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF0000000000005F3040\
0000005F30400F0F0F1F101F5F3040000000000000000000000000000000000000
50304050304F503040CFCFCFC0C0C03F5F40C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000\
0000000000000000000000000000000000000000003040602F3F60304060CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000000000001F20303F5070000000000000\
0000001F20303F507F2F304F1F202F0F0F0F1F2F3F405F803F507F405F7F3F507F
405F803F507F405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF5F7F8F1F1F200000000000000000005F70805F7F8F608090CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
608F8F1F20200F1010506F6F4050503040406080805F8080608080608080608080000000000000\
5F80806080805F7F7F6080805F807F304040000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000000005F30400F0F0F\
5F30404F2F3F4020305F304F0000000000001F10101F101F100F105F304F503040
5F3F4FCFCFCFCFCFCFCFCFCF3F5F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000000000000000000000000000\
0000000000000000000000000000000000001F203F2F3F5F303F602F3F5F303F60
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF0000000000000000003F50700F0F10000000000000\
0000002F3F5F3F507F3F507F405F7F3F507F40507F3F507F405F7F3F507F405F7F
3F507F405F7F3F507F405F7FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C05F7F905F7F8F607F905F7F8F5F7F905F7F8F5F7F905F7F8F000000
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
6080806080805F7F803040401F2F2F101010000000101F1F5F7F7F6080805F7F7F1F1F1F000000\
5F80805F7F7F5F807F5F7F7F5F807F2F3030000000000000000000000000CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000000000603F4F5F30403F202F402030\
5F304F5F2F405F304F00000060304F5F30405F304F5F304F5F304F5030405F304F
CFCFCFCFCFCFC0C0C03F5F40CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000\
00000000000000000020304F2F3F5F303F602F3F5F303F602F3F5F3040602F3F60
304060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF00000000000010101F3F5070000000000000000000\
000000405F7F3F507F405F7F3F507F405F7F3F507F405F7F3F507F405F7F3F507F
405F7F3F507F405F803F5F7F405F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF6080905F7F8F000000000000000000000000000000000000
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6080805F80806080805F80806080805F8080608080101F1F0F0F0F5070706080801F2F2F000000\
5F7F7F6080805F7F7F5F80805F7F7F000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000603F4F5F3040603F4F5F304F5F304F\
502F40603F5060304F603F4F5F304F603F4F5F304F5F3F4F5F304FCFCFCFCFCFCF
CFCFCFCFCFCF3F5F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000\
00000000000000000F202F40303F5F2F3F5F303F602F3F5F303F602F3F5F303F60
2F3F5FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000000000\
0000003F4F7040507F3F50703F507F3F507F405F7F3F50703F507F3F507F405F7F
3F507F3F5F7F3F507F405F7F3F507F3F5F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF5F7F8F5F7F8F5F7F8F5F7F905F7F8F5F7F8F5F7F8F0F0F0F
000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
5F7F806080805F7F806080805F7F7F5F80805F7F7F5F80800000000F10105F7F7F5F80805F7F7F\
5F7F7F5F7F7F5F807F507F7F5F7F7F000000000000000000000000C0C0C0CFCFCF
C0C0C0CFCFCF000000000000000000000000000000603F4F5F30405F304060304F603F4F5F304F\
60304F5F304F60304F5F304F5F304F5F304F5F304FC0C0C0CFCFCFC0C0C0CFCFCF
0000003F5F40C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
0000000000000F0F10303F602F3F5F303F602F3F5F303F602F3F5F303F602F3F5F
3040602F3F60CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000\
0F0F103F50703F507F405F7F3F507F405F7F3F507F405F7F3F507F405F7F3F507F
405F7F3F507F405F7F3F507F405F803F5F7F405F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF6080905F7F8F5F7F905F7F8F6080905F7F8F5F7F90000000
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
6080805F80806080805F7F7F6080805F7F805F80805F7F7F3040400000005F80805F7F7F5F8080\
5F7F7F5F807F5F7F7F5F8080000000000000000000000000000000CFCFCFCFCFCF
000000000000000000000000000000402F306F3F50603F4F6F3F50603F4F603F5060304F603F50\
502F3F5F304F4F2F3F100F0F20101F603F4FCFCFCFCFCFCFCFCFCFCFCFCF000000
0F0F0FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF000000\
0000000000000000002F3F5F303F5F2F3F5F303F5F2F3F5F303F602F3F5F303F60
2F3F5F303F602F3F5F201F10CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000\
000F0F3F50703F507F3F50703F507F3F50703F507F3F50703F5F7F3F50703F507F
3F507F405F7F3F507F3F5F7F3F507F405F7F3F507F3F5F7FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF5F7F8F5F7F8F5F7F8F5F7F8F5F7F8F5F7F8F000000000000
5F7F8F000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
5F7F7F5F80805F7F7F5F7F805F7F7F5F80805F7F7F5F7F7F5F7F7F3F50504F60605F7F7F507F7F\
5F7F7F507F7F5F7F7F203030000000000000000000000000000000000000000000
0000000000000000006F3F50603F4F6F3F50603F4F603F4F603F4F2F1F20301F2F000000000000\
40203000000020101F5F304FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F30303F3F3F
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
0000000000000000002F3F502F30501F2030000000000000000000000000000000
0000000000000000000000005F4030705F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
0000000F0F0F3F507F405F7F3F5070405F7F3F507F405F7F3F507F405F7F3F507F
405F7F3F507F405F7F3F507F405F7F3F507F405F7F3F5F7F405F7F000000000000CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF5F7F8F5F7F8F5F7F901F202F0F0F0F0000000000004F606F
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F7F80
5F80805F7F7F5F80805F7F7F5F80805F7F7F5F80805F7F7F203030000000008080002020000000\
507F7F5F7F7F000000000000000000000000000000000000000000000000000000
6F40506F3F4F6F40506F3F506F4050603F4F20101F000000000000603F4F6F3F5060304F603F50\
60304F603F5060304FCFCFCFCFCFCFCFCFCFCFCFCF0000003F30303F3F3FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
0000000000000000000000001F20300F101F0F0F10000000000000000000000000
000000000000101F200F101F3F50703F5F7F3F507F3F5F7F3F507F3F5F7F3F507F000000000000\
000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF5070800F0F0F5070800F0F0F0000000000002F3F40000000
000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C05F7F80
507F7F5F7F80507F7F5F7F80507F7F50707000000000000000505F000000000000507F7F507F70\
0F0F0F0000000000000000000000000000000000000000006F40506F3F4F6F4050
6F3F4F6F4050603F4F2F1F20000000301F2F603F4F6F3F5060304F603F4F60304F603F5060304F\
60304FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000003F30303F3F3FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003F507F405F7F3F507F405F7F3F507F405F7F3F5F7F405F7F3F5F7F000000\
000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF5F7F8F5F7F8F0F0F0F0000000000005F7F8F000000000000
0000005F7F8F000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F7F7F
5F80805F7F7F5F7F80507F7F5F7F80507F7F4F6F6F1F2F2F5F7F7F507F7F5F7F7F000000000000\
0000000000000000000000000000000000000000000000006F40506F40506F3F50
6F40500000004F2F3F6F3F4F6F40506F3F506F3F50603F4F6F3F50603F4F603F5060304F603F50\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000003F30303F3F3FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000000000
0000000000000000000000000000000F0F0F101F20000000000000000000C0C0C0CFCFCFCFCFCF\
CFCFCF000000000000000000000000000000000000000000000000000000202F3F
2F3F503F5F7F3F50703F5F7F3F507F3F5F7F3F50703F5F7F3F507F3F5F7F3F507F3F5F7F3F507F\
000000000000000000000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C05F7F8F507080000000000000000000507080000000000000304F50
000000000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05F7F7F
507F7F5F7F7F507F7F507F7F507F7F5F7F7F50707F507F7F1F2020000000000000000000000000\
0000000000000000000000000000000000000000000000004F303F000000000000
6F3F4F6F4050603F4F6F3F50603F4F6F3F50603F4F603F5060304F603F50C0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF0000000000003F30303F3F3FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000
202F402F3F5F3040602F3F5F303F602F3F5F3040602F3F60000000000000000000000000CFCFCF\
CFCFCFCFCFCF0000000000000000000000000000003F5F7F3F5070405F7F3F507F
3F5F7F3F5070405F7F3F507F3F5F7F3F507F405F7F3F507F405F7F3F507F405F7F3F5F7F405F7F\
3F5F7F405F7F000000000000000000000000000000000000000000000000CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF5F7F8F50708F000000000000000000304F50000000000000101F20000000
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F7F80507F7F
5F7F7F507F7F5F7F7F203030000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000000000000F0F0F603F4F7040506F4050
6F40506F3F506F40504020306F3F504020305F3F4F603F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF000000201F1F302F2F3F30303F3F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000000000000000000000
2F30502F3F5F2F3F5F303F5F2F3F5F303F5F2F3F5F303F602F3F5F70503F000000000000000000\
000000C0C0C0CFCFCF0000000000003F50703F507F3F50703F507F3F50703F507F
3F50703F507F3F50703F5F7F3F50701F2F3F000F0F10101F0000000000001F20303F5F7F3F5070\
3F5F7F3F507F3F5F7F3F507F000000000000000000000000000000000000000000
000000000000000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
50708F5070804F6F7F0000000000000000003F505F0000003F505F000000000000
000000000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF50707F000000
000000000000000000000000000000000000000000000000000000000000102020507F7F4F7070\
4060600000000000000000000000000000007040506F3F4F6F40506F3F5020101F
0000000000000000002F1F201F1010402F30C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
000000302F2F403F3F3F30303F3F3FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000
202F3F2F3F5F303F5F000000000000202F40303F602F3F5F3040602F3F5F3040607F5F403F2F20\
000000000000000000CFCFCF3F50703F507F3F5070405F7F3F50703F5F7F3F5070
405F7F3F507F101F1F000000000000000000000F0F20304F101F200000000000000000001F2F30\
3F5F7F405F7F3F5F7F405F7F3F5F7F405F7F00000010100F5F403F000000000000
000000000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF507F8F\
50708050708F0000000000000000000000000F0F0F5F7F8F000000000000000000
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000
0000000000000000000000000000000000002F4040507070507F7F507070507F7F2F4040000000\
000000000000000000000000000000000000000000000000100F102F1F206F3F50
6F40506F3F506F40506F3F50CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000\
000000403F3F3F30303F3F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000
2F3050303F5F2F3050000000003F5F000000000000303F5F2F3F5F303F602F3F5F303F600F0000\
000000402F207F5F407F5040000000C0C0C03F50703F50703F507F3F50703F507F
0000000000000000000000000000003F5F7F3F2F3F3F00003F5070000000000000000000000000\
0000000000001F2F3F3F5F7F3F5F7F3F507F3F5F7F3F5F7F906F608F6F5F8F6F60
8F6F5F8F6F608F6F5F8F6F608F6F5F50403F000000000000000000000000000000000000000000\
000000C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C05070804F7080\
5070804F708050708000000050708F4F00005070800F1010000000000000000000
000000CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000000000
507070507F7F4F707050707F4F7070507F7F4F707050707F000000000000000000000000000000\
0000000000000000000000000000006F40507040506F40507040506F3F4F6F4050
6F3F506F4050C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF1F10100000000000002F2020000000\
403F3F3F3030CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000
0000001F2030303F5F2F3050000000002030000F0F000000303F602F3F5F3040602F3F5F304060\
000000000000000000000000000000000000000000000000203040000000000000
0000000000000000000000003F50703F5070401F203F2F3F304F6F000000000000000000000000\
000000000000000000000000304F60405F7F3F5F7F405F7F3F5F7F40607F001010
00201F302F209070608F6F608F6F608F6F60906F608F6F608F6F608F6F5F8F6F6050403F4F3F30\
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F708050708F4F7080507080\
4F708050708F50708050708F0000000000004F70800F0F0F000000000000000000
000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000\
000000000000000000000000000000000000000000507F7F50707F507F7F50707F
507F7F4F7070507F7F507070507F7F0F0F10000000000000000000000000000000000000000000\
00000000000000000000000000000070405F6F40507040506F40507040506F4050
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000001F10100000004F30300000000F0F0F403F3F\
3F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000
0000000000002F3F5F303F5F2F3F5F2F3F5F2F3F5F303F5F2F3F5F303F5F2F3F5F303F602F3F5F\
303F60000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000001F2F3F101F200000000F0F103F5F7F000000000000101F20\
0000000000000000000000000000000000002F40503F5F703F5F7F3F5F7F3F5F7F
3F5F7F3F5F7F1010108F6F608F6F5F8F6F608F6F5F8F6F608F605F8F6F608F605F8F6F608F605F\
8F6F608F605F8F6F60201F101F1010000000000000000000000000000000000000
0000000000000000000000000000000000000000004F70804F6F804F70804F6F805070804F6F80\
10101F4F6F805070804F6F805070804F7080000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
00000000000000000000000000000000000050707F4F707050707F4F707050707F
1F2F2F000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000007040506F40507040506F3F506F4050C0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF000000302020000000000000000000000000000000403F3F3F3030\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF000000000000101F2F303F5F2F3F5F303F5F2F3F5F0000002F3F5F3040602F3F5F304060\
2F3F5F3040602F3F5F805F4F7F5F40302F1F5F40301F1010000000000000000000
0000000000000000000000000000000000000000000000000000000000003F5F7F0F1010000000\
2F3F503F5F7F000000000000000000000000000000405F7F3F5F7F405F7F2F4050
0F0F0F000F0F40607F3F5F7F40607F3F302F0000008F6F60906F608F6F6000000050403F8F6F60\
8F6F608F6F608F605F8F6F608F60608F6F608F605F8F6F608F605F0F0F0F000000
0000000000000000000000004F70804F6F804F70804F6F805070804F6F805070804F6F80507080\
4F70800000000000005070804F70805070804F7080507080000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000201F1F7F505F0000000000000000004F707050707F4F70704F6F70000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000704F5F704050CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF000000000000000000503F3F0000000F0F0F4F30304F3F3FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF000000000000000000303F5F2F3F502F3F5F2F3F5F303F5F000000303F5F2F3F5F\
303F602F3F5F303F5F2F3F5F7F5F407F50407F5F407F50407F5F407F50407F5F40
0000000000000000000000000000000000000000000000000000000000000000003F5F7F1F2030\
0000000000003F5F7F3F5070000000000000000000000000101F200000001F2F30
3F5F7010101F3F5F703F5F7F3F5F7F40607F3F5F7F40607F3F5F7F40607F80605F8F6F608F605F\
2F2020000000000000302F208F6F6080605F8F606080605F8F606080605F8F6060
40607F4F6F8040607F4F6F8000000F3F5060406F7F4F6F804F6F7F4F70804F6F7F4F6F804F6F80\
4F70804F6F7F0000000000000000004F6F805070804F6F80507080000000000000
0000000000000000000000000000000000000000000000000000000000000000000000004F3030\
7F505F7F50507F505F0000000000004F6F704F70704F70704F70704F6F7000000F
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
000000000000201F1F0000000000000000000F0F0F1F1010000000403F3FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF000000000000303F5F2F3F5F303F5F2F3F5F0000002F3F5F000000\
2F3F5F3040602F3F5F3040602F3F5F3040607F5F40805F407F5F40805F407F5F40
7F5F407F50400F0F0F0000000000000000000000000000000000000000000000000000000F0F0F\
3F50700F0F0F000000101F203F5F70000000000000000000000000101F20000F0F
0F0F0F3F5F70304F603F5F7F40607F3F5F7F40607F3F5F7F40607F3F5F7F40608040607F40607F\
40607F40608040607F40608040607F406F8040607F406F8040607F4F6F80406F7F
4F6F80406F7F4F6F80304F5F0F0F0F0F101F2F3F40000F0F40607F4F6F804F70804F6F804F7080\
4F6F805070804F6F800000000000000000000000000000004F7080000000000000
00000000000000000000000000000000000000000000000000000000000060404F7F505F805F5F\
7F505F805F5F0000000000004F70704F707F4F70704F707F4F70704F707F4F7070
4F7070000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000
0000000000000000004F30300000000000004F3F3F4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0000000000000202F3F2F3F502F3F5F2F3F5F1F20300F101F\
1F2030000000303F5F2F3F5F303F5F2F3F5F3040607F50407F5F407F50407F5F40
7F50407F5F407F5040000000000000000000000000000000000000000000000000000000000000\
0000003F50700000000000000000001F2F3F3F5F7F0000000000000000000F0F0F
000000000F0F203F4F405F7F3F5F703F5F7F3F5F7F40607F3F5F7F40607F3F5F7F40607F3F5F7F\
40607F3F607F40607F3F607F40607F40607F40607F40607F1F2F302F3F4F000000
0000000000003F5F701F2F300000000F0F0F00000F0F0F0F0000000F0F0F4F6F7F4F6F804F6F7F\
4F6F804F6F7F4F70804F6F7F000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000805F5F7F505F805F5F7F5050\
7F505F000000000000406F704F70704F6F700000002030304F70704F6F704F7070
406F6F4F6F70000000101F2040606F406F6F0000000F0F0F101F1F000000000000000000000000\
000000C0C0C0CFCFCFC0C0C0CFCFCFC0C0C000000000000000000000000050303F
00000050303F0000000F0F0F0000000F0000403030CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000101F2F303F5F2F3F5F30405F\
00000030405F2F3F5F0F10102F3F5F3040602F3F5F3040602F3F5F805F407F5F40
805F4F7F5F407F5F407F5040000000000000000000000000000000000000000000000000000000\
0000000000003F50700000000000000000000000003F5F70101F2F000000000000
000000000F0F0F0F0F0F101F40607F3F5F7F40607F3F5F7F40607F3F5F7F40607F3F5F7F2F3F4F\
40607F40607F40607F40608030405F000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000704F50805F5F7F505F805F5F7F505F805F5F\
000000000000406F704F70704F6F704F7070406F700000000000004F7070406F70
4F70704F6F7040606F40606F40606040606F1F2F2F0F1010203F3F101F1F000000000000000000\
CFCFCFCFCFCFCFCFCF0000000000000000000000000000005030300000000F0F0F
0000000000003F2F2F4F30304F3F3F4F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF0000000000000000002F3F5F2F3F5F\
303F5F1010200000002F3F5F303F5F000000303F5F2F3F5F3040602F3F5F7F5F40
7F5F407F5F407F50407F5F403F2F20000000000000000000000000000000000000000000000000\
0000000000000000003F50700000000000000000000000000F0F10304F60000000
0000000000003F5F703F5F7F3F5F70405F7F3F5F703F5F7F3F5F7F40607F3F5F701F2F30000000\
40607F3F5F7F40607F3F607F000000000000000000000000000000000000000000
203F402F404F40607F1F2F30304F604F6F8040607F4F6F7F406F7F4F6F80406F7F4F6F800F1010\
000000000000000000000000000000805F5F805F5F503F3F0F0F0F000000000000
0000000000000000000000000000000000004F3030805F5F7F505F805F5F7F505F80505F7F505F\
000000406F6F4F6F70406F704F6F70406F6F4F6F70406F6F0000003F5F604F6F70
406F6F4F6F703F5F5F40606F304F50406F6F3F505F2F3F3F2F4040406F6F000000000000CFCFCF\
0000000000000000000000000000000000005030300000000000004F2F30000000
0000000F0F0F0000004F3F3FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000000F0F10\
2F3F5F30405F2F3F5F000000202F403040602F3F5F202F40202F403040602F3F5F
3040607F5F40805F4F7F5F40805F400000000000000000000000000000007F5F407F5040000000\
0000000000000000000000003F5F700000000000000000000000000000003F5F7F
0000000000000000003F5F7F405F7F3F5F7F40607F3F5F7F40607F3F5F7F40607F000000000000\
3F5F7F40607F40607F40607F00000040607F405F70000000000000000000304050
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000008F6060805F5F8F6060805F5F806060805F5F8F6060805F5F
0F0F0F0000007F5F5F7F5050805F5F805F5F805F60805F5F805F5F7F505F805F607F505F4F6F70\
0F10101F2020406F704F6F70406F704F6F70406F704F6F70102020304F50406F6F
4F6F70406F6F40606040606F1F20202F3F3F406F6F40606F000000000000000000000000000000\
0000000000000000000F0F0F402F2F000000100F0F503F3F0000000000003F2F2F
503F3F4F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000\
0000002F3F50303F5F2F3F502F3F5F00000030405F2F3F5F303F5F2F3F5F0F0F0F
0000000000002F3F5F805F407F50407F5F400000000000000000000000000000007F5F40705040\
7F50407F5040000000000000000000304F6F000000000000000000000000000000
3F5F700000000000000000003F5F703F5F7F3F5F70405F7F3F5F70405F7F000000000000000000\
40607F3F5F7F40607F2F40500000003F607F4000003F607F0F101F3F5F7040607F
40607F40607F40607F406F7F40607F000000000000000000000000000000000000000000000000\
000000805F5F7F5F506F4F4F2F1F1F302020806060805F5F805F5F805F5F805F60
805F5F805F5F805F5F805F5F7F505F805F5F80505F805F5F7F505F805F5F7F505F406F7040606F\
3F606000000000000040606F406F70406F6F406F7040606F0000000F0F0F406F70
40606F406F6F3F5F6040606F3F6060000000000000000000000000000000000000000000000000\
0000002F1F1F1F101000000050303F00000000000000000050303F4F3030503F3F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
000000000000000000303F5F2F3F5F30405F2F3F5F101F2F0F101F30405F2F3F5F
3040602F3F5F3040602F3F5F202F3F7F5F40805F400000000000000000000000007F50407F5F40\
7F50407F5F407F50407F5F407F50400000001F2030000000000000000000000000
0000003F5F700000000000000000003F5F7040607F3F5F7040607F3F5F70000000000000000000\
3F607F40607F3F5F7F40607F00000000000040607F40607F000000000000000000
0000000000000000000000000000000000000000000000000F1010304F5F4F6F8080605F8F6060\
8F60608F6060805F5F8F6060805F5F8F6060805F5F8F6060CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF406F7040606F406F70\
406F6F305050000000406F70406F6F406F70406F6F304F4F000000406F70406F6F
406F70406F6F101F1F0000000000000000000000000000000000000000000000000000005F3F3F\
000000402F2F503030000000000000000000503030503F3F50303FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C00000000000000000000000002F3F5F2F3F5F30405F2F3F5F000000
2F304F30405F2F3F5F30405F2F3F5F3040602F3F5F304060604F3F0000000000007050407F5040\
7F5F407F50407F50407F50407F5040000000000000000000000000000000000000
0000000000002F3F50000000000000000000000000000F0F2F405F000000000000000000000000\
40607F3F5F7040607F3F5F70000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF3F606F406F703F606F406F6F3F606F\
406F703F606F0000003F606F406F6F3F606F304F4F000000406F6F3F606F406F6F
40606F406F6F0000000000000000000000000000000000000000000000003F202040202F5F3030\
0000000000000000000000000F0F0F50303050303FCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000202F402F3F5030405F2F3F5F
30405F2F3F5F2F304F0000003040602F3F5F3040602F405F3040602F405F1F10100F0000805F40\
7F5040805F407F50407F5F407F50405F4030000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
3F5F7040607F40607F40607F000000000000000000000000000000000000000000
000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF406F703F606F406F703F606F406F7040606F406F70\
3F606F0000000F1010406F6F3F606F0F0F0F000000406F6F40606F406F6F40606F
406F6F00000000000000000000000000000040202F2F1F1F0F0F0F000000000000000000000000\
0000000000005030305F3F3F503030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C00000000000000000002F3F5010101F
2F3F502F3F502F3F5F30405F2F3F5F2F3F5F0000000F0F100000002F3F502F3F5F3040607F5040\
7F5F407F50407F5F407F50407F50407F50400000000000000000006F4F3F7F5040
704F3F7F504030201F000000000000000000000000000000000000000000000000000000000000\
40607F3F5F7040607F3F5F70000000000000000000000000000000000000000000
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF3F606F3F606F3F606F40606F3F606F3F606F3F606F40606F3F606F\
2F4F4F101F1F40606F3F606F0000001F303040606F3F606F40606F3F606F40606F
0000000000000000000000005F0000502F2F0000000000000000000000000000000000003F2020\
5F303F5030305F3F3FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF0000000000002F3F50
1F1F2F2F3F500F0F0F2F3F5030405F2F405F30405F2F3F5F3040602F405F3040602F3F50304060\
2F405F3040607F5040805F4F7F50407F5F407F50400F0F0F0000003F2F207F5040
7F00007F00007F50407F5040000000000000000000000000000000000000000000000000000000\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
0000000000003F606F3F606F40606F3F606F3F606F3F606F40606F3F606F40606F3F606F2F404F\
1F202F40606F3F606F304F4F3F606F40606F3F606F40606F3F606F203030000000
000000000000600F0F5F30300F0F0F0000000000000000000000002F1F1F5F3F3F5F30305F3F3F\
5F303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0000000
0F101F0F0F0F2F3F500000002F3F500000002F3F500F101F00000030405F2F3F5F30405F2F405F\
30405F2F3F5F30405F2F405F3040607F50407F50407F50407F50400000000F0F0F
7F50407F4F407040306F403F0000003F2F20704F3F5F3F30704F40402F200000000F0F0F000000\
000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006F403F6F403F704F40201010000000000000000000000000000000000000\
000000000000000000000000000000000000000000100F0F0000000000003F606F
3F5F6F3F606F305F603F606F3F5F6F3F606F3F5F603F5F602F404F30505F3F5F6030505F203F40\
3F606F3F5F603F606F3F60603F606F3F5F603F606F3F6060000000000000000000
0000000F0F0F0000000000000000000000000000005F303F5F30305F3F3F503030CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF000000000000000000202F400F0F0F2F3F50000F0F000000000000000000000000\
0000000000000000000000002F3F50304F602F405F304F607F5040805F4F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
402F20201F10704F3F7F4F40704F407F4F40704F407F4F4070403F7F4F406F403F
1F1010000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000F0F0F6F3F3F402F20000000000000000000000000000000
0000003F5F6F3F606F3F5F6F3F606F3F5F6F3F606F000000000F0F005F6F00606F005F60000000\
3F5F603F606F3F606F3F606F3F6060304F4F000000000000000000000000000000
0000000000000F0F0F4F2F2F603F3F5F3030603F3F5F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000001F20303040602F405F30405F2F405F000000
0000001F1010000000000000000000000000000000000000000000000000201F1030201F704F3F\
7F4F4050302F000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000050302F7040406F403F\
402F2F4020200F0F0F000000305F6F305F603F5F6F000000000000000000000000
0000000000000000000000000000003F5F6F305F603F606F305F603F5F6F3F5F603F606F305F60\
3F606F3F5F60203F4000000000000000000000000000000060303F603030603F3F
5F303060303F5F303060303F5F3030CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000\
0000000000000000000000000000000000002F405F304F602F405F304F60000000
000000000000000000000000304F602F4F60304F60000000000000000000000000000000000000\
0000000000000000004F302F30201F7F4F40704F3F7F4F40704F405F3F30704F3F
7F4F40704F40704F4070403F503F3070403F503030000000000000000000000000000000305F6F\
305F60305F6F3050603F5F6F305F603F5F6F305F603F5F6F305F603F5F6F000000
000000000000000000000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000000000000000006030306F3F3F60303F603F3F603030
603F3F603030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF000000000000000000000000\
0000000000000000000000000000000000000000002F405F30405F2F405F000000
0000000000000000000000001F2F30304F602F405F304F602F405F304F602F405F304F602F4F5F\
304F60000000000000000000000000000000000000000000000000000000000000
0000000000000000000000002F50603050602F50603050602F50603050602F5060305060305060\
305F602F5060000000000000000000000000000000305060305F6F305060305F60
305F603F5F6F0F1F1F000000000000000000000000000000000000000000000000000000000000\
0000000000000000000000006F3F3F60303060303F603030603F3F603030CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000\
000000000000000000000000000000000000000000000F0F2F405F304F602F405F
0000000000000000000000000000002F405F304F602F4F5F304F602F405F304F602F4F5F304F60\
2F4F5F304F602F4F60304F602F4F603050602F4F603050602F4F60305060304F60
305060000000203F403050603050602F5060305060305060305060102F301F2F30305060305F60\
305060305F60000000000000000000000000000000000000000000000000000000
000F0F305F603F5F6F305F603F5F6F305F603F5F6F000000000000000000000000000000000000\
0000006F3F3F60303F6F3F3F60303FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF000000000000000000000000000000000000202F4030405F
00000000000000000000000000000030405F2F405F304F5F000000000000000000304F602F405F\
304F602F4F5F304F602F4F5F304F602F4F5F304F602F4F5F304F602F405F000000
0000002F4F5F2F4F5F3050602F4F5F3050602F4F602F4F500000002F40502F50603050602F5060\
305060000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000603030\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000
10101F0000000000000000000000000000000000002F405F10101F00000000405F000000304F60\
1F2F3F10101F0000000000000F0F0F0000000000001020202F4F5F3050602F4F5F
3050602F4F603050602F4F603050602F40500F0F0F20404F3050603050603050602F50602F404F\
000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF000000000000
00000000000000000000000000000000000000000000000000000030405F2F405F304F5F2F405F\
304F5F2F405F304F5F2F405F000000000000000000000000000000000000000000
0000000000000F10101F2F30203F4F3050602F4F5F3050602F4F5F2F4F5F2F4F5F2F4F5F2F4F5F\
0F0F1000000020303F000000000000000000000000000000000000000000000000
000000000000000000000000CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF000000000000000000000000000000000000000000000000304F5F\
2F405F304F602F4F5F304F5F2F4F5F304F602F4F5F304F602F4F5F304F602F4F5F
304F602F4F5F3050602F4F5F3050602F4F5F2F3F4F2F4F5F304F5F00000020303F20404F304F5F\
203F4F1F2F30000000000000000000000000000000000000000000000000000000
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF000000000000000000\
0000000000000000000000000000000000000000000000000000000F101F0F1010
00000000000000000000000F0000000000000000000000000000000000000000002F404F10202F\
0F0F0F000000000000000000000000000000CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF000000
000000000000000000000000000000000000000000000000000000CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[28]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7
[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7
[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl7
20L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7
ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P7
00071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7NJl00g5[RP0O[`0306^_07B_20L000L7
09j_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_RW40
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`80076_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PM^[`03Jg4003:_00<0Jjl0
K:l81`001`L0PZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[g6_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`L80027L@N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`2:[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061fN_00=a[jl03:lD0009
CWMaMg5bLG8001V_00<0[jl0Hjl81`001`L0LJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[g6_1jl7[`N_1jl7[`N_01P000MgLGMaMg5g00P007V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
00N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061f:_00=[0:l02JlI000:
LF^:JhY[RW2:L0`0MJl81`001`L0JJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_0:l7
[`N_1jl7[`0L009a00Z:LHYaRW6:LHYa2@000f/00002001b1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7GJl00g6_[`06[`03L@0001800V/01HY[
RV^:00@000Y[RV^:JhY[RW1a20002G2:JhY`RV^:L00406f_20L000L706>_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[g6_1jl7[`N_1jl010001g9aLG5gLGL02P001f^:JhA[RV/00`0036^:LHYaRW6:
LHYaRP80015`RW6:LHYaRW6:LHYaRW6:L@02RP03L@0000@006N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061eR_00=[[jl01jlE0003L00000D002V:L8Y[RV^:JhY[
RV^:JhY`RW2:L8Y`RW2:JhY`RW2:L8Y`RW2:L8Y`L@040003L00006>_20L000L705b_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7L@N_1jl7[`N_1`l000F:L@00L@07009a00>:LHX010002digLGMaMg5gLGMa0080
015bLGMaMg5gLGMaMg5gLGMaM`0400::00EaRXZ:L@02RPT005f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl020L000H7Fjl:009[00>:00000`000hX000040005JhY[RV/00P001g9aLW5bLG80
0g406G9aLG5bLG9aLW5bLG9aLW5bLG9aLW5bLG800W440003LDh0008000N:L8Y`RP1[00@000=`0000
Fjl81`001`L0F:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L3009a00=g00000P001g40001aRV/0
1P02L@0_RW6:LGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGMaMg5gLGL010000g5g
00050003L@2:00::1000G8X0Jjl7[`N_1jl7[g6_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7F:l80005
J`0006/0200018Y[RV/8L@03LW5b00f_0W404W9aLG5bLG9aLW5bLG9aLW5bL@X000=a00000`000g40
00020004[hX0RPH02:l00aR_[`1;[`P700071`1E[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`06009a
10000g5[0003009a00j:LHYaMg5gLGMa1jl7[aH001@7[gMaMg5gLGMaMg5bLGMaLW5gL@D000=>L@00
10000gLH000300031jn:00<005R:00001jl7[`N_1`07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7Djl60003
L@0000@000=[00000P002HY[RV/0LG9aLP02L@:_7`04[`07LTibLG9aLP02L@X000<HLP000`02L@D0
0jl00g2:00040003RP2_00B_00<H0:l0BJl81`001`L0DJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0[`N_1jl7[`040003
RP0000@000=a00000`000g6:L@020005Mg5gL@L03`03LAL000L7[g9aLW5b009a00@HL@1>1P000g5g
000200=a1P001:l7[`L2RP8005=`06/7[`N_1jma[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PM@[`X000=[0000
0`000f]aJ`02L@03075b00>_3P001i1aMg6@LGL00W4G0005[g5bLG800W4014ibLG86009a0`000aQa
000600J_00=`RP001004[`03L@2_04J_20L000L704j_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L90003RP000080
00]aRW6:LGMaMg47[`0400=a1`003gMbT76@LI1aT76@LI1bT00F0004[`MaLP9a2P000g5g0002009a
2000F0N_1jl7[hZ906/0RP2_1jl7L@N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PM>[`L000=a00000`00175[LF/5L@03
[`0000@00W400gMaM`05000AMg5gLGMaT75gLGMaMg5gLGL01`000g40000>00:_00AaLW5b0W400aQ>
000400046000009a2007[`03L8X0008000>:0:l00Zl00aR_[`13[`P700071`1;[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00<000=[00001P002F]aRW5gLGMa1`090005T76@LI000g402I1aT76@LI0HL@030005T76@LY00
0P000g6@000500036900018000L7LG9aLW5b00D00W401G80075g00P005F_1jl7[`N_1jl7[hX0RP07
[fb_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061dZ_0`000g4000070008LF]gLGMaMjl>00Aa00F@LI1aT005000:675gLGMa
M`1D009a00=g00000`000i1a000D0006[g9aLW5b1@03L@<000=b00001P0<[`0308X000:_00=a0:l0
@:l81`001`L0BZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`1a1P0016/0L@03L@05Mg5g[`L02`000g9aT003L@05T76@LI00
0W420006LGMaT7400W420007L@1>076@L@050003LI1a00h00W400`1aT00300061g5bLG9a1P001GM>
075b00T000`7[`N_1jl7[`N_1g0200141jl760N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLn[`030:n_00J_00<0J`0010002f]a
0740LF]gLG9a00:_40001g9aMg6@LDh00P0014j@LI02L@@000A>Mg5g0`000gMaM`050003Mg5g00`0
00Aa001g0P000g4000020006[g5bLG9a1P001G8H075>00T03:l300:_00<H0:l0?Jl81`001`L0AJl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`050005K01/L@000W401gMaMg5gLGL03P00276@LGMaMg6@0W402@1aLI1aT75b0002L@05
T76@LI000W401@2@LI1a00D000=aT7802P000gL0M`020004LG80T0H000F_LW5bL@090003L@0000P0
00`7[`N_1jl7[`N_1jl30011[`MaLJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7A:l5000>LF/0Jg40L@1gLGMaLW42[a000W4017Ma
Mg42000GT76@LI1aM`1gLGMaMg5gLGMaMg5gLAP01@000gMaM`09009a00P0CWL0Mg40L@L00Zl0175b
LDhB00f_00N:0000[aQa03b_20L000L704>_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00<0011/LF`0K75/L@1gLGMaMg474`02L@0:
T740LI1aT76@L@8000n@LI1aT76@LI1aT76@LI001@000i1aM`0;000:T76@LI1aT76@L@L000B_1g5g
0W480003L@0000L000n_1jl7[`N_1jl7[`N_1g000P00?PMaL@07[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7@Zl3000@LF]aJg5[LF]gL@1a
Mg5b[aD001eaMg5gLI1aT76@L@1>Mg5gLGMaMg5gLGMaMg5gL@050004LGMaL@L000/H07MaMg6@LGMa
M`02L@T000F_LG9aLP06009a00A[LW5a1P0>[`03L00000<0>Zl81`001`L0?jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`03000AJg5/LFaa
K75/T76@LGMaMjl05@007W6@LGL0T76@LI1a076@LI1aT76@LGMaT75gLI1aT0@000EaT76@L@04000C
Mg6@LY1aT76@LI1aT76@LY1bT0060007[`MaMg5gCP040006LV][LG9a1`00CJl7[`N_1jl7[`N_1jl7
[hX0L007[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00P700061cn_0P00275[LF]aJg5[0W401gMaMg5gLJl05P007g5gLGM>076@LGLH075gLGMaMg5g
LGMaMg5gLGMaMg4010007WMaMg5gLG9aMg5gLGMaMg5gLGMaT75gLGMaMg5gL@H000J_LG9aLW440003
LW5b009a2@0?[`03JhX003V_20L000L703b_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L300Ea00=/LF`00W401i1aT76@L@L02000176@074;0003
T76@008002]aT76@L@1aT76@L@N_1jl7[`N_1jl7[`N_1jl7[`1aT75gLGMaMg6@LWMa00T000YaT7:@
LY1bT7:@1@0020MaMg5gLGMa3P00C76_1jl7[`N_1jl7[`N_1jl7LHX01jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLj[`<000UaJg5[LF]aJg40
0P000g5gL@03L@03[`0000X000N@LGMaT75g00D000>@LGL00P000g5gL@0S[`0ALGMaMg5gLGMaMg5g
LGMaMg400P02L@06Mg6@LGMa1@005jmbLG9aLW5bLG9aLW5bLG9aLW5bLG9a016_00=`0:l0=Jl81`00
1`L0>:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
0`002W5/LFaaK75/RW420007T76@LI2_1`0=0019T7:@LI1bT740LI1aT76@LI2_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jn@LI1aT7:@LY1bT7:@LY1bT0020007691bT7:@LP05001O
[`MaLW5gLGMaMg5gLGMaMg5gLGMbMg87[`N_1jl7[`N_1jl7[`N_1g40[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7>Jl20008LF]aJg5[LF/2L@80
1G400jl0000?000=LGMaMg5gLI1aMg6@L@0/[`05Mg5bLG800W40201a679aMg5g0`001WMaT75gL@D0
0Zl0575bLG9aLW5bLG9aLW5bLG9aLW5b4Zl00g00[`0c[`P700071`0g[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02000:LFaaK75/LFb@L@<000EaT76@
[`0B000kLG:@LI1aT76@L@N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[g400`03L@04LWMaL@8000YaT79aLY1bT7:@1P009jl7LGMaMg5gLGMaMg5gLGMbMg5gLWN_1jl7
[`N_1jl7[`N_1jl7[`03L@0c[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061cN_0`001f]aJg5[LF/01@03L@03MZl000D000N@LG5aT76@00P000IaMg5g
LI0d[`08LG9aMg5gLGL2L@<00W401gMaMg5gLGL01P002ZmbLG9aLTib07420003L@0H00L04Jl00g5`
RP0b[`P700071`0d[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1`@000Q/LFaaK740L@@000EaT76@[`05000:LI1aT76@LI1aT0P004AaT747[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LY1bT7:@LY1bT09a00X0LG6@
LY1bT79a1P001:mgLGL2LA00016_1jl7[`N_1jl7[`N_1jl7[`02L@0bRZl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLd[`D000MaJg5[L@1f00D000AfLGJ_
1P00191aT7430004LI1aT0P000=aMjl0>:l02g9aMg5gLGMaMg5g00800W40101gLGL80007LG9aLW5b
L@0>01:_00=a08X0<:l81`001`L0<Zl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl70g430003K75/008000>@00000`00191aT:l50004LI1aT09a00<0E0000P00191a
T746001>1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1g:@LY1aT7:@LI1b04iDLI1a20002@MaMg5gLGMaM`02L@X004F_1jl7[`N_1jl7[`N_1jl7
[`N_L@07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7
<Zl00`1aJ`030003Jg5[008000=g000010000g5f[`040008CW6@LI1aT002CP08E000LI1aT743042_
00maMg5gLGMaMdh0CU@0Mg402@003:mbLG9aLW5bLG9aLPAa1P0B[`8000=`[jl0;Jl81`001`L0<:l7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`Aa0P02L@06K000T01a
10000g6@[`040009LGN@Mi1aT76@00@0059aT76@LI1a0:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g:@LI1aT79a10000g:@0009000@
[gMaMg5gLGMaMg5gLGMbM`9a1000@jl7[`N_1jl7[`N_1jl7[`N_1g40L0N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7;jl01G40L@1a00<000MaJ`1a04ia00<0
00=gLGL01@00291aT76@LI1a0P002I1aT76@LI1aT00Y[cD00Zl04G5bLG9aLW5bLG9aLW5bLG9a00<0
4Zl402b_20L000L703Z_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
K75/LF`0K75/07400W430004T76@[`@000N@LI1aT74000=a02f@LI1aT76@LI2_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L0=0000g6@0005000@[`MaLW5bLGMaMg5gLGMaM`9a0`004`N_
1jl7[`N_1jl7[`N_1jl7[g401000:jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl020L000H7;Zl02@1a0740L@1aJ`040004M`00009a00=g[jl00`001Y1aT76@L@@000V@LI1a
T76@LI007jm200:_00mbLG9aLW5bLG9aLW5bLG801@0A[`0660000:l0:Zl81`001`L0;:l7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl70P001VaaK75/L@<000>@09001000176@
[`L30006CW6@LI1a0P009i00LG6@LI1aT76@[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`17000@
1jmbLG9aMg5gLGMaMg5gL@D001>_1jl7[`N_1jl7[`N_1jl7[`Ma00<002^_RZl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061bb_0`001G40L@1a00<000>@07L01000175g
LJl40004LI1aT09a00`0T740LI1aT76@LI0J[cl000=a00002002[`0=LG9aLW5bLG9aLW5b600701:_
00<H00000Zl00hZ_[`0V[`P700071`0Z[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7100016aaK7430004T000L@<000AaT7471000:W6@LI1aT02@LI1aT76@LI1aT:l7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1cl000EaRZV?Z@0>0008LGMaMg5gLGL2L@@00W405007[`N_1jl7[`N_1jl7
[`N_1jma0`00:PN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061bZ_
1@000g40L@030003T00000D000AgLJn_10001I1aT76@008000YaT02@LI1aT76@5jm00007L@00L92:
S`0>0009LG9a075bLG9>00<000=aLP000P0B[`04L@0002V_20L000L702R_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl71@000g5/L@030005T00009000P001G6@LI2_00D000AaT76@
0P007g6@L@1aT76@LI2_1jl7[`N_1jl7[`N_1jl7[`N_1jl0@`0028n@002?ZHn@40000g400002L@@0
00=gL@000P004Zl7[`N_1jl7[`N_1jl7[`N_1`9a0P00:Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_00P700061bR_1@000g40L@030005T00009000`0017MaMjl50005LI1aT740
0P002I1aT02@LI1aT00C[d@000][092:T6^?RY2:T00A0003600000<000=bL@00100B[`04674002R_
20L000L702J_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`D000AaK76_0P001I00
001a00<000B@LI2_1P00891aT76@091aT740LI1aT:l7[`N_1jl7[`N_1jl7[`N_:0000g6:000K000;
T000RZV?RP1aRZT00XlF009a1@0050N_1jl7[`N_1jl7[`N_1jl7[`Ma0P00:0N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLV[`<000Ea0000L@080008L@00T75gLJl4000:
LI1aT76@04iaT09a00D0T76@L@0?[aD000=[00004`000i2:000K0007T740Ji2:L0020004RY2:T1D0
00=b0000100C[`04LAQa02N_20L000L702B_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`L20008K740LFaa[`L50003T000008000=aT:l00P02L@07T75aLI1gT002L@0EMi1fT740LI2_1jl7
[`N_1jl7[`N_01000g43T003RY2@00>@00BYT:V;3`001I2YT:Ua01H000J:0000ZHl20007T8nY076?
Z@02S`L00W4C000m[`N_1jl7[`N_1jl7[`N_1jl7[g80L@07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`081`001PLT[`<000E[0000L@02[`03L@00008000>@00000`00191a[`03L@80
00YaT76@L@1aT76@0W400`2@[`0;[a40019aRY2:T8Z@RY2:T8Z@RY2:T8X>00::00B@RY1a5002RP05
T00009000`002I2:ShZ@RY2:T002L@8000EaT8Z:S`0C01>_00Aa67409Jl81`001`L08jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_008000QaK01/LJl7[`@000>@000010000g470002000L
L@2@LI1aT7N@091fT7N@L@1a1jl7[`N_1jl7[`h000Ia075[T8/<T007ZI2YT:V@RP0=0006ZI2YT:V:
50008:V?ZHX0SjV:T8nYSjV?ZHn@RZUaT8nYSjV?ZHnY075[3@00?G47[`N_1jl7[`N_1jl7[`N_1jl7
L@1>LJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H79:l50004LJn_[`@0
00=a00000P000g400002[`@0012@L@00T76@L@1aT76@LI002Zl@000IT8Z@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:R`0<0004RY2:T0::5@008I2:T02@RY2:T8Z@RXn:ShZYRZV:ZHZYRZV?T8n@Si1[L@0<01>_
00Aa674H9Jl81`001`L08Zl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L50005LJl7[`L0
0`000i0000030005L@1a[`L00`006I00076@LI1gT02@Mi1fT7Ia[`N_1jl7[`L03`00292;T8^@Ri2;
39000jV@Z@03T003ZI2@00/000FYT:V@Z@02T1@002A`ZHn@SjV?ZHnYRZV:ZI2YSjV?ZHnYSjV?ZHnY
SjV?ZHn:LHX;000j1jl7[`N_1jl7[`N_1jl7[`N_LP1b00N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`P700061b>_10000g40[`02[`<000=fL@000P001W40001a[`H000MaT02@LI1a0080
00F@LI1aT007[`l001V:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY1a00::00B@RY2:2`0028Z@RY2:T8Z?
4`009G6:T8Z@RY2:T8n@RY2?T8ZYRZV:ZHZYRZV?T8n@Si2?T8n@L7402P0B[`04674HLBB_20L000L7
026_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl010001W5/1jl7[`<000=a00000P001WH0
LG6@[`@00g404`2@LI1gT01aMY1gT7H7[`N_1jl03P02L@0:T8^@Ri2;T8^@R`f@00E[0000J`02T003
ZI0000T000N?T:V@ZI2Y00:@4`009G6:ZHn@SjV@ZI2YT:V@ZHnYSjV?ZHnYSjV?ZHnYSjV?ZHnYSg00
2P004Jl7[`N_1jl7[`N_1jl7[`N_00Aa02@7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl81`001PLQ[`H01:l;0003LGJ_00P000b@LGMaT76@091aT745[a0001V:T8Z@RY2:T8Z@RY2:T8Z@
RY2:T8Z@RY2:00H000B@RY2:2P001i2:T8Z@RY000XXC000UJhZ@Si2?T8n@Si2?T8ZYRZV:ZHZYRZV?
T8n@Si2?T8n@Si2:T00;012_00Aa67408jl81`001`L08:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl70P0026`0K0N_1jl70`001Y00002@009a00>@L@L01P000g6@0002000<T7N@MP1aT7J@[`N_
4@003G6@Ri2;T8^@Ri2;T8/039000jT00002E08000EaZI2YT0090009RZV@ZI2YT:Ua018002F:SjV@
ZI2YT:V@ZI2YT:V?ZHnYSjV?ZHnYSjV?ZHnYSjV?ZHnY00:?00=a00002000=PN_1jl7[`N_1jl7[`N_
1`1bL@2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061b2_1@000g6_[`03[`80
00F@001a0002L@03091a00:_2`00276@LI00MW6@0jl40003L@0000/001f@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RY2:T8Z@J`1>E002CP8000B:T8Z@2@002G6:T8Z@RY2:T00A00::02NYRY2?T8n@Si2?T8n@RY1a
ZHZYRXm[ZHZ@Si2?T8n@Si2?ZHZ@RY000V/:00f_00Aa67408jl81`001`L07Zl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1`D000Ia[`N_1jl2000=L@00LI00T76@LI2_1`0;0009T7N@Mi00T7H700D0
0W40191aT7420003L@00008001=aT8^@Ri2;T8^@Ri2;T8^@LHYa00R@00=a04h00U@024h00:V@ZI1a
2@002:V@ZI2YT:V@3P004G6:ZI2YT:V@ZI2YT:V@ZHZY009a0P02L@<001EaJi2:LHnYSjV?ZHnYSjV?
ZHnYSg002P003Zl7[`N_1jl7[`N_1g400W408`07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00P700061aj_0`000g40L@05[`8000]a0000T76@LI1aT002[`d000F@LI1aT007000;RW6;LI1a
T76@LI000W4048Z@RY2:T8Z@RY2:T01a0743000:L@2@RY2:T76@RP800dh30004T8Z@RPX000N@RY2:
T8Z@00d000eaRZV?ZHnYSjV:ZHn@008000E[0000J`020007L76@Sg1aRP05000?LG2:SjV?T8n@Si2?
T8n@00`02Zl011Qa6005[`030:n_01Z_20L000L701b_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
1P001V`7[`N_1`D000V@LI1aT76@L@L03P00176@LI0;000ELH^@Ri2:T8Z@LI2;T8^@Ri2;T8^:009a
0`001W40L@2@R`:@0`02T0L000EaT:V@Z@0:0008JjV@ZI2YT74;000;RY2YT:V@ZI2YT:T00W4500:?
00aaSjV@ZI2YT:V?ZHl2000AL@0008maZHnYSjV?ZHnYSjT02`002PN_1jl7[`N_1jl2L@0TLP00[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H77:l700J_1002L@06T76@LI1a0Zl?0003
L@0000P000=a00001000476@LI1aT8Z@RY2:T8Z@RY070003T8Z@00::00Q[L@1/08^:L@H000J:T8Z@
RY0:0007RY2:T8Z@L@0:000;L8nYSjV?ZHnYSi000`005g4008ZYSjV?ZHnYSjV?ZHnYSjV?Z@1a00@0
00b?T8n@Si2?T8n@Si0<00V_00Aa07408Jl81`001`L06jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`0300096000LJl7[`N_00@00W40291aT76@LI2_3`000g4000080005T8^@Rg40100038^@LI2;T8^@
Ri2;T09a1@00276@T8^@Ri2:10001:V@T8X60006Ri2YT:V;2P002:V@ZI2YT:Ua2@002Y2YT:V@ZI2Y
T:T3000XL@2?SjV?ZHnYT:V@ZI2YT:V@ZHnYSjV?Z@1a0:V?ZHnYSjV?ZHnYS`d002f_1jl7[`N_1`1b
L@2_1jl7[`2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H76jl20005L@0007401Zl40008
LI1aT76@LI02[a/001>;LI1aT740LI1aRhZ@RY2:T8Z@009a100028Z@RY2:T8Ya0`02L@06T8Z@RY2:
1@001W6@RY2:T0X000R:T8Z@Si2:Z@P000^:ZHnYSjV?ZHnYL@02000YZF^@SjV?ZHnYSjV?ZHnYSjV?
ZHnYSjV?ZHnYSg6?T8n@Si2?T8n@Si003P07[`05LAQa60008:l81`001`L06Jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl00P003740001a[`N_1jl7[`800W402Y1aT76@LI1a1jlJ0007L@2;Ri2;T002R`0;
RY2;T8^@Ri2;T8/00`000g6@RP05T09a0`02L@03ZI2Y00>@00fYT75[LI2YT:V@ZI1a00T000VYT:V@
ZI2YT7401`002G6@ZI2YT:V@Z@02000FLJV?ZI2YT:V?ZHnYT:V@ZI2YT:V@Z@>?00nYRZV?ZHnYSjV?
ZHnYSg4018l?00061jl7[`Ma0P008W801jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`00
1PLJ[`<000Aa07401Zl3000:LI1aT76@LI1aT0:_2`03L@l0019a08]aT76@LI1aT8Z@RY2:J`02L@::
00J@RY2:T744000ET8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00X000Z:T8n@RY2:ZHYa1@002W6:ZHnYSjV?
T6/2000FRXnYSjV?ZHnYSjV?ZHnYSjV?ZHmaJ`8000E[001`L@02S`08LI2?T8n@Si020003LI1a0100
1:l01G4HLAP0022_20L000L701V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_008000]a001/1jl7[`N_
1`02000<LI1aT76@LI1aT:l72P000g6@T00>0003L@0000<00X/02I2:T8^@Ri2;T002L@8000>;T8/0
1I002F/00740RhZ@R`02T003ZI2@00>@00^YT:V@ZI2YT:V@Z@09000GZI2YT:V@ZI2YT:V?ZHnYT:V@
ZI2YT8l00P00576YT:V@ZI2YT:V@ZHnYT:V@ZHZY0W490003L00000800XX00hnY00030003L:T00100
00H7[`MaLP02L@0P0:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLJ[`D01jl3000:
T76@LI1aT76@L@:_2000191aT74>0003Rg6@009a0P002W6@LI1aT8Z@RY02RP8000Z:T8Z@RY2:T8Ya
0`005i2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@00::2@005I2:T8n@RZV:ZHZYRZV:ZHnYSjV?T003000C
T8ZYSjV?ZHnYSjV?ZHnYSg40S`030004LG2?009a0P000g4000040003L92?008000ED002?J`0A00:_
00Ia6000L@04[`030:n_01R_20L000L701V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Ma00<000THLJl7
[`N_1jl00`00376@LI1aT76@LI2_1`H000=aT8l00Y0@000@T76@Ri2;T8^@Ri2;T8^@R`<000U/T8^@
Ri2;T6`00P02L@04T8^@R`:@00>YT:T00i003ZV@ZI2YT:V@ZI2YT:V:2@00:Y2YT:V@ZI2YT:V@ZI2Y
T:V@ZI2YL740LHnYT:V@ZI2YT:V@ZI2YT8X0L0<000F:JjV:Z@03S`08ZG6YSg6?LG02S`8000I>Ddd0
ZHlA0005[`MaLP000g407jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H76:l00g40
000200V_1000272@LI1aT76@0Zl50005T76@LI005P002g6@LI1aT8Z@RY1/00<000QaT8Z@RY2:T0<0
01f@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T009000UT8Z@RY2?T8Z@Si2?T8n@SjV:ZHZYRZV?
ZHZYSjV?ZHnYSjV?Z@04000FJi2?ZHnYSjV?T8n@Si2?T8nYSjV?T08000M>E000Si1`018000Ma674H
LAP001j_20L000L701N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`03000:K76_1jl7[`N_1`L000R@LI1a
T747[a/017402P2@Ri2:T8^@Ri03009a0i001H^@Ri2:00800W4<T00AZI2YT:V@ZI2YT:V@ZI2YT740
20009I2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZHX00`0068nYT:V@ZHnYT:V@ZI2Y
T:V?ZHnYSjV?L@8000I>002YSjT70003Z@0000P000=aLW400W4087801jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_20L000H76:l30003LJn_00N_20001I1aT76@00:_6`000i1aT003L@07T76@LI1a
T002L@8000V:T8Z@RY2:T7400P0088^:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:2000992:
T8ZYRY2?T8n@Si2?T8nYRZV:ZHnYRZV?ZHnYSjV?ZHnYRP8001YaRZV?ZHnYSjV?T8nYSi2?T8n@SjV?
ZHnYRPD000>?T8l01`000hZ@000900066740LAP01Jl00`2_[`0F[`P700071`0F[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1`9a00`0LJl7[`N_1jl7[`L90006T76@L@N_6`003W6@Ri2;T8Z@Ri2;T8]a0P002H^@
Ri2;T8^@R`030003LI2@00b@01:YT:V@ZI2YT:V@ZI2YT:V@ZG47000ST:V_1jl7[`N_1jl7[`N_1jnY
T:V@ZI2YT:V@ZI2YT:V@ZG400P02T00IZI2YT:V@ZI2YT:V@ZI2YT:V@ZHnYSjV?Z@050003ZHnY00L0
00>YSg402P008g40LG801jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061aN_00Aa0740
2Zl90003LI1a00:_7@002VaaLI1aRg6@LI02RP<000N@RY2:T8Z@009a02@0LG6@RY2:T8Z@RY2:T8Z@
RY2:T8Z@RY2:T8Z@RY2:T8Z@RW4301n_00MaZHnYSjUa00<001b?ZHnYSjV?ZHnYSjV?T8nYSi2?T8nY
SjV?ZHn@0`02S`03T8l000H000>:T8X030000g4H000N[`P700071`0F[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1`<000]a[`N_1jl7[`N_1`09009a00>@[`L07`02L@:;00F@Ri2;T0030004T8^@R`>@00>;T000
0P00192;T8/<T00gZI2YT:V@ZI2YT:V@ZI2Y[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7LY1bT005000UL92YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V?ZHnYSjV?L@1[SjV?Z@070003ZHnY00`0
025aLP00[`N_1jl0[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H75Zl400^_0`000g4000040004
T76_[b4000MaT76@LI1a008000UaRY2:T8Z@RW400`006Y2:T8Z@RY2:T8Z@RY2:T8Z@RY2:T8Z@RY2:
5Zl?00F_00F@LI1aT007000RZHnYSjV?ZHnYSjV?ZHn@Si2?T8nYSjV?ZHnYSg6:T8n@S`L000B?T8Z?
30000aP0000M[`P700071`0E[`N_1jl7[`N_1jl7[`N_1jl7[`N_00@00147[`N_1jl7[`N_T76@LI1a
T005L@03T:l7020000MaT8^@Ri2;00<000V;T8^@Ri2;T8/00P00176;T8/>T00DZI2YT:V_1jl7[`N_
1jl7[`N_1jlO0006ZG4008]a30006Y2YT:V@ZI2YT:V@ZI2YSjV?ZHnYSjV?ZHnY0Xl70004ZHnYRPl0
01f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLE[`8000=a07402Zl03G6@LI1aT76@
LI1aT7400ZlQ0004LI1aT09a0P002g6@RY2:T8Z@RW40009a01J@RY2:T8Z@RY2:T8Z@RY2:T8Z@091a
2Zl2T003Z@0001l000>@J`004@006Zn?ZHnYSjV?T8n@SjV?ZHnYSjV?T8n@Si2:1`0018Z@Si0=0003
L@2_00F_00=b[jl05:l81`001`L05:l7[`N_1jl7[`N_1jl7[`N_1jl70`006W4H1jl7[`N_1jl7[i1a
T76@LI1aT76@L@N_80001g6;T8^@Ri000`001Y2;T8^@R`>@0P001g6@Ri2;T8/02i0200B@00f_1jl7
[`N@ZI2YT:V@024000>@[Y0040006`N_T92YT:V@ZI2YT:V?ZHnYSjV?ZHnYSjV?Z@070004ZHnYS`l0
01f_1jl7[`Ma1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLE[`@02jl0376@LI1aT76@LI1aT0:_
8@001Y1aT76@L@8000YaRY2:T8Z@RY1a0P00492:T8Z@RY2;T8Z@Ri2:T8/70005LJn_[`0029000jT0
000B0003LI0000/000>:ZHX03`02[`0HLH]aT8n@SjV?T8nYSjV?ZHnYSjV?T8n@200018n@RY0?00J_
00<0[jl05:l81`001`L05Jl7[`N_1jl7[`N_1jl7[`N_1jl06002000JLJl7[`N_1jl7[`N_T76@LI1a
T76@LI1a1jlP000FT8^@Ri2;T01aRi2;T8^@Ri2;L@2@R`N@00R;T:l7[`N_1`8000>@00001`002G6Z
T:Z@ZY2ZT00E0003ZJh000/000B@[XYa4P005W6@LY1aT8nYSjV?ZHnYSjV?ZHnYSg470003ZHnY00:?
40006jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLC[`@000=a[jl02Zl0376@LI1aT76@
LI1aT0:_7`0028]aT76@LI000W402I1aT76@RY2:T0020008Ri2;T8^@Ri09[``00Y002:V@ZI2YT:V@
50000g6?RP0<0004LJV@RQ@001=aT76@LJV?T8nYSi2?T8n@Si1`00L000F?T8n@L@0@01^_20L000L7
02>_1jl7[`N_1jl7[`N_1jl7[`L0L@1aK0N_1jl7[`N_1jl7[`06009a00F@LI1a1`0Q000DT8^@Ri2;
L@2@Ri2;T8^@Ri2;08X8T0071jl7[`N_1`0>000:T:Z@ZY2ZT:Z@ZQH000>Y00002`000hn^000G000?
LI1bT7:YSjV?ZHnYSjV?00P000FYSjV?T00@000K[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P7
00061a:_1@0=[`P000AaT76@8@02L@0KT76@LX/0T76@LI1aT76@RY00T8^@Ri2;T76@00F_4002T008
ZI2YT:V@ZI0800:@00>YT:T02@000jU`000;00:@00>YL0005`003I1aT76@Si2?T8n@Si0020001Hn@
Si2?01006jl81`001`L04jl7[`N_1jl7[`N_1jl7[`N_01P00P003W6_1jl7[`N_1jl7[`N_2000191a
T74:0007R`2;LY1aLP0@000CRi2;T8^@R`2;T8^@Ri2;T8^@R`02L@03T8^@00F@00<7[`L04`002JZ@
ZY2ZT:Z@ZP080004RjV@[`:@20000hnYT00<0005LJV^RW405`00391bT7:@0:V?ZHnYS`P000FYSjV?
L@0A000J1jl7[`2_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLB[`<000=a0:l03:l5009a0P000i1a
000;0005Rg6;LH/00W4?0005Rg6@LY000W402`2@LI1aT76@RY2:009a00Z@Ri1aT76@LI2_50002W6@
ZI2YT:V@ZI070009LJV@ZI2YRZU[00L000>@Z@0030000hZ@T00J0003LI1a008000F@Si2?T0080005
Si2?T8X030000g00S`0200B_00<0[jl04jl81`001`L08jl7[`N_1jl7[`N_1jl7[`N_L@1a00N_1jl7
[`N_1jl7[`N_00@000>@LI000P000g6@000<0008T7:@LY1bT78=0006Ri2;T8^@0P002X^@Ri2;T8^@
Ri02L@R@5`002JZ@ZY2ZT:Z@ZP070009LI2YT:n@[i1a00L00ZT00i1a000:0005LJj@ZG407`000hnY
S`080005ZHnYSg402P008G00LG5`0700[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLB[`<0
00=a[jl03:l30004T764L@8000=a00003000276;LH]aRg6;0W4;000AT7:@LI1bL@2@LI1aT76@LI00
0XX01g6@LI1aT7406P002JV@ZI2YT:V@Z@060009JjV@ZI2YT:V@00P000B:T:Ua30000hn@L@0O0003
[hnY00P000F?T8n@L@0:0005S`00L0000W000`2_[`0G[`P700071`0@[`N_1jl7[`N_1jl7[`N_1`<0
011a[`N_1jl7[`N_1jl7[`N_0`00176@Ji0B000:K91bT7:@LY1bK0T00X/01Y2;T8^@00:;00R@Ri2;
T8^@R`N@6`002X^@ZY2ZT:Z@ZY06000;ZI2YT:n@[i2_T7401`000hn@RP0=0003[Y0001/000IaT02_
1g470006SjV?ZHn:2P001G6?08ma00:?0P0060N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061a6_
00Aa07403jl30004T76:LA<000QaRg6;LH]aT09a2@001Y1aT7:@L@8000j@LI1aT76@091aT76@LAd0
00YaZI2YT:V@ZI2Y1P002i2YT:V@ZI2YT:V@00L000AaZI2Y3@000jT0000J0004T:V@Z@>_1P001Y2?
T8n@S`X000MaRP1a076?00806:l81`001`L04:l7[`N_1jl7[`N_1jl7[`L2000A60N_1jl7[`N_1jl7
[`N_1jl070002I1bT7:@Ri2;T008000=Ri2;T8^@Ri2;T8^@[`0W0009T:Z@ZY2ZT:Z@00H000F_T:n@
[`03Z@05[ZV^ZI002@000jV@L@0;0003[Y0001P000^?[Y2^T:j@[Zl7[`030006SjV?ZHnY2`0018l0
S`02S`0JL@0000Ma1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLA[`8000=a[jl03Zl50003T00001D0
00Z;LI1aRg6;LH]/1000391aT76@LI1aT76@LP>_7`000h][000500V@1@003g6@ZI2YT:V@ZI2YT:V@
Z@080003RZV@00`000>YRXl05P002XnYSjV?ZHn@S`04[`07092?T8n@S`0;0007LHZ:076?L0030003
[`2_01>_20L000L702>_1jl7[`N_1jl7[`N_1jl0676_1jl7[`N_1jl7[`N_1jl7[`050003L@0001H0
00]aLY2;T8^@Ri2;L@02000@LY2;T8^@Ri2;T:l7[`N_1ad00W42T00>L@00076@ZY2ZT:Z@ZY050005
T:n@[i000ZT02JnY[ZV^ZJjY[P0900:Y00>@L@001`001Hn@[Y2^01D000R^T:j@[Y2^T0@000X7[`N?
ZHnYSjV?2P07S`@001F_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H74:l00`1a000A[al001J@LH]a
Rg6;LH]a076@LI1aT76@LI1a1ZlM0004Ri2@T0D029000g400004000?T:V@ZI2YT:V@ZI2YT:V@00X0
00>YRXX02P00176YT902LA4000UaSjV?ZHnYSi001@03[`07Si2?T8n@RP030005L8n?Sg000W401Xma
Sg6?L@@05Jl81`001`L08jl7[`N_1jl7[`N_1jl7[`0HLJl7[`N_1jl7[`N_1jl7[`N_00@000=a0000
1@001740LAP@000MT8^@Ri2;T8^@Ri2;T8^@Ri2;T8/7[`N_1jn@07406`001G6@ZY2Z00@000UaZY2Z
T:Z@ZY001@004V^_T:n@[jV^ZJnY[ZV^ZJj@J`P000NY[ZV^T8Y`00D000M`T92^T:i[00T000=aT000
1@00276^T:j@[Y2^1`004`N_1hnYSjV?ZHmaRZV?T8n@Si000Xl018Z?Shl5000D1jl7[`N_1jl7[`N_
1jl7[`N_1jl81`001PL@[`030740016_3@001W9aLW6;L@d0019aRg6;LH]aT76@LI1aT76@LI07[`07
LG:@LH/0L@0H0003Ri2@00>@0`000g6@T006T0D0012YT:V@ZI2YT:V@ZI2YT:V@2P001W6:ZI2YL@X0
00FYT:V?T0060003SjUa00H000RYSjV?ZHnYS`P00jl028n@Si2?T8n@28l018Z?LHl2L@@05:l81`00
1`L08jl7[`N_1jl7[`N_1jl7[g4HLJl7[`N_1jl7[`N_1jl7[`N_00/000]bRg:;LY1bT7:;L@0;000F
Ri2;T8^@Ri2;T8^@Ri2;T:l7[`N_1`<000EaT96@T@02L@03074001@00Y001:Z@ZI030009RZZ@ZY2Z
T:Z:00@001=aT:n@[i2ZZJnY[ZV_ZJjY[P1[00T000F@ZJjY[P02Z@9a1@02L@:@00>^L@0010001I2^
T:j@00@00Y001jj@[Y2^T:h02@005`N_1hnYSjV?ZHnYSi2?T8n@Si2?T76?00H001>_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061`n_0P000g40[`0@[`/000abLG9aLW5bLH]aRg4;000>LH]aRg6;LI1aT76@
LI06[`H00W4018^@Ri02Ra@000F;T8B:R`02T0<01i000g400003000@ZI2YT:V@ZI2YT:V@ZI2YT0h0
0W400jV@Z@090004ZI2YT0@000FYT:V@L@040008LJV?ZHnYSjT;00>_00N?T8n@Si2:00N?00B:ShZ?
0W4501>_20L000L702>_1jl7[`N_1jl7[`N_1`1a602_1jl7[`N_1jl7[`N_1jl7[`090006K75bRg:@
0g801h]bT7:@LW402@004i2;T8^@Ri2;T8^@Ri2_1jl7[`L02P001g6AT96@TH/00g430005L@000740
0P001FaaLG4000:@00FMQ:>@ZP030008ZY2ZT:Z@ZY04000EZY2_T:n@[i2_ZJnY[ZV_ZG40L@1[00d0
00B^ZJjY2002L@0:T75a0:V@[Y2^T0@000R^T:j@[Y2^T0`00147[`N?ZHnYSjV?T8n@Si2?T003S`03
T75a00@001>_1jl7[`N_1jl7[`N_1jl7[`N_00P700061`n_0P000g40[`0@[`T00W401W9aLW5bK0@0
00B;LH]a2P002X]aRg6@LI1aT747[`d001F@691aT8^@Ri2;T1R@LX/0T8]aT7400`001Y25Q8B@RP80
00=aT9001Y03000?LI2YT:V@ZI2YT:V@ZI2Y00:@4`000hYaZ@02T003RY0000@000^?T:U`08nYQ:V@
Z@030009ZHnYSjV?ZHnY00d00jl018n@Si02S`0>RXn?L@2?RXn:Sg6?L74401:_20L000L702>_1jl7
[`N_1jl7[`N_1aQa602_1jl7[`N_1jl7[`N_1jl7[`080008Rg6@LX]bRg850005LY1bT7402P003Y2;
T8^@Ri2;1jl7[`N_40007AP0L@1aT96;TI2AT96@TI2ZT:Z@ZW40T:>4WI2Z00<000RZT:Z@ZY2ZR`@0
01BZT:n@[i2_ZJnY[jV_ZJj@06^_T08000=[07402`02Z@08[Y2@RY2:T8X2000:L75a0:j:WI2^RP<0
00V@[Y2^T:j@[Y003P001`N_1hnYSjT00W050005LG2@Si000hl05W400007[`N_1jl7[`N_1jl7[`N_
1jl81`001PL?[`03L@0001:_1P003QQbLG9aRg6;LH/0Cdi?0P001G6;LH]a00X000J@LI1aT745[a`0
00=a00000W4900:@00>:T9000P02T007ZI2YT:V@Z@02000EK92@ZI2YT:V@ZI2YT:V@ZI2YJjV@0080
00=a00003P001g6:LG5`ZI001P001W6@WHBYT0<000I[ZI2YT:T4T1400Zl2S`07L8n@L8n:S`030007
ShZ?LHmaS`0201:_20L000L702>_1jl7[`N_1jl7[`N_1aQa00N_1jl7[`N_1jl7[`N_1jl7[`06000G
LG:;LX]bT7:@LW40E4m>075bT7:@LW4020002W:@LY2_1jl7[`L/0004T:V@R`8000V@ZY2ZT:Z@ZY00
0`00692ZT:Z@ZY2_T:n@[i2_T:n@LJU`076YLA8017460004LJV4W@:@0`002Jj@[Y2^T:j@S`070003
T00000T000h7LI2?T8n@Si2?T8Z?00J?0`004Jl7[`N_1jl7[`N_1jl7[`N_00P700061`n_00=a0000
4Zl5009a00UbLG9aRg6;LH/00P000aQ?60020005Rg6;LH/00W480003LI2_00F_9@000i00000600:@
0`02T007ZI2YT:V@Z@02000ERjV@ZI2YT:V@ZI2YT:V@ZI2YT74000>@7@02T004WHZYS`8000F?T:V@
Z@04T0L000B@Si2?2P001AQaT8n@00:?00b:ShZ?L@1`Sg6?LHl3016_20L000L702>_1jl7[`N_1jl7
[`N_1faa00N_1jl7[`N_1jl7[`N_1jl7[`05000<679bRg:;LY1bT7:@0`002U@007:@LY1bT7890006
[`N_1jl71@000h/0LP0L0003LI1a00`00Y008ZZ@ZY2ZT:Z@08^ZT:Z@ZY2ZT:n@[i2_T:n@[jV_0740
[Y020005L@0007405`001:j?[Y030009T:j@[Y2^T:j@00H000J^T:j@[Y0:000;LXn@Si2?T8n@Si00
0`05S`80016_1jl7[`N_1jl7[`N_1jl7[`081`001PL?[`03L@1a016_1P003G9aLW5bLG9aRg6;LH/0
1@02L@06Rg6;LH]a1`06[`D000=aRi000W400g80000E0003LI1a00B@3004T006ZI2YT:T00Y005ZV@
ZI2YT:V@ZI2YT:V@ZI2YT01[ZI0300:@6P00192YT9030008ZI2YT:V@ZI060007T8n@Si2?T0040003
LI0000<000Aa08n@0Xl01XZ?RXmaS`8000E`Sg6?L@02016_20L000L702:_1jl7[`N_1jl7[`N_1aP7
S`N_1jl7[`N_1jl7[`N_1jl71@000g5bLP03LP05Rg:;LY000g800i0000020009LY1bT7:@LY1b00D0
00J_1jl7[`L900B@00EaLW40L@070003K00000800W400`2AT002T003ZY2Z009a2`009jZ@ZY2ZT:Z@
ZY2;T:Z@ZY2ZT:Z@[i2_T:n@[i2_T:m[[ZUa076?Z@040003L8X001@000=aT8l00`002G6@[Y2^T:j@
J`0500:@00J^T:j@[Y040003Sg4000H000R?T8n@Si2?T0:?0P04S`0CL0001jl7[`N_1jl7[`N_1jl7
[`081`001PL?[`03LJn?012_1P00475bLG9aLW6;LH]aLP1bLH/30009Rg6;LH]aRg6;00D01Zl:009a
00Z@Ri2;T6b@091a0P001740T742T003Ri2@00B@0W400aQa000:000SLI2@T:V@ZI2YT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT8X00`02T0<000F:ZI2YL@0H0009Sg2YT:V@ZI2?00H000R?T8n@Si2?T0D0
00>@00001@003i2?ShZ?RXn:ShZ?L01aS`02L@804:l81`001`L08Jl7[`N_1jl7[`N_1jl7K0N?1jl7
[`N_1jl7[`N_1jl7[`040003601b00Ib00QaT7:@LP1aT0=b00]aLY1bT7:@LY1bL@0300061jl7[`N_
3P00291aTI2AT96;0Y001I5bTI2A00>@00BARi6@0W400g80000=000RZY2ZT:Z@ZY2ZT:V@ZY2ZT:Z@
ZY2_T:n@[i2_T:nY[jV^L@800ZT02G4008n^T:n@L@0G009a00R^T:j@[Y2^RPD000Ua[Y2^T:j@[Y00
10000g6@00070007T8n@Si2?T004S`0308n?00:?0P0040N_1jl7[`N_1jl7[`N_1jl81`001PL?[`03
LJn?00n_1P00279aLW5bLG9a0P004h]a000HLH]aRg6;LH]aRg6;LH/00W400`2@L@04[a8000Yb0900
T8^@Ri1b0Y02L@8000B@000H4006T00LZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YRP8000>@RW40
0`02T006ZI2YT:U[50001g6:ZI2YT:T00i060009ZHn@Si2?T8n@00D000>@00001`002Xn:ShZ?LHma
Sg420003Sg6?00804:l81`001`L08:l7[`N_1jl7[`N_1jl760N?1jl7[`N_1jl7[`N_1jl71P07LP05
R`1aLY000P005g:@LY1bRg:@LY1bT7:@LY1bT8n@[`N_01X000F@0900L@0G000VLI2ZT:Z@ZY2ZT:Z@
ZY2ZT:Z@ZY2ZT:n@[i2_T:n@[i2_ZJj@[Y040009LI2_T:j@[i1`01<000YaT:j@[Y2^T:j@1P002Zj@
[Y2^T:j@[Y040003LI0000P000J?T8n@Si05S`0F08n?Sg401jl7[`N_1jl7[`N_1jl7[`P700061a6_
00=D[jl03:l5009a00abLG9aLW5b01QaLW42000ELW5bLH]aRg6;LH]aRg6;091aT76@00T000>@0000
:P000g6@T00<T00IZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V:L@02000ALJV@ZI2YT:V@ZI2YSi1aSg40
30000g6@Z@03T005ZHnYSg401@002i2?T8n@Si2?T8n@00@000=aT0001P001740[g050005L8maS`00
0W400hma000@[`P700071`0@[`N_1jl7[`N_1jl7[`N_1`8000f_1jl7[`N_1jl7[`N_00H000N;LX]b
Rg:;00<001F;LQP0LG:@LX]bT7:@LY1bT7:@LP0019090005TI2ALY400W4U009a0Y00>JZ@ZY2ZT:Z@
ZY2ZT:n@[i2_T:n@[i2_T:nY[jV_ZJnY[jV_ZJnYT8Z@ZJnY[ZV^ZJjY[ZV^ZJmaRP03L@800W400jT0
L@03T009[Y2^T:j@[Y2^00D000b^T:j@[Y2^T:j@[Y050003T74000@000Q`S`2_1g6?L09a0P001g6?
LG2?MXl00P003jl7[`N_1jl7[`N_1jl7[`081`001PLA[`03S`2_00b_1P00375bLG9aLW5/079aLP80
01AaLW5bLH]aRg6;LH]aRg40LI1aT0T000N@LY1bT7:@009b0`000g40000O012@01ZYT:V@ZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YL@:@016YT:V@ZI2YT:V@ZI2YT:V@Z@02T006RZV@ZHnY0i002JV@ZI2YT:V@
Z@05000=LHn@Si2?T8n@Si2?T0050003T00000@000AaL0000Zl2S`03RW40008000J?L@1aSg4200n_
20L000L701n_1jl7[`N_1jl7[`N_1jl7Si2_1jl7[`N_1jl7[`N_00D000IaLW:;LX/2LP0701QbRg40
K002LP0>Rg:@LY1bT7:@LY1bT003T0T00Y003I6@TI2AT96@TI2ALY000W4K001DT8^ZT:Z@ZY2ZT:Z@
ZY2ZT:Z@[i2_T:n@[i2_T:n@[i2_ZJnY[jV_ZJnY[jV_ZJnY[jV^ZJjY[ZV^ZJjY[Y2^T:j@[Y2^T:j@
[Y2^T:j@[Y2^T:j@10003jj@[Y2^T:j@[Y2^T:j@L@040003T74000@00W430003[i2?00:?00V@S`00
S`2?MXl00P003jl7[`N_1jl7[`N_1jl7[`081`001PLA[`03T76_00^_1P02L@06LW5bLG9/0P004g9a
LP0HLG9aLW6;LH]aRg6;LH/00P00176@LY09000@T7:@LY1bT7:@LY2;T8^@LAX00W4BT00kZI2YT:V@
ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:V@ZI2YT:T00i000jV?
T003000?T8n@Si2?T8n@Si2?T8n@00D000>@000010000hn@00030006[hn:Sg6?0P001G5`LHma0080
3jl81`001`L07Zl7[`N_1jl7[`N_1jl7[`N?T:l7[`N_1jl7[`N_1`L017800h]b00020003LX/00080
00f;LX]bRg:@LY1bT7:@00801Y08000>LI6@TI2AT96@TI2AT942T0T000=a07403002L@0LT:Z@ZY2Z
T:Z@ZY2ZT:Z@ZY2ZT:n@[i2_T:n@[`R@00=a00001P00>jnY[jV_ZJnY[ZV^ZJjY[ZV_T:j@[Y2^T:j@
[Y2^T:j@[Y2^T:j@[P1aSjj@[Y2^T:j@[Y2^T:j@[Y1a00<000=aShl01@02L@@000D7[hn?T002S`0G
076?SgJ?L@2_1jl7[`N_1jl7[`N_1jl020L000H74Jl00i1a[`0;[`H00W401G9aLW5b008000=aLW40
0P003G9aLW5bLH]aRg6;LH/00P00276@LI1bT7:@1`00475bT7:@LY1bT7:@LY2;T7420003K00000L0
0W400`1aL@040003LY1a01J@00fYLI1aT76@LI1aT76@009a4`005jV@ZI2YT:V@ZI2YT:V@ZI2YT:V@
ZI2Y00F@01JYL@2?T8n@Si2?T8n@Si2?T8n@Si1a10000i0000040003Si0000@000j_LHmaSg6?08ma
Sg6?00n_20L000L7016_1jl7[`N_1jl7[`N_1jl7[`02000:1jl7[`N_1jl7[`P00W801X]bRaP0L@9b
00ma075aRg:;LX]bT7:@LY000`08T003T@0000D000nAT96@TI2AT96@TI2AT9401@001G40000H00<0
00NAT96@TG6A00>@01>AT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2Z00:@00=aT9003I0E009a01V@ZJjY[Y2_T:j@
[Y2^T:j@[Y2^T:j@[Y2^00>@01>^T:j@[Y2^T:j@[Y2^T:j@[Y1a00@000=aT00010000i1a00040004
[`N?T0>?01H0L8mfSgH0[`N_1jl7[`N_1jl7[`N_20L000H74Zl00g40[`09[`H000HHLG5bLG82L@05
075aLW400P00379aLW5bLG9aRg6;L@<000YaT7:@LY1bT7:@1P004W6@LY1bT7:@LY1bT7:@LW40K080
00>;L@001@0JT0<001:@076:T76@LI1aT740LI1aT6`80003L@1a00d0016@LI2@ZI2YT:V@ZI2YT:V@
Z@03T00JZI2@T:V@ZI2YSi2?T8n@Si2?T8n@Si2?T7440003T00000D000>@000010003Jn?LHmaSg40
L8maS`003jl81`001`L07Jl7[`N_1jl7[`N_1jl7[`N_T707[`N_1jl7[`N_00@00W400g9aLP03LP05
R`1aLX/00W82000;LX]bRg:;LY1bT7800`0;T003T@0000@000j@TI2AT96@TI2AT96@T@8000B@002@
0P000g6ALP02000GTI2AT96@TI2AT96@ZY2ZT:Z@ZY2ZT740200:T003092@00:@1`000g6@T006T003
R`0000L00W44000U[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@[P02T0@000=aT000
10000i1a000500031hn?00B?01D0SgJ?L@2_1jl7[`N_1jl7[`N_1jl020L000H74Zl00g6@[`09[`80
00/HLW5bLG9aLW5bL@02000ALW5bL@1aLW5bLG9aLW5bLH/00`003I1a07:@LY1bT7:@LY001P003W:@
LY1bT7:@LY1bT8]a10001W:@002;L@804i0300:@1`002FaaT76@LI1aT0020007LI1aR`2@L@040008
T76@0740L@02L@03T76@00h000Z_ZI2YT:V@ZI2Y1I005JV@ZHn@Si2?T8n@Si2?T8n@Si2?T0050003
T8l000@000=`T0001@003Jn?LHmaSg6?08maS`003jl81`001`L07:l7[`N_1jl7[`N_1jl7[`N_T8l7
[`N_1jl7[`L30006LW5bLG9a1G800`1aLP03LP0<079bRg:;LY1bRg:;10003Y00001bTG:ALY5bTH^A
1P004i2AT96@TI2AT96@TI2AT740T7800P004Y1a002AT96@TI2AT96@TI2AT0D00Y0300:@00=a0000
0P000g6@T005T0<00i020006T8AaL@1a3`000i2;L@06000T[`N_1jl7[`N@[Y2^T:j@[Y2^T:j@[Y2^
T:j@[Y2^T:j@[Y2^20000g6@L0040003LHl000H000>_Shl01Hl4000?[`N_1jl7[`N_1jl7[`N_00P7
00061a:_00>:0:l02:l3000;L@00075bLG9aLW400P00179aLQP20009LW5bLG9aLW5b00/000R@LY1b
T7:@LPH001=aLY1bT7:@LY1bT7:@LY1bLH^@008000F@6000R`0;T0P000>@00000P000i1a00030008
T8^@Ri2;T7430003LI1a008000Z@Ji00Rg6@LH]a3P000i00000200^_00NYT:V@ZI2Y00b@00J?T8n@
Si030003L@0000@000>?T0000`000g2@0006009a00N?LHmaSg5>00<03jl81`001`L06jl7[`N_1jl7
[`N_1jl7[`N_1hl0[`N_1jl7[`08009a1G8200Ab00X0LW:;LY1bRg:;3@00295bTG:ALY6@1P02T00@
TI2AT96@TI2AT96@TG:AT08000j@0000TI2AT96@TI2AT1000W4400J@00Aa00000Y000g400003T080
00Ua091aT76@07403@00:PN_1jl7[`N_1jl7[`N_1jn^T:j@[Y2^T:j@[Y2^T:j@[Y2^T:j@LG2^T0H0
00>@S`0010000hl0000600F?00AfLE>?0P003jl7[`N_1jl7[`N_1jl7[`081`001PLC[`03RP2_00J_
1@02L@08LW5bLG9aLW420004LW5bL@8000UbLG9aLW5bLJl03@001i1bT7:@LY001P003g:@LY1bT7:@
LY1bT7:@R`03T08000N;L@00Ri2;00F@0Zl;0003T00000<000>@R`000P001X^@Ri2;T0<000B@LI1a
0P000i1aT009009a00>@00002@0C[`03T:V@00^@00Z?T8n@Si2?T8ma1P000hn@00040003T00000H0
00V_Sg6?LHl0Dg400P0?[`P700071`0B[`N_1jl7[`N_1jl7[`N_1jl70W401jl7[`N_1jl01002L@04
LW5bL@Eb00<0LG800g802`1bLX]bRg:@LPN_00d000QaTG:ALY6@T@D00Y004I6@TI2AT96@TI2AT96@
TI2A009b0P00376AT96@TI07[`N_1`X00W450005T74007401I0300B@0`000i00000:0003LI0000L0
02b_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[jj@[Y2^T:j@[Y2^T:j@[Y2^T:j@L@L000>?T90010000hl0
000600031hn?00:?00Af002?0P003jl7[`N_1jl7[`N_1jl7[`081`001PLC[`03T8Z_00J_10002aQb
6740LW5bLG9a008000AbLG9a0P001g9aLW5bLG800jl=0008LY1bT7:@LY05000@T7:@LY1bT7:@LY1b
T8^@R`F@00=bRh/00i09[`T000>;000010002I00002@Ri2;T0030004T76@LA8000>@00000`0J[`^@
00Z?T8n@Si2?L@1a1P000gJ@S`040003T00000H000J_Sg6?LHl20003L@0000n_20L000L701^_1jl7
[`N_1jl7[`N_1jl7[`N?T:l7[`N_1jl01P001QP0075bL@=b0P04LP03601b009b00N;LX^_1jl700h0
00AbTG:A0i000g800002000VT96@TI2AT96@TI2AT96@TI2AT96@TI2AT0N_1jl7[`N_1jl7[`L70003
T74000@000=aT0000P03T004L@0000B@4`00<Y2;0:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1i2^T:j@[Y2^T:j@[Y2^T75`0Xl700:@10000hl000070007[hn?SgH0MP03000?[`N_1jl7[`N_1jl7
[`N_00P700061a>_00>@RP001Jl80008LW5bLG9aLW420003LW5b009a00L0LG5bLG9a00B_3`001g:@
LY1bT78010004QQbT7:@LY1bT7:@LY2;T8^@R`F@00>;[jl03Zl60003T00000@00W420003Ri1a00@0
00B@LI1a4`0R[`R@00Z?T8n@Si00T8n@1`000g6?T0030003T00000L000N_LHmaS`2?00<03jl81`00
1`L04Zl7[`N_1jl7[`N_1jl7[`N_1`:?00H01jl7[`L60005LG9aLW401G8200Eb00/0LW9bRg:;RPN_
1`0@0004LY6@T@:@00=b00000P00:Y6@TI2AT96@TI2AT96@TI2AT96@TI07[`N_1jl7[`N_1jl7[`N_
1jl7[`<000>@L@001002L@T0190A000a1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1i2^T:j@[Y2^T:j@[P02T003Sg40008000=a00000P000g6@S`020003MXma00T000J?MXmfSgH3000?
[`N_1jl7[`N_1jl7[`N_00P700061`n_00L0[jn_T8n@00F_1`0041Qa679aLW5bLG80LW5bLG820006
LG9aLW6@0jlA0007LY1bT7:@L@03000?LY1bT7:@LY1bT7:@Ri2;00F@6:lC0004T76@L@h0:Zl5T008
Si2?T8n@Sg0:0003Si00008000>@000020001Zn?LHmaS`<03jl81`001`L06Zl7[`N_1jl7[`N_1jl7
00N_1jn@L02_1jl72@00179aLW44LP03075b00Ab00<0LW800W80192:T:l:009a0Y000i4000020009
Mi6@TI2AT74000:@03BAT96@TI2AT96@TI2AT96_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_2@03T003L@0000<003h7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1i2^T:j@[Y2^T08000FET740L@050003LI2E008000>@S`0020001@MfSgJ?0080
015`0:l7[`N_1jl7[`N_1jl7[`081`001PL?[`030:n_00:_00>:T8X01:l8000@LAQb679aLW5b01Qa
LW5bL@8000MbL@1aT76@00/000F@LY1bT0030008T7:@LY1bT782000?T7:@LY1bT7:@LY1bT7:@06n_
00F@Si2?T003S`03T8l000T000J@S`00T4d:0004[g6?L@8000=f[jl03Zl81`001`L06Zl7[`N_1jl7
[`N_1jl7L@N_1jl709001jl72@00179aLG44LP0301Qb00Eb00<0LW800P000hZ@000;0006LY5gRY2A
0P00Qg6AMi6@TI2ALQR@TI2AT96@TI2AT96@TJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`02L@05[Y2^T:h00Y0<0006SiD0076@2P001:n?LG02000AL@N_1jl7
[`N_1jl7[`N_1jl020L000H74:l00`2_[`02[`03T00000>_3@0014ibLG820007LG9aLW5b0002L@80
00>@L@0030001W:@JhAbT08001IbT7:@LY1bT1R@LY1bT7:@LY1bT7:@Kjl20007LXn@Si2?T0050003
L@0000D000IaS`00MXl:0003[g6?008000==MZl03jl81`001`L06Zl7[`N_1jl7[`N_1jl7[`2_1jl7
Si2?1jl73@000g5bLP02LP03075b00Ab00H0LI2?T8X>0007LWN4JhB@T@020022Mi6@TI2AT96@TI2A
T96@TI2AT0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`D0
00J;T:j@[Y040003T7JE00H000=aU@000P000i1=00080003[hma008001:?00N_1jl7[`N_1jl7[`N_
1jl81`001PL@[`030:n_00>_00>:T0000Zl>0004LW5bL@8000]bLG9aLW40LI1aT00?0006T7::JhAb
0P004i1bT7:@LY1bT7:@LY1bT7:@LY00KZl60007LH^?T8n@L002S`05T8n@Si001P001G1f002?00X0
00>_LG400P000g00[`0?[`P700071`0E[`N_1jl7[`N_1jl7[`N_0:l7[`N_00<000>_1`0030000g5b
LP02LP03075b00Eb00EaT8n@RP0>0007L@00Mi1bT0030020TI2AT96@TI2AT96@TI2AT96_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g46009a0Y002Zj@Si2ET9F@
UG46009a00<0CI002P000jn?0002000BS`07[`N_1jl7[`N_1jl7[`N_20L000H74:l00g2_[`03[`03
RY1a00:_3P02L@0>LW5b079aLW5bLI1aT74>0003T1R@008000<HT7800P004G5bT7:@LY1bT7:@LY1b
T7:@06j_00=aR`001P03L@0:Si2?T8n@Si2?T0L000F@001fL@030003L@0000<000N_L@00MW00012_
20L000L701Z_1jl7[`N_1jl7[`N_1jl700N_1jl7Si001`l017800g5bLP04LP04Si2?T0d000EaLY00
L@070021TGNAT96@TI2AT96@TI2A[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`MaT76@00X000V@UI2ET9F@UG401`001I00091`008000>@L@000`000g5fL@02000C
L@001jl7[`N_1jl7[`N_1jl7[`081`001PLA[`030:n_00>_00B@RXn_3`003aQbLG80LW5bLG9aT76@
L@0<0004T76@CPX000n@LY1bT7:@LY1bT7:@LI00Jjl01X]aRg6;L@/000R?T8n@Si2?T0/000Ea091f
L@030003Sg6@00<000=f00004:l81`001`L06Zl7[`N_1jl7[`N_1jl7[`M`1jl7[`MaT8X73`000g5b
LP07LP05T8n@Si002`02LP03T01a00X000aaTGNAT96@TI2AT944T01b[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g6@LI1aT75/2P001i2ET9F@UI002P001G40T8n@00@0
00>?T7H00`004g4000N_1jl7[`N_1jl7[`N_1jl020L000H74Jl00hl0[`04[`03Si00010000abLG9a
075bLG9aT74<009b00E>0000LP09000:LG:@LY1bT7:@LP@00W400i1a[`1T[`0<LH]aRg6;LH]aRg6;
2002[`07Si2?T8n@L@0:0003RV^@00<000F@MY1fS`020003L@00016_20L000L701R_1jl7[`N_1jl7
[`N_1jl7[`2_1jl7[i0C00=b00H0LW9b[`L;0005LG:@LWL00`000g9aT@08000AMi6@TI2ALY5gT8Z@
Si2?T8l00Y00I`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_T76@LI000`001i1a
T76@LH/01P001Zl7[`N@U@>@00>?L@001`001924RXl30008Si2?T7H0LHl2000A[`N_1jl7[`N_1jl7
[`N_1jl020L000H74Zl00g2_[`04[`03T000014000IbLG9aLW43[`L000LHLG5bT79a00@000AaT01a
1P005W6@LY1bT7:@LY1aT76@LI1aT8Z@LI1N[`06LI1aT76;1@001h]aRg6;LG801006[`06Si2?T8ma
20000gJ4MP030009LI1fT7J?07Ia00804Jl81`001`L04Jl7[`N_1jl7[`N_1jl7[`N_008000H7[`N_
1hlC0003LG9b00=b00>_1jl01P000g5bT003LP03L@0000<00W800g400005000:LGNAMi6@TGNAS`>@
00F?T92@S`06T01Q1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g6@LI1aT0090003L@0000@0
00N_1jl7[`N_00F@0W460003LI2?00<000V?T8n@MY001g400`004Jl7[`N_1jl7[`N_1jl7[`N_00P7
00061a:_00<0Sjl01:l00i1a000A009a00AbLG9a1:l3000467MbM`9a1`00191bT1P7000JLY1bT7:@
LY1aT76@LI1aT76@092:T76@RY1G[`07T76@08]aT00@00R_00J?T8n@Si070003T7Ia00<000F@MY1f
T0020004[gH001:_20L000L701:_1jl7[`N_1jl7[`N_1jl7[hl20006[`N_1jn@4`05LP08[`N_1jma
LY02LP03600000H000A>LY5a20002Y1gTGNAMi6:T8X:00N@05d01jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_T76@L@1aT7403`002Zl7[`N_1jl7[`L5T09a1@02S`8000IaT7J@MY03000F1g4000N_1jl7[`N_
1jl7[`N_1jl7[`P700061a>_00>?L:l01:l00hX0000A0004LG9aLPJ_00=b00002P001G5bT79a00L0
0W402I1bT7:@LP1aT00;0006T8Z@LI2:0P1A[`08LI1aT740LI02L@h02jl01i2?T8n@Si002@001Y1f
T7J@MP<000>_00004jl81`001`L04Zl7[`N_1jl7[`N_1jl7[`N_1`8000J_1jl7[g470004675aCP9b
00=a00000`04LP051jl7[`L03P000dj@L@09000;695gTGNAMi5a08X02P000g6@T004T0@005R_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`1aT76@L@1aT76@3`0030N_1jl7[`N_1jl7[`F@00>?LG401P001Y2?T7J@MP@0
01If002_1jl7[`N_1jl7[`N_1jl7[`N_20L000H74jl00g00S`04[`03S`0000@000]b679>LTibLG80
L@020004CW9aLPF_3@0017:@LW48000<675bT7:@LY00T02@2P001W6@LI2:T0D000>@LJl0Ajl03I1a
L@2@LH]aT02@LI003P0>[`06Si2?T8n@0Xl40007MY1fT7J@MP050003L@0001>_20L000L701>_1jl7
[`N_1jl7[`N_1jl7[`N:008000D7[`N_1`05009a00=bLG803G801@N_1jl700d000>@L@002@02LP0:
T7:AMi5gLHX0L@X01Y05001FSi2:T:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7RY2:R`2@LI1aT02@LI0>000@[`N_1jl7[`N_1jl7
[`N_1`B@00>?T8l00W402I2?T8n@MY1fT005000FMP07[`N_1jl7[`N_1jl7[`N_1jl7[`P700061aB_
00=`00001Jl8000>679>LQQbCW9>LTibLG85[``000=gLP002@0027MbT7:@LY1b3`001G6@LI1a00D0
00YaT76@LI1aT76@>jl0591aT76@L@1aT76@LI1aRg6;LH]/2P0@[`0BL@00MY1fT7J@MY1fT7J@MY1f
1`000g1a[`0C[`P700071`0D[`N_1jl7[`N_1jl7[`N_1jl7[`L20005[`N_1jl02@000aQbL@0:LP05
1jl7[`L02@00175bTG4:0008LG:ALY5bTG8@00F@1@0;T01?RY2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[i2;T8^@Ri1a076@LI1aT02@RY1aT76@L@08000A
1jl7[`N_1jl7[`N_1jl7[h/00`003YF?T8n@Si2?T8n@Si1a1`005g5f[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00P700061aF_00>?L0001:l=0009LQQbCW9>LW5b00B_2@03LP/000XHLY1bT7:@LY2_4000191a
T745000@LI1aT76@LI1aT76@LI1aT2f_00b@LI1aT76@LI1aT7420009T76@LH/0T76;009a2P0A[`9a
0P02[`0;T7J@MY1fT7J@MY001@001W40001fL1F_20L000L701F_1jl7[`N_1jl7[`N_1jl7[`N_1hX0
0P0010N_1jl?00Mb00@7[`N_1P000g5bT002LPT000iaLY1bT7:ALY5bTJl7[`l00i000g4000020003
LI2@00^@00F;T92@RP02T00i1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jn@RY2;
T8^@Ri2;T8^@R`1aT76@009a00P0T76@LI1aR`9a1P004Zl7[`N_1jl7[`N_1jl7[`N_R`9a00l01jl7
Si2?T8n@Si2?T7401@006gH0001a1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLF[`0308l000>_
4@00179>LTh5[`@000HHLW9gLW49000<6000LWMbT7:@LY1b1:l>0003T76@00D000B@LI1a0P02L@0A
T76@LI1aT76@LI1aT76@LI000W4K[`0AT76@LI1aT76@LI1aT76@LI000W40102@LI030005T76@LH/0
2@0B[`=a00<0[jl00Zl02Y1fT7J@MY1fT4d4009a00F?0000MP0F[`P700071`0F[`N_1jl7[`N_1jl7
[`N_1jl7[`N_S`8000>_1jl04P03LP05[`N_1jl00`001G:@LW8H00X000<H00000P003G:ALY5bTG87
[`N_1jl03@03T0D00i0400F@00N;T92@Ri2;00F@00=a00001I004H^@RY2:T:l7[`N_1jn@Ri2;00@0
01MaRY2:T8Z@RY2;T8^@Ri2;T8]a091aT002000:LI1aT76@LI1aT0D002@7[`N_1jl7[`N_1jl7[`N_
1jn;LH]a1jl7[`N_T8n@Si2?T8l40004LGJ@MP<001Ma1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`00
1PLG[`03072_00:_4`000g9>[`04[`<000=a00004@001aQbT7:@LY002:l<009a1@000i1aT002000B
CP00T76@LI1aT76@LI1aT76@0`005i1aT76@LI1aT76@LI1aT76@LI1aT76@00D000n@LI1aT76@LI1a
T76@LI000`000i1aL@020006LI1aT76;0`000g400003016_0W40101aLG46[`07MY1fT7J@L@040004
Sg6?L@@05jl81`001`L05Zl7[`N_1jl7[`N_1jl7[`N_1jl7[`L300031jl701<000Ib[`N_1jlD0004
CW80009b00fALY6:1jl7[`N_1jl700`000>@00000`02T007L@1DE02@L@05T006Ri2@T8^@0`09T00A
RY1a08Z@Ri2:T8^@Ri2;T8/01@003XZ@RY2:T8Z@Ri2;T8^@0P000g6@R`02000;T76@RY1aT76;0740
10009Zl7[`N_1jl7[`N_1jl7[`N_1g6;08]a1jl7[`N_1jn@Si2?T8ma0P001WJ@MY1fT0@001N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLG[`03S`2?00>_1`0011Qb61P900F_3@000aP000020004
LP00M`8000M>T7:@LY1a00Z_2P000g4000040006L@00E4h00W460003J`0000<000eaT76@LI1aT76@
LI00009a00f@LI1aT76@LI1aT76@00D000F@LI1aT0030004T76@L@8001:@LI1a076@LI1aT02;LI1a
Rg4301:_0W420003LJn_00N_00mfT7J@MXmfT7J?LHmaSg40100G[`P700071`0G[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl00P000`N_1`05009a00IbLG9aLTh70005[`N_1jl02P02L@08T7800900LG82000E
T000LY5bTG6@RY2_1jl7[`N_1jl700l00i000`1aT003T005LI00L@000g4=T009L@2@RY2;T8Ya00<0
00J@Ri2;T8/60004T8Z@L@<000R;L@1aRi00L@<000F@RY2:T009000E1jl7[`N_1jl7[`N_1jl7[`N_
1g6;00<003@7[`N_1jl7[`N_1hn@Si2?T7J@MY1fT7J@07401jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
20L000H76:l300:_1@002Tib678HLQQb678500J_2002LP04Mg9g608000Ag075b0P001GL0T7:@00<0
00>@LI002jl=000ULI1aT76@LI1aT02@LI1aT76@LI1aT76@LI1aT740LI1aT76@L@060005T76@LI00
1P000g6@L@02000AE02@K91aT01a091aT76@LI00200D[`=a0P0<[`0=T7J@MY1fSgJ?LHmaS`0401R_
20L000L701R_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L300031jl000<000EbLG4060020003LG9b00=b
0`001@N_1jl700@0011bMg9gLWMbMg:@LW4HT79a0W801@1aLY1b00@0011aSi2:1jl7[`N_1jl7[`N_
3003T003Ri2@00>@0W4;000;L@2@RY2:T8^@Rg4020001W6@RY2:L@D000B@Ri1a0`0038Z@RY2:T8Z@
RY1aT0P001R_1jl7[`N_1jl7[`N_1jl7[`N_1g6;LH/2000K[`N_1jl7[`N_1jl7[`N?T8n@MY1fT7J@
MY1a00<001V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061aV_00>?00000Zl3008H00EbCP1>
C`020008679>LTibCW45[`@001MgLWMbMg9gLWMbMg9gLWMbMg:@LWMbM`020008LI1aT76@LI0<[`/0
00^@LI1aT76@LI2:T003L@0?T76@LI1aT76@LI1aT76@00`000EaT76@L@060004T76@009a00^@LI1a
T76@LI1aT00601J_00AaRg6:0P0?[`0<T7J@MY1aT76@LHma100I[`P700071`0J[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[hl20003[`00008000MbLG9a04i?008000a>LW9bLG9a1jl7[`L4000FLWMbMg9g
LWMbT7:@LY1bT7:@LY5bT0H001=aT8Z@RY2_1jl7[`N_1jl7[`N_00X00W400i2;T00:T00<Ri2:T8^@
RY2;T8Z@3P00176@RY07000>T8^@Ri2:T8Z@RY2:LF`6000i1jl7[`N_1jl7[`N_1jl7[`N_1jl7[h]a
Rg6;L@2_1jl7[`N_1jl7[`N_1jl7[i1fT7J@MY1fT7Ia00<001X7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`P700061aZ_00E[S`00[`030005679>LTh01@001TibCW9>LPF_1@005W9gLWMbMg9gLWMbMg9g
LWMbMg9gLW43009a00R?LI1aT76@L@f_30006G6@RY1aT8Z@LI1aT76@LI1aT76@LI1aT7403@00176@
LG480004LI1aT0=a2@0F[`08LH]aRg6;L@0B[`0:MY1fT76@LI1aS`8000=a0:l06Jl81`001`L06Jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00XX00hl01`040005LG9aLW401003LP08LG9a1jl7[`L6000F
LWMbMg9gLY1bT7:@LY1bT7:@LY2?L@<001R@Si2:T8Z@RY2:1jl7[`N_1jl7[`N_1jl;0003LI2@00N@
00R;T76@LI2:T0>:0P000i1a000=0003T8X001D001d7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g6;LH]a
R`02000O[`N_1jl7[`N_1jl7[`N_1jl7[i1fT7J@MY1fT7H0C@02000J1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl81`001PLJ[`04L8m`RPD000mbCW9>LTib674HLTibCW801Jl8000AMg9gLWMbMg9gLWMb
Mg9gLWL00P000g400002000;T76@LI1aT76@LI004:l90006L@00LI2:0W420003T76@00=a00J@001a
074@0003L@0001406:l01H]aRg6;00<05Jl02GJ@MY1aT76?L@0401^_20L000L701b_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1hZ?0XX40004CW9aLP9a00=>LW400g80275bL@N_1jl72@00475bMg9gLY1b
T7:@LY1bTG830003RP00008001f@RY2:T8Z@RY2:T8X7[`N_1jl7[`N_1jl7[`N_1`0A0003LHZ@009a
00F:T02@L@0R000L1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jn;00=a0`008@N_1jl7[`N_1jl7[`N_
1jl7[`N_1jn@MY1fT7J@MY00MP02000K[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061a^_
00B?L8m`10003g5>LTibCP1>LTibCW8HLP05[``000igLWMbMg9gLWMbMg9g[`<0019a0000LI1aT76@
LI1aT76@LI0B[`l000B:LI000W400i00000N01b_00=a00001@0H[`06MY1fT76@0W4501^_20L000L7
01n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jn?RXn:00<001LHLW5bLG9a075bLG9aLW5b[`N_1g9g
LP0:000>Mg9gLY1bT7:@LY1b1jl3000UL@0008Z@RY2:T8Z@RY2:T8Z@RPN_1jl7[`N_1jl7[`N_1jl7
[`0>0004T8Z@LAl0026_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[i00Rg400`008740[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[i1fT7J@10007g40076_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0
20L000H77:l01G2?L8m`00<000m>LTia0780LDibCW9>LQP01Jl3LP03Mg8000L000]gLWMbMg9gLWMb
M`04[`D0016@LI1aT76@LI1aT76@LI1aT00G[bH07Zl00h]a00050003L@2_01Z_00EfT76@L@060003
MP2_01Z_20L000L7026_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[hn:ShX00P000g9aLP02L@05
0780LW401G804:l7[`N_Mg9gLWMbMg9gLQP3000ALG9gLWMbMg9gLPN_1jl7[`L010001Hn@Si2?00L0
02j:T8Z@RY2:1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl76@005@N_1jl7[`N_
1jl7[`N_1jl7[`N_R`02L@<000>;00000P008Zl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_T7J@
MW42000QL@00MP07[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061aj_00AaSg6?0P0011Qb
6782008H00L0LQQbCW9>00J_00]gLWMbMg9gLWMbM`030009679gLWMbMg9g00V_0`02L@T000V@LI1a
T76@LI008ZlH0003M`0000@03Zl00h]aR`030003R`0000<07Zl0101aSg440003C@1f00807:l81`00
1`L08Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1hZ?RP020006675bCW800W4200Ib01<7[`N_
1g5gLWMbMg9gLWMbMg9a0080015bMg9gLWMb1jl7[`N_1jl7[`0:009a02Z@RY2:T8Z@RY2:T8Z@[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl50003692@00D000>@00001P001W40UP2FM`L000f_1jl7
[`N_1jl7LI1a00800W45000U[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`MfT7J@L@020003
L@00008001f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLP[`03LHma008000=b6780
0QP02P1b600HLTibCW85[`@000MaLWMbMg9g00Eb00MaLWMbMg9g00b_20004g6@LI1aT76@LI1aT76@
LI1aT7405jl50006TGNAMi5g10000i1g00060006L@1g07N@2@04[`04Rg6;L@@000J;0000Rg4R[`05
T76?LGH010000g40[`0M[`P700071`0R[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[hn:S`80
00Y>LW5b601aLP1a1G801@N_1jl700L001igLWMbMg9gLWMbMg9gLWMb1jl7[`N_1jl7[`N_1jl60009
T8Z@RY2:T8Z@009a0P00776@RY2:T8Z@[`N_1jl7[`N_1jl7[`N_1jl7[`L40009T9J@UY2FT02@00<0
00>@UP001@002WMa09IbUY1aCY080003LI1/00@000F@0000T002000X[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7MY1fL@8000=a00000P007PN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`P700061b6_00>?L@000P0011Qb678200096780LQQbCW9>00B_2`07LP05Mg9gLWL04:l40007
LI1aT76@L@0<0005Sg6@LI004:l00`2@L@02M`04TGNACP9g00>A09000`000i1g0005000=T7N@Mi1g
T7N@Mi1gL@09009a0P02L@809Zl00hmaMP030004MP1f01n_20L000L702>_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1hZ?RP030006CW9>LP1b0W46LP04[`N_1``001hHMg9gLWMbMg9gLPN_1jl7
[`N_1jl7[`N_1jl7[`L2T005RY2:T8X01000192:RXX7000AJi2:T8Z@RY2_1jl7[`N_1jl00`001I2A
T9J@008000MaT75a091a008000>@UY001@03T00;UY2FT9J@UY2FT9H01P0017400745000Y1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7MY000W400`1f0002000P1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLR[`03RXl000800QP03G8H61Qb601>LQQbCP000jl90003
600000801W8017MbMg8C[`06LI1aT76@10001G64JhYa00X00W401i1aT76@LI000jl3000?L@1gTG80
Mi5b07NA095b008000>@M`000`004TigTGN@Mi1gT7N@Mi1gT7N@CP/0:Jl00gH0L@020004L@00022_
20L000L702>_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7RP040008LTibCP1>LP06LP04
0:l7[`H000IaLWMbMg82000QLG9gLWMbMg9g[`N_1jl7[`N_1jl7[`N_1jl7[`N_T8Z@00H000J?T8Ya
090=00::00B@RY2:10003gL0UY1aT9IaMi2FT9J@L@020003T9H000<000B@UY2F1P00276FT9J@UY2F
2000;Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7MP03000UL@000:l7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLS[`03S`0000<000=b67800P0021Qb
079>LQQb0P02[`L000=gLWL00V/01G80000H00Ab00=gLZl05jl=0003LI0000=a2@000i1a0005000>
LGL0Mi40T7NAMi5gT7420003T7Ma008000egT@00CY1gT7N@M`1a009g00B@Mi1g200[[`03L@000080
0W42026_20L000L702B_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[hl40005CW9>LP00
0W400g80LP03LP04ZHl0[`P000IgLW5[Mg84000OMg9gLWN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1`0>0008L@0008Z@RY0=00:@00ZFT7:@UY2FT9J@0`00692FT96@UP2FT9J@UY2FT9J@LI2FT9J@UPT0
02d7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jn@MW401@008PN_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLT[`03RXl00080019b678HLTib6780CQQb602:
0:l90003L@0000H00W800gMb[`0P[``000=aRP001P00191gTGL4000HTGMgLI1aT76@LI1aT76@Mi5g
TGN@Mi1g0P03M`0807L0T01D090;02N_00@0MW5f0P000g40000S[`P700071`0V[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7RXl2000CCW9>LW5b079>LQQaCW80LHl0[`080003Mg8000H0
02THMg9gLPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`0;0003RW4000H000FAMi6@
T@0300::00Z@RY2:T8Z@RY2:0`004i2FT9H0LI2FT9J@UY2F06H0UW4030009@N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L00P03MP8002Ia00N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`P700061bJ_01F:S`1b600HLAP0678HLQP06780T8l02@000aQa00070004LWMbMbN_
2P000i0000050004TGNAJ`L000MaT76@LI1a00H000^@091gT7N@Mi1gT0030003T7L000<000>@0000
1`0R[`@000=fL@000`0U[`P700071`0W[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N:0080019b075>LDib079>LTh0CZT0ZHl80005Mg80CWL01P00;g9gLPN_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P000BA095g1@000g6@L@0A000?RPN@UY2FT9J@UWL0
MiJ@00@000>FT00020008Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`0400=f02T0MW40
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLX[`0ARP0H01Qb678HL@1b
678H01P00P000i2?0006000667L001Qa1@0017MbMg8a[`07095g04jAM`0D00N_00]gT7N@Mi1gT7N@
L@040003T7Mg00X08:l20004K01aMPD09Jl81`001`L0::l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`L3000BLQP0CW9>LAQbCW9>08l0SjV?1P00175b6002LPD0061bMg9g[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`1gTGNAM`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[iJ@UY2FT9H50003UY0000/001h7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl30004L@00MPD002H7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P7
00061bZ_0P001QP0078HLP<H00Qb678HT8maS`L000=>MaP00P000g8H0003009b00=gLWL0<:l01@2A
Mi5g022_00F@Mi1gT0040004LI1gM``07Jl40006L@1fL@1a0P0W[`P700071`0Y[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl01@003aQbCW9>04ibCW9aZHm`S`060003LG800080
00=bM`001000GGMbMg87[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl0Mi5gTJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[iJ@UP040004T9J@UPd001`7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_100017400742000ZMP07[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_20L000H7:jl60004LQQb608000Qb67:_T8l0S`D000<HL@000P000aQ>
000400=b00=g[jl0;Zl0102AM`0S[`03T7L0008000=gT7L02P000gL0M`0201Z_1P0016`0LG420003
CZn_02N_20L000L702^_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00@0
011bCW9>LTib079aLZnYSg2?10000g4000030003M`0000<005agLWMbMjl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`2AM`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`8000BFT9J@20008I00T02@091>0:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`060004
K000MP8002]a1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061bf_
10002W4001Qb6780LTh2[`04T8maS`<000=b60000`000g80000200Ib<:l00dibT@0U[`04T7N@M`P0
00R@0900T76@M`805Zl80003L@1f00<0:jl81`001`L0;Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`09000:LG80LTh7[`N_Z@<000=b00000`000g5b0005009b039gLWN_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`8002/7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7T9J@TI2A00F@00NFMi1gT7N@00D001@7[`N_1jl7[`N_
1jl7[`N_1jl7[`L003=aRW5f07J_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl020L000H7;jl:0004LQQb60B_00>@LAP01`000g8000070003Mdn_032_00=a[jl08Zl0491g
T7N@Mi1gT7N@Mi1gT742M`D04Zl90005LFaaCP00;Zl81`001`L0;Zl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L:000<CW9>LTh7[`N_1hXH1@000aQ>0005009b0P00=7Mb
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl2LP0UTGL00740
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[i6@L@070008T7NEMi1gT4h4000@1jl7[`N_1jl7[`N_
1jl7[`X003B:LHZ_0:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
20L000H7<:l:0005678HLQP01Jl70003LP0000<000Qb0000LDmgCcF_0P000g80000201V_00QgT7N@
Mi1gT09a00P0T7N@Mi1gT0@03Zl;0005LFaa[gH0<:l81`001`L0<:l7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`/000QbCW:_1jl7[`H000=aLP000`00@Th0CW9gLWMb
Mjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1gL0L@D0
01N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`05T00>Mi1gT7Ma091gT7N@Mg43000<1jl7[`N_1jl7[`N_
0P001Faa002:00D003J:LHZ_MW47[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl81`001PLa[``000=b67801Jl50003CW8000@000PHCg9?MdmgCcN_00Eg0000M`0301J_
00]gT7N@Mi1gT7N@M`020005T7N@Mi00100:[`8000=/06`00P000hX000030003LFaa00:_00=>[jl0
;jl81`001`L0<Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0
3000279>LZl7[`N_10000g5b00020014LW4001QgLWMbMg87[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7Mg43000MT7L7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1gMa09000W4017N@Mi020006Mi1gT7Ma0`002@N_1jl7[`N_1`020006L@1aLFb:0`00>VaaRW47[`Ma
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLc[`X0
00E>678HLP05[`<000=>00000P002W9?MdmbCgM?Mdl2LSJ_00R@Mi1gT7N@01J_00>@Mi000Th30009
T7L0091gT7N@00@01jl2000>RVaaJhY/RP2:K8Y/LF`3[`03CZn_036_20L000L703>_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl01P003G5>LTibCW9>LZl7[`L0
0`000gL000020015LWMbMg9gLWMbMg9gLPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[i5g00H001l7[`N_1jl7[`N_1jl7[`N_1jl7[`MgT7N@Mi1gL@2@0080
00EgT7N@M`0400051jl7[`L01000AG6:LHY/Rg6:LHYaRZl7[`Ma1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLe[`D000THLQQb678HLQP01Jl7000:
LTibCgM>LTmgC`=b00=?[jl0<jl01WN@Mi1gT0805Jl04I1gT7N@Mi1g07N@Mi1gT7N@00<01:l5000<
K8Y/RVb:K8Y/RVb:0jl00`1>[`0d[`P700071`0f[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl70`002G9aLW5bLG9aLP02L@05[`N_1jl01@00J79gLWMbMg9g
LWMbMg9gLWN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jnA
Mi5gTGL0M`2_1jl7[`N_1jl7[`N_1jl7[`N_1gN@Mi1gT7N@0gL01I1gT7Ma00T000N;LH]aRP1a0080
00Q/002:[`N_1`9a03N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061cR_0P000aQa00080003RW6_00B_0`0:LP07MdmgCgM?M`02032_00AgT7M>
0P000g40000C[`0=Mi1gT7N@Mi1gLEA>0002M`P000n:K8Y/RVa[08Y/RVb:K8X01:l00`1a[`0g[`P7
00071`0h[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`L?00051jl7[`L010000aQ>00060008CW9gLWMbMg82000a1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[i00T006000J[`N_1jl7[`N_1jl7[`N_1jl7[i1gT7N@Mi020004
CP1gT0L000F;RX^:R`02L@19075aRg6:LHYa1jl7[`L0LJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLi[a401:l<0009CgM?MdmgCgM?00@0
:jl00gN@00020005T000M`004Zl01WN@Mi1gT09g00<0T7L02@003g6:K8X0LFb:K8Y/RVb:K005[`03
076_03V_20L000L703V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl04@001:l7[`L8000=679bMg9gLWMbMg9gLP06000Y1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7Mi000`000i00LP02000L[`N_1jl7[`N_1jl7[`N_1jl7[i1gT7N@
Mi1gT0/0019a002:Rg6;LH]aRg47[`N_1jl2009a03X7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7?:l60007678HLQQb600400:_1@004Dmb
LW9?MdmgCgM?MdmgCgMb00P08jl2M`<000=a00000`0B[`05T7N@Mi0040002fb:LHY/RV^:JhY/00F_
00D0J`1a000k[`P700071`0l[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl71@002G9>LTibCW9>S`04000@[`00Mg9gLWMbMg9gLW40L@8000HHMg9g
LWL:000Q[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7Mi1g00<000=a07400`004jl7[`N_1jl7
[`N_1jl7[`N_1gL02P00175gMg450004RX^:L@<000Qa076_1jl7[`801740?0N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061cj_1@002g80
01Qb678HLXYa008000>_LTl00g800dmbLP05008H0`001TibMg9gLP8000>:0000200H[`9g00=b0000
0`000gL0000201:_2002L@9g00=a00001`02K004RVb:K0F_10000f]a000m[`P700071`0n[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L30004
LW40CP8000EbCW9>LP020009RXl01g9gLWMb00@000AgK6]b1@003QQgLWMbMhnYSjV?ZHn:1`0050N_
1jl7[`N_1jl7[`N_1jl7[`N_17L01@1gJgM>00<000h7[`N_1jl7[`N_1jl7[`L000QgT7N@Mi1gT0T0
00f;RX^:RhZ;[`N_1jl700@00W40?jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PM0[`<000=>60000`001G8HLQQb00L000=a0000
0`001G9>K01b00L00W8017MbMg82000:T8Z@RY2:T8Z@L0H03Jl01g9gLWMbMg800P000gL0000200V_
2@02M`06LWMaMg5g30001G6:LHYa00J_0P001G4006]a042_20L000L7042_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl70`002g9>LTibCW9>
LTib00d000Ea0000M`08000DCWMbMg9g0:V?ZHnYRZV:ZHnYSg4=00Eg00<0MgL00gLF00=g00>@0000
3P0038Z;RX^_1jl7[`L0L@@00W40@Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061dB_0P004QQa678H01Qb678HShZ?RXn:S`X0
00E?CP1gCP04000?601>07M?Mg9gLWMbT8Ya00<001B@RY2:T8YgLWMb04igLWMbMg9gLP<000AbMg9g
300016^;RX/20004LGMaMa001jl20003J`0000<000=a[jl0@Zl81`001`L0A:l7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl70P004W5b
LG80LP1bCW9>LXZ?RXn:S`d000=aLP000`005W40Mg5gLWMbMg9gLWMbMg9gLWMbMg84M`06LP1>CP1b
17L00i1g00040003M`0000T000B:T8Z@0P07M`d000F_1jl7[`060003L@00009a04F_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0
20L000H7AZl30003LAQb00<000Ub678HLXZ?RXl00XX:0003CP00008000=g00000`00575bMdmgLWMb
Mg9gCgM?Mg9gLThH0P000gL000050006Mg9gLWMb3`0018Z@RX/20003LGMa008000AgLGMa0P000gMa
000600F_0`000f/0J`040003LJn_04B_20L000L704N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02009a00=bL@000dh02G9>
LTj?RXn:S`0:0003LP00008000=g00000`003TibMg9gLWMbMg9gLWMa8P001H^:T8Z@00801GL00`1g
M`02M`05LWMaM`000Th30003[`N_00@000>:00000P00Bf]aLG6_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PM9[`<0
00`HLQP06780LQQb68l3RP03L@0000P000=>00000`004Dh007M?Mg9gCgMbM`1gCgM?00H000AaCWL0
0W801GMbMg9g00H000>:T6/020004g6:RhZ;RP1aMg5gLGMaMg5gLGL00g400g80L@020003[`0000@0
00=a074010000f^_[`17[`P700071`1:[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L3000>LTh0679>LAQbCW::ShX50003
RXl000<000=b00000`004g8007MbMg9gLWMb079gLWL0Mg804P001jV:ZHZ@RY000P0038Z@RY2:T8Z@
RWL0L@Ig00QaMgMgLGL0L@9g20000f/0000200=a04^_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7CJl20008
LQQb078HLQP20004LXZ:RPD00hX30003CP0000H000IgCgMbMg82000?MdmgCP1?Jdl0CWMbMg9g00P0
01J:Sg5[092:T8Z@RY2:RhZ;RX^:Mg5g0P00375gLGMa075gLGMaM`/000Ia001[LF]=[`P700071`1=
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl00P003W9>LTia079>LW5b68n:100028n:ShZ?RP0H10002GL0001gLWMb
M`020004LWMbM`8000=bM`002P00679gZHZYRZV:ZHZY[`N_1jl7[`N_1jl7LP=g00<0MgL00WL00g5g
M`02M`T000B:075a0`00DG6:[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PM?[`@000eb678H01QbCW9>
LTj:008000=aRXX00hX70003L@0000<000Qb0000MdigCPl05Jl047MaMdigCP1>Mdia07MaMg480003
LF/000@000=[LJl0Cjl81`001`L0D:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl70`003G5>LTibCP1aLW5bLG80
0P0018Z?RXl2RQ0000AbMg9g2`005`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl700Ag00=bMg800P0017Mb
0004M`D000=[L@001@00EG6:L@N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7EJl2000>L@00679>
LQQbCW8HLTh5RP<000B:Q8Z:600@[`8000UgCWM>MdigCWL00Th01W5gCWMaM`9a0`000hA[00040004
JhY[RUJ_20L000L705N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`000Th00`1>0002CP0=079>LTib
LG9aShZ?RP0200>:00<0LHX00XX00g40000=00>:3@003F/0Mg9gLGMbMg5gCWL00W40279gLWMbMg9g
2@00G8YaRZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7FJl30003600H00P000AaCW9>
0XX<0003L@2:00N:40000hY[00040006CWM>Mdig0P002DiD07M>MdigCP070005JhY[RV/0Fjl81`00
1`L0FZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`d00W401G9aLP1a00T000=aRXX04P00172:
Jg420003LWMb00P000YgLGMaMg9gLWM>1@00IG6:LHYaRW47[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061ef_3000179>LTh40003LTib00P000=[L8X00hX026^:LHY`LG1a10002Tig
CW9>MdigCWLC0006RV^:JhY[HJl81`001`L0Gjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00X000=bLG8010002AQbCW9>LTibL@0=000:CWM>MdigCWMaM`D000EgLWMaM`0>00::06MaRW47
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061f>_20000g9>LP04000K
CW9>LTibCW9>LTibCW9>LTib075>LTibCW80009>00=bCP0020001TigCWM>M`H000A[RV^:J:l81`00
1`L0IJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00H00W440003LTib00<0
01QbLG9aLW5bLG9aLP1bCWM>Mdib079>Mg4D001^RZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PMY[`/000I>001?04h7000>LTibCW9>LTh0CW9>LThC072_
20L000L706^_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`0:0007
LTibLG9aLP090009LDigCWM>LTib008000=>00002@00MJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PMa[`P001MbCW9>LTibCW9>LTibCW9>
LDib075>L@0907V_20L000L708:_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl72@00QPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7ojlC[`P7
00071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B
[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7
[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7
[`N_1jl720L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"MathLinkMode"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Writing input files for Persistence of Vision Ray Tracer", "Section",
  CellTags->"POVRay"],

Cell[TextData[{
  StyleBox["The program ",
    FontFamily->"Times New Roman"],
  "can write input files for the Persistence of Vision Ray Tracer (",
  ButtonBox["POVRay)",
    ButtonData:>{
      URL[ "http://www.povray.org"], None},
    ButtonStyle->"Hyperlink"],
  ". If the OpenGL viewer is running in MathLink mode the function ",
  StyleBox["MVWritePOVRay[]",
    FontFamily->"Courier"],
  " will create a input file for POVRay. The first argument is the name of \
the POVRay file. If the second argument is a ",
  StyleBox["Graphics3D[]",
    FontFamily->"Courier"],
  " object it is first transferred to the  OpenGL Viewer and than the POVRay \
file is written. Is the second argument not a ",
  StyleBox["Graphics3D[]",
    FontFamily->"Courier"],
  " object the current contents of the OpenGL viewer window is stored in the \
file."
}], "Text",
  CellTags->"POVRay"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\tMVWritePOVRay["\<f:/jpk/Develop/mview3d/sicosa.pov\>", 
      Graphics3D[Stellate[Icosahedron[]]], \n\tMVPolygonShading -> MVFlat]
      \)\)], "Input",
  CellLabel->"In[4]:=",
  CellTags->"POVRay"],

Cell[BoxData[
    \("f:\\jpk\\Develop\\mview3d\\sicosa.pov"\)], "Output",
  CellLabel->"Out[4]=",
  CellTags->"POVRay"]
}, Open  ]],

Cell[TextData[{
  "Since the surface normals are calculated by the program, POVRay files with \
smooth surfaces can be written. Here is an example from the \"",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " Programmer II\""
}], "Text",
  CellTags->"POVRay"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(Needs["\<Graphics`ParametricPlot3D`\>"]\), 
    \(r8[t_] := 1/100 + 1/4\ t/\((2  \[Pi])\); \n
    z8[t_] := \(-9\)/32\ \((t/\((2  \[Pi])\))\)^2; \nrev = 3 + 5/8; \n
    r0 = 1/20; \n
    grMaeder2 = 
      ParametricPlot3D[\n
        \t\t\ \ \ \ \ \ \ \ \ {
          \((r0 + r8[t]*\((105/100 + Cos[\[Phi]])\))\)*Cos[t], \n
          \t\t\t\ \ \ \ \ \ \ \ \ 
          r8[t]*\((r0 + r8[t]*\((105/100 + Cos[\[Phi]])\))\)*Sin[t], \n
          \t\t\t\ \ \ \ \ \ \ \ \ z8[t] + r8[t]*Sin[\[Phi]]}, \n
        \t\t\ \ \ \ \ \ \ \ {t, 0.1, 2  \[Pi]*rev, 2  \[Pi]/22}, \n
        \t\t\ \ \ \ \ \ \ \ {\[Phi], \(-\[Pi]\) + 0.1, \[Pi] + 0.1, 
          2  \[Pi]/24}, \n\t\t\ \ \ \ \ \ \ \ Axes -> None, PlotRange -> All, 
        \n\t\t\ \ \ \ \ \ \ ViewPoint -> {0, \(-3\), 0.1}]; \n\t\t\t\)}], 
  "Input",
  CellLabel->"In[34]:=",
  CellTags->"POVRay"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.34968 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.183589 1.36718 2.22045e-16 1.36718 [
[ 0 0 0 0 ]
[ 1 1.34968 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.0061 0 m
0 1.34968 L
s
0 1.34968 m
.08751 1.2545 L
s
.08751 1.2545 m
.09167 .13961 L
s
.09167 .13961 m
.0061 0 L
s
.9939 0 m
.90833 .13961 L
s
.90833 .13961 m
.91249 1.2545 L
s
.91249 1.2545 m
1 1.34968 L
s
1 1.34968 m
.9939 0 L
s
.0061 0 m
0 1.34968 L
s
0 1.34968 m
1 1.34968 L
s
1 1.34968 m
.9939 0 L
s
.9939 0 m
.0061 0 L
s
.09167 .13961 m
.90833 .13961 L
s
.90833 .13961 m
.91249 1.2545 L
s
.91249 1.2545 m
.08751 1.2545 L
s
.08751 1.2545 m
.09167 .13961 L
s
0 0 m
1 0 L
1 1.34968 L
0 1.34968 L
closepath
clip
newpath
.5 Mabswid
.38782 .52241 m .38838 .47025 L .50612 .49174 L p .726 .591 .705 r
F P
0 g
s
.50612 .49174 m .50603 .54304 L .38782 .52241 L p .726 .591 .705 r
F P
0 g
s
.50703 .44152 m .50612 .49174 L .38838 .47025 L p .716 .622 .751 r
F P
0 g
s
.50603 .54304 m .50612 .49174 L .62303 .51059 L p .674 .549 .705 r
F P
0 g
s
.50612 .49174 m .50703 .44152 L .62006 .46092 L p .664 .58 .75 r
F P
0 g
s
.62006 .46092 m .62303 .51059 L .50612 .49174 L p .664 .58 .75 r
F P
0 g
s
.39155 .57254 m .38782 .52241 L .50603 .54304 L p .732 .557 .655 r
F P
0 g
s
.50603 .54304 m .50675 .59238 L .39155 .57254 L p .732 .557 .655 r
F P
0 g
s
.50675 .59238 m .50603 .54304 L .6234 .56142 L p .68 .515 .656 r
F P
0 g
s
.62303 .51059 m .6234 .56142 L .50603 .54304 L p .674 .549 .705 r
F P
0 g
s
.27594 .49896 m .27696 .44554 L .38838 .47025 L p .779 .635 .702 r
F P
0 g
s
.38838 .47025 m .38782 .52241 L .27594 .49896 L p .779 .635 .702 r
F P
0 g
s
.3932 .41919 m .38838 .47025 L .27696 .44554 L p .767 .666 .751 r
F P
0 g
s
.38838 .47025 m .3932 .41919 L .50703 .44152 L p .716 .622 .751 r
F P
0 g
s
.5087 .3954 m .50703 .44152 L .3932 .41919 L p .7 .651 .797 r
F P
0 g
s
.50703 .44152 m .5087 .3954 L .61463 .41542 L p .649 .611 .798 r
F P
0 g
s
.61463 .41542 m .62006 .46092 L .50703 .44152 L p .649 .611 .798 r
F P
0 g
s
.28259 .55034 m .27594 .49896 L .38782 .52241 L p .786 .598 .646 r
F P
0 g
s
.38782 .52241 m .39155 .57254 L .28259 .55034 L p .786 .598 .646 r
F P
0 g
s
.39939 .61761 m .39155 .57254 L .50675 .59238 L p .735 .515 .593 r
F P
0 g
s
.50675 .59238 m .50826 .63678 L .39939 .61761 L p .735 .515 .593 r
F P
0 g
s
.50826 .63678 m .50675 .59238 L .62114 .61039 L p .684 .475 .597 r
F P
0 g
s
.6234 .56142 m .62114 .61039 L .50675 .59238 L p .68 .515 .656 r
F P
0 g
s
.6234 .56142 m .62303 .51059 L .73203 .52741 L p .616 .502 .701 r
F P
0 g
s
.62303 .51059 m .62006 .46092 L .72536 .47802 L p .606 .536 .749 r
F P
0 g
s
.72536 .47802 m .73203 .52741 L .62303 .51059 L p .606 .536 .749 r
F P
0 g
s
.62114 .61039 m .6234 .56142 L .73284 .57812 L p .623 .467 .65 r
F P
0 g
s
.73203 .52741 m .73284 .57812 L .6234 .56142 L p .616 .502 .701 r
F P
0 g
s
.27696 .44554 m .28558 .39333 L .3932 .41919 L p .767 .666 .751 r
F P
0 g
s
.40203 .37229 m .3932 .41919 L .28558 .39333 L p .749 .694 .8 r
F P
0 g
s
.3932 .41919 m .40203 .37229 L .5087 .3954 L p .7 .651 .797 r
F P
0 g
s
.29656 .59653 m .28259 .55034 L .39155 .57254 L p .789 .55 .574 r
F P
0 g
s
.39155 .57254 m .39939 .61761 L .29656 .59653 L p .789 .55 .574 r
F P
0 g
s
.62006 .46092 m .61463 .41542 L .71319 .433 L p .588 .57 .801 r
F P
0 g
s
.71319 .433 m .72536 .47802 L .62006 .46092 L p .588 .57 .801 r
F P
0 g
s
.62114 .61039 m .61636 .65452 L .50826 .63678 L p .684 .475 .597 r
F P
0 g
s
.61636 .65452 m .62114 .61039 L .72773 .62708 L p .628 .426 .59 r
F P
0 g
s
.73284 .57812 m .72773 .62708 L .62114 .61039 L p .623 .467 .65 r
F P
0 g
s
.51104 .35626 m .5087 .3954 L .40203 .37229 L p .671 .68 .849 r
F P
0 g
s
.5087 .3954 m .51104 .35626 L .60704 .37695 L p .619 .644 .853 r
F P
0 g
s
.60704 .37695 m .61463 .41542 L .5087 .3954 L p .619 .644 .853 r
F P
0 g
s
.41093 .65477 m .39939 .61761 L .50826 .63678 L p .734 .454 .504 r
F P
0 g
s
.50826 .63678 m .51047 .67345 L .41093 .65477 L p .734 .454 .504 r
F P
0 g
s
.51047 .67345 m .50826 .63678 L .61636 .65452 L p .684 .42 .516 r
F P
0 g
s
.28558 .39333 m .30133 .34552 L .40203 .37229 L p .749 .694 .8 r
F P
0 g
s
.41439 .3325 m .40203 .37229 L .30133 .34552 L p .718 .72 .851 r
F P
0 g
s
.40203 .37229 m .41439 .3325 L .51104 .35626 L p .671 .68 .849 r
F P
0 g
s
.31711 .63458 m .29656 .59653 L .39939 .61761 L p .784 .476 .467 r
F P
0 g
s
.39939 .61761 m .41093 .65477 L .31711 .63458 L p .784 .476 .467 r
F P
0 g
s
.61463 .41542 m .60704 .37695 L .6962 .3952 L p .554 .608 .861 r
F P
0 g
s
.6962 .3952 m .71319 .433 L .61463 .41542 L p .554 .608 .861 r
F P
0 g
s
.17772 .47221 m .17916 .41711 L .27696 .44554 L p .839 .685 .692 r
F P
0 g
s
.27696 .44554 m .27594 .49896 L .17772 .47221 L p .839 .685 .692 r
F P
0 g
s
.28558 .39333 m .27696 .44554 L .17916 .41711 L p .823 .719 .75 r
F P
0 g
s
.61636 .65452 m .60931 .69102 L .51047 .67345 L p .684 .42 .516 r
F P
0 g
s
.60931 .69102 m .61636 .65452 L .71698 .67127 L p .63 .373 .51 r
F P
0 g
s
.72773 .62708 m .71698 .67127 L .61636 .65452 L p .628 .426 .59 r
F P
0 g
s
.18707 .52533 m .17772 .47221 L .27594 .49896 L p .848 .641 .623 r
F P
0 g
s
.27594 .49896 m .28259 .55034 L .18707 .52533 L p .848 .641 .623 r
F P
0 g
s
.17916 .41711 m .19129 .36346 L .28558 .39333 L p .823 .719 .75 r
F P
0 g
s
.30133 .34552 m .28558 .39333 L .19129 .36346 L p .799 .747 .804 r
F P
0 g
s
.20668 .57313 m .18707 .52533 L .28259 .55034 L p .85 .58 .53 r
F P
0 g
s
.28259 .55034 m .29656 .59653 L .20668 .57313 L p .85 .58 .53 r
F P
0 g
s
.73284 .57812 m .73203 .52741 L .82607 .54285 L p .543 .443 .691 r
F P
0 g
s
.73203 .52741 m .72536 .47802 L .81608 .49352 L p .529 .481 .747 r
F P
0 g
s
.81608 .49352 m .82607 .54285 L .73203 .52741 L p .529 .481 .747 r
F P
0 g
s
.51393 .32672 m .51104 .35626 L .41439 .3325 L p .614 .708 .912 r
F P
0 g
s
.51104 .35626 m .51393 .32672 L .59772 .34809 L p .555 .678 .921 r
F P
0 g
s
.59772 .34809 m .60704 .37695 L .51104 .35626 L p .555 .678 .921 r
F P
0 g
s
.72773 .62708 m .73284 .57812 L .82728 .59372 L p .552 .404 .634 r
F P
0 g
s
.82607 .54285 m .82728 .59372 L .73284 .57812 L p .543 .443 .691 r
F P
0 g
s
.42554 .6815 m .41093 .65477 L .51047 .67345 L p .718 .348 .348 r
F P
0 g
s
.51047 .67345 m .51326 .69988 L .42554 .6815 L p .718 .348 .348 r
F P
0 g
s
.51326 .69988 m .51047 .67345 L .60931 .69102 L p .676 .332 .384 r
F P
0 g
s
.30133 .34552 m .32334 .30513 L .41439 .3325 L p .718 .72 .851 r
F P
0 g
s
.42959 .30247 m .41439 .3325 L .32334 .30513 L p .661 .741 .909 r
F P
0 g
s
.41439 .3325 m .42959 .30247 L .51393 .32672 L p .614 .708 .912 r
F P
0 g
s
.72536 .47802 m .71319 .433 L .79787 .44884 L p .506 .521 .806 r
F P
0 g
s
.79787 .44884 m .81608 .49352 L .72536 .47802 L p .506 .521 .806 r
F P
0 g
s
.71698 .67127 m .72773 .62708 L .81961 .64297 L p .559 .36 .568 r
F P
0 g
s
.82728 .59372 m .81961 .64297 L .72773 .62708 L p .552 .404 .634 r
F P
0 g
s
.34306 .66191 m .31711 .63458 L .41093 .65477 L p .751 .337 .266 r
F P
0 g
s
.41093 .65477 m .42554 .6815 L .34306 .66191 L p .751 .337 .266 r
F P
0 g
s
.19129 .36346 m .21341 .31462 L .30133 .34552 L p .799 .747 .804 r
F P
0 g
s
.32334 .30513 m .30133 .34552 L .21341 .31462 L p .762 .768 .858 r
F P
0 g
s
.60704 .37695 m .59772 .34809 L .67537 .36716 L p .473 .647 .934 r
F P
0 g
s
.67537 .36716 m .6962 .3952 L .60704 .37695 L p .473 .647 .934 r
F P
0 g
s
.23546 .61252 m .20668 .57313 L .29656 .59653 L p .834 .479 .383 r
F P
0 g
s
.29656 .59653 m .31711 .63458 L .23546 .61252 L p .834 .479 .383 r
F P
0 g
s
.60931 .69102 m .60038 .7174 L .51326 .69988 L p .676 .332 .384 r
F P
0 g
s
.60038 .7174 m .60931 .69102 L .70115 .70787 L p .628 .296 .389 r
F P
0 g
s
.71698 .67127 m .70115 .70787 L .60931 .69102 L p .63 .373 .51 r
F P
0 g
s
.71319 .433 m .6962 .3952 L .77252 .41168 L p .46 .564 .874 r
F P
0 g
s
.77252 .41168 m .79787 .44884 L .71319 .433 L p .46 .564 .874 r
F P
0 g
s
.70115 .70787 m .71698 .67127 L .80349 .68748 L p .565 .308 .485 r
F P
0 g
s
.81961 .64297 m .80349 .68748 L .71698 .67127 L p .559 .36 .568 r
F P
0 g
s
.5172 .30898 m .51393 .32672 L .42959 .30247 L p .468 .714 .976 r
F P
0 g
s
.51393 .32672 m .5172 .30898 L .58723 .33099 L p .369 .681 .978 r
F P
0 g
s
.58723 .33099 m .59772 .34809 L .51393 .32672 L p .369 .681 .978 r
F P
0 g
s
.21341 .31462 m .24422 .27373 L .32334 .30513 L p .762 .768 .858 r
F P
0 g
s
.35035 .27486 m .32334 .30513 L .24422 .27373 L p .7 .78 .913 r
F P
0 g
s
.32334 .30513 m .35035 .27486 L .42959 .30247 L p .661 .741 .909 r
F P
0 g
s
.42554 .6815 m .51326 .69988 L p .46038 .6968 L .625 .11 0 r
F P
0 g
s
.51326 .69988 m .51647 .71406 L .44234 .69575 L p .625 .11 0 r
F P
0 g
s
.51647 .71406 m .51326 .69988 L .60038 .7174 L p .629 .158 .116 r
F P
0 g
s
.60038 .7174 m .5901 .73167 L .51647 .71406 L p .629 .158 .116 r
F P
0 g
s
.5199 .71462 m .4603 .69615 L p .50753 .71121 L 0 .401 .699 r
F P
0 g
s
.51647 .71406 m .5173 .7142 L p .50753 .71121 L 0 .401 .699 r
F P
0 g
s
.5173 .7142 m .5199 .71462 L p .50753 .71121 L 0 .401 .699 r
F P
0 g
s
.44234 .69575 m .43411 .68877 L p .46038 .6968 L .625 .11 0 r
F P
0 g
s
.43411 .68877 m .42554 .6815 L p .46038 .6968 L .625 .11 0 r
F P
0 g
s
.27168 .6408 m .23546 .61252 L .31711 .63458 L p .755 .274 .097 r
F P
0 g
s
.31711 .63458 m .34306 .66191 L .27168 .6408 L p .755 .274 .097 r
F P
0 g
s
.44672 .28444 m .42959 .30247 L .35035 .27486 L p .538 .74 .969 r
F P
0 g
s
.42959 .30247 m .44672 .28444 L .5172 .30898 L p .468 .714 .976 r
F P
0 g
s
.34306 .66191 m .42554 .6815 L p .38634 .67773 L .568 0 0 r
F P
0 g
s
.42554 .6815 m .44234 .69575 L .37286 .67644 L p .568 0 0 r
F P
0 g
s
.4603 .69615 m .40463 .67682 L p .43634 .69273 L .043 .555 .874 r
F P
0 g
s
.44234 .69575 m .44671 .69584 L p .43634 .69273 L .043 .555 .874 r
F P
0 g
s
.44671 .69584 m .4603 .69615 L p .43634 .69273 L .043 .555 .874 r
F P
0 g
s
.37286 .67644 m .35821 .66929 L p .38634 .67773 L .568 0 0 r
F P
0 g
s
.35821 .66929 m .34306 .66191 L p .38634 .67773 L .568 0 0 r
F P
0 g
s
.59772 .34809 m .58723 .33099 L .65197 .35096 L p .221 .628 .955 r
F P
0 g
s
.65197 .35096 m .67537 .36716 L .59772 .34809 L p .221 .628 .955 r
F P
0 g
s
.6962 .3952 m .67537 .36716 L .74154 .38456 L p .346 .602 .943 r
F P
0 g
s
.74154 .38456 m .77252 .41168 L .6962 .3952 L p .346 .602 .943 r
F P
0 g
s
.5901 .73167 m .60038 .7174 L .68117 .73439 L p .607 .165 .174 r
F P
0 g
s
.70115 .70787 m .68117 .73439 L .60038 .7174 L p .628 .296 .389 r
F P
0 g
s
.68117 .73439 m .70115 .70787 L .77985 .72439 L p .57 .239 .368 r
F P
0 g
s
.80349 .68748 m .77985 .72439 L .70115 .70787 L p .565 .308 .485 r
F P
0 g
s
.10072 .44179 m .10253 .38459 L .17916 .41711 L p .912 .747 .664 r
F P
0 g
s
.17916 .41711 m .17772 .47221 L .10072 .44179 L p .912 .747 .664 r
F P
0 g
s
.19129 .36346 m .17916 .41711 L .10253 .38459 L p .889 .788 .743 r
F P
0 g
s
.11237 .49716 m .10072 .44179 L .17772 .47221 L p .923 .688 .565 r
F P
0 g
s
.17772 .47221 m .18707 .52533 L .11237 .49716 L p .923 .688 .565 r
F P
0 g
s
.10253 .38459 m .11765 .32923 L .19129 .36346 L p .889 .788 .743 r
F P
0 g
s
.21341 .31462 m .19129 .36346 L .11765 .32923 L p .855 .815 .81 r
F P
0 g
s
.13678 .5471 m .11237 .49716 L .18707 .52533 L p .915 .595 .423 r
F P
0 g
s
.18707 .52533 m .20668 .57313 L .13678 .5471 L p .915 .595 .423 r
F P
0 g
s
.82728 .59372 m .82607 .54285 L .89832 .55755 L p .429 .352 .668 r
F P
0 g
s
.82607 .54285 m .81608 .49352 L .8856 .50811 L p .407 .398 .737 r
F P
0 g
s
.8856 .50811 m .89832 .55755 L .82607 .54285 L p .407 .398 .737 r
F P
0 g
s
.81961 .64297 m .82728 .59372 L .89985 .60879 L p .445 .306 .598 r
F P
0 g
s
.89832 .55755 m .89985 .60879 L .82728 .59372 L p .429 .352 .668 r
F P
0 g
s
.81608 .49352 m .79787 .44884 L .86247 .46365 L p .37 .445 .808 r
F P
0 g
s
.86247 .46365 m .8856 .50811 L .81608 .49352 L p .37 .445 .808 r
F P
0 g
s
.24422 .27373 m .28188 .24355 L .35035 .27486 L p .7 .78 .913 r
F P
0 g
s
.38072 .25696 m .35035 .27486 L .28188 .24355 L p .589 .766 .964 r
F P
0 g
s
.35035 .27486 m .38072 .25696 L .44672 .28444 L p .538 .74 .969 r
F P
0 g
s
.80349 .68748 m .81961 .64297 L .89007 .65854 L p .458 .258 .52 r
F P
0 g
s
.89985 .60879 m .89007 .65854 L .81961 .64297 L p .445 .306 .598 r
F P
0 g
s
.11765 .32923 m .14516 .27931 L .21341 .31462 L p .855 .815 .81 r
F P
0 g
s
.24422 .27373 m .21341 .31462 L .14516 .27931 L p .805 .829 .87 r
F P
0 g
s
.31312 .65588 m .27168 .6408 L .34306 .66191 L p .405 0 0 r
F P
0 g
s
.34306 .66191 m .37286 .67644 L .31312 .65588 L p .405 0 0 r
F P
0 g
s
.52064 .30462 m .5172 .30898 L .44672 .28444 L p .035 .541 .875 r
F P
0 g
s
.46472 .28 m .52064 .30462 L p .50294 .29979 L .035 .541 .875 r
F P
0 g
s
.44672 .28444 m .46472 .28 L p .50294 .29979 L .035 .541 .875 r
F P
0 g
s
.38072 .25696 m .41252 .25297 L .46472 .28 L p .235 .635 .958 r
F P
0 g
s
.46472 .28 m .41252 .25297 L p .41312 .25329 L .36 0 0 r
F P
0 g
s
.52064 .30462 m .46472 .28 L p .46545 .28043 L .594 .023 0 r
F P
0 g
s
.5172 .30898 m .52064 .30462 L p .56214 .32282 L 0 .391 .692 r
F P
0 g
s
.50905 .30575 m .52064 .30462 L .46545 .28043 L .594 .023 0 r
F
.58723 .33099 m .5172 .30898 L p .56214 .32282 L 0 .391 .692 r
F P
0 g
s
.52401 .31444 m .52251 .31008 L p .50905 .30575 L .594 .023 0 r
F P
0 g
s
.52251 .31008 m .52064 .30462 L p .50905 .30575 L .594 .023 0 r
F P
0 g
s
.57622 .32716 m .58347 .32968 L p .56214 .32282 L 0 .391 .692 r
F P
0 g
s
.58347 .32968 m .58723 .33099 L p .56214 .32282 L 0 .391 .692 r
F P
0 g
s
.46472 .28 m .44672 .28444 L .38072 .25696 L p .235 .635 .958 r
F P
0 g
s
.48235 .28995 m .52401 .31444 L p .50598 .30397 L .594 .023 0 r
F P
0 g
s
.46472 .28 m .48235 .28995 L p .50598 .30397 L .594 .023 0 r
F P
0 g
s
.48235 .28995 m .46472 .28 L p .41312 .25329 L .36 0 0 r
F P
0 g
s
.44361 .26361 m .48235 .28995 L p .46131 .27881 L .36 0 0 r
F P
0 g
s
.41252 .25297 m .44361 .26361 L p .46131 .27881 L .36 0 0 r
F P
0 g
s
.36805 .22334 m .41084 .23531 L .44361 .26361 L p 0 .451 .861 r
F P
0 g
s
.44361 .26361 m .41084 .23531 L p .42197 .24504 L .484 0 0 r
F P
0 g
s
.48235 .28995 m .44361 .26361 L p .4538 .273 L .77 .262 0 r
F P
0 g
s
.49836 .31405 m .48235 .28995 L p .4538 .273 L .77 .262 0 r
F P
0 g
s
.44361 .26361 m .41252 .25297 L .36805 .22334 L p 0 .451 .861 r
F P
0 g
s
.47175 .28857 m .49836 .31405 L p .48235 .2993 L .77 .262 0 r
F P
0 g
s
.44361 .26361 m .47175 .28857 L p .48235 .2993 L .77 .262 0 r
F P
0 g
s
.47175 .28857 m .44361 .26361 L p .42197 .24504 L .484 0 0 r
F P
0 g
s
.4494 .26169 m .47175 .28857 L p .4543 .27331 L .484 0 0 r
F P
0 g
s
.41084 .23531 m .4494 .26169 L p .4543 .27331 L .484 0 0 r
F P
0 g
s
.4494 .26169 m .43324 .23342 L p .44378 .24835 L 0 .103 .656 r
F P
0 g
s
.47175 .28857 m .4494 .26169 L p .45969 .27636 L .797 .335 0 r
F P
0 g
s
.4948 .3264 m .47175 .28857 L p .45969 .27636 L .797 .335 0 r
F P
0 g
s
.5115 .35097 m .49836 .31405 L .47175 .28857 L p .89 .502 .29 r
F P
0 g
s
.4948 .3264 m .5115 .35097 L p .50185 .33583 L .89 .502 .29 r
F P
0 g
s
.47175 .28857 m .48845 .31598 L p .50185 .33583 L .89 .502 .29 r
F P
0 g
s
.48845 .31598 m .4948 .3264 L p .50185 .33583 L .89 .502 .29 r
F P
0 g
s
.48088 .30088 m .46444 .28041 L p .44378 .24835 L 0 .103 .656 r
F P
0 g
s
.46444 .28041 m .4494 .26169 L p .44543 .25226 L 0 .103 .656 r
F P
0 g
s
.46444 .28041 m .44543 .25226 L .44378 .24835 L 0 .103 .656 r
F
.52064 .30462 m .57622 .32716 L p .56214 .32282 L 0 .391 .692 r
F P
0 g
s
.17246 .58829 m .13678 .5471 L .20668 .57313 L p .86 .43 .191 r
F P
0 g
s
.20668 .57313 m .23546 .61252 L .17246 .58829 L p .86 .43 .191 r
F P
0 g
s
.67537 .36716 m .65197 .35096 L .70686 .36944 L p .012 .526 .862 r
F P
0 g
s
.70686 .36944 m .74154 .38456 L .67537 .36716 L p .012 .526 .862 r
F P
0 g
s
.4603 .69615 m .44234 .69575 L .51647 .71406 L p 0 .401 .699 r
F P
0 g
s
.5199 .71462 m .51647 .71406 L .5901 .73167 L p .38 0 0 r
F P
0 g
s
.79787 .44884 m .77252 .41168 L .83037 .4271 L p .296 .492 .88 r
F P
0 g
s
.83037 .4271 m .86247 .46365 L .79787 .44884 L p .296 .492 .88 r
F P
0 g
s
.68117 .73439 m .65822 .74884 L .5901 .73167 L p .607 .165 .174 r
F P
0 g
s
.65822 .74884 m .68117 .73439 L .75008 .75116 L p .57 .143 .191 r
F P
0 g
s
.77985 .72439 m .75008 .75116 L .68117 .73439 L p .57 .239 .368 r
F P
0 g
s
.77985 .72439 m .80349 .68748 L .86958 .70356 L p .473 .206 .428 r
F P
0 g
s
.89007 .65854 m .86958 .70356 L .80349 .68748 L p .458 .258 .52 r
F P
0 g
s
.40463 .67682 m .37286 .67644 L .44234 .69575 L p .043 .555 .874 r
F P
0 g
s
.58723 .33099 m .57622 .32716 L .62749 .34801 L p .378 0 0 r
F P
0 g
s
.62749 .34801 m .65197 .35096 L .58723 .33099 L p .378 0 0 r
F P
0 g
s
.5901 .73167 m .57912 .73246 L .5199 .71462 L p .38 0 0 r
F P
0 g
s
.57912 .73246 m .5901 .73167 L .65822 .74884 L p .497 0 0 r
F P
0 g
s
.14516 .27931 m .18333 .2381 L .24422 .27373 L p .805 .829 .87 r
F P
0 g
s
.28188 .24355 m .24422 .27373 L .18333 .2381 L p .733 .826 .923 r
F P
0 g
s
.21717 .61792 m .17246 .58829 L .23546 .61252 L p .655 .099 0 r
F P
0 g
s
.23546 .61252 m .27168 .6408 L .21717 .61792 L p .655 .099 0 r
F P
0 g
s
.77252 .41168 m .74154 .38456 L .7913 .40094 L p .125 .512 .911 r
F P
0 g
s
.7913 .40094 m .83037 .4271 L .77252 .41168 L p .125 .512 .911 r
F P
0 g
s
.28188 .24355 m .32407 .22627 L .38072 .25696 L p .589 .766 .964 r
F P
0 g
s
.41252 .25297 m .38072 .25696 L .32407 .22627 L p .376 .687 .98 r
F P
0 g
s
.75008 .75116 m .77985 .72439 L .8396 .74088 L p .491 .15 .315 r
F P
0 g
s
.86958 .70356 m .8396 .74088 L .77985 .72439 L p .473 .206 .428 r
F P
0 g
s
.3571 .65643 m .31312 .65588 L .37286 .67644 L p .233 .655 .959 r
F P
0 g
s
.37286 .67644 m .40463 .67682 L .3571 .65643 L p .233 .655 .959 r
F P
0 g
s
.65197 .35096 m .62749 .34801 L .67073 .36762 L p .476 0 0 r
F P
0 g
s
.67073 .36762 m .70686 .36944 L .65197 .35096 L p .476 0 0 r
F P
0 g
s
.18333 .2381 m .22976 .20839 L .28188 .24355 L p .733 .826 .923 r
F P
0 g
s
.32407 .22627 m .28188 .24355 L .22976 .20839 L p .627 .796 .964 r
F P
0 g
s
.65822 .74884 m .63377 .74991 L .57912 .73246 L p .497 0 0 r
F P
0 g
s
.63377 .74991 m .65822 .74884 L .71604 .76584 L p .547 .009 0 r
F P
0 g
s
.75008 .75116 m .71604 .76584 L .65822 .74884 L p .57 .143 .191 r
F P
0 g
s
.52064 .30462 m .52401 .31444 L .56544 .33733 L p .665 .156 .025 r
F P
0 g
s
.56544 .33733 m .57622 .32716 L .52064 .30462 L p .665 .156 .025 r
F P
0 g
s
.26803 .63385 m .21717 .61792 L .27168 .6408 L p 0 .415 .771 r
F P
0 g
s
.27168 .6408 m .31312 .65588 L .26803 .63385 L p 0 .415 .771 r
F P
0 g
s
.47821 .68219 m .4603 .69615 L .5199 .71462 L p .32 .641 .966 r
F P
0 g
s
.5199 .71462 m .52331 .70102 L .47821 .68219 L p .32 .641 .966 r
F P
0 g
s
.52331 .70102 m .5199 .71462 L .57912 .73246 L p .109 .546 .915 r
F P
0 g
s
.74779 .38703 m .7913 .40094 L .74154 .38456 L p .229 0 0 r
F P
0 g
s
.74154 .38456 m .70686 .36944 L .74779 .38703 L p .229 0 0 r
F P
0 g
s
.43541 .89996 m .43579 .86158 L .52385 .87672 L p .712 .578 .704 r
F P
0 g
s
.52385 .87672 m .52382 .91433 L .43541 .89996 L p .712 .578 .704 r
F P
0 g
s
.52419 .84075 m .52385 .87672 L .43579 .86158 L p .705 .601 .738 r
F P
0 g
s
.52382 .91433 m .52385 .87672 L .60967 .89189 L p .673 .547 .703 r
F P
0 g
s
.52385 .87672 m .52419 .84075 L .60707 .85653 L p .666 .57 .738 r
F P
0 g
s
.60707 .85653 m .60967 .89189 L .52385 .87672 L p .666 .57 .738 r
F P
0 g
s
.43803 .93755 m .43541 .89996 L .52382 .91433 L p .716 .554 .668 r
F P
0 g
s
.52382 .91433 m .52411 .9512 L .43803 .93755 L p .716 .554 .668 r
F P
0 g
s
.52411 .9512 m .52382 .91433 L .60998 .92889 L p .678 .523 .668 r
F P
0 g
s
.60967 .89189 m .60998 .92889 L .52382 .91433 L p .673 .547 .703 r
F P
0 g
s
.35109 .8856 m .3518 .8463 L .43579 .86158 L p .752 .611 .702 r
F P
0 g
s
.43579 .86158 m .43541 .89996 L .35109 .8856 L p .752 .611 .702 r
F P
0 g
s
.43914 .82485 m .43579 .86158 L .3518 .8463 L p .744 .635 .739 r
F P
0 g
s
.43579 .86158 m .43914 .82485 L .52419 .84075 L p .705 .601 .738 r
F P
0 g
s
.35596 .92406 m .35109 .8856 L .43541 .89996 L p .756 .584 .664 r
F P
0 g
s
.43541 .89996 m .43803 .93755 L .35596 .92406 L p .756 .584 .664 r
F P
0 g
s
.43622 .66259 m .40463 .67682 L .4603 .69615 L p .452 .693 .97 r
F P
0 g
s
.4603 .69615 m .47821 .68219 L .43622 .66259 L p .452 .693 .97 r
F P
0 g
s
.57622 .32716 m .56544 .33733 L .60358 .35891 L p .661 .206 .161 r
F P
0 g
s
.60358 .35891 m .62749 .34801 L .57622 .32716 L p .661 .206 .161 r
F P
0 g
s
.71604 .76584 m .75008 .75116 L .80202 .76797 L p .518 .093 .173 r
F P
0 g
s
.8396 .74088 m .80202 .76797 L .75008 .75116 L p .491 .15 .315 r
F P
0 g
s
.52481 .80873 m .52419 .84075 L .43914 .82485 L p .694 .625 .776 r
F P
0 g
s
.52419 .84075 m .52481 .80873 L .60234 .82508 L p .653 .595 .777 r
F P
0 g
s
.60234 .82508 m .60707 .85653 L .52419 .84075 L p .653 .595 .777 r
F P
0 g
s
.3518 .8463 m .35806 .80866 L .43914 .82485 L p .744 .635 .739 r
F P
0 g
s
.44526 .79214 m .43914 .82485 L .35806 .80866 L p .732 .66 .778 r
F P
0 g
s
.43914 .82485 m .44526 .79214 L .52481 .80873 L p .694 .625 .776 r
F P
0 g
s
.44349 .97193 m .43803 .93755 L .52411 .9512 L p .719 .525 .627 r
F P
0 g
s
.52411 .9512 m .52468 .98496 L .44349 .97193 L p .719 .525 .627 r
F P
0 g
s
.52468 .98496 m .52411 .9512 L .60798 .96516 L p .682 .495 .628 r
F P
0 g
s
.60998 .92889 m .60798 .96516 L .52411 .9512 L p .678 .523 .668 r
F P
0 g
s
.57912 .73246 m .56816 .71927 L .52331 .70102 L p .109 .546 .915 r
F P
0 g
s
.56816 .71927 m .57912 .73246 L .63377 .74991 L p 0 .366 .737 r
F P
0 g
s
.36613 .95919 m .35596 .92406 L .43803 .93755 L p .759 .552 .617 r
F P
0 g
s
.43803 .93755 m .44349 .97193 L .36613 .95919 L p .759 .552 .617 r
F P
0 g
s
.60998 .92889 m .60967 .89189 L .68716 .90721 L p .63 .512 .7 r
F P
0 g
s
.60967 .89189 m .60707 .85653 L .68188 .87233 L p .621 .537 .738 r
F P
0 g
s
.68188 .87233 m .68716 .90721 L .60967 .89189 L p .621 .537 .738 r
F P
0 g
s
.60798 .96516 m .60998 .92889 L .68778 .94371 L p .637 .487 .663 r
F P
0 g
s
.68716 .90721 m .68778 .94371 L .60998 .92889 L p .63 .512 .7 r
F P
0 g
s
.52569 .78276 m .52481 .80873 L .44526 .79214 L p .672 .653 .823 r
F P
0 g
s
.52481 .80873 m .52569 .78276 L .59578 .79962 L p .628 .625 .828 r
F P
0 g
s
.59578 .79962 m .60234 .82508 L .52481 .80873 L p .628 .625 .828 r
F P
0 g
s
.89985 .60879 m .89832 .55755 L .9426 .57213 L p .196 .166 .596 r
F P
0 g
s
.89832 .55755 m .8856 .50811 L .928 .52246 L p .148 .222 .684 r
F P
0 g
s
.928 .52246 m .9426 .57213 L .89832 .55755 L p .148 .222 .684 r
F P
0 g
s
.35806 .80866 m .36947 .77512 L .44526 .79214 L p .732 .66 .778 r
F P
0 g
s
.45377 .76558 m .44526 .79214 L .36947 .77512 L p .711 .686 .824 r
F P
0 g
s
.44526 .79214 m .45377 .76558 L .52569 .78276 L p .672 .653 .823 r
F P
0 g
s
.32407 .22627 m .36805 .22334 L .41252 .25297 L p .376 .687 .98 r
F P
0 g
s
.89007 .65854 m .89985 .60879 L .94435 .62387 L p .233 .113 .503 r
F P
0 g
s
.9426 .57213 m .94435 .62387 L .89985 .60879 L p .196 .166 .596 r
F P
0 g
s
.60707 .85653 m .60234 .82508 L .67228 .84135 L p .604 .564 .782 r
F P
0 g
s
.67228 .84135 m .68188 .87233 L .60707 .85653 L p .604 .564 .782 r
F P
0 g
s
.0526 .40757 m .05467 .34789 L .10253 .38459 L p .99 .816 .574 r
F P
0 g
s
.10253 .38459 m .10072 .44179 L .0526 .40757 L p .99 .816 .574 r
F P
0 g
s
.11765 .32923 m .10253 .38459 L .05467 .34789 L p .961 .876 .705 r
F P
0 g
s
.45147 1.00087 m .44349 .97193 L .52468 .98496 L p .721 .486 .57 r
F P
0 g
s
.52468 .98496 m .52552 1.01339 L .45147 1.00087 L p .721 .486 .57 r
F P
0 g
s
.52552 1.01339 m .52468 .98496 L .60379 .99837 L p .685 .46 .575 r
F P
0 g
s
.60798 .96516 m .60379 .99837 L .52468 .98496 L p .682 .495 .628 r
F P
0 g
s
.27722 .87109 m .27824 .83068 L .3518 .8463 L p .799 .651 .698 r
F P
0 g
s
.3518 .8463 m .35109 .8856 L .27722 .87109 L p .799 .651 .698 r
F P
0 g
s
.35806 .80866 m .3518 .8463 L .27824 .83068 L p .79 .678 .741 r
F P
0 g
s
.22976 .20839 m .28146 .19229 L .32407 .22627 L p .627 .796 .964 r
F P
0 g
s
.36805 .22334 m .32407 .22627 L .28146 .19229 L p .472 .721 .978 r
F P
0 g
s
.28411 .91058 m .27722 .87109 L .35109 .8856 L p .804 .619 .651 r
F P
0 g
s
.35109 .8856 m .35596 .92406 L .28411 .91058 L p .804 .619 .651 r
F P
0 g
s
.06594 .46567 m .0526 .40757 L .10072 .44179 L p .985 .709 .395 r
F P
0 g
s
.10072 .44179 m .11237 .49716 L .06594 .46567 L p .985 .709 .395 r
F P
0 g
s
.60379 .99837 m .60798 .96516 L .6837 .97949 L p .642 .46 .621 r
F P
0 g
s
.68778 .94371 m .6837 .97949 L .60798 .96516 L p .637 .487 .663 r
F P
0 g
s
.38098 .98867 m .36613 .95919 L .44349 .97193 L p .759 .506 .551 r
F P
0 g
s
.44349 .97193 m .45147 1.00087 L .38098 .98867 L p .759 .506 .551 r
F P
0 g
s
.8856 .50811 m .86247 .46365 L .90153 .47815 L p .077 .277 .764 r
F P
0 g
s
.90153 .47815 m .928 .52246 L .8856 .50811 L p .077 .277 .764 r
F P
0 g
s
.05467 .34789 m .07199 .29062 L .11765 .32923 L p .961 .876 .705 r
F P
0 g
s
.14516 .27931 m .11765 .32923 L .07199 .29062 L p .911 .904 .805 r
F P
0 g
s
.40064 .64207 m .3571 .65643 L .40463 .67682 L p .538 .732 .964 r
F P
0 g
s
.40463 .67682 m .43622 .66259 L .40064 .64207 L p .538 .732 .964 r
F P
0 g
s
.27824 .83068 m .28711 .79197 L .35806 .80866 L p .79 .678 .741 r
F P
0 g
s
.36947 .77512 m .35806 .80866 L .28711 .79197 L p .775 .704 .785 r
F P
0 g
s
.86958 .70356 m .89007 .65854 L .93312 .67423 L p .27 .065 .406 r
F P
0 g
s
.94435 .62387 m .93312 .67423 L .89007 .65854 L p .233 .113 .503 r
F P
0 g
s
.62749 .34801 m .60358 .35891 L .63559 .37953 L p .631 .21 .226 r
F P
0 g
s
.63559 .37953 m .67073 .36762 L .62749 .34801 L p .631 .21 .226 r
F P
0 g
s
.29851 .94657 m .28411 .91058 L .35596 .92406 L p .807 .578 .591 r
F P
0 g
s
.35596 .92406 m .36613 .95919 L .29851 .94657 L p .807 .578 .591 r
F P
0 g
s
.09381 .51825 m .06594 .46567 L .11237 .49716 L p .917 .525 .136 r
F P
0 g
s
.11237 .49716 m .13678 .5471 L .09381 .51825 L p .917 .525 .136 r
F P
0 g
s
.32169 .63482 m .26803 .63385 L .31312 .65588 L p .368 .716 .987 r
F P
0 g
s
.31312 .65588 m .3571 .65643 L .32169 .63482 L p .368 .716 .987 r
F P
0 g
s
.70686 .36944 m .67073 .36762 L .70268 .38648 L p .509 0 0 r
F P
0 g
s
.70268 .38648 m .74779 .38703 L .70686 .36944 L p .509 0 0 r
F P
0 g
s
.60234 .82508 m .59578 .79962 L .65894 .8163 L p .57 .598 .841 r
F P
0 g
s
.65894 .8163 m .67228 .84135 L .60234 .82508 L p .57 .598 .841 r
F P
0 g
s
.86247 .46365 m .83037 .4271 L .86491 .44218 L p 0 .321 .808 r
F P
0 g
s
.86491 .44218 m .90153 .47815 L .86247 .46365 L p 0 .321 .808 r
F P
0 g
s
.07199 .29062 m .10338 .23963 L .14516 .27931 L p .911 .904 .805 r
F P
0 g
s
.18333 .2381 m .14516 .27931 L .10338 .23963 L p .842 .906 .881 r
F P
0 g
s
.63377 .74991 m .60946 .73711 L .56816 .71927 L p 0 .366 .737 r
F P
0 g
s
.71604 .76584 m .67991 .76715 L .63377 .74991 L p .547 .009 0 r
F P
0 g
s
.60946 .73711 m .63377 .74991 L .67991 .76715 L p .47 0 0 r
F P
0 g
s
.60379 .99837 m .59767 1.02634 L .52552 1.01339 L p .685 .46 .575 r
F P
0 g
s
.59767 1.02634 m .60379 .99837 L .67517 1.01221 L p .647 .425 .568 r
F P
0 g
s
.6837 .97949 m .67517 1.01221 L .60379 .99837 L p .642 .46 .621 r
F P
0 g
s
.52676 .76461 m .52569 .78276 L .45377 .76558 L p .621 .686 .89 r
F P
0 g
s
.52569 .78276 m .52676 .76461 L .58778 .78187 L p .564 .663 .904 r
F P
0 g
s
.58778 .78187 m .59578 .79962 L .52569 .78276 L p .564 .663 .904 r
F P
0 g
s
.36947 .77512 m .38533 .74789 L .45377 .76558 L p .711 .686 .824 r
F P
0 g
s
.46415 .74698 m .45377 .76558 L .38533 .74789 L p .664 .714 .885 r
F P
0 g
s
.45377 .76558 m .46415 .74698 L .52676 .76461 L p .621 .686 .89 r
F P
0 g
s
.28711 .79197 m .30326 .75749 L .36947 .77512 L p .775 .704 .785 r
F P
0 g
s
.38533 .74789 m .36947 .77512 L .30326 .75749 L p .749 .728 .833 r
F P
0 g
s
.67991 .76715 m .71604 .76584 L .75924 .78286 L p .558 .053 .012 r
F P
0 g
s
.80202 .76797 m .75924 .78286 L .71604 .76584 L p .518 .093 .173 r
F P
0 g
s
.8396 .74088 m .86958 .70356 L .90963 .71984 L p .313 .025 .305 r
F P
0 g
s
.93312 .67423 m .90963 .71984 L .86958 .70356 L p .27 .065 .406 r
F P
0 g
s
.46147 1.02238 m .45147 1.00087 L .52552 1.01339 L p .718 .423 .476 r
F P
0 g
s
.52552 1.01339 m .52656 1.03458 L .46147 1.02238 L p .718 .423 .476 r
F P
0 g
s
.52656 1.03458 m .52552 1.01339 L .59767 1.02634 L p .686 .406 .493 r
F P
0 g
s
.3195 .97667 m .29851 .94657 L .36613 .95919 L p .803 .516 .503 r
F P
0 g
s
.36613 .95919 m .38098 .98867 L .3195 .97667 L p .803 .516 .503 r
F P
0 g
s
.13439 .56173 m .09381 .51825 L .13678 .5471 L p .718 .216 0 r
F P
0 g
s
.13678 .5471 m .17246 .58829 L .13439 .56173 L p .718 .216 0 r
F P
0 g
s
.39958 1.01051 m .38098 .98867 L .45147 1.00087 L p .75 .424 .432 r
F P
0 g
s
.45147 1.00087 m .46147 1.02238 L .39958 1.01051 L p .75 .424 .432 r
F P
0 g
s
.52707 .33822 m .52401 .31444 L .48235 .28995 L p .796 .372 .241 r
F P
0 g
s
.52401 .31444 m .52707 .33822 L .55567 .36124 L p .759 .393 .362 r
F P
0 g
s
.55567 .36124 m .56544 .33733 L .52401 .31444 L p .759 .393 .362 r
F P
0 g
s
.68778 .94371 m .68716 .90721 L .75078 .92274 L p .573 .466 .694 r
F P
0 g
s
.68716 .90721 m .68188 .87233 L .74325 .88825 L p .558 .494 .739 r
F P
0 g
s
.74325 .88825 m .75078 .92274 L .68716 .90721 L p .558 .494 .739 r
F P
0 g
s
.6837 .97949 m .68778 .94371 L .75166 .95884 L p .583 .439 .651 r
F P
0 g
s
.75078 .92274 m .75166 .95884 L .68778 .94371 L p .573 .466 .694 r
F P
0 g
s
.10338 .23963 m .14674 .19833 L .18333 .2381 L p .842 .906 .881 r
F P
0 g
s
.22976 .20839 m .18333 .2381 L .14674 .19833 L p .756 .881 .936 r
F P
0 g
s
.83037 .4271 m .7913 .40094 L .82056 .41697 L p 0 .308 .718 r
F P
0 g
s
.82056 .41697 m .86491 .44218 L .83037 .4271 L p 0 .308 .718 r
F P
0 g
s
.49478 .65438 m .47821 .68219 L .52331 .70102 L p .479 .617 .909 r
F P
0 g
s
.52331 .70102 m .52647 .67375 L .49478 .65438 L p .479 .617 .909 r
F P
0 g
s
.52647 .67375 m .52331 .70102 L .56816 .71927 L p .331 .554 .916 r
F P
0 g
s
.48235 .28995 m .49836 .31405 L .52707 .33822 L p .796 .372 .241 r
F P
0 g
s
.59578 .79962 m .58778 .78187 L .64271 .7989 L p .472 .637 .927 r
F P
0 g
s
.64271 .7989 m .65894 .8163 L .59578 .79962 L p .472 .637 .927 r
F P
0 g
s
.68188 .87233 m .67228 .84135 L .72956 .85764 L p .532 .524 .792 r
F P
0 g
s
.72956 .85764 m .74325 .88825 L .68188 .87233 L p .532 .524 .792 r
F P
0 g
s
.67517 1.01221 m .6837 .97949 L .74583 .99418 L p .592 .41 .604 r
F P
0 g
s
.75166 .95884 m .74583 .99418 L .6837 .97949 L p .583 .439 .651 r
F P
0 g
s
.46538 .63432 m .43622 .66259 L .47821 .68219 L p .571 .661 .898 r
F P
0 g
s
.47821 .68219 m .49478 .65438 L .46538 .63432 L p .571 .661 .898 r
F P
0 g
s
.56544 .33733 m .55567 .36124 L .58196 .3833 L p .712 .38 .415 r
F P
0 g
s
.58196 .3833 m .60358 .35891 L .56544 .33733 L p .712 .38 .415 r
F P
0 g
s
.59767 1.02634 m .58998 1.04717 L .52656 1.03458 L p .686 .406 .493 r
F P
0 g
s
.58998 1.04717 m .59767 1.02634 L .6627 1.03972 L p .651 .377 .491 r
F P
0 g
s
.67517 1.01221 m .6627 1.03972 L .59767 1.02634 L p .647 .425 .568 r
F P
0 g
s
.30326 .75749 m .32569 .72952 L .38533 .74789 L p .749 .728 .833 r
F P
0 g
s
.40465 .72881 m .38533 .74789 L .32569 .72952 L p .7 .749 .89 r
F P
0 g
s
.38533 .74789 m .40465 .72881 L .46415 .74698 L p .664 .714 .885 r
F P
0 g
s
.80202 .76797 m .8396 .74088 L .8754 .75762 L p .369 .002 .204 r
F P
0 g
s
.90963 .71984 m .8754 .75762 L .8396 .74088 L p .313 .025 .305 r
F P
0 g
s
.56816 .71927 m .55802 .69257 L .52647 .67375 L p .331 .554 .916 r
F P
0 g
s
.55802 .69257 m .56816 .71927 L .60946 .73711 L p .059 .425 .869 r
F P
0 g
s
.28146 .19229 m .33503 .19107 L .36805 .22334 L p .472 .721 .978 r
F P
0 g
s
.41084 .23531 m .36805 .22334 L .33503 .19107 L p .258 .578 .942 r
F P
0 g
s
.18495 .59312 m .13439 .56173 L .17246 .58829 L p 0 .208 .649 r
F P
0 g
s
.17246 .58829 m .21717 .61792 L .18495 .59312 L p 0 .208 .649 r
F P
0 g
s
.52796 .7556 m .52676 .76461 L .46415 .74698 L p .45 .704 .975 r
F P
0 g
s
.52676 .76461 m .52796 .7556 L .57887 .77313 L p .297 .657 .971 r
F P
0 g
s
.58778 .78187 m .52676 .76461 L p .56212 .77039 L .297 .657 .971 r
F P
0 g
s
.52921 .75649 m .56962 .77411 L p .53785 .75999 L .528 0 0 r
F P
0 g
s
.52796 .7556 m .52852 .756 L p .53785 .75999 L .528 0 0 r
F P
0 g
s
.52852 .756 m .52921 .75649 L p .53785 .75999 L .528 0 0 r
F P
0 g
s
.57887 .77313 m .58689 .781 L p .56212 .77039 L .297 .657 .971 r
F P
0 g
s
.58689 .781 m .58778 .78187 L p .56212 .77039 L .297 .657 .971 r
F P
0 g
s
.34577 .99884 m .3195 .97667 L .38098 .98867 L p .779 .398 .335 r
F P
0 g
s
.38098 .98867 m .39958 1.01051 L .34577 .99884 L p .779 .398 .335 r
F P
0 g
s
.47574 .7377 m .46415 .74698 L .40465 .72881 L p .538 .729 .963 r
F P
0 g
s
.46415 .74698 m .47574 .7377 L .52796 .7556 L p .45 .704 .975 r
F P
0 g
s
.67228 .84135 m .65894 .8163 L .71057 .83293 L p .474 .56 .863 r
F P
0 g
s
.71057 .83293 m .72956 .85764 L .67228 .84135 L p .474 .56 .863 r
F P
0 g
s
.21997 .8562 m .22124 .81452 L .27824 .83068 L p .864 .705 .684 r
F P
0 g
s
.27824 .83068 m .27722 .87109 L .21997 .8562 L p .864 .705 .684 r
F P
0 g
s
.28711 .79197 m .27824 .83068 L .22124 .81452 L p .85 .739 .741 r
F P
0 g
s
.22851 .89689 m .21997 .8562 L .27722 .87109 L p .871 .661 .617 r
F P
0 g
s
.27722 .87109 m .28411 .91058 L .22851 .89689 L p .871 .661 .617 r
F P
0 g
s
.47286 1.03494 m .46147 1.02238 L .52656 1.03458 L p .69 .278 .261 r
F P
0 g
s
.52656 1.03458 m .52774 1.04699 L .47286 1.03494 L p .69 .278 .261 r
F P
0 g
s
.52774 1.04699 m .52656 1.03458 L .58998 1.04717 L p .675 .299 .326 r
F P
0 g
s
.6627 1.03972 m .67517 1.01221 L .73365 1.02645 L p .6 .376 .548 r
F P
0 g
s
.74583 .99418 m .73365 1.02645 L .67517 1.01221 L p .592 .41 .604 r
F P
0 g
s
.22124 .81452 m .23223 .77459 L .28711 .79197 L p .85 .739 .741 r
F P
0 g
s
.30326 .75749 m .28711 .79197 L .23223 .77459 L p .827 .767 .796 r
F P
0 g
s
.42076 1.02314 m .39958 1.01051 L .46147 1.02238 L p .683 .213 .13 r
F P
0 g
s
.46147 1.02238 m .47286 1.03494 L .42076 1.02314 L p .683 .213 .13 r
F P
0 g
s
.67073 .36762 m .63559 .37953 L .65906 .39954 L p .584 .182 .246 r
F P
0 g
s
.65906 .39954 m .70268 .38648 L .67073 .36762 L p .584 .182 .246 r
F P
0 g
s
.14674 .19833 m .19916 .16953 L .22976 .20839 L p .756 .881 .936 r
F P
0 g
s
.28146 .19229 m .22976 .20839 L .19916 .16953 L p .653 .829 .968 r
F P
0 g
s
.37448 .62052 m .32169 .63482 L .3571 .65643 L p .599 .769 .961 r
F P
0 g
s
.3571 .65643 m .40064 .64207 L .37448 .62052 L p .599 .769 .961 r
F P
0 g
s
.24632 .93388 m .22851 .89689 L .28411 .91058 L p .87 .601 .528 r
F P
0 g
s
.28411 .91058 m .29851 .94657 L .24632 .93388 L p .87 .601 .528 r
F P
0 g
s
.7913 .40094 m .74779 .38703 L .77144 .40428 L p .42 0 0 r
F P
0 g
s
.77144 .40428 m .82056 .41697 L .7913 .40094 L p .42 0 0 r
F P
0 g
s
.58778 .78187 m .57887 .77313 L .62464 .79039 L p .026 .532 .871 r
F P
0 g
s
.64271 .7989 m .58778 .78187 L p .61277 .78765 L .026 .532 .871 r
F P
0 g
s
.56962 .77411 m .60592 .79147 L p .5853 .77748 L .629 .121 .022 r
F P
0 g
s
.57887 .77313 m .57471 .77357 L p .5853 .77748 L .629 .121 .022 r
F P
0 g
s
.57471 .77357 m .56962 .77411 L p .5853 .77748 L .629 .121 .022 r
F P
0 g
s
.62464 .79039 m .64095 .79807 L p .61277 .78765 L .026 .532 .871 r
F P
0 g
s
.64095 .79807 m .64271 .7989 L p .61277 .78765 L .026 .532 .871 r
F P
0 g
s
.44064 .61346 m .40064 .64207 L .43622 .66259 L p .637 .702 .893 r
F P
0 g
s
.43622 .66259 m .46538 .63432 L .44064 .61346 L p .637 .702 .893 r
F P
0 g
s
.60358 .35891 m .58196 .3833 L .60395 .40464 L p .661 .348 .434 r
F P
0 g
s
.60395 .40464 m .63559 .37953 L .60358 .35891 L p .661 .348 .434 r
F P
0 g
s
.23223 .77459 m .25222 .73905 L .30326 .75749 L p .827 .767 .796 r
F P
0 g
s
.32569 .72952 m .30326 .75749 L .25222 .73905 L p .791 .788 .85 r
F P
0 g
s
.32569 .72952 m .35295 .70995 L .40465 .72881 L p .7 .749 .89 r
F P
0 g
s
.42619 .71927 m .40465 .72881 L .35295 .70995 L p .593 .753 .954 r
F P
0 g
s
.40465 .72881 m .42619 .71927 L .47574 .7377 L p .538 .729 .963 r
F P
0 g
s
.71409 .78436 m .66964 .77217 L .64416 .7547 L p .618 .068 0 r
F P
0 g
s
.64416 .7547 m .66964 .77217 L p .65882 .76417 L .715 .18 0 r
F P
0 g
s
.67991 .76715 m .64416 .7547 L p .63146 .74649 L .47 0 0 r
F P
0 g
s
.64416 .7547 m .67991 .76715 L .71409 .78436 L p .618 .068 0 r
F P
0 g
s
.61131 .72904 m .64416 .7547 L p .65882 .76417 L .715 .18 0 r
F P
0 g
s
.64416 .7547 m .60946 .73711 L p .63146 .74649 L .47 0 0 r
F P
0 g
s
.75924 .78286 m .71409 .78436 L .67991 .76715 L p .558 .053 .012 r
F P
0 g
s
.65894 .8163 m .64271 .7989 L .68751 .81582 L p .293 .581 .94 r
F P
0 g
s
.68751 .81582 m .71057 .83293 L .65894 .8163 L p .293 .581 .94 r
F P
0 g
s
.58998 1.04717 m .58123 1.05936 L .52774 1.04699 L p .675 .299 .326 r
F P
0 g
s
.58123 1.05936 m .58998 1.04717 L .64709 1.06015 L p .652 .295 .353 r
F P
0 g
s
.6627 1.03972 m .64709 1.06015 L .58998 1.04717 L p .651 .377 .491 r
F P
0 g
s
.75924 .78286 m .80202 .76797 L .83268 .785 L p .444 .008 .115 r
F P
0 g
s
.8754 .75762 m .83268 .785 L .80202 .76797 L p .369 .002 .204 r
F P
0 g
s
.27225 .96468 m .24632 .93388 L .29851 .94657 L p .853 .503 .389 r
F P
0 g
s
.29851 .94657 m .3195 .97667 L .27225 .96468 L p .853 .503 .389 r
F P
0 g
s
.60946 .73711 m .58703 .71095 L .55802 .69257 L p .059 .425 .869 r
F P
0 g
s
.58703 .71095 m .60946 .73711 L .64416 .7547 L p 0 .142 .602 r
F P
0 g
s
.24208 .61026 m .18495 .59312 L .21717 .61792 L p .128 .574 .923 r
F P
0 g
s
.21717 .61792 m .26803 .63385 L .24208 .61026 L p .128 .574 .923 r
F P
0 g
s
.64709 1.06015 m .6627 1.03972 L .71589 1.05349 L p .61 .333 .472 r
F P
0 g
s
.73365 1.02645 m .71589 1.05349 L .6627 1.03972 L p .6 .376 .548 r
F P
0 g
s
.37561 1.01149 m .34577 .99884 L .39958 1.01051 L p .612 .061 0 r
F P
0 g
s
.39958 1.01051 m .42076 1.02314 L .37561 1.01149 L p .612 .061 0 r
F P
0 g
s
.52921 .75649 m .52796 .7556 L .47574 .7377 L p .272 0 0 r
F P
0 g
s
.56962 .77411 m .57887 .77313 L .52796 .7556 L p .528 0 0 r
F P
0 g
s
.48777 .7385 m .47574 .7377 L .42619 .71927 L p .074 .561 .897 r
F P
0 g
s
.47574 .7377 m .48777 .7385 L .52921 .75649 L p .272 0 0 r
F P
0 g
s
.25222 .73905 m .27991 .71029 L .32569 .72952 L p .791 .788 .85 r
F P
0 g
s
.35295 .70995 m .32569 .72952 L .27991 .71029 L p .732 .797 .904 r
F P
0 g
s
.19916 .16953 m .25712 .1552 L .28146 .19229 L p .653 .829 .968 r
F P
0 g
s
.33503 .19107 m .28146 .19229 L .25712 .1552 L p .538 .748 .973 r
F P
0 g
s
.75166 .95884 m .75078 .92274 L .79583 .93848 L p .474 .388 .678 r
F P
0 g
s
.75078 .92274 m .74325 .88825 L .78665 .90432 L p .444 .418 .735 r
F P
0 g
s
.78665 .90432 m .79583 .93848 L .75078 .92274 L p .444 .418 .735 r
F P
0 g
s
.74583 .99418 m .75166 .95884 L .7969 .97422 L p .496 .358 .623 r
F P
0 g
s
.79583 .93848 m .7969 .97422 L .75166 .95884 L p .474 .388 .678 r
F P
0 g
s
.74779 .38703 m .70268 .38648 L .72082 .40504 L p .502 0 0 r
F P
0 g
s
.72082 .40504 m .77144 .40428 L .74779 .38703 L p .502 0 0 r
F P
0 g
s
.53022 1.04217 m .49675 1.02979 L p .52282 1.04584 L .348 .656 .97 r
F P
0 g
s
.47286 1.03494 m .52774 1.04699 L p .49732 1.04027 L .388 0 0 r
F P
0 g
s
.48489 1.03752 m .47677 1.03578 L p .49732 1.04027 L .388 0 0 r
F P
0 g
s
.47677 1.03578 m .47286 1.03494 L p .49732 1.04027 L .388 0 0 r
F P
0 g
s
.52774 1.04699 m .52899 1.04964 L .48489 1.03752 L p .388 0 0 r
F P
0 g
s
.52899 1.04964 m .52774 1.04699 L .58123 1.05936 L p .548 0 0 r
F P
0 g
s
.52899 1.04964 m .52938 1.04728 L p .52282 1.04584 L .348 .656 .97 r
F P
0 g
s
.52938 1.04728 m .53022 1.04217 L p .52282 1.04584 L .348 .656 .97 r
F P
0 g
s
.44311 1.02555 m .42076 1.02314 L .47286 1.03494 L p 0 .458 .751 r
F P
0 g
s
.47286 1.03494 m .48489 1.03752 L .44311 1.02555 L p 0 .458 .751 r
F P
0 g
s
.74325 .88825 m .72956 .85764 L .76997 .87401 L p .388 .449 .802 r
F P
0 g
s
.76997 .87401 m .78665 .90432 L .74325 .88825 L p .388 .449 .802 r
F P
0 g
s
.30463 .9872 m .27225 .96468 L .3195 .97667 L p .772 .297 .108 r
F P
0 g
s
.3195 .97667 m .34577 .99884 L .30463 .9872 L p .772 .297 .108 r
F P
0 g
s
.66189 .8075 m .68751 .81582 L .64271 .7989 L p .333 0 0 r
F P
0 g
s
.64271 .7989 m .62464 .79039 L .66189 .8075 L p .333 0 0 r
F P
0 g
s
.33503 .19107 m .38683 .20503 L .41084 .23531 L p .258 .578 .942 r
F P
0 g
s
.4494 .26169 m .41084 .23531 L .38683 .20503 L p .002 .362 .833 r
F P
0 g
s
.5296 .37466 m .52707 .33822 L .49836 .31405 L p .837 .517 .447 r
F P
0 g
s
.52707 .33822 m .5296 .37466 L .54766 .39755 L p .778 .497 .51 r
F P
0 g
s
.54766 .39755 m .55567 .36124 L .52707 .33822 L p .778 .497 .51 r
F P
0 g
s
.60592 .79147 m .62464 .79039 L .57887 .77313 L p .629 .121 .022 r
F P
0 g
s
.73365 1.02645 m .74583 .99418 L .78978 1.00917 L p .513 .328 .567 r
F P
0 g
s
.7969 .97422 m .78978 1.00917 L .74583 .99418 L p .496 .358 .623 r
F P
0 g
s
.49836 .31405 m .5115 .35097 L .5296 .37466 L p .837 .517 .447 r
F P
0 g
s
.50873 .61435 m .49478 .65438 L .52647 .67375 L p .519 .561 .839 r
F P
0 g
s
.52647 .67375 m .52913 .6344 L .50873 .61435 L p .519 .561 .839 r
F P
0 g
s
.52913 .6344 m .52647 .67375 L .55802 .69257 L p .385 .491 .844 r
F P
0 g
s
.35295 .70995 m .3833 .70019 L .42619 .71927 L p .593 .753 .954 r
F P
0 g
s
.44854 .72004 m .42619 .71927 L .3833 .70019 L p .32 .67 .975 r
F P
0 g
s
.42619 .71927 m .44854 .72004 L .48777 .7385 L p .074 .561 .897 r
F P
0 g
s
.64709 1.06015 m .62934 1.07204 L .58123 1.05936 L p .652 .295 .353 r
F P
0 g
s
.62934 1.07204 m .64709 1.06015 L .69369 1.07347 L p .623 .272 .356 r
F P
0 g
s
.71589 1.05349 m .69369 1.07347 L .64709 1.06015 L p .61 .333 .472 r
F P
0 g
s
.72956 .85764 m .71057 .83293 L .74688 .84958 L p .258 .472 .876 r
F P
0 g
s
.74688 .84958 m .76997 .87401 L .72956 .85764 L p .258 .472 .876 r
F P
0 g
s
.55567 .36124 m .54766 .39755 L .56427 .41976 L p .723 .465 .536 r
F P
0 g
s
.56427 .41976 m .58196 .3833 L .55567 .36124 L p .723 .465 .536 r
F P
0 g
s
.48989 .5937 m .46538 .63432 L .49478 .65438 L p .603 .611 .832 r
F P
0 g
s
.49478 .65438 m .50873 .61435 L .48989 .5937 L p .603 .611 .832 r
F P
0 g
s
.58123 1.05936 m .57199 1.06196 L .52899 1.04964 L p .548 0 0 r
F P
0 g
s
.57199 1.06196 m .58123 1.05936 L .62934 1.07204 L p .613 .114 .042 r
F P
0 g
s
.3019 .6119 m .24208 .61026 L .26803 .63385 L p .458 .757 .994 r
F P
0 g
s
.26803 .63385 m .32169 .63482 L .3019 .6119 L p .458 .757 .994 r
F P
0 g
s
.71409 .78436 m .75924 .78286 L .78431 .80004 L p .543 .062 .064 r
F P
0 g
s
.83268 .785 m .78431 .80004 L .75924 .78286 L p .444 .008 .115 r
F P
0 g
s
.63559 .37953 m .60395 .40464 L .61996 .4255 L p .602 .297 .426 r
F P
0 g
s
.61996 .4255 m .65906 .39954 L .63559 .37953 L p .602 .297 .426 r
F P
0 g
s
.27991 .71029 m .3135 .69025 L .35295 .70995 L p .732 .797 .904 r
F P
0 g
s
.3833 .70019 m .35295 .70995 L .3135 .69025 L p .631 .783 .954 r
F P
0 g
s
.4227 .59175 m .37448 .62052 L .40064 .64207 L p .69 .746 .894 r
F P
0 g
s
.40064 .64207 m .44064 .61346 L .4227 .59175 L p .69 .746 .894 r
F P
0 g
s
.55802 .69257 m .54948 .65392 L .52913 .6344 L p .385 .491 .844 r
F P
0 g
s
.54948 .65392 m .55802 .69257 L .58703 .71095 L p .123 .351 .815 r
F P
0 g
s
.71589 1.05349 m .73365 1.02645 L .77493 1.04099 L p .531 .296 .503 r
F P
0 g
s
.78978 1.00917 m .77493 1.04099 L .73365 1.02645 L p .513 .328 .567 r
F P
0 g
s
.40704 1.01365 m .37561 1.01149 L .42076 1.02314 L p .156 .629 .927 r
F P
0 g
s
.42076 1.02314 m .44311 1.02555 L .40704 1.01365 L p .156 .629 .927 r
F P
0 g
s
.64416 .7547 m .61131 .72904 L p .60484 .72459 L 0 .142 .602 r
F P
0 g
s
.62901 .74694 m .61131 .72904 L p .62985 .74275 L .715 .18 0 r
F P
0 g
s
.66964 .77217 m .62901 .74694 L p .62985 .74275 L .715 .18 0 r
F P
0 g
s
.62901 .74694 m .63879 .76472 L p .62812 .75141 L .872 .498 .334 r
F P
0 g
s
.61131 .72904 m .62901 .74694 L p .61827 .73384 L .842 .379 .008 r
F P
0 g
s
.58379 .69178 m .61131 .72904 L p .61827 .73384 L .842 .379 .008 r
F P
0 g
s
.61131 .72904 m .58703 .71095 L p .60484 .72459 L 0 .142 .602 r
F P
0 g
s
.56818 .67302 m .58703 .71095 L .61131 .72904 L p 0 .012 .562 r
F P
0 g
s
.58379 .69178 m .56818 .67302 L p .57953 .68777 L 0 .012 .562 r
F P
0 g
s
.61131 .72904 m .58379 .69178 L p .57953 .68777 L 0 .012 .562 r
F P
0 g
s
.59514 .71029 m .58379 .69178 L p .59596 .70662 L .842 .379 .008 r
F P
0 g
s
.62901 .74694 m .59514 .71029 L p .59596 .70662 L .842 .379 .008 r
F P
0 g
s
.59514 .71029 m .60134 .7286 L p .5943 .71401 L .912 .614 .466 r
F P
0 g
s
.58379 .69178 m .59514 .71029 L p .58816 .69588 L .938 .569 .199 r
F P
0 g
s
.56371 .64543 m .58379 .69178 L p .58816 .69588 L .938 .569 .199 r
F P
0 g
s
.55437 .62587 m .56818 .67302 L .58379 .69178 L p 0 0 .412 r
F P
0 g
s
.56371 .64543 m .55437 .62587 L p .55998 .63843 L 0 0 .412 r
F P
0 g
s
.58379 .69178 m .56371 .64543 L p .55998 .63843 L 0 0 .412 r
F P
0 g
s
.57049 .66471 m .56371 .64543 L p .56984 .65809 L .938 .569 .199 r
F P
0 g
s
.59514 .71029 m .57049 .66471 L p .56984 .65809 L .938 .569 .199 r
F P
0 g
s
.57049 .66471 m .57421 .68374 L p .57113 .67122 L .963 .699 .486 r
F P
0 g
s
.56371 .64543 m .57049 .66471 L p .56741 .65238 L .902 .605 .106 r
F P
0 g
s
.56741 .65238 m .55269 .59363 L .56371 .64543 L .902 .605 .106 r
F
.55262 .59334 m .55267 .59356 L p .55269 .59363 L F P
0 g
s
.55267 .59356 m .56371 .64543 L p .55269 .59363 L .902 .605 .106 r
F P
0 g
s
.55545 .56014 m .55141 .53933 L p .55616 .60314 L .336 .277 .642 r
F P
0 g
s
.55693 .61344 m .55262 .59334 L p .55498 .60275 L .902 .605 .106 r
F P
0 g
s
.57049 .66471 m .55698 .61365 L p .55511 .60327 L .902 .605 .106 r
F P
0 g
s
.57113 .67122 m .557 .61372 L .57049 .66471 L .963 .699 .486 r
F
.55693 .61344 m .55698 .61365 L p .557 .61372 L F P
0 g
s
.55698 .61365 m .57049 .66471 L p .557 .61372 L .963 .699 .486 r
F P
0 g
s
.56371 .64543 m .55262 .59334 L .54673 .57293 L p 0 0 .51 r
F P
0 g
s
.54673 .57293 m .55437 .62587 L .56371 .64543 L p 0 0 .51 r
F P
0 g
s
.55932 .63326 m .55693 .61344 L p .55927 .62298 L .963 .699 .486 r
F P
0 g
s
.57421 .68374 m .55932 .63326 L p .55927 .62298 L .963 .699 .486 r
F P
0 g
s
.5432 .60595 m .54948 .65392 L .56818 .67302 L p .11 .247 .726 r
F P
0 g
s
.56818 .67302 m .55437 .62587 L .5432 .60595 L p .11 .247 .726 r
F P
0 g
s
.55932 .63326 m .55961 .65284 L p .55936 .64316 L .83 .659 .114 r
F P
0 g
s
.55932 .63326 m .55907 .62519 L p .55881 .62379 L 0 0 .487 r
F P
0 g
s
.55907 .62519 m .55787 .58643 L p .55881 .62379 L 0 0 .487 r
F P
0 g
s
.55787 .58643 m .5577 .58068 L p .55784 .58631 L 0 0 .487 r
F P
0 g
s
.55787 .58643 m .55784 .58631 L .55881 .62379 L 0 0 .487 r
F
.56981 .55042 m .56945 .52966 L p .55941 .59211 L .292 .173 .54 r
F P
0 g
s
.55693 .61344 m .55932 .63326 L p .55881 .62379 L 0 0 .487 r
F P
0 g
s
.5577 .58068 m .55545 .56014 L p .5559 .56869 L 0 0 .487 r
F P
0 g
s
.55878 .62314 m .5577 .58068 L .5559 .56869 L 0 0 .487 r
F
.55936 .64316 m .55826 .60208 L .55932 .63326 L .83 .659 .114 r
F
.55826 .60208 m .55793 .58939 L .55828 .6007 L F
.55826 .60208 m .55828 .6007 L .55932 .63326 L F
.55793 .58939 m .55792 .58908 L .55932 .63326 L F
.55787 .58643 m .55932 .63326 L p .55789 .58797 L F P
0 g
s
.5588 .62325 m .5577 .58068 L .55878 .62314 L 0 0 .487 r
F
.55698 .61365 m .55693 .61344 L p .55511 .60327 L .902 .605 .106 r
F P
0 g
s
.55545 .56014 m .55693 .61344 L p .55881 .62379 L 0 0 .487 r
F P
0 g
s
.55881 .62379 m .5588 .62325 L .55878 .62314 L 0 0 .487 r
F
.55961 .65284 m .55798 .60096 L .5577 .58068 L p .83 .659 .114 r
F P
0 g
s
.58703 .71095 m .56818 .67302 L .54948 .65392 L p .123 .351 .815 r
F P
0 g
s
.55283 .64087 m .55631 .62102 L p .55602 .63016 L .893 .716 .655 r
F P
0 g
s
.55409 .69133 m .55283 .64087 L p .55602 .63016 L .893 .716 .655 r
F P
0 g
s
.53108 .58563 m .52913 .6344 L .54948 .65392 L p .388 .422 .775 r
F P
0 g
s
.54948 .65392 m .5432 .60595 L .53108 .58563 L p .388 .422 .775 r
F P
0 g
s
.55283 .64087 m .54785 .6605 L p .54975 .6512 L .952 .832 .688 r
F P
0 g
s
.55631 .62102 m .55283 .64087 L p .55482 .63122 L .939 .867 .444 r
F P
0 g
s
.55961 .65284 m .55782 .6722 L p .55793 .62228 L .985 .788 .557 r
F P
0 g
s
.55793 .62228 m .55796 .6099 L .55839 .62097 L .985 .788 .557 r
F
.55793 .62228 m .55839 .62097 L .55961 .65284 L F
.55796 .6099 m .55802 .60953 L .55961 .65284 L F
.55816 .6066 m .55961 .65284 L p .55797 .60813 L F P
0 g
s
.55782 .6722 m .55631 .62102 L .55798 .60096 L p .985 .788 .557 r
F P
0 g
s
.55782 .6722 m .55409 .69133 L p .55602 .63016 L .893 .716 .655 r
F P
0 g
s
.55631 .62102 m .55653 .62847 L p .55602 .63016 L .893 .716 .655 r
F P
0 g
s
.55653 .62847 m .55782 .6722 L p .55602 .63016 L .893 .716 .655 r
F P
0 g
s
.54975 .6512 m .54984 .65078 L .55283 .64087 L .952 .832 .688 r
F
.55828 .61199 m .55635 .62084 L p .55482 .63122 L .939 .867 .444 r
F P
0 g
s
.55635 .62084 m .55631 .62102 L p .55482 .63122 L .939 .867 .444 r
F P
0 g
s
.55636 .62077 m .5581 .61282 L .55812 .61154 L 0 0 .261 r
F
.55631 .62102 m .55635 .62084 L p .55636 .62077 L F P
0 g
s
.55635 .62084 m .5581 .61282 L p .55636 .62077 L 0 0 .261 r
F P
0 g
s
.56724 .57101 m .56981 .55042 L p .56772 .56132 L 0 0 .261 r
F P
0 g
s
.55812 .61154 m .56724 .57101 L .56772 .56132 L 0 0 .261 r
F
.5581 .61282 m .56724 .57101 L p .55812 .61154 L F P
0 g
s
.56199 .59144 m .56724 .57101 L p .56492 .58226 L .939 .867 .444 r
F P
0 g
s
.55482 .63122 m .56199 .59144 L .56492 .58226 L .939 .867 .444 r
F
.55798 .60096 m .55816 .6066 L p .55797 .60813 L .985 .788 .557 r
F P
0 g
s
.55796 .6099 m .55797 .60813 L .55802 .60953 L .985 .788 .557 r
F
.55798 .60096 m .55631 .62102 L p .55812 .61154 L 0 0 .261 r
F P
0 g
s
.56981 .55042 m .55803 .60078 L p .55812 .61154 L 0 0 .261 r
F P
0 g
s
.55803 .60078 m .55798 .60096 L p .55812 .61154 L 0 0 .261 r
F P
0 g
s
.55511 .60327 m .55693 .61344 L .55498 .60275 L .902 .605 .106 r
F
.55693 .61344 m .55668 .60457 L p .55616 .60314 L .336 .277 .642 r
F P
0 g
s
.55141 .53933 m .55262 .59334 L .55693 .61344 L p .336 .277 .642 r
F P
0 g
s
.53972 .55218 m .5432 .60595 L .55437 .62587 L p .257 .28 .694 r
F P
0 g
s
.55437 .62587 m .54673 .57293 L .53972 .55218 L p .257 .28 .694 r
F P
0 g
s
.54785 .6605 m .55453 .61173 L .56199 .59144 L p .952 .832 .688 r
F P
0 g
s
.56199 .59144 m .55453 .61173 L p .55921 .60014 L .959 .931 .721 r
F P
0 g
s
.56724 .57101 m .56199 .59144 L p .56678 .5801 L .67 .772 .294 r
F P
0 g
s
.5899 .52541 m .56724 .57101 L p .56678 .5801 L .67 .772 .294 r
F P
0 g
s
.56199 .59144 m .55286 .64069 L p .54984 .65078 L .952 .832 .688 r
F P
0 g
s
.55439 .63247 m .56199 .59144 L p .55482 .63122 L .939 .867 .444 r
F P
0 g
s
.58101 .5462 m .5899 .52541 L p .58415 .539 L .67 .772 .294 r
F P
0 g
s
.56199 .59144 m .58101 .5462 L p .58415 .539 L .67 .772 .294 r
F P
0 g
s
.59428 .50461 m .56981 .55042 L .56724 .57101 L p .068 0 .311 r
F P
0 g
s
.56724 .57101 m .5899 .52541 L .59428 .50461 L p .068 0 .311 r
F P
0 g
s
.58101 .5462 m .56199 .59144 L p .55921 .60014 L .959 .931 .721 r
F P
0 g
s
.55793 .58939 m .55789 .58797 L .55792 .58908 L .83 .659 .114 r
F
.5577 .58068 m .55787 .58643 L p .55789 .58797 L F P
0 g
s
.5577 .58068 m .55798 .60096 L p .55941 .59211 L .292 .173 .54 r
F P
0 g
s
.55803 .6007 m .55976 .59339 L .55941 .59211 L .292 .173 .54 r
F
.55798 .60096 m .55803 .60078 L p .55803 .6007 L F P
0 g
s
.55803 .60078 m .55976 .59339 L p .55803 .6007 L .292 .173 .54 r
F P
0 g
s
.55453 .61173 m .56842 .56697 L .58101 .5462 L p .959 .931 .721 r
F P
0 g
s
.58101 .5462 m .56842 .56697 L p .57811 .55281 L .898 .986 .78 r
F P
0 g
s
.60875 .50805 m .58101 .5462 L p .57811 .55281 L .898 .986 .78 r
F P
0 g
s
.55286 .64069 m .55283 .64087 L p .54984 .65078 L .952 .832 .688 r
F P
0 g
s
.55283 .64087 m .55439 .63247 L p .55482 .63122 L .939 .867 .444 r
F P
0 g
s
.56945 .52966 m .5577 .58068 L p .55941 .59211 L .292 .173 .54 r
F P
0 g
s
.55976 .59339 m .56981 .55042 L p .55941 .59211 L .292 .173 .54 r
F P
0 g
s
.55668 .60457 m .55545 .56014 L p .55616 .60314 L .336 .277 .642 r
F P
0 g
s
.57049 .66471 m .59514 .71029 L p .5943 .71401 L .912 .614 .466 r
F P
0 g
s
.59514 .71029 m .62901 .74694 L p .62812 .75141 L .872 .498 .334 r
F P
0 g
s
.71057 .83293 m .68751 .81582 L .71889 .83269 L p 0 .383 .803 r
F P
0 g
s
.71889 .83269 m .74688 .84958 L .71057 .83293 L p 0 .383 .803 r
F P
0 g
s
.34131 .99985 m .30463 .9872 L .34577 .99884 L p .378 0 0 r
F P
0 g
s
.34577 .99884 m .37561 1.01149 L .34131 .99985 L p .378 0 0 r
F P
0 g
s
.25712 .1552 m .31673 .15633 L .33503 .19107 L p .538 .748 .973 r
F P
0 g
s
.38683 .20503 m .33503 .19107 L .31673 .15633 L p .417 .64 .948 r
F P
0 g
s
.53042 .76739 m .52921 .75649 L .48777 .7385 L p .736 .285 .179 r
F P
0 g
s
.52921 .75649 m .53042 .76739 L .56067 .78492 L p .731 .349 .327 r
F P
0 g
s
.56067 .78492 m .56962 .77411 L .52921 .75649 L p .731 .349 .327 r
F P
0 g
s
.47416 .57239 m .44064 .61346 L .46538 .63432 L p .667 .656 .83 r
F P
0 g
s
.46538 .63432 m .48989 .5937 L .47416 .57239 L p .667 .656 .83 r
F P
0 g
s
.49943 .74953 m .48777 .7385 L .44854 .72004 L p .652 .099 0 r
F P
0 g
s
.48777 .7385 m .49943 .74953 L .53042 .76739 L p .736 .285 .179 r
F P
0 g
s
.62464 .79039 m .60592 .79147 L .63541 .80864 L p .655 .207 .175 r
F P
0 g
s
.63541 .80864 m .66189 .8075 L .62464 .79039 L p .655 .207 .175 r
F P
0 g
s
.58196 .3833 m .56427 .41976 L .57814 .44144 L p .669 .424 .542 r
F P
0 g
s
.57814 .44144 m .60395 .40464 L .58196 .3833 L p .669 .424 .542 r
F P
0 g
s
.70268 .38648 m .65906 .39954 L .67216 .41928 L p .521 .122 .226 r
F P
0 g
s
.67216 .41928 m .72082 .40504 L .70268 .38648 L p .521 .122 .226 r
F P
0 g
s
.69369 1.07347 m .71589 1.05349 L .75332 1.06754 L p .554 .263 .427 r
F P
0 g
s
.77493 1.04099 m .75332 1.06754 L .71589 1.05349 L p .531 .296 .503 r
F P
0 g
s
.18498 .84073 m .18642 .79761 L .22124 .81452 L p .958 .786 .627 r
F P
0 g
s
.22124 .81452 m .21997 .8562 L .18498 .84073 L p .958 .786 .627 r
F P
0 g
s
.23223 .77459 m .22124 .81452 L .18642 .79761 L p .934 .836 .724 r
F P
0 g
s
.49675 1.02979 m .48489 1.03752 L .52899 1.04964 L p .348 .656 .97 r
F P
0 g
s
.53022 1.04217 m .52899 1.04964 L .57199 1.06196 L p .084 .544 .905 r
F P
0 g
s
.46511 1.01743 m .44311 1.02555 L .48489 1.03752 L p .487 .7 .963 r
F P
0 g
s
.48489 1.03752 m .49675 1.02979 L .46511 1.01743 L p .487 .7 .963 r
F P
0 g
s
.19462 .88277 m .18498 .84073 L .21997 .8562 L p .963 .71 .504 r
F P
0 g
s
.21997 .8562 m .22851 .89689 L .19462 .88277 L p .963 .71 .504 r
F P
0 g
s
.18642 .79761 m .19883 .75633 L .23223 .77459 L p .934 .836 .724 r
F P
0 g
s
.25222 .73905 m .23223 .77459 L .19883 .75633 L p .893 .866 .805 r
F P
0 g
s
.62934 1.07204 m .61062 1.07446 L .57199 1.06196 L p .613 .114 .042 r
F P
0 g
s
.61062 1.07446 m .62934 1.07204 L .6685 1.08496 L p .632 .179 .158 r
F P
0 g
s
.69369 1.07347 m .6685 1.08496 L .62934 1.07204 L p .623 .272 .356 r
F P
0 g
s
.56962 .77411 m .56067 .78492 L .58784 .80219 L p .708 .366 .396 r
F P
0 g
s
.58784 .80219 m .60592 .79147 L .56962 .77411 L p .708 .366 .396 r
F P
0 g
s
.3135 .69025 m .35077 .68033 L .3833 .70019 L p .631 .783 .954 r
F P
0 g
s
.4147 .701 m .3833 .70019 L .35077 .68033 L p .458 .715 .979 r
F P
0 g
s
.3833 .70019 m .4147 .701 L .44854 .72004 L p .32 .67 .975 r
F P
0 g
s
.47017 .73125 m .44854 .72004 L .4147 .701 L p 0 .279 .681 r
F P
0 g
s
.44503 .71248 m .47017 .73125 L p .45366 .72225 L 0 .279 .681 r
F P
0 g
s
.4147 .701 m .44503 .71248 L p .45366 .72225 L 0 .279 .681 r
F P
0 g
s
.44503 .71248 m .42625 .69314 L p .43422 .70022 L 0 .261 .754 r
F P
0 g
s
.47017 .73125 m .44503 .71248 L p .45323 .7198 L .751 .234 0 r
F P
0 g
s
.44854 .72004 m .47017 .73125 L p .48385 .7405 L .652 .099 0 r
F P
0 g
s
.48956 .75227 m .47017 .73125 L p .45323 .7198 L .751 .234 0 r
F P
0 g
s
.47218 .73395 m .44503 .71248 L p .43422 .70022 L 0 .261 .754 r
F P
0 g
s
.47017 .73125 m .49943 .74953 L p .48385 .7405 L .652 .099 0 r
F P
0 g
s
.3603 .5979 m .3019 .6119 L .32169 .63482 L p .643 .81 .963 r
F P
0 g
s
.32169 .63482 m .37448 .62052 L .3603 .5979 L p .643 .81 .963 r
F P
0 g
s
.2147 .92089 m .19462 .88277 L .22851 .89689 L p .939 .589 .331 r
F P
0 g
s
.22851 .89689 m .24632 .93388 L .2147 .92089 L p .939 .589 .331 r
F P
0 g
s
.68751 .81582 m .66189 .8075 L .68787 .82452 L p .573 .011 0 r
F P
0 g
s
.68787 .82452 m .71889 .83269 L .68751 .81582 L p .573 .011 0 r
F P
0 g
s
.19883 .75633 m .22136 .71966 L .25222 .73905 L p .893 .866 .805 r
F P
0 g
s
.27991 .71029 m .25222 .73905 L .22136 .71966 L p .835 .876 .873 r
F P
0 g
s
.66964 .77217 m .71409 .78436 L .73357 .80165 L p .664 .175 .08 r
F P
0 g
s
.78431 .80004 m .73357 .80165 L .71409 .78436 L p .543 .062 .064 r
F P
0 g
s
.57199 1.06196 m .56288 1.05462 L .53022 1.04217 L p .084 .544 .905 r
F P
0 g
s
.56288 1.05462 m .57199 1.06196 L .61062 1.07446 L p .317 0 0 r
F P
0 g
s
.38683 .20503 m .43324 .23342 L .4494 .26169 L p .002 .362 .833 r
F P
0 g
s
.48088 .30088 m .47093 .27443 L p .47732 .28962 L 0 0 .427 r
F P
0 g
s
.48088 .30088 m .4948 .3264 L p .48413 .31118 L .797 .335 0 r
F P
0 g
s
.51093 .37458 m .4948 .3264 L .48088 .30088 L p .939 .577 .186 r
F P
0 g
s
.51093 .37458 m .52071 .39826 L p .5139 .37937 L .919 .623 .465 r
F P
0 g
s
.4948 .3264 m .50637 .36097 L p .5139 .37937 L .919 .623 .465 r
F P
0 g
s
.50637 .36097 m .51093 .37458 L p .5139 .37937 L .919 .623 .465 r
F P
0 g
s
.50283 .35026 m .49227 .3265 L p .47732 .28962 L 0 0 .427 r
F P
0 g
s
.49227 .3265 m .48088 .30088 L p .47781 .29115 L 0 0 .427 r
F P
0 g
s
.49227 .3265 m .47781 .29115 L .47732 .28962 L 0 0 .427 r
F
.4494 .26169 m .48088 .30088 L p .48413 .31118 L .797 .335 0 r
F P
0 g
s
.43792 1.00503 m .40704 1.01365 L .44311 1.02555 L p .567 .729 .951 r
F P
0 g
s
.44311 1.02555 m .46511 1.01743 L .43792 1.00503 L p .567 .729 .951 r
F P
0 g
s
.37983 1.00169 m .34131 .99985 L .37561 1.01149 L p .339 .713 .984 r
F P
0 g
s
.37561 1.01149 m .40704 1.01365 L .37983 1.00169 L p .339 .713 .984 r
F P
0 g
s
.94435 .62387 m .9426 .57213 L .9541 .58716 L p 0 0 .259 r
F P
0 g
s
.9426 .57213 m .928 .52246 L .93874 .53722 L p 0 0 .35 r
F P
0 g
s
.93874 .53722 m .9541 .58716 L .9426 .57213 L p 0 0 .35 r
F P
0 g
s
.24389 .95249 m .2147 .92089 L .24632 .93388 L p .844 .379 .057 r
F P
0 g
s
.24632 .93388 m .27225 .96468 L .24389 .95249 L p .844 .379 .057 r
F P
0 g
s
.6685 1.08496 m .69369 1.07347 L .72636 1.08701 L p .585 .23 .334 r
F P
0 g
s
.75332 1.06754 m .72636 1.08701 L .69369 1.07347 L p .554 .263 .427 r
F P
0 g
s
.928 .52246 m .90153 .47815 L .91094 .49302 L p 0 0 .405 r
F P
0 g
s
.91094 .49302 m .93874 .53722 L .928 .52246 L p 0 0 .405 r
F P
0 g
s
.22136 .71966 m .25251 .69009 L .27991 .71029 L p .835 .876 .873 r
F P
0 g
s
.3135 .69025 m .27991 .71029 L .25251 .69009 L p .756 .862 .927 r
F P
0 g
s
.93312 .67423 m .94435 .62387 L .95595 .63942 L p 0 0 .159 r
F P
0 g
s
.9541 .58716 m .95595 .63942 L .94435 .62387 L p 0 0 .259 r
F P
0 g
s
.31673 .15633 m .37393 .17291 L .38683 .20503 L p .417 .64 .948 r
F P
0 g
s
.43324 .23342 m .38683 .20503 L .37393 .17291 L p .304 .513 .894 r
F P
0 g
s
.60395 .40464 m .57814 .44144 L .58821 .46272 L p .607 .37 .531 r
F P
0 g
s
.58821 .46272 m .61996 .4255 L .60395 .40464 L p .607 .37 .531 r
F P
0 g
s
.46288 .55037 m .4227 .59175 L .44064 .61346 L p .722 .704 .832 r
F P
0 g
s
.44064 .61346 m .47416 .57239 L .46288 .55037 L p .722 .704 .832 r
F P
0 g
s
.53137 .42132 m .5296 .37466 L .5115 .35097 L p .844 .593 .558 r
F P
0 g
s
.5296 .37466 m .53137 .42132 L .54206 .44382 L p .78 .555 .594 r
F P
0 g
s
.54206 .44382 m .54766 .39755 L .5296 .37466 L p .78 .555 .594 r
F P
0 g
s
.90153 .47815 m .86491 .44218 L .87264 .45758 L p .587 .06 0 r
F P
0 g
s
.87264 .45758 m .91094 .49302 L .90153 .47815 L p .587 .06 0 r
F P
0 g
s
.60592 .79147 m .58784 .80219 L .60987 .81925 L p .68 .361 .428 r
F P
0 g
s
.60987 .81925 m .63541 .80864 L .60592 .79147 L p .68 .361 .428 r
F P
0 g
s
.52071 .39826 m .5115 .35097 L .4948 .3264 L p .919 .623 .465 r
F P
0 g
s
.5115 .35097 m .52071 .39826 L .53137 .42132 L p .844 .593 .558 r
F P
0 g
s
.90963 .71984 m .93312 .67423 L .94403 .69027 L p 0 0 .068 r
F P
0 g
s
.94383 .69064 m .90963 .71984 L .94403 .69027 L 0 0 .068 r
F
.8754 .75762 m .90963 .71984 L .91943 .73657 L p 0 0 .003 r
F P
0 g
s
.94412 .69039 m .91943 .73657 L .90963 .71984 L p 0 0 .068 r
F P
0 g
s
.91943 .73657 m .88358 .77476 L .8754 .75762 L p 0 0 .003 r
F P
0 g
s
.92071 .70728 m .89693 .75389 L .91923 .73672 L p .854 .946 .585 r
F P
0 g
s
.91923 .73672 m .91943 .73657 L p .91944 .73634 L .854 .946 .585 r
F P
0 g
s
.92071 .70728 m .91923 .73672 L .91944 .73634 L .854 .946 .585 r
F
.94403 .69027 m .94412 .69039 L p .94383 .69064 L 0 0 .068 r
F P
0 g
s
.95595 .63942 m .94412 .69039 L .93312 .67423 L p 0 0 .159 r
F P
0 g
s
.51897 .56483 m .50873 .61435 L .52913 .6344 L p .524 .508 .78 r
F P
0 g
s
.52913 .6344 m .53108 .58563 L .51897 .56483 L p .524 .508 .78 r
F P
0 g
s
.65906 .39954 m .61996 .4255 L .62878 .4461 L p .524 .217 .388 r
F P
0 g
s
.62878 .4461 m .67216 .41928 L .65906 .39954 L p .524 .217 .388 r
F P
0 g
s
.54766 .39755 m .54206 .44382 L .55191 .46583 L p .724 .516 .608 r
F P
0 g
s
.55191 .46583 m .56427 .41976 L .54766 .39755 L p .724 .516 .608 r
F P
0 g
s
.50784 .54352 m .48989 .5937 L .50873 .61435 L p .609 .565 .78 r
F P
0 g
s
.50873 .61435 m .51897 .56483 L .50784 .54352 L p .609 .565 .78 r
F P
0 g
s
.53152 .78773 m .53042 .76739 L .49943 .74953 L p .788 .465 .443 r
F P
0 g
s
.53042 .76739 m .53152 .78773 L .55264 .80497 L p .748 .465 .502 r
F P
0 g
s
.55264 .80497 m .56067 .78492 L .53042 .76739 L p .748 .465 .502 r
F P
0 g
s
.50991 .77018 m .49943 .74953 L .47017 .73125 L p .824 .427 .298 r
F P
0 g
s
.49943 .74953 m .50991 .77018 L .53152 .78773 L p .788 .465 .443 r
F P
0 g
s
.66189 .8075 m .63541 .80864 L .65589 .82567 L p .652 .245 .259 r
F P
0 g
s
.65589 .82567 m .68787 .82452 L .66189 .8075 L p .652 .245 .259 r
F P
0 g
s
.7969 .97422 m .79583 .93848 L .81887 .95436 L p .217 .182 .603 r
F P
0 g
s
.79583 .93848 m .78665 .90432 L .80878 .9205 L p .115 .197 .673 r
F P
0 g
s
.80878 .9205 m .81887 .95436 L .79583 .93848 L p .115 .197 .673 r
F P
0 g
s
.86491 .44218 m .82056 .41697 L .82649 .43328 L p .566 0 0 r
F P
0 g
s
.82649 .43328 m .87264 .45758 L .86491 .44218 L p .566 0 0 r
F P
0 g
s
.28021 .9754 m .24389 .95249 L .27225 .96468 L p .561 0 0 r
F P
0 g
s
.27225 .96468 m .30463 .9872 L .28021 .9754 L p .561 0 0 r
F P
0 g
s
.35077 .68033 m .38922 .68129 L .4147 .701 L p .458 .715 .979 r
F P
0 g
s
.44503 .71248 m .4147 .701 L .38922 .68129 L p .18 .547 .929 r
F P
0 g
s
.78978 1.00917 m .7969 .97422 L .82004 .98976 L p .286 .163 .532 r
F P
0 g
s
.81887 .95436 m .82004 .98976 L .7969 .97422 L p .217 .182 .603 r
F P
0 g
s
.78665 .90432 m .76997 .87401 L .79047 .89045 L p 0 .191 .718 r
F P
0 g
s
.79047 .89045 m .80878 .9205 L .78665 .90432 L p 0 .191 .718 r
F P
0 g
s
.61599 1.07966 m .5922 1.06713 L p .60824 1.07356 L .602 .04 0 r
F P
0 g
s
.642 1.08711 m .61599 1.07966 L p .60824 1.07356 L .602 .04 0 r
F P
0 g
s
.61599 1.07966 m .63248 1.09216 L p .62476 1.08657 L .781 .301 .088 r
F P
0 g
s
.5922 1.06713 m .61599 1.07966 L p .60849 1.07426 L .487 0 0 r
F P
0 g
s
.61062 1.07446 m .5922 1.06713 L p .57797 1.06089 L .317 0 0 r
F P
0 g
s
.5922 1.06713 m .61062 1.07446 L .642 1.08711 L p .602 .04 0 r
F P
0 g
s
.57539 1.05041 m .5922 1.06713 L p .60849 1.07426 L .487 0 0 r
F P
0 g
s
.59229 1.06304 m .61599 1.07966 L p .62476 1.08657 L .781 .301 .088 r
F P
0 g
s
.5922 1.06713 m .56288 1.05462 L p .57797 1.06089 L .317 0 0 r
F P
0 g
s
.6685 1.08496 m .642 1.08711 L .61062 1.07446 L p .632 .179 .158 r
F P
0 g
s
.25251 .69009 m .29016 .66961 L .3135 .69025 L p .756 .862 .927 r
F P
0 g
s
.35077 .68033 m .3135 .69025 L .29016 .66961 L p .656 .819 .962 r
F P
0 g
s
.5076 1.0121 m .49675 1.02979 L .53022 1.04217 L p .533 .625 .889 r
F P
0 g
s
.53022 1.04217 m .53134 1.02493 L .5076 1.0121 L p .533 .625 .889 r
F P
0 g
s
.53134 1.02493 m .53022 1.04217 L .56288 1.05462 L p .403 .58 .912 r
F P
0 g
s
.41327 .56916 m .3603 .5979 L .37448 .62052 L p .736 .798 .902 r
F P
0 g
s
.37448 .62052 m .4227 .59175 L .41327 .56916 L p .736 .798 .902 r
F P
0 g
s
.77493 1.04099 m .78978 1.00917 L .81221 1.02431 L p .342 .145 .462 r
F P
0 g
s
.82004 .98976 m .81221 1.02431 L .78978 1.00917 L p .286 .163 .532 r
F P
0 g
s
.48523 .99918 m .46511 1.01743 L .49675 1.02979 L p .606 .655 .872 r
F P
0 g
s
.49675 1.02979 m .5076 1.0121 L .48523 .99918 L p .606 .655 .872 r
F P
0 g
s
.56067 .78492 m .55264 .80497 L .57164 .82195 L p .713 .451 .528 r
F P
0 g
s
.57164 .82195 m .58784 .80219 L .56067 .78492 L p .713 .451 .528 r
F P
0 g
s
.76997 .87401 m .74688 .84958 L .76516 .86625 L p 0 .116 .623 r
F P
0 g
s
.76516 .86625 m .79047 .89045 L .76997 .87401 L p 0 .116 .623 r
F P
0 g
s
.47017 .73125 m .48956 .75227 L .50991 .77018 L p .824 .427 .298 r
F P
0 g
s
.49861 .52165 m .47416 .57239 L .48989 .5937 L p .674 .614 .78 r
F P
0 g
s
.48989 .5937 m .50784 .54352 L .49861 .52165 L p .674 .614 .78 r
F P
0 g
s
.642 1.08711 m .6685 1.08496 L .69584 1.09803 L p .63 .208 .224 r
F P
0 g
s
.72636 1.08701 m .69584 1.09803 L .6685 1.08496 L p .585 .23 .334 r
F P
0 g
s
.62901 .74694 m .66964 .77217 L .68391 .78962 L p .785 .336 .181 r
F P
0 g
s
.73357 .80165 m .68391 .78962 L .66964 .77217 L p .664 .175 .08 r
F P
0 g
s
.82056 .41697 m .77144 .40428 L .77565 .42175 L p .518 0 0 r
F P
0 g
s
.77565 .42175 m .82649 .43328 L .82056 .41697 L p .518 0 0 r
F P
0 g
s
.56427 .41976 m .55191 .46583 L .56015 .48743 L p .669 .472 .609 r
F P
0 g
s
.56015 .48743 m .57814 .44144 L .56427 .41976 L p .669 .472 .609 r
F P
0 g
s
.41755 .9925 m .37983 1.00169 L .40704 1.01365 L p .619 .76 .946 r
F P
0 g
s
.40704 1.01365 m .43792 1.00503 L .41755 .9925 L p .619 .76 .946 r
F P
0 g
s
.56288 1.05462 m .55454 1.0377 L .53134 1.02493 L p .403 .58 .912 r
F P
0 g
s
.55454 1.0377 m .56288 1.05462 L .5922 1.06713 L p .105 .458 .887 r
F P
0 g
s
.75332 1.06754 m .77493 1.04099 L .7959 1.05565 L p .397 .134 .394 r
F P
0 g
s
.81221 1.02431 m .7959 1.05565 L .77493 1.04099 L p .342 .145 .462 r
F P
0 g
s
.83268 .785 m .8754 .75762 L .88358 .77476 L p .146 0 0 r
F P
0 g
s
.74688 .84958 m .71889 .83269 L .73455 .84954 L p .615 .047 0 r
F P
0 g
s
.73455 .84954 m .76516 .86625 L .74688 .84958 L p .615 .047 0 r
F P
0 g
s
.32123 .98804 m .28021 .9754 L .30463 .9872 L p 0 .504 .854 r
F P
0 g
s
.30463 .9872 m .34131 .99985 L .32123 .98804 L p 0 .504 .854 r
F P
0 g
s
.46609 .98612 m .43792 1.00503 L .46511 1.01743 L p .657 .685 .865 r
F P
0 g
s
.46511 1.01743 m .48523 .99918 L .46609 .98612 L p .657 .685 .865 r
F P
0 g
s
.37393 .17291 m .42482 .20386 L .43324 .23342 L p .304 .513 .894 r
F P
0 g
s
.47093 .27443 m .43324 .23342 L .42482 .20386 L p .211 .382 .816 r
F P
0 g
s
.43324 .23342 m .47093 .27443 L .48088 .30088 L p 0 .103 .656 r
F P
0 g
s
.29016 .66961 m .3318 .65964 L .35077 .68033 L p .656 .819 .962 r
F P
0 g
s
.38922 .68129 m .35077 .68033 L .3318 .65964 L p .537 .743 .971 r
F P
0 g
s
.49713 .32531 m .50283 .35026 L p .4976 .33783 L 0 0 .427 r
F P
0 g
s
.47093 .27443 m .49713 .32531 L p .4976 .33783 L 0 0 .427 r
F P
0 g
s
.49713 .32531 m .49434 .29976 L p .49662 .31195 L 0 0 .532 r
F P
0 g
s
.50283 .35026 m .49713 .32531 L p .49963 .33768 L .842 .575 .032 r
F P
0 g
s
.48088 .30088 m .50283 .35026 L p .50585 .36214 L .939 .577 .186 r
F P
0 g
s
.51353 .4063 m .50283 .35026 L p .49963 .33768 L .842 .575 .032 r
F P
0 g
s
.50283 .35026 m .51093 .37458 L p .50585 .36214 L .939 .577 .186 r
F P
0 g
s
.5188 .42963 m .51093 .37458 L .50283 .35026 L p .979 .721 .47 r
F P
0 g
s
.51353 .4063 m .5188 .42963 L p .51681 .41974 L .979 .721 .47 r
F P
0 g
s
.50283 .35026 m .51353 .4063 L p .51681 .41974 L .979 .721 .47 r
F P
0 g
s
.50986 .38248 m .51353 .4063 L p .51153 .3964 L .842 .575 .032 r
F P
0 g
s
.49713 .32531 m .50986 .38248 L p .51153 .3964 L .842 .575 .032 r
F P
0 g
s
.50811 .35819 m .50986 .38248 L p .50801 .37264 L 0 0 .532 r
F P
0 g
s
.49434 .29976 m .50811 .35819 L p .50801 .37264 L 0 0 .532 r
F P
0 g
s
.50847 .33346 m .50811 .35819 L p .50657 .34842 L .285 .289 .688 r
F P
0 g
s
.49475 .27369 m .50847 .33346 L p .50657 .34842 L .285 .289 .688 r
F P
0 g
s
.50847 .33346 m .51093 .30838 L p .51046 .31779 L 0 0 .407 r
F P
0 g
s
.50811 .35819 m .50847 .33346 L p .50822 .34301 L .824 .689 .15 r
F P
0 g
s
.5098 .38217 m .49713 .32531 L .49662 .31195 L 0 0 .532 r
F
.50986 .38248 m .50981 .38227 L p .5098 .38217 L F P
0 g
s
.50981 .38227 m .49713 .32531 L p .5098 .38217 L 0 0 .532 r
F P
0 g
s
.50626 .41839 m .50814 .44187 L p .50812 .36707 L .995 .819 .556 r
F P
0 g
s
.50811 .35819 m .5079 .36513 L p .50812 .36707 L .995 .819 .556 r
F P
0 g
s
.5079 .36513 m .50626 .41839 L p .50812 .36707 L .995 .819 .556 r
F P
0 g
s
.49434 .29976 m .49475 .27369 L p .50657 .34842 L .285 .289 .688 r
F P
0 g
s
.50811 .35819 m .50621 .35011 L p .50657 .34842 L .285 .289 .688 r
F P
0 g
s
.50621 .35011 m .49434 .29976 L p .50657 .34842 L .285 .289 .688 r
F P
0 g
s
.50964 .33401 m .50847 .33346 L .51046 .31779 L 0 0 .407 r
F
.50626 .41839 m .5076 .37496 L p .50716 .38369 L .824 .689 .15 r
F P
0 g
s
.5076 .37496 m .50811 .35819 L p .50716 .38369 L .824 .689 .15 r
F P
0 g
s
.50716 .38369 m .50811 .35819 L .50822 .34301 L .824 .689 .15 r
F
.50986 .38248 m .50811 .35819 L p .50812 .36791 L .995 .819 .556 r
F P
0 g
s
.50662 .39447 m .50626 .41839 L p .50653 .40773 L .824 .689 .15 r
F P
0 g
s
.50847 .33346 m .50662 .39447 L p .50653 .40773 L .824 .689 .15 r
F P
0 g
s
.50662 .39447 m .50825 .34053 L p .50964 .33401 L 0 0 .407 r
F P
0 g
s
.50814 .44187 m .50986 .38248 L p .50812 .36791 L .995 .819 .556 r
F P
0 g
s
.50925 .37019 m .50662 .39447 L p .50714 .384 L 0 0 .407 r
F P
0 g
s
.51093 .30838 m .50925 .37019 L p .50714 .384 L 0 0 .407 r
F P
0 g
s
.51533 .283 m .51396 .34561 L .50925 .37019 L p .276 .209 .599 r
F P
0 g
s
.50925 .37019 m .51396 .34561 L p .51142 .35591 L .951 .891 .496 r
F P
0 g
s
.50925 .37019 m .51093 .30838 L .51533 .283 L p .276 .209 .599 r
F P
0 g
s
.49346 .42878 m .50925 .37019 L p .51142 .35591 L .951 .891 .496 r
F P
0 g
s
.50662 .39447 m .50925 .37019 L p .50671 .38069 L .964 .861 .689 r
F P
0 g
s
.48931 .45271 m .50662 .39447 L p .50671 .38069 L .964 .861 .689 r
F P
0 g
s
.50825 .34053 m .50847 .33346 L p .50964 .33401 L 0 0 .407 r
F P
0 g
s
.51353 .4063 m .50986 .38248 L p .51013 .39235 L .915 .749 .662 r
F P
0 g
s
.51209 .46489 m .51353 .4063 L p .51013 .39235 L .915 .749 .662 r
F P
0 g
s
.63541 .80864 m .60987 .81925 L .62514 .83614 L p .649 .341 .437 r
F P
0 g
s
.62514 .83614 m .65589 .82567 L .63541 .80864 L p .649 .341 .437 r
F P
0 g
s
.4274 1.15289 m .42781 1.12939 L .48054 1.13835 L p .724 .587 .702 r
F P
0 g
s
.48054 1.13835 m .48035 1.16105 L .4274 1.15289 L p .724 .587 .702 r
F P
0 g
s
.48236 1.11677 m .48054 1.13835 L .42781 1.12939 L p .72 .601 .723 r
F P
0 g
s
.48035 1.16105 m .48054 1.13835 L .53428 1.14751 L p .7 .567 .702 r
F P
0 g
s
.48054 1.13835 m .48236 1.11677 L .53427 1.12666 L p .696 .581 .722 r
F P
0 g
s
.53427 1.12666 m .53428 1.14751 L .48054 1.13835 L p .696 .581 .722 r
F P
0 g
s
.77144 .40428 m .72082 .40504 L .72359 .42373 L p .466 0 0 r
F P
0 g
s
.72359 .42373 m .77565 .42175 L .77144 .40428 L p .466 0 0 r
F P
0 g
s
.58784 .80219 m .57164 .82195 L .58704 .83871 L p .679 .429 .536 r
F P
0 g
s
.58704 .83871 m .60987 .81925 L .58784 .80219 L p .679 .429 .536 r
F P
0 g
s
.43026 1.17596 m .4274 1.15289 L .48035 1.16105 L p .726 .571 .68 r
F P
0 g
s
.48035 1.16105 m .48178 1.18336 L .43026 1.17596 L p .726 .571 .68 r
F P
0 g
s
.48178 1.18336 m .48035 1.16105 L .53429 1.16945 L p .702 .553 .681 r
F P
0 g
s
.53428 1.14751 m .53429 1.16945 L .48035 1.16105 L p .7 .567 .702 r
F P
0 g
s
.42781 1.12939 m .43144 1.10702 L .48236 1.11677 L p .72 .601 .723 r
F P
0 g
s
.48569 1.09775 m .48236 1.11677 L .43144 1.10702 L p .714 .616 .746 r
F P
0 g
s
.48236 1.11677 m .48569 1.09775 L .53425 1.1083 L p .69 .596 .745 r
F P
0 g
s
.53425 1.1083 m .53427 1.12666 L .48236 1.11677 L p .69 .596 .745 r
F P
0 g
s
.04282 .3073 m .06117 .248 L .07199 .29062 L p .934 .908 .518 r
F P
0 g
s
.07199 .29062 m .05467 .34789 L .04282 .3073 L p .934 .908 .518 r
F P
0 g
s
.10338 .23963 m .07199 .29062 L .06117 .248 L p .91 .982 .729 r
F P
0 g
s
.61996 .4255 m .58821 .46272 L .59372 .48374 L p .527 .289 .497 r
F P
0 g
s
.59372 .48374 m .62878 .4461 L .61996 .4255 L p .527 .289 .497 r
F P
0 g
s
.53429 1.16945 m .53428 1.14751 L .5847 1.15682 L p .677 .548 .701 r
F P
0 g
s
.53428 1.14751 m .53427 1.12666 L .58297 1.13667 L p .673 .562 .722 r
F P
0 g
s
.58297 1.13667 m .5847 1.15682 L .53428 1.14751 L p .673 .562 .722 r
F P
0 g
s
.06117 .248 m .09431 .19602 L .10338 .23963 L p .91 .982 .729 r
F P
0 g
s
.14674 .19833 m .10338 .23963 L .09431 .19602 L p .843 .984 .866 r
F P
0 g
s
.72636 1.08701 m .75332 1.06754 L .7722 1.08167 L p .458 .135 .331 r
F P
0 g
s
.7959 1.05565 m .7722 1.08167 L .75332 1.06754 L p .397 .134 .394 r
F P
0 g
s
.53429 1.16945 m .53429 1.19101 L .48178 1.18336 L p .702 .553 .681 r
F P
0 g
s
.53429 1.19101 m .53429 1.16945 L .5849 1.17802 L p .68 .534 .681 r
F P
0 g
s
.5847 1.15682 m .5849 1.17802 L .53429 1.16945 L p .677 .548 .701 r
F P
0 g
s
.04061 .36977 m .04282 .3073 L .05467 .34789 L p .885 .74 .232 r
F P
0 g
s
.05467 .34789 m .0526 .40757 L .04061 .36977 L p .885 .74 .232 r
F P
0 g
s
.43618 1.19706 m .43026 1.17596 L .48178 1.18336 L p .728 .553 .655 r
F P
0 g
s
.48178 1.18336 m .48476 1.20379 L .43618 1.19706 L p .728 .553 .655 r
F P
0 g
s
.48476 1.20379 m .48178 1.18336 L .53429 1.19101 L p .704 .536 .658 r
F P
0 g
s
.38922 .68129 m .42625 .69314 L .44503 .71248 L p .18 .547 .929 r
F P
0 g
s
.47218 .73395 m .4593 .71516 L p .46725 .7263 L 0 0 .49 r
F P
0 g
s
.47218 .73395 m .48956 .75227 L p .47642 .74053 L .751 .234 0 r
F P
0 g
s
.50533 .78176 m .48956 .75227 L .47218 .73395 L p .901 .5 .234 r
F P
0 g
s
.4942 .76402 m .47218 .73395 L p .46725 .7263 L 0 0 .49 r
F P
0 g
s
.44503 .71248 m .47218 .73395 L p .47642 .74053 L .751 .234 0 r
F P
0 g
s
.49209 .4992 m .46288 .55037 L .47416 .57239 L p .731 .662 .782 r
F P
0 g
s
.47416 .57239 m .49861 .52165 L .49209 .4992 L p .731 .662 .782 r
F P
0 g
s
.53427 1.12666 m .53425 1.1083 L .57984 1.1189 L p .665 .578 .747 r
F P
0 g
s
.57984 1.1189 m .58297 1.13667 L .53427 1.12666 L p .665 .578 .747 r
F P
0 g
s
.71889 .83269 m .68787 .82452 L .70072 .84147 L p .659 .177 .098 r
F P
0 g
s
.70072 .84147 m .73455 .84954 L .71889 .83269 L p .659 .177 .098 r
F P
0 g
s
.57814 .44144 m .56015 .48743 L .56615 .50868 L p .607 .417 .599 r
F P
0 g
s
.56615 .50868 m .58821 .46272 L .57814 .44144 L p .607 .417 .599 r
F P
0 g
s
.37996 1.14494 m .38055 1.1206 L .42781 1.12939 L p .753 .611 .701 r
F P
0 g
s
.42781 1.12939 m .4274 1.15289 L .37996 1.14494 L p .753 .611 .701 r
F P
0 g
s
.43144 1.10702 m .42781 1.12939 L .38055 1.1206 L p .749 .628 .725 r
F P
0 g
s
.78431 .80004 m .83268 .785 L .83905 .80235 L p .331 0 0 r
F P
0 g
s
.88358 .77476 m .83905 .80235 L .83268 .785 L p .146 0 0 r
F P
0 g
s
.09431 .19602 m .13983 .15493 L .14674 .19833 L p .843 .984 .866 r
F P
0 g
s
.19916 .16953 m .14674 .19833 L .13983 .15493 L p .757 .939 .943 r
F P
0 g
s
.3841 1.16881 m .37996 1.14494 L .4274 1.15289 L p .756 .593 .675 r
F P
0 g
s
.4274 1.15289 m .43026 1.17596 L .3841 1.16881 L p .756 .593 .675 r
F P
0 g
s
.05474 .431 m .04061 .36977 L .0526 .40757 L p .741 .475 0 r
F P
0 g
s
.0526 .40757 m .06594 .46567 L .05474 .431 L p .741 .475 0 r
F P
0 g
s
.43144 1.10702 m .43805 1.0873 L .48569 1.09775 L p .714 .616 .746 r
F P
0 g
s
.49029 1.08257 m .48569 1.09775 L .43805 1.0873 L p .703 .634 .777 r
F P
0 g
s
.48569 1.09775 m .49029 1.08257 L .53424 1.09364 L p .678 .614 .777 r
F P
0 g
s
.53424 1.09364 m .53425 1.1083 L .48569 1.09775 L p .678 .614 .777 r
F P
0 g
s
.53429 1.19101 m .53429 1.21076 L .48476 1.20379 L p .704 .536 .658 r
F P
0 g
s
.53429 1.21076 m .53429 1.19101 L .58356 1.19885 L p .682 .518 .657 r
F P
0 g
s
.5849 1.17802 m .58356 1.19885 L .53429 1.19101 L p .68 .534 .681 r
F P
0 g
s
.38055 1.1206 m .38582 1.0974 L .43144 1.10702 L p .749 .628 .725 r
F P
0 g
s
.43805 1.0873 m .43144 1.10702 L .38582 1.0974 L p .741 .644 .752 r
F P
0 g
s
.5922 1.06713 m .57539 1.05041 L .55454 1.0377 L p .105 .458 .887 r
F P
0 g
s
.53225 .47481 m .53137 .42132 L .52071 .39826 L p .821 .625 .638 r
F P
0 g
s
.53137 .42132 m .53225 .47481 L .53933 .49671 L p .759 .58 .654 r
F P
0 g
s
.53933 .49671 m .54206 .44382 L .53137 .42132 L p .759 .58 .654 r
F P
0 g
s
.45714 .52765 m .41327 .56916 L .4227 .59175 L p .775 .759 .838 r
F P
0 g
s
.4227 .59175 m .46288 .55037 L .45714 .52765 L p .775 .759 .838 r
F P
0 g
s
.53241 .8162 m .53152 .78773 L .50991 .77018 L p .795 .54 .552 r
F P
0 g
s
.53152 .78773 m .53241 .8162 L .54613 .83296 L p .75 .52 .583 r
F P
0 g
s
.54613 .83296 m .55264 .80497 L .53152 .78773 L p .75 .52 .583 r
F P
0 g
s
.51843 .79914 m .50991 .77018 L .48956 .75227 L p .853 .55 .475 r
F P
0 g
s
.50991 .77018 m .51843 .79914 L .53241 .8162 L p .795 .54 .552 r
F P
0 g
s
.61599 1.07966 m .642 1.08711 L .66382 1.09981 L p .695 .221 .121 r
F P
0 g
s
.69584 1.09803 m .66382 1.09981 L .642 1.08711 L p .63 .208 .224 r
F P
0 g
s
.13983 .15493 m .19449 .12751 L .19916 .16953 L p .757 .939 .943 r
F P
0 g
s
.25712 .1552 m .19916 .16953 L .19449 .12751 L p .668 .865 .974 r
F P
0 g
s
.53425 1.1083 m .53424 1.09364 L .5755 1.10473 L p .65 .599 .783 r
F P
0 g
s
.5755 1.10473 m .57984 1.1189 L .53425 1.1083 L p .65 .599 .783 r
F P
0 g
s
.52521 .45246 m .52071 .39826 L .51093 .37458 L p .896 .676 .596 r
F P
0 g
s
.52071 .39826 m .52521 .45246 L .53225 .47481 L p .821 .625 .638 r
F P
0 g
s
.39269 1.19059 m .3841 1.16881 L .43026 1.17596 L p .758 .571 .643 r
F P
0 g
s
.43026 1.17596 m .43618 1.19706 L .39269 1.19059 L p .758 .571 .643 r
F P
0 g
s
.36413 .98953 m .32123 .98804 L .34131 .99985 L p .447 .757 .995 r
F P
0 g
s
.34131 .99985 m .37983 1.00169 L .36413 .98953 L p .447 .757 .995 r
F P
0 g
s
.44479 1.21477 m .43618 1.19706 L .48476 1.20379 L p .73 .529 .62 r
F P
0 g
s
.48476 1.20379 m .48909 1.22096 L .44479 1.21477 L p .73 .529 .62 r
F P
0 g
s
.48909 1.22096 m .48476 1.20379 L .53429 1.21076 L p .706 .515 .627 r
F P
0 g
s
.0842 .48667 m .05474 .431 L .06594 .46567 L p 0 0 .437 r
F P
0 g
s
.06594 .46567 m .09381 .51825 L .0842 .48667 L p 0 0 .437 r
F P
0 g
s
.52462 .50946 m .51897 .56483 L .53108 .58563 L p .565 .499 .74 r
F P
0 g
s
.53108 .58563 m .53215 .53103 L .52462 .50946 L p .565 .499 .74 r
F P
0 g
s
.53215 .53103 m .53108 .58563 L .5432 .60595 L p .457 .421 .73 r
F P
0 g
s
.38582 1.0974 m .39543 1.07691 L .43805 1.0873 L p .741 .644 .752 r
F P
0 g
s
.44722 1.07153 m .43805 1.0873 L .39543 1.07691 L p .729 .661 .783 r
F P
0 g
s
.43805 1.0873 m .44722 1.07153 L .49029 1.08257 L p .703 .634 .777 r
F P
0 g
s
.51774 .48741 m .50784 .54352 L .51897 .56483 L p .639 .555 .743 r
F P
0 g
s
.51897 .56483 m .52462 .50946 L .51774 .48741 L p .639 .555 .743 r
F P
0 g
s
.72082 .40504 m .67216 .41928 L .67384 .43905 L p .426 .016 .156 r
F P
0 g
s
.67384 .43905 m .72359 .42373 L .72082 .40504 L p .426 .016 .156 r
F P
0 g
s
.54206 .44382 m .53933 .49671 L .54589 .5182 L p .704 .536 .658 r
F P
0 g
s
.54589 .5182 m .55191 .46583 L .54206 .44382 L p .704 .536 .658 r
F P
0 g
s
.3318 .65964 m .37459 .66085 L .38922 .68129 L p .537 .743 .971 r
F P
0 g
s
.42625 .69314 m .38922 .68129 L .37459 .66085 L p .411 .635 .947 r
F P
0 g
s
.5849 1.17802 m .5847 1.15682 L .62794 1.16621 L p .651 .527 .7 r
F P
0 g
s
.5847 1.15682 m .58297 1.13667 L .62475 1.14672 L p .645 .543 .724 r
F P
0 g
s
.62475 1.14672 m .62794 1.16621 L .5847 1.15682 L p .645 .543 .724 r
F P
0 g
s
.53429 1.21076 m .53428 1.22736 L .48909 1.22096 L p .706 .515 .627 r
F P
0 g
s
.53428 1.22736 m .53429 1.21076 L .58077 1.21791 L p .684 .497 .627 r
F P
0 g
s
.58356 1.19885 m .58077 1.21791 L .53429 1.21076 L p .682 .518 .657 r
F P
0 g
s
.68391 .78962 m .63879 .76472 L .62901 .74694 L p .785 .336 .181 r
F P
0 g
s
.58356 1.19885 m .5849 1.17802 L .6283 1.18669 L p .655 .512 .677 r
F P
0 g
s
.62794 1.16621 m .6283 1.18669 L .5849 1.17802 L p .651 .527 .7 r
F P
0 g
s
.45187 .97285 m .41755 .9925 L .43792 1.00503 L p .699 .722 .867 r
F P
0 g
s
.43792 1.00503 m .46609 .98612 L .45187 .97285 L p .699 .722 .867 r
F P
0 g
s
.51093 .37458 m .5188 .42963 L .52521 .45246 L p .896 .676 .596 r
F P
0 g
s
.49588 1.07227 m .49029 1.08257 L .44722 1.07153 L p .678 .657 .822 r
F P
0 g
s
.49029 1.08257 m .49588 1.07227 L .53422 1.08371 L p .65 .642 .829 r
F P
0 g
s
.53422 1.08371 m .53424 1.09364 L .49029 1.08257 L p .65 .642 .829 r
F P
0 g
s
.55264 .80497 m .54613 .83296 L .55849 .84947 L p .712 .497 .595 r
F P
0 g
s
.55849 .84947 m .57164 .82195 L .55264 .80497 L p .712 .497 .595 r
F P
0 g
s
.5432 .60595 m .53972 .55218 L .53215 .53103 L p .457 .421 .73 r
F P
0 g
s
.58297 1.13667 m .57984 1.1189 L .61894 1.12953 L p .634 .56 .754 r
F P
0 g
s
.61894 1.12953 m .62475 1.14672 L .58297 1.13667 L p .634 .56 .754 r
F P
0 g
s
.19449 .12751 m .25445 .11549 L .25712 .1552 L p .668 .865 .974 r
F P
0 g
s
.31673 .15633 m .25712 .1552 L .25445 .11549 L p .585 .775 .97 r
F P
0 g
s
.51666 .98557 m .5076 1.0121 L .53134 1.02493 L p .573 .575 .817 r
F P
0 g
s
.53134 1.02493 m .53227 .999 L .51666 .98557 L p .573 .575 .817 r
F P
0 g
s
.53227 .999 m .53134 1.02493 L .55454 1.0377 L p .471 .528 .834 r
F P
0 g
s
.48956 .75227 m .50533 .78176 L .51843 .79914 L p .853 .55 .475 r
F P
0 g
s
.502 .97195 m .48523 .99918 L .5076 1.0121 L p .634 .609 .808 r
F P
0 g
s
.5076 1.0121 m .51666 .98557 L .502 .97195 L p .634 .609 .808 r
F P
0 g
s
.69584 1.09803 m .72636 1.08701 L .7427 1.10059 L p .53 .158 .28 r
F P
0 g
s
.7722 1.08167 m .7427 1.10059 L .72636 1.08701 L p .458 .135 .331 r
F P
0 g
s
.40518 1.20881 m .39269 1.19059 L .43618 1.19706 L p .759 .539 .599 r
F P
0 g
s
.43618 1.19706 m .44479 1.21477 L .40518 1.20881 L p .759 .539 .599 r
F P
0 g
s
.12687 .53289 m .0842 .48667 L .09381 .51825 L p 0 .181 .706 r
F P
0 g
s
.09381 .51825 m .13439 .56173 L .12687 .53289 L p 0 .181 .706 r
F P
0 g
s
.53424 1.09364 m .53422 1.08371 L .57025 1.09511 L p .609 .632 .848 r
F P
0 g
s
.57025 1.09511 m .5755 1.10473 L .53424 1.09364 L p .609 .632 .848 r
F P
0 g
s
.68787 .82452 m .65589 .82567 L .66594 .84259 L p .635 .253 .302 r
F P
0 g
s
.66594 .84259 m .70072 .84147 L .68787 .82452 L p .635 .253 .302 r
F P
0 g
s
.58077 1.21791 m .58356 1.19885 L .62581 1.20679 L p .658 .496 .652 r
F P
0 g
s
.6283 1.18669 m .62581 1.20679 L .58356 1.19885 L p .655 .512 .677 r
F P
0 g
s
.51209 .46489 m .49861 .52165 L .50784 .54352 L p .7 .603 .745 r
F P
0 g
s
.50784 .54352 m .51774 .48741 L .51209 .46489 L p .7 .603 .745 r
F P
0 g
s
.42482 .20386 m .46591 .24712 L .47093 .27443 L p .211 .382 .816 r
F P
0 g
s
.49713 .32531 m .47093 .27443 L .46591 .24712 L p .123 .246 .72 r
F P
0 g
s
.73357 .80165 m .78431 .80004 L .78893 .81744 L p .51 .04 .075 r
F P
0 g
s
.83905 .80235 m .78893 .81744 L .78431 .80004 L p .331 0 0 r
F P
0 g
s
.39543 1.07691 m .40873 1.0605 L .44722 1.07153 L p .729 .661 .783 r
F P
0 g
s
.45834 1.0608 m .44722 1.07153 L .40873 1.0605 L p .703 .681 .825 r
F P
0 g
s
.44722 1.07153 m .45834 1.0608 L .49588 1.07227 L p .678 .657 .822 r
F P
0 g
s
.57984 1.1189 m .5755 1.10473 L .6109 1.11579 L p .612 .584 .798 r
F P
0 g
s
.6109 1.11579 m .61894 1.12953 L .57984 1.1189 L p .612 .584 .798 r
F P
0 g
s
.45551 1.22788 m .44479 1.21477 L .48909 1.22096 L p .729 .486 .559 r
F P
0 g
s
.48909 1.22096 m .49448 1.2337 L .45551 1.22788 L p .729 .486 .559 r
F P
0 g
s
.49448 1.2337 m .48909 1.22096 L .53428 1.22736 L p .707 .479 .576 r
F P
0 g
s
.55191 .46583 m .54589 .5182 L .55141 .53933 L p .647 .489 .655 r
F P
0 g
s
.55141 .53933 m .56015 .48743 L .55191 .46583 L p .647 .489 .655 r
F P
0 g
s
.60987 .81925 m .58704 .83871 L .59772 .85526 L p .642 .395 .53 r
F P
0 g
s
.59772 .85526 m .62514 .83614 L .60987 .81925 L p .642 .395 .53 r
F P
0 g
s
.34238 1.13716 m .34312 1.11193 L .38055 1.1206 L p .796 .647 .697 r
F P
0 g
s
.38055 1.1206 m .37996 1.14494 L .34238 1.13716 L p .796 .647 .697 r
F P
0 g
s
.38582 1.0974 m .38055 1.1206 L .34312 1.11193 L p .789 .668 .73 r
F P
0 g
s
.55454 1.0377 m .54758 1.01227 L .53227 .999 L p .471 .528 .834 r
F P
0 g
s
.54758 1.01227 m .55454 1.0377 L .57539 1.05041 L p .243 .421 .839 r
F P
0 g
s
.25445 .11549 m .31559 .11957 L .31673 .15633 L p .585 .775 .97 r
F P
0 g
s
.37393 .17291 m .31673 .15633 L .31559 .11957 L p .515 .679 .94 r
F P
0 g
s
.34755 1.16185 m .34238 1.13716 L .37996 1.14494 L p .801 .623 .661 r
F P
0 g
s
.37996 1.14494 m .3841 1.16881 L .34755 1.16185 L p .801 .623 .661 r
F P
0 g
s
.34312 1.11193 m .34972 1.08784 L .38582 1.0974 L p .789 .668 .73 r
F P
0 g
s
.39543 1.07691 m .38582 1.0974 L .34972 1.08784 L p .778 .686 .763 r
F P
0 g
s
.53428 1.22736 m .53427 1.23968 L .49448 1.2337 L p .707 .479 .576 r
F P
0 g
s
.53427 1.23968 m .53428 1.22736 L .57671 1.23392 L p .686 .465 .58 r
F P
0 g
s
.58077 1.21791 m .57671 1.23392 L .53428 1.22736 L p .684 .497 .627 r
F P
0 g
s
.48954 .95812 m .46609 .98612 L .48523 .99918 L p .681 .642 .805 r
F P
0 g
s
.48523 .99918 m .502 .97195 L .48954 .95812 L p .681 .642 .805 r
F P
0 g
s
.57671 1.23392 m .58077 1.21791 L .62063 1.22516 L p .661 .474 .621 r
F P
0 g
s
.62581 1.20679 m .62063 1.22516 L .58077 1.21791 L p .658 .496 .652 r
F P
0 g
s
.17969 .56645 m .12687 .53289 L .13439 .56173 L p .062 .458 .883 r
F P
0 g
s
.13439 .56173 m .18495 .59312 L .17969 .56645 L p .062 .458 .883 r
F P
0 g
s
.50207 1.06755 m .49588 1.07227 L .45834 1.0608 L p .601 .69 .905 r
F P
0 g
s
.49588 1.07227 m .50207 1.06755 L .5342 1.07918 L p .526 .684 .938 r
F P
0 g
s
.5342 1.07918 m .53422 1.0836 L p .52169 1.07692 L .526 .684 .938 r
F P
0 g
s
.5342 1.07918 m .50207 1.06755 L p .52651 1.07732 L .521 0 0 r
F P
0 g
s
.53419 1.08039 m .5342 1.07955 L p .52651 1.07732 L .521 0 0 r
F P
0 g
s
.5342 1.07955 m .5342 1.07918 L p .52651 1.07732 L .521 0 0 r
F P
0 g
s
.50207 1.06755 m .50529 1.06817 L p .51029 1.07084 L .521 0 0 r
F P
0 g
s
.50529 1.06817 m .50845 1.06878 L p .51029 1.07084 L .521 0 0 r
F P
0 g
s
.45834 1.0608 m .47066 1.05585 L .50207 1.06755 L p .601 .69 .905 r
F P
0 g
s
.50845 1.06878 m .50207 1.06755 L .47066 1.05585 L p .05 .542 .886 r
F P
0 g
s
.47066 1.05585 m .48335 1.05705 L .50845 1.06878 L p .05 .542 .886 r
F P
0 g
s
.50845 1.06878 m .48335 1.05705 L p .50808 1.07198 L .764 .337 .237 r
F P
0 g
s
.53422 1.08371 m .49588 1.07227 L p .52169 1.07692 L .526 .684 .938 r
F P
0 g
s
.51459 1.07591 m .51225 1.07319 L p .50808 1.07198 L .764 .337 .237 r
F P
0 g
s
.51225 1.07319 m .50845 1.06878 L p .50808 1.07198 L .764 .337 .237 r
F P
0 g
s
.50845 1.06878 m .53419 1.08039 L p .51029 1.07084 L .521 0 0 r
F P
0 g
s
.53422 1.0836 m .53422 1.08371 L p .52169 1.07692 L .526 .684 .938 r
F P
0 g
s
.42072 1.22224 m .40518 1.20881 L .44479 1.21477 L p .756 .48 .515 r
F P
0 g
s
.44479 1.21477 m .45551 1.22788 L .42072 1.22224 L p .756 .48 .515 r
F P
0 g
s
.3583 1.18431 m .34755 1.16185 L .3841 1.16881 L p .803 .591 .615 r
F P
0 g
s
.3841 1.16881 m .39269 1.19059 L .3583 1.18431 L p .803 .591 .615 r
F P
0 g
s
.67216 .41928 m .62878 .4461 L .62975 .46663 L p .404 .081 .3 r
F P
0 g
s
.62975 .46663 m .67384 .43905 L .67216 .41928 L p .404 .081 .3 r
F P
0 g
s
.57164 .82195 m .55849 .84947 L .56855 .86573 L p .676 .469 .597 r
F P
0 g
s
.56855 .86573 m .58704 .83871 L .57164 .82195 L p .676 .469 .597 r
F P
0 g
s
.58821 .46272 m .56615 .50868 L .56945 .52966 L p .523 .338 .572 r
F P
0 g
s
.56945 .52966 m .59372 .48374 L .58821 .46272 L p .523 .338 .572 r
F P
0 g
s
.34972 1.08784 m .36173 1.06653 L .39543 1.07691 L p .778 .686 .763 r
F P
0 g
s
.40873 1.0605 m .39543 1.07691 L .36173 1.06653 L p .76 .703 .798 r
F P
0 g
s
.5755 1.10473 m .57025 1.09511 L .60117 1.10645 L p .535 .622 .885 r
F P
0 g
s
.60117 1.10645 m .6109 1.11579 L .5755 1.10473 L p .535 .622 .885 r
F P
0 g
s
.406 .97973 m .36413 .98953 L .37983 1.00169 L p .655 .797 .951 r
F P
0 g
s
.37983 1.00169 m .41755 .9925 L .406 .97973 L p .655 .797 .951 r
F P
0 g
s
.53422 1.08371 m .5342 1.07918 L .56443 1.09072 L p .347 .655 .97 r
F P
0 g
s
.57025 1.09511 m .53422 1.08371 L p .55455 1.08843 L .347 .655 .97 r
F P
0 g
s
.53419 1.08039 m .55844 1.09187 L p .54043 1.08244 L .697 .245 .174 r
F P
0 g
s
.5342 1.07918 m .53419 1.07979 L p .54043 1.08244 L .697 .245 .174 r
F P
0 g
s
.53419 1.07979 m .53419 1.08039 L p .54043 1.08244 L .697 .245 .174 r
F P
0 g
s
.56443 1.09072 m .57015 1.09504 L p .55455 1.08843 L .347 .655 .97 r
F P
0 g
s
.57015 1.09504 m .57025 1.09511 L p .55455 1.08843 L .347 .655 .97 r
F P
0 g
s
.42625 .69314 m .4593 .71516 L .47218 .73395 L p 0 .261 .754 r
F P
0 g
s
.4593 .71516 m .48604 .7459 L .4942 .76402 L p 0 0 .49 r
F P
0 g
s
.4942 .76402 m .48604 .7459 L p .49105 .75761 L .748 .385 0 r
F P
0 g
s
.47218 .73395 m .4942 .76402 L p .49705 .76982 L .901 .5 .234 r
F P
0 g
s
.4942 .76402 m .50533 .78176 L p .49705 .76982 L .901 .5 .234 r
F P
0 g
s
.51629 .81771 m .50533 .78176 L .4942 .76402 L p .939 .633 .432 r
F P
0 g
s
.5095 .80066 m .51629 .81771 L p .5121 .80753 L .939 .633 .432 r
F P
0 g
s
.4942 .76402 m .5095 .80066 L p .5121 .80753 L .939 .633 .432 r
F P
0 g
s
.5095 .80066 m .4942 .76402 L p .49105 .75761 L .748 .385 0 r
F P
0 g
s
.50457 .78327 m .5095 .80066 L p .50521 .79065 L .748 .385 0 r
F P
0 g
s
.48604 .7459 m .50457 .78327 L p .50521 .79065 L .748 .385 0 r
F P
0 g
s
.50457 .78327 m .50197 .76553 L p .50389 .7753 L 0 0 .472 r
F P
0 g
s
.5095 .80066 m .50457 .78327 L p .50671 .79331 L .984 .709 .34 r
F P
0 g
s
.51693 .8413 m .5095 .80066 L p .50671 .79331 L .984 .709 .34 r
F P
0 g
s
.51357 .82469 m .50457 .78327 L p .50389 .7753 L 0 0 .472 r
F P
0 g
s
.50814 .44187 m .49209 .4992 L .49861 .52165 L p .757 .652 .745 r
F P
0 g
s
.49861 .52165 m .51209 .46489 L .50814 .44187 L p .757 .652 .745 r
F P
0 g
s
.40873 1.0605 m .42484 1.04929 L .45834 1.0608 L p .703 .681 .825 r
F P
0 g
s
.47066 1.05585 m .45834 1.0608 L .42484 1.04929 L p .642 .701 .89 r
F P
0 g
s
.48887 .47621 m .45714 .52765 L .46288 .55037 L p .787 .715 .784 r
F P
0 g
s
.46288 .55037 m .49209 .4992 L .48887 .47621 L p .787 .715 .784 r
F P
0 g
s
.31559 .11957 m .37378 .13931 L .37393 .17291 L p .515 .679 .94 r
F P
0 g
s
.42482 .20386 m .37393 .17291 L .37378 .13931 L p .461 .586 .892 r
F P
0 g
s
.66382 1.09981 m .69584 1.09803 L .70939 1.11109 L p .615 .213 .256 r
F P
0 g
s
.7427 1.10059 m .70939 1.11109 L .69584 1.09803 L p .53 .158 .28 r
F P
0 g
s
.53215 .53103 m .53225 .47481 L .52521 .45246 L p .714 .582 .705 r
F P
0 g
s
.53225 .47481 m .53215 .53103 L .53972 .55218 L p .657 .535 .704 r
F P
0 g
s
.53972 .55218 m .53933 .49671 L .53225 .47481 L p .657 .535 .704 r
F P
0 g
s
.52462 .50946 m .52521 .45246 L .5188 .42963 L p .772 .629 .702 r
F P
0 g
s
.52521 .45246 m .52462 .50946 L .53215 .53103 L p .714 .582 .705 r
F P
0 g
s
.37459 .66085 m .41564 .67321 L .42625 .69314 L p .411 .635 .947 r
F P
0 g
s
.4593 .71516 m .42625 .69314 L .41564 .67321 L p .297 .508 .892 r
F P
0 g
s
.6283 1.18669 m .62794 1.16621 L .6609 1.17557 L p .615 .498 .697 r
F P
0 g
s
.62794 1.16621 m .62475 1.14672 L .65659 1.15673 L p .604 .516 .727 r
F P
0 g
s
.65659 1.15673 m .6609 1.17557 L .62794 1.16621 L p .604 .516 .727 r
F P
0 g
s
.63248 1.09216 m .604 1.07554 L .59229 1.06304 L p .781 .301 .088 r
F P
0 g
s
.59229 1.06304 m .604 1.07554 L p .59626 1.06637 L .877 .447 .169 r
F P
0 g
s
.61599 1.07966 m .59229 1.06304 L p .58854 1.05988 L .487 0 0 r
F P
0 g
s
.57257 1.03831 m .59229 1.06304 L p .59626 1.06637 L .877 .447 .169 r
F P
0 g
s
.59229 1.06304 m .57539 1.05041 L p .58854 1.05988 L .487 0 0 r
F P
0 g
s
.66382 1.09981 m .63248 1.09216 L .61599 1.07966 L p .695 .221 .121 r
F P
0 g
s
.5188 .42963 m .51774 .48741 L .52462 .50946 L p .772 .629 .702 r
F P
0 g
s
.56015 .48743 m .55141 .53933 L .55545 .56014 L p .579 .43 .643 r
F P
0 g
s
.55545 .56014 m .56615 .50868 L .56015 .48743 L p .579 .43 .643 r
F P
0 g
s
.46763 1.23548 m .45551 1.22788 L .49448 1.2337 L p .715 .37 .394 r
F P
0 g
s
.49448 1.2337 m .50057 1.24113 L .46763 1.23548 L p .715 .37 .394 r
F P
0 g
s
.50057 1.24113 m .49448 1.2337 L .53427 1.23968 L p .702 .395 .454 r
F P
0 g
s
.62581 1.20679 m .6283 1.18669 L .66139 1.19534 L p .622 .481 .67 r
F P
0 g
s
.6609 1.17557 m .66139 1.19534 L .6283 1.18669 L p .615 .498 .697 r
F P
0 g
s
.3739 1.20303 m .3583 1.18431 L .39269 1.19059 L p .802 .543 .546 r
F P
0 g
s
.39269 1.19059 m .40518 1.20881 L .3739 1.20303 L p .802 .543 .546 r
F P
0 g
s
.65589 .82567 m .62514 .83614 L .63258 .85286 L p .611 .304 .425 r
F P
0 g
s
.63258 .85286 m .66594 .84259 L .65589 .82567 L p .611 .304 .425 r
F P
0 g
s
.68391 .78962 m .73357 .80165 L .73666 .81905 L p .662 .222 .195 r
F P
0 g
s
.78893 .81744 m .73666 .81905 L .73357 .80165 L p .51 .04 .075 r
F P
0 g
s
.62475 1.14672 m .61894 1.12953 L .64875 1.14008 L p .585 .536 .766 r
F P
0 g
s
.64875 1.14008 m .65659 1.15673 L .62475 1.14672 L p .585 .536 .766 r
F P
0 g
s
.23892 .58512 m .17969 .56645 L .18495 .59312 L p .314 .663 .973 r
F P
0 g
s
.18495 .59312 m .24208 .61026 L .23892 .58512 L p .314 .663 .973 r
F P
0 g
s
.57671 1.23392 m .57165 1.24578 L .53427 1.23968 L p .686 .465 .58 r
F P
0 g
s
.57165 1.24578 m .57671 1.23392 L .6131 1.24055 L p .665 .443 .574 r
F P
0 g
s
.62063 1.22516 m .6131 1.24055 L .57671 1.23392 L p .661 .474 .621 r
F P
0 g
s
.36173 1.06653 m .37836 1.04942 L .40873 1.0605 L p .76 .703 .798 r
F P
0 g
s
.42484 1.04929 m .40873 1.0605 L .37836 1.04942 L p .728 .717 .84 r
F P
0 g
s
.53933 .49671 m .53972 .55218 L .54673 .57293 L p .59 .481 .698 r
F P
0 g
s
.54673 .57293 m .54589 .5182 L .53933 .49671 L p .59 .481 .698 r
F P
0 g
s
.51774 .48741 m .5188 .42963 L .51353 .4063 L p .836 .683 .692 r
F P
0 g
s
.57539 1.05041 m .56137 1.02537 L .54758 1.01227 L p .243 .421 .839 r
F P
0 g
s
.56137 1.02537 m .57539 1.05041 L .59229 1.06304 L p 0 .028 .57 r
F P
0 g
s
.53427 1.23968 m .53426 1.24687 L .50057 1.24113 L p .702 .395 .454 r
F P
0 g
s
.53426 1.24687 m .53427 1.23968 L .57165 1.24578 L p .686 .397 .479 r
F P
0 g
s
.51353 .4063 m .51209 .46489 L .51774 .48741 L p .836 .683 .692 r
F P
0 g
s
.62063 1.22516 m .62581 1.20679 L .65802 1.21471 L p .627 .463 .641 r
F P
0 g
s
.66139 1.19534 m .65802 1.21471 L .62581 1.20679 L p .622 .481 .67 r
F P
0 g
s
.61894 1.12953 m .6109 1.11579 L .63789 1.12676 L p .537 .562 .828 r
F P
0 g
s
.63789 1.12676 m .64875 1.14008 L .61894 1.12953 L p .537 .562 .828 r
F P
0 g
s
.57025 1.09511 m .56443 1.09072 L .5904 1.10215 L p 0 .409 .748 r
F P
0 g
s
.5904 1.10215 m .60117 1.10645 L .57025 1.09511 L p 0 .409 .748 r
F P
0 g
s
.63879 .76472 m .60134 .7286 L .59514 .71029 L p .872 .498 .334 r
F P
0 g
s
.46591 .24712 m .49434 .29976 L .49713 .32531 L p .123 .246 .72 r
F P
0 g
s
.43828 1.22994 m .42072 1.22224 L .45551 1.22788 L p .712 .294 .253 r
F P
0 g
s
.45551 1.22788 m .46763 1.23548 L .43828 1.22994 L p .712 .294 .253 r
F P
0 g
s
.52436 .83444 m .51843 .79914 L .50533 .78176 L p .854 .612 .572 r
F P
0 g
s
.51843 .79914 m .52436 .83444 L .53303 .85086 L p .792 .581 .615 r
F P
0 g
s
.53303 .85086 m .53241 .8162 L .51843 .79914 L p .792 .581 .615 r
F P
0 g
s
.53241 .8162 m .53303 .85086 L .5416 .867 L p .747 .553 .632 r
F P
0 g
s
.5416 .867 m .54613 .83296 L .53241 .8162 L p .747 .553 .632 r
F P
0 g
s
.48037 .94402 m .45187 .97285 L .46609 .98612 L p .723 .68 .808 r
F P
0 g
s
.46609 .98612 m .48954 .95812 L .48037 .94402 L p .723 .68 .808 r
F P
0 g
s
.37378 .13931 m .42517 .17326 L .42482 .20386 L p .461 .586 .892 r
F P
0 g
s
.46591 .24712 m .42482 .20386 L .42517 .17326 L p .423 .5 .834 r
F P
0 g
s
.54589 .5182 m .54673 .57293 L .55262 .59334 L p .497 .406 .683 r
F P
0 g
s
.55262 .59334 m .55141 .53933 L .54589 .5182 L p .497 .406 .683 r
F P
0 g
s
.62878 .4461 m .59372 .48374 L .59428 .50461 L p .396 .149 .418 r
F P
0 g
s
.59428 .50461 m .62975 .46663 L .62878 .4461 L p .396 .149 .418 r
F P
0 g
s
.55844 1.09187 m .56443 1.09072 L .5342 1.07918 L p .697 .245 .174 r
F P
0 g
s
.39332 1.21673 m .3739 1.20303 L .40518 1.20881 L p .787 .443 .404 r
F P
0 g
s
.40518 1.20881 m .42072 1.22224 L .39332 1.21673 L p .787 .443 .404 r
F P
0 g
s
.30042 .58774 m .23892 .58512 L .24208 .61026 L p .519 .792 .995 r
F P
0 g
s
.24208 .61026 m .3019 .6119 L .30042 .58774 L p .519 .792 .995 r
F P
0 g
s
.6131 1.24055 m .62063 1.22516 L .65101 1.23237 L p .632 .441 .605 r
F P
0 g
s
.65802 1.21471 m .65101 1.23237 L .62063 1.22516 L p .627 .463 .641 r
F P
0 g
s
.42484 1.04929 m .44269 1.04406 L .47066 1.05585 L p .642 .701 .89 r
F P
0 g
s
.48335 1.05705 m .47066 1.05585 L .44269 1.04406 L p .414 .679 .971 r
F P
0 g
s
.50986 .38248 m .50814 .44187 L .51209 .46489 L p .915 .749 .662 r
F P
0 g
s
.37836 1.04942 m .39849 1.03769 L .42484 1.04929 L p .728 .717 .84 r
F P
0 g
s
.44269 1.04406 m .42484 1.04929 L .39849 1.03769 L p .669 .723 .889 r
F P
0 g
s
.50533 .78176 m .51629 .81771 L .52436 .83444 L p .854 .612 .572 r
F P
0 g
s
.54613 .83296 m .5416 .867 L .54937 .88288 L p .708 .525 .638 r
F P
0 g
s
.54937 .88288 m .55849 .84947 L .54613 .83296 L p .708 .525 .638 r
F P
0 g
s
.6109 1.11579 m .60117 1.10645 L .62476 1.11767 L p .32 .577 .934 r
F P
0 g
s
.62476 1.11767 m .63789 1.12676 L .6109 1.11579 L p .32 .577 .934 r
F P
0 g
s
.58704 .83871 m .56855 .86573 L .57555 .88176 L p .637 .432 .589 r
F P
0 g
s
.57555 .88176 m .59772 .85526 L .58704 .83871 L p .637 .432 .589 r
F P
0 g
s
.52325 .95199 m .51666 .98557 L .53227 .999 L p .585 .537 .768 r
F P
0 g
s
.53227 .999 m .53295 .96613 L .52325 .95199 L p .585 .537 .768 r
F P
0 g
s
.53295 .96613 m .53227 .999 L .54758 1.01227 L p .493 .482 .774 r
F P
0 g
s
.5142 .9376 m .502 .97195 L .51666 .98557 L p .643 .576 .764 r
F P
0 g
s
.51666 .98557 m .52325 .95199 L .5142 .9376 L p .643 .576 .764 r
F P
0 g
s
.57165 1.24578 m .56593 1.25268 L .53426 1.24687 L p .686 .397 .479 r
F P
0 g
s
.56593 1.25268 m .57165 1.24578 L .60372 1.25191 L p .668 .386 .484 r
F P
0 g
s
.6131 1.24055 m .60372 1.25191 L .57165 1.24578 L p .665 .443 .574 r
F P
0 g
s
.17843 .77981 m .19135 .73708 L .19883 .75633 L p .943 .912 .533 r
F P
0 g
s
.19883 .75633 m .18642 .79761 L .17854 .78008 L p .943 .912 .533 r
F P
0 g
s
.19883 .75633 m .17854 .78008 L .17873 .77946 L .943 .912 .533 r
F
.22136 .71966 m .19883 .75633 L .19135 .73708 L p .91 .981 .73 r
F P
0 g
s
.19135 .73708 m .21477 .69924 L .22136 .71966 L p .91 .981 .73 r
F P
0 g
s
.20035 .76104 m .2127 .71682 L .19165 .73679 L p 0 0 .213 r
F P
0 g
s
.19165 .73679 m .19135 .73708 L p .19147 .7374 L 0 0 .213 r
F P
0 g
s
.19147 .7374 m .20035 .76104 L .19165 .73679 L 0 0 .213 r
F
.17854 .78008 m .17843 .77981 L p .17873 .77946 L .943 .912 .533 r
F P
0 g
s
.25251 .69009 m .22136 .71966 L .21477 .69924 L p .841 .985 .866 r
F P
0 g
s
.50626 .41839 m .48887 .47621 L .49209 .4992 L p .818 .707 .743 r
F P
0 g
s
.49209 .4992 m .50814 .44187 L .50626 .41839 L p .818 .707 .743 r
F P
0 g
s
.17692 .82448 m .17843 .77981 L .18642 .79761 L p .916 .764 .277 r
F P
0 g
s
.18642 .79761 m .18498 .84073 L .17692 .82448 L p .916 .764 .277 r
F P
0 g
s
.44394 .95929 m .406 .97973 L .41755 .9925 L p .737 .77 .879 r
F P
0 g
s
.41755 .9925 m .45187 .97285 L .44394 .95929 L p .737 .77 .879 r
F P
0 g
s
.48033 1.23702 m .46763 1.23548 L .50057 1.24113 L p .205 0 0 r
F P
0 g
s
.50057 1.24113 m .50696 1.2427 L .48033 1.23702 L p .205 0 0 r
F P
0 g
s
.50696 1.2427 m .50057 1.24113 L .53426 1.24687 L p .498 0 0 r
F P
0 g
s
.63248 1.09216 m .66382 1.09981 L .67454 1.11244 L p .709 .304 .278 r
F P
0 g
s
.70939 1.11109 m .67454 1.11244 L .66382 1.09981 L p .615 .213 .256 r
F P
0 g
s
.63879 .76472 m .68391 .78962 L .68583 .80709 L p .767 .387 .335 r
F P
0 g
s
.73666 .81905 m .68583 .80709 L .68391 .78962 L p .662 .222 .195 r
F P
0 g
s
.42517 .17326 m .46638 .21909 L .46591 .24712 L p .423 .5 .834 r
F P
0 g
s
.49434 .29976 m .46591 .24712 L .46638 .21909 L p .39 .419 .771 r
F P
0 g
s
.56615 .50868 m .55545 .56014 L .5577 .58068 L p .479 .342 .615 r
F P
0 g
s
.5577 .58068 m .56945 .52966 L .56615 .50868 L p .479 .342 .615 r
F P
0 g
s
.21477 .69924 m .24706 .66887 L .25251 .69009 L p .841 .985 .866 r
F P
0 g
s
.29016 .66961 m .25251 .69009 L .24706 .66887 L p .754 .94 .944 r
F P
0 g
s
.18696 .86801 m .17692 .82448 L .18498 .84073 L p .811 .534 0 r
F P
0 g
s
.18498 .84073 m .19462 .88277 L .18696 .86801 L p .811 .534 0 r
F P
0 g
s
.53426 1.24687 m .53424 1.24841 L .50696 1.2427 L p .498 0 0 r
F P
0 g
s
.53424 1.24841 m .53426 1.24687 L .56593 1.25268 L p .611 .124 .073 r
F P
0 g
s
.62514 .83614 m .59772 .85526 L .60293 .8716 L p .595 .344 .507 r
F P
0 g
s
.60293 .8716 m .63258 .85286 L .62514 .83614 L p .595 .344 .507 r
F P
0 g
s
.54758 1.01227 m .54251 .98004 L .53295 .96613 L p .493 .482 .774 r
F P
0 g
s
.54251 .98004 m .54758 1.01227 L .56137 1.02537 L p .29 .365 .767 r
F P
0 g
s
.56443 1.09072 m .55844 1.09187 L .57931 1.10321 L p .717 .353 .36 r
F P
0 g
s
.57931 1.10321 m .5904 1.10215 L .56443 1.09072 L p .717 .353 .36 r
F P
0 g
s
.41564 .67321 m .45215 .69589 L .4593 .71516 L p .297 .508 .892 r
F P
0 g
s
.48604 .7459 m .4593 .71516 L .45215 .69589 L p .209 .38 .815 r
F P
0 g
s
.50657 .92294 m .48954 .95812 L .502 .97195 L p .688 .61 .763 r
F P
0 g
s
.502 .97195 m .5142 .9376 L .50657 .92294 L p .688 .61 .763 r
F P
0 g
s
.35996 .57426 m .30042 .58774 L .3019 .6119 L p .673 .854 .969 r
F P
0 g
s
.3019 .6119 m .3603 .5979 L .35996 .57426 L p .673 .854 .969 r
F P
0 g
s
.60372 1.25191 m .6131 1.24055 L .64084 1.24712 L p .639 .41 .555 r
F P
0 g
s
.65101 1.23237 m .64084 1.24712 L .6131 1.24055 L p .632 .441 .605 r
F P
0 g
s
.31853 1.12945 m .31937 1.10328 L .34312 1.11193 L p .874 .713 .679 r
F P
0 g
s
.34312 1.11193 m .34238 1.13716 L .31853 1.12945 L p .874 .713 .679 r
F P
0 g
s
.34972 1.08784 m .34312 1.11193 L .31937 1.10328 L p .859 .743 .734 r
F P
0 g
s
.31937 1.10328 m .32684 1.07826 L .34972 1.08784 L p .859 .743 .734 r
F P
0 g
s
.36173 1.06653 m .34972 1.08784 L .32684 1.07826 L p .836 .764 .783 r
F P
0 g
s
.45667 1.23136 m .43828 1.22994 L .46763 1.23548 L p .168 .65 .926 r
F P
0 g
s
.46763 1.23548 m .48033 1.23702 L .45667 1.23136 L p .168 .65 .926 r
F P
0 g
s
.24706 .66887 m .28596 .64802 L .29016 .66961 L p .754 .94 .944 r
F P
0 g
s
.3318 .65964 m .29016 .66961 L .28596 .64802 L p .666 .862 .974 r
F P
0 g
s
.41524 1.22447 m .39332 1.21673 L .42072 1.22224 L p .617 .072 0 r
F P
0 g
s
.42072 1.22224 m .43828 1.22994 L .41524 1.22447 L p .617 .072 0 r
F P
0 g
s
.32439 1.15498 m .31853 1.12945 L .34238 1.13716 L p .883 .672 .613 r
F P
0 g
s
.34238 1.13716 m .34755 1.16185 L .32439 1.15498 L p .883 .672 .613 r
F P
0 g
s
.60117 1.10645 m .5904 1.10215 L .61023 1.11343 L p .638 .104 0 r
F P
0 g
s
.61023 1.11343 m .62476 1.11767 L .60117 1.10645 L p .638 .104 0 r
F P
0 g
s
.55849 .84947 m .54937 .88288 L .55573 .89849 L p .671 .494 .637 r
F P
0 g
s
.55573 .89849 m .56855 .86573 L .55849 .84947 L p .671 .494 .637 r
F P
0 g
s
.20784 .90737 m .18696 .86801 L .19462 .88277 L p 0 0 .338 r
F P
0 g
s
.19462 .88277 m .2147 .92089 L .20784 .90737 L p 0 0 .338 r
F P
0 g
s
.50845 1.06878 m .51459 1.07591 L .53417 1.0873 L p .756 .433 .438 r
F P
0 g
s
.53417 1.0873 m .53419 1.08039 L .50845 1.06878 L p .756 .433 .438 r
F P
0 g
s
.39849 1.03769 m .42076 1.03213 L .44269 1.04406 L p .669 .723 .889 r
F P
0 g
s
.46107 1.04519 m .44269 1.04406 L .42076 1.03213 L p .545 .699 .941 r
F P
0 g
s
.44269 1.04406 m .46107 1.04519 L .48335 1.05705 L p .414 .679 .971 r
F P
0 g
s
.32684 1.07826 m .34044 1.05606 L .36173 1.06653 L p .836 .764 .783 r
F P
0 g
s
.37836 1.04942 m .36173 1.06653 L .34044 1.05606 L p .802 .776 .829 r
F P
0 g
s
.59229 1.06304 m .57257 1.03831 L p .57 1.03589 L 0 .028 .57 r
F P
0 g
s
.58035 1.05107 m .57257 1.03831 L p .58164 1.04905 L .877 .447 .169 r
F P
0 g
s
.604 1.07554 m .58035 1.05107 L p .58164 1.04905 L .877 .447 .169 r
F P
0 g
s
.58035 1.05107 m .58417 1.06359 L p .57907 1.05308 L .88 .611 .528 r
F P
0 g
s
.57257 1.03831 m .58035 1.05107 L p .57517 1.04077 L .945 .585 .276 r
F P
0 g
s
.55825 1.00718 m .57257 1.03831 L p .57517 1.04077 L .945 .585 .276 r
F P
0 g
s
.57257 1.03831 m .56137 1.02537 L p .57 1.03589 L 0 .028 .57 r
F P
0 g
s
.55117 .99372 m .56137 1.02537 L .57257 1.03831 L p 0 0 .527 r
F P
0 g
s
.55825 1.00718 m .55117 .99372 L p .55568 1.00312 L 0 0 .527 r
F P
0 g
s
.57257 1.03831 m .55825 1.00718 L p .55568 1.00312 L 0 0 .527 r
F P
0 g
s
.56322 1.0204 m .55825 1.00718 L p .56308 1.01677 L .945 .585 .276 r
F P
0 g
s
.58035 1.05107 m .56322 1.0204 L p .56308 1.01677 L .945 .585 .276 r
F P
0 g
s
.56322 1.0204 m .5657 1.03337 L p .56335 1.02392 L .946 .692 .521 r
F P
0 g
s
.55825 1.00718 m .56322 1.0204 L p .56077 1.01116 L .832 .534 0 r
F P
0 g
s
.56077 1.01116 m .55043 .97198 L .55825 1.00718 L .832 .534 0 r
F
.55038 .97181 m .55041 .97194 L p .55043 .97198 L F P
0 g
s
.55041 .97194 m .55825 1.00718 L p .55043 .97198 L .832 .534 0 r
F P
0 g
s
.55278 .94931 m .54952 .9347 L p .55328 .97944 L .438 .358 .669 r
F P
0 g
s
.5538 .98568 m .55038 .97181 L p .5524 .97947 L .832 .534 0 r
F P
0 g
s
.56322 1.0204 m .55384 .9858 L p .55248 .97977 L .832 .534 0 r
F P
0 g
s
.56335 1.02392 m .55576 .99354 L .56322 1.0204 L .946 .692 .521 r
F
.5538 .98568 m .55588 .99333 L p .55576 .99354 L F P
0 g
s
.55588 .99333 m .56322 1.0204 L p .55576 .99354 L .946 .692 .521 r
F P
0 g
s
.54556 .95768 m .55117 .99372 L .55825 1.00718 L p .107 .178 .656 r
F P
0 g
s
.55825 1.00718 m .55038 .97181 L .54556 .95768 L p .107 .178 .656 r
F P
0 g
s
.55557 .99927 m .5538 .98568 L p .55576 .99354 L .946 .692 .521 r
F P
0 g
s
.5657 1.03337 m .55557 .99927 L p .55576 .99354 L .946 .692 .521 r
F P
0 g
s
.56137 1.02537 m .55117 .99372 L .54251 .98004 L p .29 .365 .767 r
F P
0 g
s
.55557 .99927 m .55556 1.01258 L p .55538 1.0047 L .928 .739 .256 r
F P
0 g
s
.5538 .98568 m .55557 .99927 L p .55514 .99162 L .092 .082 .552 r
F P
0 g
s
.5552 .99643 m .55516 .99463 L .55557 .99927 L .928 .739 .256 r
F
.55538 1.0047 m .5552 .99643 L .55557 .99927 L F
.55447 .96365 m .55278 .94931 L p .55318 .95648 L .092 .082 .552 r
F P
0 g
s
.56237 .94377 m .56247 .92895 L p .55568 .97036 L .359 .236 .573 r
F P
0 g
s
.55514 .99162 m .55454 .96682 L .55359 .96384 L .092 .082 .552 r
F
.55536 .9926 m .55447 .96365 L p .55514 .99162 L F P
0 g
s
.55278 .94931 m .5538 .98568 L p .55514 .99162 L .092 .082 .552 r
F P
0 g
s
.55248 .97977 m .5538 .98568 L .5524 .97947 L .832 .534 0 r
F
.55384 .9858 m .5538 .98568 L p .55248 .97977 L F P
0 g
s
.5538 .98568 m .55365 .98043 L p .55328 .97944 L .438 .358 .669 r
F P
0 g
s
.54952 .9347 m .55038 .97181 L .5538 .98568 L p .438 .358 .669 r
F P
0 g
s
.55282 .99144 m .55447 .97769 L p .55437 .98534 L .93 .746 .632 r
F P
0 g
s
.5538 1.02558 m .55282 .99144 L p .55437 .98534 L .93 .746 .632 r
F P
0 g
s
.55282 .99144 m .54966 1.00488 L p .55136 .99719 L .982 .869 .649 r
F P
0 g
s
.55447 .97769 m .55282 .99144 L p .55438 .98367 L 0 0 .174 r
F P
0 g
s
.55516 .99463 m .55479 .97811 L .55541 .99417 L .928 .739 .256 r
F
.55479 .97811 m .55463 .97103 L .55485 .97737 L F
.55479 .97811 m .55485 .97737 L .55541 .99417 L F
.55463 .97103 m .55465 .97044 L .55541 .99417 L F
.55457 .9669 m .55541 .99417 L p .55456 .96787 L F P
0 g
s
.55454 .96682 m .55447 .96365 L .55359 .96384 L .092 .082 .552 r
F
.55556 1.01258 m .55447 .97769 L .55447 .96365 L p .928 .739 .256 r
F P
0 g
s
.55556 1.01258 m .5538 1.02558 L p .55437 .98534 L .93 .746 .632 r
F P
0 g
s
.55447 .97769 m .55467 .98413 L p .55437 .98534 L .93 .746 .632 r
F P
0 g
s
.55467 .98413 m .55556 1.01258 L p .55437 .98534 L .93 .746 .632 r
F P
0 g
s
.55136 .99719 m .55142 .99695 L .55282 .99144 L .982 .869 .649 r
F
.55997 .9583 m .56237 .94377 L p .56064 .95239 L 0 0 .174 r
F P
0 g
s
.55438 .98367 m .55997 .9583 L .56064 .95239 L 0 0 .174 r
F
.55282 .99144 m .55431 .98452 L p .55438 .98367 L F P
0 g
s
.55431 .98452 m .55997 .9583 L p .55438 .98367 L 0 0 .174 r
F P
0 g
s
.55438 .98367 m .55458 .98266 L .55438 .98357 L 0 0 .174 r
F
.55458 .98266 m .56237 .94377 L .55438 .98357 L F
.56237 .94377 m .5545 .97759 L p .55438 .98357 L F P
0 g
s
.5545 .97759 m .55447 .97769 L p .55438 .98367 L 0 0 .174 r
F P
0 g
s
.55997 .9583 m .55284 .99134 L p .55142 .99695 L .982 .869 .649 r
F P
0 g
s
.5545 .97754 m .55597 .97127 L .55568 .97036 L .359 .236 .573 r
F
.5545 .97759 m .55597 .97127 L p .5545 .97754 L F P
0 g
s
.55597 .97127 m .56237 .94377 L p .55568 .97036 L .359 .236 .573 r
F P
0 g
s
.55552 .97252 m .55997 .9583 L p .558 .96721 L .982 .869 .649 r
F P
0 g
s
.54966 1.00488 m .55552 .97252 L p .558 .96721 L .982 .869 .649 r
F P
0 g
s
.55463 .97103 m .55456 .96787 L .55465 .97044 L .928 .739 .256 r
F
.55447 .96365 m .55457 .9669 L p .55456 .96787 L F P
0 g
s
.55447 .96365 m .55447 .97769 L p .55568 .97036 L .359 .236 .573 r
F P
0 g
s
.55365 .98043 m .55278 .94931 L p .55328 .97944 L .438 .358 .669 r
F P
0 g
s
.53972 .9433 m .54251 .98004 L .55117 .99372 L p .432 .402 .725 r
F P
0 g
s
.55117 .99372 m .54556 .95768 L .53972 .9433 L p .432 .402 .725 r
F P
0 g
s
.55552 .97252 m .54942 .98641 L p .55321 .97745 L .91 .841 .763 r
F P
0 g
s
.55997 .9583 m .55552 .97252 L p .55936 .96372 L .842 .904 .493 r
F P
0 g
s
.57479 .92835 m .55997 .9583 L p .55936 .96372 L .842 .904 .493 r
F P
0 g
s
.56766 .9433 m .55552 .97252 L p .55321 .97745 L .91 .841 .763 r
F P
0 g
s
.55284 .99134 m .55282 .99144 L p .55142 .99695 L .982 .869 .649 r
F P
0 g
s
.55447 .97769 m .5545 .97759 L p .5545 .97754 L .359 .236 .573 r
F P
0 g
s
.55359 .96384 m .55447 .96365 L .55318 .95648 L .092 .082 .552 r
F
.56247 .92895 m .55447 .96365 L p .55568 .97036 L .359 .236 .573 r
F P
0 g
s
.55557 .99927 m .55536 .9926 L p .55514 .99162 L .092 .082 .552 r
F P
0 g
s
.55541 .99417 m .55557 .99927 L p .55516 .99463 L .928 .739 .256 r
F P
0 g
s
.56322 1.0204 m .58035 1.05107 L p .57907 1.05308 L .88 .611 .528 r
F P
0 g
s
.59372 .48374 m .56945 .52966 L .56981 .55042 L p .384 .201 .506 r
F P
0 g
s
.56981 .55042 m .59428 .50461 L .59372 .48374 L p .384 .201 .506 r
F P
0 g
s
.60134 .7286 m .57421 .68374 L .57049 .66471 L p .912 .614 .466 r
F P
0 g
s
.81887 .95436 m .80878 .9205 L .80799 .93669 L p .881 .56 .053 r
F P
0 g
s
.80878 .9205 m .79047 .89045 L .78956 .9069 L p .89 .47 .083 r
F P
0 g
s
.78956 .9069 m .80799 .93669 L .80878 .9205 L p .89 .47 .083 r
F P
0 g
s
.46638 .21909 m .49475 .27369 L .49434 .29976 L p .39 .419 .771 r
F P
0 g
s
.82004 .98976 m .81887 .95436 L .81815 .97025 L p .765 .605 .037 r
F P
0 g
s
.80799 .93669 m .81815 .97025 L .81887 .95436 L p .881 .56 .053 r
F P
0 g
s
.53419 1.08039 m .53417 1.0873 L .55268 1.09852 L p .734 .459 .511 r
F P
0 g
s
.55268 1.09852 m .55844 1.09187 L .53419 1.08039 L p .734 .459 .511 r
F P
0 g
s
.33655 1.17814 m .32439 1.15498 L .34755 1.16185 L p .883 .612 .523 r
F P
0 g
s
.34755 1.16185 m .3583 1.18431 L .33655 1.17814 L p .883 .612 .523 r
F P
0 g
s
.79047 .89045 m .76516 .86625 L .76413 .88292 L p .834 .377 .126 r
F P
0 g
s
.76413 .88292 m .78956 .9069 L .79047 .89045 L p .834 .377 .126 r
F P
0 g
s
.66139 1.19534 m .6609 1.17557 L .68145 1.18478 L p .542 .44 .689 r
F P
0 g
s
.6609 1.17557 m .65659 1.15673 L .67645 1.16657 L p .515 .458 .732 r
F P
0 g
s
.67645 1.16657 m .68145 1.18478 L .6609 1.17557 L p .515 .458 .732 r
F P
0 g
s
.49555 1.06435 m .51459 1.07591 L p .50459 1.06987 L .764 .337 .237 r
F P
0 g
s
.48335 1.05705 m .49555 1.06435 L p .50459 1.06987 L .764 .337 .237 r
F P
0 g
s
.49555 1.06435 m .48335 1.05705 L .46107 1.04519 L p .49 0 0 r
F P
0 g
s
.47872 1.05262 m .49555 1.06435 L p .48499 1.05848 L .49 0 0 r
F P
0 g
s
.46107 1.04519 m .47872 1.05262 L p .48499 1.05848 L .49 0 0 r
F P
0 g
s
.47872 1.05262 m .46565 1.0407 L p .47046 1.0449 L 0 .172 .666 r
F P
0 g
s
.49555 1.06435 m .47872 1.05262 L p .48391 1.05724 L .837 .434 .28 r
F P
0 g
s
.50643 1.07728 m .49555 1.06435 L p .48391 1.05724 L .837 .434 .28 r
F P
0 g
s
.49445 1.06589 m .47872 1.05262 L p .47046 1.0449 L 0 .172 .666 r
F P
0 g
s
.28596 .64802 m .32878 .63808 L .3318 .65964 L p .666 .862 .974 r
F P
0 g
s
.37459 .66085 m .3318 .65964 L .32878 .63808 L p .587 .769 .966 r
F P
0 g
s
.41354 .54577 m .35996 .57426 L .3603 .5979 L p .775 .86 .913 r
F P
0 g
s
.3603 .5979 m .41327 .56916 L .41354 .54577 L p .775 .86 .913 r
F P
0 g
s
.65659 1.15673 m .64875 1.14008 L .66733 1.15046 L p .457 .476 .79 r
F P
0 g
s
.66733 1.15046 m .67645 1.16657 L .65659 1.15673 L p .457 .476 .79 r
F P
0 g
s
.56593 1.25268 m .55993 1.25411 L .53424 1.24841 L p .611 .124 .073 r
F P
0 g
s
.55993 1.25411 m .56593 1.25268 L .59313 1.25846 L p .645 .219 .218 r
F P
0 g
s
.60372 1.25191 m .59313 1.25846 L .56593 1.25268 L p .668 .386 .484 r
F P
0 g
s
.65802 1.21471 m .66139 1.19534 L .68202 1.20384 L p .559 .422 .651 r
F P
0 g
s
.68145 1.18478 m .68202 1.20384 L .66139 1.19534 L p .542 .44 .689 r
F P
0 g
s
.81221 1.02431 m .82004 .98976 L .81933 1.00528 L p 0 0 0 r
F P
0 g
s
.81815 .97025 m .81933 1.00528 L .82004 .98976 L p .765 .605 .037 r
F P
0 g
s
.23812 .93985 m .20784 .90737 L .2147 .92089 L p 0 .097 .627 r
F P
0 g
s
.2147 .92089 m .24389 .95249 L .23812 .93985 L p 0 .097 .627 r
F P
0 g
s
.34044 1.05606 m .35925 1.03821 L .37836 1.04942 L p .802 .776 .829 r
F P
0 g
s
.39849 1.03769 m .37836 1.04942 L .35925 1.03821 L p .755 .775 .869 r
F P
0 g
s
.76516 .86625 m .73455 .84954 L .73346 .86636 L p .755 .304 .182 r
F P
0 g
s
.73346 .86636 m .76413 .88292 L .76516 .86625 L p .755 .304 .182 r
F P
0 g
s
.60134 .7286 m .63879 .76472 L .63989 .78242 L p .822 .511 .465 r
F P
0 g
s
.68583 .80709 m .63989 .78242 L .63879 .76472 L p .767 .387 .335 r
F P
0 g
s
.59313 1.25846 m .60372 1.25191 L .62818 1.25796 L p .647 .362 .472 r
F P
0 g
s
.64084 1.24712 m .62818 1.25796 L .60372 1.25191 L p .639 .41 .555 r
F P
0 g
s
.64875 1.14008 m .63789 1.12676 L .65473 1.13754 L p .274 .471 .87 r
F P
0 g
s
.65473 1.13754 m .66733 1.15046 L .64875 1.14008 L p .274 .471 .87 r
F P
0 g
s
.65101 1.23237 m .65802 1.21471 L .6781 1.22248 L p .572 .402 .613 r
F P
0 g
s
.68202 1.20384 m .6781 1.22248 L .65802 1.21471 L p .559 .422 .651 r
F P
0 g
s
.50102 .90799 m .48037 .94402 L .48954 .95812 L p .731 .646 .765 r
F P
0 g
s
.48954 .95812 m .50657 .92294 L .50102 .90799 L p .731 .646 .765 r
F P
0 g
s
.3542 1.19734 m .33655 1.17814 L .3583 1.18431 L p .864 .508 .374 r
F P
0 g
s
.3583 1.18431 m .3739 1.20303 L .3542 1.19734 L p .864 .508 .374 r
F P
0 g
s
.7959 1.05565 m .81221 1.02431 L .81144 1.03939 L p 0 0 0 r
F P
0 g
s
.81933 1.00528 m .81144 1.03939 L .81221 1.02431 L p 0 0 0 r
F P
0 g
s
.49275 1.23236 m .48033 1.23702 L .50696 1.2427 L p .594 .698 .915 r
F P
0 g
s
.50696 1.2427 m .51321 1.23829 L .49275 1.23236 L p .594 .698 .915 r
F P
0 g
s
.51321 1.23829 m .50696 1.2427 L .53424 1.24841 L p .518 .692 .947 r
F P
0 g
s
.52725 .87358 m .52436 .83444 L .51629 .81771 L p .828 .637 .644 r
F P
0 g
s
.52436 .83444 m .52725 .87358 L .53334 .88927 L p .772 .598 .663 r
F P
0 g
s
.53334 .88927 m .53303 .85086 L .52436 .83444 L p .772 .598 .663 r
F P
0 g
s
.5904 1.10215 m .57931 1.10321 L .59529 1.11436 L p .711 .397 .444 r
F P
0 g
s
.59529 1.11436 m .61023 1.11343 L .5904 1.10215 L p .711 .397 .444 r
F P
0 g
s
.55844 1.09187 m .55268 1.09852 L .56866 1.10956 L p .715 .464 .544 r
F P
0 g
s
.56866 1.10956 m .57931 1.10321 L .55844 1.09187 L p .715 .464 .544 r
F P
0 g
s
.45761 .5043 m .41354 .54577 L .41327 .56916 L p .829 .827 .846 r
F P
0 g
s
.41327 .56916 m .45714 .52765 L .45761 .5043 L p .829 .827 .846 r
F P
0 g
s
.53303 .85086 m .53334 .88927 L .5394 .90469 L p .73 .566 .669 r
F P
0 g
s
.5394 .90469 m .5416 .867 L .53303 .85086 L p .73 .566 .669 r
F P
0 g
s
.32878 .63808 m .37262 .63968 L .37459 .66085 L p .587 .769 .966 r
F P
0 g
s
.41564 .67321 m .37459 .66085 L .37262 .63968 L p .525 .673 .931 r
F P
0 g
s
.73455 .84954 m .70072 .84147 L .69963 .85833 L p .676 .258 .245 r
F P
0 g
s
.69963 .85833 m .73346 .86636 L .73455 .84954 L p .676 .258 .245 r
F P
0 g
s
.604 1.07554 m .63248 1.09216 L .64053 1.10452 L p .793 .42 .349 r
F P
0 g
s
.67454 1.11244 m .64053 1.10452 L .63248 1.09216 L p .709 .304 .278 r
F P
0 g
s
.47535 .92961 m .44394 .95929 L .45187 .97285 L p .766 .728 .817 r
F P
0 g
s
.45187 .97285 m .48037 .94402 L .47535 .92961 L p .766 .728 .817 r
F P
0 g
s
.53424 1.24841 m .53422 1.24417 L .51321 1.23829 L p .518 .692 .947 r
F P
0 g
s
.53422 1.24417 m .53424 1.24841 L .55993 1.25411 L p .346 .665 .974 r
F P
0 g
s
.43818 1.22569 m .41524 1.22447 L .43828 1.22994 L p .392 .737 .992 r
F P
0 g
s
.43828 1.22994 m .45667 1.23136 L .43818 1.22569 L p .392 .737 .992 r
F P
0 g
s
.56855 .86573 m .55573 .89849 L .56021 .91385 L p .63 .457 .63 r
F P
0 g
s
.56021 .91385 m .57555 .88176 L .56855 .86573 L p .63 .457 .63 r
F P
0 g
s
.35925 1.03821 m .38199 1.02591 L .39849 1.03769 L p .755 .775 .869 r
F P
0 g
s
.42076 1.03213 m .39849 1.03769 L .38199 1.02591 L p .69 .756 .902 r
F P
0 g
s
.52162 .85759 m .51629 .81771 L .5095 .80066 L p .91 .689 .59 r
F P
0 g
s
.51629 .81771 m .52162 .85759 L .52725 .87358 L p .828 .637 .644 r
F P
0 g
s
.27568 .96321 m .23812 .93985 L .24389 .95249 L p 0 .408 .845 r
F P
0 g
s
.24389 .95249 m .28021 .9754 L .27568 .96321 L p 0 .408 .845 r
F P
0 g
s
.63789 1.12676 m .62476 1.11767 L .63949 1.12868 L p .491 0 0 r
F P
0 g
s
.63949 1.12868 m .65473 1.13754 L .63789 1.12676 L p .491 0 0 r
F P
0 g
s
.48887 .47621 m .50626 .41839 L .50662 .39447 L p .887 .776 .732 r
F P
0 g
s
.59772 .85526 m .57555 .88176 L .579 .89755 L p .585 .378 .567 r
F P
0 g
s
.579 .89755 m .60293 .8716 L .59772 .85526 L p .585 .378 .567 r
F P
0 g
s
.64084 1.24712 m .65101 1.23237 L .66996 1.23942 L p .584 .379 .569 r
F P
0 g
s
.6781 1.22248 m .66996 1.23942 L .65101 1.23237 L p .572 .402 .613 r
F P
0 g
s
.42076 1.03213 m .44367 1.03315 L .46107 1.04519 L p .545 .699 .941 r
F P
0 g
s
.47872 1.05262 m .46107 1.04519 L .44367 1.03315 L p .255 .57 .939 r
F P
0 g
s
.47465 1.22637 m .45667 1.23136 L .48033 1.23702 L p .637 .708 .898 r
F P
0 g
s
.48033 1.23702 m .49275 1.23236 L .47465 1.22637 L p .637 .708 .898 r
F P
0 g
s
.48931 .45271 m .45761 .5043 L .45714 .52765 L p .849 .781 .785 r
F P
0 g
s
.45714 .52765 m .48887 .47621 L .48931 .45271 L p .849 .781 .785 r
F P
0 g
s
.45215 .69589 m .4816 .72737 L .48604 .7459 L p .209 .38 .815 r
F P
0 g
s
.50457 .78327 m .48604 .7459 L .4816 .72737 L p .115 .241 .718 r
F P
0 g
s
.5416 .867 m .5394 .90469 L .54494 .91983 L p .691 .535 .67 r
F P
0 g
s
.54494 .91983 m .54937 .88288 L .5416 .867 L p .691 .535 .67 r
F P
0 g
s
.52091 .89854 m .5142 .9376 L .52325 .95199 L p .668 .569 .734 r
F P
0 g
s
.52325 .95199 m .52687 .91374 L .52091 .89854 L p .668 .569 .734 r
F P
0 g
s
.52687 .91374 m .52325 .95199 L .53295 .96613 L p .62 .532 .734 r
F P
0 g
s
.53295 .96613 m .53331 .92865 L .52687 .91374 L p .62 .532 .734 r
F P
0 g
s
.53331 .92865 m .53295 .96613 L .54251 .98004 L p .552 .484 .733 r
F P
0 g
s
.7722 1.08167 m .7959 1.05565 L .79502 1.07022 L p .001 0 .031 r
F P
0 g
s
.81144 1.03939 m .79502 1.07022 L .7959 1.05565 L p 0 0 0 r
F P
0 g
s
.50662 .39447 m .48931 .45271 L .48887 .47621 L p .887 .776 .732 r
F P
0 g
s
.37614 1.21129 m .3542 1.19734 L .3739 1.20303 L p .758 .261 .041 r
F P
0 g
s
.3739 1.20303 m .39332 1.21673 L .37614 1.21129 L p .758 .261 .041 r
F P
0 g
s
.52006 1.08847 m .51459 1.07591 L .49555 1.06435 L p .793 .508 .506 r
F P
0 g
s
.51459 1.07591 m .52006 1.08847 L .53416 1.09945 L p .755 .516 .57 r
F P
0 g
s
.53416 1.09945 m .53417 1.0873 L .51459 1.07591 L p .755 .516 .57 r
F P
0 g
s
.70072 .84147 m .66594 .84259 L .66496 .85938 L p .609 .238 .314 r
F P
0 g
s
.66496 .85938 m .69963 .85833 L .70072 .84147 L p .609 .238 .314 r
F P
0 g
s
.57421 .68374 m .60134 .7286 L .60197 .74676 L p .842 .59 .558 r
F P
0 g
s
.63989 .78242 m .60197 .74676 L .60134 .7286 L p .822 .511 .465 r
F P
0 g
s
.51593 .88306 m .50657 .92294 L .5142 .9376 L p .711 .603 .734 r
F P
0 g
s
.5142 .9376 m .52091 .89854 L .51593 .88306 L p .711 .603 .734 r
F P
0 g
s
.37262 .63968 m .41449 .65267 L .41564 .67321 L p .525 .673 .931 r
F P
0 g
s
.45215 .69589 m .41564 .67321 L .41449 .65267 L p .482 .585 .881 r
F P
0 g
s
.54251 .98004 m .53972 .9433 L .53331 .92865 L p .552 .484 .733 r
F P
0 g
s
.53417 1.0873 m .53416 1.09945 L .54755 1.11024 L p .729 .511 .596 r
F P
0 g
s
.54755 1.11024 m .55268 1.09852 L .53417 1.0873 L p .729 .511 .596 r
F P
0 g
s
.5095 .80066 m .51693 .8413 L .52162 .85759 L p .91 .689 .59 r
F P
0 g
s
.59313 1.25846 m .58203 1.25974 L .55993 1.25411 L p .645 .219 .218 r
F P
0 g
s
.58203 1.25974 m .59313 1.25846 L .61389 1.26414 L p .653 .264 .295 r
F P
0 g
s
.62818 1.25796 m .61389 1.26414 L .59313 1.25846 L p .647 .362 .472 r
F P
0 g
s
.55993 1.25411 m .55407 1.24996 L .53422 1.24417 L p .346 .665 .974 r
F P
0 g
s
.55407 1.24996 m .55993 1.25411 L .58203 1.25974 L p 0 .462 .794 r
F P
0 g
s
.49555 1.06435 m .50643 1.07728 L .52006 1.08847 L p .793 .508 .506 r
F P
0 g
s
.62476 1.11767 m .61023 1.11343 L .62264 1.1245 L p .725 .329 .299 r
F P
0 g
s
.62264 1.1245 m .63949 1.12868 L .62476 1.11767 L p .725 .329 .299 r
F P
0 g
s
.31792 .97588 m .27568 .96321 L .28021 .9754 L p .267 .647 .966 r
F P
0 g
s
.28021 .9754 m .32123 .98804 L .31792 .97588 L p .267 .647 .966 r
F P
0 g
s
.62818 1.25796 m .64084 1.24712 L .65814 1.25352 L p .598 .352 .515 r
F P
0 g
s
.66996 1.23942 m .65814 1.25352 L .64084 1.24712 L p .584 .379 .569 r
F P
0 g
s
.54937 .88288 m .54494 .91983 L .54952 .9347 L p .652 .502 .667 r
F P
0 g
s
.54952 .9347 m .55573 .89849 L .54937 .88288 L p .652 .502 .667 r
F P
0 g
s
.38199 1.02591 m .40712 1.02 L .42076 1.03213 L p .69 .756 .902 r
F P
0 g
s
.44367 1.03315 m .42076 1.03213 L .40712 1.02 L p .604 .711 .92 r
F P
0 g
s
.7427 1.10059 m .7722 1.08167 L .77119 1.09568 L p .24 0 .089 r
F P
0 g
s
.79502 1.07022 m .77119 1.09568 L .7722 1.08167 L p .001 0 .031 r
F P
0 g
s
.57931 1.10321 m .56866 1.10956 L .58095 1.12039 L p .698 .46 .56 r
F P
0 g
s
.58095 1.12039 m .59529 1.11436 L .57931 1.10321 L p .698 .46 .56 r
F P
0 g
s
.49843 .24721 m .51093 .30838 L .50847 .33346 L p .469 .424 .724 r
F P
0 g
s
.50847 .33346 m .49475 .27369 L .49843 .24721 L p .469 .424 .724 r
F P
0 g
s
.4725 .19053 m .49843 .24721 L .49475 .27369 L p .523 .507 .78 r
F P
0 g
s
.49475 .27369 m .46638 .21909 L .4725 .19053 L p .523 .507 .78 r
F P
0 g
s
.55932 .63326 m .57421 .68374 L .5746 .70257 L p .881 .658 .596 r
F P
0 g
s
.60197 .74676 m .5746 .70257 L .57421 .68374 L p .842 .59 .558 r
F P
0 g
s
.66594 .84259 m .63258 .85286 L .63179 .86941 L p .558 .239 .384 r
F P
0 g
s
.63179 .86941 m .66496 .85938 L .66594 .84259 L p .558 .239 .384 r
F P
0 g
s
.40088 1.21901 m .37614 1.21129 L .39332 1.21673 L p 0 .39 .716 r
F P
0 g
s
.39332 1.21673 m .41524 1.22447 L .40088 1.21901 L p 0 .39 .716 r
F P
0 g
s
.52687 .91374 m .52725 .87358 L .52162 .85759 L p .748 .608 .702 r
F P
0 g
s
.52725 .87358 m .52687 .91374 L .53331 .92865 L p .707 .574 .703 r
F P
0 g
s
.53331 .92865 m .53334 .88927 L .52725 .87358 L p .707 .574 .703 r
F P
0 g
s
.55268 1.09852 m .54755 1.11024 L .55919 1.1208 L p .708 .502 .608 r
F P
0 g
s
.55919 1.1208 m .56866 1.10956 L .55268 1.09852 L p .708 .502 .608 r
F P
0 g
s
.46056 1.2203 m .43818 1.22569 L .45667 1.23136 L p .666 .728 .896 r
F P
0 g
s
.45667 1.23136 m .47465 1.22637 L .46056 1.2203 L p .666 .728 .896 r
F P
0 g
s
.5746 .70257 m .55961 .65284 L .55932 .63326 L p .881 .658 .596 r
F P
0 g
s
.51236 .86726 m .50102 .90799 L .50657 .92294 L p .753 .64 .736 r
F P
0 g
s
.50657 .92294 m .51593 .88306 L .51236 .86726 L p .753 .64 .736 r
F P
0 g
s
.53334 .88927 m .53331 .92865 L .53972 .9433 L p .667 .541 .702 r
F P
0 g
s
.53972 .9433 m .5394 .90469 L .53334 .88927 L p .667 .541 .702 r
F P
0 g
s
.50403 1.22179 m .49275 1.23236 L .51321 1.23829 L p .66 .637 .816 r
F P
0 g
s
.51321 1.23829 m .51888 1.22816 L .50403 1.22179 L p .66 .637 .816 r
F P
0 g
s
.51888 1.22816 m .51321 1.23829 L .53422 1.24417 L p .624 .625 .83 r
F P
0 g
s
.41449 .65267 m .45157 .67612 L .45215 .69589 L p .482 .585 .881 r
F P
0 g
s
.4816 .72737 m .45215 .69589 L .45157 .67612 L p .454 .508 .824 r
F P
0 g
s
.49807 .89271 m .47535 .92961 L .48037 .94402 L p .776 .691 .77 r
F P
0 g
s
.48037 .94402 m .50102 .90799 L .49807 .89271 L p .776 .691 .77 r
F P
0 g
s
.52162 .85759 m .52091 .89854 L .52687 .91374 L p .748 .608 .702 r
F P
0 g
s
.52091 .89854 m .52162 .85759 L .51693 .8413 L p .797 .649 .698 r
F P
0 g
s
.58035 1.05107 m .604 1.07554 L .6097 1.08786 L p .849 .53 .447 r
F P
0 g
s
.64053 1.10452 m .6097 1.08786 L .604 1.07554 L p .793 .42 .349 r
F P
0 g
s
.44367 1.03315 m .46565 1.0407 L .47872 1.05262 L p .255 .57 .939 r
F P
0 g
s
.46565 1.0407 m .48522 1.05427 L .49445 1.06589 L p 0 .172 .666 r
F P
0 g
s
.49445 1.06589 m .48522 1.05427 L p .49059 1.06158 L .795 .331 0 r
F P
0 g
s
.47872 1.05262 m .49445 1.06589 L p .49764 1.06946 L .837 .434 .28 r
F P
0 g
s
.50715 1.0841 m .49445 1.06589 L p .49059 1.06158 L .795 .331 0 r
F P
0 g
s
.49445 1.06589 m .50643 1.07728 L p .49764 1.06946 L .837 .434 .28 r
F P
0 g
s
.53422 1.24417 m .53421 1.23439 L .51888 1.22816 L p .624 .625 .83 r
F P
0 g
s
.53421 1.23439 m .53422 1.24417 L .55407 1.24996 L p .566 .614 .86 r
F P
0 g
s
.61023 1.11343 m .59529 1.11436 L .60532 1.12527 L p .7 .416 .489 r
F P
0 g
s
.60532 1.12527 m .62264 1.1245 L .61023 1.11343 L p .7 .416 .489 r
F P
0 g
s
.36192 .97702 m .31792 .97588 L .32123 .98804 L p .511 .79 .996 r
F P
0 g
s
.32123 .98804 m .36413 .98953 L .36192 .97702 L p .511 .79 .996 r
F P
0 g
s
.4816 .72737 m .50197 .76553 L .50457 .78327 L p .115 .241 .718 r
F P
0 g
s
.51184 .80775 m .51357 .82469 L p .51194 .81641 L 0 0 .472 r
F P
0 g
s
.50197 .76553 m .51184 .80775 L p .51194 .81641 L 0 0 .472 r
F P
0 g
s
.51192 .79046 m .51184 .80775 L p .51051 .79962 L .321 .313 .694 r
F P
0 g
s
.50197 .74742 m .51192 .79046 L p .51051 .79962 L .321 .313 .694 r
F P
0 g
s
.51192 .79046 m .51383 .77283 L p .51341 .78075 L 0 0 .475 r
F P
0 g
s
.51184 .80775 m .51192 .79046 L p .51173 .79854 L .893 .744 .241 r
F P
0 g
s
.50457 .78327 m .51357 .82469 L p .51513 .83287 L .984 .709 .34 r
F P
0 g
s
.50197 .76553 m .50197 .74742 L p .51181 .80754 L .321 .313 .694 r
F P
0 g
s
.51184 .80775 m .51181 .8076 L p .51181 .80754 L .321 .313 .694 r
F P
0 g
s
.51181 .8076 m .50197 .76553 L p .51181 .80754 L .321 .313 .694 r
F P
0 g
s
.51072 .84234 m .51179 .79621 L .51058 .83467 L .893 .744 .241 r
F
.51178 .79494 m .51058 .83467 L p .51179 .79621 L F P
0 g
s
.51072 .84234 m .51184 .80775 L .51173 .79854 L .893 .744 .241 r
F
.51051 .85113 m .51075 .84352 L p .51072 .84234 L F P
0 g
s
.51075 .84352 m .51184 .80775 L p .51072 .84234 L .893 .744 .241 r
F P
0 g
s
.51192 .79046 m .51178 .79494 L p .51179 .79621 L .893 .744 .241 r
F P
0 g
s
.51357 .82469 m .51184 .80775 L p .51191 .81604 L .961 .788 .623 r
F P
0 g
s
.51184 .80775 m .51051 .85113 L .51236 .86726 L p .961 .788 .623 r
F P
0 g
s
.51058 .83467 m .51192 .79046 L p .51341 .78075 L 0 0 .475 r
F P
0 g
s
.51236 .86726 m .51357 .82469 L p .51191 .81604 L .961 .788 .623 r
F P
0 g
s
.51259 .81787 m .51058 .83467 L p .51103 .82607 L 0 0 .475 r
F P
0 g
s
.51383 .77283 m .51259 .81787 L p .51103 .82607 L 0 0 .475 r
F P
0 g
s
.51744 .75487 m .51641 .80072 L .51259 .81787 L p .378 .295 .635 r
F P
0 g
s
.51259 .81787 m .51641 .80072 L p .51422 .80932 L .903 .867 .428 r
F P
0 g
s
.51259 .81787 m .51383 .77283 L .51744 .75487 L p .378 .295 .635 r
F P
0 g
s
.50098 .8611 m .51259 .81787 L p .51422 .80932 L .903 .867 .428 r
F P
0 g
s
.51058 .83467 m .51259 .81787 L p .51044 .82662 L .955 .851 .7 r
F P
0 g
s
.49803 .87709 m .51058 .83467 L p .51044 .82662 L .955 .851 .7 r
F P
0 g
s
.51058 .83467 m .51051 .85113 L p .51072 .84234 L .893 .744 .241 r
F P
0 g
s
.51357 .82469 m .51693 .8413 L p .51513 .83287 L .984 .709 .34 r
F P
0 g
s
.5394 .90469 m .53972 .9433 L .54556 .95768 L p .62 .504 .699 r
F P
0 g
s
.54556 .95768 m .54494 .91983 L .5394 .90469 L p .62 .504 .699 r
F P
0 g
s
.51693 .8413 m .51593 .88306 L .52091 .89854 L p .797 .649 .698 r
F P
0 g
s
.57555 .88176 m .56021 .91385 L .56247 .92895 L p .572 .401 .611 r
F P
0 g
s
.56247 .92895 m .579 .89755 L .57555 .88176 L p .572 .401 .611 r
F P
0 g
s
.61389 1.26414 m .62818 1.25796 L .64344 1.26382 L p .619 .32 .441 r
F P
0 g
s
.65814 1.25352 m .64344 1.26382 L .62818 1.25796 L p .598 .352 .515 r
F P
0 g
s
.70939 1.11109 m .7427 1.10059 L .7416 1.114 L p .438 .03 .167 r
F P
0 g
s
.77119 1.09568 m .7416 1.114 L .7427 1.10059 L p .24 0 .089 r
F P
0 g
s
.49096 1.21529 m .47465 1.22637 L .49275 1.23236 L p .687 .655 .813 r
F P
0 g
s
.49275 1.23236 m .50403 1.22179 L .49096 1.21529 L p .687 .655 .813 r
F P
0 g
s
.55573 .89849 m .54952 .9347 L .55278 .94931 L p .603 .459 .659 r
F P
0 g
s
.55278 .94931 m .56021 .91385 L .55573 .89849 L p .603 .459 .659 r
F P
0 g
s
.40712 1.02 m .43293 1.02089 L .44367 1.03315 L p .604 .711 .92 r
F P
0 g
s
.46565 1.0407 m .44367 1.03315 L .43293 1.02089 L p .502 .638 .915 r
F P
0 g
s
.51593 .88306 m .51693 .8413 L .51357 .82469 L p .863 .705 .683 r
F P
0 g
s
.58203 1.25974 m .57119 1.25562 L p .56328 1.25318 L 0 .462 .794 r
F P
0 g
s
.61389 1.26414 m .59893 1.26521 L .58203 1.25974 L p .653 .264 .295 r
F P
0 g
s
.57119 1.25562 m .58203 1.25974 L .59893 1.26521 L p .566 0 0 r
F P
0 g
s
.58433 1.2611 m .57119 1.25562 L p .58062 1.25888 L .566 0 0 r
F P
0 g
s
.59893 1.26521 m .58433 1.2611 L p .58062 1.25888 L .566 0 0 r
F P
0 g
s
.60953 1.27046 m .59263 1.26634 L .58433 1.2611 L p .699 .259 .201 r
F P
0 g
s
.58433 1.2611 m .59263 1.26634 L p .58841 1.26354 L .765 .28 .083 r
F P
0 g
s
.59893 1.26521 m .60953 1.27046 L p .59108 1.26361 L .699 .259 .201 r
F P
0 g
s
.58433 1.2611 m .59272 1.26346 L p .59108 1.26361 L .699 .259 .201 r
F P
0 g
s
.59272 1.26346 m .59893 1.26521 L p .59108 1.26361 L .699 .259 .201 r
F P
0 g
s
.57109 1.25208 m .58433 1.2611 L p .58841 1.26354 L .765 .28 .083 r
F P
0 g
s
.57119 1.25562 m .58433 1.2611 L p .58005 1.25836 L 0 .377 .817 r
F P
0 g
s
.56135 1.24639 m .57119 1.25562 L p .58005 1.25836 L 0 .377 .817 r
F P
0 g
s
.57109 1.25208 m .56135 1.24639 L p .56939 1.25153 L 0 .377 .817 r
F P
0 g
s
.58433 1.2611 m .57109 1.25208 L p .56939 1.25153 L 0 .377 .817 r
F P
0 g
s
.57731 1.25751 m .57109 1.25208 L p .57904 1.25734 L .765 .28 .083 r
F P
0 g
s
.58881 1.26381 m .57731 1.25751 L .57904 1.25734 L .765 .28 .083 r
F
.57731 1.25751 m .57968 1.26263 L p .57555 1.25768 L .827 .54 .502 r
F P
0 g
s
.57109 1.25208 m .57731 1.25751 L p .57286 1.25264 L .839 .372 .051 r
F P
0 g
s
.56012 1.23874 m .57109 1.25208 L p .57286 1.25264 L .839 .372 .051 r
F P
0 g
s
.5532 1.23264 m .56135 1.24639 L .57109 1.25208 L p .195 .413 .847 r
F P
0 g
s
.56012 1.23874 m .5532 1.23264 L p .55882 1.23875 L .195 .413 .847 r
F P
0 g
s
.57109 1.25208 m .56012 1.23874 L p .55882 1.23875 L .195 .413 .847 r
F P
0 g
s
.56463 1.24457 m .56012 1.23874 L p .56585 1.24499 L .839 .372 .051 r
F P
0 g
s
.57731 1.25751 m .56463 1.24457 L p .56585 1.24499 L .839 .372 .051 r
F P
0 g
s
.56463 1.24457 m .56645 1.25009 L p .56344 1.24415 L .861 .605 .55 r
F P
0 g
s
.56012 1.23874 m .56463 1.24457 L p .56141 1.23873 L .839 .423 0 r
F P
0 g
s
.56141 1.23873 m .55372 1.22478 L .56012 1.23874 L .839 .423 0 r
F
.55219 1.22201 m .55361 1.225 L p .55372 1.22478 L F P
0 g
s
.55361 1.225 m .56012 1.23874 L p .55372 1.22478 L .839 .423 0 r
F P
0 g
s
.54954 1.20872 m .55042 1.21011 L .54951 1.20861 L .089 .173 .658 r
F
.55413 1.22398 m .55044 1.21016 L .54954 1.20872 L F
.55044 1.21016 m .55042 1.21011 L .54954 1.20872 L F
.55042 1.19894 m .55108 1.19852 L .55026 1.19651 L .339 .278 .642 r
F
.55128 1.21151 m .55109 1.19961 L .55042 1.19894 L F
.55109 1.19961 m .55108 1.19852 L .55042 1.19894 L F
.55291 1.19399 m .55483 1.18866 L .55419 1.18656 L .414 .287 .598 r
F
.55181 1.20033 m .55377 1.19277 L .55291 1.19399 L F
.55377 1.19277 m .55483 1.18866 L .55291 1.19399 L F
.55538 1.19047 m .55994 1.18165 L .56029 1.17938 L .143 0 .355 r
F
.55624 1.18926 m .55994 1.18165 L p .55538 1.19047 L F P
0 g
s
.55207 1.20059 m .55483 1.18866 L p .55181 1.20033 L .414 .287 .598 r
F P
0 g
s
.55149 1.21185 m .55108 1.19852 L p .55128 1.21151 L .339 .278 .642 r
F P
0 g
s
.54988 1.19096 m .55042 1.21011 L p .55128 1.21151 L .339 .278 .642 r
F P
0 g
s
.55042 1.21011 m .54783 1.20302 L p .54951 1.20861 L .089 .173 .658 r
F P
0 g
s
.55219 1.22201 m .54783 1.20302 L .54405 1.19564 L p .506 .451 .729 r
F P
0 g
s
.54736 1.18309 m .54783 1.20302 L .55042 1.21011 L p .558 .453 .691 r
F P
0 g
s
.55042 1.21011 m .54988 1.19096 L .54736 1.18309 L p .558 .453 .691 r
F P
0 g
s
.55546 1.2284 m .55219 1.22201 L p .55559 1.22818 L .839 .423 0 r
F P
0 g
s
.56463 1.24457 m .55546 1.2284 L p .55559 1.22818 L .839 .423 0 r
F P
0 g
s
.54729 1.21533 m .5532 1.23264 L .56012 1.23874 L p .334 .4 .781 r
F P
0 g
s
.56012 1.23874 m .55219 1.22201 L .54729 1.21533 L p .334 .4 .781 r
F P
0 g
s
.55546 1.2284 m .55689 1.23449 L p .55533 1.2286 L .96 .691 .483 r
F P
0 g
s
.54783 1.20302 m .55219 1.22201 L .55546 1.2284 L p .089 .173 .658 r
F P
0 g
s
.55546 1.2284 m .5542 1.22384 L p .55413 1.22398 L .089 .173 .658 r
F P
0 g
s
.5542 1.22384 m .55042 1.21011 L p .55413 1.22398 L .089 .173 .658 r
F P
0 g
s
.55533 1.2286 m .55198 1.21596 L .55546 1.2284 L .96 .691 .483 r
F
.55198 1.21596 m .55044 1.21018 L .55215 1.21639 L F
.55198 1.21596 m .55215 1.21639 L .55546 1.2284 L F
.55042 1.21011 m .55044 1.21016 L p .55044 1.21018 L F P
0 g
s
.55044 1.21016 m .55546 1.2284 L p .55044 1.21018 L .96 .691 .483 r
F P
0 g
s
.55165 1.2169 m .55042 1.21011 L p .55198 1.21596 L .96 .691 .483 r
F P
0 g
s
.55689 1.23449 m .55165 1.2169 L p .55198 1.21596 L .96 .691 .483 r
F P
0 g
s
.54405 1.19564 m .54729 1.21533 L .55219 1.22201 L p .506 .451 .729 r
F P
0 g
s
.55165 1.2169 m .55142 1.22337 L p .55134 1.21776 L 1 .799 .482 r
F P
0 g
s
.55134 1.21776 m .55129 1.2143 L .55165 1.2169 L 1 .799 .482 r
F
.55042 1.21011 m .55165 1.2169 L p .55128 1.21151 L .339 .278 .642 r
F P
0 g
s
.55149 1.21459 m .55149 1.21185 L .55128 1.21151 L .339 .278 .642 r
F
.55159 1.21501 m .55149 1.21185 L p .55149 1.21459 L F P
0 g
s
.55129 1.2143 m .55119 1.20659 L .55159 1.21501 L 1 .799 .482 r
F
.55119 1.20659 m .55116 1.20439 L .55124 1.20655 L F
.55119 1.20659 m .55124 1.20655 L .55159 1.21501 L F
.55124 1.20396 m .55159 1.21501 L p .55116 1.20439 L F P
0 g
s
.55088 1.20577 m .55108 1.19852 L p .55116 1.20439 L 1 .799 .482 r
F P
0 g
s
.55142 1.22337 m .55088 1.20577 L p .55116 1.20439 L 1 .799 .482 r
F P
0 g
s
.54113 1.20839 m .54433 1.2263 L .5532 1.23264 L p .545 .508 .766 r
F P
0 g
s
.5532 1.23264 m .54729 1.21533 L .54113 1.20839 L p .545 .508 .766 r
F P
0 g
s
.54979 1.22951 m .54932 1.2127 L .55088 1.20577 L p .833 .67 .681 r
F P
0 g
s
.55088 1.20577 m .54932 1.2127 L p .55062 1.20701 L .983 .895 .558 r
F P
0 g
s
.55088 1.20577 m .55142 1.22337 L .54979 1.22951 L p .833 .67 .681 r
F P
0 g
s
.55062 1.20701 m .55386 1.19289 L .55088 1.20577 L .983 .895 .558 r
F
.55394 1.19253 m .55089 1.20573 L p .55386 1.19289 L F P
0 g
s
.55089 1.20573 m .55088 1.20577 L p .55089 1.20573 L .983 .895 .558 r
F P
0 g
s
.55386 1.19289 m .55089 1.20573 L .55089 1.20573 L .983 .895 .558 r
F
.55483 1.18866 m .55394 1.19253 L p .55386 1.19289 L F P
0 g
s
.55108 1.19852 m .55124 1.20396 L p .55116 1.20439 L 1 .799 .482 r
F P
0 g
s
.55108 1.19852 m .55088 1.20577 L p .55181 1.20033 L .414 .287 .598 r
F P
0 g
s
.55522 1.18061 m .55108 1.19852 L p .55181 1.20033 L .414 .287 .598 r
F P
0 g
s
.53937 1.18799 m .54113 1.20839 L .54729 1.21533 L p .597 .509 .726 r
F P
0 g
s
.54729 1.21533 m .54405 1.19564 L .53937 1.18799 L p .597 .509 .726 r
F P
0 g
s
.54656 1.21928 m .54928 1.20374 L .55278 1.19637 L p .826 .694 .718 r
F P
0 g
s
.55278 1.19637 m .54928 1.20374 L p .55226 1.19758 L .901 .826 .764 r
F P
0 g
s
.55064 1.20692 m .55278 1.19637 L .55338 1.19499 L .983 .895 .558 r
F
.55088 1.20577 m .55207 1.20059 L p .55181 1.20033 L .414 .287 .598 r
F P
0 g
s
.55053 1.20699 m .55278 1.19637 L p .55064 1.20692 L .983 .895 .558 r
F P
0 g
s
.54932 1.2127 m .54656 1.21928 L p .5517 1.20034 L .826 .694 .718 r
F P
0 g
s
.55278 1.19637 m .55201 1.19998 L p .5517 1.20034 L .826 .694 .718 r
F P
0 g
s
.55201 1.19998 m .54932 1.2127 L p .5517 1.20034 L .826 .694 .718 r
F P
0 g
s
.55994 1.18165 m .55278 1.19637 L p .55226 1.19758 L .901 .826 .764 r
F P
0 g
s
.55278 1.19637 m .55483 1.18866 L p .55338 1.19499 L .983 .895 .558 r
F P
0 g
s
.55345 1.19485 m .55624 1.18926 L .55538 1.19047 L .143 0 .355 r
F
.55278 1.19637 m .5534 1.1951 L p .55345 1.19485 L F P
0 g
s
.5534 1.1951 m .55624 1.18926 L p .55345 1.19485 L .143 0 .355 r
F P
0 g
s
.55483 1.18866 m .55376 1.1927 L p .55508 1.18948 L .143 0 .355 r
F P
0 g
s
.55376 1.1927 m .55278 1.19637 L p .55508 1.18948 L .143 0 .355 r
F P
0 g
s
.55278 1.19637 m .55538 1.19047 L .55508 1.18948 L .143 0 .355 r
F
.56302 1.17319 m .55483 1.18866 L p .55538 1.19047 L F P
0 g
s
.55108 1.19852 m .54988 1.19096 L p .55026 1.19651 L .339 .278 .642 r
F P
0 g
s
.55385 1.17226 m .54988 1.19096 L .55108 1.19852 L p .582 .446 .661 r
F P
0 g
s
.55108 1.19852 m .55522 1.18061 L .55385 1.17226 L p .582 .446 .661 r
F P
0 g
s
.55492 1.18975 m .54846 1.19748 L p .55334 1.19107 L .809 .737 .784 r
F P
0 g
s
.55492 1.18975 m .55994 1.18165 L p .55679 1.18818 L .901 .826 .764 r
F P
0 g
s
.55994 1.18165 m .55492 1.18975 L p .55966 1.18352 L .883 .992 .771 r
F P
0 g
s
.57032 1.16952 m .55994 1.18165 L p .55966 1.18352 L .883 .992 .771 r
F P
0 g
s
.56309 1.17824 m .55492 1.18975 L p .55334 1.19107 L .809 .737 .784 r
F P
0 g
s
.54995 1.20236 m .55492 1.18975 L .55679 1.18818 L .901 .826 .764 r
F
.54928 1.20374 m .54974 1.20259 L p .54995 1.20236 L F P
0 g
s
.54974 1.20259 m .55492 1.18975 L p .54995 1.20236 L .901 .826 .764 r
F P
0 g
s
.54932 1.2127 m .55053 1.20699 L p .55064 1.20692 L .983 .895 .558 r
F P
0 g
s
.55165 1.2169 m .55159 1.21501 L p .55149 1.21459 L .339 .278 .642 r
F P
0 g
s
.55159 1.21501 m .55165 1.2169 L p .55129 1.2143 L 1 .799 .482 r
F P
0 g
s
.55994 1.18165 m .56302 1.17319 L p .56029 1.17938 L .143 0 .355 r
F P
0 g
s
.55483 1.18866 m .55522 1.18061 L p .55419 1.18656 L .414 .287 .598 r
F P
0 g
s
.55546 1.2284 m .56463 1.24457 L p .56344 1.24415 L .861 .605 .55 r
F P
0 g
s
.56463 1.24457 m .57731 1.25751 L p .57555 1.25768 L .827 .54 .502 r
F P
0 g
s
.59263 1.26634 m .58831 1.26385 L p .58881 1.26381 L .765 .28 .083 r
F P
0 g
s
.58831 1.26385 m .57731 1.25751 L p .58881 1.26381 L .765 .28 .083 r
F P
0 g
s
.57119 1.25562 m .55407 1.24996 L p .56328 1.25318 L 0 .462 .794 r
F P
0 g
s
.63258 .85286 m .60293 .8716 L .60238 .88773 L p .522 .255 .452 r
F P
0 g
s
.60238 .88773 m .63179 .86941 L .63258 .85286 L p .522 .255 .452 r
F P
0 g
s
.43465 .14196 m .4725 .19053 L .46638 .21909 L p .539 .567 .831 r
F P
0 g
s
.46638 .21909 m .42517 .17326 L .43465 .14196 L p .539 .567 .831 r
F P
0 g
s
.52449 1.10563 m .52006 1.08847 L .50643 1.07728 L p .788 .562 .593 r
F P
0 g
s
.52006 1.08847 m .52449 1.10563 L .53415 1.11605 L p .75 .55 .624 r
F P
0 g
s
.53415 1.11605 m .53416 1.09945 L .52006 1.08847 L p .75 .55 .624 r
F P
0 g
s
.51357 .82469 m .51236 .86726 L .51593 .88306 L p .863 .705 .683 r
F P
0 g
s
.45157 .67612 m .48139 .70843 L .4816 .72737 L p .454 .508 .824 r
F P
0 g
s
.50197 .76553 m .4816 .72737 L .48139 .70843 L p .427 .437 .768 r
F P
0 g
s
.54494 .91983 m .54556 .95768 L .55038 .97181 L p .556 .452 .692 r
F P
0 g
s
.55038 .97181 m .54952 .9347 L .54494 .91983 L p .556 .452 .692 r
F P
0 g
s
.55407 1.24996 m .54875 1.24048 L .53421 1.23439 L p .566 .614 .86 r
F P
0 g
s
.54875 1.24048 m .55407 1.24996 L .57119 1.25562 L p .431 .591 .909 r
F P
0 g
s
.42674 1.21998 m .40088 1.21901 L .41524 1.22447 L p .498 .764 .99 r
F P
0 g
s
.41524 1.22447 m .43818 1.22569 L .42674 1.21998 L p .498 .764 .99 r
F P
0 g
s
.53416 1.09945 m .53415 1.11605 L .54342 1.12624 L p .724 .536 .637 r
F P
0 g
s
.54342 1.12624 m .54755 1.11024 L .53416 1.09945 L p .724 .536 .637 r
F P
0 g
s
.62975 .46663 m .59428 .50461 L .5899 .52541 L p .119 0 .21 r
F P
0 g
s
.56866 1.10956 m .55919 1.1208 L .5682 1.13114 L p .69 .489 .611 r
F P
0 g
s
.5682 1.13114 m .58095 1.12039 L .56866 1.10956 L p .69 .489 .611 r
F P
0 g
s
.51522 1.09498 m .50643 1.07728 L .49445 1.06589 L p .856 .565 .495 r
F P
0 g
s
.50643 1.07728 m .51522 1.09498 L .52449 1.10563 L p .788 .562 .593 r
F P
0 g
s
.68202 1.20384 m .68145 1.18478 L .68861 1.1937 L p .228 .191 .606 r
F P
0 g
s
.68145 1.18478 m .67645 1.16657 L .68336 1.17614 L p .021 .131 .64 r
F P
0 g
s
.68336 1.17614 m .68861 1.1937 L .68145 1.18478 L p .021 .131 .64 r
F P
0 g
s
.67645 1.16657 m .66733 1.15046 L .67381 1.16057 L p 0 0 .508 r
F P
0 g
s
.67381 1.16057 m .68336 1.17614 L .67645 1.16657 L p 0 0 .508 r
F P
0 g
s
.40466 .9666 m .36192 .97702 L .36413 .98953 L p .678 .844 .963 r
F P
0 g
s
.36413 .98953 m .406 .97973 L .40466 .9666 L p .678 .844 .963 r
F P
0 g
s
.50514 .22045 m .51533 .283 L .51093 .30838 L p .575 .502 .736 r
F P
0 g
s
.51093 .30838 m .49843 .24721 L .50514 .22045 L p .575 .502 .736 r
F P
0 g
s
.51051 .85113 m .49807 .89271 L .50102 .90799 L p .803 .687 .736 r
F P
0 g
s
.50102 .90799 m .51236 .86726 L .51051 .85113 L p .803 .687 .736 r
F P
0 g
s
.6781 1.22248 m .68202 1.20384 L .6892 1.21205 L p .332 .207 .556 r
F P
0 g
s
.68861 1.1937 m .6892 1.21205 L .68202 1.20384 L p .228 .191 .606 r
F P
0 g
s
.66733 1.15046 m .65473 1.13754 L .66061 1.14805 L p .809 .321 0 r
F P
0 g
s
.66061 1.14805 m .67381 1.16057 L .66733 1.15046 L p .809 .321 0 r
F P
0 g
s
.67454 1.11244 m .70939 1.11109 L .70828 1.12394 L p .591 .195 .261 r
F P
0 g
s
.7416 1.114 m .70828 1.12394 L .70939 1.11109 L p .438 .03 .167 r
F P
0 g
s
.59529 1.11436 m .58095 1.12039 L .58872 1.13097 L p .682 .451 .565 r
F P
0 g
s
.58872 1.13097 m .60532 1.12527 L .59529 1.11436 L p .682 .451 .565 r
F P
0 g
s
.67384 .43905 m .62975 .46663 L .62285 .48725 L p .174 0 .094 r
F P
0 g
s
.5899 .52541 m .62285 .48725 L .62975 .46663 L p .119 0 .21 r
F P
0 g
s
.59893 1.26521 m .61389 1.26414 L .62687 1.2696 L p .651 .284 .337 r
F P
0 g
s
.64344 1.26382 m .62687 1.2696 L .61389 1.26414 L p .619 .32 .441 r
F P
0 g
s
.48086 1.20865 m .46056 1.2203 L .47465 1.22637 L p .713 .682 .818 r
F P
0 g
s
.47465 1.22637 m .49096 1.21529 L .48086 1.20865 L p .713 .682 .818 r
F P
0 g
s
.54755 1.11024 m .54342 1.12624 L .55154 1.13618 L p .703 .523 .641 r
F P
0 g
s
.55154 1.13618 m .55919 1.1208 L .54755 1.11024 L p .703 .523 .641 r
F P
0 g
s
.66996 1.23942 m .6781 1.22248 L .68509 1.22994 L p .398 .209 .505 r
F P
0 g
s
.6892 1.21205 m .68509 1.22994 L .6781 1.22248 L p .332 .207 .556 r
F P
0 g
s
.65473 1.13754 m .63949 1.12868 L .64465 1.13942 L p .784 .39 .309 r
F P
0 g
s
.64465 1.13942 m .66061 1.14805 L .65473 1.13754 L p .784 .39 .309 r
F P
0 g
s
.60293 .8716 m .579 .89755 L .57871 .91308 L p .497 .279 .514 r
F P
0 g
s
.57871 .91308 m .60238 .88773 L .60293 .8716 L p .497 .279 .514 r
F P
0 g
s
.6097 1.08786 m .58417 1.06359 L .58035 1.05107 L p .849 .53 .447 r
F P
0 g
s
.43293 1.02089 m .45768 1.02853 L .46565 1.0407 L p .502 .638 .915 r
F P
0 g
s
.48522 1.05427 m .46565 1.0407 L .45768 1.02853 L p .394 .538 .881 r
F P
0 g
s
.49445 1.06589 m .50715 1.0841 L p .50928 1.08667 L .856 .565 .495 r
F P
0 g
s
.501 1.07296 m .50715 1.0841 L p .50123 1.07605 L .795 .331 0 r
F P
0 g
s
.48522 1.05427 m .501 1.07296 L p .50123 1.07605 L .795 .331 0 r
F P
0 g
s
.501 1.07296 m .49739 1.06156 L p .50074 1.06939 L .114 .243 .72 r
F P
0 g
s
.50715 1.0841 m .501 1.07296 L p .50471 1.08117 L .955 .615 .32 r
F P
0 g
s
.51594 1.10601 m .50715 1.0841 L p .50471 1.08117 L .955 .615 .32 r
F P
0 g
s
.51594 1.10601 m .5213 1.11625 L p .51792 1.10856 L .844 .617 .594 r
F P
0 g
s
.50715 1.0841 m .51594 1.10601 L p .51792 1.10856 L .844 .617 .594 r
F P
0 g
s
.51192 1.0955 m .51594 1.10601 L p .51256 1.09853 L .955 .615 .32 r
F P
0 g
s
.501 1.07296 m .51192 1.0955 L p .51256 1.09853 L .955 .615 .32 r
F P
0 g
s
.50715 1.0841 m .51522 1.09498 L p .50928 1.08667 L .856 .565 .495 r
F P
0 g
s
.5213 1.11625 m .51522 1.09498 L .50715 1.0841 L p .844 .617 .594 r
F P
0 g
s
.51192 1.0955 m .50964 1.08472 L p .51122 1.09214 L 0 0 .159 r
F P
0 g
s
.51594 1.10601 m .51192 1.0955 L p .51377 1.10322 L .91 .691 .594 r
F P
0 g
s
.51722 1.12035 m .5202 1.13012 L p .51857 1.1233 L .91 .691 .594 r
F P
0 g
s
.51192 1.0955 m .51722 1.12035 L p .51857 1.1233 L .91 .691 .594 r
F P
0 g
s
.51722 1.12035 m .51192 1.0955 L p .51122 1.09214 L 0 0 .159 r
F P
0 g
s
.5202 1.13012 m .51594 1.10601 L p .51377 1.10322 L .91 .691 .594 r
F P
0 g
s
.51558 1.11029 m .51722 1.12035 L p .51581 1.11374 L 0 0 .159 r
F P
0 g
s
.50964 1.08472 m .51558 1.11029 L p .51581 1.11374 L 0 0 .159 r
F P
0 g
s
.51544 1.09994 m .51558 1.11029 L p .51449 1.10391 L .372 .348 .703 r
F P
0 g
s
.50934 1.07363 m .51544 1.09994 L p .51449 1.10391 L .372 .348 .703 r
F P
0 g
s
.52755 1.12623 m .52449 1.10563 L .51522 1.09498 L p .78 .588 .641 r
F P
0 g
s
.51522 1.09498 m .5213 1.11625 L .52755 1.12623 L p .78 .588 .641 r
F P
0 g
s
.51544 1.09994 m .51685 1.08927 L p .51648 1.09563 L .195 .144 .569 r
F P
0 g
s
.51558 1.11029 m .51544 1.09994 L p .51532 1.10655 L .978 .81 .4 r
F P
0 g
s
.50964 1.08472 m .50934 1.07363 L p .51556 1.11019 L .372 .348 .703 r
F P
0 g
s
.51558 1.11029 m .51556 1.11022 L p .51556 1.11019 L .372 .348 .703 r
F P
0 g
s
.51556 1.11022 m .50964 1.08472 L p .51556 1.11019 L .372 .348 .703 r
F P
0 g
s
.5149 1.12978 m .51539 1.10269 L .51462 1.12699 L .978 .81 .4 r
F
.51538 1.10205 m .51462 1.12699 L p .51539 1.10269 L F P
0 g
s
.5149 1.12978 m .51558 1.11029 L .51532 1.10655 L .978 .81 .4 r
F
.51478 1.13654 m .51497 1.13036 L p .5149 1.12978 L F P
0 g
s
.51497 1.13036 m .51558 1.11029 L p .5149 1.12978 L .978 .81 .4 r
F P
0 g
s
.51544 1.09994 m .51538 1.10205 L p .51539 1.10269 L .978 .81 .4 r
F P
0 g
s
.51722 1.12035 m .51558 1.11029 L p .51576 1.11717 L .886 .723 .674 r
F P
0 g
s
.51558 1.11029 m .51478 1.13654 L .51651 1.14579 L p .886 .723 .674 r
F P
0 g
s
.51462 1.12699 m .51544 1.09994 L p .51648 1.09563 L .195 .144 .569 r
F P
0 g
s
.51651 1.14579 m .51722 1.12035 L p .51576 1.11717 L .886 .723 .674 r
F P
0 g
s
.52426 1.13961 m .5213 1.11625 L .51594 1.10601 L p .81 .63 .659 r
F P
0 g
s
.51594 1.10601 m .5202 1.13012 L .52426 1.13961 L p .81 .63 .659 r
F P
0 g
s
.51608 1.11713 m .51462 1.12699 L p .51501 1.12044 L .195 .144 .569 r
F P
0 g
s
.51685 1.08927 m .51608 1.11713 L p .51501 1.12044 L .195 .144 .569 r
F P
0 g
s
.51608 1.11713 m .51906 1.10693 L p .51721 1.11362 L 0 0 0 r
F P
0 g
s
.51462 1.12699 m .51608 1.11713 L p .51432 1.124 L .953 .852 .704 r
F P
0 g
s
.50693 1.15296 m .51462 1.12699 L p .51432 1.124 L .953 .852 .704 r
F P
0 g
s
.50882 1.14395 m .50693 1.15296 L p .50889 1.14527 L .953 .852 .704 r
F P
0 g
s
.51608 1.11713 m .50882 1.14395 L p .50889 1.14527 L .953 .852 .704 r
F P
0 g
s
.50882 1.14395 m .51608 1.11713 L p .51721 1.11362 L 0 0 0 r
F P
0 g
s
.51462 1.12699 m .50693 1.15296 L .5073 1.16165 L p .844 .723 .731 r
F P
0 g
s
.51288 1.1346 m .50882 1.14395 L p .51091 1.13638 L 0 0 0 r
F P
0 g
s
.51906 1.10693 m .51288 1.1346 L p .51091 1.13638 L 0 0 0 r
F P
0 g
s
.51462 1.12699 m .51478 1.13654 L p .5149 1.12978 L .978 .81 .4 r
F P
0 g
s
.51963 1.15475 m .5202 1.13012 L .51722 1.12035 L p .802 .652 .696 r
F P
0 g
s
.51722 1.12035 m .51651 1.14579 L .51963 1.15475 L p .802 .652 .696 r
F P
0 g
s
.5073 1.16165 m .51478 1.13654 L .51462 1.12699 L p .844 .723 .731 r
F P
0 g
s
.51288 1.1346 m .51886 1.12489 L p .51492 1.13276 L .042 0 .253 r
F P
0 g
s
.50882 1.14395 m .51288 1.1346 L p .50874 1.14256 L .925 .969 .68 r
F P
0 g
s
.49554 1.16791 m .50882 1.14395 L p .50874 1.14256 L .925 .969 .68 r
F P
0 g
s
.49287 1.17607 m .50693 1.15296 L .50882 1.14395 L p .879 .812 .779 r
F P
0 g
s
.49554 1.16791 m .49287 1.17607 L p .49694 1.16787 L .879 .812 .779 r
F P
0 g
s
.50882 1.14395 m .49554 1.16791 L p .49694 1.16787 L .879 .812 .779 r
F P
0 g
s
.50158 1.1594 m .49554 1.16791 L p .49978 1.15977 L .925 .969 .68 r
F P
0 g
s
.51288 1.1346 m .50158 1.1594 L p .49978 1.15977 L .925 .969 .68 r
F P
0 g
s
.51478 1.13654 m .5073 1.16165 L .50984 1.17007 L p .784 .661 .727 r
F P
0 g
s
.49365 1.18392 m .5073 1.16165 L .50693 1.15296 L p .806 .714 .763 r
F P
0 g
s
.50693 1.15296 m .49287 1.17607 L .49365 1.18392 L p .806 .714 .763 r
F P
0 g
s
.50158 1.1594 m .51063 1.15051 L p .50339 1.15904 L 0 0 .042 r
F P
0 g
s
.49554 1.16791 m .50158 1.1594 L p .49411 1.16796 L .845 .999 .792 r
F P
0 g
s
.47713 1.1874 m .49554 1.16791 L p .49411 1.16796 L .845 .999 .792 r
F P
0 g
s
.4734 1.19475 m .49287 1.17607 L .49554 1.16791 L p .852 .848 .839 r
F P
0 g
s
.47713 1.1874 m .4734 1.19475 L p .47924 1.18768 L .852 .848 .839 r
F P
0 g
s
.49554 1.16791 m .47713 1.1874 L p .47924 1.18768 L .852 .848 .839 r
F P
0 g
s
.48589 1.17968 m .47713 1.1874 L p .4833 1.18033 L .845 .999 .792 r
F P
0 g
s
.50158 1.1594 m .48589 1.17968 L p .4833 1.18033 L .845 .999 .792 r
F P
0 g
s
.48589 1.17968 m .49918 1.17154 L p .48839 1.17905 L 0 0 0 r
F P
0 g
s
.47713 1.1874 m .48589 1.17968 L p .47508 1.18712 L .724 .987 .884 r
F P
0 g
s
.45482 1.20107 m .47713 1.1874 L p .47508 1.18712 L .724 .987 .884 r
F P
0 g
s
.45482 1.20107 m .44984 1.20773 L p .45549 1.20352 L .797 .882 .908 r
F P
0 g
s
.47713 1.1874 m .45482 1.20107 L p .45549 1.20352 L .797 .882 .908 r
F P
0 g
s
.46685 1.19404 m .45482 1.20107 L p .46089 1.19689 L .724 .987 .884 r
F P
0 g
s
.48589 1.17968 m .46685 1.19404 L p .46089 1.19689 L .724 .987 .884 r
F P
0 g
s
.49287 1.17607 m .4734 1.19475 L .47477 1.20182 L p .795 .743 .806 r
F P
0 g
s
.44984 1.20773 m .4734 1.19475 L .47713 1.1874 L p .797 .882 .908 r
F P
0 g
s
.47477 1.20182 m .49365 1.18392 L .49287 1.17607 L p .795 .743 .806 r
F P
0 g
s
.49918 1.17154 m .48528 1.18653 L .46685 1.19404 L p 0 0 0 r
F P
0 g
s
.46685 1.19404 m .48528 1.18653 L p .47224 1.19146 L .375 .837 .78 r
F P
0 g
s
.46685 1.19404 m .48589 1.17968 L p .48839 1.17905 L 0 0 0 r
F P
0 g
s
.44573 1.20147 m .46685 1.19404 L p .47224 1.19146 L .375 .837 .78 r
F P
0 g
s
.45482 1.20107 m .46685 1.19404 L p .4542 1.19884 L .636 .932 .969 r
F P
0 g
s
.43011 1.20799 m .45482 1.20107 L p .4542 1.19884 L .636 .932 .969 r
F P
0 g
s
.48589 1.17968 m .50158 1.1594 L p .50339 1.15904 L 0 0 .042 r
F P
0 g
s
.50158 1.1594 m .51288 1.1346 L p .51492 1.13276 L .042 0 .253 r
F P
0 g
s
.51192 1.0955 m .501 1.07296 L p .50074 1.06939 L .114 .243 .72 r
F P
0 g
s
.48139 .70843 m .50197 .74742 L .50197 .76553 L p .427 .437 .768 r
F P
0 g
s
.55961 .65284 m .5746 .70257 L .57167 .72119 L p .812 .613 .634 r
F P
0 g
s
.57167 .72119 m .55782 .6722 L .55961 .65284 L p .812 .613 .634 r
F P
0 g
s
.51338 1.20602 m .50403 1.22179 L .51888 1.22816 L p .673 .604 .769 r
F P
0 g
s
.51888 1.22816 m .52358 1.21297 L .51338 1.20602 L p .673 .604 .769 r
F P
0 g
s
.52358 1.21297 m .51888 1.22816 L .53421 1.23439 L p .643 .587 .776 r
F P
0 g
s
.50925 .37019 m .49346 .42878 L p .49272 .43856 L .964 .861 .689 r
F P
0 g
s
.51396 .34561 m .5011 .40453 L .49346 .42878 L p .951 .891 .496 r
F P
0 g
s
.49346 .42878 m .5011 .40453 L p .49422 .41877 L .954 .951 .699 r
F P
0 g
s
.46447 .48041 m .49346 .42878 L p .49422 .41877 L .954 .951 .699 r
F P
0 g
s
.49346 .42878 m .48931 .45271 L p .49272 .43856 L .964 .861 .689 r
F P
0 g
s
.45761 .5043 m .48931 .45271 L .49346 .42878 L p .917 .866 .774 r
F P
0 g
s
.56021 .91385 m .55278 .94931 L .55447 .96365 L p .527 .391 .638 r
F P
0 g
s
.55447 .96365 m .56247 .92895 L .56021 .91385 L p .527 .391 .638 r
F P
0 g
s
.44324 .94538 m .40466 .9666 L .406 .97973 L p .773 .835 .899 r
F P
0 g
s
.406 .97973 m .44394 .95929 L .44324 .94538 L p .773 .835 .899 r
F P
0 g
s
.53421 1.23439 m .53419 1.21974 L .52358 1.21297 L p .643 .587 .776 r
F P
0 g
s
.53419 1.21974 m .53421 1.23439 L .54875 1.24048 L p .6 .569 .792 r
F P
0 g
s
.72359 .42373 m .67384 .43905 L .664 .45909 L p .269 0 .004 r
F P
0 g
s
.62285 .48725 m .664 .45909 L .67384 .43905 L p .174 0 .094 r
F P
0 g
s
.65814 1.25352 m .66996 1.23942 L .67655 1.24617 L p .453 .21 .454 r
F P
0 g
s
.68509 1.22994 m .67655 1.24617 L .66996 1.23942 L p .398 .209 .505 r
F P
0 g
s
.50448 1.19887 m .49096 1.21529 L .50403 1.22179 L p .699 .623 .768 r
F P
0 g
s
.50403 1.22179 m .51338 1.20602 L .50448 1.19887 L p .699 .623 .768 r
F P
0 g
s
.63949 1.12868 m .62264 1.1245 L .62703 1.13528 L p .727 .409 .442 r
F P
0 g
s
.62703 1.13528 m .64465 1.13942 L .63949 1.12868 L p .727 .409 .442 r
F P
0 g
s
.57119 1.25562 m .56135 1.24639 L .54875 1.24048 L p .431 .591 .909 r
F P
0 g
s
.45194 1.21411 m .42674 1.21998 L .43818 1.22569 L p .687 .76 .907 r
F P
0 g
s
.43818 1.22569 m .46056 1.2203 L .45194 1.21411 L p .687 .76 .907 r
F P
0 g
s
.32013 1.06851 m .33428 1.0454 L .34044 1.05606 L p .926 .932 .79 r
F P
0 g
s
.34044 1.05606 m .32684 1.07826 L .32013 1.06851 L p .926 .932 .79 r
F P
0 g
s
.35925 1.03821 m .34044 1.05606 L .33428 1.0454 L p .855 .916 .873 r
F P
0 g
s
.31234 1.09452 m .32013 1.06851 L .32684 1.07826 L p .979 .906 .662 r
F P
0 g
s
.32684 1.07826 m .31937 1.10328 L .31247 1.09468 L p .979 .906 .662 r
F P
0 g
s
.32684 1.07826 m .31247 1.09468 L .31262 1.09421 L .979 .906 .662 r
F
.32394 1.08545 m .33138 1.05843 L .32033 1.06832 L p 0 0 .332 r
F P
0 g
s
.32033 1.06832 m .32013 1.06851 L p .32019 1.06879 L 0 0 .332 r
F P
0 g
s
.32019 1.06879 m .32394 1.08545 L .32033 1.06832 L 0 0 .332 r
F
.31247 1.09468 m .31234 1.09452 L p .31262 1.09421 L .979 .906 .662 r
F P
0 g
s
.64053 1.10452 m .67454 1.11244 L .67351 1.12484 L p .699 .339 .361 r
F P
0 g
s
.70828 1.12394 m .67351 1.12484 L .67454 1.11244 L p .591 .195 .261 r
F P
0 g
s
.33428 1.0454 m .35383 1.02677 L .35925 1.03821 L p .855 .916 .873 r
F P
0 g
s
.38199 1.02591 m .35925 1.03821 L .35383 1.02677 L p .777 .869 .916 r
F P
0 g
s
.31147 1.12165 m .31234 1.09452 L .31937 1.10328 L p .998 .825 .483 r
F P
0 g
s
.31937 1.10328 m .31853 1.12945 L .31148 1.12166 L p .998 .825 .483 r
F P
0 g
s
.31937 1.10328 m .31148 1.12166 L .31148 1.12163 L .998 .825 .483 r
F
.3231 1.11359 m .32394 1.08545 L .31236 1.0945 L p .121 .082 .536 r
F P
0 g
s
.31236 1.0945 m .31234 1.09452 L p .31236 1.09454 L .121 .082 .536 r
F P
0 g
s
.31236 1.09454 m .3231 1.11359 L .31236 1.0945 L .121 .082 .536 r
F
.31148 1.12166 m .31147 1.12165 L p .31148 1.12163 L .998 .825 .483 r
F P
0 g
s
.48394 .16171 m .50514 .22045 L .49843 .24721 L p .609 .567 .782 r
F P
0 g
s
.49843 .24721 m .4725 .19053 L .48394 .16171 L p .609 .567 .782 r
F P
0 g
s
.55919 1.1208 m .55154 1.13618 L .55791 1.14587 L p .684 .507 .641 r
F P
0 g
s
.55791 1.14587 m .5682 1.13114 L .55919 1.1208 L p .684 .507 .641 r
F P
0 g
s
.58095 1.12039 m .5682 1.13114 L .57397 1.1412 L p .671 .472 .608 r
F P
0 g
s
.57397 1.1412 m .58872 1.13097 L .58095 1.12039 L p .671 .472 .608 r
F P
0 g
s
.47507 .91485 m .44324 .94538 L .44394 .95929 L p .815 .793 .832 r
F P
0 g
s
.44394 .95929 m .47535 .92961 L .47507 .91485 L p .815 .793 .832 r
F P
0 g
s
.52449 1.10563 m .52755 1.12623 L .53415 1.13596 L p .744 .568 .656 r
F P
0 g
s
.53415 1.13596 m .53415 1.11605 L .52449 1.10563 L p .744 .568 .656 r
F P
0 g
s
.54875 1.24048 m .54433 1.2263 L .53419 1.21974 L p .6 .569 .792 r
F P
0 g
s
.54433 1.2263 m .54875 1.24048 L .56135 1.24639 L p .512 .54 .822 r
F P
0 g
s
.579 .89755 m .56247 .92895 L .56237 .94377 L p .468 .295 .562 r
F P
0 g
s
.56237 .94377 m .57871 .91308 L .579 .89755 L p .468 .295 .562 r
F P
0 g
s
.35383 1.02677 m .37744 1.01388 L .38199 1.02591 L p .777 .869 .916 r
F P
0 g
s
.40712 1.02 m .38199 1.02591 L .37744 1.01388 L p .703 .803 .928 r
F P
0 g
s
.31757 1.14806 m .31147 1.12165 L .31853 1.12945 L p .96 .67 .24 r
F P
0 g
s
.31853 1.12945 m .32439 1.15498 L .31757 1.14806 L p .96 .67 .24 r
F P
0 g
s
.62687 1.2696 m .60953 1.27046 L .59893 1.26521 L p .651 .284 .337 r
F P
0 g
s
.5746 .70257 m .60197 .74676 L .59704 .76478 L p .781 .558 .597 r
F P
0 g
s
.59704 .76478 m .57167 .72119 L .5746 .70257 L p .781 .558 .597 r
F P
0 g
s
.53415 1.11605 m .53415 1.13596 L .54056 1.14543 L p .718 .551 .663 r
F P
0 g
s
.54056 1.14543 m .54342 1.12624 L .53415 1.11605 L p .718 .551 .663 r
F P
0 g
s
.64344 1.26382 m .65814 1.25352 L .66417 1.25961 L p .508 .217 .407 r
F P
0 g
s
.67655 1.24617 m .66417 1.25961 L .65814 1.25352 L p .453 .21 .454 r
F P
0 g
s
.58417 1.06359 m .5657 1.03337 L .56322 1.0204 L p .88 .611 .528 r
F P
0 g
s
.62264 1.1245 m .60532 1.12527 L .60893 1.1359 L p .688 .422 .514 r
F P
0 g
s
.60893 1.1359 m .62703 1.13528 L .62264 1.1245 L p .688 .422 .514 r
F P
0 g
s
.38715 .10474 m .43465 .14196 L .42517 .17326 L p .556 .633 .883 r
F P
0 g
s
.42517 .17326 m .37378 .13931 L .38715 .10474 L p .556 .633 .883 r
F P
0 g
s
.49807 .89271 m .51051 .85113 L .51058 .83467 L p .867 .752 .733 r
F P
0 g
s
.45768 1.02853 m .47966 1.04239 L .48522 1.05427 L p .394 .538 .881 r
F P
0 g
s
.501 1.07296 m .48522 1.05427 L .47966 1.04239 L p .285 .418 .821 r
F P
0 g
s
.49803 .87709 m .47507 .91485 L .47535 .92961 L p .83 .751 .777 r
F P
0 g
s
.47535 .92961 m .49807 .89271 L .49803 .87709 L p .83 .751 .777 r
F P
0 g
s
.49768 1.19151 m .48086 1.20865 L .49096 1.21529 L p .725 .649 .773 r
F P
0 g
s
.49096 1.21529 m .50448 1.19887 L .49768 1.19151 L p .725 .649 .773 r
F P
0 g
s
.37744 1.01388 m .40349 1.0076 L .40712 1.02 L p .703 .803 .928 r
F P
0 g
s
.43293 1.02089 m .40712 1.02 L .40349 1.0076 L p .638 .729 .914 r
F P
0 g
s
.51058 .83467 m .49803 .87709 L .49807 .89271 L p .867 .752 .733 r
F P
0 g
s
.77565 .42175 m .72359 .42373 L .71068 .44289 L p .399 0 0 r
F P
0 g
s
.664 .45909 m .71068 .44289 L .72359 .42373 L p .269 0 .004 r
F P
0 g
s
.33023 1.17192 m .31757 1.14806 L .32439 1.15498 L p .825 .418 0 r
F P
0 g
s
.32439 1.15498 m .33655 1.17814 L .33023 1.17192 L p .825 .418 0 r
F P
0 g
s
.6097 1.08786 m .64053 1.10452 L .63968 1.11664 L p .765 .453 .458 r
F P
0 g
s
.67351 1.12484 m .63968 1.11664 L .64053 1.10452 L p .699 .339 .361 r
F P
0 g
s
.54342 1.12624 m .54056 1.14543 L .54625 1.15465 L p .697 .535 .664 r
F P
0 g
s
.54625 1.15465 m .55154 1.13618 L .54342 1.12624 L p .697 .535 .664 r
F P
0 g
s
.47477 1.20182 m .45194 1.21411 L .46056 1.2203 L p .739 .721 .834 r
F P
0 g
s
.46056 1.2203 m .48086 1.20865 L .47477 1.20182 L p .739 .721 .834 r
F P
0 g
s
.52015 1.18611 m .51338 1.20602 L .52358 1.21297 L p .68 .584 .741 r
F P
0 g
s
.52358 1.21297 m .52698 1.19377 L .52015 1.18611 L p .68 .584 .741 r
F P
0 g
s
.52698 1.19377 m .52358 1.21297 L .53419 1.21974 L p .652 .565 .744 r
F P
0 g
s
.40349 1.0076 m .4302 1.00835 L .43293 1.02089 L p .638 .729 .914 r
F P
0 g
s
.45768 1.02853 m .43293 1.02089 L .4302 1.00835 L p .588 .656 .884 r
F P
0 g
s
.62687 1.2696 m .64344 1.26382 L .64878 1.26936 L p .57 .239 .369 r
F P
0 g
s
.66417 1.25961 m .64878 1.26936 L .64344 1.26382 L p .508 .217 .407 r
F P
0 g
s
.48558 .68908 m .50463 .72894 L .50197 .74742 L p .55 .514 .769 r
F P
0 g
s
.50197 .74742 m .48139 .70843 L .48558 .68908 L p .55 .514 .769 r
F P
0 g
s
.51192 .79046 m .50197 .74742 L .50463 .72894 L p .502 .443 .723 r
F P
0 g
s
.60532 1.12527 m .58872 1.13097 L .59159 1.14126 L p .664 .434 .562 r
F P
0 g
s
.59159 1.14126 m .60893 1.1359 L .60532 1.12527 L p .664 .434 .562 r
F P
0 g
s
.34858 1.19161 m .33023 1.17192 L .33655 1.17814 L p .546 .055 0 r
F P
0 g
s
.33655 1.17814 m .3542 1.19734 L .34858 1.19161 L p .546 .055 0 r
F P
0 g
s
.53419 1.21974 m .53417 1.2012 L .52698 1.19377 L p .652 .565 .744 r
F P
0 g
s
.53417 1.2012 m .53419 1.21974 L .54433 1.2263 L p .614 .543 .751 r
F P
0 g
s
.50463 .72894 m .51383 .77283 L .51192 .79046 L p .502 .443 .723 r
F P
0 g
s
.56135 1.24639 m .5532 1.23264 L .54433 1.2263 L p .512 .54 .822 r
F P
0 g
s
.51426 1.17821 m .50448 1.19887 L .51338 1.20602 L p .704 .604 .741 r
F P
0 g
s
.51338 1.20602 m .52015 1.18611 L .51426 1.17821 L p .704 .604 .741 r
F P
0 g
s
.5682 1.13114 m .55791 1.14587 L .56207 1.15528 L p .663 .487 .637 r
F P
0 g
s
.56207 1.15528 m .57397 1.1412 L .5682 1.13114 L p .663 .487 .637 r
F P
0 g
s
.55782 .6722 m .57167 .72119 L .56571 .73962 L p .755 .573 .65 r
F P
0 g
s
.56571 .73962 m .55409 .69133 L .55782 .6722 L p .755 .573 .65 r
F P
0 g
s
.60197 .74676 m .63989 .78242 L .6323 .80004 L p .769 .496 .522 r
F P
0 g
s
.6323 .80004 m .59704 .76478 L .60197 .74676 L p .769 .496 .522 r
F P
0 g
s
.45789 .65587 m .48558 .68908 L .48139 .70843 L p .562 .565 .814 r
F P
0 g
s
.48139 .70843 m .45157 .67612 L .45789 .65587 L p .562 .565 .814 r
F P
0 g
s
.58417 1.06359 m .6097 1.08786 L .60909 1.09992 L p .797 .535 .541 r
F P
0 g
s
.63968 1.11664 m .60909 1.09992 L .6097 1.08786 L p .765 .453 .458 r
F P
0 g
s
.55154 1.13618 m .54625 1.15465 L .5508 1.16359 L p .677 .518 .663 r
F P
0 g
s
.5508 1.16359 m .55791 1.14587 L .55154 1.13618 L p .677 .518 .663 r
F P
0 g
s
.52904 1.14884 m .52755 1.12623 L .5213 1.11625 L p .761 .597 .675 r
F P
0 g
s
.52755 1.12623 m .52904 1.14884 L .53415 1.1578 L p .73 .575 .681 r
F P
0 g
s
.53415 1.1578 m .53415 1.13596 L .52755 1.12623 L p .73 .575 .681 r
F P
0 g
s
.54433 1.2263 m .54113 1.20839 L .53417 1.2012 L p .614 .543 .751 r
F P
0 g
s
.4302 1.00835 m .45577 1.01607 L .45768 1.02853 L p .588 .656 .884 r
F P
0 g
s
.47966 1.04239 m .45768 1.02853 L .45577 1.01607 L p .553 .589 .844 r
F P
0 g
s
.47966 1.04239 m .49739 1.06156 L .501 1.07296 L p .285 .418 .821 r
F P
0 g
s
.37136 1.20577 m .34858 1.19161 L .3542 1.19734 L p 0 .35 .792 r
F P
0 g
s
.3542 1.19734 m .37614 1.21129 L .37136 1.20577 L p 0 .35 .792 r
F P
0 g
s
.5213 1.11625 m .52426 1.13961 L .52904 1.14884 L p .761 .597 .675 r
F P
0 g
s
.53415 1.13596 m .53415 1.1578 L .53917 1.1665 L p .706 .557 .682 r
F P
0 g
s
.53917 1.1665 m .54056 1.14543 L .53415 1.13596 L p .706 .557 .682 r
F P
0 g
s
.50984 1.17007 m .49768 1.19151 L .50448 1.19887 L p .731 .628 .744 r
F P
0 g
s
.50448 1.19887 m .51426 1.17821 L .50984 1.17007 L p .731 .628 .744 r
F P
0 g
s
.82649 .43328 m .77565 .42175 L .75979 .43987 L p .546 .027 0 r
F P
0 g
s
.71068 .44289 m .75979 .43987 L .77565 .42175 L p .399 0 0 r
F P
0 g
s
.60953 1.27046 m .62687 1.2696 L .63144 1.27474 L p .642 .287 .354 r
F P
0 g
s
.64878 1.26936 m .63144 1.27474 L .62687 1.2696 L p .57 .239 .369 r
F P
0 g
s
.58872 1.13097 m .57397 1.1412 L .57619 1.15097 L p .648 .446 .597 r
F P
0 g
s
.57619 1.15097 m .59159 1.14126 L .58872 1.13097 L p .648 .446 .597 r
F P
0 g
s
.5657 1.03337 m .58417 1.06359 L .58381 1.07584 L p .813 .589 .599 r
F P
0 g
s
.60909 1.09992 m .58381 1.07584 L .58417 1.06359 L p .797 .535 .541 r
F P
0 g
s
.49365 1.18392 m .47477 1.20182 L .48086 1.20865 L p .755 .686 .785 r
F P
0 g
s
.48086 1.20865 m .49768 1.19151 L .49365 1.18392 L p .755 .686 .785 r
F P
0 g
s
.54056 1.14543 m .53917 1.1665 L .54369 1.17494 L p .685 .54 .682 r
F P
0 g
s
.54369 1.17494 m .54625 1.15465 L .54056 1.14543 L p .685 .54 .682 r
F P
0 g
s
.49346 .42878 m .46447 .48041 L .45761 .5043 L p .917 .866 .774 r
F P
0 g
s
.41354 .54577 m .45761 .5043 L .46447 .48041 L p .879 .913 .847 r
F P
0 g
s
.55557 .99927 m .5657 1.03337 L .56556 1.04605 L p .852 .645 .622 r
F P
0 g
s
.56556 1.04605 m .55556 1.01258 L .55557 .99927 L p .852 .645 .622 r
F P
0 g
s
.52386 1.16345 m .52015 1.18611 L .52698 1.19377 L p .694 .578 .721 r
F P
0 g
s
.52698 1.19377 m .52885 1.17189 L .52386 1.16345 L p .694 .578 .721 r
F P
0 g
s
.52885 1.17189 m .52698 1.19377 L .53417 1.2012 L p .669 .559 .721 r
F P
0 g
s
.58381 1.07584 m .56556 1.04605 L .5657 1.03337 L p .813 .589 .599 r
F P
0 g
s
.45577 1.01607 m .47845 1.03023 L .47966 1.04239 L p .553 .589 .844 r
F P
0 g
s
.49739 1.06156 m .47966 1.04239 L .47845 1.03023 L p .527 .53 .802 r
F P
0 g
s
.39702 1.21345 m .37136 1.20577 L .37614 1.21129 L p .275 .649 .967 r
F P
0 g
s
.37614 1.21129 m .40088 1.21901 L .39702 1.21345 L p .275 .649 .967 r
F P
0 g
s
.53417 1.2012 m .53416 1.18007 L .52885 1.17189 L p .669 .559 .721 r
F P
0 g
s
.53416 1.18007 m .53417 1.2012 L .54113 1.20839 L p .64 .538 .723 r
F P
0 g
s
.51963 1.15475 m .51426 1.17821 L .52015 1.18611 L p .718 .599 .722 r
F P
0 g
s
.52015 1.18611 m .52386 1.16345 L .51963 1.15475 L p .718 .599 .722 r
F P
0 g
s
.42335 .63156 m .45789 .65587 L .45157 .67612 L p .573 .625 .865 r
F P
0 g
s
.45157 .67612 m .41449 .65267 L .42335 .63156 L p .573 .625 .865 r
F P
0 g
s
.55283 .64087 m .55409 .69133 L .54877 .71023 L p .818 .656 .682 r
F P
0 g
s
.54877 .71023 m .54785 .6605 L .55283 .64087 L p .818 .656 .682 r
F P
0 g
s
.55791 1.14587 m .5508 1.16359 L .55385 1.17226 L p .654 .496 .658 r
F P
0 g
s
.55385 1.17226 m .56207 1.15528 L .55791 1.14587 L p .654 .496 .658 r
F P
0 g
s
.50979 .71012 m .51744 .75487 L .51383 .77283 L p .59 .507 .729 r
F P
0 g
s
.51383 .77283 m .50463 .72894 L .50979 .71012 L p .59 .507 .729 r
F P
0 g
s
.52885 1.17189 m .52904 1.14884 L .52426 1.13961 L p .726 .589 .701 r
F P
0 g
s
.52904 1.14884 m .52885 1.17189 L .53416 1.18007 L p .702 .569 .702 r
F P
0 g
s
.53416 1.18007 m .53415 1.1578 L .52904 1.14884 L p .702 .569 .702 r
F P
0 g
s
.52386 1.16345 m .52426 1.13961 L .5202 1.13012 L p .757 .614 .7 r
F P
0 g
s
.52426 1.13961 m .52386 1.16345 L .52885 1.17189 L p .726 .589 .701 r
F P
0 g
s
.49739 1.06156 m .50964 1.08472 L .51192 1.0955 L p .114 .243 .72 r
F P
0 g
s
.53415 1.1578 m .53416 1.18007 L .53937 1.18799 L p .679 .55 .701 r
F P
0 g
s
.53937 1.18799 m .53917 1.1665 L .53415 1.1578 L p .679 .55 .701 r
F P
0 g
s
.54625 1.15465 m .54369 1.17494 L .54736 1.18309 L p .663 .52 .68 r
F P
0 g
s
.54736 1.18309 m .5508 1.16359 L .54625 1.15465 L p .663 .52 .68 r
F P
0 g
s
.54113 1.20839 m .53937 1.18799 L .53416 1.18007 L p .64 .538 .723 r
F P
0 g
s
.59263 1.26634 m .60953 1.27046 L .61332 1.27539 L p .718 .363 .375 r
F P
0 g
s
.63144 1.27474 m .61332 1.27539 L .60953 1.27046 L p .642 .287 .354 r
F P
0 g
s
.57397 1.1412 m .56207 1.15528 L .56379 1.1644 L p .635 .457 .625 r
F P
0 g
s
.56379 1.1644 m .57619 1.15097 L .57397 1.1412 L p .635 .457 .625 r
F P
0 g
s
.51651 1.14579 m .50984 1.17007 L .51426 1.17821 L p .746 .624 .724 r
F P
0 g
s
.51426 1.17821 m .51963 1.15475 L .51651 1.14579 L p .746 .624 .724 r
F P
0 g
s
.57871 .91308 m .56237 .94377 L .55997 .9583 L p .169 0 .381 r
F P
0 g
s
.60238 .88773 m .57871 .91308 L .57479 .92835 L p .281 .035 .352 r
F P
0 g
s
.55997 .9583 m .57479 .92835 L .57871 .91308 L p .169 0 .381 r
F P
0 g
s
.5202 1.13012 m .51963 1.15475 L .52386 1.16345 L p .757 .614 .7 r
F P
0 g
s
.47845 1.03023 m .49672 1.04986 L .49739 1.06156 L p .527 .53 .802 r
F P
0 g
s
.50964 1.08472 m .49739 1.06156 L .49672 1.04986 L p .494 .471 .761 r
F P
0 g
s
.63989 .78242 m .68583 .80709 L .67522 .82455 L p .73 .394 .41 r
F P
0 g
s
.67522 .82455 m .6323 .80004 L .63989 .78242 L p .73 .394 .41 r
F P
0 g
s
.51259 .81787 m .50098 .8611 L p .50074 .86606 L .955 .851 .7 r
F P
0 g
s
.50678 .84475 m .50098 .8611 L p .50376 .85023 L .903 .867 .428 r
F P
0 g
s
.51641 .80072 m .50678 .84475 L p .50376 .85023 L .903 .867 .428 r
F P
0 g
s
.50678 .84475 m .51499 .82802 L p .50985 .83916 L 0 0 0 r
F P
0 g
s
.50098 .8611 m .50678 .84475 L p .50124 .856 L .95 .956 .706 r
F P
0 g
s
.47972 .89969 m .50098 .8611 L p .50124 .856 L .95 .956 .706 r
F P
0 g
s
.50098 .8611 m .48741 .88573 L p .47507 .91485 L .9 .841 .778 r
F P
0 g
s
.48741 .88573 m .47972 .89969 L p .48028 .9007 L .9 .841 .778 r
F P
0 g
s
.47784 .90583 m .47507 .91485 L p .48741 .88573 L .9 .841 .778 r
F P
0 g
s
.48909 .88412 m .47972 .89969 L p .48588 .88719 L .95 .956 .706 r
F P
0 g
s
.50678 .84475 m .48909 .88412 L p .48588 .88719 L .95 .956 .706 r
F P
0 g
s
.50098 .8611 m .49803 .87709 L p .50074 .86606 L .955 .851 .7 r
F P
0 g
s
.47507 .91485 m .49803 .87709 L .50098 .8611 L p .9 .841 .778 r
F P
0 g
s
.48909 .88412 m .50254 .86812 L p .49228 .88107 L 0 0 0 r
F P
0 g
s
.47972 .89969 m .48909 .88412 L p .4783 .89712 L .87 .995 .794 r
F P
0 g
s
.45015 .93104 m .47972 .89969 L p .4783 .89712 L .87 .995 .794 r
F P
0 g
s
.47972 .89969 m .4573 .92347 L p .44324 .94538 L .869 .89 .847 r
F P
0 g
s
.4573 .92347 m .45015 .93104 L p .45063 .93169 L .869 .89 .847 r
F P
0 g
s
.44869 .93408 m .44324 .94538 L p .4573 .92347 L .869 .89 .847 r
F P
0 g
s
.46442 .91622 m .45015 .93104 L p .45955 .91972 L .87 .995 .794 r
F P
0 g
s
.48909 .88412 m .46442 .91622 L p .45955 .91972 L .87 .995 .794 r
F P
0 g
s
.44324 .94538 m .47507 .91485 L .47972 .89969 L p .869 .89 .847 r
F P
0 g
s
.46442 .91622 m .48513 .90089 L p .46906 .91289 L 0 0 0 r
F P
0 g
s
.45015 .93104 m .46442 .91622 L p .44809 .92818 L .747 .991 .88 r
F P
0 g
s
.41418 .953 m .45015 .93104 L p .44809 .92818 L .747 .991 .88 r
F P
0 g
s
.43431 .93885 m .46442 .91622 L p .46906 .91289 L 0 0 0 r
F P
0 g
s
.45015 .93104 m .44869 .93408 L p .45063 .93169 L .869 .89 .847 r
F P
0 g
s
.46442 .91622 m .48677 .88713 L p .49228 .88107 L 0 0 0 r
F P
0 g
s
.48677 .88713 m .48909 .88412 L p .48966 .88357 L 0 0 0 r
F P
0 g
s
.48677 .88713 m .48966 .88357 L .49228 .88107 L 0 0 0 r
F
.47972 .89969 m .47784 .90583 L p .48028 .9007 L .9 .841 .778 r
F P
0 g
s
.48909 .88412 m .50411 .85068 L p .50985 .83916 L 0 0 0 r
F P
0 g
s
.50411 .85068 m .50678 .84475 L p .50797 .84257 L 0 0 0 r
F P
0 g
s
.50411 .85068 m .50797 .84257 L .50985 .83916 L 0 0 0 r
F
.42379 1.21411 m .39702 1.21345 L .40088 1.21901 L p .552 .783 .985 r
F P
0 g
s
.40088 1.21901 m .42674 1.21998 L .42379 1.21411 L p .552 .783 .985 r
F P
0 g
s
.53917 1.1665 m .53937 1.18799 L .54405 1.19564 L p .655 .53 .7 r
F P
0 g
s
.54405 1.19564 m .54369 1.17494 L .53917 1.1665 L p .655 .53 .7 r
F P
0 g
s
.5073 1.16165 m .49365 1.18392 L .49768 1.19151 L p .762 .662 .751 r
F P
0 g
s
.49768 1.19151 m .50984 1.17007 L .5073 1.16165 L p .762 .662 .751 r
F P
0 g
s
.68336 1.17614 m .67381 1.16057 L .66838 1.1703 L p .869 .596 .521 r
F P
0 g
s
.67381 1.16057 m .66061 1.14805 L .65572 1.15819 L p .815 .53 .507 r
F P
0 g
s
.65572 1.15819 m .66838 1.1703 L .67381 1.16057 L p .815 .53 .507 r
F P
0 g
s
.66061 1.14805 m .64465 1.13942 L .64042 1.14981 L p .763 .479 .504 r
F P
0 g
s
.64042 1.14981 m .65572 1.15819 L .66061 1.14805 L p .763 .479 .504 r
F P
0 g
s
.63179 .86941 m .60238 .88773 L .59629 .9036 L p .367 .062 .309 r
F P
0 g
s
.57479 .92835 m .59629 .9036 L .60238 .88773 L p .281 .035 .352 r
F P
0 g
s
.68861 1.1937 m .68336 1.17614 L .67755 1.18531 L p .924 .678 .547 r
F P
0 g
s
.66838 1.1703 m .67755 1.18531 L .68336 1.17614 L p .869 .596 .521 r
F P
0 g
s
.49395 .66935 m .50979 .71012 L .50463 .72894 L p .62 .561 .767 r
F P
0 g
s
.50463 .72894 m .48558 .68908 L .49395 .66935 L p .62 .561 .767 r
F P
0 g
s
.62285 .48725 m .5899 .52541 L .58101 .5462 L p 0 0 0 r
F P
0 g
s
.64465 1.13942 m .62703 1.13528 L .62353 1.14572 L p .716 .442 .51 r
F P
0 g
s
.62353 1.14572 m .64042 1.14981 L .64465 1.13942 L p .716 .442 .51 r
F P
0 g
s
.6892 1.21205 m .68861 1.1937 L .68259 1.20221 L p .971 .777 .586 r
F P
0 g
s
.67755 1.18531 m .68259 1.20221 L .68861 1.1937 L p .924 .678 .547 r
F P
0 g
s
.55409 .69133 m .56571 .73962 L .55726 .75784 L p .704 .534 .655 r
F P
0 g
s
.55726 .75784 m .54877 .71023 L .55409 .69133 L p .704 .534 .655 r
F P
0 g
s
.62703 1.13528 m .60893 1.1359 L .6062 1.1462 L p .674 .418 .525 r
F P
0 g
s
.6062 1.1462 m .62353 1.14572 L .62703 1.13528 L p .674 .418 .525 r
F P
0 g
s
.5508 1.16359 m .54736 1.18309 L .54988 1.19096 L p .634 .494 .674 r
F P
0 g
s
.54988 1.19096 m .55385 1.17226 L .5508 1.16359 L p .634 .494 .674 r
F P
0 g
s
.54369 1.17494 m .54405 1.19564 L .54783 1.20302 L p .621 .503 .698 r
F P
0 g
s
.54783 1.20302 m .54736 1.18309 L .54369 1.17494 L p .621 .503 .698 r
F P
0 g
s
.87264 .45758 m .82649 .43328 L .80803 .45039 L p .689 .181 .023 r
F P
0 g
s
.75979 .43987 m .80803 .45039 L .82649 .43328 L p .546 .027 0 r
F P
0 g
s
.49672 1.04986 m .50934 1.07363 L .50964 1.08472 L p .494 .471 .761 r
F P
0 g
s
.50984 1.17007 m .51651 1.14579 L .51478 1.13654 L p .784 .661 .727 r
F P
0 g
s
.68509 1.22994 m .6892 1.21205 L .68316 1.21983 L p .987 .89 .628 r
F P
0 g
s
.68259 1.20221 m .68316 1.21983 L .6892 1.21205 L p .971 .777 .586 r
F P
0 g
s
.44984 1.20773 m .42379 1.21411 L .42674 1.21998 L p .7 .807 .931 r
F P
0 g
s
.42674 1.21998 m .45194 1.21411 L .44984 1.20773 L p .7 .807 .931 r
F P
0 g
s
.57731 1.25751 m .59263 1.26634 L .59567 1.27126 L p .784 .456 .432 r
F P
0 g
s
.61332 1.27539 m .59567 1.27126 L .59263 1.26634 L p .718 .363 .375 r
F P
0 g
s
.56207 1.15528 m .55385 1.17226 L .55522 1.18061 L p .62 .462 .646 r
F P
0 g
s
.55522 1.18061 m .56379 1.1644 L .56207 1.15528 L p .62 .462 .646 r
F P
0 g
s
.55556 1.01258 m .56556 1.04605 L .56287 1.05842 L p .792 .608 .652 r
F P
0 g
s
.56287 1.05842 m .5538 1.02558 L .55556 1.01258 L p .792 .608 .652 r
F P
0 g
s
.66496 .85938 m .63179 .86941 L .62305 .88574 L p .465 .116 .288 r
F P
0 g
s
.59629 .9036 m .62305 .88574 L .63179 .86941 L p .367 .062 .309 r
F P
0 g
s
.57167 .72119 m .59704 .76478 L .58702 .78264 L p .729 .523 .613 r
F P
0 g
s
.58702 .78264 m .56571 .73962 L .57167 .72119 L p .729 .523 .613 r
F P
0 g
s
.60893 1.1359 m .59159 1.14126 L .5896 1.15121 L p .64 .404 .546 r
F P
0 g
s
.5896 1.15121 m .6062 1.1462 L .60893 1.1359 L p .64 .404 .546 r
F P
0 g
s
.33295 .08149 m .38715 .10474 L .37378 .13931 L p .582 .713 .932 r
F P
0 g
s
.37378 .13931 m .31559 .11957 L .33295 .08149 L p .582 .713 .932 r
F P
0 g
s
.51641 .80072 m .51744 .75487 L .52248 .73657 L p .523 .415 .673 r
F P
0 g
s
.5171 .69097 m .52248 .73657 L .51744 .75487 L p .649 .552 .731 r
F P
0 g
s
.51744 .75487 m .50979 .71012 L .5171 .69097 L p .649 .552 .731 r
F P
0 g
s
.67655 1.24617 m .68509 1.22994 L .67921 1.23697 L p .889 .961 .629 r
F P
0 g
s
.68316 1.21983 m .67921 1.23697 L .68509 1.22994 L p .987 .89 .628 r
F P
0 g
s
.3842 .61784 m .42335 .63156 L .41449 .65267 L p .592 .702 .919 r
F P
0 g
s
.41449 .65267 m .37262 .63968 L .3842 .61784 L p .592 .702 .919 r
F P
0 g
s
.54785 .6605 m .54877 .71023 L .54232 .7289 L p .76 .61 .691 r
F P
0 g
s
.54232 .7289 m .54179 .67991 L .54785 .6605 L p .76 .61 .691 r
F P
0 g
s
.55453 .61173 m .54785 .6605 L .54179 .67991 L p .863 .74 .725 r
F P
0 g
s
.59159 1.14126 m .57619 1.15097 L .57488 1.16042 L p .611 .399 .569 r
F P
0 g
s
.57488 1.16042 m .5896 1.15121 L .59159 1.14126 L p .611 .399 .569 r
F P
0 g
s
.4734 1.19475 m .44984 1.20773 L .45194 1.21411 L p .769 .782 .864 r
F P
0 g
s
.45194 1.21411 m .47477 1.20182 L .4734 1.19475 L p .769 .782 .864 r
F P
0 g
s
.66417 1.25961 m .67655 1.24617 L .67101 1.25247 L p 0 0 0 r
F P
0 g
s
.67921 1.23697 m .67101 1.25247 L .67655 1.24617 L p .889 .961 .629 r
F P
0 g
s
.59567 1.27126 m .57968 1.26263 L .57731 1.25751 L p .784 .456 .432 r
F P
0 g
s
.45287 .11041 m .48394 .16171 L .4725 .19053 L p .616 .616 .828 r
F P
0 g
s
.4725 .19053 m .43465 .14196 L .45287 .11041 L p .616 .616 .828 r
F P
0 g
s
.56556 1.04605 m .58381 1.07584 L .57942 1.08777 L p .766 .565 .627 r
F P
0 g
s
.57942 1.08777 m .56287 1.05842 L .56556 1.04605 L p .766 .565 .627 r
F P
0 g
s
.69963 .85833 m .66496 .85938 L .65329 .876 L p .569 .196 .294 r
F P
0 g
s
.62305 .88574 m .65329 .876 L .66496 .85938 L p .465 .116 .288 r
F P
0 g
s
.57619 1.15097 m .56379 1.1644 L .56302 1.17319 L p .585 .397 .593 r
F P
0 g
s
.56302 1.17319 m .57488 1.16042 L .57619 1.15097 L p .585 .397 .593 r
F P
0 g
s
.55282 .99144 m .5538 1.02558 L .55047 1.03826 L p .834 .67 .679 r
F P
0 g
s
.55047 1.03826 m .54966 1.00488 L .55282 .99144 L p .834 .67 .679 r
F P
0 g
s
.48274 1.26751 m .48429 1.26054 L .50238 1.26421 L p .708 .576 .705 r
F P
0 g
s
.50238 1.26421 m .50142 1.27043 L .48274 1.26751 L p .708 .576 .705 r
F P
0 g
s
.50413 1.25871 m .50238 1.26421 L .48429 1.26054 L p .707 .58 .711 r
F P
0 g
s
.50142 1.27043 m .50238 1.26421 L .5202 1.26766 L p .699 .569 .704 r
F P
0 g
s
.50238 1.26421 m .50413 1.25871 L .52095 1.26282 L p .698 .572 .71 r
F P
0 g
s
.52095 1.26282 m .5202 1.26766 L .50238 1.26421 L p .698 .572 .71 r
F P
0 g
s
.4659 1.2644 m .46799 1.25666 L .48429 1.26054 L p .72 .588 .707 r
F P
0 g
s
.48429 1.26054 m .48274 1.26751 L .4659 1.2644 L p .72 .588 .707 r
F P
0 g
s
.4871 1.25436 m .48429 1.26054 L .46799 1.25666 L p .718 .593 .715 r
F P
0 g
s
.48429 1.26054 m .4871 1.25436 L .50413 1.25871 L p .707 .58 .711 r
F P
0 g
s
.48191 1.01774 m .49918 1.03784 L .49672 1.04986 L p .605 .567 .785 r
F P
0 g
s
.49672 1.04986 m .47845 1.03023 L .48191 1.01774 L p .605 .567 .785 r
F P
0 g
s
.50934 1.07363 m .49672 1.04986 L .49918 1.03784 L p .591 .528 .752 r
F P
0 g
s
.48258 1.27481 m .48274 1.26751 L .50142 1.27043 L p .709 .573 .7 r
F P
0 g
s
.50142 1.27043 m .50132 1.27693 L .48258 1.27481 L p .709 .573 .7 r
F P
0 g
s
.50132 1.27693 m .50142 1.27043 L .51979 1.27312 L p .7 .566 .7 r
F P
0 g
s
.5202 1.26766 m .51979 1.27312 L .50142 1.27043 L p .699 .569 .704 r
F P
0 g
s
.49918 1.03784 m .51109 1.06223 L .50934 1.07363 L p .591 .528 .752 r
F P
0 g
s
.51544 1.09994 m .50934 1.07363 L .51109 1.06223 L p .549 .471 .72 r
F P
0 g
s
.46799 1.25666 m .47179 1.24981 L .4871 1.25436 L p .718 .593 .715 r
F P
0 g
s
.49099 1.24941 m .4871 1.25436 L .47179 1.24981 L p .715 .597 .723 r
F P
0 g
s
.4871 1.25436 m .49099 1.24941 L .50655 1.2543 L p .704 .584 .718 r
F P
0 g
s
.50655 1.2543 m .50413 1.25871 L .4871 1.25436 L p .704 .584 .718 r
F P
0 g
s
.50413 1.25871 m .50655 1.2543 L .52198 1.25895 L p .696 .576 .716 r
F P
0 g
s
.52198 1.25895 m .52095 1.26282 L .50413 1.25871 L p .696 .576 .716 r
F P
0 g
s
.46568 1.27249 m .4659 1.2644 L .48274 1.26751 L p .722 .584 .7 r
F P
0 g
s
.48274 1.26751 m .48258 1.27481 L .46568 1.27249 L p .722 .584 .7 r
F P
0 g
s
.4838 1.28193 m .48258 1.27481 L .50132 1.27693 L p .709 .569 .694 r
F P
0 g
s
.50132 1.27693 m .50208 1.28328 L .4838 1.28193 L p .709 .569 .694 r
F P
0 g
s
.50208 1.28328 m .50132 1.27693 L .51975 1.27883 L p .7 .562 .695 r
F P
0 g
s
.51979 1.27312 m .51975 1.27883 L .50132 1.27693 L p .7 .566 .7 r
F P
0 g
s
.64878 1.26936 m .66417 1.25961 L .65913 1.26527 L p .091 0 0 r
F P
0 g
s
.67101 1.25247 m .65913 1.26527 L .66417 1.25961 L p 0 0 0 r
F P
0 g
s
.47179 1.24981 m .47706 1.2443 L .49099 1.24941 L p .715 .597 .723 r
F P
0 g
s
.49569 1.24601 m .49099 1.24941 L .47706 1.2443 L p .709 .601 .734 r
F P
0 g
s
.49099 1.24941 m .49569 1.24601 L .50948 1.25128 L p .7 .589 .728 r
F P
0 g
s
.50948 1.25128 m .50655 1.2543 L .49099 1.24941 L p .7 .589 .728 r
F P
0 g
s
.68583 .80709 m .73666 .81905 L .72297 .83654 L p .637 .23 .254 r
F P
0 g
s
.72297 .83654 m .67522 .82455 L .68583 .80709 L p .637 .23 .254 r
F P
0 g
s
.50655 1.2543 m .50948 1.25128 L .52323 1.25629 L p .691 .581 .727 r
F P
0 g
s
.52323 1.25629 m .52198 1.25895 L .50655 1.2543 L p .691 .581 .727 r
F P
0 g
s
.46733 1.28038 m .46568 1.27249 L .48258 1.27481 L p .722 .578 .692 r
F P
0 g
s
.48258 1.27481 m .4838 1.28193 L .46733 1.28038 L p .722 .578 .692 r
F P
0 g
s
.46043 1.00328 m .48191 1.01774 L .47845 1.03023 L p .617 .612 .823 r
F P
0 g
s
.47845 1.03023 m .45577 1.01607 L .46043 1.00328 L p .617 .612 .823 r
F P
0 g
s
.57968 1.26263 m .56645 1.25009 L .56463 1.24457 L p .827 .54 .502 r
F P
0 g
s
.48632 1.2884 m .4838 1.28193 L .50208 1.28328 L p .709 .564 .688 r
F P
0 g
s
.50208 1.28328 m .50365 1.28905 L .48632 1.2884 L p .709 .564 .688 r
F P
0 g
s
.50365 1.28905 m .50208 1.28328 L .52008 1.28441 L p .7 .558 .69 r
F P
0 g
s
.51975 1.27883 m .52008 1.28441 L .50208 1.28328 L p .7 .562 .695 r
F P
0 g
s
.51109 1.06223 m .51685 1.08927 L .51544 1.09994 L p .549 .471 .72 r
F P
0 g
s
.52248 .73657 m .52177 .78323 L .51641 .80072 L p .523 .415 .673 r
F P
0 g
s
.50678 .84475 m .51641 .80072 L .52177 .78323 L p 0 0 .291 r
F P
0 g
s
.56379 1.1644 m .55522 1.18061 L .55483 1.18866 L p .547 .384 .611 r
F P
0 g
s
.55483 1.18866 m .56302 1.17319 L .56379 1.1644 L p .547 .384 .611 r
F P
0 g
s
.51979 1.27312 m .5202 1.26766 L .53592 1.27084 L p .692 .563 .704 r
F P
0 g
s
.5202 1.26766 m .52095 1.26282 L .53583 1.26668 L p .691 .566 .71 r
F P
0 g
s
.53583 1.26668 m .53592 1.27084 L .5202 1.26766 L p .691 .566 .71 r
F P
0 g
s
.47706 1.2443 m .48342 1.24051 L .49569 1.24601 L p .709 .601 .734 r
F P
0 g
s
.50089 1.2444 m .49569 1.24601 L .48342 1.24051 L p .699 .606 .749 r
F P
0 g
s
.49569 1.24601 m .50089 1.2444 L .51272 1.24985 L p .69 .596 .746 r
F P
0 g
s
.51272 1.24985 m .50948 1.25128 L .49569 1.24601 L p .69 .596 .746 r
F P
0 g
s
.51975 1.27883 m .51979 1.27312 L .53597 1.27555 L p .693 .56 .7 r
F P
0 g
s
.53592 1.27084 m .53597 1.27555 L .51979 1.27312 L p .692 .563 .704 r
F P
0 g
s
.50948 1.25128 m .51272 1.24985 L .52461 1.25504 L p .679 .592 .752 r
F P
0 g
s
.52461 1.25504 m .52323 1.25629 L .50948 1.25128 L p .679 .592 .752 r
F P
0 g
s
.47074 1.28754 m .46733 1.28038 L .4838 1.28193 L p .723 .571 .683 r
F P
0 g
s
.4838 1.28193 m .48632 1.2884 L .47074 1.28754 L p .723 .571 .683 r
F P
0 g
s
.52095 1.26282 m .52198 1.25895 L .53571 1.26333 L p .688 .571 .718 r
F P
0 g
s
.53571 1.26333 m .53583 1.26668 L .52095 1.26282 L p .688 .571 .718 r
F P
0 g
s
.52008 1.28441 m .51975 1.27883 L .53598 1.28048 L p .693 .556 .695 r
F P
0 g
s
.53597 1.27555 m .53598 1.28048 L .51975 1.27883 L p .693 .56 .7 r
F P
0 g
s
.52177 .78323 m .52248 .73657 L .52856 .71796 L p .603 .481 .687 r
F P
0 g
s
.52598 .67151 m .52856 .71796 L .52248 .73657 L p .698 .591 .732 r
F P
0 g
s
.52248 .73657 m .5171 .69097 L .52598 .67151 L p .698 .591 .732 r
F P
0 g
s
.52198 1.25895 m .52323 1.25629 L .53555 1.26104 L p .682 .578 .732 r
F P
0 g
s
.53555 1.26104 m .53571 1.26333 L .52198 1.25895 L p .682 .578 .732 r
F P
0 g
s
.4556 1.2526 m .46019 1.24505 L .47179 1.24981 L p .738 .616 .723 r
F P
0 g
s
.47179 1.24981 m .46799 1.25666 L .4556 1.2526 L p .738 .616 .723 r
F P
0 g
s
.47706 1.2443 m .47179 1.24981 L .46019 1.24505 L p .731 .62 .735 r
F P
0 g
s
.45308 1.2611 m .4556 1.2526 L .46799 1.25666 L p .742 .61 .712 r
F P
0 g
s
.46799 1.25666 m .4659 1.2644 L .45308 1.2611 L p .742 .61 .712 r
F P
0 g
s
.48998 1.29377 m .48632 1.2884 L .50365 1.28905 L p .709 .557 .68 r
F P
0 g
s
.50365 1.28905 m .50593 1.29385 L .48998 1.29377 L p .709 .557 .68 r
F P
0 g
s
.50593 1.29385 m .50365 1.28905 L .52075 1.28948 L p .7 .552 .683 r
F P
0 g
s
.52008 1.28441 m .52075 1.28948 L .50365 1.28905 L p .7 .558 .69 r
F P
0 g
s
.46019 1.24505 m .46655 1.23898 L .47706 1.2443 L p .731 .62 .735 r
F P
0 g
s
.48342 1.24051 m .47706 1.2443 L .46655 1.23898 L p .721 .622 .747 r
F P
0 g
s
.45281 1.26999 m .45308 1.2611 L .4659 1.2644 L p .745 .603 .7 r
F P
0 g
s
.4659 1.2644 m .46568 1.27249 L .45281 1.26999 L p .745 .603 .7 r
F P
0 g
s
.48342 1.24051 m .49046 1.23871 L .50089 1.2444 L p .699 .606 .749 r
F P
0 g
s
.50623 1.24469 m .50089 1.2444 L .49046 1.23871 L p .676 .61 .775 r
F P
0 g
s
.50089 1.2444 m .50623 1.24469 L .51604 1.25011 L p .654 .612 .795 r
F P
0 g
s
.51604 1.25011 m .51272 1.24985 L .51135 1.24922 L p .654 .612 .795 r
F P
0 g
s
.51604 1.25011 m .51135 1.24922 L .51099 1.24821 L .654 .612 .795 r
F
.51272 1.24985 m .51604 1.25011 L .52603 1.25527 L p .549 .637 .89 r
F P
0 g
s
.51923 1.25205 m .51604 1.25011 L .51436 1.24918 L p .717 .2 0 r
F P
0 g
s
.51436 1.24918 m .50623 1.24469 L p .51441 1.24932 L .717 .2 0 r
F P
0 g
s
.51923 1.25205 m .51436 1.24918 L .51441 1.24932 L .717 .2 0 r
F
.52603 1.25527 m .52461 1.25504 L .52326 1.25445 L p .549 .637 .89 r
F P
0 g
s
.52326 1.25445 m .51272 1.24985 L p .52164 1.25348 L .549 .637 .89 r
F P
0 g
s
.52603 1.25527 m .52326 1.25445 L .52164 1.25348 L .549 .637 .89 r
F
.52323 1.25629 m .52461 1.25504 L .53539 1.25995 L p .658 .596 .774 r
F P
0 g
s
.51923 1.25205 m .52739 1.25697 L p .51997 1.25248 L .741 .522 .596 r
F P
0 g
s
.51604 1.25011 m .51871 1.25173 L p .51997 1.25248 L .741 .522 .596 r
F P
0 g
s
.51871 1.25173 m .51923 1.25205 L p .51997 1.25248 L .741 .522 .596 r
F P
0 g
s
.53539 1.25995 m .53374 1.26034 L .53121 1.25869 L .658 .596 .774 r
F
.53539 1.25995 m .5342 1.25941 L p .53258 1.25888 L .752 .38 .35 r
F P
0 g
s
.5342 1.25941 m .52461 1.25504 L p .53177 1.25849 L .752 .38 .35 r
F P
0 g
s
.53258 1.25888 m .5342 1.25941 L .53177 1.25849 L .752 .38 .35 r
F
.53374 1.26034 m .52323 1.25629 L p .53121 1.25869 L .658 .596 .774 r
F P
0 g
s
.52739 1.25697 m .53505 1.26161 L p .52924 1.25752 L .72 .539 .644 r
F P
0 g
s
.52591 1.25525 m .52603 1.25527 L p .52836 1.25684 L .752 .38 .35 r
F P
0 g
s
.52603 1.25527 m .52722 1.25676 L p .52924 1.25752 L .72 .539 .644 r
F P
0 g
s
.52722 1.25676 m .52739 1.25697 L p .52924 1.25752 L .72 .539 .644 r
F P
0 g
s
.52321 1.25381 m .51604 1.25011 L p .52254 1.25404 L .741 .522 .596 r
F P
0 g
s
.51135 1.24685 m .51923 1.25205 L p .51441 1.24932 L .717 .2 0 r
F P
0 g
s
.51923 1.25205 m .51135 1.24685 L p .52089 1.25459 L .752 .549 .622 r
F P
0 g
s
.52206 1.25554 m .52121 1.25449 L p .52089 1.25459 L .752 .549 .622 r
F P
0 g
s
.52121 1.25449 m .51923 1.25205 L p .52089 1.25459 L .752 .549 .622 r
F P
0 g
s
.52461 1.25504 m .52591 1.25525 L p .52836 1.25684 L .752 .38 .35 r
F P
0 g
s
.52603 1.25527 m .52321 1.25381 L p .52667 1.25654 L .741 .522 .596 r
F P
0 g
s
.52667 1.25654 m .52321 1.25381 L .52254 1.25404 L .741 .522 .596 r
F
.52739 1.25697 m .52722 1.25676 L p .52667 1.25654 L F P
0 g
s
.52722 1.25676 m .52603 1.25527 L p .52667 1.25654 L .741 .522 .596 r
F P
0 g
s
.53522 1.26015 m .53536 1.25998 L p .53258 1.25888 L .752 .38 .35 r
F P
0 g
s
.53536 1.25998 m .53539 1.25995 L p .53258 1.25888 L .752 .38 .35 r
F P
0 g
s
.52603 1.25527 m .53522 1.26015 L p .52836 1.25684 L .752 .38 .35 r
F P
0 g
s
.53539 1.25995 m .53555 1.26104 L .53374 1.26034 L p .658 .596 .774 r
F P
0 g
s
.51441 1.24932 m .51063 1.24718 L .51135 1.24685 L .717 .2 0 r
F
.49768 1.23901 m .5046 1.24139 L .51135 1.24685 L p .594 .61 .838 r
F P
0 g
s
.51135 1.24685 m .5046 1.24139 L p .50549 1.24213 L .76 .246 0 r
F P
0 g
s
.51036 1.24644 m .51135 1.24685 L p .51063 1.24718 L .717 .2 0 r
F P
0 g
s
.51135 1.24685 m .50623 1.24469 L .49768 1.23901 L p .594 .61 .838 r
F P
0 g
s
.51453 1.24962 m .51135 1.24685 L .50549 1.24213 L .76 .246 0 r
F
.5159 1.25075 m .51451 1.24956 L p .51453 1.24962 L F P
0 g
s
.51451 1.24956 m .51135 1.24685 L p .51453 1.24962 L .76 .246 0 r
F P
0 g
s
.50623 1.24469 m .51036 1.24644 L p .51063 1.24718 L .717 .2 0 r
F P
0 g
s
.51135 1.24922 m .50089 1.2444 L p .51099 1.24821 L .654 .612 .795 r
F P
0 g
s
.63144 1.27474 m .64878 1.26936 L .64437 1.27448 L p .45 .047 .181 r
F P
0 g
s
.65913 1.26527 m .64437 1.27448 L .64878 1.26936 L p .091 0 0 r
F P
0 g
s
.52075 1.28948 m .52008 1.28441 L .53594 1.28529 L p .693 .552 .69 r
F P
0 g
s
.53598 1.28048 m .53594 1.28529 L .52008 1.28441 L p .693 .556 .695 r
F P
0 g
s
.47569 1.29348 m .47074 1.28754 L .48632 1.2884 L p .722 .561 .672 r
F P
0 g
s
.48632 1.2884 m .48998 1.29377 L .47569 1.29348 L p .722 .561 .672 r
F P
0 g
s
.56645 1.25009 m .55689 1.23449 L .55546 1.2284 L p .861 .605 .55 r
F P
0 g
s
.5538 1.02558 m .56287 1.05842 L .5579 1.07046 L p .749 .579 .664 r
F P
0 g
s
.5579 1.07046 m .55047 1.03826 L .5538 1.02558 L p .749 .579 .664 r
F P
0 g
s
.46655 1.23898 m .47423 1.23479 L .48342 1.24051 L p .721 .622 .747 r
F P
0 g
s
.49046 1.23871 m .48342 1.24051 L .47423 1.23479 L p .707 .622 .759 r
F P
0 g
s
.43619 .99547 m .46043 1.00328 L .45577 1.01607 L p .634 .672 .868 r
F P
0 g
s
.45577 1.01607 m .4302 1.00835 L .43619 .99547 L p .634 .672 .868 r
F P
0 g
s
.4548 1.27865 m .45281 1.26999 L .46568 1.27249 L p .747 .594 .686 r
F P
0 g
s
.46568 1.27249 m .46733 1.28038 L .4548 1.27865 L p .747 .594 .686 r
F P
0 g
s
.54179 .67991 m .54551 .63186 L .55453 .61173 L p .863 .74 .725 r
F P
0 g
s
.54179 .67991 m .54232 .7289 L .53527 .7473 L p .71 .569 .694 r
F P
0 g
s
.53527 .7473 m .53518 .69907 L .54179 .67991 L p .71 .569 .694 r
F P
0 g
s
.54551 .63186 m .54179 .67991 L .53518 .69907 L p .797 .677 .731 r
F P
0 g
s
.54877 .71023 m .55726 .75784 L .54706 .77582 L p .652 .491 .652 r
F P
0 g
s
.54706 .77582 m .54232 .7289 L .54877 .71023 L p .652 .491 .652 r
F P
0 g
s
.52075 1.28948 m .52172 1.29369 L .50593 1.29385 L p .7 .552 .683 r
F P
0 g
s
.52172 1.29369 m .52075 1.28948 L .53586 1.28965 L p .693 .546 .683 r
F P
0 g
s
.53594 1.28529 m .53586 1.28965 L .52075 1.28948 L p .693 .552 .69 r
F P
0 g
s
.49452 1.29767 m .48998 1.29377 L .50593 1.29385 L p .708 .546 .667 r
F P
0 g
s
.50593 1.29385 m .50876 1.29734 L .49452 1.29767 L p .708 .546 .667 r
F P
0 g
s
.50876 1.29734 m .50593 1.29385 L .52172 1.29369 L p .699 .542 .671 r
F P
0 g
s
.47423 1.23479 m .48272 1.23278 L .49046 1.23871 L p .707 .622 .759 r
F P
0 g
s
.49768 1.23901 m .49046 1.23871 L .48272 1.23278 L p .687 .617 .772 r
F P
0 g
s
.49046 1.23871 m .49768 1.23901 L .50623 1.24469 L p .676 .61 .775 r
F P
0 g
s
.91094 .49302 m .87264 .45758 L .85207 .47388 L p .814 .346 .106 r
F P
0 g
s
.80803 .45039 m .85207 .47388 L .87264 .45758 L p .689 .181 .023 r
F P
0 g
s
.53597 1.27555 m .53592 1.27084 L .54816 1.27374 L p .685 .558 .705 r
F P
0 g
s
.53592 1.27084 m .53583 1.26668 L .54747 1.27023 L p .683 .562 .712 r
F P
0 g
s
.54747 1.27023 m .54816 1.27374 L .53592 1.27084 L p .683 .562 .712 r
F P
0 g
s
.45892 1.2865 m .4548 1.27865 L .46733 1.28038 L p .747 .582 .671 r
F P
0 g
s
.46733 1.28038 m .47074 1.28754 L .45892 1.2865 L p .747 .582 .671 r
F P
0 g
s
.48183 1.29779 m .47569 1.29348 L .48998 1.29377 L p .72 .545 .653 r
F P
0 g
s
.48998 1.29377 m .49452 1.29767 L .48183 1.29779 L p .72 .545 .653 r
F P
0 g
s
.53583 1.26668 m .53571 1.26333 L .54652 1.26742 L p .679 .567 .722 r
F P
0 g
s
.54652 1.26742 m .54747 1.27023 L .53583 1.26668 L p .679 .567 .722 r
F P
0 g
s
.53598 1.28048 m .53597 1.27555 L .54854 1.27769 L p .686 .554 .699 r
F P
0 g
s
.54816 1.27374 m .54854 1.27769 L .53597 1.27555 L p .685 .558 .705 r
F P
0 g
s
.50594 .64928 m .5171 .69097 L .50979 .71012 L p .672 .599 .766 r
F P
0 g
s
.50979 .71012 m .49395 .66935 L .50594 .64928 L p .672 .599 .766 r
F P
0 g
s
.47087 .6352 m .49395 .66935 L .48558 .68908 L p .627 .604 .808 r
F P
0 g
s
.48558 .68908 m .45789 .65587 L .47087 .6352 L p .627 .604 .808 r
F P
0 g
s
.58381 1.07584 m .60909 1.09992 L .60239 1.11164 L p .756 .524 .581 r
F P
0 g
s
.60239 1.11164 m .57942 1.08777 L .58381 1.07584 L p .756 .524 .581 r
F P
0 g
s
.73346 .86636 m .69963 .85833 L .68499 .87507 L p .667 .291 .323 r
F P
0 g
s
.65329 .876 m .68499 .87507 L .69963 .85833 L p .569 .196 .294 r
F P
0 g
s
.53571 1.26333 m .53555 1.26104 L .54537 1.26549 L p .667 .577 .745 r
F P
0 g
s
.54537 1.26549 m .54652 1.26742 L .53571 1.26333 L p .667 .577 .745 r
F P
0 g
s
.53594 1.28529 m .53598 1.28048 L .54858 1.28183 L p .686 .55 .694 r
F P
0 g
s
.54854 1.27769 m .54858 1.28183 L .53598 1.28048 L p .686 .554 .699 r
F P
0 g
s
.61332 1.27539 m .63144 1.27474 L .62775 1.27948 L p .627 .272 .35 r
F P
0 g
s
.64437 1.27448 m .62775 1.27948 L .63144 1.27474 L p .45 .047 .181 r
F P
0 g
s
.52172 1.29369 m .52293 1.29676 L .50876 1.29734 L p .699 .542 .671 r
F P
0 g
s
.52293 1.29676 m .52172 1.29369 L .53575 1.29328 L p .691 .535 .671 r
F P
0 g
s
.53586 1.28965 m .53575 1.29328 L .52172 1.29369 L p .693 .546 .683 r
F P
0 g
s
.48272 1.23278 m .49143 1.23308 L .49768 1.23901 L p .687 .617 .772 r
F P
0 g
s
.5046 1.24139 m .49768 1.23901 L .49143 1.23308 L p .654 .605 .787 r
F P
0 g
s
.53555 1.26104 m .53539 1.25995 L .5441 1.26457 L p .549 .619 .874 r
F P
0 g
s
.54377 1.26477 m .53555 1.26104 L p .54116 1.26336 L .549 .619 .874 r
F P
0 g
s
.53522 1.26015 m .5428 1.26473 L p .53813 1.26172 L .73 .513 .598 r
F P
0 g
s
.53539 1.25995 m .53522 1.26015 L p .53813 1.26172 L .73 .513 .598 r
F P
0 g
s
.53522 1.26015 m .53522 1.26015 L p .53813 1.26172 L .73 .513 .598 r
F P
0 g
s
.5441 1.26457 m .54537 1.26549 L .54377 1.26477 L p .549 .619 .874 r
F P
0 g
s
.5441 1.26457 m .54377 1.26477 L .54116 1.26336 L .549 .619 .874 r
F
.53505 1.26161 m .53522 1.26015 L .52603 1.25527 L p .72 .539 .644 r
F P
0 g
s
.49964 1.29985 m .49452 1.29767 L .50876 1.29734 L p .704 .519 .636 r
F P
0 g
s
.50876 1.29734 m .51194 1.29928 L .49964 1.29985 L p .704 .519 .636 r
F P
0 g
s
.51194 1.29928 m .50876 1.29734 L .52293 1.29676 L p .695 .518 .644 r
F P
0 g
s
.53586 1.28965 m .53594 1.28529 L .54828 1.28586 L p .686 .545 .688 r
F P
0 g
s
.54858 1.28183 m .54828 1.28586 L .53594 1.28529 L p .686 .55 .694 r
F P
0 g
s
.46489 1.293 m .45892 1.2865 L .47074 1.28754 L p .746 .565 .649 r
F P
0 g
s
.47074 1.28754 m .47569 1.29348 L .46489 1.293 L p .746 .565 .649 r
F P
0 g
s
.51135 1.24685 m .5159 1.25075 L p .51719 1.25159 L .752 .549 .622 r
F P
0 g
s
.51075 1.2457 m .5159 1.25075 L p .51153 1.24713 L .76 .246 0 r
F P
0 g
s
.5046 1.24139 m .51075 1.2457 L p .51153 1.24713 L .76 .246 0 r
F P
0 g
s
.51075 1.2457 m .5072 1.24037 L p .50952 1.24345 L .399 .475 .822 r
F P
0 g
s
.5159 1.25075 m .51075 1.2457 L p .51397 1.24951 L .797 .574 .599 r
F P
0 g
s
.51956 1.25612 m .5159 1.25075 L p .51397 1.24951 L .797 .574 .599 r
F P
0 g
s
.5157 1.25163 m .51956 1.25612 L p .51588 1.25176 L .797 .574 .599 r
F P
0 g
s
.51075 1.2457 m .5157 1.25163 L p .51588 1.25176 L .797 .574 .599 r
F P
0 g
s
.5157 1.25163 m .51075 1.2457 L p .50952 1.24345 L .399 .475 .822 r
F P
0 g
s
.5072 1.24037 m .51317 1.24686 L .5157 1.25163 L p .399 .475 .822 r
F P
0 g
s
.5157 1.25163 m .51317 1.24686 L p .51546 1.25146 L .944 .58 .243 r
F P
0 g
s
.51912 1.25879 m .5157 1.25163 L p .51546 1.25146 L .944 .58 .243 r
F P
0 g
s
.5159 1.25075 m .52206 1.25554 L p .51719 1.25159 L .752 .549 .622 r
F P
0 g
s
.51923 1.25205 m .52206 1.25554 L .5286 1.26004 L p .723 .554 .661 r
F P
0 g
s
.5286 1.26004 m .52739 1.25697 L .51923 1.25205 L p .723 .554 .661 r
F P
0 g
s
.48875 1.30018 m .48183 1.29779 L .49452 1.29767 L p .714 .507 .608 r
F P
0 g
s
.49452 1.29767 m .49964 1.29985 L .48875 1.30018 L p .714 .507 .608 r
F P
0 g
s
.52823 .76542 m .52856 .71796 L .53518 .69907 L p .661 .529 .692 r
F P
0 g
s
.53572 .6518 m .53518 .69907 L .52856 .71796 L p .745 .63 .733 r
F P
0 g
s
.52856 .71796 m .52598 .67151 L .53572 .6518 L p .745 .63 .733 r
F P
0 g
s
.52856 .71796 m .52823 .76542 L .52177 .78323 L p .603 .481 .687 r
F P
0 g
s
.5428 1.26473 m .5441 1.26457 L .53539 1.25995 L p .73 .513 .598 r
F P
0 g
s
.51608 1.11713 m .51685 1.08927 L .51972 1.07826 L p .495 .393 .668 r
F P
0 g
s
.51481 1.05049 m .51972 1.07826 L .51685 1.08927 L p .616 .519 .721 r
F P
0 g
s
.51685 1.08927 m .51109 1.06223 L .51481 1.05049 L p .616 .519 .721 r
F P
0 g
s
.53575 1.29328 m .53586 1.28965 L .54767 1.28953 L p .685 .538 .68 r
F P
0 g
s
.54828 1.28586 m .54767 1.28953 L .53586 1.28965 L p .686 .545 .688 r
F P
0 g
s
.52293 1.29676 m .52429 1.29847 L .51194 1.29928 L p .695 .518 .644 r
F P
0 g
s
.52429 1.29847 m .52293 1.29676 L .5356 1.29591 L p .686 .508 .64 r
F P
0 g
s
.53575 1.29328 m .5356 1.29591 L .52293 1.29676 L p .691 .535 .671 r
F P
0 g
s
.53518 .69907 m .53572 .6518 L .54551 .63186 L p .797 .677 .731 r
F P
0 g
s
.53518 .69907 m .53527 .7473 L .52823 .76542 L p .661 .529 .692 r
F P
0 g
s
.41082 .99485 m .43619 .99547 L .4302 1.00835 L p .662 .755 .919 r
F P
0 g
s
.4302 1.00835 m .40349 1.0076 L .41082 .99485 L p .662 .755 .919 r
F P
0 g
s
.49143 1.23308 m .49978 1.23567 L .5046 1.24139 L p .654 .605 .787 r
F P
0 g
s
.51075 1.2457 m .5046 1.24139 L .49978 1.23567 L p .594 .577 .804 r
F P
0 g
s
.56766 .9433 m .57479 .92835 L p .57043 .93834 L .842 .904 .493 r
F P
0 g
s
.55552 .97252 m .56766 .9433 L p .57043 .93834 L .842 .904 .493 r
F P
0 g
s
.54942 .98641 m .55801 .95794 L .56766 .9433 L p .91 .841 .763 r
F P
0 g
s
.56766 .9433 m .55801 .95794 L p .56518 .94777 L .903 .931 .822 r
F P
0 g
s
.58531 .9192 m .56766 .9433 L p .56518 .94777 L .903 .931 .822 r
F P
0 g
s
.59629 .9036 m .57479 .92835 L .56766 .9433 L p 0 0 0 r
F P
0 g
s
.50467 1.02547 m .51481 1.05049 L .51109 1.06223 L p .64 .559 .747 r
F P
0 g
s
.51109 1.06223 m .49918 1.03784 L .50467 1.02547 L p .64 .559 .747 r
F P
0 g
s
.52739 1.25697 m .5286 1.26004 L .53491 1.26424 L p .711 .552 .672 r
F P
0 g
s
.53491 1.26424 m .53505 1.26161 L .52739 1.25697 L p .711 .552 .672 r
F P
0 g
s
.59567 1.27126 m .61332 1.27539 L .6104 1.27994 L p .714 .41 .461 r
F P
0 g
s
.62775 1.27948 m .6104 1.27994 L .61332 1.27539 L p .627 .272 .35 r
F P
0 g
s
.47231 1.29771 m .46489 1.293 L .47569 1.29348 L p .742 .536 .615 r
F P
0 g
s
.47569 1.29348 m .48183 1.29779 L .47231 1.29771 L p .742 .536 .615 r
F P
0 g
s
.54966 1.00488 m .55047 1.03826 L .54585 1.05061 L p .777 .625 .691 r
F P
0 g
s
.54585 1.05061 m .54528 1.01799 L .54966 1.00488 L p .777 .625 .691 r
F P
0 g
s
.55552 .97252 m .54966 1.00488 L .54528 1.01799 L p .866 .738 .719 r
F P
0 g
s
.45409 1.24009 m .46108 1.23345 L .46655 1.23898 L p .761 .67 .761 r
F P
0 g
s
.46655 1.23898 m .46019 1.24505 L .45409 1.24009 L p .761 .67 .761 r
F P
0 g
s
.47423 1.23479 m .46655 1.23898 L .46108 1.23345 L p .74 .661 .773 r
F P
0 g
s
.50498 1.30015 m .49964 1.29985 L .51194 1.29928 L p .639 .298 .377 r
F P
0 g
s
.51194 1.29928 m .51527 1.29956 L .50498 1.30015 L p .639 .298 .377 r
F P
0 g
s
.51527 1.29956 m .51194 1.29928 L .52429 1.29847 L p .627 .293 .386 r
F P
0 g
s
.46108 1.23345 m .46952 1.22885 L .47423 1.23479 L p .74 .661 .773 r
F P
0 g
s
.48272 1.23278 m .47423 1.23479 L .46952 1.22885 L p .717 .647 .778 r
F P
0 g
s
.44904 1.24834 m .45409 1.24009 L .46019 1.24505 L p .78 .672 .745 r
F P
0 g
s
.46019 1.24505 m .4556 1.2526 L .44904 1.24834 L p .78 .672 .745 r
F P
0 g
s
.53522 1.26015 m .53505 1.26161 L .54155 1.26595 L p .71 .542 .659 r
F P
0 g
s
.54155 1.26595 m .5428 1.26473 L .53522 1.26015 L p .71 .542 .659 r
F P
0 g
s
.5159 1.25075 m .51956 1.25612 L .52434 1.26033 L p .735 .568 .666 r
F P
0 g
s
.52434 1.26033 m .52206 1.25554 L .5159 1.25075 L p .735 .568 .666 r
F P
0 g
s
.5356 1.29591 m .53575 1.29328 L .54678 1.29256 L p .682 .524 .665 r
F P
0 g
s
.54767 1.28953 m .54678 1.29256 L .53575 1.29328 L p .685 .538 .68 r
F P
0 g
s
.55165 1.2169 m .55689 1.23449 L .55643 1.24025 L p .825 .633 .643 r
F P
0 g
s
.55643 1.24025 m .55142 1.22337 L .55165 1.2169 L p .825 .633 .643 r
F P
0 g
s
.52206 1.25554 m .52434 1.26033 L .52958 1.26424 L p .716 .562 .679 r
F P
0 g
s
.52958 1.26424 m .5286 1.26004 L .52206 1.25554 L p .716 .562 .679 r
F P
0 g
s
.46952 1.22885 m .47884 1.22661 L .48272 1.23278 L p .717 .647 .778 r
F P
0 g
s
.49143 1.23308 m .48272 1.23278 L .47884 1.22661 L p .695 .629 .778 r
F P
0 g
s
.54854 1.27769 m .54816 1.27374 L .55607 1.27631 L p .675 .551 .707 r
F P
0 g
s
.54816 1.27374 m .54747 1.27023 L .55506 1.27347 L p .671 .556 .717 r
F P
0 g
s
.55506 1.27347 m .55607 1.27631 L .54816 1.27374 L p .671 .556 .717 r
F P
0 g
s
.54747 1.27023 m .54652 1.26742 L .55368 1.27119 L p .66 .563 .735 r
F P
0 g
s
.55368 1.27119 m .55506 1.27347 L .54747 1.27023 L p .66 .563 .735 r
F P
0 g
s
.49598 1.30049 m .48875 1.30018 L .49964 1.29985 L p .603 .191 .234 r
F P
0 g
s
.49964 1.29985 m .50498 1.30015 L .49598 1.30049 L p .603 .191 .234 r
F P
0 g
s
.44627 1.25763 m .44904 1.24834 L .4556 1.2526 L p .796 .668 .722 r
F P
0 g
s
.4556 1.2526 m .45308 1.2611 L .44627 1.25763 L p .796 .668 .722 r
F P
0 g
s
.54858 1.28183 m .54854 1.27769 L .55662 1.27951 L p .677 .547 .699 r
F P
0 g
s
.55607 1.27631 m .55662 1.27951 L .54854 1.27769 L p .675 .551 .707 r
F P
0 g
s
.54652 1.26742 m .54537 1.26549 L .552 1.26963 L p .61 .58 .795 r
F P
0 g
s
.55368 1.27119 m .54652 1.26742 L p .55162 1.26947 L .61 .58 .795 r
F P
0 g
s
.5428 1.26473 m .54824 1.269 L p .54414 1.26461 L .717 .534 .641 r
F P
0 g
s
.5441 1.26457 m .54408 1.26457 L p .54414 1.26461 L .717 .534 .641 r
F P
0 g
s
.54408 1.26457 m .5428 1.26473 L p .54414 1.26461 L .717 .534 .641 r
F P
0 g
s
.552 1.26963 m .55233 1.26994 L p .55162 1.26947 L .61 .58 .795 r
F P
0 g
s
.55233 1.26994 m .55368 1.27119 L p .55162 1.26947 L .61 .58 .795 r
F P
0 g
s
.52429 1.29847 m .52571 1.29871 L .51527 1.29956 L p .627 .293 .386 r
F P
0 g
s
.52571 1.29871 m .52429 1.29847 L .53544 1.29738 L p .526 .106 .186 r
F P
0 g
s
.5356 1.29591 m .53544 1.29738 L .52429 1.29847 L p .686 .508 .64 r
F P
0 g
s
.54828 1.28586 m .54858 1.28183 L .55668 1.28286 L p .677 .541 .692 r
F P
0 g
s
.55662 1.27951 m .55668 1.28286 L .54858 1.28183 L p .677 .547 .699 r
F P
0 g
s
.47884 1.22661 m .48842 1.2269 L .49143 1.23308 L p .695 .629 .778 r
F P
0 g
s
.49978 1.23567 m .49143 1.23308 L .48842 1.2269 L p .673 .608 .774 r
F P
0 g
s
.34302 .61574 m .3842 .61784 L .37262 .63968 L p .62 .799 .968 r
F P
0 g
s
.37262 .63968 m .32878 .63808 L .34302 .61574 L p .62 .799 .968 r
F P
0 g
s
.49978 1.23567 m .5072 1.24037 L .51075 1.2457 L p .594 .577 .804 r
F P
0 g
s
.55014 1.26888 m .552 1.26963 L .54537 1.26549 L p .765 .463 .475 r
F P
0 g
s
.54537 1.26549 m .5441 1.26457 L .55014 1.26888 L p .765 .463 .475 r
F P
0 g
s
.57968 1.26263 m .59567 1.27126 L .5935 1.27582 L p .756 .496 .539 r
F P
0 g
s
.6104 1.27994 m .5935 1.27582 L .59567 1.27126 L p .714 .41 .461 r
F P
0 g
s
.5286 1.26004 m .52958 1.26424 L .53479 1.26787 L p .706 .558 .683 r
F P
0 g
s
.53479 1.26787 m .53491 1.26424 L .5286 1.26004 L p .706 .558 .683 r
F P
0 g
s
.44597 1.26731 m .44627 1.25763 L .45308 1.2611 L p .808 .656 .694 r
F P
0 g
s
.45308 1.2611 m .45281 1.26999 L .44597 1.26731 L p .808 .656 .694 r
F P
0 g
s
.48066 1.3003 m .47231 1.29771 L .48183 1.29779 L p .727 .464 .53 r
F P
0 g
s
.48183 1.29779 m .48875 1.30018 L .48066 1.3003 L p .727 .464 .53 r
F P
0 g
s
.57488 1.16042 m .56302 1.17319 L .55994 1.18165 L p .415 .19 .462 r
F P
0 g
s
.53505 1.26161 m .53491 1.26424 L .54044 1.26815 L p .704 .55 .676 r
F P
0 g
s
.54044 1.26815 m .54155 1.26595 L .53505 1.26161 L p .704 .55 .676 r
F P
0 g
s
.5896 1.15121 m .57488 1.16042 L .57032 1.16952 L p .523 .277 .484 r
F P
0 g
s
.55994 1.18165 m .57032 1.16952 L .57488 1.16042 L p .415 .19 .462 r
F P
0 g
s
.53544 1.29738 m .5356 1.29591 L .54567 1.29477 L p .671 .481 .62 r
F P
0 g
s
.54678 1.29256 m .54567 1.29477 L .5356 1.29591 L p .682 .524 .665 r
F P
0 g
s
.54767 1.28953 m .54828 1.28586 L .55625 1.28612 L p .676 .534 .684 r
F P
0 g
s
.55668 1.28286 m .55625 1.28612 L .54828 1.28586 L p .677 .541 .692 r
F P
0 g
s
.51019 1.29854 m .50498 1.30015 L .51527 1.29956 L p .709 .658 .797 r
F P
0 g
s
.51527 1.29956 m .51851 1.29813 L .51019 1.29854 L p .709 .658 .797 r
F P
0 g
s
.51851 1.29813 m .51527 1.29956 L .52571 1.29871 L p .704 .638 .78 r
F P
0 g
s
.55689 1.23449 m .56645 1.25009 L .56556 1.25527 L p .789 .588 .629 r
F P
0 g
s
.56556 1.25527 m .55643 1.24025 L .55689 1.23449 L p .789 .588 .629 r
F P
0 g
s
.54824 1.269 m .55014 1.26888 L .5441 1.26457 L p .717 .534 .641 r
F P
0 g
s
.51499 .82802 m .52177 .78323 L .52823 .76542 L p .332 .197 .543 r
F P
0 g
s
.52177 .78323 m .51499 .82802 L .50678 .84475 L p 0 0 .291 r
F P
0 g
s
.48842 1.2269 m .49759 1.22967 L .49978 1.23567 L p .673 .608 .774 r
F P
0 g
s
.5072 1.24037 m .49978 1.23567 L .49759 1.22967 L p .652 .584 .766 r
F P
0 g
s
.52209 1.26259 m .51956 1.25612 L .5157 1.25163 L p .763 .59 .664 r
F P
0 g
s
.51956 1.25612 m .52209 1.26259 L .52592 1.2661 L p .726 .574 .683 r
F P
0 g
s
.52592 1.2661 m .52434 1.26033 L .51956 1.25612 L p .726 .574 .683 r
F P
0 g
s
.52434 1.26033 m .52592 1.2661 L .53025 1.26931 L p .712 .566 .688 r
F P
0 g
s
.53025 1.26931 m .52958 1.26424 L .52434 1.26033 L p .712 .566 .688 r
F P
0 g
s
.56645 1.25009 m .57968 1.26263 L .57821 1.2674 L p .776 .551 .594 r
F P
0 g
s
.5935 1.27582 m .57821 1.2674 L .57968 1.26263 L p .756 .496 .539 r
F P
0 g
s
.50302 1.29869 m .49598 1.30049 L .50498 1.30015 L p .715 .691 .826 r
F P
0 g
s
.50498 1.30015 m .51019 1.29854 L .50302 1.29869 L p .715 .691 .826 r
F P
0 g
s
.44816 1.27675 m .44597 1.26731 L .45281 1.26999 L p .815 .636 .66 r
F P
0 g
s
.45281 1.26999 m .4548 1.27865 L .44816 1.27675 L p .815 .636 .66 r
F P
0 g
s
.57821 1.2674 m .56556 1.25527 L .56645 1.25009 L p .776 .551 .594 r
F P
0 g
s
.54678 1.29256 m .54767 1.28953 L .55535 1.28908 L p .673 .522 .672 r
F P
0 g
s
.55625 1.28612 m .55535 1.28908 L .54767 1.28953 L p .676 .534 .684 r
F P
0 g
s
.52571 1.29871 m .52709 1.29746 L .51851 1.29813 L p .704 .638 .78 r
F P
0 g
s
.53544 1.29738 m .53527 1.29758 L .52571 1.29871 L p .526 .106 .186 r
F P
0 g
s
.52709 1.29746 m .52571 1.29871 L .53527 1.29758 L p .701 .627 .771 r
F P
0 g
s
.6062 1.1462 m .5896 1.15121 L .58322 1.16081 L p .598 .342 .5 r
F P
0 g
s
.57032 1.16952 m .58322 1.16081 L .5896 1.15121 L p .523 .277 .484 r
F P
0 g
s
.5428 1.26473 m .54155 1.26595 L .54641 1.26998 L p .706 .544 .667 r
F P
0 g
s
.54641 1.26998 m .54824 1.269 L .5428 1.26473 L p .706 .544 .667 r
F P
0 g
s
.51972 1.07826 m .51906 1.10693 L .51608 1.11713 L p .495 .393 .668 r
F P
0 g
s
.53491 1.26424 m .53479 1.26787 L .53955 1.27119 L p .701 .554 .685 r
F P
0 g
s
.53955 1.27119 m .54044 1.26815 L .53491 1.26424 L p .701 .554 .685 r
F P
0 g
s
.5157 1.25163 m .51912 1.25879 L p .5194 1.25797 L .763 .59 .664 r
F P
0 g
s
.5173 1.25471 m .51912 1.25879 L p .51698 1.25449 L .944 .58 .243 r
F P
0 g
s
.51317 1.24686 m .5173 1.25471 L p .51698 1.25449 L .944 .58 .243 r
F P
0 g
s
.5173 1.25471 m .51682 1.25033 L p .5177 1.25498 L .371 .349 .704 r
F P
0 g
s
.51912 1.25879 m .5173 1.25471 L p .51878 1.25981 L .824 .642 .656 r
F P
0 g
s
.5193 1.26337 m .52078 1.26669 L p .51953 1.26239 L .824 .642 .656 r
F P
0 g
s
.5173 1.25471 m .5193 1.26337 L p .51953 1.26239 L .824 .642 .656 r
F P
0 g
s
.5193 1.26337 m .5173 1.25471 L p .5177 1.25498 L .371 .349 .704 r
F P
0 g
s
.52078 1.26669 m .51912 1.25879 L p .51878 1.25981 L .824 .642 .656 r
F P
0 g
s
.51901 1.25975 m .5193 1.26337 L p .51853 1.25935 L .371 .349 .704 r
F P
0 g
s
.51682 1.25033 m .51901 1.25975 L p .51853 1.25935 L .371 .349 .704 r
F P
0 g
s
.51912 1.25879 m .52209 1.26259 L p .5194 1.25797 L .763 .59 .664 r
F P
0 g
s
.52332 1.26972 m .52209 1.26259 L .51912 1.25879 L p .744 .592 .688 r
F P
0 g
s
.51912 1.25879 m .52078 1.26669 L .52332 1.26972 L p .744 .592 .688 r
F P
0 g
s
.51901 1.25975 m .51998 1.25584 L p .51958 1.26022 L .503 .4 .67 r
F P
0 g
s
.5193 1.26337 m .51901 1.25975 L p .51902 1.26453 L .947 .775 .637 r
F P
0 g
s
.51903 1.27226 m .5193 1.26337 L p .51902 1.26453 L .947 .775 .637 r
F P
0 g
s
.51872 1.26944 m .51903 1.27226 L p .51902 1.26834 L .947 .775 .637 r
F P
0 g
s
.51901 1.25975 m .51872 1.26944 L p .51902 1.26834 L .947 .775 .637 r
F P
0 g
s
.52056 1.27479 m .52078 1.26669 L .5193 1.26337 L p .769 .624 .698 r
F P
0 g
s
.5193 1.26337 m .51903 1.27226 L .52056 1.27479 L p .769 .624 .698 r
F P
0 g
s
.51872 1.26944 m .51969 1.26632 L p .51838 1.27068 L 0 0 0 r
F P
0 g
s
.51596 1.27872 m .51651 1.28077 L p .51724 1.27705 L .813 .687 .725 r
F P
0 g
s
.51872 1.26944 m .51596 1.27872 L p .51724 1.27705 L .813 .687 .725 r
F P
0 g
s
.51596 1.27872 m .51872 1.26944 L p .51838 1.27068 L 0 0 0 r
F P
0 g
s
.51651 1.28077 m .51903 1.27226 L .51872 1.26944 L p .813 .687 .725 r
F P
0 g
s
.51692 1.27638 m .51596 1.27872 L p .51704 1.27513 L 0 0 0 r
F P
0 g
s
.51969 1.26632 m .51692 1.27638 L p .51704 1.27513 L 0 0 0 r
F P
0 g
s
.51847 1.28253 m .52056 1.27479 L .51903 1.27226 L p .75 .618 .713 r
F P
0 g
s
.51903 1.27226 m .51651 1.28077 L .51847 1.28253 L p .75 .618 .713 r
F P
0 g
s
.51692 1.27638 m .51939 1.27373 L p .5168 1.27774 L .438 .244 .519 r
F P
0 g
s
.51596 1.27872 m .51692 1.27638 L p .51455 1.28057 L .929 .925 .785 r
F P
0 g
s
.51091 1.28698 m .51596 1.27872 L p .51455 1.28057 L .929 .925 .785 r
F P
0 g
s
.51192 1.28833 m .51651 1.28077 L .51596 1.27872 L p .786 .681 .748 r
F P
0 g
s
.51596 1.27872 m .51091 1.28698 L .51192 1.28833 L p .786 .681 .748 r
F P
0 g
s
.51187 1.28534 m .51692 1.27638 L p .5168 1.27774 L .438 .244 .519 r
F P
0 g
s
.51872 1.26944 m .51901 1.25975 L p .51958 1.26022 L .503 .4 .67 r
F P
0 g
s
.52958 1.26424 m .53025 1.26931 L .53471 1.27223 L p .703 .561 .69 r
F P
0 g
s
.53471 1.27223 m .53479 1.26787 L .52958 1.26424 L p .703 .561 .69 r
F P
0 g
s
.53527 1.29758 m .53544 1.29738 L .54442 1.29599 L p .496 .874 .976 r
F P
0 g
s
.54567 1.29477 m .54442 1.29599 L .53544 1.29738 L p .671 .481 .62 r
F P
0 g
s
.48938 1.3006 m .48066 1.3003 L .48875 1.30018 L p .432 0 0 r
F P
0 g
s
.48875 1.30018 m .49598 1.30049 L .48938 1.3006 L p .432 0 0 r
F P
0 g
s
.49759 1.22967 m .50573 1.23476 L .5072 1.24037 L p .652 .584 .766 r
F P
0 g
s
.51317 1.24686 m .5072 1.24037 L .50573 1.23476 L p .626 .557 .757 r
F P
0 g
s
.5149 1.29514 m .51019 1.29854 L .51851 1.29813 L p .711 .608 .739 r
F P
0 g
s
.51851 1.29813 m .52144 1.29511 L .5149 1.29514 L p .711 .608 .739 r
F P
0 g
s
.52144 1.29511 m .51851 1.29813 L .52709 1.29746 L p .704 .596 .733 r
F P
0 g
s
.54567 1.29477 m .54678 1.29256 L .55404 1.29153 L p .665 .496 .646 r
F P
0 g
s
.55535 1.28908 m .55404 1.29153 L .54678 1.29256 L p .673 .522 .672 r
F P
0 g
s
.45269 1.28528 m .44816 1.27675 L .4548 1.27865 L p .817 .605 .616 r
F P
0 g
s
.4548 1.27865 m .45892 1.2865 L .45269 1.28528 L p .817 .605 .616 r
F P
0 g
s
.54155 1.26595 m .54044 1.26815 L .54479 1.27175 L p .7 .549 .678 r
F P
0 g
s
.54479 1.27175 m .54641 1.26998 L .54155 1.26595 L p .7 .549 .678 r
F P
0 g
s
.52668 1.27246 m .52592 1.2661 L .52209 1.26259 L p .72 .577 .693 r
F P
0 g
s
.52592 1.2661 m .52668 1.27246 L .53058 1.2749 L p .708 .568 .695 r
F P
0 g
s
.53058 1.2749 m .53025 1.26931 L .52592 1.2661 L p .708 .568 .695 r
F P
0 g
s
.52209 1.26259 m .52332 1.26972 L .52668 1.27246 L p .72 .577 .693 r
F P
0 g
s
.50939 1.29491 m .50302 1.29869 L .51019 1.29854 L p .722 .628 .753 r
F P
0 g
s
.51019 1.29854 m .5149 1.29514 L .50939 1.29491 L p .722 .628 .753 r
F P
0 g
s
.52709 1.29746 m .52834 1.2948 L .52144 1.29511 L p .704 .596 .733 r
F P
0 g
s
.52834 1.2948 m .52709 1.29746 L .5351 1.2965 L p .699 .59 .731 r
F P
0 g
s
.53527 1.29758 m .5351 1.2965 L .52709 1.29746 L p .701 .627 .771 r
F P
0 g
s
.53479 1.26787 m .53471 1.27223 L .53894 1.27485 L p .698 .557 .69 r
F P
0 g
s
.53894 1.27485 m .53955 1.27119 L .53479 1.26787 L p .698 .557 .69 r
F P
0 g
s
.55607 1.27631 m .55506 1.27347 L .55839 1.27638 L p .619 .534 .737 r
F P
0 g
s
.55506 1.27347 m .55368 1.27119 L .55694 1.27463 L p .467 .511 .82 r
F P
0 g
s
.55694 1.27463 m .55839 1.27638 L .55506 1.27347 L p .467 .511 .82 r
F P
0 g
s
.55368 1.27119 m .552 1.26963 L .55519 1.27343 L p .79 .518 .526 r
F P
0 g
s
.55519 1.27343 m .55694 1.27463 L .55368 1.27119 L p .79 .518 .526 r
F P
0 g
s
.55662 1.27951 m .55607 1.27631 L .55944 1.27855 L p .642 .531 .713 r
F P
0 g
s
.55839 1.27638 m .55944 1.27855 L .55607 1.27631 L p .619 .534 .737 r
F P
0 g
s
.5351 1.2965 m .53527 1.29758 L .54312 1.29615 L p .699 .618 .763 r
F P
0 g
s
.54442 1.29599 m .54312 1.29615 L .53527 1.29758 L p .496 .874 .976 r
F P
0 g
s
.54528 1.01799 m .54942 .98641 L .55552 .97252 L p .866 .738 .719 r
F P
0 g
s
.53025 1.26931 m .53058 1.2749 L .53468 1.27705 L p .701 .563 .695 r
F P
0 g
s
.53468 1.27705 m .53471 1.27223 L .53025 1.26931 L p .701 .563 .695 r
F P
0 g
s
.62353 1.14572 m .6062 1.1462 L .59776 1.15611 L p .657 .401 .521 r
F P
0 g
s
.58322 1.16081 m .59776 1.15611 L .6062 1.1462 L p .598 .342 .5 r
F P
0 g
s
.56842 .56697 m .55453 .61173 L .54551 .63186 L p .889 .817 .771 r
F P
0 g
s
.552 1.26963 m .55014 1.26888 L .55326 1.27285 L p .728 .54 .637 r
F P
0 g
s
.55326 1.27285 m .55519 1.27343 L .552 1.26963 L p .728 .54 .637 r
F P
0 g
s
.50573 1.23476 m .51229 1.2418 L .51317 1.24686 L p .626 .557 .757 r
F P
0 g
s
.5173 1.25471 m .51317 1.24686 L .51229 1.2418 L p .58 .515 .746 r
F P
0 g
s
.55668 1.28286 m .55662 1.27951 L .56001 1.28099 L p .649 .525 .699 r
F P
0 g
s
.55944 1.27855 m .56001 1.28099 L .55662 1.27951 L p .642 .531 .713 r
F P
0 g
s
.54442 1.29599 m .54567 1.29477 L .55242 1.29331 L p .614 .366 .518 r
F P
0 g
s
.55404 1.29153 m .55242 1.29331 L .54567 1.29477 L p .665 .496 .646 r
F P
0 g
s
.48996 1.00488 m .50467 1.02547 L .49918 1.03784 L p .648 .591 .776 r
F P
0 g
s
.49918 1.03784 m .48191 1.01774 L .48996 1.00488 L p .648 .591 .776 r
F P
0 g
s
.49787 1.29859 m .48938 1.3006 L .49598 1.30049 L p .717 .748 .876 r
F P
0 g
s
.49598 1.30049 m .50302 1.29869 L .49787 1.29859 L p .717 .748 .876 r
F P
0 g
s
.54044 1.26815 m .53955 1.27119 L .54349 1.27421 L p .697 .551 .685 r
F P
0 g
s
.54349 1.27421 m .54479 1.27175 L .54044 1.26815 L p .697 .551 .685 r
F P
0 g
s
.54232 .7289 m .54706 .77582 L .53599 .79353 L p .589 .436 .641 r
F P
0 g
s
.53599 .79353 m .53527 .7473 L .54232 .7289 L p .589 .436 .641 r
F P
0 g
s
.5188 1.29019 m .5149 1.29514 L .52144 1.29511 L p .711 .592 .721 r
F P
0 g
s
.52144 1.29511 m .52387 1.2907 L .5188 1.29019 L p .711 .592 .721 r
F P
0 g
s
.52387 1.2907 m .52144 1.29511 L .52834 1.2948 L p .703 .583 .718 r
F P
0 g
s
.38605 1.00149 m .41082 .99485 L .40349 1.0076 L p .699 .868 .962 r
F P
0 g
s
.40349 1.0076 m .37744 1.01388 L .38605 1.00149 L p .699 .868 .962 r
F P
0 g
s
.55014 1.26888 m .54824 1.269 L .55127 1.27294 L p .711 .544 .661 r
F P
0 g
s
.55127 1.27294 m .55326 1.27285 L .55014 1.26888 L p .711 .544 .661 r
F P
0 g
s
.55047 1.03826 m .5579 1.07046 L .55111 1.08213 L p .714 .553 .669 r
F P
0 g
s
.55111 1.08213 m .54585 1.05061 L .55047 1.03826 L p .714 .553 .669 r
F P
0 g
s
.55625 1.28612 m .55668 1.28286 L .56007 1.28355 L p .65 .515 .686 r
F P
0 g
s
.56001 1.28099 m .56007 1.28355 L .55668 1.28286 L p .649 .525 .699 r
F P
0 g
s
.51906 1.10693 m .51972 1.07826 L .52383 1.06692 L p .587 .47 .686 r
F P
0 g
s
.52026 1.03841 m .52383 1.06692 L .51972 1.07826 L p .656 .549 .721 r
F P
0 g
s
.51972 1.07826 m .51481 1.05049 L .52026 1.03841 L p .656 .549 .721 r
F P
0 g
s
.45925 1.29233 m .45269 1.28528 L .45892 1.2865 L p .81 .557 .555 r
F P
0 g
s
.45892 1.2865 m .46489 1.293 L .45925 1.29233 L p .81 .557 .555 r
F P
0 g
s
.52658 1.27897 m .52668 1.27246 L .52332 1.26972 L p .715 .578 .7 r
F P
0 g
s
.52668 1.27246 m .52658 1.27897 L .53053 1.28062 L p .705 .57 .7 r
F P
0 g
s
.53053 1.28062 m .53058 1.2749 L .52668 1.27246 L p .705 .57 .7 r
F P
0 g
s
.52315 1.27703 m .52332 1.26972 L .52078 1.26669 L p .732 .592 .7 r
F P
0 g
s
.52332 1.26972 m .52315 1.27703 L .52658 1.27897 L p .715 .578 .7 r
F P
0 g
s
.52834 1.2948 m .52938 1.29092 L .52387 1.2907 L p .703 .583 .718 r
F P
0 g
s
.52938 1.29092 m .52834 1.2948 L .53495 1.2942 L p .697 .578 .717 r
F P
0 g
s
.5351 1.2965 m .53495 1.2942 L .52834 1.2948 L p .699 .59 .731 r
F P
0 g
s
.51466 1.2894 m .50939 1.29491 L .5149 1.29514 L p .722 .608 .729 r
F P
0 g
s
.5149 1.29514 m .5188 1.29019 L .51466 1.2894 L p .722 .608 .729 r
F P
0 g
s
.53471 1.27223 m .53468 1.27705 L .53864 1.27889 L p .696 .558 .695 r
F P
0 g
s
.53864 1.27889 m .53894 1.27485 L .53471 1.27223 L p .696 .558 .695 r
F P
0 g
s
.46447 .48041 m .42394 .5217 L .41354 .54577 L p .879 .913 .847 r
F P
0 g
s
.35996 .57426 m .41354 .54577 L .42394 .5217 L p .8 .932 .919 r
F P
0 g
s
.54824 1.269 m .54641 1.26998 L .54937 1.27369 L p .703 .547 .673 r
F P
0 g
s
.54937 1.27369 m .55127 1.27294 L .54824 1.269 L p .703 .547 .673 r
F P
0 g
s
.53058 1.2749 m .53053 1.28062 L .53468 1.28198 L p .699 .564 .7 r
F P
0 g
s
.53468 1.28198 m .53468 1.27705 L .53058 1.2749 L p .699 .564 .7 r
F P
0 g
s
.55535 1.28908 m .55625 1.28612 L .55962 1.28604 L p .645 .498 .668 r
F P
0 g
s
.56007 1.28355 m .55962 1.28604 L .55625 1.28612 L p .65 .515 .686 r
F P
0 g
s
.53495 1.2942 m .5351 1.2965 L .54185 1.29524 L p .695 .587 .731 r
F P
0 g
s
.54312 1.29615 m .54185 1.29524 L .5351 1.2965 L p .699 .618 .763 r
F P
0 g
s
.60909 1.09992 m .63968 1.11664 L .63024 1.12838 L p .734 .458 .51 r
F P
0 g
s
.63024 1.12838 m .60239 1.11164 L .60909 1.09992 L p .734 .458 .51 r
F P
0 g
s
.55242 1.29331 m .5506 1.29429 L .54442 1.29599 L p .614 .366 .518 r
F P
0 g
s
.54312 1.29615 m .54442 1.29599 L .5506 1.29429 L p .716 .711 .845 r
F P
0 g
s
.76413 .88292 m .73346 .86636 L .71599 .88306 L p .752 .39 .368 r
F P
0 g
s
.68499 .87507 m .71599 .88306 L .73346 .86636 L p .667 .291 .323 r
F P
0 g
s
.73666 .81905 m .78893 .81744 L .77236 .83505 L p .455 0 .046 r
F P
0 g
s
.77236 .83505 m .72297 .83654 L .73666 .81905 L p .455 0 .046 r
F P
0 g
s
.52161 1.28401 m .5188 1.29019 L .52387 1.2907 L p .711 .584 .712 r
F P
0 g
s
.52387 1.2907 m .52562 1.28519 L .52161 1.28401 L p .711 .584 .712 r
F P
0 g
s
.52562 1.28519 m .52387 1.2907 L .52938 1.29092 L p .702 .576 .71 r
F P
0 g
s
.53955 1.27119 m .53894 1.27485 L .54259 1.27716 L p .694 .553 .69 r
F P
0 g
s
.54259 1.27716 m .54349 1.27421 L .53955 1.27119 L p .694 .553 .69 r
F P
0 g
s
.52562 1.28519 m .52658 1.27897 L .52315 1.27703 L p .712 .58 .706 r
F P
0 g
s
.52658 1.27897 m .52562 1.28519 L .53013 1.28609 L p .703 .572 .705 r
F P
0 g
s
.53013 1.28609 m .53053 1.28062 L .52658 1.27897 L p .703 .572 .705 r
F P
0 g
s
.51229 1.2418 m .51682 1.25033 L .5173 1.25471 L p .58 .515 .746 r
F P
0 g
s
.56287 1.05842 m .57942 1.08777 L .57147 1.09933 L p .729 .543 .639 r
F P
0 g
s
.57147 1.09933 m .5579 1.07046 L .56287 1.05842 L p .729 .543 .639 r
F P
0 g
s
.52078 1.26669 m .52056 1.27479 L .52315 1.27703 L p .732 .592 .7 r
F P
0 g
s
.55142 1.22337 m .55643 1.24025 L .55418 1.24568 L p .769 .601 .67 r
F P
0 g
s
.55418 1.24568 m .54979 1.22951 L .55142 1.22337 L p .769 .601 .67 r
F P
0 g
s
.54641 1.26998 m .54479 1.27175 L .54768 1.27504 L p .698 .548 .68 r
F P
0 g
s
.54768 1.27504 m .54937 1.27369 L .54641 1.26998 L p .698 .548 .68 r
F P
0 g
s
.52938 1.29092 m .53013 1.28609 L .52562 1.28519 L p .702 .576 .71 r
F P
0 g
s
.53013 1.28609 m .52938 1.29092 L .53482 1.29086 L p .697 .571 .71 r
F P
0 g
s
.53495 1.2942 m .53482 1.29086 L .52938 1.29092 L p .697 .578 .717 r
F P
0 g
s
.50556 1.2944 m .49787 1.29859 L .50302 1.29869 L p .739 .667 .78 r
F P
0 g
s
.50302 1.29869 m .50939 1.29491 L .50556 1.2944 L p .739 .667 .78 r
F P
0 g
s
.52315 1.27703 m .52161 1.28401 L .52562 1.28519 L p .712 .58 .706 r
F P
0 g
s
.52161 1.28401 m .52315 1.27703 L .52056 1.27479 L p .725 .593 .708 r
F P
0 g
s
.55404 1.29153 m .55535 1.28908 L .55868 1.28829 L p .626 .456 .633 r
F P
0 g
s
.55962 1.28604 m .55868 1.28829 L .55535 1.28908 L p .645 .498 .668 r
F P
0 g
s
.51847 1.28253 m .51466 1.2894 L .5188 1.29019 L p .723 .598 .717 r
F P
0 g
s
.5188 1.29019 m .52161 1.28401 L .51847 1.28253 L p .723 .598 .717 r
F P
0 g
s
.53053 1.28062 m .53013 1.28609 L .53473 1.28669 L p .697 .567 .704 r
F P
0 g
s
.53473 1.28669 m .53468 1.28198 L .53053 1.28062 L p .697 .567 .704 r
F P
0 g
s
.53468 1.27705 m .53468 1.28198 L .53868 1.28303 L p .694 .56 .7 r
F P
0 g
s
.53868 1.28303 m .53864 1.27889 L .53468 1.27705 L p .694 .56 .7 r
F P
0 g
s
.4674 1.29742 m .45925 1.29233 L .46489 1.293 L p .787 .472 .454 r
F P
0 g
s
.46489 1.293 m .47231 1.29771 L .4674 1.29742 L p .787 .472 .454 r
F P
0 g
s
.53482 1.29086 m .53473 1.28669 L .53013 1.28609 L p .697 .571 .71 r
F P
0 g
s
.53482 1.29086 m .53495 1.2942 L .5407 1.2933 L p .693 .575 .718 r
F P
0 g
s
.54185 1.29524 m .5407 1.2933 L .53495 1.2942 L p .695 .587 .731 r
F P
0 g
s
.52056 1.27479 m .51847 1.28253 L .52161 1.28401 L p .725 .593 .708 r
F P
0 g
s
.48932 1.22045 m .49851 1.2234 L .49759 1.22967 L p .684 .614 .772 r
F P
0 g
s
.49759 1.22967 m .48842 1.2269 L .48932 1.22045 L p .684 .614 .772 r
F P
0 g
s
.50573 1.23476 m .49759 1.22967 L .49851 1.2234 L p .671 .59 .756 r
F P
0 g
s
.54185 1.29524 m .54312 1.29615 L .54869 1.29441 L p .699 .61 .753 r
F P
0 g
s
.5506 1.29429 m .54869 1.29441 L .54312 1.29615 L p .716 .711 .845 r
F P
0 g
s
.49851 1.2234 m .50667 1.22886 L .50573 1.23476 L p .671 .59 .756 r
F P
0 g
s
.51229 1.2418 m .50573 1.23476 L .50667 1.22886 L p .66 .569 .742 r
F P
0 g
s
.47972 1.22019 m .48932 1.22045 L .48842 1.2269 L p .703 .645 .79 r
F P
0 g
s
.48842 1.2269 m .47884 1.22661 L .47972 1.22019 L p .703 .645 .79 r
F P
0 g
s
.53894 1.27485 m .53864 1.27889 L .54215 1.28043 L p .692 .555 .695 r
F P
0 g
s
.54215 1.28043 m .54259 1.27716 L .53894 1.27485 L p .692 .555 .695 r
F P
0 g
s
.54479 1.27175 m .54349 1.27421 L .54632 1.2769 L p .695 .55 .685 r
F P
0 g
s
.54632 1.2769 m .54768 1.27504 L .54479 1.27175 L p .695 .55 .685 r
F P
0 g
s
.55242 1.29331 m .55404 1.29153 L .55732 1.29016 L p .498 .233 .443 r
F P
0 g
s
.55868 1.28829 m .55732 1.29016 L .55404 1.29153 L p .626 .456 .633 r
F P
0 g
s
.53468 1.28198 m .53473 1.28669 L .53906 1.28699 L p .693 .563 .704 r
F P
0 g
s
.53906 1.28699 m .53868 1.28303 L .53468 1.28198 L p .693 .563 .704 r
F P
0 g
s
.53473 1.28669 m .53482 1.29086 L .53975 1.29049 L p .692 .568 .71 r
F P
0 g
s
.5407 1.2933 m .53975 1.29049 L .53482 1.29086 L p .693 .575 .718 r
F P
0 g
s
.51192 1.28833 m .50556 1.2944 L .50939 1.29491 L p .741 .637 .745 r
F P
0 g
s
.50939 1.29491 m .51466 1.2894 L .51192 1.28833 L p .741 .637 .745 r
F P
0 g
s
.64042 1.14981 m .62353 1.14572 L .61295 1.15576 L p .703 .455 .546 r
F P
0 g
s
.59776 1.15611 m .61295 1.15576 L .62353 1.14572 L p .657 .401 .521 r
F P
0 g
s
.50667 1.22886 m .51325 1.23645 L .51229 1.2418 L p .66 .569 .742 r
F P
0 g
s
.51682 1.25033 m .51229 1.2418 L .51325 1.23645 L p .645 .548 .73 r
F P
0 g
s
.53975 1.29049 m .53906 1.28699 L .53473 1.28669 L p .692 .568 .71 r
F P
0 g
s
.47037 1.22266 m .47972 1.22019 L .47884 1.22661 L p .73 .688 .81 r
F P
0 g
s
.47884 1.22661 m .46952 1.22885 L .47037 1.22266 L p .73 .688 .81 r
F P
0 g
s
.5407 1.2933 m .54185 1.29524 L .54684 1.29366 L p .694 .586 .73 r
F P
0 g
s
.54869 1.29441 m .54684 1.29366 L .54185 1.29524 L p .699 .61 .753 r
F P
0 g
s
.54528 1.01799 m .54585 1.05061 L .54036 1.0626 L p .737 .593 .695 r
F P
0 g
s
.54036 1.0626 m .54005 1.03075 L .54528 1.01799 L p .737 .593 .695 r
F P
0 g
s
.54942 .98641 m .54528 1.01799 L .54005 1.03075 L p .796 .669 .723 r
F P
0 g
s
.55839 1.27638 m .55694 1.27463 L .5566 1.27773 L p .742 .572 .663 r
F P
0 g
s
.55694 1.27463 m .55519 1.27343 L .55517 1.2769 L p .726 .561 .667 r
F P
0 g
s
.55517 1.2769 m .5566 1.27773 L .55694 1.27463 L p .726 .561 .667 r
F P
0 g
s
.55519 1.27343 m .55326 1.27285 L .55358 1.2765 L p .716 .555 .671 r
F P
0 g
s
.55358 1.2765 m .55517 1.2769 L .55519 1.27343 L p .716 .555 .671 r
F P
0 g
s
.54349 1.27421 m .54259 1.27716 L .54539 1.27916 L p .692 .551 .69 r
F P
0 g
s
.54539 1.27916 m .54632 1.2769 L .54349 1.27421 L p .692 .551 .69 r
F P
0 g
s
.53864 1.27889 m .53868 1.28303 L .54221 1.28377 L p .69 .557 .7 r
F P
0 g
s
.54221 1.28377 m .54215 1.28043 L .53864 1.27889 L p .69 .557 .7 r
F P
0 g
s
.55944 1.27855 m .55839 1.27638 L .55778 1.27894 L p .766 .591 .661 r
F P
0 g
s
.5566 1.27773 m .55778 1.27894 L .55839 1.27638 L p .742 .572 .663 r
F P
0 g
s
.51466 1.2894 m .51847 1.28253 L .51651 1.28077 L p .743 .623 .726 r
F P
0 g
s
.55732 1.29016 m .55564 1.29151 L .55242 1.29331 L p .498 .233 .443 r
F P
0 g
s
.5506 1.29429 m .55242 1.29331 L .55564 1.29151 L p .759 .764 .857 r
F P
0 g
s
.51651 1.28077 m .51192 1.28833 L .51466 1.2894 L p .743 .623 .726 r
F P
0 g
s
.55326 1.27285 m .55127 1.27294 L .55196 1.27656 L p .708 .552 .675 r
F P
0 g
s
.55196 1.27656 m .55358 1.2765 L .55326 1.27285 L p .708 .552 .675 r
F P
0 g
s
.56001 1.28099 m .55944 1.27855 L .55864 1.28044 L p .797 .623 .665 r
F P
0 g
s
.55778 1.27894 m .55864 1.28044 L .55944 1.27855 L p .766 .591 .661 r
F P
0 g
s
.47658 1.3002 m .4674 1.29742 L .47231 1.29771 L p .697 .259 .205 r
F P
0 g
s
.47231 1.29771 m .48066 1.3003 L .47658 1.3002 L p .697 .259 .205 r
F P
0 g
s
.55127 1.27294 m .54937 1.27369 L .5504 1.27707 L p .702 .551 .679 r
F P
0 g
s
.5504 1.27707 m .55196 1.27656 L .55127 1.27294 L p .702 .551 .679 r
F P
0 g
s
.56007 1.28355 m .56001 1.28099 L .55911 1.28212 L p .822 .662 .684 r
F P
0 g
s
.55864 1.28044 m .55911 1.28212 L .56001 1.28099 L p .797 .623 .665 r
F P
0 g
s
.53975 1.29049 m .5407 1.2933 L .54516 1.29209 L p .691 .574 .719 r
F P
0 g
s
.54684 1.29366 m .54516 1.29209 L .5407 1.2933 L p .694 .586 .73 r
F P
0 g
s
.51325 1.23645 m .51779 1.24565 L .51682 1.25033 L p .645 .548 .73 r
F P
0 g
s
.51901 1.25975 m .51682 1.25033 L .51779 1.24565 L p .615 .513 .715 r
F P
0 g
s
.53868 1.28303 m .53906 1.28699 L .54276 1.28697 L p .689 .561 .705 r
F P
0 g
s
.54276 1.28697 m .54221 1.28377 L .53868 1.28303 L p .689 .561 .705 r
F P
0 g
s
.54259 1.27716 m .54215 1.28043 L .54493 1.28165 L p .689 .552 .694 r
F P
0 g
s
.54493 1.28165 m .54539 1.27916 L .54259 1.27716 L p .689 .552 .694 r
F P
0 g
s
.54869 1.29441 m .5506 1.29429 L .55373 1.29226 L p .714 .634 .767 r
F P
0 g
s
.55564 1.29151 m .55373 1.29226 L .5506 1.29429 L p .759 .764 .857 r
F P
0 g
s
.54937 1.27369 m .54768 1.27504 L .54902 1.278 L p .698 .549 .682 r
F P
0 g
s
.54902 1.278 m .5504 1.27707 L .54937 1.27369 L p .698 .549 .682 r
F P
0 g
s
.4619 1.22768 m .47037 1.22266 L .46952 1.22885 L p .773 .748 .83 r
F P
0 g
s
.46952 1.22885 m .46108 1.23345 L .4619 1.22768 L p .773 .748 .83 r
F P
0 g
s
.525 .81095 m .52823 .76542 L .53527 .7473 L p .499 .353 .614 r
F P
0 g
s
.52823 .76542 m .525 .81095 L .51499 .82802 L p .332 .197 .543 r
F P
0 g
s
.53906 1.28699 m .53975 1.29049 L .54377 1.28981 L p .689 .566 .711 r
F P
0 g
s
.54516 1.29209 m .54377 1.28981 L .53975 1.29049 L p .691 .574 .719 r
F P
0 g
s
.55962 1.28604 m .56007 1.28355 L .55916 1.28389 L p .82 .685 .714 r
F P
0 g
s
.55911 1.28212 m .55916 1.28389 L .56007 1.28355 L p .822 .662 .684 r
F P
0 g
s
.54377 1.28981 m .54276 1.28697 L .53906 1.28699 L p .689 .566 .711 r
F P
0 g
s
.51287 1.01274 m .52026 1.03841 L .51481 1.05049 L p .674 .583 .745 r
F P
0 g
s
.51481 1.05049 m .50467 1.02547 L .51287 1.01274 L p .674 .583 .745 r
F P
0 g
s
.54768 1.27504 m .54632 1.2769 L .54791 1.27928 L p .693 .549 .685 r
F P
0 g
s
.54791 1.27928 m .54902 1.278 L .54768 1.27504 L p .693 .549 .685 r
F P
0 g
s
.40466 .9666 m .44324 .94538 L .45015 .93104 L p .798 .921 .92 r
F P
0 g
s
.55868 1.28829 m .55962 1.28604 L .55879 1.2856 L p .792 .677 .737 r
F P
0 g
s
.55916 1.28389 m .55879 1.2856 L .55962 1.28604 L p .82 .685 .714 r
F P
0 g
s
.54215 1.28043 m .54221 1.28377 L .54499 1.2842 L p .686 .555 .7 r
F P
0 g
s
.54499 1.2842 m .54493 1.28165 L .54215 1.28043 L p .686 .555 .7 r
F P
0 g
s
.54684 1.29366 m .54869 1.29441 L .55175 1.29234 L p .7 .6 .741 r
F P
0 g
s
.55373 1.29226 m .55175 1.29234 L .54869 1.29441 L p .714 .634 .767 r
F P
0 g
s
.48616 1.30048 m .47658 1.3002 L .48066 1.3003 L p 0 .571 .723 r
F P
0 g
s
.48066 1.3003 m .48938 1.3006 L .48616 1.30048 L p 0 .571 .723 r
F P
0 g
s
.54932 1.2127 m .54979 1.22951 L .54693 1.23531 L p .768 .618 .694 r
F P
0 g
s
.54693 1.23531 m .54656 1.21928 L .54932 1.2127 L p .768 .618 .694 r
F P
0 g
s
.54632 1.2769 m .54539 1.27916 L .54714 1.28084 L p .689 .548 .689 r
F P
0 g
s
.54714 1.28084 m .54791 1.27928 L .54632 1.2769 L p .689 .548 .689 r
F P
0 g
s
.55732 1.29016 m .55868 1.28829 L .55802 1.28716 L p .761 .654 .743 r
F P
0 g
s
.55879 1.2856 m .55802 1.28716 L .55868 1.28829 L p .792 .677 .737 r
F P
0 g
s
.55643 1.24025 m .56556 1.25527 L .56217 1.2601 L p .752 .573 .654 r
F P
0 g
s
.56217 1.2601 m .55418 1.24568 L .55643 1.24025 L p .752 .573 .654 r
F P
0 g
s
.54221 1.28377 m .54276 1.28697 L .54557 1.28665 L p .686 .559 .706 r
F P
0 g
s
.54557 1.28665 m .54499 1.2842 L .54221 1.28377 L p .686 .559 .706 r
F P
0 g
s
.51779 1.24565 m .51998 1.25584 L .51901 1.25975 L p .615 .513 .715 r
F P
0 g
s
.54516 1.29209 m .54684 1.29366 L .54981 1.29177 L p .694 .584 .728 r
F P
0 g
s
.55175 1.29234 m .54981 1.29177 L .54684 1.29366 L p .7 .6 .741 r
F P
0 g
s
.45489 1.23492 m .4619 1.22768 L .46108 1.23345 L p .843 .838 .841 r
F P
0 g
s
.46108 1.23345 m .45409 1.24009 L .45489 1.23492 L p .843 .838 .841 r
F P
0 g
s
.54276 1.28697 m .54377 1.28981 L .54662 1.28882 L p .687 .565 .712 r
F P
0 g
s
.54662 1.28882 m .54557 1.28665 L .54276 1.28697 L p .687 .565 .712 r
F P
0 g
s
.54539 1.27916 m .54493 1.28165 L .54677 1.28255 L p .685 .549 .694 r
F P
0 g
s
.54677 1.28255 m .54714 1.28084 L .54539 1.27916 L p .685 .549 .694 r
F P
0 g
s
.54377 1.28981 m .54516 1.29209 L .54806 1.29056 L p .689 .573 .719 r
F P
0 g
s
.54981 1.29177 m .54806 1.29056 L .54516 1.29209 L p .694 .584 .728 r
F P
0 g
s
.53527 .7473 m .53599 .79353 L .525 .81095 L p .499 .353 .614 r
F P
0 g
s
.55564 1.29151 m .55732 1.29016 L .55691 1.28844 L p .737 .631 .741 r
F P
0 g
s
.55802 1.28716 m .55691 1.28844 L .55732 1.29016 L p .761 .654 .743 r
F P
0 g
s
.54806 1.29056 m .54662 1.28882 L .54377 1.28981 L p .689 .573 .719 r
F P
0 g
s
.5566 1.27773 m .55517 1.2769 L .55251 1.28003 L p .713 .567 .688 r
F P
0 g
s
.55517 1.2769 m .55358 1.2765 L .55158 1.27981 L p .71 .564 .687 r
F P
0 g
s
.55158 1.27981 m .55251 1.28003 L .55517 1.2769 L p .71 .564 .687 r
F P
0 g
s
.55358 1.2765 m .55196 1.27656 L .55062 1.27984 L p .707 .56 .686 r
F P
0 g
s
.55062 1.27984 m .55158 1.27981 L .55358 1.2765 L p .707 .56 .686 r
F P
0 g
s
.55778 1.27894 m .5566 1.27773 L .55336 1.28049 L p .715 .571 .69 r
F P
0 g
s
.55251 1.28003 m .55336 1.28049 L .5566 1.27773 L p .713 .567 .688 r
F P
0 g
s
.49549 1.29824 m .48616 1.30048 L .48938 1.3006 L p .696 .838 .951 r
F P
0 g
s
.48938 1.3006 m .49787 1.29859 L .49549 1.29824 L p .696 .838 .951 r
F P
0 g
s
.55196 1.27656 m .5504 1.27707 L .5497 1.28012 L p .703 .557 .686 r
F P
0 g
s
.5497 1.28012 m .55062 1.27984 L .55196 1.27656 L p .703 .557 .686 r
F P
0 g
s
.55864 1.28044 m .55778 1.27894 L .55406 1.28115 L p .718 .574 .693 r
F P
0 g
s
.55336 1.28049 m .55406 1.28115 L .55778 1.27894 L p .715 .571 .69 r
F P
0 g
s
.5504 1.27707 m .54902 1.278 L .54888 1.28063 L p .699 .553 .686 r
F P
0 g
s
.54888 1.28063 m .5497 1.28012 L .5504 1.27707 L p .699 .553 .686 r
F P
0 g
s
.52383 1.06692 m .52336 1.09639 L .51906 1.10693 L p .587 .47 .686 r
F P
0 g
s
.51288 1.1346 m .51906 1.10693 L .52336 1.09639 L p .342 .205 .546 r
F P
0 g
s
.54493 1.28165 m .54499 1.2842 L .54682 1.28432 L p .682 .551 .7 r
F P
0 g
s
.54682 1.28432 m .54677 1.28255 L .54493 1.28165 L p .682 .551 .7 r
F P
0 g
s
.55373 1.29226 m .55564 1.29151 L .55553 1.28937 L p .721 .614 .737 r
F P
0 g
s
.55691 1.28844 m .55553 1.28937 L .55564 1.29151 L p .737 .631 .741 r
F P
0 g
s
.55911 1.28212 m .55864 1.28044 L .55456 1.28198 L p .719 .578 .695 r
F P
0 g
s
.55406 1.28115 m .55456 1.28198 L .55864 1.28044 L p .718 .574 .693 r
F P
0 g
s
.54902 1.278 m .54791 1.27928 L .54822 1.28134 L p .693 .549 .686 r
F P
0 g
s
.54822 1.28134 m .54888 1.28063 L .54902 1.278 L p .693 .549 .686 r
F P
0 g
s
.50556 1.2944 m .51192 1.28833 L .51091 1.28698 L p .779 .704 .78 r
F P
0 g
s
.55916 1.28389 m .55911 1.28212 L .55484 1.28291 L p .72 .581 .699 r
F P
0 g
s
.55456 1.28198 m .55484 1.28291 L .55911 1.28212 L p .719 .578 .695 r
F P
0 g
s
.50393 1.29363 m .49549 1.29824 L .49787 1.29859 L p .768 .752 .839 r
F P
0 g
s
.49787 1.29859 m .50556 1.2944 L .50393 1.29363 L p .768 .752 .839 r
F P
0 g
s
.54499 1.2842 m .54557 1.28665 L .54729 1.286 L p .682 .557 .706 r
F P
0 g
s
.54729 1.286 m .54682 1.28432 L .54499 1.2842 L p .682 .557 .706 r
F P
0 g
s
.55175 1.29234 m .55373 1.29226 L .55397 1.28989 L p .71 .601 .732 r
F P
0 g
s
.55553 1.28937 m .55397 1.28989 L .55373 1.29226 L p .721 .614 .737 r
F P
0 g
s
.54791 1.27928 m .54714 1.28084 L .54777 1.28219 L p .686 .545 .688 r
F P
0 g
s
.54777 1.28219 m .54822 1.28134 L .54791 1.27928 L p .686 .545 .688 r
F P
0 g
s
.51091 1.28698 m .50393 1.29363 L .50556 1.2944 L p .779 .704 .78 r
F P
0 g
s
.55879 1.2856 m .55916 1.28389 L .55487 1.28388 L p .719 .583 .702 r
F P
0 g
s
.55484 1.28291 m .55487 1.28388 L .55916 1.28389 L p .72 .581 .699 r
F P
0 g
s
.664 .45909 m .62285 .48725 L .60875 .50805 L p 0 0 0 r
F P
0 g
s
.58101 .5462 m .60875 .50805 L p .61348 .50045 L 0 0 0 r
F P
0 g
s
.56842 .56697 m .58874 .52906 L .60875 .50805 L p .898 .986 .78 r
F P
0 g
s
.60875 .50805 m .58874 .52906 L p .60458 .51475 L .751 .993 .809 r
F P
0 g
s
.64352 .47957 m .60875 .50805 L p .60458 .51475 L .751 .993 .809 r
F P
0 g
s
.60875 .50805 m .62285 .48725 L p .61348 .50045 L 0 0 0 r
F P
0 g
s
.65572 1.15819 m .64042 1.14981 L .62777 1.15978 L p .738 .502 .572 r
F P
0 g
s
.61295 1.15576 m .62777 1.15978 L .64042 1.14981 L p .703 .455 .546 r
F P
0 g
s
.54557 1.28665 m .54662 1.28882 L .54815 1.2875 L p .685 .564 .713 r
F P
0 g
s
.54815 1.2875 m .54729 1.286 L .54557 1.28665 L p .685 .564 .713 r
F P
0 g
s
.52336 1.09639 m .52383 1.06692 L .52887 1.05522 L p .637 .511 .692 r
F P
0 g
s
.52701 1.02596 m .52887 1.05522 L .52383 1.06692 L p .688 .573 .721 r
F P
0 g
s
.52383 1.06692 m .52026 1.03841 L .52701 1.02596 L p .688 .573 .721 r
F P
0 g
s
.54714 1.28084 m .54677 1.28255 L .54755 1.28314 L p .677 .541 .692 r
F P
0 g
s
.54755 1.28314 m .54777 1.28219 L .54714 1.28084 L p .677 .541 .692 r
F P
0 g
s
.54981 1.29177 m .55175 1.29234 L .55235 1.28994 L p .702 .59 .728 r
F P
0 g
s
.55397 1.28989 m .55235 1.28994 L .55175 1.29234 L p .71 .601 .732 r
F P
0 g
s
.51015 1.22267 m .51608 1.23079 L .51325 1.23645 L p .676 .577 .737 r
F P
0 g
s
.51325 1.23645 m .50667 1.22886 L .51015 1.22267 L p .676 .577 .737 r
F P
0 g
s
.51779 1.24565 m .51325 1.23645 L .51608 1.23079 L p .668 .56 .724 r
F P
0 g
s
.51998 1.25584 m .51969 1.26632 L .51872 1.26944 L p .503 .4 .67 r
F P
0 g
s
.51608 1.23079 m .52017 1.24066 L .51779 1.24565 L p .668 .56 .724 r
F P
0 g
s
.51998 1.25584 m .51779 1.24565 L .52017 1.24066 L p .654 .539 .712 r
F P
0 g
s
.54662 1.28882 m .54806 1.29056 L .54933 1.28871 L p .689 .572 .718 r
F P
0 g
s
.54933 1.28871 m .54815 1.2875 L .54662 1.28882 L p .689 .572 .718 r
F P
0 g
s
.55802 1.28716 m .55879 1.2856 L .55465 1.28482 L p .718 .585 .705 r
F P
0 g
s
.55487 1.28388 m .55465 1.28482 L .55879 1.2856 L p .719 .583 .702 r
F P
0 g
s
.54806 1.29056 m .54981 1.29177 L .55077 1.28954 L p .695 .581 .723 r
F P
0 g
s
.55235 1.28994 m .55077 1.28954 L .54981 1.29177 L p .702 .59 .728 r
F P
0 g
s
.55077 1.28954 m .54933 1.28871 L .54806 1.29056 L p .695 .581 .723 r
F P
0 g
s
.54677 1.28255 m .54682 1.28432 L .54758 1.28411 L p .671 .543 .699 r
F P
0 g
s
.54758 1.28411 m .54755 1.28314 L .54677 1.28255 L p .671 .543 .699 r
F P
0 g
s
.93874 .53722 m .91094 .49302 L .88878 .50884 L p .91 .509 .208 r
F P
0 g
s
.85207 .47388 m .88878 .50884 L .91094 .49302 L p .814 .346 .106 r
F P
0 g
s
.44983 1.24388 m .45489 1.23492 L .45409 1.24009 L p .933 .948 .775 r
F P
0 g
s
.45409 1.24009 m .44904 1.24834 L .44983 1.24388 L p .933 .948 .775 r
F P
0 g
s
.55691 1.28844 m .55802 1.28716 L .5542 1.28568 L p .717 .586 .708 r
F P
0 g
s
.55465 1.28482 m .5542 1.28568 L .55802 1.28716 L p .718 .585 .705 r
F P
0 g
s
.54682 1.28432 m .54729 1.286 L .54786 1.28504 L p .674 .552 .708 r
F P
0 g
s
.54786 1.28504 m .54758 1.28411 L .54682 1.28432 L p .674 .552 .708 r
F P
0 g
s
.55553 1.28937 m .55691 1.28844 L .55354 1.28638 L p .714 .585 .71 r
F P
0 g
s
.5542 1.28568 m .55354 1.28638 L .55691 1.28844 L p .717 .586 .708 r
F P
0 g
s
.54729 1.286 m .54815 1.2875 L .54836 1.28586 L p .683 .563 .714 r
F P
0 g
s
.54836 1.28586 m .54786 1.28504 L .54729 1.286 L p .683 .563 .714 r
F P
0 g
s
.50279 1.21684 m .51015 1.22267 L .50667 1.22886 L p .684 .597 .752 r
F P
0 g
s
.50667 1.22886 m .49851 1.2234 L .50279 1.21684 L p .684 .597 .752 r
F P
0 g
s
.52065 .62894 m .52598 .67151 L .5171 .69097 L p .717 .636 .767 r
F P
0 g
s
.5171 .69097 m .50594 .64928 L .52065 .62894 L p .717 .636 .767 r
F P
0 g
s
.55397 1.28989 m .55553 1.28937 L .55272 1.2869 L p .711 .584 .712 r
F P
0 g
s
.55354 1.28638 m .55272 1.2869 L .55553 1.28937 L p .714 .585 .71 r
F P
0 g
s
.54815 1.2875 m .54933 1.28871 L .54906 1.28653 L p .692 .572 .715 r
F P
0 g
s
.54906 1.28653 m .54836 1.28586 L .54815 1.2875 L p .692 .572 .715 r
F P
0 g
s
.51969 1.26632 m .51998 1.25584 L .52215 1.25161 L p .621 .498 .692 r
F P
0 g
s
.52017 1.24066 m .52215 1.25161 L .51998 1.25584 L p .654 .539 .712 r
F P
0 g
s
.54979 1.22951 m .55418 1.24568 L .55038 1.25075 L p .738 .582 .68 r
F P
0 g
s
.55038 1.25075 m .54693 1.23531 L .54979 1.22951 L p .738 .582 .68 r
F P
0 g
s
.55235 1.28994 m .55397 1.28989 L .5518 1.28718 L p .708 .583 .713 r
F P
0 g
s
.55272 1.2869 m .5518 1.28718 L .55397 1.28989 L p .711 .584 .712 r
F P
0 g
s
.54933 1.28871 m .55077 1.28954 L .54991 1.28698 L p .699 .577 .715 r
F P
0 g
s
.54991 1.28698 m .54906 1.28653 L .54933 1.28871 L p .699 .577 .715 r
F P
0 g
s
.55077 1.28954 m .55235 1.28994 L .55084 1.28721 L p .704 .58 .714 r
F P
0 g
s
.5518 1.28718 m .55084 1.28721 L .55235 1.28994 L p .708 .583 .713 r
F P
0 g
s
.55084 1.28721 m .54991 1.28698 L .55077 1.28954 L p .704 .58 .714 r
F P
0 g
s
.56571 .73962 m .58702 .78264 L .57284 .80032 L p .678 .483 .615 r
F P
0 g
s
.57284 .80032 m .55726 .75784 L .56571 .73962 L p .678 .483 .615 r
F P
0 g
s
.36355 1.01497 m .38605 1.00149 L .37744 1.01388 L p .709 .976 .919 r
F P
0 g
s
.37744 1.01388 m .35383 1.02677 L .36355 1.01497 L p .709 .976 .919 r
F P
0 g
s
.44705 1.25395 m .44983 1.24388 L .44904 1.24834 L p .815 .825 .348 r
F P
0 g
s
.44904 1.24834 m .44627 1.25763 L .44705 1.25395 L p .815 .825 .348 r
F P
0 g
s
.4945 1.21371 m .50279 1.21684 L .49851 1.2234 L p .694 .624 .774 r
F P
0 g
s
.49851 1.2234 m .48932 1.22045 L .4945 1.21371 L p .694 .624 .774 r
F P
0 g
s
.56556 1.25527 m .57821 1.2674 L .57324 1.27179 L p .745 .548 .628 r
F P
0 g
s
.57324 1.27179 m .56217 1.2601 L .56556 1.25527 L p .745 .548 .628 r
F P
0 g
s
.54005 1.03075 m .54222 .99995 L .54942 .98641 L p .796 .669 .723 r
F P
0 g
s
.54005 1.03075 m .54036 1.0626 L .53445 1.07423 L p .704 .566 .696 r
F P
0 g
s
.53445 1.07423 m .53443 1.04317 L .54005 1.03075 L p .704 .566 .696 r
F P
0 g
s
.54222 .99995 m .54005 1.03075 L .53443 1.04317 L p .751 .628 .723 r
F P
0 g
s
.56309 1.17824 m .57032 1.16952 L p .56616 1.17498 L .883 .992 .771 r
F P
0 g
s
.55492 1.18975 m .56309 1.17824 L p .56616 1.17498 L .883 .992 .771 r
F P
0 g
s
.54846 1.19748 m .55396 1.18658 L .56309 1.17824 L p .809 .737 .784 r
F P
0 g
s
.56309 1.17824 m .55396 1.18658 L p .56068 1.18081 L .831 .861 .868 r
F P
0 g
s
.57325 1.17001 m .56309 1.17824 L p .56068 1.18081 L .831 .861 .868 r
F P
0 g
s
.58322 1.16081 m .57032 1.16952 L .56309 1.17824 L p 0 0 0 r
F P
0 g
s
.52215 1.25161 m .52189 1.26288 L .51969 1.26632 L p .621 .498 .692 r
F P
0 g
s
.51692 1.27638 m .51969 1.26632 L .52189 1.26288 L p .547 .405 .639 r
F P
0 g
s
.54656 1.21928 m .54693 1.23531 L .54309 1.24077 L p .736 .593 .697 r
F P
0 g
s
.54309 1.24077 m .54284 1.22553 L .54656 1.21928 L p .736 .593 .697 r
F P
0 g
s
.54928 1.20374 m .54656 1.21928 L .54284 1.22553 L p .762 .63 .714 r
F P
0 g
s
.52065 1.22482 m .52383 1.23536 L .52017 1.24066 L p .682 .569 .722 r
F P
0 g
s
.52017 1.24066 m .51608 1.23079 L .52065 1.22482 L p .682 .569 .722 r
F P
0 g
s
.52215 1.25161 m .52017 1.24066 L .52383 1.23536 L p .672 .552 .71 r
F P
0 g
s
.55492 1.18975 m .54928 1.20374 L .54467 1.21077 L p .787 .676 .741 r
F P
0 g
s
.54284 1.22553 m .54467 1.21077 L .54928 1.20374 L p .762 .63 .714 r
F P
0 g
s
.51187 1.28534 m .51484 1.2834 L p .51098 1.28695 L .374 .1 .363 r
F P
0 g
s
.50393 1.29363 m .51091 1.28698 L .51187 1.28534 L p .864 .937 .868 r
F P
0 g
s
.50487 1.29258 m .50393 1.29363 L p .50611 1.29135 L .864 .937 .868 r
F P
0 g
s
.51187 1.28534 m .50487 1.29258 L p .50611 1.29135 L .864 .937 .868 r
F P
0 g
s
.50487 1.29258 m .50853 1.29124 L p .50362 1.29382 L .321 0 .146 r
F P
0 g
s
.49549 1.29824 m .50393 1.29363 L .50487 1.29258 L p .744 .964 .937 r
F P
0 g
s
.49641 1.29762 m .49549 1.29824 L p .49716 1.29723 L .744 .964 .937 r
F P
0 g
s
.50487 1.29258 m .49641 1.29762 L p .49716 1.29723 L .744 .964 .937 r
F P
0 g
s
.49641 1.29762 m .5009 1.29672 L p .4957 1.29799 L .074 0 0 r
F P
0 g
s
.48616 1.30048 m .49549 1.29824 L .49641 1.29762 L p .521 .899 .959 r
F P
0 g
s
.49641 1.29762 m .48706 1.3001 L .48616 1.30048 L p .521 .899 .959 r
F P
0 g
s
.48774 1.29993 m .49641 1.29762 L .4957 1.29799 L .074 0 0 r
F
.48761 1.29995 m .49641 1.29762 L p .48774 1.29993 L F P
0 g
s
.5009 1.29672 m .49246 1.29946 L .48706 1.3001 L p .074 0 0 r
F P
0 g
s
.48378 1.29926 m .47745 1.29986 L p .49048 1.29951 L .439 .808 .994 r
F P
0 g
s
.49246 1.29946 m .49103 1.29942 L p .49048 1.29951 L .439 .808 .994 r
F P
0 g
s
.49103 1.29942 m .48378 1.29926 L p .49048 1.29951 L .439 .808 .994 r
F P
0 g
s
.48706 1.3001 m .48761 1.29995 L p .48774 1.29993 L .074 0 0 r
F P
0 g
s
.49641 1.29762 m .50487 1.29258 L p .50362 1.29382 L .321 0 .146 r
F P
0 g
s
.50487 1.29258 m .51187 1.28534 L p .51098 1.28695 L .374 .1 .363 r
F P
0 g
s
.51187 1.28534 m .51091 1.28698 L p .51271 1.2838 L .929 .925 .785 r
F P
0 g
s
.51692 1.27638 m .51187 1.28534 L p .51271 1.2838 L .929 .925 .785 r
F P
0 g
s
.54585 1.05061 m .55111 1.08213 L .54312 1.09342 L p .682 .527 .669 r
F P
0 g
s
.54312 1.09342 m .54036 1.0626 L .54585 1.05061 L p .682 .527 .669 r
F P
0 g
s
.52189 1.26288 m .52215 1.25161 L .52537 1.24706 L p .654 .526 .695 r
F P
0 g
s
.52383 1.23536 m .52537 1.24706 L .52215 1.25161 L p .672 .552 .71 r
F P
0 g
s
.44674 1.26444 m .44705 1.25395 L .44627 1.25763 L p 0 0 .284 r
F P
0 g
s
.44627 1.25763 m .44597 1.26731 L .44674 1.26444 L p 0 0 .284 r
F P
0 g
s
.52863 1.0855 m .52887 1.05522 L .53443 1.04317 L p .673 .54 .695 r
F P
0 g
s
.53453 1.01314 m .53443 1.04317 L .52887 1.05522 L p .718 .598 .722 r
F P
0 g
s
.52887 1.05522 m .52701 1.02596 L .53453 1.01314 L p .718 .598 .722 r
F P
0 g
s
.52887 1.05522 m .52863 1.0855 L .52336 1.09639 L p .637 .511 .692 r
F P
0 g
s
.66838 1.1703 m .65572 1.15819 L .64119 1.16789 L p .765 .544 .599 r
F P
0 g
s
.62777 1.15978 m .64119 1.16789 L .65572 1.15819 L p .738 .502 .572 r
F P
0 g
s
.51605 1.21616 m .52065 1.22482 L .51608 1.23079 L p .688 .585 .735 r
F P
0 g
s
.51608 1.23079 m .51015 1.22267 L .51605 1.21616 L p .688 .585 .735 r
F P
0 g
s
.62305 .88574 m .59629 .9036 L .58531 .9192 L p 0 0 0 r
F P
0 g
s
.56766 .9433 m .58531 .9192 L p .58959 .91289 L 0 0 0 r
F P
0 g
s
.55801 .95794 m .57051 .93446 L .58531 .9192 L p .903 .931 .822 r
F P
0 g
s
.58531 .9192 m .57051 .93446 L p .58165 .92458 L .792 .998 .851 r
F P
0 g
s
.60731 .90182 m .58531 .9192 L p .58165 .92458 L .792 .998 .851 r
F P
0 g
s
.58531 .9192 m .59629 .9036 L p .58959 .91289 L 0 0 0 r
F P
0 g
s
.48583 1.21349 m .4945 1.21371 L .48932 1.22045 L p .711 .668 .806 r
F P
0 g
s
.48932 1.22045 m .47972 1.22019 L .48583 1.21349 L p .711 .668 .806 r
F P
0 g
s
.53443 1.04317 m .53453 1.01314 L .54222 .99995 L p .751 .628 .723 r
F P
0 g
s
.53443 1.04317 m .53445 1.07423 L .52863 1.0855 L p .673 .54 .695 r
F P
0 g
s
.52189 1.26288 m .51939 1.27373 L .51692 1.27638 L p .547 .405 .639 r
F P
0 g
s
.44894 1.27464 m .44674 1.26444 L .44597 1.26731 L p 0 .061 .588 r
F P
0 g
s
.44597 1.26731 m .44816 1.27675 L .44894 1.27464 L p 0 .061 .588 r
F P
0 g
s
.55801 .95794 m .54942 .98641 L .54222 .99995 L p .823 .729 .76 r
F P
0 g
s
.47163 .9901 m .48996 1.00488 L .48191 1.01774 L p .655 .631 .814 r
F P
0 g
s
.48191 1.01774 m .46043 1.00328 L .47163 .9901 L p .655 .631 .814 r
F P
0 g
s
.52537 1.24706 m .52516 1.25913 L .52189 1.26288 L p .654 .526 .695 r
F P
0 g
s
.51939 1.27373 m .52189 1.26288 L .52516 1.25913 L p .625 .483 .669 r
F P
0 g
s
.54693 1.23531 m .55038 1.25075 L .54542 1.25547 L p .718 .567 .684 r
F P
0 g
s
.54542 1.25547 m .54309 1.24077 L .54693 1.23531 L p .718 .567 .684 r
F P
0 g
s
.55418 1.24568 m .56217 1.2601 L .55668 1.26455 L p .728 .561 .665 r
F P
0 g
s
.55668 1.26455 m .55038 1.25075 L .55418 1.24568 L p .728 .561 .665 r
F P
0 g
s
.5011 .40453 m .47738 .45616 L .46447 .48041 L p .954 .951 .699 r
F P
0 g
s
.63968 1.11664 m .67351 1.12484 L .66111 1.13684 L p .68 .347 .405 r
F P
0 g
s
.66111 1.13684 m .63024 1.12838 L .63968 1.11664 L p .68 .347 .405 r
F P
0 g
s
.52516 1.25913 m .52537 1.24706 L .52938 1.24219 L p .672 .541 .697 r
F P
0 g
s
.5285 1.22972 m .52938 1.24219 L .52537 1.24706 L p .686 .562 .71 r
F P
0 g
s
.52537 1.24706 m .52383 1.23536 L .5285 1.22972 L p .686 .562 .71 r
F P
0 g
s
.78956 .9069 m .76413 .88292 L .74417 .89948 L p .822 .487 .422 r
F P
0 g
s
.71599 .88306 m .74417 .89948 L .76413 .88292 L p .752 .39 .368 r
F P
0 g
s
.52667 1.21851 m .5285 1.22972 L .52383 1.23536 L p .694 .578 .721 r
F P
0 g
s
.52383 1.23536 m .52065 1.22482 L .52667 1.21851 L p .694 .578 .721 r
F P
0 g
s
.51886 1.12489 m .52336 1.09639 L .52863 1.0855 L p .525 .382 .628 r
F P
0 g
s
.52336 1.09639 m .51886 1.12489 L .51288 1.1346 L p .342 .205 .546 r
F P
0 g
s
.54284 1.22553 m .54309 1.24077 L .53862 1.24588 L p .716 .577 .698 r
F P
0 g
s
.53862 1.24588 m .53849 1.23142 L .54284 1.22553 L p .716 .577 .698 r
F P
0 g
s
.54467 1.21077 m .54284 1.22553 L .53849 1.23142 L p .732 .602 .712 r
F P
0 g
s
.54467 1.21077 m .54846 1.19748 L .55492 1.18975 L p .787 .676 .741 r
F P
0 g
s
.45349 1.28386 m .44894 1.27464 L .44816 1.27675 L p .084 .252 .739 r
F P
0 g
s
.44816 1.27675 m .45269 1.28528 L .45349 1.28386 L p .084 .252 .739 r
F P
0 g
s
.51034 1.20996 m .51605 1.21616 L .51015 1.22267 L p .694 .605 .752 r
F P
0 g
s
.51015 1.22267 m .50279 1.21684 L .51034 1.20996 L p .694 .605 .752 r
F P
0 g
s
.57821 1.2674 m .5935 1.27582 L .58663 1.27996 L p .734 .508 .585 r
F P
0 g
s
.58663 1.27996 m .57324 1.27179 L .57821 1.2674 L p .734 .508 .585 r
F P
0 g
s
.52316 .99962 m .52701 1.02596 L .52026 1.03841 L p .703 .606 .745 r
F P
0 g
s
.52026 1.03841 m .51287 1.01274 L .52316 .99962 L p .703 .606 .745 r
F P
0 g
s
.54846 1.19748 m .54467 1.21077 L .53936 1.21743 L p .745 .628 .729 r
F P
0 g
s
.53849 1.23142 m .53936 1.21743 L .54467 1.21077 L p .732 .602 .712 r
F P
0 g
s
.59704 .76478 m .6323 .80004 L .61674 .81756 L p .719 .469 .547 r
F P
0 g
s
.61674 .81756 m .58702 .78264 L .59704 .76478 L p .719 .469 .547 r
F P
0 g
s
.47658 1.3002 m .48616 1.30048 L .48706 1.3001 L p .314 .766 .955 r
F P
0 g
s
.46007 1.29145 m .45349 1.28386 L .45269 1.28528 L p .137 .387 .84 r
F P
0 g
s
.45269 1.28528 m .45925 1.29233 L .46007 1.29145 L p .137 .387 .84 r
F P
0 g
s
.47739 1.2162 m .48583 1.21349 L .47972 1.22019 L p .745 .762 .867 r
F P
0 g
s
.47972 1.22019 m .47037 1.22266 L .47739 1.2162 L p .745 .762 .867 r
F P
0 g
s
.52927 1.25505 m .52938 1.24219 L .53388 1.23698 L p .686 .553 .697 r
F P
0 g
s
.53382 1.22375 m .53388 1.23698 L .52938 1.24219 L p .699 .573 .71 r
F P
0 g
s
.52938 1.24219 m .5285 1.22972 L .53382 1.22375 L p .699 .573 .71 r
F P
0 g
s
.52938 1.24219 m .52927 1.25505 L .52516 1.25913 L p .672 .541 .697 r
F P
0 g
s
.48706 1.3001 m .47745 1.29986 L .47658 1.3002 L p .314 .766 .955 r
F P
0 g
s
.4674 1.29742 m .47658 1.3002 L .47745 1.29986 L p .205 .635 .95 r
F P
0 g
s
.51939 1.27373 m .51484 1.2834 L .51187 1.28534 L p .438 .244 .519 r
F P
0 g
s
.46825 1.29691 m .46007 1.29145 L .45925 1.29233 L p .164 .51 .913 r
F P
0 g
s
.45925 1.29233 m .4674 1.29742 L .46825 1.29691 L p .164 .51 .913 r
F P
0 g
s
.52323 1.27076 m .52516 1.25913 L .52927 1.25505 L p .654 .512 .678 r
F P
0 g
s
.52516 1.25913 m .52323 1.27076 L .51939 1.27373 L p .625 .483 .669 r
F P
0 g
s
.53849 1.23142 m .53862 1.24588 L .53387 1.25064 L p .7 .564 .698 r
F P
0 g
s
.53387 1.25064 m .53388 1.23698 L .53849 1.23142 L p .7 .564 .698 r
F P
0 g
s
.53936 1.21743 m .53849 1.23142 L .53388 1.23698 L p .713 .585 .71 r
F P
0 g
s
.52403 1.20931 m .52667 1.21851 L .52065 1.22482 L p .7 .594 .735 r
F P
0 g
s
.52065 1.22482 m .51605 1.21616 L .52403 1.20931 L p .7 .594 .735 r
F P
0 g
s
.59776 1.15611 m .58322 1.16081 L .57325 1.17001 L p .483 .164 .35 r
F P
0 g
s
.56309 1.17824 m .57325 1.17001 L p .58009 1.16352 L 0 0 0 r
F P
0 g
s
.56079 1.1788 m .57325 1.17001 L p .56956 1.17318 L .831 .861 .868 r
F P
0 g
s
.55396 1.18658 m .56079 1.1788 L p .56956 1.17318 L .831 .861 .868 r
F P
0 g
s
.5438 1.19454 m .54706 1.18718 L .56079 1.1788 L p .769 .712 .799 r
F P
0 g
s
.56079 1.1788 m .54706 1.18718 L p .55437 1.18292 L .777 .864 .914 r
F P
0 g
s
.56079 1.1788 m .55396 1.18658 L .5438 1.19454 L p .769 .712 .799 r
F P
0 g
s
.56849 1.17467 m .56079 1.1788 L p .55437 1.18292 L .777 .864 .914 r
F P
0 g
s
.57325 1.17001 m .56079 1.1788 L p .56843 1.17459 L 0 0 0 r
F P
0 g
s
.58471 1.16561 m .57325 1.17001 L p .56843 1.17459 L 0 0 0 r
F P
0 g
s
.57325 1.17001 m .58322 1.16081 L p .58009 1.16352 L 0 0 0 r
F P
0 g
s
.5337 1.21186 m .53382 1.22375 L .5285 1.22972 L p .707 .589 .722 r
F P
0 g
s
.5285 1.22972 m .52667 1.21851 L .5337 1.21186 L p .707 .589 .722 r
F P
0 g
s
.47745 1.29986 m .46825 1.29691 L .4674 1.29742 L p .205 .635 .95 r
F P
0 g
s
.67755 1.18531 m .66838 1.1703 L .65231 1.17956 L p .785 .583 .626 r
F P
0 g
s
.64119 1.16789 m .65231 1.17956 L .66838 1.1703 L p .765 .544 .599 r
F P
0 g
s
.53388 1.23698 m .53382 1.22375 L .53936 1.21743 L p .713 .585 .71 r
F P
0 g
s
.53388 1.23698 m .53387 1.25064 L .52927 1.25505 L p .686 .553 .697 r
F P
0 g
s
.51484 1.2834 m .51939 1.27373 L .52323 1.27076 L p .598 .432 .628 r
F P
0 g
s
.54551 .63186 m .55327 .58772 L .56842 .56697 L p .889 .817 .771 r
F P
0 g
s
.55327 .58772 m .54551 .63186 L .53572 .6518 L p .818 .735 .772 r
F P
0 g
s
.53936 1.21743 m .54117 1.20486 L .54846 1.19748 L p .745 .628 .729 r
F P
0 g
s
.54117 1.20486 m .53936 1.21743 L .53382 1.22375 L p .722 .604 .724 r
F P
0 g
s
.54309 1.24077 m .54542 1.25547 L .53973 1.25982 L p .702 .555 .685 r
F P
0 g
s
.53973 1.25982 m .53862 1.24588 L .54309 1.24077 L p .702 .555 .685 r
F P
0 g
s
.53382 1.22375 m .5337 1.21186 L .54117 1.20486 L p .722 .604 .724 r
F P
0 g
s
.54036 1.0626 m .54312 1.09342 L .53461 1.10431 L p .646 .496 .665 r
F P
0 g
s
.53461 1.10431 m .53445 1.07423 L .54036 1.0626 L p .646 .496 .665 r
F P
0 g
s
.3449 1.03439 m .36355 1.01497 L .35383 1.02677 L p .543 .859 .565 r
F P
0 g
s
.35383 1.02677 m .33428 1.0454 L .3449 1.03439 L p .543 .859 .565 r
F P
0 g
s
.52816 1.26746 m .52927 1.25505 L .53387 1.25064 L p .672 .53 .683 r
F P
0 g
s
.52927 1.25505 m .52816 1.26746 L .52323 1.27076 L p .654 .512 .678 r
F P
0 g
s
.5039 1.20666 m .51034 1.20996 L .50279 1.21684 L p .703 .636 .779 r
F P
0 g
s
.50279 1.21684 m .4945 1.21371 L .5039 1.20666 L p .703 .636 .779 r
F P
0 g
s
.52629 1.11479 m .52863 1.0855 L .53445 1.07423 L p .6 .453 .654 r
F P
0 g
s
.52863 1.0855 m .52629 1.11479 L .51886 1.12489 L p .525 .382 .628 r
F P
0 g
s
.46974 1.22166 m .47739 1.2162 L .47037 1.22266 L p .729 .982 .909 r
F P
0 g
s
.47037 1.22266 m .4619 1.22768 L .46974 1.22166 L p .729 .982 .909 r
F P
0 g
s
.5935 1.27582 m .6104 1.27994 L .60144 1.28404 L p .705 .431 .507 r
F P
0 g
s
.60144 1.28404 m .58663 1.27996 L .5935 1.27582 L p .705 .431 .507 r
F P
0 g
s
.53862 1.24588 m .53973 1.25982 L .53381 1.26382 L p .687 .543 .685 r
F P
0 g
s
.53381 1.26382 m .53387 1.25064 L .53862 1.24588 L p .687 .543 .685 r
F P
0 g
s
.53387 1.25064 m .53381 1.26382 L .52816 1.26746 L p .672 .53 .683 r
F P
0 g
s
.50214 .99163 m .51287 1.01274 L .50467 1.02547 L p .68 .613 .773 r
F P
0 g
s
.50467 1.02547 m .48996 1.00488 L .50214 .99163 L p .68 .613 .773 r
F P
0 g
s
.78893 .81744 m .83905 .80235 L .81999 .82002 L p .147 0 0 r
F P
0 g
s
.81999 .82002 m .77236 .83505 L .78893 .81744 L p .147 0 0 r
F P
0 g
s
.51484 1.2834 m .50853 1.29124 L .50487 1.29258 L p .374 .1 .363 r
F P
0 g
s
.55396 1.18658 m .54846 1.19748 L .54117 1.20486 L p .756 .658 .754 r
F P
0 g
s
.53353 1.20211 m .5337 1.21186 L .52667 1.21851 L p .712 .606 .737 r
F P
0 g
s
.52667 1.21851 m .52403 1.20931 L .53353 1.20211 L p .712 .606 .737 r
F P
0 g
s
.55038 1.25075 m .55668 1.26455 L .54965 1.26863 L p .71 .55 .67 r
F P
0 g
s
.54965 1.26863 m .54542 1.25547 L .55038 1.25075 L p .71 .55 .67 r
F P
0 g
s
.53445 1.07423 m .53461 1.10431 L .52629 1.11479 L p .6 .453 .654 r
F P
0 g
s
.42379 1.21411 m .44984 1.20773 L .45482 1.20107 L p .698 .871 .963 r
F P
0 g
s
.51969 1.28114 m .52323 1.27076 L .52816 1.26746 L p .641 .482 .653 r
F P
0 g
s
.52323 1.27076 m .51969 1.28114 L .51484 1.2834 L p .598 .432 .628 r
F P
0 g
s
.53687 .60839 m .53572 .6518 L .52598 .67151 L p .764 .679 .769 r
F P
0 g
s
.52598 .67151 m .52065 .62894 L .53687 .60839 L p .764 .679 .769 r
F P
0 g
s
.68259 1.20221 m .67755 1.18531 L .66037 1.194 L p .802 .619 .655 r
F P
0 g
s
.65231 1.17956 m .66037 1.194 L .67755 1.18531 L p .785 .583 .626 r
F P
0 g
s
.56217 1.2601 m .57324 1.27179 L .5654 1.27578 L p .723 .541 .644 r
F P
0 g
s
.5654 1.27578 m .55668 1.26455 L .56217 1.2601 L p .723 .541 .644 r
F P
0 g
s
.52075 1.20274 m .52403 1.20931 L .51605 1.21616 L p .705 .616 .754 r
F P
0 g
s
.51605 1.21616 m .51034 1.20996 L .52075 1.20274 L p .705 .616 .754 r
F P
0 g
s
.30261 .62554 m .34302 .61574 L .32878 .63808 L p .651 .906 .981 r
F P
0 g
s
.32878 .63808 m .28596 .64802 L .30261 .62554 L p .651 .906 .981 r
F P
0 g
s
.5579 1.07046 m .57147 1.09933 L .56073 1.11047 L p .696 .518 .643 r
F P
0 g
s
.56073 1.11047 m .55111 1.08213 L .5579 1.07046 L p .696 .518 .643 r
F P
0 g
s
.54222 .99995 m .54672 .97221 L .55801 .95794 L p .823 .729 .76 r
F P
0 g
s
.54672 .97221 m .54222 .99995 L .53453 1.01314 L p .771 .67 .752 r
F P
0 g
s
.53476 .98611 m .53453 1.01314 L .52701 1.02596 L p .734 .633 .747 r
F P
0 g
s
.52701 1.02596 m .52316 .99962 L .53476 .98611 L p .734 .633 .747 r
F P
0 g
s
.50853 1.29124 m .51484 1.2834 L .51969 1.28114 L p .584 .381 .572 r
F P
0 g
s
.5438 1.19454 m .54117 1.20486 L .5337 1.21186 L p .729 .625 .742 r
F P
0 g
s
.5337 1.21186 m .53353 1.20211 L .5438 1.19454 L p .729 .625 .742 r
F P
0 g
s
.54117 1.20486 m .5438 1.19454 L .55396 1.18658 L p .756 .658 .754 r
F P
0 g
s
.46341 1.22949 m .46974 1.22166 L .4619 1.22768 L p .297 .005 .288 r
F P
0 g
s
.4619 1.22768 m .45489 1.23492 L .46341 1.22949 L p .297 .005 .288 r
F P
0 g
s
.50853 1.29124 m .5009 1.29672 L .49641 1.29762 L p .321 0 .146 r
F P
0 g
s
.6104 1.27994 m .62775 1.27948 L .61665 1.28374 L p .602 .234 .316 r
F P
0 g
s
.61665 1.28374 m .60144 1.28404 L .6104 1.27994 L p .602 .234 .316 r
F P
0 g
s
.49717 1.20648 m .5039 1.20666 L .4945 1.21371 L p .72 .701 .831 r
F P
0 g
s
.4945 1.21371 m .48583 1.21349 L .49717 1.20648 L p .72 .701 .831 r
F P
0 g
s
.52613 1.27856 m .52816 1.26746 L .53381 1.26382 L p .663 .508 .664 r
F P
0 g
s
.52816 1.26746 m .52613 1.27856 L .51969 1.28114 L p .641 .482 .653 r
F P
0 g
s
.53453 1.01314 m .53476 .98611 L .54672 .97221 L p .771 .67 .752 r
F P
0 g
s
.54542 1.25547 m .54965 1.26863 L .54175 1.27231 L p .694 .539 .671 r
F P
0 g
s
.54175 1.27231 m .53973 1.25982 L .54542 1.25547 L p .694 .539 .671 r
F P
0 g
s
.53572 .6518 m .53687 .60839 L .55327 .58772 L p .818 .735 .772 r
F P
0 g
s
.61295 1.15576 m .59776 1.15611 L .58471 1.16561 L p .63 .363 .495 r
F P
0 g
s
.57325 1.17001 m .58471 1.16561 L .59776 1.15611 L p .483 .164 .35 r
F P
0 g
s
.2755 .07407 m .33295 .08149 L .31559 .11957 L p .62 .804 .97 r
F P
0 g
s
.31559 .11957 m .25445 .11549 L .2755 .07407 L p .62 .804 .97 r
F P
0 g
s
.68316 1.21983 m .68259 1.20221 L .66479 1.2102 L p .815 .655 .685 r
F P
0 g
s
.66037 1.194 m .66479 1.2102 L .68259 1.20221 L p .802 .619 .655 r
F P
0 g
s
.57942 1.08777 m .60239 1.11164 L .59033 1.12293 L p .722 .507 .599 r
F P
0 g
s
.59033 1.12293 m .57147 1.09933 L .57942 1.08777 L p .722 .507 .599 r
F P
0 g
s
.9541 .58716 m .93874 .53722 L .91553 .55293 L p .972 .66 .317 r
F P
0 g
s
.88878 .50884 m .91553 .55293 L .93874 .53722 L p .91 .509 .208 r
F P
0 g
s
.53368 1.27562 m .53381 1.26382 L .53973 1.25982 L p .68 .525 .669 r
F P
0 g
s
.53381 1.26382 m .53368 1.27562 L .52613 1.27856 L p .663 .508 .664 r
F P
0 g
s
.53973 1.25982 m .54175 1.27231 L .53368 1.27562 L p .68 .525 .669 r
F P
0 g
s
.67351 1.12484 m .70828 1.12394 L .69291 1.13639 L p .553 .151 .234 r
F P
0 g
s
.69291 1.13639 m .66111 1.13684 L .67351 1.12484 L p .553 .151 .234 r
F P
0 g
s
.80799 .93669 m .78956 .9069 L .76757 .92324 L p .878 .581 .481 r
F P
0 g
s
.74417 .89948 m .76757 .92324 L .78956 .9069 L p .822 .487 .422 r
F P
0 g
s
.53332 1.19516 m .53353 1.20211 L .52403 1.20931 L p .719 .631 .76 r
F P
0 g
s
.52403 1.20931 m .52075 1.20274 L .53332 1.19516 L p .719 .631 .76 r
F P
0 g
s
.45883 1.23917 m .46341 1.22949 L .45489 1.23492 L p .503 .317 .562 r
F P
0 g
s
.45489 1.23492 m .44983 1.24388 L .45883 1.23917 L p .503 .317 .562 r
F P
0 g
s
.62775 1.27948 m .64437 1.27448 L .63124 1.27909 L p 0 0 0 r
F P
0 g
s
.63124 1.27909 m .61665 1.28374 L .62775 1.27948 L p 0 0 0 r
F P
0 g
s
.47745 1.29986 m .48706 1.3001 L .49246 1.29946 L p .439 .808 .994 r
F P
0 g
s
.5148 1.28959 m .51969 1.28114 L .52613 1.27856 L p .634 .451 .618 r
F P
0 g
s
.51969 1.28114 m .5148 1.28959 L .50853 1.29124 L p .584 .381 .572 r
F P
0 g
s
.33138 1.05843 m .3449 1.03439 L .33428 1.0454 L p 0 0 0 r
F P
0 g
s
.33428 1.0454 m .32013 1.06851 L .33138 1.05843 L p 0 0 0 r
F P
0 g
s
.51706 1.19926 m .52075 1.20274 L .51034 1.20996 L p .714 .654 .788 r
F P
0 g
s
.51034 1.20996 m .5039 1.20666 L .51706 1.19926 L p .714 .654 .788 r
F P
0 g
s
.5009 1.29672 m .50853 1.29124 L .5148 1.28959 L p .564 .3 .473 r
F P
0 g
s
.67921 1.23697 m .68316 1.21983 L .66529 1.22708 L p .826 .695 .72 r
F P
0 g
s
.66479 1.2102 m .66529 1.22708 L .68316 1.21983 L p .815 .655 .685 r
F P
0 g
s
.55668 1.26455 m .5654 1.27578 L .55551 1.27935 L p .706 .532 .651 r
F P
0 g
s
.55551 1.27935 m .54965 1.26863 L .55668 1.26455 L p .706 .532 .651 r
F P
0 g
s
.64437 1.27448 m .65913 1.26527 L .64419 1.27039 L p .789 .941 .926 r
F P
0 g
s
.64419 1.27039 m .63124 1.27909 L .64437 1.27448 L p .789 .941 .926 r
F P
0 g
s
.57324 1.27179 m .58663 1.27996 L .57595 1.28366 L p .716 .509 .608 r
F P
0 g
s
.57595 1.28366 m .5654 1.27578 L .57324 1.27179 L p .716 .509 .608 r
F P
0 g
s
.45482 1.20107 m .43011 1.20799 L .42379 1.21411 L p .698 .871 .963 r
F P
0 g
s
.39702 1.21345 m .42379 1.21411 L .43011 1.20799 L p .579 .805 .985 r
F P
0 g
s
.45632 1.25003 m .45883 1.23917 L .44983 1.24388 L p .553 .41 .641 r
F P
0 g
s
.44983 1.24388 m .44705 1.25395 L .45632 1.25003 L p .553 .41 .641 r
F P
0 g
s
.49061 1.20943 m .49717 1.20648 L .48583 1.21349 L p .719 .911 .961 r
F P
0 g
s
.48583 1.21349 m .47739 1.2162 L .49061 1.20943 L p .719 .911 .961 r
F P
0 g
s
.54706 1.18718 m .5438 1.19454 L .53353 1.20211 L p .737 .658 .771 r
F P
0 g
s
.53353 1.20211 m .53332 1.19516 L .54706 1.18718 L p .737 .658 .771 r
F P
0 g
s
.46825 1.29691 m .47745 1.29986 L .48378 1.29926 L p .565 .691 .926 r
F P
0 g
s
.67101 1.25247 m .67921 1.23697 L .66182 1.24347 L p .835 .745 .764 r
F P
0 g
s
.66529 1.22708 m .66182 1.24347 L .67921 1.23697 L p .826 .695 .72 r
F P
0 g
s
.44198 .60997 m .47087 .6352 L .45789 .65587 L p .632 .657 .857 r
F P
0 g
s
.45789 .65587 m .42335 .63156 L .44198 .60997 L p .632 .657 .857 r
F P
0 g
s
.65913 1.26527 m .67101 1.25247 L .65462 1.25824 L p .839 .818 .829 r
F P
0 g
s
.65462 1.25824 m .64419 1.27039 L .65913 1.26527 L p .839 .818 .829 r
F P
0 g
s
.66182 1.24347 m .65462 1.25824 L .67101 1.25247 L p .835 .745 .764 r
F P
0 g
s
.45092 .98217 m .47163 .9901 L .46043 1.00328 L p .666 .691 .864 r
F P
0 g
s
.46043 1.00328 m .43619 .99547 L .45092 .98217 L p .666 .691 .864 r
F P
0 g
s
.45015 .93104 m .41418 .953 L .40466 .9666 L p .798 .921 .92 r
F P
0 g
s
.36192 .97702 m .40466 .9666 L .41418 .953 L p .682 .898 .974 r
F P
0 g
s
.52333 1.28759 m .52613 1.27856 L .53368 1.27562 L p .659 .485 .638 r
F P
0 g
s
.52613 1.27856 m .52333 1.28759 L .5148 1.28959 L p .634 .451 .618 r
F P
0 g
s
.45605 1.26133 m .45632 1.25003 L .44705 1.25395 L p .573 .458 .684 r
F P
0 g
s
.44705 1.25395 m .44674 1.26444 L .45605 1.26133 L p .573 .458 .684 r
F P
0 g
s
.48966 .61419 m .50594 .64928 L .49395 .66935 L p .676 .639 .805 r
F P
0 g
s
.49395 .66935 m .47087 .6352 L .48966 .61419 L p .676 .639 .805 r
F P
0 g
s
.48378 1.29926 m .47547 1.29615 L .46825 1.29691 L p .565 .691 .926 r
F P
0 g
s
.46007 1.29145 m .46825 1.29691 L .47547 1.29615 L p .584 .608 .842 r
F P
0 g
s
.62777 1.15978 m .61295 1.15576 L .59669 1.16536 L p .688 .454 .562 r
F P
0 g
s
.58471 1.16561 m .59669 1.16536 L .61295 1.15576 L p .63 .363 .495 r
F P
0 g
s
.5148 1.28959 m .50887 1.29551 L .5009 1.29672 L p .564 .3 .473 r
F P
0 g
s
.49246 1.29946 m .5009 1.29672 L .50887 1.29551 L p .44 .01 .124 r
F P
0 g
s
.54965 1.26863 m .55551 1.27935 L .54455 1.2825 L p .691 .521 .652 r
F P
0 g
s
.54455 1.2825 m .54175 1.27231 L .54965 1.26863 L p .691 .521 .652 r
F P
0 g
s
.45803 1.2723 m .45605 1.26133 L .44674 1.26444 L p .583 .492 .718 r
F P
0 g
s
.44674 1.26444 m .44894 1.27464 L .45803 1.2723 L p .583 .492 .718 r
F P
0 g
s
.47547 1.29615 m .46809 1.29033 L .46007 1.29145 L p .584 .608 .842 r
F P
0 g
s
.45349 1.28386 m .46007 1.29145 L .46809 1.29033 L p .588 .559 .788 r
F P
0 g
s
.51063 1.15051 m .51886 1.12489 L .52629 1.11479 L p .442 .258 .535 r
F P
0 g
s
.51886 1.12489 m .51063 1.15051 L .50158 1.1594 L p .042 0 .253 r
F P
0 g
s
.83905 .80235 m .88358 .77476 L .86253 .79235 L p 0 0 0 r
F P
0 g
s
.86253 .79235 m .81999 .82002 L .83905 .80235 L p 0 0 0 r
F P
0 g
s
.53351 1.28524 m .53368 1.27562 L .54175 1.27231 L p .676 .506 .648 r
F P
0 g
s
.53368 1.27562 m .53351 1.28524 L .52333 1.28759 L p .659 .485 .638 r
F P
0 g
s
.53309 1.19148 m .53332 1.19516 L .52075 1.20274 L p .729 .682 .804 r
F P
0 g
s
.52075 1.20274 m .51706 1.19926 L .53309 1.19148 L p .729 .682 .804 r
F P
0 g
s
.46214 1.2822 m .45803 1.2723 L .44894 1.27464 L p .588 .523 .75 r
F P
0 g
s
.44894 1.27464 m .45349 1.28386 L .46214 1.2822 L p .588 .523 .75 r
F P
0 g
s
.50887 1.29551 m .5148 1.28959 L .52333 1.28759 L p .624 .399 .556 r
F P
0 g
s
.46809 1.29033 m .46214 1.2822 L .45349 1.28386 L p .588 .559 .788 r
F P
0 g
s
.49717 1.20648 m .49061 1.20943 L .49157 1.20906 L p .047 0 0 r
F P
0 g
s
.51706 1.19926 m .5039 1.20666 L p .51053 1.20127 L .731 .755 .872 r
F P
0 g
s
.5039 1.20666 m .49717 1.20648 L .51319 1.19911 L p .731 .755 .872 r
F P
0 g
s
.49157 1.20906 m .50943 1.2023 L p .50127 1.20508 L .047 0 0 r
F P
0 g
s
.49717 1.20648 m .49157 1.20906 L .50127 1.20508 L .047 0 0 r
F
.51319 1.19911 m .51556 1.1992 L p .51053 1.20127 L .731 .755 .872 r
F P
0 g
s
.51556 1.1992 m .51706 1.19926 L p .51053 1.20127 L .731 .755 .872 r
F P
0 g
s
.55111 1.08213 m .56073 1.11047 L .54816 1.12117 L p .662 .489 .64 r
F P
0 g
s
.54816 1.12117 m .54312 1.09342 L .55111 1.08213 L p .662 .489 .64 r
F P
0 g
s
.32013 1.06851 m .31234 1.09452 L .32394 1.08545 L p 0 0 .332 r
F P
0 g
s
.54175 1.27231 m .54455 1.2825 L .53351 1.28524 L p .676 .506 .648 r
F P
0 g
s
.48467 1.21532 m .49061 1.20943 L .47739 1.2162 L p .432 .108 .314 r
F P
0 g
s
.47739 1.2162 m .46974 1.22166 L .48467 1.21532 L p .432 .108 .314 r
F P
0 g
s
.51758 .97796 m .52316 .99962 L .51287 1.01274 L p .709 .637 .774 r
F P
0 g
s
.51287 1.01274 m .50214 .99163 L .51758 .97796 L p .709 .637 .774 r
F P
0 g
s
.70828 1.12394 m .7416 1.114 L .72346 1.127 L p .256 0 0 r
F P
0 g
s
.72346 1.127 m .69291 1.13639 L .70828 1.12394 L p .256 0 0 r
F P
0 g
s
.81815 .97025 m .80799 .93669 L .78456 .95273 L p .921 .671 .544 r
F P
0 g
s
.76757 .92324 m .78456 .95273 L .80799 .93669 L p .878 .581 .481 r
F P
0 g
s
.43011 1.20799 m .40467 1.20764 L .39702 1.21345 L p .579 .805 .985 r
F P
0 g
s
.37136 1.20577 m .39702 1.21345 L .40467 1.20764 L p .469 .704 .971 r
F P
0 g
s
.65329 .876 m .62305 .88574 L .60731 .90182 L p .259 0 .045 r
F P
0 g
s
.58531 .9192 m .60731 .90182 L p .61659 .89147 L 0 0 0 r
F P
0 g
s
.58611 .91758 m .60731 .90182 L p .60048 .90787 L .792 .998 .851 r
F P
0 g
s
.57051 .93446 m .58611 .91758 L p .60048 .90787 L .792 .998 .851 r
F P
0 g
s
.55327 .94936 m .56145 .93296 L .58611 .91758 L p .824 .871 .88 r
F P
0 g
s
.58611 .91758 m .56145 .93296 L p .57458 .92536 L .694 .969 .93 r
F P
0 g
s
.58611 .91758 m .57051 .93446 L .55327 .94936 L p .824 .871 .88 r
F P
0 g
s
.60379 .90846 m .58611 .91758 L p .57458 .92536 L .694 .969 .93 r
F P
0 g
s
.60731 .90182 m .58611 .91758 L p .59996 .91001 L 0 0 0 r
F P
0 g
s
.63221 .89238 m .60731 .90182 L p .59996 .91001 L 0 0 0 r
F P
0 g
s
.60731 .90182 m .62305 .88574 L p .61659 .89147 L 0 0 0 r
F P
0 g
s
.55726 .75784 m .57284 .80032 L .55577 .81775 L p .622 .434 .607 r
F P
0 g
s
.55577 .81775 m .54706 .77582 L .55726 .75784 L p .622 .434 .607 r
F P
0 g
s
.58663 1.27996 m .60144 1.28404 L .58762 1.28765 L p .693 .437 .531 r
F P
0 g
s
.58762 1.28765 m .57595 1.28366 L .58663 1.27996 L p .693 .437 .531 r
F P
0 g
s
.56079 1.1788 m .56849 1.17467 L .58471 1.16561 L p 0 0 0 r
F P
0 g
s
.95595 .63942 m .9541 .58716 L .93035 .60309 L p .995 .792 .424 r
F P
0 g
s
.91553 .55293 m .93035 .60309 L .9541 .58716 L p .972 .66 .317 r
F P
0 g
s
.57051 .93446 m .55801 .95794 L .54672 .97221 L p .832 .787 .81 r
F P
0 g
s
.50887 1.29551 m .50232 1.2985 L .49246 1.29946 L p .44 .01 .124 r
F P
0 g
s
.48378 1.29926 m .49246 1.29946 L .50232 1.2985 L p .48 .823 .998 r
F P
0 g
s
.5654 1.27578 m .57595 1.28366 L .56261 1.2869 L p .701 .503 .617 r
F P
0 g
s
.56261 1.2869 m .55551 1.27935 L .5654 1.27578 L p .701 .503 .617 r
F P
0 g
s
.53309 1.19148 m .53284 1.19134 L .5356 1.1903 L p .407 .858 .81 r
F P
0 g
s
.55074 1.18329 m .54825 1.18592 L p .54546 1.18689 L .751 .736 .838 r
F P
0 g
s
.53332 1.19516 m .53309 1.19148 L .55074 1.18329 L p .751 .736 .838 r
F P
0 g
s
.54706 1.18718 m .55074 1.18329 L .56849 1.17467 L p .777 .864 .914 r
F P
0 g
s
.55074 1.18329 m .53309 1.19148 L p .54858 1.18547 L .407 .858 .81 r
F P
0 g
s
.55459 1.18314 m .55233 1.18323 L p .54858 1.18547 L .407 .858 .81 r
F P
0 g
s
.55233 1.18323 m .55074 1.18329 L p .54858 1.18547 L .407 .858 .81 r
F P
0 g
s
.53309 1.19148 m .51706 1.19926 L p .52838 1.19358 L .728 .859 .943 r
F P
0 g
s
.53284 1.19134 m .53299 1.19143 L p .52838 1.19358 L .728 .859 .943 r
F P
0 g
s
.53299 1.19143 m .53309 1.19148 L p .52838 1.19358 L .728 .859 .943 r
F P
0 g
s
.51319 1.19911 m .50943 1.2023 L .51045 1.20196 L p .518 .185 .343 r
F P
0 g
s
.51706 1.19926 m .51319 1.19911 L .53284 1.19134 L p .728 .859 .943 r
F P
0 g
s
.51045 1.20196 m .53261 1.19475 L p .51966 1.19766 L .518 .185 .343 r
F P
0 g
s
.51319 1.19911 m .51045 1.20196 L .51966 1.19766 L .518 .185 .343 r
F
.53284 1.19134 m .53261 1.19475 L .53358 1.19445 L p .612 .347 .491 r
F P
0 g
s
.54706 1.18718 m .53332 1.19516 L p .54546 1.18689 L .751 .736 .838 r
F P
0 g
s
.53358 1.19445 m .55834 1.18675 L p .54132 1.18982 L .612 .347 .491 r
F P
0 g
s
.53309 1.19148 m .5356 1.1903 L .54056 1.18858 L .407 .858 .81 r
F
.53284 1.19134 m .53358 1.19445 L .54132 1.18982 L .612 .347 .491 r
F
.5356 1.1903 m .55459 1.18314 L p .54056 1.18858 L .407 .858 .81 r
F P
0 g
s
.54825 1.18592 m .54706 1.18718 L p .54546 1.18689 L .751 .736 .838 r
F P
0 g
s
.59669 1.16536 m .58471 1.16561 L .56849 1.17467 L p .575 .272 .416 r
F P
0 g
s
.51993 1.29396 m .52333 1.28759 L .53351 1.28524 L p .652 .446 .592 r
F P
0 g
s
.52333 1.28759 m .51993 1.29396 L .50887 1.29551 L p .624 .399 .556 r
F P
0 g
s
.47975 1.22375 m .48467 1.21532 L .46974 1.22166 L p .588 .385 .574 r
F P
0 g
s
.46974 1.22166 m .46341 1.22949 L .47975 1.22375 L p .588 .385 .574 r
F P
0 g
s
.31234 1.09452 m .31147 1.12165 L .3231 1.11359 L p .121 .082 .536 r
F P
0 g
s
.64119 1.16789 m .62777 1.15978 L .60837 1.16927 L p .718 .507 .603 r
F P
0 g
s
.59669 1.16536 m .60837 1.16927 L .62777 1.15978 L p .688 .454 .562 r
F P
0 g
s
.52202 1.14118 m .52629 1.11479 L .53461 1.10431 L p .561 .387 .602 r
F P
0 g
s
.52629 1.11479 m .52202 1.14118 L .51063 1.15051 L p .442 .258 .535 r
F P
0 g
s
.48876 .9765 m .50214 .99163 L .48996 1.00488 L p .685 .652 .812 r
F P
0 g
s
.48996 1.00488 m .47163 .9901 L .48876 .9765 L p .685 .652 .812 r
F P
0 g
s
.40467 1.20764 m .38026 1.20002 L .37136 1.20577 L p .469 .704 .971 r
F P
0 g
s
.34858 1.19161 m .37136 1.20577 L .38026 1.20002 L p .382 .589 .925 r
F P
0 g
s
.54312 1.09342 m .54816 1.12117 L .53488 1.13141 L p .621 .449 .628 r
F P
0 g
s
.53488 1.13141 m .53461 1.10431 L .54312 1.09342 L p .621 .449 .628 r
F P
0 g
s
.50232 1.2985 m .50887 1.29551 L .51993 1.29396 L p .562 .218 .343 r
F P
0 g
s
.50943 1.2023 m .51319 1.19911 L .49717 1.20648 L p .047 0 0 r
F P
0 g
s
.50254 .86812 m .51499 .82802 L .525 .81095 L p .172 0 .365 r
F P
0 g
s
.51499 .82802 m .50254 .86812 L .48909 .88412 L p 0 0 0 r
F P
0 g
s
.50232 1.2985 m .49558 1.29836 L .48378 1.29926 L p .48 .823 .998 r
F P
0 g
s
.47547 1.29615 m .48378 1.29926 L .49558 1.29836 L p .632 .676 .874 r
F P
0 g
s
.88358 .77476 m .91943 .73657 L .89693 .75389 L p .623 .893 .576 r
F P
0 g
s
.89693 .75389 m .86253 .79235 L .88358 .77476 L p .623 .893 .576 r
F P
0 g
s
.53461 1.10431 m .53488 1.13141 L .52202 1.14118 L p .561 .387 .602 r
F P
0 g
s
.7416 1.114 m .77119 1.09568 L .75065 1.10929 L p 0 0 0 r
F P
0 g
s
.75065 1.10929 m .72346 1.127 L .7416 1.114 L p 0 0 0 r
F P
0 g
s
.81933 1.00528 m .81815 .97025 L .79393 .98594 L p .952 .762 .61 r
F P
0 g
s
.78456 .95273 m .79393 .98594 L .81815 .97025 L p .921 .671 .544 r
F P
0 g
s
.55551 1.27935 m .56261 1.2869 L .54794 1.28968 L p .686 .491 .618 r
F P
0 g
s
.54794 1.28968 m .54455 1.2825 L .55551 1.27935 L p .686 .491 .618 r
F P
0 g
s
.5333 1.29202 m .53351 1.28524 L .54455 1.2825 L p .671 .474 .61 r
F P
0 g
s
.53351 1.28524 m .5333 1.29202 L .51993 1.29396 L p .652 .446 .592 r
F P
0 g
s
.32893 1.14089 m .3231 1.11359 L .31147 1.12165 L p .198 .231 .671 r
F P
0 g
s
.31147 1.12165 m .31757 1.14806 L .32893 1.14089 L p .198 .231 .671 r
F P
0 g
s
.42394 .5217 m .46447 .48041 L .47738 .45616 L p .881 .991 .793 r
F P
0 g
s
.47619 1.23415 m .47975 1.22375 L .46341 1.22949 L p .619 .455 .639 r
F P
0 g
s
.46341 1.22949 m .45883 1.23917 L .47619 1.23415 L p .619 .455 .639 r
F P
0 g
s
.53511 .96387 m .53476 .98611 L .52316 .99962 L p .74 .667 .779 r
F P
0 g
s
.52316 .99962 m .51758 .97796 L .53511 .96387 L p .74 .667 .779 r
F P
0 g
s
.60144 1.28404 m .61665 1.28374 L .59961 1.28748 L p .554 .151 .231 r
F P
0 g
s
.59961 1.28748 m .58762 1.28765 L .60144 1.28404 L p .554 .151 .231 r
F P
0 g
s
.38026 1.20002 m .35854 1.18565 L .34858 1.19161 L p .382 .589 .925 r
F P
0 g
s
.33023 1.17192 m .34858 1.19161 L .35854 1.18565 L p .315 .473 .858 r
F P
0 g
s
.26575 .64667 m .30261 .62554 L .28596 .64802 L p .647 .968 .882 r
F P
0 g
s
.28596 .64802 m .24706 .66887 L .26575 .64667 L p .647 .968 .882 r
F P
0 g
s
.54455 1.2825 m .54794 1.28968 L .5333 1.29202 L p .671 .474 .61 r
F P
0 g
s
.94412 .69039 m .95595 .63942 L .93212 .65579 L p .964 .895 .52 r
F P
0 g
s
.93035 .60309 m .93212 .65579 L .95595 .63942 L p .995 .792 .424 r
F P
0 g
s
.34102 1.16547 m .32893 1.14089 L .31757 1.14806 L p .257 .356 .774 r
F P
0 g
s
.31757 1.14806 m .33023 1.17192 L .34102 1.16547 L p .257 .356 .774 r
F P
0 g
s
.35854 1.18565 m .34102 1.16547 L .33023 1.17192 L p .315 .473 .858 r
F P
0 g
s
.54672 .97221 m .55327 .94936 L .57051 .93446 L p .832 .787 .81 r
F P
0 g
s
.55327 .94936 m .54672 .97221 L .53476 .98611 L p .778 .712 .79 r
F P
0 g
s
.49558 1.29836 m .48913 1.29509 L .47547 1.29615 L p .632 .676 .874 r
F P
0 g
s
.46809 1.29033 m .47547 1.29615 L .48913 1.29509 L p .642 .608 .8 r
F P
0 g
s
.65231 1.17956 m .64119 1.16789 L .61897 1.17709 L p .735 .542 .631 r
F P
0 g
s
.60837 1.16927 m .61897 1.17709 L .64119 1.16789 L p .718 .507 .603 r
F P
0 g
s
.56849 1.17467 m .57655 1.17449 L .59669 1.16536 L p .575 .272 .416 r
F P
0 g
s
.57655 1.17449 m .56849 1.17467 L .55074 1.18329 L p .398 0 .143 r
F P
0 g
s
.91943 .73657 m .94412 .69039 L .92071 .70728 L p .854 .946 .585 r
F P
0 g
s
.77119 1.09568 m .79502 1.07022 L .77259 1.08444 L p .795 .993 .76 r
F P
0 g
s
.77259 1.08444 m .75065 1.10929 L .77119 1.09568 L p .795 .993 .76 r
F P
0 g
s
.81144 1.03939 m .81933 1.00528 L .79502 1.02055 L p .966 .856 .68 r
F P
0 g
s
.79393 .98594 m .79502 1.02055 L .81933 1.00528 L p .952 .762 .61 r
F P
0 g
s
.57595 1.28366 m .58762 1.28765 L .57046 1.29076 L p .68 .43 .537 r
F P
0 g
s
.57046 1.29076 m .56261 1.2869 L .57595 1.28366 L p .68 .43 .537 r
F P
0 g
s
.47424 1.2458 m .47619 1.23415 L .45883 1.23917 L p .631 .489 .671 r
F P
0 g
s
.45883 1.23917 m .45632 1.25003 L .47424 1.2458 L p .631 .489 .671 r
F P
0 g
s
.53476 .98611 m .53511 .96387 L .55327 .94936 L p .778 .712 .79 r
F P
0 g
s
.42921 .98165 m .45092 .98217 L .43619 .99547 L p .678 .789 .934 r
F P
0 g
s
.43619 .99547 m .41082 .99485 L .42921 .98165 L p .678 .789 .934 r
F P
0 g
s
.51993 1.29396 m .51616 1.2972 L .50232 1.2985 L p .562 .218 .343 r
F P
0 g
s
.49558 1.29836 m .50232 1.2985 L .51616 1.2972 L p .47 .833 .993 r
F P
0 g
s
.93212 .65579 m .92071 .70728 L .94412 .69039 L p .964 .895 .52 r
F P
0 g
s
.51616 1.2972 m .51993 1.29396 L .5333 1.29202 L p .615 .32 .446 r
F P
0 g
s
.50601 1.20861 m .50943 1.2023 L .49061 1.20943 L p .597 .361 .529 r
F P
0 g
s
.49061 1.20943 m .48467 1.21532 L .50601 1.20861 L p .597 .361 .529 r
F P
0 g
s
.57147 1.09933 m .59033 1.12293 L .57409 1.13374 L p .689 .484 .604 r
F P
0 g
s
.57409 1.13374 m .56073 1.11047 L .57147 1.09933 L p .689 .484 .604 r
F P
0 g
s
.60239 1.11164 m .63024 1.12838 L .61323 1.13965 L p .704 .448 .534 r
F P
0 g
s
.61323 1.13965 m .59033 1.12293 L .60239 1.11164 L p .704 .448 .534 r
F P
0 g
s
.60837 1.16927 m .59669 1.16536 L .57655 1.17449 L p .672 .441 .562 r
F P
0 g
s
.48913 1.29509 m .48339 1.28891 L .46809 1.29033 L p .642 .608 .8 r
F P
0 g
s
.46214 1.2822 m .46809 1.29033 L .48339 1.28891 L p .643 .573 .761 r
F P
0 g
s
.55074 1.18329 m .55459 1.18314 L .57655 1.17449 L p .398 0 .143 r
F P
0 g
s
.79502 1.07022 m .81144 1.03939 L .78774 1.05416 L p .943 .952 .747 r
F P
0 g
s
.78774 1.05416 m .77259 1.08444 L .79502 1.07022 L p .943 .952 .747 r
F P
0 g
s
.79502 1.02055 m .78774 1.05416 L .81144 1.03939 L p .966 .856 .68 r
F P
0 g
s
.47403 1.25792 m .47424 1.2458 L .45632 1.25003 L p .637 .512 .694 r
F P
0 g
s
.45632 1.25003 m .45605 1.26133 L .47403 1.25792 L p .637 .512 .694 r
F P
0 g
s
.61665 1.28374 m .63124 1.27909 L .61111 1.28314 L p .652 .917 .978 r
F P
0 g
s
.61111 1.28314 m .59961 1.28748 L .61665 1.28374 L p .652 .917 .978 r
F P
0 g
s
.53261 1.19475 m .53284 1.19134 L .51319 1.19911 L p .518 .185 .343 r
F P
0 g
s
.54706 .77582 m .55577 .81775 L .53731 .8349 L p .548 .363 .582 r
F P
0 g
s
.53731 .8349 m .53599 .79353 L .54706 .77582 L p .548 .363 .582 r
F P
0 g
s
.48339 1.28891 m .47876 1.28024 L .46214 1.2822 L p .643 .573 .761 r
F P
0 g
s
.45803 1.2723 m .46214 1.2822 L .47876 1.28024 L p .642 .55 .735 r
F P
0 g
s
.47557 1.26967 m .47403 1.25792 L .45605 1.26133 L p .641 .531 .714 r
F P
0 g
s
.45605 1.26133 m .45803 1.2723 L .47557 1.26967 L p .641 .531 .714 r
F P
0 g
s
.42394 .5217 m .37432 .54977 L .35996 .57426 L p .8 .932 .919 r
F P
0 g
s
.30042 .58774 m .35996 .57426 L .37432 .54977 L p .687 .902 .973 r
F P
0 g
s
.47876 1.28024 m .47557 1.26967 L .45803 1.2723 L p .642 .55 .735 r
F P
0 g
s
.41418 .953 m .37416 .964 L .36192 .97702 L p .682 .898 .974 r
F P
0 g
s
.31792 .97588 m .36192 .97702 L .37416 .964 L p .548 .821 .997 r
F P
0 g
s
.51905 .8517 m .525 .81095 L .53599 .79353 L p .429 .242 .524 r
F P
0 g
s
.525 .81095 m .51905 .8517 L .50254 .86812 L p .172 0 .365 r
F P
0 g
s
.53306 1.29551 m .5333 1.29202 L .54794 1.28968 L p .644 .377 .5 r
F P
0 g
s
.5333 1.29202 m .53306 1.29551 L .51616 1.2972 L p .615 .32 .446 r
F P
0 g
s
.66037 1.194 m .65231 1.17956 L .62774 1.18829 L p .746 .569 .654 r
F P
0 g
s
.61897 1.17709 m .62774 1.18829 L .65231 1.17956 L p .735 .542 .631 r
F P
0 g
s
.56261 1.2869 m .57046 1.29076 L .55169 1.29337 L p .665 .411 .528 r
F P
0 g
s
.55169 1.29337 m .54794 1.28968 L .56261 1.2869 L p .665 .411 .528 r
F P
0 g
s
.51616 1.2972 m .5123 1.29711 L .49558 1.29836 L p .47 .833 .993 r
F P
0 g
s
.48913 1.29509 m .49558 1.29836 L .5123 1.29711 L p .66 .673 .851 r
F P
0 g
s
.50319 1.21764 m .50601 1.20861 L .48467 1.21532 L p .64 .459 .622 r
F P
0 g
s
.48467 1.21532 m .47975 1.22375 L .50319 1.21764 L p .64 .459 .622 r
F P
0 g
s
.54794 1.28968 m .55169 1.29337 L .53306 1.29551 L p .644 .377 .5 r
F P
0 g
s
.58874 .52906 m .56842 .56697 L .55327 .58772 L p .877 .877 .832 r
F P
0 g
s
.53599 .79353 m .53731 .8349 L .51905 .8517 L p .429 .242 .524 r
F P
0 g
s
.63124 1.27909 m .64419 1.27039 L .62133 1.27493 L p .771 .76 .843 r
F P
0 g
s
.62133 1.27493 m .61111 1.28314 L .63124 1.27909 L p .771 .76 .843 r
F P
0 g
s
.71068 .44289 m .664 .45909 L .64352 .47957 L p 0 0 0 r
F P
0 g
s
.60875 .50805 m .64352 .47957 L p .65437 .46763 L 0 0 0 r
F P
0 g
s
.61427 .50055 m .64352 .47957 L p .63345 .48867 L .751 .993 .809 r
F P
0 g
s
.58874 .52906 m .61427 .50055 L p .63345 .48867 L .751 .993 .809 r
F P
0 g
s
.56462 .55027 m .57892 .52202 L .61427 .50055 L p .831 .941 .897 r
F P
0 g
s
.61427 .50055 m .57892 .52202 L p .59835 .51041 L .67 .97 .915 r
F P
0 g
s
.64352 .47957 m .61427 .50055 L p .6346 .4894 L .409 .838 .689 r
F P
0 g
s
.68314 .46276 m .64352 .47957 L p .6346 .4894 L .409 .838 .689 r
F P
0 g
s
.61427 .50055 m .58874 .52906 L .56462 .55027 L p .831 .941 .897 r
F P
0 g
s
.61427 .50055 m .64345 .48347 L .68314 .46276 L p .409 .838 .689 r
F P
0 g
s
.64345 .48347 m .61427 .50055 L p .59835 .51041 L .67 .97 .915 r
F P
0 g
s
.64352 .47957 m .664 .45909 L p .65437 .46763 L 0 0 0 r
F P
0 g
s
.41364 .06991 m .45287 .11041 L .43465 .14196 L p .623 .674 .878 r
F P
0 g
s
.43465 .14196 m .38715 .10474 L .41364 .06991 L p .623 .674 .878 r
F P
0 g
s
.5123 1.29711 m .51616 1.2972 L .53306 1.29551 L p .386 .822 .956 r
F P
0 g
s
.58762 1.28765 m .59961 1.28748 L .57853 1.29066 L p .451 0 .041 r
F P
0 g
s
.57853 1.29066 m .57046 1.29076 L .58762 1.28765 L p .451 0 .041 r
F P
0 g
s
.55834 1.18675 m .55459 1.18314 L .53284 1.19134 L p .612 .347 .491 r
F P
0 g
s
.57655 1.17449 m .58442 1.17826 L .60837 1.16927 L p .672 .441 .562 r
F P
0 g
s
.58442 1.17826 m .57655 1.17449 L .55459 1.18314 L p .649 .41 .543 r
F P
0 g
s
.61897 1.17709 m .60837 1.16927 L .58442 1.17826 L p .701 .502 .615 r
F P
0 g
s
.66479 1.2102 m .66037 1.194 L .6341 1.20212 L p .754 .591 .675 r
F P
0 g
s
.62774 1.18829 m .6341 1.20212 L .66037 1.194 L p .746 .569 .654 r
F P
0 g
s
.49918 1.17154 m .51063 1.15051 L .52202 1.14118 L p .419 .174 .433 r
F P
0 g
s
.51063 1.15051 m .49918 1.17154 L .48589 1.17968 L p 0 0 .042 r
F P
0 g
s
.51285 .59295 m .52065 .62894 L .50594 .64928 L p .72 .677 .807 r
F P
0 g
s
.50594 .64928 m .48966 .61419 L .51285 .59295 L p .72 .677 .807 r
F P
0 g
s
.53239 1.20149 m .53261 1.19475 L .50943 1.2023 L p .643 .435 .586 r
F P
0 g
s
.50943 1.2023 m .50601 1.20861 L .53239 1.20149 L p .643 .435 .586 r
F P
0 g
s
.51062 .96244 m .51758 .97796 L .50214 .99163 L p .713 .679 .815 r
F P
0 g
s
.50214 .99163 m .48876 .9765 L .51062 .96244 L p .713 .679 .815 r
F P
0 g
s
.68499 .87507 m .65329 .876 L .63221 .89238 L p .505 .111 .228 r
F P
0 g
s
.60731 .90182 m .63221 .89238 L .65329 .876 L p .259 0 .045 r
F P
0 g
s
.5123 1.29711 m .50859 1.29367 L .48913 1.29509 L p .66 .673 .851 r
F P
0 g
s
.48339 1.28891 m .48913 1.29509 L .50859 1.29367 L p .667 .61 .783 r
F P
0 g
s
.50115 1.22876 m .50319 1.21764 L .47975 1.22375 L p .653 .496 .658 r
F P
0 g
s
.47975 1.22375 m .47619 1.23415 L .50115 1.22876 L p .653 .496 .658 r
F P
0 g
s
.55459 1.18314 m .55834 1.18675 L .58442 1.17826 L p .649 .41 .543 r
F P
0 g
s
.64419 1.27039 m .65462 1.25824 L .62957 1.2634 L p .771 .693 .777 r
F P
0 g
s
.62957 1.2634 m .62133 1.27493 L .64419 1.27039 L p .771 .693 .777 r
F P
0 g
s
.66529 1.22708 m .66479 1.2102 L .6376 1.21761 L p .759 .612 .695 r
F P
0 g
s
.6341 1.20212 m .6376 1.21761 L .66479 1.2102 L p .754 .591 .675 r
F P
0 g
s
.53282 1.29544 m .53306 1.29551 L .55169 1.29337 L p .151 .694 .771 r
F P
0 g
s
.53306 1.29551 m .53282 1.29544 L .5123 1.29711 L p .386 .822 .956 r
F P
0 g
s
.6323 .80004 m .67522 .82455 L .65304 .84195 L p .688 .377 .443 r
F P
0 g
s
.65304 .84195 m .61674 .81756 L .6323 .80004 L p .688 .377 .443 r
F P
0 g
s
.65462 1.25824 m .66182 1.24347 L .63525 1.24935 L p .768 .658 .741 r
F P
0 g
s
.63525 1.24935 m .62957 1.2634 L .65462 1.25824 L p .768 .658 .741 r
F P
0 g
s
.56073 1.11047 m .57409 1.13374 L .55517 1.14402 L p .655 .453 .599 r
F P
0 g
s
.55517 1.14402 m .54816 1.12117 L .56073 1.11047 L p .655 .453 .599 r
F P
0 g
s
.23504 .6778 m .26575 .64667 L .24706 .66887 L p .541 .86 .569 r
F P
0 g
s
.24706 .66887 m .21477 .69924 L .23504 .6778 L p .541 .86 .569 r
F P
0 g
s
.47362 .96841 m .48876 .9765 L .47163 .9901 L p .692 .717 .868 r
F P
0 g
s
.47163 .9901 m .45092 .98217 L .47362 .96841 L p .692 .717 .868 r
F P
0 g
s
.57046 1.29076 m .57853 1.29066 L .55555 1.29331 L p .212 0 0 r
F P
0 g
s
.55555 1.29331 m .55169 1.29337 L .57046 1.29076 L p .212 0 0 r
F P
0 g
s
.66182 1.24347 m .66529 1.22708 L .63799 1.23373 L p .764 .633 .716 r
F P
0 g
s
.6376 1.21761 m .63799 1.23373 L .66529 1.22708 L p .759 .612 .695 r
F P
0 g
s
.50859 1.29367 m .50529 1.28712 L .48339 1.28891 L p .667 .61 .783 r
F P
0 g
s
.47876 1.28024 m .48339 1.28891 L .50529 1.28712 L p .667 .581 .749 r
F P
0 g
s
.63799 1.23373 m .63525 1.24935 L .66182 1.24347 L p .764 .633 .716 r
F P
0 g
s
.50003 1.2412 m .50115 1.22876 L .47619 1.23415 L p .66 .518 .679 r
F P
0 g
s
.47619 1.23415 m .47424 1.2458 L .50003 1.2412 L p .66 .518 .679 r
F P
0 g
s
.50859 1.29367 m .5123 1.29711 L .53282 1.29544 L p .679 .678 .843 r
F P
0 g
s
.40797 .98864 m .42921 .98165 L .41082 .99485 L p .635 .929 .972 r
F P
0 g
s
.41082 .99485 m .38605 1.00149 L .40797 .98864 L p .635 .929 .972 r
F P
0 g
s
.55169 1.29337 m .55555 1.29331 L .53282 1.29544 L p .151 .694 .771 r
F P
0 g
s
.59961 1.28748 m .61111 1.28314 L .58627 1.2866 L p .717 .799 .915 r
F P
0 g
s
.58627 1.2866 m .57853 1.29066 L .59961 1.28748 L p .717 .799 .915 r
F P
0 g
s
.58702 .78264 m .61674 .81756 L .59462 .8349 L p .67 .432 .552 r
F P
0 g
s
.59462 .8349 m .57284 .80032 L .58702 .78264 L p .67 .432 .552 r
F P
0 g
s
.62774 1.18829 m .61897 1.17709 L .59155 1.18575 L p .715 .535 .645 r
F P
0 g
s
.58442 1.17826 m .59155 1.18575 L .61897 1.17709 L p .701 .502 .615 r
F P
0 g
s
.53222 1.21109 m .53239 1.20149 L .50601 1.20861 L p .665 .491 .64 r
F P
0 g
s
.50601 1.20861 m .50319 1.21764 L .53222 1.21109 L p .665 .491 .64 r
F P
0 g
s
.56175 1.19388 m .55834 1.18675 L .53261 1.19475 L p .667 .47 .609 r
F P
0 g
s
.53261 1.19475 m .53239 1.20149 L .56175 1.19388 L p .667 .47 .609 r
F P
0 g
s
.50529 1.28712 m .50263 1.27791 L .47876 1.28024 L p .667 .581 .749 r
F P
0 g
s
.47557 1.26967 m .47876 1.28024 L .50263 1.27791 L p .666 .562 .728 r
F P
0 g
s
.49991 1.25413 m .50003 1.2412 L .47424 1.2458 L p .663 .534 .696 r
F P
0 g
s
.47424 1.2458 m .47403 1.25792 L .49991 1.25413 L p .663 .534 .696 r
F P
0 g
s
.59155 1.18575 m .58442 1.17826 L .55834 1.18675 L p .685 .49 .617 r
F P
0 g
s
.51608 1.1629 m .52202 1.14118 L .53488 1.13141 L p .549 .333 .539 r
F P
0 g
s
.52202 1.14118 m .51608 1.1629 L .49918 1.17154 L p .419 .174 .433 r
F P
0 g
s
.50263 1.27791 m .50079 1.26665 L .47557 1.26967 L p .666 .562 .728 r
F P
0 g
s
.47403 1.25792 m .47557 1.26967 L .50079 1.26665 L p .665 .548 .711 r
F P
0 g
s
.50079 1.26665 m .49991 1.25413 L .47403 1.25792 L p .665 .548 .711 r
F P
0 g
s
.21868 .08332 m .2755 .07407 L .25445 .11549 L p .667 .9 .979 r
F P
0 g
s
.25445 .11549 m .19449 .12751 L .21868 .08332 L p .667 .9 .979 r
F P
0 g
s
.40911 .59544 m .44198 .60997 L .42335 .63156 L p .639 .731 .915 r
F P
0 g
s
.42335 .63156 m .3842 .61784 L .40911 .59544 L p .639 .731 .915 r
F P
0 g
s
.37416 .964 m .33278 .96316 L .31792 .97588 L p .548 .821 .997 r
F P
0 g
s
.27568 .96321 m .31792 .97588 L .33278 .96316 L p .423 .709 .983 r
F P
0 g
s
.55834 1.18675 m .56175 1.19388 L .59155 1.18575 L p .685 .49 .617 r
F P
0 g
s
.54816 1.12117 m .55517 1.14402 L .53526 1.15374 L p .613 .408 .58 r
F P
0 g
s
.53526 1.15374 m .53488 1.13141 L .54816 1.12117 L p .613 .408 .58 r
F P
0 g
s
.53282 1.29544 m .53258 1.29182 L .50859 1.29367 L p .679 .678 .843 r
F P
0 g
s
.50529 1.28712 m .50859 1.29367 L .53258 1.29182 L p .684 .617 .775 r
F P
0 g
s
.63024 1.12838 m .66111 1.13684 L .63866 1.14831 L p .654 .335 .42 r
F P
0 g
s
.63866 1.14831 m .61323 1.13965 L .63024 1.12838 L p .654 .335 .42 r
F P
0 g
s
.53258 1.29182 m .53282 1.29544 L .55555 1.29331 L p .694 .695 .847 r
F P
0 g
s
.53488 1.13141 m .53526 1.15374 L .51608 1.1629 L p .549 .333 .539 r
F P
0 g
s
.53554 .94793 m .53511 .96387 L .51758 .97796 L p .743 .716 .826 r
F P
0 g
s
.51758 .97796 m .51062 .96244 L .53554 .94793 L p .743 .716 .826 r
F P
0 g
s
.61111 1.28314 m .62133 1.27493 L .59315 1.27885 L p .738 .685 .799 r
F P
0 g
s
.59315 1.27885 m .58627 1.2866 L .61111 1.28314 L p .738 .685 .799 r
F P
0 g
s
.57853 1.29066 m .58627 1.2866 L .55924 1.28949 L p .709 .73 .867 r
F P
0 g
s
.55924 1.28949 m .55555 1.29331 L .57853 1.29066 L p .709 .73 .867 r
F P
0 g
s
.55327 .58772 m .56462 .55027 L .58874 .52906 L p .877 .877 .832 r
F P
0 g
s
.56462 .55027 m .55327 .58772 L .53687 .60839 L p .817 .785 .823 r
F P
0 g
s
.53209 1.22292 m .53222 1.21109 L .50319 1.21764 L p .673 .518 .666 r
F P
0 g
s
.50319 1.21764 m .50115 1.22876 L .53209 1.22292 L p .673 .518 .666 r
F P
0 g
s
.6341 1.20212 m .62774 1.18829 L .59746 1.19645 L p .722 .557 .665 r
F P
0 g
s
.59155 1.18575 m .59746 1.19645 L .62774 1.18829 L p .715 .535 .645 r
F P
0 g
s
.53857 .5716 m .53687 .60839 L .52065 .62894 L p .765 .723 .813 r
F P
0 g
s
.52065 .62894 m .51285 .59295 L .53857 .5716 L p .765 .723 .813 r
F P
0 g
s
.55555 1.29331 m .55924 1.28949 L .53258 1.29182 L p .694 .695 .847 r
F P
0 g
s
.56458 1.20405 m .56175 1.19388 L .53239 1.20149 L p .682 .511 .648 r
F P
0 g
s
.53239 1.20149 m .53222 1.21109 L .56458 1.20405 L p .682 .511 .648 r
F P
0 g
s
.46685 1.19404 m .44573 1.20147 L p .4303 1.20791 L .636 .932 .969 r
F P
0 g
s
.48528 1.18653 m .46984 1.19443 L .44573 1.20147 L p .375 .837 .78 r
F P
0 g
s
.44573 1.20147 m .46984 1.19443 L p .45841 1.19617 L .574 .863 .996 r
F P
0 g
s
.42395 1.20142 m .44573 1.20147 L p .45841 1.19617 L .574 .863 .996 r
F P
0 g
s
.44573 1.20147 m .43011 1.20799 L p .4303 1.20791 L .636 .932 .969 r
F P
0 g
s
.40467 1.20764 m .43011 1.20799 L .44573 1.20147 L p .588 .833 .991 r
F P
0 g
s
.59746 1.19645 m .59155 1.18575 L .56175 1.19388 L p .698 .525 .649 r
F P
0 g
s
.56145 .93296 m .55327 .94936 L .53511 .96387 L p .78 .773 .846 r
F P
0 g
s
.53258 1.29182 m .53237 1.2849 L .50529 1.28712 L p .684 .617 .775 r
F P
0 g
s
.50263 1.27791 m .50529 1.28712 L .53237 1.2849 L p .684 .59 .744 r
F P
0 g
s
.59033 1.12293 m .61323 1.13965 L .59033 1.15034 L p .674 .427 .54 r
F P
0 g
s
.59033 1.15034 m .57409 1.13374 L .59033 1.12293 L p .674 .427 .54 r
F P
0 g
s
.53511 .96387 m .53554 .94793 L .56145 .93296 L p .78 .773 .846 r
F P
0 g
s
.53202 1.23615 m .53209 1.22292 L .50115 1.22876 L p .678 .535 .683 r
F P
0 g
s
.50115 1.22876 m .50003 1.2412 L .53202 1.23615 L p .678 .535 .683 r
F P
0 g
s
.62133 1.27493 m .62957 1.2634 L .59869 1.26793 L p .737 .643 .756 r
F P
0 g
s
.59869 1.26793 m .59315 1.27885 L .62133 1.27493 L p .737 .643 .756 r
F P
0 g
s
.53687 .60839 m .53857 .5716 L .56462 .55027 L p .817 .785 .823 r
F P
0 g
s
.2127 .71682 m .23504 .6778 L .21477 .69924 L p 0 0 0 r
F P
0 g
s
.21477 .69924 m .19135 .73708 L .2127 .71682 L p 0 0 0 r
F P
0 g
s
.56175 1.19388 m .56458 1.20405 L .59746 1.19645 L p .698 .525 .649 r
F P
0 g
s
.6376 1.21761 m .6341 1.20212 L .60174 1.20964 L p .727 .574 .682 r
F P
0 g
s
.59746 1.19645 m .60174 1.20964 L .6341 1.20212 L p .722 .557 .665 r
F P
0 g
s
.53237 1.2849 m .53258 1.29182 L .55924 1.28949 L p .699 .628 .774 r
F P
0 g
s
.46922 .58803 m .48966 .61419 L .47087 .6352 L p .678 .691 .855 r
F P
0 g
s
.47087 .6352 m .44198 .60997 L .46922 .58803 L p .678 .691 .855 r
F P
0 g
s
.38865 1.0027 m .40797 .98864 L .38605 1.00149 L p 0 0 0 r
F P
0 g
s
.38605 1.00149 m .36355 1.01497 L .38865 1.0027 L p 0 0 0 r
F P
0 g
s
.53237 1.2849 m .5322 1.27513 L .50263 1.27791 L p .684 .59 .744 r
F P
0 g
s
.50079 1.26665 m .50263 1.27791 L .5322 1.27513 L p .683 .573 .725 r
F P
0 g
s
.58627 1.2866 m .59315 1.27885 L .56253 1.28216 L p .716 .648 .781 r
F P
0 g
s
.56253 1.28216 m .55924 1.28949 L .58627 1.2866 L p .716 .648 .781 r
F P
0 g
s
.53202 1.24989 m .53202 1.23615 L .50003 1.2412 L p .68 .548 .697 r
F P
0 g
s
.50003 1.2412 m .49991 1.25413 L .53202 1.24989 L p .68 .548 .697 r
F P
0 g
s
.71599 .88306 m .68499 .87507 L .65837 .89159 L p .646 .292 .358 r
F P
0 g
s
.63221 .89238 m .65837 .89159 L .68499 .87507 L p .505 .111 .228 r
F P
0 g
s
.56662 1.21656 m .56458 1.20405 L .53222 1.21109 L p .689 .533 .669 r
F P
0 g
s
.53222 1.21109 m .53209 1.22292 L .56662 1.21656 L p .689 .533 .669 r
F P
0 g
s
.62957 1.2634 m .63525 1.24935 L .60251 1.25459 L p .735 .62 .731 r
F P
0 g
s
.60251 1.25459 m .59869 1.26793 L .62957 1.2634 L p .735 .62 .731 r
F P
0 g
s
.5322 1.27513 m .53208 1.26319 L .50079 1.26665 L p .683 .573 .725 r
F P
0 g
s
.49991 1.25413 m .50079 1.26665 L .53208 1.26319 L p .682 .56 .71 r
F P
0 g
s
.55924 1.28949 m .56253 1.28216 L .53237 1.2849 L p .699 .628 .774 r
F P
0 g
s
.63799 1.23373 m .6376 1.21761 L .60409 1.2244 L p .73 .588 .697 r
F P
0 g
s
.60174 1.20964 m .60409 1.2244 L .6376 1.21761 L p .727 .574 .682 r
F P
0 g
s
.53208 1.26319 m .53202 1.24989 L .49991 1.25413 L p .682 .56 .71 r
F P
0 g
s
.33278 .96316 m .29288 .95042 L .27568 .96321 L p .423 .709 .983 r
F P
0 g
s
.23812 .93985 m .27568 .96321 L .29288 .95042 L p .315 .578 .935 r
F P
0 g
s
.60174 1.20964 m .59746 1.19645 L .56458 1.20405 L p .705 .545 .669 r
F P
0 g
s
.63525 1.24935 m .63799 1.23373 L .60436 1.23974 L p .733 .603 .712 r
F P
0 g
s
.60436 1.23974 m .60251 1.25459 L .63525 1.24935 L p .733 .603 .712 r
F P
0 g
s
.60409 1.2244 m .60436 1.23974 L .63799 1.23373 L p .73 .588 .697 r
F P
0 g
s
.5322 1.27513 m .53237 1.2849 L .56253 1.28216 L p .699 .601 .743 r
F P
0 g
s
.48528 1.18653 m .49918 1.17154 L .51608 1.1629 L p .361 .022 .248 r
F P
0 g
s
.56458 1.20405 m .56662 1.21656 L .60174 1.20964 L p .705 .545 .669 r
F P
0 g
s
.45773 .96796 m .47362 .96841 L .45092 .98217 L p .685 .835 .956 r
F P
0 g
s
.45092 .98217 m .42921 .98165 L .45773 .96796 L p .685 .835 .956 r
F P
0 g
s
.50274 .95415 m .51062 .96244 L .48876 .9765 L p .716 .752 .88 r
F P
0 g
s
.48876 .9765 m .47362 .96841 L .50274 .95415 L p .716 .752 .88 r
F P
0 g
s
.59315 1.27885 m .59869 1.26793 L .56517 1.27182 L p .715 .617 .746 r
F P
0 g
s
.56517 1.27182 m .56253 1.28216 L .59315 1.27885 L p .715 .617 .746 r
F P
0 g
s
.56775 1.23057 m .56662 1.21656 L .53209 1.22292 L p .693 .548 .685 r
F P
0 g
s
.53209 1.22292 m .53202 1.23615 L .56775 1.23057 L p .693 .548 .685 r
F P
0 g
s
.56253 1.28216 m .56517 1.27182 L .5322 1.27513 L p .699 .601 .743 r
F P
0 g
s
.60409 1.2244 m .60174 1.20964 L .56662 1.21656 L p .708 .56 .684 r
F P
0 g
s
.53208 1.26319 m .5322 1.27513 L .56517 1.27182 L p .698 .584 .724 r
F P
0 g
s
.37432 .54977 m .31874 .56249 L .30042 .58774 L p .687 .902 .973 r
F P
0 g
s
.23892 .58512 m .30042 .58774 L .31874 .56249 L p .56 .825 .995 r
F P
0 g
s
.44573 1.20147 m .42395 1.20142 L .40467 1.20764 L p .588 .833 .991 r
F P
0 g
s
.38026 1.20002 m .40467 1.20764 L .42395 1.20142 L p .557 .723 .951 r
F P
0 g
s
.19135 .73708 m .17843 .77981 L .20035 .76104 L p 0 0 .213 r
F P
0 g
s
.56788 1.24511 m .56775 1.23057 L .53202 1.23615 L p .695 .56 .698 r
F P
0 g
s
.53202 1.23615 m .53202 1.24989 L .56788 1.24511 L p .695 .56 .698 r
F P
0 g
s
.56662 1.21656 m .56775 1.23057 L .60409 1.2244 L p .708 .56 .684 r
F P
0 g
s
.59869 1.26793 m .60251 1.25459 L .567 1.25919 L p .714 .599 .726 r
F P
0 g
s
.567 1.25919 m .56517 1.27182 L .59869 1.26793 L p .714 .599 .726 r
F P
0 g
s
.66111 1.13684 m .69291 1.13639 L .66491 1.14823 L p .493 .069 .163 r
F P
0 g
s
.66491 1.14823 m .63866 1.14831 L .66111 1.13684 L p .493 .069 .163 r
F P
0 g
s
.56517 1.27182 m .567 1.25919 L .53208 1.26319 L p .698 .584 .724 r
F P
0 g
s
.53202 1.24989 m .53208 1.26319 L .567 1.25919 L p .697 .571 .71 r
F P
0 g
s
.567 1.25919 m .56788 1.24511 L .53202 1.24989 L p .697 .571 .71 r
F P
0 g
s
.60436 1.23974 m .60409 1.2244 L .56775 1.23057 L p .711 .573 .698 r
F P
0 g
s
.48513 .90089 m .50254 .86812 L .51905 .8517 L p .161 0 .216 r
F P
0 g
s
.50254 .86812 m .48513 .90089 L .46442 .91622 L p 0 0 0 r
F P
0 g
s
.57284 .80032 m .59462 .8349 L .56795 .85198 L p .613 .38 .54 r
F P
0 g
s
.56795 .85198 m .55577 .81775 L .57284 .80032 L p .613 .38 .54 r
F P
0 g
s
.37261 1.0229 m .38865 1.0027 L .36355 1.01497 L p .271 0 .243 r
F P
0 g
s
.36355 1.01497 m .3449 1.03439 L .37261 1.0229 L p .271 0 .243 r
F P
0 g
s
.29288 .95042 m .25725 .92654 L .23812 .93985 L p .315 .578 .935 r
F P
0 g
s
.20784 .90737 m .23812 .93985 L .25725 .92654 L p .224 .439 .86 r
F P
0 g
s
.57409 1.13374 m .59033 1.15034 L .56369 1.16039 L p .639 .392 .529 r
F P
0 g
s
.56369 1.16039 m .55517 1.14402 L .57409 1.13374 L p .639 .392 .529 r
F P
0 g
s
.60251 1.25459 m .60436 1.23974 L .56788 1.24511 L p .713 .585 .711 r
F P
0 g
s
.56788 1.24511 m .567 1.25919 L .60251 1.25459 L p .713 .585 .711 r
F P
0 g
s
.56775 1.23057 m .56788 1.24511 L .60436 1.23974 L p .711 .573 .698 r
F P
0 g
s
.75979 .43987 m .71068 .44289 L .68314 .46276 L p .288 0 0 r
F P
0 g
s
.64352 .47957 m .68314 .46276 L .71068 .44289 L p 0 0 0 r
F P
0 g
s
.50884 1.17846 m .51608 1.1629 L .53526 1.15374 L p .519 .235 .424 r
F P
0 g
s
.51608 1.1629 m .50884 1.17846 L .48528 1.18653 L p .361 .022 .248 r
F P
0 g
s
.58611 .91758 m .60379 .90846 L .63221 .89238 L p 0 0 0 r
F P
0 g
s
.47738 .45616 m .44407 .49718 L .42394 .5217 L p .881 .991 .793 r
F P
0 g
s
.19891 .80733 m .20035 .76104 L .17843 .77981 L p 0 0 .456 r
F P
0 g
s
.17843 .77981 m .17692 .82448 L .19891 .80733 L p 0 0 .456 r
F P
0 g
s
.25725 .92654 m .22843 .89311 L .20784 .90737 L p .224 .439 .86 r
F P
0 g
s
.18696 .86801 m .20784 .90737 L .22843 .89311 L p .142 .293 .759 r
F P
0 g
s
.74417 .89948 m .71599 .88306 L .68399 .89956 L p .727 .415 .451 r
F P
0 g
s
.65837 .89159 m .68399 .89956 L .71599 .88306 L p .646 .292 .358 r
F P
0 g
s
.42395 1.20142 m .403 1.19386 L .38026 1.20002 L p .557 .723 .951 r
F P
0 g
s
.35854 1.18565 m .38026 1.20002 L .403 1.19386 L p .537 .63 .891 r
F P
0 g
s
.53603 .9394 m .53554 .94793 L .51062 .96244 L p .74 .804 .903 r
F P
0 g
s
.51062 .96244 m .50274 .95415 L .53603 .9394 L p .74 .804 .903 r
F P
0 g
s
.55517 1.14402 m .56369 1.16039 L .53572 1.16976 L p .593 .335 .496 r
F P
0 g
s
.53572 1.16976 m .53526 1.15374 L .55517 1.14402 L p .593 .335 .496 r
F P
0 g
s
.2085 .85242 m .19891 .80733 L .17692 .82448 L p .058 .134 .629 r
F P
0 g
s
.17692 .82448 m .18696 .86801 L .2085 .85242 L p .058 .134 .629 r
F P
0 g
s
.67522 .82455 m .72297 .83654 L .69362 .85403 L p .599 .208 .272 r
F P
0 g
s
.69362 .85403 m .65304 .84195 L .67522 .82455 L p .599 .208 .272 r
F P
0 g
s
.65837 .89159 m .63221 .89238 L .60379 .90846 L p .387 0 .077 r
F P
0 g
s
.53526 1.15374 m .53572 1.16976 L .50884 1.17846 L p .519 .235 .424 r
F P
0 g
s
.16646 .10897 m .21868 .08332 L .19449 .12751 L p .706 .975 .919 r
F P
0 g
s
.19449 .12751 m .13983 .15493 L .16646 .10897 L p .706 .975 .919 r
F P
0 g
s
.22843 .89311 m .2085 .85242 L .18696 .86801 L p .142 .293 .759 r
F P
0 g
s
.61323 1.13965 m .63866 1.14831 L .60839 1.15909 L p .622 .304 .41 r
F P
0 g
s
.60839 1.15909 m .59033 1.15034 L .61323 1.13965 L p .622 .304 .41 r
F P
0 g
s
.36097 1.04787 m .37261 1.0229 L .3449 1.03439 L p .402 .193 .478 r
F P
0 g
s
.3449 1.03439 m .33138 1.05843 L .36097 1.04787 L p .402 .193 .478 r
F P
0 g
s
.37439 .59277 m .40911 .59544 L .3842 .61784 L p .641 .833 .974 r
F P
0 g
s
.3842 .61784 m .34302 .61574 L .37439 .59277 L p .641 .833 .974 r
F P
0 g
s
.44215 .97525 m .45773 .96796 L .42921 .98165 L p .377 .84 .796 r
F P
0 g
s
.42921 .98165 m .40797 .98864 L .44215 .97525 L p .377 .84 .796 r
F P
0 g
s
.51073 .88504 m .51905 .8517 L .53731 .8349 L p .415 .164 .422 r
F P
0 g
s
.51905 .8517 m .51073 .88504 L .48513 .90089 L p .161 0 .216 r
F P
0 g
s
.56145 .93296 m .57073 .92415 L .60379 .90846 L p .694 .969 .93 r
F P
0 g
s
.57073 .92415 m .56145 .93296 L .53554 .94793 L p .756 .884 .937 r
F P
0 g
s
.46381 .92408 m .43431 .93885 L p .44358 .93192 L 0 0 0 r
F P
0 g
s
.48513 .90089 m .46381 .92408 L p .44358 .93192 L 0 0 0 r
F P
0 g
s
.46381 .92408 m .50051 .90867 L p .4817 .91714 L 0 0 0 r
F P
0 g
s
.43431 .93885 m .46381 .92408 L p .44385 .93238 L .488 .902 .845 r
F P
0 g
s
.43992 .93595 m .46381 .92408 L p .4817 .91714 L 0 0 0 r
F P
0 g
s
.46442 .91622 m .43431 .93885 L p .42364 .94608 L .747 .991 .88 r
F P
0 g
s
.40067 .95034 m .43431 .93885 L p .44385 .93238 L .488 .902 .845 r
F P
0 g
s
.43431 .93885 m .41418 .953 L p .42364 .94608 L .747 .991 .88 r
F P
0 g
s
.37416 .964 m .41418 .953 L .43431 .93885 L p .64 .941 .96 r
F P
0 g
s
.55577 .81775 m .56795 .85198 L .53914 .86872 L p .537 .301 .504 r
F P
0 g
s
.53914 .86872 m .53731 .8349 L .55577 .81775 L p .537 .301 .504 r
F P
0 g
s
.53554 .94793 m .53603 .9394 L .57073 .92415 L p .756 .884 .937 r
F P
0 g
s
.37432 .54977 m .42394 .5217 L .44407 .49718 L p .775 .991 .888 r
F P
0 g
s
.69291 1.13639 m .72346 1.127 L .69018 1.13937 L p 0 0 0 r
F P
0 g
s
.69018 1.13937 m .66491 1.14823 L .69291 1.13639 L p 0 0 0 r
F P
0 g
s
.50305 .56594 m .51285 .59295 L .48966 .61419 L p .719 .73 .859 r
F P
0 g
s
.48966 .61419 m .46922 .58803 L .50305 .56594 L p .719 .73 .859 r
F P
0 g
s
.403 1.19386 m .38435 1.17925 L .35854 1.18565 L p .537 .63 .891 r
F P
0 g
s
.34102 1.16547 m .35854 1.18565 L .38435 1.17925 L p .523 .557 .832 r
F P
0 g
s
.46984 1.19443 m .48528 1.18653 L .50884 1.17846 L p .08 0 0 r
F P
0 g
s
.53731 .8349 m .53914 .86872 L .51073 .88504 L p .415 .164 .422 r
F P
0 g
s
.35456 1.0759 m .36097 1.04787 L .33138 1.05843 L p .454 .307 .591 r
F P
0 g
s
.33138 1.05843 m .32394 1.08545 L .35456 1.0759 L p .454 .307 .591 r
F P
0 g
s
.50274 .95415 m .47362 .96841 L p .48929 .95736 L .669 .892 .978 r
F P
0 g
s
.47362 .96841 m .45773 .96796 L .49445 .95373 L p .669 .892 .978 r
F P
0 g
s
.44215 .97525 m .48632 .96127 L p .46691 .96581 L .065 0 0 r
F P
0 g
s
.45773 .96796 m .44318 .97477 L p .46691 .96581 L .065 0 0 r
F P
0 g
s
.44318 .97477 m .44215 .97525 L p .46691 .96581 L .065 0 0 r
F P
0 g
s
.49445 .95373 m .49838 .95393 L p .48929 .95736 L .669 .892 .978 r
F P
0 g
s
.49838 .95393 m .50274 .95415 L p .48929 .95736 L .669 .892 .978 r
F P
0 g
s
.61674 .81756 m .65304 .84195 L .62129 .85919 L p .64 .341 .448 r
F P
0 g
s
.62129 .85919 m .59462 .8349 L .61674 .81756 L p .64 .341 .448 r
F P
0 g
s
.38435 1.17925 m .36928 1.15858 L .34102 1.16547 L p .523 .557 .832 r
F P
0 g
s
.32893 1.14089 m .34102 1.16547 L .36928 1.15858 L p .511 .496 .776 r
F P
0 g
s
.76757 .92324 m .74417 .89948 L .70733 .91582 L p .776 .502 .521 r
F P
0 g
s
.68399 .89956 m .70733 .91582 L .74417 .89948 L p .727 .415 .451 r
F P
0 g
s
.35384 1.10504 m .35456 1.0759 L .32394 1.08545 L p .481 .381 .664 r
F P
0 g
s
.32394 1.08545 m .3231 1.11359 L .35384 1.10504 L p .481 .381 .664 r
F P
0 g
s
.36856 .04309 m .41364 .06991 L .38715 .10474 L p .633 .746 .929 r
F P
0 g
s
.38715 .10474 m .33295 .08149 L .36856 .04309 L p .633 .746 .929 r
F P
0 g
s
.31874 .56249 m .26083 .55866 L .23892 .58512 L p .56 .825 .995 r
F P
0 g
s
.17969 .56645 m .23892 .58512 L .26083 .55866 L p .434 .716 .984 r
F P
0 g
s
.36928 1.15858 m .35886 1.13326 L .32893 1.14089 L p .511 .496 .776 r
F P
0 g
s
.3231 1.11359 m .32893 1.14089 L .35886 1.13326 L p .498 .44 .722 r
F P
0 g
s
.35886 1.13326 m .35384 1.10504 L .3231 1.11359 L p .498 .44 .722 r
F P
0 g
s
.59033 1.15034 m .60839 1.15909 L .57317 1.16912 L p .576 .244 .37 r
F P
0 g
s
.57317 1.16912 m .56369 1.16039 L .59033 1.15034 L p .576 .244 .37 r
F P
0 g
s
.44589 .5727 m .46922 .58803 L .44198 .60997 L p .677 .766 .918 r
F P
0 g
s
.44198 .60997 m .40911 .59544 L .44589 .5727 L p .677 .766 .918 r
F P
0 g
s
.72346 1.127 m .75065 1.10929 L .71271 1.12226 L p .754 .976 .921 r
F P
0 g
s
.71271 1.12226 m .69018 1.13937 L .72346 1.127 L p .754 .976 .921 r
F P
0 g
s
.5008 1.18675 m .50884 1.17846 L .53572 1.16976 L p .368 0 .095 r
F P
0 g
s
.50884 1.17846 m .5008 1.18675 L .46984 1.19443 L p .08 0 0 r
F P
0 g
s
.42797 .98985 m .44215 .97525 L .40797 .98864 L p .332 0 .153 r
F P
0 g
s
.40797 .98864 m .38865 1.0027 L .42797 .98985 L p .332 0 .153 r
F P
0 g
s
.60379 .90846 m .62239 .9078 L .65837 .89159 L p .387 0 .077 r
F P
0 g
s
.62239 .9078 m .60379 .90846 L .57073 .92415 L p .143 0 0 r
F P
0 g
s
.68399 .89956 m .65837 .89159 L .62239 .9078 L p .615 .268 .359 r
F P
0 g
s
.5407 .54386 m .53857 .5716 L .51285 .59295 L p .76 .779 .869 r
F P
0 g
s
.51285 .59295 m .50305 .56594 L .5407 .54386 L p .76 .779 .869 r
F P
0 g
s
.46984 1.19443 m .45389 1.19464 L .42395 1.20142 L p .574 .863 .996 r
F P
0 g
s
.403 1.19386 m .42395 1.20142 L .45389 1.19464 L p .607 .741 .94 r
F P
0 g
s
.63866 1.14831 m .66491 1.14823 L .62706 1.15932 L p .39 0 .02 r
F P
0 g
s
.62706 1.15932 m .60839 1.15909 L .63866 1.14831 L p .39 0 .02 r
F P
0 g
s
.57892 .52202 m .56462 .55027 L .53857 .5716 L p .801 .848 .886 r
F P
0 g
s
.12265 .14955 m .16646 .10897 L .13983 .15493 L p .698 .965 .727 r
F P
0 g
s
.13983 .15493 m .09431 .19602 L .12265 .14955 L p .698 .965 .727 r
F P
0 g
s
.78456 .95273 m .76757 .92324 L .72675 .93932 L p .807 .569 .578 r
F P
0 g
s
.70733 .91582 m .72675 .93932 L .76757 .92324 L p .776 .502 .521 r
F P
0 g
s
.53603 .9394 m .50274 .95415 L p .52821 .94269 L .575 .925 .954 r
F P
0 g
s
.50274 .95415 m .49445 .95373 L .53655 .93894 L p .575 .925 .954 r
F P
0 g
s
.48632 .96127 m .53706 .94667 L p .50813 .95146 L .355 0 .047 r
F P
0 g
s
.49445 .95373 m .48689 .96074 L p .50813 .95146 L .355 0 .047 r
F P
0 g
s
.48689 .96074 m .48632 .96127 L p .50813 .95146 L .355 0 .047 r
F P
0 g
s
.53655 .93894 m .5363 .93915 L p .52821 .94269 L .575 .925 .954 r
F P
0 g
s
.5363 .93915 m .53603 .9394 L p .52821 .94269 L .575 .925 .954 r
F P
0 g
s
.80803 .45039 m .75979 .43987 L .72503 .45898 L p .532 .044 .042 r
F P
0 g
s
.68314 .46276 m .72503 .45898 L .75979 .43987 L p .288 0 0 r
F P
0 g
s
.56369 1.16039 m .57317 1.16912 L .53623 1.17833 L p .504 .139 .281 r
F P
0 g
s
.53623 1.17833 m .53572 1.16976 L .56369 1.16039 L p .504 .139 .281 r
F P
0 g
s
.53572 1.16976 m .53623 1.17833 L .5008 1.18675 L p .368 0 .095 r
F P
0 g
s
.53857 .5716 m .5407 .54386 L .57892 .52202 L p .801 .848 .886 r
F P
0 g
s
.75065 1.10929 m .77259 1.08444 L .73092 1.09804 L p .855 .881 .857 r
F P
0 g
s
.73092 1.09804 m .71271 1.12226 L .75065 1.10929 L p .855 .881 .857 r
F P
0 g
s
.45389 1.19464 m .46984 1.19443 L .5008 1.18675 L p .517 .882 .978 r
F P
0 g
s
.34018 .60234 m .37439 .59277 L .34302 .61574 L p .584 .927 .956 r
F P
0 g
s
.34302 .61574 m .30261 .62554 L .34018 .60234 L p .584 .927 .956 r
F P
0 g
s
.72297 .83654 m .77236 .83505 L .73579 .85273 L p .37 0 0 r
F P
0 g
s
.73579 .85273 m .69362 .85403 L .72297 .83654 L p .37 0 0 r
F P
0 g
s
.48632 .96127 m .49445 .95373 L .45773 .96796 L p .065 0 0 r
F P
0 g
s
.57073 .92415 m .58049 .92361 L .62239 .9078 L p .143 0 0 r
F P
0 g
s
.58049 .92361 m .57073 .92415 L .53603 .9394 L p .272 .766 .706 r
F P
0 g
s
.43431 .93885 m .40067 .95034 L .37416 .964 L p .64 .941 .96 r
F P
0 g
s
.33278 .96316 m .37416 .964 L .40067 .95034 L p .565 .853 .997 r
F P
0 g
s
.53603 .9394 m .53655 .93894 L .58049 .92361 L p .272 .766 .706 r
F P
0 g
s
.79393 .98594 m .78456 .95273 L .74086 .96845 L p .829 .626 .628 r
F P
0 g
s
.72675 .93932 m .74086 .96845 L .78456 .95273 L p .807 .569 .578 r
F P
0 g
s
.41618 1.01079 m .42797 .98985 L .38865 1.0027 L p .499 .245 .46 r
F P
0 g
s
.38865 1.0027 m .37261 1.0229 L .41618 1.01079 L p .499 .245 .46 r
F P
0 g
s
.26083 .55866 m .20458 .53817 L .17969 .56645 L p .434 .716 .984 r
F P
0 g
s
.12687 .53289 m .17969 .56645 L .20458 .53817 L p .313 .583 .939 r
F P
0 g
s
.46381 .92408 m .48513 .90089 L .51073 .88504 L p .083 0 0 r
F P
0 g
s
.77259 1.08444 m .78774 1.05416 L .7435 1.06837 L p .863 .796 .784 r
F P
0 g
s
.7435 1.06837 m .73092 1.09804 L .77259 1.08444 L p .863 .796 .784 r
F P
0 g
s
.45389 1.19464 m .43852 1.18708 L .403 1.19386 L p .607 .741 .94 r
F P
0 g
s
.38435 1.17925 m .403 1.19386 L .43852 1.18708 L p .606 .65 .867 r
F P
0 g
s
.79502 1.02055 m .79393 .98594 L .74865 1.00122 L p .845 .678 .676 r
F P
0 g
s
.74086 .96845 m .74865 1.00122 L .79393 .98594 L p .829 .626 .628 r
F P
0 g
s
.59462 .8349 m .62129 .85919 L .5829 .87614 L p .581 .281 .426 r
F P
0 g
s
.5829 .87614 m .56795 .85198 L .59462 .8349 L p .581 .281 .426 r
F P
0 g
s
.78774 1.05416 m .79502 1.02055 L .74956 1.03533 L p .856 .732 .726 r
F P
0 g
s
.74956 1.03533 m .7435 1.06837 L .78774 1.05416 L p .856 .732 .726 r
F P
0 g
s
.74865 1.00122 m .74956 1.03533 L .79502 1.02055 L p .845 .678 .676 r
F P
0 g
s
.70733 .91582 m .68399 .89956 L .64063 .91573 L p .698 .413 .487 r
F P
0 g
s
.62239 .9078 m .64063 .91573 L .68399 .89956 L p .615 .268 .359 r
F P
0 g
s
.66491 1.14823 m .69018 1.13937 L .64506 1.15091 L p .514 .912 .906 r
F P
0 g
s
.64506 1.15091 m .62706 1.15932 L .66491 1.14823 L p .514 .912 .906 r
F P
0 g
s
.60839 1.15909 m .62706 1.15932 L .58298 1.16952 L p .204 0 0 r
F P
0 g
s
.57317 1.16912 m .60839 1.15909 L p .58631 1.16815 L .204 0 0 r
F P
0 g
s
.64506 1.15091 m .59244 1.16147 L p .61954 1.15979 L .658 .908 .98 r
F P
0 g
s
.62706 1.15932 m .63368 1.15623 L p .61954 1.15979 L .658 .908 .98 r
F P
0 g
s
.63368 1.15623 m .64506 1.15091 L p .61954 1.15979 L .658 .908 .98 r
F P
0 g
s
.58298 1.16952 m .58118 1.16944 L p .58631 1.16815 L .204 0 0 r
F P
0 g
s
.58118 1.16944 m .57317 1.16912 L p .58631 1.16815 L .204 0 0 r
F P
0 g
s
.09057 .20236 m .12265 .14955 L .09431 .19602 L p 0 0 0 r
F P
0 g
s
.09431 .19602 m .06117 .248 L .09057 .20236 L p 0 0 0 r
F P
0 g
s
.49248 1.18714 m .5008 1.18675 L .53623 1.17833 L p .366 .835 .876 r
F P
0 g
s
.5008 1.18675 m .49248 1.18714 L .45389 1.19464 L p .517 .882 .978 r
F P
0 g
s
.64063 .91573 m .62239 .9078 L .58049 .92361 L p .569 .212 .323 r
F P
0 g
s
.40762 1.03667 m .41618 1.01079 L .37261 1.0229 L p .549 .358 .575 r
F P
0 g
s
.37261 1.0229 m .36097 1.04787 L .40762 1.03667 L p .549 .358 .575 r
F P
0 g
s
.53706 .94667 m .53655 .93894 L .49445 .95373 L p .355 0 .047 r
F P
0 g
s
.43852 1.18708 m .42482 1.17221 L .38435 1.17925 L p .606 .65 .867 r
F P
0 g
s
.36928 1.15858 m .38435 1.17925 L .42482 1.17221 L p .601 .589 .811 r
F P
0 g
s
.43852 1.18708 m .45389 1.19464 L .49248 1.18714 L p .641 .764 .938 r
F P
0 g
s
.47891 .97634 m .48632 .96127 L .44215 .97525 L p .504 .186 .36 r
F P
0 g
s
.44215 .97525 m .42797 .98985 L .47891 .97634 L p .504 .186 .36 r
F P
0 g
s
.49184 .54985 m .50305 .56594 L .46922 .58803 L p .708 .808 .927 r
F P
0 g
s
.46922 .58803 m .44589 .5727 L .49184 .54985 L p .708 .808 .927 r
F P
0 g
s
.50051 .90867 m .51073 .88504 L .53914 .86872 L p .358 .014 .238 r
F P
0 g
s
.51073 .88504 m .50051 .90867 L .46381 .92408 L p .083 0 0 r
F P
0 g
s
.20458 .53817 m .15401 .50213 L .12687 .53289 L p .313 .583 .939 r
F P
0 g
s
.0842 .48667 m .12687 .53289 L .15401 .50213 L p .195 .431 .861 r
F P
0 g
s
.85207 .47388 m .80803 .45039 L .76638 .4687 L p .698 .259 .205 r
F P
0 g
s
.72503 .45898 m .76638 .4687 L .80803 .45039 L p .532 .044 .042 r
F P
0 g
s
.65304 .84195 m .69362 .85403 L .65121 .87137 L p .545 .155 .253 r
F P
0 g
s
.65121 .87137 m .62129 .85919 L .65304 .84195 L p .545 .155 .253 r
F P
0 g
s
.57317 1.16912 m .58298 1.16952 L .53675 1.17879 L p 0 0 0 r
F P
0 g
s
.53623 1.17833 m .57317 1.16912 L p .54146 1.17754 L 0 0 0 r
F P
0 g
s
.59244 1.16147 m .53725 1.171 L p .57312 1.16984 L .678 .846 .964 r
F P
0 g
s
.58298 1.16952 m .5864 1.16661 L p .57312 1.16984 L .678 .846 .964 r
F P
0 g
s
.5864 1.16661 m .59244 1.16147 L p .57312 1.16984 L .678 .846 .964 r
F P
0 g
s
.53675 1.17879 m .53665 1.17871 L p .54146 1.17754 L 0 0 0 r
F P
0 g
s
.53665 1.17871 m .53623 1.17833 L p .54146 1.17754 L 0 0 0 r
F P
0 g
s
.53623 1.17833 m .53675 1.17879 L .49248 1.18714 L p .366 .835 .876 r
F P
0 g
s
.56795 .85198 m .5829 .87614 L .54139 .89267 L p .497 .184 .366 r
F P
0 g
s
.54139 .89267 m .53914 .86872 L .56795 .85198 L p .497 .184 .366 r
F P
0 g
s
.30884 .62369 m .34018 .60234 L .30261 .62554 L p 0 0 0 r
F P
0 g
s
.30261 .62554 m .26575 .64667 L .30884 .62369 L p 0 0 0 r
F P
0 g
s
.44407 .49718 m .40303 .52471 L .37432 .54977 L p .775 .991 .888 r
F P
0 g
s
.31874 .56249 m .37432 .54977 L .40303 .52471 L p .671 .942 .963 r
F P
0 g
s
.4029 1.06569 m .40762 1.03667 L .36097 1.04787 L p .571 .421 .639 r
F P
0 g
s
.36097 1.04787 m .35456 1.0759 L .4029 1.06569 L p .571 .421 .639 r
F P
0 g
s
.42117 .56943 m .44589 .5727 L .40911 .59544 L p .648 .872 .982 r
F P
0 g
s
.40911 .59544 m .37439 .59277 L .42117 .56943 L p .648 .872 .982 r
F P
0 g
s
.77236 .83505 m .81999 .82002 L .77666 .83789 L p 0 0 0 r
F P
0 g
s
.77666 .83789 m .73579 .85273 L .77236 .83505 L p 0 0 0 r
F P
0 g
s
.40067 .95034 m .36575 .94972 L .33278 .96316 L p .565 .853 .997 r
F P
0 g
s
.29288 .95042 m .33278 .96316 L .36575 .94972 L p .514 .746 .979 r
F P
0 g
s
.42482 1.17221 m .41374 1.15101 L .36928 1.15858 L p .601 .589 .811 r
F P
0 g
s
.35886 1.13326 m .36928 1.15858 L .41374 1.15101 L p .596 .543 .765 r
F P
0 g
s
.58049 .92361 m .59008 .93147 L .64063 .91573 L p .569 .212 .323 r
F P
0 g
s
.59008 .93147 m .58049 .92361 L .53655 .93894 L p .494 .107 .234 r
F P
0 g
s
.07274 .2637 m .09057 .20236 L .06117 .248 L p 0 0 .025 r
F P
0 g
s
.06117 .248 m .04282 .3073 L .07274 .2637 L p 0 0 .025 r
F P
0 g
s
.53655 .93894 m .53706 .94667 L .59008 .93147 L p .494 .107 .234 r
F P
0 g
s
.53914 .86872 m .54139 .89267 L .50051 .90867 L p .358 .014 .238 r
F P
0 g
s
.72675 .93932 m .70733 .91582 L .65727 .9318 L p .737 .494 .561 r
F P
0 g
s
.64063 .91573 m .65727 .9318 L .70733 .91582 L p .698 .413 .487 r
F P
0 g
s
.69018 1.13937 m .71271 1.12226 L .66112 1.13437 L p .778 .864 .912 r
F P
0 g
s
.66112 1.13437 m .64506 1.15091 L .69018 1.13937 L p .778 .864 .912 r
F P
0 g
s
.72503 .45898 m .68314 .46276 L .64345 .48347 L p .111 0 0 r
F P
0 g
s
.40237 1.09582 m .4029 1.06569 L .35456 1.0759 L p .583 .465 .685 r
F P
0 g
s
.35456 1.0759 m .35384 1.10504 L .40237 1.09582 L p .583 .465 .685 r
F P
0 g
s
.41374 1.15101 m .40607 1.12494 L .35886 1.13326 L p .596 .543 .765 r
F P
0 g
s
.35384 1.10504 m .35886 1.13326 L .40607 1.12494 L p .59 .504 .725 r
F P
0 g
s
.3204 .03219 m .36856 .04309 L .33295 .08149 L p .645 .835 .973 r
F P
0 g
s
.33295 .08149 m .2755 .07407 L .3204 .03219 L p .645 .835 .973 r
F P
0 g
s
.15401 .50213 m .11288 .45285 L .0842 .48667 L p .195 .431 .861 r
F P
0 g
s
.05474 .431 m .0842 .48667 L .11288 .45285 L p .074 .256 .746 r
F P
0 g
s
.40607 1.12494 m .40237 1.09582 L .35384 1.10504 L p .59 .504 .725 r
F P
0 g
s
.59244 1.16147 m .58298 1.16952 L .62706 1.15932 L p .658 .908 .98 r
F P
0 g
s
.0706 .32911 m .07274 .2637 L .04282 .3073 L p 0 0 .351 r
F P
0 g
s
.04282 .3073 m .04061 .36977 L .0706 .32911 L p 0 0 .351 r
F P
0 g
s
.49248 1.18714 m .48446 1.17951 L .43852 1.18708 L p .641 .764 .938 r
F P
0 g
s
.42482 1.17221 m .43852 1.18708 L .48446 1.17951 L p .649 .669 .855 r
F P
0 g
s
.48446 1.17951 m .49248 1.18714 L .53675 1.17879 L p .665 .797 .946 r
F P
0 g
s
.47275 .99797 m .47891 .97634 L .42797 .98985 L p .586 .355 .533 r
F P
0 g
s
.42797 .98985 m .41618 1.01079 L .47275 .99797 L p .586 .355 .533 r
F P
0 g
s
.65727 .9318 m .64063 .91573 L .59008 .93147 L p .668 .394 .496 r
F P
0 g
s
.11288 .45285 m .08433 .39375 L .05474 .431 L p .074 .256 .746 r
F P
0 g
s
.04061 .36977 m .05474 .431 L .08433 .39375 L p 0 .05 .583 r
F P
0 g
s
.54315 .52719 m .5407 .54386 L .50305 .56594 L p .733 .863 .942 r
F P
0 g
s
.50305 .56594 m .49184 .54985 L .54315 .52719 L p .733 .863 .942 r
F P
0 g
s
.53753 .96215 m .53706 .94667 L .48632 .96127 L p .584 .296 .446 r
F P
0 g
s
.48632 .96127 m .47891 .97634 L .53753 .96215 L p .584 .296 .446 r
F P
0 g
s
.08433 .39375 m .0706 .32911 L .04061 .36977 L p 0 .05 .583 r
F P
0 g
s
.5407 .54386 m .54315 .52719 L .59529 .50499 L p .737 .929 .953 r
F P
0 g
s
.59529 .50499 m .54315 .52719 L p .55958 .52092 L .398 .854 .825 r
F P
0 g
s
.57892 .52202 m .59529 .50499 L p .64318 .48363 L .67 .97 .915 r
F P
0 g
s
.59529 .50499 m .57892 .52202 L .5407 .54386 L p .737 .929 .953 r
F P
0 g
s
.6127 .50062 m .59529 .50499 L p .55958 .52092 L .398 .854 .825 r
F P
0 g
s
.59529 .50499 m .64345 .48347 L p .64318 .48363 L .67 .97 .915 r
F P
0 g
s
.74086 .96845 m .72675 .93932 L .67113 .95496 L p .759 .549 .612 r
F P
0 g
s
.65727 .9318 m .67113 .95496 L .72675 .93932 L p .737 .494 .561 r
F P
0 g
s
.71271 1.12226 m .73092 1.09804 L .67411 1.11079 L p .796 .768 .827 r
F P
0 g
s
.67411 1.11079 m .66112 1.13437 L .71271 1.12226 L p .796 .768 .827 r
F P
0 g
s
.88878 .50884 m .85207 .47388 L .8043 .49152 L p .803 .425 .34 r
F P
0 g
s
.76638 .4687 m .8043 .49152 L .85207 .47388 L p .698 .259 .205 r
F P
0 g
s
.28265 .65551 m .30884 .62369 L .26575 .64667 L p .056 0 .01 r
F P
0 g
s
.26575 .64667 m .23504 .6778 L .28265 .65551 L p .056 0 .01 r
F P
0 g
s
.53725 1.171 m .53675 1.17879 L .58298 1.16952 L p .678 .846 .964 r
F P
0 g
s
.53675 1.17879 m .53725 1.171 L .48446 1.17951 L p .665 .797 .946 r
F P
0 g
s
.36575 .94972 m .33193 .93684 L .29288 .95042 L p .514 .746 .979 r
F P
0 g
s
.25725 .92654 m .29288 .95042 L .33193 .93684 L p .478 .642 .928 r
F P
0 g
s
.81999 .82002 m .86253 .79235 L .81333 .81032 L p .648 .968 .836 r
F P
0 g
s
.81333 .81032 m .77666 .83789 L .81999 .82002 L p .648 .968 .836 r
F P
0 g
s
.46381 .92408 m .43992 .93595 L .40067 .95034 L p .488 .902 .845 r
F P
0 g
s
.59008 .93147 m .59883 .94728 L .65727 .9318 L p .668 .394 .496 r
F P
0 g
s
.59883 .94728 m .59008 .93147 L .53706 .94667 L p .632 .357 .484 r
F P
0 g
s
.48446 1.17951 m .47729 1.16432 L .42482 1.17221 L p .649 .669 .855 r
F P
0 g
s
.41374 1.15101 m .42482 1.17221 L .47729 1.16432 L p .647 .612 .8 r
F P
0 g
s
.53706 .94667 m .53753 .96215 L .59883 .94728 L p .632 .357 .484 r
F P
0 g
s
.46827 1.02469 m .47275 .99797 L .41618 1.01079 L p .614 .429 .609 r
F P
0 g
s
.41618 1.01079 m .40762 1.03667 L .46827 1.02469 L p .614 .429 .609 r
F P
0 g
s
.36575 .94972 m .40067 .95034 L .43992 .93595 L p .557 .882 .99 r
F P
0 g
s
.74865 1.00122 m .74086 .96845 L .68121 .98365 L p .773 .591 .653 r
F P
0 g
s
.67113 .95496 m .68121 .98365 L .74086 .96845 L p .759 .549 .612 r
F P
0 g
s
.73092 1.09804 m .7435 1.06837 L .6831 1.08178 L p .794 .709 .77 r
F P
0 g
s
.6831 1.08178 m .67411 1.11079 L .73092 1.09804 L p .794 .709 .77 r
F P
0 g
s
.62129 .85919 m .65121 .87137 L .5997 .88839 L p .465 .06 .187 r
F P
0 g
s
.5997 .88839 m .5829 .87614 L .62129 .85919 L p .465 .06 .187 r
F P
0 g
s
.64506 1.15091 m .66112 1.13437 L .60088 1.14545 L p .747 .779 .879 r
F P
0 g
s
.60088 1.14545 m .59244 1.16147 L .64506 1.15091 L p .747 .779 .879 r
F P
0 g
s
.47729 1.16432 m .48446 1.17951 L .53725 1.171 L p .683 .693 .853 r
F P
0 g
s
.67113 .95496 m .65727 .9318 L .59883 .94728 L p .704 .477 .577 r
F P
0 g
s
.53792 .98438 m .53753 .96215 L .47891 .97634 L p .635 .414 .565 r
F P
0 g
s
.47891 .97634 m .47275 .99797 L .53792 .98438 L p .635 .414 .565 r
F P
0 g
s
.74956 1.03533 m .74865 1.00122 L .68678 1.01588 L p .782 .628 .69 r
F P
0 g
s
.68121 .98365 m .68678 1.01588 L .74865 1.00122 L p .773 .591 .653 r
F P
0 g
s
.69362 .85403 m .73579 .85273 L .68242 .87031 L p .238 0 0 r
F P
0 g
s
.68242 .87031 m .65121 .87137 L .69362 .85403 L p .238 0 0 r
F P
0 g
s
.39671 .57873 m .42117 .56943 L .37439 .59277 L p .424 .869 .842 r
F P
0 g
s
.37439 .59277 m .34018 .60234 L .39671 .57873 L p .424 .869 .842 r
F P
0 g
s
.7435 1.06837 m .74956 1.03533 L .68743 1.04938 L p .789 .665 .727 r
F P
0 g
s
.68743 1.04938 m .6831 1.08178 L .7435 1.06837 L p .789 .665 .727 r
F P
0 g
s
.47729 1.16432 m .47149 1.14257 L .41374 1.15101 L p .647 .612 .8 r
F P
0 g
s
.40607 1.12494 m .41374 1.15101 L .47149 1.14257 L p .644 .572 .759 r
F P
0 g
s
.4658 1.05465 m .46827 1.02469 L .40762 1.03667 L p .627 .474 .656 r
F P
0 g
s
.40762 1.03667 m .4029 1.06569 L .4658 1.05465 L p .627 .474 .656 r
F P
0 g
s
.68678 1.01588 m .68743 1.04938 L .74956 1.03533 L p .782 .628 .69 r
F P
0 g
s
.26353 .69565 m .28265 .65551 L .23504 .6778 L p .238 .001 .339 r
F P
0 g
s
.23504 .6778 m .2127 .71682 L .26353 .69565 L p .238 .001 .339 r
F P
0 g
s
.47993 .54605 m .49184 .54985 L .44589 .5727 L p .631 .909 .983 r
F P
0 g
s
.44589 .5727 m .42117 .56943 L .47993 .54605 L p .631 .909 .983 r
F P
0 g
s
.91553 .55293 m .88878 .50884 L .83604 .52605 L p .868 .552 .45 r
F P
0 g
s
.8043 .49152 m .83604 .52605 L .88878 .50884 L p .803 .425 .34 r
F P
0 g
s
.33193 .93684 m .30161 .91239 L .25725 .92654 L p .478 .642 .928 r
F P
0 g
s
.22843 .89311 m .25725 .92654 L .30161 .91239 L p .451 .548 .864 r
F P
0 g
s
.86253 .79235 m .89693 .75389 L .8431 .77179 L p .862 .98 .854 r
F P
0 g
s
.8431 .77179 m .81333 .81032 L .86253 .79235 L p .862 .98 .854 r
F P
0 g
s
.59244 1.16147 m .60088 1.14545 L .53769 1.15543 L p .715 .727 .86 r
F P
0 g
s
.53769 1.15543 m .53725 1.171 L .59244 1.16147 L p .715 .727 .86 r
F P
0 g
s
.53725 1.171 m .53769 1.15543 L .47729 1.16432 L p .683 .693 .853 r
F P
0 g
s
.47149 1.14257 m .46747 1.11575 L .40607 1.12494 L p .644 .572 .759 r
F P
0 g
s
.40237 1.09582 m .40607 1.12494 L .46747 1.11575 L p .64 .54 .725 r
F P
0 g
s
.46553 1.08573 m .4658 1.05465 L .4029 1.06569 L p .635 .509 .692 r
F P
0 g
s
.4029 1.06569 m .40237 1.09582 L .46553 1.08573 L p .635 .509 .692 r
F P
0 g
s
.76638 .4687 m .72503 .45898 L .67442 .47923 L p .488 .013 .053 r
F P
0 g
s
.64345 .48347 m .67442 .47923 L .72503 .45898 L p .111 0 0 r
F P
0 g
s
.40303 .52471 m .3567 .53654 L .31874 .56249 L p .671 .942 .963 r
F P
0 g
s
.26083 .55866 m .31874 .56249 L .3567 .53654 L p .584 .858 .995 r
F P
0 g
s
.48902 .9208 m .50051 .90867 L .54139 .89267 L p .131 0 0 r
F P
0 g
s
.50051 .90867 m .48902 .9208 L .43992 .93595 L p 0 0 0 r
F P
0 g
s
.46747 1.11575 m .46553 1.08573 L .40237 1.09582 L p .64 .54 .725 r
F P
0 g
s
.67442 .47923 m .64345 .48347 L .59529 .50499 L p 0 0 0 r
F P
0 g
s
.59883 .94728 m .60612 .97002 L .67113 .95496 L p .704 .477 .577 r
F P
0 g
s
.60612 .97002 m .59883 .94728 L .53753 .96215 L p .672 .451 .577 r
F P
0 g
s
.53753 .96215 m .53792 .98438 L .60612 .97002 L p .672 .451 .577 r
F P
0 g
s
.66112 1.13437 m .67411 1.11079 L .60771 1.1225 L p .751 .704 .807 r
F P
0 g
s
.60771 1.1225 m .60088 1.14545 L .66112 1.13437 L p .751 .704 .807 r
F P
0 g
s
.47149 1.14257 m .47729 1.16432 L .53769 1.15543 L p .683 .635 .796 r
F P
0 g
s
.5829 .87614 m .5997 .88839 L .54391 .90492 L p .34 0 .052 r
F P
0 g
s
.54391 .90492 m .54139 .89267 L .5829 .87614 L p .34 0 .052 r
F P
0 g
s
.68121 .98365 m .67113 .95496 L .60612 .97002 L p .722 .527 .626 r
F P
0 g
s
.53821 1.01186 m .53792 .98438 L .47275 .99797 L p .655 .472 .624 r
F P
0 g
s
.47275 .99797 m .46827 1.02469 L .53821 1.01186 L p .655 .472 .624 r
F P
0 g
s
.25295 .74132 m .26353 .69565 L .2127 .71682 L p .324 .173 .517 r
F P
0 g
s
.2127 .71682 m .20035 .76104 L .25295 .74132 L p .324 .173 .517 r
F P
0 g
s
.27238 .03849 m .3204 .03219 L .2755 .07407 L p .64 .926 .974 r
F P
0 g
s
.2755 .07407 m .21868 .08332 L .27238 .03849 L p .64 .926 .974 r
F P
0 g
s
.30161 .91239 m .277 .87795 L .22843 .89311 L p .451 .548 .864 r
F P
0 g
s
.2085 .85242 m .22843 .89311 L .277 .87795 L p .427 .463 .794 r
F P
0 g
s
.93035 .60309 m .91553 .55293 L .85925 .56998 L p .908 .655 .545 r
F P
0 g
s
.83604 .52605 m .85925 .56998 L .91553 .55293 L p .868 .552 .45 r
F P
0 g
s
.54139 .89267 m .54391 .90492 L .48902 .9208 L p .131 0 0 r
F P
0 g
s
.89693 .75389 m .92071 .70728 L .86374 .72496 L p .925 .908 .787 r
F P
0 g
s
.86374 .72496 m .8431 .77179 L .89693 .75389 L p .925 .908 .787 r
F P
0 g
s
.25171 .78923 m .25295 .74132 L .20035 .76104 L p .371 .288 .632 r
F P
0 g
s
.20035 .76104 m .19891 .80733 L .25171 .78923 L p .371 .288 .632 r
F P
0 g
s
.67411 1.11079 m .6831 1.08178 L .61244 1.0942 L p .749 .658 .761 r
F P
0 g
s
.61244 1.0942 m .60771 1.1225 L .67411 1.11079 L p .749 .658 .761 r
F P
0 g
s
.277 .87795 m .25994 .8359 L .2085 .85242 L p .427 .463 .794 r
F P
0 g
s
.19891 .80733 m .2085 .85242 L .25994 .8359 L p .403 .38 .719 r
F P
0 g
s
.53805 1.1331 m .53769 1.15543 L .60088 1.14545 L p .716 .664 .798 r
F P
0 g
s
.53769 1.15543 m .53805 1.1331 L .47149 1.14257 L p .683 .635 .796 r
F P
0 g
s
.60088 1.14545 m .60771 1.1225 L .53805 1.1331 L p .716 .664 .798 r
F P
0 g
s
.68678 1.01588 m .68121 .98365 L .61143 .99816 L p .733 .563 .663 r
F P
0 g
s
.60612 .97002 m .61143 .99816 L .68121 .98365 L p .722 .527 .626 r
F P
0 g
s
.46747 1.11575 m .47149 1.14257 L .53805 1.1331 L p .68 .597 .756 r
F P
0 g
s
.93212 .65579 m .93035 .60309 L .87212 .62024 L p .931 .744 .629 r
F P
0 g
s
.85925 .56998 m .87212 .62024 L .93035 .60309 L p .908 .655 .545 r
F P
0 g
s
.43992 .93595 m .41503 .93546 L .36575 .94972 L p .557 .882 .99 r
F P
0 g
s
.33193 .93684 m .36575 .94972 L .41503 .93546 L p .573 .779 .976 r
F P
0 g
s
.53838 1.04268 m .53821 1.01186 L .46827 1.02469 L p .666 .51 .663 r
F P
0 g
s
.46827 1.02469 m .4658 1.05465 L .53838 1.04268 L p .666 .51 .663 r
F P
0 g
s
.25994 .8359 m .25171 .78923 L .19891 .80733 L p .403 .38 .719 r
F P
0 g
s
.92071 .70728 m .93212 .65579 L .87367 .67318 L p .938 .827 .709 r
F P
0 g
s
.87367 .67318 m .86374 .72496 L .92071 .70728 L p .938 .827 .709 r
F P
0 g
s
.61143 .99816 m .60612 .97002 L .53792 .98438 L p .689 .502 .629 r
F P
0 g
s
.54315 .52719 m .49184 .54985 L p .53303 .52933 L .562 .922 .949 r
F P
0 g
s
.49184 .54985 m .47993 .54605 L .54576 .52299 L p .562 .922 .949 r
F P
0 g
s
.46812 .5551 m .54836 .53189 L p .50151 .54158 L .098 0 0 r
F P
0 g
s
.47993 .54605 m .46955 .55401 L p .50151 .54158 L .098 0 0 r
F P
0 g
s
.46955 .55401 m .46812 .5551 L p .50151 .54158 L .098 0 0 r
F P
0 g
s
.54576 .52299 m .54462 .52483 L p .53303 .52933 L .562 .922 .949 r
F P
0 g
s
.54462 .52483 m .54315 .52719 L p .53303 .52933 L .562 .922 .949 r
F P
0 g
s
.53792 .98438 m .53821 1.01186 L .61143 .99816 L p .689 .502 .629 r
F P
0 g
s
.87212 .62024 m .87367 .67318 L .93212 .65579 L p .931 .744 .629 r
F P
0 g
s
.6831 1.08178 m .68743 1.04938 L .61471 1.06253 L p .745 .624 .726 r
F P
0 g
s
.61471 1.06253 m .61244 1.0942 L .6831 1.08178 L p .745 .624 .726 r
F P
0 g
s
.41503 .93546 m .43992 .93595 L .48902 .9208 L p .509 .893 .959 r
F P
0 g
s
.59529 .50499 m .6127 .50062 L .67442 .47923 L p 0 0 0 r
F P
0 g
s
.68743 1.04938 m .68678 1.01588 L .61437 1.02974 L p .74 .594 .694 r
F P
0 g
s
.61143 .99816 m .61437 1.02974 L .68678 1.01588 L p .733 .563 .663 r
F P
0 g
s
.37424 .60023 m .39671 .57873 L .34018 .60234 L p .106 0 0 r
F P
0 g
s
.34018 .60234 m .30884 .62369 L .37424 .60023 L p .106 0 0 r
F P
0 g
s
.53805 1.1331 m .53829 1.10553 L .46747 1.11575 L p .68 .597 .756 r
F P
0 g
s
.46553 1.08573 m .46747 1.11575 L .53829 1.10553 L p .677 .567 .724 r
F P
0 g
s
.5384 1.07466 m .53838 1.04268 L .4658 1.05465 L p .672 .539 .695 r
F P
0 g
s
.4658 1.05465 m .46553 1.08573 L .5384 1.07466 L p .672 .539 .695 r
F P
0 g
s
.61437 1.02974 m .61471 1.06253 L .68743 1.04938 L p .74 .594 .694 r
F P
0 g
s
.53829 1.10553 m .5384 1.07466 L .46553 1.08573 L p .677 .567 .724 r
F P
0 g
s
.73579 .85273 m .77666 .83789 L .71277 .85577 L p .422 .866 .813 r
F P
0 g
s
.71277 .85577 m .68242 .87031 L .73579 .85273 L p .422 .866 .813 r
F P
0 g
s
.54315 .52719 m .54576 .52299 L .6127 .50062 L p .398 .854 .825 r
F P
0 g
s
.53829 1.10553 m .53805 1.1331 L .60771 1.1225 L p .713 .624 .757 r
F P
0 g
s
.60771 1.1225 m .61244 1.0942 L .53829 1.10553 L p .713 .624 .757 r
F P
0 g
s
.61437 1.02974 m .61143 .99816 L .53821 1.01186 L p .699 .537 .665 r
F P
0 g
s
.65121 .87137 m .68242 .87031 L .61725 .88758 L p .045 0 0 r
F P
0 g
s
.5997 .88839 m .65121 .87137 L p .62257 .88504 L .045 0 0 r
F P
0 g
s
.71277 .85577 m .63436 .87341 L p .67077 .87106 L .588 .931 .952 r
F P
0 g
s
.68242 .87031 m .69633 .86365 L p .67077 .87106 L .588 .931 .952 r
F P
0 g
s
.69633 .86365 m .71277 .85577 L p .67077 .87106 L .588 .931 .952 r
F P
0 g
s
.61725 .88758 m .61317 .88777 L p .62257 .88504 L .045 0 0 r
F P
0 g
s
.61317 .88777 m .5997 .88839 L p .62257 .88504 L .045 0 0 r
F P
0 g
s
.53821 1.01186 m .53838 1.04268 L .61437 1.02974 L p .699 .537 .665 r
F P
0 g
s
.8043 .49152 m .76638 .4687 L .7051 .48842 L p .669 .275 .291 r
F P
0 g
s
.67442 .47923 m .7051 .48842 L .76638 .4687 L p .488 .013 .053 r
F P
0 g
s
.46812 .5551 m .47993 .54605 L .42117 .56943 L p 0 0 0 r
F P
0 g
s
.42117 .56943 m .39671 .57873 L .46812 .5551 L p 0 0 0 r
F P
0 g
s
.5384 1.07466 m .53829 1.10553 L .61244 1.0942 L p .709 .594 .725 r
F P
0 g
s
.61244 1.0942 m .61471 1.06253 L .5384 1.07466 L p .709 .594 .725 r
F P
0 g
s
.61471 1.06253 m .61437 1.02974 L .53838 1.04268 L p .705 .566 .695 r
F P
0 g
s
.53838 1.04268 m .5384 1.07466 L .61471 1.06253 L p .705 .566 .695 r
F P
0 g
s
.3567 .53654 m .30806 .53133 L .26083 .55866 L p .584 .858 .995 r
F P
0 g
s
.20458 .53817 m .26083 .55866 L .30806 .53133 L p .515 .756 .984 r
F P
0 g
s
.7051 .48842 m .67442 .47923 L .6127 .50062 L p .411 0 .004 r
F P
0 g
s
.41503 .93546 m .39083 .92234 L .33193 .93684 L p .573 .779 .976 r
F P
0 g
s
.30161 .91239 m .33193 .93684 L .39083 .92234 L p .568 .68 .915 r
F P
0 g
s
.2279 .06214 m .27238 .03849 L .21868 .08332 L p .549 .927 .809 r
F P
0 g
s
.21868 .08332 m .16646 .10897 L .2279 .06214 L p .549 .927 .809 r
F P
0 g
s
.47703 .9203 m .48902 .9208 L .54391 .90492 L p .392 .851 .864 r
F P
0 g
s
.48902 .9208 m .47703 .9203 L .41503 .93546 L p .509 .893 .959 r
F P
0 g
s
.3554 .63264 m .37424 .60023 L .30884 .62369 L p .365 .052 .296 r
F P
0 g
s
.30884 .62369 m .28265 .65551 L .3554 .63264 L p .365 .052 .296 r
F P
0 g
s
.5997 .88839 m .61725 .88758 L .54654 .90431 L p 0 0 0 r
F P
0 g
s
.54654 .90431 m .54391 .90492 L .5997 .88839 L p 0 0 0 r
F P
0 g
s
.54391 .90492 m .54654 .90431 L .47703 .9203 L p .392 .851 .864 r
F P
0 g
s
.77666 .83789 m .81333 .81032 L .7401 .82847 L p .763 .952 .937 r
F P
0 g
s
.7401 .82847 m .71277 .85577 L .77666 .83789 L p .763 .952 .937 r
F P
0 g
s
.39083 .92234 m .41503 .93546 L .47703 .9203 L p .613 .815 .978 r
F P
0 g
s
.83604 .52605 m .8043 .49152 L .73334 .51074 L p .756 .432 .436 r
F P
0 g
s
.7051 .48842 m .73334 .51074 L .8043 .49152 L p .669 .275 .291 r
F P
0 g
s
.54836 .53189 m .54576 .52299 L .47993 .54605 L p .098 0 0 r
F P
0 g
s
.39083 .92234 m .36906 .89727 L .30161 .91239 L p .568 .68 .915 r
F P
0 g
s
.277 .87795 m .30161 .91239 L .36906 .89727 L p .56 .601 .851 r
F P
0 g
s
.6127 .50062 m .62999 .50955 L .7051 .48842 L p .411 0 .004 r
F P
0 g
s
.62999 .50955 m .6127 .50062 L .54576 .52299 L p .289 0 0 r
F P
0 g
s
.63436 .87341 m .61725 .88758 L .68242 .87031 L p .588 .931 .952 r
F P
0 g
s
.45725 .57671 m .46812 .5551 L .39671 .57873 L p .349 0 .115 r
F P
0 g
s
.39671 .57873 m .37424 .60023 L .45725 .57671 L p .349 0 .115 r
F P
0 g
s
.34162 .67381 m .3554 .63264 L .28265 .65551 L p .458 .236 .489 r
F P
0 g
s
.28265 .65551 m .26353 .69565 L .34162 .67381 L p .458 .236 .489 r
F P
0 g
s
.30806 .53133 m .26041 .50886 L .20458 .53817 L p .515 .756 .984 r
F P
0 g
s
.15401 .50213 m .20458 .53817 L .26041 .50886 L p .459 .65 .942 r
F P
0 g
s
.54576 .52299 m .54836 .53189 L .62999 .50955 L p .289 0 0 r
F P
0 g
s
.19033 .10196 m .2279 .06214 L .16646 .10897 L p 0 0 0 r
F P
0 g
s
.16646 .10897 m .12265 .14955 L .19033 .10196 L p 0 0 0 r
F P
0 g
s
.81333 .81032 m .8431 .77179 L .76235 .79012 L p .836 .872 .87 r
F P
0 g
s
.76235 .79012 m .7401 .82847 L .81333 .81032 L p .836 .872 .87 r
F P
0 g
s
.73334 .51074 m .7051 .48842 L .62999 .50955 L p .625 .254 .32 r
F P
0 g
s
.36906 .89727 m .35134 .86182 L .277 .87795 L p .56 .601 .851 r
F P
0 g
s
.25994 .8359 m .277 .87795 L .35134 .86182 L p .55 .536 .792 r
F P
0 g
s
.85925 .56998 m .83604 .52605 L .75706 .54487 L p .801 .534 .535 r
F P
0 g
s
.73334 .51074 m .75706 .54487 L .83604 .52605 L p .756 .432 .436 r
F P
0 g
s
.33397 .72083 m .34162 .67381 L .26353 .69565 L p .5 .341 .597 r
F P
0 g
s
.26353 .69565 m .25295 .74132 L .33397 .72083 L p .5 .341 .597 r
F P
0 g
s
.46533 .90688 m .47703 .9203 L .54654 .90431 L p .638 .857 .983 r
F P
0 g
s
.47703 .9203 m .46533 .90688 L .39083 .92234 L p .613 .815 .978 r
F P
0 g
s
.36906 .89727 m .39083 .92234 L .46533 .90688 L p .63 .715 .908 r
F P
0 g
s
.61725 .88758 m .63436 .87341 L .54911 .89052 L p .64 .903 .984 r
F P
0 g
s
.54911 .89052 m .54654 .90431 L .61725 .88758 L p .64 .903 .984 r
F P
0 g
s
.35134 .86182 m .33903 .81844 L .25994 .8359 L p .55 .536 .792 r
F P
0 g
s
.25171 .78923 m .25994 .8359 L .33903 .81844 L p .539 .476 .734 r
F P
0 g
s
.33308 .77026 m .33397 .72083 L .25295 .74132 L p .524 .415 .672 r
F P
0 g
s
.25295 .74132 m .25171 .78923 L .33308 .77026 L p .524 .415 .672 r
F P
0 g
s
.8431 .77179 m .86374 .72496 L .77781 .74337 L p .85 .797 .799 r
F P
0 g
s
.77781 .74337 m .76235 .79012 L .8431 .77179 L p .85 .797 .799 r
F P
0 g
s
.54654 .90431 m .54911 .89052 L .46533 .90688 L p .638 .857 .983 r
F P
0 g
s
.33903 .81844 m .33308 .77026 L .25171 .78923 L p .539 .476 .734 r
F P
0 g
s
.26041 .50886 m .21726 .47018 L .15401 .50213 L p .459 .65 .942 r
F P
0 g
s
.11288 .45285 m .15401 .50213 L .21726 .47018 L p .41 .545 .88 r
F P
0 g
s
.87212 .62024 m .85925 .56998 L .77443 .58855 L p .826 .61 .61 r
F P
0 g
s
.75706 .54487 m .77443 .58855 L .85925 .56998 L p .801 .534 .535 r
F P
0 g
s
.55076 .55365 m .54836 .53189 L .46812 .5551 L p .483 .105 .246 r
F P
0 g
s
.46812 .5551 m .45725 .57671 L .55076 .55365 L p .483 .105 .246 r
F P
0 g
s
.44811 .60964 m .45725 .57671 L .37424 .60023 L p .508 .221 .412 r
F P
0 g
s
.37424 .60023 m .3554 .63264 L .44811 .60964 L p .508 .221 .412 r
F P
0 g
s
.71277 .85577 m .7401 .82847 L .64979 .84653 L p .754 .871 .932 r
F P
0 g
s
.64979 .84653 m .63436 .87341 L .71277 .85577 L p .754 .871 .932 r
F P
0 g
s
.16267 .15541 m .19033 .10196 L .12265 .14955 L p 0 0 .105 r
F P
0 g
s
.12265 .14955 m .09057 .20236 L .16267 .15541 L p 0 0 .105 r
F P
0 g
s
.62999 .50955 m .64594 .53152 L .73334 .51074 L p .625 .254 .32 r
F P
0 g
s
.64594 .53152 m .62999 .50955 L .54836 .53189 L p .567 .2 .306 r
F P
0 g
s
.86374 .72496 m .87367 .67318 L .78525 .69162 L p .849 .734 .735 r
F P
0 g
s
.78525 .69162 m .77781 .74337 L .86374 .72496 L p .849 .734 .735 r
F P
0 g
s
.87367 .67318 m .87212 .62024 L .78409 .63869 L p .841 .674 .675 r
F P
0 g
s
.77443 .58855 m .78409 .63869 L .87212 .62024 L p .826 .61 .61 r
F P
0 g
s
.78409 .63869 m .78525 .69162 L .87367 .67318 L p .841 .674 .675 r
F P
0 g
s
.54836 .53189 m .55076 .55365 L .64594 .53152 L p .567 .2 .306 r
F P
0 g
s
.75706 .54487 m .73334 .51074 L .64594 .53152 L p .706 .411 .475 r
F P
0 g
s
.46533 .90688 m .45479 .88117 L .36906 .89727 L p .63 .715 .908 r
F P
0 g
s
.35134 .86182 m .36906 .89727 L .45479 .88117 L p .629 .641 .843 r
F P
0 g
s
.21726 .47018 m .18192 .4177 L .11288 .45285 L p .41 .545 .88 r
F P
0 g
s
.08433 .39375 m .11288 .45285 L .18192 .4177 L p .364 .438 .804 r
F P
0 g
s
.45479 .88117 m .46533 .90688 L .54911 .89052 L p .678 .754 .909 r
F P
0 g
s
.14724 .21868 m .16267 .15541 L .09057 .20236 L p .144 0 .402 r
F P
0 g
s
.09057 .20236 m .07274 .2637 L .14724 .21868 L p .144 0 .402 r
F P
0 g
s
.44142 .65169 m .44811 .60964 L .3554 .63264 L p .565 .35 .547 r
F P
0 g
s
.3554 .63264 m .34162 .67381 L .44142 .65169 L p .565 .35 .547 r
F P
0 g
s
.63436 .87341 m .64979 .84653 L .55142 .86419 L p .72 .804 .917 r
F P
0 g
s
.55142 .86419 m .54911 .89052 L .63436 .87341 L p .72 .804 .917 r
F P
0 g
s
.7401 .82847 m .76235 .79012 L .66238 .80861 L p .782 .786 .856 r
F P
0 g
s
.66238 .80861 m .64979 .84653 L .7401 .82847 L p .782 .786 .856 r
F P
0 g
s
.18192 .4177 m .15728 .35508 L .08433 .39375 L p .364 .438 .804 r
F P
0 g
s
.0706 .32911 m .08433 .39375 L .15728 .35508 L p .312 .322 .71 r
F P
0 g
s
.14538 .28699 m .14724 .21868 L .07274 .2637 L p .245 .182 .585 r
F P
0 g
s
.07274 .2637 m .0706 .32911 L .14538 .28699 L p .245 .182 .585 r
F P
0 g
s
.55278 .58701 m .55076 .55365 L .45725 .57671 L p .593 .315 .464 r
F P
0 g
s
.45725 .57671 m .44811 .60964 L .55278 .58701 L p .593 .315 .464 r
F P
0 g
s
.54911 .89052 m .55142 .86419 L .45479 .88117 L p .678 .754 .909 r
F P
0 g
s
.15728 .35508 m .14538 .28699 L .0706 .32911 L p .312 .322 .71 r
F P
0 g
s
.45479 .88117 m .44618 .84476 L .35134 .86182 L p .629 .641 .843 r
F P
0 g
s
.33903 .81844 m .35134 .86182 L .44618 .84476 L p .624 .583 .788 r
F P
0 g
s
.77443 .58855 m .75706 .54487 L .65935 .56527 L p .743 .504 .568 r
F P
0 g
s
.64594 .53152 m .65935 .56527 L .75706 .54487 L p .706 .411 .475 r
F P
0 g
s
.65935 .56527 m .64594 .53152 L .55076 .55365 L p .654 .373 .482 r
F P
0 g
s
.43771 .69988 m .44142 .65169 L .34162 .67381 L p .592 .427 .627 r
F P
0 g
s
.34162 .67381 m .33397 .72083 L .43771 .69988 L p .592 .427 .627 r
F P
0 g
s
.55076 .55365 m .55278 .58701 L .65935 .56527 L p .654 .373 .482 r
F P
0 g
s
.44618 .84476 m .44019 .80017 L .33903 .81844 L p .624 .583 .788 r
F P
0 g
s
.33308 .77026 m .33903 .81844 L .44019 .80017 L p .617 .533 .737 r
F P
0 g
s
.44618 .84476 m .45479 .88117 L .55142 .86419 L p .682 .68 .841 r
F P
0 g
s
.76235 .79012 m .77781 .74337 L .67113 .76228 L p .785 .722 .793 r
F P
0 g
s
.67113 .76228 m .66238 .80861 L .76235 .79012 L p .785 .722 .793 r
F P
0 g
s
.43729 .75065 m .43771 .69988 L .33397 .72083 L p .607 .484 .686 r
F P
0 g
s
.33397 .72083 m .33308 .77026 L .43729 .75065 L p .607 .484 .686 r
F P
0 g
s
.44019 .80017 m .43729 .75065 L .33308 .77026 L p .617 .533 .737 r
F P
0 g
s
.78409 .63869 m .77443 .58855 L .6692 .60858 L p .763 .568 .634 r
F P
0 g
s
.65935 .56527 m .6692 .60858 L .77443 .58855 L p .743 .504 .568 r
F P
0 g
s
.64979 .84653 m .66238 .80861 L .5533 .82694 L p .731 .725 .845 r
F P
0 g
s
.5533 .82694 m .55142 .86419 L .64979 .84653 L p .731 .725 .845 r
F P
0 g
s
.55427 .62977 m .55278 .58701 L .44811 .60964 L p .635 .418 .571 r
F P
0 g
s
.44811 .60964 m .44142 .65169 L .55427 .62977 L p .635 .418 .571 r
F P
0 g
s
.55142 .86419 m .5533 .82694 L .44618 .84476 L p .682 .68 .841 r
F P
0 g
s
.77781 .74337 m .78525 .69162 L .67534 .71091 L p .782 .67 .74 r
F P
0 g
s
.67534 .71091 m .67113 .76228 L .77781 .74337 L p .782 .67 .74 r
F P
0 g
s
.78525 .69162 m .78409 .63869 L .67468 .65835 L p .775 .621 .689 r
F P
0 g
s
.6692 .60858 m .67468 .65835 L .78409 .63869 L p .763 .568 .634 r
F P
0 g
s
.6692 .60858 m .65935 .56527 L .55278 .58701 L p .69 .466 .577 r
F P
0 g
s
.67468 .65835 m .67534 .71091 L .78525 .69162 L p .775 .621 .689 r
F P
0 g
s
.44019 .80017 m .44618 .84476 L .5533 .82694 L p .68 .625 .787 r
F P
0 g
s
.55278 .58701 m .55427 .62977 L .6692 .60858 L p .69 .466 .577 r
F P
0 g
s
.55511 .67889 m .55427 .62977 L .44142 .65169 L p .655 .483 .639 r
F P
0 g
s
.44142 .65169 m .43771 .69988 L .55511 .67889 L p .655 .483 .639 r
F P
0 g
s
.66238 .80861 m .67113 .76228 L .5546 .78134 L p .731 .668 .788 r
F P
0 g
s
.5546 .78134 m .5533 .82694 L .66238 .80861 L p .731 .668 .788 r
F P
0 g
s
.5533 .82694 m .5546 .78134 L .44019 .80017 L p .68 .625 .787 r
F P
0 g
s
.43729 .75065 m .44019 .80017 L .5546 .78134 L p .675 .578 .739 r
F P
0 g
s
.67468 .65835 m .6692 .60858 L .55427 .62977 L p .709 .528 .642 r
F P
0 g
s
.55522 .73072 m .55511 .67889 L .43771 .69988 L p .667 .533 .692 r
F P
0 g
s
.43771 .69988 m .43729 .75065 L .55522 .73072 L p .667 .533 .692 r
F P
0 g
s
.5546 .78134 m .55522 .73072 L .43729 .75065 L p .675 .578 .739 r
F P
0 g
s
.55427 .62977 m .55511 .67889 L .67468 .65835 L p .709 .528 .642 r
F P
0 g
s
.67113 .76228 m .67534 .71091 L .55522 .73072 L p .727 .621 .74 r
F P
0 g
s
.55522 .73072 m .5546 .78134 L .67113 .76228 L p .727 .621 .74 r
F P
0 g
s
.67534 .71091 m .67468 .65835 L .55511 .67889 L p .72 .576 .693 r
F P
0 g
s
.55511 .67889 m .55522 .73072 L .67534 .71091 L p .72 .576 .693 r
F P
0 g
s
.25 Mabswid
.0061 0 m
0 1.34968 L
s
0 1.34968 m
1 1.34968 L
s
1 1.34968 m
.9939 0 L
s
.9939 0 m
.0061 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[34]:=",
  ImageSize->{213.312, 287.875},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40003E00017`P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoIP031`000=400@L000031`000<l70P011`00
00<70000c`L20047000010L01`3=1`0300L00047000010L01`3<1`0400L7004700001@L01`L00</7
00@01`L00@L000031`07008700<01`L0a`L01@071`L0004700000`L01`021`0300L703L720271`03
00L700870@011`0000<700L00`L00`071`0`1`H000RL]Ibe0;FG]@D0PPL00`071`021`400@L00003
1`0700@700<01`L0;@L2000C]INeUkFG]INe0;FG]INeUkFG]@0307h700<01`L00`L1004700000`L0
1`041`0300L702/70`005kFL]IbeW;FL]@2eUkFG]INeUkFG]@2`0080N`L00`071`041`400@L00003
1`0700D700<01`L0:0L30009UkFG]INeUkFG0080016eUkFG]INeUkFG]@2`U[2F001j1`0300L700@7
0@011`0000<700L01@L00`071`0V1`8001nK09beW;FL]IbeW02G]INeUkFG]INeUkD0/9N`Uk2G0080
M`L00`071`051`400@L000031`0700H700<01`L090L09@2KUP2G]INeUkFG]IL0UkFG]INeUkFG]INe
0;2F/9J`U[2F/000M@L00`071`061`400@L000031`0700L700<01`L08PL09P2eV`2L]IbeW;FL]I`0
UkFG]INeUkFG]INe0;2G/9N`Uk2G/9N`0P1c1`0300L700H70@011`0000<700L01`L00`071`0P1`80
02VK002G]INeUkFG]IL0UkFG]INeUkFG]INe0;2F/9J`U[2F/9J`U[2F001a1`0300L700L70@011`00
00<700L020L00`071`0N1`8000Je0;FL]I`<000I]INeUkFG]@2`Uk2G/9N`Uk2G/9N`U`2F001`1`03
00L700L70@011`0000<700L02@L00`071`0L1`8000>K0000100039NeUkFG]@2`Uk2G/0P001:F/9J`
U[2F/9J`U[2F/02`T@1^1`0300L700P70@011`0000<700L02@L00`071`0J1`H02[D03`2eUkFG]INe
UkFG]IL0U`02000A/9N`Uk2G/9N`Uk00/9J`UP00K0L00`071`091`400@L000031`0700X700<01`L0
60L4000N]02G]INeUkFG]INeU`2G/9N`Uk2G/9N`U`2FUiJG0P003iJ`U[2F/9J`0;2A/96`T@0206X7
00<01`L02@L1004700000`L01`0:1`0300L701L70`03]@030;Fe00Re01@0UkFG]INeUkFG]INe0;2G
/9N`U`<000nG/9N`0;2F/9J`U[2F/000J0L00`071`0:1`400@L000031`0700/700<01`L050L4000V
]KBe0;FG]INeUkFG]INe0;2G/9N`Uk2G/9N`U`2FUiJGUYNFUiH2000>/9H0TK2A/96`TK2A/01V1`03
00L700/70@011`0000<700L030L00`071`0B1`D00kD00`2e]@08]@0J0;FG]INeUkFG]INeU`2G/9N`
Uk2G/9N`Uk02000=U[2F/9J`U[2F/9J`001U1`0300L700/70@011`0000<700L030L00`071`0A1`D0
02Zd]K@0UkFG]INeUkFG]IL0Uk2G/9N`Uk2G/9N`09NFUiJGUYNFUiJGUYL2000=TK2A/96`TK2A/96`
001S1`0300L700`70@011`0000<700L03@L00`071`0?1`@000<H0;D00[D00`2e]@06]@T001NG]INe
UkD0/9N`Uk2G/9N`Uk2G/02`TP02000</9J`U[2F/9J`0;00H@L00`071`0=1`400@L000031`0700h7
00<01`L03@L60009]KBe0;FG]INe00@000>G]@001YL8000MUiJGUYNFUiJGUYNF096GTIL0/96`TK2A
/940T@00H@L00`071`0=1`400@L000031`0700h700<01`L030L40003600000>e00<0]@000`005KF`
]K2e/02G]INeUkFG]INeUkD0U`03000K/9N`Uk2G/9L0T[2B/9:`0;2F/9J`UP2A/94005l700<01`L0
3PL1004700000`L01`0?1`0300L700X71`000kBe0003000;UkFG]INeUkFG]@0039L01@2AUi6G00<0
01RGUYNFU`2GTINAUi6G0;2A/96`0;2A/01N1`0300L700l70@011`0000<700L03`L00`071`091`80
00DH01P06003000L]@2e/;F`]K2e/;F`]K00UkFG]INeUkFG]INe00NG0`002k00/9:`T[2B/9:`0080
00VF/02`TK2A/000G@L00`071`0?1`400@L000031`0701074`004K2e/02G]INeUkFG]INeUkD000bG
00/0TINAUi6GTINAU`02000CTINAUi6GTINAUi40T@2A/96`T@1N01870@011`0000<700L040L00`07
1`061`8000DH01P0600200Be01/0]K2e/;F`]K2e/;F`09NeUkFG]INeUkFG]@002iL3000C/9:`T[2B
/9:`TP2A/96`TK2A001K1`0300L701070@011`0000<700L040L00`071`051`T001>`]K2e/02G]INe
UkFG]INeUkD000bG01T0TINAUi6GTINAUi40N980Ui6GTINAUi6G008000NA/96`TK0005/700<01`L0
40L1004700000`L01`0@1`0300L700@70P0021P0600H01P01[D06P2`]K2e/;F`]K2e/02G]INeUkFG
]INeUkD02iL05`2GTYL0/9:`T[2B/02`T@2A/96`TK0005X700<01`L040L1004700000`L01`0@1`03
00L700<72P004KF`]K2e/;D0]INeUkFG]INe014000nGTINAUi6GTINA07RBN9800P00496GTINA096B
0;2A/96`T@1J1`0300L701070@011`0000<700L040L00`071`031`0601P0600H1007]@060;2e/;F`
10004;F`]@2`Uk2G/9N`Uk2G/9L500JG01P0Ui:GTYNB09:`T[00/96`0;2A/96`T@1I1`0300L70107
0@011`0000<700L040L00`071`021`P000`HK@2e/;F`]K2e0;D40008UkFG]INeU`0<U`0509MiUgT0
10006YNA07RBN99hTWP0TINA096BTI80/96`TK00F@L00`071`0@1`400@L000031`07010700X01`L7
01P0600H0P000aQ]0008]@80026e/;F`]K2e/;F`]K00Uk2G/9N`Uk2G/9N`09NBUi:GTYL00`0029NB
Ui:GTYNB0`003I6`TK2A/02`TK00/000F0L00`071`0@1`400@L000031`07010700@01`L71P002fdH
KAP0/;F`]K2e008000n_09NeUkFG]INeUkFG]@0039L02P1iUgVGNIMiUgT3000FN99hTWRBN980TY6B
TI:ATP2`T@2A05P700<01`L040L1004700000`L01`0@1`0300L7008000XH0000KAQ]66d01;D2000b
]K2e0;F`]K2e/;F`]K2e/;D0/9N`Uk2G/9N`Uk2G09:GTYNBUi:GTP1i09:GTYNBUi82000=/96`TK2A
096`0;2A001G1`0300L701070@011`0000<700L040L00`070004000:66dHKAQ]602`]@8001F_]Jne
[kD0]INeUkFG]INeUkFG]@002iL03@2GNIMiUgVGNIL0NGP00P003Y9hTWRB0980TY6BTI:A0P001:ZA
ZP1G1`0300L701070@011`0000<700L040L0100701P2000966dHKAQ]602e008001fe/;F`]K2e/;D0
]K2e/;F`]K2e/;F`09N`Uk2G/008000<TYNBUi:GTYL0TWVB0P004Y:GTP2ATP2`TK2A/940TK2A05L7
00<01`L040L1004700000`L01`0@1`@000U]66dHKAQ]66d00P006:ne[kF_]Jne[kF_]@2eUkFG]INe
UkFG]@L01iL5000MUgVGN@1hNGQiN01hTP2BN980TY6BTI80ZY6ZT@00EPL00`071`0@1`400@L00003
1`0701070`0086dHKAQ]66dHKAP0/;F`]K2e/;F`]K2e/02`]K2e/;F`10000kFG000<U`0407iiOP@0
00NB07VBNI9i00<000VATY6B0;2A/9400P001;2A/01F1`0300L701070@011`0000<700L040L07`0H
KAQ]66dHKAQ]66d0]Jne[kF_]Jne[kF_]Jl0UkD010001iN`Uk2G/9L00P0;U`0307Ui00Ai0P004GUh
NGQiN7T0TWRBN980TY6B008000BAZY40EPL00`071`0@1`400@L000031`0700l70P0076dHKAQ]66dH
KAQ]00L0]K2e/;F`]K2e/;F`]K02000>]INeUkFG]INeUkFG]@0;U`0807iiOWUnNGh20005NI9iTWT0
0P003i:ATY6B0;00TY40TK2A001E1`0300L701070@011`0000<700L03PL2000KKAQ]66dHKAQ]66dH
0:l70;F_]Jne[kF_]Jne0080016_002`Uk2G/9N`Uk2G/9N`000;U`0307Ui00Ai00l0D01hNGQi07P0
TWRBN9800P02T@050:ZAZP00E@L00`071`0@1`400@L000031`0700d70`006AQ]66dHKAQ]66dHK@07
]02`]K2e/;F`]K000P004kF`]K2e0;FG]INeUkFG]INeUkD00P0:U`0907iiOWUnN@1@008001BBN@1c
N@2BTI:ATP2BT@2A/96`05@700<01`L040L1004700000`L01`0=1`<000Y]66dHKAQ]66dH0`005:l7
[`2_]Jne[kF_0:ne[kF_]Jne0P003YN`Uk2G/9N`Uk2G/9L01IL700Ei00l0D5H0NGP0D7P0TWRBN000
0i401@2AZY4005@700<01`L040L1004700000`L01`0<1`<000U]66dHKAQ]66d00P00379c00Nd1k@0
/;F`]@8000^`]K2e/;F`]K2e/002000;]INeUkFG]INeUkD01P001WjGOYMnU`@000IiOWT0D5H30007
NG<0TI:ATP020007TI80/96`001D1`0300L701070@011`0000<700L02`L4000E66dHKAQ]66d0LVeb
K@2_1jl7[`2_008001Ne[kF_]Jne[kF_]Jne[kD0/9N`Uk2G/003000@/9L009MiUgVGNIMiUgT0D0@0
00aFD5I@051hD01hTWP200>A00@0ZP2AE0L00`071`0@1`400@L000031`0700X71@001VdHKAQ]6080
01mcLW=bL`07]0Nd1k@0/;F`]K2e/;F`]K2e/;F`]K2e008000>G]@000P001kFG]INeUkD00P002GjG
OYMnUgjGOP02000HEUD005I@EU00LgUc0940N980TY40T@2AE0L00`071`0@1`400@L000031`0700X7
1@005AQ]66dH06ebKG9]LVd0[`N_1jl0[`02000B]Jne[kF_]Jne[kF_]Jne[kF_0P003;2G/9N`Uk2G
/9N`U`8000fGNIMiUgVGNIL0EU1F008000M@EP00D7Q@008000>AL`000i40102AT@1C1`0300L70107
0@011`0000<700L02@L5000466dHK@80019bLg9cLW=bL`07]0Nd0:l7[`L2000B/;F`]K2e/;F`]K2e
/;F`0;2e0P0039NeUkFG]INeUkFG]@8000QnUgjGOYMnU`8000IEEP00D5H2000>LgUc07RBN02ATP2Z
T@1C1`0300L701070@011`0000<700L02@L5000^KAP0079]LVebKG9]LVd0[`N_0:ne[kF_]Jl0[kF_
]Jne[kF_]Jne[`2F/9J`UP8000b`Uk2G/9N`Uk2G/9L20005UgVGNIL01@002EH0001FB`1@N0020003
Li4000:A00@0TI40D`L00`071`0@1`400@L000031`0700T71@006aP0LW=bLg9cLW=bLg80]0Nd0:l7
[`N_1jl7[`020009]K2e/;F`]K2e008000R`]K2e/;F`]@8000bG]INeUkFG]INeUkD50004OWUnJ`<0
00=EQ0000P003GUc002BN980TP2ZT@00D`L00`071`0@1`400@L000031`0700P71`00<FebKG9]LVeb
KG9]LP07[`L0]Jne[kF_]Jne[kD0]Jne[kF_]@2`U[2F/9J`U[2F/9H00P002K2G/9N`Uk2G/00200:G
0P001UInEWiFOP<00X@20008N01c095c09420003T@07058700<01`L040L1004700000`L01`081`D0
01mcLW=bLg9cLW=bLg9cLP2d1`07[`N_1jl7[`N_1jl7008001B`]K2e0;F`]K2e/;F`]K2e/;F`]@80
00fG]INeUkD0/9N`Uk2G008000QnNGiiOWV400:40P001WT007RBN0<000>A00L0DPL00`071`0@1`40
0@L000031`0700P71@008FebKG9]LVebKG9]LVeb00L0]Jne[kF_]Jne[kF_]Jne[`02000D]@2`U[2F
/9J`U[2F/9J`U[2F/9H20004/9L000VG0`002GiFOUInJh@0C0020004D01cT@<00Y400`071`1@1`03
00L701070@011`0000<700L020L5000>Lg9cLW=bLg9cLW=bLg82000Y1jl7[`N_1jl7[`N_1jl7[`N_
1`2e/;F`]K2e/;F`]K2e/;F`]K2e/;D00P003YN`Uk2G/9N`Uk2G/9N`0P003WUnNGiiQ01@07<0TWRB
0P000i6Z001B1`0300L701070@011`0000<700L01`L4000W601]LVebKG9]LVebKG9]LP2e[kF_]Jne
[kF_]Jne[kF_]Jne0;2_008001F`U[2F/9J`U[2F/9J`U[2F/9H0U[000`0<U`8000=nEWh01@002G<0
TG<0L`2A001B1`0300L701070@011`0000<700L01`L3000A64/0Lg9cLW=bLg9cLW=bLg800`005Zl7
[`N_1jl7[`N_1jl7[`N_0:ne[kD3000?]K2e/;F`]K2e/;F`]K0000J`0`002k2G/9N`Uk2G/9N`00<0
00=WEFL00P002E00N980T@2Z001B1`0300L701070@011`0000<700L01`L3008H0180KG9]LVebKG9]
LVebKG80aZl2000G]Jne[kF_]Jne[kF_]Jl0[k2_/:n`[k000P005YJ`U[2F/9J`U[2F09J`U[2F/9J`
U[0200VG00<0OYL00P001EFR001@00<000Ac0940DPL00`071`0@1`400@L000031`0700L70P0044/H
B`1cLW=bLg9cLW=bLg830004[`N_1`8001N_1jl7[`N_1jl7[`2_]Jne[kF_]Jne[`02000:]K2e/;F`
]K2e00b`0`001iN`Uk2G/0001IL30004L`00L`8000>Z00L0D@L00`071`0@1`400@L000031`0700H7
0P04600;06ebKG9]LVebKG800P02L`070<J_aZo6[`02000G]Jne[kF_]Jne0;2_/:n`[k2_/:n`[k00
0`005[2F/9J`0;2F/9J`U[2F/9J`U[2F/9H300:G00T0UgjGOYMnUgh00P001G<0L`2A0080DPL00`07
1`0@1`400@L000031`0700H70P003AQ;601bLg9cLW=bLg800P003I5cTG>A00N_1jl7[`L00P006:l7
[`N_1`2e[kF_]Jne[kF_]Jne[kF_]@8000>`]@005;0200VG00NR0500L`2A00@0D0L00`071`0@1`40
0@L000031`0700H700<061P00QP0201bKG9]LVeb0P07L`0:0:o6[lJ_aZo6[`8001Re[kD0/:n`[k2_
/:n`[k2_/:n`[k2_/:l3000GU[2F/9J`U[2F/9J`U[2F/9J`U[00/9H010001IMnUgjG00D000FA0000
XP0204h700<01`L040L1004700000`L01`061`0R04/HBaP0LW=bLg9c095cTG>ALi5cT@07[`N_1jl7
[`N_1`8001V_0:ne[kF_]Jne[kF_]Jne[kF_]Jne0;2_00@04;001`2G/9N`Uk000`02U`03_P000080
00>A00000P001::M001<1`0300L701070@011`0000<700L01PL00`0H60026005079]LVd00P0:L`0>
0:o6[lJ_aZo6[lJ_aZl2000K/:n`[k2_/:n`[k2_/:n`[k2_/:l0[k2_/:n`00@001JF/9J`U[2F/9J`
UP2F/9J`U[2F/9J`1`0019400:<20003XXBR00<0B@L00`071`0@1`400@L000031`0700H700P0BaQ;
601bL`8001acTG>ALi5cTG>ALi401jl7[`N_1jl7[`N_1`370`007;F_]Jne[kF_]Jne[kF_]Jl0[k2_
/:n`[k2_/:l400J`0100Uk2G/9N`Uk2G/9N`Uk2G0P001g<0002o0:T00P0019fRWJ8204L700<01`L0
40L1004700000`L01`061`0301PH008H0`0?L`0=0<J_aZo6[lJ_aZo60005[`<001f`[k2_/:n`[k2_
/:n`0;2_/:n`[k2_/:n`[k2_/004000BU[00/9J`U[2F/9J`U[2F/9J`1P001J<0XjVS008000BRQ::4
0`141`0300L701070@011`0000<700L01PL0;01;64/H07>ALi5cTG>ALi5cTG>ALi5c0:l7[`N_1jl7
[`N_0:o7[lN_ajo70`007[F_]Jne[kF_]@2`[k2_/:n`[k2_/:n`[k2_/:n`[`@000j`Uk2G/9N`Uk2G
/9N`U`H000No0:VSZJ>Y008000FMXP2SXP02048700<01`L040L1004700000`L01`061`0401PH6080
4G<03036[lJ_aZo6[lJ_00b_0`007k2_/:n`0;2_/:n`[k2_/:n`[k2_/:n`[k2_/:l0TK0010002iJ`
U[2F/9J`U[2F00L000NSZJ>YXjVS008000FSQ:>4X`0403h700<01`L040L1004700000`L01`051`05
01Q;64/00P00;7>ALi5cTG>ALi5cTG>ALi5c0:l7[`N_1jl7[`L0ajo7[lN_ajo7[lN_ajo70`008;F_
0:n`[k2_/:n`[k2_/:n`[k2_/:n`[k00/:n`[k2_10001k2G/9N`Uk001@002l00ZJ>YXjVSZ@2Y0080
00FRXj:SXP0503/700<01`L040L1004700000`L01`051`0501PH60000T/00`1cL`0?L`0:0:o6[lJ_
aZo601>_0`008K2_/:n`[k2_/:n`[k2_/:n`[k2_/02`TK2A/96`TK2A/0040004U[2F/0@001BX0000
XjVSZJ<0XjVS08BSQ:>4X`8000=[00000P0g1`0300L701070@011`0000<700L01@L06@0HB`1<Bd`0
TG>ALi5cTG>ALi5cTG>ALi400P007Zl7[`N_1jl0[lN_ajo7[lN_ajo7[lN_ajo7[lL0/0@0026_/:n`
[k2_/:n`[k2_/:n`[`2_/:n`[k2_/:n`[k2_/:l020003/00`:T0XjVS0:BYY:VT0P001Z>RXj:SXPP0
=0L00`071`0@1`400@L000031`0700D700<0600014/00`1cL`0>L`8000R_0036[lJ_01B_00L0[jZ_
ZZnZ00@0026`[k2_/:n`[k2_/:n`0;2A/96`TK2A/96`TK2A/96`TK0010000jP0Z@02Z@8000VY0:VS
ZJ>YXjT00P001hBS0024Xh@00P03Q0<0<@L00`071`0@1`400@L000031`0700D701H0601;C4]<095c
TG>ALi5cTG>ALi5c0P001I80b/K:008002:_1`37[lN_ajo7[lN_ajo7[lN_ajo7[`2_/:n`[k2_/:n`
0`007[2_/:n`[k2_/:l0[k2_/:n`[k2_/:n`[k2_/:n`[`<001>Y0:T0ZL2Y`02YY:VTZJBYY:VT0080
00NSRZ>:XhZS008018@302h700<01`L040L1004700000`L01`051`801T/00`1cL`09L`8000UcTG>A
0<J_aZl00`0D[`0>0:nZ[jZ_ZZnZ[jZ_ZZl4000K[k2_/:l0TK2A/96`TK2A/96`TK2A/96`TK2A00@0
00?00:T00jT2000:ZJ>YXjVSZJ<0RP8000RSQ:>4XhBSQ0801H@302/700<01`L040L1004700000`L0
1`051`0B04]<Bda;C4/0Li5cTG>ALi5c0P003Y9hTWRBN036b/K:aP2_0`008jo7[lN_ajo7[lN_ajo7
[lN_0:n`[k2_/:n`[k2_/:n`[k2_00@001N`0;2_/:n`[k2_/:n`[k2_/:n`[k2_/003000NaP00ZL2Y
`02Z0:VTZJBYY02TZJBY0:>:XhZSRZ>:300X1`0300L701070@011`0000<700L01@L00`1;B`05B`03
07=c00Ec0P003g>ALi5cTG>AL`2_aZo60004[`@03:l05`2Z[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_00@0
01BA/96`TK2A/96`TK2A/96`TK2Ab@8000>Ya@001:T00`2TZ@02000:XjVS08ZTRZB:Y08000>4X`00
0`000hF:Q@0200J40`0V1`0300L701070@011`0000<700L01@L03P1;C4]<Bda;07>ALi5c0P006I9h
TWRBN99hTWRB0<[6bP36[lJ_aZo6[lH00`009<N_ajo7[lN_a`2`[k2_/:n`[k2_/:n`[k2_/:n`[k2_
/02`[`@000n`[k2_/:n`[k2_/:n`[k000`006<JY0030ZL00ZZVZZ@2T0:BYY:VTZJBYY08000VSRZ>:
XhZSRZ<00P06Q0<090L00`071`0@1`400@L000031`0700H700<0Bd/01D/00`1cL`02000BLi5cTG>A
Li5cTG>AL`2_aZl03:l300J_01l0ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jX0TJnA[i6_00D000Z`TK2A
/96`TK2A0`009JW50:X0Z@2YY:VTZJ@0RZB:Y8ZTRZB:08X0QHZ5RXF:QHZ5RXD00P06Q0@08@L00`07
1`0@1`400@L000031`0700H700P0C4]<Bda;C0<002>BN99hTWRBN99hTWRBN99h0<K:0<J_aZo6[lJ_
aZo6[lJ_aP04000V[`2_/:n`[k2_/:n`[k2_/:n`[k2_/:n`[`2_/:n`[k2_/:n`[k050005/:n`[k00
10009JT0Z@2Y0:VZZJZYZZT0Y:VTZJBYY02;ZH/0RZ>:XhZSRZ>:XhX03P0N1`0300L701070@011`00
00<700L01PL00`1;B`05B`0G07>ALi5cTG>ALi5cTG>ALi5cT@36[`004jl3000XZZnZ[jZ_ZZnZ[jZ_
ZZnZ[jZ_ZZnZ[`2_TJnA[i6_TJnA[i6_TJnA[`T000C50:ZY0P006jVTZJBY0:V;08ZTRZB:08ZTRZB:
08F:QHZ5RP040003Q8X000R41@0L1`0300L701070@011`0000<700L01PL01`1<Bda;C4/00P00;99h
TWRBN99hTWRBN99hTWRBN036bP36[lJ_aZo6[lJ_aZo6[lJ_aZo6[`2_1000:;2_/:n`[k2_/:n`[k2_
/:n`[`2_/:n`[k2_/:n`[k2_/:n`[k2_/:l5000NZ@2YaZT0ZJZYZZT0ZJZY0:BYY02;ZH^YRjV;08ZS
0`0028Z5RXF:QHX028@20003Q00000806@L00`071`0@1`400@L000031`0700L700<0Bd/00d/0601@
095cTG>ALi5cTG>ALi5cTG>AL`2_01B_00H0[jZ_ZZl4000TZZnZ[jZ_ZZnZ[jZ_ZZl0[i6_TJnA[i6_
TJnA[i6_TJnA[i6_1@001LD0ZZVZ008001FTZJ@0RjV;ZH/0RP2:Y8ZTRZB:Y8X00P0038Z4RXB:Q8Z4
RXB:00V400D0Q6^4J`0201L700<01`L040L1004700000`L01`071`0K04]<Bd`0Le00N99hTWRBN99h
TWRBN99hTWRB008001n_aZo6[lJ_aZo6[lJ_aZo6[lJ_aP36[lJ_aZo6[lJ_00@0022`[k2_/:n`[k2_
0:n`[k2_/:n`[k2_/:n`[k2_/:n`[`L000>YaZT00P004jZYZP2ZZJZYZZT0RjV;ZH^YRjT00P003HX0
QHZ5RXF:QHZ5RXD03P03Q0805@L00`071`0@1`400@L000031`0700L700D0Bd];0004D00D07>ALi5c
TG>ALi5cTG>ALi5cT@0D[`0?0:Z_ZZnZ[jZ_ZZnZ[jZ_00@001^Z[jZ_ZP2A[i6_TJnA[i6_TJnA[i6_
TJnA[i4020005ZZYZP2Z0:T0ZH^YRjV;08/0RZB:Y8X2000;RhF;08Z4RXB:Q8X0100036/0JhA[Q6^4
JhA[00:41`0C1`0300L701070@011`0000<700L01`L0701;C01cD7=@L`2BN99hTWRBN99hTWRBN99h
TWP2000VaZo6[lJ_aZo6[lJ_aZo6[lJ_aP36[lJ_aZo6[lJ_aZo6[lJ_aZl4000G/02`[k2_/:n`[k2_
/:n`[k2_/:n`[k000`009<H00:T0ZLJY0:T0Z@2YZZVZZJX0ZP2YRjV;08Z;RX^:R`2:Q@@000J5Q8F4
Q@0:Q00308B400F40`0A1`0300L701070@011`0000<700L020L00`1;0006D00@095cTG>ALi5cTG>A
Li5cT@8000=h00000P0@[`0G0:Z_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl010005i6_TJnA[i6_TJnA[i6_
TJnA[lT0[`2^008001F_0:ZYZP2ZZ@2;ZH^YRjT0ZH^Y0:@00P0028F;QH^5RhF;0P003f^5JhE[QF^5
J`1[Q6^4J`0700R40`0?1`0300L701070@011`0000<700L020L2000ED7=@Le1cD01hTWRBN99hTWRB
N99h008000NGN9L0aZo600@002N_aZo6[lJ_aZo6[lH0aZo6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lH0[jX0
1@003Jn`[k2_/:n`[k2_/:l00P000lX0aP02000MZ@00aP2ZZJZY0:VZZJX0ZX^ZRjX0RhZ;RX^:RhX0
0`002HB5Q8F4QHB5Q0070007JhA[Q02MQ00400B400@0Q9d03PL00`071`0@1`400@L000031`0700P7
00<0D5001e002`1cTG>ALi5cTG>A008000ehTWRBN980aZW6ZLJY00<02Jl0802Z[jZ_ZZnZ[jZ_ZZnZ
[jZ_ZZnZ[jZ_ZP2Z[jZ_ZZnZ10002JnA[i6_TJnA[`040003[P0000<0012Y0:VZZJX0ZH^Y0:V;ZH^Y
0P001XF;QH^5R`8000Q/Q@25JhE[Q@D000^4JhA[Q6^4JhA[0006Q0@000B408B40P0<1`0300L70107
0@011`0000<700L020L04`1cD7=@Le1cD7<0TWRBN99hTWP00P004YMhUgRGN9MhU`36[lJ_aZo6[`<0
02Z_aZo6[`2_aZo6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lH0[jZ_ZZnZ[jZ_ZZl50003/:n`008000C60036
10004JX00:ZY0:VZ0:ZY08^ZRjX000:;00@0RX^:0P001h]/Rfb;08D00`04J`0F06^4JhA[Q6^4J`24
WHBMQ9f4WHBMQ08000FMQ9f4000;1`0300L701070@011`0000<700L02@L00`1@D007D006095cTG>A
0P005gRBN99hTWRBN99h0:W6ZLJYaZW6ZLJY00@002^_0:nZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZP2Z
[jZ_ZZnZ[jZ_ZZnZ[jZ_00D000G60000aP03000@ZP00ZJX0ZY00R`2;ZH^Y00>;00/0R`00K8E/QFb5
K00200I[00D0JhA[Q00800V40P000j800003Q0802@L00`071`0@1`400@L000031`0700T700h0D7=@
Le1cD7=@07RBN08001fGN9MhUgRGN9MhUgRGN02_aZo6[lJ_aZo6[lJ_aP03000ZaZo6[lJ_aZo6[lJ_
aZo6[lJ_aZo6[lH0[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ10001<H00:l3000@ZZl0ZZT0ZJZY08^ZRjZ;00B;
00P0Rfb;K8]/R`801V/7009[0P0019f4WH@50004Q9f400FR0P0019f4W@081`0300L701070@011`00
00<700L02@L200U@0`008WRBN99hTWRBN99hTWRBN99h0:W6ZLJYaZW6ZLJYaZW60:l4000VZZnZ[jZ_
ZZnZ[jZ_ZZnZ[jZ_0:nZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl20003[`2_008001?60000ZZT0ZJX0ZY2Z0:V;
ZH/000B;0P001Fb5K8E/008000A[06][1@06J`802h@30008Q::4XXBRQ:8200:400<01`L01@L00`07
1`0@1`400@L000031`0700X703<0Le1cD7=@Le1c09MhUgRGN9MhUgRGN9MhUgRGN9L0aZo6[lJ_aZo6
[lJ_aZl0Z/JZaZX010009lJ_aZo6[lJ_aZo6[lJ_0:Z_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[`070<X0aP02000B
[`00aZZ_ZP2Y0:VZ0:X0ZX/018/01`2;08]/Rf`00P000f`0000600E[0P0039f4WHBMQ9f4WHBMQ080
0Z<200RR10061`0300L701070@011`0000<700L02PL200M@02h0N@2BN99hTWRBN99hTWRBN99hTWRB
0<JYaZW6ZLJYaZW6ZLH0[jV_ZJnY[jV_10007JZ_ZZnZ[jZ_ZZl0[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ00@0
00F_0:l0aP020004ZJZYZP8000R@ZY00R`2;00B;00H0Rf`0K8D30003J`0000D00V/700Z40P001XVS
RJ>9X`8000V4XXBRQ::4XX@00P051`0300L701070@011`0000<700L02`L0>P1@Le1cD7<0NGQi09Mh
UgRGN9MhUgRGN9MhUgRGN9L0aZo6[lJ_aZo6[lJ_0:[6Z/JZaZ[6Z/JZaZX4000JaZo6[lJ_0:Z_ZZnZ
[jZ_ZZnZ[jZ_ZZnZ0<H30003[`2_008000o60000[jZ_0:ZY092ZT0001H/00`1]R`030003K00000d0
18@40004WHBMQ0802Z<200FR0P001:8000031`0300L701070@011`0000<700L02`L200A@03P0NE1i
D7T0TWRBN99hTWRBN99hTWRBN99h0:W6ZLJYaZW6ZLJYaP2_ZJnY[jV_ZJnY[jV_ZJnY[`@001>Z[`2_
ZZnZ[jZ_ZZnZ[jZ_ZZnZ00<000J_002_0:l2000?Z@000:VZT02@ZP2ZRjX000B;00@0K8]/0P000f/0
000<00Z410003XVSRJ>9XhVSRJ>9XhVS0P000h@00004XP0308@0008700<01`L040L1004700000`L0
1`0;1`8001ecD7=@07QiN7UhN@2GN9MhUgRGN9MhUgRGN9MhU`02000Q[lJ_aZo6[lJ_aZl0Z/JZaZ[6
Z/JZaZ[6Z/JZaZ[6Z/JZ00H000b_ZZnZ[jZ_ZZnZ[jX200081`070<J_0:l20003aP2_008000^Z[`2Z
0:Z@ZP2A0002R`0406f;K@8000=/00001@000f/0J`0300Z40`02X`803Z<2000<X[nR_j:oXP2400L0
4PL1004700000`L01`0<1`800U00601@NE1iD7U@07RBN99hTWRBN99hTWRBN08002VGN@2YaZW6ZLJY
aZW60:nY[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_0:nY[`060006ZZnZ[jZ_0`003<J_0:l0ZP2A0:T0Z@80
012ZT02@08^ZR`2;08/0K8]/3`08Q0@001>:XhZSRZ>:08VSRJ>9XhVSRJ>9008000>XX`001j840187
0@011`0000<700L030L2000HD01hNGQiN7UhNGP0N9MhUgRGN9MhUgRG0P05U`0/0:o6[lJ_aZo6[`2Z
aZ[6Z/JZaZ[6Z/JZaZ[6Z/JZaZ[6ZP2Z[jZ_ZZnZ[jX70003aP2_008000W60:l0[`360:l00P004JX0
T@2@ZY00Ri6;06f;K@0H00h01H@300bS0P06X`<000jX`:S0Z02R_j:oX[nR_`804PL1004700000`L0
1`0=1`8001A@NE1iD7U@NE1iD01hTWRBN99hTP<003NGNIMiUgVGN@2YaZW6ZLJYaP2_ZJnY[jV_ZJnY
[jV_ZJnY[jV_ZJnY[`2_ZJnY[jV_ZJnY[jV_00@000f_aP00[`2Z0940Z@2Y008000RZ0:X0ZX^Z00:;
0P001H]/R`1/00X000A[Q6^41@004J>:XhZSRZ>:XhZSRZ>:XhZS008000j9X`00XjRSZ:>XXjRS00RR
00<01`L040L1004700000`L01`0=1`<0015hNGQiN7UhNGQiN01hUgRGN00200bG0`00:Zo6[lJ_0:[6
Z/JZaZ[6Z/JZaZ[6Z/JZaZ[6Z/JZ0:Z_ZZnZ[jZ_ZZnZ[`8000Nd00070<N_008000N_0:l0aP2_0080
00ZZ0940T:Z@08^A0P0016f;06`80006JhA[Q6^40`000jB:000BX`<001K0Z<2X`:S0Z<2X`:P0X[nR
_j:oX[l04@L1004700000`L01`0>1`<000i@NE1iD7U@NE1iD01hTP80011iUgVGNIMiUgVGNIMiUgVG
1@009LH0[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJl0[jV_ZJnY[jV_ZJl00P001Zl00<J_00:_00D0[`2_
T@03000:Z@00ZP2ZR`2;00:;00@0R`2;1`001XA[Q6^4J`@001BSRZ>:XhX0RZ>:XhZSRZ>:XhZSRP@0
016YX`2SZ:>XXjRSZ:>XXjRS0006XP030;l001070@011`0000<700L03PL2000=D01hNGQiN7UhNGQi
N00201>G00@0aZW610007j[6Z/JZaZ[6Z/JZaZ[6Z/JZaZ[6ZP2Z[jZ_ZZnZ[jX00`003[@7002_0:o7
[`2_0:Z_0`003Zl00:X0TJX0ZP2AR`1b1@001f^4JhA[Q6/00`002hZTRZB:Y8ZTRZB:00801j<400NY
0180Z<2X`:S0Z<2X`:S0Z02R_j8200:o00<01`L03PL1004700000`L01`0?1`0?051<051iD7U@NE1i
D01h008001ZGNIMiUgVGNIMiUgVGNIMiUgVG0<JYaP2_Z@D001^Y[jV_ZJnY[jV_ZJnY[jV_0:nY[jV_
ZJnY0;@00P001PN_0:o600:_0P002:l0[`2ZT@2Y0P001ZX0ZX^A00:;00<0K@000`02J`8000>4J`00
0P004hZSRZ>:XhZSRZ>:XhZSRZ<0XhX010008jVSZJ>YXjVSZJ>YX`2SZ:>XXjRSZ:>XXjRS0:80YknW
_jL000l70@011`0000<700L03`L04P1@D01iN7UhNGQiN01iUgVG01:G0P002/JY0:W6ZLJYaZT4000C
aZ[6Z/JZaZ[6Z/JZ0:Z_ZZnZ[`020005]@001k@00P001@L0ajo7008000J_0:X0TJl20009ZP00ZY6Z
T02;00@01V/3000CRZB:Y8ZTRZB:Y8ZTRZB:Y8ZTRP03016Y00d0Z<2X`:S0Z<2X`:S000801[l00`07
1`0=1`400@L000031`0701070240C500NE1iD7U@07QiN7UhN@2GNIMiUgVGNIMiUgVGNIL00P02U`<0
00^_ZJnY[jV_ZJnY[`05000@[jV_ZJnY[`2_ZJnY[`2e]08000L7[`2_0:l000:_00T0[jX0TJl0ZY40
0P001jX00900Ri400`05J`@001:5RP2SRZ>:XhZSRZ>:XhZSRZ<4000MY:T0ZJ>YXjVSZJ>YXjVSZJ>Y
XjVS0:>XXjRSZ:<010002/2X0:NoYknW_`0>1`400@L000031`0701070180D51@07UhNGQi09MiUgVG
NIL200^G0P06U`8000o6ZLJYaZW6ZLJYaZW6ZLH01@001/JZ0:Z_ZP8000T7]@L01k@0[`L00P003:o7
[`2_/02_ZP2A[`8000BZ002Z0`05J`<001:TRZB:Y8ZT0:B:Y8ZTRZB:Y8X40007ZJBYY:VT000CZ@03
0:S000@000NYaJW5ZLD000No00<01`L02`L1004700000`L01`0@1`8001]@C500NE1i07UhNGQiN7Uh
NGP0NIMiUgVGNIL00P0:U`8001:Y[jV_ZJnY[jV_ZJnY[jV_ZJl60006];Fd]@070P000jl7[`0200:_
00X0ZZnZ096_0:ZA0P000jX00006J`<001::QHZ5RXF:QHZ508ZSRZ>:XhX3000MY:VTZJBYY:VTZJ@0
XjVSZJ>YXjVSZJ>YXjVSZJ<010004l2X`:S0Z<2X`:S00;nW_jNoY`003@L1004700000`L01`0@1`03
00L000=@0100N@2GNIMiUgVGNIMiUgT01IL200jG0P004<JYaZW6ZLJYaZW6ZLJYaZT50003]@Ne0080
01D700N_1jl70<N_a`2`[k00[jX0TJl00`0016/0Jf/4000AY8ZTRZB:Y8ZTRZB:Y8X0RZ@010004:BY
Y:VTZJBYY:VTZJBYY00<Z@@0016YaP35ZLFYaJW5ZLFYaJW50006_`0300L700X70@011`0000<700L0
40L05P071`1@C500NGQiN7UhNGQiN7UhNGP20003U`0001:G1P001jV_ZJnY[jT00`001YbeW02d]@80
00R_0007[`N_1`800Zl01P2_ZZl0[``00165RXF:QHZ5RXF:QHZ5RXF:Q@04000LZJBYY:VTZJBYY:VT
ZJBYY:VTZJ@0XjVSZJ>YX`@01jT05@30Z<2X`:S0Z<2X`:P0YknW_jNo000<1`400@L000031`070107
00D01`L70002D0<000iiUgVGNIMiUgVGNIMiU`804IL20006]INeUkFG1`05]@060;D000Ne0P0040N_
1jl70<L0ajl0[k2_0:X20003J`1[00@000n508ZTRZB:Y8ZTRZB:Y8X01@006JBY0:VTZJBYY:VTZJBY
Y:VTZJBYY:VTZ@000jT4000JZLJYaZW6ZLJYaZT0ZLFYaJW5ZLFYaJW5Z@06_`0300L700T70@011`00
00<700L040L00`071`021`05051<05000P003WUhNGQiN7UhNGQi09Mn0`0<U`8000^GW9NLUibGW9NL
U`020006]IbeW;FL0`003Zne[kD01jl7[`L0/:n`0P000jZ_ZP0800F500P0RXF:QHZ5RPH001b;ZH^Y
RjV;Z@2YY:VTZJBYY:VTZJBYY:VTZJBY1@0@Z@0B0:S0Z<2X`:S0Z<2X`:P0YknW0`0;1`400@L00003
1`07010700<01`L00`L04@1@07UFN@2GNIMiUgVGNIL000JG0`07U`8000^eUkFG]INeUkFG]@0200Ne
0P002PNe1kD700N_1jl20004ajo7[`H000m[0000RXF:QHZ5RXD0RZ@01@007ZVTZJBYY:VTZJBYY:VT
0:BYY:VTZJBYY:VTZJBYY0@002FZZ@2YaZW6ZLJYaZW6ZLJYaZW6ZLH0aJW5ZLFYaJW5ZLFYa@2o0080
00C4aL@02PL1004700000`L01`0@1`0300L700@70P00151FD5H2000AN7UhNGQiN01nUgjGOYMnUgh0
0`02U`8001:GW9NLUibGW9NLUi`0]IbeW;D40005[kF_]Jl00`002Zl7[`2_/:n`[k0700Z510007jV;
ZH^YRjV;ZH^YRjV;ZH^YRjT0ZJBYY:VTZJBYY:T01007Z@030:VY012Y00d0`:S0Z<2X`:S0Z<2X0080
00K4_lBoa00:1`400@L000031`07010700<01`L01@L20005EWUFNEH00P001IMiUgT000fG1`001iNe
UkFG]IL02002]@050;D7]@L00P002[0700N_0:o7[lL5000;RXF:QHZ5RXF:QHX010007h/0Y:VTZJBY
Y:VTZJBYY:VTZJBYY:VT0:BYY:VTZJ@01000::ZYZZVZZJZYZZVZ0<JYaZW6ZLJYaZW6ZLJYaZW6Z@2Y
aJW5ZLFYaJT30008a@34aLC5a00:1`400@L000031`07010700<01`L01PL0201FD5I@EU1F0P005gP0
OYMnUgjGOYMnUgjGOYL0W9NLUibG00T000^G]INeUkFG]@2e[`02000:]Jne[`2_0:n`[`D000=[08D0
1XD4000OY8ZTRZ@0ZH^YRjV;ZH^YRjV;ZH^YRjV;ZH^YR`2TZ@04012Y00<0ZJT04:T01P2X`:S0Z0<0
00^YaJW5Z@2oa;o4000:1`400@L000031`07010700<01`L01`L20007EWUFNEIiEP0200fG00h0]INe
UkFG]INeUkFG00Ze00X01`00/0N`1k070P000jo700030008JhE[08F:QHX4000NRjV;ZH^YRjV;0:BY
Y:VTZJBYY:VTZJBYY:VTZJBY1@00:ZZYZZVZZJZYZZVZZJZYZZVZZJZY0:W6ZLJYaZW6ZLJYaZW6ZLJY
aP35Z@<000o5a/G6aLK5aP35a<G4a@002@L1004700000`L01`0@1`0300L700P70P00251FD5I@EP1n
0`008YMnUgjGOYMnUgh0UibGW9NLUibGW9NL0;FG]INeUkFG]IL20008]Jne[kF_]Jl500I[1@007:B:
Y8ZTRZB:Y8ZTRZB:08^YRjV;ZH^YRjV;ZH/60003ZH/001JY00<0ZJT03jT4000BZLFYaJW5ZLFYaJT0
_lBoa;l02@L1004700000`L01`0@1`0300L700X70P002GUFNEH0OYMnU`0300JG00h0UkFG]INeUkFG
]INe00Re0`002;07/0N`1k0710001f^5JhE[QF/00`006h^YRjV;ZH^YRjV;ZH^YRjV;Z@2YY:VTZJBY
Y005000]ZJZYZZVZZJZY0:VZZJZYZZVZZJZYZZVZZJZYZZVZZJX0aZW6ZLJYaZW6ZLJY00@001K6Z@35
a/G6aLK5a/G6aLH0aLC5a<D02@L1004700000`L01`0@1`0300L700/70`002U1F07iiOWUnNGh3000E
Ugh0UibGW9NLUibGW9NL0;FG]INe00@000VG]@2e[kF_]Jl00P07J`@001R508ZTRZB:Y8ZTRZB:Y8ZT
RZB:Y8X0RjT6000?RjV;ZH^YRjV;ZH^YRjT001JY00<0ZJT01JT5000KZLJYaZW6Z@2YaJW5ZLFYaJW5
ZLD0a;o4_l@000P70@011`0000<700L040L00`071`0<1`<000aFN@2GOYMnUgjGOYL80009UkFG]INe
U`2e00@000^e/;F`]K00/0N`1`030004QF^5J`@018/04`2;ZH^YRjV;ZH^YRjV;ZH^YRjT01@00;jZY
ZZVZZJZYZZVZZJZYZZVZZJZY0:VZZJZYZZVZZJZYZZVZZJZYZZVZZJZY0:W600@0022YaZW6ZLJYaZW6
ZLH0a/G6aLK5a/G6aLK50<C5a<G400P70@011`0000<700L040L00`071`0>1`@000]iOWUnNGiiOWUn
0006U`T000bG]INeUkFG]INe0;D5009[0`005hF;QH^5RhF;0:B:Y8ZTRZB:Y8ZTRZB:00@001^YR`2;
ZH^YRjV;ZH^YRjV;ZH^YRjV;ZH^YR`005:T5000RaZW6ZLJYaZW6ZLJYaZW6Z@2YaJW5ZLFYaJW5ZLD0
a;o4_`8020L1004700000`L01`0@1`0300L700l710008gjGOYMnUgjGOP2GW9NLUibGW9NLUi`0]K2e
/;F`]K2e/;F`008000=[0000100;R`0:08^YRjV;ZH^YR`D002f@ZY2ZT:X0ZZVZZJZYZZVZZJZYZZVZ
ZJZYZZVZZJX0ZZVZZJZYZZVZZJZYZZT01P009JnY0:W6ZLJYaZW6ZLJYaZW6ZLJYaP36aLK5a/G6aLK5
a/D0a<D00P000lD01`071`400@L000031`07010700<01`L04@L60007OWUnNGii000;U`0;09NeUkFG
]INeUkD01`005H^5RhF;QH^5RhF;QH^5R`2TRZB:Y004000TRjV;ZH^YRjV;ZH/0RjV;ZH^YRjV;ZH^Y
RjV;ZH^YRjV;ZH/02JT5000YZJZYZZVZZJX0aZW6ZLJYaZW6ZLJYaZW6ZLJYaP35ZLFYaJW5ZLFYaJT0
0P001<G4a@081`400@L000031`07010700<01`L04`L20003EWh000<001EnU`2LUibGW9NLUibGW9L0
/;F`]K0010000f/00002012;00<0ZH/01000:jZ@ZY2ZT:Z@ZY2ZT:Z@ZP2ZZJZYZZVZZJZYZZVZZJZY
ZZVZZJZYZZT0ZJX01P00;:V_ZJnY[jV_ZJnY[jV_0<JYaZW6ZLJYaZW6ZLJYaZW6Z@35a/G6aLK5a/G6
0`001LW5bLD000P70@011`0000<700L040L00`071`0D1`8000EENEEiE@0900JG00<0]@0020004XE/
Q@2;QH^5RhF;QH^5RhF;Q@D002V;ZH^YRjV;ZH^YRjV;ZH^YRjV;08^YRjV;ZH^YRjV;ZH^YRjV;ZH^Y
R`06000^ZZVZZJZYZZVZZJZYZZVZZJZYZZT0ZLJYaZW6ZLJYaZW6ZLJYaZW60<FYaJW5Z@@000S6[P34
aLC500P70@011`0000<700L040L00`071`0F1`8000MnEWiFOUH000FG20000kFG00030006QVb6K8H0
2h/4000XRjX0ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2Z0:ZYZZVZZJZYZZVZZJZYZPH0032YZZT0ZJnY[jV_
ZJnY[jV_ZJnY[jV_ZJnY[`36ZLJYaZW6ZLJYaZW6ZLJYaZW60<H4000<aLK5a/G60<W5bLD020L10047
00000`L01`0@1`0300L701P70P004EEiEGUE07jGOYMnUgjGOYMn00D000m/QFb5K8E/08F;QH^5RhD0
1006R`0P08^YRjV;ZH^YRjV;ZH^YRjV;ZH^YR`2;ZH^YRjV;ZH/6000bZZVZZJZYZZVZ0:ZYZZVZZJZY
ZZVZZJZYZZVZZJZYZZT0ZLJYaZW6ZLJYaZW6ZLJYaZT4000@aZk6[/J^aZk6[P34aLC500P70@011`00
00<700L040L00`071`0J1`<000AFOUH02YL2000:K8I/QVb6K8I/00B;100098^ZRjZ;ZX^ZR`2@ZY2Z
T:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2Z0:ZYZPH003>YZZVZZJZYZZVZZJZYZZT0ZJnY[jV_ZJnY[jV_ZJnY[jV_
ZJnY[`36ZLJYaZW6ZLJYaZT01@005:o60<K5a/G6aLK5a/G60<W5bLD020L1004700000`L01`0@1`03
00L701`71P001gjGOYMnUgh00P002hE/QFb5K8E/QFb500D03H/05@2YRjV;ZH^YRjV;ZH^YRjV;ZH^Y
R`06000eZZVZZJZYZZVZZJZYZZVZZJZYZZVZ0:ZYZZVZZJZYZZVZZJZYZZVZZJZYZZVZ0<JYaZW6ZLH0
10006JW6ZLJYaZT0[/J^aZk6[/J^aZh0a<G4a@0020L1004700000`L01`0@1`0300L701h70P000eH0
00080009K8I/QVb6K8I/00<0026;ZX^ZRjZ;ZX^ZRjZ;ZX^ZR`2@ZY2ZT:Z@ZY2ZT:Z@ZY001@00=ZVZ
0:ZYZZVZZJZYZZVZZJZYZZVZZJZYZZVZZ@2Y[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_Z@2YaPD001g6[lJ_
aZo6[lJ_aP36aLK5a/G6aLK5aP39aLW500081`400@L000031`07010700<01`L080L20003D00000D0
00M/QFb5K8E/00<000=b08/04H/02`2YRjV;ZH^YRjV;00D003J@ZY2ZT:Z@0:VZZJZYZZVZZJZYZZVZ
ZJZYZZVZZJZYZZT0ZJZYZZVZZJZYZZVZZJZYZZVZZJX5000RaZW6ZLJYaZW6ZLJYaZW6Z@2^aZk6[/J^
aZk6[P34aLC500P70@011`0000<700L040L00`071`0R1`L000E/QVb6K00300>;01/0RjZ;ZX^ZRjZ;
ZX^ZRjZ;ZX^Z0:Z@ZY2ZT:X01000=jVZZJZYZZVZZJZYZP2ZZJZYZZVZZJZYZZVZZJZYZZVZZJZYZZVZ
0:nY[jV_ZJnY[jV_ZJnY[jT010009lJZ0:o6[lJ_aZo6[lJ_aZo6[lJ_aP36aLK5a/G6aLK5aP39aLW5
00081`400@L000031`07010700<01`L08PL60003K8E/00<000MbRg:;LX/001:;00<0RjT01@00=ZZ@
ZY2ZT:Z@ZY2ZT:Z@ZY00ZJZYZZVZZJZYZZVZZJZYZZVZZJZYZZVZZ@2YZZVZZJZYZZVZZ@D002RY[jV_
ZJl0aZW6ZLJYaZW6ZLJYaZW6ZLJY0:k6[/J^aZk6[/J^0<C50`081`400@L000031`07010700<01`L0
8@L:00V;0140RjZ;ZX^ZRjZ;ZX^ZRjZ;ZP05000fZZVZZJZYZZVZZJZYZZVZZJZYZZVZ0:ZYZZVZZJZY
ZZVZZJZYZZVZZJZYZZVZZJX0[jV_ZJnY1@00:Z[6Z/JZaZ[6Z/JZ0:o6[lJ_aZo6[lJ_aZo6[lJ_aP36
aLK5a/G6aLK5aP<00/T00`071`061`400@L000031`07010700<01`L080L8000<LX]bRg:;LX]bRg80
3H/4000gZX/0T:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY00ZJZYZZVZZJZYZZVZZJZYZZVZZJZYZZVZZ@2Y
ZP04000^ZJnY[jV_ZJnY[jV_ZJnY[`36ZLJYaZW6ZLJYaZW6ZLJYaZT0[/J^aZk6[/J^aP8000C5bLD0
2@L1004700000`L01`0@1`0300L701l71`0=R`0:0:Z;ZX^ZRjZ;ZP@003F@ZY2ZT:X0ZZVZZJZYZZVZ
ZJZYZZVZZJZYZZVZZJX0ZZVZZJZYZZVZZJZYZZVZZJZYZZVZZ@06000_aZ[6Z/JZaZ[6Z/JZaZ[6Z/JZ
aZX0[lJ_aZo6[lJ_aZo6[lJ_aZo60<K5a/G6aLH00`000lX0b@02b@0300L700L70@011`0000<700L0
40L00`071`0N1`D00V`03`1bRg:;LX]bRg:;LX]b0005R`@003>ZRjZ;ZX^ZRjZ;092ZT:Z@ZY2ZT:Z@
ZY2ZT:Z@ZY2ZT:Z@0:VZZJZYZZVZZJZYZZVZZJX01P00<ZVZZ@2Y[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_
0<JYaZW6ZLJYaZW6ZLJYaZW6Z@2^aZk60`002LZ_bZl0aLW500091`400@L000031`07010700<01`L0
7@L40004K6e/00f;00<0RjX01000<92ZT:Z@ZY2ZT:Z@ZY2Z0:ZYZZVZZJZYZZVZZJZYZZVZZJZYZZVZ
0:ZYZZVZZJZYZPL003F_ZZnZ[jZ_ZZl0aZ[6Z/JZaZ[6Z/JZaZ[6Z/JZaZ[6ZP2_aZo6[lJ_aZo6[lJ_
aZo6[lH0aP030007b/K:a/[60004b@0300L700L70@011`0000<700L040L00`071`0L1`<01F`0302;
LX]bRg:;LX]bR`D002jZRjZ;ZX^ZRjZ;ZX^ZRjZ;ZX/0T:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY00ZJZY
1P00>:ZYZZVZZJZYZZVZZJZYZZT0ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJnY[`36ZLJYaZW6ZLJYaZW6ZLJY
10003lZ_bZo:[lZ_bP39aLW500091`400@L0000300L7010700<01`L06`L20007K6e/KFa]0008R`@0
02bAR`2@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@0:VZZJZYZZVZZJZYZZVZZJZYZZVZZ@L003VZ[jZ_ZZnZ[jZ_
ZZnZ[jZ_ZZnZ[jZ_0<JZaZ[6Z/JZaZ[6Z/JZaZ[6Z/JZaZX0[lJ_aZo6[lJ_aZl01@003Zo60<[6b/K:
a/[6b/H00lT00`071`091`4000000`071`0@1`0300L701X70P06K00608]bRg:;1000:X^<Rhb;S02Z
RjZ;ZX^ZRjZ;ZX^ZRjZ;ZX^Z0:Z@ZY2ZT:Z@ZY2ZT:Z@ZPH003b@ZY00ZJZYZZVZZJZYZZVZZJZYZZVZ
ZJZYZZVZZ@2Y[jV_ZJnY[jV_ZJnY[jV_ZJnY[jV_0<JYaZW6ZLH4000G[lJ_aZo6[`2_bZo:[lZ_bZo:
0<W5b@002`L1000000<01`L040L00`071`0I1`8000U/KFa]K6e/08/01000:96;TH^ARi6;T@2ZT:Z@
ZY2ZT:Z@ZY2ZT:Z@ZY2ZT02YZZVZZJZYZZT6000m[jZ_ZZnZ[jZ_0:nZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ
[jZ_ZZl0aZ[6Z/JZaZ[6Z/JZaZ[6Z/JZaZ[6ZP2_aP05000GaZo6[lJ_aZo6[`36b/K:a/[6b/K:aP00
0lT00`071`091`4000000`071`0@1`0300L701P70`05K0@002J;S8^<Rhb;S8^<Rhb;08^ZRjZ;ZX^Z
RjZ;ZX^ZRjZ;ZX^Z0:Z@ZPH003j@ZY2ZT:Z@ZY2ZT:Z@ZY00ZJZYZZVZZJZYZZVZZJZYZZVZZJZYZZVZ
Z@2Y[jV_ZJnY[jV_ZJnY[jV_ZJnY[`D001k6[lJ_aZo6[lJ_aZo6[lH0bZo:[lZ_bZo:[lX0bLD200/7
0@00000300L700l700<01`L060L30004KFa]K08002B;096;TH^ARi6;TH^ARi6;T@2ZT:Z@ZY2ZT:Z@
ZY2ZT:Z@ZY06000n[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_0:nZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl0
aZ[6Z/JZaZ[6Z/JZaZX6000QaZl0[lJ_aZo6[lJ_aZo6[lJ_aZl0a/[6b/K:a/[6bP3900<030L10000
00<01`L03`L00`071`0G1`@00V`2000RKG9]08^<Rhb;S8^<Rhb;S8^<R`2;ZX^ZRjZ;ZX^ZRjZ;ZP@0
03j@ZP2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY00ZJZYZZVZZJZYZZVZZJZYZZVZZJZYZZVZZ@2Y[jV_
ZJnY[`H002FZ[jZ_ZZnZ[`36[lJ_aZo6[lJ_aZo6[lJ_aP3:[lZ_bZo:[lZ_008000?9/`0030L10000
00<01`L03`L00`071`0G1`<0029/002;KH]]08^ARi6;TH^ARi6;TH^ARi40ZY2ZT:Z@ZY2Z1@00?:ZA
ZY6ZT@2Z[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl0[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[`36ZPH0
02[6[lJ_aZo6[lJ_aZo6[`2_aZo6[lJ_aZo6[lJ_aZo6[`36b/K:a/[6b/H20003b[<000d70@000003
00L700l700<01`L05PL5000MKG9]LVeb08b;S8^<Rhb;S8^<Rhb;S02ZRjZ;ZX/01000>jZ@ZY2ZT:Z@
ZY2Z0:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT02YZZVZZJZYZZVZZJZYZZVZZJZYZZVZ00H002fZ[jZ_
ZZnZ[jZ_ZZnZ[jZ_ZZnZ[`36[lJ_aZo6[lJ_aZo6[lJ_aP3:[lZ_bZl00`001Jl0/lT000d70@000003
00L700l700<01`L05PL3000KRff;KH]]Rfd0Ri6;TH^ARi6;TH^ARi6;092Z00@003VAZY6ZTJZAZY6Z
TJZAZY40ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_0:nZ[jZ_ZZnZ[jZ_ZZnZ[jX01P00<:nZ[`36[lJ_
aZo6[lJ_aZo6[lJ_aZo6[lJ_0:o6[lJ_aZo6[lJ_aZo6[lH0b/K:aP<000O:1lX0/lX000h70@000003
00L700l700<01`L05@L4000FKG9]LVebKG80S8^<Rhb;S8^<Rhb;S0D003JZT:Z@ZY2ZT:Z@ZY2ZT:Z@
ZY2Z0:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT02YZZVZZJZYZZT7000cZZnZ[jZ_ZZnZ0:Z_ZZnZ[jZ_
ZZnZ[jZ_ZZnZ[jZ_ZP2_aZo6[lJ_aZo6[lJ_aZo6[`2_00<000Z_bZo:[lX0bK<03PL1000000<01`L0
3`L00`071`0D1`@001=]Rff;KH]]Rfd0Ri6;TH^ARi6;00@00Y40<`2AZY6ZTJZAZY6ZTJZAZY6ZTJZA
ZY40ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_0:nZ[`06000fZZnZ[jZ_ZZnZ[jZ_ZZnZ[`36[lJ_aZo6
[lJ_aZo6[lJ_aZo6[lH0aZo6[lJ_aZo6[lJ_aZo610003<X7bPO:1lX70<Zc00l70@00000300L700l7
00<01`L050L2000A601bKG9]LVebK@2;S8^<Rh`01000<X^ARi6;T@2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2Z
T:X0ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2Z1`00>:nZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jX0ZZnZ[jZ_ZZnZ
[jZ_ZZnZ[jZ_ZZnZ0:o6[lJ_aZo6[lJ_10004<J_0:o:[lZ_bZo:[`2cb@0?1`4000000`071`0?1`03
00L701<7010061PH06f;KH]]Rff;096;100:T@0V096ZTJZAZY6ZTJZAZY6ZTJZAZY6ZT@2Z[jZ_ZZnZ
[jZ_ZZnZ[jX6000j[jZ_0:nZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl0aZo6[lJ_aZo6[lJ_aZo6[lJ_
aZo60<J_aZo6[`D001>_1jl7[`3:1lX7bPO:1lX0/lX001070@00000300L700l700<01`L04`L0300H
001]LVebKG9]LP@002j;TH^ARi6;TH^ARi6;T@2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:X0ZY2ZT:Z@ZY2Z
1P00?96ZTJZAZY6ZT@2Z[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl0[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_
ZP2_aP@001N_aZo6[lJ_aZo60<Z_bZo:[lZ_bP39000A1`4000000`071`0?1`0300L7018700X061PH
08]]Rff;0`000h`0T@0?T@0J096ZTJZAZY6ZTJZAZY6ZTJZAZY6ZT@2Z[jX6000m[jZ_ZZnZ[jZ_ZZnZ
[jZ_0:nZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZP2_aZo6[lJ_aZo6[lJ_aZo6[lJ_aP05000G1jl7[`N_
1jl7[`N_1`07bPO:1lX7bPL00`0B1`4000000`071`0?1`0300L701870P001QP006ebK@8002V<LXab
08^ARi6;TH^ARi6;TH^ARi40ZY2ZT:Z@ZY2ZT:Z@ZY2ZT:Z@ZP06000mTJZAZY6ZTJZAZY6ZTJZAZY6Z
TJZA0:Z_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[`2_ZZnZ[jZ_ZZnZ[jZ_ZP0600:_01X0[lJ_aZo6[lJ_
aZo6[lJ_0:o:[lZ_bZo:[`<04PL1000000<01`L03`L00`071`0B1`0501PH06d00`001g:<LXabS000
49403P2AZY6ZTJZAZY6ZTJZA1P00?JnA0:Z_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[`2_ZZnZ[jZ_ZZnZ
[jZ_ZZnZ[jZ_ZZnZ[jX0[lJ_aZo6[lH01P008:o7[lN_ajl0[`N_1jl7[`N_1jl7[`N_1`07bPO:1lX7
0`000`L01`0A1`4000000`071`0?1`0300L7014700@0600H0P008g:<LXabS7:<LP2;TH^ARi6;TH^A
Ri6;TH^A0:Z@ZY2ZT:Z@00D003fAZY6ZTJZAZP2ZTJZAZY6ZTJZAZY6ZTJZAZY6ZTJZAZY40ZZnZ[jZ_
ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_0:nZ00H03Jl05P36[lJ_aZo6[lJ_aZo6[lH0bZo:[lX5008700<01`L0
40L1000000<01`L03`L00`071`0A1`0301PH008000]bS7:<LXabS7:<000@T@03096Z00H003bA[i6_
TJnA[i6_TJnA0:Z_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[`2_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl6000W
[lN_ajo7[lN_ajo7[lN_ajo700N_1jl7[`N_1jl7[`N_1jl0bPO:008000@7]0000`L00`071`0@1`40
00000`071`0?1`0300L701470`006d/0S7:<LXabS7:<LP2;TH^ARi6;TH^ARi6;T@05000lZY6ZTJZA
ZY6ZTJZAZY6ZTJZAZP2ZTJZAZY6ZTJZAZY6ZTJZAZY6ZTJZAZY40ZZnZ[jZ_ZZnZ[jZ_ZZnZ1P001:nZ
[`0D[`0B0:o6[lJ_aZo6[lJ_aZo6[`2_0`000k@7]00200@700<01`L040L1000000<01`L03`L00`07
1`0@1`800d/0301bS7:<LXabS7:<00ZA1000>jZA096_TJnA[i6_TJnA[i6_TJnA[i6_T@2Z[jZ_ZZnZ
[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl0[jZ_ZZnZ[jZ_00H002b_/:n`[k2_/:l0[lN_ajo7[lN_ajo7[lN_ajo7
[lL01jl7[`N_1jl7[`N_1`@000Nd1k@7]0L000D700<01`L040L1000000<01`L03`L00`071`0@1`0G
01Q;64/0S7:<LXabS7:<LP2;TH^ARi401006T@0c0:ZAZY6ZTJZAZY6ZTJZAZY6ZTJZAZY40TJZAZY6Z
TJZAZY6ZTJZAZY6ZTJZAZY40ZZnZ00H000n_ZZnZ[jZ_ZZnZ[jZ_ZP005:l02`36[lJ_aZo6[lJ_00@0
00X700Nd1k@7]0L01PL00`071`0@1`4000000`071`0?1`0300L7010700<0Bd/00T/0301bS7:<LXab
S7:<00:A1000=ZZAZY6ZTJZAZY40TJnA[i6_TJnA[i6_TJnA[i6_TJnA[`2_ZZnZ[jZ_ZZnZ[jZ_ZZnZ
[jZ_ZPL002n_/:n`[k2_/:n`[k2_/:n`[k2_/037[lN_ajo7[lN_ajo7[lN_ajo7[`2_1jl7[`05000=
[`N_1`07]0Nd1k@700071`0300L701070@00000300L700l700<01`L040L03`0HBaQ;08abS7:<LXab
S00400jA02H0ZY6ZTJZAZY6ZTJZAZY6ZTJZAZY6ZT@2AZY6ZTJZAZY6ZTJZAZPH001bA[i40ZZnZ[jZ_
ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jX04jl00`2_aP04000A[`N_1jl7[`L01k@7]0Nd1`0020L00`071`0@1`40
00000`071`0?1`0300L7010700<0Bd/00T/01`1bS7:<LX`00`00<Y40ZY6ZTJZAZY6ZTJZAZY6ZT@2A
[i6_TJnA[i6_TJnA[i6_TJnA[i6_0:nZ[jZ_ZZnZ1P00<[2Z/:Z`Z[2Z/02`[k2_/:n`[k2_/:n`[k2_
/:n`[k2_/037[lN_ajo7[lN_ajo7[lN_1P004PN_1jl7[`N_1jl7[`2d1k@7]0802@L00`071`0@1`40
00000`071`0?1`0300L7010700X064/HB`2<LXab0P001I5cTG<0016A01T0TJZAZY6ZTJZAZY6ZTJZA
ZY6ZTJZA096Z00H002NA[i6_TJnA[i6_TJnA[i40ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[`002Zl7000E
[k2_0:l7[`N_1jl7[`N_1jl0]0Nd00@02PL00`071`0@1`4000000`071`0?1`0300L7010700<0Bd/0
0T/00`1b0002000/Li5cTG>AL`2AZY6ZTJZAZY6ZTJZAZY6Z0:nA[i6_TJnA[i6_TJnA[i6_TJl5000`
/:Z`Z[2Z/:Z`Z[2Z/:Z`Z[2Z/:Z`0;2_/:n`[k2_/:n`[k2_/:n`[k2_/:l0[lN_1`006Zl7[`N_1jl7
[`2_1jl7[`N_1jl7[`N_0;@70P000k@01`0;1`0300L701070@00000300L700l700<01`L040L01@0H
BaQ;008000]cTG>ALi5cTG>A000@T@0>096ZTJZAZY6ZTJZAZY46000_ZY40TJnA[i6_TJnA[i6_TJnA
[i6_TJnA[i40ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZl01P007Jn`[k2_/:n`[k2_/:n`[k001jl7[`N_1jl7
[`N_00@000>e00003@L00`071`0@1`4000000`071`0?1`0300L7010700@0Bd];0P009G>ALi5cTG>A
Li5c096ZTJZAZY6ZTJZAZY6ZTJX0[i6_TJnA[i401@00;i6`TK2A/96`0;2Z/:Z`Z[2Z/:Z`Z[2Z/:Z`
Z[2Z/:Z`0;2_/:n`[k2_/:n`[k2_00H0026`[`2_1jl7[`N_1jl7[`N_1jl7[`L01jl7[`N_1jl7[`L0
0`001;Fd]@0?1`0300L701070@00000300L700l700<01`L040L04P0H001@095cTG>ALi5cTG>A012A
00<0TJX01P00;Y6ZTJZAZY6ZTJZAZY40TJnA[i6_TJnA[i6_TJnA[i6_TJnA[i40ZZnZ[jZ_ZZl6000S
[k2_/:n`[`2_/:n`[k2_/:n`[k2_/:n`[k2_0:l7[`N_1jl01@000i^eV`02010700<01`L040L10000
00<01`L03`L00`071`0@1`8001i;D4]@095cTG>ALi5cTG<0TJZAZY6ZTJZAZY6ZTJX5000^/96`TK2A
/96`TK2A/96`TK2A/02`Z[2Z/:Z`Z[2Z/:Z`Z[2Z/:Z`Z[2Z/02`[`H002F`[k2_/:n`[k2_/:n`00N_
1jl7[`N_1jl7[`N_1jl7[`2_1jl700@00[D0102e];D2018700<01`L040L1000000<01`L03`L00`07
1`0@1`0C04]@Be1;07>ALi5cTG>ALi5c0007T@H00Y40;02AZY6ZTJZAZY6ZTJZAZY6ZTJZAZY6Z0:nA
[i6_TJnA[i6_TJnA[i6_TJnA1@009Zn`[k2_/:n`[k2_/:n`[k2_/:l0[k2_/:n`[k2_/:n`[k2_/:n`
1@002;Fd]KBe0;FK0P0D1`0300L701070@00000300L700l700<01`L040L05@1@Be1;D02ALi5cTG>A
Li5cT@2ZT@05000^TK2A/96`TK2A096`TK2A/96`TK2A/96`TK2A/96`T@2Z/:Z`Z[2Z/:Z`Z[2Z/0H0
02R_/02`[k2_/:n`[k2_/:n`[k2_/:n`[`2_1jl7[`N_1jl7[`N_1jl71007]@D05PL00`071`0@1`40
00000`071`0?1`0300L701070140Be1;D4/0Li5cTG>ALi5cT@0400jA01l0ZY6ZTJZAZY6ZTJZAZY6Z
TJZAZY6Z0:nA[i6_TJnA00H002V`TK2A/96`T@2_/:n`[k2_/:n`[k2_/:n`[k2_/02`[k2_/:n`[k2_
/005000:]@2e];Fd]KBe]0<06@L00`071`0@1`4000000`071`0?1`0300L7014700d0Be1;D02ALi5c
TG>A00<002^B0;2A/96`TK2A/96`TK2A/940TK2A/96`TK2A/96`TK2A/96`TK2A0:Z`00H002Z_/:n`
[k2_/:n`[k2_/02`[k2_/:n`[k2_/:n`[k2_/:n`[`2_1jl7[`L50004]K2e/0801[D401/700<01`L0
40L1000000<01`L03`L00`071`0A1`0;051;D4]@095cTG<00P001Y5cTG>A012A01<0TJZAZY6ZTJZA
ZY6ZTJZAZY6Z00D002^`TK2A/96`TK2A/96`TK2A/96`T@2_/:n`[k2_/:n`[k2_/:n`[k2_0:n`00D0
00fe[kF_]Jne[`2d]KBe00D07PL00`071`0@1`4000000`071`0?1`0300L7014700P0Be1;D4/0L`<0
02JATY6BTI:A096`TK2A/96`TK2A/96`TK00/96`TK2A/96`TK2A/0D002Z`T@2_/:n`[k2_/:n`[k2_
/:n`[k2_/:n`0;2_/:n`[k2_/:n`[k2_/:l6000>]K2e/;F`]K2e/;F`0;D6024700<01`L040L10000
00<01`L03`L00`071`0A1`06051;D4]@0P002g>ALi5cTG>ALi40012A00T0TJZAZY6ZTJX01@00:;2A
/96`TK00/96`TK2A/96`TK2A/96`TK2A/96`T@2_/:n`[k2_/:l7000?[k00]Jne[kF_]Jne[kF_00D0
00Im0000Q00Q1`0300L701070@00000300L700l700<01`L04PL00`1@B`03000PTI:ATY6BTI:ATY6B
0;2A/96`TK2A/96`TK2A/940TK05000W/96`TK2A/96`TK2A/02`[k2_/:n`[k2_/:n`[k2_/:n`[k00
/:n`00H001:`]K2e/;F`]K00/;F`]K2e/;D60003UP00008000>XX`000P0P1`0300L701070@000003
00L700l700<01`L04PL3000@D7=@07>ALi5cTG>ALi5c00jA1@009[2A/96`TK2A/96`TK2A/96`T@2A
/96`TK2A/96`TK2A/96`TK2A1P005;2_/:n`[k2_/:n`[k2_0:ne[kF_10000ibG00060006RP2RXj:S
0P0O1`0300L701070@00000300L700l700<01`L04PL06`1cD7=@L`2BTI:ATY6BTI:ATP2`TK2A/96`
T@06000V/96`0;2A/96`TK2A/96`TK2A/96`TK00/:n`[k2_/:n`[k2_/:l500:`01@0]K2e/;F`]K2e
/;F`]K2e/;D0]@@000BL]Ibe10001J>TXjBS00<000JSZ:>XX`0N1`0300L701070@00000300L700l7
00<01`L04`L04`1cD7=@L`2ALi5cTG>ALi5cT@000Y45000VTINAUi6GTINA096`TK2A/96`TK2A/96`
TK2A/940TK2A/96`TK06000FU[2F/9J`UP2_/:n`[k2_/:n`[k2_/0D000FLUibGW005000:XhZSRZ>:
X`2YX`8000>SXZ<00P0M1`0300L701070@00000300L700l700<01`L04`L04@1@Le1cD02ATY6BTI:A
TY6B00@002JA/96`TK2A/96`TK2A/940TK2A/96`TK2A/96`TK2A/96`T@2_/0D03K002`2e/;F`]K2e
/;F`00@000Je0;FL]I`;000:XjBSY02YXjVSZ@8000BS0:8070L00`071`0@1`4000000`071`0?1`03
00L701@700`0D7=@Le00Li5cTG<4000TTWP0TINAUi6GTINAUi6GTINAU`2`TK2A/96`TK2A/96`TK2A
1P006;2F/9J`U[2F/9J`U[2F/9J`0;2_/:n`[`D000JG]INe09`40004Q8F4Q@8000F:Y8ZTRP040008
ZJ>YXjVSZJ<20003_j8001/700<01`L040L1000000<01`L03`L00`071`0D1`0907=@Le1c09:A00<0
02:AUi6GTINA096`TK2A/96`TK2A/96`TK00/96`TK2A/96`1P0019N`U`0C/0030;2e00@000NL]Ibe
W;FL00@000I[08Z5RXD20008Y8ZTRZB:Y8X2000?Z@2YXjVSZJ>Y0<00_j8001X700<01`L040L10000
00<01`L03`L00`071`0E1`0607=@Le1c0P007gRBN99hTWRBN99h096GTINAUi6GTINAUi6GT@2A/940
1P006[2F/9J`U[2F/9H0U[2F/9J`U[2F/9J`U[2F1P001kFG]INeUkD01P000f/000060007Y8ZTRZB:
Y0020004XjVSZ@8000>SZJ<00P001jVXZ@2oXP006@L00`071`0@1`4000000`071`0?1`0300L701D7
0P0087=@001i09NAUi6GTINAUi6G0;2A/96`TK2A/96`TK2A1@004IN`Uk2G/9N`Uk2G/9N`Uk0000V`
1P002INeU`2L]IbeW0040003K@2;00D000>5JhD00P000hZT00020004Y8X000RY00/0X`2Y`:W0ZL00
_`0201P700<01`L040L1000000<01`L03`L00`071`0F1`0L0500D7U@N@2BN99hTWRBN99hTP2GTINA
Ui6GT@D001VAUi40U[2F/9J`U[2F/9J`U[2F/9J`0;2F00H000^eUkFG]INeUkD0]@0500:;1`000f^4
J`020007Y8ZTRZB:Y002000:XjVSZJ>YXjVSZ@8000RXZJRYZ:VXZ@<060L00`071`0@1`4000000`07
1`0?1`0300L701H70P005WQiN7UhN@2GTINAUi6GTINAU`2`TK04000IUk2G/9N`Uk00/9N`Uk2G/9N`
Uk2G/9N`U`05000?UkFG]INeUkFG]INeUkFG008000>Z0:X00P000h/0K005000:JhD008ZTRZB:Y0<0
0ZT200JY0P003/2Y0:W0ZL2Y`:W00;l05`L00`071`0@1`4000000`071`0?1`0300L701L70P004E1i
D7U@07RBN99hTWRBN99h00@001VGTINAUi6GTINAUi6G0;2F/9J`U[2F/9J`00@000fG/9L0UkFG]INe
UkFG00L000F@ZP00R`030003K00000L000B:Y8ZT0`001jV;ZH^YRjT00P001:>Y0005Z@0<0:RYZ:VX
Z@300;h05PL00`071`0@1`4000000`071`0?1`0300L701L70`0037QiN7Uh096GTINAU`@001VBU`2`
Uk2G/9N`Uk2G/9N`Uk00/9N`Uk2G00@000beUkFG]INeU`2G]IL50004ZP00[`<000RZT@2;08/0K0L0
00>:Y8X0100;Z@<000c0ZL2Y`:W0Z@2Y`:T20005aJT0_P005PL00`071`0@1`4000000`071`0?1`03
00L701P700/0B`1@NE1iD01hTP03000IUi6GTINAU`2GTINAUi6GTINAUi6GT@2F/004000=Uk2G/9N`
Uk2G/9N`U`050004ZP2Z[`L000FZ0940R`060004QHZ5RP<000^YR`2;ZH^YRjV;Z@0300:Y00<0ZJT0
1jT02`2X0:W0ZL2Y0:8001D700<01`L040L1000000<01`L03`L00`071`0I1`8000IiN7UhNGP2000H
Ui:GTYNBUi:GTYL0/9N`Uk2G/9N`Uk2G10002kFG]INeUkFG]INe00L000W7[`2Z/02_0:l01P02T@D0
00JTRZB:Y8X20006ZJBYY:T01ZT2000?ZZVZZJZY0:W0ZL2Y`:W0008000VYaJW5ZLD0_`005@L00`07
1`0@1`4000000`071`0?1`0300L701X70P000gU@N@02000FN02AUi6GTINAUi6GTIL0Ui6GTINAU`@0
19L30004/9N`U`L000Z_/02_ZP2Z[`2_0P000i40T@090005Q8D0RXD00`002h^YRjV;ZH^Y0:V;00<0
2:T00`2YZ@03Z@<00ZT01`2Y`:W0ZL000`0D1`0300L701070@00000300L700l700<01`L06`L20006
N@2GNIMi0P003YNBUi:GTYNBUi:G0;2G10002kFG]INeUkFG]INe00X000^_1`00[`2Z/:X0ZP020003
[`2_00D000J:QHZ5RXD3000:Y:VTZJBYY:VTZ@@000nYZZVZZJZYZZVZZJX0`:T00P003<JYaZW60<FY
aJW5Z@<050L00`071`0@1`4000000`071`0?1`0300L701/701D0C01iN7UhNGQi09NAUi6GTINAUi40
0`0;U`@000>LUi`00`001[D00:n`00:_0P002:nZ096_T@2A0`001XB5Q8F4Q@@000V;Z@2YRjV;ZH/0
10001:V@Z@0<Z@<02:T01`30ZL2Y`:T00P0D1`0300L701070@00000300L700l700<01`L070L4000=
NIMiUgVG09NBUi:GTP0300>G10000kFG]@040005UkFG]IL010004@Ne0:l70<N_a`2`Z[00[jZ_00<0
00B408Z510002JV;ZH^YR`2TZ@04000AZJZYZZVZZ@2YZZVZZJZYZZT00`004:T0ZLJYaZW6ZLJYaP35
ZLD401@700<01`L040L1000000<01`L03`L00`071`0M1`8000Q@07QiN7UhN@8000BAU`002YL50005
W9NLUi`010000jne00020003/02_00:_00H0[`2_0:l30004Q6^4J`<000V;ZH^YRjV;ZH/010003JV@
ZI2YT:V@ZI2YT0001JT30006ZJZYZZT02JT01`2Y0035ZLD00P0C1`0300L701070@00000300L700l7
00<01`L07PL00`1FD0020005UgVGNIL00`0:U`8000FeUkFG]@0700061kD000N`0P001/N_a`2`[`D0
00=[Q0000`0028/0RjV;ZH^Y10004jVZ0:ZYZZVZZJZYZZVZZJX0ZZT00`004ZV_ZJnY[jV_0<JYaZW6
ZLJYaP8000G5a/G6a@0201<700<01`L040L1000000<01`L03`L00`071`0O1`8000E@EP00N0020003
UgjG00@019L20007UibGW9NLU`040003]@2e008000F_]Jne0003[`030:nZ00P000R;RX^:R`2YR`@0
01>YRjV;ZH/0T:V@ZI2YT:V@ZI2Y00@000fYZZVZZJZYZZVZZJX000FY0`001lH0aJW5ZLD00P0C1`03
00L701070@00000300L700l700<01`L08@L01@1@EU0000VG3@05]@8000@7/0N`0`001:o70;060007
RjV;ZH^YR`05000BZJZYZZVZZJZYZZT0ZJZYZZVZ1@004jZY0:V_ZJnY[jV_ZJnY[jT0ZLH00`002/JY
aZW60<K5a/D201<700<01`L040L1000000<01`L03`L00`071`0R1`8000=F07h00`001iMnUgjGOP00
29L01@2eUkFG00<000B_]Jne0P000jn`[`080005RhZ;RX/010004h^Y0:V;ZH^YRjV;ZH^YRjT0ZI00
10005JZYZZVZZJZY0:VZZJZYZZVZZJZYZP03000?aZW6ZLJYaZT0ZLFYaJT001<700<01`L040L10000
00<01`L03`L00`071`0T1`<000=nNGh00`02U`0;09NeUkFG]INe0;D010002k2e00N`1`2e[kF_00@0
00J;RX^:08/4000BZY2ZT:Z@0:VZZJZYZZVZZJZY1@005ZVZZJZYZZVZZJZYZP2_ZJnY[jV_ZJl3000B
aP36ZLJYaZW6ZLH0a/G6aLH04`L00`071`0@1`4000000`071`0?1`0300L702H70P001GiiOWUn00d0
00NeUkFG]INe00<000>`[k000`001H^5RhF;00@0016;ZH^YRjV;ZH^Y0:V;ZH^YR`05000GT:X0ZZVZ
ZJZYZZVZZJZYZZT0ZJZYZZT010004ZnY[jT0ZLJYaZW6ZLJY0:W5Z@@04PL00`071`0@1`4000000`07
1`0?1`0300L702P70`0057iiOP2LUibGW9NLU`2`]K2e/;F`0`000jne[`030003RX^:00@0016Z0:Z@
ZY2ZT:Z@ZY2ZT02YZP04000IZJZYZZVZZ@2YZZVZZJZYZZVZZJZYZZT0Z@03000CZ/JZaZ[6Z/H0aZW6
ZLJYaZW6Z@030004aLH0018700<01`L040L1000000<01`L03`L00`071`0Z1`@000=n09L01iL00`2G
]@040003Uk2G00<000=[00000`0048^@Ri2;08^YRjV;ZH^YRjT5000IZY2ZT:Z@ZY2ZT:X0ZZVZZJZY
ZZVZZJZYZP04000EZJnY[jV_ZJnY[jV_0<JYaZW6ZLJY008000C6aLK50P0B1`0300L701070@000003
00L700l700<01`L0;@L>0005]K2e/;D00`004hD0002ZRjZ;ZX^ZR`2@ZY2ZT:X01@006JZY0:VZZJZY
ZZVZZJZYZZVZ0:ZYZZVZZJX01@005JnZ0:[6Z/JZaZ[6Z/JZaZX0ZLJYaP040006[`35a/G60P0B1`03
00L701070@00000300L700l700<01`L0;`L30007UgjGOYMnU`020003Uk2G00H000j;092;T8^@Ri2;
T02YR`@001VZT:Z@ZY2Z0:Z@ZY2ZT:Z@ZY2ZT:Z@0:VZ00D001NZZJZYZZVZ0:nY[jV_ZJnY[jV_ZJl0
aP03000:aZW6ZLH0a/G6a@804PL00`071`0@1`4000000`071`0?1`0300L703873`03R`080:Z;ZX^Z
RjX6000HZJZYZZVZZJZYZZT0ZJZYZZVZZJZYZZVZ1@006:nZ[jZ_ZZnZ[jZ_ZP2ZaZ[6Z/JZaZ[6ZP@0
00k6[lJ_aZo6[`35a/G601<700<01`L040L1000000<01`L03`L00`071`0e1`800gh50007RX^:RhX0
R`06000IZX^Z0:Z@ZY2ZT:Z@ZY2ZT:X0ZY2ZT:Z@ZP05000IZY00ZJZYZZVZZJZYZZVZZJZY0:V_ZJnY
[`04000BZJl0aZW6ZLJYaZW60<K5a/D04`L00`071`0@1`4000000`071`0?1`0300L703L71P05R`@0
01Z@ZY2ZT:Z@ZY00ZJZYZZVZZJZYZZVZZ@2YZPD001Z_ZZnZ[jZ_0:nZ[jZ_ZZnZ[jZ_ZZnZ[`36ZP@0
01?6Z/JZaZX0[lJ_aZo6[lJ_0<G600<04`L00`071`0@1`4000000`071`0?1`0300L703P70P02K003
08Z;00<001R@0:Z;ZX^ZRjZ;ZX/0T:Z@ZY2ZT:Z@ZY06000JZY2ZT:Z@ZY2ZT:Z@0:VZZJZYZZVZZJZY
ZZT5000DZJnY[jV_ZJnY[`36ZLJYaZW6ZLH30003b/D001<700<01`L040L1000000<01`L03`L00`07
1`0g1`0406a/K0@001>ZRjX0ZY2ZT:Z@ZY2ZT:X0ZZVZ00P00jX0602Z[jZ_ZZnZ[jZ_ZZnZ[jX0ZZnZ
[jZ_ZPD001RZ[`36Z/JZaZ[6Z/JZaZX0[lJ_aZo6[lH20005a/[6bP004`L00`071`0@1`4000000`07
1`0?1`0300L703H700<0K6`00`00492;T8^@R`2;ZX^ZRjZ;ZX/6000NZY2ZT:Z@ZY2ZT:X0ZY2ZT:Z@
ZY2ZT:Z@ZY2Z0:ZY1@006jV_ZJnY[jT0ZJnY[jV_ZJnY[jV_Z@2YaZW6Z@020006aP3:aLX050L00`07
1`0@1`4000000`071`0?1`0300L703D71@003X^ZRjZ;ZX^Z0:Z@ZY2Z10000i6Z000>ZP0>0:Z_ZZnZ
[jZ_ZZnZ[jX5000NZZnZ[jZ_ZZnZ[jZ_0<JZaZ[6Z/JZaZ[6Z/H0aZo60P002:o6[`36b/H050L00`07
1`0@1`4000000`071`0?1`0300L703@70P003X]]R`2@Ri2;T8^@0:Z;10007JZ@ZY2ZT02@ZY2ZT:Z@
ZY2ZT:Z@ZP2ZT:Z@ZY2Z00H0026@ZP2_ZJnY[jV_ZJnY[jV_Z@2Y[jV_ZJnY[jV_ZJnY0:T00P002/J_
aZo60<[5bP0D1`0300L701070@00000300L700l700<01`L0<`L2000:Rg:;0:Z;ZX^ZR`D000^AZY6Z
TJZAZY6Z000>ZP030:Z_00D002>_ZZnZ[jZ_ZP2Z[jZ_ZZnZ[jZ_ZZnZ[`36Z/JZaZ[6Z/JZaP04000<
[lJ_aZo6[`36b/H050L00`071`0@1`4000000`071`0?1`0300L703870P001X]]Rfd0R`D001b;T@2Z
T:Z@ZY2ZT:Z@ZP2ZT:Z@ZY2ZT:Z@ZY2Z1@008jZ@ZY2ZT:Z@ZY2ZT:X0[jV_ZJnY[jV_ZJnY[jT0ZJnY
[jV_00D000n_ZP2_aZo6[lJ_0<G:a@005@L00`071`0@1`4000000`071`0?1`0300L703470`000g:;
0003000DT:Z@ZY2ZT02AZY6ZTJZAZY6ZT@07ZPH002B_ZP2Z[jZ_ZZnZ[jZ_ZZnZ[jX0ZZnZ[jZ_ZZnZ
[jZ_ZZl0aZX4000AaZo6[lJ_0:o6[lJ_aZo60<X00`0E1`0300L701070@00000300L700l700<01`L0
<0L3000MK@00S02ARi6;TH^ARi40ZY2ZT:Z@ZY2ZT:X0ZY001@00992ZT:Z@ZY2Z0:Z@ZY2ZT:Z@ZY2Z
T:Z@ZP2_ZJnY[jV_ZJnY[`H001>Z[jZ_ZZnZ[jZ_0<J_aZo6[lJ_008000?:00L05@L00`071`0@1`40
00000`071`0?1`0300L7030710005X^AR`2@ZY2ZT:Z@ZP2ZTJZAZY6ZTJX6000QZZnZ[jZ_ZZnZ[jZ_
ZP2Z[jZ_ZZnZ[jZ_ZZnZ[jX0ZZnZ00P001R_ZZl0aZo6[lJ_aZo6[lJ_0:o6[lJ_aZl200;:00<01`L0
50L00`071`0@1`4000000`071`0?1`0300L702l70`004XabS7:<096;TH^ARi6;092ZT0H0026ZT:X0
ZY2ZT:Z@ZY2ZT:Z@ZY2Z0:Z@ZY2ZT:Z@ZY2ZT:X01P007:Z_ZZnZ[jZ_ZZnZ0:Z_ZZnZ[jZ_ZZnZ[`36
[lH40004[`3:01L700<01`L040L1000000<01`L03`L00`071`0_1`<000b;TH^A0:Z@ZY2ZT:X6000Q
TJZAZY6ZTJZA0:Z_ZZnZ[jZ_ZZnZ[jZ_ZP2Z[jZ_ZZnZ00H001n_ZP2Z[jZ_ZZnZ[jZ_ZZnZ[`36[lJ_
aZo6[lJ_aP3600<000D7[`N_0002bP0300L701D700<01`L040L1000000<01`L03`L00`071`0^1`<0
00QbS7:<LP2;T@D002:AR`2@ZY2ZT:Z@ZY2ZT:X0ZY2ZT:Z@ZY2ZT:Z@ZY2Z0:Z@1@008:Z_ZZnZ[jZ_
0:nZ[jZ_ZZnZ[jZ_ZZnZ0:Z_ZZnZ[jZ_1@002/J_aZo6[lH0bP0H1`0300L701070@00000300L700l7
00<01`L0;PL01`1;096;TH/010008:ZAZY6ZTJX0ZY6ZTJZAZY6ZTJZA0:Z_ZZnZ[jZ_ZZnZ1`007jZ_
ZZnZ[jZ_ZZnZ[jX0ZZnZ[jZ_ZZnZ[jZ_ZZl0aZl01P002lN_a`07[`N_1jl700<06@L00`071`0@1`40
00000`071`0?1`0300L702h70P000h`00002000LR`2;TH^ARi6;TH/0T:Z@ZY2ZT:Z@ZY2Z0:Z@ZPP0
026AZY6Z0:nZ[jZ_ZZnZ[jZ_ZZnZ[`2_ZZnZ[jZ_ZZnZ[jX01@08[`090<J_aZo6[lJ_00806PL00`07
1`0@1`4000000`071`0?1`0300L702d701d06000THbAS940ZY6ZTJZAZY6Z0:ZAZY6ZTJZAZP07000T
[jZ_ZZnZ[jZ_ZZnZ0:Z_ZZnZ[jZ_ZZnZ[jZ_ZP2Z[jZ_ZZnZ1@02[`0B0<N_ajo7[lN_ajl0[`N_1jl7
0`0K1`0300L701070@00000300L700l700<01`L0;@L3000DTH^ARi6;08^ARi6;TH^AR`2@ZY06000V
ZY6Z0:ZAZY6ZTJZAZY6ZTJZAZP2_ZZnZ[jZ_ZZnZ[jZ_ZZl0[jX50007ZZnZ[jZ_000;[`040<J_aPD0
70L00`071`0@1`4000000`071`0?1`0300L702d70140B`2<THbAS940ZY6ZTJZAZP06000WTJnA[i6_
TJnA0:Z_ZZnZ[jZ_ZZnZ[jZ_ZP2Z[jZ_ZZnZ[jZ_ZZnZ00D02jl03@2_ajo7[lN_ajo700L00`0010N_
000M1`0300L701070@00000300L700l700<01`L0;@L0301]096;TH^AR`2;T@D002NZT@2AZY6ZTJZA
ZY6ZTJX0ZY6ZTJZAZY6ZTJZAZY6Z0:nZ[jZ_ZZl01P004JZ_0:nZ[jZ_ZZnZ[jZ_ZZl000R_100010N_
1jl201h700<01`L040L1000000<01`L03`L00`071`0]1`0804/0S96<TH`4000WZY6ZTJZAZP2_TJnA
[i6_TJnA[i6_0:nZ[jZ_ZZnZ[jZ_ZZnZ0:Z_00D000V_ZZnZ[jZ_ZP003Jl01@2_ajo700@000R_0:l7
[`N_020700<01`L040L1000000<01`L03`L00`071`0]1`0406d0T@@002F<096ZTJZAZY6ZT@2AZY6Z
TJZAZY6ZTJZA096ZTJZAZY6ZTJZA00H001b_ZZnZ[jZ_ZZnZ[jZ_0:nZ[jZ_ZZnZ[jZ_ZP2_10002K2_
/:l0[`N_1`02024700<01`L040L1000000<01`L03`L00`071`0]1`0404/000FA01`0ZY6ZTJZAZY6Z
T@2A[i6_TJnA[i6_TJl0[jZ_20004jZ_ZP2Z[jZ_ZZnZ[jZ_ZZnZ[`002Zl5000:[`N_1jl7[`L01`@0
8PL00`071`0@1`4000000`071`0?1`0300L702d70`006i6<THbAS02AZY6ZTJZAZY6Z0:ZAZY6ZTJZA
ZP07000N[i6_TJnA[i6_TJl0[jZ_ZZnZ[jZ_ZZnZ[jX0ZZnZ1`003:Z_0;2_/:n`[k2_/0<000>e00L0
8`L00`071`0@1`4000000`071`0?1`0300L702d700<0L`001i403P2AZY6ZTJZAZY40TJnA1P007ZnA
[`2_ZZnZ[jZ_ZZnZ[jZ_ZP2Z[jZ_ZZnZ[jZ_ZPL0016_/:n`[k2_/:l0[`N_1jl7[`030003]@0002D7
00<01`L040L1000000<01`L03`L00`071`0]1`0?07<0THbAS96<T@2ZTJZA00T001bAZY6ZTJZAZY40
TJnA[i6_TJnA[i6_TJl0[jZ_20005:Z_ZP2Z[jZ_ZZnZ[jZ_ZP2_/:n`1@000kD0000W1`0300L70107
0@00000300L700l700<01`L0;@L0101cL`05T@H001jA/96`T@2A[i6_TJnA[i6_TJnA0:Z_ZZnZ[jZ_
ZZl7000G/:n`[k2_/:n`[`2_/:n`[k2_/:n`[k001@000`N`000302T700<01`L040L1000000<01`L0
3`L00`071`0]1`0507=;08`010007I40ZY6ZTJZAZY6Z0:ZAZY6ZTJZAZY6ZTJX0[i6_00L001Z_TJl0
[jZ_ZZnZ[jZ_ZZnZ[`2_ZZnZ[jZ_ZPD000F_]Jne[`0302/700<01`L040L1000000<01`L03`L00`07
1`0^1`0N07<0096BTI:ATP2`TK2A/96`TK00[i6_TJnA[i6_1`006k2Z/:Z`Z[2Z/:X0[k2_/:n`[k2_
/:n`0;2_/0060007[`001k07/00402d700<01`L040L1000000<01`L03`L00`071`0^1`8000=h0940
1I403P2AZY6ZTJZAZY40TJZA1P006jnA[`2_TJnA[i6_TJnA[i6_0:nZ[jZ_ZZnZ[`070008/:n`[k2_
0:l6030700<01`L040L1000000<01`L03`L00`071`0^1`0B07Uc096BTI:ATP2`TK2A/96`1P006i6`
TK2A/96`T@2Z/:Z`Z[2Z/:Z`Z[2Z0:n`[`06000:/:l0[kF_]Jne[`T0<PL00`071`0@1`4000000`07
1`0?1`0300L702l700@0N7<01Y400`2ZT@04000J[i6_0:nA[i6_TJnA[i6_T@2A[i6_TJnA[i47000:
[k2_/:n`[k00/0/000BY0000<PL00`071`0@1`4000000`071`0?1`0300L702l700L0LgT0TY6B00H0
01RA/96`TK2A096`TK2A/96`TK2A/02`Z[06000<[k2_0:n`[k2_/:n`1@03]@@000>YZZT00P001<FY
a@0a1`0300L701070@00000300L700l700<01`L0<0L00`1cN003000HTY6BT@2A[i6_TJnA[i40TJnA
[i6_TJnA1P003;2A/96`TK2A/02`[`H000>e0;D01P001JVZ002Y008000JZ0:T0a@0`1`0300L70107
0@00000300L700l700<01`L0<0L0601i0980Ui6GTIL0/96`TK2A/96`0;2A/0H000b_/02`[k2_/:n`
[k060004/;F`]@D000VZZ@00ZZVZZJX00P002:W6ZLH0aJT0<0L00`071`0@1`4000000`071`0?1`03
00L703470100NGQi09:ATY6B0:nA[i6_1`003;2A/96`TK2A/02`T@H000F`Uk00]@070003ZH^Y0080
00F@ZP2ZZ@020004ZZT0Z@8000FZ0<FY000_1`0300L701070@00000300L700l700<01`L0<PL02@2B
N02AUi6GT@06000>/9J`U[00/:n`[k2_/:l50008/;F`]K2e/;D30003ZP00008000>TZP000P001:ZY
ZZT30004ZJnY[`8000FYaZT0Z@0202l700<01`L040L1000000<01`L03`L00`071`0c1`0607Uh096B
0P00496G0;2A/96`TK2A/02`TK040007U[2F09N`U`0;0003RjT0008000BZ0:Z@0`001JVZ0:ZY0080
00VZZ@2YZZT0Z@00;`L00`071`0@1`4000000`071`0?1`0300L703@700<0TP000P001K2B/9:`0080
00BF/9J`1`002;2G/9N`Uk2G1@001:X00:l70003ZZT000@000FY[jV_Z@030006ZZnZ[`360`000lK5
000^1`0300L701070@00000300L700l700<01`L0=@L02P2GN02AUi6GTIL50006/9J`U[2F20001jnZ
0:Z_0:l01P000h^Z00030006T:Z@0:VZ10001jT0ZJZYZZT00P001ZnY0:W502h700<01`L040L10000
00<01`L03`L00`071`0f1`0309Mi008000N`TP00/9N`008000>G/9L01`001jl00;2_0:l01P000jX0
ZP030003ZP2Z00:Z1@002:nZ[jZ_ZP2Z0`001JX0Z/JZ00<0;PL00`071`0@1`4000000`071`0?1`03
00L703L70P0017P00007U`T000R_/02_ZP2Z[`T000>ZT:X01@002ZZ@ZP2ZZJZYZZT3000;ZJnY[`2_
ZJl0a@00;PL00`071`0@1`4000000`071`0?1`0300L703P70P04U`d000D7[`00/0020003[`0000@0
00>ZZJX01@002jl0[jZ_ZZnZ[`2_00@000RZ0:Z_ZZnZ[`@000?500L0;@L00`071`0@1`4000000`07
1`0?1`0300L703X72003U`050;D009H00`000jn`00050003R`2@00D000Z@ZY2ZT:X0ZY2Z1@002JZY
ZZVZZ@2Y[`040005ZJl0a@00;PL00`071`0@1`4000000`071`0?1`0300L703`710000iNe00030003
/;D0008000>`[`001@000jX0000200>Z00@0ZZnZ20002ZnZ[`2_ZZnZ[jX40007[jZ_0<J_aP0202h7
00<01`L040L1000000<01`L03`L00`071`0o1`T000>`UP001@002Y00002Z0:Z@ZY05000>T:Z@ZP2Z
ZJZYZZVZ0:X40009Z@2Y[jV_Z@2Y00@0;PL00`071`0@1`4000000`071`0?1`0300L704872003/0D0
0jX00`2Z0002000=ZP2Z[jZ_ZZnZ0:Z_ZP06000:[jZ_ZZl0[jZ_ZP<000>_aP00;`L00`071`0@1`40
00000`071`0?1`0300L704H72@0018/0T:X40008T:Z@ZP2ZT:X8000=ZJZYZP2ZZJZYZZT0Z@030004
ZP2ZaP80;`L00`071`0@1`4000000`071`0?1`0300L704/700<0R`001@02ZP030:Z_00L0012_ZZnZ
0:Z_ZZnZ[jX0ZZnZ1@001ZZ_ZZl0aP80<0L00`071`0@1`4000000`071`0?1`0300L704/70`001:Z@
0906000?ZY2Z0:Z@ZY2ZT:X0ZZVZ00P000RZ[jX0ZZnZ[`<000?600L0;`L00`071`0@1`4000000`07
1`0?1`0300L704X700<0ZP001@003JZ_ZP2Z[jZ_ZP2Z[jX020003JnZ[jX0ZZnZ[jX0ZZl010000jo6
000a1`0300L701070@00000300L700l700<01`L0BPL3000;T:X0ZY2ZT:X0ZY001`001I6ZTJX000JZ
00@0ZZnZ1P001jnZ0:o6[`00<@L00`071`0@1`4000000`071`0?1dd000RA096ZT@2Z[`H000nZ[jX0
ZZnZ[jZ_ZP2Z[jX01`002:Z_ZP2_aZo6=`0B1`4000000`071`0>1`0300L704X700@0T@2Z1@003JZA
ZP2ZTJZAZP2ZTJX020003:Z_ZP2Z[jZ_ZP2Z[`H0=0L00`071`0?1`4000000`071`0=1`0300L704/7
10002jnA0:Z_ZZnZ0:Z_00L000j_ZZnZ0:Z_ZZnZ[`2_ZPH000C700L0=`L00`071`0>1`4000000`07
1`0<1`0300L704`700D0T@2ZT@0:000>ZY6Z0:ZAZY6ZT@2Z[jX70007ZZl0[jZ_ZP0203T700<01`L0
3@L1000000<01`L02`L00`071`1=1`D0016A0:Z_ZZnZ0:Z_ZZnZ0:Z_ZP070005ZZnZ0:X02@0l1`03
00L700`70@00000300L700/700<01`L0CPL02@2A096ZT@2AZP0<0006TJnA0:Z_1P001:Z_0;0303l7
00<01`L02`L1000000<01`L02PL00`071`1?1`T000b_ZP2Z[jZ_ZP2Z[jX60005/:n`0;001@131`03
00L700X70@00000300L700T700<01`L0D@L01@2A0:ZA00l000FA[i40ZP0804L700<01`L02@L10000
00<01`L020L00`071`1C1`@000BA0:Z`501<1`0300L700P70@00000300L700L700<01`L0E@L70003
/96`00T0E@L00`071`071`4000000`071`061`0300L705P7301J1`0300L700L70@00000300L700D7
00<01`L0`0L00`071`061`4000000`071`041`0300L70<8700<01`L01@L1000000<01`L00`L00`07
1`341`0300L700@70@00000300L7008700<01`L0aPL00`071`031`4000000`071`021`0300L70<L7
00<01`L00PL1000000D01`L7003;1`0400L71`40000010071`3=1`0300L7004000000`07003?1`40
0@L100000P3A1`80003E0000\
\>"],
  ImageRangeCache->{{{0, 212.312}, {286.875, 0}} -> {0.134279, -0.000557988, 
  0.00344511, 0.00344511}},
  CellTags->"POVRay"]
}, Open  ]],

Cell[BoxData[
    \(MVClear[]; MVShow3D[grMaeder2, MVGrayBackground -> True]; \)], "Input",
  CellLabel->"In[36]:=",
  CellTags->"POVRay"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(MVPasteGraphics[]; \)\)], "Input",
  CellLabel->"In[10]:=",
  CellTags->"POVRay"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF5F6060CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF5F605F5F60605F6060606F605F6060606F6F5F6060606F6F606F6F606F6F\
606F6F606F70606F6F6F6F70606F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
505F505F605F505F5F5F60605F605F5F60605F6060606F6F5F6060606F6F5F6060606F6F5F606F\
606F6F606F6F606F70606F6F606F70606F6F6F7070606F6FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F505F605F505F5F
5F60605F605F5F60605F6060606F6F5F6060606F6F5F6060606F6F606F6F606F6F606F6F606F70\
606F6F6F7070606F6F6F7070606F706F7070606F706F70706F70706F707FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0505F50505F50505F5F505F505F605F
50605F5F60605F60605F60605F6060606F6F5F6060606F6F5F606F606F6F5F606F606F6F606F6F\
606F70606F6F606F70606F6F6F7070606F6F6F7070606F706F7070606F706F7070
6F70706F707FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF505F50505F50505F5F505F5F5F60605F605F
5F60605F6060606F6F5F6060606F6F5F6060606F6F606F6F606F6F606F6F606F70606F6F606F70\
606F6F6F7070606F706F7070606F706F70706F6F706F70706F6F706F707F6F7070
6F707F6F707070707F6F707FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF4F504F505F50505F50505F5F505F505F605F505F5F5F6060
5F605F5F60605F6060606F6F5F6060606F6F5F6060606F6F5F606F606F6F606F6F606F70606F6F\
606F70606F6F6F6F70606F6F606F70606F706F7070606F706F70706F6F706F707F
6F6F706F707F6F70706F707F6F7070CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF4F504F505F50505F505F605F505F5F5F60605F605F5F60605F6060
606F6F5F6060606F6F5F6060606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F6F7070\
606F706F7070606F706F7070606F706F7070606F706F707F6F70706F707F6F7070
70707F6F707F70707F6F707F70707F6F707FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF4F504F505F50505F50505F5F505F505F605F505F5F5F60605F60605F6060
5F6060606F6F5F6060606F6F5F606F606F6F5F606F606F6F606F6F606F70606F6F606F70606F70\
6F7070606F706F6F70606F706F7070606F706F7070606F706F707F6F6F706F707F
6F70706F707F6F70706F707F6F707F70707F6F707F70707FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
4F504F505F50505F505F605F505F5F5F605F5F605F5F60605F6060606F6F5F6060
606F6F5F6060606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F6F7070606F706F7070\
606F706F7070606F706F7070606F706F707F6F6F706F707F6F6F706F707F6F707F
6F707F6F707F70707F6F707F70707F6F707F707F7F70707F707F7FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF4F4F40\
505F50505F50505F5F505F505F605F505F5F5F60605F605F5F60605F6060606F6F
5F6060606F6F5F606F606F6F5F606F606F6F606F6F606F70606F6F606F70606F6F606F70606F6F\
606F70606F706F7070606F706F7070606F706F707F606F706F707F6F6F706F707F
6F6F706F707F6F707F6F707F6F6F706F707F6F707F70707F6F707F70707F70707FCFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F303F4F5040505F50\
505F505F605F505F5F5F60605F605F5F60605F6060606F6F5F6060606F6F5F6060
606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F6F7070606F706F7070606F706F7070\
606F706F7070606F706F707F6F6F706F707F6F6F706F707F6F707F6F707F6F707F
70707F6F707F70707F6F707F70707F6F707F70707F6F707F707F8070707F707F8070707FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF4F4F40505F50505F50\
5F5F5F505F505F605F505F5F5F60605F605F5F60605F6060606F6F5F6060606F6F
5F606F606F6F5F606F606F6F606F6F606F70606F6F606F70606F70606F70606F70606F70606F70\
6F7070606F706F7070606F706F707F606F706F707F6F6F706F707F6F6F706F707F
6F707F6F707F6F6F7F6F707F6F707F70707F6F707F70707F6F707F70708070707F70708070707F\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F5040505F50505F505F605F\
505F5F5F60605F605F5F60605F6060606F6F5F6060606F6F5F6060606F6F5F6F6F
606F6F5F6F6F606F70606F6F606F70606F6F607070606F70606F70606F706F7070606F706F7070\
606F706F707F606F706F707F606F706F707F6F707F6F707F6F707F6F707F6F707F
6F707F6F707F70707F6F707F70707F6F707F70708070707F70708070707F707F8070707F707F80\
70707FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3030304F504F505F50505F5F505F50\
5F605F505F5F5F60605F605F5F60605F6060606F6F5F60605F6F6F5F6060606F6F
5F606F606F6F5F606F606F705F6F6F606F70606F6F606F70606F6F606F70606F70606F70606F70\
606F70606F706F707F606F706F707F606F706F707F606F706F707F6F6F7F6F707F
6F6F706F707F6F707F6F707F6F6F7F6F707F6F707F70707F6F707F70707F70707F70708070707F\
70708070707FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303030505F50505F505F605F505F5F5F6060\
5F605F5F60605F6060606F605F6060606F6F5F6060606F6F5F6F6F606F6F5F606F
606F70606F6F606F70606F6F606F70606F70606F70606F706F7070606F70607070606F706F707F\
606F706F707F606F706F707F6F6F7F6F707F6F6F7F6F707F6F707F6F707F6F707F
70707F6F707F70707F6F707F7070806F707F7070806F707F707F8070707F707F8070707F707F80\
70707F7F7F8070707FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF30303050504F505F505F605F505F505F605F505F5F\
5F60605F605F5F60605F6060606F6F5F6060606F6F5F6060606F6F5F6060606F6F
5F606F606F6F5F606F606F70606F6F606F70606F6F606F70606F70606F70606F70606F70606F70\
6F6F70606F706F6F70606F706F707F606F706F707F606F706F707F6F6F706F707F
6F707F6F707F6F6F7F6F707F6F707F70707F6F707F70707F6F707F7070806F707F70708070707F\
707F8070707F707F80CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030303F3F3F505F505F605F505F5F5F605F5F605F5F6060\
5F605F5F60605F6060606F605F6060606F6F5F606F606F6F5F606F606F6F606F6F
606F6F606F6F606F70606F6F606F70606F70607070606F70606F70606F706F7070606F70606F70\
606F706F707F606F706F707F606F706F707F6F6F7F6F707F606F7F6F707F6F707F
6F707F6F707F70707F6F707F70707F6F707F7070806F707F70708070707F707F8070707F707080\
70707F7F7F8070707F7F7F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F303F50504F505F5F505F505F605F505F5F5F605F505F5F\
5F60605F60605F60605F60605F6F6F5F6060606F6F5F6060606F6F5F606F606F6F
5F606F606F6F5F6F6F606F705F6F6F606F70606F70606F70606F6F606F70606F70606F70606F70\
606F70606F70606F7F606F70606F7F606F706F707F606F706F707F606F706F707F
606F706F707F6F6F7F6F707F6F6F7F6F707F6F707F70707F6F707F70707F6F707F7070806F707F\
70708070707F707F8070707F7F7F80C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF3F3F3F3F303F5F5F5F505F505F605F5F605F5F60605F605F5F6060\
5F6060606F605F6060606F6F5F6F6F606F6F5F6F6F606F6F606F6F606F6F606F6F
606F70606F6F606F70606F6F606F70606F70606F70606F706F7070606F70606F70606F706F707F\
606F706F707F606F706F707F606F7F6F707F606F7F6F707F6F6F7F6F707F6F6F7F
6F707F6F707F6F707F6F707F7070806F707F7070806F707F70708070707F70708070707F707F80\
70707F707F8070707F7F7F807070807F7F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C03F303F303030505F50505F505F605F505F5F5F605F505F5F5F60605F605F\
5F60605F60605F60605F6060606F6F5F6060606F6F5F606F606F6F5F606F606F6F
5F6F6F606F705F606F606F70606F6F606F70606F70606F70606F70606F70606F70606F70606F70\
606F70606F70606F70606F706F707F606F706F6F7F606F706F707F606F706F6F7F
606F7F6F707F606F7F6F707F6F6F7F6F707F6F6F7F70707F6F707F7070806F707F70708070707F\
70708070707F70708070707F7F7F8070707F7F7F80CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3F303F30303F505F50505F505F605F5F605F5F605F5F605F5F60605F60605F6060\
5F6060606F6F5F6060606F6F5F606F606F6F606F6F606F6F606F6F606F70606F6F
606F70606F6F606F70606F70606F70606F70607070606F70606F70606F7060707F606F70606F7F\
606F706F707F606F70606F7F606F706F707F606F7F6F707F606F706F707F606F7F
6F707F606F7F6F707F6F6F7F6F707F6F6F7F7070806F707F7070806F707F70708070707F707080\
70707F707F8070707F7F7F8070707F7F7F807F70807F7F80CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C03030303030303F303F505F50505F5F505F5F5F605F505F5F5F60605F605F5F60605F6060\
5F60605F6060606F6F5F6060606F6F5F606F606F6F5F606F606F6F5F6F6F606F70
5F606F606F70606F6F606F705F6F6F606F70606F70606F70606F70606F70606F70606F70606F70\
606F70606F70606F7F606F70606F7F606F70606F7F606F70606F7F606F706F6F7F
606F706F6F7F606F7F6F707F606F7F6F707F6F6F7F6F707F6F6F7F70707F6F707F7070806F707F\
70708070707F70708070707F70708070707F7F7F8070707FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
3F303F30303F3F303F30303F5F605F5F5F5F5F605F5F605F6060605F60606060605F6060606F6F\
5F6060606F6F5F606F606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F
606F70606F70606F70606F70607070606F70606F70606F7060707F606F70606F7F606F706F707F\
606F70606F7F606F706F707F606F706F6F7F606F706F707F606F7F6F707F606F7F
6F707F606F7F6F707F606F7F6F707F6F6F7F6F707F6F6F7F7070806F707F70708070707F707080\
70707F70708070707F7F7F807070807F7F807070807F7F807F7F80CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F3030
3030303F303F30303F505F50505F505F605F5F5F5F5F60605F605F5F60605F60605F60605F6060\
606F6F5F6060606F6F606F6F606F6F5F606F606F6F606F6F606F705F6F6F606F70
606F6F606F70606F6F606F70606F70606F705F6F70606F70606F70606F70606F70606F70606F70\
606F7F606F70606F70606F70606F7F606F70606F7F606F70606F7F606F70606F7F
606F706F6F7F606F706F6F7F606F7F6F707F606F7F6F707F6F6F7F6F707F6F6F7F70707F6F707F\
7070806F707F70708070707F70708070707F7F708070707F7F7F80C0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303030
3F303F30303F3F3F40505F505F605F5F5F5F6060605F605F6060605F6060606F605F6060606F6F\
5F6060606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F606F70606F6F
606F70606F6F606F70606F70606F70606F70606F70606F70606F70606F70606F70606F70606F70\
606F70606F7F606F70606F7F606F70606F7F606F70606F7F606F706F6F7F606F7F
6F6F7F606F7F6F707F606F7F6F6F7F606F7F6F707F6F6F7F6F707F6F6F7F7070806F707F707080\
70707F70708070707F70708070707F7F7F807070807F7F807070807F7F80CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3030303F303F
3030303F30403030405F605F505F505F605F5F605F5F60605F605F5F60605F6060606F6F5F6060\
606F6F5F6060606F6F5F606F606F6F5F606F606F6F5F606F606F6F5F6F6F606F70
5F606F606F705F6F6F606F705F606F606F705F6F70606F705F606F606F705F6070606F705F6070\
606F705F6F70606F705F6070606F705F6F70606F7F5F6070606F7F606F70606F7F
606F70606F7F606F70606F7F606F706F6F7F606F7F6F6F7F606F7F6F6F7F6F6F7F6F707F6F6F7F\
70707F6F707F7070806F707F70708070707F7F708070707F7F7080707080CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030303F303F30303F
3F3F403F30403F304F5F5F5F5F60605F605F6060605F6060606F6F606060606F6F606060606F6F\
606F6F606F6F606F6F606F70606F6F606F70606F6F606F70606F6F606F70606F6F
606F70606F70606F70606F70606F70606F70606F70606F70606F70606F70606F705F6F70606F7F\
606F70606F70606F70606F7F606F70606F7F606F70606F7F606F70606F7F606F70
606F7F606F7F6F6F7F606F7F6F6F7F606F7F6F6F7F606F7F6F707F6F6F7F6F707F6F6F7F707080\
6F707F70708070707F7070807070807F708070707F7F7F807F70807F7F807F7080
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F303F3030303F303F
3030403F304F505F505F605F5F605F6060605F605F6060605F6060606F6F5F6060606F6F5F6060\
606F6F5F6060606F6F606F6F606F6F5F606F606F6F606F6F606F705F606F606F70
5F6F6F606F705F606F606F705F6F70606F705F606F606F705F6F70606F705F6070606F705F6070\
606F705F6070606F705F6070606F705F6070606F705F6070606F705F6070606F7F
606070606F7F606070606F7F606F70606F7F606F706F6F7F606F7F6F6F7F606F7F6F6F7F6F6F7F\
6F707F6F6F7F70708070707F70708070707F70708070707F7F708070707F7F7F80
7F7080CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F303F3030303F3040303040
3F304F30304F5F605F5F605F6060605F6060606F60606060606F605F6060606F6F606F60606F6F\
606F6F606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F606F70606F6F
606F70606F6F606F70606F70606F70606F70606F70606F70606F705F6F70606F705F6F70606F70\
5F6070606F705F6F70606F705F6070606F7F5F6070606F705F6070606F7F5F6070
606F7F5F6070606F7F606F70606F7F606F706F6F7F606F7F6F6F7F606F7F6F6F7F6F6F7F6F6F7F\
6F6F7F7070806F6F7F70708070707F70708070707F7F708070707F7F7F807F7080
7F7F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F303F3030303F303F30303F3F3040
3030403F304F5F5F5F5F60605F605F6060605F6060606F605F6060606F605F6060606F6F5F6060\
606F6F60606F606F6F5F606F606F6F606F6F606F705F606F606F6F5F6F6F606F70
5F606F606F705F606F606F705F606F606F705F6070606F705F606F606F705F6070606F705F606F\
5F6F705F60705F6F705F60705F6F705F6070606F705F60705F60705F60705F6F70
5F60706060705F6070606F7F5F6070606F7F606070606F7F606070606F7F606F7F6F6F7F606F7F\
6F6F7F6F6F7F6F707F6F6F7F70707F70707F70708070707F70708070707F7F7080
7070807F7F80C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030303F3F403030403F304030304F
3F3F503F30506060605F605F606F60606060606F60606060606F6F606F60606F6F606060606F6F\
606F6F606F6F606F6F606F70606F6F606F70606F6F606F70606F6F606F70606F6F
606F70606F70606F70606F70606F70606F70606F705F6F70606F705F6F70606F705F6070606F70\
5F6070606F705F6070606F705F6070606F705F6070606F7F5F6070606F705F6070
606F7F5F6070606F7F5F6070606F7F606070606F7F6060706F6F7F606F7F6F6F7F606F7F6F6F7F\
6F6F7F6F707F6F6F7F7070806F6F7F70708070707F7F70807070807F7080707080
7F7F807F7080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3030303F303F30303F3F30403030403F304F
30304F5F605F5F605F6060605F6060606F605F6060606F605F6060606F6F5F6060606F6F606060\
606F6F5F606F606F6F606F6F606F70606F6F606F70606F6F606F705F6F6F606F70
606F6F606F705F606F606F705F6F6F606F705F606F606F705F6070606F705F606F5F6F705F6070\
606F705F60705F6F705F60705F6F705F60705F60705F60705F60705F60705F6070
5F60705F60705F60705F60705F6070606F7F5F607060607F5F6070606F7F606070606F7F606F7F\
6F6F7F606F7F6F6F7F6F6F7F70707F6F6F7F70708070707F70708070707F7F7080
7070807F7F807F7080CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF3030303F303F30303F3F30403030403F3F4F30304F
3F30505F605F606F605F6060606F60606060606F6F606060606F6F606060606F6F606F6F606F6F\
606F6F606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F606F70606F70
606F70606F6F606F70606F70606F705F6F70606F705F6F70606F705F6070606F705F6070606F70\
5F6070606F705F60705F6F705F60705F6F705F60705F60705F60705F6F705F6070
5F60705F60705F60705F60705F60705F6070606F7F5F6070606F7F5F6070606F7F606F7F6F6F7F\
606F7F6F6F7F6F6F7F6F6F7F6F6F7F707080706F7F70708070707F7F7080707080
7F70807F70807F7F8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C030303F30303F3F30403030403F304F30304F3F3050
3030505F605F5F605F6060605F605F606F60606060606F6F5F6060606F6F606060606F6F5F6060\
606F6F606F6F606F6F60606F606F6F606F6F606F70606F6F606F70606F6F606F70
5F606F606F705F6F6F606F705F606F606F705F606F606F705F606F5F6F705F606F5F6F705F606F\
5F60705F606F5F60705F606F5F60705F606F5F607050606F5F60705F60705F6070
50606F5F60705060705F6070505F705F60705F60705F60705F60705F60705F6070606F7F606070\
606F7F60607F6F6F7F60607F6F6F7F6F6F7F706F7F6F6F7F707080706F7F707080
70707F7F70807F7080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F303F30303F3F30403030403F3F4F3F304F3F3050303050
3F3F5F5F605F606F60606060606F60606060606F6F606F60606F6F606F6F606F6F606F6F606F6F\
606F6F606F6F606F6F6F6F70606F6F606F70606F6F6F6F70606F70606F70606F6F
606F70606F70606F70606F70606F70606F70606F705F6F70606F705F6070606F705F606F606F70\
5F60705F6F705F60705F6F705F60705F60705F60705F60705F60705F6070506070
5F60705F60705F60705F60705F60705F60705F60705F607060607F5F607060607F5F6070606F7F\
60607F606F7F606F7F6F6F7F6F6F7F6F6F7F6F6F7F70708070707F70708070707F
7F70807F70807F7F807F7080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C03F303F30303F3F30403030403F304F30304F3F30503030503F305F
5F605F6060605F605F606060606060606F60606060606F6F606060606F6F606060606F6F606F6F\
606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F606F70606F6F606F70
606F6F606F705F606F606F705F606F606F705F606F606F705F606F606F705F606F5F60705F606F\
5F607050606F5F607050606F5F607050606F5F607050606F5F6070505F6F506070
505F6F5F6070505F6F5F6070505F705F6070505F6F5F60705F5F705F60705F5F705F60705F6070\
60607F5F6070606F7F60607F6F6F7F60607F6F6F7F6F6F7F707080706F7F707080
70707F7F708070707F7F7080CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF30303F3F30403030403F304F30304F3F30503030503F3F5F3F305F
6060605F605F606F60606060606F60606060606F6F606F60606F6F606F6F6F6F6F606F6F606F6F\
606F6F6F6F70606F6F606F70606F6F6F6F70606F6F606F70606F6F606F70606F6F
606F70606F6F606F70606F70606F705F606F606F705F6F70606F705F606F606F705F60705F6F70\
5F606F5F60705F606F5F607050606F5F607050606F5F6070505F6F5F6070505F6F
506070505F6F5F6070505F6F5F6070505F6F5F6070505F705F60705F5F705F60705F60705F6070\
5F607060607F606070606F7F60607F6F6F7F6F6F7F6F6F7F6F6F7F70708070707F
7F708070707F7F70807F70807F7F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C03F303F30303F3F30403030403F304F30304F3F30503030503F3F5F
5F605F6060605F605F606F605F6060606F60606060606F6F606060606F6F606F6F606F6F606F6F\
606F6F606F6F606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F606F70
5F606F606F705F606F606F705F606F606F705F606F606F705F606F5F6F705F606F5F607050606F\
5F607050606F5F6070505F6F50606F505F6F5F6070505F6F50606F505F6F50606F
505F6F505F6F505F6F505F70505F6F505F6F505F6F505F70505F6F5F6070505F705F6070505F6F\
5F60705F60706060705F607060607F6060706F6F7F6060706F6F7F6F6F7F707080
706F7F70708070707F7F70807F7080CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF30303F3F3F403030403F304F30304F3F3F5F3F305F3F3F5F3F3F5F
606F6060605F606F60606060606F6F606F606F6F6F606F606F6F6F606F6F6F6F6F606F6F6F7070\
606F6F6F6F70606F6F6F7070606F6F6F6F70606F6F6F7070606F70606F70606F6F
606F70606F70606F705F6F6F606F705F6F70606F705F606F606F705F6070606F705F606F5F6F70\
5F60705F607050606F5F607050606F5F6070505F6F5F607050606F506070505F6F
506070505F6F505F6F505F6F506070505F6F505F70505F6F506070505F705F6070505F705F6070\
5F5F705F60705F607060607F6060706F6F7F60607F6F6F7F6F6F7F706F7F6F6F7F
70708070707F7F70807070807F7F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF30303F3F30403030403F304F30304F3F30503030503F3F5F3F305F3F3F60
5F605F6060605F605F606F60606060606F6F606060606F6F606F6F6F6F6F606F6F606F6F606F6F\
6F6F6F606F6F606F6F606F6F606F70606F6F606F70606F6F606F70606F6F606F70
606F6F606F705F606F606F705F606F606F705F606F5F6F705F606F5F6F705F606F5F60705F606F\
5F607050606F5F607050606F5F6070505F6F50606F505F6F50606F505F6F505F6F
4F5F6F505F6F4F5F6F505F6F4F5F6F505F6F4F506F505F6F4F5F6F505F6F505F6F505F70505F6F\
5F6070505F6F5F60705F5F706060705F607060607F6060706F6F7F6F607F6F6F7F
6F6F7F707080706F7F7F708070707F7F7080C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF3F30403030403F304F30304F3F3F503F30503F3F5F3F305F3F3F603F3F60
606F6060605F606F60606060606F60606F606F6F6F606F6F6F6F6F606F6F6F6F6F606F6F6F6F6F\
606F6F6F6F70606F6F6F6F70606F6F6F7070606F6F606F70606F6F6F7070606F70
606F70606F6F606F70606F70606F705F606F606F705F6070606F705F606F606F705F60705F6F70\
5F606F5F60705F606F5F607050606F5F607050606F506070505F6F506070505F6F
505F6F4F5F6F505F6F4F5F6F505F6F4F506F505F6F4F5F6F505F6F4F506F505F70505F6F505F70\
505F6F5F6070505F705F60705F5F7060607F6060706F6F7F60607F6F6F7F6F6F7F
706F7F706F7F7F708070707F7F70807F7080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF3030403F304F3030403F304F3030503F3F5F3030503F3F5F3F3F5F3F3F60
5F605F606060606060606F60606060606F60606F606F6F6F606F606F6F6F606F6F6F6F6F606F6F\
606F6F606F6F606F6F606F6F606F6F606F6F606F70606F6F606F70606F6F606F70
606F6F606F70606F6F606F705F606F606F705F606F606F705F606F5F6F705F606F5F6F705F606F\
5F606F50606F5F6070505F6F50606F505F6F50606F505F6F505F6F4F5F6F505F6F
4F50604F5F6F4F506F4F5F6F4F50604F5F6F4F50604F5F6F4F50604F506F4F506F505F6F4F506F\
505F6F505F6F505F70505F6F5F5F705F5F706060705F607060607F6060706F6F7F
6F607F706F7F706F7F707080706F7F7F7080C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
3030403F304F30304F3F3F503F30503F3F5F3F305F3F3F603F3F60403F603F3F60
606F606060606F6F60606F606F6F6F606F606F6F6F606F6F6F706F6F6F6F6F706F606F6F6F7070\
606F6F6F6F70606F6F6F7070606F6F6F6F70606F6F6F7070606F706F6F70606F70
6F7070606F70606F70606F70606F70606F70606F705F6F70606F705F6F70606F705F606F606F70\
5F606F5F607050606F5F607050606F5F606F505F6F506070505F6F50606F4F5F6F
505F6F4F5F6F4F5F6F4F50604F5F6F4F506F4F5F6F4F50604F5F6F4F506F4F5F6F4F506F505F6F\
4F506F505F6F505F6F5F5F705F5F705F60705F5F7060607F6060706F6F7F6F607F
706F7F6F6F7F707080706F7F7F70807F70807F7080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
3F30403030403F304F30304F3F3F503030503F3F5F3F3F5F3F3F603F3F6040406F
60605F606F60606060606F60606F606F6F6F606F606F6F6F606F6F6F6F6F606F6F6F6F6F606F6F\
6F6F6F606F6F6F6F6F606F6F6F6F70606F6F606F70606F6F606F70606F6F606F70
606F6F606F70606F6F606F70606F6F606F705F606F606F705F606F606F705F606F5F60705F606F\
5F607050606F5F606F50606F5F606F505F6F50606F505F6F505F6F4F5F60505F6F
4F5F604F5F6F4F50604F506F4050604F506F4050604F50604050604F506F4050604F506F4F5060\
4F506F4F50604F506F4F506F505F70505F6F5F5F705F5F706060706060706F607F
6F60706F6F7F6F6F7F706F7F706F7F7F7080706F7FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
3030403F304F3F304F3F3F503F30503F3F5F3F3F5F3F3F603F3F6040406F3F406F
6F6F606060606F6F6F606F606F6F6F606F606F6F6F606F6F6F6F6F606F6F6F706F6F6F6F6F7070\
606F6F6F7070606F6F6F7070606F6F6F7070606F6F6F6F70606F6F6F7070606F70
606F70606F706F6F70606F70606F70606F6F606F70606F70606F705F606F606F705F606F5F6F70\
5F606F5F6F705F606F5F607050606F5F6070505F6F50606F4F5F6F505F6F4F5F6F
505F6F4F5F604F5F6F4F50604F506F4050604F506F4050604F50604050604F506F4050604F506F\
4050604F506F4F506F505F6F50506F5F5F70505F705F5F705F5F7060607F606070
6F6F7F6F607F706F7F706F7F707080706F7F7F7080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF303040\
30304F30304F3F3F503030503F3F5F3F3F5F3F3F603F3F5F3F3F603F3F6F404070
60605F606F606060606F6F6F6060606F6F6F606F606F6F6F606F606F6F6F606F6F6F6F6F606F6F\
6F6F6F606F6F6F6F70606F6F6F6F70606F6F6F6F70606F6F606F70606F6F6F6F70
606F6F606F70606F6F606F70606F6F606F705F606F606F705F606F5F6F705F606F5F6F705F606F\
5F606F5F606F5F607050606F5F606F505F6F50606F4F5F60505F6F4F5F60505F6F
4F50604F5F6F4F50604F5F6F405060405060404F60405060404F5F405060404F60405060404F60\
405060404F604F506F404F604F506F4F506F505F7050506F5F5F705F5F70606070
6060706F607F6F6070706F7F706F7F706F7F706F7FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F4F\
3F304F3F3F503F30503F3F5F3F3F5F3F3F603F3F6040406F3F406F40407040406F
6F6F60606F606F6F6F606F606F6F6F6F6F6F6F6F6F6F6F6F6F706F6F6F6F6F706F6F6F6F6F7070\
6F6F6F6F70706F6F6F6F70706F6F706F7070606F6F6F7070606F706F7070606F70
6F7070606F706F7070606F706F7070606F70606F70606F70606F705F6F70606F705F606F606F70\
5F606F5F6F705F606F5F6F705F606F5F607050606F5F606F505F6F50606F4F5F6F
505F6F4F5F604F5F6F4F50604F5F6F4050604F5060404F60405060404F604050603F4F60405060\
404F60405060404F604F506F4F50604F506F50506F5F5F70505F705F5F705F5F70
60607F6060706F6F7F6F6F7F707080706F7F7F7080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF303040\
3F3F4F30304F3F3F503F3F503F3F5F3F3F5F3F3F603F3F6040406F3F406F404070
6060606F6F606060606F6F60606F606F6F6F606F606F6F6F606F6F6F6F6F606F6F6F6F6F6F6F6F\
6F7070606F6F6F70706F6F6F6F7070606F6F6F6F70606F6F6F7070606F6F6F6F70
606F6F6F6F70606F6F606F70606F70606F70606F6F606F705F6F6F606F705F606F5F6F705F606F\
5F6F705F606F5F606F50606F5F606F50606F5F606F505F6F50606F4F5F60505F6F
4F5F604F5F6F4F50604F5F604050604F50604050604050603F4F5F404F603F4F5F3F4F603F4F5F\
404F603F4F5F404F60404F60405060404F604F506F4F506F505F7050506F5F5F70
5F5F7060607F6060706F6F7F6F6F7F706F7F706F7F706F7FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030403F304F\
30304F3F3F503F3F503F3F5F3F3F5F4040603F3F6040406F3F406F404070404070
6F6F60606F606F6F6F6F6F606F6F6F606F606F6F6F6F6F6F6F6F6F6F6F6F6F706F6F6F6F6F7070\
6F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F706F7070606F706F7070606F70
6F7070606F706F7070606F706F6F70606F706F6F70606F70606F70606F6F606F705F6F70606F70\
5F606F606F705F606F5F60705F606F5F607050606F5F606F505F6F50606F505F6F
505F6F4F5F60505F6F4F50604F5F604050604F50604050604050603F4F5F4050603F4F5F3F4F5F\
3F405F3F4F603F4F5F404F60404F604F506F404F604F506F4F506F5F5F705F5F6F
5F5F70605F706F607F6F6070706F7F706F7F706F7F706F7FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F304F30304F\
3F30503030503F3F5F3F305F3F3F603F3F6040406F3F3F6F40407040407040407F
6060606F6F60606F606F6F6F606F606F6F6F606F606F6F6F606F606F6F6F6F6F6F6F6F6F606F6F\
6F6F6F6F6F6F6F7070606F6F6F70706F6F6F6F7070606F6F6F6F70606F6F6F7070
606F6F6F6F70606F6F6F6F70606F6F606F70606F6F606F70606F6F606F705F6F6F606F705F606F\
5F6F705F606F5F6F705F606F5F606F50606F5F606F505F6F50606F505F6F505F6F
4F5F604F5F6F4F50604F5F604050604F50604050604050603F4F5F404F5F3F4F5F3F4F5F30405F\
3F4F5F30405F3F405F30405F3F4F5F3F4F5F404F60404F604F506F4F506F505F6F
50506F5F5F70605F706F607F6F6070706F7F6F6070706F7FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F3F3F50\
3F30503F3F5F3F3F5F403F603F3F6040406F3F3F6F40407040407040407F404070
6F6F6F6F6F606F6F6F6F6F606F706F6F6F6F6F6F6F6F6F6F6F706F6F6F6F6F706F6F6F6F6F7070\
6F6F6F6F70706F6F6F6F70706F70706F70706F6F706F70706F6F706F7070606F70
6F70706F6F706F7070606F706F7070606F706F6F70606F706F7070606F70606F70606F70606F70\
606F70606F705F606F606F705F606F5F6F705F606F5F607050606F5F606F505F6F
50606F4F5F6F505F6F4F5F604F5F6F4F50604F5060405060405060404F5F404F5F3F4F5F3F4F5F\
3F405F3F4F5F30405F3F405F3F405F3F4F5F3F4F604F4F60404F604F506F50506F
5F5F705F5F706060706060706F6F7F6F6F7F706F7F6F6070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F304F30304F\
3F3F503030503F3F5F3F3F5F3F3F603F3F6040406F3F3F6F404070404070404F7F
404F7F6F6F60606F606F6F6F6F6F606F6F6F6F6F606F6F6F6F6F6F6F706F6F6F6F6F6F6F6F6F6F\
6F70706F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F706F7070606F6F6F7070
606F706F7070606F6F6F6F70606F706F6F70606F6F606F70606F70606F70606F6F606F70606F70\
606F705F606F606F705F606F5F6F705F606F5F606F50606F5F606F505F6F50606F
505F6F50606F4F5F604F5F6F4F50604F5F6040505F40506040505F4050603F4F5F3F4F5F3F4050\
3F4F5F3040503F405030405030405F303F503F405F3F405F404F60404F604F506F
50506F5F5F705F5F6F6060706060706F607F6F60706F607FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F3F3F50\
3030503F3F5F3F3F5F3F3F603F3F6040406F3F3F6F404070404070404F7F404F7F
4F4F7F6F6F606F6F6F6F6F606F6F6F6F6F606F706F6F6F6F6F706F6F6F6F6F70706F6F6F6F7070\
6F6F6F7070706F70706F70706F6F6F7070706F70706F70706F6F706F70706F6F70
6F70706F6F706F7070606F706F7070606F706F7070606F706F6F70606F706F6F70606F70606F70\
606F70606F705F6F70606F705F606F606F705F606F5F607050606F5F607050606F
5F606F505F6F50606F505F6F505F6F4F50604F5F6040506040506040505F4050603F4F5F3F4F5F\
3040503F405F3040503F4050303F5030405F303F503F405F3F405F404F60404F60
50506F50506F5F5F70605F706060706060706F607F6F6070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF30303F3F3050303050\
3F305F30305F3F3F603F3F603F3F6F3F3F6F4040703F3F6F404070404070404F7F
404F7F6F6F606F6F606F6F6F6F6F606F6F6F6F6F606F6F6F6F6F6F6F6F6F6F6F6F6F706F6F6F6F\
6F70706F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F6F6F70706F6F706F7070
606F6F6F7070606F706F7070606F6F6F6F70606F706F6F70606F6F606F70606F70606F70606F6F\
606F70606F6F606F705F606F5F6F705F606F5F607050606F5F606F50606F5F606F
505F6F50606F505F6F50606F4F5F604F5F6F4F50604F5F6040505F405060404F5F40505F3F4F5F\
3F4F5F3040503040502F3F4F303F50303F4F303F50303F503F405F3F405F404F60
404F6050506F50506F5F5F705F5F70606070605F706F6070605F70CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F30403030503F3F5F\
3F305F3F3F603F3F60403F6F3F3F6F4040703F4070404F7F404070404F7F404F7F
4F50806F6F606F6F6F6F6F6070706F6F6F6F70706F6F6F6F7070706F6F6F7070706F6F6F707070\
6F706F7070706F6F6F7070706F70707070706F70707070706F70706F70706F7070
7070706F70706F70706F6F706F70706F6F706F7070606F706F7070606F706F7070606F706F7070\
606F70606F70606F70606F70606F70606F705F606F606F705F606F5F60705F606F
5F607050606F5F606F505F6F50606F505F6F505F6F4F5F604F5F6F4F50604F506040505F405060\
3F4F5F3F4F5F3040503F405F303F50304050303F4F303F50303F503F405F3F405F
4F4F604F4F6050506F50506F606070605F70606070605F706F6070CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3030403F305030305F\
3F3F6030305F3F3F603F3F603F3F6F3F3F6F404070404070404F7F40407F4F4F7F
404F7F6F6F606F6F606F6F6F6F6F606F706F6F6F606F706F6F6F6F70706F6F6F6F6F706F6F6F6F\
7070706F6F6F6F70706F6F6F7070706F6F6F6F70706F6F706F70706F6F706F7070
6F6F706F70706F6F706F70706F6F706F7070606F706F7070606F706F7070606F70606F70606F70\
606F70606F70606F70606F70606F705F606F606F705F606F5F6F705F606F5F606F
50606F5F606F505F6F50606F505F6F50606F4F5F60505F6F4F5F604F5F6040506040506040505F\
4050603F4F5F3F4F5F3F40503F40502F3F4F2F3F4F30304F303F50303F503F405F
3F405F4F4F604F4F6050506F5F506F605F705F5F6F605F70605F6FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F304F30305F3F3F5F\
30305F3F3F603F3F603F3F6F3F3F6F404070404070404F7F40407F4F4F7F404F7F
4F50806F6F606F6F6F6F6F606F6F6F6F6F6070706F6F6F6F70706F6F6F6F7070706F706F707070\
6F6F6F7070706F70707070706F6F6F7070706F70707070706F6F707070706F7070
7070706F70707070706F70706F70706F6F706F70706F6F706F7070606F706F7070606F706F7070\
606F706F7070606F70606F70606F70606F70606F70606F705F606F606F705F606F
5F6F705F606F5F607050606F5F606F505F6F50606F505F6F505F6F4F5F60505F6F4F50604F5060\
40505F4050603F4F5F3F4F5F3F4F503F4050303F4F303F4F2F3040303F50303F50
3F405F40405F4F506F50506F5F5F6F5F506F605F70605F6F605F70CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C030303F30304F3F305F30305F\
3F3F603F30603F3F6F3F3F6F3F406F3F406F404070404070404F7F404F7F4F5080
404F7F4F50806F6F606F6F6F6F6F606F6F6F6F6F6070706F6F6F6F70706F6F6F6F7070706F6F6F\
7070706F6F6F7070706F6F6F6F70706F6F6F7070706F6F6F6F70706F6F706F7070
6F6F706F70706F6F706F70706F6F706F70706F6F706F7070606F706F7070606F706F7070606F70\
606F70606F70606F70606F6F606F70606F70606F705F606F606F705F606F606F70
5F606F5F60705F606F5F606F505F6F50606F505F6F50606F4F5F60505F6F4F5F604F5F6F4F5060\
4F506040505F40505F3F4F503F4F5F3040503F4F50303F4F303F4F2F2F40303F50
3030503F405F40405F50506F504F605F506F5F506F605F6F5F506FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30303F3F305030305F3F3F60\
3F30603F3F6F3F3F6F4040703F40704040704040704F4F7F404F7F4F5080404F80
4F508F4F50806F6F6F6F6F6070706F6F6F6F70706F6F6F6F70707070706F7070706F706F707070\
70706F7070706F706F7070706F70707070706F70707070706F70707070706F7070
7070706F70707070706F707070707F6F707070707F6F70706F707F6F70706F70706F6F706F707F\
6F6F706F7070606F706F7070606F70606F70606F70606F70606F70606F70606F70
606F705F6F70606F705F606F5F6F705F606F5F606F50606F5F606F505F6F50606F4F5F60505F6F\
4F5F604F5F604050604F506040505F40505F3F4F503F4F5F304050303F4F2F2F40
303F503F3F50404060404F6050506F50506F5F506F5F506F605F70CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F304030304F3F3060303060\
3F3F6F3F30603F3F6F3F3F6F4040703F4070404F7F40407F4F4F7F404F7F4F5080
4F50804F508F6F6F606F6F606F6F6070706F6F6F6070706F6F6F6F70706F6F6F6F7070706F6F6F\
7070706F6F6F7070706F6F6F7070706F6F6F7070706F70707070706F6F70707070
6F70707070706F6F706F70706F70707070706F6F706F70706F6F706F70706F6F706F70706F6F70\
6F7070606F706F6F70606F706F6F70606F70606F70606F70606F7060606F606F70
5F6F70606F705F606F5F6F705F606F5F607050606F5F606F505F6F50606F505F60505F6F4F5F60\
505F6F4F50604F5F604F50604F5060404F5F40505F3F4F5F3F4F5F304050304050
2F3040303050303F504040604F4F6050506F504F605F506F5F506FCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30303F3F30503030603F3F60\
3F3F603F3F6F3F3F6F4040703F4070404F7F404070404F7F404F7F4F50804F5080
4F508F4F508F706F6F6F6F6070706F6F6F6070706F6F6F6F70706F6F6F6F70707070706F707070\
70706F7070707070707070706F70707070707070707070706F70707070706F7070
7070706F707070707F6F707070707F6F707070707F6F707070707F6F707070707F6F70706F7070\
6F6F706F707F6F6F706F7070606F706F7070606F706F6F70606F706F6F70606F70
606F70606F70606F70606F70606F705F606F606F705F606F5F606F50606F5F606F505F6F50606F\
4F5F60505F6F4F5F604F5F604F50604F5F604050604050603F4F5F404F5F3F4F50
3F4F502F30403F40503F405F4F4F604F4F6050506F5050605F506FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03030403030503F3060303060\
3F3F6F3F3F6F3F406F3F3F6F404070404070404F7F404F7F404F7F404F7F4F5080
4F50804F5F8F6F6F606F6F6F6F6F60706F6F6F6F6070706F6F6F6F70706F6F6F6F7070706F6F6F\
7070706F706F7070706F6F6F7070706F706F7070706F6F6F7070706F7070707070
6F6F707070706F70707070706F6F707070706F70707070706F6F706F70706F6F706F70706F6F70\
6F70706F6F706F7070606F706F6F70606F706F6F70606F70606F70606F70606F70
606F70606F70606F70606F705F606F606F705F606F606F705F606F5F606F50606F5F606F505F6F\
50606F505F60505F6F4F5F604F5F604F50604F5F6040505F405060404F5F404F5F
3F40503F4F502F2F403F3F5F3F405F4F4F604F4F60504F60504F5FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030403F305F3030603F3F6F\
3F3F6F3F406F3F3F6F404070404070404F7F404F7F4F50804F4F804F50804F5080
505F8F4F5F8F505F906F6F6070706F706F6F70706F706F6F70707070706F70707070706F707070\
70706F707070707070707070707070707070707070707070707070707070707070
70707F70707070707F70707070707F70707070707F6F707070707F6F707070707F6F707070707F\
6F707070707F6F70706F707F6F70706F707F6F6F706F707F6F6F706F7070606F70
6F707F606F70606F70606F70606F70606F70606F705F6070606F705F6070606F705F606F5F6F70\
5F606F5F606F505F6F50606F505F6F505F6F4F5F60505F6F4F5F604F5F60405060
4F5060404F5F303040303F503F405F40405F4F4F604F4F60504F60CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F30403030503F3F60303F60\
3F3F6F3F3F6F3F40703F4070404F7F404070404F7F404F7F4F5080404F804F508F
4F508F505F8F4F5F8F706F606F6F6070706F6F6F6070706F6F6F6F70706F6F6F6F70706F706F6F\
7070706F6F6F70707070706F70707070706F7070707070707070706F7070707070
7070707070706F70707070706F70707070706F70707070706F707070707F6F6F707070706F7070\
70707F6F6F706F70706F6F706F707F6F6F706F70706F6F706F7070606F706F6F70
606F706F6F70606F70606F70606F70606F705F606F606F705F606F606F705F606F5F6F705F606F\
5F607050606F5F606F505F6F50606F505F60505F6F4F5F60505F6F4F50604F5F60
4050604F506040505F30304F303F5040405F40405F4F4F604F4050CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030403F3F5F3030603F3F6F\
3F3F6F3F40703F40704040703F4070404F7F404F7F4F5080404F804F508F4F508F
4F5F8F4F5F8F5060906F6F6070706F6F6F6070706F706F6F70706F706F6F70707070706F707070\
70706F707070707070707070707070707070707070707070707070707070707070
70707070707070707F70707070707F70707070707F70707070707F6F707070707F6F707070707F\
6F707070707F6F707070707F6F707070707F6F70706F707F6F6F706F707F6F6F70
6F7070606F706F707F606F70606F70606F70606F70606F70606F705F6070606F705F6070606F70\
5F606F606F705F606F5F607050606F5F606F505F6F50606F505F60505F6F4F5F60
4F5F604F50604F5F60405060303F503F3F504F405F40405F4F40505F3F30403030CFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03030403030503F3F60303F60\
3F3F6F3F3F6F3F40703F4070404F7F404F7F404F7F404F7F4F50804050804F5F8F
4F508F4F5F8F4F5F8F706F6F6F6F60706F6F6F6F6F70706F6F6F6F70706F706F6F70706F6F6F6F\
70706F706F6F707070706F6F70707070706F70707070706F707070707070707070
7070707070707070707070706F70707070706F70707070706F70707070706F707070707F6F6F70\
7070706F707070707F6F6F706F70706F6F706F707F6F6F706F70706F6F706F7070
606F706F6F70606F706F6F70606F70606F70606F70606F7060606F606F705F606F606F705F606F\
5F6F705F606F5F6F705F606F5F606F5F606F5F606F505F6F50606F505F60505F6F
4F5F604F5F604F50604F5F602F3040303F503F3F5040405F403F4F6F403F6F3F30603F30403030\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030403F3F5F303F5F3F406F\
3F3F6F3F40703F4070404F7F404F7F404F7F404F7F4F50804050804F5F8F4F508F
505F904F5F8F506090505F9070706F706F6F70706F706F6F70706F70706F70706F70706F707070\
70706F70707070706F7070707070707070707070707F7F70707070707070707070
7F7F7F70707070707F707070707F7F70707070707F70707070707F70707070707F70707070707F\
70707070707F6F707070707F6F707070707F6F707070707F6F707070707F6F7070
70707F6F70706F707F6F6F706F707F6F6F706F7070606F706F707F606F70606F70606F70606F70\
606F70606F705F6070606F705F6070606F705F606F5F6F705F606F5F607050606F
5F606F505F6F50606F4F5F60505F6F30304F403F503F3F5040404F704F40704F406F403F70403F\
5F3F30503F3F4F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F304F3030503F3F60303F6F\
3F40703F3F6F3F40703F4070404F7F3F4F7F405080404F804F508F4050804F5F8F
4F5F8F505F904F5F8F506090706F6070706F6F6F6070706F706F6F70706F706F6F70706F706F6F\
707070706F6F70707070706F70707070706F707070707070707070707070707070
70707070707070707070707070707070707F70707070707070707070707F6F70707070706F7070\
70707F6F707070707F6F707070707F6F707070707F6F707070707F6F6F706F707F
6F6F706F707F6F6F706F707F6F6F706F707F606F706F6F70606F706F6F70606F70606F70606F70\
606F705F6070606F705F6070606F705F606F5F6F705F606F5F60705F606F5F606F
50606F5F606F505F6F50606F505F603030403F304F3F3F503F3040805F4F7F50407F5040704F3F\
704F3F6F403060403F503F30503F304F2F30CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030403F3F50303F5F3F406F\
3F3F6F3F40703F4070404F7F3F4F7F405080404F804050804050804F5F8F4F5F8F
4F5F904F5F8F50609050609050609F706F6070706F706F6F70706F706F6F70707070706F707070\
70706F70707070706F70707070706F7F70707070707070707070707F7F70707070
7F70707070707F7F7F707070707F7F7070707F7F7F70707070707F707070707F7F70707070707F\
70707070707F70707070707F6F707070707F6F707070707F6F707070707F6F7070
70707F6F707070707F6F70706F707F6F6F706F707F6F6F706F707F606F706F707F606F706F6F70\
606F706F6F70606F70606F70606F70606F705F6070606F705F606F606F705F606F
5F60705F606F5F607050606F5F606F2F2F3F403F4F3F304F3F3040805F408F604F805F4F80504F\
7F4F407F5040704F3F704F3F60403F60403F5F3F30503030CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03030403030503F3F60303F60\
3F40703F40703F4F7F3F4070404F7F3F4F7F405080404F8040508F4050804F5F8F
4F5F8F4F5F904F5F90506090506090706F6F706F6070706F706F6070706F706F6F70706F706F6F\
70706F70706F707070706F6F70707070706F70707070706F707070707070707070
70707070707070707070707070707070707070707070707F70707070707070707070707F707070\
70707F70707070707F6F707070707F6F707070707F6F707070707F6F707070707F
6F6F706F707F6F6F706F707F6F6F706F707F6F6F706F707F6F6F706F7070606F706F7070606F70\
6F6F70606F70606F70606F70606F70606070606F705F606F606F705F606F5F6F70
5F606F5F606F5F606F5F606F505F6F5F606F30304030304F2F2F3F8F604F8F604F8F6050805F4F\
805F4F7F50407F5040704F3F7F4F40704F3F70403F603F305F3F3F503030503030
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030403F3F50303F5F3F406F\
3F40703F4F7F3F4070404F7F3F4F7F4050804050804F508F40508F4F5F8F4F5F8F
5060904F5F905060905060905F6F9F50609F70706F706F6F7F706F70706F70706F70706F7F7070\
70706F7F707070706F7F70707070707F70707070707F7F707070707F7F70707070
7F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F\
70707F707F7F707070707F7F70707F70707F707070707F7F70707F70707F6F7070
70707F6F707F70707F6F707070707F6F707070707F6F707070707F6F70706F707F6F6F706F707F\
6F6F706F707F606F706F707F606F70606F70606F70606F70606F70606F705F6070
606F705F6070606F705F606F5F6F705F606F30304F302F40906F508F6F50906F508F6050906F50\
8F60508F5F50805F4F805F4F8050408050407F4F407F504070403F6F4040603F3F
603F3F5F30305F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F3F4030304F3F3F5F303F60\
3F40703040703F4F7F3F4F7F404F803F4F7F4050804050804F5F8F4050804F5F8F
4F5F8F5060904F5F90506090506090506F9F706F6070706F706F6F70706F706F6F70706F70706F\
707070706F6F70707070706F7F707070706F7070707070707F7070707070707070
7070707F70707070707070707070707F707F70707070707070707070707F70707070707F707070\
70707F70707070707F70707070707F70707070707F70707070707F6F707070707F
6F707070707F6F6F7070707F6F707070707F6F6F706F707F6F6F706F707F6F6F706F707F6F6F70\
6F707F606F706F707F606F706F707F606F70606F70606F70606F70606070606F70
5F6070606F705F606F5F6F705F606F5F6F705F606F302F408F6F4F906F5F906F50906F5F906050\
906F5F8F60508F6050805F508F5F5080504F80504F7F50407F5040704F40704F40
6F403F6F4040603F3F603F3F5F3030CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030403F3F50303F5F3F406F\
3040703F4F7F3F4F7F404F803F4F7F40508040508040508F4050804F5F8F4F5F8F
4F60904F5F90506090506090506F9F50609F70706F706F6F70706F706F6F7F706F70706F7F7070\
70706F7F707070706F7F707070706F7F70707070707F70707070707F7F70707070
7F70707070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F\
7070707F7F7F70707F707F7F707070707F7F70707F70707F707070707F7F70707F
70707F6F707070707F6F707F70707F6F707070707F6F707070707F6F707070707F6F70706F707F\
6F6F706F707F6F6F706F707F6F6F706F707F6F6F706F707F606F706F707F606F70
606F70606F70606F70606F70606F705F6070606F703F3040907050906F509F705F90705F9F705F\
906F5F9F6F5F906F5F906F5F8F60508F6050805F4F805F5080504F805F4F7F504F
7F504F704F40704F406F403F6F4040603F3F603F3F5F3030503030CFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C030304030304F3F405F303F60\
3F40703040703F4F7F3F4F7F404F803F4F80405080405080405F8F40508F4F5F8F
4F5F8F4F60904F5F9050609F506090506F9F506F9F70706F706F6070706F706F6F70706F706F6F\
70706F70706F7F707070706F70707070706F7F707070706F7F70707070707F7070
7070707070707070707F70707070707070707070707F707F70707070707070707070707F707070\
70707F70707070707F70707070707F70707070707F70707070707F70707070707F
6F707070707F6F707070707F6F707070707F6F707070707F6F6F706F707F6F6F706F707F6F6F70\
6F707F6F6F706F707F6F6F706F707F6F6F706F707F606F706F6F70606F706F6F70
606F70606F70606F70606F705F6070606F705F60703F30408F6F509F705F9F705F9F7060906F5F\
9F7060906F5F9F7060906F5F906F5F8F605F8F605F805F50805F5080504F805F50
7F504F80504F7F4F407F4F4F7040406F4040603F3F6F3F3F603030603F305F3030403030302F30\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030403F3F50303F5F3F406F\
3F40703F4F7F3F4F7F4050803F4F804050804050804F5F8F405F8F4F5F8F4F5F8F
5060904F6090506090506090506F9F506F9F5F6FA0706F607F706F706F6F7F706F706F6F7F7070\
70706F7F707070706F7F70707070707F70707070707F7F707F70707F7F70707070
7F7F7F7070707F7F707070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F\
70707F7F7F7F7070707F7F7F70707F7F7F7F70707F7F7F7F70707F707F7F70707F
707F7F70707F70707F70707F707F7F70707F70707F70707F70707F70707F70707F6F707F70707F\
6F707F70707F6F707070707F6F707F6F707F6F6F706F707F6F6F706F707F6F6F70
6F707F6F6F706F707F606F706F707F606F70606F705F503F90705F9F705FA07F609F7F60A07F6F\
9F70609F706F9F70609F706F9F6F609F706090605F906F5F8F605F8F605F8F5F50
8F605F805F50805F5080504F8050507F4F4F704F4F6F40407040406F3F3F6F403F603F3F6F3F3F\
5F3030503030403030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F304030304F3F3F5F303F60\
3F40703040703F4F7F3F4F7F4050803F4F804050804050804F5F8F405F8F4F5F8F
4F5F8F5060904F609050609F506090506F9F506F9F5F6FA0706F6070706F706F6070706F706F6F\
7F706F706F6F70706F70706F7F707070706F7F70707070707F70707070707F7070
7070707F70707070707F70707070707F70707070707F70707070707F7F7F7070707F707F707070\
7F7F7F7070707F707F7070707F7F7F70707070707F707070707F7F70707070707F
70707F70707F70707070707F70707070707F6F707070707F6F707070707F6F707070707F6F7070\
70707F6F6F706F707F6F6F706F707F6F6F706F707F6F6F706F707F606F706F707F
606F706F707F606F706F6F70606F706F6F7F606F706F60409070509F7F609F7F60A07F6FA07F60\
A07F6F9F7060A07F6F9F70609F706F906F609F7060906F60906F608F605F8F605F
8F5F5F8F5F5F805050805F5080504F8050507F4F4F7F4F4F704040704F406F403F7040406F3F3F\
6F3F3F603030603F305F30305F30304F3030403030CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F50303F5F3F4060\
30406F3F4F7F3F4F7F4050803F4F80405080405080405F8F405F8F4F5F904F5F8F
4F60904F609050609F506090506F9F506F9F5F70A05F6FA070706F706F607F706F706F6F7F706F\
706F6F7F707070706F7F707070706F7F70707F70707F70707070707F7F707F7070
7F7F707070707F7F7F7F70707F7F707070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F\
7070707F7F7F70707F7F7F7F7070707F7F7F70707F7F7F7F70707F7F7F7F70707F
7F7F7F70707F7F7F7F70707F707F7F70707F707F7F70707F70707F70707F70707F70707F70707F\
6F707F70707F6F707F70707F6F707070707F6F707F6F707F6F6F706F707F6F6F70
6F707F6F6F706F707F6F6F706F707F606F706F707F706F408F704F9F7F5FA0806FA0806FAF8070\
A07F6FA07F70A07F6FA07F709F706FA07F6F9F706F9F706F906F609F6F6F906F60
9060608F5F5F8F605F805F5F8F5F5F805050805F508050508050507F4F4F7F4F4F704F40704040\
6F403F6F40406F3F3F6F3F3F603030603F3F5F3030503F304F30304F3030CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF30304F3F3F5F303F60\
3F40703040703F4F7F3F4F7F4050803F508040508F405080405F8F405F8F4F5F90
4F5F8F4F60904F6090506F9F506090506F9F506F9F5F70A0706F60706F6F706F6070706F706F60\
70706F706F6F7F706F706F6F7F707070706F7F707070706F7F70707070707F7070
7070707F70707070707F70707070707F70707070707F7F707070707F70707070707F7F7F707070\
7F707F7070707F7F7F7070707F707F7070707F7F7F70707070707F707070707F7F
70707070707F70707070707F70707070707F70707F70707F70707070707F70707F70707F6F7070\
70707F6F707070707F6F6F7070707F6F6F7070707F6F6F706F707F6F6F706F707F
6F6F706F707F6F6F706F707F606F706F6F70606F70706F5F8F704F9F7F60A07F60AF806FA0806F\
AF8070A07F70AF8070A07F6FA07F709F706FA070709F706F9F706F9F6F6F9F6F6F
8F60609060608F5F5F8F605F805F5F8F5F5F805F50805F5F7F50508050507F4F4F7F4F4F704040\
7040406F40407040406F3F3F6F3F3F6F3F306F3F3F6030305F3F305030305F3F30
4F30304F3030C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F4F303F503F4060\
3F406F3F4F703F4F7F4050803F508040508F4050804F5F8F405F8F4F5F904F5F8F
5060904F6090506F9F506090506F9F506F9F5F70A05F6FA05F70A0706F607F706F706F6F7F706F\
70706F7F706F70706F7F70707F706F7F70707070707F7F707F70707F7F707F7070
7F7F707F70707F7F707F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F\
7F707F7F7F7F7070707F7F7F7F707F7F7F7F70707F7F7F7F70707F7F7F7F70707F
7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F707F7F70707F707F7F\
70707F70707F70707F70707F70707F70707F6F707F70707F6F707F70707F6F707F
70707F6F707F6F707F6F6F706F707F6F6F7F3030307F6F5F907F5F9F7F5FAF806FAF8070AF8F7F\
AF8070AF807FAF8070AF807FA07F70A07F70A07F70A07F709F7070A070709F6F6F
9F706F906F60906F6F8F60609060608F60608F6060805F5F8F5F5F80505080505F7F4F507F5050\
704F4F7F4F4F704040704F4F7040407040406F3F3F70403F6F3F3F6F3F3F603F30
603F3F5F30305F3F3F5030305F3F3F503030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF30304F303F5F303F60\
3F406F30406F3F4F7F3F4F7F4050803F5080405F8F40508F4F5F8F405F8F4F6090
4F5F905060904F6090506F9F506F9F506F9F506F9F5F70A05F70A05F70A0706F6070706F706F6F\
7F706F706F6F7F706F70706F7F707070706F7F707070706F7F707070706F7F7070
7070707F70707070707F70707070707F7F707070707F7F707070707F7F7F7070707F7F7F707070\
7F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F707F
7070707F7F7F70707070707F70707F7F7F7F70707070707F70707F70707F70707F70707F70707F\
70707F70707F70707F70707F70707F6F707070707F6F707F70707F6F6F7070707F
6F6F7070707F6F6F706F707F6F6F706F707F706F4F80706090705F9F8060A0806FAF8F70AF8070\
AF8F7FAF807FAF8F7FAF807FAF807FA07F70A07F7FA07F70A07F709F70709F7070
9F6F6F9F6F6F906060906F6F8F60609060608F5F5F8F5F5F805F5F805F5F80505080505F7F4F50\
7F5050704F4F7F4F4F70404F704F4F6F40406F40406F3F3F6F3F3F603F306F3F3F
603030603F3F5F30305F303F5F30305F303F503030503030CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F304F3030503F3F60\
303F6F3F4F703F4F704050803F5080405F8F405080405F8F405F8F4F60904F5F90
4F60904F6090506F9F506F9F506F9F506F9F5F70A05F70A05F70A05F70A07F706F706F6F7F706F\
706F6F7F706F70706F7F707070706F7F70707F706F7F707070706F7F7F707F7070
7F70707F70707F7F707F70707F7F707F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F\
7F70707F7F7F7F707F7F7F7F7070707F7F7F7F707F7F7F7F70707F7F7F7F70707F
7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F707F7F\
70707F7F7F7F70707F70707F70707F70707F70707F70707F70707F70707F70707F
70707F6F707F70707F6F707F70707F6F6F70807F5F8F7F6F907F5F9F8060AF8F70AF8F70B08F7F\
AF8F7FB08F80AF8F7FAF8F80AF807FAF807FA07F7FAF807FA07F70A07F7F9F7070
9F70709F6F709F7070906F6F906F6F8F60608F60608F5F608F5F60805F5F8F5F5F80505F80505F\
7F50508050507F4F4F7F4F4F70404070404F6F40407040406F3F3F6F40406F3F3F
6F3F3F603F3F603F3F60303F603F3F5F3030603F3F5F30305F303F5F303050303FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F3F5030305F\
3F406F3F406F3F4F703F4F7F4050803F5080405F8F405080405F8F405F8F4F6090
4F5F904F60904F6090506F9F50609F506F9F506F9F5F70A0506F9F5F70A05F70A070706F706F60\
70706F706F6F7F706F706F6F7F706F70706F7F707070706F7F707070706F7F7070
70706F7F70707070707F70707070707F70707070707F7F707070707F7F707F70707F7F7F707070\
7F7F7F7F70707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F7070707F7F7F
7070707F7F7F70707F7F7F7F707070707F7F70707F7F7F7F70707F707F7F70707F707F7F70707F\
70707F70707F70707F70707F70707F70707F70707F6F707070707F70707F70707F
6F6F7070707F6F707F70707F6F6F706F707F8F805F908070907F5FA0806FAF8F6FB0907FAF8F7F\
B08F80AF8F7FB08F80AF807FAF8080AF807FAF8080A07F7FA07F7F9F7070A07F7F
9F70709F70709F6F709F6F7090606F90606F8F60608F6060805F5F8F5F6080505F805F5F7F5050\
7F505F7F4F507F4F5070404F70404F6F404070404F6F3F406F40406F3F3F6F3F40
6F3F3F6F3F3F60303F6F3F3F603030603F3F60303060303F5F30305F30305F30305F3030503030\
403030C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030503F3F60\
3F3F603F406F3F4F7040507F3F5080405F8F40508F4F5F8F405F8F4F60904F6090
50609F4F6090506F9F506F9F506F9F506F9F5F70A0506F9F5F70A05F70A05F7FAF5F70A07F706F\
706F6F7F706F70706F7F70707F706F7F70707F706F7F70707F70707F70707F7070
7F7F707F70707F7F707F70707F7F707F70707F7F707F70707F7F7F7F70707F7F7F7F70707F7F7F\
7F7F7F7F7F7F7F70707F7F7F7F7F7F7F7F7F7F707F7F7F7F7F7F7F7F7F7F7F707F
7F7F7F7F7F7F7F7F7F70707F7F7F7F7F7F7F7F7F7F70707F7F7F80707F7F7F7F7F70707F7F7F80\
70707F7F7F7F70707F7F7F8070707F707F7F70707F707F7F70707F70707F70707F
707F7F70707F70707F70707F70707F70704F908F6F908F709F8F6FA08F6FB0907FB0907FB09080\
B09080B09080B08F80B0908FB08F80B08F80AF8080AF8080A07F7FA07F7FA07F7F
A07F7F9F707F9F707F906F709F6F7090606F90606F8F60608F606F8F5F608F5F6080505F805F5F\
7F505F7F505F7F4F507F4F5070404F704F4F70404F70404F6F40406F40406F3F3F
6F3F406F3F3F6F3F3F60303F6F3F3F60303F603F3F5F3030603F3F60303060303F5F3030603F3F\
5F30305F303F503030503F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C030305030305F\
3F40603F40603F4F703F4F704050803F5080405F8F405F8F4F5F90405F8F4F6090
4F6090506F9F4F6090506F9F506F9F506F9F506F9F5F70A05070A05F70A05F70A05F7FA0706F60\
7F706F706F607F706F706F6F7F706F706F6F7F706F70706F7F707070706F7F7070
70706F7F707070706F7F70707F70707F70707070707F70707F70707F7F707070707F7F707F7070\
7F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F707F7F7F7F7070707F7F7F
70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F\
7F7F7F70707F707F7F70707F707F7F70707F70707F70707F707F7F70707F70707F
70707F70707F6F707F70707F6F707F8F8060908F6FA0907F9F8F6FAF9070AF8F70B0907FB09080\
B0908FB08F80B0908FB08F80B08F8FAF8080AF8F80A08080AF8080A07F7FA07F7F
9F707F9F7F7F906F709F7070906F70906F708F606F8F606F8F5F6F8F5F6F80505F8050607F505F\
7F505F7F4F507F4F50704050704F506F404F70404F6F3F406F3F4060303F6F3F40
60303F6F3F3F60303060303F60303060303F60303060303F60303060303F603030603030603030\
6030305F30305F30305F30305F3030503030503030CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3030503F3F5F\
3F3F603F406F3F406F404F7F3F4F7F405F8F405F8F405F8F405F8F4F60904F6090
4F60904F6090506F9F506F9F506F9F506F9F5F70A050709F5F70A05F70A05F7FAF5F70A0607FAF\
706F607F706F70706F7F706F70706F7F70707F706F7F70707F706F7F70707F706F
7F70707F706F7F7F707F70707F70707F70707F7F707F70707F7F707F70707F7F7F7F70707F7F7F\
7F70707F7F7F7F7F707F7F7F7F70707F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F
7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F807F7F7F7F7F7F707F7F7F7F807F7F7F7F7F80\
707F7F7F7F80707F7F7F7F7F70707F7F7F8070707F707F7F70707F7F7F8070707F
707F7F70707F707F7F70707F7F7F50908F60A09F7FA09080A09070AF907FB09F80B09080BF9F8F\
B09080BF9F8FB0908FB0908FB08F8FB0908FAF8F8FAF8F8FA08080AF8080A07F80
A07F809F707F9F707F90707F9F707F906F70906F708F606F8F606F805F60805F607F5060805060\
7F4F5F7F505F704F50704F506F404F70404F6F3F4F6F3F4F603F406F3F4060303F
6F3F4060303F6F3F3F60303F60303F603030603F3F60303060303F603030603F3F60303060303F\
6030306F3F3F60303060303F5F3030603F3F5F30305F3030503030CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF303050\
3F3F603F3F603F4F6F3F4F7040507F3F5080405F8F405F8F405F8F405F8F4F6090
4F60904F60904F6090506F9F4F6F90506F9F506F9F5F70A0506F9F5F70A05F70A05F7FA05F70A0\
607FAF706F607F706F706F6F7F706F706F6F7F706F706F6F7F706F70706F7F7070
70706F7F707070706F7F707070706F7F70707F70707F70707070707F70707F70707F7F707F7070\
7F7F707F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F707F7F7F7F
70707F7F7F7F7F7F7F7F7F7F70707F7F7F7F707F7F7F7F7F707F7F7F7F7F707F7F7F7F7F70707F\
7F7F7F70707F7F7F7F70707F707F7F70707F7F7F7F70707F707F7F70707F707F7F
70707F70707F70707F70707F7F7F5F9F9070A09F7FAF9F8FA09070B09F7FB09080BF9F80B09080\
BF9F8FB0908FBF908FB08F8FB0908FB08F8FB08F8FAF8080AF808FA07F80A08080
9F707F9F707F9F707F9F707F906F70906F708F60708F6070805F60805F607F50608050607F4F5F\
7F505F704F50704F506F404F6F404F6F3F406F3F4F603040603F4060303F603F40
5F303060303F5F303060303F5F30306030305F30306030305F3030603030603030603030603030\
6030306030306030306030306030306030306030306030306030305F3030503030
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F5F\
3F3F5F3F40603F406F404F70404F7F405080405F804F6090405F8F4F60904F6090
4F6F9F4F6090506F9F4F6F9F506F9F506F9F5F70A050709F5F70A05F70A05F7FA05F70A0607FAF\
5F7FAF6080AF706F6F7F706F7F706F7F70707F706F7F70707F706F7F70707F706F
7F70707F70707F70707F70707F7F707F70707F7F707F70707F7F7F7F70707F7F7F7F7070807F7F\
7F70707F7F7F7F7070807F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F
7F7F7F7F7F7F7F7F7F7F7F7F7F80807F7F7F7F7F807F7F7F7F80807F7F7F7F7F807F7F7F7F7F80\
7F7F7F7F7F80707F7F7F7F80707F7F7F7F80707F7F7F7F80707F7F7F7F8070707F
7F7F8070707F707F80707F508F8F6F9F9070AFA080B0A08FB09F80B09F80BF9F8FBF9F8FBF9F90\
BF9F8FBF9F90B0908FBF9090B09090B09090AF8F8FB08F90AF808FAF808FA07F80
A07F809F70809F708090707F9F707F906F70906F708F606F8F606F805F6F805F6F7F50607F5060\
704F5F704F5F6F40506F40506F3F4F6F3F4F603F406F3F4060303F603F3F5F303F
603F3F5F3030603F3F5F3030603F3F5F3030603F3F603030603F3F6030306F3F3F6030306F3F3F\
6030306F3F3F6030306F3F3F6030306F3F3F6030306F3F3F6030306F3F3F603030
5F3F3F4F3030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
3F3F5F303F603F406F3F406F404F7F3F507F405F80405F8F4F6090405F8F4F6090
4F60904F6F904F6090506F9F4F6F9F506F9F506F9F5070A050709F5F70A05070A05F7FA05F7FA0\
607FAF5F7FA07F6F6F706F607F706F706F6F7F706F7F706F7F7070706F6F7F7070
7F706F7F70707F706F7F70707F706F7F70707F706F7F70707F70707F7F707F70707F70707F7070\
7F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F7F70707F7F7F
7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F807F7F7F7F7F807F7F7F\
7F7F80707F7F7F7F80707F7F7F7F8070707F7F7F8070707F7F7F8070707F707F7F
70707F707F8070707F808060909070A09F7FAF9F80BFAF90B09F80BFA08FB09F8FC09F8FBF9F8F\
BF9F90BF9F90BF9F90B09090BF9090AF8F8FB08F90AF8F8FAF8F8FA0808FA0808F
9F7F809F7F8090707F9F707F906F70906F7F8F60708F6070805F6F805F6F7F50607F5060704F5F\
704F5F6F40506F4050603F4F603F4F5F3040603F405F303F5F303F5030305F303F
5F30305F303F5F30305F30305F303060303F5F30306030306030306F3F3F6030306F3F306F3030\
6F3F3F6F30306F3F306F3F306F3F3F6F30306F3F306F3F30703F3F6F30306F3F30
6030306F3F304F3030CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
30305F3F3F603F40603F4F703F4F70405080405080405F8F405F8F4F60904F6090
4F60904F6090506F9F4F6F9F506F9F506F9F5070A050709F5F70A05070A05F7FA05F7FA05F7FAF\
5F7FA06080AF607FAF7F706F706F6F7F706F7F706F7F706F7F706F7F70707F706F
7F70707F706F7F70707F70707F70707F70707F7F707F70707F7F707F70707F7F707F70707F7F70\
7F7070807F7F7F70707F7F7F7F7070807F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F
7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F807F7F7F7F7F807F7F7F7F80807F7F7F7F7F80\
7F7F7F7F7F807F7F7F7F7F80707F7F7F7F807F7F7F7F7F8070707F7F7F8070707F
7F7F8070707F7F80608F8F6FA09F7FAF9F80B0AF90BFAF90BFA090BFA08FC0A090C09F8FC0A090\
BF9F90C0A09FBF9F90BF9F9FB09090B09090B08F90B08F90AF808FAF8090A07F8F
A07F8F9F70809F7F80906F7F906F7F8F60708F6070805F6F805F6F7F50607F5060704F5F704F5F\
6F40506F4050603F4F603F4F5F30405F3F4050303F5F303F5030305F3F3F503030
5F3F3F5030305F3F3F5F30305F3F3F5F3030603F3F6030306F3F3F6030306F3F3F6F3F306F3F3F\
6F3F30703F3F6F3F30703F3F703F3070403F703F3070403F703F3070403F6F3F30
6F3F3F6F30306F3F3F603030503030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF30305F3F406F30406F3F4F703F4F7F405080405080405F8F405F8F4F6090
4060904F60904F6090506F9F4F6F90506F9F506F9F5070A050709F5F70A05070A05F7FA05F70A0\
5F7FAF5F7FA06080AF5F7FAF7F6F6F706F6F7F706F706F6F7F706F706F6F7F706F
706F6F7F706F7F706F7F70707F706F7F70707F706F7F70707F706F7F70707F70707F70707F7070\
7F70707F70707F7F7F7F70707F7F707F70707F7F7F7F70707F7F7F7F70707F7F7F
7F70707F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F7F807F7F7F\
7F7F807F7F7F7F7F80707F7F7F7F80707F7F7F7F8070707F7F7F8070707F7F7F80
70707F707F808080609F9070A09F7FB0A08FB0A08FC0B09FBFA090C0AF90BFA090C0A090BF9F90\
C0A09FBF9F90BF9F9FB09090BF909FB08F90B09090AF8F90AF8F90A0808FA0808F
9F70809F7F80907080906F7F8F60708F6070805F70805F707F50607050606F4F5F704F5F6F3F50\
603F505F304F5F3F4050303F50303F50303F50303F4F30305030304F303050303F
4F30305030305030305F303F5F30306030306030306F3F3F6030306F3F306F30306F3F3F6F3F30\
703F3F6F3F3070403F703F3070403F703F307F403F703F307F4030703F3070403F
6F3F306F3F306F3F306F3F30603030603030C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF303F603F406F3F407040507F3F507F405F8F405F8F4F60904F6090
4F6F904F6090506F9F4F6F9050709F506F9F5070A050709F5F70A05F70A05F7FA05F7FA0607FAF\
5F7FA06080AF607FAF6080B06080AF7F706F7F6F6F7F706F7F706F7F706F7F706F
7F70707F706F7F70707F706F7F70707F70707F70707F7070807F707F7070807F707F7070807F7F\
7F7070807F7F7F7070807F7F7F7070807F7F7F7070807F7F7F7F7F807F7F7F7F7F
807F7F7F7F7F807F7F7F7F7F807F7F7F7F7F807F7F7F7F7F8080807F7F7F8080807F7F7F808080\
7F7F7F807F807F7F7F807F807F7F807F7F807F7F7F7F7F807F7F807F7F80707F80
7F7F807F80609090709F9F7FAFA08FB0AF90BFB09FC0B09FC0B0A0CFB09FC0AF9FC0A09FCFAF9F\
C0A09FC0A09FBF9F9FC09F9FBF909FB0909FB08F90B0909FAF8F90AF8090A07F8F
A07F8F9F70809F70808F6F7F906F7F8F60708060707F506F7F506F704F60704F5F6F405F604050\
5F3F4F5F3F4050303F503F3F4F303F503F3F4F30304F3F3F4F3030503F3F503030
5F3F3F5F3030603F3F6030306F3F3F6F3F3F6F3F3F6F3F3F70403F703F3F70403F703F3F7F403F\
70403F7F403F7F403F7F4F407F403F7F4F3F7F403F7F4F3F7F403F7F4F3F7F4030
7F4F3F70403070403F703F3070403F6F3F306F3F3F603030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0303F60303F6F3F40703F4F7040507F3F5080405F8F405F8F4F6090
4F60904F6F904F6090506F9F4F6F9050709F506F9F50709F50709F5F70A05070A05F7FA05F7FA0\
5F7FAF5F7FA06080AF5F7FAF6080AF6080AF7F706F706F6F7F706F706F6F7F706F
7F6F6F7F706F706F6F7F706F7F706F7F70707F706F7F70707F70707F70707F70707F70707F7070\
7F7F707F70707F7F707F70707F7F7F7F70707F7F7F7F7070807F7F7F70707F7F7F
7F7F70807F7F7F7F70807F7F7F7F7F807F7F7F7F7F807F7F7F7F7F80807F7F7F7F807F7F7F7F7F\
807F807F7F7F7F7F807F7F7F7F7F807F7F7F7F7F807F7F7F7F7F8070707F7F7F80
6F70508F8F6F909070A0A080AFA08FB0AF90BFAF9FC0B0A0C0B0A0D0B0A0BFA09FCFAF9FC0A09F\
C0A0A0BF9F9FC0A09FBF9F9FBF9F9FB09090B0909FAF8F90AF8F90A0808FA0808F
9F7F8F9F708F8F6F7F8F6F7F80607F8060707F506F70506F6F4F606F4F60603F50603F50503040\
503F404F303F4F303F4030304F303F4030304F303F4F3030503F3F5030305F303F
5F3030603F3F6030306F3F3F6F3F3F6F3F3F6F3F3F70403F703F3F7F403F70403F7F40407F403F\
804F407F403F804F3F7F403F804F3F7F403F7F4F3F7F403F804F3F7F40307F4F3F
7F4030804F3F7F40307F4F3F70403070403F6F3F306F3F30603F30CFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F406F3F406F3F4F7F3F4F7F4050804050804F5F8F40608F
4F60904F6090506F9F4F6F90506F9F506F9F5070A050709F5F70A05070A05F7FA05F7FA05F7FAF\
5F7FA06080AF5F7FAF6080AF6080AF6080AF6080AF7F706F706F6F7F706F7F6F6F
7F706F7F6F6F7F70707F706F7F70707F706F7F70707F70707F70707F7070807F707F7070807F70\
7F7070807F7F7F7070807F7F7F7070807F7F7F7F70807F7F7F7F70807F7F7F7F7F
807F7F7F7F7F807F7F7F7F7F807F7F7F7F7F80807F807F7F80807F7F7F7F808080807F7F808080\
7F7F7F8080807F7F7F807F807F7F7F807F807F7F7F7F7F807F7F7F7F7F806F7F50
808F6F909070A0A080AFA08FB0AF9FB0AF90C0BFA0C0BFA0CFBFAFD0BFA0C0AFAFCFAFA0CFAFA0\
C0A09FC0A0A0BF9FA0C0A0A0B0909FB0909FB0909FAF8F9FA08090A08090A07F8F
9F7F8F9070809070808F6F7F8F607F7F5F6F7F506F6F4F606F4F60603F505F3F50503040503F40\
40303F4F3F3F403030403F3F4030304F3F3F4F3030503F3F5F3030603F3F603F3F
603F3F603F3F6F403F6F3F3F70403F6F3F3F70404070403F7F40407F403F804F40804F40804F40\
804F40805040804F40804F40804F3F804F40804F3F804F3F804F3F804F3F804F3F
804F3F804F3F80503F804F30804F3F7F4F307F4F3F70403070403F6F3F30CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF303F603F406F3F40703F4F7F3F4F7F4050804050804F5F8F
405F8F4F6F904F6090506F9F4F6F90506F9F506F9F50709F50709F5F70A05070A05F7FA05F70A0\
5F7FA05F7FA06080AF5F7FA06080AF6080AF6080AF6080AF7F6F6F706F607F706F
706F607F706F706F6F7F706F706F6F7F706F7F706F7F70707F6F6F7F70707F706F7F70707F706F\
7F70707F7070807F707F7070807F707F7070807F7F7F7070807F7F7F7070807F7F
7F7070807F7F7F7F70807F7F7F7F70807F7F7F7F7F807F7F7F7F7F807F7F807F7F807F7F7F7F7F\
807F7F7F7F7F807F807F7F7F807F7F7F7F7F807F807F707F7F7F807F707F708060
8F8F6F9F9F80A0A080B0AF90B0AF90BFB09FBFB0A0CFBFAFCFBFAFDFBFAFBFAFA0C0AFA0CFAFA0\
C0A0A0BF9F9FBFA0A0BF9F9FBF9FA0B0909FB0909FA08090A08090A080909F7F90
9070809070808F6F7F8F607F7F5F707F506F6F4F606F40605F3F505F3F4F4F30404F30403F3030\
3F30303F303040303F4030304F303F5030305F3F3F5F3030603F3F603F3F6F3F3F
603F3F6F403F6F3F3F70404070403F7F40407F4040804F407F4040804F40804F40804F40804F40\
805040804F408F504F804F408F5040804F408F5040804F3F8F5040804F3F8F5040
804F3F80503F804F3F80503F804F3080503F7F4F307F4F3F704030704030603F30CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF30406F3F40703F4F7040507F40507F405F80405F80
4F60904F60904F6F9F4F6F9050709F506F9F50709F50709F5F7FA05070A05F7FA05F7FA05F7FAF\
5F7FA06080AF5F7FA06080AF6080AF6080AF6080AF6F8FAF6080AF7F706F7F6F6F
7F706F7F6F6F7F706F7F6F6F7F70707F706F7F70707F706F8070707F70707F70707F7070807F70\
7F7070807F707F7070807F7F7F7070807F7F7F7070807F7F807F70807F7F807F70
807F7F807F7F807F7F807F7F80807F807F7F80807F807F7F808080807F7F808080807F7F808080\
807F7F808080807F7F808080807F7F807F807F7F7F807F807F7F7F7F8060808F6F
9F9F80A0A080AFAF90B0AF90BFB0A0BFB0A0C0BFAFC0BFAFD0C0BFDFC0B0C0AFB0C0AFA0D0B0AF\
C0A0A0C0A0A0BF9FA0C0A0A0B09FA0B09FA0AF8F9FAF8F9FA08090A080909F7F8F
9F708F8F6F808F6F80805F707F5F7070506F6F4F605F3F505F3F504F30404F3F403F30303F3F3F\
4030304F3F3F5030305F3F3F5F3F3F603F3F603F3F6F40406F3F3F704040704040
7F40407F4040804F4F7F4040804F4F804F40804F4F804F408F504F804F4F8F504F8050408F504F\
8F504F8F504F8F504F905F4F8F5F4F905F4F8F504F905F4F8F5F40905F4F8F5040
905F4F8F5F408F5F408F503F8F5F408F503F8F503F8050307F4F3F70403070403F6F3F30CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF3F406F303F6F3F4F703F4F7040507F3F507F405F80
405F8F4F60904F60904F6F9F4F6F9050709F4F6F9F50709F50709F5F7FA050709F5F7FA05F7FA0\
5F7FA05F7FA05F7FA05F7FA06080AF5F7FA06080AF6080AF608FAF6080AF7F6F6F
706F607F706F706F6F7F706F7F6F6F7F706F7F6F6F7F706F7F706F7F70707F706F7F70707F706F\
7F70707F70707F70707F7070807F707F7070807F707F7070807F7F7F7070807F7F
807F70807F7F807F70807F7F807F7F807F7F807F7F807F7F807F7F80807F807F7F807F7F807F7F\
80807F807F7F807F7F807F7F807F807F7F7F807F807F7F7F7080607F806F90907F
9F9F80AFAF8FAFAF90BFB09FBFB0A0C0BFAFC0BFAFCFC0B0CFBFB0E0CFBFBFAFB0BFAFAFD0AFAF\
CFAFAFBF9FA0BFA0A0BF9FA0BF9FA0B0909FAF909FA08090A0809F9F7F8F907F8F
8F6F808F6F80805F7F7F5F7070506F6F4F6F5F3F50503F4F4F30404F303F30303040303F4F3030\
503F3F5030305F3F3F5F3F3F603F40603F3F6F40406F40407040407040407F4F4F
7F4040804F4F804F40804F4F804F408F504F804F408F504F804F4F8F504F8F504F8F504F8F504F\
905F4F8F504F905F4F905F4F905F4F905F4F905F4F905F4F90604F905F4F905F4F
905F40905F4F8F5F408F5F408F503F8F5F4080503F8F503F80503F80503F7040307040306F3F30\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F40703F40703F4F703F4F7F405080405080
4F5F8F405F8F4F6F904F6F9050709F4F6F9F50709F50709F5070A050709F5F7FA05F7FA05F7FA0\
5F7FA05F80A05F7FA06080AF5F7FA06080AF6080AF6080AF6080AF6F8FAF608FAF
7F706F706F6F7F706F7F6F6F7F706F7F6F6F7F70707F706F7F70707F706F7F70707F706F7F7070\
7F70708070707F7070807F707F7070807F707F7070807F707F7070807F7F807F70
807F7F807F70807F7F807F7F807F7F807F7F8F807F807F7F8F807F807F7F8F807F807F7F8F807F\
807F7F808080807F7F807F80807F7F807F80807F7F70806F7F8F6F8F90709F9F80
A0A08FAFAF90BFB0A0BFB0A0C0BFAFC0BFAFCFC0B0CFC0B0D0C0BFE0CFBFC0B0BFBFA0AFD0AFAF\
C0A0AFC0A0AFBF9FA0BF9FA0B09FA0B09FA0AF8F9FAF8F9FA080909F7F9090708F
90708F80607F7F607F7050706F4F6F5F3F5F503F504030403F3F3F403030503F3F503F3F5F3F3F\
5F3F3F604040603F3F6F40406F404070404F7F4040804F4F7F404F804F4F804F4F
8F504F804F4F8F50508F504F8F50508F504F905F508F5050905F508F504F905F50905F50905F50\
905F509F6050905F509F6050905F509F60509F60509F60509F60509F605090604F
9F604F905F4F90604F905F40905F408F5F40905F408F5F3F8F5F3F80503F80503F7F4F3F704F3F\
6F40306F403FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF303F603F407030406F3F4F7F3F4F7F405080
405080405F8F40608F4F6F904F6F9050709F4F6F9F50709F50709F50709F50709F5F7FA050709F\
5F7FA05F7FA05F7FA05F7FA05F80A05F7FA06080AF5F80A06080AF6080A0608FAF
6080AF608FAF706F607F706F706F607F706F7F6F6F7F706F706F6F7F706F7F6F6F7F70707F6F6F\
7F70707F706F7F70707F706F7F70707F7070807F707F7070807F707F7070807F70
7F7070807F70807F70807F7F807F70807F7F807F708F7F7F807F708F7F7F807F7F8F807F807F7F\
807F7F807F7F807F7F807F7F807F7F807F7F70806F7F8F6F8090708F907FA0A08F
A0A08FB0AF9FB0B09FBFBFA0BFB0A0C0C0B0C0C0B0CFC0BFCFBFBFEFCFC0C0AFBFBFAFAFCFAFAF\
CFAFB0BFA0AFBFA0AFB090A0B09FA0AF90A0AF8F9FA080909F809090708F90708F
8060807F607F7050706F506F5F3F5F503F503F30304F3F3F4F3030503F3F5F3F3F603F40603F3F\
6F40406F3F407040407F40407F404F7F404F804F4F804F4F804F4F804F4F8F5050
8F504F8F50508F5050905F508F5050905F50905F50905F5F905F50906050905F509F605F905F50\
9F605F9F60509F605F9F60509F60509F60509F6F509F60509F60509F60509F6050
90604F9F604F905F4F90604F8F5F409060408F5F40905F408F5F3F8F5F3F80503F80503F7F4F30\
704F3F6F40306F4030C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303F603F4F703F4F7040507F3F507F
405F80405F804F608F4F608F506F9F4F6F9F50709F50709F507FA050709F5F7FA050709F5F7FA0\
5F7FA05F7FA05F7FA06080A05F80A06080AF5F80A0608FAF6080AF608FAF6080AF
6F8FAF608FAF6F8FAF7F6F6F7F706F7F6F6F7F706F7F6F6F7F70707F706F7F70707F706F807070\
7F70708070707F7070807F707F7070807F707F7070807F7F807070807F7F807070
807F7F807F70807F7F807F708F7F7F807F7F8F7F7F807F7F8F807F807F7F8F807F807F7F8F8080\
807F7F8F8080807F7F8F8080807F7F70806F7F8F708F907F8F9F7F9FA08FA0A08F
B0B09FB0B0A0BFBFA0BFBFAFC0C0B0C0C0B0CFCFBFCFC0BFD0C0C0E0CFC0CFBFC0BFAFB0CFAFB0\
CFAFB0C0A0B0BF9FAFB09FAFB090A0B090A0A08F9FA08F9F9F7F90907F908F6F8F
806080705F707050703F30304F3F3F503F3F5F3F405F3F3F604040603F407040407040407F404F\
7F404F804F4F7F404F804F50804F4F8F5050804F4F8F50508F50508F5F508F5050
905F5F905F50905F5F905F5F9F605F9F605F9F605F9F605FA06F609F605FA06F609F605FA06F60\
9F6F5FA06F5F9F6F5FA06F5FA06F5FA06F5F9F6F50A06F5FA06F50A06F5F9F6F50
A06F5F9F6F509F6F5090604F9F6F4F90604F90604F9060409060408F5F409060408F5F3F8F5F40\
7F503F7F4F3F6F40306F403FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF30406F3F4F70304F703F507F
3F507F405F80405F804F608F4F60904F6F9F4F6F9050709F50709F50709F50709F5F7FA050709F\
5F7FA0507F9F5F7FA05F7FA05F80A05F7FA06080A05F80A06080AF6080A0608FAF
6080AF6F8FAF608FAF6F8FAF706F607F706F706F607F706F7F6F6F7F706F7F6F6F7F706F7F6F6F\
7F70707F6F6F7F70707F706F8070707F706F8070707F7070807F707F7070807F70
807070807F70807070807F70807F70807F7F807F70807F7F807F708F7F7F807F70807F7F807F7F\
807F7F807F7F807F7F6F7F6F70806F7F8F7080907F8F907F9FA0809FA08FAFAF90
AFAF9FBFB0A0BFB0A0C0BFAFC0BFB0CFCFBFC0CFBFCFC0C0D0C0C0E0CFC0CFB0C0BFAFBFC0A0AF\
CFAFB0BFA0AFBFA0AFAF90A0B09FAFAF8FA0A08FA09F809F907F908F708F806F80
705F7F4F3F3F4F303F503F3F5F3F3F603F40603F406F4040703F407F404F7F40407F404F7F404F\
804F4F7F4F4F804F50804F4F8F50508050508F50508F5050905F5F8F5050905F5F
905F5F9F605F905F5F9F605F9F605FA06F609F605FA06F60A06F60A06F609F6F5FA06F60A06F5F\
A06F609F6F5FA06F5FA06F5FA06F5F9F6F5FA06F5FA06F5FA0705FA06F50A06F5F
A06F50A06F5F9F6F509F6F509F604F9F6F4F90604F90604F9060409060408F5F3F9060408F5F3F\
8F5F3F7F503F704F3F6F4030CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F5F30406F3F4F7F3F4F7F
4050803F5F8040608F40608F4F6F904F6F9050709F4F6F9050709F50709F5F7FA0507F9F5F7FA0\
507F9F5F80A05F7FA05F80A05F7FA06080AF5F80A06080AF6080A0608FAF608FAF
608FAF608FA06F8FAF608FAF6F8FAF608FA06F90AF7F6F6F7F706F7F6F6F7F70707F6F6F7F7070\
7F6F6F8070707F706F8070707F706F8070707F7070807F707F7070807F70807070
807F70807070807F7F807070807F7F807070807F7F807F70807F7F807F708F7F7F807F7F807F7F\
807F7F807F7F6F806F708F707F8F708F907F8F9F7F9FA08F9FA08FAFAF90AFAF9F
B0B0A0B0B0A0C0BFAFC0BFB0CFC0BFC0CFBFCFD0C0C0C0C0D0CFCFDFC0C0D0C0D0BFAFBFC0AFB0\
C0A0B0C0AFB0B09FAFB09FAFAF90A0AF90A0A08F9F9F809F907090403F3F4F3F3F
503F3F503F3F604040603F406F4040703F407F404F7F404F804F4F7F4F4F804F50804F508F5050\
8050508F505F8F5050905F5F8F505F905F5F905F5F9F605F905F5F9F60609F6060
9F6F609F6060A06F60A06F60A06F6FA06F60AF706FA06F60AF706FA06F60AF7060A06F60AF7060\
A06F60AF7060A07060AF7060A06F5FAF7060A0705FAF7060A0705FAF7060A0705F
AF705FA0705FA0705FA06F50A070509F6F50A06F509F6F4F9F6F4F9060409F6F4F906040906040\
8F603F8F5F407F503F704F3F604030CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0303F5F30406F3F507F
304F7F3F5F803F5F8040608F40608F4F6F904F6F904F6F904F6F9050709F50709F5F7F9F50709F\
5F7FA0507F9F5F7FA05F7F9F5F80A05F7FA06080A05F80A06080A06080A0608FAF
6080A0608FAF608FA0608FAF608FA0608FA0608FA06F90AF608FA07F6F6F7F6F6F7F706F7F6F6F\
7F706F7F6F6F7F70707F6F6F7F70707F706F8070707F706F8070707F7070807070
7F70708070707F7070807F707F7070807F70807070807F7F807070807F7F807070807F7F807070\
6F80706F806F708F707F8F7080907F8F907F90A08F9FA08FA0AF90A0AF90B0B0A0
B0B0A0BFBFAFBFBFAFC0C0B0C0C0B0CFCFBFCFCFC0CFC0CFD0C0CFDFC0C0CFBFCFBFAFC0BFA0B0\
BFA0B0BFA0B0B09FB0AF90A0A090A0A08F9F403F3F40303F4F3F3F503F3F5F3F40
603F3F6F3F40703F4070404F7040407F404F7F404F804F50804F508050508050508F505F8F5050\
905F5F8F5F5F905F60905F5F9F606090605F9F60609F60609F6F609F6060A06F6F
9F6F60A06F6FA06F60A0706FA06F60A0706FA06F60AF706FA07060AF706FA07060AF706FA07060\
AF7060A07060AF7060A07060AF7060A0705FAF7060A0705FAF7060A0705FAF7060
A0705FAF705FA0705FAF705FA07050A070509F6F50A070509F6F4F9F6F4F906F4F9F6F4F906040\
9060408F5F3F805F3F704F306F4F3FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F3F4060304F70
3F507F3F507F405F803F5F8040608F40608F4F6F904F6F9050709F50709050709F50709F5F7FA0\
5F7F9F5F7FA05F7F9F5F80A05F80A06080A05F80A0608FAF6080A0608FAF6080A0
608FAF608FA0608FAF608FA06F8FAF608FA06F8FAF608FA06F90AF6F8FA06F90A07F6F6F7F7070\
7F6F6F7F70707F6F6F8070707F706F8070707F706F8070707F70708070707F7070
807F70807070807F70807070807F7F807070807F7F8070708F7F7F807F70807F7F807070708F7F\
708F70708F7F7F8F70809F808F9F8090A08F9FA08FA0AF90A0AF9FB0B0A0B0B0A0
BFBFAFBFBFAFC0C0B0C0BFB0CFCFC0CFD0C0CFD0CFC0C0CFDFCFD0D0C0C0CFBFD0BFAFC0BFAFBF\
BFA0B0BFA0B0B09FAF403F3F40303F503F40503F3F603F406F3F40703F40703F40
7F404F70404F7F4F507F404F804F50804F508F505F8050508F5F5F8F5F5F905F60905F609F6060\
9060609F606F9F6060A06F6F9F6F6FA06F6FA06F6FA0706FA06F6FA0706FA06F6F
AF706FA0706FAF706FA0706FAF7F70AF706FAF7F6FAF706FB07F70AF7F6FB07F6FAF706FB07F6F\
AF7F6FAF7F6FAF7060B07F6FAF7F60AF7F6FAF7F60B07F6FAF7F60AF7F60AF7F60
AF7F60AF7F60AF7F60AF7F5FAF7F60AF705FAF7F5FA07050A070509F70509F70509F6F4F9F6F4F\
906F40906F408F5F3F805F3F704F3F6F4F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF30304F304060
304F703F507F3F507F3F5F803F5F8040608F40608F4F6F904F6F904F70904F6F9050709F50709F\
5F7FA0507F9F5F7FA05F7F9F5F80A05F7F9F5F80A05F80A06080A05F80A06080A0
6080A0608FA06080A0608FA0608FA0608FA0608FA0608FA0608FA06F90A0608FA06F90A0608FA0\
7F706F7F6F6F7F706F7F6F6F7F70707F6F6F7F70707F6F6F8070707F6F6F807070
7F706F8070707F706F8070707F70708070707F70708070708070706F8F7F6F80706F8F7F708F70\
7F8F7F7F8F708090808F908090A08F90A08F9FAF90A0AF90AFB0A0AFB0A0B0BFAF
B0B0AFBFBFB0BFBFB0C0C0BFC0CFBFCFDFC0C0CFCFC0C0D0D0C0D0D0C0C0C0B0CFBFB0CFB0A0B0\
403F3F40303050303F5F303F603F406F303F6F3F406F3F4070404F7040407F404F
7F404F804F507F4F508050508050508F505F8F505F905F60905F5F906060905F609F606F9F606F\
A06F6F9F6F6FA06F6FA06F6FA06F6FA06F6FA0706FA06F6FAF7070A06F6FAF706F
A0706FAF7070AF706FAF7F70AF706FAF7F70AF706FB07F70AF7F6FB07F70AF7F6FB07F70AF7F6F\
B07F70AF7F6FB07F6FAF7F6FB07F6FAF7F60B07F6FAF7F60B07F6FAF7F60B07F60
AF7F60B07F60AF7F60AF7F60AF7F5FAF7F60A0705FA07F5FA07050A070509F6F4F9F704F906F4F\
9F6F4F906040906F40805F3F7F503F6F4F30604030CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303F50
3F4F6F30507F3F5F803F5F804060803F5F8040608F40608F4F6F904F6F90507090507090507F9F\
50709F5F7FA05F7F9F5F80A05F7FA06080A05F80A06080A05F80A0608FA06080A0
608FA06080A0608FA0608FA0608FA0608FA06F8FA0608FA06F8FA0608FA06F90A06F8FA06F90A0\
6F8F9F6F90A06F909F6F909F6F8F9F8070707F6F6F8070707F6F6F8070707F706F
8070707F706F8070708070708070706F8F806F8F806F8F7F6F8F806F8F7F708F7F708F7F7F907F\
7F907F809F808F9F8090A08F90A090A0AF9FA0AF9FAFB0A0AFB0A0B0BFAFB0BFAF
BFBFB0BFBFB0C0C0BFC0C0BFCFD0C0CFD0C0CFD0D0C0C0D0D0CFD0CFBFC03F3F3F40303F503F3F\
50303F603F406F3F406F3F406F3F4070404F70404F7F4F507F404F804F50804F50
80505F80505F8F5F5F8F505F905F60905F609F606F9060609F6F6F9F606FA06F70A06F6FA06F70\
A06F6FA07070A06F70AF7070A0706FAF7070A07070AF7070AF7070AF7F70AF7070
AF7F70AF7070B07F70AF7F70B07F70AF7F70B08070B07F70B08070B07F70BF807FB08070B08070\
B08070BF8070B08070B08070B0806FB08070B0806FB0806FB0806FB0806FB0806F
B0806FB07F60B0806FB07F60B08060AF7F60AF7F60A07F5FAF7F5FA07050A070509F70509F7050\
906F4F9F704F906F40906F40805F3F7F503F6F4F30CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
303F50304F6F30507F3F5F803F507F3F5F803F5F8040608F40608F4F6F8F4F6F8F4F70904F7090\
50709F50709F5F7F9F507F9F5F80A05F7F9F5F80A05F7F9F5F80A05F809F6080A0
5F809F6080A05F809F608FA060809F608FA0608F9F608FA0608F9F608FA0608F9F6F90A0608F9F\
6F8F9F608F9F6F909F608F906F8F9F608F906F9090608F906F8F90608F8F6F9090
608F8F6F8F8F608F806F8F8F608F806F8F806F8F7F6F8F806F8F7F708F7F708F7F7F90807F907F\
809080809F8090A08F90A08F9FAF909FAF90AFB0A0AFAFA0B0B0AFB0B0AFBFBFB0
BFBFB0BFBFBFBFBFB0C0C0BFC0CFC0CFD0CFC0CFD03F3F3F4030304F303F50303F603F4060303F\
6F3F406F3F4070404F70404F7F40507F4050804F507F4F5080505F80505F8F5F60
8F505F905F60905F6090606F9060609F606F9F606F9F6F709F606FA06F70A06F6FA07070A06F70\
A07070A07070AF7070A07070AF7070AF7070AF7F70AF7070AF7F70AF7070B07F70
AF7F70B07F70AF7F70B08070AF7F70B08070B07F70B0807FB07F70B0807FB08070BF807FB08070\
BF8070B08070BF8070B08070BF8070B0806FBF8070B0806FB0806FB0806FB0806F
B0806FB0806FB08060B0806FAF7F60B08060AF7F5FAF7F60A07F5FA07F5FA07050A070509F704F\
9F704F906F4F9F6F4F906F408F60407F5F3F70503F604030CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF303F5F3F4F7030507F3F5F803F5F8040608F4060804F6F8F406F8F4F6F904F6F9050709F\
507090507F9F507F9F5F80A05F7F9F5F80A05F809F6080A05F809F6080A05F809F
608FA06080A0608FA060809F608FA0608FA06F8FA0608F9F6F90A0608FA06F90A0608F9F6F90A0\
6F8F9F6F909F608F906F909F6F8F906F9090608F906F90906F8F906F90906F8F8F
6F90906F8F8F6F8F8F6F8F8070908F6F8F80709080708F807F90807F90807F90807F9080809F8F\
8F9F8F90A09090A0909FAF9F9FAF9FA0B0A0AFB0A0B0BFAFB0B0AFBFBFB0B0BFB0
C0C0BFBFBFBFC0C0BFC0C0BFCFD0CF40303F4F3F3F50303F603F40603F406F404F6F3F4070404F\
70404F7F4F507F4F5080505F80505F8F506080505F8F5F608F5F6090606F905F6F
9F606F9F606F9F6F709F606FA06F709F6F70A06F70A06F70A0707FA07070AF707FA07070AF7F7F\
AF707FAF7F7FAF7F7FB07F7FAF7F7FB07F7FAF7F7FB0807FB07F7FB0807FB07F7F
B0807FB0807FBF807FB0807FBF807FB0807FBF8F7FB0807FBF8F7FBF8F7FBF8F7FBF8F7FC08F7F\
BF8F7FBF8F7FBF8F70BF8F7FBF8F70BF8F70B08070BF8F70B08070BF8F70B0806F
BF8F70B0806FB08F6FB0806FB08F6FB0806FB0806FAF8060AF8060AF7F5FAF7F5FA07050A07F50\
9F70509F705090704F9F704F906F408F60407F5F3F70503F604F30CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF303F5F304F703050703F5F803F5F80406080406080406F8F40608F4F6F904F6F90\
5070904F7090507F9F507F9F5F7F9F507F9F5F809F5F809F6080A05F809F6080A0
5F809F608FA05F809F608FA060809F608FA0608F9F608FA0608F9F6F8FA0608F9F6F8F9F608F9F\
6F8F9F608F906F8F90608F906F8F90608F906F8F906F8F906F90906F8F8F6F8F8F
6F8F8F6F8F8F6F8F806F8F806F8F80709080708F807F90807090807F90807F9080809F8F809F8F\
90A09090A0909FAF9F9FAF9FA0B0A0A0AFA0AFB0AFAFB0AFB0BFB0B0BFAFBFBFB0
BFBFB0BFC0BF3F30304F303F50303F5F3F405F3F3F603F40603F406F404F6F404F704F50704F50\
7F505F7F4F5F8050608050608F5F608F5F60905F6F905F6F9F606F90606F9F6070
9F606F9F6F709F606FA06F709F6F70A07070A06F70A0707FA07070AF707FA07070AF7F7FAF7F7F\
B07F7FAF7F7FB07F7FAF7F7FB08080AF7F7FB0807FB0807FB08080B0807FB0807F
B0807FBF8F80B0807FBF8F80BF807FBF8F80BF807FBF8F80BF8F7FBF8F80BF8F7FBF8F7FBF8F7F\
C08F7FBF8F7FBF8F7FBF8F70BF8F7FBF8F70BF8F70BF8F70BF8F70B0806FBF8F70
B0806FBF8F70B0806FB08F6FB0806FB08F6FB08060B08060AF8060AF8060A07F5FA07F5FA07F50\
A07F509F704F9F704F90704F9F704F8F6F4080604070503F6F503FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF3040603F4F703050703F5F803F5F8040608F4060804F6F8F4F6F8F4F7090\
4F7090507F9F507090507F9F507F905F809F5F7F9F5F809F5F809F6080A05F809F
608FA060809F608FA0608F9F608FA0608F9F6F8FA0608F9F6F8F9F608F9F6F8F9F608F9F6F8F9F\
608F906F909F6F8F906F909F6F8F906F90906F8F906F90906F8F8F6F90906F8F8F
70908F6F8F8F70908F70908F70908F7090807F908F7F908F7F9F8F80908F8F9F908F9F9090A090\
90A0909FAF9F9FAF9FA0B0A0A0B0A0AFB0AFAFB0AFB0BFB0B0BFAFBFC0BF3F303F
4F3F3F50303F5F3F405F3F406040406040406F4F4F6F404F704F50704F507F505F7F505F805060\
8050608F5F6F8F5F6F90606F90606F9F607090606F9F6F709F60709F6F709F6F70
A06F709F6F70A0707FA07070A0707FA07070AF7F7FAF707FAF7F7FAF7F7FB07F80AF7F7FB08080\
B07F7FB08080B08080B08080B08080BF8F80B08080BF8F80B08080BF8F80BF8F80
BF8F80BF8F80C08F80BF8F80C08F80BF8F80C09080BF8F80C09080BF8F80C09080BF8F80C09080\
BF8F7FC09080BF8F7FC0907FBF8F7FC0907FBF8F70BF8F7FBF8F70BF907FBF8F70
BF8F70B08F6FBF8F70B08F6FBF8F70B08F6FBF8F6FB08F6FB08F6FAF8060AF8060AF7F5FAF805F\
A07F50A07F509F70509F705090704F9F704F8F6F408060406F503FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF303F603F4F70304F703F5F7F3F5F804060804060804F6F8F406F8F\
4F70904F70905070904F7090507F905070905F7F9F507F905F809F5F7F9F5F809F
5F809F60809F5F809F608F9F60809F608F9F608F9F608F9F608090608F9F608F906F8F9F608F90\
6F8F9F608F906F8F90608F906F8F906F8F906F90906F8F8F6F8F8F6F8F8F70908F
6F8F8F70908F708F8F70908F7090807F908F7F908F7F9F8F7F908F8F9F908F9F9090A0908FA090\
90AF9F9FA09FA0AFA0A0AFA0AFB0AFAFB0AFAFB0AFAFB0AF40303F5030305F3F3F
5F3F3F603F40603F406F404F6F404F704F50704F507F505F7F4F5F8050608050608F5F6F8F5F60\
905F6F8F5F6F90607090606F9F60709060709F6F709F60709F6F709F6F70A0707F
9F6F70A0707FA0707FA0707FA0707FAF7F7FAF707FAF7F80AF7F7FAF7F80AF7F7FB08080B07F80\
B08080B08080BF8F8FB08080BF8F80B08080BF8F8FBF8080BF8F80BF8F80BF8F8F
BF8F80BF8F80BF8F80C09080BF8F80C09080BF8F80C09080BF8F80C09080BF8F80C09080BF8F7F\
C09080BF907FC09080BF8F7FC0907FBF8F7FC0907FBF8F70BF907FBF8F70BF9070
B08F70BF8F70B08F70BF8F70B08F6FBF8F6FB08F6FBF8F6FB08060B08F60AF8060AF8060A07F5F\
A0805FA07F50A07F509F704F9F704F90704F90704F7F603F705F3F604F30CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3040603F4F703F5070405F7F405F80406F8F40608F4F6F90\
4F6F8F4F70904F7090507F90507090507F90507F905F7F9F507F905F809F5F7F90
60809F5F809F608F9F60809F608FA0608F9F6F8F9F608F9F6F909F608F9F6F8F9F608F9F6F909F\
6F8F906F909F6F8F906F909F6F8F906F90906F8F907090906F90907090906F908F
70909070908F70908F70908F7F9F907F908F809F90809F908FA09F8FA09090A09F90A0909FAF9F\
9FAF9FA0AFA0A0AFA0AFB0AFAFB0AFAFBFAF4F303F503F405F3F4060404F603F4F
6F40506F4050704F50704F507F505F7F505F8050608050608F5F6F8F5F6F905F6F8F5F6F906070\
9060709F6F709060709F6F7F9F6F7FA0707F9F6F7FA0707FA0707FA0707FA0707F
AF7F80A0707FAF7F80AF7F7FB07F80AF7F80B08080AF7F80B0808FB08080B0808FB08080BF8F8F\
BF8F8FBF8F8FBF8F8FC0908FBF8F8FC0908FBF8F8FC0908FBF8F8FC0908FBF8F8F
C0908FC0908FC0908FBF9080C0908FC09080C0908FC09080C0908FC09080C0908FC09080C09F8F\
C09080C09F80C09080C09F80C09080C09080BF907FC09080BF907FC0907FBF9070
C0907FBF9070BF9070BF8F70BF9070BF8F70BF9070BF8F6FBF9070B08F6FB08F6FAF8060B08F60\
AF805FAF805FA07F50A080509F7F509F7F4F90704F8F6F407F5F3F6F503FCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF303F5F303F603F4F6F3F4F70405F7F3F5F7F406080406F8F\
4F6F8F406F8F4F70904F708F5070904F708F507F90507090507F90507F905F809F
5F7F905F809F5F8090608F9F60809F608F9F608090608F9F608F906F8F9F608F906F8F9F608F90\
6F8F9F608F906F8F906F8F906F90906F8F906F90906F8F8F7090906F8F8F70908F
70908F70909070908F7F9F907F908F809F90809F908FA0908F9F908FA09F8FA09090AF9F90AF9F\
9FAFA09FAFA0A0B0AF403030503F3F503F3F5F3F405F3F406F40506F404F704F50
704F507F4F5F7F4F5F8050608050608F5F6F8050608F5F6F8F5F6F9060708F5F6F906070906070\
9F6F7F9060709F6F7F9F6F7FA0707F9F6F7FA07080A0707FAF7F80A0707FAF7F80
AF7F80AF7F80AF7F80AF8080AF7F80B0808FB08080B0808FB0808FBF8F8FB0808FBF8F8FBF8F8F\
BF8F8FBF8F8FBF8F8FBF8F8FC0908FBF8F8FC0908FBF8F8FC0908FBF8F8FC0908F
BF908FC0908FBF9080C0908FC0908FC0908FC09080C0908FC09080C09F8FC09080C09F8FC09080\
C09F8FC09080C09F80C09080C09F80C0907FC09080C0907FC09080BF907FC0907F
BF907FC0907FBF9070BF9070BF9070BF9070BF8F6FBF9070BF8F6FBF906FB08F60B08F6FAF8060\
AF8F60A0805FA0805F9F7F50A07F5090704F9F704F8F6F407F60406F5030CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3040603F4F6F3F4F70405F7F405F7F40608F\
406F8F4F70904F6F8F4F70904F708F507F90507090507F90507F905F809F5F7F90
5F809F5F809060809F5F809F608F9F608090608F9F608F90608F9F608F906F8F9F608F906F8F9F\
608F906F909F6F8F906F909F6F8F907090906F90907090906F8F8F709090709090
7F90907090907F9F907F9F90809F9F809F908FA09F8FA09F8FA09F8FA09F90AFA090AFA09FAFA0\
4030304F3F404F3F405040405F3F40604050604050704F5F70405F7F4F607F4F60
8050608050608F5F6F80506F8F5F6F8F5F6F9060708F5F709060709060709F6F7F906F709F6F7F\
9F6F7FA07080A0707FA07080A07080AF7F80A07F80AF7F80AF7F80AF808FAF7F80
B0808FAF8080B0808FB0808FB08F8FB0808FBF8F90BF8F8FBF8F90BF8F8FBF9090BF8F8FBF9090\
BF8F8FC09090BF9090C09090BF908FC09090C0908FC09090C0908FC09F90C0908F
C09F90C0908FC09F90C0908FC09F8FC0908FCF9F8FC09F8FCF9F8FC0908FCF9F8FC09F8FCF9F8F\
C09F8FCF9F8FC09F8FCF9F8FC09F80CF9F8FC09F80C09F80C0907FC09F80C0907F
C09F7FBF907FC09F7FBF907FC09F7FBF9070C09F7FBF9070BF9070BF906FBF9070B08F6FB08F6F\
AF8060AF8F60A0805FA0805F9F7F509F7F509F7F4F90704F7F603F705F3F604F30
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF303F603F4F6F3F4F6F3F5F7F3F5F7F\
4060804060804F6F8F406F8F4F708F4F708F5070904F708F507F90507F905F7F90
507F905F809F5F809060809F5F809060809F5F8090608F905F8090608F90608090608F90608F90\
6F8F906F8F906F909F6F8F906F90906F8F907090906F8F8F7090907090907F9F90
7090907F9F907F9F90809F9F809F9080A09F809F908FA09F8FA09F90A0A090A0A0403F3F40303F\
4F3F40503F405F404F5F404F6F40506F4050704F5F704F5F7F50607F4F6080506F
805060805F6F805F6F8F5F708F5F6F8F60708F60709060709060709F6F7F906F7F9F6F7F9F6F7F\
A07080A07080A07F80A07080AF7F8FA07F80AF7F8FAF7F80AF808FAF7F8FB0808F
AF808FB0808FB0808FBF8F90B0808FBF8F90B08F8FBF8F90B08F8FBF8F90BF8F8FBF9090BF8F8F\
C09090BF9090C09090BF9090C09090C09090C09F90C0908FC09090C0908FC09F90
C0908FC09F90C0908FC09F90C0908FC09F8FC09F8FCF9F8FC09F8FCF9F8FC09F8FCF9F8FC09F8F\
CF9F8FC09F8FCFA08FC09F80CF9F8FC09F80CF9F8FC09F80C09F80C09F80C09F80
C0907FC09F7FC0907FC09F7FBF9070C09F7FBF9070BF9F70BF9070BF9070B0906FB0906FAF8F60\
AF8F60AF805FAF805FA07F50A080509F7F4F9F7F50907040806F40705F3F6F503F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F4F6F3F4F703F5F7F\
3F5F7F4060804060804F6F8F4F6F8F4F70904F708F507F90507F905F7F90507F90
5F809F5F80905F809F5F809060809F5F809060809F608090608F9F608F906F8F9F608F906F909F\
6F8F906F909F6F8F9070909F6F90907090907090907F9F9F7090907F9F9F7F9F90
809F9F7F9F9F80A09F809F9F8FA09F80A09F8FA09F8FA09F403F3F403F3F4F3F40503F405F404F\
5F404F6040506F4050704F5F704F5F7F4F607F4F6080506F80506F805F6F805F6F
8F5F708F5F709060708F6070906F7F9060709F6F7F906F7F9F70809F6F7FA070809F7080A07F8F\
A07080AF7F8FA07F8FAF808FAF808FB08090AF808FB08F90B08090B08F90B08090
BF8F90B08F90BF8F90B08F90BF9090BF8F90BF9090BF8F90C09090BF9090C09090BF9090C09F9F\
C09090C09F9FC09090CF9F9FC09F90CF9F9FC09F90CF9F9FC09F90CF9F90C09F90
CF9F90C09F90CF9F90C09F90CFA090CF9F90CFA090CF9F8FCFA090CF9F90CFA090CF9F8FCFA090\
CFA08FCFA090CFA08FCFA090CFA08FCFA08FCF9F8FCFA08FCF9F80CFA08FC09F80
CFA080C09F80C09F80C09F7FC09F80C09F7FC09F7FBF9070C09F7FBF9070BF9F70B0906FB0906F\
AF8F60AF8F60A0805FAF8F5FA08050A080509F7F4F9F7F508F70407F60406F503F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF304F6F\
3F5F7F3F5F7F406080406080406F80406F804F708F4F708F507F9050708F507F90
507F905F80905F7F905F80905F80906080905F8090608090608090608F9F6080906F8F9F608F90\
6F909F6F8F906F909F6F8F9070909F70909070909F7090907F9F9F7090907F9F9F
7F9F90809F9F7F9F9F80A09F809F9F8FA0A03F303F4F3F404F3040503F4F503F4F604050603F50\
6F405F6F405F704F60704F607F50607F506080506F80506F8F5F70805F6F8F6070
8F5F7090607F8F6070906F7F906F7F9F6F80906F7F9F70809F7080A070809F7080A07F8FA07F8F\
AF7F8FAF7F8FAF8090AF808FB08090AF808FB08090B08090B08F90B08090B08F90
B08F90BF8F90B08F90BF8F90BF8F90BF9090BF8F90BF9090BF9090C0909FBF9090C0909FC09090\
C09F9FC09090C09F9FC09F90CF9F9FC09F90CF9F9FC09F90CF9F9FC09F90CF9F90
C09F90CFA09FC09F90CFA090CF9F90CFA090CF9F90CFA090CF9F90CFA090CF9F8FCFA090CFA08F\
CFA090CFA08FCFA090CFA08FCFA090CFA08FCFA08FCF9F8FCFA08FC09F80CFA080
C09F80C0A080C09F7FC09F80C09F7FC09F7FBF9F70C09F7FBF9070BF9F7FBF9070BF9070B0906F\
B0906FAF8F60AF8F60A0805FA0805F9F7F509F8050907F4F90704F7F6040705F3F
5F4F30CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
30506F3F5F7F3F5F7F4060803F607F406F80406F804F708F4F6F8F50709050708F
5F7F90507F905F809F5F80906080905F8090608F9F608090608F9F608F906F8F9F6F8F906F909F\
6F8F9070909F6F909F70909F70909070909F70909F7F9F9F7090907F9F9F7F9F9F
809F9F7F9F9F80A0A03F303F403F404F3040503F4F503F4F603F50603F506F405F6F405F704F5F\
70405F7F4F607F4F6080506F7F506F805F70805F708F60708F5F7090607F90607F
906F7F906F7F9F6F809F6F809F70809F7080A0708FA07080A07F8FA07F8FAF8090AF7F8FAF8090\
AF808FB08090AF8090B08F90B08090B08F9FB08F90BF8F9FB08F90BF8F9FB08F90
BF909FBF8F90BF909FBF9090BF909FBF9090C0909FBF909FC09F9FC0909FC09F9FC09F9FC09F9F\
C09F9FCF9F9FC09F9FCFA09FCF9F9FCFA09FCF9F9FCFA09FCF9F9FCFA09FCFA09F
CFA09FCFA090CFA09FCFA090CFA09FCFA090D0A09FCFA090D0A09FCFA090D0AF9FCFA090D0AF90\
CFA090D0AF90CFA090D0AF90CFA08FD0AF90CFA08FCFA08FCFA08FCFA08FC0A080
CFA08FC09F80CFA080C09F80C0A080C09F7FC0A080BF9F7FC09F7FBF9F70C09F7FBF9070BF9070\
B0906FB0906FAF8F60AF8F60A0805FA0805F9F80509F8050907F4F8F704F70603F
6F503FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C03F5F7F305F703F607F3F607F4060804060804F6F804F6F8050708F
4F708F507F90507F8F5F8090507F8F5F80905F80906080905F8090608F906080906F8F90608F90\
6F8F9F6F8F9070909F6F8F9070909F70909070909F70909070909F70909F7F9F9F
7F9F9F3F303F40303F4F30405030405F3F4F5F304F603F50603F506F405F6F405F704F60704F5F\
7F4F607F506080506F7F506F805F70805F708F607F8F607F90607F90607F906F80
906F7F9F70809F6F809F708F9F7080A07F8FA0708FA07F8FA07F8FAF7F90AF7F8FAF8090AF808F\
AF8090AF8090B08F90AF8090B08F9FB08F90BF8F9FB08F90BF8F9FB08F90BF909F
B08F90BF909FBF9090BF909FBF9090C0909FBF909FC09F9FBF909FC09F9FC09F9FC09F9FC09F9F\
CF9F9FC09F9FCFA09FC09F9FCFA09FCF9F9FCFA09FCF9F9FCFA09FCFA09FCFA09F
CFA09FCFA09FCFA09FD0A09FCFA090CFA09FCFA090D0AF9FCFA090D0A09FCFA090D0AF9FCFA090\
D0AF90CFA090D0AF90CFA090CFAF90CFA08FCFA090CFA08FCFA08FC0A08FCFA08F
C09F80C0A080C09F80C0A080C09F7FC0A080BF9F7FC09F7FBF9F70BF9F7FBF9F70BF9F70B0906F\
B0906FAF8F60AF9060A08F5FA08F5F9F8050A0805F907F4F9F7F4F8F70407F6F40
605030CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F5F7F3F5F7F3F607F3F607F406F80406F804F708F4F6F80
507F8F507F8F507F90507F8F5F80905F7F905F80905F8090608090608090608F906080906F8F9F\
6F8F9F6F909F6F8F9F70909F70909F70909F7090907F9F9F7F909F7F9F9F3F303F
4F3F404F30405F3F4F5F304F603F50603F506F405F6F405F704F60704F60704F60704F607F506F\
7F506F805F70805F708F607F8F607F906F8090607F906F80906F809F708F9F7080
A0708F9F708FA07F8FA07F8FA07F90A07F8FAF8090A07F90AF8090AF8090AF8090AF8090B08F9F\
B08F90B08F9FB08F9FBF909FB08F9FBF909FB08F9FBF909FBF909FBF909FBF909F
C09FA0BF909FC09FA0BF909FC09FA0C09F9FC09FA0C09F9FCF9FA0C09F9FCFA0A0C09F9FCFA0A0\
CFA0A0CFA0A0CFA09FD0A0A0CFA0A0D0A0A0CFA0A0D0AFA0CFA0A0D0AFA0CFA09F
D0AFA0D0A09FD0AFA0CFA09FD0AFA0D0AF9FD0AF9FCFA09FD0AF9FD0AF9FD0AF9FCFAF90D0AF9F\
D0AF90D0AF9FCFAF90D0AF9FD0AF90D0AF90CFA090D0AF90CFA08FCFAF8FCFA08F
CFAF8FC0A080CFA08FC0A080CFA080C0A080C0A080C09F7FC0A080BF9F7FC0A07FBF9F70BF9F70\
B0906FB0906FAF9060AF906FA08F60A08F5F9F8050A0805F9F8050907F4F7F6F40
6F5F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3050703F5F7F3F607F406F804060804F6F80
4F6F8050708F4F708F507F8F507F8F5F7F90507F8F5F80905F7F8F6080905F8090608F90608F90\
6F8F9F6F8F906F909F6F8F9070909F6F8F9070909F70909F3F3F3F40303F4F3040
5030405F3F4F5F304F603F50603F506F405F6F405F704F60704F607F506F704F607F506F7F506F\
805F70805F708F607F8F607F906F80906F80906F80906F809F708F9F708F9F708F
9F708FA07F909F708FA07F90A07F8FAF8090A07F90AF8090AF8090AF809FAF8090B08F9FAF8090\
B08F9FB08F9FB08F9FB08F9FBF90A0B08F9FBF909FBF909FBF90A0BF909FBF90A0
BF909FC09FA0BF909FC09FA0C09F9FC09FA0C09F9FC09FA0C09F9FCFA0A0C09F9FCFA0A0CFA0A0\
CFA0A0CFA0A0CFA0A0CFA0A0D0AFA0CFA0A0D0AFA0CFA0A0D0AFA0CFA09FD0AFA0
CFA09FD0AFA0CFA09FD0AFA0CFA09FD0AF9FCFA09FD0AF9FCFAF9FD0AF9FCFA090D0AF9FCFAF90\
D0AF9FCFAF90D0AF9FCFAF90D0AF9FCFAF90D0AF90CFA08FCFAF90CFA08FCFAF8F
C0A08FCFAF8FC0A080C0A080C0A080C0A080C09F7FC0A080BF9F7FC0A07FBF9F7FBF9F7FBF9F70\
BF9F70B0906FB0906FAF8F60AF9060A08F5FA08F5F9F80509F8050907F4F7F6F40
605F3FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF304F6F3F5F703F5F70406080406F80
4F708F4F6F80507F8F50708F507F8F507F8F5F7F905F7F8F5F80905F8090608F906080906F8F9F\
608F906F909F6F8F9F70909F6F8F9F70909F3F303F403F404F3040503F4F503F4F
5F3F50603F506F405F6F405F6F4F606F4F6070506F704F6F7F506F7F506F805F70805F708F607F\
8F607F906F808F6F80906F80906F809F708F9F708F9F708F9F708FA07F909F7F8F
A07F90A07F90AF8090A07F90AF8090A08090AF809FAF8090B08F9FAF8090B08F9FB08F9FB08F9F\
B08F9FBF90A0B0909FBF90A0BF909FBF9FA0BF90A0C09FA0BF90A0C09FA0BF9FA0
C09FA0C09FA0C09FA0C09FA0C0A0A0C09FA0CFA0A0C0A0A0CFA0A0CFA0A0CFA0AFCFA0A0CFA0AF\
CFA0A0D0AFAFCFA0A0D0AFAFD0A0A0D0AFAFD0AFA0D0AFA0D0AFA0D0AFA0D0AFA0
D0AFA0D0AFA0D0AFA0D0AFA0D0AFA0D0AF9FD0AFA0D0AF9FD0AFA0D0AF9FD0B0A0D0AF9FD0AF9F\
D0AF9FD0B09FD0AF9FD0B09FD0AF90D0B09FCFAF90D0AF90CFAF90D0AF90CFAF8F
CFAF8FCFA08FCFAF8FC0A080CFAF8FC0A080C0AF80C0A080C0A080BFA07FC0A080BFA07FBFA07F\
B09F70BF9F70B0906FB0906FAF8F60AF9060A08F5FA08F5F9F80509080507F6F40
6F5F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F506F3F506F405F70
405F7F4F6F804F6F8050708F4F708F507F8F507F8F5F7F90507F8F5F80905F8090608090608090\
608F9F608F906F8F9F6F8F9070909F3F303F4F30404F3040503F4F503F4F604050
603F506F405F6F4060704F606F4F6070506F70506F7F5F707F5070805F7F805F7F8F607F8F607F\
8F6F808F6F80906F8F906F809F708F9F708F9F7F909F708F9F7F909F7F8FA07F90
A07F90A07F90A07F90AF809FA08090AF809FAF8090AF8F9FAF809FB08F9FAF8F9FB08F9FB08F9F\
B090A0B0909FBF90A0B0909FBF90A0BF90A0C09FA0BF90A0C09FA0BF9FA0C09FA0
BF9FA0C09FA0C09FA0C0A0AFC09FA0CFA0A0C0A0A0CFA0AFC0A0A0CFA0AFCFA0A0CFAFAFCFA0A0\
D0AFAFCFA0A0D0AFAFD0A0A0D0AFAFD0AFA0D0AFAFD0AFA0D0AFA0D0AFA0D0AFA0
CFAFA0D0AFA0D0AFA0D0AFA0CFAF9FD0AFA0D0AF9FD0AFA0CFAF9FD0AF9FD0AF9FD0B09FCFAF9F\
D0AF9FD0AF9FD0B09FD0AF90D0B09FD0AF90D0B09FCFAF90CFAF90CFAF8FCFAF90
CFA08FCFAF8FC0A08FCFAF8FC0A080C0A080C0A080C0A080BFA07FC0A080BFA07FC0A07FBF9F70\
BF9F70B09F6FB09F70AF9060AF9060A08F5FA08F5F9F80509F8050907F4F807040
605F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F4F603F4F6F
405F70405F704F607F406F804F708F4F708F507F8F50708F5F7F905F7F8F5F80905F8090608090\
6080906F8F9F6F8F90403F3F40303F4F3F404F3F405F40505F405060405F60405F
6F4F606F4F6070506F70506F7F5F707F5070805F7F805F7F8F60808F60808F6F808F6080906F8F\
906F8F90708F90708F9F7F909F7090A07F909F7F90A07F90A07F90A0809FA07F90
AF809FA0809FAF809FAF809FAF8F9FAF8F9FB08F9FAF8F9FB090A0B08F9FB090A0B090A0BF90A0\
BF90A0BF9FA0BF90A0C09FAFBF9FA0C09FAFBF9FA0C0A0AFC09FAFC0A0AFC09FA0
CFA0AFC0A0AFCFA0AFC0A0AFCFA0AFCFA0AFCFAFAFCFA0AFD0AFAFCFA0AFD0AFAFCFA0AFD0AFAF\
D0AFAFD0AFAFD0AFAFDFB0AFD0AFAFDFB0AFD0AFAFDFB0AFD0AFAFDFB0AFD0AFA0
DFB0AFD0AFA0D0B0A0D0AFA0D0B0A0D0AFA0D0B0A0D0AFA0DFB0A0D0B0A0D0B0A0D0AF9FDFB0A0\
D0B09FD0B0A0D0B09FDFB0A0D0B09FD0B09FD0AF90D0B09FD0AF90D0B090CFAF90
D0B090CFAF8FCFAF90CFAF8FCFAF8FC0AF8FCFAF8FC0A080CFAF80C0A080C0AF80C0A07FC0A080\
BFA07FBFA070B09F6FB09F6FAF9060AF9060A08F5FA08F5F9F80509080507F6F40
6F6040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
3040603F506F3F506F40607F40607F4F70804F6F804F708F4F7080507F8F50708F5F7F8F5F7F8F\
6080905F8090403F3F403F404F3F4F4F3F4F5F40505F4050604F5F60405F6F4F6F
704F6F70507070506F7F5F707F5F7080607F805F7F8F60808060808F6F808F6080906F8F906F8F\
90708F90708F9F70909F708F9F7F909F7F90A07F90A07F90A0809FA07F90A0809F
A0809FAF809FAF809FAF8F9FAF8F9FB08FA0AF8F9FB08FA0B08FA0B090A0B08FA0BF90A0B090A0\
BF9FAFBF90A0BF9FAFBF9FA0C09FAFBF9FA0C09FAFC09FAFC0A0AFC09FAFC0A0AF
C0A0AFCFA0AFC0A0AFCFA0AFCFA0AFCFAFAFCFA0AFD0AFAFCFA0AFD0AFAFCFA0AFD0AFAFD0AFAF\
DFAFAFD0AFAFDFAFAFD0AFAFDFB0AFDFAFAFDFB0AFDFAFAFDFB0AFD0AFA0DFB0AF
D0AFA0DFB0AFD0AFA0DFB0A0D0AFA0DFB0A0D0AFA0D0B0A0D0B0A0D0B0A0D0AF9FD0B0A0D0B09F\
D0B0A0D0B09FD0B09FD0B09FD0B09FD0B09FD0B09FD0B090D0B09FCFAF90D0B090
CFAF90CFB090CFAF8FCFAF8FC0AF8FCFAF8FC0A080C0AF8FC0A080C0AF80BFA07FC0A080BFA07F\
BFA07FB09F70B09F70AF906FAF906FA08F60A08F609F8F5F9F8F5F90805080704F
605F3FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3F40603F506F405F7040607F4F6F80406F804F70804F708050708F50708F5F7F8F\
3F303F4F3F404F3F4050404F50404F5F4F5F5F405F604F606F4F6070506F705070
7F50707F5070805F7F805F7F80608080607F8F6F808F60808F6F8F8F6F8090708F906F8F907090\
90708F9F7F909F7F90A07F9F9F7F90A0809FA0809FA0809FA0809FAF8FA0AF809F
AF8FA0AF809FB08FA0AF8FA0B08FA0B08FA0B090A0B090A0BF90AFB090A0BF9FAFBF90A0BF9FAF\
BF9FAFC09FAFBF9FAFC0A0AFC09FAFC0A0B0C0A0AFCFA0B0C0A0AFCFA0B0CFA0AF
CFA0B0CFA0AFD0AFB0CFA0AFD0AFB0CFA0AFD0AFB0D0AFAFD0AFB0D0AFAFDFAFB0D0AFAFDFB0B0\
D0AFAFDFB0B0DFAFAFDFB0B0DFAFAFE0B0B0DFB0AFE0B0AFDFB0AFE0B0AFDFB0AF
DFB0AFDFB0A0DFB0AFDFB0A0DFB0AFD0B0A0DFB0AFD0B0A0DFB0A0D0B0A0DFB0A0D0B0A0DFB0A0\
D0B0A0DFBFA0D0B09FDFB0A0D0B09FDFBFA0D0B09FD0B09FD0B090D0B09FD0B090
D0B090CFAF90D0B090CFAF8FCFB090C0AF8FCFAF8FC0AF80C0AF8FC0AF80C0AF80C0A080C0AF80\
BFA07FBFA07FB09F70B09F6FAF9060AF906FA08F609F8F5F9080509F8F507F704F
6F6040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F4F603F5F7040607F40607F406F80406F803F303F3F303F\
4F3F404F3F4F504050503F505F405F5F405F6F4F606F4F607050707050707F5F7F
7F50707F5F7F7F5F7F806080805F7F8F60808060808F6F8F8F6F80906F8F906F8F90709090708F\
9F7F909F70909F7F9F9F7F90A0809FA0809FA0809FA0809FAF80A0A0809FAF8FA0
AF809FAF8FA0AF8FA0B090A0AF8FA0B090A0B090A0BF90AFB090A0BF9FAFBF90AFBF9FAFBF9FAF\
BF9FAFBF9FAFC0A0AFBF9FAFC0A0B0C0A0AFCFA0B0C0A0AFCFA0B0CFA0AFCFA0B0
CFA0AFCFAFAFCFA0AFD0AFB0CFA0AFD0AFB0D0AFAFD0AFB0D0AFAFD0AFB0D0AFAFDFB0B0D0AFAF\
DFB0B0DFAFAFDFB0B0DFAFAFE0B0B0DFB0AFE0B0B0DFB0AFE0B0AFDFB0AFE0B0AF
DFB0AFDFB0AFDFB0AFDFB0AFDFB0A0DFB0AFDFB0A0DFB0AFD0B0A0DFB0A0D0B0A0DFB0A0D0B0A0\
DFB0A0D0B0A0DFBFA0D0B09FD0B0A0D0B09FD0BFA0D0B09FD0B09FD0B090D0B09F
CFAF90CFB090CFAF8FCFB090C0AF8FCFAF8FC0AF8FC0AF8FC0AF80C0AF80BFA080C0AF80BFA07F\
BFA07FB09F70B09F70AF906FAF906FA09060A090609F8F5F9F8F5F90805080704F
605F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF303F503F506F40607F403F404030404F3F4F\
4F3F4F5040505F405F604F606040606F4F6F6F4F6F7050707050707F5F7F7F5F7F
8060808060808060808060808F6F8F8F6F8F8F6F8F8F6F8F90709090708F9F70909070909F7F9F\
9F7F90A0809F9F7F9FA080A0A0809FAF8FA0A080A0AF8FA0AF8FA0AF8FA0AF8FA0
B090AFB08FA0B090AFB090A0BF9FAFB090AFBF9FAFB090AFBF9FB0BF9FAFBF9FB0BF9FAFC0A0B0\
C09FAFC0A0B0C0A0AFC0A0B0C0A0B0CFA0B0C0A0B0CFAFB0CFA0B0CFAFB0CFA0B0
D0AFB0D0AFB0D0AFB0D0AFB0DFAFB0D0AFB0DFB0B0D0AFB0DFB0B0DFAFB0DFB0B0DFAFB0E0B0B0\
DFB0B0E0B0B0DFB0B0E0BFB0E0B0B0E0BFB0E0B0B0EFBFB0E0B0AFE0BFB0E0B0AF
E0BFB0E0B0AFE0BFAFDFB0AFE0BFAFDFB0AFE0BFAFDFB0AFE0BFAFDFBFA0DFBFAFDFB0A0DFBFAF\
DFBFA0DFBFA0DFBFA0DFBFA0DFBFA0DFBFA0D0B09FDFBFA0D0B09FD0BF9FD0B090
D0BF9FCFB090D0B090CFB090CFB090CFAF8FCFB08FC0AF8FCFB08FC0AF80C0AF80BFAF80C0AF80\
BFA07FBFA07FB09F70B09F70AF906FA0906F9F8F60A090609F8F5F908F5F7F704F
6F6F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF3F303F4030404030404F3F4F503F50\
5F3F5F5F3F5F6040606040606F4F6F6F4F6F7050707050707F5F7F7F5F7F806080
8060808060808060808F6F8F8F6F8F8F6F8F8F6F8F90709090708F9070909070909F7F9F9F7F90\
A0809F9F7F9FA080A0A0809FAF8FA0A080A0AF8FA0AF8FA0AF8FAFAF8FA0B090AF
AF8FA0B090AFB090A0B090AFB090AFBF9FAFB090AFBF9FAFBF9FAFBF9FB0BF9FAFBF9FB0BF9FAF\
C0A0B0BF9FAFC0A0B0C0A0B0CFA0B0C0A0B0CFAFB0CFA0B0CFAFB0CFA0B0D0AFB0
CFAFB0D0AFB0D0AFB0D0AFB0D0AFB0DFB0B0D0AFB0DFB0B0DFAFB0DFB0B0DFAFB0E0B0B0DFB0B0\
E0B0B0DFB0B0E0B0B0E0B0B0EFBFB0E0B0B0EFBFB0E0B0B0EFBFB0E0B0AFE0BFB0
E0B0AFE0BFB0E0B0AFE0BFAFE0B0AFE0BFAFDFB0AFE0BFAFDFBFAFE0BFAFDFB0A0E0BFAFDFBFA0\
DFBFAFDFB0A0DFBFA0DFBFA0DFBFA0D0B09FD0BFA0D0B09FD0BF9FCFB09FD0B09F
CFB090D0B09FCFB090CFB090CFAF8FCFB090C0AF8FC0AF8FC0AF80C0AF8FBFAF80C0AF80BFA07F\
BFAF7FB09F70B0A070AF9F6FAF9F6FA09060A090609F8F5F9F8F5F908050807F4F
6F603FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F3F3F303F403F4F4F304F503F50503F505F3F5F\
5F3F6060406F60406F6F4F706F4F7070507F70507F7F5F7F7F5F7F806080806080
806F8F80608F8F6F8F8F6F8F8F70908F6F8F9070909070909F7F9F9070909F7F9F9F7F9FA0809F\
9F809FA080A0A080A0AF8FA0A08FA0AF8FAFAF8FA0B090AFAF8FAFB090AFB090AF
B090AFB090AFBF9FB0B090AFBF9FB0B09FAFBF9FB0BF9FAFBF9FB0BF9FAFC0A0B0BFA0B0C0A0B0\
C0A0B0C0A0B0C0A0B0CFAFB0C0A0B0CFAFBFCFAFB0CFAFBFCFAFB0D0AFBFD0AFB0
D0AFBFD0AFB0DFB0BFD0AFB0DFB0B0DFAFB0DFB0BFDFB0B0E0B0B0DFB0B0E0B0BFE0B0B0E0BFB0\
E0B0B0EFBFBFE0B0B0EFBFB0E0B0B0EFBFBFEFBFB0EFBFB0EFBFB0EFBFB0EFBFB0
EFBFB0E0BFAFEFBFB0E0BFAFEFBFB0E0BFAFEFBFB0E0BFAFE0BFAFE0BFAFE0BFAFDFBFAFE0BFAF\
DFBFA0E0BFAFDFBFA0DFBFAFD0BFA0DFBFA0D0BFA0D0BFA0D0B09FD0BFA0D0B09F
D0BF9FCFB090D0B09FCFB090CFB090C0B08FCFB090C0AF8FC0B08FC0AF80C0B08FBFAF80C0AF80\
BFA07FBFA07FB0A070AF9F70A09F6FA09F6F9F90609F9060908F5F908F5F807F4F
6F6F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F303F3F303F4F304F4F304F50305050305F5F3F605F3F60\
60406F60406F6F4F706F4F6F70507070507F7F5F7F7F5F7F806080806080806F8F
80608F8F6F8F8F6F8F8F70908F6F8F907090907090907F9F9070909F7F9F9F7F9FA080A09F809F\
A080A0A080A0AF8FAFA08FA0AF8FAFAF8FA0AF90AFAF8FAFB090AFAF90AFB090AF
B090AFB090AFB090AFBF9FB0B090AFBF9FB0BF9FAFBF9FB0BF9FAFBFA0B0BF9FB0C0A0B0BFA0B0\
C0A0B0C0A0B0C0AFBFC0A0B0CFAFBFC0AFB0CFAFBFCFAFB0CFAFBFCFAFB0D0AFBF
CFAFB0D0B0BFD0AFB0DFB0BFD0AFB0DFB0B0DFB0B0DFB0BFDFB0B0E0B0B0DFB0B0E0BFBFE0B0B0\
E0BFB0E0B0B0EFBFBFE0B0B0EFBFB0EFBFB0EFBFBFEFBFB0EFBFB0EFBFB0EFBFB0
E0BFB0EFBFB0E0BFAFEFBFB0E0BFAFE0BFB0E0BFAFE0BFAFDFBFAFE0BFAFDFBFAFE0BFAFDFBFA0\
DFBFAFDFBFA0DFBFAFD0BFA0DFBFA0D0BFA0DFBFA0D0B09FD0BFA0CFB09FD0BF9F
CFB090CFB09FCFB090CFB090C0B08FCFB090C0AF8FC0B08FC0AF80C0AF8FBFAF80C0AF80BFAF7F\
BFAF7FB0A07FB0A070AF9F6FAF9F6FA09060A090609F8F5F908F5F908050807F4F
5F5F3FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF3F3F3F3F30404F3F4F4F30505F3F5F503F5F5F40605F4060604F6F\
604F6F6F4F706F4F70705F7F705F7F7F5F807F5F8080608F80608F8F6F8F806F8F
8F70908F6F909070908F7090907F9F907F9F9F7F9F907F9F9F80A09F80A0A080A0A080A0A08FAF\
A08FA0AF8FAFA08FAFAF90AFAF8FAFB090AFAF90AFB090B0B090AFB09FB0B090AF
BF9FB0B09FB0BF9FB0B09FB0BFA0B0BF9FB0BFA0B0BF9FB0C0A0BFC0A0B0C0A0BFC0A0B0C0AFBF\
C0A0B0CFAFBFC0AFB0CFAFBFCFAFBFCFAFBFCFAFBFD0B0BFD0AFBFD0B0BFD0AFBF
DFB0BFD0B0BFDFB0BFDFB0BFDFB0BFDFB0BFE0BFBFDFB0BFE0BFBFE0B0BFE0BFBFE0B0BFEFBFBF\
EFBFBFEFBFBFEFBFB0F0C0BFEFBFBFF0C0BFEFBFB0F0C0BFEFBFB0F0C0BFEFBFB0
EFC0BFEFBFB0EFC0B0E0BFB0EFC0B0E0BFB0EFC0B0E0BFAFE0C0B0E0BFAFE0C0AFDFBFAFE0C0AF\
DFBFAFDFC0AFDFBFA0DFC0AFDFBFA0DFC0AFD0BFA0DFC0A0D0BFA0D0BFA0CFBF9F
D0BF9FCFB09FCFBF9FCFB090CFBF90CFB090CFB090C0B08FC0B08FC0B08FC0B08FBFAF80C0B08F\
BFAF80BFAF7FB0A070B0A070AF9F6FA09F6F9F90609F9060908F5F908F5F70704F
5F5F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C03F303F3F30404F304F4F304F503F5F503F5F5F40605F3F6060406F60406F\
6F4F706F4F7070507F70507F7F5F807F5F8080608F80608F8F6F90806F8F8F7090
8F6F9090709F8F7090907F9F907F9F9F7F9F907F9F9F80A09F80A0A080A09F80A0A08FA0A080A0\
AF8FAFA08FA0AF8FAFAF8FAFAF90B0AF90AFB090B0AF90AFB09FB0B090AFB09FB0
B090B0BF9FB0B09FB0BF9FB0BF9FB0BFA0B0BF9FB0C0A0B0BFA0B0C0A0BFBFA0B0C0A0BFC0A0B0\
CFAFBFC0A0B0CFAFBFCFAFBFCFAFBFCFAFBFCFB0BFCFAFBFD0B0BFCFAFBFD0B0BF
D0B0BFDFB0BFD0B0BFDFB0BFDFB0BFDFBFBFDFB0BFE0BFBFDFB0BFE0BFBFE0B0BFE0BFBFE0BFBF\
EFBFBFE0BFBFEFBFBFEFBFBFF0C0BFEFBFB0F0C0BFEFBFB0F0C0BFEFBFB0EFC0B0
EFBFB0EFC0B0E0BFB0EFC0B0E0BFB0EFC0B0E0BFAFE0C0B0E0BFAFE0C0B0DFBFAFE0C0AFDFBFAF\
DFC0AFDFBFA0DFC0AFD0BFA0DFC0AFD0BFA0D0C0A0D0BFA0D0BFA0CFBF9FD0BF9F
CFBF9FCFBF9FCFB090CFBF90C0B090CFB090C0B08FC0B08FC0B08FC0B08FBFAF80C0B08FBFAF80\
BFAF80B0A07FB0A07FAF9F70AF9F70A0906FA090609F905F90905F8F8050707040
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3F3F3F3F30404F3F4F4F3050503F5F503F605F406F5F4060604F70604F706F5070\
6F5070705F7F705F7F7F5F807F5F80806F8F80608F8F6F90806F908F709F8F7090
90709F90709F907FA0907F9F9F7FA0907F9F9F80A09F80A0A080A09F80A0A08FAFA08FAFA08FAF\
A08FAFAF90B0AF90AFAF90B0AF90AFB09FB0B090B0B09FB0B090B0B09FB0B09FB0
BF9FB0B09FB0BFA0BFBF9FB0BFA0BFBFA0B0C0A0BFBFA0B0C0A0BFC0A0BFC0AFBFC0AFBFCFAFBF\
C0AFBFCFAFC0CFAFBFCFB0C0CFAFBFD0B0C0CFB0BFD0B0C0D0B0BFD0B0C0D0B0BF
DFB0C0D0B0BFDFBFC0DFB0BFDFBFC0DFB0BFE0BFC0E0BFBFE0BFC0E0BFBFEFBFC0E0BFBFEFC0BF\
EFBFBFEFC0BFEFC0BFF0C0BFEFC0BFF0C0BFF0C0BFF0C0BFEFC0BFF0C0BFEFC0B0
EFC0BFEFBFB0EFC0BFE0C0B0EFC0B0E0BFB0EFC0B0E0C0B0E0C0B0E0BFAFE0C0B0DFC0AFE0C0AF\
DFBFAFDFC0AFDFC0AFDFC0AFD0BFA0DFC0AFD0BFA0D0C0A0D0BFA0D0C0A0CFBF9F
D0BF9FCFBF9FCFBF9FCFB090CFBF90C0B090CFBF90C0B08FC0B090C0B08FC0B08FBFB080C0B08F\
B0AF80BFAF80B0A07FAFA070A09F6FA09F6F9F90609F9060908F5F8F8F50606F40
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C03F303F3F3040403F4F40304F503F5F503F5F5F406F5F406F604F6F604F706F50706F4F70\
70507F705F7F7F5F80705F807F608F7F608F806F90806F908F709F8F7090907F9F
8F709F907F9F907F9F9F7FA0907F9F9F80A09F80A09F80A09F80A0A08FAFA080A0A08FAFA08FAF\
AF8FAFA08FAFAF90B0AF90AFAF90B0AF90B0B09FB0AF90B0B09FB0B09FB0BF9FB0
B09FB0BF9FB0BF9FB0BFA0BFBF9FB0BFA0BFBFA0B0C0A0BFBFA0B0C0AFBFC0A0BFCFAFBFC0AFBF\
CFAFBFC0AFBFCFB0C0CFAFBFCFB0C0CFAFBFD0B0C0CFB0BFD0B0C0D0B0BFDFB0C0
D0B0BFDFB0C0DFB0BFDFBFC0DFB0BFDFBFC0DFBFBFE0BFC0E0BFBFE0BFBFE0BFBFEFC0BFE0BFBF\
EFC0BFEFC0BFEFC0BFEFC0BFF0C0BFEFC0BFF0CFBFEFC0BFF0C0BFEFC0B0EFC0BF
EFC0B0EFC0B0E0C0B0EFC0B0E0BFB0E0C0B0E0BFB0E0C0B0DFBFAFE0C0B0DFBFAFE0C0B0DFBFAF\
DFC0AFDFBFAFDFC0AFD0BFA0DFC0AFD0BFA0D0C0AFD0BFA0D0C0A0CFBF9FD0BFA0
CFBF9FCFBF9FCFBF90CFBF9FC0B090CFBF90C0B090C0B090BFB08FC0B08FBFB08FC0B08FBFAF80\
BFAF80B0A07FB0A07FA09F70A09F6F9F90609F9060908F5F908F5F7F7F4F5F603F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
3F3F3F3F3040403F4F4F3F5050405F5040605F406F5F406F604F70604F706F507F6F507F705F80\
705F807F608F7F608F806F90806F908F7090806F908F709F8F709F907F9F8F709F
907FA0907FA09F80A0907FA09F80AF9F80A09F8FAF9F80A0A08FAFA08FAFA08FAFA08FAFAF90B0\
AF90B0AF90B0AF90B0B09FBFAF9FB0B09FBFB09FB0B09FBFB09FB0BFA0BFB09FB0
BFA0BFBFA0BFBFA0BFBFA0BFC0AFBFBFA0BFC0AFBFC0A0BFC0AFC0C0AFBFCFAFC0C0AFBFCFB0C0\
CFAFC0CFB0C0CFB0C0D0B0C0CFB0C0D0B0C0D0B0C0D0BFC0D0B0C0DFBFC0D0B0C0
DFBFC0DFBFC0DFBFC0DFBFC0E0C0C0DFBFC0E0C0C0E0BFBFE0C0C0E0C0C0EFC0C0E0C0BFEFCFC0\
EFC0BFEFCFC0EFC0BFF0CFC0F0CFBFF0CFC0F0CFBFF0CFC0F0CFBFF0CFBFEFC0BF
F0CFBFEFC0BFEFCFBFEFC0B0EFCFBFE0C0B0EFC0B0E0C0B0E0C0B0E0C0B0E0C0B0DFC0AFE0C0B0\
DFC0AFDFC0B0DFC0AFDFC0AFDFC0AFDFC0AFD0C0A0DFC0AFD0C0A0D0C0A0CFBFA0
D0C0A0CFBF9FCFBF9FCFBF9FCFBF9FC0BF90CFBF90C0B090C0BF90C0B08FC0B08FBFB08FC0B08F\
B0AF80B0AF80AFA070AFA070A09F6FA09F6F9090609090608F8F5F7F7F4F505F3F
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C03F303F
3F3040403F4F403F4F4F405F50405F5F406F5F406F604F70604F706F507F6F507F705F80705F80\
7F608F7F608F7F6F907F6F8F806F90806F908F709F8F709F907F9F8F709F907FA0
907F9F9080A0907FA09F80A09F80A09F80AF9F80A0A08FAF9F8FAFA08FAFA08FAFA090B0A08FAF\
AF90B0AF90B0AF9FB0AF90B0B09FBFAF9FB0B09FBFB09FB0BFA0BFB09FB0BFA0BF
B0A0BFBFA0BFBFA0BFBFA0BFBFA0BFC0AFC0BFA0BFC0AFC0C0AFBFC0AFC0C0AFBFCFAFC0C0AFBF\
CFB0C0CFAFC0CFB0C0CFB0C0D0B0C0CFB0C0D0BFC0D0B0C0D0BFC0D0B0C0DFBFC0
D0BFC0DFBFC0DFBFBFDFC0C0DFBFC0E0C0C0DFBFBFE0C0C0E0C0BFE0C0C0E0C0BFEFC0C0E0C0BF\
EFCFC0EFC0BFEFCFC0EFCFBFF0CFC0EFCFBFF0CFC0F0CFBFF0CFBFEFCFBFEFCFBF
EFCFBFEFCFBFE0C0B0EFCFBFE0C0B0EFCFBFE0C0B0E0CFB0E0C0B0E0CFB0DFC0AFE0C0B0DFC0AF\
DFC0B0DFC0AFDFC0AFD0C0AFDFC0AFD0C0AFD0C0AFD0C0A0D0C0A0CFBFA0D0C0A0
CFBF9FCFC0A0CFBF9FCFBF9FC0BF90CFBF9FC0B090C0BF90BFB08FC0B090BFB08FBFB08FB0AF80\
B0AF80AFA07FAFA07FA09F70A09F709F90609090608F8F5F8F8F5F6F7040CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3040
403F4F403F5050405F5040605F4F6F5F406F604F7F604F7F6F507F6F507F705F80705F807F608F\
7F608F806F907F6F90806F90806F908F709F8F709F907FA08F7F9F907FA0907FA0
9080A0907FA09F80AF9F80A09F8FAF9F80AFA08FAFA08FAFA08FB0A08FAFAF90B0A090B0AF90B0\
AF90B0B09FBFAF9FB0B09FBFB09FBFB0A0BFB09FBFBFA0BFB09FBFBFA0BFBFA0BF
BFA0BFBFA0BFC0AFC0BFA0BFC0AFC0C0AFBFC0AFC0C0AFC0C0AFC0C0AFC0CFB0C0CFB0C0CFB0C0\
CFB0C0D0B0CFCFB0C0D0BFCFCFB0C0D0BFCFD0BFC0D0BFCFD0BFC0DFC0CFDFBFC0
DFC0C0DFBFC0E0C0C0DFC0C0E0C0C0DFC0C0E0CFC0E0C0C0E0CFC0E0C0C0EFCFC0EFCFC0EFCFC0\
EFCFC0F0CFC0EFCFC0F0D0C0EFCFBFF0D0C0F0D0C0F0D0C0F0CFBFF0D0C0EFCFBF
F0D0BFEFCFBFEFD0BFEFCFBFEFCFBFE0CFB0EFCFBFE0CFB0E0CFB0E0C0B0E0CFB0DFC0B0E0CFB0\
DFC0AFDFCFB0DFC0AFDFCFB0D0C0AFDFCFAFD0C0AFD0C0AFD0C0A0D0C0AFCFC0A0
D0C0A0CFBF9FCFC0A0CFBF9FCFBF9FC0BF90C0BF9FC0B090C0BF90BFB08FC0B08FBFB08FBFB08F\
B0AF7FB0AF7FA0A070A0A0709F9F6F9F9F6F9090609090608080506F7040CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF30303F403F4F
403F4F4F405F4F405F5F4F6F5F406F604F7F604F7F6F50806F507F705F80705F8070608F70608F\
7F608F7F608F806F90806F9080709F80709F8F7FA08F709F907FA08F7FA09080A0
907FA09080AF9080A09F80AF9F80A09F8FAF9F8FAFA08FB09F8FAFA090B0A08FB0AF90B0A090B0\
AF9FB0AF90B0B09FBFAF9FB0B09FBFB09FBFB0A0BFB09FBFBFA0BFB0A0BFBFA0C0
BFA0BFBFA0C0BFA0BFC0AFC0BFA0BFC0AFC0C0AFC0C0AFC0C0AFC0C0B0C0C0AFC0CFB0CFC0B0C0\
CFB0CFCFB0C0D0BFCFCFB0C0D0BFCFCFBFCFD0BFCFD0BFC0D0BFCFD0BFC0DFC0CF
D0BFC0DFC0C0DFC0C0E0C0C0DFC0C0E0C0C0E0C0C0E0CFC0E0C0C0E0CFC0E0CFC0EFCFC0E0CFC0\
EFCFC0EFCFC0F0D0C0EFCFBFF0D0C0F0D0BFF0D0C0F0D0BFF0D0C0EFCFBFF0D0C0
EFCFBFEFD0BFEFCFBFEFD0BFE0CFB0EFCFBFE0CFB0E0CFBFE0CFB0E0CFB0DFCFB0E0CFB0DFC0AF\
DFCFB0DFC0AFDFCFB0D0C0AFDFCFAFD0C0AFD0CFAFD0C0A0D0C0AFCFC0A0D0C0A0
CFBF9FCFBF9FC0BF9FCFBF9FC0BF90C0BF90BFB090C0BF90BFB08FBFB08FBFB08FBFB08FB0AF80\
B0AF80AFA07FA0A07F9F9F6F9F9F6F9090609090608F8F5F7F80505F603FCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30303F3F3F4F403F50
4F405F4F40605F4F6F5F4F7060507F60507F6F50806F5080705F8F705F8070608F70608F7F6F90\
7F6F90806F9F806F908F709F8F709F8F7FA08F7FA09080AF907FA09080AF9080A0
9F80AF9F80AF9F8FAF9F80AFA08FB09F8FAFA090B0A08FB0AF90B0A090B0AF90BFAF90B0AF9FBF\
AF9FBFB09FBFAF9FBFB0A0C0B0A0BFB0A0C0B0A0BFBFA0C0BFA0C0BFAFC0BFA0C0
C0AFC0BFAFC0C0AFC0BFAFC0C0B0CFC0AFC0C0B0CFC0AFC0CFB0CFC0B0CFCFB0CFCFB0CFCFBFCF\
CFB0CFD0BFCFCFBFCFD0BFD0D0BFCFD0BFD0D0BFCFDFC0D0D0C0CFDFC0CFDFC0CF
DFC0CFDFC0CFE0CFCFDFC0C0E0CFCFE0CFC0E0CFCFE0CFC0EFD0CFE0CFC0EFD0CFEFCFC0EFD0CF\
EFD0C0F0D0C0EFD0C0F0D0CFF0D0C0F0DFC0F0D0C0FFDFC0F0D0C0F0DFC0EFD0BF
F0DFC0EFD0BFEFD0C0EFD0BFEFD0BFEFD0BFEFD0BFE0CFB0EFD0BFE0CFB0E0D0BFDFCFB0E0D0BF\
DFCFB0DFCFB0DFCFB0DFCFB0D0CFAFDFCFB0D0C0AFD0CFAFD0C0A0D0C0AFCFC0A0
CFC0A0CFBF9FCFC0A0C0BF9FCFBF9FC0BF90C0BF90BFB090C0BF90BFB08FBFB08FB0B08FB0B08F\
AFAF80AFAF7FA0A070A0A0709F9F6F909F6F8F90608F9060707F4FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF3030303F3F4F3F304F403F5F
4F3F5F50406F50406F5F4F7F604F7F6F508060507F6F5F806F5F8070608F705F8F7F60907F6090\
806F907F6F9080709F80709F8F7FA08F709F907FA08F7FA09080AF907FA09080AF
9080AF9F8FAF9F80AF9F8FB09F8FAFA090B09F8FB0A090B0A090B0AF90BFA090B0AF9FBFAF9FBF\
B09FBFAF9FBFB0A0BFB09FBFB0A0C0B0A0BFB0A0C0B0A0C0BFAFC0B0A0C0BFAFC0
BFAFC0C0AFCFBFAFC0C0AFCFBFAFC0C0B0CFC0AFC0C0B0CFC0B0CFCFB0CFC0B0CFCFB0CFCFB0CF\
CFBFD0CFB0CFD0BFD0CFBFCFD0BFD0D0BFCFD0C0D0D0BFCFDFC0D0D0C0CFDFC0CF
DFC0CFDFCFCFDFC0CFE0CFCFDFCFCFE0CFCFE0CFC0E0CFCFE0CFC0EFD0CFE0CFC0EFD0C0EFD0C0\
EFD0CFEFD0C0F0D0C0EFD0C0F0DFC0F0D0C0F0DFC0F0D0C0F0DFC0EFD0BFF0DFC0
EFD0BFF0DFC0EFD0BFEFD0BFEFD0BFEFD0BFE0D0BFE0D0BFE0D0B0E0D0BFDFCFB0E0D0B0DFCFB0\
DFCFB0DFCFAFDFCFB0D0CFAFDFCFB0D0C0AFD0CFAFCFC0AFD0C0AFCFC0A0CFC0A0
C0BF9FCFC0A0C0BF9FC0BF9FC0BF90C0BF9FBFB090BFBF90BFB08FBFB090B0B08FB0B08FAFAF80\
AFAF80A0A070A0A0709F9F6F9F9F6F8F90608F9060808F5F6F704FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F4F3F304F40405F4F3F5F
50406F50406F5F4F7F604F7F6F50806050806F5F8F6F5F8070608F70608F7F6F907F60907F6F90\
7F6F9080709F80709F8F7FA08F7FA09080AF907FA09080AF9080AF9F8FAF9080AF
9F8FB09F8FAFA08FB09F8FB0A090B0A08FB0A090BFA090B0AF9FBFA090B0AF9FBFAF9FBFAF9FBF\
AF9FBFB0A0C0B0A0BFB0A0C0B0A0C0BFAFC0B0A0C0BFAFCFBFA0C0BFAFCFBFAFC0
C0AFCFBFAFC0C0B0CFC0AFCFC0B0CFC0B0CFCFB0D0C0B0CFCFB0D0C0B0CFCFBFD0CFBFD0CFBFD0\
CFBFD0D0BFD0CFBFD0D0C0D0D0BFD0D0C0DFD0C0D0DFC0D0D0C0D0DFCFD0DFC0D0
DFCFD0DFC0CFE0CFD0DFCFCFE0CFD0E0CFCFE0D0D0E0CFCFEFD0CFE0D0CFEFD0CFEFD0CFEFDFCF\
EFD0C0F0DFCFEFDFC0F0DFCFF0DFC0F0DFCFF0DFC0F0DFCFF0DFC0F0DFC0F0DFC0
F0DFC0EFDFBFF0DFC0EFDFBFEFDFC0E0D0BFEFDFBFE0D0BFE0D0BFE0D0B0E0D0BFDFCFB0DFD0B0\
DFCFB0DFD0B0D0CFB0DFCFB0D0CFAFD0CFB0D0C0AFD0CFAFCFC0A0CFC0AFCFC0A0
CFC0A0C0BF9FCFC0A0C0BF9FC0BF9FBFBF90C0BF9FBFB090BFBF90B0B08FBFBF8FB0B08FB0B080\
AFAF7FA0AF7F9FA0709FA070909F6F909F6F8F9060808F5FCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F304030304F403F5F403F5F4F4060
4F406F5F4F705F4F706050806050806F5F806F5F8070608F6F5F8F7060907060907F6F907F6F90\
80709F80709F8F7FA080709F8F7FA08F7FA09080AF9080AF9080AF9080AF9F8FB0
9080AF9F8FB09F8FB0A090B09F8FB0A090B0A090B0A090BFA090B0AF9FBFAF9FBFAF9FBFAF9FBF\
AF9FC0AF9FBFB0A0C0B0A0BFB0A0C0B0A0C0BFAFCFB0A0C0BFAFCFBFAFC0BFAFCF
BFAFCFBFAFCFBFAFCFC0B0CFBFAFCFC0B0CFC0B0CFC0B0D0C0B0CFCFBFD0C0B0D0CFBFD0CFBFD0\
CFBFD0CFBFD0D0C0DFCFBFD0D0C0DFD0BFD0D0C0DFD0C0D0DFC0D0D0C0D0DFCFD0
DFC0D0DFCFD0DFCFD0E0CFD0DFCFCFE0CFD0E0CFCFE0D0D0E0CFCFEFD0CFE0D0CFEFDFCFEFD0CF\
EFDFCFEFD0CFF0DFCFEFDFC0F0DFCFF0DFC0F0E0CFF0DFC0F0DFC0F0DFC0F0DFC0
EFDFBFEFDFC0EFD0BFEFDFC0E0D0BFE0D0BFE0D0BFE0D0BFDFD0B0E0D0BFDFCFB0DFD0BFDFCFB0\
DFCFB0D0CFB0DFCFB0D0C0AFD0CFB0CFC0AFD0CFAFCFC0A0CFC0AFCFC0A0CFC0A0
C0BF9FC0C09FC0BF9FC0BF9FBFBF90BFBF90BFB090BFBF90B0B08FB0B08FB0B08FB0B08FAFAF7F\
A0AF7F9FA0709FA070909F6F909F6F8F90608F90605F6F40CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30304F403F5F403F5F4F40604F406F
5F4F705F4F706050806050806F5F8F6F5F8F70608F70608F7F6F907060907F6F9F7F6F9080709F\
80709F8F7FA0807FA08F80AF8F7FAF9080AF9080AF9F8FB09080AF9F8FB09F8FB0
9F90B09F8FB0A090BF9F90B0A090BFA090BFA09FBFA090BFAF9FC0AF9FBFAF9FC0AF9FBFB0A0C0\
AFA0C0B0A0C0B0A0C0B0AFCFB0A0C0BFAFCFB0AFC0BFAFCFBFAFCFBFAFCFBFAFCF
C0B0D0BFB0CFC0B0D0C0B0CFC0BFD0C0B0D0CFBFD0C0B0D0CFBFDFCFBFD0CFBFDFCFBFD0D0C0DF\
CFBFDFD0C0DFCFC0DFD0C0DFD0C0DFD0C0DFD0C0DFDFCFDFDFC0DFDFCFDFDFCFD0
E0CFDFDFCFD0E0D0DFDFCFD0E0D0DFE0D0D0E0D0D0E0D0D0EFDFD0EFD0D0EFDFD0EFD0CFF0DFD0\
EFDFCFF0DFD0F0DFCFF0E0D0F0DFCFF0E0CFF0E0C0FFE0CFF0DFC0F0E0CFEFDFC0
F0E0C0EFDFC0EFDFC0EFDFBFEFDFC0E0D0BFE0DFBFE0D0BFE0DFBFDFD0BFE0D0BFDFCFB0DFD0BF\
DFCFB0DFD0B0D0CFB0D0CFB0D0CFB0D0CFB0CFC0AFD0CFAFCFC0AFCFC0AFC0C0A0
CFC0A0C0C09FC0C0A0BFBF9FC0C09FBFBF90BFBF90BFB090BFBF90B0B08FB0B08FAFB080AFB080\
A0AF7FA0AF7F9F9F709FA0708F906F8F9060808F5FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3030403F3F503F3F5F4F40604F4060504F6F
504F705F507F5F507F6F5F8F6F5F8F6F608F6F5F8F7060907060907F6F9F7F6F907F709F7F709F\
807FA08070A08F7FA08F7FA09080AF8F80AF9080B09080AF9F8FB0908FB09F8FB0
9F8FB09F90BF9F8FB0A090BF9F90B0A090BFA090BFA09FBFA09FBFAF9FC0A09FBFAF9FC0AF9FC0\
B0A0C0AFA0C0B0A0C0B0A0C0B0AFCFB0A0C0BFAFCFB0AFCFBFAFCFBFAFCFBFB0CF
BFAFCFC0B0D0BFB0CFC0B0D0C0B0D0C0BFD0C0B0D0CFBFDFC0BFD0CFBFDFC0BFD0CFBFDFCFBFDF\
D0C0DFCFBFDFD0C0DFD0C0DFD0C0E0D0C0DFD0CFDFD0C0DFDFCFDFD0CFDFDFCFDF
DFCFDFE0D0DFDFCFD0E0D0DFE0D0D0E0D0DFE0D0D0E0D0D0E0D0D0EFDFD0E0D0D0EFDFD0EFDFCF\
EFDFD0EFDFCFEFDFCFEFDFCFF0E0CFEFDFC0F0E0CFEFDFC0F0E0CFEFDFC0EFDFC0
EFDFC0EFDFC0E0D0BFEFDFC0E0D0BFE0DFBFE0D0BFE0D0BFDFD0BFE0D0BFDFCFB0DFD0BFD0CFB0\
DFD0B0D0CFB0D0CFB0D0CFAFD0CFB0CFC0AFCFCFAFCFC0AFCFC0AFC0C0A0C0C0A0
C0BF9FC0C0A0BFBF9FBFBF9FBFBF90BFBF9FB0B090BFBF90B0B08FB0BF90AFB08FAFB08FA0AF80\
A0AF7F9FA070909F708F906F8F90607F8F5F506F40CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF30303F3F3F503F3F5F404060404060504F6F504F70
5F507F5F507F6F5F8F6F5F8F6F608F6F608F706F907060907F6F9F7F6F9F8070A07F709F807FA0\
8070A08F7FAF8F7FAF9080AF8F80AF908FB09080B09F8FB0908FB09F90BF9F8FB0
9F90BF9F8FB0A090BFA090BFA09FBFA090BFAF9FC0A09FBFAF9FC0AF9FBFAFA0C0AF9FC0B0A0C0\
AFA0C0B0AFCFB0A0C0B0AFCFB0A0CFBFAFCFB0AFCFBFAFCFBFAFCFBFB0D0BFB0D0
C0B0D0BFB0D0C0BFD0C0B0D0C0BFDFC0B0D0CFBFDFC0BFDFCFBFDFCFBFDFCFC0DFCFBFDFD0C0E0\
CFC0DFD0C0E0D0C0DFD0CFE0D0C0E0DFCFE0D0CFE0DFCFE0D0CFDFDFD0E0DFCFDF
DFD0E0DFCFDFE0D0DFE0D0DFE0D0DFE0D0DFE0DFDFE0D0D0EFDFDFE0D0D0EFDFD0E0DFD0EFDFD0\
E0DFCFEFE0D0EFDFCFEFE0CFEFDFCFF0E0CFEFE0CFEFE0CFEFDFC0EFE0CFEFDFC0
EFE0C0E0DFC0EFDFC0E0DFBFE0DFC0E0D0BFE0DFC0DFD0BFE0D0BFDFD0BFDFD0BFDFD0B0DFD0BF\
D0CFB0D0D0B0D0CFB0D0CFB0CFCFAFD0CFB0CFCFAFCFCFAFC0C0A0CFCFAFC0C0A0
C0C0A0BFBF9FC0C0A0BFBF9FBFBF9FBFBF90BFBF9FB0BF90B0BF90B0B08FB0BF8FAFAF80A0AF80\
9FA07F9FA0708F9F6F8F9F6F808F607F8F5FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C030304F303050403F5F403F604F406F4F4F6F505070
5F507F605F80605F806F608F6F5F8F706090706090706F9F706F907F709F7F709F8070A08070A0\
807FA0807FA08F80AF8F7FAF9080B09080AF908FB09080B09F8FB0908FB09F90BF
9F8FB09F90BF9F90BFA090BF9F90BFA09FBFA090BFAF9FC0A09FBFAF9FC0AF9FC0AFA0C0AF9FC0\
B0A0CFAFA0C0B0AFCFB0A0C0B0AFCFB0AFCFBFAFCFB0AFCFBFB0D0BFAFCFBFB0D0
BFB0CFC0B0D0BFB0D0C0BFDFC0B0D0C0BFDFC0BFD0CFBFDFC0BFDFCFBFDFCFBFDFCFC0E0CFBFDF\
CFC0E0CFC0DFD0CFE0CFC0E0D0CFE0D0C0E0DFCFE0D0CFDFDFCFE0DFCFDFDFD0E0
DFCFDFDFD0DFDFD0DFE0D0DFDFD0D0E0D0DFDFD0D0E0DFDFDFD0D0E0DFD0E0DFD0E0DFD0E0DFCF\
E0DFD0E0DFCFEFE0D0E0DFCFEFE0CFE0E0CFEFE0CFE0DFC0EFE0CFE0DFC0EFE0C0
E0DFC0E0DFC0E0DFBFE0DFC0DFD0BFE0DFBFDFD0BFE0D0BFDFD0B0DFD0BFD0D0B0DFD0BFD0CFB0\
D0D0B0D0CFB0D0CFB0CFCFAFCFCFB0CFC0AFCFCFAFC0C0A0C0C0AFC0C0A0C0C0A0
BFBF9FC0C0A0BFBF9FBFC09FB0BF90BFBF90B0BF90B0BF90AFB08FB0B08FAFAF80A0AF809FA07F\
9FA0708F9F6F8F9F6F808F607F8F604F603FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F3F403F30503F3F5F403F604F406F4F406F504F7050507F
605F80605F806F60906F6090706F907060907F6F9F706F9F7F70A07F70A0807FA08070A08F7FAF\
807FAF8F80AF8F80AF908FB09080B0908FB0908FB09F90BF9F8FB09F90BF9F8FBF
A090BF9F90BFA09FC09F90BFA09FC0A09FC0AF9FC0A09FC0AFA0CFAF9FC0AFA0CFAFA0C0B0AFCF\
AFA0CFB0AFCFB0A0CFBFAFD0B0AFCFBFB0D0B0AFCFBFB0D0BFB0D0BFB0D0BFB0D0
C0BFDFC0B0D0C0BFDFC0B0D0C0BFDFC0BFDFCFBFDFC0BFDFCFC0E0CFC0DFCFC0E0CFC0E0D0CFE0\
CFC0E0D0CFE0D0C0E0D0CFEFD0CFE0D0CFEFD0CFE0DFD0EFD0CFE0DFD0E0D0CFE0
DFD0E0DFD0DFDFD0E0DFD0DFE0DFDFDFD0DFE0DFDFDFD0D0E0DFDFDFDFD0E0DFD0E0DFD0E0E0D0\
E0DFD0E0E0D0E0DFCFEFE0D0E0E0CFEFE0CFE0E0CFEFE0CFE0E0CFE0E0CFE0DFC0
E0E0CFE0DFC0E0DFC0DFDFC0E0DFC0DFDFBFDFDFC0DFD0BFDFDFBFDFD0BFDFD0BFD0D0B0DFD0BF\
D0D0B0D0D0B0CFCFB0D0D0B0CFCFB0CFCFB0CFCFAFCFCFAFC0C0AFC0CFAFC0C0A0
C0C0A0BFC0A0BFC0A0BFBF9FBFC09FB0BF90B0BF90B0B090B0BF90AFB08FAFB08F9FAF7F9FAF7F\
909F708F9F7080906080906070805FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF3030403F305F3F305F403F6040406F504F70504F705F5080
5F5080605F8F6F5F8F6F60906F6090706F9F706F9F7F709F7F6F9F7F70A07F70A0807FA0807FA0\
8F80AF807FAF8F80AF8F80AF908FB09080B0908FB0908FB09F90BF908FB09F90BF
9F90BFA090BF9F90BFA09FC0A090BFA09FC0A09FC0AF9FC0A09FC0AFA0CFAF9FC0AFA0CFAFA0CF\
B0AFCFAFA0CFB0AFCFB0AFCFB0AFD0B0AFCFBFB0D0B0AFD0BFB0D0BFB0D0BFB0D0
BFB0D0C0BFDFBFB0D0C0BFDFC0BFDFC0BFDFC0BFDFCFC0DFC0BFDFCFC0E0CFC0DFCFC0E0CFC0E0\
D0CFE0CFC0E0D0CFE0D0CFE0D0CFEFD0CFE0D0CFEFD0CFE0D0D0E0D0CFDFD0D0E0
D0D0DFDFD0E0D0D0DFDFD0DFD0D0DFDFDFDFD0D0D0DFDFDFDFDFD0DFDFDFDFDFD0DFDFD0DFDFD0\
E0E0D0DFDFCFE0E0D0DFE0CFE0E0CFE0E0CFE0E0CFE0E0C0E0E0CFDFDFC0E0E0C0
DFDFC0DFE0C0DFDFBFDFDFC0DFDFBFDFDFC0D0D0BFDFDFBFD0D0BFDFD0BFD0D0B0D0D0BFD0CFB0\
D0D0BFCFCFB0CFCFB0CFCFB0CFCFB0C0C0AFCFCFAFC0C0AFC0CFAFBFC0A0C0C0A0
BFC09FBFC0A0B0BF9FB0BF9FB0BF90B0BF90AFB08FAFB090AFB08FAFB08F9FAF7F9FA07F909F70\
8F9F7080906F809060708F5FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF30303F3F3F503F305F403F6F403F6F4F4F704F4F705F50805F5080
605F8F605F8F7060906F6090706F9F706F9F7F70A07F6F9F7F70A07F70A0807FAF807FA0807FAF\
807FAF8F80B08F80AF908FB0908FB0908FBF908FB09F90BF908FBF9F90BF9F90BF
9F90C09F90BFA09FC0A09FC0A09FC0A09FC0AFA0CFA09FC0AFA0CFAFA0C0AFA0CFAFA0CFB0AFCF\
AFA0CFB0AFD0B0AFCFB0AFD0B0AFD0BFB0D0BFB0D0BFB0D0BFB0D0C0BFDFBFB0D0
C0BFDFBFB0DFC0BFDFC0BFDFC0BFDFC0BFDFCFC0E0C0C0DFCFC0E0C0C0E0CFC0E0CFC0E0CFCFE0\
CFC0E0D0CFEFCFCFE0D0CFEFD0CFEFD0D0EFD0CFEFD0D0EFD0CFE0D0D0E0D0D0E0
D0D0E0D0D0DFD0DFE0D0D0DFDFDFDFD0D0DFDFDFDFD0DFDFDFDFDFD0DFD0DFE0DFDFDFD0DFE0D0\
DFDFD0E0E0D0DFE0CFE0E0D0DFE0CFE0EFD0DFE0CFE0E0CFDFE0C0E0E0CFDFE0C0
DFE0CFDFDFC0DFE0C0DFDFC0DFDFC0D0DFBFDFDFC0D0D0BFD0DFBFD0D0BFD0DFBFD0D0BFD0D0BF\
CFCFB0D0D0BFCFCFB0CFCFB0CFCFB0CFCFB0C0CFAFC0CFAFBFC0A0C0C0AFBFC0A0
BFC0A0B0BF9FBFC09FB0BF90B0BF9FAFB090B0BF90AFB08FAFB08FA0AF809FAF8090A07F90A070\
80906F80906F7F8F606F805FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C03F30403030503F3F603F3F604F40704F4070504F7F504F7F5F5F8F
5F5F8F6F60906F6090706F9F6F6090706F9F706F9F7F70A07F709F7F70A07F70A0807FAF807FA0\
8F80AF8F80AF8F80B08F80B0908FB0908FB09090BF908FB09F90BF9090BF9F90BF
9F90BF9F9FC09F90BFA09FC09F9FC0A09FC0A09FC0AFA0CFA09FC0AFA0CFAFA0CFAFAFCFAFA0CF\
B0AFCFAFAFCFB0AFD0B0AFCFB0B0D0B0AFD0BFB0D0B0B0D0BFB0DFBFB0D0BFBFDF
BFB0D0C0BFDFBFBFDFC0BFDFBFBFDFC0BFE0C0BFDFC0C0E0C0BFDFCFC0E0C0C0E0CFCFE0C0C0E0\
CFCFEFCFCFE0CFCFEFCFCFE0D0CFEFCFCFEFD0D0EFCFCFE0D0D0E0CFD0E0D0D0E0
CFD0DFD0D0E0D0D0DFD0DFE0CFD0DFD0DFDFD0D0DFD0DFDFD0DFD0D0DFD0D0DFD0DFE0D0D0DFD0\
DFE0D0D0DFCFDFE0D0D0E0CFDFE0CFDFE0CFDFEFCFDFE0C0DFE0CFD0E0C0DFE0CF
D0DFC0D0E0C0D0DFC0D0DFC0D0DFBFD0DFC0CFD0BFD0DFBFCFD0BFCFD0BFCFD0BFCFD0BFCFCFB0\
CFD0B0C0CFB0CFD0B0C0CFAFC0CFB0C0C0AFC0CFAFBFC0A0BFC0A0BFC0A0BFC0A0
B0BF9FB0BF9FAFBF90B0BF9FAFB090AFB090A0B08FAFB08F9FAF809FAF8090A07F90A07F80906F\
80906F7F8F60708F60C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF3030403F3F603F3F6F4040704F4070504F7F504F7F5F50805F508F
6060906F6090706F9F6F6F9F7070A0706F9F7F70A07F70A07F7FAF7F7FA0807FAF807FAF8F80B0\
8F80AF8F80B08F80B0908FBF908FB09090BF908FBF9F90C09F90BF9F9FC09F90BF
A09FC09F9FC0A09FC0A09FC0A0A0CFA09FC0AFA0CFA0A0CFAFA0CFAFA0CFAFAFD0AFA0CFB0AFD0\
B0AFD0B0AFD0B0AFD0BFB0DFB0B0D0BFB0DFB0B0D0BFBFDFBFB0DFBFBFDFBFB0DF
C0BFE0BFBFDFC0BFE0BFBFDFC0C0E0C0C0E0C0C0E0C0C0E0CFCFEFC0C0E0CFCFEFC0C0E0CFCFEF\
CFCFEFCFCFEFCFCFEFD0D0F0CFCFEFD0D0EFCFD0EFD0D0EFD0D0EFD0D0EFCFD0E0
D0DFEFD0D0E0D0DFE0D0DFDFD0DFE0D0DFDFD0DFDFD0DFDFD0E0DFD0DFD0D0E0DFD0DFD0DFE0DF\
D0E0D0DFE0D0D0E0CFDFEFD0DFE0CFDFEFD0DFE0CFDFEFCFD0E0CFDFE0CFD0E0C0
DFE0CFD0E0C0D0E0C0D0DFC0D0E0C0CFDFC0D0DFC0CFDFBFD0DFC0CFD0BFCFDFBFCFD0BFCFD0BF\
CFD0B0CFD0B0C0CFB0CFD0B0C0CFAFC0CFB0BFC0AFC0CFAFBFC0A0BFC0A0B0BF9F
B0C0A0B0BF9FB0BF9FAFB090AFBF90AFB090AFB090A0B08FA0B08F90A07F90A07F8F9F70809F70\
7F906F7F9060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF3030303F30503F3060403F6F403F6F4F40704F4F7F505080505080605F8F
605F8F6F60906F6090706F9F706F9F7070A0706F9F7F70A07F70A0807FAF7F7FA0807FAF807FAF\
8F80B08F80AF8F8FB08F8FB0908FBF908FBF9090BF9090BF9F90C09090BF9F9FC0
9F90C0A09FC09F9FC0A09FC0A09FC0A0A0CFA09FC0AFA0CFA0A0CFAFAFCFAFA0CFAFAFD0AFAFCF\
B0AFD0AFAFD0B0B0D0B0AFD0B0B0DFB0B0D0BFB0DFB0B0D0BFBFDFB0B0DFBFBFDF
BFBFDFBFBFE0BFBFDFC0C0E0BFBFDFC0C0E0BFC0E0C0C0E0C0C0E0C0CFEFC0C0E0CFCFEFC0CFE0\
CFCFEFC0CFEFCFCFEFCFCFEFCFD0F0CFCFEFCFD0EFCFD0EFD0D0EFCFD0E0D0D0EF
CFD0E0D0DFE0CFD0E0D0DFE0CFDFDFD0DFE0CFDFDFD0DFDFCFDFDFD0E0DFCFDFD0D0E0D0D0E0D0\
D0E0D0D0E0CFD0E0D0D0E0CFD0EFD0D0E0CFD0EFCFD0E0CFD0E0CFD0E0C0D0E0CF
CFE0C0D0E0C0CFDFC0CFDFC0CFDFBFCFDFC0CFD0BFCFDFBFCFD0BFCFD0BFC0D0B0CFD0BFC0CFB0\
CFD0B0C0CFB0C0CFB0C0CFAFC0CFB0BFC0AFBFC0AFB0C0A0BFC0A0B0BF9FB0BF9F
AFBF9FAFBF9FAFB090AFBF90A0B08FA0B0909FAF8F9FB08F90AF8090A07F809F70809F707F906F\
708F60C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF3F3F403F30503F3F6F403F6F4F40704F407F504F80504F805F5F8F5F5F8F
6060906F6090706F9F706F9F7070A07070A07F70A07F70A07F7FAF7F7FAF8080AF807FAF8F80B0\
8080B08F8FBF8F8FB0908FBF908FBF9F90C09090BF9F90C09F90BF9F9FC09F9FC0
A09FC09F9FC0A0A0CFA09FC0A0A0CFA0A0CFAFA0CFA0A0CFAFAFD0AFA0CFAFAFD0AFAFD0B0AFD0\
AFAFD0B0B0DFB0AFD0B0B0DFB0B0D0BFB0DFB0B0DFBFBFDFB0B0DFBFBFE0BFBFDF
BFBFE0BFBFDFC0C0E0BFBFE0C0C0E0BFC0E0C0C0EFC0C0E0C0CFEFC0C0E0CFCFEFC0CFEFCFCFEF\
C0CFEFCFD0F0CFCFEFCFD0F0CFD0EFD0D0F0CFD0EFCFD0EFCFD0EFD0DFEFCFD0EF
D0DFEFCFDFE0D0DFE0CFDFE0D0DFE0CFDFDFD0E0E0CFDFDFD0E0DFCFDFDFD0E0DFD0E0D0D0E0DF\
CFE0D0D0EFD0D0E0D0D0EFD0D0E0CFD0EFD0D0EFCFD0EFCFCFE0CFD0EFCFCFE0C0
D0E0CFCFDFC0D0E0C0CFDFC0CFDFC0CFDFBFCFDFC0CFD0BFCFDFBFC0D0BFCFD0BFC0D0B0C0D0BF\
C0CFB0C0D0B0C0CFB0C0CFB0BFC0AFBFCFAFBFC0A0BFC0A0B0BFA0B0C0A0AFBF9F
AFBF9FAFBF90AFBF9FA0B090A0B0909FB08FA0B08F9FAF8090AF808FA07F8F9F707F906F708F60\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF3030403F3F5F3F306F403F704040704F4F7F4F4F7F5050805F5080605F90
605F906F6F9F6F6F9F706FA0706F9F7070A07070A07F7FAF7F70A0807FAF7F7FAF8080B08080AF\
8F80B08F80B08F8FBF8F8FB09090BF9090BF9F90C09090BF9F9FC09F90C09F9FC0
9F9FC0A09FC09F9FC0A0A0CFA09FC0A0A0CFA0A0CFAFA0CFA0A0CFAFAFD0AFA0CFAFAFD0AFAFD0\
AFAFD0AFAFD0B0B0DFAFAFD0B0B0DFB0B0D0B0B0DFB0B0DFBFBFDFB0B0DFBFBFE0
B0BFDFBFBFE0BFBFE0BFC0E0BFBFE0BFC0E0BFC0E0C0CFEFBFC0E0C0CFEFC0CFEFC0CFEFC0CFEF\
C0CFEFC0CFEFCFD0F0C0CFEFCFD0F0CFD0EFCFD0F0CFD0EFCFD0EFCFD0EFCFDFEF
CFD0E0CFDFE0CFDFE0CFDFE0CFDFDFCFDFE0CFDFDFD0E0E0CFDFDFCFE0DFCFE0DFD0E0DFCFE0D0\
D0E0D0CFE0D0D0EFD0CFE0CFD0EFD0CFE0CFD0EFCFCFE0C0CFE0CFCFE0C0CFE0CF
CFDFC0CFDFC0C0DFC0CFDFC0C0D0BFCFDFBFC0D0BFC0D0BFC0D0B0C0D0BFC0CFB0C0D0BFBFCFB0\
C0CFB0BFCFAFC0CFB0BFC0AFBFC0AFB0C0A0B0C0AFB0BFA0B0C0A0AFBF9FAFBF9F
A0B090A0BF90A0B090A0B0909FAF8F9FB08F90AF8090AF808F9F7F809F7070906F708F6060805F\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF3F3F4F3F305F403F704040704F4F7F4F4F7F5050805050805F5F8F5F5F90
6F6F9F6F6F9F7070A0706FA07F70A07070A07F7FAF7F7FAF8080AF807FAF8080B08080B08F8FB0\
8F8FB08F8FBF8F8FBF9090C09090BF9F90C09090C09F9FC09F9FC09F9FC09F9FC0
A0A0CF9F9FCFA0A0CFA09FCFA0A0CFA0A0CFAFAFD0A0A0CFAFAFD0AFAFD0AFAFD0AFAFD0B0B0DF\
AFAFD0B0B0DFB0B0D0B0B0DFB0B0DFB0BFDFB0B0DFBFBFE0B0BFDFBFBFE0B0BFDF
BFC0E0BFBFE0BFC0E0BFC0E0C0C0EFBFC0E0C0CFEFBFC0EFC0CFEFC0CFEFC0CFEFC0CFEFCFD0F0\
C0CFEFCFD0F0C0D0F0CFD0F0CFD0F0CFDFF0CFD0F0CFDFF0CFDFEFCFDFF0CFDFEF
CFDFEFCFDFEFCFE0EFCFDFE0D0E0EFCFDFE0CFE0E0CFE0DFD0E0E0CFE0DFD0E0DFCFE0DFD0EFDF\
CFE0D0D0E0D0CFE0D0D0EFD0CFE0CFD0EFD0CFE0CFD0E0CFCFE0CFCFE0CFCFE0C0
CFE0CFCFDFC0CFDFC0C0DFBFCFDFC0C0DFBFC0DFBFC0D0BFC0D0BFC0D0BFC0D0BFBFCFB0C0D0BF\
BFCFB0C0CFB0BFCFAFBFCFB0B0C0AFBFC0AFB0C0A0B0C0A0AFBFA0AFBFA0AFBF9F
AFBF9FA0B090A0BF909FB08FA0B0909FB08F90AF8080A07F809F7070906F6F8F6060805FCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
3F303F30304F3F3F603F3F6F4040704040704F4F80504F805F5F8F5F5F8F606090
6060906F6F9F6F6F9F7070A07070A07F7FAF7070A07F7FAF7F7FAF8080B07F7FAF8080B08080B0\
8F8FBF8F8FB08F8FBF8F8FBF9090C09090BF9090C09090C09F9FC09090C09F9FCF
9F9FC0A0A0CF9F9FC0A0A0CFA0A0CFA0A0D0A0A0CFA0AFD0A0A0CFAFAFD0A0AFD0AFAFD0AFAFD0\
B0B0DFAFAFD0B0B0DFB0B0DFB0B0DFB0B0DFB0BFDFB0B0DFBFBFE0B0BFDFBFBFE0
B0BFE0BFC0E0B0BFE0BFC0E0BFC0E0BFC0EFBFC0E0BFCFEFBFCFEFC0CFEFBFCFEFC0CFEFC0CFEF\
C0D0F0C0CFEFC0D0F0C0D0F0CFDFF0C0D0F0CFDFF0C0D0EFCFDFF0C0D0EFCFDFEF
CFDFEFCFDFEFC0DFE0CFE0EFCFDFE0CFE0E0CFDFDFCFE0E0CFDFDFCFE0DFCFDFD0CFE0DFCFDFD0\
CFE0D0CFDFD0CFE0D0CFE0CFCFE0D0CFE0CFCFE0CFCFDFC0CFE0CFC0DFC0CFDFC0
C0DFC0CFDFC0C0DFBFC0DFC0C0D0BFC0DFBFBFD0BFC0D0BFBFD0B0C0D0BFBFCFB0BFCFB0BFCFB0\
BFCFB0BFC0AFBFCFAFB0C0AFB0C0AFAFBFA0B0C0A0AFBF9FAFBFA0A0BF9FA0BF9F
A0B090A0BF909FB08F9FB08F90AF8F90AF80809F7F7F9F70708F6F6F8F605F7F5FCFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
30303F3F3F503F3F604040704040704F4F7F4F4F8050508F50508F606090606090
6F6F9F6F6F9F7070A07070A0707FAF7070A07F7FAF7F7FAF8080B07F80AF8080B08080B08F8FBF\
8F8FB08F90BF8F8FBF9090C09090BF9F9FC09090C09F9FC0909FC09F9FCF9F9FC0
9FA0CF9F9FCFA0A0CFA0A0CFA0A0D0A0A0CFAFAFD0A0A0D0AFAFD0A0AFD0AFAFDFAFAFD0AFB0DF\
AFAFD0B0B0DFAFB0DFB0B0DFB0B0DFB0BFE0B0B0DFB0BFE0B0BFDFBFC0E0B0BFE0
BFC0E0B0BFE0BFC0EFBFC0E0BFC0EFBFC0EFC0CFEFBFCFEFC0CFEFBFCFEFC0D0F0C0CFEFC0D0F0\
C0D0F0C0D0F0C0D0F0CFDFF0C0D0F0CFDFFFC0DFF0CFDFF0C0DFEFCFDFF0C0DFEF
CFDFEFC0DFEFCFE0EFCFDFE0CFDFE0C0DFE0CFE0E0CFDFDFCFE0E0C0DFDFCFE0DFCFDFD0CFE0DF\
CFDFD0CFE0D0CFDFD0CFE0D0CFDFCFCFE0CFCFDFCFCFE0CFC0DFC0CFDFCFC0DFC0
C0DFC0C0DFC0C0DFC0C0D0BFC0DFC0BFD0BFC0DFBFBFD0BFC0D0BFBFCFB0BFD0BFBFCFB0BFCFB0\
BFCFAFBFCFB0B0C0AFB0C0AFB0C0A0B0C0AFAFBFA0AFC0A0A0BF9FAFBF9FA0B090
A0B0909FB08F9FB09090AF8F8FAF80809F7F7F9F706F8F6F6F8F605F7F5FCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
3030403F3F503F3F603F3F7040407F404F7F5050805050805F5F8F5F5F8F60609F
60609F6F6FA06F6FA07070AF7070A07F7FAF707FAF7F7FAF7F7FAF8080B08080B0808FB08080B0\
8F8FBF8F8FBF8F90BF8F8FBF9090C09090BF909FC09090C09F9FC0909FC09F9FCF
9F9FC09FA0CF9F9FCFA0A0CF9FA0CFA0A0D0A0A0CFA0AFD0A0A0D0AFAFD0A0AFD0AFAFDFAFAFD0\
AFB0DFAFAFDFB0B0DFAFB0DFB0B0DFB0B0DFB0BFE0B0B0DFB0BFE0B0BFE0BFC0E0
B0BFE0BFC0E0B0C0E0BFC0EFB0C0E0BFC0EFBFC0EFBFCFEFBFCFEFBFCFEFBFCFEFC0D0F0BFCFEF\
C0D0F0C0D0F0C0D0F0C0D0F0C0DFF0C0D0F0C0DFF0C0D0EFC0DFF0C0D0EFC0DFEF
C0D0E0C0DFEFC0D0E0CFDFE0C0D0E0C0DFE0C0DFDFCFDFE0C0DFDFC0DFDFC0DFD0CFDFDFC0DFD0\
CFDFD0C0DFCFCFDFD0C0DFCFCFDFCFC0DFCFCFDFCFC0DFC0C0DFC0C0D0C0C0DFC0
BFD0BFC0D0C0BFD0BFC0D0C0BFD0BFBFD0BFBFCFB0BFD0BFB0CFB0BFCFB0B0CFB0BFCFB0B0C0AF\
B0CFB0B0C0AFB0C0AFAFC0A0AFC0A0AFBFA0AFBFA0A0B09FA0BF9F9FB0909FB090
90AF8F90AF8F8FAF808FAF807F9F7F7090706F8F6F608F60507F50CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
3030403F3F5F3F3F6040407F404F7F4F50804F50805F5F8F5F5F8F60609060609F
6F6FA06F70A07070AF7070A07F7FAF707FAF7F7FB07F7FAF8080B07F80B0808FB08080B08F8FBF\
8F8FBF8F90BF8F8FBF9090C09090C0909FC0909FC09F9FCF909FC09FA0CF9F9FCF
9FA0CF9FA0CFA0A0D09FA0CFA0AFD0A0A0D0A0AFD0A0AFD0AFAFDFA0AFD0AFB0DFAFAFD0AFB0DF\
AFB0DFB0B0DFAFB0DFB0BFE0B0B0DFB0BFE0B0BFE0BFBFE0B0BFE0BFC0E0B0BFE0
BFC0EFBFC0E0BFC0EFBFC0EFBFCFEFBFCFEFBFCFF0BFCFEFC0D0F0BFCFF0C0D0F0BFCFEFC0D0F0\
C0D0F0C0D0F0C0D0F0C0DFF0C0D0F0C0DFF0C0D0EFCFDFF0C0D0EFC0DFF0C0D0EF
CFDFEFC0D0EFC0DFEFC0D0E0CFDFE0C0DFE0C0DFE0C0D0DFCFDFDFC0DFDFCFDFDFC0DFD0CFDFDF\
C0DFD0CFDFD0C0DFCFCFDFD0C0DFCFCFDFCFC0DFC0C0DFCFC0D0C0C0DFC0BFD0C0
C0DFC0BFD0C0C0D0C0BFD0BFC0D0C0BFD0BFBFD0BFBFCFB0BFD0BFBFCFB0BFCFB0B0CFB0BFCFB0\
B0CFAFB0CFB0AFC0AFAFC0AFAFBFA0A0BFA0A0B09FA0BF9F9FB0909FB09090AF8F
90AF8F8FAF808FAF807F9F7F7F9F706F8F6F608F60507F5FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
3F3F403F3F503F406F3F40704F4F804F4F8050508F50508F5F60905F5F90606F9F
606F9F6F70A06F70A07070AF7070AF7F7FAF707FAF7F80B07F7FAF8080B07F80B0808FB0808FB0\
8F8FBF808FBF8F90BF8F90BF9090C09090C0909FC0909FC09F9FCF909FC09FA0CF
9F9FCF9FA0CF9FA0CFA0A0D09FA0CFA0AFD0A0A0D0A0AFD0A0AFD0AFAFDFA0AFD0AFB0DFAFAFDF\
AFB0DFAFB0DFAFB0DFAFB0DFB0BFE0AFB0DFB0BFE0B0BFE0B0BFE0B0BFE0B0C0EF
B0BFE0BFC0EFB0C0E0BFC0EFB0C0EFBFCFEFBFC0EFBFCFEFBFCFEFBFCFEFBFCFEFBFCFEFBFCFEF\
C0D0F0BFCFEFC0D0EFBFD0EFC0D0F0BFD0EFC0D0EFC0D0EFC0D0EFBFD0E0C0D0EF
C0D0E0C0D0EFBFD0E0C0D0E0C0D0E0C0D0E0BFD0DFC0D0DFC0D0DFC0DFDFC0D0D0C0DFD0C0D0D0\
C0DFD0C0D0CFC0DFCFC0D0CFC0DFCFBFD0C0C0DFCFBFD0C0C0D0C0BFD0C0BFD0C0
BFD0BFBFD0C0BFCFBFBFD0BFBFCFBFBFCFBFB0CFB0BFCFBFB0CFB0B0CFB0B0C0AFB0CFB0AFC0AF\
B0C0AFAFBFA0AFC0A0A0BFA0A0BFA09FB09F9FB09F90AF9090B0908FAF8F8FAF8F
80A08080A0807F9F7F709070608F6F5F8060C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
3030403F405F3F406F404F7F404F7F4F508F50508F505F905F5F9060609F60609F
6F6FA06F6FA0707FAF7070AF707FAF707FAF7F80B07F7FAF7F80B07F80B0808FBF808FB0808FBF\
808FBF8F90C08F90BF9090C08F90C0909FC0909FC09F9FCF909FC09FA0CF9F9FCF
9FA0CF9FA0CFA0A0D09FA0CFA0AFD09FA0D0A0AFD0A0AFD0A0AFDFA0AFD0AFB0DFAFAFDFAFB0DF\
AFB0DFB0B0E0AFB0DFB0BFE0AFB0E0B0BFE0B0BFE0B0BFE0B0BFE0B0C0EFB0BFE0
BFC0EFB0C0E0BFC0EFB0C0E0BFC0EFB0C0E0BFCFEFBFCFEFBFCFEFBFCFEFBFCFEFBFCFEFBFD0EF\
BFCFEFC0D0EFBFD0EFC0D0EFBFD0EFC0D0EFC0D0EFC0D0EFBFD0E0C0D0EFC0D0E0
C0D0E0BFD0E0C0D0E0C0D0E0C0D0E0BFD0DFC0D0E0C0D0DFC0D0DFBFD0D0C0DFDFC0D0D0C0D0D0\
BFD0CFC0DFD0C0D0CFC0DFCFBFD0CFC0DFCFBFD0C0C0D0CFBFD0C0C0D0C0BFCFC0
BFD0C0BFCFBFBFD0C0BFCFBFBFCFBFB0CFB0BFCFBFB0C0B0B0CFB0AFC0B0B0CFB0AFC0AFAFC0AF\
A0BFA0AFC0AFA0BFA0A0BFA09FB09F9FB09F90AF9090AF908FAF8F8FAF8F80A080
80A080709F7F709070608F6F4F7050CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
3F3F4F3F3F5F4040703F407F404F804F4F8050508F50508F5F5F905F6090606F9F
606F9F6F70A06F70A0707FAF7070AF707FAF707FAF7F80B07F7FAF7F80B07F80B0808FBF808FB0\
808FBF808FBF8F90C08F90BF9090C08F90C0909FCF909FC0909FCF909FC09FA0CF
909FCF9FA0CF9FA0CF9FA0D09FA0CFA0AFD09FA0D0A0AFD0A0AFD0A0AFDFA0AFD0AFB0DFA0AFDF\
AFB0DFAFB0DFAFB0E0AFB0DFB0BFE0AFB0DFB0BFE0AFB0DFB0BFE0B0BFE0B0C0E0
B0BFDFB0C0E0B0BFE0B0C0E0B0C0E0B0C0E0B0C0E0BFCFE0B0C0E0BFCFE0B0CFE0BFCFEFBFCFE0\
BFCFEFBFCFE0BFD0EFBFCFE0BFD0EFBFD0E0C0D0EFBFCFE0BFD0E0BFCFE0C0D0E0
BFCFDFBFD0E0BFCFDFC0D0DFBFCFDFBFD0DFBFCFD0C0D0DFBFCFD0BFD0D0BFCFD0C0D0D0BFCFCF\
BFD0D0BFD0CFC0D0CFBFD0C0BFD0CFBFCFC0BFD0C0BFCFC0BFD0C0BFCFBFBFCFC0
B0CFBFB0CFBFB0CFBFB0CFBFAFC0B0B0CFB0AFC0B0AFC0B0AFC0AFAFC0AFA0C0AFAFC0AFA0BFA0\
A0BFA09FB09F9FBF9F90B09090B0908FAF908FAF9080A08F80A08F7FA0807FA080
70907F6F8F6F4F7050CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
30304F3F40603F406F404F80404F804F508F4F508F505F90505F905F609F5F609F
6F70A06F70A0707FAF707FAF707FB0707FAF7F80B07F80B07F80B07F80B0808FBF808FBF8F90BF\
808FBF8F90C08F90BF909FC08F90C0909FCF909FC09FA0CF909FCF9FA0CF90A0CF
9FA0D09FA0CF9FAFD09FA0D0A0AFD0A0AFD0A0AFDFA0AFD0AFB0DFA0AFDFAFB0DFA0B0DFAFB0DF\
AFB0DFAFB0E0AFB0DFB0BFE0AFBFE0B0BFE0AFBFDFB0BFE0B0BFDFB0BFE0B0BFDF
B0C0E0B0BFE0B0C0E0B0C0E0BFC0E0B0C0E0BFC0E0B0C0E0BFCFE0BFCFE0BFCFE0BFCFE0BFCFEF\
BFCFE0BFD0E0BFCFE0C0D0EFBFD0E0C0D0E0BFD0E0C0D0E0BFD0E0C0D0E0BFD0DF
C0D0E0BFD0DFC0D0DFBFD0DFC0D0DFBFD0D0C0D0DFBFCFD0C0D0DFBFD0D0C0D0D0BFCFCFC0D0D0\
BFD0CFC0D0CFBFCFCFC0D0CFBFD0CFBFD0CFBFCFC0BFD0C0B0CFC0BFCFC0B0CFBF
B0CFC0B0CFBFB0CFBFAFC0B0B0CFBFAFC0B0AFC0B0AFC0AFAFC0B0A0C0AFA0C0AFA0BFA0A0BFAF\
9FBFA09FBFA090B09F90B09F8FAF908FAF9080A08F80AF8F7FA0807FA0806F9070
5F7F60CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
3F3F4F3F3F5F3F406F3F407F405080404F804F508F505F8F5F60905F6090606F9F
606F9F6F70A06F70A0707FAF707FAF7080B0707FAF7F80B07F80B07F8FBF7F80B0808FBF808FBF\
8090BF808FBF8F90C08F90C0909FC08F90C0909FCF909FC090A0CF909FCF9FA0CF
90A0CF9FA0D09FA0CF9FAFD09FA0D0A0AFD09FAFD0A0AFD0A0AFD0A0B0DFA0AFD0A0B0DFA0B0DF\
AFB0DFA0B0DFAFB0DFAFB0DFAFBFE0AFB0DFAFBFE0AFBFDFB0BFE0AFBFDFB0BFDF
AFBFDFB0C0E0AFBFDFB0C0DFB0BFDFB0C0E0B0C0DFB0C0E0B0C0DFBFCFE0B0C0DFBFCFE0B0CFDF\
BFCFE0B0CFDFBFCFE0BFCFE0BFD0E0BFCFDFBFD0E0BFCFDFBFD0DFBFCFDFBFD0DF
BFCFDFBFD0DFBFCFD0BFD0D0BFCFD0BFD0D0BFCFD0BFD0D0BFCFCFBFD0D0BFCFCFBFD0CFBFCFCF\
BFD0CFB0CFC0BFD0CFBFCFC0BFD0CFB0CFC0B0CFC0B0CFBFB0CFC0AFC0BFB0CFBF
AFC0B0AFC0BFAFC0B0AFC0B0A0C0AFAFC0B0A0BFAFA0C0AFA0BFAFA0BFAF9FBFA09FBFA090B09F\
90B09F8FAF908FAF9080AF8F8FAF907FA0807FA08F709F7F709F7F507F60406F50
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
30304F3F405F3F406F404F7F404F804F508F4F508F505F90505F905F6F9F5F6F9F
6070A06070A06F7FAF6F7FAF707FB0707FAF7F80B07F80B07F80B07F80B0808FBF808FBF8090BF\
808FBF8F90C08F90C08F9FC08F90C0909FCF909FCF90A0CF909FCF9FA0CF90A0CF
9FA0D09FA0CF9FAFD09FA0D09FAFD09FAFD0A0AFD09FAFD0A0B0DFA0AFD0A0B0DFA0B0D0AFB0DF\
A0B0DFAFBFDFAFB0DFAFBFDFAFB0DFAFBFE0AFBFDFAFBFE0AFBFDFB0BFE0AFBFDF
B0C0DFAFBFDFB0C0E0B0BFDFB0C0DFB0C0DFB0C0DFB0C0DFB0C0DFB0C0DFBFCFE0B0C0DFBFCFE0\
B0CFDFBFCFE0BFCFDFBFCFE0BFCFDFBFD0E0BFCFDFBFD0DFBFCFDFBFD0DFBFCFD0
BFD0DFBFCFD0BFD0DFBFCFD0BFD0D0BFCFCFBFD0D0BFCFCFBFD0D0B0CFCFBFD0CFB0CFCFBFD0CF\
B0CFC0BFD0CFB0CFC0BFCFC0B0CFC0B0CFC0B0CFBFB0CFC0AFC0BFAFCFBFAFC0B0
AFC0BFA0C0B0AFC0B0A0C0B0A0C0B0A0BFAFA0C0AF9FBFAF9FBFAF9FBFA09FBFA090B09F90B09F\
8FAF908FB09F80AF9080AF8F7F9F80709F806F907F608F7040705FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
303F4F3F3F503F406F3F4070404F804050804F5F8F4F5F8F5060905060905F6F9F
5F6F9F6070A06F70A0707FAF6F7FAF7080B0707FB07F80B07080B07F8FBF7F8FB0808FBF7F8FBF\
8090BF8090BF8F90C08090C08F9FC08F9FC0909FCF8F9FC090A0CF909FCF90A0CF
90A0CF9FA0CF90A0CF9FAFD09FA0CF9FAFD09FAFCFA0AFD09FAFD0A0AFD09FAFD0A0B0DFA0AFD0\
A0B0DFA0B0D0AFB0DFA0B0DFAFBFDFAFB0DFAFBFDFAFB0DFAFBFDFAFBFDFAFBFDF
AFBFDFAFBFDFAFBFDFB0C0DFAFBFD0B0C0DFAFC0DFB0C0DFAFC0D0B0C0DFB0C0D0B0C0DFB0C0D0\
B0CFDFB0C0DFBFCFDFB0CFDFBFCFDFB0CFDFBFCFDFB0CFD0BFCFD0B0CFD0BFCFD0
B0CFD0BFCFD0B0CFCFBFCFD0B0CFCFB0CFCFB0CFCFB0CFCFB0CFC0B0CFCFB0CFC0B0CFCFB0C0C0\
B0CFC0B0CFC0B0CFC0AFC0BFAFCFC0AFC0BFAFCFBFA0C0B0AFC0BFA0C0B0A0C0B0
A0BFAFA0C0B09FBFAFA0BFAF9FBFAF9FBFAF9FBFA09FBFA090B0A090BFA090B09F90B09F80AF90\
80AF907FA08F7F9F806F907F6F907F5F8F704F7F60C0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
30304F3F405F3F4060404F7F4050804F508F4F508F5060905060905F6F9F5F6F9F
6070A06070A06F7FAF6F7FAF7080B07080B07F80B07080B07F8FBF7F8FBF808FBF7F8FBF8090C0\
8090BF8F90C08090C08F9FC08F9FC0909FCF8F9FC090A0CF90A0CF90A0CF90A0CF
9FAFD090A0CF9FAFD09FA0CF9FAFD09FAFD09FAFD09FAFD0A0B0D09FAFD0A0B0D0A0B0D0A0B0DF\
A0B0D0AFB0DFA0B0DFAFBFDFAFBFDFAFBFDFAFBFDFAFBFDFAFBFDFAFBFDFAFBFDF
B0C0DFAFBFDFB0C0DFAFBFDFB0C0DFAFC0DFB0C0DFAFC0DFB0CFDFB0C0DFB0CFDFB0C0DFB0CFDF\
B0CFDFB0CFDFB0CFD0BFCFDFB0CFDFBFCFDFB0CFD0BFCFDFB0CFD0BFCFD0B0CFD0
BFCFD0B0CFD0B0CFD0B0CFCFB0CFD0B0CFCFB0CFCFB0CFCFB0CFCFB0CFCFB0CFCFAFCFC0B0CFCF\
AFCFC0B0CFC0AFC0C0B0CFC0AFC0BFAFCFBFA0C0BFAFC0BFA0C0B0A0C0B0A0BFB0
A0C0B09FBFAF9FBFAF9FBFAF9FBFAF90BFA09FBFAF90B0A090BFA08FB09F8FAF9F7FA0907FA08F\
709F806F9F80608F7F608F7050806FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCF303F503F40603F4F704050804050804F5F8F4F5F8F5060905060905F6F9F
5F6FA06070AF6070A06F7FAF6F7FAF7080B07080B07F80B07080B07F8FBF7F8FB07F8FBF7F8FBF\
8090C08090BF8090C08090C08F9FC08F9FC08F9FCF8F9FC090A0CF8F9FC090A0CF
90A0CF90A0CF90A0CF9FAFCF90A0CF9FAFD09FAFCF9FAFD09FAFCFA0B0D09FAFD0A0B0D09FB0D0\
A0B0DFA0B0D0A0B0DFA0B0D0AFBFDFA0B0D0AFBFDFA0BFDFAFBFDFA0BFD0AFBFDF
AFBFD0AFC0DFAFBFD0AFC0DFAFBFD0B0C0DFAFBFD0B0C0DFAFC0D0B0C0DFAFC0D0B0C0DFB0C0D0\
B0CFDFAFC0D0B0CFDFB0C0D0B0CFDFB0C0D0B0CFD0B0C0D0B0CFD0AFC0CFB0CFD0
AFC0CFB0CFD0AFC0CFB0CFCFAFC0CFB0CFCFAFC0C0AFCFCFAFC0C0AFCFC0AFC0C0AFCFC0AFC0C0\
AFCFC0A0C0BFAFC0C0A0C0BFAFC0BFA0C0B0A0C0BF9FBFB0A0C0B09FBFAF9FBFB0
9FBFAF9FBFAF90B0A090BFAF90B0A090B0A08FB09F8FB09F80AF907FA090709F80709F8060907F\
60907F50806F50806FCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF3F3F5F3F4060404F70404F804F5F8F4F5F8F5060905060905F6F9F5F6F9F
6070A06070A06F7FAF6F7FAF7080B0707FB07080B07080B07F8FBF7080B07F8FBF7F8FBF8090BF\
7F8FBF8090C08090C08F9FC08090C08F9FCF8F9FC090A0CF8F9FCF90A0CF90A0CF
90A0CF90A0CF9FAFD090A0CF9FAFD090AFCF9FAFD09FAFCF9FB0D09FAFD0A0B0D09FB0D0A0B0D0\
9FB0D0A0BFDFA0B0D0A0BFDFA0B0D0AFBFDFA0BFDFAFBFDFA0BFD0AFC0DFAFBFDF
AFC0DFAFBFD0AFC0DFAFC0D0AFC0DFAFC0D0B0C0DFAFC0D0B0C0DFAFC0D0B0CFDFAFC0D0B0CFDF\
AFC0D0B0CFDFAFC0D0B0CFD0AFC0D0B0CFD0AFC0D0B0CFD0AFC0CFB0CFD0AFC0CF
AFCFCFAFC0CFAFCFCFAFC0CFAFCFCFAFC0C0AFCFCFAFC0C0AFCFC0A0C0C0AFCFC0A0C0C0AFC0C0\
A0C0BFAFCFC0A0C0BFA0C0BFA0C0B0A0C0BF9FBFB09FC0B09FBFAF9FBFB090BFAF
9FBFAF90B0A090B0A08FB0A08FB09F80AF9F80AF9F7FA09070A08F6F90806F90805F8F7F5F8F70\
4F7F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCF3030503F40603F406F40508040508F4F5F904F5F8F50609050609F5F6FA0
5F6F9F6070A06070A06F7FAF6F7FAF7080B06F80B07080B07080B07F8FBF708FB07F8FBF7F8FBF\
8090BF7F90BF8090C08090BF809FC0809FC08F9FCF8F9FC08FA0CF8F9FC090A0CF
8FA0CF90A0CF90A0CF90AFCF90A0CF9FAFCF90AFCF9FAFD090AFCF9FAFD09FAFCF9FB0D09FAFCF\
A0B0D09FB0D0A0B0D09FB0D0A0BFD0A0B0D0A0BFDFA0B0D0A0BFDFA0BFD0AFBFDF
A0BFD0AFBFD0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0AFC0D0AFC0D0AFC0D0\
AFC0D0AFC0D0AFC0D0AFC0CFAFC0D0AFC0CFAFC0D0A0C0CFAFC0CFA0C0CFAFC0CF
A0C0C0AFC0CFA0C0C0AFC0CFA0C0C0A0C0C0A0C0C0A0C0C0A0C0BFA0C0C0A0C0BFA0C0C09FBFBF\
A0C0BF9FC0BFA0C0BF9FBFB09FC0B09FBFB09FBFB090B0AF90BFAF8FB0A08FB0A0
80AF9F80AF9F80AF9F80AF9F7FA0907FA0906F9F8F6F9F80608F7F5F8F7F508070CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCF3F3F503F3F60404F704050804F5F8F4F5F8F5060905060905F6F9F5F6F9F
6070A06070A0607FAF607FAF7080B06F80B07080B07080B07F8FBF708FB07F8FBF7F8FBF8090C0\
7F90BF8090C08090BF8F9FC0809FC08F9FCF8F9FC08FA0CF8FA0CF90A0CF8FA0CF
90AFCF90A0CF90AFCF90A0CF9FAFD090AFCF9FAFD090AFCF9FB0D09FAFD09FB0D09FB0D0A0B0D0\
9FB0D0A0BFD0A0B0D0A0BFDFA0BFD0A0BFDFA0BFD0AFBFDFA0BFD0AFBFDFA0BFD0
AFC0DFA0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0BFD0AFC0D0A0C0D0AFC0D0A0C0D0AFC0D0\
AFC0D0AFC0D0AFC0CFAFCFD0AFC0CFAFC0D0A0C0CFAFCFD0A0C0CFAFC0CFA0C0CF
AFC0CFA0C0C0A0C0CFA0C0C0A0C0CFA0C0C0A0C0C0A0C0C0A0C0C0A0C0BFA0C0C09FC0BFA0C0BF\
9FC0BFA0C0BF9FBFB09FBFB090BFAF90BFAF8FB0AF8FB0AF80AFA080AFA07FAF9F
80AF9F7FA0907FA090709F8F70A08F6F90806090807F503F6F4F3FCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C03F3F603F406F40507F4050804F5F8F4F5F8F5060905060905F6F9F
5F6F9F6070A05F70A0607FAF6F7FAF7080B06F80B07080B07080B0708FBF708FB07F8FBF7F8FBF\
7F90BF7F8FBF8090C08090BF809FC08090C08F9FC08F9FC08FA0CF8F9FC08FA0CF
8FA0CF90A0CF8FA0CF90AFCF90A0CF90AFD090AFCF9FAFD090AFCF9FB0D090AFCF9FB0D09FB0D0\
9FB0D09FB0D0A0B0D09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD0
A0BFD0A0BFD09FBFCFA0BFD0A0BFCFA0BFD0A0BFCFA0BFD0A0BFCFA0C0D0A0BFCFA0C0D0A0BFCF\
A0C0D0A0BFCFA0C0CFA0C0CFAFC0CFA0C0CFA0C0CFA0C0CFA0C0CFA0BFC0A0C0CF
A0C0C0A0C0C09FBFC0A0C0C09FBFC0A0C0C09FBFBF9FC0BF9FBFBF9FC0BF9FBFB09FBFBF90BFB0\
9FBFB090B0AF90B0AF8FB0AF8FB0AF80AFA080AFA07FA09F7FAF9F70A09070A090
6F9F8F709F8F6090806090804060509060508F5F4F8F5F407F503F7F4F3FC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCF3F3F5F3F406F3F4F7040508F405F8F4F60904F6090506F9F506F9F
5F6FA05F6FA0607FAF607FAF6F7FAF6F7FAF7080B07080B0708FBF7080B07F8FBF708FBF7F90BF\
7F8FBF8090C07F90BF809FC08090C08F9FC0809FC08FA0CF8F9FC08FA0CF8FA0CF
90A0CF8FA0CF90AFD090A0CF90AFD090AFCF9FAFD090AFCF9FB0D090AFCF9FB0D09FB0D09FB0D0\
9FB0D09FB0D09FB0D09FBFD09FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD09FBFD0
A0BFD09FBFCFA0BFD09FBFCFA0BFD09FBFCFA0C0D0A0BFCFA0C0D09FBFCFA0C0D0A0BFCFA0C0CF\
A0BFCFA0C0D0A0C0CFA0C0CFA0C0CFA0C0CFA0C0CFA0C0CF9FBFC0A0C0CF9FC0C0
A0C0C09FBFC0A0C0C09FBFC09FC0C09FBFBF9FC0BF90BFBF9FBFBF90BFB090BFB090B0B090BFB0\
8FB0AF8FB0AF80AFA080AFA07FA09F7FAF9F70A09070A0906F9F8F709F906F9F8F
6F9F8F5F8F7FA0706F9F7060A070609F6F5F9F6F5F8F604F8F5F4F80504080503FCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C03F3F5F303F603F4F703F4F7F405F8F405F8F4F60904F6090506F9F
506F9F5F70A05F70A0607FAF607FAF6F80B06F80B07080B06F80B0708FB0708FB07F8FBF708FBF\
7F90BF7F90BF7F90C07F90BF809FC0809FC0809FC0809FC08FA0CF8FA0C08FA0CF
8FA0C08FA0CF8FA0CF90AFCF8FA0CF90AFCF8FAFCF90AFCF90AFCF90AFCF90AFCF90B0D090AFCF\
9FB0D090B0CF9FB0D090B0CF9FB0D09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD0
9FB0CF9FBFCF9FB0CF9FBFD09FB0CF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCFA0BFCF9FBFC0\
9FBFCF9FBFCFA0C0CF9FBFC0A0C0CF9FBFC0A0C0CF9FBFC09FC0C09FBFC09FBFC0
90BFBF9FBFBF90BFBF90BFBF8FB0B090BFB08FB0B08FB0B08FB0AF8FB0AF80AFAF8FB0AF80AFA0\
80AFA07FA09F7FAF9F70A09070A0906F9F906F9F9060908F609080508070A0706F
9F6F60A0706FA07060A0706F9F7060A070609F6F5F9F6F5F90604F8F604F805F407F503FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F40603F406F40507F4050804F60904F609050609F50609F
5F70A05F6FA05F70A05F70A06F7FAF6F7FAF6F80B06F80B0708FBF708FB0708FBF708FBF7F90BF\
7F90BF7F90C07F90BF809FC0809FC0809FC0809FC08FA0CF809FC08FA0CF8FA0CF
8FA0CF8FA0CF8FAFCF8FA0CF90AFCF8FAFCF90AFCF8FAFCF90AFD090AFCF90B0CF90AFCF9FB0D0\
90B0CF9FB0D090B0CF9FB0D09FB0CF9FBFD09FB0CF9FBFD09FB0CF9FBFD09FB0CF
9FBFD09FBFCF9FBFD09FBFCF9FBFD09FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCF9FBFCFA0C0CF\
9FBFCF9FC0CF9FBFC0A0C0CF9FBFC09FC0C090BFC09FBFC090BFBF90BFBF90B0BF
90BFBF8FB0B090B0BF8FB0B08FB0B08FB0AF8FB0B080AFAF80B0AF80AFA080AFAF7FAFA07FAFA0\
70A09F70A09F6F9F9070A0906090805F8F7F4F706F406F60A07070A07070A07070
AF7F70A07F70AF7F70A07F6FAF7F6FA07060A07F60A07060A070609F6F50906F50805F407F503F\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF30305F3F406F3F4F70405080405F8F4F60904F6090506F9F
506F9F5F70A05F70A05F70A0607FAF6F80AF607FAF6F80B06F80B0708FBF6F8FB0708FBF708FBF\
7F90BF708FBF7F90BF7F90BF809FC07F90BF809FC0809FC080A0C0809FC08FA0CF
80A0C08FA0CF80A0C08FA0CF8FA0C08FAFCF8FA0C08FAFCF8FAFCF90AFCF8FAFC090AFCF8FAFCF\
90B0CF8FAFCF90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF9FBFCF90B0CF9FB0CF
90B0CF9FBFCF90B0CF9FBFCF90B0CF9FBFCF90B0C09FBFCF90BFC09FBFCF90BFC09FBFCF90BFC0\
9FBFC090B0C090BFC090BFC090BFC08FB0BF90BFBF8FB0BF8FB0BF8FB0B08FB0B0
80AFB08FB0B080AFAF80AFAF7FAFA080AFAF7FA0A07FAFA07FA0A07FAFA070A09F70A09F6F9F90\
60908F50807F507F70406F609F6F709F6F70A07070A07070A07F7FA07070AF7F70
A07F70AF7F70A07F70AF7F70A07F6FAF7F6FA07060A07F60A0705FA070609F7050906F50805040\
704F3FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF303F60404F703F507F405F8F405F8F4F609F4F6090
506F9F506F9F5F70A05F70A0607FAF607FAF6F80B06F80B06F8FB06F80B0708FBF708FB07090BF\
708FBF7F90BF7090BF7F90C07F90BF809FC07F9FC0809FC0809FC080A0CF80A0C0
8FA0CF80A0C08FA0CF80A0C08FAFCF8FA0C08FAFCF8FAFCF8FAFCF8FAFC090AFCF8FAFCF90AFCF\
8FAFCF90B0CF90AFCF90B0CF90B0CF90B0CF90B0CF90B0CF90B0CF9FBFCF90B0CF
90BFCF90B0CF9FBFCF90B0CF9FBFCF90B0C09FBFCF90B0C090BFC090B0C090BFC090B0C090BFC0\
8FB0BF90BFC08FB0BF90BFBF8FB0BF8FB0BF8FB0B08FB0B080AFB080B0B080AFAF
80AFAF7FAFAF80AFAF7FA0A07FAFA070A0A07FAFA070A09F70A09F6F9F9060908F50807F4F7F70\
3F6F6090606F9060709F6F709F6F70A0707FA07070A07F7FA07070AF7F7FAF7F70
AF807FAF807FB0807FAF8070AF8070AF7F6FAF8070A07F6FAF7F6FA07F60A07F609F705F906F50\
7F5040704F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0303F603F406F404F7F3F5080405F904060904F6F9F
4F6090506F9F506F9F5F70A05F70A0607FAF607FAF6F80B06080AF6F8FB06F80B0708FBF6F8FB0\
708FBF708FB07090BF7090BF7F90BF7F90BF7F9FC07F90BF7F9FC07F9FC080A0C0
809FC080A0C080A0C080A0C080A0C080A0C080A0C08FAFCF80A0C08FAFCF8FAFC08FAFCF8FAFC0\
8FAFCF8FAFC08FB0CF8FAFC090B0CF8FAFC090B0CF8FB0C090B0CF8FB0C090B0CF
8FB0C090B0C08FB0C090B0C08FB0C08FB0C08FB0BF8FB0C08FAFBF8FB0BF8FAFBF8FB0BF80AFB0\
8FB0BF80AFB08FB0BF80AFB080B0B080AFB080AFB07FAFAF7FAFAF7FA0A07FAFAF
70A0A07FA0A070A09F70A0A06F9F906090905F8F8F5F8F804F807F407F707F4F5F805F608F5F60\
90606F90606F9F6F709F6F70A07070A07070A07F7FA07070AF7F7FAF7F7FB0807F
AF807FB0807FAF807FB08F7FAF8070B08070AF8070AF8070AF7F6FAF806FA07F60A07F609F7050\
906F507F5040704F3FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F406F3F4070405080405F8F4F60904F6090
506F9F506F9F5F70A05070A05F7FAF5F7FAF6080AF6080AF6F8FB06F80B06F8FB06F8FB07090BF\
708FB07090BF7090BF7F90BF7090BF7F9FBF7F90BF7F9FC07F9FC0809FC07F9FC0
80A0C080A0C080A0C080A0C08FAFCF80A0C08FAFCF80A0C08FAFCF8FAFC08FAFCF8FAFC08FAFCF\
8FAFC08FB0CF8FAFC090B0CF8FB0CF90B0CF8FB0C090B0CF8FB0C08FB0C08FAFC0
8FB0C08FAFC08FB0C080AFBF8FB0C080AFBF8FB0BF80AFBF8FB0BF80AFBF80B0BF80AFB080B0BF\
80AFB080AFB080AFB080B0B07FAFAF7FAFAF7FA0AF7FAFAF70A0A07FA0A06F9F9F
6F9F9F60908F5F908F50808050807F407F70407F706F40507F4F5F7F505F805F608F5F60906070\
9F6F70A0707F9F707FA07F7FA07F7FAF7F80AF7F7FB08080B08080B08F80B08080
B08F80B08F7FB08F80B08F7FBF8F80B08F7FB08F7FB08F70B08F70AF806FAF806FA07F60A07F60\
9F705F906F507F5F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3030603F40703F4F7F405F8F405F8F4F6090
4F6090506F9F4F6F9F5070A05070A05F7FA05F7FA06080AF6080AF6F8FB06080AF6F8FB06F8FB0\
708FB06F8FB07090BF708FB07090BF7090BF7090BF7090BF7F9FC07F9FBF7F9FC0
7F9FBF80A0C07F9FC080A0C080A0C080A0C080A0C080A0C080A0C080AFC080A0C08FAFC080AFC0\
8FAFC080AFC08FAFC080AFC08FAFC080AFBF8FAFC080AFC08FAFC080AFBF80AFBF
80AFBF80AFBF80AFBF80AFBF80AFB080AFBF7FA0B080AFB07FA0B080AFB07FA0AF7FAFB07FA0AF\
7FAFB070A0AF7FAFAF70A0AF7FA0AF6F9F9F6F9F9F609090609090508F80508080
4F7F7F4F7F7F3F706F5F40505F3F4F6040506F3F4F704F50704F50805F608F5F6F906070906F70\
9F707F9F7070A07F7FA07F7FAF8080AF8080B08080B08080B08F80B08080B08F80
B08F80B08F80B08F7FB08F80B08F7FBF9080B08F7FBF907FB08F70B08F70AF806FAF806FA07F60\
A07F5F9F7050906F4F6F4F3FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F6F3F4F7F3F4F7F405F8F406090
4F6F904F6F9050709F506F9F5070A05070A05F7FAF5F7FAF6080AF6080AF6F8FB06F8FB06F8FB0\
6F8FB07090BF6F8FB07090BF7090B07090BF7090BF7F9FBF7090BF7F9FC07F9FBF
7FA0C07F9FBF80A0C07FA0C080A0C07FA0C080A0C080A0C080A0C07FA0BF80AFC080A0BF80AFC0\
80A0BF80AFC080A0BF80AFC080A0BF80AFC080AFBF80AFBF80AFBF80AFBF80AFBF
80AFBF7FA0B080AFBF7FA0B07FAFB07FA0B07FAFB07FA0B07FAFB070A0AF7FAFB070A0AF7FA0AF\
70A0A070A0A06F9F9F6F9F9F5F90905F9090508F805080804F7F7F407F707F608F
70507F5F405F503F4F503F40603F4F60404F704F507F505F8F5F6F90606F9F6F709F6F709F707F\
9F707FAF7F80AF7F80B08080B08080B08F8FB08F8FBF8F8FB08F8FBF908FB08F8F
BF908FB08F80BF908FBF9080BF908FBF9080BF9080BF9080BF9080B08F70B08F70AF806FAF806F\
A07F5FA0805F9F7F5080604F5F403FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF303F6F3F4F7F3F507F405F8F
40608F4F6F904F6F9050709F4F6F9F5070A05070A05F7FA05F7FA06080AF6080AF6F8FB06080AF\
6F8FB06F8FB06F90B06F8FB07090B06F90B07090BF7090B0709FBF7090BF7F9FBF
709FBF7F9FBF7F9FBF7FA0C0709FBF7FA0BF7F9FBF7FA0BF709FBF7FA0BF7F9FBF7FA0BF7F9FB0\
7FA0BF7FA0B07FA0BF7FA0B07FA0BF7FA0B07FA0BF7FA0B07FA0BF7FA0B07FAFB0
70A0B07FA0B070A0AF7FA0B070A0AF70A0AF70A0AF70A0AF6F9FA06F9FA060909F609F9F5F9090\
5F90905F8F8F5F8F8F4F80804F807F407F70A090B09080A08F7F9F7F608F6F507F
4F3F4F403F404F303F603F406F404F7F505F80506090606F90606F9F6F709F6F70A0707FA0707F\
AF8080AF8080B08F8FB0808FB08F8FB08F8FBF908FB08F8FBF908FBF908FBF908F
BF908FBF908FBF908FBF9F8FBF9080BF9F8FBF9080BF9F80B0907FBF907FB08F70B08F70AF806F\
AF806FA0805FA0805F90704F705040C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F4070404F7F3F5080
40608F4060904F6F9F4F6F9F5070A050709F5F7FA0507FA05F80AF5F80AF6080AF6080AF6F8FB0\
6F8FB06F8FB06F8FB07090BF6F90B07090BF6F90B0709FBF7090BF709FBF7090BF
7F9FBF709FBF7F9FBF709FBF7FA0BF709FBF7FA0BF709FBF7FA0BF709FBF7FA0BF709FB07FA0BF\
70A0B07FA0BF70A0B07FA0BF70A0B07FA0B070A0B07FA0B070A0B07FA0B070A0B0
7FA0B070A0AF70A0AF6F9FA06F9FA060909F60909F5F90906090905F8F8F5F8F8F508080508080\
4F807FBFB0D0B0AFC0B0AFC0AFA0BFAF9FBF9F8FB09080A07F6F906F507F503F50
403F405F3F406F4F5070505F8F5F6F8F5F6F906070906F70A07080A07080AF7F80AF8080B08F90\
B08F8FB08F90B08F8FBF9090B0908FBF9090BF9090BF9F90BF9090C09F90BF9F90
C09F90C09F90C09F90BF9F8FC09F90BF9F8FC09F8FBF9F80C09F8FBF9F80BF9F80B0907FB0907F\
AF8F6FAF8F6FA0805FA0805F8F6F4F604F40CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF3F406F3040703F507F
3F508040608F4060904F6F9F4F6F9F50709F50709F507FA0507FA05F80AF5F7FA06080AF6080AF\
608FB0608FAF6F8FB0608FAF6F90B0608FAF6F90B06F90B06F90B06F90B06F90B0
6F90B0709FB06F90B0709FB0709FB0709FBF709FB0709FB0709FB0709FB0709FB0709FB0709FB0\
70A0B0709FAF709FB0709FAF70A0B06F9FAF70A0B0709FAF70A0AF6F9FA06F9FA0
6090A0609FA05F909060909F5F8F905F8F9050808F50808F4F80804F8080BFCFDFBFC0EFB0B0D0\
BFB0CFAFAFC0B0AFC0A09FBFA09FBF9F90B0908FAF80709F6F508050405F4F3F40
60404F8050608050608F5F6F8F606F906F709F6F70A07080A07F80AF808FAF808FB08F90B08F8F\
B08F90B08F8FBF9090B09090BF9090BF9090BF9F90BF9F90C09F9FBF9F90C0A09F
C09F90C0A090BF9F90C09F90BF9F8FC0A08FBF9F8FBF9F8FBF9F80BF9F80BF9F80BF9F80B0907F\
B09070AF8F6FA08F609F7F5F9F7F50705F40CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F40703F4F70
4050803F5F804F6090406F904F709F4F709F507FA050709F5F7FA0507FA05F80AF5F80A06080AF\
6080AF608FB0608FAF608FB0608FAF6F90B0608FAF6F90B0608FAF6F90B06F90AF
6F90B06F90AF709FB06F90B0709FB06F9FB0709FB06F9FB0709FB06F9FAF709FB06F9FAF709FB0\
6F9FAF709FB06F9FAF6F9FAF6F90A06F9FAF6090A06090A060909F60909F5F8F90
5F90905F8F8F5F8F9050808F50808F305F50C0DFD0BFDFD0BFD0E0BFCFE0C0CFDFB0BFE0B0B0DF\
B0AFC0AFAFCFAFA0C0A09FBF9F90B09F90B08F7FA070608F4F3F406F4F507F4F5F
805060805F60906070906F70A07080A07F80AF808FAF808FB08F90B08F90BF9090B08F90BF909F\
BF9090BF9F9FBF909FC09F9FBF9F9FC0A09FC09F9FC0A09FC0A09FC0A09FC0A09F
C0A09FC0A090C0A09FC0A090C0A090C0A08FC0A090C0A08FC0A08FBF9F80C0A08FBF9F80BF9F80\
B09070B09070AF8F60A080609F7F509070505F403FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F4070
304F703F50803F5F80406F90406F904F709F4F709050709F4F709F507FA0507F9F5F80A0507FA0\
5F80A05F80A0608FAF5F80A0608FAF608FAF608FAF608FAF608FAF608FAF6F90AF
608FAF6F90AF6090AF6F90AF6090AF6F90AF6F90AF6F9FAF6F90AF6F9FAF6090A06F90AF6090A0\
6090A05F8F9F60909F5F8F905F8F9F5F8F905F8F9050808F508F8F50808F50808F
4F80803F605FAFBFB0B0C0C0B0C0C0BFCFCFBFD0CFBFD0DFB0CFE0BFCFE0BFC0D0B0BFE0AFB0D0\
AFAFCFA0A0C0A0A0C09F90B0908FAF3F30305F404F7040507F4F5F8050608F6070
9060709F707F9F7080AF808FAF7F8FAF8090AF808FB08F90B08F90B09090B09090BF909FBF909F\
BF9F9FBF9F9FC09FA0BF9F9FC0A0A0C0A09FC0A0A0C0A09FC0A09FC0A09FC0A09F
C0A09FC0A09FC0A09FC0AF9FC0A090C0A090C0A090C0A090BFA08FC0A08FBFA08FC0A08FBF9F80\
BF9F7FAF9070AF906FA080609F805F907F50806F4FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
3F4F703F507F405F8040608F4F6F90406F904F709F4F7090507F9F4F709F507F9F507F9F5F80A0\
5F80A05F80A05F80A0608FAF5F8FAF608FAF608FA06090AF608FAF6090AF608FAF
6F90AF6090AF6F90AF6090AF6F90AF6090AF6F90AF6090A06F90AF6090A06090A05F8F9F60909F\
5F8F9F5F8F9F508F905F8F9050808F508F8F50808F4F70703F6F609FB0A0A0BFAF
AFBFB0AFBFBFB0C0C0B0C0C0BFCFCFBFCFCFBFD0D0BFD0D0BFD0EFB0C0E0BFC0DFB0BFDFB0B0D0\
A0AFCFA0A0C09F90BF403F406F3F4F704F507F505F8F5F6F8F60709F6F7F9F7080
AF808FAF808FAF8090AF808FB08F90B08F90B0909FB08F90BF909FBF909FBF9FA0BF9F9FC0A0A0\
C0A0A0C0A0A0C0A0A0CFAFA0C0A0A0CFAFA0C0A0A0CFAFA0C0A0A0CFAFA0C0A09F
CFAFA0CFAF9FCFAFA0C0AF9FCFAF9FC0AF9FCFAF9FC0A090CFAF90C0A08FC0AF8FBFA08FC0A08F\
BFA080BF9F7FAF9070AF906FA08F609F805F907F4FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCF3F4F703050703F5F803F5F8040608F406F8F4F70904F70904F709F4F7090507F9F4F7F9F\
507F9F507F9F5F80A05080A05F8FA05F80A05F8FA05F8FA0608FAF5F8FA0608FA0
5F8FA0608FA05F8F9F5F8FA05F8F9F5F8FA05F8F9F5F8F9F5F8F905F8F9F5080905F8F90508090\
508F9050808F50808F4070704070703F6F6080A0908FA09F9FB0A09FB0AFA0BFB0
A0BFB0AFC0BFAFC0BFB0C0C0B0C0C0BFD0D0B0CFCFB0CFDFB0C0E0B0C0E0B0BFD0AFBFDFA0AFCF\
A0AFCF5F303F6F404F704F507F5060805F6F906F709F707FA07F8FA07F8FAF8090
AF8090B08F90AF8F90B08F9FB08F9FBF909FB0909FBF9FA0BF9F9FC0A0A0BF9FA0C0A0A0C0A0A0\
C0A0AFC0A0A0C0AFA0C0A0A0CFAFA0C0A0A0CFAFA0C0AFA0CFAFA0C0AFA0CFAFA0
C0AFA0CFAFA0C0AF9FCFAFA0C0AF9FCFAF9FC0AF90C0AF9FC0A090C0AF90BFA08FC0A08FBFA080\
BFA08FB09F7FB09F7FAF906FA08F6F9F805F908050CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCF3F40603050703F5F7F3F5F8040608F406F8F4F70904F70904F7F9F4F7090507F9F\
4F7F9F50809F507F9F5F80A050809F5F8FA05F80A05F8FA05F809F5F8FA05F809F
5F8F9F50809F5F8F9F5080905F8F9F5080905F8F9050809050809050808F508F9050808F4F7F80\
4070704070703F6F6F6F907F70908080A0908FA09F90B0A09FB0AFA0BFB0A0BFB0
AFC0BFAFBFBFB0C0C0B0C0C0B0CFCFB0CFCFBFD0D0B0CFDFB0CFE0AFBFDFB0BFDFAFB0D060404F\
704F50805F60805F6F906F7F9F7080A07F8FA07F8FAF8090A0808FAF8F90AF8F90
B0909FB08F9FBF90A0BF90A0BF9FA0BF9FA0C0A0AFC0A0A0C0A0AFC0A0A0C0AFAFC0A0AFCFAFAF\
C0A0AFCFAFAFCFAFAFCFAFAFC0AFA0CFAFAFCFAFA0CFB0AFCFAFA0CFB0AFCFAFA0
CFB0A0CFAFA0CFB0A0CFB0A0CFB0A0CFAF9FCFB09FC0AF9FCFAF9FC0AF90C0AF90C0AF8FC0AF8F\
BFA080C0AF8FB09F7FB09F70A0906FA08F609080506F5F40CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF3050703F5F7F3F5F7F4060803F6080406F8F4070904F70904F7090\
4F7F9F4F7F90507F9F4F7F9050809F507F9050809F507F9050809F50809050809F
5080905080905080905080904F808F5080904F808F50808F40707F407F7F4070704070703F706F\
4F7F6F5F8070709080709F8F80A09080A09F90AFA090B0AF9FB0B09FB0B0A0BFBF
A0BFBFAFC0C0AFBFC0B0C0CFAFC0CFB0CFCFAFC0CFB0C0DFAFBFDF5F4040604050705060805F6F\
906F7F906F7F9F70809F7080A07F8FA07F8FAF8090A08090AF8F9FAF8F9FB0909F
B0909FBF9FA0BF9FA0BFA0AFBF9FA0C0A0AFBFA0A0C0A0AFC0A0A0C0A0AFC0A0AFCFAFAFC0A0AF\
CFAFAFC0AFAFCFAFAFC0AFAFCFAFAFCFAFAFCFB0AFCFAFA0CFB0AFCFAFA0CFB0AF
CFAFA0CFB0A0CFB0A0CFB0A0CFAFA0CFB0A0C0AF9FCFB09FC0AF90C0AF90C0AF90C0AF90BFA08F\
BFAF8FBFA080BFA080B09F70AF906F9F8F609F805F8F7F4FCFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF30405F3F5F7F3F5F7F406F8F406F8F40708F4070904F7F90\
4F70904F7F904F7090507F904F7F90507F904F7F905080904F7F905080904F7F8F
5080905080905080904F7F8F50808F4F7F804F7F804F7F7F4F7F7F407F70507F70507F705F8F7F\
608F80709F8F7F9F9080A09F80A09F90AFA090B0AF9FBFB09FB0B0A0BFBFA0BFBF
AFC0C0AFBFC0AFC0CFAFC0CFB0CFD0B0C0CF5040405F40507050607F5F6F906F7F906F7F9F708F\
9F708FA07F90A07F8FAF8090AF8090AF8F9FAF8F9FB090A0B090A0BF9FAFBF9FA0
C0A0AFBFA0AFC0A0AFBFA0AFC0AFB0C0A0AFC0AFB0C0A0AFCFAFB0C0AFAFCFAFB0C0AFAFCFB0B0\
CFAFAFCFB0B0CFB0AFD0B0B0CFB0AFD0B0B0CFB0AFD0BFB0CFB0AFD0B0AFCFB0AF
D0BFAFCFB0AFD0BFAFCFB0A0D0BFAFCFB0A0CFB0A0CFB09FCFB09FC0B09FCFB09FC0AF90C0B090\
C0AF8FC0AF8FBFA080BFA080AF9F70A0906F9F805F908050CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0305070305F7F3F60803F6080406F8F406F8F\
40708F40708F4F7F904F708F4F7F904F708F4F7F8F4F7F8F507F904F7F8F50808F
4F7F8F4F7F8F4F7F804F7F804F7F7F4F7F7F4F7F7F4F7F7F507F7F5F807F5F807F608F806F9080\
709F90709F9080A09F80A09F90AFAF8FAFA090B0AF9FB0B0A0BFBF9FB0BFA0BFBF
A0BFBFAFC0C0A0BFC04F3F405F40507050607F506F8F60708F607F906F80906F809F708F9F7F8F\
A08090A08090AF8F9FAF8F9FB090A0B090A0BF9FAFB09FA0BF9FAFBF9FAFBFA0AF
BFA0AFC0A0B0BFA0AFC0A0B0C0A0AFC0AFB0C0A0AFCFAFB0C0AFAFCFAFB0C0AFAFCFB0B0CFAFB0\
CFB0B0CFB0AFCFB0B0CFB0B0D0BFB0CFB0AFD0B0B0CFB0AFD0BFB0CFB0AFD0B0AF
CFB0AFD0BFAFCFB0A0CFBFAFCFB0A0CFBFAFCFB0A0CFB0A0C0B09FCFB09FC0AF90C0B090C0AF90\
C0AF90BFAF8FBFAF8FB0A07FAF9F70A08F609F8F5F8F7F50CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F4F60305F7F3F60803F60803F6F80\
3F6F8040708F406F804F708F40708F4F7F8F4F7F8F4F7F8F4F7F8F4F7F8F4F7F80
4F7F8F4F7F805080804F7F80508080507F7F5F808050807F5F8F80608F806F908F6F908F709F90\
709F9080A0A080A0A08FAFAF8FAFAF90B0B090B0B09FBFBF9FB0BFA0BFBFA0BFBF
403F4050404F70506080506F805F708F5F70906F80906F809F708F9F708FA08090A08090AF8F9F\
AF8F9FB090AFB090A0BF9FAFB09FAFBF9FAFBF9FAFBFA0AFBF9FAFC0A0B0BFA0AF
C0AFB0C0A0B0C0AFB0C0AFB0CFAFB0C0AFB0CFB0B0CFAFB0CFB0B0CFAFB0CFB0BFCFB0B0D0B0BF\
CFB0B0D0BFBFD0B0B0D0BFB0D0B0B0D0BFB0D0B0B0D0BFB0CFB0AFD0BFB0CFBFAF
D0BFB0CFB0AFD0BFAFCFBFAFD0BFAFCFB0A0D0BFAFCFB0A0CFBFA0C0B09FCFB09FC0B090C0B090\
BFAF8FC0B090BFAF8FBFAF80AF9F70AF9F709F8F6090805F807040CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF304F6F3F5F703F5F70\
3F6F803F6F80407080406F804070804070804F7F8F4070804F7F804F70804F7F80
4F707F4F7F804F7F80508080507F805F80805F8080608F8F608F8F6F908F6F908F709F90709F90\
7FA0A080A0A080AFAF80A0A08FAFAF8FAFAF9FB0B090B0B03F303F4F3F406F405F
704F6080506F805F708F607F8F607F90708F90708F9F7F909F7F90A0809FAF8F9FB090A0AF90A0\
B090AFB090AFBF9FAFB09FAFBF9FAFBF9FAFBFA0B0BF9FAFBFA0B0BFA0B0C0AFB0
BFA0B0C0AFB0C0AFB0CFAFBFC0AFB0CFB0BFCFAFB0CFB0BFCFB0B0CFB0BFCFB0B0D0B0BFCFB0B0\
D0BFBFD0B0B0D0BFBFD0B0B0D0BFB0D0BFB0D0BFB0D0B0B0D0BFB0D0BFAFD0BFB0
CFB0AFD0BFAFCFBFAFD0BFAFCFB0A0CFBFAFCFBFA0CFBFAFCFB0A0CFBFA0C0B09FC0B09FBFAF90\
C0B090BFAF8FBFAF8FB0A080B0A07FA0906F9F90609080508F7F4FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F506F\
3F506F3F607F3F607F4070804070804F70804070804F7F8F4F7F80507F8F4F7F80
50808F50808F5F808F5F808F608F8F5F8F8F608F90608F8F6F90906F9090709F9F709F9F80A0A0\
7FA0A080AFAF80A0A08FAFB08FAFB090B0B04F30406F40506F405F7F506F7F506F
8F607F8F6F809F708F9F708FA0809FA0809FAF8FA0AF8FA0B090AFAF90A0B090AFB090AFBF9FB0\
B09FAFBFA0B0BF9FAFBFA0B0BFA0B0C0A0B0BFA0B0C0AFBFC0AFB0C0AFBFC0AFB0
CFB0BFC0AFBFCFB0BFCFB0BFCFB0BFCFB0BFD0B0BFCFB0BFD0BFBFD0B0BFD0BFBFD0B0BFD0BFBF\
D0BFBFDFBFBFD0BFB0DFBFBFDFBFBFDFBFBFDFBFB0DFC0BFD0BFB0DFC0B0D0BFB0
DFC0B0D0BFB0D0C0B0D0BFAFD0C0B0D0BFAFD0C0AFCFBFA0CFBFAFCFBFA0CFBFA0C0B09FC0B09F\
C0B090C0B090BFAF8FBFB08FB0A07FAF9F709F90609F8F608F7F4FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C03F5F703F5F703F6F7F3F6F7F40708040707F4F70804070804F7F80
4F7F8050808F50808F5F808F5F808F608F90608F8F6F90906F9090709F9F6F909F7F9FA07F9FA0\
7FA0A07FA0A080AFAF80AFAF603F4F603F50704F60704F60805F70805F708F6F80
90708F9F7F909F7F9FA080A0A0809FAF8FA0A08FA0AF90AFAF90A0B090AFAF90AFB09FB0B09FAF\
BFA0B0B09FB0BFA0B0BFA0B0C0AFBFBFA0B0C0AFBFC0AFB0C0AFBFC0AFB0C0B0BF
C0AFBFCFB0BFCFB0BFCFB0BFCFB0BFD0BFBFCFB0BFD0BFBFD0B0BFD0BFBFD0B0BFD0BFBFD0BFBF\
DFBFBFD0BFB0DFBFBFDFBFBFDFC0BFDFBFB0DFC0BFDFBFB0DFC0BFD0BFB0DFC0B0
D0BFB0DFC0B0D0BFAFD0C0B0D0BFAFD0C0AFCFBFAFCFBFAFCFBFA0CFBFA0C0B09FC0B09FC0B090\
C0B090BFAF8FBFB08FBFAF8FB0AF80A09F70A09F6F908F5F8F804FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F5F703F6F7F406F80406F7F4F7F804F7080
507F8F507F8F5F80905F808F608F905F8F90609090608F906F909F6F9090709FA0709F9F7FA0A0\
7FA0A0405F5F5F3F4F6F4F60704F607F5F70805F7F8F6F80906F8F9F7F9F9F7F9F
A080A0A080A0AF8FA0A08FA0AF90AFAF8FA0B090AFAF90AFB09FB0B09FAFBFA0B0B09FB0BFA0BF\
BFA0B0C0AFBFBFA0BFC0AFBFBFAFBFC0B0BFC0AFBFC0B0BFC0AFBFCFB0C0CFB0BF
CFB0C0CFB0BFD0BFC0CFB0BFD0BFC0D0B0BFD0BFC0D0BFBFD0BFC0D0BFBFDFBFC0DFBFBFDFBFBF\
DFBFBFDFC0BFDFBFBFE0C0BFDFBFBFE0C0BFDFC0BFE0C0BFDFBFB0E0C0BFDFC0B0
DFC0B0D0BFB0DFC0B0D0BFB0D0C0B0D0BFAFD0C0B0CFBFAFCFC0AFCFBFA0CFBFA0C0BF9FC0BF9F\
BFB090C0B090BFB08FBFB08FB0AF80AFA07F9F906F908F5F807F4FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F60703F6070406F7F
406F7F4F70804F708050808F50808F5F80905F808F608F90608F906F909F6F909F709FA0403F40\
5F4050604F5F7050707F5F7F8F60808F6F8F9070909070909F7F9F9F7F9FA080A0
A080A0AF8FAFA08FA0AF90AFAF90AFAF90AFAF90AFB09FB0B09FAFB0A0B0B09FB0BFA0BFBFA0B0\
BFAFBFBFAFBFC0AFBFBFAFBFC0AFBFC0AFBFCFB0C0C0AFBFCFB0C0CFB0BFCFB0C0
CFB0BFD0B0C0CFB0BFD0BFC0D0B0BFD0BFC0D0BFBFDFBFC0D0BFBFDFBFC0DFBFBFDFC0C0DFBFBF\
DFC0BFDFBFBFE0C0BFDFBFBFE0C0BFE0C0BFE0C0BFDFBFB0E0C0BFDFC0B0DFC0BF
D0BFB0DFC0B0D0BFB0D0C0B0D0BFAFD0C0AFCFBFAFCFC0AFCFBFA0CFBFA0C0BFA0C0BFA0BFB09F\
C0B09FBFB090BFB090B0AF8FB0AF8FA0A0709F9F6F8F805F80804FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
3F5F703F6070406F7F4F7080507F8F507F8F5F80905F8090608F9F608F904F3F4F5F40506F506F\
705F708060808F6F8F907090907090907F90907F909F80A09F80A0A08FA0A08FA0
AF90AFAF90AFB09FB0AF90B0B09FB0B09FB0B0A0BFB0A0B0BFA0BFBFA0BFBFAFBFBFAFBFC0AFC0\
BFAFBFC0B0C0C0AFBFCFB0C0C0B0C0CFB0C0CFB0C0CFB0C0CFB0C0D0BFC0CFB0C0
D0BFC0D0BFC0D0BFC0D0BFC0DFC0C0DFBFC0DFC0C0DFBFC0DFC0C0DFBFC0E0C0C0DFBFC0E0C0C0\
E0C0C0E0C0C0E0C0BFE0CFC0E0C0BFE0CFC0E0C0BFE0CFC0E0C0BFE0C0BFDFC0B0
DFC0BFDFC0B0DFC0B0D0C0B0D0C0B0D0C0B0D0C0B0CFC0AFD0C0AFCFBFA0CFC0A0C0BF9FC0BFA0\
C0BF9FC0BF9FBFB090BFB090B0AF8FAFA07F9F906F908F60807F4FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCF305060405F6F4060704F70804F30405F405F6F4F607F50707F5F7F\
80608F8F6F8F9070909070909F7F9F907F9F9F80A09F80A0A08FAFA08FA0AF90AF
AF90B0B09FB0AF9FB0B09FBFB09FB0B0A0BFB0A0BFBFA0BFB0A0BFBFAFC0BFA0BFBFAFC0BFAFBF\
C0B0C0BFAFBFC0B0C0C0AFC0CFB0C0C0B0C0CFB0C0CFB0C0D0BFC0CFB0C0D0BFC0
D0BFC0D0BFC0D0BFC0DFBFC0DFBFC0DFC0C0DFBFC0DFC0C0DFBFC0E0C0C0DFBFC0E0C0C0E0C0C0\
E0C0C0E0C0BFE0C0C0E0C0BFE0CFC0E0C0BFE0CFBFE0C0BFE0CFBFDFC0B0DFC0BF
DFC0B0DFC0BFD0C0B0D0C0B0D0C0B0D0C0B0CFC0AFCFC0AFCFC0AFCFC0AFC0BFA0C0BFA0C0BF9F\
C0BF9FBFB090BFB090B0AF8FB0B08FA0A0709F9F6F8F8F5F80804FC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF403F40603F5F6F4F6F6F4F6F7F5F7F7F5F808F6F8F\
8F6F90907F9F907F9F9F80A09F80A0A08FAFA08FAFA090AFA090AFAF9FB0AF9FB0
B09FBFAF9FB0B0A0BFB0A0BFB0A0BFB0A0BFBFAFC0BFA0BFBFAFC0BFAFBFC0AFC0BFAFC0C0B0C0\
BFAFC0C0B0CFC0B0C0CFB0C0C0B0C0CFBFCFCFB0C0D0BFCFCFB0C0D0BFCFD0BFC0
DFBFCFD0BFC0DFC0CFDFBFC0DFC0CFDFBFC0E0C0CFDFC0C0E0C0CFDFC0C0E0CFCFE0C0C0E0CFC0\
E0C0C0E0CFC0E0C0C0E0CFC0E0CFBFEFCFC0E0CFBFE0CFC0E0CFBFE0CFBFDFCFBF
DFCFBFDFC0B0DFCFBFD0C0B0D0CFB0D0C0B0D0CFB0CFC0AFCFC0B0CFC0AFCFC0AFC0BFA0C0BFA0\
BFB09FBFBF9FBFB090BFB08FB0AF80AFAF7F9F906F8F8F5F7F7F4FCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCF303030503F506040606F4F706F50707F5F7F7F60808F6F8F8F7090\
907F9F907FA09F80A09F80A0A08FAFA08FAFAF90B0A090B0AF9FB0AF9FB0B09FBF
AF9FB0B0A0BFB09FBFB0A0BFB0A0BFBFA0BFB0A0BFBFAFC0BFA0BFBFAFC0BFAFC0C0B0C0BFAFC0\
C0B0C0C0B0C0C0B0CFC0B0C0CFB0CFCFB0C0CFBFCFCFB0C0D0BFCFD0BFC0D0BFCF
D0BFC0DFC0CFDFBFC0DFC0CFDFBFC0E0C0CFDFC0C0E0C0CFDFC0C0E0C0C0DFC0C0E0CFC0E0C0C0\
E0CFC0E0C0C0E0CFC0E0C0BFE0CFC0E0CFBFEFCFC0E0CFBFE0CFBFDFCFBFE0CFBF
DFC0B0DFCFBFD0C0B0DFCFB0D0C0B0D0CFB0CFC0AFD0C0B0CFC0AFCFC0AFC0BFA0C0BFA0BFBF9F\
BFBF9FB0B090BFB090B0AF8FB0B08FA0A07F9F9F6F808F5F7F7F4FC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCF403F405F405F604F6F70507F705F7F80608F806F8F8F70908F709F9F80A0\
9F80AFA08FAF9F8FAFA090B0A090B0AF90B0A090B0AF9FBFAF9FBFB0A0BFAF9FBF
B0A0C0B0A0BFB0AFC0B0A0BFBFAFC0BFAFC0BFAFC0BFAFC0C0B0CFBFAFC0C0B0CFBFB0C0C0B0CF\
C0B0CFCFBFCFC0B0CFCFBFCFCFBFCFCFBFCFCFBFCFD0C0CFD0BFCFD0C0CFD0BFCF
DFC0CFDFC0CFDFC0CFDFC0CFE0CFD0E0C0CFE0CFCFE0C0CFE0CFCFE0CFCFE0CFCFE0CFC0EFCFCF\
E0CFC0EFCFC0E0CFC0EFD0C0E0CFC0EFD0C0E0CFC0EFD0C0E0CFBFE0D0C0E0CFBF
E0D0BFDFCFBFDFCFBFD0CFB0DFCFBFD0CFB0D0CFB0CFC0AFCFCFB0CFC0AFC0C0AFC0BFA0C0C0A0\
BFBF9FBFBF9FB0B090B0B090AFB08FA0AF7F909F6F808F5F707040CFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF403F4F504F6060506F705F807F5F80806F8F806F90907F9F8F7F9F9080A09F80AF\
9F8FAF9F8FAFA08FB0A08FB0A090B0A090B0AF9FBFAF9FB0AFA0BFAF9FBFB0A0C0
B0A0BFB0AFC0B0A0BFBFAFC0B0AFC0BFAFC0BFAFC0BFAFC0BFAFC0C0B0CFBFB0C0C0B0CFC0B0CF\
C0BFCFC0B0CFCFBFCFCFBFCFCFBFCFCFBFCFD0BFCFCFBFCFD0C0CFD0BFCFDFC0CF
D0C0CFDFC0D0DFC0CFE0CFCFDFC0CFE0CFD0E0C0CFE0CFCFE0CFCFE0CFCFE0CFC0E0CFCFE0CFC0\
EFD0CFE0CFC0EFD0C0E0CFC0EFD0C0E0CFC0EFD0C0E0D0C0EFD0C0E0CFBFE0D0BF
DFCFBFDFD0BFD0CFB0DFCFBFD0CFB0D0CFB0CFC0AFCFCFB0C0C0AFCFC0AFC0BFA0C0C0A0BFBF9F\
BFBF9FB0B090B0B090AFB08FAFB08F9FA07F90906F7F8050CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF403F505F4F6F6F50707F608F7F608F80709F80709F8F7FA08F7FA09F80AF9F80AF9F8FB0\
9F8FAFA090B0A090B0A090BFA090B0AF9FBFAF9FBFAFA0BFAF9FBFB0A0C0B0A0C0
B0AFC0B0A0C0BFAFCFB0AFC0BFAFCFBFAFC0BFB0CFBFB0CFC0B0CFBFB0CFC0BFCFC0B0CFC0BFD0\
C0B0CFCFBFD0CFBFCFCFBFD0CFBFCFD0C0D0D0C0CFD0C0D0D0C0CFDFCFD0DFC0CF
DFCFD0DFC0CFE0CFD0E0CFD0E0CFD0E0CFCFEFD0D0E0CFCFE0D0D0E0CFCFEFD0D0E0CFCFEFD0CF\
E0CFCFEFD0CFE0D0C0EFD0CFE0D0C0EFD0CFEFD0C0EFD0C0E0D0C0E0D0C0E0D0BF
E0D0BFDFCFBFDFD0BFD0CFB0D0CFBFCFCFB0D0CFB0CFC0B0CFCFB0C0C0AFC0C0AFBFBFA0BFBFA0\
B0BF9FBFBF9FB0B090B0B090AFAF809FA07F809060708050CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
4F3F505F4F6F6F507F7F608F7F608F806F90806F908F7F9F8F7F9F9080A09080AF9F8FB09F8FAF\
A090B0A090B0A09FBFA090B0AF9FBFA09FBFAFA0C0AF9FBFB0A0C0AFA0C0B0AFC0
B0A0C0B0AFC0B0AFC0BFAFCFB0AFC0BFB0CFBFAFCFBFB0CFBFB0CFC0B0CFBFB0CFC0BFD0C0B0CF\
C0BFD0C0BFCFCFBFD0C0BFCFCFC0D0CFBFCFD0C0D0D0C0CFD0C0D0D0C0CFDFCFD0
DFC0CFDFCFD0DFCFD0E0CFD0E0CFCFE0D0D0E0CFCFE0D0D0E0CFCFE0D0D0E0D0CFEFD0CFE0D0CF\
EFD0CFE0D0CFEFD0CFE0D0C0EFD0CFE0D0C0EFDFC0E0D0C0E0D0C0DFD0BFE0D0C0
DFCFBFDFCFBFD0CFB0D0CFBFCFCFB0D0CFB0CFC0B0CFCFB0C0C0AFC0C0AFBFC0A0BFC0A0B0BF9F\
BFBF9FB0B090B0B090AFAF8FAFAF80909F6F808F5F6F7040CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F3F5F
5F4F6F60507F70608F7F6F9080709F80709F8F7FA08F7FA09080AF9080AF9F8FB09F90B0A090BF\
A090BFA09FBFA09FBFAFA0C0AF9FBFAFA0C0AFA0C0B0AFCFB0A0C0B0AFCFB0AFC0
BFB0CFB0AFCFBFB0CFBFAFCFBFB0D0BFB0CFBFB0D0BFB0CFC0BFD0C0B0D0C0BFD0C0BFD0CFBFDF\
C0BFD0CFC0DFCFBFD0CFC0DFCFC0D0D0C0DFD0C0D0DFCFDFD0C0D0DFCFDFDFCFD0
E0D0DFDFCFD0E0D0D0E0CFD0EFD0DFE0D0D0EFD0D0E0D0D0EFDFD0EFD0D0EFDFD0E0D0CFEFDFD0\
E0D0CFEFDFCFE0D0CFEFDFCFE0DFCFE0DFCFE0DFC0E0DFCFE0D0C0E0DFC0DFD0BF
DFD0C0DFD0BFDFD0BFD0CFB0D0D0BFCFCFB0CFCFB0CFCFB0CFCFB0C0C0AFC0C0AFBFC0A0BFC0A0\
B0BF9FB0BF9FAFB08FAFB08FA0AF80909F6F7080506F7F40CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF40305F50406F
5F5070705F8070608F80709F80709F8F7FA08F7FA09080AF9080AF9F8FB09F8FB0A090BFA090BF\
A09FBFA09FBFAF9FC0A09FBFAFA0C0AFA0C0AFA0C0AFA0C0B0AFCFAFA0C0B0AFCF
B0AFCFBFB0CFB0AFCFBFB0CFBFB0CFBFB0D0BFB0CFBFB0D0BFB0D0C0BFD0BFB0D0C0BFD0C0BFD0\
C0BFDFC0BFD0CFC0DFCFC0D0D0C0DFCFC0D0D0C0DFD0C0D0DFCFDFD0CFD0DFCFDF
DFCFD0E0D0DFDFCFD0E0D0DFE0D0D0EFD0DFE0D0D0E0D0D0E0D0D0E0DFD0E0D0CFE0DFD0E0D0CF\
E0DFCFDFD0CFE0DFCFDFDFC0E0DFCFDFD0C0E0DFC0DFDFC0DFDFC0DFD0BFDFD0C0
D0D0BFD0D0BFCFCFB0D0D0BFCFCFB0CFCFB0C0CFB0CFCFB0C0C0AFC0CFAFBFC0A0BFC0A0B0BF9F\
B0BF9FAFB090AFB08FA0AF809FA07F809060708050CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF403F5050406F5F4F70
6F5F8070608F8070A08070A08F7FA08F7FA09080B0908FAF9F8FB09F90B0A09FBFA090BFA09FC0\
A09FBFAFA0C0A09FC0AFA0C0AFA0C0B0AFCFAFA0C0B0AFCFAFAFCFB0AFCFB0AFCF
BFB0CFB0AFCFBFB0D0BFB0D0BFB0D0BFB0D0C0BFDFBFB0D0C0BFDFBFBFD0C0BFDFC0BFDFC0C0DF\
C0BFDFCFC0DFCFC0DFCFC0DFCFC0DFD0CFDFD0C0DFD0CFDFD0CFDFDFD0DFDFCFDF
DFD0DFDFD0DFE0D0DFE0D0DFE0DFDFE0D0DFE0DFDFE0DFD0E0DFD0E0DFD0E0DFD0E0DFD0E0DFD0\
DFDFCFE0DFD0DFDFCFE0DFCFDFDFC0E0E0CFDFDFC0DFE0C0DFDFC0DFDFC0D0DFBF
D0DFC0D0D0BFD0D0BFCFD0BFCFD0BFC0CFB0CFD0B0C0CFB0C0CFB0BFC0AFBFC0AFB0BFA0B0BF9F\
AFB090AFBF90A0B08FA0B08F909F7080906060704FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F304050406F5F4F706F5F80
6F608F7F709F8070A08F7FAF8F7FA09080AF9080AF908FB0908FB09F90BF9F90BFA09FC0A09FBF\
A09FC0A09FC0AFA0C0A09FC0AFA0C0AFA0C0AFAFCFAFA0C0B0AFCFB0AFCFB0B0CF
B0AFCFB0B0D0B0B0CFBFB0D0B0B0D0BFB0D0BFB0D0C0BFDFBFB0D0C0BFDFC0BFDFC0C0DFC0BFDF\
C0C0DFC0C0DFCFC0E0CFC0DFCFC0DFCFC0DFD0CFDFCFCFDFD0CFDFD0CFDFDFD0DF
D0D0DFDFD0DFDFD0DFE0DFDFDFD0DFE0DFDFDFDFD0E0DFDFDFDFD0E0DFD0DFDFD0DFDFD0DFDFCF\
DFDFCFDFDFCFDFDFCFD0DFC0DFDFCFD0DFC0DFE0CFD0DFC0D0DFC0D0DFBFD0DFC0
CFD0BFCFDFBFCFD0BFCFD0BFC0CFB0C0CFB0C0CFB0C0CFB0BFC0AFBFC0AFB0BFA0B0BFA0AFB090\
AFB090A0B08FA0B08F9FAF808F9F706F8050CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF50406F5F4F70605F806F608F
7F709F8070A08F7FAF8F7FAF9080B09080B0908FB0908FB09F90BF9F90BFA09FC0A09FBFA0A0C0\
A09FC0AFA0C0A0A0C0AFAFCFAFA0CFAFAFCFAFAFCFB0AFD0B0AFCFB0B0D0B0AFCF
BFB0D0B0B0D0BFB0D0BFB0D0BFBFDFBFBFD0C0BFDFBFBFDFC0C0DFC0BFDFC0C0E0C0BFDFCFC0E0\
C0C0E0CFC0E0C0C0E0CFCFE0CFCFE0D0CFE0CFCFDFD0D0E0D0D0DFD0D0E0D0D0DF
DFDFE0DFDFDFDFDFE0DFDFDFE0E0E0DFDFDFE0E0DFDFDFDFE0E0DFDFE0D0DFE0DFDFDFD0DFE0D0\
DFE0D0DFE0D0D0DFCFDFE0CFD0E0CFDFE0CFD0E0C0DFE0CFD0E0C0D0E0C0CFDFC0
D0DFC0CFDFBFCFDFBFC0D0BFCFD0BFC0CFB0C0CFB0BFCFAFC0CFB0BFC0AFBFC0A0AFBF9FB0BF9F\
A0B090A0B0909FAF809FAF807F90605F7040CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0403F50504070605080605F8F706F90
7F6F9F807FAF807FA08F80AF8F80AF908FB0908FB09F90BF9F90BF9F9FC09F9FBFA09FC0A09FC0\
A0A0C0A09FC0AFA0CFAFA0C0AFAFCFAFA0CFAFAFCFAFAFCFB0B0D0AFAFCFB0B0D0
B0B0D0BFB0D0B0B0D0BFBFDFBFB0D0BFBFDFBFBFDFBFBFDFBFBFDFC0C0E0BFBFDFC0C0E0C0C0E0\
C0C0E0C0C0E0CFCFE0C0CFE0CFCFE0C0CFDFCFD0E0CFD0DFD0D0E0CFD0DFD0DFE0
D0DFDFD0DFE0D0DFDFDFE0E0DFDFDFDFE0E0DFE0DFDFE0DFDFE0DFDFE0DFD0E0D0DFE0D0D0E0D0\
DFE0D0D0E0CFD0E0D0D0E0CFD0E0CFD0E0C0D0E0CFD0E0C0D0E0C0CFDFBFCFDFC0
CFDFBFCFDFBFC0D0B0C0D0BFC0CFB0C0CFB0BFCFAFBFCFB0BFC0AFBFC0AFAFBF9FAFBF9FA0B090\
A0B0909FAF8F9FAF80809F70607F50C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F404F406F5F4F7F6050806F6090706F9F
807FA0807FAF8F80B08F80AF908FB0908FB09F90BF9090BF9F9FC09F9FBFA0A0C0A09FC0A0A0CF\
A0A0C0AFAFCFAFA0CFAFAFCFAFAFCFB0AFD0AFAFCFB0B0D0B0AFD0B0B0D0B0B0D0
B0B0DFB0B0D0BFBFDFBFBFDFBFBFDFBFBFDFBFC0E0BFBFE0C0C0E0BFBFE0C0C0EFC0C0E0C0C0EF\
C0C0E0CFCFEFC0CFEFCFCFEFC0CFE0CFD0EFCFD0E0CFD0E0CFD0E0D0DFE0CFDFE0
D0E0E0D0DFE0D0E0E0D0E0E0DFE0E0D0E0DFDFEFE0D0E0DFDFEFDFD0E0DFDFEFDFD0E0D0D0EFDF\
D0E0D0D0EFD0D0E0CFD0E0D0D0E0CFD0E0CFCFE0C0D0E0CFCFDFC0CFE0C0CFDFBF
CFDFC0C0D0BFCFD0BFC0D0B0C0D0BFBFCFB0C0CFB0BFC0AFBFC0AFB0C0A0B0C0A0AFBF9FAFBF9F\
9FB08F9FB08F90A080708F604F6F3FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F3F50504F7F5F4F80605F8F6F60907F70A0
7F70A08080AF8080AF8F8FB08F8FB0908FBF9090BF9F90BF9F90BF9F9FC09F9FC0A0A0CFA09FC0\
A0A0CFA0A0CFAFAFCFA0A0CFAFAFCFAFAFCFAFAFD0AFAFCFB0B0D0AFB0D0B0B0DF
B0B0D0B0B0DFB0B0DFBFBFE0B0BFDFBFBFE0BFBFE0BFC0E0BFBFE0BFC0E0BFC0E0C0C0EFBFC0EF\
C0C0EFC0C0EFC0CFEFC0CFEFCFD0EFC0CFE0CFD0EFC0D0E0CFDFEFCFDFE0CFDFE0
CFDFE0D0E0E0CFE0E0D0EFE0D0E0DFD0EFE0D0EFDFD0EFE0D0E0DFD0EFDFD0E0D0D0EFDFCFE0D0\
D0E0D0CFE0CFD0E0D0CFE0CFCFE0CFCFDFC0CFE0CFCFDFC0CFDFC0C0DFBFCFDFC0
C0D0BFC0D0BFBFD0B0C0D0BFBFCFB0BFCFB0BFC0AFBFCFB0B0C0A0B0C0A0AFBF9FA0BF9F9FAF8F\
9FAF8F8FA0807F9F7050704FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF4F4060504F80605F8F6F6090706F9F7F70A0
8F80B08F80B08F8FB08F8FB09090BF9090BF9F9FC09F90BFA0A0C09F9FC0A0A0CFA0A0C0AFAFCF\
A0A0CFAFAFCFA0AFCFAFAFD0AFAFD0AFB0D0AFAFD0B0B0DFB0B0D0B0B0DFB0B0DF
BFBFDFB0BFDFBFBFE0B0BFDFBFC0E0BFBFE0BFC0EFBFC0E0C0C0EFBFC0EFC0C0EFBFC0EFC0CFF0\
C0CFF0C0CFF0C0CFEFCFD0F0C0D0EFCFD0F0C0D0EFCFDFF0CFDFEFCFDFEFCFDFEF
D0E0EFCFE0EFD0EFEFD0E0E0D0EFEFD0EFE0D0EFE0D0EFDFD0EFE0D0EFDFD0EFDFCFE0D0D0EFDF\
CFE0D0D0E0D0CFE0CFD0E0D0CFE0CFCFE0CFCFDFC0CFE0CFC0DFC0CFDFC0C0D0BF
C0DFC0C0D0BFC0D0BFBFCFB0BFD0BFBFCFB0BFCFB0B0C0AFB0C0AFAFBFA0A0BF9F9FB0909FB08F\
90AF808FA08060805FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF3F3F404F406F5F508F605F8F6F6F9F706F9F7F7FAF
8080B08F8FB08F8FB0908FBF8F8FBF9090BF9090BF9F9FC09F9FC0A0A0C09F9FC0A0A0CFA0A0CF\
A0AFCFA0A0CFAFAFD0A0AFCFAFAFD0AFAFD0AFB0D0AFAFD0B0B0DFAFB0D0B0B0DF
B0B0DFB0BFE0B0B0DFB0BFE0B0BFE0BFC0EFB0BFE0BFC0EFBFC0EFBFC0F0BFC0EFBFCFF0BFC0F0\
C0CFF0BFCFF0C0CFF0C0CFF0C0D0F0C0D0EFCFDFF0C0D0EFCFDFF0C0DFEFCFE0EF
CFDFEFCFE0EFCFE0E0CFEFE0CFE0E0D0EFE0CFE0DFD0EFDFCFE0DFCFEFDFCFE0D0CFE0DFCFE0D0\
CFE0D0C0DFCFCFE0CFC0DFCFCFE0CFC0DFC0C0DFC0C0DFC0C0DFC0BFD0BFC0D0BF
BFD0BFBFD0BFBFCFB0BFCFB0B0CFB0BFCFB0B0C0AFB0C0AFAFBFA0A0BF9F9FB09090AF8F8FA080\
8FA080708F60CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF3F3F505F50805F5F8F6F609F6F6F9F7F7FAF7F7FAF
8F8FB08080B08F8FBF8F8FBF9090BF9090BF9F9FC09F9FC09FA0C09F9FC0A0A0CFA0A0CFA0AFCF\
A0A0CFAFAFD0A0AFD0AFAFD0AFAFD0AFB0DFAFB0D0B0B0DFAFB0DFB0BFDFB0B0DF
B0BFE0B0BFE0BFBFE0B0BFE0BFC0EFB0BFE0BFC0EFBFC0EFBFC0F0BFC0EFC0CFF0BFCFF0C0CFFF\
BFCFF0C0D0FFC0CFF0C0D0F0C0D0F0CFDFF0C0D0F0CFDFF0C0DFEFCFE0EFC0DFEF
CFE0EFCFE0E0CFEFEFCFE0E0CFEFE0CFEFDFCFEFE0CFE0DFCFEFDFCFE0DFCFEFDFCFE0D0CFE0D0\
C0E0D0CFE0D0C0DFCFCFE0CFC0DFCFC0E0CFC0DFC0C0DFC0BFD0C0C0DFC0BFD0BF
BFD0BFBFCFB0BFD0BFB0CFB0B0CFB0B0C0AFB0C0AFAFBFA0A0BFA09FB0909FB0908FAF808FA080\
7F9F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0404060505080605F906060906F6F9F7070A08080AF
8080B08F8FB08F8FB08F90BF8F8FBF9090BF9090BF9F9FC09F9FC09FA0CF9FA0C0A0A0CF9FA0CF\
A0AFCFA0A0CFAFAFD0A0AFD0AFAFD0AFAFD0AFB0DFAFAFD0AFB0DFAFB0DFB0BFE0
AFB0DFB0BFE0B0BFE0B0BFEFB0BFE0B0C0EFB0BFEFBFC0F0B0C0EFBFC0F0BFC0F0BFCFF0BFC0F0\
BFCFFFBFCFF0C0D0F0BFCFF0C0D0F0BFD0EFC0D0F0BFD0EFC0DFEFC0DFEFC0DFEF
C0DFE0C0E0E0C0E0E0CFE0E0C0E0DFCFEFE0C0E0DFCFEFDFC0E0D0C0E0DFC0E0D0C0E0D0C0DFCF\
C0E0D0C0DFCFC0DFCFBFDFC0C0DFCFBFDFC0C0DFC0BFD0BFBFD0C0BFD0BFBFD0BF
B0CFB0B0CFB0B0C0B0B0CFB0AFC0AFAFC0AFAFBFA0A0BFA09FB09090B0908FAF8F8FA0807F9F70\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F3F404F4F605F5F8F5F5F906F6F9F6F70A07F7FAF7F80AF
8F8FBF8F8FBF8F90BF8F90BF909FC09090C09F9FC0909FC09FA0CF9FA0CFA0AFCFA0A0CFA0AFD0\
A0AFD0AFAFD0A0AFD0AFB0DFAFAFD0AFB0DFAFB0DFB0B0E0AFB0DFB0BFE0AFB0E0
B0BFE0B0BFE0B0BFEFB0BFE0BFC0EFB0C0EFBFC0EFB0C0EFBFCFF0BFC0F0BFCFF0BFCFF0C0CFFF\
BFCFF0C0D0FFBFCFF0C0D0F0BFD0F0C0D0F0BFD0EFC0DFF0C0DFEFC0DFEFC0DFE0
C0E0EFC0E0E0C0E0E0C0E0E0CFEFE0C0EFDFCFEFE0C0E0DFCFEFDFC0E0D0C0E0DFC0E0D0C0E0D0\
C0DFCFC0E0D0BFDFCFC0DFCFBFDFCFC0DFCFBFD0C0BFDFC0BFD0BFBFD0BFB0CFB0
B0D0BFB0CFB0B0CFB0AFC0AFAFC0AFAFC0AFA0C0A09FB09F9FB09F8FAF8F8FA0807F9F70CFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF30303F4F4F6F50508F606090606F9F7070A07070A08080B0
808FB08F8FBF8F8FBF8F90BF8F90BF909FC0909FC09F9FC09F9FC09FA0CF9FA0CFA0AFCFA0A0CF\
A0AFD0A0AFCFA0AFD0A0AFD0AFB0DFA0AFD0AFB0DFAFB0DFAFB0E0AFB0DFB0BFE0
AFBFE0B0BFE0AFBFE0B0BFEFB0BFE0B0C0EFB0BFEFB0C0EFB0C0EFBFCFF0B0C0EFBFCFF0BFCFF0\
BFCFF0BFCFF0BFD0F0BFCFEFBFD0F0BFD0EFBFD0EFBFD0EFC0DFEFBFD0E0C0DFEF
BFDFE0C0E0E0BFDFDFC0E0E0C0E0DFC0EFDFBFE0DFC0E0DFBFE0D0C0E0DFBFDFD0C0E0D0BFDFCF\
C0DFD0BFDFCFBFDFCFBFD0C0BFDFCFBFD0C0BFD0C0B0CFBFB0D0BFB0CFB0B0CFB0
AFC0B0AFCFB0A0C0AFAFC0AFA0BFA0A0C0AF9FB09F90B09F8FAF8F80A0806F9070CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F3F404F4F6F5F60906060906F6FA06F70A07F7FAF7F80B0
8F8FBF808FBF8F90C08F90BF909FC0909FC09FA0CF909FC09FA0CF9FA0CFA0AFD09FA0CFA0AFD0\
A0AFCFA0AFD0A0AFD0AFB0DFA0AFD0AFB0DFAFB0DFAFB0DFAFB0DFB0BFE0AFBFE0
B0BFE0AFBFE0B0C0EFB0BFE0B0C0EFB0BFE0B0C0EFB0C0EFB0C0EFB0C0EFBFCFF0B0CFF0BFCFF0\
BFCFF0BFD0F0BFCFF0BFD0F0BFD0EFBFD0F0BFD0EFBFDFEFBFD0E0C0DFEFBFDFE0
C0E0E0BFDFE0C0E0E0BFE0DFC0E0E0BFE0DFC0E0DFBFE0DFC0E0DFBFDFD0C0E0D0BFDFD0BFDFD0\
BFDFCFBFDFCFBFD0C0BFD0CFB0D0C0BFD0C0B0CFBFB0D0BFB0CFBFB0CFBFAFC0B0
AFCFB0A0C0AFAFC0AFA0BFAFA0C0AF9FBF9F90B09F80A08F7F9F806F9070CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCF3030404F4F70505F90606F9F606F9F7070A0707FAF808FB0
808FB08F90BF8F90BF8F90C08F90BF909FC0909FC09FA0CF909FC09FA0CF9FA0CFA0AFCF9FA0CF\
A0AFD0A0AFD0A0AFD0A0AFD0A0B0DFA0AFD0AFB0DFA0B0DFAFB0DFAFB0DFAFBFE0
AFB0DFAFBFE0AFBFE0B0C0E0AFBFE0B0C0EFAFC0E0B0C0EFB0C0EFB0C0EFB0C0EFB0CFF0B0C0EF\
BFCFF0B0CFF0BFD0F0B0CFEFBFD0EFB0D0EFBFD0EFB0D0E0BFDFEFBFDFE0BFDFE0
BFDFDFBFDFE0BFDFDFBFE0DFBFDFDFBFE0DFBFDFD0BFE0D0B0DFD0BFDFD0B0DFCFBFDFD0B0D0CF\
BFD0CFB0D0C0BFD0CFB0CFC0B0D0C0B0CFBFB0CFBFAFC0B0AFCFBFAFC0B0AFC0B0
A0BFAFA0C0AF9FBFA09FBFA090B09F8FAF907F9F80709F7F608F6FCFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F3F504F4F705F609F606F9F6F70A06F7FAF7F80B07F80B0
8F90BF8090BF8F90C08F90BF909FC0909FC090A0CF909FC09FA0CF9FA0CF9FAFCF9FA0CFA0AFD0\
A0AFD0A0AFD0A0AFD0A0B0DFA0B0D0AFB0DFA0B0DFAFBFE0AFB0DFAFBFE0AFBFE0
B0BFE0AFBFE0B0C0E0AFBFE0B0C0EFB0C0EFB0C0EFB0C0EFB0CFEFB0C0EFB0CFF0B0CFEFBFCFF0\
B0CFF0BFD0F0B0CFEFBFD0F0BFD0EFBFD0EFB0D0EFBFDFEFBFD0E0BFDFE0BFD0E0
BFDFE0BFDFDFBFDFDFBFDFDFBFE0DFBFDFD0BFDFD0B0DFD0BFDFD0B0DFCFBFDFD0B0D0CFBFDFCF\
B0D0CFB0D0CFB0CFC0B0D0C0B0CFC0B0CFC0AFCFBFB0CFBFAFC0B0AFC0B0A0BFAF
A0C0AF9FBFA09FBFA090B09F8FAF907F9F8070907F5F8F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCF3F3F4F4F5070505F90606F9F606F9F6F7FAF707FAF7F80B0
808FBF8090BF8090BF8F90C08F90BF909FC08F9FC090A0CF90A0C09FA0CF9FA0CF9FAFCF9FAFCF\
A0AFD09FAFCFA0AFD0A0AFD0A0B0DFA0AFD0A0B0DFA0B0DFAFBFE0A0B0DFAFBFE0
AFBFE0AFBFE0AFBFE0AFBFE0AFBFE0B0C0EFAFC0E0B0C0EFAFC0EFB0CFEFB0C0EFB0CFEFB0CFEF\
B0CFF0B0CFEFB0CFEFB0CFEFB0D0EFB0CFE0B0D0E0B0CFE0BFD0E0B0D0DFB0D0DF
B0D0DFBFDFDFB0D0D0BFDFDFB0D0D0BFDFD0B0D0CFB0D0D0B0D0CFB0D0CFB0CFCFB0D0CFB0CFC0\
B0D0C0AFCFC0B0CFC0AFCFBFB0CFC0AFC0BFAFC0BFA0BFB0A0C0B09FBFA09FBFA0
90B0A090B0A08FAF9080AF90709F806F907FC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCF3F3F4F404F6F5F60905F6F9F6070A06F70A07080B07F80B0
808FBF808FBF8F90C08F90BF8F9FC08F9FC090A0CF909FC090A0CF90A0CF9FAFCF9FAFCF9FAFD0\
9FAFCFA0B0D09FAFD0A0B0DFA0AFD0A0B0DFA0B0DFAFB0DFA0B0DFAFBFE0AFBFE0
AFBFE0AFBFE0AFC0E0AFBFE0AFC0E0AFBFE0B0C0EFAFC0E0B0C0EFAFC0E0B0CFEFB0C0EFB0CFEF\
B0CFEFB0CFEFB0CFE0B0CFEFB0CFE0B0CFE0B0CFE0B0CFE0B0CFDFB0D0E0B0CFDF
B0D0DFB0D0D0BFD0DFB0D0D0B0D0D0B0D0CFB0D0D0B0D0CFB0D0CFB0CFCFB0D0CFAFCFC0B0CFC0\
AFCFC0B0CFC0AFC0BFAFCFBFA0C0BFA0C0BF9FBFB09FBFAF90B0A090BFA08FB09F
8FB09F80AF9080AF906F907F4F705FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0404F6F4F5F8F5F6F9F5F6F9F6F70A06F7FAF7F80B0
7F8FB08090BF8090BF8F90C08090BF8F9FC08F9FC090A0CF909FC090A0CF90A0CF9FAFCF9FA0CF\
9FAFD09FAFD0A0B0D09FAFD0A0B0DFA0B0D0A0B0DFA0B0D0A0B0DFA0B0DFAFBFE0
A0B0DFAFBFDFA0BFDFAFBFE0A0BFDFAFBFE0AFBFDFAFC0E0AFBFDFAFC0E0AFC0E0B0C0E0AFC0E0\
B0CFE0AFC0E0B0CFE0AFC0DFB0CFE0AFC0DFB0CFDFAFC0DFB0CFDFAFCFD0B0CFDF
AFCFD0B0CFD0AFCFD0B0D0D0AFCFCFB0CFCFAFCFCFB0CFCFAFC0C0AFCFC0AFC0C0AFCFC0A0C0BF\
AFC0BFA0C0BFA0C0BFA0BFB0A0BFB09FBFAF9FBFAF8FB0A08FB0A080AF9F80AF9F
7FA0906F90804F6F5FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF3F4060505F805F6F9F6070A06F70A07080AF7080B0
8090BF8090BF8090C08090BF8F9FC08F9FC090A0CF8F9FC090A0CF90A0CF9FAFCF90AFCF9FAFD0\
9FAFCF9FB0D09FAFD0A0B0D09FB0D0A0B0DFA0B0D0A0BFDFA0B0DFAFBFDFA0B0DF
AFBFE0A0BFDFAFBFDFA0BFDFAFC0DFAFBFDFAFC0DFAFBFDFAFC0E0AFC0DFAFC0E0AFC0DFB0CFE0\
AFC0DFB0CFE0AFC0DFB0CFE0AFCFDFB0CFDFAFCFD0B0CFDFAFCFD0B0CFD0AFCFD0
B0CFD0AFCFCFB0CFD0AFCFCFB0CFCFAFCFCFAFCFCFAFC0C0AFCFC0A0C0C0AFC0C0A0C0BFA0C0BF\
A0C0BFA0C0BF9FBFB09FC0B090BFAF90BFAF8FB0A08FB0A080AF9F7FA0906F9F80
507F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCF3F40604F507F5F6F9F5F6F9F6F7FAF6F7FAF7080B0
7F8FB08090BF7F90BF8090C08090BF8F9FC08F9FC08FA0C08FA0C090A0CF90A0C090AFCF90AFCF\
9FAFCF90AFCF9FAFD09FAFCF9FB0D09FAFD0A0B0D09FB0D0A0B0DFA0B0D0A0BFDF
A0B0D0A0BFDFA0B0D0A0BFDFA0BFD0AFBFDFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFAFC0DF\
AFC0DFAFC0D0AFC0DFAFC0D0AFCFD0AFC0D0AFCFD0AFC0CFAFCFD0AFC0CFAFCFCF
AFC0CFAFCFCFA0C0C0AFCFCFA0C0C0AFCFC0A0C0C0A0C0C0A0C0BFA0C0BF9FBFB09FC0BF9FBFB0\
9FBFB090BFAF9FBFB090B0AF90B0AF80AF9F7FA0906F9F80609080507F6FCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCF303F5F4F507F50609F5F70A06070A06F80AF6F80AF
7F8FB07F8FBF8090C08090BF809FC0809FC08FA0C08F9FC090A0CF8FA0C090AFCF90AFCF9FAFCF\
90AFCF9FAFD09FAFCF9FB0D09FAFCFA0B0D09FB0D0A0B0D09FB0D0A0BFDFA0B0D0
A0BFDFA0B0D0A0BFDFA0BFD0A0BFDFA0BFD0AFC0DFA0BFD0AFC0DFA0BFD0AFC0DFAFC0D0AFC0DF\
AFC0D0AFC0DFAFC0D0AFC0D0A0C0CFAFC0D0A0C0CFAFC0D0A0C0CFAFC0CFA0C0CF
AFC0CFA0C0C0AFCFCFA0C0C0A0C0C0A0C0BFA0C0C09FBFBF9FC0BF9FBFB09FC0BF90BFB09FBFB0\
90B0AF90BFB080AFA07FA09F6F9F8F6F90805F8F7F508070CFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F406F50608F506F9F5F70A06070A06F80AF
6F80AF7F8FBF7F90BF8090BF7F90BF809FC0809FBF8F9FC08F9FC08FA0C08FA0C090AFCF90A0C0\
90AFCF90AFCF9FAFCF90AFCF9FAFCF90AFCF9FB0D09FAFCF9FB0D09FB0CFA0BFD0
9FB0D0A0BFD09FB0D0A0BFD09FB0D0A0BFD0A0BFD0A0BFD0A0BFCFA0BFD0A0BFD0A0C0D0A0BFCF\
A0C0D0A0BFCFA0C0D0A0BFCFA0C0CFA0BFCFA0C0CF9FBFC0A0C0CF9FBFC0A0C0CF
9FBFC0A0C0C09FBFC0A0C0C09FBFBF9FBFBF90BFBF9FBFBF90B0B090BFB08FB0AF8FB0AF80AFA0\
80AFA0709F906F9F8F5F8F7F508070C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F406040507F5F6F9F5F70A0607FAF607FAF
708FB0708FB07F90BF7F90BF809FC0809FBF809FC0809FC08FA0C08FA0C090A0CF8FA0C090AFCF\
90AFCF90AFCF90AFCF9FB0D090AFCF9FB0D09FB0CF9FB0D09FB0D09FBFD09FB0D0
A0BFD09FB0D0A0BFD09FB0CFA0BFD09FBFCFA0BFD09FBFCFA0BFD0A0BFCFA0C0D09FBFCFA0C0D0\
A0BFCFA0C0D09FBFCFA0C0CF9FBFCFA0C0CF9FBFC0A0C0CF9FBFC0A0C0C09FBFC0
A0C0C09FBFC09FC0C09FBFBF9FBFBF90B0B090B0B080AFAF8FB0AF7FAFA07FAFA070A09070A090\
60908FB09F8FA08F70CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0404F704F5F8F5F70A05F70A0607FAF
6F80AF708FB0708FB07F90BF7F90BF809FC07F90BF809FC0809FC08FA0C08FA0C08FA0CF8FA0C0\
90AFCF8FA0C090AFCF90AFCF90AFCF90AFCF90B0CF90AFCF9FB0CF90B0CF9FB0CF
9FB0CF9FB0CF90B0CF9FB0CF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FB0CF9FBFCF9FBFCF\
9FBFCF9FBFC09FBFCF9FBFC09FBFC090BFC09FBFC090BFC09FBFC090BFBF90BFBF
90B0BF90BFBF8FB0B08FB0B080AFAF80AFAF70A09F70A09F6F9F906F9F9060908FCFBFB0C0B0AF\
CFB0AFBFA090B09F80A08F6FCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F3F604F5080506F9F5F7FA05F7FA0
6F80AF6F80AF7090BF7F90BF7F90BF7F90BF809FC0809FBF80A0C0809FC08FA0C08FA0C08FAFCF\
8FA0C090AFCF8FAFC090AFCF8FAFC090B0CF90AFCF90B0CF90AFCF9FB0CF90B0CF
9FB0CF90B0CF9FB0CF90B0CF9FB0CF90B0C09FBFCF90B0CF9FBFCF90B0C09FBFCF90BFC09FBFCF\
90BFC09FBFCF90BFC09FBFC090B0BF90BFC08FB0BF90B0BF8FB0B08FB0B080AFB0
80B0B080AFAF80AFAF7FA0A070A0A06F9F906F9F90507F70406050CFBFBFD0BFBFCFBFB0CFBFB0\
C0B0AFCFBFAFBFAF90AF9F7F907F50CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3F40704F608F506F9F5F7FA0
5F7FA06080AF6F80AF708FB0708FB07F90BF7F90BF7F9FBF7F90BF809FBF809FBF80A0C080A0BF\
8FA0C080A0C08FAFC080A0C08FAFC08FAFC08FAFC08FAFC090AFC08FAFC090B0CF
8FAFC090B0CF90B0C090B0CF8FB0C090B0C090B0C090B0C090B0C090B0C090B0C090B0C08FB0BF\
90B0BF8FB0BF8FB0BF80AFB080AFB080AFAF80AFB07FA0AF7FAFAF7FA0A07FA0A0
70A09F70A09F608F8F507F703F5F50CFBFC0CFBFBFD0BFC0CFBFBFD0BFBFD0BFBFD0C0BFCFBFB0\
CFBFAFC0B0A0C0BFA0AF9F809F8060C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF404F704F609050709F
5F7FA05F7FA06F80AF6F8FAF7090B07090B07F90BF7090BF7F9FBF7F9FBF80A0C07F9FBF80A0C0\
80A0C08FAFC080A0C08FAFC08FAFC08FAFC08FAFC08FAFC08FAFC08FB0C08FAFC0
90B0CF8FB0C090B0C08FAFC08FB0C08FAFBF8FB0C08FAFBF8FB0BF80AFBF8FB0BF80AFB08FB0BF\
80AFB080AFB07FAFAF80AFB07FA0AF7FA0AF70A0A070A0A06F9090608F8F507F7F
406F60C0B0C0CFBFC0CFBFC0D0C0C0D0BFC0DFC0C0D0C0C0DFC0C0D0C0BFDFC0C0D0C0BFDFC0BF\
D0BFB0CFC0B0C0BFA0C0B0A0A090708F7F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCF3F4F704F6F90
50709F5F7FA05F7FA06080AF608FAF7090B06F8FB07090B07090B07F9FBF709FB07F9FBF7F9FBF\
80A0BF7FA0BF80A0C080A0BF80A0C080A0BF80A0BF80A0BF80AFBF80A0BF80AFBF
80A0BF80AFBF80A0BF80AFBF80A0B080AFBF7FA0B07FAFB07FA0B07FAFB07FA0AF7FA0AF70A0AF\
7FA0AF70A0A06F9F9F608F8F5F80804F7F7040706FB09FAFBFAFBFBFB0BFC0B0C0
C0B0C0CFBFC0CFBFC0D0C0C0D0BFC0DFC0CFDFC0C0DFC0C0DFC0C0DFC0C0D0C0BFDFC0C0D0C0BF\
D0C0BFCFBFAFCFBFAFC0B0A0B0AF8F8F7F50CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF3F406F40507F
4F6F9050709F5F7FA05F80A0608FAF608FAF7090B06F90B07090B07090B0709FBF709FB07F9FBF\
709FB07FA0BF7F9FBF7FA0BF7F9FB07FA0BF7FA0B07FA0BF7FA0B07FA0BF7FA0B0
7FA0B07FA0B07FA0B07FA0B07FA0B070A0AF7FA0B070A0AF70A0AF6F909F6F909F5F8F905F8F8F\
4F80804F7F7F403F409F8090AF90A0BFA0BFBFAFC0C0B0C0C0B0C0CFBFCFCFBFC0
D0C0CFD0BFCFDFC0CFDFC0CFDFC0CFDFC0C0E0CFCFE0CFC0E0CFCFE0C0C0E0CFC0DFCFC0DFCFC0\
D0C0BFD0C0B0CFBFAFC0BFAFBFB09F908F6FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C03F4F70
3F5F7F4F6F904F7090507F9F507F9F5F80A06080A06F8FAF608FAF6F90AF6F8FAF6F90AF6F90AF\
6F90AF6F90AF709FB06F90AF709FB0709FAF709FB06F9FAF709FAF709FAF709FAF
6F9FAF709FAF6F90A06F90A05F8F905F8F9050808F508080407F7F407F7FAFC0E0B0C0EFAFBFDF\
805F708F6F80AF90AFB0AFBFBFB0C0BFB0C0C0BFCFC0B0C0CFBFCFCFBFCFD0C0CF
D0BFCFD0C0CFD0C0CFDFCFCFDFC0CFE0CFCFE0CFCFEFCFCFE0CFC0E0CFC0DFCFC0DFCFC0D0C0BF\
D0CFBFCFC0B0CFC0B0C0BFA0BFBF9F9F90706F603FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCF3F507040608F4F7090507F9F507F9F5F80A05F809F608FAF608FA06090AF608FA06F90AF\
6090AF6F90AF6F90AF6F9FAF6F90AF6F90A0608F9F6090A05F8F9F5F8F9050808F
508F904F808F4F80804F7F7FAFC0C0AFC0C0AFCFD0AFCFD0B0CFDFAFCFE0B0CFE07F6070AF9FAF\
B0A0BFBFAFC0BFAFC0C0B0CFC0B0CFCFBFCFC0BFCFCFC0D0CFC0CFD0C0D0D0C0D0
DFCFD0DFCFD0E0CFD0DFCFD0E0D0D0E0CFCFEFD0D0E0D0CFEFD0CFE0D0CFE0D0CFDFCFC0DFD0C0\
D0CFBFD0CFBFCFC0B0CFC0AFBFBFA0AFAF8F6F6F40CFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCF3050703F607F406F804F7F904F7F90507F9F50809F5F809F5F809F5F8F9F50809F\
5F809F50809050809050808F5080904F7F8F50808F4F7F804F7F80305F5080AFA0
8FAFAF9FBFB09FBFBFA0C0C0A0BFC0AFC0CFA0C0CFAFC0D05F4F50A08FA0AF90AFB0A0BFB0AFC0\
BFB0CFBFB0C0C0BFCFC0BFCFC0BFD0C0BFCFCFC0D0CFC0D0D0C0D0D0C0D0D0CFD0
D0CFD0DFCFDFDFCFD0E0D0D0DFCFD0E0D0D0E0D0CFE0D0CFE0D0CFE0DFCFDFD0C0DFD0C0DFCFC0\
D0D0C0CFCFBFCFCFB0C0BFAFBFBFA0B0B0906F6F40CFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCF3F607F3F607F40708F4F708F4F7F8F4F708F4F7F8F4F7F8F4F7F8F\
4F7F8F50808F4F7F8040707F3F6F6F3F6F6F5080706F908F709F9080AFAF8FAFAF
90B0B090B0BFA0BFC0A0BFC0A0C0C0A0C0CF806F80A08FA0AF9FB0B0A0BFBFB0CFBFB0CFBFBFCF\
BFB0CFC0BFD0C0BFD0CFC0D0C0BFD0CFC0DFCFC0DFD0CFDFD0C0DFD0CFDFD0CFDF
DFD0DFDFCFDFE0D0DFDFD0D0E0DFDFE0D0D0E0DFD0E0DFD0EFDFD0E0DFCFE0DFCFDFD0C0DFD0C0\
D0CFBFD0D0C0CFCFB0C0C0AFBFBFA0B0BF9F60604FCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF3040503F6F7F3F6F7F40708040708040708040707F\
40707F4070704F7F7F4F7F7F5F8F8F608F8F6F9F90709F9F80A0AF80AFAF8FB0B0
8FAFB090B0BF90B0BF9FBFC07F5F7F90809FAF9FB0B0AFC0B0AFC0BFB0CFB0B0CFBFBFD0BFB0D0\
C0BFD0C0BFD0C0C0DFC0BFD0CFC0DFC0C0DFCFCFE0CFC0DFD0CFDFD0CFDFD0D0DF
D0CFDFDFD0DFDFD0DFDFDFDFDFD0D0DFDFD0DFDFD0E0DFD0DFDFCFE0E0CFDFDFCFDFDFCFD0D0C0\
D0D0C0CFCFBFCFCFBFBFC0AFB0BFA0AFB09060604FC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF305F6F3F6F7F3F6F7F4F707F\
4F7F8050808F50808F608F906090906F9F9F6F9F9F7FA0AF80A0AF8FAFB08FAFB0
90B0BF8FB0BF7F6F80A090B0AFA0C0AFA0C0B0B0CFB0B0CFBFBFD0BFB0D0BFBFDFBFBFD0C0C0DF\
C0BFDFC0C0E0C0C0DFCFCFE0C0C0E0CFCFE0CFCFE0D0D0EFCFCFE0D0D0E0D0D0E0
DFDFE0D0DFE0DFDFE0DFDFDFDFE0DFDFDFDFDFE0DFDFE0D0E0E0D0DFE0CFDFE0CFD0DFCFD0DFCF\
CFD0C0CFD0C0C0CFB0BFC0AFB0BF9FAFB090CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
3F50603F5F6F4F707F507F8F5F8F905F8F906F909F6F909F709FA0709FA080A0B0
70507F9080A0A09FBFAFA0C0AFAFC0B0B0CFB0B0CFBFB0D0B0B0D0BFBFDFBFBFDFBFBFDFBFBFDF\
C0C0E0BFC0E0C0C0E0C0C0E0CFCFEFC0CFE0CFCFEFCFCFE0CFD0EFCFD0E0D0DFE0
D0D0E0D0DFE0D0DFDFDFE0E0D0DFDFDFE0DFD0E0D0DFE0D0D0E0D0DFEFD0D0E0CFD0E0CFCFDFC0\
CFDFC0C0CFBFC0CFBFB0C0AFB0BF9FA0AF8FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF40506050809050809060909F806F90
9F8FAFA0A0C0B0AFCFAFAFCFB0B0D0B0B0D0BFBFDFB0BFDFBFBFDFBFBFDFBFC0E0BFC0E0C0C0EF\
BFC0E0C0CFEFC0C0EFC0CFEFC0CFEFCFD0EFCFCFEFCFD0EFCFD0EFD0DFEFCFDFE0
D0DFEFD0DFE0D0E0E0D0E0E0D0E0E0D0E0DFDFEFDFD0E0D0D0EFD0D0E0CFD0EFD0CFDFCFCFDFC0\
C0D0C0C0D0C0BFCFB0B0C0AFA0B09F909F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF8F7FA09F90BF
A09FBFAFAFCFAFAFCFB0B0D0AFB0D0B0B0DFB0B0DFBFBFE0B0BFDFBFBFE0BFBFE0BFC0EFBFC0E0\
BFC0EFBFC0EFC0CFEFBFCFEFC0CFEFC0CFEFCFD0F0C0D0EFCFD0EFCFD0EFCFDFEF
CFDFE0CFE0E0CFE0E0D0E0E0CFE0DFD0E0DFCFE0D0D0EFD0CFE0CFD0EFD0CFE0CFCFE0CFC0DFC0\
C0DFC0BFCFBFBFCFB0AFBF9FA0B09060705FCFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF60607F9F9FBFA09FC0
AFAFCFAFB0D0B0B0D0AFB0D0B0BFDFB0BFDFB0BFE0B0BFE0BFC0EFBFC0EFBFC0EFBFC0EFBFCFF0\
BFC0EFC0CFF0BFCFEFC0D0F0C0D0F0C0D0F0C0D0F0CFDFF0C0DFEFCFDFEFC0DFEF
CFE0EFCFE0E0CFEFEFCFE0E0D0EFE0CFEFDFCFEFDFCFEFD0CFEFD0CFEFCFCFEFCFC0E0C0C0DFC0\
BFD0BFBFD0BFB0C0AFAFBFA09FAF8FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF4F40609090B09F9FC0A0AFCF
A0AFCFAFB0D0AFB0D0B0B0DFAFB0DFB0BFE0B0BFE0B0BFEFB0BFE0BFC0EFB0C0EFBFC0F0BFC0F0\
BFCFF0BFCFEFBFCFF0BFCFF0C0D0F0BFD0EFC0D0F0C0D0EFC0DFF0C0DFEFC0E0EF
C0E0E0CFE0EFC0E0E0CFEFE0C0E0DFCFEFDFC0E0D0CFEFD0C0E0CFCFEFD0C0E0C0C0DFC0BFD0BF\
BFD0BFB0CFB0AFC0AF9FB090809070C0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F60909F9FC0A0AFCFA0AFCF
AFB0D0AFB0D0B0BFDFAFB0DFB0BFE0B0BFE0B0C0EFB0BFEFB0C0EFB0C0EFBFC0F0B0C0F0BFCFF0\
BFCFF0BFCFF0BFCFF0BFD0F0BFD0F0C0DFF0BFD0F0C0DFF0C0DFEFC0E0F0C0E0EF
C0E0EFC0E0E0CFEFEFC0E0E0C0EFE0C0E0DFC0EFDFC0E0D0C0EFD0BFE0CFC0E0CFBFDFC0BFD0C0\
B0CFBFB0CFBF9FB09F8FA080CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF7F7FA09FA0CF9FA0CFA0AFD0
A0AFD0AFB0DFAFB0DFAFBFE0AFB0E0B0BFE0AFBFE0B0C0EFB0BFEFB0C0F0B0C0EFBFCFF0B0C0F0\
BFCFF0B0CFF0BFD0F0B0CFF0BFD0F0BFD0F0BFDFF0BFD0EFBFDFEFBFDFEFC0E0EF
BFDFE0C0E0E0BFE0E0C0E0E0BFE0DFC0E0DFBFE0D0BFE0D0BFDFCFBFE0CFB0DFC0B0D0C0AFCFBF\
AFCFB09FB0A08FAF8FC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF5F5F90808FB09FA0CFA0AFD0A0AFD0
AFB0DFAFB0DFAFBFE0AFB0E0B0BFEFAFBFEFB0C0EFB0BFEFB0C0F0B0C0F0B0CFF0B0C0F0BFCFFF\
B0CFF0BFCFF0B0CFF0BFD0F0BFD0F0BFD0F0BFD0F0BFDFF0BFDFEFBFDFEFBFDFE0
C0E0EFBFE0E0C0EFE0BFE0DFC0EFDFBFE0D0BFE0D0BFDFCFBFE0CFB0DFC0B0D0C0AFCFBFB0CFBF\
A0C0B090B09F6F9070CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0606F9F808FB09FAFCF9FAFCFA0B0D0
A0B0D0AFB0DFA0B0DFAFBFE0AFBFE0AFBFEFAFBFE0B0C0EFAFBFEFB0C0F0AFC0EFB0C0F0B0C0F0\
B0CFF0B0CFEFB0CFF0B0CFEFBFD0F0B0D0EFBFD0F0B0D0EFBFDFEFB0DFE0BFDFE0
BFDFE0BFE0E0B0DFDFBFE0DFB0DFD0BFDFD0B0D0CFB0DFCFB0D0C0B0D0C0AFCFBFAFC0BFA0BFAF\
8FAF9060806FCFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF606F9F808FBF9FA0CFA0AFD0A0AFD0
A0B0DFA0B0DFAFBFE0AFBFE0AFBFE0AFBFE0B0C0EFAFBFEFB0C0EFAFC0EFB0CFF0B0C0EFB0CFF0\
B0CFEFB0CFF0B0CFEFB0D0F0B0CFEFBFD0F0B0D0EFBFDFEFB0D0E0BFDFE0B0DFE0
BFDFE0B0DFDFBFDFDFB0DFD0B0DFD0B0D0CFB0D0CFB0D0C0B0CFC0A0C0BFA0C0B09FBFAF80A090\
508060CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C04F5F707080AF9FA0CF9FAFCFA0B0D0
9FAFD0A0B0DFA0B0DFAFBFE0A0B0DFAFBFE0AFBFE0AFC0EFAFBFE0AFC0EFAFC0EFB0C0EFAFC0EF\
B0CFEFAFC0EFB0CFEFAFCFE0B0CFEFB0CFE0B0D0E0AFCFE0B0D0E0B0D0DFB0D0DF
AFD0D0B0D0D0AFD0D0B0D0D0AFCFCFAFCFCFAFCFC0AFCFC09FBFB09FBFAF8FB0A07F9F8FC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF708FB08F9FC09FAFCF9FAFCF
A0B0D0A0B0D0A0BFDFA0B0DFAFBFE0A0BFE0AFBFE0AFBFE0AFC0EFAFC0E0AFC0EFAFC0E0B0CFEF\
AFC0E0B0CFEFAFCFE0B0CFEFAFCFE0B0D0E0AFCFDFB0D0E0AFCFDFB0D0DFAFCFD0
B0D0D0AFCFCFAFCFCFAFCFC0AFCFCFA0C0C0A0C0C09FBFB08FB0A0709F90CFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF5F7FA08090BF90A0C09FAFCF
9FAFCFA0B0D09FB0D0A0B0DFA0B0D0A0BFDFA0BFDFAFBFDFA0BFDFAFC0E0A0BFDFAFC0E0AFC0DF\
AFC0E0A0C0DFAFC0DFAFC0DFAFCFDFAFC0D0AFCFD0AFC0D0AFCFD0A0C0CFAFC0CF
A0C0CFAFC0CF9FBFBF9FBFBF90B0B090B0B0709F90609080CFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF6F80AF8FA0C08FAFC0
90AFCF90AFCF9FB0D09FB0CFA0BFD09FB0D0A0BFDFA0BFD0A0BFD0A0BFD0AFC0DFA0BFD0AFC0DF\
A0BFD0AFC0D0A0C0D0AFC0D0A0C0CFAFC0D0A0C0CFA0C0CF9FBFC09FC0C09FBFBF
90BFBF8FB0B08FB0AF7FA0A0404F4FCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF6080A080A0C0
80A0BF90AFC090AFC09FB0CF90AFC09FB0CF90B0CF9FB0CF90B0C09FB0CF9FB0CF9FBFCF9FB0C0\
9FBFCF90B0C09FBFC090B0C090BFC090B0BF90B0BF8FAFB07FA0A06090904F706F
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF507F90
7090B07F9FB080A0BF80A0BF8FAFC08FAFBF8FAFC08FAFBF8FAFBF80A0B080A0B07F9FAF7FA0AF\
6F909F608F905080804F7F80CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCF\
CFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCFCFCFCFCFCFCFC0C0C0CFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF\
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
CFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0\
CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
C0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCFC0C0C0CFCFCF
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[10]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7
[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7
[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl7
20L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7
ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P7
00071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7GJl00i2_[`2b[`P7
00071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jn@LI1fT7N@Mi1gT7N@M`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7F:l04g5gLI1aT76@
LI1aT76@LI1aT740Yjl81`001`L0JJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g6@LI1aT7N@Mi1gT7N@
Mi1gT7N@M`03T02U[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P7
00061e:_01maMW5gLI1aT76@LI1aT76@LI1aT76@LI1aT7:@LI1b0:6_20L000L706>_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_MW6@LI1aT76@LI1aT76@LY1gT7L00i000gN@T008T02P1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061dn_0W408WIaMg6@LI1aT76@LI1aT76@LI1aT76@LI1b
T7:@LY1bT7:O[`P700071`1S[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LI1aT76@LI1aT76@LY1bT7N@Mi1g016@
09f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061dj_02QaMW5gLI1aT76@
LI1aT76@LI1aT76@LI1aT7:@LY1bT7:@LY1bT7:@W:l81`001`L0Hjl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`MaT76@
LI1aT76@Mi1bT7:@Mi1gT7N@M`0DT02J1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P7
00061d^_0W40;7IaMg6@LI1aT76@LI1aT76@LI1aT76@LI1bT7:@LY1bT7:@LY1bT7:@LY1bVJl81`00
1`L0Hjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7LI1aT76@LI1aT7N@Mi1bT7N@Mi1gT7N@M`0ET02I[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061dZ_031aMW6@LI1aT76@LI1aT76@LI1aT76@LI1aT7:@
LY1bT7:@LY1bT7:@LY1bT7:@LY2H[`P700071`1S[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LGIaT76@LI1aT7N@Mi1gT7:@
Mi1gT7N@Mi1g01N@09N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061dR_0340MW6@
LI1aT76@LI1aT76@LI1aT76@LI1aT76@LY1bT7:@LY1bT7:@LY1bT7:@LY1b00>@UZl81`001`L0AJl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`02L@0HT76@LI1aT76@LI1bT7:@Mi1gT7N@Mi1g0i001GN@Mi1g00>@00=gT9003i00UJl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061dJ_03P0LG5fLGMaMg6@LI1aT76@LI1aT76@
LI1aT76@LI1aT76@LI1aT76@LY1bT7:@LY1bT7:@LY1bT9B_20L000L704B_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`=a01J@LI1a
T76@LI1gT7N@Mi1gT7N@Mi1g0i002GN@T91gT7N@M`0ET02C[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00P700061dB_03`0LG5fLGMaMg6@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1bT7:@
LY1bT7:@LY1bT7:@LY2B[`P700071`13[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@0XT76@LI1aT76@LI1bT7N@Mi1gT7N@Mi1g
T7N@Mi1gT7N@Mi1gT7N@Ma>@096_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061d>_03ia07Ia
Mg5gLI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LY1bT7:@LY1bT7:@LY1bT7:@LY6_
20L000L7046_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`04L@0LT76@LI1aT7N@LY1gT7N@Mi1gT7N@Mi1gT7N@M`>@00UgT7N@Mi1gT7L0
5900T0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061d6_045a075aMW6@LI1aT76@LI1aT76@LI1a
T76@LI1aT76@LI1aT76@LI1aT76@LI1bT7:@LY1bT7:@LY1bT7:@LY1bT02@[`P700071`10[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`=a
02Z@LI1aT76@MY1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7LET02?[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl020L000H7@:l00`1a0002L@0nMW6@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1a
T76@LI1aT76@LI1bT7:@LY1bT7:@LY1bT7:@LY1bT7:?[`P700071`0n[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L5L@0dT76@LI1aT76@LI1g
T7N@Mi1gT7N@LY1gT7:@Mi1bT7:@Mi1gT7N@Mi1gT7:@Mi1gT7N@M`b@08h7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_20L000H7?Zl01@1a674H009a03b@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1a
T7:@LI1bT76@LI1aT7:@LY1bT7:@LY1bT8/3T8j_20L000L703b_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L7L@0dT76@LI1aT76@LI1gT7N@Mi1g
T7N@LY1bT7:@LY1bT7:@Mi1gT7N@LY1gT7:@LY1bT7:@M`f@08f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0
20L000H7?Jl01W4HLAQa609a042@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LY1a
T76@LI1aT76@LY1bT7:@LY1bT8^@Ri2;SJl81`001`L0>jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00Qa03B@LI1aT76@LI1aT7N@Mi1gT7N@Mi1b
T7N@LY1bT7:@LY1gT7N@LY1bT7:@LY1bT7:@LY1b3Y00S0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7
>jl0Bg40LAQa674HT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1a
T76@LI1bT7:@LY1bT7:@Ri2;T02<[`P700071`0j[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`Ma03QbLI1aT76@LI1aT76@Mi1gT7N@Mi1gT7N@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@M`b@08`7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061cZ_
04d0LAQa674HLAR@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT75gLI1aT76@
LI1aT76@LY1bT7:@LY1bT8^@R`2;[`P700071`0h[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L7L@0lLW6@LI1aT76@LI1aT7N@Mi1gT7N@Mi1gT7:@LY1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1g2i00Rjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7
>Jl0Cg4HLAQa674HLW6@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT75gLGMaMg5gLI1a
T76@LI1aT76@LY1bT7:@LY1bT7:@Ri00RZl81`001`L0>:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl71g40?W9aT76@LI1aT76@Mi1gT7N@Mi1gT7N@Mi1b
T7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LY1b2Y00RPN_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P7
00061cV_051a674HLAQb679aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT75gLGMa
Mg5gLGMaLW5gLGMaT76@LI1aT76@LY1bT8^@RhV_20L000L703R_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`=a00YbLG9aLW6@LI1f2I00<gN@Mi1gT7N@
Mi1gT7N@Mi1gT7N@Mi1bT7:@LY1bT79gLWMbMg9gLWMbT7:@LY1bT7:@LP07T029[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl020L000H7>Jl0D74HLAQb678HLW6@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1a
T75gLGMaMg5gLG9aLW5bLG9aLW5gLGMaT76@LI1aT7:@LY2;RJl81`001`L0=jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W402W9aLW5bLG9aT74;T00c
Mi1gT7N@Mi1gT7N@Mi1gT7N@Mi1gT7:@LY1bT7:@LWMbMg9gLWMbMg9gLY1bT7:@LY1b00R@08P7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_20L000H7>:l0DQQa674HLQQb679aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@
LI1aT76@LI1aT75gLGMaMg5bLG9aLW5bLG9aLW5bLGMaT76@LI1bT7:@Ri28[`P700071`0g[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@06LW5bLG9a
0W800i1aT00:T009Mi1gT7N@Mi1g00>@02MgT7N@Mi1gT7N@LY1bT7:@LY1bMg9gLWMbMg5gLWMaMg:@
LY1bT7802900R0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLg[`1CLAQa678HLQQb679aT76@LI1aT76@LI1a
T76@LI1aT76@LI1aT76@LI1aT76@LI1aT75gLGMaMg5bLG9aLW5bLG9aLW5bLG9aMg6@LI1bT7:@LY00
R:l81`001`L0=Zl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1`=a00AbLG9a17800i1fT00>T005Mi1gT7L01I009gN@Mi1gT7N@LY1bT7:@LY1bMg9gLWMaLW5b
LG9aMg5gLGMbT7:@LP06T0281jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061cN_05Aa674HLQQb679>LW6@LI1a
T76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aMg5gLG9aLW5bLG9aLTibCW9aLW5bLGMa
T7:@LY1bT7:7[`P700071`0e[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl01740179aLW44LP03T7J@012@00=gT7L01i009gN@Mi1gT7:@LY1bT7:@LWMbMg5g
LGMaLW5bLG9aLW5gLGMbT7:@R`05T027[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061cJ_00THLAQbK79aLW40
0g80BG6@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aMg5gLGMaMg5bLG9aLTib
CW9>LTibLG9aT76@LI1bT740Qjl81`001`L0=Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_009a00AbLG9a1g801Hn@Si2?016@02UgT7N@Mi1gT7N@Mi1gT7N@
LY1bT7:@LGMaMg5gLGMaLW5bLG9aLW5bL@02LP2>T7:@LY2;T:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061cJ_
00THLAQb679aLW401780B91aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT75gLGMa
Mg5gLGMaLW5bCW5>LDiaCW5>LW5bLI1aT76@LHN_20L000L703V_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LG9aLW402G801Hn@Si2?016@029gT7N@Mi1gT7N@
Mi1gT7N@LY1bT7:@LY1aMg5gLGMaMg5b1G40TW9aLW6@LY1bT7:@LPN_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7
=Jl0274HLQQbK79a1W80BI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1a
T75gLGMaMg5bLG9>LDiaCW4HLDibLG9aT76@LI00QZl81`001`L0>Jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`MaLW5bL@09LP05Si2?T8l05I0087N@Mi1gT7N@
Mi1gT7N@Mi1bT7:@LY1aT75gLGMaLW5b0W40TdiaLG9aLW6@LY1bT787[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`00
1PLe[`08LAQb678HLW46LP19T76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1a
T76@LI1aMg5gLGMaLW5bCW5>LAQa679aLW6@LI1aT026[`P700071`0c[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_009a00=bLG802g80192?T8lGT00PMi1gT7N@
Mi1gT7N@Mi1gT7:@LY1bT76@LGMaMg5bLG85L@2>LW6@LI1bT787[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061cB_
00T0LAQb679aLW401g80>W6@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@
LI1aT76@LGMaMg5bLG82L@0<CW4HLAQbLG9aRg6@QZl81`001`L0<jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@03LW5b00Ub00IgLY2?T8lQT02YMi1gT7N@
LY1bT7:@LY1aT79gLGMaLW5bLG9>LW5bLI1aT787[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLd[`07074HLQQb
L@09LP0nLI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1bT76@LI1aT76@LI1aT76@LI1aT75g
LGMaLW5bLG82L@08675>LW5bLH^6[`P700071`0c[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_009a27801WMbMg:ALRV@0:EgT7N@Mi1gT7N@Mi1bT7:@LY1aT75g
LG9aLW5bLG9a1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7=:l02AQa678HLTibL@06LP14Mg6@LI1aT76@LI1a
T76@LI1aT76@LI1aT76@LI1bT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LGMaMg5gLG9aLQP2L@03
LW5a08J_20L000L703>_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl00W48LP06Mg9hLWQb;I00XGN@Mi1gT7N@LY1bT7:@LY1aT75gLGMaLW5bLHYa1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00P700061cB_00LHLAQb679>00Ib04EgLWMaT76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1b
T76@LI1aT76@LI1aT76@LI1aT76@LI1aT76@LI1aMg5gLG800W401aQaLG5`RV/0Pjl81`001`L0<jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@Qb00MgLWQb
TG:A03:@015gT7N@LY1bT7:@LY1bT79gLP02L@03LW6:00::08=aRW47[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLd[`09
674HLTibCW9?00Ab04]gLWMbT76@LI1aT76@LI1aT76@LI1aT76@LI1bT76@LY1aT76@LI1aT76@LI1b
T7:@LY1aT76@LI1aT76@LI1aT75gLGMaMaQa67400XX01g2:L8Y[LF/0OJl81`001`L0<jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@Ib00YgLWMbN79hLY5b
<i003gN@Mi1gT7:@LY1bT7:@LP04L@05ShZ?RXl00XX0Og6:LHZ_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7=:l02AQa679>LTib
C`04LP17Mg9gLWMaT76@LI1aT76@LI1aT76@LI1aT76@LI1bT76@LY1aT76@LI1aT76@LY1bT7:@LY1a
T76@LI1aT76@LI1aT76@LGL00W401QQaRXn:S`::00M`RW2:JhY[07V_20L000L703>_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W46LP0;Mg9hLWQbTG:AMi40
=9003WN@Mi1gT7N@Mi1gT7N@0W4038nYSjV:ZHZ?RXn:S`::07UaRW47[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061cB_00THLAQbCW9?LTl0
0W80EGMbMg9gLWMbT76@LI1aT76@LI1aT76@LI1aT76@Ri1aT7:@LY1aT76@LI1aT76@LY1bT7:@LY1b
T7:@LI1aT76@LI1aT76@LI1aT1R?RXn:T8Z?RXl018X01W6:LHY[RWB_20L000L703>_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W46LP0;Mg9hLWQbN7:AMi40
>Y0067N@Mi1gT7N@Mg6?ZHnYSjV?ZHZYRY2:S`J:07=aRW47[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061cB_00DHLAQbCP06LP0HMg9gLWMbN7:ALI1a
T76@LI2:T8Z@Si2?0i002G6@LI1aT76@L@09T00`Mi1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LI1aT76@
L92:ZHnYRZV:ZHZ@RY2:T8Z?18X0276:JhY[RV]aJZl81`001`L0<jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@03LW5b00Eb00igLWQbN7:ALY5gTHn@ScN@
00MgT7N@Mi1g00:?01>YSjV@ZI2YT:V?ZHnYRZV:T8Z@00R:00=aRW400W40IPN_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061cF_00Aa679>1W8097MbMg9gLWQbTG:@LI1a
T76@RY2:T8Z@Si2?T8Z@LI1aT76@L@^@01ebT7:@LY1bT7:@LY1bT7:@LY1bT7:@LI1aT76@L@02S`0C
ZHnYSjV?ZHnYRY2:T8Z@RY2:S`04RP0:LHYaRV^:JhY[LFB_20L000L703N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LG9a00Ab019gLWMbN79hLY5bTGNASi2?
T8liT007Mi1gT8nYS`05Z@0@T:V@ZI2YRZV:ZHZYRY2:T0Z:06EaRW5a[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLe[`04LAQbCPIb02AgLWMbMg9hLY5bTG6@LI1aT8Z@RY2?
T8n@Si2:T8Z@RY2:T8X;T00dLY1bT7:@LY1bT7:@LY1bT7:@LY1bT7:@LI1bLHZ@SjV@ZI2YT:V?ZHnY
RZV:T8Z@RY2:T0J:00eaRW6:L8Y[RV^:JhY[05f_20L000L703N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LG9a00Ab011gLWQbN79hLY5gTGNAMi6??I001Hn@
SjV@00NY012@ZI2YT:V:ZHZYRZV:T8Z;28X0IG6:LHYaRW6:LHZ_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061cF_00Aa678H1W8097MbMg9gLWQbTG:ALY1aT76@RY2:T8Z@Si2:T8Z@RY2:
T8Z@RP^@00]bT7:@LY1bT7:@LP03T00ALY1bT7:@LY1bT7:@LI2?T8l01JT04Y2YT:V@ZHZYRZV:T8Z@
RY2:R`J:00maRW6:JhY[RV^:JhY[RV/0FJl81`001`L0=jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LW4017804WMbMg9hLWQbTG:AMi5gTGN@Sc^@00>?ZHl0
0ZT00jjY[P06Z@0@T:V@ZI2YRjV:ZHZYRY2:R`Z:065aRW6:LHYaRW6:L@N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061cJ_00<HLQP01W8097MbMg9gLWQbTG:ALY5bT76@RY2:T8Z@RY2?T8Z@RY2:
T8Z@RPf@00UbT92@LY1gT7801I003gN@Mi1bT7:@LY1aT8nYS`09Z@0@T:V@ZH^YRZV:T8Z@RX^:R`B:
01EaRW6:JhY[RV^:JhY[RV^:JhY[LF/0Djl81`001`L0=jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LW4017804WMbN79hLWQgTGNAMi5gTGNAScZ@00^YT:j@
[ZV^ZJjY[P06Z@0@T:V@ZI2YRjV;ZHZ@RX^:R`B:06EaRW6:LHYaRW6:LHYaRW6:LHYaRW47[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`081`001PLg[`0fLAQbCW9?Mg9gLWMbMg9gLWQbN7:AMi5gT76@LI1aT76@LI2:
T8Z@RY2:T8n@RY2:T8Z@RY2;3I004G:@LY1bT7:@LY1bT7:@Sjj@00BY00>^ZJT01:T0;i2YT:V;ZH^@
RY2:T8Z;RX]aRW6:K8Y/RVb:JhY[RV^:JhY[RV^:JhY[RV^:JhY[04f_20L000L703N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Ma00Ab01=gLWQbN79hLWQg
TGNAMi5gTGRA03V@00n^ZJnY[ZV_ZJnY[jV_ZJl00ZT0492YT:V@ZI2YRi2:T8Z;RX/2RP0GLHYaRW6:
LHYaRW6:LHYaRW6:LHYaRW400hX2L@1;[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061cR_03DHLTibCgMbMg9g
LWMbN79hLWQbN7NAMi1aT76@RY1aT8Z@RY2:T8Z@Si2:T8Z@RY2:T8Z@RP0=T00=Mi1gT7:@LY1bT7N@
S`02T009[ZV^ZJjY[ZV^00:Y00J@ZI2YT:T2T00/Ri2;T8Z@RX]aRg6:K8Y/RV]aJhY[RV^:JhY[RV^:
JhY[RV^:JhY[RV^:Jg5:[`P700071`0i[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[g9a009b01AgLWQbN79hLWQgTGNAMi5gTGRAN3F@00>?[Y000Zh0;:nY
[jW6ZLJY[jV_ZJnYZY2YT:V@ZH^@Ri2:RhZ;LHYaRW6:LHYaRW6:LHYa4hX0BG47[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl0
20L000H7>Jl0179>LTl2LP0^Mg9gLWMbMg9hLWQbN7NAMi5gT76@LI1aT76@RY2:T8Z@RY2:T8Z@RY2:
T8Z@RPn@00egT7N@LY1bT7N@SjjY00>^00RY[ZV_ZJnYZP:Y00B@ZI2Y0Y00;h^@RY2:Rg6;K8Y/RV]a
Jg5[LF]aJg5[RV^:JhY[RV^:JhY[RV^:RHY`RV^:JhY[04N_20L000L703V_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g400W805WMbN79hLWQbN7NAMi5g
TGNAN95hTGPdT00K[ZV_[/J^aZk6ZLJYaZV_ZJnYZY2YT:V@ZH^@00:;00R:Rg6:LHYaRP=a00B:LHYa
6hX0AJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`081`001PLi[`0dLQQbCW9?Mg9gLWMbMg9hLWQbN7NAMi5gTGN@LI1aT8Z@RY2:
T8Z@RY2:T8Z@RY2:T8Z@RQF@00=bShl00Y00AZjY[jj_[ZnY[jV_ZJnYZY2YT:V;T8^@RX]aRg6:K75[
LF]aJg5[LF^:JhY[RV^:L8Y`RXV:RHZ9RXV:RHZ9RXV:RHY`RV]5[`P700071`0i[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00Ab01IgLWQbN79hMi5g
TGNAMi5hTGRAN95h<9006jj@[jj_[/J_aZo6ZLJYaZV_ZJn@ZY2YT:V;T002R`04RX]aRPMa00B:LHYa
38X00hn:RP0?RP141jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PLj[`05679>LTl00W80;7MbMg9gLWQbN79hMi5gTGNAMi5g
T76@LI2:T8Z@RY2:T8Z@RY2:T8Z@RY2:4i000h^@R`02T00C[ZV^[Zn^[jj_ZJnY[jV_ZJZ@Z@02T00I
Ri2;T8Z;LHY/L@1a075[LF^:JhY[RW2:L@0CRP09RHZ9RXV:RHY[04>_20L000L703Z_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl70g8067MbMg9hLWQg
TGNAMi5gTGRAMi5hTGRFN2j@0Zh05Jn^[jo6[lZ_aZW6ZLJY[jV_T:Z@Z@02T004Ri2:R`Ma00B:LHYa
38X05JR:Z8ZXRZR:Z8ZXRZR:Z8ZXRZR:S`03RP121jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7>jl0<W8HLTibCgMbMg9gLWQb
N7MhMi5gTGNAMi5gTGN@LI2:T8Z@RY2:T8Z@RY2:T8Z@RY2:5I000hn^T002[P03[jj_00B_00ZY[jV_
ZJnYZY2Z0Y004h^@RX]aRVaaJg5[LF^:JhY[RW403XX03Hn:ShZ?RXn:ShZ?RXl018X01HV:L8Y[046_
20L000L703^_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`06LP0JMg9hMi5gTGNAMi5gTGNAMi5hTGRAN96?T8lHT005ZI2YT:T01i003iF@[Y2_[Zn^
bZo:[`N_aP02[`0>ZZnY[jVZT:Z;T8^@LH/3L@03RW6:00V:026YRZV:ZHZYRZV:ZHZYRZV:ZHZYSjV:
ZHZXRZR:Z8ZXRXl01HX0?PN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_20L000H7?Jl00g9>LP03LP0VMg9hLWQbN7:ALY5gTGNAMi5gTGNAMi6:
T8Z@RY2:T8Z@RY2?T8l5T003Si2?00f@00UfT92ET:n^[jh01:l06<J_[jZ_ZZnYZZVZT:V;T76;075[
RV^:K0Z:02>?RY2:T8ZYRZV:ZHZYRZV:ZHZYRZR:ZHZYRZR:Z8Z?RXn:S`02RP03RHY`03f_20L000L7
03f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00Ab01agLWQbN7NAMi5gTGNAMi5gTGNAN95hTGRFSi2?69000jV@T005T00AUI2^T:n^[jk6[lZ_
1jl7[lH00Zl01ZZ_ZZn@ZP:@00N;LG6:LHYa00Z:02V@RZV:ZHZYRZV:ZHZYSjV?ZHnYSjV?ZHnYSjV?
ZHnYRZR:Z8ZXRZR:S`02RP0m[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PLn[`03679?009b03AgLWMbN79hLY5gTGNAMi5gTGNAMi5g
TGNAMi6:T8Z@RY2:T8Z@RY2?T8n@RY2:T8Z@Si2?1I002GJ@T9F@[Y2_[P04[`0=]:o6[lNZ[jZ_ZJZ@
ZP02T007Jg5[RVb:K00:RP0YRhZ@RZV:ZHZYRZV:ZHZYRZV:ZHZYRZV:ZHZYRZV:ZHZXRZR:ShZ?RXl0
0XX00g2:[`0k[`P700071`0o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`Ma00Ab01ahLWQgN7NAMi5gTGNAMi5gTGNAN95gUWNFMiIg790029F@
[Y2_UJn^0Zl01lZ_bZl7[`L00Zl01ZZ_ZZn@[`=a00>:LHX01HX068^:T8ZYRZV:ZHZYRZV:ZHnYSjV?
ZHnYS`>Y01RXZJRYZ:VXZHnYSjV?ZHnYSjV?Z8nXRZP3RP0l1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7?jl0?75>LTmbCgMbMg9hLWQb
TGNAMi5gTGNAMi5gTGNAMi5gTGNARY2:T8Z@RY2:T8Z@RY2:T8Z@RY2?T8Z@M`:@00RET:j@[jj_[PB_
01:d[lN_/:Z_ZZm[LFaaK8Y/RW44RP0hRhZ@RY2:T8ZYRZV:ZHZYRZV:ZHZYRZV:ZJRYZ:V?ZJRYSjV?
ZHnYSjV?ZHZYRZR:Z8Z?RXn9RW0k[`P700071`0o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_009a0W8087MbN79hMgQgTGNAMi5gTGNAMi5g
TGNFMiIgUWNFMiIg4Y0039F@UI2ET:j@[iF_[PB_00O:[`N_1jl700:_0g400hYaRP05RP0FRhZ@RZV:
ZHZYRjV@ZI2YT:V@ZI2YT1>Y012XZJRYZ:VXZHnYSjR?Z8nX0Xl2RP0kL@N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7@Jl0>75>LTmgCgMb
Mg9gLWQbTGNAMi5gTGNAMi5gTGNAMi5gTGNAMi5gTHZ@RY2:T8Z@RY1gT7N@Mi1g0Y0029F@UI2_UJn^
1:l03[B_]Jo7[g5[LFb:K8Ya18X07X^:T8Z@RZV:ZHZYRZV:ZHZYRZV:ZHZYRZV?ZHnYZ0>Y01jXZJRY
Z:VXZJRYZ:VXZJRYSjV?Z8ZXRXn:ShZ?RHXj[`P700071`11[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W80?7MbMg9hLWQgN7NAMi5g
TGNAMi5gTGNFMiIgUWNFMiIgUY2FT9J@UY2FT9IgUWNET9F@UI2ET9F@[iF_[PJ_00C:[`N_1746RP0B
RhZ;RZV:ZHZYRZV;ZH^YT:V@8JT0BjRYZ:VXZHnYSjR?Z8nXRXn:1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PM3[`0bLTmgCgMbMg9gLWQb
N7NAMi5gTGNAMi5gTGNAMi5gTGNAMi1gT7N@Mi1gT7N@Mi1gT7L4T008UI2ET:nE[jh6[`08]6]aJhY/
RW44RP0DRhZ;RY2:ZHZYRZV:ZHZYRZV@ZI0IZ@0IZ:VXZJRYZ:VXZJRYSjV?Z8nXRXn:ShZ?L00i[`P7
00071`13[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`02LP04Mg9gLP9g039hMgQgTGNAMi5gTGNFMiIgUY2FT9J@UY2FT9J@UY2FT9J@
UY2FT9F@UY2FT:n@[iF_U@B_00JdLHYaRW44RP0BRhZ@RZV;ZH^YRjV;ZI2YT:V@3:T0;JjY[ZV^ZJjY
[ZV^ZJjYaJW5ZLFYaJW5ZJjY[ZV^ZJjY[ZV^Z:VXZHnYSjV?Z002S`0i[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061dF_02abCW9?Mg9gLWMb
Mg9gLY5gTGNAMi5gTGNAMi1gT7N@Mi1gT7N@Mi1gT7N@M`J@01:ET9J@[i2_UJnELF^:K8YaRW42RP0H
RhZ;RY2:T8ZYRZV:ZH^YRjV;ZI2YT:V@8JT05ZRYZ:VXZJRYZ:V?ZHnYSjR:Z8Z?L8Xh[`P700071`15
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_009b03AgLWMgN7MhMi5gTGNAMi5gTGNAMiJ@UY2FT9J@UY2FT9J@UY2FT9J@UY2FT9J@
[i2_UJnE0jl00g6:L@02RP0DRhZ;RY2:T8^YRjV;ZH^YT:V@ZI0<Z@0c[jV_ZLJYaZW6ZJnYaZV^ZLFY
aJW6ZLJYaJW5ZLFY[ZV^ZJjY[ZV^ZJjY[ZR^Z:V?ZHnX00:?03V:1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7AZl00aQbCP02LP0:Mg9gLWMb
Mg9gLP9g01RAMi1gTGN@Mi1gT7N@Mi1gT7N@Mi1gT7L4T00@UI2ET9F@[i2_Jg5[RVb:L@::01R;RX^:
RhZ@RY2:ZHZYRjV;ZH^YT:V@ZI0LZ@07[ZV^ZJjY[P08Z@0@Z:VXZJRYZ:V?ZHnYSjR:Z0::>:l81`00
1`L0Ajl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_00Ab02egLWQgN7NAMi5gTGNAMi5gTGNFMiJ@UY2FT9J@UY2FT9J@UY2FT9J@
[i2_UJl00g406XYaRg6;RX^:T8^@RjV;ZH^YT:V@ZI2YT:V@1ZT0>jZYZZV_ZJnY[jW6ZLJYaZW6ZLJY
aZW6ZLJYaZW6ZLJYaZW6ZLJYaJW5ZJjY[ZV^ZJjY[ZV^Z:j?ZHnY00:?0XX0=jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7BJl0379>LTmgLWMb
Mg9gLP9g042@Mi1gT7N@Mi1gT7N@Mi5gT7N@Mi1gUI2ET9F@UI2FT9I[LFb:LHYaRg6;RX^:T8Z@RY2:
T8^@RjV;ZH^YT:V@5:T00jjYZ@07Z@0?[ZV^ZJjY[ZV^ZJjY[ZV^00RY00nXZJRYSjV?ZHnXSjR:Sg00
=jl81`001`L0BZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl70g809gMbN7MhMi5gTGNAMi5gTGNAMiJ@UY2FT9J@UY2FT9J@
UY2FT9J@[`03L@2JRW6;RX^:RhZ@Ri2;ZH^YT:V@ZI2YT:Z@ZY2ZZJZY[jVZZJnY[jV_ZJnY[jW6ZLJY
aZW6ZLJYaZW6ZLJYaZW6ZLJYaZW6ZLJYaZW5ZLFYaJW5ZLFY[ZV^ZJjX[Xn^SjV?Z8Z?[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061df_00YbCgMb
Mg9gLWMb0WL07Y1gT7N@Mi1gT7N@Mi5gTGNAMiIgUY2FT9J@LFaaK09a01b;LH]aRhZ;RY2:T8^@Ri2;
T8^YRjV@ZI2YT:V@3ZT0>JZY[jV_ZJnY[jV^ZJjY[ZV^ZJjY[ZV^ZJjY[ZV^ZJjY[ZV^ZJjY[ZV^ZJjX
ZJRYSjV?ZHnXRXm`RP0f[`P700071`1A[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jmgLWMb009g01ahMi5gTGNAMi5g
TGNFT9J@UY2FT9J@UY2FT9J@1740IHYaRhZ;RX^:T8^@RjV;ZI2YT:V@ZI2YT:ZYZZVZZJnY[jV_ZJnY
[jV_ZJnY[jW6ZLJYaZW6ZLJYaZW6ZLJYaZW6ZLJYaZW6ZLJYaZW6ZLJYaJW5ZLFYaJV^ZLFY[ZV^Z:j?
ZHnX00:?03N:1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00P700061dn_00YgCWM?Mg9gLWMb17L0>i1gT7N@Mi1gTGNAMi5gTGNF075/LFb:K8YaRg6;RX^:
T8Z@Ri2;T8^YRjV@ZI2YT:V@ZI2YT:ZYZZVZ00NY03VZZJZY[jV_ZJnY[jV_ZJnY[jV_ZJjY[ZV^ZLJY
aJW5ZJjY[ZV^ZJjY[ZV^ZJjY[ZR^Z:V?ZHnYSjT00Xl2RSJ_20L000L705>_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_Mg9gLP02M`0GN7NAMi5gTGNAMi5gTI2FT9J@UY2FT9H00g4028YaRhZ;RX^:0X/0Gi2;ZH^YRjV@
ZY2ZT:Z@ZZVZZJZY[jV_ZJnY[jV_ZJnY[jV_ZLJYaZW6ZLJYaZW6ZLJYaZk6[/J^aZk6[/J^aZk6[/J^
aZk6ZLJ^aZk5ZLFYaJW5ZLFY[ZV^Z:j?[XnY00:?03N:1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061e>_00QbCWMbMg9gLP9g06>@Mi1gT7NAMi5g
TGMaK75/LFb;LH]aRg6;RY2:T8^@Ri2;ZH^YT:V@ZI2YT:Z@ZY2ZZJZYZZVZZJZYZZV_ZJnY[jV_ZJnY
[jV_ZJnYaZW6ZLJYaZW6ZLJYaZW6[/JYaZk6[/D00Zh05ZW5ZJjY[ZV^ZJjY[ZV^ZJjXZJRYSjT2S`03
RXn_03F_20L000L705N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`MbMg800WL03Y5gTGNAMi5gTGNFT9J@
1740KHYaRhZ;RY2;T8^@RjV@ZI2YT:Z@ZY2ZT:Z@ZY2ZZJnY[jV_ZJnY[jV_ZJnYaZW6ZLJYaZW6ZLJ_
aZo6[lJ_aZo6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZk6ZLJYaZW5ZLFY[ZV^ZJjY[ZR^SjT00Xl0=g47
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7
EZl04dibLGMbMg9gLY1gT7N@Mi1[LF`00W4048YaRg6;LI2:T8^@Ri2;T8/2T01KZI2YT:V@ZY2ZT:Z@
ZY2ZZJZYZZVZZJnY[jV_ZJnY[jV_ZJnYaZV_ZLJYaZk6[/J^aZW6ZLJYaZW6ZLJ^aZk6[/J^aZk5ZLFY
aJV^ZJjY[ZV^ZJjY[ZR^Z:V?Z@02S`03RXn_03F_20L000L705N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl00W802gMbMgMhMi5gTGNA00Ea06>;LH^;T8^@Ri2;ZH^YRjZ@ZY2ZT:Z@ZY2ZT:ZY[jV_ZJnY
[jV_ZZnZ[jZ_Z/JZaZ[6Z/JZaZ[6Z/J_aZo6[lJ_aZo6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/J^
aZk6ZLH00Zh02ZV^ZJjY[ZV^Sjh2S`0gL@N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PMI[`0<LTigLWMbMg9gLY1/174048]aRg6;RY2;T8^@Ri2;
T8/2T01QZI2YT:Z@ZY2ZT:Z@ZZVZZJZY[jV_ZJnY[jZ_ZZnY[jW6ZLJYaZW6ZLJ_aZo6[lJ_aZk6[/J^
aZk6[/J^aZk6[/J^aZk6[/J^aZk5ZLFY[ZV^ZJjY[ZV^ZJjXZHnYSjR:S`0f[`P700071`1K[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02LP04Mg9gM`Aa00B;LH^:0X/0?92;T8^@RjV;ZY2ZT:Z@
ZY2ZT:Z@ZZVZZJnY[jV_ZJnY[jZ_ZZnZ[j[6Z/JZaZ[6Z/JZaZo6[lJ_aZo6[`G6022_aZo6[lJ^aZk6
[/J^aZk6[/J^aZk6[/J^aZk6[/J^aPB^00RY[ZV^ZJj?[P:?03Ma1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061ej_00DHLW9aK002L@06Rg6;LH]a
0X/0Fi2;T8^@Ri2;T8^YRjV@ZY2ZT:Z@ZY2ZZJZY[jV_ZJnY[jZ_ZZnZ[jZ_ZZnZaZ[6Z/JZaZ[6Z/J_
aZo6[lJ_aZo6[/J^aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/D00Zh04ZV^ZJjY[ZV^ZJjXZHnYSi2:ScJ_
20L000L705j_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl71G400h]aR`03R`0hT8^@Ri2;
ZY2ZT:Z@ZY2ZT:Z@ZZVZZJnZ[jV_ZJnZ[jZ_ZZnZ[jZ_Z/JZaZo6[lJ_aZo6[lJ_aZlCaP03[lJ_00?6
01:_aZk6[/J^aZk6[/J^aZk6[/H6[P06ZJjY[Xn^0Xl0=hX7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7GJl03740LFaaK8]aRg6;LP:;05f@Ri2;
T8^@Ri2;ZX^ZT:Z@ZY2ZT:ZYZZVZZJnY[jV_ZJnY[jZ_ZZnZ[jZ_ZZnZaZ[6Z/JZaZo6[lJ_aZo6[lJ_
aZo6[lJ^aZk6[/J^aZk6[/J^aZk6[/J^aZk6[/H01:h03ZV^[ZjY[ZV^Sjj?ZHn@0Xlf[`P700071`1K
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`04L@04Rg6;L@B;00B@Ri2;0Y003:Z@ZY2ZT:Z@
ZY2ZT0:Z02J_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[jZ_Z/J_aZo6[lJ_aZo6[lJ_aZo6[aG601J_a/K6[lJ_aZk6
[/J^aZk6[/J^aZk61Zh01ZV^ZJj?[P:?03Ma1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061eZ_01D0LFaaK79aRg6;LH]bT8^@Ri2;T8/00Y004Y6@
TI2ZT:Z@ZY2ZT:Z@ZY2_Z@:Z046_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[j[6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_
aZo6[lJ_aZo6[lJ^aZk6[/J^aZk6[Zn^[`08[P0:ZJjY[ZV^Sjj?Z@:?00=a[jl0=Jl81`001`L0F:l7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`L3L@06LW6;LX]b0X/0>Y2;T8^ARjZ@ZY2ZT:Z@ZY2ZT:ZA
[i6_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[j[6[lJ_ajo7[lN_ajo6[lJ_aZl6aP0;b/K:a/[6b/K:a/X00lH07l[6
a/K:a/Z_bZo:[lZ^bZk6[/J^aZk6[/J^aZk6[/H00Zh01ZV^T:j?[P:?03N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061eV_06eaK75/LW6;LH]a
Rg:@LY2;T8^ARi6@TI2AT:Z@ZY2ZT:Z@ZY2ZTJnA[jZ_ZZnZ[jZ_ZZnZ[jZ_ZZnZ[j[6[lJ_aZo6[lJ_
aZo6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_aZk6[/J^aZk6[/J^[jj_00Z^00RY[ZV^Sjj?Z@:?00=`[jl0
=Jl81`001`L0Ejl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@06LW6;LX]b0X/0992;TH^AT:Z@
ZY2ZT:Z@ZY6ZTJZA[i6_ZZnZ[jZ_ZZnZ[jZ_ZPB_01G6[lJ_ajo7[lN_ajo7[lN_ajo7a/L00lH0;`O6
1lH7aPO61lH7a/[6b/K:a/[6b/K:[lZ_bZo:[lZ^bZk6[/J^aZk6[/J^aZk600:^00JY[Y2^Sjh2S`0g
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`00
1PMG[`0jLAQaLG9aRg6;LX]bT8^@Ri2;TH^ARi6@TI2ZT:Z@ZY2ZTJZA[i6_TJnZ[jZ_ZZnZ[jZ_ZZnZ
[jZ_ZP:_01S6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_aZl3aP0J[lJ_aZo6[lJ_aZk6[/J^aZk6[/J^[jj_
[Zl8[P08ZJj@[Xn^Si02ScR_20L000L705J_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`=a00B;LX]b
0X/0992;T8^AT96@ZY2ZT:Z@ZY6ZTJZA[i6_ZZnZ[jZ_ZZnZ[jZ_ZPB_04W6[lN_ajo7[lN_ajo7[lN_
ajo7aPO61lH7aPO61lH7aPO61lH7a/[6b/K:a/[6b/K:[lZ_bZo:[lZ_bZo:[/J^aZk6[/J^aZj_00:^
03n@[Y2^Si2?1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl020L000H7EZl0=@1aK79aRg:;LX]bT7:@Ri2;TH^ARi6@TI2AT:Z@ZY6ZTJZA[i6_TJnZ[jZ_
ZZnZ[jZ_ZZnZ00J_01_6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_bZo:a/X01<H06:o6[lJ_aZo6[lJ^aZk6
[/J^aZj_[Zn^[`R^00RY[Y2^Sjj?T0:?>:l81`001`L0E:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1`=a
0W80Mh]bT8^@Ri2;TI2AT:Z@ZY2ZT:Z@ZY6ZTJnA[jZ_ZZnZ[jZ`Z[2Z/:Z`[k2_/:o7[lN_ajo7[lN_
ajo7[`N_1jl7aPO61lH7aPO61lH7aPO61lH7a/[6b/K:a/[6b/K:[lZ_bZo:[lZ_bZo:[/J^aZk6[Zn^
[jj_00:^00B@[Y2^0Xl0>Jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`081`001PMD[`05074HLV`00W80;8]bRg:@LY2;T8^ARi6;TI2AT96@ZY6ZTJZA
ZY6ZTJnA[jZ_ZZnZ[jZ_ZZnZ2:l0>K2_ajo7[lN_ajo6[lJ_aZo6[lJ_bZo:[lZ_b/K:a/[6b/K:[lZ_
bZo:[lZ_aZk6[/J^aZj_[Zn^[`08[P09T:j?[Xn^Si1a03V_20L000L705>_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`02L@04LW6;LP:;07>@Ri2;TI2AT96@ZY2ZT:ZA[i6_TJnA[i6_ZZnZ[jZ`Z[2Z/:Z`Z[2_
/:o7[lN_ajo7[lN_ajl7[`N_1jl7[`O61lH7aPO61lH7aPO61lH7aPO61lK:a/[6b/K:a/Z_bZo:[lZ_
bZo:[lZ^aZk6[Zn^[jj_00B^03n@[Y2^S`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7Djl0<W4HLVabLH]bRg:@LY1bT8^@Ri6;TI2AT96@
ZY6ZTJZAZY6ZTJnA[i6_ZZnZ[jZ_ZZnZ2:l0>k2_/:o7[lN_ajo7[lN_ajo6[lJ_aZo:[lZ_bZo:a/[6
b/K:[lZ_bZo:[lZ_bZo6[lJ^aZk6[Zn^[jj_00R^00R@[Xn^SiF?T3Z_20L000L7056_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl00W40Ng9aLW:;LY6;TH^ARi6@TI2AT:ZAZY6ZTJnA[i6_TJnA[i6`TK2Z/:Z`[k2_
/:n`[lN_ajo7[lN_ajo7[lN_1k07[`N_1jl7[`O61lH7aPO61lH7aPO:1lX7aPO61lH7a/[6bZo:[lZ_
bZo:[lZ_bZk:[/Z^[jj_[Zn^[`02[P11UJj@[XnEMPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7DJl01W4HLW5bL@9b00j@LY1bT7:ARi6;
TI2AT0JA01BZTJZA[i6_TJnA[i6_ZZnZ[jZ_ZP:_03^`[k2_/:n`[k2_/:n`[lN_ajo7[lN_1jl7[`N_
1jl7[`N_bZo:[l[6bZo:[lZ_bZo:[lZ_bZo:[lJ_aP04[`0:[Zn^[jj_[Zn^[`B^00N@[XnESi2?03^_
20L000L705>_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g5bL@04LP0<T7:ARi6;TI2AT96@0Y40IjZAZY6ZTJnA
[i6_TJnA/96`Z[2Z/:n`[k2_/:n`[lN_ajo7[lN_ajo7/0N`1k07/0N`1jl7[`N_1jl7[`O61lH7bPO:
1lX7aPO61lH7[lZ_bZo:[lZ_bZo:[lZ_bZk:[/Z^]:j_[Zl00Zh0@iF^T:j@UHmf[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7D:l01aQa
679aLW400W80391bT7:ALY6;TH^AT0JA01:ZTJZA[i6_TJnA[i6_TJnZ[jX4[`0k/:n`[k2_/:n`[k2_
/:n`[lN_ajo7[lN_1jl7[`N_1jl7[`N_1jo:[lZ_bZo:[lZ_bZo:[lZ_bZo:[lX01Zl02:j_[Zn^[jj_
1:h029F^T9F?T8n@?:l81`001`L0Cjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W404W9aLW:;LY2;TH^AT96@
TI2AT0:A06NZTJZA[i6_TJnA/96`TK2A/:Z`[k2_/:n`[k2_/:o7[lN_ajl7[`N_1k07/0N`1k07/0N_
1jl7[`N_1jl7[`O:1lX7bPO:1jl7[`N_1jl7[lZ_bZo:[lZ_bZo:[lZ^bZjd[Zn^[jj_00:^04>@[Y2E
SiF_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00P700061dn_00Ia679/LW42LP0<Rg:@LY5bTH^ARi6@2Y40CJZA[i6_TJnA[i6_TJnZ[jZ`[k2_
/:n`[k2_/:n`[k2_/:n`[k2_ajo7[lN_1jl7[`N_1jl7[`N_1jo:[lZ_bZo:[lZ_bZo:[lZ_bZo:00R_
01>^[jj_[Zn^[iF^UJj@UHn@Si1f03f_20L000L7056_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`MaLW401780_8]b
TH^ARi6@TI2AT96@ZY6ZTJnA[i6_TK2A/96`TK2A/:Z`[k2_/:n`[k2_/:o7[lN_ajl7[`N`1k07/0N`
1k07/0N_1jl7[`N_1jl7]0Nd1k@7]0Nd1k@7[`N_1jl7[lZ_bZo:[lZ_bZo:[kB^]:j_[Zn^[iF_UJj@
UI2ES`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_20L000H7Cjl0174HLW44LP0<Rg:ALY5bTH^AT96@29409ZZAZY6_TJnA[i6_TK2A/:Z`[k2_
/:n`[k2_/:n`[k2_/:n`[k2_0k00::ne[`N_1jl7[`N_1jl7[`N_1jo:[lZ_bZo:[lZ_bZo:[lZ_bZnd
[k@6[`0:[Zn^[jj_UJjE[P:E00F@UHn@S`0o[`P700071`1?[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7L@06LP0<
T7:ARi6@TI2AT96@0Y40[ZZA[i6_TJnA[i6`TK2A/96`TK2Z/:n`[k2_/:n`[k2_/:l7[`N`1k07/0N`
1k07/0N`1k07/0N`1jl7[`N_1k@7]0Nd1k@7]0Nd1k@7[`N_1jo:[lZ_bZo:[lZ_]:nd[Zn^[jj_UJnE
[i2ET9F?1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`P700061dn_00Aa679a1780391bTG:ALY6;TI2AT0RA02BZTJnA[i6_TJnA/96`TK2_
/:n`[k2_/:n`[k2_/:n`[k2_/:l4/00W]K2e[kF_1jl7[`N_1jl7[`N_]Jl7[lZ_bZo:[lZ_bZnd[kB_
]:nd00J_016^[jj_UJnE[iF^UJj@UHn@MP11[`P700071`1?[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7L@04LP08
T7:ALY6@TI08T@0N[i6`TK2A/96`TK2A/96`[k2_/:n`[k2_/:n`[k2_1;00AKF`]K2e/0N`1k07/0N`
1k07/0N`1kD7]@Ne1kD7]0Nd1k@7]0Nd1k@7]0N_bZo:[lZ_bZo:[lZ_]:nd[Zn^[jj_UJnE[`02U@15
T9Ef1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061dj_00D0LQQbL@02LP0:T7:@LY5bTI2AT0bA04f_TJnA[i6_TJnA[i6`TK2A
/:n`[k2_/:n`[k2_/:n`[k2_/:ne[kF`]K2e[kF_]Jne[kF_]Jne[kF_]Jne[kF_]:nd[kB_]:nd[kB_
]004[`0<[Zn^[iF_UJnE[YF^0YD01I2ESi1f04>_20L000L704f_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`02L@Ab
00JALY5bTI0:T@0O[i6`TK2A/96`TK2A/96`[k2_/:n`[k2_/:n`[k2_]@03/013]K2e/;F`]K07/0N`
1k07/0N`1k07]@N`1kD7]@N_1jl7[`N_1jl7[`N_1jo:[kB_]:nd[kB_]:nd[[B^[jj_UJnE[`02U@17
T9Ef1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl020L000H7CZl00aQb6004LP0:T7:ALY5bTI2AT0ZA04^FTJnA[i6_TJnA[i6`TK2A
/96`U[2_/:n`[k2_/:n`[k2_/:n`[kF`]K2e/;F_]Jne[kF_]Jne[kF_]Jne[kF_]:nd[kB_]:nd[k@0
1Zl02Zj_[ZnE[iF_UJh2U@07T9F@UGJ@MP15[`P700071`1=[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W42LP06
Mg:ALY5b39406ZnA/96`TK2A/96`TK2F/9J`[k2_/:n`[k2_1;00AKF`]K2e/;F`]K2e/0N`1k07/0N`
1k07/0N`1k07/0N_1jl7[`N_1jl7[`N_]Jne[kB_]:nd[kB_]:nd[kB^[iF_UJnE[P02U@19T9EfT:l7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`081`001PM>[`03679>00Ab00R@LY5bTG:AT0bA04NFTJnA[i6_TJnA/96`TK2F/9J`U[2F
/:n`[k2_/:n`[k2_/:n`[kF_]Jne[kF_]Jne[kF_]Jne[kF_]Jne[kB_]:nd[kB_]008[`0:[ZnE[iF_
UJnE[`:E00N@UI2EMY1f04N_20L000L704j_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[g43LP06N7:ALY5b2Y406YJA
/96`TK2A/96`TK2A/9J`U[2F/:n`[k2_1[00@KF`]K2e/;F`]K2e/;F`1k07/0N`1k07/;F`1k2e/0N_
]Jl7[`N_1jl7[`N_]Jnd[kB_]:nd[kB_]9FdUJnE[iF_00:E04f@UI2EMPN_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`00
1PM>[`03679?00Ab00R@LY5bTGNAM`bA04>FTIJA[i6_TJnA/96`TK2F/9J`U[2F/9J`[k2_/:n`[k2_
/:n`[k2_]Jne[kF_]Jn`[k2_/:n`[k2_]Jne[kB_]:nd00Z_00RF[iF_UJnE[`:E00B@UI2E0Y02MTZ_
20L000L704j_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[g43LP06N7:ALY5b2Y406YJA/96`TK2A/96`TK2F/9J`U[2_
/:n`[k2_0k000jn`[`04/02<]K2e/;F`]K2e/;F`1k07/0N`]K2e[kF_]Jne[kF_1jl7[`N_]Jne[kB_
]:nd[kB_]:ndUJnE[iF_UJn@UI2EMPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PM>[`03679?00Ab00R@LY5b
TGNAN0ZA03VFTIJAUY6_TJnA/96`TK2F/9J`U[2F/9J`U[2F/:n`[k2_/:n`[k2_/:n`[kF_]Jne[k2_
/:n`[k000jl01kB_]:nd[k@01Zl04YJ_UZnF[iF_UJnEUY2ET9F@U@9g00=a[jl0Bjl81`001`L0CZl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_L@=b00IhLY5bTGL:T@0RUY6`TK2A/96`TK2A/9J`U[2F/9J`U[2F/:n`[k2_
/:n`[`>`08^_]Jne[kF`]Jne[kF_]Jne[kF_]Jne[kF_]Jne[kF_]Jne[kF_]Jnd[kB_]9J_UZnF[iF_
UJnEUY2ET7N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061dj_00<HLW8017802WMbTG:AMi5hTGP8T@0g
UY6FTIJAUY6`TK2A/96`TK2F/9J`U[2F/9J`U[2F/9J`U[2F/:n`[k2_/:n`[k2_/:n`[k2_/00<[`0F
UZnF[iJ_UJnE[iF_UIJ@UI2EMi1fMe:_20L000L704n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Ma009b00QgLWQb
TGNAM`JA05NFTIJAUY6GTK2A/96`TK2A/96`U[2F/9J`U[2F/:n`[k2_/:ne[kF_]Jne[kF_]Jne[kF_
]Jne[kF_]Jne[kF_]Jne[kF_]:nd[kBF[iJ_UZnE[iF_UJl00YD0FI2EMi1g1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_00P700061dn_019bLG9bMg9gLY5bTGNAN95hTGP8T@0]UY6FTIJA/96`TK2A/96`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F/:n`[k2_/:n`00:_02BF[iJ_UZnF[iJ_UZnF[iJ_UZnF[iJ_UJnEUYFF
T9J@UGN@MWMF[`P700071`1?[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7L@02LP08N79hLY5gTGL6T@2bUY6FTIJA
UY6GTK2A/96`TK2A/9J`U[2F/9J`U[2F/9J`U[2F/9JeU[FF]IJe[kF_]Jne[kF_]JneU[FF]IJeU[FF
]IJ_U[BF[iJ_UZnE[iFFUIJ@UY2EMi2_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061dn_
019bLG9bMg9gLY5bTGNAN95hTGP6T@11UY6FTIJAUY6FTK2A/96`TK2F/9J`U[2F/9J`U[2F/9J`U[2F
/9J`U[2F[iJ_UZnF[iJ_UZnF[iJ_UZnF[iJ_UZl00YH03YFFUIJ@UY2EMiEgT8V:FZl81`001`L0Cjl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1g400W8027MbN7:AMi5g1Y40BIJAUY6FTIJAUi6`TK2A/96`TK2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F]IJ`U[2F]JneU[FF]IJeU[FF]IJdU[BF[iJ_UZl00YH0IiFFT9J@UWNE
LJV:Z8X7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7Cjl0579>LW9gLWMbN7:AMi5g
TGRAN95h1Y40>IJAUY6FTINAUi6`TK2A/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F[iJ_UZnF[iJ_UZnF
[iJ_UZnF[`04UP0CUIJ@UY2EMiEgUGNYRZV:ZHZ?R@1I[`P700071`1?[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
L@04LP08N79hMi5gTGP4T@13UY6FTIJAUY6GTINA/96`TK2A/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J`UZnF[iJ_UZnF[`02UP1]UIJEUY2FT9IgZHnYSjV?ZHnYRXn:1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PM@[`03CW9b009b00igLWQbTG:AMi5hTGRAN0BA01:FTIJA
UY6FTIJAUY6FTINAUY44UP0G[iJ_UZnF/9J`U[2F[iJ_UZnF[iJ_UZl02YH039FFT9J@UY2FMiEgT09g
00maZHZYSjV?ZHZYRZV:ShT0Ejl81`001`L0Cjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl017802WQbN7NAMi5h
TGP2T@0kUY6FTIJAUY6FTINAUi6`TK2A/96`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`UZnF
[iJ_UZl00YH02YFFT9J@UY2FMi02M`0@T:V@ZI2YT:V@ZI2YSjV?Z@:?05N:1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl020L000H7DJl05W9>LTmgLWQbN7:AMi5hTGRAN95hTGP2T@0FUY6FTIJAUY6FTIJA
UY6FTIJAUY6FT@nF01:AUY6FTIJAUY2FT9J@UWNEMi02M`0;MY2:T8ZYRZV:ZHX01jT02:RYSjV?ZHZ?
0XYD[`P700071`1A[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_009b00YgLWQbTGNAN95h19408iJAUY6FTIJA
Ui6GTINAUiJGUYNF/9J`U[2F/9J`U[2F/9J`012F00RAUY2FT9IgT0=g00Z;RY2:ZI2YT:V@1ZT01jjY
[ZV^ZJh00ZT0FHnYSjV:1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061e:_04HHLTmgLWMbN79h
Mi5gTGRAN95hTGRAN95hUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJ@UWNFMiIgUWN@
17L01G6:LHYa00::00R@RZV:ZI2YT0fY00N?ZHnYRXm`05>_20L000L705:_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl70g802WQbN7NAMi5hTGP2T@0DUY6FTIJAUY6FTINAUY6GTINAUi43UP06TINAUi6G49H02Y2F
T9IgUWN@Mi42R`03LHYa00::00J@RjV@ZI04Z@0?[jV_ZJnY[ZV^ZJjY[ZV^00:Y05N?ZHn?L@N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl020L000H7E:l0?QQbCgMbMg9hLWQgTGNAN95hTGRAN95hUWRFTIJAUY6F
TIJAUY6FTIJAUY6FTIJAUY6FTIIgUWNFMi5gT7N@0WL02GJZT95bLFaaK002RP08RhZ@RZV;ZI0AZ@06
Z:V?ZHnY0XYB[`P700071`1D[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl70g802WQgN7NAMi5hTGP2T@0f
UY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJAUY6FTIJ@UWNFMiIgT7N`[k2Z[i6ZT8]a0XX01Y2;
ZH^YT0:Y06j_ZJnY[jV_ZJnY[jW6ZJnY[jV^ZJjY[ZV^SjV?RZl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P7
00061eF_03=bCW9?Mg9hLWQgN7NAMi5gTGNAMi5hTGRAN95hUWRFN9IhUWRFMiIgUWNFMiIgTGN@Mi00
1GL01k2_[i6_TJl00Y42R`0;K8Z:RhZ@RZV;ZI002:T03jjY[ZV^ZJjY[ZV^ZJjY[P02Z@05SjV?Z740
DJl81`001`L0EJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W80;gMbN7MhMi5gTGRAN9IhUWRFN9Ih
UY6FTIJAUY6FTIJAUY6FTIJ@UWNFMiIgUWNA009g01F_]K2e[k2_/:Z_TJn@TG6;RY2;ZI000ZT0LZZY
[jV_ZJnYaZW6ZLJYaZW6ZLJYaZW6ZLJY[ZV^Z:V?ZG47[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061eN_
02UbCWMbMg9gLWQgN7MhMi5gTGNAMi5gTGNAMi5gTGNAMi5gTGNAMi5gT004M`03CZn_00:_00Vd[k2_
/9J_TJl00Y402`2:LH^:T8^YT:V@00BY01fZZJnY[jV_ZJnY[jV_ZJnY[ZV^ZJjY[ZV^SjV?Z@02Se2_
20L000L705N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W809GQgN7MhMgQgTGNAMiIhUWNFN9Ih
UY6FTIJAUY6FMiIgUWNFMiH00WL00gJ_[`02[`0B]Jne[kF`]Jn`U[2ALHZ;RZV;0ZT0N:ZYZZV_ZJnY
[jW6ZLJYaZW6ZLJ^aZk6[/J^aZk6[/JYaZV^ZJj?ZHl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061eV_
00QbCgMbMg9gLP9g01=hMi5gTGNAMi5gTGNAMi5gTGN@00Ig00UaMdjET:nE[iH01Zl03[2_/96_JhYa
RhZ@RjV@1:T08jZYZZV_ZJnY[jV_ZJnY[jV_ZJnY[jV_ZJjY[ZV^ZJjY[XnY00:?D:l81`001`L0M:l7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g9gLWMbN7MhMgQgN7NAMi5gTGNAMi5gTGNA1GL2T00B
UY2FUJnF[iJe[kF_]Jne[hYa0Y00:jV@ZI2ZZJZY[jV_ZJnYaZW6ZLJ_aZo6[lJ_aZo6[/J^aZk6[/J^
aZk6[Zl00Zh0EJV^ZJj?1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7G:l01digCgMbMg804gL02GIg
MY1gUI2FT002UP08[iJ_UZnF/:l2L@0HRhZ@RjV@ZI2YT:ZYZZV_ZJnY[jV_ZJnY2Zl02lJ^aZk6[/J^
[jj_00B^00JY[ZV^SjU@[`P700071`1^[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_Mg9g
LWQgN7MhMgQgN7MhMgP7M`06T7NET9J@0YH02:nF[iJ_U[FF0W40=I2;ZH^YT:Z@ZZV_ZJnZ[jZ_Z/JZ
aZ[6Z/J_aZo6[lJ_aZo6[lJ_aZo6[/J^aZk6[/J^aZj_00:^05>Y[Xl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`081`001PMO[`08LTmgCgMbMg8<M`0:T7N@Mi1gUI2FT0BF02:_UW5aRhZ@RY2;ZH^YT:ZYZZV_ZJnY
[jV_ZZnZ[jZ_ZZnZ1:l04LJ_aZo6[lJ^aZk6[/J^aZj_00J^00FY[XnYRP1?[`P700071`1S[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LWMb00ag00Z@Mi1gUWNFT9J@19H00jnF[`02L@0j
RY2;ZH^ZT:Z@ZZV_ZZnZ[jZ_ZZnZaZ[6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_aZo6[/J^aZk6[Zn^[`:^
05>Y[XnX[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PMS[`06LTmgLWMb1WL03Y1gT7N@Mi1gUWNFT9JA
0YH2L@0<RhZ@RjV;ZY2ZT:Z@0ZX03:nZ[jZ_ZZnZ[jZ_ZPB_01_6[lJ_aZo6[lJ_aZo6[lJ_aZo6[/J^
aZj_[Zl01Zh00hnYS`1?[`P700071`1U[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1g8027L0395gTGNFMiJ@UY2FT0>F00=aRhX00X/2T00>ZY2ZT:n@[jV_ZZnZ[jX4[`0YaZo6[lJ_
aZo6[lJ_aZo6[lJ_aZo6[lJ_bZo:[lZ_bZo6[/J^aZj_[Zl00Zh0DI2X[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`081`001PMX[`03CWMb00Ag01:@Mi1gT7NAMiIgUW6:LH^:T8/2T00@ZY2ZT:Z@ZY2_ZJnZ[jZ_ZPR_
01o6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_aZo6[/J^[jj_00J^00>@ZHl0Cjl81`001`L0Jjl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Mb00Ag00ZAMi5gUWNFT9Ia0X/04Y2;
ZY2ZT:Z@[i2_ZZnZ[jZ_ZPJ_01;6[lJ_ajo7[lN_ajo7[lN_aZl4aP0Eb/K6[lZ_bZo:[lZ_bZk6[Zn^
[jj_00:^056@Sjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7KZl00digLP02M`04T7NAM`9a00B@Ri2;
19003ZZ@ZY2_TJnZ[jZ_ZZnZ2Zl08lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_aZo6[/J^[jj_00B^
00>@ZHl0Cjl81`001`L0Ljl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1g4018/2T00>ZY2ZT:nA[i6_ZZnZ[jX6[`0aajo7[lN_ajo7[lN_ajo7[lO61lH7aPO6
1lH7aPO61lK:[lZ_bZo:[lZ_bZk6[Zn^[`02[P1AT8n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061g6_
00AaK8]a0X/0692;TI2ZT:Z@ZY6_TJnZ[jZ_ZZnZ[jZ_ZPR_01g6[lJ_aZo6[lJ_aZo6[lJ_aZo6[lJ_
aZo6[lJ_aP04[`04[Zn^[`:^00FY[Y2^S`1?[`P700071`1_[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_009a0X/0492;TI2ZTJnA[i6_TJnZ[jX4[`0g/:n`[k2_
ajo7[lN_ajl7[lN_1jl7aPO61lH7aPO61lH7aPO61lK:a/Z_bZo:[lZ_bZk:[Zn^[`02[P1AT8n_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_00P700061fn_0W4018]bT8/2T00DTI2ZT:ZAZY6_TJnA[jZ_ZZnZ[jX:[`0O
aZo6[lJ_aZo6[lJ_aZo6[lZ_bZo:[lZ_aZo6[lJ_aP04[`04[Zn^[`:^00BE[Y2@D:l81`001`L0KJl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W80Ci2;TI2AT:ZA
ZY6_TJnA[jZ_Z[2_/:n`[k2_/:n`[lN_ajo7[`N_1jl7[`O61lH7aPO61lH7aPO61lH7aPO6b/K:[lZ_
bZo:[lZ^[jj_[Zl00Zh0DHl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PM^[`05LH^;T8/00Y004Y6@
TI6ZTJnA[i6_TJnA[jZ_ZPR_02>`[k2_/:o7[lN_aZo7[lJ_aZo6[lZ_bZo:[lZ_bZo:[lJ_aP06[`04
[Zn^[`:^00B@[Xn@D:l81`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1g6@Ri6@TI2ZTJZA[i6_TK2Z/:Z`Z[2_/:n`[k2_/:o7[lN_1jl7[`N_1jl7aPO61lH7
aPO61lH7aPO61lH7aPO6bZo:[lZ_bZo:[lZ^]:j_[ZnE[Xl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004Zl7[`N_1jl7[`N_1jl7[`N_1`P7
00061fb_00U/LW:@Ri6@TI000Y403:ZA[i6_TJnA[i6_T@R_02F`[k2_/:n`[lN_ajo7[lN_ajl7[`N_
1jl7[lZ_bZo:[lZ_bZo:00R_00R^[jj_UJj@[P:?D:l81`001`L0ojl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`MaRg:AT96@ZY6_TJnA[i6`TK2Z/:Z`[k2_/:n`[k2_/:n`
[`N_1jl7[`N_1jl7[`O61lH7aPO61lH7[`N_1jl7[`N_bZo:[lZ_bZo:[kB^[jj_UJn@UGH7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`004Zl7
[`N_1jl7[`N_1jl7[`N_1`P700061fZ_00THLW6@Ri6@TI0019403:ZA[i6_TJnA[i6_ZPB_02^`[k2_
/:n`[k2_/:n`[lN_ajl7[`N_1jl7[`N_1jo:[lZ_bZo:[lZ_]:nd00J_00V^[iF^UJj@UGH0DJl81`00
1`L0KJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[h]bT8/00Y40;:ZA
ZY6_TJnA[i6`TK2Z/:n`[k2_/:n`[k2_/:o7/0N`1k07/0N_1jl7/0N`0`L08<X7bPN_1k@7[`Nd1k@7
]0N_bZo:[lZ_]:nd[ZnE[iF^0Xl0DJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PMX[`09679bRg:A
T96@00:A00bZTJnA[i6_TJnA[jX2[`0c/:n`[k2_/:n`[k2_/:n`[k2_/:ne[kF_1jl7[`N_1jl7[`N_
1jo:[lZ_]:nd[kB_]:nd00:_00J^[jj_UJh2U@9fDZl81`001`L0Ijl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W8018]bTI02T@0H[i6_TJnA/96`TK2_/:n`[k2_/:n`[k2_
0[00R;F`]K2e/0N`1k07/0N`1k07]@Nd1k@7]0Nd1k@7]0Nd1k@7[`N_bZo:[kB_]:j_UJnE[Y2?[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PMX[`05LG9bT7801Y406ZnA[i6_TJnA[i6`U[2_/:n`[k2_
/:n`[k2_1[008KF`]K2e[`N_1jl7[`N_1jne[kF_]:nd[kB_]:nd[kB_]002[`04[ZnE[P:E0Y1D[`P7
00071`1W[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7L@02LP0LTI2ATJnA
[i6_TK2A/96`U[2_/:n`[k2_/:n`[`:`08Ze/;F`]K2e/;F`1k07/0N`1k07]@Ne1k@7]0Nd1k@7]0Nd
1jl7[lZ_]:nd[kB_]:j_UJj@UJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl81`001PMW[`06LG:@LY6@
1Y406ZnA[i6_TJnA/96`U[2_/:n`[k2_/:n`[k2_1[007kF`]K2e[kF_]Jne[kF_]Jne[kF_]:nd[kB_
]:nd[k@00Zl01:j_UJl2U@03T9F_05F_20L000L706^_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`MaT7:AT002T@0F[i6_TJnA/96`TK2F/:n`[k2_/:n`[`B`08Ze/;F`]K2e/;F`
]K2e/;F`1k07]@Nd1k@7]0Nd1k@7[kF_]Jnd[kB_]:nd[kB^[iF_T9F_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl81`001PMV[`07679bT7:AT004T@0JUY6_TJnA[i6`TK2F/9J`[k2_/:n`[k2_/:l8/00M
]K2e/;F_]Jne[kF_]Jne[kF_]:nd[kB_]:nd[k@00Zl029F_UJnE[i2EF:l81`001`L0Jjl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LG:ALY6@00:A01:_TJnA/96`TK2A/9J`U[2_
/:l8/00_]K2e/;F`]K2e/;F`]K2e/0N`1jl7]0Nd1k@7];F_]Jne[kF_]:nd[kBF]9FdUJl00YD0FJl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061fJ_00LHLW:@LY6@00BA01bFTJnA[i6`TJnA
/9J`U[2F/:n`U[2_/:n`[k2_1[005kF`]Jne[kF_]Jne[kF_]Jne[kB_]:nd00B_00ZF[iF_UJnEUY2E
FZl81`001`L0Jjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_LG:AMi6@00:A
01J_TK2A/96`TK2A/9J`U[2F/:n`[k2_0[00S;F`]K2e/;F`]K2e/;F`]K2e/;F_]Jne[kF_]Jne[kF_
]Jne[kB_]:ndU[BE[iF_T9Eg1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7IZl01AQb
LY5b00JA01bFTIJA[i6_TK2A/9J`U[2F/9J`U[2F/:n`[k2_1[004kF_]Jne[kF_]Jne[kF_]Jnd[k@0
2:l019F_UJl2U@:@00=f[jl0FJl81`001`L0Jjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_LG:AMi6@00:A01BFTJnA/96`TK2F/9J`U[2F/9J`UPR`02Ne/;F`]K2e/;F`]K2e[kF_
]Jne[kF_]Jne[kF_]JndU[BF[iF_UIH00YD0GgH7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_00P700061fN_0W80195bTGL6T@0YUY6FTIJA/96`TK2F/9J`U[2F/9J`U[2F/9J`U[2F/9J`U[2F
/:n`[k001Zl03YJ_UZnF[iJ_UJnEUY2E0WMP[`P700071`1[[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7LI1gTI000Y4059JAUY6`TK2A/9J`U[2F/9J`U[2F0k001Jn`[k2_00:`
08Je/;F_]Jne[kF_]Jne[kF_]Jne[kF_]IJdU[BF]9J_UJn@UGL7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`P700061fN_0W80195bTGL6T@0iUY6FTIJA[i6`TK2A/9J`U[2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J_UZnF[iJ_UZnF[iJ_UJnEUY2E009gI:l81`001`L0Jjl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g5gLY5g00:A0:BFTIJAUi6`TK2F/9J`U[2F/9J`U[2F
/9J`U[2_/:n`[kF_]JneU[FF]IJe[kF_]IJdUZnF[iJ_UZnEUY2EMi2_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061fR_00E>Mg:ALP06T@0[UY6FTIJAUY6_TJnA/96`U[2F
/9J`U[2F[iJ`U[2F/9J_UZnF[iJ_UZnF[`06UP04T9IgU@9gJ:l81`001`L0Ijl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl00W80=I5gTGRFTIJAUY6GTK2A/9J`U[2F/9J`U[2F
/9J`U[2F/9J`U[FF]IJeU[FF]IJ`U[2F[iJ_00BF06n@UWN_T0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7JJl2LP04TG:AN0JA00nFTIJAUY6FTIJAUY6_UZl0
0iH04JnF[iJ_UZnF[iJ_UZnF[iJ_00JF00^@UWNEMjn^[jV^S`1Y[`P700071`1Y[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Ma009g02^AMi6AUY6FTIJAUi6`TK2F/9J`U[2F
/9J`U[2F/9J`U[2F/9J`U[2F/9J_00JF00Z@UY2@LLJ_aZo60Zh0JHl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`081`001PMZ[`07CWMbTGNAN006T@0<UY6FTIJAUY6FTIJA
3iH02Y6FT9J@UY2FMi02MPF_00O6[jn^[i2Y06R_20L000L707n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[g9gTGNAN9JAUY6FTIJAUY6FTIJA00>F012AUY6FTIJAUY6F
TIJAUY2F0gL2[`1jajo7[`N_1jl7[`N_bZo6[lJ^ZJl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl81`001PM/[`0RCW9bTGNAMi5hTGRFN9JAUY6FTIJAUY6FTIJAUWNFMiIgT@=g
00AaT:ZZ1:l03LJ_aZo6[lJ_aZo6[lH00Zl00jj_S`1W[`P700071`2:[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7LWQgTGNAMiIhUY6FTIJAUY6FTIJAUY6FT9IgUWN@
0WL2/003RjVZ00:_01F`[lN_1jl7[`N_1lH7aPN_bZo:[lH00Zh0IhX7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7KZl054igLWQgTGNAMi5gTGNAMi5gTGNA1GL02ZnF
/9J`U[2@[jX8[`0=ajo7[`N_1jo:[lJ_aP02[`04[Zn@SfJ_20L000L706n_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1g800WL03GQgTGNAMiIgTGNAMgP017L0SiJ_
U[FF]IJeLJnZ/:n`[lN_ajl7[`N_1jl7[`O61lH7[`N_bZo6[Zn?1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_00P700061g6_00IgCgMbMg85M`03LWM>009g00RET9JF[iJ_UP:@
00>_ZZl00jl05K2_/:n`[lN_1jl7[`N_1jl7[lZ_aP02[`04[Zn^SfJ_20L000L707>_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Ma00Yg00BFT9JA0YH01:nF
/8/2ZP28/:n`[k2_ajo7[`N`1jl7[`N_1jl7[`N_1jl7[lZ_]:j_L@N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061gJ_00=>Mdl017L01Y1gUY2FT@:F00B@TJnA0Zl06k2_
/:n`[k2_/:n`[k2_1jl7[`N_1jl7[lZ_]002[`03[ZnE06N_20L000L707V_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`Ma009g01:AMiJ@UY2FRjZA
/:n`[k2_/:l2/00J]K07/0N`1k07/0N_1jl7]0Nd1k@7[lZ_]:h2[`1V1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7Ojl09g9gMgNZTJnA/:n`[k2_/:n`[k2_/:n`[kF_
1jl7[`N_1jl7[kB_]002[`03[Zn@06N_20L000L708^_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jnZTK2_/:n`[k2_00:`01Ve/;F`
]K07/0N`1k07]@Nd1k@7]0Nd1jnd00:_06VET:l7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl020L000H7P:l03G6ATJnF/:n`[k2_/:l00k005Zn`/;F_]Jne[kF_]Jne[kF_]:nd
[k@2[`03UI2_06N_20L000L707n_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_00:A00B`[k2_1[00Q[F`]K2e/;F`1k07/0Ne1kD7]@Nd
1kBe[kF_]9F^[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_20L000H7
P:l02i2_TK2F/9J`[k2_00J`01Fe/;F`]Jne[kF_]Jne[kF_]:ndUZl00YEZ[`P700071`23[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
LYJA/:l02;003;F`]K2e/;F`]K2e/0Fe07Fd]KBe];F_]9J_T0N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl020L000H7Ojl2T@0:[i6`U[2F/9J`[`F`01:_/:ne[kF_]Jne
[kF_]Jnd[k@2[`:E00=f[jl0JJl81`001`L0Pjl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1gNFTK2F00R`01Ne/;F`]K2e/;F`]K2e/;F_
]Jne[kF_]002UP1]MPN_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`081`001PMo[`9g01^_TK2F/9J`U[2F/9J`U[2F/9J`[kFF]Jne[kD00Zl019J_UZl2UFj_20L000L7
08F_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_UY6`U[2F00J`08Je/;F`]K2e/;F_]Jne[kF_]JneUZnEUZl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_
1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`P700061h2_01agTI6_U[2F/9J`U[2F/9J`
U[2F/9J`U[2F[iJ_0iH00iEg[`1`[`P700071`2J[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`NA/9J`U[2F/9J`U[2_]IJeU[FF]IJe
UZnF[`:F07Ea1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl020L000H7Pjl2T@0;UY6FTIJAUY6FT9400gMo[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B
[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7
[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7
[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl7
20L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7
ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P7
00071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_000B[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B
[`N_1jl7[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7
[`N_1jl7[`N_1jl720L000H7ojlC[`P700071`3o[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7
[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_1jl7[`N_000B[`N_1jl7[`N_1jl7
[`N_1jl720L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"POVRay"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVWritePOVRay["\<f:/Develop/mview3d/shell.pov\>"]\)], "Input",
  CellLabel->"In[33]:=",
  CellTags->"POVRay"],

Cell[BoxData[
    \("f:\\Develop\\mview3d\\shell.pov"\)], "Output",
  CellLabel->"Out[33]=",
  CellTags->"POVRay"]
}, Open  ]],

Cell["\<\
The output is designed to uses a aspect ratio of 1 for 3d graphics. \
For POVRay the width and the height of the image should be set to the same \
value. The triangles written to the file are grouped by the surface color and \
a table of color entries is declared. With this feature a redefinition of the \
surface colors or a change to volume  textures is possible. Since the concept \
of textures in OpenGL and POVRay is very different the texture export is \
limited to planar, spherical and cylindrical mapping.\
\>", "Text",
  CellTags->"POVRay"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVWrite3DMF["\<i:/images/shell.3dmf\>"]\)], "Input",
  CellLabel->"In[5]:=",
  CellTags->"POVRay"],

Cell[BoxData[
    \("i:/images/shell.3dmf"\)], "Output",
  CellLabel->"Out[5]=",
  CellTags->"POVRay"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Writing DXF files", "Section",
  CellTags->"DXFFiles"],

Cell[TextData[{
  "For more complex scenes a modeler or an alternative render engine may be \
helpful. The 3D file format supported by nearly all modelers is the DXF \
format. ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " can write DXF files with the ",
  StyleBox["Utilities`DXF`",
    FontFamily->"Arial"],
  " package but this is slow. One may use the OpenGL viewer to write the DXF \
files. With the command"
}], "Text",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVWriteDXF["\<f:/tmp/sdodeca.dxf\>", 
      Graphics3D[Stellate[Dodecahedron[]]]]\)], "Input",
  CellLabel->"In[7]:=",
  CellTags->"DXFFiles"],

Cell[BoxData[
    \("f:\\tmp\\sdodeca.dxf"\)], "Output",
  CellLabel->"Out[7]=",
  CellTags->"DXFFiles"]
}, Open  ]],

Cell[TextData[{
  "a DXF file is written. With out the ",
  StyleBox["Graphics3D[]",
    FontFamily->"Courier New"],
  " object in the second argument the current contents of the OpenGL viewer \
window is written to the DXF file. ",
  StyleBox["MVWriteDXF[]",
    FontFamily->"Courier New"],
  " has the same options as the ",
  StyleBox["MVWritePOVRay[]",
    FontFamily->"Courier New"],
  " function except ",
  StyleBox["MVPolygonShading",
    FontFamily->"Courier New"],
  ", because DXF files store no normal vectors of the surface. Most modelers \
include an option to smooth imported DXF files."
}], "Text",
  CellTags->"DXFFiles"]
}, Closed]],

Cell[CellGroupData[{

Cell["Reading DXF files", "Section",
  CellTags->"DXFFiles"],

Cell["\<\
The OpenGL viewer can also read DXF files.  At first the graphic \
must be loaded into MathGL3d .\
\>", "Text",
  CellTags->"DXFFiles"],

Cell[BoxData[
    \(MVClear[]; MVLoadDXF["\<mathlogo.dxf\>"]; \)], "Input",
  CellLabel->"In[2]:=",
  CellTags->"DXFFiles"],

Cell[BoxData[
    \(\(mlogo = MVGetGraphics3D[]; \)\)], "Input",
  CellLabel->"In[3]:=",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[mlogo, ViewPoint -> {0.000, \ 3.293, \ 0.777}]; \)\)], "Input",
  CellLabel->"In[4]:=",
  CellTags->"DXFFiles"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .13387 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-2.22045e-16 1 -0.432636 1 [
[ 0 0 0 0 ]
[ 1 .13387 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.9855 .01778 m
.9895 .13387 L
s
.9895 .13387 m
1 .11842 L
s
1 .11842 m
.99582 0 L
s
.99582 0 m
.9855 .01778 L
s
.0145 .01778 m
.00418 0 L
s
.00418 0 m
0 .11842 L
s
0 .11842 m
.0105 .13387 L
s
.0105 .13387 m
.0145 .01778 L
s
.9855 .01778 m
.9895 .13387 L
s
.9895 .13387 m
.0105 .13387 L
s
.0105 .13387 m
.0145 .01778 L
s
.0145 .01778 m
.9855 .01778 L
s
.99582 0 m
.00418 0 L
s
.00418 0 m
0 .11842 L
s
0 .11842 m
1 .11842 L
s
1 .11842 m
.99582 0 L
s
0 0 m
1 0 L
1 .13387 L
0 .13387 L
closepath
clip
newpath
.5 Mabswid
.83556 .02267 m .83598 .01868 L .82635 .02237 L closepath p .819 .994 1 r
F P
0 g
s
.82711 .01804 m .82635 .02237 L .83598 .01868 L closepath p .824 .977 1 r
F P
0 g
s
.82635 .02237 m .82711 .01804 L .81941 .02301 L closepath p .878 1 1 r
F P
0 g
s
.82635 .02237 m .83556 .02267 L .83163 .02564 L closepath p .955 .786 .906 r
F P
0 g
s
.82635 .02237 m .83163 .02564 L .82159 .02796 L closepath p .955 .786 .906 r
F P
0 g
s
.81941 .02301 m .82635 .02237 L .81957 .02942 L closepath p .955 .786 .906 r
F P
0 g
s
.82159 .02796 m .81957 .02942 L .82635 .02237 L closepath p .955 .786 .906 r
F P
0 g
s
.10121 .02331 m .10116 .01916 L .10015 .02237 L closepath p .445 .651 1 r
F P
0 g
s
.09922 .01804 m .10015 .02237 L .10116 .01916 L closepath p .552 .772 1 r
F P
0 g
s
.10015 .02237 m .09922 .01804 L .09928 .02309 L closepath p 1 1 1 r
F P
0 g
s
.09442 .04756 m .09566 .04068 L .10015 .02237 L closepath p .955 .786 .906 r
F P
0 g
s
.09972 .02273 m .10015 .02237 L .09566 .04068 L closepath p .955 .786 .906 r
F P
0 g
s
.10456 .02987 m .10015 .02237 L .10121 .02331 L closepath p .955 .786 .906 r
F P
0 g
s
.10015 .02237 m .10456 .02987 L .08867 .07285 L closepath p .955 .786 .906 r
F P
0 g
s
.84355 .02359 m .84484 .01932 L .83556 .02267 L closepath p .809 .964 1 r
F P
0 g
s
.83598 .01868 m .83556 .02267 L .84484 .01932 L closepath p .804 .981 1 r
F P
0 g
s
.83554 .02636 m .83163 .02564 L .83556 .02267 L closepath p .955 .786 .906 r
F P
0 g
s
.83556 .02267 m .84355 .02359 L .83554 .02636 L closepath p .955 .786 .906 r
F P
0 g
s
.09566 .04068 m .09658 .03525 L .09972 .02273 L closepath p .955 .786 .906 r
F P
0 g
s
.09928 .02309 m .09972 .02273 L .09658 .03525 L closepath p .955 .786 .906 r
F P
0 g
s
.53961 .0242 m .54118 .02044 L .53791 .02275 L closepath p .559 .71 1 r
F P
0 g
s
.538 .01842 m .53791 .02275 L .54118 .02044 L closepath p .587 .769 1 r
F P
0 g
s
.53791 .02275 m .538 .01842 L .53718 .02336 L closepath p 1 1 1 r
F P
0 g
s
.53359 .0447 m .53452 .03861 L .53791 .02275 L closepath p .955 .786 .906 r
F P
0 g
s
.53755 .02306 m .53791 .02275 L .53452 .03861 L closepath p .955 .786 .906 r
F P
0 g
s
.53791 .02275 m .53961 .0242 L .54584 .03585 L closepath p .955 .786 .906 r
F P
0 g
s
.53791 .02275 m .54188 .0293 L .52925 .06672 L closepath p .955 .786 .906 r
F P
0 g
s
.8197 .01875 m .81941 .02301 L .82711 .01804 L closepath p .879 1 1 r
F P
0 g
s
.81941 .02301 m .8197 .01875 L .81335 .02492 L closepath p .947 1 1 r
F P
0 g
s
.81335 .02492 m .81941 .02301 L .81755 .03087 L closepath p .955 .786 .906 r
F P
0 g
s
.81957 .02942 m .81755 .03087 L .81941 .02301 L closepath p .955 .786 .906 r
F P
0 g
s
.53452 .03861 m .5352 .03384 L .53755 .02306 L closepath p .955 .786 .906 r
F P
0 g
s
.53718 .02336 m .53755 .02306 L .5352 .03384 L closepath p .955 .786 .906 r
F P
0 g
s
.09722 .01904 m .09928 .02309 L .09922 .01804 L closepath p 1 1 1 r
F P
0 g
s
.09928 .02309 m .09722 .01904 L .0984 .02523 L closepath p 1 1 .9 r
F P
0 g
s
.09658 .03525 m .0972 .03129 L .09928 .02309 L closepath p .955 .786 .906 r
F P
0 g
s
.09884 .02416 m .09928 .02309 L .0972 .03129 L closepath p .955 .786 .906 r
F P
0 g
s
.10303 .02612 m .10362 .02239 L .10121 .02331 L closepath p .254 .439 .928 r
F P
0 g
s
.10116 .01916 m .10121 .02331 L .10362 .02239 L closepath p .267 .479 .959 r
F P
0 g
s
.10121 .02331 m .10303 .02612 L .10288 .02659 L closepath p .955 .786 .906 r
F P
0 g
s
.53615 .01932 m .53718 .02336 L .538 .01842 L closepath p 1 1 1 r
F P
0 g
s
.53718 .02336 m .53615 .01932 L .53648 .0252 L closepath p 1 1 .814 r
F P
0 g
s
.5352 .03384 m .53563 .0304 L .53718 .02336 L closepath p .955 .786 .906 r
F P
0 g
s
.53683 .02428 m .53718 .02336 L .53563 .0304 L closepath p .955 .786 .906 r
F P
0 g
s
.85033 .02511 m .85194 .02093 L .84355 .02359 L closepath p .773 .926 1 r
F P
0 g
s
.84484 .01932 m .84355 .02359 L .85194 .02093 L closepath p .773 .926 1 r
F P
0 g
s
.83946 .02707 m .83554 .02636 L .84355 .02359 L closepath p .955 .786 .906 r
F P
0 g
s
.84355 .02359 m .85033 .02511 L .83946 .02707 L closepath p .955 .786 .906 r
F P
0 g
s
.48308 .02208 m .48491 .02488 L .48308 .02173 L 1 1 .916 r
F
.48299 .02219 m .48491 .02488 L .48298 .02193 L F
.48302 .02325 m .48356 .02372 L p .48299 .02219 L F P
0 g
s
.48356 .02372 m .48491 .02488 L p .48299 .02219 L 1 1 .916 r
F P
0 g
s
.48491 .02488 m .48365 .02061 L .48308 .02173 L p 1 1 .916 r
F P
0 g
s
.48365 .02061 m .48546 .02442 L .48598 .01964 L closepath p .986 1 1 r
F P
0 g
s
.48546 .02442 m .48365 .02061 L .48491 .02488 L closepath p 1 1 1 r
F P
0 g
s
.48598 .01964 m .48577 .007 L .48365 .02061 L closepath p .817 1 1 r
F P
0 g
s
.48577 .007 m .4834 .00798 L .48343 .00933 L p .817 1 1 r
F P
0 g
s
.48343 .00933 m .48365 .02061 L .48525 .0103 L p .817 1 1 r
F P
0 g
s
.48525 .0103 m .48577 .007 L p .48343 .00933 L .817 1 1 r
F P
0 g
s
.48077 .00699 m .48106 .01963 L .48277 .0142 L p .11 .256 .799 r
F P
0 g
s
.48365 .02061 m .4834 .00798 L .48283 .01756 L p .939 1 .63 r
F P
0 g
s
.48267 .00942 m .48077 .00699 L p .48145 .00945 L .11 .256 .799 r
F P
0 g
s
.48145 .00945 m .48277 .0142 L .48267 .00942 L .11 .256 .799 r
F
.482 .00857 m .48267 .00942 L p .48118 .00813 L .857 .376 .09 r
F P
0 g
s
.48277 .0142 m .4815 .01823 L p .48291 .02112 L .11 .256 .799 r
F P
0 g
s
.4815 .01823 m .48106 .01963 L .48293 .02203 L p .11 .256 .799 r
F P
0 g
s
.48293 .02203 m .48365 .02061 L p .48283 .01756 L .939 1 .63 r
F P
0 g
s
.48308 .02173 m .48298 .02193 L p .48308 .02208 L 1 1 .916 r
F P
0 g
s
.48293 .02203 m .48405 .02347 L p .4815 .01823 L .11 .256 .799 r
F P
0 g
s
.48106 .01963 m .48023 .02396 L .48304 .0236 L p .425 .573 1 r
F P
0 g
s
.48298 .02193 m .48293 .02203 L p .48302 .02325 L 1 1 .916 r
F P
0 g
s
.48301 .02324 m .48302 .02325 L p .48293 .02203 L 1 1 .916 r
F P
0 g
s
.4855 .02677 m .48491 .02488 L .48308 .0233 L p 1 .738 .314 r
F P
0 g
s
.48308 .0233 m .48312 .0234 L .4855 .02677 L 1 .738 .314 r
F
.48326 .0238 m .4855 .02677 L p .48303 .02327 L F P
0 g
s
.48301 .02324 m .48316 .02358 L .48303 .02327 L 1 .738 .314 r
F
.48316 .02358 m .48326 .0238 L .48303 .02327 L F
.48309 .02529 m .48353 .02445 L p .48304 .0236 L .426 .553 .986 r
F P
0 g
s
.48353 .02445 m .48309 .02529 L p .48368 .02592 L .373 .487 .94 r
F P
0 g
s
.48423 .02623 m .48368 .02592 L .48442 .02672 L .373 .487 .94 r
F
.48353 .02445 m .48368 .02592 L .48423 .02623 L F
.48727 .0277 m .4855 .02677 L .48436 .0263 L p .394 .018 .018 r
F P
0 g
s
.48473 .02651 m .48727 .0277 L p .48436 .0263 L .394 .018 .018 r
F P
0 g
s
.487 .02791 m .48727 .0277 L .48541 .02683 L p .461 .03 .03 r
F P
0 g
s
.487 .02791 m .48541 .02683 L .4861 .02735 L .461 .03 .03 r
F
.48631 .02751 m .487 .02791 L p .48473 .02651 L F P
0 g
s
.51267 .02397 m .5127 .01964 L .48601 .02397 L closepath p .848 1 1 r
F P
0 g
s
.48598 .01964 m .48601 .02397 L .5127 .01964 L closepath p .848 1 1 r
F P
0 g
s
.48601 .02397 m .48598 .01964 L .48546 .02442 L closepath p 1 1 1 r
F P
0 g
s
.49153 .0277 m .48999 .0277 L .48601 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.48564 .02665 m .48601 .02397 L .48999 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.48601 .02397 m .49934 .02397 L .50818 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.48601 .02397 m .50818 .0277 L .50297 .03242 L closepath p .955 .786 .906 r
F P
0 g
s
.48601 .02397 m .49449 .02819 L .49153 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.48999 .0277 m .48845 .0277 L .48564 .02665 L closepath p .955 .786 .906 r
F P
0 g
s
.48727 .0277 m .48564 .02665 L .48845 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.5127 .01964 m .5129 .007 L .48598 .01964 L closepath p .623 .977 1 r
F P
0 g
s
.48577 .007 m .48598 .01964 L .5129 .007 L closepath p .623 .977 1 r
F P
0 g
s
.48619 .02046 m .48405 .02347 L .48631 .02751 L p .09 .09 .652 r
F P
0 g
s
.48353 .02445 m .48421 .02618 L .48434 .02626 L .373 .487 .94 r
F
.48541 .02683 m .48473 .02651 L p .4861 .02735 L .461 .03 .03 r
F P
0 g
s
.48444 .02633 m .4855 .02677 L .4833 .02384 L p 1 .654 .198 r
F P
0 g
s
.48444 .02633 m .4833 .02384 L .48355 .02442 L 1 .654 .198 r
F
.4836 .02454 m .48465 .02696 L .48804 .0306 L .373 .487 .94 r
F
.48353 .02445 m .48465 .02696 L .4836 .02454 L F
.48304 .0236 m .48023 .02396 L p .48309 .02529 L .426 .553 .986 r
F P
0 g
s
.48252 .02282 m .48377 .02606 L .48436 .0263 L p 1 .654 .198 r
F P
0 g
s
.48436 .0263 m .48444 .02633 L p .48414 .0258 L 1 .654 .198 r
F P
0 g
s
.48436 .0263 m .48414 .0258 L .48252 .02282 L 1 .654 .198 r
F
.4831 .02743 m .48553 .02899 L p .48309 .02529 L .373 .487 .94 r
F P
0 g
s
.48553 .02899 m .48804 .0306 L p .48465 .02696 L .373 .487 .94 r
F P
0 g
s
.48309 .02529 m .48553 .02899 L .48465 .02696 L .373 .487 .94 r
F
.48353 .02445 m .48355 .02442 L p .48325 .02378 L .426 .553 .986 r
F P
0 g
s
.48316 .02358 m .48304 .0236 L p .48353 .02445 L .426 .553 .986 r
F P
0 g
s
.48325 .02378 m .48252 .02282 L p .48444 .02633 L 1 .654 .198 r
F P
0 g
s
.48405 .02347 m .48316 .02358 L p .48325 .02378 L .426 .553 .986 r
F P
0 g
s
.4833 .02384 m .48325 .02378 L p .48355 .02442 L 1 .654 .198 r
F P
0 g
s
.48355 .02442 m .48405 .02347 L p .48325 .02378 L .426 .553 .986 r
F P
0 g
s
.48804 .0306 m .48405 .02347 L .48353 .02445 L p .373 .487 .94 r
F P
0 g
s
.48405 .02347 m .48293 .02203 L p .48301 .02324 L .425 .573 1 r
F P
0 g
s
.48308 .0233 m .48301 .02324 L p .48312 .0234 L 1 .738 .314 r
F P
0 g
s
.48316 .02358 m .48405 .02347 L p .48301 .02324 L .425 .573 1 r
F P
0 g
s
.48283 .01756 m .48332 .00934 L p .48268 .00988 L .939 1 .63 r
F P
0 g
s
.48405 .02347 m .48381 .01088 L .48277 .0142 L p .11 .256 .799 r
F P
0 g
s
.48405 .02347 m .48277 .0142 L .48291 .02112 L .11 .256 .799 r
F
.48381 .01088 m .48405 .02347 L .48619 .02046 L p .09 .09 .652 r
F P
0 g
s
.487 .02791 m .48626 .02484 L p .48679 .01573 L .262 .642 1 r
F P
0 g
s
.48681 .01538 m .48681 .01576 L p .48679 .01573 L .262 .642 1 r
F P
0 g
s
.48681 .01576 m .487 .02791 L p .48679 .01573 L .262 .642 1 r
F P
0 g
s
.4861 .01498 m .48681 .01538 L p .48626 .02484 L .262 .642 1 r
F P
0 g
s
.48727 .0277 m .487 .02791 L .48845 .0277 L closepath p .078 .02 .02 r
F P
0 g
s
.48839 .02791 m .48845 .0277 L .487 .02791 L closepath p .078 .02 .02 r
F P
0 g
s
.48839 .02791 m .487 .02791 L .48681 .01576 L p .619 .975 1 r
F P
0 g
s
.48681 .01576 m .48681 .01538 L .48686 .0158 L p .619 .975 1 r
F P
0 g
s
.48686 .0158 m .48839 .02791 L p .48681 .01576 L .619 .975 1 r
F P
0 g
s
.48822 .01538 m .48839 .02791 L .48681 .01538 L closepath p .619 .975 1 r
F P
0 g
s
.48619 .02046 m .48786 .01812 L .4861 .01498 L p .09 .09 .652 r
F P
0 g
s
.4861 .01498 m .48381 .01088 L p .48619 .02046 L .09 .09 .652 r
F P
0 g
s
.48626 .02484 m .48353 .01351 L .48604 .01495 L p .262 .642 1 r
F P
0 g
s
.48604 .01495 m .4861 .01498 L p .4861 .01501 L .262 .642 1 r
F P
0 g
s
.4861 .01501 m .48626 .02484 L .48604 .01495 L .262 .642 1 r
F
.48293 .02203 m .48106 .01963 L p .48304 .0236 L .425 .573 1 r
F P
0 g
s
.48226 .01022 m .48252 .02282 L .48283 .01756 L p .939 1 .63 r
F P
0 g
s
.48226 .01022 m .48283 .01756 L .48268 .00988 L .939 1 .63 r
F
.48267 .00942 m .48226 .01022 L p .48332 .00934 L F P
0 g
s
.48277 .0142 m .48381 .01088 L .48289 .0097 L p .11 .256 .799 r
F P
0 g
s
.48289 .0097 m .48267 .00942 L p .48268 .00993 L .11 .256 .799 r
F P
0 g
s
.48268 .00993 m .48277 .0142 L .48289 .0097 L .11 .256 .799 r
F
.4834 .00798 m .48517 .00762 L .48333 .00927 L p .115 .115 .139 r
F P
0 g
s
.48517 .00762 m .4846 .00809 L .48327 .0093 L p .507 .86 .6 r
F P
0 g
s
.48267 .00942 m .48289 .0097 L p .4827 .0096 L .857 .376 .09 r
F P
0 g
s
.48317 .00941 m .48274 .00979 L p .48267 .00942 L .115 .115 .139 r
F P
0 g
s
.48274 .00979 m .48226 .01022 L .48267 .00942 L p .115 .115 .139 r
F P
0 g
s
.4827 .0096 m .48244 .00945 L .48267 .00942 L .857 .376 .09 r
F
.48289 .0097 m .48381 .01088 L p .48276 .00978 L F P
0 g
s
.48276 .00978 m .48244 .00945 L .48289 .0097 L .857 .376 .09 r
F
.48327 .0093 m .48271 .00981 L p .48317 .00941 L .507 .86 .6 r
F P
0 g
s
.48271 .00981 m .48226 .01022 L .48274 .00979 L p .507 .86 .6 r
F P
0 g
s
.48274 .00979 m .48317 .00941 L p .48271 .00981 L .507 .86 .6 r
F P
0 g
s
.48317 .00941 m .48517 .00762 L p .48327 .0093 L .507 .86 .6 r
F P
0 g
s
.48304 .0236 m .48316 .02358 L p .48301 .02324 L .425 .573 1 r
F P
0 g
s
.48301 .02324 m .48252 .02282 L .48326 .0238 L p 1 .738 .314 r
F P
0 g
s
.48293 .02203 m .48252 .02282 L .48301 .02324 L p 1 1 .916 r
F P
0 g
s
.48283 .01756 m .48252 .02282 L .48293 .02203 L p .939 1 .63 r
F P
0 g
s
.48252 .02282 m .48226 .01022 L .48377 .02606 L closepath p .066 .066 .488 r
F P
0 g
s
.48353 .01351 m .48377 .02606 L .48226 .01022 L closepath p .066 .066 .488 r
F P
0 g
s
.4819 .01013 m .48118 .00813 L p .48255 .00957 L .857 .376 .09 r
F P
0 g
s
.48381 .01088 m .48329 .01068 L p .48255 .00957 L .857 .376 .09 r
F P
0 g
s
.48353 .01351 m .48226 .01022 L .48271 .00981 L p .274 .368 .855 r
F P
0 g
s
.48271 .00981 m .48288 .01006 L .48353 .01351 L .274 .368 .855 r
F
.48386 .01183 m .48353 .01351 L p .48281 .00972 L F P
0 g
s
.48329 .01068 m .4819 .01013 L p .48255 .00957 L .857 .376 .09 r
F P
0 g
s
.48271 .00981 m .48281 .00972 L p .48288 .01006 L .274 .368 .855 r
F P
0 g
s
.48604 .01495 m .48353 .01351 L .48409 .01233 L p .608 .793 1 r
F P
0 g
s
.48454 .01327 m .48436 .01268 L p .4861 .01498 L .972 .549 .113 r
F P
0 g
s
.48436 .01268 m .48381 .01088 L .4861 .01498 L p .972 .549 .113 r
F P
0 g
s
.4852 .01001 m .48681 .01538 L .4861 .01498 L p .608 .793 1 r
F P
0 g
s
.4861 .01498 m .48604 .01495 L p .48603 .01489 L .608 .793 1 r
F P
0 g
s
.48603 .01489 m .4852 .01001 L .4861 .01498 L .608 .793 1 r
F
.48386 .01183 m .48436 .01268 L .48414 .01239 L .967 .546 .109 r
F
.48409 .01233 m .4852 .01001 L p .48604 .01495 L .608 .793 1 r
F P
0 g
s
.48381 .01088 m .48436 .01268 L p .48386 .01183 L .967 .546 .109 r
F P
0 g
s
.48436 .01268 m .48447 .01306 L p .48414 .01239 L .967 .546 .109 r
F P
0 g
s
.48329 .01068 m .48333 .0107 L p .48386 .01183 L .967 .546 .109 r
F P
0 g
s
.48333 .0107 m .48381 .01088 L p .48386 .01183 L .967 .546 .109 r
F P
0 g
s
.4846 .00809 m .4852 .01001 L .48409 .01233 L p .2 .29 .798 r
F P
0 g
s
.48409 .01233 m .48353 .01351 L .48386 .01183 L p .2 .29 .798 r
F P
0 g
s
.48386 .01183 m .4846 .00809 L p .48409 .01233 L .2 .29 .798 r
F P
0 g
s
.48281 .00972 m .4846 .00809 L .48386 .01183 L p .274 .368 .855 r
F P
0 g
s
.4819 .01013 m .48329 .01068 L p .48454 .01327 L .967 .546 .109 r
F P
0 g
s
.48447 .01306 m .48454 .01327 L p .48329 .01068 L .967 .546 .109 r
F P
0 g
s
.48454 .01327 m .48569 .01705 L .48217 .01064 L p .967 .546 .109 r
F P
0 g
s
.48217 .01064 m .4819 .01013 L p .48454 .01327 L .967 .546 .109 r
F P
0 g
s
.4861 .01498 m .48633 .01538 L p .48454 .01327 L .972 .549 .113 r
F P
0 g
s
.48633 .01538 m .48786 .01812 L p .4856 .01482 L .972 .549 .113 r
F P
0 g
s
.4856 .01482 m .48454 .01327 L .48633 .01538 L .972 .549 .113 r
F
.48786 .01812 m .48569 .01705 L .48454 .01327 L p F P
0 g
s
.48377 .02606 m .48353 .01351 L .48626 .02484 L p .262 .642 1 r
F P
0 g
s
.48626 .02484 m .487 .02791 L .48631 .02751 L p .262 .642 1 r
F P
0 g
s
.48473 .02651 m .48377 .02606 L .48631 .02751 L p .461 .03 .03 r
F P
0 g
s
.48421 .02618 m .48442 .02672 L .48434 .02626 L .373 .487 .94 r
F
.48436 .0263 m .48377 .02606 L .48473 .02651 L p .394 .018 .018 r
F P
0 g
s
.48631 .02751 m .48377 .02606 L p .48626 .02484 L .262 .642 1 r
F P
0 g
s
.48786 .01812 m .48619 .02046 L p .48631 .02751 L .09 .09 .652 r
F P
0 g
s
.48631 .02751 m .48804 .0306 L .48801 .0288 L p .09 .09 .652 r
F P
0 g
s
.48801 .0288 m .48786 .01812 L p .48631 .02751 L .09 .09 .652 r
F P
0 g
s
.48434 .02626 m .48442 .02672 L .48465 .02696 L .373 .487 .94 r
F
.48244 .00945 m .48118 .00813 L .48267 .00942 L .857 .376 .09 r
F
.48332 .00934 m .4834 .00798 L .48267 .00942 L p .939 1 .63 r
F P
0 g
s
.48333 .00927 m .48317 .00941 L p .4834 .00798 L .115 .115 .139 r
F P
0 g
s
.48267 .00942 m .4834 .00798 L p .48317 .00941 L .115 .115 .139 r
F P
0 g
s
.48118 .00813 m .48077 .00699 L .482 .00857 L p .857 .376 .09 r
F P
0 g
s
.48309 .02529 m .48225 .02689 L .4831 .02743 L p .373 .487 .94 r
F P
0 g
s
.48023 .02396 m .48225 .02689 L .48309 .02529 L p .426 .553 .986 r
F P
0 g
s
.48023 .02396 m .48106 .01963 L .41459 .02396 L closepath p .848 1 1 r
F P
0 g
s
.41439 .01963 m .41459 .02396 L .48106 .01963 L closepath p .848 1 1 r
F P
0 g
s
.41459 .02396 m .41439 .01963 L .4142 .02427 L closepath p 1 1 1 r
F P
0 g
s
.40665 .02459 m .40815 .0212 L .40571 .02396 L closepath p .573 .756 1 r
F P
0 g
s
.40549 .01963 m .40571 .02396 L .40815 .0212 L closepath p .59 .782 1 r
F P
0 g
s
.40571 .02396 m .40549 .01963 L .36826 .02396 L closepath p .848 1 1 r
F P
0 g
s
.36795 .01963 m .36826 .02396 L .40549 .01963 L closepath p .848 1 1 r
F P
0 g
s
.36826 .02396 m .36795 .01963 L .36782 .02427 L closepath p 1 1 1 r
F P
0 g
s
.35421 .02459 m .35557 .02119 L .35321 .02396 L closepath p .578 .767 1 r
F P
0 g
s
.35287 .01963 m .35321 .02396 L .35557 .02119 L closepath p .589 .785 1 r
F P
0 g
s
.35321 .02396 m .35287 .01963 L .31575 .02396 L closepath p .848 1 1 r
F P
0 g
s
.31532 .01963 m .31575 .02396 L .35287 .01963 L closepath p .848 1 1 r
F P
0 g
s
.31575 .02396 m .31532 .01963 L .31535 .02427 L closepath p 1 1 1 r
F P
0 g
s
.29538 .02459 m .29662 .0212 L .29445 .02396 L closepath p .555 .749 1 r
F P
0 g
s
.29397 .01963 m .29445 .02396 L .29662 .0212 L closepath p .574 .776 1 r
F P
0 g
s
.29445 .02396 m .29397 .01963 L .25675 .02396 L closepath p .848 1 1 r
F P
0 g
s
.25618 .01963 m .25675 .02396 L .29397 .01963 L closepath p .848 1 1 r
F P
0 g
s
.25675 .02396 m .25618 .01963 L .25635 .02427 L closepath p 1 1 1 r
F P
0 g
s
.25042 .02455 m .25146 .0211 L .24918 .02396 L closepath p .617 .821 1 r
F P
0 g
s
.2486 .01963 m .24918 .02396 L .25146 .0211 L closepath p .609 .809 1 r
F P
0 g
s
.24918 .02396 m .2486 .01963 L .21495 .02396 L closepath p .848 1 1 r
F P
0 g
s
.21269 .01963 m .21495 .02396 L .2486 .01963 L closepath p .848 1 1 r
F P
0 g
s
.21495 .02396 m .21269 .01963 L .21478 .02442 L closepath p 1 1 .889 r
F P
0 g
s
.19451 .02459 m .19544 .02116 L .19334 .02396 L closepath p .588 .796 1 r
F P
0 g
s
.19262 .01963 m .19334 .02396 L .19544 .02116 L closepath p .587 .795 1 r
F P
0 g
s
.19334 .02396 m .19262 .01963 L .12786 .02396 L closepath p .848 1 1 r
F P
0 g
s
.127 .01963 m .12786 .02396 L .19262 .01963 L closepath p .848 1 1 r
F P
0 g
s
.12786 .02396 m .127 .01963 L .12722 .02449 L closepath p 1 1 1 r
F P
0 g
s
.07272 .02474 m .07335 .02129 L .07115 .02396 L closepath p .589 .808 1 r
F P
0 g
s
.07015 .01963 m .07115 .02396 L .07335 .02129 L closepath p .583 .799 1 r
F P
0 g
s
.07115 .02396 m .07015 .01963 L .04078 .02396 L closepath p .848 1 1 r
F P
0 g
s
.03971 .01963 m .04078 .02396 L .07015 .01963 L closepath p .848 1 1 r
F P
0 g
s
.04078 .02396 m .03971 .01963 L .04013 .02446 L closepath p 1 1 1 r
F P
0 g
s
.04025 .02679 m .04078 .02396 L .07272 .02474 L closepath p .955 .786 .906 r
F P
0 g
s
.07115 .02396 m .07272 .02474 L .04078 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.13343 .02815 m .13195 .02813 L .12786 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.12729 .02685 m .12786 .02396 L .13195 .02813 L closepath p .955 .786 .906 r
F P
0 g
s
.12786 .02396 m .19334 .02396 L .1609 .02567 L closepath p .955 .786 .906 r
F P
0 g
s
.19451 .02459 m .1609 .02567 L .19334 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.22158 .02777 m .22004 .02776 L .21495 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.21527 .02661 m .21495 .02396 L .22004 .02776 L closepath p .955 .786 .906 r
F P
0 g
s
.21495 .02396 m .24918 .02396 L .23116 .02583 L closepath p .955 .786 .906 r
F P
0 g
s
.25042 .02455 m .23116 .02583 L .24918 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.19128 .02769 m .12786 .02396 L .19395 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.24213 .02922 m .21495 .02396 L .24738 .02769 L closepath p .955 .786 .906 r
F P
0 g
s
.23914 .03379 m .21495 .02396 L .24213 .02922 L closepath p .955 .786 .906 r
F P
0 g
s
.12786 .02396 m .19128 .02769 L .15589 .02615 L closepath p .955 .786 .906 r
F P
0 g
s
.21495 .02396 m .22704 .02887 L .22158 .02777 L closepath p .955 .786 .906 r
F P
0 g
s
.12786 .02396 m .15589 .02615 L .13065 .02605 L closepath p .955 .786 .906 r
F P
0 g
s
.32067 .02785 m .31646 .02695 L .35321 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.31575 .02396 m .35321 .02396 L .31646 .02695 L closepath p .955 .786 .906 r
F P
0 g
s
.35307 .02567 m .35321 .02396 L .35421 .02459 L closepath p .955 .786 .906 r
F P
0 g
s
.34821 .02784 m .35321 .02396 L .35294 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.36782 .02427 m .36826 .02396 L .40665 .02459 L closepath p .955 .786 .906 r
F P
0 g
s
.40571 .02396 m .40665 .02459 L .36826 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.40571 .02396 m .36826 .02396 L .40665 .02459 L closepath p .955 .786 .906 r
F P
0 g
s
.35321 .02396 m .34137 .02865 L .32067 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.32953 .03334 m .35321 .02396 L .34348 .02897 L closepath p .955 .786 .906 r
F P
0 g
s
.34821 .02784 m .34348 .02897 L .35321 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.25635 .02427 m .25675 .02396 L .29538 .02459 L closepath p .955 .786 .906 r
F P
0 g
s
.29445 .02396 m .29538 .02459 L .25675 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.29445 .02396 m .25675 .02396 L .29538 .02459 L closepath p .955 .786 .906 r
F P
0 g
s
.48023 .02396 m .4138 .02457 L .41459 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.48023 .02396 m .48225 .02689 L .48447 .03417 L closepath p .955 .786 .906 r
F P
0 g
s
.48023 .02396 m .48447 .03417 L .47098 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.4138 .02457 m .48023 .02396 L .45787 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.47098 .02785 m .45787 .02785 L .48023 .02396 L closepath p .955 .786 .906 r
F P
0 g
s
.944 .02456 m .9467 .0211 L .94274 .02397 L closepath p .685 .841 1 r
F P
0 g
s
.94377 .01964 m .94274 .02397 L .9467 .0211 L closepath p .681 .831 1 r
F P
0 g
s
.94274 .02397 m .94377 .01964 L .9085 .02397 L closepath p .848 1 1 r
F P
0 g
s
.90786 .01964 m .9085 .02397 L .94377 .01964 L closepath p .848 1 1 r
F P
0 g
s
.9085 .02397 m .90786 .01964 L .90836 .02442 L closepath p 1 1 .72 r
F P
0 g
s
.88803 .0246 m .89059 .02113 L .88665 .02397 L closepath p .684 .844 1 r
F P
0 g
s
.88755 .01964 m .88665 .02397 L .89059 .02113 L closepath p .679 .832 1 r
F P
0 g
s
.88665 .02397 m .88755 .01964 L .86086 .02397 L closepath p .848 1 1 r
F P
0 g
s
.8617 .01964 m .86086 .02397 L .88755 .01964 L closepath p .848 1 1 r
F P
0 g
s
.86086 .02397 m .8617 .01964 L .86055 .02442 L closepath p 1 1 .842 r
F P
0 g
s
.78978 .0246 m .79218 .02119 L .78876 .02397 L closepath p .636 .789 1 r
F P
0 g
s
.78943 .01964 m .78876 .02397 L .79218 .02119 L closepath p .644 .804 1 r
F P
0 g
s
.78876 .02397 m .78943 .01964 L .75129 .02397 L closepath p .848 1 1 r
F P
0 g
s
.75188 .01964 m .75129 .02397 L .78943 .01964 L closepath p .848 1 1 r
F P
0 g
s
.75129 .02397 m .75188 .01964 L .7509 .02427 L closepath p 1 1 1 r
F P
0 g
s
.72712 .0246 m .72942 .02123 L .72629 .02397 L closepath p .599 .746 1 r
F P
0 g
s
.72681 .01964 m .72629 .02397 L .72942 .02123 L closepath p .621 .787 1 r
F P
0 g
s
.72629 .02397 m .72681 .01964 L .68866 .02397 L closepath p .848 1 1 r
F P
0 g
s
.6891 .01964 m .68866 .02397 L .72681 .01964 L closepath p .848 1 1 r
F P
0 g
s
.68866 .02397 m .6891 .01964 L .68823 .02427 L closepath p 1 1 1 r
F P
0 g
s
.68125 .02456 m .68334 .02111 L .68006 .02397 L closepath p .656 .827 1 r
F P
0 g
s
.68048 .01964 m .68006 .02397 L .68334 .02111 L closepath p .652 .82 1 r
F P
0 g
s
.68006 .02397 m .68048 .01964 L .64598 .02397 L closepath p .848 1 1 r
F P
0 g
s
.64487 .01964 m .64598 .02397 L .68048 .01964 L closepath p .848 1 1 r
F P
0 g
s
.64598 .02397 m .64487 .01964 L .64552 .02488 L closepath p 1 1 .909 r
F P
0 g
s
.6255 .0246 m .62743 .02113 L .62413 .02397 L closepath p .661 .838 1 r
F P
0 g
s
.62442 .01964 m .62413 .02397 L .62743 .02113 L closepath p .655 .826 1 r
F P
0 g
s
.62413 .02397 m .62442 .01964 L .56213 .02397 L closepath p .848 1 1 r
F P
0 g
s
.56228 .01964 m .56213 .02397 L .62442 .01964 L closepath p .848 1 1 r
F P
0 g
s
.56213 .02397 m .56228 .01964 L .56158 .02442 L closepath p 1 1 1 r
F P
0 g
s
.51392 .02467 m .51564 .02125 L .51267 .02397 L closepath p .618 .797 1 r
F P
0 g
s
.5127 .01964 m .51267 .02397 L .51564 .02125 L closepath p .62 .802 1 r
F P
0 g
s
.5128 .02747 m .50818 .0277 L .49934 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.49934 .02397 m .51267 .02397 L .50607 .02572 L closepath p .955 .786 .906 r
F P
0 g
s
.51392 .02467 m .50607 .02572 L .51267 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.56712 .0277 m .5627 .02688 L .62413 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.56213 .02397 m .62413 .02397 L .5627 .02688 L closepath p .955 .786 .906 r
F P
0 g
s
.62413 .02397 m .6255 .0246 L .62186 .02583 L closepath p .955 .786 .906 r
F P
0 g
s
.65249 .02778 m .65097 .02777 L .64598 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.64619 .02661 m .64598 .02397 L .65097 .02777 L closepath p .955 .786 .906 r
F P
0 g
s
.64598 .02397 m .68006 .02397 L .66214 .02583 L closepath p .955 .786 .906 r
F P
0 g
s
.68125 .02456 m .66214 .02583 L .68006 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.62413 .02397 m .59889 .02642 L .56712 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.67314 .02922 m .64598 .02397 L .67829 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.67032 .03379 m .64598 .02397 L .67314 .02922 L closepath p .955 .786 .906 r
F P
0 g
s
.64598 .02397 m .65815 .02888 L .65249 .02778 L closepath p .955 .786 .906 r
F P
0 g
s
.62413 .02397 m .61959 .0277 L .60492 .02854 L closepath p .955 .786 .906 r
F P
0 g
s
.59884 .02724 m .62413 .02397 L .60492 .02854 L closepath p .955 .786 .906 r
F P
0 g
s
.57364 .02888 m .62413 .02397 L .59158 .028 L closepath p .955 .786 .906 r
F P
0 g
s
.59884 .02724 m .59158 .028 L .62413 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.86213 .02732 m .86075 .02659 L .86086 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.86025 .02488 m .86086 .02397 L .86075 .02659 L closepath p .955 .786 .906 r
F P
0 g
s
.86086 .02397 m .87376 .02397 L .88502 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.88767 .02739 m .88502 .0277 L .87376 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.87376 .02397 m .88665 .02397 L .88071 .02568 L closepath p .955 .786 .906 r
F P
0 g
s
.88803 .0246 m .88071 .02568 L .88665 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.90836 .02442 m .9085 .02397 L .944 .02456 L closepath p .955 .786 .906 r
F P
0 g
s
.94274 .02397 m .944 .02456 L .9085 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.86086 .02397 m .86375 .02606 L .86213 .02732 L closepath p .955 .786 .906 r
F P
0 g
s
.88226 .0277 m .86086 .02397 L .88502 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.86665 .02815 m .86086 .02397 L .87731 .02859 L closepath p .955 .786 .906 r
F P
0 g
s
.88226 .0277 m .87731 .02859 L .86086 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.69029 .02762 m .68844 .02659 L .72629 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.70747 .02397 m .72629 .02397 L .68844 .02659 L closepath p .955 .786 .906 r
F P
0 g
s
.68844 .02659 m .6878 .02458 L .70747 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.68866 .02397 m .70747 .02397 L .6878 .02458 L closepath p .955 .786 .906 r
F P
0 g
s
.72609 .02568 m .72629 .02397 L .72712 .0246 L closepath p .955 .786 .906 r
F P
0 g
s
.72122 .02785 m .72629 .02397 L .72589 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.70268 .03334 m .70009 .02926 L .72629 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.70829 .02579 m .72629 .02397 L .70009 .02926 L closepath p .955 .786 .906 r
F P
0 g
s
.70268 .03334 m .72629 .02397 L .71656 .02897 L closepath p .955 .786 .906 r
F P
0 g
s
.72122 .02785 m .71656 .02897 L .72629 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.75602 .02785 m .75192 .02703 L .75129 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.75051 .02458 m .75129 .02397 L .75192 .02703 L closepath p .955 .786 .906 r
F P
0 g
s
.75051 .02458 m .75192 .02703 L .75129 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.75129 .02397 m .77003 .02397 L .78395 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.78861 .02739 m .78395 .02785 L .77003 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.77003 .02397 m .78876 .02397 L .77932 .02568 L closepath p .955 .786 .906 r
F P
0 g
s
.78978 .0246 m .77932 .02568 L .78876 .02397 L closepath p .955 .786 .906 r
F P
0 g
s
.77929 .02897 m .75129 .02397 L .78395 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.75129 .02397 m .76529 .02647 L .75602 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.0972 .03129 m .0975 .0288 L .09884 .02416 L closepath p .955 .786 .906 r
F P
0 g
s
.0984 .02523 m .09884 .02416 L .0975 .0288 L closepath p .955 .786 .906 r
F P
0 g
s
.54219 .02854 m .5444 .02549 L .53961 .0242 L closepath p .414 .521 .96 r
F P
0 g
s
.54118 .02044 m .53961 .0242 L .5444 .02549 L closepath p .404 .527 .968 r
F P
0 g
s
.54493 .03402 m .54584 .03585 L .53961 .0242 L closepath p .955 .786 .906 r
F P
0 g
s
.53961 .0242 m .54219 .02854 L .54493 .03402 L closepath p .955 .786 .906 r
F P
0 g
s
.36583 .02049 m .36782 .02427 L .36795 .01963 L closepath p .982 1 1 r
F P
0 g
s
.36782 .02427 m .36583 .02049 L .36738 .02457 L closepath p 1 1 1 r
F P
0 g
s
.31324 .02051 m .31535 .02427 L .31532 .01963 L closepath p .988 1 1 r
F P
0 g
s
.31535 .02427 m .31324 .02051 L .31496 .02457 L closepath p 1 1 1 r
F P
0 g
s
.25409 .02051 m .25635 .02427 L .25618 .01963 L closepath p .989 1 1 r
F P
0 g
s
.25635 .02427 m .25409 .02051 L .25595 .02457 L closepath p 1 1 1 r
F P
0 g
s
.36738 .02457 m .36782 .02427 L .40696 .02648 L closepath p .955 .786 .906 r
F P
0 g
s
.40665 .02459 m .40696 .02648 L .36782 .02427 L closepath p .955 .786 .906 r
F P
0 g
s
.25595 .02457 m .25635 .02427 L .29567 .02647 L closepath p .955 .786 .906 r
F P
0 g
s
.29538 .02459 m .29567 .02647 L .25635 .02427 L closepath p .955 .786 .906 r
F P
0 g
s
.41231 .02051 m .4142 .02427 L .41439 .01963 L closepath p .987 1 1 r
F P
0 g
s
.4142 .02427 m .41231 .02051 L .4138 .02457 L closepath p 1 1 1 r
F P
0 g
s
.74982 .02052 m .7509 .02427 L .75188 .01964 L closepath p .979 1 1 r
F P
0 g
s
.7509 .02427 m .74982 .02052 L .75051 .02458 L closepath p 1 1 1 r
F P
0 g
s
.687 .02049 m .68823 .02427 L .6891 .01964 L closepath p .975 1 1 r
F P
0 g
s
.68823 .02427 m .687 .02049 L .6878 .02458 L closepath p 1 1 1 r
F P
0 g
s
.53563 .0304 m .53581 .02827 L .53683 .02428 L closepath p .955 .786 .906 r
F P
0 g
s
.53648 .0252 m .53683 .02428 L .53581 .02827 L closepath p .955 .786 .906 r
F P
0 g
s
.21193 .02145 m .21478 .02442 L .21269 .01963 L closepath p 1 1 .915 r
F P
0 g
s
.21478 .02442 m .21193 .02145 L .21462 .02488 L closepath p 1 1 .894 r
F P
0 g
s
.90719 .02146 m .90836 .02442 L .90786 .01964 L closepath p 1 1 .779 r
F P
0 g
s
.90836 .02442 m .90719 .02146 L .90821 .02488 L closepath p 1 1 .736 r
F P
0 g
s
.85961 .02074 m .86055 .02442 L .8617 .01964 L closepath p 1 1 1 r
F P
0 g
s
.86055 .02442 m .85961 .02074 L .86025 .02488 L closepath p 1 1 .953 r
F P
0 g
s
.55995 .02061 m .56158 .02442 L .56228 .01964 L closepath p .984 1 1 r
F P
0 g
s
.56158 .02442 m .55995 .02061 L .56103 .02488 L closepath p 1 1 1 r
F P
0 g
s
.90821 .02488 m .90836 .02442 L .94447 .02633 L closepath p .955 .786 .906 r
F P
0 g
s
.944 .02456 m .94447 .02633 L .90836 .02442 L closepath p .955 .786 .906 r
F P
0 g
s
.03724 .02061 m .04013 .02446 L .03971 .01963 L closepath p .989 1 1 r
F P
0 g
s
.04013 .02446 m .03724 .02061 L .03948 .02495 L closepath p 1 1 1 r
F P
0 g
s
.12452 .02065 m .12722 .02449 L .127 .01963 L closepath p .992 1 1 r
F P
0 g
s
.12722 .02449 m .12452 .02065 L .12657 .02503 L closepath p 1 1 1 r
F P
0 g
s
.2508 .02632 m .25252 .02426 L .25042 .02455 L closepath p .161 .163 .699 r
F P
0 g
s
.25146 .0211 m .25042 .02455 L .25252 .02426 L closepath p .16 .219 .757 r
F P
0 g
s
.23116 .02583 m .25042 .02455 L .24738 .02769 L closepath p .955 .786 .906 r
F P
0 g
s
.2508 .02632 m .24738 .02769 L .25042 .02455 L closepath p .955 .786 .906 r
F P
0 g
s
.94447 .02633 m .94792 .02427 L .944 .02456 L closepath p .423 .433 .863 r
F P
0 g
s
.9467 .0211 m .944 .02456 L .94792 .02427 L closepath p .424 .47 .903 r
F P
0 g
s
.68167 .02633 m .68447 .02427 L .68125 .02456 L closepath p .319 .348 .815 r
F P
0 g
s
.68334 .02111 m .68125 .02456 L .68447 .02427 L closepath p .322 .392 .862 r
F P
0 g
s
.66214 .02583 m .68125 .02456 L .67829 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.68167 .02633 m .67829 .0277 L .68125 .02456 L closepath p .955 .786 .906 r
F P
0 g
s
.3647 .02251 m .36738 .02457 L .36583 .02049 L closepath p 1 1 .944 r
F P
0 g
s
.36738 .02457 m .3647 .02251 L .36844 .02609 L closepath p .884 .423 .092 r
F P
0 g
s
.31215 .02251 m .31496 .02457 L .31324 .02051 L closepath p 1 1 .945 r
F P
0 g
s
.31496 .02457 m .31215 .02251 L .31566 .0258 L closepath p .985 .559 .129 r
F P
0 g
s
.25299 .02251 m .25595 .02457 L .25409 .02051 L closepath p 1 1 .95 r
F P
0 g
s
.25595 .02457 m .25299 .02251 L .25676 .0268 L closepath p 1 .754 .378 r
F P
0 g
s
.37376 .02784 m .37103 .02779 L .36738 .02457 L closepath p .955 .786 .906 r
F P
0 g
s
.36949 .02762 m .36738 .02457 L .37103 .02779 L closepath p .955 .786 .906 r
F P
0 g
s
.37057 .02621 m .36738 .02457 L .40538 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.40696 .02648 m .40538 .02739 L .36738 .02457 L closepath p .955 .786 .906 r
F P
0 g
s
.26599 .02815 m .25929 .02769 L .25595 .02457 L closepath p .955 .786 .906 r
F P
0 g
s
.25676 .0268 m .25595 .02457 L .25929 .02769 L closepath p .955 .786 .906 r
F P
0 g
s
.25595 .02457 m .29567 .02647 L .26599 .02815 L closepath p .955 .786 .906 r
F P
0 g
s
.41124 .02251 m .4138 .02457 L .41231 .02051 L closepath p 1 1 .933 r
F P
0 g
s
.4138 .02457 m .41124 .02251 L .41457 .02684 L closepath p 1 .734 .322 r
F P
0 g
s
.42325 .02823 m .41694 .02785 L .4138 .02457 L closepath p .955 .786 .906 r
F P
0 g
s
.41457 .02684 m .4138 .02457 L .41694 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.4138 .02457 m .45787 .02785 L .44103 .0289 L closepath p .955 .786 .906 r
F P
0 g
s
.4138 .02457 m .44103 .0289 L .42325 .02823 L closepath p .955 .786 .906 r
F P
0 g
s
.74879 .02251 m .75051 .02458 L .74982 .02052 L closepath p 1 1 .877 r
F P
0 g
s
.75051 .02458 m .74879 .02251 L .75192 .02703 L closepath p .772 .351 .045 r
F P
0 g
s
.68592 .02251 m .6878 .02458 L .687 .02049 L closepath p 1 1 .896 r
F P
0 g
s
.6878 .02458 m .68592 .02251 L .68844 .02659 L closepath p 1 .664 .182 r
F P
0 g
s
.35452 .02647 m .35651 .02441 L .35421 .02459 L closepath p .168 .174 .697 r
F P
0 g
s
.35557 .02119 m .35421 .02459 L .35651 .02441 L closepath p .167 .228 .755 r
F P
0 g
s
.29567 .02647 m .29751 .02441 L .29538 .02459 L closepath p .16 .16 .666 r
F P
0 g
s
.29662 .0212 m .29538 .02459 L .29751 .02441 L closepath p .16 .189 .727 r
F P
0 g
s
.19486 .02647 m .19645 .02442 L .19451 .02459 L closepath p .152 .152 .656 r
F P
0 g
s
.19544 .02116 m .19451 .02459 L .19645 .02442 L closepath p .151 .165 .716 r
F P
0 g
s
.1609 .02567 m .19451 .02459 L .19395 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.19486 .02647 m .19395 .02739 L .19451 .02459 L closepath p .955 .786 .906 r
F P
0 g
s
.35421 .02459 m .35452 .02647 L .35307 .02567 L closepath p .955 .786 .906 r
F P
0 g
s
.40696 .02648 m .40907 .02442 L .40665 .02459 L closepath p .172 .193 .708 r
F P
0 g
s
.40815 .0212 m .40665 .02459 L .40907 .02442 L closepath p .171 .246 .766 r
F P
0 g
s
.88854 .02648 m .89185 .02442 L .88803 .0246 L closepath p .406 .422 .859 r
F P
0 g
s
.89059 .02113 m .88803 .0246 L .89185 .02442 L closepath p .405 .456 .897 r
F P
0 g
s
.79015 .02648 m .79322 .02442 L .78978 .0246 L closepath p .351 .359 .813 r
F P
0 g
s
.79218 .02119 m .78978 .0246 L .79322 .02442 L closepath p .352 .401 .86 r
F P
0 g
s
.72743 .02648 m .73034 .02442 L .72712 .0246 L closepath p .313 .319 .784 r
F P
0 g
s
.72942 .02123 m .72712 .0246 L .73034 .02442 L closepath p .315 .364 .834 r
F P
0 g
s
.62597 .02648 m .62863 .02442 L .6255 .0246 L closepath p .299 .339 .813 r
F P
0 g
s
.62743 .02113 m .6255 .0246 L .62863 .02442 L closepath p .3 .379 .857 r
F P
0 g
s
.62348 .02661 m .62186 .02583 L .6255 .0246 L closepath p .955 .786 .906 r
F P
0 g
s
.6255 .0246 m .62597 .02648 L .62348 .02661 L closepath p .955 .786 .906 r
F P
0 g
s
.88071 .02568 m .88803 .0246 L .88767 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.88854 .02648 m .88767 .02739 L .88803 .0246 L closepath p .955 .786 .906 r
F P
0 g
s
.72712 .0246 m .72743 .02648 L .72609 .02568 L closepath p .955 .786 .906 r
F P
0 g
s
.77932 .02568 m .78978 .0246 L .78861 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.79015 .02648 m .78861 .02739 L .78978 .0246 L closepath p .955 .786 .906 r
F P
0 g
s
.51433 .02678 m .51672 .02473 L .51392 .02467 L closepath p .226 .265 .76 r
F P
0 g
s
.51564 .02125 m .51392 .02467 L .51672 .02473 L closepath p .222 .306 .807 r
F P
0 g
s
.50607 .02572 m .51392 .02467 L .5128 .02747 L closepath p .955 .786 .906 r
F P
0 g
s
.51433 .02678 m .5128 .02747 L .51392 .02467 L closepath p .955 .786 .906 r
F P
0 g
s
.07318 .02708 m .07447 .02502 L .07272 .02474 L closepath p .137 .137 .612 r
F P
0 g
s
.07335 .02129 m .07272 .02474 L .07447 .02502 L closepath p .133 .133 .657 r
F P
0 g
s
.07272 .02474 m .07318 .02708 L .04025 .02679 L closepath p .955 .786 .906 r
F P
0 g
s
.21157 .02249 m .21462 .02488 L .21193 .02145 L closepath p 1 1 .887 r
F P
0 g
s
.21462 .02488 m .21157 .02249 L .21523 .02672 L closepath p 1 .777 .407 r
F P
0 g
s
.90687 .0225 m .90821 .02488 L .90719 .02146 L closepath p 1 1 .721 r
F P
0 g
s
.90821 .02488 m .90687 .0225 L .9089 .02673 L closepath p .809 .464 .041 r
F P
0 g
s
.8588 .02282 m .86025 .02488 L .85961 .02074 L closepath p 1 1 .783 r
F P
0 g
s
.86025 .02488 m .8588 .02282 L .86119 .0261 L closepath p .011 .011 .471 r
F P
0 g
s
.64355 .02225 m .64552 .02488 L .64487 .01964 L closepath p 1 1 .909 r
F P
0 g
s
.64552 .02488 m .64355 .02225 L .64613 .02673 L closepath p .975 .632 .136 r
F P
0 g
s
.55884 .02282 m .56103 .02488 L .55995 .02061 L closepath p 1 1 .904 r
F P
0 g
s
.56103 .02488 m .55884 .02282 L .5618 .02594 L closepath p .76 .294 .054 r
F P
0 g
s
.91524 .02778 m .91372 .02777 L .90821 .02488 L closepath p .955 .786 .906 r
F P
0 g
s
.9089 .02673 m .90821 .02488 L .91372 .02777 L closepath p .955 .786 .906 r
F P
0 g
s
.90821 .02488 m .92634 .0256 L .91524 .02778 L closepath p .955 .786 .906 r
F P
0 g
s
.81309 .02087 m .81335 .02492 L .8197 .01875 L closepath p .947 1 1 r
F P
0 g
s
.81335 .02492 m .81309 .02087 L .80817 .02811 L closepath p 1 1 1 r
F P
0 g
s
.80817 .02811 m .81335 .02492 L .81586 .0329 L closepath p .955 .786 .906 r
F P
0 g
s
.81755 .03087 m .81586 .0329 L .81335 .02492 L closepath p .955 .786 .906 r
F P
0 g
s
.03599 .02289 m .03948 .02495 L .03724 .02061 L closepath p 1 1 .963 r
F P
0 g
s
.03948 .02495 m .03599 .02289 L .04009 .02692 L closepath p 1 .826 .513 r
F P
0 g
s
.1233 .02296 m .12657 .02503 L .12452 .02065 L closepath p 1 1 .952 r
F P
0 g
s
.12657 .02503 m .1233 .02296 L .12713 .02699 L closepath p 1 .83 .503 r
F P
0 g
s
.85589 .02724 m .85779 .02317 L .85033 .02511 L closepath p .725 .872 1 r
F P
0 g
s
.85194 .02093 m .85033 .02511 L .85779 .02317 L closepath p .725 .872 1 r
F P
0 g
s
.84323 .02921 m .83946 .02707 L .85033 .02511 L closepath p .955 .786 .906 r
F P
0 g
s
.85033 .02511 m .85589 .02724 L .84323 .02921 L closepath p .955 .786 .906 r
F P
0 g
s
.53474 .02181 m .53648 .0252 L .53615 .01932 L closepath p 1 1 .941 r
F P
0 g
s
.53648 .0252 m .53474 .02181 L .53581 .02827 L closepath p 1 1 .738 r
F P
0 g
s
.09561 .02183 m .0984 .02523 L .09722 .01904 L closepath p 1 1 .972 r
F P
0 g
s
.0984 .02523 m .09561 .02183 L .0975 .0288 L closepath p 1 1 .842 r
F P
0 g
s
.92634 .0256 m .94447 .02633 L .91911 .02865 L closepath p .955 .786 .906 r
F P
0 g
s
.91911 .02865 m .91524 .02778 L .92634 .0256 L closepath p .955 .786 .906 r
F P
0 g
s
.82628 .02622 m .82159 .02796 L .83163 .02564 L closepath p .955 .786 .906 r
F P
0 g
s
.82628 .02622 m .83163 .02564 L .82159 .02796 L closepath p .955 .786 .906 r
F P
0 g
s
.82628 .02622 m .82801 .02684 L .83163 .02564 L closepath p .21 .073 .073 r
F P
0 g
s
.83251 .02585 m .83163 .02564 L .82801 .02684 L closepath p .037 .037 .037 r
F P
0 g
s
.82159 .02796 m .82352 .02784 L .82628 .02622 L closepath p .052 .052 .052 r
F P
0 g
s
.82801 .02684 m .82628 .02622 L .82352 .02784 L closepath p .195 .087 .087 r
F P
0 g
s
.83251 .02585 m .82352 .02784 L .82482 .02451 L p .703 1 1 r
F P
0 g
s
.83725 .02667 m .83163 .02564 L .83251 .02585 L p .143 .004 .004 r
F P
0 g
s
.82711 .01804 m .84484 .01932 L .8499 .0071 L p .592 .946 1 r
F P
0 g
s
.8499 .0071 m .85007 .00668 L .84935 .00703 L p .592 .946 1 r
F P
0 g
s
.84935 .00703 m .82711 .01804 L p .8499 .0071 L .592 .946 1 r
F P
0 g
s
.83206 .00538 m .82711 .01804 L .85007 .00668 L closepath p .592 .946 1 r
F P
0 g
s
.85007 .00668 m .84484 .01932 L .85728 .0083 L closepath p .524 .873 1 r
F P
0 g
s
.85007 .00668 m .85051 .00677 L .84107 .00603 L closepath p .072 .005 .005 r
F P
0 g
s
.85051 .00677 m .84236 .00584 L .84107 .00603 L p .04 .607 .817 r
F P
0 g
s
.84107 .00603 m .84107 .00603 L .84107 .00603 L p .04 .607 .817 r
F P
0 g
s
.84107 .00603 m .85051 .00677 L p .84107 .00603 L .04 .607 .817 r
F P
0 g
s
.82482 .02451 m .82842 .01532 L .83251 .02585 L p .703 1 1 r
F P
0 g
s
.83251 .02585 m .83946 .02707 L .83725 .02667 L p .143 .004 .004 r
F P
0 g
s
.35294 .02739 m .35307 .02567 L .35452 .02647 L closepath p .955 .786 .906 r
F P
0 g
s
.72589 .02739 m .72609 .02568 L .72743 .02648 L closepath p .955 .786 .906 r
F P
0 g
s
.70009 .02926 m .69471 .02785 L .70829 .02579 L closepath p .955 .786 .906 r
F P
0 g
s
.69029 .02762 m .70829 .02579 L .69471 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.31276 .02484 m .31566 .0258 L .31215 .02251 L closepath p 1 .842 .513 r
F P
0 g
s
.31566 .0258 m .31276 .02484 L .31636 .02703 L closepath p 1 .647 .287 r
F P
0 g
s
.62186 .02583 m .62509 .02739 L .61959 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.55952 .02501 m .5618 .02594 L .55884 .02282 L closepath p 1 .766 .368 r
F P
0 g
s
.5618 .02594 m .55952 .02501 L .56256 .02699 L closepath p .929 .465 .111 r
F P
0 g
s
.13343 .02815 m .13065 .02605 L .15589 .02615 L closepath p .955 .786 .906 r
F P
0 g
s
.86375 .02606 m .86665 .02815 L .8652 .02783 L closepath p .955 .786 .906 r
F P
0 g
s
.8652 .02783 m .86213 .02732 L .86375 .02606 L closepath p .955 .786 .906 r
F P
0 g
s
.36575 .02575 m .36844 .02609 L .3647 .02251 L closepath p 1 .803 .465 r
F P
0 g
s
.36844 .02609 m .36575 .02575 L .36949 .02762 L closepath p 1 .59 .24 r
F P
0 g
s
.85779 .02317 m .85589 .02724 L .86048 .0269 L p .542 .653 1 r
F P
0 g
s
.86001 .02589 m .86004 .02589 L p .85981 .02547 L .968 .604 .113 r
F P
0 g
s
.86004 .02589 m .86119 .0261 L p .85981 .02547 L .968 .604 .113 r
F P
0 g
s
.86119 .0261 m .8588 .02282 L .85981 .02547 L p .968 .604 .113 r
F P
0 g
s
.88755 .01964 m .89342 .007 L .8617 .01964 L closepath p .623 .977 1 r
F P
0 g
s
.86718 .007 m .8617 .01964 L .89342 .007 L closepath p .623 .977 1 r
F P
0 g
s
.8617 .01964 m .86718 .007 L .85961 .02074 L closepath p .771 1 1 r
F P
0 g
s
.86506 .00812 m .85961 .02074 L .86718 .007 L closepath p .771 1 1 r
F P
0 g
s
.85961 .02074 m .86506 .00812 L .8588 .02282 L closepath p .008 .008 .008 r
F P
0 g
s
.86424 .01022 m .8588 .02282 L .86506 .00812 L closepath p .008 .008 .008 r
F P
0 g
s
.86476 .01609 m .8632 .01768 L p .86046 .02597 L .182 .457 .95 r
F P
0 g
s
.8588 .02282 m .86424 .01022 L .86054 .02381 L p .132 .132 .697 r
F P
0 g
s
.85981 .02547 m .8588 .02282 L p .86054 .02381 L .132 .132 .697 r
F P
0 g
s
.86046 .02597 m .8632 .01768 L .85994 .02517 L .182 .457 .95 r
F
.86502 .0155 m .86291 .02661 L p .86045 .026 L .369 .693 1 r
F P
0 g
s
.85981 .02547 m .86039 .02613 L p .85994 .02517 L .182 .457 .95 r
F P
0 g
s
.86119 .0261 m .85997 .02588 L .86213 .02732 L closepath p .819 .34 .076 r
F P
0 g
s
.86103 .02686 m .86421 .03622 L .8672 .02952 L p .128 .128 .663 r
F P
0 g
s
.86045 .026 m .86291 .02661 L .86039 .02613 L .369 .693 1 r
F
.86291 .02661 m .86274 .0275 L .86039 .02613 L p F P
0 g
s
.86213 .02732 m .86274 .0275 L .8652 .02783 L closepath p .146 .008 .008 r
F P
0 g
s
.86561 .02797 m .8652 .02783 L .86274 .0275 L closepath p .142 .007 .007 r
F P
0 g
s
.8652 .02783 m .86561 .02797 L .86665 .02815 L closepath p .172 .005 .005 r
F P
0 g
s
.8667 .02819 m .86665 .02815 L .86561 .02797 L closepath p .145 .002 .002 r
F P
0 g
s
.86964 .0304 m .86665 .02815 L .87572 .03113 L closepath p .955 .786 .906 r
F P
0 g
s
.86665 .02815 m .8667 .02819 L .86964 .0304 L closepath p 0 .117 .567 r
F P
0 g
s
.87731 .02859 m .87572 .03113 L .86665 .02815 L closepath p .955 .786 .906 r
F P
0 g
s
.86893 .02989 m .86964 .0304 L .86758 .02885 L p .437 .002 .002 r
F P
0 g
s
.86758 .02885 m .8667 .02819 L .86893 .02989 L p .437 .002 .002 r
F P
0 g
s
.8703 .03136 m .86964 .0304 L .87643 .0346 L closepath p .955 .786 .906 r
F P
0 g
s
.87572 .03113 m .87643 .0346 L .86964 .0304 L closepath p .955 .786 .906 r
F P
0 g
s
.86893 .02989 m .87453 .0174 L .87138 .03177 L p .151 .338 .865 r
F P
0 g
s
.86964 .0304 m .86893 .02989 L .87225 .03425 L closepath p .69 .244 .029 r
F P
0 g
s
.87095 .03233 m .8703 .03136 L .87841 .03961 L closepath p .955 .786 .906 r
F P
0 g
s
.87643 .0346 m .87841 .03961 L .8703 .03136 L closepath p .955 .786 .906 r
F P
0 g
s
.8716 .03329 m .87095 .03233 L .88168 .04615 L closepath p .955 .786 .906 r
F P
0 g
s
.87841 .03961 m .88168 .04615 L .87095 .03233 L closepath p .955 .786 .906 r
F P
0 g
s
.87107 .03317 m .87225 .03425 L .86947 .0306 L p .726 .283 .038 r
F P
0 g
s
.86947 .0306 m .86893 .02989 L .87107 .03317 L p .726 .283 .038 r
F P
0 g
s
.87572 .03113 m .879 .02908 L .87731 .02859 L closepath p .135 .135 .184 r
F P
0 g
s
.87643 .0346 m .87967 .03221 L .87572 .03113 L closepath p .388 .388 .829 r
F P
0 g
s
.879 .02908 m .87572 .03113 L .87967 .03221 L closepath p .386 .39 .831 r
F P
0 g
s
.87225 .03425 m .8716 .03329 L .88622 .05425 L closepath p .955 .786 .906 r
F P
0 g
s
.88168 .04615 m .88622 .05425 L .8716 .03329 L closepath p .955 .786 .906 r
F P
0 g
s
.87225 .03425 m .87107 .03317 L .87768 .04509 L closepath p .819 .419 .054 r
F P
0 g
s
.87768 .04509 m .87225 .03425 L .88636 .05448 L closepath p .955 .786 .906 r
F P
0 g
s
.88622 .05425 m .88636 .05448 L .87225 .03425 L closepath p .955 .786 .906 r
F P
0 g
s
.87841 .03961 m .88157 .03691 L .87643 .0346 L closepath p .447 .487 .912 r
F P
0 g
s
.87967 .03221 m .87643 .0346 L .88157 .03691 L closepath p .444 .488 .914 r
F P
0 g
s
.87107 .03317 m .8767 .02072 L .87653 .04192 L p .142 .232 .782 r
F P
0 g
s
.87652 .04403 m .87768 .04509 L .87107 .03317 L closepath p .818 .419 .054 r
F P
0 g
s
.87653 .04192 m .87652 .04403 L .87107 .03317 L p .142 .232 .782 r
F P
0 g
s
.87138 .03177 m .87107 .03317 L .86893 .02989 L p .151 .338 .865 r
F P
0 g
s
.86824 .01497 m .86274 .0275 L .86502 .0155 L p .369 .693 1 r
F P
0 g
s
.86054 .02381 m .86393 .01136 L p .86527 .01291 L .132 .132 .697 r
F P
0 g
s
.86393 .01136 m .86424 .01022 L .86527 .01291 L p .132 .132 .697 r
F P
0 g
s
.86424 .01022 m .86755 .00809 L .86554 .01306 L p .418 .464 .9 r
F P
0 g
s
.86585 .01358 m .86527 .01291 L p .86476 .01609 L .182 .457 .95 r
F P
0 g
s
.86554 .01306 m .86543 .01333 L .86527 .01291 L p .418 .464 .9 r
F P
0 g
s
.86527 .01291 m .86424 .01022 L p .86554 .01306 L .418 .464 .9 r
F P
0 g
s
.86527 .01291 m .86321 .01058 L p .86363 .01206 L .182 .457 .95 r
F P
0 g
s
.86363 .01206 m .86476 .01609 L .86527 .01291 L .182 .457 .95 r
F
.86518 .0129 m .86321 .01058 L p .86527 .01291 L .551 .069 .005 r
F P
0 g
s
.86585 .01358 m .86824 .01497 L p .86724 .01514 L .369 .693 1 r
F P
0 g
s
.86724 .01514 m .86502 .0155 L .86585 .01358 L .369 .693 1 r
F
.86476 .01609 m .86651 .01432 L .86585 .01358 L p .182 .457 .95 r
F P
0 g
s
.86755 .00809 m .86851 .00933 L .86552 .01321 L p .583 .652 1 r
F P
0 g
s
.86552 .01321 m .86543 .01333 L .86549 .01317 L p .583 .652 1 r
F P
0 g
s
.86549 .01317 m .86755 .00809 L p .86552 .01321 L .583 .652 1 r
F P
0 g
s
.86595 .0137 m .86614 .01393 L .86549 .01318 L .551 .069 .005 r
F
.86651 .01432 m .86644 .01425 L p .86595 .0137 L F P
0 g
s
.86644 .01425 m .86614 .01393 L p .86595 .0137 L .551 .069 .005 r
F P
0 g
s
.86851 .00933 m .86824 .01497 L .86585 .01358 L p .695 .812 1 r
F P
0 g
s
.86585 .01358 m .86543 .01333 L .86552 .01321 L p .695 .812 1 r
F P
0 g
s
.86552 .01321 m .86851 .00933 L p .86585 .01358 L .695 .812 1 r
F P
0 g
s
.86585 .01358 m .86549 .01318 L .8654 .01307 L .551 .069 .005 r
F
.86527 .01291 m .86585 .01358 L p .8654 .01307 L F P
0 g
s
.86585 .01358 m .86651 .01432 L p .86549 .01318 L .551 .069 .005 r
F P
0 g
s
.86321 .01058 m .85779 .02317 L .86476 .01609 L p .182 .457 .95 r
F P
0 g
s
.86527 .01291 m .86543 .01333 L p .86054 .02381 L .132 .132 .697 r
F P
0 g
s
.86614 .01393 m .86518 .0129 L p .86527 .01291 L .551 .069 .005 r
F P
0 g
s
.86502 .0155 m .86543 .01333 L .86585 .01358 L p .369 .693 1 r
F P
0 g
s
.86561 .02797 m .86274 .0275 L .86768 .01626 L p .549 .902 1 r
F P
0 g
s
.86768 .01626 m .86824 .01497 L .86791 .01661 L p .549 .902 1 r
F P
0 g
s
.86791 .01661 m .86561 .02797 L p .86768 .01626 L .549 .902 1 r
F P
0 g
s
.86274 .0275 m .86213 .02732 L .86045 .0262 L p .015 .255 .666 r
F P
0 g
s
.86274 .0275 m .86045 .0262 L .8608 .02638 L .015 .255 .666 r
F
.86824 .01497 m .87115 .01545 L .86956 .01906 L p .549 .902 1 r
F P
0 g
s
.86956 .01906 m .86561 .02797 L .86791 .01661 L p .549 .902 1 r
F P
0 g
s
.86791 .01661 m .86824 .01497 L p .86956 .01906 L .549 .902 1 r
F P
0 g
s
.86561 .02797 m .87115 .01545 L .8667 .02819 L closepath p .535 .886 1 r
F P
0 g
s
.87225 .01567 m .8667 .02819 L .87115 .01545 L closepath p .535 .886 1 r
F P
0 g
s
.8667 .02819 m .87225 .01567 L .86893 .02989 L closepath p .303 .61 1 r
F P
0 g
s
.86974 .02382 m .86651 .01432 L .86117 .02655 L p .128 .128 .663 r
F P
0 g
s
.86046 .02597 m .85994 .02517 L .86039 .02613 L .182 .457 .95 r
F
.86002 .02571 m .85981 .02547 L p .85998 .02584 L .542 .653 1 r
F P
0 g
s
.85998 .02584 m .86009 .02607 L .86002 .02571 L .542 .653 1 r
F
.86009 .02596 m .86004 .02593 L p .86039 .02613 L .015 .255 .666 r
F P
0 g
s
.86039 .02613 m .86002 .02571 L p .86007 .02594 L .542 .653 1 r
F P
0 g
s
.86004 .02593 m .85997 .02588 L .86039 .02613 L p .015 .255 .666 r
F P
0 g
s
.86007 .02594 m .86009 .02607 L .86039 .02613 L .542 .653 1 r
F
.86543 .01333 m .85997 .02588 L .86054 .02381 L p .132 .132 .697 r
F P
0 g
s
.85997 .02588 m .86543 .01333 L .86502 .0155 L p .369 .693 1 r
F P
0 g
s
.86088 .02641 m .86274 .0275 L p .8608 .02638 L .015 .255 .666 r
F P
0 g
s
.86103 .02686 m .86651 .01432 L .86476 .01609 L p .182 .457 .95 r
F P
0 g
s
.86838 .02232 m .86518 .0129 L .86605 .01384 L p .911 .568 .072 r
F P
0 g
s
.86605 .01384 m .86651 .01432 L .86657 .01462 L p .911 .568 .072 r
F P
0 g
s
.86657 .01462 m .86838 .02232 L p .86605 .01384 L .911 .568 .072 r
F P
0 g
s
.86117 .02655 m .86103 .02686 L .86147 .02671 L p .128 .128 .663 r
F P
0 g
s
.86147 .02671 m .86974 .02382 L p .86117 .02655 L .128 .128 .663 r
F P
0 g
s
.86039 .02613 m .86088 .02641 L p .86009 .02596 L .015 .255 .666 r
F P
0 g
s
.8608 .02638 m .86009 .02596 L .86088 .02641 L .015 .255 .666 r
F
.86974 .02382 m .86838 .02232 L .86657 .01462 L p .91 .568 .071 r
F P
0 g
s
.86657 .01462 m .86651 .01432 L .86664 .01471 L p .91 .568 .071 r
F P
0 g
s
.86664 .01471 m .86974 .02382 L p .86657 .01462 L .91 .568 .071 r
F P
0 g
s
.8672 .02952 m .86974 .02382 L .86147 .02671 L p .128 .128 .663 r
F P
0 g
s
.86147 .02671 m .86103 .02686 L p .8672 .02952 L .128 .128 .663 r
F P
0 g
s
.86045 .0262 m .86009 .02596 L p .8608 .02638 L .015 .255 .666 r
F P
0 g
s
.86039 .02613 m .86103 .02686 L p .86066 .02627 L .182 .457 .95 r
F P
0 g
s
.86066 .02627 m .86046 .02597 L .86039 .02613 L .182 .457 .95 r
F
.86039 .02613 m .85997 .02588 L p .86045 .026 L .369 .693 1 r
F P
0 g
s
.85994 .02517 m .85831 .02263 L .85981 .02547 L .182 .457 .95 r
F
.8632 .01768 m .85831 .02263 L p .85994 .02517 L F P
0 g
s
.85981 .02547 m .85779 .02317 L p .86048 .0269 L .542 .653 1 r
F P
0 g
s
.85981 .02547 m .85997 .02588 L .86001 .02589 L p .968 .604 .113 r
F P
0 g
s
.86054 .02381 m .85997 .02588 L .85981 .02547 L p .132 .132 .697 r
F P
0 g
s
.85997 .02588 m .86502 .0155 L .86045 .026 L .369 .693 1 r
F
.86103 .02686 m .86476 .01609 L .86046 .02597 L .182 .457 .95 r
F
.85831 .02263 m .85779 .02317 L .85981 .02547 L p F P
0 g
s
.86048 .0269 m .86103 .02686 L .86039 .02613 L p .542 .653 1 r
F P
0 g
s
.86009 .02607 m .86048 .0269 L .86039 .02613 L .542 .653 1 r
F
.10562 .0308 m .10661 .02751 L .10303 .02612 L closepath p .22 .389 .89 r
F P
0 g
s
.10362 .02239 m .10303 .02612 L .10661 .02751 L closepath p .21 .392 .895 r
F P
0 g
s
.10561 .0308 m .10288 .02659 L .10303 .02612 L closepath p .955 .786 .906 r
F P
0 g
s
.18735 .02779 m .15589 .02615 L .19128 .02769 L closepath p .955 .786 .906 r
F P
0 g
s
.15589 .02615 m .18735 .02779 L .18392 .02835 L closepath p .955 .786 .906 r
F P
0 g
s
.18329 .02866 m .15589 .02615 L .18392 .02835 L closepath p .955 .786 .906 r
F P
0 g
s
.14912 .02906 m .138 .02847 L .15589 .02615 L closepath p .955 .786 .906 r
F P
0 g
s
.13343 .02815 m .15589 .02615 L .138 .02847 L closepath p .955 .786 .906 r
F P
0 g
s
.1434 .03645 m .14296 .03396 L .15589 .02615 L closepath p .955 .786 .906 r
F P
0 g
s
.14235 .03196 m .15589 .02615 L .14296 .03396 L closepath p .955 .786 .906 r
F P
0 g
s
.15402 .04279 m .14393 .04092 L .15589 .02615 L closepath p .955 .786 .906 r
F P
0 g
s
.1434 .03645 m .15589 .02615 L .14393 .04092 L closepath p .955 .786 .906 r
F P
0 g
s
.15059 .11967 m .14949 .1063 L .15589 .02615 L closepath p .955 .786 .906 r
F P
0 g
s
.14839 .09295 m .15589 .02615 L .14949 .1063 L closepath p .955 .786 .906 r
F P
0 g
s
.15589 .02615 m .18329 .02866 L .17014 .02755 L closepath p .955 .786 .906 r
F P
0 g
s
.15589 .02615 m .17014 .02755 L .1646 .02823 L closepath p .955 .786 .906 r
F P
0 g
s
.15589 .02615 m .1646 .02823 L .15865 .0297 L closepath p .955 .786 .906 r
F P
0 g
s
.15059 .11967 m .15589 .02615 L .15657 .0341 L closepath p .955 .786 .906 r
F P
0 g
s
.15865 .0297 m .15657 .0341 L .15589 .02615 L closepath p .955 .786 .906 r
F P
0 g
s
.37376 .02784 m .37057 .02621 L .40065 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.40538 .02739 m .40065 .02785 L .37057 .02621 L closepath p .955 .786 .906 r
F P
0 g
s
.24993 .02735 m .25063 .02699 L .2508 .02632 L closepath p .029 .029 .029 r
F P
0 g
s
.25252 .02426 m .2508 .02632 L .25063 .02699 L closepath p .024 .024 .024 r
F P
0 g
s
.24738 .02769 m .2508 .02632 L .24993 .02735 L closepath p .955 .786 .906 r
F P
0 g
s
.94365 .02736 m .94616 .027 L .94447 .02633 L closepath p .138 .138 .144 r
F P
0 g
s
.94792 .02427 m .94447 .02633 L .94616 .027 L closepath p .14 .14 .201 r
F P
0 g
s
.68083 .02736 m .68266 .027 L .68167 .02633 L closepath p .103 .103 .103 r
F P
0 g
s
.68447 .02427 m .68167 .02633 L .68266 .027 L closepath p .104 .104 .104 r
F P
0 g
s
.67829 .0277 m .68167 .02633 L .68083 .02736 L closepath p .955 .786 .906 r
F P
0 g
s
.94112 .0277 m .94447 .02633 L .94365 .02736 L closepath p .955 .786 .906 r
F P
0 g
s
.92047 .03128 m .91911 .02865 L .94447 .02633 L closepath p .955 .786 .906 r
F P
0 g
s
.93247 .0288 m .94447 .02633 L .93594 .02922 L closepath p .955 .786 .906 r
F P
0 g
s
.94112 .0277 m .93594 .02922 L .94447 .02633 L closepath p .955 .786 .906 r
F P
0 g
s
.59889 .02642 m .57364 .02888 L .5712 .02799 L closepath p .955 .786 .906 r
F P
0 g
s
.5712 .02799 m .56712 .0277 L .59889 .02642 L closepath p .955 .786 .906 r
F P
0 g
s
.76529 .02647 m .77929 .02897 L .76243 .02922 L closepath p .955 .786 .906 r
F P
0 g
s
.76243 .02922 m .75602 .02785 L .76529 .02647 L closepath p .955 .786 .906 r
F P
0 g
s
.35373 .02693 m .35577 .02605 L .35452 .02647 L closepath p .025 .025 .025 r
F P
0 g
s
.35651 .02441 m .35452 .02647 L .35577 .02605 L closepath p .087 .087 .087 r
F P
0 g
s
.29488 .02693 m .29677 .02605 L .29567 .02647 L closepath p .017 .017 .017 r
F P
0 g
s
.29751 .02441 m .29567 .02647 L .29677 .02605 L closepath p .076 .076 .076 r
F P
0 g
s
.26914 .0299 m .26599 .02815 L .29567 .02647 L closepath p .955 .786 .906 r
F P
0 g
s
.26914 .0299 m .29567 .02647 L .27688 .02943 L closepath p .955 .786 .906 r
F P
0 g
s
.29408 .02739 m .27688 .02943 L .29567 .02647 L closepath p .955 .786 .906 r
F P
0 g
s
.19395 .02739 m .19448 .02706 L .19486 .02647 L closepath p .014 .014 .014 r
F P
0 g
s
.19645 .02442 m .19486 .02647 L .19448 .02706 L closepath p .005 .005 .005 r
F P
0 g
s
.40617 .02693 m .40834 .02605 L .40696 .02648 L closepath p .031 .031 .031 r
F P
0 g
s
.40907 .02442 m .40696 .02648 L .40834 .02605 L closepath p .097 .097 .097 r
F P
0 g
s
.88767 .02739 m .89001 .02707 L .88854 .02648 L closepath p .122 .122 .122 r
F P
0 g
s
.89185 .02442 m .88854 .02648 L .89001 .02707 L closepath p .127 .127 .127 r
F P
0 g
s
.78938 .02694 m .79254 .02605 L .79015 .02648 L closepath p .077 .077 .077 r
F P
0 g
s
.79322 .02442 m .79015 .02648 L .79254 .02605 L closepath p .154 .154 .349 r
F P
0 g
s
.72666 .02694 m .72966 .02605 L .72743 .02648 L closepath p .07 .07 .07 r
F P
0 g
s
.73034 .02442 m .72743 .02648 L .72966 .02605 L closepath p .146 .146 .314 r
F P
0 g
s
.62509 .02739 m .62674 .02707 L .62597 .02648 L closepath p .087 .087 .087 r
F P
0 g
s
.62863 .02442 m .62597 .02648 L .62674 .02707 L closepath p .087 .087 .087 r
F P
0 g
s
.62509 .02739 m .62348 .02661 L .62597 .02648 L closepath p .955 .786 .906 r
F P
0 g
s
.10288 .02659 m .10561 .0308 L .10456 .02987 L closepath p .955 .786 .906 r
F P
0 g
s
.68718 .02586 m .68844 .02659 L .68592 .02251 L closepath p .94 .58 .084 r
F P
0 g
s
.68844 .02659 m .68718 .02586 L .69029 .02762 L closepath p .003 .232 .628 r
F P
0 g
s
.22004 .02776 m .2185 .02775 L .21527 .02661 L closepath p .955 .786 .906 r
F P
0 g
s
.21716 .02769 m .21527 .02661 L .2185 .02775 L closepath p .955 .786 .906 r
F P
0 g
s
.65097 .02777 m .64946 .02776 L .64619 .02661 L closepath p .955 .786 .906 r
F P
0 g
s
.64807 .0277 m .64619 .02661 L .64946 .02776 L closepath p .955 .786 .906 r
F P
0 g
s
.2128 .026 m .21523 .02672 L .21157 .02249 L closepath p 1 .758 .38 r
F P
0 g
s
.21523 .02672 m .2128 .026 L .21716 .02769 L closepath p .472 .042 .042 r
F P
0 g
s
.90825 .02599 m .9089 .02673 L .90687 .0225 L closepath p .772 .421 .032 r
F P
0 g
s
.91372 .02777 m .9122 .02776 L .9089 .02673 L closepath p .955 .786 .906 r
F P
0 g
s
.91082 .0277 m .9089 .02673 L .9122 .02776 L closepath p .955 .786 .906 r
F P
0 g
s
.9089 .02673 m .90825 .02599 L .91082 .0277 L closepath p .036 .366 .759 r
F P
0 g
s
.6448 .02595 m .64613 .02673 L .64355 .02225 L closepath p .967 .62 .121 r
F P
0 g
s
.64613 .02673 m .6448 .02595 L .64807 .0277 L closepath p .01 .289 .669 r
F P
0 g
s
.51356 .02713 m .51599 .0263 L .51433 .02678 L closepath p .022 .022 .022 r
F P
0 g
s
.51672 .02473 m .51433 .02678 L .51599 .0263 L closepath p .112 .112 .146 r
F P
0 g
s
.04207 .02785 m .04025 .02679 L .07318 .02708 L closepath p .955 .786 .906 r
F P
0 g
s
.25455 .02621 m .25676 .0268 L .25299 .02251 L closepath p 1 .695 .294 r
F P
0 g
s
.25676 .0268 m .25455 .02621 L .25929 .02769 L closepath p .302 .018 .018 r
F P
0 g
s
.41273 .0262 m .41457 .02684 L .41124 .02251 L closepath p 1 .662 .222 r
F P
0 g
s
.41457 .02684 m .41273 .0262 L .41694 .02785 L closepath p .329 .014 .014 r
F P
0 g
s
.13195 .02813 m .13046 .02811 L .12729 .02685 L closepath p .955 .786 .906 r
F P
0 g
s
.12894 .028 m .12729 .02685 L .13046 .02811 L closepath p .955 .786 .906 r
F P
0 g
s
.48598 .03367 m .48804 .0306 L .48225 .02689 L closepath p .378 .487 .939 r
F P
0 g
s
.48669 .03645 m .48447 .03417 L .48225 .02689 L closepath p .955 .786 .906 r
F P
0 g
s
.48225 .02689 m .48598 .03367 L .48669 .03645 L closepath p .955 .786 .906 r
F P
0 g
s
.03719 .02596 m .04009 .02692 L .03599 .02289 L closepath p 1 .764 .42 r
F P
0 g
s
.04009 .02692 m .03719 .02596 L .04108 .02738 L closepath p .436 .036 .036 r
F P
0 g
s
.35294 .02739 m .35385 .02714 L .35373 .02693 L closepath p .023 .023 .023 r
F P
0 g
s
.35577 .02605 m .35373 .02693 L .35385 .02714 L closepath p .023 .023 .023 r
F P
0 g
s
.29408 .02739 m .29484 .02714 L .29488 .02693 L closepath p .016 .016 .016 r
F P
0 g
s
.29677 .02605 m .29488 .02693 L .29484 .02714 L closepath p .016 .016 .016 r
F P
0 g
s
.40538 .02739 m .40643 .02714 L .40617 .02693 L closepath p .03 .03 .03 r
F P
0 g
s
.40834 .02605 m .40617 .02693 L .40643 .02714 L closepath p .03 .03 .03 r
F P
0 g
s
.78861 .02739 m .79067 .02714 L .78938 .02694 L closepath p .076 .076 .076 r
F P
0 g
s
.79254 .02605 m .78938 .02694 L .79067 .02714 L closepath p .075 .075 .075 r
F P
0 g
s
.72589 .02739 m .72778 .02714 L .72666 .02694 L closepath p .069 .069 .069 r
F P
0 g
s
.72966 .02605 m .72666 .02694 L .72778 .02714 L closepath p .068 .068 .068 r
F P
0 g
s
.12444 .02627 m .12713 .02699 L .1233 .02296 L closepath p 1 .783 .431 r
F P
0 g
s
.12713 .02699 m .12444 .02627 L .12894 .028 L closepath p .571 .063 .063 r
F P
0 g
s
.56135 .0266 m .56256 .02699 L .55952 .02501 L closepath p .633 .095 .047 r
F P
0 g
s
.56256 .02699 m .56135 .0266 L .56712 .0277 L closepath p .02 .546 .772 r
F P
0 g
s
.31447 .02659 m .31636 .02703 L .31276 .02484 L closepath p .813 .297 .092 r
F P
0 g
s
.31636 .02703 m .31447 .02659 L .32067 .02785 L closepath p .008 .484 .724 r
F P
0 g
s
.75115 .02658 m .75192 .02703 L .74879 .02251 L closepath p .76 .337 .042 r
F P
0 g
s
.75192 .02703 m .75115 .02658 L .75602 .02785 L closepath p .029 .544 .799 r
F P
0 g
s
.847 .03135 m .84323 .02921 L .85589 .02724 L closepath p .955 .786 .906 r
F P
0 g
s
.83955 .02711 m .847 .03135 L .84517 .03032 L p .001 .221 .62 r
F P
0 g
s
.84517 .03032 m .83946 .02707 L .83955 .02711 L p .001 .221 .62 r
F P
0 g
s
.83955 .02711 m .83946 .02707 L .83251 .02585 L closepath p .132 .002 .002 r
F P
0 g
s
.83955 .02711 m .83251 .02585 L .83341 .0236 L p .541 .895 1 r
F P
0 g
s
.83341 .0236 m .83755 .01329 L .83955 .02711 L p .541 .895 1 r
F P
0 g
s
.07318 .02708 m .04325 .02785 L .04207 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.07318 .02708 m .04696 .02785 L .04325 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.06007 .02747 m .07318 .02708 L .06589 .02815 L closepath p .955 .786 .906 r
F P
0 g
s
.07226 .02748 m .07362 .02667 L .07318 .02708 L closepath p .36 .824 .723 r
F P
0 g
s
.07447 .02502 m .07318 .02708 L .07362 .02667 L closepath p .019 .019 .019 r
F P
0 g
s
.06589 .02815 m .07318 .02708 L .07134 .02788 L closepath p .955 .786 .906 r
F P
0 g
s
.06954 .02762 m .06589 .02815 L .07318 .02708 L closepath p .955 .786 .906 r
F P
0 g
s
.06954 .02762 m .07318 .02708 L .06589 .02815 L closepath p .955 .786 .906 r
F P
0 g
s
.5128 .02747 m .51407 .02727 L .51356 .02713 L closepath p .033 .033 .033 r
F P
0 g
s
.51599 .0263 m .51356 .02713 L .51407 .02727 L closepath p .034 .034 .034 r
F P
0 g
s
.85792 .0296 m .86103 .02686 L .85589 .02724 L closepath p .543 .65 1 r
F P
0 g
s
.59158 .028 m .59217 .02819 L .59884 .02724 L closepath p .024 .024 .024 r
F P
0 g
s
.59895 .02748 m .59884 .02724 L .59217 .02819 L closepath p .024 .024 .024 r
F P
0 g
s
.59884 .02724 m .60492 .02854 L .60186 .02791 L closepath p .955 .786 .906 r
F P
0 g
s
.59884 .02724 m .59895 .02748 L .60492 .02854 L closepath p .223 .02 .02 r
F P
0 g
s
.85589 .02724 m .85691 .02842 L .85721 .0421 L closepath p .955 .786 .906 r
F P
0 g
s
.85589 .02724 m .85721 .0421 L .847 .03135 L closepath p .955 .786 .906 r
F P
0 g
s
.24738 .02769 m .24671 .0279 L .24993 .02735 L closepath p .012 .012 .012 r
F P
0 g
s
.25063 .02699 m .24993 .02735 L .24671 .0279 L closepath p .232 .754 .758 r
F P
0 g
s
.94112 .0277 m .94229 .02791 L .94365 .02736 L closepath p .04 .035 .035 r
F P
0 g
s
.94616 .027 m .94365 .02736 L .94229 .02791 L closepath p .022 .022 .022 r
F P
0 g
s
.67829 .0277 m .67877 .02791 L .68083 .02736 L closepath p .026 .026 .026 r
F P
0 g
s
.68266 .027 m .68083 .02736 L .67877 .02791 L closepath p .006 .006 .006 r
F P
0 g
s
.03854 .02644 m .04108 .02738 L .03719 .02596 L closepath p .004 .366 .683 r
F P
0 g
s
.04108 .02738 m .03854 .02644 L .04207 .02785 L closepath p .387 .023 .023 r
F P
0 g
s
.34821 .02784 m .3479 .02805 L .35294 .02739 L closepath p .017 .017 .017 r
F P
0 g
s
.35385 .02714 m .35294 .02739 L .3479 .02805 L closepath p .148 .693 .739 r
F P
0 g
s
.27688 .02943 m .29408 .02739 L .28461 .02897 L closepath p .955 .786 .906 r
F P
0 g
s
.28935 .02784 m .28461 .02897 L .29408 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.28935 .02784 m .29408 .02739 L .28461 .02897 L closepath p .955 .786 .906 r
F P
0 g
s
.28935 .02784 m .28888 .02805 L .29408 .02739 L closepath p .016 .016 .016 r
F P
0 g
s
.29484 .02714 m .29408 .02739 L .28888 .02805 L closepath p .156 .7 .747 r
F P
0 g
s
.19128 .02769 m .19048 .0279 L .19395 .02739 L closepath p .011 .011 .011 r
F P
0 g
s
.19448 .02706 m .19395 .02739 L .19048 .0279 L closepath p .224 .75 .768 r
F P
0 g
s
.40065 .02785 m .40047 .02806 L .40538 .02739 L closepath p .018 .018 .018 r
F P
0 g
s
.40643 .02714 m .40538 .02739 L .40047 .02806 L closepath p .14 .686 .731 r
F P
0 g
s
.88502 .0277 m .88604 .02791 L .88767 .02739 L closepath p .04 .031 .031 r
F P
0 g
s
.89001 .02707 m .88767 .02739 L .88604 .02791 L closepath p .016 .016 .016 r
F P
0 g
s
.62243 .0277 m .62276 .02791 L .62509 .02739 L closepath p .024 .024 .024 r
F P
0 g
s
.62674 .02707 m .62509 .02739 L .62276 .02791 L closepath p .002 .002 .002 r
F P
0 g
s
.62243 .0277 m .61959 .0277 L .62509 .02739 L closepath p .955 .786 .906 r
F P
0 g
s
.78395 .02785 m .78479 .02806 L .78861 .02739 L closepath p .038 .028 .028 r
F P
0 g
s
.79067 .02714 m .78861 .02739 L .78479 .02806 L closepath p .007 .007 .007 r
F P
0 g
s
.72122 .02785 m .72189 .02806 L .72589 .02739 L closepath p .032 .026 .026 r
F P
0 g
s
.72778 .02714 m .72589 .02739 L .72189 .02806 L closepath p .004 .004 .004 r
F P
0 g
s
.05301 .03129 m .05049 .02871 L .06007 .02747 L closepath p .955 .786 .906 r
F P
0 g
s
.04696 .02785 m .06007 .02747 L .05049 .02871 L closepath p .955 .786 .906 r
F P
0 g
s
.06147 .02948 m .05988 .03348 L .06007 .02747 L closepath p .955 .786 .906 r
F P
0 g
s
.05301 .03129 m .06007 .02747 L .05988 .03348 L closepath p .955 .786 .906 r
F P
0 g
s
.06147 .02948 m .06007 .02747 L .06589 .02815 L closepath p .955 .786 .906 r
F P
0 g
s
.50818 .0277 m .5082 .02791 L .5128 .02747 L closepath p .043 .02 .02 r
F P
0 g
s
.51407 .02727 m .5128 .02747 L .5082 .02791 L closepath p .098 .655 .724 r
F P
0 g
s
.07134 .02788 m .07146 .02769 L .07226 .02748 L closepath p .331 .801 .696 r
F P
0 g
s
.07362 .02667 m .07226 .02748 L .07146 .02769 L closepath p .361 .818 .696 r
F P
0 g
s
.1646 .02823 m .17014 .02755 L .16792 .02806 L closepath p .955 .786 .906 r
F P
0 g
s
.17014 .02755 m .16737 .02789 L .16737 .02798 L p .047 .035 .035 r
F P
0 g
s
.16853 .02796 m .17014 .02755 L p .16737 .02798 L .047 .035 .035 r
F P
0 g
s
.16967 .02778 m .17014 .02755 L .16853 .02796 L p .014 .014 .014 r
F P
0 g
s
.16853 .02796 m .16738 .02825 L .16967 .02778 L p .014 .014 .014 r
F P
0 g
s
.17014 .02755 m .16967 .02778 L .17463 .02937 L closepath p .405 .038 .038 r
F P
0 g
s
.17463 .02937 m .17014 .02755 L .18329 .02866 L closepath p .955 .786 .906 r
F P
0 g
s
.16737 .02789 m .1646 .02823 L .17014 .02755 L closepath p .955 .786 .906 r
F P
0 g
s
.19262 .01963 m .18797 .00699 L .127 .01963 L closepath p .623 .977 1 r
F P
0 g
s
.16967 .02778 m .16467 .01526 L .17161 .02749 L p .24 .663 1 r
F P
0 g
s
.17161 .02749 m .17183 .02789 L p .16967 .02778 L .24 .663 1 r
F P
0 g
s
.17251 .02909 m .17463 .02937 L .17186 .02848 L p .509 .059 .059 r
F P
0 g
s
.17186 .02848 m .16967 .02778 L .17251 .02909 L p .509 .059 .059 r
F P
0 g
s
.17463 .02937 m .17251 .02909 L .17819 .03425 L closepath p .966 .506 .127 r
F P
0 g
s
.17525 .03318 m .17819 .03425 L .17395 .03039 L p 1 .579 .21 r
F P
0 g
s
.17395 .03039 m .17251 .02909 L .17525 .03318 L p 1 .579 .21 r
F P
0 g
s
.17819 .03425 m .17525 .03318 L .18313 .04509 L closepath p 1 .723 .371 r
F P
0 g
s
.18019 .04404 m .18313 .04509 L .17525 .03318 L closepath p 1 .723 .371 r
F P
0 g
s
.17525 .03318 m .17032 .02073 L .17901 .04124 L p .011 .011 .332 r
F P
0 g
s
.17901 .04124 m .18019 .04404 L .17525 .03318 L p .011 .011 .332 r
F P
0 g
s
.17251 .02909 m .16755 .01659 L .17399 .03047 L p .045 .045 .552 r
F P
0 g
s
.17399 .03047 m .17525 .03318 L .17251 .02909 L p .045 .045 .552 r
F P
0 g
s
.17183 .02789 m .17251 .02909 L .16999 .02793 L p .24 .663 1 r
F P
0 g
s
.16999 .02793 m .16967 .02778 L p .17183 .02789 L .24 .663 1 r
F P
0 g
s
.16737 .02798 m .16738 .02825 L .16853 .02796 L p .047 .035 .035 r
F P
0 g
s
.36853 .02774 m .36949 .02762 L .36575 .02575 L closepath p .632 .07 .062 r
F P
0 g
s
.36949 .02762 m .36853 .02774 L .37103 .02779 L closepath p .164 .023 .023 r
F P
0 g
s
.69021 .02776 m .69029 .02762 L .68718 .02586 L closepath p .433 .016 .016 r
F P
0 g
s
.69029 .02762 m .69021 .02776 L .69167 .02776 L closepath p .122 .014 .014 r
F P
0 g
s
.69029 .02762 m .69471 .02785 L .69167 .02776 L closepath p .955 .786 .906 r
F P
0 g
s
.25848 .0279 m .25929 .02769 L .25455 .02621 L closepath p .452 .048 .048 r
F P
0 g
s
.25929 .02769 m .25848 .0279 L .26599 .02815 L closepath p .143 .025 .025 r
F P
0 g
s
.25063 .02699 m .24671 .0279 L .24666 .02774 L p .761 1 1 r
F P
0 g
s
.2486 .01963 m .24479 .00699 L .21269 .01963 L closepath p .623 .977 1 r
F P
0 g
s
.24666 .02774 m .24287 .01538 L .2504 .02664 L p .761 1 1 r
F P
0 g
s
.2504 .02664 m .25063 .02699 L p .24666 .02774 L .761 1 1 r
F P
0 g
s
.24738 .02769 m .24213 .02922 L .24279 .02888 L p .251 .761 .731 r
F P
0 g
s
.24279 .02888 m .24281 .02887 L .24293 .02884 L p .251 .761 .731 r
F P
0 g
s
.24293 .02884 m .24738 .02769 L p .24279 .02888 L .251 .761 .731 r
F P
0 g
s
.24671 .0279 m .24738 .02769 L .24281 .02887 L closepath p .004 .004 .004 r
F P
0 g
s
.21598 .02786 m .21716 .02769 L .2128 .026 L closepath p .602 .069 .069 r
F P
0 g
s
.21716 .02769 m .21598 .02786 L .2185 .02775 L closepath p .111 .021 .021 r
F P
0 g
s
.18735 .02779 m .18671 .02799 L .19128 .02769 L closepath p .052 .018 .018 r
F P
0 g
s
.19048 .0279 m .19128 .02769 L .18671 .02799 L closepath p .055 .018 .018 r
F P
0 g
s
.91149 .02787 m .91082 .0277 L .90825 .02599 L closepath p .018 .267 .678 r
F P
0 g
s
.91082 .0277 m .91149 .02787 L .9122 .02776 L closepath p .084 .014 .014 r
F P
0 g
s
.879 .02908 m .87969 .02818 L .87777 .02851 L p .132 .132 .132 r
F P
0 g
s
.87777 .02851 m .87731 .02859 L .879 .02908 L p .132 .132 .132 r
F P
0 g
s
.879 .02908 m .87967 .03221 L .88075 .02988 L p .117 .117 .556 r
F P
0 g
s
.88044 .02803 m .87731 .02859 L .87969 .02818 L p .002 .002 .002 r
F P
0 g
s
.87939 .02824 m .88475 .01657 L .87969 .02818 L p .74 .953 .602 r
F P
0 g
s
.87969 .02818 m .88226 .0277 L .88044 .02803 L p .002 .002 .002 r
F P
0 g
s
.88328 .02791 m .88226 .0277 L .87969 .02818 L closepath p .053 .028 .028 r
F P
0 g
s
.88226 .0277 m .88328 .02791 L .88502 .0277 L closepath p .078 .02 .02 r
F P
0 g
s
.87969 .02818 m .88545 .01566 L .88328 .02791 L closepath p .648 1 1 r
F P
0 g
s
.88909 .01538 m .88328 .02791 L .88545 .01566 L closepath p .648 1 1 r
F P
0 g
s
.88545 .01566 m .87969 .02818 L .88475 .01657 L closepath p .74 .953 .602 r
F P
0 g
s
.88075 .02988 m .88543 .01975 L .879 .02908 L p .117 .117 .556 r
F P
0 g
s
.87969 .02818 m .879 .02908 L .87939 .02824 L p .74 .953 .602 r
F P
0 g
s
.88604 .02791 m .88502 .0277 L .88328 .02791 L closepath p .078 .02 .02 r
F P
0 g
s
.93843 .02887 m .93594 .02922 L .93654 .03235 L closepath p .152 .152 .292 r
F P
0 g
s
.93843 .02887 m .93654 .03235 L .93804 .02952 L p .63 .791 .347 r
F P
0 g
s
.93871 .0284 m .93594 .02922 L .93843 .02887 L p .035 .035 .035 r
F P
0 g
s
.93804 .02952 m .94316 .01989 L .93843 .02887 L p .63 .791 .347 r
F P
0 g
s
.93843 .02887 m .94112 .0277 L .93871 .0284 L p .035 .035 .035 r
F P
0 g
s
.94229 .02791 m .94112 .0277 L .93843 .02887 L closepath p .046 .046 .046 r
F P
0 g
s
.64805 .02787 m .64807 .0277 L .6448 .02595 L closepath p .417 .016 .016 r
F P
0 g
s
.64807 .0277 m .64805 .02787 L .64946 .02776 L closepath p .094 .017 .017 r
F P
0 g
s
.61457 .02859 m .60492 .02854 L .61959 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.61776 .02802 m .61457 .02859 L .61627 .02818 L p .165 .704 .736 r
F P
0 g
s
.61554 .02908 m .61627 .02818 L .61485 .02853 L p .094 .094 .094 r
F P
0 g
s
.61485 .02853 m .61457 .02859 L .61554 .02908 L p .094 .094 .094 r
F P
0 g
s
.61554 .02908 m .61616 .03215 L .61649 .02983 L p .065 .065 .37 r
F P
0 g
s
.61627 .02818 m .61554 .02908 L .61566 .02824 L p .889 1 .77 r
F P
0 g
s
.62442 .01964 m .62631 .007 L .56228 .01964 L closepath p .623 .977 1 r
F P
0 g
s
.61566 .02824 m .61729 .01657 L .61627 .02818 L p .889 1 .77 r
F P
0 g
s
.61627 .02818 m .61959 .0277 L .61776 .02802 L p .165 .704 .736 r
F P
0 g
s
.61992 .02791 m .61959 .0277 L .61627 .02818 L closepath p .034 .023 .023 r
F P
0 g
s
.61959 .0277 m .61992 .02791 L .62243 .0277 L closepath p .078 .02 .02 r
F P
0 g
s
.61627 .02818 m .61803 .01566 L .61992 .02791 L closepath p .656 1 1 r
F P
0 g
s
.61803 .01566 m .61627 .02818 L .61729 .01657 L closepath p .889 1 .77 r
F P
0 g
s
.61649 .02983 m .61792 .01969 L .61554 .02908 L p .065 .065 .37 r
F P
0 g
s
.62276 .02791 m .62243 .0277 L .61992 .02791 L closepath p .078 .02 .02 r
F P
0 g
s
.5673 .02791 m .56712 .0277 L .56135 .0266 L closepath p .212 .016 .016 r
F P
0 g
s
.56712 .0277 m .5673 .02791 L .5712 .02799 L closepath p .123 .019 .019 r
F P
0 g
s
.50629 .02827 m .50428 .02888 L .50581 .02833 L p .237 .747 .708 r
F P
0 g
s
.50428 .02888 m .50297 .03242 L .50818 .0277 L closepath p .955 .786 .906 r
F P
0 g
s
.50428 .02888 m .50818 .0277 L .50297 .03242 L closepath p .955 .786 .906 r
F P
0 g
s
.50533 .03038 m .50581 .02833 L .50509 .02858 L p .117 .117 .284 r
F P
0 g
s
.50509 .02858 m .50428 .02888 L .50533 .03038 L p .117 .117 .284 r
F P
0 g
s
.50533 .03038 m .50563 .03393 L .50564 .0327 L p .018 .018 .124 r
F P
0 g
s
.50581 .02833 m .50533 .03038 L .50534 .02838 L p .847 .822 .318 r
F P
0 g
s
.50534 .02838 m .50541 .01789 L .50581 .02833 L p .847 .822 .318 r
F P
0 g
s
.50541 .01789 m .5059 .01581 L .50589 .01639 L p .847 .822 .318 r
F P
0 g
s
.50589 .01639 m .50581 .02833 L .50541 .01789 L p .847 .822 .318 r
F P
0 g
s
.50571 .0215 m .50541 .01789 L .50541 .0182 L p .018 .018 .124 r
F P
0 g
s
.50541 .0182 m .50533 .03038 L .50571 .0215 L p .018 .018 .124 r
F P
0 g
s
.50581 .02833 m .50818 .0277 L .50629 .02827 L p .237 .747 .708 r
F P
0 g
s
.5082 .02791 m .50818 .0277 L .50581 .02833 L closepath p .02 .02 .02 r
F P
0 g
s
.50581 .02833 m .5059 .01581 L .5082 .02791 L closepath p .711 1 1 r
F P
0 g
s
.5066 .02639 m .50571 .0215 L .50571 .02174 L p .036 .036 .257 r
F P
0 g
s
.50571 .02174 m .50563 .03393 L .5066 .02639 L p .036 .036 .257 r
F P
0 g
s
.5066 .02639 m .5065 .03876 L .50805 .03254 L closepath p .046 .046 .328 r
F P
0 g
s
.5066 .02639 m .50805 .03254 L .50773 .03232 L p 1 .79 .372 r
F P
0 g
s
.5065 .03876 m .5066 .02639 L .50563 .03393 L closepath p .036 .036 .257 r
F P
0 g
s
.50773 .03232 m .50583 .03096 L .5066 .02639 L p 1 .79 .372 r
F P
0 g
s
.50571 .0215 m .5066 .02639 L .50613 .02602 L p 1 .836 .435 r
F P
0 g
s
.50564 .0327 m .50571 .0215 L .50533 .03038 L p .018 .018 .124 r
F P
0 g
s
.50613 .02602 m .50428 .02458 L .50571 .0215 L p 1 .836 .435 r
F P
0 g
s
.50541 .01789 m .50571 .0215 L .50456 .02053 L p 1 .904 .53 r
F P
0 g
s
.50456 .02053 m .50334 .01952 L .50541 .01789 L p 1 .904 .53 r
F P
0 g
s
.49153 .0277 m .48845 .0277 L .48841 .02785 L p .078 .02 .02 r
F P
0 g
s
.48905 .02786 m .49153 .0277 L p .48841 .02785 L .078 .02 .02 r
F P
0 g
s
.49151 .02791 m .49153 .0277 L .48839 .02791 L closepath p .078 .02 .02 r
F P
0 g
s
.49151 .02791 m .49138 .01538 L .49345 .02794 L p .504 .876 1 r
F P
0 g
s
.49153 .0277 m .49151 .02791 L .49465 .02844 L closepath p .24 .02 .02 r
F P
0 g
s
.49351 .02829 m .49465 .02844 L .49265 .0281 L p .155 .006 .006 r
F P
0 g
s
.49265 .0281 m .49151 .02791 L .49351 .02829 L p .155 .006 .006 r
F P
0 g
s
.49465 .02844 m .49153 .0277 L .49449 .02819 L closepath p .955 .786 .906 r
F P
0 g
s
.49449 .02819 m .50297 .03242 L .49465 .02844 L closepath p .955 .786 .906 r
F P
0 g
s
.49465 .02844 m .49351 .02829 L .4969 .03067 L closepath p .722 .2 .065 r
F P
0 g
s
.4969 .03067 m .49465 .02844 L .50297 .03242 L closepath p .955 .786 .906 r
F P
0 g
s
.49351 .02829 m .49341 .01577 L .49473 .02855 L p .121 .349 .868 r
F P
0 g
s
.49486 .02976 m .4969 .03067 L .49428 .02883 L p .793 .286 .08 r
F P
0 g
s
.49428 .02883 m .49351 .02829 L .49486 .02976 L p .793 .286 .08 r
F P
0 g
s
.4969 .03067 m .49486 .02976 L .49776 .03269 L closepath p 1 .652 .218 r
F P
0 g
s
.49559 .0315 m .49776 .03269 L .49501 .02992 L p 1 .664 .233 r
F P
0 g
s
.49501 .02992 m .49486 .02976 L .49559 .0315 L p 1 .664 .233 r
F P
0 g
s
.49776 .03269 m .49559 .0315 L .49911 .03661 L closepath p 1 .724 .303 r
F P
0 g
s
.49686 .03526 m .49911 .03661 L .49563 .03155 L p 1 .727 .306 r
F P
0 g
s
.49563 .03155 m .49559 .0315 L .49686 .03526 L p 1 .727 .306 r
F P
0 g
s
.4969 .03067 m .50297 .03242 L .49776 .03269 L closepath p .955 .786 .906 r
F P
0 g
s
.49911 .03661 m .49776 .03269 L .50297 .03242 L closepath p .955 .786 .906 r
F P
0 g
s
.50297 .03242 m .50328 .0361 L .49911 .03661 L closepath p .955 .786 .906 r
F P
0 g
s
.50328 .0361 m .50419 .04106 L .50058 .04104 L closepath p .955 .786 .906 r
F P
0 g
s
.50058 .04104 m .49911 .03661 L .50328 .0361 L closepath p .955 .786 .906 r
F P
0 g
s
.49911 .03661 m .5109 .07208 L .50497 .05432 L closepath p .955 .786 .906 r
F P
0 g
s
.49911 .03661 m .49686 .03526 L .5109 .07208 L closepath p 1 .732 .313 r
F P
0 g
s
.50205 .04546 m .50058 .04104 L .50419 .04106 L closepath p .955 .786 .906 r
F P
0 g
s
.50419 .04106 m .50571 .04733 L .50205 .04546 L closepath p .955 .786 .906 r
F P
0 g
s
.50352 .04989 m .50205 .04546 L .50571 .04733 L closepath p .955 .786 .906 r
F P
0 g
s
.50868 .07082 m .5109 .07208 L .49857 .03974 L p 1 .732 .313 r
F P
0 g
s
.49857 .03974 m .49686 .03526 L .50868 .07082 L p 1 .732 .313 r
F P
0 g
s
.50647 .05876 m .50499 .05432 L .51058 .06377 L closepath p .955 .786 .906 r
F P
0 g
s
.50784 .0549 m .51058 .06377 L .50499 .05432 L closepath p .955 .786 .906 r
F P
0 g
s
.50794 .0632 m .50647 .05876 L .51394 .07395 L closepath p .955 .786 .906 r
F P
0 g
s
.51058 .06377 m .51394 .07395 L .50647 .05876 L closepath p .955 .786 .906 r
F P
0 g
s
.50942 .06764 m .50794 .0632 L .51791 .08544 L closepath p .955 .786 .906 r
F P
0 g
s
.51394 .07395 m .51791 .08544 L .50794 .0632 L closepath p .955 .786 .906 r
F P
0 g
s
.5109 .07208 m .50942 .06764 L .52249 .09826 L closepath p .955 .786 .906 r
F P
0 g
s
.51791 .08544 m .52249 .09826 L .50942 .06764 L closepath p .955 .786 .906 r
F P
0 g
s
.5109 .07208 m .50868 .07082 L .51507 .085 L closepath p 1 .74 .322 r
F P
0 g
s
.51507 .085 m .5109 .07208 L .52198 .0997 L closepath p .955 .786 .906 r
F P
0 g
s
.52249 .09826 m .52198 .0997 L .5109 .07208 L closepath p .955 .786 .906 r
F P
0 g
s
.51284 .0837 m .51507 .085 L .50906 .07167 L p 1 .739 .321 r
F P
0 g
s
.50906 .07167 m .50868 .07082 L .51284 .0837 L p 1 .739 .321 r
F P
0 g
s
.51507 .085 m .51284 .0837 L .51805 .09464 L closepath p 1 .75 .333 r
F P
0 g
s
.51579 .0932 m .51805 .09464 L .51329 .08464 L p 1 .747 .329 r
F P
0 g
s
.51329 .08464 m .51284 .0837 L .51579 .0932 L p 1 .747 .329 r
F P
0 g
s
.51805 .09464 m .51507 .085 L .52147 .10114 L closepath p .955 .786 .906 r
F P
0 g
s
.52198 .0997 m .52147 .10114 L .51507 .085 L closepath p .955 .786 .906 r
F P
0 g
s
.51805 .09464 m .51579 .0932 L .51984 .10099 L closepath p 1 .769 .356 r
F P
0 g
s
.51755 .09934 m .51984 .10099 L .51647 .09451 L p 1 .763 .347 r
F P
0 g
s
.51647 .09451 m .51579 .0932 L .51755 .09934 L p 1 .763 .347 r
F P
0 g
s
.51984 .10099 m .51755 .09934 L .52044 .10403 L closepath p 1 .83 .433 r
F P
0 g
s
.51813 .10215 m .52044 .10403 L .5184 .10073 L p 1 .82 .416 r
F P
0 g
s
.5184 .10073 m .51755 .09934 L .51813 .10215 L p 1 .82 .416 r
F P
0 g
s
.51755 .09934 m .51782 .08785 L .51813 .10194 L p .041 .041 .291 r
F P
0 g
s
.51813 .10194 m .51813 .10215 L .51755 .09934 L p .041 .041 .291 r
F P
0 g
s
.51984 .10099 m .51805 .09464 L .52095 .10258 L closepath p .955 .786 .906 r
F P
0 g
s
.52147 .10114 m .52095 .10258 L .51805 .09464 L closepath p .955 .786 .906 r
F P
0 g
s
.51841 .09069 m .51813 .10215 L .51784 .08879 L p .041 .041 .291 r
F P
0 g
s
.51784 .08879 m .51782 .08785 L .51841 .09069 L p .041 .041 .291 r
F P
0 g
s
.54602 .04317 m .52044 .10403 L .52345 .0951 L closepath p .955 .786 .906 r
F P
0 g
s
.52249 .09826 m .52345 .0951 L .52044 .10403 L closepath p .955 .786 .906 r
F P
0 g
s
.52249 .09826 m .52044 .10403 L .52345 .0951 L closepath p .955 .786 .906 r
F P
0 g
s
.52345 .0951 m .52225 .09142 L .52226 .09187 L p 1 1 .755 r
F P
0 g
s
.52226 .09187 m .52249 .09826 L .52308 .09633 L p 1 1 .755 r
F P
0 g
s
.52308 .09633 m .52345 .0951 L p .52226 .09187 L 1 1 .755 r
F P
0 g
s
.54584 .03585 m .52345 .0951 L .5249 .08878 L closepath p .955 .786 .906 r
F P
0 g
s
.54584 .03585 m .54602 .04317 L .52345 .0951 L closepath p .955 .786 .906 r
F P
0 g
s
.52225 .09142 m .52345 .0951 L .52297 .08597 L closepath p 1 .976 .575 r
F P
0 g
s
.52225 .09142 m .52297 .08597 L .52319 .07846 L p .772 .706 .161 r
F P
0 g
s
.52319 .07846 m .52332 .07428 L .52303 .07885 L p .772 .706 .161 r
F P
0 g
s
.52303 .07885 m .52225 .09142 L p .52319 .07846 L .772 .706 .161 r
F P
0 g
s
.52259 .07981 m .52225 .09142 L .52312 .07747 L p .772 .706 .161 r
F P
0 g
s
.52312 .07747 m .52332 .07428 L .52266 .07926 L p .772 .706 .161 r
F P
0 g
s
.52266 .07926 m .52259 .07981 L p .52312 .07747 L .772 .706 .161 r
F P
0 g
s
.52225 .09142 m .52259 .07981 L .52268 .09301 L closepath p .052 .052 .371 r
F P
0 g
s
.52302 .08142 m .52268 .09301 L .5226 .08176 L p .052 .052 .371 r
F P
0 g
s
.5226 .08176 m .52259 .07981 L .52302 .08142 L p .052 .052 .371 r
F P
0 g
s
.52463 .07687 m .525 .06505 L .52297 .08597 L closepath p .799 .756 .227 r
F P
0 g
s
.52332 .07428 m .52297 .08597 L .525 .06505 L closepath p .799 .756 .227 r
F P
0 g
s
.52302 .08142 m .52393 .07965 L .52351 .08104 L p .033 .033 .033 r
F P
0 g
s
.52268 .09301 m .52302 .08142 L .5194 .07615 L p .06 .06 .431 r
F P
0 g
s
.52351 .08104 m .52295 .08287 L .52302 .08142 L p .033 .033 .033 r
F P
0 g
s
.51557 .05834 m .51534 .07025 L .52302 .08142 L closepath p .06 .06 .431 r
F P
0 g
s
.51827 .0698 m .51929 .06987 L p .52231 .0806 L .939 .586 .083 r
F P
0 g
s
.51929 .06987 m .51929 .06987 L .5193 .06987 L p .939 .586 .083 r
F P
0 g
s
.5193 .06987 m .52231 .0806 L p .51929 .06987 L .939 .586 .083 r
F P
0 g
s
.52231 .0806 m .52295 .08287 L .52224 .08088 L p .939 .586 .083 r
F P
0 g
s
.52224 .08088 m .51827 .0698 L p .52231 .0806 L .939 .586 .083 r
F P
0 g
s
.52302 .08142 m .52295 .08287 L .52231 .0806 L p .837 .511 .047 r
F P
0 g
s
.52231 .0806 m .51929 .06987 L .52256 .08 L p .837 .511 .047 r
F P
0 g
s
.52256 .08 m .52302 .08142 L p .52231 .0806 L .837 .511 .047 r
F P
0 g
s
.5194 .07615 m .51534 .07025 L .52221 .09156 L p .06 .06 .431 r
F P
0 g
s
.52221 .09156 m .52268 .09301 L p .5194 .07615 L .06 .06 .431 r
F P
0 g
s
.52268 .09301 m .52249 .09826 L .52225 .09142 L closepath p .129 .129 .629 r
F P
0 g
s
.52249 .09826 m .52268 .09301 L .51791 .08544 L closepath p .152 .197 .745 r
F P
0 g
s
.51901 .08162 m .51791 .08544 L .52268 .09301 L closepath p .165 .234 .765 r
F P
0 g
s
.51841 .09069 m .52086 .08875 L .5189 .09232 L p .156 .156 .487 r
F P
0 g
s
.51813 .10215 m .51841 .09069 L .51771 .10438 L p .78 .719 .178 r
F P
0 g
s
.51792 .09409 m .51765 .1055 L .51829 .09293 L p .78 .719 .178 r
F P
0 g
s
.51829 .09293 m .51841 .09069 L .51792 .09409 L p .78 .719 .178 r
F P
0 g
s
.51579 .0932 m .51603 .08161 L .51753 .09917 L p .053 .053 .385 r
F P
0 g
s
.51753 .09917 m .51755 .09934 L .51579 .0932 L p .053 .053 .385 r
F P
0 g
s
.51782 .08785 m .51755 .09934 L .51603 .08161 L closepath p .053 .053 .385 r
F P
0 g
s
.51765 .1055 m .51792 .09409 L .51595 .10695 L closepath p .878 1 .927 r
F P
0 g
s
.51619 .09556 m .51595 .10695 L .51775 .09519 L p .878 1 .927 r
F P
0 g
s
.51775 .09519 m .51792 .09409 L .51619 .09556 L p .878 1 .927 r
F P
0 g
s
.51595 .10695 m .51619 .09556 L .51349 .10724 L closepath p .646 1 1 r
F P
0 g
s
.5137 .09585 m .51349 .10724 L .51614 .09577 L p .646 1 1 r
F P
0 g
s
.51614 .09577 m .51619 .09556 L .5137 .09585 L p .646 1 1 r
F P
0 g
s
.51349 .10724 m .5137 .09585 L .50809 .10735 L closepath p .592 .97 1 r
F P
0 g
s
.50821 .09596 m .50809 .10735 L .5137 .09585 L closepath p .592 .97 1 r
F P
0 g
s
.5137 .09585 m .51405 .09623 L .50931 .09601 L p .119 .035 .035 r
F P
0 g
s
.50931 .09601 m .50821 .09596 L .5137 .09585 L p .119 .035 .035 r
F P
0 g
s
.51619 .09556 m .51706 .09581 L .515 .09584 L p .044 .033 .033 r
F P
0 g
s
.515 .09584 m .5137 .09585 L .51619 .09556 L p .044 .033 .033 r
F P
0 g
s
.51792 .09409 m .51991 .09334 L .51774 .09463 L p .064 .064 .064 r
F P
0 g
s
.51774 .09463 m .51619 .09556 L .51792 .09409 L p .064 .064 .064 r
F P
0 g
s
.5189 .09232 m .51792 .09409 L .51841 .09069 L p .156 .156 .487 r
F P
0 g
s
.51771 .10438 m .51765 .1055 L .51813 .10215 L p .78 .719 .178 r
F P
0 g
s
.51284 .0837 m .51303 .07197 L .51577 .09304 L p .057 .057 .412 r
F P
0 g
s
.51577 .09304 m .51579 .0932 L .51284 .0837 L p .057 .057 .412 r
F P
0 g
s
.50868 .07082 m .50881 .05891 L .5128 .0835 L p .059 .059 .426 r
F P
0 g
s
.5128 .0835 m .51284 .0837 L .50868 .07082 L p .059 .059 .426 r
F P
0 g
s
.49686 .03526 m .49682 .02284 L .50849 .07003 L p .06 .06 .437 r
F P
0 g
s
.50849 .07003 m .50868 .07082 L .49686 .03526 L p .06 .06 .437 r
F P
0 g
s
.49559 .0315 m .49552 .01903 L .49684 .03495 L p .061 .061 .445 r
F P
0 g
s
.49684 .03495 m .49686 .03526 L .49559 .0315 L p .061 .061 .445 r
F P
0 g
s
.49486 .02976 m .49478 .01726 L .49555 .03087 L p .073 .073 .53 r
F P
0 g
s
.49555 .03087 m .49559 .0315 L .49486 .02976 L p .073 .073 .53 r
F P
0 g
s
.49473 .02855 m .49486 .02976 L .49351 .02829 L p .121 .349 .868 r
F P
0 g
s
.49345 .02794 m .49351 .02829 L .49151 .02791 L p .504 .876 1 r
F P
0 g
s
.48841 .02785 m .48839 .02791 L .48905 .02786 L p .078 .02 .02 r
F P
0 g
s
.67496 .02887 m .67314 .02922 L .67301 .03235 L closepath p .119 .119 .135 r
F P
0 g
s
.67496 .02887 m .67301 .03235 L .67361 .02953 L p .831 .956 .552 r
F P
0 g
s
.67593 .0284 m .67314 .02922 L .67496 .02887 L p .017 .017 .017 r
F P
0 g
s
.67361 .02953 m .67564 .01989 L .67496 .02887 L p .831 .956 .552 r
F P
0 g
s
.67496 .02887 m .67829 .0277 L .67593 .0284 L p .017 .017 .017 r
F P
0 g
s
.67877 .02791 m .67829 .0277 L .67496 .02887 L closepath p .031 .031 .031 r
F P
0 g
s
.21761 .02795 m .2185 .02775 L .21598 .02786 L closepath p .131 .024 .024 r
F P
0 g
s
.2185 .02775 m .21761 .02795 L .22158 .02777 L closepath p .082 .02 .02 r
F P
0 g
s
.91316 .02796 m .9122 .02776 L .91149 .02787 L closepath p .094 .015 .015 r
F P
0 g
s
.9122 .02776 m .91316 .02796 L .91524 .02778 L closepath p .078 .019 .019 r
F P
0 g
s
.64972 .02796 m .64946 .02776 L .64805 .02787 L closepath p .107 .018 .018 r
F P
0 g
s
.64946 .02776 m .64972 .02796 L .65249 .02778 L closepath p .08 .02 .02 r
F P
0 g
s
.69192 .02795 m .69167 .02776 L .69021 .02776 L closepath p .138 .016 .016 r
F P
0 g
s
.69167 .02776 m .69192 .02795 L .69471 .02785 L closepath p .091 .018 .018 r
F P
0 g
s
.22084 .02798 m .22158 .02777 L .21797 .02794 L p .085 .021 .021 r
F P
0 g
s
.20834 .00699 m .21269 .01963 L .24479 .00699 L closepath p .623 .977 1 r
F P
0 g
s
.22084 .02798 m .21661 .01546 L .22316 .02801 L p .563 .934 1 r
F P
0 g
s
.22158 .02777 m .22084 .02798 L .22541 .02864 L closepath p .29 .035 .035 r
F P
0 g
s
.22316 .02801 m .22324 .02816 L .22084 .02798 L p .563 .934 1 r
F P
0 g
s
.21797 .02794 m .21761 .02795 L .22084 .02798 L p .085 .021 .021 r
F P
0 g
s
.22541 .02864 m .22158 .02777 L .22704 .02887 L closepath p .955 .786 .906 r
F P
0 g
s
.91634 .02798 m .91524 .02778 L .91316 .02796 L closepath p .08 .019 .019 r
F P
0 g
s
.91524 .02778 m .91634 .02798 L .91911 .02865 L closepath p .004 .448 .704 r
F P
0 g
s
.6529 .02798 m .65249 .02778 L .64972 .02796 L closepath p .082 .02 .02 r
F P
0 g
s
.65249 .02778 m .6529 .02798 L .65635 .02865 L closepath p .199 .011 .011 r
F P
0 g
s
.65635 .02865 m .65249 .02778 L .65815 .02888 L closepath p .955 .786 .906 r
F P
0 g
s
.37097 .0279 m .37103 .02779 L .36853 .02774 L closepath p .088 .011 .011 r
F P
0 g
s
.37103 .02779 m .37097 .0279 L .37376 .02784 L closepath p .057 .011 .011 r
F P
0 g
s
.18552 .02792 m .18392 .02835 L .18735 .02779 L closepath p .955 .786 .906 r
F P
0 g
s
.18552 .02792 m .18735 .02779 L .18392 .02835 L closepath p .955 .786 .906 r
F P
0 g
s
.19048 .0279 m .18671 .02799 L .18671 .02798 L p .635 .987 1 r
F P
0 g
s
.18735 .02779 m .18552 .02792 L .18513 .02809 L p .014 .014 .014 r
F P
0 g
s
.18518 .02808 m .18735 .02779 L p .18513 .02809 L .014 .014 .014 r
F P
0 g
s
.18671 .02799 m .18735 .02779 L .18512 .02809 L closepath p .015 .015 .015 r
F P
0 g
s
.18671 .02798 m .18196 .01547 L .19036 .02772 L p .635 .987 1 r
F P
0 g
s
.19036 .02772 m .19048 .0279 L p .18671 .02798 L .635 .987 1 r
F P
0 g
s
.18513 .02809 m .18512 .02809 L .18518 .02808 L p .014 .014 .014 r
F P
0 g
s
.37341 .02805 m .37376 .02784 L .37131 .02789 L p .13 .022 .022 r
F P
0 g
s
.37376 .02784 m .37341 .02805 L .37944 .02925 L closepath p .274 .028 .028 r
F P
0 g
s
.37944 .02925 m .37376 .02784 L .39828 .02841 L closepath p .955 .786 .906 r
F P
0 g
s
.40065 .02785 m .39828 .02841 L .37376 .02784 L closepath p .955 .786 .906 r
F P
0 g
s
.40549 .01963 m .40406 .00699 L .36795 .01963 L closepath p .623 .977 1 r
F P
0 g
s
.36595 .00699 m .36795 .01963 L .40406 .00699 L closepath p .623 .977 1 r
F P
0 g
s
.37341 .02805 m .37149 .01553 L .37752 .02837 L p .489 .869 1 r
F P
0 g
s
.37752 .02837 m .37777 .0289 L .37341 .02805 L p .489 .869 1 r
F P
0 g
s
.37131 .02789 m .37097 .0279 L .37341 .02805 L p .13 .022 .022 r
F P
0 g
s
.34608 .02835 m .34348 .02897 L .34453 .02849 L p .245 .763 .767 r
F P
0 g
s
.34453 .02849 m .34348 .02897 L .34382 .02983 L closepath p .066 .066 .066 r
F P
0 g
s
.34453 .02849 m .34382 .02983 L .34359 .02858 L p 1 1 .716 r
F P
0 g
s
.35287 .01963 m .35064 .00699 L .31532 .01963 L closepath p .623 .977 1 r
F P
0 g
s
.34359 .02858 m .34146 .01734 L .34453 .02849 L p 1 1 .716 r
F P
0 g
s
.34453 .02849 m .34821 .02784 L .34608 .02835 L p .245 .763 .767 r
F P
0 g
s
.3479 .02805 m .34821 .02784 L .34453 .02849 L closepath p .016 .016 .016 r
F P
0 g
s
.28722 .02835 m .28461 .02897 L .28552 .02849 L p .258 .774 .779 r
F P
0 g
s
.28552 .02849 m .28461 .02897 L .2848 .02984 L closepath p .054 .054 .054 r
F P
0 g
s
.28552 .02849 m .2848 .02984 L .28447 .02858 L p 1 1 .744 r
F P
0 g
s
.29397 .01963 m .29085 .00699 L .25618 .01963 L closepath p .623 .977 1 r
F P
0 g
s
.28447 .02858 m .28154 .01734 L .28552 .02849 L p 1 1 .744 r
F P
0 g
s
.28552 .02849 m .28935 .02784 L .28722 .02835 L p .258 .774 .779 r
F P
0 g
s
.28888 .02805 m .28935 .02784 L .28552 .02849 L closepath p .014 .014 .014 r
F P
0 g
s
.39828 .02841 m .39592 .02897 L .37944 .02925 L closepath p .955 .786 .906 r
F P
0 g
s
.39852 .02836 m .39592 .02897 L .39711 .0285 L p .233 .754 .756 r
F P
0 g
s
.39641 .02985 m .39711 .0285 L .39641 .02878 L p .077 .077 .077 r
F P
0 g
s
.39641 .02878 m .39592 .02897 L .39641 .02985 L p .077 .077 .077 r
F P
0 g
s
.39656 .03221 m .39641 .02985 L .39631 .02994 L p .164 .164 .629 r
F P
0 g
s
.39631 .02994 m .39432 .03189 L .39656 .03221 L p .164 .164 .629 r
F P
0 g
s
.39704 .03654 m .39656 .03221 L .39649 .03228 L p .168 .168 .667 r
F P
0 g
s
.39649 .03228 m .39448 .03436 L .39704 .03654 L p .168 .168 .667 r
F P
0 g
s
.39704 .03654 m .39785 .04287 L .39784 .04277 L p .001 .001 .092 r
F P
0 g
s
.39448 .03436 m .39656 .03221 L .39432 .03189 L closepath p .164 .164 .628 r
F P
0 g
s
.39784 .04277 m .3963 .03056 L .39704 .03654 L p .001 .001 .092 r
F P
0 g
s
.39656 .03221 m .39704 .03654 L .39702 .03637 L p .714 .519 .003 r
F P
0 g
s
.39432 .03189 m .39641 .02985 L .39592 .02897 L closepath p .078 .078 .078 r
F P
0 g
s
.39702 .03637 m .39548 .02414 L .39656 .03221 L p .714 .519 .003 r
F P
0 g
s
.39641 .02985 m .39656 .03221 L .3964 .03096 L p .751 .577 .012 r
F P
0 g
s
.3964 .03096 m .39499 .01975 L .39641 .02985 L p .751 .577 .012 r
F P
0 g
s
.39711 .0285 m .39641 .02985 L .39625 .02859 L p .98 1 .686 r
F P
0 g
s
.39625 .02859 m .39484 .01735 L .39711 .0285 L p .98 1 .686 r
F P
0 g
s
.39711 .0285 m .40065 .02785 L .39852 .02836 L p .233 .754 .756 r
F P
0 g
s
.40047 .02806 m .40065 .02785 L .39711 .0285 L closepath p .018 .018 .018 r
F P
0 g
s
.3202 .02806 m .32067 .02785 L .31447 .02659 L closepath p .291 .031 .031 r
F P
0 g
s
.32067 .02785 m .3202 .02806 L .32679 .02922 L closepath p .266 .03 .03 r
F P
0 g
s
.32679 .02922 m .32067 .02785 L .34137 .02865 L closepath p .955 .786 .906 r
F P
0 g
s
.78083 .02984 m .78149 .0285 L .78028 .02876 L p .14 .14 .256 r
F P
0 g
s
.78028 .02876 m .77929 .02897 L .78083 .02984 L p .14 .14 .256 r
F P
0 g
s
.78083 .02984 m .78105 .03261 L .78153 .0312 L p .079 .079 .339 r
F P
0 g
s
.78186 .02836 m .77929 .02897 L .78149 .0285 L p .005 .005 .005 r
F P
0 g
s
.78943 .01964 m .79382 .007 L .75188 .01964 L closepath p .623 .977 1 r
F P
0 g
s
.78126 .02858 m .78509 .01735 L .78149 .0285 L p .735 .853 .404 r
F P
0 g
s
.78149 .0285 m .78395 .02785 L .78186 .02836 L p .005 .005 .005 r
F P
0 g
s
.78479 .02806 m .78395 .02785 L .78149 .0285 L closepath p .03 .03 .03 r
F P
0 g
s
.78149 .0285 m .78576 .01598 L .78479 .02806 L closepath p .674 1 1 r
F P
0 g
s
.78509 .01735 m .78576 .01598 L .78552 .01667 L p .735 .853 .404 r
F P
0 g
s
.78552 .01667 m .78149 .0285 L .78509 .01735 L p .735 .853 .404 r
F P
0 g
s
.78149 .0285 m .78083 .02984 L .78126 .02858 L p .735 .853 .404 r
F P
0 g
s
.78509 .01735 m .78083 .02984 L .78531 .02015 L closepath p .079 .079 .339 r
F P
0 g
s
.78509 .01735 m .78531 .02015 L .78418 .01878 L p 1 .847 .397 r
F P
0 g
s
.78418 .01878 m .78365 .01814 L .78509 .01735 L p 1 .847 .397 r
F P
0 g
s
.78153 .0312 m .78531 .02015 L .78083 .02984 L p .079 .079 .339 r
F P
0 g
s
.75669 .02806 m .75602 .02785 L .75115 .02658 L closepath p .193 .003 .003 r
F P
0 g
s
.75602 .02785 m .75669 .02806 L .75649 .02785 L closepath p .078 .02 .02 r
F P
0 g
s
.75717 .02806 m .75649 .02785 L .75669 .02806 L closepath p .078 .02 .02 r
F P
0 g
s
.75649 .02785 m .75717 .02806 L .76249 .02922 L closepath p .177 .005 .005 r
F P
0 g
s
.71793 .02984 m .71859 .0285 L .7174 .02878 L p .131 .131 .216 r
F P
0 g
s
.7174 .02878 m .71656 .02897 L .71793 .02984 L p .131 .131 .216 r
F P
0 g
s
.71793 .02984 m .71813 .03203 L .71844 .03087 L p .068 .068 .312 r
F P
0 g
s
.71859 .0285 m .71793 .02984 L .71826 .02858 L p .784 .894 .457 r
F P
0 g
s
.71913 .02836 m .71656 .02897 L .71859 .0285 L p .001 .001 .001 r
F P
0 g
s
.72681 .01964 m .73025 .007 L .6891 .01964 L closepath p .623 .977 1 r
F P
0 g
s
.71826 .02858 m .72123 .01735 L .71859 .0285 L p .784 .894 .457 r
F P
0 g
s
.71859 .0285 m .72122 .02785 L .71913 .02836 L p .001 .001 .001 r
F P
0 g
s
.72189 .02806 m .72122 .02785 L .71859 .0285 L closepath p .028 .028 .028 r
F P
0 g
s
.71859 .0285 m .72191 .01598 L .72189 .02806 L closepath p .678 1 1 r
F P
0 g
s
.72123 .01735 m .72191 .01598 L .72172 .01669 L p .784 .894 .457 r
F P
0 g
s
.72172 .01669 m .71859 .0285 L .72123 .01735 L p .784 .894 .457 r
F P
0 g
s
.72144 .01957 m .72123 .01735 L .72121 .01743 L p .068 .068 .312 r
F P
0 g
s
.72121 .01743 m .71793 .02984 L .72144 .01957 L p .068 .068 .312 r
F P
0 g
s
.72526 .01554 m .72189 .02806 L .72191 .01598 L closepath p .678 1 1 r
F P
0 g
s
.72144 .01957 m .71813 .03203 L .72203 .0239 L closepath p .076 .076 .365 r
F P
0 g
s
.72144 .01957 m .72203 .0239 L .72189 .02376 L p 1 .819 .372 r
F P
0 g
s
.71844 .03087 m .72144 .01957 L .71793 .02984 L p .068 .068 .312 r
F P
0 g
s
.72189 .02376 m .72022 .022 L .72144 .01957 L p 1 .819 .372 r
F P
0 g
s
.72123 .01735 m .72144 .01957 L .72042 .01845 L p 1 .853 .417 r
F P
0 g
s
.72042 .01845 m .71962 .01756 L .72123 .01735 L p 1 .853 .417 r
F P
0 g
s
.69519 .02806 m .69471 .02785 L .69192 .02795 L closepath p .095 .019 .019 r
F P
0 g
s
.69471 .02785 m .69519 .02806 L .70009 .02926 L closepath p .209 .008 .008 r
F P
0 g
s
.44477 .02785 m .44103 .0289 L .45787 .02785 L closepath p .955 .786 .906 r
F P
0 g
s
.44477 .02785 m .45787 .02785 L .44103 .0289 L closepath p .955 .786 .906 r
F P
0 g
s
.44477 .02785 m .44462 .02806 L .46749 .02788 L p .078 .02 .02 r
F P
0 g
s
.46749 .02788 m .47098 .02785 L .46697 .02785 L p .078 .02 .02 r
F P
0 g
s
.46697 .02785 m .44477 .02785 L p .46749 .02788 L .078 .02 .02 r
F P
0 g
s
.4709 .02806 m .47098 .02785 L .44462 .02806 L closepath p .078 .02 .02 r
F P
0 g
s
.47098 .02785 m .4709 .02806 L .47518 .02992 L closepath p .391 .021 .021 r
F P
0 g
s
.47518 .02992 m .47098 .02785 L .48447 .03417 L closepath p .955 .786 .906 r
F P
0 g
s
.48106 .01963 m .48077 .00699 L .41439 .01963 L closepath p .623 .977 1 r
F P
0 g
s
.4709 .02806 m .47046 .01554 L .47354 .0282 L p .26 .643 1 r
F P
0 g
s
.47392 .02976 m .47518 .02992 L .47219 .02862 L p .514 .046 .046 r
F P
0 g
s
.47219 .02862 m .4709 .02806 L .47392 .02976 L p .514 .046 .046 r
F P
0 g
s
.48447 .03417 m .48873 .04439 L .47518 .02992 L closepath p .955 .786 .906 r
F P
0 g
s
.48099 .03612 m .47518 .02992 L .48873 .04439 L closepath p .955 .786 .906 r
F P
0 g
s
.48099 .03612 m .48873 .04439 L .47518 .02992 L closepath p .955 .786 .906 r
F P
0 g
s
.47518 .02992 m .47392 .02976 L .47809 .03302 L closepath p .771 .26 .075 r
F P
0 g
s
.47392 .02976 m .47352 .01727 L .47604 .03117 L p .122 .383 .891 r
F P
0 g
s
.4762 .03206 m .47809 .03302 L .47438 .03012 L p .716 .196 .062 r
F P
0 g
s
.47438 .03012 m .47392 .02976 L .4762 .03206 L p .716 .196 .062 r
F P
0 g
s
.47809 .03302 m .4762 .03206 L .48099 .03612 L closepath p .746 .235 .067 r
F P
0 g
s
.47909 .03547 m .48099 .03612 L .4762 .03206 L closepath p .846 .353 .092 r
F P
0 g
s
.4762 .03206 m .47584 .0196 L .479 .03503 L p .113 .295 .828 r
F P
0 g
s
.479 .03503 m .47909 .03547 L .4762 .03206 L p .113 .295 .828 r
F P
0 g
s
.47604 .03117 m .4762 .03206 L .47392 .02976 L p .122 .383 .891 r
F P
0 g
s
.47354 .0282 m .47392 .02976 L .4709 .02806 L p .26 .643 1 r
F P
0 g
s
.4709 .02806 m .44462 .02806 L .4446 .02772 L p .619 .975 1 r
F P
0 g
s
.4446 .02772 m .44378 .01554 L .46 .02303 L p .619 .975 1 r
F P
0 g
s
.46 .02303 m .4709 .02806 L p .4446 .02772 L .619 .975 1 r
F P
0 g
s
.44477 .02785 m .44103 .0289 L .44211 .02848 L p .248 .76 .735 r
F P
0 g
s
.44211 .02848 m .44238 .02837 L .44267 .02831 L p .248 .76 .735 r
F P
0 g
s
.44267 .02831 m .44477 .02785 L p .44211 .02848 L .248 .76 .735 r
F P
0 g
s
.44462 .02806 m .44477 .02785 L .44238 .02837 L closepath p .018 .018 .018 r
F P
0 g
s
.41651 .02806 m .41694 .02785 L .41273 .0262 L closepath p .44 .035 .035 r
F P
0 g
s
.41694 .02785 m .41651 .02806 L .42325 .02823 L closepath p .126 .022 .022 r
F P
0 g
s
.04017 .02574 m .04207 .02785 L .03854 .02644 L closepath p .914 1 1 r
F P
0 g
s
.04207 .02785 m .04017 .02574 L .04266 .02785 L closepath p .631 .979 1 r
F P
0 g
s
.04091 .02517 m .04266 .02785 L .04017 .02574 L closepath p 1 1 1 r
F P
0 g
s
.04266 .02785 m .04091 .02517 L .04325 .02785 L closepath p .717 1 1 r
F P
0 g
s
.04201 .02607 m .04325 .02785 L .04091 .02517 L closepath p .057 .261 .748 r
F P
0 g
s
.04325 .02785 m .04201 .02607 L .04696 .02785 L closepath p .564 .954 1 r
F P
0 g
s
.04576 .02806 m .04696 .02785 L .04201 .02607 L closepath p .577 .069 .069 r
F P
0 g
s
.04696 .02785 m .04576 .02806 L .05049 .02871 L closepath p .34 .046 .046 r
F P
0 g
s
.06589 .02815 m .06147 .02948 L .06179 .02902 L p .36 .846 .823 r
F P
0 g
s
.06179 .02902 m .06186 .02893 L .0623 .02884 L p .36 .846 .823 r
F P
0 g
s
.0623 .02884 m .06589 .02815 L p .06179 .02902 L .36 .846 .823 r
F P
0 g
s
.07146 .02769 m .06474 .02836 L .06457 .02804 L p .692 1 1 r
F P
0 g
s
.07015 .01963 m .06364 .00699 L .03971 .01963 L closepath p .623 .977 1 r
F P
0 g
s
.06457 .02804 m .05814 .01584 L .06979 .0262 L p .692 1 1 r
F P
0 g
s
.06979 .0262 m .07146 .02769 L p .06457 .02804 L .692 1 1 r
F P
0 g
s
.06589 .02815 m .06474 .02836 L .07134 .02788 L closepath p .023 .015 .015 r
F P
0 g
s
.07146 .02769 m .07134 .02788 L .06474 .02836 L closepath p .138 .689 .769 r
F P
0 g
s
.1646 .02823 m .16371 .02844 L .16737 .02789 L closepath p .01 .01 .01 r
F P
0 g
s
.16738 .02825 m .16737 .02789 L .16371 .02844 L closepath p .097 .035 .035 r
F P
0 g
s
.18671 .02799 m .18512 .02809 L .18501 .02781 L p .664 1 1 r
F P
0 g
s
.18501 .02781 m .18034 .01557 L .18631 .02721 L p .664 1 1 r
F P
0 g
s
.18631 .02721 m .18671 .02799 L p .18501 .02781 L .664 1 1 r
F P
0 g
s
.18552 .02792 m .18277 .02903 L .18346 .02851 L p .362 .832 .749 r
F P
0 g
s
.18346 .02851 m .18358 .02842 L .18376 .02837 L p .362 .832 .749 r
F P
0 g
s
.18376 .02837 m .18552 .02792 L p .18346 .02851 L .362 .832 .749 r
F P
0 g
s
.18512 .02809 m .18552 .02792 L .18358 .02842 L closepath p .008 .008 .008 r
F P
0 g
s
.81755 .03087 m .81998 .03042 L .82159 .02796 L closepath p .093 .093 .093 r
F P
0 g
s
.82352 .02784 m .82159 .02796 L .81998 .03042 L closepath p .092 .092 .092 r
F P
0 g
s
.57076 .02809 m .5712 .02799 L .56841 .02793 L p .084 .012 .012 r
F P
0 g
s
.56322 .007 m .56228 .01964 L .62631 .007 L closepath p .623 .977 1 r
F P
0 g
s
.57076 .02809 m .57183 .01557 L .57239 .02821 L p .431 .802 1 r
F P
0 g
s
.5712 .02799 m .57076 .02809 L .57364 .02888 L closepath p .329 .024 .024 r
F P
0 g
s
.57241 .02863 m .57364 .02888 L .57176 .02837 L p .271 .014 .014 r
F P
0 g
s
.57176 .02837 m .57076 .02809 L .57241 .02863 L p .271 .014 .014 r
F P
0 g
s
.57364 .02888 m .57241 .02863 L .57506 .03096 L closepath p .851 .393 .08 r
F P
0 g
s
.57241 .02863 m .57351 .01612 L .57322 .02896 L p .108 .159 .726 r
F P
0 g
s
.5732 .02994 m .57506 .03096 L .57287 .02904 L p .916 .479 .095 r
F P
0 g
s
.57287 .02904 m .57241 .02863 L .5732 .02994 L p .916 .479 .095 r
F P
0 g
s
.59158 .028 m .58851 .0296 L .57364 .02888 L closepath p .955 .786 .906 r
F P
0 g
s
.57506 .03096 m .57364 .02888 L .58851 .0296 L closepath p .955 .786 .906 r
F P
0 g
s
.57506 .03096 m .5732 .02994 L .57608 .03486 L closepath p 1 .753 .312 r
F P
0 g
s
.57395 .03317 m .57608 .03486 L .57351 .03047 L p 1 .788 .364 r
F P
0 g
s
.57351 .03047 m .5732 .02994 L .57395 .03317 L p 1 .788 .364 r
F P
0 g
s
.58851 .0296 m .57608 .03486 L .57506 .03096 L closepath p .955 .786 .906 r
F P
0 g
s
.57562 .03268 m .57506 .03096 L .57608 .03486 L closepath p .955 .786 .906 r
F P
0 g
s
.57562 .03268 m .57608 .03486 L .57506 .03096 L closepath p .955 .786 .906 r
F P
0 g
s
.57608 .03486 m .57395 .03317 L .57844 .05125 L closepath p 1 .854 .451 r
F P
0 g
s
.5763 .04952 m .57844 .05125 L .57464 .03595 L p 1 .854 .451 r
F P
0 g
s
.57464 .03595 m .57395 .03317 L .5763 .04952 L p 1 .854 .451 r
F P
0 g
s
.57844 .05125 m .57608 .03486 L .58801 .03124 L closepath p .955 .786 .906 r
F P
0 g
s
.57844 .05125 m .5763 .04952 L .58078 .07116 L closepath p 1 .873 .476 r
F P
0 g
s
.57864 .06942 m .58078 .07116 L .57682 .05204 L p 1 .873 .476 r
F P
0 g
s
.57682 .05204 m .5763 .04952 L .57864 .06942 L p 1 .873 .476 r
F P
0 g
s
.58078 .07116 m .57844 .05125 L .58751 .03288 L closepath p .955 .786 .906 r
F P
0 g
s
.58456 .10772 m .5836 .09856 L .58078 .07116 L closepath p .955 .786 .906 r
F P
0 g
s
.58265 .08942 m .58078 .07116 L .5836 .09856 L closepath p .955 .786 .906 r
F P
0 g
s
.58078 .07116 m .57864 .06942 L .58267 .08942 L closepath p 1 .883 .489 r
F P
0 g
s
.58048 .08741 m .58267 .08942 L .57864 .06942 L closepath p 1 .885 .494 r
F P
0 g
s
.57864 .06942 m .57984 .05749 L .58047 .08684 L p .038 .038 .209 r
F P
0 g
s
.58047 .08684 m .58048 .08741 L .57864 .06942 L p .038 .038 .209 r
F P
0 g
s
.5763 .04952 m .57746 .03731 L .5786 .0683 L p .041 .041 .229 r
F P
0 g
s
.5786 .0683 m .57864 .06942 L .5763 .04952 L p .041 .041 .229 r
F P
0 g
s
.57395 .03317 m .57507 .02072 L .57619 .04695 L p .045 .045 .264 r
F P
0 g
s
.57619 .04695 m .5763 .04952 L .57395 .03317 L p .045 .045 .264 r
F P
0 g
s
.5732 .02994 m .57431 .01745 L .57397 .03209 L p .061 .061 .376 r
F P
0 g
s
.57397 .03209 m .57395 .03317 L .5732 .02994 L p .061 .061 .376 r
F P
0 g
s
.57322 .02896 m .5732 .02994 L .57241 .02863 L p .108 .159 .726 r
F P
0 g
s
.57239 .02821 m .57241 .02863 L .57076 .02809 L p .431 .802 1 r
F P
0 g
s
.56841 .02793 m .5673 .02791 L .57076 .02809 L p .084 .012 .012 r
F P
0 g
s
.12747 .02816 m .12894 .028 L .12444 .02627 L closepath p .662 .089 .083 r
F P
0 g
s
.12894 .028 m .12747 .02816 L .13046 .02811 L closepath p .152 .026 .026 r
F P
0 g
s
.59044 .02891 m .58852 .0296 L .59008 .03083 L closepath p .141 .141 .403 r
F P
0 g
s
.59044 .02891 m .59008 .03083 L .59025 .02925 L p .755 .725 .19 r
F P
0 g
s
.59069 .02847 m .58852 .0296 L .59044 .02891 L p .027 .027 .027 r
F P
0 g
s
.59025 .02925 m .59144 .01835 L .59044 .02891 L p .755 .725 .19 r
F P
0 g
s
.59044 .02891 m .59158 .028 L .59069 .02847 L p .027 .027 .027 r
F P
0 g
s
.59217 .02819 m .59158 .028 L .59044 .02891 L closepath p .038 .038 .038 r
F P
0 g
s
.80739 .02436 m .80817 .02811 L .81309 .02087 L closepath p 1 1 1 r
F P
0 g
s
.80817 .02811 m .80739 .02436 L .80385 .03257 L closepath p 1 1 1 r
F P
0 g
s
.80385 .03257 m .80817 .02811 L .81417 .03494 L closepath p .955 .786 .906 r
F P
0 g
s
.81586 .0329 m .81417 .03494 L .80817 .02811 L closepath p .955 .786 .906 r
F P
0 g
s
.1293 .02831 m .13046 .02811 L .12747 .02816 L closepath p .169 .029 .029 r
F P
0 g
s
.13046 .02811 m .1293 .02831 L .13343 .02815 L closepath p .089 .021 .021 r
F P
0 g
s
.26501 .02834 m .26599 .02815 L .26261 .02803 L p .143 .025 .025 r
F P
0 g
s
.26599 .02815 m .26501 .02834 L .26914 .0299 L closepath p .559 .06 .06 r
F P
0 g
s
.25249 .00699 m .25618 .01963 L .29085 .00699 L closepath p .623 .977 1 r
F P
0 g
s
.26501 .02834 m .26145 .01582 L .26659 .02854 L p .217 .63 1 r
F P
0 g
s
.26692 .02934 m .26914 .0299 L .26669 .02897 L p .513 .051 .051 r
F P
0 g
s
.26669 .02897 m .26501 .02834 L .26692 .02934 L p .513 .051 .051 r
F P
0 g
s
.26914 .0299 m .26692 .02934 L .27072 .03333 L closepath p 1 .672 .272 r
F P
0 g
s
.26784 .03173 m .27072 .03333 L .2676 .03006 L p 1 .745 .375 r
F P
0 g
s
.2676 .03006 m .26692 .02934 L .26784 .03173 L p 1 .745 .375 r
F P
0 g
s
.26692 .02934 m .26338 .01684 L .26732 .02999 L p .019 .019 .325 r
F P
0 g
s
.26732 .02999 m .26784 .03173 L .26692 .02934 L p .019 .019 .325 r
F P
0 g
s
.26659 .02854 m .26692 .02934 L .26501 .02834 L p .217 .63 1 r
F P
0 g
s
.26338 .01684 m .26692 .02934 L .26145 .01582 L closepath p .217 .63 1 r
F P
0 g
s
.26432 .01926 m .26784 .03173 L .26354 .01738 L p .019 .019 .325 r
F P
0 g
s
.26354 .01738 m .26338 .01684 L .26432 .01926 L p .019 .019 .325 r
F P
0 g
s
.26432 .01926 m .26607 .0167 L .26643 .0244 L p .158 .158 .664 r
F P
0 g
s
.26784 .03173 m .26432 .01926 L .26929 .04209 L p .781 .544 .023 r
F P
0 g
s
.26693 .03503 m .27042 .04727 L .26453 .02024 L p .781 .544 .023 r
F P
0 g
s
.26453 .02024 m .26432 .01926 L .26693 .03503 L p .781 .544 .023 r
F P
0 g
s
.26929 .04209 m .27042 .04727 L .26784 .03173 L p .781 .544 .023 r
F P
0 g
s
.27072 .03333 m .26784 .03173 L .27331 .04891 L closepath p 1 .887 .559 r
F P
0 g
s
.27042 .04727 m .27331 .04891 L .26784 .03173 L closepath p 1 .887 .559 r
F P
0 g
s
.27042 .04727 m .26693 .03503 L .27296 .06628 L closepath p .807 .583 .029 r
F P
0 g
s
.26951 .05431 m .27296 .06628 L .26702 .03547 L p .807 .583 .029 r
F P
0 g
s
.26702 .03547 m .26693 .03503 L .26951 .05431 L p .807 .583 .029 r
F P
0 g
s
.27296 .06628 m .27585 .06794 L .27042 .04727 L closepath p 1 .905 .584 r
F P
0 g
s
.27296 .06628 m .26951 .05431 L .27725 .10123 L closepath p .815 .596 .031 r
F P
0 g
s
.27386 .08976 m .27725 .10123 L .27006 .05769 L p .815 .596 .031 r
F P
0 g
s
.27006 .05769 m .26951 .05431 L .27386 .08976 L p .815 .596 .031 r
F P
0 g
s
.27725 .10123 m .28014 .1028 L .27296 .06628 L closepath p 1 .911 .592 r
F P
0 g
s
.27725 .10123 m .27386 .08976 L .27764 .10547 L closepath p .836 .631 .069 r
F P
0 g
s
.27425 .09406 m .27764 .10547 L .27434 .09178 L p .836 .631 .069 r
F P
0 g
s
.27434 .09178 m .27386 .08976 L .27425 .09406 L p .836 .631 .069 r
F P
0 g
s
.27764 .10547 m .27425 .09406 L .2771 .10684 L closepath p 1 1 .624 r
F P
0 g
s
.2737 .09545 m .2771 .10684 L .27457 .09551 L p 1 1 .624 r
F P
0 g
s
.27457 .09551 m .27425 .09406 L .2737 .09545 L p 1 1 .624 r
F P
0 g
s
.27422 .09472 m .2737 .09545 L .27425 .09406 L p .086 .086 .086 r
F P
0 g
s
.27459 .09513 m .2737 .09545 L .27608 .09212 L closepath p .059 .059 .059 r
F P
0 g
s
.2737 .09545 m .27459 .09513 L .27297 .09539 L p .153 .701 .783 r
F P
0 g
s
.27297 .09539 m .27088 .09573 L .2737 .09545 L p .153 .701 .783 r
F P
0 g
s
.27425 .09406 m .27608 .09212 L .27422 .09472 L p .086 .086 .086 r
F P
0 g
s
.27386 .08976 m .27566 .08749 L .27432 .09371 L p .151 .151 .598 r
F P
0 g
s
.27432 .09371 m .27425 .09406 L .27386 .08976 L p .151 .151 .598 r
F P
0 g
s
.26951 .05431 m .2713 .05197 L .27351 .08466 L p .154 .154 .626 r
F P
0 g
s
.27351 .08466 m .27386 .08976 L .26951 .05431 L p .154 .154 .626 r
F P
0 g
s
.26693 .03503 m .26871 .03258 L .26934 .04983 L p .155 .155 .636 r
F P
0 g
s
.26934 .04983 m .26951 .05431 L .26693 .03503 L p .155 .155 .636 r
F P
0 g
s
.26643 .0244 m .26693 .03503 L .26432 .01926 L p .158 .158 .664 r
F P
0 g
s
.26261 .02803 m .25848 .0279 L .26501 .02834 L p .143 .025 .025 r
F P
0 g
s
.13246 .02836 m .13343 .02815 L .12951 .0283 L p .094 .022 .022 r
F P
0 g
s
.13343 .02815 m .13246 .02836 L .138 .02847 L closepath p .152 .027 .027 r
F P
0 g
s
.12135 .00699 m .127 .01963 L .18797 .00699 L closepath p .623 .977 1 r
F P
0 g
s
.13246 .02836 m .12689 .01584 L .13631 .02843 L p .578 .946 1 r
F P
0 g
s
.13642 .02857 m .138 .02847 L .1342 .02839 L p .107 .018 .018 r
F P
0 g
s
.1342 .02839 m .13246 .02836 L .13642 .02857 L p .107 .018 .018 r
F P
0 g
s
.138 .02847 m .13642 .02857 L .14066 .02944 L closepath p .46 .06 .06 r
F P
0 g
s
.13642 .02857 m .1309 .01606 L .13801 .0287 L p .348 .761 1 r
F P
0 g
s
.13829 .02919 m .14066 .02944 L .13796 .02889 L p .393 .047 .047 r
F P
0 g
s
.13796 .02889 m .13642 .02857 L .13829 .02919 L p .393 .047 .047 r
F P
0 g
s
.138 .02847 m .14912 .02906 L .14066 .02944 L closepath p .955 .786 .906 r
F P
0 g
s
.14066 .02944 m .13829 .02919 L .14235 .03196 L closepath p 1 .574 .202 r
F P
0 g
s
.13939 .03094 m .14235 .03196 L .13927 .02986 L p 1 .613 .251 r
F P
0 g
s
.13927 .02986 m .13829 .02919 L .13939 .03094 L p 1 .613 .251 r
F P
0 g
s
.13829 .02919 m .1328 .01668 L .13879 .02964 L p .032 .032 .49 r
F P
0 g
s
.13879 .02964 m .13939 .03094 L .13829 .02919 L p .032 .032 .49 r
F P
0 g
s
.13801 .0287 m .13829 .02919 L .13642 .02857 L p .348 .761 1 r
F P
0 g
s
.1328 .01668 m .13829 .02919 L .1309 .01606 L closepath p .348 .761 1 r
F P
0 g
s
.13392 .01846 m .13939 .03094 L .13314 .01741 L p .032 .032 .49 r
F P
0 g
s
.13314 .01741 m .1328 .01668 L .13392 .01846 L p .032 .032 .49 r
F P
0 g
s
.13392 .01846 m .1351 .0153 L .13487 .01974 L p .142 .142 .634 r
F P
0 g
s
.13939 .03094 m .13392 .01846 L .13963 .03333 L p .848 .577 .044 r
F P
0 g
s
.13473 .02234 m .14019 .03476 L .13407 .01886 L p .848 .577 .044 r
F P
0 g
s
.13407 .01886 m .13392 .01846 L .13473 .02234 L p .848 .577 .044 r
F P
0 g
s
.14019 .03476 m .13473 .02234 L .14071 .03921 L closepath p .914 .683 .144 r
F P
0 g
s
.13631 .02843 m .13642 .02857 L .13246 .02836 L p .578 .946 1 r
F P
0 g
s
.1309 .01606 m .13642 .02857 L .12689 .01584 L closepath p .578 .946 1 r
F P
0 g
s
.13525 .02685 m .14071 .03921 L .13476 .02243 L p .914 .683 .144 r
F P
0 g
s
.13476 .02243 m .13473 .02234 L .13525 .02685 L p .914 .683 .144 r
F P
0 g
s
.14071 .03921 m .13525 .02685 L .14134 .04565 L closepath p .926 .703 .168 r
F P
0 g
s
.1359 .03338 m .14134 .04565 L .13528 .02693 L p .926 .703 .168 r
F P
0 g
s
.13528 .02693 m .13525 .02685 L .1359 .03338 L p .926 .703 .168 r
F P
0 g
s
.14134 .04565 m .1359 .03338 L .1421 .05411 L closepath p .932 .713 .181 r
F P
0 g
s
.13666 .04197 m .1421 .05411 L .13592 .03346 L p .932 .713 .181 r
F P
0 g
s
.13592 .03346 m .1359 .03338 L .13666 .04197 L p .932 .713 .181 r
F P
0 g
s
.1421 .05411 m .13666 .04197 L .14298 .0646 L closepath p .936 .719 .189 r
F P
0 g
s
.13755 .0526 m .14298 .0646 L .13667 .04199 L p .936 .719 .189 r
F P
0 g
s
.13667 .04199 m .13666 .04197 L .13755 .0526 L p .936 .719 .189 r
F P
0 g
s
.1218 .063 m .13301 .08528 L .14629 .11449 L closepath p .955 .786 .906 r
F P
0 g
s
.14759 .11686 m .14629 .11449 L .13301 .08528 L closepath p .955 .786 .906 r
F P
0 g
s
.13301 .08528 m .1426 .10418 L .14759 .11686 L closepath p .955 .786 .906 r
F P
0 g
s
.14247 .10418 m .14344 .10074 L .13297 .08528 L closepath p .196 .356 .865 r
F P
0 g
s
.1341 .08217 m .13297 .08528 L .14344 .10074 L closepath p .196 .356 .865 r
F P
0 g
s
.14298 .0646 m .13755 .0526 L .14084 .07415 L p .937 .721 .191 r
F P
0 g
s
.14889 .11924 m .14759 .11686 L .1426 .10418 L closepath p .955 .786 .906 r
F P
0 g
s
.1426 .10418 m .15059 .11967 L .14889 .11924 L closepath p .955 .786 .906 r
F P
0 g
s
.14706 .11037 m .14247 .10418 L .14782 .1148 L p .197 .357 .866 r
F P
0 g
s
.14344 .10074 m .14247 .10418 L .14706 .11037 L p .197 .357 .866 r
F P
0 g
s
.15096 .11907 m .15059 .11967 L .15657 .0341 L closepath p .955 .786 .906 r
F P
0 g
s
.14621 .06633 m .14298 .0646 L .15059 .11967 L closepath p 1 .939 .658 r
F P
0 g
s
.14971 .11327 m .14298 .0646 L .14642 .10667 L p 1 .939 .658 r
F P
0 g
s
.14642 .10667 m .14084 .07415 L .14642 .10667 L .937 .721 .191 r
F
.14642 .10667 m .14344 .10074 L p .14706 .11037 L .197 .357 .866 r
F P
0 g
s
.14642 .10667 m .14642 .10667 L p .14642 .10667 L .937 .721 .191 r
F P
0 g
s
.14642 .10667 m .14298 .0646 L p .14084 .07415 L .937 .721 .191 r
F P
0 g
s
.14084 .07415 m .14084 .07415 L .14642 .10667 L .937 .721 .191 r
F
.14429 .10215 m .14369 .10117 L p .14598 .10573 L .005 .005 .337 r
F P
0 g
s
.14369 .10117 m .14344 .10074 L .14642 .10667 L p .005 .005 .337 r
F P
0 g
s
.14715 .1155 m .14642 .10667 L p .14642 .10667 L .937 .721 .191 r
F P
0 g
s
.14084 .07415 m .1455 .10472 L p .14642 .10667 L .937 .721 .191 r
F P
0 g
s
.14598 .10573 m .14369 .10117 L .14642 .10667 L .005 .005 .337 r
F
.1455 .10472 m .14715 .1155 L p .14642 .10667 L .937 .721 .191 r
F P
0 g
s
.14642 .10667 m .15091 .11555 L p .14598 .10573 L .005 .005 .337 r
F P
0 g
s
.1341 .08217 m .14344 .10074 L .13897 .09131 L p .005 .005 .335 r
F P
0 g
s
.13897 .09131 m .138 .08927 L .13646 .08646 L p .005 .005 .335 r
F P
0 g
s
.13646 .08646 m .1341 .08217 L p .13897 .09131 L .005 .005 .335 r
F P
0 g
s
.14344 .10074 m .14429 .10215 L p .13944 .09221 L .005 .005 .337 r
F P
0 g
s
.13897 .09131 m .14344 .10074 L p .13944 .09221 L .005 .005 .337 r
F P
0 g
s
.14103 .09527 m .138 .08927 L p .13944 .09221 L .005 .005 .337 r
F P
0 g
s
.13944 .09221 m .14429 .10215 L .14103 .09527 L .005 .005 .337 r
F
.1455 .10472 m .14215 .08271 L p .14103 .09527 L .937 .721 .191 r
F P
0 g
s
.14215 .08271 m .13755 .0526 L .14103 .09527 L p .937 .721 .191 r
F P
0 g
s
.14103 .09527 m .14177 .10424 L p .1455 .10472 L .937 .721 .191 r
F P
0 g
s
.14167 .10235 m .14177 .10424 L .14103 .09527 L p .132 .132 .528 r
F P
0 g
s
.14429 .10215 m .14558 .10429 L p .14103 .09527 L .005 .005 .337 r
F P
0 g
s
.15091 .11555 m .14558 .10429 L .14429 .10215 L p .005 .005 .337 r
F P
0 g
s
.15091 .11555 m .14429 .10215 L .14598 .10573 L .005 .005 .337 r
F
.14103 .09527 m .14558 .10429 L p .14205 .09941 L 1 .7 .355 r
F P
0 g
s
.14177 .10424 m .14715 .1155 L .1455 .10472 L p .937 .721 .191 r
F P
0 g
s
.15059 .11967 m .14971 .11327 L p .14642 .10667 L 1 .939 .658 r
F P
0 g
s
.14706 .11037 m .14792 .11153 L p .14642 .10667 L .197 .357 .866 r
F P
0 g
s
.1502 .12162 m .14889 .11924 L .15059 .11967 L closepath p .955 .786 .906 r
F P
0 g
s
.15028 .11967 m .15033 .11935 L p .15033 .11935 L .197 .357 .866 r
F P
0 g
s
.15028 .11967 m .15033 .11935 L .14706 .11037 L .197 .357 .866 r
F
.15028 .11967 m .15059 .11967 L .15033 .11935 L p .832 1 1 r
F P
0 g
s
.14782 .1148 m .15028 .11967 L p .14706 .11037 L .197 .357 .866 r
F P
0 g
s
.15033 .11935 m .14715 .1155 L .15028 .11967 L p .832 1 1 r
F P
0 g
s
.14792 .11153 m .14706 .11037 L p .15033 .11935 L .197 .357 .866 r
F P
0 g
s
.14715 .1155 m .15033 .11935 L p .14792 .11153 L 1 .939 .658 r
F P
0 g
s
.14642 .10667 m .14642 .10667 L p .14792 .11153 L 1 .939 .658 r
F P
0 g
s
.14642 .10667 m .14715 .1155 L p .14792 .11153 L 1 .939 .658 r
F P
0 g
s
.15033 .11935 m .15059 .11967 L p .14642 .10667 L 1 .939 .658 r
F P
0 g
s
.15033 .11935 m .15091 .11555 L .14792 .11153 L p .197 .357 .866 r
F P
0 g
s
.14792 .11153 m .15091 .11555 L .14642 .10667 L p .197 .357 .866 r
F P
0 g
s
.14558 .10429 m .15091 .11555 L .14177 .10424 L closepath p .566 .952 1 r
F P
0 g
s
.14558 .10429 m .14315 .10469 L .14315 .10469 L p 1 .7 .355 r
F P
0 g
s
.14558 .10429 m .14177 .10424 L .14333 .10466 L p .161 .04 .04 r
F P
0 g
s
.14333 .10466 m .14177 .10424 L .14138 .09669 L p .132 .132 .528 r
F P
0 g
s
.14111 .09558 m .14167 .10235 L .14103 .09527 L .132 .132 .528 r
F
.14205 .09941 m .14138 .09669 L .14103 .09527 L 1 .7 .355 r
F
.14134 .09589 m .14167 .10235 L p .1411 .09541 L .132 .132 .528 r
F P
0 g
s
.14315 .10469 m .14082 .09771 L p .14205 .09941 L 1 .7 .355 r
F P
0 g
s
.14558 .10429 m .14333 .10466 L .14205 .09941 L 1 .7 .355 r
F
.14333 .10466 m .14315 .10469 L .14205 .09941 L F
.14315 .10469 m .14558 .10429 L .14346 .10469 L closepath p .161 .04 .04 r
F P
0 g
s
.14333 .10466 m .14346 .10469 L .14558 .10429 L p .161 .04 .04 r
F P
0 g
s
.13902 .05033 m .14346 .10469 L .14333 .10466 L p .132 .132 .528 r
F P
0 g
s
.14138 .09669 m .13902 .05033 L p .14333 .10466 L .132 .132 .528 r
F P
0 g
s
.14205 .09941 m .14205 .09941 L .14138 .09669 L 1 .7 .355 r
F
.14109 .09539 m .14111 .09558 L .14103 .09527 L .132 .132 .528 r
F
.14205 .09941 m .14082 .09771 L .14103 .09527 L 1 .7 .355 r
F
.138 .08927 m .13897 .09131 L p .13944 .09221 L .005 .005 .337 r
F P
0 g
s
.14082 .09771 m .138 .08927 L .14103 .09527 L p 1 .7 .355 r
F P
0 g
s
.14109 .09539 m .14134 .09589 L .1411 .09541 L .132 .132 .528 r
F
.13755 .0526 m .14134 .09586 L .14109 .09539 L F
.14134 .09586 m .14134 .09589 L .14109 .09539 L F
.14103 .09527 m .13755 .0526 L p .14109 .09539 L F P
0 g
s
.13993 .06819 m .14134 .09589 L p .13755 .0526 L .132 .132 .528 r
F P
0 g
s
.13755 .0526 m .13902 .05033 L .13993 .06819 L p .132 .132 .528 r
F P
0 g
s
.13666 .04197 m .13813 .03965 L .13757 .05228 L p .132 .132 .529 r
F P
0 g
s
.13757 .05228 m .13755 .0526 L .13666 .04197 L p .132 .132 .529 r
F P
0 g
s
.1359 .03338 m .13736 .03102 L .13669 .04161 L p .133 .133 .535 r
F P
0 g
s
.13669 .04161 m .13666 .04197 L .1359 .03338 L p .133 .133 .535 r
F P
0 g
s
.13525 .02685 m .13671 .02443 L .13594 .03297 L p .134 .134 .545 r
F P
0 g
s
.13594 .03297 m .1359 .03338 L .13525 .02685 L p .134 .134 .545 r
F P
0 g
s
.13473 .02234 m .13617 .01988 L .13539 .02583 L p .136 .136 .564 r
F P
0 g
s
.13539 .02583 m .13525 .02685 L .13473 .02234 L p .136 .136 .564 r
F P
0 g
s
.13487 .01974 m .13473 .02234 L .13392 .01846 L p .142 .142 .634 r
F P
0 g
s
.13963 .03333 m .14019 .03476 L .13939 .03094 L p .848 .577 .044 r
F P
0 g
s
.12951 .0283 m .1293 .02831 L .13246 .02836 L p .094 .022 .022 r
F P
0 g
s
.06474 .02836 m .06589 .02815 L .06186 .02893 L closepath p .143 .683 .7 r
F P
0 g
s
.16114 .0286 m .15865 .0297 L .1646 .02823 L closepath p .955 .786 .906 r
F P
0 g
s
.16114 .0286 m .1646 .02823 L .15865 .0297 L closepath p .955 .786 .906 r
F P
0 g
s
.16738 .02825 m .16371 .02844 L .16356 .02807 L p .655 1 1 r
F P
0 g
s
.16356 .02807 m .15861 .01593 L .16662 .02719 L p .655 1 1 r
F P
0 g
s
.16662 .02719 m .16738 .02825 L p .16356 .02807 L .655 1 1 r
F P
0 g
s
.1646 .02823 m .15865 .0297 L .15904 .02936 L p .298 .805 .809 r
F P
0 g
s
.15904 .02936 m .15923 .0292 L .16024 .02902 L p .298 .805 .809 r
F P
0 g
s
.16024 .02902 m .1646 .02823 L p .15904 .02936 L .298 .805 .809 r
F P
0 g
s
.16371 .02844 m .1646 .02823 L .15923 .0292 L closepath p .006 .006 .006 r
F P
0 g
s
.42266 .02842 m .42325 .02823 L .42025 .02815 L p .125 .022 .022 r
F P
0 g
s
.42624 .02997 m .42325 .02823 L .44103 .0289 L closepath p .955 .786 .906 r
F P
0 g
s
.4131 .00699 m .41439 .01963 L .48077 .00699 L closepath p .623 .977 1 r
F P
0 g
s
.42266 .02842 m .42149 .01591 L .42424 .0286 L p .248 .638 1 r
F P
0 g
s
.42325 .02823 m .42266 .02842 L .42624 .02997 L closepath p .512 .043 .043 r
F P
0 g
s
.42441 .0294 m .42624 .02997 L .42403 .02901 L p .481 .037 .037 r
F P
0 g
s
.42403 .02901 m .42266 .02842 L .42441 .0294 L p .481 .037 .037 r
F P
0 g
s
.42624 .02997 m .42441 .0294 L .4278 .03334 L closepath p 1 .62 .178 r
F P
0 g
s
.42441 .0294 m .42326 .01689 L .42509 .03008 L p .054 .054 .447 r
F P
0 g
s
.42531 .03174 m .4278 .03334 L .425 .03008 L p 1 .707 .295 r
F P
0 g
s
.425 .03008 m .42441 .0294 L .42531 .03174 L p 1 .707 .295 r
F P
0 g
s
.4278 .03334 m .42531 .03174 L .42809 .03499 L closepath p 1 .862 .492 r
F P
0 g
s
.4256 .03337 m .42809 .03499 L .42534 .03177 L p 1 .862 .493 r
F P
0 g
s
.42534 .03177 m .42531 .03174 L .4256 .03337 L p 1 .862 .493 r
F P
0 g
s
.42809 .03499 m .4256 .03337 L .42856 .038 L closepath p 1 .872 .506 r
F P
0 g
s
.42607 .03637 m .42856 .038 L .42562 .0334 L p 1 .873 .506 r
F P
0 g
s
.42562 .0334 m .4256 .03337 L .42607 .03637 L p 1 .873 .506 r
F P
0 g
s
.42856 .038 m .42607 .03637 L .42922 .04237 L closepath p 1 .876 .511 r
F P
0 g
s
.42673 .04073 m .42922 .04237 L .42609 .0364 L p 1 .876 .511 r
F P
0 g
s
.42609 .0364 m .42607 .03637 L .42673 .04073 L p 1 .876 .511 r
F P
0 g
s
.42922 .04237 m .42673 .04073 L .43008 .0481 L closepath p 1 .878 .514 r
F P
0 g
s
.42758 .04646 m .43008 .0481 L .42674 .04076 L p 1 .878 .514 r
F P
0 g
s
.42674 .04076 m .42673 .04073 L .42758 .04646 L p 1 .878 .514 r
F P
0 g
s
.43008 .0481 m .42758 .04646 L .43112 .0552 L closepath p 1 .879 .516 r
F P
0 g
s
.42862 .05356 m .43112 .0552 L .42759 .04649 L p 1 .879 .516 r
F P
0 g
s
.42759 .04649 m .42758 .04646 L .42862 .05356 L p 1 .879 .516 r
F P
0 g
s
.43112 .0552 m .42862 .05356 L .43235 .06367 L closepath p 1 .88 .517 r
F P
0 g
s
.42985 .06204 m .43235 .06367 L .42863 .05359 L p 1 .88 .517 r
F P
0 g
s
.42863 .05359 m .42862 .05356 L .42985 .06204 L p 1 .88 .517 r
F P
0 g
s
.43235 .06367 m .42985 .06204 L .43377 .07352 L closepath p 1 .881 .518 r
F P
0 g
s
.43127 .0719 m .43377 .07352 L .42986 .06207 L p 1 .881 .518 r
F P
0 g
s
.42986 .06207 m .42985 .06204 L .43127 .0719 L p 1 .881 .518 r
F P
0 g
s
.43377 .07352 m .43127 .0719 L .43538 .08475 L closepath p 1 .881 .518 r
F P
0 g
s
.43288 .08315 m .43538 .08475 L .43134 .07212 L p 1 .881 .518 r
F P
0 g
s
.43134 .07212 m .43127 .0719 L .43288 .08315 L p 1 .881 .518 r
F P
0 g
s
.43538 .08475 m .43288 .08315 L .43652 .09292 L closepath p 1 .884 .522 r
F P
0 g
s
.43402 .0913 m .43652 .09292 L .43299 .08344 L p 1 .884 .521 r
F P
0 g
s
.43299 .08344 m .43288 .08315 L .43402 .0913 L p 1 .884 .521 r
F P
0 g
s
.43652 .09292 m .43402 .0913 L .43734 .09916 L closepath p 1 .889 .529 r
F P
0 g
s
.43483 .09748 m .43734 .09916 L .43416 .09163 L p 1 .889 .528 r
F P
0 g
s
.43416 .09163 m .43402 .0913 L .43483 .09748 L p 1 .889 .528 r
F P
0 g
s
.43734 .09916 m .43483 .09748 L .43782 .10346 L closepath p 1 .9 .543 r
F P
0 g
s
.43531 .10169 m .43782 .10346 L .43506 .09793 L p 1 .899 .542 r
F P
0 g
s
.43506 .09793 m .43483 .09748 L .43531 .10169 L p 1 .899 .542 r
F P
0 g
s
.43433 .09022 m .43531 .10169 L .43386 .08616 L p .005 .005 .099 r
F P
0 g
s
.43386 .08616 m .43384 .08595 L .43433 .09022 L p .005 .005 .099 r
F P
0 g
s
.43538 .08475 m .43977 .03206 L .43652 .09292 L closepath p .955 .786 .906 r
F P
0 g
s
.43433 .09022 m .43656 .08817 L .43457 .09233 L p .166 .166 .639 r
F P
0 g
s
.43457 .09233 m .43448 .09251 L .43433 .09022 L p .166 .166 .639 r
F P
0 g
s
.43531 .10169 m .43433 .09022 L .43546 .10385 L p .717 .546 .003 r
F P
0 g
s
.43546 .10385 m .43547 .10395 L .43531 .10169 L p .717 .546 .003 r
F P
0 g
s
.43483 .09748 m .43384 .08595 L .4353 .10161 L p .005 .005 .099 r
F P
0 g
s
.4353 .10161 m .43531 .10169 L .43483 .09748 L p .005 .005 .099 r
F P
0 g
s
.43402 .0913 m .43302 .07969 L .43482 .09741 L p .008 .008 .121 r
F P
0 g
s
.43482 .09741 m .43483 .09748 L .43402 .0913 L p .008 .008 .121 r
F P
0 g
s
.43288 .08315 m .43186 .07142 L .43402 .09126 L p .01 .01 .133 r
F P
0 g
s
.43402 .09126 m .43402 .0913 L .43288 .08315 L p .01 .01 .133 r
F P
0 g
s
.43127 .0719 m .43023 .06001 L .43288 .08314 L p .01 .01 .138 r
F P
0 g
s
.43288 .08314 m .43288 .08315 L .43127 .0719 L p .01 .01 .138 r
F P
0 g
s
.42985 .06204 m .42878 .05 L .43127 .07189 L p .011 .011 .139 r
F P
0 g
s
.43127 .07189 m .43127 .0719 L .42985 .06204 L p .011 .011 .139 r
F P
0 g
s
.42862 .05356 m .42753 .0414 L .42985 .06202 L p .011 .011 .14 r
F P
0 g
s
.42985 .06202 m .42985 .06204 L .42862 .05356 L p .011 .011 .14 r
F P
0 g
s
.42758 .04646 m .42648 .0342 L .42862 .05354 L p .011 .011 .142 r
F P
0 g
s
.42862 .05354 m .42862 .05356 L .42758 .04646 L p .011 .011 .142 r
F P
0 g
s
.42673 .04073 m .42562 .02839 L .42758 .04643 L p .011 .011 .145 r
F P
0 g
s
.42758 .04643 m .42758 .04646 L .42673 .04073 L p .011 .011 .145 r
F P
0 g
s
.42607 .03637 m .42495 .02396 L .42673 .0407 L p .012 .012 .149 r
F P
0 g
s
.42673 .0407 m .42673 .04073 L .42607 .03637 L p .012 .012 .149 r
F P
0 g
s
.4256 .03337 m .42447 .02093 L .42606 .03632 L p .013 .013 .157 r
F P
0 g
s
.42606 .03632 m .42607 .03637 L .4256 .03337 L p .013 .013 .157 r
F P
0 g
s
.42531 .03174 m .42418 .01927 L .42556 .03305 L p .016 .016 .179 r
F P
0 g
s
.42556 .03305 m .4256 .03337 L .42531 .03174 L p .016 .016 .179 r
F P
0 g
s
.42509 .03008 m .42531 .03174 L .42441 .0294 L p .054 .054 .447 r
F P
0 g
s
.42424 .0286 m .42441 .0294 L .42266 .02842 L p .248 .638 1 r
F P
0 g
s
.42025 .02815 m .41651 .02806 L .42266 .02842 L p .125 .022 .022 r
F P
0 g
s
.53384 .02603 m .53581 .02827 L .53474 .02181 L closepath p 1 1 .733 r
F P
0 g
s
.53581 .02827 m .53384 .02603 L .53518 .03257 L closepath p 1 1 .684 r
F P
0 g
s
.53581 .02827 m .53563 .0304 L .53518 .03257 L closepath p .955 .786 .906 r
F P
0 g
s
.86076 .0464 m .85721 .0421 L .85691 .02842 L closepath p .955 .786 .906 r
F P
0 g
s
.85691 .02842 m .85792 .0296 L .86076 .0464 L closepath p .955 .786 .906 r
F P
0 g
s
.54584 .03585 m .54806 .03281 L .54219 .02854 L closepath p .38 .474 .925 r
F P
0 g
s
.5444 .02549 m .54219 .02854 L .54806 .03281 L closepath p .38 .474 .925 r
F P
0 g
s
.60444 .02864 m .60492 .02854 L .60261 .02813 L p .222 .02 .02 r
F P
0 g
s
.61457 .02859 m .61374 .02986 L .60492 .02854 L closepath p .955 .786 .906 r
F P
0 g
s
.60703 .03042 m .60492 .02854 L .61374 .02986 L closepath p .955 .786 .906 r
F P
0 g
s
.60444 .02864 m .60603 .01613 L .60564 .02898 L p .135 .405 .911 r
F P
0 g
s
.60492 .02854 m .60444 .02864 L .60703 .03042 L closepath p .609 .076 .038 r
F P
0 g
s
.60561 .02988 m .60703 .03042 L .60483 .02891 L p .727 .216 .061 r
F P
0 g
s
.60483 .02891 m .60444 .02864 L .60561 .02988 L p .727 .216 .061 r
F P
0 g
s
.61374 .02986 m .61291 .03113 L .60703 .03042 L closepath p .955 .786 .906 r
F P
0 g
s
.60703 .03042 m .60561 .02988 L .60938 .03425 L closepath p .88 .439 .084 r
F P
0 g
s
.60938 .03425 m .60703 .03042 L .61291 .03113 L closepath p .955 .786 .906 r
F P
0 g
s
.60754 .03318 m .60938 .03425 L .60604 .03037 L p .917 .484 .094 r
F P
0 g
s
.60604 .03037 m .60561 .02988 L .60754 .03318 L p .917 .484 .094 r
F P
0 g
s
.61291 .03113 m .61357 .03454 L .60938 .03425 L closepath p .955 .786 .906 r
F P
0 g
s
.60938 .03425 m .60754 .03318 L .61465 .04509 L closepath p .968 .567 .104 r
F P
0 g
s
.60938 .03425 m .62361 .06312 L .61465 .04509 L closepath p .955 .786 .906 r
F P
0 g
s
.61649 .04867 m .60938 .03425 L .61357 .03454 L closepath p .955 .786 .906 r
F P
0 g
s
.61282 .04404 m .61465 .04509 L .60754 .03318 L closepath p .968 .567 .104 r
F P
0 g
s
.60754 .03318 m .60917 .02074 L .61259 .04261 L p .102 .102 .657 r
F P
0 g
s
.61259 .04261 m .61282 .04404 L .60754 .03318 L p .102 .102 .657 r
F P
0 g
s
.60561 .02988 m .60722 .01739 L .60752 .03225 L p .112 .164 .731 r
F P
0 g
s
.60752 .03225 m .60754 .03318 L .60561 .02988 L p .112 .164 .731 r
F P
0 g
s
.60564 .02898 m .60561 .02988 L .60444 .02864 L p .135 .405 .911 r
F P
0 g
s
.60261 .02813 m .59895 .02748 L .60444 .02864 L p .222 .02 .02 r
F P
0 g
s
.54402 .03219 m .54493 .03402 L .54219 .02854 L closepath p .955 .786 .906 r
F P
0 g
s
.54402 .03219 m .54219 .02854 L .54493 .03402 L closepath p .955 .786 .906 r
F P
0 g
s
.61291 .03113 m .61554 .02908 L .61457 .02859 L closepath p .096 .096 .096 r
F P
0 g
s
.22324 .02816 m .21904 .01564 L .22319 .02798 L p .003 .003 .243 r
F P
0 g
s
.22324 .02816 m .22541 .02864 L .22328 .02834 L p .063 .629 .817 r
F P
0 g
s
.22361 .02923 m .22664 .03127 L .22382 .02869 L p 1 .772 .407 r
F P
0 g
s
.22382 .02869 m .22324 .02816 L .22361 .02923 L p 1 .772 .407 r
F P
0 g
s
.22319 .02798 m .22361 .02923 L .22324 .02816 L p .003 .003 .243 r
F P
0 g
s
.22328 .02834 m .22084 .02798 L .22324 .02816 L p .063 .629 .817 r
F P
0 g
s
.22541 .02864 m .22324 .02816 L .22664 .03127 L closepath p 1 .665 .263 r
F P
0 g
s
.21904 .01564 m .22324 .02816 L .21661 .01546 L closepath p .563 .934 1 r
F P
0 g
s
.21942 .01672 m .22361 .02923 L .2192 .0161 L p .003 .003 .243 r
F P
0 g
s
.2192 .0161 m .21904 .01564 L .21942 .01672 L p .003 .003 .243 r
F P
0 g
s
.21942 .01672 m .2211 .0146 L .22001 .01835 L p .119 .119 .369 r
F P
0 g
s
.22361 .02923 m .21942 .01672 L .22249 .03209 L p .981 .86 .369 r
F P
0 g
s
.21661 .01546 m .22084 .02798 L .21333 .01543 L closepath p .613 .971 1 r
F P
0 g
s
.21761 .02795 m .21333 .01543 L .22084 .02798 L closepath p .613 .971 1 r
F P
0 g
s
.21883 .02239 m .22303 .03482 L .21959 .01758 L p .981 .86 .369 r
F P
0 g
s
.21959 .01758 m .21942 .01672 L .21883 .02239 L p .981 .86 .369 r
F P
0 g
s
.22303 .03482 m .21883 .02239 L .22145 .04169 L closepath p 1 .955 .513 r
F P
0 g
s
.21598 .02786 m .21167 .01534 L .21761 .02795 L closepath p .579 .946 1 r
F P
0 g
s
.21333 .01543 m .21761 .02795 L .21167 .01534 L closepath p .579 .946 1 r
F P
0 g
s
.21722 .02936 m .22145 .04169 L .21888 .02275 L p 1 .955 .513 r
F P
0 g
s
.21888 .02275 m .21883 .02239 L .21722 .02936 L p 1 .955 .513 r
F P
0 g
s
.22145 .04169 m .21722 .02936 L .21978 .0479 L closepath p 1 .98 .553 r
F P
0 g
s
.21553 .03566 m .21978 .0479 L .21723 .02941 L p 1 .98 .553 r
F P
0 g
s
.21723 .02941 m .21722 .02936 L .21553 .03566 L p 1 .98 .553 r
F P
0 g
s
.21978 .0479 m .21553 .03566 L .21843 .05282 L closepath p 1 .983 .559 r
F P
0 g
s
.21416 .04065 m .21843 .05282 L .21631 .04026 L p 1 .983 .559 r
F P
0 g
s
.21631 .04026 m .21553 .03566 L .21416 .04065 L p 1 .983 .559 r
F P
0 g
s
.21843 .05282 m .21416 .04065 L .19578 .05282 L closepath p .607 .971 1 r
F P
0 g
s
.19116 .04065 m .19578 .05282 L .21416 .04065 L closepath p .607 .971 1 r
F P
0 g
s
.21416 .04065 m .21525 .0409 L .21293 .04087 L p .088 .023 .023 r
F P
0 g
s
.21293 .04087 m .19116 .04065 L .21416 .04065 L p .088 .023 .023 r
F P
0 g
s
.21553 .03566 m .21704 .03435 L .21419 .04057 L p .096 .096 .175 r
F P
0 g
s
.21419 .04057 m .21416 .04065 L .21553 .03566 L p .096 .096 .175 r
F P
0 g
s
.21722 .02936 m .21882 .02781 L .21583 .03494 L p .098 .098 .187 r
F P
0 g
s
.21583 .03494 m .21553 .03566 L .21722 .02936 L p .098 .098 .187 r
F P
0 g
s
.21883 .02239 m .22048 .02058 L .2181 .027 L p .102 .102 .227 r
F P
0 g
s
.2181 .027 m .21722 .02936 L .21883 .02239 L p .102 .102 .227 r
F P
0 g
s
.22001 .01835 m .21883 .02239 L .21942 .01672 L p .119 .119 .369 r
F P
0 g
s
.22249 .03209 m .22303 .03482 L .22361 .02923 L p .981 .86 .369 r
F P
0 g
s
.22664 .03127 m .22541 .02864 L .23914 .03379 L closepath p .955 .786 .906 r
F P
0 g
s
.22704 .02887 m .23914 .03379 L .22541 .02864 L closepath p .955 .786 .906 r
F P
0 g
s
.34137 .02865 m .32953 .03334 L .32679 .02922 L closepath p .955 .786 .906 r
F P
0 g
s
.91875 .02817 m .91911 .02865 L .91634 .02798 L closepath p .115 .668 .862 r
F P
0 g
s
.91911 .02865 m .91875 .02817 L .92047 .03128 L closepath p .665 .278 .01 r
F P
0 g
s
.68048 .01964 m .68321 .007 L .64487 .01964 L closepath p .623 .977 1 r
F P
0 g
s
.6553 .02817 m .65766 .01565 L .65588 .02799 L p .099 .099 .591 r
F P
0 g
s
.6553 .02817 m .65635 .02865 L .65474 .02834 L p .095 .654 .846 r
F P
0 g
s
.65635 .02865 m .6553 .02817 L .65766 .03128 L closepath p .862 .459 .067 r
F P
0 g
s
.65571 .02923 m .65766 .03128 L .6557 .0287 L p .976 .612 .126 r
F P
0 g
s
.6557 .0287 m .6553 .02817 L .65571 .02923 L p .976 .612 .126 r
F P
0 g
s
.65766 .03128 m .65571 .02923 L .65725 .03715 L closepath p 1 .977 .613 r
F P
0 g
s
.65532 .03483 m .65725 .03715 L .65605 .03098 L p 1 .977 .614 r
F P
0 g
s
.65605 .03098 m .65571 .02923 L .65532 .03483 L p 1 .977 .614 r
F P
0 g
s
.65766 .03128 m .65635 .02865 L .67032 .03379 L closepath p .955 .786 .906 r
F P
0 g
s
.65815 .02888 m .67032 .03379 L .65635 .02865 L closepath p .955 .786 .906 r
F P
0 g
s
.65767 .0224 m .65532 .03483 L .65794 .01755 L p .024 .024 .037 r
F P
0 g
s
.65794 .01755 m .65807 .01673 L .65767 .0224 L p .024 .024 .037 r
F P
0 g
s
.65767 .0224 m .66044 .02059 L .65741 .02708 L p .166 .166 .503 r
F P
0 g
s
.65741 .02708 m .65635 .02936 L .65767 .0224 L p .166 .166 .503 r
F P
0 g
s
.65532 .03483 m .65767 .0224 L .65421 .04063 L p .695 .676 .133 r
F P
0 g
s
.65421 .04063 m .65401 .04169 L .65532 .03483 L p .695 .676 .133 r
F P
0 g
s
.65571 .02923 m .65807 .01673 L .65573 .03212 L p .024 .024 .037 r
F P
0 g
s
.65573 .03212 m .65532 .03483 L .65571 .02923 L p .024 .024 .037 r
F P
0 g
s
.65588 .02799 m .65571 .02923 L .6553 .02817 L p .099 .099 .591 r
F P
0 g
s
.65474 .02834 m .6529 .02798 L .6553 .02817 L p .095 .654 .846 r
F P
0 g
s
.17463 .02937 m .18329 .02866 L .18277 .02902 L closepath p .955 .786 .906 r
F P
0 g
s
.04818 .02856 m .05049 .02871 L .04767 .02833 L p .255 .03 .03 r
F P
0 g
s
.05049 .02871 m .04818 .02856 L .05301 .03129 L closepath p .907 .406 .126 r
F P
0 g
s
.03274 .00699 m .03971 .01963 L .06364 .00699 L closepath p .623 .977 1 r
F P
0 g
s
.04818 .02856 m .04133 .01605 L .04896 .02892 L p .052 .126 .659 r
F P
0 g
s
.04982 .03038 m .05301 .03129 L .04948 .0293 L p .951 .466 .138 r
F P
0 g
s
.04948 .0293 m .04818 .02856 L .04982 .03038 L p .951 .466 .138 r
F P
0 g
s
.04896 .02892 m .04982 .03038 L .04818 .02856 L p .052 .126 .659 r
F P
0 g
s
.043 .01789 m .04982 .03038 L .04183 .01689 L p .052 .126 .659 r
F P
0 g
s
.04183 .01689 m .04133 .01605 L .043 .01789 L p .052 .126 .659 r
F P
0 g
s
.043 .01789 m .04395 .01462 L .04388 .01864 L p .156 .232 .773 r
F P
0 g
s
.04982 .03038 m .043 .01789 L .05025 .03171 L p .759 .41 .028 r
F P
0 g
s
.04386 .02006 m .05067 .03252 L .04309 .01806 L p .759 .41 .028 r
F P
0 g
s
.04309 .01806 m .043 .01789 L .04386 .02006 L p .759 .41 .028 r
F P
0 g
s
.04133 .01605 m .04818 .02856 L .03888 .01554 L closepath p .444 .845 1 r
F P
0 g
s
.04767 .02833 m .04576 .02806 L .04818 .02856 L p .255 .03 .03 r
F P
0 g
s
.04576 .02806 m .03888 .01554 L .04818 .02856 L closepath p .444 .845 1 r
F P
0 g
s
.05067 .03252 m .04386 .02006 L .05203 .03683 L closepath p .828 .507 .043 r
F P
0 g
s
.04523 .02444 m .05203 .03683 L .0439 .02014 L p .828 .507 .043 r
F P
0 g
s
.0439 .02014 m .04386 .02006 L .04523 .02444 L p .828 .507 .043 r
F P
0 g
s
.05203 .03683 m .04523 .02444 L .06415 .07754 L closepath p .841 .527 .047 r
F P
0 g
s
.05752 .06573 m .06415 .07754 L .04576 .02593 L p .841 .527 .047 r
F P
0 g
s
.04576 .02593 m .04523 .02444 L .05752 .06573 L p .841 .527 .047 r
F P
0 g
s
.06415 .07754 m .05752 .06573 L .0684 .09233 L closepath p .849 .539 .048 r
F P
0 g
s
.06183 .08073 m .0684 .09233 L .05765 .06605 L p .849 .539 .048 r
F P
0 g
s
.05765 .06605 m .05752 .06573 L .06183 .08073 L p .849 .539 .048 r
F P
0 g
s
.0684 .09233 m .06183 .08073 L .07141 .10326 L closepath p .859 .553 .051 r
F P
0 g
s
.06488 .09182 m .07141 .10326 L .062 .08113 L p .859 .553 .051 r
F P
0 g
s
.062 .08113 m .06183 .08073 L .06488 .09182 L p .859 .553 .051 r
F P
0 g
s
.07141 .10326 m .06488 .09182 L .07319 .11035 L closepath p .876 .58 .055 r
F P
0 g
s
.06668 .09901 m .07319 .11035 L .06519 .09252 L p .876 .58 .055 r
F P
0 g
s
.06519 .09252 m .06488 .09182 L .06668 .09901 L p .876 .58 .055 r
F P
0 g
s
.07319 .11035 m .06668 .09901 L .07375 .1136 L closepath p .929 .666 .133 r
F P
0 g
s
.06725 .10231 m .07375 .1136 L .06722 .10012 L p .929 .666 .133 r
F P
0 g
s
.06722 .10012 m .06668 .09901 L .06725 .10231 L p .929 .666 .133 r
F P
0 g
s
.07375 .1136 m .06725 .10231 L .07303 .11763 L closepath p 1 .967 .541 r
F P
0 g
s
.06651 .10639 m .07303 .11763 L .06828 .10503 L p 1 .967 .541 r
F P
0 g
s
.06828 .10503 m .06725 .10231 L .06651 .10639 L p 1 .967 .541 r
F P
0 g
s
.06651 .10639 m .06732 .10568 L .06572 .10701 L p .462 .826 .565 r
F P
0 g
s
.06572 .10701 m .06433 .10817 L .06651 .10639 L p .462 .826 .565 r
F P
0 g
s
.06725 .10231 m .06855 .10039 L .06728 .10414 L p .083 .083 .181 r
F P
0 g
s
.06728 .10414 m .06651 .10639 L .06725 .10231 L p .083 .083 .181 r
F P
0 g
s
.06668 .09901 m .06797 .09684 L .06732 .10179 L p .131 .131 .58 r
F P
0 g
s
.06732 .10179 m .06725 .10231 L .06668 .09901 L p .131 .131 .58 r
F P
0 g
s
.06488 .09182 m .06614 .08942 L .06665 .09838 L p .141 .141 .655 r
F P
0 g
s
.06665 .09838 m .06668 .09901 L .06488 .09182 L p .141 .141 .655 r
F P
0 g
s
.06183 .08073 m .06304 .07814 L .06478 .09112 L p .144 .144 .678 r
F P
0 g
s
.06478 .09112 m .06488 .09182 L .06183 .08073 L p .144 .144 .678 r
F P
0 g
s
.05752 .06573 m .0587 .06303 L .06165 .0797 L p .146 .146 .689 r
F P
0 g
s
.06165 .0797 m .06183 .08073 L .05752 .06573 L p .146 .146 .689 r
F P
0 g
s
.04523 .02444 m .04638 .02164 L .0551 .05617 L p .147 .147 .698 r
F P
0 g
s
.0551 .05617 m .05752 .06573 L .04523 .02444 L p .147 .147 .698 r
F P
0 g
s
.04386 .02006 m .04494 .01708 L .0452 .02356 L p .148 .158 .711 r
F P
0 g
s
.0452 .02356 m .04523 .02444 L .04386 .02006 L p .148 .158 .711 r
F P
0 g
s
.04388 .01864 m .04386 .02006 L .043 .01789 L p .156 .232 .773 r
F P
0 g
s
.05025 .03171 m .05067 .03252 L .04982 .03038 L p .759 .41 .028 r
F P
0 g
s
.09431 .02608 m .0975 .0288 L .09561 .02183 L closepath p 1 1 .877 r
F P
0 g
s
.0975 .0288 m .09431 .02608 L .09659 .0338 L closepath p 1 1 .818 r
F P
0 g
s
.0975 .0288 m .0972 .03129 L .09659 .0338 L closepath p .955 .786 .906 r
F P
0 g
s
.92047 .03128 m .93247 .0288 L .93318 .03379 L closepath p .955 .786 .906 r
F P
0 g
s
.93594 .02922 m .93318 .03379 L .93247 .0288 L closepath p .955 .786 .906 r
F P
0 g
s
.50297 .03242 m .50533 .03038 L .50428 .02888 L closepath p .125 .125 .247 r
F P
0 g
s
.44025 .03087 m .44009 .03127 L .43112 .0552 L closepath p .955 .786 .906 r
F P
0 g
s
.44025 .03087 m .43112 .0552 L .44009 .03127 L closepath p .955 .786 .906 r
F P
0 g
s
.43112 .0552 m .43008 .0481 L .44025 .03087 L closepath p .955 .786 .906 r
F P
0 g
s
.44009 .03127 m .43993 .03166 L .43235 .06367 L closepath p .955 .786 .906 r
F P
0 g
s
.44009 .03127 m .43235 .06367 L .43993 .03166 L closepath p .955 .786 .906 r
F P
0 g
s
.43235 .06367 m .43112 .0552 L .44009 .03127 L closepath p .955 .786 .906 r
F P
0 g
s
.43993 .03166 m .43977 .03206 L .43377 .07352 L closepath p .955 .786 .906 r
F P
0 g
s
.43993 .03166 m .43377 .07352 L .43977 .03206 L closepath p .955 .786 .906 r
F P
0 g
s
.43377 .07352 m .43235 .06367 L .43993 .03166 L closepath p .955 .786 .906 r
F P
0 g
s
.43977 .03206 m .44197 .03001 L .44103 .0289 L closepath p .111 .111 .171 r
F P
0 g
s
.44197 .03001 m .43977 .03206 L .4421 .03206 L closepath p .168 .168 .642 r
F P
0 g
s
.44197 .03001 m .4421 .03206 L .44205 .03137 L p .713 .547 .002 r
F P
0 g
s
.44189 .02856 m .44103 .0289 L .44197 .03001 L p .104 .104 .223 r
F P
0 g
s
.44205 .03137 m .44122 .0196 L .44197 .03001 L p .713 .547 .002 r
F P
0 g
s
.44197 .03001 m .44238 .02837 L .44189 .02856 L p .104 .104 .223 r
F P
0 g
s
.4278 .03334 m .4272 .03145 L .44103 .0289 L closepath p .955 .786 .906 r
F P
0 g
s
.42624 .02997 m .44103 .0289 L .4272 .03145 L closepath p .955 .786 .906 r
F P
0 g
s
.44103 .0289 m .44087 .0293 L .4278 .03334 L closepath p .955 .786 .906 r
F P
0 g
s
.34187 .03188 m .34382 .02983 L .34348 .02897 L closepath p .067 .067 .067 r
F P
0 g
s
.34348 .02897 m .34307 .02969 L .32953 .03334 L closepath p .955 .786 .906 r
F P
0 g
s
.28299 .03188 m .2848 .02984 L .28461 .02897 L closepath p .055 .055 .055 r
F P
0 g
s
.27072 .03333 m .27012 .0314 L .28461 .02897 L closepath p .955 .786 .906 r
F P
0 g
s
.26914 .0299 m .28461 .02897 L .27012 .0314 L closepath p .955 .786 .906 r
F P
0 g
s
.28461 .02897 m .2838 .03042 L .27072 .03333 L closepath p .955 .786 .906 r
F P
0 g
s
.38209 .03333 m .37944 .02925 L .39592 .02897 L closepath p .955 .786 .906 r
F P
0 g
s
.39592 .02897 m .39512 .03043 L .38209 .03333 L closepath p .955 .786 .906 r
F P
0 g
s
.71656 .02897 m .71618 .0297 L .70268 .03334 L closepath p .955 .786 .906 r
F P
0 g
s
.76535 .03334 m .76243 .02922 L .77929 .02897 L closepath p .955 .786 .906 r
F P
0 g
s
.77929 .02897 m .77892 .0297 L .76535 .03334 L closepath p .955 .786 .906 r
F P
0 g
s
.77779 .03189 m .78083 .02984 L .77929 .02897 L closepath p .14 .14 .245 r
F P
0 g
s
.71505 .03189 m .71793 .02984 L .71656 .02897 L closepath p .131 .131 .205 r
F P
0 g
s
.18277 .02902 m .18207 .02995 L .17463 .02937 L closepath p .955 .786 .906 r
F P
0 g
s
.18181 .03112 m .18337 .02907 L .18277 .02903 L closepath p .054 .054 .054 r
F P
0 g
s
.18388 .03232 m .18337 .02907 L .18295 .02963 L p .147 .147 .624 r
F P
0 g
s
.18295 .02963 m .18181 .03112 L .18388 .03232 L p .147 .147 .624 r
F P
0 g
s
.18388 .03232 m .18235 .03471 L .18554 .03705 L closepath p .166 .254 .782 r
F P
0 g
s
.18388 .03232 m .18554 .03705 L .18523 .03625 L p .69 .372 .006 r
F P
0 g
s
.18235 .03471 m .18388 .03232 L .18181 .03112 L closepath p .148 .148 .621 r
F P
0 g
s
.18523 .03625 m .18077 .02466 L .18388 .03232 L p .69 .372 .006 r
F P
0 g
s
.18337 .02907 m .18388 .03232 L .18329 .03079 L p .855 .612 .053 r
F P
0 g
s
.1834 .02856 m .18277 .02903 L .18337 .02907 L p .056 .056 .056 r
F P
0 g
s
.18329 .03079 m .17908 .01986 L .18337 .02907 L p .855 .612 .053 r
F P
0 g
s
.18337 .02907 m .18358 .02842 L .1834 .02856 L p .056 .056 .056 r
F P
0 g
s
.14235 .03196 m .14066 .02944 L .14912 .02906 L closepath p .955 .786 .906 r
F P
0 g
s
.23914 .03379 m .24074 .03235 L .24213 .02922 L closepath p .053 .053 .053 r
F P
0 g
s
.24281 .02887 m .24213 .02922 L .24074 .03235 L closepath p .039 .039 .039 r
F P
0 g
s
.93318 .03379 m .93654 .03235 L .93594 .02922 L closepath p .157 .157 .294 r
F P
0 g
s
.76184 .02888 m .76581 .01637 L .76409 .02903 L p .139 .294 .833 r
F P
0 g
s
.76184 .02888 m .76249 .02922 L .76053 .02879 L p .022 .539 .778 r
F P
0 g
s
.76249 .02922 m .76184 .02888 L .76535 .03334 L closepath p .697 .245 .033 r
F P
0 g
s
.76372 .03175 m .76535 .03334 L .76306 .03043 L p .759 .313 .047 r
F P
0 g
s
.76306 .03043 m .76184 .02888 L .76372 .03175 L p .759 .313 .047 r
F P
0 g
s
.76535 .03334 m .76372 .03175 L .76579 .03542 L closepath p 1 .744 .27 r
F P
0 g
s
.76416 .03382 m .76579 .03542 L .76376 .03182 L p 1 .744 .27 r
F P
0 g
s
.76376 .03182 m .76372 .03175 L .76416 .03382 L p 1 .744 .27 r
F P
0 g
s
.76579 .03542 m .76535 .03334 L .77892 .0297 L closepath p .955 .786 .906 r
F P
0 g
s
.76579 .03542 m .76416 .03382 L .76642 .03874 L closepath p 1 .762 .292 r
F P
0 g
s
.76479 .03712 m .76642 .03874 L .76418 .03388 L p 1 .762 .293 r
F P
0 g
s
.76418 .03388 m .76416 .03382 L .76479 .03712 L p 1 .762 .293 r
F P
0 g
s
.77892 .0297 m .77854 .03043 L .76579 .03542 L closepath p .955 .786 .906 r
F P
0 g
s
.76642 .03874 m .76579 .03542 L .77854 .03043 L closepath p .955 .786 .906 r
F P
0 g
s
.76642 .03874 m .76479 .03712 L .76725 .04329 L closepath p 1 .77 .303 r
F P
0 g
s
.76561 .04167 m .76725 .04329 L .76481 .03719 L p 1 .77 .303 r
F P
0 g
s
.76481 .03719 m .76479 .03712 L .76561 .04167 L p 1 .77 .303 r
F P
0 g
s
.77854 .03043 m .77817 .03116 L .76642 .03874 L closepath p .955 .786 .906 r
F P
0 g
s
.77817 .03116 m .77779 .03189 L .76725 .04329 L closepath p .955 .786 .906 r
F P
0 g
s
.76725 .04329 m .76642 .03874 L .77817 .03116 L closepath p .955 .786 .906 r
F P
0 g
s
.76827 .04909 m .76725 .04329 L .77779 .03189 L closepath p .955 .786 .906 r
F P
0 g
s
.76725 .04329 m .76561 .04167 L .76828 .04909 L closepath p 1 .775 .308 r
F P
0 g
s
.76663 .04746 m .76828 .04909 L .76563 .04173 L p 1 .775 .309 r
F P
0 g
s
.76563 .04173 m .76561 .04167 L .76663 .04746 L p 1 .775 .309 r
F P
0 g
s
.77779 .03189 m .77801 .03475 L .76827 .04909 L closepath p .955 .786 .906 r
F P
0 g
s
.76888 .05261 m .76827 .04909 L .77801 .03475 L closepath p .955 .786 .906 r
F P
0 g
s
.77859 .0393 m .77953 .04554 L .76949 .05613 L closepath p .955 .786 .906 r
F P
0 g
s
.76949 .05613 m .76888 .05261 L .77859 .0393 L closepath p .955 .786 .906 r
F P
0 g
s
.76828 .04909 m .76663 .04746 L .76949 .05613 L closepath p 1 .778 .312 r
F P
0 g
s
.76785 .05449 m .76949 .05613 L .76665 .04752 L p 1 .778 .312 r
F P
0 g
s
.76665 .04752 m .76663 .04746 L .76785 .05449 L p 1 .778 .312 r
F P
0 g
s
.77801 .03475 m .77859 .0393 L .76888 .05261 L closepath p .955 .786 .906 r
F P
0 g
s
.7702 .06027 m .76949 .05613 L .77953 .04554 L closepath p .955 .786 .906 r
F P
0 g
s
.78083 .05348 m .78399 .0722 L .7709 .06441 L closepath p .955 .786 .906 r
F P
0 g
s
.7709 .06441 m .7702 .06027 L .78083 .05348 L closepath p .955 .786 .906 r
F P
0 g
s
.76949 .05613 m .76785 .05449 L .77091 .06441 L closepath p 1 .78 .315 r
F P
0 g
s
.76926 .06278 m .77091 .06441 L .76787 .05455 L p 1 .78 .315 r
F P
0 g
s
.76787 .05455 m .76785 .05449 L .76926 .06278 L p 1 .78 .315 r
F P
0 g
s
.77953 .04554 m .78083 .05348 L .7702 .06027 L closepath p .955 .786 .906 r
F P
0 g
s
.77445 .07698 m .77506 .08492 L .7709 .06441 L closepath p .955 .786 .906 r
F P
0 g
s
.77251 .07394 m .7709 .06441 L .77506 .08492 L closepath p .955 .786 .906 r
F P
0 g
s
.78514 .11481 m .7709 .06441 L .78399 .0722 L closepath p .955 .786 .906 r
F P
0 g
s
.77091 .06441 m .76926 .06278 L .77251 .07394 L closepath p 1 .781 .317 r
F P
0 g
s
.77087 .07233 m .77251 .07394 L .76928 .06284 L p 1 .781 .317 r
F P
0 g
s
.76928 .06284 m .76926 .06278 L .77087 .07233 L p 1 .781 .317 r
F P
0 g
s
.77506 .08492 m .77566 .09287 L .77251 .07394 L closepath p .955 .786 .906 r
F P
0 g
s
.77432 .08473 m .77251 .07394 L .77566 .09287 L closepath p .955 .786 .906 r
F P
0 g
s
.77432 .08473 m .77566 .09287 L .77251 .07394 L closepath p .955 .786 .906 r
F P
0 g
s
.77251 .07394 m .77087 .07233 L .77432 .08473 L closepath p 1 .783 .319 r
F P
0 g
s
.77268 .08313 m .77432 .08473 L .77092 .07249 L p 1 .783 .319 r
F P
0 g
s
.77092 .07249 m .77087 .07233 L .77268 .08313 L p 1 .783 .319 r
F P
0 g
s
.77801 .08957 m .77662 .0991 L .77445 .07698 L closepath p .955 .786 .906 r
F P
0 g
s
.77566 .09287 m .77445 .07698 L .77662 .0991 L closepath p .955 .786 .906 r
F P
0 g
s
.77432 .08473 m .77268 .08313 L .77566 .09287 L closepath p 1 .785 .322 r
F P
0 g
s
.77401 .09126 m .77566 .09287 L .77277 .08343 L p 1 .785 .322 r
F P
0 g
s
.77277 .08343 m .77268 .08313 L .77401 .09126 L p 1 .785 .322 r
F P
0 g
s
.77566 .09287 m .77401 .09126 L .77662 .0991 L closepath p 1 .793 .332 r
F P
0 g
s
.77497 .09742 m .77662 .0991 L .77412 .09159 L p 1 .792 .331 r
F P
0 g
s
.77412 .09159 m .77401 .09126 L .77497 .09742 L p 1 .792 .331 r
F P
0 g
s
.78157 .10218 m .77721 .10341 L .77801 .08957 L closepath p .955 .786 .906 r
F P
0 g
s
.77662 .0991 m .77801 .08957 L .77721 .10341 L closepath p .955 .786 .906 r
F P
0 g
s
.77662 .0991 m .77497 .09742 L .77721 .10341 L closepath p 1 .808 .351 r
F P
0 g
s
.77555 .10164 m .77721 .10341 L .77514 .09788 L p 1 .806 .348 r
F P
0 g
s
.77514 .09788 m .77497 .09742 L .77555 .10164 L p 1 .806 .348 r
F P
0 g
s
.77721 .10341 m .77555 .10164 L .77743 .1058 L closepath p 1 .845 .4 r
F P
0 g
s
.77576 .10393 m .77743 .1058 L .77614 .10295 L p 1 .843 .397 r
F P
0 g
s
.77614 .10295 m .77555 .10164 L .77576 .10393 L p 1 .843 .397 r
F P
0 g
s
.77555 .10164 m .77975 .09018 L .77579 .10383 L p .079 .079 .349 r
F P
0 g
s
.77579 .10383 m .77576 .10393 L .77555 .10164 L p .079 .079 .349 r
F P
0 g
s
.77996 .09249 m .77576 .10393 L .77946 .09117 L p .079 .079 .349 r
F P
0 g
s
.77946 .09117 m .77975 .09018 L .77996 .09249 L p .079 .079 .349 r
F P
0 g
s
.77996 .09249 m .78308 .09055 L .78036 .09355 L p .158 .158 .379 r
F P
0 g
s
.77576 .10393 m .77996 .09249 L .77518 .10539 L p .671 .754 .263 r
F P
0 g
s
.77897 .09509 m .77478 .10648 L .77902 .09503 L p .671 .754 .263 r
F P
0 g
s
.77902 .09503 m .77996 .09249 L .77897 .09509 L p .671 .754 .263 r
F P
0 g
s
.77497 .09742 m .77916 .0859 L .77557 .10156 L p .086 .086 .403 r
F P
0 g
s
.77557 .10156 m .77555 .10164 L .77497 .09742 L p .086 .086 .403 r
F P
0 g
s
.77975 .09018 m .77555 .10164 L .77916 .0859 L closepath p .086 .086 .403 r
F P
0 g
s
.77478 .10648 m .77897 .09509 L .77053 .10719 L closepath p .65 1 1 r
F P
0 g
s
.77401 .09126 m .77818 .07964 L .77498 .09735 L p .088 .088 .422 r
F P
0 g
s
.77498 .09735 m .77497 .09742 L .77401 .09126 L p .088 .088 .422 r
F P
0 g
s
.77916 .0859 m .77497 .09742 L .77818 .07964 L closepath p .088 .088 .422 r
F P
0 g
s
.77466 .09581 m .77053 .10719 L .77855 .09569 L p .65 1 1 r
F P
0 g
s
.77855 .09569 m .77897 .09509 L .77466 .09581 L p .65 1 1 r
F P
0 g
s
.77053 .10719 m .77466 .09581 L .76872 .10719 L closepath p .58 .961 1 r
F P
0 g
s
.77281 .09581 m .76872 .10719 L .77466 .09581 L closepath p .58 .961 1 r
F P
0 g
s
.77466 .09581 m .77542 .09621 L .77312 .09585 L p .14 .037 .037 r
F P
0 g
s
.77312 .09585 m .77281 .09581 L .77466 .09581 L p .14 .037 .037 r
F P
0 g
s
.77897 .09509 m .78122 .09475 L .77894 .09512 L p .004 .004 .004 r
F P
0 g
s
.77894 .09512 m .77466 .09581 L .77897 .09509 L p .004 .004 .004 r
F P
0 g
s
.78036 .09355 m .77897 .09509 L .77996 .09249 L p .158 .158 .379 r
F P
0 g
s
.77518 .10539 m .77478 .10648 L .77576 .10393 L p .671 .754 .263 r
F P
0 g
s
.77268 .08313 m .77683 .0714 L .77402 .09122 L p .089 .089 .433 r
F P
0 g
s
.77402 .09122 m .77401 .09126 L .77268 .08313 L p .089 .089 .433 r
F P
0 g
s
.77087 .07233 m .77499 .06044 L .77268 .08311 L p .09 .09 .436 r
F P
0 g
s
.77268 .08311 m .77268 .08313 L .77087 .07233 L p .09 .09 .436 r
F P
0 g
s
.76926 .06278 m .77335 .05076 L .77087 .0723 L p .09 .09 .437 r
F P
0 g
s
.77087 .0723 m .77087 .07233 L .76926 .06278 L p .09 .09 .437 r
F P
0 g
s
.76785 .05449 m .77192 .04235 L .76926 .06275 L p .09 .09 .439 r
F P
0 g
s
.76926 .06275 m .76926 .06278 L .76785 .05449 L p .09 .09 .439 r
F P
0 g
s
.76663 .04746 m .77068 .03521 L .76785 .05446 L p .091 .091 .442 r
F P
0 g
s
.76785 .05446 m .76785 .05449 L .76663 .04746 L p .091 .091 .442 r
F P
0 g
s
.76561 .04167 m .76964 .02934 L .76664 .04742 L p .091 .091 .446 r
F P
0 g
s
.76664 .04742 m .76663 .04746 L .76561 .04167 L p .091 .091 .446 r
F P
0 g
s
.76479 .03712 m .7688 .02473 L .76562 .04162 L p .092 .092 .452 r
F P
0 g
s
.76562 .04162 m .76561 .04167 L .76479 .03712 L p .092 .092 .452 r
F P
0 g
s
.76416 .03382 m .76816 .02138 L .7648 .03705 L p .093 .093 .462 r
F P
0 g
s
.7648 .03705 m .76479 .03712 L .76416 .03382 L p .093 .093 .462 r
F P
0 g
s
.76372 .03175 m .76772 .01929 L .76435 .03302 L p .096 .096 .485 r
F P
0 g
s
.76435 .03302 m .76416 .03382 L .76372 .03175 L p .096 .096 .485 r
F P
0 g
s
.76409 .02903 m .76372 .03175 L .76184 .02888 L p .139 .294 .833 r
F P
0 g
s
.76053 .02879 m .75717 .02806 L .76184 .02888 L p .022 .539 .778 r
F P
0 g
s
.67032 .03379 m .67301 .03235 L .67314 .02922 L closepath p .126 .126 .146 r
F P
0 g
s
.31253 .00699 m .31532 .01963 L .35064 .00699 L closepath p .623 .977 1 r
F P
0 g
s
.325 .02888 m .32235 .01637 L .326 .029 L p .068 .068 .612 r
F P
0 g
s
.325 .02888 m .32679 .02922 L .32442 .0288 L p .004 .485 .712 r
F P
0 g
s
.32679 .02922 m .325 .02888 L .32953 .03334 L closepath p .944 .493 .111 r
F P
0 g
s
.32679 .03174 m .32953 .03334 L .3266 .03045 L p .98 .541 .124 r
F P
0 g
s
.3266 .03045 m .325 .02888 L .32679 .03174 L p .98 .541 .124 r
F P
0 g
s
.326 .029 m .32679 .03174 L .325 .02888 L p .068 .068 .612 r
F P
0 g
s
.32417 .01927 m .32679 .03174 L .32272 .01766 L p .068 .068 .612 r
F P
0 g
s
.32272 .01766 m .32235 .01637 L .32417 .01927 L p .068 .068 .612 r
F P
0 g
s
.32417 .01927 m .32607 .01671 L .32488 .02007 L p .165 .165 .69 r
F P
0 g
s
.32679 .03174 m .32417 .01927 L .32696 .03287 L p .725 .494 .008 r
F P
0 g
s
.32449 .02117 m .32711 .03361 L .32417 .0193 L p .725 .494 .008 r
F P
0 g
s
.32417 .0193 m .32417 .01927 L .32449 .02117 L p .725 .494 .008 r
F P
0 g
s
.32711 .03361 m .32449 .02117 L .3276 .03678 L closepath p .74 .517 .011 r
F P
0 g
s
.32499 .02438 m .3276 .03678 L .3245 .02119 L p .74 .517 .011 r
F P
0 g
s
.3245 .02119 m .32449 .02117 L .32499 .02438 L p .74 .517 .011 r
F P
0 g
s
.3276 .03678 m .32499 .02438 L .32826 .04124 L closepath p .746 .526 .012 r
F P
0 g
s
.32566 .02891 m .32826 .04124 L .32499 .0244 L p .746 .526 .012 r
F P
0 g
s
.32499 .0244 m .32499 .02438 L .32566 .02891 L p .746 .526 .012 r
F P
0 g
s
.32696 .03287 m .32711 .03361 L .32679 .03174 L p .725 .494 .008 r
F P
0 g
s
.32953 .03334 m .32679 .03174 L .32985 .03523 L closepath p 1 .877 .533 r
F P
0 g
s
.32711 .03361 m .32985 .03523 L .32679 .03174 L closepath p 1 .877 .533 r
F P
0 g
s
.32826 .04124 m .32566 .02891 L .32909 .04701 L closepath p .75 .531 .013 r
F P
0 g
s
.32649 .03476 m .32909 .04701 L .32566 .02893 L p .75 .531 .013 r
F P
0 g
s
.32566 .02893 m .32566 .02891 L .32649 .03476 L p .75 .531 .013 r
F P
0 g
s
.32985 .03523 m .32711 .03361 L .33034 .03841 L closepath p 1 .887 .547 r
F P
0 g
s
.3276 .03678 m .33034 .03841 L .32711 .03361 L closepath p 1 .888 .547 r
F P
0 g
s
.33034 .03841 m .3276 .03678 L .331 .04288 L closepath p 1 .892 .553 r
F P
0 g
s
.32826 .04124 m .331 .04288 L .3276 .03678 L closepath p 1 .892 .553 r
F P
0 g
s
.32909 .04701 m .32649 .03476 L .33008 .05409 L closepath p .751 .534 .014 r
F P
0 g
s
.3275 .04194 m .33008 .05409 L .3265 .03478 L p .751 .534 .014 r
F P
0 g
s
.3265 .03478 m .32649 .03476 L .3275 .04194 L p .751 .534 .014 r
F P
0 g
s
.331 .04288 m .32826 .04124 L .33183 .04866 L closepath p 1 .894 .556 r
F P
0 g
s
.32909 .04701 m .33183 .04866 L .32826 .04124 L closepath p 1 .894 .556 r
F P
0 g
s
.33008 .05409 m .3275 .04194 L .33124 .06248 L closepath p .753 .536 .014 r
F P
0 g
s
.32868 .05045 m .33124 .06248 L .3275 .04196 L p .753 .536 .014 r
F P
0 g
s
.3275 .04196 m .3275 .04194 L .32868 .05045 L p .753 .536 .014 r
F P
0 g
s
.33183 .04866 m .32909 .04701 L .33283 .05573 L closepath p 1 .896 .558 r
F P
0 g
s
.33008 .05409 m .33283 .05573 L .32909 .04701 L closepath p 1 .896 .558 r
F P
0 g
s
.33283 .05573 m .33008 .05409 L .33399 .06411 L closepath p 1 .897 .56 r
F P
0 g
s
.33124 .06248 m .33399 .06411 L .33008 .05409 L closepath p 1 .897 .56 r
F P
0 g
s
.33124 .06248 m .32868 .05045 L .33258 .07218 L closepath p .754 .537 .014 r
F P
0 g
s
.33003 .06029 m .33258 .07218 L .32868 .05046 L p .754 .537 .014 r
F P
0 g
s
.32868 .05046 m .32868 .05045 L .33003 .06029 L p .754 .537 .014 r
F P
0 g
s
.33399 .06411 m .33124 .06248 L .33532 .0738 L closepath p 1 .897 .561 r
F P
0 g
s
.33258 .07218 m .33532 .0738 L .33124 .06248 L closepath p 1 .897 .561 r
F P
0 g
s
.33258 .07218 m .33003 .06029 L .33408 .08321 L closepath p .755 .539 .015 r
F P
0 g
s
.33156 .07148 m .33408 .08321 L .33005 .06037 L p .755 .539 .015 r
F P
0 g
s
.33005 .06037 m .33003 .06029 L .33156 .07148 L p .755 .539 .015 r
F P
0 g
s
.33532 .0738 m .33258 .07218 L .33683 .0848 L closepath p 1 .898 .562 r
F P
0 g
s
.33408 .08321 m .33683 .0848 L .33258 .07218 L closepath p 1 .898 .562 r
F P
0 g
s
.33408 .08321 m .33156 .07148 L .33513 .09116 L closepath p .758 .544 .015 r
F P
0 g
s
.33262 .07955 m .33513 .09116 L .33158 .0716 L p .758 .544 .015 r
F P
0 g
s
.33158 .0716 m .33156 .07148 L .33262 .07955 L p .758 .544 .015 r
F P
0 g
s
.33513 .09116 m .33262 .07955 L .33587 .09728 L closepath p .766 .557 .017 r
F P
0 g
s
.33337 .08575 m .33587 .09728 L .33265 .07969 L p .766 .557 .017 r
F P
0 g
s
.33265 .07969 m .33262 .07955 L .33337 .08575 L p .766 .557 .017 r
F P
0 g
s
.33683 .0848 m .33408 .08321 L .33788 .09278 L closepath p 1 .901 .565 r
F P
0 g
s
.33513 .09116 m .33788 .09278 L .33408 .08321 L closepath p 1 .9 .565 r
F P
0 g
s
.33587 .09728 m .33337 .08575 L .33631 .10156 L closepath p .781 .58 .021 r
F P
0 g
s
.33381 .09009 m .33631 .10156 L .33342 .08598 L p .781 .58 .021 r
F P
0 g
s
.33342 .08598 m .33337 .08575 L .33381 .09009 L p .781 .58 .021 r
F P
0 g
s
.33631 .10156 m .33381 .09009 L .33644 .10403 L closepath p .814 .635 .07 r
F P
0 g
s
.33395 .09259 m .33644 .10403 L .33391 .0906 L p .814 .635 .07 r
F P
0 g
s
.33391 .0906 m .33381 .09009 L .33395 .09259 L p .814 .635 .07 r
F P
0 g
s
.33788 .09278 m .33513 .09116 L .33863 .09896 L closepath p 1 .907 .574 r
F P
0 g
s
.33587 .09728 m .33863 .09896 L .33513 .09116 L closepath p 1 .906 .573 r
F P
0 g
s
.33644 .10403 m .33395 .09259 L .33613 .10596 L closepath p .927 .848 .348 r
F P
0 g
s
.33363 .09456 m .33613 .10596 L .33417 .09398 L p .927 .848 .348 r
F P
0 g
s
.33417 .09398 m .33395 .09259 L .33363 .09456 L p .927 .848 .348 r
F P
0 g
s
.33613 .10596 m .33363 .09456 L .33489 .10684 L closepath p .9 1 1 r
F P
0 g
s
.33237 .09545 m .33489 .10684 L .33369 .09513 L p .9 1 1 r
F P
0 g
s
.33369 .09513 m .33363 .09456 L .33237 .09545 L p .9 1 1 r
F P
0 g
s
.33489 .10684 m .33237 .09545 L .33295 .10718 L closepath p .691 1 1 r
F P
0 g
s
.3304 .09579 m .33295 .10718 L .33237 .09545 L closepath p .691 1 1 r
F P
0 g
s
.33237 .09545 m .33278 .09569 L .33105 .09576 L p .029 .029 .029 r
F P
0 g
s
.33105 .09576 m .3304 .09579 L .33237 .09545 L p .029 .029 .029 r
F P
0 g
s
.33363 .09456 m .33512 .09384 L .33332 .09489 L p .025 .025 .025 r
F P
0 g
s
.33332 .09489 m .33237 .09545 L .33363 .09456 L p .025 .025 .025 r
F P
0 g
s
.33395 .09259 m .33593 .09066 L .33397 .09397 L p .13 .13 .378 r
F P
0 g
s
.33397 .09397 m .33363 .09456 L .33395 .09259 L p .13 .13 .378 r
F P
0 g
s
.33381 .09009 m .33579 .08804 L .33404 .09238 L p .154 .154 .588 r
F P
0 g
s
.33404 .09238 m .33395 .09259 L .33381 .09009 L p .154 .154 .588 r
F P
0 g
s
.33337 .08575 m .33534 .08359 L .33387 .08988 L p .158 .158 .631 r
F P
0 g
s
.33387 .08988 m .33381 .09009 L .33337 .08575 L p .158 .158 .631 r
F P
0 g
s
.33262 .07955 m .33458 .07729 L .33341 .08552 L p .16 .16 .647 r
F P
0 g
s
.33341 .08552 m .33337 .08575 L .33262 .07955 L p .16 .16 .647 r
F P
0 g
s
.33156 .07148 m .33351 .06916 L .33263 .0794 L p .161 .161 .656 r
F P
0 g
s
.33263 .0794 m .33262 .07955 L .33156 .07148 L p .161 .161 .656 r
F P
0 g
s
.33003 .06029 m .33198 .05794 L .33156 .07141 L p .162 .162 .66 r
F P
0 g
s
.33156 .07141 m .33156 .07148 L .33003 .06029 L p .162 .162 .66 r
F P
0 g
s
.32868 .05045 m .33062 .04807 L .33004 .06023 L p .162 .162 .661 r
F P
0 g
s
.33004 .06023 m .33003 .06029 L .32868 .05045 L p .162 .162 .661 r
F P
0 g
s
.3275 .04194 m .32943 .03953 L .32868 .05038 L p .162 .162 .662 r
F P
0 g
s
.32868 .05038 m .32868 .05045 L .3275 .04194 L p .162 .162 .662 r
F P
0 g
s
.32649 .03476 m .32842 .03232 L .32751 .04187 L p .162 .162 .663 r
F P
0 g
s
.32751 .04187 m .3275 .04194 L .32649 .03476 L p .162 .162 .663 r
F P
0 g
s
.32566 .02891 m .32758 .02643 L .3265 .03468 L p .162 .162 .665 r
F P
0 g
s
.3265 .03468 m .32649 .03476 L .32566 .02891 L p .162 .162 .665 r
F P
0 g
s
.32499 .02438 m .3269 .02187 L .32567 .02882 L p .163 .163 .669 r
F P
0 g
s
.32567 .02882 m .32566 .02891 L .32499 .02438 L p .163 .163 .669 r
F P
0 g
s
.32449 .02117 m .3264 .01863 L .32502 .02425 L p .163 .163 .675 r
F P
0 g
s
.32502 .02425 m .32499 .02438 L .32449 .02117 L p .163 .163 .675 r
F P
0 g
s
.32488 .02007 m .32449 .02117 L .32417 .01927 L p .165 .165 .69 r
F P
0 g
s
.32442 .0288 m .3202 .02806 L .325 .02888 L p .004 .485 .712 r
F P
0 g
s
.37777 .0289 m .37592 .01639 L .37887 .02908 L p .075 .075 .625 r
F P
0 g
s
.37777 .0289 m .37944 .02925 L .37713 .02879 L p .003 .466 .704 r
F P
0 g
s
.37944 .02925 m .37777 .0289 L .38209 .03333 L closepath p .93 .48 .105 r
F P
0 g
s
.3795 .03176 m .38209 .03333 L .37925 .03041 L p .977 .543 .119 r
F P
0 g
s
.37925 .03041 m .37777 .0289 L .3795 .03176 L p .977 .543 .119 r
F P
0 g
s
.38209 .03333 m .3795 .03176 L .38266 .03665 L closepath p 1 .87 .513 r
F P
0 g
s
.38005 .035 m .38266 .03665 L .37959 .0319 L p 1 .87 .513 r
F P
0 g
s
.37959 .0319 m .3795 .03176 L .38005 .035 L p 1 .87 .513 r
F P
0 g
s
.38266 .03665 m .38005 .035 L .38369 .04351 L closepath p 1 .884 .531 r
F P
0 g
s
.38103 .04154 m .38369 .04351 L .38012 .03516 L p 1 .884 .53 r
F P
0 g
s
.38012 .03516 m .38005 .035 L .38103 .04154 L p 1 .884 .53 r
F P
0 g
s
.381 .04135 m .38103 .04154 L .38005 .035 L p .001 .001 .107 r
F P
0 g
s
.38103 .04154 m .37923 .02921 L .38239 .05105 L p .703 .492 .001 r
F P
0 g
s
.38239 .05105 m .3824 .05115 L .38103 .04154 L p .703 .492 .001 r
F P
0 g
s
.38005 .035 m .37823 .02258 L .381 .04135 L p .001 .001 .107 r
F P
0 g
s
.3795 .03176 m .37767 .01929 L .37986 .03372 L p .005 .005 .137 r
F P
0 g
s
.37986 .03372 m .38005 .035 L .3795 .03176 L p .005 .005 .137 r
F P
0 g
s
.37887 .02908 m .3795 .03176 L .37777 .0289 L p .075 .075 .625 r
F P
0 g
s
.37713 .02879 m .37341 .02805 L .37777 .0289 L p .003 .466 .704 r
F P
0 g
s
.69924 .02888 m .70226 .01638 L .70112 .02905 L p .125 .208 .767 r
F P
0 g
s
.69924 .02888 m .70009 .02926 L .69816 .02879 L p .019 .512 .764 r
F P
0 g
s
.70009 .02926 m .69924 .02888 L .70268 .03334 L closepath p .771 .33 .05 r
F P
0 g
s
.70088 .03173 m .70268 .03334 L .7003 .03026 L p .848 .419 .07 r
F P
0 g
s
.7003 .03026 m .69924 .02888 L .70088 .03173 L p .848 .419 .07 r
F P
0 g
s
.70268 .03334 m .70088 .03173 L .70349 .0373 L closepath p 1 .772 .318 r
F P
0 g
s
.70169 .03568 m .70349 .0373 L .70095 .03188 L p 1 .772 .318 r
F P
0 g
s
.70095 .03188 m .70088 .03173 L .70169 .03568 L p 1 .772 .318 r
F P
0 g
s
.70349 .0373 m .70268 .03334 L .71618 .0297 L closepath p .955 .786 .906 r
F P
0 g
s
.70349 .0373 m .70169 .03568 L .70448 .04261 L closepath p 1 .786 .336 r
F P
0 g
s
.70269 .04098 m .70448 .04261 L .70175 .03583 L p 1 .787 .336 r
F P
0 g
s
.70175 .03583 m .70169 .03568 L .70269 .04098 L p 1 .787 .336 r
F P
0 g
s
.71618 .0297 m .7158 .03043 L .70349 .0373 L closepath p .955 .786 .906 r
F P
0 g
s
.70448 .0426 m .70349 .0373 L .7158 .03043 L closepath p .955 .786 .906 r
F P
0 g
s
.70448 .04261 m .70269 .04098 L .70567 .04925 L closepath p 1 .795 .347 r
F P
0 g
s
.70387 .04763 m .70567 .04925 L .70274 .04114 L p 1 .795 .346 r
F P
0 g
s
.70274 .04114 m .70269 .04098 L .70387 .04763 L p 1 .795 .346 r
F P
0 g
s
.7158 .03043 m .71543 .03116 L .70448 .0426 L closepath p .955 .786 .906 r
F P
0 g
s
.71543 .03116 m .71505 .03189 L .70566 .04925 L closepath p .955 .786 .906 r
F P
0 g
s
.70566 .04925 m .70448 .0426 L .71543 .03116 L closepath p .955 .786 .906 r
F P
0 g
s
.70703 .05725 m .70566 .04925 L .71505 .03189 L closepath p .955 .786 .906 r
F P
0 g
s
.70567 .04925 m .70387 .04763 L .70703 .05725 L closepath p 1 .8 .353 r
F P
0 g
s
.70523 .05562 m .70703 .05725 L .70391 .04777 L p 1 .8 .354 r
F P
0 g
s
.70391 .04777 m .70387 .04763 L .70523 .05562 L p 1 .8 .354 r
F P
0 g
s
.70703 .05725 m .71505 .03189 L .70859 .06659 L closepath p .955 .786 .906 r
F P
0 g
s
.71472 .06725 m .70859 .06659 L .71505 .03189 L closepath p .955 .786 .906 r
F P
0 g
s
.70703 .05725 m .70523 .05562 L .70859 .06659 L closepath p 1 .804 .358 r
F P
0 g
s
.70679 .06497 m .70859 .06659 L .70528 .05577 L p 1 .804 .358 r
F P
0 g
s
.70528 .05577 m .70523 .05562 L .70679 .06497 L p 1 .804 .358 r
F P
0 g
s
.70859 .06659 m .70679 .06497 L .71033 .0773 L closepath p 1 .807 .362 r
F P
0 g
s
.70859 .06659 m .71165 .06692 L .71033 .07729 L closepath p .955 .786 .906 r
F P
0 g
s
.70853 .07568 m .71033 .0773 L .70683 .06512 L p 1 .807 .362 r
F P
0 g
s
.70683 .06512 m .70679 .06497 L .70853 .07568 L p 1 .807 .362 r
F P
0 g
s
.71226 .08936 m .71033 .07729 L .71165 .06692 L closepath p .955 .786 .906 r
F P
0 g
s
.71033 .0773 m .70853 .07568 L .71226 .08936 L closepath p 1 .809 .365 r
F P
0 g
s
.71046 .08777 m .71226 .08936 L .70857 .07584 L p 1 .809 .365 r
F P
0 g
s
.70857 .07584 m .70853 .07568 L .71046 .08777 L p 1 .809 .365 r
F P
0 g
s
.71165 .06692 m .71472 .06725 L .71226 .08936 L closepath p .955 .786 .906 r
F P
0 g
s
.71438 .1028 m .71226 .08936 L .71472 .06725 L closepath p .955 .786 .906 r
F P
0 g
s
.71226 .08936 m .71046 .08777 L .71438 .1028 L closepath p 1 .811 .367 r
F P
0 g
s
.71259 .10123 m .71438 .1028 L .71162 .0922 L p 1 .811 .367 r
F P
0 g
s
.71162 .0922 m .71046 .08777 L .71259 .10123 L p 1 .811 .367 r
F P
0 g
s
.71505 .03189 m .71525 .03418 L .71438 .1028 L closepath p .955 .786 .906 r
F P
0 g
s
.71525 .03418 m .71584 .03854 L .71448 .10393 L closepath p .955 .786 .906 r
F P
0 g
s
.71448 .10393 m .71438 .1028 L .71525 .03418 L closepath p .955 .786 .906 r
F P
0 g
s
.71584 .03854 m .7168 .04497 L .71458 .10507 L closepath p .955 .786 .906 r
F P
0 g
s
.71458 .10507 m .71448 .10393 L .71584 .03854 L closepath p .955 .786 .906 r
F P
0 g
s
.7168 .04497 m .71814 .05348 L .71467 .1062 L closepath p .955 .786 .906 r
F P
0 g
s
.71467 .1062 m .71458 .10507 L .7168 .04497 L closepath p .955 .786 .906 r
F P
0 g
s
.71477 .10734 m .71467 .1062 L .71814 .05348 L closepath p .955 .786 .906 r
F P
0 g
s
.71438 .1028 m .71259 .10123 L .71477 .10734 L closepath p 1 .866 .44 r
F P
0 g
s
.71295 .10547 m .71477 .10734 L .71355 .10391 L p 1 .866 .44 r
F P
0 g
s
.71355 .10391 m .71259 .10123 L .71295 .10547 L p 1 .866 .44 r
F P
0 g
s
.71259 .10123 m .71583 .08976 L .71301 .10514 L p .065 .065 .292 r
F P
0 g
s
.71301 .10514 m .71295 .10547 L .71259 .10123 L p .065 .065 .292 r
F P
0 g
s
.71477 .10734 m .71814 .05348 L .72209 .0781 L closepath p .955 .786 .906 r
F P
0 g
s
.71619 .09406 m .71295 .10547 L .71549 .09159 L p .065 .065 .292 r
F P
0 g
s
.71549 .09159 m .71583 .08976 L .71619 .09406 L p .065 .065 .292 r
F P
0 g
s
.72209 .0781 m .72604 .1028 L .71477 .10734 L closepath p .955 .786 .906 r
F P
0 g
s
.71619 .09406 m .71915 .09212 L .7164 .09478 L p .155 .155 .382 r
F P
0 g
s
.71295 .10547 m .71619 .09406 L .71262 .10633 L p .71 .771 .275 r
F P
0 g
s
.71571 .09545 m .71247 .10684 L .71577 .09552 L p .71 .771 .275 r
F P
0 g
s
.71577 .09552 m .71619 .09406 L .71571 .09545 L p .71 .771 .275 r
F P
0 g
s
.71778 .09513 m .71571 .09545 L .71915 .09212 L closepath p .135 .135 .248 r
F P
0 g
s
.71571 .09545 m .71778 .09513 L .71566 .09539 L p .112 .667 .743 r
F P
0 g
s
.71583 .08976 m .71259 .10123 L .71367 .0761 L closepath p .076 .076 .381 r
F P
0 g
s
.71566 .09539 m .71296 .09573 L .71571 .09545 L p .112 .667 .743 r
F P
0 g
s
.7164 .09478 m .71571 .09545 L .71619 .09406 L p .155 .155 .382 r
F P
0 g
s
.71262 .10633 m .71247 .10684 L .71295 .10547 L p .71 .771 .275 r
F P
0 g
s
.71046 .08777 m .71367 .0761 L .71259 .10117 L p .076 .076 .381 r
F P
0 g
s
.71259 .10117 m .71259 .10123 L .71046 .08777 L p .076 .076 .381 r
F P
0 g
s
.70853 .07568 m .7117 .06385 L .71047 .08771 L p .076 .076 .383 r
F P
0 g
s
.71047 .08771 m .71046 .08777 L .70853 .07568 L p .076 .076 .383 r
F P
0 g
s
.70679 .06497 m .70993 .05298 L .70853 .07562 L p .077 .077 .387 r
F P
0 g
s
.70853 .07562 m .70853 .07568 L .70679 .06497 L p .077 .077 .387 r
F P
0 g
s
.70523 .05562 m .70835 .04349 L .70679 .06491 L p .077 .077 .391 r
F P
0 g
s
.70679 .06491 m .70679 .06497 L .70523 .05562 L p .077 .077 .391 r
F P
0 g
s
.70387 .04763 m .70696 .03538 L .70524 .05555 L p .078 .078 .396 r
F P
0 g
s
.70524 .05555 m .70523 .05562 L .70387 .04763 L p .078 .078 .396 r
F P
0 g
s
.70269 .04098 m .70576 .02865 L .70387 .04754 L p .079 .079 .404 r
F P
0 g
s
.70387 .04754 m .70387 .04763 L .70269 .04098 L p .079 .079 .404 r
F P
0 g
s
.70169 .03568 m .70475 .02327 L .7027 .04087 L p .08 .08 .416 r
F P
0 g
s
.7027 .04087 m .70269 .04098 L .70169 .03568 L p .08 .08 .416 r
F P
0 g
s
.70088 .03173 m .70393 .01926 L .70189 .03426 L p .083 .083 .436 r
F P
0 g
s
.70189 .03426 m .70169 .03568 L .70088 .03173 L p .083 .083 .436 r
F P
0 g
s
.70112 .02905 m .70088 .03173 L .69924 .02888 L p .125 .208 .767 r
F P
0 g
s
.69816 .02879 m .69519 .02806 L .69924 .02888 L p .019 .512 .764 r
F P
0 g
s
.44087 .0293 m .44072 .02969 L .42809 .03499 L closepath p .955 .786 .906 r
F P
0 g
s
.42809 .03499 m .4278 .03334 L .44087 .0293 L closepath p .955 .786 .906 r
F P
0 g
s
.54188 .0293 m .54584 .03585 L .52683 .07932 L closepath p .955 .786 .906 r
F P
0 g
s
.52683 .07932 m .52925 .06672 L .54188 .0293 L closepath p .955 .786 .906 r
F P
0 g
s
.18013 .0321 m .17916 .03317 L .17463 .02937 L closepath p .955 .786 .906 r
F P
0 g
s
.17819 .03425 m .17463 .02937 L .17916 .03317 L closepath p .955 .786 .906 r
F P
0 g
s
.18013 .0321 m .17463 .02937 L .18207 .02995 L closepath p .955 .786 .906 r
F P
0 g
s
.05988 .03348 m .06114 .03144 L .06147 .02948 L closepath p .04 .04 .04 r
F P
0 g
s
.06186 .02893 m .06147 .02948 L .06114 .03144 L closepath p .021 .021 .021 r
F P
0 g
s
.58751 .03288 m .59008 .03083 L .58852 .0296 L closepath p .145 .145 .36 r
F P
0 g
s
.58851 .0296 m .58801 .03124 L .57608 .03486 L closepath p .955 .786 .906 r
F P
0 g
s
.86106 .03884 m .86421 .03622 L .85792 .0296 L closepath p .426 .457 .889 r
F P
0 g
s
.86103 .02686 m .85792 .0296 L .86421 .03622 L closepath p .426 .457 .889 r
F P
0 g
s
.85792 .0296 m .86106 .03884 L .86242 .04783 L closepath p .955 .786 .906 r
F P
0 g
s
.86242 .04783 m .86076 .0464 L .85792 .0296 L closepath p .955 .786 .906 r
F P
0 g
s
.44072 .02969 m .44056 .03008 L .42856 .038 L closepath p .955 .786 .906 r
F P
0 g
s
.42856 .038 m .42809 .03499 L .44072 .02969 L closepath p .955 .786 .906 r
F P
0 g
s
.34307 .02969 m .34267 .03042 L .32985 .03523 L closepath p .955 .786 .906 r
F P
0 g
s
.32985 .03523 m .32953 .03334 L .34307 .02969 L closepath p .955 .786 .906 r
F P
0 g
s
.15657 .0341 m .15806 .03206 L .15865 .0297 L closepath p .046 .046 .046 r
F P
0 g
s
.15923 .0292 m .15865 .0297 L .15806 .03206 L closepath p .034 .034 .034 r
F P
0 g
s
.10897 .03737 m .10456 .02987 L .10561 .0308 L closepath p .955 .786 .906 r
F P
0 g
s
.10456 .02987 m .10897 .03737 L .08547 .0872 L closepath p .955 .786 .906 r
F P
0 g
s
.08547 .0872 m .08867 .07285 L .10456 .02987 L closepath p .955 .786 .906 r
F P
0 g
s
.18207 .02995 m .18181 .03112 L .17864 .03386 L closepath p .955 .786 .906 r
F P
0 g
s
.17842 .03405 m .18207 .02995 L .17864 .03386 L closepath p .955 .786 .906 r
F P
0 g
s
.17819 .03425 m .18207 .02995 L .17842 .03405 L closepath p .955 .786 .906 r
F P
0 g
s
.44056 .03008 m .4404 .03048 L .42922 .04237 L closepath p .955 .786 .906 r
F P
0 g
s
.42922 .04237 m .42856 .038 L .44056 .03008 L closepath p .955 .786 .906 r
F P
0 g
s
.53563 .0304 m .5352 .03384 L .53446 .03759 L closepath p .955 .786 .906 r
F P
0 g
s
.53446 .03759 m .53518 .03257 L .53563 .0304 L closepath p .955 .786 .906 r
F P
0 g
s
.34267 .03042 m .34227 .03115 L .33034 .03841 L closepath p .955 .786 .906 r
F P
0 g
s
.33034 .03841 m .32985 .03523 L .34267 .03042 L closepath p .955 .786 .906 r
F P
0 g
s
.2838 .03042 m .28299 .03188 L .27331 .04891 L closepath p .955 .786 .906 r
F P
0 g
s
.27331 .04891 m .27072 .03333 L .2838 .03042 L closepath p .955 .786 .906 r
F P
0 g
s
.39512 .03043 m .39432 .03189 L .38266 .03665 L closepath p .955 .786 .906 r
F P
0 g
s
.38266 .03665 m .38209 .03333 L .39512 .03043 L closepath p .955 .786 .906 r
F P
0 g
s
.4404 .03048 m .44025 .03087 L .43008 .0481 L closepath p .955 .786 .906 r
F P
0 g
s
.43008 .0481 m .42922 .04237 L .4404 .03048 L closepath p .955 .786 .906 r
F P
0 g
s
.10897 .03737 m .11012 .03431 L .10562 .0308 L closepath p .201 .363 .87 r
F P
0 g
s
.10661 .02751 m .10562 .0308 L .11012 .03431 L closepath p .197 .362 .87 r
F P
0 g
s
.81417 .03494 m .81694 .03414 L .81755 .03087 L closepath p .131 .131 .131 r
F P
0 g
s
.81694 .03414 m .81417 .03494 L .8133 .04248 L closepath p .162 .162 .376 r
F P
0 g
s
.81694 .03414 m .8133 .04248 L .81414 .0403 L p .683 .789 .316 r
F P
0 g
s
.81851 .03069 m .81755 .03087 L .81694 .03414 L p .13 .13 .13 r
F P
0 g
s
.81414 .0403 m .81806 .03016 L .81694 .03414 L p .683 .789 .316 r
F P
0 g
s
.81694 .03414 m .81998 .03042 L .81851 .03069 L p .13 .13 .13 r
F P
0 g
s
.18181 .03112 m .18235 .03471 L .17819 .03425 L closepath p .955 .786 .906 r
F P
0 g
s
.18091 .0319 m .17864 .03386 L .18181 .03112 L closepath p .955 .786 .906 r
F P
0 g
s
.18091 .0319 m .18181 .03112 L .17864 .03386 L closepath p .955 .786 .906 r
F P
0 g
s
.61357 .03454 m .61616 .03215 L .61291 .03113 L closepath p .277 .302 .781 r
F P
0 g
s
.61554 .02908 m .61291 .03113 L .61616 .03215 L closepath p .275 .304 .783 r
F P
0 g
s
.34227 .03115 m .34187 .03188 L .331 .04288 L closepath p .955 .786 .906 r
F P
0 g
s
.331 .04288 m .33034 .03841 L .34227 .03115 L closepath p .955 .786 .906 r
F P
0 g
s
.58801 .03124 m .58751 .03288 L .57844 .05125 L closepath p .955 .786 .906 r
F P
0 g
s
.22664 .03127 m .22361 .02923 L .22603 .03714 L closepath p 1 1 .758 r
F P
0 g
s
.22664 .03127 m .23914 .03379 L .22603 .03714 L closepath p .955 .786 .906 r
F P
0 g
s
.91917 .02923 m .92047 .03128 L .91904 .0287 L p .809 .452 .043 r
F P
0 g
s
.92047 .03128 m .91917 .02923 L .92017 .03715 L closepath p 1 .911 .469 r
F P
0 g
s
.94377 .01964 m .95049 .007 L .90786 .01964 L closepath p .623 .977 1 r
F P
0 g
s
.91917 .02923 m .92553 .01673 L .91988 .03212 L p .086 .086 .277 r
F P
0 g
s
.91889 .03483 m .92017 .03715 L .91939 .03098 L p 1 .911 .468 r
F P
0 g
s
.91939 .03098 m .91917 .02923 L .91889 .03483 L p 1 .911 .468 r
F P
0 g
s
.92047 .03128 m .93318 .03379 L .92017 .03715 L closepath p .955 .786 .906 r
F P
0 g
s
.92017 .03715 m .91889 .03483 L .91896 .04424 L closepath p 1 .992 .596 r
F P
0 g
s
.91771 .04169 m .91896 .04424 L .9189 .03609 L p 1 .993 .599 r
F P
0 g
s
.9189 .03609 m .91889 .03483 L .91771 .04169 L p 1 .993 .599 r
F P
0 g
s
.91896 .04424 m .92017 .03715 L .93318 .03379 L closepath p .955 .786 .906 r
F P
0 g
s
.91896 .04424 m .91771 .04169 L .91753 .05066 L closepath p 1 1 .649 r
F P
0 g
s
.9163 .0479 m .91753 .05066 L .91771 .04169 L closepath p 1 1 .655 r
F P
0 g
s
.91771 .04169 m .92405 .02936 L .91648 .04748 L p .054 .054 .054 r
F P
0 g
s
.91648 .04748 m .9163 .0479 L .91771 .04169 L p .054 .054 .054 r
F P
0 g
s
.91889 .03483 m .92524 .0224 L .91812 .04063 L p .064 .064 .117 r
F P
0 g
s
.91812 .04063 m .91771 .04169 L .91889 .03483 L p .064 .064 .117 r
F P
0 g
s
.91988 .03212 m .91889 .03483 L .91917 .02923 L p .086 .086 .277 r
F P
0 g
s
.91904 .0287 m .91875 .02817 L .91917 .02923 L p .809 .452 .043 r
F P
0 g
s
.65766 .03128 m .67032 .03379 L .65725 .03715 L closepath p .955 .786 .906 r
F P
0 g
s
.05301 .03129 m .04982 .03038 L .05398 .0337 L closepath p 1 .775 .457 r
F P
0 g
s
.0972 .03129 m .09658 .03525 L .09556 .03969 L closepath p .955 .786 .906 r
F P
0 g
s
.09556 .03969 m .09659 .0338 L .0972 .03129 L closepath p .955 .786 .906 r
F P
0 g
s
.05301 .03129 m .05988 .03348 L .05344 .03224 L closepath p .955 .786 .906 r
F P
0 g
s
.84642 .03106 m .847 .03135 L .84278 .02895 L p .366 .003 .003 r
F P
0 g
s
.85197 .03591 m .847 .03135 L .85721 .0421 L closepath p .955 .786 .906 r
F P
0 g
s
.85197 .03591 m .85721 .0421 L .847 .03135 L closepath p .955 .786 .906 r
F P
0 g
s
.847 .03135 m .84642 .03106 L .85197 .03591 L closepath p .54 .018 .017 r
F P
0 g
s
.85194 .02093 m .85728 .0083 L .84484 .01932 L closepath p .524 .873 1 r
F P
0 g
s
.84642 .03106 m .85167 .01859 L .85121 .03384 L p .24 .538 1 r
F P
0 g
s
.85116 .03541 m .85197 .03591 L .84789 .03234 L p .54 .018 .017 r
F P
0 g
s
.84789 .03234 m .84642 .03106 L .85116 .03541 L p .54 .018 .017 r
F P
0 g
s
.85197 .03591 m .85116 .03541 L .85721 .0421 L closepath p .641 .155 .028 r
F P
0 g
s
.85116 .03541 m .85649 .02299 L .85625 .04017 L p .161 .436 .936 r
F P
0 g
s
.85624 .04141 m .85721 .0421 L .85116 .03541 L closepath p .645 .159 .029 r
F P
0 g
s
.85625 .04017 m .85624 .04141 L .85116 .03541 L p .161 .436 .936 r
F P
0 g
s
.85121 .03384 m .85116 .03541 L .84642 .03106 L p .24 .538 1 r
F P
0 g
s
.84278 .02895 m .83955 .02711 L .84642 .03106 L p .366 .003 .003 r
F P
0 g
s
.34221 .04003 m .34416 .03782 L .34187 .03188 L closepath p .156 .156 .589 r
F P
0 g
s
.34382 .02983 m .34187 .03188 L .34416 .03782 L closepath p .156 .156 .589 r
F P
0 g
s
.33183 .04866 m .331 .04288 L .34187 .03188 L closepath p .955 .786 .906 r
F P
0 g
s
.33935 .05829 m .33283 .05573 L .34187 .03188 L closepath p .955 .786 .906 r
F P
0 g
s
.33183 .04866 m .34187 .03188 L .33283 .05573 L closepath p .955 .786 .906 r
F P
0 g
s
.34221 .04003 m .33683 .0848 L .34187 .03188 L closepath p .955 .786 .906 r
F P
0 g
s
.28313 .03434 m .28494 .0322 L .28299 .03188 L closepath p .15 .15 .576 r
F P
0 g
s
.28538 .03653 m .28494 .0322 L .28488 .03227 L p .155 .155 .62 r
F P
0 g
s
.28488 .03227 m .28313 .03434 L .28538 .03653 L p .155 .155 .62 r
F P
0 g
s
.28614 .04284 m .28538 .03653 L .28534 .03658 L p .157 .157 .636 r
F P
0 g
s
.28534 .03658 m .28358 .03876 L .28614 .04284 L p .157 .157 .636 r
F P
0 g
s
.28614 .04284 m .28434 .04514 L .28722 .05116 L closepath p .158 .158 .645 r
F P
0 g
s
.28614 .04284 m .28722 .05116 L .28721 .0511 L p .796 .577 .026 r
F P
0 g
s
.28434 .04514 m .28614 .04284 L .28358 .03876 L closepath p .157 .157 .637 r
F P
0 g
s
.28721 .0511 m .28399 .03897 L .28614 .04284 L p .796 .577 .026 r
F P
0 g
s
.28538 .03653 m .28614 .04284 L .28612 .04274 L p .803 .588 .028 r
F P
0 g
s
.28472 .02993 m .28299 .03188 L .28494 .0322 L p .15 .15 .577 r
F P
0 g
s
.28358 .03876 m .28538 .03653 L .28313 .03434 L closepath p .155 .155 .62 r
F P
0 g
s
.28612 .04274 m .2829 .03053 L .28538 .03653 L p .803 .588 .028 r
F P
0 g
s
.28494 .0322 m .28538 .03653 L .28534 .03635 L p .817 .609 .041 r
F P
0 g
s
.28534 .03635 m .28213 .02413 L .28494 .0322 L p .817 .609 .041 r
F P
0 g
s
.28494 .0322 m .2848 .02984 L .28472 .02993 L p .15 .15 .577 r
F P
0 g
s
.27585 .06794 m .27331 .04891 L .28299 .03188 L closepath p .955 .786 .906 r
F P
0 g
s
.27908 .09406 m .27802 .08534 L .28299 .03188 L closepath p .955 .786 .906 r
F P
0 g
s
.27585 .06794 m .28299 .03188 L .27802 .08534 L closepath p .955 .786 .906 r
F P
0 g
s
.28055 .10734 m .28014 .1028 L .28299 .03188 L closepath p .955 .786 .906 r
F P
0 g
s
.27908 .09406 m .28299 .03188 L .28014 .1028 L closepath p .955 .786 .906 r
F P
0 g
s
.28055 .10734 m .28299 .03188 L .28218 .09692 L closepath p .955 .786 .906 r
F P
0 g
s
.28313 .03434 m .28218 .09692 L .28299 .03188 L closepath p .955 .786 .906 r
F P
0 g
s
.77801 .03475 m .78105 .03261 L .77779 .03189 L closepath p .305 .281 .743 r
F P
0 g
s
.78083 .02984 m .77779 .03189 L .78105 .03261 L closepath p .305 .282 .743 r
F P
0 g
s
.71525 .03418 m .71813 .03203 L .71505 .03189 L closepath p .282 .269 .739 r
F P
0 g
s
.71793 .02984 m .71505 .03189 L .71813 .03203 L closepath p .282 .27 .74 r
F P
0 g
s
.38369 .0435 m .38266 .03665 L .39432 .03189 L closepath p .955 .786 .906 r
F P
0 g
s
.39432 .03189 m .39448 .03436 L .38369 .0435 L closepath p .955 .786 .906 r
F P
0 g
s
.1791 .03347 m .18091 .0319 L .18 .03268 L closepath p .955 .786 .906 r
F P
0 g
s
.17864 .03386 m .18091 .0319 L .1791 .03347 L closepath p .955 .786 .906 r
F P
0 g
s
.17864 .03386 m .1791 .03347 L .18091 .0319 L closepath p .955 .786 .906 r
F P
0 g
s
.14235 .03196 m .13939 .03094 L .1434 .03645 L closepath p 1 .858 .538 r
F P
0 g
s
.43992 .03442 m .4421 .03206 L .43977 .03206 L closepath p .168 .168 .643 r
F P
0 g
s
.43538 .08475 m .43377 .07352 L .43977 .03206 L closepath p .955 .786 .906 r
F P
0 g
s
.4388 .06767 m .43652 .09292 L .43977 .03206 L closepath p .955 .786 .906 r
F P
0 g
s
.43977 .03206 m .43992 .03442 L .43782 .10346 L closepath p .955 .786 .906 r
F P
0 g
s
.05988 .03348 m .05398 .0337 L .05344 .03224 L closepath p .955 .786 .906 r
F P
0 g
s
.05369 .03291 m .05344 .03224 L .05398 .0337 L closepath p .955 .786 .906 r
F P
0 g
s
.05369 .03291 m .05398 .0337 L .05344 .03224 L closepath p .955 .786 .906 r
F P
0 g
s
.50328 .0361 m .50563 .03393 L .50297 .03242 L closepath p .176 .176 .679 r
F P
0 g
s
.50533 .03038 m .50297 .03242 L .50563 .03393 L closepath p .176 .176 .68 r
F P
0 g
s
.80271 .02916 m .80385 .03257 L .80739 .02436 L closepath p 1 1 1 r
F P
0 g
s
.80385 .03257 m .80271 .02916 L .79875 .04308 L closepath p 1 1 .87 r
F P
0 g
s
.53293 .03026 m .53518 .03257 L .53384 .02603 L closepath p 1 1 .763 r
F P
0 g
s
.53518 .03257 m .53293 .03026 L .53446 .03759 L closepath p 1 1 .714 r
F P
0 g
s
.80385 .03257 m .80589 .04834 L .79875 .04308 L closepath p .955 .786 .906 r
F P
0 g
s
.80385 .03257 m .81417 .03494 L .80693 .04802 L closepath p .955 .786 .906 r
F P
0 g
s
.81001 .06349 m .81204 .09587 L .80385 .03257 L closepath p .955 .786 .906 r
F P
0 g
s
.58805 .04314 m .59061 .04095 L .58751 .03288 L closepath p .202 .19 .681 r
F P
0 g
s
.59008 .03083 m .58751 .03288 L .59061 .04095 L closepath p .202 .19 .681 r
F P
0 g
s
.58751 .03288 m .58805 .04314 L .58456 .10772 L closepath p .955 .786 .906 r
F P
0 g
s
.58078 .07116 m .58751 .03288 L .58456 .10772 L closepath p .955 .786 .906 r
F P
0 g
s
.06031 .03849 m .06155 .03625 L .05988 .03348 L closepath p .122 .122 .496 r
F P
0 g
s
.06114 .03144 m .05988 .03348 L .06155 .03625 L closepath p .122 .122 .495 r
F P
0 g
s
.05988 .03348 m .06031 .03849 L .0554 .03818 L closepath p .955 .786 .906 r
F P
0 g
s
.05988 .03348 m .0554 .03818 L .05398 .0337 L closepath p .955 .786 .906 r
F P
0 g
s
.48941 .04048 m .49161 .03778 L .48598 .03367 L closepath p .355 .459 .92 r
F P
0 g
s
.48804 .0306 m .48598 .03367 L .49161 .03778 L closepath p .358 .46 .919 r
F P
0 g
s
.48941 .04048 m .48669 .03645 L .48598 .03367 L closepath p .955 .786 .906 r
F P
0 g
s
.05067 .03252 m .05398 .0337 L .04982 .03038 L closepath p 1 .781 .465 r
F P
0 g
s
.05398 .0337 m .05067 .03252 L .0554 .03818 L closepath p 1 .825 .519 r
F P
0 g
s
.05463 .03568 m .05398 .0337 L .0554 .03818 L closepath p .955 .786 .906 r
F P
0 g
s
.05463 .03568 m .0554 .03818 L .05398 .0337 L closepath p .955 .786 .906 r
F P
0 g
s
.23544 .0471 m .23706 .0456 L .23914 .03379 L closepath p .102 .102 .2 r
F P
0 g
s
.24074 .03235 m .23914 .03379 L .23706 .0456 L closepath p .102 .102 .2 r
F P
0 g
s
.22535 .04054 m .22603 .03714 L .23914 .03379 L closepath p .955 .786 .906 r
F P
0 g
s
.21201 .09872 m .22441 .04424 L .23914 .03379 L closepath p .955 .786 .906 r
F P
0 g
s
.22535 .04054 m .23914 .03379 L .22441 .04424 L closepath p .955 .786 .906 r
F P
0 g
s
.21201 .09872 m .23914 .03379 L .23544 .0471 L closepath p .955 .786 .906 r
F P
0 g
s
.23729 .04012 m .23544 .0471 L .23914 .03379 L closepath p .955 .786 .906 r
F P
0 g
s
.23729 .04012 m .23914 .03379 L .23544 .0471 L closepath p .955 .786 .906 r
F P
0 g
s
.93017 .04718 m .93355 .04568 L .93318 .03379 L closepath p .27 .187 .581 r
F P
0 g
s
.93654 .03235 m .93318 .03379 L .93355 .04568 L closepath p .27 .187 .581 r
F P
0 g
s
.64525 .09872 m .66866 .04016 L .66371 .06297 L closepath p .955 .786 .906 r
F P
0 g
s
.667 .04714 m .66371 .06297 L .66866 .04016 L closepath p .955 .786 .906 r
F P
0 g
s
.667 .04714 m .66866 .04016 L .66371 .06297 L closepath p .955 .786 .906 r
F P
0 g
s
.667 .04714 m .66971 .04565 L .67032 .03379 L closepath p .164 .164 .47 r
F P
0 g
s
.67301 .03235 m .67032 .03379 L .66971 .04565 L closepath p .164 .164 .47 r
F P
0 g
s
.65591 .04424 m .65725 .03715 L .67032 .03379 L closepath p .955 .786 .906 r
F P
0 g
s
.67032 .03379 m .66866 .04016 L .65591 .04424 L closepath p .955 .786 .906 r
F P
0 g
s
.93318 .03379 m .93167 .04019 L .91896 .04424 L closepath p .955 .786 .906 r
F P
0 g
s
.09327 .03147 m .09659 .0338 L .09431 .02608 L closepath p 1 1 .827 r
F P
0 g
s
.09659 .0338 m .09327 .03147 L .09556 .03969 L closepath p 1 1 .819 r
F P
0 g
s
.5352 .03384 m .53452 .03861 L .53323 .04495 L closepath p .955 .786 .906 r
F P
0 g
s
.53323 .04495 m .53446 .03759 L .5352 .03384 L closepath p .955 .786 .906 r
F P
0 g
s
.53323 .04495 m .5352 .03384 L .53446 .03759 L closepath p .955 .786 .906 r
F P
0 g
s
.15679 .04581 m .15829 .04364 L .15657 .0341 L closepath p .128 .128 .478 r
F P
0 g
s
.15806 .03206 m .15657 .0341 L .15829 .04364 L closepath p .128 .128 .478 r
F P
0 g
s
.15569 .07012 m .15657 .0341 L .15679 .04581 L closepath p .955 .786 .906 r
F P
0 g
s
.15173 .11652 m .15657 .0341 L .1548 .10632 L closepath p .955 .786 .906 r
F P
0 g
s
.15096 .11907 m .15657 .0341 L .15173 .11652 L closepath p .955 .786 .906 r
F P
0 g
s
.48447 .03417 m .48669 .03645 L .48873 .04439 L closepath p .955 .786 .906 r
F P
0 g
s
.71584 .03854 m .71871 .0363 L .71525 .03418 L closepath p .3 .298 .766 r
F P
0 g
s
.71813 .03203 m .71525 .03418 L .71871 .0363 L closepath p .3 .298 .766 r
F P
0 g
s
.17819 .03425 m .19153 .06312 L .18313 .04509 L closepath p .955 .786 .906 r
F P
0 g
s
.18486 .04867 m .17819 .03425 L .18235 .03471 L closepath p .955 .786 .906 r
F P
0 g
s
.28218 .09692 m .28313 .03434 L .28381 .08653 L closepath p .955 .786 .906 r
F P
0 g
s
.28358 .03876 m .28381 .08653 L .28313 .03434 L closepath p .955 .786 .906 r
F P
0 g
s
.39496 .03878 m .39704 .03654 L .39448 .03436 L closepath p .168 .168 .667 r
F P
0 g
s
.39448 .03436 m .39496 .03878 L .38443 .0487 L closepath p .955 .786 .906 r
F P
0 g
s
.38443 .0487 m .38369 .0435 L .39448 .03436 L closepath p .955 .786 .906 r
F P
0 g
s
.44461 .06737 m .44653 .06313 L .43992 .03442 L closepath p .175 .197 .707 r
F P
0 g
s
.4421 .03206 m .43992 .03442 L .44653 .06313 L closepath p .175 .197 .707 r
F P
0 g
s
.43992 .03442 m .44226 .05088 L .43798 .10582 L closepath p .955 .786 .906 r
F P
0 g
s
.43798 .10582 m .43782 .10346 L .43992 .03442 L closepath p .955 .786 .906 r
F P
0 g
s
.61549 .03953 m .618 .03683 L .61357 .03454 L closepath p .357 .425 .883 r
F P
0 g
s
.61616 .03215 m .61357 .03454 L .618 .03683 L closepath p .354 .425 .885 r
F P
0 g
s
.61357 .03454 m .61549 .03953 L .61649 .04867 L closepath p .955 .786 .906 r
F P
0 g
s
.18409 .03976 m .18554 .03705 L .18235 .03471 L closepath p .167 .254 .781 r
F P
0 g
s
.18235 .03471 m .18409 .03976 L .18486 .04867 L closepath p .955 .786 .906 r
F P
0 g
s
.77859 .0393 m .78162 .03707 L .77801 .03475 L closepath p .324 .314 .773 r
F P
0 g
s
.78105 .03261 m .77801 .03475 L .78162 .03707 L closepath p .324 .314 .773 r
F P
0 g
s
.81028 .04388 m .8133 .04248 L .81417 .03494 L closepath p .161 .161 .375 r
F P
0 g
s
.81028 .04388 m .80693 .04802 L .81417 .03494 L closepath p .955 .786 .906 r
F P
0 g
s
.09658 .03525 m .09566 .04068 L .09389 .04816 L closepath p .955 .786 .906 r
F P
0 g
s
.09389 .04816 m .09556 .03969 L .09658 .03525 L closepath p .955 .786 .906 r
F P
0 g
s
.09389 .04816 m .09658 .03525 L .09556 .03969 L closepath p .955 .786 .906 r
F P
0 g
s
.54584 .03585 m .55779 .05825 L .54602 .04317 L closepath p .955 .786 .906 r
F P
0 g
s
.55293 .05636 m .54602 .04317 L .55779 .05825 L closepath p .955 .786 .906 r
F P
0 g
s
.55778 .05825 m .56001 .05522 L .54584 .03585 L closepath p .392 .491 .938 r
F P
0 g
s
.54806 .03281 m .54584 .03585 L .56001 .05522 L closepath p .392 .491 .938 r
F P
0 g
s
.5249 .08878 m .52683 .07932 L .54584 .03585 L closepath p .955 .786 .906 r
F P
0 g
s
.50419 .04106 m .5065 .03876 L .50328 .0361 L closepath p .214 .251 .749 r
F P
0 g
s
.50563 .03393 m .50328 .0361 L .5065 .03876 L closepath p .215 .251 .749 r
F P
0 g
s
.48099 .03612 m .47909 .03547 L .48624 .04232 L closepath p .846 .353 .092 r
F P
0 g
s
.48099 .03612 m .48873 .04439 L .48624 .04232 L closepath p .955 .786 .906 r
F P
0 g
s
.48964 .04393 m .48873 .04439 L .48669 .03645 L closepath p .955 .786 .906 r
F P
0 g
s
.48669 .03645 m .48941 .04048 L .48964 .04393 L closepath p .955 .786 .906 r
F P
0 g
s
.14019 .03476 m .1434 .03645 L .13939 .03094 L closepath p 1 .875 .568 r
F P
0 g
s
.1434 .03645 m .14019 .03476 L .14393 .04092 L closepath p 1 .923 .634 r
F P
0 g
s
.22303 .03482 m .22603 .03714 L .22361 .02923 L closepath p 1 1 .758 r
F P
0 g
s
.22603 .03714 m .22303 .03482 L .22441 .04424 L closepath p 1 1 .827 r
F P
0 g
s
.65725 .03715 m .65532 .03483 L .65591 .04424 L closepath p 1 1 .713 r
F P
0 g
s
.1218 .063 m .12297 .05998 L .10897 .03737 L closepath p .195 .355 .864 r
F P
0 g
s
.11012 .03431 m .10897 .03737 L .12297 .05998 L closepath p .195 .355 .864 r
F P
0 g
s
.08294 .09799 m .08547 .0872 L .10897 .03737 L closepath p .955 .786 .906 r
F P
0 g
s
.10897 .03737 m .08106 .10519 L .08294 .09799 L closepath p .955 .786 .906 r
F P
0 g
s
.08036 .10745 m .08106 .10519 L .10897 .03737 L closepath p .955 .786 .906 r
F P
0 g
s
.09505 .0712 m .10897 .03737 L .08106 .10519 L closepath p .955 .786 .906 r
F P
0 g
s
.08008 .10823 m .08036 .10745 L .10897 .03737 L closepath p .955 .786 .906 r
F P
0 g
s
.0947 .07232 m .10897 .03737 L .08036 .10745 L closepath p .955 .786 .906 r
F P
0 g
s
.0947 .07232 m .08036 .10745 L .10897 .03737 L closepath p .955 .786 .906 r
F P
0 g
s
.10897 .03737 m .1218 .063 L .14499 .11211 L closepath p .955 .786 .906 r
F P
0 g
s
.10897 .03737 m .14499 .11211 L .10872 .04584 L closepath p .955 .786 .906 r
F P
0 g
s
.10897 .03737 m .10872 .04584 L .09456 .07271 L closepath p .955 .786 .906 r
F P
0 g
s
.53222 .03524 m .53446 .03759 L .53293 .03026 L closepath p 1 1 .714 r
F P
0 g
s
.53446 .03759 m .53222 .03524 L .53323 .04495 L closepath p 1 1 .731 r
F P
0 g
s
.05203 .03683 m .0554 .03818 L .05067 .03252 L closepath p 1 .826 .522 r
F P
0 g
s
.0554 .03818 m .0675 .07879 L .06142 .05847 L closepath p .955 .786 .906 r
F P
0 g
s
.06144 .05845 m .0554 .03818 L .06031 .03849 L closepath p .955 .786 .906 r
F P
0 g
s
.06031 .03849 m .06092 .04204 L .06144 .05845 L closepath p .955 .786 .906 r
F P
0 g
s
.0675 .07879 m .06144 .05845 L .06092 .04204 L closepath p .955 .786 .906 r
F P
0 g
s
.0554 .03818 m .05203 .03683 L .0675 .07879 L closepath p 1 .836 .534 r
F P
0 g
s
.0618 .04628 m .06303 .04386 L .06031 .03849 L closepath p .137 .137 .609 r
F P
0 g
s
.06155 .03625 m .06031 .03849 L .06303 .04386 L closepath p .136 .136 .608 r
F P
0 g
s
.7168 .04497 m .71967 .04268 L .71584 .03854 L closepath p .306 .309 .775 r
F P
0 g
s
.71871 .0363 m .71584 .03854 L .71967 .04268 L closepath p .306 .309 .775 r
F P
0 g
s
.53452 .03861 m .53359 .0447 L .5315 .05465 L closepath p .955 .786 .906 r
F P
0 g
s
.5315 .05465 m .53323 .04495 L .53452 .03861 L closepath p .955 .786 .906 r
F P
0 g
s
.28381 .08653 m .28358 .03876 L .28543 .07614 L closepath p .955 .786 .906 r
F P
0 g
s
.28434 .04514 m .28543 .07614 L .28358 .03876 L closepath p .955 .786 .906 r
F P
0 g
s
.39578 .04516 m .39785 .04287 L .39496 .03878 L closepath p .169 .169 .682 r
F P
0 g
s
.39704 .03654 m .39496 .03878 L .39785 .04287 L closepath p .169 .169 .682 r
F P
0 g
s
.39496 .03878 m .39578 .04516 L .38517 .0539 L closepath p .955 .786 .906 r
F P
0 g
s
.38517 .0539 m .38443 .0487 L .39496 .03878 L closepath p .955 .786 .906 r
F P
0 g
s
.86317 .0463 m .86636 .04383 L .86106 .03884 L closepath p .408 .428 .864 r
F P
0 g
s
.86421 .03622 m .86106 .03884 L .86636 .04383 L closepath p .408 .428 .864 r
F P
0 g
s
.86288 .04745 m .86242 .04783 L .86106 .03884 L closepath p .955 .786 .906 r
F P
0 g
s
.86106 .03884 m .86317 .0463 L .86288 .04745 L closepath p .955 .786 .906 r
F P
0 g
s
.77953 .04554 m .78255 .04325 L .77859 .0393 L closepath p .333 .329 .786 r
F P
0 g
s
.78162 .03707 m .77859 .0393 L .78255 .04325 L closepath p .333 .329 .787 r
F P
0 g
s
.61868 .0461 m .62109 .04314 L .61549 .03953 L closepath p .396 .48 .926 r
F P
0 g
s
.618 .03683 m .61549 .03953 L .62109 .04314 L closepath p .394 .48 .927 r
F P
0 g
s
.62361 .06312 m .61649 .04867 L .61549 .03953 L closepath p .955 .786 .906 r
F P
0 g
s
.61549 .03953 m .61868 .0461 L .62361 .06312 L closepath p .955 .786 .906 r
F P
0 g
s
.88168 .04615 m .88474 .04322 L .87841 .03961 L closepath p .476 .533 .948 r
F P
0 g
s
.88157 .03691 m .87841 .03961 L .88474 .04322 L closepath p .474 .533 .949 r
F P
0 g
s
.09223 .03734 m .09556 .03969 L .09327 .03147 L closepath p 1 1 .819 r
F P
0 g
s
.09556 .03969 m .09223 .03734 L .09389 .04816 L closepath p 1 1 .829 r
F P
0 g
s
.18703 .04627 m .1884 .04331 L .18409 .03976 L closepath p .202 .342 .851 r
F P
0 g
s
.18554 .03705 m .18409 .03976 L .1884 .04331 L closepath p .198 .341 .851 r
F P
0 g
s
.19153 .06312 m .18486 .04867 L .18409 .03976 L closepath p .955 .786 .906 r
F P
0 g
s
.18409 .03976 m .18703 .04627 L .19153 .06312 L closepath p .955 .786 .906 r
F P
0 g
s
.3434 .04988 m .34534 .04751 L .34221 .04003 L closepath p .163 .163 .659 r
F P
0 g
s
.34534 .04751 m .3434 .04988 L .34677 .05628 L closepath p .167 .169 .693 r
F P
0 g
s
.34534 .04751 m .34677 .05628 L .34671 .05598 L p .719 .495 .006 r
F P
0 g
s
.34383 .03819 m .34221 .04003 L .34534 .04751 L p .163 .163 .659 r
F P
0 g
s
.34671 .05598 m .34444 .04417 L .34534 .04751 L p .719 .495 .006 r
F P
0 g
s
.34534 .04751 m .34416 .03782 L .34383 .03819 L p .163 .163 .659 r
F P
0 g
s
.3434 .04988 m .33683 .0848 L .34221 .04003 L closepath p .955 .786 .906 r
F P
0 g
s
.64525 .09872 m .65591 .04424 L .66866 .04016 L closepath p .955 .786 .906 r
F P
0 g
s
.93167 .04019 m .93017 .04718 L .90168 .09211 L closepath p .955 .786 .906 r
F P
0 g
s
.9161 .05708 m .91896 .04424 L .93167 .04019 L closepath p .955 .786 .906 r
F P
0 g
s
.90168 .09211 m .9161 .05708 L .93167 .04019 L closepath p .955 .786 .906 r
F P
0 g
s
.49055 .04347 m .49287 .04145 L .48941 .04048 L closepath p .281 .373 .857 r
F P
0 g
s
.49161 .03778 m .48941 .04048 L .49287 .04145 L closepath p .289 .365 .846 r
F P
0 g
s
.49055 .04347 m .48964 .04393 L .48941 .04048 L closepath p .955 .786 .906 r
F P
0 g
s
.09566 .04068 m .09442 .04756 L .09159 .0592 L closepath p .955 .786 .906 r
F P
0 g
s
.09159 .0592 m .09389 .04816 L .09566 .04068 L closepath p .955 .786 .906 r
F P
0 g
s
.14071 .03921 m .14393 .04092 L .14019 .03476 L closepath p 1 .923 .634 r
F P
0 g
s
.14393 .04092 m .14071 .03921 L .14457 .04739 L closepath p 1 .931 .646 r
F P
0 g
s
.14393 .04092 m .15402 .04279 L .14457 .04739 L closepath p .955 .786 .906 r
F P
0 g
s
.50571 .04733 m .50793 .04482 L .50419 .04106 L closepath p .232 .288 .783 r
F P
0 g
s
.5065 .03876 m .50419 .04106 L .50793 .04482 L closepath p .238 .289 .782 r
F P
0 g
s
.06092 .04204 m .0618 .04628 L .0675 .07879 L closepath p .955 .786 .906 r
F P
0 g
s
.85721 .0421 m .85624 .04141 L .86076 .0464 L closepath p .655 .172 .03 r
F P
0 g
s
.48491 .04216 m .48624 .04232 L .47909 .03547 L closepath p .808 .309 .082 r
F P
0 g
s
.48624 .04232 m .48491 .04216 L .48749 .04336 L closepath p .687 .141 .067 r
F P
0 g
s
.15215 .05947 m .14457 .04739 L .15402 .04279 L closepath p .955 .786 .906 r
F P
0 g
s
.79727 .04038 m .79875 .04308 L .80271 .02916 L closepath p 1 1 .868 r
F P
0 g
s
.79875 .04308 m .79727 .04038 L .79725 .05646 L closepath p 1 .975 .583 r
F P
0 g
s
.79875 .04308 m .80076 .0564 L .79725 .05646 L closepath p .955 .786 .906 r
F P
0 g
s
.80076 .0564 m .79875 .04308 L .80589 .04834 L closepath p .955 .786 .906 r
F P
0 g
s
.58955 .05562 m .5921 .05333 L .58805 .04314 L closepath p .233 .242 .729 r
F P
0 g
s
.59061 .04095 m .58805 .04314 L .5921 .05333 L closepath p .233 .241 .729 r
F P
0 g
s
.58871 .09299 m .58456 .10772 L .58805 .04314 L closepath p .955 .786 .906 r
F P
0 g
s
.58805 .04314 m .58955 .05562 L .58871 .09299 L closepath p .955 .786 .906 r
F P
0 g
s
.5427 .05883 m .54488 .05759 L .54602 .04317 L closepath p .146 .146 .414 r
F P
0 g
s
.54602 .04317 m .54654 .04416 L .55293 .05636 L closepath p 0 0 .328 r
F P
0 g
s
.54654 .04416 m .54602 .04317 L .54488 .05759 L closepath p .06 .06 .133 r
F P
0 g
s
.54602 .04317 m .5427 .05883 L .52442 .11487 L closepath p .955 .786 .906 r
F P
0 g
s
.52044 .10403 m .54602 .04317 L .52442 .11487 L closepath p .955 .786 .906 r
F P
0 g
s
.48708 .044 m .48749 .04336 L .48491 .04216 L closepath p .699 .156 .07 r
F P
0 g
s
.48749 .04336 m .48708 .044 L .48873 .04439 L closepath p .694 .149 .07 r
F P
0 g
s
.48964 .04393 m .49146 .04379 L .49055 .04347 L closepath p .065 .065 .065 r
F P
0 g
s
.49287 .04145 m .49055 .04347 L .49146 .04379 L closepath p .086 .086 .086 r
F P
0 g
s
.38369 .04351 m .38103 .04154 L .38517 .0539 L closepath p 1 .888 .537 r
F P
0 g
s
.80693 .04802 m .81028 .04388 L .81001 .06349 L closepath p .955 .786 .906 r
F P
0 g
s
.80915 .05455 m .81001 .06349 L .81028 .04388 L closepath p .955 .786 .906 r
F P
0 g
s
.80915 .05455 m .81028 .04388 L .81001 .06349 L closepath p .955 .786 .906 r
F P
0 g
s
.80915 .05455 m .81227 .05256 L .81028 .04388 L closepath p .249 .186 .627 r
F P
0 g
s
.8133 .04248 m .81028 .04388 L .81227 .05256 L closepath p .243 .185 .626 r
F P
0 g
s
.48873 .04439 m .4887 .04465 L .48964 .04393 L closepath p .021 .021 .021 r
F P
0 g
s
.49146 .04379 m .48964 .04393 L .4887 .04465 L closepath p .039 .039 .039 r
F P
0 g
s
.22145 .04169 m .22441 .04424 L .22303 .03482 L closepath p 1 1 .829 r
F P
0 g
s
.22441 .04424 m .22145 .04169 L .22267 .05065 L closepath p 1 1 .85 r
F P
0 g
s
.22441 .04424 m .21201 .09872 L .22092 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.65401 .04169 m .65591 .04424 L .65532 .03483 L closepath p 1 1 .714 r
F P
0 g
s
.65591 .04424 m .65401 .04169 L .65436 .05066 L closepath p 1 1 .752 r
F P
0 g
s
.65591 .04424 m .64525 .09872 L .65281 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.4887 .04465 m .48873 .04439 L .48708 .044 L closepath p .35 .024 .024 r
F P
0 g
s
.52925 .06672 m .5315 .05465 L .53359 .0447 L closepath p .955 .786 .906 r
F P
0 g
s
.53099 .04258 m .53323 .04495 L .53222 .03524 L closepath p 1 1 .731 r
F P
0 g
s
.53323 .04495 m .53099 .04258 L .5315 .05465 L closepath p 1 1 .74 r
F P
0 g
s
.71814 .05348 m .72101 .05117 L .7168 .04497 L closepath p .31 .314 .78 r
F P
0 g
s
.71967 .04268 m .7168 .04497 L .72101 .05117 L closepath p .31 .314 .78 r
F P
0 g
s
.18313 .04509 m .18019 .04404 L .18923 .05834 L closepath p 1 .719 .367 r
F P
0 g
s
.18313 .04509 m .19372 .06796 L .18923 .05834 L closepath p .955 .786 .906 r
F P
0 g
s
.19372 .06796 m .18313 .04509 L .19153 .06312 L closepath p .955 .786 .906 r
F P
0 g
s
.87768 .04509 m .87652 .04403 L .88439 .05834 L closepath p .815 .414 .053 r
F P
0 g
s
.61465 .04509 m .61282 .04404 L .62116 .05834 L closepath p .966 .563 .104 r
F P
0 g
s
.61465 .04509 m .62595 .06796 L .62116 .05834 L closepath p .955 .786 .906 r
F P
0 g
s
.62595 .06796 m .61465 .04509 L .62361 .06312 L closepath p .955 .786 .906 r
F P
0 g
s
.88439 .05834 m .87768 .04509 L .8865 .05471 L closepath p .955 .786 .906 r
F P
0 g
s
.88636 .05448 m .8865 .05471 L .87768 .04509 L closepath p .955 .786 .906 r
F P
0 g
s
.28543 .07614 m .28434 .04514 L .28706 .06578 L closepath p .955 .786 .906 r
F P
0 g
s
.28543 .05348 m .28706 .06578 L .28434 .04514 L closepath p .955 .786 .906 r
F P
0 g
s
.28543 .05348 m .28434 .04514 L .28706 .06578 L closepath p .955 .786 .906 r
F P
0 g
s
.28543 .05348 m .28722 .05116 L .28434 .04514 L closepath p .158 .158 .645 r
F P
0 g
s
.39693 .05351 m .399 .0512 L .39578 .04516 L closepath p .17 .173 .69 r
F P
0 g
s
.39785 .04287 m .39578 .04516 L .399 .0512 L closepath p .17 .173 .69 r
F P
0 g
s
.39578 .04516 m .39693 .05351 L .38614 .06087 L closepath p .955 .786 .906 r
F P
0 g
s
.38614 .06087 m .38517 .0539 L .39578 .04516 L closepath p .955 .786 .906 r
F P
0 g
s
.78083 .05348 m .78385 .05116 L .77953 .04554 L closepath p .338 .337 .794 r
F P
0 g
s
.78255 .04325 m .77953 .04554 L .78385 .05116 L closepath p .338 .337 .794 r
F P
0 g
s
.15799 .06006 m .15948 .05778 L .15679 .04581 L closepath p .137 .137 .546 r
F P
0 g
s
.15829 .04364 m .15679 .04581 L .15948 .05778 L closepath p .137 .137 .546 r
F P
0 g
s
.15679 .04581 m .15799 .06006 L .15569 .07012 L closepath p .955 .786 .906 r
F P
0 g
s
.10635 .05545 m .10675 .05468 L .10838 .04737 L p .02 .02 .02 r
F P
0 g
s
.10836 .04728 m .10635 .05545 L p .10838 .04737 L .02 .02 .02 r
F P
0 g
s
.10838 .04737 m .10872 .04584 L .10836 .04728 L p .02 .02 .02 r
F P
0 g
s
.10872 .04584 m .10635 .05545 L .0814 .12786 L closepath p .955 .786 .906 r
F P
0 g
s
.10872 .04584 m .0814 .12786 L .07717 .11545 L closepath p .955 .786 .906 r
F P
0 g
s
.08008 .10823 m .09456 .07271 L .10872 .04584 L closepath p .955 .786 .906 r
F P
0 g
s
.10872 .04584 m .10479 .05451 L .08008 .10823 L closepath p .955 .786 .906 r
F P
0 g
s
.10804 .04681 m .10872 .04584 L .10838 .04737 L p .873 .806 .29 r
F P
0 g
s
.10838 .04737 m .10675 .05468 L .10804 .04681 L p .873 .806 .29 r
F P
0 g
s
.10209 .03455 m .10804 .04681 L .10477 .04823 L p 1 .95 .512 r
F P
0 g
s
.10477 .04823 m .09945 .05053 L .10209 .03455 L p 1 .95 .512 r
F P
0 g
s
.10209 .03455 m .10078 .02945 L .10143 .03113 L p .072 .072 .53 r
F P
0 g
s
.10804 .04681 m .10209 .03455 L .10648 .0442 L p .689 .337 .011 r
F P
0 g
s
.07995 .10855 m .08008 .10823 L .10479 .05451 L closepath p .955 .786 .906 r
F P
0 g
s
.07995 .10855 m .10479 .05451 L .09299 .08056 L closepath p .955 .786 .906 r
F P
0 g
s
.10086 .06318 m .09299 .08056 L .10479 .05451 L closepath p .955 .786 .906 r
F P
0 g
s
.10086 .06318 m .10479 .05451 L .09299 .08056 L closepath p .955 .786 .906 r
F P
0 g
s
.08301 .12799 m .0814 .12786 L .10635 .05545 L closepath p .955 .786 .906 r
F P
0 g
s
.10635 .05545 m .10438 .0638 L .08301 .12799 L closepath p .955 .786 .906 r
F P
0 g
s
.10439 .0638 m .10545 .06256 L .10635 .05545 L closepath p .07 .07 .1 r
F P
0 g
s
.10675 .05468 m .10635 .05545 L .10545 .06256 L closepath p .026 .026 .026 r
F P
0 g
s
.10804 .04681 m .10545 .06256 L .10416 .05997 L p 1 .95 .512 r
F P
0 g
s
.10416 .05997 m .09945 .05053 L .10312 .04894 L p 1 .95 .512 r
F P
0 g
s
.10312 .04894 m .10804 .04681 L p .10416 .05997 L 1 .95 .512 r
F P
0 g
s
.10648 .0442 m .1087 .04909 L p .10804 .04681 L .689 .337 .011 r
F P
0 g
s
.10845 .04777 m .10804 .04681 L p .1087 .04909 L .689 .337 .011 r
F P
0 g
s
.10872 .04584 m .10804 .04681 L .11609 .06096 L closepath p .936 .534 .092 r
F P
0 g
s
.11403 .06083 m .11609 .06096 L .10804 .04681 L closepath p 1 .697 .301 r
F P
0 g
s
.10817 .04878 m .11403 .06083 L .10209 .03455 L closepath p .689 .337 .011 r
F P
0 g
s
.1083 .04485 m .10817 .04878 L .10143 .03113 L p .144 .231 .781 r
F P
0 g
s
.10143 .03113 m .10078 .02945 L .10138 .03068 L p .144 .231 .781 r
F P
0 g
s
.10138 .03068 m .1083 .04485 L p .10143 .03113 L .144 .231 .781 r
F P
0 g
s
.10143 .03113 m .10817 .04878 L .10209 .03455 L p .072 .072 .53 r
F P
0 g
s
.1087 .04909 m .11403 .06083 L .10845 .04777 L p .689 .337 .011 r
F P
0 g
s
.10872 .04584 m .1502 .12162 L .11609 .06096 L closepath p .955 .786 .906 r
F P
0 g
s
.13979 .1026 m .1294 .08363 L .10872 .04584 L closepath p .955 .786 .906 r
F P
0 g
s
.11905 .06471 m .10872 .04584 L .1294 .08363 L closepath p .955 .786 .906 r
F P
0 g
s
.13979 .1026 m .10872 .04584 L .14499 .11211 L closepath p .955 .786 .906 r
F P
0 g
s
.62314 .05425 m .62549 .05116 L .61868 .0461 L closepath p .418 .511 .949 r
F P
0 g
s
.62109 .04314 m .61868 .0461 L .62549 .05116 L closepath p .418 .511 .949 r
F P
0 g
s
.63074 .0776 m .62361 .06312 L .61868 .0461 L closepath p .955 .786 .906 r
F P
0 g
s
.61868 .0461 m .62314 .05425 L .63074 .0776 L closepath p .955 .786 .906 r
F P
0 g
s
.88622 .05425 m .88923 .05119 L .88168 .04615 L closepath p .493 .559 .967 r
F P
0 g
s
.88474 .04322 m .88168 .04615 L .88923 .05119 L closepath p .493 .559 .967 r
F P
0 g
s
.19117 .05425 m .19249 .05116 L .18703 .04627 L closepath p .242 .392 .888 r
F P
0 g
s
.1884 .04331 m .18703 .04627 L .19249 .05116 L closepath p .241 .392 .888 r
F P
0 g
s
.19822 .0776 m .19153 .06312 L .18703 .04627 L closepath p .955 .786 .906 r
F P
0 g
s
.18703 .04627 m .19117 .05425 L .19822 .0776 L closepath p .955 .786 .906 r
F P
0 g
s
.06436 .05686 m .06557 .0543 L .0618 .04628 L closepath p .143 .143 .658 r
F P
0 g
s
.06303 .04386 m .0618 .04628 L .06557 .0543 L closepath p .143 .143 .657 r
F P
0 g
s
.06436 .05686 m .0675 .07879 L .0618 .04628 L closepath p .955 .786 .906 r
F P
0 g
s
.86335 .04707 m .8666 .04483 L .86317 .0463 L closepath p .393 .401 .842 r
F P
0 g
s
.86636 .04383 m .86317 .0463 L .8666 .04483 L closepath p .393 .401 .842 r
F P
0 g
s
.86335 .04707 m .86288 .04745 L .86317 .0463 L closepath p .955 .786 .906 r
F P
0 g
s
.86037 .04622 m .86076 .0464 L .85624 .04141 L closepath p .599 .115 .017 r
F P
0 g
s
.86076 .0464 m .86037 .04622 L .86242 .04783 L closepath p .505 .012 .012 r
F P
0 g
s
.86288 .04745 m .86556 .04714 L .86335 .04707 L closepath p .121 .121 .121 r
F P
0 g
s
.8666 .04483 m .86335 .04707 L .86556 .04714 L closepath p .154 .154 .336 r
F P
0 g
s
.23172 .06296 m .23335 .06141 L .23544 .0471 L closepath p .108 .108 .247 r
F P
0 g
s
.23706 .0456 m .23544 .0471 L .23335 .06141 L closepath p .108 .108 .247 r
F P
0 g
s
.21635 .11543 m .23544 .0471 L .23172 .06296 L closepath p .955 .786 .906 r
F P
0 g
s
.21635 .11543 m .21201 .09872 L .23544 .0471 L closepath p .955 .786 .906 r
F P
0 g
s
.66371 .06297 m .66644 .06142 L .667 .04714 L closepath p .168 .168 .503 r
F P
0 g
s
.66971 .04565 m .667 .04714 L .66644 .06142 L closepath p .167 .167 .503 r
F P
0 g
s
.92719 .06297 m .90168 .09211 L .93017 .04718 L closepath p .955 .786 .906 r
F P
0 g
s
.92719 .06297 m .93058 .06143 L .93017 .04718 L closepath p .282 .19 .605 r
F P
0 g
s
.93355 .04568 m .93017 .04718 L .93058 .06143 L closepath p .281 .19 .605 r
F P
0 g
s
.50784 .0549 m .50989 .05211 L .50571 .04733 L closepath p .233 .303 .8 r
F P
0 g
s
.50793 .04482 m .50571 .04733 L .50989 .05211 L closepath p .245 .307 .8 r
F P
0 g
s
.50571 .04733 m .50784 .0549 L .50352 .04989 L closepath p .955 .786 .906 r
F P
0 g
s
.14134 .04565 m .14457 .04739 L .14071 .03921 L closepath p 1 .931 .646 r
F P
0 g
s
.14457 .04739 m .14134 .04565 L .14533 .05585 L closepath p 1 .936 .652 r
F P
0 g
s
.14457 .04739 m .15215 .05947 L .14533 .05585 L closepath p .955 .786 .906 r
F P
0 g
s
.86242 .04783 m .86338 .04809 L .86288 .04745 L closepath p .074 .074 .074 r
F P
0 g
s
.86556 .04714 m .86288 .04745 L .86338 .04809 L closepath p .072 .072 .072 r
F P
0 g
s
.08867 .07285 m .09159 .0592 L .09442 .04756 L closepath p .955 .786 .906 r
F P
0 g
s
.86338 .04809 m .86242 .04783 L .86037 .04622 L closepath p .028 .272 .701 r
F P
0 g
s
.09057 .04579 m .09389 .04816 L .09223 .03734 L closepath p 1 1 .829 r
F P
0 g
s
.09389 .04816 m .09057 .04579 L .09159 .0592 L closepath p 1 1 .834 r
F P
0 g
s
.80589 .04834 m .80794 .06415 L .80076 .0564 L closepath p .955 .786 .906 r
F P
0 g
s
.27331 .04891 m .27042 .04727 L .27585 .06794 L closepath p 1 .905 .584 r
F P
0 g
s
.34629 .06785 m .3434 .04988 L .34487 .05888 L closepath p .955 .786 .906 r
F P
0 g
s
.34485 .05886 m .34677 .05628 L .3434 .04988 L closepath p .168 .171 .693 r
F P
0 g
s
.3434 .04988 m .34629 .06785 L .33683 .0848 L closepath p .955 .786 .906 r
F P
0 g
s
.50499 .05432 m .50352 .04989 L .50784 .0549 L closepath p .955 .786 .906 r
F P
0 g
s
.21978 .0479 m .22267 .05065 L .22145 .04169 L closepath p 1 1 .847 r
F P
0 g
s
.22267 .05065 m .21978 .0479 L .22092 .05708 L closepath p 1 1 .849 r
F P
0 g
s
.91753 .05066 m .9163 .0479 L .9161 .05708 L closepath p 1 1 .652 r
F P
0 g
s
.65249 .0479 m .65436 .05066 L .65401 .04169 L closepath p 1 1 .757 r
F P
0 g
s
.65436 .05066 m .65249 .0479 L .65281 .05708 L closepath p 1 1 .755 r
F P
0 g
s
.44461 .06737 m .43798 .10582 L .44226 .05088 L closepath p .955 .786 .906 r
F P
0 g
s
.46916 .05207 m .47055 .04804 L .46833 .05157 L closepath p .642 .795 1 r
F P
0 g
s
.46825 .0473 m .46833 .05157 L .47055 .04804 L closepath p .713 .89 1 r
F P
0 g
s
.46833 .05157 m .46825 .0473 L .46714 .05239 L closepath p 1 1 1 r
F P
0 g
s
.46833 .05157 m .46999 .05257 L .46714 .05239 L closepath p .955 .786 .906 r
F P
0 g
s
.46999 .05257 m .4722 .04999 L .46916 .05207 L closepath p .554 .767 1 r
F P
0 g
s
.47055 .04804 m .46916 .05207 L .4722 .04999 L closepath p .484 .618 1 r
F P
0 g
s
.46532 .04907 m .46714 .05239 L .46825 .0473 L closepath p 1 1 1 r
F P
0 g
s
.46714 .05239 m .46532 .04907 L .46674 .05485 L closepath p 1 1 .698 r
F P
0 g
s
.46674 .05485 m .46714 .05239 L .46999 .05257 L closepath p .955 .786 .906 r
F P
0 g
s
.47076 .06014 m .47326 .05847 L .46999 .05257 L closepath p .2 .209 .705 r
F P
0 g
s
.4722 .04999 m .46999 .05257 L .47326 .05847 L closepath p .176 .194 .701 r
F P
0 g
s
.46999 .05257 m .47076 .06014 L .46674 .05485 L closepath p .955 .786 .906 r
F P
0 g
s
.29197 .1028 m .29379 .10059 L .28543 .05348 L closepath p .158 .158 .648 r
F P
0 g
s
.28722 .05116 m .28543 .05348 L .29379 .10059 L closepath p .158 .158 .648 r
F P
0 g
s
.78399 .0722 m .78702 .06991 L .78083 .05348 L closepath p .34 .341 .797 r
F P
0 g
s
.78385 .05116 m .78083 .05348 L .78702 .06991 L closepath p .34 .341 .797 r
F P
0 g
s
.72604 .1028 m .72894 .1006 L .71814 .05348 L closepath p .311 .316 .781 r
F P
0 g
s
.72101 .05117 m .71814 .05348 L .72894 .1006 L closepath p .311 .316 .781 r
F P
0 g
s
.72011 .06578 m .72209 .0781 L .71814 .05348 L closepath p .955 .786 .906 r
F P
0 g
s
.39846 .06428 m .40053 .06199 L .39693 .05351 L closepath p .171 .176 .693 r
F P
0 g
s
.399 .0512 m .39693 .05351 L .40053 .06199 L closepath p .171 .176 .693 r
F P
0 g
s
.40031 .08414 m .39693 .05351 L .39846 .06428 L closepath p .955 .786 .906 r
F P
0 g
s
.38711 .06785 m .38614 .06087 L .39693 .05351 L closepath p .955 .786 .906 r
F P
0 g
s
.39693 .05351 m .40371 .11489 L .38771 .07168 L closepath p .955 .786 .906 r
F P
0 g
s
.38771 .07168 m .38711 .06785 L .39693 .05351 L closepath p .955 .786 .906 r
F P
0 g
s
.3824 .05115 m .38517 .0539 L .38103 .04154 L closepath p 1 .888 .536 r
F P
0 g
s
.38517 .0539 m .3824 .05115 L .38711 .06785 L closepath p 1 .89 .54 r
F P
0 g
s
.1931 .05669 m .19343 .05267 L .19117 .05425 L closepath p .326 .52 .982 r
F P
0 g
s
.19249 .05116 m .19117 .05425 L .19343 .05267 L closepath p .294 .456 .934 r
F P
0 g
s
.88834 .05669 m .89028 .05266 L .88622 .05425 L closepath p .539 .65 1 r
F P
0 g
s
.88923 .05119 m .88622 .05425 L .89028 .05266 L closepath p .529 .614 1 r
F P
0 g
s
.62509 .05669 m .62648 .05269 L .62314 .05425 L closepath p .466 .599 1 r
F P
0 g
s
.62549 .05116 m .62314 .05425 L .62648 .05269 L closepath p .449 .554 .979 r
F P
0 g
s
.19117 .05425 m .19172 .05517 L .19702 .06742 L closepath p .955 .786 .906 r
F P
0 g
s
.19117 .05425 m .19409 .06083 L .20493 .09211 L closepath p .955 .786 .906 r
F P
0 g
s
.20493 .09211 m .19822 .0776 L .19117 .05425 L closepath p .955 .786 .906 r
F P
0 g
s
.62314 .05425 m .62371 .05517 L .62774 .06406 L closepath p .955 .786 .906 r
F P
0 g
s
.62921 .06742 m .62314 .05425 L .62774 .06406 L closepath p .955 .786 .906 r
F P
0 g
s
.62314 .05425 m .62921 .06742 L .6305 .07315 L closepath p .955 .786 .906 r
F P
0 g
s
.63789 .09211 m .63074 .0776 L .62314 .05425 L closepath p .955 .786 .906 r
F P
0 g
s
.81001 .06349 m .81311 .06125 L .80915 .05455 L closepath p .323 .303 .76 r
F P
0 g
s
.81227 .05256 m .80915 .05455 L .81311 .06125 L closepath p .324 .303 .76 r
F P
0 g
s
.52926 .05227 m .5315 .05465 L .53099 .04258 L closepath p 1 1 .74 r
F P
0 g
s
.5315 .05465 m .52926 .05227 L .52925 .06672 L closepath p 1 1 .745 r
F P
0 g
s
.8865 .05471 m .88664 .05494 L .88439 .05834 L closepath p .955 .786 .906 r
F P
0 g
s
.4643 .05286 m .46674 .05485 L .46532 .04907 L closepath p 1 1 .806 r
F P
0 g
s
.46674 .05485 m .4643 .05286 L .46629 .0612 L closepath p 1 .994 .679 r
F P
0 g
s
.46629 .0612 m .46674 .05485 L .47076 .06014 L closepath p .955 .786 .906 r
F P
0 g
s
.51058 .06377 m .51237 .06059 L .50784 .0549 L closepath p .218 .302 .804 r
F P
0 g
s
.50989 .05211 m .50784 .0549 L .51237 .06059 L closepath p .239 .311 .806 r
F P
0 g
s
.88664 .05494 m .88679 .05517 L .89239 .074 L closepath p .955 .786 .906 r
F P
0 g
s
.89239 .074 m .88439 .05834 L .88664 .05494 L closepath p .955 .786 .906 r
F P
0 g
s
.82711 .01804 m .83206 .00538 L .8197 .01875 L closepath p .662 1 1 r
F P
0 g
s
.10116 .01916 m .09512 .00652 L .09922 .01804 L closepath p .119 .556 .979 r
F P
0 g
s
.09315 .00538 m .09922 .01804 L .09512 .00652 L closepath p .119 .556 .979 r
F P
0 g
s
.09922 .01804 m .09315 .00538 L .09722 .01904 L closepath p .902 1 1 r
F P
0 g
s
.19172 .05517 m .19222 .05589 L .19564 .06406 L closepath p .955 .786 .906 r
F P
0 g
s
.19564 .06406 m .19702 .06742 L .19172 .05517 L closepath p .955 .786 .906 r
F P
0 g
s
.62371 .05517 m .62423 .05589 L .62724 .06258 L closepath p .955 .786 .906 r
F P
0 g
s
.62724 .06258 m .62774 .06406 L .62371 .05517 L closepath p .955 .786 .906 r
F P
0 g
s
.90168 .09211 m .89239 .074 L .88679 .05517 L closepath p .955 .786 .906 r
F P
0 g
s
.90168 .09211 m .88679 .05517 L .88834 .05669 L closepath p .955 .786 .906 r
F P
0 g
s
.88733 .05589 m .88834 .05669 L .88679 .05517 L closepath p .955 .786 .906 r
F P
0 g
s
.88733 .05589 m .88679 .05517 L .88834 .05669 L closepath p .955 .786 .906 r
F P
0 g
s
.54118 .02044 m .54181 .00781 L .538 .01842 L closepath p .248 .616 1 r
F P
0 g
s
.53858 .00576 m .538 .01842 L .54181 .00781 L closepath p .248 .616 1 r
F P
0 g
s
.538 .01842 m .53858 .00576 L .53615 .01932 L closepath p .831 1 1 r
F P
0 g
s
.59096 .06524 m .59352 .06298 L .58955 .05562 L closepath p .246 .262 .748 r
F P
0 g
s
.5921 .05333 m .58955 .05562 L .59352 .06298 L closepath p .246 .262 .747 r
F P
0 g
s
.58955 .05562 m .59026 .06043 L .59286 .07829 L closepath p .955 .786 .906 r
F P
0 g
s
.59286 .07829 m .58871 .09299 L .58955 .05562 L closepath p .955 .786 .906 r
F P
0 g
s
.33283 .05573 m .33609 .05701 L .33399 .06411 L closepath p .955 .786 .906 r
F P
0 g
s
.82454 .0061 m .8197 .01875 L .83206 .00538 L closepath p .662 1 1 r
F P
0 g
s
.8197 .01875 m .82454 .0061 L .81309 .02087 L closepath p .737 1 1 r
F P
0 g
s
.1421 .05411 m .14533 .05585 L .14134 .04565 L closepath p 1 .936 .653 r
F P
0 g
s
.14533 .05585 m .1421 .05411 L .14621 .06633 L closepath p 1 .938 .657 r
F P
0 g
s
.14533 .05585 m .15027 .07619 L .14621 .06633 L closepath p .955 .786 .906 r
F P
0 g
s
.15027 .07619 m .14533 .05585 L .15215 .05947 L closepath p .955 .786 .906 r
F P
0 g
s
.19222 .05589 m .1931 .05669 L .19518 .06258 L closepath p .955 .786 .906 r
F P
0 g
s
.19518 .06258 m .19564 .06406 L .19222 .05589 L closepath p .955 .786 .906 r
F P
0 g
s
.62509 .05669 m .62724 .06258 L .62423 .05589 L closepath p .955 .786 .906 r
F P
0 g
s
.09112 .00639 m .09722 .01904 L .09315 .00538 L closepath p .902 1 1 r
F P
0 g
s
.09722 .01904 m .09112 .00639 L .09561 .02183 L closepath p 1 1 .84 r
F P
0 g
s
.10362 .02239 m .09761 .00979 L .10116 .01916 L closepath p .045 .045 .59 r
F P
0 g
s
.09512 .00652 m .10116 .01916 L .09761 .00979 L closepath p .045 .045 .59 r
F P
0 g
s
.55202 .05624 m .55293 .05636 L .54654 .04416 L closepath p .886 .499 .072 r
F P
0 g
s
.55293 .05636 m .55202 .05624 L .56142 .07252 L closepath p .861 .449 .069 r
F P
0 g
s
.56142 .07252 m .55293 .05636 L .56821 .07771 L closepath p .955 .786 .906 r
F P
0 g
s
.55779 .05825 m .56821 .07771 L .55293 .05636 L closepath p .955 .786 .906 r
F P
0 g
s
.80076 .0564 m .80365 .06964 L .79834 .06661 L closepath p .955 .786 .906 r
F P
0 g
s
.80365 .06964 m .80076 .0564 L .80794 .06415 L closepath p .955 .786 .906 r
F P
0 g
s
.79834 .06661 m .79725 .05646 L .80076 .0564 L closepath p .955 .786 .906 r
F P
0 g
s
.5367 .00668 m .53615 .01932 L .53858 .00576 L closepath p .831 1 1 r
F P
0 g
s
.53615 .01932 m .5367 .00668 L .53474 .02181 L closepath p .918 1 .648 r
F P
0 g
s
.79564 .05448 m .79725 .05646 L .79727 .04038 L closepath p 1 .981 .595 r
F P
0 g
s
.79725 .05646 m .79564 .05448 L .79834 .06661 L closepath p 1 .823 .365 r
F P
0 g
s
.19479 .05688 m .19406 .05277 L .1931 .05669 L closepath p .787 .967 1 r
F P
0 g
s
.19343 .05267 m .1931 .05669 L .19406 .05277 L closepath p .771 .95 1 r
F P
0 g
s
.89001 .05688 m .8909 .05277 L .88834 .05669 L closepath p .797 .965 1 r
F P
0 g
s
.89028 .05266 m .88834 .05669 L .8909 .05277 L closepath p .779 .939 1 r
F P
0 g
s
.62674 .05688 m .62704 .05278 L .62509 .05669 L closepath p .792 .965 1 r
F P
0 g
s
.62648 .05269 m .62509 .05669 L .62704 .05278 L closepath p .783 .953 1 r
F P
0 g
s
.1931 .05669 m .19649 .05708 L .19751 .06105 L closepath p .955 .786 .906 r
F P
0 g
s
.19751 .06105 m .19518 .06258 L .1931 .05669 L closepath p .955 .786 .906 r
F P
0 g
s
.62509 .05669 m .62838 .05708 L .62724 .06258 L closepath p .955 .786 .906 r
F P
0 g
s
.89167 .05708 m .90168 .09211 L .88834 .05669 L closepath p .955 .786 .906 r
F P
0 g
s
.41439 .01963 m .4131 .00699 L .41231 .02051 L closepath p .83 1 1 r
F P
0 g
s
.40815 .0212 m .40676 .00858 L .40549 .01963 L closepath p .227 .618 1 r
F P
0 g
s
.40406 .00699 m .40549 .01963 L .40676 .00858 L closepath p .227 .618 1 r
F P
0 g
s
.36795 .01963 m .36595 .00699 L .36583 .02049 L closepath p .829 1 1 r
F P
0 g
s
.35557 .02119 m .35338 .00857 L .35287 .01963 L closepath p .217 .616 1 r
F P
0 g
s
.35064 .00699 m .35287 .01963 L .35338 .00857 L closepath p .217 .616 1 r
F P
0 g
s
.31532 .01963 m .31253 .00699 L .31324 .02051 L closepath p .843 1 1 r
F P
0 g
s
.29662 .0212 m .29354 .00858 L .29397 .01963 L closepath p .185 .593 1 r
F P
0 g
s
.29085 .00699 m .29397 .01963 L .29354 .00858 L closepath p .185 .593 1 r
F P
0 g
s
.25618 .01963 m .25249 .00699 L .25409 .02051 L closepath p .85 1 1 r
F P
0 g
s
.25146 .0211 m .24769 .00848 L .2486 .01963 L closepath p .229 .641 1 r
F P
0 g
s
.24479 .00699 m .2486 .01963 L .24769 .00848 L closepath p .229 .641 1 r
F P
0 g
s
.21269 .01963 m .20834 .00699 L .21193 .02145 L closepath p 1 1 .694 r
F P
0 g
s
.19544 .02116 m .19083 .00854 L .19262 .01963 L closepath p .187 .609 1 r
F P
0 g
s
.18797 .00699 m .19262 .01963 L .19083 .00854 L closepath p .187 .609 1 r
F P
0 g
s
.127 .01963 m .12135 .00699 L .12452 .02065 L closepath p .861 1 1 r
F P
0 g
s
.07335 .02129 m .06688 .00868 L .07015 .01963 L closepath p .161 .6 1 r
F P
0 g
s
.06364 .00699 m .07015 .01963 L .06688 .00868 L closepath p .161 .6 1 r
F P
0 g
s
.03971 .01963 m .03274 .00699 L .03724 .02061 L closepath p .865 1 1 r
F P
0 g
s
.9467 .0211 m .95347 .00848 L .94377 .01964 L closepath p .426 .745 1 r
F P
0 g
s
.95049 .007 m .94377 .01964 L .95347 .00848 L closepath p .426 .745 1 r
F P
0 g
s
.91404 .007 m .90786 .01964 L .95049 .007 L closepath p .623 .977 1 r
F P
0 g
s
.90786 .01964 m .91404 .007 L .90719 .02146 L closepath p .022 .022 .022 r
F P
0 g
s
.89059 .02113 m .89651 .00851 L .88755 .01964 L closepath p .416 .743 1 r
F P
0 g
s
.89342 .007 m .88755 .01964 L .89651 .00851 L closepath p .416 .743 1 r
F P
0 g
s
.79218 .02119 m .79661 .00858 L .78943 .01964 L closepath p .361 .694 1 r
F P
0 g
s
.79382 .007 m .78943 .01964 L .79661 .00858 L closepath p .361 .694 1 r
F P
0 g
s
.75569 .007 m .75188 .01964 L .79382 .007 L closepath p .623 .977 1 r
F P
0 g
s
.75188 .01964 m .75569 .007 L .74982 .02052 L closepath p .775 1 1 r
F P
0 g
s
.72942 .02123 m .7329 .00861 L .72681 .01964 L closepath p .323 .663 1 r
F P
0 g
s
.73025 .007 m .72681 .01964 L .7329 .00861 L closepath p .323 .663 1 r
F P
0 g
s
.69196 .007 m .6891 .01964 L .73025 .007 L closepath p .623 .977 1 r
F P
0 g
s
.6891 .01964 m .69196 .007 L .687 .02049 L closepath p .781 1 1 r
F P
0 g
s
.68334 .02111 m .68612 .00849 L .68048 .01964 L closepath p .355 .707 1 r
F P
0 g
s
.68321 .007 m .68048 .01964 L .68612 .00849 L closepath p .355 .707 1 r
F P
0 g
s
.64707 .007 m .64487 .01964 L .68321 .007 L closepath p .623 .977 1 r
F P
0 g
s
.64487 .01964 m .64707 .007 L .64355 .02225 L closepath p .843 .946 .528 r
F P
0 g
s
.62743 .02113 m .62936 .00851 L .62442 .01964 L closepath p .35 .71 1 r
F P
0 g
s
.62631 .007 m .62442 .01964 L .62936 .00851 L closepath p .35 .71 1 r
F P
0 g
s
.56228 .01964 m .56322 .007 L .55995 .02061 L closepath p .805 1 1 r
F P
0 g
s
.51564 .02125 m .51587 .00864 L .5127 .01964 L closepath p .287 .662 1 r
F P
0 g
s
.5129 .007 m .5127 .01964 L .51587 .00864 L closepath p .287 .662 1 r
F P
0 g
s
.06436 .05686 m .06477 .0584 L .0675 .07879 L closepath p .955 .786 .906 r
F P
0 g
s
.068 .07024 m .0692 .06762 L .06436 .05686 L closepath p .147 .147 .685 r
F P
0 g
s
.06557 .0543 m .06436 .05686 L .0692 .06762 L closepath p .147 .147 .685 r
F P
0 g
s
.19649 .05708 m .19578 .05282 L .19479 .05688 L closepath p .796 .966 1 r
F P
0 g
s
.19406 .05277 m .19479 .05688 L .19578 .05282 L closepath p .826 .996 1 r
F P
0 g
s
.89167 .05708 m .89259 .05282 L .89001 .05688 L closepath p .806 .964 1 r
F P
0 g
s
.8909 .05277 m .89001 .05688 L .89259 .05282 L closepath p .828 .995 1 r
F P
0 g
s
.62838 .05708 m .62868 .05282 L .62674 .05688 L closepath p .801 .964 1 r
F P
0 g
s
.62704 .05278 m .62674 .05688 L .62868 .05282 L closepath p .828 .996 1 r
F P
0 g
s
.33609 .05701 m .33935 .05829 L .33532 .0738 L closepath p .955 .786 .906 r
F P
0 g
s
.33532 .0738 m .33399 .06411 L .33609 .05701 L closepath p .955 .786 .906 r
F P
0 g
s
.21843 .05282 m .22092 .05708 L .21978 .0479 L closepath p 1 1 .851 r
F P
0 g
s
.22092 .05708 m .21843 .05282 L .19649 .05708 L closepath p .845 1 1 r
F P
0 g
s
.19578 .05282 m .19649 .05708 L .21843 .05282 L closepath p .845 1 1 r
F P
0 g
s
.91617 .05418 m .9163 .0479 L .91522 .05282 L p 1 1 .648 r
F P
0 g
s
.92262 .03566 m .9163 .0479 L .92405 .02936 L closepath p .054 .054 .054 r
F P
0 g
s
.9163 .0479 m .92262 .03566 L .91522 .05282 L closepath p .055 .055 .055 r
F P
0 g
s
.92152 .04065 m .91522 .05282 L .92064 .04026 L p .055 .055 .055 r
F P
0 g
s
.92064 .04026 m .92262 .03566 L .92152 .04065 L p .055 .055 .055 r
F P
0 g
s
.92152 .04065 m .92455 .0409 L .92203 .04087 L p .088 .023 .023 r
F P
0 g
s
.91522 .05282 m .92152 .04065 L .91009 .04546 L p .607 .971 1 r
F P
0 g
s
.92405 .02936 m .91771 .04169 L .92524 .0224 L closepath p .064 .064 .117 r
F P
0 g
s
.89259 .05282 m .89855 .04065 L .8909 .05277 L closepath p .597 .961 1 r
F P
0 g
s
.89684 .04061 m .8909 .05277 L .89855 .04065 L closepath p .597 .961 1 r
F P
0 g
s
.89855 .04065 m .89962 .0409 L .89689 .04061 L p .096 .021 .021 r
F P
0 g
s
.89962 .0409 m .89793 .0407 L .89686 .04061 L p .003 .517 .713 r
F P
0 g
s
.89689 .04061 m .89684 .04061 L .89855 .04065 L p .096 .021 .021 r
F P
0 g
s
.89686 .04061 m .89684 .04061 L .89689 .04061 L p .003 .517 .713 r
F P
0 g
s
.89689 .04061 m .89962 .0409 L p .89686 .04061 L .003 .517 .713 r
F P
0 g
s
.89855 .04065 m .89848 .0408 L p .92035 .04114 L .607 .971 1 r
F P
0 g
s
.89848 .0408 m .89259 .05282 L .91624 .04287 L p .607 .971 1 r
F P
0 g
s
.91624 .04287 m .92035 .04114 L p .89848 .0408 L .607 .971 1 r
F P
0 g
s
.92035 .04114 m .92152 .04065 L .89855 .04065 L p .607 .971 1 r
F P
0 g
s
.92203 .04087 m .89855 .04065 L .92152 .04065 L p .088 .023 .023 r
F P
0 g
s
.91009 .04546 m .89259 .05282 L .91522 .05282 L p .607 .971 1 r
F P
0 g
s
.91522 .05282 m .9161 .05708 L .91617 .05418 L p 1 1 .648 r
F P
0 g
s
.9161 .05708 m .91522 .05282 L .89167 .05708 L closepath p .845 1 1 r
F P
0 g
s
.89259 .05282 m .89167 .05708 L .91522 .05282 L closepath p .845 1 1 r
F P
0 g
s
.65131 .05282 m .65281 .05708 L .65249 .0479 L closepath p 1 1 .752 r
F P
0 g
s
.65281 .05708 m .65131 .05282 L .62838 .05708 L closepath p .845 1 1 r
F P
0 g
s
.62868 .05282 m .62838 .05708 L .65131 .05282 L closepath p .845 1 1 r
F P
0 g
s
.19649 .05708 m .2026 .05708 L .2087 .06109 L closepath p .955 .786 .906 r
F P
0 g
s
.2087 .05708 m .2087 .06109 L .2026 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.21989 .06113 m .2087 .06109 L .22092 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.2087 .05708 m .22092 .05708 L .2087 .06109 L closepath p .955 .786 .906 r
F P
0 g
s
.2087 .06109 m .19751 .06105 L .19649 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.21201 .09872 m .21596 .0799 L .22092 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.21989 .06113 m .22092 .05708 L .21596 .0799 L closepath p .955 .786 .906 r
F P
0 g
s
.62838 .05708 m .64059 .05708 L .62952 .06105 L closepath p .955 .786 .906 r
F P
0 g
s
.64067 .06109 m .62952 .06105 L .64059 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.65183 .06113 m .64067 .06109 L .65281 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.64059 .05708 m .65281 .05708 L .64067 .06109 L closepath p .955 .786 .906 r
F P
0 g
s
.62952 .06105 m .62724 .06258 L .62838 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.65281 .05708 m .64525 .09872 L .65183 .06113 L closepath p .955 .786 .906 r
F P
0 g
s
.89167 .05708 m .90389 .05708 L .90168 .09211 L closepath p .955 .786 .906 r
F P
0 g
s
.9161 .05708 m .90168 .09211 L .90389 .05708 L closepath p .955 .786 .906 r
F P
0 g
s
.5444 .02549 m .54507 .01293 L .54118 .02044 L closepath p .107 .176 .74 r
F P
0 g
s
.54181 .00781 m .54118 .02044 L .54507 .01293 L closepath p .107 .176 .74 r
F P
0 g
s
.3638 .00786 m .36583 .02049 L .36595 .00699 L closepath p .829 1 1 r
F P
0 g
s
.36583 .02049 m .3638 .00786 L .3647 .02251 L closepath p .997 1 .726 r
F P
0 g
s
.68983 .00786 m .687 .02049 L .69196 .007 L closepath p .781 1 1 r
F P
0 g
s
.687 .02049 m .68983 .00786 L .68592 .02251 L closepath p .821 .938 .523 r
F P
0 g
s
.31041 .00788 m .31324 .02051 L .31253 .00699 L closepath p .843 1 1 r
F P
0 g
s
.31324 .02051 m .31041 .00788 L .31215 .02251 L closepath p 1 1 .742 r
F P
0 g
s
.25036 .00788 m .25409 .02051 L .25249 .00699 L closepath p .85 1 1 r
F P
0 g
s
.25409 .02051 m .25036 .00788 L .25299 .02251 L closepath p 1 1 .769 r
F P
0 g
s
.41098 .00788 m .41231 .02051 L .4131 .00699 L closepath p .83 1 1 r
F P
0 g
s
.41231 .02051 m .41098 .00788 L .41124 .02251 L closepath p .976 1 .691 r
F P
0 g
s
.7536 .00789 m .74982 .02052 L .75569 .007 L closepath p .775 1 1 r
F P
0 g
s
.74982 .02052 m .7536 .00789 L .74879 .02251 L closepath p .771 .892 .46 r
F P
0 g
s
.03023 .00798 m .03724 .02061 L .03274 .00699 L closepath p .865 1 1 r
F P
0 g
s
.03724 .02061 m .03023 .00798 L .03599 .02289 L closepath p 1 1 .84 r
F P
0 g
s
.56086 .00798 m .55995 .02061 L .56322 .007 L closepath p .805 1 1 r
F P
0 g
s
.55995 .02061 m .56086 .00798 L .55884 .02282 L closepath p .895 .983 .58 r
F P
0 g
s
.11883 .00802 m .12452 .02065 L .12135 .00699 L closepath p .861 1 1 r
F P
0 g
s
.12452 .02065 m .11883 .00802 L .1233 .02296 L closepath p 1 1 .802 r
F P
0 g
s
.81783 .00825 m .81309 .02087 L .82454 .0061 L closepath p .737 1 1 r
F P
0 g
s
.81309 .02087 m .81783 .00825 L .80739 .02436 L closepath p .794 1 .982 r
F P
0 g
s
.85779 .02317 m .86321 .01058 L .85194 .02093 L closepath p .455 .794 1 r
F P
0 g
s
.85728 .0083 m .85194 .02093 L .86321 .01058 L closepath p .455 .794 1 r
F P
0 g
s
.25252 .02426 m .24877 .01169 L .25146 .0211 L closepath p .007 .007 .252 r
F P
0 g
s
.24769 .00848 m .25146 .0211 L .24877 .01169 L closepath p .007 .007 .252 r
F P
0 g
s
.94792 .02427 m .95471 .01169 L .9467 .0211 L closepath p .144 .179 .735 r
F P
0 g
s
.95347 .00848 m .9467 .0211 L .95471 .01169 L closepath p .144 .179 .735 r
F P
0 g
s
.68447 .02427 m .68726 .01169 L .68334 .02111 L closepath p .101 .101 .588 r
F P
0 g
s
.68612 .00849 m .68334 .02111 L .68726 .01169 L closepath p .101 .101 .588 r
F P
0 g
s
.89185 .02442 m .89779 .01185 L .89059 .02113 L closepath p .137 .147 .711 r
F P
0 g
s
.89651 .00851 m .89059 .02113 L .89779 .01185 L closepath p .137 .147 .711 r
F P
0 g
s
.62863 .02442 m .63058 .01185 L .62743 .02113 L closepath p .092 .092 .562 r
F P
0 g
s
.62936 .00851 m .62743 .02113 L .63058 .01185 L closepath p .092 .092 .562 r
F P
0 g
s
.56821 .07771 m .57713 .09421 L .56142 .07252 L closepath p .955 .786 .906 r
F P
0 g
s
.56817 .07771 m .57035 .07458 L .55778 .05825 L closepath p .392 .492 .939 r
F P
0 g
s
.56001 .05522 m .55778 .05825 L .57035 .07458 L closepath p .393 .492 .939 r
F P
0 g
s
.19645 .02442 m .19185 .01184 L .19544 .02116 L closepath p .717 .416 .011 r
F P
0 g
s
.19083 .00854 m .19544 .02116 L .19185 .01184 L closepath p .717 .416 .011 r
F P
0 g
s
.35651 .02441 m .35433 .01184 L .35557 .02119 L closepath p .023 .023 .28 r
F P
0 g
s
.35338 .00857 m .35557 .02119 L .35433 .01184 L closepath p .023 .023 .28 r
F P
0 g
s
.33683 .0848 m .33532 .0738 L .33935 .05829 L closepath p .955 .786 .906 r
F P
0 g
s
.79322 .02442 m .79767 .01185 L .79218 .02119 L closepath p .116 .116 .615 r
F P
0 g
s
.79661 .00858 m .79218 .02119 L .79767 .01185 L closepath p .116 .116 .615 r
F P
0 g
s
.29751 .02441 m .29444 .01184 L .29662 .0212 L closepath p .006 .006 .213 r
F P
0 g
s
.29354 .00858 m .29662 .0212 L .29444 .01184 L closepath p .006 .006 .213 r
F P
0 g
s
.40907 .02442 m .40769 .01185 L .40815 .0212 L closepath p .034 .034 .315 r
F P
0 g
s
.40676 .00858 m .40815 .0212 L .40769 .01185 L closepath p .034 .034 .315 r
F P
0 g
s
.73034 .02442 m .73383 .01185 L .72942 .02123 L closepath p .101 .101 .547 r
F P
0 g
s
.7329 .00861 m .72942 .02123 L .73383 .01185 L closepath p .101 .101 .547 r
F P
0 g
s
.19649 .074 m .18923 .05834 L .19372 .06796 L closepath p .955 .786 .906 r
F P
0 g
s
.1863 .05731 m .18923 .05834 L .18019 .04404 L closepath p 1 .72 .368 r
F P
0 g
s
.18923 .05834 m .1863 .05731 L .19649 .074 L closepath p 1 .717 .365 r
F P
0 g
s
.6289 .074 m .62116 .05834 L .62595 .06796 L closepath p .955 .786 .906 r
F P
0 g
s
.88324 .05731 m .88439 .05834 L .87652 .04403 L closepath p .815 .414 .053 r
F P
0 g
s
.88439 .05834 m .88324 .05731 L .89239 .074 L closepath p .813 .41 .053 r
F P
0 g
s
.61933 .05732 m .62116 .05834 L .61282 .04404 L closepath p .966 .563 .104 r
F P
0 g
s
.62116 .05834 m .61933 .05732 L .6289 .074 L closepath p .964 .56 .103 r
F P
0 g
s
.51672 .02473 m .51697 .01216 L .51564 .02125 L closepath p .062 .062 .429 r
F P
0 g
s
.51587 .00864 m .51564 .02125 L .51697 .01216 L closepath p .062 .062 .429 r
F P
0 g
s
.07447 .02502 m .06802 .01246 L .07335 .02129 L closepath p .821 .508 .04 r
F P
0 g
s
.06688 .00868 m .07335 .02129 L .06802 .01246 L closepath p .821 .508 .04 r
F P
0 g
s
.07176 .09362 m .0675 .07879 L .06477 .0584 L closepath p .955 .786 .906 r
F P
0 g
s
.06477 .0584 m .06517 .05994 L .07176 .09362 L closepath p .955 .786 .906 r
F P
0 g
s
.20757 .00884 m .21193 .02145 L .20834 .00699 L closepath p 1 1 .694 r
F P
0 g
s
.21193 .02145 m .20757 .00884 L .21157 .02249 L closepath p 1 1 .638 r
F P
0 g
s
.91336 .00884 m .90719 .02146 L .91404 .007 L closepath p .022 .022 .022 r
F P
0 g
s
.90719 .02146 m .91336 .00884 L .90687 .0225 L closepath p .036 .036 .036 r
F P
0 g
s
.54065 .07009 m .54308 .0685 L .5427 .05883 L closepath p .157 .157 .474 r
F P
0 g
s
.54488 .05759 m .5427 .05883 L .54308 .0685 L closepath p .15 .15 .455 r
F P
0 g
s
.5427 .05883 m .54065 .07009 L .52725 .11504 L closepath p .955 .786 .906 r
F P
0 g
s
.52725 .11504 m .52442 .11487 L .5427 .05883 L closepath p .955 .786 .906 r
F P
0 g
s
.34629 .06785 m .34793 .06361 L .34485 .05886 L closepath p .167 .169 .692 r
F P
0 g
s
.34677 .05628 m .34485 .05886 L .34793 .06361 L closepath p .168 .169 .692 r
F P
0 g
s
.53526 .0092 m .53474 .02181 L .5367 .00668 L closepath p .918 1 .648 r
F P
0 g
s
.53474 .02181 m .53526 .0092 L .53293 .03026 L closepath p .815 .786 .269 r
F P
0 g
s
.08948 .00922 m .09561 .02183 L .09112 .00639 L closepath p 1 1 .84 r
F P
0 g
s
.09561 .02183 m .08948 .00922 L .09431 .02608 L closepath p 1 1 .655 r
F P
0 g
s
.08827 .05683 m .09159 .0592 L .09057 .04579 L closepath p 1 1 .834 r
F P
0 g
s
.09159 .0592 m .08827 .05683 L .08867 .07285 L closepath p 1 1 .838 r
F P
0 g
s
.64572 .00965 m .64355 .02225 L .64707 .007 L closepath p .843 .946 .528 r
F P
0 g
s
.64355 .02225 m .64572 .00965 L .6448 .02595 L closepath p .092 .092 .545 r
F P
0 g
s
.10661 .02751 m .10064 .01498 L .10362 .02239 L closepath p .019 .019 .424 r
F P
0 g
s
.09761 .00979 m .10362 .02239 L .10064 .01498 L closepath p .019 .019 .424 r
F P
0 g
s
.20719 .00989 m .21157 .02249 L .20757 .00884 L closepath p 1 1 .638 r
F P
0 g
s
.21157 .02249 m .20719 .00989 L .2128 .026 L closepath p .001 .001 .239 r
F P
0 g
s
.91304 .0099 m .90687 .0225 L .91336 .00884 L closepath p .036 .036 .036 r
F P
0 g
s
.90687 .0225 m .91304 .0099 L .90825 .02599 L closepath p .14 .166 .727 r
F P
0 g
s
.36265 .00991 m .3647 .02251 L .3638 .00786 L closepath p .997 1 .726 r
F P
0 g
s
.3647 .02251 m .36265 .00991 L .36575 .02575 L closepath p .031 .031 .328 r
F P
0 g
s
.3093 .00991 m .31215 .02251 L .31041 .00788 L closepath p 1 1 .742 r
F P
0 g
s
.31215 .02251 m .3093 .00991 L .31276 .02484 L closepath p .007 .007 .201 r
F P
0 g
s
.24925 .00991 m .25299 .02251 L .25036 .00788 L closepath p 1 1 .769 r
F P
0 g
s
.25299 .02251 m .24925 .00991 L .25455 .02621 L closepath p .023 .023 .358 r
F P
0 g
s
.40989 .00991 m .41124 .02251 L .41098 .00788 L closepath p .976 1 .691 r
F P
0 g
s
.41124 .02251 m .40989 .00991 L .41273 .0262 L closepath p .054 .054 .457 r
F P
0 g
s
.75256 .00991 m .74879 .02251 L .7536 .00789 L closepath p .771 .892 .46 r
F P
0 g
s
.74879 .02251 m .75256 .00991 L .75115 .02658 L closepath p .133 .236 .789 r
F P
0 g
s
.68874 .00991 m .68592 .02251 L .68983 .00786 L closepath p .821 .938 .523 r
F P
0 g
s
.68592 .02251 m .68874 .00991 L .68718 .02586 L closepath p .104 .104 .605 r
F P
0 g
s
.55973 .01022 m .55884 .02282 L .56086 .00798 L closepath p .895 .983 .58 r
F P
0 g
s
.55884 .02282 m .55973 .01022 L .55952 .02501 L closepath p .071 .071 .46 r
F P
0 g
s
.06517 .05994 m .06605 .06319 L .0748 .10468 L closepath p .955 .786 .906 r
F P
0 g
s
.0748 .10468 m .07176 .09362 L .06517 .05994 L closepath p .955 .786 .906 r
F P
0 g
s
.02896 .0103 m .03599 .02289 L .03023 .00798 L closepath p 1 1 .84 r
F P
0 g
s
.03599 .02289 m .02896 .0103 L .03719 .02596 L closepath p .771 .422 .031 r
F P
0 g
s
.15799 .06006 m .15874 .06652 L .16093 .08596 L closepath p .955 .786 .906 r
F P
0 g
s
.15948 .073 m .16093 .08596 L .15874 .06652 L closepath p .955 .786 .906 r
F P
0 g
s
.15948 .073 m .15874 .06652 L .16093 .08596 L closepath p .955 .786 .906 r
F P
0 g
s
.16093 .08596 m .16243 .08377 L .15799 .06006 L closepath p .14 .14 .576 r
F P
0 g
s
.15948 .05778 m .15799 .06006 L .16243 .08377 L closepath p .14 .14 .576 r
F P
0 g
s
.1548 .10632 m .15799 .06006 L .16093 .08596 L closepath p .955 .786 .906 r
F P
0 g
s
.1548 .10632 m .15569 .07012 L .15799 .06006 L closepath p .955 .786 .906 r
F P
0 g
s
.11759 .01037 m .1233 .02296 L .11883 .00802 L closepath p 1 1 .802 r
F P
0 g
s
.1233 .02296 m .11759 .01037 L .12444 .02627 L closepath p .743 .419 .02 r
F P
0 g
s
.47154 .06771 m .47381 .06553 L .47076 .06014 L closepath p .174 .178 .685 r
F P
0 g
s
.47326 .05847 m .47076 .06014 L .47381 .06553 L closepath p .191 .192 .688 r
F P
0 g
s
.47076 .06014 m .47154 .06771 L .46629 .0612 L closepath p .955 .786 .906 r
F P
0 g
s
.59262 .07658 m .59286 .07829 L .59026 .06043 L closepath p .955 .786 .906 r
F P
0 g
s
.59026 .06043 m .59096 .06524 L .59262 .07658 L closepath p .955 .786 .906 r
F P
0 g
s
.20952 .10197 m .20493 .09211 L .19409 .06083 L closepath p .955 .786 .906 r
F P
0 g
s
.19409 .06083 m .19702 .06742 L .20952 .10197 L closepath p .955 .786 .906 r
F P
0 g
s
.11609 .06096 m .11403 .06083 L .1252 .07946 L closepath p 1 .657 .26 r
F P
0 g
s
.11609 .06096 m .1502 .12162 L .1252 .07946 L closepath p .955 .786 .906 r
F P
0 g
s
.20862 .09264 m .20765 .09201 L .20594 .08707 L closepath p .955 .786 .906 r
F P
0 g
s
.20619 .08511 m .20379 .07886 L p .20506 .08249 L .45 .52 .945 r
F P
0 g
s
.20317 .07792 m .2086 .09264 L p .20379 .07886 L .45 .52 .945 r
F P
0 g
s
.20545 .08254 m .20409 .0791 L p .20504 .08165 L .652 .652 .985 r
F P
0 g
s
.20409 .0791 m .20408 .07909 L .2077 .08879 L .652 .652 .985 r
F
.20368 .07858 m .20662 .08624 L p .20407 .07904 L F P
0 g
s
.21064 .09653 m .21151 .1034 L .20862 .09264 L closepath p .955 .786 .906 r
F P
0 g
s
.20662 .08624 m .21058 .09653 L p .2077 .08879 L .652 .652 .985 r
F P
0 g
s
.20407 .07904 m .20662 .08624 L .2077 .08879 L .652 .652 .985 r
F
.19518 .06258 m .19552 .06368 L p .20375 .07878 L .658 .603 .927 r
F P
0 g
s
.19552 .06368 m .19564 .06406 L .19712 .06676 L p .658 .603 .927 r
F P
0 g
s
.19712 .06676 m .20375 .07878 L p .19552 .06368 L .658 .603 .927 r
F P
0 g
s
.20375 .07878 m .20376 .07879 L .19518 .06258 L .658 .603 .927 r
F
.20409 .0791 m .20407 .07904 L p .20408 .07909 L .652 .652 .985 r
F P
0 g
s
.20377 .07881 m .20619 .08511 L p .20396 .07931 L .736 .398 .02 r
F P
0 g
s
.20662 .08624 m .20398 .07935 L p .20619 .08511 L .736 .398 .02 r
F P
0 g
s
.20398 .07935 m .2036 .07836 L p .20396 .07931 L .736 .398 .02 r
F P
0 g
s
.20619 .08511 m .20398 .07935 L .20396 .07931 L .736 .398 .02 r
F
.20416 .07922 m .20419 .07921 L .20409 .07902 L .921 .724 .859 r
F
.20407 .07904 m .20409 .07902 L .19558 .06232 L F
.20416 .07922 m .20409 .07902 L .20407 .07904 L F
.20056 .07255 m .20406 .07905 L p .19558 .06232 L F P
0 g
s
.20406 .07905 m .20407 .07904 L .19558 .06232 L .921 .724 .859 r
F
.20407 .07904 m .20391 .07906 L .20385 .07868 L .996 .845 .938 r
F
.19751 .06105 m .19558 .06232 L p .20415 .07914 L .921 .724 .859 r
F P
0 g
s
.20415 .07914 m .20415 .07915 L .20415 .07914 L .921 .724 .859 r
F
.20415 .07914 m .20415 .07914 L .19751 .06105 L F
.21105 .09677 m .20545 .08254 L .2077 .08879 L .652 .652 .985 r
F
.21154 .09794 m .21032 .09486 L p .21105 .09677 L F P
0 g
s
.21032 .09484 m .20545 .08254 L p .21031 .09489 L .652 .652 .985 r
F P
0 g
s
.20545 .08254 m .20504 .08165 L .2077 .08879 L .652 .652 .985 r
F
.20389 .0792 m .21045 .09527 L .20387 .07922 L .996 .845 .938 r
F
.20395 .07915 m .20834 .09004 L .20389 .0792 L F
.21151 .1034 m .21064 .09653 L .21635 .11543 L closepath p .955 .786 .906 r
F P
0 g
s
.21201 .09872 m .21635 .11543 L .21064 .09653 L closepath p .955 .786 .906 r
F P
0 g
s
.21058 .09653 m .21185 .09872 L .21154 .09794 L p .652 .652 .985 r
F P
0 g
s
.21154 .09794 m .2077 .08879 L .21058 .09653 L .652 .652 .985 r
F
.20834 .09004 m .21185 .09872 L .20389 .0792 L .996 .845 .938 r
F
.2013 .07239 m .20333 .07769 L p .19997 .06891 L .736 .398 .02 r
F P
0 g
s
.20039 .07003 m .19997 .06891 L .19678 .06061 L .736 .398 .02 r
F
.2013 .07239 m .19997 .06891 L .20039 .07003 L F
.20339 .07781 m .20056 .07255 L p .20337 .07777 L .006 .006 .358 r
F P
0 g
s
.19558 .06232 m .19518 .06258 L .20056 .07255 L p .921 .724 .859 r
F P
0 g
s
.20056 .07255 m .19518 .06258 L .20087 .07313 L p .006 .006 .358 r
F P
0 g
s
.20087 .07313 m .20337 .07777 L p .20056 .07255 L .006 .006 .358 r
F P
0 g
s
.20333 .07769 m .20334 .07769 L .19678 .06061 L .736 .398 .02 r
F
.21247 .09476 m .21431 .07887 L p .20416 .07922 L .957 .827 .957 r
F P
0 g
s
.20416 .07922 m .21431 .07887 L .20542 .07884 L .957 .827 .957 r
F
.21989 .06113 m .19751 .06105 L p .20542 .07884 L .947 .735 .845 r
F P
0 g
s
.21063 .09055 m .21431 .07887 L p .20542 .07884 L .947 .735 .845 r
F P
0 g
s
.20542 .07884 m .21431 .07887 L .21989 .06113 L .947 .735 .845 r
F
.21989 .06113 m .21632 .06142 L .21431 .07887 L p 1 1 .819 r
F P
0 g
s
.21431 .07887 m .21201 .09872 L .21989 .06113 L p 1 1 .819 r
F P
0 g
s
.21431 .07887 m .21989 .06113 L p .21431 .07887 L .947 .735 .845 r
F P
0 g
s
.20415 .07915 m .2042 .07921 L .19751 .06105 L .921 .724 .859 r
F
.2042 .07921 m .20447 .07953 L .20434 .07917 L F
.2042 .07921 m .20434 .07917 L .19751 .06105 L F
.20482 .08045 m .20676 .08344 L .19751 .06105 L F
.20784 .08605 m .20987 .08954 L .19751 .06105 L F
.2105 .09099 m .21063 .09055 L .19751 .06105 L F
.20485 .07756 m .20448 .07808 L .20509 .07809 L .947 .735 .845 r
F
.19751 .06105 m .20448 .07808 L .20485 .07756 L F
.19751 .06105 m .20354 .07807 L p .20448 .07808 L F P
0 g
s
.20354 .07807 m .20355 .07809 L .20509 .07809 L .947 .735 .845 r
F
.20404 .07901 m .20339 .07781 L p .20344 .07788 L .006 .006 .358 r
F P
0 g
s
.20407 .07904 m .20391 .07865 L p .20368 .07858 L .652 .652 .985 r
F P
0 g
s
.20354 .07809 m .20368 .07836 L .20371 .07839 L .947 .735 .845 r
F
.20404 .07901 m .20344 .07788 L .20371 .0784 L .006 .006 .358 r
F
.20354 .07809 m .20386 .07867 L .20509 .07809 L .947 .735 .845 r
F
.21431 .07887 m .21431 .07885 L p .20542 .07884 L .957 .827 .957 r
F P
0 g
s
.21431 .07885 m .21463 .07604 L p .20543 .07883 L .957 .827 .957 r
F P
0 g
s
.20542 .07884 m .21431 .07885 L .20543 .07883 L .957 .827 .957 r
F
.20509 .07809 m .19751 .06105 L p .20508 .07809 L .921 .724 .859 r
F P
0 g
s
.21989 .06113 m .21063 .09055 L p .21063 .09053 L 1 1 .819 r
F P
0 g
s
.21872 .06122 m .21988 .06113 L p .21063 .09053 L 1 1 .819 r
F P
0 g
s
.21988 .06113 m .21989 .06113 L p .21063 .09053 L 1 1 .819 r
F P
0 g
s
.21231 .09434 m .20509 .07809 L .20508 .07809 L .921 .724 .859 r
F
.21231 .09434 m .21242 .09454 L .20509 .07809 L p F P
0 g
s
.20391 .07865 m .20407 .07904 L p .2039 .07865 L .996 .845 .938 r
F P
0 g
s
.20391 .07865 m .20391 .07865 L p .2036 .07826 L .652 .652 .985 r
F P
0 g
s
.20406 .07905 m .20416 .07922 L p .20407 .07904 L .921 .724 .859 r
F P
0 g
s
.20395 .07883 m .20416 .07922 L .20442 .07883 L .957 .827 .957 r
F
.19882 .06612 m .20119 .07255 L p .2004 .07013 L .45 .52 .945 r
F P
0 g
s
.20391 .07906 m .20385 .07868 L .20368 .07836 L .658 .603 .927 r
F
.20354 .07836 m .20368 .07836 L .19518 .06258 L F
.20391 .07865 m .20368 .07836 L .20368 .07858 L .652 .652 .985 r
F
.20368 .07836 m .20354 .07807 L .20371 .0784 L .947 .735 .845 r
F
.20262 .07563 m .19678 .06061 L .20223 .0747 L .652 .652 .985 r
F
.20337 .07777 m .20344 .07788 L p .20339 .07781 L .006 .006 .358 r
F P
0 g
s
.20344 .07788 m .19518 .06258 L p .20385 .07868 L .658 .603 .927 r
F P
0 g
s
.20262 .07563 m .20223 .0747 L .2036 .07826 L .652 .652 .985 r
F
.20354 .07807 m .20396 .07862 L .20371 .0784 L .947 .735 .845 r
F
.20344 .07788 m .20369 .07836 L p .20371 .0784 L .006 .006 .358 r
F P
0 g
s
.20368 .07836 m .20368 .07836 L .20354 .07809 L .658 .603 .927 r
F
.20368 .07836 m .2036 .07826 L .20363 .07836 L .652 .652 .985 r
F
.20368 .07836 m .20363 .07836 L .20368 .07858 L F
.20376 .07879 m .20377 .07881 L p .2036 .07836 L .736 .398 .02 r
F P
0 g
s
.20391 .07906 m .20368 .07836 L .20354 .07836 L .658 .603 .927 r
F
.20354 .07809 m .20377 .07872 L p .20385 .07868 L .947 .735 .845 r
F P
0 g
s
.2036 .07826 m .20354 .07809 L p .20368 .07836 L .947 .735 .845 r
F P
0 g
s
.20368 .07836 m .20366 .07829 L .2036 .07826 L .652 .652 .985 r
F
.20366 .07829 m .20262 .07563 L .20352 .07804 L F
.20354 .07807 m .20354 .07809 L p .20355 .07809 L .947 .735 .845 r
F P
0 g
s
.20369 .07836 m .20344 .07788 L p .20354 .07809 L .658 .603 .927 r
F P
0 g
s
.20354 .07809 m .20354 .07807 L p .20368 .07836 L .947 .735 .845 r
F P
0 g
s
.20368 .07836 m .20369 .07836 L .20354 .07809 L .658 .603 .927 r
F
.20366 .07829 m .20352 .07804 L .2036 .07826 L .652 .652 .985 r
F
.20396 .07862 m .20391 .07865 L .20371 .0784 L .947 .735 .845 r
F
.20391 .07865 m .2038 .07836 L p .20368 .07836 L .652 .652 .985 r
F P
0 g
s
.20354 .07807 m .20319 .0771 L p .20357 .07807 L .947 .735 .845 r
F P
0 g
s
.20319 .0771 m .19751 .06105 L p .20357 .07807 L .947 .735 .845 r
F P
0 g
s
.19751 .06105 m .20509 .07809 L p .20357 .07807 L .947 .735 .845 r
F P
0 g
s
.20354 .07807 m .20509 .07809 L .20396 .07862 L .947 .735 .845 r
F
.2038 .07836 m .19678 .06061 L p .20368 .07836 L .652 .652 .985 r
F P
0 g
s
.2036 .07826 m .20269 .07591 L .20351 .07815 L .652 .652 .985 r
F
.19678 .06061 m .20274 .07614 L p .20269 .07591 L F P
0 g
s
.20274 .07614 m .20351 .07815 L p .20269 .07591 L .652 .652 .985 r
F P
0 g
s
.19564 .06406 m .19702 .06742 L .20313 .07792 L p .389 .454 .901 r
F P
0 g
s
.20376 .07879 m .20374 .07875 L .19518 .06258 L .658 .603 .927 r
F
.20352 .07836 m .20232 .07619 L p .20296 .07715 L .131 .185 .747 r
F P
0 g
s
.20232 .07619 m .19712 .06676 L p .2025 .07633 L .131 .185 .747 r
F P
0 g
s
.19712 .06676 m .19564 .06406 L .19749 .06737 L p .131 .185 .747 r
F P
0 g
s
.19749 .06737 m .2025 .07633 L p .19712 .06676 L .131 .185 .747 r
F P
0 g
s
.2025 .07633 m .20296 .07715 L p .20232 .07619 L .131 .185 .747 r
F P
0 g
s
.20296 .07715 m .19749 .06737 L p .20293 .07756 L .389 .454 .901 r
F P
0 g
s
.19749 .06737 m .19564 .06406 L p .19864 .06962 L .389 .454 .901 r
F P
0 g
s
.20293 .07756 m .19749 .06737 L .19864 .06962 L .389 .454 .901 r
F
.20298 .0774 m .20317 .07792 L p .20298 .07739 L .45 .52 .945 r
F P
0 g
s
.20296 .07715 m .20302 .07726 L p .20352 .07836 L .131 .185 .747 r
F P
0 g
s
.20353 .07831 m .20307 .07711 L .20311 .07745 L .45 .52 .945 r
F
.20302 .07726 m .20344 .078 L p .20349 .07824 L .131 .185 .747 r
F P
0 g
s
.20352 .07836 m .20302 .07726 L .20349 .07824 L .131 .185 .747 r
F
.20366 .07868 m .20353 .07831 L .20317 .07792 L .45 .52 .945 r
F
.20353 .07831 m .20311 .07745 L .20317 .07792 L F
.203 .07733 m .20302 .07727 L .20296 .07715 L .389 .454 .901 r
F
.20307 .07711 m .2004 .07013 L .20298 .07739 L .45 .52 .945 r
F
.20307 .07711 m .20298 .07739 L .20317 .07792 L F
.20305 .07756 m .20361 .07853 L .20302 .07727 L .389 .454 .901 r
F
.20305 .07754 m .20302 .07727 L .203 .07733 L F
.2035 .07856 m .20262 .07643 L .20274 .07652 L .141 .19 .747 r
F
.20305 .07755 m .20302 .07727 L .20305 .07754 L .389 .454 .901 r
F
.20376 .07879 m .2013 .07239 L p .20353 .07831 L .45 .52 .945 r
F P
0 g
s
.2036 .07836 m .19678 .06061 L p .20376 .07879 L .736 .398 .02 r
F P
0 g
s
.19678 .06061 m .20193 .074 L .19938 .06755 L .996 .845 .938 r
F
.2013 .07239 m .2004 .07013 L .20353 .07831 L .45 .52 .945 r
F
.2004 .07013 m .20119 .07255 L .20086 .07141 L F
.19938 .06755 m .20193 .074 L .20086 .07147 L .996 .845 .938 r
F
.20123 .07241 m .20359 .07815 L p .20075 .07088 L .957 .827 .957 r
F P
0 g
s
.1979 .06362 m .20075 .07088 L .19704 .06135 L .957 .827 .957 r
F
.2013 .07239 m .19924 .06703 L p .2004 .07013 L .45 .52 .945 r
F P
0 g
s
.19924 .06703 m .19871 .06568 L .2004 .07013 L .45 .52 .945 r
F
.1972 .06174 m .19882 .06612 L p .2004 .07013 L F P
0 g
s
.19674 .06066 m .19678 .06061 L p .20086 .07147 L .996 .845 .938 r
F P
0 g
s
.20021 .06978 m .1979 .06362 L .19868 .06569 L .141 .19 .747 r
F
.20274 .07652 m .19986 .06884 L .20021 .06978 L F
.19986 .06884 m .1979 .06362 L .20021 .06978 L F
.20178 .07364 m .20322 .0774 L .20195 .07433 L .996 .845 .938 r
F
.20086 .07141 m .20298 .0774 L .20298 .07739 L .45 .52 .945 r
F
.20332 .07817 m .20325 .07759 L .20345 .07835 L .163 .299 .827 r
F
.20317 .07792 m .20298 .0774 L p .20352 .07861 L .141 .19 .747 r
F P
0 g
s
.20352 .07863 m .20317 .07792 L .20352 .07861 L .141 .19 .747 r
F
.20351 .07859 m .20298 .0774 L .20345 .07844 L F
.20356 .07871 m .20352 .07861 L .20351 .07859 L F
.20352 .07861 m .20298 .0774 L .20351 .07859 L F
.20416 .07922 m .20482 .08045 L p .20419 .07921 L .921 .724 .859 r
F P
0 g
s
.20404 .07901 m .20371 .0784 L .20652 .08361 L .006 .006 .358 r
F
.20379 .07878 m .2038 .07881 L .20509 .07809 L .947 .735 .845 r
F
.21032 .09486 m .21032 .09484 L p .21105 .09677 L .652 .652 .985 r
F P
0 g
s
.21032 .09484 m .21031 .09489 L .21105 .09677 L .652 .652 .985 r
F
.20545 .08254 m .21185 .09872 L p .20395 .07915 L .996 .845 .938 r
F P
0 g
s
.20985 .09361 m .20394 .0791 L p .21008 .09422 L .838 .764 .988 r
F P
0 g
s
.20407 .07904 m .20545 .08254 L p .20395 .07915 L .996 .845 .938 r
F P
0 g
s
.20386 .07867 m .20389 .07873 L .20509 .07809 L .947 .735 .845 r
F
.2038 .07881 m .20383 .07883 L .20509 .07809 L F
.20416 .07922 m .20542 .07884 L .20442 .07883 L .957 .827 .957 r
F
.20371 .07839 m .20385 .07868 L .20386 .07867 L .947 .735 .845 r
F
.20368 .07836 m .20385 .07868 L .20371 .07839 L F
.20368 .07836 m .20371 .0784 L .20393 .07883 L F
.20416 .07922 m .20404 .07901 L p .20415 .07921 L .006 .006 .358 r
F P
0 g
s
.20407 .07904 m .20387 .07867 L .2039 .07865 L .996 .845 .938 r
F
.20407 .07904 m .20385 .07868 L .20387 .07867 L F
.20383 .07883 m .20386 .07883 L .20509 .07809 L .947 .735 .845 r
F
.20386 .07883 m .20528 .07884 L .20528 .07883 L F
.20386 .07883 m .20528 .07883 L .20509 .07809 L F
.20395 .07883 m .20525 .08205 L .21247 .09476 L .957 .827 .957 r
F
.20427 .07964 m .20625 .08331 L .20399 .07896 L .658 .603 .927 r
F
.20407 .07904 m .20395 .07915 L .20391 .07906 L .996 .845 .938 r
F
.20395 .07905 m .20441 .08017 L p .20394 .07883 L .957 .827 .957 r
F P
0 g
s
.20441 .08017 m .21201 .09872 L p .20525 .08205 L .957 .827 .957 r
F P
0 g
s
.20441 .08017 m .20525 .08205 L .20394 .07883 L .957 .827 .957 r
F
.20394 .07883 m .20525 .08205 L .20395 .07883 L F
.21193 .09872 m .21201 .09872 L .20441 .08017 L p F P
0 g
s
.20441 .08017 m .20395 .07905 L p .20436 .08005 L .957 .827 .957 r
F P
0 g
s
.20441 .08017 m .20436 .08005 L .21193 .09872 L .957 .827 .957 r
F
.20397 .07917 m .20486 .08073 L p .20395 .07905 L .658 .603 .927 r
F P
0 g
s
.20427 .07964 m .20399 .07896 L .20397 .07894 L .658 .603 .927 r
F
.20397 .07894 m .20399 .07896 L .20393 .07883 L F
.20395 .07905 m .20394 .07883 L .20393 .07883 L .957 .827 .957 r
F
.20395 .07905 m .20427 .07964 L .20393 .07883 L .658 .603 .927 r
F
.20397 .07917 m .20393 .07883 L .20385 .07869 L F
.20395 .07905 m .20386 .07883 L p .20395 .07906 L .957 .827 .957 r
F P
0 g
s
.20397 .07917 m .20397 .07917 L p .20395 .07905 L .658 .603 .927 r
F P
0 g
s
.20385 .07868 m .20391 .07906 L .20377 .07872 L .996 .845 .938 r
F
.20383 .07883 m .20394 .0791 L p .20386 .07883 L .957 .827 .957 r
F P
0 g
s
.20394 .0791 m .21193 .09872 L p .20435 .08003 L .957 .827 .957 r
F P
0 g
s
.20385 .07868 m .20393 .07883 L .20393 .07883 L .947 .735 .845 r
F
.20368 .07836 m .20385 .07868 L .20385 .07868 L F
.20377 .07872 m .20379 .07878 L p .20389 .07873 L F P
0 g
s
.20385 .07868 m .20393 .07883 L .20385 .07868 L .947 .735 .845 r
F
.20387 .07867 m .20385 .07868 L .20385 .07868 L .996 .845 .938 r
F
.20387 .07867 m .20385 .07868 L .20387 .07867 L F
.20415 .07915 m .20482 .08045 L .20447 .07953 L .921 .724 .859 r
F
.20482 .08045 m .20784 .08605 L p .20676 .08344 L F P
0 g
s
.20403 .07898 m .20652 .08361 L .20562 .08016 L .947 .735 .845 r
F
.20393 .07883 m .20562 .08016 L .20528 .07884 L F
.20399 .07895 m .20562 .08016 L .20393 .07883 L F
.20542 .07884 m .2067 .08394 L .21063 .09055 L F
.2067 .08394 m .2077 .08579 L .21063 .09055 L F
.21032 .09065 m .20416 .07922 L .20657 .0837 L .006 .006 .358 r
F
.20784 .08605 m .2105 .09099 L p .20987 .08954 L .921 .724 .859 r
F P
0 g
s
.21032 .09066 m .20784 .08606 L .21032 .09065 L .006 .006 .358 r
F
.21032 .09066 m .20416 .07922 L p .20784 .08606 L F P
0 g
s
.2105 .09099 m .21231 .09434 L p .21063 .09055 L .921 .724 .859 r
F P
0 g
s
.21144 .09272 m .2105 .09098 L .21145 .09274 L .006 .006 .358 r
F
.2105 .09098 m .20657 .0837 L .21049 .09097 L F
.2105 .09098 m .21049 .09097 L .21145 .09274 L F
.20416 .07922 m .20415 .07921 L .20652 .08361 L F
.21049 .09097 m .21063 .09055 L p .2077 .08579 L .947 .735 .845 r
F P
0 g
s
.20416 .07922 m .20652 .08361 L .20657 .0837 L .006 .006 .358 r
F
.20395 .07905 m .20395 .07906 L .20435 .08003 L .957 .827 .957 r
F
.20436 .08005 m .20435 .08003 L .21193 .09872 L F
.20395 .07905 m .20435 .08003 L .20436 .08005 L F
.20528 .07884 m .20663 .08401 L .20542 .07884 L .947 .735 .845 r
F
.21008 .09422 m .208 .08911 L .20801 .08913 L .838 .764 .988 r
F
.208 .08911 m .20394 .0791 L .20801 .08913 L F
.20397 .07917 m .20384 .07863 L .20391 .07906 L .658 .603 .927 r
F
.20351 .07815 m .20351 .07815 L p .2036 .07826 L .652 .652 .985 r
F P
0 g
s
.20351 .07815 m .20368 .07858 L p .2036 .07826 L .652 .652 .985 r
F P
0 g
s
.20333 .07769 m .20376 .07879 L p .20334 .07769 L .736 .398 .02 r
F P
0 g
s
.20395 .07915 m .20344 .078 L .20333 .07765 L .996 .845 .938 r
F
.20389 .07903 m .20391 .07905 L .20391 .07906 L .131 .185 .747 r
F
.20391 .07905 m .20393 .07908 L .20391 .07906 L F
.20391 .07906 m .20393 .07908 L .20394 .07911 L F
.20389 .07903 m .20394 .07911 L .20394 .07912 L F
.20389 .07903 m .20386 .07896 L .20396 .07914 L F
.20801 .08913 m .20394 .0791 L .20384 .07889 L .838 .764 .988 r
F
.20374 .07875 m .20381 .07882 L .20391 .07905 L .131 .185 .747 r
F
.20375 .07878 m .20391 .07907 L p .20378 .07881 L .658 .603 .927 r
F P
0 g
s
.20391 .07907 m .20397 .07917 L p .20391 .07906 L .658 .603 .927 r
F P
0 g
s
.20391 .07907 m .20391 .07906 L .20378 .07881 L .658 .603 .927 r
F
.20397 .07917 m .20391 .07907 L p .20394 .07912 L .131 .185 .747 r
F P
0 g
s
.20391 .07907 m .20376 .0788 L p .20389 .07903 L .131 .185 .747 r
F P
0 g
s
.20391 .07907 m .20389 .07903 L .20394 .07912 L .131 .185 .747 r
F
.20374 .07875 m .20367 .07856 L .20381 .07882 L F
.20367 .07856 m .20367 .07856 L .20379 .07878 L F
.20351 .07827 m .20344 .078 L .20348 .07811 L F
.20352 .07836 m .20349 .07824 L .20351 .07827 L F
.20349 .07824 m .20344 .078 L .20351 .07827 L F
.20376 .0788 m .20352 .07836 L p .20374 .07875 L F P
0 g
s
.20375 .07878 m .20353 .07831 L .20366 .07868 L .45 .52 .945 r
F
.20376 .07879 m .20353 .07831 L .20375 .07878 L F
.20344 .078 m .20296 .07715 L p .20361 .07853 L .389 .454 .901 r
F P
0 g
s
.20387 .07922 m .20365 .07867 L p .20391 .07907 L .996 .845 .938 r
F P
0 g
s
.20395 .07915 m .20393 .07917 L .20391 .07907 L .996 .845 .938 r
F
.20393 .07917 m .20387 .07922 L .20391 .07907 L F
.2038 .07881 m .21008 .09422 L .20381 .07882 L .838 .764 .988 r
F
.20377 .07881 m .20396 .07931 L .2036 .07836 L .736 .398 .02 r
F
.20376 .07879 m .20378 .07881 L .20374 .07875 L .658 .603 .927 r
F
.20376 .0788 m .20374 .07875 L .20386 .07896 L .131 .185 .747 r
F
.20376 .0788 m .20386 .07896 L .20389 .07903 L F
.20379 .07886 m .20376 .07879 L p .20366 .07868 L .45 .52 .945 r
F P
0 g
s
.20372 .07876 m .21008 .09422 L p .2038 .07881 L .838 .764 .988 r
F P
0 g
s
.20948 .09292 m .20387 .07922 L p .21045 .09527 L .996 .845 .938 r
F P
0 g
s
.21185 .09872 m .21193 .09872 L .20391 .07921 L p .957 .827 .957 r
F P
0 g
s
.21185 .09872 m .20948 .09292 L p .21045 .09527 L .996 .845 .938 r
F P
0 g
s
.20387 .07922 m .21185 .09872 L p .20391 .07921 L .957 .827 .957 r
F P
0 g
s
.21201 .09872 m .21214 .09755 L p .20525 .08205 L .957 .827 .957 r
F P
0 g
s
.21214 .09755 m .21247 .09476 L p .20525 .08205 L .957 .827 .957 r
F P
0 g
s
.20619 .08511 m .21058 .09653 L .20662 .08624 L p .736 .398 .02 r
F P
0 g
s
.2086 .09264 m .21058 .09653 L .20619 .08511 L p .45 .52 .945 r
F P
0 g
s
.20663 .08401 m .20937 .09454 L .2068 .08432 L .947 .735 .845 r
F
.20397 .07917 m .20452 .08013 L .20465 .08036 L .131 .185 .747 r
F
.20937 .09454 m .21046 .09107 L p .2068 .08432 L .947 .735 .845 r
F P
0 g
s
.20625 .08331 m .20397 .07917 L p .20465 .08036 L .131 .185 .747 r
F P
0 g
s
.20394 .07912 m .20437 .07986 L .20452 .08013 L .131 .185 .747 r
F
.20397 .07917 m .20394 .07912 L .20452 .08013 L F
.20386 .07883 m .20394 .0791 L .20435 .08003 L .957 .827 .957 r
F
.20394 .07911 m .20396 .07914 L .20397 .07916 L .131 .185 .747 r
F
.20393 .07908 m .20396 .07914 L .20394 .07911 L F
.20394 .07911 m .20437 .07986 L .20394 .07912 L F
.20397 .07916 m .20396 .07914 L .20437 .07986 L F
.20394 .0791 m .20383 .07883 L p .20384 .07889 L .838 .764 .988 r
F P
0 g
s
.20391 .07905 m .20381 .07882 L .20437 .07986 L .131 .185 .747 r
F
.20381 .07883 m .20402 .07944 L .20663 .08401 L .947 .735 .845 r
F
.20367 .07856 m .20379 .07878 L .20437 .07986 L .131 .185 .747 r
F
.2038 .07881 m .20663 .08401 L .20381 .07882 L .947 .735 .845 r
F
.20381 .07882 m .20525 .08147 L .20703 .08494 L F
.20619 .08511 m .20506 .08249 L .2086 .09264 L .45 .52 .945 r
F
.20402 .07944 m .20937 .09454 L .20673 .08439 L .947 .735 .845 r
F
.20604 .08313 m .20703 .08494 L .21035 .09143 L F
.20604 .08313 m .21035 .09143 L .21034 .09145 L F
.20663 .08401 m .2068 .08432 L .20542 .07884 L F
.20395 .07883 m .21054 .09037 L .21041 .09082 L F
.21054 .09037 m .21063 .09053 L .21041 .09082 L F
.21041 .09082 m .21063 .09053 L .21049 .09097 L F
.20657 .0837 m .20652 .08361 L .21042 .09083 L .006 .006 .358 r
F
.21063 .09055 m .21049 .09097 L p .21063 .09053 L 1 1 .819 r
F P
0 g
s
.20396 .07883 m .21063 .09053 L .20395 .07883 L .947 .735 .845 r
F
.20509 .07809 m .20528 .07884 L .20542 .07884 L F
.20509 .07809 m .20543 .07883 L p .20396 .07862 L F P
0 g
s
.20396 .07862 m .20543 .07883 L .20413 .07883 L .947 .735 .845 r
F
.21463 .07604 m .21632 .06142 L p .20442 .07883 L .957 .827 .957 r
F P
0 g
s
.20543 .07883 m .21463 .07604 L .20442 .07883 L .957 .827 .957 r
F
.20718 .08277 m .21063 .09053 L p .20396 .07883 L .947 .735 .845 r
F P
0 g
s
.21872 .06122 m .21063 .09053 L .21049 .09097 L 1 1 .819 r
F
.21872 .06122 m .21049 .09097 L .21041 .09082 L F
.21145 .09274 m .20657 .0837 L .2116 .09302 L .006 .006 .358 r
F
.2105 .09098 m .21032 .09066 L .21032 .09065 L F
.21063 .09053 m .21049 .09097 L .21049 .09097 L 1 1 .819 r
F
.21063 .09053 m .21242 .09454 L p .21049 .09097 L .947 .735 .845 r
F P
0 g
s
.21042 .09083 m .21042 .09083 L .2116 .09302 L .006 .006 .358 r
F
.20657 .0837 m .21042 .09083 L .21042 .09083 L F
.21144 .09272 m .2105 .09099 L p .2105 .09098 L F P
0 g
s
.2105 .09099 m .21032 .09066 L p .2105 .09098 L .006 .006 .358 r
F P
0 g
s
.2067 .08394 m .2068 .08432 L .2077 .08579 L .947 .735 .845 r
F
.20425 .07945 m .20393 .07883 L .21242 .09454 L F
.20377 .07872 m .20388 .07873 L .20385 .07867 L F
.20388 .07873 m .20389 .07873 L .20385 .07867 L F
.20385 .07867 m .20389 .07873 L .20386 .07867 L F
.20393 .07883 m .20371 .0784 L .20395 .07883 L F
.20395 .07883 m .20394 .07883 L .20395 .07883 L .957 .827 .957 r
F
.20394 .07883 m .20394 .07883 L .20395 .07883 L F
.20399 .07895 m .20652 .08361 L .20403 .07898 L .947 .735 .845 r
F
.20393 .07883 m .20395 .07883 L .21242 .09454 L F
.20392 .07878 m .20394 .07883 L .20395 .07883 L .957 .827 .957 r
F
.20393 .07883 m .20394 .07883 L .20392 .07878 L F
.20399 .07895 m .20663 .08401 L .20652 .08361 L .947 .735 .845 r
F
.20371 .0784 m .20369 .07836 L .20652 .08361 L .006 .006 .358 r
F
.20393 .07883 m .21029 .09059 L .20392 .07878 L .658 .603 .927 r
F
.2068 .08432 m .21049 .09097 L .2077 .08579 L .947 .735 .845 r
F
.21029 .09059 m .21029 .09059 L .20392 .07878 L .658 .603 .927 r
F
.20371 .0784 m .20396 .07883 L .20395 .07883 L .947 .735 .845 r
F
.20387 .07867 m .20385 .07868 L .20383 .07859 L .996 .845 .938 r
F
.20384 .07863 m .20369 .07836 L p .20384 .07867 L .658 .603 .927 r
F P
0 g
s
.20385 .07868 m .20384 .07863 L .20384 .07867 L .658 .603 .927 r
F
.20383 .07859 m .20385 .07868 L .20322 .0774 L .996 .845 .938 r
F
.20396 .07862 m .20413 .07883 L .20391 .07865 L .947 .735 .845 r
F
.19678 .06061 m .20391 .07865 L p .20193 .074 L .996 .845 .938 r
F P
0 g
s
.2039 .07865 m .20322 .0774 L .20178 .07364 L .996 .845 .938 r
F
.20387 .07867 m .20322 .0774 L .2039 .07865 L F
.20371 .0784 m .20372 .07841 L .20374 .07845 L .947 .735 .845 r
F
.20372 .07841 m .20406 .07883 L .20374 .07845 L F
.20374 .07845 m .20406 .07883 L .20396 .07883 L F
.20391 .07865 m .2039 .07865 L .20086 .07147 L .996 .845 .938 r
F
.20359 .07815 m .20395 .07883 L .21632 .06142 L .957 .827 .957 r
F
.20369 .07836 m .2116 .09302 L p .20652 .08361 L .006 .006 .358 r
F P
0 g
s
.21041 .09081 m .20392 .07878 L p .21041 .09081 L .658 .603 .927 r
F P
0 g
s
.20402 .07944 m .20673 .08439 L .20663 .08401 L .947 .735 .845 r
F
.20381 .07882 m .20703 .08494 L .20604 .08313 L F
.20625 .08331 m .20663 .08401 L .20399 .07895 L F
.20393 .07883 m .20393 .07883 L .20425 .07945 L F
.20384 .07889 m .20383 .07883 L .20381 .07882 L .838 .764 .988 r
F
.20382 .07883 m .20625 .08331 L .20383 .07883 L .947 .735 .845 r
F
.20625 .08331 m .20465 .08036 L .20465 .08036 L .131 .185 .747 r
F
.21046 .09107 m .21049 .09097 L p .2068 .08432 L .947 .735 .845 r
F P
0 g
s
.21049 .09097 m .21046 .09107 L p .21049 .09097 L 1 1 .819 r
F P
0 g
s
.21049 .09097 m .21046 .09107 L .21046 .09107 L 1 1 .819 r
F
.21242 .09454 m .21235 .09454 L p .20467 .08027 L .947 .735 .845 r
F P
0 g
s
.20828 .08704 m .20467 .08027 L .21235 .09454 L .947 .735 .845 r
F
.20467 .08027 m .20425 .07945 L .21242 .09454 L F
.20625 .08331 m .20626 .08332 L p .20393 .07883 L .658 .603 .927 r
F P
0 g
s
.20625 .08331 m .20465 .08036 L .20625 .08331 L .131 .185 .747 r
F
.20626 .08332 m .20393 .07884 L .20393 .07883 L .658 .603 .927 r
F
.21046 .09107 m .20937 .09454 L p .21046 .09107 L 1 1 .819 r
F P
0 g
s
.20703 .08494 m .20525 .08147 L .21226 .09454 L .947 .735 .845 r
F
.20386 .07883 m .20395 .07905 L p .20393 .07883 L .957 .827 .957 r
F P
0 g
s
.20383 .07883 m .20625 .08331 L .20399 .07895 L .947 .735 .845 r
F
.20386 .07883 m .20386 .07883 L p .20383 .07883 L .957 .827 .957 r
F P
0 g
s
.2039 .07883 m .2036 .07826 L .20391 .07883 L .947 .735 .845 r
F
.2036 .07826 m .20368 .07836 L .20385 .07868 L F
.2036 .07826 m .20385 .07868 L .20383 .07868 L F
.20385 .07868 m .20369 .07836 L .20368 .07836 L .658 .603 .927 r
F
.20385 .07868 m .20377 .07872 L .20322 .0774 L .996 .845 .938 r
F
.20386 .07883 m .20326 .07736 L p .20383 .07883 L .957 .827 .957 r
F P
0 g
s
.20326 .07736 m .20123 .07241 L p .20335 .07766 L .957 .827 .957 r
F P
0 g
s
.20326 .07736 m .20335 .07766 L .20383 .07883 L .957 .827 .957 r
F
.20352 .07836 m .20348 .07809 L .20374 .07875 L .131 .185 .747 r
F
.20381 .07882 m .20381 .07882 L .2032 .07747 L .838 .764 .988 r
F
.20348 .07809 m .20348 .07808 L .2035 .07814 L .131 .185 .747 r
F
.20348 .07809 m .2035 .07814 L .20374 .07875 L F
.20525 .08147 m .2039 .07883 L .20828 .08704 L .947 .735 .845 r
F
.2039 .07883 m .20391 .07883 L .20828 .08704 L F
.20525 .08147 m .20828 .08704 L .21226 .09454 L F
.20541 .0817 m .20647 .08354 L .20393 .07884 L .658 .603 .927 r
F
.20626 .08332 m .20647 .08354 L .20541 .0817 L F
.20626 .08332 m .21041 .09085 L p .20647 .08354 L F P
0 g
s
.21041 .09085 m .21041 .09081 L .20393 .07883 L .658 .603 .927 r
F
.20625 .08331 m .20465 .08036 L .20989 .08953 L .131 .185 .747 r
F
.20381 .07883 m .20401 .07942 L p .20381 .07883 L .947 .735 .845 r
F P
0 g
s
.20401 .07942 m .20937 .09454 L p .20402 .07944 L .947 .735 .845 r
F P
0 g
s
.20401 .07942 m .20402 .07944 L .20381 .07883 L .947 .735 .845 r
F
.2038 .07881 m .20381 .07883 L p .20381 .07883 L F P
0 g
s
.204 .07937 m .2038 .07881 L p .20597 .083 L .947 .735 .845 r
F P
0 g
s
.20423 .07981 m .20597 .083 L .21195 .09454 L .947 .735 .845 r
F
.204 .07937 m .20597 .083 L .20423 .07981 L F
.20362 .07865 m .20395 .07914 L .20306 .07759 L .389 .454 .901 r
F
.20423 .07981 m .20395 .07914 L .20362 .07865 L F
.2038 .07881 m .20381 .07882 L .20791 .08674 L .947 .735 .845 r
F
.20785 .08664 m .20791 .08674 L .21195 .09454 L F
.2038 .07881 m .20791 .08674 L .20785 .08664 L F
.20423 .07981 m .20989 .08953 L p .20395 .07914 L .389 .454 .901 r
F P
0 g
s
.20344 .078 m .20266 .07626 L .20333 .07765 L .996 .845 .938 r
F
.2032 .07747 m .20381 .07882 L .20162 .07343 L .838 .764 .988 r
F
.20391 .07907 m .20372 .07878 L .20344 .078 L .996 .845 .938 r
F
.20372 .07878 m .20365 .07867 L .20356 .07838 L F
.20372 .07878 m .20356 .07838 L .20344 .078 L F
.20372 .07876 m .20346 .07805 L .20353 .07827 L .838 .764 .988 r
F
.20349 .07809 m .20346 .07805 L p .20364 .0786 L .389 .454 .901 r
F P
0 g
s
.20344 .078 m .20348 .07808 L p .20348 .07811 L .131 .185 .747 r
F P
0 g
s
.20346 .07805 m .20344 .078 L p .20351 .07821 L .389 .454 .901 r
F P
0 g
s
.20364 .0786 m .20346 .07805 L .20351 .07821 L .389 .454 .901 r
F
.20379 .07878 m .2038 .07881 L p .2038 .07881 L .947 .735 .845 r
F P
0 g
s
.20379 .07878 m .20381 .07882 L .20381 .07882 L .947 .735 .845 r
F
.2038 .07881 m .20379 .07878 L p .20381 .07882 L F P
0 g
s
.20383 .07883 m .20383 .07883 L p .20386 .07883 L .957 .827 .957 r
F P
0 g
s
.20381 .07882 m .20382 .07883 L .20383 .07883 L .947 .735 .845 r
F
.20383 .07883 m .20399 .07895 L .20393 .07883 L F
.20379 .07878 m .2036 .07826 L p .20525 .08147 L F P
0 g
s
.20367 .07856 m .20348 .07808 L .20437 .07986 L .131 .185 .747 r
F
.20372 .07876 m .20381 .07882 L .20346 .07805 L .838 .764 .988 r
F
.20365 .07867 m .20267 .07629 L p .20344 .078 L .996 .845 .938 r
F P
0 g
s
.20333 .07765 m .20276 .07647 L .20195 .07433 L .996 .845 .938 r
F
.20235 .07537 m .20242 .07558 L p .20155 .07324 L .141 .19 .747 r
F P
0 g
s
.20242 .07558 m .20356 .07871 L p .20195 .07434 L .141 .19 .747 r
F P
0 g
s
.20242 .07558 m .20195 .07434 L .20155 .07324 L .141 .19 .747 r
F
.20155 .07324 m .20195 .07434 L .19869 .0657 L F
.20195 .07434 m .20351 .07858 L .20023 .06982 L F
.20195 .07434 m .20023 .06982 L .19869 .0657 L F
.20119 .07255 m .20298 .0774 L p .20086 .07141 L .45 .52 .945 r
F P
0 g
s
.20255 .07622 m .1979 .06362 L p .20268 .07648 L .141 .19 .747 r
F P
0 g
s
.20086 .07147 m .20111 .07232 L .2002 .06992 L .996 .845 .938 r
F
.20178 .07364 m .20139 .07287 L .20086 .07147 L F
.20274 .07652 m .1992 .06711 L .19986 .06884 L .141 .19 .747 r
F
.20139 .07287 m .20111 .07232 L .20086 .07147 L .996 .845 .938 r
F
.20351 .07858 m .2035 .07857 L .19869 .0657 L .141 .19 .747 r
F
.20274 .07652 m .20268 .07648 L .1992 .06711 L F
.20293 .07758 m .19702 .06742 L .2021 .07566 L p .163 .299 .827 r
F P
0 g
s
.20242 .07558 m .20232 .07533 L .20261 .07708 L .304 .419 .897 r
F
.2021 .07566 m .20235 .07606 L p .20293 .07758 L .163 .299 .827 r
F P
0 g
s
.20238 .07621 m .2021 .07566 L .20239 .07637 L .251 .371 .865 r
F
.20039 .07041 m .20239 .07642 L p .20232 .07533 L .304 .419 .897 r
F P
0 g
s
.20235 .07606 m .20237 .0761 L p .20307 .0778 L .163 .299 .827 r
F P
0 g
s
.20307 .0776 m .20306 .07759 L .20305 .07756 L .389 .454 .901 r
F
.20313 .07792 m .204 .07937 L .20305 .07756 L F
.203 .07733 m .20295 .07733 L .20298 .0774 L F
.20313 .07792 m .203 .07733 L .20293 .07756 L F
.20298 .0774 m .20262 .07643 L p .20345 .07844 L .141 .19 .747 r
F P
0 g
s
.203 .07733 m .20296 .07715 L .20295 .07733 L .389 .454 .901 r
F
.20305 .07756 m .20305 .07755 L .20305 .07754 L F
.2035 .07856 m .20262 .07643 L .2035 .07856 L .141 .19 .747 r
F
.20345 .07844 m .20262 .07643 L .2035 .07857 L F
.20262 .07643 m .20255 .07622 L p .20268 .07648 L F P
0 g
s
.2035 .07854 m .20242 .07558 L p .20253 .07645 L .304 .419 .897 r
F P
0 g
s
.20177 .07456 m .20039 .07041 L p .20221 .0753 L .168 .309 .829 r
F P
0 g
s
.20221 .0753 m .20039 .07041 L .20169 .07383 L .168 .309 .829 r
F
.20177 .07406 m .20169 .07383 L .20175 .074 L F
.20239 .07637 m .2021 .07566 L .20221 .07604 L .251 .371 .865 r
F
.20298 .0774 m .20295 .07733 L .20293 .07756 L .389 .454 .901 r
F
.20307 .0778 m .20237 .0761 L .20311 .07771 L .163 .299 .827 r
F
.20232 .07533 m .20237 .0761 L .20249 .07637 L .304 .419 .897 r
F
.20237 .0761 m .20239 .07642 L .20249 .07637 L F
.20366 .07877 m .20359 .07861 L .20389 .07919 L .163 .299 .827 r
F
.20305 .07756 m .204 .07937 L .20305 .07756 L .389 .454 .901 r
F
.20356 .07871 m .2086 .09264 L .20317 .07792 L p .141 .19 .747 r
F P
0 g
s
.20313 .07792 m .20311 .07785 L .20345 .07835 L .163 .299 .827 r
F
.20313 .07792 m .20307 .0778 L .20311 .07785 L F
.20313 .07792 m .20345 .07835 L .20346 .07842 L F
.20307 .0778 m .20313 .07774 L .20332 .07817 L F
.20293 .07758 m .20235 .07606 L .20307 .0778 L F
.20313 .07792 m .20293 .07758 L p .20307 .0778 L F P
0 g
s
.2035 .07854 m .20253 .07645 L .20261 .07708 L .304 .419 .897 r
F
.20366 .07876 m .20238 .07621 L .20239 .07637 L .251 .371 .865 r
F
.20392 .07925 m .20368 .07879 L .20366 .07876 L F
.20313 .07774 m .20325 .07759 L .20332 .07817 L .163 .299 .827 r
F
.20368 .07879 m .20238 .07621 L .20366 .07876 L .251 .371 .865 r
F
.20195 .07433 m .20244 .07567 L .20169 .07383 L .996 .845 .938 r
F
.20195 .07433 m .20169 .07383 L .20111 .07232 L F
.20162 .07363 m .20372 .07876 L p .20111 .07229 L .838 .764 .988 r
F P
0 g
s
.20423 .07981 m .20306 .07759 L .20307 .0776 L .389 .454 .901 r
F
.20353 .07827 m .20346 .07805 L .2032 .07747 L .838 .764 .988 r
F
.20361 .07853 m .20361 .07854 L .20344 .078 L .389 .454 .901 r
F
.20353 .0783 m .20363 .07862 L .20387 .07922 L .957 .827 .957 r
F
.20361 .07854 m .20364 .0786 L .20349 .07814 L .389 .454 .901 r
F
.20361 .07854 m .20349 .07814 L .20344 .078 L F
.20391 .07921 m .20353 .0783 L p .20387 .07922 L .957 .827 .957 r
F P
0 g
s
.20344 .078 m .20267 .07629 L .20267 .07627 L .996 .845 .938 r
F
.20353 .0783 m .20244 .07563 L p .20363 .07862 L .957 .827 .957 r
F P
0 g
s
.20345 .07835 m .20327 .07756 L .20375 .07881 L .163 .299 .827 r
F
.20307 .0778 m .20322 .07747 L .20325 .07759 L F
.20392 .07925 m .20295 .07704 L .20334 .07811 L .251 .371 .865 r
F
.20364 .07837 m .20349 .07809 L p .20361 .07849 L .389 .454 .901 r
F P
0 g
s
.20364 .0786 m .20364 .07837 L .20361 .07849 L .389 .454 .901 r
F
.20467 .08027 m .2036 .07826 L .20425 .07945 L .947 .735 .845 r
F
.20372 .07876 m .20372 .07876 L p .2038 .07881 L .838 .764 .988 r
F P
0 g
s
.20381 .07882 m .20525 .08147 L .20381 .07882 L .947 .735 .845 r
F
.20465 .08036 m .20452 .08013 L .20989 .08953 L .131 .185 .747 r
F
.20452 .08013 m .20437 .07986 L .20989 .08953 L F
.20123 .07241 m .20135 .07272 L .20383 .07883 L .957 .827 .957 r
F
.20178 .07364 m .20195 .07433 L .20139 .07287 L .996 .845 .938 r
F
.20098 .07185 m .20383 .07883 L p .20049 .07059 L .957 .827 .957 r
F P
0 g
s
.20239 .07642 m .20261 .07708 L p .20249 .07637 L .304 .419 .897 r
F P
0 g
s
.20313 .07792 m .20423 .07981 L p .204 .07937 L .389 .454 .901 r
F P
0 g
s
.20375 .07899 m .20366 .07877 L .20389 .07919 L .163 .299 .827 r
F
.20346 .07842 m .20346 .0784 L .20349 .07845 L F
.20346 .0784 m .20345 .07835 L .20349 .07845 L F
.20375 .07899 m .20313 .07792 L p .20359 .07861 L F P
0 g
s
.20392 .07925 m .20221 .07604 L .20292 .07754 L .251 .371 .865 r
F
.20351 .07858 m .20356 .07871 L .2035 .07857 L .141 .19 .747 r
F
.20356 .07871 m .20352 .07863 L .20352 .07861 L F
.2035 .07857 m .2035 .07856 L .2035 .07856 L F
.20356 .07871 m .20345 .07844 L .2035 .07857 L F
.20356 .07871 m .2035 .07854 L p .20261 .07708 L .304 .419 .897 r
F P
0 g
s
.20356 .07874 m .20347 .07848 L .20292 .07754 L .251 .371 .865 r
F
.20261 .07708 m .20594 .08707 L p .20356 .07871 L .304 .419 .897 r
F P
0 g
s
.20316 .07804 m .20292 .07754 L .19702 .06742 L .251 .371 .865 r
F
.20356 .07874 m .20292 .07754 L .20316 .07804 L F
.20339 .0787 m .20302 .07775 L .20337 .07861 L .168 .309 .829 r
F
.20302 .07775 m .20177 .07456 L .20311 .07794 L F
.20302 .07775 m .20311 .07794 L .20337 .07861 L F
.20322 .07818 m .20177 .07456 L .20221 .0753 L F
.20337 .07861 m .20311 .07794 L .20322 .07818 L F
.20311 .07794 m .20177 .07456 L .20322 .07818 L F
.20237 .0761 m .20322 .07747 L p .20311 .07771 L .163 .299 .827 r
F P
0 g
s
.20295 .07704 m .2021 .07566 L p .20334 .07811 L .251 .371 .865 r
F P
0 g
s
.20298 .07723 m .20353 .07905 L p .20175 .074 L .168 .309 .829 r
F P
0 g
s
.2028 .07765 m .20239 .07642 L p .20341 .07877 L .168 .309 .829 r
F P
0 g
s
.20353 .07905 m .20342 .07885 L .20341 .07877 L .168 .309 .829 r
F
.20342 .07885 m .2028 .07765 L .20341 .07877 L F
.19702 .06742 m .19933 .07268 L .20333 .07842 L p .251 .371 .865 r
F P
0 g
s
.20353 .07905 m .20594 .08707 L .2028 .07765 L p .168 .309 .829 r
F P
0 g
s
.20335 .07844 m .20333 .07837 L p .20214 .07719 L .224 .361 .863 r
F P
0 g
s
.20335 .07844 m .20214 .07719 L .20214 .07719 L .224 .361 .863 r
F
.20324 .07819 m .20243 .07712 L .20316 .078 L .196 .343 .852 r
F
.20333 .07842 m .20243 .07712 L p .20331 .07827 L F P
0 g
s
.20353 .07905 m .20239 .07642 L .20302 .07775 L .168 .309 .829 r
F
.20333 .07842 m .20316 .07804 L .19702 .06742 L .251 .371 .865 r
F
.20333 .07837 m .20293 .07751 L .20213 .07716 L .224 .361 .863 r
F
.2021 .07566 m .19702 .06742 L p .20237 .07661 L .251 .371 .865 r
F P
0 g
s
.20243 .07712 m .19933 .07268 L .20146 .07462 L p .196 .343 .852 r
F P
0 g
s
.19933 .07268 m .20257 .07982 L p .202 .07638 L .205 .345 .853 r
F P
0 g
s
.20214 .07719 m .20213 .07716 L .20209 .07714 L .224 .361 .863 r
F
.20333 .07837 m .20213 .07716 L .20214 .07719 L F
.20175 .07489 m .19933 .07268 L p .202 .07638 L .205 .345 .853 r
F P
0 g
s
.20316 .078 m .20243 .07712 L .20266 .07709 L .196 .343 .852 r
F
.20135 .07306 m .20163 .07462 L p .20308 .07757 L .224 .361 .863 r
F P
0 g
s
.20239 .07642 m .20177 .07456 L p .20302 .07775 L .168 .309 .829 r
F P
0 g
s
.20316 .078 m .20266 .07709 L .20274 .07708 L .196 .343 .852 r
F
.20146 .07462 m .20175 .07489 L p .20243 .07712 L F P
0 g
s
.20175 .07489 m .20222 .07532 L p .20232 .07614 L .196 .343 .852 r
F P
0 g
s
.20243 .07712 m .20175 .07489 L .20232 .07614 L .196 .343 .852 r
F
.20292 .07754 m .20221 .07604 L .20237 .07661 L .251 .371 .865 r
F
.20163 .07462 m .20209 .07714 L p .20293 .07751 L .224 .361 .863 r
F P
0 g
s
.20308 .07757 m .20163 .07462 L .20293 .07751 L .224 .361 .863 r
F
.20331 .07827 m .20316 .078 L .20305 .07704 L .196 .343 .852 r
F
.20333 .07837 m .20308 .07757 L p .20293 .07751 L .224 .361 .863 r
F P
0 g
s
.20316 .078 m .20274 .07708 L .20305 .07704 L .196 .343 .852 r
F
.20243 .07712 m .20222 .07532 L .20291 .07706 L F
.20222 .07532 m .20175 .07489 L p .20334 .07824 L .205 .345 .853 r
F P
0 g
s
.20334 .07824 m .20175 .07489 L .202 .07638 L .205 .345 .853 r
F
.20331 .07827 m .20324 .07819 L .20316 .078 L .196 .343 .852 r
F
.20209 .07714 m .20257 .07982 L p .20335 .07844 L .224 .361 .863 r
F P
0 g
s
.20257 .07982 m .20334 .07824 L p .202 .07638 L .205 .345 .853 r
F P
0 g
s
.20214 .07719 m .20214 .07719 L .20209 .07714 L .224 .361 .863 r
F
.20333 .07842 m .20356 .07874 L p .20316 .07804 L .251 .371 .865 r
F P
0 g
s
.20355 .07873 m .20333 .07842 L p .20331 .07827 L .196 .343 .852 r
F P
0 g
s
.20383 .07883 m .20155 .07324 L p .20381 .07882 L .838 .764 .988 r
F P
0 g
s
.20023 .06993 m .20039 .07041 L p .20174 .07379 L .304 .419 .897 r
F P
0 g
s
.20139 .07287 m .2014 .07289 L .2014 .07288 L .996 .845 .938 r
F
.2014 .07289 m .20195 .07433 L .2014 .07288 L F
.2014 .07288 m .20195 .07433 L .20162 .07355 L F
.20242 .07558 m .20235 .07537 L p .20039 .07041 L .304 .419 .897 r
F P
0 g
s
.20223 .07506 m .20162 .07343 L .20109 .07213 L .838 .764 .988 r
F
.20223 .07506 m .20109 .07213 L .20212 .07479 L F
.20235 .07537 m .20219 .07493 L p .20174 .07379 L .304 .419 .897 r
F P
0 g
s
.20139 .07287 m .20162 .07355 L .20123 .07255 L .996 .845 .938 r
F
.20235 .07537 m .20174 .07379 L .20039 .07041 L .304 .419 .897 r
F
.20123 .07255 m .20162 .07355 L .20111 .07232 L .996 .845 .938 r
F
.2002 .06992 m .20111 .07232 L .19644 .06104 L F
.20212 .07479 m .20067 .07109 L .20111 .07229 L .838 .764 .988 r
F
.20381 .07882 m .2016 .07338 L .20161 .0734 L F
.20381 .07882 m .20155 .07324 L .2016 .07338 L F
.19644 .06104 m .19669 .06072 L p .20086 .07147 L .996 .845 .938 r
F P
0 g
s
.20109 .07213 m .19837 .06543 L .19841 .06554 L .838 .764 .988 r
F
.20381 .07882 m .20107 .07207 L .20109 .07213 L F
.20107 .07207 m .19837 .06543 L .20109 .07213 L F
.20123 .07241 m .20075 .07088 L .1979 .06362 L .957 .827 .957 r
F
.20108 .07199 m .19869 .0657 L .1972 .06174 L .141 .19 .747 r
F
.20219 .07493 m .20155 .07324 L .20108 .07199 L F
.20155 .07324 m .19869 .0657 L .20108 .07199 L F
.20155 .07324 m .20098 .07185 L p .2016 .07338 L .838 .764 .988 r
F P
0 g
s
.2016 .07338 m .20098 .07185 L .20105 .07202 L .838 .764 .988 r
F
.19406 .05277 m .18941 .04061 L .19343 .05267 L closepath p .486 .878 1 r
F P
0 g
s
.19343 .05267 m .18877 .0405 L .19249 .05116 L closepath p .039 .039 .512 r
F P
0 g
s
.18877 .0405 m .19343 .05267 L .18941 .04061 L closepath p .486 .878 1 r
F P
0 g
s
.19217 .04855 m .19678 .06061 L .19183 .049 L closepath p 1 1 .894 r
F P
0 g
s
.19678 .06061 m .19217 .04855 L .19548 .05724 L p .664 .338 .001 r
F P
0 g
s
.19706 .06135 m .19678 .06061 L p .19548 .05724 L .664 .338 .001 r
F P
0 g
s
.19706 .06135 m .21632 .06142 L p .19698 .06115 L .6 .968 1 r
F P
0 g
s
.19698 .06115 m .19696 .06115 L .19706 .06135 L .6 .968 1 r
F
.19548 .05724 m .19217 .04856 L p .19245 .04931 L .664 .338 .001 r
F P
0 g
s
.19217 .04856 m .19217 .04855 L .19222 .0487 L p .664 .338 .001 r
F P
0 g
s
.19222 .0487 m .19245 .04931 L p .19217 .04856 L .664 .338 .001 r
F P
0 g
s
.19678 .06061 m .19706 .06135 L p .19708 .06139 L .736 .398 .02 r
F P
0 g
s
.19669 .06072 m .19674 .06066 L p .19882 .06612 L .996 .845 .938 r
F P
0 g
s
.19882 .06612 m .20086 .07147 L .19669 .06072 L .996 .845 .938 r
F
.19578 .05282 m .19116 .04065 L .19406 .05277 L closepath p .589 .958 1 r
F P
0 g
s
.18941 .04061 m .19406 .05277 L .19116 .04065 L closepath p .589 .958 1 r
F P
0 g
s
.19644 .06104 m .19183 .049 L .19678 .06061 L closepath p 1 1 .894 r
F P
0 g
s
.21201 .04937 m .19245 .04931 L p .19247 .04935 L .6 .968 1 r
F P
0 g
s
.1967 .06135 m .20936 .05145 L p .19247 .04935 L .6 .968 1 r
F P
0 g
s
.20936 .05145 m .21201 .04937 L p .19247 .04935 L .6 .968 1 r
F P
0 g
s
.21632 .06142 m .21201 .04937 L .19697 .06114 L p .6 .968 1 r
F P
0 g
s
.21632 .06142 m .19697 .06114 L .19698 .06115 L .6 .968 1 r
F
.21632 .06142 m .19706 .06135 L p .20359 .07815 L .957 .827 .957 r
F P
0 g
s
.19706 .06135 m .2013 .07239 L .19708 .06139 L .736 .398 .02 r
F
.19841 .06487 m .20359 .07815 L .19706 .06135 L .957 .827 .957 r
F
.20325 .07759 m .20327 .07756 L .20345 .07835 L .163 .299 .827 r
F
.20322 .07747 m .20327 .07756 L p .20325 .07759 L F P
0 g
s
.20327 .07756 m .20295 .07704 L p .20392 .07925 L .251 .371 .865 r
F P
0 g
s
.20276 .07647 m .20266 .07626 L .20195 .07433 L .996 .845 .938 r
F
.20267 .07629 m .20387 .07922 L p .20244 .07563 L .957 .827 .957 r
F P
0 g
s
.20348 .07808 m .20989 .08953 L p .20437 .07986 L .131 .185 .747 r
F P
0 g
s
.20989 .08953 m .20364 .07837 L p .20364 .0786 L .389 .454 .901 r
F P
0 g
s
.20615 .08244 m .20592 .08206 L .20733 .08414 L .163 .299 .827 r
F
.20389 .07919 m .20592 .08206 L .20615 .08244 L F
.20389 .07919 m .20359 .07861 L .20592 .08206 L F
.20401 .07942 m .20389 .07919 L .20389 .07919 L F
.20349 .07845 m .20345 .07835 L .20733 .08414 L F
.20413 .07883 m .20406 .07883 L .20391 .07865 L .947 .735 .845 r
F
.20395 .07883 m .20442 .07883 L .21632 .06142 L .957 .827 .957 r
F
.20383 .07868 m .20385 .07868 L .20393 .07883 L .947 .735 .845 r
F
.20391 .07883 m .20393 .07883 L .20425 .07945 L F
.20383 .07868 m .20393 .07883 L .20391 .07883 L F
.20392 .07878 m .20384 .07863 L p .20385 .07869 L .658 .603 .927 r
F P
0 g
s
.20393 .07883 m .20392 .07878 L .20385 .07869 L .658 .603 .927 r
F
.20359 .07815 m .20386 .07883 L p .20392 .07878 L .957 .827 .957 r
F P
0 g
s
.20386 .07883 m .20393 .07883 L .20392 .07878 L .957 .827 .957 r
F
.20406 .07883 m .20718 .08277 L .20396 .07883 L .947 .735 .845 r
F
.20395 .07883 m .20442 .07883 L .20442 .07883 L .957 .827 .957 r
F
.20395 .07883 m .20442 .07883 L .20395 .07883 L F
.20327 .07756 m .20733 .08414 L p .20375 .07881 L .163 .299 .827 r
F P
0 g
s
.20392 .07925 m .20618 .08249 L p .20327 .07756 L .251 .371 .865 r
F P
0 g
s
.20618 .08249 m .20733 .08414 L .20474 .07995 L p .251 .371 .865 r
F P
0 g
s
.20474 .07995 m .20327 .07756 L p .20618 .08249 L .251 .371 .865 r
F P
0 g
s
.20331 .07827 m .20305 .07704 L .20399 .07693 L .196 .343 .852 r
F
.19924 .06703 m .2013 .07239 L p .19706 .06135 L .736 .398 .02 r
F P
0 g
s
.19852 .06481 m .20219 .07493 L p .20108 .07199 L .141 .19 .747 r
F P
0 g
s
.19706 .06135 m .19704 .06135 L p .19841 .06487 L .957 .827 .957 r
F P
0 g
s
.19706 .06135 m .1972 .06174 L .19924 .06703 L p .736 .398 .02 r
F P
0 g
s
.1972 .06174 m .19737 .06213 L p .20108 .07199 L .141 .19 .747 r
F P
0 g
s
.19924 .06703 m .1972 .06174 L p .19871 .06568 L .45 .52 .945 r
F P
0 g
s
.19737 .06213 m .19852 .06481 L p .20108 .07199 L .141 .19 .747 r
F P
0 g
s
.2032 .07747 m .20162 .07343 L .20223 .07506 L .838 .764 .988 r
F
.20219 .07493 m .20235 .07537 L p .20155 .07324 L .141 .19 .747 r
F P
0 g
s
.20123 .07241 m .20049 .07059 L p .20135 .07272 L .957 .827 .957 r
F P
0 g
s
.20175 .074 m .20169 .07383 L .2015 .07329 L .168 .309 .829 r
F
.20049 .07059 m .20123 .07241 L p .19966 .06843 L .957 .827 .957 r
F P
0 g
s
.20023 .06993 m .19966 .06843 L .19686 .06135 L .957 .827 .957 r
F
.20049 .07059 m .19966 .06843 L .20023 .06993 L F
.19986 .06884 m .1992 .06711 L .1979 .06362 L .141 .19 .747 r
F
.19704 .06135 m .19698 .06135 L p .20123 .07241 L .957 .827 .957 r
F P
0 g
s
.19966 .06843 m .20123 .07241 L .19698 .06135 L .957 .827 .957 r
F
.20018 .0698 m .20023 .06993 L p .20098 .07185 L .304 .419 .897 r
F P
0 g
s
.20098 .07185 m .20023 .06993 L .20049 .07059 L .304 .419 .897 r
F
.19852 .06481 m .20107 .07207 L p .20219 .07493 L F P
0 g
s
.2016 .07338 m .1993 .06771 L .20161 .0734 L .838 .764 .988 r
F
.20098 .07185 m .20023 .06993 L .20174 .07379 L .304 .419 .897 r
F
.20018 .0698 m .19852 .06481 L p .2015 .07329 L .168 .309 .829 r
F P
0 g
s
.20169 .07383 m .20061 .07095 L .2015 .07329 L .168 .309 .829 r
F
.20061 .07095 m .20018 .0698 L .2015 .07329 L F
.20219 .07493 m .20018 .0698 L .20098 .07185 L .304 .419 .897 r
F
.20157 .07347 m .2015 .07329 L .19852 .06481 L .168 .309 .829 r
F
.20175 .074 m .2015 .07329 L .20157 .07347 L F
.19698 .06135 m .19686 .06135 L p .19966 .06843 L .957 .827 .957 r
F P
0 g
s
.20023 .06993 m .20023 .06993 L p .20049 .07059 L .304 .419 .897 r
F P
0 g
s
.19686 .06135 m .1967 .06135 L .20049 .07059 L p .957 .827 .957 r
F P
0 g
s
.20219 .07493 m .20098 .07185 L .20174 .07379 L .304 .419 .897 r
F
.20266 .07626 m .20142 .07322 L p .20244 .07567 L .996 .845 .938 r
F P
0 g
s
.20244 .07567 m .20142 .07322 L .20133 .07298 L .996 .845 .938 r
F
.20267 .07629 m .20266 .07626 L p .20267 .07627 L F P
0 g
s
.20135 .07306 m .20135 .07303 L p .20332 .07829 L .205 .346 .853 r
F P
0 g
s
.20244 .07567 m .20091 .07196 L .20053 .07101 L .996 .845 .938 r
F
.20244 .07563 m .20162 .07363 L p .20267 .07629 L .957 .827 .957 r
F P
0 g
s
.20111 .07232 m .20169 .07383 L .20103 .07222 L .996 .845 .938 r
F
.20031 .07038 m .20103 .07222 L .19644 .06104 L F
.20111 .07232 m .20103 .07222 L .20031 .07038 L F
.20162 .07363 m .20179 .07409 L .20267 .07629 L .957 .827 .957 r
F
.20298 .07723 m .20195 .07452 L .19852 .06481 L .168 .309 .829 r
F
.20212 .07479 m .20109 .07213 L .20067 .07109 L .838 .764 .988 r
F
.20195 .07452 m .20174 .07397 L .19852 .06481 L .168 .309 .829 r
F
.19852 .06481 m .20053 .06931 L p .20298 .07723 L F P
0 g
s
.20101 .07117 m .20121 .0723 L p .20168 .07301 L .224 .361 .863 r
F P
0 g
s
.20121 .0723 m .20135 .07303 L p .20168 .07301 L .224 .361 .863 r
F P
0 g
s
.20193 .07371 m .20168 .07301 L .20135 .07303 L .224 .361 .863 r
F
.20135 .07303 m .20135 .07306 L p .20136 .07307 L F P
0 g
s
.20193 .07371 m .20135 .07303 L .20136 .07307 L .224 .361 .863 r
F
.20162 .07363 m .20104 .07222 L p .20179 .07409 L .957 .827 .957 r
F P
0 g
s
.20195 .07452 m .20175 .074 L .20174 .07397 L .168 .309 .829 r
F
.20104 .07222 m .20107 .07231 L .20267 .07629 L .957 .827 .957 r
F
.20138 .07313 m .20267 .07629 L p .20093 .07198 L F P
0 g
s
.20308 .07757 m .203 .07729 L p .20135 .07306 L .224 .361 .863 r
F P
0 g
s
.20221 .0753 m .20169 .07383 L .20177 .07406 L .168 .309 .829 r
F
.203 .07729 m .20193 .07371 L p .20136 .07307 L .224 .361 .863 r
F P
0 g
s
.203 .07729 m .20136 .07307 L .20135 .07306 L .224 .361 .863 r
F
.20109 .07094 m .20298 .07723 L p .20053 .06931 L .168 .309 .829 r
F P
0 g
s
.20222 .07532 m .20399 .07693 L p .20291 .07706 L .196 .343 .852 r
F P
0 g
s
.20334 .07824 m .20352 .07788 L p .20222 .07532 L .205 .345 .853 r
F P
0 g
s
.20281 .07586 m .20222 .07532 L p .20352 .07788 L .205 .345 .853 r
F P
0 g
s
.20193 .07371 m .20077 .06986 L .20101 .07117 L p .224 .361 .863 r
F P
0 g
s
.20039 .07041 m .20018 .0698 L p .20169 .07383 L .168 .309 .829 r
F P
0 g
s
.20161 .0734 m .19987 .06912 L .19988 .06916 L .838 .764 .988 r
F
.1967 .06135 m .19644 .06104 L .20162 .07363 L p F P
0 g
s
.20162 .07363 m .20111 .07229 L .1967 .06135 L .838 .764 .988 r
F
.20135 .07303 m .20077 .06986 L p .20319 .07786 L .205 .346 .853 r
F P
0 g
s
.20332 .07829 m .20135 .07303 L .20319 .07786 L .205 .346 .853 r
F
.20104 .07222 m .20093 .07198 L p .20107 .07231 L .957 .827 .957 r
F P
0 g
s
.20142 .07322 m .20091 .07196 L p .20133 .07298 L .996 .845 .938 r
F P
0 g
s
.20091 .07196 m .19644 .06104 L p .20053 .07101 L .996 .845 .938 r
F P
0 g
s
.20093 .07198 m .19644 .06104 L .20138 .07313 L p .957 .827 .957 r
F P
0 g
s
.20352 .07788 m .20399 .07693 L .20281 .07586 L p .205 .345 .853 r
F P
0 g
s
.20053 .06931 m .20077 .06986 L .20109 .07094 L p .168 .309 .829 r
F P
0 g
s
.20077 .06986 m .20399 .07693 L .20332 .07829 L p .205 .346 .853 r
F P
0 g
s
.19394 .05282 m .19852 .06481 L .19262 .04976 L p .008 .008 .31 r
F P
0 g
s
.19262 .04976 m .1926 .0497 L .19394 .05282 L p .008 .008 .31 r
F P
0 g
s
.19245 .04931 m .19209 .04931 L .19251 .05039 L p .6 .968 1 r
F P
0 g
s
.19251 .05039 m .1967 .06135 L p .19245 .04931 L .6 .968 1 r
F P
0 g
s
.1967 .06135 m .19209 .04931 L .19644 .06104 L closepath p .067 .152 .694 r
F P
0 g
s
.19183 .049 m .19644 .06104 L .19209 .04931 L closepath p .067 .152 .694 r
F P
0 g
s
.19209 .04931 m .19609 .05899 L p .19195 .04929 L 1 .771 .77 r
F P
0 g
s
.1926 .0497 m .1972 .06174 L .19548 .05724 L p .664 .338 .001 r
F P
0 g
s
.19548 .05724 m .1972 .06174 L .19706 .06135 L p .664 .338 .001 r
F P
0 g
s
.19697 .06114 m .19696 .06115 L p .19698 .06115 L .6 .968 1 r
F P
0 g
s
.19245 .04931 m .1926 .0497 L p .19548 .05724 L .664 .338 .001 r
F P
0 g
s
.1972 .06174 m .1926 .0497 L .19852 .06481 L closepath p .008 .008 .31 r
F P
0 g
s
.19852 .06481 m .19394 .05282 L .20077 .06986 L closepath p .011 .011 .328 r
F P
0 g
s
.19622 .05794 m .20077 .06986 L .19396 .05287 L p .011 .011 .328 r
F P
0 g
s
.19396 .05287 m .19394 .05282 L .19622 .05794 L p .011 .011 .328 r
F P
0 g
s
.1979 .06362 m .1972 .06174 L p .19868 .06569 L .141 .19 .747 r
F P
0 g
s
.20077 .06986 m .19622 .05794 L .20399 .07693 L closepath p .013 .013 .339 r
F P
0 g
s
.19948 .06511 m .20399 .07693 L .19625 .05801 L p .013 .013 .339 r
F P
0 g
s
.19625 .05801 m .19622 .05794 L .19948 .06511 L p .013 .013 .339 r
F P
0 g
s
.20111 .07229 m .20067 .07109 L .1967 .06135 L .838 .764 .988 r
F
.1993 .06771 m .19837 .06543 L p .19932 .06778 L F P
0 g
s
.19837 .06543 m .1967 .06135 L p .19841 .06554 L .838 .764 .988 r
F P
0 g
s
.20049 .07059 m .1967 .06135 L .20098 .07185 L p .957 .827 .957 r
F P
0 g
s
.20442 .07883 m .20543 .07883 L .20442 .07883 L .957 .827 .957 r
F
.20542 .07884 m .20543 .07883 L .20442 .07883 L F
.20413 .07883 m .20718 .08277 L .20406 .07883 L .947 .735 .845 r
F
.20543 .07883 m .20718 .08277 L p .20413 .07883 L F P
0 g
s
.20399 .07693 m .19948 .06511 L .20733 .08414 L closepath p .014 .014 .348 r
F P
0 g
s
.20287 .07242 m .20733 .08414 L .19953 .06521 L p .014 .014 .348 r
F P
0 g
s
.19953 .06521 m .19948 .06511 L .20287 .07242 L p .014 .014 .348 r
F P
0 g
s
.19948 .06511 m .19815 .06464 L p .20272 .07222 L 1 .711 .353 r
F P
0 g
s
.20332 .07829 m .20257 .07982 L .20135 .07306 L p .205 .346 .853 r
F P
0 g
s
.19988 .06916 m .1993 .06771 L .19932 .06778 L .838 .764 .988 r
F
.19987 .06912 m .1993 .06771 L .19988 .06916 L F
.20098 .07185 m .1993 .06771 L p .20105 .07202 L F P
0 g
s
.20353 .07905 m .20339 .0787 L .20337 .07861 L .168 .309 .829 r
F
.20257 .07982 m .20594 .08707 L .20335 .07844 L p .224 .361 .863 r
F P
0 g
s
.2086 .09264 m .20356 .07871 L p .20496 .08363 L .304 .419 .897 r
F P
0 g
s
.20409 .08059 m .20494 .08297 L .20445 .08184 L p .983 .84 .947 r
F P
0 g
s
.20445 .08184 m .19272 .05502 L p .20409 .08059 L .983 .84 .947 r
F P
0 g
s
.20409 .08059 m .19272 .05502 L .19278 .05514 L .983 .84 .947 r
F
.19907 .06654 m .20409 .08059 L p .19394 .05776 L F P
0 g
s
.19717 .06298 m .19394 .05776 L .1919 .05315 L .983 .84 .947 r
F
.19907 .06654 m .19394 .05776 L .19717 .06298 L F
.19907 .06654 m .1919 .05315 L .1935 .05613 L F
.19804 .06461 m .1935 .05613 L .19412 .05727 L F
.19907 .06654 m .1935 .05613 L .19804 .06461 L F
.20355 .07873 m .20347 .07857 L .20399 .07693 L .196 .343 .852 r
F
.20347 .07857 m .20331 .07827 L .20333 .07822 L F
.20347 .07857 m .20333 .07822 L .20399 .07693 L F
.20356 .07874 m .20392 .07925 L p .20347 .07848 L .251 .371 .865 r
F P
0 g
s
.20399 .07693 m .20733 .08414 L .20355 .07873 L p .196 .343 .852 r
F P
0 g
s
.21008 .09422 m .21193 .09872 L .20985 .09361 L p .838 .764 .988 r
F P
0 g
s
.21195 .09454 m .20937 .09454 L .204 .07937 L p .947 .735 .845 r
F P
0 g
s
.20423 .07981 m .20401 .07942 L .20733 .08414 L .163 .299 .827 r
F
.20401 .07942 m .20389 .07919 L .20733 .08414 L F
.20733 .08414 m .20989 .08953 L .20423 .07981 L p F P
0 g
s
.20733 .08414 m .20287 .07242 L .20989 .08953 L closepath p .016 .016 .361 r
F P
0 g
s
.20547 .07789 m .20989 .08953 L .20293 .07256 L p .016 .016 .361 r
F P
0 g
s
.20293 .07256 m .20287 .07242 L .20547 .07789 L p .016 .016 .361 r
F P
0 g
s
.21201 .04937 m .2149 .05745 L p .20681 .07819 L 1 .942 .493 r
F P
0 g
s
.20989 .08953 m .20547 .07789 L .20683 .08124 L p .019 .019 .379 r
F P
0 g
s
.21051 .09036 m .21053 .0904 L .21041 .09082 L 1 1 .819 r
F
.21632 .06142 m .21053 .0904 L .21051 .09036 L F
.21632 .06142 m .21872 .06122 L p .21053 .0904 L F P
0 g
s
.21023 .09046 m .21632 .06142 L p .21041 .09082 L 1 1 .819 r
F P
0 g
s
.21026 .09029 m .20989 .08953 L p .20683 .08124 L .019 .019 .379 r
F P
0 g
s
.20989 .08953 m .21026 .09029 L p .20625 .08331 L .131 .185 .747 r
F P
0 g
s
.21126 .08553 m .21026 .09029 L p .21309 .06754 L 1 .942 .493 r
F P
0 g
s
.21028 .09063 m .20625 .08331 L p .21026 .09029 L .131 .185 .747 r
F P
0 g
s
.21026 .09029 m .21023 .09046 L p .21026 .09022 L 1 .942 .493 r
F P
0 g
s
.21026 .09022 m .21309 .06754 L .21026 .09029 L 1 .942 .493 r
F
.21029 .09059 m .21041 .09081 L .21029 .09059 L .658 .603 .927 r
F
.21041 .09085 m .21028 .09063 L p .21026 .09029 L .131 .185 .747 r
F P
0 g
s
.21 .08999 m .21004 .0897 L .21023 .09046 L 1 .942 .493 r
F
.21004 .0897 m .21309 .06754 L .21025 .09027 L F
.21004 .0897 m .21025 .09027 L .21023 .09046 L F
.20683 .08124 m .20547 .07789 L .20585 .07867 L p .019 .019 .379 r
F P
0 g
s
.21235 .09454 m .21226 .09454 L p .20828 .08704 L .947 .735 .845 r
F P
0 g
s
.21226 .09454 m .21195 .09454 L p .20703 .08494 L .947 .735 .845 r
F P
0 g
s
.20641 .0802 m .2077 .07775 L p .2084 .08484 L 1 .942 .493 r
F P
0 g
s
.20721 .08143 m .20683 .08124 L .20664 .08075 L .019 .019 .379 r
F
.2059 .07864 m .20681 .07819 L .20623 .07967 L 1 .942 .493 r
F
.2059 .07864 m .20623 .07967 L .20588 .07874 L F
.20503 .08258 m .20593 .07831 L p .20505 .0827 L F P
0 g
s
.20597 .07822 m .20505 .0827 L .20593 .07831 L 1 .942 .493 r
F
.20688 .07393 m .20505 .0827 L .2068 .07413 L F
.20937 .09454 m .20494 .08297 L .20641 .0802 L p F P
0 g
s
.20937 .09454 m .20641 .0802 L .2084 .08484 L 1 .942 .493 r
F
.20588 .07874 m .20623 .07967 L .20505 .0827 L F
.21364 .06649 m .20494 .08297 L p .2149 .05745 L F P
0 g
s
.20423 .07981 m .20375 .07899 L p .20389 .07919 L .163 .299 .827 r
F P
0 g
s
.19815 .06464 m .19843 .06537 L .20272 .07222 L 1 .711 .353 r
F
.20494 .08297 m .19907 .06654 L p .19975 .06777 L 1 .866 .918 r
F P
0 g
s
.1919 .05315 m .19182 .05297 L p .19412 .05727 L .983 .84 .947 r
F P
0 g
s
.19182 .05297 m .19185 .05304 L p .19336 .05587 L .957 .829 .959 r
F P
0 g
s
.19799 .06459 m .19394 .05696 L .19336 .05587 L .957 .829 .959 r
F
.19394 .05696 m .19182 .05297 L .19336 .05587 L F
.20085 .06977 m .20272 .07222 L p .19843 .06537 L 1 .711 .353 r
F P
0 g
s
.19838 .06523 m .19898 .06603 L .19961 .06751 L 1 .711 .353 r
F
.20579 .07896 m .19975 .06777 L .20085 .06977 L 1 .866 .918 r
F
.19844 .06474 m .19948 .06511 L .19961 .06679 L p 1 .711 .353 r
F P
0 g
s
.19907 .06654 m .19907 .06653 L p .19957 .06744 L 1 .866 .918 r
F P
0 g
s
.19957 .06744 m .19957 .06744 L .20582 .07879 L 1 .866 .918 r
F
.19907 .06654 m .19957 .06744 L .19957 .06744 L F
.19844 .06474 m .19864 .06511 L .19859 .0651 L 1 .711 .353 r
F
.19864 .06511 m .19961 .06679 L .19859 .0651 L F
.19957 .06744 m .20579 .07872 L .20582 .07879 L 1 .866 .918 r
F
.19907 .06653 m .19957 .06743 L .19957 .06744 L F
.19957 .06743 m .20579 .07872 L .19957 .06744 L F
.19961 .06679 m .19967 .06765 L p .19859 .0651 L 1 .711 .353 r
F P
0 g
s
.19815 .06464 m .19804 .06461 L p .19838 .06524 L 1 .711 .353 r
F P
0 g
s
.19824 .0649 m .19844 .06474 L .19898 .06603 L 1 .711 .353 r
F
.19804 .06461 m .19813 .06464 L p .19824 .0649 L F P
0 g
s
.19813 .06464 m .19844 .06474 L p .19824 .0649 L 1 .711 .353 r
F P
0 g
s
.19804 .06461 m .19898 .06603 L .19838 .06523 L 1 .711 .353 r
F
.19907 .06654 m .19102 .05114 L .19874 .06588 L .983 .84 .947 r
F
.19967 .06761 m .19961 .06679 L p .19841 .06525 L 1 .712 .348 r
F P
0 g
s
.19883 .06604 m .19967 .06761 L .19841 .06525 L 1 .712 .348 r
F
.19481 .05615 m .19601 .05904 L p .19509 .05688 L .947 .734 .843 r
F P
0 g
s
.19182 .05297 m .19123 .05162 L p .19907 .06654 L .983 .84 .947 r
F P
0 g
s
.19907 .06654 m .19123 .05162 L .19636 .06136 L .983 .84 .947 r
F
.19791 .06405 m .19828 .06503 L .19833 .06504 L 1 .712 .348 r
F
.19796 .06409 m .19833 .06504 L .1995 .06528 L F
.19791 .06405 m .19833 .06504 L .19796 .06409 L F
.19394 .05282 m .1926 .0497 L .19545 .0571 L p .992 .61 .145 r
F P
0 g
s
.19545 .0571 m .19609 .05873 L .19394 .05282 L .992 .61 .145 r
F
.19832 .06504 m .19836 .06515 L .19837 .06515 L 1 .712 .348 r
F
.19833 .06504 m .19837 .06515 L .1995 .06528 L F
.19832 .06504 m .19837 .06515 L .19833 .06504 L F
.19509 .05688 m .19601 .05904 L .19621 .05954 L .947 .734 .843 r
F
.19631 .05981 m .19509 .05688 L p .19621 .05954 L F P
0 g
s
.19609 .05899 m .19686 .06101 L .19195 .04929 L 1 .771 .77 r
F
.19245 .04931 m .19217 .04855 L .19243 .04922 L p .057 .057 .452 r
F P
0 g
s
.1926 .04921 m .19252 .04922 L p .21308 .04725 L .841 1 1 r
F P
0 g
s
.19252 .04922 m .19209 .04931 L p .19292 .04922 L .841 1 1 r
F P
0 g
s
.21308 .04725 m .19252 .04922 L .19292 .04922 L .841 1 1 r
F
.21201 .04937 m .21308 .04725 L p .19245 .04927 L F P
0 g
s
.19245 .04931 m .21201 .04937 L p .19245 .04927 L .841 1 1 r
F P
0 g
s
.19836 .06515 m .19838 .06518 L .1995 .06528 L 1 .712 .348 r
F
.19961 .06679 m .1995 .06528 L p .19846 .06519 L F P
0 g
s
.20272 .07222 m .20287 .07242 L .19948 .06511 L p 1 .711 .353 r
F P
0 g
s
.1984 .06522 m .19966 .06724 L .19903 .06641 L 1 .723 .371 r
F
.19836 .06512 m .20287 .07242 L p .1984 .06522 L F P
0 g
s
.19597 .05886 m .19672 .06067 L .19499 .05628 L .902 .713 .863 r
F
.19696 .06115 m .1967 .06135 L .19706 .06135 L p .6 .968 1 r
F P
0 g
s
.19677 .06086 m .19707 .06157 L p .19707 .06159 L .947 .734 .843 r
F P
0 g
s
.19622 .05794 m .19791 .06405 L p .19885 .06477 L 1 .712 .348 r
F P
0 g
s
.19622 .05794 m .19885 .06477 L .1995 .06528 L 1 .712 .348 r
F
.19734 .06227 m .19842 .06508 L .19785 .06278 L 1 .688 .293 r
F
.19745 .06238 m .19622 .05794 L p .19785 .06278 L 1 .723 .389 r
F P
0 g
s
.19631 .05981 m .19621 .05954 L .19724 .06202 L .947 .734 .843 r
F
.19733 .06224 m .19785 .06278 L .19689 .06042 L 1 .688 .293 r
F
.19733 .06225 m .19785 .06278 L .19733 .06224 L F
.19734 .06227 m .19785 .06278 L .19733 .06225 L F
.19631 .05981 m .19724 .06202 L .19733 .06226 L .947 .734 .843 r
F
.19601 .05904 m .19677 .06086 L p .19621 .05954 L F P
0 g
s
.19621 .05954 m .19677 .06086 L .19707 .06159 L .947 .734 .843 r
F
.19631 .05981 m .19744 .06253 L p .19549 .05769 L .902 .713 .863 r
F P
0 g
s
.19729 .06214 m .19733 .06225 L .19732 .06223 L 1 .688 .293 r
F
.19733 .06225 m .19733 .06225 L .19733 .06224 L F
.19728 .06212 m .19394 .05282 L p .19849 .06512 L 1 .755 .489 r
F P
0 g
s
.19842 .06508 m .19845 .06516 L .19884 .06574 L 1 .688 .293 r
F
.19966 .06724 m .20287 .07242 L .19982 .06765 L 1 .723 .371 r
F
.20287 .07242 m .20085 .06977 L p .19903 .06641 L F P
0 g
s
.20304 .07278 m .19963 .06719 L .20519 .0774 L 1 .684 .303 r
F
.19966 .06724 m .19982 .06765 L .19903 .06641 L 1 .723 .371 r
F
.19827 .06498 m .19836 .06512 L p .1984 .06522 L F P
0 g
s
.19963 .06719 m .19827 .06498 L .19963 .06741 L 1 .684 .303 r
F
.19963 .06719 m .19963 .06741 L .20519 .0774 L F
.19846 .06518 m .19904 .06645 L .19864 .06566 L .983 .84 .947 r
F
.19804 .06454 m .19782 .06372 L .19859 .06553 L 1 .723 .389 r
F
.19836 .06518 m .19961 .06679 L .19846 .06519 L 1 .712 .348 r
F
.19846 .06518 m .19864 .06566 L .19839 .06515 L .983 .84 .947 r
F
.19844 .06513 m .19846 .06518 L .19839 .06515 L F
.1989 .06605 m .19904 .06645 L p .19844 .06513 L F P
0 g
s
.19838 .06523 m .19961 .06751 L .19907 .06653 L 1 .711 .353 r
F
.19904 .06645 m .19907 .06654 L p .19874 .06588 L .983 .84 .947 r
F P
0 g
s
.20233 .07155 m .19963 .06717 L p .20304 .07278 L 1 .684 .303 r
F P
0 g
s
.19963 .06717 m .19827 .06498 L p .19963 .06719 L 1 .684 .303 r
F P
0 g
s
.19963 .06717 m .19963 .06719 L .20304 .07278 L 1 .684 .303 r
F
.20559 .07991 m .20503 .08258 L p .20034 .06886 L 1 .866 .918 r
F P
0 g
s
.20503 .08258 m .20494 .08297 L p .20087 .07103 L 1 .866 .918 r
F P
0 g
s
.19989 .06827 m .1985 .06553 L .20518 .08297 L .957 .829 .959 r
F
.19839 .06522 m .19883 .06604 L .19841 .06525 L 1 .712 .348 r
F
.19836 .06518 m .19841 .06528 L .19839 .06522 L F
.19841 .06528 m .19967 .06761 L .19839 .06522 L F
.19836 .06515 m .19837 .06518 L .19838 .06518 L F
.19828 .06503 m .19836 .06518 L .19837 .06518 L F
.19828 .06503 m .19837 .06518 L .19832 .06504 L F
.19808 .06467 m .19827 .06498 L p .19903 .06641 L 1 .723 .371 r
F P
0 g
s
.19828 .06503 m .19833 .06517 L .19836 .06518 L 1 .712 .348 r
F
.20503 .08258 m .19975 .06777 L .20034 .06886 L 1 .866 .918 r
F
.20579 .07896 m .20559 .07991 L p .20034 .06886 L F P
0 g
s
.20087 .07103 m .19975 .06777 L .20503 .08258 L 1 .866 .918 r
F
.19904 .06632 m .19884 .06574 L p .19845 .06516 L 1 .688 .293 r
F P
0 g
s
.19884 .06574 m .19785 .06278 L p .19842 .06508 L 1 .688 .293 r
F P
0 g
s
.19849 .06513 m .19838 .06484 L .19881 .06565 L 1 .755 .489 r
F
.20142 .07092 m .20222 .07229 L p .20004 .0683 L 1 .684 .303 r
F P
0 g
s
.19843 .06537 m .1985 .06553 L .20085 .06977 L 1 .711 .353 r
F
.19907 .06653 m .19904 .06645 L p .19967 .06761 L 1 .866 .918 r
F P
0 g
s
.19911 .06661 m .19962 .06752 L .19967 .06765 L 1 .711 .353 r
F
.19907 .06653 m .19962 .06752 L .19911 .06661 L F
.19907 .06653 m .19961 .06751 L .19962 .06752 L F
.20518 .08297 m .2069 .08297 L p .19989 .06827 L .957 .829 .959 r
F P
0 g
s
.20145 .07134 m .19989 .06827 L .2069 .08297 L .957 .829 .959 r
F
.20133 .07095 m .19981 .06804 L .20738 .08297 L F
.19981 .06804 m .1985 .06553 L .19989 .06827 L F
.19981 .06804 m .19989 .06827 L .20738 .08297 L F
.19941 .06788 m .2001 .06878 L p .1985 .06553 L 1 .711 .353 r
F P
0 g
s
.2001 .06878 m .20049 .06929 L p .19976 .06791 L 1 .711 .353 r
F P
0 g
s
.1985 .06553 m .2001 .06878 L .19976 .06791 L 1 .711 .353 r
F
.20049 .06929 m .20085 .06977 L p .1985 .06553 L F P
0 g
s
.19827 .06498 m .199 .06632 L .20519 .0774 L 1 .684 .303 r
F
.19967 .06761 m .19958 .06737 L .20579 .07872 L 1 .866 .918 r
F
.19958 .06737 m .20294 .07344 L .20579 .07872 L F
.20411 .07555 m .20519 .0774 L p .199 .06632 L 1 .684 .303 r
F P
0 g
s
.19966 .06737 m .20565 .07839 L .20579 .07872 L 1 .866 .918 r
F
.19985 .06772 m .20565 .07839 L .19966 .06737 L F
.19394 .05282 m .19715 .0615 L p .19649 .05875 L 1 .688 .293 r
F P
0 g
s
.19394 .05282 m .1988 .06561 L p .19838 .06484 L 1 .755 .489 r
F P
0 g
s
.19846 .06506 m .19394 .05282 L .19849 .06513 L 1 .755 .489 r
F
.1988 .06562 m .19394 .05282 L p .19747 .06228 L .992 .61 .145 r
F P
0 g
s
.19857 .06512 m .1988 .06562 L .19747 .06228 L .992 .61 .145 r
F
.19858 .06514 m .1989 .06605 L p .19856 .0651 L .983 .84 .947 r
F P
0 g
s
.19857 .06512 m .19858 .06514 L .19857 .06512 L .992 .61 .145 r
F
.19858 .06514 m .1988 .06562 L .19857 .06512 L F
.19815 .06402 m .19833 .06454 L .19686 .06101 L 1 .771 .77 r
F
.19863 .06514 m .19833 .06454 L .19815 .06402 L F
.19863 .06514 m .19858 .06512 L .19833 .06454 L F
.19777 .06305 m .19863 .06514 L p .19807 .06384 L F P
0 g
s
.19609 .05899 m .19777 .06305 L p .19686 .06101 L 1 .771 .77 r
F P
0 g
s
.19864 .0651 m .19878 .06519 L .19967 .06761 L 1 .866 .918 r
F
.19777 .06305 m .19631 .05953 L p .19637 .05948 L .858 .69 .878 r
F P
0 g
s
.19622 .0591 m .19779 .06314 L .19754 .06247 L .992 .61 .145 r
F
.19609 .05873 m .19754 .06247 L .19394 .05282 L F
.19622 .0591 m .19754 .06247 L .19609 .05873 L F
.19863 .06514 m .19777 .06305 L p .1964 .05946 L .858 .69 .878 r
F P
0 g
s
.19777 .06305 m .19637 .05948 L .1964 .05946 L .858 .69 .878 r
F
.19844 .06439 m .19875 .06516 L p .19865 .06505 L 1 .785 .643 r
F P
0 g
s
.19846 .06458 m .19844 .06439 L .19865 .06505 L 1 .785 .643 r
F
.19785 .06278 m .19649 .05875 L p .19689 .06042 L 1 .688 .293 r
F P
0 g
s
.1987 .06502 m .19871 .06505 L .19245 .04931 L .057 .057 .452 r
F
.19868 .06495 m .1987 .06502 L .19245 .04931 L F
.1995 .06528 m .19948 .06511 L .19622 .05794 L p 1 .712 .348 r
F P
0 g
s
.19867 .06494 m .19647 .0594 L .20054 .0561 L .858 .69 .878 r
F
.19245 .04927 m .19708 .06094 L .19245 .04931 L .057 .057 .452 r
F
.21308 .04725 m .2142 .04503 L .1926 .04921 L p .841 1 1 r
F P
0 g
s
.19649 .05875 m .19622 .05794 L .19394 .05282 L p 1 .688 .293 r
F P
0 g
s
.19244 .04924 m .20239 .04731 L p .19645 .05933 L .858 .69 .878 r
F P
0 g
s
.19243 .04922 m .19245 .04927 L .19245 .04927 L .057 .057 .452 r
F
.19243 .04922 m .19245 .04927 L .19244 .04924 L F
.1926 .0497 m .19844 .06439 L p .19846 .06458 L 1 .785 .643 r
F P
0 g
s
.19545 .0571 m .1926 .0497 L p .19846 .06458 L 1 .785 .643 r
F P
0 g
s
.1926 .0497 m .19245 .04931 L p .19871 .06505 L .057 .057 .452 r
F P
0 g
s
.19871 .06505 m .19891 .06554 L .19871 .06505 L .057 .057 .452 r
F
.19871 .06505 m .19871 .06505 L .1926 .0497 L F
.19637 .05948 m .19572 .05777 L .19639 .05944 L .858 .69 .878 r
F
.19647 .0594 m .20239 .04731 L .20054 .0561 L F
.19637 .05948 m .19645 .05933 L .19647 .0594 L F
.19645 .05933 m .20239 .04731 L .19647 .0594 L F
.19708 .06094 m .19868 .06495 L .19245 .04931 L .057 .057 .452 r
F
.19863 .06483 m .19867 .06494 L .19708 .06094 L F
.19867 .06494 m .19868 .06495 L .19708 .06094 L F
.19863 .06514 m .1964 .05946 L .19865 .06505 L .858 .69 .878 r
F
.197 .06111 m .19857 .06502 L .19729 .06177 L .902 .713 .863 r
F
.19209 .04931 m .19245 .04931 L p .19245 .04927 L .841 1 1 r
F P
0 g
s
.19376 .05276 m .19243 .04924 L .19637 .05948 L .858 .69 .878 r
F
.19631 .05953 m .19376 .05276 L .19637 .05948 L F
.19545 .0571 m .19622 .0591 L p .19609 .05873 L .992 .61 .145 r
F P
0 g
s
.19631 .05953 m .19209 .04931 L .19243 .04924 L p .858 .69 .878 r
F P
0 g
s
.19622 .0591 m .19729 .06177 L .19213 .0486 L .902 .713 .863 r
F
.197 .06111 m .19729 .06177 L .19622 .0591 L F
.19847 .06469 m .19545 .0571 L .19846 .06458 L 1 .785 .643 r
F
.19849 .06487 m .19668 .06024 L .19847 .06469 L F
.19668 .06024 m .19545 .0571 L .19847 .06469 L F
.1988 .06561 m .1988 .06562 L p .19838 .06484 L 1 .755 .489 r
F P
0 g
s
.19857 .06512 m .19866 .06515 L .19967 .06761 L 1 .866 .918 r
F
.1988 .06562 m .19881 .06565 L p .19878 .06559 L 1 .755 .489 r
F P
0 g
s
.19838 .06484 m .1988 .06562 L .19878 .06559 L 1 .755 .489 r
F
.19904 .06645 m .199 .06632 L p .19967 .06761 L 1 .866 .918 r
F P
0 g
s
.199 .06632 m .19947 .06717 L .19967 .06761 L 1 .866 .918 r
F
.1985 .06553 m .19799 .06459 L .19837 .0656 L .957 .829 .959 r
F
.19799 .06459 m .19729 .06326 L .20494 .08297 L F
.19272 .05502 m .1919 .05315 L p .19278 .05514 L .983 .84 .947 r
F P
0 g
s
.19185 .05304 m .19272 .05502 L p .19729 .06326 L .957 .829 .959 r
F P
0 g
s
.19729 .06326 m .19727 .06322 L .20494 .08297 L .957 .829 .959 r
F
.19383 .05758 m .20494 .08297 L p .19727 .06322 L F P
0 g
s
.19791 .06405 m .19808 .06469 L p .19811 .06475 L 1 .712 .348 r
F P
0 g
s
.19791 .06405 m .19811 .06475 L .19833 .06517 L 1 .712 .348 r
F
.199 .06632 m .1981 .0647 L .20222 .07229 L 1 .684 .303 r
F
.19807 .06464 m .19804 .06454 L p .19812 .06475 L 1 .723 .389 r
F P
0 g
s
.19804 .06461 m .19652 .06408 L p .19838 .06524 L 1 .711 .353 r
F P
0 g
s
.19799 .06459 m .19804 .06461 L p .19808 .06476 L 1 .711 .353 r
F P
0 g
s
.19799 .06452 m .1973 .06329 L .19845 .06545 L .957 .829 .959 r
F
.19966 .06772 m .1985 .06553 L .19977 .06793 L F
.19838 .06524 m .19832 .0652 L .1985 .06553 L 1 .711 .353 r
F
.19804 .06461 m .19968 .06776 L .19808 .06476 L F
.19851 .06553 m .1985 .06553 L .19966 .06772 L .957 .829 .959 r
F
.1985 .06553 m .1985 .06553 L .19966 .06772 L F
.19967 .06765 m .19968 .06776 L p .19804 .06461 L 1 .711 .353 r
F P
0 g
s
.20937 .09454 m .2077 .07775 L .21197 .07564 L 1 .942 .493 r
F
.20721 .08143 m .2116 .09302 L .20683 .08124 L p .019 .019 .379 r
F P
0 g
s
.20683 .08124 m .2116 .09302 L .21026 .09029 L p .019 .019 .379 r
F P
0 g
s
.20761 .08297 m .20767 .08297 L p .19977 .06793 L .957 .829 .959 r
F P
0 g
s
.20767 .08297 m .20776 .08297 L p .20049 .06929 L .957 .829 .959 r
F P
0 g
s
.20049 .06929 m .19977 .06793 L .20767 .08297 L .957 .829 .959 r
F
.19099 .0504 m .19839 .06515 L .18899 .04651 L .983 .84 .947 r
F
.19762 .06299 m .19815 .06437 L p .19782 .06372 L 1 .723 .389 r
F P
0 g
s
.19733 .06227 m .19844 .06515 L .19734 .06227 L 1 .688 .293 r
F
.19844 .06515 m .19845 .06515 L .19734 .06227 L F
.19815 .06437 m .19859 .06553 L p .19782 .06372 L 1 .723 .389 r
F P
0 g
s
.19857 .06548 m .19858 .06549 L .19845 .06516 L 1 .688 .293 r
F
.19845 .06515 m .19845 .06516 L .19734 .06227 L F
.19857 .06548 m .19845 .06516 L .19845 .06515 L F
.19852 .06514 m .19665 .06062 L p .19853 .06513 L .947 .734 .843 r
F P
0 g
s
.19977 .06793 m .19976 .06791 L .20129 .07088 L .957 .829 .959 r
F
.19976 .06791 m .19862 .06576 L .1998 .06803 L F
.19976 .06791 m .1998 .06803 L .20129 .07088 L F
.1989 .06552 m .19871 .06505 L .1987 .06502 L .057 .057 .452 r
F
.1998 .06804 m .19981 .06804 L .20133 .07095 L .957 .829 .959 r
F
.1998 .06803 m .20133 .07095 L .20129 .07088 L F
.19862 .06576 m .1998 .06804 L .1998 .06803 L F
.1998 .06804 m .20133 .07095 L .1998 .06803 L F
.19875 .06516 m .19865 .06483 L .19878 .06519 L 1 .866 .918 r
F
.19891 .06554 m .19893 .0656 L .1926 .0497 L .057 .057 .452 r
F
.19865 .06505 m .19647 .0594 L .19867 .06494 L .858 .69 .878 r
F
.19859 .06498 m .19864 .06485 L .19875 .06516 L 1 .866 .918 r
F
.19864 .06485 m .19865 .06483 L .19875 .06516 L F
.19977 .06793 m .20129 .07088 L .20086 .07002 L .957 .829 .959 r
F
.19932 .06768 m .20046 .06923 L .19938 .06784 L 1 .723 .371 r
F
.20049 .06929 m .19938 .06784 L p .20046 .06923 L F P
0 g
s
.19862 .06576 m .19862 .06576 L .19981 .06804 L .957 .829 .959 r
F
.19862 .06576 m .19981 .06804 L .1998 .06804 L F
.19862 .06576 m .1985 .06553 L .19862 .06576 L F
.19968 .06776 m .19971 .06817 L p .19808 .06476 L 1 .711 .353 r
F P
0 g
s
.19974 .06785 m .19977 .06793 L .20049 .06929 L .957 .829 .959 r
F
.19665 .06062 m .19867 .06545 L .19867 .06545 L .947 .734 .843 r
F
.19858 .06549 m .19858 .06549 L .19845 .06516 L 1 .688 .293 r
F
.19845 .06516 m .19845 .06516 L .19734 .06227 L F
.19858 .06549 m .19845 .06516 L .19845 .06516 L F
.19665 .06062 m .19662 .06055 L .1985 .06504 L .947 .734 .843 r
F
.19665 .06062 m .1985 .06504 L .1985 .06504 L F
.19968 .06776 m .19967 .06761 L p .19833 .06517 L 1 .712 .348 r
F P
0 g
s
.19852 .06552 m .19976 .06791 L .19974 .06785 L .957 .829 .959 r
F
.19848 .06545 m .19852 .06552 L .19852 .06552 L F
.19852 .06552 m .19976 .06791 L .19852 .06552 L F
.19856 .0651 m .19881 .06581 L .19847 .06515 L .983 .84 .947 r
F
.19851 .06511 m .19847 .06515 L .19845 .06513 L F
.19856 .0651 m .19847 .06515 L .19851 .06511 L F
.19848 .06545 m .19976 .06791 L .19848 .06545 L .957 .829 .959 r
F
.19976 .06791 m .19977 .06793 L .19974 .06785 L F
.19929 .06691 m .20061 .06933 L .19995 .06848 L 1 .723 .371 r
F
.19847 .06545 m .19845 .06545 L .1985 .06553 L .957 .829 .959 r
F
.19847 .06545 m .1985 .06553 L .19851 .06553 L F
.19833 .06517 m .19843 .06543 L .19968 .06776 L 1 .712 .348 r
F
.20494 .08297 m .20503 .08258 L p .20505 .0827 L 1 .942 .493 r
F P
0 g
s
.20494 .08297 m .20518 .08297 L p .1985 .06553 L .957 .829 .959 r
F P
0 g
s
.1985 .06553 m .19837 .0656 L .20494 .08297 L .957 .829 .959 r
F
.19652 .06408 m .19941 .06788 L p .1985 .06553 L 1 .711 .353 r
F P
0 g
s
.19832 .0652 m .19652 .06408 L .1985 .06553 L 1 .711 .353 r
F
.19808 .06476 m .19971 .06817 L .19866 .06682 L F
.19931 .06732 m .19971 .06817 L .19932 .06768 L 1 .723 .371 r
F
.19971 .06817 m .19968 .06776 L p .19843 .06543 L 1 .712 .348 r
F P
0 g
s
.19833 .06517 m .19967 .06761 L .19836 .06518 L 1 .712 .348 r
F
.19842 .06531 m .19929 .06691 L .19932 .06768 L 1 .723 .371 r
F
.19929 .06691 m .19971 .06817 L .19931 .06732 L F
.19848 .06545 m .19974 .06785 L .1985 .06545 L .957 .829 .959 r
F
.19974 .06785 m .20049 .06929 L .19972 .0678 L F
.19974 .06785 m .19972 .0678 L .1985 .06545 L F
.19893 .0656 m .19875 .06516 L .19873 .06511 L .057 .057 .452 r
F
.20061 .06933 m .20083 .06973 L .19932 .06768 L 1 .723 .371 r
F
.19976 .06785 m .20049 .06929 L .20474 .07729 L .957 .829 .959 r
F
.19865 .06505 m .19878 .06519 L .19864 .0651 L 1 .866 .918 r
F
.19879 .06554 m .19863 .06514 L p .19867 .06509 L .858 .69 .878 r
F P
0 g
s
.19879 .06554 m .19867 .06509 L .19876 .06531 L .858 .69 .878 r
F
.1985 .06545 m .19972 .0678 L .19967 .06768 L .957 .829 .959 r
F
.19972 .0678 m .20046 .06924 L .19967 .06768 L F
.19843 .06543 m .19943 .06809 L .19971 .06817 L 1 .712 .348 r
F
.19811 .06475 m .19848 .06545 L .19848 .06545 L .957 .829 .959 r
F
.19799 .06452 m .19845 .06545 L .19847 .06545 L F
.19833 .06517 m .19811 .06475 L .19843 .06543 L 1 .712 .348 r
F
.19808 .06467 m .1984 .06537 L .19809 .0647 L 1 .723 .371 r
F
.19818 .06487 m .19841 .06532 L .1984 .06537 L F
.19808 .06467 m .19841 .06532 L .19818 .06487 L F
.19102 .05114 m .18899 .04651 L p .19874 .06588 L .983 .84 .947 r
F P
0 g
s
.19827 .06498 m .19808 .06467 L p .199 .06632 L 1 .684 .303 r
F P
0 g
s
.19807 .06461 m .19837 .06521 L .19808 .06464 L 1 .723 .389 r
F
.19809 .0647 m .19848 .06545 L .19811 .06475 L .957 .829 .959 r
F
.19848 .06545 m .19977 .06793 L .19848 .06545 L F
.19848 .06545 m .19848 .06545 L .19811 .06475 L F
.19804 .06395 m .19867 .06545 L .19804 .06395 L .947 .734 .843 r
F
.19772 .06402 m .19799 .06452 L .19774 .06404 L .957 .829 .959 r
F
.19854 .0651 m .19869 .06545 L p .19853 .06511 L .947 .734 .843 r
F P
0 g
s
.19847 .06545 m .19977 .06793 L .19848 .06545 L .957 .829 .959 r
F
.19806 .06465 m .19848 .06545 L .19809 .0647 L F
.19848 .06545 m .19977 .06793 L .19848 .06545 L F
.19848 .06545 m .19848 .06545 L .19809 .0647 L F
.19852 .06514 m .19853 .06513 L .19867 .06545 L .947 .734 .843 r
F
.19967 .06768 m .20046 .06924 L .19976 .06785 L .957 .829 .959 r
F
.20046 .06924 m .20049 .06929 L .19976 .06785 L F
.20083 .06973 m .20049 .06929 L p .20046 .06923 L 1 .723 .371 r
F P
0 g
s
.20585 .07867 m .20721 .08143 L p .20664 .08075 L .019 .019 .379 r
F P
0 g
s
.2116 .09302 m .20721 .08143 L .21242 .09454 L closepath p .026 .026 .427 r
F P
0 g
s
.20804 .08297 m .21242 .09454 L .20721 .08143 L closepath p .026 .026 .427 r
F P
0 g
s
.20776 .08297 m .20788 .08297 L p .20142 .07092 L .957 .829 .959 r
F P
0 g
s
.20142 .07092 m .19935 .06705 L .20354 .07498 L .957 .829 .959 r
F
.20142 .07092 m .20354 .07498 L .20776 .08297 L F
.19933 .06694 m .20222 .07229 L .1993 .06687 L 1 .684 .303 r
F
.2053 .07795 m .20085 .06977 L .2049 .07711 L 1 .866 .918 r
F
.20222 .07229 m .20142 .07092 L p .19975 .0676 L 1 .757 .441 r
F P
0 g
s
.19971 .06817 m .19995 .07147 L p .19866 .06682 L 1 .711 .353 r
F P
0 g
s
.19962 .06757 m .19962 .06777 L .20142 .07092 L 1 .684 .303 r
F
.19929 .06691 m .19995 .06848 L .19971 .06817 L 1 .723 .371 r
F
.19913 .06697 m .20142 .07092 L p .19897 .06663 L 1 .684 .303 r
F P
0 g
s
.1997 .06772 m .20474 .07729 L .20354 .07498 L .957 .829 .959 r
F
.1985 .06545 m .19976 .06785 L .1997 .06772 L F
.19976 .06785 m .20474 .07729 L .1997 .06772 L F
.19845 .06516 m .19858 .06549 L .19904 .06632 L 1 .688 .293 r
F
.19849 .06545 m .19897 .06663 L .19962 .06777 L 1 .684 .303 r
F
.19799 .06452 m .19847 .06545 L .19806 .06465 L .957 .829 .959 r
F
.19841 .06527 m .19877 .06621 L .1984 .06541 L 1 .723 .371 r
F
.19877 .06621 m .19938 .06784 L .1984 .06541 L F
.19938 .06784 m .1983 .06641 L p .1983 .06636 L F P
0 g
s
.19932 .06767 m .19938 .06784 L .1983 .06636 L 1 .723 .371 r
F
.19799 .06459 m .19808 .06476 L .19866 .06682 L 1 .711 .353 r
F
.19861 .06676 m .19866 .06682 L .19995 .07147 L F
.19799 .06459 m .19866 .06682 L .19861 .06676 L F
.19809 .0647 m .1984 .06537 L .1983 .06636 L 1 .723 .371 r
F
.19829 .06635 m .1983 .06636 L .1983 .06641 L F
.19652 .06408 m .19799 .06459 L p .19861 .06676 L 1 .711 .353 r
F P
0 g
s
.19809 .0647 m .1983 .06636 L .19829 .06635 L 1 .723 .371 r
F
.1984 .06541 m .19932 .06767 L .1983 .06636 L F
.19995 .07147 m .19652 .06408 L p .19861 .06676 L 1 .711 .353 r
F P
0 g
s
.19811 .06475 m .19857 .06643 L .19843 .06543 L 1 .712 .348 r
F
.19857 .06643 m .19995 .07147 L .19878 .06684 L F
.19857 .06643 m .19878 .06684 L .19843 .06543 L F
.19813 .06478 m .19816 .06485 L .19882 .06667 L 1 .723 .389 r
F
.19995 .07147 m .19977 .06897 L p .19843 .06543 L 1 .712 .348 r
F P
0 g
s
.19816 .06485 m .19913 .06697 L .19882 .06667 L 1 .723 .389 r
F
.19799 .06452 m .19908 .06688 L .19897 .06663 L 1 .684 .303 r
F
.19904 .06632 m .19912 .06656 L p .19807 .06461 L 1 .723 .389 r
F P
0 g
s
.19799 .06452 m .19935 .06705 L .19866 .06575 L 1 .684 .303 r
F
.19935 .06705 m .19962 .06757 L .19866 .06575 L F
.19866 .06575 m .19962 .06757 L .19962 .06752 L F
.19807 .06464 m .19812 .06475 L .19813 .06478 L 1 .723 .389 r
F
.19807 .06464 m .19813 .06478 L .1981 .06474 L F
.19808 .06467 m .19841 .06527 L .19841 .06532 L 1 .723 .371 r
F
.19808 .06469 m .19809 .06471 L p .19811 .06475 L 1 .712 .348 r
F P
0 g
s
.19809 .06471 m .19807 .06464 L p .1981 .06474 L 1 .723 .389 r
F P
0 g
s
.19804 .06461 m .19802 .06457 L .19962 .06752 L 1 .684 .303 r
F
.19962 .06752 m .19962 .06752 L .20222 .07229 L F
.19804 .06461 m .19962 .06752 L .19962 .06752 L F
.19808 .06467 m .19804 .06461 L p .199 .06632 L F P
0 g
s
.19807 .06461 m .19912 .06656 L .19837 .06521 L 1 .723 .389 r
F
.19804 .06454 m .19807 .06461 L .19808 .06464 L F
.19908 .06649 m .19785 .06418 L .1985 .06545 L .957 .829 .959 r
F
.19785 .06418 m .1957 .06015 L .19771 .06396 L F
.19785 .06418 m .19771 .06396 L .1985 .06545 L F
.19867 .06545 m .19867 .06545 L .19804 .06395 L .947 .734 .843 r
F
.19804 .06395 m .19867 .06545 L .19867 .06545 L F
.19707 .06157 m .19854 .0651 L p .19718 .06186 L F P
0 g
s
.19851 .06511 m .19854 .06508 L .19672 .06067 L .902 .713 .863 r
F
.19851 .06511 m .19849 .06512 L .19855 .06507 L .983 .84 .947 r
F
.19849 .06512 m .19839 .06515 L .19848 .0651 L F
.199 .06632 m .19868 .06543 L p .19947 .06717 L 1 .866 .918 r
F P
0 g
s
.19881 .06565 m .19894 .06597 L p .19849 .06517 L 1 .755 .489 r
F P
0 g
s
.19908 .06649 m .1985 .06545 L .19935 .06705 L .957 .829 .959 r
F
.19849 .06545 m .19962 .06777 L .19962 .06757 L 1 .684 .303 r
F
.19912 .06656 m .19932 .06717 L p .19837 .06521 L 1 .723 .389 r
F P
0 g
s
.19841 .06527 m .20083 .06973 L .19842 .06531 L 1 .723 .371 r
F
.1981 .0647 m .1981 .06472 L .20222 .07229 L 1 .684 .303 r
F
.19962 .06752 m .2001 .06846 L .20004 .0683 L F
.19962 .06757 m .2001 .06846 L .19962 .06752 L F
.20061 .06936 m .19908 .06649 L .19935 .06705 L .957 .829 .959 r
F
.19858 .06549 m .1991 .06653 L .19904 .06632 L 1 .688 .293 r
F
.19812 .06414 m .19791 .06363 L .19868 .06545 L .947 .734 .843 r
F
.19867 .06564 m .19857 .06545 L .20061 .06936 L .957 .829 .959 r
F
.20085 .06977 m .20083 .06973 L p .19841 .06527 L 1 .723 .371 r
F P
0 g
s
.19857 .06545 m .19855 .06545 L .20061 .06936 L .957 .829 .959 r
F
.19804 .06454 m .19807 .0646 L .19807 .06461 L 1 .723 .389 r
F
.19812 .06464 m .19785 .06418 L .19841 .06523 L .957 .829 .959 r
F
.199 .06632 m .19933 .06694 L .1993 .06687 L 1 .684 .303 r
F
.19849 .06534 m .19846 .06534 L .20061 .06936 L .957 .829 .959 r
F
.19844 .06526 m .19842 .06525 L .18899 .04651 L .983 .84 .947 r
F
.19904 .06645 m .19842 .06525 L .19844 .06526 L F
.19904 .06645 m .19874 .06588 L .19842 .06525 L F
.19846 .06527 m .19841 .06523 L .19846 .06534 L .957 .829 .959 r
F
.19846 .06527 m .19846 .06534 L .19849 .06534 L F
.19857 .06545 m .19812 .06464 L .19855 .06545 L F
.19858 .06549 m .19871 .06584 L .1991 .06653 L 1 .688 .293 r
F
.19847 .06545 m .19848 .06545 L .19806 .06465 L .957 .829 .959 r
F
.19815 .06481 m .19897 .06663 L .19849 .06545 L 1 .684 .303 r
F
.19816 .06485 m .19932 .06717 L .19913 .06697 L 1 .723 .389 r
F
.19922 .06695 m .19908 .06688 L .1989 .06626 L 1 .757 .441 r
F
.19913 .06697 m .19908 .06688 L p .19961 .06745 L F P
0 g
s
.19871 .06584 m .19923 .06718 L .19926 .06703 L 1 .688 .293 r
F
.2001 .06846 m .20142 .07092 L .20004 .0683 L 1 .684 .303 r
F
.20142 .07092 m .19913 .06697 L p .19962 .06747 L 1 .757 .441 r
F P
0 g
s
.19977 .06897 m .19975 .06869 L p .19943 .06809 L 1 .712 .348 r
F P
0 g
s
.19975 .06869 m .19971 .06817 L p .19943 .06809 L 1 .712 .348 r
F P
0 g
s
.19932 .06717 m .20267 .07715 L p .19882 .06667 L 1 .723 .389 r
F P
0 g
s
.19923 .06718 m .19962 .06819 L .19944 .06762 L 1 .688 .293 r
F
.19923 .06718 m .19944 .06762 L .19926 .06703 L F
.19844 .06515 m .19857 .06548 L .19845 .06515 L F
.19733 .06227 m .19848 .06529 L .19857 .06548 L F
.19744 .06253 m .19842 .06508 L .19734 .06227 L F
.19734 .06227 m .19744 .06253 L .19734 .06227 L F
.19744 .06253 m .1977 .06316 L p .19729 .06214 L .902 .713 .863 r
F P
0 g
s
.19728 .06212 m .19729 .06214 L p .19732 .06223 L 1 .688 .293 r
F P
0 g
s
.19744 .06253 m .19729 .06214 L .19549 .05769 L .902 .713 .863 r
F
.19729 .06214 m .19733 .06227 L p .19733 .06225 L 1 .688 .293 r
F P
0 g
s
.19733 .06226 m .19728 .06212 L .1974 .06242 L .947 .734 .843 r
F
.19733 .06227 m .19734 .06227 L .19733 .06225 L 1 .688 .293 r
F
.19843 .06512 m .19839 .06515 L .19099 .0504 L .983 .84 .947 r
F
.19733 .06227 m .19826 .06484 L p .19848 .06529 L 1 .688 .293 r
F P
0 g
s
.19807 .0646 m .19904 .06632 L .19807 .06461 L 1 .723 .389 r
F
.19812 .06464 m .19841 .06523 L .19846 .06527 L .957 .829 .959 r
F
.19808 .06464 m .19829 .06507 L .19813 .06478 L 1 .723 .389 r
F
.19837 .06521 m .19829 .06507 L .19808 .06464 L F
.1983 .06641 m .19652 .06408 L p .19808 .06467 L 1 .723 .371 r
F P
0 g
s
.1981 .0647 m .19804 .06461 L .1981 .06472 L 1 .684 .303 r
F
.19652 .06408 m .19322 .0568 L .19808 .06467 L p 1 .723 .371 r
F P
0 g
s
.19804 .06461 m .19799 .06452 L p .19802 .06457 L 1 .684 .303 r
F P
0 g
s
.1957 .06015 m .19848 .06545 L .1985 .06545 L .957 .829 .959 r
F
.19815 .06481 m .19816 .06485 L .19813 .06478 L 1 .723 .389 r
F
.19799 .06452 m .19806 .06465 L .19774 .06404 L .957 .829 .959 r
F
.19774 .06404 m .19806 .06465 L .19809 .0647 L F
.1981 .06474 m .19813 .06478 L .20267 .07715 L 1 .723 .389 r
F
.1957 .06015 m .1973 .06329 L .19799 .06452 L .957 .829 .959 r
F
.19809 .06471 m .19855 .0664 L p .19811 .06475 L 1 .712 .348 r
F P
0 g
s
.19855 .0664 m .19995 .07147 L p .19857 .06643 L 1 .712 .348 r
F P
0 g
s
.19855 .0664 m .19857 .06643 L .19811 .06475 L 1 .712 .348 r
F
.19817 .0647 m .19853 .06543 L .19546 .06044 L 1 .757 .441 r
F
.19523 .06001 m .19546 .06044 L .19322 .0568 L F
.19817 .0647 m .19546 .06044 L .19523 .06001 L F
.19823 .06478 m .1983 .06493 L .19817 .0647 L F
.19867 .06564 m .19102 .05114 L .19609 .06088 L .957 .829 .959 r
F
.1983 .06493 m .19853 .06543 L .19817 .0647 L 1 .757 .441 r
F
.19826 .06484 m .19827 .06487 L p .1983 .06493 L 1 .688 .293 r
F P
0 g
s
.19826 .06484 m .1983 .06493 L .19848 .06529 L 1 .688 .293 r
F
.19823 .06478 m .19913 .06642 L p .19853 .06543 L 1 .757 .441 r
F P
0 g
s
.19839 .065 m .19836 .06498 L .198 .06427 L .983 .84 .947 r
F
.19839 .065 m .198 .06427 L .19099 .0504 L F
.19817 .06461 m .19728 .06212 L p .19849 .06512 L 1 .755 .489 r
F P
0 g
s
.19852 .06509 m .19839 .065 L .19099 .0504 L .983 .84 .947 r
F
.19865 .06534 m .19854 .0651 L .19859 .06517 L 1 .771 .77 r
F
.19812 .06414 m .19868 .06545 L .19867 .06545 L .947 .734 .843 r
F
.1974 .06242 m .19791 .06363 L .19812 .06414 L F
.1974 .06242 m .19724 .06202 L .19791 .06363 L F
.19665 .06062 m .19804 .06395 L .19804 .06395 L F
.19865 .06545 m .19861 .06537 L .19865 .06545 L F
.19843 .06503 m .19881 .06565 L .19849 .06517 L 1 .755 .489 r
F
.19865 .06545 m .19852 .06514 L p .19861 .06537 L .947 .734 .843 r
F P
0 g
s
.19621 .05954 m .19854 .0651 L .19853 .06511 L .947 .734 .843 r
F
.1977 .06316 m .19865 .06545 L p .19672 .06067 L .902 .713 .863 r
F P
0 g
s
.19868 .06543 m .19857 .06512 L p .19876 .06556 L 1 .866 .918 r
F P
0 g
s
.19858 .06512 m .19601 .05904 L .19585 .0586 L 1 .771 .77 r
F
.19857 .06512 m .19857 .06512 L .19876 .06556 L 1 .866 .918 r
F
.19868 .06536 m .19872 .06547 L .19878 .06557 L .992 .61 .145 r
F
.19857 .06513 m .19872 .06547 L .19868 .06536 L F
.19857 .06513 m .19864 .0653 L .19872 .06547 L F
.20222 .07229 m .20411 .07555 L p .1993 .06687 L 1 .684 .303 r
F P
0 g
s
.19958 .06737 m .19941 .06691 L .20294 .07344 L 1 .866 .918 r
F
.19913 .06642 m .19985 .06772 L p .20222 .07229 L 1 .757 .441 r
F P
0 g
s
.19881 .06581 m .1989 .06605 L .19845 .06513 L .983 .84 .947 r
F
.20267 .07715 m .19995 .07147 L .19809 .06471 L p 1 .723 .389 r
F P
0 g
s
.19799 .06452 m .19767 .064 L p .19908 .06688 L 1 .684 .303 r
F P
0 g
s
.19808 .06467 m .1983 .06641 L .19808 .06467 L 1 .723 .371 r
F
.19767 .064 m .19322 .0568 L .19837 .06567 L p 1 .684 .303 r
F P
0 g
s
.19837 .06567 m .19908 .06688 L p .19767 .064 L 1 .684 .303 r
F P
0 g
s
.19913 .06642 m .19855 .06548 L .19853 .06543 L 1 .757 .441 r
F
.19934 .0668 m .19826 .06483 L p .19945 .06705 L 1 .599 .157 r
F P
0 g
s
.19826 .06483 m .19829 .06491 L .19945 .06705 L 1 .599 .157 r
F
.20411 .07555 m .20222 .07229 L p .19985 .06772 L 1 .757 .441 r
F P
0 g
s
.19934 .0668 m .19945 .06705 L .20573 .07872 L 1 .599 .157 r
F
.2026 .07291 m .19859 .06538 L .19865 .06558 L .957 .829 .959 r
F
.19551 .05959 m .19102 .05114 L .19867 .06564 L F
.19812 .06464 m .1957 .06015 L .19785 .06418 L F
.1957 .06015 m .19772 .06402 L .19774 .06404 L F
.1957 .06015 m .19182 .05297 L .1973 .06329 L F
.19272 .05502 m .19383 .05758 L p .19727 .06322 L F P
0 g
s
.19322 .0568 m .19086 .05144 L .19823 .06478 L p 1 .757 .441 r
F P
0 g
s
.19826 .06483 m .19806 .06448 L p .19829 .06491 L 1 .599 .157 r
F P
0 g
s
.19845 .06513 m .1989 .06605 L .19844 .06513 L .983 .84 .947 r
F
.19817 .06461 m .19894 .06597 L .19829 .06491 L 1 .755 .489 r
F
.19827 .06487 m .19817 .06461 L p .19829 .06491 L F P
0 g
s
.19859 .06538 m .19551 .05959 L .19849 .0653 L .957 .829 .959 r
F
.19829 .06491 m .1983 .06493 L .19831 .06495 L 1 .755 .489 r
F
.19913 .06642 m .19975 .0676 L .19879 .06601 L 1 .757 .441 r
F
.20043 .06891 m .19865 .06558 L .19867 .06564 L .957 .829 .959 r
F
.19865 .06558 m .19849 .0653 L .19867 .06564 L F
.1985 .06529 m .19858 .06544 L .19857 .06543 L 1 .599 .157 r
F
.19806 .06448 m .19814 .06462 L .1983 .06492 L F
.19858 .06545 m .19859 .06545 L .19945 .06705 L F
.19829 .06491 m .19859 .06545 L .19858 .06545 L F
.19851 .06531 m .19849 .0653 L .19865 .06558 L .957 .829 .959 r
F
.19853 .06537 m .19855 .06539 L .19851 .06532 L 1 .599 .157 r
F
.19851 .06531 m .19851 .06532 L .1985 .06529 L F
.19853 .06537 m .19851 .06532 L .19851 .06531 L F
.19855 .06539 m .19851 .06532 L .1985 .06529 L F
.19805 .06445 m .19853 .06537 L .1985 .06529 L F
.19871 .06584 m .19858 .06549 L .19858 .06549 L 1 .688 .293 r
F
.19871 .06584 m .19858 .06549 L .19871 .06584 L F
.19913 .06642 m .19879 .06601 L .19855 .06548 L 1 .757 .441 r
F
.19829 .06507 m .19816 .06485 L .19815 .06481 L 1 .723 .389 r
F
.19809 .06471 m .1981 .06474 L .20267 .07715 L F
.19855 .06548 m .19568 .06086 L .19853 .06543 L 1 .757 .441 r
F
.19853 .06543 m .19568 .06086 L .19546 .06044 L F
.19855 .06548 m .19879 .06601 L .19568 .06086 L F
.19829 .06507 m .19932 .06717 L .19816 .06485 L 1 .723 .389 r
F
.19908 .06688 m .19651 .06247 L p .19879 .06607 L 1 .757 .441 r
F P
0 g
s
.19799 .06452 m .19897 .06663 L .19815 .06481 L 1 .684 .303 r
F
.19879 .06601 m .19651 .06247 L .19568 .06086 L 1 .757 .441 r
F
.19975 .0676 m .1989 .06626 L .19879 .06601 L F
.19837 .06521 m .19932 .06717 L .19829 .06507 L 1 .723 .389 r
F
.1989 .06626 m .19651 .06247 L .19879 .06601 L 1 .757 .441 r
F
.19962 .06863 m .19962 .06851 L .19853 .06544 L 1 .688 .293 r
F
.19852 .06543 m .19853 .06544 L .19848 .06529 L F
.19962 .06863 m .19853 .06544 L .19852 .06543 L F
.19827 .06487 m .19871 .0661 L p .19852 .06543 L F P
0 g
s
.19871 .0661 m .19962 .06863 L p .19886 .06642 L 1 .688 .293 r
F P
0 g
s
.19871 .0661 m .19886 .06642 L .19852 .06543 L 1 .688 .293 r
F
.19847 .06537 m .19849 .06539 L .19878 .06615 L 1 .755 .489 r
F
.19829 .06491 m .19849 .06539 L .19847 .06537 L F
.19829 .06491 m .19831 .06495 L .19849 .06539 L F
.19829 .06491 m .1983 .06492 L .19859 .06545 L 1 .599 .157 r
F
.19806 .06448 m .1983 .06492 L .19829 .06491 L F
.1983 .06493 m .19852 .06543 L .19848 .06529 L 1 .688 .293 r
F
.19827 .06487 m .19852 .06543 L .1983 .06493 L F
.1983 .06493 m .19894 .06597 L .19831 .06495 L 1 .755 .489 r
F
.19806 .06448 m .19805 .06445 L p .19814 .06462 L 1 .599 .157 r
F P
0 g
s
.19985 .06772 m .20547 .07789 L .20411 .07555 L p 1 .757 .441 r
F P
0 g
s
.20547 .07789 m .19985 .06772 L p .20553 .07808 L 1 .599 .157 r
F P
0 g
s
.19941 .06691 m .19868 .06491 L .20565 .07839 L 1 .866 .918 r
F
.19868 .06491 m .19871 .06474 L .20565 .07839 L F
.19871 .06474 m .19908 .063 L .19945 .06386 L F
.20287 .07242 m .20233 .07155 L p .20304 .07278 L 1 .684 .303 r
F P
0 g
s
.20519 .0774 m .20547 .07789 L .20287 .07242 L p 1 .684 .303 r
F P
0 g
s
.19871 .06474 m .19945 .06386 L .20565 .07839 L 1 .866 .918 r
F
.19137 .04495 m .21201 .04937 L .20585 .07867 L p F P
0 g
s
.20585 .07867 m .20582 .07879 L p .20549 .07802 L 1 .866 .918 r
F P
0 g
s
.20585 .07867 m .20547 .07789 L p .20576 .0785 L 1 .599 .157 r
F P
0 g
s
.20549 .07802 m .19137 .04495 L .20585 .07867 L 1 .866 .918 r
F
.20576 .0785 m .20584 .07867 L .20585 .07867 L 1 .599 .157 r
F
.19887 .06544 m .19871 .06504 L .19879 .06494 L .858 .69 .878 r
F
.19888 .06543 m .19879 .06494 L .21063 .05062 L F
.19887 .06544 m .19879 .06494 L .19888 .06543 L F
.19857 .06512 m .19601 .05904 L .19858 .06512 L 1 .771 .77 r
F
.19857 .06512 m .19677 .06086 L .19667 .06061 L F
.19856 .0651 m .19856 .0651 L .19967 .06761 L 1 .866 .918 r
F
.19857 .06512 m .19878 .06557 L .19858 .06514 L .992 .61 .145 r
F
.19856 .06512 m .19857 .06512 L .19857 .06512 L 1 .771 .77 r
F
.1985 .06499 m .19852 .06509 L .19099 .0504 L .983 .84 .947 r
F
.19854 .0651 m .19677 .06086 L p .19856 .06512 L 1 .771 .77 r
F P
0 g
s
.19859 .06517 m .19854 .0651 L .19856 .06512 L 1 .771 .77 r
F
.19869 .06546 m .19854 .0651 L .19865 .06534 L F
.19868 .06543 m .19876 .06556 L .19967 .06761 L 1 .866 .918 r
F
.1987 .06546 m .19894 .06597 L .19864 .0653 L .992 .61 .145 r
F
.1987 .06546 m .19865 .06534 L .19871 .06547 L 1 .771 .77 r
F
.19869 .06546 m .19865 .06534 L .1987 .06546 L F
.19894 .06597 m .1988 .06562 L p .19864 .0653 L .992 .61 .145 r
F P
0 g
s
.19869 .06545 m .19854 .0651 L p .19869 .06546 L 1 .771 .77 r
F P
0 g
s
.19853 .06511 m .19869 .06545 L .19868 .06545 L .947 .734 .843 r
F
.19865 .06545 m .19868 .06543 L .19854 .06508 L .902 .713 .863 r
F
.1986 .06524 m .19868 .06543 L .19861 .06525 L F
.20553 .07808 m .20453 .07625 L .2056 .07831 L 1 .599 .157 r
F
.20453 .07625 m .19934 .0668 L .2056 .07831 L F
.20582 .07879 m .20581 .07883 L p .2049 .07711 L 1 .866 .918 r
F P
0 g
s
.2056 .07831 m .19934 .0668 L .20573 .07872 L 1 .599 .157 r
F
.19928 .0665 m .19869 .06539 L p .19889 .06577 L .016 .016 .392 r
F P
0 g
s
.19985 .06772 m .19934 .0668 L p .20553 .07808 L 1 .599 .157 r
F P
0 g
s
.20685 .0807 m .20585 .07867 L p .20642 .07988 L 1 .599 .157 r
F P
0 g
s
.20642 .07988 m .20679 .08066 L .20685 .0807 L 1 .599 .157 r
F
.20679 .08066 m .20721 .08143 L .20685 .0807 L p F P
0 g
s
.20573 .07872 m .20679 .08066 L p .20584 .07867 L 1 .599 .157 r
F P
0 g
s
.20547 .07789 m .20573 .07872 L .20584 .07867 L 1 .599 .157 r
F
.20804 .08297 m .19908 .06583 L p .19878 .06555 L .957 .829 .959 r
F P
0 g
s
.20804 .08297 m .19878 .06555 L .19878 .06556 L .957 .829 .959 r
F
.19928 .0665 m .19889 .06577 L .19941 .06675 L .016 .016 .392 r
F
.20575 .07869 m .19928 .0665 L p .20572 .07861 L F P
0 g
s
.19872 .06548 m .19857 .06512 L .19872 .06548 L 1 .771 .77 r
F
.19777 .06305 m .19807 .06384 L .19686 .06101 L F
.1987 .06548 m .19768 .06286 L .19565 .05789 L .902 .713 .863 r
F
.19863 .06514 m .19866 .0652 L p .19858 .06512 L 1 .771 .77 r
F P
0 g
s
.19872 .06547 m .20575 .07869 L .20267 .07305 L 1 .811 .573 r
F
.19872 .06546 m .20575 .07869 L .19872 .06547 L F
.19869 .06539 m .20575 .07869 L p .19869 .06541 L F P
0 g
s
.20028 .06846 m .19874 .06552 L .19867 .06544 L .957 .829 .959 r
F
.19872 .06548 m .19857 .06512 L .19873 .06549 L 1 .771 .77 r
F
.19869 .06541 m .19872 .06546 L .19872 .06547 L 1 .811 .573 r
F
.19865 .06505 m .19894 .06565 L .19883 .06549 L 1 .785 .643 r
F
.19894 .06564 m .19875 .06516 L .19893 .0656 L .057 .057 .452 r
F
.19894 .06565 m .19875 .06516 L p .19894 .06564 L F P
0 g
s
.1989 .06554 m .19871 .06505 L .1989 .06552 L .057 .057 .452 r
F
.1989 .06554 m .19891 .06554 L .19871 .06505 L F
.19858 .06514 m .19878 .06557 L .1988 .06562 L .992 .61 .145 r
F
.19902 .06572 m .19884 .06535 L p .19867 .06544 L .957 .829 .959 r
F P
0 g
s
.19868 .06543 m .1987 .06548 L .19868 .06543 L .902 .713 .863 r
F
.19868 .06543 m .19868 .06543 L .19861 .06525 L F
.19853 .0651 m .1987 .06546 L .19864 .0653 L .992 .61 .145 r
F
.19871 .06547 m .19865 .06534 L .19859 .06517 L 1 .771 .77 r
F
.19851 .06505 m .19872 .06547 L .19851 .06506 L 1 .811 .573 r
F
.19872 .06547 m .20267 .07305 L .19873 .06548 L F
.19874 .06554 m .19871 .06547 L .19873 .06552 L 1 .771 .77 r
F
.19871 .06547 m .19856 .06512 L .19872 .06548 L F
.19908 .06583 m .19896 .06567 L .19867 .06544 L .957 .829 .959 r
F
.19871 .06547 m .19872 .06548 L .19873 .06552 L 1 .771 .77 r
F
.19872 .06547 m .19873 .06548 L .1987 .06543 L 1 .811 .573 r
F
.19869 .06541 m .1987 .06543 L .19851 .06506 L F
.19872 .06547 m .1987 .06543 L .19869 .06541 L F
.19857 .06512 m .19858 .06514 L p .19857 .06512 L .983 .84 .947 r
F P
0 g
s
.19873 .06552 m .19856 .06512 L .19857 .06512 L 1 .771 .77 r
F
.19856 .06512 m .19856 .06512 L .19857 .06512 L F
.19857 .06512 m .19857 .06512 L .19857 .06512 L F
.19856 .06512 m .19857 .06512 L .19857 .06512 L F
.19856 .0651 m .19857 .06512 L p .19856 .0651 L .983 .84 .947 r
F P
0 g
s
.19856 .06512 m .19677 .06086 L .19857 .06512 L 1 .771 .77 r
F
.19857 .06512 m .19857 .06512 L .19856 .0651 L .983 .84 .947 r
F
.19857 .06512 m .19856 .0651 L p .19857 .06512 L 1 .866 .918 r
F P
0 g
s
.19856 .0651 m .19855 .06507 L p .19856 .0651 L 1 .866 .918 r
F P
0 g
s
.19779 .06314 m .19857 .06512 L .19747 .06228 L .992 .61 .145 r
F
.19855 .06508 m .19857 .06512 L .19779 .06314 L F
.19855 .06507 m .19856 .0651 L p .19855 .06507 L .983 .84 .947 r
F P
0 g
s
.19868 .06543 m .1987 .06543 L .19768 .06286 L .902 .713 .863 r
F
.19866 .0652 m .19872 .06546 L .19858 .06512 L 1 .771 .77 r
F
.1987 .06543 m .19873 .06541 L .19859 .06508 L .902 .713 .863 r
F
.19857 .06513 m .19878 .06557 L .19857 .06512 L .992 .61 .145 r
F
.19873 .06549 m .19857 .06512 L .19858 .06512 L 1 .771 .77 r
F
.19855 .06507 m .19866 .06515 L .19857 .06512 L 1 .866 .918 r
F
.1987 .06543 m .19859 .06508 L .19768 .06286 L .902 .713 .863 r
F
.19852 .06514 m .19867 .06545 L .19865 .06545 L .947 .734 .843 r
F
.1985 .06504 m .1985 .06504 L .19867 .06545 L F
.19856 .0651 m .19851 .06511 L .19855 .06507 L .983 .84 .947 r
F
.19865 .06545 m .19854 .06508 L .19851 .06511 L .902 .713 .863 r
F
.19852 .06508 m .19855 .06514 L .19853 .0651 L .992 .61 .145 r
F
.19855 .06514 m .1987 .06546 L .19853 .0651 L F
.1985 .06504 m .19853 .0651 L .19864 .0653 L F
.19854 .06507 m .19852 .06509 L .1985 .06499 L .983 .84 .947 r
F
.1986 .06524 m .19861 .06525 L .19565 .05789 L .902 .713 .863 r
F
.1984 .06477 m .19864 .0653 L .19857 .06513 L .992 .61 .145 r
F
.19852 .06508 m .19869 .06539 L .19869 .06541 L 1 .811 .573 r
F
.19851 .06505 m .19852 .06508 L p .19852 .06508 L F P
0 g
s
.19854 .06508 m .1986 .06524 L .19854 .06508 L .902 .713 .863 r
F
.1984 .06477 m .1985 .06504 L p .19864 .0653 L .992 .61 .145 r
F P
0 g
s
.19852 .06508 m .19869 .06539 L p .19852 .06508 L 1 .811 .573 r
F P
0 g
s
.1985 .06504 m .19852 .06508 L p .19853 .0651 L .992 .61 .145 r
F P
0 g
s
.19875 .06516 m .19894 .06565 L p .19865 .06505 L 1 .785 .643 r
F P
0 g
s
.1984 .06477 m .19855 .06508 L .19779 .06314 L .992 .61 .145 r
F
.19855 .06507 m .19854 .06507 L .1985 .06499 L .983 .84 .947 r
F
.1984 .06477 m .19857 .06513 L .19857 .06512 L .992 .61 .145 r
F
.19796 .06368 m .19854 .06508 L .19802 .0638 L .902 .713 .863 r
F
.19796 .06368 m .19802 .0638 L .19565 .05789 L F
.19672 .06067 m .19796 .06368 L .19565 .05789 L F
.19849 .06512 m .19848 .0651 L .19855 .06507 L .983 .84 .947 r
F
.1987 .06526 m .19865 .06505 L .19883 .06549 L 1 .785 .643 r
F
.19875 .06516 m .1926 .0497 L p .19873 .06511 L .057 .057 .452 r
F P
0 g
s
.2142 .04503 m .21201 .04937 L .20239 .04731 L p .841 1 1 r
F P
0 g
s
.20239 .04731 m .2142 .04503 L p .20239 .04731 L .858 .69 .878 r
F P
0 g
s
.20239 .04731 m .19137 .04495 L .2142 .04503 L p .841 1 1 r
F P
0 g
s
.20054 .0561 m .20239 .04731 L .2142 .04503 L .858 .69 .878 r
F
.21063 .05062 m .19863 .06514 L .21105 .04758 L F
.19867 .06494 m .19869 .065 L .19868 .06495 L .057 .057 .452 r
F
.19863 .06514 m .20054 .0561 L .21105 .04758 L .858 .69 .878 r
F
.19855 .06507 m .19859 .06498 L .19866 .06515 L 1 .866 .918 r
F
.19873 .06541 m .19876 .0654 L .19772 .06284 L .902 .713 .863 r
F
.19849 .06487 m .19859 .06498 L .1987 .06526 L 1 .785 .643 r
F
.19859 .06498 m .19865 .06505 L .19864 .0651 L 1 .866 .918 r
F
.19859 .06498 m .19865 .06505 L .1987 .06526 L 1 .785 .643 r
F
.19863 .06483 m .19883 .06534 L p .19869 .065 L .057 .057 .452 r
F P
0 g
s
.19863 .06483 m .19869 .065 L .19867 .06494 L .057 .057 .452 r
F
.19883 .06534 m .19887 .06545 L .19868 .06495 L F
.19887 .06545 m .1987 .06502 L .19868 .06495 L F
.19639 .05944 m .19244 .04924 L .19645 .05933 L .858 .69 .878 r
F
.19245 .04927 m .19863 .06483 L p .19708 .06094 L .057 .057 .452 r
F P
0 g
s
.19245 .04927 m .19708 .06094 L .19245 .04927 L .057 .057 .452 r
F
.19865 .06505 m .1964 .05946 L .19647 .0594 L .858 .69 .878 r
F
.19772 .06284 m .19884 .06535 L .19217 .04855 L .902 .713 .863 r
F
.19876 .06531 m .19867 .06509 L .19871 .06504 L .858 .69 .878 r
F
.1989 .06551 m .1987 .06502 L .19887 .06545 L .057 .057 .452 r
F
.1989 .06551 m .1989 .06552 L .1987 .06502 L F
.19879 .06554 m .19871 .06504 L .19887 .06544 L .858 .69 .878 r
F
.19849 .06487 m .19846 .06458 L .19865 .06505 L 1 .785 .643 r
F
.19876 .0654 m .19884 .06535 L .19772 .06284 L .902 .713 .863 r
F
.19857 .06502 m .19858 .06504 L .19857 .06502 L F
.19858 .06504 m .19873 .06541 L .19858 .06504 L F
.19849 .06487 m .19883 .06549 L .19864 .06524 L 1 .785 .643 r
F
.19854 .06508 m .19854 .06508 L .19565 .05789 L .902 .713 .863 r
F
.19855 .06507 m .19855 .06507 L .19099 .0504 L .983 .84 .947 r
F
.19095 .05031 m .19099 .0504 L .18899 .04651 L F
.19855 .06507 m .19099 .0504 L .19095 .05031 L F
.18964 .04609 m .19846 .0649 L .19095 .05031 L F
.18902 .04649 m .18964 .04609 L p .19095 .05031 L F P
0 g
s
.19842 .06478 m .19855 .06507 L .19846 .0649 L .983 .84 .947 r
F
.1985 .06504 m .1984 .06477 L p .19847 .06485 L 1 .785 .643 r
F P
0 g
s
.19855 .06507 m .19847 .06485 L p .19849 .06487 L 1 .866 .918 r
F P
0 g
s
.19849 .06487 m .19847 .06485 L .19849 .06487 L 1 .785 .643 r
F
.19847 .06485 m .19855 .06507 L p .19842 .06478 L .983 .84 .947 r
F P
0 g
s
.1985 .06504 m .19847 .06485 L .19849 .06487 L 1 .785 .643 r
F
.19884 .06535 m .19883 .06534 L .19772 .06284 L .902 .713 .863 r
F
.19887 .06534 m .19866 .0648 L p .19884 .06535 L F P
0 g
s
.19866 .0648 m .19217 .04855 L p .19864 .06485 L .902 .713 .863 r
F P
0 g
s
.19884 .06535 m .19866 .0648 L .19864 .06485 L .902 .713 .863 r
F
.19875 .06459 m .19137 .04495 L .19908 .063 L 1 .866 .918 r
F
.19244 .04922 m .19866 .0648 L p .19243 .04922 L .947 .735 .844 r
F P
0 g
s
.20028 .06846 m .19866 .06543 L .2026 .07291 L .957 .829 .959 r
F
.19941 .06691 m .19985 .06772 L .19966 .06737 L 1 .866 .918 r
F
.19866 .06543 m .19859 .06538 L .2026 .07291 L .957 .829 .959 r
F
.19908 .06688 m .19913 .06697 L p .19897 .06663 L 1 .684 .303 r
F P
0 g
s
.1989 .06626 m .19908 .06688 L .19879 .06607 L 1 .757 .441 r
F
.19962 .06819 m .19871 .06584 L .19871 .06584 L 1 .688 .293 r
F
.19857 .06548 m .19962 .0683 L .19962 .06819 L F
.19848 .06529 m .19962 .0683 L .19857 .06548 L F
.19962 .06819 m .19962 .06819 L .19858 .06549 L F
.19857 .06548 m .19962 .06819 L .19858 .06549 L F
.19962 .06819 m .19997 .0691 L .19962 .06819 L F
.19997 .0691 m .19962 .06806 L p .19962 .06819 L F P
0 g
s
.19962 .06806 m .19904 .06632 L p .19962 .06819 L 1 .688 .293 r
F P
0 g
s
.19878 .06615 m .19894 .06626 L .20304 .07677 L 1 .755 .489 r
F
.19831 .06495 m .19894 .06626 L .19878 .06615 L F
.19871 .06584 m .19926 .06703 L .1991 .06653 L 1 .688 .293 r
F
.1989 .06619 m .19896 .06629 L p .19886 .06611 L 1 .599 .157 r
F P
0 g
s
.19975 .0676 m .20142 .07092 L .19962 .06747 L 1 .757 .441 r
F
.19855 .06539 m .19884 .06604 L .19889 .06615 L 1 .599 .157 r
F
.19975 .0676 m .19961 .06745 L .19939 .06702 L 1 .757 .441 r
F
.19975 .0676 m .19962 .06747 L .19961 .06745 L F
.20573 .07872 m .20333 .07431 L p .20575 .07869 L 1 .811 .573 r
F P
0 g
s
.20371 .07505 m .19867 .06564 L .20061 .06936 L .957 .829 .959 r
F
.20342 .07449 m .19867 .06564 L .20371 .07505 L F
.19869 .06545 m .19876 .06564 L p .19868 .06545 L .947 .734 .843 r
F P
0 g
s
.20043 .06891 m .19867 .06564 L .20427 .07608 L .957 .829 .959 r
F
.19939 .06702 m .19922 .06695 L .1989 .06626 L 1 .757 .441 r
F
.19896 .06629 m .19946 .06722 L p .19892 .06621 L 1 .599 .157 r
F P
0 g
s
.19961 .06745 m .19908 .06688 L .19939 .06702 L 1 .757 .441 r
F
.19946 .06722 m .19968 .06762 L p .19946 .06722 L 1 .599 .157 r
F P
0 g
s
.19893 .06623 m .19946 .06722 L .19946 .06722 L 1 .599 .157 r
F
.19968 .06762 m .20145 .07086 L p .19966 .06758 L F P
0 g
s
.19861 .06552 m .19889 .06615 L .19946 .0672 L 1 .599 .157 r
F
.19861 .06552 m .19946 .0672 L .19946 .06712 L F
.19852 .06534 m .19867 .06561 L .19851 .06532 L F
.19859 .06538 m .19851 .06531 L .19865 .06558 L .957 .829 .959 r
F
.19894 .06597 m .19909 .06637 L p .19831 .06495 L 1 .755 .489 r
F P
0 g
s
.19901 .0662 m .19908 .06636 L .19894 .06597 L .992 .61 .145 r
F
.19909 .06637 m .19894 .06597 L p .19908 .06636 L F P
0 g
s
.19952 .06723 m .20145 .07086 L .19951 .06721 L 1 .599 .157 r
F
.19952 .06723 m .19951 .06721 L .1986 .06549 L F
.19951 .06721 m .20145 .07086 L .1995 .06719 L F
.19951 .06721 m .1995 .06719 L .19867 .06561 L F
.19945 .06705 m .20145 .07086 L .20057 .06915 L F
.19857 .06543 m .1995 .06718 L .19945 .06705 L F
.1995 .06718 m .20145 .07086 L .19945 .06705 L F
.20333 .07666 m .20129 .07154 L .20304 .07594 L .057 .057 .452 r
F
.19952 .06723 m .19965 .06755 L .20145 .07086 L 1 .599 .157 r
F
.20285 .0734 m .19967 .0676 L .19967 .0676 L 1 .811 .573 r
F
.20145 .07086 m .20573 .07872 L p .20057 .06915 L 1 .599 .157 r
F P
0 g
s
.20333 .07431 m .19946 .06722 L p .20285 .0734 L 1 .811 .573 r
F P
0 g
s
.20285 .0734 m .19946 .06722 L .19945 .0672 L 1 .811 .573 r
F
.19884 .06604 m .19888 .06614 L .19946 .06721 L 1 .599 .157 r
F
.19999 .06861 m .20203 .07349 L .19998 .06861 L .947 .734 .843 r
F
.20203 .07349 m .20203 .07349 L .19998 .06861 L F
.19858 .06544 m .1987 .06567 L .19857 .06543 L 1 .599 .157 r
F
.1995 .06719 m .20145 .07086 L .1995 .06718 L F
.19925 .06655 m .20055 .06917 L .19955 .06728 L 1 .811 .573 r
F
.19965 .06755 m .19966 .06758 L .20145 .07086 L 1 .599 .157 r
F
.20575 .07869 m .20333 .07431 L .20285 .0734 L 1 .811 .573 r
F
.20579 .07896 m .20085 .06977 L .2053 .07795 L 1 .866 .918 r
F
.20427 .07608 m .20342 .07449 L .20371 .07505 L .957 .829 .959 r
F
.20581 .07883 m .20579 .07896 L p .2049 .07711 L 1 .866 .918 r
F P
0 g
s
.20804 .08297 m .20043 .06891 L .20427 .07608 L .957 .829 .959 r
F
.1987 .06567 m .1995 .06719 L .19869 .06565 L 1 .599 .157 r
F
.1987 .06567 m .19869 .06565 L .19857 .06543 L F
.19861 .06552 m .19863 .06555 L .1986 .06549 L F
.19863 .06555 m .19952 .06723 L .19861 .06551 L F
.19863 .06555 m .19861 .06551 L .1986 .06549 L F
.1986 .06549 m .19951 .06721 L .19867 .06561 L F
.19855 .06539 m .19861 .06552 L .1986 .06549 L F
.19892 .06621 m .19893 .06623 L .19893 .06623 L F
.19855 .06539 m .19867 .06561 L .19852 .06534 L F
.19855 .06539 m .19889 .06615 L .19861 .06552 L F
.19886 .06611 m .19896 .06629 L .19892 .06621 L F
.19819 .06472 m .1989 .06619 L .19855 .06539 L F
.19831 .06495 m .19909 .06637 L .19894 .06626 L 1 .755 .489 r
F
.19884 .06595 m .19887 .06601 L .19889 .06613 L 1 .811 .573 r
F
.19946 .06712 m .19946 .0672 L .19965 .06755 L 1 .599 .157 r
F
.19946 .06712 m .19965 .06755 L .19952 .06723 L F
.19887 .06601 m .20055 .06917 L .19893 .06621 L 1 .811 .573 r
F
.19887 .06601 m .19893 .06621 L .19889 .06613 L F
.19867 .06545 m .19881 .0658 L .19867 .06545 L .947 .734 .843 r
F
.19884 .06604 m .19946 .06721 L .19946 .0672 L 1 .599 .157 r
F
.19945 .06719 m .19896 .06629 L .19896 .06629 L 1 .811 .573 r
F
.19945 .06719 m .19896 .06629 L .19945 .06719 L F
.19893 .06623 m .19946 .06722 L .19893 .06623 L 1 .599 .157 r
F
.19896 .06629 m .19894 .06626 L p .19896 .06629 L 1 .811 .573 r
F P
0 g
s
.19962 .06863 m .20267 .07715 L p .19962 .06851 L 1 .688 .293 r
F P
0 g
s
.20267 .07715 m .20036 .07025 L p .19962 .0683 L 1 .688 .293 r
F P
0 g
s
.19962 .06851 m .20267 .07715 L .19962 .0683 L 1 .688 .293 r
F
.20468 .0811 m .20267 .07715 L .19871 .0661 L p 1 .755 .489 r
F P
0 g
s
.19871 .0661 m .19827 .06487 L p .19877 .06614 L 1 .755 .489 r
F P
0 g
s
.19848 .06529 m .19962 .06851 L .19962 .0683 L 1 .688 .293 r
F
.19962 .0683 m .20036 .07025 L .19962 .06819 L F
.19871 .0661 m .19877 .06614 L .20468 .0811 L 1 .755 .489 r
F
.2028 .0733 m .19895 .06627 L .20267 .07305 L 1 .811 .573 r
F
.19827 .06487 m .19829 .06491 L .20468 .0811 L 1 .755 .489 r
F
.19805 .06445 m .19086 .05144 L .19855 .06555 L p 1 .599 .157 r
F P
0 g
s
.19895 .06627 m .19759 .06379 L .19895 .06627 L 1 .811 .573 r
F
.19895 .06627 m .19895 .06627 L .20267 .07305 L F
.19893 .06621 m .19759 .06379 L .19746 .06351 L F
.20267 .07305 m .19895 .06626 L .19893 .06621 L F
.19884 .06604 m .19886 .06611 L .19888 .06614 L 1 .599 .157 r
F
.19895 .06626 m .19759 .06379 L .19893 .06621 L 1 .811 .573 r
F
.19999 .06861 m .19998 .06861 L .19881 .06581 L .947 .734 .843 r
F
.19881 .06581 m .19881 .06581 L .19867 .06545 L F
.19946 .0672 m .19946 .06721 L .19966 .06758 L 1 .599 .157 r
F
.19946 .0672 m .19966 .06758 L .19965 .06755 L F
.19892 .06621 m .19968 .06762 L .19966 .06758 L F
.20575 .07869 m .19896 .06631 L .19896 .06628 L 1 .811 .573 r
F
.19909 .06637 m .19946 .06735 L p .19898 .06635 L 1 .755 .489 r
F P
0 g
s
.19946 .06736 m .19909 .06637 L .19946 .06736 L .992 .61 .145 r
F
.19927 .06693 m .19946 .06736 L .19908 .06636 L F
.19879 .06573 m .20026 .06923 L .19879 .06574 L .947 .734 .843 r
F
.19927 .06693 m .19908 .06636 L .19921 .06677 L .992 .61 .145 r
F
.19999 .06861 m .19881 .06581 L .19881 .06581 L .947 .734 .843 r
F
.19905 .06637 m .19999 .06861 L .19905 .06638 L F
.19876 .06563 m .1992 .06663 L .19888 .06601 L 1 .811 .573 r
F
.19881 .0658 m .19881 .06581 L .19867 .06545 L .947 .734 .843 r
F
.19865 .06545 m .19867 .06545 L .19881 .06581 L F
.19865 .06545 m .19881 .06581 L .19883 .06588 L F
.19865 .06545 m .19886 .06596 L p .19868 .06543 L .902 .713 .863 r
F P
0 g
s
.19907 .06636 m .19865 .06558 L .19908 .06638 L .957 .829 .959 r
F
.19907 .06636 m .19908 .06638 L .20043 .06891 L F
.2026 .07291 m .19907 .06636 L .20043 .06891 L F
.20788 .08297 m .20804 .08297 L p .19935 .06705 L F P
0 g
s
.20121 .07141 m .19869 .06545 L .19869 .06546 L 1 .771 .77 r
F
.19873 .06548 m .20267 .07305 L .20055 .06917 L 1 .811 .573 r
F
.20804 .08297 m .19876 .06554 L .19874 .06552 L .957 .829 .959 r
F
.20721 .08143 m .20575 .07869 L p .20723 .08145 L .016 .016 .392 r
F P
0 g
s
.20572 .07861 m .20804 .08297 L p .20575 .07869 L .016 .016 .392 r
F P
0 g
s
.20804 .08297 m .19874 .06552 L .20028 .06846 L .957 .829 .959 r
F
.19941 .06675 m .19889 .06577 L .1996 .06711 L .016 .016 .392 r
F
.19902 .06572 m .19867 .06544 L .19867 .06544 L .957 .829 .959 r
F
.19894 .06565 m .199 .06578 L p .19883 .06549 L 1 .785 .643 r
F P
0 g
s
.19893 .06573 m .19911 .06608 L .19895 .06588 L 1 .862 .779 r
F
.20804 .08297 m .19923 .06641 L p .19911 .06608 L F P
0 g
s
.19923 .06641 m .19895 .06588 L p .19911 .06608 L 1 .862 .779 r
F P
0 g
s
.19894 .06561 m .19891 .06554 L .1989 .06554 L .057 .057 .452 r
F
.19908 .06583 m .20804 .08297 L p .19911 .06608 L 1 .862 .779 r
F P
0 g
s
.199 .06578 m .19911 .06608 L .19893 .06573 L 1 .862 .779 r
F
.19908 .06583 m .19911 .06608 L .199 .06578 L F
.19929 .06651 m .19894 .06562 L .19929 .06651 L .057 .057 .452 r
F
.19923 .06641 m .20572 .07861 L p .19921 .06638 L .016 .016 .392 r
F P
0 g
s
.19928 .0665 m .1996 .06711 L .20572 .07861 L .016 .016 .392 r
F
.199 .06578 m .20312 .07613 L p .19893 .06573 L 1 .785 .643 r
F P
0 g
s
.19894 .06562 m .19891 .06554 L .19894 .06561 L .057 .057 .452 r
F
.1993 .06652 m .19904 .06602 L .19887 .06544 L .858 .69 .878 r
F
.19904 .06602 m .19879 .06554 L .19883 .06549 L F
.19883 .06549 m .199 .06578 L .19893 .06573 L 1 .785 .643 r
F
.19908 .06583 m .19902 .06572 L p .19896 .06567 L .957 .829 .959 r
F P
0 g
s
.19888 .06543 m .19908 .06583 L p .199 .06578 L 1 .862 .779 r
F P
0 g
s
.19888 .06547 m .199 .06578 L .19893 .06573 L 1 .862 .779 r
F
.19888 .06543 m .199 .06578 L .19888 .06547 L F
.20575 .07869 m .2028 .0733 L .20267 .07305 L 1 .811 .573 r
F
.20575 .07869 m .20721 .08143 L .20573 .07872 L p F P
0 g
s
.20804 .08297 m .20721 .08143 L p .20723 .08145 L .016 .016 .392 r
F P
0 g
s
.20804 .08297 m .20028 .06846 L .2026 .07291 L .957 .829 .959 r
F
.20354 .07498 m .20474 .07729 L .20776 .08297 L F
.20129 .07088 m .20133 .07095 L .20761 .08297 L F
.20086 .07002 m .20129 .07088 L .20761 .08297 L F
.2069 .08297 m .20738 .08297 L p .20145 .07134 L F P
0 g
s
.20593 .07831 m .2068 .07413 L p .20597 .07822 L 1 .942 .493 r
F P
0 g
s
.2068 .07413 m .21201 .04937 L p .20688 .07393 L 1 .942 .493 r
F P
0 g
s
.20584 .07867 m .20679 .08066 L .20585 .07867 L 1 .599 .157 r
F
.21201 .04937 m .20681 .07819 L .2059 .07864 L 1 .942 .493 r
F
.20738 .08297 m .20761 .08297 L p .20133 .07095 L .957 .829 .959 r
F P
0 g
s
.20623 .08271 m .20018 .06873 L .21063 .05062 L .858 .69 .878 r
F
.20018 .06873 m .19887 .06544 L .21063 .05062 L F
.20129 .07154 m .20018 .06873 L .20623 .08271 L F
.20195 .07305 m .19887 .06534 L p .19947 .06693 L .902 .713 .863 r
F P
0 g
s
.19884 .06536 m .20089 .07052 L .19888 .06548 L .057 .057 .452 r
F
.20128 .07151 m .19929 .06651 L .20126 .07146 L F
.19929 .06651 m .19929 .06651 L .20126 .07146 L F
.1995 .06719 m .1995 .06718 L .19869 .06565 L 1 .599 .157 r
F
.19925 .06655 m .19955 .06728 L .1992 .06663 L 1 .811 .573 r
F
.20099 .07086 m .19879 .06554 L p .20018 .06873 L .858 .69 .878 r
F P
0 g
s
.20129 .07154 m .20127 .07149 L .19891 .06554 L .057 .057 .452 r
F
.20129 .07154 m .1989 .06554 L .2013 .07155 L F
.2013 .07155 m .1989 .06554 L .2013 .07157 L F
.20125 .07143 m .19894 .06565 L .20125 .07144 L F
.20125 .07143 m .19942 .06685 L .20125 .07143 L F
.20125 .07144 m .19894 .06564 L .20126 .07146 L F
.20099 .07086 m .20018 .06873 L .20129 .07154 L .858 .69 .878 r
F
.20129 .07154 m .20128 .07151 L .20129 .07154 L .057 .057 .452 r
F
.20128 .07151 m .20126 .07146 L .20129 .07154 L F
.20126 .07146 m .19894 .06562 L .19894 .06561 L F
.19891 .06581 m .19879 .06554 L .1993 .06652 L .858 .69 .878 r
F
.19895 .06588 m .19923 .06641 L p .19886 .06571 L .016 .016 .392 r
F P
0 g
s
.19879 .06559 m .19923 .06641 L .19921 .06638 L .016 .016 .392 r
F
.20346 .07706 m .19883 .06549 L .20373 .07773 L 1 .785 .643 r
F
.1988 .06545 m .19883 .06549 L .20346 .07706 L F
.19975 .06808 m .19998 .06859 L .20288 .07552 L .947 .734 .843 r
F
.20494 .08297 m .20937 .09454 L .20804 .08297 L closepath p .586 .963 1 r
F P
0 g
s
.203 .07584 m .20288 .07552 L .20614 .08334 L .947 .734 .843 r
F
.20373 .07773 m .20149 .07214 L .20457 .07979 L 1 .785 .643 r
F
.20117 .07132 m .20076 .07033 L .20116 .0713 L 1 .771 .77 r
F
.20149 .07214 m .19883 .06549 L .20124 .07151 L 1 .785 .643 r
F
.20149 .07214 m .20124 .07151 L .20457 .07979 L F
.20116 .0713 m .20076 .07033 L .20077 .07037 L 1 .771 .77 r
F
.20178 .07277 m .20117 .07132 L .20116 .0713 L F
.20026 .06923 m .20174 .07277 L p .2001 .06887 L .947 .734 .843 r
F P
0 g
s
.2001 .06887 m .20174 .07277 L .20192 .07323 L .947 .734 .843 r
F
.19878 .06556 m .19878 .06555 L .19876 .06554 L .957 .829 .959 r
F
.19866 .0652 m .19889 .06577 L p .19872 .06545 L 1 .771 .77 r
F P
0 g
s
.19879 .06559 m .19895 .06588 L .19889 .06577 L .016 .016 .392 r
F
.19884 .06568 m .19889 .06577 L .19886 .06571 L F
.19879 .06559 m .19889 .06577 L .19884 .06568 L F
.19872 .06545 m .20098 .07082 L .19872 .06546 L 1 .771 .77 r
F
.19873 .06548 m .19925 .06655 L .1992 .06663 L 1 .811 .573 r
F
.19967 .0676 m .19945 .06719 L .19945 .06719 L F
.19889 .06577 m .20098 .07082 L p .19872 .06545 L 1 .771 .77 r
F P
0 g
s
.20098 .07082 m .20178 .07277 L p .20114 .07123 L 1 .771 .77 r
F P
0 g
s
.19967 .0676 m .19896 .06629 L .19967 .0676 L 1 .811 .573 r
F
.20098 .07082 m .20114 .07123 L .19872 .06546 L 1 .771 .77 r
F
.19891 .06581 m .19895 .06588 L p .19879 .06559 L .016 .016 .392 r
F P
0 g
s
.20001 .06866 m .2001 .06887 L .2028 .07533 L .947 .734 .843 r
F
.20286 .07548 m .2028 .07533 L .20614 .08334 L F
.20001 .06866 m .2028 .07533 L .20286 .07548 L F
.2036 .07735 m .20457 .07979 L .20127 .07149 L .057 .057 .452 r
F
.204 .07835 m .20457 .07979 L .2036 .07735 L F
.20614 .08334 m .20606 .08334 L .2041 .07859 L p .947 .734 .843 r
F P
0 g
s
.19946 .06735 m .20304 .07677 L p .19894 .06626 L 1 .755 .489 r
F P
0 g
s
.20045 .06997 m .19946 .06736 L .19946 .06736 L .992 .61 .145 r
F
.20328 .07745 m .20046 .07001 L .20045 .06997 L F
.20046 .07001 m .19946 .06736 L .20045 .06997 L F
.20304 .07677 m .20073 .07069 L p .20328 .07745 L F P
0 g
s
.20073 .07069 m .19946 .06735 L p .20051 .07012 L .992 .61 .145 r
F P
0 g
s
.20328 .07745 m .20073 .07069 L .20051 .07012 L .992 .61 .145 r
F
.19865 .06545 m .19906 .06639 L .19918 .06671 L .947 .734 .843 r
F
.19878 .0657 m .19886 .06596 L .19884 .06595 L 1 .811 .573 r
F
.19886 .06596 m .19888 .06601 L .19887 .06601 L F
.19886 .06596 m .19887 .06601 L .19884 .06595 L F
.19865 .06545 m .19865 .06545 L .19885 .06593 L .947 .734 .843 r
F
.19885 .06594 m .19885 .06593 L .19925 .06688 L F
.19865 .06545 m .19885 .06593 L .19885 .06594 L F
.20328 .07745 m .19946 .06735 L .19946 .06736 L .992 .61 .145 r
F
.19946 .06753 m .20328 .07745 L p .19946 .06736 L F P
0 g
s
.20312 .07613 m .20457 .07979 L p .20124 .07151 L 1 .785 .643 r
F P
0 g
s
.19865 .06505 m .20098 .07082 L .20124 .07151 L 1 .785 .643 r
F
.20098 .07082 m .20312 .07613 L .20124 .07151 L F
.20114 .07123 m .20077 .07037 L .20082 .07048 L 1 .771 .77 r
F
.20178 .07277 m .20116 .0713 L .20114 .07123 L F
.20116 .0713 m .20077 .07037 L .20114 .07123 L F
.20301 .07587 m .203 .07584 L .20614 .08334 L .947 .734 .843 r
F
.2041 .07859 m .20301 .07587 L .20614 .08334 L F
.20288 .07552 m .20286 .07548 L .20614 .08334 L F
.20178 .07277 m .20614 .08334 L .20269 .07504 L p 1 .771 .77 r
F P
0 g
s
.20606 .08334 m .20195 .07305 L p .20139 .07178 L .902 .713 .863 r
F P
0 g
s
.2023 .07408 m .20363 .07741 L .20304 .07594 L .057 .057 .452 r
F
.20133 .07164 m .20304 .07594 L .2013 .07157 L F
.20142 .07185 m .20304 .07594 L .20133 .07164 L F
.20334 .0767 m .20127 .07149 L .20129 .07154 L F
.2036 .07735 m .20127 .07149 L .20334 .0767 L F
.20224 .07403 m .20288 .07552 L .203 .07584 L .947 .734 .843 r
F
.20333 .07666 m .20129 .07154 L .20129 .07154 L .057 .057 .452 r
F
.20334 .0767 m .20129 .07154 L .20333 .07666 L F
.20178 .07277 m .20121 .07143 L .20121 .07141 L 1 .771 .77 r
F
.20203 .07349 m .20286 .07548 L .20203 .07349 L .947 .734 .843 r
F
.20129 .07154 m .2013 .07155 L .20304 .07594 L .057 .057 .452 r
F
.20304 .07594 m .2013 .07155 L .2013 .07157 L F
.20177 .07275 m .20125 .07143 L .20125 .07144 L F
.20177 .07275 m .20125 .07143 L .20125 .07143 L F
.20177 .07275 m .20125 .07144 L .20126 .07146 L F
.19921 .06677 m .19908 .06636 L .19901 .0662 L .992 .61 .145 r
F
.19946 .06736 m .19909 .06637 L .19908 .06636 L F
.19962 .06747 m .19913 .06697 L .19961 .06745 L 1 .757 .441 r
F
.20177 .07275 m .20178 .07278 L .19942 .06685 L .057 .057 .452 r
F
.19926 .06692 m .19925 .06688 L .203 .07584 L .947 .734 .843 r
F
.20178 .07278 m .20482 .08042 L .20312 .07613 L .057 .057 .452 r
F
.2041 .07859 m .19936 .06717 L p .20301 .07587 L .947 .734 .843 r
F P
0 g
s
.19936 .06717 m .19946 .06736 L .19927 .06693 L .992 .61 .145 r
F
.19946 .06743 m .19946 .06743 L .19936 .06717 L F
.19946 .06743 m .19946 .06736 L .19936 .06717 L F
.19865 .06545 m .19925 .06688 L .19926 .06692 L .947 .734 .843 r
F
.19936 .06717 m .19865 .06545 L p .19926 .06692 L F P
0 g
s
.19883 .06588 m .19881 .06581 L .19906 .06639 L .947 .734 .843 r
F
.19876 .06563 m .19888 .06601 L .19886 .06596 L 1 .811 .573 r
F
.19876 .06563 m .19886 .06596 L .19876 .06564 L F
.19867 .06545 m .19868 .06545 L .2001 .06887 L .947 .734 .843 r
F
.19876 .06564 m .19886 .06596 L .19878 .0657 L 1 .811 .573 r
F
.19881 .06579 m .2001 .06887 L .20001 .06866 L .947 .734 .843 r
F
.19867 .06545 m .19881 .06579 L .19881 .06579 L F
.19881 .06579 m .2001 .06887 L .19881 .06579 L F
.19881 .0658 m .19905 .06637 L .19881 .06581 L F
.19905 .06637 m .19905 .06638 L .19881 .06581 L F
.19876 .06564 m .19878 .06567 L p .19879 .06573 L F P
0 g
s
.19879 .06573 m .19878 .06567 L .19892 .06604 L .947 .734 .843 r
F
.19886 .06596 m .19936 .06717 L p .19875 .0656 L .902 .713 .863 r
F P
0 g
s
.19865 .06545 m .19998 .06859 L .19975 .06808 L .947 .734 .843 r
F
.19936 .06717 m .19956 .06765 L p .19876 .06563 L .902 .713 .863 r
F P
0 g
s
.19936 .06717 m .19876 .06563 L .19875 .0656 L .902 .713 .863 r
F
.19936 .06717 m .19926 .06692 L .20301 .07587 L .947 .734 .843 r
F
.19867 .06545 m .20286 .07548 L .19867 .06545 L F
.19867 .06545 m .19867 .06545 L .20286 .07548 L F
.19956 .06765 m .2041 .07859 L p .20009 .06884 L .902 .713 .863 r
F P
0 g
s
.2041 .07859 m .20606 .08334 L p .2038 .07785 L .902 .713 .863 r
F P
0 g
s
.20382 .0779 m .204 .07835 L .2036 .07735 L .057 .057 .452 r
F
.2041 .07859 m .2038 .07785 L .20009 .06884 L .902 .713 .863 r
F
.20598 .08334 m .20521 .08139 L p .20482 .08042 L .057 .057 .452 r
F P
0 g
s
.20521 .08139 m .20312 .07613 L p .20482 .08042 L .057 .057 .452 r
F P
0 g
s
.20312 .07613 m .19942 .06685 L p .20178 .07278 L .057 .057 .452 r
F P
0 g
s
.20614 .08334 m .20178 .07277 L p .20623 .08271 L .858 .69 .878 r
F P
0 g
s
.20178 .07277 m .20099 .07086 L p .20129 .07154 L .858 .69 .878 r
F P
0 g
s
.20178 .07277 m .20129 .07154 L .20623 .08271 L .858 .69 .878 r
F
.20269 .07504 m .19869 .06545 L p .20178 .07277 L 1 .771 .77 r
F P
0 g
s
.20064 .07008 m .20328 .07654 L .20139 .07178 L .902 .713 .863 r
F
.20174 .07277 m .20614 .08334 L p .20192 .07323 L .947 .734 .843 r
F P
0 g
s
.20328 .07654 m .20606 .08334 L .20139 .07178 L .902 .713 .863 r
F
.20142 .07185 m .20457 .07978 L p .20363 .07741 L .057 .057 .452 r
F P
0 g
s
.20142 .07185 m .20363 .07741 L .2023 .07408 L .057 .057 .452 r
F
.19926 .06692 m .203 .07584 L .20301 .07587 L .947 .734 .843 r
F
.20457 .07978 m .2036 .07735 L .20334 .0767 L .057 .057 .452 r
F
.19869 .06539 m .19877 .06555 L .1996 .06711 L .016 .016 .392 r
F
.19876 .0654 m .1988 .06545 L .20346 .07706 L 1 .785 .643 r
F
.19885 .06538 m .19888 .06548 L .19887 .06545 L .057 .057 .452 r
F
.2013 .07157 m .1989 .06554 L .1989 .06552 L F
.20125 .07144 m .19894 .06565 L .19894 .06564 L F
.19942 .06685 m .19894 .06565 L p .20125 .07143 L F P
0 g
s
.20126 .07146 m .19894 .06564 L .19893 .0656 L .057 .057 .452 r
F
.20142 .07185 m .1989 .06552 L .1989 .06551 L F
.19891 .06554 m .20177 .07275 L .19893 .0656 L F
.20018 .06873 m .19891 .06581 L .1993 .06652 L .858 .69 .878 r
F
.20077 .07037 m .19873 .06552 L .20082 .07048 L 1 .771 .77 r
F
.19895 .06588 m .19889 .06577 L p .19888 .06543 L 1 .862 .779 r
F P
0 g
s
.19873 .06541 m .1989 .06583 L .19889 .06574 L .902 .713 .863 r
F
.19904 .06602 m .19883 .06549 L .19887 .06544 L .858 .69 .878 r
F
.19888 .06543 m .19889 .06577 L .19884 .06536 L 1 .862 .779 r
F
.19884 .06535 m .19888 .06548 L .19885 .06538 L .057 .057 .452 r
F
.19884 .06536 m .19888 .06548 L .19884 .06535 L F
.19884 .06535 m .19888 .06543 L p .19884 .06536 L 1 .862 .779 r
F P
0 g
s
.19873 .06541 m .19947 .06693 L .19884 .06535 L .902 .713 .863 r
F
.19864 .06524 m .19876 .0654 L .19888 .0657 L 1 .785 .643 r
F
.19886 .06569 m .19888 .0657 L .19895 .06588 L F
.19864 .06524 m .19888 .0657 L .19886 .06569 L F
.19975 .06808 m .20212 .07373 L .20224 .07403 L .947 .734 .843 r
F
.19867 .06545 m .20286 .07548 L .20288 .07552 L F
.20009 .06884 m .20278 .07538 L .20064 .07008 L .902 .713 .863 r
F
.20121 .07143 m .19869 .06545 L .20121 .07141 L 1 .771 .77 r
F
.20077 .07037 m .19874 .06554 L .19873 .06552 L F
.20076 .07033 m .19874 .06554 L .20077 .07037 L F
.19872 .06548 m .19872 .06548 L .20082 .07048 L F
.19946 .06722 m .19897 .06631 L p .19945 .0672 L 1 .811 .573 r
F P
0 g
s
.20082 .07048 m .19872 .06548 L .19873 .06549 L 1 .771 .77 r
F
.19871 .06548 m .19901 .0662 L .1987 .06546 L .992 .61 .145 r
F
.19874 .06558 m .19869 .06546 L .1987 .06546 L 1 .771 .77 r
F
.19875 .0656 m .19901 .0662 L .19868 .06543 L .992 .61 .145 r
F
.1987 .06546 m .19901 .0662 L .19894 .06597 L F
.19873 .06548 m .1992 .06663 L .19876 .06563 L 1 .811 .573 r
F
.20178 .07277 m .19874 .06558 L .19871 .06547 L 1 .771 .77 r
F
.19874 .06558 m .1987 .06546 L .19871 .06547 L F
.19868 .06543 m .19871 .06548 L .1987 .06546 L .992 .61 .145 r
F
.20009 .06884 m .20064 .07008 L .19873 .06541 L .902 .713 .863 r
F
.19874 .0655 m .20197 .07337 L .20236 .07436 L 1 .785 .643 r
F
.19884 .06535 m .19866 .06543 L .19867 .06544 L .957 .829 .959 r
F
.19864 .06524 m .19875 .0655 L .19874 .0655 L 1 .785 .643 r
F
.19875 .0655 m .20197 .07337 L .19874 .0655 L F
.19864 .06524 m .20346 .07706 L .20365 .07754 L F
.19875 .0656 m .19875 .0656 L .19868 .06543 L .902 .713 .863 r
F
.19875 .0656 m .19874 .06558 L .1987 .06548 L F
.19875 .0656 m .1987 .06548 L .19868 .06543 L F
.19243 .04922 m .19866 .0648 L .19866 .0648 L .947 .735 .844 r
F
.19863 .06483 m .19243 .04922 L p .19866 .0648 L F P
0 g
s
.19855 .06507 m .19852 .06497 L .19864 .06485 L 1 .866 .918 r
F
.19852 .06497 m .19759 .06194 L .19863 .06483 L F
.19852 .06497 m .19863 .06483 L .19864 .06485 L F
.1985 .06504 m .19849 .06487 L .19864 .06524 L 1 .785 .643 r
F
.19847 .06501 m .19806 .06421 L .19839 .06491 L .957 .829 .959 r
F
.19806 .06421 m .18899 .04651 L .19745 .06307 L F
.19806 .06421 m .19745 .06307 L .19839 .06491 L F
.1987 .06542 m .19645 .06118 L .19877 .06555 L .016 .016 .392 r
F
.1985 .06504 m .19851 .06507 L .19851 .06507 L 1 .785 .643 r
F
.19869 .06539 m .19851 .06505 L p .1987 .06542 L .016 .016 .392 r
F P
0 g
s
.19851 .06505 m .19824 .06456 L p .19852 .06509 L .016 .016 .392 r
F P
0 g
s
.19851 .06505 m .19852 .06509 L .1987 .06542 L .016 .016 .392 r
F
.19824 .06456 m .19645 .06118 L p .1987 .06542 L F P
0 g
s
.18899 .04651 m .19102 .05114 L p .19867 .06544 L .957 .829 .959 r
F P
0 g
s
.19849 .06512 m .19846 .06506 L .19849 .06513 L 1 .755 .489 r
F
.19852 .06509 m .19843 .06512 L .19836 .06498 L .983 .84 .947 r
F
.19817 .06461 m .19843 .06503 L .19849 .06517 L 1 .755 .489 r
F
.19866 .06543 m .19102 .05114 L .19859 .06538 L .957 .829 .959 r
F
.19123 .05162 m .19102 .05114 L p .19636 .06136 L .983 .84 .947 r
F P
0 g
s
.19102 .05114 m .19182 .05297 L p .19609 .06088 L .957 .829 .959 r
F P
0 g
s
.19086 .05144 m .18945 .04801 L .19824 .06456 L p 1 .811 .573 r
F P
0 g
s
.19824 .06456 m .19851 .06505 L p .19844 .06493 L 1 .811 .573 r
F P
0 g
s
.19838 .06483 m .19785 .0638 L .19847 .06501 L .957 .829 .959 r
F
.19824 .06456 m .19844 .06493 L .19086 .05144 L 1 .811 .573 r
F
.19702 .06187 m .18899 .04651 L p .19785 .0638 L .957 .829 .959 r
F P
0 g
s
.19702 .06187 m .19785 .0638 L .19838 .06483 L .957 .829 .959 r
F
.19852 .06508 m .19847 .06501 L .19868 .06542 L F
.19702 .06187 m .19847 .06501 L .19852 .06508 L F
.19918 .06671 m .19906 .06639 L .19998 .06859 L .947 .734 .843 r
F
.19852 .06508 m .20241 .07449 L .20267 .07514 L 1 .785 .643 r
F
.19877 .06564 m .19921 .06677 L .19901 .0662 L .992 .61 .145 r
F
.19903 .06628 m .20026 .06923 L p .19953 .0675 L .947 .734 .843 r
F P
0 g
s
.19946 .06735 m .19909 .06637 L p .19946 .06736 L .992 .61 .145 r
F P
0 g
s
.19892 .06604 m .19903 .06628 L .19953 .0675 L .947 .734 .843 r
F
.19878 .06567 m .19903 .06628 L p .19892 .06604 L F P
0 g
s
.19956 .06765 m .20009 .06884 L .19874 .06558 L .902 .713 .863 r
F
.19852 .06508 m .20241 .07449 L .19852 .06508 L 1 .785 .643 r
F
.19868 .06542 m .19847 .06501 L .19866 .06543 L .957 .829 .959 r
F
.19855 .06514 m .19875 .0656 L .19868 .06543 L .992 .61 .145 r
F
.19872 .06548 m .19875 .0656 L .19871 .06548 L 1 .811 .573 r
F
.19873 .06548 m .19876 .06564 L .19872 .06548 L F
.19868 .06545 m .19879 .06573 L .19879 .06574 L .947 .734 .843 r
F
.19868 .06545 m .19876 .06564 L .19879 .06573 L F
.19886 .06596 m .19875 .0656 L .19868 .06543 L .902 .713 .863 r
F
.19867 .06544 m .19866 .06543 L .19867 .06544 L .957 .829 .959 r
F
.19852 .06508 m .19921 .06677 L .19877 .06564 L .992 .61 .145 r
F
.19956 .06765 m .19874 .06558 L .19875 .0656 L .902 .713 .863 r
F
.19879 .06574 m .20026 .06923 L .2001 .06887 L .947 .734 .843 r
F
.19855 .06514 m .1987 .06549 L .19855 .06551 L 1 .811 .573 r
F
.19852 .06508 m .19901 .0662 L .19855 .06514 L .992 .61 .145 r
F
.19871 .06547 m .20009 .06884 L .19873 .06541 L .902 .713 .863 r
F
.1987 .06548 m .19874 .06558 L .19871 .06547 L F
.19874 .06558 m .20009 .06884 L .19871 .06547 L F
.1985 .06504 m .19864 .06524 L .19852 .06508 L 1 .785 .643 r
F
.19847 .06485 m .19545 .0571 L .19849 .06487 L F
.19842 .06478 m .19846 .0649 L .19839 .06474 L .983 .84 .947 r
F
.1984 .06477 m .19839 .06474 L p .19847 .06485 L 1 .785 .643 r
F P
0 g
s
.19426 .0545 m .19873 .06541 L .19203 .04873 L .902 .713 .863 r
F
.19622 .0591 m .1984 .06477 L p .19779 .06314 L .992 .61 .145 r
F P
0 g
s
.19839 .06474 m .19545 .0571 L p .19847 .06485 L 1 .785 .643 r
F P
0 g
s
.19677 .06086 m .19601 .05904 L p .19667 .06061 L 1 .771 .77 r
F P
0 g
s
.19572 .05777 m .19244 .04924 L .19639 .05944 L .858 .69 .878 r
F
.19243 .04922 m .19245 .04927 L p .19245 .04927 L .057 .057 .452 r
F P
0 g
s
.19243 .04922 m .19245 .04927 L .19243 .04922 L .057 .057 .452 r
F
.19244 .04924 m .19245 .04927 L .19245 .04931 L F
.19243 .04924 m .19244 .04924 L p .19637 .05948 L .858 .69 .878 r
F P
0 g
s
.19665 .06062 m .19631 .05981 L p .19662 .06055 L .947 .734 .843 r
F P
0 g
s
.19728 .06212 m .19724 .06202 L .1974 .06242 L .947 .734 .843 r
F
.19707 .06159 m .19707 .06157 L .19718 .06186 L F
.19729 .06214 m .19672 .06067 L .19549 .05769 L .902 .713 .863 r
F
.1977 .06316 m .19672 .06067 L .19729 .06214 L F
.19672 .06067 m .19565 .05789 L .19499 .05628 L F
.19549 .05769 m .19597 .05886 L .19426 .0545 L F
.19597 .05886 m .19499 .05628 L .19426 .0545 L F
.19243 .04922 m .19217 .04855 L .19244 .04922 L p .947 .735 .844 r
F P
0 g
s
.19518 .05708 m .19631 .05981 L p .19211 .04925 L .902 .713 .863 r
F P
0 g
s
.19211 .04925 m .19211 .04925 L .19194 .04884 L .902 .713 .863 r
F
.19518 .05708 m .19211 .04925 L .19211 .04925 L F
.19631 .05981 m .19426 .0545 L .19194 .04884 L F
.19217 .04855 m .19213 .0486 L p .19857 .06502 L F P
0 g
s
.19213 .0486 m .19203 .04873 L p .197 .06111 L .902 .713 .863 r
F P
0 g
s
.19857 .06502 m .19857 .06502 L .19217 .04855 L .902 .713 .863 r
F
.19858 .06504 m .19858 .06504 L .19217 .04855 L F
.19203 .04873 m .19194 .04884 L p .19426 .0545 L F P
0 g
s
.19509 .05688 m .19183 .049 L .19481 .05615 L p .947 .734 .843 r
F P
0 g
s
.19601 .05904 m .19195 .04929 L p .19585 .0586 L 1 .771 .77 r
F P
0 g
s
.19195 .04929 m .19183 .049 L .19209 .04931 L p 1 .771 .77 r
F P
0 g
s
.19194 .04884 m .19183 .049 L .19518 .05708 L p .902 .713 .863 r
F P
0 g
s
.19249 .05116 m .18782 .03896 L .1884 .04331 L closepath p .024 .024 .413 r
F P
0 g
s
.18877 .0405 m .18782 .03896 L .18832 .04029 L p .039 .039 .512 r
F P
0 g
s
.18832 .04029 m .19249 .05116 L .18877 .0405 L p .039 .039 .512 r
F P
0 g
s
.18877 .0405 m .18941 .04061 L .18907 .04065 L p .189 .021 .021 r
F P
0 g
s
.18907 .04065 m .1886 .0407 L .18877 .0405 L p .189 .021 .021 r
F P
0 g
s
.18964 .04609 m .19137 .04495 L p .19842 .06478 L .983 .84 .947 r
F P
0 g
s
.19842 .06478 m .19839 .06474 L .18964 .04609 L .983 .84 .947 r
F
.19137 .04495 m .19847 .06485 L p .19842 .06478 L F P
0 g
s
.19849 .06487 m .19137 .04495 L .19759 .06194 L 1 .866 .918 r
F
.19847 .06485 m .19137 .04495 L p .19849 .06487 L F P
0 g
s
.19884 .06535 m .19702 .06187 L p .19868 .06542 L .957 .829 .959 r
F P
0 g
s
.19855 .06514 m .19873 .06548 L .1987 .06549 L 1 .811 .573 r
F
.1987 .06548 m .19873 .06541 L .19868 .06543 L .902 .713 .863 r
F
.19852 .06508 m .19864 .06524 L .20241 .07449 L 1 .785 .643 r
F
.19847 .06501 m .19839 .06491 L .19866 .06543 L .957 .829 .959 r
F
.19484 .05815 m .19885 .06568 L p .19645 .06118 L .016 .016 .392 r
F P
0 g
s
.19885 .06568 m .19891 .06581 L p .19884 .06567 L .016 .016 .392 r
F P
0 g
s
.19645 .06118 m .19885 .06568 L .19884 .06567 L .016 .016 .392 r
F
.19884 .06536 m .19888 .0657 L .18945 .04801 L 1 .862 .779 r
F
.19889 .06577 m .19885 .06568 L p .19888 .0657 L F P
0 g
s
.19864 .06524 m .19895 .06588 L .19875 .0655 L 1 .785 .643 r
F
.19885 .06568 m .18945 .04801 L p .19888 .0657 L 1 .862 .779 r
F P
0 g
s
.19889 .06574 m .20064 .07008 L .19947 .06693 L .902 .713 .863 r
F
.1989 .06583 m .20064 .07008 L .19889 .06574 L F
.19884 .06535 m .19947 .06693 L .19887 .06534 L F
.19883 .06534 m .19884 .06536 L .19884 .06535 L .057 .057 .452 r
F
.20278 .07538 m .20153 .07228 L .20064 .07008 L .902 .713 .863 r
F
.19875 .0655 m .19895 .06588 L .20197 .07337 L 1 .785 .643 r
F
.20236 .07436 m .20197 .07337 L .20346 .07706 L F
.20365 .07754 m .20346 .07706 L .20373 .07773 L F
.20212 .07373 m .20288 .07552 L .20224 .07403 L .947 .734 .843 r
F
.2038 .07785 m .20606 .08334 L .20328 .07654 L .902 .713 .863 r
F
.20457 .07978 m .20334 .0767 L .20333 .07666 L .057 .057 .452 r
F
.20142 .07185 m .1989 .06551 L .19887 .06545 L F
.20142 .07185 m .2013 .07157 L .1989 .06552 L F
.20064 .07008 m .20139 .07178 L .19947 .06693 L .902 .713 .863 r
F
.20139 .07178 m .20195 .07305 L .19947 .06693 L F
.19883 .06534 m .20142 .07185 L p .20089 .07052 L .057 .057 .452 r
F P
0 g
s
.19883 .06534 m .20089 .07052 L .19884 .06536 L .057 .057 .452 r
F
.20267 .07514 m .20241 .07449 L .20365 .07754 L 1 .785 .643 r
F
.19871 .06548 m .19889 .06613 L .19855 .06551 L 1 .811 .573 r
F
.19875 .0656 m .19889 .06613 L .19871 .06548 L F
.19876 .06564 m .19889 .06613 L .19872 .06548 L F
.19896 .06628 m .19896 .06629 L .19895 .06627 L F
.19895 .06627 m .19909 .06637 L .19898 .06635 L 1 .755 .489 r
F
.19852 .06508 m .19946 .06753 L p .19921 .06677 L .992 .61 .145 r
F P
0 g
s
.19896 .06631 m .19896 .06629 L .19896 .06628 L 1 .811 .573 r
F
.19896 .06629 m .19894 .06626 L .19895 .06627 L F
.19896 .06629 m .19895 .06627 L .19895 .06627 L F
.19894 .06626 m .19895 .06627 L .19898 .06635 L 1 .755 .489 r
F
.19946 .06753 m .19927 .06693 L .19921 .06677 L .992 .61 .145 r
F
.20278 .07538 m .2038 .07785 L .20153 .07228 L .902 .713 .863 r
F
.204 .07842 m .20365 .07754 L .20457 .07979 L 1 .785 .643 r
F
.1985 .06504 m .20365 .07754 L .204 .07842 L F
.20457 .07978 m .20333 .07666 L .20304 .07594 L .057 .057 .452 r
F
.2038 .07785 m .20328 .07654 L .20064 .07008 L .902 .713 .863 r
F
.20457 .07978 m .20457 .07979 L .20382 .0779 L .057 .057 .452 r
F
.20457 .07979 m .20457 .07979 L .20382 .0779 L F
.19946 .06753 m .19946 .06743 L .19927 .06693 L .992 .61 .145 r
F
.1989 .06619 m .1989 .06619 L p .19886 .06611 L 1 .599 .157 r
F P
0 g
s
.19866 .0648 m .20195 .07305 L p .19866 .0648 L .947 .735 .844 r
F P
0 g
s
.20195 .07305 m .20606 .08334 L p .20195 .07304 L .947 .735 .844 r
F P
0 g
s
.19866 .0648 m .20195 .07305 L .20195 .07304 L .947 .735 .844 r
F
.19855 .06555 m .1989 .06619 L p .19819 .06472 L 1 .599 .157 r
F P
0 g
s
.19868 .06491 m .19865 .06483 L .19866 .0648 L 1 .866 .918 r
F
.19871 .06474 m .19866 .0648 L .19875 .06459 L F
.19868 .06491 m .19866 .0648 L .19871 .06474 L F
.19865 .06483 m .19137 .04495 L .19875 .06459 L F
.18945 .04801 m .18899 .04651 L .19884 .06535 L p 1 .862 .779 r
F P
0 g
s
.19884 .06535 m .19884 .06536 L .18945 .04801 L 1 .862 .779 r
F
.19884 .06535 m .19884 .06535 L .19883 .06534 L .902 .713 .863 r
F
.19851 .06506 m .19855 .06514 L .19855 .06551 L 1 .811 .573 r
F
.21372 .04734 m .20614 .08334 L p .21063 .05062 L .858 .69 .878 r
F P
0 g
s
.21063 .05062 m .21105 .04758 L .2132 .04739 L .858 .69 .878 r
F
.21063 .05062 m .2132 .04739 L .21372 .04734 L F
.19883 .06534 m .19863 .06483 L p .19887 .06534 L .947 .735 .844 r
F P
0 g
s
.19863 .06483 m .19866 .0648 L .19887 .06534 L .947 .735 .844 r
F
.19805 .06445 m .19812 .06461 L .19819 .06472 L 1 .599 .157 r
F
.19812 .06461 m .19855 .06555 L .19819 .06472 L F
.19851 .06505 m .19855 .06551 L .19746 .06351 L 1 .811 .573 r
F
.19633 .06117 m .19746 .06351 L .19086 .05144 L F
.19851 .06505 m .19746 .06351 L .19633 .06117 L F
.20606 .08334 m .20598 .08334 L .20142 .07185 L p .947 .735 .844 r
F P
0 g
s
.20142 .07185 m .19883 .06534 L p .20172 .07253 L .947 .735 .844 r
F P
0 g
s
.20142 .07185 m .20172 .07253 L .20606 .08334 L .947 .735 .844 r
F
.20304 .07677 m .20468 .0811 L p .20198 .07424 L 1 .755 .489 r
F P
0 g
s
.19962 .06819 m .20036 .07025 L .19997 .0691 L 1 .688 .293 r
F
.19878 .06615 m .20304 .07677 L .20198 .07424 L 1 .755 .489 r
F
.19894 .06626 m .19759 .06379 L p .19895 .06627 L 1 .811 .573 r
F P
0 g
s
.19759 .06379 m .19086 .05144 L p .19746 .06351 L 1 .811 .573 r
F P
0 g
s
.19883 .06534 m .19887 .06534 L .20606 .08334 L .947 .735 .844 r
F
.19546 .06044 m .19568 .06086 L .19322 .0568 L 1 .757 .441 r
F
.19651 .06247 m .19322 .0568 L p .19568 .06086 L F P
0 g
s
.19645 .06118 m .19891 .06581 L .19879 .06559 L .016 .016 .392 r
F
.18899 .04651 m .18902 .04649 L p .19095 .05031 L .983 .84 .947 r
F P
0 g
s
.19864 .06485 m .19759 .06194 L .19865 .06483 L 1 .866 .918 r
F
.20328 .07745 m .20468 .0811 L .20304 .07677 L p .992 .61 .145 r
F P
0 g
s
.19645 .06118 m .18945 .04801 L .19484 .05815 L p .016 .016 .392 r
F P
0 g
s
.20437 .0803 m .19946 .06753 L p .20457 .07979 L 1 .785 .643 r
F P
0 g
s
.19946 .06753 m .1985 .06504 L p .20009 .06889 L 1 .785 .643 r
F P
0 g
s
.19946 .06753 m .20009 .06889 L .20457 .07979 L 1 .785 .643 r
F
.20598 .08334 m .20468 .0811 L .20437 .0803 L p F P
0 g
s
.20457 .07979 m .20598 .08334 L p .20437 .0803 L 1 .785 .643 r
F P
0 g
s
.2077 .07775 m .21309 .06754 L p .21197 .07564 L 1 .942 .493 r
F P
0 g
s
.20457 .07978 m .20598 .08334 L p .20457 .07979 L .057 .057 .452 r
F P
0 g
s
.2142 .04503 m .21372 .04734 L p .21105 .04758 L .858 .69 .878 r
F P
0 g
s
.21041 .09081 m .21041 .09082 L .21041 .09085 L 1 1 .819 r
F
.21041 .09085 m .2116 .09302 L p .21041 .09081 L .658 .603 .927 r
F P
0 g
s
.20937 .09454 m .21023 .09046 L p .21041 .09081 L 1 1 .819 r
F P
0 g
s
.21041 .09081 m .21041 .09085 L .20937 .09454 L 1 1 .819 r
F
.2149 .05745 m .21632 .06142 L .21364 .06649 L p 1 .942 .493 r
F P
0 g
s
.2116 .09302 m .21041 .09081 L p .21041 .09081 L .658 .603 .927 r
F P
0 g
s
.21041 .09082 m .21049 .09097 L .21039 .0913 L 1 1 .819 r
F
.21041 .09085 m .21039 .0913 L .20937 .09454 L F
.21041 .09082 m .21039 .0913 L .21041 .09085 L F
.21026 .09029 m .2116 .09302 L .21041 .09085 L p .131 .185 .747 r
F P
0 g
s
.21034 .09145 m .21035 .09143 L .21195 .09454 L .947 .735 .845 r
F
.21023 .09046 m .20937 .09454 L p .21 .08999 L 1 .942 .493 r
F P
0 g
s
.21309 .06754 m .21632 .06142 L .21126 .08553 L p 1 .942 .493 r
F P
0 g
s
.2116 .09302 m .21242 .09454 L .21144 .09272 L p .006 .006 .358 r
F P
0 g
s
.20594 .08707 m .20781 .09099 L p .20496 .08363 L .304 .419 .897 r
F P
0 g
s
.20781 .09099 m .2086 .09264 L p .20496 .08363 L .304 .419 .897 r
F P
0 g
s
.90404 .06109 m .90966 .06111 L .89263 .06741 L closepath p .955 .786 .906 r
F P
0 g
s
.90966 .06111 m .91527 .06113 L .89531 .07286 L closepath p .955 .786 .906 r
F P
0 g
s
.89531 .07286 m .89263 .06741 L .90966 .06111 L closepath p .955 .786 .906 r
F P
0 g
s
.91527 .06113 m .91907 .05689 L .89281 .06105 L closepath p .843 1 1 r
F P
0 g
s
.89373 .0568 m .89281 .06105 L .91907 .05689 L closepath p .843 1 1 r
F P
0 g
s
.89281 .06105 m .89373 .0568 L .89104 .06143 L closepath p .911 1 1 r
F P
0 g
s
.6436 .09653 m .64494 .09872 L .64477 .0983 L p .699 .68 .989 r
F P
0 g
s
.62952 .06105 m .62927 .06122 L p .63715 .079 L .922 .726 .86 r
F P
0 g
s
.62927 .06122 m .62763 .06232 L p .63631 .07754 L .922 .726 .86 r
F P
0 g
s
.63631 .07754 m .63715 .079 L .62927 .06122 L .922 .726 .86 r
F
.63328 .07259 m .63715 .079 L p .62763 .06232 L F P
0 g
s
.63715 .079 m .63718 .07904 L .62952 .06105 L .922 .726 .86 r
F
.63718 .07904 m .63733 .0792 L .63733 .07889 L F
.63718 .07904 m .63733 .07889 L .62952 .06105 L F
.63858 .08259 m .63721 .07912 L p .63796 .08128 L .699 .68 .989 r
F P
0 g
s
.63721 .07912 m .6372 .0791 L .63972 .08634 L .699 .68 .989 r
F
.63676 .07858 m .63972 .08634 L p .63718 .07904 L F P
0 g
s
.63717 .07905 m .6385 .08242 L .63706 .07915 L .997 .845 .937 r
F
.63721 .07912 m .63858 .08259 L p .63717 .07905 L F P
0 g
s
.64365 .09565 m .63858 .08259 L .63972 .08634 L .699 .68 .989 r
F
.64477 .0983 m .6422 .09177 L p .64365 .09565 L F P
0 g
s
.6422 .09177 m .63858 .08259 L p .64365 .09565 L .699 .68 .989 r
F P
0 g
s
.63858 .08259 m .64494 .09872 L p .6385 .08242 L .997 .845 .937 r
F P
0 g
s
.63718 .07904 m .63702 .07865 L p .63676 .07858 L .699 .68 .989 r
F P
0 g
s
.62952 .06105 m .63684 .07843 L p .63702 .07865 L .947 .735 .845 r
F P
0 g
s
.63709 .07862 m .63702 .07865 L .63718 .07883 L .947 .735 .845 r
F
.62952 .06105 m .63702 .07865 L .63709 .07862 L F
.63721 .07856 m .63718 .07883 L .63733 .07883 L F
.62952 .06105 m .63709 .07862 L .63721 .07856 L F
.63709 .07862 m .63718 .07883 L .63721 .07856 L F
.63972 .08634 m .6436 .09653 L p .64477 .0983 L .699 .68 .989 r
F P
0 g
s
.63858 .08259 m .6385 .08242 L .63717 .07905 L .997 .845 .937 r
F
.63723 .07923 m .63871 .08284 L p .63911 .08265 L .957 .827 .957 r
F P
0 g
s
.63728 .07923 m .63911 .08265 L .64747 .07887 L .957 .827 .957 r
F
.63723 .07923 m .63911 .08265 L .63728 .07923 L F
.64535 .09782 m .64567 .09493 L p .63871 .08284 L F P
0 g
s
.64567 .09493 m .64747 .07887 L p .63911 .08265 L .957 .827 .957 r
F P
0 g
s
.63871 .08284 m .64567 .09493 L .63911 .08265 L .957 .827 .957 r
F
.63728 .07922 m .64747 .07887 L .63854 .07884 L F
.63723 .07923 m .63728 .07923 L .63728 .07922 L F
.63728 .07923 m .64747 .07887 L .63728 .07922 L F
.63718 .07904 m .63716 .07903 L .63701 .07865 L .997 .845 .937 r
F
.63715 .079 m .63697 .07867 L .63712 .07896 L .094 .094 .669 r
F
.63716 .07903 m .63697 .07867 L .63701 .07865 L .997 .845 .937 r
F
.63721 .07912 m .63718 .07904 L p .6372 .0791 L .699 .68 .989 r
F P
0 g
s
.63718 .07904 m .63721 .07912 L p .63717 .07905 L .997 .845 .937 r
F P
0 g
s
.63715 .079 m .63801 .08043 L p .6376 .07951 L .922 .726 .86 r
F P
0 g
s
.63736 .0798 m .63716 .07926 L .63714 .07924 L .846 .769 .986 r
F
.63858 .08259 m .63796 .08128 L .63972 .08634 L .699 .68 .989 r
F
.6385 .08242 m .64494 .09872 L .63706 .07915 L .997 .845 .937 r
F
.64181 .09074 m .63755 .08021 L .63736 .0798 L .846 .769 .986 r
F
.63755 .08021 m .63716 .07926 L .63736 .0798 L F
.6376 .07951 m .63853 .08003 L .638 .07891 L .922 .726 .86 r
F
.63742 .07909 m .638 .07891 L .62952 .06105 L F
.6376 .07951 m .638 .07891 L .63742 .07909 L F
.6362 .07792 m .64151 .09264 L p .63688 .07886 L .556 .59 .97 r
F P
0 g
s
.64494 .09872 m .63699 .07922 L p .63706 .07915 L .997 .845 .937 r
F P
0 g
s
.64494 .09872 m .64509 .09872 L .63706 .0792 L p .957 .827 .957 r
F P
0 g
s
.62952 .06105 m .63729 .07876 L .63793 .07876 L .947 .735 .845 r
F
.63733 .0792 m .63745 .07917 L .63733 .07889 L .922 .726 .86 r
F
.64136 .08599 m .6437 .08961 L .63838 .07888 L F
.63804 .07899 m .63838 .07888 L .62952 .06105 L F
.64136 .08599 m .63838 .07888 L .63804 .07899 L F
.63717 .07905 m .63715 .079 L p .63717 .07904 L .094 .094 .669 r
F P
0 g
s
.63716 .07903 m .63706 .07915 L .637 .0789 L .997 .845 .937 r
F
.63718 .07904 m .63717 .07905 L .63716 .07903 L F
.63717 .07905 m .63706 .07915 L .63716 .07903 L F
.63706 .07915 m .63704 .07917 L .63702 .07907 L F
.63706 .07915 m .63702 .07907 L .637 .0789 L F
.63871 .08284 m .63723 .07923 L p .63838 .08209 L .957 .827 .957 r
F P
0 g
s
.63723 .07923 m .63706 .07883 L p .63716 .07926 L .957 .827 .957 r
F P
0 g
s
.63697 .07867 m .63697 .07867 L .63712 .07896 L .094 .094 .669 r
F
.63716 .07903 m .637 .0789 L .63697 .07867 L .997 .845 .937 r
F
.637 .0789 m .63696 .07868 L .63697 .07867 L F
.63678 .07835 m .63695 .07863 L p .63715 .079 L .094 .094 .669 r
F P
0 g
s
.63715 .079 m .63676 .07832 L .6368 .07839 L .094 .094 .669 r
F
.63686 .07849 m .63689 .07856 L p .63707 .07883 L .947 .735 .845 r
F P
0 g
s
.63709 .07917 m .63705 .07883 L .63697 .07869 L .701 .636 .937 r
F
.63689 .07856 m .63701 .07883 L p .63705 .07883 L .947 .735 .845 r
F P
0 g
s
.63695 .07863 m .63704 .07878 L p .63712 .07896 L .094 .094 .669 r
F P
0 g
s
.63704 .07879 m .63704 .07878 L p .63707 .07904 L .957 .827 .957 r
F P
0 g
s
.64446 .09114 m .64456 .09072 L .62952 .06105 L .922 .726 .86 r
F
.63854 .07884 m .63845 .07986 L .63939 .08052 L .947 .735 .845 r
F
.62952 .06105 m .63796 .07883 L .63854 .07884 L F
.63796 .07883 m .63845 .07986 L .63854 .07884 L F
.65183 .06113 m .62952 .06105 L p .63854 .07884 L F P
0 g
s
.63642 .07745 m .63542 .07507 L p .63497 .07388 L .947 .735 .845 r
F P
0 g
s
.63542 .07507 m .62952 .06105 L p .63497 .07388 L .947 .735 .845 r
F P
0 g
s
.63854 .07884 m .63939 .08052 L .64047 .07884 L .947 .735 .845 r
F
.63854 .07884 m .64047 .07884 L .65183 .06113 L F
.64456 .09072 m .64593 .08513 L p .63939 .08052 L F P
0 g
s
.63939 .08052 m .64593 .08513 L .65183 .06113 L .947 .735 .845 r
F
.64779 .07603 m .64747 .07885 L p .65104 .06122 L 1 1 .703 r
F P
0 g
s
.64747 .07885 m .64567 .09493 L p .64976 .06954 L 1 1 .703 r
F P
0 g
s
.64976 .06954 m .64567 .09493 L .64578 .09528 L 1 1 .703 r
F
.64525 .09872 m .65183 .06113 L p .64975 .06958 L F P
0 g
s
.64652 .09454 m .6464 .09434 L p .6464 .09434 L .922 .726 .86 r
F P
0 g
s
.63816 .07809 m .63817 .07809 L .62952 .06105 L p .922 .726 .86 r
F P
0 g
s
.63715 .079 m .63704 .07878 L .63727 .07921 L .094 .094 .669 r
F
.63717 .07905 m .63717 .07904 L .63727 .07921 L F
.63717 .07905 m .63727 .07921 L .63727 .07921 L F
.63728 .07922 m .63717 .07905 L p .63727 .07921 L F P
0 g
s
.63706 .07883 m .63723 .07923 L p .6375 .07883 L .957 .827 .957 r
F P
0 g
s
.63723 .07923 m .63854 .07884 L .6375 .07883 L .957 .827 .957 r
F
.63801 .08043 m .64012 .08339 L .63853 .08003 L .922 .726 .86 r
F
.63801 .08043 m .63853 .08003 L .6376 .07951 L F
.63733 .0792 m .6376 .07951 L .63745 .07917 L F
.63801 .08043 m .64136 .08599 L p .64012 .08339 L F P
0 g
s
.64049 .0827 m .63716 .07883 L .64456 .09072 L .947 .735 .845 r
F
.63702 .07865 m .63718 .07904 L p .63701 .07865 L .997 .845 .937 r
F P
0 g
s
.63716 .07883 m .63686 .07849 L .63708 .07883 L .947 .735 .845 r
F
.63716 .07883 m .63708 .07883 L .64456 .09072 L F
.63701 .07865 m .63694 .07869 L .63701 .07865 L .997 .845 .937 r
F
.63684 .07843 m .63686 .07849 L p .64049 .0827 L .947 .735 .845 r
F P
0 g
s
.64136 .08599 m .64446 .09114 L p .6437 .08961 L .922 .726 .86 r
F P
0 g
s
.64446 .09112 m .64456 .09072 L p .63686 .07849 L .947 .735 .845 r
F P
0 g
s
.63817 .07809 m .63855 .07883 L p .63697 .07867 L .947 .735 .845 r
F P
0 g
s
.63793 .07876 m .63784 .07884 L .63796 .07884 L .947 .735 .845 r
F
.63729 .07876 m .63748 .07918 L .63793 .07876 L F
.63697 .07867 m .63855 .07883 L .63707 .07883 L F
.63691 .07847 m .63068 .06332 L .63817 .07809 L p F P
0 g
s
.6366 .07835 m .63667 .07835 L .62724 .06258 L .701 .636 .937 r
F
.63667 .07835 m .63676 .07835 L .63648 .07788 L F
.63667 .07835 m .63648 .07788 L .62724 .06258 L F
.63643 .07782 m .63328 .07259 L p .63644 .07778 L .094 .094 .669 r
F P
0 g
s
.62763 .06232 m .62724 .06258 L .63328 .07259 L p .922 .726 .86 r
F P
0 g
s
.63328 .07259 m .62724 .06258 L .63365 .07318 L p .094 .094 .669 r
F P
0 g
s
.63365 .07318 m .63644 .07778 L p .63328 .07259 L .094 .094 .669 r
F P
0 g
s
.63644 .07778 m .63678 .07835 L p .63643 .07782 L .094 .094 .669 r
F P
0 g
s
.6365 .07789 m .63365 .07318 L p .63696 .07868 L .701 .636 .937 r
F P
0 g
s
.63365 .07318 m .62724 .06258 L p .63374 .07335 L .701 .636 .937 r
F P
0 g
s
.63696 .07868 m .63365 .07318 L .63374 .07335 L .701 .636 .937 r
F
.63688 .07886 m .6393 .08522 L p .63676 .07858 L .078 .078 .537 r
F P
0 g
s
.63972 .08634 m .63676 .07858 L p .6393 .08522 L .078 .078 .537 r
F P
0 g
s
.62724 .06258 m .62761 .06367 L p .63684 .07877 L .701 .636 .937 r
F P
0 g
s
.62944 .06681 m .63684 .07877 L p .62761 .06367 L .701 .636 .937 r
F P
0 g
s
.63684 .07877 m .63684 .07877 L .62724 .06258 L .701 .636 .937 r
F
.62774 .06406 m .62922 .06742 L .63616 .07792 L p .516 .542 .939 r
F P
0 g
s
.63677 .07835 m .63574 .0757 L .6366 .07815 L .699 .68 .989 r
F
.63582 .07612 m .6366 .07815 L p .63574 .0757 L F P
0 g
s
.62991 .06061 m .63582 .07612 L p .63574 .0757 L .699 .68 .989 r
F P
0 g
s
.63667 .07835 m .63582 .07612 L p .63685 .07878 L .078 .078 .537 r
F P
0 g
s
.63582 .07612 m .62991 .06061 L p .63611 .07685 L .078 .078 .537 r
F P
0 g
s
.63685 .07878 m .63582 .07612 L .63611 .07685 L .078 .078 .537 r
F
.63688 .07886 m .63676 .07858 L .63688 .07885 L F
.63686 .07881 m .63685 .07877 L .62724 .06258 L .701 .636 .937 r
F
.63658 .07836 m .63526 .07621 L p .63601 .07715 L .23 .385 .884 r
F P
0 g
s
.63674 .07866 m .63609 .07722 L .63637 .07815 L .556 .59 .97 r
F
.63608 .07725 m .63601 .07715 L p .63605 .07732 L .516 .542 .939 r
F P
0 g
s
.63609 .07722 m .63536 .07559 L .63605 .07732 L .556 .59 .97 r
F
.63609 .07722 m .63605 .07732 L .63637 .07815 L F
.63688 .07886 m .63688 .07886 L .63688 .07885 L .078 .078 .537 r
F
.63688 .07886 m .63685 .07878 L p .63674 .07866 L .556 .59 .97 r
F P
0 g
s
.63684 .07877 m .63686 .07881 L p .63684 .07877 L .701 .636 .937 r
F P
0 g
s
.63686 .07881 m .63702 .07907 L p .63685 .07877 L .701 .636 .937 r
F P
0 g
s
.63699 .07922 m .6368 .07875 L p .637 .0789 L .997 .845 .937 r
F P
0 g
s
.63644 .07771 m .63688 .07886 L p .63644 .07771 L .078 .078 .537 r
F P
0 g
s
.63676 .07858 m .63667 .07835 L p .63688 .07885 L .078 .078 .537 r
F P
0 g
s
.63688 .07886 m .63667 .07835 L .63685 .07878 L .078 .078 .537 r
F
.63681 .07835 m .63235 .06688 L .63677 .07835 L .699 .68 .989 r
F
.63235 .06688 m .62991 .06061 L .63242 .0671 L F
.63235 .06688 m .63242 .0671 L .63677 .07835 L F
.63616 .07792 m .63609 .07753 L .62774 .06406 L .516 .542 .939 r
F
.63637 .07815 m .63605 .07732 L .6362 .07792 L .556 .59 .97 r
F
.63605 .07732 m .63601 .07715 L .63601 .07738 L .516 .542 .939 r
F
.63536 .07559 m .63601 .07738 L p .63605 .07732 L .556 .59 .97 r
F P
0 g
s
.63601 .07738 m .6362 .07792 L p .63605 .07732 L .556 .59 .97 r
F P
0 g
s
.63601 .07715 m .63551 .07636 L p .63601 .07738 L .516 .542 .939 r
F P
0 g
s
.63609 .07753 m .63605 .07732 L .63601 .07738 L .516 .542 .939 r
F
.63526 .07621 m .62944 .06681 L p .63551 .07635 L .23 .385 .884 r
F P
0 g
s
.63526 .07621 m .63551 .07635 L .63601 .07715 L .23 .385 .884 r
F
.62761 .06367 m .62774 .06406 L .62944 .06681 L p .701 .636 .937 r
F P
0 g
s
.62944 .06681 m .62774 .06406 L .62987 .06743 L p .23 .385 .884 r
F P
0 g
s
.62987 .06743 m .63601 .07715 L p .62944 .06681 L .23 .385 .884 r
F P
0 g
s
.63551 .07636 m .62774 .06406 L p .63609 .07753 L .516 .542 .939 r
F P
0 g
s
.63652 .07843 m .63568 .07645 L .63569 .07646 L .261 .384 .875 r
F
.63715 .079 m .63643 .07782 L p .63676 .07832 L .094 .094 .669 r
F P
0 g
s
.63702 .07865 m .63677 .07835 L .63676 .07858 L .699 .68 .989 r
F
.6366 .07815 m .63667 .07835 L p .63677 .07835 L F P
0 g
s
.63667 .07835 m .63676 .07858 L p .63677 .07835 L .699 .68 .989 r
F P
0 g
s
.63696 .07868 m .63686 .07849 L .63697 .07867 L .947 .735 .845 r
F
.63689 .07856 m .63686 .07849 L p .63696 .07868 L F P
0 g
s
.63695 .07863 m .63678 .07835 L p .63696 .07868 L .701 .636 .937 r
F P
0 g
s
.63678 .07835 m .6365 .07789 L p .63677 .07835 L .701 .636 .937 r
F P
0 g
s
.6368 .07839 m .63676 .07832 L .63678 .07835 L .094 .094 .669 r
F
.63696 .07868 m .63678 .07835 L .63677 .07835 L .701 .636 .937 r
F
.63702 .07907 m .63676 .07835 L .6366 .07835 L F
.63702 .07907 m .63683 .07876 L p .63698 .07896 L .23 .385 .884 r
F P
0 g
s
.63683 .07876 m .63684 .07876 L .63698 .07896 L .23 .385 .884 r
F
.6368 .07876 m .63699 .07922 L p .63688 .07876 L .957 .827 .957 r
F P
0 g
s
.63683 .07876 m .63658 .07836 L p .63679 .07868 L .23 .385 .884 r
F P
0 g
s
.63683 .07876 m .63679 .07868 L .63683 .07876 L .23 .385 .884 r
F
.63497 .07388 m .62952 .06105 L .63063 .06324 L .947 .735 .845 r
F
.63685 .07878 m .63442 .07239 L p .63661 .07831 L .556 .59 .97 r
F P
0 g
s
.63442 .07239 m .63644 .07771 L p .63304 .06879 L .078 .078 .537 r
F P
0 g
s
.63347 .0699 m .63304 .06879 L .62991 .06061 L .078 .078 .537 r
F
.63442 .07239 m .63304 .06879 L .63347 .0699 L F
.63685 .07878 m .63661 .07831 L .63684 .07877 L .556 .59 .97 r
F
.63658 .07836 m .63668 .07828 L .63685 .07876 L .23 .385 .884 r
F
.63691 .0787 m .63687 .07872 L .63509 .07409 L .997 .845 .937 r
F
.63683 .07876 m .63679 .07868 L .63684 .07876 L .23 .385 .884 r
F
.63685 .07876 m .63668 .07828 L .63698 .07896 L F
.63687 .07872 m .6368 .07875 L .6364 .07766 L .997 .845 .937 r
F
.63687 .07872 m .6364 .07766 L .63509 .07409 L F
.6368 .07876 m .6368 .07875 L .63694 .07869 L F
.63685 .07868 m .63655 .078 L .63685 .07868 L .23 .385 .884 r
F
.63655 .078 m .63656 .07801 L p .63685 .07868 L F P
0 g
s
.63688 .07876 m .63653 .07791 L p .6368 .07875 L .957 .827 .957 r
F P
0 g
s
.63657 .07802 m .63655 .078 L p .63658 .07806 L .516 .542 .939 r
F P
0 g
s
.63658 .07836 m .63655 .078 L .63668 .07828 L .23 .385 .884 r
F
.63509 .07409 m .63672 .07853 L .63577 .0762 L .997 .845 .937 r
F
.63637 .07784 m .63615 .0771 L .63609 .07722 L .556 .59 .97 r
F
.63658 .07836 m .63608 .07725 L .63657 .07824 L .23 .385 .884 r
F
.63674 .07866 m .63661 .07831 L .63637 .07784 L .556 .59 .97 r
F
.63661 .07831 m .63615 .0771 L .63637 .07784 L F
.63601 .07715 m .63608 .07725 L p .63658 .07836 L .23 .385 .884 r
F P
0 g
s
.63608 .07725 m .63655 .078 L p .63657 .07824 L .23 .385 .884 r
F P
0 g
s
.63655 .078 m .63608 .07725 L p .63675 .07862 L .516 .542 .939 r
F P
0 g
s
.63644 .07771 m .63644 .07771 L .62991 .06061 L .078 .078 .537 r
F
.63442 .07239 m .6338 .07081 L .63661 .07831 L .556 .59 .97 r
F
.63675 .07863 m .63577 .0762 L p .63672 .07853 L .997 .845 .937 r
F P
0 g
s
.63684 .07843 m .63681 .07835 L .63677 .07835 L .699 .68 .989 r
F
.63702 .07907 m .63696 .07868 L .63676 .07835 L .701 .636 .937 r
F
.63702 .07865 m .63702 .07865 L p .63684 .07843 L .699 .68 .989 r
F P
0 g
s
.63677 .07828 m .63688 .07848 L p .63688 .07848 L .947 .735 .845 r
F P
0 g
s
.63686 .07849 m .63684 .07843 L .63688 .07848 L .947 .735 .845 r
F
.63684 .07843 m .63677 .07828 L .63688 .07848 L F
.63694 .07869 m .63689 .07856 L p .63696 .07868 L F P
0 g
s
.63686 .07849 m .63817 .07809 L .63697 .07867 L .947 .735 .845 r
F
.63697 .07867 m .63695 .07863 L .63697 .07867 L .094 .094 .669 r
F
.63701 .07865 m .63691 .0787 L .63509 .07409 L .997 .845 .937 r
F
.63702 .07865 m .62991 .06061 L p .63684 .07843 L .699 .68 .989 r
F P
0 g
s
.64942 .06142 m .63502 .06137 L p .63706 .07883 L .957 .827 .957 r
F P
0 g
s
.63615 .0771 m .63349 .07003 L .63609 .07722 L .556 .59 .97 r
F
.62991 .06061 m .63702 .07865 L p .63508 .07405 L .997 .845 .937 r
F P
0 g
s
.63702 .07865 m .63701 .07865 L .63488 .0736 L .997 .845 .937 r
F
.6358 .07577 m .63706 .07883 L p .63087 .06315 L .957 .827 .957 r
F P
0 g
s
.62991 .06061 m .63442 .07239 L .63251 .06754 L .997 .845 .937 r
F
.63239 .06705 m .63184 .06564 L .63349 .07003 L .556 .59 .97 r
F
.6338 .07081 m .63349 .07003 L .63661 .07831 L F
.63239 .06705 m .63349 .07003 L .6338 .07081 L F
.63349 .07003 m .6326 .06776 L .63609 .07722 L F
.63251 .06754 m .63488 .0736 L .63397 .07142 L .997 .845 .937 r
F
.63442 .07239 m .63488 .0736 L .63251 .06754 L F
.63609 .07722 m .6326 .06776 L .63536 .07559 L .556 .59 .97 r
F
.63037 .06174 m .63195 .06612 L p .6326 .06776 L F P
0 g
s
.63195 .06612 m .63536 .07559 L p .6326 .06776 L .556 .59 .97 r
F P
0 g
s
.63246 .06753 m .63195 .06612 L p .63295 .06879 L .261 .384 .875 r
F P
0 g
s
.63591 .07672 m .63246 .06753 L .63295 .06879 L .261 .384 .875 r
F
.6356 .07624 m .63246 .06753 L p .63591 .07672 L F P
0 g
s
.63591 .07672 m .63195 .06612 L .63298 .06883 L .261 .384 .875 r
F
.62988 .06066 m .62991 .06061 L p .63397 .07142 L .997 .845 .937 r
F P
0 g
s
.63334 .06978 m .63105 .06363 L .63184 .06571 L .261 .384 .875 r
F
.63591 .07672 m .63298 .06883 L .63334 .06978 L F
.63298 .06883 m .63105 .06363 L .63334 .06978 L F
.6366 .07858 m .63712 .07937 L p .63612 .0776 L .516 .542 .939 r
F P
0 g
s
.6366 .07858 m .63659 .07858 L p .63657 .07843 L .345 .453 .917 r
F P
0 g
s
.6366 .07862 m .63651 .07841 L .63652 .07843 L .261 .384 .875 r
F
.6362 .07792 m .63601 .07738 L p .63661 .07864 L F P
0 g
s
.63664 .07871 m .63661 .07865 L .63661 .07864 L .261 .384 .875 r
F
.63661 .07865 m .6362 .07792 L .63661 .07864 L F
.63612 .0776 m .63632 .07796 L .6361 .07744 L .516 .542 .939 r
F
.63632 .07796 m .63712 .07937 L .6361 .07744 L F
.63675 .07868 m .63673 .07863 L .63674 .07866 L .556 .59 .97 r
F
.63684 .07877 m .63673 .07863 L .63675 .07868 L F
.63684 .07877 m .63661 .07831 L .63673 .07863 L F
.6361 .07744 m .63712 .07937 L .63608 .07725 L .516 .542 .939 r
F
.63651 .07841 m .63568 .07645 L .63652 .07843 L .261 .384 .875 r
F
.63616 .07792 m .63608 .07725 L .63605 .07732 L .516 .542 .939 r
F
.63601 .07738 m .63568 .07645 L p .63651 .07841 L .261 .384 .875 r
F P
0 g
s
.63568 .07645 m .6356 .07624 L p .63569 .07646 L .261 .384 .875 r
F P
0 g
s
.6366 .0786 m .6356 .07624 L .63591 .07672 L .261 .384 .875 r
F
.63664 .07871 m .6366 .07862 L .6366 .0786 L F
.6366 .07862 m .6356 .07624 L .6366 .0786 L F
.63664 .07871 m .63601 .07738 L .63651 .07841 L F
.63616 .07792 m .63659 .07858 L p .63612 .0776 L .516 .542 .939 r
F P
0 g
s
.63616 .07792 m .6361 .07779 L .63655 .07836 L .345 .453 .917 r
F
.63704 .07877 m .63706 .07883 L .63502 .06137 L .957 .827 .957 r
F
.63586 .07565 m .63706 .07883 L .63147 .06136 L F
.63796 .07884 m .63748 .07918 L .63845 .07986 L .947 .735 .845 r
F
.63784 .07884 m .63748 .07918 L .63796 .07884 L F
.64747 .07887 m .64779 .07603 L p .63854 .07884 L .957 .827 .957 r
F P
0 g
s
.63733 .07883 m .63718 .07883 L .63748 .07918 L .947 .735 .845 r
F
.63939 .08052 m .63748 .07918 L .64049 .0827 L F
.64446 .09114 m .6464 .09434 L p .64456 .09072 L .922 .726 .86 r
F P
0 g
s
.63855 .07883 m .64456 .09069 L p .63707 .07883 L .947 .735 .845 r
F P
0 g
s
.63707 .07883 m .64456 .09069 L .64446 .09112 L .947 .735 .845 r
F
.63706 .07883 m .6375 .07883 L .64942 .06142 L .957 .827 .957 r
F
.63705 .07883 m .63712 .07895 L .63707 .07883 L .947 .735 .845 r
F
.63689 .07856 m .63705 .07883 L .63707 .07883 L F
.63701 .07883 m .63694 .07869 L p .63707 .07883 L F P
0 g
s
.63706 .07883 m .63704 .07879 L p .63705 .07883 L .957 .827 .957 r
F P
0 g
s
.64428 .09075 m .64441 .09096 L p .64428 .09075 L .094 .094 .669 r
F P
0 g
s
.63727 .07921 m .64441 .09096 L .64441 .09097 L .094 .094 .669 r
F
.63704 .07878 m .64428 .09075 L p .63727 .07921 L F P
0 g
s
.63705 .07883 m .64428 .09075 L .63704 .07878 L .701 .636 .937 r
F
.64428 .09075 m .64441 .09096 L .64428 .09075 L F
.64428 .09075 m .64428 .09075 L .63704 .07878 L F
.64441 .09096 m .63704 .07878 L p .64441 .09096 L F P
0 g
s
.64446 .09114 m .63728 .07922 L p .64441 .09097 L .094 .094 .669 r
F P
0 g
s
.65104 .06122 m .64747 .07885 L .64976 .06954 L 1 1 .703 r
F
.65183 .06113 m .65104 .06122 L p .64975 .06958 L F P
0 g
s
.64779 .07603 m .64942 .06142 L p .6375 .07883 L .957 .827 .957 r
F P
0 g
s
.63854 .07884 m .64779 .07603 L .6375 .07883 L .957 .827 .957 r
F
.65104 .06122 m .64942 .06142 L .64779 .07603 L p 1 1 .703 r
F P
0 g
s
.64942 .06142 m .65182 .06113 L p .64446 .09112 L 1 1 .703 r
F P
0 g
s
.64446 .09114 m .64441 .09097 L .64441 .09097 L .094 .094 .669 r
F
.64446 .09114 m .64441 .09097 L .64446 .09114 L F
.63706 .07883 m .63705 .07883 L .63716 .07926 L .957 .827 .957 r
F
.63716 .07926 m .63707 .07904 L p .63714 .07924 L .846 .769 .986 r
F P
0 g
s
.63707 .07904 m .63699 .07883 L p .63707 .07906 L .846 .769 .986 r
F P
0 g
s
.63714 .07924 m .63707 .07904 L .63707 .07906 L .846 .769 .986 r
F
.63707 .07904 m .63716 .07926 L p .63704 .07879 L .957 .827 .957 r
F P
0 g
s
.63709 .07917 m .63728 .07949 L .63708 .079 L .701 .636 .937 r
F
.63707 .07904 m .63708 .079 L .63707 .07897 L F
.63709 .07917 m .63708 .079 L .63707 .07904 L F
.63716 .07926 m .63757 .08027 L p .63723 .07923 L .957 .827 .957 r
F P
0 g
s
.63757 .08027 m .64509 .09872 L p .63838 .08209 L .957 .827 .957 r
F P
0 g
s
.63723 .07923 m .63757 .08027 L .63838 .08209 L .957 .827 .957 r
F
.63734 .07958 m .63965 .08331 L p .63712 .07895 L .701 .636 .937 r
F P
0 g
s
.63871 .08284 m .64525 .09872 L .64535 .09782 L p .957 .827 .957 r
F P
0 g
s
.64362 .09454 m .64438 .09146 L p .63965 .08331 L .947 .735 .845 r
F P
0 g
s
.63712 .07895 m .63736 .07953 L .63965 .08331 L .947 .735 .845 r
F
.63965 .08331 m .63709 .07917 L p .63786 .08036 L .23 .385 .884 r
F P
0 g
s
.63709 .07917 m .63709 .07917 L .63785 .08034 L .23 .385 .884 r
F
.63712 .07895 m .63965 .08331 L .64446 .09112 L .947 .735 .845 r
F
.64443 .09122 m .64446 .09112 L p .63965 .08331 L F P
0 g
s
.64443 .09122 m .63701 .07883 L .64445 .09114 L .947 .735 .845 r
F
.63786 .08036 m .63785 .08034 L .63785 .08034 L .23 .385 .884 r
F
.63701 .07883 m .63716 .07921 L p .63712 .07895 L .947 .735 .845 r
F P
0 g
s
.63716 .07921 m .63727 .07938 L .63712 .07895 L .947 .735 .845 r
F
.63965 .08331 m .63966 .08333 L p .63705 .07883 L .701 .636 .937 r
F P
0 g
s
.63707 .07883 m .63712 .07895 L .64446 .09112 L .947 .735 .845 r
F
.63966 .08333 m .63705 .07883 L .63705 .07883 L .701 .636 .937 r
F
.63706 .07888 m .63712 .07895 L p .63705 .07883 L F P
0 g
s
.64593 .08513 m .65183 .06113 L p .64593 .08513 L .947 .735 .845 r
F P
0 g
s
.65182 .06113 m .65183 .06113 L .64446 .09112 L p 1 1 .703 r
F P
0 g
s
.65182 .06113 m .64446 .09112 L .64446 .09112 L 1 1 .703 r
F
.63701 .07883 m .63712 .07895 L .63705 .07883 L .947 .735 .845 r
F
.64438 .09146 m .64443 .09122 L p .63965 .08331 L F P
0 g
s
.64443 .09122 m .64438 .09146 L .64438 .09146 L 1 1 .703 r
F
.63727 .07921 m .63727 .07921 L .64441 .09097 L .094 .094 .669 r
F
.65182 .06113 m .64446 .09112 L .64446 .09112 L 1 1 .703 r
F
.63727 .07921 m .64441 .09097 L .64441 .09097 L .094 .094 .669 r
F
.64289 .08855 m .64374 .08993 L .64652 .09454 L .947 .735 .845 r
F
.63701 .07883 m .64374 .08993 L .64289 .08855 L F
.63701 .07883 m .63707 .07883 L .64374 .08993 L F
.63966 .08333 m .64109 .08552 L p .63705 .07883 L .701 .636 .937 r
F P
0 g
s
.63727 .07921 m .64428 .09075 L .64428 .09075 L .094 .094 .669 r
F
.64439 .09091 m .64196 .08684 L .64433 .09047 L .23 .385 .884 r
F
.6443 .09063 m .64433 .09047 L p .64446 .09112 L 1 1 .703 r
F P
0 g
s
.63728 .07922 m .63727 .07921 L .64441 .09097 L .094 .094 .669 r
F
.64456 .09069 m .64652 .09454 L p .64446 .09112 L .947 .735 .845 r
F P
0 g
s
.64441 .09097 m .64441 .09097 L .64566 .09302 L .094 .094 .669 r
F
.63728 .07922 m .64441 .09097 L .64441 .09097 L F
.6444 .091 m .64441 .09096 L .64359 .08961 L .701 .636 .937 r
F
.64109 .08552 m .64128 .08581 L .63705 .07883 L F
.6444 .091 m .63966 .08333 L p .64196 .08684 L .23 .385 .884 r
F P
0 g
s
.6444 .091 m .64196 .08684 L .64439 .09091 L .23 .385 .884 r
F
.64433 .09047 m .64942 .06142 L p .64448 .09102 L 1 1 .703 r
F P
0 g
s
.64448 .09102 m .64446 .09112 L .64433 .09047 L 1 1 .703 r
F
.64593 .08202 m .64566 .09302 L .64433 .09047 L p .107 .11 .687 r
F P
0 g
s
.64626 .08029 m .64634 .07977 L p .64433 .09047 L .003 .003 .003 r
F P
0 g
s
.64634 .07977 m .64942 .06142 L .64656 .07773 L p .003 .003 .003 r
F P
0 g
s
.64656 .07773 m .64433 .09047 L p .64634 .07977 L .003 .003 .003 r
F P
0 g
s
.65131 .05282 m .65361 .04065 L .62868 .05282 L closepath p .607 .971 1 r
F P
0 g
s
.64942 .06142 m .64626 .08029 L p .6479 .07092 L .003 .003 .003 r
F P
0 g
s
.64919 .06358 m .64925 .06303 L .6479 .07092 L .003 .003 .003 r
F
.64925 .06303 m .64942 .06142 L .6479 .07092 L F
.64653 .07885 m .65169 .04938 L p .64659 .0784 L F P
0 g
s
.64659 .0784 m .64626 .08029 L .64653 .07885 L .003 .003 .003 r
F
.64763 .0827 m .64566 .09302 L .646 .0792 L p .107 .11 .687 r
F P
0 g
s
.63438 .07241 m .63463 .07318 L .63397 .07142 L .997 .845 .937 r
F
.6358 .07577 m .63087 .06315 L .63105 .06363 L .957 .827 .957 r
F
.63349 .07003 m .63037 .06174 L .6326 .06776 L .556 .59 .97 r
F
.62991 .06061 m .63508 .07405 L .63442 .07239 L .997 .845 .937 r
F
.63147 .06136 m .63045 .06135 L p .63586 .07565 L .957 .827 .957 r
F P
0 g
s
.63045 .06135 m .63019 .06135 L p .63335 .06942 L .957 .827 .957 r
F P
0 g
s
.63087 .06315 m .63176 .06541 L .63019 .06135 L .957 .827 .957 r
F
.63176 .06541 m .63335 .06942 L .6311 .06366 L F
.63022 .06135 m .63442 .07239 L .63021 .06139 L .078 .078 .537 r
F
.62991 .06061 m .63022 .06135 L p .63021 .06139 L F P
0 g
s
.63061 .0408 m .62868 .05282 L .64902 .04289 L p .607 .971 1 r
F P
0 g
s
.62868 .05282 m .63063 .04065 L .62704 .05278 L closepath p .595 .96 1 r
F P
0 g
s
.62897 .04061 m .62704 .05278 L .63059 .04078 L p .595 .96 1 r
F P
0 g
s
.63059 .04078 m .63063 .04065 L .62917 .04061 L p .595 .96 1 r
F P
0 g
s
.62917 .04061 m .62897 .04061 L p .63059 .04078 L .595 .96 1 r
F P
0 g
s
.62648 .05269 m .6284 .04053 L .62549 .05116 L closepath p .118 .216 .773 r
F P
0 g
s
.6274 .03897 m .62549 .05116 L .6284 .04053 L closepath p .118 .216 .773 r
F P
0 g
s
.62704 .05278 m .62897 .04061 L .62648 .05269 L closepath p .527 .899 1 r
F P
0 g
s
.6284 .04053 m .62648 .05269 L .62897 .04061 L closepath p .527 .899 1 r
F P
0 g
s
.62931 .0407 m .62763 .04051 L .62838 .04052 L p .107 .007 .007 r
F P
0 g
s
.62838 .04052 m .6284 .04053 L .62843 .04053 L p .107 .007 .007 r
F P
0 g
s
.62843 .04053 m .62931 .0407 L p .62838 .04052 L .107 .007 .007 r
F P
0 g
s
.63099 .0409 m .62931 .0407 L .62898 .04061 L p .101 .005 .005 r
F P
0 g
s
.62898 .04061 m .62897 .04061 L .62901 .04061 L p .101 .005 .005 r
F P
0 g
s
.62901 .04061 m .63099 .0409 L p .62898 .04061 L .101 .005 .005 r
F P
0 g
s
.63063 .04065 m .63061 .0408 L p .65266 .04111 L .607 .971 1 r
F P
0 g
s
.64902 .04289 m .65266 .04111 L p .63061 .0408 L .607 .971 1 r
F P
0 g
s
.65266 .04111 m .65361 .04065 L .63063 .04065 L p .607 .971 1 r
F P
0 g
s
.63063 .04065 m .63099 .0409 L .62901 .04061 L p .102 .022 .022 r
F P
0 g
s
.62901 .04061 m .62897 .04061 L .63063 .04065 L p .102 .022 .022 r
F P
0 g
s
.62991 .06061 m .63189 .04855 L .63089 .05723 L p .095 .095 .593 r
F P
0 g
s
.64942 .06142 m .65121 .05189 L p .63024 .06118 L .601 .968 1 r
F P
0 g
s
.64942 .06142 m .63024 .06118 L .63024 .06119 L .601 .968 1 r
F
.63022 .06135 m .62991 .06061 L p .63089 .05723 L .095 .095 .593 r
F P
0 g
s
.63022 .06135 m .64942 .06142 L p .63024 .06119 L .601 .968 1 r
F P
0 g
s
.63024 .06119 m .63022 .06119 L .63022 .06135 L .601 .968 1 r
F
.62959 .06104 m .62983 .06072 L p .63008 .06139 L .997 .845 .937 r
F P
0 g
s
.63438 .07241 m .63534 .07505 L .63463 .07318 L .997 .845 .937 r
F
.63653 .07791 m .63529 .07488 L p .63675 .07863 L .957 .827 .957 r
F P
0 g
s
.63509 .07409 m .63577 .0762 L .63438 .07241 L .997 .845 .937 r
F
.63552 .0755 m .63555 .07558 L p .63517 .07458 L .261 .384 .875 r
F P
0 g
s
.63555 .07558 m .63664 .07871 L p .63494 .0741 L .261 .384 .875 r
F P
0 g
s
.63534 .07505 m .63494 .0741 L .63334 .06978 L .261 .384 .875 r
F
.63555 .07558 m .63494 .0741 L .63534 .07505 L F
.6366 .07858 m .63555 .07558 L p .63557 .07646 L .46 .526 .947 r
F P
0 g
s
.63008 .06139 m .63397 .07142 L .62973 .06137 L .997 .845 .937 r
F
.63508 .07566 m .63539 .07609 L p .63593 .07757 L .345 .453 .917 r
F P
0 g
s
.636 .07695 m .63639 .07749 L p .63623 .07751 L .345 .453 .917 r
F P
0 g
s
.63609 .07754 m .63639 .07749 L .63639 .07815 L .345 .453 .917 r
F
.63639 .07749 m .636 .07695 L p .63668 .07818 L .426 .494 .928 r
F P
0 g
s
.63539 .07609 m .636 .07695 L p .63609 .07754 L .345 .453 .917 r
F P
0 g
s
.636 .07695 m .63537 .07606 L p .63611 .07725 L .426 .494 .928 r
F P
0 g
s
.63668 .07818 m .636 .07695 L .63611 .07725 L .426 .494 .928 r
F
.63672 .07838 m .6369 .07877 L .63506 .07422 L .846 .769 .986 r
F
.63352 .07041 m .63508 .07544 L p .63555 .07559 L .46 .526 .947 r
F P
0 g
s
.63593 .07757 m .63539 .07609 L .63609 .07754 L .345 .453 .917 r
F
.63557 .07646 m .63539 .07609 L .63559 .07709 L .46 .526 .947 r
F
.63593 .07757 m .62922 .06742 L .63508 .07566 L p .345 .453 .917 r
F P
0 g
s
.63537 .07606 m .62922 .06742 L p .63535 .07693 L .426 .494 .928 r
F P
0 g
s
.63539 .07612 m .63537 .07606 L .63535 .07693 L .426 .494 .928 r
F
.63508 .07544 m .63538 .07641 L p .63539 .07609 L .46 .526 .947 r
F P
0 g
s
.63538 .07641 m .63559 .07709 L p .63539 .07609 L .46 .526 .947 r
F P
0 g
s
.63664 .07871 m .64151 .09264 L .6362 .07792 L p .261 .384 .875 r
F P
0 g
s
.63559 .07709 m .63669 .08065 L p .63664 .07871 L .46 .526 .947 r
F P
0 g
s
.63669 .08065 m .63869 .08707 L p .63786 .08369 L .46 .526 .947 r
F P
0 g
s
.63664 .07871 m .63669 .08065 L .63786 .08369 L .46 .526 .947 r
F
.63576 .07765 m .63538 .07641 L p .63646 .07878 L .365 .457 .914 r
F P
0 g
s
.63658 .07905 m .63647 .07886 L .63646 .07878 L .365 .457 .914 r
F
.63647 .07886 m .63576 .07765 L .63646 .07878 L F
.63672 .07838 m .63506 .07422 L .63514 .07446 L .846 .769 .986 r
F
.63532 .07496 m .63672 .07838 L p .63514 .07446 L F P
0 g
s
.63603 .07687 m .63675 .07863 L p .63529 .07488 L .957 .827 .957 r
F P
0 g
s
.63716 .07921 m .63701 .07883 L p .63833 .08104 L .947 .735 .845 r
F P
0 g
s
.63716 .07921 m .63833 .08104 L .63966 .08333 L .947 .735 .845 r
F
.63706 .07913 m .63706 .07912 L .63782 .0803 L .23 .385 .884 r
F
.63706 .07913 m .63782 .0803 L .63785 .08034 L F
.63716 .07921 m .63744 .07985 L p .63727 .07938 L .947 .735 .845 r
F P
0 g
s
.63335 .06942 m .63706 .07883 L .63045 .06135 L .957 .827 .957 r
F
.63532 .07494 m .63552 .0755 L p .63517 .07458 L .261 .384 .875 r
F P
0 g
s
.63681 .07834 m .63419 .07194 L .63673 .07819 L .957 .827 .957 r
F
.63442 .07239 m .63508 .07405 L .63488 .0736 L .997 .845 .937 r
F
.6358 .07577 m .63382 .07097 L p .63681 .07834 L .957 .827 .957 r
F P
0 g
s
.63433 .07232 m .63517 .07458 L .63185 .06574 L .261 .384 .875 r
F
.63552 .0755 m .63529 .07488 L .63517 .07458 L F
.63529 .07488 m .63517 .07458 L .63517 .07458 L F
.63534 .07505 m .63577 .0762 L .63463 .07318 L .997 .845 .937 r
F
.63513 .07449 m .63532 .07496 L p .63501 .07419 L .846 .769 .986 r
F P
0 g
s
.63535 .07505 m .63334 .06978 L .63184 .06571 L .261 .384 .875 r
F
.63555 .07558 m .63545 .07532 L .63535 .07505 L F
.63545 .07532 m .63334 .06978 L .63535 .07505 L F
.63532 .07496 m .63514 .07446 L .63493 .07393 L .846 .769 .986 r
F
.63525 .07479 m .63493 .07393 L .63443 .0727 L F
.63532 .07496 m .63493 .07393 L .63525 .07479 L F
.6345 .07292 m .63352 .07041 L .63331 .06987 L .997 .845 .937 r
F
.63463 .07318 m .63504 .07442 L .63397 .07142 L F
.63537 .07513 m .6346 .07313 L .63352 .07041 L .46 .526 .947 r
F
.63555 .07559 m .63508 .07544 L .63557 .07646 L F
.63593 .07757 m .63609 .07754 L .63639 .07815 L .345 .453 .917 r
F
.63659 .07858 m .63659 .07858 L .63612 .0776 L .516 .542 .939 r
F
.63616 .07792 m .63655 .07836 L .63657 .07843 L .345 .453 .917 r
F
.63659 .07858 m .63616 .07792 L p .63645 .07829 L F P
0 g
s
.63616 .07792 m .63593 .07757 L p .6361 .07779 L .345 .453 .917 r
F P
0 g
s
.63659 .07857 m .63557 .07646 L .63559 .07709 L .46 .526 .947 r
F
.63568 .077 m .63639 .07793 L .63537 .07606 L .426 .494 .928 r
F
.63609 .07754 m .636 .07695 L .63623 .07751 L .345 .453 .917 r
F
.63639 .07793 m .63625 .07755 L .63537 .07606 L .426 .494 .928 r
F
.63538 .07641 m .63352 .07041 L p .63629 .07818 L .365 .457 .914 r
F P
0 g
s
.63501 .07419 m .63443 .0727 L .63391 .0714 L .846 .769 .986 r
F
.63416 .07206 m .63391 .0714 L .62986 .06135 L F
.63397 .07142 m .6345 .07292 L .63331 .06987 L .997 .845 .937 r
F
.63501 .07419 m .63391 .0714 L .63416 .07206 L .846 .769 .986 r
F
.63537 .07513 m .63352 .07041 L .63555 .07559 L .46 .526 .947 r
F
.63577 .0762 m .63504 .07442 L p .63463 .07318 L .997 .845 .937 r
F P
0 g
s
.63532 .07496 m .63443 .0727 L .63501 .07419 L .846 .769 .986 r
F
.63555 .07558 m .63532 .07494 L p .6346 .07313 L .46 .526 .947 r
F P
0 g
s
.63629 .07818 m .63352 .07041 L .63455 .07313 L .365 .457 .914 r
F
.63498 .07431 m .63455 .07313 L .63488 .074 L F
.63629 .07818 m .63455 .07313 L .63498 .07431 L F
.63598 .07789 m .63539 .07612 L .63535 .07693 L .426 .494 .928 r
F
.63664 .07871 m .6366 .07863 L .6366 .07862 L .261 .384 .875 r
F
.6366 .07863 m .63651 .07841 L .6366 .07862 L F
.63664 .07871 m .6366 .07858 L p .63559 .07709 L .46 .526 .947 r
F P
0 g
s
.63658 .07905 m .63538 .07641 L .63629 .07818 L .365 .457 .914 r
F
.63598 .07789 m .63641 .07842 L p .63568 .077 L .426 .494 .928 r
F P
0 g
s
.62922 .06742 m .63168 .07268 L .63598 .07789 L p .426 .494 .928 r
F P
0 g
s
.63641 .07842 m .63534 .07712 L p .6358 .07709 L .39 .478 .926 r
F P
0 g
s
.63659 .07858 m .63557 .07646 L .63659 .07857 L .46 .526 .947 r
F
.63641 .07842 m .63644 .07845 L p .63568 .077 L .426 .494 .928 r
F P
0 g
s
.63644 .07845 m .63671 .07878 L p .63659 .07858 L .426 .494 .928 r
F P
0 g
s
.6366 .07858 m .63659 .07858 L .63659 .07857 L .46 .526 .947 r
F
.63644 .07845 m .63659 .07858 L .63568 .077 L .426 .494 .928 r
F
.63439 .07463 m .63477 .0749 L p .63534 .07712 L .39 .478 .926 r
F P
0 g
s
.63615 .07737 m .63611 .07725 L .63537 .07606 L .426 .494 .928 r
F
.63657 .07843 m .63657 .07842 L .63658 .07845 L .345 .453 .917 r
F
.63657 .07842 m .63655 .07836 L .63658 .07845 L F
.63659 .07858 m .6366 .07858 L p .63659 .07858 L .516 .542 .939 r
F P
0 g
s
.63659 .07858 m .63645 .07829 L .63657 .07843 L .345 .453 .917 r
F
.63671 .07878 m .63642 .07796 L .63568 .077 L .426 .494 .928 r
F
.63577 .07597 m .63658 .07905 L p .63532 .07532 L .365 .457 .914 r
F P
0 g
s
.63447 .07304 m .63462 .07459 L p .63477 .07485 L .408 .49 .931 r
F P
0 g
s
.63611 .07723 m .63467 .07355 L .63477 .07485 L .408 .49 .931 r
F
.63467 .07355 m .63447 .07304 L .63477 .07485 L F
.63477 .0749 m .63534 .07532 L p .63534 .07612 L .39 .478 .926 r
F P
0 g
s
.63534 .07712 m .63477 .0749 L .63534 .07612 L .39 .478 .926 r
F
.63534 .07712 m .63168 .07268 L .63439 .07463 L p F P
0 g
s
.63477 .0749 m .63168 .07268 L .63512 .07982 L p .396 .479 .925 r
F P
0 g
s
.63611 .07723 m .63462 .07459 L .63586 .07722 L .408 .49 .931 r
F
.63534 .07712 m .63534 .07532 L .63603 .07707 L .39 .478 .926 r
F
.63462 .07459 m .63487 .07716 L p .63586 .07722 L .408 .49 .931 r
F P
0 g
s
.63487 .07716 m .63512 .07982 L p .63583 .07795 L .408 .49 .931 r
F P
0 g
s
.63611 .07723 m .63487 .07716 L .63583 .07795 L .408 .49 .931 r
F
.63641 .07839 m .63639 .07833 L p .63639 .07834 L F P
0 g
s
.63639 .07833 m .63611 .07723 L p .63639 .07834 L .408 .49 .931 r
F P
0 g
s
.63639 .07834 m .63611 .07723 L .636 .07752 L .408 .49 .931 r
F
.63641 .07839 m .6358 .07709 L .63598 .07707 L .39 .478 .926 r
F
.63517 .07458 m .63517 .07458 L .63185 .06574 L .261 .384 .875 r
F
.63176 .06541 m .6311 .06366 L .63019 .06135 L .957 .827 .957 r
F
.63382 .07097 m .63185 .06574 L .63184 .06571 L .261 .384 .875 r
F
.63239 .06705 m .63442 .07239 L p .63022 .06135 L .078 .078 .537 r
F P
0 g
s
.62983 .06072 m .62988 .06066 L p .63195 .06612 L .997 .845 .937 r
F P
0 g
s
.63195 .06612 m .63105 .06363 L p .63298 .06883 L .261 .384 .875 r
F P
0 g
s
.6345 .07292 m .63504 .07442 L .63352 .07041 L .997 .845 .937 r
F
.63352 .07041 m .63337 .06993 L p .63488 .074 L .365 .457 .914 r
F P
0 g
s
.63016 .06135 m .63014 .06135 L p .6358 .07577 L .957 .827 .957 r
F P
0 g
s
.63135 .06457 m .6358 .07577 L .63014 .06135 L .957 .827 .957 r
F
.63433 .07232 m .63185 .06574 L .63382 .07097 L .261 .384 .875 r
F
.63195 .06612 m .63397 .07142 L .62983 .06072 L .997 .845 .937 r
F
.63442 .07239 m .63239 .06706 L p .6338 .07081 L .556 .59 .97 r
F P
0 g
s
.63248 .0677 m .63307 .06915 L p .63007 .06135 L .957 .827 .957 r
F P
0 g
s
.63307 .06915 m .6358 .07577 L p .63135 .06457 L .957 .827 .957 r
F P
0 g
s
.63307 .06915 m .63135 .06457 L .63007 .06135 L .957 .827 .957 r
F
.637 .0789 m .6368 .07876 L .63696 .07868 L .997 .845 .937 r
F
.6368 .07876 m .63694 .07869 L .63697 .07867 L F
.6369 .07877 m .63696 .07881 L .63445 .07265 L .846 .769 .986 r
F
.6369 .07877 m .63445 .07265 L .63466 .0732 L F
.6369 .07877 m .63291 .06885 L .63305 .06923 L F
.63331 .06987 m .63381 .07132 L .63242 .06775 L .997 .845 .937 r
F
.6342 .07209 m .63137 .06506 L .63155 .06554 L .846 .769 .986 r
F
.6369 .07877 m .63411 .07184 L .6342 .07209 L F
.63337 .06993 m .63352 .07041 L p .6346 .07313 L .46 .526 .947 r
F P
0 g
s
.63411 .07184 m .63137 .06506 L .6342 .07209 L .846 .769 .986 r
F
.63514 .07446 m .63506 .07422 L .63302 .06917 L F
.63312 .06944 m .63302 .06917 L .62986 .06135 L F
.63514 .07446 m .63302 .06917 L .63312 .06944 L F
.63529 .07488 m .63513 .07449 L p .63552 .0755 L .957 .827 .957 r
F P
0 g
s
.63555 .07558 m .63528 .07489 L .63537 .07513 L .46 .526 .947 r
F
.63528 .07489 m .63513 .07449 L .63537 .07513 L F
.63513 .07449 m .63384 .07138 L p .63552 .0755 L .957 .827 .957 r
F P
0 g
s
.63242 .06775 m .63381 .07132 L .62959 .06104 L .997 .845 .937 r
F
.63384 .07138 m .62959 .06104 L p .63403 .07187 L .957 .827 .957 r
F P
0 g
s
.63331 .06987 m .63455 .07313 L .63381 .07132 L .997 .845 .937 r
F
.63384 .07138 m .63403 .07187 L .63552 .0755 L .957 .827 .957 r
F
.63448 .07309 m .63447 .07304 L p .63642 .0783 L .396 .48 .925 r
F P
0 g
s
.63438 .07208 m .63442 .07247 L .63178 .06482 L .365 .457 .914 r
F
.63382 .07097 m .63419 .07194 L .63681 .07834 L .957 .827 .957 r
F
.63696 .07881 m .63444 .0726 L .63445 .07265 L .846 .769 .986 r
F
.63433 .07232 m .63673 .07819 L p .63419 .07194 L .957 .827 .957 r
F P
0 g
s
.63696 .07881 m .63433 .07232 L .63444 .0726 L .846 .769 .986 r
F
.63577 .07597 m .63442 .07247 L .63438 .07208 L .365 .457 .914 r
F
.63532 .07494 m .63517 .07458 L .63433 .07232 L .261 .384 .875 r
F
.63555 .07558 m .63537 .07513 L .63555 .07559 L .46 .526 .947 r
F
.63577 .07597 m .63548 .07554 L .63442 .07247 L .365 .457 .914 r
F
.63532 .07494 m .63356 .07041 L .6346 .07313 L .46 .526 .947 r
F
.63504 .07442 m .63446 .07299 L p .63455 .07313 L .997 .845 .937 r
F P
0 g
s
.63446 .07299 m .62959 .06104 L p .63397 .07171 L .997 .845 .937 r
F P
0 g
s
.63455 .07313 m .63446 .07299 L .63397 .07171 L .997 .845 .937 r
F
.63447 .07304 m .63447 .07304 L p .63628 .07778 L .396 .48 .925 r
F P
0 g
s
.63398 .07183 m .63446 .07299 L p .6334 .07034 L .957 .827 .957 r
F P
0 g
s
.63446 .07299 m .63603 .07687 L p .6342 .07232 L .957 .827 .957 r
F P
0 g
s
.6334 .07034 m .63446 .07299 L .6342 .07232 L .957 .827 .957 r
F
.63548 .07554 m .63544 .07549 L .6347 .07347 L .365 .457 .914 r
F
.63544 .07549 m .63532 .07532 L .63487 .07398 L F
.63544 .07549 m .63487 .07398 L .6347 .07347 L F
.63552 .0755 m .63461 .07329 L .63603 .07687 L .957 .827 .957 r
F
.63461 .07329 m .6349 .07405 L .63603 .07687 L F
.63532 .07494 m .63433 .07232 L .6337 .07061 L .261 .384 .875 r
F
.63528 .07489 m .6346 .07313 L .63513 .07449 L .46 .526 .947 r
F
.6337 .07061 m .63184 .06571 L .63037 .06174 L .261 .384 .875 r
F
.63433 .07232 m .63184 .06571 L .6337 .07061 L F
.63356 .07041 m .63337 .06993 L .6346 .07313 L .46 .526 .947 r
F
.63466 .0732 m .63445 .07265 L .63291 .06885 L .846 .769 .986 r
F
.63382 .07097 m .63307 .06915 L p .63419 .07194 L .957 .827 .957 r
F P
0 g
s
.63307 .06915 m .63287 .06865 L p .63315 .06935 L .957 .827 .957 r
F P
0 g
s
.63307 .06915 m .63315 .06935 L .63419 .07194 L .957 .827 .957 r
F
.63337 .06993 m .63337 .06993 L p .6347 .07347 L .365 .457 .914 r
F P
0 g
s
.63694 .07869 m .6368 .07875 L .63701 .07865 L .997 .845 .937 r
F
.63699 .07883 m .63433 .07232 L p .63696 .07881 L .846 .769 .986 r
F P
0 g
s
.63022 .06135 m .63037 .06174 L .63239 .06705 L p .078 .078 .537 r
F P
0 g
s
.63239 .06705 m .63037 .06174 L p .63184 .06564 L .556 .59 .97 r
F P
0 g
s
.63037 .06174 m .63178 .06482 L .63532 .07494 L p .261 .384 .875 r
F P
0 g
s
.63178 .06482 m .63439 .07255 L p .63532 .07494 L .46 .526 .947 r
F P
0 g
s
.63442 .07279 m .63513 .07449 L p .6342 .07219 L .846 .769 .986 r
F P
0 g
s
.63287 .06865 m .63335 .06989 L .63419 .07194 L .957 .827 .957 r
F
.63445 .07265 m .6313 .06489 L .63137 .06506 L .846 .769 .986 r
F
.63444 .0726 m .6313 .06489 L .63445 .07265 L F
.62986 .06135 m .63433 .07232 L p .63419 .07194 L .957 .827 .957 r
F P
0 g
s
.63433 .07232 m .63347 .0702 L p .63444 .0726 L .846 .769 .986 r
F P
0 g
s
.63347 .0702 m .6313 .06489 L p .63354 .07039 L .846 .769 .986 r
F P
0 g
s
.63444 .0726 m .63347 .0702 L .63354 .07039 L .846 .769 .986 r
F
.63513 .07449 m .63501 .07419 L .63416 .07206 L F
.6342 .07219 m .63416 .07206 L .62986 .06135 L F
.63513 .07449 m .63416 .07206 L .6342 .07219 L F
.63488 .074 m .63364 .07068 L .6347 .07347 L .365 .457 .914 r
F
.63364 .07068 m .63337 .06993 L .63347 .0702 L F
.63364 .07068 m .63347 .0702 L .6347 .07347 L F
.6347 .07347 m .63347 .0702 L .63178 .06482 L F
.63337 .06993 m .63336 .0699 L p .63347 .0702 L F P
0 g
s
.63336 .0699 m .63178 .06482 L p .63347 .0702 L .365 .457 .914 r
F P
0 g
s
.6347 .07347 m .63337 .06993 L .6347 .07347 L .365 .457 .914 r
F
.63548 .07554 m .6347 .07347 L .63178 .06482 L F
.63178 .06482 m .63417 .06986 L .63577 .07597 L p F P
0 g
s
.63611 .07723 m .63417 .06986 L p .63447 .07304 L .408 .49 .931 r
F P
0 g
s
.63502 .06137 m .63147 .06136 L p .63704 .07877 L .957 .827 .957 r
F P
0 g
s
.63704 .07878 m .6358 .07577 L p .63681 .07834 L .957 .827 .957 r
F P
0 g
s
.63534 .07532 m .63759 .07693 L p .63603 .07707 L .39 .478 .926 r
F P
0 g
s
.63569 .07557 m .63537 .07534 L p .63624 .07763 L .396 .479 .925 r
F P
0 g
s
.63447 .07304 m .63417 .06986 L p .63626 .07769 L .396 .48 .925 r
F P
0 g
s
.63537 .07534 m .6353 .07528 L p .63647 .07824 L .396 .479 .925 r
F P
0 g
s
.63628 .07778 m .63447 .07304 L .63626 .07769 L .396 .48 .925 r
F
.63461 .07329 m .6334 .07034 L .6349 .07405 L .957 .827 .957 r
F
.63701 .07876 m .63697 .07867 L .63707 .07883 L .947 .735 .845 r
F
.63696 .07868 m .63697 .07867 L .63701 .07876 L F
.63704 .07878 m .63695 .07863 L p .63697 .07869 L .701 .636 .937 r
F P
0 g
s
.63702 .07907 m .63706 .07913 L p .63695 .07863 L .701 .636 .937 r
F P
0 g
s
.6369 .07877 m .63736 .0798 L .63696 .07881 L .846 .769 .986 r
F
.63706 .07913 m .63709 .07917 L p .63707 .07909 L .701 .636 .937 r
F P
0 g
s
.63706 .07913 m .63707 .07909 L .63695 .07863 L .701 .636 .937 r
F
.63707 .07908 m .63699 .07883 L .63696 .07881 L .846 .769 .986 r
F
.63714 .07924 m .63707 .07906 L .63707 .07908 L F
.63707 .07906 m .63699 .07883 L .63707 .07908 L F
.63694 .07869 m .63696 .07868 L .63701 .07876 L .947 .735 .845 r
F
.63705 .07883 m .63704 .07878 L .63697 .07869 L .701 .636 .937 r
F
.63673 .07819 m .63707 .07904 L p .63704 .07878 L .957 .827 .957 r
F P
0 g
s
.63661 .07809 m .64384 .08953 L p .63756 .07986 L .23 .385 .884 r
F P
0 g
s
.63966 .08333 m .63833 .08104 L .64443 .09122 L .947 .735 .845 r
F
.63785 .08034 m .63782 .0803 L .64384 .08953 L .23 .385 .884 r
F
.63703 .07905 m .63694 .07883 L .63756 .07986 L F
.63694 .07883 m .63661 .07809 L .63756 .07986 L F
.63786 .08036 m .63785 .08034 L .64384 .08953 L F
.63786 .08036 m .63785 .08034 L .63786 .08036 L F
.63709 .07917 m .63785 .08034 L .63786 .08036 L F
.63736 .07953 m .63762 .08016 L .63965 .08331 L .947 .735 .845 r
F
.63762 .08016 m .63764 .08021 L .63778 .08043 L F
.63762 .08016 m .63778 .08043 L .63965 .08331 L F
.63966 .08333 m .63965 .08331 L p .63786 .08036 L .23 .385 .884 r
F P
0 g
s
.63965 .08331 m .63786 .08036 L .63786 .08036 L .23 .385 .884 r
F
.63708 .07916 m .63703 .07905 L .64384 .08953 L F
.64441 .09133 m .63716 .07921 L .64443 .09122 L .947 .735 .845 r
F
.63764 .08021 m .64362 .09454 L .63993 .08412 L F
.63764 .08021 m .63993 .08412 L .63965 .08331 L F
.64509 .09872 m .64525 .09872 L .63871 .08284 L p .957 .827 .957 r
F P
0 g
s
.63871 .08284 m .63838 .08209 L .64509 .09872 L .957 .827 .957 r
F
.63744 .07985 m .64362 .09454 L p .63762 .08016 L .947 .735 .845 r
F P
0 g
s
.64362 .09454 m .63744 .07985 L p .64441 .09133 L .947 .735 .845 r
F P
0 g
s
.63744 .07985 m .63762 .08016 L .63712 .07895 L .947 .735 .845 r
F
.63744 .07985 m .63716 .07921 L p .64441 .09133 L F P
0 g
s
.63709 .07917 m .63706 .07913 L p .63709 .07917 L .23 .385 .884 r
F P
0 g
s
.63704 .07917 m .63699 .07922 L .637 .0789 L .997 .845 .937 r
F
.63702 .07903 m .63701 .07901 L .63702 .07902 L .23 .385 .884 r
F
.63702 .07903 m .63702 .07902 L .63703 .07905 L F
.63706 .07912 m .63702 .07903 L .63703 .07905 L F
.63706 .07912 m .63703 .07905 L .63708 .07916 L F
.63699 .07922 m .64494 .09872 L p .63706 .0792 L .957 .827 .957 r
F P
0 g
s
.63702 .07907 m .63698 .07896 L .63701 .07901 L .23 .385 .884 r
F
.63702 .07907 m .63701 .07901 L .63706 .07912 L F
.63706 .07913 m .63702 .07907 L p .63706 .07912 L F P
0 g
s
.63656 .07801 m .63661 .07809 L p .63701 .07901 L .23 .385 .884 r
F P
0 g
s
.63701 .07901 m .63661 .07809 L .63703 .07905 L .23 .385 .884 r
F
.63698 .07896 m .63685 .07868 L .63685 .07868 L F
.63698 .07896 m .63656 .07801 L .63701 .07901 L F
.63672 .07838 m .64181 .09074 L p .63736 .0798 L .846 .769 .986 r
F P
0 g
s
.64384 .08953 m .63661 .07809 L p .63705 .07919 L .516 .542 .939 r
F P
0 g
s
.63966 .08333 m .63786 .08036 L .64384 .08953 L .23 .385 .884 r
F
.63741 .07981 m .64384 .08953 L p .63705 .07919 L .516 .542 .939 r
F P
0 g
s
.64113 .08414 m .64384 .08953 L .63741 .07981 L p .345 .453 .917 r
F P
0 g
s
.63712 .07937 m .63741 .07981 L p .63705 .07919 L .516 .542 .939 r
F P
0 g
s
.63741 .07981 m .63712 .07937 L p .6372 .07947 L .345 .453 .917 r
F P
0 g
s
.63741 .07981 m .6372 .07947 L .64113 .08414 L .345 .453 .917 r
F
.63661 .07809 m .63661 .07809 L p .63661 .07809 L .516 .542 .939 r
F P
0 g
s
.63712 .07937 m .63661 .07809 L .63661 .07809 L .516 .542 .939 r
F
.6368 .07875 m .63675 .07863 L p .63672 .07853 L .997 .845 .937 r
F P
0 g
s
.63675 .07863 m .6368 .07875 L p .63653 .07791 L .957 .827 .957 r
F P
0 g
s
.63661 .07809 m .63657 .07802 L p .63712 .07937 L .516 .542 .939 r
F P
0 g
s
.63712 .07937 m .63657 .07802 L .63658 .07806 L .516 .542 .939 r
F
.63639 .07749 m .63812 .07992 L p .63655 .07836 L .345 .453 .917 r
F P
0 g
s
.63812 .07992 m .64113 .08414 L p .63913 .08161 L .345 .453 .917 r
F P
0 g
s
.63688 .07876 m .63688 .07876 L p .6368 .07876 L .957 .827 .957 r
F P
0 g
s
.6368 .07875 m .6368 .07876 L p .63688 .07876 L .957 .827 .957 r
F P
0 g
s
.63706 .0792 m .63688 .07876 L p .63699 .07922 L .957 .827 .957 r
F P
0 g
s
.63712 .07937 m .63655 .078 L .63675 .07862 L .516 .542 .939 r
F
.63655 .07836 m .63812 .07992 L .63913 .08161 L .345 .453 .917 r
F
.63679 .07845 m .63668 .07818 L .63611 .07725 L .426 .494 .928 r
F
.63637 .07807 m .63598 .07707 L .63624 .07705 L .39 .478 .926 r
F
.63647 .07824 m .63569 .07557 L .63624 .07763 L .396 .479 .925 r
F
.63679 .07845 m .63611 .07725 L .63615 .07737 L .426 .494 .928 r
F
.63639 .07815 m .63639 .07749 L .63655 .07836 L .345 .453 .917 r
F
.63645 .07757 m .63639 .07749 L p .63668 .07818 L .426 .494 .928 r
F P
0 g
s
.63658 .07845 m .63655 .07836 L .64113 .08414 L .345 .453 .917 r
F
.63712 .07937 m .6366 .07858 L p .63694 .0789 L F P
0 g
s
.6366 .07858 m .63657 .07843 L .63694 .0789 L .345 .453 .917 r
F
.63712 .07937 m .63694 .0789 L .64113 .08414 L F
.63697 .0791 m .63688 .07886 L p .63695 .07908 L .556 .59 .97 r
F P
0 g
s
.63697 .0791 m .63695 .07908 L .63812 .08254 L .556 .59 .97 r
F
.6393 .08522 m .63697 .0791 L p .63812 .08254 L F P
0 g
s
.63671 .07878 m .64113 .08414 L p .63642 .07796 L .426 .494 .928 r
F P
0 g
s
.63639 .07793 m .63679 .07845 L .63625 .07755 L .426 .494 .928 r
F
.64113 .08414 m .63645 .07757 L p .63679 .07845 L F P
0 g
s
.63641 .07842 m .63624 .07705 L .63759 .07693 L .39 .478 .926 r
F
.63647 .07824 m .63759 .07693 L .63569 .07557 L p .396 .479 .925 r
F P
0 g
s
.63641 .07839 m .63598 .07707 L .63637 .07807 L .39 .478 .926 r
F
.63642 .0783 m .63647 .07824 L p .6353 .07528 L .396 .479 .925 r
F P
0 g
s
.63642 .0783 m .63447 .07304 L .63628 .07778 L .396 .48 .925 r
F
.63417 .06986 m .63759 .07693 L .63642 .0783 L p F P
0 g
s
.63641 .07842 m .6358 .07709 L .63641 .07839 L .39 .478 .926 r
F
.63658 .07905 m .63869 .08707 L .63576 .07765 L p .365 .457 .914 r
F P
0 g
s
.63641 .07839 m .636 .07752 L .63512 .07982 L .408 .49 .931 r
F
.63512 .07982 m .63642 .0783 L p .63477 .0749 L .396 .479 .925 r
F P
0 g
s
.6353 .07528 m .63477 .0749 L p .63642 .0783 L .396 .479 .925 r
F P
0 g
s
.63642 .0783 m .63512 .07982 L .63448 .07309 L p .396 .48 .925 r
F P
0 g
s
.63014 .06135 m .63007 .06135 L p .63135 .06457 L .957 .827 .957 r
F P
0 g
s
.63007 .06135 m .62986 .06135 L .63248 .0677 L p .957 .827 .957 r
F P
0 g
s
.63305 .06923 m .63291 .06885 L .63137 .06506 L .846 .769 .986 r
F
.63155 .06554 m .63137 .06506 L .62986 .06135 L F
.63512 .07982 m .63869 .08707 L .63641 .07839 L p .408 .49 .931 r
F P
0 g
s
.62986 .06135 m .62959 .06104 L .63442 .07279 L p .846 .769 .986 r
F P
0 g
s
.63442 .07279 m .6342 .07219 L .62986 .06135 L .846 .769 .986 r
F
.62959 .06104 m .63398 .07183 L p .6334 .07034 L .957 .827 .957 r
F P
0 g
s
.63287 .06865 m .62986 .06135 L p .63335 .06989 L .957 .827 .957 r
F P
0 g
s
.6313 .06489 m .62986 .06135 L p .63137 .06506 L .846 .769 .986 r
F P
0 g
s
.63759 .07693 m .64113 .08414 L .63658 .07862 L p .39 .478 .926 r
F P
0 g
s
.63658 .07862 m .63641 .07842 L p .63647 .07834 L .39 .478 .926 r
F P
0 g
s
.63658 .07862 m .63647 .07834 L .63759 .07693 L .39 .478 .926 r
F
.63008 .06139 m .62973 .06137 L .62959 .06104 L .997 .845 .937 r
F
.6479 .05145 m .62986 .06135 L p .65121 .05189 L .601 .968 1 r
F P
0 g
s
.63219 .04931 m .63219 .04935 L .63191 .04862 L .095 .095 .593 r
F
.63089 .05723 m .63188 .04865 L p .63219 .04935 L F P
0 g
s
.63188 .04865 m .63189 .04855 L p .63219 .04935 L .095 .095 .593 r
F P
0 g
s
.62986 .06135 m .6479 .05145 L p .63217 .04935 L .601 .968 1 r
F P
0 g
s
.63184 .04931 m .63166 .05036 L p .63217 .04935 L .601 .968 1 r
F P
0 g
s
.63166 .05036 m .62986 .06135 L p .63217 .04935 L .601 .968 1 r
F P
0 g
s
.6479 .05145 m .65169 .04938 L .63184 .04931 L p .601 .968 1 r
F P
0 g
s
.65121 .05189 m .65169 .04938 L .6479 .05145 L p .601 .968 1 r
F P
0 g
s
.63962 .06546 m .63759 .07693 L .63623 .05812 L p .103 .103 .659 r
F P
0 g
s
.64113 .08414 m .63759 .07693 L .63962 .06546 L p .104 .104 .666 r
F P
0 g
s
.63962 .06546 m .63968 .06511 L .63972 .06569 L p .104 .104 .666 r
F P
0 g
s
.63972 .06569 m .64113 .08414 L p .63962 .06546 L .104 .104 .666 r
F P
0 g
s
.64328 .07242 m .64325 .0726 L p .64066 .06718 L .104 .104 .666 r
F P
0 g
s
.64325 .0726 m .64113 .08414 L .63969 .06521 L p .104 .104 .666 r
F P
0 g
s
.63969 .06521 m .63968 .06511 L p .64325 .0726 L .104 .104 .666 r
F P
0 g
s
.63621 .05794 m .63417 .06986 L .63378 .05282 L closepath p .102 .102 .651 r
F P
0 g
s
.63759 .07693 m .63417 .06986 L .63619 .05805 L p .103 .103 .659 r
F P
0 g
s
.63619 .05805 m .63621 .05794 L .63623 .05812 L p .103 .103 .659 r
F P
0 g
s
.63623 .05812 m .63759 .07693 L p .63619 .05805 L .103 .103 .659 r
F P
0 g
s
.64384 .08953 m .64113 .08414 L .64316 .07309 L p .105 .105 .674 r
F P
0 g
s
.64316 .07309 m .64328 .07242 L .64331 .07343 L p .105 .105 .674 r
F P
0 g
s
.64331 .07343 m .64384 .08953 L p .64316 .07309 L .105 .105 .674 r
F P
0 g
s
.63968 .06511 m .63962 .06546 L p .63661 .05882 L .103 .103 .659 r
F P
0 g
s
.63621 .05794 m .63968 .06511 L p .63661 .05882 L .103 .103 .659 r
F P
0 g
s
.63623 .05812 m .63621 .05794 L p .63962 .06546 L .103 .103 .659 r
F P
0 g
s
.63968 .06511 m .63886 .06464 L p .64315 .07223 L .955 .551 .099 r
F P
0 g
s
.63886 .06464 m .63881 .06462 L p .63899 .06493 L .955 .551 .099 r
F P
0 g
s
.63886 .06464 m .63899 .06493 L .64315 .07223 L .955 .551 .099 r
F
.64602 .08167 m .64626 .08029 L p .6443 .09063 L .003 .003 .003 r
F P
0 g
s
.64433 .09047 m .6443 .09063 L p .64626 .08029 L .003 .003 .003 r
F P
0 g
s
.64433 .09047 m .64589 .07896 L .64593 .08202 L .107 .11 .687 r
F
.64603 .07789 m .646 .0792 L p .64589 .07896 L F P
0 g
s
.646 .0792 m .64593 .08202 L p .64589 .07896 L .107 .11 .687 r
F P
0 g
s
.64394 .08901 m .64603 .07789 L p .64433 .09047 L .107 .11 .687 r
F P
0 g
s
.64441 .09097 m .64441 .09097 L .64566 .09302 L .094 .094 .669 r
F
.64446 .09114 m .64441 .09097 L .64566 .09302 L F
.6444 .091 m .64566 .09302 L p .64441 .09096 L .701 .636 .937 r
F P
0 g
s
.64441 .09096 m .64566 .09302 L p .64441 .09097 L .094 .094 .669 r
F P
0 g
s
.64566 .09302 m .64441 .09096 L p .64441 .09096 L .701 .636 .937 r
F P
0 g
s
.64433 .09047 m .64566 .09302 L .6444 .091 L p .23 .385 .884 r
F P
0 g
s
.64362 .09454 m .64581 .08297 L .64602 .08167 L p .003 .003 .003 r
F P
0 g
s
.6443 .09063 m .64362 .09454 L p .64602 .08167 L .003 .003 .003 r
F P
0 g
s
.63178 .06482 m .63378 .05282 L .63417 .06986 L closepath p .102 .102 .651 r
F P
0 g
s
.64603 .07789 m .64384 .08953 L .64328 .07242 L closepath p .105 .105 .674 r
F P
0 g
s
.64009 .06705 m .64581 .08297 L .63481 .05787 L p .983 .841 .946 r
F P
0 g
s
.64009 .06705 m .63481 .05787 L .63735 .06204 L .983 .841 .946 r
F
.63896 .06488 m .63897 .0649 L .63965 .06608 L .955 .551 .099 r
F
.64009 .06705 m .63371 .05536 L .63801 .06312 L .983 .841 .946 r
F
.63879 .06461 m .63801 .06312 L .63803 .06317 L F
.64009 .06705 m .63801 .06312 L .63879 .06461 L F
.646 .0792 m .64603 .07789 L p .64763 .0827 L .107 .11 .687 r
F P
0 g
s
.64875 .08297 m .64788 .08143 L .64653 .07887 L p .105 .13 .703 r
F P
0 g
s
.64626 .08029 m .64625 .08034 L p .64653 .07885 L .003 .003 .003 r
F P
0 g
s
.64625 .08034 m .64582 .08291 L p .6465 .07901 L .003 .003 .003 r
F P
0 g
s
.64653 .07885 m .64625 .08034 L .6465 .07901 L .003 .003 .003 r
F
.64582 .08291 m .64581 .08297 L .64583 .08285 L p F P
0 g
s
.64583 .08285 m .64653 .07885 L p .64582 .08291 L .003 .003 .003 r
F P
0 g
s
.64566 .09302 m .64788 .08143 L .64652 .09454 L closepath p .111 .15 .72 r
F P
0 g
s
.65169 .04938 m .65121 .05189 L p .64919 .06358 L .003 .003 .003 r
F P
0 g
s
.65121 .05189 m .64942 .06142 L p .64919 .06358 L .003 .003 .003 r
F P
0 g
s
.65169 .04938 m .64653 .07885 L p .63981 .06311 L 1 .867 .915 r
F P
0 g
s
.64603 .07789 m .64653 .07885 L p .64646 .07919 L .107 .11 .687 r
F P
0 g
s
.64653 .07885 m .64651 .07898 L p .64616 .07815 L 1 .867 .915 r
F P
0 g
s
.64616 .07815 m .63981 .06311 L .64653 .07885 L 1 .867 .915 r
F
.63965 .06608 m .63983 .06657 L .64315 .07223 L .955 .551 .099 r
F
.64152 .06971 m .64315 .07223 L p .63983 .06657 L F P
0 g
s
.63881 .06462 m .63888 .06477 L .6389 .0648 L .955 .551 .099 r
F
.63879 .06461 m .63909 .06477 L p .64037 .06757 L .954 .55 .099 r
F P
0 g
s
.63891 .06477 m .63896 .06486 L .63904 .06499 L .947 .546 .096 r
F
.63634 .05895 m .63506 .05629 L p .63615 .05854 L .947 .734 .843 r
F P
0 g
s
.63909 .06477 m .63968 .06511 L p .64036 .06755 L .954 .55 .099 r
F P
0 g
s
.64009 .06705 m .64008 .06704 L p .64034 .06752 L 1 .867 .915 r
F P
0 g
s
.63378 .05282 m .63178 .06482 L .63236 .04971 L closepath p .101 .101 .638 r
F P
0 g
s
.63371 .05536 m .63278 .05325 L p .63803 .06317 L .983 .841 .946 r
F P
0 g
s
.63278 .05325 m .63282 .05334 L p .63407 .05571 L .957 .829 .959 r
F P
0 g
s
.63877 .06459 m .63481 .05709 L .63407 .05571 L .957 .829 .959 r
F
.63481 .05709 m .63278 .05325 L .63407 .05571 L F
.63481 .05787 m .63371 .05536 L p .63735 .06204 L .983 .841 .946 r
F P
0 g
s
.63282 .05334 m .64581 .08297 L p .63877 .06459 L .957 .829 .959 r
F P
0 g
s
.64581 .08297 m .64009 .06705 L p .64051 .06784 L 1 .867 .915 r
F P
0 g
s
.63881 .06462 m .63877 .06459 L p .639 .06503 L .955 .551 .099 r
F P
0 g
s
.63877 .06459 m .63879 .06461 L p .63885 .06475 L .954 .55 .099 r
F P
0 g
s
.64647 .07916 m .64583 .08285 L p .64051 .06784 L 1 .867 .915 r
F P
0 g
s
.64583 .08285 m .64581 .08297 L p .64295 .07481 L 1 .867 .915 r
F P
0 g
s
.64295 .07481 m .64051 .06784 L .64583 .08285 L 1 .867 .915 r
F
.63909 .06477 m .64036 .06755 L .64034 .06752 L .954 .55 .099 r
F
.64028 .06718 m .64041 .06766 L p .64024 .06712 L F P
0 g
s
.64037 .06757 m .64024 .06712 L .64041 .06766 L .954 .55 .099 r
F
.64009 .06705 m .63497 .05737 L .63836 .06374 L .983 .841 .946 r
F
.63037 .06174 m .63236 .04971 L .63178 .06482 L closepath p .101 .101 .638 r
F P
0 g
s
.64009 .06705 m .63218 .05187 L .6376 .06224 L .983 .841 .946 r
F
.64581 .08297 m .64761 .08297 L p .63877 .06459 L .957 .829 .959 r
F P
0 g
s
.63897 .0649 m .63922 .06545 L .63965 .06608 L .955 .551 .099 r
F
.64235 .07167 m .63877 .06459 L .64761 .08297 L .957 .829 .959 r
F
.63922 .06545 m .64031 .06784 L .63988 .06669 L .955 .551 .099 r
F
.63922 .06545 m .63988 .06669 L .63965 .06608 L F
.64009 .06705 m .6376 .06224 L .63987 .06662 L .983 .841 .946 r
F
.63235 .04973 m .63205 .05155 L .63201 .05148 L .095 .095 .593 r
F
.63205 .05155 m .63037 .06174 L .63201 .05148 L F
.63037 .06174 m .63089 .05723 L p .63201 .05148 L F P
0 g
s
.63235 .04973 m .63037 .06174 L p .63235 .04973 L .095 .095 .593 r
F P
0 g
s
.63219 .04931 m .63236 .04971 L p .63089 .05723 L .095 .095 .593 r
F P
0 g
s
.6376 .06224 m .63197 .05139 L .63857 .06411 L .983 .841 .946 r
F
.6376 .06224 m .63857 .06411 L .63987 .06662 L F
.63886 .0647 m .64006 .06698 L .63896 .06486 L .947 .546 .096 r
F
.63887 .06471 m .63903 .065 L .63924 .06542 L .97 .57 .104 r
F
.64041 .06764 m .64017 .0668 L p .63896 .06486 L .947 .546 .096 r
F P
0 g
s
.63903 .065 m .63912 .06514 L .63917 .06525 L .97 .57 .104 r
F
.63903 .065 m .63917 .06525 L .63924 .06542 L F
.64006 .06698 m .64041 .06764 L .63896 .06486 L .947 .546 .096 r
F
.63887 .06471 m .64041 .06764 L .63886 .0647 L F
.64007 .067 m .63987 .06662 L .6397 .06627 L .983 .841 .946 r
F
.64007 .067 m .6397 .06627 L .63987 .06659 L F
.64005 .06694 m .64007 .067 L p .63987 .06659 L F P
0 g
s
.63983 .06657 m .63986 .06664 L .64021 .06728 L .955 .551 .099 r
F
.64788 .08143 m .64763 .0827 L p .64646 .07919 L .107 .11 .687 r
F P
0 g
s
.64875 .08297 m .64652 .09454 L .64788 .08143 L closepath p .111 .15 .72 r
F P
0 g
s
.64125 .06929 m .64053 .06792 L .64808 .08297 L .957 .829 .959 r
F
.64053 .06792 m .64065 .06828 L .64808 .08297 L F
.63927 .06555 m .63926 .06556 L .64069 .06839 L F
.64065 .06828 m .64069 .06839 L .64808 .08297 L F
.63927 .06555 m .64069 .06839 L .64065 .06828 L F
.63912 .06514 m .64029 .06718 L p .63938 .06566 L .97 .57 .104 r
F P
0 g
s
.64029 .06718 m .64328 .07242 L p .64039 .06741 L .97 .57 .104 r
F P
0 g
s
.64029 .06718 m .64039 .06741 L .63938 .06566 L .97 .57 .104 r
F
.64053 .06792 m .63927 .06555 L .64065 .06828 L .957 .829 .959 r
F
.63986 .06664 m .64031 .06784 L .64021 .06728 L .955 .551 .099 r
F
.64041 .06766 m .64044 .06776 L p .63879 .06461 L .954 .55 .099 r
F P
0 g
s
.64003 .06691 m .64005 .06694 L .64004 .06693 L .97 .57 .104 r
F
.63927 .06555 m .63877 .06459 L .63926 .06556 L .957 .829 .959 r
F
.64042 .06772 m .63877 .06459 L .64056 .06798 L F
.64044 .06776 m .64055 .06815 L p .63885 .06475 L .954 .55 .099 r
F P
0 g
s
.64031 .06784 m .64152 .06971 L p .64021 .06728 L .955 .551 .099 r
F P
0 g
s
.6414 .06953 m .6407 .06845 L p .64116 .06907 L .97 .57 .104 r
F P
0 g
s
.63787 .06408 m .64017 .06763 L p .639 .06503 L .955 .551 .099 r
F P
0 g
s
.64017 .06763 m .64031 .06784 L p .6389 .0648 L .955 .551 .099 r
F P
0 g
s
.6407 .06845 m .63938 .06641 L p .64055 .06815 L .97 .57 .104 r
F P
0 g
s
.64055 .06815 m .63938 .06641 L .63935 .06633 L .97 .57 .104 r
F
.64055 .06815 m .64151 .07147 L p .63966 .0668 L .954 .55 .099 r
F P
0 g
s
.63888 .06477 m .64017 .06763 L .6389 .0648 L .955 .551 .099 r
F
.63881 .06462 m .64031 .06784 L .63896 .06488 L F
.63885 .06475 m .64055 .06815 L .63966 .0668 L .954 .55 .099 r
F
.63879 .06461 m .64044 .06776 L .63885 .06475 L F
.64055 .06815 m .64044 .06776 L p .63891 .06479 L .947 .546 .096 r
F P
0 g
s
.63884 .06465 m .63886 .06469 L p .63896 .06486 L .947 .546 .096 r
F P
0 g
s
.63884 .06465 m .63896 .06486 L .63891 .06477 L .947 .546 .096 r
F
.63884 .06465 m .6388 .06456 L p .639 .06495 L .991 .583 .139 r
F P
0 g
s
.63876 .06453 m .63877 .06459 L .64042 .06772 L .957 .829 .959 r
F
.63937 .06562 m .63885 .06466 L .63892 .06481 L .907 .5 .083 r
F
.63886 .0647 m .64056 .06798 L .63891 .06479 L .957 .829 .959 r
F
.63876 .06452 m .63885 .06468 L .63886 .0647 L F
.63885 .06468 m .64056 .06798 L .63886 .0647 L F
.63902 .06494 m .63904 .06499 L .63906 .06502 L .947 .546 .096 r
F
.63904 .06497 m .63906 .06502 L .64017 .0668 L F
.63902 .06494 m .63906 .06502 L .63904 .06497 L F
.63968 .06511 m .63621 .05794 L .63846 .06368 L p F P
0 g
s
.63846 .06368 m .6385 .06377 L .63968 .06511 L .947 .546 .096 r
F
.63849 .06376 m .6389 .06468 L .63968 .06511 L F
.63968 .06511 m .64028 .06718 L p .64024 .06712 L .954 .55 .099 r
F P
0 g
s
.64017 .0668 m .63968 .06511 L p .63887 .06467 L .947 .546 .096 r
F P
0 g
s
.63892 .06478 m .64017 .0668 L .63887 .06467 L .947 .546 .096 r
F
.63621 .05794 m .63979 .06634 L p .63918 .06524 L .991 .583 .139 r
F P
0 g
s
.63846 .06368 m .63621 .05794 L p .63918 .06524 L .991 .583 .139 r
F P
0 g
s
.63884 .06465 m .63892 .06478 L .63887 .06467 L .947 .546 .096 r
F
.63929 .06543 m .63901 .06495 L p .63937 .06562 L .907 .5 .083 r
F P
0 g
s
.63849 .06376 m .63884 .06465 L p .6389 .06468 L .947 .546 .096 r
F P
0 g
s
.63912 .06514 m .63938 .06566 L .63917 .06525 L .97 .57 .104 r
F
.63938 .06566 m .64042 .06746 L .64005 .06694 L F
.64042 .06746 m .64328 .07242 L .64063 .06793 L F
.64042 .06746 m .64063 .06793 L .64005 .06694 L F
.63939 .06553 m .63827 .06277 L .63784 .06183 L .888 .493 .074 r
F
.64029 .06718 m .63929 .06543 L p .63949 .06585 L .907 .5 .083 r
F P
0 g
s
.63994 .06664 m .64 .06681 L p .6392 .06515 L .983 .841 .946 r
F P
0 g
s
.64 .06681 m .64005 .06694 L p .63986 .06654 L .983 .841 .946 r
F P
0 g
s
.64 .06681 m .63986 .06654 L .6392 .06515 L .983 .841 .946 r
F
.64328 .07242 m .64152 .06971 L p .64005 .06694 L .97 .57 .104 r
F P
0 g
s
.64008 .06704 m .64 .06681 L p .64647 .07889 L 1 .867 .915 r
F P
0 g
s
.64029 .06718 m .63949 .06585 L .64038 .06751 L .907 .5 .083 r
F
.63979 .06634 m .63827 .06277 L p .6395 .06575 L .888 .493 .074 r
F P
0 g
s
.63815 .06283 m .63881 .06431 L .6387 .06404 L .888 .493 .074 r
F
.63812 .06273 m .6387 .06404 L .63752 .06106 L F
.63815 .06283 m .6387 .06404 L .63812 .06273 L F
.63236 .04971 m .63235 .04973 L p .63235 .04973 L .095 .095 .593 r
F P
0 g
s
.63378 .05282 m .63236 .04971 L .63312 .0514 L p .714 .344 .02 r
F P
0 g
s
.63215 .04495 m .64259 .04731 L p .64259 .04731 L 1 .867 .915 r
F P
0 g
s
.65169 .04938 m .65128 .04928 L .63219 .04927 L .841 1 1 r
F
.65128 .04928 m .63269 .04922 L .63219 .04927 L F
.63219 .04931 m .65169 .04938 L p .63219 .04927 L F P
0 g
s
.63219 .04931 m .63189 .04855 L p .6322 .04922 L .137 .162 .724 r
F P
0 g
s
.63194 .04869 m .63219 .04931 L p .63191 .04862 L .095 .095 .593 r
F P
0 g
s
.63237 .04921 m .63184 .04931 L p .64559 .04799 L .841 1 1 r
F P
0 g
s
.64553 .04798 m .63237 .04921 L .64559 .04799 L .841 1 1 r
F
.65128 .04928 m .64559 .04799 L .63269 .04922 L F
.64259 .04731 m .65169 .04938 L p .64259 .04731 L 1 .867 .915 r
F P
0 g
s
.65401 .04169 m .65635 .02936 L .65249 .0479 L closepath p .727 .731 .207 r
F P
0 g
s
.6548 .03566 m .65249 .0479 L .65635 .02936 L closepath p .727 .731 .207 r
F P
0 g
s
.65249 .0479 m .6548 .03566 L .65131 .05282 L closepath p .724 .726 .2 r
F P
0 g
s
.65361 .04065 m .65131 .05282 L .65387 .04026 L p .724 .726 .2 r
F P
0 g
s
.65387 .04026 m .6548 .03566 L .65361 .04065 L p .724 .726 .2 r
F P
0 g
s
.65635 .02936 m .65401 .04169 L .65758 .0229 L p .695 .676 .133 r
F P
0 g
s
.65758 .0229 m .65767 .0224 L .65635 .02936 L p .695 .676 .133 r
F P
0 g
s
.65635 .02936 m .65907 .02782 L .65533 .03469 L p .161 .161 .459 r
F P
0 g
s
.65533 .03469 m .6548 .03566 L .65635 .02936 L p .161 .161 .459 r
F P
0 g
s
.65361 .04065 m .65591 .0409 L .65346 .04087 L p .088 .023 .023 r
F P
0 g
s
.65346 .04087 m .63063 .04065 L .65361 .04065 L p .088 .023 .023 r
F P
0 g
s
.65491 .04503 m .65169 .04938 L .63215 .04495 L closepath p .841 1 1 r
F P
0 g
s
.65326 .04726 m .65381 .04651 L p .63237 .04921 L .841 1 1 r
F P
0 g
s
.65381 .04651 m .65491 .04503 L .64883 .04615 L p .841 1 1 r
F P
0 g
s
.64883 .04615 m .63237 .04921 L p .65381 .04651 L .841 1 1 r
F P
0 g
s
.64328 .07242 m .64029 .06718 L p .64576 .0774 L .907 .5 .083 r
F P
0 g
s
.64 .06681 m .64524 .07649 L .64647 .07889 L 1 .867 .915 r
F
.64009 .06705 m .64034 .06752 L .64035 .06753 L F
.64035 .06753 m .64034 .06752 L .64141 .0695 L F
.64034 .06752 m .64036 .06755 L .64037 .06757 L .954 .55 .099 r
F
.64041 .06763 m .64116 .06907 L .63998 .06685 L .97 .57 .104 r
F
.64029 .06718 m .64037 .06749 L .64576 .0774 L .907 .5 .083 r
F
.64035 .06753 m .64141 .0695 L .64152 .06971 L 1 .867 .915 r
F
.64651 .07898 m .64647 .07916 L p .64051 .06784 L F P
0 g
s
.64152 .06971 m .6414 .06953 L p .64005 .06694 L .97 .57 .104 r
F P
0 g
s
.64218 .07093 m .64447 .07506 L p .6404 .06759 L .907 .5 .083 r
F P
0 g
s
.64447 .07506 m .64576 .0774 L p .64037 .06749 L .907 .5 .083 r
F P
0 g
s
.63952 .06522 m .63762 .0611 L .63952 .06522 L 1 .756 .566 r
F
.63926 .06516 m .63459 .0546 L .63957 .06567 L 1 .667 .308 r
F
.63438 .05419 m .63783 .06209 L p .63752 .06106 L .888 .493 .074 r
F P
0 g
s
.63871 .06376 m .63459 .0546 L p .6387 .06373 L .714 .344 .02 r
F P
0 g
s
.63944 .06519 m .63944 .06519 L .63966 .0656 L 1 .867 .915 r
F
.63621 .05794 m .63378 .05282 L .63438 .05419 L p .888 .493 .074 r
F P
0 g
s
.63459 .0546 m .63455 .05453 L p .63459 .0546 L .714 .344 .02 r
F P
0 g
s
.63478 .05502 m .63752 .06106 L .6364 .05839 L .888 .493 .074 r
F
.63438 .05419 m .63752 .06106 L .63478 .05502 L F
.63679 .0595 m .63544 .0565 L .63544 .0565 L .714 .344 .02 r
F
.63544 .0565 m .63455 .05453 L .63455 .05452 L F
.63544 .0565 m .63455 .05453 L .63544 .0565 L F
.63616 .05813 m .63679 .0595 L .63406 .05342 L F
.63634 .05852 m .63679 .0595 L .63616 .05813 L F
.63378 .05282 m .63459 .0546 L p .63926 .06516 L 1 .667 .308 r
F P
0 g
s
.634 .05329 m .63406 .05342 L .63378 .05282 L .714 .344 .02 r
F
.63629 .05842 m .63406 .05342 L .634 .05329 L F
.63455 .05453 m .63378 .05282 L p .63455 .05452 L F P
0 g
s
.63783 .06209 m .63812 .06273 L .63752 .06106 L .888 .493 .074 r
F
.63949 .06577 m .63942 .0656 L .63947 .06573 L F
.63942 .0656 m .63827 .06277 L .6394 .06556 L F
.63942 .0656 m .6394 .06556 L .63947 .06573 L F
.63429 .05539 m .63915 .06516 L .63366 .05418 L .983 .841 .946 r
F
.6387 .06404 m .63939 .06553 L .63752 .06106 L .888 .493 .074 r
F
.63872 .06408 m .63878 .06422 L .6387 .06404 L F
.63878 .06422 m .63939 .06553 L .6387 .06404 L F
.63784 .06213 m .63783 .06209 L p .63925 .06514 L 1 .667 .308 r
F P
0 g
s
.63783 .06209 m .63378 .05282 L p .63918 .06498 L 1 .667 .308 r
F P
0 g
s
.63783 .06209 m .63918 .06498 L .63925 .06514 L 1 .667 .308 r
F
.63956 .06566 m .63459 .0546 L .63459 .0546 L .714 .344 .02 r
F
.63783 .06209 m .63815 .06283 L p .63812 .06273 L .888 .493 .074 r
F P
0 g
s
.6394 .06556 m .63827 .06277 L .63939 .06553 L .888 .493 .074 r
F
.63918 .06498 m .63378 .05282 L .63926 .06516 L 1 .667 .308 r
F
.63881 .06431 m .63902 .06476 L .63872 .06408 L .888 .493 .074 r
F
.63902 .06476 m .63939 .06553 L .63872 .06408 L F
.63893 .06463 m .63907 .0649 L .63881 .06431 L F
.63815 .06283 m .63893 .06463 L p .63881 .06431 L F P
0 g
s
.63893 .06463 m .63784 .06213 L p .63925 .06514 L 1 .667 .308 r
F P
0 g
s
.6394 .06512 m .63947 .06532 L p .63937 .0651 L .983 .841 .946 r
F P
0 g
s
.63947 .06532 m .63994 .06664 L p .6394 .0652 L .983 .841 .946 r
F P
0 g
s
.63947 .06532 m .6394 .0652 L .63937 .0651 L .983 .841 .946 r
F
.64603 .07789 m .64447 .07506 L p .64024 .06711 L 1 .636 .21 r
F P
0 g
s
.63974 .06616 m .64021 .06704 L p .64024 .06711 L 1 .636 .21 r
F P
0 g
s
.64021 .06704 m .64603 .07789 L p .64024 .06711 L 1 .636 .21 r
F P
0 g
s
.64 .06681 m .63984 .06635 L p .64024 .06711 L 1 .867 .915 r
F P
0 g
s
.64071 .06813 m .64024 .06711 L .64447 .07506 L 1 .867 .915 r
F
.64 .06681 m .64024 .06711 L .64071 .06813 L F
.63956 .06566 m .63459 .0546 L .63956 .06566 L .714 .344 .02 r
F
.63634 .05895 m .63615 .05854 L .6388 .06398 L .947 .734 .843 r
F
.64603 .07789 m .64021 .06704 L p .64039 .06737 L .752 .352 .035 r
F P
0 g
s
.64447 .07506 m .64371 .07359 L .64524 .07649 L 1 .867 .915 r
F
.64603 .07789 m .64039 .06737 L .64617 .07817 L .752 .352 .035 r
F
.64371 .07359 m .6394 .06512 L .64453 .07511 L 1 .867 .915 r
F
.64371 .07359 m .64453 .07511 L .64524 .07649 L F
.64024 .06711 m .64022 .06707 L .64447 .07506 L F
.64022 .06707 m .64371 .07359 L .64447 .07506 L F
.64039 .06737 m .64023 .06711 L .64652 .0789 L .752 .352 .035 r
F
.64021 .06704 m .64021 .06705 L .64039 .06737 L F
.64021 .06705 m .64023 .06711 L .64039 .06737 L F
.64022 .06707 m .6394 .06512 L .64371 .07359 L 1 .867 .915 r
F
.63984 .06635 m .63967 .06588 L p .64024 .06711 L F P
0 g
s
.63967 .06588 m .6394 .06512 L p .64024 .06711 L 1 .867 .915 r
F P
0 g
s
.63189 .04855 m .62991 .06061 L .63156 .049 L closepath p .868 1 .712 r
F P
0 g
s
.62959 .06104 m .63156 .049 L .62991 .06061 L closepath p .868 1 .712 r
F P
0 g
s
.62986 .06135 m .63184 .04931 L .62959 .06104 L closepath p .132 .373 .889 r
F P
0 g
s
.63156 .049 m .62959 .06104 L .63184 .04931 L closepath p .132 .373 .889 r
F P
0 g
s
.63197 .05139 m .63148 .05029 L p .63987 .06662 L .983 .841 .946 r
F P
0 g
s
.63148 .05029 m .62983 .04651 L p .63735 .06156 L .983 .841 .946 r
F P
0 g
s
.63891 .06477 m .63904 .06499 L .63902 .06494 L .947 .546 .096 r
F
.63987 .06662 m .63148 .05029 L .63735 .06156 L .983 .841 .946 r
F
.63885 .06468 m .63912 .06514 L .63887 .06471 L .97 .57 .104 r
F
.63885 .06466 m .63912 .06514 L p .63885 .06468 L F P
0 g
s
.63901 .06495 m .63885 .06466 L p .63937 .06562 L .907 .5 .083 r
F P
0 g
s
.63883 .06462 m .63989 .06656 L .639 .06495 L .991 .583 .139 r
F
.6388 .06456 m .63883 .0646 L .63883 .06462 L F
.63883 .0646 m .63989 .06656 L .63883 .06462 L F
.63885 .06468 m .63884 .06465 L p .63885 .06467 L F P
0 g
s
.63882 .06462 m .63885 .06466 L p .63885 .06468 L .97 .57 .104 r
F P
0 g
s
.64043 .06766 m .64027 .06734 L .64044 .06767 L .957 .829 .959 r
F
.64027 .06734 m .63652 .06022 L .64009 .067 L F
.64027 .06734 m .64009 .067 L .64044 .06767 L F
.64005 .06694 m .63987 .06659 L .63915 .06516 L .983 .841 .946 r
F
.64 .06681 m .63949 .06585 L .63955 .06598 L .907 .5 .083 r
F
.63846 .06368 m .63849 .06376 L p .6385 .06377 L .947 .546 .096 r
F P
0 g
s
.6388 .06456 m .63846 .06368 L p .63918 .06524 L .991 .583 .139 r
F P
0 g
s
.6392 .06515 m .64005 .06694 L .63915 .06516 L .983 .841 .946 r
F
.63979 .06634 m .63989 .06656 L p .6388 .06456 L .991 .583 .139 r
F P
0 g
s
.63987 .06659 m .6397 .06627 L .62983 .04651 L .983 .841 .946 r
F
.63927 .06532 m .63652 .06022 L .64043 .06766 L .957 .829 .959 r
F
.6392 .06515 m .63915 .06516 L .63769 .06222 L .983 .841 .946 r
F
.63278 .05325 m .63218 .05187 L p .63836 .06374 L F P
0 g
s
.63927 .06532 m .63908 .06495 L .63278 .05325 L .957 .829 .959 r
F
.63908 .06495 m .63469 .05621 L .63397 .05509 L F
.63218 .05187 m .63197 .05139 L p .6376 .06224 L .983 .841 .946 r
F P
0 g
s
.63908 .06495 m .63397 .05509 L .63278 .05325 L .957 .829 .959 r
F
.63069 .04847 m .63278 .05325 L p .63469 .05621 L F P
0 g
s
.63652 .06022 m .63834 .06371 L .63837 .06376 L .957 .829 .959 r
F
.63497 .05737 m .63278 .05325 L .63836 .06374 L .983 .841 .946 r
F
.63652 .06022 m .63819 .06346 L .63834 .06371 L .957 .829 .959 r
F
.63652 .06022 m .63278 .05325 L .63877 .06459 L F
.64037 .06749 m .6404 .06759 L .64447 .07506 L .907 .5 .083 r
F
.64288 .07221 m .64218 .07093 L p .63984 .06635 L 1 .636 .21 r
F P
0 g
s
.64005 .06694 m .64041 .06763 L .63998 .06685 L .97 .57 .104 r
F
.63969 .06625 m .64125 .0693 L .64553 .0774 L .957 .829 .959 r
F
.63924 .06542 m .63938 .06566 L .64005 .06694 L .97 .57 .104 r
F
.63938 .06641 m .63787 .06408 L p .63935 .06633 L F P
0 g
s
.63877 .06459 m .63787 .06408 L p .639 .06503 L .955 .551 .099 r
F P
0 g
s
.64151 .07147 m .63787 .06408 L .63877 .06459 L p .954 .55 .099 r
F P
0 g
s
.63877 .06453 m .63887 .06471 L .63787 .06408 L .97 .57 .104 r
F
.64007 .067 m .64009 .06705 L p .63987 .06662 L .983 .841 .946 r
F P
0 g
s
.64044 .06776 m .64041 .06764 L p .63889 .06477 L .947 .546 .096 r
F P
0 g
s
.63891 .06479 m .64044 .06776 L .63889 .06477 L .947 .546 .096 r
F
.6387 .0644 m .64056 .06798 L .63876 .06452 L .957 .829 .959 r
F
.63834 .06371 m .64056 .06798 L .63837 .06376 L F
.63886 .06469 m .63886 .06471 L p .64041 .06764 L .947 .546 .096 r
F P
0 g
s
.63887 .06471 m .64003 .06691 L .63979 .06659 L .97 .57 .104 r
F
.63877 .06459 m .63885 .06475 L .64151 .07147 L .954 .55 .099 r
F
.63887 .06471 m .63979 .06659 L .6392 .06582 L .97 .57 .104 r
F
.63886 .06471 m .63891 .06479 L .63889 .06477 L .947 .546 .096 r
F
.63891 .06479 m .64056 .06798 L .64125 .0693 L .957 .829 .959 r
F
.63998 .06685 m .64116 .06907 L .64103 .06887 L .97 .57 .104 r
F
.63998 .06685 m .64103 .06887 L .63935 .06633 L F
.63998 .06685 m .64103 .06887 L .63998 .06685 L F
.6392 .06583 m .63935 .06633 L .63787 .06408 L F
.63998 .06685 m .63935 .06633 L .6392 .06583 L F
.63884 .06465 m .63885 .06466 L .63885 .06467 L .991 .583 .139 r
F
.63885 .06466 m .64054 .06809 L .63885 .06467 L F
.63886 .06471 m .6392 .06582 L .63787 .06408 L .97 .57 .104 r
F
.63887 .06471 m .6392 .06582 L .63886 .06471 L F
.63929 .06543 m .63945 .0658 L .6397 .06628 L .907 .5 .083 r
F
.63917 .06525 m .63938 .06566 L .63924 .06542 L .97 .57 .104 r
F
.63937 .06562 m .63892 .06481 L .63945 .0658 L .907 .5 .083 r
F
.63926 .06545 m .63948 .06586 L .63927 .06545 L .957 .829 .959 r
F
.63925 .06545 m .63948 .06586 L .63926 .06545 L F
.63931 .06553 m .6395 .0659 L .63969 .06625 L F
.63927 .06545 m .63931 .06553 L .63931 .06553 L F
.63312 .0514 m .63563 .05695 L .634 .05329 L .714 .344 .02 r
F
.63312 .0514 m .63629 .05842 L p .63563 .05695 L F P
0 g
s
.63312 .0514 m .634 .05329 L .63378 .05282 L .714 .344 .02 r
F
.63549 .05665 m .63236 .04971 L p .63913 .06437 L 1 .756 .566 r
F P
0 g
s
.63948 .06511 m .63949 .06513 L .63236 .04971 L .137 .162 .724 r
F
.63827 .06277 m .6364 .05839 L p .63784 .06183 L .888 .493 .074 r
F P
0 g
s
.63952 .06522 m .63236 .04971 L p .63949 .06513 L .137 .162 .724 r
F P
0 g
s
.63947 .06509 m .63948 .06511 L .63219 .04931 L .137 .162 .724 r
F
.64333 .0729 m .64002 .06649 L .64026 .06717 L .957 .829 .959 r
F
.63991 .06641 m .64289 .07203 L .64028 .06724 L 1 .75 .424 r
F
.64617 .07817 m .64039 .06737 L .64652 .0789 L .752 .352 .035 r
F
.64289 .07203 m .64736 .08046 L .644 .07418 L 1 .75 .424 r
F
.64447 .07506 m .64288 .07221 L p .63984 .06635 L 1 .636 .21 r
F P
0 g
s
.64008 .06704 m .64647 .07889 L .64651 .07898 L 1 .867 .915 r
F
.64536 .0768 m .64026 .06717 L .64364 .07368 L .957 .829 .959 r
F
.64751 .08075 m .64407 .07432 L p .64736 .08046 L 1 .75 .424 r
F P
0 g
s
.64023 .06711 m .64024 .06712 L .64652 .0789 L .752 .352 .035 r
F
.64407 .07432 m .64026 .06717 L .64536 .0768 L .957 .829 .959 r
F
.64117 .06886 m .64026 .06717 L .64407 .07432 L F
.6405 .06764 m .64407 .07432 L p .64024 .06712 L .752 .352 .035 r
F P
0 g
s
.64407 .07432 m .64652 .0789 L p .64024 .06712 L .752 .352 .035 r
F P
0 g
s
.64002 .06649 m .64001 .06648 L .64026 .06717 L .957 .829 .959 r
F
.63957 .06594 m .63929 .06525 L .63951 .06581 L .888 .493 .074 r
F
.63929 .06525 m .63923 .06513 L .63958 .06599 L F
.63977 .06621 m .63994 .06664 L .63922 .0652 L .983 .841 .946 r
F
.63984 .06635 m .64218 .07093 L .6395 .06575 L 1 .636 .21 r
F
.63947 .06573 m .6394 .06556 L .63939 .06553 L .888 .493 .074 r
F
.64026 .06717 m .64011 .06688 L .63927 .06532 L .957 .829 .959 r
F
.63974 .06616 m .63984 .06635 L .6395 .06575 L 1 .636 .21 r
F
.64032 .06735 m .63927 .06532 L .64043 .06766 L .957 .829 .959 r
F
.64026 .06717 m .63944 .06564 L .64032 .06735 L F
.63944 .06564 m .63927 .06532 L .64032 .06735 L F
.63927 .06545 m .63931 .06553 L .63931 .06553 L F
.63837 .06376 m .6387 .0644 L .63876 .06452 L F
.63884 .06465 m .639 .06495 L .63892 .06481 L .991 .583 .139 r
F
.63834 .06371 m .63854 .06412 L .63876 .06453 L .957 .829 .959 r
F
.63854 .06412 m .63877 .06459 L .63876 .06453 L F
.63819 .06346 m .63877 .06459 L .63834 .06371 L F
.63886 .06471 m .63952 .0664 L p .64055 .06815 L .947 .546 .096 r
F P
0 g
s
.63952 .0664 m .64151 .07147 L .64071 .06869 L p .947 .546 .096 r
F P
0 g
s
.64071 .06869 m .64055 .06815 L p .63952 .0664 L .947 .546 .096 r
F P
0 g
s
.64003 .06691 m .64004 .06693 L .63787 .06408 L .97 .57 .104 r
F
.6444 .07715 m .64151 .07147 L .63885 .06468 L p .991 .583 .139 r
F P
0 g
s
.63877 .06453 m .63882 .06462 L p .63887 .06471 L .97 .57 .104 r
F P
0 g
s
.63886 .06469 m .63887 .06471 L .63886 .0647 L .947 .546 .096 r
F
.63885 .06468 m .63885 .06467 L .6444 .07715 L .991 .583 .139 r
F
.63787 .06408 m .63435 .0568 L .63876 .06452 L p .97 .57 .104 r
F P
0 g
s
.63876 .06452 m .63877 .06453 L p .63877 .06453 L .97 .57 .104 r
F P
0 g
s
.63886 .06471 m .64055 .06815 L .63891 .06479 L .947 .546 .096 r
F
.63876 .06452 m .63877 .06453 L .63787 .06408 L .97 .57 .104 r
F
.63892 .06481 m .63435 .0568 L .63974 .06648 L .907 .5 .083 r
F
.63885 .06466 m .63876 .06452 L p .63892 .06481 L F P
0 g
s
.63932 .06544 m .63975 .06634 L .63677 .06107 L 1 .636 .21 r
F
.63643 .06041 m .63677 .06107 L .63435 .0568 L F
.63876 .06452 m .63435 .0568 L p .63892 .06481 L .907 .5 .083 r
F P
0 g
s
.63932 .06544 m .63677 .06107 L .63643 .06041 L 1 .636 .21 r
F
.64048 .06776 m .64553 .0774 L .64435 .07512 L .957 .829 .959 r
F
.63931 .06553 m .64055 .06788 L .64048 .06776 L F
.64055 .06788 m .64553 .0774 L .64048 .06776 L F
.64 .06681 m .63955 .06598 L .6397 .06628 L .907 .5 .083 r
F
.63931 .06553 m .64055 .06789 L .63931 .06553 L .957 .829 .959 r
F
.64055 .06789 m .64553 .0774 L .64055 .06788 L F
.64038 .06751 m .64 .06681 L .64011 .06704 L .907 .5 .083 r
F
.64055 .06789 m .64055 .06788 L .63997 .06678 L .957 .829 .959 r
F
.64055 .06789 m .63997 .06678 L .63996 .06677 L F
.6404 .06759 m .64011 .06704 L .64047 .06784 L .907 .5 .083 r
F
.64038 .06751 m .64011 .06704 L .6404 .06759 L F
.63996 .06677 m .63997 .06678 L .63931 .06553 L .957 .829 .959 r
F
.63957 .06594 m .63979 .06634 L .6395 .06575 L .888 .493 .074 r
F
.63885 .06467 m .63998 .06695 L .63973 .06665 L .991 .583 .139 r
F
.63945 .0658 m .63985 .06673 L .6397 .06628 L .907 .5 .083 r
F
.63985 .06673 m .64047 .06784 L p .6397 .06628 L F P
0 g
s
.63925 .06545 m .64125 .0693 L .63948 .06586 L .957 .829 .959 r
F
.6395 .0659 m .64125 .0693 L .63969 .06625 L F
.63931 .06553 m .64125 .0693 L .63931 .06553 L F
.64011 .06704 m .6397 .06628 L .64047 .06784 L .907 .5 .083 r
F
.64 .06681 m .6397 .06628 L .64011 .06704 L F
.63945 .0658 m .63892 .06481 L .63974 .06648 L F
.63989 .06656 m .64054 .06809 L p .639 .06495 L .991 .583 .139 r
F P
0 g
s
.63892 .06481 m .639 .06495 L .64054 .06809 L .991 .583 .139 r
F
.63974 .06648 m .63765 .06272 L .63985 .06673 L .907 .5 .083 r
F
.63435 .0568 m .63748 .06244 L p .63765 .06272 L F P
0 g
s
.63748 .06244 m .63985 .06673 L p .63765 .06272 L .907 .5 .083 r
F P
0 g
s
.63998 .06696 m .63435 .0568 L p .63975 .06634 L 1 .636 .21 r
F P
0 g
s
.63998 .06682 m .63985 .06649 L .63969 .06621 L .888 .493 .074 r
F
.64037 .06743 m .64014 .06715 L .63975 .06634 L 1 .636 .21 r
F
.64014 .06715 m .63998 .06696 L .63975 .06634 L F
.63969 .06621 m .63985 .06649 L .63957 .06594 L .888 .493 .074 r
F
.6407 .06867 m .63998 .06682 L .63969 .06621 L F
.63985 .06649 m .63979 .06634 L .63957 .06594 L F
.63927 .06545 m .63931 .06553 L .63927 .06545 L .957 .829 .959 r
F
.63944 .06579 m .64553 .0774 L .63945 .06579 L F
.63968 .06511 m .64328 .07242 L p .64066 .06718 L .104 .104 .666 r
F P
0 g
s
.64315 .07223 m .64328 .07242 L .63968 .06511 L p .955 .551 .099 r
F P
0 g
s
.64576 .0774 m .64603 .07789 L .64328 .07242 L p .907 .5 .083 r
F P
0 g
s
.63966 .0656 m .64181 .06786 L .64623 .07833 L 1 .867 .915 r
F
.63955 .06562 m .63679 .0595 L .63679 .0595 L .714 .344 .02 r
F
.63962 .06577 m .63994 .06664 L .63932 .06513 L .983 .841 .946 r
F
.63932 .06513 m .63977 .06621 L .63922 .0652 L F
.6393 .06512 m .63922 .0652 L .6392 .06515 L F
.63957 .06567 m .63971 .06597 L p .63893 .06463 L 1 .667 .308 r
F P
0 g
s
.6397 .06596 m .63955 .06562 L .63955 .06562 L .714 .344 .02 r
F
.63955 .06562 m .63679 .0595 L .63955 .06562 L F
.63937 .0651 m .63962 .06577 L .63932 .06513 L .983 .841 .946 r
F
.63949 .06545 m .63941 .06527 L .6395 .06545 L .947 .734 .843 r
F
.63947 .06542 m .64659 .07899 L p .64093 .06826 L 1 .75 .424 r
F P
0 g
s
.63987 .06634 m .63957 .06567 L .63957 .06566 L .714 .344 .02 r
F
.63966 .06588 m .6397 .06596 L .63957 .06567 L F
.63987 .06633 m .63956 .06566 L .63956 .06566 L F
.64001 .06648 m .63946 .06545 L .64024 .06713 L .957 .829 .959 r
F
.63948 .06545 m .63946 .06545 L .64001 .06648 L F
.63948 .06545 m .64093 .06826 L .63954 .06561 L 1 .75 .424 r
F
.63947 .06542 m .64093 .06826 L .63948 .06545 L F
.63971 .06564 m .63948 .06544 L .63997 .06636 L .957 .829 .959 r
F
.63952 .06522 m .64181 .06786 L .63966 .0656 L 1 .867 .915 r
F
.64093 .06826 m .64247 .07124 L .63986 .0663 L 1 .75 .424 r
F
.63937 .06497 m .63913 .06437 L .63972 .06564 L 1 .756 .566 r
F
.6394 .06512 m .63941 .06512 L .64524 .07649 L 1 .867 .915 r
F
.63972 .06564 m .63971 .06564 L .64001 .06648 L .957 .829 .959 r
F
.63941 .06513 m .63952 .06522 L .63966 .0656 L 1 .867 .915 r
F
.63945 .06514 m .63952 .06522 L .63941 .06513 L F
.64407 .07432 m .64029 .06724 L p .644 .07418 L 1 .75 .424 r
F P
0 g
s
.6399 .06652 m .64029 .06724 L p .63948 .06569 L .752 .352 .035 r
F P
0 g
s
.64029 .06724 m .6398 .06633 L p .64029 .06724 L 1 .75 .424 r
F P
0 g
s
.644 .07418 m .64029 .06724 L .64029 .06724 L 1 .75 .424 r
F
.64021 .06704 m .6398 .06629 L p .64022 .06706 L .752 .352 .035 r
F P
0 g
s
.63946 .06545 m .63941 .06548 L .64024 .06713 L .957 .829 .959 r
F
.63989 .06656 m .63979 .06634 L p .63985 .06649 L .888 .493 .074 r
F P
0 g
s
.63998 .06682 m .63989 .06656 L .63985 .06649 L .888 .493 .074 r
F
.6407 .06867 m .64001 .0669 L .63998 .06682 L F
.64001 .0669 m .63989 .06656 L .63998 .06682 L F
.64047 .06784 m .64109 .06896 L p .6404 .06759 L .907 .5 .083 r
F P
0 g
s
.64152 .06971 m .64141 .0695 L .64651 .07898 L 1 .867 .915 r
F
.64653 .07885 m .64788 .08143 L p .64646 .07919 L .107 .11 .687 r
F P
0 g
s
.64026 .06717 m .64044 .06767 L .64847 .08297 L .957 .829 .959 r
F
.64109 .06896 m .64218 .07093 L p .64087 .06848 L .907 .5 .083 r
F P
0 g
s
.6404 .06759 m .64109 .06896 L .64087 .06848 L .907 .5 .083 r
F
.64218 .07093 m .64047 .06784 L p .64037 .06743 L 1 .636 .21 r
F P
0 g
s
.64047 .06784 m .63998 .06696 L p .64014 .06715 L 1 .636 .21 r
F P
0 g
s
.64037 .06743 m .64047 .06784 L .64014 .06715 L 1 .636 .21 r
F
.6391 .06499 m .63957 .06597 L .63947 .06573 L .888 .493 .074 r
F
.63884 .06449 m .6391 .06499 L .63435 .0568 L 1 .636 .21 r
F
.63907 .06492 m .63947 .06573 L .63939 .06553 L .888 .493 .074 r
F
.6391 .06499 m .63947 .06573 L .63907 .06492 L F
.6394 .06556 m .6391 .06499 L .63927 .06532 L .957 .829 .959 r
F
.6391 .06499 m .63943 .06562 L .63928 .06537 L 1 .636 .21 r
F
.63652 .06022 m .63927 .06545 L .63927 .06545 L .957 .829 .959 r
F
.63652 .06022 m .63925 .06545 L .63927 .06545 L F
.6391 .06499 m .63928 .06537 L .63435 .0568 L 1 .636 .21 r
F
.63922 .0652 m .63994 .06664 L .63922 .0652 L .983 .841 .946 r
F
.63907 .0649 m .63939 .06553 L .63881 .06431 L .888 .493 .074 r
F
.63884 .06449 m .63974 .06616 L p .63943 .06562 L 1 .636 .21 r
F P
0 g
s
.63904 .06487 m .6391 .06499 L .63907 .06492 L .888 .493 .074 r
F
.63904 .06487 m .63957 .06597 L .6391 .06499 L F
.6395 .06575 m .64037 .06743 L .63975 .06634 L 1 .636 .21 r
F
.63969 .06621 m .63957 .06594 L .63958 .06599 L .888 .493 .074 r
F
.63957 .06594 m .63951 .06581 L .63958 .06599 L F
.63963 .06611 m .63949 .06577 L .63947 .06573 L F
.6395 .06575 m .63975 .06634 L .63932 .06544 L 1 .636 .21 r
F
.63943 .06562 m .63932 .06544 L .63928 .06537 L F
.63974 .06616 m .63947 .0657 L .63943 .06562 L F
.63947 .0657 m .63932 .06544 L .63943 .06562 L F
.63947 .0657 m .6395 .06575 L .63932 .06544 L F
.63974 .06616 m .6395 .06575 L .63947 .0657 L F
.64011 .06688 m .63899 .06477 L .6391 .06499 L .957 .829 .959 r
F
.64011 .06688 m .6391 .06499 L .6394 .06556 L F
.64022 .06706 m .63903 .06484 L .64023 .06711 L .752 .352 .035 r
F
.63903 .06484 m .63885 .0645 L .63906 .0649 L F
.63903 .06484 m .63906 .0649 L .64023 .06711 L F
.63874 .06424 m .63769 .06222 L .63131 .0494 L .983 .841 .946 r
F
.63131 .0494 m .63429 .05539 L .62983 .04651 L F
.6309 .04857 m .62983 .04651 L p .63469 .05621 L .957 .829 .959 r
F P
0 g
s
.63524 .05726 m .63469 .05621 L .63927 .06532 L .957 .829 .959 r
F
.6309 .04857 m .63469 .05621 L .63524 .05726 L F
.63885 .0645 m .63892 .06465 L .64023 .06711 L .752 .352 .035 r
F
.6398 .06629 m .63882 .06445 L p .64022 .06706 L F P
0 g
s
.63925 .06514 m .6392 .06515 L .63893 .06463 L .983 .841 .946 r
F
.63925 .06514 m .63893 .06463 L .63131 .0494 L F
.63893 .06463 m .63971 .06597 L .63907 .06491 L 1 .667 .308 r
F
.63893 .06463 m .63904 .06487 L p .63939 .06553 L .888 .493 .074 r
F P
0 g
s
.63904 .06487 m .63893 .06463 L p .63907 .06491 L 1 .667 .308 r
F P
0 g
s
.64011 .06688 m .63748 .06165 L .63899 .06477 L .957 .829 .959 r
F
.63748 .06165 m .6309 .04857 L .63658 .05995 L F
.6392 .06515 m .63769 .06222 L .63874 .06424 L .983 .841 .946 r
F
.63748 .06165 m .63658 .05995 L .63899 .06477 L .957 .829 .959 r
F
.63957 .06567 m .63871 .06376 L p .63957 .06566 L .714 .344 .02 r
F P
0 g
s
.63934 .06516 m .63934 .06518 L .63863 .06361 L .714 .344 .02 r
F
.6397 .06596 m .63871 .06376 L .6387 .06373 L F
.64024 .06713 m .63941 .06548 L .64026 .06717 L .957 .829 .959 r
F
.63967 .06609 m .6399 .06652 L p .63921 .06518 L .752 .352 .035 r
F P
0 g
s
.6398 .06633 m .63976 .06625 L p .6398 .06633 L 1 .75 .424 r
F P
0 g
s
.63982 .06654 m .63969 .06621 L .63963 .06611 L .888 .493 .074 r
F
.63969 .06621 m .63958 .06599 L .63963 .06611 L F
.63928 .06537 m .63932 .06544 L .63435 .0568 L 1 .636 .21 r
F
.64116 .06907 m .6407 .06845 L .64055 .06815 L .97 .57 .104 r
F
.64034 .06738 m .64652 .0811 L p .63907 .06492 L 1 .667 .308 r
F P
0 g
s
.63976 .06625 m .63973 .06619 L p .63976 .06625 L 1 .75 .424 r
F P
0 g
s
.6398 .06633 m .63976 .06625 L .63976 .06625 L 1 .75 .424 r
F
.63907 .06492 m .63971 .06597 L .63988 .06636 L 1 .667 .308 r
F
.63948 .06569 m .64029 .06724 L .64023 .06711 L .752 .352 .035 r
F
.63991 .06641 m .64028 .06724 L .63976 .06625 L 1 .75 .424 r
F
.63921 .06518 m .6399 .06652 L .63948 .06569 L .752 .352 .035 r
F
.63907 .06492 m .63988 .06636 L .63976 .06625 L 1 .667 .308 r
F
.63986 .06634 m .63976 .06625 L .63938 .06555 L 1 .75 .424 r
F
.63991 .06641 m .63976 .06625 L .63986 .06634 L F
.63971 .06597 m .63971 .06597 L .63988 .06636 L 1 .667 .308 r
F
.63988 .06636 m .63988 .06636 L .64034 .06737 L F
.63971 .06597 m .63988 .06636 L .63988 .06636 L F
.63976 .06625 m .63988 .06636 L .64034 .06737 L F
.64652 .0789 m .64788 .08143 L .64653 .07885 L p .752 .352 .035 r
F P
0 g
s
.64659 .07899 m .64736 .08046 L .63987 .06635 L 1 .75 .424 r
F
.63895 .06444 m .63925 .06514 L .63148 .04973 L .983 .841 .946 r
F
.63895 .06444 m .63148 .04973 L .63131 .0494 L F
.63948 .06545 m .63932 .06512 L .63941 .06527 L .947 .734 .843 r
F
.63932 .06512 m .63634 .05895 L .63934 .06511 L F
.63932 .06512 m .63934 .06511 L .63941 .06527 L F
.6397 .06596 m .63871 .06376 L .6397 .06596 L .714 .344 .02 r
F
.63946 .06545 m .63634 .05895 L p .63948 .06545 L .947 .734 .843 r
F P
0 g
s
.63852 .06341 m .63895 .06444 L .63131 .0494 L .983 .841 .946 r
F
.63937 .06497 m .6394 .06503 L .63697 .06012 L 1 .764 .752 r
F
.63863 .06361 m .63932 .06513 L .63679 .0595 L .714 .344 .02 r
F
.63932 .06513 m .63934 .06518 L .63932 .06514 L F
.63932 .06513 m .63932 .06514 L .63679 .0595 L F
.63945 .06514 m .6394 .06512 L .63697 .06012 L 1 .764 .752 r
F
.63934 .06499 m .63916 .06451 L .63937 .06497 L 1 .756 .566 r
F
.63896 .06434 m .63952 .06545 L .6395 .06545 L .947 .734 .843 r
F
.63896 .06434 m .6388 .06398 L .63952 .06545 L F
.64659 .07899 m .63987 .06635 L .63989 .06636 L 1 .75 .424 r
F
.63955 .06546 m .63455 .05517 L .63452 .05509 L 1 .764 .752 r
F
.644 .07418 m .6398 .06633 L .6398 .06633 L 1 .75 .424 r
F
.63978 .06629 m .63973 .06619 L .63973 .06619 L F
.644 .07418 m .63978 .06629 L .63978 .06629 L F
.63978 .06629 m .63973 .06619 L .63978 .06629 L F
.64333 .0729 m .64117 .06886 L .64407 .07432 L .957 .829 .959 r
F
.64289 .07203 m .644 .07418 L .64028 .06724 L 1 .75 .424 r
F
.63966 .06561 m .63966 .0656 L .64048 .06719 L 1 .867 .915 r
F
.64043 .06711 m .64048 .06719 L .64623 .07833 L F
.63966 .06561 m .64048 .06719 L .64043 .06711 L F
.63937 .06505 m .64623 .07833 L .64524 .07649 L F
.64875 .08297 m .63985 .06582 L p .64 .06641 L .957 .829 .959 r
F P
0 g
s
.64003 .06642 m .64654 .07879 L p .64309 .07227 L .105 .13 .703 r
F P
0 g
s
.63996 .06636 m .63997 .06637 L .64309 .07227 L .105 .13 .703 r
F
.63997 .06637 m .64003 .06642 L .64309 .07227 L F
.64653 .07885 m .64603 .07789 L p .64652 .0789 L .752 .352 .035 r
F P
0 g
s
.64653 .07887 m .63996 .06636 L p .64654 .07879 L .105 .13 .703 r
F P
0 g
s
.64875 .08297 m .64 .06641 L .64002 .06649 L .957 .829 .959 r
F
.63985 .06582 m .63963 .0654 L p .63972 .06565 L F P
0 g
s
.63985 .06582 m .63972 .06565 L .64002 .06649 L .957 .829 .959 r
F
.64576 .07741 m .64659 .07899 L .64525 .07647 L 1 .75 .424 r
F
.64576 .07741 m .64525 .07647 L .64247 .07124 L F
.64093 .06826 m .64576 .07741 L .64247 .07124 L F
.64654 .07879 m .64875 .08297 L p .64653 .07887 L .105 .13 .703 r
F P
0 g
s
.64875 .08297 m .64002 .06649 L .64333 .0729 L .957 .829 .959 r
F
.63955 .06546 m .63455 .05517 L .63955 .06546 L 1 .764 .752 r
F
.63963 .06553 m .63937 .06497 L .63964 .06554 L 1 .756 .566 r
F
.63935 .06511 m .63936 .06511 L .63938 .06518 L .947 .734 .843 r
F
.63936 .06511 m .63937 .0651 L .63938 .06518 L F
.6361 .05818 m .63937 .06497 L p .63697 .06012 L 1 .764 .752 r
F P
0 g
s
.63801 .06223 m .63837 .06296 L .63761 .06135 L .904 .716 .866 r
F
.63823 .0627 m .63837 .06296 L .63801 .06223 L F
.63823 .0627 m .6385 .06323 L .63837 .06296 L F
.63969 .06557 m .64001 .06648 L .64002 .06649 L .957 .829 .959 r
F
.6376 .06106 m .63952 .06522 L p .63762 .0611 L 1 .756 .566 r
F P
0 g
s
.6397 .06558 m .63968 .06554 L .63967 .06552 L .137 .162 .724 r
F
.63985 .06582 m .64875 .08297 L .64003 .06642 L p 1 .857 .741 r
F P
0 g
s
.63985 .06582 m .64003 .06642 L .63982 .06587 L 1 .857 .741 r
F
.63969 .06557 m .64001 .06648 L .63969 .06557 L .957 .829 .959 r
F
.63979 .06581 m .63978 .06594 L .63975 .06581 L 1 .857 .741 r
F
.63985 .06582 m .63982 .06587 L .63979 .06581 L F
.63982 .06587 m .63978 .06594 L .63979 .06581 L F
.63972 .06564 m .63952 .06522 L .63949 .06513 L .137 .162 .724 r
F
.64045 .0672 m .63976 .06573 L .63972 .06564 L F
.63976 .06573 m .63952 .06522 L .63972 .06564 L F
.63977 .06577 m .6397 .06581 L .63972 .06573 L 1 .857 .741 r
F
.63985 .06582 m .63979 .06581 L .63977 .06577 L F
.63979 .06581 m .6397 .06581 L .63977 .06577 L F
.6364 .05839 m .63621 .05794 L p .63438 .05419 L .888 .493 .074 r
F P
0 g
s
.63236 .04971 m .63219 .04931 L p .63948 .06511 L .137 .162 .724 r
F P
0 g
s
.6322 .04922 m .63222 .04927 L p .63219 .04931 L .137 .162 .724 r
F P
0 g
s
.63218 .04924 m .6322 .04924 L p .63616 .05814 L .863 .698 .883 r
F P
0 g
s
.63981 .06311 m .64259 .04731 L .65169 .04938 L 1 .867 .915 r
F
.65326 .04726 m .64553 .04798 L .64559 .04799 L .841 1 1 r
F
.65169 .04938 m .65326 .04726 L p .64559 .04799 L F P
0 g
s
.6322 .04924 m .64259 .04731 L p .63625 .05807 L .863 .698 .883 r
F P
0 g
s
.63625 .05807 m .64259 .04731 L .64132 .05453 L .863 .698 .883 r
F
.63952 .06522 m .6396 .06463 L .64181 .06786 L 1 .867 .915 r
F
.63947 .06508 m .6396 .06463 L .63952 .06522 L F
.63974 .06561 m .63985 .06582 L p .63974 .06562 L 1 .857 .741 r
F P
0 g
s
.64044 .06719 m .63969 .06556 L .63992 .06528 L .863 .698 .883 r
F
.6396 .06463 m .64015 .06391 L .64181 .06786 L 1 .867 .915 r
F
.64064 .06682 m .63992 .06528 L .64526 .05842 L .863 .698 .883 r
F
.64044 .06719 m .63992 .06528 L .64064 .06682 L F
.63937 .06497 m .63972 .06564 L .63964 .06554 L 1 .756 .566 r
F
.63937 .06497 m .63945 .06514 L p .6394 .06503 L 1 .764 .752 r
F P
0 g
s
.63916 .06451 m .63915 .06448 L .63937 .06497 L 1 .756 .566 r
F
.63915 .06448 m .63913 .06437 L .63937 .06497 L F
.63945 .06514 m .63734 .06075 L p .63947 .06507 L .863 .698 .883 r
F P
0 g
s
.63734 .06075 m .6361 .05818 L p .63713 .06027 L .863 .698 .883 r
F P
0 g
s
.63655 .059 m .63841 .06311 L .63679 .0595 L .714 .344 .02 r
F
.63734 .06075 m .63713 .06027 L .63947 .06507 L .863 .698 .883 r
F
.63652 .05894 m .63867 .06369 L .63841 .06311 L .714 .344 .02 r
F
.63652 .05894 m .63841 .06311 L .63655 .059 L F
.63937 .06505 m .63945 .06483 L .63945 .06514 L 1 .867 .915 r
F
.63947 .06507 m .63625 .05807 L .64132 .05453 L .863 .698 .883 r
F
.63945 .06483 m .63956 .06453 L .63952 .06474 L 1 .867 .915 r
F
.63945 .06483 m .63952 .06474 L .63945 .06514 L F
.63956 .06453 m .63981 .06311 L .63995 .06344 L F
.6396 .06463 m .63995 .06344 L .64015 .06391 L F
.63956 .06453 m .63995 .06344 L .6396 .06463 L F
.63952 .06474 m .6396 .06463 L .63945 .06514 L F
.63956 .06453 m .6396 .06463 L .63952 .06474 L F
.63967 .06476 m .64132 .05453 L .65119 .04763 L .863 .698 .883 r
F
.63947 .06507 m .63956 .06453 L .63967 .06476 L F
.63956 .06453 m .64132 .05453 L .63967 .06476 L F
.6395 .06493 m .6322 .04922 L p .63944 .06498 L .904 .716 .866 r
F P
0 g
s
.6322 .04922 m .63945 .06483 L p .6322 .04922 L .947 .735 .844 r
F P
0 g
s
.63968 .06549 m .63956 .06496 L .64526 .05842 L .863 .698 .883 r
F
.63968 .06549 m .63974 .06561 L p .63968 .0655 L 1 .857 .741 r
F P
0 g
s
.63974 .06561 m .63974 .06562 L .63965 .06545 L 1 .857 .741 r
F
.63969 .06556 m .63968 .06549 L .63989 .06523 L .863 .698 .883 r
F
.63992 .06528 m .63989 .06523 L .64526 .05842 L F
.63969 .06556 m .63989 .06523 L .63992 .06528 L F
.63967 .06551 m .63947 .06507 L .63956 .06496 L F
.63967 .06551 m .63956 .06496 L .63968 .06549 L F
.64526 .05842 m .63947 .06507 L .65119 .04763 L F
.64736 .08046 m .64407 .07432 L .644 .07418 L 1 .75 .424 r
F
.64524 .07649 m .64623 .07833 L .64647 .07889 L 1 .867 .915 r
F
.64875 .08297 m .64333 .0729 L .64536 .0768 L .957 .829 .959 r
F
.64435 .07512 m .64553 .0774 L .64847 .08297 L F
.64788 .08143 m .64751 .08075 L p .64736 .08046 L 1 .75 .424 r
F P
0 g
s
.64659 .07899 m .64788 .08143 L p .64736 .08046 L 1 .75 .424 r
F P
0 g
s
.64847 .08297 m .64859 .08297 L p .64364 .07368 L .957 .829 .959 r
F P
0 g
s
.64859 .08297 m .64875 .08297 L p .64364 .07368 L .957 .829 .959 r
F P
0 g
s
.64808 .08297 m .64847 .08297 L p .64125 .06929 L .957 .829 .959 r
F P
0 g
s
.64824 .08296 m .64044 .06719 L .64526 .05842 L .863 .698 .883 r
F
.64081 .06797 m .6395 .06493 L .63944 .06498 L .904 .716 .866 r
F
.6394 .06489 m .63959 .0653 L p .63942 .06493 L .137 .162 .724 r
F P
0 g
s
.63961 .06535 m .6322 .04922 L .6322 .04922 L .904 .716 .866 r
F
.63961 .06535 m .6322 .04922 L .63961 .06535 L F
.6322 .04922 m .63945 .06483 L .63945 .06483 L .947 .735 .844 r
F
.63876 .0635 m .63854 .06287 L .63945 .06483 L F
.63876 .0635 m .63945 .06483 L .63941 .06489 L F
.63222 .04927 m .6394 .06489 L p .63686 .0594 L .137 .162 .724 r
F P
0 g
s
.6394 .06489 m .63942 .06493 L .63219 .04931 L .137 .162 .724 r
F
.63762 .0611 m .63952 .06522 L .63952 .06522 L 1 .756 .566 r
F
.63942 .06493 m .63947 .06509 L .63219 .04931 L .137 .162 .724 r
F
.63947 .06507 m .63836 .06281 L .63625 .05807 L .863 .698 .883 r
F
.63945 .06514 m .63947 .06508 L .63952 .06522 L 1 .867 .915 r
F
.63937 .06505 m .63944 .06519 L .63944 .06519 L F
.63945 .06514 m .63962 .06552 L .6394 .06512 L 1 .764 .752 r
F
.63944 .06519 m .63966 .0656 L .63966 .06561 L 1 .867 .915 r
F
.63962 .06552 m .63957 .06548 L .6394 .06512 L 1 .764 .752 r
F
.63919 .06466 m .63916 .06451 L .63934 .06499 L 1 .756 .566 r
F
.63836 .06281 m .63713 .06027 L .63682 .0595 L .863 .698 .883 r
F
.63919 .06466 m .63549 .05665 L .63916 .06451 L 1 .756 .566 r
F
.63971 .06578 m .63971 .06579 L .63706 .06033 L 1 .764 .752 r
F
.63932 .0651 m .63937 .06505 L .6385 .06323 L .904 .716 .866 r
F
.6394 .06512 m .63937 .06505 L p .63941 .06512 L 1 .867 .915 r
F P
0 g
s
.63556 .05703 m .63954 .06541 L .63877 .06377 L .904 .716 .866 r
F
.63924 .06487 m .63696 .05987 L .63919 .06466 L 1 .756 .566 r
F
.63679 .0595 m .63544 .0565 L .63679 .0595 L .714 .344 .02 r
F
.63696 .05987 m .63549 .05665 L .63919 .06466 L 1 .756 .566 r
F
.63634 .05852 m .63655 .059 L .63679 .0595 L .714 .344 .02 r
F
.63549 .05665 m .63913 .06437 L .63916 .06451 L 1 .756 .566 r
F
.63713 .06027 m .63616 .05814 L .63625 .05807 L .863 .698 .883 r
F
.63713 .06027 m .6361 .05818 L .63616 .05814 L F
.63629 .05842 m .63634 .05852 L .63616 .05813 L .714 .344 .02 r
F
.6361 .05818 m .63365 .05258 L .63616 .05814 L .863 .698 .883 r
F
.63877 .06377 m .6396 .06538 L .63352 .05247 L .904 .716 .866 r
F
.63954 .06541 m .63954 .06541 L .63877 .06377 L F
.63956 .06544 m .63937 .06497 L .63962 .06552 L 1 .756 .566 r
F
.63937 .06505 m .63945 .06514 L .63941 .06513 L 1 .867 .915 r
F
.63937 .06505 m .63937 .06505 L .63944 .06519 L F
.63954 .06541 m .6396 .06538 L .63877 .06377 L .904 .716 .866 r
F
.63221 .04927 m .63686 .0594 L .63219 .04931 L .137 .162 .724 r
F
.63236 .04971 m .6376 .06106 L p .63762 .0611 L 1 .756 .566 r
F P
0 g
s
.63616 .05814 m .6322 .04924 L .63625 .05807 L .863 .698 .883 r
F
.63836 .06281 m .63682 .0595 L .63625 .05807 L F
.63365 .05258 m .63218 .04924 L .63616 .05814 L F
.63184 .04931 m .63219 .04931 L p .63219 .04927 L .841 1 1 r
F P
0 g
s
.63204 .04927 m .63218 .04924 L p .63459 .05471 L .863 .698 .883 r
F P
0 g
s
.63189 .04855 m .63194 .04869 L p .63191 .04862 L .095 .095 .593 r
F P
0 g
s
.63189 .04855 m .63176 .04873 L p .63963 .06546 L .904 .716 .866 r
F P
0 g
s
.63222 .04927 m .63686 .0594 L .63221 .04927 L .137 .162 .724 r
F
.63963 .0654 m .63967 .06552 L .63947 .06509 L F
.63962 .06536 m .63963 .0654 L .63947 .06509 L F
.63962 .06536 m .63947 .06509 L .63942 .06493 L F
.63967 .06552 m .63948 .06511 L .63947 .06509 L F
.63968 .06554 m .63948 .06511 L .63967 .06552 L F
.63968 .06554 m .63968 .06554 L .63948 .06511 L F
.63963 .06546 m .63961 .06535 L .63189 .04855 L .904 .716 .866 r
F
.63937 .06505 m .63215 .04495 L p .6329 .04673 L 1 .867 .915 r
F P
0 g
s
.6394 .06489 m .63876 .0635 L p .63941 .06489 L .947 .735 .844 r
F P
0 g
s
.63937 .06505 m .6329 .04673 L .63941 .0649 L 1 .867 .915 r
F
.63937 .06505 m .63981 .06311 L .63956 .06453 L F
.63952 .06522 m .63977 .06577 L p .63952 .06522 L 1 .756 .566 r
F P
0 g
s
.63971 .06579 m .63955 .06546 L .63955 .06546 L 1 .764 .752 r
F
.63932 .0651 m .6385 .06323 L .63823 .0627 L .904 .716 .866 r
F
.63937 .06506 m .63931 .06511 L .63915 .06476 L .983 .841 .946 r
F
.63924 .06487 m .63919 .06466 L .63933 .065 L 1 .756 .566 r
F
.63629 .05842 m .63652 .05894 L p .63655 .059 L .714 .344 .02 r
F P
0 g
s
.63629 .05842 m .63655 .059 L .63634 .05852 L .714 .344 .02 r
F
.63934 .06518 m .63957 .06567 L .63956 .06565 L F
.63932 .06514 m .63956 .06565 L .63679 .0595 L F
.63934 .06518 m .63956 .06565 L .63932 .06514 L F
.63937 .06505 m .63937 .06504 L .63846 .06329 L .983 .841 .946 r
F
.63846 .06327 m .63846 .06329 L .63131 .0494 L F
.63938 .06512 m .63684 .05989 L .63938 .06512 L 1 .764 .752 r
F
.63934 .06514 m .63892 .06421 L .63871 .06379 L .904 .716 .866 r
F
.6397 .06596 m .63956 .06566 L .6397 .06596 L .714 .344 .02 r
F
.63956 .06566 m .63679 .0595 L .63956 .06565 L F
.63956 .06566 m .63956 .06565 L .6397 .06596 L F
.63892 .06421 m .63761 .06135 L .6387 .06375 L .904 .716 .866 r
F
.63871 .06379 m .63761 .06135 L .63556 .05703 L F
.63963 .06553 m .63964 .06554 L .64234 .07132 L 1 .756 .566 r
F
.63971 .06597 m .63988 .06636 L p .63988 .06636 L 1 .667 .308 r
F P
0 g
s
.63986 .06631 m .6397 .06596 L .63986 .06631 L .714 .344 .02 r
F
.63986 .06631 m .6397 .06596 L .6397 .06596 L F
.63987 .06633 m .6397 .06596 L .63987 .06633 L F
.63988 .06636 m .63957 .06567 L p .63987 .06634 L F P
0 g
s
.63976 .06609 m .63987 .06633 L .6397 .06596 L .714 .344 .02 r
F
.63976 .06609 m .6397 .06596 L .63966 .06588 L F
.63954 .06561 m .64093 .06826 L .63986 .0663 L 1 .75 .424 r
F
.64036 .06714 m .63937 .06497 L .64194 .07054 L 1 .756 .566 r
F
.63957 .06548 m .63706 .06033 L .63697 .06012 L 1 .764 .752 r
F
.63962 .06552 m .63948 .06544 L .63971 .06564 L .957 .829 .959 r
F
.63956 .06547 m .63706 .06033 L .63957 .06548 L 1 .764 .752 r
F
.63954 .06541 m .63951 .06543 L .63962 .06552 L .957 .829 .959 r
F
.63937 .0651 m .63954 .06545 L p .63938 .06518 L .947 .734 .843 r
F P
0 g
s
.63955 .06556 m .6397 .06579 L .64003 .06642 L .105 .13 .703 r
F
.63996 .06636 m .63947 .06542 L p .64003 .06642 L F P
0 g
s
.63947 .06542 m .63946 .06541 L p .63947 .06542 L .105 .13 .703 r
F P
0 g
s
.63947 .06542 m .63947 .06542 L .64003 .06642 L .105 .13 .703 r
F
.63962 .06567 m .64036 .06714 L .64081 .06811 L .947 .734 .843 r
F
.63938 .06518 m .63965 .06572 L .63962 .06567 L F
.63965 .06572 m .64036 .06714 L .63962 .06567 L F
.63948 .06544 m .63948 .06545 L .63997 .06636 L .957 .829 .959 r
F
.63948 .06545 m .6395 .06545 L .64102 .06856 L .947 .734 .843 r
F
.64121 .06885 m .63956 .06547 L .63957 .06548 L 1 .764 .752 r
F
.6441 .0748 m .64128 .069 L .64121 .06885 L F
.63964 .06554 m .63972 .06564 L .64234 .07132 L 1 .756 .566 r
F
.64128 .069 m .63956 .06547 L .64121 .06885 L 1 .764 .752 r
F
.64761 .08297 m .64808 .08297 L p .64235 .07167 L .957 .829 .959 r
F P
0 g
s
.6445 .07599 m .64542 .078 L .64043 .06717 L .137 .162 .724 r
F
.64449 .07596 m .64043 .06717 L .64053 .06739 L F
.64498 .07702 m .6445 .07599 L .64449 .07596 L F
.6445 .07599 m .64043 .06717 L .64449 .07596 L F
.63963 .06552 m .64005 .06646 L .63962 .06552 L 1 .764 .752 r
F
.63969 .06557 m .63967 .06552 L .63963 .0654 L .137 .162 .724 r
F
.63971 .06561 m .63968 .06554 L .6397 .06558 L F
.63952 .06522 m .63977 .06577 L .63977 .06577 L 1 .756 .566 r
F
.63963 .0654 m .63967 .06552 L .63967 .06552 L .137 .162 .724 r
F
.63948 .06511 m .64045 .0672 L .63949 .06513 L F
.63967 .06551 m .63969 .06557 L .63969 .06557 L .957 .829 .959 r
F
.63963 .0654 m .63967 .06551 L .63967 .06551 L F
.63967 .06551 m .63969 .06557 L .63967 .06551 L F
.63967 .06551 m .63945 .06514 L .63947 .06507 L .863 .698 .883 r
F
.63967 .06551 m .63966 .06551 L .63971 .06564 L .957 .829 .959 r
F
.63972 .06565 m .63952 .06522 L p .63971 .06563 L .137 .162 .724 r
F P
0 g
s
.63965 .06554 m .63945 .06514 L p .63967 .06551 L .863 .698 .883 r
F P
0 g
s
.63961 .06537 m .63954 .06541 L .63971 .06564 L .957 .829 .959 r
F
.63969 .06556 m .63965 .06554 L .63967 .06551 L .863 .698 .883 r
F
.63969 .06556 m .63967 .06551 L .63968 .06549 L F
.63963 .0654 m .63961 .06537 L p .63966 .06551 L .957 .829 .959 r
F P
0 g
s
.63963 .0654 m .63966 .06551 L .63967 .06551 L .957 .829 .959 r
F
.63968 .06551 m .63968 .0657 L .63962 .06539 L 1 .857 .741 r
F
.63977 .06577 m .63969 .06556 L .63968 .06551 L F
.64017 .06655 m .64013 .06652 L .6397 .06558 L .137 .162 .724 r
F
.63969 .06557 m .6397 .06558 L .63967 .06552 L F
.64017 .06655 m .6397 .06558 L .63969 .06557 L F
.63985 .06582 m .63977 .06577 L .63968 .06551 L 1 .857 .741 r
F
.63977 .06577 m .63973 .06574 L .63969 .06556 L F
.63963 .0654 m .63971 .06564 L .63972 .06564 L .957 .829 .959 r
F
.63972 .0657 m .63965 .06554 L p .63969 .06556 L .863 .698 .883 r
F P
0 g
s
.63972 .06573 m .64044 .06719 L .64005 .06646 L 1 .764 .752 r
F
.63973 .06574 m .63963 .06553 L .63973 .06574 L 1 .756 .566 r
F
.63945 .06514 m .63972 .0657 L p .63972 .06573 L 1 .764 .752 r
F P
0 g
s
.63973 .06574 m .63972 .06573 L .63972 .0657 L 1 .857 .741 r
F
.63972 .0657 m .64044 .06719 L p .63972 .06573 L 1 .764 .752 r
F P
0 g
s
.63972 .0657 m .63968 .0657 L .63968 .06551 L 1 .857 .741 r
F
.64003 .06642 m .63978 .06594 L p .63982 .06587 L F P
0 g
s
.63978 .06594 m .6397 .06581 L p .63975 .06581 L 1 .857 .741 r
F P
0 g
s
.63962 .06552 m .64005 .06646 L .63957 .06548 L 1 .764 .752 r
F
.63956 .06544 m .63957 .06545 L .6402 .06683 L 1 .756 .566 r
F
.63948 .06545 m .64102 .06856 L .64142 .06942 L .947 .734 .843 r
F
.63946 .06541 m .63954 .06553 L .63955 .06556 L .105 .13 .703 r
F
.63947 .06542 m .63955 .06556 L .64003 .06642 L F
.63946 .06541 m .63955 .06556 L .63947 .06542 L F
.63954 .06545 m .63973 .06585 L p .63952 .06545 L .947 .734 .843 r
F P
0 g
s
.63973 .06585 m .64036 .06714 L p .63965 .06572 L .947 .734 .843 r
F P
0 g
s
.63952 .06545 m .63973 .06585 L .63965 .06572 L .947 .734 .843 r
F
.63951 .06543 m .63948 .06544 L .6395 .06545 L .957 .829 .959 r
F
.63954 .06545 m .6395 .06545 L .63962 .06552 L F
.63938 .06518 m .63954 .06545 L .63952 .06545 L .947 .734 .843 r
F
.63951 .06543 m .6395 .06545 L .63954 .06545 L .957 .829 .959 r
F
.63977 .06577 m .64234 .07133 L p .63977 .06577 L 1 .756 .566 r
F P
0 g
s
.64194 .07054 m .63973 .06574 L .64234 .07132 L 1 .756 .566 r
F
.63997 .06636 m .63948 .06545 L .64001 .06648 L .957 .829 .959 r
F
.64247 .07158 m .64055 .06743 L .64248 .07161 L .137 .162 .724 r
F
.64055 .06743 m .63968 .06554 L .64064 .06762 L F
.64055 .06743 m .64064 .06762 L .64248 .07161 L F
.64157 .06963 m .63937 .0651 L .63938 .06511 L 1 .764 .752 r
F
.64642 .07962 m .64164 .06978 L .64157 .06963 L F
.64164 .06978 m .63937 .0651 L .64157 .06963 L F
.64234 .07132 m .63952 .06522 L .64234 .07133 L 1 .756 .566 r
F
.63952 .06522 m .63952 .06522 L .64234 .07133 L F
.6395 .06545 m .63952 .06545 L .64204 .07065 L .947 .734 .843 r
F
.64446 .09112 m .64443 .09122 L .64443 .09122 L 1 1 .703 r
F
.64446 .09112 m .64443 .09122 L .64446 .09112 L F
.64443 .09122 m .64438 .09146 L .64443 .09122 L F
.64362 .09454 m .6443 .09063 L p .64441 .09096 L F P
0 g
s
.64441 .09096 m .64446 .09112 L .64438 .09146 L 1 1 .703 r
F
.64441 .09096 m .64438 .09146 L .64362 .09454 L F
.64446 .09112 m .64438 .09146 L p .64446 .09112 L F P
0 g
s
.64438 .09146 m .64362 .09454 L p .64438 .09146 L 1 1 .703 r
F P
0 g
s
.64566 .09302 m .64652 .09454 L .64446 .09114 L p .094 .094 .669 r
F P
0 g
s
.64362 .09454 m .64445 .09114 L .64652 .09454 L .947 .735 .845 r
F
.64581 .08297 m .64362 .09454 L .64875 .08297 L closepath p .586 .963 1 r
F P
0 g
s
.64219 .07097 m .64204 .07065 L .64821 .08334 L .947 .734 .843 r
F
.6395 .06545 m .64204 .07065 L .64219 .07097 L F
.64154 .06955 m .63971 .06579 L .63971 .06579 L 1 .764 .752 r
F
.64642 .07962 m .64156 .0696 L .64154 .06955 L F
.64156 .0696 m .63971 .06579 L .64154 .06955 L F
.64055 .06743 m .63968 .06554 L .63968 .06554 L .137 .162 .724 r
F
.64496 .07699 m .64053 .06739 L .64055 .06743 L F
.64053 .06739 m .63968 .06554 L .64055 .06743 L F
.64635 .08 m .64498 .07702 L .64496 .07699 L F
.6441 .0748 m .64367 .07395 L .64154 .06955 L 1 .764 .752 r
F
.64248 .07161 m .64064 .06762 L .64091 .0682 L .137 .162 .724 r
F
.64137 .0692 m .64154 .06955 L .63971 .06579 L 1 .764 .752 r
F
.6441 .0748 m .64154 .06955 L .64137 .0692 L F
.64498 .07702 m .64053 .06739 L .64496 .07699 L .137 .162 .724 r
F
.64053 .06739 m .63968 .06554 L .64053 .06739 L F
.64053 .06739 m .64053 .06739 L .64496 .07699 L F
.63954 .06553 m .6397 .06579 L .63955 .06556 L .105 .13 .703 r
F
.63949 .06545 m .63964 .06569 L .63954 .06541 L .904 .716 .866 r
F
.63947 .06542 m .63941 .06527 L .63949 .06545 L .947 .734 .843 r
F
.63935 .06519 m .63966 .06573 L .63954 .06553 L .105 .13 .703 r
F
.63966 .06573 m .6397 .06579 L .63954 .06553 L F
.63948 .06544 m .63949 .06545 L .63954 .06541 L .904 .716 .866 r
F
.63949 .06545 m .63972 .06594 L .63964 .06569 L F
.6397 .06581 m .63969 .06578 L p .63972 .06573 L 1 .857 .741 r
F P
0 g
s
.63972 .06594 m .6423 .07132 L .63968 .06571 L .904 .716 .866 r
F
.63971 .06579 m .63955 .06546 L .63971 .06579 L 1 .764 .752 r
F
.64026 .06693 m .63938 .06512 L .63938 .06512 L F
.63972 .06594 m .63968 .06571 L .63954 .06541 L .904 .716 .866 r
F
.64044 .06719 m .6441 .0748 L p .64121 .06885 L 1 .764 .752 r
F P
0 g
s
.64017 .06655 m .64294 .07257 L p .64013 .06652 L .137 .162 .724 r
F P
0 g
s
.64821 .08334 m .64805 .08334 L .64407 .07505 L p .947 .734 .843 r
F P
0 g
s
.64296 .07274 m .63954 .06561 L p .64172 .07007 L .947 .734 .843 r
F P
0 g
s
.64296 .07274 m .64266 .07209 L .63948 .06544 L .904 .716 .866 r
F
.64041 .06755 m .6419 .07046 L .64635 .08001 L 1 .756 .566 r
F
.64234 .07133 m .6462 .07969 L p .6419 .07046 L F P
0 g
s
.6462 .07969 m .64635 .08001 L p .6419 .07046 L 1 .756 .566 r
F P
0 g
s
.64436 .07568 m .64247 .07158 L .64431 .07558 L .137 .162 .724 r
F
.64496 .07699 m .64247 .07158 L .64466 .07634 L F
.64789 .08334 m .6424 .07145 L p .64045 .0672 L F P
0 g
s
.64407 .07505 m .64296 .07274 L p .6436 .07404 L .947 .734 .843 r
F P
0 g
s
.64407 .07505 m .64374 .07434 L .6423 .07132 L .904 .716 .866 r
F
.64266 .07209 m .6423 .07132 L .63948 .06544 L F
.64431 .07558 m .64247 .07158 L .64248 .07161 L .137 .162 .724 r
F
.64296 .07274 m .64268 .07204 L .6436 .07404 L .947 .734 .843 r
F
.64407 .07505 m .6423 .07132 L .64266 .07209 L .904 .716 .866 r
F
.64296 .07274 m .64172 .07007 L .64268 .07204 L .947 .734 .843 r
F
.6436 .07404 m .64268 .07204 L .64821 .08334 L F
.64294 .07257 m .64635 .08 L p .64466 .07634 L .137 .162 .724 r
F P
0 g
s
.64174 .06999 m .64466 .07634 L .64091 .0682 L .137 .162 .724 r
F
.64374 .07434 m .64805 .08334 L .64268 .07203 L .904 .716 .866 r
F
.64407 .07505 m .6436 .07404 L .64821 .08334 L .947 .734 .843 r
F
.64407 .07505 m .64805 .08334 L p .64374 .07434 L .904 .716 .866 r
F P
0 g
s
.64294 .07257 m .64466 .07634 L .64407 .07505 L .137 .162 .724 r
F
.63954 .06561 m .63948 .06545 L .64821 .08334 L .947 .734 .843 r
F
.6423 .0712 m .64407 .07505 L .64174 .06999 L .137 .162 .724 r
F
.64294 .07257 m .64407 .07505 L .6423 .0712 L F
.63973 .06574 m .63973 .06574 L .64234 .07132 L 1 .756 .566 r
F
.6424 .07145 m .63972 .06565 L p .64045 .0672 L .137 .162 .724 r
F P
0 g
s
.6441 .0748 m .63971 .06579 L .63971 .06578 L 1 .764 .752 r
F
.64294 .07257 m .64091 .0682 L .64013 .06652 L .137 .162 .724 r
F
.64642 .07964 m .64174 .06999 L p .64642 .07962 L 1 .764 .752 r
F P
0 g
s
.64174 .06999 m .63973 .06585 L p .64165 .0698 L 1 .764 .752 r
F P
0 g
s
.64642 .07962 m .64174 .06999 L .64165 .0698 L 1 .764 .752 r
F
.63973 .06585 m .63937 .0651 L p .63974 .06586 L F P
0 g
s
.64642 .07962 m .63973 .06585 L .63974 .06586 L 1 .764 .752 r
F
.62986 .04649 m .63215 .04495 L p .63131 .0494 L .983 .841 .946 r
F P
0 g
s
.63937 .0651 m .63922 .0652 L .6393 .06512 L .983 .841 .946 r
F
.63925 .06514 m .63918 .06498 L .63926 .06516 L 1 .667 .308 r
F
.63946 .06543 m .6309 .04857 L .63939 .06545 L .957 .829 .959 r
F
.63712 .06051 m .63937 .0651 L .6372 .06069 L .947 .734 .843 r
F
.63684 .05989 m .63578 .0577 L .63684 .05989 L 1 .764 .752 r
F
.63684 .05989 m .63684 .05989 L .63938 .06512 L F
.63938 .06512 m .63684 .05989 L .63683 .05986 L F
.63684 .05989 m .63578 .0577 L .63683 .05986 L F
.63583 .05781 m .63937 .0651 L p .63641 .05904 L .947 .734 .843 r
F P
0 g
s
.63628 .05874 m .63948 .06544 L .63492 .05568 L .904 .716 .866 r
F
.63534 .05752 m .6397 .06581 L p .63599 .05879 L .105 .13 .703 r
F P
0 g
s
.6397 .06581 m .64003 .06642 L p .63969 .06578 L .105 .13 .703 r
F P
0 g
s
.63929 .06525 m .63827 .06277 L .63923 .06513 L .888 .493 .074 r
F
.6393 .06512 m .63932 .06511 L .63836 .06312 L .904 .716 .866 r
F
.63932 .06511 m .63932 .0651 L .63836 .06312 L F
.63931 .06512 m .63925 .06514 L .63852 .06341 L .983 .841 .946 r
F
.63915 .06476 m .63918 .06483 L p .63915 .06477 L .714 .344 .02 r
F P
0 g
s
.63927 .06503 m .63915 .06476 L p .63924 .06487 L 1 .756 .566 r
F P
0 g
s
.63756 .06149 m .6393 .06512 L p .63722 .06074 L .904 .716 .866 r
F P
0 g
s
.63846 .06329 m .63852 .06341 L .63131 .0494 L .983 .841 .946 r
F
.63634 .05895 m .6388 .06398 L .63896 .06434 L .947 .734 .843 r
F
.6361 .05818 m .63184 .04931 L .63204 .04927 L p .863 .698 .883 r
F P
0 g
s
.63578 .05752 m .63556 .05703 L .63492 .05568 L .904 .716 .866 r
F
.63615 .05854 m .63712 .06051 L .6372 .06069 L .947 .734 .843 r
F
.63609 .05839 m .63937 .0651 L .63615 .05854 L F
.63836 .06312 m .63911 .06467 L .63704 .06034 L .904 .716 .866 r
F
.63911 .06467 m .63932 .0651 L .63905 .06453 L F
.63911 .06467 m .63905 .06453 L .63704 .06034 L F
.63867 .06369 m .63934 .06516 L .63863 .06361 L .714 .344 .02 r
F
.63918 .06483 m .63933 .06516 L .63867 .06369 L F
.63933 .06516 m .63934 .06516 L .63867 .06369 L F
.63722 .06074 m .63704 .06034 L .63576 .05765 L .904 .716 .866 r
F
.6393 .06512 m .63704 .06034 L .63722 .06074 L F
.63683 .05986 m .63578 .0577 L .63576 .05766 L 1 .764 .752 r
F
.63717 .06068 m .63756 .06149 L p .63576 .05765 L .904 .716 .866 r
F P
0 g
s
.63756 .06149 m .63722 .06074 L .63576 .05765 L .904 .716 .866 r
F
.63578 .0577 m .63455 .05517 L p .63576 .05766 L 1 .764 .752 r
F P
0 g
s
.63176 .04873 m .63167 .04884 L p .63408 .05391 L .904 .716 .866 r
F P
0 g
s
.63506 .05629 m .63609 .05839 L .63615 .05854 L .947 .734 .843 r
F
.6361 .05818 m .63697 .06012 L .6317 .04929 L 1 .764 .752 r
F
.63576 .05765 m .63628 .05874 L .63408 .05391 L .904 .716 .866 r
F
.63184 .04931 m .6361 .05818 L p .6317 .04929 L 1 .764 .752 r
F P
0 g
s
.63628 .05874 m .63492 .05568 L .63408 .05391 L .904 .716 .866 r
F
.63634 .05895 m .63717 .06068 L p .63576 .05765 L F P
0 g
s
.63455 .05517 m .63583 .05781 L p .63506 .05629 L .947 .734 .843 r
F P
0 g
s
.63506 .05629 m .63583 .05781 L .63641 .05904 L .947 .734 .843 r
F
.63697 .06012 m .63706 .06033 L .6317 .04929 L 1 .764 .752 r
F
.63697 .06012 m .6317 .04929 L .6317 .04929 L F
.63506 .05629 m .63634 .05895 L p .63167 .04884 L .904 .716 .866 r
F P
0 g
s
.63634 .05895 m .63408 .05391 L .63167 .04884 L .904 .716 .866 r
F
.63506 .05629 m .63156 .049 L .63455 .05517 L p .947 .734 .843 r
F P
0 g
s
.63455 .05517 m .63156 .049 L p .63452 .05509 L 1 .764 .752 r
F P
0 g
s
.6317 .04929 m .6317 .04929 L .63156 .049 L 1 .764 .752 r
F
.63156 .049 m .63184 .04931 L p .6317 .04929 L F P
0 g
s
.63167 .04884 m .63156 .049 L .63506 .05629 L p .904 .716 .866 r
F P
0 g
s
.63151 .04834 m .63846 .06327 L .63131 .0494 L .983 .841 .946 r
F
.6393 .0651 m .63902 .06456 L p .63929 .06507 L .105 .13 .703 r
F P
0 g
s
.63782 .06191 m .6309 .04857 L p .63889 .06433 L .957 .829 .959 r
F P
0 g
s
.6393 .0651 m .63933 .06517 L .63919 .0649 L 1 .75 .424 r
F
.63902 .06456 m .6393 .0651 L p .63919 .0649 L F P
0 g
s
.63929 .06507 m .63889 .06433 L .63946 .06543 L .957 .829 .959 r
F
.63782 .06191 m .63889 .06433 L .63929 .06507 L F
.63599 .05879 m .6397 .06581 L .63969 .06578 L .105 .13 .703 r
F
.63969 .06578 m .63964 .06568 L p .63968 .0657 L 1 .857 .741 r
F P
0 g
s
.63972 .06573 m .63969 .06578 L .63968 .0657 L 1 .857 .741 r
F
.63998 .06635 m .6402 .06683 L .6416 .06988 L 1 .756 .566 r
F
.63956 .06544 m .6402 .06683 L .63998 .06635 L F
.64027 .06696 m .63938 .06512 L .64026 .06693 L 1 .764 .752 r
F
.6423 .07132 m .64374 .07434 L .63977 .06589 L .904 .716 .866 r
F
.6423 .07132 m .63977 .06589 L .63954 .06541 L F
.64142 .06942 m .64102 .06856 L .64219 .07097 L .947 .734 .843 r
F
.64142 .06942 m .64219 .07097 L .64247 .07157 L F
.64374 .07434 m .64268 .07203 L .63954 .06541 L .904 .716 .866 r
F
.64466 .07634 m .64436 .07568 L .64431 .07558 L .137 .162 .724 r
F
.64635 .08 m .64496 .07699 L .64466 .07634 L F
.63957 .06545 m .63962 .06552 L .64036 .06714 L 1 .756 .566 r
F
.63957 .06545 m .64036 .06714 L .6416 .06988 L F
.63972 .06573 m .63968 .0657 L .63972 .0657 L 1 .857 .741 r
F
.64053 .06739 m .64043 .06717 L .63968 .06554 L .137 .162 .724 r
F
.64045 .0672 m .63972 .06565 L .63971 .06563 L F
.64044 .06719 m .63972 .0657 L p .63969 .06556 L .863 .698 .883 r
F P
0 g
s
.64044 .06719 m .64121 .06885 L .64005 .06646 L 1 .764 .752 r
F
.63954 .06541 m .64268 .07203 L .6396 .06538 L .904 .716 .866 r
F
.64247 .07157 m .64219 .07097 L .64821 .08334 L .947 .734 .843 r
F
.64036 .06714 m .64821 .08334 L p .64081 .06811 L F P
0 g
s
.64642 .07962 m .64027 .06696 L .64026 .06693 L 1 .764 .752 r
F
.64268 .07203 m .64805 .08334 L .64081 .06797 L .904 .716 .866 r
F
.64064 .06762 m .63971 .06561 L .64091 .0682 L .137 .162 .724 r
F
.64268 .07203 m .64081 .06797 L .6396 .06538 L .904 .716 .866 r
F
.63954 .06561 m .64296 .07274 L p .63948 .06544 L F P
0 g
s
.63948 .06545 m .63947 .06542 L .63949 .06545 L .947 .734 .843 r
F
.63948 .06545 m .63948 .06544 L .63946 .06541 L .904 .716 .866 r
F
.63948 .06544 m .63948 .06544 L .63946 .06541 L F
.63953 .0656 m .63957 .06567 L .63937 .06524 L .714 .344 .02 r
F
.63986 .0663 m .63987 .06635 L .63125 .05009 L 1 .75 .424 r
F
.63989 .06636 m .63988 .06636 L .63986 .0663 L F
.63988 .06636 m .63987 .06635 L .63986 .0663 L F
.64 .06665 m .63999 .06662 L .63954 .06561 L .714 .344 .02 r
F
.63953 .0656 m .63946 .06545 L p .63947 .06545 L .947 .734 .843 r
F P
0 g
s
.63946 .06545 m .63953 .0656 L p .63946 .06545 L .904 .716 .866 r
F P
0 g
s
.63951 .06543 m .63939 .06545 L .63946 .06545 L .957 .829 .959 r
F
.63933 .06517 m .63986 .0663 L .63919 .0649 L 1 .75 .424 r
F
.63999 .06662 m .63957 .06567 L .63953 .0656 L .714 .344 .02 r
F
.63953 .0656 m .63947 .06545 L .63948 .06545 L .947 .734 .843 r
F
.63946 .06545 m .63954 .06561 L .63947 .06545 L .904 .716 .866 r
F
.63953 .0656 m .63954 .06561 L p .63946 .06545 L F P
0 g
s
.63954 .06561 m .63953 .0656 L p .63948 .06545 L .947 .734 .843 r
F P
0 g
s
.63937 .06524 m .63957 .06567 L .63934 .06518 L .714 .344 .02 r
F
.64041 .06755 m .64 .06665 L .63937 .06524 L F
.63954 .06561 m .63944 .06536 L .63932 .0651 L .904 .716 .866 r
F
.63937 .06505 m .63939 .0651 L p .63937 .06506 L .983 .841 .946 r
F P
0 g
s
.63948 .06544 m .63954 .06541 L .63937 .06506 L .904 .716 .866 r
F
.63949 .06542 m .63947 .06542 L .63951 .06543 L .957 .829 .959 r
F
.63939 .0651 m .63931 .06512 L .63931 .06511 L .983 .841 .946 r
F
.6372 .06069 m .63937 .0651 L .63935 .06511 L .947 .734 .843 r
F
.63939 .0651 m .63931 .06511 L .63937 .06506 L .983 .841 .946 r
F
.63937 .0651 m .63578 .0577 L p .63938 .06511 L 1 .764 .752 r
F P
0 g
s
.63934 .06514 m .63937 .06506 L .63937 .06505 L .904 .716 .866 r
F
.63934 .06514 m .63937 .06505 L .63932 .0651 L F
.63939 .0651 m .6394 .06512 L p .63937 .0651 L .983 .841 .946 r
F P
0 g
s
.63948 .06544 m .63937 .06506 L .63934 .06514 L .904 .716 .866 r
F
.63949 .06535 m .63927 .06503 L .63934 .06499 L 1 .756 .566 r
F
.63945 .06521 m .63934 .06499 L .63935 .06498 L F
.63953 .06536 m .63935 .06498 L .63937 .06497 L F
.63945 .06514 m .63963 .06552 L .63945 .06514 L 1 .764 .752 r
F
.63915 .06476 m .63915 .06477 L .63867 .06369 L .714 .344 .02 r
F
.63937 .06505 m .63937 .06506 L .63852 .06341 L .983 .841 .946 r
F
.63862 .06359 m .63889 .06412 L .63924 .06487 L 1 .756 .566 r
F
.63937 .06505 m .63852 .06341 L .63846 .06329 L .983 .841 .946 r
F
.63652 .05894 m .63915 .06476 L p .63867 .06369 L .714 .344 .02 r
F P
0 g
s
.63556 .05703 m .63877 .06377 L .63386 .05323 L .904 .716 .866 r
F
.63839 .06308 m .63549 .05665 L p .6383 .06283 L 1 .756 .566 r
F P
0 g
s
.63839 .06308 m .6383 .06283 L .63889 .06412 L 1 .756 .566 r
F
.63862 .06359 m .63839 .06308 L p .63889 .06412 L F P
0 g
s
.6361 .05818 m .63204 .04927 L .63459 .05471 L .863 .698 .883 r
F
.63386 .05323 m .63877 .06377 L .63352 .05247 L .904 .716 .866 r
F
.63876 .0635 m .6322 .04922 L p .63854 .06287 L .947 .735 .844 r
F P
0 g
s
.63934 .06514 m .63556 .05703 L .63578 .05752 L .904 .716 .866 r
F
.6322 .04922 m .63189 .04855 L p .6322 .04922 L F P
0 g
s
.63189 .04855 m .6322 .04922 L p .6322 .04922 L .137 .162 .724 r
F P
0 g
s
.6322 .04922 m .63189 .04855 L .6322 .04922 L p .947 .735 .844 r
F P
0 g
s
.63408 .05391 m .63556 .05703 L .63199 .04924 L .904 .716 .866 r
F
.63199 .04924 m .63199 .04924 L .63176 .04873 L F
.63408 .05391 m .63199 .04924 L .63199 .04924 L F
.63556 .05703 m .63352 .05247 L .63176 .04873 L F
.63215 .04495 m .63846 .06327 L .63151 .04834 L .983 .841 .946 r
F
.63892 .06421 m .6387 .06375 L .63871 .06379 L .904 .716 .866 r
F
.63915 .06476 m .63862 .06359 L p .63924 .06487 L 1 .756 .566 r
F P
0 g
s
.63937 .06505 m .63846 .06329 L .63846 .06327 L .983 .841 .946 r
F
.63935 .06502 m .63915 .06476 L .63852 .06341 L F
.63937 .06506 m .63915 .06476 L .63935 .06502 L F
.63933 .065 m .63919 .06466 L .63934 .06499 L 1 .756 .566 r
F
.63927 .06503 m .63924 .06487 L .63933 .065 L F
.63927 .06476 m .63937 .06505 L p .63922 .06476 L .983 .841 .946 r
F P
0 g
s
.63927 .06476 m .63922 .06476 L .63846 .06327 L .983 .841 .946 r
F
.63215 .04495 m .63927 .06476 L p .63846 .06327 L F P
0 g
s
.63937 .06505 m .63937 .06505 L p .63937 .06504 L .983 .841 .946 r
F P
0 g
s
.63961 .06537 m .63782 .06191 L p .63929 .06508 L .957 .829 .959 r
F P
0 g
s
.63947 .06543 m .63939 .06545 L .63947 .06544 L .957 .829 .959 r
F
.63946 .06545 m .63939 .06545 L .63941 .06548 L F
.63928 .06508 m .63986 .0663 L .63933 .06532 L 1 .75 .424 r
F
.6393 .06512 m .63943 .06538 L .63931 .06511 L .904 .716 .866 r
F
.63918 .06483 m .64041 .06755 L p .63938 .06525 L .714 .344 .02 r
F P
0 g
s
.63933 .06516 m .63938 .06525 L .63934 .06518 L .714 .344 .02 r
F
.63918 .06483 m .63938 .06525 L .63933 .06516 L F
.6393 .06512 m .63946 .06545 L p .63943 .06538 L .904 .716 .866 r
F P
0 g
s
.63931 .06511 m .63943 .06538 L .63932 .0651 L .904 .716 .866 r
F
.6393 .0651 m .63929 .06507 L .63933 .06515 L .105 .13 .703 r
F
.6393 .0651 m .63947 .06542 L p .63948 .06545 L 1 .75 .424 r
F P
0 g
s
.6393 .0651 m .63948 .06545 L .63933 .06517 L 1 .75 .424 r
F
.6393 .0651 m .63933 .06515 L .63935 .06519 L .105 .13 .703 r
F
.63946 .06541 m .6393 .0651 L p .63935 .06519 L F P
0 g
s
.63947 .06543 m .63947 .06544 L .63951 .06543 L .957 .829 .959 r
F
.63953 .0654 m .63942 .06536 L .63949 .06542 L F
.63961 .06537 m .63951 .06537 L .63953 .0654 L F
.63951 .06537 m .63942 .06536 L .63953 .0654 L F
.63952 .06538 m .63934 .06499 L .63945 .06521 L 1 .756 .566 r
F
.64041 .06755 m .63949 .06535 L .63952 .06538 L F
.63949 .06535 m .63934 .06499 L .63952 .06538 L F
.63946 .06541 m .63948 .06544 L .63944 .06536 L .904 .716 .866 r
F
.63948 .06544 m .63935 .06519 L .63954 .06553 L .105 .13 .703 r
F
.63946 .06541 m .63935 .06519 L .63948 .06544 L F
.63946 .06545 m .63947 .06545 L .63932 .0651 L .904 .716 .866 r
F
.63944 .06536 m .63929 .06508 L .63942 .06536 L .957 .829 .959 r
F
.63951 .06537 m .63929 .06508 L .63944 .06536 L F
.63929 .06508 m .63929 .06508 L .63944 .06536 L F
.63931 .06511 m .63927 .06503 L .63949 .06535 L 1 .756 .566 r
F
.63961 .06537 m .63929 .06508 L .63951 .06537 L .957 .829 .959 r
F
.63961 .06563 m .63931 .06511 L .63949 .06535 L 1 .756 .566 r
F
.63933 .06515 m .63931 .06511 L .6397 .06579 L .105 .13 .703 r
F
.63931 .06511 m .63961 .06563 L .6397 .06579 L F
.63947 .06542 m .63947 .06543 L .63948 .06543 L .957 .829 .959 r
F
.63947 .06542 m .63948 .06543 L .63951 .06543 L F
.63949 .06542 m .63942 .06536 L .63946 .06543 L F
.63933 .06517 m .63948 .06545 L .63954 .06561 L 1 .75 .424 r
F
.64034 .06737 m .63988 .06636 L .64034 .06738 L 1 .667 .308 r
F
.63988 .06636 m .64034 .06738 L p .63988 .06636 L F P
0 g
s
.64034 .06738 m .6397 .06596 L .6397 .06596 L .714 .344 .02 r
F
.63933 .06515 m .6397 .06579 L .63935 .06519 L .105 .13 .703 r
F
.63954 .06561 m .63948 .06544 L .63948 .06545 L .904 .716 .866 r
F
.64041 .06755 m .64034 .06738 L .63987 .06633 L .714 .344 .02 r
F
.64041 .06755 m .63987 .06633 L .63976 .06609 L F
.64034 .06738 m .63988 .06636 L p .64034 .06738 L F P
0 g
s
.64034 .06738 m .63988 .06636 L .63987 .06634 L .714 .344 .02 r
F
.64034 .06738 m .63987 .06633 L .63987 .06633 L F
.64041 .06755 m .63929 .06507 L p .63961 .06563 L 1 .756 .566 r
F P
0 g
s
.63929 .06507 m .63927 .06503 L p .63961 .06563 L 1 .756 .566 r
F P
0 g
s
.63931 .06511 m .63906 .0646 L .63961 .06563 L .105 .13 .703 r
F
.63929 .06507 m .63906 .0646 L .63933 .06515 L F
.64296 .07274 m .64407 .07505 L p .64266 .07209 L .904 .716 .866 r
F P
0 g
s
.64498 .07702 m .64635 .08001 L .64542 .078 L .137 .162 .724 r
F
.64498 .07702 m .64542 .078 L .6445 .07599 L F
.64635 .08 m .64635 .08 L .64498 .07702 L F
.64635 .08 m .64635 .08001 L .64498 .07702 L F
.63954 .06541 m .63949 .06542 L .63951 .06543 L .957 .829 .959 r
F
.63963 .06552 m .63962 .06552 L .63945 .06514 L 1 .764 .752 r
F
.63961 .06537 m .63953 .0654 L .63954 .06541 L .957 .829 .959 r
F
.63953 .0654 m .63949 .06542 L .63954 .06541 L F
.63956 .06544 m .63935 .06498 L .63953 .06536 L 1 .756 .566 r
F
.64041 .06755 m .63952 .06538 L .63956 .06544 L F
.63952 .06538 m .63935 .06498 L .63956 .06544 L F
.63429 .05539 m .63366 .05418 L .62983 .04651 L .983 .841 .946 r
F
.62983 .04651 m .63069 .04847 L p .63469 .05621 L .957 .829 .959 r
F P
0 g
s
.63964 .06568 m .63033 .04801 L p .63962 .06539 L 1 .857 .741 r
F P
0 g
s
.63968 .0657 m .63964 .06568 L .63962 .06539 L 1 .857 .741 r
F
.63965 .06544 m .63965 .06545 L .63962 .06539 L F
.63974 .06561 m .63965 .06545 L .63965 .06544 L F
.63974 .06561 m .63962 .06539 L .63962 .06538 L F
.64081 .06797 m .63961 .06535 L .63961 .06535 L .904 .716 .866 r
F
.64017 .06655 m .63969 .06557 L .63962 .06536 L .137 .162 .724 r
F
.63969 .06557 m .63963 .0654 L .63962 .06536 L F
.64091 .0682 m .63971 .06561 L .6397 .06558 L F
.63963 .06546 m .64081 .06797 L .63961 .06535 L .904 .716 .866 r
F
.63941 .0649 m .6329 .04673 L .63981 .06311 L 1 .867 .915 r
F
.63959 .0653 m .64017 .06655 L p .63961 .06534 L .137 .162 .724 r
F P
0 g
s
.63959 .0653 m .63961 .06534 L .63942 .06493 L .137 .162 .724 r
F
.63962 .06538 m .63962 .06539 L .63033 .04801 L 1 .857 .741 r
F
.6441 .0748 m .64044 .06719 L p .64824 .08296 L .863 .698 .883 r
F P
0 g
s
.64805 .08334 m .63968 .06534 L p .64081 .06797 L .904 .716 .866 r
F P
0 g
s
.63945 .06483 m .64805 .08334 L p .63945 .06483 L .947 .735 .844 r
F P
0 g
s
.63968 .06534 m .6395 .06493 L p .63967 .06534 L .904 .716 .866 r
F P
0 g
s
.64081 .06797 m .63968 .06534 L .63967 .06534 L .904 .716 .866 r
F
.63941 .06489 m .63945 .06483 L .64805 .08334 L .947 .735 .844 r
F
.64259 .04731 m .65491 .04503 L p .64132 .05453 L .863 .698 .883 r
F P
0 g
s
.64805 .08334 m .64789 .08334 L .6394 .06489 L p .947 .735 .844 r
F P
0 g
s
.6394 .06489 m .63941 .06489 L .64805 .08334 L .947 .735 .844 r
F
.63033 .04801 m .62983 .04651 L .63968 .06549 L p 1 .857 .741 r
F P
0 g
s
.63968 .06549 m .63968 .0655 L .63033 .04801 L 1 .857 .741 r
F
.6416 .06988 m .64036 .06714 L .64194 .07054 L 1 .756 .566 r
F
.63929 .06507 m .63599 .05879 L .63906 .0646 L .105 .13 .703 r
F
.63902 .06456 m .63599 .05879 L p .63929 .06507 L F P
0 g
s
.63919 .0649 m .63928 .06508 L .63125 .05009 L 1 .75 .424 r
F
.63928 .06508 m .63933 .06532 L .63125 .05009 L F
.63907 .06491 m .63907 .06492 L .64652 .0811 L 1 .667 .308 r
F
.64467 .077 m .63987 .06633 L .63987 .06633 L .714 .344 .02 r
F
.64626 .08053 m .64467 .077 L .64467 .077 L F
.64467 .077 m .63987 .06633 L .64467 .077 L F
.64029 .06724 m .6405 .06764 L p .64024 .06712 L .752 .352 .035 r
F P
0 g
s
.64024 .06712 m .6405 .06764 L .64024 .06712 L .752 .352 .035 r
F
.64041 .06755 m .63956 .06544 L .6416 .06988 L 1 .756 .566 r
F
.65089 .05129 m .64526 .05842 L .65119 .04763 L .863 .698 .883 r
F
.64824 .08296 m .64526 .05842 L .65089 .05129 L F
.63931 .06553 m .63944 .06579 L .63931 .06553 L .957 .829 .959 r
F
.63944 .06579 m .63945 .06579 L .63931 .06553 L F
.63931 .06553 m .63931 .06553 L .63927 .06545 L F
.63998 .06695 m .64054 .06809 L .63973 .06665 L .991 .583 .139 r
F
.64054 .06809 m .63989 .06656 L p .6407 .06867 L .888 .493 .074 r
F P
0 g
s
.63904 .06487 m .6407 .06867 L p .63957 .06597 L .888 .493 .074 r
F P
0 g
s
.63885 .0645 m .63882 .06445 L .63892 .06465 L .752 .352 .035 r
F
.63938 .06555 m .63967 .06609 L p .63882 .06445 L F P
0 g
s
.63973 .06619 m .63938 .06555 L p .63973 .06619 L 1 .75 .424 r
F P
0 g
s
.64054 .06809 m .6444 .07715 L p .63973 .06665 L .991 .583 .139 r
F P
0 g
s
.63435 .0568 m .63184 .05144 L .63884 .06449 L p 1 .636 .21 r
F P
0 g
s
.63762 .0619 m .63902 .06456 L p .63125 .05009 L 1 .75 .424 r
F P
0 g
s
.63902 .06456 m .63919 .0649 L .63125 .05009 L 1 .75 .424 r
F
.63599 .05879 m .63033 .04801 L .63534 .05752 L p .105 .13 .703 r
F P
0 g
s
.63934 .06534 m .63991 .06641 L .63938 .06555 L 1 .75 .424 r
F
.63184 .05144 m .63125 .05009 L p .63934 .06534 L F P
0 g
s
.6407 .06867 m .63982 .06654 L .63963 .06611 L .888 .493 .074 r
F
.6407 .06867 m .6444 .07715 L .64054 .06809 L p F P
0 g
s
.64652 .0811 m .6444 .07715 L .63904 .06487 L p 1 .667 .308 r
F P
0 g
s
.63904 .06487 m .63907 .06491 L .64652 .0811 L 1 .667 .308 r
F
.63882 .06445 m .63967 .06609 L .63921 .06518 L .752 .352 .035 r
F
.63882 .06445 m .63184 .05144 L .63938 .06555 L p F P
0 g
s
.63938 .06555 m .63184 .05144 L p .63934 .06534 L 1 .75 .424 r
F P
0 g
s
.63125 .05009 m .63033 .04801 L .63762 .0619 L p 1 .75 .424 r
F P
0 g
s
.63981 .06311 m .63215 .04495 L .64259 .04731 L 1 .867 .915 r
F
.63215 .04495 m .63215 .04495 L .64259 .04731 L F
.62983 .04651 m .62986 .04649 L p .63131 .0494 L .983 .841 .946 r
F P
0 g
s
.64635 .08 m .64789 .08334 L p .64635 .08001 L .137 .162 .724 r
F P
0 g
s
.64626 .08053 m .64034 .06738 L p .64626 .08053 L .714 .344 .02 r
F P
0 g
s
.64626 .08053 m .64034 .06738 L .64034 .06738 L .714 .344 .02 r
F
.64821 .08334 m .6441 .0748 L p .64824 .08296 L .863 .698 .883 r
F P
0 g
s
.65491 .04503 m .65451 .04734 L p .65119 .04763 L .863 .698 .883 r
F P
0 g
s
.65451 .04734 m .64821 .08334 L p .65119 .04763 L .863 .698 .883 r
F P
0 g
s
.63089 .05723 m .63037 .06174 L .63022 .06135 L p .095 .095 .593 r
F P
0 g
s
.63024 .06119 m .63024 .06118 L .63022 .06119 L .601 .968 1 r
F
.62986 .06135 m .63022 .06135 L p .63022 .06119 L F P
0 g
s
.63019 .06135 m .63016 .06135 L p .63087 .06315 L .957 .827 .957 r
F P
0 g
s
.6441 .0748 m .64642 .07962 L p .64367 .07395 L 1 .764 .752 r
F P
0 g
s
.64041 .06755 m .64626 .08053 L p .64034 .06738 L .714 .344 .02 r
F P
0 g
s
.64598 .07991 m .64041 .06755 L p .646 .07927 L 1 .756 .566 r
F P
0 g
s
.64598 .07991 m .646 .07927 L .64635 .08001 L 1 .756 .566 r
F
.64626 .08053 m .64598 .07991 L p .64635 .08001 L F P
0 g
s
.64626 .08053 m .64652 .0811 L .64626 .08053 L p .714 .344 .02 r
F P
0 g
s
.64789 .08334 m .64652 .0811 L .64626 .08053 L p 1 .756 .566 r
F P
0 g
s
.64635 .08001 m .64789 .08334 L p .64626 .08053 L 1 .756 .566 r
F P
0 g
s
.64642 .07962 m .64821 .08334 L .64642 .07964 L p 1 .764 .752 r
F P
0 g
s
.64433 .09047 m .64384 .08953 L .64394 .08901 L p .107 .11 .687 r
F P
0 g
s
.64384 .08953 m .64433 .09047 L p .63966 .08333 L .23 .385 .884 r
F P
0 g
s
.63105 .06363 m .63087 .06315 L .63016 .06135 L .957 .827 .957 r
F
.63105 .06363 m .63037 .06174 L p .63184 .06571 L .261 .384 .875 r
F P
0 g
s
.64567 .09493 m .64525 .09872 L p .64578 .09528 L 1 1 .703 r
F P
0 g
s
.64652 .09454 m .64362 .09454 L .64362 .09454 L p .947 .735 .845 r
F P
0 g
s
.64146 .08979 m .63757 .08027 L p .64181 .09074 L .846 .769 .986 r
F P
0 g
s
.63757 .08027 m .63716 .07926 L p .63755 .08021 L .846 .769 .986 r
F P
0 g
s
.64181 .09074 m .63757 .08027 L .63755 .08021 L .846 .769 .986 r
F
.6393 .08522 m .6436 .09653 L .63972 .08634 L p .078 .078 .537 r
F P
0 g
s
.64181 .09074 m .64509 .09872 L .64146 .08979 L p .846 .769 .986 r
F P
0 g
s
.64151 .09264 m .6436 .09653 L .6393 .08522 L p .556 .59 .97 r
F P
0 g
s
.6393 .08522 m .63812 .08254 L .64151 .09264 L .556 .59 .97 r
F
.63869 .08707 m .64151 .09264 L .63664 .07871 L p .46 .526 .947 r
F P
0 g
s
.89281 .06105 m .89843 .06107 L .89104 .06143 L closepath p .955 .786 .906 r
F P
0 g
s
.89843 .06107 m .90404 .06109 L .89431 .06226 L closepath p .955 .786 .906 r
F P
0 g
s
.89431 .06226 m .89104 .06143 L .89843 .06107 L closepath p .955 .786 .906 r
F P
0 g
s
.89834 .06425 m .89431 .06226 L .90404 .06109 L closepath p .955 .786 .906 r
F P
0 g
s
.92719 .06297 m .9223 .09257 L .90168 .09211 L closepath p .955 .786 .906 r
F P
0 g
s
.91861 .11512 m .90168 .09211 L .9223 .09257 L closepath p .955 .786 .906 r
F P
0 g
s
.90916 .09872 m .91221 .09912 L .91527 .06113 L closepath p .284 .19 .619 r
F P
0 g
s
.91907 .05689 m .91527 .06113 L .91221 .09912 L closepath p .284 .19 .619 r
F P
0 g
s
.90799 .09725 m .91527 .06113 L .90916 .09872 L closepath p .955 .786 .906 r
F P
0 g
s
.90593 .0937 m .91527 .06113 L .90799 .09725 L closepath p .955 .786 .906 r
F P
0 g
s
.90295 .08807 m .91527 .06113 L .90593 .0937 L closepath p .955 .786 .906 r
F P
0 g
s
.89905 .08036 m .91527 .06113 L .90295 .08807 L closepath p .955 .786 .906 r
F P
0 g
s
.89905 .08036 m .89531 .07286 L .91527 .06113 L closepath p .955 .786 .906 r
F P
0 g
s
.46397 .05861 m .46629 .0612 L .4643 .05286 L closepath p 1 .986 .659 r
F P
0 g
s
.46629 .0612 m .46397 .05861 L .46497 .06469 L closepath p 1 1 .867 r
F P
0 g
s
.46497 .06469 m .46629 .0612 L .47154 .06771 L closepath p .955 .786 .906 r
F P
0 g
s
.25063 .02699 m .24685 .01445 L .25252 .02426 L closepath p 1 1 .85 r
F P
0 g
s
.24877 .01169 m .25252 .02426 L .24685 .01445 L closepath p 1 1 .85 r
F P
0 g
s
.94616 .027 m .95292 .01446 L .94792 .02427 L closepath p .662 .854 .448 r
F P
0 g
s
.95471 .01169 m .94792 .02427 L .95292 .01446 L closepath p .662 .854 .448 r
F P
0 g
s
.68266 .027 m .68543 .01446 L .68447 .02427 L closepath p .845 1 .639 r
F P
0 g
s
.68726 .01169 m .68447 .02427 L .68543 .01446 L closepath p .845 1 .639 r
F P
0 g
s
.89053 .05786 m .89104 .06143 L .89373 .0568 L closepath p .934 1 1 r
F P
0 g
s
.89104 .06143 m .89053 .05786 L .89048 .06258 L closepath p 1 1 .818 r
F P
0 g
s
.89104 .06143 m .89431 .06226 L .89048 .06258 L closepath p .955 .786 .906 r
F P
0 g
s
.81205 .01178 m .80739 .02436 L .81783 .00825 L closepath p .794 1 .982 r
F P
0 g
s
.80739 .02436 m .81205 .01178 L .80271 .02916 L closepath p .805 1 .776 r
F P
0 g
s
.35577 .02605 m .35359 .0135 L .35651 .02441 L closepath p .994 1 .648 r
F P
0 g
s
.35433 .01184 m .35651 .02441 L .35359 .0135 L closepath p .994 1 .648 r
F P
0 g
s
.29677 .02605 m .29369 .0135 L .29751 .02441 L closepath p 1 1 .681 r
F P
0 g
s
.29444 .01184 m .29751 .02441 L .29369 .0135 L closepath p 1 1 .681 r
F P
0 g
s
.19448 .02706 m .18985 .01453 L .19645 .02442 L closepath p 1 1 .888 r
F P
0 g
s
.19185 .01184 m .19645 .02442 L .18985 .01453 L closepath p 1 1 .888 r
F P
0 g
s
.40834 .02605 m .40695 .0135 L .40907 .02442 L closepath p .969 1 .615 r
F P
0 g
s
.40769 .01185 m .40907 .02442 L .40695 .0135 L closepath p .969 1 .615 r
F P
0 g
s
.89001 .02707 m .89592 .01453 L .89185 .02442 L closepath p .718 .914 .534 r
F P
0 g
s
.89779 .01185 m .89185 .02442 L .89592 .01453 L closepath p .718 .914 .534 r
F P
0 g
s
.79254 .02605 m .79698 .0135 L .79322 .02442 L closepath p .698 .792 .314 r
F P
0 g
s
.79767 .01185 m .79322 .02442 L .79698 .0135 L closepath p .698 .792 .314 r
F P
0 g
s
.72966 .02605 m .73314 .0135 L .73034 .02442 L closepath p .751 .837 .369 r
F P
0 g
s
.73383 .01185 m .73034 .02442 L .73314 .0135 L closepath p .751 .837 .369 r
F P
0 g
s
.62674 .02707 m .62866 .01453 L .62863 .02442 L closepath p .88 1 .706 r
F P
0 g
s
.63058 .01185 m .62863 .02442 L .62866 .01453 L closepath p .88 1 .706 r
F P
0 g
s
.51599 .0263 m .51623 .01375 L .51672 .02473 L closepath p .912 .978 .564 r
F P
0 g
s
.51697 .01216 m .51672 .02473 L .51623 .01375 L closepath p .912 .978 .564 r
F P
0 g
s
.30992 .01228 m .31276 .02484 L .3093 .00991 L closepath p .007 .007 .201 r
F P
0 g
s
.31276 .02484 m .30992 .01228 L .31447 .02659 L closepath p .101 .313 .831 r
F P
0 g
s
.56042 .01244 m .55952 .02501 L .55973 .01022 L closepath p .071 .071 .46 r
F P
0 g
s
.55952 .02501 m .56042 .01244 L .56135 .0266 L closepath p .139 .485 .962 r
F P
0 g
s
.07362 .02667 m .06716 .01413 L .07447 .02502 L closepath p 1 1 .81 r
F P
0 g
s
.06802 .01246 m .07447 .02502 L .06716 .01413 L closepath p 1 1 .81 r
F P
0 g
s
.89431 .06226 m .89834 .06425 L .89102 .06398 L closepath p .955 .786 .906 r
F P
0 g
s
.89102 .06398 m .89048 .06258 L .89431 .06226 L closepath p .955 .786 .906 r
F P
0 g
s
.04017 .02574 m .0332 .01318 L .04091 .02517 L closepath p .989 1 1 r
F P
0 g
s
.03395 .01261 m .04091 .02517 L .0332 .01318 L closepath p .989 1 1 r
F P
0 g
s
.04091 .02517 m .03395 .01261 L .04201 .02607 L closepath p .082 .337 .825 r
F P
0 g
s
.8891 .06061 m .89048 .06258 L .89053 .05786 L closepath p 1 1 .844 r
F P
0 g
s
.89048 .06258 m .8891 .06061 L .89102 .06398 L closepath p .846 .489 .054 r
F P
0 g
s
.54806 .03281 m .54879 .02036 L .5444 .02549 L closepath p .093 .093 .636 r
F P
0 g
s
.54507 .01293 m .5444 .02549 L .54879 .02036 L closepath p .093 .093 .636 r
F P
0 g
s
.03854 .02644 m .03155 .0139 L .04017 .02574 L closepath p .879 1 1 r
F P
0 g
s
.0332 .01318 m .04017 .02574 L .03155 .0139 L closepath p .879 1 1 r
F P
0 g
s
.36372 .0132 m .36575 .02575 L .36265 .00991 L closepath p .031 .031 .328 r
F P
0 g
s
.36575 .02575 m .36372 .0132 L .36853 .02774 L closepath p .128 .52 .974 r
F P
0 g
s
.83755 .01329 m .83251 .02585 L .82842 .01532 L closepath p .703 1 1 r
F P
0 g
s
.22545 .09257 m .22708 .09106 L .23172 .06296 L closepath p .111 .111 .272 r
F P
0 g
s
.23335 .06141 m .23172 .06296 L .22708 .09106 L closepath p .111 .111 .272 r
F P
0 g
s
.21635 .11543 m .23172 .06296 L .22545 .09257 L closepath p .955 .786 .906 r
F P
0 g
s
.9223 .09257 m .92569 .09106 L .92719 .06297 L closepath p .289 .191 .621 r
F P
0 g
s
.93058 .06143 m .92719 .06297 L .92569 .09106 L closepath p .289 .191 .621 r
F P
0 g
s
.65835 .09257 m .66107 .09106 L .66371 .06297 L closepath p .17 .17 .525 r
F P
0 g
s
.66644 .06142 m .66371 .06297 L .66107 .09106 L closepath p .17 .17 .525 r
F P
0 g
s
.65423 .11512 m .66371 .06297 L .65835 .09257 L closepath p .955 .786 .906 r
F P
0 g
s
.65423 .11512 m .64525 .09872 L .66371 .06297 L closepath p .955 .786 .906 r
F P
0 g
s
.69001 .01331 m .68718 .02586 L .68874 .00991 L closepath p .104 .104 .605 r
F P
0 g
s
.68718 .02586 m .69001 .01331 L .69021 .02776 L closepath p .299 .645 1 r
F P
0 g
s
.14629 .11449 m .14499 .11211 L .1218 .063 L closepath p .955 .786 .906 r
F P
0 g
s
.13297 .08528 m .1341 .08217 L .1218 .063 L closepath p .195 .355 .864 r
F P
0 g
s
.12297 .05998 m .1218 .063 L .1341 .08217 L closepath p .195 .355 .865 r
F P
0 g
s
.647 .0134 m .6448 .02595 L .64572 .00965 L closepath p .092 .092 .545 r
F P
0 g
s
.6448 .02595 m .647 .0134 L .64805 .02787 L closepath p .303 .658 1 r
F P
0 g
s
.03018 .01341 m .03719 .02596 L .02896 .0103 L closepath p .771 .422 .031 r
F P
0 g
s
.03719 .02596 m .03018 .01341 L .03854 .02644 L closepath p .304 .73 1 r
F P
0 g
s
.91444 .01344 m .90825 .02599 L .91304 .0099 L closepath p .14 .166 .727 r
F P
0 g
s
.90825 .02599 m .91444 .01344 L .91149 .02787 L closepath p .383 .701 1 r
F P
0 g
s
.20845 .01345 m .2128 .026 L .20719 .00989 L closepath p .001 .001 .239 r
F P
0 g
s
.2128 .026 m .20845 .01345 L .21598 .02786 L closepath p .152 .574 .997 r
F P
0 g
s
.19153 .06312 m .20493 .09211 L .19372 .06796 L closepath p .955 .786 .906 r
F P
0 g
s
.62361 .06312 m .63789 .09211 L .62595 .06796 L closepath p .955 .786 .906 r
F P
0 g
s
.35385 .02714 m .35164 .0146 L .35577 .02605 L closepath p .885 1 1 r
F P
0 g
s
.35359 .0135 m .35577 .02605 L .35164 .0146 L closepath p .885 1 1 r
F P
0 g
s
.29484 .02714 m .29173 .01461 L .29677 .02605 L closepath p .894 1 1 r
F P
0 g
s
.29369 .0135 m .29677 .02605 L .29173 .01461 L closepath p .894 1 1 r
F P
0 g
s
.40643 .02714 m .40501 .01461 L .40834 .02605 L closepath p .876 1 1 r
F P
0 g
s
.40695 .0135 m .40834 .02605 L .40501 .01461 L closepath p .876 1 1 r
F P
0 g
s
.79067 .02714 m .79507 .01461 L .79254 .02605 L closepath p .793 1 .998 r
F P
0 g
s
.79698 .0135 m .79254 .02605 L .79507 .01461 L closepath p .793 1 .998 r
F P
0 g
s
.72778 .02714 m .73123 .01461 L .72966 .02605 L closepath p .809 1 1 r
F P
0 g
s
.73314 .0135 m .72966 .02605 L .73123 .01461 L closepath p .809 1 1 r
F P
0 g
s
.03507 .01352 m .04201 .02607 L .03395 .01261 L closepath p .082 .337 .825 r
F P
0 g
s
.04201 .02607 m .03507 .01352 L .04576 .02806 L closepath p .135 .58 .989 r
F P
0 g
s
.068 .07024 m .0748 .10468 L .06605 .06319 L closepath p .955 .786 .906 r
F P
0 g
s
.08816 .01353 m .09431 .02608 L .08948 .00922 L closepath p 1 1 .655 r
F P
0 g
s
.09431 .02608 m .08816 .01353 L .09327 .03147 L closepath p 1 .974 .552 r
F P
0 g
s
.4114 .01366 m .41273 .0262 L .40989 .00991 L closepath p .054 .054 .457 r
F P
0 g
s
.41273 .0262 m .4114 .01366 L .41651 .02806 L closepath p .292 .682 1 r
F P
0 g
s
.25083 .01366 m .25455 .02621 L .24925 .00991 L closepath p .023 .023 .358 r
F P
0 g
s
.25455 .02621 m .25083 .01366 L .25848 .0279 L closepath p .292 .701 1 r
F P
0 g
s
.11875 .01372 m .12444 .02627 L .11759 .01037 L closepath p .743 .419 .02 r
F P
0 g
s
.12444 .02627 m .11875 .01372 L .12747 .02816 L closepath p .112 .523 .96 r
F P
0 g
s
.51407 .02727 m .51428 .01474 L .51599 .0263 L closepath p .838 1 1 r
F P
0 g
s
.51623 .01375 m .51599 .0263 L .51428 .01474 L closepath p .838 1 1 r
F P
0 g
s
.81222 .07258 m .81524 .07004 L .81001 .06349 L closepath p .373 .39 .838 r
F P
0 g
s
.81311 .06125 m .81001 .06349 L .81524 .07004 L closepath p .373 .39 .838 r
F P
0 g
s
.81001 .06349 m .81222 .07258 L .81102 .07966 L closepath p .955 .786 .906 r
F P
0 g
s
.03155 .0139 m .03854 .02644 L .03018 .01341 L closepath p .304 .73 1 r
F P
0 g
s
.75495 .01404 m .75115 .02658 L .75256 .00991 L closepath p .133 .236 .789 r
F P
0 g
s
.75115 .02658 m .75495 .01404 L .75669 .02806 L closepath p .49 .846 1 r
F P
0 g
s
.31166 .01405 m .31447 .02659 L .30992 .01228 L closepath p .101 .313 .831 r
F P
0 g
s
.31447 .02659 m .31166 .01405 L .3202 .02806 L closepath p .437 .827 1 r
F P
0 g
s
.56228 .01406 m .56135 .0266 L .56042 .01244 L closepath p .139 .485 .962 r
F P
0 g
s
.56135 .0266 m .56228 .01406 L .5673 .02791 L closepath p .493 .863 1 r
F P
0 g
s
.51394 .07395 m .51534 .07025 L .51058 .06377 L closepath p .185 .287 .799 r
F P
0 g
s
.51237 .06059 m .51058 .06377 L .51534 .07025 L closepath p .217 .301 .803 r
F P
0 g
s
.07146 .02769 m .06496 .01516 L .07362 .02667 L closepath p .892 1 1 r
F P
0 g
s
.06716 .01413 m .07362 .02667 L .06496 .01516 L closepath p .892 1 1 r
F P
0 g
s
.10438 .0638 m .10283 .07088 L .08462 .12811 L closepath p .955 .786 .906 r
F P
0 g
s
.08462 .12811 m .08301 .12799 L .10438 .0638 L closepath p .955 .786 .906 r
F P
0 g
s
.10283 .07088 m .10423 .06884 L .10439 .0638 L closepath p .085 .085 .165 r
F P
0 g
s
.10545 .06256 m .10439 .0638 L .10423 .06884 L closepath p .075 .075 .144 r
F P
0 g
s
.09697 .06327 m .10545 .06256 L .10423 .06884 L p 1 .969 .542 r
F P
0 g
s
.10169 .07669 m .10301 .07512 L .10283 .07088 L closepath p .09 .09 .199 r
F P
0 g
s
.10423 .06884 m .10283 .07088 L .10301 .07512 L closepath p .089 .089 .197 r
F P
0 g
s
.10423 .06884 m .10301 .07512 L .10294 .07499 L p 1 .969 .542 r
F P
0 g
s
.08738 .12827 m .08462 .12811 L .10283 .07088 L closepath p .955 .786 .906 r
F P
0 g
s
.10283 .07088 m .10169 .07669 L .08738 .12827 L closepath p .955 .786 .906 r
F P
0 g
s
.10294 .07499 m .09697 .06327 L p .10423 .06884 L 1 .969 .542 r
F P
0 g
s
.88969 .06236 m .89102 .06398 L .8891 .06061 L closepath p .906 .566 .07 r
F P
0 g
s
.89102 .06398 m .88969 .06236 L .89263 .06741 L closepath p .812 .425 .049 r
F P
0 g
s
.89263 .06741 m .89102 .06398 L .89834 .06425 L closepath p .955 .786 .906 r
F P
0 g
s
.24685 .01445 m .25063 .02699 L .24287 .01538 L closepath p .761 1 1 r
F P
0 g
s
.94229 .02791 m .94899 .01538 L .94616 .027 L closepath p .694 1 1 r
F P
0 g
s
.95292 .01446 m .94616 .027 L .94899 .01538 L closepath p .694 1 1 r
F P
0 g
s
.67877 .02791 m .68148 .01538 L .68266 .027 L closepath p .722 1 1 r
F P
0 g
s
.68543 .01446 m .68266 .027 L .68148 .01538 L closepath p .722 1 1 r
F P
0 g
s
.80794 .06415 m .80999 .07999 L .80365 .06964 L closepath p .955 .786 .906 r
F P
0 g
s
.19048 .0279 m .18579 .01538 L .19448 .02706 L closepath p .753 1 1 r
F P
0 g
s
.18985 .01453 m .19448 .02706 L .18579 .01538 L closepath p .753 1 1 r
F P
0 g
s
.88604 .02791 m .89189 .01538 L .89001 .02707 L closepath p .693 1 1 r
F P
0 g
s
.89592 .01453 m .89001 .02707 L .89189 .01538 L closepath p .693 1 1 r
F P
0 g
s
.62276 .02791 m .62462 .01538 L .62674 .02707 L closepath p .718 1 1 r
F P
0 g
s
.62866 .01453 m .62674 .02707 L .62462 .01538 L closepath p .718 1 1 r
F P
0 g
s
.8447 .01457 m .83955 .02711 L .83755 .01329 L closepath p .541 .895 1 r
F P
0 g
s
.83955 .02711 m .8447 .01457 L .84642 .03106 L closepath p .366 .694 1 r
F P
0 g
s
.3479 .02805 m .34559 .01553 L .35385 .02714 L closepath p .711 1 1 r
F P
0 g
s
.35164 .0146 m .35385 .02714 L .34559 .01553 L closepath p .711 1 1 r
F P
0 g
s
.28888 .02805 m .28568 .01553 L .29484 .02714 L closepath p .714 1 1 r
F P
0 g
s
.29173 .01461 m .29484 .02714 L .28568 .01553 L closepath p .714 1 1 r
F P
0 g
s
.40047 .02806 m .39897 .01554 L .40643 .02714 L closepath p .707 1 1 r
F P
0 g
s
.40501 .01461 m .40643 .02714 L .39897 .01554 L closepath p .707 1 1 r
F P
0 g
s
.78479 .02806 m .78911 .01554 L .79067 .02714 L closepath p .683 1 1 r
F P
0 g
s
.79507 .01461 m .79067 .02714 L .78911 .01554 L closepath p .683 1 1 r
F P
0 g
s
.72189 .02806 m .72526 .01554 L .72778 .02714 L closepath p .687 1 1 r
F P
0 g
s
.73123 .01461 m .72778 .02714 L .72526 .01554 L closepath p .687 1 1 r
F P
0 g
s
.3998 .07375 m .40188 .07147 L .39846 .06428 L closepath p .171 .176 .693 r
F P
0 g
s
.40053 .06199 m .39846 .06428 L .40188 .07147 L closepath p .171 .176 .693 r
F P
0 g
s
.39846 .06428 m .3998 .07375 L .40031 .08414 L closepath p .955 .786 .906 r
F P
0 g
s
.5082 .02791 m .50833 .01538 L .51407 .02727 L closepath p .677 1 1 r
F P
0 g
s
.51428 .01474 m .51407 .02727 L .50833 .01538 L closepath p .677 1 1 r
F P
0 g
s
.59217 .02819 m .59357 .01567 L .59895 .02748 L closepath p .672 1 1 r
F P
0 g
s
.60045 .01495 m .59895 .02748 L .59357 .01567 L closepath p .672 1 1 r
F P
0 g
s
.59895 .02748 m .60045 .01495 L .60444 .02864 L closepath p .501 .868 1 r
F P
0 g
s
.11012 .03431 m .10421 .02188 L .10661 .02751 L closepath p .007 .007 .351 r
F P
0 g
s
.10064 .01498 m .10661 .02751 L .10421 .02188 L closepath p .007 .007 .351 r
F P
0 g
s
.46306 .06129 m .46497 .06469 L .46397 .05861 L closepath p 1 1 .839 r
F P
0 g
s
.46497 .06469 m .46306 .06129 L .46125 .0667 L closepath p 1 1 1 r
F P
0 g
s
.46497 .06469 m .47154 .06771 L .46125 .0667 L closepath p .955 .786 .906 r
F P
0 g
s
.06496 .01516 m .07146 .02769 L .05814 .01584 L closepath p .692 1 1 r
F P
0 g
s
.36653 .01522 m .36853 .02774 L .36372 .0132 L closepath p .128 .52 .974 r
F P
0 g
s
.36853 .02774 m .36653 .01522 L .37341 .02805 L closepath p .578 .943 1 r
F P
0 g
s
.69309 .01524 m .69021 .02776 L .69001 .01331 L closepath p .299 .645 1 r
F P
0 g
s
.69021 .02776 m .69309 .01524 L .69192 .02795 L closepath p .564 .924 1 r
F P
0 g
s
.16738 .02825 m .16234 .01573 L .16967 .02778 L closepath p .752 1 1 r
F P
0 g
s
.16467 .01526 m .16967 .02778 L .16234 .01573 L closepath p .752 1 1 r
F P
0 g
s
.81998 .03042 m .82483 .01793 L .82352 .02784 L closepath p .797 1 .918 r
F P
0 g
s
.82842 .01532 m .82352 .02784 L .82483 .01793 L closepath p .797 1 .918 r
F P
0 g
s
.21167 .01534 m .21598 .02786 L .20845 .01345 L closepath p .152 .574 .997 r
F P
0 g
s
.91773 .01535 m .91149 .02787 L .91444 .01344 L closepath p .383 .701 1 r
F P
0 g
s
.91149 .02787 m .91773 .01535 L .91316 .02796 L closepath p .599 .954 1 r
F P
0 g
s
.6503 .01535 m .64805 .02787 L .647 .0134 L closepath p .303 .658 1 r
F P
0 g
s
.64805 .02787 m .6503 .01535 L .64972 .02796 L closepath p .592 .952 1 r
F P
0 g
s
.25482 .01537 m .25848 .0279 L .25083 .01366 L closepath p .292 .701 1 r
F P
0 g
s
.25848 .0279 m .25482 .01537 L .26501 .02834 L closepath p .571 .939 1 r
F P
0 g
s
.24281 .02887 m .23891 .01636 L .24671 .0279 L closepath p .77 1 1 r
F P
0 g
s
.24287 .01538 m .24671 .0279 L .23891 .01636 L closepath p .77 1 1 r
F P
0 g
s
.18579 .01538 m .19048 .0279 L .18196 .01547 L closepath p .635 .987 1 r
F P
0 g
s
.88328 .02791 m .88909 .01538 L .88604 .02791 L closepath p .619 .975 1 r
F P
0 g
s
.89189 .01538 m .88604 .02791 L .88909 .01538 L closepath p .619 .975 1 r
F P
0 g
s
.93843 .02887 m .94508 .01636 L .94229 .02791 L closepath p .698 1 1 r
F P
0 g
s
.94899 .01538 m .94229 .02791 L .94508 .01636 L closepath p .698 1 1 r
F P
0 g
s
.62174 .01538 m .61992 .02791 L .61803 .01566 L closepath p .656 1 1 r
F P
0 g
s
.61992 .02791 m .62174 .01538 L .62276 .02791 L closepath p .619 .975 1 r
F P
0 g
s
.62462 .01538 m .62276 .02791 L .62174 .01538 L closepath p .619 .975 1 r
F P
0 g
s
.56832 .01538 m .5673 .02791 L .56228 .01406 L closepath p .493 .863 1 r
F P
0 g
s
.5673 .02791 m .56832 .01538 L .57076 .02809 L closepath p .589 .95 1 r
F P
0 g
s
.50833 .01538 m .5082 .02791 L .5059 .01581 L closepath p .711 1 1 r
F P
0 g
s
.48839 .02791 m .48822 .01538 L .49151 .02791 L closepath p .619 .975 1 r
F P
0 g
s
.49138 .01538 m .49151 .02791 L .48822 .01538 L closepath p .619 .975 1 r
F P
0 g
s
.67496 .02887 m .67761 .01636 L .67877 .02791 L closepath p .729 1 1 r
F P
0 g
s
.68148 .01538 m .67877 .02791 L .67761 .01636 L closepath p .729 1 1 r
F P
0 g
s
.69483 .01543 m .69192 .02795 L .69309 .01524 L closepath p .564 .924 1 r
F P
0 g
s
.69192 .02795 m .69483 .01543 L .69519 .02806 L closepath p .602 .96 1 r
F P
0 g
s
.91943 .01544 m .91316 .02796 L .91773 .01535 L closepath p .599 .954 1 r
F P
0 g
s
.91316 .02796 m .91943 .01544 L .91634 .02798 L closepath p .616 .972 1 r
F P
0 g
s
.65199 .01544 m .64972 .02796 L .6503 .01535 L closepath p .592 .952 1 r
F P
0 g
s
.64972 .02796 m .65199 .01544 L .6529 .02798 L closepath p .615 .972 1 r
F P
0 g
s
.92265 .01546 m .91634 .02798 L .91943 .01544 L closepath p .616 .972 1 r
F P
0 g
s
.91634 .02798 m .92265 .01546 L .91875 .02817 L closepath p .589 .943 1 r
F P
0 g
s
.65522 .01546 m .6529 .02798 L .65199 .01544 L closepath p .615 .972 1 r
F P
0 g
s
.6529 .02798 m .65522 .01546 L .6553 .02817 L closepath p .58 .94 1 r
F P
0 g
s
.18196 .01547 m .18671 .02799 L .18034 .01557 L closepath p .664 1 1 r
F P
0 g
s
.34453 .02849 m .34218 .01598 L .3479 .02805 L closepath p .698 1 1 r
F P
0 g
s
.34559 .01553 m .3479 .02805 L .34218 .01598 L closepath p .698 1 1 r
F P
0 g
s
.28552 .02849 m .28227 .01598 L .28888 .02805 L closepath p .701 1 1 r
F P
0 g
s
.28568 .01553 m .28888 .02805 L .28227 .01598 L closepath p .701 1 1 r
F P
0 g
s
.37149 .01553 m .37341 .02805 L .36653 .01522 L closepath p .578 .943 1 r
F P
0 g
s
.41525 .01554 m .41651 .02806 L .4114 .01366 L closepath p .292 .682 1 r
F P
0 g
s
.41651 .02806 m .41525 .01554 L .42266 .02842 L closepath p .581 .945 1 r
F P
0 g
s
.39711 .0285 m .39555 .01598 L .40047 .02806 L closepath p .695 1 1 r
F P
0 g
s
.39897 .01554 m .40047 .02806 L .39555 .01598 L closepath p .695 1 1 r
F P
0 g
s
.78911 .01554 m .78479 .02806 L .78576 .01598 L closepath p .674 1 1 r
F P
0 g
s
.31747 .01554 m .3202 .02806 L .31166 .01405 L closepath p .437 .827 1 r
F P
0 g
s
.3202 .02806 m .31747 .01554 L .325 .02888 L closepath p .501 .881 1 r
F P
0 g
s
.69814 .01554 m .69519 .02806 L .69483 .01543 L closepath p .602 .96 1 r
F P
0 g
s
.69519 .02806 m .69814 .01554 L .69924 .02888 L closepath p .516 .877 1 r
F P
0 g
s
.76059 .01554 m .75669 .02806 L .75495 .01404 L closepath p .49 .846 1 r
F P
0 g
s
.75669 .02806 m .76059 .01554 L .75717 .02806 L closepath p .619 .975 1 r
F P
0 g
s
.76107 .01554 m .75717 .02806 L .76059 .01554 L closepath p .619 .975 1 r
F P
0 g
s
.75717 .02806 m .76107 .01554 L .76184 .02888 L closepath p .536 .894 1 r
F P
0 g
s
.44238 .02837 m .44151 .01585 L .44462 .02806 L closepath p .696 1 1 r
F P
0 g
s
.44378 .01554 m .44462 .02806 L .44151 .01585 L closepath p .696 1 1 r
F P
0 g
s
.47046 .01554 m .4709 .02806 L .44378 .01554 L closepath p .619 .975 1 r
F P
0 g
s
.03888 .01554 m .04576 .02806 L .03507 .01352 L closepath p .135 .58 .989 r
F P
0 g
s
.18358 .02842 m .17878 .0159 L .18512 .02809 L closepath p .755 1 1 r
F P
0 g
s
.18034 .01557 m .18512 .02809 L .17878 .0159 L closepath p .755 1 1 r
F P
0 g
s
.57183 .01557 m .57076 .02809 L .56832 .01538 L closepath p .589 .95 1 r
F P
0 g
s
.59199 .07223 m .59262 .07658 L .59096 .06524 L closepath p .955 .786 .906 r
F P
0 g
s
.59199 .07223 m .59096 .06524 L .59262 .07658 L closepath p .955 .786 .906 r
F P
0 g
s
.59199 .07223 m .59455 .07 L .59096 .06524 L closepath p .246 .261 .747 r
F P
0 g
s
.59352 .06298 m .59096 .06524 L .59455 .07 L closepath p .246 .261 .747 r
F P
0 g
s
.12182 .01564 m .12747 .02816 L .11875 .01372 L closepath p .112 .523 .96 r
F P
0 g
s
.12747 .02816 m .12182 .01564 L .1293 .02831 L closepath p .553 .928 1 r
F P
0 g
s
.9251 .01565 m .91875 .02817 L .92265 .01546 L closepath p .589 .943 1 r
F P
0 g
s
.91875 .02817 m .9251 .01565 L .91917 .02923 L closepath p .141 .174 .733 r
F P
0 g
s
.65766 .01565 m .6553 .02817 L .65522 .01546 L closepath p .58 .94 1 r
F P
0 g
s
.59044 .02891 m .59181 .0164 L .59217 .02819 L closepath p .797 1 1 r
F P
0 g
s
.59357 .01567 m .59217 .02819 L .59181 .0164 L closepath p .797 1 1 r
F P
0 g
s
.16234 .01573 m .16738 .02825 L .15861 .01593 L closepath p .655 1 1 r
F P
0 g
s
.49341 .01577 m .49351 .02829 L .49138 .01538 L closepath p .504 .876 1 r
F P
0 g
s
.12368 .01579 m .1293 .02831 L .12182 .01564 L closepath p .553 .928 1 r
F P
0 g
s
.1293 .02831 m .12368 .01579 L .13246 .02836 L closepath p .607 .967 1 r
F P
0 g
s
.26145 .01582 m .26501 .02834 L .25482 .01537 L closepath p .571 .939 1 r
F P
0 g
s
.12689 .01584 m .13246 .02836 L .12368 .01579 L closepath p .607 .967 1 r
F P
0 g
s
.06186 .02893 m .05521 .01642 L .06474 .02836 L closepath p .755 1 1 r
F P
0 g
s
.05814 .01584 m .06474 .02836 L .05521 .01642 L closepath p .755 1 1 r
F P
0 g
s
.44197 .03001 m .44109 .01752 L .44238 .02837 L closepath p .9 .873 .387 r
F P
0 g
s
.44151 .01585 m .44238 .02837 L .44109 .01752 L closepath p .9 .873 .387 r
F P
0 g
s
.18337 .02907 m .17857 .01657 L .18358 .02842 L closepath p 1 1 .623 r
F P
0 g
s
.17878 .0159 m .18358 .02842 L .17857 .01657 L closepath p 1 1 .623 r
F P
0 g
s
.42149 .01591 m .42266 .02842 L .41525 .01554 L closepath p .581 .945 1 r
F P
0 g
s
.15923 .0292 m .15407 .01669 L .16371 .02844 L closepath p .731 1 1 r
F P
0 g
s
.15861 .01593 m .16371 .02844 L .15407 .01669 L closepath p .731 1 1 r
F P
0 g
s
.34218 .01598 m .34453 .02849 L .34146 .01734 L closepath p 1 1 .716 r
F P
0 g
s
.28227 .01598 m .28552 .02849 L .28154 .01734 L closepath p 1 1 .744 r
F P
0 g
s
.39555 .01598 m .39711 .0285 L .39484 .01735 L closepath p .98 1 .686 r
F P
0 g
s
.57351 .01612 m .57241 .02863 L .57183 .01557 L closepath p .431 .802 1 r
F P
0 g
s
.60603 .01613 m .60444 .02864 L .60045 .01495 L closepath p .501 .868 1 r
F P
0 g
s
.28055 .10734 m .28706 .06578 L .29197 .1028 L closepath p .955 .786 .906 r
F P
0 g
s
.28869 .07809 m .29197 .1028 L .28706 .06578 L closepath p .955 .786 .906 r
F P
0 g
s
.24074 .03235 m .23681 .01989 L .24281 .02887 L closepath p 1 1 .798 r
F P
0 g
s
.23891 .01636 m .24281 .02887 L .23681 .01989 L closepath p 1 1 .798 r
F P
0 g
s
.67761 .01636 m .67496 .02887 L .67564 .01989 L closepath p .831 .956 .552 r
F P
0 g
s
.94456 .01734 m .93843 .02887 L .94316 .01989 L p .63 .791 .347 r
F P
0 g
s
.94316 .01989 m .93654 .03235 L .94013 .03341 L closepath p .058 .058 .06 r
F P
0 g
s
.94316 .01989 m .9389 .03081 L .9401 .02544 L p 1 1 .635 r
F P
0 g
s
.9401 .02544 m .94196 .01717 L .94316 .01989 L p 1 1 .635 r
F P
0 g
s
.94316 .01989 m .94508 .01636 L .94456 .01734 L p .63 .791 .347 r
F P
0 g
s
.32235 .01637 m .325 .02888 L .31747 .01554 L closepath p .501 .881 1 r
F P
0 g
s
.76581 .01637 m .76184 .02888 L .76107 .01554 L closepath p .536 .894 1 r
F P
0 g
s
.70226 .01638 m .69924 .02888 L .69814 .01554 L closepath p .516 .877 1 r
F P
0 g
s
.37592 .01639 m .37777 .0289 L .37149 .01553 L closepath p .489 .869 1 r
F P
0 g
s
.59181 .0164 m .59044 .02891 L .59144 .01835 L closepath p .755 .725 .19 r
F P
0 g
s
.06114 .03144 m .05448 .01897 L .06186 .02893 L closepath p 1 1 .653 r
F P
0 g
s
.05521 .01642 m .06186 .02893 L .05448 .01897 L closepath p 1 1 .653 r
F P
0 g
s
.17857 .01657 m .18337 .02907 L .17908 .01986 L closepath p .855 .612 .053 r
F P
0 g
s
.88475 .01657 m .879 .02908 L .88543 .01975 L closepath p .117 .117 .556 r
F P
0 g
s
.61729 .01657 m .61554 .02908 L .61792 .01969 L closepath p .065 .065 .37 r
F P
0 g
s
.16755 .01659 m .17251 .02909 L .16467 .01526 L closepath p .24 .663 1 r
F P
0 g
s
.8073 .01666 m .80271 .02916 L .81205 .01178 L closepath p .805 1 .776 r
F P
0 g
s
.80271 .02916 m .8073 .01666 L .79727 .04038 L closepath p .717 .837 .385 r
F P
0 g
s
.15806 .03206 m .15287 .01959 L .15923 .0292 L closepath p 1 1 .709 r
F P
0 g
s
.15407 .01669 m .15923 .0292 L .15287 .01959 L closepath p 1 1 .709 r
F P
0 g
s
.14298 .0646 m .14621 .06633 L .1421 .05411 L closepath p 1 .938 .656 r
F P
0 g
s
.14839 .09295 m .14621 .06633 L .15027 .07619 L closepath p .955 .786 .906 r
F P
0 g
s
.92553 .01673 m .91917 .02923 L .9251 .01565 L closepath p .141 .174 .733 r
F P
0 g
s
.65807 .01673 m .65571 .02923 L .65766 .01565 L closepath p .099 .099 .591 r
F P
0 g
s
.42326 .01689 m .42441 .0294 L .42149 .01591 L closepath p .248 .638 1 r
F P
0 g
s
.79678 .06498 m .79834 .06661 L .79564 .05448 L closepath p 1 .82 .361 r
F P
0 g
s
.79834 .06661 m .79678 .06498 L .80116 .07656 L closepath p .986 .663 .166 r
F P
0 g
s
.80116 .07656 m .79834 .06661 L .80365 .06964 L closepath p .955 .786 .906 r
F P
0 g
s
.4603 .06266 m .46125 .0667 L .46306 .06129 L closepath p 1 1 1 r
F P
0 g
s
.46125 .0667 m .4603 .06266 L .45451 .06737 L closepath p .875 1 1 r
F P
0 g
s
.45451 .06737 m .46125 .0667 L .47154 .06771 L closepath p .955 .786 .906 r
F P
0 g
s
.52701 .06434 m .52925 .06672 L .52926 .05227 L closepath p 1 1 .745 r
F P
0 g
s
.52925 .06672 m .52701 .06434 L .52683 .07932 L closepath p 1 1 .749 r
F P
0 g
s
.49478 .01726 m .49486 .02976 L .49341 .01577 L closepath p .121 .349 .868 r
F P
0 g
s
.47352 .01727 m .47392 .02976 L .47046 .01554 L closepath p .26 .643 1 r
F P
0 g
s
.34416 .03782 m .3418 .02544 L .34382 .02983 L closepath p .813 .64 .076 r
F P
0 g
s
.34146 .01734 m .34382 .02983 L .3418 .02544 L closepath p .813 .64 .076 r
F P
0 g
s
.28494 .0322 m .28167 .01974 L .2848 .02984 L closepath p .85 .663 .107 r
F P
0 g
s
.28154 .01734 m .2848 .02984 L .28167 .01974 L closepath p .85 .663 .107 r
F P
0 g
s
.39484 .01735 m .39641 .02985 L .39499 .01975 L closepath p .751 .577 .012 r
F P
0 g
s
.60722 .01739 m .60561 .02988 L .60603 .01613 L closepath p .135 .405 .911 r
F P
0 g
s
.87453 .0174 m .86893 .02989 L .87225 .01567 L closepath p .303 .61 1 r
F P
0 g
s
.57431 .01745 m .5732 .02994 L .57351 .01612 L closepath p .108 .159 .726 r
F P
0 g
s
.44109 .01752 m .44197 .03001 L .44122 .0196 L closepath p .713 .547 .002 r
F P
0 g
s
.45441 .06313 m .45451 .06737 L .4603 .06266 L closepath p .873 1 1 r
F P
0 g
s
.45451 .06737 m .45441 .06313 L .44461 .06737 L closepath p .843 1 1 r
F P
0 g
s
.44653 .06313 m .44461 .06737 L .45441 .06313 L closepath p .843 1 1 r
F P
0 g
s
.44461 .06737 m .45451 .06737 L .44534 .07128 L closepath p .955 .786 .906 r
F P
0 g
s
.47154 .06771 m .44534 .07128 L .45451 .06737 L closepath p .955 .786 .906 r
F P
0 g
s
.44958 .10073 m .44461 .06737 L .44534 .07128 L closepath p .955 .786 .906 r
F P
0 g
s
.44461 .06737 m .44958 .10073 L .43798 .10582 L closepath p .955 .786 .906 r
F P
0 g
s
.53343 .01777 m .53293 .03026 L .53526 .0092 L closepath p .815 .786 .269 r
F P
0 g
s
.53293 .03026 m .53343 .01777 L .53222 .03524 L closepath p .776 .716 .174 r
F P
0 g
s
.89138 .06609 m .89263 .06741 L .88969 .06236 L closepath p .839 .456 .056 r
F P
0 g
s
.89263 .06741 m .89138 .06609 L .89531 .07286 L closepath p .815 .42 .052 r
F P
0 g
s
.19702 .06742 m .19933 .07268 L .20184 .07939 L closepath p .955 .786 .906 r
F P
0 g
s
.21635 .11543 m .20952 .10197 L .19702 .06742 L closepath p .955 .786 .906 r
F P
0 g
s
.63167 .07268 m .6305 .07315 L .62921 .06742 L closepath p .955 .786 .906 r
F P
0 g
s
.81694 .03414 m .82175 .0217 L .81998 .03042 L closepath p .783 .998 .673 r
F P
0 g
s
.82483 .01793 m .81998 .03042 L .82175 .0217 L closepath p .783 .998 .673 r
F P
0 g
s
.43782 .10346 m .43734 .09916 L .4388 .06767 L closepath p .955 .786 .906 r
F P
0 g
s
.43652 .09292 m .4388 .06767 L .43734 .09916 L closepath p .955 .786 .906 r
F P
0 g
s
.47154 .06771 m .47226 .0725 L .45325 .07128 L closepath p .955 .786 .906 r
F P
0 g
s
.45891 .07173 m .45325 .07128 L .47226 .0725 L closepath p .955 .786 .906 r
F P
0 g
s
.47226 .0725 m .47298 .07729 L .45891 .07173 L closepath p .955 .786 .906 r
F P
0 g
s
.4633 .07307 m .45891 .07173 L .47298 .07729 L closepath p .955 .786 .906 r
F P
0 g
s
.4633 .07307 m .47298 .07729 L .45891 .07173 L closepath p .955 .786 .906 r
F P
0 g
s
.47298 .07729 m .47371 .08209 L .4633 .07307 L closepath p .955 .786 .906 r
F P
0 g
s
.46642 .0753 m .4633 .07307 L .47371 .08209 L closepath p .955 .786 .906 r
F P
0 g
s
.46642 .0753 m .47371 .08209 L .4633 .07307 L closepath p .955 .786 .906 r
F P
0 g
s
.47371 .08209 m .47443 .08688 L .46642 .0753 L closepath p .955 .786 .906 r
F P
0 g
s
.47443 .08688 m .4767 .08463 L .47154 .06771 L closepath p .185 .216 .721 r
F P
0 g
s
.47381 .06553 m .47154 .06771 L .4767 .08463 L closepath p .185 .216 .721 r
F P
0 g
s
.47154 .06771 m .45325 .07128 L .44534 .07128 L closepath p .955 .786 .906 r
F P
0 g
s
.49161 .03778 m .49148 .02539 L .48804 .0306 L closepath p .082 .082 .6 r
F P
0 g
s
.48786 .01812 m .48804 .0306 L .49148 .02539 L closepath p .082 .082 .6 r
F P
0 g
s
.38414 .06361 m .38711 .06785 L .3824 .05115 L closepath p 1 .89 .54 r
F P
0 g
s
.38711 .06785 m .38414 .06361 L .34629 .06785 L closepath p .843 1 1 r
F P
0 g
s
.34793 .06361 m .34629 .06785 L .38414 .06361 L closepath p .843 1 1 r
F P
0 g
s
.34687 .07168 m .33683 .0848 L .34629 .06785 L closepath p .955 .786 .906 r
F P
0 g
s
.34629 .06785 m .38711 .06785 L .34687 .07168 L closepath p .955 .786 .906 r
F P
0 g
s
.38771 .07168 m .34687 .07168 L .38711 .06785 L closepath p .955 .786 .906 r
F P
0 g
s
.27585 .06794 m .27296 .06628 L .28014 .1028 L closepath p 1 .911 .592 r
F P
0 g
s
.19372 .06796 m .20493 .09211 L .19649 .074 L closepath p .955 .786 .906 r
F P
0 g
s
.62595 .06796 m .63789 .09211 L .6289 .074 L closepath p .955 .786 .906 r
F P
0 g
s
.59061 .04095 m .59198 .02861 L .59008 .03083 L closepath p .032 .032 .153 r
F P
0 g
s
.59144 .01835 m .59008 .03083 L .59198 .02861 L closepath p .032 .032 .153 r
F P
0 g
s
.85167 .01859 m .84642 .03106 L .8447 .01457 L closepath p .366 .694 1 r
F P
0 g
s
.06155 .03625 m .0549 .02385 L .06114 .03144 L closepath p .982 .757 .246 r
F P
0 g
s
.05448 .01897 m .06114 .03144 L .0549 .02385 L closepath p .982 .757 .246 r
F P
0 g
s
.0871 .019 m .09327 .03147 L .08816 .01353 L closepath p 1 .974 .552 r
F P
0 g
s
.09327 .03147 m .0871 .019 L .09223 .03734 L closepath p 1 .963 .534 r
F P
0 g
s
.49552 .01903 m .49559 .0315 L .49478 .01726 L closepath p .073 .073 .53 r
F P
0 g
s
.70393 .01926 m .70088 .03173 L .70226 .01638 L closepath p .125 .208 .767 r
F P
0 g
s
.42418 .01927 m .42531 .03174 L .42326 .01689 L closepath p .054 .054 .447 r
F P
0 g
s
.76772 .01929 m .76372 .03175 L .76581 .01637 L closepath p .139 .294 .833 r
F P
0 g
s
.37767 .01929 m .3795 .03176 L .37592 .01639 L closepath p .075 .075 .625 r
F P
0 g
s
.71871 .0363 m .72203 .0239 L .71813 .03203 L closepath p .076 .076 .365 r
F P
0 g
s
.15829 .04364 m .1531 .03134 L .15806 .03206 L closepath p .962 .776 .26 r
F P
0 g
s
.15287 .01959 m .15806 .03206 L .1531 .03134 L closepath p .962 .776 .26 r
F P
0 g
s
.47584 .0196 m .4762 .03206 L .47352 .01727 L closepath p .122 .383 .891 r
F P
0 g
s
.45441 .06313 m .45371 .05111 L .44653 .06313 L closepath p .602 .969 1 r
F P
0 g
s
.44571 .05111 m .44653 .06313 L .45371 .05111 L closepath p .602 .969 1 r
F P
0 g
s
.44571 .05111 m .45359 .05139 L .44591 .05139 L p .098 .026 .026 r
F P
0 g
s
.44458 .04948 m .44653 .06313 L .44571 .05111 L p .014 .014 .151 r
F P
0 g
s
.44591 .05139 m .44349 .05139 L .44571 .05111 L p .098 .026 .026 r
F P
0 g
s
.44571 .05111 m .4421 .03206 L .44458 .04948 L p .014 .014 .151 r
F P
0 g
s
.44122 .0196 m .4421 .03206 L .44571 .05111 L closepath p .014 .014 .151 r
F P
0 g
s
.618 .03683 m .61979 .02443 L .61616 .03215 L closepath p .093 .093 .58 r
F P
0 g
s
.61792 .01969 m .61616 .03215 L .61979 .02443 L closepath p .093 .093 .58 r
F P
0 g
s
.28167 .01974 m .28494 .0322 L .28213 .02413 L closepath p .817 .609 .041 r
F P
0 g
s
.39499 .01975 m .39656 .03221 L .39548 .02414 L closepath p .714 .519 .003 r
F P
0 g
s
.88157 .03691 m .88736 .02452 L .87967 .03221 L closepath p .136 .158 .721 r
F P
0 g
s
.88543 .01975 m .87967 .03221 L .88736 .02452 L closepath p .136 .158 .721 r
F P
0 g
s
.17924 .02027 m .18388 .03232 L .18077 .02466 L p .69 .372 .006 r
F P
0 g
s
.18367 .03101 m .18077 .02466 L .1809 .02498 L p .013 .013 .34 r
F P
0 g
s
.1809 .02498 m .18554 .03705 L .18367 .03101 L p .013 .013 .34 r
F P
0 g
s
.18782 .03896 m .18367 .03101 L .18393 .0317 L p .024 .024 .413 r
F P
0 g
s
.18393 .0317 m .1884 .04331 L .18782 .03896 L p .024 .024 .413 r
F P
0 g
s
.18782 .03896 m .18877 .0405 L .18864 .0405 L p .994 .559 .151 r
F P
0 g
s
.1884 .04331 m .18367 .03101 L .18554 .03705 L closepath p .013 .013 .34 r
F P
0 g
s
.18864 .0405 m .18687 .04051 L .18782 .03896 L p .994 .559 .151 r
F P
0 g
s
.18367 .03101 m .18782 .03896 L .18725 .03878 L p 1 .673 .31 r
F P
0 g
s
.18725 .03878 m .1849 .03801 L .18367 .03101 L p 1 .673 .31 r
F P
0 g
s
.18077 .02466 m .18367 .03101 L .18288 .03071 L p 1 .715 .357 r
F P
0 g
s
.18288 .03071 m .18068 .02988 L .18077 .02466 L p 1 .715 .357 r
F P
0 g
s
.18077 .02466 m .17908 .01986 L .17924 .02027 L p .69 .372 .006 r
F P
0 g
s
.93355 .04568 m .94013 .03341 L .93654 .03235 L closepath p .058 .058 .06 r
F P
0 g
s
.23706 .0456 m .23307 .03333 L .24074 .03235 L closepath p 1 .98 .552 r
F P
0 g
s
.23681 .01989 m .24074 .03235 L .23307 .03333 L closepath p 1 .98 .552 r
F P
0 g
s
.66971 .04565 m .67229 .03338 L .67301 .03235 L closepath p .714 .722 .197 r
F P
0 g
s
.67564 .01989 m .67301 .03235 L .67229 .03338 L closepath p .714 .722 .197 r
F P
0 g
s
.80365 .06964 m .80786 .08315 L .80116 .07656 L closepath p .955 .786 .906 r
F P
0 g
s
.80786 .08315 m .80365 .06964 L .80999 .07999 L closepath p .955 .786 .906 r
F P
0 g
s
.78162 .03707 m .78589 .02468 L .78105 .03261 L closepath p .087 .087 .401 r
F P
0 g
s
.78531 .02015 m .78105 .03261 L .78589 .02468 L closepath p .087 .087 .401 r
F P
0 g
s
.56001 .05522 m .56092 .04309 L .54806 .03281 L closepath p .097 .097 .663 r
F P
0 g
s
.54879 .02036 m .54806 .03281 L .56092 .04309 L closepath p .097 .097 .663 r
F P
0 g
s
.53805 .08446 m .54047 .08291 L .54065 .07009 L closepath p .157 .157 .475 r
F P
0 g
s
.54308 .0685 m .54065 .07009 L .54047 .08291 L closepath p .157 .157 .475 r
F P
0 g
s
.54065 .07009 m .53805 .08446 L .53213 .11527 L closepath p .955 .786 .906 r
F P
0 g
s
.53213 .11527 m .52725 .11504 L .54065 .07009 L closepath p .955 .786 .906 r
F P
0 g
s
.07393 .0903 m .07487 .08735 L .068 .07024 L closepath p .141 .141 .681 r
F P
0 g
s
.0692 .06762 m .068 .07024 L .07487 .08735 L closepath p .148 .148 .699 r
F P
0 g
s
.068 .07024 m .07097 .08046 L .0748 .10468 L closepath p .955 .786 .906 r
F P
0 g
s
.8767 .02072 m .87107 .03317 L .87453 .0174 L closepath p .151 .338 .865 r
F P
0 g
s
.57507 .02072 m .57395 .03317 L .57431 .01745 L closepath p .061 .061 .376 r
F P
0 g
s
.17032 .02073 m .17525 .03318 L .16755 .01659 L closepath p .045 .045 .552 r
F P
0 g
s
.60917 .02074 m .60754 .03318 L .60722 .01739 L closepath p .112 .164 .731 r
F P
0 g
s
.42447 .02093 m .4256 .03337 L .42418 .01927 L closepath p .016 .016 .179 r
F P
0 g
s
.76816 .02138 m .76416 .03382 L .76772 .01929 L closepath p .096 .096 .485 r
F P
0 g
s
.82175 .0217 m .81694 .03414 L .81806 .03016 L closepath p .683 .789 .316 r
F P
0 g
s
.44958 .10073 m .4518 .09852 L .44534 .07128 L closepath p .175 .199 .708 r
F P
0 g
s
.44791 .07161 m .44534 .07128 L .4518 .09852 L closepath p .175 .199 .708 r
F P
0 g
s
.44791 .07161 m .45106 .08701 L .44912 .0736 L p .015 .015 .154 r
F P
0 g
s
.44791 .07161 m .45325 .07128 L .44771 .07128 L p .119 .031 .031 r
F P
0 g
s
.45313 .07161 m .45325 .07128 L .44859 .07157 L p .119 .031 .031 r
F P
0 g
s
.44859 .07157 m .44791 .07161 L .45313 .07161 L p .119 .031 .031 r
F P
0 g
s
.45325 .07128 m .45313 .07161 L .45891 .07173 L closepath p .177 .032 .032 r
F P
0 g
s
.45313 .07161 m .45241 .05971 L .45813 .07166 L p .558 .935 1 r
F P
0 g
s
.45826 .07194 m .45891 .07173 L .45496 .07165 L p .138 .024 .024 r
F P
0 g
s
.45496 .07165 m .45313 .07161 L .45826 .07194 L p .138 .024 .024 r
F P
0 g
s
.45891 .07173 m .45826 .07194 L .4633 .07307 L closepath p .344 .037 .037 r
F P
0 g
s
.4603 .06266 m .4597 .05064 L .45441 .06313 L closepath p .648 1 1 r
F P
0 g
s
.45826 .07194 m .45763 .06005 L .46176 .07225 L p .422 .815 1 r
F P
0 g
s
.462 .07296 m .4633 .07307 L .46072 .07249 L p .271 .023 .023 r
F P
0 g
s
.46072 .07249 m .45826 .07194 L .462 .07296 L p .271 .023 .023 r
F P
0 g
s
.4633 .07307 m .462 .07296 L .46642 .0753 L closepath p .63 .067 .062 r
F P
0 g
s
.46306 .06129 m .4625 .04925 L .4603 .06266 L closepath p .832 1 1 r
F P
0 g
s
.462 .07296 m .46142 .06108 L .46508 .07409 L p .112 .326 .848 r
F P
0 g
s
.46397 .07508 m .46642 .0753 L .462 .07296 L closepath p .93 .438 .131 r
F P
0 g
s
.46642 .0753 m .46397 .07508 L .46827 .07843 L closepath p 1 .631 .267 r
F P
0 g
s
.46595 .0772 m .46827 .07843 L .46531 .07612 L p .783 .273 .078 r
F P
0 g
s
.46531 .07612 m .46397 .07508 L .46595 .0772 L p .783 .273 .078 r
F P
0 g
s
.46827 .07843 m .46642 .0753 L .47443 .08688 L closepath p .955 .786 .906 r
F P
0 g
s
.46827 .07843 m .46595 .0772 L .46971 .08256 L closepath p 1 .734 .325 r
F P
0 g
s
.46753 .08116 m .46971 .08256 L .46595 .0772 L closepath p 1 .67 .233 r
F P
0 g
s
.46595 .0772 m .46544 .06538 L .46839 .08131 L p .06 .06 .474 r
F P
0 g
s
.46971 .08256 m .46753 .08116 L .47087 .08552 L closepath p 1 .673 .236 r
F P
0 g
s
.4691 .08513 m .47087 .08552 L .46753 .08116 L closepath p 1 .669 .23 r
F P
0 g
s
.46839 .08131 m .4691 .08513 L .46595 .0772 L p .06 .06 .474 r
F P
0 g
s
.46508 .07409 m .46595 .0772 L .46397 .07508 L p .112 .326 .848 r
F P
0 g
s
.46397 .07508 m .462 .07296 L p .46243 .07312 L .112 .326 .848 r
F P
0 g
s
.46243 .07312 m .46508 .07409 L .46397 .07508 L .112 .326 .848 r
F
.46176 .07225 m .462 .07296 L .45826 .07194 L p .422 .815 1 r
F P
0 g
s
.45813 .07166 m .45826 .07194 L .45313 .07161 L p .558 .935 1 r
F P
0 g
s
.44791 .07161 m .44712 .05971 L .45313 .07161 L closepath p .598 .968 1 r
F P
0 g
s
.44912 .0736 m .44712 .05971 L .44791 .07161 L p .015 .015 .154 r
F P
0 g
s
.44771 .07128 m .44534 .07128 L .44791 .07161 L p .119 .031 .031 r
F P
0 g
s
.12297 .05998 m .11724 .04791 L .11012 .03431 L closepath p .004 .004 .333 r
F P
0 g
s
.10421 .02188 m .11012 .03431 L .11724 .04791 L closepath p .004 .004 .333 r
F P
0 g
s
.34852 .08378 m .35063 .08262 L .34687 .07168 L closepath p .165 .165 .672 r
F P
0 g
s
.35181 .09108 m .35063 .08262 L .35059 .08265 L p .165 .165 .674 r
F P
0 g
s
.35059 .08265 m .34852 .08378 L .35181 .09108 L p .165 .165 .674 r
F P
0 g
s
.3527 .09715 m .35181 .09108 L .35174 .09116 L p .166 .166 .678 r
F P
0 g
s
.35174 .09116 m .34981 .09302 L .3527 .09715 L p .166 .166 .678 r
F P
0 g
s
.3527 .09715 m .35074 .09939 L .35325 .10057 L closepath p .167 .167 .69 r
F P
0 g
s
.3527 .09715 m .35325 .10057 L .35324 .10053 L p .722 .498 .006 r
F P
0 g
s
.34981 .09302 m .35181 .09108 L .34852 .08378 L closepath p .165 .165 .674 r
F P
0 g
s
.35074 .09939 m .3527 .09715 L .34981 .09302 L closepath p .166 .166 .678 r
F P
0 g
s
.35324 .10053 m .35101 .08909 L .3527 .09715 L p .722 .498 .006 r
F P
0 g
s
.35181 .09108 m .3527 .09715 L .35269 .09712 L p .733 .516 .009 r
F P
0 g
s
.35269 .09712 m .35045 .08562 L .35181 .09108 L p .733 .516 .009 r
F P
0 g
s
.35063 .08262 m .35181 .09108 L .35181 .09106 L p .737 .522 .01 r
F P
0 g
s
.34915 .07201 m .34687 .07168 L .35063 .08262 L p .165 .165 .671 r
F P
0 g
s
.35181 .09106 m .34956 .07947 L .35063 .08262 L p .737 .522 .01 r
F P
0 g
s
.35063 .08262 m .34918 .07201 L .34915 .07201 L p .165 .165 .671 r
F P
0 g
s
.34918 .07201 m .35063 .08262 L .34887 .07355 L p .74 .526 .011 r
F P
0 g
s
.34924 .07168 m .34687 .07168 L .34918 .07201 L p .12 .031 .031 r
F P
0 g
s
.34887 .07355 m .34836 .07088 L .34918 .07201 L p .74 .526 .011 r
F P
0 g
s
.38533 .0717 m .38771 .07168 L .38533 .07168 L p .12 .031 .031 r
F P
0 g
s
.38541 .07201 m .38771 .07168 L .34918 .07201 L closepath p .12 .031 .031 r
F P
0 g
s
.38771 .07168 m .39051 .08869 L .38909 .0802 L closepath p .955 .786 .906 r
F P
0 g
s
.38771 .07168 m .38541 .07201 L .38911 .08018 L closepath p 1 .876 .521 r
F P
0 g
s
.38653 .07888 m .38911 .08018 L .38541 .07201 L closepath p 1 .877 .525 r
F P
0 g
s
.38541 .07201 m .38367 .06012 L .38558 .07267 L p .003 .003 .124 r
F P
0 g
s
.38558 .07267 m .38653 .07888 L .38541 .07201 L p .003 .003 .124 r
F P
0 g
s
.34918 .07201 m .38533 .0717 L p .34924 .07168 L .12 .031 .031 r
F P
0 g
s
.38533 .07168 m .34924 .07168 L p .38533 .0717 L .12 .031 .031 r
F P
0 g
s
.34687 .07168 m .34852 .08378 L .34877 .09326 L closepath p .955 .786 .906 r
F P
0 g
s
.34687 .07168 m .35068 .11489 L .3392 .1059 L closepath p .955 .786 .906 r
F P
0 g
s
.39051 .08869 m .38771 .07168 L .40371 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.33907 .10333 m .34687 .07168 L .3392 .1059 L closepath p .955 .786 .906 r
F P
0 g
s
.33863 .09896 m .34687 .07168 L .33907 .10333 L closepath p .955 .786 .906 r
F P
0 g
s
.33788 .09278 m .34687 .07168 L .33863 .09896 L closepath p .955 .786 .906 r
F P
0 g
s
.33683 .0848 m .34687 .07168 L .33788 .09278 L closepath p .955 .786 .906 r
F P
0 g
s
.92524 .0224 m .91889 .03483 L .92553 .01673 L closepath p .086 .086 .277 r
F P
0 g
s
.37823 .02258 m .38005 .035 L .37767 .01929 L closepath p .005 .005 .137 r
F P
0 g
s
.78399 .0722 m .78648 .08667 L .78457 .09347 L closepath p .955 .786 .906 r
F P
0 g
s
.78648 .08667 m .78952 .0844 L .78399 .0722 L closepath p .341 .343 .799 r
F P
0 g
s
.78702 .06991 m .78399 .0722 L .78952 .0844 L closepath p .341 .343 .799 r
F P
0 g
s
.59262 .07658 m .59519 .07438 L .59199 .07223 L closepath p .245 .26 .746 r
F P
0 g
s
.59455 .07 m .59199 .07223 L .59519 .07438 L closepath p .245 .26 .746 r
F P
0 g
s
.53271 .02282 m .53222 .03524 L .53342 .01791 L p .776 .716 .174 r
F P
0 g
s
.53222 .03524 m .53271 .02282 L .53099 .04258 L closepath p .79 .74 .206 r
F P
0 g
s
.53147 .03026 m .53099 .04258 L .5327 .02295 L p .79 .74 .206 r
F P
0 g
s
.5327 .02295 m .53271 .02282 L .53147 .03026 L p .79 .74 .206 r
F P
0 g
s
.53099 .04258 m .53147 .03026 L .52926 .05227 L closepath p .797 .753 .223 r
F P
0 g
s
.5297 .04009 m .52926 .05227 L .53145 .03037 L p .797 .753 .223 r
F P
0 g
s
.53145 .03037 m .53147 .03026 L .5297 .04009 L p .797 .753 .223 r
F P
0 g
s
.52926 .05227 m .5297 .04009 L .52701 .06434 L closepath p .801 .76 .233 r
F P
0 g
s
.52742 .05234 m .52701 .06434 L .5297 .04016 L p .801 .76 .233 r
F P
0 g
s
.5297 .04016 m .5297 .04009 L .52742 .05234 L p .801 .76 .233 r
F P
0 g
s
.52701 .06434 m .52742 .05234 L .52463 .07687 L closepath p .802 .763 .237 r
F P
0 g
s
.525 .06505 m .52463 .07687 L .52742 .05234 L closepath p .802 .763 .237 r
F P
0 g
s
.52742 .05234 m .52984 .05072 L .52508 .06482 L p .153 .153 .452 r
F P
0 g
s
.52508 .06482 m .525 .06505 L .52742 .05234 L p .153 .153 .452 r
F P
0 g
s
.5297 .04009 m .53214 .03843 L .52759 .05184 L p .153 .153 .455 r
F P
0 g
s
.52759 .05184 m .52742 .05234 L .5297 .04009 L p .153 .153 .455 r
F P
0 g
s
.53147 .03026 m .53391 .02854 L .5299 .03955 L p .154 .154 .461 r
F P
0 g
s
.5299 .03955 m .5297 .04009 L .53147 .03026 L p .154 .154 .461 r
F P
0 g
s
.53271 .02282 m .53516 .02104 L .53172 .02963 L p .155 .155 .472 r
F P
0 g
s
.53172 .02963 m .53147 .03026 L .53271 .02282 L p .155 .155 .472 r
F P
0 g
s
.53342 .01791 m .53343 .01777 L .53271 .02282 L p .776 .716 .174 r
F P
0 g
s
.49682 .02284 m .49686 .03526 L .49552 .01903 L closepath p .061 .061 .445 r
F P
0 g
s
.55979 .07185 m .56142 .07252 L .55202 .05624 L closepath p .959 .554 .102 r
F P
0 g
s
.56142 .07252 m .55979 .07185 L .57149 .09166 L closepath p .946 .532 .099 r
F P
0 g
s
.57149 .09166 m .56142 .07252 L .57713 .09421 L closepath p .955 .786 .906 r
F P
0 g
s
.85649 .02299 m .85116 .03541 L .85167 .01859 L closepath p .24 .538 1 r
F P
0 g
s
.81576 .08181 m .81102 .07966 L .81222 .07258 L closepath p .955 .786 .906 r
F P
0 g
s
.81576 .08181 m .8187 .079 L .81222 .07258 L closepath p .42 .466 .901 r
F P
0 g
s
.81524 .07004 m .81222 .07258 L .8187 .079 L closepath p .419 .466 .901 r
F P
0 g
s
.47877 .02306 m .47909 .03547 L .47584 .0196 L closepath p .113 .295 .828 r
F P
0 g
s
.47909 .03547 m .47877 .02306 L .48491 .04216 L closepath p .115 .31 .839 r
F P
0 g
s
.20357 .08258 m .20184 .07939 L .19933 .07268 L closepath p .955 .786 .906 r
F P
0 g
s
.19933 .07268 m .20257 .07982 L .20357 .08258 L closepath p .955 .786 .906 r
F P
0 g
s
.6305 .07315 m .63167 .07268 L .63789 .09211 L closepath p .955 .786 .906 r
F P
0 g
s
.63512 .07982 m .63789 .09211 L .63167 .07268 L closepath p .955 .786 .906 r
F P
0 g
s
.70475 .02327 m .70169 .03568 L .70393 .01926 L closepath p .083 .083 .436 r
F P
0 g
s
.08533 .07049 m .08867 .07285 L .08827 .05683 L closepath p 1 1 .838 r
F P
0 g
s
.08867 .07285 m .08533 .07049 L .08547 .0872 L closepath p 1 1 .841 r
F P
0 g
s
.89414 .07178 m .89531 .07286 L .89138 .06609 L closepath p .824 .43 .054 r
F P
0 g
s
.89531 .07286 m .89414 .07178 L .89905 .08036 L closepath p .816 .417 .053 r
F P
0 g
s
.86636 .04383 m .87192 .03154 L .86421 .03622 L closepath p .122 .122 .613 r
F P
0 g
s
.86974 .02382 m .86421 .03622 L .87192 .03154 L closepath p .122 .122 .613 r
F P
0 g
s
.06303 .04386 m .05639 .03156 L .06155 .03625 L closepath p .914 .642 .104 r
F P
0 g
s
.0549 .02385 m .06155 .03625 L .05639 .03156 L closepath p .914 .642 .104 r
F P
0 g
s
.71967 .04268 m .723 .03037 L .71871 .0363 L closepath p .078 .078 .383 r
F P
0 g
s
.72203 .0239 m .71871 .0363 L .723 .03037 L closepath p .078 .078 .383 r
F P
0 g
s
.42495 .02396 m .42607 .03637 L .42447 .02093 L closepath p .013 .013 .157 r
F P
0 g
s
.28213 .02413 m .28538 .03653 L .2829 .03053 L closepath p .803 .588 .028 r
F P
0 g
s
.39747 .03901 m .3963 .03056 L .39631 .03059 L p .002 .002 .105 r
F P
0 g
s
.39631 .03059 m .39785 .04287 L .39747 .03901 L p .002 .002 .105 r
F P
0 g
s
.39902 .04995 m .39747 .03901 L .39747 .03901 L p .003 .003 .111 r
F P
0 g
s
.39747 .03901 m .399 .0512 L .39902 .04995 L p .003 .003 .111 r
F P
0 g
s
.40039 .05957 m .39902 .04995 L .39902 .04996 L p .003 .003 .111 r
F P
0 g
s
.39902 .04996 m .40053 .06199 L .40039 .05957 L p .003 .003 .111 r
F P
0 g
s
.40158 .06785 m .40039 .05957 L .40039 .05958 L p .003 .003 .112 r
F P
0 g
s
.40039 .05958 m .40188 .07147 L .40158 .06785 L p .003 .003 .112 r
F P
0 g
s
.40257 .0748 m .40158 .06785 L .40158 .06786 L p .003 .003 .113 r
F P
0 g
s
.40158 .06786 m .40305 .07964 L .40257 .0748 L p .003 .003 .113 r
F P
0 g
s
.40338 .0804 m .40257 .0748 L .40257 .07481 L p .004 .004 .114 r
F P
0 g
s
.40257 .07481 m .40403 .08648 L .40338 .0804 L p .004 .004 .114 r
F P
0 g
s
.404 .08466 m .40338 .0804 L .40338 .08041 L p .004 .004 .116 r
F P
0 g
s
.40338 .08041 m .40483 .09201 L .404 .08466 L p .004 .004 .116 r
F P
0 g
s
.40444 .08757 m .404 .08466 L .40401 .08468 L p .004 .004 .121 r
F P
0 g
s
.40401 .08468 m .40545 .0962 L .40444 .08757 L p .004 .004 .121 r
F P
0 g
s
.40469 .08913 m .40444 .08757 L .40448 .08793 L p .006 .006 .132 r
F P
0 g
s
.40448 .08793 m .40587 .09907 L .40469 .08913 L p .006 .006 .132 r
F P
0 g
s
.40688 .09442 m .40469 .08913 L .40518 .09312 L p .048 .048 .442 r
F P
0 g
s
.40518 .09312 m .40612 .10061 L .40688 .09442 L p .048 .048 .442 r
F P
0 g
s
.40688 .09442 m .40828 .10583 L .41142 .09592 L closepath p .354 .764 1 r
F P
0 g
s
.40688 .09442 m .41118 .09631 L .40771 .09506 L p .35 .03 .03 r
F P
0 g
s
.40545 .0962 m .404 .08466 L .40483 .09201 L closepath p .004 .004 .116 r
F P
0 g
s
.40587 .09907 m .40444 .08757 L .40545 .0962 L closepath p .004 .004 .121 r
F P
0 g
s
.40612 .10061 m .40469 .08913 L .40587 .09907 L closepath p .006 .006 .132 r
F P
0 g
s
.40828 .10583 m .40688 .09442 L .40612 .10061 L closepath p .048 .048 .442 r
F P
0 g
s
.40771 .09506 m .40513 .09413 L .40688 .09442 L p .35 .03 .03 r
F P
0 g
s
.40469 .08913 m .40688 .09442 L .40554 .0942 L p 1 .65 .207 r
F P
0 g
s
.40554 .0942 m .40513 .09413 L .40469 .08913 L p 1 .65 .207 r
F P
0 g
s
.40444 .08757 m .40469 .08913 L .40467 .08912 L p 1 .874 .511 r
F P
0 g
s
.40195 .08875 m .40403 .08648 L .40097 .0819 L closepath p .171 .177 .694 r
F P
0 g
s
.40403 .08648 m .40195 .08875 L .40483 .09201 L closepath p .171 .178 .695 r
F P
0 g
s
.40483 .09201 m .40338 .0804 L .40403 .08648 L closepath p .004 .004 .114 r
F P
0 g
s
.40467 .08912 m .40208 .08759 L .40444 .08757 L p 1 .874 .511 r
F P
0 g
s
.404 .08466 m .40444 .08757 L .40443 .08757 L p 1 .879 .519 r
F P
0 g
s
.40305 .07964 m .40097 .0819 L .40403 .08648 L closepath p .171 .177 .694 r
F P
0 g
s
.40403 .08648 m .40257 .0748 L .40305 .07964 L closepath p .003 .003 .113 r
F P
0 g
s
.40443 .08757 m .40183 .08601 L .404 .08466 L p 1 .879 .519 r
F P
0 g
s
.40338 .0804 m .404 .08466 L .404 .08466 L p 1 .881 .521 r
F P
0 g
s
.404 .08466 m .40139 .08308 L .40338 .0804 L p 1 .881 .521 r
F P
0 g
s
.40257 .0748 m .40338 .0804 L .40338 .0804 L p 1 .882 .523 r
F P
0 g
s
.40097 .0819 m .40305 .07964 L .3998 .07375 L closepath p .171 .177 .694 r
F P
0 g
s
.40188 .07147 m .3998 .07375 L .40305 .07964 L closepath p .171 .177 .694 r
F P
0 g
s
.40305 .07964 m .40158 .06785 L .40188 .07147 L closepath p .003 .003 .112 r
F P
0 g
s
.40338 .0804 m .40076 .0788 L .40257 .0748 L p 1 .882 .523 r
F P
0 g
s
.40158 .06785 m .40257 .0748 L .40257 .0748 L p 1 .883 .524 r
F P
0 g
s
.40188 .07147 m .40039 .05957 L .40053 .06199 L closepath p .003 .003 .111 r
F P
0 g
s
.40257 .0748 m .39995 .07317 L .40158 .06785 L p 1 .883 .524 r
F P
0 g
s
.40039 .05957 m .40158 .06785 L .40157 .06785 L p 1 .883 .524 r
F P
0 g
s
.40053 .06199 m .39902 .04995 L .399 .0512 L closepath p .003 .003 .111 r
F P
0 g
s
.40157 .06785 m .39895 .0662 L .40039 .05957 L p 1 .883 .524 r
F P
0 g
s
.39902 .04995 m .40039 .05957 L .40039 .05957 L p 1 .884 .525 r
F P
0 g
s
.399 .0512 m .39747 .03901 L .39785 .04287 L closepath p .002 .002 .105 r
F P
0 g
s
.40039 .05957 m .39776 .05789 L .39902 .04995 L p 1 .884 .525 r
F P
0 g
s
.39747 .03901 m .39902 .04995 L .39897 .04992 L p 1 .884 .525 r
F P
0 g
s
.39897 .04992 m .39639 .04824 L .39747 .03901 L p 1 .884 .525 r
F P
0 g
s
.3963 .03056 m .39747 .03901 L .39738 .03895 L p 1 .886 .528 r
F P
0 g
s
.39548 .0242 m .39704 .03654 L .3963 .03056 L p .001 .001 .092 r
F P
0 g
s
.39738 .03895 m .39484 .03727 L .3963 .03056 L p 1 .886 .528 r
F P
0 g
s
.3963 .03056 m .39548 .02414 L .39548 .0242 L p .001 .001 .092 r
F P
0 g
s
.3998 .07375 m .40096 .0819 L .40078 .08401 L closepath p .955 .786 .906 r
F P
0 g
s
.40371 .11489 m .40031 .08414 L .3998 .07375 L closepath p .955 .786 .906 r
F P
0 g
s
.51791 .08544 m .51901 .08162 L .51394 .07395 L closepath p .168 .259 .784 r
F P
0 g
s
.51534 .07025 m .51394 .07395 L .51901 .08162 L closepath p .18 .28 .793 r
F P
0 g
s
.62109 .04314 m .62293 .03084 L .618 .03683 L closepath p .104 .104 .665 r
F P
0 g
s
.61979 .02443 m .618 .03683 L .62293 .03084 L closepath p .104 .104 .665 r
F P
0 g
s
.06415 .07754 m .0675 .07879 L .05203 .03683 L closepath p 1 .836 .534 r
F P
0 g
s
.19358 .07301 m .19649 .074 L .1863 .05731 L closepath p 1 .717 .365 r
F P
0 g
s
.19649 .074 m .19358 .07301 L .20493 .09211 L closepath p 1 .716 .363 r
F P
0 g
s
.89126 .07301 m .89239 .074 L .88324 .05731 L closepath p .813 .41 .053 r
F P
0 g
s
.89239 .074 m .89126 .07301 L .90168 .09211 L closepath p .812 .408 .052 r
F P
0 g
s
.62709 .07302 m .6289 .074 L .61933 .05732 L closepath p .964 .56 .103 r
F P
0 g
s
.6289 .074 m .62709 .07302 L .63789 .09211 L closepath p .963 .558 .103 r
F P
0 g
s
.88474 .04322 m .89058 .03092 L .88157 .03691 L closepath p .143 .237 .787 r
F P
0 g
s
.88736 .02452 m .88157 .03691 L .89058 .03092 L closepath p .143 .237 .787 r
F P
0 g
s
.78255 .04325 m .78684 .03095 L .78162 .03707 L closepath p .091 .091 .428 r
F P
0 g
s
.78589 .02468 m .78162 .03707 L .78684 .03095 L closepath p .091 .091 .428 r
F P
0 g
s
.7688 .02473 m .76479 .03712 L .76816 .02138 L closepath p .093 .093 .462 r
F P
0 g
s
.08605 .02495 m .09223 .03734 L .08714 .01916 L p 1 .963 .534 r
F P
0 g
s
.09223 .03734 m .08605 .02495 L .09057 .04579 L closepath p 1 .976 .555 r
F P
0 g
s
.08435 .03352 m .09057 .04579 L .08608 .02509 L p 1 .976 .555 r
F P
0 g
s
.08608 .02509 m .08605 .02495 L .08435 .03352 L p 1 .976 .555 r
F P
0 g
s
.09057 .04579 m .08435 .03352 L .08827 .05683 L closepath p 1 .983 .566 r
F P
0 g
s
.08201 .04472 m .08827 .05683 L .08437 .03365 L p 1 .983 .566 r
F P
0 g
s
.08437 .03365 m .08435 .03352 L .08201 .04472 L p 1 .983 .566 r
F P
0 g
s
.08827 .05683 m .08201 .04472 L .08533 .07049 L closepath p 1 .987 .572 r
F P
0 g
s
.07903 .05857 m .08533 .07049 L .08203 .04486 L p 1 .987 .572 r
F P
0 g
s
.08203 .04486 m .08201 .04472 L .07903 .05857 L p 1 .987 .572 r
F P
0 g
s
.08533 .07049 m .07903 .05857 L .08218 .08477 L closepath p 1 .99 .578 r
F P
0 g
s
.07582 .07306 m .08218 .08477 L .07903 .05857 L closepath p 1 .99 .578 r
F P
0 g
s
.07903 .05857 m .0803 .05697 L .07597 .07255 L p .08 .08 .145 r
F P
0 g
s
.07597 .07255 m .07582 .07306 L .07903 .05857 L p .08 .08 .145 r
F P
0 g
s
.08201 .04472 m .0833 .04306 L .07919 .05799 L p .08 .08 .151 r
F P
0 g
s
.07919 .05799 m .07903 .05857 L .08201 .04472 L p .08 .08 .151 r
F P
0 g
s
.08435 .03352 m .08565 .03181 L .08219 .0441 L p .081 .081 .159 r
F P
0 g
s
.08219 .0441 m .08201 .04472 L .08435 .03352 L p .081 .081 .159 r
F P
0 g
s
.08605 .02495 m .08735 .02318 L .08456 .03281 L p .083 .083 .172 r
F P
0 g
s
.08456 .03281 m .08435 .03352 L .08605 .02495 L p .083 .083 .172 r
F P
0 g
s
.08714 .01916 m .0871 .019 L .08605 .02495 L p 1 .963 .534 r
F P
0 g
s
.49287 .04145 m .49277 .02912 L .49161 .03778 L closepath p .061 .061 .445 r
F P
0 g
s
.49148 .02539 m .49161 .03778 L .49277 .02912 L closepath p .061 .061 .445 r
F P
0 g
s
.34534 .04751 m .34299 .03527 L .34416 .03782 L closepath p .753 .545 .014 r
F P
0 g
s
.3418 .02544 m .34416 .03782 L .34299 .03527 L closepath p .753 .545 .014 r
F P
0 g
s
.50793 .04482 m .50805 .03254 L .5065 .03876 L closepath p .046 .046 .328 r
F P
0 g
s
.79972 .07533 m .80116 .07656 L .79678 .06498 L closepath p .985 .661 .164 r
F P
0 g
s
.80116 .07656 m .79972 .07533 L .80573 .08632 L closepath p .884 .492 .072 r
F P
0 g
s
.80573 .08632 m .80116 .07656 L .80786 .08315 L closepath p .955 .786 .906 r
F P
0 g
s
.59286 .07829 m .59543 .07612 L .59262 .07658 L closepath p .242 .256 .742 r
F P
0 g
s
.59519 .07438 m .59262 .07658 L .59543 .07612 L closepath p .242 .256 .742 r
F P
0 g
s
.09808 .09375 m .09939 .09221 L .10169 .07669 L closepath p .087 .087 .181 r
F P
0 g
s
.10301 .07512 m .10169 .07669 L .09939 .09221 L closepath p .088 .088 .181 r
F P
0 g
s
.10169 .07669 m .09808 .09375 L .09014 .12843 L closepath p .955 .786 .906 r
F P
0 g
s
.09014 .12843 m .08738 .12827 L .10169 .07669 L closepath p .955 .786 .906 r
F P
0 g
s
.80178 .02803 m .79727 .04038 L .8073 .01666 L closepath p .717 .837 .385 r
F P
0 g
s
.79727 .04038 m .80178 .02803 L .79564 .05448 L closepath p .048 .048 .095 r
F P
0 g
s
.57713 .09421 m .58456 .10772 L .57149 .09166 L closepath p .955 .786 .906 r
F P
0 g
s
.577 .09421 m .579 .09074 L .56817 .07771 L closepath p .393 .492 .939 r
F P
0 g
s
.57035 .07458 m .56817 .07771 L .579 .09074 L closepath p .393 .492 .939 r
F P
0 g
s
.42562 .02839 m .42673 .04073 L .42495 .02396 L closepath p .012 .012 .149 r
F P
0 g
s
.5921 .05333 m .5935 .04117 L .59061 .04095 L closepath p .045 .045 .246 r
F P
0 g
s
.59198 .02861 m .59061 .04095 L .5935 .04117 L closepath p .045 .045 .246 r
F P
0 g
s
.70576 .02865 m .70269 .04098 L .70475 .02327 L closepath p .08 .08 .416 r
F P
0 g
s
.59419 .08961 m .59676 .08745 L .59286 .07829 L closepath p .233 .24 .728 r
F P
0 g
s
.59543 .07612 m .59286 .07829 L .59676 .08745 L closepath p .233 .24 .728 r
F P
0 g
s
.59044 .08715 m .59286 .07829 L .59419 .08961 L closepath p .955 .786 .906 r
F P
0 g
s
.58456 .10772 m .59286 .07829 L .58317 .11381 L closepath p .955 .786 .906 r
F P
0 g
s
.47135 .08265 m .47156 .08497 L .46827 .07843 L closepath p .955 .786 .906 r
F P
0 g
s
.46971 .08256 m .46827 .07843 L .47156 .08497 L closepath p .955 .786 .906 r
F P
0 g
s
.86164 .02908 m .85624 .04141 L .85649 .02299 L closepath p .161 .436 .936 r
F P
0 g
s
.85624 .04141 m .86164 .02908 L .86037 .04622 L closepath p .166 .444 .941 r
F P
0 g
s
.49146 .04379 m .49133 .0315 L .49287 .04145 L closepath p .936 1 .693 r
F P
0 g
s
.49277 .02912 m .49287 .04145 L .49133 .0315 L closepath p .936 1 .693 r
F P
0 g
s
.37923 .02921 m .38133 .02707 L .38064 .03863 L p .167 .167 .683 r
F P
0 g
s
.37923 .02921 m .38103 .04154 L .37824 .02264 L p .001 .001 .107 r
F P
0 g
s
.38064 .03863 m .38062 .03896 L .37923 .02921 L p .167 .167 .683 r
F P
0 g
s
.37824 .02264 m .37823 .02258 L .37923 .02921 L p .001 .001 .107 r
F P
0 g
s
.0675 .07879 m .06415 .07754 L .07176 .09362 L closepath p 1 .841 .541 r
F P
0 g
s
.76964 .02934 m .76561 .04167 L .7688 .02473 L closepath p .092 .092 .452 r
F P
0 g
s
.48468 .02984 m .48491 .04216 L .47877 .02306 L closepath p .115 .31 .839 r
F P
0 g
s
.48491 .04216 m .48468 .02984 L .48708 .044 L closepath p .13 .468 .947 r
F P
0 g
s
.52463 .07687 m .52683 .07932 L .52701 .06434 L closepath p 1 1 .744 r
F P
0 g
s
.52683 .07932 m .52463 .07687 L .5249 .08878 L closepath p 1 1 .754 r
F P
0 g
s
.20184 .07939 m .20357 .08258 L .20667 .09139 L closepath p .955 .786 .906 r
F P
0 g
s
.12244 .0788 m .1252 .07946 L .11403 .06083 L closepath p 1 .715 .362 r
F P
0 g
s
.1252 .07946 m .12244 .0788 L .13605 .10138 L closepath p 1 .698 .343 r
F P
0 g
s
.1252 .07946 m .1502 .12162 L .13605 .10138 L closepath p .955 .786 .906 r
F P
0 g
s
.81227 .05256 m .81701 .04039 L .8133 .04248 L closepath p .054 .054 .127 r
F P
0 g
s
.81806 .03016 m .8133 .04248 L .81701 .04039 L closepath p .054 .054 .127 r
F P
0 g
s
.81102 .07966 m .81576 .08181 L .81204 .09587 L closepath p .955 .786 .906 r
F P
0 g
s
.20594 .08707 m .20357 .08258 L .20257 .07982 L closepath p .955 .786 .906 r
F P
0 g
s
.63512 .07982 m .63869 .08707 L .63789 .09211 L closepath p .955 .786 .906 r
F P
0 g
s
.72101 .05117 m .72437 .03898 L .71967 .04268 L closepath p .079 .079 .393 r
F P
0 g
s
.723 .03037 m .71967 .04268 L .72437 .03898 L closepath p .079 .079 .393 r
F P
0 g
s
.80999 .07999 m .81204 .09587 L .80573 .08632 L closepath p .955 .786 .906 r
F P
0 g
s
.2829 .03053 m .28614 .04284 L .28399 .03897 L closepath p .796 .577 .026 r
F P
0 g
s
.38911 .08018 m .38653 .07888 L .39051 .08869 L closepath p 1 .876 .523 r
F P
0 g
s
.62549 .05116 m .6274 .03897 L .62109 .04314 L closepath p .11 .137 .709 r
F P
0 g
s
.62293 .03084 m .62109 .04314 L .6274 .03897 L closepath p .11 .137 .709 r
F P
0 g
s
.89793 .07938 m .89905 .08036 L .89414 .07178 L closepath p .817 .419 .053 r
F P
0 g
s
.89905 .08036 m .89793 .07938 L .90295 .08807 L closepath p .814 .412 .052 r
F P
0 g
s
.88923 .05119 m .89514 .039 L .88474 .04322 L closepath p .147 .281 .821 r
F P
0 g
s
.89058 .03092 m .88474 .04322 L .89514 .039 L closepath p .147 .281 .821 r
F P
0 g
s
.78385 .05116 m .78816 .03897 L .78255 .04325 L closepath p .093 .093 .443 r
F P
0 g
s
.78684 .03095 m .78255 .04325 L .78816 .03897 L closepath p .093 .093 .443 r
F P
0 g
s
.07097 .08046 m .07245 .08543 L .0766 .11196 L closepath p .955 .786 .906 r
F P
0 g
s
.0766 .11196 m .0748 .10468 L .07097 .08046 L closepath p .955 .786 .906 r
F P
0 g
s
.09299 .08056 m .08509 .09798 L .07995 .10855 L closepath p .955 .786 .906 r
F P
0 g
s
.15948 .05778 m .1543 .04568 L .15829 .04364 L closepath p .924 .709 .174 r
F P
0 g
s
.1531 .03134 m .15829 .04364 L .1543 .04568 L closepath p .924 .709 .174 r
F P
0 g
s
.4887 .04465 m .48853 .03236 L .49146 .04379 L closepath p .766 1 1 r
F P
0 g
s
.49133 .0315 m .49146 .04379 L .48853 .03236 L closepath p .766 1 1 r
F P
0 g
s
.8666 .04483 m .87216 .03255 L .86636 .04383 L closepath p .117 .117 .568 r
F P
0 g
s
.87192 .03154 m .86636 .04383 L .87216 .03255 L closepath p .117 .117 .568 r
F P
0 g
s
.06557 .0543 m .05897 .04215 L .06303 .04386 L closepath p .878 .585 .055 r
F P
0 g
s
.05639 .03156 m .06303 .04386 L .05897 .04215 L closepath p .878 .585 .055 r
F P
0 g
s
.48688 .03171 m .48708 .044 L .48468 .02984 L closepath p .13 .468 .947 r
F P
0 g
s
.48708 .044 m .48688 .03171 L .4887 .04465 L closepath p .36 .745 1 r
F P
0 g
s
.88223 .03174 m .87652 .04403 L .8767 .02072 L closepath p .142 .232 .782 r
F P
0 g
s
.87652 .04403 m .88223 .03174 L .88324 .05731 L closepath p .143 .236 .786 r
F P
0 g
s
.17534 .03174 m .18019 .04404 L .17032 .02073 L closepath p .011 .011 .332 r
F P
0 g
s
.18019 .04404 m .17534 .03174 L .1863 .05731 L closepath p .012 .012 .338 r
F P
0 g
s
.61453 .03175 m .61282 .04404 L .60917 .02074 L closepath p .102 .102 .657 r
F P
0 g
s
.61282 .04404 m .61453 .03175 L .61933 .05732 L closepath p .103 .103 .66 r
F P
0 g
s
.54488 .05759 m .54556 .0455 L .54654 .04416 L closepath p .751 .684 .134 r
F P
0 g
s
.54725 .03187 m .54654 .04416 L .54556 .0455 L closepath p .751 .684 .134 r
F P
0 g
s
.54654 .04416 m .54725 .03187 L .55202 .05624 L closepath p .087 .087 .591 r
F P
0 g
s
.82065 .09118 m .8235 .08818 L .81576 .08181 L closepath p .461 .53 .951 r
F P
0 g
s
.8187 .079 m .81576 .08181 L .8235 .08818 L closepath p .46 .53 .951 r
F P
0 g
s
.82065 .09118 m .81204 .09587 L .81576 .08181 L closepath p .955 .786 .906 r
F P
0 g
s
.48853 .03236 m .4887 .04465 L .48688 .03171 L closepath p .36 .745 1 r
F P
0 g
s
.40176 .08914 m .40078 .08401 L .40096 .0819 L closepath p .955 .786 .906 r
F P
0 g
s
.40096 .0819 m .40195 .08875 L .40176 .08914 L closepath p .955 .786 .906 r
F P
0 g
s
.50989 .05211 m .51004 .03993 L .50793 .04482 L closepath p .052 .052 .369 r
F P
0 g
s
.50805 .03254 m .50793 .04482 L .51004 .03993 L closepath p .052 .052 .369 r
F P
0 g
s
.86556 .04714 m .87111 .03489 L .8666 .04483 L closepath p .001 .001 .001 r
F P
0 g
s
.87216 .03255 m .8666 .04483 L .87111 .03489 L closepath p .001 .001 .001 r
F P
0 g
s
.47156 .08497 m .47176 .08729 L .46971 .08256 L closepath p .955 .786 .906 r
F P
0 g
s
.47087 .08552 m .46971 .08256 L .47176 .08729 L closepath p .955 .786 .906 r
F P
0 g
s
.20765 .09201 m .20667 .09139 L .20357 .08258 L closepath p .955 .786 .906 r
F P
0 g
s
.20357 .08258 m .20594 .08707 L .20765 .09201 L closepath p .955 .786 .906 r
F P
0 g
s
.47443 .08688 m .474 .08763 L .47135 .08265 L closepath p .955 .786 .906 r
F P
0 g
s
.47156 .08497 m .47135 .08265 L .474 .08763 L closepath p .955 .786 .906 r
F P
0 g
s
.23335 .06141 m .2293 .04937 L .23706 .0456 L closepath p 1 .952 .508 r
F P
0 g
s
.23307 .03333 m .23706 .0456 L .2293 .04937 L closepath p 1 .952 .508 r
F P
0 g
s
.66644 .06142 m .66896 .04938 L .66971 .04565 L closepath p 0 0 0 r
F P
0 g
s
.67229 .03338 m .66971 .04565 L .66896 .04938 L closepath p 0 0 0 r
F P
0 g
s
.93058 .06143 m .93712 .04938 L .93355 .04568 L closepath p .064 .064 .108 r
F P
0 g
s
.94013 .03341 m .93355 .04568 L .93712 .04938 L closepath p .064 .064 .108 r
F P
0 g
s
.86584 .03396 m .86037 .04622 L .86164 .02908 L closepath p .166 .444 .941 r
F P
0 g
s
.86037 .04622 m .86584 .03396 L .86338 .04809 L closepath p .344 .669 1 r
F P
0 g
s
.42648 .0342 m .42758 .04646 L .42562 .02839 L closepath p .011 .011 .145 r
F P
0 g
s
.34984 .09704 m .34877 .09326 L .34852 .08378 L closepath p .955 .786 .906 r
F P
0 g
s
.34852 .08378 m .34981 .09302 L .34984 .09704 L closepath p .955 .786 .906 r
F P
0 g
s
.40078 .08401 m .40274 .09427 L .40175 .09429 L closepath p .955 .786 .906 r
F P
0 g
s
.86338 .04809 m .86889 .03586 L .86556 .04714 L closepath p .743 1 1 r
F P
0 g
s
.87111 .03489 m .86556 .04714 L .86889 .03586 L closepath p .743 1 1 r
F P
0 g
s
.53609 .0958 m .53851 .09423 L .53805 .08446 L closepath p .158 .158 .482 r
F P
0 g
s
.54047 .08291 m .53805 .08446 L .53851 .09423 L closepath p .158 .158 .482 r
F P
0 g
s
.53242 .1149 m .53213 .11527 L .53805 .08446 L closepath p .955 .786 .906 r
F P
0 g
s
.53805 .08446 m .53609 .0958 L .53242 .1149 L closepath p .955 .786 .906 r
F P
0 g
s
.47055 .04804 m .47011 .0358 L .46825 .0473 L closepath p .405 .792 1 r
F P
0 g
s
.46777 .03506 m .46825 .0473 L .47011 .0358 L closepath p .405 .792 1 r
F P
0 g
s
.46825 .0473 m .46777 .03506 L .46532 .04907 L closepath p .866 1 1 r
F P
0 g
s
.77068 .03521 m .76663 .04746 L .76964 .02934 L closepath p .091 .091 .446 r
F P
0 g
s
.34299 .03527 m .34534 .04751 L .34444 .04417 L closepath p .719 .495 .006 r
F P
0 g
s
.70696 .03538 m .70387 .04763 L .70576 .02865 L closepath p .079 .079 .404 r
F P
0 g
s
.474 .08763 m .47237 .08788 L .47156 .08497 L closepath p .955 .786 .906 r
F P
0 g
s
.47176 .08729 m .47156 .08497 L .47237 .08788 L closepath p .955 .786 .906 r
F P
0 g
s
.4722 .04999 m .47178 .03779 L .47055 .04804 L closepath p .11 .284 .819 r
F P
0 g
s
.47011 .0358 m .47055 .04804 L .47178 .03779 L closepath p .11 .284 .819 r
F P
0 g
s
.86889 .03586 m .86338 .04809 L .86584 .03396 L closepath p .344 .669 1 r
F P
0 g
s
.07393 .0903 m .0766 .11196 L .07245 .08543 L closepath p .955 .786 .906 r
F P
0 g
s
.47087 .08552 m .4691 .08513 L .47176 .08729 L closepath p 1 .597 .157 r
F P
0 g
s
.16203 .09871 m .16353 .09655 L .16093 .08596 L closepath p .137 .137 .549 r
F P
0 g
s
.16243 .08377 m .16093 .08596 L .16353 .09655 L closepath p .137 .137 .549 r
F P
0 g
s
.15787 .09613 m .16093 .08596 L .16203 .09871 L closepath p .955 .786 .906 r
F P
0 g
s
.4648 .03685 m .46532 .04907 L .46777 .03506 L closepath p .866 1 1 r
F P
0 g
s
.46532 .04907 m .4648 .03685 L .4643 .05286 L closepath p .887 .872 .388 r
F P
0 g
s
.80444 .08546 m .80573 .08632 L .79972 .07533 L closepath p .889 .496 .074 r
F P
0 g
s
.80573 .08632 m .80444 .08546 L .81204 .09587 L closepath p .785 .335 .056 r
F P
0 g
s
.78829 .09688 m .78457 .09347 L .78648 .08667 L closepath p .955 .786 .906 r
F P
0 g
s
.78829 .09688 m .79133 .0946 L .78648 .08667 L closepath p .343 .346 .802 r
F P
0 g
s
.78952 .0844 m .78648 .08667 L .79133 .0946 L closepath p .343 .346 .802 r
F P
0 g
s
.57746 .03731 m .5763 .04952 L .57507 .02072 L closepath p .045 .045 .264 r
F P
0 g
s
.474 .08763 m .4754 .0873 L .47443 .08688 L closepath p .093 .093 .093 r
F P
0 g
s
.4767 .08463 m .47443 .08688 L .4754 .0873 L closepath p .094 .094 .094 r
F P
0 g
s
.64197 .09785 m .63789 .09211 L .63869 .08707 L closepath p .955 .786 .906 r
F P
0 g
s
.63869 .08707 m .64151 .09264 L .64197 .09785 L closepath p .955 .786 .906 r
F P
0 g
s
.59419 .08961 m .59515 .09779 L .59044 .08715 L closepath p .955 .786 .906 r
F P
0 g
s
.58802 .09603 m .59044 .08715 L .59515 .09779 L closepath p .955 .786 .906 r
F P
0 g
s
.47326 .05847 m .47286 .04638 L .4722 .04999 L closepath p .017 .017 .151 r
F P
0 g
s
.47178 .03779 m .4722 .04999 L .47286 .04638 L closepath p .017 .017 .151 r
F P
0 g
s
.08218 .08477 m .08547 .0872 L .08533 .07049 L closepath p 1 1 .839 r
F P
0 g
s
.08547 .0872 m .08218 .08477 L .08294 .09799 L closepath p 1 1 .846 r
F P
0 g
s
.47068 .08742 m .47176 .08729 L .4691 .08513 L closepath p .871 .411 .087 r
F P
0 g
s
.47176 .08729 m .47068 .08742 L .47237 .08788 L closepath p .755 .229 .077 r
F P
0 g
s
.47237 .08788 m .4723 .08825 L .474 .08763 L closepath p .034 .034 .034 r
F P
0 g
s
.4754 .0873 m .474 .08763 L .4723 .08825 L closepath p .009 .009 .009 r
F P
0 g
s
.4723 .08825 m .47237 .08788 L .47068 .08742 L closepath p .448 .035 .035 r
F P
0 g
s
.90191 .08722 m .90295 .08807 L .89793 .07938 L closepath p .809 .408 .051 r
F P
0 g
s
.90295 .08807 m .90191 .08722 L .90591 .0937 L closepath p .799 .392 .05 r
F P
0 g
s
.38062 .03896 m .3824 .05115 L .37923 .02926 L p .703 .492 .001 r
F P
0 g
s
.3824 .05115 m .38062 .03896 L .38414 .06361 L closepath p .706 .497 .001 r
F P
0 g
s
.38062 .03896 m .38284 .03766 L .38239 .05134 L p .167 .167 .68 r
F P
0 g
s
.38238 .0516 m .38414 .06361 L .38231 .05081 L p .706 .497 .001 r
F P
0 g
s
.38231 .05081 m .38062 .03896 L .38238 .0516 L p .706 .497 .001 r
F P
0 g
s
.38414 .06361 m .38238 .0516 L .34793 .06361 L closepath p .602 .969 1 r
F P
0 g
s
.38238 .0516 m .38482 .05187 L .3824 .05186 L p .099 .026 .026 r
F P
0 g
s
.34562 .0516 m .34444 .04417 L .3457 .05071 L p .722 .499 .006 r
F P
0 g
s
.3457 .05071 m .34677 .05628 L .34562 .0516 L p .722 .499 .006 r
F P
0 g
s
.34793 .06361 m .34562 .0516 L .34677 .05628 L closepath p .722 .499 .006 r
F P
0 g
s
.34562 .0516 m .34793 .06361 L .38238 .0516 L closepath p .602 .969 1 r
F P
0 g
s
.34562 .0516 m .38482 .05187 L .3456 .05187 L p .099 .026 .026 r
F P
0 g
s
.3456 .05187 m .34317 .05187 L .34562 .0516 L p .099 .026 .026 r
F P
0 g
s
.3824 .05186 m .34562 .0516 L .38238 .0516 L p .099 .026 .026 r
F P
0 g
s
.38239 .05134 m .38238 .0516 L .38062 .03896 L p .167 .167 .68 r
F P
0 g
s
.37923 .02926 m .37923 .02921 L .38062 .03896 L p .703 .492 .001 r
F P
0 g
s
.29379 .10059 m .29065 .08911 L .28722 .05116 L closepath p .794 .573 .025 r
F P
0 g
s
.28399 .03897 m .28722 .05116 L .29065 .08911 L closepath p .794 .573 .025 r
F P
0 g
s
.78702 .06991 m .79139 .05799 L .78385 .05116 L closepath p .093 .093 .449 r
F P
0 g
s
.78816 .03897 m .78385 .05116 L .79139 .05799 L closepath p .093 .093 .449 r
F P
0 g
s
.72894 .1006 m .73243 .08912 L .72101 .05117 L closepath p .08 .08 .396 r
F P
0 g
s
.72437 .03898 m .72101 .05117 L .73243 .08912 L closepath p .08 .08 .396 r
F P
0 g
s
.89028 .05266 m .89621 .04049 L .88923 .05119 L closepath p .154 .377 .894 r
F P
0 g
s
.89514 .039 m .88923 .05119 L .89621 .04049 L closepath p .154 .377 .894 r
F P
0 g
s
.38704 .08014 m .38653 .07888 L .38791 .08714 L p 1 .873 .517 r
F P
0 g
s
.38653 .07888 m .3848 .06709 L .38791 .08714 L closepath p .004 .004 .131 r
F P
0 g
s
.3862 .07547 m .38791 .08714 L .38488 .06757 L p .004 .004 .131 r
F P
0 g
s
.38488 .06757 m .3848 .06709 L .3862 .07547 L p .004 .004 .131 r
F P
0 g
s
.38791 .08714 m .39051 .08869 L .38704 .08014 L p 1 .873 .517 r
F P
0 g
s
.3862 .07547 m .38828 .07311 L .3875 .08321 L p .166 .166 .669 r
F P
0 g
s
.38791 .08714 m .3862 .07547 L .38899 .09586 L p .716 .514 .004 r
F P
0 g
s
.38738 .08487 m .38907 .09641 L .38635 .07652 L p .716 .514 .004 r
F P
0 g
s
.38635 .07652 m .3862 .07547 L .38738 .08487 L p .716 .514 .004 r
F P
0 g
s
.38899 .09586 m .38907 .09641 L .38791 .08714 L p .716 .514 .004 r
F P
0 g
s
.39051 .08869 m .38791 .08714 L .39169 .09811 L closepath p 1 .899 .551 r
F P
0 g
s
.38907 .09641 m .39169 .09811 L .38791 .08714 L closepath p 1 .898 .551 r
F P
0 g
s
.38907 .09641 m .38738 .08487 L .38941 .10401 L closepath p .775 .607 .035 r
F P
0 g
s
.38738 .08487 m .38948 .08272 L .38808 .09059 L p .159 .159 .603 r
F P
0 g
s
.38773 .09258 m .38941 .10401 L .38758 .08683 L p .775 .607 .035 r
F P
0 g
s
.38758 .08683 m .38738 .08487 L .38773 .09258 L p .775 .607 .035 r
F P
0 g
s
.38941 .10401 m .38773 .09258 L .3891 .10594 L closepath p .891 .82 .309 r
F P
0 g
s
.38741 .09454 m .3891 .10594 L .38787 .09398 L p .891 .82 .309 r
F P
0 g
s
.38787 .09398 m .38773 .09258 L .38741 .09454 L p .891 .82 .309 r
F P
0 g
s
.3891 .10594 m .38741 .09454 L .3878 .10683 L closepath p .883 1 1 r
F P
0 g
s
.38609 .09544 m .3878 .10683 L .38743 .0951 L p .883 1 1 r
F P
0 g
s
.38743 .0951 m .38741 .09454 L .38609 .09544 L p .883 1 1 r
F P
0 g
s
.3878 .10683 m .38609 .09544 L .38588 .10717 L closepath p .689 1 1 r
F P
0 g
s
.38414 .09579 m .38588 .10717 L .38609 .09544 L closepath p .689 1 1 r
F P
0 g
s
.38609 .09544 m .38665 .09568 L .3848 .09576 L p .031 .031 .031 r
F P
0 g
s
.3848 .09576 m .38414 .09579 L .38609 .09544 L p .031 .031 .031 r
F P
0 g
s
.38741 .09454 m .38903 .09383 L .38714 .09486 L p .03 .03 .03 r
F P
0 g
s
.38714 .09486 m .38609 .09544 L .38741 .09454 L p .03 .03 .03 r
F P
0 g
s
.38773 .09258 m .38984 .09064 L .38776 .09397 L p .138 .138 .405 r
F P
0 g
s
.38776 .09397 m .38741 .09454 L .38773 .09258 L p .138 .138 .405 r
F P
0 g
s
.38808 .09059 m .38773 .09258 L .38738 .08487 L p .159 .159 .603 r
F P
0 g
s
.3875 .08321 m .38738 .08487 L .3862 .07547 L p .166 .166 .669 r
F P
0 g
s
.40371 .11489 m .39748 .11195 L .39051 .08869 L closepath p .955 .786 .906 r
F P
0 g
s
.39169 .09811 m .39051 .08869 L .39748 .11195 L closepath p .955 .786 .906 r
F P
0 g
s
.40274 .09427 m .40483 .09201 L .40195 .08875 L closepath p .171 .178 .695 r
F P
0 g
s
.40274 .09427 m .40176 .08914 L .40195 .08875 L closepath p .955 .786 .906 r
F P
0 g
s
.52297 .08597 m .5249 .08878 L .52463 .07687 L closepath p 1 1 .716 r
F P
0 g
s
.5249 .08878 m .52297 .08597 L .52345 .0951 L closepath p 1 1 .752 r
F P
0 g
s
.51237 .06059 m .51256 .04854 L .50989 .05211 L closepath p .056 .056 .396 r
F P
0 g
s
.51004 .03993 m .50989 .05211 L .51256 .04854 L closepath p .056 .056 .396 r
F P
0 g
s
.58267 .08942 m .58048 .08741 L .58456 .10772 L closepath p 1 .884 .493 r
F P
0 g
s
.95326 .0905 m .95521 .08652 L .94741 .08942 L closepath p .775 .94 1 r
F P
0 g
s
.94845 .08522 m .94741 .08942 L .95521 .08652 L closepath p .776 .937 1 r
F P
0 g
s
.94741 .08942 m .94845 .08522 L .94167 .09052 L closepath p .9 1 1 r
F P
0 g
s
.94239 .10201 m .9372 .1 L .94741 .08942 L closepath p .955 .786 .906 r
F P
0 g
s
.94167 .09052 m .94741 .08942 L .9372 .1 L closepath p .955 .786 .906 r
F P
0 g
s
.94741 .08942 m .95326 .0905 L .93737 .11462 L closepath p .955 .786 .906 r
F P
0 g
s
.59515 .09779 m .59773 .09564 L .59419 .08961 L closepath p .233 .24 .728 r
F P
0 g
s
.59676 .08745 m .59419 .08961 L .59773 .09564 L closepath p .233 .24 .728 r
F P
0 g
s
.81311 .06125 m .81786 .0492 L .81227 .05256 L closepath p .088 .088 .383 r
F P
0 g
s
.81701 .04039 m .81227 .05256 L .81786 .0492 L closepath p .088 .088 .383 r
F P
0 g
s
.8909 .05277 m .89684 .04061 L .89028 .05266 L closepath p .53 .888 1 r
F P
0 g
s
.89621 .04049 m .89028 .05266 L .89684 .04061 L closepath p .53 .888 1 r
F P
0 g
s
.87149 .09084 m .87393 .08677 L .87032 .09003 L closepath p .664 .78 1 r
F P
0 g
s
.87118 .08584 m .87032 .09003 L .87393 .08677 L closepath p .72 .884 1 r
F P
0 g
s
.87032 .09003 m .87118 .08584 L .86985 .09049 L closepath p 1 1 1 r
F P
0 g
s
.87032 .09003 m .86939 .09095 L .86973 .09026 L closepath p .955 .786 .906 r
F P
0 g
s
.87032 .09003 m .87266 .09165 L .8737 .10126 L closepath p .955 .786 .906 r
F P
0 g
s
.87032 .09003 m .87116 .09284 L .86939 .09095 L closepath p .955 .786 .906 r
F P
0 g
s
.46376 .04069 m .4643 .05286 L .4648 .03685 L closepath p .887 .872 .388 r
F P
0 g
s
.4643 .05286 m .46376 .04069 L .46397 .05861 L closepath p .78 .67 .113 r
F P
0 g
s
.07393 .0903 m .07541 .09507 L .0766 .11196 L closepath p .955 .786 .906 r
F P
0 g
s
.07688 .09954 m .07724 .09601 L .07393 .0903 L closepath p .121 .121 .631 r
F P
0 g
s
.07487 .08735 m .07393 .0903 L .07724 .09601 L closepath p .138 .138 .659 r
F P
0 g
s
.86916 .08671 m .86985 .09049 L .87118 .08584 L closepath p .975 1 1 r
F P
0 g
s
.86985 .09049 m .86916 .08671 L .86939 .09095 L closepath p 1 1 1 r
F P
0 g
s
.95829 .09376 m .96105 .09024 L .95326 .0905 L closepath p .652 .786 1 r
F P
0 g
s
.95521 .08652 m .95326 .0905 L .96105 .09024 L closepath p .65 .788 1 r
F P
0 g
s
.95024 .10944 m .93737 .11462 L .95326 .0905 L closepath p .955 .786 .906 r
F P
0 g
s
.95326 .0905 m .95829 .09376 L .95175 .09996 L closepath p .955 .786 .906 r
F P
0 g
s
.59352 .06298 m .59494 .05095 L .5921 .05333 L closepath p .049 .049 .281 r
F P
0 g
s
.5935 .04117 m .5921 .05333 L .59494 .05095 L closepath p .049 .049 .281 r
F P
0 g
s
.94182 .08655 m .94167 .09052 L .94845 .08522 L closepath p .901 1 1 r
F P
0 g
s
.94167 .09052 m .94182 .08655 L .93681 .09384 L closepath p 1 1 1 r
F P
0 g
s
.9372 .1 m .93354 .09865 L .94167 .09052 L closepath p .955 .786 .906 r
F P
0 g
s
.93681 .09384 m .94167 .09052 L .93354 .09865 L closepath p .955 .786 .906 r
F P
0 g
s
.42753 .0414 m .42862 .05356 L .42648 .0342 L closepath p .011 .011 .142 r
F P
0 g
s
.87266 .09165 m .87589 .08922 L .87149 .09084 L closepath p .586 .755 1 r
F P
0 g
s
.87393 .08677 m .87149 .09084 L .87589 .08922 L closepath p .562 .649 1 r
F P
0 g
s
.86801 .08852 m .86939 .09095 L .86916 .08671 L closepath p 1 1 .919 r
F P
0 g
s
.86939 .09095 m .86801 .08852 L .86813 .09663 L closepath p 1 1 .664 r
F P
0 g
s
.86813 .09663 m .86939 .09095 L .87116 .09284 L closepath p .955 .786 .906 r
F P
0 g
s
.82729 .10074 m .82983 .09739 L .82065 .09118 L closepath p .499 .594 .998 r
F P
0 g
s
.8235 .08818 m .82065 .09118 L .82983 .09739 L closepath p .503 .594 .998 r
F P
0 g
s
.82065 .09118 m .82729 .10074 L .81204 .09587 L closepath p .955 .786 .906 r
F P
0 g
s
.20667 .09139 m .20862 .09264 L .21151 .1034 L closepath p .955 .786 .906 r
F P
0 g
s
.0692 .06762 m .06265 .05566 L .06557 .0543 L closepath p .856 .551 .049 r
F P
0 g
s
.05897 .04215 m .06557 .0543 L .06265 .05566 L closepath p .856 .551 .049 r
F P
0 g
s
.87322 .09655 m .8737 .10126 L .87266 .09165 L closepath p .955 .786 .906 r
F P
0 g
s
.8737 .10126 m .87699 .09922 L .87266 .09165 L closepath p .353 .33 .778 r
F P
0 g
s
.87589 .08922 m .87266 .09165 L .87699 .09922 L closepath p .352 .33 .778 r
F P
0 g
s
.56961 .09082 m .57149 .09166 L .55979 .07185 L closepath p .976 .565 .113 r
F P
0 g
s
.57149 .09166 m .56961 .09082 L .58317 .11381 L closepath p .972 .558 .109 r
F P
0 g
s
.58317 .11381 m .57149 .09166 L .58456 .10772 L closepath p .955 .786 .906 r
F P
0 g
s
.80013 .04233 m .79564 .05448 L .80178 .02803 L closepath p .048 .048 .095 r
F P
0 g
s
.79564 .05448 m .80013 .04233 L .79678 .06498 L closepath p .087 .087 .386 r
F P
0 g
s
.77192 .04235 m .76785 .05449 L .77068 .03521 L closepath p .091 .091 .442 r
F P
0 g
s
.95225 .09286 m .95424 .0889 L .94749 .09195 L closepath p .773 .939 1 r
F P
0 g
s
.94853 .08776 m .94749 .09195 L .95424 .0889 L closepath p .773 .935 1 r
F P
0 g
s
.94749 .09195 m .94853 .08776 L .94277 .09286 L closepath p .9 1 1 r
F P
0 g
s
.94348 .10237 m .93929 .10069 L .94749 .09195 L closepath p .955 .786 .906 r
F P
0 g
s
.94277 .09286 m .94749 .09195 L .93929 .10069 L closepath p .955 .786 .906 r
F P
0 g
s
.94749 .09195 m .95225 .09286 L .94348 .10237 L closepath p .955 .786 .906 r
F P
0 g
s
.20203 .09115 m .20493 .09211 L .19359 .07305 L p 1 .716 .363 r
F P
0 g
s
.20493 .09211 m .20203 .09115 L .20952 .10197 L closepath p 1 .716 .363 r
F P
0 g
s
.20666 .1011 m .20952 .10197 L .20203 .09115 L closepath p 1 .716 .363 r
F P
0 g
s
.20203 .09115 m .19749 .07953 L .20666 .1011 L p .013 .013 .344 r
F P
0 g
s
.20666 .1011 m .20666 .1011 L .20203 .09115 L p .013 .013 .344 r
F P
0 g
s
.19359 .07305 m .19358 .07301 L .20203 .09115 L p 1 .716 .363 r
F P
0 g
s
.90056 .09114 m .90168 .09211 L .89126 .07301 L closepath p .811 .408 .052 r
F P
0 g
s
.90168 .09211 m .90056 .09114 L .9068 .10201 L closepath p .809 .404 .052 r
F P
0 g
s
.6361 .09115 m .63789 .09211 L .62711 .07305 L p .963 .558 .103 r
F P
0 g
s
.63789 .09211 m .6361 .09115 L .64279 .10197 L closepath p .963 .558 .103 r
F P
0 g
s
.65423 .11512 m .65333 .11581 L .63789 .09211 L closepath p .955 .786 .906 r
F P
0 g
s
.64279 .10197 m .63789 .09211 L .65333 .11581 L closepath p .955 .786 .906 r
F P
0 g
s
.64104 .1011 m .64279 .10197 L .6361 .09115 L closepath p .963 .558 .103 r
F P
0 g
s
.6361 .09115 m .63817 .07953 L .64104 .1011 L p .103 .103 .665 r
F P
0 g
s
.64104 .1011 m .64104 .1011 L .6361 .09115 L p .103 .103 .665 r
F P
0 g
s
.62711 .07305 m .62709 .07302 L .6361 .09115 L p .963 .558 .103 r
F P
0 g
s
.91861 .11512 m .91776 .11581 L .90168 .09211 L closepath p .955 .786 .906 r
F P
0 g
s
.9068 .10201 m .90168 .09211 L .91776 .11581 L closepath p .955 .786 .906 r
F P
0 g
s
.57035 .07458 m .57142 .06273 L .56001 .05522 L closepath p .097 .097 .664 r
F P
0 g
s
.56092 .04309 m .56001 .05522 L .57142 .06273 L closepath p .097 .097 .664 r
F P
0 g
s
.22333 .10395 m .22496 .10243 L .22545 .09257 L closepath p .114 .114 .298 r
F P
0 g
s
.22708 .09106 m .22545 .09257 L .22496 .10243 L closepath p .114 .114 .298 r
F P
0 g
s
.22545 .09257 m .22439 .09826 L .21635 .11543 L closepath p .955 .786 .906 r
F P
0 g
s
.92219 .09345 m .92558 .09192 L .9223 .09257 L closepath p .298 .201 .642 r
F P
0 g
s
.92569 .09106 m .9223 .09257 L .92558 .09192 L closepath p .298 .201 .642 r
F P
0 g
s
.65821 .09345 m .66094 .09192 L .65835 .09257 L closepath p .172 .172 .549 r
F P
0 g
s
.66107 .09106 m .65835 .09257 L .66094 .09192 L closepath p .172 .172 .549 r
F P
0 g
s
.65835 .09257 m .65821 .09345 L .65629 .10383 L closepath p .955 .786 .906 r
F P
0 g
s
.9223 .09257 m .92218 .09345 L .92046 .10383 L closepath p .955 .786 .906 r
F P
0 g
s
.64606 .1036 m .64197 .09785 L .64151 .09264 L closepath p .955 .786 .906 r
F P
0 g
s
.64151 .09264 m .6436 .09653 L .64606 .1036 L closepath p .955 .786 .906 r
F P
0 g
s
.70835 .04349 m .70523 .05562 L .70696 .03538 L closepath p .078 .078 .396 r
F P
0 g
s
.87116 .09284 m .87201 .09564 L .86813 .09663 L closepath p .955 .786 .906 r
F P
0 g
s
.95225 .09286 m .95543 .09501 L p .95807 .09332 L .644 .776 1 r
F P
0 g
s
.95543 .09501 m .95625 .09556 L .95795 .09347 L p .644 .776 1 r
F P
0 g
s
.95795 .09347 m .95807 .09332 L p .95543 .09501 L .644 .776 1 r
F P
0 g
s
.95225 .09286 m .95625 .09556 L .95123 .10096 L closepath p .955 .786 .906 r
F P
0 g
s
.9618 .09855 m .96507 .09581 L .95829 .09376 L closepath p .533 .62 1 r
F P
0 g
s
.96105 .09024 m .95829 .09376 L .96507 .09581 L closepath p .535 .619 1 r
F P
0 g
s
.9618 .09855 m .95175 .09996 L .95829 .09376 L closepath p .955 .786 .906 r
F P
0 g
s
.95175 .09996 m .9618 .09855 L .95024 .10944 L closepath p .955 .786 .906 r
F P
0 g
s
.95899 .09954 m .96227 .09684 L .95753 .09584 L p .524 .606 1 r
F P
0 g
s
.95753 .09584 m .95625 .09556 L .95701 .09668 L p .524 .606 1 r
F P
0 g
s
.95701 .09668 m .95899 .09954 L p .95753 .09584 L .524 .606 1 r
F P
0 g
s
.95625 .09556 m .96227 .09684 L .95945 .09268 L p .526 .605 .998 r
F P
0 g
s
.95795 .09347 m .95625 .09556 L p .95945 .09268 L .526 .605 .998 r
F P
0 g
s
.95899 .09954 m .95123 .10096 L .95625 .09556 L closepath p .955 .786 .906 r
F P
0 g
s
.95491 .09587 m .95524 .09954 L .95945 .09268 L p .161 .159 .303 r
F P
0 g
s
.95945 .09268 m .95958 .09246 L p .95929 .09267 L .161 .159 .303 r
F P
0 g
s
.95807 .09332 m .95843 .09266 L .95225 .09286 L .644 .776 1 r
F
.95865 .09215 m .95225 .09286 L p .95843 .09266 L F P
0 g
s
.95892 .09191 m .95876 .09191 L p .95852 .09261 L .828 .994 1 r
F P
0 g
s
.9599 .09194 m .95892 .09191 L p .95884 .09206 L .828 .994 1 r
F P
0 g
s
.95852 .09261 m .95902 .09237 L p .95884 .09206 L .828 .994 1 r
F P
0 g
s
.95902 .09237 m .9599 .09194 L p .95884 .09206 L .828 .994 1 r
F P
0 g
s
.96105 .09024 m .96807 .07861 L .95521 .08652 L closepath p .348 .668 1 r
F P
0 g
s
.9599 .09194 m .96691 .08034 L .96478 .0819 L p .577 .952 1 r
F P
0 g
s
.9653 .08103 m .95626 .08746 L p .95885 .09176 L .337 .654 1 r
F P
0 g
s
.95626 .08746 m .95424 .0889 L .95876 .09191 L p .337 .654 1 r
F P
0 g
s
.95876 .09191 m .9599 .09194 L p .96478 .0819 L .577 .952 1 r
F P
0 g
s
.95876 .09191 m .95906 .09211 L p .95885 .09176 L .337 .654 1 r
F P
0 g
s
.95885 .09176 m .95626 .08746 L .95876 .09191 L .337 .654 1 r
F
.95424 .0889 m .95225 .09286 L .95865 .09215 L p .643 .779 1 r
F P
0 g
s
.95876 .09191 m .95424 .0889 L p .95865 .09215 L .643 .779 1 r
F P
0 g
s
.95521 .08652 m .96214 .07484 L .94845 .08522 L closepath p .517 .881 1 r
F P
0 g
s
.96214 .07484 m .95521 .08652 L .96807 .07861 L closepath p .348 .668 1 r
F P
0 g
s
.95424 .0889 m .96115 .07725 L .94853 .08776 L closepath p .513 .877 1 r
F P
0 g
s
.96478 .0819 m .96691 .08034 L .96574 .0803 L p .577 .952 1 r
F P
0 g
s
.95424 .0889 m .96605 .0805 L .96574 .0803 L p .337 .654 1 r
F P
0 g
s
.96574 .0803 m .96115 .07725 L .95959 .07987 L p .337 .654 1 r
F P
0 g
s
.95959 .07987 m .95424 .0889 L p .96574 .0803 L .337 .654 1 r
F P
0 g
s
.96574 .0803 m .95836 .08008 L p .96478 .0819 L .577 .952 1 r
F P
0 g
s
.96417 .08044 m .95836 .08008 L .96574 .0803 L p .086 .017 .017 r
F P
0 g
s
.96574 .0803 m .96691 .08034 L .96647 .08035 L p .086 .017 .017 r
F P
0 g
s
.96553 .08012 m .96115 .07725 L .96574 .0803 L p .003 .171 .6 r
F P
0 g
s
.96574 .0803 m .96605 .0805 L .96587 .08037 L p .003 .171 .6 r
F P
0 g
s
.96587 .08037 m .96553 .08012 L p .96574 .0803 L .003 .171 .6 r
F P
0 g
s
.96647 .08035 m .96587 .08037 L p .96574 .0803 L .086 .017 .017 r
F P
0 g
s
.95906 .09211 m .96605 .0805 L .9653 .08103 L p .337 .654 1 r
F P
0 g
s
.96807 .07861 m .96105 .09024 L .97215 .08426 L closepath p .159 .396 .907 r
F P
0 g
s
.96605 .0805 m .95906 .09211 L .96029 .09129 L p .157 .371 .889 r
F P
0 g
s
.9599 .09194 m .96012 .09156 L p .96645 .08109 L .609 .8 .376 r
F P
0 g
s
.96645 .08108 m .96645 .08109 L .96645 .08109 L .609 .8 .376 r
F
.96645 .08109 m .9599 .09194 L .96645 .08109 L F
.96645 .08109 m .96645 .08108 L p .96012 .09156 L F P
0 g
s
.96649 .08102 m .9599 .09194 L p .96645 .08108 L .609 .8 .376 r
F P
0 g
s
.96507 .09581 m .97215 .08426 L .96105 .09024 L closepath p .159 .396 .907 r
F P
0 g
s
.96029 .09129 m .96931 .0853 L .96645 .08109 L p .157 .371 .889 r
F P
0 g
s
.96605 .0805 m .96029 .09129 L .96643 .08111 L .157 .371 .889 r
F
.96645 .08109 m .96643 .08106 L p .96643 .08111 L F P
0 g
s
.96643 .08106 m .96605 .0805 L p .96643 .08111 L .157 .371 .889 r
F P
0 g
s
.96478 .0819 m .95148 .09169 L .95876 .09191 L p .577 .952 1 r
F P
0 g
s
.95906 .09211 m .9653 .08103 L .95885 .09176 L .337 .654 1 r
F
.96217 .08804 m .95924 .08796 L .96632 .08092 L p 1 1 .91 r
F P
0 g
s
.96643 .08106 m .96645 .08109 L p .96628 .08088 L .644 .203 .016 r
F P
0 g
s
.96632 .08092 m .96691 .08034 L .96645 .08109 L p 1 1 .91 r
F P
0 g
s
.96645 .08109 m .96217 .08804 L p .96632 .08092 L 1 1 .91 r
F P
0 g
s
.96605 .0805 m .96643 .08106 L p .96632 .08092 L .644 .203 .016 r
F P
0 g
s
.96628 .08088 m .96628 .08087 L p .96632 .08092 L .644 .203 .016 r
F P
0 g
s
.96628 .08087 m .96605 .0805 L p .96632 .08092 L .644 .203 .016 r
F P
0 g
s
.96643 .08111 m .96628 .08088 L p .96645 .08109 L .644 .203 .016 r
F P
0 g
s
.95836 .08008 m .95148 .09169 L .96478 .0819 L p .577 .952 1 r
F P
0 g
s
.95924 .08796 m .96417 .08044 L .96588 .08037 L p 1 1 .896 r
F P
0 g
s
.96588 .08037 m .96641 .08035 L p .96611 .08067 L 1 1 .896 r
F P
0 g
s
.96587 .08037 m .96605 .0805 L p .96597 .0805 L .652 .206 .019 r
F P
0 g
s
.96587 .08038 m .96611 .08067 L .95924 .08796 L 1 1 .896 r
F
.96586 .08037 m .96597 .0805 L .96578 .08049 L .652 .206 .019 r
F
.96587 .08037 m .96597 .0805 L .96586 .08037 L F
.96578 .08049 m .96597 .0805 L .96643 .08111 L F
.96605 .0805 m .96628 .08088 L p .96597 .0805 L F P
0 g
s
.96632 .08092 m .95924 .08796 L p .96611 .08067 L 1 1 .896 r
F P
0 g
s
.96628 .08088 m .96643 .08111 L p .96597 .0805 L .652 .206 .019 r
F P
0 g
s
.96642 .08082 m .96632 .08092 L p .96641 .08035 L 1 1 .896 r
F P
0 g
s
.96641 .08035 m .96632 .08092 L .96611 .08067 L 1 1 .896 r
F
.96588 .08037 m .96611 .08067 L .96611 .08067 L F
.96588 .08037 m .96611 .08067 L .96587 .08038 L F
.96641 .08035 m .96691 .08034 L .96642 .08082 L p F P
0 g
s
.96217 .08804 m .96645 .08109 L p .9635 .08596 L .609 .8 .376 r
F P
0 g
s
.96012 .09156 m .96217 .08804 L p .9635 .08596 L .609 .8 .376 r
F P
0 g
s
.95958 .09246 m .9599 .09194 L .95801 .09343 L p .161 .159 .303 r
F P
0 g
s
.95801 .09343 m .9599 .09194 L .95852 .09261 L p .826 1 1 r
F P
0 g
s
.95852 .09261 m .95843 .09266 L p .95807 .09332 L .826 1 1 r
F P
0 g
s
.95801 .09343 m .95852 .09261 L .95807 .09332 L .826 1 1 r
F
.9582 .09315 m .95807 .09332 L p .95806 .09341 L .526 .605 .998 r
F P
0 g
s
.95807 .09332 m .95795 .09347 L p .95799 .09344 L .526 .605 .998 r
F P
0 g
s
.95807 .09332 m .95799 .09344 L .95806 .09341 L .526 .605 .998 r
F
.95945 .09268 m .95906 .09211 L .9582 .09315 L p F P
0 g
s
.95929 .09267 m .95886 .09299 L .95945 .09268 L .161 .159 .303 r
F
.9582 .09315 m .95806 .09341 L .95945 .09268 L .526 .605 .998 r
F
.96012 .09156 m .95944 .09267 L .95945 .09268 L p .157 .371 .889 r
F P
0 g
s
.96028 .09129 m .96012 .09156 L p .96029 .09129 L .157 .371 .889 r
F P
0 g
s
.96016 .09137 m .95906 .09211 L p .95941 .09262 L .157 .371 .889 r
F P
0 g
s
.95886 .09299 m .95491 .09587 L .95945 .09268 L .161 .159 .303 r
F
.95801 .09343 m .95491 .09587 L p .95958 .09246 L F P
0 g
s
.95524 .09954 m .96217 .08804 L .96012 .09156 L p .609 .8 .376 r
F P
0 g
s
.96012 .09156 m .95999 .09179 L p .9597 .09226 L .609 .8 .376 r
F P
0 g
s
.95524 .09954 m .96012 .09156 L .9597 .09226 L .609 .8 .376 r
F
.95999 .09179 m .9599 .09194 L .95943 .09272 L p F P
0 g
s
.96931 .0853 m .96029 .09129 L p .95945 .09268 L .157 .371 .889 r
F P
0 g
s
.95945 .09268 m .96227 .09684 L .96466 .09291 L p .157 .371 .889 r
F P
0 g
s
.96466 .09291 m .96931 .0853 L p .95945 .09268 L .157 .371 .889 r
F P
0 g
s
.95943 .09272 m .95524 .09954 L p .95999 .09179 L .609 .8 .376 r
F P
0 g
s
.96217 .08804 m .95524 .09954 L .96383 .08988 L closepath p .202 .474 .961 r
F P
0 g
s
.96217 .08804 m .96293 .08938 L .96164 .08888 L p .554 .084 .084 r
F P
0 g
s
.96164 .08888 m .95924 .08796 L .96217 .08804 L p .554 .084 .084 r
F P
0 g
s
.96931 .0853 m .96834 .08418 L .96643 .08111 L p .644 .203 .016 r
F P
0 g
s
.96645 .08109 m .96931 .0853 L p .96643 .08111 L .644 .203 .016 r
F P
0 g
s
.96645 .08108 m .96691 .08034 L .96649 .08102 L p .609 .8 .376 r
F P
0 g
s
.96643 .08111 m .96834 .08418 L .96632 .08126 L p .652 .206 .019 r
F P
0 g
s
.96632 .08126 m .96578 .08049 L p .96643 .08111 L .652 .206 .019 r
F P
0 g
s
.96587 .08037 m .96417 .08044 L .96574 .0803 L .086 .017 .017 r
F
.96578 .08049 m .96553 .08012 L .96587 .08037 L p .652 .206 .019 r
F P
0 g
s
.96587 .08037 m .96586 .08037 L p .96587 .08037 L .086 .017 .017 r
F P
0 g
s
.96586 .08037 m .96417 .08044 L p .96587 .08037 L .086 .017 .017 r
F P
0 g
s
.95906 .09211 m .9587 .09214 L p .95843 .09266 L .644 .776 1 r
F P
0 g
s
.95807 .09332 m .9582 .09315 L p .95843 .09266 L .644 .776 1 r
F P
0 g
s
.95843 .09266 m .95852 .09261 L p .95845 .0926 L .828 .994 1 r
F P
0 g
s
.9582 .09315 m .95906 .09211 L p .95843 .09266 L .644 .776 1 r
F P
0 g
s
.95876 .09191 m .95876 .09191 L p .95874 .09196 L .828 .994 1 r
F P
0 g
s
.95874 .09196 m .95852 .09261 L .95876 .09191 L .828 .994 1 r
F
.95876 .09191 m .95148 .09169 L p .95846 .09261 L F P
0 g
s
.95846 .09261 m .95852 .09261 L .95876 .09191 L .828 .994 1 r
F
.95865 .09215 m .95906 .09211 L .95876 .09191 L p .643 .779 1 r
F P
0 g
s
.9587 .09214 m .95865 .09215 L p .95843 .09266 L .644 .776 1 r
F P
0 g
s
.95123 .10096 m .94348 .10237 L .95225 .09286 L closepath p .955 .786 .906 r
F P
0 g
s
.95148 .09169 m .95182 .09587 L .95843 .09266 L p .828 .994 1 r
F P
0 g
s
.95148 .09169 m .95843 .09266 L .95845 .0926 L .828 .994 1 r
F
.95182 .09587 m .95491 .09587 L .95801 .09343 L p .826 1 1 r
F P
0 g
s
.95843 .09266 m .95182 .09587 L p .95801 .09343 L .826 1 1 r
F P
0 g
s
.9429 .0889 m .94277 .09286 L .94853 .08776 L closepath p .901 1 1 r
F P
0 g
s
.94277 .09286 m .9429 .0889 L .93892 .09556 L closepath p 1 1 1 r
F P
0 g
s
.93929 .10069 m .9364 .09954 L .94277 .09286 L closepath p .955 .786 .906 r
F P
0 g
s
.93892 .09556 m .94277 .09286 L .9364 .09954 L closepath p .955 .786 .906 r
F P
0 g
s
.35074 .09938 m .34984 .09704 L .34981 .09302 L closepath p .955 .786 .906 r
F P
0 g
s
.34877 .09326 m .34984 .09704 L .35068 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.55281 .04412 m .55202 .05624 L .54725 .03187 L closepath p .087 .087 .591 r
F P
0 g
s
.55202 .05624 m .55281 .04412 L .55979 .07185 L closepath p .092 .092 .629 r
F P
0 g
s
.9219 .09564 m .92529 .09411 L .92219 .09345 L closepath p .299 .202 .643 r
F P
0 g
s
.92558 .09192 m .92219 .09345 L .92529 .09411 L closepath p .299 .202 .643 r
F P
0 g
s
.65789 .09564 m .66061 .09411 L .65821 .09345 L closepath p .172 .172 .55 r
F P
0 g
s
.66094 .09192 m .65821 .09345 L .66061 .09411 L closepath p .172 .172 .55 r
F P
0 g
s
.65683 .10149 m .65629 .10383 L .65821 .09345 L closepath p .955 .786 .906 r
F P
0 g
s
.65821 .09345 m .65789 .09564 L .65683 .10149 L closepath p .955 .786 .906 r
F P
0 g
s
.92095 .10149 m .92046 .10383 L .92218 .09345 L closepath p .955 .786 .906 r
F P
0 g
s
.92218 .09345 m .9219 .09564 L .92095 .10149 L closepath p .955 .786 .906 r
F P
0 g
s
.78457 .09347 m .78829 .09688 L .78514 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.0684 .09233 m .07176 .09362 L .06415 .07754 L closepath p 1 .841 .54 r
F P
0 g
s
.07176 .09362 m .0684 .09233 L .0748 .10468 L closepath p 1 .848 .549 r
F P
0 g
s
.9051 .09314 m .90591 .0937 L .90191 .08722 L closepath p .772 .363 .042 r
F P
0 g
s
.90591 .0937 m .9051 .09314 L .90793 .09725 L closepath p .758 .34 .04 r
F P
0 g
s
.09536 .10693 m .09667 .1054 L .09808 .09375 L closepath p .088 .088 .187 r
F P
0 g
s
.09939 .09221 m .09808 .09375 L .09667 .1054 L closepath p .088 .088 .188 r
F P
0 g
s
.09049 .128 m .09014 .12843 L .09808 .09375 L closepath p .955 .786 .906 r
F P
0 g
s
.09808 .09375 m .09536 .10693 L .09049 .128 L closepath p .955 .786 .906 r
F P
0 g
s
.93618 .09032 m .93681 .09384 L .94182 .08655 L closepath p 1 1 1 r
F P
0 g
s
.93681 .09384 m .93618 .09032 L .93354 .09865 L closepath p 1 1 .935 r
F P
0 g
s
.91907 .05689 m .92544 .04478 L .89373 .0568 L closepath p .604 .969 1 r
F P
0 g
s
.89971 .04469 m .89373 .0568 L .92544 .04478 L closepath p .604 .969 1 r
F P
0 g
s
.89373 .0568 m .89971 .04469 L .89053 .05786 L closepath p .711 1 1 r
F P
0 g
s
.91221 .09912 m .91849 .08762 L .91907 .05689 L closepath p .067 .067 .133 r
F P
0 g
s
.92544 .04478 m .91907 .05689 L .91849 .08762 L closepath p .067 .067 .133 r
F P
0 g
s
.58148 .09893 m .58253 .10004 L p .58137 .09699 L .393 .493 .94 r
F P
0 g
s
.58456 .10772 m .58286 .09927 L p .58126 .09496 L 1 .882 .489 r
F P
0 g
s
.58286 .09927 m .58048 .08741 L .58126 .09496 L p 1 .882 .489 r
F P
0 g
s
.5817 .07574 m .58192 .08886 L p .58126 .09496 L .038 .038 .211 r
F P
0 g
s
.58588 .10356 m .58719 .09213 L .58167 .09119 L p .097 .097 .666 r
F P
0 g
s
.58588 .10356 m .58167 .09119 L .58135 .09419 L .097 .097 .666 r
F
.58126 .09496 m .58048 .08741 L .58073 .08504 L p .038 .038 .211 r
F P
0 g
s
.58073 .08504 m .5817 .07574 L p .58126 .09496 L .038 .038 .211 r
F P
0 g
s
.58126 .09496 m .58588 .10356 L p .58135 .09419 L .097 .097 .666 r
F P
0 g
s
.58135 .09419 m .57974 .09087 L .58126 .09496 L .097 .097 .666 r
F
.58437 .1074 m .58456 .10772 L p .58126 .09497 L 1 .882 .489 r
F P
0 g
s
.58253 .10004 m .58254 .10005 L p .58207 .09819 L .393 .493 .94 r
F P
0 g
s
.58127 .09498 m .58456 .10772 L .58126 .09496 L 1 .882 .489 r
F
.58254 .10005 m .58588 .10356 L p .58207 .09819 L .393 .493 .94 r
F P
0 g
s
.58588 .10356 m .58126 .09496 L p .58207 .09819 L .393 .493 .94 r
F P
0 g
s
.58126 .09496 m .579 .09074 L p .58253 .10004 L .393 .493 .94 r
F P
0 g
s
.577 .09421 m .58424 .10772 L .58437 .1074 L p .393 .493 .94 r
F P
0 g
s
.58253 .10004 m .577 .09421 L p .58437 .1074 L .393 .493 .94 r
F P
0 g
s
.58126 .09497 m .58216 .10356 L .58437 .1074 L p 1 .882 .489 r
F P
0 g
s
.58437 .1074 m .58588 .10356 L .58253 .10004 L p .393 .493 .94 r
F P
0 g
s
.58192 .08886 m .58177 .07976 L p .5825 .0834 L .038 .038 .211 r
F P
0 g
s
.58177 .07976 m .5817 .07574 L .5825 .0834 L p .038 .038 .211 r
F P
0 g
s
.5817 .07574 m .58048 .08741 L .57984 .05749 L closepath p .038 .038 .209 r
F P
0 g
s
.5817 .07574 m .58435 .07386 L .58379 .08486 L p .215 .22 .712 r
F P
0 g
s
.579 .09074 m .58719 .09213 L .58251 .08342 L p .097 .097 .666 r
F P
0 g
s
.58251 .08342 m .5825 .0834 L p .5825 .08341 L .097 .097 .666 r
F P
0 g
s
.5825 .08341 m .579 .09074 L .58251 .08342 L .097 .097 .666 r
F
.58719 .09213 m .58588 .10356 L .58341 .09213 L closepath p .582 .961 1 r
F P
0 g
s
.58341 .09213 m .58629 .09252 L .58719 .09213 L closepath p .137 .036 .036 r
F P
0 g
s
.58606 .09248 m .58341 .09213 L .5837 .08647 L p .211 .217 .71 r
F P
0 g
s
.5825 .0834 m .58719 .09213 L p .5839 .08697 L .96 .542 .106 r
F P
0 g
s
.58379 .08486 m .5837 .08647 L p .5825 .0834 L .215 .22 .712 r
F P
0 g
s
.5837 .08647 m .58341 .09213 L .5825 .0834 L p .215 .22 .712 r
F P
0 g
s
.5839 .08697 m .58291 .08541 L .5825 .0834 L .96 .542 .106 r
F
.58568 .09187 m .58291 .08541 L p .5839 .08697 L F P
0 g
s
.58719 .09213 m .58583 .0919 L .5839 .08697 L .96 .542 .106 r
F
.58583 .0919 m .58568 .09187 L .5839 .08697 L F
.58719 .09213 m .58596 .09252 L .58568 .09187 L p F P
0 g
s
.58596 .09252 m .58719 .09213 L .58629 .09252 L closepath p .137 .036 .036 r
F P
0 g
s
.58629 .09252 m .58606 .09248 L p .58379 .08486 L .211 .217 .71 r
F P
0 g
s
.5837 .08647 m .58379 .08486 L p .58606 .09248 L .211 .217 .71 r
F P
0 g
s
.58379 .08486 m .58435 .07386 L .58572 .08703 L p .211 .217 .71 r
F P
0 g
s
.58572 .08703 m .58629 .09252 L p .58379 .08486 L .211 .217 .71 r
F P
0 g
s
.5825 .0834 m .5817 .07574 L p .58379 .08486 L .215 .22 .712 r
F P
0 g
s
.5825 .0834 m .5802 .07912 L .58005 .08062 L p .097 .097 .666 r
F P
0 g
s
.58005 .08062 m .579 .09074 L p .5825 .0834 L .097 .097 .666 r
F P
0 g
s
.5825 .0834 m .58341 .09213 L p .58192 .08886 L .038 .038 .211 r
F P
0 g
s
.58291 .08541 m .5802 .07912 L .5825 .0834 L p .96 .542 .106 r
F P
0 g
s
.58341 .09213 m .58216 .10356 L .58192 .08886 L p .038 .038 .211 r
F P
0 g
s
.58126 .09496 m .58126 .09497 L p .58127 .09498 L 1 .882 .489 r
F P
0 g
s
.58192 .08886 m .58216 .10356 L .58126 .09496 L p .038 .038 .211 r
F P
0 g
s
.58167 .09119 m .57974 .09087 L p .58135 .09419 L .097 .097 .666 r
F P
0 g
s
.57974 .09087 m .579 .09074 L .58126 .09496 L p .097 .097 .666 r
F P
0 g
s
.579 .09074 m .577 .09421 L .58148 .09893 L p .393 .493 .94 r
F P
0 g
s
.579 .09074 m .58148 .09893 L .58137 .09699 L .393 .493 .94 r
F
.40336 .09847 m .40545 .0962 L .40274 .09427 L closepath p .171 .18 .696 r
F P
0 g
s
.40483 .09201 m .40274 .09427 L .40545 .0962 L closepath p .171 .18 .696 r
F P
0 g
s
.40336 .09847 m .40175 .09429 L .40274 .09427 L closepath p .955 .786 .906 r
F P
0 g
s
.40175 .09429 m .40336 .09847 L .40273 .10458 L closepath p .955 .786 .906 r
F P
0 g
s
.88906 .04521 m .88324 .05731 L .88223 .03174 L closepath p .143 .236 .786 r
F P
0 g
s
.89053 .05786 m .89646 .04577 L .8891 .06061 L closepath p .647 .795 .345 r
F P
0 g
s
.89501 .04855 m .8891 .06061 L .89646 .04577 L closepath p .647 .795 .345 r
F P
0 g
s
.89646 .04577 m .89053 .05786 L .89971 .04469 L closepath p .711 1 1 r
F P
0 g
s
.8891 .06061 m .89501 .04855 L .88969 .06236 L closepath p .13 .13 .667 r
F P
0 g
s
.89561 .05033 m .88969 .06236 L .89501 .04855 L closepath p .13 .13 .667 r
F P
0 g
s
.88969 .06236 m .89561 .05033 L .89138 .06609 L closepath p .14 .196 .753 r
F P
0 g
s
.88324 .05731 m .88906 .04521 L .89126 .07301 L closepath p .143 .239 .788 r
F P
0 g
s
.18153 .04521 m .1863 .05731 L .17534 .03174 L closepath p .012 .012 .338 r
F P
0 g
s
.1863 .05731 m .18153 .04521 L .19358 .07301 L closepath p .012 .012 .342 r
F P
0 g
s
.62115 .04521 m .61933 .05732 L .61453 .03175 L closepath p .103 .103 .66 r
F P
0 g
s
.61933 .05732 m .62115 .04521 L .62709 .07302 L closepath p .103 .103 .663 r
F P
0 g
s
.75763 .09464 m .75804 .09478 L .75725 .11304 L closepath p .955 .786 .906 r
F P
0 g
s
.75801 .10574 m .75725 .11304 L .75804 .09478 L closepath p .955 .786 .906 r
F P
0 g
s
.75804 .09478 m .75846 .09491 L .75801 .10574 L closepath p .955 .786 .906 r
F P
0 g
s
.75955 .09572 m .75801 .10574 L .75846 .09491 L closepath p .955 .786 .906 r
F P
0 g
s
.75859 .09518 m .76016 .09108 L .75763 .09464 L closepath p .665 .811 1 r
F P
0 g
s
.75823 .09046 m .75763 .09464 L .76016 .09108 L closepath p .722 .892 1 r
F P
0 g
s
.75763 .09464 m .75823 .09046 L .75644 .0957 L closepath p 1 1 1 r
F P
0 g
s
.75763 .09464 m .75477 .09887 L .75644 .0957 L closepath p .955 .786 .906 r
F P
0 g
s
.75763 .09464 m .75744 .10383 L .75477 .09887 L closepath p .955 .786 .906 r
F P
0 g
s
.32454 .09493 m .32552 .09574 L .32244 .10624 L closepath p .955 .786 .906 r
F P
0 g
s
.32454 .09493 m .32244 .10624 L .32081 .09889 L closepath p .955 .786 .906 r
F P
0 g
s
.32466 .0952 m .32526 .09112 L .32379 .09466 L closepath p .597 .776 1 r
F P
0 g
s
.32338 .09048 m .32379 .09466 L .32526 .09112 L closepath p .686 .879 1 r
F P
0 g
s
.32379 .09466 m .32338 .09048 L .32257 .09572 L closepath p 1 1 1 r
F P
0 g
s
.24769 .09573 m .24827 .0919 L .2457 .09466 L closepath p .606 .801 1 r
F P
0 g
s
.24511 .09048 m .2457 .09466 L .24827 .0919 L closepath p .626 .833 1 r
F P
0 g
s
.2457 .09466 m .24511 .09048 L .24494 .09513 L closepath p 1 1 1 r
F P
0 g
s
.2522 .11284 m .24977 .10677 L .2457 .09466 L closepath p .955 .786 .906 r
F P
0 g
s
.24548 .09472 m .2457 .09466 L .24977 .10677 L closepath p .955 .786 .906 r
F P
0 g
s
.32379 .09466 m .32454 .09493 L .32257 .09572 L closepath p .955 .786 .906 r
F P
0 g
s
.24769 .09573 m .2522 .11284 L .2457 .09466 L closepath p .955 .786 .906 r
F P
0 g
s
.24977 .10677 m .24762 .10187 L .24548 .09472 L closepath p .955 .786 .906 r
F P
0 g
s
.24526 .09478 m .24548 .09472 L .24762 .10187 L closepath p .955 .786 .906 r
F P
0 g
s
.24762 .10187 m .24576 .09815 L .24526 .09478 L closepath p .955 .786 .906 r
F P
0 g
s
.24503 .09484 m .24526 .09478 L .24576 .09815 L closepath p .955 .786 .906 r
F P
0 g
s
.54308 .0685 m .54373 .05656 L .54488 .05759 L closepath p .775 .727 .19 r
F P
0 g
s
.54556 .0455 m .54488 .05759 L .54373 .05656 L closepath p .775 .727 .19 r
F P
0 g
s
.68171 .09585 m .68332 .09202 L .6797 .0948 L closepath p .657 .818 1 r
F P
0 g
s
.68012 .09061 m .6797 .0948 L .68332 .09202 L closepath p .669 .846 1 r
F P
0 g
s
.6797 .0948 m .68012 .09061 L .67895 .09526 L closepath p 1 1 1 r
F P
0 g
s
.6866 .11288 m .68383 .10658 L .6797 .0948 L closepath p .955 .786 .906 r
F P
0 g
s
.67948 .09485 m .6797 .0948 L .68383 .10658 L closepath p .955 .786 .906 r
F P
0 g
s
.68171 .09585 m .6866 .11288 L .6797 .0948 L closepath p .955 .786 .906 r
F P
0 g
s
.24576 .09815 m .24418 .09559 L .24503 .09484 L closepath p .955 .786 .906 r
F P
0 g
s
.24481 .0949 m .24503 .09484 L .24418 .09559 L closepath p .955 .786 .906 r
F P
0 g
s
.68383 .10658 m .68151 .10162 L .67948 .09485 L closepath p .955 .786 .906 r
F P
0 g
s
.67925 .09491 m .67948 .09485 L .68151 .10162 L closepath p .955 .786 .906 r
F P
0 g
s
.68151 .10162 m .67963 .098 L .67925 .09491 L closepath p .955 .786 .906 r
F P
0 g
s
.67903 .09497 m .67925 .09491 L .67963 .098 L closepath p .955 .786 .906 r
F P
0 g
s
.32081 .09889 m .32257 .09572 L .32454 .09493 L closepath p .955 .786 .906 r
F P
0 g
s
.67963 .098 m .6782 .09572 L .67903 .09497 L closepath p .955 .786 .906 r
F P
0 g
s
.67881 .09503 m .67903 .09497 L .6782 .09572 L closepath p .955 .786 .906 r
F P
0 g
s
.16243 .08377 m .15729 .07205 L .15948 .05778 L closepath p .905 .678 .135 r
F P
0 g
s
.1543 .04568 m .15948 .05778 L .15729 .07205 L closepath p .905 .678 .135 r
F P
0 g
s
.07614 .09742 m .0766 .11196 L .07541 .09507 L closepath p .955 .786 .906 r
F P
0 g
s
.24292 .09129 m .24494 .09513 L .24511 .09048 L closepath p .973 1 1 r
F P
0 g
s
.24494 .09513 m .24292 .09129 L .24418 .09559 L closepath p 1 1 1 r
F P
0 g
s
.75955 .09572 m .76273 .093 L .75859 .09518 L closepath p .627 .801 1 r
F P
0 g
s
.76016 .09108 m .75859 .09518 L .76273 .093 L closepath p .611 .75 1 r
F P
0 g
s
.32552 .09574 m .32764 .09307 L .32466 .0952 L closepath p .55 .759 1 r
F P
0 g
s
.32526 .09112 m .32466 .0952 L .32764 .09307 L closepath p .523 .701 1 r
F P
0 g
s
.49253 .0958 m .4944 .09201 L .49162 .09526 L closepath p .645 .798 1 r
F P
0 g
s
.4916 .09107 m .49162 .09526 L .4944 .09201 L closepath p .701 .885 1 r
F P
0 g
s
.49162 .09526 m .4916 .09107 L .49012 .09632 L closepath p 1 1 1 r
F P
0 g
s
.67795 .09142 m .67895 .09526 L .68012 .09061 L closepath p .964 1 1 r
F P
0 g
s
.67895 .09526 m .67795 .09142 L .6782 .09572 L closepath p 1 1 1 r
F P
0 g
s
.49162 .09526 m .49345 .09634 L .49012 .09632 L closepath p .955 .786 .906 r
F P
0 g
s
.93826 .0921 m .93892 .09556 L .9429 .0889 L closepath p 1 1 1 r
F P
0 g
s
.93892 .09556 m .93826 .0921 L .9364 .09954 L closepath p 1 1 .914 r
F P
0 g
s
.24079 .09391 m .24418 .09559 L .24292 .09129 L closepath p 1 1 1 r
F P
0 g
s
.24418 .09559 m .24079 .09391 L .24576 .09815 L closepath p 1 .621 .261 r
F P
0 g
s
.87201 .09564 m .87286 .09845 L .86694 .10114 L closepath p .955 .786 .906 r
F P
0 g
s
.86694 .10114 m .86813 .09663 L .87201 .09564 L closepath p .955 .786 .906 r
F P
0 g
s
.92144 .09914 m .92484 .09761 L .9219 .09564 L closepath p .299 .202 .644 r
F P
0 g
s
.92529 .09411 m .9219 .09564 L .92484 .09761 L closepath p .299 .202 .644 r
F P
0 g
s
.65737 .09914 m .66009 .09761 L .65789 .09564 L closepath p .172 .172 .551 r
F P
0 g
s
.66061 .09411 m .65789 .09564 L .66009 .09761 L closepath p .172 .172 .551 r
F P
0 g
s
.65736 .09914 m .65683 .10149 L .65789 .09564 L closepath p .955 .786 .906 r
F P
0 g
s
.92144 .09914 m .92095 .10149 L .9219 .09564 L closepath p .955 .786 .906 r
F P
0 g
s
.92144 .09914 m .9219 .09564 L .92095 .10149 L closepath p .955 .786 .906 r
F P
0 g
s
.47381 .06553 m .47341 .05354 L .47326 .05847 L closepath p .009 .009 .086 r
F P
0 g
s
.47286 .04638 m .47326 .05847 L .47341 .05354 L closepath p .009 .009 .086 r
F P
0 g
s
.75553 .09212 m .75644 .0957 L .75823 .09046 L closepath p 1 1 1 r
F P
0 g
s
.75644 .0957 m .75553 .09212 L .75477 .09887 L closepath p 1 1 .879 r
F P
0 g
s
.32064 .09214 m .32257 .09572 L .32338 .09048 L closepath p 1 1 1 r
F P
0 g
s
.32257 .09572 m .32064 .09214 L .32081 .09889 L closepath p 1 1 .947 r
F P
0 g
s
.67591 .09403 m .6782 .09572 L .67795 .09142 L closepath p 1 1 .986 r
F P
0 g
s
.6782 .09572 m .67591 .09403 L .67963 .098 L closepath p .904 .46 .093 r
F P
0 g
s
.75801 .10574 m .76102 .10402 L .75955 .09572 L closepath p .213 .18 .582 r
F P
0 g
s
.76273 .093 m .75955 .09572 L .76102 .10402 L closepath p .216 .181 .583 r
F P
0 g
s
.25078 .09893 m .25198 .09556 L .24769 .09573 L closepath p .438 .619 1 r
F P
0 g
s
.24827 .0919 m .24769 .09573 L .25198 .09556 L closepath p .435 .627 1 r
F P
0 g
s
.25078 .09893 m .2522 .11284 L .24769 .09573 L closepath p .955 .786 .906 r
F P
0 g
s
.32368 .10574 m .3256 .10404 L .32552 .09574 L closepath p .13 .13 .363 r
F P
0 g
s
.32764 .09307 m .32552 .09574 L .3256 .10404 L closepath p .131 .131 .365 r
F P
0 g
s
.32368 .10574 m .32244 .10624 L .32552 .09574 L closepath p .955 .786 .906 r
F P
0 g
s
.53477 .10407 m .5372 .10247 L .53609 .0958 L closepath p .159 .159 .493 r
F P
0 g
s
.5372 .10247 m .53477 .10407 L .53653 .10762 L closepath p .162 .162 .519 r
F P
0 g
s
.5372 .10247 m .53653 .10762 L .53654 .1075 L p .756 .691 .143 r
F P
0 g
s
.53826 .09439 m .53609 .0958 L .5372 .10247 L p .159 .159 .493 r
F P
0 g
s
.53654 .1075 m .53709 .09624 L .5372 .10247 L p .756 .691 .143 r
F P
0 g
s
.5372 .10247 m .53851 .09423 L .53826 .09439 L p .159 .159 .493 r
F P
0 g
s
.53271 .11452 m .53242 .1149 L .53609 .0958 L closepath p .955 .786 .906 r
F P
0 g
s
.53609 .0958 m .53477 .10407 L .53271 .11452 L closepath p .955 .786 .906 r
F P
0 g
s
.53609 .0958 m .53271 .11452 L .53477 .10407 L closepath p .955 .786 .906 r
F P
0 g
s
.49345 .09634 m .49579 .09444 L .49253 .0958 L closepath p .479 .74 1 r
F P
0 g
s
.4944 .09201 m .49253 .0958 L .49579 .09444 L closepath p .413 .514 .953 r
F P
0 g
s
.46342 .04652 m .46397 .05861 L .46378 .0425 L p .78 .67 .113 r
F P
0 g
s
.46397 .05861 m .46342 .04652 L .46306 .06129 L closepath p .908 .923 .466 r
F P
0 g
s
.4625 .04925 m .46306 .06129 L .46337 .04858 L p .908 .923 .466 r
F P
0 g
s
.46337 .04858 m .46342 .04652 L .4625 .04925 L p .908 .923 .466 r
F P
0 g
s
.4597 .05064 m .4603 .06266 L .46231 .05037 L p .832 1 1 r
F P
0 g
s
.46231 .05037 m .4625 .04925 L .4597 .05064 L p .832 1 1 r
F P
0 g
s
.45371 .05111 m .45441 .06313 L .4597 .05064 L closepath p .648 1 1 r
F P
0 g
s
.4597 .05064 m .46046 .0507 L .45707 .05091 L p .012 .012 .012 r
F P
0 g
s
.45707 .05091 m .45371 .05111 L .4597 .05064 L p .012 .012 .012 r
F P
0 g
s
.4625 .04925 m .46426 .04866 L .46197 .04966 L p .024 .024 .024 r
F P
0 g
s
.46197 .04966 m .4597 .05064 L .4625 .04925 L p .024 .024 .024 r
F P
0 g
s
.46342 .04652 m .4656 .0451 L .46327 .04822 L p .128 .128 .272 r
F P
0 g
s
.46327 .04822 m .4625 .04925 L .46342 .04652 L p .128 .128 .272 r
F P
0 g
s
.46378 .0425 m .46376 .04069 L .46342 .04652 L p .78 .67 .113 r
F P
0 g
s
.68491 .09903 m .68714 .09565 L .68171 .09585 L closepath p .543 .676 1 r
F P
0 g
s
.68332 .09202 m .68171 .09585 L .68714 .09565 L closepath p .541 .683 1 r
F P
0 g
s
.68491 .09903 m .6866 .11288 L .68171 .09585 L closepath p .955 .786 .906 r
F P
0 g
s
.81093 .09531 m .81204 .09587 L .80444 .08546 L closepath p .79 .34 .058 r
F P
0 g
s
.81204 .09587 m .81093 .09531 L .82018 .10489 L closepath p .675 .174 .042 r
F P
0 g
s
.94348 .10237 m .95899 .09954 L .93946 .11281 L closepath p .955 .786 .906 r
F P
0 g
s
.95182 .09587 m .95148 .09169 L .94785 .10326 L closepath p 1 1 .866 r
F P
0 g
s
.94785 .10326 m .95148 .09169 L .94827 .09767 L p 1 1 .866 r
F P
0 g
s
.94794 .09876 m .94775 .09859 L p .94758 .09999 L .357 .307 .747 r
F P
0 g
s
.94775 .09859 m .94468 .09587 L p .94724 .09952 L .357 .307 .747 r
F P
0 g
s
.94758 .09999 m .94775 .09859 L .94724 .09952 L .357 .307 .747 r
F
.94758 .09999 m .94785 .10326 L p .94827 .09767 L 1 1 .866 r
F P
0 g
s
.94808 .09169 m .94468 .09587 L .94831 .09909 L closepath p .357 .307 .747 r
F P
0 g
s
.94827 .09767 m .9475 .09909 L .94758 .09999 L p 1 1 .866 r
F P
0 g
s
.94831 .09909 m .95514 .08759 L .95377 .08839 L p .104 .104 .398 r
F P
0 g
s
.94827 .09767 m .95638 .08353 L .94957 .09547 L .009 .009 .009 r
F
.95836 .08008 m .95632 .08364 L p .95638 .08353 L F P
0 g
s
.95632 .08364 m .94957 .09547 L p .95638 .08353 L .009 .009 .009 r
F P
0 g
s
.95311 .08893 m .95836 .08008 L p .94827 .09767 L .009 .009 .009 r
F P
0 g
s
.95528 .07352 m .94845 .08522 L .96214 .07484 L closepath p .517 .881 1 r
F P
0 g
s
.95536 .0761 m .94853 .08776 L .96115 .07725 L closepath p .513 .877 1 r
F P
0 g
s
.94957 .09547 m .95836 .08008 L .9551 .08614 L p .009 .009 .009 r
F P
0 g
s
.9549 .08008 m .94808 .09169 L .95377 .08839 L p .104 .104 .398 r
F P
0 g
s
.95377 .08839 m .95514 .08759 L .9551 .08614 L p .104 .104 .398 r
F P
0 g
s
.95514 .08759 m .94831 .09909 L .95432 .08759 L closepath p .584 .962 1 r
F P
0 g
s
.95432 .08759 m .95697 .08796 L .95514 .08759 L closepath p .132 .035 .035 r
F P
0 g
s
.94827 .09767 m .94831 .09909 L p .95377 .08839 L .104 .104 .398 r
F P
0 g
s
.95432 .08759 m .9475 .09909 L .94957 .09547 L p .009 .009 .009 r
F P
0 g
s
.95514 .08759 m .95697 .08796 L .95423 .08796 L p .132 .035 .035 r
F P
0 g
s
.95423 .08796 m .95398 .08796 L .95425 .08788 L p .132 .035 .035 r
F P
0 g
s
.95425 .08788 m .95514 .08759 L p .95423 .08796 L .132 .035 .035 r
F P
0 g
s
.95432 .08759 m .94957 .09547 L .95438 .08735 L .009 .009 .009 r
F
.9551 .08614 m .95432 .08759 L p .95438 .08735 L F P
0 g
s
.96101 .08044 m .95697 .08796 L .95586 .08781 L p .199 .168 .371 r
F P
0 g
s
.9551 .08614 m .95514 .08759 L .95489 .08767 L p 1 .828 .346 r
F P
0 g
s
.95836 .08008 m .96101 .08044 L .955 .08686 L p .199 .168 .371 r
F P
0 g
s
.955 .08686 m .95432 .08759 L .9551 .08614 L p .199 .168 .371 r
F P
0 g
s
.9551 .08614 m .95836 .08008 L p .955 .08686 L .199 .168 .371 r
F P
0 g
s
.9551 .08614 m .9551 .08614 L p .95425 .08788 L 1 .828 .346 r
F P
0 g
s
.95489 .08767 m .95432 .08759 L .955 .08686 L p .199 .168 .371 r
F P
0 g
s
.95489 .08767 m .95425 .08788 L p .9551 .08614 L 1 .828 .346 r
F P
0 g
s
.95586 .08781 m .95489 .08767 L p .95529 .08656 L .199 .168 .371 r
F P
0 g
s
.955 .08686 m .95529 .08656 L p .95489 .08767 L .199 .168 .371 r
F P
0 g
s
.95529 .08656 m .96101 .08044 L p .95586 .08781 L .199 .168 .371 r
F P
0 g
s
.9551 .08614 m .9549 .08008 L p .95377 .08839 L .104 .104 .398 r
F P
0 g
s
.94845 .08522 m .95528 .07352 L .94182 .08655 L closepath p .655 1 1 r
F P
0 g
s
.94853 .08776 m .95536 .0761 L .9429 .0889 L closepath p .655 1 1 r
F P
0 g
s
.94808 .09169 m .9549 .08008 L .94058 .09169 L closepath p .588 .964 1 r
F P
0 g
s
.94854 .07486 m .94182 .08655 L .95528 .07352 L closepath p .655 1 1 r
F P
0 g
s
.94965 .07725 m .9429 .0889 L .95536 .0761 L closepath p .655 1 1 r
F P
0 g
s
.9429 .0889 m .94965 .07725 L .93826 .0921 L closepath p .72 1 .875 r
F P
0 g
s
.94182 .08655 m .94854 .07486 L .93618 .09032 L closepath p .721 1 .888 r
F P
0 g
s
.94493 .08049 m .93826 .0921 L .94965 .07725 L closepath p .72 1 .875 r
F P
0 g
s
.94729 .08008 m .94058 .09169 L .9549 .08008 L closepath p .588 .964 1 r
F P
0 g
s
.9549 .08008 m .95373 .08044 L .94729 .08008 L closepath p .125 .033 .033 r
F P
0 g
s
.9509 .08044 m .94729 .08008 L .95373 .08044 L closepath p .125 .033 .033 r
F P
0 g
s
.94827 .09767 m .95148 .09169 L .95311 .08893 L p .009 .009 .009 r
F P
0 g
s
.95377 .08839 m .94808 .09169 L .94827 .09767 L p .104 .104 .398 r
F P
0 g
s
.94957 .09547 m .9475 .09909 L .94827 .09767 L p .009 .009 .009 r
F P
0 g
s
.94758 .09999 m .94831 .09909 L .94794 .09876 L p .357 .307 .747 r
F P
0 g
s
.94468 .09587 m .94808 .09169 L .9419 .09587 L closepath p .84 1 1 r
F P
0 g
s
.94058 .09169 m .9419 .09587 L .94808 .09169 L closepath p .84 1 1 r
F P
0 g
s
.94468 .09587 m .94491 .10326 L .94758 .09999 L p .357 .307 .747 r
F P
0 g
s
.94831 .09909 m .94491 .10326 L .9475 .09909 L closepath p .839 1 1 r
F P
0 g
s
.94243 .11258 m .9419 .09587 L .94491 .10326 L closepath p .955 .786 .906 r
F P
0 g
s
.94468 .09587 m .94491 .10326 L .9419 .09587 L closepath p .955 .786 .906 r
F P
0 g
s
.94785 .10326 m .95182 .09587 L .95007 .10326 L closepath p .955 .786 .906 r
F P
0 g
s
.94785 .10326 m .9475 .09909 L .94491 .10326 L closepath p .839 1 1 r
F P
0 g
s
.94491 .10326 m .94785 .10326 L .94243 .11258 L closepath p .955 .786 .906 r
F P
0 g
s
.94365 .10102 m .94497 .10518 L .94927 .10102 L closepath p .839 1 1 r
F P
0 g
s
.9475 .09909 m .95432 .08759 L .94831 .09909 L closepath p .584 .962 1 r
F P
0 g
s
.95491 .09587 m .95007 .10326 L .95182 .09587 L closepath p .955 .786 .906 r
F P
0 g
s
.94927 .10102 m .95612 .08954 L .94365 .10102 L closepath p .583 .962 1 r
F P
0 g
s
.95041 .08954 m .94365 .10102 L .95612 .08954 L closepath p .583 .962 1 r
F P
0 g
s
.96002 .10426 m .93946 .11281 L .95899 .09954 L closepath p .955 .786 .906 r
F P
0 g
s
.95007 .10326 m .94243 .11258 L .94785 .10326 L closepath p .955 .786 .906 r
F P
0 g
s
.94822 .10518 m .94927 .10102 L .94497 .10518 L closepath p .839 1 1 r
F P
0 g
s
.94625 .10791 m .95046 .10922 L .94243 .11258 L closepath p .955 .786 .906 r
F P
0 g
s
.94974 .10853 m .95162 .11263 L .93946 .11281 L closepath p .955 .786 .906 r
F P
0 g
s
.94497 .10518 m .94365 .10102 L .94514 .1105 L closepath p 1 .836 .36 r
F P
0 g
s
.9522 .10803 m .95129 .10988 L .94497 .10518 L closepath p .955 .786 .906 r
F P
0 g
s
.94506 .10784 m .94497 .10518 L .95129 .10988 L closepath p .955 .786 .906 r
F P
0 g
s
.95116 .10589 m .94497 .10518 L .94822 .10518 L closepath p .955 .786 .906 r
F P
0 g
s
.9522 .10803 m .94497 .10518 L .95116 .10589 L closepath p .955 .786 .906 r
F P
0 g
s
.94396 .11093 m .94514 .1105 L .94365 .10102 L closepath p 1 .836 .36 r
F P
0 g
s
.94365 .10102 m .95041 .08954 L .94396 .11093 L closepath p .103 .103 .395 r
F P
0 g
s
.95007 .10326 m .95524 .09954 L .95491 .09587 L closepath p .189 .163 .307 r
F P
0 g
s
.95368 .10464 m .95686 .10135 L .95007 .10326 L closepath p .704 .869 1 r
F P
0 g
s
.95524 .09954 m .95007 .10326 L .95686 .10135 L closepath p .698 .723 1 r
F P
0 g
s
.95007 .10326 m .95368 .10464 L .94625 .10791 L closepath p .955 .786 .906 r
F P
0 g
s
.94927 .10102 m .94822 .10518 L .95372 .10245 L closepath p .729 .892 1 r
F P
0 g
s
.95129 .10988 m .94839 .1105 L .94506 .10784 L closepath p .955 .786 .906 r
F P
0 g
s
.94514 .1105 m .94506 .10784 L .94839 .1105 L closepath p .955 .786 .906 r
F P
0 g
s
.95686 .10135 m .96383 .08988 L .95524 .09954 L closepath p .202 .474 .961 r
F P
0 g
s
.95372 .10245 m .96063 .091 L .94927 .10102 L closepath p .458 .815 1 r
F P
0 g
s
.95612 .08954 m .94927 .10102 L .96063 .091 L closepath p .458 .815 1 r
F P
0 g
s
.95073 .0996 m .94396 .11093 L .95041 .08954 L closepath p .103 .103 .395 r
F P
0 g
s
.94729 .08008 m .9509 .08044 L .94797 .10171 L closepath p .348 .299 .742 r
F P
0 g
s
.95612 .08954 m .95736 .08993 L .95041 .08954 L closepath p .134 .035 .035 r
F P
0 g
s
.95404 .08993 m .95041 .08954 L .95736 .08993 L closepath p .134 .035 .035 r
F P
0 g
s
.95041 .08954 m .95404 .08993 L .95073 .0996 L closepath p .349 .3 .742 r
F P
0 g
s
.96146 .07736 m .96115 .07725 L .96553 .08012 L closepath p .007 .178 .613 r
F P
0 g
s
.96115 .07725 m .96146 .07736 L .95536 .0761 L closepath p .156 .005 .005 r
F P
0 g
s
.95536 .0761 m .9566 .07644 L .94965 .07725 L closepath p .067 .052 .052 r
F P
0 g
s
.95373 .08044 m .9549 .08008 L .95398 .08796 L p 1 .828 .346 r
F P
0 g
s
.95398 .08796 m .95398 .08796 L .95375 .08081 L p 1 .828 .346 r
F P
0 g
s
.95375 .08081 m .95373 .08044 L p .95398 .08796 L 1 .828 .346 r
F P
0 g
s
.95145 .09746 m .9509 .08044 L .95398 .08796 L closepath p .955 .785 .905 r
F P
0 g
s
.95373 .08044 m .95398 .08796 L .9509 .08044 L closepath p .955 .785 .905 r
F P
0 g
s
.9514 .08669 m .9511 .08657 L p .95373 .08039 L .955 .785 .905 r
F P
0 g
s
.95374 .0807 m .95373 .08039 L .95526 .0768 L .955 .785 .905 r
F
.95374 .0807 m .95526 .0768 L .95527 .07679 L F
.94797 .10171 m .9509 .08044 L .95111 .08696 L p .348 .299 .742 r
F P
0 g
s
.95111 .08696 m .95112 .0874 L p .95112 .0874 L .348 .299 .742 r
F P
0 g
s
.95112 .0874 m .94797 .10171 L .95111 .08696 L .348 .299 .742 r
F
.9372 .1 m .94239 .10201 L .93386 .10989 L closepath p .955 .786 .906 r
F P
0 g
s
.93929 .10069 m .94348 .10237 L .93669 .1089 L closepath p .955 .786 .906 r
F P
0 g
s
.93737 .11462 m .93386 .10989 L .94239 .10201 L closepath p .955 .786 .906 r
F P
0 g
s
.93946 .11281 m .93669 .1089 L .94348 .10237 L closepath p .955 .786 .906 r
F P
0 g
s
.9419 .09587 m .94058 .09169 L .94243 .11258 L closepath p 1 .838 .362 r
F P
0 g
s
.94348 .11547 m .93946 .11281 L .95162 .11263 L closepath p .955 .786 .906 r
F P
0 g
s
.94124 .11301 m .94243 .11258 L .94058 .09169 L closepath p 1 .838 .362 r
F P
0 g
s
.94058 .09169 m .94729 .08008 L .94124 .11301 L closepath p .103 .103 .393 r
F P
0 g
s
.94797 .10171 m .94124 .11301 L .94729 .08008 L closepath p .103 .103 .393 r
F P
0 g
s
.94101 .09093 m .94454 .08898 L .9423 .0965 L closepath p .399 .404 .842 r
F P
0 g
s
.9423 .0965 m .94454 .08898 L .94528 .09234 L p .39 .395 .836 r
F P
0 g
s
.94528 .09234 m .94559 .09371 L .94559 .09371 L p .39 .395 .836 r
F P
0 g
s
.94559 .09371 m .9423 .0965 L p .94528 .09234 L .39 .395 .836 r
F P
0 g
s
.94973 .08199 m .95178 .07736 L .95539 .07667 L p .955 .785 .905 r
F P
0 g
s
.94529 .08418 m .95178 .07736 L .94973 .08199 L p .955 .785 .905 r
F P
0 g
s
.9511 .08657 m .94746 .08517 L p .9478 .08462 L .955 .785 .905 r
F P
0 g
s
.9511 .08657 m .9478 .08462 L .94914 .0828 L .955 .785 .905 r
F
.94749 .08505 m .94529 .08418 L p .94973 .08199 L F P
0 g
s
.9511 .08657 m .94905 .08263 L .95092 .08088 L .955 .785 .905 r
F
.94973 .08199 m .94824 .08535 L .94749 .08505 L p F P
0 g
s
.94914 .0828 m .9478 .08462 L .94905 .08263 L .955 .785 .905 r
F
.95092 .08088 m .94905 .08263 L .95088 .0797 L F
.94454 .08898 m .94529 .08418 L .94749 .08505 L p F P
0 g
s
.94749 .08505 m .94748 .0851 L .94454 .08898 L .955 .785 .905 r
F
.94746 .08517 m .9461 .08465 L .94807 .0826 L p F P
0 g
s
.94807 .0826 m .95088 .0797 L p .94746 .08517 L .955 .785 .905 r
F P
0 g
s
.94749 .08505 m .94751 .08506 L p .94748 .0851 L .955 .785 .905 r
F P
0 g
s
.94746 .08517 m .9511 .08657 L p .94609 .08933 L .955 .785 .905 r
F P
0 g
s
.94528 .09234 m .94454 .08898 L .94677 .08679 L p .955 .785 .905 r
F P
0 g
s
.94677 .08679 m .94609 .08933 L .94555 .08963 L .955 .785 .905 r
F
.94746 .08517 m .94609 .08933 L .94677 .08679 L F
.94555 .08963 m .9511 .08657 L .94526 .09224 L F
.95111 .08696 m .9427 .09472 L p .9511 .08657 L F P
0 g
s
.9423 .0965 m .94559 .09371 L .9456 .10124 L closepath p .514 .602 1 r
F P
0 g
s
.94559 .09371 m .94842 .0977 L .94842 .0977 L p .521 .608 1 r
F P
0 g
s
.94842 .0977 m .9456 .10124 L .94559 .09371 L p .521 .608 1 r
F P
0 g
s
.94559 .09371 m .94559 .09371 L p .94842 .0977 L .521 .608 1 r
F P
0 g
s
.96311 .10426 m .95024 .10944 L .9618 .09855 L closepath p .955 .786 .906 r
F P
0 g
s
.96311 .10426 m .96663 .10238 L .9618 .09855 L closepath p .415 .416 .847 r
F P
0 g
s
.96507 .09581 m .9618 .09855 L .96663 .10238 L closepath p .41 .415 .849 r
F P
0 g
s
.96227 .09684 m .95899 .09954 L .96353 .10238 L closepath p .407 .41 .844 r
F P
0 g
s
.96002 .10426 m .96353 .10238 L .95899 .09954 L closepath p .411 .409 .842 r
F P
0 g
s
.96002 .10426 m .95929 .1089 L .94974 .10853 L closepath p .955 .786 .906 r
F P
0 g
s
.96663 .10238 m .97374 .09093 L .96507 .09581 L closepath p .131 .131 .615 r
F P
0 g
s
.97215 .08426 m .96507 .09581 L .97374 .09093 L closepath p .131 .131 .615 r
F P
0 g
s
.96353 .10238 m .97059 .09093 L .96227 .09684 L closepath p .129 .129 .605 r
F P
0 g
s
.96931 .0853 m .96227 .09684 L .97059 .09093 L closepath p .129 .129 .605 r
F P
0 g
s
.95847 .10585 m .96546 .09445 L .95686 .10135 L closepath p .139 .145 .707 r
F P
0 g
s
.96383 .08988 m .95686 .10135 L .96546 .09445 L closepath p .139 .145 .707 r
F P
0 g
s
.96311 .10426 m .96218 .10994 L .95024 .10944 L closepath p .955 .786 .906 r
F P
0 g
s
.95497 .10772 m .95847 .10585 L .95368 .10464 L closepath p .45 .496 .92 r
F P
0 g
s
.95686 .10135 m .95368 .10464 L .95847 .10585 L closepath p .461 .485 .903 r
F P
0 g
s
.9557 .10616 m .96264 .09476 L .95372 .10245 L closepath p .15 .276 .816 r
F P
0 g
s
.96063 .091 m .95372 .10245 L .96264 .09476 L closepath p .15 .276 .816 r
F P
0 g
s
.95368 .10464 m .95497 .10772 L .95046 .10922 L closepath p .955 .786 .906 r
F P
0 g
s
.95162 .11263 m .94974 .10853 L .95929 .1089 L closepath p .955 .786 .906 r
F P
0 g
s
.95424 .10954 m .95791 .10881 L .95497 .10772 L closepath p .319 .191 .527 r
F P
0 g
s
.95847 .10585 m .95497 .10772 L .95791 .10881 L closepath p .289 .191 .61 r
F P
0 g
s
.9535 .11136 m .95046 .10922 L .95497 .10772 L closepath p .955 .786 .906 r
F P
0 g
s
.95257 .11441 m .95024 .10944 L .96218 .10994 L closepath p .955 .786 .906 r
F P
0 g
s
.95392 .10963 m .96084 .09828 L .9557 .10616 L closepath p .015 .015 .015 r
F P
0 g
s
.96264 .09476 m .9557 .10616 L .96084 .09828 L closepath p .015 .015 .015 r
F P
0 g
s
.95046 .10922 m .94625 .10791 L .95368 .10464 L closepath p .955 .786 .906 r
F P
0 g
s
.9522 .10803 m .9557 .10616 L .95116 .10589 L closepath p .461 .521 .942 r
F P
0 g
s
.95372 .10245 m .95116 .10589 L .9557 .10616 L closepath p .458 .531 .953 r
F P
0 g
s
.9557 .10616 m .9522 .10803 L .95392 .10963 L closepath p .162 .162 .349 r
F P
0 g
s
.96084 .09828 m .95392 .10963 L .95644 .0996 L closepath p .666 1 1 r
F P
0 g
s
.96084 .09828 m .96049 .09471 L .96264 .09476 L closepath p 1 1 .806 r
F P
0 g
s
.95644 .0996 m .95753 .09534 L .96084 .09828 L closepath p .926 1 1 r
F P
0 g
s
.96049 .09471 m .96084 .09828 L .95753 .09534 L closepath p .882 1 1 r
F P
0 g
s
.95073 .0996 m .95421 .09534 L .95644 .0996 L closepath p .836 1 1 r
F P
0 g
s
.95753 .09534 m .95644 .0996 L .95421 .09534 L p .836 1 1 r
F P
0 g
s
.95421 .09534 m .95421 .09534 L .95673 .09534 L p .836 1 1 r
F P
0 g
s
.95673 .09534 m .95753 .09534 L p .95421 .09534 L .836 1 1 r
F P
0 g
s
.95073 .0996 m .95404 .08993 L p .95405 .09036 L .349 .3 .742 r
F P
0 g
s
.95421 .09534 m .95073 .0996 L p .95405 .09036 L .349 .3 .742 r
F P
0 g
s
.95144 .09746 m .94797 .10171 L p .95122 .09041 L .348 .299 .742 r
F P
0 g
s
.95046 .10922 m .9535 .11136 L .9456 .11258 L closepath p .955 .786 .906 r
F P
0 g
s
.95129 .10988 m .95392 .10963 L .9522 .10803 L closepath p .162 .162 .354 r
F P
0 g
s
.95024 .10944 m .95257 .11441 L .93737 .11462 L closepath p .955 .786 .906 r
F P
0 g
s
.9456 .11258 m .94243 .11258 L .95046 .10922 L closepath p .955 .786 .906 r
F P
0 g
s
.94839 .1105 m .94959 .11093 L .95129 .10988 L closepath p .097 .062 .062 r
F P
0 g
s
.95392 .10963 m .95129 .10988 L .94959 .11093 L closepath p .048 .048 .048 r
F P
0 g
s
.94959 .11093 m .95644 .0996 L .95392 .10963 L closepath p .666 1 1 r
F P
0 g
s
.95116 .10589 m .95372 .10245 L .94822 .10518 L closepath p .732 .919 1 r
F P
0 g
s
.94514 .1105 m .94396 .11093 L .94839 .1105 L closepath p .156 .041 .041 r
F P
0 g
s
.94959 .11093 m .94839 .1105 L .94396 .11093 L closepath p .156 .041 .041 r
F P
0 g
s
.94396 .11093 m .95073 .0996 L .94959 .11093 L closepath p .578 .96 1 r
F P
0 g
s
.95644 .0996 m .94959 .11093 L .95073 .0996 L closepath p .578 .96 1 r
F P
0 g
s
.94797 .10171 m .95145 .09746 L .95684 .10171 L closepath p .836 1 1 r
F P
0 g
s
.95118 .089 m .94797 .10171 L .95112 .0874 L .348 .299 .742 r
F
.95122 .09041 m .95122 .09041 L .95118 .089 L F
.95122 .09041 m .94797 .10171 L .95118 .089 L F
.9427 .09472 m .9514 .08669 L .95048 .08901 L p .955 .785 .905 r
F P
0 g
s
.95048 .08901 m .94629 .09955 L .94347 .09575 L p .955 .785 .905 r
F P
0 g
s
.94347 .09575 m .9427 .09472 L p .95048 .08901 L .955 .785 .905 r
F P
0 g
s
.94559 .09371 m .94824 .08535 L .9511 .08649 L p .955 .785 .905 r
F P
0 g
s
.9511 .08657 m .9514 .08669 L .95111 .08696 L p .955 .785 .905 r
F P
0 g
s
.95123 .09079 m .95122 .09041 L .95122 .09041 L .348 .299 .742 r
F
.95144 .09746 m .95122 .09041 L .95123 .09079 L F
.95137 .09511 m .95122 .09041 L .95122 .09041 L F
.95144 .09746 m .95122 .09041 L .95137 .09511 L F
.95112 .0874 m .95122 .09041 L p .95122 .09041 L F P
0 g
s
.95122 .09041 m .95145 .09746 L .95144 .09746 L p .348 .299 .742 r
F P
0 g
s
.95398 .08796 m .9549 .08008 L .9551 .08614 L p 1 .828 .346 r
F P
0 g
s
.95425 .08788 m .95398 .08796 L p .9551 .08614 L 1 .828 .346 r
F P
0 g
s
.96807 .07861 m .96762 .07828 L .96214 .07484 L closepath p .003 .188 .609 r
F P
0 g
s
.96248 .07496 m .96214 .07484 L .96762 .07828 L closepath p .008 .196 .623 r
F P
0 g
s
.96146 .07736 m .9566 .07644 L .9566 .07644 L p .17 .01 .01 r
F P
0 g
s
.9566 .07644 m .95536 .0761 L .96146 .07736 L p .17 .01 .01 r
F P
0 g
s
.96146 .07736 m .96146 .07736 L p .9566 .07644 L .17 .01 .01 r
F P
0 g
s
.95178 .07736 m .94965 .07725 L .9566 .07644 L p .05 .047 .047 r
F P
0 g
s
.9566 .07644 m .9566 .07644 L .95539 .07667 L p .05 .047 .047 r
F P
0 g
s
.95539 .07667 m .95178 .07736 L p .9566 .07644 L .05 .047 .047 r
F P
0 g
s
.95088 .0797 m .95344 .07705 L p .9511 .08657 L .955 .785 .905 r
F P
0 g
s
.95344 .07705 m .95442 .07603 L p .95414 .07691 L .955 .785 .905 r
F P
0 g
s
.9511 .08657 m .95344 .07705 L .95414 .07691 L .955 .785 .905 r
F
.95442 .07603 m .95651 .07386 L p .9555 .07624 L F P
0 g
s
.9511 .08657 m .95442 .07603 L .9555 .07624 L .955 .785 .905 r
F
.96101 .08044 m .95836 .08008 L .96417 .08044 L closepath p .125 .033 .033 r
F P
0 g
s
.95651 .07386 m .96248 .07496 L .94629 .09955 L closepath p .955 .785 .905 r
F P
0 g
s
.9514 .08669 m .95143 .08663 L .95143 .08663 L .955 .785 .905 r
F
.95143 .08663 m .9553 .0767 L .95143 .08663 L F
.95143 .08663 m .9553 .0767 L .95532 .0767 L F
.9514 .08669 m .95373 .08039 L .95374 .0807 L F
.95558 .07619 m .9514 .08669 L p .95552 .07618 L F P
0 g
s
.9511 .08649 m .95251 .08706 L .95116 .08836 L p .955 .785 .905 r
F P
0 g
s
.95116 .08836 m .94559 .09371 L p .9511 .08649 L .955 .785 .905 r
F P
0 g
s
.94878 .11258 m .94997 .11301 L .9535 .11136 L closepath p .087 .066 .066 r
F P
0 g
s
.94878 .11258 m .9456 .11258 L .9535 .11136 L closepath p .955 .786 .906 r
F P
0 g
s
.95684 .10171 m .94997 .11301 L .94797 .10171 L closepath p .577 .959 1 r
F P
0 g
s
.95252 .10041 m .95052 .10443 L .94842 .0977 L p .636 .778 1 r
F P
0 g
s
.94842 .0977 m .94842 .0977 L .94935 .09831 L p .636 .778 1 r
F P
0 g
s
.94935 .09831 m .95252 .10041 L p .94842 .0977 L .636 .778 1 r
F P
0 g
s
.94896 .09847 m .94935 .09831 L p .94969 .0995 L .955 .785 .905 r
F P
0 g
s
.94935 .09831 m .95146 .09746 L p .95067 .09837 L .955 .785 .905 r
F P
0 g
s
.95792 .09746 m .95684 .10171 L .95145 .09746 L closepath p .836 1 1 r
F P
0 g
s
.95427 .09704 m .95422 .09706 L .95424 .09634 L .955 .785 .905 r
F
.95422 .09706 m .9537 .09714 L .95424 .09634 L F
.95555 .09687 m .95468 .09746 L .95146 .09746 L p F P
0 g
s
.95347 .09746 m .95263 .09746 L .95424 .09634 L .955 .785 .905 r
F
.95146 .09746 m .95424 .09634 L p .95263 .09746 L F P
0 g
s
.95412 .09263 m .95404 .08993 L .96049 .09471 L closepath p .955 .785 .905 r
F P
0 g
s
.96049 .09471 m .95753 .09534 L .95412 .09263 L closepath p .955 .785 .905 r
F P
0 g
s
.95421 .09534 m .95412 .09263 L .95753 .09534 L closepath p .955 .785 .905 r
F P
0 g
s
.95424 .09634 m .95942 .09426 L p .95427 .09704 L .955 .785 .905 r
F P
0 g
s
.95754 .09866 m .95252 .10041 L .94935 .09831 L p .955 .785 .905 r
F P
0 g
s
.95171 .09841 m .95151 .09947 L .95148 .09854 L .955 .785 .905 r
F
.95741 .09534 m .95295 .09847 L .95171 .09841 L F
.95295 .09847 m .95151 .09947 L .95171 .09841 L F
.95778 .09534 m .95151 .09947 L .95741 .09534 L F
.96546 .09445 m .95847 .10585 L .9649 .09745 L closepath p .072 .072 .137 r
F P
0 g
s
.95684 .10171 m .95792 .09746 L .96313 .09974 L closepath p .93 1 1 r
F P
0 g
s
.96274 .09623 m .96313 .09974 L .95792 .09746 L closepath p .896 1 1 r
F P
0 g
s
.96274 .09623 m .95468 .09746 L .95555 .09687 L p .955 .785 .905 r
F P
0 g
s
.95792 .09746 m .95468 .09746 L .96274 .09623 L closepath p .955 .785 .905 r
F P
0 g
s
.96218 .10994 m .96551 .10894 L .96311 .10426 L closepath p .291 .191 .624 r
F P
0 g
s
.95929 .1089 m .96262 .10787 L .96002 .10426 L closepath p .293 .192 .628 r
F P
0 g
s
.96663 .10238 m .96311 .10426 L .96551 .10894 L closepath p .298 .192 .623 r
F P
0 g
s
.96353 .10238 m .96002 .10426 L .96262 .10787 L closepath p .299 .192 .627 r
F P
0 g
s
.95791 .10881 m .9649 .09745 L .95847 .10585 L closepath p .072 .072 .137 r
F P
0 g
s
.95929 .1089 m .9568 .11281 L .95162 .11263 L closepath p .955 .786 .906 r
F P
0 g
s
.95791 .10881 m .95424 .10954 L .95617 .11107 L closepath p .252 .169 .309 r
F P
0 g
s
.95617 .11107 m .96313 .09974 L .95791 .10881 L closepath p .646 .88 .513 r
F P
0 g
s
.9649 .09745 m .95791 .10881 L .96313 .09974 L closepath p .646 .88 .513 r
F P
0 g
s
.96313 .09974 m .95617 .11107 L .95684 .10171 L closepath p .668 1 1 r
F P
0 g
s
.96313 .09974 m .96274 .09623 L .9649 .09745 L closepath p 1 1 .951 r
F P
0 g
s
.96611 .0977 m .96221 .10041 L .96115 .09819 L p .955 .785 .905 r
F P
0 g
s
.96115 .09819 m .96124 .09814 L .96611 .0977 L .955 .785 .905 r
F
.97059 .09093 m .96353 .10238 L .96968 .09649 L closepath p .077 .077 .171 r
F P
0 g
s
.96968 .09649 m .96865 .09371 L .97059 .09093 L closepath p 1 .963 .538 r
F P
0 g
s
.96264 .09476 m .96142 .09283 L .96063 .091 L closepath p .682 .287 .016 r
F P
0 g
s
.9649 .09745 m .96349 .09437 L .96546 .09445 L closepath p 1 1 .643 r
F P
0 g
s
.96968 .09649 m .96611 .0977 L .96807 .09463 L p 1 1 .901 r
F P
0 g
s
.96807 .09463 m .96865 .09371 L .96865 .09371 L p 1 1 .901 r
F P
0 g
s
.96865 .09371 m .96968 .09649 L p .96807 .09463 L 1 1 .901 r
F P
0 g
s
.96966 .095 m .96791 .09487 L .9704 .09472 L .955 .785 .905 r
F
.96347 .09443 m .9649 .09745 L .96274 .09623 L closepath p 1 1 .819 r
F P
0 g
s
.96069 .09432 m .96113 .09343 L p .96142 .09283 L 1 1 .8 r
F P
0 g
s
.96113 .09343 m .96142 .09283 L .96142 .09283 L p 1 1 .8 r
F P
0 g
s
.96142 .09283 m .96142 .09283 L p .96113 .09343 L 1 1 .8 r
F P
0 g
s
.96347 .09443 m .96274 .09623 L .96194 .09566 L p .955 .785 .905 r
F P
0 g
s
.96264 .09476 m .96049 .09471 L .96069 .09432 L p 1 1 .8 r
F P
0 g
s
.96142 .09283 m .96264 .09476 L p .96069 .09432 L 1 1 .8 r
F P
0 g
s
.96005 .0956 m .96067 .09435 L .96766 .09486 L .955 .785 .905 r
F
.96807 .09463 m .96611 .0977 L .96273 .09758 L p F P
0 g
s
.96711 .09516 m .96766 .09486 L .96966 .095 L .955 .785 .905 r
F
.96005 .0956 m .96766 .09486 L .96711 .09516 L F
.96194 .09566 m .96217 .09547 L .96966 .095 L F
.96015 .09581 m .96274 .09623 L .95555 .09687 L F
.96015 .09581 m .96217 .09547 L .96194 .09566 L F
.96166 .09546 m .96194 .09566 L p .96015 .09581 L F P
0 g
s
.96194 .09566 m .96274 .09623 L p .96015 .09581 L .955 .785 .905 r
F P
0 g
s
.96124 .09814 m .96015 .09581 L p .96301 .09557 L .955 .785 .905 r
F P
0 g
s
.9621 .09782 m .96301 .09557 L .96966 .095 L .955 .785 .905 r
F
.96124 .09814 m .96301 .09557 L .9621 .09782 L F
.96791 .09487 m .96346 .09455 L .96805 .09467 L F
.96348 .09443 m .9704 .09472 L p .96346 .09455 L F P
0 g
s
.96221 .10041 m .95252 .10041 L .95754 .09866 L p .955 .785 .905 r
F P
0 g
s
.95993 .09534 m .96125 .09818 L p .95778 .09534 L .955 .785 .905 r
F P
0 g
s
.96125 .09818 m .96114 .09819 L .95778 .09534 L .955 .785 .905 r
F
.96115 .09819 m .96221 .10041 L p .95754 .09866 L F P
0 g
s
.96134 .09835 m .96106 .0986 L .95778 .09534 L .955 .785 .905 r
F
.95542 .0969 m .96022 .09873 L .95151 .09947 L F
.95861 .09616 m .9618 .09933 L .95542 .0969 L F
.95754 .09866 m .96083 .09751 L .96115 .09819 L p F P
0 g
s
.96125 .09818 m .96134 .09835 L p .96114 .09819 L .955 .785 .905 r
F P
0 g
s
.96134 .09835 m .9618 .09933 L p .96106 .0986 L .955 .785 .905 r
F P
0 g
s
.97374 .09093 m .96663 .10238 L .9726 .09758 L closepath p .077 .077 .165 r
F P
0 g
s
.97059 .09093 m .96939 .08898 L .96931 .0853 L closepath p .917 .627 .092 r
F P
0 g
s
.96546 .09445 m .96424 .09252 L .96383 .08988 L closepath p .826 .485 .045 r
F P
0 g
s
.96293 .08938 m .96383 .08988 L .96424 .09252 L closepath p .85 .48 .057 r
F P
0 g
s
.96939 .08898 m .97059 .09093 L .96865 .09371 L closepath p 1 .962 .538 r
F P
0 g
s
.96546 .09445 m .96349 .09437 L .96383 .09353 L p 1 1 .743 r
F P
0 g
s
.96383 .09353 m .96424 .09252 L .96424 .09252 L p 1 1 .743 r
F P
0 g
s
.96424 .09252 m .96546 .09445 L p .96383 .09353 L 1 1 .743 r
F P
0 g
s
.9716 .09477 m .95942 .09426 L .96147 .09344 L p .955 .785 .905 r
F P
0 g
s
.96147 .09344 m .97254 .08898 L .97176 .09383 L p .955 .785 .905 r
F P
0 g
s
.97176 .09383 m .9716 .09477 L p .96147 .09344 L .955 .785 .905 r
F P
0 g
s
.96142 .09283 m .96049 .09471 L .95404 .08993 L closepath p .955 .785 .905 r
F P
0 g
s
.95942 .09426 m .96348 .09443 L p .96346 .09455 L .955 .785 .905 r
F P
0 g
s
.96834 .08418 m .96931 .0853 L .96939 .08898 L closepath p .91 .626 .088 r
F P
0 g
s
.96063 .091 m .96036 .09065 L .95612 .08954 L closepath p .023 .445 .757 r
F P
0 g
s
.96036 .09065 m .96063 .091 L .96142 .09283 L closepath p .67 .294 .009 r
F P
0 g
s
.97254 .08898 m .95942 .09426 L .97121 .08317 L closepath p .955 .785 .905 r
F P
0 g
s
.96113 .09343 m .96383 .09353 L p .96071 .09531 L .955 .785 .905 r
F P
0 g
s
.96015 .09581 m .96005 .0956 L p .96217 .09547 L .955 .785 .905 r
F P
0 g
s
.96002 .09536 m .96166 .09546 L .96015 .09581 L .955 .785 .905 r
F
.96 .09571 m .96018 .09535 L .96071 .09531 L F
.96018 .09535 m .96113 .09343 L .96071 .09531 L F
.96137 .09525 m .96166 .09546 L p .96002 .09536 L F P
0 g
s
.96383 .09353 m .96347 .09443 L p .96266 .09435 L .955 .785 .905 r
F P
0 g
s
.96194 .09566 m .96137 .09525 L p .96347 .09443 L .955 .785 .905 r
F P
0 g
s
.96137 .09525 m .96266 .09435 L .96347 .09443 L .955 .785 .905 r
F
.96229 .09518 m .96383 .09353 L .96326 .09496 L F
.96 .09571 m .96071 .09531 L .96229 .09518 L F
.96071 .09531 m .96383 .09353 L .96229 .09518 L F
.95778 .09534 m .95861 .09616 L .95542 .0969 L F
.95942 .09426 m .95778 .09534 L .95741 .09534 L F
.95942 .09426 m .95993 .09534 L p .95778 .09534 L F P
0 g
s
.95421 .09534 m .95421 .09534 L p .95405 .09036 L .349 .3 .742 r
F P
0 g
s
.95404 .08993 m .95404 .08993 L p .95421 .09529 L .349 .3 .742 r
F P
0 g
s
.95404 .08993 m .95404 .08993 L .95417 .09414 L p .349 .3 .742 r
F P
0 g
s
.95417 .09414 m .95421 .09529 L p .95404 .08993 L .349 .3 .742 r
F P
0 g
s
.95421 .09529 m .95421 .09534 L p .95404 .08993 L .349 .3 .742 r
F P
0 g
s
.95612 .08954 m .96036 .09065 L .96014 .0906 L p .195 .007 .007 r
F P
0 g
s
.96014 .0906 m .95736 .08993 L .95736 .08993 L p .195 .007 .007 r
F P
0 g
s
.95736 .08993 m .95612 .08954 L p .96014 .0906 L .195 .007 .007 r
F P
0 g
s
.9607 .08617 m .95942 .09426 L .95144 .09747 L p .955 .785 .905 r
F P
0 g
s
.9589 .09334 m .95984 .09537 L p .95567 .09633 L .955 .785 .905 r
F P
0 g
s
.95984 .09537 m .95654 .09618 L .95567 .09633 L .955 .785 .905 r
F
.96002 .09536 m .96015 .09581 L .95863 .09567 L F
.95984 .09537 m .95863 .09567 L .95744 .09556 L F
.96002 .09536 m .95863 .09567 L .95984 .09537 L F
.95863 .09567 m .96015 .09581 L .95693 .09609 L F
.95555 .09687 m .95654 .09618 L p .96015 .09581 L F P
0 g
s
.95654 .09618 m .95664 .09611 L p .95693 .09609 L .955 .785 .905 r
F P
0 g
s
.95693 .09609 m .96015 .09581 L .95654 .09618 L .955 .785 .905 r
F
.95555 .09687 m .955 .09695 L .95511 .09685 L F
.955 .09695 m .95145 .09746 L .95511 .09685 L F
.95145 .09746 m .95313 .09676 L p .95511 .09685 L F P
0 g
s
.95654 .09618 m .95555 .09687 L p .95313 .09676 L .955 .785 .905 r
F P
0 g
s
.95146 .09746 m .95145 .09746 L p .95555 .09687 L .955 .785 .905 r
F P
0 g
s
.96083 .09751 m .95428 .09761 L p .95558 .09642 L .955 .785 .905 r
F P
0 g
s
.95567 .09633 m .95581 .09636 L .95558 .09642 L .955 .785 .905 r
F
.95581 .09636 m .96083 .09751 L .95558 .09642 L F
.95995 .0956 m .96083 .09751 L p .95581 .09636 L F P
0 g
s
.95654 .09618 m .95581 .09636 L .95567 .09633 L .955 .785 .905 r
F
.95995 .0956 m .95581 .09636 L .95654 .09618 L F
.96005 .0956 m .95942 .09426 L p .96067 .09435 L F P
0 g
s
.96383 .08988 m .96293 .08938 L .96217 .08804 L closepath p .636 .139 .031 r
F P
0 g
s
.96095 .08461 m .97121 .08317 L .95942 .09426 L closepath p .955 .785 .905 r
F P
0 g
s
.9589 .09334 m .96113 .09343 L p .96005 .09561 L .955 .785 .905 r
F P
0 g
s
.96 .09571 m .95995 .0956 L p .96005 .09561 L .955 .785 .905 r
F P
0 g
s
.95995 .0956 m .9589 .09334 L p .96005 .09561 L .955 .785 .905 r
F P
0 g
s
.95422 .09706 m .95347 .09746 L .9537 .09714 L .955 .785 .905 r
F
.95347 .09746 m .94969 .0995 L .95263 .09746 L F
.95263 .09746 m .95238 .09764 L .95178 .09746 L F
.95238 .09764 m .94969 .0995 L .95178 .09746 L F
.95178 .09746 m .94969 .0995 L .95146 .09746 L F
.95428 .09761 m .95145 .09765 L p .95423 .09604 L F P
0 g
s
.9589 .09334 m .95428 .09761 L .95423 .09604 L .955 .785 .905 r
F
.95313 .09676 m .95606 .09554 L p .95654 .09618 L F P
0 g
s
.95606 .09554 m .95964 .09404 L .95744 .09556 L p .955 .785 .905 r
F P
0 g
s
.95744 .09556 m .95654 .09618 L p .95606 .09554 L .955 .785 .905 r
F P
0 g
s
.94243 .11258 m .94124 .11301 L .94878 .11258 L closepath p .158 .041 .041 r
F P
0 g
s
.94997 .11301 m .94878 .11258 L .94124 .11301 L closepath p .158 .041 .041 r
F P
0 g
s
.94124 .11301 m .94797 .10171 L .94997 .11301 L closepath p .577 .959 1 r
F P
0 g
s
.9456 .10124 m .94842 .0977 L .95052 .10443 L closepath p .642 .783 1 r
F P
0 g
s
.95144 .09747 m .94629 .09955 L p .9607 .08617 L .955 .785 .905 r
F P
0 g
s
.95924 .08796 m .95145 .09746 L .95697 .08796 L closepath p .955 .785 .905 r
F P
0 g
s
.95398 .08796 m .95697 .08796 L .95145 .09746 L closepath p .955 .785 .905 r
F P
0 g
s
.95141 .09645 m .95272 .09591 L p .95145 .09765 L .955 .785 .905 r
F P
0 g
s
.95272 .09591 m .9589 .09334 L p .9531 .0967 L .955 .785 .905 r
F P
0 g
s
.95313 .09676 m .95145 .09746 L .95272 .09591 L p .955 .785 .905 r
F P
0 g
s
.9531 .0967 m .95145 .09765 L .95272 .09591 L .955 .785 .905 r
F
.96939 .08898 m .94842 .0977 L .96834 .08418 L closepath p F P
0 g
s
.95424 .09629 m .95313 .09676 L p .95417 .09414 L .955 .785 .905 r
F P
0 g
s
.95272 .09591 m .95417 .09414 L p .95313 .09676 L .955 .785 .905 r
F P
0 g
s
.95251 .08706 m .96834 .08418 L .94842 .0977 L closepath p .955 .785 .905 r
F P
0 g
s
.96939 .08898 m .96865 .09371 L .9589 .09334 L closepath p .955 .785 .905 r
F P
0 g
s
.96036 .09065 m .95404 .08993 L .95736 .08993 L closepath p .955 .785 .905 r
F P
0 g
s
.96142 .09283 m .95404 .08993 L .96036 .09065 L closepath p .955 .785 .905 r
F P
0 g
s
.95964 .09404 m .95534 .09271 L .96014 .0906 L p .955 .785 .905 r
F P
0 g
s
.96014 .0906 m .96293 .08938 L .96178 .091 L p .955 .785 .905 r
F P
0 g
s
.96178 .091 m .95964 .09404 L p .96014 .0906 L .955 .785 .905 r
F P
0 g
s
.94629 .09955 m .96248 .07496 L .9607 .08617 L p .955 .785 .905 r
F P
0 g
s
.95534 .09271 m .95964 .09404 L .95424 .09629 L p .955 .785 .905 r
F P
0 g
s
.95417 .09414 m .95534 .09271 L p .95424 .09629 L .955 .785 .905 r
F P
0 g
s
.95984 .09537 m .95995 .0956 L p .95654 .09618 L .955 .785 .905 r
F P
0 g
s
.95664 .09611 m .95744 .09556 L p .95863 .09567 L .955 .785 .905 r
F P
0 g
s
.95863 .09567 m .95693 .09609 L .95664 .09611 L .955 .785 .905 r
F
.95744 .09556 m .95964 .09404 L p .96002 .09536 L F P
0 g
s
.95964 .09404 m .96137 .09525 L p .96002 .09536 L .955 .785 .905 r
F P
0 g
s
.95964 .09404 m .96424 .09252 L .96383 .09353 L p .955 .785 .905 r
F P
0 g
s
.96266 .09435 m .95964 .09404 L .96383 .09353 L .955 .785 .905 r
F
.96383 .09353 m .96865 .09371 L p .96326 .09496 L F P
0 g
s
.96865 .09371 m .96807 .09463 L p .96347 .09646 L .955 .785 .905 r
F P
0 g
s
.96083 .09751 m .96405 .09595 L p .96347 .09646 L .955 .785 .905 r
F P
0 g
s
.96405 .09595 m .96865 .09371 L p .96347 .09646 L .955 .785 .905 r
F P
0 g
s
.96273 .09758 m .96083 .09751 L p .96807 .09463 L .955 .785 .905 r
F P
0 g
s
.96137 .09525 m .95964 .09404 L p .96266 .09435 L .955 .785 .905 r
F P
0 g
s
.96865 .09371 m .96083 .09751 L .96017 .09609 L p .955 .785 .905 r
F P
0 g
s
.96017 .09609 m .96 .09571 L p .96107 .09546 L .955 .785 .905 r
F P
0 g
s
.96017 .09609 m .96107 .09546 L .96865 .09371 L .955 .785 .905 r
F
.96791 .09487 m .96805 .09467 L .9704 .09472 L F
.96223 .09756 m .96273 .09758 L p .96124 .09814 L F P
0 g
s
.96273 .09758 m .96611 .0977 L p .96124 .09814 L .955 .785 .905 r
F P
0 g
s
.9716 .09477 m .9618 .09933 L .96124 .09814 L p .955 .785 .905 r
F P
0 g
s
.9704 .09472 m .9716 .09477 L p .96124 .09814 L .955 .785 .905 r
F P
0 g
s
.96115 .09819 m .96083 .09751 L .96223 .09756 L p .955 .785 .905 r
F P
0 g
s
.9618 .09933 m .95911 .09937 L p .96022 .09873 L .955 .785 .905 r
F P
0 g
s
.95911 .09937 m .95151 .09947 L p .96022 .09873 L .955 .785 .905 r
F P
0 g
s
.94842 .0977 m .96083 .09751 L .95754 .09866 L p .955 .785 .905 r
F P
0 g
s
.94935 .09831 m .94842 .0977 L p .95754 .09866 L .955 .785 .905 r
F P
0 g
s
.95067 .09837 m .95022 .09889 L .94935 .09831 L .955 .785 .905 r
F
.95022 .09889 m .94969 .0995 L .94935 .09831 L F
.95145 .09765 m .94842 .0977 L .95141 .09645 L p F P
0 g
s
.95151 .09947 m .94969 .0995 L p .95148 .09854 L .955 .785 .905 r
F P
0 g
s
.94969 .0995 m .94629 .09955 L .94896 .09847 L p .955 .785 .905 r
F P
0 g
s
.94824 .08535 m .94559 .09371 L .94528 .09234 L p .955 .785 .905 r
F P
0 g
s
.94677 .08679 m .94824 .08535 L p .94528 .09234 L .955 .785 .905 r
F P
0 g
s
.94751 .08506 m .94824 .08535 L .94677 .08679 L p .955 .785 .905 r
F P
0 g
s
.94677 .08679 m .94454 .08898 L p .94751 .08506 L .955 .785 .905 r
F P
0 g
s
.9461 .08465 m .94746 .08517 L p .94555 .08963 L .955 .785 .905 r
F P
0 g
s
.9427 .09472 m .94458 .08917 L p .94526 .09224 L .955 .785 .905 r
F P
0 g
s
.94458 .08917 m .9461 .08465 L p .94526 .09224 L .955 .785 .905 r
F P
0 g
s
.9566 .07644 m .94824 .08535 L .94973 .08199 L p .955 .785 .905 r
F P
0 g
s
.95539 .07667 m .9566 .07644 L p .94973 .08199 L .955 .785 .905 r
F P
0 g
s
.95527 .07679 m .95526 .0768 L .9553 .0767 L .955 .785 .905 r
F
.95532 .0767 m .9553 .0767 L .95552 .07618 L F
.95651 .07386 m .95558 .07619 L p .95552 .07618 L F P
0 g
s
.95251 .08706 m .9511 .08649 L p .9537 .07952 L .955 .785 .905 r
F P
0 g
s
.9511 .08649 m .94824 .08535 L .95097 .08244 L p .955 .785 .905 r
F P
0 g
s
.95097 .08244 m .9537 .07952 L p .9511 .08649 L .955 .785 .905 r
F P
0 g
s
.9537 .07952 m .9566 .07644 L .95384 .08362 L p .955 .785 .905 r
F P
0 g
s
.95384 .08362 m .95251 .08706 L p .9537 .07952 L .955 .785 .905 r
F P
0 g
s
.82018 .10489 m .81204 .09587 L .82729 .10074 L closepath p .955 .786 .906 r
F P
0 g
s
.59515 .09779 m .59574 .10283 L .58802 .09603 L closepath p .955 .786 .906 r
F P
0 g
s
.5856 .10491 m .58802 .09603 L .59574 .10283 L closepath p .955 .786 .906 r
F P
0 g
s
.16203 .09871 m .16284 .10801 L .15787 .09613 L closepath p .955 .786 .906 r
F P
0 g
s
.1548 .10632 m .15787 .09613 L .16284 .10801 L closepath p .955 .786 .906 r
F P
0 g
s
.48766 .0995 m .49012 .09632 L .49345 .09634 L closepath p .955 .786 .906 r
F P
0 g
s
.48876 .09259 m .49012 .09632 L .4916 .09107 L closepath p 1 1 1 r
F P
0 g
s
.49012 .09632 m .48876 .09259 L .48766 .0995 L closepath p 1 1 1 r
F P
0 g
s
.49316 .09934 m .49549 .09765 L .49345 .09634 L closepath p .16 .16 .53 r
F P
0 g
s
.49579 .09444 m .49345 .09634 L .49549 .09765 L closepath p .16 .16 .531 r
F P
0 g
s
.49254 .10329 m .49345 .09634 L .49316 .09934 L closepath p .955 .786 .906 r
F P
0 g
s
.49213 .10535 m .49345 .09634 L .49254 .10329 L closepath p .955 .786 .906 r
F P
0 g
s
.49345 .09634 m .49213 .10535 L .48766 .0995 L closepath p .955 .786 .906 r
F P
0 g
s
.65014 .10936 m .64606 .1036 L .6436 .09653 L closepath p .955 .786 .906 r
F P
0 g
s
.6436 .09653 m .64494 .09872 L .65014 .10936 L closepath p .955 .786 .906 r
F P
0 g
s
.86677 .09411 m .86813 .09663 L .86802 .08918 L p 1 1 .663 r
F P
0 g
s
.86813 .09663 m .86677 .09411 L .86694 .10114 L closepath p 1 1 .698 r
F P
0 g
s
.86563 .09845 m .86694 .10114 L .86678 .09484 L p 1 1 .694 r
F P
0 g
s
.86678 .09484 m .86677 .09411 L .86563 .09845 L p 1 1 .694 r
F P
0 g
s
.86694 .10114 m .86563 .09845 L .86583 .10448 L closepath p 1 1 .752 r
F P
0 g
s
.86461 .10159 m .86583 .10448 L .86566 .09934 L p 1 1 .746 r
F P
0 g
s
.86566 .09934 m .86563 .09845 L .86461 .10159 L p 1 1 .746 r
F P
0 g
s
.86583 .10448 m .86461 .10159 L .8648 .10665 L closepath p 1 1 .848 r
F P
0 g
s
.86369 .10355 m .8648 .10665 L .86466 .10303 L p 1 1 .843 r
F P
0 g
s
.86466 .10303 m .86461 .10159 L .86369 .10355 L p 1 1 .843 r
F P
0 g
s
.86461 .10159 m .87017 .09012 L .8638 .10331 L p .002 .002 .002 r
F P
0 g
s
.8638 .10331 m .86369 .10355 L .86461 .10159 L p .002 .002 .002 r
F P
0 g
s
.86563 .09845 m .8712 .08694 L .8647 .10139 L p .027 .027 .027 r
F P
0 g
s
.8647 .10139 m .86461 .10159 L .86563 .09845 L p .027 .027 .027 r
F P
0 g
s
.87118 .08584 m .87683 .07414 L .86916 .08671 L closepath p .72 1 1 r
F P
0 g
s
.86677 .09411 m .87235 .08254 L .86571 .09827 L p .039 .039 .039 r
F P
0 g
s
.86571 .09827 m .86563 .09845 L .86677 .09411 L p .039 .039 .039 r
F P
0 g
s
.87393 .08677 m .87962 .0751 L .87118 .08584 L closepath p .444 .803 1 r
F P
0 g
s
.87683 .07414 m .87118 .08584 L .87962 .0751 L closepath p .444 .803 1 r
F P
0 g
s
.8712 .08694 m .86563 .09845 L .87235 .08254 L closepath p .039 .039 .039 r
F P
0 g
s
.87017 .09012 m .86461 .10159 L .8712 .08694 L closepath p .027 .027 .027 r
F P
0 g
s
.86923 .09212 m .86369 .10355 L .86976 .09097 L p .002 .002 .002 r
F P
0 g
s
.86976 .09097 m .87017 .09012 L .86923 .09212 L p .002 .002 .002 r
F P
0 g
s
.86923 .09212 m .87223 .09142 L .86849 .09372 L p .115 .115 .115 r
F P
0 g
s
.86369 .10355 m .86923 .09212 L .86051 .10526 L p .743 1 .803 r
F P
0 g
s
.86443 .09622 m .85895 .1076 L .86712 .0953 L p .743 1 .803 r
F P
0 g
s
.86712 .0953 m .86923 .09212 L .86443 .09622 L p .743 1 .803 r
F P
0 g
s
.85895 .1076 m .86443 .09622 L .85094 .10889 L closepath p .64 1 1 r
F P
0 g
s
.8487 .09627 m .84346 .10765 L .85423 .09916 L p .508 .885 1 r
F P
0 g
s
.85423 .09916 m .85629 .09753 L .8487 .09627 L p .508 .885 1 r
F P
0 g
s
.84346 .10765 m .8487 .09627 L .83646 .10385 L closepath p .341 .692 1 r
F P
0 g
s
.84159 .09242 m .83646 .10385 L .8487 .09627 L closepath p .341 .692 1 r
F P
0 g
s
.8487 .09627 m .849 .09654 L .84557 .09463 L p .37 .009 .009 r
F P
0 g
s
.84557 .09463 m .84159 .09242 L .8487 .09627 L p .37 .009 .009 r
F P
0 g
s
.849 .09654 m .8487 .09627 L .8531 .09713 L p .194 .02 .02 r
F P
0 g
s
.85628 .09755 m .85094 .10889 L .86218 .09833 L p .64 1 1 r
F P
0 g
s
.85094 .10889 m .85629 .09753 L .84346 .10765 L closepath p .508 .885 1 r
F P
0 g
s
.8531 .09713 m .85726 .09793 L .85016 .09674 L p .194 .02 .02 r
F P
0 g
s
.85016 .09674 m .849 .09654 L p .8531 .09713 L .194 .02 .02 r
F P
0 g
s
.85629 .09753 m .85628 .09755 L p .86223 .09829 L .64 1 1 r
F P
0 g
s
.86218 .09833 m .86223 .09829 L p .85628 .09755 L .64 1 1 r
F P
0 g
s
.86223 .09829 m .86443 .09622 L .85629 .09753 L p .64 1 1 r
F P
0 g
s
.85629 .09753 m .85726 .09793 L .8531 .09713 L p .2 .022 .022 r
F P
0 g
s
.8531 .09713 m .8487 .09627 L .85629 .09753 L p .2 .022 .022 r
F P
0 g
s
.86443 .09622 m .86647 .0963 L .86297 .09673 L p .037 .037 .037 r
F P
0 g
s
.86297 .09673 m .85629 .09753 L .86443 .09622 L p .037 .037 .037 r
F P
0 g
s
.86849 .09372 m .86443 .09622 L .86923 .09212 L p .115 .115 .115 r
F P
0 g
s
.86051 .10526 m .85895 .1076 L .86369 .10355 L p .743 1 .803 r
F P
0 g
s
.86802 .08918 m .86801 .08852 L .86677 .09411 L p 1 1 .663 r
F P
0 g
s
.78942 .1028 m .78514 .11481 L .78829 .09688 L closepath p .955 .786 .906 r
F P
0 g
s
.78942 .1028 m .79246 .10051 L .78829 .09688 L closepath p .347 .354 .809 r
F P
0 g
s
.79133 .0946 m .78829 .09688 L .79246 .10051 L closepath p .347 .354 .809 r
F P
0 g
s
.35099 .10888 m .35068 .11489 L .34984 .09704 L closepath p .955 .786 .906 r
F P
0 g
s
.34984 .09704 m .35074 .09938 L .35099 .10888 L closepath p .955 .786 .906 r
F P
0 g
s
.1341 .08217 m .12853 .07042 L .12297 .05998 L closepath p .005 .005 .334 r
F P
0 g
s
.11724 .04791 m .12297 .05998 L .12853 .07042 L closepath p .005 .005 .334 r
F P
0 g
s
.90756 .09712 m .90793 .09725 L .9051 .09314 L closepath p .687 .261 .024 r
F P
0 g
s
.90793 .09725 m .90756 .09712 L .90901 .09872 L closepath p .647 .189 .021 r
F P
0 g
s
.07614 .09742 m .07688 .09975 L .0766 .11196 L closepath p .955 .786 .906 r
F P
0 g
s
.59574 .10283 m .59832 .10066 L .59515 .09779 L closepath p .233 .24 .728 r
F P
0 g
s
.59773 .09564 m .59515 .09779 L .59832 .10066 L closepath p .233 .24 .728 r
F P
0 g
s
.51534 .07025 m .51557 .05834 L .51237 .06059 L closepath p .058 .058 .414 r
F P
0 g
s
.51256 .04854 m .51237 .06059 L .51557 .05834 L closepath p .058 .058 .414 r
F P
0 g
s
.0799 .10868 m .07995 .10855 L .08509 .09798 L closepath p .955 .786 .906 r
F P
0 g
s
.08509 .09798 m .07717 .11545 L .0799 .10868 L closepath p .955 .786 .906 r
F P
0 g
s
.07993 .09519 m .08294 .09799 L .08218 .08477 L closepath p 1 1 .826 r
F P
0 g
s
.08294 .09799 m .07993 .09519 L .08106 .10519 L closepath p 1 1 .849 r
F P
0 g
s
.67784 .09734 m .67963 .098 L .67591 .09403 L closepath p .947 .515 .105 r
F P
0 g
s
.67963 .098 m .67784 .09734 L .68151 .10162 L closepath p .977 .566 .114 r
F P
0 g
s
.11403 .06083 m .10817 .04878 L .12244 .0788 L closepath p .653 .289 .003 r
F P
0 g
s
.39748 .11195 m .39125 .10901 L .39169 .09811 L closepath p .955 .786 .906 r
F P
0 g
s
.39204 .10588 m .39169 .09811 L .39125 .10901 L closepath p .955 .786 .906 r
F P
0 g
s
.39204 .10588 m .39125 .10901 L .39169 .09811 L closepath p .955 .786 .906 r
F P
0 g
s
.39169 .09811 m .38907 .09641 L .39204 .10588 L closepath p 1 .942 .613 r
F P
0 g
s
.24278 .09741 m .24576 .09815 L .24079 .09391 L closepath p 1 .666 .322 r
F P
0 g
s
.24576 .09815 m .24278 .09741 L .24762 .10187 L closepath p 1 .709 .369 r
F P
0 g
s
.22333 .10395 m .21635 .11543 L .22439 .09826 L closepath p .955 .786 .906 r
F P
0 g
s
.87286 .09845 m .8737 .10126 L .86583 .10448 L closepath p .955 .786 .906 r
F P
0 g
s
.86583 .10448 m .86694 .10114 L .87286 .09845 L closepath p .955 .786 .906 r
F P
0 g
s
.81524 .07004 m .82003 .05812 L .81311 .06125 L closepath p .108 .108 .544 r
F P
0 g
s
.81786 .0492 m .81311 .06125 L .82003 .05812 L closepath p .108 .108 .544 r
F P
0 g
s
.40379 .10134 m .40587 .09907 L .40336 .09847 L closepath p .171 .182 .698 r
F P
0 g
s
.40545 .0962 m .40336 .09847 L .40587 .09907 L closepath p .171 .182 .698 r
F P
0 g
s
.40379 .10134 m .40273 .10458 L .40336 .09847 L closepath p .955 .786 .906 r
F P
0 g
s
.22708 .09106 m .22292 .07944 L .23335 .06141 L closepath p 1 .937 .484 r
F P
0 g
s
.2293 .04937 m .23335 .06141 L .22292 .07944 L closepath p 1 .937 .484 r
F P
0 g
s
.66107 .09106 m .66352 .07944 L .66644 .06142 L closepath p .005 .005 .005 r
F P
0 g
s
.66896 .04938 m .66644 .06142 L .66352 .07944 L closepath p .005 .005 .005 r
F P
0 g
s
.92569 .09106 m .93217 .07944 L .93058 .06143 L closepath p .069 .069 .139 r
F P
0 g
s
.93712 .04938 m .93058 .06143 L .93217 .07944 L closepath p .069 .069 .139 r
F P
0 g
s
.93246 .0959 m .93354 .09865 L .93618 .09032 L closepath p 1 1 .925 r
F P
0 g
s
.93354 .09865 m .93246 .0959 L .93257 .10426 L closepath p 1 .981 .57 r
F P
0 g
s
.93354 .09865 m .9372 .1 L .93257 .10426 L closepath p .955 .786 .906 r
F P
0 g
s
.16284 .10801 m .16434 .10584 L .16203 .09871 L closepath p .137 .137 .55 r
F P
0 g
s
.16353 .09655 m .16203 .09871 L .16434 .10584 L closepath p .137 .137 .55 r
F P
0 g
s
.90932 .09912 m .90901 .09872 L .90756 .09712 L closepath p .54 .059 .003 r
F P
0 g
s
.90901 .09872 m .90932 .09912 L .90916 .09872 L closepath p .145 .038 .038 r
F P
0 g
s
.91221 .09912 m .90916 .09872 L .90932 .09912 L closepath p .145 .038 .038 r
F P
0 g
s
.65423 .11512 m .64494 .09872 L .64525 .09872 L closepath p .955 .786 .906 r
F P
0 g
s
.65423 .11512 m .65014 .10936 L .64494 .09872 L closepath p .955 .786 .906 r
F P
0 g
s
.75324 .09604 m .75477 .09887 L .75553 .09212 L closepath p 1 1 .922 r
F P
0 g
s
.75477 .09887 m .75324 .09604 L .75233 .10351 L closepath p 1 1 .896 r
F P
0 g
s
.75233 .10351 m .75477 .09887 L .75744 .10383 L closepath p .955 .786 .906 r
F P
0 g
s
.31823 .09604 m .32081 .09889 L .32064 .09214 L closepath p 1 1 .972 r
F P
0 g
s
.32081 .09889 m .31823 .09604 L .31817 .10352 L closepath p 1 1 .958 r
F P
0 g
s
.31817 .10352 m .32081 .09889 L .32244 .10624 L closepath p .955 .786 .906 r
F P
0 g
s
.2522 .11284 m .25078 .09893 L .25842 .10676 L closepath p .955 .786 .906 r
F P
0 g
s
.25514 .10372 m .25842 .10676 L .25078 .09893 L closepath p .955 .786 .906 r
F P
0 g
s
.25514 .10372 m .25078 .09893 L .25842 .10676 L closepath p .955 .786 .906 r
F P
0 g
s
.25514 .10372 m .25622 .10022 L .25078 .09893 L closepath p .419 .598 1 r
F P
0 g
s
.25198 .09556 m .25078 .09893 L .25622 .10022 L closepath p .42 .598 1 r
F P
0 g
s
.33863 .09896 m .33587 .09728 L .33907 .10333 L closepath p 1 .918 .589 r
F P
0 g
s
.6866 .11288 m .68491 .09903 L .69278 .1068 L closepath p .955 .786 .906 r
F P
0 g
s
.68941 .10378 m .69278 .1068 L .68491 .09903 L closepath p .955 .786 .906 r
F P
0 g
s
.68941 .10378 m .68491 .09903 L .69278 .1068 L closepath p .955 .786 .906 r
F P
0 g
s
.68941 .10378 m .69152 .10027 L .68491 .09903 L closepath p .531 .661 1 r
F P
0 g
s
.68714 .09565 m .68491 .09903 L .69152 .10027 L closepath p .532 .661 1 r
F P
0 g
s
.92082 .10395 m .92421 .10243 L .92144 .09914 L closepath p .299 .202 .644 r
F P
0 g
s
.92484 .09761 m .92144 .09914 L .92421 .10243 L closepath p .299 .202 .644 r
F P
0 g
s
.65665 .10395 m .65938 .10243 L .65737 .09914 L closepath p .172 .172 .551 r
F P
0 g
s
.66009 .09761 m .65737 .09914 L .65938 .10243 L closepath p .172 .172 .551 r
F P
0 g
s
.65683 .10149 m .65736 .09914 L .65553 .11017 L closepath p .955 .786 .906 r
F P
0 g
s
.65665 .10395 m .65553 .11017 L .65736 .09914 L closepath p .955 .786 .906 r
F P
0 g
s
.92095 .10149 m .92144 .09914 L .91981 .11017 L closepath p .955 .786 .906 r
F P
0 g
s
.92082 .10395 m .91981 .11017 L .92144 .09914 L closepath p .955 .786 .906 r
F P
0 g
s
.42878 .05 m .42985 .06204 L .42753 .0414 L closepath p .011 .011 .14 r
F P
0 g
s
.49229 .10574 m .49461 .10418 L .49316 .09934 L closepath p .157 .157 .496 r
F P
0 g
s
.49549 .09765 m .49316 .09934 L .49461 .10418 L closepath p .157 .157 .496 r
F P
0 g
s
.49316 .09934 m .49229 .10574 L .49254 .10329 L closepath p .955 .786 .906 r
F P
0 g
s
.3513 .10287 m .35099 .10888 L .35074 .09938 L closepath p .955 .786 .906 r
F P
0 g
s
.3513 .10287 m .35325 .10057 L .35074 .09939 L closepath p .167 .167 .69 r
F P
0 g
s
.48567 .09633 m .48766 .0995 L .48876 .09259 L closepath p 1 1 1 r
F P
0 g
s
.48766 .0995 m .48567 .09633 L .48418 .10418 L closepath p 1 1 .999 r
F P
0 g
s
.48418 .10418 m .48766 .0995 L .49213 .10535 L closepath p .955 .786 .906 r
F P
0 g
s
.07984 .1088 m .07903 .10354 L .07688 .09954 L closepath p .077 .077 .487 r
F P
0 g
s
.07881 .10152 m .07993 .09519 L .07824 .09214 L p 1 .964 .537 r
F P
0 g
s
.07724 .09601 m .07688 .09954 L .07874 .10299 L p .104 .104 .526 r
F P
0 g
s
.0787 .10214 m .07724 .09601 L p .07874 .10299 L .104 .104 .526 r
F P
0 g
s
.07487 .08735 m .0684 .07568 L .0692 .06762 L closepath p .844 .534 .047 r
F P
0 g
s
.07724 .09601 m .0708 .08447 L .07487 .08735 L closepath p .855 .551 .049 r
F P
0 g
s
.0684 .07568 m .07487 .08735 L .0708 .08447 L closepath p .855 .551 .049 r
F P
0 g
s
.07874 .10299 m .07903 .10354 L .0787 .10214 L p .104 .104 .526 r
F P
0 g
s
.07903 .10354 m .07262 .09211 L .07724 .09601 L closepath p .88 .59 .055 r
F P
0 g
s
.07262 .09211 m .0708 .08447 L .07598 .09376 L p .88 .59 .055 r
F P
0 g
s
.07598 .09376 m .07724 .09601 L .07262 .09211 L p .88 .59 .055 r
F P
0 g
s
.07239 .09005 m .07881 .10152 L .07354 .08363 L closepath p 1 .964 .537 r
F P
0 g
s
.07881 .10152 m .07239 .09005 L .07903 .10354 L closepath p .963 .728 .207 r
F P
0 g
s
.07262 .09211 m .07903 .10354 L .07239 .09005 L closepath p .963 .728 .207 r
F P
0 g
s
.07262 .09211 m .07128 .09361 L .0708 .08447 L closepath p 1 .772 .335 r
F P
0 g
s
.07239 .09005 m .07252 .08994 L .07262 .09211 L closepath p .013 .013 .126 r
F P
0 g
s
.07262 .09211 m .07252 .08994 L .07191 .09176 L p .918 .929 .476 r
F P
0 g
s
.07191 .09176 m .07128 .09361 L .07262 .09211 L p .918 .929 .476 r
F P
0 g
s
.07824 .09214 m .07354 .08363 L .07864 .10096 L p 1 .964 .537 r
F P
0 g
s
.07864 .10096 m .07881 .10152 L p .07824 .09214 L 1 .964 .537 r
F P
0 g
s
.93531 .09683 m .9364 .09954 L .93826 .0921 L closepath p 1 1 .905 r
F P
0 g
s
.93386 .10989 m .93257 .10426 L .9372 .1 L closepath p .955 .786 .906 r
F P
0 g
s
.9364 .09954 m .93531 .09683 L .93566 .10426 L closepath p 1 .97 .553 r
F P
0 g
s
.9364 .09954 m .93929 .10069 L .93566 .10426 L closepath p .955 .786 .906 r
F P
0 g
s
.07762 .10205 m .0766 .11196 L .07688 .09975 L closepath p .955 .786 .906 r
F P
0 g
s
.09945 .05053 m .10545 .06256 L .09697 .06327 L closepath p 1 .969 .542 r
F P
0 g
s
.77335 .05076 m .76926 .06278 L .77192 .04235 L closepath p .09 .09 .439 r
F P
0 g
s
.59455 .07 m .59599 .05808 L .59352 .06298 L closepath p .049 .049 .28 r
F P
0 g
s
.59494 .05095 m .59352 .06298 L .59599 .05808 L closepath p .049 .049 .28 r
F P
0 g
s
.93669 .1089 m .93566 .10426 L .93929 .10069 L closepath p .955 .786 .906 r
F P
0 g
s
.45053 .10434 m .45246 .10228 L .44958 .10073 L closepath p .171 .236 .756 r
F P
0 g
s
.4518 .09852 m .44958 .10073 L .45246 .10228 L closepath p .183 .226 .732 r
F P
0 g
s
.45148 .10796 m .43798 .10582 L .44958 .10073 L closepath p .955 .786 .906 r
F P
0 g
s
.83441 .10757 m .83646 .10385 L .82729 .10074 L closepath p .562 .69 1 r
F P
0 g
s
.82983 .09739 m .82729 .10074 L .83646 .10385 L closepath p .567 .69 1 r
F P
0 g
s
.82729 .10074 m .83441 .10757 L .82018 .10489 L closepath p .955 .786 .906 r
F P
0 g
s
.52095 .10258 m .52044 .10403 L .51984 .10099 L closepath p .955 .786 .906 r
F P
0 g
s
.87401 .10322 m .87728 .10104 L .8737 .10126 L closepath p .369 .359 .804 r
F P
0 g
s
.87699 .09922 m .8737 .10126 L .87728 .10104 L closepath p .369 .36 .805 r
F P
0 g
s
.8648 .10665 m .86583 .10448 L .8737 .10126 L closepath p .955 .786 .906 r
F P
0 g
s
.8737 .10126 m .87401 .10322 L .87057 .11165 L closepath p .955 .786 .906 r
F P
0 g
s
.8737 .10126 m .87057 .11165 L .8648 .10665 L closepath p .955 .786 .906 r
F P
0 g
s
.40403 .10289 m .40612 .10061 L .40379 .10134 L closepath p .172 .189 .705 r
F P
0 g
s
.40587 .09907 m .40379 .10134 L .40612 .10061 L closepath p .172 .189 .705 r
F P
0 g
s
.40273 .10458 m .40379 .10134 L .40371 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.40403 .10289 m .40371 .11489 L .40379 .10134 L closepath p .955 .786 .906 r
F P
0 g
s
.13304 .10055 m .13605 .10138 L .12244 .0788 L closepath p 1 .715 .376 r
F P
0 g
s
.13605 .10138 m .13304 .10055 L .14866 .12673 L closepath p 1 .709 .369 r
F P
0 g
s
.13605 .10138 m .1502 .12162 L .14866 .12673 L closepath p .955 .786 .906 r
F P
0 g
s
.65629 .10383 m .65683 .10149 L .65423 .11512 L closepath p .955 .786 .906 r
F P
0 g
s
.65553 .11017 m .65423 .11512 L .65683 .10149 L closepath p .955 .786 .906 r
F P
0 g
s
.92046 .10383 m .92095 .10149 L .91861 .11512 L closepath p .955 .786 .906 r
F P
0 g
s
.91981 .11017 m .91861 .11512 L .92095 .10149 L closepath p .955 .786 .906 r
F P
0 g
s
.67978 .10065 m .68151 .10162 L .6783 .09787 L p .905 .473 .09 r
F P
0 g
s
.68151 .10162 m .67978 .10065 L .68383 .10658 L closepath p .966 .57 .101 r
F P
0 g
s
.68213 .10562 m .68383 .10658 L .68003 .10102 L p .958 .56 .099 r
F P
0 g
s
.68003 .10102 m .67978 .10065 L .68213 .10562 L p .958 .56 .099 r
F P
0 g
s
.68383 .10658 m .68213 .10562 L .6866 .11288 L closepath p .975 .589 .115 r
F P
0 g
s
.68488 .11188 m .6866 .11288 L .68213 .10562 L closepath p .974 .588 .114 r
F P
0 g
s
.68213 .10562 m .68491 .09421 L .68488 .11164 L p .104 .104 .64 r
F P
0 g
s
.68488 .11164 m .68488 .11188 L .68213 .10562 L p .104 .104 .64 r
F P
0 g
s
.68332 .09202 m .68611 .08041 L .68012 .09061 L closepath p .356 .73 1 r
F P
0 g
s
.67978 .10065 m .68252 .08917 L .68214 .10508 L p .107 .107 .667 r
F P
0 g
s
.68214 .10508 m .68213 .10562 L .67978 .10065 L p .107 .107 .667 r
F P
0 g
s
.6783 .09787 m .67784 .09734 L .67978 .10065 L p .905 .473 .09 r
F P
0 g
s
.24477 .1009 m .24762 .10187 L .24359 .09815 L p 1 .641 .274 r
F P
0 g
s
.24762 .10187 m .24477 .1009 L .24977 .10677 L closepath p 1 .728 .371 r
F P
0 g
s
.24707 .10636 m .24977 .10677 L .24477 .1009 L closepath p 1 .746 .399 r
F P
0 g
s
.24977 .10677 m .24707 .10636 L .2522 .11284 L closepath p 1 .757 .412 r
F P
0 g
s
.24937 .11183 m .2522 .11284 L .24707 .10636 L closepath p 1 .737 .379 r
F P
0 g
s
.24292 .09129 m .239 .07967 L .24079 .09391 L closepath p 1 1 .888 r
F P
0 g
s
.24477 .1009 m .24088 .08942 L .24861 .10983 L p .014 .014 .324 r
F P
0 g
s
.24861 .10983 m .24937 .11183 L .24477 .1009 L p .014 .014 .324 r
F P
0 g
s
.24359 .09815 m .24278 .09741 L .24477 .1009 L p 1 .641 .274 r
F P
0 g
s
.20952 .10197 m .20666 .1011 L .21635 .11543 L closepath p 1 .689 .333 r
F P
0 g
s
.64279 .10197 m .64104 .1011 L .65003 .11543 L closepath p .943 .525 .099 r
F P
0 g
s
.65333 .11581 m .65242 .11651 L .64279 .10197 L closepath p .955 .786 .906 r
F P
0 g
s
.65003 .11543 m .64279 .10197 L .65242 .11651 L closepath p .955 .786 .906 r
F P
0 g
s
.90574 .10115 m .9068 .10201 L .90056 .09114 L closepath p .809 .404 .052 r
F P
0 g
s
.9068 .10201 m .90574 .10115 L .91441 .11543 L closepath p .782 .362 .048 r
F P
0 g
s
.91776 .11581 m .9169 .11651 L .9068 .10201 L closepath p .955 .786 .906 r
F P
0 g
s
.91441 .11543 m .9068 .10201 L .9169 .11651 L closepath p .955 .786 .906 r
F P
0 g
s
.07762 .10205 m .07836 .10432 L .0766 .11196 L closepath p .955 .786 .906 r
F P
0 g
s
.78514 .11481 m .77743 .1058 L .78157 .10218 L closepath p .955 .786 .906 r
F P
0 g
s
.77721 .10341 m .78157 .10218 L .77743 .1058 L closepath p .955 .786 .906 r
F P
0 g
s
.70993 .05298 m .70679 .06497 L .70835 .04349 L closepath p .077 .077 .391 r
F P
0 g
s
.80129 .05298 m .79678 .06498 L .80013 .04233 L closepath p .087 .087 .386 r
F P
0 g
s
.79678 .06498 m .80129 .05298 L .79972 .07533 L closepath p .11 .11 .575 r
F P
0 g
s
.4767 .08463 m .47635 .07292 L .47381 .06553 L closepath p .022 .022 .186 r
F P
0 g
s
.47341 .05354 m .47381 .06553 L .47635 .07292 L closepath p .022 .022 .186 r
F P
0 g
s
.29271 .10588 m .29427 .10376 L .29197 .1028 L closepath p .155 .155 .693 r
F P
0 g
s
.29467 .10568 m .29427 .10376 L .29389 .10428 L p .152 .152 .671 r
F P
0 g
s
.29389 .10428 m .29271 .10588 L .29467 .10568 L p .152 .152 .671 r
F P
0 g
s
.29467 .10568 m .29526 .10688 L .29518 .10659 L p .035 .035 .436 r
F P
0 g
s
.29427 .10376 m .29467 .10568 L .29465 .10559 L p .733 .48 .011 r
F P
0 g
s
.29364 .10078 m .29197 .1028 L .29427 .10376 L p .16 .16 .661 r
F P
0 g
s
.29465 .10559 m .29154 .09427 L .29427 .10376 L p .733 .48 .011 r
F P
0 g
s
.29427 .10376 m .29379 .10059 L .29364 .10078 L p .16 .16 .661 r
F P
0 g
s
.29427 .10376 m .29113 .09233 L .29379 .10059 L closepath p .782 .554 .022 r
F P
0 g
s
.29154 .09427 m .29113 .09233 L .29137 .09322 L p .733 .48 .011 r
F P
0 g
s
.29137 .09322 m .29427 .10376 L .29154 .09427 L p .733 .48 .011 r
F P
0 g
s
.29214 .09549 m .29154 .09427 L .29182 .09529 L p .035 .035 .436 r
F P
0 g
s
.29182 .09529 m .29467 .10568 L .29214 .09549 L p .035 .035 .436 r
F P
0 g
s
.29065 .08911 m .29379 .10059 L .29113 .09233 L closepath p .782 .554 .022 r
F P
0 g
s
.29214 .09549 m .29526 .10688 L .30062 .09549 L closepath p .58 .961 1 r
F P
0 g
s
.29214 .09549 m .30008 .09589 L .29275 .09589 L p .14 .036 .036 r
F P
0 g
s
.29518 .10659 m .29214 .09549 L .29467 .10568 L p .035 .035 .436 r
F P
0 g
s
.29275 .09589 m .29158 .09589 L .29214 .09549 L p .14 .036 .036 r
F P
0 g
s
.29154 .09427 m .29214 .09549 L .29178 .09575 L p .846 .448 .061 r
F P
0 g
s
.29178 .09575 m .29158 .09589 L .29154 .09427 L p .846 .448 .061 r
F P
0 g
s
.28014 .1028 m .27725 .10123 L .28055 .10734 L closepath p 1 .929 .616 r
F P
0 g
s
.79257 .10921 m .79475 .10573 L .78942 .1028 L closepath p .398 .483 .928 r
F P
0 g
s
.79246 .10051 m .78942 .1028 L .79475 .10573 L closepath p .42 .482 .918 r
F P
0 g
s
.7269 .10588 m .72952 .10375 L .72604 .1028 L closepath p .301 .35 .825 r
F P
0 g
s
.72894 .1006 m .72604 .1028 L .72952 .10375 L closepath p .319 .33 .794 r
F P
0 g
s
.29197 .1028 m .29345 .10897 L .27909 .11028 L closepath p .955 .786 .906 r
F P
0 g
s
.28055 .10734 m .29197 .1028 L .27909 .11028 L closepath p .955 .786 .906 r
F P
0 g
s
.71343 .11028 m .71477 .10734 L .72604 .1028 L closepath p .955 .786 .906 r
F P
0 g
s
.72604 .1028 m .72775 .10897 L .71343 .11028 L closepath p .955 .786 .906 r
F P
0 g
s
.78942 .1028 m .79257 .10921 L .78514 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.59597 .10472 m .59854 .10249 L .59574 .10283 L closepath p .233 .241 .729 r
F P
0 g
s
.59832 .10066 m .59574 .10283 L .59854 .10249 L closepath p .233 .241 .729 r
F P
0 g
s
.59574 .10283 m .59597 .10472 L .5856 .10491 L closepath p .955 .786 .906 r
F P
0 g
s
.35428 .1093 m .35541 .10582 L .3513 .10287 L closepath p .207 .351 .857 r
F P
0 g
s
.35325 .10057 m .3513 .10287 L .35541 .10582 L closepath p .24 .355 .852 r
F P
0 g
s
.3513 .10287 m .35428 .1093 L .35068 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.40703 .10931 m .40828 .10583 L .40403 .10289 L closepath p .234 .369 .868 r
F P
0 g
s
.40612 .10061 m .40403 .10289 L .40828 .10583 L closepath p .265 .372 .862 r
F P
0 g
s
.40403 .10289 m .40703 .10931 L .40371 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.87472 .1068 m .87798 .10455 L .87401 .10322 L closepath p .381 .381 .824 r
F P
0 g
s
.87728 .10104 m .87401 .10322 L .87798 .10455 L closepath p .381 .381 .824 r
F P
0 g
s
.87173 .11204 m .87057 .11165 L .87401 .10322 L closepath p .955 .786 .906 r
F P
0 g
s
.87401 .10322 m .87472 .1068 L .87173 .11204 L closepath p .955 .786 .906 r
F P
0 g
s
.49135 .11195 m .49254 .10329 L .49229 .10574 L closepath p .955 .786 .906 r
F P
0 g
s
.49254 .10329 m .49135 .11195 L .49213 .10535 L closepath p .955 .786 .906 r
F P
0 g
s
.89733 .05411 m .89138 .06609 L .89561 .05033 L closepath p .14 .196 .753 r
F P
0 g
s
.89138 .06609 m .89733 .05411 L .89414 .07178 L closepath p .142 .222 .774 r
F P
0 g
s
.33631 .10156 m .33907 .10333 L .33587 .09728 L closepath p 1 .917 .588 r
F P
0 g
s
.33907 .10333 m .33631 .10156 L .3392 .1059 L closepath p 1 .943 .625 r
F P
0 g
s
.43782 .10346 m .43531 .10169 L .43798 .10582 L closepath p 1 .926 .581 r
F P
0 g
s
.75099 .10036 m .75233 .10351 L .75324 .09604 L closepath p 1 1 .892 r
F P
0 g
s
.75233 .10351 m .75099 .10036 L .74975 .1068 L closepath p 1 1 .993 r
F P
0 g
s
.74975 .1068 m .75233 .10351 L .75725 .11304 L closepath p .955 .786 .906 r
F P
0 g
s
.75744 .10383 m .75725 .11304 L .75233 .10351 L closepath p .955 .786 .906 r
F P
0 g
s
.31579 .10036 m .31817 .10352 L .31823 .09604 L closepath p 1 1 .956 r
F P
0 g
s
.31817 .10352 m .31579 .10036 L .31547 .1068 L closepath p 1 1 1 r
F P
0 g
s
.31547 .1068 m .31817 .10352 L .32033 .11757 L closepath p .955 .786 .906 r
F P
0 g
s
.32244 .10624 m .32033 .11757 L .31817 .10352 L closepath p .955 .786 .906 r
F P
0 g
s
.25842 .10676 m .25933 .10311 L .25514 .10372 L closepath p .469 .654 1 r
F P
0 g
s
.25622 .10022 m .25514 .10372 L .25933 .10311 L closepath p .469 .654 1 r
F P
0 g
s
.69278 .1068 m .69472 .10314 L .68941 .10378 L closepath p .563 .702 1 r
F P
0 g
s
.69152 .10027 m .68941 .10378 L .69472 .10314 L closepath p .563 .702 1 r
F P
0 g
s
.22201 .11017 m .2236 .10882 L .22333 .10395 L closepath p .111 .111 .272 r
F P
0 g
s
.22496 .10243 m .22333 .10395 L .2236 .10882 L closepath p .111 .111 .272 r
F P
0 g
s
.22333 .10395 m .22201 .11017 L .21635 .11543 L closepath p .955 .786 .906 r
F P
0 g
s
.91981 .11017 m .92317 .10882 L .92082 .10395 L closepath p .289 .191 .623 r
F P
0 g
s
.92421 .10243 m .92082 .10395 L .92317 .10882 L closepath p .29 .191 .623 r
F P
0 g
s
.65553 .11017 m .65822 .10882 L .65665 .10395 L closepath p .17 .17 .525 r
F P
0 g
s
.65938 .10243 m .65665 .10395 L .65822 .10882 L closepath p .17 .17 .525 r
F P
0 g
s
.52442 .11487 m .51835 .11 L .52044 .10403 L closepath p .955 .786 .906 r
F P
0 g
s
.51951 .10853 m .52044 .10403 L .51835 .11 L closepath p .955 .786 .906 r
F P
0 g
s
.51951 .10853 m .51835 .11 L .52044 .10403 L closepath p .955 .786 .906 r
F P
0 g
s
.52044 .10403 m .51813 .10215 L .51952 .10853 L closepath p 1 1 .754 r
F P
0 g
s
.5341 .10926 m .53653 .10762 L .53477 .10407 L closepath p .162 .162 .519 r
F P
0 g
s
.533 .11415 m .53271 .11452 L .53477 .10407 L closepath p .955 .786 .906 r
F P
0 g
s
.53477 .10407 m .5341 .10926 L .533 .11415 L closepath p .955 .786 .906 r
F P
0 g
s
.48231 .10086 m .48418 .10418 L .48567 .09633 L closepath p 1 1 .999 r
F P
0 g
s
.48418 .10418 m .48231 .10086 L .48138 .10728 L closepath p 1 1 1 r
F P
0 g
s
.48138 .10728 m .48418 .10418 L .49082 .11438 L closepath p .955 .786 .906 r
F P
0 g
s
.49213 .10535 m .49082 .11438 L .48418 .10418 L closepath p .955 .786 .906 r
F P
0 g
s
.93129 .10238 m .93257 .10426 L .93246 .0959 L closepath p 1 .989 .587 r
F P
0 g
s
.93257 .10426 m .93129 .10238 L .93386 .10989 L closepath p .944 .65 .126 r
F P
0 g
s
.93439 .10238 m .93566 .10426 L .93531 .09683 L closepath p 1 .98 .572 r
F P
0 g
s
.93566 .10426 m .93439 .10238 L .93669 .1089 L closepath p .948 .658 .134 r
F P
0 g
s
.07836 .10432 m .0791 .10657 L .07717 .11545 L closepath p .955 .786 .906 r
F P
0 g
s
.07717 .11545 m .0766 .11196 L .07836 .10432 L closepath p .955 .786 .906 r
F P
0 g
s
.45148 .10796 m .45315 .10473 L .45053 .10434 L closepath p .173 .244 .76 r
F P
0 g
s
.45246 .10228 m .45053 .10434 L .45315 .10473 L closepath p .172 .249 .767 r
F P
0 g
s
.07141 .10326 m .0748 .10468 L .0684 .09233 L closepath p 1 .847 .547 r
F P
0 g
s
.0748 .10468 m .07141 .10326 L .0766 .11197 L closepath p 1 .862 .566 r
F P
0 g
s
.59848 .10969 m .60017 .1062 L .59597 .10472 L closepath p .335 .448 .915 r
F P
0 g
s
.59854 .10249 m .59597 .10472 L .60017 .1062 L closepath p .353 .439 .9 r
F P
0 g
s
.59597 .10472 m .59692 .10762 L .58317 .11381 L closepath p .955 .786 .906 r
F P
0 g
s
.58317 .11381 m .5856 .10491 L .59597 .10472 L closepath p .955 .786 .906 r
F P
0 g
s
.06265 .05566 m .0692 .06762 L .0684 .07568 L closepath p .844 .534 .047 r
F P
0 g
s
.81943 .10467 m .82018 .10489 L .81093 .09531 L closepath p .661 .159 .038 r
F P
0 g
s
.82018 .10489 m .81943 .10467 L .82893 .11134 L closepath p .512 .025 .025 r
F P
0 g
s
.82893 .11134 m .82018 .10489 L .83441 .10757 L closepath p .955 .786 .906 r
F P
0 g
s
.07881 .10152 m .08106 .10519 L .07993 .09519 L closepath p 1 1 .761 r
F P
0 g
s
.08106 .10519 m .07881 .10152 L .07984 .1088 L closepath p 1 1 .864 r
F P
0 g
s
.49082 .11438 m .49213 .10535 L .49135 .11195 L closepath p .955 .786 .906 r
F P
0 g
s
.54047 .08291 m .54109 .07117 L .54308 .0685 L closepath p .784 .743 .211 r
F P
0 g
s
.54373 .05656 m .54308 .0685 L .54109 .07117 L closepath p .784 .743 .211 r
F P
0 g
s
.49135 .11195 m .4936 .11069 L .49229 .10574 L closepath p .154 .154 .477 r
F P
0 g
s
.49461 .10418 m .49229 .10574 L .4936 .11069 L closepath p .155 .155 .478 r
F P
0 g
s
.75725 .11304 m .76025 .11138 L .75801 .10574 L closepath p .23 .183 .617 r
F P
0 g
s
.76102 .10402 m .75801 .10574 L .76025 .11138 L closepath p .231 .184 .618 r
F P
0 g
s
.32664 .11268 m .32566 .10818 L .32488 .10932 L p 1 1 .981 r
F P
0 g
s
.32368 .10574 m .32269 .11303 L .32415 .11176 L p .135 .135 .412 r
F P
0 g
s
.32269 .11303 m .32349 .1162 L .32433 .11255 L p .124 .124 .314 r
F P
0 g
s
.3248 .11421 m .32697 .11471 L .32472 .11192 L p 1 .786 .392 r
F P
0 g
s
.32472 .11192 m .32434 .11145 L p .32448 .1123 L 1 .786 .392 r
F P
0 g
s
.3248 .11421 m .32472 .11192 L .32448 .1123 L 1 .786 .392 r
F
.32697 .11471 m .32664 .11268 L .32433 .11116 L p 1 .882 .541 r
F P
0 g
s
.32433 .11143 m .32697 .11471 L p .32433 .11116 L 1 .882 .541 r
F P
0 g
s
.32433 .11255 m .3246 .11137 L .32434 .1116 L p .124 .124 .314 r
F P
0 g
s
.32434 .1116 m .3241 .11181 L p .32433 .11255 L .124 .124 .314 r
F P
0 g
s
.3241 .11181 m .32269 .11303 L p .32433 .11255 L .124 .124 .314 r
F P
0 g
s
.32244 .10624 m .32368 .10574 L .32033 .11757 L closepath p .955 .786 .906 r
F P
0 g
s
.32269 .11303 m .32033 .11757 L .32368 .10574 L closepath p .955 .786 .906 r
F P
0 g
s
.32115 .0995 m .32384 .11083 L .32336 .10285 L p .987 1 .8 r
F P
0 g
s
.3256 .10404 m .32368 .10574 L .32448 .11065 L p .136 .136 .413 r
F P
0 g
s
.32433 .11116 m .32664 .11268 L p .32488 .10932 L 1 1 .981 r
F P
0 g
s
.32448 .11065 m .3246 .11137 L .32488 .10931 L p .136 .136 .413 r
F P
0 g
s
.32501 .10833 m .3256 .10404 L p .32484 .10854 L .136 .136 .413 r
F P
0 g
s
.32488 .10931 m .32501 .10833 L p .32467 .10982 L .136 .136 .413 r
F P
0 g
s
.32467 .10982 m .32448 .11065 L .32488 .10931 L .136 .136 .413 r
F
.32484 .10854 m .32448 .11065 L .32501 .10833 L F
.32297 .10118 m .32346 .10172 L p .32488 .10931 L .925 .832 .326 r
F P
0 g
s
.32346 .10172 m .3256 .10404 L .32501 .10833 L p .925 .832 .326 r
F P
0 g
s
.32501 .10833 m .32488 .10931 L p .32346 .10172 L .925 .832 .326 r
F P
0 g
s
.32415 .11176 m .32432 .11121 L .32368 .10574 L .135 .135 .412 r
F
.32432 .11121 m .3243 .11087 L .32368 .10574 L F
.32435 .10986 m .32368 .10574 L p .32433 .11116 L F P
0 g
s
.32294 .09261 m .3256 .10404 L .32297 .10118 L p .925 .832 .326 r
F P
0 g
s
.32221 .09796 m .32115 .0995 L p .32336 .10285 L .987 1 .8 r
F P
0 g
s
.32297 .10118 m .32192 .10005 L .32216 .09831 L p .925 .832 .326 r
F P
0 g
s
.32226 .09816 m .32297 .10118 L .32216 .09831 L .925 .832 .326 r
F
.32221 .09796 m .32294 .09261 L p .32295 .0961 L F P
0 g
s
.32295 .0961 m .32297 .10118 L .32221 .09796 L .925 .832 .326 r
F
.323 .09681 m .32336 .10285 L p .32488 .10931 L .987 1 .8 r
F P
0 g
s
.32554 .10769 m .323 .09681 L p .32488 .10931 L .987 1 .8 r
F P
0 g
s
.3256 .10404 m .32294 .09261 L .32764 .09307 L closepath p .943 .874 .386 r
F P
0 g
s
.32675 .09592 m .323 .09681 L .32554 .10769 L p .725 1 1 r
F P
0 g
s
.32336 .10285 m .323 .09681 L .32221 .09796 L p .987 1 .8 r
F P
0 g
s
.31908 .09793 m .32294 .09261 L .32221 .09796 L p 1 1 .753 r
F P
0 g
s
.32216 .09831 m .32192 .10005 L p .32158 .09905 L 1 1 .753 r
F P
0 g
s
.32184 .09882 m .32149 .09916 L p .32191 .09871 L .022 .022 .022 r
F P
0 g
s
.32149 .09916 m .32115 .0995 L .32158 .09905 L p .022 .022 .022 r
F P
0 g
s
.32158 .09905 m .32191 .09871 L p .32149 .09916 L .022 .022 .022 r
F P
0 g
s
.32336 .10285 m .32384 .11083 L .32488 .10931 L p .987 1 .8 r
F P
0 g
s
.32432 .11121 m .32433 .11116 L .3243 .11087 L .135 .135 .412 r
F
.32448 .11065 m .32435 .10986 L p .32433 .11116 L F P
0 g
s
.32488 .10932 m .32384 .11083 L .32433 .11115 L p 1 1 .981 r
F P
0 g
s
.32433 .11115 m .32448 .11065 L .32488 .10932 L 1 1 .981 r
F
.32434 .11145 m .32433 .11143 L p .32434 .1115 L 1 .786 .392 r
F P
0 g
s
.32415 .11176 m .3246 .11137 L .32448 .11065 L p .135 .135 .412 r
F P
0 g
s
.3246 .11137 m .32192 .10005 L .32297 .10118 L p .925 .832 .326 r
F P
0 g
s
.32433 .11115 m .32433 .11116 L p .32448 .11065 L 1 1 .981 r
F P
0 g
s
.32384 .11083 m .32115 .0995 L .32302 .10706 L p .008 .008 .221 r
F P
0 g
s
.32192 .10005 m .3246 .11137 L .32368 .10982 L p .957 .906 .435 r
F P
0 g
s
.32384 .11083 m .32302 .10706 L .32425 .11224 L .008 .008 .221 r
F
.32192 .10005 m .32368 .10982 L .32173 .1016 L .957 .906 .435 r
F
.32433 .11143 m .32384 .11083 L .3244 .11282 L p 1 .786 .392 r
F P
0 g
s
.32433 .11116 m .32384 .11083 L .32433 .11143 L p 1 .882 .541 r
F P
0 g
s
.32368 .10982 m .3246 .11137 L .32433 .11255 L p .957 .906 .435 r
F P
0 g
s
.3244 .11282 m .3248 .11421 L p .32433 .11143 L 1 .786 .392 r
F P
0 g
s
.32212 .10293 m .3248 .11421 L .32302 .10706 L p .008 .008 .221 r
F P
0 g
s
.32302 .10706 m .32115 .0995 L .32165 .10125 L p .008 .008 .221 r
F P
0 g
s
.32165 .10125 m .32212 .10293 L p .32302 .10706 L .008 .008 .221 r
F P
0 g
s
.32165 .10125 m .32192 .10005 L p .32173 .1016 L .957 .906 .435 r
F P
0 g
s
.32173 .1016 m .32123 .10568 L .32165 .10125 L .957 .906 .435 r
F
.32368 .10982 m .32123 .10568 L p .32173 .1016 L F P
0 g
s
.32123 .10568 m .3208 .10494 L .32165 .10125 L p .957 .906 .435 r
F P
0 g
s
.32115 .0995 m .32158 .1001 L .32163 .10086 L .168 .231 .756 r
F
.32155 .09978 m .32192 .10005 L .32159 .10021 L p 1 1 .769 r
F P
0 g
s
.32158 .1001 m .32159 .10011 L .32159 .10021 L .168 .231 .756 r
F
.32158 .1001 m .32159 .10021 L .32163 .10086 L F
.32165 .10125 m .32115 .0995 L p .32167 .10099 L F P
0 g
s
.31846 .1017 m .31908 .09793 L .32155 .09978 L p 1 1 .769 r
F P
0 g
s
.32159 .10021 m .31846 .1017 L p .32155 .09978 L 1 1 .769 r
F P
0 g
s
.3208 .10494 m .31846 .1017 L .32159 .10021 L p 1 1 .789 r
F P
0 g
s
.32159 .10021 m .32192 .10005 L .32165 .10125 L p 1 1 .789 r
F P
0 g
s
.32165 .10125 m .3208 .10494 L p .32159 .10021 L 1 1 .789 r
F P
0 g
s
.32115 .0995 m .32149 .09916 L p .32159 .10011 L .168 .231 .756 r
F P
0 g
s
.32163 .10086 m .32159 .10011 L .32167 .10099 L .168 .231 .756 r
F
.32172 .10151 m .32165 .10125 L p .32167 .10099 L F P
0 g
s
.3231 .09757 m .32212 .10293 L .32172 .10151 L p .168 .231 .756 r
F P
0 g
s
.32192 .10005 m .32155 .09978 L p .32158 .09905 L 1 1 .753 r
F P
0 g
s
.32155 .09978 m .3215 .09919 L .32158 .09905 L 1 1 .753 r
F
.32149 .09916 m .3231 .09757 L p .32172 .10151 L .168 .231 .756 r
F P
0 g
s
.32155 .09978 m .32128 .09957 L p .3215 .09919 L 1 1 .753 r
F P
0 g
s
.32221 .09796 m .32216 .09831 L p .32158 .09905 L 1 1 .753 r
F P
0 g
s
.32158 .09905 m .32115 .0995 L .32221 .09796 L p .058 .058 .058 r
F P
0 g
s
.32128 .09957 m .31908 .09793 L p .32221 .09796 L 1 1 .753 r
F P
0 g
s
.32191 .09871 m .32158 .09905 L p .32221 .09796 L .058 .058 .058 r
F P
0 g
s
.323 .09681 m .3239 .09663 L .32191 .09871 L p .058 .058 .058 r
F P
0 g
s
.32216 .09831 m .32221 .09796 L p .32226 .09816 L .925 .832 .326 r
F P
0 g
s
.32221 .09796 m .323 .09681 L p .32191 .09871 L .058 .058 .058 r
F P
0 g
s
.3239 .09663 m .3231 .09757 L .32184 .09882 L p .022 .022 .022 r
F P
0 g
s
.32191 .09871 m .3239 .09663 L p .32184 .09882 L .022 .022 .022 r
F P
0 g
s
.32488 .10931 m .32566 .10818 L .32554 .10769 L p .987 1 .8 r
F P
0 g
s
.32554 .10769 m .32566 .10818 L .32572 .10748 L p .725 1 1 r
F P
0 g
s
.32572 .10748 m .32675 .09592 L p .32554 .10769 L .725 1 1 r
F P
0 g
s
.32488 .10931 m .3246 .11137 L p .32297 .10118 L .925 .832 .326 r
F P
0 g
s
.32433 .11255 m .32384 .11083 L p .32461 .11345 L .008 .008 .221 r
F P
0 g
s
.3208 .10494 m .32368 .10982 L p .32433 .11255 L .957 .906 .435 r
F P
0 g
s
.32302 .10706 m .32461 .11345 L p .32425 .11224 L .008 .008 .221 r
F P
0 g
s
.32461 .11345 m .3248 .11421 L .32433 .11255 L p .008 .008 .221 r
F P
0 g
s
.32433 .11255 m .32349 .1162 L .3231 .11457 L p .957 .906 .435 r
F P
0 g
s
.3231 .11457 m .3208 .10494 L p .32433 .11255 L .957 .906 .435 r
F P
0 g
s
.3248 .11421 m .32434 .11145 L .32434 .1115 L 1 .786 .392 r
F
.78514 .11481 m .76991 .11134 L .77743 .1058 L closepath p .955 .786 .906 r
F P
0 g
s
.7756 .10992 m .77743 .1058 L .76991 .11134 L closepath p .955 .786 .906 r
F P
0 g
s
.7756 .10992 m .76991 .11134 L .77743 .1058 L closepath p .955 .786 .906 r
F P
0 g
s
.77743 .1058 m .77576 .10393 L .7756 .10992 L closepath p 1 1 .827 r
F P
0 g
s
.43547 .10395 m .43798 .10582 L .43531 .10169 L closepath p 1 .925 .579 r
F P
0 g
s
.43798 .10582 m .43547 .10395 L .43721 .10898 L closepath p 1 1 .794 r
F P
0 g
s
.45148 .10796 m .45294 .10932 L .43798 .10582 L closepath p .955 .786 .906 r
F P
0 g
s
.43721 .10898 m .43798 .10582 L .45294 .10932 L closepath p .955 .786 .906 r
F P
0 g
s
.38941 .10401 m .39204 .10588 L .38907 .09641 L closepath p 1 .942 .613 r
F P
0 g
s
.39204 .10588 m .38941 .10401 L .39125 .10901 L closepath p 1 1 .81 r
F P
0 g
s
.29345 .10897 m .29467 .10568 L .29271 .10588 L closepath p .152 .152 .685 r
F P
0 g
s
.72775 .10897 m .73001 .10567 L .7269 .10588 L closepath p .294 .345 .822 r
F P
0 g
s
.72952 .10375 m .7269 .10588 L .73001 .10567 L closepath p .294 .337 .813 r
F P
0 g
s
.33644 .10403 m .3392 .1059 L .33631 .10156 L closepath p 1 .942 .624 r
F P
0 g
s
.3392 .1059 m .33644 .10403 L .33841 .10902 L closepath p 1 1 .818 r
F P
0 g
s
.33841 .10902 m .3392 .1059 L .35068 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.16284 .10801 m .16337 .11384 L .1548 .10632 L closepath p .955 .786 .906 r
F P
0 g
s
.15173 .11652 m .1548 .10632 L .16337 .11384 L closepath p .955 .786 .906 r
F P
0 g
s
.07984 .1088 m .07717 .11545 L .0791 .10657 L closepath p .955 .786 .906 r
F P
0 g
s
.8648 .10665 m .86369 .10355 L .85915 .11144 L closepath p 1 1 1 r
F P
0 g
s
.85915 .11144 m .8648 .10665 L .87057 .11165 L closepath p .955 .786 .906 r
F P
0 g
s
.57984 .05749 m .57864 .06942 L .57746 .03731 L closepath p .041 .041 .229 r
F P
0 g
s
.2638 .10996 m .26406 .10599 L .25842 .10676 L closepath p .578 .776 1 r
F P
0 g
s
.25933 .10311 m .25842 .10676 L .26406 .10599 L closepath p .579 .773 1 r
F P
0 g
s
.2638 .10996 m .2522 .11284 L .25842 .10676 L closepath p .955 .786 .906 r
F P
0 g
s
.87472 .1068 m .87522 .1099 L .87289 .11242 L closepath p .955 .786 .906 r
F P
0 g
s
.87542 .11227 m .87289 .11242 L .87522 .1099 L closepath p .955 .786 .906 r
F P
0 g
s
.87542 .11227 m .87872 .11041 L .87472 .1068 L closepath p .358 .34 .788 r
F P
0 g
s
.87798 .10455 m .87472 .1068 L .87872 .11041 L closepath p .358 .34 .787 r
F P
0 g
s
.74868 .10331 m .74975 .1068 L .75099 .10036 L closepath p 1 1 .993 r
F P
0 g
s
.74975 .1068 m .74868 .10331 L .74481 .10997 L closepath p 1 1 1 r
F P
0 g
s
.69826 .10997 m .69954 .10599 L .69278 .1068 L closepath p .636 .798 1 r
F P
0 g
s
.69472 .10314 m .69278 .1068 L .69954 .10599 L closepath p .638 .796 1 r
F P
0 g
s
.74975 .1068 m .75725 .11304 L .74481 .10997 L closepath p .955 .786 .906 r
F P
0 g
s
.6866 .11288 m .69278 .1068 L .69546 .11196 L closepath p .955 .786 .906 r
F P
0 g
s
.69826 .10997 m .69546 .11196 L .69278 .1068 L closepath p .955 .786 .906 r
F P
0 g
s
.87289 .11242 m .87173 .11204 L .87472 .1068 L closepath p .955 .786 .906 r
F P
0 g
s
.31337 .10331 m .31547 .1068 L .31579 .10036 L closepath p 1 1 1 r
F P
0 g
s
.31547 .1068 m .31337 .10331 L .31037 .10998 L closepath p 1 1 1 r
F P
0 g
s
.31547 .1068 m .32033 .11757 L .31362 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.31547 .1068 m .31362 .11482 L .31037 .10998 L closepath p .955 .786 .906 r
F P
0 g
s
.09355 .11622 m .09486 .11466 L .09536 .10693 L closepath p .089 .089 .199 r
F P
0 g
s
.09667 .1054 m .09536 .10693 L .09486 .11466 L closepath p .09 .09 .199 r
F P
0 g
s
.09085 .12757 m .09049 .128 L .09536 .10693 L closepath p .955 .786 .906 r
F P
0 g
s
.09536 .10693 m .09355 .11622 L .09085 .12757 L closepath p .955 .786 .906 r
F P
0 g
s
.09536 .10693 m .09085 .12757 L .09355 .11622 L closepath p .955 .786 .906 r
F P
0 g
s
.78952 .0844 m .79392 .07268 L .78702 .06991 L closepath p .094 .094 .453 r
F P
0 g
s
.79139 .05799 m .78702 .06991 L .79392 .07268 L closepath p .094 .094 .453 r
F P
0 g
s
.59519 .07438 m .59664 .06252 L .59455 .07 L closepath p .049 .049 .278 r
F P
0 g
s
.59599 .05808 m .59455 .07 L .59664 .06252 L closepath p .049 .049 .278 r
F P
0 g
s
.8187 .079 m .82355 .06721 L .81524 .07004 L closepath p .123 .123 .672 r
F P
0 g
s
.82003 .05812 m .81524 .07004 L .82355 .06721 L closepath p .123 .123 .672 r
F P
0 g
s
.47968 .10378 m .48138 .10728 L .48231 .10086 L closepath p 1 1 1 r
F P
0 g
s
.48138 .10728 m .47968 .10378 L .47693 .11046 L closepath p 1 1 1 r
F P
0 g
s
.48138 .10728 m .49082 .11438 L .47693 .11046 L closepath p .955 .786 .906 r
F P
0 g
s
.27764 .10547 m .28055 .10734 L .27725 .10123 L closepath p 1 .927 .613 r
F P
0 g
s
.28055 .10734 m .27764 .10547 L .27909 .11028 L closepath p 1 1 .927 r
F P
0 g
s
.71477 .10734 m .71295 .10547 L .71343 .11028 L closepath p 1 1 .849 r
F P
0 g
s
.84203 .11167 m .84346 .10765 L .83441 .10757 L closepath p .659 .815 1 r
F P
0 g
s
.83646 .10385 m .83441 .10757 L .84346 .10765 L closepath p .66 .814 1 r
F P
0 g
s
.83441 .10757 m .84203 .11167 L .82893 .11134 L closepath p .955 .786 .906 r
F P
0 g
s
.59692 .10762 m .59848 .10969 L .59943 .11487 L closepath p .955 .786 .906 r
F P
0 g
s
.59943 .11487 m .58317 .11381 L .59692 .10762 L closepath p .955 .786 .906 r
F P
0 g
s
.58456 .10772 m .58216 .10356 L .58424 .10772 L closepath p .839 1 1 r
F P
0 g
s
.58588 .10356 m .58424 .10772 L .58216 .10356 L closepath p .839 1 1 r
F P
0 g
s
.45439 .11068 m .45461 .10655 L .45148 .10796 L closepath p .448 .643 1 r
F P
0 g
s
.45315 .10473 m .45148 .10796 L .45461 .10655 L closepath p .426 .578 1 r
F P
0 g
s
.16337 .11384 m .16487 .11163 L .16284 .10801 L closepath p .138 .138 .553 r
F P
0 g
s
.16511 .11394 m .16487 .11163 L .16443 .11228 L p .139 .139 .566 r
F P
0 g
s
.16443 .11228 m .16337 .11384 L .16511 .11394 L p .139 .139 .566 r
F P
0 g
s
.16511 .11394 m .16362 .1162 L .16702 .11851 L closepath p .166 .287 .812 r
F P
0 g
s
.16511 .11394 m .16702 .11851 L .16651 .1174 L p .674 .329 .006 r
F P
0 g
s
.16362 .1162 m .16511 .11394 L .16337 .11384 L closepath p .139 .139 .566 r
F P
0 g
s
.16651 .1174 m .16194 .10729 L .16511 .11394 L p .674 .329 .006 r
F P
0 g
s
.16487 .11163 m .16511 .11394 L .1651 .11392 L p .912 .689 .148 r
F P
0 g
s
.1643 .10591 m .16284 .10801 L .16487 .11163 L p .138 .138 .554 r
F P
0 g
s
.1651 .11392 m .16 .10265 L .16487 .11163 L p .912 .689 .148 r
F P
0 g
s
.16487 .11163 m .16434 .10584 L .1643 .10591 L p .138 .138 .554 r
F P
0 g
s
.50881 .05891 m .50868 .07082 L .49682 .02284 L closepath p .06 .06 .437 r
F P
0 g
s
.51765 .1055 m .51952 .10853 L .51813 .10215 L closepath p 1 1 .679 r
F P
0 g
s
.51952 .10853 m .51765 .1055 L .51672 .11096 L closepath p 1 1 1 r
F P
0 g
s
.07984 .1088 m .0799 .10868 L .07717 .11545 L closepath p .955 .786 .906 r
F P
0 g
s
.07903 .10354 m .07984 .1088 L .07881 .10152 L closepath p .784 .578 .022 r
F P
0 g
s
.4518 .09852 m .45106 .08701 L .44791 .07161 L closepath p .015 .015 .154 r
F P
0 g
s
.45241 .05971 m .45313 .07161 L .44712 .05971 L closepath p .598 .968 1 r
F P
0 g
s
.93566 .10787 m .93669 .1089 L .93439 .10238 L closepath p .931 .639 .109 r
F P
0 g
s
.93669 .1089 m .93566 .10787 L .93946 .11281 L closepath p .664 .212 .024 r
F P
0 g
s
.96218 .10994 m .959 .11466 L .95257 .11441 L closepath p .955 .786 .906 r
F P
0 g
s
.9568 .11281 m .9597 .11254 L .95929 .1089 L closepath p .189 .163 .315 r
F P
0 g
s
.96262 .10787 m .95929 .1089 L .9597 .11254 L closepath p .181 .162 .314 r
F P
0 g
s
.29574 .11103 m .29526 .10688 L .29345 .10897 L closepath p .388 .621 1 r
F P
0 g
s
.29467 .10568 m .29345 .10897 L .29526 .10688 L closepath p .224 .37 .872 r
F P
0 g
s
.73022 .11103 m .73076 .10688 L .72775 .10897 L closepath p .519 .697 1 r
F P
0 g
s
.73001 .10567 m .72775 .10897 L .73076 .10688 L closepath p .445 .547 .974 r
F P
0 g
s
.29635 .11255 m .27909 .11028 L .29345 .10897 L closepath p .955 .786 .906 r
F P
0 g
s
.29345 .10897 m .29574 .11103 L .29635 .11255 L closepath p .955 .786 .906 r
F P
0 g
s
.70841 .11196 m .72775 .10897 L .73022 .11103 L closepath p .955 .786 .906 r
F P
0 g
s
.7117 .11102 m .71343 .11028 L .72775 .10897 L closepath p .955 .786 .906 r
F P
0 g
s
.70841 .11196 m .7117 .11102 L .72775 .10897 L closepath p .955 .786 .906 r
F P
0 g
s
.43518 .10591 m .43721 .10898 L .43547 .10395 L closepath p 1 1 .705 r
F P
0 g
s
.43721 .10898 m .43518 .10591 L .43494 .11075 L closepath p 1 1 1 r
F P
0 g
s
.45294 .10932 m .45439 .11068 L .43721 .10898 L closepath p .955 .786 .906 r
F P
0 g
s
.43494 .11075 m .43721 .10898 L .45439 .11068 L closepath p .955 .786 .906 r
F P
0 g
s
.3891 .10594 m .39125 .10901 L .38941 .10401 L closepath p 1 1 .73 r
F P
0 g
s
.39125 .10901 m .3891 .10594 L .38891 .11083 L closepath p 1 1 1 r
F P
0 g
s
.40371 .11489 m .38891 .11083 L .39125 .10901 L closepath p .955 .786 .906 r
F P
0 g
s
.39027 .11008 m .39125 .10901 L .38891 .11083 L closepath p .955 .786 .906 r
F P
0 g
s
.39027 .11008 m .38891 .11083 L .39125 .10901 L closepath p .955 .786 .906 r
F P
0 g
s
.90013 .05988 m .89414 .07178 L .89733 .05411 L closepath p .142 .222 .774 r
F P
0 g
s
.89414 .07178 m .90013 .05988 L .89793 .07938 L closepath p .143 .232 .782 r
F P
0 g
s
.33613 .10596 m .33841 .10902 L .33644 .10403 L closepath p 1 1 .742 r
F P
0 g
s
.33841 .10902 m .33613 .10596 L .33612 .11083 L closepath p 1 1 1 r
F P
0 g
s
.35068 .11489 m .33612 .11083 L .33841 .10902 L closepath p .955 .786 .906 r
F P
0 g
s
.33746 .11009 m .33841 .10902 L .33612 .11083 L closepath p .955 .786 .906 r
F P
0 g
s
.33746 .11009 m .33612 .11083 L .33841 .10902 L closepath p .955 .786 .906 r
F P
0 g
s
.5607 .05995 m .55979 .07185 L .55281 .04412 L closepath p .092 .092 .629 r
F P
0 g
s
.55979 .07185 m .5607 .05995 L .56961 .09082 L closepath p .094 .094 .646 r
F P
0 g
s
.43023 .06001 m .43127 .0719 L .42878 .05 L closepath p .011 .011 .139 r
F P
0 g
s
.45763 .06005 m .45826 .07194 L .45241 .05971 L closepath p .558 .935 1 r
F P
0 g
s
.79849 .11134 m .79919 .10719 L .79257 .10921 L closepath p .705 .876 1 r
F P
0 g
s
.79475 .10573 m .79257 .10921 L .79919 .10719 L closepath p .704 .886 1 r
F P
0 g
s
.794 .11312 m .78514 .11481 L .79257 .10921 L closepath p .955 .786 .906 r
F P
0 g
s
.79257 .10921 m .79849 .11134 L .794 .11312 L closepath p .955 .786 .906 r
F P
0 g
s
.34688 .06012 m .34918 .07201 L .34836 .07088 L closepath p .74 .526 .011 r
F P
0 g
s
.34918 .07201 m .34688 .06012 L .38541 .07201 L closepath p .598 .968 1 r
F P
0 g
s
.38367 .06012 m .38541 .07201 L .34688 .06012 L closepath p .598 .968 1 r
F P
0 g
s
.5333 .11377 m .53519 .11327 L .5341 .10926 L closepath p .141 .141 .412 r
F P
0 g
s
.53653 .10762 m .5341 .10926 L .53519 .11327 L closepath p .152 .152 .423 r
F P
0 g
s
.5333 .11377 m .533 .11415 L .5341 .10926 L closepath p .955 .786 .906 r
F P
0 g
s
.36009 .11144 m .35977 .10729 L .35428 .1093 L closepath p .673 .868 1 r
F P
0 g
s
.35541 .10582 m .35428 .1093 L .35977 .10729 L closepath p .672 .878 1 r
F P
0 g
s
.35757 .11321 m .35068 .11489 L .35428 .1093 L closepath p .955 .786 .906 r
F P
0 g
s
.35428 .1093 m .36009 .11144 L .35757 .11321 L closepath p .955 .786 .906 r
F P
0 g
s
.41295 .11145 m .41275 .1073 L .40703 .10931 L closepath p .68 .872 1 r
F P
0 g
s
.40828 .10583 m .40703 .10931 L .41275 .1073 L closepath p .679 .883 1 r
F P
0 g
s
.41051 .11321 m .40371 .11489 L .40703 .10931 L closepath p .955 .786 .906 r
F P
0 g
s
.40703 .10931 m .41295 .11145 L .41051 .11321 L closepath p .955 .786 .906 r
F P
0 g
s
.9535 .11136 m .95617 .11107 L .95424 .10954 L closepath p .167 .167 .414 r
F P
0 g
s
.77499 .06044 m .77087 .07233 L .77335 .05076 L closepath p .09 .09 .437 r
F P
0 g
s
.59848 .10969 m .60339 .11134 L .60326 .11504 L closepath p .955 .786 .906 r
F P
0 g
s
.60326 .11504 m .59943 .11487 L .59848 .10969 L closepath p .955 .786 .906 r
F P
0 g
s
.60339 .11134 m .60364 .10719 L .59848 .10969 L closepath p .703 .884 1 r
F P
0 g
s
.60017 .1062 m .59848 .10969 L .60364 .10719 L closepath p .705 .901 1 r
F P
0 g
s
.93284 .1089 m .93386 .10989 L .93129 .10238 L closepath p .928 .632 .101 r
F P
0 g
s
.93386 .10989 m .93284 .1089 L .93737 .11462 L closepath p .648 .189 .022 r
F P
0 g
s
.77478 .10648 m .7756 .10992 L .77576 .10393 L closepath p 1 1 .765 r
F P
0 g
s
.7756 .10992 m .77478 .10648 L .76991 .11134 L closepath p .902 1 1 r
F P
0 g
s
.959 .11466 m .96186 .11444 L .96218 .10994 L closepath p .184 .161 .294 r
F P
0 g
s
.96551 .10894 m .96218 .10994 L .96186 .11444 L closepath p .174 .16 .291 r
F P
0 g
s
.2693 .11103 m .26876 .10688 L .2638 .10996 L closepath p .752 .937 1 r
F P
0 g
s
.26406 .10599 m .2638 .10996 L .26876 .10688 L closepath p .752 .938 1 r
F P
0 g
s
.2693 .11103 m .2522 .11284 L .2638 .10996 L closepath p .955 .786 .906 r
F P
0 g
s
.74452 .10603 m .74481 .10997 L .74868 .10331 L closepath p 1 1 1 r
F P
0 g
s
.74481 .10997 m .74452 .10603 L .73856 .11103 L closepath p .892 1 1 r
F P
0 g
s
.69546 .11196 m .69826 .10997 L .70432 .11104 L closepath p .955 .786 .906 r
F P
0 g
s
.70379 .11103 m .70432 .11104 L .69826 .10997 L closepath p .955 .786 .906 r
F P
0 g
s
.70379 .11103 m .69826 .10997 L .70432 .11104 L closepath p .955 .786 .906 r
F P
0 g
s
.70379 .11103 m .70427 .10688 L .69826 .10997 L closepath p .766 .938 1 r
F P
0 g
s
.69954 .10599 m .69826 .10997 L .70427 .10688 L closepath p .766 .939 1 r
F P
0 g
s
.73856 .11103 m .74481 .10997 L .75725 .11304 L closepath p .955 .786 .906 r
F P
0 g
s
.30907 .10603 m .31037 .10998 L .31337 .10331 L closepath p 1 1 1 r
F P
0 g
s
.31037 .10998 m .30907 .10603 L .30407 .11103 L closepath p .9 1 1 r
F P
0 g
s
.30407 .11103 m .31037 .10998 L .31362 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.52442 .11487 m .52236 .11481 L .51835 .11 L closepath p .955 .786 .906 r
F P
0 g
s
.51672 .11096 m .51835 .11 L .52236 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.22057 .11512 m .2221 .11398 L .22201 .11017 L closepath p .101 .101 .188 r
F P
0 g
s
.2236 .10882 m .22201 .11017 L .2221 .11398 L closepath p .101 .101 .188 r
F P
0 g
s
.91861 .11512 m .92192 .11398 L .91981 .11017 L closepath p .267 .187 .569 r
F P
0 g
s
.92317 .10882 m .91981 .11017 L .92192 .11398 L closepath p .267 .187 .569 r
F P
0 g
s
.65423 .11512 m .65687 .11398 L .65553 .11017 L closepath p .163 .163 .461 r
F P
0 g
s
.65822 .10882 m .65553 .11017 L .65687 .11398 L closepath p .163 .163 .461 r
F P
0 g
s
.21757 .11597 m .21635 .11543 L .22201 .11017 L closepath p .955 .786 .906 r
F P
0 g
s
.22201 .11017 m .22057 .11512 L .21757 .11597 L closepath p .955 .786 .906 r
F P
0 g
s
.46142 .06108 m .462 .07296 L .45763 .06005 L closepath p .422 .815 1 r
F P
0 g
s
.89721 .06113 m .89126 .07301 L .88906 .04521 L closepath p .143 .239 .788 r
F P
0 g
s
.90399 .06759 m .89793 .07938 L .90013 .05988 L closepath p .143 .232 .782 r
F P
0 g
s
.89793 .07938 m .90399 .06759 L .90191 .08722 L closepath p .143 .238 .788 r
F P
0 g
s
.89126 .07301 m .89721 .06113 L .90056 .09114 L closepath p .143 .241 .79 r
F P
0 g
s
.18892 .06114 m .19358 .07301 L .18153 .04521 L closepath p .012 .012 .342 r
F P
0 g
s
.19358 .07301 m .18892 .06114 L .20203 .09115 L closepath p .013 .013 .345 r
F P
0 g
s
.62903 .06114 m .62709 .07302 L .62115 .04521 L closepath p .103 .103 .663 r
F P
0 g
s
.62709 .07302 m .62903 .06114 L .6361 .09115 L closepath p .103 .103 .665 r
F P
0 g
s
.2771 .10684 m .27909 .11028 L .27764 .10547 L closepath p 1 1 .875 r
F P
0 g
s
.31362 .11482 m .27208 .11115 L .27732 .11102 L closepath p .955 .786 .906 r
F P
0 g
s
.27485 .11127 m .27732 .11102 L .27208 .11115 L closepath p .955 .786 .906 r
F P
0 g
s
.27485 .11127 m .27208 .11115 L .27732 .11102 L closepath p .955 .786 .906 r
F P
0 g
s
.27909 .11028 m .2771 .10684 L .27485 .11127 L closepath p .91 1 1 r
F P
0 g
s
.27909 .11028 m .31362 .11482 L .27732 .11102 L closepath p .955 .786 .906 r
F P
0 g
s
.71247 .10684 m .71343 .11028 L .71295 .10547 L closepath p 1 1 .755 r
F P
0 g
s
.71343 .11028 m .71247 .10684 L .70927 .11127 L closepath p .899 1 1 r
F P
0 g
s
.47595 .10654 m .47693 .11046 L .47968 .10378 L closepath p 1 1 1 r
F P
0 g
s
.47693 .11046 m .47595 .10654 L .47233 .11153 L closepath p .919 1 1 r
F P
0 g
s
.47465 .11126 m .47693 .11046 L .49082 .11438 L closepath p .955 .786 .906 r
F P
0 g
s
.46209 .11131 m .46214 .10717 L .45439 .11068 L closepath p .805 .976 1 r
F P
0 g
s
.45461 .10655 m .45439 .11068 L .46214 .10717 L closepath p .805 .976 1 r
F P
0 g
s
.45439 .11068 m .46209 .11131 L .43494 .11075 L closepath p .955 .786 .906 r
F P
0 g
s
.43396 .10675 m .43494 .11075 L .43518 .10591 L closepath p 1 1 1 r
F P
0 g
s
.43494 .11075 m .43396 .10675 L .43282 .11117 L closepath p .911 1 1 r
F P
0 g
s
.43494 .11075 m .49082 .11438 L .43282 .11117 L closepath p .955 .786 .906 r
F P
0 g
s
.46288 .11256 m .43494 .11075 L .46209 .11131 L closepath p .955 .786 .906 r
F P
0 g
s
.3878 .10683 m .38891 .11083 L .3891 .10594 L closepath p 1 1 1 r
F P
0 g
s
.38891 .11083 m .3878 .10683 L .38651 .11129 L closepath p .91 1 1 r
F P
0 g
s
.40371 .11489 m .39158 .11379 L .38891 .11083 L closepath p .955 .786 .906 r
F P
0 g
s
.38651 .11129 m .38891 .11083 L .39158 .11379 L closepath p .955 .786 .906 r
F P
0 g
s
.33489 .10684 m .33612 .11083 L .33613 .10596 L closepath p 1 1 1 r
F P
0 g
s
.33612 .11083 m .33489 .10684 L .33369 .1113 L closepath p .91 1 1 r
F P
0 g
s
.35068 .11489 m .33866 .11379 L .33612 .11083 L closepath p .955 .786 .906 r
F P
0 g
s
.33369 .1113 m .33612 .11083 L .33866 .11379 L closepath p .955 .786 .906 r
F P
0 g
s
.51595 .10695 m .51672 .11096 L .51765 .1055 L closepath p 1 1 1 r
F P
0 g
s
.51672 .11096 m .51595 .10695 L .51376 .11136 L closepath p .887 1 1 r
F P
0 g
s
.51672 .11096 m .5075 .11446 L .51376 .11136 L closepath p .955 .786 .906 r
F P
0 g
s
.52236 .11481 m .51546 .11504 L .51672 .11096 L closepath p .955 .786 .906 r
F P
0 g
s
.5075 .11446 m .51672 .11096 L .51546 .11504 L closepath p .955 .786 .906 r
F P
0 g
s
.70841 .11196 m .70927 .11127 L .7117 .11102 L closepath p .955 .786 .906 r
F P
0 g
s
.71057 .1112 m .7117 .11102 L .70927 .11127 L closepath p .955 .786 .906 r
F P
0 g
s
.71057 .1112 m .70927 .11127 L .7117 .11102 L closepath p .955 .786 .906 r
F P
0 g
s
.27047 .11106 m .27208 .11115 L .26983 .11104 L closepath p .955 .786 .906 r
F P
0 g
s
.27208 .11115 m .2717 .107 L .2693 .11103 L closepath p .82 .99 1 r
F P
0 g
s
.26876 .10688 m .2693 .11103 L .2717 .107 L closepath p .821 .99 1 r
F P
0 g
s
.2693 .11103 m .26983 .11104 L .2522 .11284 L closepath p .955 .786 .906 r
F P
0 g
s
.30361 .10688 m .30407 .11103 L .30907 .10603 L closepath p .898 1 1 r
F P
0 g
s
.30407 .11103 m .30361 .10688 L .29574 .11103 L closepath p .838 1 1 r
F P
0 g
s
.29526 .10688 m .29574 .11103 L .30361 .10688 L closepath p .838 1 1 r
F P
0 g
s
.73912 .10688 m .73856 .11103 L .74452 .10603 L closepath p .892 1 1 r
F P
0 g
s
.73856 .11103 m .73912 .10688 L .73022 .11103 L closepath p .838 1 1 r
F P
0 g
s
.73076 .10688 m .73022 .11103 L .73912 .10688 L closepath p .838 1 1 r
F P
0 g
s
.70496 .11106 m .70653 .11115 L .70432 .11104 L closepath p .955 .786 .906 r
F P
0 g
s
.70653 .11115 m .70717 .107 L .70379 .11103 L closepath p .823 .99 1 r
F P
0 g
s
.70427 .10688 m .70379 .11103 L .70717 .107 L closepath p .823 .99 1 r
F P
0 g
s
.29574 .11103 m .29991 .11103 L .31362 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.30407 .11103 m .31362 .11482 L .29991 .11103 L closepath p .955 .786 .906 r
F P
0 g
s
.31362 .11482 m .29635 .11255 L .29574 .11103 L closepath p .955 .786 .906 r
F P
0 g
s
.73022 .11103 m .73439 .11103 L .75725 .11304 L closepath p .955 .786 .906 r
F P
0 g
s
.73856 .11103 m .75725 .11304 L .73439 .11103 L closepath p .955 .786 .906 r
F P
0 g
s
.6866 .11288 m .73022 .11103 L .75725 .11304 L closepath p .955 .786 .906 r
F P
0 g
s
.31362 .11482 m .2522 .11284 L .26983 .11104 L closepath p .955 .786 .906 r
F P
0 g
s
.31362 .11482 m .26983 .11104 L .27208 .11115 L closepath p .955 .786 .906 r
F P
0 g
s
.6866 .11288 m .70432 .11104 L .70653 .11115 L closepath p .955 .786 .906 r
F P
0 g
s
.27485 .11127 m .27433 .10712 L .27208 .11115 L closepath p .82 .99 1 r
F P
0 g
s
.2717 .107 m .27208 .11115 L .27433 .10712 L closepath p .82 .99 1 r
F P
0 g
s
.70927 .11127 m .70976 .10712 L .70653 .11115 L closepath p .823 .99 1 r
F P
0 g
s
.70717 .107 m .70653 .11115 L .70976 .10712 L closepath p .823 .989 1 r
F P
0 g
s
.70927 .11127 m .6866 .11288 L .70653 .11115 L closepath p .955 .786 .906 r
F P
0 g
s
.43228 .10706 m .43282 .11117 L .43396 .10675 L closepath p .908 1 1 r
F P
0 g
s
.43282 .11117 m .43228 .10706 L .42931 .11137 L closepath p .858 1 1 r
F P
0 g
s
.49082 .11438 m .44644 .11482 L .43282 .11117 L closepath p .955 .786 .906 r
F P
0 g
s
.42931 .11137 m .43282 .11117 L .44644 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.60909 .11162 m .61035 .10858 L .60808 .11119 L closepath p .601 .833 1 r
F P
0 g
s
.60827 .10932 m .60808 .11119 L .61035 .10858 L closepath p .747 1 1 r
F P
0 g
s
.60808 .11119 m .60827 .10932 L .60779 .11125 L closepath p .682 1 1 r
F P
0 g
s
.60808 .11119 m .60779 .11125 L .60827 .10932 L closepath p .682 1 1 r
F P
0 g
s
.6075 .11131 m .60808 .11119 L .60992 .11374 L closepath p .955 .786 .906 r
F P
0 g
s
.60999 .11221 m .60992 .11374 L .60808 .11119 L closepath p .955 .786 .906 r
F P
0 g
s
.60842 .11001 m .60779 .11125 L .60827 .10932 L closepath p .081 .081 .437 r
F P
0 g
s
.60779 .11125 m .60842 .11001 L .6075 .11131 L closepath p .515 .957 1 r
F P
0 g
s
.49082 .11438 m .47233 .11153 L .47465 .11126 L closepath p .955 .786 .906 r
F P
0 g
s
.27433 .10712 m .27485 .11127 L .2771 .10684 L closepath p .878 1 1 r
F P
0 g
s
.70976 .10712 m .70927 .11127 L .71247 .10684 L closepath p .874 1 1 r
F P
0 g
s
.70841 .11196 m .6866 .11288 L .70927 .11127 L closepath p .955 .786 .906 r
F P
0 g
s
.69794 .11207 m .70927 .11127 L .6866 .11288 L closepath p .955 .786 .906 r
F P
0 g
s
.38588 .10717 m .38651 .11129 L .3878 .10683 L closepath p .906 1 1 r
F P
0 g
s
.38651 .11129 m .38588 .10717 L .38264 .11144 L closepath p .853 1 1 r
F P
0 g
s
.39158 .11379 m .37945 .11268 L .38651 .11129 L closepath p .955 .786 .906 r
F P
0 g
s
.38264 .11144 m .38651 .11129 L .37945 .11268 L closepath p .955 .786 .906 r
F P
0 g
s
.33295 .10718 m .33369 .1113 L .33489 .10684 L closepath p .906 1 1 r
F P
0 g
s
.33369 .1113 m .33295 .10718 L .32975 .11145 L closepath p .853 1 1 r
F P
0 g
s
.33866 .11379 m .32664 .11268 L .33369 .1113 L closepath p .955 .786 .906 r
F P
0 g
s
.32975 .11145 m .33369 .1113 L .32664 .11268 L closepath p .955 .786 .906 r
F P
0 g
s
.6077 .10916 m .6075 .11131 L .60842 .11001 L closepath p .131 .355 .876 r
F P
0 g
s
.6075 .11131 m .6077 .10916 L .60339 .11134 L closepath p .639 .985 1 r
F P
0 g
s
.60339 .11134 m .6075 .11131 L .60985 .11527 L closepath p .955 .786 .906 r
F P
0 g
s
.60992 .11374 m .60985 .11527 L .6075 .11131 L closepath p .955 .786 .906 r
F P
0 g
s
.47233 .11153 m .47226 .10738 L .46209 .11131 L closepath p .83 .997 1 r
F P
0 g
s
.46214 .10717 m .46209 .11131 L .47226 .10738 L closepath p .83 .997 1 r
F P
0 g
s
.46209 .11131 m .47233 .11153 L .46288 .11256 L closepath p .955 .786 .906 r
F P
0 g
s
.8287 .11144 m .82893 .11134 L .81943 .10467 L closepath p .493 .021 .021 r
F P
0 g
s
.82893 .11134 m .8287 .11144 L .83828 .11522 L closepath p .335 .017 .017 r
F P
0 g
s
.83828 .11522 m .82893 .11134 L .84203 .11167 L closepath p .955 .786 .906 r
F P
0 g
s
.8016 .11136 m .8025 .10722 L .79849 .11134 L closepath p .836 1 1 r
F P
0 g
s
.79919 .10719 m .79849 .11134 L .8025 .10722 L closepath p .835 1 1 r
F P
0 g
s
.77053 .10719 m .76991 .11134 L .77478 .10648 L closepath p .894 1 1 r
F P
0 g
s
.76991 .11134 m .77053 .10719 L .76809 .11134 L closepath p .838 1 1 r
F P
0 g
s
.76872 .10719 m .76809 .11134 L .77053 .10719 L closepath p .838 1 1 r
F P
0 g
s
.76809 .11134 m .76872 .10719 L .76571 .11165 L closepath p .882 1 1 r
F P
0 g
s
.60364 .10719 m .60339 .11134 L .6077 .10916 L closepath p .641 .826 1 r
F P
0 g
s
.60985 .11527 m .60326 .11504 L .60339 .11134 L closepath p .955 .786 .906 r
F P
0 g
s
.76571 .11165 m .76809 .11134 L .76743 .11196 L closepath p .955 .786 .906 r
F P
0 g
s
.76991 .11134 m .76743 .11196 L .76809 .11134 L closepath p .955 .786 .906 r
F P
0 g
s
.80285 .11142 m .79849 .11134 L .8016 .11136 L closepath p .955 .786 .906 r
F P
0 g
s
.78514 .11481 m .78126 .11487 L .76991 .11134 L closepath p .955 .786 .906 r
F P
0 g
s
.76743 .11196 m .76991 .11134 L .78126 .11487 L closepath p .955 .786 .906 r
F P
0 g
s
.80285 .11142 m .794 .11312 L .79849 .11134 L closepath p .955 .786 .906 r
F P
0 g
s
.51349 .10724 m .51376 .11136 L .51595 .10695 L closepath p .882 1 1 r
F P
0 g
s
.51376 .11136 m .51349 .10724 L .50807 .1115 L closepath p .846 1 1 r
F P
0 g
s
.95617 .11107 m .9535 .11136 L .94997 .11301 L closepath p .049 .049 .049 r
F P
0 g
s
.51376 .11136 m .5075 .11446 L .50807 .1115 L closepath p .955 .786 .906 r
F P
0 g
s
.80285 .11142 m .8041 .10733 L .8016 .11136 L closepath p .82 .988 1 r
F P
0 g
s
.8025 .10722 m .8016 .11136 L .8041 .10733 L closepath p .815 .981 1 r
F P
0 g
s
.42912 .10721 m .42931 .11137 L .43228 .10706 L closepath p .857 1 1 r
F P
0 g
s
.42931 .11137 m .42912 .10721 L .42704 .11168 L closepath p .89 1 1 r
F P
0 g
s
.44644 .11482 m .44249 .11489 L .42931 .11137 L closepath p .955 .786 .906 r
F P
0 g
s
.42704 .11168 m .42931 .11137 L .44249 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.804 .11296 m .80705 .10925 L .80285 .11142 L closepath p .572 .642 1 r
F P
0 g
s
.8041 .10733 m .80285 .11142 L .80705 .10925 L closepath p .622 .777 1 r
F P
0 g
s
.80285 .11142 m .80514 .1145 L .78514 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.85895 .1076 m .85915 .11144 L .86369 .10355 L closepath p 1 1 1 r
F P
0 g
s
.85915 .11144 m .85895 .1076 L .85012 .11304 L closepath p .889 1 1 r
F P
0 g
s
.85506 .11264 m .85915 .11144 L .87006 .11173 L closepath p .955 .786 .906 r
F P
0 g
s
.87057 .11165 m .87006 .11173 L .85915 .11144 L closepath p .955 .786 .906 r
F P
0 g
s
.36321 .11146 m .36308 .10732 L .36009 .11144 L closepath p .835 1 1 r
F P
0 g
s
.35977 .10729 m .36009 .11144 L .36308 .10732 L closepath p .835 1 1 r
F P
0 g
s
.38236 .10729 m .38264 .11144 L .38588 .10717 L closepath p .851 1 1 r
F P
0 g
s
.38264 .11144 m .38236 .10729 L .38025 .11175 L closepath p .887 1 1 r
F P
0 g
s
.38264 .11144 m .37945 .11268 L .38025 .11175 L closepath p .955 .786 .906 r
F P
0 g
s
.36446 .11152 m .36009 .11144 L .36321 .11146 L closepath p .955 .786 .906 r
F P
0 g
s
.36446 .11152 m .35757 .11321 L .36009 .11144 L closepath p .955 .786 .906 r
F P
0 g
s
.41607 .11147 m .41607 .10733 L .41295 .11145 L closepath p .835 1 1 r
F P
0 g
s
.41275 .1073 m .41295 .11145 L .41607 .10733 L closepath p .835 1 1 r
F P
0 g
s
.32935 .1073 m .32975 .11145 L .33295 .10718 L closepath p .851 1 1 r
F P
0 g
s
.32975 .11145 m .32935 .1073 L .32742 .11176 L closepath p .889 1 1 r
F P
0 g
s
.32975 .11145 m .32664 .11268 L .32742 .11176 L closepath p .955 .786 .906 r
F P
0 g
s
.41732 .11153 m .41295 .11145 L .41607 .11147 L closepath p .955 .786 .906 r
F P
0 g
s
.41732 .11153 m .41051 .11321 L .41295 .11145 L closepath p .955 .786 .906 r
F P
0 g
s
.36446 .11152 m .36468 .10743 L .36321 .11146 L closepath p .817 .989 1 r
F P
0 g
s
.36308 .10732 m .36321 .11146 L .36468 .10743 L closepath p .81 .981 1 r
F P
0 g
s
.41732 .11153 m .41767 .10744 L .41607 .11147 L closepath p .817 .989 1 r
F P
0 g
s
.41607 .10733 m .41607 .11147 L .41767 .10744 L closepath p .81 .981 1 r
F P
0 g
s
.50809 .10735 m .50807 .1115 L .51349 .10724 L closepath p .846 1 1 r
F P
0 g
s
.50807 .1115 m .50809 .10735 L .50752 .11204 L closepath p 1 1 1 r
F P
0 g
s
.36446 .11152 m .36596 .11262 L .35757 .11321 L closepath p .955 .786 .906 r
F P
0 g
s
.35969 .11426 m .35757 .11321 L .36596 .11262 L closepath p .955 .786 .906 r
F P
0 g
s
.36596 .11262 m .36577 .11539 L .35969 .11426 L closepath p .955 .786 .906 r
F P
0 g
s
.36552 .11307 m .36752 .10936 L .36446 .11152 L closepath p .465 .573 .991 r
F P
0 g
s
.36468 .10743 m .36446 .11152 L .36752 .10936 L closepath p .55 .744 1 r
F P
0 g
s
.47226 .10738 m .47233 .11153 L .47595 .10654 L closepath p .919 1 1 r
F P
0 g
s
.49082 .11438 m .46288 .11256 L .47233 .11153 L closepath p .955 .786 .906 r
F P
0 g
s
.41732 .11153 m .41882 .11263 L .41051 .11321 L closepath p .955 .786 .906 r
F P
0 g
s
.41261 .11426 m .41051 .11321 L .41882 .11263 L closepath p .955 .786 .906 r
F P
0 g
s
.41882 .11263 m .41865 .11539 L .41261 .11426 L closepath p .955 .786 .906 r
F P
0 g
s
.41838 .11307 m .42051 .10937 L .41732 .11153 L closepath p .48 .582 .995 r
F P
0 g
s
.41767 .10744 m .41732 .11153 L .42051 .10937 L closepath p .559 .748 1 r
F P
0 g
s
.61011 .11206 m .61197 .10905 L .60909 .11162 L closepath p .625 .839 1 r
F P
0 g
s
.61035 .10858 m .60909 .11162 L .61197 .10905 L closepath p .651 .891 1 r
F P
0 g
s
.59543 .07612 m .59688 .06428 L .59519 .07438 L closepath p .048 .048 .271 r
F P
0 g
s
.59664 .06252 m .59519 .07438 L .59688 .06428 L closepath p .048 .048 .271 r
F P
0 g
s
.86891 .11227 m .87097 .11229 L .87057 .11165 L closepath p .068 .068 .068 r
F P
0 g
s
.86927 .11309 m .87097 .11229 L .87097 .11229 L p .081 .081 .081 r
F P
0 g
s
.87097 .11229 m .86891 .11227 L .86927 .11309 L p .081 .081 .081 r
F P
0 g
s
.86682 .11424 m .86927 .11309 L .86926 .11309 L p .091 .091 .091 r
F P
0 g
s
.86926 .11309 m .86726 .11288 L .86682 .11424 L p .091 .091 .091 r
F P
0 g
s
.86682 .11424 m .86483 .11402 L .86572 .11476 L closepath p .092 .092 .092 r
F P
0 g
s
.86682 .11424 m .86572 .11476 L .86572 .11475 L p .714 1 1 r
F P
0 g
s
.87006 .11173 m .86934 .11196 L .84822 .11651 L closepath p .955 .786 .906 r
F P
0 g
s
.87006 .11173 m .84822 .11651 L .85506 .11264 L closepath p .955 .786 .906 r
F P
0 g
s
.86934 .11196 m .86726 .11288 L .86323 .11398 L closepath p .955 .786 .906 r
F P
0 g
s
.85712 .11601 m .84822 .11651 L .86934 .11196 L closepath p .955 .786 .906 r
F P
0 g
s
.86726 .11288 m .86927 .11309 L .86891 .11227 L closepath p .08 .08 .08 r
F P
0 g
s
.86483 .11402 m .86682 .11424 L .86726 .11288 L closepath p .091 .091 .091 r
F P
0 g
s
.86483 .11402 m .86323 .11398 L .86726 .11288 L closepath p .955 .786 .906 r
F P
0 g
s
.86572 .11475 m .87129 .10348 L .86682 .11424 L p .714 1 1 r
F P
0 g
s
.86927 .11309 m .86682 .11424 L .86682 .11424 L p .713 1 1 r
F P
0 g
s
.86682 .11424 m .87241 .10295 L .86927 .11309 L p .713 1 1 r
F P
0 g
s
.87097 .11229 m .86927 .11309 L .86935 .11294 L p .713 1 1 r
F P
0 g
s
.87153 .11207 m .87057 .11165 L .87097 .11229 L p .068 .068 .068 r
F P
0 g
s
.87057 .11165 m .87155 .11209 L .87289 .11242 L closepath p .259 .01 .01 r
F P
0 g
s
.86935 .11294 m .8749 .10179 L .87097 .11229 L p .713 1 1 r
F P
0 g
s
.87097 .11229 m .87155 .11209 L .87153 .11207 L p .068 .068 .068 r
F P
0 g
s
.765 .10807 m .76571 .11165 L .76872 .10719 L closepath p .903 1 1 r
F P
0 g
s
.76571 .11165 m .765 .10807 L .76494 .11258 L closepath p 1 1 1 r
F P
0 g
s
.76743 .11196 m .76494 .11258 L .76571 .11165 L closepath p .955 .786 .906 r
F P
0 g
s
.85012 .11304 m .85094 .10889 L .84203 .11167 L closepath p .778 .946 1 r
F P
0 g
s
.84346 .10765 m .84203 .11167 L .85094 .10889 L closepath p .778 .947 1 r
F P
0 g
s
.84203 .11167 m .85012 .11304 L .83828 .11522 L closepath p .955 .786 .906 r
F P
0 g
s
.4255 .10811 m .42704 .11168 L .42912 .10721 L closepath p .917 1 1 r
F P
0 g
s
.42704 .11168 m .4255 .10811 L .42628 .11263 L closepath p 1 1 1 r
F P
0 g
s
.44249 .11489 m .43699 .1151 L .42704 .11168 L closepath p .955 .786 .906 r
F P
0 g
s
.42628 .11263 m .42704 .11168 L .43699 .1151 L closepath p .955 .786 .906 r
F P
0 g
s
.37862 .10817 m .38025 .11175 L .38236 .10729 L closepath p .913 1 1 r
F P
0 g
s
.38025 .11175 m .37862 .10817 L .37945 .11268 L closepath p 1 1 1 r
F P
0 g
s
.32566 .10818 m .32742 .11176 L .32935 .1073 L closepath p .916 1 1 r
F P
0 g
s
.32742 .11176 m .32566 .10818 L .32664 .11268 L closepath p 1 1 1 r
F P
0 g
s
.579 .09074 m .5802 .07912 L .57035 .07458 L closepath p .097 .097 .665 r
F P
0 g
s
.57142 .06273 m .57035 .07458 L .5802 .07912 L closepath p .097 .097 .665 r
F P
0 g
s
.49082 .11438 m .49286 .11364 L .49135 .11195 L closepath p .144 .144 .403 r
F P
0 g
s
.4936 .11069 m .49135 .11195 L .49286 .11364 L closepath p .146 .146 .393 r
F P
0 g
s
.78126 .11487 m .77578 .11504 L .76743 .11196 L closepath p .955 .786 .906 r
F P
0 g
s
.76494 .11258 m .76743 .11196 L .77578 .11504 L closepath p .955 .786 .906 r
F P
0 g
s
.07319 .11035 m .0766 .11197 L .07141 .10326 L closepath p 1 .859 .561 r
F P
0 g
s
.0766 .11197 m .07319 .11035 L .07717 .11545 L closepath p 1 .904 .622 r
F P
0 g
s
.50571 .10841 m .50752 .11204 L .50809 .10735 L closepath p .989 1 1 r
F P
0 g
s
.50752 .11204 m .50571 .10841 L .50696 .11258 L closepath p 1 1 1 r
F P
0 g
s
.6108 .11404 m .61344 .11219 L .61011 .11206 L closepath p .312 .382 .855 r
F P
0 g
s
.61197 .10905 m .61011 .11206 L .61344 .11219 L closepath p .305 .416 .893 r
F P
0 g
s
.87545 .11319 m .87877 .11203 L .87542 .11227 L closepath p .327 .281 .731 r
F P
0 g
s
.87872 .11041 m .87542 .11227 L .87877 .11203 L closepath p .327 .281 .731 r
F P
0 g
s
.87542 .11227 m .87514 .11366 L .87289 .11242 L closepath p .955 .786 .906 r
F P
0 g
s
.09939 .09221 m .09329 .08061 L .10301 .07512 L closepath p 1 .98 .559 r
F P
0 g
s
.09697 .06327 m .10301 .07512 L .09329 .08061 L closepath p 1 .98 .559 r
F P
0 g
s
.87269 .11255 m .87289 .11242 L .87158 .11209 L p .336 .019 .019 r
F P
0 g
s
.87446 .11331 m .87289 .11242 L .87514 .11366 L closepath p .955 .786 .906 r
F P
0 g
s
.87289 .11242 m .87269 .11255 L .87514 .11366 L closepath p .413 .02 .02 r
F P
0 g
s
.87514 .11366 m .87712 .11453 L .87545 .11319 L closepath p .156 .156 .273 r
F P
0 g
s
.87877 .11203 m .87545 .11319 L .87712 .11453 L closepath p .156 .156 .272 r
F P
0 g
s
.87798 .10455 m .88374 .09313 L .87728 .10104 L closepath p .113 .113 .533 r
F P
0 g
s
.87872 .11041 m .8845 .09907 L .87798 .10455 L closepath p .104 .104 .459 r
F P
0 g
s
.88374 .09313 m .87798 .10455 L .8845 .09907 L closepath p .104 .104 .459 r
F P
0 g
s
.87269 .11255 m .87838 .10124 L .87713 .11437 L p .386 .743 1 r
F P
0 g
s
.87713 .11437 m .87712 .11453 L .87269 .11255 L p .386 .743 1 r
F P
0 g
s
.87158 .11209 m .87155 .11209 L .87269 .11255 L p .336 .019 .019 r
F P
0 g
s
.80428 .06348 m .79972 .07533 L .80129 .05298 L closepath p .11 .11 .575 r
F P
0 g
s
.79972 .07533 m .80428 .06348 L .80444 .08546 L closepath p .127 .157 .724 r
F P
0 g
s
.76324 .11072 m .76494 .11258 L .765 .10807 L closepath p 1 1 .924 r
F P
0 g
s
.76494 .11258 m .76324 .11072 L .76533 .1146 L closepath p 1 .759 .289 r
F P
0 g
s
.50462 .11072 m .50696 .11258 L .50571 .10841 L closepath p 1 1 .892 r
F P
0 g
s
.50696 .11258 m .50462 .11072 L .50736 .1146 L closepath p 1 .831 .434 r
F P
0 g
s
.77578 .11504 m .77304 .11513 L .76494 .11258 L closepath p .955 .786 .906 r
F P
0 g
s
.76533 .1146 m .76494 .11258 L .77304 .11513 L closepath p .955 .786 .906 r
F P
0 g
s
.95298 .11547 m .95162 .11263 L .9568 .11281 L closepath p .955 .786 .906 r
F P
0 g
s
.95162 .11263 m .95298 .11547 L .94348 .11547 L closepath p .955 .786 .906 r
F P
0 g
s
.42374 .11077 m .42628 .11263 L .4255 .10811 L closepath p 1 1 .967 r
F P
0 g
s
.42628 .11263 m .42374 .11077 L .42659 .1147 L closepath p 1 .879 .515 r
F P
0 g
s
.43699 .1151 m .43424 .11521 L .42628 .11263 L closepath p .955 .786 .906 r
F P
0 g
s
.42659 .1147 m .42628 .11263 L .43424 .11521 L closepath p .955 .786 .906 r
F P
0 g
s
.85012 .11304 m .85506 .11264 L .84822 .11651 L closepath p .955 .786 .906 r
F P
0 g
s
.37679 .11082 m .37945 .11268 L .37862 .10817 L closepath p 1 1 .978 r
F P
0 g
s
.37945 .11268 m .37679 .11082 L .37981 .11471 L closepath p 1 .867 .509 r
F P
0 g
s
.39035 .1151 m .38761 .11521 L .37945 .11268 L closepath p .955 .786 .906 r
F P
0 g
s
.37981 .11471 m .37945 .11268 L .38761 .11521 L closepath p .955 .786 .906 r
F P
0 g
s
.39158 .11379 m .39582 .11489 L .37945 .11268 L closepath p .955 .786 .906 r
F P
0 g
s
.39035 .1151 m .37945 .11268 L .39582 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.33737 .1151 m .33464 .11521 L .32664 .11268 L closepath p .955 .786 .906 r
F P
0 g
s
.32697 .11471 m .32664 .11268 L .33464 .11521 L closepath p .955 .786 .906 r
F P
0 g
s
.33866 .11379 m .34283 .11489 L .32664 .11268 L closepath p .955 .786 .906 r
F P
0 g
s
.33737 .1151 m .32664 .11268 L .34283 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.93891 .11255 m .93946 .11281 L .93566 .10787 L closepath p .684 .233 .029 r
F P
0 g
s
.94242 .11783 m .93737 .11462 L .95257 .11441 L closepath p .955 .786 .906 r
F P
0 g
s
.93946 .11281 m .93891 .11255 L .94348 .11547 L closepath p .42 .008 .008 r
F P
0 g
s
.95418 .11784 m .95257 .11441 L .959 .11466 L closepath p .955 .786 .906 r
F P
0 g
s
.95257 .11441 m .95418 .11784 L .94242 .11783 L closepath p .955 .786 .906 r
F P
0 g
s
.95298 .11547 m .95508 .11569 L .9568 .11281 L closepath p .115 .115 .115 r
F P
0 g
s
.9597 .11254 m .9568 .11281 L .95508 .11569 L closepath p .117 .117 .117 r
F P
0 g
s
.2522 .11284 m .24937 .11183 L .25377 .11668 L closepath p 1 .746 .39 r
F P
0 g
s
.26201 .11482 m .25742 .11639 L .2522 .11284 L closepath p .955 .786 .906 r
F P
0 g
s
.25376 .11668 m .2522 .11284 L .25742 .11639 L closepath p .955 .786 .906 r
F P
0 g
s
.30072 .11482 m .28782 .11482 L .2522 .11284 L closepath p .955 .786 .906 r
F P
0 g
s
.26201 .11482 m .2522 .11284 L .28782 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.2522 .11284 m .31362 .11482 L .30072 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.6866 .11288 m .68488 .11188 L .68831 .1167 L closepath p .968 .579 .104 r
F P
0 g
s
.69646 .11481 m .692 .11639 L .6866 .11288 L closepath p .955 .786 .906 r
F P
0 g
s
.68831 .1167 m .6866 .11288 L .692 .11639 L closepath p .955 .786 .906 r
F P
0 g
s
.6866 .11288 m .72227 .11481 L .69646 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.6866 .11288 m .75725 .11304 L .72227 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.7117 .06385 m .70853 .07568 L .70993 .05298 L closepath p .077 .077 .387 r
F P
0 g
s
.80514 .1145 m .80827 .11265 L .804 .11296 L closepath p .458 .582 1 r
F P
0 g
s
.80705 .10925 m .804 .11296 L .80827 .11265 L closepath p .463 .487 .904 r
F P
0 g
s
.32209 .11673 m .32349 .1162 L .32269 .11303 L closepath p .112 .112 .305 r
F P
0 g
s
.32269 .11303 m .32209 .11673 L .32033 .11757 L closepath p .955 .786 .906 r
F P
0 g
s
.85094 .10889 m .85012 .11304 L .85895 .1076 L closepath p .89 1 1 r
F P
0 g
s
.75676 .11674 m .75929 .11621 L .75725 .11304 L closepath p .171 .171 .554 r
F P
0 g
s
.76025 .11138 m .75725 .11304 L .75929 .11621 L closepath p .194 .177 .546 r
F P
0 g
s
.75725 .11304 m .75676 .11674 L .75501 .11758 L closepath p .955 .786 .906 r
F P
0 g
s
.75725 .11304 m .75501 .11758 L .74807 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.75725 .11304 m .74807 .11481 L .72227 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.84822 .11651 m .83828 .11522 L .85012 .11304 L closepath p .955 .786 .906 r
F P
0 g
s
.36657 .11461 m .3686 .11276 L .36552 .11307 L closepath p .299 .476 .951 r
F P
0 g
s
.36752 .10936 m .36552 .11307 L .3686 .11276 L closepath p .296 .36 .838 r
F P
0 g
s
.41945 .11461 m .42161 .11276 L .41838 .11307 L closepath p .321 .492 .959 r
F P
0 g
s
.42051 .10937 m .41838 .11307 L .42161 .11276 L closepath p .32 .379 .849 r
F P
0 g
s
.35757 .11321 m .3618 .11532 L .35068 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.41051 .11321 m .4147 .11532 L .40371 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.59676 .08745 m .59824 .07578 L .59543 .07612 L closepath p .044 .044 .242 r
F P
0 g
s
.59688 .06428 m .59543 .07612 L .59824 .07578 L closepath p .044 .044 .242 r
F P
0 g
s
.87712 .11453 m .87514 .11366 L .87269 .11255 L closepath p .001 .294 .649 r
F P
0 g
s
.53271 .11452 m .53394 .11505 L .5333 .11377 L closepath p .11 .11 .11 r
F P
0 g
s
.53519 .11327 m .5333 .11377 L .53394 .11505 L closepath p .108 .108 .108 r
F P
0 g
s
.40371 .11489 m .39976 .11482 L .39158 .11379 L closepath p .955 .786 .906 r
F P
0 g
s
.39582 .11489 m .39158 .11379 L .39976 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.35068 .11489 m .34671 .11482 L .33866 .11379 L closepath p .955 .786 .906 r
F P
0 g
s
.34283 .11489 m .33866 .11379 L .34671 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.58128 .11301 m .58317 .11381 L .56961 .09082 L closepath p .976 .562 .112 r
F P
0 g
s
.58317 .11381 m .58128 .11301 L .58401 .11454 L closepath p .673 .134 .059 r
F P
0 g
s
.59076 .11504 m .58732 .11522 L .58317 .11381 L closepath p .955 .786 .906 r
F P
0 g
s
.58484 .11527 m .58317 .11381 L .58732 .11522 L closepath p .955 .786 .906 r
F P
0 g
s
.59943 .11487 m .59667 .11481 L .58317 .11381 L closepath p .955 .786 .906 r
F P
0 g
s
.59076 .11504 m .58317 .11381 L .59667 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.16337 .11384 m .16362 .1162 L .15173 .11652 L closepath p .955 .786 .906 r
F P
0 g
s
.86323 .11398 m .86483 .11402 L .85712 .11601 L closepath p .955 .786 .906 r
F P
0 g
s
.86381 .1145 m .86572 .11476 L .86483 .11402 L closepath p .092 .092 .092 r
F P
0 g
s
.86381 .1145 m .85712 .11601 L .86483 .11402 L closepath p .955 .786 .906 r
F P
0 g
s
.61033 .11466 m .61248 .11455 L .6108 .11404 L closepath p .132 .132 .132 r
F P
0 g
s
.61344 .11219 m .6108 .11404 L .61248 .11455 L closepath p .139 .139 .286 r
F P
0 g
s
.3618 .11532 m .35969 .11426 L .36577 .11539 L closepath p .955 .786 .906 r
F P
0 g
s
.4147 .11532 m .41261 .11426 L .41865 .11539 L closepath p .955 .786 .906 r
F P
0 g
s
.48947 .11488 m .49088 .11526 L .49082 .11438 L closepath p .076 .076 .076 r
F P
0 g
s
.49286 .11364 m .49082 .11438 L .49088 .11526 L closepath p .066 .066 .066 r
F P
0 g
s
.49082 .11438 m .48812 .11539 L .47299 .11464 L closepath p .955 .786 .906 r
F P
0 g
s
.44644 .11482 m .49082 .11438 L .45515 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.46544 .06538 m .46595 .0772 L .46142 .06108 L closepath p .112 .326 .848 r
F P
0 g
s
.51546 .11504 m .51143 .11522 L .5075 .11446 L closepath p .955 .786 .906 r
F P
0 g
s
.50855 .11527 m .5075 .11446 L .51143 .11522 L closepath p .955 .786 .906 r
F P
0 g
s
.85712 .11601 m .85846 .11641 L .86381 .1145 L closepath p .095 .063 .063 r
F P
0 g
s
.86572 .11476 m .86381 .1145 L .85846 .11641 L closepath p .088 .062 .062 r
F P
0 g
s
.80464 .11489 m .80718 .1148 L .80514 .1145 L closepath p .121 .121 .121 r
F P
0 g
s
.80827 .11265 m .80514 .1145 L .80718 .1148 L closepath p .148 .148 .288 r
F P
0 g
s
.80514 .1145 m .80413 .11527 L .79712 .11469 L closepath p .955 .786 .906 r
F P
0 g
s
.7891 .11487 m .78514 .11481 L .80514 .1145 L closepath p .955 .786 .906 r
F P
0 g
s
.53213 .11527 m .53221 .11572 L .53271 .11452 L closepath p .033 .033 .033 r
F P
0 g
s
.53394 .11505 m .53271 .11452 L .53221 .11572 L closepath p .085 .084 .084 r
F P
0 g
s
.5829 .11487 m .58401 .11454 L .58128 .11301 L closepath p .844 .347 .093 r
F P
0 g
s
.58401 .11454 m .5829 .11487 L .58484 .11527 L closepath p .762 .225 .087 r
F P
0 g
s
.7643 .1141 m .76533 .1146 L .76324 .11072 L closepath p .931 .599 .078 r
F P
0 g
s
.76533 .1146 m .7643 .1141 L .76642 .11527 L closepath p .418 .012 .012 r
F P
0 g
s
.50568 .11412 m .50736 .1146 L .50462 .11072 L closepath p 1 .71 .26 r
F P
0 g
s
.50736 .1146 m .50568 .11412 L .50855 .11527 L closepath p .49 .039 .039 r
F P
0 g
s
.77304 .11513 m .77031 .11522 L .76533 .1146 L closepath p .955 .786 .906 r
F P
0 g
s
.76642 .11527 m .76533 .1146 L .77031 .11522 L closepath p .955 .786 .906 r
F P
0 g
s
.36617 .115 m .36757 .11484 L .36657 .11461 L closepath p .078 .078 .078 r
F P
0 g
s
.3686 .11276 m .36657 .11461 L .36757 .11484 L closepath p .089 .089 .089 r
F P
0 g
s
.41905 .115 m .42058 .11485 L .41945 .11461 L closepath p .086 .086 .086 r
F P
0 g
s
.42161 .11276 m .41945 .11461 L .42058 .11485 L closepath p .099 .099 .099 r
F P
0 g
s
.93686 .1144 m .93737 .11462 L .93284 .1089 L closepath p .669 .21 .027 r
F P
0 g
s
.93737 .11462 m .93686 .1144 L .94242 .11783 L closepath p .412 .009 .009 r
F P
0 g
s
.47946 .11532 m .47299 .11464 L .48812 .11539 L closepath p .955 .786 .906 r
F P
0 g
s
.47299 .11464 m .47946 .11532 L .45515 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.95418 .11784 m .95626 .11808 L .959 .11466 L closepath p .113 .113 .113 r
F P
0 g
s
.96186 .11444 m .959 .11466 L .95626 .11808 L closepath p .115 .115 .115 r
F P
0 g
s
.60985 .11527 m .61015 .11572 L .61033 .11466 L closepath p .047 .047 .047 r
F P
0 g
s
.61248 .11455 m .61033 .11466 L .61015 .11572 L closepath p .078 .078 .078 r
F P
0 g
s
.80017 .11522 m .79712 .11469 L .80413 .11527 L closepath p .955 .786 .906 r
F P
0 g
s
.79712 .11469 m .80017 .11522 L .7891 .11487 L closepath p .955 .786 .906 r
F P
0 g
s
.42465 .11418 m .42659 .1147 L .42374 .11077 L closepath p 1 .773 .355 r
F P
0 g
s
.42659 .1147 m .42465 .11418 L .42754 .11539 L closepath p .65 .089 .067 r
F P
0 g
s
.43424 .11521 m .43149 .11532 L .42659 .1147 L closepath p .955 .786 .906 r
F P
0 g
s
.42754 .11539 m .42659 .1147 L .43149 .11532 L closepath p .955 .786 .906 r
F P
0 g
s
.42754 .11539 m .43149 .11532 L .42659 .1147 L closepath p .955 .786 .906 r
F P
0 g
s
.37779 .11422 m .37981 .11471 L .37679 .11082 L closepath p 1 .764 .353 r
F P
0 g
s
.37981 .11471 m .37779 .11422 L .38094 .11539 L closepath p .572 .058 .058 r
F P
0 g
s
.38761 .11521 m .38488 .11532 L .37981 .11471 L closepath p .955 .786 .906 r
F P
0 g
s
.38094 .11539 m .37981 .11471 L .38488 .11532 L closepath p .955 .786 .906 r
F P
0 g
s
.38094 .11539 m .38488 .11532 L .37981 .11471 L closepath p .955 .786 .906 r
F P
0 g
s
.32697 .11471 m .3248 .11421 L .32803 .11539 L closepath p .62 .068 .068 r
F P
0 g
s
.33464 .11521 m .33192 .11532 L .32697 .11471 L closepath p .955 .786 .906 r
F P
0 g
s
.32803 .11539 m .32697 .11471 L .33192 .11532 L closepath p .955 .786 .906 r
F P
0 g
s
.32803 .11539 m .33192 .11532 L .32697 .11471 L closepath p .955 .786 .906 r
F P
0 g
s
.78126 .11487 m .78206 .11531 L .78514 .11481 L closepath p .154 .043 .043 r
F P
0 g
s
.7859 .11525 m .78514 .11481 L .78206 .11531 L closepath p .154 .043 .043 r
F P
0 g
s
.78514 .11481 m .7859 .11525 L .7891 .11487 L closepath p .166 .041 .041 r
F P
0 g
s
.692 .11639 m .69402 .1163 L .69646 .11481 L closepath p .055 .055 .055 r
F P
0 g
s
.69699 .11525 m .69646 .11481 L .69402 .1163 L closepath p .055 .055 .055 r
F P
0 g
s
.69699 .11525 m .69402 .1163 L .69424 .11546 L p .721 1 1 r
F P
0 g
s
.70296 .11481 m .69646 .11481 L .69699 .11525 L p .16 .042 .042 r
F P
0 g
s
.69424 .11546 m .69698 .10505 L .69699 .11525 L p .721 1 1 r
F P
0 g
s
.69699 .11525 m .74807 .11481 L .70296 .11481 L p .16 .042 .042 r
F P
0 g
s
.74873 .11525 m .74807 .11481 L .69699 .11525 L closepath p .16 .042 .042 r
F P
0 g
s
.75294 .11639 m .74807 .11481 L .75501 .11758 L closepath p .955 .786 .906 r
F P
0 g
s
.75294 .11639 m .75501 .11758 L .74807 .11481 L closepath p .955 .786 .906 r
F P
0 g
s
.74807 .11481 m .74873 .11525 L .75294 .11639 L closepath p .296 .021 .021 r
F P
0 g
s
.51546 .11504 m .51562 .11548 L .52236 .11481 L closepath p .138 .042 .042 r
F P
0 g
s
.52242 .11525 m .52236 .11481 L .51562 .11548 L closepath p .138 .042 .042 r
F P
0 g
s
.52236 .11481 m .52242 .11525 L .52725 .11504 L closepath p .192 .042 .042 r
F P
0 g
s
.59076 .11504 m .59114 .11548 L .59667 .11481 L closepath p .137 .043 .043 r
F P
0 g
s
.59693 .11525 m .59667 .11481 L .59114 .11548 L closepath p .137 .043 .043 r
F P
0 g
s
.59667 .11481 m .59693 .11525 L .60326 .11504 L closepath p .181 .041 .041 r
F P
0 g
s
.455 .11534 m .45515 .11489 L .4466 .11525 L p .166 .042 .042 r
F P
0 g
s
.4466 .11525 m .4463 .11527 L .455 .11534 L p .166 .042 .042 r
F P
0 g
s
.44644 .11482 m .4463 .11527 L .45515 .11489 L closepath p .166 .042 .042 r
F P
0 g
s
.46732 .1151 m .45515 .11489 L .47946 .11532 L closepath p .955 .786 .906 r
F P
0 g
s
.46732 .1151 m .47946 .11532 L .45515 .11489 L closepath p .955 .786 .906 r
F P
0 g
s
.45515 .11489 m .455 .11534 L .46732 .1151 L closepath p .173 .042 .042 r
F P
0 g
s
.46718 .11555 m .46732 .1151 L .455 .11534 L closepath p .173 .042 .042 r
F P
0 g
s
.45461 .10655 m .45391 .09515 L .45315 .10473 L closepath p .098 .224 .769 r
F P
0 g
s
.46214 .10717 m .46157 .09578 L .45461 .10655 L closepath p .528 .918 1 r
F P
0 g
s
.45391 .09515 m .45461 .10655 L .46157 .09578 L closepath p .528 .918 1 r
F P
0 g
s
.455 .11534 m .45432 .10407 L .46708 .11545 L p .565 .95 1 r
F P
0 g
s
.46708 .11545 m .46718 .11555 L .455 .11534 L p .565 .95 1 r
F P
0 g
s
.455 .11534 m .4463 .11527 L .4463 .11527 L p .571 .955 1 r
F P
0 g
s
.44644 .11482 m .44249 .11489 L .4424 .11532 L p .147 .042 .042 r
F P
0 g
s
.44257 .11531 m .44644 .11482 L p .4424 .11532 L .147 .042 .042 r
F P
0 g
s
.4463 .11527 m .44644 .11482 L .4424 .11533 L closepath p .147 .042 .042 r
F P
0 g
s
.4463 .11527 m .44548 .104 L .45485 .11516 L p .571 .955 1 r
F P
0 g
s
.45485 .11516 m .455 .11534 L p .4463 .11527 L .571 .955 1 r
F P
0 g
s
.4424 .11532 m .4424 .11533 L .44257 .11531 L p .147 .042 .042 r
F P
0 g
s
.35022 .11533 m .35068 .11489 L .34645 .11525 L p .175 .042 .042 r
F P
0 g
s
.34645 .11525 m .3463 .11527 L .35022 .11533 L p .175 .042 .042 r
F P
0 g
s
.34671 .11482 m .3463 .11527 L .35068 .11489 L closepath p .175 .043 .043 r
F P
0 g
s
.35624 .1151 m .35068 .11489 L .3618 .11532 L closepath p .955 .786 .906 r
F P
0 g
s
.35068 .11489 m .35022 .11533 L .35624 .1151 L closepath p .192 .043 .043 r
F P
0 g
s
.35575 .11554 m .35624 .1151 L .35022 .11533 L closepath p .192 .043 .043 r
F P
0 g
s
.35022 .11533 m .34794 .10407 L .35569 .11545 L p .55 .939 1 r
F P
0 g
s
.35569 .11545 m .35575 .11554 L .35022 .11533 L p .55 .939 1 r
F P
0 g
s
.35022 .11533 m .3463 .11527 L .3463 .11527 L p .564 .95 1 r
F P
0 g
s
.34671 .11482 m .34283 .11489 L .34249 .11531 L p .145 .041 .041 r
F P
0 g
s
.34271 .11531 m .34671 .11482 L p .34249 .11531 L .145 .041 .041 r
F P
0 g
s
.3463 .11527 m .34671 .11482 L .34247 .11533 L closepath p .145 .041 .041 r
F P
0 g
s
.3463 .11527 m .34396 .104 L .35016 .11522 L p .564 .95 1 r
F P
0 g
s
.35016 .11522 m .35022 .11533 L p .3463 .11527 L .564 .95 1 r
F P
0 g
s
.34249 .11531 m .34247 .11533 L .34271 .11531 L p .145 .041 .041 r
F P
0 g
s
.40339 .11533 m .40371 .11489 L .39963 .11525 L p .174 .042 .042 r
F P
0 g
s
.39963 .11525 m .39949 .11527 L .40339 .11533 L p .174 .042 .042 r
F P
0 g
s
.39976 .11482 m .39949 .11527 L .40371 .11489 L closepath p .174 .042 .042 r
F P
0 g
s
.4092 .1151 m .40371 .11489 L .4147 .11532 L closepath p .955 .786 .906 r
F P
0 g
s
.40371 .11489 m .40339 .11533 L .4092 .1151 L closepath p .19 .043 .043 r
F P
0 g
s
.40885 .11554 m .4092 .1151 L .40339 .11533 L closepath p .19 .043 .043 r
F P
0 g
s
.40339 .11533 m .40191 .10407 L .4088 .11545 L p .551 .939 1 r
F P
0 g
s
.4088 .11545 m .40885 .11554 L .40339 .11533 L p .551 .939 1 r
F P
0 g
s
.40339 .11533 m .39949 .11527 L .39949 .11527 L p .564 .95 1 r
F P
0 g
s
.39976 .11482 m .39582 .11489 L .39561 .11531 L p .146 .041 .041 r
F P
0 g
s
.39584 .11531 m .39976 .11482 L p .39561 .11531 L .146 .041 .041 r
F P
0 g
s
.39949 .11527 m .39976 .11482 L .3956 .11533 L closepath p .146 .041 .041 r
F P
0 g
s
.39949 .11527 m .39796 .104 L .40333 .11522 L p .564 .95 1 r
F P
0 g
s
.40333 .11522 m .40339 .11533 L p .39949 .11527 L .564 .95 1 r
F P
0 g
s
.39561 .11531 m .3956 .11533 L .39584 .11531 L p .146 .041 .041 r
F P
0 g
s
.26201 .11482 m .25742 .11639 L .25776 .11637 L p .021 .021 .021 r
F P
0 g
s
.25946 .11585 m .26201 .11482 L p .25776 .11637 L .021 .021 .021 r
F P
0 g
s
.26138 .11527 m .26201 .11482 L .25946 .11585 L p .02 .02 .02 r
F P
0 g
s
.25946 .11585 m .25826 .11634 L .26138 .11527 L p .02 .02 .02 r
F P
0 g
s
.26201 .11482 m .26138 .11527 L .30766 .11487 L p .16 .042 .042 r
F P
0 g
s
.30766 .11487 m .31362 .11482 L .30696 .11482 L p .16 .042 .042 r
F P
0 g
s
.30696 .11482 m .26201 .11482 L p .30766 .11487 L .16 .042 .042 r
F P
0 g
s
.31312 .11527 m .31362 .11482 L .26138 .11527 L closepath p .16 .042 .042 r
F P
0 g
s
.31836 .11639 m .31362 .11482 L .32033 .11757 L closepath p .955 .786 .906 r
F P
0 g
s
.31836 .11639 m .32033 .11757 L .31362 .11482 L closepath p .955 .786 .906 r
F P
0 g
s
.31362 .11482 m .31312 .11527 L .31836 .11639 L closepath p .42 .054 .054 r
F P
0 g
s
.31337 .10331 m .31053 .09187 L .30907 .10603 L closepath p .887 1 1 r
F P
0 g
s
.31312 .11527 m .31027 .104 L .31609 .11548 L p .316 .739 1 r
F P
0 g
s
.31659 .11646 m .31836 .11639 L .3138 .11541 L p .45 .06 .06 r
F P
0 g
s
.3138 .11541 m .31312 .11527 L .31659 .11646 L p .45 .06 .06 r
F P
0 g
s
.32136 .11796 m .32033 .11757 L .32209 .11673 L closepath p .955 .786 .906 r
F P
0 g
s
.32136 .11796 m .32209 .11673 L .32033 .11757 L closepath p .955 .786 .906 r
F P
0 g
s
.31836 .11639 m .31659 .11646 L .32136 .11796 L closepath p .61 .08 .08 r
F P
0 g
s
.32088 .11841 m .32136 .11796 L .31659 .11646 L closepath p .497 .057 .057 r
F P
0 g
s
.31659 .11646 m .31379 .10521 L .32066 .118 L p .232 .66 1 r
F P
0 g
s
.32066 .118 m .32088 .11841 L .31659 .11646 L p .232 .66 1 r
F P
0 g
s
.31609 .11548 m .31659 .11646 L .31312 .11527 L p .316 .739 1 r
F P
0 g
s
.26406 .10599 m .26046 .09458 L .25933 .10311 L closepath p .115 .538 .97 r
F P
0 g
s
.26876 .10688 m .26524 .09549 L .26406 .10599 L closepath p .436 .846 1 r
F P
0 g
s
.26046 .09458 m .26406 .10599 L .26524 .09549 L closepath p .436 .846 1 r
F P
0 g
s
.2717 .107 m .26822 .09562 L .26876 .10688 L closepath p .55 .938 1 r
F P
0 g
s
.26524 .09549 m .26876 .10688 L .26822 .09562 L closepath p .55 .938 1 r
F P
0 g
s
.27433 .10712 m .27088 .09573 L .2717 .107 L closepath p .549 .937 1 r
F P
0 g
s
.26822 .09562 m .2717 .107 L .27088 .09573 L closepath p .549 .937 1 r
F P
0 g
s
.26138 .11527 m .25774 .104 L .28205 .10894 L p .576 .959 1 r
F P
0 g
s
.28205 .10894 m .28215 .10896 L p .26138 .11527 L .576 .959 1 r
F P
0 g
s
.28215 .10896 m .31312 .11527 L .26139 .11527 L p .576 .959 1 r
F P
0 g
s
.26139 .11527 m .26138 .11527 L p .28215 .10896 L .576 .959 1 r
F P
0 g
s
.25776 .11637 m .25826 .11634 L .25946 .11585 L p .021 .021 .021 r
F P
0 g
s
.79464 .11504 m .7891 .11487 L .80017 .11522 L closepath p .955 .786 .906 r
F P
0 g
s
.77578 .11504 m .77661 .11548 L .78126 .11487 L closepath p .147 .044 .044 r
F P
0 g
s
.78206 .11531 m .78126 .11487 L .77661 .11548 L closepath p .147 .044 .044 r
F P
0 g
s
.78983 .11531 m .7891 .11487 L .7859 .11525 L closepath p .166 .041 .041 r
F P
0 g
s
.7891 .11487 m .78983 .11531 L .79464 .11504 L closepath p .174 .04 .04 r
F P
0 g
s
.48812 .11539 m .48809 .11583 L .48947 .11488 L closepath p .038 .038 .038 r
F P
0 g
s
.49088 .11526 m .48947 .11488 L .48809 .11583 L closepath p .096 .06 .06 r
F P
0 g
s
.80413 .11527 m .80494 .11572 L .80464 .11489 L closepath p .077 .077 .077 r
F P
0 g
s
.80718 .1148 m .80464 .11489 L .80494 .11572 L closepath p .079 .079 .079 r
F P
0 g
s
.44249 .11489 m .43699 .1151 L .43693 .11552 L p .132 .041 .041 r
F P
0 g
s
.43719 .11551 m .44249 .11489 L p .43693 .11552 L .132 .041 .041 r
F P
0 g
s
.4424 .11533 m .44249 .11489 L .43719 .11551 L p .132 .041 .041 r
F P
0 g
s
.43719 .11551 m .43693 .11554 L .4424 .11533 L p .132 .041 .041 r
F P
0 g
s
.4424 .11533 m .44152 .10407 L .44627 .1152 L p .587 .968 1 r
F P
0 g
s
.44627 .1152 m .44627 .1152 L p .4424 .11533 L .587 .968 1 r
F P
0 g
s
.44627 .1152 m .4463 .11527 L .44241 .11533 L p .587 .968 1 r
F P
0 g
s
.44241 .11533 m .4424 .11533 L p .44627 .1152 L .587 .968 1 r
F P
0 g
s
.43693 .11552 m .43693 .11554 L .43719 .11551 L p .132 .041 .041 r
F P
0 g
s
.34283 .11489 m .33737 .1151 L .33706 .11552 L p .128 .04 .04 r
F P
0 g
s
.33732 .11551 m .34283 .11489 L p .33706 .11552 L .128 .04 .04 r
F P
0 g
s
.34247 .11533 m .34283 .11489 L .33732 .11551 L p .128 .04 .04 r
F P
0 g
s
.33732 .11551 m .33705 .11554 L .34247 .11533 L p .128 .04 .04 r
F P
0 g
s
.34247 .11533 m .34007 .10407 L .34626 .1152 L p .588 .968 1 r
F P
0 g
s
.34626 .1152 m .34626 .1152 L p .34247 .11533 L .588 .968 1 r
F P
0 g
s
.34626 .1152 m .3463 .11527 L .34248 .11533 L p .588 .968 1 r
F P
0 g
s
.34248 .11533 m .34247 .11533 L p .34626 .1152 L .588 .968 1 r
F P
0 g
s
.33706 .11552 m .33705 .11554 L .33732 .11551 L p .128 .04 .04 r
F P
0 g
s
.39582 .11489 m .39035 .1151 L .39017 .11552 L p .13 .041 .041 r
F P
0 g
s
.39043 .11551 m .39582 .11489 L p .39017 .11552 L .13 .041 .041 r
F P
0 g
s
.3956 .11533 m .39582 .11489 L .39043 .11551 L p .13 .041 .041 r
F P
0 g
s
.39043 .11551 m .39016 .11554 L .3956 .11533 L p .13 .041 .041 r
F P
0 g
s
.3956 .11533 m .39401 .10407 L .39946 .1152 L p .587 .968 1 r
F P
0 g
s
.39946 .1152 m .39946 .1152 L p .3956 .11533 L .587 .968 1 r
F P
0 g
s
.39946 .1152 m .39949 .11527 L .39561 .11533 L p .587 .968 1 r
F P
0 g
s
.39561 .11533 m .3956 .11533 L p .39946 .1152 L .587 .968 1 r
F P
0 g
s
.39017 .11552 m .39016 .11554 L .39043 .11551 L p .13 .041 .041 r
F P
0 g
s
.36577 .11539 m .36541 .11583 L .36617 .115 L closepath p .007 .007 .007 r
F P
0 g
s
.36757 .11484 m .36617 .115 L .36541 .11583 L closepath p .042 .042 .042 r
F P
0 g
s
.41865 .11539 m .41843 .11583 L .41905 .115 L closepath p .016 .016 .016 r
F P
0 g
s
.42058 .11485 m .41905 .115 L .41843 .11583 L closepath p .047 .047 .047 r
F P
0 g
s
.77031 .11522 m .77108 .11566 L .77578 .11504 L closepath p .147 .044 .044 r
F P
0 g
s
.77661 .11548 m .77578 .11504 L .77108 .11566 L closepath p .147 .044 .044 r
F P
0 g
s
.79533 .11548 m .79464 .11504 L .78983 .11531 L closepath p .174 .04 .04 r
F P
0 g
s
.79464 .11504 m .79533 .11548 L .80017 .11522 L closepath p .174 .04 .04 r
F P
0 g
s
.50855 .11527 m .50857 .11572 L .51546 .11504 L closepath p .138 .042 .042 r
F P
0 g
s
.51562 .11548 m .51546 .11504 L .50857 .11572 L closepath p .138 .042 .042 r
F P
0 g
s
.52714 .11548 m .52725 .11504 L .52242 .11525 L closepath p .192 .042 .042 r
F P
0 g
s
.52725 .11504 m .52714 .11548 L .53213 .11527 L closepath p .192 .042 .042 r
F P
0 g
s
.58484 .11527 m .58507 .11572 L .59076 .11504 L closepath p .137 .043 .043 r
F P
0 g
s
.59114 .11548 m .59076 .11504 L .58507 .11572 L closepath p .137 .043 .043 r
F P
0 g
s
.60341 .11548 m .60326 .11504 L .59693 .11525 L closepath p .182 .041 .041 r
F P
0 g
s
.60326 .11504 m .60341 .11548 L .60985 .11527 L closepath p .182 .041 .041 r
F P
0 g
s
.43149 .11532 m .43137 .11576 L .43699 .1151 L closepath p .132 .041 .041 r
F P
0 g
s
.43693 .11554 m .43699 .1151 L .43137 .11576 L closepath p .132 .041 .041 r
F P
0 g
s
.46732 .1151 m .46718 .11555 L .47946 .11532 L closepath p .173 .042 .042 r
F P
0 g
s
.33192 .11532 m .33153 .11576 L .33737 .1151 L closepath p .128 .04 .04 r
F P
0 g
s
.33705 .11554 m .33737 .1151 L .33153 .11576 L closepath p .128 .04 .04 r
F P
0 g
s
.35624 .1151 m .35575 .11554 L .3618 .11532 L closepath p .192 .043 .043 r
F P
0 g
s
.38488 .11532 m .38463 .11576 L .39035 .1151 L closepath p .13 .041 .041 r
F P
0 g
s
.39016 .11554 m .39035 .1151 L .38463 .11576 L closepath p .13 .041 .041 r
F P
0 g
s
.4092 .1151 m .40885 .11554 L .4147 .11532 L closepath p .19 .043 .043 r
F P
0 g
s
.91776 .11581 m .92042 .11606 L .91861 .11512 L closepath p .196 .143 .143 r
F P
0 g
s
.92192 .11398 m .91861 .11512 L .92042 .11606 L closepath p .151 .151 .228 r
F P
0 g
s
.65333 .11581 m .6553 .11607 L .65423 .11512 L closepath p .116 .116 .116 r
F P
0 g
s
.65687 .11398 m .65423 .11512 L .6553 .11607 L closepath p .118 .118 .118 r
F P
0 g
s
.21968 .11581 m .22049 .11606 L .22057 .11512 L closepath p .064 .064 .064 r
F P
0 g
s
.2221 .11398 m .22057 .11512 L .22049 .11606 L closepath p .049 .049 .049 r
F P
0 g
s
.21879 .11651 m .21757 .11597 L .22057 .11512 L closepath p .955 .786 .906 r
F P
0 g
s
.83864 .11556 m .83828 .11522 L .8287 .11144 L closepath p .336 .018 .018 r
F P
0 g
s
.83828 .11522 m .83864 .11556 L .84822 .11651 L closepath p .199 .029 .029 r
F P
0 g
s
.76642 .11527 m .76713 .11572 L .77031 .11522 L closepath p .154 .043 .043 r
F P
0 g
s
.77108 .11566 m .77031 .11522 L .76713 .11572 L closepath p .154 .043 .043 r
F P
0 g
s
.80092 .11566 m .80017 .11522 L .79533 .11548 L closepath p .174 .04 .04 r
F P
0 g
s
.80017 .11522 m .80092 .11566 L .80413 .11527 L closepath p .166 .041 .041 r
F P
0 g
s
.76713 .11572 m .76642 .11527 L .7643 .1141 L closepath p .363 .003 .003 r
F P
0 g
s
.80494 .11572 m .80413 .11527 L .80092 .11566 L closepath p .167 .041 .041 r
F P
0 g
s
.50857 .11572 m .50855 .11527 L .50568 .11412 L closepath p .473 .036 .036 r
F P
0 g
s
.53221 .11572 m .53213 .11527 L .52714 .11548 L closepath p .191 .042 .042 r
F P
0 g
s
.58507 .11572 m .58484 .11527 L .5829 .11487 L closepath p .374 .032 .032 r
F P
0 g
s
.61015 .11572 m .60985 .11527 L .60341 .11548 L closepath p .181 .041 .041 r
F P
0 g
s
.42754 .11539 m .42735 .11583 L .43149 .11532 L closepath p .147 .042 .042 r
F P
0 g
s
.43137 .11576 m .43149 .11532 L .42735 .11583 L closepath p .147 .042 .042 r
F P
0 g
s
.47938 .11576 m .47946 .11532 L .46718 .11555 L closepath p .173 .042 .042 r
F P
0 g
s
.47946 .11532 m .47938 .11576 L .48812 .11539 L closepath p .166 .042 .042 r
F P
0 g
s
.32803 .11539 m .32757 .11583 L .33192 .11532 L closepath p .145 .041 .041 r
F P
0 g
s
.33153 .11576 m .33192 .11532 L .32757 .11583 L closepath p .145 .041 .041 r
F P
0 g
s
.36137 .11576 m .3618 .11532 L .35575 .11554 L closepath p .192 .043 .043 r
F P
0 g
s
.3618 .11532 m .36137 .11576 L .36577 .11539 L closepath p .175 .043 .043 r
F P
0 g
s
.38094 .11539 m .38062 .11583 L .38488 .11532 L closepath p .146 .041 .041 r
F P
0 g
s
.38463 .11576 m .38488 .11532 L .38062 .11583 L closepath p .146 .041 .041 r
F P
0 g
s
.41441 .11576 m .4147 .11532 L .40885 .11554 L closepath p .19 .043 .043 r
F P
0 g
s
.4147 .11532 m .41441 .11576 L .41865 .11539 L closepath p .174 .042 .042 r
F P
0 g
s
.42735 .11583 m .42754 .11539 L .42465 .11418 L closepath p .533 .045 .045 r
F P
0 g
s
.48809 .11583 m .48812 .11539 L .47938 .11576 L closepath p .166 .042 .042 r
F P
0 g
s
.32757 .11583 m .32803 .11539 L .3248 .11421 L closepath p .56 .057 .057 r
F P
0 g
s
.36541 .11583 m .36577 .11539 L .36137 .11576 L closepath p .175 .043 .043 r
F P
0 g
s
.38062 .11583 m .38094 .11539 L .37779 .11422 L closepath p .533 .051 .051 r
F P
0 g
s
.41843 .11583 m .41865 .11539 L .41441 .11576 L closepath p .174 .042 .042 r
F P
0 g
s
.21358 .11473 m .21635 .11543 L .20666 .1011 L closepath p 1 .689 .333 r
F P
0 g
s
.21804 .11696 m .21879 .11651 L .21358 .11473 L closepath p .556 .067 .067 r
F P
0 g
s
.21879 .11651 m .21804 .11696 L .21968 .11581 L closepath p .477 .845 .596 r
F P
0 g
s
.22049 .11606 m .21968 .11581 L .21834 .11675 L p .048 .048 .048 r
F P
0 g
s
.21834 .11675 m .21804 .11696 L .21834 .11684 L p .048 .048 .048 r
F P
0 g
s
.21834 .11684 m .22049 .11606 L p .21834 .11675 L .048 .048 .048 r
F P
0 g
s
.21635 .11543 m .21358 .11473 L .21834 .11635 L p .454 .046 .046 r
F P
0 g
s
.21834 .11635 m .21879 .11651 L .21834 .11631 L p .454 .046 .046 r
F P
0 g
s
.21834 .11631 m .21635 .11543 L p .21834 .11635 L .454 .046 .046 r
F P
0 g
s
.91344 .11474 m .91441 .11543 L .90574 .10115 L closepath p .782 .362 .048 r
F P
0 g
s
.91441 .11543 m .91344 .11474 L .9169 .11651 L closepath p .024 .375 .736 r
F P
0 g
s
.64837 .11473 m .65003 .11543 L .64104 .1011 L closepath p .943 .525 .099 r
F P
0 g
s
.65003 .11543 m .64837 .11473 L .65242 .11651 L closepath p .308 .005 .005 r
F P
0 g
s
.07375 .1136 m .07717 .11545 L .07319 .11035 L closepath p 1 .899 .612 r
F P
0 g
s
.07717 .11545 m .07375 .1136 L .07597 .12064 L closepath p 1 1 .844 r
F P
0 g
s
.07597 .12064 m .07717 .11545 L .0814 .12786 L closepath p .955 .786 .906 r
F P
0 g
s
.94375 .11569 m .94348 .11547 L .93891 .11255 L closepath p .398 .004 .004 r
F P
0 g
s
.94348 .11547 m .94375 .11569 L .94826 .11635 L closepath p .19 .016 .016 r
F P
0 g
s
.94826 .11635 m .94945 .1168 L .95298 .11547 L closepath p .109 .061 .061 r
F P
0 g
s
.95508 .11569 m .95298 .11547 L .94945 .1168 L closepath p .09 .058 .058 r
F P
0 g
s
.94826 .11635 m .94348 .11547 L .95298 .11547 L closepath p .955 .786 .906 r
F P
0 g
s
.9169 .11651 m .91801 .11696 L .91776 .11581 L closepath p .092 .092 .092 r
F P
0 g
s
.92042 .11606 m .91776 .11581 L .91801 .11696 L closepath p .163 .1 .1 r
F P
0 g
s
.65242 .11651 m .65283 .11696 L .65333 .11581 L closepath p .056 .056 .056 r
F P
0 g
s
.6553 .11607 m .65333 .11581 L .65283 .11696 L closepath p .089 .081 .081 r
F P
0 g
s
.84822 .11651 m .84915 .11696 L .85712 .11601 L closepath p .141 .047 .047 r
F P
0 g
s
.85846 .11641 m .85712 .11601 L .84915 .11696 L closepath p .133 .045 .045 r
F P
0 g
s
.11669 .067 m .12244 .0788 L .10817 .04878 L closepath p .653 .289 .003 r
F P
0 g
s
.12244 .0788 m .11669 .067 L .13304 .10055 L closepath p 0 0 .31 r
F P
0 g
s
.3848 .06709 m .38653 .07888 L .38367 .06012 L closepath p .003 .003 .124 r
F P
0 g
s
.16636 .12198 m .16702 .11851 L .16362 .1162 L closepath p .159 .285 .817 r
F P
0 g
s
.16362 .1162 m .16461 .11957 L .14866 .12673 L closepath p .955 .786 .906 r
F P
0 g
s
.14866 .12673 m .15173 .11652 L .16362 .1162 L closepath p .955 .786 .906 r
F P
0 g
s
.09264 .12159 m .09396 .11999 L .09355 .11622 L closepath p .093 .093 .229 r
F P
0 g
s
.09486 .11466 m .09355 .11622 L .09396 .11999 L closepath p .093 .093 .229 r
F P
0 g
s
.0912 .12715 m .09085 .12757 L .09355 .11622 L closepath p .955 .786 .906 r
F P
0 g
s
.09355 .11622 m .09264 .12159 L .0912 .12715 L closepath p .955 .786 .906 r
F P
0 g
s
.09355 .11622 m .0912 .12715 L .09264 .12159 L closepath p .955 .786 .906 r
F P
0 g
s
.8235 .08818 m .82842 .07652 L .8187 .079 L closepath p .134 .211 .768 r
F P
0 g
s
.82355 .06721 m .8187 .079 L .82842 .07652 L closepath p .134 .211 .768 r
F P
0 g
s
.94945 .1168 m .94826 .11635 L .94375 .11569 L closepath p .211 .02 .02 r
F P
0 g
s
.25742 .11639 m .25566 .11757 L .25376 .11668 L closepath p .955 .786 .906 r
F P
0 g
s
.2545 .11796 m .25376 .11668 L .25566 .11757 L closepath p .955 .786 .906 r
F P
0 g
s
.2545 .11796 m .25566 .11757 L .25376 .11668 L closepath p .955 .786 .906 r
F P
0 g
s
.2545 .11796 m .25384 .11841 L .25742 .11639 L closepath p .009 .009 .009 r
F P
0 g
s
.25826 .11634 m .25742 .11639 L .25384 .11841 L closepath p .029 .029 .029 r
F P
0 g
s
.692 .11639 m .69034 .11758 L .68831 .1167 L closepath p .955 .786 .906 r
F P
0 g
s
.6892 .11797 m .68831 .1167 L .69034 .11758 L closepath p .955 .786 .906 r
F P
0 g
s
.6892 .11797 m .68971 .11842 L .692 .11639 L closepath p .06 .06 .06 r
F P
0 g
s
.69402 .1163 m .692 .11639 L .68971 .11842 L closepath p .074 .074 .074 r
F P
0 g
s
.75233 .11648 m .75294 .11639 L .74928 .1154 L p .325 .026 .026 r
F P
0 g
s
.75607 .11797 m .75501 .11758 L .75676 .11674 L closepath p .955 .786 .906 r
F P
0 g
s
.75294 .11639 m .75233 .11648 L .75607 .11797 L closepath p .441 .033 .033 r
F P
0 g
s
.75607 .11797 m .75675 .11842 L .75676 .11674 L closepath p .075 .075 .075 r
F P
0 g
s
.75233 .11648 m .75618 .10523 L .75673 .11802 L p .358 .731 1 r
F P
0 g
s
.75675 .11842 m .75607 .11797 L .75233 .11648 L closepath p .335 .013 .013 r
F P
0 g
s
.75673 .11802 m .75675 .11842 L .75233 .11648 L p .358 .731 1 r
F P
0 g
s
.74928 .1154 m .74873 .11525 L .75233 .11648 L p .325 .026 .026 r
F P
0 g
s
.91801 .11696 m .9169 .11651 L .91344 .11474 L closepath p .008 .294 .666 r
F P
0 g
s
.84915 .11696 m .84822 .11651 L .83864 .11556 L closepath p .209 .031 .031 r
F P
0 g
s
.65283 .11696 m .65242 .11651 L .64837 .11473 L closepath p .393 .021 .021 r
F P
0 g
s
.25146 .11642 m .25377 .11668 L .24937 .11183 L closepath p 1 .698 .32 r
F P
0 g
s
.25377 .11668 m .25146 .11642 L .2545 .11796 L closepath p 1 .626 .246 r
F P
0 g
s
.68717 .11645 m .68831 .1167 L .68488 .11188 L closepath p .908 .502 .083 r
F P
0 g
s
.68831 .1167 m .68717 .11645 L .6892 .11797 L closepath p .828 .365 .073 r
F P
0 g
s
.32136 .11796 m .32088 .11841 L .32209 .11673 L closepath p .675 .846 .424 r
F P
0 g
s
.32349 .1162 m .32209 .11673 L .32088 .11841 L closepath p .047 .047 .047 r
F P
0 g
s
.75929 .11621 m .75676 .11674 L .75675 .11842 L closepath p .108 .108 .108 r
F P
0 g
s
.94272 .11807 m .94242 .11783 L .93686 .1144 L closepath p .392 .005 .005 r
F P
0 g
s
.94242 .11783 m .94272 .11807 L .94834 .1189 L closepath p .193 .018 .018 r
F P
0 g
s
.94834 .1189 m .94242 .11783 L .95418 .11784 L closepath p .955 .786 .906 r
F P
0 g
s
.94834 .1189 m .94954 .11935 L .95418 .11784 L closepath p .113 .061 .061 r
F P
0 g
s
.95626 .11808 m .95418 .11784 L .94954 .11935 L closepath p .096 .058 .058 r
F P
0 g
s
.25384 .11841 m .2545 .11796 L .25146 .11642 L closepath p .697 .152 .07 r
F P
0 g
s
.68971 .11842 m .6892 .11797 L .68717 .11645 L closepath p .45 .005 .005 r
F P
0 g
s
.94954 .11935 m .94834 .1189 L .94272 .11807 L closepath p .212 .022 .022 r
F P
0 g
s
.1502 .12162 m .15059 .11967 L .15096 .11907 L closepath p .955 .786 .906 r
F P
0 g
s
.12853 .07042 m .1341 .08217 L .138 .08927 L closepath p .005 .005 .335 r
F P
0 g
s
.16636 .12198 m .14866 .12673 L .16461 .11957 L closepath p .955 .786 .906 r
F P
0 g
s
.15091 .11555 m .15028 .11967 L .14715 .1155 L closepath p .832 1 1 r
F P
0 g
s
.34836 .07088 m .35063 .08262 L .34956 .07947 L closepath p .737 .522 .01 r
F P
0 g
s
.53851 .09423 m .5391 .08266 L .54047 .08291 L closepath p .779 .735 .201 r
F P
0 g
s
.54109 .07117 m .54047 .08291 L .5391 .08266 L closepath p .779 .735 .201 r
F P
0 g
s
.77683 .0714 m .77268 .08313 L .77499 .06044 L closepath p .09 .09 .436 r
F P
0 g
s
.43186 .07142 m .43288 .08315 L .43023 .06001 L closepath p .01 .01 .138 r
F P
0 g
s
.07303 .11763 m .07597 .12064 L .07375 .1136 L closepath p 1 1 .805 r
F P
0 g
s
.07597 .12064 m .07303 .11763 L .07273 .12335 L closepath p 1 1 1 r
F P
0 g
s
.0814 .12786 m .07273 .12335 L .07597 .12064 L closepath p .955 .786 .906 r
F P
0 g
s
.0746 .1223 m .07597 .12064 L .07273 .12335 L closepath p .955 .786 .906 r
F P
0 g
s
.0746 .1223 m .07273 .12335 L .07597 .12064 L closepath p .955 .786 .906 r
F P
0 g
s
.51303 .07197 m .51284 .0837 L .50881 .05891 L closepath p .059 .059 .426 r
F P
0 g
s
.16353 .09655 m .1584 .08502 L .16243 .08377 L closepath p .922 .707 .171 r
F P
0 g
s
.15729 .07205 m .16243 .08377 L .1584 .08502 L closepath p .922 .707 .171 r
F P
0 g
s
.09156 .12672 m .0923 .12625 L .09264 .12159 L closepath p .059 .059 .075 r
F P
0 g
s
.09396 .11999 m .09264 .12159 L .0923 .12625 L closepath p .079 .079 .118 r
F P
0 g
s
.09156 .12672 m .0912 .12715 L .09264 .12159 L closepath p .955 .786 .906 r
F P
0 g
s
.09156 .12672 m .09264 .12159 L .0912 .12715 L closepath p .955 .786 .906 r
F P
0 g
s
.79133 .0946 m .79577 .08303 L .78952 .0844 L closepath p .095 .095 .459 r
F P
0 g
s
.79392 .07268 m .78952 .0844 L .79577 .08303 L closepath p .095 .095 .459 r
F P
0 g
s
.4754 .0873 m .47503 .07563 L .4767 .08463 L closepath p .922 .995 .594 r
F P
0 g
s
.47635 .07292 m .4767 .08463 L .47503 .07563 L closepath p .922 .995 .594 r
F P
0 g
s
.17209 .12391 m .17132 .11979 L .16636 .12198 L closepath p .673 .879 1 r
F P
0 g
s
.16702 .11851 m .16636 .12198 L .17132 .11979 L closepath p .673 .893 1 r
F P
0 g
s
.16636 .12198 m .17209 .12391 L .14866 .12673 L closepath p .955 .786 .906 r
F P
0 g
s
.08218 .08477 m .07582 .07306 L .07993 .09519 L closepath p 1 .987 .573 r
F P
0 g
s
.46863 .07342 m .4691 .08513 L .46544 .06538 L closepath p .06 .06 .474 r
F P
0 g
s
.4691 .08513 m .46863 .07342 L .47068 .08742 L closepath p .094 .154 .717 r
F P
0 g
s
.80908 .07376 m .80444 .08546 L .80428 .06348 L closepath p .127 .157 .724 r
F P
0 g
s
.80444 .08546 m .80908 .07376 L .81093 .09531 L closepath p .139 .299 .837 r
F P
0 g
s
.07087 .11938 m .07273 .12335 L .07303 .11763 L closepath p 1 1 1 r
F P
0 g
s
.07273 .12335 m .07087 .11938 L .0694 .12387 L closepath p .9 1 1 r
F P
0 g
s
.0814 .12786 m .08026 .12782 L .07273 .12335 L closepath p .955 .786 .906 r
F P
0 g
s
.0694 .12387 m .07273 .12335 L .08026 .12782 L closepath p .955 .786 .906 r
F P
0 g
s
.17846 .12424 m .17874 .1212 L .17734 .12376 L closepath p .549 .817 1 r
F P
0 g
s
.17658 .12194 m .17734 .12376 L .17874 .1212 L closepath p .787 1 1 r
F P
0 g
s
.17734 .12376 m .17658 .12194 L .1771 .12382 L closepath p .734 1 1 r
F P
0 g
s
.17734 .12376 m .1771 .12382 L .17658 .12194 L closepath p .734 1 1 r
F P
0 g
s
.17686 .12388 m .17734 .12376 L .17577 .12441 L closepath p .955 .786 .906 r
F P
0 g
s
.17944 .1249 m .17577 .12441 L .17734 .12376 L closepath p .955 .786 .906 r
F P
0 g
s
.1768 .12261 m .1771 .12382 L .17658 .12194 L closepath p .008 .008 .247 r
F P
0 g
s
.1771 .12382 m .1768 .12261 L .17686 .12388 L closepath p .579 .995 1 r
F P
0 g
s
.06804 .11978 m .0694 .12387 L .07087 .11938 L closepath p .896 1 1 r
F P
0 g
s
.0694 .12387 m .06804 .11978 L .06294 .12405 L closepath p .847 1 1 r
F P
0 g
s
.08026 .12782 m .07911 .12778 L .0694 .12387 L closepath p .955 .786 .906 r
F P
0 g
s
.06294 .12405 m .0694 .12387 L .07911 .12778 L closepath p .955 .786 .906 r
F P
0 g
s
.17602 .12176 m .17686 .12388 L .1768 .12261 L closepath p .076 .203 .736 r
F P
0 g
s
.17686 .12388 m .17602 .12176 L .17209 .12391 L closepath p .634 .983 1 r
F P
0 g
s
.17577 .12441 m .17209 .12391 L .17686 .12388 L closepath p .955 .786 .906 r
F P
0 g
s
.17132 .11979 m .17209 .12391 L .17602 .12176 L closepath p .627 .842 1 r
F P
0 g
s
.17944 .1249 m .14866 .12673 L .17209 .12391 L closepath p .955 .786 .906 r
F P
0 g
s
.87478 .07503 m .86916 .08671 L .87683 .07414 L closepath p .72 1 1 r
F P
0 g
s
.86916 .08671 m .87478 .07503 L .86801 .08852 L closepath p .689 .873 .468 r
F P
0 g
s
.06191 .11992 m .06294 .12405 L .06804 .11978 L closepath p .847 1 1 r
F P
0 g
s
.06294 .12405 m .06191 .11992 L .06224 .12467 L closepath p 1 1 1 r
F P
0 g
s
.07911 .12778 m .07582 .12786 L .06294 .12405 L closepath p .955 .786 .906 r
F P
0 g
s
.06252 .12577 m .06294 .12405 L .07582 .12786 L closepath p .955 .786 .906 r
F P
0 g
s
.87589 .08922 m .88162 .07757 L .87393 .08677 L closepath p .153 .41 .918 r
F P
0 g
s
.87962 .0751 m .87393 .08677 L .88162 .07757 L closepath p .153 .41 .918 r
F P
0 g
s
.17958 .12471 m .18042 .12173 L .17846 .12424 L closepath p .573 .825 1 r
F P
0 g
s
.17874 .1212 m .17846 .12424 L .18042 .12173 L closepath p .604 .87 1 r
F P
0 g
s
.90803 .07555 m .90191 .08722 L .90399 .06759 L closepath p .143 .238 .788 r
F P
0 g
s
.90191 .08722 m .90803 .07555 L .9051 .09314 L closepath p .145 .261 .806 r
F P
0 g
s
.4723 .08825 m .47188 .07659 L .4754 .0873 L closepath p .749 1 1 r
F P
0 g
s
.47503 .07563 m .4754 .0873 L .47188 .07659 L closepath p .749 1 1 r
F P
0 g
s
.05933 .12101 m .06224 .12467 L .06191 .11992 L closepath p .993 1 1 r
F P
0 g
s
.06224 .12467 m .05933 .12101 L .06155 .12529 L closepath p 1 1 1 r
F P
0 g
s
.47023 .07575 m .47068 .08742 L .46863 .07342 L closepath p .094 .154 .717 r
F P
0 g
s
.47068 .08742 m .47023 .07575 L .4723 .08825 L closepath p .258 .658 1 r
F P
0 g
s
.18029 .12695 m .18185 .12513 L .17958 .12471 L closepath p .156 .194 .737 r
F P
0 g
s
.18042 .12173 m .17958 .12471 L .18185 .12513 L closepath p .152 .227 .773 r
F P
0 g
s
.59773 .09564 m .59922 .08409 L .59676 .08745 L closepath p .044 .044 .242 r
F P
0 g
s
.59824 .07578 m .59676 .08745 L .59922 .08409 L closepath p .044 .044 .242 r
F P
0 g
s
.17944 .1249 m .17914 .12843 L .17339 .12638 L closepath p .955 .786 .906 r
F P
0 g
s
.17944 .1249 m .16734 .12786 L .15743 .12811 L closepath p .955 .786 .906 r
F P
0 g
s
.14866 .12673 m .17944 .1249 L .15743 .12811 L closepath p .955 .786 .906 r
F P
0 g
s
.71367 .0761 m .71046 .08777 L .7117 .06385 L closepath p .076 .076 .383 r
F P
0 g
s
.05809 .12346 m .06155 .12529 L .05933 .12101 L closepath p 1 1 .947 r
F P
0 g
s
.06155 .12529 m .05809 .12346 L .06193 .12764 L closepath p 1 .905 .624 r
F P
0 g
s
.82983 .09739 m .83486 .08587 L .8235 .08818 L closepath p .143 .325 .857 r
F P
0 g
s
.82842 .07652 m .8235 .08818 L .83486 .08587 L closepath p .143 .325 .857 r
F P
0 g
s
.47188 .07659 m .4723 .08825 L .47023 .07575 L closepath p .258 .658 1 r
F P
0 g
s
.07582 .12786 m .07119 .12811 L .06252 .12577 L closepath p .955 .786 .906 r
F P
0 g
s
.06211 .12749 m .06252 .12577 L .07119 .12811 L closepath p .955 .786 .906 r
F P
0 g
s
.87361 .07686 m .86801 .08852 L .87478 .07503 L closepath p .689 .873 .468 r
F P
0 g
s
.86801 .08852 m .87361 .07686 L .86677 .09411 L closepath p .046 .046 .046 r
F P
0 g
s
.17605 .12835 m .17339 .12638 L .17914 .12843 L closepath p .955 .786 .906 r
F P
0 g
s
.17339 .12638 m .17605 .12835 L .16734 .12786 L closepath p .955 .786 .906 r
F P
0 g
s
.87699 .09922 m .88273 .08772 L .87589 .08922 L closepath p .101 .101 .439 r
F P
0 g
s
.88273 .08772 m .87699 .09922 L .88303 .08957 L closepath p .108 .108 .494 r
F P
0 g
s
.88273 .08772 m .88303 .08957 L .88297 .0895 L p 1 .742 .246 r
F P
0 g
s
.88141 .078 m .87589 .08922 L .88273 .08772 L p .101 .101 .439 r
F P
0 g
s
.88297 .0895 m .88163 .08793 L .88273 .08772 L p 1 .742 .246 r
F P
0 g
s
.88273 .08772 m .88162 .07757 L .88141 .078 L p .101 .101 .439 r
F P
0 g
s
.09085 .12757 m .09089 .12819 L .09156 .12672 L closepath p .041 .041 .041 r
F P
0 g
s
.0923 .12625 m .09156 .12672 L .09089 .12819 L closepath p .032 .032 .032 r
F P
0 g
s
.14564 .12595 m .14866 .12673 L .13304 .10055 L closepath p 1 .711 .374 r
F P
0 g
s
.14866 .12673 m .14564 .12595 L .14963 .12758 L closepath p .845 .318 .116 r
F P
0 g
s
.15743 .12811 m .15346 .12835 L .14866 .12673 L closepath p .955 .786 .906 r
F P
0 g
s
.1506 .12843 m .14866 .12673 L .15346 .12835 L closepath p .955 .786 .906 r
F P
0 g
s
.17972 .12769 m .18078 .12767 L .18029 .12695 L closepath p .076 .076 .076 r
F P
0 g
s
.18185 .12513 m .18029 .12695 L .18078 .12767 L closepath p .068 .068 .068 r
F P
0 g
s
.07119 .12811 m .06657 .12835 L .06211 .12749 L closepath p .955 .786 .906 r
F P
0 g
s
.06328 .12843 m .06211 .12749 L .06657 .12835 L closepath p .955 .786 .906 r
F P
0 g
s
.08462 .12811 m .0833 .12857 L .09014 .12843 L closepath p .235 .051 .051 r
F P
0 g
s
.08904 .12891 m .09014 .12843 L .0833 .12857 L closepath p .234 .051 .051 r
F P
0 g
s
.09014 .12843 m .08904 .12891 L .09085 .12757 L closepath p .733 .964 .632 r
F P
0 g
s
.09089 .12819 m .09085 .12757 L .08904 .12891 L closepath p .055 .055 .055 r
F P
0 g
s
.14732 .128 m .14963 .12758 L .14564 .12595 L closepath p 1 .541 .21 r
F P
0 g
s
.14963 .12758 m .14732 .128 L .1506 .12843 L closepath p .911 .401 .141 r
F P
0 g
s
.94282 .07869 m .93618 .09032 L .94854 .07486 L closepath p .721 1 .888 r
F P
0 g
s
.93618 .09032 m .94282 .07869 L .93246 .0959 L closepath p .645 .848 .447 r
F P
0 g
s
.05909 .1272 m .06193 .12764 L .05809 .12346 L closepath p 1 .839 .514 r
F P
0 g
s
.06193 .12764 m .05909 .1272 L .06328 .12843 L closepath p .676 .102 .092 r
F P
0 g
s
.17605 .12835 m .1751 .12882 L .17827 .12851 L p .198 .047 .047 r
F P
0 g
s
.17827 .12851 m .17914 .12843 L .17827 .12841 L p .198 .047 .047 r
F P
0 g
s
.17827 .12841 m .17605 .12835 L p .17827 .12851 L .198 .047 .047 r
F P
0 g
s
.17829 .12891 m .17914 .12843 L .1751 .12882 L closepath p .198 .047 .047 r
F P
0 g
s
.17914 .12843 m .17829 .12891 L .17972 .12769 L closepath p .698 .923 .565 r
F P
0 g
s
.18078 .12767 m .17972 .12769 L .17829 .12891 L p .043 .043 .043 r
F P
0 g
s
.17829 .12891 m .17829 .12891 L .17829 .12891 L p .043 .043 .043 r
F P
0 g
s
.17829 .12891 m .18078 .12767 L p .17829 .12891 L .043 .043 .043 r
F P
0 g
s
.76016 .09108 m .76413 .07946 L .75823 .09046 L closepath p .433 .805 1 r
F P
0 g
s
.76216 .07883 m .75823 .09046 L .76413 .07946 L closepath p .433 .805 1 r
F P
0 g
s
.75823 .09046 m .76216 .07883 L .75553 .09212 L closepath p .777 1 .966 r
F P
0 g
s
.32526 .09112 m .3226 .07951 L .32338 .09048 L closepath p .336 .751 1 r
F P
0 g
s
.32069 .07885 m .32338 .09048 L .3226 .07951 L closepath p .336 .751 1 r
F P
0 g
s
.32338 .09048 m .32069 .07885 L .32064 .09214 L closepath p .881 1 1 r
F P
0 g
s
.24827 .0919 m .24444 .08029 L .24511 .09048 L closepath p .232 .662 1 r
F P
0 g
s
.24123 .07885 m .24511 .09048 L .24444 .08029 L closepath p .232 .662 1 r
F P
0 g
s
.24511 .09048 m .24123 .07885 L .24292 .09129 L closepath p .806 1 1 r
F P
0 g
s
.0833 .12857 m .08462 .12811 L .07799 .12826 L closepath p .235 .051 .051 r
F P
0 g
s
.0833 .12857 m .07799 .12826 L .07799 .12826 L p .519 .92 1 r
F P
0 g
s
.07911 .12778 m .07582 .12786 L .0749 .12828 L p .146 .042 .042 r
F P
0 g
s
.07523 .12828 m .07911 .12778 L p .0749 .12828 L .146 .042 .042 r
F P
0 g
s
.07799 .12826 m .07911 .12778 L .07477 .12834 L closepath p .146 .042 .042 r
F P
0 g
s
.07911 .12778 m .07799 .12826 L .08462 .12811 L closepath p .234 .051 .051 r
F P
0 g
s
.07303 .11763 m .06651 .10639 L .07087 .11938 L closepath p .975 1 1 r
F P
0 g
s
.07799 .12826 m .07154 .11718 L .08294 .12822 L p .519 .92 1 r
F P
0 g
s
.08294 .12822 m .0833 .12857 L p .07799 .12826 L .519 .92 1 r
F P
0 g
s
.0749 .12828 m .07477 .12834 L .07523 .12828 L p .146 .042 .042 r
F P
0 g
s
.16425 .12778 m .15743 .12811 L .16734 .12786 L closepath p .955 .786 .906 r
F P
0 g
s
.16425 .12778 m .16734 .12786 L .15743 .12811 L closepath p .955 .786 .906 r
F P
0 g
s
.16425 .12778 m .15743 .12811 L .15684 .12848 L p .126 .041 .041 r
F P
0 g
s
.15802 .12844 m .16425 .12778 L p .15684 .12848 L .126 .041 .041 r
F P
0 g
s
.16336 .12826 m .16425 .12778 L .15802 .12844 L p .125 .041 .041 r
F P
0 g
s
.15802 .12844 m .15669 .12858 L .16336 .12826 L p .125 .041 .041 r
F P
0 g
s
.16425 .12778 m .16336 .12826 L .16734 .12786 L closepath p .198 .047 .047 r
F P
0 g
s
.16637 .12834 m .16734 .12786 L .16351 .12825 L p .198 .047 .047 r
F P
0 g
s
.16351 .12825 m .16336 .12826 L .16637 .12834 L p .198 .047 .047 r
F P
0 g
s
.1717 .12811 m .16734 .12786 L .17605 .12835 L closepath p .955 .786 .906 r
F P
0 g
s
.16734 .12786 m .16637 .12834 L .1717 .12811 L closepath p .227 .05 .05 r
F P
0 g
s
.17067 .12858 m .1717 .12811 L .16637 .12834 L closepath p .228 .05 .05 r
F P
0 g
s
.16637 .12834 m .16127 .11726 L .17058 .12848 L p .525 .923 1 r
F P
0 g
s
.17058 .12848 m .17067 .12858 L .16637 .12834 L p .525 .923 1 r
F P
0 g
s
.16336 .12826 m .15822 .11718 L .16624 .12817 L p .549 .941 1 r
F P
0 g
s
.16624 .12817 m .16624 .12817 L p .16336 .12826 L .549 .941 1 r
F P
0 g
s
.16624 .12817 m .16637 .12834 L .16338 .12826 L p .549 .941 1 r
F P
0 g
s
.16338 .12826 m .16336 .12826 L p .16624 .12817 L .549 .941 1 r
F P
0 g
s
.15684 .12848 m .15669 .12858 L .15802 .12844 L p .126 .041 .041 r
F P
0 g
s
.07582 .12786 m .07119 .12811 L .07025 .12855 L p .116 .04 .04 r
F P
0 g
s
.0705 .12854 m .07582 .12786 L p .07025 .12855 L .116 .04 .04 r
F P
0 g
s
.07477 .12834 m .07582 .12786 L .0705 .12854 L p .116 .04 .04 r
F P
0 g
s
.0705 .12854 m .0702 .12858 L .07477 .12834 L p .116 .04 .04 r
F P
0 g
s
.07087 .11938 m .06433 .10817 L .06804 .11978 L closepath p .682 1 1 r
F P
0 g
s
.07477 .12834 m .06828 .11726 L .07793 .12819 L p .589 .97 1 r
F P
0 g
s
.07793 .12819 m .07793 .12819 L p .07477 .12834 L .589 .97 1 r
F P
0 g
s
.07793 .12819 m .07799 .12826 L .07479 .12834 L p .589 .97 1 r
F P
0 g
s
.07479 .12834 m .07477 .12834 L p .07793 .12819 L .589 .97 1 r
F P
0 g
s
.07025 .12855 m .0702 .12858 L .0705 .12854 L p .116 .04 .04 r
F P
0 g
s
.68286 .07899 m .68012 .09061 L .68611 .08041 L closepath p .356 .73 1 r
F P
0 g
s
.68012 .09061 m .68286 .07899 L .67795 .09142 L closepath p .741 1 1 r
F P
0 g
s
.06657 .12835 m .0655 .12883 L .07119 .12811 L closepath p .116 .04 .04 r
F P
0 g
s
.0702 .12858 m .07119 .12811 L .0655 .12883 L closepath p .117 .04 .04 r
F P
0 g
s
.1506 .12843 m .14967 .12891 L .15743 .12811 L closepath p .125 .041 .041 r
F P
0 g
s
.15669 .12858 m .15743 .12811 L .14967 .12891 L closepath p .126 .041 .041 r
F P
0 g
s
.1717 .12811 m .17067 .12858 L .17605 .12835 L closepath p .228 .05 .05 r
F P
0 g
s
.57067 .0792 m .56961 .09082 L .5607 .05995 L closepath p .094 .094 .646 r
F P
0 g
s
.56961 .09082 m .57067 .0792 L .58128 .11301 L closepath p .095 .095 .653 r
F P
0 g
s
.22496 .10243 m .22076 .09098 L .22708 .09106 L closepath p 1 .92 .457 r
F P
0 g
s
.22292 .07944 m .22708 .09106 L .22076 .09098 L closepath p 1 .92 .457 r
F P
0 g
s
.92558 .09192 m .93206 .08032 L .92569 .09106 L closepath p .074 .074 .182 r
F P
0 g
s
.93177 .08253 m .93206 .08032 L .93206 .08032 L p .075 .075 .184 r
F P
0 g
s
.93206 .08032 m .92558 .09192 L .93177 .08253 L p .075 .075 .184 r
F P
0 g
s
.93177 .08253 m .92529 .09411 L .93131 .08608 L closepath p .075 .075 .184 r
F P
0 g
s
.93177 .08253 m .93131 .08608 L .93131 .08608 L p 1 .96 .54 r
F P
0 g
s
.92529 .09411 m .93177 .08253 L .92558 .09192 L closepath p .075 .075 .184 r
F P
0 g
s
.93131 .08608 m .93003 .08377 L .93177 .08253 L p 1 .96 .54 r
F P
0 g
s
.93206 .08032 m .93177 .08253 L .93177 .08253 L p 1 .96 .541 r
F P
0 g
s
.93217 .07944 m .92569 .09106 L .93206 .08032 L p .074 .074 .182 r
F P
0 g
s
.93177 .08253 m .93049 .0802 L .93206 .08032 L p 1 .96 .541 r
F P
0 g
s
.93206 .08032 m .93217 .07944 L .93217 .07944 L p .074 .074 .182 r
F P
0 g
s
.66094 .09192 m .66339 .08032 L .66107 .09106 L closepath p .011 .011 .011 r
F P
0 g
s
.66306 .08253 m .66339 .08032 L .66339 .08032 L p .011 .011 .011 r
F P
0 g
s
.66339 .08032 m .66094 .09192 L .66306 .08253 L p .011 .011 .011 r
F P
0 g
s
.66253 .08608 m .66306 .08253 L .66306 .08253 L p .011 .011 .011 r
F P
0 g
s
.66306 .08253 m .66061 .09411 L .66253 .08608 L p .011 .011 .011 r
F P
0 g
s
.66253 .08608 m .66009 .09761 L .66181 .09097 L closepath p .011 .011 .011 r
F P
0 g
s
.66253 .08608 m .66181 .09097 L .66181 .09097 L p 1 1 .671 r
F P
0 g
s
.66352 .07944 m .66107 .09106 L .66339 .08032 L p .011 .011 .011 r
F P
0 g
s
.66061 .09411 m .66306 .08253 L .66094 .09192 L closepath p .011 .011 .011 r
F P
0 g
s
.66009 .09761 m .66253 .08608 L .66061 .09411 L closepath p .011 .011 .011 r
F P
0 g
s
.66181 .09097 m .65984 .08867 L .66253 .08608 L p 1 1 .671 r
F P
0 g
s
.66306 .08253 m .66253 .08608 L .66253 .08608 L p 1 1 .671 r
F P
0 g
s
.66253 .08608 m .66057 .08377 L .66306 .08253 L p 1 1 .671 r
F P
0 g
s
.66339 .08032 m .66306 .08253 L .66306 .08253 L p 1 1 .672 r
F P
0 g
s
.66306 .08253 m .6611 .0802 L .66339 .08032 L p 1 1 .672 r
F P
0 g
s
.66339 .08032 m .66352 .07944 L .66352 .07944 L p .011 .011 .011 r
F P
0 g
s
.06328 .12843 m .06211 .12891 L .06657 .12835 L closepath p .146 .043 .043 r
F P
0 g
s
.0655 .12883 m .06657 .12835 L .06211 .12891 L closepath p .146 .043 .043 r
F P
0 g
s
.1751 .12882 m .17605 .12835 L .17067 .12858 L closepath p .227 .049 .049 r
F P
0 g
s
.4944 .09201 m .49432 .0804 L .4916 .09107 L closepath p .377 .774 1 r
F P
0 g
s
.49148 .07946 m .4916 .09107 L .49432 .0804 L closepath p .377 .774 1 r
F P
0 g
s
.4916 .09107 m .49148 .07946 L .48876 .09259 L closepath p .826 1 1 r
F P
0 g
s
.76273 .093 m .76673 .08141 L .76016 .09108 L closepath p .236 .577 1 r
F P
0 g
s
.76413 .07946 m .76016 .09108 L .76673 .08141 L closepath p .236 .577 1 r
F P
0 g
s
.34956 .07947 m .35181 .09108 L .35045 .08562 L closepath p .733 .516 .009 r
F P
0 g
s
.32764 .09307 m .32501 .08148 L .32526 .09112 L closepath p .107 .413 .897 r
F P
0 g
s
.3226 .07951 m .32526 .09112 L .32501 .08148 L closepath p .107 .413 .897 r
F P
0 g
s
.06211 .12891 m .06328 .12843 L .05909 .1272 L closepath p .655 .088 .088 r
F P
0 g
s
.14967 .12891 m .1506 .12843 L .14732 .128 L closepath p .514 .071 .071 r
F P
0 g
s
.90666 .07953 m .90056 .09114 L .89721 .06113 L closepath p .143 .241 .79 r
F P
0 g
s
.91127 .08155 m .9051 .09314 L .90803 .07555 L closepath p .145 .261 .806 r
F P
0 g
s
.9051 .09314 m .91127 .08155 L .90756 .09712 L closepath p .149 .313 .847 r
F P
0 g
s
.90056 .09114 m .90666 .07953 L .90574 .10115 L closepath p .143 .244 .792 r
F P
0 g
s
.19749 .07953 m .20203 .09115 L .18892 .06114 L closepath p .013 .013 .345 r
F P
0 g
s
.63817 .07953 m .6361 .09115 L .62903 .06114 L closepath p .103 .103 .665 r
F P
0 g
s
.77818 .07964 m .77401 .09126 L .77683 .0714 L closepath p .089 .089 .433 r
F P
0 g
s
.239 .07967 m .24292 .09129 L .24123 .07885 L closepath p .806 1 1 r
F P
0 g
s
.43302 .07969 m .43402 .0913 L .43186 .07142 L closepath p .01 .01 .133 r
F P
0 g
s
.68066 .0798 m .67795 .09142 L .68286 .07899 L closepath p .741 1 1 r
F P
0 g
s
.67795 .09142 m .68066 .0798 L .67591 .09403 L closepath p .832 1 .7 r
F P
0 g
s
.25198 .09556 m .2482 .084 L .24827 .0919 L closepath p .085 .273 .79 r
F P
0 g
s
.24444 .08029 m .24827 .0919 L .2482 .084 L closepath p .085 .273 .79 r
F P
0 g
s
.49579 .09444 m .49572 .08287 L .4944 .09201 L closepath p .086 .086 .644 r
F P
0 g
s
.49432 .0804 m .4944 .09201 L .49572 .08287 L closepath p .086 .086 .644 r
F P
0 g
s
.68714 .09565 m .68999 .0841 L .68332 .09202 L closepath p .141 .461 .949 r
F P
0 g
s
.68611 .08041 m .68332 .09202 L .68999 .0841 L closepath p .141 .461 .949 r
F P
0 g
s
.93826 .0921 m .94493 .08049 L .93531 .09683 L closepath p .629 .819 .401 r
F P
0 g
s
.75942 .08052 m .75553 .09212 L .76216 .07883 L closepath p .777 1 .966 r
F P
0 g
s
.75553 .09212 m .75942 .08052 L .75324 .09604 L closepath p .757 .909 .498 r
F P
0 g
s
.3179 .08054 m .32064 .09214 L .32069 .07885 L closepath p .881 1 1 r
F P
0 g
s
.32064 .09214 m .3179 .08054 L .31823 .09604 L closepath p .997 1 .767 r
F P
0 g
s
.09667 .1054 m .09052 .09399 L .09939 .09221 L closepath p 1 .976 .554 r
F P
0 g
s
.09329 .08061 m .09939 .09221 L .09052 .09399 L closepath p 1 .976 .554 r
F P
0 g
s
.48859 .08099 m .48876 .09259 L .49148 .07946 L closepath p .826 1 1 r
F P
0 g
s
.48876 .09259 m .48859 .08099 L .48567 .09633 L closepath p .92 1 .812 r
F P
0 g
s
.76102 .10402 m .765 .09259 L .76273 .093 L closepath p .034 .034 .034 r
F P
0 g
s
.76673 .08141 m .76273 .093 L .765 .09259 L closepath p .034 .034 .034 r
F P
0 g
s
.32501 .08148 m .32764 .09307 L .32294 .09261 L closepath p .943 .874 .386 r
F P
0 g
s
.51603 .08161 m .51579 .0932 L .51303 .07197 L closepath p .057 .057 .412 r
F P
0 g
s
.23684 .08234 m .24079 .09391 L .239 .07967 L closepath p 1 1 .888 r
F P
0 g
s
.24079 .09391 m .23684 .08234 L .24477 .1009 L closepath p .036 .036 .48 r
F P
0 g
s
.67859 .08246 m .67591 .09403 L .68066 .0798 L closepath p .832 1 .7 r
F P
0 g
s
.67591 .09403 m .67859 .08246 L .67978 .10065 L closepath p .117 .182 .746 r
F P
0 g
s
.92484 .09761 m .93131 .08608 L .92529 .09411 L closepath p .075 .075 .184 r
F P
0 g
s
.87235 .08254 m .86677 .09411 L .87361 .07686 L closepath p .046 .046 .046 r
F P
0 g
s
.5372 .10247 m .53777 .09102 L .53851 .09423 L closepath p .772 .722 .183 r
F P
0 g
s
.5391 .08266 m .53851 .09423 L .53777 .09102 L closepath p .772 .722 .183 r
F P
0 g
s
.49549 .09765 m .49542 .08613 L .49579 .09444 L closepath p .772 .684 .131 r
F P
0 g
s
.49572 .08287 m .49579 .09444 L .49542 .08613 L closepath p .772 .684 .131 r
F P
0 g
s
.21242 .09454 m .20804 .08297 L .20937 .09454 L closepath p .586 .963 1 r
F P
0 g
s
.64652 .09454 m .64875 .08297 L .64362 .09454 L closepath p .586 .963 1 r
F P
0 g
s
.79246 .10051 m .79692 .08902 L .79133 .0946 L closepath p .096 .096 .473 r
F P
0 g
s
.79577 .08303 m .79133 .0946 L .79692 .08902 L closepath p .096 .096 .473 r
F P
0 g
s
.07354 .08363 m .07993 .09519 L .07582 .07306 L closepath p 1 .987 .573 r
F P
0 g
s
.81567 .08375 m .81093 .09531 L .80908 .07376 L closepath p .139 .299 .837 r
F P
0 g
s
.81093 .09531 m .81567 .08375 L .81943 .10467 L closepath p .149 .436 .936 r
F P
0 g
s
.25622 .10022 m .25251 .08873 L .25198 .09556 L closepath p .076 .203 .737 r
F P
0 g
s
.2482 .084 m .25198 .09556 L .25251 .08873 L closepath p .076 .203 .737 r
F P
0 g
s
.59832 .10066 m .59982 .08918 L .59773 .09564 L closepath p .044 .044 .243 r
F P
0 g
s
.59982 .08918 m .59832 .10066 L .60004 .09104 L closepath p .045 .045 .245 r
F P
0 g
s
.59982 .08918 m .60004 .09104 L .60004 .09104 L p 1 .864 .458 r
F P
0 g
s
.59922 .08409 m .59773 .09564 L .59982 .08918 L p .044 .044 .243 r
F P
0 g
s
.60004 .09104 m .59792 .08946 L .59982 .08918 L p 1 .864 .458 r
F P
0 g
s
.59982 .08918 m .59922 .08409 L .59922 .08409 L p .044 .044 .243 r
F P
0 g
s
.69152 .10027 m .69444 .08879 L .68714 .09565 L closepath p .137 .413 .917 r
F P
0 g
s
.68999 .0841 m .68714 .09565 L .69444 .08879 L closepath p .137 .413 .917 r
F P
0 g
s
.93905 .08435 m .93246 .0959 L .94282 .07869 L closepath p .645 .848 .447 r
F P
0 g
s
.93246 .0959 m .93905 .08435 L .93129 .10238 L closepath p .068 .068 .125 r
F P
0 g
s
.7571 .0845 m .75324 .09604 L .75942 .08052 L closepath p .757 .909 .498 r
F P
0 g
s
.75324 .09604 m .7571 .0845 L .75099 .10036 L closepath p .741 .872 .436 r
F P
0 g
s
.31546 .0845 m .31823 .09604 L .3179 .08054 L closepath p .997 1 .767 r
F P
0 g
s
.31823 .09604 m .31546 .0845 L .31579 .10036 L closepath p .999 1 .735 r
F P
0 g
s
.48545 .08479 m .48567 .09633 L .48859 .08099 L closepath p .92 1 .812 r
F P
0 g
s
.48567 .09633 m .48545 .08479 L .48231 .10086 L closepath p .924 1 .767 r
F P
0 g
s
.16434 .10584 m .15923 .09444 L .16353 .09655 L closepath p .922 .705 .169 r
F P
0 g
s
.1584 .08502 m .16353 .09655 L .15923 .09444 L closepath p .922 .705 .169 r
F P
0 g
s
.94194 .08529 m .93531 .09683 L .94493 .08049 L closepath p .629 .819 .401 r
F P
0 g
s
.93531 .09683 m .94194 .08529 L .93439 .10238 L closepath p .071 .071 .147 r
F P
0 g
s
.91377 .08559 m .90756 .09712 L .91127 .08155 L closepath p .149 .313 .847 r
F P
0 g
s
.90756 .09712 m .91377 .08559 L .90932 .09912 L closepath p .175 .455 .948 r
F P
0 g
s
.35061 .0864 m .3527 .09715 L .35101 .08909 L p .722 .498 .006 r
F P
0 g
s
.3532 .09441 m .35101 .08909 L .35178 .09305 L p .036 .036 .4 r
F P
0 g
s
.35178 .09305 m .35325 .10057 L .3532 .09441 L p .036 .036 .4 r
F P
0 g
s
.3532 .09441 m .35541 .10582 L .35763 .09591 L closepath p .336 .752 1 r
F P
0 g
s
.3532 .09441 m .35724 .09631 L .3539 .09508 L p .375 .035 .035 r
F P
0 g
s
.35541 .10582 m .3532 .09441 L .35325 .10057 L closepath p .036 .036 .4 r
F P
0 g
s
.3539 .09508 m .35131 .09412 L .3532 .09441 L p .375 .035 .035 r
F P
0 g
s
.35101 .08909 m .3532 .09441 L .35178 .09419 L p 1 .67 .242 r
F P
0 g
s
.35178 .09419 m .35131 .09412 L .35101 .08909 L p 1 .67 .242 r
F P
0 g
s
.35101 .08909 m .35045 .08562 L .35061 .0864 L p .722 .498 .006 r
F P
0 g
s
.83646 .10385 m .84159 .09242 L .82983 .09739 L closepath p .182 .491 .972 r
F P
0 g
s
.84159 .09242 m .849 .09654 L .8483 .09616 L p .376 .011 .011 r
F P
0 g
s
.83388 .08811 m .82983 .09739 L .84159 .09242 L p .182 .491 .972 r
F P
0 g
s
.8483 .09616 m .84123 .09236 L .84159 .09242 L p .376 .011 .011 r
F P
0 g
s
.84159 .09242 m .83486 .08587 L .83388 .08811 L p .182 .491 .972 r
F P
0 g
s
.43384 .08595 m .43483 .09748 L .43302 .07969 L closepath p .008 .008 .121 r
F P
0 g
s
.65938 .10243 m .66181 .09097 L .66009 .09761 L closepath p .011 .011 .011 r
F P
0 g
s
.92421 .10243 m .93068 .09097 L .92484 .09761 L closepath p .075 .075 .185 r
F P
0 g
s
.93131 .08608 m .92484 .09761 L .93068 .09097 L closepath p .075 .075 .185 r
F P
0 g
s
.49461 .10418 m .49453 .09275 L .49549 .09765 L closepath p .795 .727 .186 r
F P
0 g
s
.49542 .08613 m .49549 .09765 L .49453 .09275 L closepath p .795 .727 .186 r
F P
0 g
s
.45246 .10228 m .45174 .09083 L .4518 .09852 L closepath p .021 .021 .198 r
F P
0 g
s
.45174 .09083 m .45246 .10228 L .45244 .09331 L closepath p .038 .038 .327 r
F P
0 g
s
.45174 .09083 m .45244 .09331 L .4522 .09324 L p 1 .755 .328 r
F P
0 g
s
.45108 .08736 m .4518 .09852 L .45174 .09083 L p .021 .021 .198 r
F P
0 g
s
.4522 .09324 m .4505 .09276 L .45174 .09083 L p 1 .755 .328 r
F P
0 g
s
.45174 .09083 m .45106 .08701 L .45108 .08736 L p .021 .021 .198 r
F P
0 g
s
.91556 .08762 m .90932 .09912 L .91377 .08559 L closepath p .175 .455 .948 r
F P
0 g
s
.90932 .09912 m .91556 .08762 L .91221 .09912 L closepath p .584 .962 1 r
F P
0 g
s
.91849 .08762 m .91221 .09912 L .91556 .08762 L closepath p .584 .962 1 r
F P
0 g
s
.87728 .10104 m .88303 .08957 L .87699 .09922 L closepath p .108 .108 .494 r
F P
0 g
s
.25933 .10311 m .25566 .09167 L .25622 .10022 L closepath p .089 .308 .816 r
F P
0 g
s
.25251 .08873 m .25622 .10022 L .25566 .09167 L closepath p .089 .308 .816 r
F P
0 g
s
.69472 .10314 m .69769 .0917 L .69152 .10027 L closepath p .142 .485 .965 r
F P
0 g
s
.69444 .08879 m .69152 .10027 L .69769 .0917 L closepath p .142 .485 .965 r
F P
0 g
s
.31298 .08888 m .31579 .10036 L .31548 .08543 L p .999 1 .735 r
F P
0 g
s
.31579 .10036 m .31298 .08888 L .31337 .10331 L closepath p .98 1 .869 r
F P
0 g
s
.31053 .09187 m .31337 .10331 L .31301 .09007 L p .98 1 .869 r
F P
0 g
s
.31301 .09007 m .31298 .08888 L .31053 .09187 L p .98 1 .869 r
F P
0 g
s
.30616 .09463 m .30907 .10603 L .31033 .09378 L p .887 1 1 r
F P
0 g
s
.31033 .09378 m .31053 .09187 L .30616 .09463 L p .887 1 1 r
F P
0 g
s
.30907 .10603 m .30616 .09463 L .30361 .10688 L closepath p .682 1 1 r
F P
0 g
s
.30062 .09549 m .30361 .10688 L .30616 .09463 L closepath p .682 1 1 r
F P
0 g
s
.30616 .09463 m .30651 .09481 L .30355 .09516 L p .022 .022 .022 r
F P
0 g
s
.30355 .09516 m .30062 .09549 L .30616 .09463 L p .022 .022 .022 r
F P
0 g
s
.31053 .09187 m .31171 .09158 L .30843 .09338 L p .036 .036 .036 r
F P
0 g
s
.30843 .09338 m .30616 .09463 L .31053 .09187 L p .036 .036 .036 r
F P
0 g
s
.31298 .08888 m .31447 .08823 L .31139 .09106 L p .066 .066 .066 r
F P
0 g
s
.31139 .09106 m .31053 .09187 L .31298 .08888 L p .066 .066 .066 r
F P
0 g
s
.31548 .08543 m .31546 .0845 L .31298 .08888 L p .999 1 .735 r
F P
0 g
s
.75481 .08888 m .75099 .10036 L .75673 .08546 L p .741 .872 .436 r
F P
0 g
s
.75099 .10036 m .75481 .08888 L .74868 .10331 L closepath p .787 .992 .656 r
F P
0 g
s
.75247 .09187 m .74868 .10331 L .75431 .09006 L p .787 .992 .656 r
F P
0 g
s
.75431 .09006 m .75481 .08888 L .75247 .09187 L p .787 .992 .656 r
F P
0 g
s
.74868 .10331 m .75247 .09187 L .74452 .10603 L closepath p .779 1 .946 r
F P
0 g
s
.74825 .09463 m .74452 .10603 L .75139 .09379 L p .779 1 .946 r
F P
0 g
s
.75139 .09379 m .75247 .09187 L .74825 .09463 L p .779 1 .946 r
F P
0 g
s
.74452 .10603 m .74825 .09463 L .73912 .10688 L closepath p .649 1 1 r
F P
0 g
s
.74276 .09549 m .73912 .10688 L .7477 .09535 L p .649 1 1 r
F P
0 g
s
.7477 .09535 m .74825 .09463 L .74276 .09549 L p .649 1 1 r
F P
0 g
s
.73912 .10688 m .74276 .09549 L .73076 .10688 L closepath p .58 .961 1 r
F P
0 g
s
.73428 .09549 m .73076 .10688 L .73673 .10122 L p .58 .961 1 r
F P
0 g
s
.73954 .09856 m .74276 .09549 L .73499 .09549 L p .58 .961 1 r
F P
0 g
s
.73761 .09564 m .73428 .09549 L .74276 .09549 L p .14 .036 .036 r
F P
0 g
s
.72952 .10375 m .73302 .09232 L .72894 .1006 L closepath p .083 .083 .421 r
F P
0 g
s
.73001 .10567 m .73351 .09427 L .72952 .10375 L closepath p .092 .092 .499 r
F P
0 g
s
.73302 .09232 m .72952 .10375 L .73351 .09427 L closepath p .092 .092 .499 r
F P
0 g
s
.73351 .09427 m .73001 .10567 L .73428 .09549 L closepath p .126 .236 .789 r
F P
0 g
s
.73673 .10122 m .73954 .09856 L p .73428 .09549 L .58 .961 1 r
F P
0 g
s
.73499 .09549 m .73428 .09549 L p .73954 .09856 L .58 .961 1 r
F P
0 g
s
.73492 .09589 m .73428 .09549 L .74342 .09589 L closepath p .14 .036 .036 r
F P
0 g
s
.73492 .09589 m .73239 .09379 L .73257 .09387 L p .569 .032 .032 r
F P
0 g
s
.73257 .09387 m .73351 .09427 L .734 .09483 L p .569 .032 .032 r
F P
0 g
s
.73428 .09549 m .73492 .09589 L .734 .09483 L p .031 .031 .485 r
F P
0 g
s
.734 .09483 m .73351 .09427 L .73428 .09549 L p .031 .031 .485 r
F P
0 g
s
.734 .09483 m .73492 .09589 L p .73257 .09387 L .569 .032 .032 r
F P
0 g
s
.74276 .09549 m .74342 .09589 L .73761 .09564 L p .14 .036 .036 r
F P
0 g
s
.74825 .09463 m .7498 .09481 L .74628 .09515 L p .041 .039 .039 r
F P
0 g
s
.74628 .09515 m .74276 .09549 L .74825 .09463 L p .041 .039 .039 r
F P
0 g
s
.75247 .09187 m .75484 .09157 L .75093 .09339 L p .09 .09 .09 r
F P
0 g
s
.75093 .09339 m .74825 .09463 L .75247 .09187 L p .09 .09 .09 r
F P
0 g
s
.75481 .08888 m .75747 .08822 L .7536 .09104 L p .131 .131 .131 r
F P
0 g
s
.7536 .09104 m .75247 .09187 L .75481 .08888 L p .131 .131 .131 r
F P
0 g
s
.75673 .08546 m .7571 .0845 L .75481 .08888 L p .741 .872 .436 r
F P
0 g
s
.79475 .10573 m .79924 .09432 L .79246 .10051 L closepath p .122 .122 .695 r
F P
0 g
s
.79919 .10719 m .80375 .09581 L .79475 .10573 L closepath p .426 .798 1 r
F P
0 g
s
.79924 .09432 m .79475 .10573 L .80375 .09581 L closepath p .426 .798 1 r
F P
0 g
s
.80375 .09581 m .8037 .09593 L p .80411 .09592 L .576 .957 1 r
F P
0 g
s
.80375 .09581 m .80457 .09621 L .80129 .09504 L p .264 .011 .011 r
F P
0 g
s
.80129 .09504 m .79924 .09432 L .80375 .09581 L p .264 .011 .011 r
F P
0 g
s
.80711 .09584 m .80375 .09581 L p .80411 .09592 L .576 .957 1 r
F P
0 g
s
.8037 .09593 m .79919 .10719 L .80636 .09691 L p .576 .957 1 r
F P
0 g
s
.80636 .09691 m .80711 .09584 L p .8037 .09593 L .576 .957 1 r
F P
0 g
s
.80457 .09621 m .80113 .09497 L p .80031 .09468 L .003 .358 .677 r
F P
0 g
s
.79924 .09432 m .80129 .09504 L p .80031 .09468 L .003 .358 .677 r
F P
0 g
s
.80129 .09504 m .80457 .09621 L p .80031 .09468 L .003 .358 .677 r
F P
0 g
s
.79924 .09432 m .79692 .08902 L .79537 .093 L p .122 .122 .695 r
F P
0 g
s
.79924 .09432 m .79853 .09403 L .79692 .08902 L closepath p .817 .442 .048 r
F P
0 g
s
.80113 .09497 m .79853 .09403 L .79924 .09432 L p .003 .358 .677 r
F P
0 g
s
.79537 .093 m .79246 .10051 L .79924 .09432 L p .122 .122 .695 r
F P
0 g
s
.12745 .08907 m .13304 .10055 L .11669 .067 L closepath p 0 0 .31 r
F P
0 g
s
.13304 .10055 m .12745 .08907 L .14564 .12595 L closepath p .001 .001 .32 r
F P
0 g
s
.73243 .08912 m .72894 .1006 L .73302 .09232 L closepath p .083 .083 .421 r
F P
0 g
s
.68252 .08917 m .67978 .10065 L .67859 .08246 L closepath p .117 .182 .746 r
F P
0 g
s
.59854 .10249 m .60004 .09104 L .59832 .10066 L closepath p .045 .045 .245 r
F P
0 g
s
.48205 .08938 m .48231 .10086 L .48533 .0854 L p .924 1 .767 r
F P
0 g
s
.48231 .10086 m .48205 .08938 L .47968 .10378 L closepath p .915 1 .846 r
F P
0 g
s
.47937 .09235 m .47968 .10378 L .48192 .09017 L p .915 1 .846 r
F P
0 g
s
.48192 .09017 m .48205 .08938 L .47937 .09235 L p .915 1 .846 r
F P
0 g
s
.47968 .10378 m .47937 .09235 L .47595 .10654 L closepath p .872 1 .977 r
F P
0 g
s
.47558 .09515 m .47595 .10654 L .47894 .09412 L p .872 1 .977 r
F P
0 g
s
.47894 .09412 m .47937 .09235 L .47558 .09515 L p .872 1 .977 r
F P
0 g
s
.47595 .10654 m .47558 .09515 L .47226 .10738 L closepath p .704 1 1 r
F P
0 g
s
.47184 .09599 m .47226 .10738 L .47558 .09515 L closepath p .704 1 1 r
F P
0 g
s
.47558 .09515 m .47646 .09531 L .47372 .09572 L p .032 .032 .032 r
F P
0 g
s
.47372 .09572 m .47184 .09599 L .47558 .09515 L p .032 .032 .032 r
F P
0 g
s
.47937 .09235 m .481 .09207 L .47712 .09427 L p .066 .066 .066 r
F P
0 g
s
.47712 .09427 m .47558 .09515 L .47937 .09235 L p .066 .066 .066 r
F P
0 g
s
.48205 .08938 m .48386 .0889 L .47999 .09187 L p .088 .088 .088 r
F P
0 g
s
.47999 .09187 m .47937 .09235 L .48205 .08938 L p .088 .088 .088 r
F P
0 g
s
.48533 .0854 m .48545 .08479 L .48205 .08938 L p .924 1 .767 r
F P
0 g
s
.24088 .08942 m .24477 .1009 L .23684 .08234 L closepath p .036 .036 .48 r
F P
0 g
s
.88303 .08957 m .87728 .10104 L .88374 .09313 L closepath p .113 .113 .533 r
F P
0 g
s
.20219 .08963 m .20666 .1011 L .19749 .07953 L closepath p .013 .013 .344 r
F P
0 g
s
.20666 .1011 m .20219 .08963 L .21358 .11473 L closepath p .019 .019 .389 r
F P
0 g
s
.64319 .08963 m .64104 .1011 L .63817 .07953 L closepath p .103 .103 .665 r
F P
0 g
s
.64104 .1011 m .64319 .08963 L .64837 .11473 L closepath p .107 .121 .696 r
F P
0 g
s
.91192 .08968 m .90574 .10115 L .90666 .07953 L closepath p .143 .244 .792 r
F P
0 g
s
.90574 .10115 m .91192 .08968 L .91344 .11474 L closepath p .146 .279 .82 r
F P
0 g
s
.45315 .10473 m .45244 .09331 L .45246 .10228 L closepath p .038 .038 .327 r
F P
0 g
s
.93786 .09093 m .93129 .10238 L .93905 .08435 L closepath p .068 .068 .125 r
F P
0 g
s
.94101 .09093 m .93439 .10238 L .94194 .08529 L closepath p .071 .071 .147 r
F P
0 g
s
.93439 .10238 m .94101 .09093 L .93566 .10787 L closepath p .125 .125 .593 r
F P
0 g
s
.93129 .10238 m .93786 .09093 L .93284 .1089 L closepath p .125 .125 .598 r
F P
0 g
s
.96551 .10894 m .9726 .09758 L .96663 .10238 L closepath p .077 .077 .165 r
F P
0 g
s
.96262 .10787 m .96968 .09649 L .96353 .10238 L closepath p .077 .077 .171 r
F P
0 g
s
.92317 .10882 m .92963 .09746 L .92421 .10243 L closepath p .069 .069 .143 r
F P
0 g
s
.93068 .09097 m .92421 .10243 L .92963 .09746 L closepath p .069 .069 .143 r
F P
0 g
s
.65822 .10882 m .66063 .09746 L .65938 .10243 L closepath p .006 .006 .006 r
F P
0 g
s
.66181 .09097 m .65938 .10243 L .66063 .09746 L closepath p .006 .006 .006 r
F P
0 g
s
.2236 .10882 m .21939 .09746 L .22496 .10243 L closepath p 1 .937 .483 r
F P
0 g
s
.22076 .09098 m .22496 .10243 L .21939 .09746 L closepath p 1 .937 .483 r
F P
0 g
s
.53777 .09102 m .5372 .10247 L .53709 .09624 L closepath p .756 .691 .143 r
F P
0 g
s
.60017 .1062 m .6017 .0948 L .59854 .10249 L closepath p .09 .09 .595 r
F P
0 g
s
.60364 .10719 m .60522 .09581 L .60017 .1062 L closepath p .416 .807 1 r
F P
0 g
s
.6017 .0948 m .60017 .1062 L .60522 .09581 L closepath p .416 .807 1 r
F P
0 g
s
.60522 .09581 m .60364 .10719 L .60934 .0978 L closepath p .302 .689 1 r
F P
0 g
s
.6017 .0948 m .6055 .09621 L .60307 .09539 L p .268 .011 .011 r
F P
0 g
s
.59966 .09392 m .59854 .10249 L .6017 .0948 L p .09 .09 .595 r
F P
0 g
s
.60307 .09539 m .60049 .09452 L .6017 .0948 L p .268 .011 .011 r
F P
0 g
s
.6017 .0948 m .60004 .09104 L .59966 .09392 L p .09 .09 .595 r
F P
0 g
s
.25566 .09167 m .25933 .10311 L .26046 .09458 L closepath p .115 .538 .97 r
F P
0 g
s
.69954 .10599 m .70258 .09459 L .69472 .10314 L closepath p .276 .643 1 r
F P
0 g
s
.69769 .0917 m .69472 .10314 L .70258 .09459 L closepath p .276 .643 1 r
F P
0 g
s
.58216 .10356 m .58341 .09213 L .58588 .10356 L closepath p .582 .961 1 r
F P
0 g
s
.43448 .09251 m .43547 .10395 L .43436 .09069 L p .717 .546 .003 r
F P
0 g
s
.43547 .10395 m .43448 .09251 L .43518 .10591 L closepath p .85 .778 .253 r
F P
0 g
s
.43448 .09251 m .43672 .09057 L .43457 .09391 L p .145 .145 .445 r
F P
0 g
s
.43419 .0945 m .43518 .10591 L .43456 .09395 L p .85 .778 .253 r
F P
0 g
s
.43456 .09395 m .43448 .09251 L .43419 .0945 L p .85 .778 .253 r
F P
0 g
s
.43518 .10591 m .43419 .0945 L .43396 .10675 L closepath p .872 1 1 r
F P
0 g
s
.43295 .09535 m .43396 .10675 L .43418 .09503 L p .872 1 1 r
F P
0 g
s
.43418 .09503 m .43419 .0945 L .43295 .09535 L p .872 1 1 r
F P
0 g
s
.43396 .10675 m .43295 .09535 L .43228 .10706 L closepath p .689 1 1 r
F P
0 g
s
.43125 .09567 m .43228 .10706 L .43294 .09555 L p .689 1 1 r
F P
0 g
s
.43294 .09555 m .43295 .09535 L .43125 .09567 L p .689 1 1 r
F P
0 g
s
.43228 .10706 m .43125 .09567 L .42912 .10721 L closepath p .61 .984 1 r
F P
0 g
s
.42804 .09583 m .42912 .10721 L .43125 .09567 L closepath p .61 .984 1 r
F P
0 g
s
.43125 .09567 m .43145 .09603 L .42858 .09586 L p .096 .034 .034 r
F P
0 g
s
.42858 .09586 m .42804 .09583 L .43125 .09567 L p .096 .034 .034 r
F P
0 g
s
.43295 .09535 m .43362 .0956 L .43181 .09566 L p .034 .034 .034 r
F P
0 g
s
.43181 .09566 m .43125 .09567 L .43295 .09535 L p .034 .034 .034 r
F P
0 g
s
.43419 .0945 m .43593 .09379 L .43399 .09481 L p .037 .037 .037 r
F P
0 g
s
.43399 .09481 m .43295 .09535 L .43419 .0945 L p .037 .037 .037 r
F P
0 g
s
.43457 .09391 m .43419 .0945 L .43448 .09251 L p .145 .145 .445 r
F P
0 g
s
.43436 .09069 m .43433 .09022 L .43448 .09251 L p .717 .546 .003 r
F P
0 g
s
.76025 .11138 m .76422 .10005 L .76102 .10402 L closepath p .043 .043 .084 r
F P
0 g
s
.765 .09259 m .76102 .10402 L .76422 .10005 L closepath p .043 .043 .084 r
F P
0 g
s
.4936 .11069 m .4935 .09936 L .49461 .10418 L closepath p .806 .747 .214 r
F P
0 g
s
.49453 .09275 m .49461 .10418 L .4935 .09936 L closepath p .806 .747 .214 r
F P
0 g
s
.8243 .09324 m .81943 .10467 L .81567 .08375 L closepath p .149 .436 .936 r
F P
0 g
s
.81943 .10467 m .8243 .09324 L .8287 .11144 L closepath p .259 .595 1 r
F P
0 g
s
.45391 .09515 m .46132 .09618 L .4558 .09572 L p .201 .038 .038 r
F P
0 g
s
.45253 .09473 m .45315 .10473 L .45391 .09515 L p .098 .224 .769 r
F P
0 g
s
.4558 .09572 m .45346 .09553 L .45391 .09515 L p .201 .038 .038 r
F P
0 g
s
.45391 .09515 m .45244 .09331 L .45253 .09473 L p .098 .224 .769 r
F P
0 g
s
.09486 .11466 m .08868 .10338 L .09667 .1054 L closepath p 1 .97 .544 r
F P
0 g
s
.09052 .09399 m .09667 .1054 L .08868 .10338 L closepath p 1 .97 .544 r
F P
0 g
s
.68491 .09421 m .68213 .10562 L .68252 .08917 L closepath p .107 .107 .667 r
F P
0 g
s
.73076 .10688 m .73428 .09549 L .73001 .10567 L closepath p .126 .236 .789 r
F P
0 g
s
.35977 .10729 m .35763 .09591 L .35541 .10582 L closepath p .336 .752 1 r
F P
0 g
s
.41275 .1073 m .41142 .09592 L .40828 .10583 L closepath p .354 .764 1 r
F P
0 g
s
.16487 .11163 m .15976 .10031 L .16434 .10584 L closepath p .92 .702 .165 r
F P
0 g
s
.15923 .09444 m .16434 .10584 L .15976 .10031 L closepath p .92 .702 .165 r
F P
0 g
s
.70427 .10688 m .70739 .09549 L .69954 .10599 L closepath p .484 .869 1 r
F P
0 g
s
.70258 .09459 m .69954 .10599 L .70739 .09549 L closepath p .484 .869 1 r
F P
0 g
s
.2771 .10684 m .2737 .09545 L .27433 .10712 L closepath p .648 1 1 r
F P
0 g
s
.71247 .10684 m .71571 .09545 L .70976 .10712 L closepath p .627 1 1 r
F P
0 g
s
.30361 .10688 m .30062 .09549 L .29526 .10688 L closepath p .58 .961 1 r
F P
0 g
s
.70717 .107 m .71033 .09562 L .70427 .10688 L closepath p .56 .942 1 r
F P
0 g
s
.70739 .09549 m .70738 .09551 L p .71015 .09596 L .56 .942 1 r
F P
0 g
s
.70738 .09551 m .70427 .10688 L .71013 .09599 L p .56 .942 1 r
F P
0 g
s
.71013 .09599 m .71015 .09596 L p .70738 .09551 L .56 .942 1 r
F P
0 g
s
.70795 .09589 m .7023 .09481 L .70249 .09466 L p .226 .025 .025 r
F P
0 g
s
.70249 .09466 m .70258 .09459 L .70382 .09489 L p .226 .025 .025 r
F P
0 g
s
.70382 .09489 m .70795 .09589 L p .70249 .09466 L .226 .025 .025 r
F P
0 g
s
.70739 .09549 m .70795 .09589 L .70382 .09489 L p .23 .027 .027 r
F P
0 g
s
.70382 .09489 m .70258 .09459 L .70739 .09549 L p .23 .027 .027 r
F P
0 g
s
.71015 .09596 m .71033 .09562 L .70739 .09549 L p .56 .942 1 r
F P
0 g
s
.70976 .10712 m .71296 .09573 L .70717 .107 L closepath p .559 .941 1 r
F P
0 g
s
.71033 .09562 m .70717 .107 L .71296 .09573 L closepath p .559 .941 1 r
F P
0 g
s
.27088 .09573 m .27433 .10712 L .2737 .09545 L closepath p .648 1 1 r
F P
0 g
s
.71296 .09573 m .70976 .10712 L .71554 .09578 L p .627 1 1 r
F P
0 g
s
.71296 .09573 m .71354 .09613 L .71078 .09569 L p .161 .034 .034 r
F P
0 g
s
.71078 .09569 m .71033 .09562 L .71296 .09573 L p .161 .034 .034 r
F P
0 g
s
.71554 .09578 m .71571 .09545 L .71296 .09573 L p .627 1 1 r
F P
0 g
s
.47226 .10738 m .47184 .09599 L .46214 .10717 L closepath p .567 .95 1 r
F P
0 g
s
.46157 .09578 m .46214 .10717 L .47184 .09599 L closepath p .567 .95 1 r
F P
0 g
s
.38588 .10717 m .38414 .09579 L .38236 .10729 L closepath p .602 .977 1 r
F P
0 g
s
.33295 .10718 m .3304 .09579 L .32935 .1073 L closepath p .602 .977 1 r
F P
0 g
s
.8025 .10722 m .80711 .09584 L .79919 .10719 L closepath p .576 .957 1 r
F P
0 g
s
.76872 .10719 m .77281 .09581 L .765 .10807 L closepath p .675 1 1 r
F P
0 g
s
.6077 .10916 m .60934 .0978 L .60364 .10719 L closepath p .302 .689 1 r
F P
0 g
s
.42912 .10721 m .42804 .09583 L .4255 .10811 L closepath p .719 1 1 r
F P
0 g
s
.8041 .10733 m .80873 .09595 L .8025 .10722 L closepath p .55 .931 1 r
F P
0 g
s
.80711 .09584 m .8025 .10722 L .80873 .09595 L closepath p .55 .931 1 r
F P
0 g
s
.36308 .10732 m .36099 .09594 L .35977 .10729 L closepath p .574 .956 1 r
F P
0 g
s
.35763 .09591 m .35977 .10729 L .36099 .09594 L closepath p .574 .956 1 r
F P
0 g
s
.38057 .09591 m .38236 .10729 L .38414 .09579 L closepath p .602 .977 1 r
F P
0 g
s
.38236 .10729 m .38057 .09591 L .37862 .10817 L closepath p .717 1 1 r
F P
0 g
s
.41607 .10733 m .41479 .09594 L .41275 .1073 L closepath p .574 .956 1 r
F P
0 g
s
.41142 .09592 m .41275 .1073 L .41479 .09594 L closepath p .574 .956 1 r
F P
0 g
s
.32675 .09592 m .32935 .1073 L .3304 .09579 L closepath p .602 .977 1 r
F P
0 g
s
.32935 .1073 m .32675 .09592 L .32566 .10818 L closepath p .725 1 1 r
F P
0 g
s
.36468 .10743 m .36262 .09605 L .36308 .10732 L closepath p .533 .924 1 r
F P
0 g
s
.36099 .09594 m .36308 .10732 L .36262 .09605 L closepath p .533 .924 1 r
F P
0 g
s
.41767 .10744 m .41641 .09606 L .41607 .10733 L closepath p .535 .924 1 r
F P
0 g
s
.41479 .09594 m .41607 .10733 L .41641 .09606 L closepath p .535 .924 1 r
F P
0 g
s
.80705 .10925 m .81173 .09789 L .8041 .10733 L closepath p .285 .632 1 r
F P
0 g
s
.80873 .09595 m .8041 .10733 L .81173 .09789 L closepath p .285 .632 1 r
F P
0 g
s
.50809 .10735 m .50821 .09596 L .50571 .10841 L closepath p .789 1 1 r
F P
0 g
s
.36752 .10936 m .3655 .09801 L .36468 .10743 L closepath p .119 .512 .962 r
F P
0 g
s
.36262 .09605 m .36468 .10743 L .3655 .09801 L closepath p .119 .512 .962 r
F P
0 g
s
.42051 .10937 m .4193 .09801 L .41767 .10744 L closepath p .124 .528 .975 r
F P
0 g
s
.41641 .09606 m .41767 .10744 L .4193 .09801 L closepath p .124 .528 .975 r
F P
0 g
s
.53519 .11327 m .53573 .10197 L .53653 .10762 L closepath p .809 .794 .282 r
F P
0 g
s
.53709 .09624 m .53653 .10762 L .53573 .10197 L closepath p .809 .794 .282 r
F P
0 g
s
.9597 .11254 m .96671 .10123 L .96262 .10787 L closepath p .603 .798 .378 r
F P
0 g
s
.96968 .09649 m .96262 .10787 L .96671 .10123 L closepath p .603 .798 .378 r
F P
0 g
s
.9423 .0965 m .93566 .10787 L .94101 .09093 L closepath p .125 .125 .593 r
F P
0 g
s
.93566 .10787 m .9423 .0965 L .93891 .11255 L closepath p .154 .368 .887 r
F P
0 g
s
.76904 .09669 m .765 .10807 L .77281 .09581 L closepath p .675 1 1 r
F P
0 g
s
.765 .10807 m .76904 .09669 L .76324 .11072 L closepath p .759 .936 .552 r
F P
0 g
s
.42437 .09674 m .4255 .10811 L .42804 .09583 L closepath p .719 1 1 r
F P
0 g
s
.4255 .10811 m .42437 .09674 L .42374 .11077 L closepath p .949 1 .745 r
F P
0 g
s
.37677 .0968 m .37862 .10817 L .38057 .09591 L closepath p .717 1 1 r
F P
0 g
s
.37862 .10817 m .37677 .0968 L .37679 .11082 L closepath p .967 1 .78 r
F P
0 g
s
.50579 .09704 m .50571 .10841 L .50821 .09596 L closepath p .789 1 1 r
F P
0 g
s
.50571 .10841 m .50579 .09704 L .50462 .11072 L closepath p .897 .972 .557 r
F P
0 g
s
.61197 .10905 m .61368 .09769 L .61035 .10858 L closepath p .416 .807 1 r
F P
0 g
s
.61203 .09721 m .61035 .10858 L .61368 .09769 L closepath p .416 .807 1 r
F P
0 g
s
.61035 .10858 m .61203 .09721 L .60827 .10932 L closepath p .741 1 1 r
F P
0 g
s
.92192 .11398 m .92836 .1027 L .92317 .10882 L closepath p .055 .055 .055 r
F P
0 g
s
.92963 .09746 m .92317 .10882 L .92836 .1027 L closepath p .055 .055 .055 r
F P
0 g
s
.65687 .11398 m .65926 .1027 L .65822 .10882 L closepath p .717 .732 .21 r
F P
0 g
s
.66063 .09746 m .65822 .10882 L .65926 .1027 L closepath p .717 .732 .21 r
F P
0 g
s
.2221 .11398 m .21786 .1027 L .2236 .10882 L closepath p 1 .986 .563 r
F P
0 g
s
.21939 .09746 m .2236 .10882 L .21786 .1027 L closepath p 1 .986 .563 r
F P
0 g
s
.93944 .09754 m .93284 .1089 L .93786 .09093 L closepath p .125 .125 .598 r
F P
0 g
s
.93284 .1089 m .93944 .09754 L .93686 .1144 L closepath p .155 .386 .9 r
F P
0 g
s
.96186 .11444 m .96891 .10316 L .96551 .10894 L closepath p .612 .818 .411 r
F P
0 g
s
.9726 .09758 m .96551 .10894 L .96891 .10316 L closepath p .612 .818 .411 r
F P
0 g
s
.61344 .11219 m .61517 .10088 L .61197 .10905 L closepath p .094 .094 .622 r
F P
0 g
s
.61368 .09769 m .61197 .10905 L .61517 .10088 L closepath p .094 .094 .622 r
F P
0 g
s
.60992 .09797 m .60827 .10932 L .61203 .09721 L closepath p .741 1 1 r
F P
0 g
s
.60827 .10932 m .60992 .09797 L .60842 .11001 L closepath p .062 .062 .371 r
F P
0 g
s
.61007 .09866 m .60842 .11001 L .60992 .09797 L closepath p .062 .062 .371 r
F P
0 g
s
.61007 .09866 m .60992 .09797 L .60842 .11001 L closepath p .062 .062 .371 r
F P
0 g
s
.60842 .11001 m .61007 .09866 L .6077 .10916 L closepath p .122 .326 .854 r
F P
0 g
s
.60934 .0978 m .6077 .10916 L .61007 .09866 L closepath p .122 .326 .854 r
F P
0 g
s
.80827 .11265 m .81297 .10135 L .80705 .10925 L closepath p .117 .117 .638 r
F P
0 g
s
.81173 .09789 m .80705 .10925 L .81297 .10135 L closepath p .117 .117 .638 r
F P
0 g
s
.3686 .11276 m .3666 .10146 L .36752 .10936 L closepath p .024 .024 .301 r
F P
0 g
s
.3655 .09801 m .36752 .10936 L .3666 .10146 L closepath p .024 .024 .301 r
F P
0 g
s
.42161 .11276 m .42041 .10146 L .42051 .10937 L closepath p .037 .037 .349 r
F P
0 g
s
.4193 .09801 m .42051 .10937 L .42041 .10146 L closepath p .037 .037 .349 r
F P
0 g
s
.87877 .11203 m .88454 .10072 L .87872 .11041 L closepath p .09 .09 .345 r
F P
0 g
s
.8845 .09907 m .87872 .11041 L .88454 .10072 L closepath p .09 .09 .345 r
F P
0 g
s
.49286 .11364 m .49275 .10235 L .4936 .11069 L closepath p .848 .831 .331 r
F P
0 g
s
.4935 .09936 m .4936 .11069 L .49275 .10235 L closepath p .848 .831 .331 r
F P
0 g
s
.76725 .09939 m .76324 .11072 L .76904 .09669 L closepath p .759 .936 .552 r
F P
0 g
s
.76324 .11072 m .76725 .09939 L .7643 .1141 L closepath p .105 .105 .576 r
F P
0 g
s
.50469 .09939 m .50462 .11072 L .50579 .09704 L closepath p .897 .972 .557 r
F P
0 g
s
.50462 .11072 m .50469 .09939 L .50568 .11412 L closepath p .054 .054 .403 r
F P
0 g
s
.42258 .09944 m .42374 .11077 L .42437 .09674 L closepath p .949 1 .745 r
F P
0 g
s
.42374 .11077 m .42258 .09944 L .42465 .11418 L closepath p .028 .028 .284 r
F P
0 g
s
.37491 .09949 m .37679 .11082 L .37677 .0968 L closepath p .967 1 .78 r
F P
0 g
s
.37679 .11082 m .37491 .09949 L .37779 .11422 L closepath p .022 .022 .28 r
F P
0 g
s
.94997 .11301 m .95684 .10171 L .95617 .11107 L closepath p .668 1 1 r
F P
0 g
s
.75929 .11621 m .76324 .10495 L .76025 .11138 L closepath p .027 .027 .027 r
F P
0 g
s
.76422 .10005 m .76025 .11138 L .76324 .10495 L closepath p .027 .027 .027 r
F P
0 g
s
.83371 .10012 m .8287 .11144 L .8243 .09324 L closepath p .259 .595 1 r
F P
0 g
s
.8287 .11144 m .83371 .10012 L .83864 .11556 L closepath p .391 .755 1 r
F P
0 g
s
.15976 .10031 m .16487 .11163 L .16 .10265 L closepath p .912 .689 .148 r
F P
0 g
s
.24555 .10051 m .24937 .11183 L .24088 .08942 L closepath p .014 .014 .324 r
F P
0 g
s
.24937 .11183 m .24555 .10051 L .25146 .11642 L closepath p .019 .019 .362 r
F P
0 g
s
.6877 .10056 m .68488 .11188 L .68491 .09421 L closepath p .104 .104 .64 r
F P
0 g
s
.68488 .11188 m .6877 .10056 L .68717 .11645 L closepath p .11 .11 .687 r
F P
0 g
s
.87712 .11453 m .88287 .10325 L .87877 .11203 L closepath p .676 .871 .472 r
F P
0 g
s
.88454 .10072 m .87877 .11203 L .88287 .10325 L closepath p .676 .871 .472 r
F P
0 g
s
.87097 .11229 m .87663 .10098 L .87155 .11209 L closepath p .692 1 1 r
F P
0 g
s
.87722 .10077 m .87155 .11209 L .87663 .10098 L closepath p .692 1 1 r
F P
0 g
s
.87155 .11209 m .87722 .10077 L .87269 .11255 L closepath p .403 .763 1 r
F P
0 g
s
.61248 .11455 m .61419 .10327 L .61344 .11219 L closepath p .809 .876 .415 r
F P
0 g
s
.61517 .10088 m .61344 .11219 L .61419 .10327 L closepath p .809 .876 .415 r
F P
0 g
s
.87663 .10098 m .87097 .11229 L .8749 .10179 L closepath p .713 1 1 r
F P
0 g
s
.95508 .11569 m .96203 .10442 L .9597 .11254 L closepath p .699 1 .862 r
F P
0 g
s
.96671 .10123 m .9597 .11254 L .96203 .10442 L closepath p .699 1 .862 r
F P
0 g
s
.9456 .10124 m .93891 .11255 L .9423 .0965 L closepath p .154 .368 .887 r
F P
0 g
s
.93891 .11255 m .9456 .10124 L .94375 .11569 L closepath p .327 .653 1 r
F P
0 g
s
.87838 .10124 m .87269 .11255 L .87722 .10077 L closepath p .403 .763 1 r
F P
0 g
s
.80718 .1148 m .81186 .10353 L .80827 .11265 L closepath p .687 .823 .374 r
F P
0 g
s
.81297 .10135 m .80827 .11265 L .81186 .10353 L closepath p .687 .823 .374 r
F P
0 g
s
.36757 .11484 m .36555 .10357 L .3686 .11276 L closepath p .972 1 .656 r
F P
0 g
s
.3666 .10146 m .3686 .11276 L .36555 .10357 L closepath p .972 1 .656 r
F P
0 g
s
.42058 .11485 m .41937 .10357 L .42161 .11276 L closepath p .946 1 .624 r
F P
0 g
s
.42041 .10146 m .42161 .11276 L .41937 .10357 L closepath p .946 1 .624 r
F P
0 g
s
.58252 .10171 m .58128 .11301 L .57067 .0792 L closepath p .095 .095 .653 r
F P
0 g
s
.58128 .11301 m .58252 .10171 L .5829 .11487 L closepath p .119 .322 .849 r
F P
0 g
s
.8749 .10179 m .86927 .11309 L .87241 .10295 L closepath p .713 1 1 r
F P
0 g
s
.53394 .11505 m .53445 .10378 L .53519 .11327 L closepath p .895 1 .717 r
F P
0 g
s
.53573 .10197 m .53519 .11327 L .53445 .10378 L closepath p .895 1 .717 r
F P
0 g
s
.49088 .11526 m .49075 .10399 L .49286 .11364 L closepath p .878 1 .942 r
F P
0 g
s
.49275 .10235 m .49286 .11364 L .49075 .10399 L closepath p .878 1 .942 r
F P
0 g
s
.16158 .10615 m .16511 .11394 L .16194 .10729 L p .674 .329 .006 r
F P
0 g
s
.16194 .10729 m .16702 .11851 L .16631 .10859 L closepath p .324 .758 1 r
F P
0 g
s
.16194 .10729 m .1654 .10902 L .16213 .10792 L p .403 .048 .048 r
F P
0 g
s
.16213 .10792 m .15955 .10705 L .16194 .10729 L p .403 .048 .048 r
F P
0 g
s
.16194 .10729 m .16 .10265 L .16158 .10615 L p .674 .329 .006 r
F P
0 g
s
.92042 .11606 m .92683 .1048 L .92192 .11398 L closepath p .658 .876 .496 r
F P
0 g
s
.92836 .1027 m .92192 .11398 L .92683 .1048 L closepath p .658 .876 .496 r
F P
0 g
s
.6553 .11607 m .65767 .10481 L .65687 .11398 L closepath p .833 1 .685 r
F P
0 g
s
.65926 .1027 m .65687 .11398 L .65767 .10481 L closepath p .833 1 .685 r
F P
0 g
s
.22049 .11606 m .21623 .1048 L .2221 .11398 L closepath p 1 1 .875 r
F P
0 g
s
.21786 .1027 m .2221 .11398 L .21623 .1048 L closepath p 1 1 .875 r
F P
0 g
s
.76833 .10282 m .7643 .1141 L .76725 .09939 L closepath p .105 .105 .576 r
F P
0 g
s
.7643 .1141 m .76833 .10282 L .76713 .11572 L closepath p .302 .663 1 r
F P
0 g
s
.50577 .10284 m .50568 .11412 L .50469 .09939 L closepath p .054 .054 .403 r
F P
0 g
s
.50568 .11412 m .50577 .10284 L .50857 .11572 L closepath p .227 .628 1 r
F P
0 g
s
.4235 .10289 m .42465 .11418 L .42258 .09944 L closepath p .028 .028 .284 r
F P
0 g
s
.42465 .11418 m .4235 .10289 L .42735 .11583 L closepath p .159 .571 1 r
F P
0 g
s
.3248 .11421 m .32212 .10293 L .32757 .11583 L closepath p .141 .569 .991 r
F P
0 g
s
.37593 .10294 m .37779 .11422 L .37491 .09949 L closepath p .022 .022 .28 r
F P
0 g
s
.37779 .11422 m .37593 .10294 L .38062 .11583 L closepath p .171 .591 1 r
F P
0 g
s
.87241 .10295 m .86682 .11424 L .87129 .10348 L closepath p .714 1 1 r
F P
0 g
s
.94352 .10312 m .93686 .1144 L .93944 .09754 L closepath p .155 .386 .9 r
F P
0 g
s
.93686 .1144 m .94352 .10312 L .94272 .11807 L closepath p .333 .663 1 r
F P
0 g
s
.95626 .11808 m .96322 .10685 L .96186 .11444 L closepath p .697 1 .88 r
F P
0 g
s
.96891 .10316 m .96186 .11444 L .96322 .10685 L closepath p .697 1 .88 r
F P
0 g
s
.88287 .10325 m .87712 .11453 L .87838 .10124 L closepath p .386 .743 1 r
F P
0 g
s
.61015 .11572 m .61183 .10446 L .61248 .11455 L closepath p .781 1 1 r
F P
0 g
s
.61419 .10327 m .61248 .11455 L .61183 .10446 L closepath p .781 1 1 r
F P
0 g
s
.09396 .11999 m .08777 .10879 L .09486 .11466 L closepath p 1 .954 .518 r
F P
0 g
s
.08868 .10338 m .09486 .11466 L .08777 .10879 L closepath p 1 .954 .518 r
F P
0 g
s
.20921 .10346 m .21358 .11473 L .20219 .08963 L closepath p .019 .019 .389 r
F P
0 g
s
.21358 .11473 m .20921 .10346 L .21804 .11696 L closepath p .165 .608 1 r
F P
0 g
s
.65064 .10346 m .64837 .11473 L .64319 .08963 L closepath p .107 .121 .696 r
F P
0 g
s
.64837 .11473 m .65064 .10346 L .65283 .11696 L closepath p .302 .685 1 r
F P
0 g
s
.91975 .10347 m .91344 .11474 L .91192 .08968 L closepath p .146 .279 .82 r
F P
0 g
s
.91344 .11474 m .91975 .10347 L .91801 .11696 L closepath p .38 .728 1 r
F P
0 g
s
.85846 .11641 m .86393 .10516 L .86572 .11476 L closepath p .657 1 1 r
F P
0 g
s
.87129 .10348 m .86572 .11476 L .86393 .10516 L closepath p .657 1 1 r
F P
0 g
s
.80494 .11572 m .80959 .10446 L .80718 .1148 L closepath p .715 1 1 r
F P
0 g
s
.81186 .10353 m .80718 .1148 L .80959 .10446 L closepath p .715 1 1 r
F P
0 g
s
.36541 .11583 m .36336 .10457 L .36757 .11484 L closepath p .816 1 1 r
F P
0 g
s
.36555 .10357 m .36757 .11484 L .36336 .10457 L closepath p .816 1 1 r
F P
0 g
s
.41843 .11583 m .41719 .10457 L .42058 .11485 L closepath p .806 1 1 r
F P
0 g
s
.41937 .10357 m .42058 .11485 L .41719 .10457 L closepath p .806 1 1 r
F P
0 g
s
.58416 .1036 m .5829 .11487 L .58252 .10171 L closepath p .119 .322 .849 r
F P
0 g
s
.5829 .11487 m .58416 .1036 L .58507 .11572 L closepath p .347 .742 1 r
F P
0 g
s
.53221 .11572 m .53271 .10446 L .53394 .11505 L closepath p .761 1 1 r
F P
0 g
s
.53445 .10378 m .53394 .11505 L .53271 .10446 L closepath p .761 1 1 r
F P
0 g
s
.78206 .11531 m .78636 .10404 L .7859 .11525 L closepath p .582 .965 1 r
F P
0 g
s
.79027 .10399 m .7859 .11525 L .78636 .10404 L closepath p .582 .965 1 r
F P
0 g
s
.7859 .11525 m .79027 .10399 L .78983 .11531 L closepath p .569 .953 1 r
F P
0 g
s
.69999 .10399 m .69699 .11525 L .69698 .10505 L closepath p .721 1 1 r
F P
0 g
s
.69699 .11525 m .69999 .10399 L .74873 .11525 L closepath p .576 .959 1 r
F P
0 g
s
.75252 .10399 m .74873 .11525 L .69999 .10399 L closepath p .576 .959 1 r
F P
0 g
s
.74873 .11525 m .75252 .10399 L .75233 .11648 L closepath p .408 .787 1 r
F P
0 g
s
.51562 .11548 m .51586 .10422 L .52242 .11525 L closepath p .595 .975 1 r
F P
0 g
s
.52276 .10399 m .52242 .11525 L .51586 .10422 L closepath p .595 .975 1 r
F P
0 g
s
.52242 .11525 m .52276 .10399 L .52714 .11548 L closepath p .548 .935 1 r
F P
0 g
s
.59114 .11548 m .59253 .10422 L .59693 .11525 L closepath p .597 .977 1 r
F P
0 g
s
.59841 .10399 m .59693 .11525 L .59253 .10422 L closepath p .597 .977 1 r
F P
0 g
s
.59693 .11525 m .59841 .10399 L .60341 .11548 L closepath p .556 .942 1 r
F P
0 g
s
.48809 .11583 m .4879 .10457 L .49088 .11526 L closepath p .688 1 1 r
F P
0 g
s
.49075 .10399 m .49088 .11526 L .4879 .10457 L closepath p .688 1 1 r
F P
0 g
s
.44548 .104 m .4463 .11527 L .44152 .10407 L closepath p .587 .968 1 r
F P
0 g
s
.34396 .104 m .3463 .11527 L .34007 .10407 L closepath p .588 .968 1 r
F P
0 g
s
.39796 .104 m .39949 .11527 L .39401 .10407 L closepath p .587 .968 1 r
F P
0 g
s
.25826 .11634 m .25457 .10508 L .26138 .11527 L closepath p .783 1 1 r
F P
0 g
s
.25774 .104 m .26138 .11527 L .25457 .10508 L closepath p .783 1 1 r
F P
0 g
s
.31027 .104 m .31312 .11527 L .25774 .104 L closepath p .576 .959 1 r
F P
0 g
s
.77661 .11548 m .78082 .10422 L .78206 .11531 L closepath p .59 .972 1 r
F P
0 g
s
.78636 .10404 m .78206 .11531 L .78082 .10422 L closepath p .59 .972 1 r
F P
0 g
s
.79425 .10404 m .78983 .11531 L .79027 .10399 L closepath p .569 .953 1 r
F P
0 g
s
.78983 .11531 m .79425 .10404 L .79533 .11548 L closepath p .562 .945 1 r
F P
0 g
s
.43693 .11554 m .43597 .10428 L .4424 .11533 L closepath p .599 .977 1 r
F P
0 g
s
.44152 .10407 m .4424 .11533 L .43597 .10428 L closepath p .599 .977 1 r
F P
0 g
s
.33705 .11554 m .33456 .10428 L .34247 .11533 L closepath p .602 .978 1 r
F P
0 g
s
.34007 .10407 m .34247 .11533 L .33456 .10428 L closepath p .602 .978 1 r
F P
0 g
s
.34794 .10407 m .35022 .11533 L .34396 .104 L closepath p .564 .95 1 r
F P
0 g
s
.39016 .11554 m .38849 .10428 L .3956 .11533 L closepath p .6 .978 1 r
F P
0 g
s
.39401 .10407 m .3956 .11533 L .38849 .10428 L closepath p .6 .978 1 r
F P
0 g
s
.40191 .10407 m .40339 .11533 L .39796 .104 L closepath p .564 .95 1 r
F P
0 g
s
.45432 .10407 m .455 .11534 L .44548 .104 L closepath p .571 .955 1 r
F P
0 g
s
.52756 .10421 m .52714 .11548 L .52276 .10399 L closepath p .548 .935 1 r
F P
0 g
s
.52714 .11548 m .52756 .10421 L .53221 .11572 L closepath p .548 .936 1 r
F P
0 g
s
.58507 .11572 m .58637 .10446 L .59114 .11548 L closepath p .597 .977 1 r
F P
0 g
s
.59253 .10422 m .59114 .11548 L .58637 .10446 L closepath p .597 .977 1 r
F P
0 g
s
.50857 .11572 m .5087 .10446 L .51562 .11548 L closepath p .595 .975 1 r
F P
0 g
s
.51586 .10422 m .51562 .11548 L .5087 .10446 L closepath p .595 .975 1 r
F P
0 g
s
.60499 .10422 m .60341 .11548 L .59841 .10399 L closepath p .556 .942 1 r
F P
0 g
s
.60341 .11548 m .60499 .10422 L .61015 .11572 L closepath p .557 .942 1 r
F P
0 g
s
.77108 .11566 m .77521 .1044 L .77661 .11548 L closepath p .59 .972 1 r
F P
0 g
s
.78082 .10422 m .77661 .11548 L .77521 .1044 L closepath p .59 .972 1 r
F P
0 g
s
.79984 .10422 m .79533 .11548 L .79425 .10404 L closepath p .562 .945 1 r
F P
0 g
s
.79533 .11548 m .79984 .10422 L .80092 .11566 L closepath p .562 .945 1 r
F P
0 g
s
.14715 .1155 m .14177 .10424 L .15091 .11555 L closepath p .566 .952 1 r
F P
0 g
s
.43137 .11576 m .43032 .1045 L .43693 .11554 L closepath p .599 .977 1 r
F P
0 g
s
.43597 .10428 m .43693 .11554 L .43032 .1045 L closepath p .599 .977 1 r
F P
0 g
s
.33153 .11576 m .32896 .1045 L .33705 .11554 L closepath p .602 .978 1 r
F P
0 g
s
.33456 .10428 m .33705 .11554 L .32896 .1045 L closepath p .602 .978 1 r
F P
0 g
s
.35355 .10428 m .35575 .11554 L .34794 .10407 L closepath p .55 .939 1 r
F P
0 g
s
.35575 .11554 m .35355 .10428 L .36137 .11576 L closepath p .55 .939 1 r
F P
0 g
s
.38463 .11576 m .38287 .1045 L .39016 .11554 L closepath p .6 .978 1 r
F P
0 g
s
.38849 .10428 m .39016 .11554 L .38287 .1045 L closepath p .6 .978 1 r
F P
0 g
s
.40746 .10428 m .40885 .11554 L .40191 .10407 L closepath p .551 .939 1 r
F P
0 g
s
.40885 .11554 m .40746 .10428 L .41441 .11576 L closepath p .551 .939 1 r
F P
0 g
s
.46668 .10428 m .46718 .11555 L .45432 .10407 L closepath p .565 .95 1 r
F P
0 g
s
.46718 .11555 m .46668 .10428 L .47938 .11576 L closepath p .565 .95 1 r
F P
0 g
s
.84381 .1043 m .83864 .11556 L .83371 .10012 L closepath p .391 .755 1 r
F P
0 g
s
.83864 .11556 m .84381 .1043 L .84915 .11696 L closepath p .518 .9 1 r
F P
0 g
s
.76713 .11572 m .7712 .10446 L .77108 .11566 L closepath p .582 .965 1 r
F P
0 g
s
.77521 .1044 m .77108 .11566 L .7712 .10446 L closepath p .582 .965 1 r
F P
0 g
s
.80551 .1044 m .80092 .11566 L .79984 .10422 L closepath p .562 .945 1 r
F P
0 g
s
.80092 .11566 m .80551 .1044 L .80494 .11572 L closepath p .569 .953 1 r
F P
0 g
s
.94945 .1168 m .95631 .10555 L .95508 .11569 L closepath p .638 1 1 r
F P
0 g
s
.96203 .10442 m .95508 .11569 L .95631 .10555 L closepath p .638 1 1 r
F P
0 g
s
.95052 .10443 m .94375 .11569 L .9456 .10124 L closepath p .327 .653 1 r
F P
0 g
s
.94375 .11569 m .95052 .10443 L .94945 .1168 L closepath p .501 .873 1 r
F P
0 g
s
.7712 .10446 m .76713 .11572 L .76833 .10282 L closepath p .302 .663 1 r
F P
0 g
s
.80959 .10446 m .80494 .11572 L .80551 .1044 L closepath p .569 .953 1 r
F P
0 g
s
.5087 .10446 m .50857 .11572 L .50577 .10284 L closepath p .227 .628 1 r
F P
0 g
s
.53271 .10446 m .53221 .11572 L .52756 .10421 L closepath p .548 .936 1 r
F P
0 g
s
.58637 .10446 m .58507 .11572 L .58416 .1036 L closepath p .347 .742 1 r
F P
0 g
s
.61183 .10446 m .61015 .11572 L .60499 .10422 L closepath p .557 .942 1 r
F P
0 g
s
.42735 .11583 m .42624 .10457 L .43137 .11576 L closepath p .587 .968 1 r
F P
0 g
s
.43032 .1045 m .43137 .11576 L .42624 .10457 L closepath p .587 .968 1 r
F P
0 g
s
.32757 .11583 m .32494 .10457 L .33153 .11576 L closepath p .588 .968 1 r
F P
0 g
s
.32896 .1045 m .33153 .11576 L .32494 .10457 L closepath p .588 .968 1 r
F P
0 g
s
.35926 .1045 m .36137 .11576 L .35355 .10428 L closepath p .55 .939 1 r
F P
0 g
s
.36137 .11576 m .35926 .1045 L .36541 .11583 L closepath p .564 .949 1 r
F P
0 g
s
.38062 .11583 m .3788 .10457 L .38463 .11576 L closepath p .587 .968 1 r
F P
0 g
s
.38287 .1045 m .38463 .11576 L .3788 .10457 L closepath p .587 .968 1 r
F P
0 g
s
.4131 .1045 m .41441 .11576 L .40746 .10428 L closepath p .551 .939 1 r
F P
0 g
s
.41441 .11576 m .4131 .1045 L .41843 .11583 L closepath p .564 .95 1 r
F P
0 g
s
.47907 .1045 m .47938 .11576 L .46668 .10428 L closepath p .565 .95 1 r
F P
0 g
s
.47938 .11576 m .47907 .1045 L .48809 .11583 L closepath p .571 .955 1 r
F P
0 g
s
.42624 .10457 m .42735 .11583 L .4235 .10289 L closepath p .159 .571 1 r
F P
0 g
s
.4879 .10457 m .48809 .11583 L .47907 .1045 L closepath p .571 .955 1 r
F P
0 g
s
.32494 .10457 m .32757 .11583 L .32212 .10293 L closepath p .141 .569 .991 r
F P
0 g
s
.36336 .10457 m .36541 .11583 L .35926 .1045 L closepath p .564 .949 1 r
F P
0 g
s
.3788 .10457 m .38062 .11583 L .37593 .10294 L closepath p .171 .591 1 r
F P
0 g
s
.41719 .10457 m .41843 .11583 L .4131 .1045 L closepath p .564 .95 1 r
F P
0 g
s
.21804 .11696 m .21374 .10571 L .22049 .11606 L closepath p .8 1 1 r
F P
0 g
s
.21623 .1048 m .22049 .11606 L .21374 .10571 L closepath p .8 1 1 r
F P
0 g
s
.91801 .11696 m .92439 .10571 L .92042 .11606 L closepath p .683 1 1 r
F P
0 g
s
.92683 .1048 m .92042 .11606 L .92439 .10571 L closepath p .683 1 1 r
F P
0 g
s
.65283 .11696 m .65516 .10571 L .6553 .11607 L closepath p .73 1 1 r
F P
0 g
s
.65767 .10481 m .6553 .11607 L .65516 .10571 L closepath p .73 1 1 r
F P
0 g
s
.32088 .11841 m .31815 .10719 L .32349 .1162 L closepath p .93 1 .972 r
F P
0 g
s
.3208 .10494 m .32349 .1162 L .31815 .10719 L closepath p .93 1 .972 r
F P
0 g
s
.75675 .11842 m .76067 .1072 L .75929 .11621 L closepath p .784 1 .831 r
F P
0 g
s
.76324 .10495 m .75929 .11621 L .76067 .1072 L closepath p .784 1 .831 r
F P
0 g
s
.68971 .11842 m .6926 .1072 L .69402 .1163 L closepath p .759 1 1 r
F P
0 g
s
.69698 .10505 m .69402 .1163 L .6926 .1072 L closepath p .759 1 1 r
F P
0 g
s
.25384 .11841 m .25009 .10719 L .25826 .11634 L closepath p .838 1 1 r
F P
0 g
s
.25457 .10508 m .25826 .11634 L .25009 .10719 L closepath p .838 1 1 r
F P
0 g
s
.84915 .11696 m .85448 .10571 L .85846 .11641 L closepath p .598 .982 1 r
F P
0 g
s
.86393 .10516 m .85846 .11641 L .85448 .10571 L closepath p .598 .982 1 r
F P
0 g
s
.24767 .10517 m .25146 .11642 L .24555 .10051 L closepath p .019 .019 .362 r
F P
0 g
s
.25146 .11642 m .24767 .10517 L .25384 .11841 L closepath p .093 .363 .853 r
F P
0 g
s
.69002 .1052 m .68717 .11645 L .6877 .10056 L closepath p .11 .11 .687 r
F P
0 g
s
.68717 .11645 m .69002 .1052 L .68971 .11842 L closepath p .182 .539 .996 r
F P
0 g
s
.31379 .10521 m .31659 .11646 L .31027 .104 L closepath p .316 .739 1 r
F P
0 g
s
.75618 .10523 m .75233 .11648 L .75252 .10399 L closepath p .408 .787 1 r
F P
0 g
s
.95631 .10555 m .94945 .1168 L .95052 .10443 L closepath p .501 .873 1 r
F P
0 g
s
.21374 .10571 m .21804 .11696 L .20921 .10346 L closepath p .165 .608 1 r
F P
0 g
s
.92439 .10571 m .91801 .11696 L .91975 .10347 L closepath p .38 .728 1 r
F P
0 g
s
.85448 .10571 m .84915 .11696 L .84381 .1043 L closepath p .518 .9 1 r
F P
0 g
s
.65516 .10571 m .65283 .11696 L .65064 .10346 L closepath p .302 .685 1 r
F P
0 g
s
.94947 .10684 m .94272 .11807 L .94352 .10312 L closepath p .333 .663 1 r
F P
0 g
s
.94272 .11807 m .94947 .10684 L .94954 .11935 L closepath p .502 .876 1 r
F P
0 g
s
.94954 .11935 m .9564 .10814 L .95626 .11808 L closepath p .634 1 1 r
F P
0 g
s
.96322 .10685 m .95626 .11808 L .9564 .10814 L closepath p .634 1 1 r
F P
0 g
s
.25009 .10719 m .25384 .11841 L .24767 .10517 L closepath p .093 .363 .853 r
F P
0 g
s
.31815 .10719 m .32088 .11841 L .31379 .10521 L closepath p .232 .66 1 r
F P
0 g
s
.6926 .1072 m .68971 .11842 L .69002 .1052 L closepath p .182 .539 .996 r
F P
0 g
s
.76067 .1072 m .75675 .11842 L .75618 .10523 L closepath p .358 .731 1 r
F P
0 g
s
.17132 .11979 m .16631 .10859 L .16702 .11851 L closepath p .324 .758 1 r
F P
0 g
s
.9564 .10814 m .94954 .11935 L .94947 .10684 L closepath p .502 .876 1 r
F P
0 g
s
.06433 .10817 m .07087 .11938 L .06651 .10639 L closepath p .975 1 1 r
F P
0 g
s
.06732 .10568 m .06651 .10639 L .06855 .10039 L closepath p .057 .057 .057 r
F P
0 g
s
.06433 .10817 m .06732 .10568 L .06731 .10568 L p .001 .001 .001 r
F P
0 g
s
.06223 .10991 m .06804 .11978 L .06433 .10817 L p .682 1 1 r
F P
0 g
s
.06731 .10568 m .06401 .10844 L .06433 .10817 L p .001 .001 .001 r
F P
0 g
s
.06433 .10817 m .06145 .10858 L .06223 .10991 L p .682 1 1 r
F P
0 g
s
.06804 .11978 m .06145 .10858 L .06191 .11992 L closepath p .592 .971 1 r
F P
0 g
s
.17602 .12176 m .17108 .11059 L .17132 .11979 L closepath p .214 .66 1 r
F P
0 g
s
.16631 .10859 m .17132 .11979 L .17108 .11059 L closepath p .214 .66 1 r
F P
0 g
s
.05523 .10872 m .06191 .11992 L .06145 .10858 L closepath p .592 .971 1 r
F P
0 g
s
.06191 .11992 m .05523 .10872 L .05933 .12101 L closepath p .845 1 1 r
F P
0 g
s
.0923 .12625 m .08608 .11514 L .09396 .11999 L closepath p 1 1 .607 r
F P
0 g
s
.08777 .10879 m .09396 .11999 L .08608 .11514 L closepath p 1 1 .607 r
F P
0 g
s
.0526 .10982 m .05933 .12101 L .05523 .10872 L closepath p .845 1 1 r
F P
0 g
s
.05933 .12101 m .0526 .10982 L .05809 .12346 L closepath p 1 1 .795 r
F P
0 g
s
.18042 .12173 m .17554 .11055 L .17874 .1212 L closepath p .308 .744 1 r
F P
0 g
s
.17384 .11002 m .17874 .1212 L .17554 .11055 L closepath p .308 .744 1 r
F P
0 g
s
.17874 .1212 m .17384 .11002 L .17658 .12194 L closepath p .79 1 1 r
F P
0 g
s
.18185 .12513 m .17699 .114 L .18042 .12173 L closepath p .661 .316 .003 r
F P
0 g
s
.17554 .11055 m .18042 .12173 L .17699 .114 L closepath p .661 .316 .003 r
F P
0 g
s
.1768 .12261 m .17187 .11145 L .17602 .12176 L closepath p .061 .155 .688 r
F P
0 g
s
.17108 .11059 m .17602 .12176 L .17187 .11145 L closepath p .061 .155 .688 r
F P
0 g
s
.17164 .11076 m .17658 .12194 L .17384 .11002 L closepath p .79 1 1 r
F P
0 g
s
.17658 .12194 m .17164 .11076 L .1768 .12261 L closepath p .746 .433 .02 r
F P
0 g
s
.17187 .11145 m .1768 .12261 L .17164 .11076 L closepath p .746 .433 .02 r
F P
0 g
s
.17187 .11145 m .17164 .11076 L .1768 .12261 L closepath p .746 .433 .02 r
F P
0 g
s
.05134 .11231 m .05809 .12346 L .0526 .10982 L closepath p 1 1 .795 r
F P
0 g
s
.05809 .12346 m .05134 .11231 L .05909 .1272 L closepath p .876 .571 .056 r
F P
0 g
s
.18078 .12767 m .17591 .11658 L .18185 .12513 L closepath p 1 1 .689 r
F P
0 g
s
.17699 .114 m .18185 .12513 L .17591 .11658 L closepath p 1 1 .689 r
F P
0 g
s
.14023 .11483 m .14564 .12595 L .12745 .08907 L closepath p .001 .001 .32 r
F P
0 g
s
.14564 .12595 m .14023 .11483 L .14732 .128 L closepath p .045 .061 .607 r
F P
0 g
s
.09089 .12819 m .08465 .11711 L .0923 .12625 L closepath p 1 1 .89 r
F P
0 g
s
.08608 .11514 m .0923 .12625 L .08465 .11711 L closepath p 1 1 .89 r
F P
0 g
s
.05236 .1161 m .05909 .1272 L .05134 .11231 L closepath p .876 .571 .056 r
F P
0 g
s
.05909 .1272 m .05236 .1161 L .06211 .12891 L closepath p .095 .507 .929 r
F P
0 g
s
.17829 .12891 m .17337 .11784 L .18078 .12767 L closepath p .858 1 1 r
F P
0 g
s
.17591 .11658 m .18078 .12767 L .17337 .11784 L closepath p .858 1 1 r
F P
0 g
s
.14194 .11691 m .14732 .128 L .14023 .11483 L closepath p .045 .061 .607 r
F P
0 g
s
.14732 .128 m .14194 .11691 L .14967 .12891 L closepath p .23 .679 1 r
F P
0 g
s
.08904 .12891 m .08277 .11784 L .09089 .12819 L closepath p .823 1 1 r
F P
0 g
s
.08465 .11711 m .09089 .12819 L .08277 .11784 L closepath p .823 1 1 r
F P
0 g
s
.07154 .11718 m .07799 .12826 L .06828 .11726 L closepath p .589 .97 1 r
F P
0 g
s
.15669 .12858 m .15144 .1175 L .16336 .12826 L closepath p .605 .982 1 r
F P
0 g
s
.15822 .11718 m .16336 .12826 L .15144 .1175 L closepath p .605 .982 1 r
F P
0 g
s
.16127 .11726 m .16637 .12834 L .15822 .11718 L closepath p .549 .941 1 r
F P
0 g
s
.0702 .12858 m .06364 .1175 L .07477 .12834 L closepath p .611 .985 1 r
F P
0 g
s
.06828 .11726 m .07477 .12834 L .06364 .1175 L closepath p .611 .985 1 r
F P
0 g
s
.07694 .1175 m .0833 .12857 L .07154 .11718 L closepath p .519 .92 1 r
F P
0 g
s
.0833 .12857 m .07694 .1175 L .08904 .12891 L closepath p .52 .921 1 r
F P
0 g
s
.14967 .12891 m .14432 .11784 L .15669 .12858 L closepath p .605 .982 1 r
F P
0 g
s
.15144 .1175 m .15669 .12858 L .14432 .11784 L closepath p .605 .982 1 r
F P
0 g
s
.16564 .1175 m .17067 .12858 L .16127 .11726 L closepath p .525 .923 1 r
F P
0 g
s
.17067 .12858 m .16564 .1175 L .1751 .12882 L closepath p .525 .924 1 r
F P
0 g
s
.0655 .12883 m .05886 .11775 L .0702 .12858 L closepath p .611 .985 1 r
F P
0 g
s
.06364 .1175 m .0702 .12858 L .05886 .11775 L closepath p .611 .985 1 r
F P
0 g
s
.17014 .11775 m .1751 .12882 L .16564 .1175 L closepath p .525 .924 1 r
F P
0 g
s
.1751 .12882 m .17014 .11775 L .17829 .12891 L closepath p .548 .941 1 r
F P
0 g
s
.06211 .12891 m .05543 .11784 L .0655 .12883 L closepath p .589 .97 1 r
F P
0 g
s
.05886 .11775 m .0655 .12883 L .05543 .11784 L closepath p .589 .97 1 r
F P
0 g
s
.05543 .11784 m .06211 .12891 L .05236 .1161 L closepath p .095 .507 .929 r
F P
0 g
s
.08277 .11784 m .08904 .12891 L .07694 .1175 L closepath p .52 .921 1 r
F P
0 g
s
.14432 .11784 m .14967 .12891 L .14194 .11691 L closepath p .23 .679 1 r
F P
0 g
s
.17337 .11784 m .17829 .12891 L .17014 .11775 L closepath p .548 .941 1 r
F P
0 g
s
.84107 .00603 m .84236 .00584 L .83206 .00538 L closepath p .05 .619 .803 r
F P
0 g
s
.83296 .00553 m .83206 .00538 L .84236 .00584 L closepath p .067 .012 .012 r
F P
0 g
s
.83206 .00538 m .83296 .00553 L .82454 .0061 L closepath p .022 .022 .022 r
F P
0 g
s
.09512 .00652 m .09313 .00649 L .09315 .00538 L closepath p .648 .084 .084 r
F P
0 g
s
.09315 .00538 m .09206 .00553 L .09112 .00639 L closepath p .444 .879 .753 r
F P
0 g
s
.09206 .00553 m .09315 .00538 L .09313 .00649 L closepath p .726 .189 .074 r
F P
0 g
s
.54181 .00781 m .54042 .00739 L .53858 .00576 L closepath p .509 .036 .036 r
F P
0 g
s
.53868 .00592 m .53858 .00576 L .54042 .00739 L closepath p .475 .005 .005 r
F P
0 g
s
.53858 .00576 m .53868 .00592 L .5367 .00668 L closepath p .017 .017 .017 r
F P
0 g
s
.82588 .00618 m .82454 .0061 L .83296 .00553 L closepath p .019 .019 .019 r
F P
0 g
s
.82454 .0061 m .82588 .00618 L .81783 .00825 L closepath p .044 .044 .044 r
F P
0 g
s
.09117 .00626 m .09112 .00639 L .09206 .00553 L closepath p .476 .837 .575 r
F P
0 g
s
.09112 .00639 m .09117 .00626 L .08948 .00922 L closepath p .623 .799 .366 r
F P
0 g
s
.09761 .00979 m .09499 .00935 L .09512 .00652 L closepath p .985 .524 .157 r
F P
0 g
s
.09313 .00649 m .09512 .00652 L .09499 .00935 L closepath p .992 .55 .15 r
F P
0 g
s
.53793 .00654 m .5367 .00668 L .53868 .00592 L closepath p .062 .062 .062 r
F P
0 g
s
.5367 .00668 m .53793 .00654 L .53526 .0092 L closepath p .088 .088 .088 r
F P
0 g
s
.85728 .0083 m .85743 .00832 L .85007 .00668 L closepath p .002 .442 .696 r
F P
0 g
s
.85051 .00677 m .85007 .00668 L .85743 .00832 L closepath p .001 .44 .693 r
F P
0 g
s
.47983 .00715 m .48077 .00699 L .4819 .01013 L closepath p .839 .378 .077 r
F P
0 g
s
.48077 .00699 m .47983 .00715 L .4131 .00699 L closepath p .056 .015 .015 r
F P
0 g
s
.41286 .00715 m .4131 .00699 L .47983 .00715 L closepath p .056 .015 .015 r
F P
0 g
s
.4131 .00699 m .41286 .00715 L .41098 .00788 L closepath p .005 .005 .005 r
F P
0 g
s
.40676 .00858 m .40476 .00779 L .40406 .00699 L closepath p .471 .031 .031 r
F P
0 g
s
.4038 .00715 m .40406 .00699 L .40476 .00779 L closepath p .486 .026 .026 r
F P
0 g
s
.40406 .00699 m .4038 .00715 L .36595 .00699 L closepath p .056 .015 .015 r
F P
0 g
s
.36559 .00715 m .36595 .00699 L .4038 .00715 L closepath p .056 .015 .015 r
F P
0 g
s
.36595 .00699 m .36559 .00715 L .3638 .00786 L closepath p .001 .001 .001 r
F P
0 g
s
.35338 .00857 m .35125 .00779 L .35064 .00699 L closepath p .482 .035 .035 r
F P
0 g
s
.35024 .00715 m .35064 .00699 L .35125 .00779 L closepath p .493 .032 .032 r
F P
0 g
s
.35064 .00699 m .35024 .00715 L .31253 .00699 L closepath p .056 .015 .015 r
F P
0 g
s
.31253 .00699 m .31202 .00715 L .31041 .00788 L closepath p .292 .772 .67 r
F P
0 g
s
.31202 .00715 m .31253 .00699 L .35024 .00715 L closepath p .056 .015 .015 r
F P
0 g
s
.29354 .00858 m .29124 .00779 L .29085 .00699 L closepath p .521 .045 .045 r
F P
0 g
s
.29028 .00715 m .29085 .00699 L .29124 .00779 L closepath p .541 .041 .041 r
F P
0 g
s
.29085 .00699 m .29028 .00715 L .25249 .00699 L closepath p .056 .015 .015 r
F P
0 g
s
.25249 .00699 m .25182 .00715 L .25036 .00788 L closepath p .314 .79 .692 r
F P
0 g
s
.25182 .00715 m .25249 .00699 L .29028 .00715 L closepath p .056 .015 .015 r
F P
0 g
s
.24769 .00848 m .24536 .00775 L .24479 .00699 L closepath p .462 .038 .038 r
F P
0 g
s
.2441 .00715 m .24479 .00699 L .24536 .00775 L closepath p .452 .04 .04 r
F P
0 g
s
.24479 .00699 m .2441 .00715 L .20834 .00699 L closepath p .056 .015 .015 r
F P
0 g
s
.20918 .00715 m .20834 .00699 L .2441 .00715 L closepath p .056 .015 .015 r
F P
0 g
s
.20834 .00699 m .20918 .00715 L .20757 .00884 L closepath p .074 .074 .074 r
F P
0 g
s
.19083 .00854 m .18832 .00779 L .18797 .00699 L closepath p .512 .049 .049 r
F P
0 g
s
.18712 .00715 m .18797 .00699 L .18832 .00779 L closepath p .511 .049 .049 r
F P
0 g
s
.18797 .00699 m .18712 .00715 L .12135 .00699 L closepath p .056 .015 .015 r
F P
0 g
s
.12135 .00699 m .12033 .00715 L .11883 .00802 L closepath p .354 .825 .738 r
F P
0 g
s
.12033 .00715 m .12135 .00699 L .18712 .00715 L closepath p .056 .015 .015 r
F P
0 g
s
.06688 .00868 m .06407 .00795 L .06364 .00699 L closepath p .535 .058 .058 r
F P
0 g
s
.06246 .00715 m .06364 .00699 L .06407 .00795 L closepath p .528 .06 .06 r
F P
0 g
s
.06364 .00699 m .06246 .00715 L .03274 .00699 L closepath p .056 .015 .015 r
F P
0 g
s
.03274 .00699 m .03148 .00715 L .03023 .00798 L closepath p .374 .843 .766 r
F P
0 g
s
.03148 .00715 m .03274 .00699 L .06246 .00715 L closepath p .056 .015 .015 r
F P
0 g
s
.95347 .00848 m .95299 .00776 L .95049 .007 L closepath p .041 .388 .78 r
F P
0 g
s
.9517 .00716 m .95049 .007 L .95299 .00776 L closepath p .035 .389 .767 r
F P
0 g
s
.95049 .007 m .9517 .00716 L .91404 .007 L closepath p .056 .015 .015 r
F P
0 g
s
.91678 .00716 m .91404 .007 L .9517 .00716 L closepath p .056 .015 .015 r
F P
0 g
s
.91404 .007 m .91678 .00716 L .91336 .00884 L closepath p .197 .173 .455 r
F P
0 g
s
.89651 .00851 m .89589 .0078 L .89342 .007 L closepath p .035 .374 .761 r
F P
0 g
s
.89448 .00716 m .89342 .007 L .89589 .0078 L closepath p .028 .374 .746 r
F P
0 g
s
.89342 .007 m .89448 .00716 L .86718 .007 L closepath p .056 .015 .015 r
F P
0 g
s
.86817 .00716 m .86718 .007 L .89448 .00716 L closepath p .056 .015 .015 r
F P
0 g
s
.86718 .007 m .86817 .00716 L .86506 .00812 L closepath p .054 .054 .054 r
F P
0 g
s
.79661 .00858 m .79565 .0078 L .79382 .007 L closepath p .02 .28 .688 r
F P
0 g
s
.79461 .00716 m .79382 .007 L .79565 .0078 L closepath p .027 .273 .699 r
F P
0 g
s
.79382 .007 m .79461 .00716 L .75569 .007 L closepath p .056 .015 .015 r
F P
0 g
s
.75638 .00716 m .75569 .007 L .79461 .00716 L closepath p .056 .015 .015 r
F P
0 g
s
.75569 .007 m .75638 .00716 L .7536 .00789 L closepath p .038 .038 .038 r
F P
0 g
s
.7329 .00861 m .73172 .0078 L .73025 .007 L closepath p .008 .219 .634 r
F P
0 g
s
.73087 .00716 m .73025 .007 L .73172 .0078 L closepath p .023 .19 .651 r
F P
0 g
s
.73025 .007 m .73087 .00716 L .69196 .007 L closepath p .056 .015 .015 r
F P
0 g
s
.69248 .00716 m .69196 .007 L .73087 .00716 L closepath p .056 .015 .015 r
F P
0 g
s
.69196 .007 m .69248 .00716 L .68983 .00786 L closepath p .032 .032 .032 r
F P
0 g
s
.68612 .00849 m .68492 .00776 L .68321 .007 L closepath p .01 .282 .667 r
F P
0 g
s
.68371 .00716 m .68321 .007 L .68492 .00776 L closepath p .007 .285 .661 r
F P
0 g
s
.68321 .007 m .68371 .00716 L .64707 .007 L closepath p .056 .015 .015 r
F P
0 g
s
.64894 .00716 m .64707 .007 L .68371 .00716 L closepath p .056 .015 .015 r
F P
0 g
s
.64707 .007 m .64894 .00716 L .64572 .00965 L closepath p .135 .135 .226 r
F P
0 g
s
.62936 .00851 m .62804 .0078 L .62631 .007 L closepath p .005 .278 .653 r
F P
0 g
s
.62665 .00716 m .62631 .007 L .62804 .0078 L closepath p .295 0 0 r
F P
0 g
s
.62631 .007 m .62665 .00716 L .56322 .007 L closepath p .056 .015 .015 r
F P
0 g
s
.56339 .00716 m .56322 .007 L .62665 .00716 L closepath p .056 .015 .015 r
F P
0 g
s
.56322 .007 m .56339 .00716 L .56086 .00798 L closepath p .02 .02 .02 r
F P
0 g
s
.51587 .00864 m .5142 .00787 L .5129 .007 L closepath p .389 .013 .013 r
F P
0 g
s
.51293 .00716 m .5129 .007 L .5142 .00787 L closepath p .39 .011 .011 r
F P
0 g
s
.5129 .007 m .51293 .00716 L .48577 .007 L closepath p .056 .015 .015 r
F P
0 g
s
.48573 .00716 m .48577 .007 L .51293 .00716 L closepath p .056 .015 .015 r
F P
0 g
s
.48577 .007 m .48573 .00716 L .4834 .00798 L closepath p .012 .012 .012 r
F P
0 g
s
.54507 .01293 m .54305 .01181 L .54181 .00781 L closepath p .906 .459 .095 r
F P
0 g
s
.54042 .00739 m .54181 .00781 L .54305 .01181 L closepath p .891 .456 .086 r
F P
0 g
s
.36514 .00746 m .3638 .00786 L .36559 .00715 L closepath p .041 .041 .041 r
F P
0 g
s
.3638 .00786 m .36514 .00746 L .36265 .00991 L closepath p .086 .086 .086 r
F P
0 g
s
.69204 .00747 m .68983 .00786 L .69248 .00716 L closepath p .083 .083 .083 r
F P
0 g
s
.68983 .00786 m .69204 .00747 L .68874 .00991 L closepath p .136 .136 .222 r
F P
0 g
s
.31161 .00746 m .31041 .00788 L .31202 .00715 L closepath p .037 .037 .037 r
F P
0 g
s
.31041 .00788 m .31161 .00746 L .3093 .00991 L closepath p .076 .076 .076 r
F P
0 g
s
.25141 .00746 m .25036 .00788 L .25182 .00715 L closepath p .028 .028 .028 r
F P
0 g
s
.25183 .01004 m .25083 .01366 L .25099 .00813 L p .16 .234 .77 r
F P
0 g
s
.25099 .00813 m .251 .00777 L .2511 .00803 L p .16 .234 .77 r
F P
0 g
s
.2511 .00803 m .25183 .01004 L p .25099 .00813 L .16 .234 .77 r
F P
0 g
s
.25036 .00788 m .25141 .00746 L .24989 .00919 L p .064 .064 .064 r
F P
0 g
s
.24989 .00919 m .24925 .00991 L .24938 .00967 L p .064 .064 .064 r
F P
0 g
s
.24938 .00967 m .25036 .00788 L p .24989 .00919 L .064 .064 .064 r
F P
0 g
s
.41246 .00746 m .41098 .00788 L .41286 .00715 L closepath p .051 .051 .051 r
F P
0 g
s
.41284 .01009 m .4114 .01366 L .41201 .00814 L p .175 .283 .798 r
F P
0 g
s
.41201 .00814 m .41205 .00777 L .41215 .00804 L p .175 .283 .798 r
F P
0 g
s
.41215 .00804 m .41284 .01009 L p .41201 .00814 L .175 .283 .798 r
F P
0 g
s
.41098 .00788 m .41246 .00746 L .41069 .00915 L p .094 .094 .094 r
F P
0 g
s
.41069 .00915 m .40989 .00991 L .41003 .00965 L p .094 .094 .094 r
F P
0 g
s
.41003 .00965 m .41098 .00788 L p .41069 .00915 L .094 .094 .094 r
F P
0 g
s
.75598 .00747 m .7536 .00789 L .75638 .00716 L closepath p .096 .096 .096 r
F P
0 g
s
.7536 .00789 m .75598 .00747 L .75256 .00991 L closepath p .145 .145 .267 r
F P
0 g
s
.03082 .00765 m .03023 .00798 L .03148 .00715 L closepath p .009 .009 .009 r
F P
0 g
s
.03077 .01016 m .03018 .01341 L .03016 .00877 L p .126 .126 .639 r
F P
0 g
s
.03023 .00798 m .03082 .00765 L .0294 .00967 L p .033 .033 .033 r
F P
0 g
s
.0294 .00967 m .02896 .0103 L .02912 .01 L p .033 .033 .033 r
F P
0 g
s
.02912 .01 m .03023 .00798 L p .0294 .00967 L .033 .033 .033 r
F P
0 g
s
.03015 .00816 m .02896 .0103 L .03082 .00765 L closepath p .745 1 .856 r
F P
0 g
s
.03016 .00877 m .03015 .00816 L .03029 .00861 L p .126 .126 .639 r
F P
0 g
s
.03029 .00861 m .03077 .01016 L p .03016 .00877 L .126 .126 .639 r
F P
0 g
s
.56283 .00762 m .56086 .00798 L .56339 .00716 L closepath p .085 .085 .085 r
F P
0 g
s
.56086 .00798 m .56283 .00762 L .55973 .01022 L closepath p .127 .127 .191 r
F P
0 g
s
.48517 .00762 m .4834 .00798 L .48573 .00716 L closepath p .075 .075 .075 r
F P
0 g
s
.11967 .00769 m .11883 .00802 L .12033 .00715 L closepath p .026 .026 .026 r
F P
0 g
s
.11958 .01023 m .11875 .01372 L .11898 .00886 L p .139 .139 .668 r
F P
0 g
s
.11898 .00886 m .11901 .00824 L .11914 .00869 L p .139 .139 .668 r
F P
0 g
s
.11914 .00869 m .11958 .01023 L p .11898 .00886 L .139 .139 .668 r
F P
0 g
s
.11883 .00802 m .11967 .00769 L .118 .00984 L p .054 .054 .054 r
F P
0 g
s
.118 .00984 m .11759 .01037 L .11772 .01012 L p .054 .054 .054 r
F P
0 g
s
.11772 .01012 m .11883 .00802 L p .118 .00984 L .054 .054 .054 r
F P
0 g
s
.86786 .00762 m .86506 .00812 L .86817 .00716 L closepath p .15 .15 .242 r
F P
0 g
s
.86506 .00812 m .86786 .00762 L .86424 .01022 L closepath p .167 .167 .418 r
F P
0 g
s
.8197 .00813 m .81783 .00825 L .82588 .00618 L closepath p .041 .041 .041 r
F P
0 g
s
.81783 .00825 m .8197 .00813 L .81205 .01178 L closepath p .076 .076 .076 r
F P
0 g
s
.86321 .01058 m .8631 .01049 L .85728 .0083 L closepath p .004 .345 .674 r
F P
0 g
s
.85743 .00832 m .85728 .0083 L .8631 .01049 L closepath p .004 .345 .674 r
F P
0 g
s
.24877 .01169 m .24575 .00956 L .24769 .00848 L closepath p 1 .771 .398 r
F P
0 g
s
.24536 .00775 m .24769 .00848 L .24575 .00956 L closepath p 1 .83 .447 r
F P
0 g
s
.95471 .01169 m .95347 .00956 L .95347 .00848 L closepath p .787 .438 .036 r
F P
0 g
s
.95299 .00776 m .95347 .00848 L .95347 .00956 L closepath p .776 .487 .026 r
F P
0 g
s
.68726 .01169 m .68535 .00956 L .68612 .00849 L closepath p .969 .615 .12 r
F P
0 g
s
.68492 .00776 m .68612 .00849 L .68535 .00956 L closepath p .974 .674 .164 r
F P
0 g
s
.89779 .01185 m .89641 .00972 L .89651 .00851 L closepath p .829 .476 .048 r
F P
0 g
s
.89589 .0078 m .89651 .00851 L .89641 .00972 L closepath p .819 .521 .038 r
F P
0 g
s
.63058 .01185 m .62852 .00972 L .62936 .00851 L closepath p .991 .632 .152 r
F P
0 g
s
.62804 .0078 m .62936 .00851 L .62852 .00972 L closepath p .995 .686 .19 r
F P
0 g
s
.19185 .01184 m .18868 .00971 L .19083 .00854 L closepath p 1 .802 .449 r
F P
0 g
s
.18832 .00779 m .19083 .00854 L .18868 .00971 L closepath p 1 .859 .497 r
F P
0 g
s
.35433 .01184 m .35158 .00971 L .35338 .00857 L closepath p 1 .784 .394 r
F P
0 g
s
.35125 .00779 m .35338 .00857 L .35158 .00971 L closepath p 1 .844 .444 r
F P
0 g
s
.79767 .01185 m .79603 .00972 L .79661 .00858 L closepath p .938 .601 .085 r
F P
0 g
s
.79565 .0078 m .79661 .00858 L .79603 .00972 L closepath p .934 .657 .126 r
F P
0 g
s
.29444 .01184 m .29153 .00971 L .29354 .00858 L closepath p 1 .806 .435 r
F P
0 g
s
.29124 .00779 m .29354 .00858 L .29153 .00971 L closepath p 1 .866 .485 r
F P
0 g
s
.40769 .01185 m .40507 .00971 L .40676 .00858 L closepath p 1 .777 .374 r
F P
0 g
s
.40476 .00779 m .40676 .00858 L .40507 .00971 L closepath p 1 .837 .424 r
F P
0 g
s
.73383 .01185 m .73204 .00972 L .7329 .00861 L closepath p .994 .667 .175 r
F P
0 g
s
.73172 .0078 m .7329 .00861 L .73204 .00972 L closepath p .992 .726 .221 r
F P
0 g
s
.51697 .01216 m .51462 .01003 L .51587 .00864 L closepath p 1 .725 .287 r
F P
0 g
s
.5142 .00787 m .51587 .00864 L .51462 .01003 L closepath p 1 .777 .324 r
F P
0 g
s
.06802 .01246 m .06454 .01033 L .06688 .00868 L closepath p 1 .824 .502 r
F P
0 g
s
.06407 .00795 m .06688 .00868 L .06454 .01033 L closepath p 1 .87 .536 r
F P
0 g
s
.20901 .00762 m .20757 .00884 L .20918 .00715 L closepath p .082 .082 .082 r
F P
0 g
s
.20757 .00884 m .20901 .00762 L .20719 .00989 L closepath p .084 .084 .085 r
F P
0 g
s
.91663 .00762 m .91336 .00884 L .91678 .00716 L closepath p .218 .178 .499 r
F P
0 g
s
.91336 .00884 m .91663 .00762 L .91304 .0099 L closepath p .222 .179 .51 r
F P
0 g
s
.53722 .00842 m .53526 .0092 L .53793 .00654 L closepath p .124 .124 .236 r
F P
0 g
s
.53526 .0092 m .53722 .00842 L .53435 .01348 L closepath p .136 .136 .372 r
F P
0 g
s
.09026 .00844 m .08948 .00922 L .09117 .00626 L closepath p .039 .039 .039 r
F P
0 g
s
.08948 .00922 m .09026 .00844 L .08816 .01353 L closepath p .048 .048 .048 r
F P
0 g
s
.64847 .00809 m .64572 .00965 L .64894 .00716 L closepath p .135 .135 .226 r
F P
0 g
s
.64572 .00965 m .64847 .00809 L .647 .0134 L closepath p .294 .366 .845 r
F P
0 g
s
.10064 .01498 m .09762 .01412 L .09761 .00979 L closepath p 1 .648 .294 r
F P
0 g
s
.09499 .00935 m .09761 .00979 L .09762 .01412 L closepath p 1 .653 .287 r
F P
0 g
s
.20883 .00808 m .20719 .00989 L .20901 .00762 L closepath p .081 .081 .081 r
F P
0 g
s
.20719 .00989 m .20883 .00808 L .20845 .01345 L closepath p .159 .223 .762 r
F P
0 g
s
.91648 .00809 m .91304 .0099 L .91663 .00762 L closepath p .214 .177 .496 r
F P
0 g
s
.91304 .0099 m .91648 .00809 L .91444 .01344 L closepath p .422 .47 .905 r
F P
0 g
s
.3647 .00777 m .36265 .00991 L .36514 .00746 L closepath p .609 .966 .772 r
F P
0 g
s
.36265 .00991 m .3647 .00777 L .36372 .0132 L closepath p .191 .288 .798 r
F P
0 g
s
.31121 .00777 m .3093 .00991 L .31161 .00746 L closepath p .619 .961 .741 r
F P
0 g
s
.3093 .00991 m .31121 .00777 L .30992 .01228 L closepath p .168 .219 .744 r
F P
0 g
s
.251 .00777 m .24925 .00991 L .25141 .00746 L closepath p .649 .983 .77 r
F P
0 g
s
.24925 .00991 m .251 .00777 L .25083 .01366 L closepath p .172 .311 .827 r
F P
0 g
s
.41205 .00777 m .40989 .00991 L .41246 .00746 L closepath p .565 .918 .686 r
F P
0 g
s
.40989 .00991 m .41205 .00777 L .4114 .01366 L closepath p .247 .356 .851 r
F P
0 g
s
.75559 .00778 m .75256 .00991 L .75598 .00747 L closepath p .017 .017 .017 r
F P
0 g
s
.75256 .00991 m .75559 .00778 L .75495 .01404 L closepath p .427 .525 .959 r
F P
0 g
s
.6916 .00778 m .68874 .00991 L .69204 .00747 L closepath p .44 .831 .606 r
F P
0 g
s
.68874 .00991 m .6916 .00778 L .69001 .01331 L closepath p .356 .421 .879 r
F P
0 g
s
.86755 .00809 m .86424 .01022 L .86786 .00762 L closepath p .125 .125 .125 r
F P
0 g
s
.56227 .00809 m .55973 .01022 L .56283 .00762 L closepath p .46 .821 .553 r
F P
0 g
s
.55973 .01022 m .56227 .00809 L .56042 .01244 L closepath p .283 .349 .831 r
F P
0 g
s
.02896 .0103 m .03015 .00816 L .03018 .01341 L closepath p .147 .191 .744 r
F P
0 g
s
.11901 .00824 m .11759 .01037 L .11967 .00769 L closepath p .702 1 .782 r
F P
0 g
s
.11759 .01037 m .11901 .00824 L .11875 .01372 L closepath p .154 .198 .744 r
F P
0 g
s
.8631 .01049 m .86321 .01058 L .86518 .0129 L closepath p .545 .068 .003 r
F P
0 g
s
.24685 .01445 m .24487 .0106 L .24877 .01169 L closepath p 1 1 .993 r
F P
0 g
s
.24575 .00956 m .24877 .01169 L .24487 .0106 L closepath p 1 1 1 r
F P
0 g
s
.95292 .01446 m .95264 .01061 L .95471 .01169 L closepath p 1 1 .903 r
F P
0 g
s
.95347 .00956 m .95471 .01169 L .95264 .01061 L closepath p 1 1 .945 r
F P
0 g
s
.68543 .01446 m .6845 .01061 L .68726 .01169 L closepath p 1 1 .951 r
F P
0 g
s
.68535 .00956 m .68726 .01169 L .6845 .01061 L closepath p 1 1 .988 r
F P
0 g
s
.81441 .01137 m .81205 .01178 L .8197 .00813 L closepath p .078 .078 .078 r
F P
0 g
s
.81205 .01178 m .81441 .01137 L .8073 .01666 L closepath p .115 .115 .115 r
F P
0 g
s
.35359 .0135 m .35077 .01018 L .35433 .01184 L closepath p 1 1 .925 r
F P
0 g
s
.35158 .00971 m .35433 .01184 L .35077 .01018 L closepath p .992 1 1 r
F P
0 g
s
.29369 .0135 m .29072 .01018 L .29444 .01184 L closepath p 1 1 .931 r
F P
0 g
s
.29153 .00971 m .29444 .01184 L .29072 .01018 L closepath p .995 1 1 r
F P
0 g
s
.18985 .01453 m .18775 .01064 L .19185 .01184 L closepath p 1 1 1 r
F P
0 g
s
.18868 .00971 m .19185 .01184 L .18775 .01064 L closepath p 1 1 1 r
F P
0 g
s
.40695 .0135 m .40426 .01018 L .40769 .01185 L closepath p 1 1 .92 r
F P
0 g
s
.40507 .00971 m .40769 .01185 L .40426 .01018 L closepath p .989 1 1 r
F P
0 g
s
.89592 .01453 m .89553 .01065 L .89779 .01185 L closepath p 1 1 .937 r
F P
0 g
s
.89641 .00972 m .89779 .01185 L .89553 .01065 L closepath p 1 1 .997 r
F P
0 g
s
.79698 .0135 m .79525 .01018 L .79767 .01185 L closepath p 1 1 .862 r
F P
0 g
s
.79603 .00972 m .79767 .01185 L .79525 .01018 L closepath p .961 1 1 r
F P
0 g
s
.73314 .0135 m .73125 .01018 L .73383 .01185 L closepath p 1 1 .874 r
F P
0 g
s
.73204 .00972 m .73383 .01185 L .73125 .01018 L closepath p .967 1 1 r
F P
0 g
s
.62866 .01453 m .62763 .01065 L .63058 .01185 L closepath p 1 1 .975 r
F P
0 g
s
.62852 .00972 m .63058 .01185 L .62763 .01065 L closepath p 1 1 1 r
F P
0 g
s
.51623 .01375 m .51384 .01038 L .51697 .01216 L closepath p 1 1 .917 r
F P
0 g
s
.51462 .01003 m .51697 .01216 L .51384 .01038 L closepath p .929 1 1 r
F P
0 g
s
.31193 .00902 m .30992 .01228 L .31121 .00777 L closepath p .376 .492 .945 r
F P
0 g
s
.30992 .01228 m .31193 .00902 L .31166 .01405 L closepath p .496 .651 1 r
F P
0 g
s
.56305 .00916 m .56042 .01244 L .56227 .00809 L closepath p .495 .598 1 r
F P
0 g
s
.56042 .01244 m .56305 .00916 L .56228 .01406 L closepath p .574 .711 1 r
F P
0 g
s
.06716 .01413 m .0636 .01074 L .06802 .01246 L closepath p 1 1 .961 r
F P
0 g
s
.06454 .01033 m .06802 .01246 L .0636 .01074 L closepath p .954 1 1 r
F P
0 g
s
.0332 .01318 m .03279 .01112 L .03395 .01261 L closepath p .931 1 1 r
F P
0 g
s
.0334 .01112 m .03395 .01261 L .03279 .01112 L closepath p .484 .916 1 r
F P
0 g
s
.03395 .01261 m .0334 .01112 L .03507 .01352 L closepath p .074 .312 .8 r
F P
0 g
s
.54879 .02036 m .54677 .01927 L .54507 .01293 L closepath p .979 .573 .117 r
F P
0 g
s
.54305 .01181 m .54507 .01293 L .54677 .01927 L closepath p .979 .573 .117 r
F P
0 g
s
.03155 .0139 m .03178 .01064 L .0332 .01318 L closepath p .923 1 1 r
F P
0 g
s
.03279 .01112 m .0332 .01318 L .03178 .01064 L closepath p .267 .674 1 r
F P
0 g
s
.36577 .00932 m .36372 .0132 L .3647 .00777 L closepath p .471 .577 .993 r
F P
0 g
s
.36372 .0132 m .36577 .00932 L .36653 .01522 L closepath p .616 .758 1 r
F P
0 g
s
.83299 .01431 m .83289 .00945 L .83755 .01329 L closepath p .936 1 1 r
F P
0 g
s
.83835 .00886 m .83755 .01329 L .83289 .00945 L closepath p .882 1 1 r
F P
0 g
s
.83755 .01329 m .83835 .00886 L .8447 .01457 L closepath p .784 .942 1 r
F P
0 g
s
.69226 .00983 m .69001 .01331 L .6916 .00778 L closepath p .307 .367 .842 r
F P
0 g
s
.69001 .01331 m .69226 .00983 L .69309 .01524 L closepath p .619 .781 1 r
F P
0 g
s
.64909 .00997 m .647 .0134 L .64847 .00809 L closepath p .287 .359 .84 r
F P
0 g
s
.647 .0134 m .64909 .00997 L .6503 .01535 L closepath p .617 .79 1 r
F P
0 g
s
.03018 .01341 m .03077 .01016 L .03155 .0139 L closepath p .613 .86 1 r
F P
0 g
s
.91718 .00997 m .91444 .01344 L .91648 .00809 L closepath p .406 .453 .893 r
F P
0 g
s
.91444 .01344 m .91718 .00997 L .91773 .01535 L closepath p .651 .802 1 r
F P
0 g
s
.20946 .00996 m .20845 .01345 L .20883 .00808 L closepath p .155 .195 .74 r
F P
0 g
s
.20845 .01345 m .20946 .00996 L .21167 .01534 L closepath p .564 .774 1 r
F P
0 g
s
.53654 .01154 m .53435 .01348 L .53722 .00842 L closepath p .135 .135 .366 r
F P
0 g
s
.53435 .01348 m .53654 .01154 L .53343 .01777 L closepath p .136 .136 .37 r
F P
0 g
s
.35164 .0146 m .34996 .01064 L .35359 .0135 L closepath p 1 1 1 r
F P
0 g
s
.35077 .01018 m .35359 .0135 L .34996 .01064 L closepath p 1 1 1 r
F P
0 g
s
.29173 .01461 m .28991 .01064 L .29369 .0135 L closepath p 1 1 1 r
F P
0 g
s
.29072 .01018 m .29369 .0135 L .28991 .01064 L closepath p 1 1 1 r
F P
0 g
s
.40501 .01461 m .40346 .01065 L .40695 .0135 L closepath p 1 1 1 r
F P
0 g
s
.40426 .01018 m .40695 .0135 L .40346 .01065 L closepath p 1 1 1 r
F P
0 g
s
.79507 .01461 m .79446 .01065 L .79698 .0135 L closepath p 1 1 1 r
F P
0 g
s
.79525 .01018 m .79698 .0135 L .79446 .01065 L closepath p 1 1 1 r
F P
0 g
s
.73123 .01461 m .73046 .01065 L .73314 .0135 L closepath p 1 1 1 r
F P
0 g
s
.73125 .01018 m .73314 .0135 L .73046 .01065 L closepath p 1 1 1 r
F P
0 g
s
.034 .01112 m .03507 .01352 L .0334 .01112 L closepath p .666 .99 1 r
F P
0 g
s
.03507 .01352 m .034 .01112 L .03888 .01554 L closepath p .223 .632 1 r
F P
0 g
s
.08935 .01208 m .08816 .01353 L .09026 .00844 L closepath p .071 .071 .093 r
F P
0 g
s
.08816 .01353 m .08935 .01208 L .0871 .019 L closepath p .079 .079 .158 r
F P
0 g
s
.4114 .01366 m .41284 .01009 L .41525 .01554 L closepath p .628 .821 1 r
F P
0 g
s
.25083 .01366 m .25183 .01004 L .25482 .01537 L closepath p .636 .841 1 r
F P
0 g
s
.11875 .01372 m .11958 .01023 L .12182 .01564 L closepath p .539 .755 1 r
F P
0 g
s
.51428 .01474 m .51306 .01073 L .51623 .01375 L closepath p 1 1 1 r
F P
0 g
s
.51384 .01038 m .51623 .01375 L .51306 .01073 L closepath p .986 1 1 r
F P
0 g
s
.03178 .01064 m .03155 .0139 L .03077 .01016 L closepath p .54 .798 1 r
F P
0 g
s
.75702 .01028 m .75495 .01404 L .75559 .00778 L closepath p .421 .519 .955 r
F P
0 g
s
.75495 .01404 m .75702 .01028 L .76059 .01554 L closepath p .73 .91 1 r
F P
0 g
s
.31264 .01028 m .31166 .01405 L .31193 .00902 L closepath p .262 .422 .91 r
F P
0 g
s
.31166 .01405 m .31264 .01028 L .31747 .01554 L closepath p .714 .912 1 r
F P
0 g
s
.56383 .01024 m .56228 .01406 L .56305 .00916 L closepath p .423 .559 .992 r
F P
0 g
s
.56228 .01406 m .56383 .01024 L .56832 .01538 L closepath p .741 .927 1 r
F P
0 g
s
.06496 .01516 m .06266 .01114 L .06716 .01413 L closepath p 1 1 1 r
F P
0 g
s
.0636 .01074 m .06716 .01413 L .06266 .01114 L closepath p .994 1 1 r
F P
0 g
s
.82842 .01532 m .8281 .01123 L .83299 .01431 L closepath p .912 1 1 r
F P
0 g
s
.83289 .00945 m .83299 .01431 L .8281 .01123 L closepath p .978 1 1 r
F P
0 g
s
.24287 .01538 m .24226 .01095 L .24685 .01445 L closepath p .931 1 1 r
F P
0 g
s
.24487 .0106 m .24685 .01445 L .24226 .01095 L closepath p .88 1 1 r
F P
0 g
s
.94899 .01538 m .95006 .01096 L .95292 .01446 L closepath p .917 1 1 r
F P
0 g
s
.95264 .01061 m .95292 .01446 L .95006 .01096 L closepath p .869 1 1 r
F P
0 g
s
.68148 .01538 m .68191 .01096 L .68543 .01446 L closepath p .923 1 1 r
F P
0 g
s
.6845 .01061 m .68543 .01446 L .68191 .01096 L closepath p .874 1 1 r
F P
0 g
s
.18579 .01538 m .18503 .01095 L .18985 .01453 L closepath p .924 1 1 r
F P
0 g
s
.18775 .01064 m .18985 .01453 L .18503 .01095 L closepath p .874 1 1 r
F P
0 g
s
.89189 .01538 m .89282 .01096 L .89592 .01453 L closepath p .912 1 1 r
F P
0 g
s
.89553 .01065 m .89592 .01453 L .89282 .01096 L closepath p .864 1 1 r
F P
0 g
s
.62462 .01538 m .62491 .01096 L .62866 .01453 L closepath p .917 1 1 r
F P
0 g
s
.62763 .01065 m .62866 .01453 L .62491 .01096 L closepath p .868 1 1 r
F P
0 g
s
.84633 .01032 m .8447 .01457 L .83835 .00886 L closepath p .781 .941 1 r
F P
0 g
s
.8447 .01457 m .84633 .01032 L .85167 .01859 L closepath p .662 .806 1 r
F P
0 g
s
.34559 .01553 m .34514 .01111 L .35164 .0146 L closepath p .903 1 1 r
F P
0 g
s
.34996 .01064 m .35164 .0146 L .34514 .01111 L closepath p .867 1 1 r
F P
0 g
s
.28568 .01553 m .28509 .01111 L .29173 .01461 L closepath p .904 1 1 r
F P
0 g
s
.28991 .01064 m .29173 .01461 L .28509 .01111 L closepath p .868 1 1 r
F P
0 g
s
.39897 .01554 m .39864 .01111 L .40501 .01461 L closepath p .902 1 1 r
F P
0 g
s
.40346 .01065 m .40501 .01461 L .39864 .01111 L closepath p .867 1 1 r
F P
0 g
s
.78911 .01554 m .7897 .01111 L .79507 .01461 L closepath p .897 1 1 r
F P
0 g
s
.79446 .01065 m .79507 .01461 L .7897 .01111 L closepath p .862 1 1 r
F P
0 g
s
.72526 .01554 m .7257 .01111 L .73123 .01461 L closepath p .898 1 1 r
F P
0 g
s
.73046 .01065 m .73123 .01461 L .7257 .01111 L closepath p .863 1 1 r
F P
0 g
s
.50833 .01538 m .50835 .01096 L .51428 .01474 L closepath p .885 1 1 r
F P
0 g
s
.51306 .01073 m .51428 .01474 L .50835 .01096 L closepath p .847 1 1 r
F P
0 g
s
.59357 .01567 m .59344 .01127 L .60045 .01495 L closepath p .883 1 1 r
F P
0 g
s
.60084 .01049 m .60045 .01495 L .59344 .01127 L closepath p .883 1 1 r
F P
0 g
s
.60045 .01495 m .60084 .01049 L .60603 .01613 L closepath p .768 .93 1 r
F P
0 g
s
.86946 .01057 m .86824 .01497 L .86851 .00933 L closepath p .454 .586 1 r
F P
0 g
s
.86824 .01497 m .86946 .01057 L .87115 .01545 L closepath p .791 .947 1 r
F P
0 g
s
.10421 .02188 m .10104 .02081 L .10064 .01498 L closepath p 1 .693 .347 r
F P
0 g
s
.09762 .01412 m .10064 .01498 L .10104 .02081 L closepath p 1 .693 .345 r
F P
0 g
s
.05814 .01584 m .0571 .01142 L .06496 .01516 L closepath p .886 1 1 r
F P
0 g
s
.06266 .01114 m .06496 .01516 L .0571 .01142 L closepath p .85 1 1 r
F P
0 g
s
.36684 .01087 m .36653 .01522 L .36577 .00932 L closepath p .286 .471 .949 r
F P
0 g
s
.36653 .01522 m .36684 .01087 L .36901 .01538 L closepath p .819 .982 1 r
F P
0 g
s
.69414 .01088 m .69309 .01524 L .69226 .00983 L closepath p .652 .809 1 r
F P
0 g
s
.69309 .01524 m .69414 .01088 L .69483 .01543 L closepath p .806 .966 1 r
F P
0 g
s
.16234 .01573 m .16063 .01115 L .16467 .01526 L closepath p .932 1 1 r
F P
0 g
s
.16346 .0108 m .16467 .01526 L .16063 .01115 L closepath p .897 1 1 r
F P
0 g
s
.16467 .01526 m .16346 .0108 L .16755 .01659 L closepath p .606 .824 1 r
F P
0 g
s
.82483 .01793 m .82398 .01419 L .82842 .01532 L closepath p 1 1 1 r
F P
0 g
s
.8281 .01123 m .82842 .01532 L .82398 .01419 L closepath p 1 1 1 r
F P
0 g
s
.21143 .01095 m .21167 .01534 L .20946 .00996 L closepath p .622 .815 1 r
F P
0 g
s
.21167 .01534 m .21143 .01095 L .21333 .01543 L closepath p .822 .985 1 r
F P
0 g
s
.91914 .01096 m .91773 .01535 L .91718 .00997 L closepath p .684 .83 1 r
F P
0 g
s
.91773 .01535 m .91914 .01096 L .91943 .01544 L closepath p .828 .985 1 r
F P
0 g
s
.65107 .01096 m .6503 .01535 L .64909 .00997 L closepath p .664 .827 1 r
F P
0 g
s
.6503 .01535 m .65107 .01096 L .65199 .01544 L closepath p .826 .985 1 r
F P
0 g
s
.25441 .01095 m .25482 .01537 L .25183 .01004 L closepath p .69 .876 1 r
F P
0 g
s
.25482 .01537 m .25441 .01095 L .26145 .01582 L closepath p .818 .98 1 r
F P
0 g
s
.36841 .01105 m .36901 .01538 L .36684 .01087 L closepath p .793 .967 1 r
F P
0 g
s
.36901 .01538 m .36841 .01105 L .37149 .01553 L closepath p .814 .983 1 r
F P
0 g
s
.23891 .01636 m .2369 .0125 L .24287 .01538 L closepath p .928 1 1 r
F P
0 g
s
.24226 .01095 m .24287 .01538 L .2369 .0125 L closepath p .949 1 1 r
F P
0 g
s
.18196 .01547 m .18102 .01105 L .18579 .01538 L closepath p .855 1 1 r
F P
0 g
s
.18503 .01095 m .18579 .01538 L .18102 .01105 L closepath p .856 1 1 r
F P
0 g
s
.88545 .01566 m .88495 .01187 L .88909 .01538 L closepath p .843 1 1 r
F P
0 g
s
.89 .01096 m .88909 .01538 L .88495 .01187 L closepath p .903 1 1 r
F P
0 g
s
.88909 .01538 m .89 .01096 L .89189 .01538 L closepath p .846 1 1 r
F P
0 g
s
.89282 .01096 m .89189 .01538 L .89 .01096 L closepath p .846 1 1 r
F P
0 g
s
.94508 .01636 m .94477 .01251 L .94899 .01538 L closepath p .912 1 1 r
F P
0 g
s
.95006 .01096 m .94899 .01538 L .94477 .01251 L closepath p .933 1 1 r
F P
0 g
s
.61803 .01566 m .61689 .01187 L .62174 .01538 L closepath p .846 1 1 r
F P
0 g
s
.62201 .01096 m .62174 .01538 L .61689 .01187 L closepath p .907 1 1 r
F P
0 g
s
.62174 .01538 m .62201 .01096 L .62462 .01538 L closepath p .846 1 1 r
F P
0 g
s
.62491 .01096 m .62462 .01538 L .62201 .01096 L closepath p .846 1 1 r
F P
0 g
s
.56848 .01096 m .56832 .01538 L .56383 .01024 L closepath p .787 .951 1 r
F P
0 g
s
.56832 .01538 m .56848 .01096 L .57183 .01557 L closepath p .826 .984 1 r
F P
0 g
s
.5059 .01581 m .50436 .01216 L .50833 .01538 L closepath p .885 1 1 r
F P
0 g
s
.50835 .01096 m .50833 .01538 L .50436 .01216 L closepath p .947 1 1 r
F P
0 g
s
.48701 .01096 m .48681 .01538 L .4852 .01001 L closepath p .634 .812 1 r
F P
0 g
s
.48681 .01538 m .48701 .01096 L .48822 .01538 L closepath p .846 1 1 r
F P
0 g
s
.48822 .01096 m .48822 .01538 L .48701 .01096 L closepath p .846 1 1 r
F P
0 g
s
.48822 .01538 m .48822 .01096 L .49138 .01538 L closepath p .846 1 1 r
F P
0 g
s
.49136 .01096 m .49138 .01538 L .48822 .01096 L closepath p .846 1 1 r
F P
0 g
s
.49138 .01538 m .49136 .01096 L .49341 .01577 L closepath p .771 .939 1 r
F P
0 g
s
.67761 .01636 m .67665 .01251 L .68148 .01538 L closepath p .92 1 1 r
F P
0 g
s
.68191 .01096 m .68148 .01538 L .67665 .01251 L closepath p .941 1 1 r
F P
0 g
s
.69555 .01102 m .69483 .01543 L .69414 .01088 L closepath p .811 .97 1 r
F P
0 g
s
.69483 .01543 m .69555 .01102 L .69814 .01554 L closepath p .833 .991 1 r
F P
0 g
s
.2128 .01101 m .21333 .01543 L .21143 .01095 L closepath p .828 .988 1 r
F P
0 g
s
.21333 .01543 m .2128 .01101 L .21661 .01546 L closepath p .842 .999 1 r
F P
0 g
s
.92055 .01102 m .91943 .01544 L .91914 .01096 L closepath p .832 .988 1 r
F P
0 g
s
.91943 .01544 m .92055 .01102 L .92265 .01546 L closepath p .843 .999 1 r
F P
0 g
s
.65248 .01102 m .65199 .01544 L .65107 .01096 L closepath p .831 .988 1 r
F P
0 g
s
.65199 .01544 m .65248 .01102 L .65522 .01546 L closepath p .843 .999 1 r
F P
0 g
s
.8726 .01109 m .87115 .01545 L .86946 .01057 L closepath p .79 .946 1 r
F P
0 g
s
.87115 .01545 m .8726 .01109 L .87225 .01567 L closepath p .78 .935 1 r
F P
0 g
s
.21594 .01103 m .21661 .01546 L .2128 .01101 L closepath p .843 .999 1 r
F P
0 g
s
.21661 .01546 m .21594 .01103 L .21904 .01564 L closepath p .814 .978 1 r
F P
0 g
s
.92365 .01104 m .92265 .01546 L .92055 .01102 L closepath p .844 .999 1 r
F P
0 g
s
.92265 .01546 m .92365 .01104 L .9251 .01565 L closepath p .82 .976 1 r
F P
0 g
s
.65558 .01104 m .65522 .01546 L .65248 .01102 L closepath p .844 .999 1 r
F P
0 g
s
.65522 .01546 m .65558 .01104 L .65766 .01565 L closepath p .818 .977 1 r
F P
0 g
s
.18034 .01557 m .17914 .01119 L .18196 .01547 L closepath p .871 1 1 r
F P
0 g
s
.18102 .01105 m .18196 .01547 L .17914 .01119 L closepath p .875 1 1 r
F P
0 g
s
.34218 .01598 m .3403 .01225 L .34559 .01553 L closepath p .872 1 1 r
F P
0 g
s
.34514 .01111 m .34559 .01553 L .3403 .01225 L closepath p .931 1 1 r
F P
0 g
s
.28227 .01598 m .28025 .01225 L .28568 .01553 L closepath p .873 1 1 r
F P
0 g
s
.28509 .01111 m .28568 .01553 L .28025 .01225 L closepath p .932 1 1 r
F P
0 g
s
.3712 .01111 m .37149 .01553 L .36841 .01105 L closepath p .838 .995 1 r
F P
0 g
s
.37149 .01553 m .3712 .01111 L .37592 .01639 L closepath p .766 .937 1 r
F P
0 g
s
.41525 .01112 m .41525 .01554 L .41284 .01009 L closepath p .668 .849 1 r
F P
0 g
s
.41525 .01554 m .41525 .01112 L .42149 .01591 L closepath p .823 .983 1 r
F P
0 g
s
.39555 .01598 m .39381 .01226 L .39897 .01554 L closepath p .871 1 1 r
F P
0 g
s
.39864 .01111 m .39897 .01554 L .39381 .01226 L closepath p .93 1 1 r
F P
0 g
s
.78576 .01598 m .78495 .01226 L .78911 .01554 L closepath p .865 1 1 r
F P
0 g
s
.7897 .01111 m .78911 .01554 L .78495 .01226 L closepath p .923 1 1 r
F P
0 g
s
.72191 .01598 m .72095 .01226 L .72526 .01554 L closepath p .866 1 1 r
F P
0 g
s
.7257 .01111 m .72526 .01554 L .72095 .01226 L closepath p .925 1 1 r
F P
0 g
s
.31704 .01111 m .31747 .01554 L .31264 .01028 L closepath p .766 .938 1 r
F P
0 g
s
.31747 .01554 m .31704 .01111 L .32235 .01637 L closepath p .775 .946 1 r
F P
0 g
s
.69865 .01111 m .69814 .01554 L .69555 .01102 L closepath p .835 .991 1 r
F P
0 g
s
.69814 .01554 m .69865 .01111 L .70226 .01638 L closepath p .771 .934 1 r
F P
0 g
s
.7612 .01111 m .76059 .01554 L .75702 .01028 L closepath p .774 .935 1 r
F P
0 g
s
.76059 .01554 m .7612 .01111 L .76107 .01554 L closepath p .846 1 1 r
F P
0 g
s
.76168 .01111 m .76107 .01554 L .7612 .01111 L closepath p .846 1 1 r
F P
0 g
s
.76107 .01554 m .76168 .01111 L .76581 .01637 L closepath p .783 .943 1 r
F P
0 g
s
.44151 .01585 m .43984 .01219 L .44378 .01554 L closepath p .87 1 1 r
F P
0 g
s
.44365 .01112 m .44378 .01554 L .43984 .01219 L closepath p .942 1 1 r
F P
0 g
s
.44378 .01554 m .44365 .01112 L .47046 .01554 L closepath p .846 1 1 r
F P
0 g
s
.47039 .01112 m .47046 .01554 L .44365 .01112 L closepath p .846 1 1 r
F P
0 g
s
.47046 .01554 m .47039 .01112 L .47352 .01727 L closepath p .605 .793 1 r
F P
0 g
s
.03778 .01112 m .03888 .01554 L .034 .01112 L closepath p .846 1 1 r
F P
0 g
s
.03888 .01554 m .03778 .01112 L .04133 .01605 L closepath p .75 .933 1 r
F P
0 g
s
.17878 .0159 m .17634 .01231 L .18034 .01557 L closepath p .907 1 1 r
F P
0 g
s
.17914 .01119 m .18034 .01557 L .17634 .01231 L closepath p .985 1 1 r
F P
0 g
s
.57264 .01126 m .57183 .01557 L .56848 .01096 L closepath p .815 .979 1 r
F P
0 g
s
.57183 .01557 m .57264 .01126 L .57351 .01612 L closepath p .724 .89 1 r
F P
0 g
s
.12142 .01126 m .12182 .01564 L .11958 .01023 L closepath p .585 .789 1 r
F P
0 g
s
.12182 .01564 m .12142 .01126 L .12368 .01579 L closepath p .809 .975 1 r
F P
0 g
s
.21984 .01192 m .21904 .01564 L .21594 .01103 L closepath p .717 .922 1 r
F P
0 g
s
.21904 .01564 m .21984 .01192 L .21942 .01672 L closepath p .151 .188 .737 r
F P
0 g
s
.92761 .01193 m .9251 .01565 L .92365 .01104 L closepath p .747 .925 1 r
F P
0 g
s
.9251 .01565 m .92761 .01193 L .92553 .01673 L closepath p .4 .457 .9 r
F P
0 g
s
.65952 .01193 m .65766 .01565 L .65558 .01104 L closepath p .736 .925 1 r
F P
0 g
s
.65766 .01565 m .65952 .01193 L .65807 .01673 L closepath p .294 .38 .86 r
F P
0 g
s
.88475 .01657 m .88333 .01445 L .88545 .01566 L closepath p 1 1 .948 r
F P
0 g
s
.88495 .01187 m .88545 .01566 L .88333 .01445 L closepath p 1 1 .906 r
F P
0 g
s
.61729 .01657 m .6152 .01445 L .61803 .01566 L closepath p 1 1 .993 r
F P
0 g
s
.61689 .01187 m .61803 .01566 L .6152 .01445 L closepath p 1 1 .955 r
F P
0 g
s
.87407 .01142 m .87225 .01567 L .8726 .01109 L closepath p .77 .926 1 r
F P
0 g
s
.87225 .01567 m .87407 .01142 L .87453 .0174 L closepath p .619 .75 1 r
F P
0 g
s
.59181 .0164 m .59031 .0129 L .59357 .01567 L closepath p .968 1 1 r
F P
0 g
s
.59344 .01127 m .59357 .01567 L .59031 .0129 L closepath p 1 1 1 r
F P
0 g
s
.15861 .01593 m .1578 .0115 L .16234 .01573 L closepath p .867 1 1 r
F P
0 g
s
.16063 .01115 m .16234 .01573 L .1578 .0115 L closepath p .903 1 1 r
F P
0 g
s
.49454 .01171 m .49341 .01577 L .49136 .01096 L closepath p .744 .921 1 r
F P
0 g
s
.49341 .01577 m .49454 .01171 L .49478 .01726 L closepath p .468 .625 1 r
F P
0 g
s
.12297 .01138 m .12368 .01579 L .12142 .01126 L closepath p .814 .978 1 r
F P
0 g
s
.12368 .01579 m .12297 .01138 L .12689 .01584 L closepath p .839 .997 1 r
F P
0 g
s
.50541 .01789 m .50303 .01577 L .5059 .01581 L closepath p 1 1 .774 r
F P
0 g
s
.50436 .01216 m .5059 .01581 L .50303 .01577 L closepath p 1 1 .827 r
F P
0 g
s
.26125 .01142 m .26145 .01582 L .25441 .01095 L closepath p .818 .98 1 r
F P
0 g
s
.26145 .01582 m .26125 .01142 L .26338 .01684 L closepath p .614 .807 1 r
F P
0 g
s
.12601 .01142 m .12689 .01584 L .12297 .01138 L closepath p .84 .998 1 r
F P
0 g
s
.12689 .01584 m .12601 .01142 L .1309 .01606 L closepath p .823 .985 1 r
F P
0 g
s
.05521 .01642 m .05258 .01278 L .05814 .01584 L closepath p .904 1 1 r
F P
0 g
s
.0571 .01142 m .05814 .01584 L .05258 .01278 L closepath p .956 1 1 r
F P
0 g
s
.44109 .01752 m .43855 .0154 L .44151 .01585 L closepath p 1 1 .797 r
F P
0 g
s
.43984 .01219 m .44151 .01585 L .43855 .0154 L closepath p 1 1 .857 r
F P
0 g
s
.17857 .01657 m .17537 .01445 L .17878 .0159 L closepath p 1 1 .87 r
F P
0 g
s
.17634 .01231 m .17878 .0159 L .17537 .01445 L closepath p 1 1 .92 r
F P
0 g
s
.42169 .0115 m .42149 .01591 L .41525 .01112 L closepath p .822 .983 1 r
F P
0 g
s
.42149 .01591 m .42169 .0115 L .42326 .01689 L closepath p .619 .799 1 r
F P
0 g
s
.15407 .01669 m .15173 .01299 L .15861 .01593 L closepath p .891 1 1 r
F P
0 g
s
.1578 .0115 m .15861 .01593 L .15173 .01299 L closepath p .937 1 1 r
F P
0 g
s
.34146 .01734 m .33866 .01522 L .34218 .01598 L closepath p 1 1 .935 r
F P
0 g
s
.3403 .01225 m .34218 .01598 L .33866 .01522 L closepath p 1 1 .941 r
F P
0 g
s
.28154 .01734 m .2786 .01522 L .28227 .01598 L closepath p 1 1 .941 r
F P
0 g
s
.28025 .01225 m .28227 .01598 L .2786 .01522 L closepath p 1 1 .948 r
F P
0 g
s
.78509 .01735 m .78342 .01523 L .78576 .01598 L closepath p 1 1 .853 r
F P
0 g
s
.78495 .01226 m .78576 .01598 L .78342 .01523 L closepath p 1 1 .863 r
F P
0 g
s
.72123 .01735 m .71941 .01523 L .72191 .01598 L closepath p 1 1 .869 r
F P
0 g
s
.72095 .01226 m .72191 .01598 L .71941 .01523 L closepath p 1 1 .878 r
F P
0 g
s
.39484 .01735 m .39218 .01523 L .39555 .01598 L closepath p 1 1 .928 r
F P
0 g
s
.39381 .01226 m .39555 .01598 L .39218 .01523 L closepath p 1 1 .935 r
F P
0 g
s
.04138 .01199 m .04133 .01605 L .03778 .01112 L closepath p .723 .917 1 r
F P
0 g
s
.04133 .01605 m .04138 .01199 L .043 .01789 L closepath p .334 .552 1 r
F P
0 g
s
.13067 .01175 m .1309 .01606 L .12601 .01142 L closepath p .812 .98 1 r
F P
0 g
s
.1309 .01606 m .13067 .01175 L .1328 .01668 L closepath p .698 .886 1 r
F P
0 g
s
.57513 .01216 m .57351 .01612 L .57264 .01126 L closepath p .706 .876 1 r
F P
0 g
s
.57351 .01612 m .57513 .01216 L .57431 .01745 L closepath p .392 .509 .956 r
F P
0 g
s
.60705 .01181 m .60603 .01613 L .60084 .01049 L closepath p .768 .93 1 r
F P
0 g
s
.60603 .01613 m .60705 .01181 L .60722 .01739 L closepath p .49 .64 1 r
F P
0 g
s
.23681 .01989 m .23385 .01716 L .23891 .01636 L closepath p 1 1 .969 r
F P
0 g
s
.2369 .0125 m .23891 .01636 L .23385 .01716 L closepath p 1 1 .981 r
F P
0 g
s
.67564 .01989 m .67377 .01717 L .67761 .01636 L closepath p 1 1 .92 r
F P
0 g
s
.67665 .01251 m .67761 .01636 L .67377 .01717 L closepath p 1 1 .933 r
F P
0 g
s
.94316 .01989 m .94196 .01717 L .94508 .01636 L closepath p 1 1 .869 r
F P
0 g
s
.94477 .01251 m .94508 .01636 L .94196 .01717 L closepath p 1 1 .881 r
F P
0 g
s
.32328 .01251 m .32235 .01637 L .31704 .01111 L closepath p .732 .925 1 r
F P
0 g
s
.32235 .01637 m .32328 .01251 L .32417 .01927 L closepath p .294 .457 .935 r
F P
0 g
s
.76781 .01251 m .76581 .01637 L .76168 .01111 L closepath p .748 .925 1 r
F P
0 g
s
.76581 .01637 m .76781 .01251 L .76772 .01929 L closepath p .451 .557 .981 r
F P
0 g
s
.70415 .01255 m .70226 .01638 L .69865 .01111 L closepath p .734 .912 1 r
F P
0 g
s
.70226 .01638 m .70415 .01255 L .70393 .01926 L closepath p .407 .512 .954 r
F P
0 g
s
.37699 .01254 m .37592 .01639 L .3712 .01111 L closepath p .725 .916 1 r
F P
0 g
s
.37592 .01639 m .37699 .01254 L .37767 .01929 L closepath p .304 .457 .932 r
F P
0 g
s
.59144 .01835 m .58928 .01624 L .59181 .0164 L closepath p 1 1 .724 r
F P
0 g
s
.59031 .0129 m .59181 .0164 L .58928 .01624 L closepath p 1 1 .777 r
F P
0 g
s
.05448 .01897 m .05096 .01685 L .05521 .01642 L closepath p 1 1 .869 r
F P
0 g
s
.05258 .01278 m .05521 .01642 L .05096 .01685 L closepath p 1 1 .902 r
F P
0 g
s
.17908 .01986 m .17591 .0181 L .17857 .01657 L closepath p 1 .898 .587 r
F P
0 g
s
.17537 .01445 m .17857 .01657 L .17591 .0181 L closepath p 1 .901 .59 r
F P
0 g
s
.88543 .01975 m .88406 .01799 L .88475 .01657 L closepath p .973 .684 .172 r
F P
0 g
s
.88333 .01445 m .88475 .01657 L .88406 .01799 L closepath p .972 .687 .174 r
F P
0 g
s
.61792 .01969 m .61587 .01793 L .61729 .01657 L closepath p 1 .796 .361 r
F P
0 g
s
.6152 .01445 m .61729 .01657 L .61587 .01793 L closepath p 1 .8 .364 r
F P
0 g
s
.16804 .01266 m .16755 .01659 L .16346 .0108 L closepath p .653 .852 1 r
F P
0 g
s
.16755 .01659 m .16804 .01266 L .17032 .02073 L closepath p .252 .444 .932 r
F P
0 g
s
.81001 .01593 m .8073 .01666 L .81441 .01137 L closepath p .117 .117 .117 r
F P
0 g
s
.8073 .01666 m .81001 .01593 L .80178 .02803 L closepath p .154 .154 .321 r
F P
0 g
s
.13338 .01274 m .1328 .01668 L .13067 .01175 L closepath p .675 .87 1 r
F P
0 g
s
.1328 .01668 m .13338 .01274 L .13392 .01846 L closepath p .243 .426 .918 r
F P
0 g
s
.15287 .01959 m .14961 .01748 L .15407 .01669 L closepath p 1 1 .909 r
F P
0 g
s
.15173 .01299 m .15407 .01669 L .14961 .01748 L closepath p 1 1 .925 r
F P
0 g
s
.2211 .0146 m .21942 .01672 L .21984 .01192 L closepath p .174 .329 .845 r
F P
0 g
s
.92899 .01461 m .92553 .01673 L .92761 .01193 L closepath p .463 .53 .95 r
F P
0 g
s
.92553 .01673 m .92899 .01461 L .92524 .0224 L closepath p .311 .24 .687 r
F P
0 g
s
.66086 .01461 m .65807 .01673 L .65952 .01193 L closepath p .377 .472 .924 r
F P
0 g
s
.65807 .01673 m .66086 .01461 L .65767 .0224 L closepath p .175 .175 .602 r
F P
0 g
s
.26446 .0132 m .26338 .01684 L .26125 .01142 L closepath p .594 .792 1 r
F P
0 g
s
.26338 .01684 m .26446 .0132 L .26432 .01926 L closepath p .166 .272 .798 r
F P
0 g
s
.42474 .01327 m .42326 .01689 L .42169 .0115 L closepath p .606 .789 1 r
F P
0 g
s
.42326 .01689 m .42474 .01327 L .42418 .01927 L closepath p .22 .332 .835 r
F P
0 g
s
.49683 .01399 m .49478 .01726 L .49454 .01171 L closepath p .507 .66 1 r
F P
0 g
s
.49478 .01726 m .49683 .01399 L .49552 .01903 L closepath p .317 .409 .881 r
F P
0 g
s
.47468 .01322 m .47352 .01727 L .47039 .01112 L closepath p .655 .827 1 r
F P
0 g
s
.47352 .01727 m .47468 .01322 L .47584 .0196 L closepath p .493 .651 1 r
F P
0 g
s
.3418 .02544 m .33901 .02353 L .34146 .01734 L closepath p 1 .946 .624 r
F P
0 g
s
.33866 .01522 m .34146 .01734 L .33901 .02353 L closepath p 1 .946 .624 r
F P
0 g
s
.28167 .01974 m .27873 .01773 L .28154 .01734 L closepath p 1 .943 .631 r
F P
0 g
s
.2786 .01522 m .28154 .01734 L .27873 .01773 L closepath p 1 .943 .632 r
F P
0 g
s
.78342 .01523 m .78509 .01735 L .78365 .01814 L closepath p 1 .847 .398 r
F P
0 g
s
.71941 .01523 m .72123 .01735 L .71962 .01756 L closepath p 1 .854 .418 r
F P
0 g
s
.39499 .01975 m .39234 .01774 L .39484 .01735 L closepath p 1 .929 .589 r
F P
0 g
s
.39218 .01523 m .39484 .01735 L .39234 .01774 L closepath p 1 .93 .59 r
F P
0 g
s
.60919 .01373 m .60722 .01739 L .60705 .01181 L closepath p .554 .698 1 r
F P
0 g
s
.60722 .01739 m .60919 .01373 L .60917 .02074 L closepath p .408 .515 .956 r
F P
0 g
s
.87714 .01371 m .87453 .0174 L .87407 .01142 L closepath p .624 .754 1 r
F P
0 g
s
.87453 .0174 m .87714 .01371 L .8767 .02072 L closepath p .5 .585 .989 r
F P
0 g
s
.57658 .01429 m .57431 .01745 L .57513 .01216 L closepath p .439 .557 .986 r
F P
0 g
s
.57431 .01745 m .57658 .01429 L .57507 .02072 L closepath p .248 .294 .785 r
F P
0 g
s
.44122 .0196 m .4387 .01781 L .44109 .01752 L closepath p 1 .926 .575 r
F P
0 g
s
.43855 .0154 m .44109 .01752 L .4387 .01781 L closepath p 1 .925 .574 r
F P
0 g
s
.53589 .01593 m .53343 .01777 L .53654 .01154 L closepath p .157 .157 .49 r
F P
0 g
s
.53343 .01777 m .53589 .01593 L .53271 .02282 L closepath p .157 .157 .494 r
F P
0 g
s
.50303 .01577 m .50541 .01789 L .50334 .01952 L closepath p 1 .904 .53 r
F P
0 g
s
.04395 .01462 m .043 .01789 L .04138 .01199 L closepath p .384 .594 1 r
F P
0 g
s
.82175 .0217 m .82054 .01833 L .82483 .01793 L closepath p 1 1 1 r
F P
0 g
s
.82398 .01419 m .82483 .01793 L .82054 .01833 L closepath p 1 1 1 r
F P
0 g
s
.49148 .02539 m .48919 .02398 L .48786 .01812 L closepath p .996 .592 .148 r
F P
0 g
s
.48569 .01705 m .48786 .01812 L .48919 .02398 L closepath p .999 .593 .153 r
F P
0 g
s
.59198 .02861 m .58983 .02669 L .59144 .01835 L closepath p 1 .91 .522 r
F P
0 g
s
.58928 .01624 m .59144 .01835 L .58983 .02669 L closepath p 1 .91 .522 r
F P
0 g
s
.1351 .0153 m .13392 .01846 L .13338 .01274 L closepath p .286 .466 .944 r
F P
0 g
s
.85403 .01468 m .85167 .01859 L .84633 .01032 L closepath p .668 .809 1 r
F P
0 g
s
.85167 .01859 m .85403 .01468 L .85649 .02299 L closepath p .591 .71 1 r
F P
0 g
s
.0549 .02385 m .05139 .02196 L .05448 .01897 L closepath p 1 .939 .669 r
F P
0 g
s
.05096 .01685 m .05448 .01897 L .05139 .02196 L closepath p 1 .94 .67 r
F P
0 g
s
.08842 .01717 m .0871 .019 L .08935 .01208 L closepath p .085 .085 .19 r
F P
0 g
s
.0871 .019 m .08842 .01717 L .08605 .02495 L closepath p .086 .086 .197 r
F P
0 g
s
.49772 .01604 m .49552 .01903 L .49683 .01399 L closepath p .326 .418 .888 r
F P
0 g
s
.49552 .01903 m .49772 .01604 L .49682 .02284 L closepath p .283 .359 .841 r
F P
0 g
s
.26607 .0167 m .26432 .01926 L .26446 .0132 L closepath p .228 .363 .864 r
F P
0 g
s
.70678 .0167 m .70393 .01926 L .70415 .01255 L closepath p .447 .55 .976 r
F P
0 g
s
.70393 .01926 m .70678 .0167 L .70475 .02327 L closepath p .315 .334 .801 r
F P
0 g
s
.32607 .01671 m .32417 .01927 L .32328 .01251 L closepath p .332 .491 .956 r
F P
0 g
s
.42634 .01671 m .42418 .01927 L .42474 .01327 L closepath p .301 .412 .89 r
F P
0 g
s
.42418 .01927 m .42634 .01671 L .42447 .02093 L closepath p .174 .208 .72 r
F P
0 g
s
.77073 .0167 m .76772 .01929 L .76781 .01251 L closepath p .478 .582 .995 r
F P
0 g
s
.76772 .01929 m .77073 .0167 L .76816 .02138 L closepath p .343 .357 .814 r
F P
0 g
s
.3797 .01669 m .37767 .01929 L .37699 .01254 L closepath p .351 .499 .958 r
F P
0 g
s
.37767 .01929 m .3797 .01669 L .37823 .02258 L closepath p .17 .186 .705 r
F P
0 g
s
.71962 .01756 m .72144 .01957 L .72022 .022 L closepath p 1 .819 .372 r
F P
0 g
s
.1531 .03134 m .14983 .02941 L .15287 .01959 L closepath p 1 .965 .691 r
F P
0 g
s
.14961 .01748 m .15287 .01959 L .14983 .02941 L closepath p 1 .965 .691 r
F P
0 g
s
.47764 .01638 m .47584 .0196 L .47468 .01322 L closepath p .463 .626 1 r
F P
0 g
s
.47584 .0196 m .47764 .01638 L .47877 .02306 L closepath p .439 .595 1 r
F P
0 g
s
.44571 .05111 m .44349 .05139 L .44122 .0196 L closepath p 1 .877 .505 r
F P
0 g
s
.4387 .01781 m .44122 .0196 L .44349 .05139 L closepath p 1 .877 .505 r
F P
0 g
s
.61979 .02443 m .61783 .02302 L .61792 .01969 L closepath p 1 .632 .165 r
F P
0 g
s
.61587 .01793 m .61792 .01969 L .61783 .02302 L closepath p .999 .633 .163 r
F P
0 g
s
.28213 .02413 m .27919 .02223 L .28167 .01974 L closepath p 1 .919 .596 r
F P
0 g
s
.27873 .01773 m .28167 .01974 L .27919 .02223 L closepath p 1 .919 .597 r
F P
0 g
s
.39548 .02414 m .39283 .02225 L .39499 .01975 L closepath p 1 .903 .551 r
F P
0 g
s
.39234 .01774 m .39499 .01975 L .39283 .02225 L closepath p 1 .903 .551 r
F P
0 g
s
.88736 .02452 m .88609 .02309 L .88543 .01975 L closepath p .857 .491 .058 r
F P
0 g
s
.88406 .01799 m .88543 .01975 L .88609 .02309 L closepath p .852 .491 .056 r
F P
0 g
s
.18077 .02466 m .17768 .02325 L .17908 .01986 L closepath p 1 .783 .436 r
F P
0 g
s
.17591 .0181 m .17908 .01986 L .17768 .02325 L closepath p 1 .785 .437 r
F P
0 g
s
.94013 .03341 m .9389 .03081 L .94316 .01989 L closepath p 1 1 .635 r
F P
0 g
s
.23307 .03333 m .23007 .03073 L .23681 .01989 L closepath p 1 1 .848 r
F P
0 g
s
.23385 .01716 m .23681 .01989 L .23007 .03073 L closepath p 1 1 .848 r
F P
0 g
s
.67229 .03338 m .67039 .03077 L .67564 .01989 L closepath p 1 1 .748 r
F P
0 g
s
.67377 .01717 m .67564 .01989 L .67039 .03077 L closepath p 1 1 .748 r
F P
0 g
s
.04494 .01708 m .04386 .02006 L .04395 .01462 L closepath p .158 .243 .781 r
F P
0 g
s
.78589 .02468 m .78424 .02278 L .78531 .02015 L closepath p 1 .804 .341 r
F P
0 g
s
.78365 .01814 m .78531 .02015 L .78424 .02278 L closepath p 1 .804 .341 r
F P
0 g
s
.56092 .04309 m .55895 .04209 L .54879 .02036 L closepath p .962 .545 .106 r
F P
0 g
s
.54677 .01927 m .54879 .02036 L .55895 .04209 L closepath p .962 .545 .106 r
F P
0 g
s
.8798 .01763 m .8767 .02072 L .87714 .01371 L closepath p .514 .598 .996 r
F P
0 g
s
.8767 .02072 m .8798 .01763 L .88223 .03174 L closepath p .471 .531 .947 r
F P
0 g
s
.57762 .01826 m .57507 .02072 L .57658 .01429 L closepath p .282 .33 .811 r
F P
0 g
s
.57507 .02072 m .57762 .01826 L .57746 .03731 L closepath p .229 .247 .738 r
F P
0 g
s
.17167 .01763 m .17032 .02073 L .16804 .01266 L closepath p .329 .506 .97 r
F P
0 g
s
.17032 .02073 m .17167 .01763 L .17534 .03174 L closepath p .192 .337 .848 r
F P
0 g
s
.6116 .01763 m .60917 .02074 L .60919 .01373 L closepath p .432 .537 .969 r
F P
0 g
s
.60917 .02074 m .6116 .01763 L .61453 .03175 L closepath p .389 .476 .924 r
F P
0 g
s
.42663 .01839 m .42447 .02093 L .42634 .01671 L closepath p .174 .209 .72 r
F P
0 g
s
.42447 .02093 m .42663 .01839 L .42495 .02396 L closepath p .173 .195 .708 r
F P
0 g
s
.3264 .01863 m .32449 .02117 L .32607 .01671 L closepath p .165 .165 .69 r
F P
0 g
s
.77118 .01883 m .76816 .02138 L .77073 .0167 L closepath p .343 .357 .815 r
F P
0 g
s
.76816 .02138 m .77118 .01883 L .7688 .02473 L closepath p .335 .344 .803 r
F P
0 g
s
.50334 .01952 m .50571 .0215 L .50428 .02458 L closepath p 1 .835 .435 r
F P
0 g
s
.81806 .03016 m .81657 .02745 L .82175 .0217 L closepath p 1 1 .833 r
F P
0 g
s
.82054 .01833 m .82175 .0217 L .81657 .02745 L closepath p 1 1 .834 r
F P
0 g
s
.11724 .04791 m .11412 .04694 L .10421 .02188 L closepath p 1 .703 .359 r
F P
0 g
s
.10104 .02081 m .10421 .02188 L .11412 .04694 L closepath p 1 .703 .359 r
F P
0 g
s
.13617 .01988 m .13473 .02234 L .1351 .0153 L closepath p .15 .15 .673 r
F P
0 g
s
.22048 .02058 m .21883 .02239 L .2211 .0146 L closepath p .119 .119 .369 r
F P
0 g
s
.92869 .02059 m .92524 .0224 L .92899 .01461 L closepath p .311 .24 .687 r
F P
0 g
s
.92524 .0224 m .92869 .02059 L .92405 .02936 L closepath p .271 .188 .607 r
F P
0 g
s
.66044 .02059 m .65767 .0224 L .66086 .01461 L closepath p .175 .175 .602 r
F P
0 g
s
.38028 .02008 m .37823 .02258 L .3797 .01669 L closepath p .17 .186 .705 r
F P
0 g
s
.37823 .02258 m .38028 .02008 L .37923 .02921 L closepath p .168 .168 .688 r
F P
0 g
s
.53516 .02104 m .53271 .02282 L .53589 .01593 L closepath p .157 .157 .494 r
F P
0 g
s
.49909 .02004 m .49682 .02284 L .49772 .01604 L closepath p .286 .361 .843 r
F P
0 g
s
.49682 .02284 m .49909 .02004 L .50881 .05891 L closepath p .282 .355 .838 r
F P
0 g
s
.8591 .01932 m .85649 .02299 L .85403 .01468 L closepath p .591 .71 1 r
F P
0 g
s
.85649 .02299 m .8591 .01932 L .86164 .02908 L closepath p .547 .649 1 r
F P
0 g
s
.48061 .01954 m .47877 .02306 L .47764 .01638 L closepath p .493 .64 1 r
F P
0 g
s
.47877 .02306 m .48061 .01954 L .48468 .02984 L closepath p .476 .619 1 r
F P
0 g
s
.70761 .02074 m .70475 .02327 L .70678 .0167 L closepath p .315 .334 .801 r
F P
0 g
s
.70475 .02327 m .70761 .02074 L .70576 .02865 L closepath p .308 .323 .791 r
F P
0 g
s
.87192 .03154 m .87054 .02992 L .86974 .02382 L closepath p .944 .624 .106 r
F P
0 g
s
.86838 .02232 m .86974 .02382 L .87054 .02992 L closepath p .944 .624 .106 r
F P
0 g
s
.05639 .03156 m .05291 .02989 L .0549 .02385 L closepath p 1 .888 .597 r
F P
0 g
s
.05139 .02196 m .0549 .02385 L .05291 .02989 L closepath p 1 .889 .597 r
F P
0 g
s
.723 .03037 m .7212 .02856 L .72203 .0239 L closepath p 1 .807 .356 r
F P
0 g
s
.72022 .022 m .72203 .0239 L .7212 .02856 L closepath p 1 .807 .356 r
F P
0 g
s
.42711 .02146 m .42495 .02396 L .42663 .01839 L closepath p .173 .195 .708 r
F P
0 g
s
.42495 .02396 m .42711 .02146 L .42562 .02839 L closepath p .173 .19 .704 r
F P
0 g
s
.2829 .03053 m .27997 .02873 L .28213 .02413 L closepath p 1 .909 .583 r
F P
0 g
s
.27919 .02223 m .28213 .02413 L .27997 .02873 L closepath p 1 .909 .583 r
F P
0 g
s
.3963 .03056 m .39366 .02876 L .39548 .02414 L closepath p 1 .892 .536 r
F P
0 g
s
.39283 .02225 m .39548 .02414 L .39366 .02876 L closepath p 1 .892 .536 r
F P
0 g
s
.3269 .02187 m .32499 .02438 L .3264 .01863 L closepath p .163 .163 .675 r
F P
0 g
s
.62293 .03084 m .62108 .02971 L .61979 .02443 L closepath p .954 .55 .099 r
F P
0 g
s
.61783 .02302 m .61979 .02443 L .62108 .02971 L closepath p .951 .549 .098 r
F P
0 g
s
.04638 .02164 m .04523 .02444 L .04494 .01708 L closepath p .149 .161 .714 r
F P
0 g
s
.89058 .03092 m .88942 .02977 L .88736 .02452 L closepath p .799 .4 .048 r
F P
0 g
s
.88609 .02309 m .88736 .02452 L .88942 .02977 L closepath p .794 .397 .046 r
F P
0 g
s
.17768 .02325 m .18077 .02466 L .18068 .02988 L closepath p 1 .716 .356 r
F P
0 g
s
.78684 .03095 m .7852 .02914 L .78589 .02468 L closepath p 1 .785 .315 r
F P
0 g
s
.78424 .02278 m .78589 .02468 L .7852 .02914 L closepath p 1 .785 .315 r
F P
0 g
s
.77183 .02221 m .7688 .02473 L .77118 .01883 L closepath p .335 .345 .804 r
F P
0 g
s
.7688 .02473 m .77183 .02221 L .76964 .02934 L closepath p .332 .339 .798 r
F P
0 g
s
.08735 .02318 m .08605 .02495 L .08842 .01717 L closepath p .086 .086 .197 r
F P
0 g
s
.49277 .02912 m .49036 .02703 L .49148 .02539 L closepath p 1 .705 .267 r
F P
0 g
s
.48919 .02398 m .49148 .02539 L .49036 .02703 L closepath p 1 .688 .258 r
F P
0 g
s
.34299 .03527 m .34023 .03356 L .3418 .02544 L closepath p 1 .903 .562 r
F P
0 g
s
.33901 .02353 m .3418 .02544 L .34023 .03356 L closepath p 1 .903 .562 r
F P
0 g
s
.50428 .02458 m .5066 .02639 L .50583 .03096 L closepath p 1 .789 .374 r
F P
0 g
s
.13671 .02443 m .13525 .02685 L .13617 .01988 L closepath p .136 .136 .564 r
F P
0 g
s
.80481 .02663 m .80178 .02803 L .81001 .01593 L closepath p .153 .153 .317 r
F P
0 g
s
.80178 .02803 m .80481 .02663 L .80013 .04233 L closepath p .22 .182 .605 r
F P
0 g
s
.42779 .02591 m .42562 .02839 L .42711 .02146 L closepath p .173 .19 .704 r
F P
0 g
s
.42562 .02839 m .42779 .02591 L .42648 .0342 L closepath p .172 .188 .701 r
F P
0 g
s
.5935 .04117 m .59137 .03942 L .59198 .02861 L closepath p 1 .864 .457 r
F P
0 g
s
.58983 .02669 m .59198 .02861 L .59137 .03942 L closepath p 1 .864 .457 r
F P
0 g
s
.70863 .02615 m .70576 .02865 L .70761 .02074 L closepath p .308 .323 .791 r
F P
0 g
s
.70576 .02865 m .70863 .02615 L .70696 .03538 L closepath p .304 .316 .785 r
F P
0 g
s
.32758 .02643 m .32566 .02891 L .3269 .02187 L closepath p .163 .163 .669 r
F P
0 g
s
.86446 .02564 m .86164 .02908 L .8591 .01932 L closepath p .548 .65 1 r
F P
0 g
s
.86164 .02908 m .86446 .02564 L .86584 .03396 L closepath p .551 .654 1 r
F P
0 g
s
.49133 .0315 m .48943 .0275 L .49277 .02912 L closepath p 1 1 .965 r
F P
0 g
s
.49036 .02703 m .49277 .02912 L .48943 .0275 L closepath p .954 1 1 r
F P
0 g
s
.38133 .02707 m .37923 .02921 L .38028 .02008 L closepath p .168 .168 .688 r
F P
0 g
s
.77267 .02685 m .76964 .02934 L .77183 .02221 L closepath p .332 .339 .798 r
F P
0 g
s
.76964 .02934 m .77267 .02685 L .77068 .03521 L closepath p .33 .335 .795 r
F P
0 g
s
.21882 .02781 m .21722 .02936 L .22048 .02058 L closepath p .103 .103 .231 r
F P
0 g
s
.92746 .02782 m .92405 .02936 L .92869 .02059 L closepath p .272 .189 .608 r
F P
0 g
s
.92405 .02936 m .92746 .02782 L .92262 .03566 L closepath p .255 .185 .569 r
F P
0 g
s
.65907 .02782 m .65635 .02936 L .66044 .02059 L closepath p .166 .166 .505 r
F P
0 g
s
.48596 .02586 m .48468 .02984 L .48061 .01954 L closepath p .452 .602 1 r
F P
0 g
s
.48468 .02984 m .48596 .02586 L .48688 .03171 L closepath p .542 .703 1 r
F P
0 g
s
.81701 .04039 m .81543 .03833 L .81806 .03016 L closepath p 1 .967 .567 r
F P
0 g
s
.81657 .02745 m .81806 .03016 L .81543 .03833 L closepath p 1 .966 .564 r
F P
0 g
s
.53391 .02854 m .53147 .03026 L .53516 .02104 L closepath p .155 .155 .472 r
F P
0 g
s
.72437 .03898 m .72257 .03723 L .723 .03037 L closepath p 1 .8 .348 r
F P
0 g
s
.7212 .02856 m .723 .03037 L .72257 .03723 L closepath p 1 .8 .348 r
F P
0 g
s
.28399 .03897 m .28107 .03723 L .2829 .03053 L closepath p 1 .904 .576 r
F P
0 g
s
.27997 .02873 m .2829 .03053 L .28107 .03723 L closepath p 1 .904 .576 r
F P
0 g
s
.39366 .02876 m .3963 .03056 L .39484 .03727 L closepath p 1 .886 .528 r
F P
0 g
s
.6274 .03897 m .62564 .03801 L .62293 .03084 L closepath p .924 .502 .093 r
F P
0 g
s
.62108 .02971 m .62293 .03084 L .62564 .03801 L closepath p .923 .502 .093 r
F P
0 g
s
.89514 .039 m .89406 .03801 L .89058 .03092 L closepath p .765 .348 .042 r
F P
0 g
s
.88942 .02977 m .89058 .03092 L .89406 .03801 L closepath p .763 .346 .041 r
F P
0 g
s
.78816 .03897 m .78653 .03723 L .78684 .03095 L closepath p 1 .773 .301 r
F P
0 g
s
.7852 .02914 m .78684 .03095 L .78653 .03723 L closepath p 1 .773 .301 r
F P
0 g
s
.18068 .02988 m .18367 .03101 L .1849 .03801 L closepath p 1 .673 .308 r
F P
0 g
s
.1543 .04568 m .15104 .04394 L .1531 .03134 L closepath p 1 .936 .647 r
F P
0 g
s
.14983 .02941 m .1531 .03134 L .15104 .04394 L closepath p 1 .936 .647 r
F P
0 g
s
.48853 .03236 m .4885 .02797 L .49133 .0315 L closepath p .948 1 1 r
F P
0 g
s
.48943 .0275 m .49133 .0315 L .4885 .02797 L closepath p 1 1 1 r
F P
0 g
s
.87216 .03255 m .87072 .03069 L .87192 .03154 L closepath p .972 .673 .162 r
F P
0 g
s
.87054 .02992 m .87192 .03154 L .87072 .03069 L closepath p .972 .672 .161 r
F P
0 g
s
.05897 .04215 m .05552 .04067 L .05639 .03156 L closepath p 1 .862 .563 r
F P
0 g
s
.05291 .02989 m .05639 .03156 L .05552 .04067 L closepath p 1 .862 .563 r
F P
0 g
s
.48723 .02692 m .48688 .03171 L .48596 .02586 L closepath p .55 .709 1 r
F P
0 g
s
.48688 .03171 m .48723 .02692 L .48853 .03236 L closepath p .716 .866 1 r
F P
0 g
s
.88534 .02868 m .88223 .03174 L .8798 .01763 L closepath p .471 .531 .947 r
F P
0 g
s
.88223 .03174 m .88534 .02868 L .88906 .04521 L closepath p .473 .533 .949 r
F P
0 g
s
.1767 .02868 m .17534 .03174 L .17167 .01763 L closepath p .192 .337 .848 r
F P
0 g
s
.17534 .03174 m .1767 .02868 L .18153 .04521 L closepath p .195 .34 .85 r
F P
0 g
s
.61698 .02868 m .61453 .03175 L .6116 .01763 L closepath p .389 .476 .924 r
F P
0 g
s
.61453 .03175 m .61698 .02868 L .62115 .04521 L closepath p .391 .478 .926 r
F P
0 g
s
.54556 .0455 m .54356 .04269 L .54725 .03187 L closepath p 1 1 .665 r
F P
0 g
s
.54696 .02673 m .54725 .03187 L .54356 .04269 L closepath p 1 .972 .545 r
F P
0 g
s
.54725 .03187 m .54696 .02673 L .55281 .04412 L closepath p .145 .219 .77 r
F P
0 g
s
.4885 .02797 m .48853 .03236 L .48723 .02692 L closepath p .482 .68 1 r
F P
0 g
s
.51004 .03993 m .508 .03868 L .50805 .03254 L closepath p 1 .753 .318 r
F P
0 g
s
.50583 .03096 m .50805 .03254 L .508 .03868 L closepath p 1 .754 .327 r
F P
0 g
s
.87111 .03489 m .87024 .03108 L .87216 .03255 L closepath p 1 1 .828 r
F P
0 g
s
.87072 .03069 m .87216 .03255 L .87024 .03108 L closepath p .996 1 1 r
F P
0 g
s
.2293 .04937 m .22627 .0469 L .23307 .03333 L closepath p 1 1 .826 r
F P
0 g
s
.23007 .03073 m .23307 .03333 L .22627 .0469 L closepath p 1 1 .826 r
F P
0 g
s
.66896 .04938 m .66704 .0469 L .67229 .03338 L closepath p 1 1 .718 r
F P
0 g
s
.67039 .03077 m .67229 .03338 L .66704 .0469 L closepath p 1 1 .718 r
F P
0 g
s
.13736 .03102 m .1359 .03338 L .13671 .02443 L closepath p .134 .134 .545 r
F P
0 g
s
.93712 .04938 m .93587 .0469 L .94013 .03341 L closepath p 1 .996 .6 r
F P
0 g
s
.9389 .03081 m .94013 .03341 L .93587 .0469 L closepath p 1 .996 .6 r
F P
0 g
s
.08565 .03181 m .08435 .03352 L .08735 .02318 L closepath p .083 .083 .172 r
F P
0 g
s
.86808 .03001 m .86584 .03396 L .86446 .02564 L closepath p .531 .637 1 r
F P
0 g
s
.86584 .03396 m .86808 .03001 L .86889 .03586 L closepath p .652 .791 1 r
F P
0 g
s
.42866 .03175 m .42648 .0342 L .42779 .02591 L closepath p .172 .188 .701 r
F P
0 g
s
.42648 .0342 m .42866 .03175 L .42753 .0414 L closepath p .172 .186 .7 r
F P
0 g
s
.10078 .04254 m .09836 .03924 L .10209 .03455 L closepath p 1 1 .752 r
F P
0 g
s
.10078 .02945 m .10209 .03455 L .09836 .03924 L closepath p 1 1 .751 r
F P
0 g
s
.32842 .03232 m .32649 .03476 L .32758 .02643 L closepath p .162 .162 .666 r
F P
0 g
s
.86889 .03586 m .86977 .03147 L .87111 .03489 L closepath p .97 1 1 r
F P
0 g
s
.87024 .03108 m .87111 .03489 L .86977 .03147 L closepath p 1 1 1 r
F P
0 g
s
.26871 .03258 m .26693 .03503 L .26607 .0167 L closepath p .158 .158 .664 r
F P
0 g
s
.47011 .0358 m .46853 .03579 L .46777 .03506 L closepath p .375 .043 .043 r
F P
0 g
s
.46768 .03529 m .46777 .03506 L .46853 .03579 L closepath p .446 .023 .023 r
F P
0 g
s
.46777 .03506 m .46768 .03529 L .4648 .03685 L closepath p .014 .014 .014 r
F P
0 g
s
.77372 .03276 m .77068 .03521 L .77267 .02685 L closepath p .33 .335 .795 r
F P
0 g
s
.77068 .03521 m .77372 .03276 L .77192 .04235 L closepath p .328 .333 .793 r
F P
0 g
s
.34444 .04417 m .3417 .04271 L .34299 .03527 L closepath p 1 .879 .531 r
F P
0 g
s
.34023 .03356 m .34299 .03527 L .3417 .04271 L closepath p 1 .879 .53 r
F P
0 g
s
.70984 .03292 m .70696 .03538 L .70863 .02615 L closepath p .304 .316 .785 r
F P
0 g
s
.70696 .03538 m .70984 .03292 L .70835 .04349 L closepath p .301 .312 .781 r
F P
0 g
s
.21704 .03435 m .21553 .03566 L .21882 .02781 L closepath p .096 .096 .177 r
F P
0 g
s
.92601 .03435 m .92262 .03566 L .92746 .02782 L closepath p .26 .186 .574 r
F P
0 g
s
.92262 .03566 m .92601 .03435 L .92152 .04065 L closepath p .26 .186 .576 r
F P
0 g
s
.65749 .03435 m .6548 .03566 L .65907 .02782 L closepath p .162 .162 .465 r
F P
0 g
s
.6548 .03566 m .65749 .03435 L .65361 .04065 L closepath p .163 .163 .467 r
F P
0 g
s
.47178 .03779 m .46938 .0363 L .47011 .0358 L closepath p .802 .308 .077 r
F P
0 g
s
.46853 .03579 m .47011 .0358 L .46938 .0363 L closepath p .613 .071 .071 r
F P
0 g
s
.86977 .03147 m .86889 .03586 L .86808 .03001 L closepath p .542 .7 1 r
F P
0 g
s
.46647 .03612 m .4648 .03685 L .46768 .03529 L closepath p .033 .033 .033 r
F P
0 g
s
.4648 .03685 m .46647 .03612 L .46376 .04069 L closepath p .117 .117 .265 r
F P
0 g
s
.58003 .03496 m .57746 .03731 L .57762 .01826 L closepath p .229 .247 .738 r
F P
0 g
s
.57746 .03731 m .58003 .03496 L .57984 .05749 L closepath p .216 .227 .72 r
F P
0 g
s
.47286 .04638 m .47017 .04401 L .47178 .03779 L closepath p 1 .893 .536 r
F P
0 g
s
.46938 .0363 m .47178 .03779 L .47017 .04401 L closepath p 1 .897 .527 r
F P
0 g
s
.38284 .03766 m .38062 .03896 L .38133 .02707 L closepath p .167 .167 .682 r
F P
0 g
s
.1849 .03801 m .18782 .03896 L .18687 .04051 L closepath p .964 .495 .132 r
F P
0 g
s
.6284 .04053 m .62763 .04051 L .6274 .03897 L closepath p .794 .347 .058 r
F P
0 g
s
.62564 .03801 m .6274 .03897 L .62763 .04051 L closepath p .789 .303 .069 r
F P
0 g
s
.29065 .08911 m .28773 .08749 L .28399 .03897 L closepath p 1 .902 .574 r
F P
0 g
s
.28107 .03723 m .28399 .03897 L .28773 .08749 L closepath p 1 .902 .574 r
F P
0 g
s
.79139 .05799 m .78976 .05631 L .78816 .03897 L closepath p 1 .769 .295 r
F P
0 g
s
.78653 .03723 m .78816 .03897 L .78976 .05631 L closepath p 1 .769 .295 r
F P
0 g
s
.73243 .08912 m .73065 .08749 L .72437 .03898 L closepath p 1 .798 .345 r
F P
0 g
s
.72257 .03723 m .72437 .03898 L .73065 .08749 L closepath p 1 .798 .345 r
F P
0 g
s
.89621 .04049 m .89623 .04051 L .89514 .039 L closepath p 0 0 .413 r
F P
0 g
s
.89406 .03801 m .89514 .039 L .89623 .04051 L closepath p .592 .107 .016 r
F P
0 g
s
.39484 .03727 m .39747 .03901 L .39639 .04824 L closepath p 1 .884 .525 r
F P
0 g
s
.51256 .04854 m .5108 .04772 L .51004 .03993 L closepath p 1 .716 .26 r
F P
0 g
s
.508 .03868 m .51004 .03993 L .5108 .04772 L closepath p 1 .722 .281 r
F P
0 g
s
.53214 .03843 m .5297 .04009 L .53391 .02854 L closepath p .154 .154 .462 r
F P
0 g
s
.81786 .0492 m .81631 .04744 L .81701 .04039 L closepath p 1 .82 .355 r
F P
0 g
s
.81543 .03833 m .81701 .04039 L .81631 .04744 L closepath p 1 .82 .356 r
F P
0 g
s
.89684 .04061 m .89793 .0407 L .89621 .04049 L closepath p .01 .494 .732 r
F P
0 g
s
.89623 .04051 m .89621 .04049 L .89793 .0407 L closepath p .087 .002 .002 r
F P
0 g
s
.18687 .04051 m .18877 .0405 L .1886 .0407 L closepath p .155 .02 .02 r
F P
0 g
s
.62897 .04061 m .62931 .0407 L .6284 .04053 L closepath p .119 .004 .004 r
F P
0 g
s
.19116 .04065 m .19033 .0409 L .18941 .04061 L closepath p .113 .025 .025 r
F P
0 g
s
.1886 .0407 m .18941 .04061 L .19033 .0409 L closepath p .146 .018 .018 r
F P
0 g
s
.21525 .0409 m .21416 .04065 L .21704 .03435 L closepath p .097 .097 .18 r
F P
0 g
s
.19033 .0409 m .19116 .04065 L .21525 .0409 L closepath p .088 .023 .023 r
F P
0 g
s
.92455 .0409 m .92152 .04065 L .92601 .03435 L closepath p .258 .186 .573 r
F P
0 g
s
.89962 .0409 m .89855 .04065 L .92455 .0409 L closepath p .088 .023 .023 r
F P
0 g
s
.65591 .0409 m .65361 .04065 L .65749 .03435 L closepath p .162 .162 .463 r
F P
0 g
s
.63099 .0409 m .63063 .04065 L .65591 .0409 L closepath p .088 .023 .023 r
F P
0 g
s
.46606 .03863 m .46376 .04069 L .46647 .03612 L closepath p .146 .146 .439 r
F P
0 g
s
.46376 .04069 m .46606 .03863 L .46342 .04652 L closepath p .158 .158 .542 r
F P
0 g
s
.59494 .05095 m .59281 .04922 L .5935 .04117 L closepath p 1 .844 .431 r
F P
0 g
s
.59137 .03942 m .5935 .04117 L .59281 .04922 L closepath p 1 .844 .431 r
F P
0 g
s
.42972 .03899 m .42753 .0414 L .42866 .03175 L closepath p .172 .186 .7 r
F P
0 g
s
.42753 .0414 m .42972 .03899 L .42878 .05 L closepath p .172 .185 .699 r
F P
0 g
s
.32943 .03953 m .3275 .04194 L .32842 .03232 L closepath p .162 .162 .663 r
F P
0 g
s
.13813 .03965 m .13666 .04197 L .13736 .03102 L closepath p .133 .133 .536 r
F P
0 g
s
.06265 .05566 m .05922 .05431 L .05897 .04215 L closepath p 1 .847 .543 r
F P
0 g
s
.05552 .04067 m .05897 .04215 L .05922 .05431 L closepath p 1 .847 .543 r
F P
0 g
s
.80329 .04027 m .80013 .04233 L .80481 .02663 L closepath p .23 .183 .613 r
F P
0 g
s
.80013 .04233 m .80329 .04027 L .80129 .05298 L closepath p .314 .3 .76 r
F P
0 g
s
.77496 .03993 m .77192 .04235 L .77372 .03276 L closepath p .328 .333 .793 r
F P
0 g
s
.77192 .04235 m .77496 .03993 L .77335 .05076 L closepath p .327 .332 .792 r
F P
0 g
s
.09945 .05053 m .09635 .04775 L .10078 .04254 L closepath p 1 1 .796 r
F P
0 g
s
.09836 .03924 m .10078 .04254 L .09635 .04775 L closepath p 1 1 .8 r
F P
0 g
s
.57142 .06273 m .56956 .06193 L .56092 .04309 L closepath p .961 .545 .106 r
F P
0 g
s
.55895 .04209 m .56092 .04309 L .56956 .06193 L closepath p .961 .545 .106 r
F P
0 g
s
.83296 .00553 m .84236 .00584 L .83835 .00886 L closepath p .955 .785 .905 r
F P
0 g
s
.83296 .00553 m .83835 .00886 L .8281 .01123 L closepath p .955 .785 .905 r
F P
0 g
s
.82588 .00618 m .83296 .00553 L .82604 .01271 L closepath p .955 .785 .905 r
F P
0 g
s
.8281 .01123 m .82604 .01271 L .83296 .00553 L closepath p .955 .785 .905 r
F P
0 g
s
.08619 .0312 m .08746 .02418 L .09206 .00553 L closepath p .955 .785 .905 r
F P
0 g
s
.09161 .0059 m .09206 .00553 L .08746 .02418 L closepath p .955 .785 .905 r
F P
0 g
s
.09655 .01317 m .09206 .00553 L .09313 .00649 L closepath p .955 .785 .905 r
F P
0 g
s
.09206 .00553 m .09655 .01317 L .0803 .05697 L closepath p .955 .785 .905 r
F P
0 g
s
.71124 .04107 m .70835 .04349 L .70984 .03292 L closepath p .301 .312 .781 r
F P
0 g
s
.70835 .04349 m .71124 .04107 L .70993 .05298 L closepath p .299 .309 .779 r
F P
0 g
s
.84234 .00959 m .83835 .00886 L .84236 .00584 L closepath p .955 .785 .905 r
F P
0 g
s
.84236 .00584 m .85051 .00677 L .84234 .00959 L closepath p .955 .785 .905 r
F P
0 g
s
.08746 .02418 m .08841 .01866 L .09161 .0059 L closepath p .955 .785 .905 r
F P
0 g
s
.09117 .00626 m .09161 .0059 L .08841 .01866 L closepath p .955 .785 .905 r
F P
0 g
s
.53427 .02828 m .53522 .02207 L .53868 .00592 L closepath p .955 .785 .905 r
F P
0 g
s
.53831 .00623 m .53868 .00592 L .53522 .02207 L closepath p .955 .785 .905 r
F P
0 g
s
.53868 .00592 m .54042 .00739 L .54677 .01927 L closepath p .955 .785 .905 r
F P
0 g
s
.53868 .00592 m .54272 .01259 L .52984 .05072 L closepath p .955 .785 .905 r
F P
0 g
s
.8197 .00813 m .82588 .00618 L .82398 .01419 L closepath p .955 .785 .905 r
F P
0 g
s
.82604 .01271 m .82398 .01419 L .82588 .00618 L closepath p .955 .785 .905 r
F P
0 g
s
.53522 .02207 m .53591 .01722 L .53831 .00623 L closepath p .955 .785 .905 r
F P
0 g
s
.53793 .00654 m .53831 .00623 L .53591 .01722 L closepath p .955 .785 .905 r
F P
0 g
s
.08841 .01866 m .08903 .01462 L .09117 .00626 L closepath p .955 .785 .905 r
F P
0 g
s
.09071 .00735 m .09117 .00626 L .08903 .01462 L closepath p .955 .785 .905 r
F P
0 g
s
.55401 .04017 m .55281 .04412 L .54696 .02673 L closepath p .284 .42 .903 r
F P
0 g
s
.55281 .04412 m .55401 .04017 L .5607 .05995 L closepath p .267 .401 .89 r
F P
0 g
s
.34562 .0516 m .34317 .05187 L .34444 .04417 L closepath p 1 .881 .532 r
F P
0 g
s
.3417 .04271 m .34444 .04417 L .34317 .05187 L closepath p 1 .88 .533 r
F P
0 g
s
.09313 .00649 m .09499 .00935 L .09484 .00983 L closepath p .955 .785 .905 r
F P
0 g
s
.53591 .01722 m .53635 .01371 L .53793 .00654 L closepath p .955 .785 .905 r
F P
0 g
s
.53758 .00748 m .53793 .00654 L .53635 .01371 L closepath p .955 .785 .905 r
F P
0 g
s
.84633 .01032 m .84234 .00959 L .85051 .00677 L closepath p .955 .785 .905 r
F P
0 g
s
.85051 .00677 m .85743 .00832 L .84633 .01032 L closepath p .955 .785 .905 r
F P
0 g
s
.92544 .04478 m .92371 .04503 L .89971 .04469 L closepath p .093 .024 .024 r
F P
0 g
s
.90079 .04495 m .89971 .04469 L .92371 .04503 L p .093 .024 .024 r
F P
0 g
s
.91225 .04499 m .90079 .04495 L p .92371 .04503 L .093 .024 .024 r
F P
0 g
s
.89971 .04469 m .90079 .04495 L .89646 .04577 L closepath p .053 .053 .053 r
F P
0 g
s
.89898 .04534 m .89646 .04577 L .90079 .04495 L closepath p .01 .01 .01 r
F P
0 g
s
.90717 .06463 m .90399 .06759 L .90334 .05699 L closepath p .472 .531 .947 r
F P
0 g
s
.90399 .06759 m .90717 .06463 L .90803 .07555 L closepath p .474 .534 .949 r
F P
0 g
s
.91115 .07248 m .90803 .07555 L .90717 .06463 L closepath p .472 .532 .949 r
F P
0 g
s
.90803 .07555 m .91115 .07248 L .91127 .08155 L closepath p .481 .546 .959 r
F P
0 g
s
.89962 .0409 m .91209 .0409 L .90986 .0766 L closepath p .955 .785 .905 r
F P
0 g
s
.92371 .04503 m .92371 .04503 L .91225 .04499 L p .093 .024 .024 r
F P
0 g
s
.91417 .07822 m .91127 .08155 L .91115 .07248 L closepath p .469 .536 .953 r
F P
0 g
s
.92455 .0409 m .90986 .0766 L .91209 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.90643 .04821 m .90232 .04618 L .91224 .04499 L p .955 .785 .905 r
F P
0 g
s
.91224 .04499 m .91225 .04499 L .91225 .04499 L p .955 .785 .905 r
F P
0 g
s
.91225 .04499 m .90643 .04821 L p .91224 .04499 L .955 .785 .905 r
F P
0 g
s
.0833 .04306 m .08201 .04472 L .08565 .03181 L closepath p .081 .081 .159 r
F P
0 g
s
.91849 .08762 m .9175 .08334 L .92544 .04478 L closepath p 1 .983 .58 r
F P
0 g
s
.92371 .04503 m .92544 .04478 L .9175 .08334 L closepath p 1 .983 .58 r
F P
0 g
s
.03094 .01003 m .03148 .00715 L .06407 .00795 L closepath p .955 .785 .905 r
F P
0 g
s
.06246 .00715 m .06407 .00795 L .03148 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.12601 .01142 m .12449 .0114 L .12033 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.11974 .0101 m .12033 .00715 L .12449 .0114 L closepath p .955 .785 .905 r
F P
0 g
s
.12033 .00715 m .18712 .00715 L .15403 .0089 L closepath p .955 .785 .905 r
F P
0 g
s
.18832 .00779 m .15403 .0089 L .18712 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.21594 .01103 m .21437 .01102 L .20918 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.2095 .00985 m .20918 .00715 L .21437 .01102 L closepath p .955 .785 .905 r
F P
0 g
s
.20918 .00715 m .2441 .00715 L .22572 .00905 L closepath p .955 .785 .905 r
F P
0 g
s
.24536 .00775 m .22572 .00905 L .2441 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.18503 .01095 m .12033 .00715 L .18775 .01064 L closepath p .955 .785 .905 r
F P
0 g
s
.2369 .0125 m .20918 .00715 L .24226 .01095 L closepath p .955 .785 .905 r
F P
0 g
s
.23385 .01716 m .20918 .00715 L .2369 .0125 L closepath p .955 .785 .905 r
F P
0 g
s
.12033 .00715 m .18503 .01095 L .14892 .00938 L closepath p .955 .785 .905 r
F P
0 g
s
.20918 .00715 m .22151 .01215 L .21594 .01103 L closepath p .955 .785 .905 r
F P
0 g
s
.12033 .00715 m .14892 .00938 L .12317 .00928 L closepath p .955 .785 .905 r
F P
0 g
s
.31704 .01111 m .31274 .0102 L .35024 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.31202 .00715 m .35024 .00715 L .31274 .0102 L closepath p .955 .785 .905 r
F P
0 g
s
.3501 .0089 m .35024 .00715 L .35125 .00779 L closepath p .955 .785 .905 r
F P
0 g
s
.34514 .01111 m .35024 .00715 L .34996 .01064 L closepath p .955 .785 .905 r
F P
0 g
s
.36514 .00746 m .36559 .00715 L .40476 .00779 L closepath p .955 .785 .905 r
F P
0 g
s
.4038 .00715 m .40476 .00779 L .36559 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.4038 .00715 m .36559 .00715 L .40476 .00779 L closepath p .955 .785 .905 r
F P
0 g
s
.35024 .00715 m .33816 .01193 L .31704 .01111 L closepath p .955 .785 .905 r
F P
0 g
s
.32607 .01671 m .35024 .00715 L .3403 .01225 L closepath p .955 .785 .905 r
F P
0 g
s
.34514 .01111 m .3403 .01225 L .35024 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.25141 .00746 m .25182 .00715 L .29124 .00779 L closepath p .955 .785 .905 r
F P
0 g
s
.29028 .00715 m .29124 .00779 L .25182 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.29028 .00715 m .25182 .00715 L .29124 .00779 L closepath p .955 .785 .905 r
F P
0 g
s
.47983 .00715 m .41205 .00777 L .41286 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.47983 .00715 m .4819 .01013 L .48416 .01755 L closepath p .955 .785 .905 r
F P
0 g
s
.47983 .00715 m .48416 .01755 L .47039 .01112 L closepath p .955 .785 .905 r
F P
0 g
s
.41205 .00777 m .47983 .00715 L .45702 .01112 L closepath p .955 .785 .905 r
F P
0 g
s
.47039 .01112 m .45702 .01112 L .47983 .00715 L closepath p .955 .785 .905 r
F P
0 g
s
.49136 .01096 m .48979 .01096 L .48573 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.48534 .00989 m .48573 .00716 L .48979 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.48573 .00716 m .49933 .00716 L .50835 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.51306 .01073 m .50835 .01096 L .49933 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.49933 .00716 m .51293 .00716 L .50619 .00894 L closepath p .955 .785 .905 r
F P
0 g
s
.5142 .00787 m .50619 .00894 L .51293 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.56848 .01096 m .56397 .01012 L .62665 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.56339 .00716 m .62665 .00716 L .56397 .01012 L closepath p .955 .785 .905 r
F P
0 g
s
.62665 .00716 m .62804 .0078 L .62433 .00906 L closepath p .955 .785 .905 r
F P
0 g
s
.65558 .01104 m .65403 .01103 L .64894 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.64915 .00985 m .64894 .00716 L .65403 .01103 L closepath p .955 .785 .905 r
F P
0 g
s
.64894 .00716 m .68371 .00716 L .66542 .00906 L closepath p .955 .785 .905 r
F P
0 g
s
.68492 .00776 m .66542 .00906 L .68371 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.48573 .00716 m .50835 .01096 L .50303 .01577 L closepath p .955 .785 .905 r
F P
0 g
s
.62665 .00716 m .60089 .00966 L .56848 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.67665 .01251 m .64894 .00716 L .68191 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.67377 .01717 m .64894 .00716 L .67665 .01251 L closepath p .955 .785 .905 r
F P
0 g
s
.48573 .00716 m .49438 .01146 L .49136 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.64894 .00716 m .66135 .01216 L .65558 .01104 L closepath p .955 .785 .905 r
F P
0 g
s
.62665 .00716 m .62201 .01096 L .60705 .01181 L closepath p .955 .785 .905 r
F P
0 g
s
.60084 .01049 m .62665 .00716 L .60705 .01181 L closepath p .955 .785 .905 r
F P
0 g
s
.57513 .01216 m .62665 .00716 L .59344 .01127 L closepath p .955 .785 .905 r
F P
0 g
s
.60084 .01049 m .59344 .01127 L .62665 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.86946 .01057 m .86806 .00983 L .86817 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.86755 .00809 m .86817 .00716 L .86806 .00983 L closepath p .955 .785 .905 r
F P
0 g
s
.86817 .00716 m .88133 .00716 L .89282 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.89553 .01065 m .89282 .01096 L .88133 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.88133 .00716 m .89448 .00716 L .88843 .0089 L closepath p .955 .785 .905 r
F P
0 g
s
.89589 .0078 m .88843 .0089 L .89448 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.91663 .00762 m .91678 .00716 L .95299 .00776 L closepath p .955 .785 .905 r
F P
0 g
s
.9517 .00716 m .95299 .00776 L .91678 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.86817 .00716 m .87112 .00929 L .86946 .01057 L closepath p .955 .785 .905 r
F P
0 g
s
.89 .01096 m .86817 .00716 L .89282 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.87407 .01142 m .86817 .00716 L .88495 .01187 L closepath p .955 .785 .905 r
F P
0 g
s
.89 .01096 m .88495 .01187 L .86817 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.69414 .01088 m .69226 .00983 L .73087 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.71167 .00716 m .73087 .00716 L .69226 .00983 L closepath p .955 .785 .905 r
F P
0 g
s
.69226 .00983 m .6916 .00778 L .71167 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.69248 .00716 m .71167 .00716 L .6916 .00778 L closepath p .955 .785 .905 r
F P
0 g
s
.73067 .0089 m .73087 .00716 L .73172 .0078 L closepath p .955 .785 .905 r
F P
0 g
s
.7257 .01111 m .73087 .00716 L .73046 .01065 L closepath p .955 .785 .905 r
F P
0 g
s
.70678 .0167 m .70415 .01255 L .73087 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.71251 .00902 m .73087 .00716 L .70415 .01255 L closepath p .955 .785 .905 r
F P
0 g
s
.70678 .0167 m .73087 .00716 L .72095 .01226 L closepath p .955 .785 .905 r
F P
0 g
s
.7257 .01111 m .72095 .01226 L .73087 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.7612 .01111 m .75702 .01028 L .75638 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.75559 .00778 m .75638 .00716 L .75702 .01028 L closepath p .955 .785 .905 r
F P
0 g
s
.75559 .00778 m .75702 .01028 L .75638 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.75638 .00716 m .7755 .00716 L .7897 .01111 L closepath p .955 .785 .905 r
F P
0 g
s
.79446 .01065 m .7897 .01111 L .7755 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.7755 .00716 m .79461 .00716 L .78498 .0089 L closepath p .955 .785 .905 r
F P
0 g
s
.79565 .0078 m .78498 .0089 L .79461 .00716 L closepath p .955 .785 .905 r
F P
0 g
s
.78495 .01226 m .75638 .00716 L .7897 .01111 L closepath p .955 .785 .905 r
F P
0 g
s
.75638 .00716 m .77067 .00971 L .7612 .01111 L closepath p .955 .785 .905 r
F P
0 g
s
.08903 .01462 m .08934 .01208 L .09071 .00735 L closepath p .954 .784 .905 r
F P
0 g
s
.09026 .00844 m .09071 .00735 L .08934 .01208 L closepath p .955 .785 .905 r
F P
0 g
s
.54584 .0174 m .54677 .01927 L .54042 .00739 L closepath p .955 .785 .905 r
F P
0 g
s
.54042 .00739 m .54305 .01181 L .54584 .0174 L closepath p .955 .785 .905 r
F P
0 g
s
.8922 .04218 m .88906 .04521 L .88534 .02868 L closepath p .473 .533 .949 r
F P
0 g
s
.88906 .04521 m .8922 .04218 L .89721 .06113 L closepath p .474 .535 .95 r
F P
0 g
s
.18292 .04218 m .18153 .04521 L .1767 .02868 L closepath p .195 .341 .85 r
F P
0 g
s
.18153 .04521 m .18292 .04218 L .18892 .06114 L closepath p .197 .343 .852 r
F P
0 g
s
.62362 .04218 m .62115 .04521 L .61698 .02868 L closepath p .391 .478 .926 r
F P
0 g
s
.62115 .04521 m .62362 .04218 L .62903 .06114 L closepath p .392 .48 .927 r
F P
0 g
s
.3647 .00777 m .36514 .00746 L .40507 .00971 L closepath p .955 .785 .905 r
F P
0 g
s
.40476 .00779 m .40507 .00971 L .36514 .00746 L closepath p .955 .785 .905 r
F P
0 g
s
.251 .00777 m .25141 .00746 L .29153 .00971 L closepath p .955 .785 .905 r
F P
0 g
s
.29124 .00779 m .29153 .00971 L .25141 .00746 L closepath p .955 .785 .905 r
F P
0 g
s
.53635 .01371 m .53654 .01154 L .53758 .00748 L closepath p .954 .784 .905 r
F P
0 g
s
.53722 .00842 m .53758 .00748 L .53654 .01154 L closepath p .955 .785 .905 r
F P
0 g
s
.91648 .00809 m .91663 .00762 L .95347 .00956 L closepath p .955 .785 .905 r
F P
0 g
s
.95299 .00776 m .95347 .00956 L .91663 .00762 L closepath p .955 .785 .905 r
F P
0 g
s
.54373 .05656 m .54148 .05416 L .54556 .0455 L closepath p 1 1 .721 r
F P
0 g
s
.54356 .04269 m .54556 .0455 L .54148 .05416 L closepath p 1 1 .713 r
F P
0 g
s
.22572 .00905 m .24536 .00775 L .24226 .01095 L closepath p .955 .785 .905 r
F P
0 g
s
.24575 .00956 m .24226 .01095 L .24536 .00775 L closepath p .955 .785 .905 r
F P
0 g
s
.66542 .00906 m .68492 .00776 L .68191 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.68535 .00956 m .68191 .01096 L .68492 .00776 L closepath p .955 .785 .905 r
F P
0 g
s
.3712 .01111 m .36841 .01105 L .3647 .00777 L closepath p .955 .785 .905 r
F P
0 g
s
.36684 .01087 m .3647 .00777 L .36841 .01105 L closepath p .955 .785 .905 r
F P
0 g
s
.36795 .00944 m .3647 .00777 L .40346 .01065 L closepath p .955 .785 .905 r
F P
0 g
s
.40507 .00971 m .40346 .01065 L .3647 .00777 L closepath p .955 .785 .905 r
F P
0 g
s
.26125 .01142 m .25441 .01095 L .251 .00777 L closepath p .955 .785 .905 r
F P
0 g
s
.25183 .01004 m .251 .00777 L .25441 .01095 L closepath p .955 .785 .905 r
F P
0 g
s
.251 .00777 m .29153 .00971 L .26125 .01142 L closepath p .955 .785 .905 r
F P
0 g
s
.42169 .0115 m .41525 .01112 L .41205 .00777 L closepath p .955 .785 .905 r
F P
0 g
s
.41284 .01009 m .41205 .00777 L .41525 .01112 L closepath p .955 .785 .905 r
F P
0 g
s
.41205 .00777 m .45702 .01112 L .43984 .01219 L closepath p .955 .785 .905 r
F P
0 g
s
.41205 .00777 m .43984 .01219 L .42169 .0115 L closepath p .955 .785 .905 r
F P
0 g
s
.15403 .0089 m .18832 .00779 L .18775 .01064 L closepath p .955 .785 .905 r
F P
0 g
s
.18868 .00971 m .18775 .01064 L .18832 .00779 L closepath p .955 .785 .905 r
F P
0 g
s
.35125 .00779 m .35158 .00971 L .3501 .0089 L closepath p .955 .785 .905 r
F P
0 g
s
.62598 .00985 m .62433 .00906 L .62804 .0078 L closepath p .955 .785 .905 r
F P
0 g
s
.62804 .0078 m .62852 .00972 L .62598 .00985 L closepath p .955 .785 .905 r
F P
0 g
s
.88843 .0089 m .89589 .0078 L .89553 .01065 L closepath p .955 .785 .905 r
F P
0 g
s
.89641 .00972 m .89553 .01065 L .89589 .0078 L closepath p .955 .785 .905 r
F P
0 g
s
.73172 .0078 m .73204 .00972 L .73067 .0089 L closepath p .955 .785 .905 r
F P
0 g
s
.78498 .0089 m .79565 .0078 L .79446 .01065 L closepath p .955 .785 .905 r
F P
0 g
s
.79603 .00972 m .79446 .01065 L .79565 .0078 L closepath p .955 .785 .905 r
F P
0 g
s
.50619 .00894 m .5142 .00787 L .51306 .01073 L closepath p .955 .785 .905 r
F P
0 g
s
.51462 .01003 m .51306 .01073 L .5142 .00787 L closepath p .955 .785 .905 r
F P
0 g
s
.15729 .07205 m .15403 .07034 L .1543 .04568 L closepath p 1 .922 .627 r
F P
0 g
s
.15104 .04394 m .1543 .04568 L .15403 .07034 L closepath p 1 .922 .627 r
F P
0 g
s
.06407 .00795 m .06454 .01033 L .03094 .01003 L closepath p .955 .785 .905 r
F P
0 g
s
.89646 .04577 m .89898 .04534 L .89501 .04855 L closepath p .149 .149 .288 r
F P
0 g
s
.92365 .01104 m .9221 .01103 L .91648 .00809 L closepath p .955 .785 .905 r
F P
0 g
s
.91718 .00997 m .91648 .00809 L .9221 .01103 L closepath p .955 .785 .905 r
F P
0 g
s
.91648 .00809 m .93498 .00882 L .92365 .01104 L closepath p .955 .785 .905 r
F P
0 g
s
.81441 .01137 m .8197 .00813 L .82226 .01626 L closepath p .955 .785 .905 r
F P
0 g
s
.82398 .01419 m .82226 .01626 L .8197 .00813 L closepath p .955 .785 .905 r
F P
0 g
s
.85018 .0125 m .84633 .01032 L .85743 .00832 L closepath p .955 .785 .905 r
F P
0 g
s
.85743 .00832 m .8631 .01049 L .85018 .0125 L closepath p .955 .785 .905 r
F P
0 g
s
.47341 .05354 m .47096 .05174 L .47286 .04638 L closepath p 1 .913 .551 r
F P
0 g
s
.47017 .04401 m .47286 .04638 L .47096 .05174 L closepath p 1 .907 .556 r
F P
0 g
s
.4656 .0451 m .46342 .04652 L .46606 .03863 L closepath p .153 .153 .516 r
F P
0 g
s
.93498 .00882 m .95347 .00956 L .92761 .01193 L closepath p .955 .785 .905 r
F P
0 g
s
.92761 .01193 m .92365 .01104 L .93498 .00882 L closepath p .955 .785 .905 r
F P
0 g
s
.83289 .00945 m .8281 .01123 L .83835 .00886 L closepath p .955 .785 .905 r
F P
0 g
s
.83289 .00945 m .83835 .00886 L .8281 .01123 L closepath p .955 .785 .905 r
F P
0 g
s
.34996 .01064 m .3501 .0089 L .35158 .00971 L closepath p .955 .785 .905 r
F P
0 g
s
.73046 .01065 m .73067 .0089 L .73204 .00972 L closepath p .955 .785 .905 r
F P
0 g
s
.70415 .01255 m .69865 .01111 L .71251 .00902 L closepath p .955 .785 .905 r
F P
0 g
s
.69414 .01088 m .71251 .00902 L .69865 .01111 L closepath p .955 .785 .905 r
F P
0 g
s
.62433 .00906 m .62763 .01065 L .62201 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.12601 .01142 m .12317 .00928 L .14892 .00938 L closepath p .955 .785 .905 r
F P
0 g
s
.87112 .00929 m .87407 .01142 L .8726 .01109 L closepath p .955 .785 .905 r
F P
0 g
s
.8726 .01109 m .86946 .01057 L .87112 .00929 L closepath p .955 .785 .905 r
F P
0 g
s
.09762 .01412 m .09484 .00983 L .09499 .00935 L closepath p .955 .785 .905 r
F P
0 g
s
.18102 .01105 m .14892 .00938 L .18503 .01095 L closepath p .955 .785 .905 r
F P
0 g
s
.14892 .00938 m .18102 .01105 L .17751 .01162 L closepath p .955 .785 .905 r
F P
0 g
s
.17687 .01193 m .14892 .00938 L .17751 .01162 L closepath p .955 .785 .905 r
F P
0 g
s
.14201 .01234 m .13067 .01175 L .14892 .00938 L closepath p .955 .785 .905 r
F P
0 g
s
.12601 .01142 m .14892 .00938 L .13067 .01175 L closepath p .955 .785 .905 r
F P
0 g
s
.13617 .01988 m .13572 .01734 L .14892 .00938 L closepath p .955 .785 .905 r
F P
0 g
s
.1351 .0153 m .14892 .00938 L .13572 .01734 L closepath p .955 .785 .905 r
F P
0 g
s
.147 .02634 m .13671 .02443 L .14892 .00938 L closepath p .955 .785 .905 r
F P
0 g
s
.13617 .01988 m .14892 .00938 L .13671 .02443 L closepath p .955 .785 .905 r
F P
0 g
s
.14346 .10469 m .14235 .09106 L .14892 .00938 L closepath p .955 .785 .905 r
F P
0 g
s
.14124 .07746 m .14892 .00938 L .14235 .09106 L closepath p .955 .785 .905 r
F P
0 g
s
.14892 .00938 m .17687 .01193 L .16346 .0108 L closepath p .955 .785 .905 r
F P
0 g
s
.14892 .00938 m .16346 .0108 L .1578 .0115 L closepath p .955 .785 .905 r
F P
0 g
s
.14892 .00938 m .1578 .0115 L .15173 .01299 L closepath p .955 .785 .905 r
F P
0 g
s
.14346 .10469 m .14892 .00938 L .14961 .01748 L closepath p .955 .785 .905 r
F P
0 g
s
.15173 .01299 m .14961 .01748 L .14892 .00938 L closepath p .955 .785 .905 r
F P
0 g
s
.3712 .01111 m .36795 .00944 L .39864 .01111 L closepath p .955 .785 .905 r
F P
0 g
s
.40346 .01065 m .39864 .01111 L .36795 .00944 L closepath p .955 .785 .905 r
F P
0 g
s
.24226 .01095 m .24575 .00956 L .24486 .0106 L closepath p .955 .785 .905 r
F P
0 g
s
.68191 .01096 m .68535 .00956 L .6845 .01061 L closepath p .955 .785 .905 r
F P
0 g
s
.95006 .01096 m .95347 .00956 L .95264 .01061 L closepath p .955 .785 .905 r
F P
0 g
s
.92899 .01461 m .92761 .01193 L .95347 .00956 L closepath p .955 .785 .905 r
F P
0 g
s
.94124 .01208 m .95347 .00956 L .94477 .01251 L closepath p .955 .785 .905 r
F P
0 g
s
.95006 .01096 m .94477 .01251 L .95347 .00956 L closepath p .955 .785 .905 r
F P
0 g
s
.60089 .00966 m .57513 .01216 L .57264 .01126 L closepath p .955 .785 .905 r
F P
0 g
s
.57264 .01126 m .56848 .01096 L .60089 .00966 L closepath p .955 .785 .905 r
F P
0 g
s
.77067 .00971 m .78495 .01226 L .76775 .01251 L closepath p .955 .785 .905 r
F P
0 g
s
.76775 .01251 m .7612 .01111 L .77067 .00971 L closepath p .955 .785 .905 r
F P
0 g
s
.26446 .0132 m .26125 .01142 L .29153 .00971 L closepath p .955 .785 .905 r
F P
0 g
s
.26446 .0132 m .29153 .00971 L .27235 .01273 L closepath p .955 .785 .905 r
F P
0 g
s
.28991 .01064 m .27235 .01273 L .29153 .00971 L closepath p .955 .785 .905 r
F P
0 g
s
.62763 .01065 m .62598 .00985 L .62852 .00972 L closepath p .955 .785 .905 r
F P
0 g
s
.09484 .00983 m .09762 .01412 L .09655 .01317 L closepath p .955 .785 .905 r
F P
0 g
s
.21437 .01102 m .2128 .01101 L .2095 .00985 L closepath p .955 .785 .905 r
F P
0 g
s
.21143 .01095 m .2095 .00985 L .2128 .01101 L closepath p .955 .785 .905 r
F P
0 g
s
.65403 .01103 m .65248 .01102 L .64915 .00985 L closepath p .955 .785 .905 r
F P
0 g
s
.65107 .01096 m .64915 .00985 L .65248 .01102 L closepath p .955 .785 .905 r
F P
0 g
s
.48979 .01096 m .48822 .01096 L .48534 .00989 L closepath p .955 .785 .905 r
F P
0 g
s
.48701 .01096 m .48534 .00989 L .48822 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.9221 .01103 m .92055 .01102 L .91718 .00997 L closepath p .955 .785 .905 r
F P
0 g
s
.91914 .01096 m .91718 .00997 L .92055 .01102 L closepath p .955 .785 .905 r
F P
0 g
s
.03279 .01112 m .03094 .01003 L .06454 .01033 L closepath p .955 .785 .905 r
F P
0 g
s
.12449 .0114 m .12297 .01138 L .11974 .0101 L closepath p .955 .785 .905 r
F P
0 g
s
.12142 .01126 m .11974 .0101 L .12297 .01138 L closepath p .955 .785 .905 r
F P
0 g
s
.48642 .01987 m .48416 .01755 L .4819 .01013 L closepath p .955 .785 .905 r
F P
0 g
s
.4819 .01013 m .48569 .01705 L .48642 .01987 L closepath p .955 .785 .905 r
F P
0 g
s
.12853 .07042 m .1255 .06965 L .11724 .04791 L closepath p 1 .702 .358 r
F P
0 g
s
.11412 .04694 m .11724 .04791 L .1255 .06965 L closepath p 1 .702 .358 r
F P
0 g
s
.0571 .01142 m .06454 .01033 L .06266 .01114 L closepath p .955 .785 .905 r
F P
0 g
s
.06454 .01033 m .034 .01112 L .03279 .01112 L closepath p .955 .785 .905 r
F P
0 g
s
.06454 .01033 m .03778 .01112 L .034 .01112 L closepath p .955 .785 .905 r
F P
0 g
s
.05116 .01072 m .06454 .01033 L .0571 .01142 L closepath p .955 .785 .905 r
F P
0 g
s
.06082 .01087 m .0571 .01142 L .06454 .01033 L closepath p .955 .785 .905 r
F P
0 g
s
.06082 .01087 m .06454 .01033 L .0571 .01142 L closepath p .955 .785 .905 r
F P
0 g
s
.60084 .01049 m .60705 .01181 L .60392 .01117 L closepath p .955 .785 .905 r
F P
0 g
s
.8631 .01049 m .86414 .01169 L .86446 .02564 L closepath p .955 .785 .905 r
F P
0 g
s
.8631 .01049 m .86446 .02564 L .85403 .01468 L closepath p .955 .785 .905 r
F P
0 g
s
.85403 .01468 m .85018 .0125 L .8631 .01049 L closepath p .955 .785 .905 r
F P
0 g
s
.27235 .01273 m .28991 .01064 L .28025 .01225 L closepath p .955 .785 .905 r
F P
0 g
s
.28509 .01111 m .28025 .01225 L .28991 .01064 L closepath p .955 .785 .905 r
F P
0 g
s
.28509 .01111 m .28991 .01064 L .28025 .01225 L closepath p .955 .785 .905 r
F P
0 g
s
.62491 .01096 m .62201 .01096 L .62763 .01065 L closepath p .955 .785 .905 r
F P
0 g
s
.04395 .01462 m .04138 .01199 L .05116 .01072 L closepath p .955 .785 .905 r
F P
0 g
s
.03778 .01112 m .05116 .01072 L .04138 .01199 L closepath p .955 .785 .905 r
F P
0 g
s
.05258 .01278 m .05096 .01685 L .05116 .01072 L closepath p .955 .785 .905 r
F P
0 g
s
.04395 .01462 m .05116 .01072 L .05096 .01685 L closepath p .955 .785 .905 r
F P
0 g
s
.05258 .01278 m .05116 .01072 L .0571 .01142 L closepath p .955 .785 .905 r
F P
0 g
s
.51557 .05834 m .51422 .05809 L .51256 .04854 L closepath p 1 .67 .188 r
F P
0 g
s
.5108 .04772 m .51256 .04854 L .51422 .05809 L closepath p 1 .687 .228 r
F P
0 g
s
.89841 .04651 m .89501 .04855 L .89898 .04534 L closepath p .155 .155 .324 r
F P
0 g
s
.89501 .04855 m .89841 .04651 L .89561 .05033 L closepath p .413 .447 .883 r
F P
0 g
s
.1578 .0115 m .16346 .0108 L .16119 .01133 L closepath p .955 .785 .905 r
F P
0 g
s
.16804 .01266 m .16346 .0108 L .17687 .01193 L closepath p .955 .785 .905 r
F P
0 g
s
.16063 .01115 m .1578 .0115 L .16346 .0108 L closepath p .955 .785 .905 r
F P
0 g
s
.69414 .01088 m .69865 .01111 L .69555 .01102 L closepath p .955 .785 .905 r
F P
0 g
s
.50436 .01216 m .50303 .01577 L .50835 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.50436 .01216 m .50835 .01096 L .50303 .01577 L closepath p .955 .785 .905 r
F P
0 g
s
.49454 .01171 m .49136 .01096 L .49438 .01146 L closepath p .955 .785 .905 r
F P
0 g
s
.61689 .01187 m .60705 .01181 L .62201 .01096 L closepath p .955 .785 .905 r
F P
0 g
s
.10817 .04878 m .1083 .04485 L .11669 .067 L closepath p .141 .209 .764 r
F P
0 g
s
.21984 .01192 m .21594 .01103 L .22151 .01215 L closepath p .955 .785 .905 r
F P
0 g
s
.65952 .01193 m .65558 .01104 L .66135 .01216 L closepath p .955 .785 .905 r
F P
0 g
s
.17914 .01119 m .17751 .01162 L .18102 .01105 L closepath p .955 .785 .905 r
F P
0 g
s
.17914 .01119 m .18102 .01105 L .17751 .01162 L closepath p .955 .785 .905 r
F P
0 g
s
.37699 .01254 m .3712 .01111 L .39622 .01168 L closepath p .955 .785 .905 r
F P
0 g
s
.39864 .01111 m .39622 .01168 L .3712 .01111 L closepath p .955 .785 .905 r
F P
0 g
s
.32328 .01251 m .31704 .01111 L .33816 .01193 L closepath p .955 .785 .905 r
F P
0 g
s
.47468 .01322 m .47039 .01112 L .48416 .01755 L closepath p .955 .785 .905 r
F P
0 g
s
.44365 .01112 m .43984 .01219 L .45702 .01112 L closepath p .955 .785 .905 r
F P
0 g
s
.44365 .01112 m .45702 .01112 L .43984 .01219 L closepath p .955 .785 .905 r
F P
0 g
s
.59344 .01127 m .59031 .0129 L .57513 .01216 L closepath p .955 .785 .905 r
F P
0 g
s
.81001 .01593 m .81441 .01137 L .82054 .01833 L closepath p .955 .785 .905 r
F P
0 g
s
.82226 .01626 m .82054 .01833 L .81441 .01137 L closepath p .955 .785 .905 r
F P
0 g
s
.87713 .01371 m .87407 .01142 L .88333 .01445 L closepath p .955 .785 .905 r
F P
0 g
s
.88495 .01187 m .88333 .01445 L .87407 .01142 L closepath p .955 .785 .905 r
F P
0 g
s
.82003 .05812 m .81856 .05669 L .81786 .0492 L closepath p .993 .687 .188 r
F P
0 g
s
.81631 .04744 m .81786 .0492 L .81856 .05669 L closepath p .993 .687 .189 r
F P
0 g
s
.49438 .01146 m .50303 .01577 L .49454 .01171 L closepath p .955 .785 .905 r
F P
0 g
s
.46426 .04866 m .4625 .04925 L .4656 .0451 L closepath p .118 .118 .21 r
F P
0 g
s
.15427 .01187 m .15173 .01299 L .1578 .0115 L closepath p .955 .785 .905 r
F P
0 g
s
.15427 .01187 m .1578 .0115 L .15173 .01299 L closepath p .955 .785 .905 r
F P
0 g
s
.42474 .01327 m .42169 .0115 L .43984 .01219 L closepath p .955 .785 .905 r
F P
0 g
s
.53654 .01154 m .53635 .01371 L .53589 .01593 L closepath p .954 .784 .905 r
F P
0 g
s
.22292 .07944 m .21986 .07707 L .2293 .04937 L closepath p 1 1 .814 r
F P
0 g
s
.22627 .0469 m .2293 .04937 L .21986 .07707 L closepath p 1 1 .814 r
F P
0 g
s
.66352 .07944 m .66157 .07707 L .66896 .04938 L closepath p 1 1 .697 r
F P
0 g
s
.66704 .0469 m .66896 .04938 L .66157 .07707 L closepath p 1 1 .697 r
F P
0 g
s
.93217 .07944 m .9309 .07707 L .93712 .04938 L closepath p 1 .982 .577 r
F P
0 g
s
.93587 .0469 m .93712 .04938 L .9309 .07707 L closepath p 1 .982 .577 r
F P
0 g
s
.39622 .01168 m .39381 .01226 L .37699 .01254 L closepath p .955 .785 .905 r
F P
0 g
s
.86808 .03001 m .86446 .02564 L .86414 .01169 L closepath p .955 .785 .905 r
F P
0 g
s
.86414 .01169 m .86518 .0129 L .86808 .03001 L closepath p .955 .785 .905 r
F P
0 g
s
.49683 .01399 m .49454 .01171 L .50303 .01577 L closepath p .955 .785 .905 r
F P
0 g
s
.13067 .01175 m .14201 .01234 L .13338 .01274 L closepath p .955 .785 .905 r
F P
0 g
s
.54491 .01554 m .54584 .0174 L .54305 .01181 L closepath p .955 .785 .905 r
F P
0 g
s
.54491 .01554 m .54305 .01181 L .54584 .0174 L closepath p .955 .785 .905 r
F P
0 g
s
.61689 .01187 m .61604 .01316 L .60705 .01181 L closepath p .955 .785 .905 r
F P
0 g
s
.60919 .01373 m .60705 .01181 L .61604 .01316 L closepath p .955 .785 .905 r
F P
0 g
s
.2211 .0146 m .21984 .01192 L .23385 .01716 L closepath p .955 .785 .905 r
F P
0 g
s
.22151 .01215 m .23385 .01716 L .21984 .01192 L closepath p .955 .785 .905 r
F P
0 g
s
.33816 .01193 m .32607 .01671 L .32328 .01251 L closepath p .955 .785 .905 r
F P
0 g
s
.66086 .01461 m .65952 .01193 L .67377 .01717 L closepath p .955 .785 .905 r
F P
0 g
s
.66135 .01216 m .67377 .01717 L .65952 .01193 L closepath p .955 .785 .905 r
F P
0 g
s
.16804 .01266 m .17687 .01193 L .17634 .01231 L closepath p .955 .785 .905 r
F P
0 g
s
.08934 .01208 m .08903 .01462 L .08842 .01717 L closepath p .954 .784 .905 r
F P
0 g
s
.92899 .01461 m .94124 .01208 L .94196 .01717 L closepath p .955 .785 .905 r
F P
0 g
s
.94477 .01251 m .94196 .01717 L .94124 .01208 L closepath p .955 .785 .905 r
F P
0 g
s
.57658 .01429 m .57513 .01216 L .59031 .0129 L closepath p .955 .785 .905 r
F P
0 g
s
.39639 .04824 m .39902 .04995 L .39776 .05789 L closepath p 1 .884 .525 r
F P
0 g
s
.42634 .01671 m .42572 .01478 L .43984 .01219 L closepath p .955 .785 .905 r
F P
0 g
s
.42474 .01327 m .43984 .01219 L .42572 .01478 L closepath p .955 .785 .905 r
F P
0 g
s
.43984 .01219 m .43968 .01259 L .42634 .01671 L closepath p .955 .785 .905 r
F P
0 g
s
.43097 .04762 m .42878 .05 L .42972 .03899 L closepath p .172 .185 .699 r
F P
0 g
s
.42878 .05 m .43097 .04762 L .43023 .06001 L closepath p .172 .184 .698 r
F P
0 g
s
.3403 .01225 m .3399 .01299 L .32607 .01671 L closepath p .955 .785 .905 r
F P
0 g
s
.26607 .0167 m .26546 .01474 L .28025 .01225 L closepath p .955 .785 .905 r
F P
0 g
s
.26446 .0132 m .28025 .01225 L .26546 .01474 L closepath p .955 .785 .905 r
F P
0 g
s
.28025 .01225 m .27942 .01374 L .26607 .0167 L closepath p .955 .785 .905 r
F P
0 g
s
.3797 .01669 m .37699 .01254 L .39381 .01226 L closepath p .955 .785 .905 r
F P
0 g
s
.39381 .01226 m .39299 .01374 L .3797 .01669 L closepath p .955 .785 .905 r
F P
0 g
s
.72095 .01226 m .72056 .013 L .70678 .0167 L closepath p .955 .785 .905 r
F P
0 g
s
.77073 .0167 m .76775 .01251 L .78495 .01226 L closepath p .955 .785 .905 r
F P
0 g
s
.78495 .01226 m .78457 .013 L .77073 .0167 L closepath p .955 .785 .905 r
F P
0 g
s
.17634 .01231 m .17562 .01325 L .16804 .01266 L closepath p .955 .785 .905 r
F P
0 g
s
.1351 .0153 m .13338 .01274 L .14201 .01234 L closepath p .955 .785 .905 r
F P
0 g
s
.89897 .04793 m .89561 .05033 L .89841 .04651 L closepath p .436 .478 .907 r
F P
0 g
s
.89561 .05033 m .89897 .04793 L .89733 .05411 L closepath p .451 .505 .929 r
F P
0 g
s
.43968 .01259 m .43952 .01299 L .42663 .01839 L closepath p .955 .785 .905 r
F P
0 g
s
.42663 .01839 m .42634 .01671 L .43968 .01259 L closepath p .955 .785 .905 r
F P
0 g
s
.54272 .01259 m .54677 .01927 L .52738 .06356 L closepath p .955 .785 .905 r
F P
0 g
s
.52738 .06356 m .52984 .05072 L .54272 .01259 L closepath p .955 .785 .905 r
F P
0 g
s
.17365 .01544 m .17266 .01654 L .16804 .01266 L closepath p .955 .785 .905 r
F P
0 g
s
.17167 .01763 m .16804 .01266 L .17266 .01654 L closepath p .955 .785 .905 r
F P
0 g
s
.17365 .01544 m .16804 .01266 L .17562 .01325 L closepath p .955 .785 .905 r
F P
0 g
s
.33062 .04807 m .32868 .05045 L .32943 .03953 L closepath p .162 .162 .662 r
F P
0 g
s
.09821 .0569 m .09476 .05496 L .09945 .05053 L closepath p 1 1 .823 r
F P
0 g
s
.09635 .04775 m .09945 .05053 L .09476 .05496 L closepath p 1 1 .826 r
F P
0 g
s
.46046 .0507 m .4597 .05064 L .46426 .04866 L closepath p .038 .038 .038 r
F P
0 g
s
.59031 .0129 m .5898 .01457 L .57762 .01826 L closepath p .955 .785 .905 r
F P
0 g
s
.59031 .0129 m .57762 .01826 L .57658 .01429 L closepath p .955 .785 .905 r
F P
0 g
s
.86518 .0129 m .86838 .02232 L .86977 .03147 L closepath p .955 .785 .905 r
F P
0 g
s
.86977 .03147 m .86808 .03001 L .86518 .0129 L closepath p .955 .785 .905 r
F P
0 g
s
.77641 .04837 m .77335 .05076 L .77496 .03993 L closepath p .328 .332 .792 r
F P
0 g
s
.77335 .05076 m .77641 .04837 L .77499 .06044 L closepath p .327 .331 .791 r
F P
0 g
s
.43952 .01299 m .43936 .01339 L .42711 .02146 L closepath p .955 .785 .905 r
F P
0 g
s
.42711 .02146 m .42663 .01839 L .43952 .01299 L closepath p .955 .785 .905 r
F P
0 g
s
.3399 .01299 m .33948 .01373 L .3264 .01863 L closepath p .955 .785 .905 r
F P
0 g
s
.3264 .01863 m .32607 .01671 L .3399 .01299 L closepath p .955 .785 .905 r
F P
0 g
s
.72056 .013 m .72018 .01374 L .70761 .02074 L closepath p .955 .785 .905 r
F P
0 g
s
.70761 .02074 m .70678 .0167 L .72056 .013 L closepath p .955 .785 .905 r
F P
0 g
s
.78457 .013 m .78419 .01374 L .77118 .01883 L closepath p .955 .785 .905 r
F P
0 g
s
.77118 .01883 m .77073 .0167 L .78457 .013 L closepath p .955 .785 .905 r
F P
0 g
s
.61604 .01316 m .6152 .01445 L .60919 .01373 L closepath p .955 .785 .905 r
F P
0 g
s
.10104 .02081 m .09655 .01317 L .09762 .01412 L closepath p .955 .785 .905 r
F P
0 g
s
.07704 .0716 m .07582 .07306 L .0803 .05697 L closepath p .077 .077 .133 r
F P
0 g
s
.09655 .01317 m .10104 .02081 L .07704 .0716 L closepath p .955 .785 .905 r
F P
0 g
s
.07704 .0716 m .0803 .05697 L .09655 .01317 L closepath p .955 .785 .905 r
F P
0 g
s
.59599 .05808 m .59386 .05634 L .59494 .05095 L closepath p 1 .845 .432 r
F P
0 g
s
.59281 .04922 m .59494 .05095 L .59386 .05634 L closepath p 1 .845 .432 r
F P
0 g
s
.48416 .01755 m .4885 .02797 L .47468 .01322 L closepath p .955 .785 .905 r
F P
0 g
s
.48061 .01954 m .47468 .01322 L .4885 .02797 L closepath p .955 .785 .905 r
F P
0 g
s
.48061 .01954 m .4885 .02797 L .47468 .01322 L closepath p .955 .785 .905 r
F P
0 g
s
.17562 .01325 m .17537 .01445 L .17213 .01723 L closepath p .955 .785 .905 r
F P
0 g
s
.1719 .01743 m .17562 .01325 L .17213 .01723 L closepath p .955 .785 .905 r
F P
0 g
s
.17167 .01763 m .17562 .01325 L .1719 .01743 L closepath p .955 .785 .905 r
F P
0 g
s
.45359 .05139 m .45371 .05111 L .46046 .0507 L closepath p .024 .024 .024 r
F P
0 g
s
.45371 .05111 m .45359 .05139 L .44571 .05111 L closepath p .098 .026 .026 r
F P
0 g
s
.43936 .01339 m .4392 .01379 L .42779 .02591 L closepath p .955 .785 .905 r
F P
0 g
s
.42779 .02591 m .42711 .02146 L .43936 .01339 L closepath p .955 .785 .905 r
F P
0 g
s
.53635 .01371 m .53591 .01722 L .53516 .02104 L closepath p .953 .783 .905 r
F P
0 g
s
.53516 .02104 m .53589 .01593 L .53635 .01371 L closepath p .954 .784 .905 r
F P
0 g
s
.8778 .01469 m .87713 .01371 L .88406 .01799 L closepath p .955 .785 .905 r
F P
0 g
s
.88333 .01445 m .88406 .01799 L .87713 .01371 L closepath p .955 .785 .905 r
F P
0 g
s
.6116 .01763 m .60919 .01373 L .6152 .01445 L closepath p .955 .785 .905 r
F P
0 g
s
.33948 .01373 m .33907 .01448 L .3269 .02187 L closepath p .955 .785 .905 r
F P
0 g
s
.3269 .02187 m .3264 .01863 L .33948 .01373 L closepath p .955 .785 .905 r
F P
0 g
s
.27942 .01374 m .2786 .01522 L .26871 .03258 L closepath p .955 .785 .905 r
F P
0 g
s
.26871 .03258 m .26607 .0167 L .27942 .01374 L closepath p .955 .785 .905 r
F P
0 g
s
.72018 .01374 m .71979 .01448 L .70863 .02615 L closepath p .955 .785 .905 r
F P
0 g
s
.70863 .02615 m .70761 .02074 L .72018 .01374 L closepath p .955 .785 .905 r
F P
0 g
s
.78419 .01374 m .78381 .01448 L .77183 .02221 L closepath p .955 .785 .905 r
F P
0 g
s
.77183 .02221 m .77118 .01883 L .78419 .01374 L closepath p .955 .785 .905 r
F P
0 g
s
.39299 .01374 m .39218 .01523 L .38028 .02008 L closepath p .955 .785 .905 r
F P
0 g
s
.38028 .02008 m .3797 .01669 L .39299 .01374 L closepath p .955 .785 .905 r
F P
0 g
s
.4392 .01379 m .43904 .01419 L .42866 .03175 L closepath p .955 .785 .905 r
F P
0 g
s
.42866 .03175 m .42779 .02591 L .4392 .01379 L closepath p .955 .785 .905 r
F P
0 g
s
.38482 .05187 m .38238 .0516 L .38284 .03766 L closepath p .167 .167 .68 r
F P
0 g
s
.49683 .01399 m .50303 .01577 L .49772 .01604 L closepath p .955 .785 .905 r
F P
0 g
s
.43904 .01419 m .43887 .01459 L .42972 .03899 L closepath p .955 .785 .905 r
F P
0 g
s
.43904 .01419 m .42972 .03899 L .43887 .01459 L closepath p .955 .785 .905 r
F P
0 g
s
.42972 .03899 m .42866 .03175 L .43904 .01419 L closepath p .955 .785 .905 r
F P
0 g
s
.57715 .01604 m .57658 .01429 L .57762 .01826 L closepath p .955 .785 .905 r
F P
0 g
s
.57715 .01604 m .57762 .01826 L .57658 .01429 L closepath p .955 .785 .905 r
F P
0 g
s
.17537 .01445 m .17591 .0181 L .17167 .01763 L closepath p .955 .785 .905 r
F P
0 g
s
.17444 .01524 m .17213 .01723 L .17537 .01445 L closepath p .955 .785 .905 r
F P
0 g
s
.17444 .01524 m .17537 .01445 L .17213 .01723 L closepath p .955 .785 .905 r
F P
0 g
s
.6152 .01445 m .61587 .01793 L .6116 .01763 L closepath p .955 .785 .905 r
F P
0 g
s
.33907 .01448 m .33866 .01522 L .32758 .02643 L closepath p .955 .785 .905 r
F P
0 g
s
.32758 .02643 m .3269 .02187 L .33907 .01448 L closepath p .955 .785 .905 r
F P
0 g
s
.71979 .01448 m .71941 .01523 L .70984 .03292 L closepath p .955 .785 .905 r
F P
0 g
s
.70984 .03292 m .70863 .02615 L .71979 .01448 L closepath p .955 .785 .905 r
F P
0 g
s
.78381 .01448 m .78342 .01523 L .77267 .02685 L closepath p .955 .785 .905 r
F P
0 g
s
.77267 .02685 m .77183 .02221 L .78381 .01448 L closepath p .955 .785 .905 r
F P
0 g
s
.52984 .05072 m .52742 .05234 L .53214 .03843 L closepath p .153 .153 .455 r
F P
0 g
s
.58003 .03496 m .57762 .01826 L .5898 .01457 L closepath p .955 .785 .905 r
F P
0 g
s
.5898 .01457 m .58928 .01624 L .58003 .03496 L closepath p .955 .785 .905 r
F P
0 g
s
.43887 .01459 m .43871 .015 L .43097 .04762 L closepath p .955 .785 .905 r
F P
0 g
s
.43887 .01459 m .43097 .04762 L .43871 .015 L closepath p .955 .785 .905 r
F P
0 g
s
.43097 .04762 m .42972 .03899 L .43887 .01459 L closepath p .955 .785 .905 r
F P
0 g
s
.2211 .0146 m .23385 .01716 L .22048 .02058 L closepath p .955 .785 .905 r
F P
0 g
s
.66086 .01461 m .67377 .01717 L .66044 .02059 L closepath p .955 .785 .905 r
F P
0 g
s
.92899 .01461 m .94196 .01717 L .92869 .02059 L closepath p .955 .785 .905 r
F P
0 g
s
.04395 .01462 m .05096 .01685 L .04439 .01559 L closepath p .955 .785 .905 r
F P
0 g
s
.08903 .01462 m .08841 .01866 L .08735 .02318 L closepath p .954 .784 .905 r
F P
0 g
s
.08735 .02318 m .08842 .01717 L .08903 .01462 L closepath p .954 .784 .905 r
F P
0 g
s
.8591 .01932 m .85403 .01468 L .86446 .02564 L closepath p .955 .785 .905 r
F P
0 g
s
.8591 .01932 m .86446 .02564 L .85403 .01468 L closepath p .955 .785 .905 r
F P
0 g
s
.87847 .01567 m .8778 .01469 L .88609 .02309 L closepath p .955 .785 .905 r
F P
0 g
s
.88406 .01799 m .88609 .02309 L .8778 .01469 L closepath p .955 .785 .905 r
F P
0 g
s
.13902 .05033 m .13755 .0526 L .13813 .03965 L closepath p .132 .132 .529 r
F P
0 g
s
.43871 .015 m .43855 .0154 L .43242 .05766 L closepath p .955 .785 .905 r
F P
0 g
s
.43871 .015 m .43242 .05766 L .43855 .0154 L closepath p .955 .785 .905 r
F P
0 g
s
.43242 .05766 m .43097 .04762 L .43871 .015 L closepath p .955 .785 .905 r
F P
0 g
s
.71283 .0506 m .70993 .05298 L .71124 .04107 L closepath p .299 .309 .779 r
F P
0 g
s
.70993 .05298 m .71283 .0506 L .7117 .06385 L closepath p .298 .307 .777 r
F P
0 g
s
.8044 .05061 m .80129 .05298 L .80329 .04027 L closepath p .312 .297 .759 r
F P
0 g
s
.80129 .05298 m .8044 .05061 L .80428 .06348 L closepath p .374 .404 .853 r
F P
0 g
s
.32842 .03232 m .32758 .02643 L .33866 .01522 L closepath p .955 .785 .905 r
F P
0 g
s
.33609 .04214 m .32943 .03953 L .33866 .01522 L closepath p .955 .785 .905 r
F P
0 g
s
.32842 .03232 m .33866 .01522 L .32943 .03953 L closepath p .955 .785 .905 r
F P
0 g
s
.33901 .02353 m .33351 .06916 L .33866 .01522 L closepath p .955 .785 .905 r
F P
0 g
s
.2713 .05197 m .26871 .03258 L .2786 .01522 L closepath p .955 .785 .905 r
F P
0 g
s
.27459 .07859 m .2735 .06971 L .2786 .01522 L closepath p .955 .785 .905 r
F P
0 g
s
.2713 .05197 m .2786 .01522 L .2735 .06971 L closepath p .955 .785 .905 r
F P
0 g
s
.27608 .09212 m .27566 .08749 L .2786 .01522 L closepath p .955 .785 .905 r
F P
0 g
s
.27459 .07859 m .2786 .01522 L .27566 .08749 L closepath p .955 .785 .905 r
F P
0 g
s
.27608 .09212 m .2786 .01522 L .27774 .08151 L closepath p .955 .785 .905 r
F P
0 g
s
.27873 .01773 m .27774 .08151 L .2786 .01522 L closepath p .955 .785 .905 r
F P
0 g
s
.71124 .04107 m .70984 .03292 L .71941 .01523 L closepath p .955 .785 .905 r
F P
0 g
s
.71908 .05127 m .71283 .0506 L .71941 .01523 L closepath p .955 .785 .905 r
F P
0 g
s
.71124 .04107 m .71941 .01523 L .71283 .0506 L closepath p .955 .785 .905 r
F P
0 g
s
.71941 .01523 m .71962 .01756 L .71875 .08749 L closepath p .955 .785 .905 r
F P
0 g
s
.77372 .03276 m .77267 .02685 L .78342 .01523 L closepath p .955 .785 .905 r
F P
0 g
s
.78342 .01523 m .78365 .01814 L .77372 .03276 L closepath p .955 .785 .905 r
F P
0 g
s
.38133 .02707 m .38028 .02008 L .39218 .01523 L closepath p .955 .785 .905 r
F P
0 g
s
.39218 .01523 m .39234 .01774 L .38133 .02707 L closepath p .955 .785 .905 r
F P
0 g
s
.17259 .01684 m .17444 .01524 L .17352 .01604 L closepath p .955 .785 .905 r
F P
0 g
s
.17213 .01723 m .17444 .01524 L .17259 .01684 L closepath p .955 .785 .905 r
F P
0 g
s
.17213 .01723 m .17259 .01684 L .17444 .01524 L closepath p .955 .785 .905 r
F P
0 g
s
.43407 .0691 m .43242 .05766 L .43855 .0154 L closepath p .955 .785 .905 r
F P
0 g
s
.43756 .05169 m .43523 .07743 L .43855 .0154 L closepath p .955 .785 .905 r
F P
0 g
s
.43407 .0691 m .43855 .0154 L .43523 .07743 L closepath p .955 .785 .905 r
F P
0 g
s
.43855 .0154 m .4387 .01781 L .43656 .08817 L closepath p .955 .785 .905 r
F P
0 g
s
.05096 .01685 m .04494 .01708 L .04439 .01559 L closepath p .955 .785 .905 r
F P
0 g
s
.04465 .01626 m .04439 .01559 L .04494 .01708 L closepath p .955 .785 .905 r
F P
0 g
s
.04465 .01626 m .04494 .01708 L .04439 .01559 L closepath p .955 .785 .905 r
F P
0 g
s
.87913 .01665 m .87847 .01567 L .88942 .02977 L closepath p .955 .785 .905 r
F P
0 g
s
.88609 .02309 m .88942 .02977 L .87847 .01567 L closepath p .955 .785 .905 r
F P
0 g
s
.47635 .07292 m .47391 .07128 L .47341 .05354 L closepath p 1 .866 .486 r
F P
0 g
s
.47096 .05174 m .47341 .05354 L .47391 .07128 L closepath p 1 .866 .486 r
F P
0 g
s
.49909 .02004 m .49772 .01604 L .50303 .01577 L closepath p .955 .785 .905 r
F P
0 g
s
.50303 .01577 m .50334 .01952 L .49909 .02004 L closepath p .955 .785 .905 r
F P
0 g
s
.81001 .01593 m .8121 .032 L .80481 .02663 L closepath p .955 .785 .905 r
F P
0 g
s
.81001 .01593 m .82054 .01833 L .81316 .03166 L closepath p .955 .785 .905 r
F P
0 g
s
.81631 .04744 m .81839 .08044 L .81001 .01593 L closepath p .955 .785 .905 r
F P
0 g
s
.58928 .01624 m .58983 .02669 L .58629 .09252 L closepath p .955 .785 .905 r
F P
0 g
s
.58242 .05525 m .58928 .01624 L .58629 .09252 L closepath p .955 .785 .905 r
F P
0 g
s
.58242 .05525 m .58003 .03496 L .58928 .01624 L closepath p .955 .785 .905 r
F P
0 g
s
.90061 .05142 m .89733 .05411 L .89897 .04793 L closepath p .461 .516 .936 r
F P
0 g
s
.89733 .05411 m .90061 .05142 L .90013 .05988 L closepath p .465 .523 .942 r
F P
0 g
s
.2713 .05197 m .26951 .05431 L .26871 .03258 L closepath p .155 .155 .636 r
F P
0 g
s
.8798 .01763 m .87913 .01665 L .89406 .03801 L closepath p .955 .785 .905 r
F P
0 g
s
.88942 .02977 m .89406 .03801 L .87913 .01665 L closepath p .955 .785 .905 r
F P
0 g
s
.05096 .01685 m .05139 .02196 L .04638 .02164 L closepath p .955 .785 .905 r
F P
0 g
s
.05096 .01685 m .04638 .02164 L .04494 .01708 L closepath p .955 .785 .905 r
F P
0 g
s
.48919 .02398 m .48642 .01987 L .48569 .01705 L closepath p .955 .785 .905 r
F P
0 g
s
.0456 .01909 m .04494 .01708 L .04638 .02164 L closepath p .955 .785 .905 r
F P
0 g
s
.0456 .01909 m .04638 .02164 L .04494 .01708 L closepath p .955 .785 .905 r
F P
0 g
s
.21978 .02404 m .22048 .02058 L .23385 .01716 L closepath p .955 .785 .905 r
F P
0 g
s
.20614 .08334 m .21882 .02781 L .23385 .01716 L closepath p .955 .785 .905 r
F P
0 g
s
.21978 .02404 m .23385 .01716 L .21882 .02781 L closepath p .955 .785 .905 r
F P
0 g
s
.20614 .08334 m .23385 .01716 L .23007 .03073 L closepath p .955 .785 .905 r
F P
0 g
s
.23196 .02362 m .23007 .03073 L .23385 .01716 L closepath p .955 .785 .905 r
F P
0 g
s
.23196 .02362 m .23385 .01716 L .23007 .03073 L closepath p .955 .785 .905 r
F P
0 g
s
.65907 .02782 m .66044 .02059 L .67377 .01717 L closepath p .955 .785 .905 r
F P
0 g
s
.67377 .01717 m .67208 .02366 L .65907 .02782 L closepath p .955 .785 .905 r
F P
0 g
s
.92746 .02782 m .92869 .02059 L .94196 .01717 L closepath p .955 .785 .905 r
F P
0 g
s
.94196 .01717 m .94042 .02369 L .92746 .02782 L closepath p .955 .785 .905 r
F P
0 g
s
.53591 .01722 m .53522 .02207 L .53391 .02854 L closepath p .955 .785 .905 r
F P
0 g
s
.53391 .02854 m .53516 .02104 L .53591 .01722 L closepath p .952 .783 .905 r
F P
0 g
s
.53391 .02854 m .53591 .01722 L .53516 .02104 L closepath p .952 .783 .905 r
F P
0 g
s
.14869 .05419 m .14961 .01748 L .14983 .02941 L closepath p .955 .785 .905 r
F P
0 g
s
.14463 .10148 m .14961 .01748 L .14777 .09108 L closepath p .955 .785 .905 r
F P
0 g
s
.14385 .10408 m .14961 .01748 L .14463 .10148 L closepath p .955 .785 .905 r
F P
0 g
s
.14385 .10408 m .14346 .10469 L .14961 .01748 L closepath p .955 .785 .905 r
F P
0 g
s
.48416 .01755 m .48642 .01987 L .4885 .02797 L closepath p .955 .785 .905 r
F P
0 g
s
.71962 .01756 m .72022 .022 L .71885 .08865 L closepath p .955 .785 .905 r
F P
0 g
s
.71885 .08865 m .71875 .08749 L .71962 .01756 L closepath p .955 .785 .905 r
F P
0 g
s
.17167 .01763 m .18527 .04705 L .1767 .02868 L closepath p .955 .785 .905 r
F P
0 g
s
.17846 .03233 m .17167 .01763 L .17591 .0181 L closepath p .955 .785 .905 r
F P
0 g
s
.6116 .01763 m .62612 .04706 L .61698 .02868 L closepath p .955 .785 .905 r
F P
0 g
s
.61885 .03233 m .6116 .01763 L .61587 .01793 L closepath p .955 .785 .905 r
F P
0 g
s
.88534 .02868 m .8798 .01763 L .8942 .03825 L closepath p .955 .785 .905 r
F P
0 g
s
.89406 .03801 m .8942 .03825 L .8798 .01763 L closepath p .955 .785 .905 r
F P
0 g
s
.27774 .08151 m .27873 .01773 L .27941 .07091 L closepath p .955 .785 .905 r
F P
0 g
s
.27919 .02223 m .27941 .07091 L .27873 .01773 L closepath p .955 .785 .905 r
F P
0 g
s
.39234 .01774 m .39283 .02225 L .38208 .03236 L closepath p .955 .785 .905 r
F P
0 g
s
.38208 .03236 m .38133 .02707 L .39234 .01774 L closepath p .955 .785 .905 r
F P
0 g
s
.4387 .01781 m .44109 .03458 L .43672 .09057 L closepath p .955 .785 .905 r
F P
0 g
s
.43672 .09057 m .43656 .08817 L .4387 .01781 L closepath p .955 .785 .905 r
F P
0 g
s
.0684 .07568 m .06526 .07476 L .06265 .05566 L closepath p 1 .832 .512 r
F P
0 g
s
.05922 .05431 m .06265 .05566 L .06526 .07476 L closepath p 1 .834 .527 r
F P
0 g
s
.61587 .01793 m .61783 .02302 L .61885 .03233 L closepath p .955 .785 .905 r
F P
0 g
s
.17591 .0181 m .17768 .02325 L .17846 .03233 L closepath p .955 .785 .905 r
F P
0 g
s
.78365 .01814 m .78424 .02278 L .77434 .03634 L closepath p .955 .785 .905 r
F P
0 g
s
.77434 .03634 m .77372 .03276 L .78365 .01814 L closepath p .955 .785 .905 r
F P
0 g
s
.81657 .02745 m .81316 .03166 L .82054 .01833 L closepath p .955 .785 .905 r
F P
0 g
s
.08841 .01866 m .08746 .02418 L .08565 .03181 L closepath p .955 .785 .905 r
F P
0 g
s
.08565 .03181 m .08735 .02318 L .08841 .01866 L closepath p .953 .783 .905 r
F P
0 g
s
.08565 .03181 m .08841 .01866 L .08735 .02318 L closepath p .953 .783 .905 r
F P
0 g
s
.54109 .07117 m .53882 .06881 L .54373 .05656 L closepath p 1 1 .732 r
F P
0 g
s
.54148 .05416 m .54373 .05656 L .53882 .06881 L closepath p 1 1 .732 r
F P
0 g
s
.09697 .06327 m .09359 .06089 L .09821 .0569 L closepath p 1 1 .823 r
F P
0 g
s
.09476 .05496 m .09821 .0569 L .09359 .06089 L closepath p 1 1 .824 r
F P
0 g
s
.5254 .07321 m .52332 .07428 L .52738 .06356 L closepath p .14 .14 .393 r
F P
0 g
s
.5254 .07321 m .52738 .06356 L .54677 .01927 L closepath p .955 .785 .905 r
F P
0 g
s
.54677 .01927 m .52393 .07965 L .5254 .07321 L closepath p .955 .785 .905 r
F P
0 g
s
.54677 .01927 m .54696 .02673 L .52393 .07965 L closepath p .955 .785 .905 r
F P
0 g
s
.54677 .01927 m .55896 .04209 L .54696 .02673 L closepath p .955 .785 .905 r
F P
0 g
s
.50334 .01952 m .50428 .02458 L .50059 .02455 L closepath p .955 .785 .905 r
F P
0 g
s
.50059 .02455 m .49909 .02004 L .50334 .01952 L closepath p .955 .785 .905 r
F P
0 g
s
.48061 .01954 m .4885 .02797 L .48596 .02586 L closepath p .955 .785 .905 r
F P
0 g
s
.58242 .05525 m .57984 .05749 L .58003 .03496 L closepath p .217 .227 .72 r
F P
0 g
s
.57984 .05749 m .58242 .05525 L .5817 .07574 L closepath p .209 .215 .709 r
F P
0 g
s
.48943 .0275 m .4885 .02797 L .48642 .01987 L closepath p .955 .785 .905 r
F P
0 g
s
.48642 .01987 m .48919 .02398 L .48943 .0275 L closepath p .955 .785 .905 r
F P
0 g
s
.51112 .05619 m .50881 .05891 L .49909 .02004 L closepath p .282 .355 .838 r
F P
0 g
s
.49909 .02004 m .51112 .05619 L .50507 .03809 L closepath p .955 .785 .905 r
F P
0 g
s
.79392 .07268 m .79231 .07106 L .79139 .05799 L closepath p 1 .766 .292 r
F P
0 g
s
.78976 .05631 m .79139 .05799 L .79231 .07106 L closepath p 1 .766 .292 r
F P
0 g
s
.59664 .06252 m .5945 .06078 L .59599 .05808 L closepath p 1 .846 .433 r
F P
0 g
s
.59386 .05634 m .59599 .05808 L .5945 .06078 L closepath p 1 .846 .433 r
F P
0 g
s
.82355 .06721 m .82218 .0661 L .82003 .05812 L closepath p .917 .554 .077 r
F P
0 g
s
.81856 .05669 m .82003 .05812 L .82218 .0661 L closepath p .915 .553 .076 r
F P
0 g
s
.51929 .06987 m .51827 .0698 L .51557 .05834 L closepath p .949 .611 .101 r
F P
0 g
s
.51422 .05809 m .51557 .05834 L .51827 .0698 L closepath p .993 .643 .16 r
F P
0 g
s
.0803 .05697 m .07903 .05857 L .0833 .04306 L closepath p .08 .08 .151 r
F P
0 g
s
.07444 .08259 m .07704 .0716 L .10104 .02081 L closepath p .955 .785 .905 r
F P
0 g
s
.10104 .02081 m .07252 .08994 L .07444 .08259 L closepath p .955 .785 .905 r
F P
0 g
s
.07181 .09223 m .07252 .08994 L .10104 .02081 L closepath p .955 .785 .905 r
F P
0 g
s
.08682 .05529 m .10104 .02081 L .07252 .08994 L closepath p .955 .785 .905 r
F P
0 g
s
.07152 .09304 m .07181 .09223 L .10104 .02081 L closepath p .955 .785 .905 r
F P
0 g
s
.08647 .05643 m .10104 .02081 L .07181 .09223 L closepath p .955 .785 .905 r
F P
0 g
s
.08647 .05643 m .07181 .09223 L .10104 .02081 L closepath p .955 .785 .905 r
F P
0 g
s
.10104 .02081 m .11412 .04694 L .13775 .09699 L closepath p .955 .785 .905 r
F P
0 g
s
.10104 .02081 m .13775 .09699 L .10078 .02945 L closepath p .955 .785 .905 r
F P
0 g
s
.10104 .02081 m .10078 .02945 L .08632 .05683 L closepath p .955 .785 .905 r
F P
0 g
s
.50881 .05891 m .51112 .05619 L .51303 .07197 L closepath p .277 .349 .833 r
F P
0 g
s
.0587 .06303 m .05752 .06573 L .04638 .02164 L closepath p .147 .147 .698 r
F P
0 g
s
.04638 .02164 m .0587 .06303 L .05251 .04232 L closepath p .955 .785 .905 r
F P
0 g
s
.05253 .0423 m .04638 .02164 L .05139 .02196 L closepath p .955 .785 .905 r
F P
0 g
s
.39776 .05789 m .40039 .05957 L .39895 .0662 L closepath p 1 .883 .524 r
F P
0 g
s
.45106 .08701 m .44856 .08539 L .44712 .05971 L closepath p 1 .875 .504 r
F P
0 g
s
.44423 .05538 m .44712 .05971 L .44856 .08539 L closepath p 1 .875 .504 r
F P
0 g
s
.44712 .05971 m .44423 .05538 L .45241 .05971 L closepath p .841 1 1 r
F P
0 g
s
.4523 .05538 m .45241 .05971 L .44423 .05538 L closepath p .841 1 1 r
F P
0 g
s
.45241 .05971 m .4523 .05538 L .45763 .06005 L closepath p .815 .982 1 r
F P
0 g
s
.05139 .02196 m .05201 .02557 L .05253 .0423 L closepath p .955 .785 .905 r
F P
0 g
s
.72022 .022 m .7212 .02856 L .71895 .08981 L closepath p .955 .785 .905 r
F P
0 g
s
.71895 .08981 m .71885 .08865 L .72022 .022 L closepath p .955 .785 .905 r
F P
0 g
s
.53522 .02207 m .53427 .02828 L .53214 .03843 L closepath p .955 .785 .905 r
F P
0 g
s
.53214 .03843 m .53391 .02854 L .53522 .02207 L closepath p .954 .784 .905 r
F P
0 g
s
.90334 .05699 m .90013 .05988 L .90061 .05142 L closepath p .469 .526 .944 r
F P
0 g
s
.90013 .05988 m .90334 .05699 L .90399 .06759 L closepath p .471 .53 .947 r
F P
0 g
s
.56267 .05663 m .5607 .05995 L .55401 .04017 L closepath p .355 .465 .926 r
F P
0 g
s
.5607 .05995 m .56267 .05663 L .57067 .0792 L closepath p .351 .461 .923 r
F P
0 g
s
.43242 .05766 m .43023 .06001 L .43097 .04762 L closepath p .172 .184 .698 r
F P
0 g
s
.43023 .06001 m .43242 .05766 L .43186 .07142 L closepath p .172 .184 .698 r
F P
0 g
s
.27941 .07091 m .27919 .02223 L .28107 .06033 L closepath p .955 .785 .905 r
F P
0 g
s
.27997 .02873 m .28107 .06033 L .27919 .02223 L closepath p .955 .785 .905 r
F P
0 g
s
.45807 .05583 m .45763 .06005 L .4523 .05538 L closepath p .805 .978 1 r
F P
0 g
s
.45763 .06005 m .45807 .05583 L .46142 .06108 L closepath p .729 .908 1 r
F P
0 g
s
.39283 .02225 m .39366 .02876 L .38284 .03766 L closepath p .955 .785 .905 r
F P
0 g
s
.38284 .03766 m .38208 .03236 L .39283 .02225 L closepath p .955 .785 .905 r
F P
0 g
s
.87024 .03108 m .86977 .03147 L .86838 .02232 L closepath p .955 .785 .905 r
F P
0 g
s
.86838 .02232 m .87054 .02992 L .87024 .03108 L closepath p .955 .785 .905 r
F P
0 g
s
.34836 .07088 m .34544 .06811 L .34688 .06012 L closepath p 1 .894 .549 r
F P
0 g
s
.34376 .05578 m .34688 .06012 L .34544 .06811 L closepath p 1 .894 .549 r
F P
0 g
s
.34688 .06012 m .34376 .05578 L .38367 .06012 L closepath p .841 1 1 r
F P
0 g
s
.38543 .05578 m .38367 .06012 L .34376 .05578 L closepath p .841 1 1 r
F P
0 g
s
.38367 .06012 m .38543 .05578 L .3848 .06709 L closepath p .169 .179 .699 r
F P
0 g
s
.33198 .05794 m .33003 .06029 L .33062 .04807 L closepath p .162 .162 .661 r
F P
0 g
s
.77805 .05809 m .77499 .06044 L .77641 .04837 L closepath p .327 .33 .791 r
F P
0 g
s
.77499 .06044 m .77805 .05809 L .77683 .0714 L closepath p .326 .33 .79 r
F P
0 g
s
.78424 .02278 m .7852 .02914 L .77496 .03993 L closepath p .955 .785 .905 r
F P
0 g
s
.77496 .03993 m .77434 .03634 L .78424 .02278 L closepath p .955 .785 .905 r
F P
0 g
s
.62612 .04706 m .61885 .03233 L .61783 .02302 L closepath p .955 .785 .905 r
F P
0 g
s
.61783 .02302 m .62108 .02971 L .62612 .04706 L closepath p .955 .785 .905 r
F P
0 g
s
.18527 .04705 m .17846 .03233 L .17768 .02325 L closepath p .955 .785 .905 r
F P
0 g
s
.17768 .02325 m .18068 .02988 L .18527 .04705 L closepath p .955 .785 .905 r
F P
0 g
s
.46255 .0572 m .46142 .06108 L .45807 .05583 L closepath p .707 .895 1 r
F P
0 g
s
.46142 .06108 m .46255 .0572 L .46343 .06323 L closepath p .456 .621 1 r
F P
0 g
s
.90037 .05815 m .89721 .06113 L .8922 .04218 L closepath p .474 .535 .95 r
F P
0 g
s
.89721 .06113 m .90037 .05815 L .90666 .07953 L closepath p .475 .536 .951 r
F P
0 g
s
.19032 .05815 m .18892 .06114 L .18292 .04218 L closepath p .197 .343 .852 r
F P
0 g
s
.18892 .06114 m .19032 .05815 L .19749 .07953 L closepath p .199 .345 .854 r
F P
0 g
s
.63152 .05815 m .62903 .06114 L .62362 .04218 L closepath p .392 .48 .927 r
F P
0 g
s
.62903 .06114 m .63152 .05815 L .63817 .07953 L closepath p .393 .481 .928 r
F P
0 g
s
.34023 .03356 m .33351 .06916 L .33901 .02353 L closepath p .955 .785 .905 r
F P
0 g
s
.64821 .08334 m .65907 .02782 L .67208 .02366 L closepath p .955 .785 .905 r
F P
0 g
s
.64821 .08334 m .67208 .02366 L .66704 .0469 L closepath p .955 .785 .905 r
F P
0 g
s
.67039 .03077 m .66704 .0469 L .67208 .02366 L closepath p .955 .785 .905 r
F P
0 g
s
.67039 .03077 m .67208 .02366 L .66704 .0469 L closepath p .955 .785 .905 r
F P
0 g
s
.94042 .02369 m .9389 .03081 L .90986 .0766 L closepath p .955 .785 .905 r
F P
0 g
s
.92455 .0409 m .92746 .02782 L .94042 .02369 L closepath p .955 .785 .905 r
F P
0 g
s
.90986 .0766 m .92455 .0409 L .94042 .02369 L closepath p .955 .785 .905 r
F P
0 g
s
.49036 .02703 m .48943 .0275 L .48919 .02398 L closepath p .955 .785 .905 r
F P
0 g
s
.08746 .02418 m .08619 .0312 L .0833 .04306 L closepath p .955 .785 .905 r
F P
0 g
s
.0833 .04306 m .08565 .03181 L .08746 .02418 L closepath p .954 .784 .905 r
F P
0 g
s
.13671 .02443 m .147 .02634 L .13736 .03102 L closepath p .955 .785 .905 r
F P
0 g
s
.50209 .02906 m .50059 .02455 L .50428 .02458 L closepath p .955 .785 .905 r
F P
0 g
s
.50428 .02458 m .50583 .03096 L .50209 .02906 L closepath p .955 .785 .905 r
F P
0 g
s
.59688 .06428 m .59475 .06252 L .59664 .06252 L closepath p 1 .85 .439 r
F P
0 g
s
.5945 .06078 m .59664 .06252 L .59475 .06252 L closepath p 1 .85 .439 r
F P
0 g
s
.5802 .07912 m .57856 .07874 L .57142 .06273 L closepath p .961 .544 .106 r
F P
0 g
s
.56956 .06193 m .57142 .06273 L .57856 .07874 L closepath p .961 .544 .106 r
F P
0 g
s
.46574 .05947 m .46343 .06323 L .46255 .0572 L closepath p .621 .761 1 r
F P
0 g
s
.46343 .06323 m .46574 .05947 L .46544 .06538 L closepath p .548 .665 1 r
F P
0 g
s
.09329 .08061 m .08989 .07827 L .09697 .06327 L closepath p 1 1 .832 r
F P
0 g
s
.09359 .06089 m .09697 .06327 L .08989 .07827 L closepath p 1 1 .832 r
F P
0 g
s
.05201 .02557 m .05291 .02989 L .0587 .06303 L closepath p .955 .785 .905 r
F P
0 g
s
.0587 .06303 m .05253 .0423 L .05201 .02557 L closepath p .955 .785 .905 r
F P
0 g
s
.80728 .06076 m .80428 .06348 L .8044 .05061 L closepath p .373 .403 .853 r
F P
0 g
s
.80428 .06348 m .80728 .06076 L .80908 .07376 L closepath p .431 .497 .929 r
F P
0 g
s
.71461 .0615 m .7117 .06385 L .71283 .0506 L closepath p .298 .307 .777 r
F P
0 g
s
.7117 .06385 m .71461 .0615 L .71367 .0761 L closepath p .297 .305 .775 r
F P
0 g
s
.14509 .04334 m .13736 .03102 L .147 .02634 L closepath p .955 .785 .905 r
F P
0 g
s
.59824 .07578 m .5961 .07405 L .59688 .06428 L closepath p 1 .866 .46 r
F P
0 g
s
.59475 .06252 m .59688 .06428 L .5961 .07405 L closepath p 1 .866 .46 r
F P
0 g
s
.80481 .02663 m .80687 .0402 L .80329 .04027 L closepath p .955 .785 .905 r
F P
0 g
s
.80687 .0402 m .80481 .02663 L .8121 .032 L closepath p .955 .785 .905 r
F P
0 g
s
.59052 .0775 m .58629 .09252 L .58983 .02669 L closepath p .955 .785 .905 r
F P
0 g
s
.58983 .02669 m .59137 .03942 L .59052 .0775 L closepath p .955 .785 .905 r
F P
0 g
s
.54696 .02673 m .54356 .04269 L .52491 .0998 L closepath p .955 .785 .905 r
F P
0 g
s
.52086 .08875 m .54696 .02673 L .52491 .0998 L closepath p .955 .785 .905 r
F P
0 g
s
.54696 .02673 m .52086 .08875 L .52393 .07965 L closepath p .955 .785 .905 r
F P
0 g
s
.55401 .04017 m .54696 .02673 L .55896 .04209 L closepath p .955 .785 .905 r
F P
0 g
s
.52738 .06356 m .525 .06505 L .52984 .05072 L closepath p .151 .151 .444 r
F P
0 g
s
.525 .06505 m .52738 .06356 L .52332 .07428 L closepath p .152 .152 .453 r
F P
0 g
s
.81316 .03166 m .81657 .02745 L .81631 .04744 L closepath p .955 .785 .905 r
F P
0 g
s
.81543 .03833 m .81631 .04744 L .81657 .02745 L closepath p .955 .785 .905 r
F P
0 g
s
.81543 .03833 m .81657 .02745 L .81631 .04744 L closepath p .955 .785 .905 r
F P
0 g
s
.46763 .06266 m .46544 .06538 L .46574 .05947 L closepath p .365 .491 .947 r
F P
0 g
s
.46544 .06538 m .46763 .06266 L .46703 .0694 L closepath p .291 .383 .865 r
F P
0 g
s
.21882 .02781 m .20614 .08334 L .21525 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.65907 .02782 m .64821 .08334 L .65591 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.52984 .05072 m .53214 .03843 L .53427 .02828 L closepath p .955 .785 .905 r
F P
0 g
s
.7212 .02856 m .72257 .03723 L .71905 .09097 L closepath p .955 .785 .905 r
F P
0 g
s
.71905 .09097 m .71895 .08981 L .7212 .02856 L closepath p .955 .785 .905 r
F P
0 g
s
.1767 .02868 m .18749 .05199 L .18292 .04218 L closepath p .951 .783 .907 r
F P
0 g
s
.18749 .05199 m .1767 .02868 L .18527 .04705 L closepath p .958 .786 .903 r
F P
0 g
s
.61698 .02868 m .62851 .05199 L .62362 .04218 L closepath p .951 .783 .907 r
F P
0 g
s
.62851 .05199 m .61698 .02868 L .62612 .04706 L closepath p .958 .786 .903 r
F P
0 g
s
.8922 .04218 m .88534 .02868 L .89435 .03849 L closepath p .955 .785 .905 r
F P
0 g
s
.8942 .03825 m .89435 .03849 L .88534 .02868 L closepath p .955 .785 .905 r
F P
0 g
s
.28107 .06033 m .27997 .02873 L .28273 .04976 L closepath p .955 .785 .905 r
F P
0 g
s
.28107 .03723 m .28273 .04976 L .27997 .02873 L closepath p .955 .785 .905 r
F P
0 g
s
.28107 .03723 m .27997 .02873 L .28273 .04976 L closepath p .955 .785 .905 r
F P
0 g
s
.39366 .02876 m .39484 .03727 L .38383 .04476 L closepath p .955 .785 .905 r
F P
0 g
s
.38383 .04476 m .38284 .03766 L .39366 .02876 L closepath p .955 .785 .905 r
F P
0 g
s
.50359 .03357 m .50209 .02906 L .50583 .03096 L closepath p .955 .785 .905 r
F P
0 g
s
.7852 .02914 m .78653 .03723 L .77568 .04415 L closepath p .955 .785 .905 r
F P
0 g
s
.77568 .04415 m .77496 .03993 L .7852 .02914 L closepath p .955 .785 .905 r
F P
0 g
s
.11758 .06371 m .11669 .067 L .1083 .04485 L closepath p .164 .311 .836 r
F P
0 g
s
.11669 .067 m .11758 .06371 L .12745 .08907 L closepath p .163 .307 .833 r
F P
0 g
s
.38686 .06444 m .3848 .06709 L .38543 .05578 L closepath p .17 .184 .703 r
F P
0 g
s
.3848 .06709 m .38686 .06444 L .3862 .07547 L closepath p .171 .187 .705 r
F P
0 g
s
.82842 .07652 m .82718 .07566 L .82355 .06721 L closepath p .842 .432 .063 r
F P
0 g
s
.82218 .0661 m .82355 .06721 L .82718 .07566 L closepath p .84 .431 .062 r
F P
0 g
s
.14983 .02941 m .15104 .04394 L .14869 .05419 L closepath p .955 .785 .905 r
F P
0 g
s
.10078 .02945 m .09836 .03924 L .07286 .11305 L closepath p .955 .785 .905 r
F P
0 g
s
.10078 .02945 m .07286 .11305 L .06855 .10039 L closepath p .955 .785 .905 r
F P
0 g
s
.10078 .02945 m .14306 .10669 L .1083 .04485 L closepath p .955 .785 .905 r
F P
0 g
s
.13245 .0873 m .12186 .06796 L .10078 .02945 L closepath p .955 .785 .905 r
F P
0 g
s
.11131 .04868 m .10078 .02945 L .12186 .06796 L closepath p .955 .785 .905 r
F P
0 g
s
.13245 .0873 m .10078 .02945 L .13775 .09699 L closepath p .955 .785 .905 r
F P
0 g
s
.07152 .09304 m .08632 .05683 L .10078 .02945 L closepath p .955 .785 .905 r
F P
0 g
s
.10078 .02945 m .09677 .03828 L .07152 .09304 L closepath p .955 .785 .905 r
F P
0 g
s
.6334 .06181 m .62612 .04706 L .62108 .02971 L closepath p .955 .785 .905 r
F P
0 g
s
.62108 .02971 m .62564 .03801 L .6334 .06181 L closepath p .955 .785 .905 r
F P
0 g
s
.19209 .06181 m .18527 .04705 L .18068 .02988 L closepath p .955 .785 .905 r
F P
0 g
s
.18068 .02988 m .1849 .03801 L .19209 .06181 L closepath p .955 .785 .905 r
F P
0 g
s
.05552 .04067 m .0587 .06303 L .05291 .02989 L closepath p .955 .785 .905 r
F P
0 g
s
.87072 .03069 m .87024 .03108 L .87054 .02992 L closepath p .955 .785 .905 r
F P
0 g
s
.39895 .0662 m .40158 .06785 L .39995 .07317 L closepath p 1 .883 .524 r
F P
0 g
s
.21057 .10037 m .23007 .03073 L .22627 .0469 L closepath p .955 .785 .905 r
F P
0 g
s
.21057 .10037 m .20614 .08334 L .23007 .03073 L closepath p .955 .785 .905 r
F P
0 g
s
.93587 .0469 m .90986 .0766 L .9389 .03081 L closepath p .955 .785 .905 r
F P
0 g
s
.50583 .03096 m .508 .03868 L .50359 .03357 L closepath p .955 .785 .905 r
F P
0 g
s
.13736 .03102 m .14509 .04334 L .13813 .03965 L closepath p .955 .785 .905 r
F P
0 g
s
.0803 .05697 m .0833 .04306 L .08619 .0312 L closepath p .955 .785 .905 r
F P
0 g
s
.46909 .06687 m .46703 .0694 L .46763 .06266 L closepath p .23 .325 .824 r
F P
0 g
s
.46703 .0694 m .46909 .06687 L .46863 .07342 L closepath p .253 .358 .851 r
F P
0 g
s
.8121 .032 m .81419 .04811 L .80687 .0402 L closepath p .955 .785 .905 r
F P
0 g
s
.138 .08927 m .13518 .08891 L .12853 .07042 L closepath p 1 .701 .357 r
F P
0 g
s
.1255 .06965 m .12853 .07042 L .13518 .08891 L closepath p 1 .701 .357 r
F P
0 g
s
.34956 .07947 m .34676 .07753 L .34836 .07088 L closepath p 1 .892 .547 r
F P
0 g
s
.34544 .06811 m .34836 .07088 L .34676 .07753 L closepath p 1 .892 .546 r
F P
0 g
s
.5391 .08266 m .53682 .08036 L .54109 .07117 L closepath p 1 1 .727 r
F P
0 g
s
.53882 .06881 m .54109 .07117 L .53682 .08036 L closepath p 1 1 .727 r
F P
0 g
s
.34317 .05187 m .34023 .03356 L .34172 .04273 L closepath p .955 .785 .905 r
F P
0 g
s
.34023 .03356 m .34317 .05187 L .33351 .06916 L closepath p .955 .785 .905 r
F P
0 g
s
.7799 .06908 m .77683 .0714 L .77805 .05809 L closepath p .326 .33 .79 r
F P
0 g
s
.77683 .0714 m .7799 .06908 L .77818 .07964 L closepath p .325 .328 .789 r
F P
0 g
s
.5051 .03809 m .50359 .03357 L .508 .03868 L closepath p .955 .785 .905 r
F P
0 g
s
.43407 .0691 m .43186 .07142 L .43242 .05766 L closepath p .172 .184 .698 r
F P
0 g
s
.43186 .07142 m .43407 .0691 L .43302 .07969 L closepath p .172 .181 .695 r
F P
0 g
s
.33351 .06916 m .33156 .07148 L .33198 .05794 L closepath p .162 .162 .66 r
F P
0 g
s
.51537 .06936 m .51303 .07197 L .51112 .05619 L closepath p .277 .348 .833 r
F P
0 g
s
.51303 .07197 m .51537 .06936 L .51603 .08161 L closepath p .271 .34 .826 r
F P
0 g
s
.1584 .08502 m .15514 .08333 L .15729 .07205 L closepath p 1 .935 .646 r
F P
0 g
s
.15403 .07034 m .15729 .07205 L .15514 .08333 L closepath p 1 .935 .646 r
F P
0 g
s
.44349 .05139 m .43672 .09057 L .44109 .03458 L closepath p .955 .785 .905 r
F P
0 g
s
.79577 .08303 m .79416 .08146 L .79392 .07268 L closepath p 1 .761 .286 r
F P
0 g
s
.79231 .07106 m .79392 .07268 L .79416 .08146 L closepath p 1 .761 .286 r
F P
0 g
s
.47503 .07563 m .47347 .07204 L .47635 .07292 L closepath p 1 1 .894 r
F P
0 g
s
.47391 .07128 m .47635 .07292 L .47347 .07204 L closepath p 1 1 .926 r
F P
0 g
s
.07582 .07306 m .07704 .0716 L .07354 .08363 L closepath p .078 .078 .141 r
F P
0 g
s
.46768 .03529 m .46938 .0363 L .46647 .03612 L closepath p .955 .785 .905 r
F P
0 g
s
.47028 .06988 m .46863 .07342 L .46909 .06687 L closepath p .249 .354 .848 r
F P
0 g
s
.46863 .07342 m .47028 .06988 L .47023 .07575 L closepath p .395 .534 .978 r
F P
0 g
s
.96214 .07484 m .96248 .07496 L .95528 .07352 L closepath p .154 .005 .005 r
F P
0 g
s
.95651 .07386 m .95528 .07352 L .96248 .07496 L closepath p .167 .01 .01 r
F P
0 g
s
.95528 .07352 m .95651 .07386 L .94854 .07486 L closepath p .065 .052 .052 r
F P
0 g
s
.81195 .0707 m .80908 .07376 L .80728 .06076 L closepath p .434 .499 .93 r
F P
0 g
s
.80908 .07376 m .81195 .0707 L .81567 .08375 L closepath p .486 .578 .988 r
F P
0 g
s
.46606 .03863 m .46647 .03612 L .46938 .0363 L closepath p .955 .785 .905 r
F P
0 g
s
.46938 .0363 m .47017 .04401 L .46606 .03863 L closepath p .955 .785 .905 r
F P
0 g
s
.87962 .0751 m .87905 .07531 L .87683 .07414 L closepath p .359 .035 .035 r
F P
0 g
s
.87785 .07448 m .87683 .07414 L .87905 .07531 L closepath p .028 .236 .684 r
F P
0 g
s
.87683 .07414 m .87785 .07448 L .87478 .07503 L closepath p .064 .064 .064 r
F P
0 g
s
.52332 .07428 m .5254 .07321 L .52259 .07981 L closepath p .146 .146 .453 r
F P
0 g
s
.39995 .07317 m .40257 .0748 L .40076 .0788 L closepath p 1 .882 .523 r
F P
0 g
s
.95066 .07498 m .94854 .07486 L .95651 .07386 L closepath p .051 .047 .047 r
F P
0 g
s
.94854 .07486 m .95066 .07498 L .94282 .07869 L closepath p .106 .106 .106 r
F P
0 g
s
.87738 .07495 m .87478 .07503 L .87785 .07448 L closepath p .136 .136 .136 r
F P
0 g
s
.87478 .07503 m .87738 .07495 L .87361 .07686 L closepath p .154 .154 .27 r
F P
0 g
s
.71915 .09212 m .71905 .09097 L .72257 .03723 L closepath p .955 .785 .905 r
F P
0 g
s
.71915 .09212 m .72257 .03723 L .7266 .06232 L closepath p .955 .785 .905 r
F P
0 g
s
.72459 .04977 m .7266 .06232 L .72257 .03723 L closepath p .955 .785 .905 r
F P
0 g
s
.78653 .03723 m .78976 .05631 L .77641 .04837 L closepath p .955 .785 .905 r
F P
0 g
s
.77641 .04837 m .77568 .04415 L .78653 .03723 L closepath p .955 .785 .905 r
F P
0 g
s
.88162 .07757 m .88024 .07613 L .87962 .0751 L closepath p .606 .135 .015 r
F P
0 g
s
.87905 .07531 m .87962 .0751 L .88024 .07613 L closepath p .564 .045 .045 r
F P
0 g
s
.39828 .06848 m .39484 .03727 L .39639 .04824 L closepath p .955 .785 .905 r
F P
0 g
s
.38482 .05187 m .38383 .04476 L .39484 .03727 L closepath p .955 .785 .905 r
F P
0 g
s
.39484 .03727 m .40174 .09983 L .38543 .05578 L closepath p .955 .785 .905 r
F P
0 g
s
.38543 .05578 m .38482 .05187 L .39484 .03727 L closepath p .955 .785 .905 r
F P
0 g
s
.38828 .07311 m .3862 .07547 L .38686 .06444 L closepath p .169 .18 .7 r
F P
0 g
s
.47188 .07659 m .47181 .07229 L .47503 .07563 L closepath p .944 1 1 r
F P
0 g
s
.47347 .07204 m .47503 .07563 L .47181 .07229 L closepath p .872 1 1 r
F P
0 g
s
.0708 .08447 m .06826 .08418 L .0684 .07568 L closepath p 1 .817 .458 r
F P
0 g
s
.06526 .07476 m .0684 .07568 L .06826 .08418 L closepath p 1 .814 .491 r
F P
0 g
s
.58435 .07386 m .5817 .07574 L .58242 .05525 L closepath p .214 .22 .712 r
F P
0 g
s
.47118 .07169 m .47023 .07575 L .47028 .06988 L closepath p .234 .38 .879 r
F P
0 g
s
.47023 .07575 m .47118 .07169 L .47188 .07659 L closepath p .64 .818 1 r
F P
0 g
s
.59922 .08409 m .59709 .0824 L .59824 .07578 L closepath p 1 .866 .46 r
F P
0 g
s
.5961 .07405 m .59824 .07578 L .59709 .0824 L closepath p 1 .866 .46 r
F P
0 g
s
.1849 .03801 m .18546 .03896 L .19086 .05144 L closepath p .955 .785 .905 r
F P
0 g
s
.1849 .03801 m .18788 .04472 L .19892 .0766 L closepath p .955 .785 .905 r
F P
0 g
s
.19892 .0766 m .19209 .06181 L .1849 .03801 L closepath p .955 .785 .905 r
F P
0 g
s
.62564 .03801 m .62622 .03896 L .63033 .04801 L closepath p .955 .785 .905 r
F P
0 g
s
.63184 .05144 m .62564 .03801 L .63033 .04801 L closepath p .955 .785 .905 r
F P
0 g
s
.62564 .03801 m .63184 .05144 L .63316 .05728 L closepath p .955 .785 .905 r
F P
0 g
s
.6407 .0766 m .6334 .06181 L .62564 .03801 L closepath p .955 .785 .905 r
F P
0 g
s
.5066 .04261 m .5051 .03809 L .5108 .04772 L closepath p .955 .785 .905 r
F P
0 g
s
.508 .03868 m .5108 .04772 L .5051 .03809 L closepath p .955 .785 .905 r
F P
0 g
s
.71658 .0738 m .71367 .0761 L .71461 .0615 L closepath p .297 .305 .775 r
F P
0 g
s
.71367 .0761 m .71658 .0738 L .71583 .08976 L closepath p .296 .303 .774 r
F P
0 g
s
.07139 .09335 m .07152 .09304 L .09677 .03828 L closepath p .955 .785 .905 r
F P
0 g
s
.07139 .09335 m .09677 .03828 L .08471 .06483 L closepath p .955 .785 .905 r
F P
0 g
s
.09275 .04712 m .08471 .06483 L .09677 .03828 L closepath p .955 .785 .905 r
F P
0 g
s
.09275 .04712 m .09677 .03828 L .08471 .06483 L closepath p .955 .785 .905 r
F P
0 g
s
.89435 .03849 m .89449 .03872 L .8922 .04218 L closepath p .955 .785 .905 r
F P
0 g
s
.4656 .0451 m .46606 .03863 L .47017 .04401 L closepath p .955 .785 .905 r
F P
0 g
s
.83486 .08587 m .83395 .0854 L .82842 .07652 L closepath p .737 .282 .044 r
F P
0 g
s
.82718 .07566 m .82842 .07652 L .83395 .0854 L closepath p .749 .292 .048 r
F P
0 g
s
.89449 .03872 m .89464 .03896 L .90037 .05815 L closepath p .955 .785 .905 r
F P
0 g
s
.90037 .05815 m .8922 .04218 L .89449 .03872 L closepath p .955 .785 .905 r
F P
0 g
s
.47181 .07229 m .47188 .07659 L .47118 .07169 L closepath p .407 .617 1 r
F P
0 g
s
.18546 .03896 m .18598 .03969 L .18945 .04801 L closepath p .955 .785 .905 r
F P
0 g
s
.18945 .04801 m .19086 .05144 L .18546 .03896 L closepath p .955 .785 .905 r
F P
0 g
s
.62622 .03896 m .62675 .03969 L .62983 .04651 L closepath p .955 .785 .905 r
F P
0 g
s
.62983 .04651 m .63033 .04801 L .62622 .03896 L closepath p .955 .785 .905 r
F P
0 g
s
.90986 .0766 m .90037 .05815 L .89464 .03896 L closepath p .955 .785 .905 r
F P
0 g
s
.90986 .0766 m .89464 .03896 L .89623 .04051 L closepath p .955 .785 .905 r
F P
0 g
s
.89519 .03969 m .89623 .04051 L .89464 .03896 L closepath p .955 .785 .905 r
F P
0 g
s
.89519 .03969 m .89464 .03896 L .89623 .04051 L closepath p .955 .785 .905 r
F P
0 g
s
.87691 .07542 m .87361 .07686 L .87738 .07495 L closepath p .108 .108 .108 r
F P
0 g
s
.87361 .07686 m .87691 .07542 L .87235 .08254 L closepath p .237 .183 .56 r
F P
0 g
s
.0745 .11317 m .07286 .11305 L .09836 .03924 L closepath p .955 .785 .905 r
F P
0 g
s
.09836 .03924 m .09635 .04774 L .0745 .11317 L closepath p .955 .785 .905 r
F P
0 g
s
.94965 .07725 m .95178 .07736 L .94493 .08049 L closepath p .108 .108 .108 r
F P
0 g
s
.59137 .03942 m .59209 .04432 L .59475 .06252 L closepath p .955 .785 .905 r
F P
0 g
s
.59475 .06252 m .59052 .0775 L .59137 .03942 L closepath p .955 .785 .905 r
F P
0 g
s
.32943 .03953 m .33276 .04083 L .33062 .04807 L closepath p .955 .785 .905 r
F P
0 g
s
.13813 .03965 m .14316 .06038 L .13902 .05033 L closepath p .955 .785 .905 r
F P
0 g
s
.14316 .06038 m .13813 .03965 L .14509 .04334 L closepath p .955 .785 .905 r
F P
0 g
s
.18598 .03969 m .18687 .04051 L .18899 .04651 L closepath p .955 .785 .905 r
F P
0 g
s
.18899 .04651 m .18945 .04801 L .18598 .03969 L closepath p .955 .785 .905 r
F P
0 g
s
.62763 .04051 m .62983 .04651 L .62675 .03969 L closepath p .955 .785 .905 r
F P
0 g
s
.88273 .08772 m .88131 .08593 L .88162 .07757 L closepath p 1 .788 .304 r
F P
0 g
s
.88024 .07613 m .88162 .07757 L .88131 .08593 L closepath p 1 .788 .303 r
F P
0 g
s
.56267 .05663 m .55401 .04017 L .5696 .06193 L closepath p .955 .785 .905 r
F P
0 g
s
.55896 .04209 m .5696 .06193 L .55401 .04017 L closepath p .955 .785 .905 r
F P
0 g
s
.80687 .0402 m .80982 .0537 L .8044 .05061 L closepath p .955 .785 .905 r
F P
0 g
s
.80982 .0537 m .80687 .0402 L .81419 .04811 L closepath p .955 .785 .905 r
F P
0 g
s
.8044 .05061 m .80329 .04027 L .80687 .0402 L closepath p .955 .785 .905 r
F P
0 g
s
.18687 .04051 m .19033 .0409 L .19137 .04495 L closepath p .955 .785 .905 r
F P
0 g
s
.19137 .04495 m .18899 .04651 L .18687 .04051 L closepath p .955 .785 .905 r
F P
0 g
s
.62763 .04051 m .63099 .0409 L .62983 .04651 L closepath p .955 .785 .905 r
F P
0 g
s
.89962 .0409 m .90986 .0766 L .89623 .04051 L closepath p .955 .785 .905 r
F P
0 g
s
.05552 .04067 m .05593 .04224 L .0587 .06303 L closepath p .955 .785 .905 r
F P
0 g
s
.97215 .08426 m .97121 .08317 L .96807 .07861 L closepath p .617 .167 .012 r
F P
0 g
s
.96762 .07828 m .96807 .07861 L .97121 .08317 L closepath p .628 .173 .016 r
F P
0 g
s
.33276 .04083 m .33609 .04214 L .33198 .05794 L closepath p .955 .785 .905 r
F P
0 g
s
.33198 .05794 m .33062 .04807 L .33276 .04083 L closepath p .955 .785 .905 r
F P
0 g
s
.9457 .07836 m .94282 .07869 L .95066 .07498 L closepath p .11 .11 .11 r
F P
0 g
s
.94282 .07869 m .9457 .07836 L .93905 .08435 L closepath p .156 .156 .271 r
F P
0 g
s
.19033 .0409 m .19656 .0409 L .20278 .04499 L closepath p .955 .785 .905 r
F P
0 g
s
.20279 .0409 m .20278 .04499 L .19656 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.2142 .04503 m .20278 .04499 L .21525 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.20279 .0409 m .21525 .0409 L .20278 .04499 L closepath p .955 .785 .905 r
F P
0 g
s
.20278 .04499 m .19137 .04495 L .19033 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.20614 .08334 m .21018 .06416 L .21525 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.2142 .04503 m .21525 .0409 L .21018 .06416 L closepath p .955 .785 .905 r
F P
0 g
s
.63099 .0409 m .64345 .0409 L .63215 .04495 L closepath p .955 .785 .905 r
F P
0 g
s
.64353 .04499 m .63215 .04495 L .64345 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.65491 .04503 m .64353 .04499 L .65591 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.64345 .0409 m .65591 .0409 L .64353 .04499 L closepath p .955 .785 .905 r
F P
0 g
s
.63215 .04495 m .62983 .04651 L .63099 .0409 L closepath p .955 .785 .905 r
F P
0 g
s
.65591 .0409 m .64821 .08334 L .65491 .04503 L closepath p .955 .785 .905 r
F P
0 g
s
.76413 .07946 m .76385 .07973 L .76216 .07883 L closepath p .333 .032 .032 r
F P
0 g
s
.76287 .07918 m .76216 .07883 L .76385 .07973 L closepath p .004 .234 .631 r
F P
0 g
s
.76216 .07883 m .76287 .07918 L .75942 .08052 L closepath p .06 .06 .06 r
F P
0 g
s
.3226 .07951 m .32109 .07975 L .32069 .07885 L closepath p .463 .064 .064 r
F P
0 g
s
.32021 .07921 m .32069 .07885 L .32109 .07975 L closepath p .553 .05 .05 r
F P
0 g
s
.32069 .07885 m .32021 .07921 L .3179 .08054 L closepath p .005 .005 .005 r
F P
0 g
s
.24444 .08029 m .24255 .08029 L .24123 .07885 L closepath p .535 .068 .068 r
F P
0 g
s
.24053 .07921 m .24123 .07885 L .24255 .08029 L closepath p .541 .058 .058 r
F P
0 g
s
.24123 .07885 m .24053 .07921 L .239 .07967 L closepath p .008 .008 .008 r
F P
0 g
s
.68611 .08041 m .68541 .08042 L .68286 .07899 L closepath p .381 .026 .026 r
F P
0 g
s
.68336 .07934 m .68286 .07899 L .68541 .08042 L closepath p .357 .008 .008 r
F P
0 g
s
.68286 .07899 m .68336 .07934 L .68066 .0798 L closepath p .046 .046 .046 r
F P
0 g
s
.57856 .07874 m .5802 .07912 L .58596 .09252 L closepath p .96 .542 .106 r
F P
0 g
s
.57295 .07614 m .57067 .0792 L .56267 .05663 L closepath p .378 .48 .932 r
F P
0 g
s
.57067 .0792 m .57295 .07614 L .58252 .10171 L closepath p .378 .48 .932 r
F P
0 g
s
.22076 .09098 m .21769 .08867 L .22292 .07944 L closepath p 1 1 .801 r
F P
0 g
s
.21986 .07707 m .22292 .07944 L .21769 .08867 L closepath p 1 1 .801 r
F P
0 g
s
.93206 .08032 m .93078 .07797 L .93217 .07944 L closepath p 1 .961 .543 r
F P
0 g
s
.9309 .07707 m .93217 .07944 L .93078 .07797 L closepath p 1 .961 .543 r
F P
0 g
s
.66339 .08032 m .66143 .07797 L .66352 .07944 L closepath p 1 1 .673 r
F P
0 g
s
.66157 .07707 m .66352 .07944 L .66143 .07797 L closepath p 1 1 .673 r
F P
0 g
s
.49432 .0804 m .49238 .08036 L .49148 .07946 L closepath p .413 .052 .052 r
F P
0 g
s
.49145 .07981 m .49148 .07946 L .49238 .08036 L closepath p .467 .029 .029 r
F P
0 g
s
.49148 .07946 m .49145 .07981 L .48859 .08099 L closepath p .028 .028 .028 r
F P
0 g
s
.76673 .08141 m .76483 .08028 L .76413 .07946 L closepath p .504 .022 .022 r
F P
0 g
s
.76385 .07973 m .76413 .07946 L .76483 .08028 L closepath p .47 .034 .034 r
F P
0 g
s
.35045 .08562 m .3477 .08402 L .34956 .07947 L closepath p 1 .888 .542 r
F P
0 g
s
.34676 .07753 m .34956 .07947 L .3477 .08402 L closepath p 1 .889 .542 r
F P
0 g
s
.32501 .08148 m .32197 .0803 L .3226 .07951 L closepath p .745 .2 .087 r
F P
0 g
s
.32109 .07975 m .3226 .07951 L .32197 .0803 L closepath p .674 .102 .087 r
F P
0 g
s
.90986 .0766 m .90666 .07953 L .90037 .05815 L closepath p .475 .536 .951 r
F P
0 g
s
.90666 .07953 m .90986 .0766 L .91192 .08968 L closepath p .476 .537 .952 r
F P
0 g
s
.19892 .0766 m .19749 .07953 L .19032 .05815 L closepath p .199 .345 .854 r
F P
0 g
s
.19749 .07953 m .19892 .0766 L .20219 .08963 L closepath p .198 .345 .854 r
F P
0 g
s
.6407 .0766 m .63817 .07953 L .63152 .05815 L closepath p .393 .481 .928 r
F P
0 g
s
.63817 .07953 m .6407 .0766 L .64319 .08963 L closepath p .393 .481 .928 r
F P
0 g
s
.33458 .07729 m .33262 .07955 L .33351 .06916 L closepath p .161 .161 .656 r
F P
0 g
s
.78127 .07738 m .77818 .07964 L .7799 .06908 L closepath p .325 .328 .789 r
F P
0 g
s
.77818 .07964 m .78127 .07738 L .77916 .0859 L closepath p .322 .322 .783 r
F P
0 g
s
.23975 .07968 m .239 .07967 L .24053 .07921 L closepath p .036 .036 .036 r
F P
0 g
s
.239 .07967 m .23975 .07968 L .23684 .08234 L closepath p .053 .053 .053 r
F P
0 g
s
.43523 .07743 m .43302 .07969 L .43407 .0691 L closepath p .172 .18 .695 r
F P
0 g
s
.43302 .07969 m .43523 .07743 L .43384 .08595 L closepath p .171 .173 .688 r
F P
0 g
s
.68259 .07981 m .68066 .0798 L .68336 .07934 L closepath p .088 .088 .088 r
F P
0 g
s
.68066 .0798 m .68259 .07981 L .67859 .08246 L closepath p .122 .122 .122 r
F P
0 g
s
.52393 .07965 m .52259 .07981 L .5254 .07321 L closepath p .107 .107 .246 r
F P
0 g
s
.52259 .07981 m .52393 .07965 L .52302 .08142 L closepath p .113 .113 .575 r
F P
0 g
s
.33351 .06916 m .33198 .05794 L .33609 .04214 L closepath p .955 .785 .905 r
F P
0 g
s
.19032 .05815 m .18292 .04218 L .18749 .05199 L closepath p .946 .781 .91 r
F P
0 g
s
.63152 .05815 m .62362 .04218 L .62851 .05199 L closepath p .948 .782 .909 r
F P
0 g
s
.06304 .07814 m .0587 .06303 L .05593 .04224 L closepath p .955 .785 .905 r
F P
0 g
s
.05593 .04224 m .05634 .04381 L .06304 .07814 L closepath p .955 .785 .905 r
F P
0 g
s
.2482 .084 m .24571 .08355 L .24444 .08029 L closepath p .871 .359 .116 r
F P
0 g
s
.24255 .08029 m .24444 .08029 L .24571 .08355 L closepath p .869 .361 .111 r
F P
0 g
s
.93078 .07797 m .93206 .08032 L .93049 .0802 L closepath p 1 .96 .541 r
F P
0 g
s
.66143 .07797 m .66339 .08032 L .6611 .0802 L closepath p 1 1 .672 r
F P
0 g
s
.49572 .08287 m .49331 .08091 L .49432 .0804 L closepath p .926 .498 .096 r
F P
0 g
s
.49238 .08036 m .49432 .0804 L .49331 .08091 L closepath p .647 .082 .082 r
F P
0 g
s
.40076 .0788 m .40338 .0804 L .40139 .08308 L closepath p 1 .881 .521 r
F P
0 g
s
.68999 .0841 m .68867 .08365 L .68611 .08041 L closepath p .673 .146 .053 r
F P
0 g
s
.68541 .08042 m .68611 .08041 L .68867 .08365 L closepath p .657 .138 .045 r
F P
0 g
s
.50811 .04713 m .5066 .04261 L .51422 .05809 L closepath p .955 .785 .905 r
F P
0 g
s
.5108 .04772 m .51422 .05809 L .5066 .04261 L closepath p .955 .785 .905 r
F P
0 g
s
.94786 .08012 m .94493 .08049 L .95178 .07736 L closepath p .112 .112 .112 r
F P
0 g
s
.94493 .08049 m .94786 .08012 L .94194 .08529 L closepath p .164 .16 .301 r
F P
0 g
s
.76166 .08026 m .75942 .08052 L .76287 .07918 L closepath p .107 .107 .107 r
F P
0 g
s
.75942 .08052 m .76166 .08026 L .7571 .0845 L closepath p .137 .137 .207 r
F P
0 g
s
.31896 .08028 m .3179 .08054 L .32021 .07921 L closepath p .045 .045 .045 r
F P
0 g
s
.3179 .08054 m .31896 .08028 L .31546 .0845 L closepath p .068 .068 .068 r
F P
0 g
s
.54356 .04269 m .54148 .05416 L .5278 .09998 L closepath p .955 .785 .905 r
F P
0 g
s
.5278 .09998 m .52491 .0998 L .54356 .04269 L closepath p .955 .785 .905 r
F P
0 g
s
.09052 .09399 m .08712 .09171 L .09329 .08061 L closepath p 1 1 .829 r
F P
0 g
s
.08989 .07827 m .09329 .08061 L .08712 .09171 L closepath p 1 1 .829 r
F P
0 g
s
.06304 .07814 m .06183 .08073 L .0587 .06303 L closepath p .145 .145 .688 r
F P
0 g
s
.48992 .08089 m .48859 .08099 L .49145 .07981 L closepath p .061 .061 .061 r
F P
0 g
s
.48859 .08099 m .48992 .08089 L .48545 .08479 L closepath p .083 .083 .083 r
F P
0 g
s
.765 .09259 m .76326 .0905 L .76673 .08141 L closepath p 1 1 .639 r
F P
0 g
s
.76483 .08028 m .76673 .08141 L .76326 .0905 L closepath p 1 1 .641 r
F P
0 g
s
.32294 .09261 m .32008 .0905 L .32501 .08148 L closepath p 1 1 .783 r
F P
0 g
s
.32197 .0803 m .32501 .08148 L .32008 .0905 L closepath p 1 1 .784 r
F P
0 g
s
.91127 .08155 m .91417 .07822 L .91377 .08559 L closepath p .495 .573 .98 r
F P
0 g
s
.51842 .07918 m .51603 .08161 L .51537 .06936 L closepath p .268 .338 .824 r
F P
0 g
s
.51603 .08161 m .51842 .07918 L .51782 .08785 L closepath p .259 .323 .812 r
F P
0 g
s
.05634 .04381 m .05723 .04713 L .06613 .08942 L closepath p .955 .785 .905 r
F P
0 g
s
.06613 .08942 m .06304 .07814 L .05634 .04381 L closepath p .955 .785 .905 r
F P
0 g
s
.15104 .04394 m .1518 .05053 L .15403 .07034 L closepath p .955 .785 .905 r
F P
0 g
s
.14777 .09108 m .15104 .04394 L .15403 .07034 L closepath p .955 .785 .905 r
F P
0 g
s
.14777 .09108 m .14869 .05419 L .15104 .04394 L closepath p .955 .785 .905 r
F P
0 g
s
.47017 .04401 m .47096 .05174 L .4656 .0451 L closepath p .955 .785 .905 r
F P
0 g
s
.5945 .06078 m .59475 .06252 L .59209 .04432 L closepath p .955 .785 .905 r
F P
0 g
s
.59209 .04432 m .59281 .04922 L .5945 .06078 L closepath p .955 .785 .905 r
F P
0 g
s
.23897 .08015 m .23684 .08234 L .23975 .07968 L closepath p .626 .994 .843 r
F P
0 g
s
.23684 .08234 m .23897 .08015 L .23886 .08588 L closepath p .299 .444 .922 r
F P
0 g
s
.68183 .08028 m .67859 .08246 L .68259 .07981 L closepath p .425 .84 .654 r
F P
0 g
s
.67859 .08246 m .68183 .08028 L .68055 .08581 L closepath p .452 .541 .965 r
F P
0 g
s
.93049 .0802 m .93177 .08253 L .93003 .08377 L closepath p 1 .96 .54 r
F P
0 g
s
.6611 .0802 m .66306 .08253 L .66057 .08377 L closepath p 1 1 .671 r
F P
0 g
s
.87562 .08121 m .87235 .08254 L .87691 .07542 L closepath p .236 .182 .559 r
F P
0 g
s
.87235 .08254 m .87562 .08121 L .8712 .08694 L closepath p .225 .18 .533 r
F P
0 g
s
.20361 .08665 m .19892 .0766 L .18788 .04472 L closepath p .955 .785 .905 r
F P
0 g
s
.18788 .04472 m .19086 .05144 L .20361 .08665 L closepath p .955 .785 .905 r
F P
0 g
s
.53777 .09102 m .53548 .08879 L .5391 .08266 L closepath p 1 1 .717 r
F P
0 g
s
.53682 .08036 m .5391 .08266 L .53548 .08879 L closepath p 1 1 .717 r
F P
0 g
s
.1083 .04485 m .14306 .10669 L .11758 .06371 L closepath p .955 .785 .905 r
F P
0 g
s
.90079 .04495 m .90652 .04497 L .89898 .04534 L closepath p .955 .785 .905 r
F P
0 g
s
.90652 .04497 m .91225 .04499 L .90232 .04618 L closepath p .955 .785 .905 r
F P
0 g
s
.90232 .04618 m .89898 .04534 L .90652 .04497 L closepath p .955 .785 .905 r
F P
0 g
s
.91225 .04499 m .91798 .04501 L .90061 .05142 L closepath p .955 .785 .905 r
F P
0 g
s
.91798 .04501 m .92371 .04503 L .90334 .05699 L closepath p .955 .785 .905 r
F P
0 g
s
.90334 .05699 m .90061 .05142 L .91798 .04501 L closepath p .955 .785 .905 r
F P
0 g
s
.49542 .08613 m .49302 .08397 L .49572 .08287 L closepath p 1 1 .681 r
F P
0 g
s
.49331 .08091 m .49572 .08287 L .49302 .08397 L closepath p 1 1 .682 r
F P
0 g
s
.9163 .08184 m .92371 .04503 L .9175 .08334 L closepath p .955 .785 .905 r
F P
0 g
s
.91419 .07822 m .92371 .04503 L .9163 .08184 L closepath p .955 .785 .905 r
F P
0 g
s
.91115 .07248 m .92371 .04503 L .91419 .07822 L closepath p .955 .785 .905 r
F P
0 g
s
.90717 .06463 m .92371 .04503 L .91115 .07248 L closepath p .955 .785 .905 r
F P
0 g
s
.90717 .06463 m .90334 .05699 L .92371 .04503 L closepath p .955 .785 .905 r
F P
0 g
s
.46426 .04866 m .4656 .0451 L .47096 .05174 L closepath p .955 .785 .905 r
F P
0 g
s
.79692 .08902 m .79532 .08749 L .79577 .08303 L closepath p 1 .749 .271 r
F P
0 g
s
.79416 .08146 m .79577 .08303 L .79532 .08749 L closepath p 1 .749 .271 r
F P
0 g
s
.89898 .04534 m .90232 .04618 L .89841 .04651 L closepath p .955 .785 .905 r
F P
0 g
s
.07444 .08259 m .07354 .08363 L .07704 .0716 L closepath p .059 .059 .059 r
F P
0 g
s
.07354 .08363 m .07444 .08259 L .07239 .09005 L closepath p .064 .064 .114 r
F P
0 g
s
.81839 .08044 m .81567 .08375 L .81195 .0707 L closepath p .488 .58 .989 r
F P
0 g
s
.81567 .08375 m .81839 .08044 L .8243 .09324 L closepath p .543 .659 1 r
F P
0 g
s
.25251 .08873 m .25015 .08843 L .2482 .084 L closepath p .912 .418 .122 r
F P
0 g
s
.24571 .08355 m .2482 .084 L .25015 .08843 L closepath p .912 .418 .122 r
F P
0 g
s
.90232 .04618 m .90643 .04821 L .89897 .04793 L closepath p .955 .785 .905 r
F P
0 g
s
.89897 .04793 m .89841 .04651 L .90232 .04618 L closepath p .955 .785 .905 r
F P
0 g
s
.59982 .08918 m .59769 .08753 L .59922 .08409 L closepath p 1 .865 .46 r
F P
0 g
s
.59709 .0824 m .59922 .08409 L .59769 .08753 L closepath p 1 .865 .46 r
F P
0 g
s
.69444 .08879 m .69327 .08849 L .68999 .0841 L closepath p .716 .204 .058 r
F P
0 g
s
.68867 .08365 m .68999 .0841 L .69327 .08849 L closepath p .717 .204 .058 r
F P
0 g
s
.97374 .09093 m .97254 .08898 L .97215 .08426 L closepath p .908 .614 .076 r
F P
0 g
s
.97121 .08317 m .97215 .08426 L .97254 .08898 L closepath p .9 .611 .069 r
F P
0 g
s
.94237 .08327 m .93905 .08435 L .9457 .07836 L closepath p .153 .153 .252 r
F P
0 g
s
.93905 .08435 m .94237 .08327 L .93786 .09093 L closepath p .265 .187 .599 r
F P
0 g
s
.06826 .08418 m .0708 .08447 L .07128 .09361 L closepath p 1 .79 .429 r
F P
0 g
s
.75996 .08349 m .7571 .0845 L .76166 .08026 L closepath p .15 .15 .29 r
F P
0 g
s
.7571 .0845 m .75996 .08349 L .75481 .08888 L closepath p .151 .151 .294 r
F P
0 g
s
.31716 .08351 m .31546 .0845 L .31896 .08028 L closepath p .087 .087 .087 r
F P
0 g
s
.31546 .0845 m .31716 .08351 L .31298 .08888 L closepath p .086 .086 .086 r
F P
0 g
s
.40139 .08308 m .404 .08466 L .40183 .08601 L closepath p 1 .879 .518 r
F P
0 g
s
.21057 .10037 m .22627 .0469 L .21986 .07707 L closepath p .955 .785 .905 r
F P
0 g
s
.65738 .10006 m .66704 .0469 L .66157 .07707 L closepath p .955 .785 .905 r
F P
0 g
s
.65738 .10006 m .64821 .08334 L .66704 .0469 L closepath p .955 .785 .905 r
F P
0 g
s
.93587 .0469 m .9309 .07707 L .90986 .0766 L closepath p .955 .785 .905 r
F P
0 g
s
.4874 .08413 m .48545 .08479 L .48992 .08089 L closepath p .097 .097 .097 r
F P
0 g
s
.48545 .08479 m .4874 .08413 L .48205 .08938 L closepath p .099 .099 .099 r
F P
0 g
s
.13908 .09941 m .13775 .09699 L .11412 .04694 L closepath p .955 .785 .905 r
F P
0 g
s
.11412 .04694 m .12554 .06965 L .13908 .09941 L closepath p .955 .785 .905 r
F P
0 g
s
.38948 .08272 m .38738 .08487 L .38828 .07311 L closepath p .166 .166 .668 r
F P
0 g
s
.18527 .04705 m .19892 .0766 L .18749 .05199 L closepath p .959 .787 .903 r
F P
0 g
s
.62612 .04706 m .6407 .0766 L .62851 .05199 L closepath p .96 .787 .902 r
F P
0 g
s
.05922 .05431 m .06613 .08942 L .05723 .04713 L closepath p .955 .785 .905 r
F P
0 g
s
.50961 .05166 m .50811 .04713 L .51827 .0698 L closepath p .955 .785 .905 r
F P
0 g
s
.51422 .05809 m .51827 .0698 L .50811 .04713 L closepath p .955 .785 .905 r
F P
0 g
s
.15923 .09444 m .15597 .09281 L .1584 .08502 L closepath p 1 .935 .645 r
F P
0 g
s
.15514 .08333 m .1584 .08502 L .15597 .09281 L closepath p 1 .935 .645 r
F P
0 g
s
.94529 .08418 m .94194 .08529 L .94786 .08012 L closepath p .157 .157 .285 r
F P
0 g
s
.94194 .08529 m .94529 .08418 L .94101 .09093 L closepath p .272 .189 .61 r
F P
0 g
s
.81631 .04744 m .81856 .05669 L .81735 .06392 L closepath p .955 .785 .905 r
F P
0 g
s
.91623 .08184 m .91377 .08559 L .91417 .07822 L closepath p .462 .542 .963 r
F P
0 g
s
.91377 .08559 m .91623 .08184 L .91556 .08762 L closepath p .547 .656 1 r
F P
0 g
s
.09635 .04774 m .09476 .05496 L .07615 .11329 L closepath p .955 .785 .905 r
F P
0 g
s
.07615 .11329 m .0745 .11317 L .09635 .04774 L closepath p .955 .785 .905 r
F P
0 g
s
.35101 .08909 m .34827 .08757 L .35045 .08562 L closepath p 1 .88 .53 r
F P
0 g
s
.3477 .08402 m .35045 .08562 L .34827 .08757 L closepath p 1 .88 .53 r
F P
0 g
s
.33534 .08359 m .33337 .08575 L .33458 .07729 L closepath p .16 .16 .646 r
F P
0 g
s
.90061 .05142 m .89897 .04793 L .90643 .04821 L closepath p .955 .785 .905 r
F P
0 g
s
.68329 .08261 m .68055 .08581 L .68183 .08028 L closepath p .474 .564 .979 r
F P
0 g
s
.68055 .08581 m .68329 .08261 L .68252 .08917 L closepath p .463 .547 .967 r
F P
0 g
s
.84159 .09242 m .84123 .09236 L .83486 .08587 L closepath p .564 .05 .019 r
F P
0 g
s
.83395 .0854 m .83486 .08587 L .84123 .09236 L closepath p .585 .066 .026 r
F P
0 g
s
.24058 .08276 m .23886 .08588 L .23897 .08015 L closepath p .338 .481 .946 r
F P
0 g
s
.23886 .08588 m .24058 .08276 L .24088 .08942 L closepath p .319 .456 .928 r
F P
0 g
s
.78225 .08373 m .77916 .0859 L .78127 .07738 L closepath p .321 .321 .783 r
F P
0 g
s
.77916 .0859 m .78225 .08373 L .77975 .09018 L closepath p .316 .311 .774 r
F P
0 g
s
.43606 .08379 m .43384 .08595 L .43523 .07743 L closepath p .171 .172 .688 r
F P
0 g
s
.43384 .08595 m .43606 .08379 L .43433 .09022 L closepath p .17 .17 .675 r
F P
0 g
s
.81419 .04811 m .81629 .06425 L .80982 .0537 L closepath p .955 .785 .905 r
F P
0 g
s
.66057 .08377 m .66253 .08608 L .65984 .08867 L closepath p 1 1 .671 r
F P
0 g
s
.93068 .09097 m .92939 .08867 L .93131 .08608 L closepath p 1 .96 .54 r
F P
0 g
s
.93003 .08377 m .93131 .08608 L .92939 .08867 L closepath p 1 .96 .54 r
F P
0 g
s
.39639 .04824 m .39776 .05789 L .39828 .06848 L closepath p .955 .785 .905 r
F P
0 g
s
.49453 .09275 m .49213 .09049 L .49542 .08613 L closepath p 1 1 .711 r
F P
0 g
s
.49302 .08397 m .49542 .08613 L .49213 .09049 L closepath p 1 1 .711 r
F P
0 g
s
.78003 .06118 m .78065 .06928 L .77641 .04837 L closepath p .955 .785 .905 r
F P
0 g
s
.77805 .05809 m .77641 .04837 L .78065 .06928 L closepath p .955 .785 .905 r
F P
0 g
s
.79096 .09974 m .77641 .04837 L .78976 .05631 L closepath p .955 .785 .905 r
F P
0 g
s
.46426 .04866 m .47096 .05174 L .46046 .0507 L closepath p .955 .785 .905 r
F P
0 g
s
.87441 .08581 m .8712 .08694 L .87562 .08121 L closepath p .221 .18 .529 r
F P
0 g
s
.8712 .08694 m .87441 .08581 L .87017 .09012 L closepath p .205 .176 .485 r
F P
0 g
s
.45174 .09083 m .44953 .08907 L .45106 .08701 L closepath p 1 .829 .417 r
F P
0 g
s
.44856 .08539 m .45106 .08701 L .44953 .08907 L closepath p 1 .794 .398 r
F P
0 g
s
.59386 .05634 m .5945 .06078 L .59281 .04922 L closepath p .955 .785 .905 r
F P
0 g
s
.59386 .05634 m .59281 .04922 L .5945 .06078 L closepath p .955 .785 .905 r
F P
0 g
s
.40183 .08601 m .40444 .08757 L .40208 .08759 L closepath p 1 .874 .511 r
F P
0 g
s
.91733 .08334 m .91556 .08762 L .91623 .08184 L closepath p .485 .595 1 r
F P
0 g
s
.91556 .08762 m .91733 .08334 L .91849 .08762 L closepath p .837 1 1 r
F P
0 g
s
.9175 .08334 m .91849 .08762 L .91733 .08334 L closepath p .837 1 1 r
F P
0 g
s
.27608 .09212 m .28273 .04976 L .28773 .08749 L closepath p .955 .785 .905 r
F P
0 g
s
.2844 .06232 m .28773 .08749 L .28273 .04976 L closepath p .955 .785 .905 r
F P
0 g
s
.88131 .08593 m .88273 .08772 L .88163 .08793 L closepath p 1 .743 .247 r
F P
0 g
s
.52025 .08565 m .51782 .08785 L .51842 .07918 L closepath p .252 .317 .807 r
F P
0 g
s
.51782 .08785 m .52025 .08565 L .51841 .09069 L closepath p .222 .266 .763 r
F P
0 g
s
.14124 .07746 m .13902 .05033 L .14316 .06038 L closepath p .955 .785 .905 r
F P
0 g
s
.15256 .05712 m .15403 .07034 L .1518 .05053 L closepath p .955 .785 .905 r
F P
0 g
s
.15256 .05712 m .1518 .05053 L .15403 .07034 L closepath p .955 .785 .905 r
F P
0 g
s
.71283 .0506 m .71595 .05093 L .71461 .0615 L closepath p .955 .785 .905 r
F P
0 g
s
.80728 .06076 m .8044 .05061 L .80982 .0537 L closepath p .955 .785 .905 r
F P
0 g
s
.45359 .05139 m .46046 .0507 L .47096 .05174 L closepath p .955 .785 .905 r
F P
0 g
s
.25566 .09167 m .2535 .09154 L .25251 .08873 L closepath p .849 .327 .112 r
F P
0 g
s
.25015 .08843 m .25251 .08873 L .2535 .09154 L closepath p .849 .327 .112 r
F P
0 g
s
.69769 .0917 m .69671 .09157 L .69444 .08879 L closepath p .651 .116 .051 r
F P
0 g
s
.69327 .08849 m .69444 .08879 L .69671 .09157 L closepath p .651 .116 .051 r
F P
0 g
s
.71595 .05093 m .71908 .05127 L .71658 .0738 L closepath p .955 .785 .905 r
F P
0 g
s
.71658 .0738 m .71461 .0615 L .71595 .05093 L closepath p .955 .785 .905 r
F P
0 g
s
.31447 .08823 m .31298 .08888 L .31716 .08351 L closepath p .084 .084 .084 r
F P
0 g
s
.75747 .08822 m .75481 .08888 L .75996 .08349 L closepath p .15 .15 .287 r
F P
0 g
s
.79532 .08749 m .79692 .08902 L .79853 .09403 L closepath p .865 .464 .068 r
F P
0 g
s
.12864 .08605 m .12745 .08907 L .11758 .06371 L closepath p .172 .337 .853 r
F P
0 g
s
.12745 .08907 m .12864 .08605 L .14023 .11483 L closepath p .173 .338 .853 r
F P
0 g
s
.34827 .08757 m .35101 .08909 L .35131 .09412 L closepath p 1 .668 .267 r
F P
0 g
s
.29113 .09233 m .28849 .09064 L .29065 .08911 L closepath p 1 .88 .52 r
F P
0 g
s
.28773 .08749 m .29065 .08911 L .28849 .09064 L closepath p 1 .84 .489 r
F P
0 g
s
.73302 .09232 m .73152 .09064 L .73243 .08912 L closepath p 1 .737 .246 r
F P
0 g
s
.73065 .08749 m .73243 .08912 L .73152 .09064 L closepath p 1 .698 .221 r
F P
0 g
s
.40208 .08759 m .40469 .08913 L .40513 .09413 L closepath p 1 .65 .235 r
F P
0 g
s
.71875 .08749 m .71658 .0738 L .71908 .05127 L closepath p .955 .785 .905 r
F P
0 g
s
.68521 .0863 m .68252 .08917 L .68329 .08261 L closepath p .416 .501 .94 r
F P
0 g
s
.68252 .08917 m .68521 .0863 L .68491 .09421 L closepath p .402 .48 .923 r
F P
0 g
s
.59769 .08753 m .59982 .08918 L .59792 .08946 L closepath p 1 .864 .458 r
F P
0 g
s
.13518 .08891 m .138 .08927 L .14315 .10469 L closepath p 1 .7 .356 r
F P
0 g
s
.44349 .05139 m .45359 .05139 L .44423 .05538 L closepath p .955 .785 .905 r
F P
0 g
s
.47096 .05174 m .44423 .05538 L .45359 .05139 L closepath p .955 .785 .905 r
F P
0 g
s
.44856 .08539 m .44349 .05139 L .44423 .05538 L closepath p .955 .785 .905 r
F P
0 g
s
.44349 .05139 m .44856 .08539 L .43672 .09057 L closepath p .955 .785 .905 r
F P
0 g
s
.19086 .05144 m .19322 .0568 L .19578 .06364 L closepath p .955 .785 .905 r
F P
0 g
s
.21057 .10037 m .20361 .08665 L .19086 .05144 L closepath p .955 .785 .905 r
F P
0 g
s
.63435 .0568 m .63316 .05728 L .63184 .05144 L closepath p .955 .785 .905 r
F P
0 g
s
.48386 .0889 m .48205 .08938 L .4874 .08413 L closepath p .099 .099 .099 r
F P
0 g
s
.24248 .08655 m .24088 .08942 L .24058 .08276 L closepath p .241 .384 .881 r
F P
0 g
s
.24088 .08942 m .24248 .08655 L .24322 .09496 L closepath p .199 .33 .838 r
F P
0 g
s
.51112 .05619 m .50961 .05166 L .52295 .08287 L closepath p .955 .785 .905 r
F P
0 g
s
.51827 .0698 m .52295 .08287 L .50961 .05166 L closepath p .955 .785 .905 r
F P
0 g
s
.88374 .09313 m .88236 .09157 L .88303 .08957 L closepath p .991 .706 .203 r
F P
0 g
s
.88163 .08793 m .88303 .08957 L .88236 .09157 L closepath p .991 .707 .203 r
F P
0 g
s
.43656 .08817 m .43606 .08379 L .43756 .05169 L closepath p .955 .785 .905 r
F P
0 g
s
.43523 .07743 m .43756 .05169 L .43606 .08379 L closepath p .955 .785 .905 r
F P
0 g
s
.47096 .05174 m .4717 .05662 L .4523 .05538 L closepath p .955 .785 .905 r
F P
0 g
s
.47096 .05174 m .4523 .05538 L .44423 .05538 L closepath p .955 .785 .905 r
F P
0 g
s
.20361 .08665 m .20219 .08963 L .19892 .0766 L closepath p .198 .344 .853 r
F P
0 g
s
.20219 .08963 m .20361 .08665 L .20921 .10346 L closepath p .221 .374 .876 r
F P
0 g
s
.6457 .08665 m .64319 .08963 L .6407 .0766 L closepath p .392 .481 .928 r
F P
0 g
s
.64319 .08963 m .6457 .08665 L .65064 .10346 L closepath p .406 .501 .943 r
F P
0 g
s
.91509 .08669 m .91192 .08968 L .90986 .0766 L closepath p .475 .537 .952 r
F P
0 g
s
.91192 .08968 m .91509 .08669 L .91975 .10347 L closepath p .488 .557 .967 r
F P
0 g
s
.27566 .08749 m .27386 .08976 L .2713 .05197 L closepath p .154 .154 .626 r
F P
0 g
s
.71875 .08749 m .71583 .08976 L .71658 .0738 L closepath p .296 .303 .774 r
F P
0 g
s
.71583 .08976 m .71875 .08749 L .71619 .09406 L closepath p .267 .254 .728 r
F P
0 g
s
.34376 .05578 m .33351 .06916 L .34317 .05187 L closepath p .955 .785 .905 r
F P
0 g
s
.34317 .05187 m .38482 .05187 L .34376 .05578 L closepath p .955 .785 .905 r
F P
0 g
s
.38543 .05578 m .34376 .05578 L .38482 .05187 L closepath p .955 .785 .905 r
F P
0 g
s
.18749 .05199 m .19892 .0766 L .19032 .05815 L closepath p .949 .782 .908 r
F P
0 g
s
.62851 .05199 m .6407 .0766 L .63152 .05815 L closepath p .95 .783 .908 r
F P
0 g
s
.07252 .08994 m .07239 .09005 L .07444 .08259 L closepath p .009 .009 .009 r
F P
0 g
s
.33579 .08804 m .33381 .09009 L .33534 .08359 L closepath p .158 .158 .629 r
F P
0 g
s
.87328 .08921 m .87017 .09012 L .87441 .08581 L closepath p .2 .175 .478 r
F P
0 g
s
.87017 .09012 m .87328 .08921 L .86923 .09212 L closepath p .169 .166 .383 r
F P
0 g
s
.78286 .08812 m .77975 .09018 L .78225 .08373 L closepath p .314 .31 .773 r
F P
0 g
s
.77975 .09018 m .78286 .08812 L .77996 .09249 L closepath p .299 .282 .746 r
F P
0 g
s
.43656 .08817 m .43433 .09022 L .43606 .08379 L closepath p .169 .169 .674 r
F P
0 g
s
.52086 .08875 m .51841 .09069 L .52025 .08565 L closepath p .211 .254 .753 r
F P
0 g
s
.44953 .08907 m .45174 .09083 L .4505 .09276 L closepath p 1 .762 .331 r
F P
0 g
s
.94138 .08898 m .93786 .09093 L .94237 .08327 L closepath p .275 .189 .608 r
F P
0 g
s
.93786 .09093 m .94138 .08898 L .93944 .09754 L closepath p .4 .407 .845 r
F P
0 g
s
.9726 .09758 m .9716 .09477 L .97374 .09093 L closepath p 1 .966 .542 r
F P
0 g
s
.97254 .08898 m .97374 .09093 L .9716 .09477 L closepath p 1 .966 .544 r
F P
0 g
s
.94454 .08898 m .94101 .09093 L .94529 .08418 L closepath p .282 .19 .621 r
F P
0 g
s
.92963 .09746 m .92837 .09501 L .93068 .09097 L closepath p 1 .98 .573 r
F P
0 g
s
.92939 .08867 m .93068 .09097 L .92837 .09501 L closepath p 1 .98 .573 r
F P
0 g
s
.66063 .09746 m .6587 .09501 L .66181 .09097 L closepath p 1 1 .696 r
F P
0 g
s
.65984 .08867 m .66181 .09097 L .6587 .09501 L closepath p 1 1 .696 r
F P
0 g
s
.21939 .09746 m .21635 .09501 L .22076 .09098 L closepath p 1 1 .814 r
F P
0 g
s
.21769 .08867 m .22076 .09098 L .21635 .09501 L closepath p 1 1 .814 r
F P
0 g
s
.53709 .09624 m .53479 .09409 L .53777 .09102 L closepath p 1 1 .695 r
F P
0 g
s
.53548 .08879 m .53777 .09102 L .53479 .09409 L closepath p 1 1 .695 r
F P
0 g
s
.6017 .0948 m .60049 .09452 L .60004 .09104 L closepath p .931 .546 .086 r
F P
0 g
s
.59792 .08946 m .60004 .09104 L .60049 .09452 L closepath p .951 .542 .099 r
F P
0 g
s
.80982 .0537 m .81412 .06747 L .80728 .06076 L closepath p .955 .785 .905 r
F P
0 g
s
.81412 .06747 m .80982 .0537 L .81629 .06425 L closepath p .955 .785 .905 r
F P
0 g
s
.26046 .09458 m .25899 .0948 L .25566 .09167 L closepath p .654 .082 .082 r
F P
0 g
s
.2535 .09154 m .25566 .09167 L .25899 .0948 L closepath p .659 .085 .085 r
F P
0 g
s
.70258 .09459 m .7023 .09481 L .69769 .0917 L closepath p .471 .03 .03 r
F P
0 g
s
.69671 .09157 m .69769 .0917 L .7023 .09481 L closepath p .481 .035 .035 r
F P
0 g
s
.06614 .08942 m .06488 .09182 L .06304 .07814 L closepath p .143 .143 .675 r
F P
0 g
s
.31171 .09158 m .31053 .09187 L .31447 .08823 L closepath p .066 .066 .066 r
F P
0 g
s
.75484 .09157 m .75247 .09187 L .75747 .08822 L closepath p .131 .131 .131 r
F P
0 g
s
.54148 .05416 m .53882 .06881 L .53278 .10021 L closepath p .955 .785 .905 r
F P
0 g
s
.53278 .10021 m .5278 .09998 L .54148 .05416 L closepath p .955 .785 .905 r
F P
0 g
s
.87223 .09142 m .86923 .09212 L .87328 .08921 L closepath p .165 .165 .377 r
F P
0 g
s
.05922 .05431 m .06224 .06473 L .06613 .08942 L closepath p .955 .785 .905 r
F P
0 g
s
.73351 .09427 m .73239 .09379 L .73302 .09232 L closepath p .965 .656 .144 r
F P
0 g
s
.73152 .09064 m .73302 .09232 L .73239 .09379 L closepath p .966 .647 .139 r
F P
0 g
s
.29154 .09427 m .28924 .09379 L .29113 .09233 L closepath p 1 .834 .452 r
F P
0 g
s
.28849 .09064 m .29113 .09233 L .28924 .09379 L closepath p 1 .821 .442 r
F P
0 g
s
.481 .09207 m .47937 .09235 L .48386 .0889 L closepath p .088 .088 .088 r
F P
0 g
s
.78308 .09055 m .77996 .09249 L .78286 .08812 L closepath p .298 .28 .745 r
F P
0 g
s
.43672 .09057 m .43448 .09251 L .43656 .08817 L closepath p .166 .166 .637 r
F P
0 g
s
.38984 .09064 m .38773 .09258 L .38948 .08272 L closepath p .159 .159 .603 r
F P
0 g
s
.76422 .10005 m .76249 .09793 L .765 .09259 L closepath p 1 .978 .594 r
F P
0 g
s
.76326 .0905 m .765 .09259 L .76249 .09793 L closepath p 1 .978 .595 r
F P
0 g
s
.33593 .09066 m .33395 .09259 L .33579 .08804 L closepath p .154 .154 .587 r
F P
0 g
s
.32008 .0905 m .32294 .09261 L .31908 .09793 L closepath p 1 1 .753 r
F P
0 g
s
.4935 .09936 m .49117 .09682 L .49453 .09275 L closepath p 1 1 .724 r
F P
0 g
s
.49213 .09049 m .49453 .09275 L .49117 .09682 L closepath p 1 1 .724 r
F P
0 g
s
.07896 .11346 m .07615 .11329 L .09476 .05496 L closepath p .955 .785 .905 r
F P
0 g
s
.09476 .05496 m .09359 .06089 L .07896 .11346 L closepath p .955 .785 .905 r
F P
0 g
s
.8845 .09907 m .88308 .09715 L .88374 .09313 L closepath p 1 .77 .28 r
F P
0 g
s
.88236 .09157 m .88374 .09313 L .88308 .09715 L closepath p 1 .77 .281 r
F P
0 g
s
.58629 .09252 m .5853 .08318 L .58242 .05525 L closepath p .955 .785 .905 r
F P
0 g
s
.58433 .07386 m .58242 .05525 L .5853 .08318 L closepath p .955 .785 .905 r
F P
0 g
s
.8267 .08963 m .8243 .09324 L .81839 .08044 L closepath p .538 .655 1 r
F P
0 g
s
.8243 .09324 m .8267 .08963 L .83371 .10012 L closepath p .611 .754 1 r
F P
0 g
s
.45807 .05583 m .4523 .05538 L .4717 .05662 L closepath p .955 .785 .905 r
F P
0 g
s
.45391 .09515 m .45346 .09553 L .45244 .09331 L closepath p .738 .255 .053 r
F P
0 g
s
.4505 .09276 m .45244 .09331 L .45346 .09553 L closepath p .757 .227 .08 r
F P
0 g
s
.38543 .05578 m .38828 .07311 L .38684 .06446 L closepath p .955 .785 .905 r
F P
0 g
s
.34376 .05578 m .34544 .06811 L .3457 .07777 L closepath p .955 .785 .905 r
F P
0 g
s
.34376 .05578 m .34764 .09983 L .33593 .09066 L closepath p .955 .785 .905 r
F P
0 g
s
.38828 .07311 m .38543 .05578 L .40174 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.33579 .08804 m .34376 .05578 L .33593 .09066 L closepath p .955 .785 .905 r
F P
0 g
s
.33534 .08359 m .34376 .05578 L .33579 .08804 L closepath p .955 .785 .905 r
F P
0 g
s
.33458 .07729 m .34376 .05578 L .33534 .08359 L closepath p .955 .785 .905 r
F P
0 g
s
.33351 .06916 m .34376 .05578 L .33458 .07729 L closepath p .955 .785 .905 r
F P
0 g
s
.4717 .05662 m .47244 .0615 L .45807 .05583 L closepath p .955 .785 .905 r
F P
0 g
s
.46255 .0572 m .45807 .05583 L .47244 .0615 L closepath p .955 .785 .905 r
F P
0 g
s
.46255 .0572 m .47244 .0615 L .45807 .05583 L closepath p .955 .785 .905 r
F P
0 g
s
.08868 .10338 m .08526 .10117 L .09052 .09399 L closepath p 1 1 .824 r
F P
0 g
s
.08712 .09171 m .09052 .09399 L .08526 .10117 L closepath p 1 1 .824 r
F P
0 g
s
.27608 .09212 m .27425 .09406 L .27566 .08749 L closepath p .15 .15 .594 r
F P
0 g
s
.71915 .09212 m .71619 .09406 L .71875 .08749 L closepath p .267 .254 .728 r
F P
0 g
s
.51537 .06936 m .51112 .05619 L .52243 .08434 L closepath p .955 .785 .905 r
F P
0 g
s
.52295 .08287 m .52243 .08434 L .51112 .05619 L closepath p .955 .785 .905 r
F P
0 g
s
.51991 .09334 m .51792 .09409 L .52086 .08875 L closepath p .139 .139 .386 r
F P
0 g
s
.78976 .05631 m .79231 .07106 L .79036 .07799 L closepath p .955 .785 .905 r
F P
0 g
s
.68758 .09135 m .68491 .09421 L .68521 .0863 L closepath p .396 .475 .92 r
F P
0 g
s
.68491 .09421 m .68758 .09135 L .6877 .10056 L closepath p .387 .46 .909 r
F P
0 g
s
.28924 .09379 m .29154 .09427 L .29158 .09589 L closepath p .804 .273 .099 r
F P
0 g
s
.35763 .09591 m .35724 .09631 L .3532 .09441 L closepath p .395 .046 .046 r
F P
0 g
s
.41142 .09592 m .41118 .09631 L .40688 .09442 L closepath p .374 .041 .041 r
F P
0 g
s
.15976 .10031 m .15651 .09875 L .15923 .09444 L closepath p 1 .933 .642 r
F P
0 g
s
.15597 .09281 m .15923 .09444 L .15651 .09875 L closepath p 1 .933 .642 r
F P
0 g
s
.43593 .09379 m .43419 .0945 L .43672 .09057 L closepath p .123 .123 .303 r
F P
0 g
s
.57295 .07614 m .56267 .05663 L .5787 .07874 L closepath p .955 .785 .905 r
F P
0 g
s
.5696 .06193 m .5787 .07874 L .56267 .05663 L closepath p .955 .785 .905 r
F P
0 g
s
.38903 .09383 m .38741 .09454 L .38984 .09064 L closepath p .115 .115 .263 r
F P
0 g
s
.33512 .09384 m .33363 .09456 L .33593 .09066 L closepath p .106 .106 .23 r
F P
0 g
s
.26524 .09549 m .2646 .09589 L .26046 .09458 L closepath p .305 .046 .046 r
F P
0 g
s
.25899 .0948 m .26046 .09458 L .2646 .09589 L closepath p .303 .045 .045 r
F P
0 g
s
.82218 .0661 m .81735 .06392 L .81856 .05669 L closepath p .955 .785 .905 r
F P
0 g
s
.30651 .09481 m .30616 .09463 L .31171 .09158 L closepath p .032 .032 .032 r
F P
0 g
s
.7498 .09481 m .74825 .09463 L .75484 .09157 L closepath p .086 .086 .086 r
F P
0 g
s
.19754 .06689 m .19578 .06364 L .19322 .0568 L closepath p .954 .785 .905 r
F P
0 g
s
.19322 .0568 m .19652 .06408 L .19754 .06689 L closepath p .955 .785 .905 r
F P
0 g
s
.63316 .05728 m .63435 .0568 L .6407 .0766 L closepath p .955 .785 .905 r
F P
0 g
s
.63787 .06408 m .6407 .0766 L .63435 .0568 L closepath p .955 .785 .905 r
F P
0 g
s
.60522 .09581 m .6055 .09621 L .6017 .0948 L closepath p .3 .028 .028 r
F P
0 g
s
.24467 .09155 m .24322 .09496 L .24248 .08655 L closepath p .229 .356 .856 r
F P
0 g
s
.24322 .09496 m .24467 .09155 L .24555 .10051 L closepath p .217 .342 .846 r
F P
0 g
s
.78122 .09475 m .77897 .09509 L .78308 .09055 L closepath p .144 .144 .293 r
F P
0 g
s
.47244 .0615 m .47317 .06639 L .46255 .0572 L closepath p .955 .785 .905 r
F P
0 g
s
.46574 .05947 m .46255 .0572 L .47317 .06639 L closepath p .955 .785 .905 r
F P
0 g
s
.46574 .05947 m .47317 .06639 L .46255 .0572 L closepath p .955 .785 .905 r
F P
0 g
s
.47646 .09531 m .47558 .09515 L .481 .09207 L closepath p .058 .058 .058 r
F P
0 g
s
.46157 .09578 m .46132 .09618 L .45391 .09515 L closepath p .201 .038 .038 r
F P
0 g
s
.43362 .0956 m .43295 .09535 L .43593 .09379 L closepath p .055 .055 .055 r
F P
0 g
s
.38665 .09568 m .38609 .09544 L .38903 .09383 L closepath p .048 .048 .048 r
F P
0 g
s
.33278 .09569 m .33237 .09545 L .33512 .09384 L closepath p .041 .041 .041 r
F P
0 g
s
.26822 .09562 m .26743 .09601 L .26524 .09549 L closepath p .178 .039 .039 r
F P
0 g
s
.2646 .09589 m .26524 .09549 L .26743 .09601 L closepath p .178 .039 .039 r
F P
0 g
s
.30008 .09589 m .30062 .09549 L .30651 .09481 L closepath p .028 .028 .028 r
F P
0 g
s
.30062 .09549 m .30008 .09589 L .29214 .09549 L closepath p .14 .036 .036 r
F P
0 g
s
.74342 .09589 m .74276 .09549 L .7498 .09481 L closepath p .06 .046 .046 r
F P
0 g
s
.71033 .09562 m .71074 .09601 L .70739 .09549 L closepath p .161 .035 .035 r
F P
0 g
s
.70795 .09589 m .70739 .09549 L .71074 .09601 L closepath p .161 .034 .034 r
F P
0 g
s
.51706 .09581 m .51619 .09556 L .51991 .09334 L closepath p .068 .068 .068 r
F P
0 g
s
.27088 .09573 m .27026 .09613 L .26822 .09562 L closepath p .178 .039 .039 r
F P
0 g
s
.26743 .09601 m .26822 .09562 L .27026 .09613 L closepath p .178 .039 .039 r
F P
0 g
s
.71074 .09601 m .71033 .09562 L .71354 .09613 L closepath p .161 .034 .034 r
F P
0 g
s
.43145 .09603 m .43125 .09567 L .43362 .0956 L closepath p .036 .036 .036 r
F P
0 g
s
.27026 .09613 m .27088 .09573 L .27459 .09513 L closepath p .023 .023 .023 r
F P
0 g
s
.71354 .09613 m .71296 .09573 L .71778 .09513 L closepath p .047 .047 .047 r
F P
0 g
s
.47184 .09599 m .47176 .09639 L .46157 .09578 L closepath p .155 .037 .037 r
F P
0 g
s
.46132 .09618 m .46157 .09578 L .47176 .09639 L closepath p .154 .037 .037 r
F P
0 g
s
.39776 .05789 m .39895 .0662 L .39876 .06835 L closepath p .955 .785 .905 r
F P
0 g
s
.40174 .09983 m .39828 .06848 L .39776 .05789 L closepath p .955 .785 .905 r
F P
0 g
s
.38419 .09615 m .38414 .09579 L .38665 .09568 L closepath p .034 .034 .034 r
F P
0 g
s
.38414 .09579 m .38419 .09615 L .38057 .09591 L closepath p .105 .033 .033 r
F P
0 g
s
.3303 .09616 m .3304 .09579 L .33278 .09569 L closepath p .031 .031 .031 r
F P
0 g
s
.3304 .09579 m .3303 .09616 L .32675 .09592 L closepath p .104 .033 .033 r
F P
0 g
s
.80711 .09584 m .80775 .09623 L .80375 .09581 L closepath p .141 .035 .035 r
F P
0 g
s
.80457 .09621 m .80375 .09581 L .80775 .09623 L closepath p .142 .036 .036 r
F P
0 g
s
.77542 .09621 m .77466 .09581 L .78122 .09475 L closepath p .052 .052 .052 r
F P
0 g
s
.77355 .09621 m .77281 .09581 L .77542 .09621 L closepath p .14 .037 .037 r
F P
0 g
s
.77281 .09581 m .77355 .09621 L .76904 .09669 L closepath p .05 .05 .05 r
F P
0 g
s
.60934 .0978 m .60969 .09617 L .60522 .09581 L closepath p .241 .656 1 r
F P
0 g
s
.6055 .09621 m .60522 .09581 L .60969 .09617 L closepath p .134 .037 .037 r
F P
0 g
s
.42787 .09623 m .42804 .09583 L .43145 .09603 L closepath p .099 .036 .036 r
F P
0 g
s
.42804 .09583 m .42787 .09623 L .42437 .09674 L closepath p .032 .032 .032 r
F P
0 g
s
.80873 .09595 m .80903 .09629 L .80711 .09584 L closepath p .159 .029 .029 r
F P
0 g
s
.80775 .09623 m .80711 .09584 L .80903 .09629 L closepath p .159 .033 .033 r
F P
0 g
s
.51405 .09623 m .5137 .09585 L .51706 .09581 L closepath p .052 .038 .038 r
F P
0 g
s
.36099 .09594 m .36042 .09633 L .35763 .09591 L closepath p .144 .036 .036 r
F P
0 g
s
.35724 .09631 m .35763 .09591 L .36042 .09633 L closepath p .145 .037 .037 r
F P
0 g
s
.38024 .09631 m .38057 .09591 L .38419 .09615 L closepath p .108 .035 .035 r
F P
0 g
s
.38057 .09591 m .38024 .09631 L .37677 .0968 L closepath p .029 .029 .029 r
F P
0 g
s
.41479 .09594 m .41436 .09633 L .41142 .09592 L closepath p .144 .036 .036 r
F P
0 g
s
.41118 .09631 m .41142 .09592 L .41436 .09633 L closepath p .145 .037 .037 r
F P
0 g
s
.32628 .09631 m .32675 .09592 L .3303 .09616 L closepath p .107 .035 .035 r
F P
0 g
s
.32675 .09592 m .32628 .09631 L .323 .09681 L closepath p .026 .026 .026 r
F P
0 g
s
.36262 .09605 m .3617 .09639 L .36099 .09594 L closepath p .188 .037 .037 r
F P
0 g
s
.36042 .09633 m .36099 .09594 L .3617 .09639 L closepath p .178 .038 .038 r
F P
0 g
s
.41641 .09606 m .41563 .09639 L .41479 .09594 L closepath p .185 .036 .036 r
F P
0 g
s
.41436 .09633 m .41479 .09594 L .41563 .09639 L closepath p .176 .037 .037 r
F P
0 g
s
.81173 .09789 m .8102 .09786 L .80873 .09595 L closepath p .635 .071 .071 r
F P
0 g
s
.80903 .09629 m .80873 .09595 L .8102 .09786 L closepath p .003 .003 .434 r
F P
0 g
s
.50824 .09636 m .50821 .09596 L .51405 .09623 L closepath p .124 .037 .037 r
F P
0 g
s
.50821 .09596 m .50824 .09636 L .50579 .09704 L closepath p .034 .034 .034 r
F P
0 g
s
.78065 .06928 m .78127 .07738 L .77805 .05809 L closepath p .955 .785 .905 r
F P
0 g
s
.7799 .06908 m .77805 .05809 L .78127 .07738 L closepath p .955 .785 .905 r
F P
0 g
s
.7799 .06908 m .78127 .07738 L .77805 .05809 L closepath p .955 .785 .905 r
F P
0 g
s
.47176 .09639 m .47184 .09599 L .47646 .09531 L closepath p .034 .034 .034 r
F P
0 g
s
.3655 .09801 m .36277 .09796 L .36262 .09605 L closepath p .794 .245 .119 r
F P
0 g
s
.3617 .09639 m .36262 .09605 L .36277 .09796 L closepath p .86 .401 .083 r
F P
0 g
s
.4193 .09801 m .41672 .09797 L .41641 .09606 L closepath p .779 .226 .114 r
F P
0 g
s
.41563 .09639 m .41641 .09606 L .41672 .09797 L closepath p .831 .371 .074 r
F P
0 g
s
.86647 .0963 m .86443 .09622 L .87223 .09142 L closepath p .112 .112 .112 r
F P
0 g
s
.53573 .10197 m .53398 .09868 L .53709 .09624 L closepath p 1 1 .741 r
F P
0 g
s
.53479 .09409 m .53709 .09624 L .53398 .09868 L closepath p 1 1 .73 r
F P
0 g
s
.96671 .10123 m .96611 .0977 L .96968 .09649 L closepath p 1 1 .899 r
F P
0 g
s
.77113 .09652 m .76904 .09669 L .77355 .09621 L closepath p .009 .009 .009 r
F P
0 g
s
.76904 .09669 m .77113 .09652 L .76725 .09939 L closepath p .129 .129 .143 r
F P
0 g
s
.42556 .09655 m .42437 .09674 L .42787 .09623 L closepath p .106 .659 .709 r
F P
0 g
s
.42437 .09674 m .42556 .09655 L .42258 .09944 L closepath p .078 .078 .078 r
F P
0 g
s
.37781 .09662 m .37677 .0968 L .38024 .09631 L closepath p .104 .658 .714 r
F P
0 g
s
.37677 .0968 m .37781 .09662 L .37491 .09949 L closepath p .068 .068 .068 r
F P
0 g
s
.3239 .09663 m .323 .09681 L .32628 .09631 L closepath p .113 .666 .721 r
F P
0 g
s
.50767 .09691 m .50579 .09704 L .50824 .09636 L closepath p .101 .101 .101 r
F P
0 g
s
.50579 .09704 m .50767 .09691 L .50469 .09939 L closepath p .129 .129 .215 r
F P
0 g
s
.61368 .09769 m .61235 .09693 L .61203 .09721 L closepath p .033 .5 .801 r
F P
0 g
s
.61132 .09649 m .61203 .09721 L .61235 .09693 L closepath p .036 .417 .78 r
F P
0 g
s
.61028 .09605 m .60999 .09611 L .61003 .09747 L p .784 1 1 r
F P
0 g
s
.60992 .09797 m .61203 .09721 L .61159 .09676 L p .548 .974 .944 r
F P
0 g
s
.61066 .09719 m .60992 .09797 L p .61159 .09676 L .548 .974 .944 r
F P
0 g
s
.61028 .09605 m .60992 .09797 L .61132 .09649 L closepath p .34 .727 1 r
F P
0 g
s
.60992 .09797 m .61028 .09605 L .61007 .09866 L closepath p .068 .068 .392 r
F P
0 g
s
.61003 .09747 m .61007 .09866 L .61017 .09741 L p .784 1 1 r
F P
0 g
s
.61017 .09741 m .61028 .09605 L p .61003 .09747 L .784 1 1 r
F P
0 g
s
.61159 .09676 m .61132 .09649 L .61066 .09719 L p .548 .974 .944 r
F P
0 g
s
.47317 .06639 m .47391 .07128 L .46574 .05947 L closepath p .955 .785 .905 r
F P
0 g
s
.46763 .06266 m .46574 .05947 L .47391 .07128 L closepath p .955 .785 .905 r
F P
0 g
s
.92836 .1027 m .92714 .10006 L .92963 .09746 L closepath p 1 1 .65 r
F P
0 g
s
.92837 .09501 m .92963 .09746 L .92714 .10006 L closepath p 1 1 .65 r
F P
0 g
s
.65926 .1027 m .65738 .10006 L .66063 .09746 L closepath p 1 1 .756 r
F P
0 g
s
.6587 .09501 m .66063 .09746 L .65738 .10006 L closepath p 1 1 .756 r
F P
0 g
s
.21786 .1027 m .21487 .10006 L .21939 .09746 L closepath p 1 1 .854 r
F P
0 g
s
.21635 .09501 m .21939 .09746 L .21487 .10006 L closepath p 1 1 .854 r
F P
0 g
s
.85726 .09793 m .85629 .09753 L .86647 .0963 L closepath p .077 .051 .051 r
F P
0 g
s
.9427 .09472 m .93944 .09754 L .94138 .08898 L closepath p .391 .398 .839 r
F P
0 g
s
.93944 .09754 m .9427 .09472 L .94352 .10312 L closepath p .52 .612 1 r
F P
0 g
s
.96891 .10316 m .96836 .09958 L .9726 .09758 L closepath p 1 1 .916 r
F P
0 g
s
.9716 .09477 m .9726 .09758 L .96836 .09958 L closepath p 1 1 .917 r
F P
0 g
s
.61517 .10088 m .61306 .09896 L .61368 .09769 L closepath p .948 .553 .094 r
F P
0 g
s
.61235 .09693 m .61368 .09769 L .61306 .09896 L closepath p .955 .606 .102 r
F P
0 g
s
.61007 .09866 m .60999 .09611 L .60934 .0978 L closepath p .13 .353 .875 r
F P
0 g
s
.60969 .09617 m .60934 .0978 L .60999 .09611 L closepath p .616 1 1 r
F P
0 g
s
.81297 .10135 m .81137 .09943 L .81173 .09789 L closepath p .921 .57 .076 r
F P
0 g
s
.8102 .09786 m .81173 .09789 L .81137 .09943 L closepath p .898 .426 .102 r
F P
0 g
s
.3666 .10146 m .36385 .09954 L .3655 .09801 L closepath p 1 .766 .375 r
F P
0 g
s
.36277 .09796 m .3655 .09801 L .36385 .09954 L closepath p 1 .607 .268 r
F P
0 g
s
.42041 .10146 m .4178 .09954 L .4193 .09801 L closepath p 1 .75 .343 r
F P
0 g
s
.41672 .09797 m .4193 .09801 L .4178 .09954 L closepath p 1 .59 .24 r
F P
0 g
s
.81195 .0707 m .80728 .06076 L .81412 .06747 L closepath p .955 .785 .905 r
F P
0 g
s
.09359 .06089 m .08989 .07827 L .08177 .11362 L closepath p .955 .785 .905 r
F P
0 g
s
.08177 .11362 m .07896 .11346 L .09359 .06089 L closepath p .955 .785 .905 r
F P
0 g
s
.06797 .09684 m .06668 .09901 L .06614 .08942 L closepath p .139 .139 .648 r
F P
0 g
s
.88454 .10072 m .8831 .09809 L .8845 .09907 L closepath p 1 .858 .397 r
F P
0 g
s
.88308 .09715 m .8845 .09907 L .8831 .09809 L closepath p 1 .858 .397 r
F P
0 g
s
.78367 .07401 m .78225 .08373 L .78003 .06118 L closepath p .955 .785 .905 r
F P
0 g
s
.78127 .07738 m .78003 .06118 L .78225 .08373 L closepath p .955 .785 .905 r
F P
0 g
s
.49275 .10235 m .49063 .0993 L .4935 .09936 L closepath p 1 1 .78 r
F P
0 g
s
.49117 .09682 m .4935 .09936 L .49063 .0993 L closepath p 1 1 .766 r
F P
0 g
s
.77034 .09746 m .76725 .09939 L .77113 .09652 L closepath p .095 .095 .095 r
F P
0 g
s
.76725 .09939 m .77034 .09746 L .76833 .10282 L closepath p .358 .402 .858 r
F P
0 g
s
.5071 .09746 m .50469 .09939 L .50767 .09691 L closepath p .028 .028 .028 r
F P
0 g
s
.50469 .09939 m .5071 .09746 L .50577 .10284 L closepath p .245 .323 .817 r
F P
0 g
s
.42478 .09751 m .42258 .09944 L .42556 .09655 L closepath p .047 .047 .047 r
F P
0 g
s
.42258 .09944 m .42478 .09751 L .4235 .10289 L closepath p .185 .262 .772 r
F P
0 g
s
.37699 .09757 m .37491 .09949 L .37781 .09662 L closepath p .029 .029 .029 r
F P
0 g
s
.37491 .09949 m .37699 .09757 L .37593 .10294 L closepath p .174 .262 .777 r
F P
0 g
s
.76324 .10495 m .76199 .10171 L .76422 .10005 L closepath p 1 1 .635 r
F P
0 g
s
.76249 .09793 m .76422 .10005 L .76199 .10171 L closepath p 1 .993 .618 r
F P
0 g
s
.83563 .09621 m .83371 .10012 L .8267 .08963 L closepath p .605 .75 1 r
F P
0 g
s
.83371 .10012 m .83563 .09621 L .84381 .1043 L closepath p .693 .858 1 r
F P
0 g
s
.7266 .06232 m .73065 .08749 L .71915 .09212 L closepath p .955 .785 .905 r
F P
0 g
s
.16 .10265 m .15676 .10116 L .15976 .10031 L closepath p 1 .927 .634 r
F P
0 g
s
.15651 .09875 m .15976 .10031 L .15676 .10116 L closepath p 1 .927 .633 r
F P
0 g
s
.59228 .07155 m .59475 .06252 L .5961 .07405 L closepath p .955 .785 .905 r
F P
0 g
s
.58629 .09252 m .59475 .06252 L .58487 .09872 L closepath p .955 .785 .905 r
F P
0 g
s
.24715 .09774 m .24555 .10051 L .24467 .09155 L closepath p .182 .31 .824 r
F P
0 g
s
.24555 .10051 m .24715 .09774 L .24767 .10517 L closepath p .211 .349 .855 r
F P
0 g
s
.6904 .09778 m .6877 .10056 L .68758 .09135 L closepath p .386 .46 .909 r
F P
0 g
s
.6877 .10056 m .6904 .09778 L .69002 .1052 L closepath p .405 .49 .932 r
F P
0 g
s
.47077 .06696 m .47098 .06933 L .46763 .06266 L closepath p .955 .785 .905 r
F P
0 g
s
.46909 .06687 m .46763 .06266 L .47098 .06933 L closepath p .955 .785 .905 r
F P
0 g
s
.88287 .10325 m .88279 .09856 L .88454 .10072 L closepath p 1 1 .933 r
F P
0 g
s
.8831 .09809 m .88454 .10072 L .88279 .09856 L closepath p 1 1 .933 r
F P
0 g
s
.87663 .10098 m .87643 .09715 L .87722 .10077 L closepath p .943 1 1 r
F P
0 g
s
.87812 .09652 m .87722 .10077 L .87643 .09715 L closepath p .948 1 1 r
F P
0 g
s
.87722 .10077 m .87812 .09652 L .87838 .10124 L closepath p .69 .859 1 r
F P
0 g
s
.61419 .10327 m .61258 .09958 L .61517 .10088 L closepath p 1 1 .857 r
F P
0 g
s
.61306 .09896 m .61517 .10088 L .61258 .09958 L closepath p 1 1 .975 r
F P
0 g
s
.8749 .10179 m .87475 .09778 L .87663 .10098 L closepath p .98 1 1 r
F P
0 g
s
.87643 .09715 m .87663 .10098 L .87475 .09778 L closepath p .947 1 1 r
F P
0 g
s
.96203 .10442 m .96221 .10041 L .96671 .10123 L closepath p 1 1 1 r
F P
0 g
s
.96611 .0977 m .96671 .10123 L .96221 .10041 L closepath p 1 1 1 r
F P
0 g
s
.88049 .09731 m .87838 .10124 L .87812 .09652 L closepath p .723 .888 1 r
F P
0 g
s
.87838 .10124 m .88049 .09731 L .88287 .10325 L closepath p .691 .847 1 r
F P
0 g
s
.81186 .10353 m .81085 .09982 L .81297 .10135 L closepath p 1 1 .874 r
F P
0 g
s
.81137 .09943 m .81297 .10135 L .81085 .09982 L closepath p .993 1 1 r
F P
0 g
s
.36555 .10357 m .36344 .09993 L .3666 .10146 L closepath p 1 1 .925 r
F P
0 g
s
.36385 .09954 m .3666 .10146 L .36344 .09993 L closepath p 1 1 1 r
F P
0 g
s
.41937 .10357 m .4174 .09993 L .42041 .10146 L closepath p 1 1 .918 r
F P
0 g
s
.4178 .09954 m .42041 .10146 L .4174 .09993 L closepath p 1 1 1 r
F P
0 g
s
.19578 .06364 m .19754 .06689 L .2007 .07586 L closepath p .954 .785 .905 r
F P
0 g
s
.11758 .06371 m .14306 .10669 L .12864 .08605 L closepath p .955 .785 .905 r
F P
0 g
s
.58487 .09872 m .58252 .10171 L .57295 .07614 L closepath p .381 .483 .933 r
F P
0 g
s
.58252 .10171 m .58487 .09872 L .58416 .1036 L closepath p .476 .618 1 r
F P
0 g
s
.87241 .10295 m .87227 .09894 L .8749 .10179 L closepath p .98 1 1 r
F P
0 g
s
.87475 .09778 m .8749 .10179 L .87227 .09894 L closepath p .98 1 1 r
F P
0 g
s
.81735 .06392 m .82218 .0661 L .81839 .08044 L closepath p .955 .785 .905 r
F P
0 g
s
.53445 .10378 m .53338 .09945 L .53573 .10197 L closepath p 1 1 .983 r
F P
0 g
s
.53398 .09868 m .53573 .10197 L .53338 .09945 L closepath p 1 1 1 r
F P
0 g
s
.19996 .07147 m .19754 .06689 L .19652 .06408 L closepath p .955 .785 .905 r
F P
0 g
s
.63787 .06408 m .64151 .07147 L .6407 .0766 L closepath p .955 .785 .905 r
F P
0 g
s
.81629 .06425 m .81839 .08044 L .81195 .0707 L closepath p .955 .785 .905 r
F P
0 g
s
.06855 .10039 m .06725 .10231 L .06797 .09684 L closepath p .129 .129 .565 r
F P
0 g
s
.49075 .10399 m .48925 .09981 L .49275 .10235 L closepath p 1 1 1 r
F P
0 g
s
.49063 .0993 m .49275 .10235 L .48925 .09981 L closepath p .944 1 1 r
F P
0 g
s
.06224 .06473 m .06375 .06979 L .06797 .09684 L closepath p .955 .785 .905 r
F P
0 g
s
.06797 .09684 m .06613 .08942 L .06224 .06473 L closepath p .955 .785 .905 r
F P
0 g
s
.16194 .10729 m .15955 .10705 L .16 .10265 L closepath p 1 .721 .343 r
F P
0 g
s
.15676 .10116 m .16 .10265 L .15955 .10705 L closepath p 1 .709 .354 r
F P
0 g
s
.92683 .1048 m .92627 .10076 L .92836 .1027 L closepath p 1 1 .963 r
F P
0 g
s
.92714 .10006 m .92836 .1027 L .92627 .10076 L closepath p 1 1 1 r
F P
0 g
s
.65767 .10481 m .65645 .10076 L .65926 .1027 L closepath p 1 1 .996 r
F P
0 g
s
.65738 .10006 m .65926 .1027 L .65645 .10076 L closepath p 1 1 1 r
F P
0 g
s
.21623 .1048 m .21396 .10076 L .21786 .1027 L closepath p 1 1 1 r
F P
0 g
s
.21487 .10006 m .21786 .1027 L .21396 .10076 L closepath p 1 1 1 r
F P
0 g
s
.08471 .06483 m .07664 .08259 L .07139 .09335 L closepath p .955 .785 .905 r
F P
0 g
s
.77074 .09953 m .76833 .10282 L .77034 .09746 L closepath p .254 .286 .772 r
F P
0 g
s
.76833 .10282 m .77074 .09953 L .7712 .10446 L closepath p .623 .797 1 r
F P
0 g
s
.50751 .09953 m .50577 .10284 L .5071 .09746 L closepath p .166 .178 .705 r
F P
0 g
s
.50577 .10284 m .50751 .09953 L .5087 .10446 L closepath p .597 .794 1 r
F P
0 g
s
.42509 .09962 m .4235 .10289 L .42478 .09751 L closepath p .153 .153 .639 r
F P
0 g
s
.4235 .10289 m .42509 .09962 L .42624 .10457 L closepath p .566 .767 1 r
F P
0 g
s
.32344 .09964 m .32212 .10293 L .3231 .09757 L closepath p .142 .142 .61 r
F P
0 g
s
.32212 .10293 m .32344 .09964 L .32494 .10457 L closepath p .562 .774 1 r
F P
0 g
s
.37736 .09964 m .37593 .10294 L .37699 .09757 L closepath p .15 .15 .644 r
F P
0 g
s
.37593 .10294 m .37736 .09964 L .3788 .10457 L closepath p .574 .782 1 r
F P
0 g
s
.87129 .10348 m .87123 .09943 L .87241 .10295 L closepath p .981 1 1 r
F P
0 g
s
.87227 .09894 m .87241 .10295 L .87123 .09943 L closepath p .981 1 1 r
F P
0 g
s
.94629 .09955 m .94352 .10312 L .9427 .09472 L closepath p .527 .619 1 r
F P
0 g
s
.94352 .10312 m .94629 .09955 L .94947 .10684 L closepath p .647 .79 1 r
F P
0 g
s
.96322 .10685 m .96345 .10283 L .96891 .10316 L closepath p 1 1 1 r
F P
0 g
s
.96836 .09958 m .96891 .10316 L .96345 .10283 L closepath p 1 1 1 r
F P
0 g
s
.88279 .09856 m .88287 .10325 L .88049 .09731 L closepath p .631 .806 1 r
F P
0 g
s
.61183 .10446 m .61209 .10021 L .61419 .10327 L closepath p .989 1 1 r
F P
0 g
s
.61258 .09958 m .61419 .10327 L .61209 .10021 L closepath p 1 1 1 r
F P
0 g
s
.08777 .10879 m .08433 .10665 L .08868 .10338 L closepath p 1 1 .811 r
F P
0 g
s
.08526 .10117 m .08868 .10338 L .08433 .10665 L closepath p 1 1 .81 r
F P
0 g
s
.21057 .10037 m .20921 .10346 L .20361 .08665 L closepath p .221 .374 .876 r
F P
0 g
s
.20921 .10346 m .21057 .10037 L .21374 .10571 L closepath p .571 .802 1 r
F P
0 g
s
.65309 .10037 m .65064 .10346 L .6457 .08665 L closepath p .406 .501 .943 r
F P
0 g
s
.65064 .10346 m .65309 .10037 L .65516 .10571 L closepath p .625 .817 1 r
F P
0 g
s
.92286 .10037 m .91975 .10347 L .91509 .08669 L closepath p .488 .557 .967 r
F P
0 g
s
.91975 .10347 m .92286 .10037 L .92439 .10571 L closepath p .655 .829 1 r
F P
0 g
s
.86393 .10516 m .8644 .10096 L .87129 .10348 L closepath p .91 1 1 r
F P
0 g
s
.87123 .09943 m .87129 .10348 L .8644 .10096 L closepath p .908 1 1 r
F P
0 g
s
.80959 .10446 m .81033 .10021 L .81186 .10353 L closepath p .959 1 1 r
F P
0 g
s
.81085 .09982 m .81186 .10353 L .81033 .10021 L closepath p 1 1 1 r
F P
0 g
s
.36336 .10457 m .36303 .10033 L .36555 .10357 L closepath p .987 1 1 r
F P
0 g
s
.36344 .09993 m .36555 .10357 L .36303 .10033 L closepath p 1 1 1 r
F P
0 g
s
.41719 .10457 m .41699 .10033 L .41937 .10357 L closepath p .985 1 1 r
F P
0 g
s
.4174 .09993 m .41937 .10357 L .41699 .10033 L closepath p 1 1 1 r
F P
0 g
s
.58572 .09947 m .58416 .1036 L .58487 .09872 L closepath p .565 .707 1 r
F P
0 g
s
.58416 .1036 m .58572 .09947 L .58637 .10446 L closepath p .705 .867 1 r
F P
0 g
s
.53271 .10446 m .53278 .10021 L .53445 .10378 L closepath p .963 1 1 r
F P
0 g
s
.53338 .09945 m .53445 .10378 L .53278 .10021 L closepath p 1 1 1 r
F P
0 g
s
.78636 .10404 m .78699 .0998 L .79027 .10399 L closepath p .84 1 1 r
F P
0 g
s
.79096 .09974 m .79027 .10399 L .78699 .0998 L closepath p .841 1 1 r
F P
0 g
s
.79027 .10399 m .79096 .09974 L .79425 .10404 L closepath p .83 .999 1 r
F P
0 g
s
.69698 .10505 m .69592 .10135 L .69999 .10399 L closepath p .949 1 1 r
F P
0 g
s
.70047 .09974 m .69999 .10399 L .69592 .10135 L closepath p .949 1 1 r
F P
0 g
s
.69999 .10399 m .70047 .09974 L .75252 .10399 L closepath p .835 1 1 r
F P
0 g
s
.75313 .09974 m .75252 .10399 L .70047 .09974 L closepath p .835 1 1 r
F P
0 g
s
.75252 .10399 m .75313 .09974 L .75618 .10523 L closepath p .707 .883 1 r
F P
0 g
s
.51586 .10422 m .51577 .09998 L .52276 .10399 L closepath p .848 1 1 r
F P
0 g
s
.52282 .09974 m .52276 .10399 L .51577 .09998 L closepath p .848 1 1 r
F P
0 g
s
.52276 .10399 m .52282 .09974 L .52756 .10421 L closepath p .817 .988 1 r
F P
0 g
s
.59253 .10422 m .59261 .09998 L .59841 .10399 L closepath p .85 1 1 r
F P
0 g
s
.59864 .09974 m .59841 .10399 L .59261 .09998 L closepath p .85 1 1 r
F P
0 g
s
.59841 .10399 m .59864 .09974 L .60499 .10422 L closepath p .822 .992 1 r
F P
0 g
s
.4879 .10457 m .48788 .10033 L .49075 .10399 L closepath p .909 1 1 r
F P
0 g
s
.48925 .09981 m .49075 .10399 L .48788 .10033 L closepath p .974 1 1 r
F P
0 g
s
.44152 .10407 m .44132 .09983 L .44548 .104 L closepath p .842 1 1 r
F P
0 g
s
.44535 .09975 m .44548 .104 L .44132 .09983 L closepath p .842 1 1 r
F P
0 g
s
.44548 .104 m .44535 .09975 L .45432 .10407 L closepath p .832 1 1 r
F P
0 g
s
.34007 .10407 m .33962 .09983 L .34396 .104 L closepath p .843 1 1 r
F P
0 g
s
.34358 .09975 m .34396 .104 L .33962 .09983 L closepath p .843 1 1 r
F P
0 g
s
.34396 .104 m .34358 .09975 L .34794 .10407 L closepath p .828 .998 1 r
F P
0 g
s
.39401 .10407 m .39369 .09983 L .39796 .104 L closepath p .842 1 1 r
F P
0 g
s
.39771 .09975 m .39796 .104 L .39369 .09983 L closepath p .842 1 1 r
F P
0 g
s
.39796 .104 m .39771 .09975 L .40191 .10407 L closepath p .828 .998 1 r
F P
0 g
s
.25457 .10508 m .25248 .10135 L .25774 .104 L closepath p .958 1 1 r
F P
0 g
s
.25716 .09975 m .25774 .104 L .25248 .10135 L closepath p .958 1 1 r
F P
0 g
s
.25774 .104 m .25716 .09975 L .31027 .104 L closepath p .835 1 1 r
F P
0 g
s
.30982 .09975 m .31027 .104 L .25716 .09975 L closepath p .835 1 1 r
F P
0 g
s
.31027 .104 m .30982 .09975 L .31379 .10521 L closepath p .676 .877 1 r
F P
0 g
s
.82718 .07566 m .81839 .08044 L .82218 .0661 L closepath p .955 .785 .905 r
F P
0 g
s
.78082 .10422 m .7814 .09998 L .78636 .10404 L closepath p .846 1 1 r
F P
0 g
s
.78699 .0998 m .78636 .10404 L .7814 .09998 L closepath p .846 1 1 r
F P
0 g
s
.795 .0998 m .79425 .10404 L .79096 .09974 L closepath p .83 .999 1 r
F P
0 g
s
.79425 .10404 m .795 .0998 L .79984 .10422 L closepath p .824 .994 1 r
F P
0 g
s
.43597 .10428 m .4357 .10004 L .44152 .10407 L closepath p .85 1 1 r
F P
0 g
s
.44132 .09983 m .44152 .10407 L .4357 .10004 L closepath p .85 1 1 r
F P
0 g
s
.33456 .10428 m .33405 .10004 L .34007 .10407 L closepath p .851 1 1 r
F P
0 g
s
.33962 .09983 m .34007 .10407 L .33405 .10004 L closepath p .851 1 1 r
F P
0 g
s
.34764 .09983 m .34794 .10407 L .34358 .09975 L closepath p .828 .998 1 r
F P
0 g
s
.34794 .10407 m .34764 .09983 L .35355 .10428 L closepath p .82 .992 1 r
F P
0 g
s
.38849 .10428 m .38811 .10004 L .39401 .10407 L closepath p .851 1 1 r
F P
0 g
s
.39369 .09983 m .39401 .10407 L .38811 .10004 L closepath p .851 1 1 r
F P
0 g
s
.40174 .09983 m .40191 .10407 L .39771 .09975 L closepath p .828 .998 1 r
F P
0 g
s
.40191 .10407 m .40174 .09983 L .40746 .10428 L closepath p .82 .992 1 r
F P
0 g
s
.45424 .09983 m .45432 .10407 L .44535 .09975 L closepath p .832 1 1 r
F P
0 g
s
.45432 .10407 m .45424 .09983 L .46668 .10428 L closepath p .828 .998 1 r
F P
0 g
s
.39976 .07357 m .39876 .06835 L .39895 .0662 L closepath p .955 .785 .905 r
F P
0 g
s
.39895 .0662 m .39995 .07317 L .39976 .07357 L closepath p .955 .785 .905 r
F P
0 g
s
.5278 .09998 m .52756 .10421 L .52282 .09974 L closepath p .817 .989 1 r
F P
0 g
s
.52756 .10421 m .5278 .09998 L .53271 .10446 L closepath p .817 .989 1 r
F P
0 g
s
.58637 .10446 m .58657 .10021 L .59253 .10422 L closepath p .85 1 1 r
F P
0 g
s
.59261 .09998 m .59253 .10422 L .58657 .10021 L closepath p .85 1 1 r
F P
0 g
s
.5087 .10446 m .50872 .10021 L .51586 .10422 L closepath p .848 1 1 r
F P
0 g
s
.51577 .09998 m .51586 .10422 L .50872 .10021 L closepath p .848 1 1 r
F P
0 g
s
.60537 .09998 m .60499 .10422 L .59864 .09974 L closepath p .822 .993 1 r
F P
0 g
s
.60499 .10422 m .60537 .09998 L .61183 .10446 L closepath p .822 .993 1 r
F P
0 g
s
.77521 .1044 m .77582 .10015 L .78082 .10422 L closepath p .846 1 1 r
F P
0 g
s
.7814 .09998 m .78082 .10422 L .77582 .10015 L closepath p .846 1 1 r
F P
0 g
s
.80064 .09998 m .79984 .10422 L .795 .0998 L closepath p .824 .994 1 r
F P
0 g
s
.79984 .10422 m .80064 .09998 L .80551 .1044 L closepath p .824 .994 1 r
F P
0 g
s
.43032 .1045 m .43009 .10026 L .43597 .10428 L closepath p .85 1 1 r
F P
0 g
s
.4357 .10004 m .43597 .10428 L .43009 .10026 L closepath p .85 1 1 r
F P
0 g
s
.32896 .1045 m .32849 .10026 L .33456 .10428 L closepath p .851 1 1 r
F P
0 g
s
.33405 .10004 m .33456 .10428 L .32849 .10026 L closepath p .851 1 1 r
F P
0 g
s
.35331 .10004 m .35355 .10428 L .34764 .09983 L closepath p .82 .992 1 r
F P
0 g
s
.35355 .10428 m .35331 .10004 L .35926 .1045 L closepath p .82 .992 1 r
F P
0 g
s
.38287 .1045 m .38253 .10026 L .38849 .10428 L closepath p .851 1 1 r
F P
0 g
s
.38811 .10004 m .38849 .10428 L .38253 .10026 L closepath p .851 1 1 r
F P
0 g
s
.40735 .10004 m .40746 .10428 L .40174 .09983 L closepath p .82 .992 1 r
F P
0 g
s
.40746 .10428 m .40735 .10004 L .4131 .1045 L closepath p .82 .992 1 r
F P
0 g
s
.46665 .10004 m .46668 .10428 L .45424 .09983 L closepath p .828 .998 1 r
F P
0 g
s
.46668 .10428 m .46665 .10004 L .47907 .1045 L closepath p .828 .998 1 r
F P
0 g
s
.84517 .10015 m .84381 .1043 L .83563 .09621 L closepath p .693 .858 1 r
F P
0 g
s
.84381 .1043 m .84517 .10015 L .85448 .10571 L closepath p .787 .959 1 r
F P
0 g
s
.7712 .10446 m .77185 .10021 L .77521 .1044 L closepath p .841 1 1 r
F P
0 g
s
.77582 .10015 m .77521 .1044 L .77185 .10021 L closepath p .84 1 1 r
F P
0 g
s
.80629 .10015 m .80551 .1044 L .80064 .09998 L closepath p .824 .994 1 r
F P
0 g
s
.80551 .1044 m .80629 .10015 L .80959 .10446 L closepath p .83 .999 1 r
F P
0 g
s
.95631 .10555 m .9574 .10131 L .96203 .10442 L closepath p .897 1 1 r
F P
0 g
s
.96221 .10041 m .96203 .10442 L .9574 .10131 L closepath p .892 1 1 r
F P
0 g
s
.95052 .10443 m .95252 .10041 L .95631 .10555 L closepath p .766 .937 1 r
F P
0 g
s
.77185 .10021 m .7712 .10446 L .77074 .09953 L closepath p .599 .776 1 r
F P
0 g
s
.81033 .10021 m .80959 .10446 L .80629 .10015 L closepath p .83 .999 1 r
F P
0 g
s
.50872 .10021 m .5087 .10446 L .50751 .09953 L closepath p .589 .788 1 r
F P
0 g
s
.53278 .10021 m .53271 .10446 L .5278 .09998 L closepath p .817 .989 1 r
F P
0 g
s
.58657 .10021 m .58637 .10446 L .58572 .09947 L closepath p .467 .665 1 r
F P
0 g
s
.61209 .10021 m .61183 .10446 L .60537 .09998 L closepath p .822 .993 1 r
F P
0 g
s
.42624 .10457 m .42606 .10033 L .43032 .1045 L closepath p .842 1 1 r
F P
0 g
s
.43009 .10026 m .43032 .1045 L .42606 .10033 L closepath p .842 1 1 r
F P
0 g
s
.32494 .10457 m .32453 .10033 L .32896 .1045 L closepath p .843 1 1 r
F P
0 g
s
.32849 .10026 m .32896 .1045 L .32453 .10033 L closepath p .843 1 1 r
F P
0 g
s
.35898 .10026 m .35926 .1045 L .35331 .10004 L closepath p .82 .992 1 r
F P
0 g
s
.35926 .1045 m .35898 .10026 L .36336 .10457 L closepath p .828 .998 1 r
F P
0 g
s
.3788 .10457 m .37851 .10033 L .38287 .1045 L closepath p .842 1 1 r
F P
0 g
s
.38253 .10026 m .38287 .1045 L .37851 .10033 L closepath p .842 1 1 r
F P
0 g
s
.41296 .10026 m .4131 .1045 L .40735 .10004 L closepath p .82 .992 1 r
F P
0 g
s
.4131 .1045 m .41296 .10026 L .41719 .10457 L closepath p .828 .998 1 r
F P
0 g
s
.47905 .10026 m .47907 .1045 L .46665 .10004 L closepath p .828 .998 1 r
F P
0 g
s
.47907 .1045 m .47905 .10026 L .4879 .10457 L closepath p .832 1 1 r
F P
0 g
s
.42606 .10033 m .42624 .10457 L .42509 .09962 L closepath p .499 .712 1 r
F P
0 g
s
.48788 .10033 m .4879 .10457 L .47905 .10026 L closepath p .832 1 1 r
F P
0 g
s
.32453 .10033 m .32494 .10457 L .32344 .09964 L closepath p .528 .747 1 r
F P
0 g
s
.36303 .10033 m .36336 .10457 L .35898 .10026 L closepath p .828 .998 1 r
F P
0 g
s
.37851 .10033 m .3788 .10457 L .37736 .09964 L closepath p .554 .766 1 r
F P
0 g
s
.41699 .10033 m .41719 .10457 L .41296 .10026 L closepath p .828 .998 1 r
F P
0 g
s
.47098 .06933 m .47118 .07169 L .46909 .06687 L closepath p .955 .785 .905 r
F P
0 g
s
.47028 .06988 m .46909 .06687 L .47118 .07169 L closepath p .955 .785 .905 r
F P
0 g
s
.21374 .10571 m .21306 .10147 L .21623 .1048 L closepath p .966 1 1 r
F P
0 g
s
.21396 .10076 m .21623 .1048 L .21306 .10147 L closepath p 1 1 1 r
F P
0 g
s
.92439 .10571 m .9254 .10147 L .92683 .1048 L closepath p .946 1 1 r
F P
0 g
s
.92627 .10076 m .92683 .1048 L .9254 .10147 L closepath p 1 1 1 r
F P
0 g
s
.65516 .10571 m .65553 .10147 L .65767 .10481 L closepath p .952 1 1 r
F P
0 g
s
.65645 .10076 m .65767 .10481 L .65553 .10147 L closepath p 1 1 1 r
F P
0 g
s
.2017 .07651 m .2007 .07586 L .19754 .06689 L closepath p .955 .785 .905 r
F P
0 g
s
.19754 .06689 m .19996 .07147 L .2017 .07651 L closepath p .955 .785 .905 r
F P
0 g
s
.47391 .07128 m .47347 .07204 L .47077 .06696 L closepath p .955 .785 .905 r
F P
0 g
s
.47098 .06933 m .47077 .06696 L .47347 .07204 L closepath p .955 .785 .905 r
F P
0 g
s
.31815 .10719 m .31772 .10295 L .3208 .10494 L closepath p 1 1 1 r
F P
0 g
s
.31846 .1017 m .3208 .10494 L .31772 .10295 L closepath p 1 1 .961 r
F P
0 g
s
.76067 .1072 m .76129 .10296 L .76324 .10495 L closepath p 1 1 1 r
F P
0 g
s
.76199 .10171 m .76324 .10495 L .76129 .10296 L closepath p 1 1 .901 r
F P
0 g
s
.6926 .1072 m .69306 .10296 L .69698 .10505 L closepath p .983 1 1 r
F P
0 g
s
.69592 .10135 m .69698 .10505 L .69306 .10296 L closepath p 1 1 1 r
F P
0 g
s
.25009 .10719 m .24949 .10295 L .25457 .10508 L closepath p .993 1 1 r
F P
0 g
s
.25248 .10135 m .25457 .10508 L .24949 .10295 L closepath p 1 1 1 r
F P
0 g
s
.85448 .10571 m .85532 .10147 L .86393 .10516 L closepath p .855 1 1 r
F P
0 g
s
.8644 .10096 m .86393 .10516 L .85532 .10147 L closepath p .853 1 1 r
F P
0 g
s
.24874 .10165 m .24767 .10517 L .24715 .09774 L closepath p .166 .281 .806 r
F P
0 g
s
.24767 .10517 m .24874 .10165 L .25009 .10719 L closepath p .484 .681 1 r
F P
0 g
s
.69215 .10167 m .69002 .1052 L .6904 .09778 L closepath p .358 .445 .904 r
F P
0 g
s
.69002 .1052 m .69215 .10167 L .6926 .1072 L closepath p .579 .733 1 r
F P
0 g
s
.31466 .10135 m .31379 .10521 L .30982 .09975 L closepath p .688 .883 1 r
F P
0 g
s
.31379 .10521 m .31466 .10135 L .31815 .10719 L closepath p .64 .835 1 r
F P
0 g
s
.7581 .10135 m .75618 .10523 L .75313 .09974 L closepath p .716 .889 1 r
F P
0 g
s
.75618 .10523 m .7581 .10135 L .76067 .1072 L closepath p .682 .848 1 r
F P
0 g
s
.9574 .10131 m .95631 .10555 L .95252 .10041 L closepath p .772 .94 1 r
F P
0 g
s
.21306 .10147 m .21374 .10571 L .21057 .10037 L closepath p .616 .832 1 r
F P
0 g
s
.9254 .10147 m .92439 .10571 L .92286 .10037 L closepath p .683 .85 1 r
F P
0 g
s
.85532 .10147 m .85448 .10571 L .84517 .10015 L closepath p .789 .96 1 r
F P
0 g
s
.65553 .10147 m .65516 .10571 L .65309 .10037 L closepath p .655 .839 1 r
F P
0 g
s
.34679 .08163 m .3457 .07777 L .34544 .06811 L closepath p .954 .784 .905 r
F P
0 g
s
.34544 .06811 m .34676 .07753 L .34679 .08163 L closepath p .955 .785 .905 r
F P
0 g
s
.39876 .06835 m .40076 .0788 L .39975 .07883 L closepath p .955 .785 .905 r
F P
0 g
s
.53308 .09983 m .53278 .10021 L .53882 .06881 L closepath p .955 .785 .905 r
F P
0 g
s
.53882 .06881 m .53682 .08036 L .53308 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.95144 .10282 m .94947 .10684 L .94629 .09955 L closepath p .642 .786 1 r
F P
0 g
s
.94947 .10684 m .95144 .10282 L .9564 .10814 L closepath p .768 .939 1 r
F P
0 g
s
.9564 .10814 m .95748 .10391 L .96322 .10685 L closepath p .894 1 1 r
F P
0 g
s
.96345 .10283 m .96322 .10685 L .95748 .10391 L closepath p .889 1 1 r
F P
0 g
s
.24949 .10295 m .25009 .10719 L .24874 .10165 L closepath p .13 .22 .777 r
F P
0 g
s
.31772 .10295 m .31815 .10719 L .31466 .10135 L closepath p .585 .798 1 r
F P
0 g
s
.69306 .10296 m .6926 .1072 L .69215 .10167 L closepath p .299 .481 .955 r
F P
0 g
s
.76129 .10296 m .76067 .1072 L .7581 .10135 L closepath p .642 .82 1 r
F P
0 g
s
.47347 .07204 m .47181 .07229 L .47098 .06933 L closepath p .955 .785 .905 r
F P
0 g
s
.47118 .07169 m .47098 .06933 L .47181 .07229 L closepath p .955 .785 .905 r
F P
0 g
s
.51842 .07918 m .51537 .06936 L .52191 .08581 L closepath p .955 .785 .905 r
F P
0 g
s
.52243 .08434 m .52191 .08581 L .51537 .06936 L closepath p .955 .785 .905 r
F P
0 g
s
.16631 .10859 m .1654 .10902 L .16194 .10729 L closepath p .422 .06 .06 r
F P
0 g
s
.14041 .10183 m .13908 .09941 L .12554 .06965 L closepath p .955 .785 .905 r
F P
0 g
s
.12554 .06965 m .13532 .08891 L .14041 .10183 L closepath p .955 .785 .905 r
F P
0 g
s
.06525 .07476 m .06797 .09684 L .06375 .06979 L closepath p .955 .785 .905 r
F P
0 g
s
.95748 .10391 m .9564 .10814 L .95144 .10282 L closepath p .773 .942 1 r
F P
0 g
s
.06433 .10817 m .06401 .10844 L .06145 .10858 L closepath p .021 .021 .021 r
F P
0 g
s
.1509 .0807 m .15403 .07034 L .15514 .08333 L closepath p .955 .785 .905 r
F P
0 g
s
.06061 .10898 m .06145 .10858 L .06401 .10844 L closepath p .025 .025 .025 r
F P
0 g
s
.06145 .10858 m .06061 .10898 L .05523 .10872 L closepath p .118 .035 .035 r
F P
0 g
s
.17108 .11059 m .17027 .10898 L .16631 .10859 L closepath p .136 .613 .984 r
F P
0 g
s
.1654 .10902 m .16631 .10859 L .17027 .10898 L closepath p .144 .039 .039 r
F P
0 g
s
.05523 .10872 m .05402 .10916 L .0526 .10982 L closepath p .299 .777 .676 r
F P
0 g
s
.05402 .10916 m .05523 .10872 L .06061 .10898 L closepath p .122 .037 .037 r
F P
0 g
s
.08608 .11514 m .08322 .11188 L .08777 .10879 L closepath p 1 1 .842 r
F P
0 g
s
.08433 .10665 m .08777 .10879 L .08322 .11188 L closepath p 1 1 .831 r
F P
0 g
s
.79416 .08146 m .79036 .07799 L .79231 .07106 L closepath p .955 .785 .905 r
F P
0 g
s
.20269 .07715 m .2017 .07651 L .19996 .07147 L closepath p .955 .785 .905 r
F P
0 g
s
.64486 .08246 m .6407 .0766 L .64151 .07147 L closepath p .955 .785 .905 r
F P
0 g
s
.64151 .07147 m .6444 .07715 L .64486 .08246 L closepath p .955 .785 .905 r
F P
0 g
s
.5961 .07405 m .59709 .0824 L .59228 .07155 L closepath p .955 .785 .905 r
F P
0 g
s
.58981 .08059 m .59228 .07155 L .59709 .0824 L closepath p .955 .785 .905 r
F P
0 g
s
.05331 .10979 m .0526 .10982 L .05402 .10916 L closepath p .04 .04 .04 r
F P
0 g
s
.05299 .11282 m .05236 .1161 L .05257 .11108 L p .099 .099 .465 r
F P
0 g
s
.05257 .11108 m .0526 .11042 L .0527 .111 L p .099 .099 .465 r
F P
0 g
s
.0527 .111 m .05299 .11282 L p .05257 .11108 L .099 .099 .465 r
F P
0 g
s
.0526 .10982 m .05331 .10979 L .05199 .11147 L p .055 .055 .055 r
F P
0 g
s
.05199 .11147 m .05134 .11231 L .05165 .11171 L p .055 .055 .055 r
F P
0 g
s
.05165 .11171 m .0526 .10982 L p .05199 .11147 L .055 .055 .055 r
F P
0 g
s
.17554 .11055 m .17304 .10983 L .17384 .11002 L closepath p .235 .006 .006 r
F P
0 g
s
.17384 .11002 m .1719 .10935 L .17164 .11076 L closepath p .63 1 1 r
F P
0 g
s
.1719 .10935 m .17384 .11002 L .17304 .10983 L closepath p .326 .013 .013 r
F P
0 g
s
.17699 .114 m .17376 .11211 L .17554 .11055 L closepath p 1 .732 .368 r
F P
0 g
s
.17304 .10983 m .17554 .11055 L .17376 .11211 L closepath p 1 .78 .402 r
F P
0 g
s
.17187 .11145 m .17051 .10892 L .17108 .11059 L closepath p .065 .168 .702 r
F P
0 g
s
.17027 .10898 m .17108 .11059 L .17051 .10892 L closepath p .683 1 1 r
F P
0 g
s
.17076 .10886 m .17164 .11076 L .1719 .10935 L closepath p .206 .654 1 r
F P
0 g
s
.17164 .11076 m .17076 .10886 L .17187 .11145 L closepath p .761 .446 .024 r
F P
0 g
s
.40174 .09983 m .39539 .09683 L .38828 .07311 L closepath p .955 .785 .905 r
F P
0 g
s
.38948 .08272 m .38828 .07311 L .39539 .09683 L closepath p .955 .785 .905 r
F P
0 g
s
.40076 .0788 m .39976 .07357 L .39995 .07317 L closepath p .955 .785 .905 r
F P
0 g
s
.17051 .10892 m .17187 .11145 L .17076 .10886 L closepath p .833 1 1 r
F P
0 g
s
.94786 .08012 m .95178 .07736 L .94529 .08418 L closepath p .955 .785 .905 r
F P
0 g
s
.95066 .07498 m .95651 .07386 L .9461 .08465 L closepath p .955 .785 .905 r
F P
0 g
s
.78731 .08686 m .78286 .08812 L .78367 .07401 L closepath p .955 .785 .905 r
F P
0 g
s
.78225 .08373 m .78367 .07401 L .78286 .08812 L closepath p .955 .785 .905 r
F P
0 g
s
.0526 .11042 m .05134 .11231 L .05331 .10979 L closepath p .668 .972 .714 r
F P
0 g
s
.05134 .11231 m .0526 .11042 L .05236 .1161 L closepath p .136 .136 .649 r
F P
0 g
s
.87785 .07448 m .87691 .07542 L .87725 .07472 L closepath p .955 .785 .905 r
F P
0 g
s
.87785 .07448 m .88024 .07613 L .88131 .08593 L closepath p .955 .785 .905 r
F P
0 g
s
.87785 .07448 m .87872 .07734 L .87691 .07542 L closepath p .955 .785 .905 r
F P
0 g
s
.06525 .07476 m .06676 .07962 L .06797 .09684 L closepath p .955 .785 .905 r
F P
0 g
s
.96248 .07496 m .96762 .07828 L .96095 .08461 L closepath p .955 .785 .905 r
F P
0 g
s
.9461 .08465 m .94237 .08327 L .95066 .07498 L closepath p .955 .785 .905 r
F P
0 g
s
.9457 .07836 m .95066 .07498 L .94237 .08327 L closepath p .955 .785 .905 r
F P
0 g
s
.87562 .08121 m .87691 .07542 L .87872 .07734 L closepath p .955 .785 .905 r
F P
0 g
s
.82718 .07566 m .83395 .0854 L .81839 .08044 L closepath p .955 .785 .905 r
F P
0 g
s
.2007 .07586 m .20269 .07715 L .20563 .08811 L closepath p .955 .785 .905 r
F P
0 g
s
.17591 .11658 m .17318 .11287 L .17699 .114 L closepath p 1 1 .915 r
F P
0 g
s
.17376 .11211 m .17699 .114 L .17318 .11287 L closepath p 1 1 1 r
F P
0 g
s
.88082 .08112 m .88131 .08593 L .88024 .07613 L closepath p .955 .785 .905 r
F P
0 g
s
.58487 .09872 m .57295 .07614 L .58629 .09252 L closepath p .955 .785 .905 r
F P
0 g
s
.5787 .07874 m .58629 .09252 L .57295 .07614 L closepath p .955 .785 .905 r
F P
0 g
s
.9566 .07644 m .96146 .07736 L .95251 .08706 L closepath p .955 .785 .905 r
F P
0 g
s
.65738 .10006 m .65645 .10076 L .6407 .0766 L closepath p .955 .785 .905 r
F P
0 g
s
.6457 .08665 m .6407 .0766 L .65645 .10076 L closepath p .955 .785 .905 r
F P
0 g
s
.92714 .10006 m .92627 .10076 L .90986 .0766 L closepath p .955 .785 .905 r
F P
0 g
s
.91509 .08669 m .90986 .0766 L .92627 .10076 L closepath p .955 .785 .905 r
F P
0 g
s
.92714 .10006 m .90986 .0766 L .9309 .07707 L closepath p .955 .785 .905 r
F P
0 g
s
.14149 .1119 m .14023 .11483 L .12864 .08605 L closepath p .178 .342 .856 r
F P
0 g
s
.14023 .11483 m .14149 .1119 L .14194 .11691 L closepath p .327 .528 .988 r
F P
0 g
s
.21986 .07707 m .21878 .08287 L .21057 .10037 L closepath p .955 .785 .905 r
F P
0 g
s
.66157 .07707 m .66143 .07797 L .65948 .08855 L closepath p .955 .785 .905 r
F P
0 g
s
.9309 .07707 m .93078 .07797 L .92902 .08855 L closepath p .955 .785 .905 r
F P
0 g
s
.20475 .0811 m .20563 .08811 L .20269 .07715 L closepath p .955 .785 .905 r
F P
0 g
s
.64903 .08832 m .64486 .08246 L .6444 .07715 L closepath p .955 .785 .905 r
F P
0 g
s
.6444 .07715 m .64652 .0811 L .64903 .08832 L closepath p .955 .785 .905 r
F P
0 g
s
.08465 .11711 m .0825 .11275 L .08608 .11514 L closepath p 1 1 1 r
F P
0 g
s
.08322 .11188 m .08608 .11514 L .0825 .11275 L closepath p 1 1 1 r
F P
0 g
s
.87872 .07734 m .87958 .0802 L .87562 .08121 L closepath p .955 .785 .905 r
F P
0 g
s
.96043 .08562 m .95251 .08706 L .96146 .07736 L closepath p .955 .785 .905 r
F P
0 g
s
.96146 .07736 m .96553 .08012 L .96043 .08562 L closepath p .955 .785 .905 r
F P
0 g
s
.3477 .08402 m .34679 .08163 L .34676 .07753 L closepath p .955 .785 .905 r
F P
0 g
s
.3457 .07777 m .34679 .08163 L .34764 .09983 L closepath p .954 .784 .905 r
F P
0 g
s
.66002 .08616 m .65948 .08855 L .66143 .07797 L closepath p .955 .785 .905 r
F P
0 g
s
.66143 .07797 m .6611 .0802 L .66002 .08616 L closepath p .955 .785 .905 r
F P
0 g
s
.92952 .08616 m .92902 .08855 L .93078 .07797 L closepath p .955 .785 .905 r
F P
0 g
s
.93078 .07797 m .93049 .0802 L .92952 .08616 L closepath p .955 .785 .905 r
F P
0 g
s
.79036 .07799 m .79416 .08146 L .79096 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.05236 .1161 m .05299 .11282 L .05543 .11784 L closepath p .522 .764 1 r
F P
0 g
s
.08214 .11319 m .08177 .11362 L .08989 .07827 L closepath p .955 .785 .905 r
F P
0 g
s
.08989 .07827 m .08712 .09171 L .08214 .11319 L closepath p .955 .785 .905 r
F P
0 g
s
.97121 .08317 m .96095 .08461 L .96762 .07828 L closepath p .955 .785 .905 r
F P
0 g
s
.17337 .11784 m .17259 .11362 L .17591 .11658 L closepath p 1 1 1 r
F P
0 g
s
.17318 .11287 m .17591 .11658 L .17259 .11362 L closepath p 1 1 1 r
F P
0 g
s
.40139 .08308 m .39975 .07883 L .40076 .0788 L closepath p .955 .785 .905 r
F P
0 g
s
.39975 .07883 m .40139 .08308 L .40075 .08932 L closepath p .955 .785 .905 r
F P
0 g
s
.14248 .11276 m .14194 .11691 L .14149 .1119 L closepath p .489 .674 1 r
F P
0 g
s
.14194 .11691 m .14248 .11276 L .14432 .11784 L closepath p .679 .863 1 r
F P
0 g
s
.08277 .11784 m .08177 .11362 L .08465 .11711 L closepath p .975 1 1 r
F P
0 g
s
.0825 .11275 m .08465 .11711 L .08177 .11362 L closepath p 1 1 1 r
F P
0 g
s
.76287 .07918 m .75996 .08349 L .76166 .08026 L closepath p .955 .785 .905 r
F P
0 g
s
.76287 .07918 m .7633 .07932 L .76249 .09793 L closepath p .955 .785 .905 r
F P
0 g
s
.76287 .07918 m .76268 .08855 L .75996 .08349 L closepath p .955 .785 .905 r
F P
0 g
s
.52025 .08565 m .51842 .07918 L .52138 .08728 L closepath p .955 .785 .905 r
F P
0 g
s
.52191 .08581 m .52138 .08728 L .51842 .07918 L closepath p .955 .785 .905 r
F P
0 g
s
.24715 .09774 m .24467 .09155 L .24053 .07921 L closepath p .955 .785 .905 r
F P
0 g
s
.2403 .07926 m .24053 .07921 L .24467 .09155 L closepath p .955 .785 .905 r
F P
0 g
s
.32021 .07921 m .32097 .07948 L .31896 .08028 L closepath p .955 .785 .905 r
F P
0 g
s
.24255 .08029 m .24715 .09774 L .24053 .07921 L closepath p .955 .785 .905 r
F P
0 g
s
.06828 .11726 m .06716 .11305 L .07154 .11718 L closepath p .844 1 1 r
F P
0 g
s
.07052 .11296 m .07154 .11718 L .06716 .11305 L closepath p .844 1 1 r
F P
0 g
s
.07154 .11718 m .07052 .11296 L .07694 .1175 L closepath p .807 .985 1 r
F P
0 g
s
.15144 .1175 m .15044 .11329 L .15822 .11718 L closepath p .854 1 1 r
F P
0 g
s
.1574 .11296 m .15822 .11718 L .15044 .11329 L closepath p .854 1 1 r
F P
0 g
s
.15822 .11718 m .1574 .11296 L .16127 .11726 L closepath p .822 .996 1 r
F P
0 g
s
.24467 .09155 m .24248 .08655 L .2403 .07926 L closepath p .955 .785 .905 r
F P
0 g
s
.24007 .07932 m .2403 .07926 L .24248 .08655 L closepath p .955 .785 .905 r
F P
0 g
s
.16055 .11305 m .16127 .11726 L .1574 .11296 L closepath p .822 .996 1 r
F P
0 g
s
.16127 .11726 m .16055 .11305 L .16564 .1175 L closepath p .809 .986 1 r
F P
0 g
s
.06364 .1175 m .06244 .11329 L .06828 .11726 L closepath p .857 1 1 r
F P
0 g
s
.06716 .11305 m .06828 .11726 L .06244 .11329 L closepath p .856 1 1 r
F P
0 g
s
.76326 .0905 m .76249 .09793 L .7633 .07932 L closepath p .955 .785 .905 r
F P
0 g
s
.7633 .07932 m .76373 .07946 L .76326 .0905 L closepath p .955 .785 .905 r
F P
0 g
s
.24248 .08655 m .24058 .08276 L .24007 .07932 L closepath p .955 .785 .905 r
F P
0 g
s
.23984 .07938 m .24007 .07932 L .24058 .08276 L closepath p .955 .785 .905 r
F P
0 g
s
.6904 .09778 m .68758 .09135 L .68336 .07934 L closepath p .955 .785 .905 r
F P
0 g
s
.68313 .0794 m .68336 .07934 L .68758 .09135 L closepath p .955 .785 .905 r
F P
0 g
s
.68541 .08042 m .6904 .09778 L .68336 .07934 L closepath p .955 .785 .905 r
F P
0 g
s
.24058 .08276 m .23897 .08015 L .23984 .07938 L closepath p .955 .785 .905 r
F P
0 g
s
.23961 .07944 m .23984 .07938 L .23897 .08015 L closepath p .955 .785 .905 r
F P
0 g
s
.68758 .09135 m .68521 .0863 L .68313 .0794 L closepath p .955 .785 .905 r
F P
0 g
s
.6829 .07946 m .68313 .0794 L .68521 .0863 L closepath p .955 .785 .905 r
F P
0 g
s
.76483 .08028 m .76326 .0905 L .76373 .07946 L closepath p .955 .785 .905 r
F P
0 g
s
.68521 .0863 m .68329 .08261 L .6829 .07946 L closepath p .955 .785 .905 r
F P
0 g
s
.68268 .07952 m .6829 .07946 L .68329 .08261 L closepath p .955 .785 .905 r
F P
0 g
s
.31716 .08351 m .31896 .08028 L .32097 .07948 L closepath p .955 .785 .905 r
F P
0 g
s
.32097 .07948 m .32197 .0803 L .31882 .09101 L closepath p .955 .785 .905 r
F P
0 g
s
.32097 .07948 m .31882 .09101 L .31716 .08351 L closepath p .955 .785 .905 r
F P
0 g
s
.68329 .08261 m .68183 .08028 L .68268 .07952 L closepath p .955 .785 .905 r
F P
0 g
s
.68245 .07958 m .68268 .07952 L .68183 .08028 L closepath p .955 .785 .905 r
F P
0 g
s
.07615 .11329 m .07694 .1175 L .07052 .11296 L closepath p .807 .986 1 r
F P
0 g
s
.07694 .1175 m .07615 .11329 L .08277 .11784 L closepath p .807 .986 1 r
F P
0 g
s
.14432 .11784 m .14347 .11362 L .15144 .1175 L closepath p .854 1 1 r
F P
0 g
s
.15044 .11329 m .15144 .1175 L .14347 .11362 L closepath p .854 1 1 r
F P
0 g
s
.165 .11329 m .16564 .1175 L .16055 .11305 L closepath p .809 .986 1 r
F P
0 g
s
.16564 .1175 m .165 .11329 L .17014 .11775 L closepath p .809 .986 1 r
F P
0 g
s
.05886 .11775 m .05772 .11354 L .06364 .1175 L closepath p .856 1 1 r
F P
0 g
s
.06244 .11329 m .06364 .1175 L .05772 .11354 L closepath p .856 1 1 r
F P
0 g
s
.06751 .08201 m .06797 .09684 L .06676 .07962 L closepath p .955 .785 .905 r
F P
0 g
s
.52295 .08287 m .52393 .07965 L .52086 .08875 L closepath p .955 .785 .905 r
F P
0 g
s
.52295 .08287 m .52086 .08875 L .52393 .07965 L closepath p .955 .785 .905 r
F P
0 g
s
.16944 .11354 m .17014 .11775 L .165 .11329 L closepath p .809 .986 1 r
F P
0 g
s
.17014 .11775 m .16944 .11354 L .17337 .11784 L closepath p .822 .996 1 r
F P
0 g
s
.05543 .11784 m .05437 .11362 L .05886 .11775 L closepath p .844 1 1 r
F P
0 g
s
.05772 .11354 m .05886 .11775 L .05437 .11362 L closepath p .844 1 1 r
F P
0 g
s
.49145 .07981 m .49331 .08091 L .48991 .08089 L closepath p .955 .785 .905 r
F P
0 g
s
.05437 .11362 m .05543 .11784 L .05299 .11282 L closepath p .508 .754 1 r
F P
0 g
s
.08177 .11362 m .08277 .11784 L .07615 .11329 L closepath p .807 .986 1 r
F P
0 g
s
.14347 .11362 m .14432 .11784 L .14248 .11276 L closepath p .346 .605 1 r
F P
0 g
s
.17259 .11362 m .17337 .11784 L .16944 .11354 L closepath p .822 .996 1 r
F P
0 g
s
.96834 .08418 m .96043 .08562 L .96553 .08012 L closepath p .955 .785 .905 r
F P
0 g
s
.87958 .0802 m .88045 .08306 L .87441 .08581 L closepath p .955 .785 .905 r
F P
0 g
s
.87441 .08581 m .87562 .08121 L .87958 .0802 L closepath p .955 .785 .905 r
F P
0 g
s
.66057 .08377 m .66002 .08616 L .6611 .0802 L closepath p .955 .785 .905 r
F P
0 g
s
.93003 .08377 m .92952 .08616 L .93049 .0802 L closepath p .955 .785 .905 r
F P
0 g
s
.93003 .08377 m .93049 .0802 L .92952 .08616 L closepath p .955 .785 .905 r
F P
0 g
s
.24571 .08355 m .24715 .09774 L .24255 .08029 L closepath p .955 .785 .905 r
F P
0 g
s
.32008 .0905 m .31882 .09101 L .32197 .0803 L closepath p .955 .785 .905 r
F P
0 g
s
.53338 .09945 m .53308 .09983 L .53682 .08036 L closepath p .955 .785 .905 r
F P
0 g
s
.53682 .08036 m .53548 .08879 L .53338 .09945 L closepath p .955 .785 .905 r
F P
0 g
s
.53682 .08036 m .53338 .09945 L .53548 .08879 L closepath p .955 .785 .905 r
F P
0 g
s
.68867 .08365 m .6904 .09778 L .68541 .08042 L closepath p .955 .785 .905 r
F P
0 g
s
.8267 .08963 m .81839 .08044 L .83395 .0854 L closepath p .955 .785 .905 r
F P
0 g
s
.95697 .08796 m .96101 .08044 L .95924 .08796 L closepath p .955 .785 .905 r
F P
0 g
s
.96417 .08044 m .95924 .08796 L .96101 .08044 L closepath p .955 .785 .905 r
F P
0 g
s
.59709 .0824 m .59769 .08753 L .58981 .08059 L closepath p .955 .785 .905 r
F P
0 g
s
.58734 .08965 m .58981 .08059 L .59769 .08753 L closepath p .955 .785 .905 r
F P
0 g
s
.15514 .08333 m .15597 .09281 L .1509 .0807 L closepath p .955 .785 .905 r
F P
0 g
s
.14777 .09109 m .1509 .0807 L .15597 .09281 L closepath p .955 .785 .905 r
F P
0 g
s
.4874 .08413 m .48991 .08089 L .49331 .08091 L closepath p .955 .785 .905 r
F P
0 g
s
.49239 .088 m .49331 .08091 L .49302 .08397 L closepath p .956 .786 .905 r
F P
0 g
s
.49197 .0901 m .49331 .08091 L .49239 .088 L closepath p .952 .782 .905 r
F P
0 g
s
.49331 .08091 m .49197 .0901 L .4874 .08413 L closepath p .955 .785 .905 r
F P
0 g
s
.20563 .08811 m .20475 .0811 L .21057 .10037 L closepath p .955 .785 .905 r
F P
0 g
s
.20614 .08334 m .21057 .10037 L .20475 .0811 L closepath p .955 .785 .905 r
F P
0 g
s
.6532 .09418 m .64903 .08832 L .64652 .0811 L closepath p .955 .785 .905 r
F P
0 g
s
.64652 .0811 m .64789 .08334 L .6532 .09418 L closepath p .955 .785 .905 r
F P
0 g
s
.79532 .08749 m .79096 .09974 L .79416 .08146 L closepath p .955 .785 .905 r
F P
0 g
s
.34795 .0937 m .34764 .09983 L .34679 .08163 L closepath p .955 .785 .905 r
F P
0 g
s
.34679 .08163 m .3477 .08402 L .34795 .0937 L closepath p .955 .785 .905 r
F P
0 g
s
.06751 .08201 m .06827 .08438 L .06797 .09684 L closepath p .955 .785 .905 r
F P
0 g
s
.07133 .09349 m .07139 .09335 L .07664 .08259 L closepath p .955 .785 .905 r
F P
0 g
s
.07664 .08259 m .06855 .10039 L .07133 .09349 L closepath p .955 .785 .905 r
F P
0 g
s
.39539 .09683 m .38903 .09383 L .38948 .08272 L closepath p .955 .785 .905 r
F P
0 g
s
.38984 .09064 m .38948 .08272 L .38903 .09383 L closepath p .955 .785 .905 r
F P
0 g
s
.38984 .09064 m .38903 .09383 L .38948 .08272 L closepath p .955 .785 .905 r
F P
0 g
s
.21769 .08867 m .21057 .10037 L .21878 .08287 L closepath p .955 .785 .905 r
F P
0 g
s
.88045 .08306 m .88131 .08593 L .87328 .08921 L closepath p .955 .785 .905 r
F P
0 g
s
.87328 .08921 m .87441 .08581 L .88045 .08306 L closepath p .955 .785 .905 r
F P
0 g
s
.40183 .08601 m .40075 .08932 L .40139 .08308 L closepath p .955 .785 .905 r
F P
0 g
s
.94237 .08327 m .9461 .08465 L .94138 .08898 L closepath p .955 .785 .905 r
F P
0 g
s
.65738 .10006 m .64789 .08334 L .64821 .08334 L closepath p .955 .785 .905 r
F P
0 g
s
.65738 .10006 m .6532 .09418 L .64789 .08334 L closepath p .955 .785 .905 r
F P
0 g
s
.75747 .08822 m .75996 .08349 L .76268 .08855 L closepath p .955 .785 .905 r
F P
0 g
s
.31447 .08823 m .31716 .08351 L .31882 .09101 L closepath p .955 .785 .905 r
F P
0 g
s
.24715 .09774 m .24571 .08355 L .2535 .09154 L closepath p .955 .785 .905 r
F P
0 g
s
.25015 .08843 m .2535 .09154 L .24571 .08355 L closepath p .955 .785 .905 r
F P
0 g
s
.25015 .08843 m .24571 .08355 L .2535 .09154 L closepath p .955 .785 .905 r
F P
0 g
s
.6904 .09778 m .68867 .08365 L .69671 .09157 L closepath p .955 .785 .905 r
F P
0 g
s
.69327 .08849 m .69671 .09157 L .68867 .08365 L closepath p .955 .785 .905 r
F P
0 g
s
.69327 .08849 m .68867 .08365 L .69671 .09157 L closepath p .955 .785 .905 r
F P
0 g
s
.66002 .08616 m .66057 .08377 L .6587 .09501 L closepath p .955 .785 .905 r
F P
0 g
s
.65984 .08867 m .6587 .09501 L .66057 .08377 L closepath p .955 .785 .905 r
F P
0 g
s
.92952 .08616 m .93003 .08377 L .92837 .09501 L closepath p .955 .785 .905 r
F P
0 g
s
.92939 .08867 m .92837 .09501 L .93003 .08377 L closepath p .955 .785 .905 r
F P
0 g
s
.49302 .08397 m .49213 .09049 L .49239 .088 L closepath p .958 .788 .905 r
F P
0 g
s
.34827 .08757 m .34795 .0937 L .3477 .08402 L closepath p .955 .785 .905 r
F P
0 g
s
.48386 .0889 m .4874 .08413 L .49197 .0901 L closepath p .955 .785 .905 r
F P
0 g
s
.06902 .08673 m .06797 .09684 L .06827 .08438 L closepath p .955 .785 .905 r
F P
0 g
s
.9427 .09472 m .94138 .08898 L .9461 .08465 L closepath p .955 .785 .905 r
F P
0 g
s
.45049 .09276 m .43672 .09057 L .44856 .08539 L closepath p .955 .785 .905 r
F P
0 g
s
.83395 .0854 m .84123 .09236 L .8267 .08963 L closepath p .955 .785 .905 r
F P
0 g
s
.52138 .08728 m .52086 .08875 L .52025 .08565 L closepath p .955 .785 .905 r
F P
0 g
s
.87223 .09142 m .87328 .08921 L .88131 .08593 L closepath p .955 .785 .905 r
F P
0 g
s
.88131 .08593 m .88163 .08793 L .87812 .09652 L closepath p .955 .785 .905 r
F P
0 g
s
.88131 .08593 m .87812 .09652 L .87223 .09142 L closepath p .955 .785 .905 r
F P
0 g
s
.40075 .08932 m .40183 .08601 L .40174 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.40208 .08759 m .40174 .09983 L .40183 .08601 L closepath p .955 .785 .905 r
F P
0 g
s
.12864 .08605 m .14306 .10669 L .14149 .1119 L closepath p .955 .785 .905 r
F P
0 g
s
.65948 .08855 m .66002 .08616 L .65738 .10006 L closepath p .955 .785 .905 r
F P
0 g
s
.6587 .09501 m .65738 .10006 L .66002 .08616 L closepath p .955 .785 .905 r
F P
0 g
s
.92902 .08855 m .92952 .08616 L .92714 .10006 L closepath p .955 .785 .905 r
F P
0 g
s
.92837 .09501 m .92714 .10006 L .92952 .08616 L closepath p .955 .785 .905 r
F P
0 g
s
.65645 .10076 m .65553 .10147 L .6457 .08665 L closepath p .955 .785 .905 r
F P
0 g
s
.65309 .10037 m .6457 .08665 L .65553 .10147 L closepath p .955 .785 .905 r
F P
0 g
s
.92627 .10076 m .9254 .10147 L .91509 .08669 L closepath p .955 .785 .905 r
F P
0 g
s
.92286 .10037 m .91509 .08669 L .9254 .10147 L closepath p .955 .785 .905 r
F P
0 g
s
.06902 .08673 m .06977 .08905 L .06797 .09684 L closepath p .955 .785 .905 r
F P
0 g
s
.79096 .09974 m .78308 .09055 L .78731 .08686 L closepath p .955 .785 .905 r
F P
0 g
s
.78286 .08812 m .78731 .08686 L .78308 .09055 L closepath p .955 .785 .905 r
F P
0 g
s
.94842 .0977 m .94559 .09371 L .95251 .08706 L closepath p .955 .785 .905 r
F P
0 g
s
.28773 .08749 m .28924 .09379 L .27459 .09513 L closepath p .955 .785 .905 r
F P
0 g
s
.27608 .09212 m .28773 .08749 L .27459 .09513 L closepath p .955 .785 .905 r
F P
0 g
s
.71778 .09513 m .71915 .09212 L .73065 .08749 L closepath p .955 .785 .905 r
F P
0 g
s
.73065 .08749 m .73239 .09379 L .71778 .09513 L closepath p .955 .785 .905 r
F P
0 g
s
.79532 .08749 m .79853 .09403 L .79096 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.59769 .08753 m .59792 .08946 L .58734 .08965 L closepath p .955 .785 .905 r
F P
0 g
s
.34827 .08757 m .35131 .09412 L .34764 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.40208 .08759 m .40513 .09413 L .40174 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.8793 .09691 m .87812 .09652 L .88163 .08793 L closepath p .955 .785 .905 r
F P
0 g
s
.88163 .08793 m .88236 .09157 L .8793 .09691 L closepath p .955 .785 .905 r
F P
0 g
s
.95924 .08796 m .96293 .08938 L .95534 .09271 L closepath p .955 .785 .905 r
F P
0 g
s
.49117 .09682 m .49239 .088 L .49213 .09049 L closepath p .957 .787 .905 r
F P
0 g
s
.49239 .088 m .49117 .09682 L .49197 .0901 L closepath p .953 .783 .905 r
F P
0 g
s
.75484 .09157 m .75747 .08822 L .76249 .09793 L closepath p .955 .785 .905 r
F P
0 g
s
.76268 .08855 m .76249 .09793 L .75747 .08822 L closepath p .955 .785 .905 r
F P
0 g
s
.31171 .09158 m .31447 .08823 L .31667 .10255 L closepath p .955 .785 .905 r
F P
0 g
s
.31882 .09101 m .31667 .10255 L .31447 .08823 L closepath p .955 .785 .905 r
F P
0 g
s
.21769 .08867 m .21635 .09501 L .21057 .10037 L closepath p .955 .785 .905 r
F P
0 g
s
.52491 .0998 m .51872 .09484 L .52086 .08875 L closepath p .955 .785 .905 r
F P
0 g
s
.51991 .09334 m .52086 .08875 L .51872 .09484 L closepath p .955 .785 .905 r
F P
0 g
s
.51991 .09334 m .51872 .09484 L .52086 .08875 L closepath p .955 .785 .905 r
F P
0 g
s
.53368 .09906 m .53338 .09945 L .53548 .08879 L closepath p .955 .785 .905 r
F P
0 g
s
.53548 .08879 m .53479 .09409 L .53368 .09906 L closepath p .955 .785 .905 r
F P
0 g
s
.481 .09207 m .48386 .0889 L .49063 .0993 L closepath p .955 .785 .905 r
F P
0 g
s
.49197 .0901 m .49063 .0993 L .48386 .0889 L closepath p .955 .785 .905 r
F P
0 g
s
.14173 .10426 m .14041 .10183 L .13532 .08891 L closepath p .955 .785 .905 r
F P
0 g
s
.13532 .08891 m .14346 .10469 L .14173 .10426 L closepath p .955 .785 .905 r
F P
0 g
s
.06977 .08905 m .07052 .09134 L .06855 .10039 L closepath p .955 .785 .905 r
F P
0 g
s
.06855 .10039 m .06797 .09684 L .06977 .08905 L closepath p .955 .785 .905 r
F P
0 g
s
.96293 .08938 m .96424 .09252 L .95964 .09404 L closepath p .955 .785 .905 r
F P
0 g
s
.59792 .08946 m .5989 .09241 L .58487 .09872 L closepath p .955 .785 .905 r
F P
0 g
s
.58487 .09872 m .58734 .08965 L .59792 .08946 L closepath p .955 .785 .905 r
F P
0 g
s
.83563 .09621 m .8267 .08963 L .84123 .09236 L closepath p .955 .785 .905 r
F P
0 g
s
.49063 .0993 m .49197 .0901 L .49117 .09682 L closepath p .955 .785 .905 r
F P
0 g
s
.31882 .09101 m .32008 .0905 L .31667 .10255 L closepath p .955 .785 .905 r
F P
0 g
s
.31908 .09793 m .31667 .10255 L .32008 .0905 L closepath p .955 .785 .905 r
F P
0 g
s
.79096 .09974 m .77542 .09621 L .78308 .09055 L closepath p .955 .785 .905 r
F P
0 g
s
.78122 .09475 m .78308 .09055 L .77542 .09621 L closepath p .955 .785 .905 r
F P
0 g
s
.78122 .09475 m .77542 .09621 L .78308 .09055 L closepath p .955 .785 .905 r
F P
0 g
s
.45049 .09276 m .45198 .09414 L .43672 .09057 L closepath p .955 .785 .905 r
F P
0 g
s
.43593 .09379 m .43672 .09057 L .45198 .09414 L closepath p .955 .785 .905 r
F P
0 g
s
.33512 .09384 m .33593 .09066 L .34764 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.15597 .09281 m .15651 .09875 L .14777 .09109 L closepath p .955 .785 .905 r
F P
0 g
s
.14463 .10148 m .14777 .09109 L .15651 .09875 L closepath p .955 .785 .905 r
F P
0 g
s
.07128 .09361 m .06855 .10039 L .07052 .09134 L closepath p .955 .785 .905 r
F P
0 g
s
.86647 .0963 m .87223 .09142 L .87812 .09652 L closepath p .955 .785 .905 r
F P
0 g
s
.25899 .0948 m .24715 .09774 L .2535 .09154 L closepath p .955 .785 .905 r
F P
0 g
s
.75484 .09157 m .76249 .09793 L .7498 .09481 L closepath p .955 .785 .905 r
F P
0 g
s
.6904 .09778 m .69671 .09157 L .69944 .09684 L closepath p .955 .785 .905 r
F P
0 g
s
.7023 .09481 m .69944 .09684 L .69671 .09157 L closepath p .955 .785 .905 r
F P
0 g
s
.88236 .09157 m .88287 .09473 L .88049 .09731 L closepath p .955 .785 .905 r
F P
0 g
s
.88049 .09731 m .8793 .09691 L .88236 .09157 L closepath p .955 .785 .905 r
F P
0 g
s
.31171 .09158 m .31667 .10255 L .30982 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.31171 .09158 m .30982 .09975 L .30651 .09481 L closepath p .955 .785 .905 r
F P
0 g
s
.0825 .11275 m .08214 .11319 L .08712 .09171 L closepath p .955 .785 .905 r
F P
0 g
s
.08712 .09171 m .08526 .10117 L .0825 .11275 L closepath p .955 .785 .905 r
F P
0 g
s
.08712 .09171 m .0825 .11275 L .08526 .10117 L closepath p .955 .785 .905 r
F P
0 g
s
.481 .09207 m .49063 .0993 L .47646 .09531 L closepath p .955 .785 .905 r
F P
0 g
s
.84123 .09236 m .849 .09654 L .83563 .09621 L closepath p .955 .785 .905 r
F P
0 g
s
.5989 .09241 m .60049 .09452 L .60146 .0998 L closepath p .955 .785 .905 r
F P
0 g
s
.60146 .0998 m .58487 .09872 L .5989 .09241 L closepath p .955 .785 .905 r
F P
0 g
s
.07128 .09361 m .07133 .09349 L .06855 .10039 L closepath p .955 .785 .905 r
F P
0 g
s
.2922 .09744 m .27459 .09513 L .28924 .09379 L closepath p .955 .785 .905 r
F P
0 g
s
.28924 .09379 m .29158 .09589 L .2922 .09744 L closepath p .955 .785 .905 r
F P
0 g
s
.71266 .09683 m .73239 .09379 L .73492 .09589 L closepath p .955 .785 .905 r
F P
0 g
s
.71602 .09588 m .71778 .09513 L .73239 .09379 L closepath p .955 .785 .905 r
F P
0 g
s
.71266 .09683 m .71602 .09588 L .73239 .09379 L closepath p .955 .785 .905 r
F P
0 g
s
.45198 .09414 m .45346 .09553 L .43593 .09379 L closepath p .955 .785 .905 r
F P
0 g
s
.43362 .0956 m .43593 .09379 L .45346 .09553 L closepath p .955 .785 .905 r
F P
0 g
s
.40174 .09983 m .38665 .09568 L .38903 .09383 L closepath p .955 .785 .905 r
F P
0 g
s
.38803 .09492 m .38903 .09383 L .38665 .09568 L closepath p .955 .785 .905 r
F P
0 g
s
.38803 .09492 m .38665 .09568 L .38903 .09383 L closepath p .955 .785 .905 r
F P
0 g
s
.34764 .09983 m .33278 .09569 L .33512 .09384 L closepath p .955 .785 .905 r
F P
0 g
s
.33414 .09493 m .33512 .09384 L .33278 .09569 L closepath p .955 .785 .905 r
F P
0 g
s
.33414 .09493 m .33278 .09569 L .33512 .09384 L closepath p .955 .785 .905 r
F P
0 g
s
.79999 .09801 m .79096 .09974 L .79853 .09403 L closepath p .955 .785 .905 r
F P
0 g
s
.79853 .09403 m .80457 .09621 L .79999 .09801 L closepath p .955 .785 .905 r
F P
0 g
s
.53397 .09868 m .53368 .09906 L .53479 .09409 L closepath p .955 .785 .905 r
F P
0 g
s
.35467 .09811 m .34764 .09983 L .35131 .09412 L closepath p .955 .785 .905 r
F P
0 g
s
.35131 .09412 m .35724 .09631 L .35467 .09811 L closepath p .955 .785 .905 r
F P
0 g
s
.40869 .09811 m .40174 .09983 L .40513 .09413 L closepath p .955 .785 .905 r
F P
0 g
s
.40513 .09413 m .41118 .09631 L .40869 .09811 L closepath p .955 .785 .905 r
F P
0 g
s
.60049 .09452 m .6055 .09621 L .60537 .09998 L closepath p .955 .785 .905 r
F P
0 g
s
.60537 .09998 m .60146 .0998 L .60049 .09452 L closepath p .955 .785 .905 r
F P
0 g
s
.88308 .09715 m .88049 .09731 L .88287 .09473 L closepath p .955 .785 .905 r
F P
0 g
s
.9716 .09477 m .96836 .09958 L .9618 .09933 L closepath p .955 .785 .905 r
F P
0 g
s
.2646 .09589 m .24715 .09774 L .25899 .0948 L closepath p .955 .785 .905 r
F P
0 g
s
.74342 .09589 m .7498 .09481 L .76249 .09793 L closepath p .955 .785 .905 r
F P
0 g
s
.69944 .09684 m .7023 .09481 L .70849 .0959 L closepath p .955 .785 .905 r
F P
0 g
s
.70795 .09589 m .70849 .0959 L .7023 .09481 L closepath p .955 .785 .905 r
F P
0 g
s
.70795 .09589 m .7023 .09481 L .70849 .0959 L closepath p .955 .785 .905 r
F P
0 g
s
.30008 .09589 m .30651 .09481 L .30982 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.52491 .0998 m .52282 .09974 L .51872 .09484 L closepath p .955 .785 .905 r
F P
0 g
s
.51706 .09581 m .51872 .09484 L .52282 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.21181 .10092 m .21057 .10037 L .21635 .09501 L closepath p .955 .785 .905 r
F P
0 g
s
.21635 .09501 m .21487 .10006 L .21181 .10092 L closepath p .955 .785 .905 r
F P
0 g
s
.27459 .09513 m .30982 .09975 L .27278 .09588 L closepath p .955 .785 .905 r
F P
0 g
s
.47413 .09612 m .47646 .09531 L .49063 .0993 L closepath p .955 .785 .905 r
F P
0 g
s
.45346 .09553 m .46132 .09618 L .43362 .0956 L closepath p .955 .785 .905 r
F P
0 g
s
.43362 .0956 m .49063 .0993 L .43145 .09603 L closepath p .955 .785 .905 r
F P
0 g
s
.46212 .09745 m .43362 .0956 L .46132 .09618 L closepath p .955 .785 .905 r
F P
0 g
s
.40174 .09983 m .38937 .0987 L .38665 .09568 L closepath p .955 .785 .905 r
F P
0 g
s
.38419 .09615 m .38665 .09568 L .38937 .0987 L closepath p .955 .785 .905 r
F P
0 g
s
.34764 .09983 m .33537 .0987 L .33278 .09569 L closepath p .955 .785 .905 r
F P
0 g
s
.3303 .09616 m .33278 .09569 L .33537 .0987 L closepath p .955 .785 .905 r
F P
0 g
s
.51706 .09581 m .50765 .09939 L .51405 .09623 L closepath p .955 .785 .905 r
F P
0 g
s
.52282 .09974 m .51577 .09998 L .51706 .09581 L closepath p .955 .785 .905 r
F P
0 g
s
.50765 .09939 m .51706 .09581 L .51577 .09998 L closepath p .955 .785 .905 r
F P
0 g
s
.30982 .09975 m .26743 .09601 L .27278 .09588 L closepath p .955 .785 .905 r
F P
0 g
s
.27026 .09613 m .27278 .09588 L .26743 .09601 L closepath p .955 .785 .905 r
F P
0 g
s
.27026 .09613 m .26743 .09601 L .27278 .09588 L closepath p .955 .785 .905 r
F P
0 g
s
.71266 .09683 m .71354 .09613 L .71602 .09588 L closepath p .955 .785 .905 r
F P
0 g
s
.71487 .09606 m .71602 .09588 L .71354 .09613 L closepath p .955 .785 .905 r
F P
0 g
s
.71487 .09606 m .71354 .09613 L .71602 .09588 L closepath p .955 .785 .905 r
F P
0 g
s
.2646 .09589 m .26514 .0959 L .24715 .09774 L closepath p .955 .785 .905 r
F P
0 g
s
.29158 .09589 m .29583 .09589 L .30982 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.30008 .09589 m .30982 .09975 L .29583 .09589 L closepath p .955 .785 .905 r
F P
0 g
s
.30982 .09975 m .2922 .09744 L .29158 .09589 L closepath p .955 .785 .905 r
F P
0 g
s
.73492 .09589 m .73917 .09589 L .76249 .09793 L closepath p .955 .785 .905 r
F P
0 g
s
.74342 .09589 m .76249 .09793 L .73917 .09589 L closepath p .955 .785 .905 r
F P
0 g
s
.6904 .09778 m .73492 .09589 L .76249 .09793 L closepath p .955 .785 .905 r
F P
0 g
s
.30982 .09975 m .24715 .09774 L .26514 .0959 L closepath p .955 .785 .905 r
F P
0 g
s
.30982 .09975 m .26514 .0959 L .26743 .09601 L closepath p .955 .785 .905 r
F P
0 g
s
.26579 .09592 m .26743 .09601 L .26514 .0959 L closepath p .955 .785 .905 r
F P
0 g
s
.6904 .09778 m .70849 .0959 L .71074 .09601 L closepath p .955 .785 .905 r
F P
0 g
s
.70913 .09592 m .71074 .09601 L .70849 .0959 L closepath p .955 .785 .905 r
F P
0 g
s
.71354 .09613 m .6904 .09778 L .71074 .09601 L closepath p .955 .785 .905 r
F P
0 g
s
.49063 .0993 m .44535 .09975 L .43145 .09603 L closepath p .955 .785 .905 r
F P
0 g
s
.42787 .09623 m .43145 .09603 L .44535 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.60969 .09617 m .61028 .09605 L .61216 .09865 L closepath p .955 .785 .905 r
F P
0 g
s
.61223 .09709 m .61216 .09865 L .61028 .09605 L closepath p .955 .785 .905 r
F P
0 g
s
.49063 .0993 m .47176 .09639 L .47413 .09612 L closepath p .955 .785 .905 r
F P
0 g
s
.71266 .09683 m .6904 .09778 L .71354 .09613 L closepath p .955 .785 .905 r
F P
0 g
s
.70197 .09695 m .71354 .09613 L .6904 .09778 L closepath p .955 .785 .905 r
F P
0 g
s
.38937 .0987 m .37699 .09757 L .38419 .09615 L closepath p .955 .785 .905 r
F P
0 g
s
.38024 .09631 m .38419 .09615 L .37699 .09757 L closepath p .955 .785 .905 r
F P
0 g
s
.33537 .0987 m .3231 .09757 L .3303 .09616 L closepath p .955 .785 .905 r
F P
0 g
s
.32628 .09631 m .3303 .09616 L .3231 .09757 L closepath p .955 .785 .905 r
F P
0 g
s
.6055 .09621 m .60969 .09617 L .61209 .10021 L closepath p .955 .785 .905 r
F P
0 g
s
.61216 .09865 m .61209 .10021 L .60969 .09617 L closepath p .955 .785 .905 r
F P
0 g
s
.46132 .09618 m .47176 .09639 L .46212 .09745 L closepath p .955 .785 .905 r
F P
0 g
s
.84517 .10015 m .83563 .09621 L .849 .09654 L closepath p .955 .785 .905 r
F P
0 g
s
.61209 .10021 m .60537 .09998 L .6055 .09621 L closepath p .955 .785 .905 r
F P
0 g
s
.77113 .09652 m .77355 .09621 L .77288 .09684 L closepath p .955 .785 .905 r
F P
0 g
s
.77542 .09621 m .77288 .09684 L .77355 .09621 L closepath p .955 .785 .905 r
F P
0 g
s
.80903 .09629 m .80457 .09621 L .80775 .09623 L closepath p .955 .785 .905 r
F P
0 g
s
.79096 .09974 m .78699 .0998 L .77542 .09621 L closepath p .955 .785 .905 r
F P
0 g
s
.77288 .09684 m .77542 .09621 L .78699 .0998 L closepath p .955 .785 .905 r
F P
0 g
s
.80903 .09629 m .79999 .09801 L .80457 .09621 L closepath p .955 .785 .905 r
F P
0 g
s
.51405 .09623 m .50765 .09939 L .50824 .09636 L closepath p .955 .785 .905 r
F P
0 g
s
.44535 .09975 m .44132 .09983 L .42787 .09623 L closepath p .955 .785 .905 r
F P
0 g
s
.42556 .09655 m .42787 .09623 L .44132 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.80903 .09629 m .81137 .09943 L .79096 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.86229 .09753 m .86647 .0963 L .8776 .0966 L closepath p .955 .785 .905 r
F P
0 g
s
.87812 .09652 m .8776 .0966 L .86647 .0963 L closepath p .955 .785 .905 r
F P
0 g
s
.38024 .09631 m .37699 .09757 L .37781 .09662 L closepath p .955 .785 .905 r
F P
0 g
s
.3617 .09639 m .35724 .09631 L .36042 .09633 L closepath p .955 .785 .905 r
F P
0 g
s
.3617 .09639 m .35467 .09811 L .35724 .09631 L closepath p .955 .785 .905 r
F P
0 g
s
.32628 .09631 m .3231 .09757 L .3239 .09663 L closepath p .955 .785 .905 r
F P
0 g
s
.41563 .09639 m .41118 .09631 L .41436 .09633 L closepath p .955 .785 .905 r
F P
0 g
s
.41563 .09639 m .40869 .09811 L .41118 .09631 L closepath p .955 .785 .905 r
F P
0 g
s
.3617 .09639 m .36322 .09751 L .35467 .09811 L closepath p .955 .785 .905 r
F P
0 g
s
.49063 .0993 m .46212 .09745 L .47176 .09639 L closepath p .955 .785 .905 r
F P
0 g
s
.41563 .09639 m .41717 .09751 L .40869 .09811 L closepath p .955 .785 .905 r
F P
0 g
s
.77288 .09684 m .77034 .09746 L .77113 .09652 L closepath p .955 .785 .905 r
F P
0 g
s
.849 .09654 m .85726 .09793 L .84517 .10015 L closepath p .955 .785 .905 r
F P
0 g
s
.44132 .09983 m .4357 .10004 L .42556 .09655 L closepath p .955 .785 .905 r
F P
0 g
s
.42478 .09751 m .42556 .09655 L .4357 .10004 L closepath p .955 .785 .905 r
F P
0 g
s
.8776 .0966 m .87687 .09683 L .85532 .10147 L closepath p .955 .785 .905 r
F P
0 g
s
.8776 .0966 m .85532 .10147 L .86229 .09753 L closepath p .955 .785 .905 r
F P
0 g
s
.87687 .09683 m .87475 .09778 L .87064 .0989 L closepath p .955 .785 .905 r
F P
0 g
s
.8644 .10096 m .85532 .10147 L .87687 .09683 L closepath p .955 .785 .905 r
F P
0 g
s
.78699 .0998 m .7814 .09998 L .77288 .09684 L closepath p .955 .785 .905 r
F P
0 g
s
.77034 .09746 m .77288 .09684 L .7814 .09998 L closepath p .955 .785 .905 r
F P
0 g
s
.88308 .09715 m .88279 .09856 L .88049 .09731 L closepath p .955 .785 .905 r
F P
0 g
s
.88209 .09821 m .88049 .09731 L .88279 .09856 L closepath p .955 .785 .905 r
F P
0 g
s
.7814 .09998 m .77861 .10007 L .77034 .09746 L closepath p .955 .785 .905 r
F P
0 g
s
.77074 .09953 m .77034 .09746 L .77861 .10007 L closepath p .955 .785 .905 r
F P
0 g
s
.35682 .09918 m .35467 .09811 L .36322 .09751 L closepath p .955 .785 .905 r
F P
0 g
s
.36322 .09751 m .36303 .10033 L .35682 .09918 L closepath p .955 .785 .905 r
F P
0 g
s
.41082 .09918 m .40869 .09811 L .41717 .09751 L closepath p .955 .785 .905 r
F P
0 g
s
.41717 .09751 m .41699 .10033 L .41082 .09918 L closepath p .955 .785 .905 r
F P
0 g
s
.4357 .10004 m .4329 .10015 L .42478 .09751 L closepath p .955 .785 .905 r
F P
0 g
s
.42509 .09962 m .42478 .09751 L .4329 .10015 L closepath p .955 .785 .905 r
F P
0 g
s
.85726 .09793 m .86229 .09753 L .85532 .10147 L closepath p .955 .785 .905 r
F P
0 g
s
.38811 .10004 m .38532 .10015 L .37699 .09757 L closepath p .955 .785 .905 r
F P
0 g
s
.37736 .09964 m .37699 .09757 L .38532 .10015 L closepath p .955 .785 .905 r
F P
0 g
s
.38937 .0987 m .39369 .09983 L .37699 .09757 L closepath p .955 .785 .905 r
F P
0 g
s
.38811 .10004 m .37699 .09757 L .39369 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.33405 .10004 m .33127 .10015 L .3231 .09757 L closepath p .955 .785 .905 r
F P
0 g
s
.32344 .09964 m .3231 .09757 L .33127 .10015 L closepath p .955 .785 .905 r
F P
0 g
s
.33537 .0987 m .33962 .09983 L .3231 .09757 L closepath p .955 .785 .905 r
F P
0 g
s
.33405 .10004 m .3231 .09757 L .33962 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.25716 .09975 m .25248 .10135 L .24715 .09774 L closepath p .955 .785 .905 r
F P
0 g
s
.24874 .10165 m .24715 .09774 L .25248 .10135 L closepath p .955 .785 .905 r
F P
0 g
s
.29665 .09975 m .28349 .09975 L .24715 .09774 L closepath p .955 .785 .905 r
F P
0 g
s
.25716 .09975 m .24715 .09774 L .28349 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.24715 .09774 m .30982 .09975 L .29665 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.70047 .09974 m .69592 .10135 L .6904 .09778 L closepath p .955 .785 .905 r
F P
0 g
s
.69215 .10167 m .6904 .09778 L .69592 .10135 L closepath p .955 .785 .905 r
F P
0 g
s
.6904 .09778 m .7268 .09974 L .70047 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.6904 .09778 m .76249 .09793 L .7268 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.87227 .09894 m .87064 .0989 L .87475 .09778 L closepath p .955 .785 .905 r
F P
0 g
s
.31908 .09793 m .31846 .1017 L .31667 .10255 L closepath p .955 .785 .905 r
F P
0 g
s
.76249 .09793 m .76199 .10171 L .76021 .10256 L closepath p .955 .785 .905 r
F P
0 g
s
.76249 .09793 m .76021 .10256 L .75313 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.76249 .09793 m .75313 .09974 L .7268 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.85532 .10147 m .84517 .10015 L .85726 .09793 L closepath p .955 .785 .905 r
F P
0 g
s
.35467 .09811 m .35898 .10026 L .34764 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.40869 .09811 m .41296 .10026 L .40174 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.40174 .09983 m .39771 .09975 L .38937 .0987 L closepath p .955 .785 .905 r
F P
0 g
s
.39369 .09983 m .38937 .0987 L .39771 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.34764 .09983 m .34358 .09975 L .33537 .0987 L closepath p .955 .785 .905 r
F P
0 g
s
.33962 .09983 m .33537 .0987 L .34358 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.59261 .09998 m .5891 .10015 L .58487 .09872 L closepath p .955 .785 .905 r
F P
0 g
s
.58657 .10021 m .58487 .09872 L .5891 .10015 L closepath p .955 .785 .905 r
F P
0 g
s
.60146 .0998 m .59864 .09974 L .58487 .09872 L closepath p .955 .785 .905 r
F P
0 g
s
.59261 .09998 m .58487 .09872 L .59864 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.15651 .09875 m .15676 .10116 L .14463 .10148 L closepath p .955 .785 .905 r
F P
0 g
s
.87064 .0989 m .87227 .09894 L .8644 .10096 L closepath p .955 .785 .905 r
F P
0 g
s
.87123 .09943 m .8644 .10096 L .87227 .09894 L closepath p .955 .785 .905 r
F P
0 g
s
.35898 .10026 m .35682 .09918 L .36303 .10033 L closepath p .955 .785 .905 r
F P
0 g
s
.41296 .10026 m .41082 .09918 L .41699 .10033 L closepath p .955 .785 .905 r
F P
0 g
s
.49063 .0993 m .48788 .10033 L .47243 .09956 L closepath p .955 .785 .905 r
F P
0 g
s
.44535 .09975 m .49063 .0993 L .45424 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.96345 .10283 m .9618 .09933 L .96836 .09958 L closepath p .955 .785 .905 r
F P
0 g
s
.9574 .10131 m .95252 .10041 L .96221 .10041 L closepath p .955 .785 .905 r
F P
0 g
s
.95144 .10282 m .94629 .09955 L .9618 .09933 L closepath p .955 .785 .905 r
F P
0 g
s
.9618 .09933 m .96345 .10283 L .95144 .10282 L closepath p .955 .785 .905 r
F P
0 g
s
.51577 .09998 m .51166 .10015 L .50765 .09939 L closepath p .955 .785 .905 r
F P
0 g
s
.50872 .10021 m .50765 .09939 L .51166 .10015 L closepath p .955 .785 .905 r
F P
0 g
s
.81137 .09943 m .81033 .10021 L .80318 .09961 L closepath p .955 .785 .905 r
F P
0 g
s
.795 .0998 m .79096 .09974 L .81137 .09943 L closepath p .955 .785 .905 r
F P
0 g
s
.77861 .10007 m .77582 .10015 L .77074 .09953 L closepath p .955 .785 .905 r
F P
0 g
s
.77185 .10021 m .77074 .09953 L .77582 .10015 L closepath p .955 .785 .905 r
F P
0 g
s
.47905 .10026 m .47243 .09956 L .48788 .10033 L closepath p .955 .785 .905 r
F P
0 g
s
.47243 .09956 m .47905 .10026 L .45424 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.80629 .10015 m .80318 .09961 L .81033 .10021 L closepath p .955 .785 .905 r
F P
0 g
s
.80318 .09961 m .80629 .10015 L .795 .0998 L closepath p .955 .785 .905 r
F P
0 g
s
.4329 .10015 m .43009 .10026 L .42509 .09962 L closepath p .955 .785 .905 r
F P
0 g
s
.42606 .10033 m .42509 .09962 L .43009 .10026 L closepath p .955 .785 .905 r
F P
0 g
s
.42606 .10033 m .43009 .10026 L .42509 .09962 L closepath p .955 .785 .905 r
F P
0 g
s
.38532 .10015 m .38253 .10026 L .37736 .09964 L closepath p .955 .785 .905 r
F P
0 g
s
.37851 .10033 m .37736 .09964 L .38253 .10026 L closepath p .955 .785 .905 r
F P
0 g
s
.37851 .10033 m .38253 .10026 L .37736 .09964 L closepath p .955 .785 .905 r
F P
0 g
s
.33127 .10015 m .32849 .10026 L .32344 .09964 L closepath p .955 .785 .905 r
F P
0 g
s
.32453 .10033 m .32344 .09964 L .32849 .10026 L closepath p .955 .785 .905 r
F P
0 g
s
.32453 .10033 m .32849 .10026 L .32344 .09964 L closepath p .955 .785 .905 r
F P
0 g
s
.7581 .10135 m .75313 .09974 L .76021 .10256 L closepath p .955 .785 .905 r
F P
0 g
s
.7581 .10135 m .76021 .10256 L .75313 .09974 L closepath p .955 .785 .905 r
F P
0 g
s
.31466 .10135 m .30982 .09975 L .31667 .10255 L closepath p .955 .785 .905 r
F P
0 g
s
.31466 .10135 m .31667 .10255 L .30982 .09975 L closepath p .955 .785 .905 r
F P
0 g
s
.80064 .09998 m .795 .0998 L .80629 .10015 L closepath p .955 .785 .905 r
F P
0 g
s
.35331 .10004 m .34764 .09983 L .35898 .10026 L closepath p .955 .785 .905 r
F P
0 g
s
.40735 .10004 m .40174 .09983 L .41296 .10026 L closepath p .955 .785 .905 r
F P
0 g
s
.46665 .10004 m .45424 .09983 L .47905 .10026 L closepath p .955 .785 .905 r
F P
0 g
s
.46665 .10004 m .47905 .10026 L .45424 .09983 L closepath p .955 .785 .905 r
F P
0 g
s
.21306 .10147 m .21181 .10092 L .21487 .10006 L closepath p .955 .785 .905 r
F P
0 g
s
.06732 .10568 m .06855 .10039 L .07286 .11305 L closepath p .955 .785 .905 r
F P
0 g
s
.15676 .10116 m .15777 .1046 L .14149 .1119 L closepath p .955 .785 .905 r
F P
0 g
s
.14149 .1119 m .14463 .10148 L .15676 .10116 L closepath p .955 .785 .905 r
F P
0 g
s
.08286 .11231 m .0825 .11275 L .08526 .10117 L closepath p .955 .785 .905 r
F P
0 g
s
.08526 .10117 m .08433 .10665 L .08286 .11231 L closepath p .955 .785 .905 r
F P
0 g
s
.08526 .10117 m .08286 .11231 L .08433 .10665 L closepath p .955 .785 .905 r
F P
0 g
s
.25248 .10135 m .25068 .10255 L .24874 .10165 L closepath p .955 .785 .905 r
F P
0 g
s
.69592 .10135 m .69422 .10256 L .69215 .10167 L closepath p .955 .785 .905 r
F P
0 g
s
.24949 .10295 m .24874 .10165 L .25068 .10255 L closepath p .955 .785 .905 r
F P
0 g
s
.24949 .10295 m .25068 .10255 L .24874 .10165 L closepath p .955 .785 .905 r
F P
0 g
s
.69306 .10296 m .69215 .10167 L .69422 .10256 L closepath p .955 .785 .905 r
F P
0 g
s
.31772 .10295 m .31667 .10255 L .31846 .1017 L closepath p .955 .785 .905 r
F P
0 g
s
.31772 .10295 m .31846 .1017 L .31667 .10255 L closepath p .955 .785 .905 r
F P
0 g
s
.76129 .10296 m .76021 .10256 L .76199 .10171 L closepath p .955 .785 .905 r
F P
0 g
s
.95748 .10391 m .95144 .10282 L .96345 .10283 L closepath p .955 .785 .905 r
F P
0 g
s
.14306 .10669 m .14346 .10469 L .14385 .10408 L closepath p .955 .785 .905 r
F P
0 g
s
.14306 .10669 m .14173 .10426 L .14346 .10469 L closepath p .955 .785 .905 r
F P
0 g
s
.15955 .10705 m .14149 .1119 L .15777 .1046 L closepath p .955 .785 .905 r
F P
0 g
s
.07286 .11305 m .06401 .10844 L .06732 .10568 L closepath p .955 .785 .905 r
F P
0 g
s
.06592 .10738 m .06732 .10568 L .06401 .10844 L closepath p .955 .785 .905 r
F P
0 g
s
.06592 .10738 m .06401 .10844 L .06732 .10568 L closepath p .955 .785 .905 r
F P
0 g
s
.08322 .11188 m .08286 .11231 L .08433 .10665 L closepath p .955 .785 .905 r
F P
0 g
s
.08322 .11188 m .08433 .10665 L .08286 .11231 L closepath p .955 .785 .905 r
F P
0 g
s
.15955 .10705 m .1654 .10902 L .14149 .1119 L closepath p .955 .785 .905 r
F P
0 g
s
.07286 .11305 m .07169 .11301 L .06401 .10844 L closepath p .955 .785 .905 r
F P
0 g
s
.06061 .10898 m .06401 .10844 L .07169 .11301 L closepath p .955 .785 .905 r
F P
0 g
s
.17027 .10898 m .17076 .10886 L .16915 .10952 L closepath p .955 .785 .905 r
F P
0 g
s
.1729 .11002 m .16915 .10952 L .17076 .10886 L closepath p .955 .785 .905 r
F P
0 g
s
.07169 .11301 m .07052 .11296 L .06061 .10898 L closepath p .955 .785 .905 r
F P
0 g
s
.05402 .10916 m .06061 .10898 L .07052 .11296 L closepath p .955 .785 .905 r
F P
0 g
s
.16915 .10952 m .1654 .10902 L .17027 .10898 L closepath p .955 .785 .905 r
F P
0 g
s
.1729 .11002 m .14149 .1119 L .1654 .10902 L closepath p .955 .785 .905 r
F P
0 g
s
.07052 .11296 m .06716 .11305 L .05402 .10916 L closepath p .955 .785 .905 r
F P
0 g
s
.05359 .11091 m .05402 .10916 L .06716 .11305 L closepath p .955 .785 .905 r
F P
0 g
s
.1729 .11002 m .17259 .11362 L .16673 .11153 L closepath p .955 .785 .905 r
F P
0 g
s
.1729 .11002 m .16055 .11305 L .15044 .11329 L closepath p .955 .785 .905 r
F P
0 g
s
.14149 .1119 m .1729 .11002 L .15044 .11329 L closepath p .955 .785 .905 r
F P
0 g
s
.06716 .11305 m .06244 .11329 L .05359 .11091 L closepath p .955 .785 .905 r
F P
0 g
s
.05317 .11266 m .05359 .11091 L .06244 .11329 L closepath p .955 .785 .905 r
F P
0 g
s
.16944 .11354 m .16673 .11153 L .17259 .11362 L closepath p .955 .785 .905 r
F P
0 g
s
.16673 .11153 m .16944 .11354 L .16055 .11305 L closepath p .955 .785 .905 r
F P
0 g
s
.15044 .11329 m .14638 .11354 L .14149 .1119 L closepath p .955 .785 .905 r
F P
0 g
s
.14347 .11362 m .14149 .1119 L .14638 .11354 L closepath p .955 .785 .905 r
F P
0 g
s
.06244 .11329 m .05772 .11354 L .05317 .11266 L closepath p .955 .785 .905 r
F P
0 g
s
.05437 .11362 m .05317 .11266 L .05772 .11354 L closepath p .955 .785 .905 r
F P
0 g
s
.1574 .11296 m .15044 .11329 L .16055 .11305 L closepath p .955 .785 .905 r
F P
0 g
s
.1574 .11296 m .16055 .11305 L .15044 .11329 L closepath p .955 .785 .905 r
F P
0 g
s
.165 .11329 m .16055 .11305 L .16944 .11354 L closepath p .955 .785 .905 r
F P
0 g
s
.25 Mabswid
.99582 0 m
.00418 0 L
s
.00418 0 m
0 .11842 L
s
0 .11842 m
1 .11842 L
s
1 .11842 m
.99582 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[4]:=",
  ImageSize->{450, 60.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000720000?0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoIPD7o`2h00H700051`80o`Nd1`801PL000D7
0P3o1k@70P061`001@L00`07000:1`l02PL300/77P081a800PLB00P74P061a800`LO00031`0000/0
2@L200X77@081a400`LB00X74P0;1`l01@L=00P74@0A1`0300L000H700051`0400L700L74P091`@0
2@LP00L79`061a@010LD00031`0002d020L300P77`071a8000<700004P081a@020L300031`0700T0
00<700000P000`L0000=00L74P0@1`0300L000H700051`0400L700L74`081`@020LR00H79`061`/0
00?71lL01`021b0000?700003`000lL7a`070003a`0000H020L400H700<0L`007P051`0307<00140
00>B00004@081a@01`L01@07003702401@L00`2B000700031lL700L03PL010071`061`001@L01@07
1`L000L72`000kD0000400L70P001:h700071a4000>e00003P061`P000<7[PL070061`l000<70000
0`021a/000>G000020000`Ne000500041kD0U`8000<700001`000`L0000400L70`000jh000051a80
00>e000030051`T000<7[PL03@001kD00007[PL00`000`L0000200L75@061`0400L01`X000>`0000
10000`L000040003U`0000P01@L@0004]@0000d700D01`L700061`001@L01@071`L000L720000lL0
000800H70P000`O7000200L730000`L0000B00H72P001<L700L60003OP0001401@L800031`0000X0
0PL60004/0001`d000=n0000:`071`8000?71`000P041b401@L<00031lL000d000<700002@071`T0
00<7a`002@051`<000<700001@000kD0]@020007/02`0<L7a`0@00D700<0M0002@000lL7000500h7
00D01`L700061`001@L00`071`021a<000Be0;Fe2`000jh0[P0C0003[P2^00<000>e00006`000jh7
000=0003U`00008000<7[PL01@000kD0000_0003]@0001<029`50003[P0000X000>e00002@000`L0
1`090003W02L00D000<7[P000`001;D0W7h60004]@00]@/000JL0000[PL50003]@2e00D000>L0000
0`000`L000040003W;D000/000=n07h02P000kD0]@080003[PL000<000Bc0;Fe0P000gjL00060005
1`0009`00P000gT000020003]@0000`000JL0000[PLG00@700<01`L010L000D700<01`L00PL00`07
1`051`h000>e00000P051`8000@700O70P071`@000NL0007a`37008000K:002e09`400031`0000L0
1PL30003W00000D000<7000020001kD00007a`L02P061`8000>G09L00P000`L0000;00872@001<L0
00L30003]@0000/000<700000`000kFL00020006W;FL]Ibe5@071`8000C70<L70P031`D000RL0000
a`O71`@000Be002e0`0010L00;D20003]@0700@71@000kD0000300031`00008000>e0;D030000`L0
00040003]@0700H71P000kD0000<00@70`000`O71`040005]@2e09L00`0017jG0;0300031`0000`0
1@L<00031lL000D03PL00`071`021`0300L700@700051`0300L7008700<01`L01PL>0004]@0000@7
0P001@N^0:h700801`L90005[PL00<D020000g@0000900D75`001Y`00:h7[PL000Be00001@L80003
1`0000X00PL80005[PL00:h01@000kD000030003U`00008000<700001@03W0<019`600031`0000`0
1PL300041`07[P<00PL20003OP0000@000<7[PL010001I`0001n008000>B00001P051`T000B^1`07
0`000kD0000;00031`0700L01`L90007M00700L01`0500@710000`N^00030005]@2e09`00`000iMn
000;0003TP0000D01@L:0005[PL00:h01@0>1`0300L7008700<01`L010L000D700<01`L00PL00`07
1`061`H000?7000010000kD0000200@70P001LL0a`O700<01`L20003]@0000@000?700001@000i`0
00090004]@0000H70P001Y`0a`371`L000<700000`001[D00007a`<000G60000]@0400D70P000iL0
U`0400031`0000T00`L:0003a`0000<000>e000010000kD000040003a`0000<000FL]IbeW0030003
W;D0014000>e00001PL400031lL000<00PL20003]@0000@000?71lL010000kD0000:0003]@0000H7
0`000i`0000200031`0700D000>e00L0200017<000L400:e0P081`050;D009L01P000`L0000600<7
0P0010O70<L500:e00H0]IL0OYL30004OP2`TPP000>`00000`061`X000@700O71@0?1`0300L70087
00<01`L010L000D700<01`L00PL00`071`071`D000<700001@001;D000041`8000@700N^10071`L0
00<7[P005P071`D000>^1`000`000lP0000200031`07008000>G00000P000`L000040003]@000080
1`L700031`0000@000>e00000P041`8000>L09`00`000`L000050003]@0000@000>GW0000`000jh0
000400NL2@000i80000<00H70P001:h0[PL20003[P0000870P000i`000040003[P0000<000BL]@2L
0`000g<0000500L70`000i`0000500041`00]@<000<700000P000i`0000300031jh000<000>e0000
2@L40003W00000<000<7[P001P041`8000>^1`001@000kD000040003Ugh0008000=A00000`000jh0
00020003M00000<01`LB010700<01`L00PL00`071`041`001@L00`071`021`0300L700T73P051`80
00C70<L71@071`P000?700004`0:1`8000?71lL010000lX0000200@71`000`L0000500`71P000`L0
000500P72P000`L0000300D71`000`L000040009W;FL]IbeW;FL01D01`L200031`0700H010L70003
1`0000T000=c0000100:1`L000<7a`L01@041`T000?70000100=1`P000G70000a`0200D700D0L`00
a`0700H70`001Wh0001iEPL000=<00000`081`X000?71`000P0B1`0300L7008700<01`L010L000@7
00<01`L00`L00`071`0;1`L02@L40003[PL000D02`L200031jh7008000C500001PL500h70P000`N^
1`030004b/P000X70`000jh00003014710001@L00<@000h71@000`L0000200/70P000`L01`0400`7
4@091`<000<7[P000`000`L000061`D000<700000P051`0404/0B`@03PL200041jh7[P<02PL;0147
30051`<000B^00070P0:1`8000IA0000DGT50004M01dB`D02`L50003[PL000804`L00`071`031`03
00L700<700041`0300L700<700<01`L02`L700T70P0010L01lL700X70`0020O7003:0<T01PL600`7
0`001<L700L500/70P000lL7000401470P000lL0a`0400l71P000lL0000;1`8000?70000100=1`T0
00<7000010091`<000?700001P071`@000?700000P051`0407<0L`D030L30003a`O700<030L60004
a`0001872@000`L01`031`D000C700003@L20003N@1i008000QF0000C01<M0H02PL500031`070080
4`L00`071`031`0300L700<700041`0300L700<700<01`L030L600T70P000jh0[P0800X70`001Jh7
003500<01`L500`70P000jh7[P0200051`3:b0002`L200031`00008000?400L040L200031`070080
00?400L03PL30004[P2^1`802`L900T70`031`L000<7000010091`<000<700001`061`<000>^1`00
0`061`8000=;00000`0<1`8000B^1jh70`0<1`<000Eb0000[P03014720000`N^00041`<000<70000
0`0>1`040540D@8000=i00000`000g@0000500X710000`L01`0301<700<01`L00`L00`071`031`00
10L00`071`031`0300L700`71P091`<000?71`00200:1`030<L000L01`L600/70P0010O70<L20004
bP3:00/70P000lL0000401470`000`O7000300l70P001LL700O7008030L200031`0000<020L30003
a00700871P000`L0L`0400P72P0017T000061`<000<700L00`061`8000=c0000100;1`8000D7a`L0
1`0200`70`001GT0000700<04@L60005a`371`000`L40003a`0000<03`L30004EP00U`@000A<001d
0P000lL000091`D000D70037000C1`0300L700<700<01`L00`L000@700<01`L00`L00`071`0<1`L0
1`L400031`0000P02PL00`07000700P700<0M0003`000jh700030004a03800/70P000`L000050147
1@000`L0000?1`8000<700L0100<1`8000>^00000`081`D00`L600031`0000@020L30003[P0000T0
1@L30003[P2^00<01`L20003L`00010000>^1jh0100=1`@000>^0:h00`0A1`D000>^0:h00`031`H0
00@700003`L40005D@0000L00`001D/0B`1d010000@7[P2^0P0C1`0300L700<700<01`L00`L000@7
00<01`L00`L00`071`0=1`H01`L40003a`0000<000?800000`091`P000?500L01`L<00041`00a`80
00<700L00`000lX0000<1`X04@L30003a`3700<03PL900`720081`D010L500031`0000D01`L?00H7
0P000`L01`0300L70P000d`0C0030003a`37008000<700L00`001@L01lL700<03PL400031`0700<0
4@L900031`07008710001<L000L2010700<0N@0010021`8000Ad07A<1@003LL7a`O70<L7a`O70<L0
0P000`L01`0C1`0300L700<700<01`L00`L000@700<01`L00`L00`071`0=1`0300L000<01`L30003
[P0000@000?:00000`091`0300N^00L02@L@00031jh000@000?4bP003@L901470`000`L01`0300h7
0`001:h700L200`72@071`D01@L300<71P071`/000=?0:h00P051`8000>^0:h00`081`0407<0La00
00B^1`070P0>1`@000>^0:h00`0A1`H000@700N^0P021`H000@700004@L500<700H0M4/0Bg@@0006
[PL01jh050L00`071`031`0300L700<700041`0300L7008700<01`L03PL600L70`000`L000020003
bP3800H020L200031`0000<000?500L020L60008a`L00;37/<L20003a`L000<000C800003@L20003
1`0000@04@L50004a`L000l70P0010L00<L200`72@071`D02`L700D74@051`@000<700000P081`80
00=<09800`001<L7a`L30008ak00a`O70<L200h71@000lL7000301471P001@O71lL000871@000`L0
000201870`051`0707@0M4`0TP0>00061`07a`L05@L00`071`021`0300L700<700041`0300L70087
00<01`L03`L600H70P000`L000060003a00000<020L20003[P0000D02PL60007a@O5ajk7[P020004
1`00[P<000?4bP003@L20003[P0000@04@L500041jh000l710000jh7000200/72@061`H030L600D7
4P041`@000>^00000P081`8000=c00000`003PO51lG7[PN^a`07[P2^0`0>1`D000<7[P000`0A1`H0
00F^1jh700021`@000@7[P2^0P0C1`0300L700@700@0B`1;2P000`N^00060004[PL001D700<01`L0
0PL00`071`031`0010L00`071`021`0300L700l71P061`8000?7000010000l[8bP0600L70P000`L0
00040003[00700P71P001`O71k37/<L00P000lL000030003b<X000h70P000`L0000401471@000lL7
000C0003a`O7008030L30004a`001`h030L600D71@000lL0000:00@70`000lL7000300T720001`O7
1k37/<L00P001<L700L300h72`0A1`0301P0008000K70007a`021`D000<700000`0J1`8000=d0000
1P000lL70002000:1`00a`371`0701D700<01`L00PL00`071`031`0010L00`071`021`0300L700l7
1P051`/000?:a0001@071`/02@L60009a@O5ajk7[P2^008000<700000P000l@01`0=1`8000>^0000
100A1`@000R^1`00[PN^1`T000@7[PN^0P000`N^1`0200`70`001PL00:h0[PP000@7[PL03@L600@7
2@021`P00`L300031jh000<02@L20003B`0000<000O51lG7[/N^008000D7[P001`0200l71@000jh0
000201470P001QP0002^1`@000<700001@001:h0000J1`<000Ed07@0TP020006[PL00:h71@000`L0
1`0201D700<01`L00PL00`071`031`0010L00`071`021`0300L701070P000lX000051`8000<70000
1P000lP01`0600H72`0:1`H00/L01[37/<L01`8000?700000P000lX01`0=1`8000<70000100A1`@0
00<7a`004`0010O71lL300/70`001<L000L>00d700<060000`041`T00PL20003C01i00<00`L30003
a`0000<02PL00`1c00050006a`N`ak370P000lL0a`03010700<0L`000P000`L00002014710001Y80
1lL7a`8000<700000`000`L0000301/70P000d`0C0070004a`L01`<000@700O70P0E1`0300L70087
00<01`L00`L000@700<01`L00PL00`071`0@1`H010L:0003b00000871@061`T000?300L02@L60005
1jk7[/L00P001:h0[PL20003a<X000l72@0B1`<000>^00000P000`Ne1`0900091kD7]@2^1jh700<0
30L@00041jh700d71P041`@000>^00000P021`0304/000H00PL300061`0000L02PL0101;04/40009
1lG7[/N^0:h7008000>^000040L50005[P001`004PL50004[PN^1`8000<700000P001I801`070080
6`L01P1;M01dB`@000H7[PL01jh200061`07[PL05PL00`071`021`0300L700<700041`0300L70087
00<01`L040L600@720000`L0bP0200871P051`H000D7002/000;1`D00/L00k37/00400071`000<[8
000?1`H000@700004@L300031`0001@000C71lL70P0<1`H000?700002@000lL0000<1`H00`L20003
a`0000D00`L20003L`2G00<00`L20005a`000<L00P091`0407<0L`D000/7/<N`a`0700L01`020107
1@0010L00<L2014700@0L`2B0P000lL7a`0200031`0000@000?700000`0K1`0704ad07@0/0030004
1lL0a`@000D7a`O7000F1`0300L7008700<01`L00`L000@700<01`L00PL00`071`0A1`D010L40003
[P0000H00`L500D71P001Jh00<<000/71P002Jk7[/L01`07[P030003a<X000l71P001:h0000A1`P0
00>L00000P0;W08000@7[PN^0P0<1`H000@7002L0`000gh0000500d71P021`@000<700000`031`80
00=;000010021`H000<700002PL20003B`0000@000K7[/N^0:h200041jh0010700<060001@000`L0
000A1`<000Ib0007[PL200031`0H008000>B00000`000`L01`0J1`0507A;04/00P001i800:h0[PL0
10000jh7[P0201H700<01`L00PL00`071`031`0010L00`071`021`0300L701471P031`X00/X01007
1`L600@71P001@L00:`000`71@001<N`ak030003a`L0008000?:000040L500041`00014720001Ibe
W;FL00<000RL]IbeW;FL]@<000H7a`L0bP0;1`H000?70000300=1`H00PL40003a`0000801@L00`1c
000500871P000lL0000:1`8000=c000010000k37/0030004a`371`8040L00`1c00050003a`000147
0`001Y800<L7a`@000=c00002@0K1`0307@000<000RB0000a`L01`8000C71lL70P0F1`0300L70087
00<01`L00`L000@700<01`L00PL00`071`0A1`@000Bn00L72`000lP01`041`D010L60003[P350080
30L50003[/N^00<000N^1`00b@34008040L50004[P0001870P001:h000L200NL0`05W0<000>^1jh0
0`0;1`H000<700000P000iL0000700d71`000`L0000700D70P001G<00007008000<700001`0;1`T0
00B^a`070P001:h7[P0A1`8000=a00000P001:h7000A1`L000>^1`001`000g@0[P0200031`0001/7
00<0B`000P000d/000030009[P000:h7[PN^00805PL00`071`021`0300L700<700041`0300L70087
00<01`L04@L40004Y`071`@000?700000P02bP801@L600<71P000`L0000200d71@000k3700020005
1lL000L00`0A1`D000@700004PL00`371`02000=a`00]IbeW;FL]IbeW0040003]@00008000?71`00
0P0;1`H000?700000P000kD000060003a00700`71P001@L0000700D01PL00`1<000500031`0000L0
2`L20003C00000@000S7/0370<L7a`804@L700041`0701871P000`O700060003C00000<000C70000
6`L30003C01<00D000S70<L7a`O71`805PL00`071`021`0300L700<700041`0300L7008700<01`L0
4@L600872@001<S:00061`D00`L60006[P350:L030L40003aLN^008000<7[PL00`000lX0000A1`@0
00<7[P000P0B1`0300L0008000<700004`000`N^000200`710000jh7000>00d72P000jh0000300H7
00<0Ld/00`000jh0000900`700@0L`1c10000jk7000200031jh700804PL30006M000[P2^0P0A1`<0
00Ed002^1`0300051`0007<00`001PL01`0701/70P0014]d07@60006[PN^1jh70P0G1`0300L70087
00<01`L00`L000@700<01`L00PL00`071`0B1`@000?300L00`000`L0000300?:00<01`L010L600<7
1`001<X0a@0=1`0307T0008000G700371`020005bP00bP004@L40003a`L000804PL400041lL000l7
10000lL7000200`710000`O7000200@7200=1`T000?7000010071`0707<007T0a`0:00`700H0C01<
09830007/0001lL7a`02018700<060000`000`L01`02014700@0601c0`002/L0000704`0C7@30004
1`00a`8050L200D700P0M4`0C7@0/08000O71lL7a`O700801PL700X700<01`L00PL00`071`031`00
10L00`071`021`0300L7018710000j801`030008[P000<T0b<X200L71@031`H000<700000P0=1`D0
00F^0:h7[P050003a007010710000`N^0003014710001Jh70<X000h71@001:h0000<1`@000>^1`00
0P051`L03PL800031`0000@01`L0101;04/800031`0000803@L0101;04/40006a`00[PN^0`0B1`04
0740M08000>^0:h00P0A1`L000T700L01`1c07<00`001I800:h700804`L400D700D0M01dB`0=00D7
0`000jh0000400P700<01`L00PL00`071`031`0010L00`071`021`0300L701872@000lL000030005
bP00bP001`L600871P000lL0000200h71P000`O71`020004bP3800@70P0<1`@000?71lL00`0<1`80
0`L400031lL000803`L400041lL000`710000`O71`0200H71P051`0300L700H71`000`O7000400T7
0P000g<000060003a`3700803@L0101c07<60004a`O71`8020L00`071`081`<000Ic0000a`L200d7
0P021`L000?700000P0010L004`800031lL001871@041`0604`0C7A<0`001lL7a`O71lL00P041`<0
00G70000a`0200031`0700H700<01`L00PL00`071`031`0010L00`071`021`0300L701872@000`L0
00040003b<X000P71P021`@000<700000P000j`01`0=1`D000@7[PN^10000lX01`031`@02`L30003
1jh700<02`L4008710000jh7000200l71@0010L0000;1`@000>^1jh00P071`D010L300P71P000jh7
000400X71P001@L00007008000>^00003@L20003B`0000@000@7[PN^0P071`<02@L30006M00000N^
0P0<1`@000<707401@000`L01`0200071`0004l0C`0400031jh7008040L600D70P000gA;M0030007
1jh7[PN^1`0200<70P001PL00:h0[P@000<700L01@L00`071`021`0300L700<700041`0300L70087
00<01`L04PL800031lL000D00/X00`071`071`H000<7000010000lL0000200l710000lL7a`050003
b00700870`000lL0000:1`<000K71lL0aP0:1`0500L01`000PL400031lL000803`L50004a`0000/7
10000`O71`0200L71@041`@01`L600041`00a`<02PL@00d70P000g<000040003a`O700<01PL500T7
0`001g@00<L7a`0030L400041`00L`H000?700000PL20005M01d0;001@000`L01`0?1`H01@L20003
C7@000@000D7a`O71`0300<72`000`L000061`0300L7008700<01`L00`L000@700<01`L00PL00`07
1`0C1`L000>^00001@000lS:00091`0307@000<000<700000`001Zh0002/00l710000`N^1`030004
1`0000<71`091`8000N^1jh70<X000T70P001:h000021`<000<7[PL0100>1`D000@700002`L30003
1jh700<01`L500<700<01jh00P071`D000<7[P001@0;1`H000>^00000P0010L0000>1`0304/000D0
00<7[PL00`061`H020L30006LP001jh70P091`H000<700000P000g80[P0200031`00008700D0M4l0
C`020005U`00[PL00`0>1`L01PL20003B`0000@000@7[PN^0`021`8000B^00072P051`0300L70087
00<01`L00`L000@700<01`L00PL00`071`0C1`H000?71`001@001<X0bP0:1`/000H70000a@0?1`D0
00<7a`000`000lX01`041`@000?700L020L200031lL700@02@L00`3700020003a00700<000?71lL0
100>1`T02`L30004a`001`8020L300<71P071`D000?71`001@0;1`X000<700000P0>1`<000=c0000
0`000`O7000200H71@000`L01`061`D000<7a`L00`091`03[@0000800`L80003a`0000<700<0M000
0P001GT00007008000?700003@L700H70`000e00D00400031lL700800PL;00061`0000L01@L00`07
1`021`0300L700<700031`0300L700<700<01`L050L500031jh000@000C8bP002`L00`1?00070003
1jh700@040L40003[P0000<000?400L010L400031`0000P70P001:h000L300P71P001<P01`L20004
1jh7[P<03PL20003[P2^00@030L50004[PL000`71`081`<000F^1`00[P0300d700<0B`000P000jh0
000600l700<0L`001@000jh7000200D70P001d/0002^1`001PL50004[P001`8020L300031`0000<7
1@000jh0000300@70`001GT0002G00@000>^1`0030L30004[PL000L70P000g@0M0040003[PN^0080
0PL30003[P0000H000<700000P051`0300L700<700<01`L00PL000<700<01`L00`L00`071`0D1`@0
00<7a`001@001<X0000<1a0040L400031`0000D01PL30003a`L000801PL30004a`O71`801PL40003
a`0000<00PL300031lL700<03PL200031`0700D02`L20003a`0000@02PL30003a`3700<020L40004
a`001`<03PL50004a`001`@03`L00`1<00060003a`0000H70P000g@000020004a`L000@71P001LL7
a`L000L70`0010L000031`H000?700000`031`8000AiEP1F1@000lL0000300T70`000lL7000200L7
00@0M01d1P001LL70007008000>B00000`0010L00<L200041`0700D700<01`L00`L00`071`021`00
0`L00`071`031`0300L701@710000jh700060003b/P000`700<0M00020000`L0000201470`002:h0
00390<@01PL300041`07[P<010L:00<70`001:h000L30003b00700<000@7002^100=1`<000<700L0
1@091``01`L30003[P0000@020L400041`00[P<03PL20006B`001`070`0010L0000>1`P000<70000
1PL700031`07010010L40003[PL000<00`L400031jh0008000>^1`000P021`0<055i07UA09L0U`2^
0P000`L000081`8000>^00000`081`8000=;00001@001:h000L60003[P2^00T01@L00`071`031`03
00L7008700031`0300L700<700<01`L05@L20003a`L000L00/X00`071`0;1`P000?71`00100A1`P0
00?800001PLM0003a`3701L01@LB00<780041`T000<700000`0?1`<000>G00000P001<L7a`L600X7
0P000d`0000600H7:0000`L0000500031lL000X000AF05Ii10000iL0000E00L70`001E00D02B00L0
00=<09802`000lL000051`0300L700<700<01`L00PL000<700<01`L00`L00`071`0E1`8000<7[P00
1`000l[8000=1`D000@H0:h71P0B1`X01`Ld00031`0000<750000`L0000O00<710000jh700040003
1`0000803`L20003L`0000d02`L0101c07<600L7;@000jh0000>0007N@1i07T0U`040003[P2^0100
20L20005M01d09805`061`0300L700<700<01`L00PL000<700<01`L00`L00`071`0E1`0400O71`H0
0/X200l73`0A1`H000C:00001`L700051kD7]@L030001KD7]@Ne01X00`LP0005]@000;D04@031`L0
00>e00001@0A1`0304`000H000>e00001P0:1`0404`0C0H01`L900071`07]@Ne1`0<0005]@Ne1kD0
0`000lL0000800031`0000/000Mi001i07Un00<000>e00002P000kD0000500P700H0M500D7@>0006
a`L00<L710061`0300L700<700<01`L00PL000<700<01`L00`L00`071`0E1`030:h700/03`L60005
[PN^0:h0100B1`D000C80<@01`L400031`0001l000<700003P031`@000<700004P000kD0000A0003
]@2e00H00`L300031`0000T04@L500031`0000L000<700002`L0101;04/400T700<0L`00>00020L0
05Ii07UF0P000iL000040003/02e00<000>e00000`000gh01`091`0507A?04l01P001g<0N@1i0;D0
2P071`0300L700<700<01`L00PL000<700<01`L00PL00`071`0E1`<000?700001P000lX0000@1`03
07T0008000@700371@0B1`T020L60003]I`000@000nL]IbeW;FL]IbeW;FL]I`02P000kD0000=00<7
1@000kD000030003]@0000`000FL0000W0020007]I`00;FL]@020003W;D000T00`L40008]@00W02L
09`301871@000kD0]@020003W00000@02`L01P1d07@0OP802@L00`1<00060003W;D000<0016L]Ibe
W;FL]IbeW;FL]IbeW009000@W02L09`0W02L09`0W;D0]@800PL03P1nN@1i09L0UibG09Ne20000kD0
00020003W00700T700L0D7@0M02L00<000@7001F3P081`0300L7008700<01`L00PL000<700<01`L0
0PL00`071`0@1`H000<700L01@001PL00<S:010710000`L01`0:01071@000lX01`081`<000T70000
W9NLUi`01P002YNLUibGW9NLUi`:0003W00000L000=n000010021`D000>L000020000i`000070003
W00000<000FL09`0W0020003U`2G008000JLU`00W9L300@7100019`009`701871@000i`000020004
W000U`803@L20003B`1i00802PL50009]@00W9NLUibG00D000^GW9NLUibGW9NLU`030003OP0000P0
00>L00001@001I`009L000@70P0035H009Mn07h0OP1nW0800Y`20003W00000@030L00`1?00020003
OP00008700@0C`1?0P000gii000700T700<01`L00PL00`071`021`000`L00`071`021`0300L700l7
4`000lX01`0@1`03054000l03`L40003bP0700T710003KD0W;FL]IbeW;FL]I`010001kFL]IbeW;D0
1`001;D009`:0003]@0000800`L00`2L00030003]@0000<000>e00001`000kD00002000:W00009`0
W02L]@8000ZL09be0;D0]@2e0`041`<000BL002e200C1`D000BL002e0P0019`0000>1`8000An0000
2PL5000>1`2eW;FL]IbeW;FL]I`40007]IbeW;FL]@030003W00000L000>e00001@000kD0000300H7
0P002YbGW02L0;D0]IL20004]I`0]@@000>L00003@L700<70`000gh0OP020004]@00W0802PL00`07
1`021`0300L7008700031`0300L7008700<01`L03`L@0004b00001873`000`L0000?1`D02`L20003
1`00008000jGW9NLUibGW9NLUibGW0D000>LU`004`000i`0000300<71P000i`000030003OP0000D0
00<7000010001Y`0002LU`8000>LUi`00`000ibGW0020003U`2G00<01@L20004W000W08000>L09`0
0`0C1`<000JL09bL09L20004W00000h70P000gh0000<1`H0012GW9NLUibGW9NLUibGW9NL10000iNL
000500031`00008000BL09bL0P0029`0W02LU`2G0P081`<000QnU`2LOP1n00:L0P000ibe000500h7
0P0017h000041`<000=n07h00P0017h00;D200X700<01`L00PL00`071`021`000`L00`071`021`03
00L700l74@000lX0000C1`@000>e00002`0?1`@030L[0003W00000L00`LD00031`0000T000>eW000
4@051`d05@L>00l7100=1`8000>e000060001I`0000701@02`L50005U`00]I`01@031`803`L500D7
00<0Ugh00P001;D009`300/700<01`L00PL00`071`021`000`L00`071`021`0300L700l7500C1a80
40L300`71@0H1`@000<700L03`L00`071`031`0300L700h700D01`L7000I1`80A`L300d7100H1`D0
90L801P70`071`@000An002e0P0<1`0300L7008700<01`L00PL000<7o`2l00@700041`0500L71`00
4PL=0006OP000<P050L40004W000W0X0o`MU1`0500L71`001@L000D700@01`L04`LA01H70P0019`0
0;D;0?l7I@L010071`061`001PL00`07000C1`0=07T009`0W02L07h0OP0401H70`0019`009`30003
W02L00@0o`MU1`0400L700H700061`0300L001@72P000iL0000201P7403o1fD700<01`001`L000L7
0P3o1k070P081`0020Oo0;802@L00001\
\>"],
  ImageRangeCache->{{{0, 449}, {59.125, 0}} -> {-0.00831485, 0.432635, 
  0.00226421, 0.00226421}},
  CellTags->"DXFFiles"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; 
    MVLoadDXF["\</export/home/kuska/MathView3D/pharao.dxf\>"]\)], "Input",
  CellLabel->"In[32]:=",
  CellTags->"DXFFiles"],

Cell[BoxData[
    \("/export/home/kuska/MathView3D/pharao.dxf"\)], "Output",
  CellLabel->"Out[32]=",
  CellTags->"DXFFiles"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVWritePOVRay["\</export/home/kuska/MathView3D/pharao.pov\>"]\)], 
  "Input",
  CellLabel->"In[38]:=",
  CellTags->"DXFFiles"],

Cell[BoxData[
    \("/export/home/kuska/MathView3D/pharao.pov"\)], "Output",
  CellLabel->"Out[38]=",
  CellTags->"DXFFiles"]
}, Open  ]],

Cell[TextData[{
  StyleBox["MVGetGraphics3D[]",
    FontFamily->"Courier"],
  "  can now transfer the 3D data to ",
  StyleBox["Mathematica'",
    FontSlant->"Italic"],
  "s kernel."
}], "Text",
  CellTags->"DXFFiles"],

Cell[BoxData[
    \(gr3d = MVGetGraphics3D[]; \n
    grData = Prepend[gr3d[\([1]\)], SurfaceColor[RGBColor[1, 1, 0]]]; \)], 
  "Input",
  CellLabel->"In[7]:=",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[Graphics3D[Prepend[grData, EdgeForm[]]], 
      ViewPoint -> {\(-1.343\), \ \(-3.104\), \ \(-0.093\)}, Boxed -> False, 
      RenderAll -> False]; \)\)], "Input",
  CellLabel->"In[9]:=",
  CellTags->"DXFFiles"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.31003 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.176499 1.3302 -0.0201674 1.3302 [
[ 0 0 0 0 ]
[ 1 1.31003 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.31003 L
0 1.31003 L
closepath
clip
newpath
.02573 .63528 m .03141 .6187 L .03259 .63795 L .008 0 0 r
F
.03259 .63795 m .02846 .6451 L .02573 .63528 L 0 .004 0 r
F
.06407 .65017 m .02846 .6451 L .03259 .63795 L .453 .669 0 r
F
.06407 .65017 m .0529 .66333 L .02846 .6451 L .549 .6 0 r
F
.03141 .6187 m .03272 .60617 L .03732 .62348 L 0 0 0 r
F
.03732 .62348 m .03259 .63795 L .03141 .6187 L F
.07304 .6362 m .03259 .63795 L .03732 .62348 L .63 .674 0 r
F
.07304 .6362 m .06407 .65017 L .03259 .63795 L .639 .643 0 r
F
.03272 .60617 m .03437 .59254 L .04252 .60934 L 0 .012 0 r
F
.04252 .60934 m .03732 .62348 L .03272 .60617 L .011 0 0 r
F
.03437 .59254 m .03859 .57578 L .04449 .58892 L 0 0 0 r
F
.04449 .58892 m .04252 .60934 L .03437 .59254 L 0 .02 0 r
F
.08314 .61758 m .03732 .62348 L .04252 .60934 L .656 .66 0 r
F
.08314 .61758 m .07304 .6362 L .03732 .62348 L .655 .662 0 r
F
.03859 .57578 m .04054 .55534 L .05044 .56631 L .124 .135 0 r
F
.05044 .56631 m .04449 .58892 L .03859 .57578 L .029 0 0 r
F
.04054 .55534 m .04238 .52811 L .05086 .54049 L .092 .091 0 r
F
.05086 .54049 m .05044 .56631 L .04054 .55534 L .124 .135 0 r
F
.04238 .52811 m .04638 .48378 L .05717 .48762 L .231 .179 0 r
F
.05717 .48762 m .05086 .54049 L .04238 .52811 L .171 .116 0 r
F
.08929 .59203 m .04252 .60934 L .04449 .58892 L .686 .645 0 r
F
.08929 .59203 m .08314 .61758 L .04252 .60934 L .682 .625 0 r
F
.09621 .57404 m .04449 .58892 L .05044 .56631 L .696 .633 0 r
F
.09621 .57404 m .08929 .59203 L .04449 .58892 L .693 .621 0 r
F
.04638 .48378 m .05063 .43836 L .05665 .44037 L 0 0 0 r
F
.05665 .44037 m .05717 .48762 L .04638 .48378 L .218 .193 0 r
F
.06823 .75569 m .0499 .71925 L .06769 .70348 L .215 .281 0 r
F
.10017 .54669 m .05044 .56631 L .05086 .54049 L .709 .628 0 r
F
.10017 .54669 m .09621 .57404 L .05044 .56631 L .704 .613 0 r
F
.05063 .43836 m .05388 .39312 L .06043 .39409 L 0 0 0 r
F
.06043 .39409 m .05665 .44037 L .05063 .43836 L F
.10935 .49335 m .05086 .54049 L .05717 .48762 L .715 .604 0 r
F
.10935 .49335 m .10017 .54669 L .05086 .54049 L F
.09461 .66536 m .0529 .66333 L .06407 .65017 L .585 .655 0 r
F
.09461 .66536 m .07971 .67902 L .0529 .66333 L .602 .621 0 r
F
.06884 .35229 m .06043 .39409 L .05388 .39312 L 0 0 0 r
F
.05388 .39312 m .06014 .35698 L .06884 .35229 L .071 0 0 r
F
.11485 .40002 m .05665 .44037 L .06043 .39409 L .73 .608 0 r
F
.11485 .40002 m .1107 .4472 L .05665 .44037 L .723 .593 0 r
F
.1107 .4472 m .05717 .48762 L .05665 .44037 L .721 .602 0 r
F
.1107 .4472 m .10935 .49335 L .05717 .48762 L .717 .594 0 r
F
.07574 .31421 m .06884 .35229 L .06014 .35698 L .051 0 0 r
F
.06014 .35698 m .06765 .31704 L .07574 .31421 L .046 0 0 r
F
.12235 .35773 m .06043 .39409 L .06884 .35229 L .73 .577 0 r
F
.12235 .35773 m .11485 .40002 L .06043 .39409 L .733 .584 0 r
F
.10884 .64997 m .06407 .65017 L .07304 .6362 L .641 .647 0 r
F
.10884 .64997 m .09461 .66536 L .06407 .65017 L .642 .646 0 r
F
.0853 .29025 m .07574 .31421 L .06765 .31704 L 0 0 0 r
F
.06765 .31704 m .07342 .29422 L .0853 .29025 L .214 .064 0 r
F
.16913 .68552 m .06769 .70348 L .1358 .68232 L .761 .678 0 r
F
.16913 .68552 m .11684 .69895 L .06769 .70348 L .732 .524 0 r
F
.11684 .69895 m .08095 .74271 L .06769 .70348 L .739 .559 0 r
F
.06769 .70348 m .08095 .74271 L .06823 .75569 L .086 .167 0 r
F
.09269 .80076 m .06823 .75569 L .08095 .74271 L .137 .251 0 r
F
.12786 .31972 m .06884 .35229 L .07574 .31421 L .733 .58 0 r
F
.12786 .31972 m .12235 .35773 L .06884 .35229 L .731 .574 0 r
F
.12459 .62588 m .07304 .6362 L .08314 .61758 L .645 .649 0 r
F
.12459 .62588 m .10884 .64997 L .07304 .6362 L .645 .645 0 r
F
.09255 .2674 m .0853 .29025 L .07342 .29422 L .2 .028 0 r
F
.07342 .29422 m .08173 .27179 L .09255 .2674 L .215 .046 0 r
F
.13455 .29921 m .07574 .31421 L .0853 .29025 L .726 .556 0 r
F
.13455 .29921 m .12786 .31972 L .07574 .31421 L .733 .582 0 r
F
.12555 .68062 m .07971 .67902 L .09461 .66536 L .608 .632 0 r
F
.11435 .68898 m .10056 .69327 L .07971 .67902 L .612 .622 0 r
F
.12555 .68062 m .11435 .68898 L .07971 .67902 L F
.11684 .69895 m .12738 .74126 L .08095 .74271 L .738 .557 0 r
F
.12738 .74126 m .10415 .78952 L .08095 .74271 L .737 .551 0 r
F
.08095 .74271 m .10415 .78952 L .09269 .80076 L .13 .245 0 r
F
.10466 .23743 m .09255 .2674 L .08173 .27179 L .197 .009 0 r
F
.08173 .27179 m .0956 .24051 L .10466 .23743 L .101 0 0 r
F
.13561 .5992 m .08314 .61758 L .08929 .59203 L .7 .644 0 r
F
.13561 .5992 m .12459 .62588 L .08314 .61758 L .685 .585 0 r
F
.14105 .27391 m .0853 .29025 L .09255 .2674 L .731 .558 0 r
F
.14105 .27391 m .13455 .29921 L .0853 .29025 L .727 .545 0 r
F
.14374 .5792 m .08929 .59203 L .09621 .57404 L .7 .629 0 r
F
.14374 .5792 m .13561 .5992 L .08929 .59203 L .702 .64 0 r
F
.15015 .25046 m .09255 .2674 L .10466 .23743 L .725 .563 0 r
F
.15015 .25046 m .14105 .27391 L .09255 .2674 L .73 .576 0 r
F
.12031 .85548 m .09269 .80076 L .10415 .78952 L .127 .24 0 r
F
.1446 .66235 m .09461 .66536 L .10884 .64997 L .627 .622 0 r
F
.1446 .66235 m .12555 .68062 L .09461 .66536 L .627 .627 0 r
F
.11723 .21607 m .10466 .23743 L .0956 .24051 L .076 0 0 r
F
.0956 .24051 m .10749 .21954 L .11723 .21607 L .026 0 0 r
F
.15007 .5534 m .09621 .57404 L .10017 .54669 L .71 .618 0 r
F
.15007 .5534 m .14374 .5792 L .09621 .57404 L .706 .607 0 r
F
.16094 .49754 m .10017 .54669 L .10935 .49335 L .722 .61 0 r
F
.16094 .49754 m .15007 .5534 L .10017 .54669 L .715 .597 0 r
F
.12738 .74126 m .14571 .78748 L .10415 .78952 L .738 .552 0 r
F
.14571 .78748 m .13227 .84208 L .10415 .78952 L .736 .54 0 r
F
.10415 .78952 m .13227 .84208 L .12031 .85548 L .179 .279 0 r
F
.16005 .22918 m .10466 .23743 L .11723 .21607 L .728 .576 0 r
F
.16005 .22918 m .15015 .25046 L .10466 .23743 L .725 .563 0 r
F
.14069 .18663 m .11723 .21607 L .10749 .21954 L 0 0 0 r
F
.10749 .21954 m .13294 .18927 L .14069 .18663 L F
.16622 .63352 m .10884 .64997 L .12459 .62588 L .672 .683 0 r
F
.16622 .63352 m .1446 .66235 L .10884 .64997 L .659 .598 0 r
F
.16528 .45457 m .10935 .49335 L .1107 .4472 L .726 .602 0 r
F
.16528 .45457 m .16094 .49754 L .10935 .49335 L .723 .596 0 r
F
.16894 .40647 m .1107 .4472 L .11485 .40002 L .733 .602 0 r
F
.16894 .40647 m .16528 .45457 L .1107 .4472 L .728 .593 0 r
F
.11435 .68898 m .12555 .68062 L .12576 .68544 L .198 0 0 r
F
.17471 .36165 m .11485 .40002 L .12235 .35773 L .734 .585 0 r
F
.17471 .36165 m .16894 .40647 L .11485 .40002 L .735 .586 0 r
F
.20068 .68826 m .11684 .69895 L .16913 .68552 L .762 .631 0 r
F
.20068 .68826 m .16443 .69936 L .11684 .69895 L .766 .678 0 r
F
.16443 .69936 m .12738 .74126 L .11684 .69895 L .767 .574 0 r
F
.17771 .20836 m .11723 .21607 L .14069 .18663 L .742 .569 0 r
F
.17771 .20836 m .16005 .22918 L .11723 .21607 L .738 .55 0 r
F
.1412 .89345 m .12031 .85548 L .13227 .84208 L .189 .294 0 r
F
.17997 .32114 m .12235 .35773 L .12786 .31972 L .735 .579 0 r
F
.17997 .32114 m .17471 .36165 L .12235 .35773 L .735 .578 0 r
F
.17994 .60847 m .12459 .62588 L .13561 .5992 L .7 .603 0 r
F
.17994 .60847 m .16622 .63352 L .12459 .62588 L .706 .631 0 r
F
.1546 .67101 m .12555 .68062 L .1446 .66235 L .251 .059 0 r
F
.12576 .68544 m .12555 .68062 L .15306 .67152 L .198 0 0 r
F
.12576 .68544 m .15306 .67152 L .12636 .68525 L F
.12636 .68525 m .15306 .67152 L .15269 .67197 L F
.15269 .67197 m .1358 .68232 L .12636 .68525 L 0 0 0 r
F
.16443 .69936 m .1731 .74371 L .12738 .74126 L .762 .567 0 r
F
.1731 .74371 m .14571 .78748 L .12738 .74126 L .762 .563 0 r
F
.18384 .30893 m .12786 .31972 L .13455 .29921 L .74 .59 0 r
F
.18384 .30893 m .17997 .32114 L .12786 .31972 L .735 .567 0 r
F
.14571 .78748 m .16568 .84415 L .13227 .84208 L .749 .554 0 r
F
.13227 .84208 m .15357 .88187 L .1412 .89345 L .191 .295 0 r
F
.16568 .84415 m .15357 .88187 L .13227 .84208 L .75 .559 0 r
F
.15791 .17371 m .14069 .18663 L .13294 .18927 L 0 0 0 r
F
.13294 .18927 m .14942 .17557 L .15791 .17371 L F
.19041 .27836 m .13455 .29921 L .14105 .27391 L .73 .549 0 r
F
.19041 .27836 m .18384 .30893 L .13455 .29921 L .742 .581 0 r
F
.19105 .5802 m .13561 .5992 L .14374 .5792 L .713 .653 0 r
F
.19105 .5802 m .17994 .60847 L .13561 .5992 L .7 .602 0 r
F
.1546 .67101 m .1358 .68232 L .15269 .67197 L 0 0 0 r
F
.16913 .68552 m .1358 .68232 L .1546 .67101 L .751 .705 0 r
F
.191 .19682 m .14069 .18663 L .15791 .17371 L .734 .569 0 r
F
.191 .19682 m .17771 .20836 L .14069 .18663 L .732 .577 0 r
F
.19477 .2661 m .14105 .27391 L .15015 .25046 L .728 .574 0 r
F
.19477 .2661 m .19041 .27836 L .14105 .27391 L .729 .581 0 r
F
.14781 .90795 m .1412 .89345 L .15357 .88187 L .147 .222 0 r
F
.19911 .56063 m .14374 .5792 L .15007 .5534 L .708 .608 0 r
F
.19911 .56063 m .19105 .5802 L .14374 .5792 L .715 .639 0 r
F
.1446 .66235 m .16622 .63352 L .17489 .64386 L .332 .167 0 r
F
.17489 .64386 m .1546 .67101 L .1446 .66235 L .267 .058 0 r
F
.1731 .74371 m .18627 .78686 L .14571 .78748 L .75 .547 0 r
F
.18627 .78686 m .16568 .84415 L .14571 .78748 L .751 .555 0 r
F
.15357 .88187 m .16308 .89895 L .14781 .90795 L .249 .329 0 r
F
.16072 .93285 m .14781 .90795 L .16308 .89895 L .244 .317 0 r
F
.17168 .16032 m .15791 .17371 L .14942 .17557 L 0 0 0 r
F
.21297 .50014 m .15007 .5534 L .16094 .49754 L .711 .593 0 r
F
.21297 .50014 m .19911 .56063 L .15007 .5534 L .711 .595 0 r
F
.20218 .24423 m .15015 .25046 L .16005 .22918 L .727 .565 0 r
F
.20218 .24423 m .19477 .2661 L .15015 .25046 L .729 .573 0 r
F
.15269 .67197 m .15306 .67152 L .1546 .67101 L .198 0 0 r
F
.16568 .84415 m .17559 .88088 L .15357 .88187 L .75 .56 0 r
F
.17559 .88088 m .16308 .89895 L .15357 .88187 L .749 .552 0 r
F
.22855 .64706 m .1546 .67101 L .17489 .64386 L .75 .69 0 r
F
.22855 .64706 m .22286 .67759 L .1546 .67101 L .748 .675 0 r
F
.22286 .67759 m .16913 .68552 L .1546 .67101 L .736 .704 0 r
F
.20233 .18471 m .15791 .17371 L .17168 .16032 L .73 .555 0 r
F
.20233 .18471 m .191 .19682 L .15791 .17371 L .724 .574 0 r
F
.21381 .22967 m .16005 .22918 L .17771 .20836 L .728 .534 0 r
F
.21381 .22967 m .20218 .24423 L .16005 .22918 L .73 .56 0 r
F
.16308 .89895 m .17462 .92189 L .16072 .93285 L .24 .313 0 r
F
.17132 .95141 m .16072 .93285 L .17462 .92189 L .253 .337 0 r
F
.21928 .45927 m .16094 .49754 L .16528 .45457 L .714 .588 0 r
F
.21928 .45927 m .21297 .50014 L .16094 .49754 L .712 .582 0 r
F
.17559 .88088 m .21236 .89943 L .16308 .89895 L .753 .557 0 r
F
.21236 .89943 m .17462 .92189 L .16308 .89895 L .753 .56 0 r
F
.23325 .69003 m .16443 .69936 L .20068 .68826 L .751 .616 0 r
F
.23325 .69003 m .21346 .70081 L .16443 .69936 L .756 .697 0 r
F
.21346 .70081 m .1731 .74371 L .16443 .69936 L .764 .568 0 r
F
.22271 .40815 m .16528 .45457 L .16894 .40647 L .723 .588 0 r
F
.22271 .40815 m .21928 .45927 L .16528 .45457 L .716 .575 0 r
F
.18627 .78686 m .19927 .845 L .16568 .84415 L .76 .565 0 r
F
.19927 .845 m .17559 .88088 L .16568 .84415 L F
.16622 .63352 m .17994 .60847 L .18866 .61915 L .368 .226 0 r
F
.18866 .61915 m .17489 .64386 L .16622 .63352 L .334 .167 0 r
F
.22709 .36288 m .16894 .40647 L .17471 .36165 L .725 .577 0 r
F
.22709 .36288 m .22271 .40815 L .16894 .40647 L .723 .573 0 r
F
.20068 .68826 m .16913 .68552 L .22286 .67759 L .735 .72 0 r
F
.17462 .92189 m .18319 .94065 L .17132 .95141 L .168 .255 0 r
F
.18697 .98192 m .17132 .95141 L .18319 .94065 L .179 .274 0 r
F
.22721 .17145 m .17168 .16032 L .20032 .14407 L .729 .561 0 r
F
.22721 .17145 m .20233 .18471 L .17168 .16032 L .73 .555 0 r
F
.21346 .70081 m .21885 .74404 L .1731 .74371 L .761 .563 0 r
F
.21885 .74404 m .18627 .78686 L .1731 .74371 L .76 .553 0 r
F
.21236 .89943 m .2132 .92002 L .17462 .92189 L .743 .539 0 r
F
.2132 .92002 m .18319 .94065 L .17462 .92189 L .745 .554 0 r
F
.23192 .32299 m .17471 .36165 L .17997 .32114 L .729 .573 0 r
F
.23192 .32299 m .22709 .36288 L .17471 .36165 L .725 .564 0 r
F
.2466 .62247 m .17489 .64386 L .18866 .61915 L .749 .665 0 r
F
.2466 .62247 m .22855 .64706 L .17489 .64386 L .752 .684 0 r
F
.19927 .845 m .19804 .87988 L .17559 .88088 L .75 .551 0 r
F
.19804 .87988 m .21236 .89943 L .17559 .88088 L .751 .559 0 r
F
.17633 .15768 m .1911 .14814 L .17959 .15583 L 0 0 0 r
F
.22173 .22217 m .17771 .20836 L .191 .19682 L .716 .549 0 r
F
.22173 .22217 m .21381 .22967 L .17771 .20836 L .719 .542 0 r
F
.17959 .15583 m .1911 .14814 L .20032 .14407 L 0 0 0 r
F
.17994 .60847 m .19105 .5802 L .20047 .58867 L .376 .249 0 r
F
.20047 .58867 m .18866 .61915 L .17994 .60847 L .359 .224 0 r
F
.23192 .32299 m .17997 .32114 L .18384 .30893 L .729 .561 0 r
F
.2132 .92002 m .23533 .94048 L .18319 .94065 L .751 .565 0 r
F
.23533 .94048 m .20149 .96974 L .18319 .94065 L .749 .545 0 r
F
.18319 .94065 m .20149 .96974 L .18697 .98192 L .288 .364 0 r
F
.23941 .28358 m .18384 .30893 L .19041 .27836 L .715 .553 0 r
F
.23941 .28358 m .23192 .32299 L .18384 .30893 L .723 .575 0 r
F
.21885 .74404 m .22657 .78912 L .18627 .78686 L .771 .569 0 r
F
.22657 .78912 m .19927 .845 L .18627 .78686 L .771 .572 0 r
F
.21107 1.03024 m .18697 .98192 L .20149 .96974 L .261 .317 0 r
F
.25388 .59108 m .18866 .61915 L .20047 .58867 L .751 .651 0 r
F
.25388 .59108 m .2466 .62247 L .18866 .61915 L .752 .652 0 r
F
.23941 .28358 m .19041 .27836 L .19477 .2661 L .714 .564 0 r
F
.23267 .20922 m .191 .19682 L .20233 .18471 L .719 .566 0 r
F
.23267 .20922 m .22173 .22217 L .191 .19682 L .727 .546 0 r
F
.19105 .5802 m .19911 .56063 L .20756 .55347 L .097 0 0 r
F
.20756 .55347 m .20047 .58867 L .19105 .5802 L .33 .248 0 r
F
.24462 .26075 m .19477 .2661 L .20218 .24423 L .722 .566 0 r
F
.24462 .26075 m .23941 .28358 L .19477 .2661 L .72 .555 0 r
F
.23239 .84589 m .19804 .87988 L .19927 .845 L .756 .557 0 r
F
.23239 .84589 m .2206 .88048 L .19804 .87988 L .75 .547 0 r
F
.2206 .88048 m .21236 .89943 L .19804 .87988 L .751 .559 0 r
F
.19911 .56063 m .21297 .50014 L .21946 .50209 L .267 .164 0 r
F
.21946 .50209 m .20756 .55347 L .19911 .56063 L .137 .032 0 r
F
.22657 .78912 m .23239 .84589 L .19927 .845 L .756 .552 0 r
F
.24463 .16158 m .20032 .14407 L .2224 .13448 L .734 .548 0 r
F
.24463 .16158 m .22721 .17145 L .20032 .14407 L .727 .561 0 r
F
.25949 .55806 m .20047 .58867 L .20756 .55347 L .76 .662 0 r
F
.25949 .55806 m .25388 .59108 L .20047 .58867 L .753 .642 0 r
F
.23325 .69003 m .20068 .68826 L .22286 .67759 L .731 .706 0 r
F
.23533 .94048 m .24641 .97126 L .20149 .96974 L .751 .549 0 r
F
.20149 .96974 m .22418 1.01743 L .21107 1.03024 L .225 .289 0 r
F
.24641 .97126 m .22418 1.01743 L .20149 .96974 L .751 .55 0 r
F
.25252 .24756 m .20218 .24423 L .21381 .22967 L .722 .549 0 r
F
.25252 .24756 m .24462 .26075 L .20218 .24423 L .722 .566 0 r
F
.25143 .19924 m .20233 .18471 L .22721 .17145 L .729 .552 0 r
F
.25143 .19924 m .23267 .20922 L .20233 .18471 L .724 .564 0 r
F
.26679 .50376 m .20756 .55347 L .21946 .50209 L .765 .637 0 r
F
.26679 .50376 m .25949 .55806 L .20756 .55347 L .765 .636 0 r
F
.23575 1.07441 m .21107 1.03024 L .22418 1.01743 L .249 .329 0 r
F
.2206 .88048 m .26201 .89941 L .21236 .89943 L .751 .559 0 r
F
.26201 .89941 m .2132 .92002 L .21236 .89943 L .749 .544 0 r
F
.21297 .50014 m .21928 .45927 L .22604 .46065 L .402 .296 0 r
F
.22604 .46065 m .21946 .50209 L .21297 .50014 L .269 .161 0 r
F
.26201 .89941 m .25127 .92119 L .2132 .92002 L .756 .56 0 r
F
.25127 .92119 m .23533 .94048 L .2132 .92002 L .756 .564 0 r
F
.26655 .69234 m .21346 .70081 L .23325 .69003 L .74 .651 0 r
F
.26655 .69234 m .26218 .70385 L .21346 .70081 L .741 .707 0 r
F
.26218 .70385 m .21885 .74404 L .21346 .70081 L .76 .563 0 r
F
.25252 .24756 m .21381 .22967 L .22173 .22217 L .723 .547 0 r
F
.26218 .70385 m .26452 .74551 L .21885 .74404 L .753 .551 0 r
F
.26452 .74551 m .22657 .78912 L .21885 .74404 L .754 .558 0 r
F
.21928 .45927 m .22271 .40815 L .22964 .4068 L .376 .283 0 r
F
.22964 .4068 m .22604 .46065 L .21928 .45927 L .399 .305 0 r
F
.27207 .46138 m .21946 .50209 L .22604 .46065 L .756 .608 0 r
F
.27207 .46138 m .26679 .50376 L .21946 .50209 L .766 .627 0 r
F
.26506 .84632 m .2206 .88048 L .23239 .84589 L .751 .548 0 r
F
.26506 .84632 m .24329 .8806 L .2206 .88048 L .743 .533 0 r
F
.24329 .8806 m .26201 .89941 L .2206 .88048 L .746 .566 0 r
F
.26216 .235 m .22173 .22217 L .23267 .20922 L .723 .541 0 r
F
.26216 .235 m .25252 .24756 L .22173 .22217 L .721 .548 0 r
F
.26634 .15446 m .2224 .13448 L .24887 .12472 L .742 .538 0 r
F
.26634 .15446 m .24463 .16158 L .2224 .13448 L .736 .548 0 r
F
.23413 .36258 m .22964 .4068 L .22271 .40815 L .367 .249 0 r
F
.22271 .40815 m .22709 .36288 L .23413 .36258 L .48 .359 0 r
F
.28203 .65132 m .22286 .67759 L .22855 .64706 L .745 .672 0 r
F
.28203 .65132 m .29205 .67859 L .22286 .67759 L .745 .673 0 r
F
.29205 .67859 m .23325 .69003 L .22286 .67759 L .736 .707 0 r
F
.24641 .97126 m .2583 1.01677 L .22418 1.01743 L .733 .529 0 r
F
.2583 1.01677 m .25016 1.06215 L .22418 1.01743 L .735 .541 0 r
F
.22418 1.01743 m .25016 1.06215 L .23575 1.07441 L .257 .335 0 r
F
.2725 .41029 m .22604 .46065 L .22964 .4068 L .764 .62 0 r
F
.2725 .41029 m .27207 .46138 L .22604 .46065 L .756 .607 0 r
F
.26452 .74551 m .26754 .7906 L .22657 .78912 L .75 .552 0 r
F
.26754 .7906 m .23239 .84589 L .22657 .78912 L .75 .548 0 r
F
.23784 .32202 m .23413 .36258 L .22709 .36288 L .48 .356 0 r
F
.22709 .36288 m .23192 .32299 L .23784 .32202 L .363 .242 0 r
F
.26765 .1905 m .22721 .17145 L .24463 .16158 L .73 .565 0 r
F
.26765 .1905 m .25143 .19924 L .22721 .17145 L .739 .552 0 r
F
.30359 .62727 m .22855 .64706 L .2466 .62247 L .749 .68 0 r
F
.30359 .62727 m .28203 .65132 L .22855 .64706 L .747 .663 0 r
F
.27705 .36599 m .22964 .4068 L .23413 .36258 L .743 .58 0 r
F
.27705 .36599 m .2725 .41029 L .22964 .4068 L .764 .618 0 r
F
.23192 .32299 m .23941 .28358 L .24287 .2805 L .084 0 0 r
F
.24287 .2805 m .23784 .32202 L .23192 .32299 L .366 .253 0 r
F
.26754 .7906 m .26506 .84632 L .23239 .84589 L .751 .55 0 r
F
.27719 .22649 m .23267 .20922 L .25143 .19924 L .715 .544 0 r
F
.27719 .22649 m .26216 .235 L .23267 .20922 L .716 .543 0 r
F
.26655 .69234 m .23325 .69003 L .29205 .67859 L .735 .671 0 r
F
.27845 .32284 m .27705 .36599 L .23413 .36258 L .742 .592 0 r
F
.27845 .32284 m .23413 .36258 L .23784 .32202 L .735 .58 0 r
F
.25127 .92119 m .28635 .93779 L .23533 .94048 L .743 .545 0 r
F
.28635 .93779 m .24641 .97126 L .23533 .94048 L F
.24446 1.09588 m .23575 1.07441 L .25016 1.06215 L .205 .244 0 r
F
.23784 .32202 m .24287 .2805 L .27845 .32284 L .735 .584 0 r
F
.25351 .25962 m .24287 .2805 L .23941 .28358 L 0 0 0 r
F
.23941 .28358 m .24462 .26075 L .25351 .25962 L .464 .306 0 r
F
.35541 .28203 m .24287 .2805 L .25351 .25962 L .763 .616 0 r
F
.24287 .2805 m .35541 .28203 L .31929 .32384 L .761 .575 0 r
F
.31929 .32384 m .27845 .32284 L .24287 .2805 L .752 .585 0 r
F
.29857 .84551 m .26592 .88124 L .24329 .8806 L .747 .555 0 r
F
.29857 .84551 m .24329 .8806 L .26506 .84632 L .719 .503 0 r
F
.26592 .88124 m .26201 .89941 L .24329 .8806 L .748 .566 0 r
F
.25016 1.06215 m .2583 1.08271 L .24446 1.09588 L .193 .233 0 r
F
.25382 1.11304 m .24446 1.09588 L .2583 1.08271 L .234 .299 0 r
F
.24462 .26075 m .25252 .24756 L .26107 .24802 L .498 .293 0 r
F
.26107 .24802 m .25351 .25962 L .24462 .26075 L .433 .194 0 r
F
.28347 .18506 m .24463 .16158 L .26634 .15446 L .738 .554 0 r
F
.28347 .18506 m .26765 .1905 L .24463 .16158 L .724 .565 0 r
F
.28635 .93779 m .29085 .97146 L .24641 .97126 L .738 .537 0 r
F
.29085 .97146 m .2583 1.01677 L .24641 .97126 L .737 .531 0 r
F
.30785 .59602 m .2466 .62247 L .25388 .59108 L .756 .656 0 r
F
.30785 .59602 m .30359 .62727 L .2466 .62247 L F
.28858 .14647 m .24887 .12472 L .27426 .11649 L .741 .531 0 r
F
.28858 .14647 m .26634 .15446 L .24887 .12472 L .737 .537 0 r
F
.2583 1.01677 m .26881 1.05847 L .25016 1.06215 L .732 .538 0 r
F
.26881 1.05847 m .2583 1.08271 L .25016 1.06215 L .728 .525 0 r
F
.30936 .89899 m .25127 .92119 L .26201 .89941 L .721 .518 0 r
F
.30943 .89961 m .25127 .92119 L .30936 .89899 L F
.30943 .89961 m .29 .91932 L .25127 .92119 L .751 .6 0 r
F
.29 .91932 m .28635 .93779 L .25127 .92119 L .744 .543 0 r
F
.28752 .21931 m .25143 .19924 L .26765 .1905 L .742 .558 0 r
F
.28752 .21931 m .27719 .22649 L .25143 .19924 L .757 .54 0 r
F
.25252 .24756 m .26216 .235 L .27006 .23545 L .485 .24 0 r
F
.27006 .23545 m .26107 .24802 L .25252 .24756 L .494 .254 0 r
F
.35541 .28203 m .25351 .25962 L .26107 .24802 L .765 .607 0 r
F
.2583 1.08271 m .2685 1.09817 L .25382 1.11304 L .247 .312 0 r
F
.25946 1.12519 m .25382 1.11304 L .2685 1.09817 L .224 .275 0 r
F
.31145 .56056 m .25388 .59108 L .25949 .55806 L .754 .643 0 r
F
.31145 .56056 m .30785 .59602 L .25388 .59108 L .757 .651 0 r
F
.29085 .97146 m .29226 1.01486 L .2583 1.01677 L .74 .535 0 r
F
.29226 1.01486 m .26881 1.05847 L .2583 1.01677 L .741 .544 0 r
F
.26881 1.05847 m .28292 1.07927 L .2583 1.08271 L .703 .496 0 r
F
.2685 1.09817 m .2583 1.08271 L .28292 1.07927 L .698 .478 0 r
F
.2685 1.09817 m .26961 1.11882 L .25946 1.12519 L .063 .096 0 r
F
.25946 1.12519 m .26961 1.11882 L .28007 1.12819 L .266 .527 0 r
F
.25946 1.12519 m .28007 1.12819 L .27278 1.13287 L .214 .555 0 r
F
.31404 .50644 m .25949 .55806 L .26679 .50376 L .756 .628 0 r
F
.31404 .50644 m .31145 .56056 L .25949 .55806 L .755 .627 0 r
F
.35541 .28203 m .26107 .24802 L .27006 .23545 L .765 .607 0 r
F
.26181 .12052 m .26651 .11826 L .26256 .12028 L 0 0 0 r
F
.26592 .88124 m .30927 .89824 L .26201 .89941 L .724 .542 0 r
F
.30927 .89824 m .30936 .89899 L .26201 .89941 L F
.27006 .23545 m .26216 .235 L .27719 .22649 L .456 .105 0 r
F
.29943 .69313 m .26218 .70385 L .26655 .69234 L .729 .693 0 r
F
.29943 .69313 m .31007 .70514 L .26218 .70385 L .724 .658 0 r
F
.31007 .70514 m .26452 .74551 L .26218 .70385 L .729 .532 0 r
F
.26256 .12028 m .26651 .11826 L .26273 .12023 L 0 0 0 r
F
.26273 .12023 m .26651 .11826 L .27426 .11649 L F
.31007 .70514 m .30995 .74567 L .26452 .74551 L .733 .54 0 r
F
.30995 .74567 m .26754 .7906 L .26452 .74551 L .733 .539 0 r
F
.30864 .79077 m .26506 .84632 L .26754 .7906 L .729 .531 0 r
F
.30864 .79077 m .29857 .84551 L .26506 .84632 L .723 .522 0 r
F
.28801 .88086 m .26592 .88124 L .29857 .84551 L .742 .549 0 r
F
.28801 .88086 m .30918 .89743 L .26592 .88124 L .736 .512 0 r
F
.30918 .89743 m .30927 .89824 L .26592 .88124 L F
.30238 .17893 m .26634 .15446 L .28858 .14647 L .744 .555 0 r
F
.30238 .17893 m .28347 .18506 L .26634 .15446 L .744 .556 0 r
F
.29943 .69313 m .26655 .69234 L .29205 .67859 L .735 .67 0 r
F
.317 .46366 m .26679 .50376 L .27207 .46138 L .76 .622 0 r
F
.317 .46366 m .31404 .50644 L .26679 .50376 L .757 .616 0 r
F
.30995 .74567 m .30864 .79077 L .26754 .7906 L .73 .534 0 r
F
.30155 .21593 m .26765 .1905 L .28347 .18506 L .722 .56 0 r
F
.30155 .21593 m .28752 .21931 L .26765 .1905 L .731 .556 0 r
F
.30977 1.06988 m .29555 1.09179 L .2685 1.09817 L .295 0 0 r
F
.29327 1.07704 m .30977 1.06988 L .2685 1.09817 L .113 0 0 r
F
.29327 1.07704 m .2685 1.09817 L .28292 1.07927 L .576 .319 0 r
F
.28016 1.11363 m .2685 1.09817 L .29555 1.09179 L .436 .337 0 r
F
.2685 1.09817 m .28016 1.11363 L .26961 1.11882 L .313 .299 0 r
F
.29226 1.01486 m .28836 1.05827 L .26881 1.05847 L .717 .513 0 r
F
.28836 1.05827 m .28292 1.07927 L .26881 1.05847 L .715 .498 0 r
F
.26961 1.11882 m .28016 1.11363 L .29507 1.11948 L .458 .7 0 r
F
.26961 1.11882 m .29507 1.11948 L .28007 1.12819 L .529 .626 0 r
F
.35541 .28203 m .27006 .23545 L .28282 .22642 L .754 .617 0 r
F
.27719 .22649 m .28282 .22642 L .27006 .23545 L .34 0 0 r
F
.31565 .41372 m .27207 .46138 L .2725 .41029 L .757 .608 0 r
F
.31565 .41372 m .317 .46366 L .27207 .46138 L .76 .613 0 r
F
.31823 .36693 m .2725 .41029 L .27705 .36599 L .755 .609 0 r
F
.31823 .36693 m .31565 .41372 L .2725 .41029 L .757 .613 0 r
F
.27278 1.13287 m .28007 1.12819 L .29446 1.13533 L .232 .589 0 r
F
.27278 1.13287 m .29446 1.13533 L .28652 1.13973 L .22 .592 0 r
F
.31254 .14358 m .27426 .11649 L .30286 .11048 L .75 .53 0 r
F
.31254 .14358 m .28858 .14647 L .27426 .11649 L .746 .533 0 r
F
.31929 .32384 m .27705 .36599 L .27845 .32284 L .752 .6 0 r
F
.31929 .32384 m .31823 .36693 L .27705 .36599 L .755 .605 0 r
F
.28282 .22642 m .27719 .22649 L .28752 .21931 L .314 0 0 r
F
.28007 1.12819 m .29507 1.11948 L .30942 1.12578 L .539 .656 0 r
F
.28007 1.12819 m .30942 1.12578 L .29446 1.13533 L .539 .657 0 r
F
.29555 1.09179 m .30077 1.10031 L .28016 1.11363 L .427 .326 0 r
F
.28016 1.11363 m .30077 1.10031 L .31444 1.10637 L .596 .68 0 r
F
.28016 1.11363 m .31444 1.10637 L .29507 1.11948 L .593 .705 0 r
F
.36061 .62487 m .28203 .65132 L .30359 .62727 L .759 .684 0 r
F
.36061 .62487 m .33583 .65147 L .28203 .65132 L .752 .65 0 r
F
.33583 .65147 m .29205 .67859 L .28203 .65132 L .748 .674 0 r
F
.28752 .21931 m .29431 .22025 L .28282 .22642 L .44 .029 0 r
F
.35541 .28203 m .28282 .22642 L .29431 .22025 L .787 .603 0 r
F
.30696 1.05466 m .28292 1.07927 L .28836 1.05827 L .667 .45 0 r
F
.30696 1.05466 m .29327 1.07704 L .28292 1.07927 L .578 .325 0 r
F
.31495 .2128 m .28347 .18506 L .30238 .17893 L .752 .579 0 r
F
.31495 .2128 m .30155 .21593 L .28347 .18506 L .77 .574 0 r
F
.29 .91932 m .31744 .92917 L .28635 .93779 L .701 .501 0 r
F
.31744 .92917 m .31963 .93133 L .28635 .93779 L F
.31963 .93133 m .32293 .93572 L .28635 .93779 L F
.32293 .93572 m .32302 .93584 L .28635 .93779 L F
.32302 .93584 m .3239 .93702 L .28635 .93779 L F
.32296 .93704 m .29085 .97146 L .28635 .93779 L .703 .512 0 r
F
.28652 1.13973 m .29446 1.13533 L .3099 1.14108 L .235 .624 0 r
F
.28652 1.13973 m .3099 1.14108 L .30061 1.14674 L .273 .607 0 r
F
.28752 .21931 m .30155 .21593 L .30829 .21751 L .532 .122 0 r
F
.30829 .21751 m .29431 .22025 L .28752 .21931 L .378 0 0 r
F
.29857 .84551 m .31259 .85232 L .28801 .88086 L .696 .499 0 r
F
.31259 .85232 m .31258 .85245 L .28801 .88086 L F
.31258 .85245 m .31248 .8533 L .28801 .88086 L F
.31248 .8533 m .31163 .86096 L .28801 .88086 L F
.31163 .86096 m .3098 .8773 L .28801 .88086 L .719 .538 0 r
F
.28801 .88086 m .30856 .87751 L .30924 .8945 L .715 .522 0 r
F
.28801 .88086 m .30924 .8945 L .30919 .89603 L F
.28801 .88086 m .30919 .89603 L .30889 .8972 L F
.32526 1.00714 m .28836 1.05827 L .29226 1.01486 L .7 .498 0 r
F
.32526 1.00714 m .30696 1.05466 L .28836 1.05827 L .669 .457 0 r
F
.32229 .17806 m .28858 .14647 L .31254 .14358 L .754 .571 0 r
F
.32229 .17806 m .30238 .17893 L .28858 .14647 L .771 .567 0 r
F
.30954 .90053 m .29 .91932 L .30943 .89961 L .751 .6 0 r
F
.29 .91932 m .31272 .91723 L .31674 .92544 L .703 .498 0 r
F
.29055 .91879 m .31271 .91721 L .29 .91932 L .71 .532 0 r
F
.31271 .91721 m .31271 .91722 L .29 .91932 L F
.31271 .91722 m .31272 .91723 L .29 .91932 L F
.29 .91932 m .31717 .92632 L .31768 .92736 L .703 .498 0 r
F
.29 .91932 m .31674 .92544 L .31717 .92632 L F
.29 .91932 m .31768 .92736 L .31839 .92881 L F
.29 .91932 m .31858 .9292 L .31881 .92967 L F
.29 .91932 m .31839 .92881 L .31858 .9292 L F
.30954 .90053 m .3111 .91392 L .29055 .91879 L .71 .532 0 r
F
.3111 .91392 m .31271 .91721 L .29055 .91879 L F
.33158 .94172 m .29085 .97146 L .32162 .93847 L .703 .512 0 r
F
.33158 .94172 m .33573 .9671 L .29085 .97146 L F
.33573 .9671 m .29226 1.01486 L .29085 .97146 L .702 .505 0 r
F
.33583 .65147 m .36077 .6784 L .29205 .67859 L .722 .611 0 r
F
.36077 .6784 m .29943 .69313 L .29205 .67859 L .716 .662 0 r
F
.33573 .9671 m .32526 1.00714 L .29226 1.01486 L .703 .507 0 r
F
.30696 1.05466 m .32108 1.04471 L .29327 1.07704 L .152 0 0 r
F
.32108 1.04471 m .30977 1.06988 L .29327 1.07704 L .282 .018 0 r
F
.35541 .28203 m .29431 .22025 L .30829 .21751 L .762 .606 0 r
F
.29446 1.13533 m .30942 1.12578 L .32463 1.1306 L .548 .679 0 r
F
.29446 1.13533 m .32463 1.1306 L .3099 1.14108 L .549 .672 0 r
F
.29507 1.11948 m .31444 1.10637 L .32779 1.11253 L .578 .667 0 r
F
.29507 1.11948 m .32779 1.11253 L .30942 1.12578 L .578 .651 0 r
F
.30977 1.06988 m .31075 1.0779 L .29555 1.09179 L 0 0 0 r
F
.30077 1.10031 m .29555 1.09179 L .31075 1.0779 L .429 .327 0 r
F
.33344 .78993 m .29857 .84551 L .30864 .79077 L .685 .486 0 r
F
.32534 .80813 m .29857 .84551 L .32367 .80551 L F
.32119 .81914 m .29857 .84551 L .32534 .80813 L F
.32053 .82126 m .29857 .84551 L .32077 .81963 L F
.32119 .81915 m .29857 .84551 L .32119 .81914 L F
.32053 .82126 m .31975 .82642 L .29857 .84551 L .675 .472 0 r
F
.31975 .82642 m .31866 .83356 L .29857 .84551 L F
.31866 .83356 m .3185 .83461 L .29857 .84551 L F
.31823 .83476 m .3184 .83528 L .29857 .84551 L F
.3184 .83528 m .31794 .8366 L .29857 .84551 L F
.31413 .83836 m .31675 .84002 L .29857 .84551 L F
.31675 .84002 m .31634 .8412 L .29857 .84551 L F
.31634 .8412 m .31563 .84325 L .29857 .84551 L F
.31563 .84325 m .31525 .84435 L .29857 .84551 L F
.31525 .84435 m .31326 .85005 L .29857 .84551 L .683 .514 0 r
F
.31112 .84939 m .31265 .85183 L .29857 .84551 L F
.31265 .85183 m .31259 .85232 L .29857 .84551 L F
.3183 .10309 m .30286 .11048 L .29934 .11122 L 0 0 0 r
F
.33192 .69293 m .29943 .69313 L .36077 .6784 L .719 .721 0 r
F
.33192 .69293 m .31007 .70514 L .29943 .69313 L .735 .659 0 r
F
.30061 1.14674 m .3099 1.14108 L .3269 1.14698 L .291 .645 0 r
F
.30061 1.14674 m .3269 1.14698 L .3142 1.15429 L .336 .609 0 r
F
.31075 1.0779 m .31545 1.08408 L .30077 1.10031 L .443 .343 0 r
F
.30077 1.10031 m .31545 1.08408 L .32893 1.09051 L .631 .623 0 r
F
.30077 1.10031 m .32893 1.09051 L .31444 1.10637 L .64 .671 0 r
F
.30155 .21593 m .31495 .2128 L .32144 .21653 L .534 .09 0 r
F
.32144 .21653 m .30829 .21751 L .30155 .21593 L .508 .037 0 r
F
.33443 .21201 m .30238 .17893 L .32229 .17806 L .771 .566 0 r
F
.30238 .17893 m .33443 .21201 L .31495 .2128 L .734 .57 0 r
F
.30286 .11048 m .31943 .10261 L .33495 .15845 L .744 .54 0 r
F
.31943 .10261 m .30286 .11048 L .3183 .10309 L 0 0 0 r
F
.33495 .15845 m .31254 .14358 L .30286 .11048 L .777 .545 0 r
F
.3601 .59678 m .30359 .62727 L .30785 .59602 L .771 .671 0 r
F
.3601 .59678 m .36061 .62487 L .30359 .62727 L .761 .643 0 r
F
.32526 1.00714 m .32952 1.01893 L .30696 1.05466 L .325 .104 0 r
F
.32952 1.01893 m .32108 1.04471 L .30696 1.05466 L .279 .053 0 r
F
.36015 .56299 m .30785 .59602 L .31145 .56056 L .788 .679 0 r
F
.36015 .56299 m .3601 .59678 L .30785 .59602 L .774 .644 0 r
F
.35541 .28203 m .30829 .21751 L .32144 .21653 L .718 .598 0 r
F
.30856 .87751 m .30972 .87806 L .30924 .8945 L .715 .522 0 r
F
.30856 .87751 m .3098 .8773 L .30972 .87806 L F
.35576 .74672 m .30864 .79077 L .30995 .74567 L .7 .509 0 r
F
.35562 .74765 m .30864 .79077 L .35576 .74672 L F
.32901 .77207 m .34716 .7875 L .30864 .79077 L .685 .485 0 r
F
.34716 .7875 m .34685 .78863 L .30864 .79077 L F
.34685 .78863 m .34663 .78948 L .30864 .79077 L F
.30919 .89603 m .30918 .89643 L .30889 .8972 L .715 .522 0 r
F
.30918 .89643 m .30916 .89723 L .30889 .8972 L F
.30889 .8972 m .30916 .89723 L .30918 .89743 L F
.31351 .89772 m .30916 .89723 L .30972 .87806 L .427 .299 0 r
F
.31606 .91275 m .3111 .91392 L .30916 .89723 L .504 .37 0 r
F
.30916 .89723 m .31351 .89772 L .31606 .91275 L .423 .329 0 r
F
.30942 1.12578 m .32779 1.11253 L .34051 1.11942 L .568 .628 0 r
F
.30942 1.12578 m .34051 1.11942 L .32463 1.1306 L .566 .675 0 r
F
.31335 .87978 m .30972 .87806 L .31265 .85183 L .247 .15 0 r
F
.30972 .87806 m .31335 .87978 L .31351 .89772 L .226 .165 0 r
F
.32108 1.04471 m .32151 1.06004 L .30977 1.06988 L .061 0 0 r
F
.30977 1.06988 m .32151 1.06004 L .31075 1.0779 L 0 0 0 r
F
.3099 1.14108 m .32463 1.1306 L .33954 1.13765 L .529 .627 0 r
F
.3099 1.14108 m .33954 1.13765 L .3269 1.14698 L .518 .67 0 r
F
.35865 .70465 m .30995 .74567 L .31007 .70514 L .7 .513 0 r
F
.35865 .70465 m .35702 .73832 L .30995 .74567 L .701 .514 0 r
F
.35697 .73863 m .35665 .74077 L .30995 .74567 L F
.35702 .73832 m .35697 .73863 L .30995 .74567 L F
.35665 .74077 m .35576 .74672 L .30995 .74567 L F
.33192 .69293 m .35865 .70465 L .31007 .70514 L .704 .583 0 r
F
.32151 1.06004 m .33285 1.05776 L .31075 1.0779 L 0 0 0 r
F
.31545 1.08408 m .31075 1.0779 L .33285 1.05776 L .468 .349 0 r
F
.3111 .91392 m .31606 .91275 L .32044 .92699 L .481 .3 0 r
F
.32044 .92699 m .31963 .93133 L .3111 .91392 L .177 .203 0 r
F
.31326 .85005 m .31265 .85183 L .31112 .84939 L .683 .514 0 r
F
.35914 .51069 m .31145 .56056 L .31404 .50644 L .788 .655 0 r
F
.35914 .51069 m .36015 .56299 L .31145 .56056 L .79 .659 0 r
F
.32229 .17806 m .31254 .14358 L .33495 .15845 L .75 .569 0 r
F
.31666 .85824 m .31265 .85183 L .3184 .83528 L .37 .231 0 r
F
.31265 .85183 m .31666 .85824 L .31335 .87978 L .337 .22 0 r
F
.31585 .87931 m .31335 .87978 L .31666 .85824 L .646 .5 0 r
F
.31335 .87978 m .31585 .87931 L .31706 .89789 L .644 .492 0 r
F
.31706 .89789 m .31351 .89772 L .31335 .87978 L .593 .459 0 r
F
.31606 .91275 m .31351 .89772 L .31706 .89789 L .593 .446 0 r
F
.36015 .464 m .31404 .50644 L .317 .46366 L .789 .645 0 r
F
.36015 .464 m .35914 .51069 L .31404 .50644 L .79 .645 0 r
F
.31708 .837 m .31762 .83754 L .31413 .83836 L .675 .472 0 r
F
.31762 .83754 m .3174 .83816 L .31413 .83836 L F
.3174 .83816 m .31707 .83911 L .31413 .83836 L F
.31707 .83911 m .31688 .83966 L .31413 .83836 L F
.31688 .83966 m .31675 .84002 L .31413 .83836 L F
.3142 1.15429 m .3269 1.14698 L .34193 1.15513 L .325 .58 0 r
F
.3142 1.15429 m .34193 1.15513 L .32777 1.16288 L .322 .583 0 r
F
.31444 1.10637 m .32893 1.09051 L .3428 1.09708 L .605 .613 0 r
F
.31444 1.10637 m .3428 1.09708 L .32779 1.11253 L .613 .662 0 r
F
.31495 .2128 m .33443 .21201 L .33795 .2159 L .595 .164 0 r
F
.33795 .2159 m .32144 .21653 L .31495 .2128 L .543 .05 0 r
F
.33285 1.05776 m .33643 1.06354 L .31545 1.08408 L .526 .423 0 r
F
.31545 1.08408 m .33643 1.06354 L .34916 1.07066 L .633 .593 0 r
F
.31545 1.08408 m .34916 1.07066 L .32893 1.09051 L .641 .62 0 r
F
.35913 .3701 m .31565 .41372 L .31823 .36693 L .775 .629 0 r
F
.35913 .3701 m .35716 .41471 L .31565 .41372 L .78 .637 0 r
F
.35716 .41471 m .317 .46366 L .31565 .41372 L .78 .629 0 r
F
.31585 .87931 m .31666 .85824 L .32035 .86809 L .124 .068 0 r
F
.32035 .86809 m .32123 .87961 L .31585 .87931 L .299 .259 0 r
F
.31706 .89789 m .31585 .87931 L .32123 .87961 L .301 .23 0 r
F
.31706 .89789 m .31808 .91074 L .31606 .91275 L .555 .414 0 r
F
.32044 .92699 m .31606 .91275 L .31808 .91074 L .473 .295 0 r
F
.3184 .83528 m .32172 .84549 L .31666 .85824 L .276 .153 0 r
F
.31666 .85824 m .32172 .84549 L .32565 .85252 L .304 .182 0 r
F
.32565 .85252 m .32035 .86809 L .31666 .85824 L .286 .151 0 r
F
.35716 .41471 m .36015 .464 L .317 .46366 L .789 .643 0 r
F
.32123 .87961 m .32701 .88834 L .31706 .89789 L .324 .253 0 r
F
.31808 .91074 m .31706 .89789 L .32701 .88834 L .308 .229 0 r
F
.31794 .8366 m .31776 .83712 L .31708 .837 L .675 .472 0 r
F
.31776 .83712 m .31763 .8375 L .31708 .837 L F
.31763 .8375 m .31762 .83754 L .31708 .837 L F
.31744 .92917 m .31916 .93039 L .31963 .93133 L .701 .501 0 r
F
.31744 .92917 m .31881 .92967 L .31916 .93039 L F
.32701 .88834 m .32403 .91021 L .31808 .91074 L .277 .195 0 r
F
.31808 .91074 m .32403 .91021 L .32408 .92193 L .286 .276 0 r
F
.31808 .91074 m .32408 .92193 L .32044 .92699 L .697 .45 0 r
F
.35617 .32673 m .31823 .36693 L .31929 .32384 L .79 .635 0 r
F
.35617 .32673 m .35913 .3701 L .31823 .36693 L .776 .611 0 r
F
.3185 .83461 m .31846 .83492 L .31823 .83476 L .675 .472 0 r
F
.31846 .83492 m .3184 .83528 L .31823 .83476 L F
.33519 .82355 m .3184 .83528 L .32062 .82067 L .404 .365 0 r
F
.33519 .82355 m .32172 .84549 L .3184 .83528 L .287 .159 0 r
F
.35617 .32673 m .31929 .32384 L .35541 .28203 L .791 .621 0 r
F
.38238 .10364 m .33495 .15845 L .31943 .10261 L .76 .549 0 r
F
.31963 .93133 m .32044 .92699 L .32755 .92888 L .23 .255 0 r
F
.32755 .92888 m .32922 .93724 L .31963 .93133 L .23 .254 0 r
F
.31963 .93133 m .32444 .93773 L .32339 .93569 L .861 .431 0 r
F
.32541 .93802 m .31963 .93133 L .32752 .93911 L 0 .265 0 r
F
.32793 .93951 m .31963 .93133 L .32922 .93724 L F
.33051 .8708 m .32035 .86809 L .32565 .85252 L .681 .559 0 r
F
.32123 .87961 m .32035 .86809 L .33051 .8708 L .684 .553 0 r
F
.32943 .92654 m .32044 .92699 L .32408 .92193 L .355 .045 0 r
F
.32044 .92699 m .32943 .92654 L .32755 .92888 L .333 0 0 r
F
.32062 .82067 m .32077 .81963 L .32084 .82006 L .685 .486 0 r
F
.33519 .82355 m .32062 .82067 L .32534 .80813 L .413 .346 0 r
F
.32084 .82006 m .32077 .81963 L .32087 .81999 L .685 .486 0 r
F
.32087 .81999 m .32077 .81963 L .32119 .81915 L F
.32952 1.01893 m .32995 1.03802 L .32108 1.04471 L .091 0 0 r
F
.32108 1.04471 m .32995 1.03802 L .32151 1.06004 L .03 0 0 r
F
.33051 .8708 m .3303 .87829 L .32123 .87961 L .625 .46 0 r
F
.32701 .88834 m .32123 .87961 L .3303 .87829 L .589 .337 0 r
F
.35541 .28203 m .32144 .21653 L .33795 .2159 L .764 .615 0 r
F
.32995 1.03802 m .33575 1.04337 L .32151 1.06004 L .012 0 0 r
F
.32151 1.06004 m .33575 1.04337 L .33285 1.05776 L 0 0 0 r
F
.32598 .93868 m .32162 .93847 L .32296 .93704 L .703 .512 0 r
F
.32665 .93945 m .32162 .93847 L .32598 .93868 L F
.32744 .94037 m .32162 .93847 L .32665 .93945 L F
.33519 .82355 m .32565 .85252 L .32172 .84549 L .307 .183 0 r
F
.32229 .17806 m .33495 .15845 L .33457 .19813 L .758 .579 0 r
F
.33457 .19813 m .33443 .21201 L .32229 .17806 L .826 .592 0 r
F
.32564 .93828 m .32598 .93868 L .32296 .93704 L .703 .512 0 r
F
.3239 .93702 m .32394 .93707 L .32296 .93704 L F
.32396 .93709 m .32404 .9372 L .32296 .93704 L F
.32394 .93707 m .32396 .93709 L .32296 .93704 L F
.32404 .9372 m .32444 .93773 L .32296 .93704 L F
.32444 .93773 m .32511 .93768 L .32339 .93569 L .744 .475 0 r
F
.32908 .80616 m .32367 .80551 L .33344 .78993 L .685 .486 0 r
F
.32534 .80813 m .32367 .80551 L .32908 .80616 L F
.3285 .90879 m .32403 .91021 L .32701 .88834 L .322 .236 0 r
F
.32408 .92193 m .32403 .91021 L .3285 .90879 L .355 .333 0 r
F
.3285 .90879 m .33402 .92319 L .32408 .92193 L .444 .429 0 r
F
.32408 .92193 m .33278 .9241 L .32943 .92654 L .486 .018 0 r
F
.33402 .92319 m .33278 .9241 L .32408 .92193 L .444 .429 0 r
F
.32444 .93773 m .32564 .93828 L .32523 .93781 L .499 .01 0 r
F
.32523 .93781 m .32511 .93768 L .32444 .93773 L .438 0 0 r
F
.32463 1.1306 m .34051 1.11942 L .35432 1.12702 L .536 .609 0 r
F
.32463 1.1306 m .35432 1.12702 L .33954 1.13765 L .533 .627 0 r
F
.33909 .96391 m .33574 .99183 L .32526 1.00714 L .368 .188 0 r
F
.33573 .9671 m .33909 .96391 L .32526 1.00714 L .198 .026 0 r
F
.33574 .99183 m .32952 1.01893 L .32526 1.00714 L .275 .079 0 r
F
.33519 .82355 m .32534 .80813 L .33414 .80348 L .479 .371 0 r
F
.32541 .93802 m .32752 .93911 L .32793 .93951 L 0 .265 0 r
F
.32546 .93808 m .32541 .93802 L .32793 .93951 L F
.32564 .93828 m .32651 .93869 L .32546 .93808 L .499 .01 0 r
F
.32744 .94037 m .32564 .93828 L .32793 .93951 L 0 .265 0 r
F
.32564 .93828 m .32651 .93869 L .32793 .93951 L F
.33519 .82355 m .33568 .8436 L .32565 .85252 L .32 .196 0 r
F
.33311 .85343 m .32565 .85252 L .33277 .84619 L .556 .578 0 r
F
.33137 .85786 m .32565 .85252 L .33311 .85343 L F
.32565 .85252 m .32672 .85352 L .33051 .8708 L .735 .592 0 r
F
.32672 .85352 m .3314 .86743 L .33051 .8708 L F
.32672 .85352 m .33144 .85792 L .33076 .86068 L F
.32672 .85352 m .33076 .86068 L .32933 .86105 L F
.32672 .85352 m .33109 .86613 L .3312 .86683 L F
.3269 1.14698 m .33954 1.13765 L .35397 1.14474 L .495 .62 0 r
F
.3269 1.14698 m .35397 1.14474 L .34193 1.15513 L .499 .605 0 r
F
.3303 .87829 m .32929 .88222 L .32701 .88834 L .464 .214 0 r
F
.32701 .88834 m .32929 .88222 L .33179 .88786 L .347 .099 0 r
F
.32701 .88834 m .33179 .88786 L .33923 .89413 L .403 .426 0 r
F
.32701 .88834 m .33294 .90694 L .3285 .90879 L .546 .404 0 r
F
.32701 .88834 m .33701 .90343 L .33294 .90694 L .557 .41 0 r
F
.32701 .88834 m .33943 .90098 L .33701 .90343 L .503 .393 0 r
F
.33923 .89413 m .33943 .90098 L .32701 .88834 L .48 .389 0 r
F
.32922 .93724 m .33645 .94332 L .32744 .94037 L 0 .377 0 r
F
.32922 .93724 m .32755 .92888 L .33503 .92466 L .417 .387 0 r
F
.33503 .92466 m .32755 .92888 L .32943 .92654 L 0 0 0 r
F
.32777 1.16288 m .34193 1.15513 L .3557 1.16458 L .309 .55 0 r
F
.32777 1.16288 m .3557 1.16458 L .33987 1.17208 L .315 .544 0 r
F
.32779 1.11253 m .3428 1.09708 L .35528 1.10303 L .591 .625 0 r
F
.32779 1.11253 m .35528 1.10303 L .34051 1.11942 L .591 .626 0 r
F
.3285 .90879 m .33294 .90694 L .33922 .91702 L .581 .491 0 r
F
.3362 .91602 m .33493 .92253 L .3285 .90879 L .547 .488 0 r
F
.33493 .92253 m .33402 .92319 L .3285 .90879 L F
.3285 .90879 m .33922 .91702 L .33922 .91735 L .581 .491 0 r
F
.3285 .90879 m .33922 .91735 L .33854 .91822 L F
.32893 1.09051 m .34916 1.07066 L .36088 1.07676 L .622 .589 0 r
F
.32893 1.09051 m .36088 1.07676 L .3428 1.09708 L .627 .606 0 r
F
.35234 .75066 m .34939 .77896 L .32901 .77207 L .685 .485 0 r
F
.3492 .77968 m .34716 .7875 L .32901 .77207 L F
.34939 .77896 m .3492 .77968 L .32901 .77207 L F
.33414 .80348 m .32908 .80616 L .33344 .78993 L .685 .486 0 r
F
.33503 .92466 m .34038 .93286 L .32922 .93724 L .444 .419 0 r
F
.33645 .94332 m .32922 .93724 L .34038 .93286 L .5 .607 0 r
F
.33488 .87515 m .32929 .88222 L .3303 .87829 L .185 0 0 r
F
.33179 .88786 m .32929 .88222 L .33488 .87515 L .268 .065 0 r
F
.33076 .86068 m .33044 .86197 L .32933 .86105 L .735 .592 0 r
F
.32933 .86105 m .33044 .86197 L .33109 .86613 L F
.32943 .92654 m .33536 .92221 L .33503 .92466 L .064 0 0 r
F
.33574 .99183 m .33572 1.01435 L .32952 1.01893 L 0 0 0 r
F
.32952 1.01893 m .33572 1.01435 L .32995 1.03802 L F
.33572 1.01435 m .33893 1.02525 L .32995 1.03802 L .066 0 0 r
F
.32995 1.03802 m .33893 1.02525 L .33575 1.04337 L 0 0 0 r
F
.3303 .87829 m .33178 .87032 L .33253 .87077 L .487 .347 0 r
F
.3303 .87829 m .33051 .8708 L .33178 .87032 L F
.3303 .87829 m .33253 .87077 L .33379 .87271 L F
.3303 .87829 m .33379 .87271 L .33488 .87515 L .249 .068 0 r
F
.34869 .85541 m .33044 .86197 L .3316 .85728 L .486 .365 0 r
F
.3312 .86683 m .33044 .86197 L .34869 .85541 L .484 .359 0 r
F
.33051 .8708 m .3314 .86743 L .33179 .8686 L .478 .334 0 r
F
.33179 .8686 m .33218 .86974 L .33051 .8708 L .486 .343 0 r
F
.33218 .86974 m .33231 .87013 L .33051 .8708 L F
.34869 .85541 m .33253 .87077 L .3312 .86683 L .444 .289 0 r
F
.33144 .85792 m .33137 .85786 L .3316 .85728 L .556 .578 0 r
F
.33223 .94194 m .33573 .9671 L .33158 .94172 L .703 .512 0 r
F
.3316 .85728 m .33325 .85308 L .33211 .85722 L .479 .327 0 r
F
.33253 .87077 m .33178 .87032 L .33231 .87013 L .487 .347 0 r
F
.33179 .88786 m .33488 .87515 L .33524 .87492 L .403 .187 0 r
F
.33179 .88786 m .33524 .87492 L .33586 .87586 L F
.33586 .87586 m .33927 .88108 L .33179 .88786 L .534 .316 0 r
F
.33927 .88108 m .3402 .88729 L .33179 .88786 L .537 .321 0 r
F
.3402 .88729 m .33923 .89413 L .33179 .88786 L .558 .434 0 r
F
.36077 .6784 m .35865 .70465 L .33192 .69293 L .681 .602 0 r
F
.33211 .85722 m .33325 .85308 L .34869 .85541 L .479 .327 0 r
F
.3378 .94394 m .33573 .9671 L .33223 .94194 L .703 .512 0 r
F
.3378 .94394 m .33223 .94194 L .33645 .94332 L F
.33869 .8745 m .33253 .87077 L .34869 .85541 L .487 .352 0 r
F
.33253 .87077 m .33869 .8745 L .33927 .88108 L .22 .39 0 r
F
.33364 .85155 m .33277 .84619 L .33568 .8436 L .556 .578 0 r
F
.33311 .85343 m .33277 .84619 L .33325 .85308 L F
.33325 .85308 m .33277 .84619 L .33338 .85259 L F
.33348 .8522 m .33277 .84619 L .33353 .85199 L F
.33356 .85186 m .33277 .84619 L .33364 .85155 L F
.33353 .85199 m .33277 .84619 L .33356 .85186 L F
.33338 .85259 m .33277 .84619 L .33348 .8522 L F
.33575 1.04337 m .35309 1.03497 L .33285 1.05776 L 0 0 0 r
F
.33643 1.06354 m .33285 1.05776 L .35309 1.03497 L .547 .431 0 r
F
.33294 .90694 m .33701 .90343 L .34244 .90992 L .59 .463 0 r
F
.33965 .91636 m .33931 .91716 L .33294 .90694 L .547 .478 0 r
F
.34148 .91213 m .33965 .91636 L .33294 .90694 L F
.33294 .90694 m .34244 .90992 L .34201 .91091 L .59 .463 0 r
F
.33294 .90694 m .34201 .91091 L .34148 .91213 L F
.33325 .85308 m .33568 .8436 L .3492 .83923 L .489 .346 0 r
F
.3492 .83923 m .34869 .85541 L .33325 .85308 L .479 .33 0 r
F
.34663 .78948 m .33475 .80333 L .33344 .78993 L .685 .486 0 r
F
.33453 .80338 m .33414 .80348 L .33344 .78993 L F
.33475 .80333 m .33453 .80338 L .33344 .78993 L F
.33379 .87271 m .3347 .8741 L .33488 .87515 L .249 .068 0 r
F
.34171 .80513 m .33414 .80348 L .33798 .80254 L .357 0 0 r
F
.33519 .82355 m .33414 .80348 L .34171 .80513 L .434 .336 0 r
F
.34615 .21706 m .33443 .21201 L .33495 .15845 L .817 .617 0 r
F
.33795 .2159 m .33443 .21201 L .34615 .21706 L .676 .133 0 r
F
.3347 .8741 m .33501 .87457 L .33488 .87515 L .249 .068 0 r
F
.33476 .80333 m .33475 .80333 L .34663 .78948 L .685 .486 0 r
F
.33476 .80333 m .33476 .80333 L .34663 .78948 L F
.33476 .80333 m .33476 .80333 L .34663 .78948 L F
.33539 .80317 m .33476 .80333 L .34662 .7895 L F
.34662 .7895 m .33476 .80333 L .34663 .78948 L F
.33524 .87492 m .33488 .87515 L .33501 .87457 L .429 .229 0 r
F
.33536 .92221 m .33493 .92253 L .3362 .91602 L .547 .488 0 r
F
.38238 .10364 m .38911 .15673 L .33495 .15845 L .768 .561 0 r
F
.38911 .15673 m .34615 .21706 L .33495 .15845 L .772 .588 0 r
F
.33909 .91932 m .33503 .92466 L .33536 .92221 L 0 0 0 r
F
.34038 .93286 m .33503 .92466 L .33909 .91932 L .501 .441 0 r
F
.33519 .82355 m .34171 .80513 L .35243 .81688 L .483 .387 0 r
F
.33519 .82355 m .35243 .81688 L .33568 .8436 L .443 .29 0 r
F
.336 .92141 m .33536 .92221 L .3362 .91602 L .547 .488 0 r
F
.33536 .92221 m .33922 .91735 L .33909 .91932 L 0 0 0 r
F
.33541 .80317 m .33539 .80317 L .34662 .7895 L .685 .486 0 r
F
.33541 .80317 m .33541 .80317 L .34662 .7895 L F
.33666 .80286 m .33541 .80317 L .34662 .7895 L F
.3492 .83923 m .33568 .8436 L .35243 .81688 L .489 .346 0 r
F
.34085 .9884 m .34125 1.00407 L .33572 1.01435 L 0 0 0 r
F
.33574 .99183 m .34085 .9884 L .33572 1.01435 L F
.33572 1.01435 m .34125 1.00407 L .33893 1.02525 L F
.33788 .94299 m .33909 .96391 L .33573 .9671 L .293 .161 0 r
F
.33909 .96391 m .34085 .9884 L .33574 .99183 L 0 0 0 r
F
.33893 1.02525 m .35294 1.02267 L .33575 1.04337 L .067 0 0 r
F
.33575 1.04337 m .35294 1.02267 L .35309 1.03497 L 0 0 0 r
F
.36061 .62487 m .36077 .6784 L .33583 .65147 L .728 .611 0 r
F
.33643 .92087 m .336 .92141 L .3362 .91602 L .547 .488 0 r
F
.33804 .91775 m .33643 .92087 L .3362 .91602 L F
.33674 .92048 m .33643 .92087 L .33804 .91775 L F
.35309 1.03497 m .35871 1.0428 L .33643 1.06354 L .606 .504 0 r
F
.33643 1.06354 m .35871 1.0428 L .36885 1.04861 L .621 .528 0 r
F
.33643 1.06354 m .36885 1.04861 L .34916 1.07066 L .644 .59 0 r
F
.34038 .93286 m .34657 .93569 L .33645 .94332 L .617 .738 0 r
F
.35296 .93566 m .34917 .94038 L .33645 .94332 L .599 .938 0 r
F
.33645 .94332 m .34657 .93569 L .35296 .93566 L .666 .912 0 r
F
.33788 .94299 m .3378 .94394 L .33645 .94332 L .703 .512 0 r
F
.33705 .80277 m .33666 .80286 L .34662 .7895 L .685 .486 0 r
F
.33844 .91834 m .33674 .92048 L .33804 .91775 L .547 .488 0 r
F
.33701 .90343 m .33943 .90098 L .34391 .90389 L .575 .502 0 r
F
.33701 .90343 m .34388 .90406 L .34382 .90446 L F
.33701 .90343 m .34391 .90389 L .34388 .90406 L F
.33701 .90343 m .34382 .90446 L .3438 .90457 L F
.33701 .90343 m .3438 .90457 L .34373 .90497 L F
.33701 .90343 m .34373 .90497 L .3436 .90577 L F
.34246 .90986 m .34244 .90992 L .33701 .90343 L .605 .466 0 r
F
.34248 .90981 m .34246 .90986 L .33701 .90343 L F
.33741 .90357 m .34248 .90981 L .33701 .90343 L F
.33732 .8027 m .33705 .80277 L .34662 .7895 L .685 .486 0 r
F
.34552 .79365 m .3377 .80261 L .33732 .8027 L .675 .472 0 r
F
.34552 .79365 m .33732 .8027 L .34662 .7895 L .685 .486 0 r
F
.34319 .90817 m .34283 .90901 L .33741 .90357 L .605 .466 0 r
F
.34217 .90526 m .34319 .90817 L .33741 .90357 L F
.34283 .90901 m .34248 .90981 L .33741 .90357 L F
.33892 .80121 m .33798 .80254 L .3377 .80261 L .675 .472 0 r
F
.33859 .94282 m .33909 .96391 L .33788 .94299 L .293 .161 0 r
F
.34615 .21706 m .34969 .21993 L .33795 .2159 L .56 0 0 r
F
.35541 .28203 m .33795 .2159 L .34969 .21993 L .805 .638 0 r
F
.33798 .80254 m .34658 .80343 L .34171 .80513 L .29 0 0 r
F
.33802 .80255 m .33798 .80254 L .33892 .80121 L .675 .472 0 r
F
.33802 .80255 m .33802 .80255 L .33892 .80121 L F
.33803 .80255 m .33802 .80255 L .33892 .80121 L F
.34552 .79365 m .33803 .80255 L .33892 .80121 L F
.34552 .79365 m .33809 .80255 L .33803 .80255 L F
.33851 .91826 m .33844 .91834 L .33804 .91775 L .547 .488 0 r
F
.33854 .91822 m .33851 .91826 L .33804 .91775 L F
.34552 .79365 m .3381 .80256 L .33809 .80255 L .675 .472 0 r
F
.34552 .79365 m .33811 .80256 L .3381 .80256 L F
.34552 .79365 m .34471 .79668 L .33811 .80256 L F
.34471 .79668 m .33847 .80259 L .33811 .80256 L F
.34471 .79668 m .33994 .80274 L .33847 .80259 L F
.33859 .94282 m .34518 .96295 L .33909 .96391 L .12 .031 0 r
F
.33993 .94251 m .34518 .96295 L .33859 .94282 L F
.34869 .85541 m .34245 .87914 L .33869 .8745 L .478 .342 0 r
F
.33869 .8745 m .34245 .87914 L .33927 .88108 L 0 .186 0 r
F
.34125 1.00407 m .35188 1.00759 L .33893 1.02525 L .143 0 0 r
F
.33893 1.02525 m .35188 1.00759 L .35294 1.02267 L .079 0 0 r
F
.34385 .91004 m .33909 .91932 L .33922 .91735 L 0 0 0 r
F
.34532 .92866 m .33909 .91932 L .34385 .91004 L .582 .467 0 r
F
.33909 .91932 m .34532 .92866 L .34038 .93286 L .493 .435 0 r
F
.33909 .96391 m .34518 .96295 L .34085 .9884 L .101 0 0 r
F
.33929 .91718 m .33922 .91702 L .33931 .91716 L .581 .491 0 r
F
.33922 .91735 m .33922 .91702 L .33929 .91718 L F
.33922 .91735 m .34319 .90817 L .34385 .91004 L .136 .008 0 r
F
.34263 .891 m .33923 .89413 L .3402 .88729 L .605 .477 0 r
F
.34263 .89101 m .33923 .89413 L .34263 .891 L F
.34263 .89102 m .33923 .89413 L .34263 .89101 L F
.34296 .89413 m .33923 .89413 L .34263 .89102 L F
.34328 .89724 m .33923 .89413 L .34314 .89589 L F
.34296 .89413 m .33923 .89413 L .34296 .89413 L F
.34314 .89589 m .33923 .89413 L .34296 .89413 L F
.33943 .90098 m .33923 .89413 L .34328 .89724 L .593 .479 0 r
F
.33927 .88108 m .34245 .87914 L .34335 .88109 L .219 .635 0 r
F
.3402 .88729 m .33927 .88108 L .34335 .88109 L .356 .198 0 r
F
.33943 .90098 m .34328 .89724 L .34343 .89866 L .593 .479 0 r
F
.33943 .90098 m .34343 .89866 L .34348 .89912 L F
.33943 .90098 m .34348 .89918 L .3435 .89924 L F
.33943 .90098 m .34343 .89915 L .34348 .89918 L F
.33943 .90098 m .34352 .89934 L .34353 .89939 L F
.33943 .90098 m .3435 .89924 L .34352 .89934 L F
.33943 .90098 m .34353 .89939 L .34402 .90164 L F
.34398 .90347 m .34391 .90389 L .33943 .90098 L .588 .499 0 r
F
.344 .90335 m .34398 .90347 L .33943 .90098 L F
.34401 .90332 m .344 .90335 L .33943 .90098 L F
.34403 .9032 m .34401 .90332 L .33943 .90098 L F
.34406 .903 m .34403 .9032 L .33943 .90098 L F
.34407 .90295 m .34406 .903 L .33943 .90098 L F
.3421 .90137 m .34407 .90295 L .33943 .90098 L F
.33954 1.13765 m .35432 1.12702 L .36694 1.13303 L .533 .625 0 r
F
.33954 1.13765 m .36694 1.13303 L .35397 1.14474 L .533 .619 0 r
F
.33987 1.17208 m .3557 1.16458 L .37979 1.178 L .333 .6 0 r
F
.33987 1.17208 m .37979 1.178 L .36133 1.18675 L .328 .603 0 r
F
.33993 .94251 m .34567 .95896 L .3446 .96072 L .428 .21 0 r
F
.3403 .94243 m .34567 .95896 L .33993 .94251 L F
.34471 .79668 m .34247 .803 L .33994 .80274 L .675 .472 0 r
F
.3402 .88729 m .34335 .88109 L .34337 .88115 L .356 .198 0 r
F
.3402 .88729 m .34337 .88115 L .34554 .88587 L .068 0 0 r
F
.34554 .88587 m .34263 .89098 L .3402 .88729 L .069 0 0 r
F
.34263 .89099 m .34263 .891 L .3402 .88729 L 0 0 0 r
F
.34263 .89098 m .34263 .89099 L .3402 .88729 L F
.34263 .89099 m .34263 .89099 L .3402 .88729 L F
.3403 .94243 m .34917 .94038 L .34567 .95896 L .428 .21 0 r
F
.34657 .93569 m .34038 .93286 L .34532 .92866 L .656 .737 0 r
F
.34051 1.11942 m .35528 1.10303 L .36877 1.10946 L .595 .633 0 r
F
.34051 1.11942 m .36877 1.10946 L .35432 1.12702 L .59 .604 0 r
F
.34518 .96295 m .34355 .98201 L .34085 .9884 L 0 0 0 r
F
.34085 .9884 m .34355 .98201 L .34125 1.00407 L F
.34355 .98201 m .35005 .99173 L .34125 1.00407 L .168 0 0 r
F
.34125 1.00407 m .35005 .99173 L .35188 1.00759 L .144 0 0 r
F
.35243 .81688 m .34171 .80513 L .34658 .80343 L .692 .402 0 r
F
.34193 1.15513 m .35397 1.14474 L .36949 1.15308 L .493 .594 0 r
F
.34193 1.15513 m .36949 1.15308 L .3557 1.16458 L .493 .595 0 r
F
.34417 .90235 m .34415 .90249 L .3421 .90137 L .588 .499 0 r
F
.34374 .9016 m .34417 .90235 L .3421 .90137 L F
.34415 .90249 m .34407 .90295 L .3421 .90137 L F
.34324 .90786 m .34319 .90817 L .34217 .90526 L .605 .466 0 r
F
.34341 .90689 m .34324 .90786 L .34217 .90526 L F
.3436 .90577 m .34341 .90689 L .34217 .90526 L F
.35216 .87503 m .34245 .87914 L .34869 .85541 L .506 .369 0 r
F
.34245 .87914 m .35216 .87503 L .35409 .89464 L .504 .365 0 r
F
.35409 .89464 m .34554 .88587 L .34245 .87914 L .522 .37 0 r
F
.34471 .79668 m .34301 .80306 L .34247 .803 L .675 .472 0 r
F
.34263 .89098 m .34554 .88587 L .34911 .89653 L .442 .311 0 r
F
.34911 .89653 m .34521 .89852 L .34263 .89098 L .529 .303 0 r
F
.34263 .89098 m .34521 .89852 L .34348 .89918 L .123 .081 0 r
F
.3428 1.09708 m .36088 1.07676 L .37272 1.0827 L .634 .616 0 r
F
.3428 1.09708 m .37272 1.0827 L .35528 1.10303 L .633 .614 0 r
F
.34716 .7875 m .34309 .80307 L .34301 .80306 L 0 0 0 r
F
.34716 .7875 m .34318 .80308 L .34309 .80307 L F
.34716 .7875 m .34355 .80312 L .34318 .80308 L F
.34319 .90817 m .34417 .90235 L .34579 .90339 L .043 0 0 r
F
.34579 .90339 m .34385 .91004 L .34319 .90817 L .042 0 0 r
F
.34343 .89915 m .34348 .89912 L .34348 .89918 L .593 .479 0 r
F
.34417 .90235 m .34348 .89918 L .34521 .89852 L .165 .198 0 r
F
.34518 .96295 m .34553 .95969 L .34355 .98201 L 0 0 0 r
F
.34553 .95969 m .34801 .97472 L .34355 .98201 L .072 0 0 r
F
.34355 .98201 m .34801 .97472 L .35005 .99173 L .116 0 0 r
F
.34716 .7875 m .34373 .80313 L .34355 .80312 L 0 0 0 r
F
.34716 .7875 m .34375 .80314 L .34373 .80313 L F
.34413 .90217 m .34417 .90235 L .34374 .9016 L .588 .499 0 r
F
.34402 .90164 m .34413 .90217 L .34374 .9016 L F
.34716 .7875 m .34389 .80315 L .34375 .80314 L 0 0 0 r
F
.35067 .91689 m .34385 .91004 L .34579 .90339 L .505 .425 0 r
F
.34385 .91004 m .35067 .91689 L .34532 .92866 L .531 .429 0 r
F
.34716 .7875 m .34846 .78974 L .34389 .80315 L 0 0 0 r
F
.34846 .78974 m .34416 .80318 L .34389 .80315 L F
.34846 .78974 m .34492 .80326 L .34416 .80318 L F
.34521 .89852 m .34579 .90339 L .34417 .90235 L F
.34567 .95896 m .34553 .95969 L .3446 .96072 L .428 .21 0 r
F
.3446 .96072 m .34553 .95969 L .34518 .96295 L F
.34846 .78974 m .34578 .80334 L .34492 .80326 L 0 0 0 r
F
.34521 .89852 m .34911 .89653 L .35115 .90634 L .606 .451 0 r
F
.35115 .90634 m .34579 .90339 L .34521 .89852 L .593 .448 0 r
F
.3517 .92996 m .34532 .92866 L .35067 .91689 L .767 .707 0 r
F
.34532 .92866 m .3517 .92996 L .34657 .93569 L .702 .768 0 r
F
.34917 .94038 m .34825 .95666 L .34553 .95969 L .242 .006 0 r
F
.34553 .95969 m .34825 .95666 L .34801 .97472 L .288 .064 0 r
F
.35409 .89464 m .34911 .89653 L .34554 .88587 L .549 .372 0 r
F
.34658 .80343 m .34578 .80334 L .34588 .80282 L 0 0 0 r
F
.34579 .90339 m .35115 .90634 L .35067 .91689 L .569 .461 0 r
F
.34767 .80436 m .34588 .80282 L .34694 .79744 L 0 0 0 r
F
.38911 .15673 m .40147 .22085 L .34615 .21706 L .763 .575 0 r
F
.34944 .21729 m .34969 .21993 L .34615 .21706 L .763 .596 0 r
F
.34657 .93569 m .3517 .92996 L .35484 .92884 L .765 .896 0 r
F
.35484 .92884 m .35296 .93566 L .34657 .93569 L .765 .893 0 r
F
.34658 .80343 m .34767 .80436 L .35243 .81688 L .888 .461 0 r
F
.34767 .80436 m .34694 .79744 L .34846 .78974 L 0 0 0 r
F
.34716 .7875 m .35568 .75494 L .35173 .77102 L F
.35173 .77102 m .34846 .78974 L .34716 .7875 L F
.35243 .81688 m .34767 .80436 L .35386 .78341 L .32 .18 0 r
F
.34846 .78974 m .35386 .78341 L .34767 .80436 L 0 0 0 r
F
.34825 .95666 m .3522 .96506 L .34801 .97472 L F
.34801 .97472 m .3522 .96506 L .35611 .97763 L .063 0 0 r
F
.34801 .97472 m .35611 .97763 L .35005 .99173 L .056 0 0 r
F
.34917 .94038 m .34883 .95314 L .34825 .95666 L 0 0 0 r
F
.34825 .95666 m .34883 .95314 L .3522 .96506 L F
.35173 .77102 m .35159 .77559 L .34846 .78974 L F
.34846 .78974 m .35159 .77559 L .35386 .78341 L .052 .019 0 r
F
.36139 .82121 m .34869 .85541 L .3492 .83923 L .198 .103 0 r
F
.367 .84288 m .34869 .85541 L .36139 .82121 L .103 .005 0 r
F
.35216 .87503 m .34869 .85541 L .367 .84288 L .179 .145 0 r
F
.34917 .94038 m .35588 .94666 L .34883 .95314 L 0 0 0 r
F
.34883 .95314 m .35588 .94666 L .35932 .95616 L .066 0 0 r
F
.34883 .95314 m .35932 .95616 L .3522 .96506 L .067 0 0 r
F
.34911 .89653 m .35409 .89464 L .35526 .90596 L .579 .443 0 r
F
.35526 .90596 m .35115 .90634 L .34911 .89653 L .609 .453 0 r
F
.34916 1.07066 m .36885 1.04861 L .37961 1.05478 L .63 .568 0 r
F
.34916 1.07066 m .37961 1.05478 L .36088 1.07676 L .636 .584 0 r
F
.34917 .94038 m .35296 .93566 L .35588 .94666 L 0 0 0 r
F
.36139 .82121 m .3492 .83923 L .35243 .81688 L .16 .057 0 r
F
.35176 .76992 m .34939 .77896 L .35234 .75066 L .685 .485 0 r
F
.3545 .25916 m .34969 .21993 L .34944 .21729 L .763 .596 0 r
F
.40147 .22085 m .3545 .25916 L .34944 .21729 L F
.35791 .28695 m .35541 .28203 L .34969 .21993 L .989 .766 0 r
F
.35005 .99173 m .35611 .97763 L .36136 .9886 L .095 0 0 r
F
.35005 .99173 m .36136 .9886 L .35188 1.00759 L .076 0 0 r
F
.35508 .9175 m .35067 .91689 L .35115 .90634 L .66 .539 0 r
F
.35067 .91689 m .35508 .9175 L .3517 .92996 L .636 .608 0 r
F
.35115 .90634 m .35526 .90596 L .35508 .9175 L .618 .515 0 r
F
.35159 .77559 m .35173 .77102 L .35644 .75658 L 0 0 0 r
F
.35644 .75658 m .35956 .76315 L .35159 .77559 L .096 0 0 r
F
.35159 .77559 m .35956 .76315 L .35386 .78341 L .236 .123 0 r
F
.35508 .9175 m .35484 .92884 L .3517 .92996 L .786 .764 0 r
F
.35568 .75494 m .35644 .75658 L .35173 .77102 L 0 0 0 r
F
.35238 .76755 m .35176 .76992 L .35234 .75066 L .685 .485 0 r
F
.35188 1.00759 m .36136 .9886 L .36647 1.00052 L .116 0 0 r
F
.35188 1.00759 m .36647 1.00052 L .35294 1.02267 L .081 0 0 r
F
.367 .84288 m .36418 .86914 L .35216 .87503 L .122 .079 0 r
F
.35409 .89464 m .35216 .87503 L .36418 .86914 L .125 .087 0 r
F
.3522 .96506 m .35932 .95616 L .36487 .96372 L .115 0 0 r
F
.3522 .96506 m .36487 .96372 L .35611 .97763 L .112 0 0 r
F
.35562 .74765 m .35238 .76755 L .35234 .75066 L .685 .485 0 r
F
.35294 .76541 m .35238 .76755 L .35562 .74765 L F
.35386 .78341 m .35784 .80204 L .35243 .81688 L .129 .028 0 r
F
.35243 .81688 m .35784 .80204 L .36139 .82121 L .158 .058 0 r
F
.35304 .7649 m .35294 .76541 L .35562 .74765 L .685 .485 0 r
F
.35304 .7649 m .35299 .7652 L .35294 .76541 L F
.35294 1.02267 m .36647 1.00052 L .37123 1.01031 L .123 0 0 r
F
.35294 1.02267 m .37123 1.01031 L .35309 1.03497 L .025 0 0 r
F
.35438 .93911 m .35296 .93566 L .35484 .92884 L 0 0 0 r
F
.35296 .93566 m .35438 .93911 L .35588 .94666 L F
.35846 .7287 m .35325 .76421 L .35299 .7652 L F
.35871 1.0428 m .35309 1.03497 L .37123 1.01031 L .626 .51 0 r
F
.35846 .7287 m .35446 .75959 L .35325 .76421 L 0 0 0 r
F
.35784 .80204 m .35386 .78341 L .35956 .76315 L .148 .039 0 r
F
.35397 1.14474 m .36694 1.13303 L .37977 1.14017 L .523 .6 0 r
F
.35397 1.14474 m .37977 1.14017 L .36949 1.15308 L .523 .593 0 r
F
.36418 .86914 m .36615 .88758 L .35409 .89464 L .182 .149 0 r
F
.35526 .90596 m .35409 .89464 L .36615 .88758 L F
.35432 1.12702 m .36877 1.10946 L .38175 1.11485 L .611 .637 0 r
F
.35432 1.12702 m .38175 1.11485 L .36694 1.13303 L .604 .613 0 r
F
.35484 .92884 m .37315 .91347 L .35438 .93911 L 0 0 0 r
F
.35968 .94438 m .35438 .93911 L .37315 .91347 L .042 0 0 r
F
.35588 .94666 m .35438 .93911 L .35968 .94438 L 0 0 0 r
F
.35472 .75861 m .35446 .75959 L .3552 .75389 L F
.40147 .22085 m .35791 .28695 L .3545 .25916 L .763 .596 0 r
F
.35568 .75494 m .35472 .75861 L .3552 .75389 L 0 0 0 r
F
.35484 .92884 m .35508 .9175 L .37315 .91347 L .211 .27 0 r
F
.35508 .9175 m .35526 .90596 L .37315 .91347 L .185 .156 0 r
F
.35586 .75454 m .3552 .75389 L .35846 .7287 L 0 0 0 r
F
.35568 .75494 m .3552 .75389 L .35585 .75457 L F
.35585 .75457 m .3552 .75389 L .35586 .75454 L F
.36615 .88758 m .37315 .91347 L .35526 .90596 L .186 .155 0 r
F
.35528 1.10303 m .37272 1.0827 L .38395 1.08884 L .613 .583 0 r
F
.35528 1.10303 m .38395 1.08884 L .36877 1.10946 L .627 .625 0 r
F
.36011 .32717 m .35617 .32673 L .35541 .28203 L .972 .769 0 r
F
.35541 .28203 m .35791 .28695 L .36011 .32717 L .984 .775 0 r
F
.36973 .72381 m .37391 .72589 L .35568 .75494 L 0 0 0 r
F
.35644 .75658 m .35568 .75494 L .37391 .72589 L F
.3557 1.16458 m .36949 1.15308 L .39352 1.16439 L .527 .66 0 r
F
.3557 1.16458 m .39352 1.16439 L .37979 1.178 L .541 .636 0 r
F
.35595 .75434 m .35586 .75454 L .35846 .7287 L 0 0 0 r
F
.35588 .94666 m .35968 .94438 L .35932 .95616 L F
.35596 .75432 m .35595 .75434 L .35846 .7287 L F
.36422 .7111 m .35596 .75432 L .35846 .7287 L F
.36422 .7111 m .35771 .75045 L .35596 .75432 L .007 0 0 r
F
.35611 .97763 m .36487 .96372 L .37078 .97182 L .163 .02 0 r
F
.35611 .97763 m .37078 .97182 L .36136 .9886 L .16 .013 0 r
F
.35913 .3701 m .35617 .32673 L .36011 .32717 L .972 .757 0 r
F
.35644 .75658 m .37391 .72589 L .38032 .73757 L .098 0 0 r
F
.38032 .73757 m .35956 .76315 L .35644 .75658 L .161 0 0 r
F
.35714 .7375 m .35702 .73832 L .35865 .70465 L 0 0 0 r
F
.35719 .73718 m .35714 .7375 L .35865 .70465 L F
.35716 .41471 m .35913 .3701 L .36146 .37055 L .988 .826 0 r
F
.36146 .37055 m .35969 .41606 L .35716 .41471 L .996 .828 0 r
F
.36015 .464 m .35716 .41471 L .35969 .41606 L .993 .786 0 r
F
.35719 .73714 m .35719 .73718 L .35865 .70465 L 0 0 0 r
F
.36077 .6784 m .35719 .73714 L .35865 .70465 L F
.35846 .7287 m .35719 .73714 L .36077 .6784 L F
.36422 .7111 m .35771 .75045 L .35771 .75045 L .007 0 0 r
F
.36422 .7111 m .36 .74537 L .35771 .75045 L F
.35956 .76315 m .36687 .77205 L .35784 .80204 L .072 0 0 r
F
.36139 .82121 m .35784 .80204 L .36687 .77205 L .14 .047 0 r
F
.40147 .22085 m .4037 .28064 L .35791 .28695 L .793 .638 0 r
F
.4037 .28064 m .36011 .32717 L .35791 .28695 L .792 .63 0 r
F
.37156 .66597 m .35846 .7287 L .36077 .6784 L .385 .258 0 r
F
.36422 .7111 m .35846 .7287 L .37156 .66597 L 0 0 0 r
F
.37123 1.01031 m .37787 1.0179 L .35871 1.0428 L .571 .45 0 r
F
.35871 1.0428 m .37787 1.0179 L .36885 1.04861 L .626 .526 0 r
F
.36011 .32717 m .36146 .37055 L .35913 .3701 L .987 .787 0 r
F
.35914 .51069 m .36015 .464 L .36127 .46428 L .857 .712 0 r
F
.36127 .46428 m .3617 .51438 L .35914 .51069 L .94 .767 0 r
F
.3617 .51438 m .36015 .56299 L .35914 .51069 L .938 .762 0 r
F
.35932 .95616 m .35968 .94438 L .36487 .96372 L 0 0 0 r
F
.36687 .77205 m .35956 .76315 L .38032 .73757 L .17 0 0 r
F
.35968 .94438 m .37315 .91347 L .38389 .92775 L .318 .228 0 r
F
.38389 .92775 m .37326 .95185 L .35968 .94438 L .25 .111 0 r
F
.36487 .96372 m .35968 .94438 L .37326 .95185 L .215 .131 0 r
F
.4071 .37299 m .35969 .41606 L .36146 .37055 L .798 .65 0 r
F
.4071 .37299 m .40308 .41213 L .35969 .41606 L .802 .658 0 r
F
.40308 .41213 m .36127 .46428 L .35969 .41606 L .803 .673 0 r
F
.35969 .41606 m .36127 .46428 L .36015 .464 L .856 .685 0 r
F
.3603 .74471 m .36 .74537 L .36422 .7111 L .334 .191 0 r
F
.3628 .56541 m .3601 .59678 L .36015 .56299 L .939 .773 0 r
F
.3628 .56541 m .36547 .62371 L .3601 .59678 L .998 .78 0 r
F
.36547 .62371 m .36061 .62487 L .3601 .59678 L .905 .756 0 r
F
.4037 .28064 m .40357 .32416 L .36011 .32717 L .794 .634 0 r
F
.40357 .32416 m .36146 .37055 L .36011 .32717 L .794 .638 0 r
F
.3628 .56541 m .36015 .56299 L .3617 .51438 L .936 .761 0 r
F
.37403 .69441 m .3603 .74471 L .36422 .7111 L .334 .191 0 r
F
.37403 .69441 m .36973 .72381 L .3603 .74471 L .154 .021 0 r
F
.36547 .62371 m .36077 .6784 L .36061 .62487 L .904 .754 0 r
F
.37156 .66597 m .36077 .6784 L .36547 .62371 L .457 .356 0 r
F
.36088 1.07676 m .37961 1.05478 L .39059 1.06048 L .622 .563 0 r
F
.36088 1.07676 m .39059 1.06048 L .37272 1.0827 L .642 .613 0 r
F
.40308 .41213 m .40197 .45746 L .36127 .46428 L .789 .651 0 r
F
.40197 .45746 m .39606 .50019 L .36127 .46428 L .789 .646 0 r
F
.37791 .50523 m .3617 .51438 L .36127 .46428 L .774 .642 0 r
F
.39606 .50019 m .37791 .50523 L .36127 .46428 L .784 .647 0 r
F
.36133 1.18675 m .37979 1.178 L .40578 1.18448 L .364 .725 0 r
F
.36133 1.18675 m .40578 1.18448 L .3856 1.19382 L .371 .722 0 r
F
.36136 .9886 m .37078 .97182 L .37865 .9795 L .229 .09 0 r
F
.36136 .9886 m .37865 .9795 L .36647 1.00052 L .18 0 0 r
F
.36687 .77205 m .37251 .78483 L .36139 .82121 L .168 .07 0 r
F
.367 .84288 m .36139 .82121 L .37251 .78483 L .102 .005 0 r
F
.40357 .32416 m .4071 .37299 L .36146 .37055 L .798 .644 0 r
F
.37791 .50523 m .36181 .51955 L .3617 .51438 L .827 .901 0 r
F
.37791 .50523 m .36181 .51964 L .36181 .51955 L .825 .891 0 r
F
.37791 .50523 m .36182 .51974 L .36181 .51964 L F
.37791 .50523 m .37773 .5114 L .36182 .51974 L F
.37773 .5114 m .3619 .52353 L .36182 .51974 L .768 .616 0 r
F
.37773 .5114 m .3628 .56541 L .3619 .52353 L F
.37773 .5114 m .37842 .55164 L .3628 .56541 L .684 .528 0 r
F
.37842 .55164 m .36547 .62371 L .3628 .56541 L .694 .545 0 r
F
.367 .84288 m .38075 .86951 L .36418 .86914 L .284 .223 0 r
F
.36418 .86914 m .38075 .86951 L .36615 .88758 L .284 .225 0 r
F
.38336 .65031 m .36422 .7111 L .37156 .66597 L .411 .26 0 r
F
.37403 .69441 m .36422 .7111 L .38336 .65031 L .296 .146 0 r
F
.36487 .96372 m .37326 .95185 L .37078 .97182 L .12 .011 0 r
F
.37842 .55164 m .37979 .60488 L .36547 .62371 L .574 .432 0 r
F
.37979 .60488 m .37156 .66597 L .36547 .62371 L .587 .449 0 r
F
.38075 .86951 m .38236 .8912 L .36615 .88758 L .29 .233 0 r
F
.36615 .88758 m .38236 .8912 L .37315 .91347 L .295 .222 0 r
F
.36647 1.00052 m .37865 .9795 L .386 .9871 L .27 .096 0 r
F
.36647 1.00052 m .386 .9871 L .37123 1.01031 L .189 0 0 r
F
.38032 .73757 m .38375 .74191 L .36687 .77205 L .17 0 0 r
F
.36687 .77205 m .38375 .74191 L .38782 .75458 L .23 .058 0 r
F
.38782 .75458 m .37251 .78483 L .36687 .77205 L .268 .115 0 r
F
.36694 1.13303 m .38175 1.11485 L .39564 1.12142 L .605 .614 0 r
F
.36694 1.13303 m .39564 1.12142 L .37977 1.14017 L .599 .592 0 r
F
.37251 .78483 m .38207 .79544 L .367 .84288 L .234 .114 0 r
F
.367 .84288 m .38207 .79544 L .39463 .80885 L .332 .211 0 r
F
.39463 .80885 m .38689 .84236 L .367 .84288 L .373 .266 0 r
F
.367 .84288 m .38689 .84236 L .37975 .85834 L .367 .208 0 r
F
.37975 .85834 m .38075 .86951 L .367 .84288 L .144 .157 0 r
F
.36877 1.10946 m .38395 1.08884 L .39579 1.09478 L .623 .619 0 r
F
.36877 1.10946 m .39579 1.09478 L .38175 1.11485 L .628 .631 0 r
F
.36885 1.04861 m .37787 1.0179 L .39129 1.02775 L .609 .508 0 r
F
.36885 1.04861 m .39129 1.02775 L .37961 1.05478 L .642 .563 0 r
F
.36949 1.15308 m .37977 1.14017 L .40511 1.14996 L .579 .678 0 r
F
.36949 1.15308 m .40511 1.14996 L .39352 1.16439 L .586 .657 0 r
F
.38618 .67818 m .36973 .72381 L .37403 .69441 L .381 .216 0 r
F
.38618 .67818 m .38199 .70392 L .36973 .72381 L .332 .167 0 r
F
.36973 .72381 m .38199 .70392 L .3899 .70439 L .385 .228 0 r
F
.37391 .72589 m .36973 .72381 L .3899 .70439 L 0 0 0 r
F
.37078 .97182 m .37326 .95185 L .38243 .95608 L .254 .124 0 r
F
.37078 .97182 m .38243 .95608 L .37865 .9795 L .227 .09 0 r
F
.37787 1.0179 m .37123 1.01031 L .386 .9871 L .609 .456 0 r
F
.38336 .65031 m .37156 .66597 L .37979 .60488 L .489 .36 0 r
F
.37251 .78483 m .38782 .75458 L .395 .76581 L .283 .132 0 r
F
.395 .76581 m .38207 .79544 L .37251 .78483 L .275 .119 0 r
F
.37272 1.0827 m .39059 1.06048 L .40117 1.06552 L .627 .591 0 r
F
.37272 1.0827 m .40117 1.06552 L .38395 1.08884 L .623 .58 0 r
F
.38236 .8912 m .38389 .92775 L .37315 .91347 L .293 .221 0 r
F
.38243 .95608 m .37326 .95185 L .38389 .92775 L .259 .12 0 r
F
.3899 .70439 m .39135 .70834 L .37391 .72589 L 0 0 0 r
F
.38032 .73757 m .37391 .72589 L .39135 .70834 L .129 0 0 r
F
.38618 .67818 m .37403 .69441 L .38336 .65031 L .408 .249 0 r
F
.39606 .50019 m .39536 .50586 L .37773 .5114 L .785 .881 0 r
F
.39606 .50019 m .37773 .5114 L .37791 .50523 L .784 .86 0 r
F
.39536 .50586 m .37842 .55164 L .37773 .5114 L .77 .597 0 r
F
.386 .9871 m .39599 .98991 L .37787 1.0179 L .592 .439 0 r
F
.37787 1.0179 m .39599 .98991 L .40913 .99548 L .62 .474 0 r
F
.37787 1.0179 m .40913 .99548 L .39129 1.02775 L .635 .502 0 r
F
.39536 .50586 m .39553 .54042 L .37842 .55164 L .69 .508 0 r
F
.39553 .54042 m .37979 .60488 L .37842 .55164 L .704 .535 0 r
F
.37865 .9795 m .38243 .95608 L .40449 .96539 L .464 .299 0 r
F
.37865 .9795 m .40449 .96539 L .386 .9871 L .351 .1 0 r
F
.37961 1.05478 m .39129 1.02775 L .40755 1.03361 L .64 .561 0 r
F
.37961 1.05478 m .40755 1.03361 L .39059 1.06048 L .637 .556 0 r
F
.3972 .85425 m .37975 .85834 L .38689 .84236 L .468 .316 0 r
F
.38523 .86099 m .37975 .85834 L .3972 .85425 L .472 .331 0 r
F
.38523 .86099 m .38075 .86951 L .37975 .85834 L .414 .367 0 r
F
.37977 1.14017 m .39564 1.12142 L .41959 1.13083 L .639 .654 0 r
F
.37977 1.14017 m .41959 1.13083 L .40511 1.14996 L .639 .661 0 r
F
.37979 1.178 m .39352 1.16439 L .41867 1.17153 L .594 .73 0 r
F
.37979 1.178 m .41867 1.17153 L .40578 1.18448 L .589 .744 0 r
F
.39553 .54042 m .39574 .587 L .37979 .60488 L .597 .428 0 r
F
.39574 .587 m .39748 .634 L .37979 .60488 L .596 .426 0 r
F
.39748 .634 m .38336 .65031 L .37979 .60488 L .56 .414 0 r
F
.39135 .70834 m .39587 .71816 L .38032 .73757 L .041 0 0 r
F
.38375 .74191 m .38032 .73757 L .39587 .71816 L .19 0 0 r
F
.39672 .85933 m .39771 .86763 L .38075 .86951 L .493 .493 0 r
F
.38075 .86951 m .38523 .86099 L .39672 .85933 L .456 .417 0 r
F
.38075 .86951 m .39771 .86763 L .39653 .87483 L .465 .291 0 r
F
.38075 .86951 m .39653 .87483 L .39594 .88461 L .441 .329 0 r
F
.39594 .88461 m .38236 .8912 L .38075 .86951 L .412 .326 0 r
F
.38175 1.11485 m .39579 1.09478 L .41022 1.10147 L .647 .659 0 r
F
.38175 1.11485 m .41022 1.10147 L .39564 1.12142 L .631 .605 0 r
F
.40152 .66441 m .39728 .68605 L .38199 .70392 L .375 .176 0 r
F
.40152 .66441 m .38199 .70392 L .38618 .67818 L .512 .326 0 r
F
.38199 .70392 m .39728 .68605 L .3899 .70439 L .384 .188 0 r
F
.38207 .79544 m .395 .76581 L .40715 .78012 L .362 .211 0 r
F
.40715 .78012 m .39463 .80885 L .38207 .79544 L .365 .215 0 r
F
.39506 .8864 m .38236 .8912 L .39594 .88461 L .145 0 0 r
F
.39367 .89412 m .38236 .8912 L .39506 .8864 L .304 .159 0 r
F
.38236 .8912 m .38646 .90107 L .38389 .92775 L .288 .217 0 r
F
.38236 .8912 m .39367 .89412 L .38646 .90107 L .28 .213 0 r
F
.43024 .10395 m .38911 .15673 L .38238 .10364 L .769 .562 0 r
F
.38389 .92775 m .39692 .93325 L .38243 .95608 L .438 .264 0 r
F
.40449 .96539 m .38243 .95608 L .39692 .93325 L .465 .297 0 r
F
.39748 .634 m .38618 .67818 L .38336 .65031 L .474 .296 0 r
F
.39587 .71816 m .3968 .72123 L .38375 .74191 L 0 0 0 r
F
.38375 .74191 m .3968 .72123 L .40221 .72953 L .219 .051 0 r
F
.40221 .72953 m .38782 .75458 L .38375 .74191 L .221 .054 0 r
F
.39536 .90028 m .39692 .93325 L .38389 .92775 L .401 .301 0 r
F
.38389 .92775 m .38646 .90107 L .39536 .90028 L .466 .373 0 r
F
.38395 1.08884 m .40117 1.06552 L .41226 1.07161 L .629 .589 0 r
F
.38395 1.08884 m .41226 1.07161 L .39579 1.09478 L .64 .614 0 r
F
.3972 .85425 m .39672 .85933 L .38523 .86099 L .415 .225 0 r
F
.3856 1.19382 m .40578 1.18448 L .43171 1.18832 L .379 .764 0 r
F
.3856 1.19382 m .43171 1.18832 L .41248 1.19742 L .366 .766 0 r
F
.39599 .98991 m .386 .9871 L .40449 .96539 L .599 .419 0 r
F
.40152 .66441 m .38618 .67818 L .39748 .634 L .526 .347 0 r
F
.38646 .90107 m .39367 .89412 L .39536 .90028 L .472 .515 0 r
F
.38689 .84236 m .39463 .80885 L .40728 .81965 L .435 .326 0 r
F
.40728 .81965 m .40079 .84462 L .38689 .84236 L .448 .343 0 r
F
.38689 .84236 m .40079 .84462 L .3972 .85425 L .455 .314 0 r
F
.38782 .75458 m .40221 .72953 L .40766 .74149 L .329 .176 0 r
F
.40766 .74149 m .395 .76581 L .38782 .75458 L .308 .14 0 r
F
.43024 .10395 m .43504 .15781 L .38911 .15673 L .795 .6 0 r
F
.43504 .15781 m .40147 .22085 L .38911 .15673 L .794 .595 0 r
F
.3899 .70439 m .39728 .68605 L .41261 .68611 L .47 .277 0 r
F
.39135 .70834 m .3899 .70439 L .41261 .68611 L 0 0 0 r
F
.39059 1.06048 m .40755 1.03361 L .42464 1.03937 L .653 .577 0 r
F
.39059 1.06048 m .42464 1.03937 L .40117 1.06552 L .654 .577 0 r
F
.39129 1.02775 m .40913 .99548 L .42383 1.00153 L .633 .5 0 r
F
.39129 1.02775 m .42383 1.00153 L .40755 1.03361 L .657 .542 0 r
F
.41261 .68611 m .41253 .69114 L .39135 .70834 L 0 0 0 r
F
.39135 .70834 m .41253 .69114 L .4133 .70081 L .034 0 0 r
F
.4133 .70081 m .39587 .71816 L .39135 .70834 L .089 0 0 r
F
.39352 1.16439 m .40511 1.14996 L .4298 1.15578 L .629 .723 0 r
F
.39352 1.16439 m .4298 1.15578 L .41867 1.17153 L .629 .721 0 r
F
.39367 .89412 m .39506 .8864 L .41119 .89188 L .457 .296 0 r
F
.41119 .89188 m .40929 .89616 L .39367 .89412 L .462 .319 0 r
F
.39367 .89412 m .40929 .89616 L .40856 .90287 L .438 .422 0 r
F
.40856 .90287 m .39536 .90028 L .39367 .89412 L .354 .432 0 r
F
.39463 .80885 m .40715 .78012 L .41817 .79297 L .45 .307 0 r
F
.41817 .79297 m .40728 .81965 L .39463 .80885 L .461 .325 0 r
F
.395 .76581 m .40766 .74149 L .41748 .75388 L .354 .191 0 r
F
.41748 .75388 m .40715 .78012 L .395 .76581 L .366 .212 0 r
F
.39506 .8864 m .39594 .88461 L .41548 .88789 L .491 .163 0 r
F
.41548 .88789 m .41119 .89188 L .39506 .8864 L .497 .187 0 r
F
.39692 .93325 m .39536 .90028 L .40856 .90287 L .416 .313 0 r
F
.41536 .49671 m .39536 .50586 L .39606 .50019 L .76 .861 0 r
F
.41536 .49671 m .41219 .50134 L .39536 .50586 L .751 .896 0 r
F
.41219 .50134 m .39553 .54042 L .39536 .50586 L .756 .561 0 r
F
.41219 .50134 m .41393 .53329 L .39553 .54042 L .725 .525 0 r
F
.41393 .53329 m .39574 .587 L .39553 .54042 L .729 .535 0 r
F
.39564 1.12142 m .41022 1.10147 L .43118 1.10967 L .642 .621 0 r
F
.39564 1.12142 m .43118 1.10967 L .41959 1.13083 L .647 .65 0 r
F
.41393 .53329 m .4151 .57215 L .39574 .587 L .629 .428 0 r
F
.4151 .57215 m .41256 .61913 L .39574 .587 L .64 .446 0 r
F
.41256 .61913 m .39748 .634 L .39574 .587 L .601 .43 0 r
F
.39579 1.09478 m .41226 1.07161 L .42281 1.07625 L .67 .657 0 r
F
.39579 1.09478 m .42281 1.07625 L .41022 1.10147 L .668 .652 0 r
F
.3968 .72123 m .39587 .71816 L .4133 .70081 L .062 0 0 r
F
.41548 .88789 m .39594 .88461 L .39653 .87483 L .483 .365 0 r
F
.40449 .96539 m .41533 .96676 L .39599 .98991 L .55 .368 0 r
F
.39599 .98991 m .43214 .96528 L .40913 .99548 L .622 .471 0 r
F
.39599 .98991 m .41533 .96676 L .43214 .96528 L .613 .454 0 r
F
.41536 .49671 m .39606 .50019 L .40197 .45746 L .788 .646 0 r
F
.42079 .87658 m .39653 .87483 L .39771 .86763 L .511 .333 0 r
F
.39653 .87483 m .42079 .87658 L .41548 .88789 L .511 .357 0 r
F
.41629 .85762 m .39672 .85933 L .3972 .85425 L .54 .331 0 r
F
.39771 .86763 m .39672 .85933 L .41629 .85762 L .568 .549 0 r
F
.3968 .72123 m .4133 .70081 L .41689 .70888 L .294 .068 0 r
F
.41689 .70888 m .40221 .72953 L .3968 .72123 L .296 .072 0 r
F
.39692 .93325 m .40856 .90287 L .42114 .90951 L .467 .367 0 r
F
.42114 .90951 m .41877 .93949 L .39692 .93325 L .45 .342 0 r
F
.39692 .93325 m .41877 .93949 L .40449 .96539 L .47 .3 0 r
F
.41465 .85353 m .3972 .85425 L .40079 .84462 L .555 .42 0 r
F
.3972 .85425 m .41465 .85353 L .41629 .85762 L .54 .326 0 r
F
.41929 .65265 m .41434 .67016 L .39728 .68605 L .375 .133 0 r
F
.41929 .65265 m .39728 .68605 L .40152 .66441 L .542 .327 0 r
F
.41261 .68611 m .39728 .68605 L .41434 .67016 L .468 .26 0 r
F
.41256 .61913 m .40152 .66441 L .39748 .634 L .56 .37 0 r
F
.41629 .85762 m .41782 .86479 L .39771 .86763 L .496 .388 0 r
F
.42672 .86917 m .39771 .86763 L .41782 .86479 L .503 .429 0 r
F
.39771 .86763 m .42672 .86917 L .42079 .87658 L .506 .34 0 r
F
.41515 .85083 m .40079 .84462 L .40728 .81965 L .539 .433 0 r
F
.40079 .84462 m .41515 .85083 L .41465 .85353 L .552 .421 0 r
F
.40117 1.06552 m .42464 1.03937 L .41226 1.07161 L .655 .579 0 r
F
.43504 .15781 m .44186 .21907 L .40147 .22085 L .804 .608 0 r
F
.44186 .21907 m .4037 .28064 L .40147 .22085 L .809 .65 0 r
F
.41929 .65265 m .40152 .66441 L .41256 .61913 L .578 .388 0 r
F
.44666 .41078 m .40197 .45746 L .40308 .41213 L .8 .66 0 r
F
.44666 .41078 m .44746 .45228 L .40197 .45746 L .788 .639 0 r
F
.44746 .45228 m .43337 .49475 L .40197 .45746 L .788 .644 0 r
F
.43337 .49475 m .41536 .49671 L .40197 .45746 L .783 .644 0 r
F
.40221 .72953 m .41689 .70888 L .42051 .71702 L .292 .068 0 r
F
.42051 .71702 m .40766 .74149 L .40221 .72953 L .373 .196 0 r
F
.44653 .37384 m .40308 .41213 L .4071 .37299 L .806 .662 0 r
F
.44653 .37384 m .44666 .41078 L .40308 .41213 L .8 .65 0 r
F
.44626 .27871 m .40357 .32416 L .4037 .28064 L .783 .625 0 r
F
.44626 .27871 m .44639 .32779 L .40357 .32416 L .798 .651 0 r
F
.44639 .32779 m .4071 .37299 L .40357 .32416 L .799 .645 0 r
F
.44186 .21907 m .44626 .27871 L .4037 .28064 L .782 .613 0 r
F
.41533 .96676 m .40449 .96539 L .41877 .93949 L .548 .391 0 r
F
.40511 1.14996 m .41959 1.13083 L .44327 1.13757 L .669 .706 0 r
F
.40511 1.14996 m .44327 1.13757 L .4298 1.15578 L .669 .701 0 r
F
.40578 1.18448 m .41867 1.17153 L .44384 1.17687 L .599 .763 0 r
F
.40578 1.18448 m .44384 1.17687 L .43171 1.18832 L .591 .78 0 r
F
.44639 .32779 m .44653 .37384 L .4071 .37299 L .806 .656 0 r
F
.40715 .78012 m .41748 .75388 L .43094 .76932 L .503 .356 0 r
F
.43094 .76932 m .41817 .79297 L .40715 .78012 L .483 .313 0 r
F
.40728 .81965 m .41817 .79297 L .42415 .80097 L .5 .367 0 r
F
.42415 .80097 m .41956 .83057 L .40728 .81965 L .543 .428 0 r
F
.40728 .81965 m .41956 .83057 L .41515 .85083 L .53 .428 0 r
F
.40755 1.03361 m .42383 1.00153 L .43873 1.00677 L .661 .546 0 r
F
.40755 1.03361 m .43873 1.00677 L .42464 1.03937 L .668 .559 0 r
F
.40766 .74149 m .42051 .71702 L .42967 .7318 L .43 .259 0 r
F
.42967 .7318 m .41748 .75388 L .40766 .74149 L .401 .2 0 r
F
.42114 .90951 m .40856 .90287 L .40929 .89616 L .406 .392 0 r
F
.40913 .99548 m .43214 .96528 L .42383 1.00153 L .639 .494 0 r
F
.42575 .8895 m .42134 .90145 L .40929 .89616 L .531 .381 0 r
F
.40929 .89616 m .41119 .89188 L .42575 .8895 L .543 .408 0 r
F
.40929 .89616 m .42134 .90145 L .42114 .90951 L .492 .413 0 r
F
.41022 1.10147 m .42281 1.07625 L .4434 1.08556 L .669 .653 0 r
F
.41022 1.10147 m .4434 1.08556 L .43118 1.10967 L .655 .612 0 r
F
.41119 .89188 m .41548 .88789 L .42622 .8876 L .456 .135 0 r
F
.42622 .8876 m .42575 .8895 L .41119 .89188 L .417 .062 0 r
F
.43337 .49475 m .41219 .50134 L .41536 .49671 L .726 .856 0 r
F
.43337 .49475 m .43038 .49937 L .41219 .50134 L .732 .843 0 r
F
.43038 .49937 m .41393 .53329 L .41219 .50134 L .764 .555 0 r
F
.41226 1.07161 m .42464 1.03937 L .44041 1.04652 L .667 .593 0 r
F
.41226 1.07161 m .44041 1.04652 L .42281 1.07625 L .695 .644 0 r
F
.41248 1.19742 m .43171 1.18832 L .45819 1.18963 L .371 .796 0 r
F
.41248 1.19742 m .45819 1.18963 L .44004 1.19878 L .376 .796 0 r
F
.41253 .69114 m .41261 .68611 L .43456 .67344 L 0 0 0 r
F
.43456 .67344 m .43747 .6785 L .41253 .69114 L F
.41253 .69114 m .43747 .6785 L .43965 .68319 L .059 0 0 r
F
.43965 .68319 m .4133 .70081 L .41253 .69114 L .134 0 0 r
F
.42916 .60576 m .41256 .61913 L .4151 .57215 L .622 .43 0 r
F
.42916 .60576 m .41929 .65265 L .41256 .61913 L .607 .407 0 r
F
.43456 .67344 m .41261 .68611 L .41434 .67016 L .498 .286 0 r
F
.43965 .68319 m .41689 .70888 L .4133 .70081 L .38 .106 0 r
F
.43038 .49937 m .43342 .52872 L .41393 .53329 L .727 .51 0 r
F
.43342 .52872 m .4151 .57215 L .41393 .53329 L .724 .503 0 r
F
.43773 .64338 m .41434 .67016 L .41929 .65265 L .57 .316 0 r
F
.43773 .64338 m .43515 .65776 L .41434 .67016 L .522 .255 0 r
F
.41434 .67016 m .43515 .65776 L .43456 .67344 L .498 .217 0 r
F
.41465 .85353 m .41515 .85083 L .43059 .85111 L .551 .42 0 r
F
.41465 .85353 m .43059 .85111 L .4303 .85655 L .552 .427 0 r
F
.4303 .85655 m .41629 .85762 L .41465 .85353 L .567 .339 0 r
F
.43342 .52872 m .43582 .5624 L .4151 .57215 L .685 .459 0 r
F
.43582 .5624 m .42916 .60576 L .4151 .57215 L .686 .461 0 r
F
.41515 .85083 m .41956 .83057 L .43104 .83633 L .606 .501 0 r
F
.43104 .83633 m .43059 .85111 L .41515 .85083 L .551 .414 0 r
F
.41877 .93949 m .44911 .94152 L .41533 .96676 L .598 .435 0 r
F
.43214 .96528 m .41533 .96676 L .44911 .94152 L .614 .463 0 r
F
.42622 .8876 m .41548 .88789 L .42079 .87658 L .506 .351 0 r
F
.41782 .86479 m .41629 .85762 L .4303 .85655 L .59 .45 0 r
F
.42051 .71702 m .41689 .70888 L .43965 .68319 L .38 .106 0 r
F
.41748 .75388 m .42967 .7318 L .44036 .74761 L .537 .354 0 r
F
.44036 .74761 m .43094 .76932 L .41748 .75388 L .54 .362 0 r
F
.4303 .85655 m .42992 .86102 L .41782 .86479 L F
.42672 .86917 m .41782 .86479 L .42992 .86102 L .553 .396 0 r
F
.42415 .80097 m .41817 .79297 L .43094 .76932 L .539 .377 0 r
F
.41867 1.17153 m .4298 1.15578 L .45578 1.15962 L .657 .762 0 r
F
.41867 1.17153 m .45578 1.15962 L .44384 1.17687 L .657 .743 0 r
F
.41877 .93949 m .42114 .90951 L .44932 .92096 L .571 .447 0 r
F
.44932 .92096 m .44911 .94152 L .41877 .93949 L .595 .494 0 r
F
.43773 .64338 m .41929 .65265 L .42916 .60576 L .629 .428 0 r
F
.41956 .83057 m .42415 .80097 L .43535 .81751 L .614 .494 0 r
F
.43535 .81751 m .43104 .83633 L .41956 .83057 L .615 .495 0 r
F
.41959 1.13083 m .43118 1.10967 L .45416 1.11553 L .682 .695 0 r
F
.41959 1.13083 m .45416 1.11553 L .44327 1.13757 L .683 .698 0 r
F
.43965 .68319 m .44627 .68867 L .42051 .71702 L .318 .04 0 r
F
.42051 .71702 m .44627 .68867 L .45178 .69616 L .427 .175 0 r
F
.45178 .69616 m .42967 .7318 L .42051 .71702 L .494 .28 0 r
F
.42079 .87658 m .42672 .86917 L .43791 .87702 L .55 .399 0 r
F
.43791 .87702 m .43084 .8838 L .42079 .87658 L .551 .407 0 r
F
.42079 .87658 m .43084 .8838 L .42622 .8876 L .599 .392 0 r
F
.44932 .92096 m .42114 .90951 L .42134 .90145 L .553 .464 0 r
F
.42134 .90145 m .42575 .8895 L .4534 .90007 L .599 .453 0 r
F
.4534 .90007 m .45019 .90991 L .42134 .90145 L F
.42134 .90145 m .45019 .90991 L .44932 .92096 L F
.44041 1.04652 m .46343 1.07551 L .42281 1.07625 L .689 .636 0 r
F
.42281 1.07625 m .46343 1.07551 L .4434 1.08556 L .688 .645 0 r
F
.42383 1.00153 m .43214 .96528 L .45002 .96917 L .64 .495 0 r
F
.42383 1.00153 m .45002 .96917 L .43873 1.00677 L .668 .535 0 r
F
.44604 .79428 m .42415 .80097 L .43094 .76932 L .636 .471 0 r
F
.43535 .81751 m .42415 .80097 L .44604 .79428 L .648 .504 0 r
F
.42464 1.03937 m .43873 1.00677 L .45423 1.01185 L .679 .571 0 r
F
.42464 1.03937 m .45423 1.01185 L .44041 1.04652 L .685 .581 0 r
F
.4534 .90007 m .42575 .8895 L .42622 .8876 L .648 .273 0 r
F
.453 .89798 m .42622 .8876 L .43084 .8838 L .629 .44 0 r
F
.42622 .8876 m .453 .89798 L .4534 .90007 L .638 .168 0 r
F
.42672 .86917 m .42992 .86102 L .44227 .86645 L .602 .449 0 r
F
.44227 .86645 m .43791 .87702 L .42672 .86917 L .585 .387 0 r
F
.44714 .59748 m .42916 .60576 L .43582 .5624 L .681 .456 0 r
F
.44714 .59748 m .43773 .64338 L .42916 .60576 L .684 .463 0 r
F
.46361 .70834 m .42967 .7318 L .45178 .69616 L .618 .429 0 r
F
.46361 .70834 m .4526 .72795 L .42967 .7318 L .585 .374 0 r
F
.44036 .74761 m .42967 .7318 L .4526 .72795 L .583 .367 0 r
F
.4298 1.15578 m .44327 1.13757 L .46724 1.14294 L .69 .733 0 r
F
.4298 1.15578 m .46724 1.14294 L .45578 1.15962 L .69 .738 0 r
F
.44227 .86645 m .42992 .86102 L .4303 .85655 L .618 .429 0 r
F
.48363 .16156 m .43504 .15781 L .43024 .10395 L .78 .589 0 r
F
.47311 .10738 m .48363 .16156 L .43024 .10395 L .803 .585 0 r
F
.44558 .84114 m .45103 .84343 L .4303 .85655 L .643 .513 0 r
F
.4303 .85655 m .43059 .85111 L .44558 .84114 L .631 .494 0 r
F
.44039 .85475 m .4303 .85655 L .45103 .84343 L .579 .393 0 r
F
.44144 .85927 m .4303 .85655 L .44039 .85475 L .583 .405 0 r
F
.4303 .85655 m .44144 .85927 L .44227 .86645 L .573 .436 0 r
F
.44779 .49126 m .44524 .4974 L .43038 .49937 L .758 .81 0 r
F
.44779 .49126 m .43038 .49937 L .43337 .49475 L .754 .877 0 r
F
.44524 .4974 m .43342 .52872 L .43038 .49937 L .754 .53 0 r
F
.43059 .85111 m .43104 .83633 L .44558 .84114 L .619 .471 0 r
F
.45734 .8926 m .43084 .8838 L .43791 .87702 L .593 .472 0 r
F
.43084 .8838 m .45734 .8926 L .453 .89798 L .605 .452 0 r
F
.43094 .76932 m .44036 .74761 L .45781 .7618 L .598 .425 0 r
F
.43094 .76932 m .45781 .7618 L .44604 .79428 L .61 .462 0 r
F
.43104 .83633 m .43535 .81751 L .46175 .83364 L .653 .533 0 r
F
.46175 .83364 m .44558 .84114 L .43104 .83633 L .636 .433 0 r
F
.43118 1.10967 m .4434 1.08556 L .46581 1.09557 L .701 .668 0 r
F
.43118 1.10967 m .46581 1.09557 L .45416 1.11553 L .703 .677 0 r
F
.43171 1.18832 m .44384 1.17687 L .46928 1.17873 L .612 .824 0 r
F
.43171 1.18832 m .46928 1.17873 L .45819 1.18963 L .608 .831 0 r
F
.45002 .96917 m .43214 .96528 L .44911 .94152 L .64 .497 0 r
F
.44779 .49126 m .43337 .49475 L .44746 .45228 L .773 .626 0 r
F
.44524 .4974 m .45094 .52771 L .43342 .52872 L .771 .55 0 r
F
.45094 .52771 m .43582 .5624 L .43342 .52872 L .764 .519 0 r
F
.46046 .6664 m .43456 .67344 L .43515 .65776 L .535 .246 0 r
F
.43747 .6785 m .43456 .67344 L .46046 .6664 L .116 0 0 r
F
.48566 .22376 m .44186 .21907 L .43504 .15781 L .774 .587 0 r
F
.48363 .16156 m .48566 .22376 L .43504 .15781 L .78 .587 0 r
F
.45661 .6361 m .43515 .65776 L .43773 .64338 L .617 .341 0 r
F
.45661 .6361 m .45681 .64944 L .43515 .65776 L .591 .306 0 r
F
.43515 .65776 m .45681 .64944 L .46046 .6664 L .54 .213 0 r
F
.43535 .81751 m .44604 .79428 L .46317 .80804 L .672 .532 0 r
F
.46317 .80804 m .46175 .83364 L .43535 .81751 L .669 .525 0 r
F
.45094 .52771 m .46053 .55929 L .43582 .5624 L .733 .482 0 r
F
.46053 .55929 m .47014 .59479 L .43582 .5624 L .735 .491 0 r
F
.47014 .59479 m .44714 .59748 L .43582 .5624 L .741 .489 0 r
F
.46046 .6664 m .4617 .66879 L .43747 .6785 L 0 0 0 r
F
.43747 .6785 m .4617 .66879 L .46365 .67166 L .059 0 0 r
F
.46365 .67166 m .43965 .68319 L .43747 .6785 L .106 0 0 r
F
.45661 .6361 m .43773 .64338 L .44714 .59748 L .669 .448 0 r
F
.44227 .86645 m .46235 .87365 L .43791 .87702 L .643 .451 0 r
F
.43791 .87702 m .46235 .87365 L .45899 .88366 L .646 .461 0 r
F
.43791 .87702 m .45899 .88366 L .45734 .8926 L .645 .462 0 r
F
.43873 1.00677 m .45002 .96917 L .46745 .97536 L .687 .555 0 r
F
.43873 1.00677 m .46745 .97536 L .45423 1.01185 L .688 .557 0 r
F
.44627 .68867 m .43965 .68319 L .46365 .67166 L .385 .027 0 r
F
.44004 1.19878 m .45819 1.18963 L .48394 1.19069 L .378 .804 0 r
F
.44004 1.19878 m .48394 1.19069 L .4666 1.19898 L .36 .802 0 r
F
.44036 .74761 m .4526 .72795 L .47204 .73158 L .623 .415 0 r
F
.45781 .7618 m .44036 .74761 L .47204 .73158 L .625 .419 0 r
F
.45103 .84343 m .45216 .84785 L .44039 .85475 L .65 .495 0 r
F
.44734 .85538 m .44039 .85475 L .45216 .84785 L .673 .543 0 r
F
.45198 .86404 m .44144 .85927 L .44039 .85475 L .633 .437 0 r
F
.45453 .85933 m .44039 .85475 L .44734 .85538 L .589 .151 0 r
F
.44039 .85475 m .45453 .85933 L .45198 .86404 L .543 .451 0 r
F
.45423 1.01185 m .48676 1.04039 L .44041 1.04652 L .674 .569 0 r
F
.44041 1.04652 m .48676 1.04039 L .48331 1.06537 L .678 .643 0 r
F
.46343 1.07551 m .44041 1.04652 L .48331 1.06537 L .689 .636 0 r
F
.45198 .86404 m .44227 .86645 L .44144 .85927 L .611 .463 0 r
F
.48673 .28294 m .44626 .27871 L .44186 .21907 L .796 .624 0 r
F
.48566 .22376 m .48673 .28294 L .44186 .21907 L .772 .62 0 r
F
.46235 .87365 m .44227 .86645 L .45198 .86404 L .619 .492 0 r
F
.44327 1.13757 m .45416 1.11553 L .47751 1.12038 L .709 .73 0 r
F
.44327 1.13757 m .47751 1.12038 L .46724 1.14294 L .706 .72 0 r
F
.4434 1.08556 m .46343 1.07551 L .46581 1.09557 L .696 .67 0 r
F
.44384 1.17687 m .45578 1.15962 L .48134 1.15994 L .702 .809 0 r
F
.44384 1.17687 m .48134 1.15994 L .46928 1.17873 L .7 .785 0 r
F
.4602 .49899 m .44524 .4974 L .44779 .49126 L .74 .788 0 r
F
.46893 .52874 m .45094 .52771 L .44524 .4974 L .791 .563 0 r
F
.4602 .49899 m .46893 .52874 L .44524 .4974 L .802 .563 0 r
F
.45103 .84343 m .44558 .84114 L .46175 .83364 L .662 .489 0 r
F
.47473 .77711 m .44604 .79428 L .45781 .7618 L .682 .539 0 r
F
.46317 .80804 m .44604 .79428 L .47473 .77711 L .678 .531 0 r
F
.48992 .32664 m .44639 .32779 L .44626 .27871 L .78 .636 0 r
F
.48673 .28294 m .48992 .32664 L .44626 .27871 L .795 .635 0 r
F
.46365 .67166 m .46522 .67424 L .44627 .68867 L .091 0 0 r
F
.44627 .68867 m .46522 .67424 L .47093 .68106 L .504 .233 0 r
F
.47093 .68106 m .45178 .69616 L .44627 .68867 L .481 .187 0 r
F
.48992 .32664 m .48788 .37441 L .44639 .32779 L .78 .629 0 r
F
.48788 .37441 m .44653 .37384 L .44639 .32779 L .772 .629 0 r
F
.4855 .40801 m .44666 .41078 L .44653 .37384 L .781 .635 0 r
F
.48619 .38358 m .47192 .3961 L .44653 .37384 L .772 .637 0 r
F
.48815 .37727 m .48619 .38358 L .44653 .37384 L F
.48788 .37441 m .48815 .37727 L .44653 .37384 L F
.48557 .413 m .44666 .41078 L .48554 .41051 L .781 .635 0 r
F
.48554 .41051 m .44666 .41078 L .48553 .40992 L F
.48553 .40992 m .44666 .41078 L .4855 .40801 L F
.48557 .413 m .48558 .41324 L .44666 .41078 L .78 .642 0 r
F
.48564 .41771 m .47055 .4318 L .44666 .41078 L F
.48558 .41324 m .48564 .41771 L .44666 .41078 L F
.48525 .45075 m .44738 .44816 L .44666 .41078 L .79 .641 0 r
F
.48533 .44541 m .48533 .44566 L .44666 .41078 L .78 .642 0 r
F
.48534 .44482 m .48533 .44541 L .44666 .41078 L F
.48533 .44566 m .48525 .45075 L .44666 .41078 L F
.47014 .59479 m .48088 .63332 L .44714 .59748 L .742 .491 0 r
F
.48088 .63332 m .45661 .6361 L .44714 .59748 L .74 .492 0 r
F
.44734 .85538 m .45216 .84785 L .45415 .85102 L .745 .641 0 r
F
.45415 .85102 m .45308 .85601 L .44734 .85538 L .672 .49 0 r
F
.44734 .85538 m .45308 .85601 L .45576 .85845 L .644 .298 0 r
F
.45576 .85845 m .45453 .85933 L .44734 .85538 L .581 .052 0 r
F
.48521 .45358 m .44738 .44816 L .48524 .4514 L .79 .641 0 r
F
.48524 .4514 m .44738 .44816 L .48525 .45075 L F
.4852 .45428 m .44738 .44816 L .4852 .45426 L F
.48519 .45443 m .44738 .44816 L .4852 .45428 L F
.48516 .45667 m .44738 .44816 L .48494 .45438 L F
.48516 .45693 m .44738 .44816 L .48516 .45667 L F
.4852 .45417 m .44738 .44816 L .4852 .45413 L F
.4852 .45409 m .44738 .44816 L .48521 .45358 L F
.4852 .45413 m .44738 .44816 L .4852 .45409 L F
.4852 .45421 m .44738 .44816 L .4852 .45417 L F
.4852 .45426 m .44738 .44816 L .4852 .45421 L F
.48516 .45693 m .44746 .45228 L .44738 .44816 L F
.48514 .45763 m .44746 .45228 L .48515 .45739 L F
.48515 .45739 m .44746 .45228 L .48515 .45732 L F
.48515 .45732 m .44746 .45228 L .48516 .45693 L F
.48516 .45693 m .44746 .45228 L .48516 .45693 L F
.44746 .45228 m .46184 .49488 L .44779 .49126 L .805 .652 0 r
F
.48514 .45763 m .46184 .49488 L .44746 .45228 L .79 .645 0 r
F
.46184 .49488 m .4602 .49899 L .44779 .49126 L .658 .782 0 r
F
.44911 .94152 m .44932 .92096 L .4751 .92103 L .701 .582 0 r
F
.4751 .92103 m .47317 .94161 L .44911 .94152 L .699 .578 0 r
F
.44911 .94152 m .47317 .94161 L .45002 .96917 L .699 .544 0 r
F
.4751 .92103 m .44932 .92096 L .45019 .90991 L .7 .542 0 r
F
.46745 .97536 m .45002 .96917 L .47317 .94161 L .696 .54 0 r
F
.45019 .90991 m .4534 .90007 L .47588 .90303 L .683 .542 0 r
F
.47588 .90303 m .47338 .91345 L .45019 .90991 L .682 .539 0 r
F
.45019 .90991 m .47338 .91345 L .4751 .92103 L .676 .566 0 r
F
.48226 .56226 m .46053 .55929 L .45094 .52771 L .811 .525 0 r
F
.46893 .52874 m .48226 .56226 L .45094 .52771 L .787 .533 0 r
F
.46175 .83364 m .4621 .8392 L .45103 .84343 L .648 .467 0 r
F
.45216 .84785 m .45103 .84343 L .4621 .8392 L .663 .504 0 r
F
.46361 .70834 m .45178 .69616 L .47093 .68106 L .644 .429 0 r
F
.45198 .86404 m .45453 .85933 L .46356 .86283 L .61 .532 0 r
F
.46356 .86283 m .46427 .86826 L .45198 .86404 L .607 .501 0 r
F
.46235 .87365 m .45198 .86404 L .46427 .86826 L .614 .492 0 r
F
.4621 .8392 m .46195 .84533 L .45216 .84785 L .627 .446 0 r
F
.45415 .85102 m .45216 .84785 L .46195 .84533 L .672 .617 0 r
F
.47204 .73158 m .4526 .72795 L .46361 .70834 L .623 .418 0 r
F
.47609 .90053 m .453 .89798 L .45734 .8926 L .667 .541 0 r
F
.4534 .90007 m .453 .89798 L .47609 .90053 L .579 .125 0 r
F
.46458 .85186 m .45308 .85601 L .45415 .85102 L .629 .448 0 r
F
.46329 .85681 m .45308 .85601 L .46458 .85186 L .638 .472 0 r
F
.46443 .85907 m .45308 .85601 L .46329 .85681 L .638 .451 0 r
F
.45576 .85845 m .45308 .85601 L .46443 .85907 L .646 .297 0 r
F
.47609 .90053 m .47588 .90303 L .4534 .90007 L .661 .307 0 r
F
.45415 .85102 m .46195 .84533 L .46281 .8484 L .711 .702 0 r
F
.46458 .85186 m .45415 .85102 L .46281 .8484 L .626 .404 0 r
F
.45416 1.11553 m .46581 1.09557 L .4873 1.10203 L .726 .708 0 r
F
.45416 1.11553 m .4873 1.10203 L .47751 1.12038 L .727 .713 0 r
F
.46745 .97536 m .49193 1.01477 L .45423 1.01185 L .671 .538 0 r
F
.45423 1.01185 m .48906 1.0304 L .48676 1.04039 L .648 .57 0 r
F
.45423 1.01185 m .49193 1.01477 L .48906 1.0304 L .669 .559 0 r
F
.46356 .86283 m .45453 .85933 L .45576 .85845 L .626 .107 0 r
F
.46443 .85907 m .46506 .86298 L .45576 .85845 L .684 .498 0 r
F
.46506 .86298 m .46356 .86283 L .45576 .85845 L .503 0 0 r
F
.45578 1.15962 m .46724 1.14294 L .49104 1.14571 L .719 .782 0 r
F
.45578 1.15962 m .49104 1.14571 L .48134 1.15994 L .719 .792 0 r
F
.48088 .63332 m .45681 .64944 L .45661 .6361 L .708 .399 0 r
F
.48088 .63332 m .48109 .63379 L .45681 .64944 L .71 .401 0 r
F
.48109 .63379 m .48101 .64329 L .45681 .64944 L F
.48101 .64329 m .481 .64474 L .45681 .64944 L F
.481 .64474 m .48098 .6467 L .45681 .64944 L F
.45681 .64944 m .48098 .6467 L .48093 .6524 L .669 .306 0 r
F
.45681 .64944 m .48091 .65485 L .4809 .6557 L F
.45681 .64944 m .48093 .6524 L .48091 .65485 L F
.4809 .6557 m .46019 .66512 L .45681 .64944 L .667 .295 0 r
F
.48073 .89671 m .45734 .8926 L .45899 .88366 L .684 .5 0 r
F
.45734 .8926 m .48073 .89671 L .47609 .90053 L .681 .525 0 r
F
.48647 .75625 m .45781 .7618 L .47204 .73158 L .71 .517 0 r
F
.45781 .7618 m .48647 .75625 L .47473 .77711 L .715 .535 0 r
F
.45819 1.18963 m .46928 1.17873 L .49474 1.18019 L .614 .844 0 r
F
.45819 1.18963 m .49474 1.18019 L .48394 1.19069 L .615 .842 0 r
F
.45899 .88366 m .46235 .87365 L .48078 .87692 L .713 .532 0 r
F
.48078 .87692 m .47979 .88652 L .45899 .88366 L .697 .49 0 r
F
.45899 .88366 m .47979 .88652 L .48073 .89671 L .697 .489 0 r
F
.48087 .65901 m .46019 .66512 L .4809 .6557 L .667 .295 0 r
F
.48086 .65944 m .46019 .66512 L .48087 .65901 L F
.48018 .66002 m .46019 .66512 L .48086 .65944 L F
.48018 .66002 m .46046 .6664 L .46019 .66512 L F
.47696 .50158 m .4602 .49899 L .46184 .49488 L .678 .806 0 r
F
.48391 .53092 m .46893 .52874 L .4602 .49899 L .833 .579 0 r
F
.47696 .50158 m .48405 .52671 L .4602 .49899 L .806 .581 0 r
F
.48405 .52671 m .48404 .52703 L .4602 .49899 L F
.48404 .52703 m .48391 .53092 L .4602 .49899 L F
.4795 .66035 m .46046 .6664 L .48018 .66002 L .667 .295 0 r
F
.47924 .66047 m .46046 .6664 L .4795 .66035 L F
.4617 .66879 m .46046 .6664 L .47924 .66047 L 0 0 0 r
F
.48226 .56226 m .47014 .59479 L .46053 .55929 L .813 .536 0 r
F
.4617 .66879 m .48086 .6597 L .48085 .66153 L 0 0 0 r
F
.48085 .66153 m .46365 .67166 L .4617 .66879 L .013 0 0 r
F
.47594 .83129 m .46175 .83364 L .46317 .80804 L .741 .587 0 r
F
.4621 .8392 m .46175 .83364 L .47594 .83129 L .727 .527 0 r
F
.48506 .46325 m .46184 .49488 L .48514 .45763 L .79 .645 0 r
F
.48499 .468 m .46184 .49488 L .48506 .46325 L F
.48499 .468 m .47995 .49748 L .46184 .49488 L F
.47995 .49748 m .47696 .50158 L .46184 .49488 L .655 .805 0 r
F
.47502 .83747 m .47587 .8441 L .46195 .84533 L .728 .543 0 r
F
.46195 .84533 m .4621 .8392 L .47502 .83747 L .716 .52 0 r
F
.46281 .8484 m .46195 .84533 L .47587 .8441 L .732 .715 0 r
F
.47594 .83129 m .47502 .83747 L .4621 .8392 L .708 .489 0 r
F
.46235 .87365 m .46427 .86826 L .48085 .87032 L .721 .609 0 r
F
.48085 .87032 m .48078 .87692 L .46235 .87365 L .71 .552 0 r
F
.46281 .8484 m .47587 .8441 L .47637 .84678 L .708 .597 0 r
F
.46458 .85186 m .46281 .8484 L .47637 .84678 L .67 .421 0 r
F
.47919 .81767 m .46317 .80804 L .47473 .77711 L .729 .589 0 r
F
.46317 .80804 m .47919 .81767 L .47594 .83129 L .735 .585 0 r
F
.47843 .85126 m .47807 .85628 L .46329 .85681 L .698 .507 0 r
F
.46329 .85681 m .46458 .85186 L .47843 .85126 L .715 .549 0 r
F
.46443 .85907 m .46329 .85681 L .47807 .85628 L .692 .473 0 r
F
.46343 1.07551 m .48331 1.06537 L .4873 1.10203 L .708 .694 0 r
F
.46581 1.09557 m .46343 1.07551 L .4873 1.10203 L .737 .699 0 r
F
.47694 .8655 m .46356 .86283 L .46506 .86298 L .211 0 0 r
F
.46427 .86826 m .46356 .86283 L .47694 .8655 L .717 .58 0 r
F
.47093 .68106 m .47479 .68722 L .46361 .70834 L .508 .296 0 r
F
.47788 .69113 m .46361 .70834 L .47479 .68722 L .551 .344 0 r
F
.46361 .70834 m .47788 .69113 L .48344 .69823 L .655 .485 0 r
F
.48996 .71538 m .46361 .70834 L .48344 .69823 L .615 .406 0 r
F
.47204 .73158 m .46361 .70834 L .48996 .71538 L .614 .413 0 r
F
.48345 .66034 m .46522 .67424 L .46365 .67166 L .093 0 0 r
F
.46365 .67166 m .48085 .66153 L .48345 .66034 L 0 0 0 r
F
.47694 .8655 m .48085 .87032 L .46427 .86826 L .722 .603 0 r
F
.47807 .85628 m .478 .86016 L .46443 .85907 L .709 .529 0 r
F
.47836 .86466 m .46506 .86298 L .46443 .85907 L .701 .51 0 r
F
.46443 .85907 m .478 .86016 L .47836 .86466 L .708 .496 0 r
F
.47637 .84678 m .47843 .85126 L .46458 .85186 L .707 .499 0 r
F
.47836 .86466 m .47694 .8655 L .46506 .86298 L .532 .009 0 r
F
.46522 .67424 m .48345 .66034 L .49111 .6634 L .429 .125 0 r
F
.49111 .6634 m .47093 .68106 L .46522 .67424 L .488 .231 0 r
F
.48053 1.19232 m .49742 1.19077 L .4666 1.19898 L .36 .807 0 r
F
.49742 1.19077 m .47831 1.19907 L .4666 1.19898 L .358 .807 0 r
F
.46724 1.14294 m .47751 1.12038 L .50035 1.124 L .727 .745 0 r
F
.46724 1.14294 m .50035 1.124 L .49104 1.14571 L .733 .77 0 r
F
.47317 .94161 m .49665 .98252 L .46745 .97536 L .673 .518 0 r
F
.46745 .97536 m .49393 1.00162 L .49193 1.01477 L .673 .538 0 r
F
.46745 .97536 m .49665 .98252 L .49393 1.00162 L .666 .538 0 r
F
.4833 .549 m .47971 .55586 L .46893 .52874 L .877 .568 0 r
F
.48341 .5458 m .48341 .54582 L .46893 .52874 L .833 .578 0 r
F
.48383 .53316 m .48343 .54524 L .46893 .52874 L F
.48343 .54524 m .48342 .54527 L .46893 .52874 L F
.48342 .54527 m .48342 .5453 L .46893 .52874 L F
.48342 .5453 m .48342 .54552 L .46893 .52874 L F
.48342 .54552 m .48341 .54571 L .46893 .52874 L F
.48341 .54571 m .48341 .5458 L .46893 .52874 L F
.48341 .54585 m .48335 .54738 L .46893 .52874 L F
.48341 .54582 m .48341 .54585 L .46893 .52874 L F
.48335 .54738 m .4833 .549 L .46893 .52874 L F
.48388 .53164 m .46893 .52874 L .48391 .53092 L .833 .579 0 r
F
.48383 .53316 m .46893 .52874 L .48388 .53164 L F
.46928 1.17873 m .48134 1.15994 L .50678 1.16078 L .708 .795 0 r
F
.46928 1.17873 m .50678 1.16078 L .49474 1.18019 L .709 .81 0 r
F
.48226 .56226 m .4829 .56484 L .47014 .59479 L .803 .524 0 r
F
.4829 .56484 m .48288 .5756 L .47014 .59479 L F
.48288 .5756 m .48288 .58026 L .47014 .59479 L F
.48288 .58026 m .48286 .59205 L .47014 .59479 L F
.48286 .59245 m .48285 .59425 L .47014 .59479 L F
.48286 .59205 m .48286 .59245 L .47014 .59479 L F
.48285 .59425 m .48285 .59524 L .47014 .59479 L F
.48285 .59524 m .48017 .59595 L .47014 .59479 L F
.48282 .59625 m .48122 .62869 L .47014 .59479 L .803 .526 0 r
F
.48122 .62869 m .48035 .63141 L .47014 .59479 L F
.48565 .41823 m .47055 .4318 L .48564 .41771 L .78 .642 0 r
F
.4857 .42197 m .47055 .4318 L .48565 .41823 L F
.4857 .42197 m .48534 .44482 L .47055 .4318 L F
.47479 .68722 m .47093 .68106 L .49111 .6634 L .542 .307 0 r
F
.48582 .38476 m .47192 .3961 L .48619 .38358 L .772 .637 0 r
F
.48554 .38566 m .47192 .3961 L .48577 .38494 L F
.48577 .38494 m .47192 .3961 L .48582 .38476 L F
.4852 .38676 m .47192 .3961 L .48554 .38566 L F
.4852 .38676 m .4855 .40801 L .47192 .3961 L F
.49691 .72368 m .47204 .73158 L .48996 .71538 L .644 .459 0 r
F
.47204 .73158 m .49691 .72368 L .50979 .73941 L .674 .547 0 r
F
.47204 .73158 m .50979 .73941 L .49656 .74152 L .683 .513 0 r
F
.47204 .73158 m .49656 .74152 L .48647 .75625 L .686 .509 0 r
F
.53368 .1685 m .48363 .16156 L .47311 .10738 L .807 .588 0 r
F
.52909 .1139 m .53368 .1685 L .47311 .10738 L .815 .585 0 r
F
.50319 .94534 m .47317 .94161 L .4751 .92103 L .748 .622 0 r
F
.47317 .94161 m .50006 .96192 L .49665 .98252 L .684 .522 0 r
F
.47317 .94161 m .50125 .949 L .50006 .96192 L .697 .517 0 r
F
.47317 .94161 m .50319 .94534 L .50125 .949 L .663 .216 0 r
F
.47338 .91345 m .47588 .90303 L .496 .90223 L .736 .593 0 r
F
.496 .90223 m .49622 .9115 L .47338 .91345 L .738 .597 0 r
F
.47338 .91345 m .49622 .9115 L .49799 .91787 L .74 .64 0 r
F
.49799 .91787 m .4751 .92103 L .47338 .91345 L .749 .612 0 r
F
.48818 .79793 m .47473 .77711 L .48647 .75625 L .785 .626 0 r
F
.47473 .77711 m .48818 .79793 L .47919 .81767 L .78 .625 0 r
F
.47479 .68722 m .49111 .6634 L .49931 .6686 L .485 .24 0 r
F
.47788 .69113 m .47479 .68722 L .49931 .6686 L .555 .345 0 r
F
.47502 .83747 m .47594 .83129 L .48918 .83158 L .755 .533 0 r
F
.48918 .83158 m .48961 .84027 L .47502 .83747 L .791 .627 0 r
F
.47587 .8441 m .47502 .83747 L .48961 .84027 L .796 .591 0 r
F
.4751 .92103 m .49799 .91787 L .50319 .94534 L .75 .621 0 r
F
.48961 .84027 m .48793 .84437 L .47587 .8441 L .744 .456 0 r
F
.47637 .84678 m .47587 .8441 L .48793 .84437 L .766 .635 0 r
F
.496 .90223 m .47588 .90303 L .47609 .90053 L .663 .309 0 r
F
.48918 .83158 m .47594 .83129 L .47919 .81767 L .761 .611 0 r
F
.49371 .89983 m .47609 .90053 L .48073 .89671 L .728 .61 0 r
F
.47609 .90053 m .49371 .89983 L .496 .90223 L .681 .371 0 r
F
.47637 .84678 m .48793 .84437 L .48707 .84907 L .775 .724 0 r
F
.47843 .85126 m .47637 .84678 L .48707 .84907 L .813 .539 0 r
F
.4881 .86568 m .47694 .8655 L .47836 .86466 L .544 .022 0 r
F
.48085 .87032 m .47694 .8655 L .4881 .86568 L .784 .609 0 r
F
.48449 .50019 m .47696 .50158 L .47995 .49748 L .639 .782 0 r
F
.48448 .50099 m .47696 .50158 L .48449 .50019 L F
.48428 .51318 m .48393 .52628 L .47696 .50158 L .857 .609 0 r
F
.48431 .511 m .4843 .51203 L .47696 .50158 L .812 .605 0 r
F
.48443 .50353 m .48433 .51034 L .47696 .50158 L F
.48433 .51034 m .48432 .51042 L .47696 .50158 L F
.48432 .51042 m .48432 .51078 L .47696 .50158 L F
.48432 .51078 m .48431 .511 L .47696 .50158 L F
.4843 .51203 m .48428 .51318 L .47696 .50158 L F
.48446 .50219 m .47696 .50158 L .48448 .50099 L .639 .782 0 r
F
.48443 .50353 m .47696 .50158 L .48446 .50219 L F
.47751 1.12038 m .4873 1.10203 L .50876 1.10615 L .744 .736 0 r
F
.47751 1.12038 m .50876 1.10615 L .50035 1.124 L .742 .727 0 r
F
.49931 .6686 m .49919 .67725 L .47788 .69113 L .657 .492 0 r
F
.48344 .69823 m .47788 .69113 L .49919 .67725 L .653 .485 0 r
F
.48979 .85885 m .478 .86016 L .47807 .85628 L .746 .56 0 r
F
.478 .86016 m .48979 .85885 L .48951 .86431 L .753 .599 0 r
F
.48951 .86431 m .47836 .86466 L .478 .86016 L .773 .545 0 r
F
.47807 .85628 m .47843 .85126 L .48864 .85194 L .805 .602 0 r
F
.48864 .85194 m .4885 .85591 L .47807 .85628 L .779 .537 0 r
F
.47807 .85628 m .4885 .85591 L .48979 .85885 L .742 .412 0 r
F
.49742 1.19077 m .47831 1.19907 L .47831 1.19907 L .331 .815 0 r
F
.47831 1.19907 m .50601 1.19178 L .47831 1.19907 L .471 .885 0 r
F
.47831 1.19907 m .50601 1.19178 L .499 1.19542 L F
.47831 1.19907 m .51054 1.18957 L .47831 1.19907 L .367 .837 0 r
F
.47831 1.19907 m .51054 1.18957 L .50601 1.19178 L F
.49742 1.19077 m .51054 1.18957 L .47831 1.19907 L .331 .815 0 r
F
.48951 .86431 m .4881 .86568 L .47836 .86466 L .634 .19 0 r
F
.48707 .84907 m .48864 .85194 L .47843 .85126 L .778 .457 0 r
F
.48882 .82603 m .47919 .81767 L .48818 .79793 L .793 .641 0 r
F
.47919 .81767 m .48882 .82603 L .48918 .83158 L .843 .616 0 r
F
.48323 .55114 m .47971 .55586 L .4833 .549 L .877 .568 0 r
F
.48305 .55641 m .47971 .55586 L .48323 .55114 L F
.48304 .55672 m .48082 .55864 L .47971 .55586 L F
.48304 .55672 m .47971 .55586 L .48305 .55641 L F
.47979 .88652 m .48078 .87692 L .49557 .87657 L .756 .543 0 r
F
.49557 .87657 m .49481 .88674 L .47979 .88652 L .751 .534 0 r
F
.48073 .89671 m .47979 .88652 L .49481 .88674 L .75 .528 0 r
F
.48475 .48317 m .47995 .49748 L .48484 .47747 L .79 .645 0 r
F
.48484 .47747 m .47995 .49748 L .48499 .468 L F
.48442 .48416 m .48473 .48453 L .47995 .49748 L .795 .65 0 r
F
.48473 .48453 m .4847 .48699 L .47995 .49748 L F
.4847 .48699 m .48462 .49189 L .47995 .49748 L F
.48462 .49189 m .48119 .49616 L .47995 .49748 L F
.48461 .49252 m .48461 .49261 L .47995 .49748 L F
.48461 .49261 m .48452 .49819 L .47995 .49748 L F
.48452 .49819 m .48452 .49826 L .47995 .49748 L F
.48452 .49826 m .48449 .50019 L .47995 .49748 L .693 .788 0 r
F
.48285 .59533 m .48017 .59595 L .48285 .59525 L .803 .524 0 r
F
.48285 .59525 m .48017 .59595 L .48285 .59524 L F
.48285 .59564 m .48017 .59595 L .48285 .59533 L F
.48285 .59568 m .48017 .59595 L .48285 .59567 L F
.48285 .59567 m .48017 .59595 L .48285 .59564 L F
.48285 .5957 m .48017 .59595 L .48285 .59568 L F
.48285 .5957 m .48282 .59625 L .48017 .59595 L F
.48018 .66002 m .48086 .65944 L .48086 .6597 L .667 .295 0 r
F
.48112 .63077 m .48035 .63141 L .48122 .62869 L .803 .526 0 r
F
.48112 .63077 m .48088 .63332 L .48035 .63141 L F
.48394 1.19069 m .49742 1.19077 L .48053 1.19232 L .36 .807 0 r
F
.49481 .88674 m .49604 .8947 L .48073 .89671 L .741 .511 0 r
F
.48073 .89671 m .49604 .8947 L .49371 .89983 L .747 .537 0 r
F
.48078 .87692 m .48085 .87032 L .49492 .8705 L .77 .602 0 r
F
.49492 .8705 m .49557 .87657 L .48078 .87692 L .761 .577 0 r
F
.48303 .55701 m .48082 .55864 L .48304 .55672 L .877 .568 0 r
F
.48299 .55811 m .48082 .55864 L .48303 .55701 L F
.48299 .55818 m .48082 .55864 L .48299 .55811 L F
.48298 .55861 m .48082 .55864 L .48299 .55819 L F
.48299 .55819 m .48082 .55864 L .48299 .55818 L F
.48297 .55881 m .48082 .55864 L .48298 .55861 L F
.48297 .55881 m .48185 .56123 L .48082 .55864 L F
.48085 .66153 m .48112 .63077 L .48729 .63108 L .217 .175 0 r
F
.48729 .63108 m .48345 .66034 L .48085 .66153 L 0 0 0 r
F
.4881 .86568 m .49492 .8705 L .48085 .87032 L .769 .576 0 r
F
.48111 .63253 m .48088 .63332 L .48112 .63077 L .803 .526 0 r
F
.48111 .63253 m .4811 .63327 L .48088 .63332 L .811 .535 0 r
F
.4811 .63329 m .4811 .63331 L .48088 .63332 L F
.4811 .63328 m .4811 .63329 L .48088 .63332 L F
.4811 .63328 m .4811 .63328 L .48088 .63332 L F
.4811 .63327 m .4811 .63328 L .48088 .63332 L F
.4811 .63327 m .4811 .63327 L .48088 .63332 L F
.4811 .63327 m .4811 .63327 L .48088 .63332 L F
.4811 .63331 m .4811 .63335 L .48088 .63332 L F
.4811 .63343 m .4811 .63343 L .48088 .63332 L .787 .435 0 r
F
.4811 .63335 m .4811 .63343 L .48088 .63332 L F
.4811 .63343 m .48109 .63383 L .48088 .63332 L F
.48729 .63108 m .48112 .63077 L .48285 .5957 L .219 .145 0 r
F
.48462 .49197 m .48119 .49616 L .48462 .4919 L .795 .65 0 r
F
.48462 .4919 m .48119 .49616 L .48462 .49189 L F
.48462 .49197 m .48461 .49252 L .48119 .49616 L F
.48134 1.15994 m .49104 1.14571 L .51445 1.14805 L .714 .784 0 r
F
.48134 1.15994 m .51445 1.14805 L .50678 1.16078 L .714 .789 0 r
F
.48292 .56029 m .48185 .56123 L .48296 .5591 L .877 .568 0 r
F
.48296 .5591 m .48185 .56123 L .48297 .55881 L F
.48291 .56068 m .48185 .56123 L .48292 .56029 L F
.48291 .56071 m .48185 .56123 L .48291 .5607 L F
.48291 .5607 m .48185 .56123 L .48291 .56068 L F
.48291 .56071 m .48185 .56123 L .48291 .56071 L F
.48291 .56073 m .48185 .56123 L .48291 .56071 L F
.48291 .56102 m .48185 .56123 L .48291 .56073 L F
.48291 .56177 m .48185 .56123 L .48291 .56102 L F
.48291 .56177 m .48226 .56226 L .48185 .56123 L F
.48226 .56226 m .4829 .56382 L .4829 .56484 L .889 .575 0 r
F
.48226 .56226 m .4829 .5635 L .4829 .56382 L F
.4829 .56264 m .4829 .56323 L .48226 .56226 L .877 .575 0 r
F
.4829 .56323 m .4829 .56326 L .48226 .56226 L F
.4829 .56326 m .4829 .5635 L .48226 .56226 L F
.48291 .56189 m .48226 .56226 L .48291 .56177 L .877 .568 0 r
F
.4829 .56247 m .48226 .56226 L .48291 .56189 L F
.4829 .5625 m .48226 .56226 L .4829 .56247 L F
.4829 .56256 m .48226 .56226 L .4829 .5625 L F
.4829 .56264 m .48226 .56226 L .4829 .56256 L F
.48637 .59537 m .48285 .5957 L .48291 .5607 L 0 0 0 r
F
.48729 .63108 m .48285 .5957 L .48637 .59537 L F
.48291 .5607 m .48291 .56068 L .48291 .56068 L .877 .568 0 r
F
.48291 .56068 m .48291 .56068 L .48291 .56068 L F
.4867 .56025 m .48291 .5607 L .48408 .52577 L 0 0 0 r
F
.48637 .59537 m .48291 .5607 L .4867 .56025 L F
.48331 1.06537 m .48676 1.04039 L .50775 1.04139 L .741 .701 0 r
F
.50775 1.04139 m .50925 1.05028 L .48331 1.06537 L .697 .622 0 r
F
.50925 1.05028 m .50925 1.05029 L .48331 1.06537 L F
.48331 1.06537 m .50925 1.05029 L .50926 1.05037 L .733 .721 0 r
F
.48331 1.06537 m .50926 1.05037 L .50928 1.05048 L F
.48331 1.06537 m .50928 1.05048 L .50948 1.05164 L F
.48331 1.06537 m .50948 1.05164 L .50951 1.05181 L F
.48331 1.06537 m .50951 1.05181 L .50952 1.05187 L F
.48331 1.06537 m .50952 1.05187 L .50952 1.05188 L F
.48331 1.06537 m .50952 1.05188 L .51022 1.05603 L F
.48331 1.06537 m .51022 1.05603 L .51057 1.05813 L F
.48331 1.06537 m .51022 1.05603 L .51022 1.05603 L F
.49482 1.06231 m .51003 1.08412 L .48331 1.06537 L .727 .657 0 r
F
.4873 1.10203 m .48331 1.06537 L .51003 1.08412 L .667 .664 0 r
F
.49919 .67725 m .50188 .68969 L .48344 .69823 L .671 .509 0 r
F
.48344 .69823 m .50188 .68969 L .49184 .70699 L .661 .486 0 r
F
.48344 .69823 m .49184 .70699 L .48996 .71538 L .755 .475 0 r
F
.49594 .64897 m .48345 .66034 L .48729 .63108 L .457 .351 0 r
F
.49111 .6634 m .48345 .66034 L .49594 .64897 L .37 .221 0 r
F
.53359 .22454 m .48566 .22376 L .48363 .16156 L .815 .615 0 r
F
.53368 .1685 m .53359 .22454 L .48363 .16156 L .807 .615 0 r
F
.48409 .52489 m .48393 .52628 L .48428 .51318 L .857 .609 0 r
F
.48408 .52564 m .48393 .52628 L .48409 .52489 L F
.48408 .52577 m .48393 .52628 L .48407 .52568 L F
.48407 .5261 m .48393 .52628 L .48408 .52577 L F
.48407 .52615 m .48393 .52628 L .48407 .5261 L F
.48406 .52623 m .48393 .52628 L .48407 .52615 L F
.48406 .52636 m .48393 .52628 L .48406 .52623 L F
.48406 .52636 m .48405 .52671 L .48393 .52628 L F
.50773 1.17946 m .49742 1.19077 L .48394 1.19069 L .633 .862 0 r
F
.49474 1.18019 m .50773 1.17946 L .48394 1.19069 L .63 .873 0 r
F
.48408 .52577 m .48407 .52568 L .48408 .52569 L .857 .609 0 r
F
.48408 .52569 m .48407 .52568 L .48408 .52568 L F
.48408 .52568 m .48407 .52568 L .48408 .52564 L F
.48849 .52469 m .48408 .52577 L .48462 .49197 L .165 .111 0 r
F
.4867 .56025 m .48408 .52577 L .48849 .52469 L .159 .091 0 r
F
.48474 .48421 m .48474 .48434 L .48442 .48416 L .795 .65 0 r
F
.48474 .48434 m .48473 .48453 L .48442 .48416 L F
.48474 .48421 m .48474 .48421 L .48442 .48416 L F
.48474 .48421 m .48474 .48421 L .48442 .48416 L F
.48474 .48421 m .48474 .48421 L .48442 .48416 L F
.48474 .48421 m .48474 .48421 L .48442 .48416 L F
.48474 .48421 m .48474 .48421 L .48442 .48416 L F
.48474 .48415 m .48474 .48421 L .48442 .48416 L F
.4845 .48392 m .48474 .48397 L .48442 .48416 L F
.48474 .48397 m .48474 .48415 L .48442 .48416 L F
.48474 .48397 m .48474 .48397 L .4845 .48392 L F
.48475 .48317 m .48474 .48397 L .4845 .48392 L F
.48833 .48922 m .48462 .49197 L .48516 .45667 L .081 .044 0 r
F
.48849 .52469 m .48462 .49197 L .48833 .48922 L .087 .055 0 r
F
.48516 .45667 m .48494 .45438 L .48519 .45443 L .79 .641 0 r
F
.48914 .4549 m .48516 .45667 L .4857 .42197 L .159 .107 0 r
F
.48833 .48922 m .48516 .45667 L .48914 .4549 L .155 .098 0 r
F
.4873 .38431 m .4852 .38676 L .48879 .3752 L 0 0 0 r
F
.48871 .4201 m .4852 .38676 L .4873 .38431 L F
.48871 .4201 m .4857 .42197 L .4852 .38676 L .014 .007 0 r
F
.53359 .22454 m .53242 .29264 L .48566 .22376 L .816 .635 0 r
F
.53242 .29264 m .48673 .28294 L .48566 .22376 L .786 .631 0 r
F
.48914 .4549 m .4857 .42197 L .48871 .4201 L .01 0 0 r
F
.49737 .59602 m .48637 .59537 L .4867 .56025 L .54 .416 0 r
F
.49628 .63079 m .48637 .59537 L .49737 .59602 L .541 .408 0 r
F
.48729 .63108 m .48637 .59537 L .49628 .63079 L .441 .347 0 r
F
.48647 .75625 m .49656 .74152 L .51765 .76318 L .724 .56 0 r
F
.51765 .76318 m .51574 .76853 L .48647 .75625 L .711 .599 0 r
F
.48647 .75625 m .49991 .78053 L .48818 .79793 L .726 .58 0 r
F
.48647 .75625 m .51574 .76853 L .49991 .78053 L .727 .58 0 r
F
.49775 .56029 m .4867 .56025 L .48849 .52469 L .565 .429 0 r
F
.49737 .59602 m .4867 .56025 L .49775 .56029 L .565 .431 0 r
F
.53287 .33106 m .48992 .32664 L .48673 .28294 L .784 .626 0 r
F
.53242 .29264 m .53287 .33106 L .48673 .28294 L .787 .626 0 r
F
.48676 1.04039 m .48906 1.0304 L .50984 1.03297 L .75 .673 0 r
F
.50984 1.03297 m .50775 1.04139 L .48676 1.04039 L .75 .671 0 r
F
.48707 .84907 m .48793 .84437 L .49635 .8457 L .77 .719 0 r
F
.48707 .84907 m .49635 .8457 L .49407 .85035 L .776 .771 0 r
F
.48864 .85194 m .48707 .84907 L .49407 .85035 L .809 .464 0 r
F
.48729 .63108 m .49628 .63079 L .49594 .64897 L .441 .344 0 r
F
.4873 1.10203 m .51003 1.08412 L .50876 1.10615 L .711 .764 0 r
F
.48879 .3752 m .49126 .37898 L .4873 .38431 L .18 0 0 r
F
.49753 .38508 m .4873 .38431 L .49126 .37898 L .538 .39 0 r
F
.4981 .421 m .48871 .4201 L .4873 .38431 L .57 .452 0 r
F
.4981 .421 m .4873 .38431 L .49753 .38508 L .535 .431 0 r
F
.49194 .36011 m .48788 .37441 L .48953 .33565 L .783 .632 0 r
F
.48908 .37383 m .48788 .37441 L .49194 .36011 L F
.48908 .37384 m .48788 .37441 L .48908 .37383 L F
.48906 .3739 m .48788 .37441 L .48908 .37384 L F
.48906 .3739 m .48788 .37441 L .48906 .3739 L F
.48906 .37394 m .48788 .37441 L .48906 .3739 L F
.48896 .37442 m .48788 .37441 L .48897 .37438 L F
.48905 .37398 m .48788 .37441 L .48906 .37394 L F
.48905 .37399 m .48788 .37441 L .48905 .37398 L F
.48905 .37398 m .48788 .37441 L .48905 .37398 L F
.48905 .37399 m .48788 .37441 L .48905 .37399 L F
.48904 .37402 m .48788 .37441 L .48905 .37399 L F
.48904 .37402 m .48788 .37441 L .48904 .37402 L F
.48903 .37406 m .48788 .37441 L .48904 .37402 L F
.48903 .37407 m .48788 .37441 L .48903 .37406 L F
.48903 .37408 m .48788 .37441 L .48903 .37407 L F
.48902 .37412 m .48788 .37441 L .48903 .37408 L F
.48898 .37433 m .48788 .37441 L .48902 .37412 L F
.48897 .37436 m .48788 .37441 L .48898 .37433 L F
.48897 .37437 m .48788 .37441 L .48897 .37436 L F
.48897 .37437 m .48788 .37441 L .48897 .37437 L F
.48897 .37437 m .48788 .37441 L .48897 .37437 L F
.48897 .37437 m .48788 .37441 L .48897 .37437 L F
.48897 .37438 m .48788 .37441 L .48897 .37437 L F
.48897 .37438 m .48788 .37441 L .48897 .37438 L F
.48816 .37724 m .48815 .37727 L .48788 .37441 L .774 .638 0 r
F
.4887 .37551 m .48816 .37724 L .48788 .37441 L F
.48878 .37524 m .4887 .37551 L .48788 .37441 L F
.48896 .37442 m .48892 .37458 L .48788 .37441 L .783 .636 0 r
F
.48885 .37491 m .48878 .37524 L .48788 .37441 L F
.48886 .3749 m .48885 .37491 L .48788 .37441 L F
.48886 .37488 m .48886 .3749 L .48788 .37441 L F
.4889 .37468 m .48886 .37488 L .48788 .37441 L F
.4889 .37468 m .4889 .37468 L .48788 .37441 L F
.48892 .37461 m .4889 .37468 L .48788 .37441 L F
.48892 .37458 m .48892 .37461 L .48788 .37441 L F
.48793 .84437 m .48961 .84027 L .49583 .84056 L .796 .517 0 r
F
.49583 .84056 m .49635 .8457 L .48793 .84437 L .795 .515 0 r
F
.4971 .8631 m .4881 .86568 L .48951 .86431 L .491 .022 0 r
F
.49492 .8705 m .4881 .86568 L .4971 .8631 L .786 .562 0 r
F
.49935 .81399 m .48818 .79793 L .49991 .78053 L .744 .605 0 r
F
.48818 .79793 m .49935 .81399 L .48882 .82603 L .749 .606 0 r
F
.49823 .49127 m .48833 .48922 L .48914 .4549 L .588 .452 0 r
F
.49813 .52573 m .48833 .48922 L .49823 .49127 L .588 .453 0 r
F
.49813 .52573 m .48849 .52469 L .48833 .48922 L .562 .436 0 r
F
.49775 .56029 m .48849 .52469 L .49813 .52573 L .563 .428 0 r
F
.48864 .85194 m .49493 .85535 L .4885 .85591 L .876 .624 0 r
F
.48979 .85885 m .4885 .85591 L .49493 .85535 L .81 .44 0 r
F
.49407 .85035 m .49493 .85535 L .48864 .85194 L .876 .624 0 r
F
.49834 .45638 m .48914 .4549 L .48871 .4201 L .582 .456 0 r
F
.49834 .45638 m .48871 .4201 L .4981 .421 L .57 .449 0 r
F
.48879 .3752 m .48878 .37524 L .48885 .37491 L .783 .636 0 r
F
.49194 .36011 m .49315 .37246 L .48879 .3752 L 0 0 0 r
F
.48879 .3752 m .49315 .37246 L .49126 .37898 L F
.48882 .82603 m .49935 .81399 L .50676 .82559 L .776 .649 0 r
F
.50676 .82559 m .50188 .83372 L .48882 .82603 L .776 .654 0 r
F
.48918 .83158 m .48882 .82603 L .50188 .83372 L .825 .603 0 r
F
.48906 1.0304 m .49193 1.01477 L .51222 1.01794 L .757 .644 0 r
F
.51222 1.01794 m .50984 1.03297 L .48906 1.0304 L .758 .647 0 r
F
.49823 .49127 m .48914 .4549 L .49834 .45638 L .583 .449 0 r
F
.49583 .84056 m .48918 .83158 L .50188 .83372 L .824 .644 0 r
F
.49583 .84056 m .48961 .84027 L .48918 .83158 L .812 .643 0 r
F
.49708 .86222 m .48951 .86431 L .48979 .85885 L .843 .677 0 r
F
.49708 .86222 m .4971 .8631 L .48951 .86431 L .335 0 0 r
F
.49222 .35945 m .48953 .33565 L .48992 .32664 L .783 .632 0 r
F
.4922 .3595 m .48953 .33565 L .49221 .35949 L F
.49221 .35949 m .48953 .33565 L .49222 .35945 L F
.49217 .35956 m .48953 .33565 L .4922 .3595 L F
.49202 .35992 m .48953 .33565 L .49217 .35956 L F
.49194 .36011 m .48953 .33565 L .49202 .35992 L F
.48979 .85885 m .49493 .85535 L .49708 .86222 L .866 .537 0 r
F
.52512 .33889 m .52262 .34214 L .48992 .32664 L .783 .632 0 r
F
.52262 .34214 m .52183 .34208 L .48992 .32664 L F
.52168 .32991 m .52512 .33889 L .48992 .32664 L F
.50891 .34315 m .50773 .34376 L .48992 .32664 L F
.50901 .3431 m .50891 .34315 L .48992 .32664 L F
.50935 .34292 m .50901 .3431 L .48992 .32664 L F
.51011 .34252 m .50936 .34291 L .48992 .32664 L F
.50936 .34291 m .50935 .34292 L .48992 .32664 L F
.51163 .34173 m .51011 .34252 L .48992 .32664 L F
.51939 .34189 m .51816 .34179 L .48992 .32664 L F
.50325 .3338 m .51239 .34134 L .48992 .32664 L F
.51239 .34134 m .51163 .34173 L .48992 .32664 L F
.52183 .34208 m .51939 .34189 L .48992 .32664 L F
.50238 .34655 m .50123 .34715 L .48992 .32664 L F
.50401 .34197 m .50238 .34655 L .48992 .32664 L F
.50123 .34715 m .49915 .34823 L .48992 .32664 L F
.50644 .34443 m .50633 .34449 L .48992 .32664 L F
.50678 .34426 m .50644 .34443 L .48992 .32664 L F
.50773 .34376 m .50772 .34377 L .48992 .32664 L F
.50772 .34377 m .50769 .34379 L .48992 .32664 L F
.50769 .34379 m .50757 .34385 L .48992 .32664 L F
.50737 .34395 m .50735 .34396 L .48992 .32664 L F
.50757 .34385 m .50737 .34395 L .48992 .32664 L F
.50735 .34396 m .50678 .34426 L .48992 .32664 L F
.49915 .34823 m .49693 .34938 L .48992 .32664 L F
.49677 .34884 m .49551 .35178 L .48992 .32664 L F
.49411 .35503 m .49222 .35945 L .48992 .32664 L F
.49421 .35482 m .49411 .35503 L .48992 .32664 L F
.49551 .35178 m .49421 .35482 L .48992 .32664 L F
.49699 .71192 m .48996 .71538 L .49184 .70699 L .754 .474 0 r
F
.49892 .71319 m .48996 .71538 L .49699 .71192 L .454 .011 0 r
F
.49772 .71963 m .48996 .71538 L .49892 .71319 L .731 .607 0 r
F
.49691 .72368 m .48996 .71538 L .49772 .71963 L .806 .446 0 r
F
.49104 1.14571 m .50035 1.124 L .52331 1.12554 L .729 .765 0 r
F
.49104 1.14571 m .52331 1.12554 L .51445 1.14805 L .73 .77 0 r
F
.49594 .64897 m .50046 .65872 L .49111 .6634 L .575 .431 0 r
F
.49931 .6686 m .49111 .6634 L .50046 .65872 L .478 .244 0 r
F
.49126 .37898 m .49315 .37246 L .49753 .38508 L .396 .374 0 r
F
.49699 .71192 m .49184 .70699 L .50188 .68969 L .666 .493 0 r
F
.51222 1.01794 m .49193 1.01477 L .49393 1.00162 L .762 .623 0 r
F
.49194 .36011 m .49642 .34965 L .49743 .36043 L .05 0 0 r
F
.49743 .36043 m .49315 .37246 L .49194 .36011 L F
.49315 .37246 m .49743 .36043 L .50574 .36979 L .48 .458 0 r
F
.50574 .36979 m .49753 .38508 L .49315 .37246 L .473 .42 0 r
F
.50837 .88962 m .50779 .89618 L .49371 .89983 L .684 .388 0 r
F
.49371 .89983 m .49604 .8947 L .50837 .88962 L .814 .618 0 r
F
.496 .90223 m .49371 .89983 L .50779 .89618 L .677 .372 0 r
F
.5126 1.00517 m .49393 1.00162 L .49665 .98252 L .732 .601 0 r
F
.49393 1.00162 m .5126 1.00517 L .51222 1.01794 L .723 .627 0 r
F
.49407 .85035 m .49635 .8457 L .50135 .8476 L .766 .757 0 r
F
.49407 .85035 m .50135 .8476 L .49871 .85045 L .705 .935 0 r
F
.49957 .85491 m .49493 .85535 L .49407 .85035 L .889 .633 0 r
F
.49407 .85035 m .49871 .85045 L .49957 .85491 L .894 .623 0 r
F
.51942 1.16044 m .50773 1.17946 L .49474 1.18019 L .723 .825 0 r
F
.50678 1.16078 m .51942 1.16044 L .49474 1.18019 L .727 .836 0 r
F
.49481 .88674 m .49557 .87657 L .50574 .87363 L .845 .619 0 r
F
.50574 .87363 m .5067 .88322 L .49481 .88674 L .815 .572 0 r
F
.49604 .8947 m .49481 .88674 L .5067 .88322 L .811 .559 0 r
F
.51057 1.05813 m .51077 1.05931 L .49482 1.06231 L .727 .657 0 r
F
.51077 1.05931 m .51193 1.06616 L .49482 1.06231 L F
.51193 1.06616 m .51003 1.08412 L .49482 1.06231 L F
.4971 .8631 m .50419 .86839 L .49492 .8705 L .804 .582 0 r
F
.49557 .87657 m .49492 .8705 L .50419 .86839 L .814 .616 0 r
F
.49708 .86222 m .49493 .85535 L .49957 .85491 L .857 .533 0 r
F
.50419 .86839 m .50574 .87363 L .49557 .87657 L .873 .732 0 r
F
.50188 .83372 m .50111 .84078 L .49583 .84056 L .851 .692 0 r
F
.49635 .8457 m .49583 .84056 L .50111 .84078 L .832 .544 0 r
F
.49594 .64897 m .49628 .63079 L .50924 .65449 L .607 .481 0 r
F
.50924 .65449 m .50046 .65872 L .49594 .64897 L .628 .455 0 r
F
.50779 .89618 m .51164 .89895 L .496 .90223 L .75 .508 0 r
F
.49622 .9115 m .496 .90223 L .51164 .89895 L .785 .635 0 r
F
.5067 .88322 m .50837 .88962 L .49604 .8947 L .745 .46 0 r
F
.51164 .89895 m .51208 .90554 L .49622 .9115 L .768 .601 0 r
F
.49622 .9115 m .51208 .90554 L .51357 .91469 L .782 .645 0 r
F
.51357 .91469 m .49799 .91787 L .49622 .9115 L .777 .661 0 r
F
.49628 .63079 m .49737 .59602 L .51718 .62765 L .672 .518 0 r
F
.50924 .65449 m .49628 .63079 L .51718 .62765 L .669 .505 0 r
F
.50111 .84078 m .50135 .8476 L .49635 .8457 L .858 .587 0 r
F
.51239 .34134 m .50107 .3518 L .49642 .34965 L .328 .007 0 r
F
.49677 .34947 m .49642 .34965 L .49677 .34884 L .783 .632 0 r
F
.50107 .3518 m .49743 .36043 L .49642 .34965 L .226 .113 0 r
F
.49656 .74152 m .50979 .73941 L .52684 .74432 L .693 .559 0 r
F
.52684 .74432 m .51765 .76318 L .49656 .74152 L .693 .56 0 r
F
.50961 .98358 m .49665 .98252 L .50006 .96192 L .714 .55 0 r
F
.49665 .98252 m .50961 .98358 L .5126 1.00517 L .71 .595 0 r
F
.49693 .34938 m .49677 .34947 L .49677 .34884 L .783 .632 0 r
F
.50683 .71859 m .50751 .72207 L .49691 .72368 L .591 .238 0 r
F
.49691 .72368 m .49772 .71963 L .50683 .71859 L .585 .229 0 r
F
.49691 .72368 m .50751 .72207 L .51022 .72923 L .683 .482 0 r
F
.51022 .72923 m .50979 .73941 L .49691 .72368 L .637 .54 0 r
F
.50188 .68969 m .50255 .70331 L .49699 .71192 L .594 .423 0 r
F
.49892 .71319 m .49699 .71192 L .50255 .70331 L .315 .094 0 r
F
.49957 .85491 m .50037 .86069 L .49708 .86222 L .911 .604 0 r
F
.50639 .85788 m .50323 .86044 L .49708 .86222 L .513 .085 0 r
F
.50323 .86044 m .4971 .8631 L .49708 .86222 L .274 0 0 r
F
.4971 .8631 m .50323 .86044 L .50911 .86571 L .838 .589 0 r
F
.50911 .86571 m .50419 .86839 L .4971 .8631 L .833 .543 0 r
F
.5159 .59292 m .49737 .59602 L .49775 .56029 L .619 .475 0 r
F
.51718 .62765 m .49737 .59602 L .5159 .59292 L .625 .503 0 r
F
.49742 1.19077 m .49742 1.19077 L .49742 1.19077 L .331 .815 0 r
F
.49742 1.19077 m .49742 1.19077 L .49742 1.19077 L F
.49742 1.19077 m .51224 1.18807 L .51054 1.18957 L .362 .834 0 r
F
.49742 1.19077 m .51224 1.18807 L .49742 1.19077 L F
.51314 1.1822 m .51224 1.18807 L .49742 1.19077 L .767 .974 0 r
F
.51314 1.1822 m .49742 1.19077 L .49742 1.19077 L F
.50773 1.17946 m .51314 1.1822 L .49742 1.19077 L .678 .977 0 r
F
.50614 .35464 m .49743 .36043 L .50107 .3518 L .38 .278 0 r
F
.50836 .36375 m .49743 .36043 L .50614 .35464 L .478 .462 0 r
F
.50574 .36979 m .49743 .36043 L .50836 .36375 L .482 .459 0 r
F
.49753 .38508 m .50574 .36979 L .51974 .37244 L .663 .655 0 r
F
.51974 .37244 m .51649 .3872 L .49753 .38508 L .672 .681 0 r
F
.51513 .42521 m .49753 .38508 L .51649 .3872 L .711 .569 0 r
F
.51513 .42521 m .4981 .421 L .49753 .38508 L .702 .565 0 r
F
.50409 .71438 m .49772 .71963 L .49892 .71319 L .602 .483 0 r
F
.49772 .71963 m .50409 .71438 L .50683 .71859 L .693 .648 0 r
F
.515 .55874 m .49775 .56029 L .49813 .52573 L .601 .459 0 r
F
.5159 .59292 m .49775 .56029 L .515 .55874 L .603 .468 0 r
F
.50319 .94534 m .49799 .91787 L .51357 .91469 L .773 .636 0 r
F
.51456 .4595 m .4981 .421 L .51513 .42521 L .708 .549 0 r
F
.51456 .4595 m .49834 .45638 L .4981 .421 L .675 .534 0 r
F
.51493 .52624 m .49813 .52573 L .49823 .49127 L .619 .477 0 r
F
.515 .55874 m .49813 .52573 L .51493 .52624 L .618 .467 0 r
F
.51415 .49334 m .49823 .49127 L .49834 .45638 L .646 .501 0 r
F
.51493 .52624 m .49823 .49127 L .51415 .49334 L .648 .49 0 r
F
.51415 .49334 m .49834 .45638 L .51456 .4595 L .679 .516 0 r
F
.49871 .85045 m .50768 .84661 L .50349 .85014 L .758 .906 0 r
F
.49871 .85045 m .50135 .8476 L .50768 .84661 L .709 .949 0 r
F
.49871 .85045 m .50349 .85014 L .50435 .85406 L .834 .572 0 r
F
.50435 .85406 m .49957 .85491 L .49871 .85045 L .829 .582 0 r
F
.50255 .70331 m .5035 .7105 L .49892 .71319 L .623 .404 0 r
F
.49892 .71319 m .5035 .7105 L .50409 .71438 L .622 .401 0 r
F
.499 1.19542 m .50601 1.19178 L .51249 1.19807 L .417 .515 0 r
F
.50625 1.19775 m .499 1.19542 L .51249 1.19807 L .043 .557 0 r
F
.51099 .66276 m .5137 .67144 L .49919 .67725 L .638 .471 0 r
F
.49919 .67725 m .49931 .6686 L .51099 .66276 L .648 .484 0 r
F
.50188 .68969 m .49919 .67725 L .5137 .67144 L .647 .494 0 r
F
.49931 .6686 m .50046 .65872 L .50924 .65449 L .552 .307 0 r
F
.50924 .65449 m .51099 .66276 L .49931 .6686 L .565 .322 0 r
F
.49991 .78053 m .51031 .80257 L .49935 .81399 L .733 .596 0 r
F
.49935 .81399 m .51031 .80257 L .5189 .81848 L .756 .635 0 r
F
.5189 .81848 m .50676 .82559 L .49935 .81399 L .753 .643 0 r
F
.50126 .86028 m .49957 .85491 L .50435 .85406 L .812 .531 0 r
F
.50037 .86069 m .49957 .85491 L .50046 .86065 L F
.50071 .86053 m .49957 .85491 L .50126 .86028 L F
.50046 .86065 m .49957 .85491 L .50071 .86053 L F
.51546 .77395 m .49991 .78053 L .51574 .76853 L .723 .573 0 r
F
.51589 .77939 m .49991 .78053 L .51546 .77395 L .718 .558 0 r
F
.51934 .78683 m .49991 .78053 L .51589 .77939 L .722 .624 0 r
F
.51031 .80257 m .49991 .78053 L .51934 .78683 L .738 .599 0 r
F
.51555 .96273 m .50006 .96192 L .50125 .949 L .743 .558 0 r
F
.50006 .96192 m .51555 .96273 L .50961 .98358 L .743 .563 0 r
F
.50035 1.124 m .50876 1.10615 L .53127 1.10669 L .753 .741 0 r
F
.50035 1.124 m .53127 1.10669 L .52331 1.12554 L .75 .731 0 r
F
.50614 .35464 m .50107 .3518 L .51239 .34134 L .485 .208 0 r
F
.50611 .84214 m .50111 .84078 L .50188 .83372 L .838 .679 0 r
F
.50135 .8476 m .50111 .84078 L .50611 .84214 L .847 .578 0 r
F
.50125 .949 m .50319 .94534 L .51486 .94803 L .589 .136 0 r
F
.50125 .949 m .51486 .94803 L .51555 .96273 L .76 .555 0 r
F
.50325 .85627 m .5013 .86026 L .50126 .86028 L .994 .875 0 r
F
.5013 .86026 m .5013 .86026 L .50325 .85627 L F
.5013 .86025 m .5013 .86026 L .50325 .85627 L F
.50328 .85623 m .5014 .86021 L .5013 .86025 L F
.50328 .85623 m .5013 .86025 L .50325 .85627 L F
.50611 .84214 m .50768 .84661 L .50135 .8476 L .832 .553 0 r
F
.50328 .85623 m .50169 .86007 L .5014 .86021 L .994 .875 0 r
F
.50328 .85623 m .50358 .85736 L .50169 .86007 L F
.50358 .85736 m .50177 .86004 L .50169 .86007 L .013 0 0 r
F
.50358 .85736 m .50186 .85999 L .50177 .86004 L F
.50358 .85736 m .50225 .85981 L .50186 .85999 L F
.5137 .67144 m .5145 .67959 L .50188 .68969 L .602 .437 0 r
F
.50188 .68969 m .5145 .67959 L .5169 .68924 L .674 .559 0 r
F
.5169 .68924 m .50255 .70331 L .50188 .68969 L .667 .479 0 r
F
.50188 .83372 m .50676 .82559 L .51251 .83602 L .797 .683 0 r
F
.50611 .84214 m .50188 .83372 L .51251 .83602 L .802 .668 0 r
F
.50646 .85499 m .50234 .85977 L .50225 .85981 L .013 0 0 r
F
.50358 .85736 m .50646 .85499 L .50225 .85981 L F
.50646 .85499 m .50264 .85963 L .50234 .85977 L F
.5026 .34643 m .50238 .34655 L .50401 .34197 L .783 .632 0 r
F
.5169 .68924 m .51503 .70065 L .50255 .70331 L .68 .499 0 r
F
.50255 .70331 m .51503 .70065 L .5035 .7105 L .656 .427 0 r
F
.50269 .34639 m .5026 .34643 L .50401 .34197 L .783 .632 0 r
F
.51085 .85139 m .50297 .85947 L .50264 .85963 L .434 .092 0 r
F
.50646 .85499 m .50537 .85689 L .50264 .85963 L .013 0 0 r
F
.50309 .34618 m .50269 .34639 L .50401 .34197 L .783 .632 0 r
F
.51085 .85139 m .50342 .85927 L .50297 .85947 L .434 .092 0 r
F
.50336 .34604 m .50309 .34618 L .50401 .34197 L .783 .632 0 r
F
.51357 .91469 m .52755 .94197 L .50319 .94534 L .762 .623 0 r
F
.50319 .94534 m .52755 .94197 L .52601 .94813 L .69 .374 0 r
F
.51486 .94803 m .50319 .94534 L .52601 .94813 L .527 0 0 r
F
.50911 .86571 m .50323 .86044 L .50639 .85788 L .854 .572 0 r
F
.50435 .85406 m .50328 .85623 L .50325 .85627 L .994 .875 0 r
F
.51816 .34179 m .51584 .34161 L .50325 .3338 L .783 .632 0 r
F
.51584 .34161 m .51282 .34137 L .50325 .3338 L F
.51282 .34137 m .51239 .34134 L .50325 .3338 L F
.50435 .85406 m .50408 .85522 L .50328 .85623 L .994 .875 0 r
F
.50408 .85522 m .50358 .85736 L .50328 .85623 L F
.50423 .34559 m .50336 .34604 L .50401 .34197 L .783 .632 0 r
F
.51085 .85139 m .50639 .85788 L .50342 .85927 L .434 .092 0 r
F
.50349 .85014 m .50768 .84661 L .51217 .84959 L .697 .73 0 r
F
.50349 .85014 m .51217 .84959 L .50715 .85168 L .711 .674 0 r
F
.50715 .85168 m .50435 .85406 L .50349 .85014 L .799 .55 0 r
F
.5035 .7105 m .51503 .70065 L .51702 .71068 L .624 .378 0 r
F
.51702 .71068 m .50409 .71438 L .5035 .7105 L .629 .406 0 r
F
.50715 .85168 m .51217 .84959 L .50358 .85736 L .544 .272 0 r
F
.51217 .84959 m .51085 .85139 L .50358 .85736 L .116 0 0 r
F
.50358 .85736 m .50435 .85406 L .50715 .85168 L .887 .723 0 r
F
.50606 .34463 m .50423 .34559 L .50401 .34197 L .783 .632 0 r
F
.50633 .34449 m .50606 .34463 L .50401 .34197 L F
.51779 .71601 m .50409 .71438 L .51702 .71068 L .693 .602 0 r
F
.51714 .7197 m .50683 .71859 L .50409 .71438 L .669 .639 0 r
F
.50409 .71438 m .51779 .71601 L .51714 .7197 L .683 .631 0 r
F
.50911 .86571 m .50574 .87363 L .50419 .86839 L .916 .749 0 r
F
.50646 .85499 m .51085 .85139 L .50537 .85689 L .013 0 0 r
F
.51359 .87667 m .50574 .87363 L .51363 .86394 L .802 .518 0 r
F
.51363 .86394 m .50574 .87363 L .50911 .86571 L .969 .838 0 r
F
.50574 .87363 m .51359 .87667 L .5067 .88322 L .796 .558 0 r
F
.50574 .36979 m .50836 .36375 L .52296 .36415 L .644 .646 0 r
F
.52296 .36415 m .51974 .37244 L .50574 .36979 L .648 .671 0 r
F
.50601 1.19178 m .51054 1.18957 L .51576 1.19678 L .229 .343 0 r
F
.51249 1.19807 m .50601 1.19178 L .51576 1.19678 L .041 .357 0 r
F
.51251 .83602 m .51808 .84617 L .50611 .84214 L .774 .616 0 r
F
.50768 .84661 m .50611 .84214 L .51808 .84617 L .794 .535 0 r
F
.51239 .34134 m .51941 .34713 L .50614 .35464 L .505 .229 0 r
F
.50836 .36375 m .50614 .35464 L .51941 .34713 L .707 .611 0 r
F
.51001 1.20091 m .50625 1.19775 L .52006 1.20096 L .413 .406 0 r
F
.50625 1.19775 m .51249 1.19807 L .52006 1.20096 L .072 .536 0 r
F
.51132 .8604 m .50639 .85788 L .51085 .85139 L .834 .583 0 r
F
.50639 .85788 m .51132 .8604 L .50911 .86571 L .838 .566 0 r
F
.50837 .88962 m .5067 .88322 L .51359 .87667 L .727 .449 0 r
F
.50676 .82559 m .5189 .81848 L .52551 .83496 L .745 .625 0 r
F
.52551 .83496 m .51251 .83602 L .50676 .82559 L .704 .65 0 r
F
.52636 1.14787 m .51942 1.16044 L .50678 1.16078 L .735 .829 0 r
F
.51445 1.14805 m .52636 1.14787 L .50678 1.16078 L .731 .813 0 r
F
.50751 .72207 m .50683 .71859 L .51714 .7197 L .654 .28 0 r
F
.51714 .7197 m .51756 .72268 L .50751 .72207 L .592 .173 0 r
F
.51022 .72923 m .50751 .72207 L .51756 .72268 L .674 .477 0 r
F
.51808 .84617 m .51217 .84959 L .50768 .84661 L .814 .706 0 r
F
.52483 1.17458 m .51314 1.1822 L .50773 1.17946 L .61 .953 0 r
F
.52483 1.17458 m .50773 1.17946 L .50773 1.17946 L F
.50773 1.17946 m .53102 1.16972 L .50773 1.17946 L .634 .96 0 r
F
.50773 1.17946 m .53102 1.16972 L .52483 1.17458 L F
.50773 1.17946 m .53173 1.16459 L .53102 1.16972 L .772 .942 0 r
F
.50773 1.17946 m .53173 1.16459 L .50773 1.17946 L F
.50773 1.17946 m .50773 1.17946 L .50773 1.17946 L .759 .959 0 r
F
.50773 1.17946 m .53105 1.15912 L .53173 1.16459 L F
.50773 1.17946 m .53105 1.15912 L .50773 1.17946 L F
.50773 1.17946 m .50773 1.17946 L .50773 1.17946 L F
.50773 1.17946 m .50773 1.17946 L .50773 1.17946 L F
.51942 1.16044 m .53105 1.15912 L .50773 1.17946 L .745 .856 0 r
F
.50775 1.04139 m .50984 1.03297 L .51042 1.03236 L 0 0 0 r
F
.51042 1.03236 m .51045 1.05519 L .50775 1.04139 L F
.50775 1.04139 m .51045 1.05519 L .51193 1.06616 L F
.52075 .88854 m .50779 .89618 L .50837 .88962 L .656 .363 0 r
F
.50779 .89618 m .52075 .88854 L .52256 .89419 L .74 .511 0 r
F
.52256 .89419 m .51164 .89895 L .50779 .89618 L .741 .516 0 r
F
.51941 .34713 m .52296 .36415 L .50836 .36375 L .662 .551 0 r
F
.51359 .87667 m .51741 .88421 L .50837 .88962 L .718 .439 0 r
F
.50837 .88962 m .51741 .88421 L .52075 .88854 L .631 .296 0 r
F
.50876 1.10615 m .51003 1.08412 L .52169 1.09142 L .637 .7 0 r
F
.50876 1.10615 m .52169 1.09142 L .53127 1.10669 L .7 .809 0 r
F
.51363 .86394 m .50911 .86571 L .51132 .8604 L .953 .728 0 r
F
.51718 .62765 m .51765 .6534 L .50924 .65449 L .719 .557 0 r
F
.50924 .65449 m .51765 .6534 L .51099 .66276 L .672 .391 0 r
F
.5214 .96769 m .50961 .98358 L .51555 .96273 L .398 .218 0 r
F
.52616 .97402 m .50961 .98358 L .5214 .96769 L .455 .289 0 r
F
.52715 .98112 m .50961 .98358 L .52616 .97402 L .449 .277 0 r
F
.52704 .99178 m .50961 .98358 L .52715 .98112 L .464 .342 0 r
F
.5126 1.00517 m .50961 .98358 L .52704 .99178 L .409 .377 0 r
F
.52241 .73312 m .50979 .73941 L .51022 .72923 L .657 .557 0 r
F
.52684 .74432 m .50979 .73941 L .52241 .73312 L .672 .594 0 r
F
.51222 1.01794 m .51042 1.03236 L .50984 1.03297 L 0 0 0 r
F
.51619 1.06972 m .51003 1.08412 L .51193 1.06616 L .727 .657 0 r
F
.51003 1.08412 m .51811 1.07132 L .51923 1.07225 L .707 .634 0 r
F
.51003 1.08412 m .51531 1.07178 L .51811 1.07132 L F
.51907 1.07246 m .52169 1.09142 L .51003 1.08412 L .744 .69 0 r
F
.51756 .72268 m .52241 .73312 L .51022 .72923 L .687 .498 0 r
F
.51934 .78683 m .52491 .79305 L .51031 .80257 L .608 .436 0 r
F
.51031 .80257 m .52491 .79305 L .52726 .81698 L .693 .601 0 r
F
.52726 .81698 m .5189 .81848 L .51031 .80257 L .671 .602 0 r
F
.51042 1.03236 m .51222 1.01794 L .52375 1.02031 L .323 .163 0 r
F
.52375 1.02031 m .52247 1.03301 L .51042 1.03236 L .428 .319 0 r
F
.51045 1.05519 m .51042 1.03236 L .52247 1.03301 L .428 .323 0 r
F
.52247 1.03301 m .52513 1.03868 L .51045 1.05519 L .341 .223 0 r
F
.51045 1.05519 m .52513 1.03868 L .52001 1.05716 L .34 .222 0 r
F
.51193 1.06616 m .51045 1.05519 L .52001 1.05716 L .334 .24 0 r
F
.51054 1.18957 m .51224 1.18807 L .51843 1.19263 L 0 .358 0 r
F
.51576 1.19678 m .51054 1.18957 L .51843 1.19263 L .137 .3 0 r
F
.51322 .85483 m .51085 .85139 L .51217 .84959 L .956 .78 0 r
F
.51085 .85139 m .51322 .85483 L .51132 .8604 L .983 .711 0 r
F
.51099 .66276 m .51765 .6534 L .52274 .66705 L .743 .483 0 r
F
.51099 .66276 m .52274 .66705 L .5137 .67144 L .733 .524 0 r
F
.51363 .86394 m .51132 .8604 L .51322 .85483 L .972 .686 0 r
F
.51164 .89895 m .52256 .89419 L .52619 .89962 L .772 .59 0 r
F
.52619 .89962 m .51208 .90554 L .51164 .89895 L .773 .605 0 r
F
.51951 1.07248 m .51193 1.06616 L .52001 1.05716 L .38 .305 0 r
F
.51208 .90554 m .52619 .89962 L .52839 .90554 L .781 .63 0 r
F
.52839 .90554 m .51357 .91469 L .51208 .90554 L .781 .645 0 r
F
.51217 .84959 m .51808 .84617 L .51322 .85483 L .813 .704 0 r
F
.52637 1.00471 m .52375 1.02031 L .51222 1.01794 L .303 .227 0 r
F
.51222 1.01794 m .5126 1.00517 L .52637 1.00471 L .368 .327 0 r
F
.51224 1.18807 m .51314 1.1822 L .51914 1.18827 L .122 .368 0 r
F
.51843 1.19263 m .51224 1.18807 L .51914 1.18827 L .109 .421 0 r
F
.51941 .34713 m .51239 .34134 L .52262 .34214 L .638 .14 0 r
F
.51249 1.19807 m .51576 1.19678 L .52511 1.19926 L .16 .631 0 r
F
.52006 1.20096 m .51249 1.19807 L .52511 1.19926 L .315 .615 0 r
F
.52551 .83496 m .51808 .84617 L .51251 .83602 L .704 .592 0 r
F
.52704 .99178 m .52637 1.00471 L .5126 1.00517 L .368 .312 0 r
F
.51314 1.1822 m .52483 1.17458 L .5293 1.1803 L .562 .55 0 r
F
.51914 1.18827 m .51314 1.1822 L .5293 1.1803 L .561 .513 0 r
F
.51808 .84617 m .51851 .85644 L .51322 .85483 L .711 .568 0 r
F
.51322 .85483 m .51851 .85644 L .51363 .86394 L .734 .483 0 r
F
.52755 .94197 m .51357 .91469 L .52839 .90554 L .774 .628 0 r
F
.51882 .87059 m .51359 .87667 L .51363 .86394 L .8 .516 0 r
F
.51741 .88421 m .51359 .87667 L .51882 .87059 L .755 .451 0 r
F
.51882 .87059 m .51363 .86394 L .51851 .85644 L .761 .518 0 r
F
.5145 .67959 m .5137 .67144 L .52274 .66705 L .738 .537 0 r
F
.53027 .48589 m .51415 .49334 L .51456 .4595 L .665 .505 0 r
F
.53237 .52093 m .51415 .49334 L .53027 .48589 L .675 .527 0 r
F
.53237 .52093 m .51493 .52624 L .51415 .49334 L .703 .535 0 r
F
.51445 1.14805 m .51445 1.14805 L .51445 1.14805 L .731 .813 0 r
F
.51445 1.14805 m .52636 1.14787 L .51445 1.14805 L F
.51445 1.14805 m .51445 1.14805 L .51445 1.14805 L F
.53479 1.12542 m .52823 1.14288 L .51445 1.14805 L .748 .791 0 r
F
.52331 1.12554 m .53479 1.12542 L .51445 1.14805 L .748 .79 0 r
F
.52823 1.14288 m .52636 1.14787 L .51445 1.14805 L .748 .791 0 r
F
.52274 .66705 m .53067 .67773 L .5145 .67959 L .787 .603 0 r
F
.5145 .67959 m .53067 .67773 L .53013 .68474 L .79 .62 0 r
F
.53013 .68474 m .5169 .68924 L .5145 .67959 L .787 .628 0 r
F
.52839 .452 m .51456 .4595 L .51513 .42521 L .62 .476 0 r
F
.53027 .48589 m .51456 .4595 L .52839 .452 L .627 .492 0 r
F
.52827 .95557 m .51555 .96273 L .51486 .94803 L .756 .552 0 r
F
.51486 .94803 m .52601 .94813 L .52827 .95557 L .773 .527 0 r
F
.53431 .55605 m .51493 .52624 L .53237 .52093 L .713 .565 0 r
F
.53431 .55605 m .515 .55874 L .51493 .52624 L .751 .574 0 r
F
.53652 .59015 m .515 .55874 L .53431 .55605 L .756 .6 0 r
F
.53652 .59015 m .5159 .59292 L .515 .55874 L .772 .603 0 r
F
.52626 .69567 m .51503 .70065 L .5169 .68924 L .746 .562 0 r
F
.51503 .70065 m .52626 .69567 L .52751 .70304 L .616 .302 0 r
F
.52751 .70304 m .51702 .71068 L .51503 .70065 L .621 .376 0 r
F
.52474 .41773 m .51513 .42521 L .51649 .3872 L .639 .508 0 r
F
.52839 .452 m .51513 .42521 L .52474 .41773 L .62 .477 0 r
F
.51531 1.07178 m .51619 1.06972 L .51811 1.07132 L .707 .634 0 r
F
.51851 .77329 m .51546 .77395 L .51574 .76853 L .163 .101 0 r
F
.51546 .77395 m .51851 .77329 L .51589 .77939 L .174 .146 0 r
F
.5214 .96769 m .51555 .96273 L .52827 .95557 L .543 .188 0 r
F
.51765 .76318 m .52147 .7646 L .51574 .76853 L .365 .231 0 r
F
.51574 .76853 m .52147 .7646 L .52288 .76709 L .228 .01 0 r
F
.51574 .76853 m .52288 .76709 L .51851 .77329 L .268 .14 0 r
F
.51576 1.19678 m .51843 1.19263 L .52733 1.19671 L .387 .641 0 r
F
.52511 1.19926 m .51576 1.19678 L .52733 1.19671 L .304 .689 0 r
F
.51934 .78683 m .51589 .77939 L .51851 .77329 L .493 .509 0 r
F
.53776 .62538 m .5159 .59292 L .53652 .59015 L .774 .615 0 r
F
.53776 .62538 m .51718 .62765 L .5159 .59292 L .765 .613 0 r
F
.51649 .3872 m .51974 .37244 L .53212 .37321 L .765 .773 0 r
F
.53212 .37321 m .52567 .38788 L .51649 .3872 L .759 .759 0 r
F
.52474 .41773 m .51649 .3872 L .52567 .38788 L .794 .6 0 r
F
.52626 .69567 m .5169 .68924 L .53013 .68474 L .761 .552 0 r
F
.51779 .71601 m .51702 .71068 L .52751 .70304 L .779 .662 0 r
F
.52876 .71802 m .51714 .7197 L .51779 .71601 L .74 .686 0 r
F
.51756 .72268 m .51714 .7197 L .52876 .71802 L .566 .154 0 r
F
.53776 .62538 m .52926 .65546 L .51718 .62765 L .762 .591 0 r
F
.52926 .65546 m .51765 .6534 L .51718 .62765 L .761 .591 0 r
F
.51882 .87059 m .52357 .87763 L .51741 .88421 L .696 .398 0 r
F
.52075 .88854 m .51741 .88421 L .52357 .87763 L .619 .295 0 r
F
.52876 .71802 m .52985 .72376 L .51756 .72268 L .678 .332 0 r
F
.51756 .72268 m .52985 .72376 L .52241 .73312 L .71 .508 0 r
F
.52388 .76053 m .51765 .76318 L .52684 .74432 L .667 .529 0 r
F
.51765 .76318 m .52388 .76053 L .52771 .76195 L .436 .086 0 r
F
.52771 .76195 m .52147 .7646 L .51765 .76318 L F
.52274 .66705 m .51765 .6534 L .52926 .65546 L .762 .492 0 r
F
.52751 .70304 m .52942 .71272 L .51779 .71601 L .731 .592 0 r
F
.51779 .71601 m .52942 .71272 L .52876 .71802 L .749 .67 0 r
F
.51808 .84617 m .52551 .83496 L .52572 .84788 L .668 .544 0 r
F
.52452 .85808 m .51851 .85644 L .51808 .84617 L .701 .56 0 r
F
.51808 .84617 m .52572 .84788 L .52452 .85808 L .667 .547 0 r
F
.51843 1.19263 m .51914 1.18827 L .52807 1.19184 L .317 .643 0 r
F
.52733 1.19671 m .51843 1.19263 L .52807 1.19184 L .327 .626 0 r
F
.52524 .77689 m .51851 .77329 L .52288 .76709 L .485 .416 0 r
F
.51851 .77329 m .52524 .77689 L .51934 .78683 L .407 .439 0 r
F
.51851 .85644 m .52074 .86427 L .51882 .87059 L .797 .547 0 r
F
.51851 .85644 m .52522 .8623 L .52074 .86427 L .561 .41 0 r
F
.51851 .85644 m .52452 .85808 L .52522 .8623 L .714 .363 0 r
F
.51882 .87059 m .52074 .86427 L .52357 .87763 L .646 .387 0 r
F
.52551 .83496 m .5189 .81848 L .52726 .81698 L .669 .587 0 r
F
.51934 1.07234 m .51945 1.07243 L .51907 1.07246 L .744 .69 0 r
F
.51923 1.07225 m .51934 1.07234 L .51907 1.07246 L F
.51951 1.07248 m .52169 1.09142 L .51907 1.07246 L F
.51945 1.07243 m .51951 1.07248 L .51907 1.07246 L F
.51914 1.18827 m .5293 1.1803 L .53991 1.18282 L .648 .684 0 r
F
.52807 1.19184 m .51914 1.18827 L .53991 1.18282 L .645 .645 0 r
F
.52951 .78637 m .51934 .78683 L .52524 .77689 L .473 .524 0 r
F
.52951 .78637 m .52491 .79305 L .51934 .78683 L .479 .414 0 r
F
.52262 .34214 m .52901 .34616 L .51941 .34713 L .451 0 0 r
F
.51941 .34713 m .52831 .34723 L .52296 .36415 L .756 .612 0 r
F
.52023 .34705 m .51941 .34713 L .52188 .34716 L .662 .842 0 r
F
.51942 1.16044 m .54288 1.13935 L .54081 1.14617 L .805 .872 0 r
F
.52636 1.14787 m .54288 1.13935 L .51942 1.16044 L .816 .944 0 r
F
.51942 1.16044 m .54081 1.14617 L .53105 1.15912 L .786 .779 0 r
F
.52428 1.0703 m .51951 1.07248 L .52001 1.05716 L .283 .224 0 r
F
.55104 1.07407 m .51951 1.07248 L .52428 1.0703 L .45 .71 0 r
F
.51951 1.07248 m .55104 1.07407 L .5404 1.0743 L .049 .61 0 r
F
.52667 1.07311 m .52169 1.09142 L .51951 1.07248 L .744 .69 0 r
F
.51974 .37244 m .52296 .36415 L .53488 .3644 L .759 .798 0 r
F
.53488 .3644 m .53212 .37321 L .51974 .37244 L .756 .785 0 r
F
.52513 1.03868 m .52948 1.04232 L .52001 1.05716 L .372 .252 0 r
F
.52948 1.04232 m .52568 1.05521 L .52001 1.05716 L .206 .056 0 r
F
.52568 1.05521 m .52428 1.0703 L .52001 1.05716 L .267 .215 0 r
F
.52067 .347 m .52023 .34705 L .52188 .34716 L .662 .842 0 r
F
.52277 .34679 m .52067 .347 L .52188 .34716 L F
.52074 .86427 m .52522 .8623 L .52648 .87179 L .554 .396 0 r
F
.52357 .87763 m .52074 .86427 L .52648 .87179 L .697 .42 0 r
F
.52357 .87763 m .53647 .88598 L .52075 .88854 L .68 .354 0 r
F
.52256 .89419 m .52075 .88854 L .53647 .88598 L .739 .511 0 r
F
.52827 .95557 m .53497 .95921 L .5214 .96769 L .404 .044 0 r
F
.52616 .97402 m .5214 .96769 L .53497 .95921 L .59 .318 0 r
F
.52147 .7646 m .52771 .76195 L .53117 .76433 L .572 .322 0 r
F
.53117 .76433 m .52288 .76709 L .52147 .7646 L .496 .103 0 r
F
.53291 .34233 m .52988 .34228 L .52168 .32991 L .783 .632 0 r
F
.53301 .34234 m .53291 .34233 L .52168 .32991 L F
.53287 .33106 m .53313 .34234 L .52168 .32991 L F
.53313 .34234 m .53301 .34234 L .52168 .32991 L F
.52988 .34228 m .52639 .34221 L .52168 .32991 L F
.53348 1.09711 m .52169 1.09142 L .52667 1.07311 L .801 .75 0 r
F
.53127 1.10669 m .52169 1.09142 L .53348 1.09711 L .631 .777 0 r
F
.52308 .34676 m .52188 .34716 L .52262 .34717 L .743 .99 0 r
F
.52277 .34679 m .52188 .34716 L .52308 .34676 L F
.52241 .73312 m .52985 .72376 L .53615 .73745 L .721 .523 0 r
F
.52684 .74432 m .52241 .73312 L .53615 .73745 L .682 .6 0 r
F
.52247 1.03301 m .52375 1.02031 L .5371 1.02398 L .659 .523 0 r
F
.5371 1.02398 m .53514 1.03704 L .52247 1.03301 L .677 .559 0 r
F
.52513 1.03868 m .52247 1.03301 L .53514 1.03704 L .714 .388 0 r
F
.52619 .89962 m .52256 .89419 L .53647 .88598 L .778 .591 0 r
F
.52512 .33889 m .52639 .34221 L .52262 .34214 L .783 .632 0 r
F
.52901 .34616 m .52262 .34214 L .53346 .34234 L .467 0 0 r
F
.52262 .34717 m .52345 .34718 L .52358 .34671 L .751 .841 0 r
F
.52262 .34717 m .52358 .34671 L .5232 .34675 L F
.52262 .34717 m .5232 .34675 L .52308 .34676 L F
.52274 .66705 m .52926 .65546 L .53154 .66746 L .782 .518 0 r
F
.53067 .67773 m .52274 .66705 L .53154 .66746 L .794 .604 0 r
F
.5326 .77277 m .52288 .76709 L .53117 .76433 L .676 .437 0 r
F
.52288 .76709 m .5326 .77277 L .52524 .77689 L .6 .49 0 r
F
.52831 .34723 m .53488 .3644 L .52296 .36415 L .802 .662 0 r
F
.54245 1.1058 m .53479 1.12542 L .52331 1.12554 L .765 .759 0 r
F
.53127 1.10669 m .54245 1.1058 L .52331 1.12554 L .749 .728 0 r
F
.524 .34718 m .52385 .34668 L .52345 .34718 L .619 .707 0 r
F
.52385 .34668 m .52363 .3467 L .52345 .34718 L F
.52363 .3467 m .52358 .34671 L .52345 .34718 L F
.52357 .87763 m .52648 .87179 L .53647 .88598 L .661 .379 0 r
F
.52375 1.02031 m .52637 1.00471 L .54 1.00446 L .666 .565 0 r
F
.54 1.00446 m .5371 1.02398 L .52375 1.02031 L .651 .54 0 r
F
.52391 .34668 m .52385 .34668 L .524 .34718 L .619 .707 0 r
F
.53685 .75954 m .52388 .76053 L .52684 .74432 L .681 .543 0 r
F
.53685 .75954 m .52771 .76195 L .52388 .76053 L .444 0 0 r
F
.52415 .34665 m .52391 .34668 L .524 .34718 L .619 .707 0 r
F
.5247 .34719 m .52415 .34665 L .524 .34718 L F
.52425 .34664 m .52415 .34665 L .5247 .34719 L F
.52485 .34658 m .52425 .34664 L .5247 .34719 L F
.53596 1.05558 m .54256 1.06728 L .52428 1.0703 L .581 .512 0 r
F
.52428 1.0703 m .52568 1.05521 L .53596 1.05558 L .543 .458 0 r
F
.52428 1.0703 m .54256 1.06728 L .55104 1.07407 L .588 .637 0 r
F
.53206 .85892 m .52452 .85808 L .52572 .84788 L .592 .479 0 r
F
.52522 .8623 m .52452 .85808 L .53206 .85892 L .585 .273 0 r
F
.52679 .34722 m .52485 .34658 L .5247 .34719 L .619 .707 0 r
F
.52567 .38788 m .52914 .42485 L .52474 .41773 L .982 .77 0 r
F
.52474 .41773 m .52914 .42485 L .52839 .452 L .989 .74 0 r
F
.52483 1.17458 m .53102 1.16972 L .53498 1.1741 L .477 .492 0 r
F
.5293 1.1803 m .52483 1.17458 L .53498 1.1741 L .474 .518 0 r
F
.52485 .34658 m .52485 .34658 L .52672 .34719 L .619 .707 0 r
F
.52486 .34658 m .52485 .34658 L .52672 .34719 L F
.52676 .34719 m .52486 .34658 L .52672 .34719 L F
.52676 .34719 m .52806 .34625 L .52486 .34658 L .46 .6 0 r
F
.52491 .79305 m .52951 .78637 L .53802 .7899 L .548 .505 0 r
F
.53564 .80294 m .52491 .79305 L .53802 .7899 L .552 .523 0 r
F
.52726 .81698 m .52491 .79305 L .53564 .80294 L .595 .528 0 r
F
.52511 1.19926 m .52733 1.19671 L .53489 1.19394 L .522 .918 0 r
F
.53308 1.19562 m .52511 1.19926 L .53489 1.19394 L .447 .871 0 r
F
.53514 1.03704 m .53604 1.03948 L .52513 1.03868 L .738 .446 0 r
F
.52948 1.04232 m .52513 1.03868 L .53604 1.03948 L .632 .171 0 r
F
.53206 .85892 m .52998 .86403 L .52522 .8623 L .68 .444 0 r
F
.52648 .87179 m .52522 .8623 L .52998 .86403 L .668 .476 0 r
F
.53737 .78312 m .52524 .77689 L .5326 .77277 L .61 .512 0 r
F
.52524 .77689 m .53737 .78312 L .52951 .78637 L .512 .546 0 r
F
.52551 .83496 m .52726 .81698 L .53525 .81786 L .562 .492 0 r
F
.53525 .81786 m .53398 .83531 L .52551 .83496 L .579 .512 0 r
F
.52572 .84788 m .52551 .83496 L .53398 .83531 L .582 .475 0 r
F
.52567 .38788 m .53212 .37321 L .54106 .37096 L .873 .906 0 r
F
.54106 .37096 m .53856 .38628 L .52567 .38788 L .864 .876 0 r
F
.52567 .38788 m .53856 .38628 L .52914 .42485 L .886 .706 0 r
F
.53596 1.05558 m .52568 1.05521 L .52948 1.04232 L .544 .384 0 r
F
.53398 .83531 m .53398 .84928 L .52572 .84788 L .593 .489 0 r
F
.52572 .84788 m .53398 .84928 L .53206 .85892 L .596 .481 0 r
F
.52601 .94813 m .52755 .94197 L .53686 .94776 L .728 .414 0 r
F
.52601 .94813 m .53686 .94776 L .53896 .95502 L .768 .565 0 r
F
.53896 .95502 m .52827 .95557 L .52601 .94813 L .785 .533 0 r
F
.53497 .95921 m .54015 .96849 L .52616 .97402 L .61 .342 0 r
F
.52715 .98112 m .52616 .97402 L .54015 .96849 L .645 .414 0 r
F
.53647 .88598 m .54432 .88471 L .52619 .89962 L .75 .55 0 r
F
.52619 .89962 m .54432 .88471 L .54624 .89175 L .788 .62 0 r
F
.54624 .89175 m .52839 .90554 L .52619 .89962 L .793 .635 0 r
F
.53013 .68474 m .53671 .69556 L .52626 .69567 L .755 .545 0 r
F
.52626 .69567 m .53671 .69556 L .54021 .70017 L .659 .252 0 r
F
.54021 .70017 m .52751 .70304 L .52626 .69567 L .67 .338 0 r
F
.52636 1.14787 m .54064 1.13472 L .54288 1.13935 L .838 .879 0 r
F
.52636 1.14787 m .53479 1.12542 L .54226 1.1276 L .774 .821 0 r
F
.54226 1.1276 m .54064 1.13472 L .52636 1.14787 L .837 .978 0 r
F
.52637 1.00471 m .52704 .99178 L .5414 .9921 L .662 .565 0 r
F
.5414 .9921 m .54 1.00446 L .52637 1.00471 L .666 .571 0 r
F
.52648 .87179 m .52998 .86403 L .54329 .86658 L .639 .444 0 r
F
.54329 .86658 m .53647 .88598 L .52648 .87179 L .608 .368 0 r
F
.53348 1.09711 m .52667 1.07311 L .52682 1.07312 L .801 .75 0 r
F
.52676 .34719 m .52672 .34719 L .52683 .34722 L .619 .707 0 r
F
.52683 .34722 m .52672 .34719 L .52679 .34722 L F
.52901 .34616 m .52806 .34625 L .52676 .34719 L .46 .6 0 r
F
.53594 .34713 m .52901 .34616 L .52676 .34719 L F
.5361 .34744 m .53594 .34713 L .52676 .34719 L F
.52683 .34722 m .52831 .34723 L .52676 .34719 L F
.52683 .34722 m .52683 .34722 L .52679 .34722 L .619 .707 0 r
F
.53348 1.09711 m .52682 1.07312 L .53205 1.07358 L .801 .75 0 r
F
.53615 .73745 m .54565 .74486 L .52684 .74432 L .7 .636 0 r
F
.52684 .74432 m .54565 .74486 L .53685 .75954 L .707 .55 0 r
F
.52704 .99178 m .52715 .98112 L .54278 .97803 L .672 .512 0 r
F
.54278 .97803 m .5414 .9921 L .52704 .99178 L .663 .497 0 r
F
.54015 .96849 m .54278 .97803 L .52715 .98112 L .638 .404 0 r
F
.53564 .80294 m .53525 .81786 L .52726 .81698 L .563 .487 0 r
F
.52733 1.19671 m .52807 1.19184 L .53261 1.18847 L .829 .987 0 r
F
.53489 1.19394 m .52733 1.19671 L .53261 1.18847 L .858 .935 0 r
F
.52942 .71272 m .52751 .70304 L .54021 .70017 L .773 .618 0 r
F
.52839 .90554 m .56398 .94167 L .52755 .94197 L .783 .635 0 r
F
.52755 .94197 m .56398 .94167 L .54775 .94781 L .744 .432 0 r
F
.53686 .94776 m .52755 .94197 L .54775 .94781 L .738 .386 0 r
F
.53117 .76433 m .52771 .76195 L .53685 .75954 L .64 .26 0 r
F
.52807 1.19184 m .53991 1.18282 L .54511 1.17918 L .479 .854 0 r
F
.53497 .95921 m .52827 .95557 L .53896 .95502 L .461 0 0 r
F
.52831 .34723 m .53854 .3475 L .53488 .3644 L .813 .667 0 r
F
.52914 .42485 m .53391 .45967 L .52839 .452 L .992 .745 0 r
F
.52839 .452 m .53391 .45967 L .53027 .48589 L .992 .771 0 r
F
.56398 .94167 m .52839 .90554 L .54624 .89175 L .793 .634 0 r
F
.53967 .71687 m .52876 .71802 L .52942 .71272 L .745 .666 0 r
F
.52985 .72376 m .52876 .71802 L .53967 .71687 L .659 .32 0 r
F
.53346 .34234 m .53854 .3475 L .52901 .34616 L .609 .071 0 r
F
.5479 .15452 m .53368 .1685 L .52909 .1139 L .802 .575 0 r
F
.52909 .1139 m .54442 .12303 L .5479 .15452 L .836 .586 0 r
F
.53856 .38628 m .53989 .42154 L .52914 .42485 L .906 .729 0 r
F
.52914 .42485 m .53989 .42154 L .53391 .45967 L .897 .691 0 r
F
.53966 .65644 m .52926 .65546 L .53776 .62538 L .765 .595 0 r
F
.52926 .65546 m .53966 .65644 L .54162 .66458 L .761 .519 0 r
F
.54162 .66458 m .53154 .66746 L .52926 .65546 L .771 .51 0 r
F
.5293 1.1803 m .53498 1.1741 L .54583 1.1767 L .588 .702 0 r
F
.53991 1.18282 m .5293 1.1803 L .54583 1.1767 L .586 .715 0 r
F
.54021 .70017 m .54109 .71152 L .52942 .71272 L .751 .584 0 r
F
.52942 .71272 m .54109 .71152 L .53967 .71687 L .758 .654 0 r
F
.53604 1.03948 m .53833 1.04051 L .52948 1.04232 L .465 0 0 r
F
.5359 1.04406 m .52948 1.04232 L .53833 1.04051 L .522 .033 0 r
F
.52948 1.04232 m .5359 1.04406 L .53596 1.05558 L .539 .382 0 r
F
.52951 .78637 m .53737 .78312 L .55379 .78522 L .533 .623 0 r
F
.55379 .78522 m .53802 .7899 L .52951 .78637 L .552 .503 0 r
F
.53967 .71687 m .54002 .72425 L .52985 .72376 L .695 .375 0 r
F
.53615 .73745 m .52985 .72376 L .54002 .72425 L .725 .525 0 r
F
.52998 .86403 m .53206 .85892 L .5417 .86209 L .611 .37 0 r
F
.5417 .86209 m .54329 .86658 L .52998 .86403 L .638 .455 0 r
F
.54358 .6845 m .53013 .68474 L .53067 .67773 L .801 .63 0 r
F
.54358 .6845 m .53671 .69556 L .53013 .68474 L .79 .553 0 r
F
.53391 .45967 m .53982 .49346 L .53027 .48589 L .942 .703 0 r
F
.53027 .48589 m .53982 .49346 L .53237 .52093 L .933 .725 0 r
F
.53067 .67773 m .53154 .66746 L .54162 .66458 L .809 .617 0 r
F
.54162 .66458 m .54271 .67428 L .53067 .67773 L .802 .606 0 r
F
.53067 .67773 m .54271 .67428 L .54358 .6845 L .806 .622 0 r
F
.53102 1.16972 m .53173 1.16459 L .53757 1.16904 L .281 .49 0 r
F
.53498 1.1741 m .53102 1.16972 L .53757 1.16904 L .284 .424 0 r
F
.53105 1.15912 m .54081 1.14617 L .5462 1.1512 L .616 .523 0 r
F
.53626 1.16282 m .53105 1.15912 L .5462 1.1512 L .645 .593 0 r
F
.53173 1.16459 m .53105 1.15912 L .53626 1.16282 L .156 .483 0 r
F
.54316 .76748 m .53117 .76433 L .53685 .75954 L .817 .523 0 r
F
.53117 .76433 m .53364 .76781 L .5326 .77277 L .47 .299 0 r
F
.54124 .76811 m .53364 .76781 L .53117 .76433 L .78 .345 0 r
F
.54124 .76811 m .53117 .76433 L .54316 .76748 L .817 .523 0 r
F
.54245 1.1058 m .53127 1.10669 L .53348 1.09711 L .69 .834 0 r
F
.53757 1.16904 m .53173 1.16459 L .53626 1.16282 L .141 .44 0 r
F
.53205 1.07358 m .54039 1.0743 L .53206 1.07371 L .742 .707 0 r
F
.5417 .86209 m .53206 .85892 L .53398 .84928 L .572 .458 0 r
F
.5404 1.0743 m .5404 1.0743 L .53206 1.07371 L .742 .707 0 r
F
.54039 1.0743 m .5404 1.0743 L .53206 1.07371 L F
.53206 1.07371 m .5404 1.0743 L .53211 1.07448 L F
.5404 1.0743 m .5404 1.0743 L .53211 1.07448 L F
.53211 1.07448 m .5404 1.0743 L .53348 1.09711 L F
.53212 .37321 m .53488 .3644 L .54323 .36494 L .87 .908 0 r
F
.54183 .36598 m .54106 .37096 L .53212 .37321 L .869 .911 0 r
F
.53982 .49346 m .54548 .52628 L .53237 .52093 L .887 .667 0 r
F
.53237 .52093 m .54548 .52628 L .53431 .55605 L .88 .693 0 r
F
.53772 .28748 m .53242 .29264 L .53359 .22454 L .57 .43 0 r
F
.53242 .29264 m .53772 .28748 L .54149 .33189 L .603 .481 0 r
F
.54149 .33189 m .53287 .33106 L .53242 .29264 L .626 .496 0 r
F
.54019 .7757 m .5326 .77277 L .53364 .76781 L .624 .443 0 r
F
.5326 .77277 m .54019 .7757 L .53737 .78312 L .58 .497 0 r
F
.53489 1.19394 m .53261 1.18847 L .53275 1.18837 L .858 .935 0 r
F
.53489 1.19394 m .53275 1.18837 L .53438 1.18715 L F
.53592 .34484 m .53287 .33106 L .53854 .3475 L .625 .506 0 r
F
.53977 .34182 m .53287 .33106 L .54149 .33189 L F
.53854 .3475 m .53287 .33106 L .53854 .3475 L F
.53854 .3475 m .53287 .33106 L .53977 .34182 L F
.53393 .34282 m .53287 .33106 L .53592 .34484 L F
.53388 .34277 m .53287 .33106 L .53393 .34282 L F
.53388 .34277 m .53287 .33106 L .53388 .34277 L F
.53387 .34276 m .53287 .33106 L .53388 .34277 L F
.53387 .34276 m .53287 .33106 L .53387 .34276 L F
.53329 .34217 m .53287 .33106 L .53387 .34276 L F
.53318 .34234 m .53287 .33106 L .53329 .34234 L F
.53287 .33106 m .53318 .34234 L .53313 .34234 L .783 .632 0 r
F
.53329 .34217 m .53346 .34234 L .53332 .34234 L .625 .506 0 r
F
.53329 .34234 m .53329 .34217 L .53332 .34234 L F
.53854 .3475 m .53346 .34234 L .54363 .35266 L .504 .006 0 r
F
.5404 1.0743 m .54366 1.07454 L .53348 1.09711 L .742 .707 0 r
F
.54722 1.0771 m .53348 1.09711 L .54366 1.07454 L .924 .973 0 r
F
.54736 1.09056 m .53348 1.09711 L .54722 1.0771 L .842 .737 0 r
F
.54863 1.1002 m .53348 1.09711 L .54736 1.09056 L .855 .815 0 r
F
.53348 1.09711 m .54863 1.1002 L .54245 1.1058 L .8 .859 0 r
F
.53368 .1685 m .54498 .21993 L .53359 .22454 L .765 .58 0 r
F
.53721 .22731 m .53359 .22454 L .54498 .21993 L .468 .046 0 r
F
.53359 .22454 m .53721 .22731 L .53772 .28748 L .167 .124 0 r
F
.54007 .7685 m .54046 .77133 L .53364 .76781 L .621 .384 0 r
F
.53364 .76781 m .54046 .77133 L .54019 .7757 L .568 .433 0 r
F
.54007 .7685 m .53364 .76781 L .54124 .76811 L .78 .345 0 r
F
.53368 .1685 m .5479 .15452 L .54715 .18719 L .829 .622 0 r
F
.53368 .1685 m .54715 .18719 L .54498 .21993 L .833 .622 0 r
F
.53989 .42154 m .54325 .45677 L .53391 .45967 L .926 .722 0 r
F
.53391 .45967 m .54325 .45677 L .53982 .49346 L .917 .689 0 r
F
.53398 .84928 m .53398 .83531 L .54215 .83526 L .519 .429 0 r
F
.54215 .83526 m .5434 .84916 L .53398 .84928 L .575 .497 0 r
F
.53398 .84928 m .5434 .84916 L .5417 .86209 L .575 .459 0 r
F
.53398 .83531 m .53525 .81786 L .54464 .81887 L .536 .475 0 r
F
.54464 .81887 m .54215 .83526 L .53398 .83531 L .519 .452 0 r
F
.54548 .52628 m .55094 .55971 L .53431 .55605 L .837 .639 0 r
F
.53431 .55605 m .55094 .55971 L .53652 .59015 L .835 .66 0 r
F
.53489 1.19394 m .53438 1.18715 L .53511 1.18662 L .956 .911 0 r
F
.53479 1.12542 m .54245 1.1058 L .55349 1.11529 L .768 .763 0 r
F
.55349 1.11529 m .54875 1.11988 L .53479 1.12542 L .765 .752 0 r
F
.53479 1.12542 m .54875 1.11988 L .5457 1.12497 L .758 .709 0 r
F
.54226 1.1276 m .53479 1.12542 L .5457 1.12497 L .672 .84 0 r
F
.53488 .3644 m .53854 .3475 L .54586 .35596 L .914 .775 0 r
F
.54586 .35596 m .54323 .36494 L .53488 .3644 L .926 .818 0 r
F
.53774 1.18512 m .53489 1.19394 L .53622 1.18579 L .905 .861 0 r
F
.53489 1.19394 m .53511 1.18662 L .53622 1.18579 L .956 .911 0 r
F
.53896 .95502 m .54614 .95712 L .53497 .95921 L .312 0 0 r
F
.54015 .96849 m .53497 .95921 L .54614 .95712 L .696 .37 0 r
F
.53498 1.1741 m .53757 1.16904 L .54679 1.17167 L .506 .7 0 r
F
.54583 1.1767 m .53498 1.1741 L .54679 1.17167 L .5 .72 0 r
F
.53514 1.03704 m .5371 1.02398 L .54654 1.02417 L .771 .648 0 r
F
.54654 1.02417 m .54516 1.03476 L .53514 1.03704 L .752 .616 0 r
F
.54516 1.03476 m .5448 1.03748 L .53514 1.03704 L F
.53604 1.03948 m .53514 1.03704 L .54457 1.03927 L .764 .459 0 r
F
.5448 1.03748 m .54469 1.03836 L .53514 1.03704 L .752 .616 0 r
F
.54469 1.03836 m .54457 1.03927 L .53514 1.03704 L F
.53525 .81786 m .53564 .80294 L .54624 .80716 L .547 .474 0 r
F
.54624 .80716 m .54464 .81887 L .53525 .81786 L .54 .462 0 r
F
.53549 1.19209 m .53663 1.19224 L .5354 1.19238 L .482 .884 0 r
F
.53578 1.1912 m .53663 1.19224 L .53549 1.19209 L .486 .885 0 r
F
.53564 .80294 m .53802 .7899 L .54731 .79661 L .544 .516 0 r
F
.54731 .79661 m .54624 .80716 L .53564 .80294 L .531 .486 0 r
F
.53853 1.18631 m .53663 1.19224 L .53578 1.1912 L .398 .831 0 r
F
.53853 1.18631 m .53578 1.1912 L .53728 1.18654 L F
.5359 1.04406 m .53833 1.04051 L .55523 1.04813 L .559 .072 0 r
F
.55487 1.04823 m .55106 1.04931 L .5359 1.04406 L .48 0 0 r
F
.53596 1.05558 m .5359 1.04406 L .55106 1.04931 L .491 .343 0 r
F
.5359 1.04406 m .55523 1.04813 L .55487 1.04823 L .559 .072 0 r
F
.5382 .34745 m .53594 .34713 L .5361 .34744 L .46 .6 0 r
F
.55106 1.04931 m .54602 1.0595 L .53596 1.05558 L .446 .246 0 r
F
.54256 1.06728 m .53596 1.05558 L .54602 1.0595 L .301 .376 0 r
F
.54553 1.0395 m .53833 1.04051 L .53604 1.03948 L .454 0 0 r
F
.53604 1.03948 m .54455 1.03945 L .54454 1.0395 L .764 .459 0 r
F
.53604 1.03948 m .54457 1.03927 L .54457 1.03928 L F
.53604 1.03948 m .54457 1.03928 L .54455 1.03945 L F
.53823 .34749 m .5382 .34745 L .5361 .34744 L .46 .6 0 r
F
.54002 .72425 m .54998 .73789 L .53615 .73745 L .734 .534 0 r
F
.54998 .73789 m .54565 .74486 L .53615 .73745 L .734 .631 0 r
F
.53774 1.18512 m .53622 1.18579 L .53663 1.18549 L .905 .861 0 r
F
.53626 1.16282 m .5462 1.1512 L .56017 1.15777 L .643 .59 0 r
F
.5481 1.16609 m .53626 1.16282 L .56017 1.15777 L .647 .635 0 r
F
.53757 1.16904 m .53626 1.16282 L .5481 1.16609 L .441 .695 0 r
F
.54329 .86658 m .54686 .86724 L .53647 .88598 L .634 .395 0 r
F
.54432 .88471 m .53647 .88598 L .54686 .86724 L .744 .529 0 r
F
.55094 .55971 m .55478 .59321 L .53652 .59015 L .815 .635 0 r
F
.53652 .59015 m .55478 .59321 L .53776 .62538 L .814 .646 0 r
F
.53663 1.19224 m .56025 1.18237 L .56089 1.1878 L .36 .832 0 r
F
.54096 1.19043 m .53663 1.19224 L .53853 1.18631 L .398 .831 0 r
F
.53774 1.18512 m .53663 1.18549 L .5375 1.18484 L .905 .861 0 r
F
.5456 .69478 m .53671 .69556 L .54358 .6845 L .827 .604 0 r
F
.53671 .69556 m .5456 .69478 L .54929 .69945 L .65 .191 0 r
F
.54929 .69945 m .54021 .70017 L .53671 .69556 L .669 .252 0 r
F
.53685 .75954 m .54565 .74486 L .54973 .75598 L .607 .425 0 r
F
.54573 .76663 m .53685 .75954 L .54973 .75598 L .635 .513 0 r
F
.54316 .76748 m .53685 .75954 L .54573 .76663 L F
.55128 .95728 m .53896 .95502 L .53686 .94776 L .786 .575 0 r
F
.53686 .94776 m .54775 .94781 L .55128 .95728 L .789 .572 0 r
F
.5371 1.02398 m .54 1.00446 L .55016 1.00415 L .762 .645 0 r
F
.55016 1.00415 m .54654 1.02417 L .5371 1.02398 L .77 .657 0 r
F
.54498 .21993 m .54908 .22421 L .53721 .22731 L .596 .199 0 r
F
.53721 .22731 m .54908 .22421 L .53772 .28748 L .782 .619 0 r
F
.53774 1.18512 m .56025 1.18237 L .53728 1.18654 L .383 .826 0 r
F
.55379 .78522 m .53737 .78312 L .54019 .7757 L .594 .513 0 r
F
.53774 1.18512 m .5375 1.18484 L .53768 1.1847 L .905 .861 0 r
F
.54679 1.17167 m .53757 1.16904 L .5481 1.16609 L .44 .7 0 r
F
.53768 1.1847 m .54556 1.18416 L .53774 1.18512 L .581 .844 0 r
F
.53793 1.18452 m .54556 1.18416 L .53768 1.1847 L F
.54908 .22421 m .56327 .22587 L .53772 .28748 L .816 .653 0 r
F
.56327 .22587 m .5725 .22752 L .53772 .28748 L .861 .71 0 r
F
.5725 .22752 m .58408 .23203 L .53772 .28748 L .821 .651 0 r
F
.58408 .23203 m .59531 .2383 L .53772 .28748 L .795 .611 0 r
F
.59531 .2383 m .60637 .24773 L .53772 .28748 L .823 .664 0 r
F
.60637 .24773 m .61533 .26042 L .53772 .28748 L .818 .652 0 r
F
.61533 .26042 m .62225 .272 L .53772 .28748 L .817 .647 0 r
F
.62225 .272 m .63206 .29309 L .53772 .28748 L .819 .659 0 r
F
.53772 .28748 m .57713 .33074 L .54149 .33189 L .793 .623 0 r
F
.60881 .33238 m .57713 .33074 L .53772 .28748 L .83 .616 0 r
F
.53772 .28748 m .63206 .29309 L .60881 .33238 L .819 .629 0 r
F
.55478 .59321 m .55559 .6278 L .53776 .62538 L .824 .66 0 r
F
.55559 .6278 m .54895 .65522 L .53776 .62538 L .826 .636 0 r
F
.53776 .62538 m .54895 .65522 L .53966 .65644 L .81 .629 0 r
F
.54014 1.18287 m .54556 1.18416 L .53793 1.18452 L .581 .844 0 r
F
.55379 .78522 m .55173 .79151 L .53802 .7899 L .554 .511 0 r
F
.54731 .79661 m .53802 .7899 L .55173 .79151 L .552 .516 0 r
F
.53842 .34748 m .5382 .34745 L .53823 .34749 L .46 .6 0 r
F
.53842 .34748 m .53823 .34749 L .53843 .34748 L F
.53852 .3475 m .53823 .34749 L .53852 .3475 L F
.53843 .34748 m .53823 .34749 L .53852 .3475 L F
.5555 1.04805 m .53833 1.04051 L .54553 1.0395 L .542 .008 0 r
F
.55523 1.04813 m .53833 1.04051 L .5555 1.04805 L F
.53852 .3475 m .53852 .3475 L .53854 .3475 L .46 .6 0 r
F
.56025 1.18237 m .54096 1.19043 L .53853 1.18631 L .398 .831 0 r
F
.54363 .35266 m .54586 .35596 L .53854 .3475 L .633 .158 0 r
F
.53875 .34771 m .53854 .3475 L .53977 .34182 L .625 .506 0 r
F
.53856 .38628 m .54177 .36899 L .54184 .3688 L .789 .759 0 r
F
.53856 .38628 m .54106 .37096 L .54177 .36899 L F
.53856 .38628 m .54184 .3688 L .54214 .36798 L F
.53856 .38628 m .53925 .38575 L .53989 .42154 L .838 .662 0 r
F
.53897 .38414 m .53932 .3857 L .53856 .38628 L .837 .583 0 r
F
.54149 .33189 m .53921 .34818 L .53875 .34771 L .571 .455 0 r
F
.54614 .95712 m .53896 .95502 L .55128 .95728 L .238 0 0 r
F
.53897 .38414 m .53914 .3833 L .53915 .38325 L .572 .473 0 r
F
.53907 .38455 m .53897 .38414 L .53915 .38325 L F
.53907 .38455 m .53915 .38325 L .53923 .3853 L F
.53914 .3833 m .53915 .38323 L .53915 .38325 L F
.54015 .39777 m .53915 .38323 L .54394 .37441 L F
.53915 .38323 m .53929 .38252 L .54394 .37441 L F
.53923 .3853 m .53915 .38325 L .53923 .3853 L F
.53923 .3853 m .53915 .38325 L .53924 .38531 L F
.53924 .38531 m .53915 .38325 L .53932 .3857 L F
.54149 .33189 m .53961 .34859 L .53921 .34818 L .571 .455 0 r
F
.53925 .38575 m .53937 .38874 L .53989 .42154 L .838 .662 0 r
F
.53925 .38575 m .53932 .3857 L .53937 .38874 L F
.53929 .38252 m .54086 .37449 L .54394 .37441 L .572 .473 0 r
F
.53932 .3857 m .54172 .42078 L .53989 .42154 L .973 .753 0 r
F
.54149 .33189 m .53964 .34861 L .53961 .34859 L .571 .455 0 r
F
.54149 .33189 m .53964 .34862 L .53964 .34861 L F
.54149 .33189 m .54006 .34757 L .53964 .34862 L F
.54006 .34757 m .53993 .34891 L .53964 .34862 L F
.53966 .65644 m .54895 .65522 L .55299 .66768 L .786 .532 0 r
F
.55299 .66768 m .54162 .66458 L .53966 .65644 L .784 .533 0 r
F
.55039 .71242 m .53967 .71687 L .54109 .71152 L .788 .686 0 r
F
.54002 .72425 m .53967 .71687 L .55039 .71242 L .634 .327 0 r
F
.54325 .45677 m .54636 .49099 L .53982 .49346 L .976 .754 0 r
F
.53982 .49346 m .54636 .49099 L .54548 .52628 L .966 .713 0 r
F
.53989 .42154 m .53988 .42152 L .53989 .42154 L .259 .228 0 r
F
.53988 .42152 m .53988 .42152 L .53989 .42154 L F
.53989 .42154 m .53989 .42154 L .53989 .42154 L F
.53989 .42154 m .53989 .42154 L .53989 .42154 L F
.53989 .42154 m .53989 .42154 L .53989 .42154 L F
.53989 .42154 m .53989 .42154 L .53989 .42154 L F
.53989 .42154 m .54159 .42083 L .54325 .45677 L .973 .741 0 r
F
.53991 1.18282 m .54583 1.1767 L .5529 1.17312 L .685 .98 0 r
F
.54511 1.17918 m .53991 1.18282 L .5529 1.17312 L .474 .851 0 r
F
.54149 .33189 m .54057 .34955 L .53993 .34891 L .571 .455 0 r
F
.54 1.00446 m .5414 .9921 L .55234 .99059 L .778 .674 0 r
F
.55234 .99059 m .55016 1.00415 L .54 1.00446 L .762 .646 0 r
F
.55039 .71242 m .55326 .71915 L .54002 .72425 L .691 .403 0 r
F
.54998 .73789 m .54002 .72425 L .55326 .71915 L .753 .537 0 r
F
.54007 .7685 m .54661 .76634 L .54453 .77248 L .648 .427 0 r
F
.54453 .77248 m .54046 .77133 L .54007 .7685 L .673 .421 0 r
F
.54121 1.18208 m .54556 1.18416 L .54014 1.18287 L .581 .844 0 r
F
.54394 .37441 m .54156 .41849 L .54015 .39777 L .576 .478 0 r
F
.54614 .95712 m .55102 .96878 L .54015 .96849 L .717 .394 0 r
F
.54278 .97803 m .54015 .96849 L .55102 .96878 L .735 .46 0 r
F
.54453 .77248 m .54019 .7757 L .54046 .77133 L .651 .505 0 r
F
.55384 .77865 m .54019 .7757 L .54453 .77248 L .647 .497 0 r
F
.54019 .7757 m .55384 .77865 L .55379 .78522 L .647 .5 0 r
F
.54109 .71152 m .54021 .70017 L .54929 .69945 L .802 .623 0 r
F
.5404 1.0743 m .55104 1.07407 L .5606 1.07575 L .043 .607 0 r
F
.54149 .33189 m .54263 .35165 L .54057 .34955 L .571 .455 0 r
F
.54064 1.13472 m .54226 1.1276 L .54792 1.13113 L .359 .521 0 r
F
.54857 1.13757 m .54064 1.13472 L .54792 1.13113 L .374 .563 0 r
F
.54288 1.13935 m .54064 1.13472 L .54857 1.13757 L .293 .566 0 r
F
.54081 1.14617 m .54288 1.13935 L .54873 1.14453 L .435 .477 0 r
F
.5462 1.1512 m .54081 1.14617 L .54873 1.14453 L .437 .494 0 r
F
.54214 .36798 m .54394 .37441 L .54086 .37449 L .571 .455 0 r
F
.54323 .36494 m .54106 .37096 L .54183 .36598 L .869 .911 0 r
F
.54929 .69945 m .5511 .70868 L .54109 .71152 L .761 .564 0 r
F
.54109 .71152 m .5511 .70868 L .55039 .71242 L .79 .677 0 r
F
.5413 1.18202 m .56025 1.18237 L .54121 1.18208 L .581 .844 0 r
F
.54121 1.18208 m .56025 1.18237 L .54556 1.18416 L F
.54306 1.18071 m .56025 1.18237 L .5413 1.18202 L F
.5414 .9921 m .54278 .97803 L .55294 .97882 L .764 .587 0 r
F
.55294 .97882 m .55234 .99059 L .5414 .9921 L .768 .594 0 r
F
.54272 .35173 m .54263 .35165 L .54149 .33189 L .805 .635 0 r
F
.54275 .35177 m .54272 .35173 L .54149 .33189 L F
.57713 .33074 m .57993 .3722 L .54149 .33189 L .794 .637 0 r
F
.54525 .35505 m .54448 .35391 L .54149 .33189 L .805 .635 0 r
F
.54532 .35516 m .54525 .35505 L .54149 .33189 L F
.54969 .34048 m .54532 .35516 L .54149 .33189 L F
.54288 .3519 m .54288 .3519 L .54149 .33189 L F
.54376 .3486 m .54288 .3519 L .54149 .33189 L F
.54288 .3519 m .54275 .35177 L .54149 .33189 L F
.54394 .37441 m .54571 .41819 L .54156 .41849 L .576 .478 0 r
F
.54159 .41891 m .54156 .41849 L .54571 .41819 L .576 .475 0 r
F
.54159 .42083 m .54172 .42078 L .54183 .42323 L .973 .741 0 r
F
.54159 .42083 m .54183 .42323 L .54325 .45677 L F
.54162 .41936 m .54159 .41891 L .54571 .41819 L .576 .475 0 r
F
.54271 .67428 m .54162 .66458 L .55299 .66768 L .778 .589 0 r
F
.54163 .41944 m .54162 .41936 L .54571 .41819 L .576 .475 0 r
F
.54527 .41988 m .54163 .41944 L .54571 .41819 L F
.54172 .42078 m .54163 .41944 L .5442 .41975 L F
.54718 .86385 m .5417 .86209 L .5434 .84916 L .638 .517 0 r
F
.54329 .86658 m .5417 .86209 L .54718 .86385 L .662 .468 0 r
F
.54172 .42078 m .5442 .41975 L .54172 .4208 L .576 .475 0 r
F
.54172 .42078 m .54508 .45539 L .54325 .45677 L .976 .743 0 r
F
.54172 .4208 m .5442 .41975 L .54172 .4208 L .576 .475 0 r
F
.54172 .4208 m .5442 .41975 L .54527 .41988 L F
.54174 .42097 m .54172 .4208 L .54527 .41988 L F
.54195 .42311 m .54174 .42097 L .54527 .41988 L F
.54243 .42806 m .54195 .42311 L .54527 .41988 L F
.54277 .36622 m .54394 .37441 L .54214 .36798 L .571 .455 0 r
F
.54215 .83526 m .54464 .81887 L .55008 .81854 L .582 .511 0 r
F
.55008 .81854 m .54741 .83643 L .54215 .83526 L .635 .574 0 r
F
.5434 .84916 m .54215 .83526 L .54741 .83643 L .647 .552 0 r
F
.54226 1.1276 m .5457 1.12497 L .55429 1.12884 L .579 .615 0 r
F
.54792 1.13113 m .54226 1.1276 L .55429 1.12884 L .617 .509 0 r
F
.54261 .42989 m .54243 .42806 L .54527 .41988 L .576 .475 0 r
F
.54245 1.1058 m .54863 1.1002 L .55699 1.10767 L .777 .796 0 r
F
.54245 1.1058 m .55699 1.10767 L .55349 1.11529 L .803 .763 0 r
F
.54602 1.0595 m .55151 1.06815 L .54256 1.06728 L .376 .466 0 r
F
.55104 1.07407 m .54256 1.06728 L .55151 1.06815 L .31 .555 0 r
F
.54571 .41819 m .54298 .43369 L .54261 .42989 L .512 .41 0 r
F
.55299 .66768 m .55554 .67788 L .54271 .67428 L .781 .595 0 r
F
.54271 .67428 m .55554 .67788 L .55815 .68484 L .779 .602 0 r
F
.55815 .68484 m .54358 .6845 L .54271 .67428 L F
.54287 .36594 m .54394 .37441 L .54277 .36622 L .571 .455 0 r
F
.55102 .96878 m .55294 .97882 L .54278 .97803 L .755 .489 0 r
F
.54325 .36494 m .54394 .37441 L .54287 .36594 L .571 .455 0 r
F
.54325 .36494 m .54287 .36594 L .54322 .36496 L F
.54363 .35266 m .54288 .3519 L .54376 .3486 L .805 .635 0 r
F
.54873 1.14453 m .54288 1.13935 L .54857 1.13757 L .265 .433 0 r
F
.54571 .41819 m .54298 .43371 L .54298 .43369 L .512 .41 0 r
F
.54571 .41819 m .54318 .43583 L .54298 .43371 L F
.5697 1.17212 m .54306 1.18071 L .54317 1.18062 L .605 .844 0 r
F
.5697 1.17212 m .56025 1.18237 L .54306 1.18071 L F
.5697 1.17212 m .54317 1.18062 L .54734 1.17762 L F
.54317 1.18062 m .54511 1.17918 L .54734 1.17762 L F
.54571 .41819 m .5435 .43914 L .54318 .43583 L .512 .41 0 r
F
.54325 .36494 m .54322 .36496 L .54323 .36494 L .571 .455 0 r
F
.54325 .36488 m .54325 .36494 L .54323 .36494 L F
.54337 .36446 m .54394 .37441 L .54325 .36488 L F
.54325 .45677 m .54506 .45541 L .54636 .49099 L .976 .743 0 r
F
.54718 .86385 m .54686 .86724 L .54329 .86658 L .633 .418 0 r
F
.57993 .3722 m .54337 .36446 L .54373 .36325 L .805 .635 0 r
F
.57993 .3722 m .54394 .37441 L .54337 .36446 L F
.54741 .83643 m .54741 .85043 L .5434 .84916 L .567 .468 0 r
F
.5434 .84916 m .54741 .85043 L .54718 .86385 L .565 .471 0 r
F
.54571 .41819 m .544 .44422 L .5435 .43914 L .512 .41 0 r
F
.55815 .68484 m .55575 .69523 L .54358 .6845 L .773 .545 0 r
F
.54358 .6845 m .55575 .69523 L .5456 .69478 L .749 .554 0 r
F
.54448 .35391 m .54363 .35266 L .54376 .3486 L .805 .635 0 r
F
.54722 1.0771 m .54366 1.07454 L .54421 1.07457 L .924 .973 0 r
F
.57993 .3722 m .54373 .36325 L .54394 .3625 L .805 .635 0 r
F
.57993 .3722 m .58413 .41875 L .54394 .37441 L .806 .652 0 r
F
.58413 .41875 m .54571 .41819 L .54394 .37441 L .799 .652 0 r
F
.57993 .3722 m .54394 .3625 L .54397 .36243 L .805 .635 0 r
F
.57993 .3722 m .54397 .36243 L .54398 .36238 L F
.57993 .3722 m .54398 .36238 L .54428 .36135 L F
.54571 .41819 m .54434 .44771 L .544 .44422 L .512 .41 0 r
F
.54722 1.0771 m .54421 1.07457 L .54721 1.07479 L .924 .973 0 r
F
.57993 .3722 m .54428 .36135 L .55163 .34252 L .805 .635 0 r
F
.54428 .36135 m .54586 .35596 L .55163 .34252 L F
.54686 .86724 m .55959 .87172 L .54432 .88471 L .8 .582 0 r
F
.54432 .88471 m .55959 .87172 L .54624 .89175 L .833 .645 0 r
F
.54571 .41819 m .54448 .4492 L .54434 .44771 L .512 .41 0 r
F
.54442 .12303 m .5705 .13074 L .5479 .15452 L .813 .57 0 r
F
.54508 .45539 m .54448 .4492 L .54571 .41819 L .787 .643 0 r
F
.55384 .77865 m .54453 .77248 L .54661 .76634 L .691 .473 0 r
F
.54454 1.0395 m .54531 1.03945 L .54553 1.0395 L .764 .459 0 r
F
.54454 1.0395 m .54654 1.02417 L .54458 1.0395 L .808 .667 0 r
F
.54539 1.03947 m .54539 1.03947 L .54458 1.0395 L .829 .701 0 r
F
.54458 1.0395 m .54539 1.03947 L .54458 1.0395 L F
.54539 1.03947 m .54539 1.03947 L .54458 1.0395 L F
.54458 1.0395 m .54654 1.02417 L .54458 1.0395 L .808 .667 0 r
F
.54458 1.0395 m .54654 1.02417 L .5457 1.03464 L F
.54458 1.0395 m .54504 1.03749 L .54547 1.03751 L F
.54458 1.0395 m .54547 1.03751 L .54543 1.03794 L F
.54458 1.0395 m .54543 1.03794 L .54539 1.03846 L F
.54539 1.03846 m .54512 1.0394 L .54458 1.0395 L .829 .701 0 r
F
.54515 1.03946 m .54539 1.03947 L .54458 1.0395 L F
.54484 1.03945 m .54531 1.03945 L .54458 1.0395 L .764 .459 0 r
F
.54464 .81887 m .54624 .80716 L .55152 .80725 L .652 .565 0 r
F
.55152 .80725 m .55008 .81854 L .54464 .81887 L .58 .475 0 r
F
.54484 1.03945 m .54512 1.0394 L .54531 1.03945 L .764 .459 0 r
F
.54715 .18719 m .56487 .19005 L .54498 .21993 L .815 .605 0 r
F
.56013 .2228 m .54498 .21993 L .56487 .19005 L .841 .643 0 r
F
.54498 .21993 m .56013 .2228 L .56327 .22587 L .546 0 0 r
F
.56327 .22587 m .54908 .22421 L .54498 .21993 L .667 .176 0 r
F
.54504 1.03749 m .5457 1.03464 L .54547 1.03751 L .808 .667 0 r
F
.54506 .45541 m .54508 .45539 L .5451 .4558 L .976 .743 0 r
F
.54506 .45541 m .5451 .4558 L .54636 .49099 L F
.54631 .46599 m .54508 .45539 L .54571 .41819 L .787 .643 0 r
F
.54508 .45539 m .54914 .49058 L .54636 .49099 L .994 .752 0 r
F
.54734 1.17762 m .54511 1.17918 L .54572 1.17871 L .605 .844 0 r
F
.54539 1.03846 m .54531 1.03945 L .54512 1.0394 L .752 .616 0 r
F
.54531 1.03945 m .54539 1.03947 L .54515 1.03946 L .829 .701 0 r
F
.54566 1.03752 m .54553 1.0395 L .54531 1.03945 L .752 .616 0 r
F
.54566 1.03752 m .54531 1.03945 L .54539 1.03846 L F
.54533 .35517 m .54532 .35516 L .54969 .34048 L .805 .635 0 r
F
.54534 .35519 m .54533 .35517 L .54969 .34048 L F
.54537 .35523 m .54534 .35519 L .54969 .34048 L F
.54586 .35596 m .54537 .35523 L .54969 .34048 L F
.54566 1.03752 m .54539 1.03846 L .54547 1.03751 L .752 .616 0 r
F
.54654 1.02417 m .54566 1.03752 L .54547 1.03751 L F
.54636 .49099 m .55136 .52632 L .54548 .52628 L .99 .742 0 r
F
.54548 .52628 m .55136 .52632 L .55094 .55971 L .987 .727 0 r
F
.54553 1.03948 m .54591 1.03945 L .54553 1.0395 L .761 .421 0 r
F
.54591 1.03945 m .54557 1.03953 L .54553 1.0395 L F
.54553 1.03946 m .54591 1.03945 L .54553 1.03948 L .829 .701 0 r
F
.54553 1.03946 m .54591 1.03945 L .54553 1.03946 L F
.54605 1.03924 m .54591 1.03945 L .54553 1.03946 L F
.54605 1.03924 m .54553 1.03946 L .54554 1.03941 L F
.54605 1.03924 m .54554 1.03941 L .54554 1.03933 L F
.55528 1.02581 m .54554 1.03933 L .54562 1.03817 L F
.55528 1.02581 m .54605 1.03924 L .54554 1.03933 L F
.54557 1.03953 m .55433 1.03878 L .54577 1.0397 L .761 .421 0 r
F
.54591 1.03945 m .55433 1.03878 L .54557 1.03953 L F
.5456 .69478 m .55575 .69523 L .55879 .70037 L .679 .288 0 r
F
.55879 .70037 m .54929 .69945 L .5456 .69478 L .66 .192 0 r
F
.54566 1.03752 m .54804 1.02445 L .54562 1.03817 L .808 .667 0 r
F
.54562 1.03817 m .54804 1.02445 L .55528 1.02581 L F
.55525 .74085 m .54565 .74486 L .54998 .73789 L .831 .77 0 r
F
.55742 .75355 m .54565 .74486 L .55525 .74085 L .755 .529 0 r
F
.54973 .75598 m .54565 .74486 L .55742 .75355 L .776 .51 0 r
F
.54591 1.03371 m .54804 1.02445 L .54566 1.03752 L .808 .667 0 r
F
.5457 1.12497 m .54875 1.11988 L .55876 1.12384 L .656 .572 0 r
F
.55429 1.12884 m .5457 1.12497 L .55876 1.12384 L .656 .582 0 r
F
.58511 .46675 m .54631 .46599 L .54571 .41819 L .787 .643 0 r
F
.58413 .41875 m .58511 .46675 L .54571 .41819 L .799 .644 0 r
F
.54734 1.17762 m .54572 1.17871 L .54701 1.1777 L .605 .844 0 r
F
.54573 .76663 m .54973 .75598 L .54657 .76636 L .129 0 0 r
F
.54577 1.0397 m .55433 1.03878 L .55423 1.04073 L .761 .421 0 r
F
.55423 1.04073 m .5458 1.03973 L .54577 1.0397 L .511 0 0 r
F
.55423 1.04073 m .54596 1.03987 L .5458 1.03973 L F
.5529 1.17312 m .54583 1.1767 L .5558 1.16638 L .826 .98 0 r
F
.54583 1.1767 m .54679 1.17167 L .5558 1.16638 L .82 .989 0 r
F
.55163 .34252 m .54586 .35596 L .54969 .34048 L .805 .635 0 r
F
.55528 1.02581 m .55433 1.03878 L .54591 1.03945 L .829 .701 0 r
F
.54591 1.03371 m .54654 1.02417 L .54804 1.02445 L .808 .667 0 r
F
.55423 1.04073 m .55504 1.04184 L .54596 1.03987 L .511 0 0 r
F
.54603 1.03992 m .54596 1.03987 L .55504 1.04184 L .543 0 0 r
F
.56111 1.05709 m .54602 1.0595 L .55106 1.04931 L .576 .391 0 r
F
.54602 1.0595 m .56111 1.05709 L .56069 1.06396 L .609 .557 0 r
F
.56069 1.06396 m .55151 1.06815 L .54602 1.0595 L .599 .568 0 r
F
.55199 1.04504 m .54603 1.03992 L .55504 1.04184 L .543 0 0 r
F
.55102 .96878 m .54614 .95712 L .55728 .96047 L .793 .428 0 r
F
.55128 .95728 m .55728 .96047 L .54614 .95712 L .388 0 0 r
F
.5462 1.1512 m .54873 1.14453 L .56466 1.15098 L .571 .645 0 r
F
.56017 1.15777 m .5462 1.1512 L .56466 1.15098 L .583 .612 0 r
F
.54624 .89175 m .55959 .87172 L .56203 .88887 L .792 .584 0 r
F
.54624 .89175 m .56203 .88887 L .56282 .92215 L .803 .636 0 r
F
.54624 .89175 m .56282 .92215 L .56398 .94167 L .785 .63 0 r
F
.54624 .80716 m .54731 .79661 L .55308 .79884 L .634 .579 0 r
F
.55308 .79884 m .55152 .80725 L .54624 .80716 L .647 .598 0 r
F
.58511 .46675 m .54634 .46626 L .54631 .46599 L .787 .645 0 r
F
.58511 .46675 m .54649 .46759 L .54634 .46626 L F
.54636 .49099 m .54914 .49058 L .55136 .52632 L .991 .737 0 r
F
.58511 .46675 m .54662 .46869 L .54649 .46759 L .787 .645 0 r
F
.54654 1.02417 m .55016 1.00415 L .55919 1.00572 L .813 .699 0 r
F
.55919 1.00572 m .55528 1.02581 L .54654 1.02417 L .803 .684 0 r
F
.54657 .76636 m .54973 .75598 L .55141 .7585 L .129 0 0 r
F
.54657 .76636 m .54658 .76634 L .54661 .76634 L 0 0 0 r
F
.54658 .76634 m .55141 .7585 L .54722 .76563 L F
.54658 .76634 m .54722 .76563 L .54663 .76631 L F
.54658 .76634 m .54663 .76631 L .54661 .76634 L F
.56071 .77043 m .54661 .76634 L .55848 .76041 L .648 .456 0 r
F
.54661 .76634 m .55117 .761 L .55848 .76041 L .732 .655 0 r
F
.55384 .77865 m .54661 .76634 L .56071 .77043 L .648 .459 0 r
F
.58802 .50931 m .54914 .49058 L .54662 .46869 L .791 .644 0 r
F
.58511 .46675 m .58802 .50931 L .54662 .46869 L .787 .645 0 r
F
.5558 1.16638 m .54679 1.17167 L .55232 1.16324 L .876 .935 0 r
F
.54679 1.17167 m .5481 1.16609 L .55232 1.16324 L .828 .982 0 r
F
.54686 .86724 m .54718 .86385 L .55914 .86787 L .791 .56 0 r
F
.55914 .86787 m .55959 .87172 L .54686 .86724 L .795 .6 0 r
F
.55492 1.1721 m .54701 1.1777 L .54883 1.17629 L .652 .662 0 r
F
.55492 1.1721 m .5697 1.17212 L .54701 1.1777 L F
.5479 .15452 m .56786 .15964 L .54715 .18719 L .826 .619 0 r
F
.56786 .15964 m .56487 .19005 L .54715 .18719 L .815 .601 0 r
F
.54718 .86385 m .54741 .85043 L .55888 .85125 L .765 .637 0 r
F
.55888 .85125 m .55914 .86787 L .54718 .86385 L .764 .636 0 r
F
.54722 1.0771 m .54721 1.07479 L .55426 1.07529 L .63 .794 0 r
F
.54722 .76563 m .55141 .7585 L .54727 .76556 L 0 0 0 r
F
.55653 1.08388 m .54722 1.0771 L .55426 1.07529 L .624 .538 0 r
F
.54736 1.09056 m .54722 1.0771 L .55653 1.08388 L .598 .541 0 r
F
.54727 .76556 m .55141 .7585 L .54762 .76516 L 0 0 0 r
F
.55173 .79151 m .55308 .79884 L .54731 .79661 L .606 .597 0 r
F
.55627 1.09435 m .54736 1.09056 L .55653 1.08388 L .638 .622 0 r
F
.54863 1.1002 m .54736 1.09056 L .55627 1.09435 L .598 .636 0 r
F
.54741 .85043 m .54741 .83643 L .55961 .83621 L .758 .623 0 r
F
.55961 .83621 m .55888 .85125 L .54741 .85043 L .765 .635 0 r
F
.56167 .81706 m .55961 .83621 L .54741 .83643 L .757 .657 0 r
F
.54741 .83643 m .55008 .81854 L .56167 .81706 L .806 .735 0 r
F
.54762 .76516 m .55141 .7585 L .55128 .75986 L 0 0 0 r
F
.54762 .76516 m .55128 .75986 L .54844 .76419 L F
.54775 .94781 m .56398 .94167 L .56478 .94698 L .767 .478 0 r
F
.55128 .95728 m .54775 .94781 L .56478 .94698 L .791 .573 0 r
F
.5705 .13074 m .56786 .15964 L .5479 .15452 L .828 .594 0 r
F
.54792 1.13113 m .55429 1.12884 L .5658 1.13146 L .66 .668 0 r
F
.54857 1.13757 m .54792 1.13113 L .5611 1.13639 L .502 .67 0 r
F
.5611 1.13639 m .54792 1.13113 L .5658 1.13146 L .672 .626 0 r
F
.5481 1.16609 m .56017 1.15777 L .56644 1.1537 L .433 .831 0 r
F
.54844 .76419 m .55128 .75986 L .5486 .76401 L 0 0 0 r
F
.56392 1.14522 m .54857 1.13757 L .5611 1.13639 L .506 .659 0 r
F
.54873 1.14453 m .54857 1.13757 L .56392 1.14522 L .521 .66 0 r
F
.5486 .76401 m .55128 .75986 L .55033 .76199 L 0 0 0 r
F
.55872 1.10238 m .54863 1.1002 L .55627 1.09435 L .619 .682 0 r
F
.55699 1.10767 m .54863 1.1002 L .55872 1.10238 L .506 .721 0 r
F
.56466 1.15098 m .54873 1.14453 L .56392 1.14522 L .528 .652 0 r
F
.54875 1.11988 m .55349 1.11529 L .5628 1.11814 L .667 .574 0 r
F
.55876 1.12384 m .54875 1.11988 L .5628 1.11814 L .666 .563 0 r
F
.55492 1.1721 m .54883 1.17629 L .5529 1.17312 L .652 .662 0 r
F
.55559 .6278 m .55853 .65341 L .54895 .65522 L .837 .648 0 r
F
.54895 .65522 m .55853 .65341 L .56125 .6665 L .793 .505 0 r
F
.55299 .66768 m .54895 .65522 L .56125 .6665 L .765 .521 0 r
F
.54914 .49058 m .55396 .5253 L .55136 .52632 L .983 .733 0 r
F
.58802 .50931 m .55202 .51135 L .54914 .49058 L .791 .644 0 r
F
.5511 .70868 m .54929 .69945 L .55879 .70037 L .785 .58 0 r
F
.54973 .75598 m .55742 .75355 L .5617 .75508 L .38 0 0 r
F
.5617 .75508 m .55141 .7585 L .54973 .75598 L .503 .086 0 r
F
.55326 .71915 m .55525 .72956 L .54998 .73789 L .752 .536 0 r
F
.54998 .73789 m .55525 .72956 L .55525 .74085 L .883 .726 0 r
F
.55008 .81854 m .55152 .80725 L .56297 .80819 L .743 .624 0 r
F
.56297 .80819 m .56167 .81706 L .55008 .81854 L .804 .757 0 r
F
.55016 1.00415 m .55234 .99059 L .56182 .99394 L .787 .67 0 r
F
.56182 .99394 m .55919 1.00572 L .55016 1.00415 L .808 .713 0 r
F
.55033 .76199 m .55128 .75986 L .55117 .761 L 0 0 0 r
F
.56043 .71702 m .55039 .71242 L .5511 .70868 L .791 .678 0 r
F
.55326 .71915 m .55039 .71242 L .56043 .71702 L .835 .463 0 r
F
.55136 .52632 m .55658 .56066 L .55094 .55971 L .991 .734 0 r
F
.55094 .55971 m .55658 .56066 L .55478 .59321 L .994 .752 0 r
F
.55728 .96047 m .55979 .9698 L .55102 .96878 L .794 .429 0 r
F
.55294 .97882 m .55102 .96878 L .55979 .9698 L .823 .534 0 r
F
.55151 1.06815 m .55654 1.06951 L .55104 1.07407 L .55 .78 0 r
F
.5606 1.07575 m .55104 1.07407 L .55933 1.06912 L .748 .992 0 r
F
.55933 1.06912 m .55104 1.07407 L .55654 1.06951 L .542 .914 0 r
F
.55106 1.04931 m .56079 1.04656 L .56111 1.05709 L .569 .393 0 r
F
.55879 .70037 m .56039 .71118 L .5511 .70868 L .806 .613 0 r
F
.5511 .70868 m .56039 .71118 L .56043 .71702 L .782 .679 0 r
F
.55141 .7585 m .55848 .76041 L .55117 .761 L .554 .109 0 r
F
.56478 .94698 m .56582 .95451 L .55128 .95728 L .747 .495 0 r
F
.55617 .95635 m .55728 .96047 L .55128 .95728 L .784 .627 0 r
F
.55136 .52632 m .55396 .5253 L .55658 .56066 L .981 .719 0 r
F
.55141 .7585 m .5617 .75508 L .56304 .75848 L .689 .415 0 r
F
.56304 .75848 m .55848 .76041 L .55141 .7585 L .54 .064 0 r
F
.55654 1.06951 m .55151 1.06815 L .56069 1.06396 L .649 .775 0 r
F
.55152 .80725 m .55308 .79884 L .56436 .80129 L .748 .695 0 r
F
.56436 .80129 m .56297 .80819 L .55152 .80725 L .736 .657 0 r
F
.56012 .78358 m .56706 .78818 L .55173 .79151 L .739 .712 0 r
F
.55173 .79151 m .55379 .78522 L .56012 .78358 L .739 .711 0 r
F
.55173 .79151 m .56706 .78818 L .56436 .80129 L .739 .695 0 r
F
.56436 .80129 m .55308 .79884 L .55173 .79151 L .749 .694 0 r
F
.55592 1.04793 m .55199 1.04504 L .55504 1.04184 L .543 0 0 r
F
.552 1.04505 m .55199 1.04504 L .55592 1.04793 L F
.5555 1.04805 m .552 1.04505 L .55592 1.04793 L F
.58802 .50931 m .55221 .51272 L .55202 .51135 L .792 .656 0 r
F
.58802 .50931 m .55233 .51355 L .55221 .51272 L F
.5558 1.16638 m .55232 1.16324 L .55242 1.16317 L .876 .935 0 r
F
.58802 .50931 m .55303 .51865 L .55233 .51355 L .792 .656 0 r
F
.55234 .99059 m .55294 .97882 L .56146 .97937 L .823 .642 0 r
F
.56146 .97937 m .56182 .99394 L .55234 .99059 L .81 .622 0 r
F
.5558 1.16638 m .55242 1.16317 L .55514 1.16134 L .876 .935 0 r
F
.55678 1.16538 m .5529 1.17312 L .55934 1.15911 L .698 .814 0 r
F
.5529 1.17312 m .5558 1.16638 L .55934 1.15911 L .689 .806 0 r
F
.55492 1.1721 m .5529 1.17312 L .55297 1.17298 L .652 .662 0 r
F
.55979 .9698 m .56146 .97937 L .55294 .97882 L .802 .504 0 r
F
.55341 1.1721 m .55492 1.1721 L .55297 1.17298 L .652 .662 0 r
F
.55299 .66768 m .56125 .6665 L .55554 .67788 L .784 .597 0 r
F
.58802 .50931 m .55326 .52024 L .55303 .51865 L .792 .656 0 r
F
.58802 .50931 m .58871 .56083 L .55326 .52024 L F
.55672 .52421 m .55396 .5253 L .55326 .52024 L .789 .656 0 r
F
.56043 .71702 m .56695 .72316 L .55326 .71915 L .771 .346 0 r
F
.56175 .73817 m .55525 .72956 L .55326 .71915 L .834 .589 0 r
F
.55326 .71915 m .56695 .72316 L .56175 .73817 L .812 .582 0 r
F
.55615 1.16664 m .5697 1.17212 L .55341 1.1721 L .619 .725 0 r
F
.55349 1.11529 m .55699 1.10767 L .56387 1.11067 L .614 .534 0 r
F
.5628 1.11814 m .55349 1.11529 L .56387 1.11067 L .639 .606 0 r
F
.56012 .78358 m .55379 .78522 L .55384 .77865 L .709 .551 0 r
F
.55384 .77865 m .56071 .77043 L .56752 .77364 L .736 .587 0 r
F
.56752 .77364 m .56774 .78016 L .55384 .77865 L .738 .593 0 r
F
.55384 .77865 m .56774 .78016 L .56012 .78358 L .74 .539 0 r
F
.55396 .5253 m .55851 .56005 L .55658 .56066 L .94 .694 0 r
F
.55403 .52586 m .55396 .5253 L .55672 .52421 L .789 .656 0 r
F
.55404 .52591 m .55403 .52586 L .55672 .52421 L F
.58871 .56083 m .55404 .52591 L .55672 .52421 L F
.58871 .56083 m .55477 .53152 L .55404 .52591 L F
.55423 1.04073 m .55433 1.03878 L .56098 1.03998 L .881 .536 0 r
F
.55504 1.04184 m .55423 1.04073 L .56059 1.04003 L .513 0 0 r
F
.55653 1.08388 m .55426 1.07529 L .55582 1.07541 L .624 .538 0 r
F
.55429 1.12884 m .55876 1.12384 L .5717 1.12707 L .68 .62 0 r
F
.5658 1.13146 m .55429 1.12884 L .5717 1.12707 L .679 .647 0 r
F
.56121 1.02129 m .56098 1.03998 L .55433 1.03878 L .909 .695 0 r
F
.55433 1.03878 m .55528 1.02581 L .56121 1.02129 L .993 .85 0 r
F
.58871 .56083 m .55851 .56005 L .55477 .53152 L .789 .656 0 r
F
.55658 .56066 m .56233 .59559 L .55478 .59321 L .984 .735 0 r
F
.55478 .59321 m .56233 .59559 L .55559 .6278 L .987 .789 0 r
F
.5607 1.04007 m .56492 1.04487 L .55504 1.04184 L .637 .102 0 r
F
.55504 1.04184 m .56059 1.04003 L .5607 1.04007 L .513 0 0 r
F
.55504 1.04184 m .56492 1.04487 L .56438 1.04591 L .5 0 0 r
F
.55592 1.04793 m .55504 1.04184 L .55613 1.04788 L .576 .026 0 r
F
.55613 1.04788 m .55504 1.04184 L .5572 1.04278 L F
.5558 1.16638 m .55514 1.16134 L .55536 1.16119 L .97 .922 0 r
F
.55525 .74085 m .55525 .72956 L .56175 .73817 L .703 .579 0 r
F
.56175 .73817 m .56428 .74497 L .55525 .74085 L .735 .683 0 r
F
.55525 .74085 m .56428 .74497 L .55742 .75355 L .819 .571 0 r
F
.55528 1.02581 m .55919 1.00572 L .56158 1.00951 L .945 .874 0 r
F
.56158 1.00951 m .56121 1.02129 L .55528 1.02581 L .986 .781 0 r
F
.5558 1.16638 m .55536 1.16119 L .55676 1.16024 L .97 .922 0 r
F
.57183 .67973 m .55554 .67788 L .56125 .6665 L .804 .622 0 r
F
.57103 .68887 m .55815 .68484 L .55554 .67788 L .816 .618 0 r
F
.55554 .67788 m .57183 .67973 L .57103 .68887 L .803 .628 0 r
F
.55559 .6278 m .56233 .59559 L .56784 .63059 L .925 .703 0 r
F
.56468 .65431 m .55853 .65341 L .55559 .6278 L .873 .673 0 r
F
.55559 .6278 m .56784 .63059 L .56468 .65431 L .925 .691 0 r
F
.56482 .69553 m .55575 .69523 L .55815 .68484 L .824 .598 0 r
F
.55575 .69523 m .56482 .69553 L .57053 .70332 L .768 .4 0 r
F
.57053 .70332 m .55879 .70037 L .55575 .69523 L .755 .307 0 r
F
.55934 1.15911 m .5558 1.16638 L .55676 1.16024 L .967 .92 0 r
F
.55653 1.08388 m .55582 1.07541 L .55888 1.07562 L .624 .538 0 r
F
.55613 1.04788 m .5572 1.04278 L .56079 1.04656 L .576 .026 0 r
F
.55717 1.16558 m .55615 1.16664 L .55648 1.16598 L .619 .725 0 r
F
.55717 1.16558 m .5697 1.17212 L .55615 1.16664 L F
.56582 .95451 m .55789 .96091 L .55617 .95635 L .784 .627 0 r
F
.55782 .96086 m .55728 .96047 L .55617 .95635 L F
.55789 .96091 m .55782 .96086 L .55617 .95635 L F
.55627 1.09435 m .55653 1.08388 L .56403 1.08706 L .654 .635 0 r
F
.56665 1.09697 m .55627 1.09435 L .56403 1.08706 L .667 .658 0 r
F
.55872 1.10238 m .55627 1.09435 L .56665 1.09697 L .627 .687 0 r
F
.55682 1.1654 m .55648 1.16598 L .55673 1.16548 L .619 .725 0 r
F
.55682 1.1654 m .55717 1.16558 L .55648 1.16598 L F
.56403 1.08706 m .55653 1.08388 L .56659 1.07458 L .692 .611 0 r
F
.55653 1.08388 m .55888 1.07562 L .56457 1.07644 L .689 .606 0 r
F
.56069 1.06396 m .56106 1.06814 L .55654 1.06951 L .745 .934 0 r
F
.55654 1.06951 m .56106 1.06814 L .55933 1.06912 L .281 .777 0 r
F
.55658 .56066 m .55851 .56005 L .56233 .59559 L .937 .675 0 r
F
.55682 1.1654 m .55673 1.16548 L .55673 1.16548 L .619 .725 0 r
F
.55678 1.16538 m .55682 1.1654 L .55673 1.16548 L F
.55934 1.15911 m .55676 1.16024 L .5572 1.15995 L .967 .92 0 r
F
.55934 1.15911 m .5697 1.17212 L .55678 1.16538 L .517 .713 0 r
F
.56387 1.11067 m .55699 1.10767 L .55872 1.10238 L .383 .587 0 r
F
.55934 1.15911 m .5572 1.15995 L .55871 1.15892 L .967 .92 0 r
F
.56113 1.04691 m .5572 1.04278 L .56438 1.04591 L .576 .026 0 r
F
.55979 .9698 m .55728 .96047 L .56897 .97003 L .889 .492 0 r
F
.56577 .96652 m .56897 .97003 L .55728 .96047 L .561 .025 0 r
F
.55742 .75355 m .56428 .74497 L .56714 .75613 L .768 .497 0 r
F
.57009 .75804 m .5617 .75508 L .55742 .75355 L .239 0 0 r
F
.55742 .75355 m .56714 .75613 L .57009 .75804 L .542 0 0 r
F
.56582 .95451 m .56097 .9631 L .55789 .96091 L .784 .627 0 r
F
.55815 .68484 m .57103 .68887 L .56482 .69553 L .82 .598 0 r
F
.56304 .75848 m .56429 .76418 L .55848 .76041 L .806 .48 0 r
F
.55848 .76041 m .56429 .76418 L .56071 .77043 L .779 .538 0 r
F
.58871 .56083 m .55886 .56253 L .55851 .56005 L .789 .656 0 r
F
.55851 .56005 m .56355 .59539 L .56233 .59559 L .871 .63 0 r
F
.56944 .66248 m .55853 .65341 L .56468 .65431 L .829 .456 0 r
F
.56125 .6665 m .55853 .65341 L .56944 .66248 L .803 .512 0 r
F
.55934 1.15911 m .55871 1.15892 L .55917 1.15861 L .967 .92 0 r
F
.57077 1.10391 m .55872 1.10238 L .56665 1.09697 L .642 .724 0 r
F
.56387 1.11067 m .55872 1.10238 L .57077 1.10391 L .64 .725 0 r
F
.55876 1.12384 m .5628 1.11814 L .5769 1.12156 L .677 .577 0 r
F
.5717 1.12707 m .55876 1.12384 L .5769 1.12156 L .68 .62 0 r
F
.56039 .71118 m .55879 .70037 L .57053 .70332 L .822 .625 0 r
F
.58871 .56083 m .55994 .57005 L .55886 .56253 L .79 .68 0 r
F
.55888 .85125 m .55961 .83621 L .56927 .8362 L .847 .705 0 r
F
.56927 .8362 m .56927 .84909 L .55888 .85125 L .834 .686 0 r
F
.55914 .86787 m .55888 .85125 L .56927 .84909 L .835 .692 0 r
F
.55888 1.07562 m .5606 1.07575 L .56457 1.07644 L .689 .606 0 r
F
.56927 .84909 m .56926 .86305 L .55914 .86787 L .826 .679 0 r
F
.55959 .87172 m .55914 .86787 L .56926 .86305 L .797 .601 0 r
F
.55917 1.15861 m .56175 1.15852 L .55934 1.15911 L .634 .868 0 r
F
.55935 1.15849 m .56175 1.15852 L .55917 1.15861 L F
.55919 1.00572 m .56182 .99394 L .5647 .99668 L .95 .9 0 r
F
.5647 .99668 m .56158 1.00951 L .55919 1.00572 L .943 .901 0 r
F
.56 1.06874 m .56372 1.07306 L .55933 1.06912 L .976 .92 0 r
F
.55933 1.06912 m .56372 1.07306 L .5606 1.07575 L .545 .892 0 r
F
.57857 1.15436 m .5697 1.17212 L .55934 1.15911 L .684 .786 0 r
F
.56031 1.15784 m .56175 1.15852 L .55935 1.15849 L .634 .868 0 r
F
.56926 .86305 m .57095 .8668 L .55959 .87172 L .815 .633 0 r
F
.55959 .87172 m .57095 .8668 L .58657 .88406 L .79 .571 0 r
F
.57818 .89104 m .55959 .87172 L .5792 .88391 L .754 .585 0 r
F
.56203 .88887 m .55959 .87172 L .57818 .89104 L .786 .581 0 r
F
.58657 .88406 m .5792 .88391 L .55959 .87172 L .803 .521 0 r
F
.55961 .83621 m .56167 .81706 L .5722 .81839 L .778 .676 0 r
F
.5722 .81839 m .56927 .8362 L .55961 .83621 L .837 .771 0 r
F
.56146 .97937 m .55979 .9698 L .56935 .9771 L .913 .582 0 r
F
.56897 .97003 m .56935 .9771 L .55979 .9698 L .904 .526 0 r
F
.58871 .56083 m .55997 .57027 L .55994 .57005 L .79 .68 0 r
F
.58871 .56083 m .56033 .5728 L .55997 .57027 L F
.56533 1.06685 m .56372 1.07306 L .56 1.06874 L .976 .92 0 r
F
.56533 1.06685 m .56 1.06874 L .56106 1.06814 L F
.56706 .78818 m .56012 .78358 L .56774 .78016 L .797 .695 0 r
F
.56487 .19005 m .58137 .19728 L .56013 .2228 L .824 .627 0 r
F
.58137 .19728 m .57259 .22626 L .56013 .2228 L .809 .603 0 r
F
.5725 .22752 m .56327 .22587 L .56013 .2228 L .549 0 0 r
F
.56013 .2228 m .57259 .22626 L .5725 .22752 L .508 0 0 r
F
.56017 1.15777 m .56466 1.15098 L .57212 1.1464 L .818 1 0 r
F
.56644 1.1537 m .56017 1.15777 L .57212 1.1464 L .77 .992 0 r
F
.56025 1.18237 m .5697 1.17212 L .58352 1.17809 L .652 .933 0 r
F
.56025 1.18237 m .58352 1.17809 L .56089 1.1878 L .36 .832 0 r
F
.56078 1.15752 m .56175 1.15852 L .56031 1.15784 L .634 .868 0 r
F
.58871 .56083 m .56079 .57604 L .56033 .5728 L .79 .68 0 r
F
.57053 .70332 m .56951 .71301 L .56039 .71118 L .851 .685 0 r
F
.56906 .71778 m .56043 .71702 L .56039 .71118 L .847 .73 0 r
F
.56039 .71118 m .56951 .71301 L .56906 .71778 L .837 .735 0 r
F
.56695 .72316 m .56043 .71702 L .56906 .71778 L .769 .351 0 r
F
.5607 1.04007 m .56059 1.04003 L .56098 1.03998 L .513 0 0 r
F
.56067 1.04006 m .56059 1.04003 L .5607 1.04007 L F
.5606 1.07575 m .56148 1.07499 L .56457 1.07644 L .689 .606 0 r
F
.56067 1.04006 m .56069 1.04007 L .5607 1.04007 L .153 .446 0 r
F
.5607 1.04007 m .56069 1.04007 L .5607 1.04007 L .513 0 0 r
F
.56069 1.04007 m .5611 1.0403 L .5607 1.04007 L .153 .446 0 r
F
.56069 1.04007 m .56147 1.04035 L .5611 1.0403 L F
.56106 1.06814 m .56069 1.06396 L .56111 1.05709 L .845 .738 0 r
F
.56134 1.04044 m .56492 1.04487 L .5607 1.04007 L .637 .102 0 r
F
.56098 1.03998 m .56119 1.04025 L .5607 1.04007 L F
.56429 .76418 m .56752 .77364 L .56071 .77043 L .771 .529 0 r
F
.56078 1.15752 m .57857 1.15436 L .56175 1.15852 L .634 .868 0 r
F
.561 1.15738 m .57857 1.15436 L .56078 1.15752 L F
.58871 .56083 m .56089 .57678 L .56079 .57604 L .79 .68 0 r
F
.56585 1.06009 m .56111 1.05709 L .56079 1.04656 L .993 .746 0 r
F
.56079 1.04656 m .56385 1.04977 L .56585 1.06009 L .99 .737 0 r
F
.56089 1.1878 m .58352 1.17809 L .58316 1.1808 L .32 .803 0 r
F
.58871 .56083 m .58753 .59572 L .56089 .57678 L .79 .68 0 r
F
.58547 .59426 m .56238 .58719 L .56089 .57678 L .803 .674 0 r
F
.56582 .95451 m .56577 .96652 L .56097 .9631 L .784 .627 0 r
F
.56098 1.03998 m .56121 1.02129 L .56555 1.02894 L .95 .758 0 r
F
.56622 1.04204 m .56098 1.03998 L .56555 1.02894 L .969 .716 0 r
F
.56119 1.04025 m .56098 1.03998 L .56307 1.04092 L F
.56307 1.04092 m .56098 1.03998 L .56622 1.04204 L F
.56551 1.15433 m .57857 1.15436 L .561 1.15738 L .634 .868 0 r
F
.56111 1.05709 m .56585 1.06009 L .56106 1.06814 L .953 .937 0 r
F
.56736 1.05907 m .56106 1.06814 L .56384 1.06347 L .976 .92 0 r
F
.56736 1.05907 m .56533 1.06685 L .56106 1.06814 L F
.5611 1.0403 m .56147 1.04035 L .56134 1.04044 L .153 .446 0 r
F
.5611 1.13639 m .5658 1.13146 L .57546 1.12631 L .803 .995 0 r
F
.56143 1.13741 m .5611 1.13639 L .56714 1.13215 L .94 .886 0 r
F
.56438 1.04591 m .56219 1.04802 L .56113 1.04691 L .435 0 0 r
F
.56158 1.00951 m .57584 1.01393 L .56121 1.02129 L .777 .594 0 r
F
.56121 1.02129 m .57584 1.01393 L .56555 1.02894 L .793 .969 0 r
F
.57024 .67061 m .56125 .6665 L .56944 .66248 L .839 .587 0 r
F
.56125 .6665 m .57024 .67061 L .57183 .67973 L .829 .62 0 r
F
.56147 1.04035 m .56144 1.04056 L .56134 1.04044 L .153 .446 0 r
F
.56392 1.14522 m .56143 1.13741 L .56714 1.13215 L .94 .886 0 r
F
.56144 1.04056 m .56207 1.04056 L .56226 1.04158 L .153 .446 0 r
F
.56147 1.04035 m .56207 1.04056 L .56144 1.04056 L F
.56935 .9771 m .56827 .98448 L .56146 .97937 L .97 .78 0 r
F
.56182 .99394 m .56146 .97937 L .56827 .98448 L .975 .757 0 r
F
.56148 1.07499 m .56206 1.07449 L .56659 1.07458 L .689 .606 0 r
F
.56457 1.07644 m .56148 1.07499 L .56659 1.07458 L F
.5647 .99668 m .57931 1.00725 L .56158 1.00951 L .773 .673 0 r
F
.56158 1.00951 m .57931 1.00725 L .57584 1.01393 L .768 .619 0 r
F
.57239 .8072 m .5722 .81839 L .56167 .81706 L .78 .669 0 r
F
.56167 .81706 m .56297 .80819 L .57239 .8072 L .834 .785 0 r
F
.5617 .75508 m .57009 .75804 L .5693 .76014 L .707 .186 0 r
F
.5693 .76014 m .56304 .75848 L .5617 .75508 L .705 .423 0 r
F
.57245 .73712 m .56175 .73817 L .56695 .72316 L .864 .644 0 r
F
.57245 .73712 m .56428 .74497 L .56175 .73817 L .875 .746 0 r
F
.56827 .98448 m .5647 .99668 L .56182 .99394 L .941 .927 0 r
F
.56282 .92215 m .56203 .88887 L .57818 .89104 L .784 .62 0 r
F
.56206 1.07449 m .56236 1.07423 L .56659 1.07458 L .689 .606 0 r
F
.56207 1.04056 m .56257 1.04196 L .56226 1.04158 L .153 .446 0 r
F
.56207 1.04056 m .56351 1.04107 L .56257 1.04196 L F
.56438 1.04591 m .56259 1.04844 L .56219 1.04802 L .435 0 0 r
F
.56233 .59559 m .56355 .59539 L .56784 .63059 L .869 .619 0 r
F
.56531 1.07447 m .56236 1.07423 L .56372 1.07306 L .692 .591 0 r
F
.58547 .59426 m .56248 .5879 L .56238 .58719 L .803 .674 0 r
F
.58547 .59426 m .56266 .58917 L .56248 .5879 L F
.56257 1.04196 m .56351 1.04107 L .56438 1.04421 L .153 .446 0 r
F
.56438 1.04591 m .56385 1.04977 L .56259 1.04844 L .435 0 0 r
F
.58547 .59426 m .56352 .59516 L .56266 .58917 L .803 .674 0 r
F
.5628 1.11814 m .56387 1.11067 L .57678 1.11355 L .652 .618 0 r
F
.5769 1.12156 m .5628 1.11814 L .57678 1.11355 L .652 .616 0 r
F
.56282 .92215 m .57818 .89104 L .59648 .90272 L .801 .641 0 r
F
.59648 .90272 m .56398 .94167 L .56282 .92215 L .801 .643 0 r
F
.56297 .80819 m .56436 .80129 L .57401 .79874 L .838 .758 0 r
F
.57401 .79874 m .57239 .8072 L .56297 .80819 L .838 .759 0 r
F
.5693 .76014 m .56429 .76418 L .56304 .75848 L .705 .414 0 r
F
.56352 1.04111 m .56351 1.04107 L .56622 1.04204 L .969 .716 0 r
F
.56355 .59539 m .56352 .59516 L .58547 .59426 L .803 .674 0 r
F
.56352 1.04112 m .56352 1.04111 L .56622 1.04204 L .969 .716 0 r
F
.56623 1.04233 m .56352 1.04112 L .56622 1.04204 L F
.56355 1.04122 m .56352 1.04112 L .56623 1.04233 L F
.56357 .59552 m .56355 .59539 L .58547 .59426 L .803 .674 0 r
F
.56357 1.0413 m .56355 1.04122 L .56623 1.04233 L .969 .716 0 r
F
.5636 .59584 m .56357 .59552 L .58547 .59426 L .803 .674 0 r
F
.56359 1.04135 m .56357 1.0413 L .56623 1.04233 L .969 .716 0 r
F
.56373 1.04186 m .56359 1.04135 L .56623 1.04233 L F
.56361 .59591 m .5636 .59584 L .58547 .59426 L .803 .674 0 r
F
.58753 .59572 m .56563 .61246 L .56361 .59591 L .801 .701 0 r
F
.58753 .59572 m .56361 .59591 L .58547 .59426 L .803 .674 0 r
F
.56531 1.07447 m .56372 1.07306 L .5644 1.07045 L .692 .591 0 r
F
.56379 1.04209 m .56373 1.04186 L .56623 1.04233 L .969 .716 0 r
F
.56438 1.04421 m .56379 1.04209 L .56623 1.04233 L F
.56384 1.06347 m .56665 1.05875 L .56736 1.05907 L .979 .915 0 r
F
.56585 1.06009 m .56385 1.04977 L .56438 1.04591 L .811 .558 0 r
F
.57678 1.11355 m .56387 1.11067 L .57077 1.10391 L .606 .662 0 r
F
.57281 1.13755 m .56392 1.14522 L .56975 1.13032 L .923 .86 0 r
F
.56392 1.14522 m .56714 1.13215 L .56975 1.13032 L .94 .886 0 r
F
.56466 1.15098 m .56392 1.14522 L .57281 1.13755 L .927 .914 0 r
F
.56398 .94167 m .59648 .90272 L .61036 .91121 L .809 .655 0 r
F
.61036 .91121 m .59762 .93987 L .56398 .94167 L .811 .659 0 r
F
.56398 .94167 m .59762 .93987 L .57046 .94612 L .769 .466 0 r
F
.56478 .94698 m .56398 .94167 L .57046 .94612 L .763 .476 0 r
F
.56403 1.08706 m .56659 1.07458 L .57663 1.09071 L .813 .731 0 r
F
.57949 1.09354 m .56403 1.08706 L .57663 1.09071 L .572 .806 0 r
F
.56665 1.09697 m .56403 1.08706 L .57949 1.09354 L .852 .765 0 r
F
.57726 .74317 m .56428 .74497 L .57245 .73712 L .893 .79 0 r
F
.57191 .75864 m .56428 .74497 L .57726 .74317 L .888 .694 0 r
F
.56714 .75613 m .56428 .74497 L .57191 .75864 L .951 .598 0 r
F
.56429 .76418 m .5693 .76014 L .5766 .76559 L .889 .758 0 r
F
.57105 .76962 m .56752 .77364 L .56429 .76418 L .777 .532 0 r
F
.5766 .76559 m .57105 .76962 L .56429 .76418 L .827 .428 0 r
F
.57401 .79874 m .56436 .80129 L .56706 .78818 L .838 .794 0 r
F
.56438 1.04591 m .56492 1.04487 L .56736 1.05907 L .862 .61 0 r
F
.56736 1.05907 m .56567 1.0583 L .56438 1.04591 L .993 .781 0 r
F
.56492 1.04487 m .56438 1.04421 L .56623 1.04233 L .969 .716 0 r
F
.56659 1.07458 m .5644 1.07045 L .5646 1.06967 L .692 .591 0 r
F
.56659 1.07458 m .56531 1.07447 L .5644 1.07045 L F
.56659 1.07458 m .5646 1.06967 L .56467 1.06942 L F
.57212 1.1464 m .56466 1.15098 L .57281 1.13755 L .927 .914 0 r
F
.56467 1.06942 m .56495 1.06833 L .58729 1.09328 L .347 .39 0 r
F
.56659 1.07458 m .56467 1.06942 L .58729 1.09328 L F
.56876 .65779 m .56468 .65431 L .56784 .63059 L .976 .752 0 r
F
.56944 .66248 m .56468 .65431 L .56876 .65779 L .831 .375 0 r
F
.5647 .99668 m .5647 .99668 L .5647 .99668 L .754 .316 0 r
F
.5647 .99668 m .5647 .99668 L .5647 .99668 L F
.5647 .99668 m .5647 .99668 L .5647 .99668 L F
.58068 .99856 m .58068 .99856 L .5647 .99668 L F
.5647 .99668 m .56907 .98891 L .56827 .98448 L .845 .857 0 r
F
.56907 .98891 m .58068 .99856 L .5647 .99668 L .754 .316 0 r
F
.5647 .99668 m .58068 .99856 L .57931 1.00725 L .849 .624 0 r
F
.56582 .95451 m .56478 .94698 L .57046 .94612 L .778 .517 0 r
F
.57103 .68887 m .57456 .70011 L .56482 .69553 L .839 .63 0 r
F
.5798 .70897 m .57053 .70332 L .56482 .69553 L .824 .364 0 r
F
.56482 .69553 m .57456 .70011 L .5798 .70897 L .837 .442 0 r
F
.56786 .15964 m .58811 .16909 L .56487 .19005 L .831 .617 0 r
F
.58811 .16909 m .58137 .19728 L .56487 .19005 L .829 .613 0 r
F
.566 1.05001 m .56492 1.04487 L .56623 1.04233 L .969 .716 0 r
F
.56548 1.04814 m .56492 1.04487 L .566 1.05001 L F
.56495 1.06833 m .56505 1.06794 L .58729 1.09328 L .347 .39 0 r
F
.57858 1.07963 m .58729 1.09328 L .56505 1.06794 L 0 .172 0 r
F
.57858 1.07963 m .56505 1.06794 L .56512 1.06768 L F
.57858 1.07963 m .56512 1.06768 L .56514 1.06761 L F
.56514 1.06761 m .56538 1.06669 L .57858 1.07963 L 0 .157 0 r
F
.56905 1.06553 m .57858 1.07963 L .56538 1.06669 L .227 .305 0 r
F
.56905 1.06553 m .56538 1.06669 L .5654 1.06659 L F
.5654 1.06659 m .56671 1.06154 L .57546 1.06367 L .266 .47 0 r
F
.566 1.05117 m .56548 1.04814 L .56677 1.05277 L .153 .446 0 r
F
.57194 1.15013 m .57857 1.15436 L .56551 1.15433 L .712 .841 0 r
F
.57194 1.15013 m .56551 1.15433 L .56644 1.1537 L F
.57584 1.01393 m .57194 1.03416 L .56555 1.02894 L .271 .227 0 r
F
.56555 1.02894 m .57015 1.04098 L .56677 1.05277 L .256 .143 0 r
F
.56555 1.02894 m .57194 1.03416 L .57015 1.04098 L .402 .217 0 r
F
.58753 .59572 m .58548 .6255 L .56563 .61246 L .801 .701 0 r
F
.58548 .6255 m .56715 .62498 L .56563 .61246 L .818 .692 0 r
F
.56584 1.05996 m .56567 1.0583 L .56665 1.05875 L .979 .915 0 r
F
.57455 .96636 m .56577 .96652 L .56582 .95451 L .742 .593 0 r
F
.56577 .96652 m .57455 .96636 L .57843 .97926 L .742 .609 0 r
F
.56897 .96974 m .56897 .97003 L .56577 .96652 L .787 .604 0 r
F
.57546 1.12631 m .5658 1.13146 L .58113 1.12186 L .678 .977 0 r
F
.5658 1.13146 m .5717 1.12707 L .58113 1.12186 L .605 .945 0 r
F
.57046 .94612 m .57346 .95733 L .56582 .95451 L .833 .59 0 r
F
.56582 .95451 m .57346 .95733 L .57455 .96636 L .832 .599 0 r
F
.56584 1.05996 m .56665 1.05875 L .56585 1.06009 L .979 .915 0 r
F
.56644 1.05376 m .56677 1.05277 L .566 1.05117 L .956 .876 0 r
F
.56677 1.05277 m .566 1.05001 L .56623 1.04233 L .969 .716 0 r
F
.57194 1.15013 m .56644 1.1537 L .5674 1.15246 L .712 .841 0 r
F
.56649 1.05401 m .56677 1.05277 L .56644 1.05376 L .956 .876 0 r
F
.56654 1.05433 m .56677 1.05277 L .56649 1.05401 L F
.56655 1.05439 m .56654 1.05433 L .56677 1.05277 L .417 .37 0 r
F
.56672 1.05535 m .56655 1.05439 L .56677 1.05277 L F
.58729 1.09328 m .5775 1.08508 L .56659 1.07458 L 0 .185 0 r
F
.56659 1.07458 m .56659 1.07458 L .56659 1.07458 L .046 .329 0 r
F
.56659 1.07458 m .56659 1.07458 L .56659 1.07458 L 0 .303 0 r
F
.56659 1.07458 m .56659 1.07458 L .56659 1.07458 L F
.57859 1.0886 m .56659 1.07458 L .56659 1.07458 L .046 .329 0 r
F
.57859 1.0886 m .57663 1.09071 L .56659 1.07458 L F
.56659 1.07458 m .5775 1.08508 L .57067 1.07935 L 0 .303 0 r
F
.58064 1.09523 m .56665 1.09697 L .57949 1.09354 L .852 .772 0 r
F
.58064 1.09523 m .56665 1.09697 L .58064 1.09523 L F
.58167 1.09856 m .56665 1.09697 L .58064 1.09523 L F
.57077 1.10391 m .56665 1.09697 L .58167 1.09856 L .84 .798 0 r
F
.56671 1.06154 m .56687 1.06094 L .57546 1.06367 L .266 .47 0 r
F
.56717 1.05797 m .56672 1.05535 L .56677 1.05277 L .417 .37 0 r
F
.57801 1.04741 m .56677 1.05277 L .57015 1.04098 L .39 .274 0 r
F
.57546 1.06367 m .56677 1.05277 L .57801 1.04741 L .436 .375 0 r
F
.56888 1.05541 m .56717 1.05797 L .56677 1.05277 L .417 .37 0 r
F
.57546 1.06367 m .56687 1.06094 L .5669 1.06083 L F
.57546 1.06367 m .5669 1.06083 L .56695 1.06064 L F
.57546 1.06367 m .56695 1.06064 L .56734 1.05913 L F
.56906 .71778 m .57461 .72286 L .56695 .72316 L .734 .311 0 r
F
.56695 .72316 m .57461 .72286 L .57245 .73712 L .852 .639 0 r
F
.56774 .78016 m .57418 .7837 L .56706 .78818 L .804 .701 0 r
F
.56706 .78818 m .57418 .7837 L .5794 .78452 L .827 .773 0 r
F
.5794 .78452 m .57782 .79185 L .56706 .78818 L .828 .788 0 r
F
.56706 .78818 m .57782 .79185 L .57401 .79874 L .824 .791 0 r
F
.56714 .75613 m .57177 .75912 L .57009 .75804 L .995 .816 0 r
F
.56714 .75613 m .57191 .75864 L .57177 .75912 L F
.56755 .62499 m .56717 .62509 L .56715 .62498 L .811 .735 0 r
F
.56755 .62499 m .56717 .62509 L .56717 .62509 L F
.56755 .62499 m .56718 .62518 L .56717 .62509 L F
.56718 1.05804 m .56717 1.05797 L .56888 1.05541 L .417 .37 0 r
F
.56755 .62499 m .56727 .62595 L .56718 .62518 L .811 .735 0 r
F
.56736 1.05907 m .56718 1.05804 L .56888 1.05541 L .417 .37 0 r
F
.56755 .62499 m .56784 .63059 L .56727 .62595 L .811 .735 0 r
F
.56734 1.05913 m .56734 1.05913 L .57434 1.06226 L .417 .37 0 r
F
.57546 1.06367 m .56734 1.05913 L .57434 1.06226 L F
.56734 1.05913 m .56736 1.05907 L .57434 1.06226 L F
.57434 1.06226 m .56736 1.05907 L .56888 1.05541 L F
.57013 1.14896 m .57194 1.15013 L .5674 1.15246 L .712 .841 0 r
F
.56774 .78016 m .56752 .77364 L .57105 .76962 L .882 .706 0 r
F
.58548 .6255 m .56815 .63682 L .56755 .62499 L .811 .735 0 r
F
.57105 .76962 m .57797 .77747 L .56774 .78016 L .855 .674 0 r
F
.57418 .7837 m .56774 .78016 L .57797 .77747 L .85 .651 0 r
F
.56784 .63059 m .56875 .6486 L .56876 .65779 L .857 .671 0 r
F
.5705 .13074 m .59308 .14176 L .56786 .15964 L .833 .598 0 r
F
.59308 .14176 m .58811 .16909 L .56786 .15964 L .836 .604 0 r
F
.58548 .6255 m .5971 .6489 L .56815 .63682 L .811 .735 0 r
F
.56874 .63706 m .56819 .63758 L .56815 .63682 L .83 .716 0 r
F
.56875 .6486 m .56819 .63758 L .56874 .63706 L F
.56827 .98448 m .56907 .98891 L .56935 .9771 L .82 .74 0 r
F
.5971 .6489 m .56904 .64979 L .56874 .63706 L .83 .716 0 r
F
.56904 .64979 m .56875 .6486 L .56874 .63706 L F
.57215 .66227 m .56876 .65779 L .56875 .6486 L .99 .729 0 r
F
.56876 .65779 m .57215 .66227 L .57618 .66855 L .629 .183 0 r
F
.56876 .65779 m .57618 .66855 L .56944 .66248 L .828 .365 0 r
F
.57843 .97926 m .56927 .97557 L .56897 .96974 L .787 .604 0 r
F
.56927 .97557 m .56897 .97003 L .56897 .96974 L F
.5971 .6489 m .57073 .65657 L .56904 .64979 L .826 .745 0 r
F
.57546 1.06367 m .57858 1.07963 L .56905 1.06553 L .227 .305 0 r
F
.57771 .71902 m .56906 .71778 L .56951 .71301 L .861 .757 0 r
F
.57461 .72286 m .56906 .71778 L .57771 .71902 L .745 .273 0 r
F
.57744 .98957 m .56907 .98891 L .56937 .97621 L .797 .639 0 r
F
.57744 .98957 m .58068 .99856 L .56907 .98891 L .74 .358 0 r
F
.58068 .99856 m .58068 .99856 L .56907 .98891 L F
.57112 .86104 m .56926 .86305 L .56927 .84909 L .995 .816 0 r
F
.57095 .8668 m .56926 .86305 L .57112 .86104 L .943 .656 0 r
F
.56927 .84909 m .56927 .8362 L .57207 .83481 L .938 .77 0 r
F
.57207 .83481 m .57086 .84876 L .56927 .84909 L .994 .848 0 r
F
.56927 .84909 m .57086 .84876 L .57112 .86104 L .995 .816 0 r
F
.57843 .97926 m .56928 .97586 L .56927 .97557 L .787 .604 0 r
F
.5741 .81795 m .57207 .83481 L .56927 .8362 L .948 .851 0 r
F
.56927 .8362 m .5722 .81839 L .5741 .81795 L .98 .914 0 r
F
.56928 .97586 m .57843 .97926 L .56929 .97609 L .772 .633 0 r
F
.56937 .97621 m .56935 .9771 L .56929 .97609 L .797 .639 0 r
F
.56929 .97609 m .57843 .97926 L .57744 .98957 L .772 .633 0 r
F
.57456 .76092 m .5693 .76014 L .57009 .75804 L .672 .145 0 r
F
.5693 .76014 m .57456 .76092 L .5766 .76559 L .875 .767 0 r
F
.57024 .67061 m .56944 .66248 L .57618 .66855 L .958 .679 0 r
F
.56951 .71301 m .57053 .70332 L .5798 .70897 L .827 .662 0 r
F
.5798 .70897 m .57767 .7148 L .56951 .71301 L .852 .777 0 r
F
.56951 .71301 m .57767 .7148 L .57771 .71902 L .863 .756 0 r
F
.5697 1.17212 m .57857 1.15436 L .59282 1.16692 L .775 .902 0 r
F
.5697 1.17212 m .59282 1.16692 L .58352 1.17809 L .678 .941 0 r
F
.57281 1.13755 m .56975 1.13032 L .57028 1.12995 L .923 .86 0 r
F
.58277 1.14013 m .57013 1.14896 L .57099 1.14786 L .766 .846 0 r
F
.58277 1.14013 m .57857 1.15436 L .57013 1.14896 L F
.57801 1.04741 m .57015 1.04098 L .57194 1.03416 L .454 .267 0 r
F
.57618 .66855 m .57803 .67709 L .57024 .67061 L .964 .699 0 r
F
.57183 .67973 m .57024 .67061 L .57803 .67709 L .964 .703 0 r
F
.57281 1.13755 m .57028 1.12995 L .57208 1.12869 L .997 .842 0 r
F
.57046 .94612 m .59762 .93987 L .59576 .94479 L .76 .444 0 r
F
.57046 .94612 m .59576 .94479 L .59685 .95324 L .795 .567 0 r
F
.59685 .95324 m .57346 .95733 L .57046 .94612 L .795 .569 0 r
F
.57067 1.07935 m .5775 1.08508 L .57859 1.0886 L 0 .303 0 r
F
.5971 .6489 m .57152 .65972 L .57073 .65657 L .826 .745 0 r
F
.57077 1.10391 m .58167 1.09856 L .58187 1.09919 L .84 .798 0 r
F
.57077 1.10391 m .58167 1.09856 L .58167 1.09856 L F
.58348 1.10433 m .57077 1.10391 L .58144 1.09937 L .836 .767 0 r
F
.57678 1.11355 m .57077 1.10391 L .58348 1.10433 L .842 .744 0 r
F
.57086 .84876 m .57207 .83481 L .57315 .83348 L .877 .765 0 r
F
.57315 .83348 m .57243 .84791 L .57086 .84876 L .938 .795 0 r
F
.57086 .84876 m .57243 .84791 L .57344 .85741 L .946 .733 0 r
F
.57344 .85741 m .57112 .86104 L .57086 .84876 L .98 .798 0 r
F
.57112 .86104 m .57272 .86605 L .57095 .8668 L .874 .587 0 r
F
.58657 .88406 m .57095 .8668 L .57272 .86605 L .854 .543 0 r
F
.58277 1.14013 m .57099 1.14786 L .57352 1.14507 L .766 .846 0 r
F
.57099 1.14786 m .57212 1.1464 L .57352 1.14507 L F
.5825 .70238 m .57103 .68887 L .57183 .67973 L .864 .684 0 r
F
.5825 .70238 m .57456 .70011 L .57103 .68887 L .911 .661 0 r
F
.57105 .76962 m .5766 .76559 L .57797 .77747 L .926 .609 0 r
F
.57272 .86605 m .57112 .86104 L .57344 .85741 L .954 .608 0 r
F
.5971 .6489 m .57195 .66147 L .57152 .65972 L .826 .745 0 r
F
.58113 1.12186 m .5717 1.12707 L .58655 1.1159 L .816 .997 0 r
F
.5717 1.12707 m .5769 1.12156 L .58655 1.1159 L .808 .998 0 r
F
.57191 .75864 m .57279 .75978 L .57177 .75912 L .729 .473 0 r
F
.57803 .67709 m .57909 .6889 L .57183 .67973 L .97 .723 0 r
F
.57183 .67973 m .57909 .6889 L .5825 .70238 L .969 .655 0 r
F
.57726 .74317 m .57995 .75912 L .57191 .75864 L .921 .738 0 r
F
.57995 .75912 m .58211 .76007 L .57191 .75864 L .29 0 0 r
F
.57191 .75864 m .5741 .75987 L .57456 .76092 L .729 .473 0 r
F
.57191 .75864 m .57456 .76092 L .57279 .75978 L F
.57191 .75864 m .58211 .76007 L .58169 .76413 L .752 .407 0 r
F
.57584 1.01393 m .58309 1.02865 L .57194 1.03416 L .352 .304 0 r
F
.57194 1.03416 m .58309 1.02865 L .57801 1.04741 L .301 .196 0 r
F
.5971 .6489 m .57212 .66215 L .57195 .66147 L .826 .745 0 r
F
.57207 .83481 m .5741 .81795 L .57521 .81712 L .862 .784 0 r
F
.57521 .81712 m .57315 .83348 L .57207 .83481 L .862 .783 0 r
F
.57472 1.13116 m .57281 1.13755 L .57208 1.12869 L .994 .824 0 r
F
.57472 1.13116 m .57208 1.12869 L .57273 1.12823 L F
.5971 .6489 m .58313 .67479 L .57212 .66215 L .826 .745 0 r
F
.58313 .67479 m .57213 .66218 L .57212 .66215 L F
.57407 1.1393 m .57212 1.1464 L .57472 1.13116 L .859 .798 0 r
F
.57212 1.1464 m .57281 1.13755 L .57472 1.13116 L .892 .824 0 r
F
.57352 1.14507 m .57212 1.1464 L .57228 1.14581 L .766 .846 0 r
F
.58313 .67479 m .57213 .66218 L .57213 .66218 L .826 .745 0 r
F
.58313 .67479 m .57213 .66219 L .57213 .66218 L F
.58313 .67479 m .58287 .67455 L .57213 .66219 L F
.58287 .67455 m .57213 .6622 L .57213 .66219 L F
.58287 .67455 m .57213 .6622 L .57213 .6622 L F
.58287 .67455 m .57213 .6622 L .57213 .6622 L F
.58287 .67455 m .58275 .67443 L .57213 .6622 L F
.58275 .67443 m .57215 .66225 L .57213 .6622 L F
.58275 .67443 m .57215 .66227 L .57215 .66225 L F
.57618 .66855 m .57215 .66227 L .59126 .68582 L .669 .193 0 r
F
.57215 .66227 m .58772 .68026 L .5859 .67922 L .725 .249 0 r
F
.58275 .67443 m .58232 .67402 L .57215 .66227 L .826 .745 0 r
F
.5722 .81839 m .57239 .8072 L .57523 .80576 L .971 .825 0 r
F
.57523 .80576 m .5741 .81795 L .5722 .81839 L .993 .877 0 r
F
.57352 1.14507 m .57228 1.14581 L .57231 1.14572 L .766 .846 0 r
F
.57407 1.1393 m .58277 1.14013 L .57231 1.14572 L .761 .799 0 r
F
.57752 .79632 m .57523 .80576 L .57239 .8072 L .959 .923 0 r
F
.57239 .8072 m .57401 .79874 L .57752 .79632 L .926 .848 0 r
F
.57243 .84791 m .57315 .83348 L .5756 .83336 L .998 .842 0 r
F
.5756 .83336 m .57511 .84673 L .57243 .84791 L .999 .832 0 r
F
.57243 .84791 m .57511 .84673 L .57536 .85628 L 1 .818 0 r
F
.57536 .85628 m .57344 .85741 L .57243 .84791 L .998 .792 0 r
F
.57245 .73712 m .57461 .72286 L .58191 .7318 L .861 .648 0 r
F
.57245 .73712 m .58191 .7318 L .57726 .74317 L .905 .787 0 r
F
.58408 .23203 m .5725 .22752 L .57259 .22626 L .566 0 0 r
F
.58137 .19728 m .59337 .20399 L .57259 .22626 L .859 .66 0 r
F
.59337 .20399 m .58283 .2308 L .57259 .22626 L .85 .645 0 r
F
.57259 .22626 m .58283 .2308 L .58408 .23203 L .445 0 0 r
F
.57344 .85741 m .57677 .8623 L .57272 .86605 L .902 .532 0 r
F
.58657 .88406 m .57272 .86605 L .57677 .8623 L .887 .505 0 r
F
.57472 1.13116 m .57273 1.12823 L .59051 1.11873 L .788 .935 0 r
F
.57273 1.12823 m .5751 1.12657 L .59051 1.11873 L F
.57766 .81543 m .5756 .83336 L .57315 .83348 L .991 .882 0 r
F
.57315 .83348 m .57521 .81712 L .57766 .81543 L .984 .885 0 r
F
.57677 .8623 m .57344 .85741 L .57536 .85628 L .902 .493 0 r
F
.57992 .95734 m .57346 .95733 L .59685 .95324 L .093 0 0 r
F
.57455 .96636 m .57346 .95733 L .57992 .95734 L .32 .231 0 r
F
.57752 .79632 m .57401 .79874 L .57782 .79185 L .914 .951 0 r
F
.57472 1.13116 m .58277 1.14013 L .57407 1.1393 L .753 .807 0 r
F
.57705 .80176 m .57521 .81712 L .5741 .81795 L .864 .781 0 r
F
.5741 .81795 m .57523 .80576 L .57705 .80176 L .909 .806 0 r
F
.5741 .75987 m .58169 .76413 L .57571 .76356 L .729 .473 0 r
F
.57797 .77747 m .5794 .78452 L .57418 .7837 L .857 .662 0 r
F
.57992 .95734 m .58408 .96292 L .57455 .96636 L .362 .281 0 r
F
.58408 .96292 m .58757 .97106 L .57455 .96636 L .37 .302 0 r
F
.57843 .97926 m .57455 .96636 L .58757 .97106 L .319 .347 0 r
F
.58521 .7095 m .57456 .70011 L .5825 .70238 L .891 .545 0 r
F
.57456 .70011 m .58521 .7095 L .58681 .71332 L .891 .547 0 r
F
.58681 .71332 m .5798 .70897 L .57456 .70011 L .86 .407 0 r
F
.57771 .71902 m .58211 .7219 L .57461 .72286 L .66 .167 0 r
F
.58191 .7318 m .57461 .72286 L .58211 .7219 L .89 .636 0 r
F
.58277 1.14013 m .57472 1.13116 L .59051 1.11873 L .767 .81 0 r
F
.5751 1.12657 m .58526 1.12109 L .59051 1.11873 L .788 .935 0 r
F
.57546 1.12631 m .58526 1.12109 L .5751 1.12657 L F
.57511 .84673 m .5756 .83336 L .57964 .84407 L .937 .776 0 r
F
.57511 .84673 m .57964 .84407 L .57937 .85426 L .934 .762 0 r
F
.57937 .85426 m .57536 .85628 L .57511 .84673 L .922 .763 0 r
F
.58269 .79142 m .57766 .81543 L .57521 .81712 L .977 .898 0 r
F
.57521 .81712 m .57705 .80176 L .58269 .79142 L .99 .881 0 r
F
.57523 .80576 m .57752 .79632 L .58074 .78871 L .868 .861 0 r
F
.58074 .78871 m .57705 .80176 L .57523 .80576 L .783 .806 0 r
F
.57857 .86039 m .57536 .85628 L .57937 .85426 L .742 .369 0 r
F
.57536 .85628 m .57857 .86039 L .57677 .8623 L .738 .37 0 r
F
.57546 1.12631 m .57697 1.12513 L .58526 1.12109 L .788 .935 0 r
F
.61351 1.04649 m .57801 1.04741 L .57546 1.06367 L .793 .711 0 r
F
.57546 1.06367 m .61096 1.06306 L .61351 1.04649 L .796 .726 0 r
F
.61096 1.06306 m .57546 1.06367 L .57858 1.07963 L .799 .709 0 r
F
.57858 1.07963 m .57546 1.06367 L .57858 1.07963 L F
.57766 .81543 m .581 .81762 L .5756 .83336 L .953 .841 0 r
F
.58405 .83424 m .5756 .83336 L .581 .81762 L .896 .763 0 r
F
.58405 .83424 m .57964 .84407 L .5756 .83336 L .895 .766 0 r
F
.57571 .76356 m .57747 .76583 L .5766 .76559 L .729 .473 0 r
F
.57571 .76356 m .58169 .76413 L .57747 .76583 L F
.58706 1.00888 m .58309 1.02865 L .57584 1.01393 L .515 .382 0 r
F
.57584 1.01393 m .57584 1.01393 L .57584 1.01393 L .448 .244 0 r
F
.57584 1.01393 m .57584 1.01393 L .57584 1.01393 L F
.57584 1.01393 m .57931 1.00725 L .58706 1.00888 L F
.57803 .67709 m .57618 .66855 L .59126 .68582 L .962 .698 0 r
F
.5766 .76559 m .58093 .77208 L .57797 .77747 L .89 .58 0 r
F
.58225 .77025 m .58093 .77208 L .5766 .76559 L .892 .579 0 r
F
.5766 .76559 m .58135 .76691 L .58225 .77025 L .892 .582 0 r
F
.57859 1.0886 m .57663 1.09071 L .57859 1.0886 L .046 .329 0 r
F
.5791 1.09025 m .57663 1.09071 L .57859 1.0886 L F
.57949 1.09354 m .57663 1.09071 L .57831 1.0904 L .964 .636 0 r
F
.57677 .8623 m .57857 .86039 L .5823 .87027 L .892 .606 0 r
F
.57677 .8623 m .5823 .87027 L .58657 .88406 L .901 .486 0 r
F
.57678 1.11355 m .58348 1.10433 L .58424 1.10619 L .842 .744 0 r
F
.58543 1.10909 m .57678 1.11355 L .58424 1.10619 L .862 .791 0 r
F
.58596 1.11038 m .57678 1.11355 L .58543 1.10909 L F
.5769 1.12156 m .57678 1.11355 L .58596 1.11038 L .862 .782 0 r
F
.5769 1.12156 m .58596 1.11038 L .58629 1.11117 L F
.58655 1.1159 m .5769 1.12156 L .5864 1.11144 L .905 .916 0 r
F
.5769 1.12156 m .58629 1.11117 L .5864 1.11144 L .862 .782 0 r
F
.57697 1.12513 m .57786 1.12443 L .58526 1.12109 L .788 .935 0 r
F
.58074 .78871 m .58269 .79142 L .57705 .80176 L .924 .927 0 r
F
.60881 .33238 m .61362 .37312 L .57713 .33074 L .832 .661 0 r
F
.61362 .37312 m .57993 .3722 L .57713 .33074 L .827 .662 0 r
F
.58191 .7318 m .58799 .73767 L .57726 .74317 L .927 .819 0 r
F
.58339 .76113 m .57726 .74317 L .58799 .73767 L .91 .738 0 r
F
.58273 .76058 m .57726 .74317 L .58339 .76113 L F
.57726 .74317 m .58273 .76058 L .57978 .75813 L .934 .745 0 r
F
.57744 .98957 m .57843 .97926 L .58912 .98841 L .229 .156 0 r
F
.58218 .9891 m .58068 .99856 L .57744 .98957 L .218 .087 0 r
F
.58068 .99856 m .58068 .99856 L .57744 .98957 L F
.58169 .76413 m .57901 .76626 L .57747 .76583 L .519 .109 0 r
F
.5775 1.08508 m .5775 1.08508 L .57859 1.0886 L 0 .303 0 r
F
.5775 1.08508 m .5775 1.08508 L .58338 1.10409 L .254 .319 0 r
F
.5775 1.08508 m .58729 1.09328 L .58338 1.10409 L F
.57782 .79185 m .58074 .78871 L .57752 .79632 L .911 .872 0 r
F
.581 .81762 m .57766 .81543 L .58269 .79142 L .952 .843 0 r
F
.5798 .70897 m .58681 .71332 L .57767 .7148 L .884 .816 0 r
F
.57767 .7148 m .58681 .71332 L .58394 .71758 L .878 .842 0 r
F
.58394 .71758 m .57771 .71902 L .57767 .7148 L .919 .799 0 r
F
.58211 .7219 m .57771 .71902 L .58394 .71758 L .661 .187 0 r
F
.58074 .78871 m .57782 .79185 L .5794 .78452 L .898 .879 0 r
F
.57827 1.12411 m .59051 1.11873 L .57786 1.12443 L .78 .944 0 r
F
.58093 .77208 m .5842 .77504 L .57797 .77747 L .956 .69 0 r
F
.57797 .77747 m .5842 .77504 L .58431 .78127 L .978 .879 0 r
F
.58431 .78127 m .5794 .78452 L .57797 .77747 L .988 .72 0 r
F
.61468 1.02895 m .58309 1.02865 L .57801 1.04741 L .8 .7 0 r
F
.57801 1.04741 m .61351 1.04649 L .61468 1.02895 L .795 .685 0 r
F
.57909 .6889 m .57803 .67709 L .59134 .68919 L .941 .702 0 r
F
.59126 .68582 m .59134 .68919 L .57803 .67709 L .933 .561 0 r
F
.59648 .90272 m .57818 .89104 L .5792 .88391 L .807 .636 0 r
F
.5833 1.12066 m .59051 1.11873 L .57827 1.12411 L .78 .944 0 r
F
.5833 1.12066 m .57827 1.12411 L .58051 1.12235 L F
.57949 1.09354 m .57831 1.0904 L .57833 1.0904 L .964 .636 0 r
F
.57949 1.09354 m .57833 1.0904 L .57837 1.09039 L F
.57874 1.09032 m .57949 1.09354 L .57837 1.09039 L .96 .668 0 r
F
.58757 .97106 m .58912 .98841 L .57843 .97926 L .202 .157 0 r
F
.57857 1.15436 m .58277 1.14013 L .6011 1.15072 L .788 .869 0 r
F
.57857 1.15436 m .6011 1.15072 L .59282 1.16692 L .759 .898 0 r
F
.58018 .86261 m .57857 .86039 L .57937 .85426 L .971 .698 0 r
F
.5823 .87027 m .57857 .86039 L .58018 .86261 L .862 .498 0 r
F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L .799 .709 0 r
F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L .802 .719 0 r
F
.57858 1.07963 m .58729 1.09328 L .57858 1.07963 L .797 .692 0 r
F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L .804 .684 0 r
F
.57858 1.07963 m .6091 1.07764 L .61096 1.06306 L .802 .719 0 r
F
.6084 1.09145 m .6091 1.07764 L .57858 1.07963 L .804 .684 0 r
F
.57858 1.07963 m .58729 1.09328 L .6084 1.09145 L .797 .692 0 r
F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L .799 .709 0 r
F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L F
.57858 1.07963 m .57858 1.07963 L .57858 1.07963 L F
.57874 1.09032 m .57949 1.09354 L .57874 1.09032 L .96 .668 0 r
F
.57966 1.09207 m .57874 1.09032 L .57888 1.0903 L F
.57966 1.09207 m .57949 1.09354 L .57874 1.09032 L F
.57966 1.09207 m .57888 1.0903 L .57888 1.0903 L F
.57966 1.09207 m .57888 1.0903 L .57891 1.09029 L F
.57966 1.09207 m .57891 1.09029 L .57892 1.09029 L F
.57966 1.09207 m .57892 1.09029 L .5791 1.09025 L F
.58169 .76413 m .57929 .76633 L .57901 .76626 L .519 .109 0 r
F
.5825 .70238 m .57909 .6889 L .58825 .70625 L .956 .654 0 r
F
.58825 .70625 m .57909 .6889 L .5909 .69685 L .927 .672 0 r
F
.59134 .68919 m .5909 .69685 L .57909 .6889 L .93 .646 0 r
F
.60045 .89108 m .59843 .8962 L .5792 .88391 L .819 .604 0 r
F
.5792 .88391 m .59843 .8962 L .59648 .90272 L .781 .638 0 r
F
.5792 .88391 m .58657 .88406 L .60045 .89108 L .819 .601 0 r
F
.58169 .76413 m .58043 .76665 L .57929 .76633 L .519 .109 0 r
F
.58706 1.00888 m .57931 1.00725 L .57931 1.00725 L .447 .25 0 r
F
.58706 1.00888 m .57931 1.00725 L .58555 1.00645 L F
.57931 1.00725 m .57931 1.00725 L .58555 1.00645 L F
.58555 1.00645 m .57931 1.00725 L .58068 .99856 L F
.58258 .85241 m .57937 .85426 L .57964 .84407 L .877 .714 0 r
F
.57937 .85426 m .58258 .85241 L .59034 .85754 L .716 .385 0 r
F
.59034 .85754 m .5872 .86187 L .57937 .85426 L .698 .538 0 r
F
.58018 .86261 m .57937 .85426 L .5872 .86187 L .758 .529 0 r
F
.58074 .78871 m .5794 .78452 L .58431 .78127 L .885 .958 0 r
F
.58064 1.09523 m .57949 1.09354 L .57966 1.09207 L .978 .695 0 r
F
.58915 .83989 m .57964 .84407 L .58405 .83424 L .823 .669 0 r
F
.58475 .84758 m .57964 .84407 L .58915 .83989 L .822 .665 0 r
F
.57964 .84407 m .58475 .84758 L .58709 .84982 L .751 .231 0 r
F
.58709 .84982 m .58258 .85241 L .57964 .84407 L .695 .627 0 r
F
.58211 .76007 m .58106 .75961 L .57978 .75813 L .934 .745 0 r
F
.57978 .75813 m .58106 .75961 L .57995 .75912 L F
.59685 .95324 m .60125 .95344 L .57992 .95734 L .138 0 0 r
F
.58408 .96292 m .57992 .95734 L .60125 .95344 L .73 .353 0 r
F
.61803 .41694 m .58413 .41875 L .57993 .3722 L .828 .667 0 r
F
.61362 .37312 m .61803 .41694 L .57993 .3722 L .827 .668 0 r
F
.5872 .86187 m .58634 .86812 L .58018 .86261 L .75 .5 0 r
F
.5823 .87027 m .58018 .86261 L .58777 .87576 L .832 .463 0 r
F
.58777 .87576 m .58018 .86261 L .58634 .86812 L .8 .46 0 r
F
.58323 .76601 m .58135 .76691 L .58043 .76665 L .023 0 0 r
F
.58169 .76413 m .58323 .76601 L .58043 .76665 L .519 .109 0 r
F
.5833 1.12066 m .58051 1.12235 L .58113 1.12186 L .78 .944 0 r
F
.58068 .99856 m .58706 1.00888 L .58068 .99856 L .303 .199 0 r
F
.58912 .98841 m .58801 .99942 L .58068 .99856 L F
.58801 .99942 m .58068 .99856 L .58068 .99856 L F
.58801 .99942 m .58706 1.00888 L .58068 .99856 L F
.58431 .78127 m .58296 .78544 L .58074 .78871 L .756 .869 0 r
F
.58074 .78871 m .58296 .78544 L .58269 .79142 L .933 .879 0 r
F
.58225 .77025 m .5842 .77504 L .58093 .77208 L .955 .695 0 r
F
.58269 .79142 m .59 .80077 L .581 .81762 L .9 .799 0 r
F
.59009 .82291 m .581 .81762 L .59 .80077 L .853 .728 0 r
F
.581 .81762 m .59009 .82291 L .58405 .83424 L .846 .734 0 r
F
.5816 1.12134 m .5833 1.12066 L .58113 1.12186 L .78 .944 0 r
F
.58323 .76601 m .58225 .77025 L .58135 .76691 L .952 .609 0 r
F
.58811 .16909 m .6056 .17848 L .58137 .19728 L .846 .631 0 r
F
.6056 .17848 m .59337 .20399 L .58137 .19728 L .86 .659 0 r
F
.58305 1.10302 m .58144 1.09937 L .58187 1.09919 L .836 .767 0 r
F
.58338 1.10409 m .58144 1.09937 L .58305 1.10302 L F
.58188 1.12104 m .5833 1.12066 L .5816 1.12134 L .78 .944 0 r
F
.5881 .76598 m .58169 .76413 L .58211 .76007 L .853 .507 0 r
F
.58323 .76601 m .58169 .76413 L .5881 .76598 L .665 .121 0 r
F
.58283 1.11998 m .59051 1.11873 L .58188 1.12104 L .846 .86 0 r
F
.58774 .72757 m .58191 .7318 L .58211 .7219 L .91 .655 0 r
F
.58774 .72757 m .58799 .73767 L .58191 .7318 L .958 .79 0 r
F
.58394 .71758 m .58714 .72099 L .58211 .7219 L .745 .282 0 r
F
.58211 .7219 m .58714 .72099 L .58774 .72757 L .918 .641 0 r
F
.58211 .76007 m .58563 .76299 L .5881 .76598 L .649 .13 0 r
F
.58225 .77025 m .58323 .76601 L .58563 .77139 L .958 .621 0 r
F
.58563 .77139 m .5842 .77504 L .58225 .77025 L .974 .677 0 r
F
.58657 .88406 m .5823 .87027 L .58777 .87576 L .839 .443 0 r
F
.58232 .67402 m .59573 .68669 L .58256 .6743 L .563 .05 0 r
F
.58521 .7095 m .5825 .70238 L .58621 .71015 L .794 .426 0 r
F
.58825 .70625 m .58621 .71015 L .5825 .70238 L .952 .746 0 r
F
.58256 .6743 m .59573 .68669 L .60264 .69404 L .563 .05 0 r
F
.60264 .69404 m .58772 .68026 L .58256 .6743 L .616 .098 0 r
F
.58258 .85241 m .58709 .84982 L .59541 .85528 L .661 .547 0 r
F
.59541 .85528 m .59034 .85754 L .58258 .85241 L .683 .428 0 r
F
.58269 .79142 m .58293 .78605 L .5969 .77851 L .699 .37 0 r
F
.5969 .77851 m .59 .80077 L .58269 .79142 L .931 .79 0 r
F
.58277 1.14013 m .59051 1.11873 L .60485 1.13427 L .804 .853 0 r
F
.58277 1.14013 m .60485 1.13427 L .6011 1.15072 L .798 .869 0 r
F
.59337 .20399 m .60749 .21354 L .58283 .2308 L .841 .634 0 r
F
.60749 .21354 m .59472 .23797 L .58283 .2308 L .814 .578 0 r
F
.58408 .23203 m .58283 .2308 L .59336 .23721 L .469 0 0 r
F
.59472 .23797 m .59336 .23721 L .58283 .2308 L .814 .578 0 r
F
.58482 1.1178 m .59051 1.11873 L .58283 1.11998 L .846 .86 0 r
F
.58294 .78571 m .5969 .77851 L .58293 .78605 L .863 .71 0 r
F
.58316 .78534 m .5969 .77851 L .58294 .78571 L F
.58296 .78544 m .58399 .78392 L .58294 .78571 L .706 .893 0 r
F
.58431 .78127 m .58399 .78392 L .58296 .78544 L F
.61421 1.00841 m .58706 1.00888 L .58309 1.02865 L .804 .663 0 r
F
.58309 1.02865 m .61468 1.02895 L .61421 1.00841 L .803 .66 0 r
F
.58309 1.02865 m .58706 1.00888 L .58706 1.00888 L .804 .663 0 r
F
.5839 .67552 m .58313 .67479 L .5971 .6489 L .826 .745 0 r
F
.58763 .77856 m .5969 .77851 L .58316 .78534 L .863 .71 0 r
F
.58763 .77856 m .58316 .78534 L .58396 .78396 L F
.58352 1.17809 m .60435 1.17157 L .58316 1.1808 L .321 .803 0 r
F
.5881 .76598 m .58563 .77139 L .58323 .76601 L .947 .642 0 r
F
.58729 1.09328 m .58753 1.11419 L .58338 1.10409 L .101 .147 0 r
F
.58799 .73767 m .58363 .76133 L .58339 .76113 L .912 .739 0 r
F
.58352 1.17809 m .59282 1.16692 L .60435 1.17157 L .69 .965 0 r
F
.58392 .76157 m .58363 .76133 L .58389 .76089 L .929 .769 0 r
F
.58799 .73767 m .59616 .73981 L .58363 .76133 L .912 .739 0 r
F
.58593 .76335 m .58389 .76089 L .59616 .73981 L .929 .769 0 r
F
.58441 .76198 m .58389 .76089 L .58563 .76299 L F
.58392 .76157 m .58389 .76089 L .58441 .76198 L F
.58393 .67555 m .5839 .67552 L .5971 .6489 L .826 .745 0 r
F
.63728 .67234 m .58393 .67555 L .5971 .6489 L F
.63728 .67234 m .58431 .67591 L .58393 .67555 L .819 .781 0 r
F
.58394 .71758 m .58681 .71332 L .58892 .71833 L .887 .86 0 r
F
.58714 .72099 m .58394 .71758 L .58892 .71833 L .751 .252 0 r
F
.58763 .77856 m .58396 .78396 L .58399 .78392 L .863 .71 0 r
F
.58431 .78127 m .58874 .77578 L .58399 .78392 L .706 .893 0 r
F
.58672 .77923 m .58763 .77856 L .58399 .78392 L .863 .71 0 r
F
.58405 .83424 m .59009 .82291 L .5947 .83176 L .797 .666 0 r
F
.58405 .83424 m .5947 .83176 L .58915 .83989 L .798 .671 0 r
F
.60125 .95344 m .60684 .95971 L .58408 .96292 L .705 .313 0 r
F
.58757 .97106 m .58408 .96292 L .60684 .95971 L .788 .502 0 r
F
.61803 .41694 m .62237 .46615 L .58413 .41875 L .827 .659 0 r
F
.62237 .46615 m .58511 .46675 L .58413 .41875 L .818 .659 0 r
F
.5842 .77504 m .58563 .77139 L .58874 .77578 L .99 .729 0 r
F
.58874 .77578 m .58431 .78127 L .5842 .77504 L .992 .881 0 r
F
.63728 .67234 m .58441 .676 L .58431 .67591 L .819 .781 0 r
F
.63728 .67234 m .58444 .67603 L .58441 .676 L F
.63728 .67234 m .58446 .67605 L .58444 .67603 L F
.63728 .67234 m .62451 .68559 L .58446 .67605 L F
.58446 .67605 m .62451 .68559 L .59075 .68198 L .83 .766 0 r
F
.58915 .83989 m .59027 .8438 L .58475 .84758 L .641 .433 0 r
F
.58475 .84758 m .59027 .8438 L .59066 .8471 L .824 .907 0 r
F
.58746 .84948 m .58475 .84758 L .59066 .8471 L .887 .741 0 r
F
.58727 .84992 m .58709 .84982 L .58475 .84758 L .524 .007 0 r
F
.58727 .84992 m .58475 .84758 L .58848 .85066 L .848 .387 0 r
F
.58871 .85079 m .58475 .84758 L .58746 .84948 L F
.58848 .85066 m .58475 .84758 L .58871 .85079 L F
.58619 1.11802 m .58482 1.1178 L .58655 1.1159 L .83 .871 0 r
F
.62237 .46615 m .6264 .507 L .58511 .46675 L .818 .674 0 r
F
.6264 .507 m .58802 .50931 L .58511 .46675 L .824 .673 0 r
F
.58621 .71015 m .58818 .71199 L .58521 .7095 L .503 0 0 r
F
.58681 .71332 m .58521 .7095 L .58818 .71199 L .925 .544 0 r
F
.63288 .62088 m .58548 .6255 L .58753 .59572 L .82 .718 0 r
F
.63288 .62088 m .64022 .64415 L .58548 .6255 L .818 .75 0 r
F
.64022 .64415 m .5971 .6489 L .58548 .6255 L .827 .74 0 r
F
.58563 .77139 m .5881 .76598 L .58805 .77255 L .943 .855 0 r
F
.58805 .77255 m .58874 .77578 L .58563 .77139 L .812 .409 0 r
F
.5859 .67922 m .58772 .68026 L .59126 .68582 L .725 .249 0 r
F
.59616 .73981 m .5875 .76525 L .58593 .76335 L .862 .674 0 r
F
.59051 1.11873 m .58619 1.11802 L .58655 1.1159 L .83 .871 0 r
F
.58825 .70625 m .59478 .71386 L .58621 .71015 L .764 .477 0 r
F
.58621 .71015 m .59478 .71386 L .58818 .71199 L .583 .009 0 r
F
.5872 .86187 m .59587 .8736 L .58634 .86812 L .737 .488 0 r
F
.58777 .87576 m .58634 .86812 L .59587 .8736 L .76 .433 0 r
F
.58655 1.1159 m .5864 1.11144 L .58704 1.11299 L .905 .916 0 r
F
.58655 1.1159 m .58704 1.11299 L .58718 1.11334 L F
.59051 1.11873 m .58655 1.1159 L .58707 1.11379 L .83 .871 0 r
F
.58657 .88406 m .58777 .87576 L .59712 .88165 L .802 .402 0 r
F
.59712 .88165 m .59981 .88602 L .58657 .88406 L .737 .29 0 r
F
.58657 .88406 m .59981 .88602 L .60045 .89108 L .828 .575 0 r
F
.58712 .77856 m .58763 .77856 L .58672 .77923 L .863 .71 0 r
F
.58818 .71199 m .58892 .71833 L .58681 .71332 L .994 .779 0 r
F
.58706 1.00888 m .58912 .98841 L .58706 1.00888 L .805 .642 0 r
F
.58706 1.00888 m .58706 1.00888 L .58706 1.00888 L F
.58706 1.00888 m .58706 1.00888 L .58706 1.00888 L F
.58706 1.00888 m .58706 1.00888 L .58706 1.00888 L .804 .64 0 r
F
.58706 1.00888 m .58706 1.00888 L .58706 1.00888 L .805 .642 0 r
F
.58706 1.00888 m .58706 1.00888 L .58706 1.00888 L F
.61232 .98765 m .58912 .98841 L .58706 1.00888 L F
.58706 1.00888 m .61421 1.00841 L .61232 .98765 L .804 .64 0 r
F
.58707 1.11379 m .59047 1.11701 L .59051 1.11873 L .893 .879 0 r
F
.58708 1.11376 m .59047 1.11701 L .58707 1.11379 L F
.58736 1.11378 m .58753 1.11419 L .58708 1.11376 L F
.58736 1.11378 m .58708 1.11376 L .58735 1.11374 L F
.58735 1.11374 m .58708 1.11376 L .5871 1.11368 L F
.59715 .85588 m .59541 .85528 L .58709 .84982 L .617 .051 0 r
F
.58734 1.11373 m .58735 1.11374 L .5871 1.11368 L .893 .879 0 r
F
.5871 1.11365 m .58734 1.11373 L .5871 1.11368 L F
.58717 1.11339 m .58723 1.11347 L .5871 1.11365 L F
.58723 1.11347 m .58734 1.11373 L .5871 1.11365 L F
.58932 .77506 m .5969 .77851 L .58712 .77856 L .861 .664 0 r
F
.58932 .77506 m .58712 .77856 L .58874 .77578 L F
.58892 .71833 m .59453 .72221 L .58714 .72099 L .714 .205 0 r
F
.58774 .72757 m .58714 .72099 L .59453 .72221 L .889 .618 0 r
F
.58717 1.11339 m .58718 1.11334 L .58723 1.11347 L .893 .879 0 r
F
.59034 .85754 m .59673 .86308 L .5872 .86187 L .673 .504 0 r
F
.59673 .86308 m .6027 .87058 L .5872 .86187 L .663 .566 0 r
F
.5872 .86187 m .6027 .87058 L .59587 .8736 L .748 .488 0 r
F
.59424 1.12063 m .58753 1.11419 L .58729 1.09328 L .223 .251 0 r
F
.58729 1.09328 m .59308 1.09559 L .59424 1.12063 L .67 .544 0 r
F
.5895 1.09417 m .58811 1.09321 L .58729 1.09328 L .898 .626 0 r
F
.59066 .8471 m .59156 .85151 L .58746 .84948 L .808 .569 0 r
F
.58746 .84948 m .59055 .8519 L .58871 .85079 L .798 .591 0 r
F
.58746 .84948 m .5919 .85272 L .59055 .8519 L F
.58746 .84948 m .59156 .85151 L .5919 .85272 L F
.59616 .73981 m .60461 .75447 L .5875 .76525 L .862 .674 0 r
F
.5881 .76598 m .5875 .76525 L .60461 .75447 L .864 .68 0 r
F
.63098 .59085 m .58753 .59572 L .58871 .56083 L .824 .709 0 r
F
.63098 .59085 m .63288 .62088 L .58753 .59572 L .824 .716 0 r
F
.60684 .95971 m .61098 .97203 L .58757 .97106 L .78 .489 0 r
F
.58912 .98841 m .58757 .97106 L .61098 .97203 L .8 .602 0 r
F
.59126 .68582 m .58772 .68026 L .60264 .69404 L .782 .289 0 r
F
.59453 .72221 m .58799 .73767 L .58774 .72757 L .957 .789 0 r
F
.59587 .8736 m .59712 .88165 L .58777 .87576 L .79 .498 0 r
F
.58799 .73767 m .59453 .72221 L .60791 .72921 L .939 .759 0 r
F
.59616 .73981 m .58799 .73767 L .60791 .72921 L .916 .676 0 r
F
.6264 .507 m .63045 .55818 L .58802 .50931 L .825 .685 0 r
F
.63045 .55818 m .58871 .56083 L .58802 .50931 L .828 .685 0 r
F
.58807 .77073 m .5969 .77851 L .58805 .77255 L .849 .689 0 r
F
.58805 .77255 m .5969 .77851 L .5885 .77469 L F
.58807 .77073 m .58808 .76874 L .60251 .76101 L .864 .68 0 r
F
.5969 .77851 m .58807 .77073 L .60251 .76101 L F
.58808 .76874 m .58808 .76847 L .60251 .76101 L F
.58808 .76847 m .5881 .76598 L .60251 .76101 L F
.60251 .76101 m .5881 .76598 L .60461 .75447 L F
.59308 .14176 m .61577 .15327 L .58811 .16909 L .837 .605 0 r
F
.61577 .15327 m .6056 .17848 L .58811 .16909 L .847 .629 0 r
F
.58826 1.0932 m .58811 1.09321 L .5895 1.09417 L .898 .626 0 r
F
.58818 .71199 m .59478 .71386 L .58892 .71833 L .864 .717 0 r
F
.58825 .70625 m .5909 .69685 L .60196 .70627 L .949 .703 0 r
F
.60196 .70627 m .59478 .71386 L .58825 .70625 L .841 .411 0 r
F
.58886 1.09315 m .58826 1.0932 L .5895 1.09417 L .898 .626 0 r
F
.5885 .77469 m .58932 .77506 L .58871 .77564 L .861 .664 0 r
F
.58932 .77506 m .58874 .77578 L .58871 .77564 L F
.63045 .55818 m .63098 .59085 L .58871 .56083 L .828 .707 0 r
F
.59038 1.09404 m .58927 1.09311 L .58886 1.09315 L .898 .626 0 r
F
.59038 1.09404 m .58886 1.09315 L .5895 1.09417 L F
.58892 .71833 m .59478 .71386 L .59453 .72221 L .695 .405 0 r
F
.61098 .97203 m .61232 .98765 L .58912 .98841 L .801 .604 0 r
F
.5947 .83176 m .59563 .83821 L .58915 .83989 L .734 .58 0 r
F
.59027 .8438 m .58915 .83989 L .59563 .83821 L .696 .465 0 r
F
.59038 1.09404 m .59073 1.09298 L .58927 1.09311 L .898 .626 0 r
F
.6084 1.09145 m .59308 1.09559 L .5895 1.09417 L .648 .246 0 r
F
.59 .80077 m .5969 .77851 L .60638 .79561 L .857 .698 0 r
F
.60638 .79561 m .59784 .81095 L .59 .80077 L .861 .713 0 r
F
.59 .80077 m .59784 .81095 L .59009 .82291 L .835 .714 0 r
F
.5947 .83176 m .59009 .82291 L .59784 .81095 L .804 .668 0 r
F
.59563 .83821 m .59842 .84424 L .59027 .8438 L .737 .526 0 r
F
.59027 .8438 m .59842 .84424 L .59665 .84757 L .648 .782 0 r
F
.59066 .8471 m .59027 .8438 L .59665 .84757 L .632 .779 0 r
F
.59034 .85754 m .59541 .85528 L .60112 .8575 L .708 .478 0 r
F
.60112 .8575 m .59673 .86308 L .59034 .85754 L .711 .496 0 r
F
.59279 1.0928 m .59073 1.09298 L .59038 1.09404 L .898 .626 0 r
F
.59285 1.09369 m .59279 1.0928 L .59038 1.09404 L F
.59047 1.11701 m .59068 1.11722 L .59051 1.11873 L .893 .879 0 r
F
.59051 1.11873 m .59068 1.11722 L .59076 1.11728 L .82 .817 0 r
F
.59051 1.11873 m .59076 1.11728 L .5913 1.1178 L F
.59051 1.11873 m .5913 1.1178 L .5914 1.1179 L F
.59051 1.11873 m .5914 1.1179 L .59168 1.11817 L F
.59051 1.11873 m .59168 1.11817 L .59187 1.11835 L F
.59051 1.11873 m .5919 1.11838 L .59237 1.11883 L .812 .854 0 r
F
.59051 1.11873 m .59187 1.11835 L .5919 1.11838 L F
.59051 1.11873 m .59247 1.11893 L .59273 1.11918 L F
.59051 1.11873 m .59237 1.11883 L .59247 1.11893 L F
.59051 1.11873 m .59273 1.11918 L .59239 1.12077 L F
.59066 .8471 m .59665 .84757 L .59755 .85144 L .737 .511 0 r
F
.59755 .85144 m .59156 .85151 L .59066 .8471 L .73 .518 0 r
F
.59075 .68198 m .62451 .68559 L .59872 .68826 L .83 .766 0 r
F
.59075 .68198 m .59872 .68826 L .59747 .68832 L .84 .764 0 r
F
.5909 .69685 m .59134 .68919 L .60196 .70627 L .953 .672 0 r
F
.59126 .68582 m .60196 .70627 L .59134 .68919 L .947 .589 0 r
F
.60264 .69404 m .60196 .70627 L .59126 .68582 L .955 .627 0 r
F
.59383 .85388 m .59156 .85151 L .59755 .85144 L .735 .556 0 r
F
.5919 .85272 m .59156 .85151 L .59304 .85341 L F
.59304 .85341 m .59156 .85151 L .59383 .85388 L F
.59273 1.11918 m .59353 1.11995 L .59239 1.12077 L .812 .854 0 r
F
.5937 1.12011 m .59424 1.12063 L .59239 1.12077 L F
.59353 1.11995 m .5937 1.12011 L .59239 1.12077 L F
.59424 1.12063 m .59558 1.12092 L .59239 1.12077 L F
.59239 1.12077 m .59558 1.12092 L .59303 1.12147 L F
.59283 1.0928 m .59279 1.0928 L .59285 1.09369 L .898 .626 0 r
F
.59282 1.16692 m .6011 1.15072 L .61394 1.15963 L .792 .941 0 r
F
.59282 1.16692 m .61394 1.15963 L .60435 1.17157 L .697 .967 0 r
F
.59283 1.0928 m .59285 1.09369 L .59288 1.0928 L .898 .626 0 r
F
.59288 1.0928 m .59285 1.09369 L .59291 1.09368 L F
.59288 1.0928 m .59291 1.09368 L .59295 1.09279 L F
.59295 1.09279 m .59291 1.09368 L .59299 1.09366 L F
.59368 1.09273 m .59295 1.09279 L .59299 1.09366 L F
.59387 1.09354 m .59368 1.09273 L .59299 1.09366 L F
.59558 1.12092 m .59661 1.12114 L .59303 1.12147 L .812 .854 0 r
F
.59661 1.12114 m .59662 1.12115 L .59303 1.12147 L F
.59662 1.12115 m .59665 1.12115 L .59303 1.12147 L F
.5977 1.12138 m .59791 1.12142 L .59303 1.12147 L F
.59665 1.12115 m .5977 1.12138 L .59303 1.12147 L F
.59303 1.12147 m .59995 1.12186 L .60485 1.13427 L F
.59791 1.12142 m .59995 1.12186 L .59303 1.12147 L F
.60403 1.10002 m .59308 1.09559 L .6084 1.09145 L .61 .183 0 r
F
.59424 1.12063 m .59308 1.09559 L .60403 1.10002 L .526 .432 0 r
F
.6056 .17848 m .61945 .18816 L .59337 .20399 L .855 .653 0 r
F
.61945 .18816 m .60749 .21354 L .59337 .20399 L .845 .63 0 r
F
.59368 1.09273 m .59387 1.09354 L .598 1.09235 L .725 .524 0 r
F
.59755 .85144 m .59477 .85445 L .59383 .85388 L .674 .445 0 r
F
.598 1.09235 m .59387 1.09354 L .5986 1.09286 L .725 .524 0 r
F
.60403 1.10002 m .59995 1.12186 L .59424 1.12063 L .513 .416 0 r
F
.59478 .71386 m .60791 .72921 L .59453 .72221 L .944 .638 0 r
F
.59784 .81095 m .60681 .82331 L .5947 .83176 L .817 .68 0 r
F
.5947 .83176 m .60681 .82331 L .60842 .82879 L .744 .531 0 r
F
.5947 .83176 m .60842 .82879 L .60188 .83703 L .76 .589 0 r
F
.60188 .83703 m .59563 .83821 L .5947 .83176 L .753 .593 0 r
F
.60749 .21354 m .59476 .23799 L .59472 .23797 L .865 .647 0 r
F
.60749 .21354 m .62181 .22466 L .59476 .23799 L F
.59488 .23793 m .59521 .23824 L .59476 .23799 L .873 .669 0 r
F
.59755 .85144 m .59649 .85549 L .59477 .85445 L .674 .445 0 r
F
.59478 .71386 m .60196 .70627 L .60791 .72921 L .942 .591 0 r
F
.62181 .22466 m .60599 .2474 L .59488 .23793 L .873 .669 0 r
F
.59531 .2383 m .59521 .23824 L .59488 .23793 L F
.60704 .85931 m .60112 .8575 L .59541 .85528 L .462 0 0 r
F
.59842 .84424 m .59563 .83821 L .60188 .83703 L .735 .525 0 r
F
.59573 .68669 m .59747 .68832 L .59728 .68833 L .84 .764 0 r
F
.59576 .94479 m .59762 .93987 L .61045 .94388 L .776 .463 0 r
F
.59685 .95324 m .59576 .94479 L .61045 .94388 L .81 .578 0 r
F
.60593 .88057 m .59587 .8736 L .6027 .87058 L .77 .536 0 r
F
.59712 .88165 m .59587 .8736 L .60593 .88057 L .795 .502 0 r
F
.60461 .75447 m .59616 .73981 L .60791 .72921 L .915 .673 0 r
F
.61036 .91121 m .59648 .90272 L .59843 .8962 L .802 .661 0 r
F
.59755 .85144 m .59971 .85677 L .59649 .85549 L .674 .445 0 r
F
.59971 .85677 m .5966 .85555 L .59649 .85549 L F
.59971 .85677 m .59698 .85578 L .5966 .85555 L F
.59665 .84757 m .59842 .84424 L .6038 .84707 L .647 .78 0 r
F
.59665 .84757 m .6038 .84707 L .6047 .85201 L .685 .498 0 r
F
.6047 .85201 m .59755 .85144 L .59665 .84757 L .696 .486 0 r
F
.60486 .86575 m .59673 .86308 L .60112 .8575 L .732 .526 0 r
F
.60486 .86575 m .6027 .87058 L .59673 .86308 L .712 .575 0 r
F
.61045 .94388 m .61207 .95109 L .59685 .95324 L .819 .595 0 r
F
.60081 .95342 m .59685 .95324 L .6006 .95271 L .512 0 0 r
F
.5969 .77851 m .60461 .75447 L .61733 .77435 L .843 .655 0 r
F
.61733 .77435 m .60638 .79561 L .5969 .77851 L .851 .697 0 r
F
.59753 .85601 m .597 .8558 L .59698 .85578 L .674 .445 0 r
F
.59971 .85677 m .59753 .85601 L .59698 .85578 L F
.59753 .85601 m .59715 .85588 L .597 .8558 L F
.64022 .64415 m .63728 .67234 L .5971 .6489 L .826 .745 0 r
F
.60593 .88057 m .60868 .88686 L .59712 .88165 L .787 .48 0 r
F
.61338 .89202 m .59981 .88602 L .59712 .88165 L .766 .295 0 r
F
.59712 .88165 m .60868 .88686 L .61338 .89202 L .785 .413 0 r
F
.59872 .68826 m .59863 .68977 L .59728 .68833 L .831 .799 0 r
F
.60054 .85706 m .59755 .85144 L .6047 .85201 L .693 .454 0 r
F
.60028 .85697 m .59755 .85144 L .60054 .85706 L F
.60021 .85694 m .59755 .85144 L .60028 .85697 L F
.59992 .85684 m .59755 .85144 L .60021 .85694 L F
.59986 .85682 m .59755 .85144 L .59992 .85684 L F
.59971 .85677 m .59755 .85144 L .59986 .85682 L F
.59762 .93987 m .61036 .91121 L .62446 .91318 L .811 .66 0 r
F
.62446 .91318 m .61862 .94124 L .59762 .93987 L .824 .684 0 r
F
.59762 .93987 m .61862 .94124 L .61045 .94388 L .76 .372 0 r
F
.59784 .81095 m .60638 .79561 L .61597 .81453 L .815 .649 0 r
F
.61597 .81453 m .60681 .82331 L .59784 .81095 L .808 .679 0 r
F
.598 1.09235 m .5986 1.09286 L .5987 1.09229 L .725 .524 0 r
F
.60188 .83703 m .60394 .84259 L .59842 .84424 L .687 .469 0 r
F
.59842 .84424 m .60394 .84259 L .6038 .84707 L .756 .776 0 r
F
.61229 .90213 m .59843 .8962 L .60045 .89108 L .813 .598 0 r
F
.59843 .8962 m .61229 .90213 L .61036 .91121 L .778 .659 0 r
F
.5987 1.09229 m .5986 1.09286 L .59932 1.09275 L .725 .524 0 r
F
.59886 .69002 m .59863 .68977 L .59872 .68826 L .812 .782 0 r
F
.5987 1.09229 m .59932 1.09275 L .59969 1.0922 L .725 .524 0 r
F
.62451 .68559 m .63728 .69534 L .59872 .68826 L .832 .749 0 r
F
.59911 .69028 m .59886 .69002 L .59872 .68826 L .812 .782 0 r
F
.59934 .68837 m .59911 .69028 L .59872 .68826 L F
.60264 .69404 m .59911 .69028 L .59934 .68837 L F
.59969 1.0922 m .59932 1.09275 L .60033 1.09261 L .725 .524 0 r
F
.63728 .69534 m .60504 .69818 L .59934 .68837 L .812 .782 0 r
F
.59969 1.0922 m .60033 1.09261 L .60068 1.09212 L .725 .524 0 r
F
.59981 .88602 m .61338 .89202 L .61115 .89495 L .811 .58 0 r
F
.61115 .89495 m .60045 .89108 L .59981 .88602 L .821 .57 0 r
F
.59995 1.12186 m .60403 1.10002 L .61093 1.12039 L .523 .426 0 r
F
.59995 1.12186 m .61093 1.12039 L .61264 1.12189 L .434 .778 0 r
F
.59995 1.12186 m .61264 1.12189 L .61338 1.12255 L F
.59995 1.12186 m .61338 1.12255 L .60768 1.12253 L F
.59995 1.12186 m .60848 1.1226 L .60485 1.13427 L .812 .854 0 r
F
.60068 1.09212 m .60033 1.09261 L .60132 1.09246 L .725 .524 0 r
F
.60045 .89108 m .61115 .89495 L .61229 .90213 L .817 .596 0 r
F
.6047 .85201 m .60125 .8573 L .60054 .85706 L .682 .438 0 r
F
.60274 .9551 m .6006 .95271 L .61207 .95109 L .512 0 0 r
F
.60081 .95342 m .6006 .95271 L .60125 .95344 L F
.60132 1.09246 m .60514 1.09173 L .60068 1.09212 L .718 .531 0 r
F
.6011 1.15072 m .60485 1.13427 L .6221 1.14441 L .824 .893 0 r
F
.6011 1.15072 m .6221 1.14441 L .61394 1.15963 L .782 .94 0 r
F
.60112 .8575 m .60974 .86014 L .60486 .86575 L .736 .527 0 r
F
.6047 .85201 m .60704 .85931 L .60125 .8573 L .682 .438 0 r
F
.60132 1.09246 m .60554 1.09186 L .60514 1.09173 L .718 .531 0 r
F
.60842 .82879 m .60838 .83632 L .60188 .83703 L .774 .609 0 r
F
.60394 .84259 m .60188 .83703 L .60838 .83632 L .749 .5 0 r
F
.60264 .69404 m .62281 .72875 L .60196 .70627 L .954 .626 0 r
F
.60196 .70627 m .61478 .72561 L .60791 .72921 L .935 .588 0 r
F
.62281 .72875 m .61478 .72561 L .60196 .70627 L .912 .509 0 r
F
.6027 .87058 m .60486 .86575 L .61545 .86716 L .751 .622 0 r
F
.61545 .86716 m .61661 .8726 L .6027 .87058 L .734 .558 0 r
F
.60593 .88057 m .6027 .87058 L .61661 .8726 L .736 .517 0 r
F
.61207 .95109 m .60295 .95534 L .60274 .9551 L .349 0 0 r
F
.61207 .95109 m .62038 .95177 L .60295 .95534 L F
.62038 .95177 m .60297 .95536 L .60295 .95534 L F
.60684 .95971 m .60297 .95536 L .62131 .95185 L .62 .196 0 r
F
.62038 .95177 m .61659 .95275 L .60297 .95536 L .349 0 0 r
F
.6038 .84707 m .60394 .84259 L .61101 .84478 L .786 .8 0 r
F
.6038 .84707 m .61101 .84478 L .61237 .84874 L .698 .445 0 r
F
.61237 .84874 m .6047 .85201 L .6038 .84707 L .695 .504 0 r
F
.60838 .83632 m .61286 .84074 L .60394 .84259 L .738 .484 0 r
F
.60394 .84259 m .61286 .84074 L .61101 .84478 L .785 .801 0 r
F
.6084 1.09145 m .61373 1.09612 L .60403 1.10002 L .587 .159 0 r
F
.60403 1.10002 m .61373 1.09612 L .62193 1.09862 L .384 0 0 r
F
.62193 1.09862 m .6191 1.10174 L .60403 1.10002 L .462 .02 0 r
F
.60403 1.10002 m .6191 1.10174 L .61761 1.11161 L .52 .398 0 r
F
.61761 1.11161 m .61093 1.12039 L .60403 1.10002 L .474 .398 0 r
F
.60435 1.17157 m .61394 1.15963 L .62435 1.16199 L .701 .983 0 r
F
.60791 .72921 m .62525 .75818 L .60461 .75447 L .875 .632 0 r
F
.61733 .77435 m .60461 .75447 L .62525 .75818 L .876 .656 0 r
F
.61376 .85207 m .6047 .85201 L .61237 .84874 L .683 .477 0 r
F
.60719 .85936 m .6047 .85201 L .6156 .85601 L .68 .437 0 r
F
.6047 .85201 m .61376 .85207 L .6156 .85601 L .678 .445 0 r
F
.6047 .85201 m .60718 .85936 L .60704 .85931 L .682 .438 0 r
F
.60718 .85936 m .6047 .85201 L .60719 .85936 L .68 .437 0 r
F
.60485 1.13427 m .60742 1.12602 L .62496 1.12665 L .838 .882 0 r
F
.60485 1.13427 m .62496 1.12665 L .6221 1.14441 L .834 .892 0 r
F
.60486 .86575 m .60974 .86014 L .61986 .86201 L .78 .595 0 r
F
.61986 .86201 m .61545 .86716 L .60486 .86575 L .757 .522 0 r
F
.63728 .69534 m .61238 .71071 L .60504 .69818 L .809 .597 0 r
F
.61238 .71071 m .61204 .71022 L .60504 .69818 L F
.60554 1.09186 m .60544 1.0917 L .60514 1.09173 L .718 .531 0 r
F
.60588 1.09166 m .60544 1.0917 L .60554 1.09186 L F
.60597 1.09179 m .60588 1.09166 L .60554 1.09186 L F
.61577 .15327 m .6337 .16358 L .6056 .17848 L .845 .625 0 r
F
.6337 .16358 m .61945 .18816 L .6056 .17848 L .855 .652 0 r
F
.60693 1.09157 m .60588 1.09166 L .60597 1.09179 L .718 .531 0 r
F
.61661 .8726 m .6194 .8825 L .60593 .88057 L .751 .544 0 r
F
.60868 .88686 m .60593 .88057 L .6194 .8825 L .744 .463 0 r
F
.607 1.09165 m .60693 1.09157 L .60597 1.09179 L .718 .531 0 r
F
.62181 .22466 m .60636 .24772 L .60599 .2474 L .86 .649 0 r
F
.62181 .22466 m .63157 .23779 L .60636 .24772 L F
.60748 .24728 m .60912 .25008 L .60636 .24772 L .859 .646 0 r
F
.60912 .25008 m .60637 .24774 L .60637 .24773 L F
.60912 .25008 m .61455 .25932 L .60637 .24774 L F
.60638 .79561 m .61733 .77435 L .62692 .80368 L .841 .683 0 r
F
.62692 .80368 m .61597 .81453 L .60638 .79561 L .851 .659 0 r
F
.60681 .82331 m .61597 .81453 L .61932 .82769 L .794 .653 0 r
F
.61932 .82769 m .60842 .82879 L .60681 .82331 L .819 .572 0 r
F
.62131 .95185 m .62525 .95638 L .60684 .95971 L .668 .266 0 r
F
.61098 .97203 m .60684 .95971 L .62525 .95638 L .771 .484 0 r
F
.6084 1.09145 m .60693 1.09157 L .607 1.09165 L .718 .531 0 r
F
.6156 .85601 m .60723 .85937 L .60719 .85936 L .745 .594 0 r
F
.6156 .85601 m .61582 .86084 L .60723 .85937 L F
.61582 .86084 m .60974 .86014 L .60723 .85937 L .341 0 0 r
F
.60742 1.12602 m .60848 1.1226 L .61891 1.1235 L .838 .882 0 r
F
.60742 1.12602 m .61891 1.1235 L .62496 1.12665 L F
.63157 .23779 m .61545 .26062 L .60748 .24728 L .859 .646 0 r
F
.61533 .26042 m .61455 .25932 L .60748 .24728 L F
.61945 .18816 m .63898 .204 L .60749 .21354 L .861 .652 0 r
F
.63898 .204 m .62181 .22466 L .60749 .21354 L .861 .651 0 r
F
.61338 1.12255 m .6134 1.12256 L .60768 1.12253 L .434 .778 0 r
F
.6134 1.12256 m .61351 1.12265 L .60768 1.12253 L F
.60768 1.12253 m .61693 1.12273 L .61891 1.1235 L F
.60791 .72921 m .61478 .72561 L .62713 .74381 L .937 .592 0 r
F
.60791 .72921 m .62713 .74381 L .62525 .75818 L .938 .598 0 r
F
.62212 .83317 m .60838 .83632 L .60842 .82879 L .795 .627 0 r
F
.61286 .84074 m .60838 .83632 L .62212 .83317 L .749 .481 0 r
F
.62538 1.08911 m .6084 1.09145 L .6091 1.07764 L .851 .725 0 r
F
.62538 1.08911 m .62502 1.09311 L .6084 1.09145 L .762 .409 0 r
F
.61373 1.09612 m .6084 1.09145 L .62502 1.09311 L .628 .125 0 r
F
.60842 .82879 m .61932 .82769 L .62212 .83317 L .81 .541 0 r
F
.6194 .8825 m .62242 .89033 L .60868 .88686 L .772 .521 0 r
F
.61338 .89202 m .60868 .88686 L .62242 .89033 L .763 .425 0 r
F
.60881 .33238 m .63206 .29309 L .63887 .33266 L .848 .671 0 r
F
.63887 .33266 m .64564 .37213 L .60881 .33238 L .848 .662 0 r
F
.64564 .37213 m .61362 .37312 L .60881 .33238 L .837 .665 0 r
F
.6091 1.07764 m .63632 1.07366 L .6091 1.07764 L .833 .735 0 r
F
.6091 1.07764 m .63632 1.07366 L .62538 1.08911 L F
.64359 1.05798 m .6091 1.07764 L .6091 1.07764 L .833 .731 0 r
F
.64359 1.05798 m .63632 1.07366 L .6091 1.07764 L F
.6091 1.07764 m .6091 1.07764 L .6091 1.07764 L .84 .754 0 r
F
.6091 1.07764 m .61096 1.06306 L .6091 1.07764 L F
.6091 1.07764 m .61096 1.06306 L .64359 1.05798 L F
.61582 .86084 m .61986 .86201 L .60974 .86014 L .481 0 0 r
F
.62446 .91318 m .61036 .91121 L .61229 .90213 L .805 .688 0 r
F
.61045 .94388 m .61862 .94124 L .62997 .94198 L .64 .18 0 r
F
.61207 .95109 m .61045 .94388 L .62997 .94198 L .823 .598 0 r
F
.61351 1.12265 m .61093 1.12039 L .61761 1.11161 L .365 .253 0 r
F
.64351 1.04103 m .61351 1.04649 L .61096 1.06306 L .848 .775 0 r
F
.61096 1.06306 m .64359 1.05798 L .64351 1.04103 L .84 .752 0 r
F
.62525 .95638 m .63137 .96856 L .61098 .97203 L .77 .482 0 r
F
.61232 .98765 m .61098 .97203 L .63137 .96856 L .81 .61 0 r
F
.61101 .84478 m .61286 .84074 L .62161 .8423 L .799 .818 0 r
F
.61237 .84874 m .61101 .84478 L .62161 .8423 L .736 .465 0 r
F
.61115 .89495 m .61338 .89202 L .62578 .89511 L .832 .612 0 r
F
.61229 .90213 m .61115 .89495 L .62618 .89809 L .805 .588 0 r
F
.62578 .89511 m .62618 .89809 L .61115 .89495 L .79 .449 0 r
F
.61204 .71022 m .61483 .71425 L .62281 .72875 L .726 .285 0 r
F
.62997 .94198 m .6321 .94795 L .61207 .95109 L .831 .617 0 r
F
.61207 .95109 m .6321 .94795 L .63237 .94973 L .844 .687 0 r
F
.61207 .95109 m .63237 .94973 L .62131 .95185 L F
.62618 .89809 m .62592 .90724 L .61229 .90213 L .829 .672 0 r
F
.61229 .90213 m .62592 .90724 L .62446 .91318 L .823 .684 0 r
F
.63137 .96856 m .63059 .98336 L .61232 .98765 L .847 .678 0 r
F
.63059 .98336 m .61232 .98765 L .61421 1.00841 L .846 .671 0 r
F
.62161 .8423 m .62249 .84679 L .61237 .84874 L .749 .487 0 r
F
.61376 .85207 m .61237 .84874 L .62249 .84679 L .757 .511 0 r
F
.63728 .69534 m .61247 .71084 L .61238 .71071 L .809 .597 0 r
F
.63728 .69534 m .64336 .73661 L .61247 .71084 L F
.6139 .71203 m .61373 .71266 L .61247 .71084 L .811 .595 0 r
F
.62212 .83317 m .62326 .83972 L .61286 .84074 L .809 .585 0 r
F
.61286 .84074 m .62326 .83972 L .62161 .8423 L .811 .807 0 r
F
.62242 .89033 m .62578 .89511 L .61338 .89202 L .832 .611 0 r
F
.61761 1.11161 m .621 1.11826 L .61351 1.12265 L .613 .517 0 r
F
.62193 1.12283 m .61351 1.12265 L .621 1.11826 L .721 .884 0 r
F
.61351 1.04649 m .64351 1.04103 L .64248 1.02339 L .848 .741 0 r
F
.64248 1.02339 m .61468 1.02895 L .61351 1.04649 L .841 .725 0 r
F
.64564 .37213 m .65353 .41513 L .61362 .37312 L .836 .658 0 r
F
.65353 .41513 m .61803 .41694 L .61362 .37312 L .819 .662 0 r
F
.62502 1.09311 m .62551 1.09739 L .61373 1.09612 L .606 .095 0 r
F
.63021 1.10024 m .62193 1.09862 L .61373 1.09612 L .275 0 0 r
F
.61373 1.09612 m .62742 1.0976 L .63021 1.10024 L .351 0 0 r
F
.61375 .71269 m .61373 .71266 L .6139 .71203 L .811 .595 0 r
F
.61377 .71271 m .61375 .71269 L .6139 .71203 L F
.62249 .84679 m .62388 .85224 L .61376 .85207 L .79 .575 0 r
F
.62433 .85551 m .6156 .85601 L .61376 .85207 L .799 .492 0 r
F
.61376 .85207 m .62388 .85224 L .62433 .85551 L .792 .589 0 r
F
.61379 .71274 m .61377 .71271 L .6139 .71203 L .811 .595 0 r
F
.61383 .7128 m .61379 .71274 L .6139 .71203 L F
.61388 .71288 m .61383 .7128 L .6139 .71203 L F
.61483 .71425 m .61388 .71288 L .6139 .71203 L F
.61561 .71566 m .61483 .71425 L .6139 .71203 L F
.6177 .71946 m .61561 .71566 L .6139 .71203 L F
.61931 .71655 m .6177 .71946 L .6139 .71203 L F
.61394 1.15963 m .6221 1.14441 L .63357 1.14903 L .814 .985 0 r
F
.61394 1.15963 m .63357 1.14903 L .62435 1.16199 L .736 .991 0 r
F
.61421 1.00841 m .63857 1.00326 L .63059 .98336 L .825 .64 0 r
F
.63857 1.00326 m .61421 1.00841 L .61468 1.02895 L .834 .685 0 r
F
.61421 1.00841 m .61421 1.00841 L .61421 1.00841 L .846 .671 0 r
F
.61468 1.02895 m .64248 1.02339 L .63857 1.00326 L .838 .691 0 r
F
.61478 .72561 m .62281 .72875 L .62713 .74381 L .913 .51 0 r
F
.63157 .23779 m .61547 .26066 L .61545 .26062 L .87 .663 0 r
F
.61545 .86716 m .61986 .86201 L .62976 .86107 L .792 .574 0 r
F
.62976 .86107 m .62828 .86548 L .61545 .86716 L .812 .628 0 r
F
.61661 .8726 m .61545 .86716 L .62828 .86548 L .808 .604 0 r
F
.63157 .23779 m .63822 .24596 L .61547 .26066 L .87 .663 0 r
F
.63822 .24596 m .61551 .26073 L .61547 .26066 L .864 .649 0 r
F
.63822 .24596 m .64882 .26053 L .61551 .26073 L F
.64882 .26053 m .6231 .27343 L .61551 .26073 L .867 .675 0 r
F
.6156 .85601 m .62433 .85551 L .6262 .86042 L .825 .581 0 r
F
.6262 .86042 m .61582 .86084 L .6156 .85601 L .804 .639 0 r
F
.6262 .86042 m .61986 .86201 L .61582 .86084 L .64 .186 0 r
F
.61597 .81453 m .62692 .80368 L .63043 .82919 L .862 .679 0 r
F
.63043 .82919 m .61932 .82769 L .61597 .81453 L .849 .684 0 r
F
.62038 .95177 m .62131 .95185 L .61659 .95275 L .349 0 0 r
F
.62828 .86548 m .62896 .87199 L .61661 .8726 L .815 .621 0 r
F
.6194 .8825 m .61661 .8726 L .62896 .87199 L .807 .576 0 r
F
.62193 1.12283 m .61891 1.1235 L .61693 1.12273 L .251 .691 0 r
F
.62525 .75818 m .63482 .79 L .61733 .77435 L .922 .724 0 r
F
.61733 .77435 m .63482 .79 L .62692 .80368 L .927 .716 0 r
F
.61761 1.11161 m .6191 1.10174 L .62342 1.10813 L .596 .467 0 r
F
.61761 1.11161 m .62342 1.10813 L .62417 1.11724 L .634 .546 0 r
F
.62417 1.11724 m .621 1.11826 L .61761 1.11161 L .673 .543 0 r
F
.61951 .72276 m .6177 .71946 L .61931 .71655 L .811 .595 0 r
F
.65353 .41513 m .65878 .46609 L .61803 .41694 L .819 .66 0 r
F
.65878 .46609 m .62237 .46615 L .61803 .41694 L .828 .659 0 r
F
.64245 .93519 m .61862 .94124 L .62446 .91318 L .848 .708 0 r
F
.61862 .94124 m .64245 .93519 L .62997 .94198 L .721 .363 0 r
F
.62193 1.12283 m .62257 1.12298 L .61891 1.1235 L .439 .863 0 r
F
.61891 1.1235 m .62151 1.12313 L .62496 1.12665 L .838 .882 0 r
F
.62193 1.09862 m .6254 1.10421 L .6191 1.10174 L .688 .285 0 r
F
.6191 1.10174 m .6254 1.10421 L .62342 1.10813 L .659 .485 0 r
F
.64336 .73661 m .62689 .73781 L .61931 .71655 L .811 .595 0 r
F
.62439 .73225 m .62281 .72875 L .61931 .71655 L F
.62689 .73781 m .62439 .73225 L .61931 .71655 L F
.62281 .72875 m .62215 .72755 L .61931 .71655 L F
.62215 .72755 m .61951 .72276 L .61931 .71655 L F
.62212 .83317 m .61932 .82769 L .63043 .82919 L .84 .548 0 r
F
.62896 .87199 m .63251 .88116 L .6194 .8825 L .829 .611 0 r
F
.62242 .89033 m .6194 .8825 L .63251 .88116 L .809 .538 0 r
F
.6337 .16358 m .65703 .18259 L .61945 .18816 L .858 .656 0 r
F
.65703 .18259 m .63898 .204 L .61945 .18816 L .858 .655 0 r
F
.6262 .86042 m .62976 .86107 L .61986 .86201 L .537 .036 0 r
F
.62417 1.11724 m .62193 1.12283 L .621 1.11826 L .689 .864 0 r
F
.62131 .95185 m .63237 .94973 L .62366 .95455 L .844 .687 0 r
F
.62151 1.12313 m .62621 1.12246 L .62496 1.12665 L .839 .882 0 r
F
.62161 .8423 m .62326 .83972 L .63083 .84321 L .828 .836 0 r
F
.62249 .84679 m .62161 .8423 L .63083 .84321 L .86 .56 0 r
F
.63898 .204 m .64825 .21626 L .62181 .22466 L .875 .675 0 r
F
.64825 .21626 m .63157 .23779 L .62181 .22466 L .867 .648 0 r
F
.62714 1.10401 m .6254 1.10421 L .62193 1.09862 L .71 .29 0 r
F
.62193 1.09862 m .6296 1.10386 L .62714 1.10401 L .691 .3 0 r
F
.63021 1.10024 m .6296 1.10386 L .62193 1.09862 L F
.63137 1.11687 m .62193 1.12283 L .6301 1.11613 L .742 .892 0 r
F
.62193 1.12283 m .62712 1.1165 L .6301 1.11613 L .746 .933 0 r
F
.62417 1.11724 m .62712 1.1165 L .62193 1.12283 L .74 .918 0 r
F
.63137 1.11687 m .6315 1.12068 L .62193 1.12283 L .742 .895 0 r
F
.6315 1.12068 m .62257 1.12298 L .62193 1.12283 L .439 .863 0 r
F
.6221 1.14441 m .62496 1.12665 L .64135 1.13521 L .87 .922 0 r
F
.6221 1.14441 m .64135 1.13521 L .63357 1.14903 L .801 .984 0 r
F
.63043 .82919 m .63157 .83361 L .62212 .83317 L .849 .569 0 r
F
.62326 .83972 m .62212 .83317 L .63157 .83361 L .859 .618 0 r
F
.6231 .27343 m .62281 .2732 L .62225 .272 L .867 .675 0 r
F
.65878 .46609 m .66561 .50618 L .62237 .46615 L .828 .667 0 r
F
.66561 .50618 m .6264 .507 L .62237 .46615 L .814 .671 0 r
F
.63251 .88116 m .63531 .89105 L .62242 .89033 L .843 .595 0 r
F
.62242 .89033 m .63531 .89105 L .63845 .89407 L .85 .673 0 r
F
.63845 .89407 m .62578 .89511 L .62242 .89033 L .854 .61 0 r
F
.63083 .84321 m .63197 .84711 L .62249 .84679 L .857 .555 0 r
F
.62388 .85224 m .62249 .84679 L .63197 .84711 L .873 .622 0 r
F
.6315 1.12068 m .62942 1.122 L .62257 1.12298 L .439 .863 0 r
F
.6231 .27343 m .62295 .2735 L .62281 .2732 L .867 .675 0 r
F
.63636 .76517 m .62713 .74381 L .62281 .72875 L .844 .481 0 r
F
.62281 .72875 m .63208 .74931 L .63636 .76517 L .843 .48 0 r
F
.64882 .26053 m .6232 .27404 L .62295 .2735 L .868 .679 0 r
F
.64882 .26053 m .6554 .28175 L .6232 .27404 L F
.6554 .28175 m .62704 .28231 L .6232 .27404 L .87 .666 0 r
F
.63157 .83361 m .63296 .8401 L .62326 .83972 L .867 .634 0 r
F
.62326 .83972 m .63296 .8401 L .63083 .84321 L .838 .832 0 r
F
.6276 1.1085 m .62342 1.10813 L .6254 1.10421 L .77 .621 0 r
F
.62342 1.10813 m .6276 1.1085 L .62712 1.1165 L .765 .655 0 r
F
.62712 1.1165 m .62417 1.11724 L .62342 1.10813 L .799 .669 0 r
F
.62366 .95455 m .63237 .94973 L .63355 .9575 L .844 .687 0 r
F
.63355 .9575 m .62453 .95555 L .62366 .95455 L .854 .646 0 r
F
.63197 .84711 m .63338 .85193 L .62388 .85224 L .889 .658 0 r
F
.62433 .85551 m .62388 .85224 L .63338 .85193 L .889 .655 0 r
F
.63338 .85193 m .63336 .85361 L .62433 .85551 L .868 .6 0 r
F
.62433 .85551 m .63336 .85361 L .63475 .85746 L .859 .573 0 r
F
.63475 .85746 m .6262 .86042 L .62433 .85551 L .862 .596 0 r
F
.62435 1.16199 m .63357 1.14903 L .64261 1.1467 L .709 .967 0 r
F
.64174 .91451 m .62446 .91318 L .62592 .90724 L .856 .719 0 r
F
.62446 .91318 m .64174 .91451 L .64245 .93519 L .857 .707 0 r
F
.63728 .67234 m .64962 .68296 L .62451 .68559 L .826 .795 0 r
F
.64962 .68296 m .67711 .69245 L .62451 .68559 L .832 .674 0 r
F
.67711 .69245 m .63728 .69534 L .62451 .68559 L .808 .756 0 r
F
.62487 .95562 m .62465 .95569 L .62453 .95555 L .854 .646 0 r
F
.62525 .95638 m .62465 .95569 L .62487 .95562 L F
.63355 .9575 m .62589 .95764 L .62487 .95562 L F
.62496 1.12665 m .64427 1.11782 L .64135 1.13521 L .85 .926 0 r
F
.62496 1.12665 m .62621 1.12246 L .63772 1.12082 L .859 .903 0 r
F
.62502 1.09311 m .62742 1.0976 L .62551 1.09739 L .606 .095 0 r
F
.62502 1.09311 m .62901 1.08734 L .62742 1.0976 L .839 .825 0 r
F
.62538 1.08911 m .62901 1.08734 L .62502 1.09311 L F
.62525 .75818 m .62713 .74381 L .63636 .76517 L .97 .653 0 r
F
.63482 .79 m .62525 .75818 L .63636 .76517 L .985 .729 0 r
F
.62564 1.08875 m .62901 1.08734 L .62538 1.08911 L .839 .825 0 r
F
.6254 1.10421 m .62714 1.10401 L .6276 1.1085 L .847 .644 0 r
F
.6326 1.08584 m .62564 1.08875 L .64506 1.07082 L .818 .692 0 r
F
.62564 1.08875 m .62599 1.08826 L .64506 1.07082 L F
.62578 .89511 m .63845 .89407 L .64079 .8975 L .801 .453 0 r
F
.64079 .8975 m .62618 .89809 L .62578 .89511 L .803 .46 0 r
F
.62699 .95984 m .62589 .95764 L .63355 .9575 L .85 .623 0 r
F
.64079 .8975 m .64319 .90699 L .62592 .90724 L .857 .688 0 r
F
.62592 .90724 m .62618 .89809 L .64079 .8975 L .861 .699 0 r
F
.62592 .90724 m .64319 .90699 L .64174 .91451 L .857 .717 0 r
F
.62599 1.08826 m .62611 1.08809 L .64506 1.07082 L .818 .692 0 r
F
.62611 1.08809 m .62616 1.08802 L .64506 1.07082 L F
.62616 1.08802 m .62718 1.08658 L .64506 1.07082 L F
.63475 .85746 m .62976 .86107 L .6262 .86042 L .548 .06 0 r
F
.66561 .50618 m .67336 .55458 L .6264 .507 L .814 .678 0 r
F
.67336 .55458 m .63045 .55818 L .6264 .507 L .815 .678 0 r
F
.62701 .7378 m .62694 .73791 L .62689 .73781 L .811 .595 0 r
F
.63043 .82919 m .62692 .80368 L .63482 .79 L .932 .724 0 r
F
.62701 .7378 m .627 .73804 L .62694 .73791 L .811 .595 0 r
F
.62844 .96273 m .62699 .95984 L .63355 .9575 L .85 .623 0 r
F
.62701 .7378 m .62717 .73843 L .627 .73804 L .811 .595 0 r
F
.62701 .7378 m .62726 .73862 L .62723 .73856 L F
.62701 .7378 m .62738 .73889 L .62726 .73862 L F
.62701 .7378 m .62723 .73856 L .62722 .73854 L F
.63208 .74931 m .62738 .73889 L .62701 .7378 L F
.62893 .73766 m .63208 .74931 L .62701 .7378 L F
.62701 .7378 m .62722 .73854 L .62717 .73843 L F
.6554 .28175 m .63311 .29316 L .62704 .28231 L .87 .666 0 r
F
.63311 .29316 m .63206 .29309 L .62704 .28231 L F
.6301 1.11613 m .62712 1.1165 L .6276 1.1085 L .865 .742 0 r
F
.62714 1.10401 m .6296 1.10386 L .63084 1.10695 L .797 .69 0 r
F
.63084 1.10695 m .6276 1.1085 L .62714 1.10401 L .861 .654 0 r
F
.62718 1.08658 m .63029 1.08219 L .64506 1.07082 L .818 .692 0 r
F
.62742 1.0976 m .62901 1.08734 L .63021 1.10024 L .839 .825 0 r
F
.6276 1.1085 m .63084 1.10695 L .6301 1.11613 L .892 .752 0 r
F
.63877 .85703 m .63875 .86243 L .62828 .86548 L .881 .688 0 r
F
.62828 .86548 m .62976 .86107 L .63877 .85703 L .9 .734 0 r
F
.62896 .87199 m .62828 .86548 L .63875 .86243 L .874 .664 0 r
F
.63355 .9575 m .63137 .96856 L .62844 .96273 L .853 .627 0 r
F
.64336 .73661 m .63666 .76396 L .62893 .73766 L .811 .595 0 r
F
.63666 .76396 m .63208 .74931 L .62893 .73766 L F
.63875 .86243 m .6402 .86812 L .62896 .87199 L .915 .751 0 r
F
.63251 .88116 m .62896 .87199 L .6402 .86812 L .882 .631 0 r
F
.62901 1.08734 m .6326 1.08584 L .63067 1.10378 L .839 .825 0 r
F
.62901 1.08734 m .63067 1.10378 L .63021 1.10024 L F
.62942 1.122 m .63088 1.12107 L .63772 1.12082 L .859 .903 0 r
F
.63084 1.10695 m .6296 1.10386 L .63021 1.10024 L .974 .67 0 r
F
.63475 .85746 m .63877 .85703 L .62976 .86107 L .331 0 0 r
F
.62997 .94198 m .64245 .93519 L .6321 .94795 L .862 .631 0 r
F
.63137 1.11687 m .6301 1.11613 L .63084 1.10695 L .959 .815 0 r
F
.63084 1.10695 m .63021 1.10024 L .63144 1.10983 L .785 .571 0 r
F
.63029 1.08219 m .63132 1.08072 L .64506 1.07082 L .818 .692 0 r
F
.63482 .79 m .6427 .81223 L .63043 .82919 L .936 .729 0 r
F
.63964 .83576 m .63043 .82919 L .6427 .81223 L .927 .71 0 r
F
.63157 .83361 m .63043 .82919 L .63964 .83576 L .936 .615 0 r
F
.67336 .55458 m .67339 .58898 L .63045 .55818 L .815 .711 0 r
F
.67339 .58898 m .63098 .59085 L .63045 .55818 L .826 .705 0 r
F
.63059 .98336 m .63063 .98255 L .63312 .98277 L .877 .751 0 r
F
.63092 .98417 m .63059 .98336 L .63312 .98277 L F
.63312 .98277 m .63063 .98255 L .64097 .97776 L F
.63063 .98255 m .63071 .98104 L .64097 .97776 L F
.63067 1.10378 m .6326 1.08584 L .63076 1.10449 L .839 .825 0 r
F
.63071 .98104 m .63081 .97928 L .64097 .97776 L .877 .751 0 r
F
.63076 1.10449 m .6326 1.08584 L .63144 1.10488 L .839 .825 0 r
F
.63084 1.10512 m .63076 1.10449 L .63144 1.10488 L .845 .823 0 r
F
.63856 .9688 m .63081 .97928 L .63113 .97311 L .872 .695 0 r
F
.63856 .9688 m .64097 .97776 L .63081 .97928 L F
.63083 .84321 m .63296 .8401 L .63618 .84531 L .845 .843 0 r
F
.63083 .84321 m .63618 .84531 L .63197 .84711 L .917 .584 0 r
F
.63084 1.10695 m .63144 1.10983 L .63137 1.11687 L .863 .688 0 r
F
.63084 1.10512 m .63144 1.10488 L .63093 1.10579 L .839 .882 0 r
F
.63088 1.12107 m .63717 1.1194 L .63772 1.12082 L .859 .903 0 r
F
.6315 1.12068 m .63717 1.1194 L .63088 1.12107 L F
.64097 .97776 m .63413 .99217 L .63092 .98417 L .858 .698 0 r
F
.63093 1.10579 m .63144 1.10488 L .63094 1.10594 L .839 .882 0 r
F
.63094 1.10594 m .63144 1.10488 L .63099 1.10628 L F
.67339 .58898 m .67872 .61678 L .63098 .59085 L .825 .721 0 r
F
.67872 .61678 m .63288 .62088 L .63098 .59085 L .828 .719 0 r
F
.63099 1.10628 m .63144 1.10488 L .631 1.10635 L .839 .882 0 r
F
.631 1.10635 m .63144 1.10488 L .631 1.10635 L F
.631 1.10635 m .63144 1.10488 L .631 1.10637 L .859 .903 0 r
F
.631 1.10637 m .63144 1.10488 L .63102 1.10655 L F
.63102 1.10655 m .63144 1.10488 L .63112 1.1073 L F
.63112 1.1073 m .63144 1.10488 L .63113 1.10742 L F
.63113 .97311 m .63128 .9703 L .63856 .9688 L .864 .675 0 r
F
.63113 1.10742 m .63144 1.10488 L .63114 1.10743 L .859 .903 0 r
F
.63114 1.10743 m .63144 1.10488 L .63141 1.10956 L F
.63133 .96942 m .63856 .9688 L .63128 .9703 L .853 .627 0 r
F
.64506 1.07082 m .63132 1.08072 L .63559 1.07469 L .872 .883 0 r
F
.63135 .96888 m .63856 .9688 L .63133 .96942 L .853 .627 0 r
F
.63831 .96823 m .63856 .9688 L .63135 .96888 L F
.63831 .96823 m .63135 .96888 L .63136 .96875 L F
.63831 .96823 m .63136 .96875 L .63137 .96856 L F
.63137 1.11687 m .63149 1.1168 L .6315 1.12068 L .859 .903 0 r
F
.63137 1.11687 m .63143 1.11139 L .63149 1.1168 L F
.63355 .9575 m .63831 .96823 L .63137 .96856 L .853 .627 0 r
F
.63141 1.10956 m .63144 1.10488 L .63144 1.10983 L .859 .903 0 r
F
.63144 1.10983 m .63146 1.1099 L .63143 1.11139 L F
.63143 1.11139 m .63146 1.1099 L .63149 1.1168 L F
.63144 1.10488 m .6326 1.08584 L .64964 1.09659 L .873 .853 0 r
F
.63144 1.10488 m .64964 1.09659 L .64427 1.11782 L .865 .903 0 r
F
.63144 1.10983 m .63144 1.10488 L .63146 1.1099 L .859 .903 0 r
F
.6315 1.12068 m .63144 1.10488 L .63717 1.1194 L F
.63772 1.12082 m .63144 1.10488 L .64427 1.11782 L F
.64825 .21626 m .65879 .22697 L .63157 .23779 L .864 .642 0 r
F
.65879 .22697 m .63822 .24596 L .63157 .23779 L .87 .663 0 r
F
.63296 .8401 m .63157 .83361 L .63767 .83821 L .958 .683 0 r
F
.63964 .83576 m .63767 .83821 L .63157 .83361 L .954 .727 0 r
F
.63618 .84531 m .63807 .8501 L .63197 .84711 L .945 .654 0 r
F
.63197 .84711 m .63807 .8501 L .63338 .85193 L .948 .68 0 r
F
.63311 .29316 m .63216 .29365 L .63206 .29309 L .87 .666 0 r
F
.64425 .78824 m .63636 .76517 L .63208 .74931 L .759 .443 0 r
F
.6321 .94795 m .64245 .93519 L .63355 .9575 L .919 .733 0 r
F
.6554 .28175 m .63231 .29454 L .63216 .29365 L .869 .663 0 r
F
.6554 .28175 m .63245 .29536 L .63231 .29454 L F
.6554 .28175 m .66066 .29409 L .63245 .29536 L F
.66066 .29409 m .63263 .29639 L .63245 .29536 L .871 .681 0 r
F
.6402 .86812 m .64495 .87716 L .63251 .88116 L .878 .625 0 r
F
.63251 .88116 m .64495 .87716 L .64778 .88627 L .888 .653 0 r
F
.64778 .88627 m .63531 .89105 L .63251 .88116 L .888 .618 0 r
F
.65059 1.0759 m .6326 1.08584 L .64506 1.07082 L .842 .735 0 r
F
.6326 1.08584 m .65059 1.0759 L .64964 1.09659 L .864 .855 0 r
F
.66066 .29409 m .63303 .29872 L .63263 .29639 L .871 .681 0 r
F
.67872 .61678 m .68425 .64098 L .63288 .62088 L .828 .731 0 r
F
.68425 .64098 m .64022 .64415 L .63288 .62088 L .812 .747 0 r
F
.63767 .83821 m .63787 .84256 L .63296 .8401 L .97 .726 0 r
F
.63296 .8401 m .63787 .84256 L .63618 .84531 L .936 .848 0 r
F
.66066 .29409 m .66808 .32306 L .63303 .29872 L .871 .681 0 r
F
.66808 .32306 m .63695 .32148 L .63303 .29872 L .867 .686 0 r
F
.63336 .85361 m .63807 .8501 L .63846 .8531 L .809 .417 0 r
F
.63338 .85193 m .63807 .8501 L .63336 .85361 L .945 .673 0 r
F
.63846 .8531 m .63475 .85746 L .63336 .85361 L .876 .58 0 r
F
.64245 .93519 m .65031 .96707 L .63355 .9575 L .886 .688 0 r
F
.63355 .9575 m .65031 .96707 L .63856 .9688 L .901 .638 0 r
F
.63357 1.14903 m .64135 1.13521 L .65075 1.13574 L .806 .999 0 r
F
.63357 1.14903 m .65075 1.13574 L .64261 1.1467 L .682 .961 0 r
F
.64097 .97776 m .6441 .99142 L .63413 .99217 L .858 .698 0 r
F
.63574 .9962 m .63413 .99217 L .6441 .99142 L .859 .717 0 r
F
.63846 .8531 m .63877 .85703 L .63475 .85746 L .481 .066 0 r
F
.63636 .76517 m .64599 .79561 L .63482 .79 L .96 .695 0 r
F
.63482 .79 m .64599 .79561 L .6427 .81223 L .96 .731 0 r
F
.63531 .89105 m .64778 .88627 L .65043 .89237 L .926 .741 0 r
F
.65043 .89237 m .63845 .89407 L .63531 .89105 L .909 .612 0 r
F
.63559 1.07469 m .63603 1.07407 L .64506 1.07082 L .634 .964 0 r
F
.6441 .99142 m .63833 1.00266 L .63574 .9962 L .866 .737 0 r
F
.63632 1.07366 m .63774 1.07345 L .63603 1.07407 L .882 .926 0 r
F
.63618 .84531 m .63787 .84256 L .63982 .84971 L .887 .756 0 r
F
.63982 .84971 m .63807 .8501 L .63618 .84531 L .962 .645 0 r
F
.63632 1.07366 m .63774 1.07345 L .63632 1.07366 L .882 .926 0 r
F
.63632 1.07366 m .63632 1.07366 L .63632 1.07366 L F
.63728 1.07159 m .63774 1.07345 L .63632 1.07366 L F
.63636 .76517 m .64425 .78824 L .64599 .79561 L .81 .493 0 r
F
.64336 .73661 m .63804 .76838 L .63666 .76396 L .811 .595 0 r
F
.64061 .33267 m .63887 .33266 L .63695 .32148 L .867 .686 0 r
F
.66808 .32306 m .64061 .33267 L .63695 .32148 L F
.67711 .69245 m .68044 .73448 L .63728 .69534 L .804 .598 0 r
F
.68044 .73448 m .64336 .73661 L .63728 .69534 L .808 .596 0 r
F
.6913 .66627 m .63728 .67234 L .64022 .64415 L .823 .742 0 r
F
.6913 .66627 m .67558 .67994 L .63728 .67234 L .817 .785 0 r
F
.63728 .67234 m .67558 .67994 L .64962 .68296 L .807 .796 0 r
F
.63886 1.06819 m .64506 1.07082 L .63728 1.07159 L .882 .926 0 r
F
.63728 1.07159 m .64506 1.07082 L .63774 1.07345 L F
.63767 .83821 m .63964 .83576 L .64765 .84421 L .966 .757 0 r
F
.63787 .84256 m .63767 .83821 L .64765 .84895 L .978 .733 0 r
F
.64765 .84421 m .64765 .84895 L .63767 .83821 L .958 .787 0 r
F
.63787 .84256 m .64465 .84821 L .63982 .84971 L .906 .764 0 r
F
.64765 .84895 m .64465 .84821 L .63787 .84256 L .817 .343 0 r
F
.64336 .73661 m .63863 .77028 L .63804 .76838 L .811 .595 0 r
F
.63846 .8531 m .63807 .8501 L .63982 .84971 L .904 .506 0 r
F
.65879 .22697 m .67582 .24835 L .63822 .24596 L .852 .627 0 r
F
.67582 .24835 m .64882 .26053 L .63822 .24596 L .854 .651 0 r
F
.6441 .99142 m .64606 1.00171 L .63833 1.00266 L .866 .737 0 r
F
.63857 1.00326 m .63833 1.00266 L .64566 1.00176 L .866 .744 0 r
F
.64079 .8975 m .63845 .89407 L .65043 .89237 L .838 .444 0 r
F
.63982 .84971 m .64342 .85432 L .63846 .8531 L .595 .126 0 r
F
.63846 .8531 m .64342 .85432 L .63877 .85703 L .609 .162 0 r
F
.63856 .9688 m .65031 .96707 L .64097 .97776 L .918 .716 0 r
F
.63857 1.00326 m .64566 1.00176 L .63858 1.00328 L .866 .744 0 r
F
.63858 1.00328 m .64566 1.00176 L .64606 1.00171 L F
.63999 1.01054 m .63858 1.00328 L .64606 1.00171 L F
.64336 .73661 m .6479 .77927 L .63863 .77028 L .811 .595 0 r
F
.6479 .77927 m .64168 .78004 L .63863 .77028 L .817 .593 0 r
F
.64342 .85432 m .64227 .85783 L .63875 .86243 L .876 .925 0 r
F
.63875 .86243 m .63877 .85703 L .64342 .85432 L .931 .731 0 r
F
.63875 .86243 m .64227 .85783 L .6402 .86812 L .991 .764 0 r
F
.64256 1.06019 m .64506 1.07082 L .63886 1.06819 L .913 .911 0 r
F
.64061 .33267 m .63897 .33325 L .63887 .33266 L .867 .686 0 r
F
.66808 .32306 m .63913 .3342 L .63897 .33325 L .852 .637 0 r
F
.65703 .18259 m .6686 .19525 L .63898 .204 L .851 .643 0 r
F
.6686 .19525 m .64825 .21626 L .63898 .204 L .86 .677 0 r
F
.66808 .32306 m .67266 .33478 L .63913 .3342 L .852 .637 0 r
F
.67266 .33478 m .63962 .33704 L .63913 .3342 L .855 .67 0 r
F
.67266 .33478 m .68093 .37394 L .63962 .33704 L F
.68093 .37394 m .64564 .37213 L .63962 .33704 L .863 .665 0 r
F
.6427 .81223 m .64462 .83094 L .63964 .83576 L .998 .801 0 r
F
.63964 .83576 m .64462 .83094 L .64765 .84421 L .975 .707 0 r
F
.63982 .84971 m .64465 .84821 L .64342 .85432 L .601 .127 0 r
F
.64606 1.00171 m .64088 1.01514 L .63999 1.01054 L .846 .711 0 r
F
.64227 .85783 m .64645 .86075 L .6402 .86812 L .988 .758 0 r
F
.6531 .86958 m .6402 .86812 L .64645 .86075 L .916 .59 0 r
F
.64495 .87716 m .6402 .86812 L .6531 .86958 L .925 .621 0 r
F
.68425 .64098 m .6913 .66627 L .64022 .64415 L .811 .752 0 r
F
.65043 .89237 m .65324 .89269 L .64079 .8975 L .297 0 0 r
F
.64319 .90699 m .64079 .8975 L .65324 .89269 L .921 .72 0 r
F
.64606 1.00171 m .64872 1.01255 L .64088 1.01514 L .846 .711 0 r
F
.64202 1.02106 m .64088 1.01514 L .64872 1.01255 L .85 .737 0 r
F
.64097 .97776 m .65031 .96707 L .6441 .99142 L .926 .733 0 r
F
.64135 1.13521 m .64427 1.11782 L .65767 1.123 L .901 .964 0 r
F
.64135 1.13521 m .65767 1.123 L .65075 1.13574 L .8 .999 0 r
F
.64229 .77996 m .64174 .78022 L .64168 .78004 L .818 .598 0 r
F
.65776 .91271 m .64174 .91451 L .64319 .90699 L .905 .768 0 r
F
.64245 .93519 m .64174 .91451 L .65776 .91271 L .904 .744 0 r
F
.64229 .77996 m .64176 .78027 L .64174 .78022 L .818 .598 0 r
F
.64229 .77996 m .64425 .78824 L .64176 .78027 L F
.64486 1.02291 m .64202 1.02106 L .64872 1.01255 L .85 .737 0 r
F
.64248 1.02339 m .64202 1.02106 L .64486 1.02291 L F
.64483 .85318 m .64645 .86075 L .64227 .85783 L .978 .685 0 r
F
.64483 .85318 m .64227 .85783 L .64342 .85432 L .796 .849 0 r
F
.6479 .77927 m .64496 .79127 L .64229 .77996 L .818 .598 0 r
F
.65776 .91271 m .65775 .93273 L .64245 .93519 L .913 .76 0 r
F
.64245 .93519 m .65775 .93273 L .66806 .95646 L .911 .736 0 r
F
.64245 .93519 m .66051 .96131 L .65031 .96707 L .905 .697 0 r
F
.66806 .95646 m .66051 .96131 L .64245 .93519 L .94 .678 0 r
F
.64251 1.024 m .64248 1.02339 L .64486 1.02291 L .85 .737 0 r
F
.64258 1.0252 m .64251 1.024 L .64391 1.02545 L .903 .49 0 r
F
.64391 1.02545 m .64251 1.024 L .64486 1.02291 L .85 .737 0 r
F
.64293 1.0594 m .64506 1.07082 L .64256 1.06019 L .913 .911 0 r
F
.64259 1.02536 m .64258 1.0252 L .64391 1.02545 L .903 .49 0 r
F
.64391 1.02545 m .6429 1.03064 L .64259 1.02536 L .922 .53 0 r
F
.64261 1.1467 m .65075 1.13574 L .6508 1.13601 L .577 .82 0 r
F
.64462 .83094 m .6427 .81223 L .64599 .79561 L .999 .801 0 r
F
.64391 1.02545 m .64516 1.03326 L .6429 1.03064 L .922 .53 0 r
F
.64299 1.03224 m .6429 1.03064 L .64516 1.03326 L .914 .717 0 r
F
.6442 1.05789 m .64293 1.0594 L .64351 1.05816 L .913 .911 0 r
F
.6442 1.05789 m .64506 1.07082 L .64293 1.0594 L F
.64305 1.03311 m .64299 1.03224 L .64516 1.03326 L .914 .717 0 r
F
.64516 1.03326 m .64327 1.03687 L .64305 1.03311 L .906 .648 0 r
F
.65324 .89269 m .65683 .90203 L .64319 .90699 L .915 .708 0 r
F
.64319 .90699 m .65683 .90203 L .65776 .91271 L .919 .725 0 r
F
.64516 1.03326 m .64336 1.03853 L .64327 1.03687 L .974 .769 0 r
F
.68044 .73448 m .68179 .7761 L .64336 .73661 L .806 .585 0 r
F
.68179 .7761 m .6479 .77927 L .64336 .73661 L .8 .587 0 r
F
.64516 1.03326 m .64712 1.03712 L .64336 1.03853 L .974 .769 0 r
F
.64712 1.03712 m .64343 1.03973 L .64336 1.03853 L .97 .905 0 r
F
.64342 .85432 m .64465 .84821 L .64483 .85318 L .9 .683 0 r
F
.64712 1.03712 m .64516 1.04073 L .64343 1.03973 L .97 .905 0 r
F
.64516 1.04073 m .64351 1.04103 L .64343 1.03973 L F
.6442 1.05789 m .64351 1.05816 L .64359 1.05798 L .913 .911 0 r
F
.64516 1.04073 m .64351 1.04143 L .64351 1.04103 L .97 .905 0 r
F
.64516 1.04073 m .64352 1.04373 L .64351 1.04143 L F
.64352 1.04373 m .64712 1.03712 L .64354 1.04672 L .964 .881 0 r
F
.64354 1.04672 m .64712 1.03712 L .64646 1.04569 L F
.64646 1.04569 m .64356 1.05128 L .64354 1.04672 L .936 .947 0 r
F
.64356 1.05128 m .64446 1.04953 L .64356 1.05215 L .885 .868 0 r
F
.64356 1.05215 m .64446 1.04953 L .65196 1.04303 L F
.65196 1.04303 m .64358 1.05632 L .64356 1.05215 L .869 .823 0 r
F
.64358 1.05664 m .64358 1.05632 L .64409 1.05621 L .869 .817 0 r
F
.65196 1.04303 m .64409 1.05621 L .64358 1.05632 L .869 .823 0 r
F
.64409 1.05621 m .64358 1.0569 L .64358 1.05664 L .862 .796 0 r
F
.64409 1.05621 m .64359 1.05798 L .64358 1.0569 L .913 .911 0 r
F
.64409 1.05621 m .6442 1.05789 L .64359 1.05798 L F
.64872 1.01255 m .65019 1.02336 L .64391 1.02545 L .836 .719 0 r
F
.64684 1.03655 m .64391 1.02545 L .65019 1.02336 L .838 .732 0 r
F
.64516 1.03326 m .64391 1.02545 L .64684 1.03655 L .8 .521 0 r
F
.64409 1.05621 m .65196 1.04303 L .66455 1.03462 L .925 .947 0 r
F
.66455 1.03462 m .64506 1.07082 L .64409 1.05621 L .927 .918 0 r
F
.65031 .96707 m .6592 1.00229 L .6441 .99142 L .936 .745 0 r
F
.6441 .99142 m .6592 1.00229 L .64606 1.00171 L .922 .767 0 r
F
.64427 1.11782 m .64964 1.09659 L .66067 1.10612 L .907 .945 0 r
F
.64427 1.11782 m .66067 1.10612 L .65767 1.123 L .896 .966 0 r
F
.64446 1.04953 m .64682 1.045 L .65196 1.04303 L .885 .868 0 r
F
.64599 .79561 m .64859 .82426 L .64462 .83094 L .996 .797 0 r
F
.64765 .84421 m .64462 .83094 L .64859 .82426 L .965 .705 0 r
F
.64465 .84821 m .64765 .84895 L .64483 .85318 L .989 .729 0 r
F
.64483 .85318 m .64765 .84895 L .64951 .8538 L .951 .605 0 r
F
.64951 .8538 m .64645 .86075 L .64483 .85318 L .976 .685 0 r
F
.64495 .87716 m .6531 .86958 L .65483 .87697 L .994 .84 0 r
F
.65483 .87697 m .64778 .88627 L .64495 .87716 L .972 .671 0 r
F
.6479 .77927 m .65061 .79288 L .64496 .79127 L .818 .598 0 r
F
.64675 .79178 m .64599 .79561 L .64496 .79127 L F
.66455 1.03462 m .66616 1.06357 L .64506 1.07082 L .879 .828 0 r
F
.66616 1.06357 m .65059 1.0759 L .64506 1.07082 L .869 .726 0 r
F
.64684 1.03655 m .64712 1.03712 L .64516 1.03326 L .631 .254 0 r
F
.68093 .37394 m .64581 .37306 L .64564 .37213 L .863 .665 0 r
F
.68093 .37394 m .64596 .37388 L .64581 .37306 L .864 .685 0 r
F
.68093 .37394 m .68788 .41652 L .64596 .37388 L F
.64968 .37766 m .6507 .39969 L .64596 .37388 L .865 .685 0 r
F
.64613 .79715 m .64599 .79561 L .64675 .79178 L .818 .598 0 r
F
.64606 1.00171 m .6592 1.00229 L .64872 1.01255 L .922 .748 0 r
F
.64633 .79855 m .64613 .79715 L .64675 .79178 L .818 .598 0 r
F
.64619 .79784 m .64613 .79715 L .64633 .79855 L F
.64619 .79784 m .6493 .81322 L .64859 .82426 L .941 .696 0 r
F
.64641 .79891 m .64633 .79855 L .64675 .79178 L .818 .598 0 r
F
.65061 .79288 m .64681 .80089 L .64641 .79891 L F
.65061 .79288 m .64641 .79891 L .64675 .79178 L F
.6531 .86958 m .64645 .86075 L .64951 .8538 L .92 .579 0 r
F
.64646 1.04569 m .6465 1.04512 L .64682 1.045 L .885 .868 0 r
F
.65144 1.03288 m .65196 1.04303 L .6465 1.04512 L .885 .78 0 r
F
.65144 1.03288 m .6465 1.04512 L .64661 1.04377 L F
.64661 1.04377 m .647 1.03866 L .64841 1.03971 L .89 .788 0 r
F
.65061 .79288 m .64757 .80469 L .64681 .80089 L .818 .598 0 r
F
.65019 1.02336 m .65144 1.03288 L .64684 1.03655 L .891 .79 0 r
F
.64841 1.03971 m .64684 1.03655 L .65144 1.03288 L .89 .788 0 r
F
.647 1.03866 m .64708 1.03772 L .64841 1.03971 L F
.64708 1.03772 m .64712 1.03712 L .64841 1.03971 L F
.65061 .79288 m .6493 .81322 L .64757 .80469 L .818 .598 0 r
F
.65186 .85045 m .64765 .84895 L .64765 .84421 L 1 .82 0 r
F
.64951 .8538 m .64765 .84895 L .65186 .85045 L .94 .577 0 r
F
.64859 .82426 m .64886 .84386 L .64765 .84421 L .844 .691 0 r
F
.64765 .84421 m .64886 .84386 L .64903 .84626 L .845 .645 0 r
F
.64778 .88627 m .65483 .87697 L .65771 .87982 L .952 .622 0 r
F
.65771 .87982 m .65043 .89237 L .64778 .88627 L .983 .714 0 r
F
.68179 .7761 m .6851 .83152 L .6479 .77927 L .801 .592 0 r
F
.6851 .83152 m .65885 .83428 L .6479 .77927 L .81 .592 0 r
F
.6686 .19525 m .68211 .20906 L .64825 .21626 L .85 .659 0 r
F
.68211 .20906 m .65879 .22697 L .64825 .21626 L .848 .65 0 r
F
.64859 .82426 m .6493 .81322 L .65077 .8378 L .857 .678 0 r
F
.65077 .8378 m .64886 .84386 L .64859 .82426 L .901 .742 0 r
F
.6592 1.00229 m .66455 1.03462 L .64872 1.01255 L .94 .795 0 r
F
.64872 1.01255 m .66455 1.03462 L .65019 1.02336 L .951 .789 0 r
F
.67582 .24835 m .68561 .269 L .64882 .26053 L .856 .657 0 r
F
.68561 .269 m .6554 .28175 L .64882 .26053 L .854 .673 0 r
F
.64886 .84386 m .65077 .8378 L .65266 .84702 L .951 .695 0 r
F
.64956 .84704 m .64903 .84626 L .64886 .84386 L .966 .743 0 r
F
.65266 .84702 m .64956 .84704 L .64886 .84386 L F
.65061 .79288 m .64991 .82349 L .6493 .81322 L .818 .598 0 r
F
.65186 .85045 m .65204 .8575 L .64951 .8538 L .961 .739 0 r
F
.64951 .8538 m .6526 .86153 L .6531 .86958 L .917 .578 0 r
F
.64951 .8538 m .65204 .8575 L .6526 .86153 L .828 .481 0 r
F
.64956 .84704 m .65266 .84702 L .65 .8477 L .864 .444 0 r
F
.67558 .67994 m .71626 .69156 L .64962 .68296 L .816 .683 0 r
F
.71626 .69156 m .67711 .69245 L .64962 .68296 L .808 .715 0 r
F
.64964 1.09659 m .65059 1.0759 L .66516 1.08602 L .917 .894 0 r
F
.64964 1.09659 m .66516 1.08602 L .66067 1.10612 L .907 .945 0 r
F
.68788 .41652 m .65367 .41646 L .64968 .37766 L .865 .685 0 r
F
.65353 .41513 m .6507 .39969 L .64968 .37766 L F
.65061 .79288 m .65885 .83428 L .64991 .82349 L .818 .598 0 r
F
.65885 .83428 m .64996 .82424 L .64991 .82349 L .817 .598 0 r
F
.65885 .83428 m .65061 .83518 L .64996 .82424 L F
.65 .8477 m .65266 .84702 L .65497 .8513 L .864 .444 0 r
F
.65497 .8513 m .65186 .85045 L .65 .8477 L .683 .181 0 r
F
.65019 1.02336 m .66455 1.03462 L .65144 1.03288 L .931 .813 0 r
F
.65031 .96707 m .66051 .96131 L .66792 .99611 L .919 .738 0 r
F
.6592 1.00229 m .65031 .96707 L .66792 .99611 L F
.65043 .89237 m .65771 .87982 L .66076 .88154 L .71 .295 0 r
F
.66076 .88154 m .65324 .89269 L .65043 .89237 L .362 0 0 r
F
.65059 1.0759 m .66616 1.06357 L .66516 1.08602 L .913 .896 0 r
F
.65885 .83428 m .65077 .8378 L .65061 .83518 L .815 .591 0 r
F
.65075 1.13574 m .65767 1.123 L .65945 1.12634 L .764 .956 0 r
F
.65075 1.13574 m .65945 1.12634 L .6508 1.13601 L .536 .828 0 r
F
.65885 .83428 m .65134 .8406 L .65077 .8378 L .815 .591 0 r
F
.6508 1.13601 m .65945 1.12634 L .65942 1.1267 L .536 .828 0 r
F
.65885 .83428 m .65242 .84586 L .65134 .8406 L .815 .591 0 r
F
.65144 1.03288 m .66455 1.03462 L .65196 1.04303 L .932 .812 0 r
F
.65204 .8575 m .65186 .85045 L .65497 .8513 L .935 .722 0 r
F
.65497 .8513 m .65311 .85951 L .65204 .8575 L .807 .791 0 r
F
.65204 .8575 m .65311 .85951 L .6526 .86153 L .748 .496 0 r
F
.65881 .83436 m .65266 .84702 L .65242 .84586 L .815 .591 0 r
F
.6526 .86153 m .65311 .85951 L .65567 .86633 L .733 .396 0 r
F
.65567 .86633 m .65525 .87003 L .6526 .86153 L .777 .48 0 r
F
.6526 .86153 m .65525 .87003 L .65483 .87697 L .926 .644 0 r
F
.65483 .87697 m .6531 .86958 L .6526 .86153 L .829 .578 0 r
F
.65881 .83436 m .65594 .84701 L .65266 .84702 L .815 .591 0 r
F
.65594 .84701 m .65497 .8513 L .65266 .84702 L F
.65312 .85949 m .65394 .86172 L .65311 .85951 L .843 .574 0 r
F
.65312 .85948 m .65312 .85949 L .65311 .85951 L F
.67784 .86619 m .65312 .85948 L .65312 .85948 L .819 .584 0 r
F
.67784 .86619 m .66062 .86773 L .65312 .85948 L F
.65312 .85949 m .66062 .86773 L .65394 .86172 L .843 .574 0 r
F
.67784 .86619 m .65312 .85948 L .65313 .85945 L .819 .584 0 r
F
.67784 .86619 m .65313 .85945 L .65325 .85891 L F
.65683 .90203 m .65324 .89269 L .66076 .88154 L .979 .693 0 r
F
.67784 .86619 m .65325 .85891 L .6535 .85781 L .819 .584 0 r
F
.67784 .86619 m .6535 .85781 L .65495 .85142 L F
.68788 .41652 m .65396 .41929 L .65367 .41646 L .865 .703 0 r
F
.65394 .86172 m .66062 .86773 L .6549 .86427 L .843 .574 0 r
F
.68788 .41652 m .6911 .46155 L .65396 .41929 L .865 .703 0 r
F
.6911 .46155 m .65867 .46507 L .65396 .41929 L .875 .701 0 r
F
.65483 .87697 m .65525 .87003 L .65615 .87349 L .92 .647 0 r
F
.65483 .87697 m .65615 .87349 L .65586 .878 L F
.65584 .86484 m .6552 .86507 L .6549 .86427 L .823 .616 0 r
F
.65885 .83428 m .65495 .85142 L .65881 .83436 L .815 .591 0 r
F
.65885 .83428 m .67784 .86619 L .65495 .85142 L F
.6555 .86587 m .6552 .86507 L .65584 .86484 L .823 .616 0 r
F
.65525 .87003 m .65567 .86633 L .65681 .87581 L .764 .553 0 r
F
.65781 .87988 m .65781 .87988 L .65525 .87003 L .69 .419 0 r
F
.65681 .87581 m .65781 .87988 L .65525 .87003 L F
.68561 .269 m .69386 .29211 L .6554 .28175 L .856 .678 0 r
F
.69386 .29211 m .66066 .29409 L .6554 .28175 L .859 .66 0 r
F
.65567 .86633 m .6555 .86587 L .65584 .86484 L .823 .616 0 r
F
.65589 .86815 m .65567 .86633 L .65584 .86484 L F
.66062 .86773 m .65589 .86815 L .65584 .86484 L F
.65586 .878 m .65615 .87349 L .65638 .87851 L .92 .647 0 r
F
.66062 .86773 m .65664 .87438 L .65589 .86815 L .823 .617 0 r
F
.65638 .87851 m .65615 .87349 L .65776 .87972 L .92 .647 0 r
F
.65638 .87851 m .65776 .87972 L .65771 .87982 L F
.65666 .87454 m .65664 .87438 L .65693 .8739 L .827 .624 0 r
F
.65667 .87461 m .65666 .87454 L .65693 .8739 L F
.65677 .87549 m .65667 .87461 L .65693 .8739 L F
.65691 .87593 m .65677 .87549 L .65693 .8739 L F
.65677 .87551 m .65677 .87549 L .65691 .87593 L F
.65681 .87581 m .65677 .87551 L .6569 .87592 L F
.6569 .87592 m .65677 .87551 L .65691 .87593 L F
.6587 .88038 m .65781 .87988 L .65681 .87581 L .634 .078 0 r
F
.66061 .88087 m .6587 .88038 L .65681 .87581 L F
.65681 .87581 m .65913 .87853 L .66237 .8832 L .594 .107 0 r
F
.66076 .88154 m .66314 .89282 L .65683 .90203 L .988 .747 0 r
F
.65683 .90203 m .66314 .89282 L .66266 .90125 L .971 .821 0 r
F
.66266 .90125 m .65776 .91271 L .65683 .90203 L .971 .746 0 r
F
.65722 .87629 m .65691 .87593 L .65693 .8739 L .827 .624 0 r
F
.65949 .87905 m .65693 .8739 L .66062 .86773 L F
.65727 .87636 m .65722 .87629 L .65693 .8739 L F
.65913 .87853 m .65693 .8739 L .65929 .87876 L F
.65929 .87876 m .65693 .8739 L .65949 .87905 L F
.65866 .87798 m .65727 .87636 L .65693 .8739 L F
.65913 .87853 m .65866 .87798 L .65693 .8739 L F
.65767 1.123 m .66067 1.10612 L .665 1.11471 L .924 .973 0 r
F
.65767 1.123 m .665 1.11471 L .65945 1.12634 L .774 .951 0 r
F
.65771 .87982 m .65776 .87972 L .65781 .87988 L .92 .647 0 r
F
.66266 .90125 m .66498 .92966 L .65775 .93273 L .992 .776 0 r
F
.65775 .93273 m .65776 .91271 L .66266 .90125 L .947 .78 0 r
F
.65775 .93273 m .66498 .92966 L .67518 .95386 L .972 .664 0 r
F
.67518 .95386 m .67389 .95575 L .65775 .93273 L .889 .5 0 r
F
.65775 .93273 m .67389 .95575 L .66806 .95646 L .983 .709 0 r
F
.6911 .46155 m .65869 .46528 L .65867 .46507 L .877 .725 0 r
F
.6911 .46155 m .6587 .46536 L .65869 .46528 L F
.66076 .88154 m .6587 .88038 L .66061 .88087 L .634 .078 0 r
F
.6911 .46155 m .69129 .46604 L .6587 .46536 L .877 .725 0 r
F
.69129 .46604 m .65878 .46609 L .6587 .46536 L F
.69129 .46604 m .6591 .46799 L .65878 .46609 L F
.68211 .20906 m .70418 .23874 L .65879 .22697 L .842 .637 0 r
F
.70418 .23874 m .67582 .24835 L .65879 .22697 L .843 .629 0 r
F
.6851 .83152 m .69606 .86426 L .65885 .83428 L .808 .587 0 r
F
.69606 .86426 m .67784 .86619 L .65885 .83428 L .806 .589 0 r
F
.6928 .50264 m .65923 .46877 L .6591 .46799 L .886 .721 0 r
F
.69129 .46604 m .6928 .50264 L .6591 .46799 L .877 .725 0 r
F
.6592 1.00229 m .66792 .99611 L .67298 1.02672 L .937 .799 0 r
F
.66455 1.03462 m .6592 1.00229 L .67298 1.02672 L .945 .797 0 r
F
.6928 .50264 m .66518 .50366 L .65923 .46877 L .886 .721 0 r
F
.65945 1.12634 m .66649 1.11907 L .65942 1.1267 L .527 .829 0 r
F
.65942 1.1267 m .66649 1.11907 L .66938 1.11921 L .574 .903 0 r
F
.65945 1.12634 m .665 1.11471 L .66649 1.11907 L .782 .962 0 r
F
.65953 .8791 m .65949 .87905 L .66062 .86773 L .827 .624 0 r
F
.65958 .87917 m .65953 .8791 L .66062 .86773 L F
.67784 .86619 m .65958 .87917 L .66062 .86773 L F
.66055 .88058 m .65958 .87917 L .66017 .87875 L .818 .605 0 r
F
.66055 .88058 m .66017 .87875 L .6606 .88066 L F
.6606 .88066 m .66017 .87875 L .66067 .88075 L F
.66067 .88075 m .66017 .87875 L .6607 .88079 L F
.6607 .88079 m .66017 .87875 L .66085 .88101 L F
.66085 .88101 m .66017 .87875 L .66053 .8785 L F
.66051 .96131 m .66806 .95646 L .67563 .98935 L .961 .739 0 r
F
.66792 .99611 m .66051 .96131 L .67563 .98935 L .939 .748 0 r
F
.66321 .88535 m .66053 .8785 L .67784 .86619 L .818 .605 0 r
F
.66237 .8832 m .66085 .88101 L .66053 .8785 L F
.66237 .8832 m .66076 .88154 L .66061 .88087 L .634 .078 0 r
F
.69386 .29211 m .70295 .31586 L .66066 .29409 L .856 .64 0 r
F
.70295 .31586 m .66808 .32306 L .66066 .29409 L .842 .666 0 r
F
.66067 1.10612 m .66516 1.08602 L .66919 1.09852 L .925 .961 0 r
F
.66067 1.10612 m .66919 1.09852 L .665 1.11471 L .892 .989 0 r
F
.66314 .89282 m .66076 .88154 L .66237 .8832 L .845 .587 0 r
F
.66237 .8832 m .66378 .88681 L .66314 .89282 L .938 .602 0 r
F
.66487 .90095 m .66266 .90125 L .66306 .89435 L .813 .6 0 r
F
.66268 .90149 m .66266 .90125 L .68264 .89857 L F
.66268 .9015 m .66268 .90149 L .68264 .89857 L F
.66268 .90152 m .66268 .9015 L .68264 .89857 L F
.69307 .90241 m .66268 .90152 L .68264 .89857 L F
.69307 .90241 m .66278 .9027 L .66268 .90152 L F
.69307 .90241 m .66285 .90354 L .66278 .9027 L F
.69307 .90241 m .66291 .90427 L .66285 .90354 L F
.69307 .90241 m .66301 .90548 L .66291 .90427 L F
.69307 .90241 m .66353 .90622 L .66301 .90548 L F
.66301 .90548 m .66353 .90622 L .66302 .90563 L .846 .599 0 r
F
.66302 .90563 m .66353 .90622 L .66305 .90606 L F
.66353 .90622 m .66307 .90623 L .66305 .90606 L .832 .665 0 r
F
.66487 .90095 m .66306 .89435 L .66314 .89282 L .813 .6 0 r
F
.66353 .90622 m .66309 .90657 L .66307 .90623 L .823 .591 0 r
F
.66449 .92359 m .66309 .90657 L .66353 .90622 L .827 .599 0 r
F
.68264 .89857 m .66314 .89282 L .66329 .89143 L .813 .6 0 r
F
.68264 .89857 m .66487 .90095 L .66314 .89282 L F
.68913 .88348 m .66321 .88535 L .67784 .86619 L .818 .605 0 r
F
.68913 .88348 m .66371 .88664 L .66321 .88535 L .815 .589 0 r
F
.68913 .88348 m .69307 .90241 L .66329 .89143 L F
.68913 .88348 m .66329 .89143 L .66332 .8911 L F
.68913 .88348 m .66332 .8911 L .66355 .88896 L F
.69307 .90241 m .68072 .92326 L .66353 .90622 L .814 .604 0 r
F
.68072 .92326 m .66449 .92359 L .66353 .90622 L .827 .599 0 r
F
.68913 .88348 m .66355 .88896 L .66371 .88749 L .815 .589 0 r
F
.68913 .88348 m .66371 .88749 L .66372 .88742 L F
.68913 .88348 m .66376 .88676 L .66371 .88664 L F
.68913 .88348 m .66372 .88742 L .66379 .88675 L F
.66379 .88675 m .66376 .88676 L .66376 .88676 L F
.66376 .88677 m .66376 .88676 L .66379 .88675 L F
.66378 .88681 m .66376 .88677 L .66379 .88675 L F
.68001 .92328 m .66498 .92966 L .66449 .92359 L .827 .602 0 r
F
.66455 1.03462 m .67298 1.02672 L .67426 1.05738 L .958 .849 0 r
F
.66616 1.06357 m .66455 1.03462 L .67426 1.05738 L .907 .847 0 r
F
.66502 .92976 m .66498 .92966 L .68001 .92328 L .827 .602 0 r
F
.665 1.11471 m .66919 1.09852 L .67297 1.10718 L .894 .989 0 r
F
.665 1.11471 m .67297 1.10718 L .66649 1.11907 L .76 .972 0 r
F
.68072 .92326 m .66502 .92976 L .68001 .92328 L .827 .602 0 r
F
.68072 .92326 m .67208 .94651 L .66502 .92976 L F
.66516 1.08602 m .66616 1.06357 L .67324 1.08002 L .932 .908 0 r
F
.66516 1.08602 m .67324 1.08002 L .66919 1.09852 L .915 .962 0 r
F
.66521 .50366 m .66561 .50618 L .66518 .50366 L .887 .739 0 r
F
.6928 .50264 m .66646 .51151 L .66521 .50366 L F
.66616 1.06357 m .67426 1.05738 L .67324 1.08002 L .913 .907 0 r
F
.6928 .50264 m .69778 .55483 L .66646 .51151 L .887 .739 0 r
F
.69622 .55267 m .67336 .55458 L .66646 .51151 L .882 .739 0 r
F
.66649 1.11907 m .67297 1.10718 L .67487 1.11072 L .751 .958 0 r
F
.66649 1.11907 m .67487 1.11072 L .66938 1.11921 L .66 .915 0 r
F
.66792 .99611 m .67563 .98935 L .68255 1.01978 L .954 .795 0 r
F
.67298 1.02672 m .66792 .99611 L .68255 1.01978 L .939 .8 0 r
F
.66806 .95646 m .67389 .95575 L .68505 .97722 L .982 .707 0 r
F
.68118 .98359 m .66806 .95646 L .68505 .97722 L .983 .717 0 r
F
.67563 .98935 m .66806 .95646 L .68118 .98359 L .97 .741 0 r
F
.70295 .31586 m .70966 .33538 L .66808 .32306 L .841 .659 0 r
F
.70966 .33538 m .67266 .33478 L .66808 .32306 L .846 .634 0 r
F
.66919 1.09852 m .67324 1.08002 L .67697 1.09144 L .926 .969 0 r
F
.66919 1.09852 m .67697 1.09144 L .67297 1.10718 L .895 .988 0 r
F
.66938 1.11921 m .67487 1.11072 L .67903 1.11203 L .681 .961 0 r
F
.68103 .95306 m .67509 .95365 L .67208 .94651 L .831 .599 0 r
F
.68072 .92326 m .68103 .95306 L .67208 .94651 L .827 .602 0 r
F
.70966 .33538 m .71743 .37135 L .67266 .33478 L .849 .67 0 r
F
.71743 .37135 m .68093 .37394 L .67266 .33478 L .851 .668 0 r
F
.67297 1.10718 m .67697 1.09144 L .68091 1.09953 L .896 .987 0 r
F
.67297 1.10718 m .68091 1.09953 L .67487 1.11072 L .744 .96 0 r
F
.67298 1.02672 m .68255 1.01978 L .68663 1.04467 L .948 .849 0 r
F
.68093 1.05046 m .67298 1.02672 L .68663 1.04467 L .957 .844 0 r
F
.67426 1.05738 m .67298 1.02672 L .68093 1.05046 L .949 .844 0 r
F
.67324 1.08002 m .67426 1.05738 L .68058 1.07261 L .946 .927 0 r
F
.67324 1.08002 m .68058 1.07261 L .67697 1.09144 L .929 .968 0 r
F
.67336 .5551 m .67336 .55458 L .69622 .55267 L .882 .739 0 r
F
.69778 .55483 m .67336 .56 L .67336 .5551 L F
.69778 .55483 m .67336 .5551 L .69622 .55267 L F
.69778 .55483 m .67337 .56518 L .67336 .56 L .884 .762 0 r
F
.69778 .55483 m .69863 .58007 L .67337 .56518 L F
.69863 .58007 m .67338 .58035 L .67337 .56518 L .866 .777 0 r
F
.69863 .58007 m .67338 .58344 L .67338 .58035 L .862 .791 0 r
F
.69863 .58007 m .69972 .59787 L .67338 .58344 L F
.68274 .58857 m .67339 .58898 L .67338 .58344 L .89 .767 0 r
F
.67367 .59042 m .67339 .58898 L .68274 .58857 L F
.69972 .59787 m .67367 .59042 L .68274 .58857 L F
.69972 .59787 m .67651 .60525 L .67367 .59042 L F
.67389 .95575 m .67518 .95386 L .68723 .97392 L .889 .502 0 r
F
.68723 .97392 m .68505 .97722 L .67389 .95575 L .977 .675 0 r
F
.67426 1.05738 m .68093 1.05046 L .68058 1.07261 L .955 .924 0 r
F
.67487 1.11072 m .68091 1.09953 L .68366 1.10307 L .752 .975 0 r
F
.67487 1.11072 m .68366 1.10307 L .67903 1.11203 L .727 .966 0 r
F
.68103 .95306 m .67518 .95386 L .67509 .95365 L .829 .594 0 r
F
.68103 .95306 m .67752 .95775 L .67518 .95386 L F
.6913 .66627 m .70198 .67648 L .67558 .67994 L .811 .768 0 r
F
.70198 .67648 m .75212 .69485 L .67558 .67994 L .814 .715 0 r
F
.75212 .69485 m .71626 .69156 L .67558 .67994 L .833 .618 0 r
F
.67563 .98935 m .68118 .98359 L .69017 1.01337 L .982 .777 0 r
F
.68255 1.01978 m .67563 .98935 L .69017 1.01337 L .962 .797 0 r
F
.70418 .23874 m .71409 .25904 L .67582 .24835 L .86 .686 0 r
F
.71409 .25904 m .68561 .269 L .67582 .24835 L .865 .659 0 r
F
.69972 .59787 m .67735 .60963 L .67651 .60525 L .891 .777 0 r
F
.67697 1.09144 m .68058 1.07261 L .68541 1.08494 L .915 .96 0 r
F
.67697 1.09144 m .68541 1.08494 L .68091 1.09953 L .865 .996 0 r
F
.71626 .69156 m .71645 .73144 L .67711 .69245 L .805 .6 0 r
F
.71645 .73144 m .68044 .73448 L .67711 .69245 L .806 .6 0 r
F
.69972 .59787 m .7023 .62378 L .67735 .60963 L .891 .777 0 r
F
.68843 .61592 m .67872 .61678 L .67735 .60963 L .879 .788 0 r
F
.68103 .95306 m .68219 .96553 L .67752 .95775 L .829 .594 0 r
F
.69606 .86426 m .68913 .88348 L .67784 .86619 L .809 .604 0 r
F
.67911 .6185 m .67872 .61678 L .68843 .61592 L .879 .788 0 r
F
.67903 1.11203 m .68366 1.10307 L .68952 1.10431 L .729 .985 0 r
F
.7023 .62378 m .68146 .62879 L .67911 .6185 L .879 .788 0 r
F
.7023 .62378 m .67911 .6185 L .68843 .61592 L F
.71645 .73144 m .71424 .77483 L .68044 .73448 L .804 .591 0 r
F
.71424 .77483 m .68179 .7761 L .68044 .73448 L .812 .59 0 r
F
.68058 1.07261 m .68093 1.05046 L .68695 1.06513 L .964 .927 0 r
F
.68058 1.07261 m .68695 1.06513 L .68541 1.08494 L .956 .948 0 r
F
.71987 .92011 m .68072 .92326 L .69307 .90241 L .823 .616 0 r
F
.71987 .92011 m .71499 .94956 L .68072 .92326 L .818 .594 0 r
F
.71499 .94956 m .68103 .95306 L .68072 .92326 L F
.68091 1.09953 m .68541 1.08494 L .68926 1.09169 L .868 .995 0 r
F
.68091 1.09953 m .68926 1.09169 L .68366 1.10307 L .758 .974 0 r
F
.71743 .37135 m .72388 .4112 L .68093 .37394 L .852 .681 0 r
F
.72388 .4112 m .68788 .41652 L .68093 .37394 L .855 .68 0 r
F
.68093 1.05046 m .68663 1.04467 L .68695 1.06513 L .959 .93 0 r
F
.71499 .94956 m .70938 .99456 L .68103 .95306 L .818 .596 0 r
F
.68532 .97074 m .68219 .96553 L .68103 .95306 L .832 .596 0 r
F
.69834 .97841 m .69242 .98665 L .68103 .95306 L F
.69242 .98665 m .68739 .97432 L .68103 .95306 L F
.68739 .97432 m .68442 .96704 L .68103 .95306 L F
.68118 .98359 m .68505 .97722 L .69517 1.00342 L .989 .734 0 r
F
.69017 1.01337 m .68118 .98359 L .69517 1.00342 L .982 .777 0 r
F
.7023 .62378 m .68295 .6353 L .68146 .62879 L .854 .898 0 r
F
.71424 .77483 m .71157 .82926 L .68179 .7761 L .814 .6 0 r
F
.71157 .82926 m .6851 .83152 L .68179 .7761 L .812 .6 0 r
F
.68255 1.01978 m .69017 1.01337 L .69137 1.03844 L .973 .857 0 r
F
.68663 1.04467 m .68255 1.01978 L .69137 1.03844 L .974 .856 0 r
F
.7023 .62378 m .69778 .64532 L .68295 .6353 L .854 .898 0 r
F
.69068 .64052 m .68425 .64098 L .68295 .6353 L .935 .867 0 r
F
.68366 1.10307 m .68926 1.09169 L .69399 1.0956 L .762 .993 0 r
F
.68366 1.10307 m .69399 1.0956 L .68952 1.10431 L .737 .987 0 r
F
.68481 .64301 m .68425 .64098 L .69068 .64052 L .935 .867 0 r
F
.68611 .97207 m .68442 .96704 L .68723 .97392 L .832 .596 0 r
F
.68611 .97207 m .68532 .97074 L .68442 .96704 L F
.69778 .64532 m .68481 .64301 L .69068 .64052 L .935 .867 0 r
F
.69778 .64532 m .68736 .65215 L .68481 .64301 L F
.68505 .97722 m .68723 .97392 L .69614 .99577 L .981 .689 0 r
F
.69614 .99577 m .69517 1.00342 L .68505 .97722 L .989 .732 0 r
F
.71157 .82926 m .71285 .86377 L .6851 .83152 L .811 .595 0 r
F
.71285 .86377 m .69606 .86426 L .6851 .83152 L .822 .593 0 r
F
.68541 1.08494 m .68695 1.06513 L .69312 1.07741 L .944 .945 0 r
F
.68541 1.08494 m .69312 1.07741 L .68926 1.09169 L .885 .989 0 r
F
.71409 .25904 m .72612 .28864 L .68561 .269 L .865 .66 0 r
F
.72612 .28864 m .69386 .29211 L .68561 .269 L .857 .679 0 r
F
.68663 1.04467 m .69137 1.03844 L .69353 1.05878 L .976 .912 0 r
F
.68695 1.06513 m .68663 1.04467 L .69353 1.05878 L .951 .928 0 r
F
.68695 1.06513 m .69353 1.05878 L .69312 1.07741 L .947 .944 0 r
F
.69778 .64532 m .68935 .6593 L .68736 .65215 L .921 .947 0 r
F
.72388 .4112 m .72792 .45587 L .68788 .41652 L .858 .696 0 r
F
.72792 .45587 m .6911 .46155 L .68788 .41652 L .856 .697 0 r
F
.71285 .86377 m .68913 .88348 L .69606 .86426 L .827 .625 0 r
F
.7267 .88027 m .68913 .88348 L .71285 .86377 L .818 .608 0 r
F
.7267 .88027 m .72274 .90196 L .68913 .88348 L .817 .605 0 r
F
.72274 .90196 m .69307 .90241 L .68913 .88348 L .823 .594 0 r
F
.68926 1.09169 m .69312 1.07741 L .6985 1.08486 L .871 .988 0 r
F
.68926 1.09169 m .6985 1.08486 L .69399 1.0956 L .767 .994 0 r
F
.69778 .64532 m .69653 .66005 L .68935 .6593 L .921 .947 0 r
F
.69653 .66005 m .69083 .66461 L .68935 .6593 L .916 .951 0 r
F
.68952 1.10431 m .69399 1.0956 L .69861 1.09366 L .723 .941 0 r
F
.69017 1.01337 m .69517 1.00342 L .69632 1.03252 L .998 .843 0 r
F
.69137 1.03844 m .69017 1.01337 L .69632 1.03252 L .983 .861 0 r
F
.69653 .66005 m .6913 .66627 L .69083 .66461 L .893 .982 0 r
F
.72792 .45587 m .72719 .49904 L .6911 .46155 L .859 .716 0 r
F
.72719 .49904 m .6928 .50264 L .6911 .46155 L .864 .715 0 r
F
.69653 .66005 m .69564 .66579 L .6913 .66627 L .893 .982 0 r
F
.69564 .66579 m .69502 .66983 L .6913 .66627 L F
.69137 1.03844 m .69632 1.03252 L .70054 1.05192 L .984 .885 0 r
F
.69353 1.05878 m .69137 1.03844 L .70054 1.05192 L .951 .911 0 r
F
.69414 .99088 m .69242 .98665 L .69834 .97841 L .832 .596 0 r
F
.72719 .49904 m .72944 .54937 L .6928 .50264 L .865 .731 0 r
F
.72944 .54937 m .69778 .55483 L .6928 .50264 L .874 .731 0 r
F
.72274 .90196 m .71987 .92011 L .69307 .90241 L .826 .613 0 r
F
.69312 1.07741 m .69353 1.05878 L .70125 1.06989 L .94 .942 0 r
F
.69312 1.07741 m .70125 1.06989 L .6985 1.08486 L .912 .98 0 r
F
.69353 1.05878 m .70054 1.05192 L .70125 1.06989 L .946 .94 0 r
F
.72612 .28864 m .7345 .31036 L .69386 .29211 L .852 .647 0 r
F
.7345 .31036 m .70295 .31586 L .69386 .29211 L .855 .639 0 r
F
.69399 1.0956 m .6985 1.08486 L .7054 1.08985 L .744 .989 0 r
F
.69399 1.0956 m .7054 1.08985 L .69861 1.09366 L .358 .805 0 r
F
.69614 .99577 m .69414 .99088 L .69834 .97841 L .832 .596 0 r
F
.6962 .67096 m .69502 .66983 L .69653 .66005 L .89 .981 0 r
F
.69517 1.00342 m .69614 .99577 L .70244 1.01814 L .99 .766 0 r
F
.70244 1.01814 m .70056 1.02518 L .69517 1.00342 L 1 .817 0 r
F
.69632 1.03252 m .69517 1.00342 L .70056 1.02518 L .987 .84 0 r
F
.69615 .99581 m .69614 .99577 L .69834 .97841 L .832 .596 0 r
F
.70938 .99456 m .69615 .99581 L .69834 .97841 L F
.69689 .99574 m .69694 .99863 L .69615 .99581 L .836 .612 0 r
F
.7121 .65833 m .6962 .67096 L .69653 .66005 L .89 .981 0 r
F
.69632 1.03252 m .70056 1.02518 L .70729 1.04451 L .989 .851 0 r
F
.70054 1.05192 m .69632 1.03252 L .70729 1.04451 L .961 .889 0 r
F
.71746 .64569 m .69653 .66005 L .69778 .64532 L .896 .932 0 r
F
.71746 .64569 m .7121 .65833 L .69653 .66005 L .841 .994 0 r
F
.70938 .99456 m .70438 1.02593 L .69689 .99574 L .836 .612 0 r
F
.70244 1.01814 m .69848 1.00409 L .69689 .99574 L F
.69689 .99574 m .69748 .99882 L .69694 .99863 L F
.69748 .99882 m .69717 .99943 L .69694 .99863 L F
.69748 .99882 m .69723 .99965 L .69717 .99943 L F
.69748 .99882 m .69735 1.00008 L .69723 .99965 L F
.69748 .99882 m .69848 1.00409 L .69735 1.00008 L F
.72944 .54937 m .72786 .57442 L .69778 .55483 L .876 .762 0 r
F
.72786 .57442 m .69863 .58007 L .69778 .55483 L .88 .759 0 r
F
.72094 .63228 m .69778 .64532 L .7023 .62378 L .889 .93 0 r
F
.72094 .63228 m .71746 .64569 L .69778 .64532 L .885 .94 0 r
F
.6985 1.08486 m .70125 1.06989 L .70816 1.07716 L .89 .97 0 r
F
.6985 1.08486 m .70816 1.07716 L .7054 1.08985 L .838 .999 0 r
F
.72786 .57442 m .72757 .59117 L .69863 .58007 L .879 .805 0 r
F
.72757 .59117 m .69972 .59787 L .69863 .58007 L .881 .803 0 r
F
.72757 .59117 m .72532 .61545 L .69972 .59787 L .88 .812 0 r
F
.72532 .61545 m .7023 .62378 L .69972 .59787 L .912 .79 0 r
F
.70054 1.05192 m .70729 1.04451 L .70911 1.06249 L .959 .927 0 r
F
.70125 1.06989 m .70054 1.05192 L .70911 1.06249 L .94 .938 0 r
F
.70056 1.02518 m .70244 1.01814 L .7073 1.03773 L .997 .829 0 r
F
.7073 1.03773 m .70729 1.04451 L .70056 1.02518 L .969 .733 0 r
F
.70125 1.06989 m .70911 1.06249 L .70816 1.07716 L .932 .959 0 r
F
.72532 .61545 m .72094 .63228 L .7023 .62378 L .89 .93 0 r
F
.7345 .31036 m .74436 .33315 L .70295 .31586 L .859 .658 0 r
F
.74436 .33315 m .70966 .33538 L .70295 .31586 L .856 .666 0 r
F
.70729 1.04451 m .7073 1.03773 L .70821 1.0536 L .771 .62 0 r
F
.74436 .33315 m .75205 .36931 L .70966 .33538 L .858 .675 0 r
F
.75205 .36931 m .71743 .37135 L .70966 .33538 L F
.73626 .82598 m .71157 .82926 L .71424 .77483 L .821 .607 0 r
F
.73626 .82598 m .72998 .86376 L .71157 .82926 L .821 .605 0 r
F
.72998 .86376 m .71285 .86377 L .71157 .82926 L .825 .606 0 r
F
.7275 .64042 m .7121 .65833 L .71746 .64569 L .833 .98 0 r
F
.72998 .86376 m .7267 .88027 L .71285 .86377 L .825 .607 0 r
F
.74329 .77486 m .71424 .77483 L .71645 .73144 L .841 .624 0 r
F
.74329 .77486 m .73626 .82598 L .71424 .77483 L .84 .614 0 r
F
.74963 .7311 m .71645 .73144 L .71626 .69156 L .839 .628 0 r
F
.75212 .69485 m .74963 .7311 L .71626 .69156 L .834 .629 0 r
F
.74963 .7311 m .74329 .77486 L .71645 .73144 L .838 .624 0 r
F
.75205 .36931 m .76081 .40785 L .71743 .37135 L .858 .673 0 r
F
.76081 .40785 m .72388 .4112 L .71743 .37135 L .849 .679 0 r
F
.73713 .63329 m .71746 .64569 L .72094 .63228 L .895 .949 0 r
F
.73713 .63329 m .7275 .64042 L .71746 .64569 L .478 .866 0 r
F
.74448 .62035 m .72094 .63228 L .72532 .61545 L .881 .92 0 r
F
.74448 .62035 m .73713 .63329 L .72094 .63228 L .85 .967 0 r
F
.76081 .40785 m .76218 .45032 L .72388 .4112 L .851 .711 0 r
F
.76218 .45032 m .72792 .45587 L .72388 .4112 L .87 .705 0 r
F
.74908 .60774 m .72532 .61545 L .72757 .59117 L .902 .831 0 r
F
.74908 .60774 m .74448 .62035 L .72532 .61545 L .878 .922 0 r
F
.7595 .49407 m .72719 .49904 L .72792 .45587 L .874 .729 0 r
F
.7595 .49407 m .76039 .54399 L .72719 .49904 L .875 .738 0 r
F
.76039 .54399 m .72944 .54937 L .72719 .49904 L .874 .738 0 r
F
.75306 .58834 m .72757 .59117 L .72786 .57442 L .893 .817 0 r
F
.75306 .58834 m .74908 .60774 L .72757 .59117 L .889 .836 0 r
F
.75598 .56742 m .72786 .57442 L .72944 .54937 L .89 .774 0 r
F
.75598 .56742 m .75306 .58834 L .72786 .57442 L .89 .819 0 r
F
.76218 .45032 m .7595 .49407 L .72792 .45587 L .872 .729 0 r
F
.76039 .54399 m .75598 .56742 L .72944 .54937 L .874 .785 0 r
F
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[9]:=",
  ImageSize->{219.812, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40003K000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI]/7003K1`00f`L00=/7003K1`00f`L00=/7
003K1`00f`L00=/7003K1`00f`L00=/7003K1`00f`L00=/7003K1`00f`L00=/7003K1`00f`L00=/7
003K1`00f`L004<700<0YjL08JMd1`00@@L00`2WY`02Y`0U[:N/YjbW[:N/YjbW`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W``1`1`00?0LgYfP7000i1`802jL0;jbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO306H7000f1cnW00>nY`L0H`L003@74:L0>;jWYjN/YjbW[:O3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>WG`L00307AJL00kjWY`05Yeh7
000^1`03[:N/01>W03^nYjNW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W``1L1`00:`MBY`04_ZNnYeX7000Y1`0>0:bW[:N/YjbW[:N/Yj`9Y`13[:N/YjbW
_ZN/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
``1H1`00:0MEY`09_ZNnYkjW_ZNn05D7000V1`1R0:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:O3YjbW
[:N/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jMC1`009@MBY`0C[:N/YjbW_ZNnYkjW`jO3Yl>W``1A1`0090L0IjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/Yl>W[:N/YjbW`jO3Yl>W`jO3YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl<0D0L0028700<0YjL0A:L04JbW[:N/YjbW[:N/Yl>W
[:N/00>W00o3Yl>W`jO3Yl>W`jO3Yl<0C`L002070P00JjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW`jO3YjbW[:O3Yl>W`jO3YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO304h7000O1`80::L00jbWY`0MY`0?[:N/YjbW[:N/YjbW`jN/00>W
01?3Yl>W`jO3Yl>W`jO3Yl>W`jO304d7000N1`1a0:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:O3Yl>W[:O3Yl>W`jN/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jN/Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl<0C0L001d700<0YjL0:JL00jbWY`0MY`0=[:N/YjbW[:N/YjbW
``05Y`0E`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO304/7000L1`1f0:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW`jN/Yl>W`jO3YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3YdT7000K1`030:NW02^W00>/YjL07jL02JbW
[:N/YjbW[007Y`0H[:N/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>WB0L001X70P003JbW[:N/YjbW[:N/Yj`0
0jL02JbW[:N/YjbW[003Y`1L[:N/YjbW[:O3YjbW[:N/Yl>W`jO3YjbW[:N/YjbW`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<>W`jM71`006PL00`2W
Y`0/Y`03[:NW01nW00V/YjbW[:N/Yj`01JL00l>W[003Y`0F`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3YdH7
000I1`0<0:bW[:N/YjbW[:N/0jL01JbW[:N/00nW05Z/YjbW[:O3YjbW[:N/Yl>W[:N/YjbW[:N/YjbW
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3[<>/`jO3Yl>W
`jM51`0060L00`2WY`0VY`R400>MQ:L07jL02JbW[:N/YjbW[005Y`03`jNW00>W00?3Yl<00jL04L>W
`jO3Yl>W`jO3Yl>W`jO304D7000G1`0<0:bW[:N/YjbW[:N/6jL0FHA[[:N/YjbW[:N/Yl>W[:N/YjbW
[:N/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jc3Yl>W
`jO3Yl>W`jO304@7000F1`809:L00f^WY`08Y`05[:N/Yj`02jL08jbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/JhV4RHBR00>400?3Yl<00jL05L>W`jO3Yl>W`jO3Yl>W`jO3Yl>W``131`005PL03@2W
[:N/YjbW[:N/Yj`02jL01jbW[:N/Yj`01ZL0GXB/Yl>W`jN/YjbW[:N/Yl>W[:N/YjbW[:N/Yl>W`jO3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yg2W`jO3Yl>/`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl=21`005@L202:W00N4JjbW[:N/00VW00>/Yj`02JL09:bW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/Yj`0[:N/Y`:/00G3YjbW[003Y`0F`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yd87000D1`8002:/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW0XT0H:bW[:N/YjbW`jN/YjbW[:O3Y`>W[:N/YjbW
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jM`Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yd47000D1`0306^W026W00>4RJL01:L00jbW[007Y`04[:N/``RW02F/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:N/0:bW[:O300:/01nW[:N/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W0407000C1`0T06^W[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/0XT0HZN/YjbW[:N/
YjbW`jN/YjbW[:L3YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>WL:O3Yl>/`jO3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jc3[<>/`jc3@0L001<70P0QY`:92jL00jbWY`05Y`03[:NW00FW
02B/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`0[:L2[00S`jN/YjbW[:N/YjbW`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl<0?`L00187065[06^W[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N9YjbW[:N/YjbW[:N/YjbW`jN/YjbW`jN/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO300>W02K3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Ycl7000B1`0306^W022W
00F9L:bW[00;Y`03[:N/00>W00>/YjL00jL0BjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/RJbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W``0o1`004@L0HP1[0:bW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[8V/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jN/YjbW`jO3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO30jL09l>/`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[<>/
`jO3Yl>W`jO3Yl>W``0n1`004@L00f/0J`0QY`0=RJN/YjbW[:N/YjbW[007Y`05[:NWYl<00jL0C:bW
[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/RJbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jLn1`004@L2J`1PYjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbWRJN/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW[<>/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
0jL09l>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<>/`jc3[<>/`jc3Yl>W``0n1`004@L00f/0Y`0PY`1Y
L:bW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yl>W[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[8V/YjbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W03h7000@1`07J`2/YjbW
[007Y`0^[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yf^W[:N/YjbW[:N/YjbW[:N/YjbW[:N/Y`?302BW`jO3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl<2Y`0Y[<>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jc3[<>/`jc3[<>/`jL0?@L0010700<0JjL08JL00f^WY`0DY`1C[:O3YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yjb9[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl<0?@L0010700E[0:bW[009Y`0^[:N/YjbW[:N/YjbW[:N/YjbW[:N/YhVW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/Y`?302BW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl<3Y`0X`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Ycd7000@1`030:NW022W00<0JjL05JL0E:bW
`jN/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/RJbW`jN/YjbW[:N/YjbW[:N/YjbW
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yc`7000?1`0mJ`2W[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/L:bW[:N/YjbW[:N/YjbW[:N/YjbW[:O3Y`03``0TYl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jO3Yl>W`jO30jL0:L>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jc3[<>/`jO3Yl>W`jO3Yl>W`jO303`7
000?1`030:NW026W00>9YjL04jL0EjbW[:O3YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/Yjb>[:N/Yl>W`jN/YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W``0k1`003PL0?V/0
[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yja`[:N/YjbW[:N/YjbW[:N/YjbW[:O3Yl>W
0l<09:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W``:W02^>`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jO3[<>/`jc3Yl>W`jO3Yl>W`jO3Yl>W03/7000>1`808ZL00f^9Y`0BY`0][:N/YjbW`jN/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/00:>02ZW[:N/Yl>W`jN/YjbW[:N/Yl>W`jO3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl<k1`003PL2000l[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YhU`[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W0l<09:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jO3Yl>W``:W02b>`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[<>W`jO3Yl>W`jO3Yl>W`cX7
000>1`808ZL00f^9Y`0@Y`0`[:N/YjbW[:O3YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/Yjb>0ZL0:ZbW[:N/Yl>W`jO3YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3YcX7000>1`80
02:/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW0XT06:bW[:N/YjbW[:N/YjbW[:N/Yl>W
`jO3Y`?302>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Y`021`0H`hk3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>/0PL05<>W`jO3Yl>W`jO3Yl>W`jO3Yl>W>@L000h700<0YjL08JL00f^9Y`0@Y`0_
[:N/YjbW[:O3YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`00jL0:jbW[:N/YjbW
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl<0>@L000d70P008jN/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW00:901R/YjbW[:N/YjbW[:N/YjbW[:O3Yl>W`jL3``1CYl>W`jO3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jL7Yl>>`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jO3Yl>W`jc3[<>/`j`0>@L000d70P0RY`03RG2W016W02j/YjbW[:N/Yl>W[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:N9Q9e[0hh0;JN/1jbW[:N/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jO3Yl>W`jO3Y`0h1`003@L2001OYjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[72>
YjbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W0jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`f/0
0hT4T`0^S/>W``O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<>/`cP7000=1`80
8ZL00hU`Y`0AY`1N[:N/YjbW[:N/`jbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[0N/08VW[8j/YjbW
`jN/YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3YcP7000=1`8005jW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/RHjW[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W`jL3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W``1`0Xh01JN/S/>W00;302d3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<<0>0L000d70P0RY`03RG2W016W02b/YjbW[:N/Yjc3
[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Y`1`SW2WSP:W00>/Yl<00Z`0;:N/YjbW[:N/YjbW[:O3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3=`L000d705l0[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/Yjb9SZN/YjbW[:N/YjbW[:N/YjbW[:O3Yl>W`jO3Y`>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jO3Yl>W`jO30004SP04[8ja[0:a02k3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>/`jc3[<>/=`L000`70P0SY`03RF^W00NW00>/YjL01jL0::bW[:N/YjbW[<>/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/Yj`0SW02SP03[8j/00>/02l3[:bW[:N/YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W``0g1`0030L2000S[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/Yj`00XT0>JN/YjbW[:N/YjbW[:N/YjbW[:O3Yl>W`jO3Y`>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jL0R@02SP08[8j/[;6/0j`2Y`0/`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[<>/`jc3[<>/
`jc3[<>/`j`g1`0030L202>W00>9JjL01JL00jbW[007Y`0^[:N/YjbW[:N/Yjc3[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/08i`[8j/SPB/0307RJbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO3Yl>W`jLf1`0030L09@2W[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`0
0XT0>JN/YjbW[:N/YjbW[:N/YjbW[:O3Yl>W`jO3Y`>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jL7Yl>W`jN>
R@07[00b/@>>`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<<f1`00
2`L202>W00=`RJL01:L00jbW[005Y`0][:N/YjbW[:N/YjbW[:N/`jbW[:N/YjbW[:N/YjbW[:N/YjbW
[:L7YjbW[8V>00NW03;3[<>9[:N/YjbW[:N/YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3YcH7000;1`8005bW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yhi`[:N/YjbW[:N/
YjbW[:N/YjbW[:O3Yl>W`jO3Y`>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W``:W00Z>[:N/YjbW`jc3
0jL0;L>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[<>/`jc3[<>/`jc3[<>/``0f1`002`L202>W
00=`RJL00ZL01JbW[:N/00>W02j/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/Yjb90Xh6Y`0c`jc3RJbW[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO303D7000;1`8005bW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YhU`[:N/YjbW[:N/
YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W``:W00V>[:N/Yjb>`j`0
0/<2Y`0^`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[3D7000;1`80
8jL0>f^9YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
Yjb900:>0jL0=XjW`l>/`hV/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3
Yl>W`cD7000;1`8005bW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YhU`[:N/YjbW[:N/
YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W``:W00B>[:N/0ZL00hh3
[002``:W02k3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/=@L000/7
0P0SY`08L8VW[:N/Yj`=Y`0V[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[8T2SP05
YhjWSZL00/<0<Zc3RJbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jc3[<>/`jc3
=@L000/705h0[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbWRG2/YjbW[:N/YjbW[:N/
YjbW[:N/Yl>W`jO3Yl>W`jc3[<>/`jc3Yl>W`jO3Yl>W`jO3Yl>W`jO30ZL018j/Yj`2Y`03S/?300;3
0ZL0;/>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`j`e1`002`L00`2W
Y`0RY`06L8VW[:N/3JL0::bW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[8T2SP>W
03J9Yl?3Yl>9[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[<<e1`00
2`L0GP2/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N9L:bW[:N/YjbW[:N/YjbW[:N/
YjbW`jO3Yl>W`jO3[<>/`jc3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl<2Y`04SZNW[0:W00>>`l<00/<0<8jW
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[3D7000;1`030:NW026W
00N9L:bW[:N/00^W02Z/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[8T2SP>W
03J9YhW3Yl>9[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<<e1`00
2`L0H@2/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[729YjbW[:N/YjbW[:N/YjbW[:N/
YjbW`jO3Yl>W`jO3[<>/`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>WL8h00ZL01JbWYhjW00?3032>Yl>W
`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO31l>/`jc3[<>/`jc3[<>/`jc3[<>/`j`e1`002`L00`2WY`0QY`05
RG2/Yj`02jL0=ZbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`0SXVWSZN9
YhW3Y`;302n/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jc3[<>/``0e1`00
2`L0H@2/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[729YjbW[:N/YjbW[:N/YjbW[:N/
YjbW`jO3Yl>W`jO3[<>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>WL8h01JL00hjW``03``0_Yl>W`jO3
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<>/`jc3[<>/`jc3[<>/`j`0=@L000/700=`YjL08JL00hU`
[00;Y`0h[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/08j9YhjWRJN9
`j`2``0_[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<<0=@L000/7
065[[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yja`RJN/YjbW[:N/YjbW[:N/YjbW[:N/
Yl>W`jO3Yl>W`jc3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yg2>00FW00>>Yl<00l<0;jO3Yl>W`jO3
Yl>W`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[<>/`jc3[<>/`jc3[<>/03D7000;1`03L:NW026W00>9L:L0
2JL0>ZbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/08j9YhjWSZN>
`jL2``0H[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW0Z`05L>/`jc3[<>/`jc3[<>/`jc3[<>/``0e1`00
2PL09@1[[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`00W00?JN/YjbW[:N/Yl>W[:N/
YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yg2>Yhh00jL00hjW``03``0_
Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[<>/`jc3[<>/`jc3[<>/`j`0=@L000X700<0L:L0
8ZL00hU[Y`07Y`0D[:N/YjbW[:N/YjbW[:N/YjbW[:L3[00SYjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
Yj`0SXVWSZN>Yhh00ZL2``0H[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW0Z`05L>/`jc3[<>/`jc3[<>/
`jc3[<>/``0e1`002PL09G1[[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`00W00>:N/
YjbW[:O3Yl>W[:N/YjbW[:N/Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO30ZL00hjW
SP03Y`03SZN>00?302nW`jO3Yl>W`jO3[<>/`jc3[<>W`jO3Yl>W`jc3[<>/`jc3[<>/`jc3[<>/`jc3
[00e1`002PL00`1`Y`0QY`9`00=[YjL01:L05ZbW[:N/YjbW[:N/YjbW[:N/YjbW[:L3[00SYjbW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/Yjb9SXVWSZN>Yhh00ZL3``0GYjbW[:N/YjbW[:N/YjbW[:N/YjbW[:L0
0Z`05L>/`jc3[<>/`jc3[<>/`jc3[<>/``0e1`002PL0Gg1[[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YhU[[:N/YjbW`jO3YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl>W`jO300:W00>>Yhh00jL00hjWSP05``0]Yl>W`jO3Yl>/`jc3[<>/`jO3Yl>W`jO3[<>/`jc3
[<>/`jc3[<>/`jc3[<>/03D7000:1`03072W026W0W05Y`0H[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
0j`09JN/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/RHj9YhjWSZN>Yj`00l<0;ZN/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:O3[<>/`jc3[<>/`jc3[<>/`jc3[<<e1`002PL0Gg00[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/Yg1[[:N/Yl>W`jO3YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W0jO3Yl>W`jO3Yl>W
`jO3Yk6//JbaTk6W`jO300:W00F>YhjWSP04Y`G302fW`jO3Yl>W`jc3[<>/`jc3Yl>W`jO3Yl>/`jc3
[<>/`jc3[<>/`jc3[<>/`j`0=@L000X700<0JjL02ZL02jbW[:N/YjbW[:N/00bW01m[L:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW00>/00FW[:N/Y`02[007/Jba[;6//@06[00ATjbC[9>/RHj9YhjW
SZN>Yj`00l<0;ZN/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:O3[<>/`jc3[<>/`jc3[<>/`jc3[<<e1`00
2PL0Bg00[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yg1[`jO3Yl>W`jO3YjbW[:N/YjbW
[:O3Yl>W`jO3Yl>W0jO3Y`>/0`02[006YjbW[:N/1JL01l>W`jO3Yl<00ZL01HjWSZN>00BW00?3Yl<0
0/<0;JO3Yl>W`jO3[<>/`jc3[<>W`jO3Yl>W`jc3[<>/`jc3[<>/`jc3[<>/`jc3[00e1`002@L00g00
J`0;Y`09[:N/YjbW[:N/00jW01m[L:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW00>/00BW/Jba
5JL030N9SXVWRJN>YhjW[0?302jW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jc3[<>/`jc3[<>/`jc3
[<>/`jc3=@L000T7059[L02/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yja[L:N/YjbW[:N/
YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jO3/@>a`jO3YjbW[:N/1JL04<>W`jO3Yl>W``2WSZN>Yhh4Y`03
`jO300;302fW`jO3Yl>W`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`j`0=@L000T7
00=`0:L02jL01jbW[:N/Yj`03jL0871[[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW0j`00jN/
Y`0BY`0>_ZNWYf^9SXVWRJN>Yhh2Y`0a`jO3`jbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yl>/`jc3[<>/
`jc3[<>/`jc3[<>/``0e1`002@L0901`YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[09[
02BW[:N/YjbW[:N/YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl<3Y`05[:NWYj`01JL04<>W`jO3Yl>W
``2WSZN>Yhh4Y`03`jO300;302fW`jO3Yl>W`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3
[<>/`j`0=@L000T700=[0:L02jL00jbW[00CY`0PL02/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:L3[0BW00>>YjL04:L02;jWJhV>RJN91ZL07<>W`l>/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:L2[00C
`jc3[<>/`jc3[<>/`jc3[<>/``0e1`002@L0BP1`YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbWL6^/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W`jO3Yl>W`jO30jL00jbWY`09Y`0D_ZNnYl>W
``2WSZN>Yhj/YjbW`j`3``0]Yl>W`jO3Yl>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3
[<>/03D700091`802ZL00jbW[00DY`0QJ`2W[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW00>/
1:L00hjWY`0;Y`03XZNW00:W00RnYf^9SXVWR@JW01c3Yl?3[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
0Z`04l>/`jc3[<>/`jc3[<>/`jc3[<<0=@L000T704H0L:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/Yg00[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jO30jL00l>WY`04Y`03SZNW00RW
01:nYkjW``2WSZN>Yhj/YjbW`j`3``0]Yl>W`jO3Yl>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3
[<>/`jc3[<>/03D700081`<000V/YjbW[:N/Yj`06:L2000OYjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/Y`03[003SZN>00VW00>9YhT00jL00j:WY`02Y`08_ZL0RHj9SXT6Y`03`jO300;302bW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jc3[<>/`jc3[<>/`jc3[<>/`cD700081`0SL01`YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`00V/08ZN/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W`jO3
Yl<:Y`03SZNW00RW01BnYkjW``2WSZN>Yhj/YjbW`jO3Y`?302^W`jO3Yl>/`jc3[<>/`jc3[<>/`jc3
[<>/`jc3[<>/`jc3[<>/0j`3[0>/03D700081`<001F/YjbW[:N/YjbW[:N/YjbW[:N/Yj`02jL2000P
[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:L3[003SZNW00:W00B>YjN90jL01hVWRJNRYj80
1ZL01P29SXV>R@JW00?3Yl<00/<0;:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:O3[<>/`jc3[<>/`jc3
[<>/`jc3=@L000P704M`072W[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[029YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:O3Yl>W`jO3Yl>W``06Y`05SZNWYhh02ZL05;jW_ZNn0:N>YhjWSZbW[:O3Yl>W
0l<0:ZO3Yl>W`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`j`3[0>/0cH700081`<000n/YjbW
[:N/YjbW[:N/Yj`03jL09:bW0:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Y`:/00>WSXh0
0jL018jWYhT3Y`03RJN900ZW00L0RHj9SXV>00FW00?3Yl<00/<01JN/YjbW01B/01;3[<>/`jc3[<>/
`jc3[<>/`j`f1`0020L0Ag00L:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:M[L:bW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/Yl>/`jc3[<>/`jL300:W00V>YjNWSZN>Yhh03:L01;jW_P02SP0<YhjWS/>W
[:O3Yl>W0l<0:Zc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/`j`3[0>/0cH700081`<0
00Z/SZb>[8j/SZb>3ZL0:ZbW[:N/Yja[L:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Y`:/
00>WSXh00jL02XjWYhVWRJN9YhT:Y`0708V>RJN9SP05Y`03`jN/00;300FW[:N/Y`0D[00B`jc3[<>/
`jc3[<>/`jc3[<>/=PL000P70P00AJbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`0RG2/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:O3[<>/`jc3Yl>W0`02Y`09SZNWYhjWSZN>00bW00BnYkh00Xh03:N>
Yhk3Yl>W[:O3Y`?302Z/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/0j`3[<>/`jc3[0?30j`3[0<f1`00
20L20016SZb>[8j/SZb>[8j/SZb>[8j/SZb>[:N/YjbW[:N/Yg1[[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/Yjc3Y`:>00f9YhV>RJN9YhVWRJN900ZW00L0RHj9YhV>00FW00[3YjbW`jN/YjbW
5:`04L>/`jc3[<>/`jc3[<>/`jc303L700071`8003VW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[729YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:L00Z`02l>/`jO3Yl>W`jL300:W00V>YjNWSZN>Yhh0
3:L04[jW_P2>RJN>Yhk3Yl>W[:O3Y`?302V/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/0j`3[0>/0j`3
[0?30l<3``0g1`001`L2001H[8j/SZb>[8j/SZb>[8j/SZb>[8j/YjbW[:N/SZN>Yf^9JjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/`jN9SXVWRHj9YhVWRJN9YhVWR@RW00H0RHj9YhT6Y`04
[:N/Y`;300B/YjbW5:`04L>/`jc3[<>/`jc3[0>/0j`303L700071`8004JW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbWRG2/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W`jL30ZL02hjW
SZN>YhjWSZN>00ZW01:nYkh0SXVWSZN>[:O3Yl>W`jL2``0Z0jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/
`j`3[0>/0j`3[0>/0l<3[0>/=`L000L70P00F:b>[8j/SZb>[8j/SZb>[8j/SZb>YhjWSZN>YhjWSZM[
L:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[<>WRHj9SXV>RHj9YhVWRJN9YhT8Y`06
08V>SZN>0ZL02:bW[:N/YjbW0/<01:bW[:LD[00@1jc3[<>/`jc3[<>/0j`3[3P700071`8004JW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yja`RG2/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3
Yl>W[:L30ZL02hjWSZN>YhjWSZN>00bW012n08j9YhjWSZbW`jO3Yl>W0l<0::c3[<>/`jc3[<>/`jc3
[<>/`jc3[<>/`j`3``?30l<3[0>/0l<3``<h1`001`L20014[8j/SZb>[8j/SZb>[8j/SZN>YhjWSZN>
YhjWSZN>L6]`YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:L2[00BYhV>RHj9SXV>RJN9
YhVWRJN92:L05P1`SXjWSZN>[:N/YjbW[:O3YjbW[:LF[00>`jc3[<>/`jc3[0>/0j`h1`001PL3001E
SZb>[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N9L8VW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jO3
Yl>W[:N/YjbW`hjWSZN9YhjWSZN>YhVWSP03Y`03SZNW00BW00BnJhj90jL02HjWYjbW`jO3Y`03``0X
[<>/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[0>/0l<3``?30l<3/@>a0cP700061`8001V>[8j/SZb>[8j/
SZN>YhjWSZN>YhjWSZN>00:W02Z/Yja[RF^/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:L2[00BSXV>RHj9SXV>RHj9SXVWRJN92:L2L0:>01:WSZN>[:N/YjbW[:O3Yl>W[:LF[00=`jc3[<>/
`j`3[0>/0`0i1`001PL20019[8j/SZb>[8j/SZb>[:N/YjbW[:N/YjbW[:N/Yja`RG29YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/Yl>W[:N/YjbW[:N/Yl>>Y`02SP0;RJN>YhjWRJN9Yhh00jL00hjWY`04Y`04
_V^>R@JW00J/Yl>W`jL5``0U[<>/`jc3[<>/`jc3[<>/`jc3[<>/`j`7[0>/0k43/@>a0k43/@0i1`00
1PL20015SZb>[8j/SZb>[8j/SZb>[8j/SZbW[:N/YjbW[:N9Jg2W[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbW00:/01B>RHj9SXV>RHj9SXV>RJN9YhVWR@JW0W004HV>YhjWSZbW[:N/YjbW
`jO301R/00d70j`3[0>/0j`3[0>a03X700051`8001>C[9>/Tjb>[8j/SZb>[8j/SZb>00Z/00b9JhU[
[:N/YjbW[:L?[003Yjb/00N/00RW[:N/Yl>>Y`:>00n9YhVWRJN9YhVWSZN>Yhh01jL01f^>RJN>Yhh0
0ZL01ZbW[:O3Y`G302B/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>a0k43/@>a0k43/@<j1`001@L20016
[8j/SZb>[8j/SZb>[8j/SZb>[8j/SZb>[8j/SW1[L6^/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/Y`:/0XT00hj9SP03R@0BSXV>RHj9YhVWRJN9Yj:WXZNR0W02R@03YhjW00FW00N/YjbW
`jO301V/00/3[0>/0j`3[0>/0`0k1`001PL04P2>Thj/SZb>[8j/SZb>[8j/SPN/00^>[8j/JhU[RJN/
Y`0=[007YjbW[:N/Y`07[00<YjbW[:N/YjbW`hVW0Xh03hVWRJN9YhVWSZN>YhVWR@02XP0FYj:WXZM[
SXVWSZN>[:N/YjbW`jO3Y`?301N/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[002``060l<3/@>a0P<00iP7
1`0j1`001PL0AgFCMI=e[8j/SZb>[8j/SZb>[8j/SZN>YhjWSZM[L6]`YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW00:/0XT018j9SW02R@05SXV>RJL00XT00hBRR@02R@06Yj:WXZNR
0W02R@:>2:L02JbW`jO3[<>/``0E[0091j`3[0>/0k4303d700071`0NSZb>[8j/SZb>[8j/SZb>[8j/
SZb>[8j/SZa[RF^96:`05ZN/YjbW[:N/YjbW[:N/YjbW`hVWRHh3R@07SXVWRJN9XP03R@0>YhVRRJ:9
XXVRRHj9Yhh2Y`0:[:N/YjbW`jO3Y`?301V/`jc3[<>/`jc3[<>/`jc3[<>/`jc3[<>/00D300>a0`L0
?@L000P71Hh05I>>Thj/SZb>[8j/SZN>YhjWSZM[L002J`0aYjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/L8U`SV^9L002R@04SXV>J`>900Z4XXV9Q::9XXVR0XT01G29RJN900RW
00F/Yl>W``0K[0050j`3/@<0?`L000X701N/SZb>[8j/SZb>[8j/SZb>[8j/SZb>R@02JbB/00bW[:N/
YjbWSXV>RG03R@03SXV900^900ZRRJ:9XW29SZN90Z86Y`0P_ZNWYkjW`jc3[<>/`jc3[<>/`jc3[<>/
`jc3[<>/`j`30`03/@<703l7000;1`0DThjCSY>>ThjCSZb>[8j/SZb>[8h3J`0cYjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yja`RG29JhU`00B900^4RHV4RHB9Q8V4R@040005
JhA[RF/00XT00jNRY`0:Y`03XZb/01R/00<3/@<0@@L000d7016>[8j/SZb>[8j/SZb>[8j/SP03Jb:/
00bW[:N/YjbW[:N/Yhh2R@04L8U`SPB900=[RHT01XT01V/00029Q0F900>>Yhh00ZL02[jW`jO3Yl>W
`jL3``0J[<>/`jc3[<>/`jc3[<>/`jc3[<>/0j`3[0=21`003`L04i>>ThjCSY>>[8j/SZb>J`07[6/0
1:`0;JN/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yja`RF]`J`03R@03JhU[00J9
1P002XV4RHB9Q8V>YhT<Y`04`jO3_QV/@`L0014700f>[8j/SZb>[8j/SV/7026/01FW[:N/YjbW[:N/
YjbW[:N9L8U[RG000hT00f^9R@04R@8000A[06][1XT04ZN>YhjWS/>W`jO3Yl>W`jO3Y`?301Z/0j`3
[0>/0j`3[0>/0j`3[0>/0j`3[0>/0d87000B1`0;SZb>[8j/SZb>J`007j`05jN/YjbW[:N/YjbW[:N/
Yjb>Yf^9Jg1[00F900A[RHV910005V^4Q8V4RHB9Q8V>YhjWSZN>[:N/Yj`5Y`07[:O3Yl>M``0<[008
YjbW[:N/Yl<5[407000D1`07[8j/SZ`0J`0N[00=YjbW[:N/YjbW[:N/Y`05[00;YjbWRG29L8U[RF/0
18T400=[01f4RHB9Q:N9YhjWSZN>[:N/Yl>W`jN/YkjW`jO3[003``0MQ<>/0j`3[<>/`jc3[<>/`jc3
[<>/`jc3Yl>/`j`0?@L001P702VW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
Y`07[00ASZN>YhjWJhU[RF]`JhU`RG0010001XA[Jf^9J`:900bWRHV>YhjWSZN9Yhh<Y`05`kk3Yl<0
1J`00l>/[00B[004Yl>/[3/7000E1`N/0ZL01jbW[:c3[<<00Z`01jN/YjbW[:L00j`01JN/YjbW00V/
012W[:N/YjbW[:N/YjbW[:N90g03R@03JhV9008001A[08U[RF^9Q:N>YhjWSZN>YhjWSP:W00>/Yj`0
0jL02<>W[:O3Yl>n1L<08J`3[0>/0j`3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Y`0g1`004PL01:N/
Yj`3Y`0D[:N/YjbW[:N/YjbW[:N/YjbW[:L=[00:YjbW[:N/YjbW[0RW00Z>Yf^9JhU[RF^910001f^4
JhA[RF/00XT05ZN9YhVWRJN>YhjWSZN9XXVRRJ:4XX@2XP09Yj;3Yl>W`jO301VW00O3Yl>W`jO303L7
000>1`11[:N/YjbW[:N/YjbW`jO3Yl>W`jN/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YhT00V/01729JhT300A[01J9JhV4RHjWSZN9YhVWRJN>YhjWRJN90ZL02/>W`jO3Yl>W
_ZL2_P?3026n1jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jL0=`L000d700=`YjL0=:L048jW
SZN>YhjWSZM[L01`JhT400=[00>4JhT00hT04jN9YhVWRJN9YhVWJjN9YhV/Yj`01JL02ZbW`jN/X[k3
_Z82_QRW00O3Yl>W`jO303L7000;1`A`02ZW[:N/YjbW[:N/Yl>W`jO3YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/Yj`EY`08RF^9JhT0J`06J`B900^WSZN>YhVWRJN9Y`02Q00gRJbW[:N/YjbW`jc3[<>/0j`3
[<>/`kk3_/>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Y`0g1`002PL01De`CG0003FW00f>YhjW
SZN>YhjW071[00D01V/5R@0PYhV>RJN9YhVWRJN>[8j/SZbW[:N/RJN9_Zc3[<>/XXT2Y`04_ZNn_QJW
00O3Yl>W`jO303L7000;1`A`02X0[:N/YjbW[:N/Yl>W`jO3YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
Yj`EY`@000=[00001F/5R@0ASZN9YhVWRJN9YhjWSZ:9XX@00Z801:NRYj84Y`0;`jNRQ::n`jO3_/<0
0[h07JO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W03L7000;1`05L4e`04d0=:L038jWSZN>YhjW
SZL0J`D000Q[06][RF^9J`:901:WRHj9YhVWRJN9YhjWSZN>Yhh8Y`06_ZO3Yl>W0l<01Zc3[;jW_QFW
00O3Yl>W`jO303L7000<1`=`02X0[:N/YjbW[:N/YjbW`jN/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
Yj`EY`06R@1[06/016/00hU[R@03R@08YhVWRJN9YhT2Y`05[:N/Yj`00ZL00hj/SP08Y`04`jO3Y`;3
00<3`l<00/<07kk3_/>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jL0=`L000`700E=L01=000cY`0;
SZN>YhjWSZN>Yf/010002V/0J`1`JhU[RF/2R@0FSXV>RJN9YhVWSZN>YhjWSZN>YhjWSP>W012/YjO3
Yl>/`jO3Yl>W`jNn5JL01l>W`jO3Yl<0=`L000d70W00:P1`YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbW[1FW0P000f/0J`03J`06RF^9JhU[0XT028j9YhVWRJN91ZL03JbW[:N/YjbW
[:N/Y`<00j`00`?3``06``0OYl>n`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Y`0g1`003@L01G00
C@1=03:W00V>YhjWSZN>Yhh010002V/0J`1`Jg1[L6/6R@04YhVWR@FW00f>YjNWSZN>Yhj/SZb>00BW
00o3Yl>/`jc3Yl>W`jNnYkh04jL01/>W`jO3YcP7000>1`0[L4e`CJbW[:N/YjbW[:N/YjbW[:N/YjbW
[:N/YjbW[:O3YjbW[:N/YjbW[00CY`061`1[06/016/03W1[RF^9Jhj9SXVWRJN92:L02JbW[:N/YjbW
[002Y`07SZNW_ZO3[003``0R[<?3`jO3_/>n`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`cP7000>1`05
04d0C@00<JL03hjWSZN>Yhh7SV/0J`1[0004J`0>L6]`Jhj9SXV>RHj9YhT;Y`03SZN>00>W00>9SV/0
1:L03L>/`jc3Yl>W`jO3_/<04jL01/>W`jO3YcP7000?1`0ZCG1=L:N/YjbW[:N/YjbW[:N/YjbW[:N/
YjbW[:O3Yl>W[:N/YjbW[:N/4JL01PNWY`1[00A[00I`Jg1[RG02R@08SXV>SZN>Yhh6Y`08[:N/YjbW
[:L3Q003SZN>00>400>WXZL00l<08Zc31l>W`jO3_/>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl<h1`00
3`L01Dd0C@1=032W00f>YhjWSZN>YhjW06/000I[00>9JhT00hT028j9SXV>RJN>1:L00jbWY`05Y`06
Q8V9L:N>0P02Y`0<`jbWQ9fW`jO3Yl>W1<<BY`05`jO3Yl<0>@L0010702U`CG1=[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW`jO3YjbW[:N/YjbW[00EY`800V/0471[L6^9L8U`SXV>RJN>Yhh6Y`08`jN/YjbW
[:L2J`:W00>nXZL00ZL03HjnYl>n`kk3[<>W`jL01L<05jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W03T7
000@1`0504d0C@00;jL02XjWSZN>YhjW06/200E[00B9JhU[0hT02G2>RJN9YhVWSP04Y`03[:NW00BW
00T0SXj>RJN>YhT01:L03;jW`jO3Yl>W`jO3[0;34ZL01L>W`jO303T7000A1`0XCG1=L:N/YjbW[:N/
YjbW[:N/YjbW[:N/Yl>W`jO3Yl>W[:N/YjbW[1>W1002J`06L6]`L8U`1HT01HjWSZN>00JW0133YjbW
[:N/YhieYhjWSZN>0ZL01JbW`j`300;300FW`jc3[007``0EYl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W03T7
000B1`0404d0CBjW00N>YhjWSZN>008000A[000016/01XU[RF^9L0:900fWRJN9YhjWSZN>[:N/00FW
00B/Yhie1Xh2Y`03[:O300;300b/`jO3Yl>W`jc3_/<AY`05`jO3Yl<0>@L0018702M`CG1=[:N/YjbW
[:N/YjbW[:N/YjbW`jO3Yl>W`jO3YjbW[:N/Yj`04ZL5009[00I`JhU`RG04R@:>00>WSZL01JL02jbW
`jN/YjbW[:N/00>>00FWSZN>Y`03[0050j`3/@<00/<01Jc3[<>/00G301FW`jO3Yl>W`jO3Yl>W`jO3
Yl>W`jL0>@L001<700A=04e=;:L51`03SZN>00H00V/04XU[RF^9L8U`SXVWRJN9YhjWSP:W00>/Yj`0
1ZL00hV>SP05SP03YjbW00B/00k3[<>W`jO3Yl>/`kk3_Q:W00?3Yl<0>@L001<700A=L4e`2JL06JbW
[:N/YjbW`jO3Yl>W`jN/YjbW[:N/Yj`04ZL01@1`06/0009[0W000hU`R@05R@04YhjWSPJW00V/Yl>W
[:N/Yj`01:L00hjWSP02SPB/0`<02L<3[<>/`jc3[005``0EYl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W03T7
000D1`0404d0L2bW00V>YhjWSZN906/00P002V/0J`1`JhU[RG02R@:>00RWSZN>YhjWSPBW00>/YjL0
1:L00hjWSP04SP07Yhj/Yl>/0`02``0;[<>/[:O3Yl>W`j`00/<BY`03`jL703T7000E1`03CG1=00JW
01^/YjbW[:N/YjbW[:O3Yl>W`jN/YjbW[:N/Yj`03jL028j9Yj9`06/00V/018U`RG04R@:>00BWSZN>
1ZL05:bW`jN/YjbW[:N/Yjb>YhjWSZbW0Z`2``0:0l<3[<>W`jO3Y`C301D3Yl>W`jO3Yl>W`jO3Yl>W
`jO3Yl<0>PL001D700A=071=:JL018jWSZL2SP09L8U`Yj9[06/0009[00I`Jg1`RG02R@:>00RWSZN>
YhjWSPBW00>/YjL00jL00jb>Y`04SP0ERJNW[:O3Y`>/`jc3YjbW[:O3Yl>/00?34jLj1`005PL00g1=
L005Y`0K[:N/YjbW[:N/YjbW`jO3YjbW[:N/YjbW[:N/00jW00m`SXV>RHh0J`1`JhU`RG0018T2SP03
YhjW00:W00>>Yhh00ZL03jbW`jN/YjbW[:N/SZb>Y`03SP05[:O3Yl<00P<2``07[<>/`jc3[004``0E
0l?3Yl>W`jO3Yl>W`jO3Yl>W`jO303X7000F1`0404e=L2RW00V>YhjWSXU`SW000XT016/00002J`9`
0129L8U`SXVWSZN>YhjWSZN>1:L01jbWYjN/Yj`00ZL4SP0:RHj/[:N/Y`>/``:/00VW[:N/Yl>W`j`0
0l<CYcX7000G1`03CG1=00BW01^/YjbW[:N/YjbW[:O3Yl>W[:N/YjbW[:N/Yj`03JL00hi`SP03R@04
SP1`009`00>9L8T00hT4SP08YhjWSZN>Yhh2Y`0?[:N/Yl>W`jN/Yjb>[8jW00>>00G3Yl>/0`02``09
[<>W`jc3Yl>/00C301D3[<>W`jO3Yl>W`jO3Yl>W`jO3Yl<0>PL001L70Td00g1=Y`0VY`04SZN9SP9`
00I[RG2906/30006L6^9L8U`0XT038j9YhjWSZN>YhjWSPBW00V/YjbW[:N/Yj`00Xh00hV>SP02Y`09
[:N/[0>/`jN/00FW00C3Yl>/0l<CYcX7000H1`03L4e`00>W01^/YjbW[:N/YjbW[:O3YjbW[:N/YjbW
[:N/Yj`03:L02f]`Jg29SXU[06/0009`00J9L8V9SXT4SP08YhjWSZN>Yhh2Y`0;[:N/YjbW`jc3Yj`0
0ZL3SP>/00BW`j`30/<02jO3YjbW`jO3Yl>W00G301:W`jO3Yl>W`jO3Yl>W`jO3Yl<j1`0060L0101`
CG0VY`07SZN9L6]`J`03R@0?06/006^9L8U`RG2>RHj900>>00N9YhVWRJN900:W00V/YjbW[:O3Yj`0
0ZL00hV>R@02SP07[:N/[<>/``02[0JW00FRYjO3_P03`a:W>`L001T7021=L4f/YjbW[:N/YjbW[:N/
YjbW[:N/YjbW[:N/YjbW[0bW015[L6]`RHj9L01[08U`RG2>R@04SP0:YhjWRJN>YhjWSPJW00VnYl>/
`jO3SZL00hh01jbW[:O3[0<00/<7Y`070j`3[0?30`02``0A1l>W`jO3Yl>W`jO3Yl>W`jL0>`L001T7
00A`CG1=9JL02XjWJg1[L8j9L8T3000:L6^9JhU`RG2>R@>>00V9SXVWSZN9YhT00jL4[003YkjW00>W
0Xh01HV>RJbW00:/00G3[<>//@02[003Tjb900BW00FnYl>/``0BYc/7000J1`0OL4e`YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[007Y`07[:NWYhU[L002SP09RHj9J`1`070000B900ZWRJN9YhVWRJN9
0ZL00jb9Y`03Y`04_ZNnY`:a00S3[<>W[8jWSP:W00F/Yl>/0`02``03Yl>W00BW01[3Yl>W`kjR_Z:W
`jO3Yl>W`jO3Yl>W`jO3Yc/7000J1`04CG1=L2:W00j>YhjWJg1[1g29L8T0L08000E[L029L00:R@:>
00JWSZN9Yhh4Y`05`jNWYl<00ZL00hV>R@02SP05[:N/Yj`00/<01j:WXZN9YhT01:L01Ki[RKk3016W
?0L001/701Y=L4f/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[0NW00>/Yj`01JL01HU[L:N>00>900=`0700
1G001HU`[8j/00>>00FWRJN>Y`02SP03Q::W00:W00J/YjbWRH@2``06YhVRRJN>0ZL01ZbW`jc3_PJW
0X@00j:4Q004``0B_/>W`jO3Yl>W`jO3Yl>W`jO3?0L001/700A`CG1=8JL02HjWSZM[08jWL002R@0:
SXU`071[L6^>L0>>00Z9SXV>RHj9SXVW0hh4Q00EXXA[YkjW`jN>RJN9YhV/YjbW[:Nn00VW00B/Yl>/
0/<00kjW[00?Yc`7000L1`0GL4e`YjbW[:N/YjbW[:N/YjbW[:N/Yj`01jL01JbW[:N/00BW00][Y`2/
Yhj9SXV>0005L0Z>00N9YhVWSZN>00:W012/YjbW`jO3Yl>W_Z:WRJN90ZL01l>W`jO3Ykh01JL01:bW
`jL5``0B_/>W`jO3Yl>W`jO3Yl>W`jO3?0L001`700A=L4e`8jL01G2>08jW00>900Z>R@00L6]`JhU`
28T028j9YhVWRJN91jL02Zc3YjNRYhVWRHh9Y`03_Z:W00NW00C3Yl>W0/<00jbWY`05Y`03[:NW00FW
?0L001d7019=L4f/YjbW[:N/YjbW[:N/Yj`7Y`09[:N/YjbW[:N/00BW00A[Yg2W0Xh01hV>RHh0L000
1702R@0>SXV>RHj9YhVWSZN>Yhh8Y`03`jO300:W00FRYhVWR@02Y`0;`jO3YjbW`jNnYkh01JL01<>W
`jL2``0A0kjMYl>W`jO3Yl>W`jO3Yl<0?0L001d700A`CG1=8ZL0172>L6/3R@03Jhj9008000Y`Jg1[
RG29L8i`18T038j9SXVWRJN>YhjWSPFW00NRYj:RRJN900:W00?3YjL01ZL01j:WXZNRYj800ZL01;jW
`jL2``03XXBW00fW?0L001h701]`CG1`[:N/YjbW[:N/YjbW[:N/YjbW[:N/Yj`02ZL01HVWL71[00:9
00H0RHh0L004L008R@29Q8V4SXT2SP06YhjWSZN>2:L01<>WYhT2XP0JYj:WYl>W`jO3Yl>W`jO3Yl>R
_ZNnYl>W`jL2``0AX/?3Yl>W`jO3Yl>W`jO3Yl<0?0L001h700@0L01`8JL2R@06L6^>L8i`0XT02700
L6]`Jg1[0W0038U`SXV>RHj9YhV>R@:>00BWSZN>1:L00jb9Y`05Y`0;[:N/YjbW[:N/Yj`01jL02<>W
`jO3Yl>n0/<=Ycd7000O1`0JCGE`MJN/YjbW[:N/YjbW[:N/YjbW[:N/Yj`:Y`0=RJM`L8V>RJL0Y`1`
0005L0B>00f9SXVWSZN>YhBRSZN>00JW0233XZNW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W_ZNnY`C3
00h3Yl>W`jO3Yl>W`jO3Ycd7000O1`05L4e`CG007jL01Xj9L6^>R@>>01n9L01`071[L7F>L7E`SXV>
RHj9SXV>RJ:9XXVRRJN900:W00>/YjL00jL07JbW[:N/YjbW[:N/YjbW[:N/Yl>W`jO3Yl>/`jc300jW
?@L0020717001jbW[:N/Yj`02jL00jbWY`0:Y`0=1hi[L8VWSZN>Y`1`0004L004SXV9L0:900R>RJN>
YhjWSPJW02B/YjbW`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jc3[<>/`j`4``0>1jO3Yl>W`jO3Yl>W
`jLm1`008@L0171=L4dOY`03RG1[00>>00V9SXi`0700L6/00W0038U`RG2>L8j9SXV>R@:>00BWSZN>
1ZL06ZbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW0Z`02<>/`jc3[<<71ZL01jbW[:N/Yj`0?@L00247
17000jN/Y`0:Y`05[:N/Yj`00jL01JbW[:N/00FW00b9Yhj>YhjWSZL0L004L00>SW29L8j9SXVWSZN>
Yhh2Y`0X[:N/YjbW[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>W`jO3Yl>/`jc3[0C300d3Yl>W`jO3Yl>W
`jO303h7000R1`04CG1=L1ZW00N>YhjW08U`00>>00V9SXi`CG00L6/00W001XU`SW2>L0:900B>RHj9
0Xh09:N>Yhj/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/Y`:/00K3[<>/`j`2``BW00R/YjbW
[:N/Ych7000R1`E`00jW[:N/YjbW[:N/YjbW[12W00B9Q6/00Xh00jc3L00200A`03R9L8V9SXV>RJN>
YhjWSZbW[:N/YjbW[:O3Yl>W[:O3Yl>W`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[0C300d3Yl>W`jO3Yl>W
`jO303h7000S1`04L4e`01VW00^>YhjWSW1[SW2>R@02[0<000=[L6/00W002XU`RG29L8j9SXT4SP0/
YhjWSZbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW[:N/YjbW`jc3[<>/`j`2``bW?PL002<7011`MG1`
CJbW[:N/YjbW[:N/4jL2L003SXV/00:a0P04L007JhV9SXV>R@02SP0`YhjWSZN>[:N/YjbW[:N/YjbW
`jN/Yl>W`jO3Yl>W`jO3Yl>W`jO3[<>/`jc3[<>/0l<03[k31l>W`jO3Yl>W`jO3?PL002@700A=L01`
5ZL02XjWSZN>YhjWJdd2SP03TiP0008000A`06/00V/01XU`RG29L0R>02nWSZN>[:N/YjbW[:N/YjbW
[:N/YjbW[:N/YjbW[:N/YjbW[:O3[<>R`jO3[<>n``0:Ycl7000T1`05MDe`CGD02JL00jbWY`0EY`:9
0P001W1[L6]`J`N902B>YhjWSZN>[:N/YjbW[:N/YjbW[:N/YjbW`jO3Yl>W`jNnYkh3Y`04_ZNnQ0:R
0ZL00l>/``04``0;_PNW`jO3Yl>W`jL0?`L002D700A`CG1`5JL03XjWSZN>YhjWSZN>L6^90`002V/0
J`29L8U`RG02R@0BSXV>RHj9YhjWSZN>YhjWSZN>1X@02:NRYj:WXZNR3:L02/>W`jc3Yl>W`j`2_PVW
?`L002D700E=L4eeL006Y`03[:N/01JW0f/20005L6]`Jg001hT02:N>YhjWSZN>3:L02jbW[:NnYl>W
`jO300>W00K3Yl>W`jL2Q008`jb4[<>/`j`6``0:_ZO3Yl>W`jO3Ycl7000U1`05L4e`L7D05ZL038jW
SZN>YhjWSW00J`<001i[06/0RG29L8U`SXV>RJN9YhjWSZN>YhjWSZN>Yhh8Y`03J`1[00>W00>nY`00
1Z803HBRRJ:W`jO3Yl>W`jL00/<00kjW_P06Ycl7000V1`05L4eeL7D00jL01JbW[:N/01JW0f/200A[
00=`RHT01HT02:N>YhjWSZN>3:L02:bWRHB9JhBR0ZL01kjW`jO3L8T01J802[jW_ZNnYl>n`j`6``08
1jO3Yl>W`jLo1`009PL01De`L7E`01FW00b>YhjWSZN>Yhi`06/50006J`29L8U`0XT058j9YhVWSZN>
YhjWSZN>YhjWSZN>0jL00f^9R@02R@:R00BWXZN40ZL00hU`R@06Y`0;XZNR1jO3Yl>/`j`00l<2_PJW
?`L002L700i=MG1eL:bW[:N/YjbW[1BW00<0J`000P0016/0Jf/6R@0:YhjWSZN>YhjWSPbW00>/YhT0
0hT05ZN9Yj:nYkjW`g1eL;jW_ZNnYkjW_ZL;``07Yl<7`jO3Y`0o1`009`L2L004MG1`CA>W00Z>YhjW
SZN>Yhj91@001F/0Jg29009[01F9SXVWRJN>YhjWSZN>YhjWSZN>Yhh01:L028U[RG2WRJNR1:L00kjW
Y`02L0FW00b/YjbW`jO3[<>/`j`5``03YkjW00:W@0L002P700=eL7000W001ZN/YjbW[1FW0V/4000G
J`1`JhU[RF^WRJN>YhjWSZN>Yhj/Yj`00ZL00hjWY`02Y`04[:N/L0B900NWRJNW_ZL300;300nWL033
Yl>W`jO3Yl<7`j`02l<01:O3Yl=01`00:0L01Te`CG1=L1:W00R>YhjWSZN>Y`L000=[06/00f/05Xj9
YhVWRJN>YhjWSZN>YhjWSZN>Yhh4Y`06RG29L:N>1:L01jc3`jN/07001JL03ZbW`jO3Yl>/`jc3[<>/
0l<2_P03Y`NW0407000Y1`E`00F/YjbW[00EY`08R@0006/0J`02J`0>L6^WRJN>YhjWSZN>Yhh2Y`0A
[:N/SZN>YhjWSZbW[72>RHh00ZL05hj/Yl>W``?3Yg1[`jO3Yl>W`jO3Yl>/00S300L3`l>W`jO30407
000Y1`06L4e`CG1=4JL02HjWSZN>YhjWJ`020006J`0006/00V/068j9YhVWRJN>YhjWSZN>YhjWSZN>
YhjWSPBW00J>L8i`Yhh2Y`05[:N/[0<00j8050NW[:N/Yl>W`jN/Yl>/`jc3[<>/1L<00jNnY`101`00
:PL5L004YjbW[1BW00@0L01[0`004V/0L01[SZN>YhjWSZN>YhjWSP:W00J/YjN>Yhh3Y`0@SZbW[:N>
L8V>[8k3Yl>/0`BR00e[`jO3Yl>W`jO3Yl>/00K300D3``?30`02``03Yl<703l7000Z1`06CG1=L4e`
4:L02HjWSZN>Yhj9J`020003J`0000<001Y[RHj9YhjWSZN>YhjWSZN>YhjWSZN>YhjWSPFW00E`RG2W
SP02Y`F/01NW`jN9YjbW[:N/YjbW[:O3[<>/`jc3[005``03Yl>W0407000[1`E`00>/YjL05:L08f]`
06/0J`1[06/0YhjWSZN>YhjWSZN>Yhj/Yjb>YhjWSZN>00:W01:/YjbWRG29SZb>[:O3[<>/`j`2R@0<
`jO3Yl>W`jc3[<<71<<01`?30l<3``<00/<00jO31`0o1`00;0L01De`CG1=00nW00R>YhjWSZN9J`<0
00=[00000`0098j9YhjWSZN>YhjWSZN>YhjWSZN>YhjWSZN>[:N/Yja`RF^WSP:W01b/YjbW[:N/JhjW
[:N/YjbW`jO3Y`N/`jc3[<>/1<<00`?3``111`00;0L5L1FW02Y`J`00J`1[06/0YhjWSZN>YhjWSZN>
Yhj/YjbW[:N/SZb>[8j/YjbW[:L2L00<[8jWYl>/`jc3Yl>W0XT03<>W`jO3Yl>/`jc3[0;300`3``?3
0l<3``?30jM11`00;@L01G1=L4e`00jW00J>YhjWSZL2J`L002^WRJN>YhjWSZN>YhjWSZN>YhjWSZN>
YhjWSZN>[:N/Yja[L8jWSZbW[:O300>a01a`RJbW[:N/YjbW`jO3Yl>/`jc3[<>/`j`3``?3@PL002d7
0g000gE`Y`0BY`03Jg1[00L002F>YhjWSZN>YhjWSZN>[8j/YjbW[:N/Yjb>[8j/YjbW[:M`RJN>00:W
00?3[<<00/<07[5`L8VW`jO3Yl>W`jc3[<>/`j`3``?30l<3``?30d87000^1`05CG1=L4d03JL01HjW
SZN9009[1`07SP0JYhjWSZN>YhjWSZN>YhjWSZN>YhjWSZbW[:L2L00<RHjWSZbW[:L3[71[0W001ZbW
[:N/Y`:/00[3[<>/`jO3[<>/0/<01`>/0l<3``<0@@L002l700A`MG1e4:L018jWJf/6000MSZN>YhjW
SZb>[8j/SZb>[:N/YjbW[:N/YjbW[8h01J`3R@06[:N/YjbW0l<04F^9Jg2/`jc3[<>/`jc3[<>/00;3
00`3``?30l<3``?30l=11`00;`L01G1=L4e`00bW00B>YhVW0V/600:>00>9Yhh00Xh05ZN>YhjWSZN>
YhjWSZN>YhjWSZN>Yhh2Y`05[:N/YhT00W002J:WXZNR`jO3C@03L007JjbW[:N/Y`02[004`jc3[0?3
00f/`j`3``?30l<3``>W0407000`1`04MG1eL0fW00>>Yhh00V/5000OSZN>YhjWSZb>[8j/SZb>[8j/
SZbW[:N/YjbW[:N/Y`05[0:900=`RH@00X@01Yf4Q8BR``:90133[<>/`jc3[<>/`j`3[0>/0`<01L<3
``?300<30/=01`00<0L01De`CG1=00^W00>9YhT00f/4000ZRHj>YhjWSZN>YhjWSZN>YhjWSZN>Yhj/
SZb>[8j/SZbW[:N/SXU[L02>18T03F^4Jl=`RJN/YjbW[:L00Z`05<>/`jc3[0>/0j`3``?30l<3``>W
@0L0034700A`MG1e3:L018jW06/4000QSZN>Yhj/SZb>[8j/SZb>[8j/SZb>[:N/YjbW[:N/YjbW00>/
01nW[8V9L6^9SXV>RJN9`g2>RL>/`jc3[<>/0jc3[<>/00<300F/0l<3``060`03``L703h7000a1`05
L4e`CG002ZL00hVW0002J`<002>9SXjWSZN>YhjWSZN>YhjWSZN>YhjWSZb>[8j/SZb>[8j/Y`04[005
L6][08h00XT01f^>RHk30g002:`05/>/`jc3[<>/0jc3[0>/0l<3[0>/0l=01`00<PL017E`MG09Y`03
SZN9009[0`008HV>SZb>[8j/SZb>[8j/SZb>[8j/SZbW[:N/YjbW[:N/Y`05[004YjbWR@9`00N>Yhie
SZb>0083012>[<>/`jc3[<>/`jc3[<>/0`<00j`30`090d07000b1`04CG1=L0BW00N>YjNWRJN900=[
0P00:W2>SZN>YhjWSZN>Yhj/SZN>YhjWSZb>[8j/SZb>[8j/SZb>[:N/YjbW[09`00^9SXj>L8jC[<=`
C@08[00F`jc3[<>/0j`3[<>/0j`3``>/0j`3/D07000c1`A`2:L018VRJf/2000QRHj>[8j/SZb>[8j/
SZb>[8j/SZb>[:N/YjbW[:N/YjbW00>/01:>[:N/YjbW[72/SZb>[7FaNP<3L00A`jc3[<>/`jc3[<>/
`k43[<<02`=11`00<`L01700L003Y`04SZNWY`:91000;g29RJN>YhjWSZN>[8j/SZN>YhjWSZb>[8j/
SZb>[8j/SZbW[:N/SZbW[:N/YjbW00>/00Ja[;6WSTd2L0N/01G3[<>/`j`3[<>/0j`3[0?30j`3/@<0
@@L003@700A`CG1=1JL00hVR0002J`0R08V>RJb>[8j/SZb>[8j/SZb>[8j/SZb>[8j/SZbW[:N/Y`F/
00N>[8j/YjbW00J/00C3SZN>0Z`04Xi``jc3[<>/`jc3[<>/`j`3[0T300>a0k40@@L003@700P0L01`
RJN9Y`:903<0Jf/0L8U`[8j/SZN>Yhj/SZb>YhjWSZN>[8j/SZb>[8j/SZb>[8j/YjbW[8j/SZb>[:L0
1:`01P>/SXj/SP:/0Td02JbW[:N/Yl>/``03[00?0j`3[0>/0k43/@>a0k430447000e1`04CG1`L0>W
1F/08hV>RJb>[8j/SZb>[8j/SZb>[8j/SZb>[8j/SZb>[8j/YjbW00N/00R>[8j/TjbC``>/00C3SZb>
0Z`04/>/`jc3[<>/`jc3[<>/0j`3[0L300>a0k400[511`00=@L01g1=L4fWR@0016/0;G29SZb>[8jC
SZN>Yhj/SZN>Yhj/SZb>[8j/SZb>[8j/SZb>[8j/YjbW[8j/SP03[00ASZbC[;73[<>>[8k3[<>n`h@0
1:L04hV4WJc3[0>/0j`3[0>a0k43/GX0@`L003H717001Hj9SXV>00:9026>[8j/SZb>[8j/SZb>[8j/
SZb>[8j/SZb>[8j/SZbW[:L03J`029>/[;6C0j`30Z`029>/[0>/0jbW0hh02XA[Q6^R[0>/0j`60`>a
A0L003H700==L4d00W00:8U`RG29SY>>[8jCSY>>[8jCSZb>[8j/SZb>[8j/SZb>[8j/SZb>[8h=[006
TjbC[9>a0j`2T`:/00[3[4d0SXV>L8j90Z802jN/0j`3[0>a0k4300:aA@L003L70W009@1`L8j9[9>/
TjbC[8j/SZb>[8j/SZb>[8j/SZb>[8j/SZb>[8h04Z`02[6//Jba[0>/0gD2T`:/00?31g000Xh00hV>
R@02SP04`jL3/@L30[561`00=`L00g1=0003L0NC01n>ThjCSY>>ThjCSZb>[8j/SZN>Yhj/SZb>[8j/
SZb>01J/00jWSY>>Tjc30:N9SXV>L0:W00W3[0>/0k43/@<00[571`00=`L2L00309>C00:C01j/Tjb>
[8j/SZb>[8j/SZb>[8j/SZb>[8j/SZb>[8hJ[00<`jc3SZb>[8h3[0>/0hh2R@05Yl?30k401@<2/DP7
000h1`0B07EeThjCSY>>ThjCSY>>ThjC1Hh04Zb>ThjCSY>>[8j/SZb>[8j/SQN/00J>YhjWSZL2/@0>
[;69TgZ9[;5j0j`3/@<2/@03[;4704P7000h1`Ae00JCSY>>Thh2T`0H[9>/Tjb>ThjCSZb>[8j/SZb>
[8j/SZb>6Z`05[6>[8j/Yjba0j`3[0>/0j`3[0>a0k430`03/@L704T7000i1`0IME9eDY>>ThjCSY>>
ThjCSY>>ThjCSY>>T`03SP0<[8j/SZb>[8j/SZb>5:`020=e[8j/S[471:`02k6//J`3[0>/0j`300>a
C0L003/7025eNWEjMJbC[9>/SY>>ThjCSY>>Thj/SZb>[8j/SZb>[9<03Z`00k6/[003[00N/Jba[0>a
TgFCSP?30j`3[0>/0j`3/@>a0k43/@>aC@L003h7011BNU9jDY>>ThjCSY>>ThjC1Hh03:b>[8j/SZbC
[9>/TaF/00Ra0gEeSY>/0`>/00^a[;6//Jba[0>/0`03/Dh7000o1`0QMGYeNWEjSY>>ThjCSY>>ThjC
SZbC[8j/SZbC[9>/TjbC00J/012a[;6//Jba[;6//Jba[0>/0Xh00i>>T`020`0>[0>/0j`3/@>a0k43
/@<2/Dh700111`06NWEjMGYe2Xh059>>ThjCSY>>[8j/SZb>[9>/TjbC3Z`00k6/Y`03SP0>Thh3[;6/
/Jba[;6//J`3/@03V;4704h700141`0JNWEjMGYeThjCSY>>[8j/SZb>[9>/TjbC[9<6[00F/Jba[;6/
/Jba[;6//J`3[0>>Yhj/SP:/00>a[;401K400`>a0`04/E0700151`Ie18h059>>ThjCSY>>[9>/TjbC
[9>/TjbC3Z`03;6/[8jWSZN>0i>/T`:/00Ja[;6//J`4/E8700171`0KMGYeNWEjSY>>ThjCSY>>[9>/
TjbC[9>/TjbC00:/00ja[;6//Jba[;6//Jba[0:a00QeSZb>V0>C[0:a00D3/K6a0`05/E<700181`Me
00JCSY>>Thh4T`0>[9>/TjbC[9>/TjbC[9<8[003/Jba00:a00EeTgFa[002T`0;[9>a[;6//IRaV;40
E@L004X702JCMI=eTgFCSY>>[9>/TjbC[9>/TjbC[9>a[;6//Jba[;6//Jba[0:a00<3/GD00Y<01K6/
TjbC00RaE`L004/727D>T`06[9>/TjbC2Z`01;5eSWD2T`08[9>HTk6C/I<3NUT7001=1`03MGYe009j
01bCNY>C[9>/TjbC[9>aTk6C/I>aTk6//Jba[;6/0[403jb>[9>/Tk6C/I>HNYQjO`1L1`00C`L029=j
TgZCNY=j29<03ZbC[9>HTk6C/I>aTk6C0k403Df>SZb>V7ZH1iQjV7X0GPL0054702UjTgZCNY=jTgZC
NY=j/I>aTk6C/I>aTk6C/I<3/@>aMHj/1`=j1gZHNP1Q1`00E0L05WZCNY=jTgZCNY=jTgZCNY=jTgZC
1i<3NP09SWE`CGYe0jbH06D7001J1`0>V7ZHNYQjV7ZHNYQjV7X61`9B0WX00iQj1`1V1`00f`L00=/7
003K1`00f`L00=/7003K1`00f`L00=/7003K1`00f`L00=/7003K1`00f`L00=/7003K1`00f`L00=/7
003K1`00f`L00=/7003K1`00f`L00=/7003K1`00\
\>"],
  ImageRangeCache->{{{0, 218.812}, {287, 0}} -> {0.132682, 0.0145566, 
  0.0034357, 0.0034357}},
  CellTags->"DXFFiles"]
}, Open  ]],

Cell["Now it is easy to import 3D models.", "Text",
  CellTags->"DXFFiles"],

Cell[BoxData[
    \(MVClear[]; MVLoadDXF["\<cow.dxf\>"]; \ngrCow = MVGetGraphics3D[]; \)], 
  "Input",
  CellLabel->"In[10]:=",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[MVRotate3D[grCow, {1, 0, 0}, \(-\[Pi]\)/2], Boxed -> False, 
      RenderAll -> False, ViewPoint -> {1.540, \ \(-2.844\), \ 0.995}]; 
    \)\)], "Input",
  CellLabel->"In[44]:=",
  CellTags->"DXFFiles"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .76631 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0275174 1.02752 -0.0995321 1.02752 [
[ 0 0 0 0 ]
[ 1 .76631 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .76631 L
0 .76631 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.11306 .35333 m .11184 .34828 L p .11377 .35389 L .411 .373 .707 r
F P
0 g
s
.11515 .35483 m .11184 .34828 L .1141 .35485 L p .84 .548 .494 r
F P
0 g
s
.11184 .34828 m .12219 .35434 L .11513 .35479 L p .865 .559 .467 r
F P
0 g
s
.12191 .36818 m .11306 .35333 L .1141 .35485 L .411 .373 .707 r
F
.12191 .36818 m .11547 .36326 L .11306 .35333 L p F P
0 g
s
.1141 .35485 m .11306 .35333 L .11377 .35389 L .411 .373 .707 r
F
.12213 .35468 m .12184 .35631 L .1141 .35485 L p .4 0 0 r
F P
0 g
s
.12983 .37783 m .1141 .35485 L .12184 .35631 L closepath p .62 .462 .647 r
F P
0 g
s
.12983 .37783 m .12191 .36818 L .1141 .35485 L closepath p .263 .329 .743 r
F P
0 g
s
.11513 .35479 m .12219 .35434 L p .11515 .35483 L .4 0 0 r
F P
0 g
s
.12219 .35434 m .12213 .35468 L p .11515 .35483 L .4 0 0 r
F P
0 g
s
.12832 .38878 m .11547 .36326 L .12191 .36818 L closepath p .447 .372 .68 r
F P
0 g
s
.12832 .38878 m .12407 .3854 L .11547 .36326 L closepath p .107 .198 .676 r
F P
0 g
s
.13135 .37106 m .12184 .35631 L .12895 .35322 L closepath p .636 .452 .616 r
F P
0 g
s
.12184 .35631 m .12219 .35434 L .12417 .3553 L p .582 .007 0 r
F P
0 g
s
.13135 .37106 m .12983 .37783 L .12184 .35631 L closepath p .544 .422 .663 r
F P
0 g
s
.13391 .4062 m .12191 .36818 L .12983 .37783 L closepath p .54 .411 .653 r
F P
0 g
s
.13391 .4062 m .12832 .38878 L .12191 .36818 L closepath p 0 0 .296 r
F P
0 g
s
.12968 .40845 m .12407 .3854 L .12832 .38878 L closepath p .24 .196 .605 r
F P
0 g
s
.13731 .4168 m .12968 .40845 L .12832 .38878 L p .355 .284 .637 r
F P
0 g
s
.137 .35938 m .12895 .35322 L .14665 .35736 L closepath p .776 .402 .347 r
F P
0 g
s
.13632 .35371 m .14665 .35736 L .12895 .35322 L closepath p .582 .023 0 r
F P
0 g
s
.14221 .37157 m .13135 .37106 L .12895 .35322 L closepath p .648 .46 .615 r
F P
0 g
s
.14221 .37157 m .12895 .35322 L .137 .35938 L closepath p .721 .473 .551 r
F P
0 g
s
.14398 .43626 m .13756 .42744 L .12968 .40845 L closepath p .771 .475 .485 r
F P
0 g
s
.14398 .43626 m .12968 .40845 L .13938 .41906 L closepath p .287 .269 .664 r
F P
0 g
s
.13756 .42744 m .1313 .41423 L .12968 .40845 L closepath p 0 0 .466 r
F P
0 g
s
.14024 .39004 m .12983 .37783 L .13135 .37106 L closepath p .529 .407 .658 r
F P
0 g
s
.13391 .4062 m .12983 .37783 L .13255 .38102 L p .537 .409 .653 r
F P
0 g
s
.13756 .42744 m .1389 .4354 L .1313 .41423 L closepath p 0 0 .492 r
F P
0 g
s
.14769 .38326 m .13135 .37106 L .14221 .37157 L closepath p .645 .423 .567 r
F P
0 g
s
.14769 .38326 m .14024 .39004 L .13135 .37106 L closepath p .597 .439 .638 r
F P
0 g
s
.13255 .38102 m .14024 .39004 L .13431 .40518 L p .537 .409 .653 r
F P
0 g
s
.13431 .40518 m .13378 .40372 L .13255 .38102 L .537 .409 .653 r
F
.13391 .4062 m .13938 .41906 L .13378 .40372 L p .429 .325 .635 r
F P
0 g
s
.13938 .41906 m .13731 .4168 L p .13391 .4062 L .355 .284 .637 r
F P
0 g
s
.13431 .40518 m .14024 .39004 L p .13475 .40639 L .457 .323 .61 r
F P
0 g
s
.14024 .39004 m .14213 .40809 L .13475 .40639 L p .457 .323 .61 r
F P
0 g
s
.13475 .40639 m .14213 .40809 L .13938 .41906 L p .452 .337 .632 r
F P
0 g
s
.137 .35938 m .14665 .35736 L .15099 .36125 L p .683 .236 .185 r
F P
0 g
s
.14221 .37157 m .137 .35938 L .15202 .36436 L p .708 .467 .558 r
F P
0 g
s
.15202 .36436 m .137 .35938 L p .15099 .36125 L .683 .236 .185 r
F P
0 g
s
.14698 .45195 m .13756 .42744 L .14785 .44687 L p .418 .371 .699 r
F P
0 g
s
.1389 .4354 m .13756 .42744 L .14698 .45195 L p 0 .014 .575 r
F P
0 g
s
.13756 .42744 m .14398 .43626 L .14799 .44604 L p .566 .438 .665 r
F P
0 g
s
.1479 .44659 m .13756 .42744 L .14799 .44604 L .566 .438 .665 r
F
.14785 .44687 m .13756 .42744 L p .1479 .44659 L F P
0 g
s
.14465 .57101 m .13818 .56832 L .13896 .54355 L closepath p .059 .041 .52 r
F P
0 g
s
.13856 .56947 m .13818 .56832 L .13856 .56848 L p .052 .047 .531 r
F P
0 g
s
.14456 .58723 m .13861 .58343 L .13818 .56832 L closepath p .053 .048 .531 r
F P
0 g
s
.14532 .50645 m .13819 .51338 L .14104 .49124 L closepath p .143 .084 .526 r
F P
0 g
s
.14409 .54065 m .13819 .51338 L .14532 .50645 L closepath p .173 .131 .566 r
F P
0 g
s
.14409 .54065 m .13896 .54355 L .13819 .51338 L closepath p 0 0 .493 r
F P
0 g
s
.13856 .56848 m .14465 .57101 L p .13856 .56947 L .052 .047 .531 r
F P
0 g
s
.14465 .57101 m .14456 .58723 L .13856 .56947 L p .052 .047 .531 r
F P
0 g
s
.14601 .6023 m .13861 .58343 L .14456 .58723 L closepath p 0 .069 .582 r
F P
0 g
s
.14601 .6023 m .13957 .59793 L .13861 .58343 L closepath p .104 .129 .602 r
F P
0 g
s
.14659 .45423 m .1389 .4354 L .14686 .45265 L p 0 .023 .58 r
F P
0 g
s
.14686 .45265 m .1389 .4354 L p .14698 .45195 L 0 .014 .575 r
F P
0 g
s
.14465 .57101 m .13896 .54355 L .14409 .54065 L closepath p 0 0 .475 r
F P
0 g
s
.14433 .42422 m .13938 .41906 L .14213 .40809 L closepath p .515 .399 .659 r
F P
0 g
s
.14724 .43459 m .13938 .41906 L p .14721 .43437 L .421 .383 .711 r
F P
0 g
s
.14702 .4333 m .13938 .41906 L .14691 .43272 L .421 .383 .711 r
F
.14691 .43272 m .13938 .41906 L .14657 .43078 L F
.14721 .43437 m .13938 .41906 L .14712 .43386 L F
.14712 .43386 m .13938 .41906 L .14702 .4333 L F
.14398 .43626 m .13938 .41906 L .14724 .43459 L p .59 .463 .675 r
F P
0 g
s
.13938 .41906 m .14433 .42422 L .14653 .43056 L p .421 .383 .711 r
F P
0 g
s
.14657 .43078 m .13938 .41906 L .14653 .43056 L .421 .383 .711 r
F
.14934 .61518 m .13957 .59793 L .14601 .6023 L closepath p .011 .147 .66 r
F P
0 g
s
.14934 .61518 m .14155 .60864 L .13957 .59793 L closepath p .161 .223 .68 r
F P
0 g
s
.14024 .39004 m .14769 .38326 L .14888 .39733 L p .593 .432 .633 r
F P
0 g
s
.14841 .39946 m .14024 .39004 L p .14888 .39733 L .593 .432 .633 r
F P
0 g
s
.14213 .40809 m .14024 .39004 L .14841 .39946 L p .609 .434 .619 r
F P
0 g
s
.14662 .48799 m .14104 .49124 L .14422 .46805 L closepath p .028 0 .431 r
F P
0 g
s
.14532 .50645 m .14104 .49124 L .14662 .48799 L closepath p .066 .039 .514 r
F P
0 g
s
.14278 .61293 m .14155 .60864 L p .15525 .6236 L .405 .481 .825 r
F P
0 g
s
.14155 .60864 m .15365 .62132 L p .15525 .6236 L .405 .481 .825 r
F P
0 g
s
.14155 .60864 m .14934 .61518 L p .15295 .62032 L 0 .133 .661 r
F P
0 g
s
.15365 .62132 m .14155 .60864 L p .15295 .62032 L 0 .133 .661 r
F P
0 g
s
.14807 .40099 m .14213 .40809 L .14841 .39946 L .609 .434 .619 r
F
.14792 .40165 m .14213 .40809 L p .14807 .40099 L F P
0 g
s
.14213 .40809 m .14601 .40378 L p .14679 .40675 L .652 .495 .658 r
F P
0 g
s
.14665 .41026 m .14213 .40809 L .14679 .40675 L .652 .495 .658 r
F
.14653 .41328 m .14213 .40809 L .14659 .41172 L F
.14651 .41393 m .14213 .40809 L p .14653 .41328 L F P
0 g
s
.14659 .41172 m .14213 .40809 L .14665 .41026 L .652 .495 .658 r
F
.14433 .42422 m .14213 .40809 L .14651 .41393 L p .65 .495 .66 r
F P
0 g
s
.14221 .37157 m .15202 .36436 L .15251 .36584 L .708 .467 .558 r
F
.15288 .36696 m .14221 .37157 L p .15251 .36584 L F P
0 g
s
.14221 .37157 m .15085 .36783 L p .15402 .37319 L .684 .419 .515 r
F P
0 g
s
.14769 .38326 m .14221 .37157 L .15432 .37702 L p .677 .436 .55 r
F P
0 g
s
.15432 .37702 m .14221 .37157 L p .15402 .37319 L .684 .419 .515 r
F P
0 g
s
.14808 .33291 m .14226 .34024 L .14437 .3144 L closepath p .285 .187 .564 r
F P
0 g
s
.15097 .35571 m .14226 .34024 L .14808 .33291 L closepath p .381 .315 .658 r
F P
0 g
s
.15097 .35571 m .15077 .36059 L .14226 .34024 L closepath p .262 .262 .67 r
F P
0 g
s
.1577 .62628 m .14519 .62129 L .14278 .61293 L p .405 .481 .825 r
F P
0 g
s
.15616 .6249 m .14278 .61293 L .15573 .62429 L .405 .481 .825 r
F
.15573 .62429 m .14278 .61293 L .15566 .62419 L F
.15566 .62419 m .14278 .61293 L .15525 .6236 L F
.14732 .43504 m .14398 .43626 L .14724 .43459 L .59 .463 .675 r
F
.14764 .43681 m .14398 .43626 L .14732 .43504 L F
.14787 .43813 m .14398 .43626 L p .14782 .43787 L F P
0 g
s
.14764 .43684 m .14398 .43626 L .14764 .43681 L .59 .463 .675 r
F
.14782 .43787 m .14398 .43626 L .14764 .43684 L F
.14799 .44604 m .14398 .43626 L .14828 .44434 L p .664 .487 .636 r
F P
0 g
s
.14398 .43626 m .14787 .43813 L p .14793 .43846 L .594 .459 .667 r
F P
0 g
s
.14794 .43852 m .14398 .43626 L .14793 .43846 L .594 .459 .667 r
F
.14795 .43857 m .14398 .43626 L .14794 .43852 L F
.1478 .44172 m .14398 .43626 L .14826 .44033 L F
.14826 .44033 m .14398 .43626 L .14795 .43857 L F
.14828 .44434 m .14398 .43626 L p .14855 .44279 L F P
0 g
s
.15203 .53929 m .14409 .54065 L .14532 .50645 L closepath p .284 .223 .611 r
F P
0 g
s
.15117 .57371 m .14409 .54065 L .15203 .53929 L closepath p .281 .211 .598 r
F P
0 g
s
.15117 .57371 m .14465 .57101 L .14409 .54065 L closepath p .042 .053 .543 r
F P
0 g
s
.15062 .47009 m .14422 .46805 L .14874 .44165 L closepath p .045 0 .422 r
F P
0 g
s
.14662 .48799 m .14422 .46805 L .15062 .47009 L closepath p .041 0 .435 r
F P
0 g
s
.14619 .42202 m .14433 .42422 L .14651 .41393 L .65 .495 .66 r
F
.14618 .42228 m .14433 .42422 L .14619 .42202 L F
.14615 .42317 m .14433 .42422 L .14618 .42228 L F
.14614 .4234 m .14433 .42422 L .14615 .42317 L F
.14613 .42352 m .14433 .42422 L .14614 .4234 L F
.14612 .42378 m .14433 .42422 L p .14613 .42352 L F P
0 g
s
.14433 .42422 m .14612 .42378 L p .14609 .42461 L .647 .485 .651 r
F P
0 g
s
.14603 .42604 m .14433 .42422 L p .14606 .42525 L .647 .485 .651 r
F P
0 g
s
.14606 .42525 m .14433 .42422 L .14607 .42517 L .647 .485 .651 r
F
.14433 .42422 m .14481 .42473 L p .14633 .42941 L .584 .478 .7 r
F P
0 g
s
.14633 .42941 m .14634 .42948 L .14433 .42422 L .584 .478 .7 r
F
.14634 .42948 m .14645 .43009 L .14433 .42422 L F
.14653 .43056 m .14433 .42422 L p .14645 .43009 L F P
0 g
s
.14607 .42517 m .14433 .42422 L .14609 .42461 L .647 .485 .651 r
F
.14946 .31122 m .14437 .3144 L .14837 .29544 L closepath p .207 .09 .491 r
F P
0 g
s
.14808 .33291 m .14437 .3144 L .14946 .31122 L closepath p .246 .161 .558 r
F P
0 g
s
.15093 .59449 m .14456 .58723 L .14465 .57101 L closepath p 0 .004 .505 r
F P
0 g
s
.15406 .61162 m .14601 .6023 L .14456 .58723 L closepath p .144 .186 .647 r
F P
0 g
s
.15406 .61162 m .14456 .58723 L .15093 .59449 L closepath p 0 0 .551 r
F P
0 g
s
.15093 .59449 m .14465 .57101 L .15117 .57371 L closepath p .056 .038 .518 r
F P
0 g
s
.14481 .42473 m .14603 .42604 L p .14601 .42661 L .584 .478 .7 r
F P
0 g
s
.14601 .42661 m .14601 .42661 L .14481 .42473 L .584 .478 .7 r
F
.14601 .42661 m .14599 .42711 L .14481 .42473 L F
.14599 .42711 m .14599 .4272 L .14481 .42473 L F
.14599 .4272 m .14598 .42742 L .14481 .42473 L F
.14598 .42742 m .14633 .42941 L .14481 .42473 L F
.14519 .62129 m .1577 .62628 L p .15786 .62642 L 0 .33 .669 r
F P
0 g
s
.15789 .62645 m .14519 .62129 L p .15786 .62642 L 0 .33 .669 r
F P
0 g
s
.15409 .50877 m .14532 .50645 L .14662 .48799 L closepath p .29 .232 .617 r
F P
0 g
s
.15203 .53929 m .14532 .50645 L .15409 .50877 L closepath p .292 .229 .612 r
F P
0 g
s
.14598 .42742 m .14679 .40675 L .15118 .42456 L p .242 .162 .561 r
F P
0 g
s
.15109 .42519 m .14598 .42742 L p .15118 .42456 L .242 .162 .561 r
F P
0 g
s
.14598 .42742 m .15026 .42555 L p .15068 .42811 L .264 .213 .611 r
F P
0 g
s
.14598 .42742 m .15068 .42811 L .15 .43285 L .264 .213 .611 r
F
.14598 .42742 m .15 .43285 L .15 .43288 L F
.14598 .42742 m .15 .43288 L .14974 .43467 L F
.14598 .42742 m .14974 .43467 L .1497 .43497 L F
.14598 .42742 m .1497 .43497 L .14959 .43575 L F
.14912 .439 m .14598 .42742 L p .14959 .43575 L F P
0 g
s
.14598 .42742 m .14839 .4363 L p .14874 .44165 L 0 0 .525 r
F P
0 g
s
.14862 .44237 m .14598 .42742 L p .14874 .44165 L 0 0 .525 r
F P
0 g
s
.15616 .6249 m .14601 .6023 L .15406 .61162 L closepath p .018 .152 .662 r
F P
0 g
s
.15616 .6249 m .14934 .61518 L .14601 .6023 L closepath p .009 .146 .66 r
F P
0 g
s
.14601 .40378 m .14792 .40165 L p .14719 .40496 L .652 .495 .658 r
F P
0 g
s
.14679 .40675 m .14601 .40378 L .14719 .40496 L .652 .495 .658 r
F
.15565 .48748 m .14662 .48799 L .15062 .47009 L closepath p .277 .188 .571 r
F P
0 g
s
.15409 .50877 m .14662 .48799 L .15565 .48748 L closepath p .28 .226 .617 r
F P
0 g
s
.14679 .40675 m .15043 .39034 L .15446 .40028 L p .328 .2 .55 r
F P
0 g
s
.15446 .40028 m .15422 .40211 L .14679 .40675 L .328 .2 .55 r
F
.15391 .4044 m .14679 .40675 L p .15422 .40211 L F P
0 g
s
.14679 .40675 m .15096 .40537 L p .15129 .42381 L .336 .22 .57 r
F P
0 g
s
.15118 .42456 m .14679 .40675 L p .15129 .42381 L .336 .22 .57 r
F P
0 g
s
.15656 .22198 m .15441 .22493 L .14763 .22361 L closepath p .637 .354 .473 r
F P
0 g
s
.14763 .22361 m .15441 .22493 L .14868 .23082 L closepath p .631 .47 .646 r
F P
0 g
s
.14769 .38326 m .15432 .37702 L .15449 .37925 L .677 .436 .55 r
F
.14769 .38326 m .15449 .37925 L .15452 .37954 L F
.1545 .37958 m .14769 .38326 L p .15452 .37954 L F P
0 g
s
.14769 .38326 m .1531 .38034 L p .15374 .38181 L .689 .46 .57 r
F P
0 g
s
.15321 .38317 m .14769 .38326 L .15374 .38181 L .689 .46 .57 r
F
.15211 .386 m .14769 .38326 L p .15229 .38555 L F P
0 g
s
.15229 .38555 m .14769 .38326 L .15288 .38402 L .689 .46 .57 r
F
.15288 .38402 m .14769 .38326 L .153 .38371 L F
.153 .38371 m .14769 .38326 L .15321 .38317 L F
.14769 .38326 m .15211 .386 L p .15172 .38701 L .66 .482 .633 r
F P
0 g
s
.15066 .38974 m .14769 .38326 L .15172 .38701 L .66 .482 .633 r
F
.14795 .38642 m .14769 .38326 L p .15066 .38974 L F P
0 g
s
.1478 .44172 m .14826 .44033 L .14835 .44087 L .594 .459 .667 r
F
.1478 .44172 m .14835 .44087 L .14839 .44106 L F
.1478 .44172 m .14843 .44132 L .1485 .44167 L F
.1478 .44172 m .14839 .44106 L .14841 .44118 L F
.1478 .44172 m .14841 .44118 L .14843 .44132 L F
.14855 .44279 m .1478 .44172 L .1485 .44167 L F
.14868 .23082 m .15333 .23509 L .14786 .23486 L closepath p .579 .365 .555 r
F P
0 g
s
.15333 .23509 m .14915 .2514 L .14786 .23486 L closepath p .582 .488 .713 r
F P
0 g
s
.14795 .38642 m .15066 .38974 L .15062 .38986 L .66 .482 .633 r
F
.14888 .39733 m .14795 .38642 L p .15043 .39034 L F P
0 g
s
.14795 .38642 m .15062 .38986 L .15043 .39034 L .66 .482 .633 r
F
.15151 .32697 m .14808 .33291 L .14946 .31122 L closepath p 0 0 .398 r
F P
0 g
s
.16207 .3558 m .14808 .33291 L .15151 .32697 L closepath p .123 .157 .624 r
F P
0 g
s
.15923 .36278 m .15097 .35571 L .14808 .33291 L closepath p .385 .318 .659 r
F P
0 g
s
.16207 .3558 m .15923 .36278 L .14808 .33291 L closepath p .073 .136 .624 r
F P
0 g
s
.15333 .29286 m .14837 .29544 L .1517 .27402 L closepath p .309 .181 .539 r
F P
0 g
s
.14946 .31122 m .14837 .29544 L .15333 .29286 L closepath p .294 .153 .511 r
F P
0 g
s
.14839 .4363 m .14882 .44113 L .14874 .44165 L 0 0 .525 r
F
.14839 .4363 m .14912 .439 L p .14886 .44085 L F P
0 g
s
.14839 .4363 m .14886 .44085 L .14882 .44113 L 0 0 .525 r
F
.14855 .44279 m .1485 .44167 L .14862 .44237 L .594 .459 .667 r
F
.15441 .22493 m .15333 .23509 L .14868 .23082 L closepath p .56 .367 .576 r
F P
0 g
s
.15331 .44132 m .14874 .44165 L .15189 .41961 L closepath p 0 0 .321 r
F P
0 g
s
.14878 .44225 m .14874 .44165 L .14886 .44164 L p 0 0 .352 r
F P
0 g
s
.14886 .44164 m .15331 .44132 L p .14878 .44225 L 0 0 .352 r
F P
0 g
s
.15331 .44132 m .15062 .47009 L .14878 .44225 L p 0 0 .352 r
F P
0 g
s
.15333 .23509 m .152 .25051 L .14915 .2514 L closepath p .384 .292 .627 r
F P
0 g
s
.15505 .27184 m .14915 .2514 L .152 .25051 L closepath p .401 .342 .677 r
F P
0 g
s
.15505 .27184 m .1517 .27402 L .14915 .2514 L closepath p .204 .216 .651 r
F P
0 g
s
.15469 .30598 m .14946 .31122 L .15333 .29286 L closepath p .198 .065 .464 r
F P
0 g
s
.15151 .32697 m .14946 .31122 L .15469 .30598 L closepath p .206 .078 .475 r
F P
0 g
s
.15026 .42555 m .15109 .42519 L p .15073 .42774 L .264 .213 .611 r
F P
0 g
s
.15026 .42555 m .15073 .42774 L .15071 .42788 L .264 .213 .611 r
F
.15026 .42555 m .15071 .42788 L .15068 .42811 L F
.15825 .63825 m .15786 .64084 L .15026 .62335 L p 0 .075 .607 r
F P
0 g
s
.1508 .62357 m .15825 .63825 L p .15026 .62335 L 0 .075 .607 r
F P
0 g
s
.1546 .38065 m .15475 .3825 L .15043 .39034 L 0 0 0 r
F
.15043 .39034 m .15374 .38181 L .1546 .38065 L p F P
0 g
s
.15475 .3825 m .15476 .38263 L .15043 .39034 L 0 0 0 r
F
.15476 .3827 m .15043 .39034 L p .15476 .38263 L F P
0 g
s
.15043 .39034 m .15469 .38284 L p .1548 .38315 L .331 .163 .497 r
F P
0 g
s
.15482 .38902 m .15043 .39034 L p .1548 .38315 L .331 .163 .497 r
F P
0 g
s
.15043 .39034 m .15457 .38909 L p .15485 .39732 L .349 .211 .548 r
F P
0 g
s
.15446 .40028 m .15043 .39034 L p .15485 .39732 L .349 .211 .548 r
F P
0 g
s
.15886 .47001 m .15062 .47009 L .15331 .44132 L closepath p .274 .184 .568 r
F P
0 g
s
.15565 .48748 m .15062 .47009 L .15886 .47001 L closepath p .274 .186 .571 r
F P
0 g
s
.15923 .36278 m .15077 .36059 L .15097 .35571 L closepath p .323 .315 .695 r
F P
0 g
s
.15923 .36278 m .15373 .36952 L .15077 .36059 L closepath p .341 .285 .648 r
F P
0 g
s
.15179 .62397 m .15825 .63825 L .1508 .62357 L p 0 .029 .585 r
F P
0 g
s
.15085 .36783 m .15288 .36696 L p .15373 .36952 L .684 .419 .515 r
F P
0 g
s
.15402 .37319 m .15085 .36783 L .15373 .36952 L .684 .419 .515 r
F
.15093 .59449 m .16333 .55902 L .16424 .59291 L closepath p .43 .394 .718 r
F P
0 g
s
.15093 .59449 m .15117 .57371 L .16333 .55902 L closepath p .227 .178 .591 r
F P
0 g
s
.15406 .61162 m .15093 .59449 L .16424 .59291 L closepath p .436 .478 .805 r
F P
0 g
s
.15096 .40537 m .15391 .4044 L p .15231 .41648 L .336 .22 .57 r
F P
0 g
s
.15096 .40537 m .15231 .41648 L .15189 .41961 L .336 .22 .57 r
F
.15096 .40537 m .15189 .41961 L .15129 .42381 L F
.15117 .57371 m .16765 .52772 L .16333 .55902 L closepath p .169 .098 .527 r
F P
0 g
s
.15117 .57371 m .15203 .53929 L .16765 .52772 L closepath p .262 .196 .591 r
F P
0 g
s
.15858 .32248 m .15151 .32697 L .15469 .30598 L closepath p .482 .319 .583 r
F P
0 g
s
.16207 .3558 m .15151 .32697 L .15858 .32248 L closepath p .517 .382 .636 r
F P
0 g
s
.15333 .29286 m .1517 .27402 L .15505 .27184 L closepath p .113 .041 .488 r
F P
0 g
s
.16106 .6355 m .15825 .63825 L .15179 .62397 L p .47 .362 .649 r
F P
0 g
s
.15401 .62487 m .16106 .6355 L p .15179 .62397 L .47 .362 .649 r
F P
0 g
s
.15551 .41658 m .15189 .41961 L .15485 .39732 L closepath p 0 0 .267 r
F P
0 g
s
.15331 .44132 m .15189 .41961 L .15551 .41658 L closepath p 0 0 .326 r
F P
0 g
s
.15583 .23587 m .152 .25051 L .15333 .23509 L closepath p .436 .337 .644 r
F P
0 g
s
.15583 .23587 m .15836 .25323 L .152 .25051 L closepath p .683 .585 .739 r
F P
0 g
s
.15951 .27041 m .152 .25051 L .15836 .25323 L closepath p .731 .522 .609 r
F P
0 g
s
.15951 .27041 m .15505 .27184 L .152 .25051 L closepath p .419 .355 .679 r
F P
0 g
s
.15203 .53929 m .17014 .50348 L .16765 .52772 L closepath p .216 .117 .521 r
F P
0 g
s
.15203 .53929 m .15409 .50877 L .17014 .50348 L closepath p .382 .306 .646 r
F P
0 g
s
.1531 .38034 m .15366 .38004 L p .15392 .38128 L .689 .46 .57 r
F P
0 g
s
.15388 .38139 m .1531 .38034 L .15388 .38138 L .689 .46 .57 r
F
.15388 .38138 m .1531 .38034 L .15389 .38138 L F
.15389 .38138 m .1531 .38034 L .15392 .38128 L F
.15382 .38158 m .1531 .38034 L .15388 .38139 L F
.15376 .38175 m .1531 .38034 L .15382 .38158 L F
.15375 .38176 m .1531 .38034 L .15376 .38175 L F
.15374 .3818 m .1531 .38034 L .15375 .38176 L F
.15374 .38181 m .1531 .38034 L .15374 .3818 L F
.16124 .4373 m .15331 .44132 L .15551 .41658 L closepath p .189 .076 .484 r
F P
0 g
s
.15886 .47001 m .15331 .44132 L .16124 .4373 L closepath p .226 .153 .56 r
F P
0 g
s
.15827 .29243 m .15333 .29286 L .15505 .27184 L closepath p .354 .244 .588 r
F P
0 g
s
.15469 .30598 m .15333 .29286 L .15827 .29243 L closepath p .339 .162 .489 r
F P
0 g
s
.15333 .23509 m .15724 .22401 L .1601 .22429 L closepath p .505 .347 .6 r
F P
0 g
s
.15441 .22493 m .15724 .22401 L .15333 .23509 L closepath p .341 .18 .512 r
F P
0 g
s
.15333 .23509 m .1601 .22429 L .15583 .23587 L closepath p .461 .294 .567 r
F P
0 g
s
.15366 .38004 m .1545 .37958 L p .15446 .3797 L .689 .46 .57 r
F P
0 g
s
.15441 .37985 m .15366 .38004 L .15446 .3797 L .689 .46 .57 r
F
.15436 .38 m .15366 .38004 L .15439 .37991 L F
.15439 .37991 m .15366 .38004 L .15441 .37985 L F
.15402 .381 m .15395 .3812 L .15366 .38004 L F
.15395 .3812 m .15394 .38122 L .15366 .38004 L F
.15394 .38122 m .15393 .38126 L .15366 .38004 L F
.15393 .38126 m .15392 .38128 L .15366 .38004 L F
.15405 .38091 m .15366 .38004 L .15405 .3809 L F
.15405 .3809 m .15366 .38004 L .15405 .3809 L F
.15405 .3809 m .15366 .38004 L .15407 .38085 L F
.15407 .38085 m .15366 .38004 L .15381 .38003 L F
.15402 .38098 m .15366 .38004 L .15405 .38091 L F
.15402 .381 m .15366 .38004 L .15402 .38098 L F
.15979 .37321 m .15373 .36952 L .15923 .36278 L closepath p .176 .062 .475 r
F P
0 g
s
.15979 .37321 m .1548 .38315 L .15373 .36952 L closepath p .153 .072 .504 r
F P
0 g
s
.15374 .38181 m .15452 .37954 L p .1546 .38056 L 0 0 .389 r
F P
0 g
s
.1546 .38065 m .15374 .38181 L p .1546 .3806 L 0 0 .389 r
F P
0 g
s
.1546 .38056 m .1546 .3806 L .15374 .38181 L 0 0 .389 r
F
.15414 .38064 m .15381 .38003 L .15436 .38 L .689 .46 .57 r
F
.15412 .3807 m .15381 .38003 L .15413 .38067 L F
.15413 .38067 m .15381 .38003 L .15414 .38064 L F
.1541 .38076 m .15381 .38003 L .15412 .3807 L F
.15407 .38085 m .15381 .38003 L .1541 .38076 L F
.16106 .6355 m .15401 .62487 L p .15526 .62538 L .461 .359 .652 r
F P
0 g
s
.15406 .61162 m .16424 .59291 L .1672 .61238 L closepath p .382 .412 .767 r
F P
0 g
s
.15616 .6249 m .15406 .61162 L .1672 .61238 L closepath p .375 .437 .797 r
F P
0 g
s
.15409 .50877 m .17028 .48646 L .17014 .50348 L closepath p .365 .26 .6 r
F P
0 g
s
.15409 .50877 m .15565 .48748 L .17028 .48646 L closepath p .458 .38 .681 r
F P
0 g
s
.15656 .22198 m .15724 .22401 L .15441 .22493 L closepath p .187 0 .143 r
F P
0 g
s
.15457 .38909 m .15482 .38902 L p .15482 .3901 L .349 .211 .548 r
F P
0 g
s
.15457 .38909 m .15482 .3901 L .15485 .39732 L .349 .211 .548 r
F
.15469 .38284 m .15476 .3827 L p .1548 .38315 L .331 .163 .497 r
F P
0 g
s
.15706 .30483 m .15469 .30598 L .15827 .29243 L closepath p .482 .297 .553 r
F P
0 g
s
.15858 .32248 m .15469 .30598 L .15706 .30483 L closepath p .498 .33 .583 r
F P
0 g
s
.15809 .39214 m .1548 .38315 L .15979 .37321 L closepath p 0 0 .347 r
F P
0 g
s
.15809 .39214 m .15485 .39732 L .1548 .38315 L closepath p 0 0 .35 r
F P
0 g
s
.15551 .41658 m .15485 .39732 L .15809 .39214 L closepath p 0 0 .257 r
F P
0 g
s
.15827 .29243 m .15505 .27184 L .15951 .27041 L closepath p .388 .267 .592 r
F P
0 g
s
.15526 .62538 m .15534 .62541 L .16106 .6355 L .461 .359 .652 r
F
.15549 .20128 m .15704 .21361 L .15533 .21095 L closepath p .043 .042 .53 r
F P
0 g
s
.15533 .21095 m .15724 .22401 L .15656 .22198 L closepath p 0 0 .239 r
F P
0 g
s
.15533 .21095 m .15704 .21361 L .15724 .22401 L closepath p 0 .022 .536 r
F P
0 g
s
.15633 .62581 m .16106 .6355 L p .15534 .62541 L .461 .359 .652 r
F P
0 g
s
.15865 .18948 m .15549 .20128 L .15732 .18875 L closepath p 0 0 .289 r
F P
0 g
s
.15865 .18948 m .15813 .20433 L .15549 .20128 L closepath p .114 .12 .586 r
F P
0 g
s
.15549 .20128 m .15813 .20433 L .15704 .21361 L closepath p .164 .13 .571 r
F P
0 g
s
.16223 .41507 m .15551 .41658 L .15809 .39214 L closepath p .068 0 .386 r
F P
0 g
s
.16124 .4373 m .15551 .41658 L .16223 .41507 L closepath p .087 .018 .474 r
F P
0 g
s
.15565 .48748 m .1753 .46893 L .17028 .48646 L closepath p .454 .347 .643 r
F P
0 g
s
.15565 .48748 m .15886 .47001 L .1753 .46893 L closepath p .465 .364 .655 r
F P
0 g
s
.16271 .23852 m .15583 .23587 L .1601 .22429 L closepath p .712 .561 .682 r
F P
0 g
s
.16276 .25304 m .15583 .23587 L .16271 .23852 L closepath p .695 .584 .727 r
F P
0 g
s
.16276 .25304 m .15836 .25323 L .15583 .23587 L closepath p .65 .561 .742 r
F P
0 g
s
.15616 .6249 m .1672 .61238 L .17558 .62726 L closepath p .483 .602 .895 r
F P
0 g
s
.16675 .63438 m .15616 .6249 L .17558 .62726 L closepath p .439 .636 .939 r
F P
0 g
s
.15678 .626 m .16554 .63335 L .15633 .62581 L .642 .44 .594 r
F
.16554 .63335 m .16106 .6355 L .15633 .62581 L p F P
0 g
s
.15788 .62644 m .16554 .63335 L .15678 .626 L .642 .44 .594 r
F
.1624 .21573 m .15704 .21361 L .15813 .20433 L closepath p .452 .39 .697 r
F P
0 g
s
.15704 .21361 m .1624 .21573 L .1601 .22429 L closepath p .429 .408 .734 r
F P
0 g
s
.15704 .21361 m .1601 .22429 L .15724 .22401 L closepath p .49 .447 .736 r
F P
0 g
s
.16605 .30361 m .15706 .30483 L .15827 .29243 L closepath p .596 .396 .58 r
F P
0 g
s
.16713 .32032 m .15706 .30483 L .16605 .30361 L closepath p .602 .416 .603 r
F P
0 g
s
.16713 .32032 m .15858 .32248 L .15706 .30483 L closepath p .625 .423 .588 r
F P
0 g
s
.15732 .18875 m .16116 .17925 L .15865 .18948 L closepath p 0 0 .229 r
F P
0 g
s
.15786 .64084 m .15825 .63825 L .17239 .64866 L closepath p .265 .408 .82 r
F P
0 g
s
.15999 .64258 m .15786 .64084 L .17307 .65281 L closepath p 0 .109 .574 r
F P
0 g
s
.15786 .64084 m .17239 .64866 L .17307 .65281 L closepath p .296 .406 .805 r
F P
0 g
s
.16494 .63276 m .15788 .62644 L .15789 .62645 L .642 .44 .594 r
F
.16494 .63276 m .16554 .63335 L .15788 .62644 L F
.16368 .38618 m .15809 .39214 L .15979 .37321 L closepath p 0 0 .312 r
F P
0 g
s
.16223 .41507 m .15809 .39214 L .16368 .38618 L closepath p 0 0 .363 r
F P
0 g
s
.16519 .20474 m .15813 .20433 L .15865 .18948 L closepath p .506 .457 .736 r
F P
0 g
s
.1624 .21573 m .15813 .20433 L .16519 .20474 L closepath p .511 .423 .693 r
F P
0 g
s
.15825 .63825 m .16106 .6355 L .17389 .64383 L closepath p .511 .423 .692 r
F P
0 g
s
.15825 .63825 m .17389 .64383 L .17239 .64866 L closepath p .514 .419 .685 r
F P
0 g
s
.16633 .29279 m .15827 .29243 L .15951 .27041 L closepath p .61 .455 .646 r
F P
0 g
s
.16605 .30361 m .15827 .29243 L .16633 .29279 L closepath p .607 .398 .573 r
F P
0 g
s
.15951 .27041 m .15836 .25323 L .16276 .25304 L closepath p .64 .452 .614 r
F P
0 g
s
.16794 .34914 m .15858 .32248 L .16713 .32032 L closepath p .627 .43 .596 r
F P
0 g
s
.16207 .3558 m .15858 .32248 L .16794 .34914 L closepath p .447 .331 .628 r
F P
0 g
s
.16889 .18948 m .15865 .18948 L .16116 .17925 L closepath p .493 .376 .648 r
F P
0 g
s
.16519 .20474 m .15865 .18948 L .16889 .18948 L closepath p .492 .449 .738 r
F P
0 g
s
.15886 .47001 m .16124 .4373 L .17642 .44151 L closepath p .467 .358 .647 r
F P
0 g
s
.15886 .47001 m .17642 .44151 L .1753 .46893 L closepath p .464 .354 .644 r
F P
0 g
s
.16523 .36778 m .15923 .36278 L .16207 .3558 L closepath p 0 0 .488 r
F P
0 g
s
.16523 .36778 m .15979 .37321 L .15923 .36278 L closepath p .011 0 .423 r
F P
0 g
s
.1658 .26977 m .15951 .27041 L .16276 .25304 L closepath p .612 .426 .606 r
F P
0 g
s
.16633 .29279 m .15951 .27041 L .1658 .26977 L closepath p .619 .459 .644 r
F P
0 g
s
.16368 .38618 m .15979 .37321 L .16523 .36778 L closepath p 0 0 .312 r
F P
0 g
s
.15999 .64258 m .17307 .65281 L .17573 .65516 L closepath p 0 .127 .604 r
F P
0 g
s
.1624 .21573 m .16762 .2257 L .1601 .22429 L closepath p .621 .605 .813 r
F P
0 g
s
.16271 .23852 m .1601 .22429 L .16762 .2257 L closepath p .658 .525 .691 r
F P
0 g
s
.16106 .6355 m .16554 .63335 L .16569 .63343 L p .606 .369 .531 r
F P
0 g
s
.16106 .6355 m .16569 .63343 L .16582 .63355 L .606 .369 .531 r
F
.16106 .6355 m .16582 .63355 L .16611 .6338 L F
.16917 .63758 m .16106 .6355 L p .16611 .6338 L F P
0 g
s
.17389 .64383 m .16106 .6355 L .16917 .63758 L p .604 .386 .558 r
F P
0 g
s
.16889 .18948 m .16116 .17925 L .17711 .17528 L closepath p .529 .386 .63 r
F P
0 g
s
.16124 .4373 m .16223 .41507 L .17919 .41914 L closepath p .485 .343 .612 r
F P
0 g
s
.16124 .4373 m .17919 .41914 L .17642 .44151 L closepath p .479 .334 .606 r
F P
0 g
s
.18039 .36731 m .16207 .3558 L .16794 .34914 L closepath p .433 .311 .614 r
F P
0 g
s
.18039 .36731 m .16523 .36778 L .16207 .3558 L closepath p .475 .294 .555 r
F P
0 g
s
.16223 .41507 m .16368 .38618 L .1834 .39065 L closepath p .529 .34 .568 r
F P
0 g
s
.16223 .41507 m .1834 .39065 L .17919 .41914 L closepath p .497 .298 .539 r
F P
0 g
s
.16519 .20474 m .17285 .21299 L .1624 .21573 L closepath p .594 .506 .725 r
F P
0 g
s
.1624 .21573 m .17285 .21299 L .16762 .2257 L closepath p .613 .601 .815 r
F P
0 g
s
.16271 .23852 m .17678 .22534 L .17117 .23808 L closepath p .587 .432 .638 r
F P
0 g
s
.16271 .23852 m .16762 .2257 L .17678 .22534 L closepath p .621 .485 .675 r
F P
0 g
s
.17117 .23808 m .16276 .25304 L .16271 .23852 L closepath p .593 .501 .719 r
F P
0 g
s
.17117 .23808 m .16823 .25317 L .16276 .25304 L closepath p .528 .421 .677 r
F P
0 g
s
.1658 .26977 m .16276 .25304 L .16823 .25317 L closepath p .527 .369 .61 r
F P
0 g
s
.18346 .53077 m .16333 .55902 L .16765 .52772 L closepath p .496 .391 .665 r
F P
0 g
s
.18113 .56993 m .16333 .55902 L .18346 .53077 L closepath p .543 .453 .702 r
F P
0 g
s
.18113 .56993 m .16424 .59291 L .16333 .55902 L closepath p .512 .461 .736 r
F P
0 g
s
.16368 .38618 m .18039 .36731 L .1834 .39065 L closepath p .531 .332 .556 r
F P
0 g
s
.16368 .38618 m .16523 .36778 L .18039 .36731 L closepath p .467 .248 .497 r
F P
0 g
s
.18917 .6006 m .16424 .59291 L .18113 .56993 L closepath p .578 .551 .788 r
F P
0 g
s
.18917 .6006 m .1672 .61238 L .16424 .59291 L closepath p .574 .555 .795 r
F P
0 g
s
.16539 .63317 m .16554 .63335 L p .16494 .63276 L .642 .44 .594 r
F P
0 g
s
.16889 .18948 m .17563 .20654 L .16519 .20474 L closepath p .62 .578 .785 r
F P
0 g
s
.16519 .20474 m .17563 .20654 L .17285 .21299 L closepath p .644 .513 .689 r
F P
0 g
s
.16549 .63325 m .16554 .63335 L .16539 .63317 L p .651 .442 .586 r
F P
0 g
s
.16554 .63335 m .16549 .63325 L p .16559 .63334 L .747 .474 .519 r
F P
0 g
s
.16559 .63334 m .16554 .63335 L p .16559 .63334 L .747 .474 .519 r
F P
0 g
s
.16562 .63337 m .16554 .63335 L p .16561 .63336 L .634 .237 .272 r
F P
0 g
s
.16554 .63335 m .16559 .63334 L p .16561 .63336 L .634 .237 .272 r
F P
0 g
s
.16569 .63343 m .16554 .63335 L .16562 .63337 L p .647 .298 .367 r
F P
0 g
s
.16823 .25317 m .17268 .27315 L .1658 .26977 L closepath p .776 .599 .66 r
F P
0 g
s
.1729 .2924 m .16633 .29279 L .1658 .26977 L closepath p .745 .56 .644 r
F P
0 g
s
.1729 .2924 m .1658 .26977 L .17268 .27315 L closepath p .786 .582 .624 r
F P
0 g
s
.17516 .30544 m .16605 .30361 L .16633 .29279 L closepath p .768 .533 .579 r
F P
0 g
s
.17725 .31945 m .16713 .32032 L .16605 .30361 L closepath p .739 .52 .597 r
F P
0 g
s
.17725 .31945 m .16605 .30361 L .17516 .30544 L closepath p .768 .524 .565 r
F P
0 g
s
.16611 .6338 m .16626 .63394 L .16917 .63758 L .606 .369 .531 r
F
.16626 .63394 m .16675 .63438 L .16917 .63758 L F
.17516 .30544 m .16633 .29279 L .1729 .2924 L closepath p .74 .529 .609 r
F P
0 g
s
.16675 .63438 m .17558 .62726 L .18729 .63502 L closepath p .479 .716 .974 r
F P
0 g
s
.16675 .63438 m .18729 .63502 L .19384 .63954 L closepath p .435 .733 .989 r
F P
0 g
s
.17389 .64383 m .16675 .63438 L .19384 .63954 L closepath p .522 .729 .968 r
F P
0 g
s
.17504 .34429 m .16713 .32032 L .17725 .31945 L closepath p .727 .474 .545 r
F P
0 g
s
.16794 .34914 m .16713 .32032 L .17504 .34429 L closepath p .584 .396 .593 r
F P
0 g
s
.20169 .6194 m .1672 .61238 L .18917 .6006 L closepath p .614 .664 .875 r
F P
0 g
s
.20169 .6194 m .17558 .62726 L .1672 .61238 L closepath p .609 .668 .882 r
F P
0 g
s
.17285 .21299 m .18151 .21545 L .16762 .2257 L closepath p .716 .721 .853 r
F P
0 g
s
.18151 .21545 m .17678 .22534 L .16762 .2257 L closepath p .624 .527 .725 r
F P
0 g
s
.18564 .51528 m .16765 .52772 L .17014 .50348 L closepath p .494 .356 .622 r
F P
0 g
s
.18346 .53077 m .16765 .52772 L .18564 .51528 L closepath p .502 .37 .633 r
F P
0 g
s
.17504 .34429 m .18039 .36731 L .16794 .34914 L closepath p .555 .348 .555 r
F P
0 g
s
.17858 .23965 m .16823 .25317 L .17117 .23808 L closepath p .831 .717 .738 r
F P
0 g
s
.17858 .23965 m .17427 .25455 L .16823 .25317 L closepath p .816 .692 .728 r
F P
0 g
s
.16823 .25317 m .17427 .25455 L .17268 .27315 L closepath p .83 .632 .634 r
F P
0 g
s
.17711 .17528 m .18109 .1943 L .16889 .18948 L closepath p .698 .594 .737 r
F P
0 g
s
.16889 .18948 m .18109 .1943 L .17563 .20654 L closepath p .68 .61 .77 r
F P
0 g
s
.18677 .50699 m .17014 .50348 L .17028 .48646 L closepath p .496 .365 .632 r
F P
0 g
s
.18564 .51528 m .17014 .50348 L .18677 .50699 L closepath p .498 .356 .617 r
F P
0 g
s
.18871 .48437 m .17028 .48646 L .1753 .46893 L closepath p .472 .365 .651 r
F P
0 g
s
.18677 .50699 m .17028 .48646 L .18871 .48437 L closepath p .471 .359 .645 r
F P
0 g
s
.17678 .22534 m .17858 .23965 L .17117 .23808 L closepath p .83 .72 .743 r
F P
0 g
s
.17239 .64866 m .17389 .64383 L .18715 .64971 L closepath p .561 .468 .707 r
F P
0 g
s
.17307 .65281 m .17239 .64866 L .18715 .64971 L closepath p .526 .584 .856 r
F P
0 g
s
.17427 .25455 m .17462 .27625 L .17268 .27315 L closepath p .999 .807 .511 r
F P
0 g
s
.17268 .27315 m .17601 .29194 L .1729 .2924 L closepath p .989 .757 .491 r
F P
0 g
s
.17268 .27315 m .17462 .27625 L .17601 .29194 L closepath p .99 .741 .374 r
F P
0 g
s
.18151 .21545 m .17285 .21299 L .17563 .20654 L closepath p .77 .661 .742 r
F P
0 g
s
.1729 .2924 m .17601 .29194 L .1802 .30504 L closepath p .957 .634 .365 r
F P
0 g
s
.1729 .2924 m .1802 .30504 L .17516 .30544 L closepath p .945 .664 .475 r
F P
0 g
s
.17307 .65281 m .18715 .64971 L .18887 .65427 L closepath p .526 .583 .855 r
F P
0 g
s
.17573 .65516 m .17307 .65281 L .18887 .65427 L closepath p .376 .687 .98 r
F P
0 g
s
.17389 .64383 m .19384 .63954 L .18715 .64971 L closepath p .523 .492 .763 r
F P
0 g
s
.17798 .24558 m .17427 .25455 L .17858 .23965 L closepath p .933 .902 .505 r
F P
0 g
s
.17798 .24558 m .17508 .25493 L .17427 .25455 L closepath p .957 .905 .532 r
F P
0 g
s
.17427 .25455 m .17508 .25493 L .17462 .27625 L closepath p .966 .779 .343 r
F P
0 g
s
.17504 .34429 m .17725 .31945 L .1834 .3427 L closepath p .634 .393 .535 r
F P
0 g
s
.19263 .36616 m .17504 .34429 L .1834 .3427 L closepath p .63 .381 .522 r
F P
0 g
s
.19263 .36616 m .18039 .36731 L .17504 .34429 L closepath p .601 .377 .548 r
F P
0 g
s
.17516 .30544 m .1802 .30504 L .18498 .31843 L closepath p .906 .552 .357 r
F P
0 g
s
.17516 .30544 m .18498 .31843 L .17725 .31945 L closepath p .864 .591 .523 r
F P
0 g
s
.18953 .4645 m .1753 .46893 L .17642 .44151 L closepath p .444 .337 .639 r
F P
0 g
s
.18871 .48437 m .1753 .46893 L .18953 .4645 L closepath p .452 .361 .662 r
F P
0 g
s
.21212 .62938 m .17558 .62726 L .20169 .6194 L closepath p .611 .725 .927 r
F P
0 g
s
.21212 .62938 m .18729 .63502 L .17558 .62726 L closepath p .582 .746 .955 r
F P
0 g
s
.18333 .2074 m .17563 .20654 L .18109 .1943 L closepath p .77 .719 .804 r
F P
0 g
s
.18151 .21545 m .17563 .20654 L .18333 .2074 L closepath p .787 .665 .729 r
F P
0 g
s
.17573 .65516 m .18887 .65427 L .19205 .65683 L closepath p .377 .686 .979 r
F P
0 g
s
.19016 .44849 m .17642 .44151 L .17919 .41914 L closepath p .474 .33 .604 r
F P
0 g
s
.18953 .4645 m .17642 .44151 L .19016 .44849 L closepath p .455 .342 .636 r
F P
0 g
s
.18927 .21536 m .17678 .22534 L .18151 .21545 L closepath p .808 .755 .804 r
F P
0 g
s
.18927 .21536 m .18277 .22901 L .17678 .22534 L closepath p .8 .735 .792 r
F P
0 g
s
.17678 .22534 m .18277 .22901 L .17858 .23965 L closepath p .83 .719 .743 r
F P
0 g
s
.17711 .17528 m .18562 .17826 L .18109 .1943 L closepath p .643 .558 .744 r
F P
0 g
s
.20106 .18125 m .18562 .17826 L .17711 .17528 L closepath p .569 .008 0 r
F P
0 g
s
.18644 .3292 m .17725 .31945 L .18498 .31843 L closepath p .848 .544 .473 r
F P
0 g
s
.1834 .3427 m .17725 .31945 L .18644 .3292 L closepath p .862 .528 .42 r
F P
0 g
s
.18293 .23066 m .17798 .24558 L .17858 .23965 L closepath p .95 .928 .577 r
F P
0 g
s
.18293 .23066 m .17858 .23965 L .18277 .22901 L closepath p .873 .909 .501 r
F P
0 g
s
.1941 .42334 m .17919 .41914 L .1834 .39065 L closepath p .535 .331 .55 r
F P
0 g
s
.19016 .44849 m .17919 .41914 L .1941 .42334 L closepath p .526 .358 .595 r
F P
0 g
s
.1973 .39376 m .18039 .36731 L .19263 .36616 L closepath p .602 .379 .55 r
F P
0 g
s
.1973 .39376 m .1834 .39065 L .18039 .36731 L closepath p .593 .376 .555 r
F P
0 g
s
.18562 .17826 m .18746 .19633 L .18109 .1943 L closepath p .8 .725 .781 r
F P
0 g
s
.18333 .2074 m .18109 .1943 L .18746 .19633 L closepath p .793 .733 .797 r
F P
0 g
s
.2066 .54828 m .18113 .56993 L .18346 .53077 L closepath p .593 .496 .713 r
F P
0 g
s
.21668 .57346 m .18113 .56993 L .2066 .54828 L closepath p .588 .486 .706 r
F P
0 g
s
.21668 .57346 m .18917 .6006 L .18113 .56993 L closepath p .572 .547 .789 r
F P
0 g
s
.18333 .2074 m .18927 .21536 L .18151 .21545 L closepath p .818 .697 .731 r
F P
0 g
s
.19105 .21766 m .18277 .22901 L .18927 .21536 L closepath p .934 .941 .775 r
F P
0 g
s
.19105 .21766 m .18293 .23066 L .18277 .22901 L closepath p .859 .98 .684 r
F P
0 g
s
.19071 .21006 m .18333 .2074 L .18746 .19633 L closepath p .801 .743 .799 r
F P
0 g
s
.18333 .2074 m .19071 .21006 L .18927 .21536 L closepath p .836 .698 .711 r
F P
0 g
s
.1834 .3427 m .18644 .3292 L .19324 .34564 L closepath p .785 .443 .408 r
F P
0 g
s
.19324 .34564 m .19263 .36616 L .1834 .3427 L closepath p .787 .456 .427 r
F P
0 g
s
.1941 .42334 m .1834 .39065 L .1973 .39376 L closepath p .595 .364 .536 r
F P
0 g
s
.20555 .52326 m .18346 .53077 L .18564 .51528 L closepath p .64 .496 .67 r
F P
0 g
s
.2066 .54828 m .18346 .53077 L .20555 .52326 L closepath p .638 .488 .663 r
F P
0 g
s
.2013 .33769 m .18644 .3292 L .18498 .31843 L closepath p .842 .539 .477 r
F P
0 g
s
.20106 .18125 m .19153 .18265 L .18562 .17826 L closepath p .813 .566 .566 r
F P
0 g
s
.18562 .17826 m .19153 .18265 L .18746 .19633 L closepath p .662 .624 .801 r
F P
0 g
s
.20743 .50386 m .18564 .51528 L .18677 .50699 L closepath p .618 .464 .65 r
F P
0 g
s
.20555 .52326 m .18564 .51528 L .20743 .50386 L closepath p .636 .501 .681 r
F P
0 g
s
.19324 .34564 m .18644 .3292 L .2013 .33769 L closepath p .841 .466 .342 r
F P
0 g
s
.21321 .4884 m .18677 .50699 L .18871 .48437 L closepath p .61 .481 .679 r
F P
0 g
s
.20743 .50386 m .18677 .50699 L .21321 .4884 L closepath p .628 .514 .704 r
F P
0 g
s
.19384 .63954 m .20302 .65005 L .18715 .64971 L closepath p .667 .686 .858 r
F P
0 g
s
.20302 .65005 m .18887 .65427 L .18715 .64971 L closepath p .673 .67 .839 r
F P
0 g
s
.22057 .63608 m .18729 .63502 L .21212 .62938 L closepath p .579 .754 .961 r
F P
0 g
s
.22057 .63608 m .19384 .63954 L .18729 .63502 L closepath p .536 .774 .985 r
F P
0 g
s
.20106 .18125 m .18746 .19633 L .19153 .18265 L closepath p .834 .81 .827 r
F P
0 g
s
.20106 .18125 m .19487 .19675 L .18746 .19633 L closepath p .799 .741 .798 r
F P
0 g
s
.19071 .21006 m .18746 .19633 L .19487 .19675 L closepath p .799 .742 .799 r
F P
0 g
s
.217 .4718 m .18871 .48437 L .18953 .4645 L closepath p .589 .476 .693 r
F P
0 g
s
.21321 .4884 m .18871 .48437 L .217 .4718 L closepath p .602 .51 .722 r
F P
0 g
s
.20302 .65005 m .19205 .65683 L .18887 .65427 L closepath p .67 .802 .946 r
F P
0 g
s
.23148 .60042 m .18917 .6006 L .21668 .57346 L closepath p .607 .606 .824 r
F P
0 g
s
.23148 .60042 m .20169 .6194 L .18917 .6006 L closepath p .591 .654 .88 r
F P
0 g
s
.19105 .21766 m .18927 .21536 L .19071 .21006 L closepath p .986 .897 .63 r
F P
0 g
s
.21421 .45455 m .18953 .4645 L .19016 .44849 L closepath p .547 .419 .657 r
F P
0 g
s
.217 .4718 m .18953 .4645 L .21421 .45455 L closepath p .577 .497 .727 r
F P
0 g
s
.20647 .42376 m .19016 .44849 L .1941 .42334 L closepath p .506 .34 .589 r
F P
0 g
s
.21421 .45455 m .19016 .44849 L .20647 .42376 L closepath p .555 .399 .625 r
F P
0 g
s
.19383 .21004 m .19071 .21006 L .19487 .19675 L closepath p .919 .882 .781 r
F P
0 g
s
.19105 .21766 m .19071 .21006 L .19383 .21004 L closepath p .923 .874 .771 r
F P
0 g
s
.20302 .65005 m .20639 .65341 L .19205 .65683 L closepath p .625 .66 .864 r
F P
0 g
s
.19584 .65672 m .19205 .65683 L .20639 .65341 L closepath p .363 .83 .896 r
F P
0 g
s
.19324 .34564 m .20943 .36194 L .19263 .36616 L closepath p .746 .42 .43 r
F P
0 g
s
.20943 .36194 m .21676 .38053 L .19263 .36616 L closepath p .75 .427 .437 r
F P
0 g
s
.21323 .38592 m .1973 .39376 L .19263 .36616 L closepath p .695 .442 .536 r
F P
0 g
s
.21323 .38592 m .19263 .36616 L .21676 .38053 L closepath p .752 .419 .42 r
F P
0 g
s
.19324 .34564 m .2013 .33769 L .20458 .3491 L closepath p .748 .331 .258 r
F P
0 g
s
.19324 .34564 m .20458 .3491 L .20943 .36194 L closepath p .769 .405 .368 r
F P
0 g
s
.19383 .21004 m .19487 .19675 L .20064 .20162 L closepath p .877 .848 .812 r
F P
0 g
s
.22917 .64117 m .19384 .63954 L .22057 .63608 L closepath p .55 .759 .974 r
F P
0 g
s
.22917 .64117 m .20302 .65005 L .19384 .63954 L closepath p .639 .678 .871 r
F P
0 g
s
.20732 .4023 m .1941 .42334 L .1973 .39376 L closepath p .609 .376 .539 r
F P
0 g
s
.20647 .42376 m .1941 .42334 L .20732 .4023 L closepath p .495 .242 .461 r
F P
0 g
s
.20618 .18994 m .19487 .19675 L .20106 .18125 L closepath p .912 .887 .795 r
F P
0 g
s
.20618 .18994 m .20064 .20162 L .19487 .19675 L closepath p .91 .837 .762 r
F P
0 g
s
.19522 .33052 m .19779 .33321 L p .19526 .33056 L .698 .181 0 r
F P
0 g
s
.19779 .33321 m .2013 .33769 L p .19526 .33056 L .698 .181 0 r
F P
0 g
s
.20732 .4023 m .1973 .39376 L .21323 .38592 L closepath p .652 .363 .467 r
F P
0 g
s
.19779 .33321 m .21465 .33031 L .2013 .33769 L closepath p .628 .274 .352 r
F P
0 g
s
.19779 .33321 m .21218 .33064 L p .21216 .33074 L .174 0 0 r
F P
0 g
s
.2013 .33769 m .22085 .32957 L .21825 .33535 L closepath p .458 .052 .181 r
F P
0 g
s
.2013 .33769 m .21465 .33031 L .22085 .32957 L closepath p .305 0 0 r
F P
0 g
s
.21825 .33535 m .20458 .3491 L .2013 .33769 L closepath p .528 .197 .35 r
F P
0 g
s
.2509 .61748 m .20169 .6194 L .23148 .60042 L closepath p .62 .735 .928 r
F P
0 g
s
.2509 .61748 m .21212 .62938 L .20169 .6194 L closepath p .624 .729 .922 r
F P
0 g
s
.23527 .64297 m .20302 .65005 L .22917 .64117 L closepath p .643 .762 .935 r
F P
0 g
s
.23527 .64297 m .20639 .65341 L .20302 .65005 L closepath p .654 .666 .85 r
F P
0 g
s
.21825 .33535 m .21674 .34354 L .20458 .3491 L closepath p .274 0 .128 r
F P
0 g
s
.21674 .34354 m .20943 .36194 L .20458 .3491 L closepath p .499 .269 .496 r
F P
0 g
s
.2281 .52462 m .20555 .52326 L .20743 .50386 L closepath p .621 .488 .679 r
F P
0 g
s
.23058 .55148 m .20555 .52326 L .2281 .52462 L closepath p .622 .461 .643 r
F P
0 g
s
.23058 .55148 m .2066 .54828 L .20555 .52326 L closepath p .599 .458 .661 r
F P
0 g
s
.23528 .6434 m .20639 .65341 L .23527 .64297 L closepath p .365 .834 .804 r
F P
0 g
s
.2247 .42978 m .20647 .42376 L .20732 .4023 L closepath p .638 .359 .479 r
F P
0 g
s
.22842 .45517 m .20647 .42376 L .2247 .42978 L closepath p .611 .448 .637 r
F P
0 g
s
.22842 .45517 m .21421 .45455 L .20647 .42376 L closepath p .65 .458 .611 r
F P
0 g
s
.23263 .57433 m .2066 .54828 L .23058 .55148 L closepath p .593 .491 .708 r
F P
0 g
s
.23263 .57433 m .21668 .57346 L .2066 .54828 L closepath p .599 .492 .704 r
F P
0 g
s
.22387 .41248 m .20732 .4023 L .21323 .38592 L closepath p .635 .345 .46 r
F P
0 g
s
.2247 .42978 m .20732 .4023 L .22387 .41248 L closepath p .625 .356 .489 r
F P
0 g
s
.226 .50239 m .20743 .50386 L .21321 .4884 L closepath p .645 .532 .711 r
F P
0 g
s
.2281 .52462 m .20743 .50386 L .226 .50239 L closepath p .64 .489 .662 r
F P
0 g
s
.21674 .34354 m .21828 .35805 L .20943 .36194 L closepath p .505 .275 .499 r
F P
0 g
s
.21828 .35805 m .22027 .37517 L .20943 .36194 L closepath p .577 .423 .636 r
F P
0 g
s
.20943 .36194 m .22027 .37517 L .21676 .38053 L closepath p .764 .434 .427 r
F P
0 g
s
.2628 .6237 m .21212 .62938 L .2509 .61748 L closepath p .617 .788 .964 r
F P
0 g
s
.2628 .6237 m .22057 .63608 L .21212 .62938 L closepath p .631 .769 .946 r
F P
0 g
s
.21216 .33074 m .21441 .31877 L p .2122 .33073 L .33 .177 .518 r
F P
0 g
s
.2122 .33073 m .21621 .31825 L .21465 .33031 L p .33 .177 .518 r
F P
0 g
s
.21441 .31877 m .21621 .31825 L .2122 .33073 L p .33 .177 .518 r
F P
0 g
s
.22483 .49017 m .21321 .4884 L .217 .4718 L closepath p .639 .546 .733 r
F P
0 g
s
.226 .50239 m .21321 .4884 L .22483 .49017 L closepath p .643 .532 .713 r
F P
0 g
s
.23071 .39768 m .21323 .38592 L .21676 .38053 L closepath p .695 .347 .383 r
F P
0 g
s
.22387 .41248 m .21323 .38592 L .23071 .39768 L closepath p .682 .368 .436 r
F P
0 g
s
.22831 .47296 m .217 .4718 L .21421 .45455 L closepath p .631 .534 .727 r
F P
0 g
s
.22831 .47296 m .21421 .45455 L .22842 .45517 L closepath p .65 .539 .715 r
F P
0 g
s
.22214 .30577 m .21605 .31128 L .2143 .3118 L closepath p .126 0 .001 r
F P
0 g
s
.21441 .31877 m .2143 .3118 L .21605 .31128 L closepath p .345 .218 .561 r
F P
0 g
s
.21441 .31877 m .21605 .31128 L .21621 .31825 L closepath p .345 .218 .561 r
F P
0 g
s
.21465 .33031 m .22061 .31771 L .22085 .32957 L closepath p .51 .334 .578 r
F P
0 g
s
.21465 .33031 m .21621 .31825 L .22061 .31771 L closepath p .51 .334 .578 r
F P
0 g
s
.22214 .30577 m .22033 .31076 L .21605 .31128 L closepath p .425 .084 .281 r
F P
0 g
s
.21621 .31825 m .21605 .31128 L .22033 .31076 L closepath p .514 .349 .596 r
F P
0 g
s
.21621 .31825 m .22033 .31076 L .22061 .31771 L closepath p .514 .349 .596 r
F P
0 g
s
.24899 .59755 m .23148 .60042 L .21668 .57346 L closepath p .602 .603 .825 r
F P
0 g
s
.24899 .59755 m .21668 .57346 L .23263 .57433 L closepath p .573 .601 .843 r
F P
0 g
s
.23791 .33102 m .23071 .34224 L .21674 .34354 L closepath p .45 .041 .171 r
F P
0 g
s
.23791 .33102 m .21674 .34354 L .21825 .33535 L closepath p .445 .035 .166 r
F P
0 g
s
.23071 .34224 m .21828 .35805 L .21674 .34354 L closepath p .56 .314 .5 r
F P
0 g
s
.23535 .3898 m .21676 .38053 L .22027 .37517 L closepath p .664 .309 .363 r
F P
0 g
s
.23071 .39768 m .21676 .38053 L .23535 .3898 L closepath p .653 .347 .442 r
F P
0 g
s
.22483 .49017 m .217 .4718 L .22831 .47296 L closepath p .629 .541 .736 r
F P
0 g
s
.21825 .33535 m .22085 .32957 L .22828 .32753 L closepath p .479 .072 .19 r
F P
0 g
s
.21825 .33535 m .22828 .32753 L .23791 .33102 L closepath p .437 .021 .151 r
F P
0 g
s
.23071 .34224 m .2311 .35679 L .21828 .35805 L closepath p .563 .318 .502 r
F P
0 g
s
.2311 .35679 m .22027 .37517 L .21828 .35805 L closepath p .592 .434 .636 r
F P
0 g
s
.2311 .35679 m .23307 .37172 L .22027 .37517 L closepath p .549 .382 .608 r
F P
0 g
s
.22027 .37517 m .23307 .37172 L .2389 .38784 L p .561 .415 .641 r
F P
0 g
s
.23535 .3898 m .22027 .37517 L .23859 .38802 L p .705 .286 .249 r
F P
0 g
s
.23859 .38802 m .22027 .37517 L p .2389 .38784 L .561 .415 .641 r
F P
0 g
s
.22214 .30577 m .22546 .30935 L .22033 .31076 L closepath p .56 .213 .337 r
F P
0 g
s
.22061 .31771 m .22033 .31076 L .22546 .30935 L closepath p .681 .479 .606 r
F P
0 g
s
.26834 .62736 m .22057 .63608 L .2628 .6237 L closepath p .634 .746 .929 r
F P
0 g
s
.26834 .62736 m .22917 .64117 L .22057 .63608 L closepath p .623 .775 .953 r
F P
0 g
s
.22061 .31771 m .22546 .30935 L .22586 .31627 L closepath p .681 .479 .606 r
F P
0 g
s
.22085 .32957 m .22061 .31771 L .22586 .31627 L closepath p .676 .465 .592 r
F P
0 g
s
.22085 .32957 m .22586 .31627 L .22828 .32753 L closepath p .676 .465 .592 r
F P
0 g
s
.23528 .6434 m .24048 .64337 L .2212 .64828 L p .402 .855 .882 r
F P
0 g
s
.22214 .30577 m .22885 .30849 L .22546 .30935 L closepath p .651 .201 .169 r
F P
0 g
s
.23713 .40852 m .22387 .41248 L .23071 .39768 L closepath p .395 .069 .291 r
F P
0 g
s
.23425 .42693 m .22387 .41248 L .23713 .40852 L closepath p .475 .232 .465 r
F P
0 g
s
.23425 .42693 m .2247 .42978 L .22387 .41248 L closepath p .453 .227 .479 r
F P
0 g
s
.23559 .45351 m .2247 .42978 L .23425 .42693 L closepath p .519 .392 .648 r
F P
0 g
s
.23559 .45351 m .22842 .45517 L .2247 .42978 L closepath p .557 .411 .638 r
F P
0 g
s
.24063 .49084 m .22483 .49017 L .22831 .47296 L closepath p .557 .471 .713 r
F P
0 g
s
.241 .50864 m .226 .50239 L .22483 .49017 L closepath p .501 .425 .703 r
F P
0 g
s
.241 .50864 m .22483 .49017 L .24063 .49084 L closepath p .559 .437 .67 r
F P
0 g
s
.22586 .31627 m .22546 .30935 L .22885 .30849 L closepath p .859 .617 .572 r
F P
0 g
s
.22586 .31627 m .22885 .30849 L .22935 .31538 L closepath p .859 .617 .572 r
F P
0 g
s
.22828 .32753 m .22586 .31627 L .22935 .31538 L closepath p .843 .57 .526 r
F P
0 g
s
.24685 .54397 m .226 .50239 L .241 .50864 L closepath p .53 .399 .646 r
F P
0 g
s
.24685 .54397 m .2281 .52462 L .226 .50239 L closepath p .501 .386 .654 r
F P
0 g
s
.25032 .57091 m .2281 .52462 L .24685 .54397 L closepath p .464 .38 .677 r
F P
0 g
s
.25032 .57091 m .23058 .55148 L .2281 .52462 L closepath p .6 .444 .643 r
F P
0 g
s
.22828 .32753 m .22935 .31538 L .2332 .32626 L closepath p .843 .57 .526 r
F P
0 g
s
.23791 .33102 m .22828 .32753 L .2332 .32626 L closepath p .425 0 0 r
F P
0 g
s
.23762 .47635 m .22831 .47296 L .22842 .45517 L closepath p .577 .479 .706 r
F P
0 g
s
.24063 .49084 m .22831 .47296 L .23762 .47635 L closepath p .578 .477 .704 r
F P
0 g
s
.23762 .47635 m .22842 .45517 L .23559 .45351 L closepath p .576 .478 .706 r
F P
0 g
s
.27612 .63146 m .22917 .64117 L .26834 .62736 L closepath p .621 .794 .965 r
F P
0 g
s
.27612 .63146 m .23527 .64297 L .22917 .64117 L closepath p .634 .764 .941 r
F P
0 g
s
.2663 .58675 m .23058 .55148 L .25032 .57091 L closepath p 0 .016 .517 r
F P
0 g
s
.2663 .58675 m .23263 .57433 L .23058 .55148 L closepath p .606 .501 .708 r
F P
0 g
s
.25425 .32963 m .24529 .33953 L .23071 .34224 L closepath p .499 .11 .233 r
F P
0 g
s
.25425 .32963 m .23071 .34224 L .23791 .33102 L closepath p .449 .04 .17 r
F P
0 g
s
.24529 .33953 m .2311 .35679 L .23071 .34224 L closepath p .614 .357 .505 r
F P
0 g
s
.24717 .39188 m .23071 .39768 L .23535 .3898 L closepath p .303 0 .227 r
F P
0 g
s
.23713 .40852 m .23071 .39768 L .24717 .39188 L closepath p .348 .054 .316 r
F P
0 g
s
.24529 .33953 m .24032 .35671 L .2311 .35679 L closepath p .596 .335 .491 r
F P
0 g
s
.24032 .35671 m .23307 .37172 L .2311 .35679 L closepath p .614 .428 .607 r
F P
0 g
s
.27927 .61222 m .23148 .60042 L .24899 .59755 L closepath p .583 .732 .945 r
F P
0 g
s
.27927 .61222 m .2509 .61748 L .23148 .60042 L closepath p .598 .728 .936 r
F P
0 g
s
.23163 .32183 m .2328 .31799 L p .2321 .32317 L .328 .141 .467 r
F P
0 g
s
.2328 .31799 m .23224 .32356 L p .2321 .32317 L .328 .141 .467 r
F P
0 g
s
.23224 .32356 m .2328 .31799 L p .23298 .32566 L .262 .086 .445 r
F P
0 g
s
.28667 .59773 m .23263 .57433 L .2663 .58675 L closepath p .063 .523 .896 r
F P
0 g
s
.28667 .59773 m .24899 .59755 L .23263 .57433 L closepath p .618 .619 .83 r
F P
0 g
s
.23585 .31763 m .2328 .31799 L .23664 .30368 L p .259 .074 .43 r
F P
0 g
s
.2328 .31799 m .23585 .31763 L .23338 .32627 L p .262 .086 .445 r
F P
0 g
s
.23318 .3262 m .2328 .31799 L .23333 .32627 L .262 .086 .445 r
F
.2331 .326 m .2328 .31799 L .23318 .3262 L F
.23333 .32627 m .2328 .31799 L .23338 .32627 L F
.23298 .32566 m .2328 .31799 L .2331 .326 L F
.24032 .35671 m .23946 .37325 L .23307 .37172 L closepath p .595 .407 .597 r
F P
0 g
s
.23307 .37172 m .23946 .37325 L .23911 .38773 L p .589 .431 .635 r
F P
0 g
s
.2389 .38784 m .23307 .37172 L p .23911 .38773 L .589 .431 .635 r
F P
0 g
s
.23318 .3262 m .23333 .32627 L .2332 .32626 L .262 .086 .445 r
F
.23791 .33102 m .2332 .32626 L .24318 .32668 L closepath p .417 0 0 r
F P
0 g
s
.23338 .32627 m .23585 .31763 L p .23414 .3263 L .538 .351 .575 r
F P
0 g
s
.23585 .31763 m .23707 .32642 L p .23414 .3263 L .538 .351 .575 r
F P
0 g
s
.25762 .38436 m .23425 .42693 L .23713 .40852 L closepath p .038 0 .303 r
F P
0 g
s
.25144 .44216 m .23425 .42693 L .25762 .38436 L closepath p .408 .283 .597 r
F P
0 g
s
.25144 .44216 m .23559 .45351 L .23425 .42693 L closepath p .377 .283 .62 r
F P
0 g
s
.28075 .63207 m .23527 .64297 L .27612 .63146 L closepath p .438 .867 .93 r
F P
0 g
s
.28075 .63207 m .23528 .6434 L .23527 .64297 L closepath p .304 .799 .825 r
F P
0 g
s
.28337 .63151 m .23528 .6434 L .28075 .63207 L closepath p .195 .719 .712 r
F P
0 g
s
.28337 .63151 m .24048 .64337 L .23528 .6434 L closepath p .368 .836 .872 r
F P
0 g
s
.24717 .39188 m .23535 .3898 L .25379 .37964 L closepath p .302 .006 .286 r
F P
0 g
s
.23942 .64364 m .24808 .6426 L .23555 .64462 L p .323 .801 .904 r
F P
0 g
s
.25489 .47017 m .23559 .45351 L .25144 .44216 L closepath p .41 .34 .668 r
F P
0 g
s
.25489 .47017 m .23762 .47635 L .23559 .45351 L closepath p .393 .34 .679 r
F P
0 g
s
.23945 .31633 m .23585 .31763 L .23687 .29964 L closepath p .54 .375 .606 r
F P
0 g
s
.23585 .31763 m .23945 .31633 L .2384 .32648 L p .526 .342 .574 r
F P
0 g
s
.23753 .32644 m .23585 .31763 L .23825 .32647 L .526 .342 .574 r
F
.23825 .32647 m .23585 .31763 L .2384 .32648 L F
.23732 .32643 m .23585 .31763 L .23753 .32644 L F
.23729 .32643 m .23585 .31763 L .23732 .32643 L F
.23727 .32643 m .23585 .31763 L .23729 .32643 L F
.23722 .32643 m .23585 .31763 L .23727 .32643 L F
.23707 .32642 m .23585 .31763 L p .23722 .32643 L F P
0 g
s
.2364 .27122 m .24923 .26835 L .24327 .28104 L closepath p .6 .379 .553 r
F P
0 g
s
.2364 .27122 m .24327 .28104 L .23895 .28085 L closepath p .393 .317 .652 r
F P
0 g
s
.23888 .28145 m .2364 .27122 L .23895 .28085 L p 0 0 .226 r
F P
0 g
s
.23895 .28085 m .24328 .29962 L .23687 .29964 L closepath p .486 .364 .638 r
F P
0 g
s
.23945 .31633 m .23687 .29964 L .24328 .29962 L closepath p .484 .337 .605 r
F P
0 g
s
.25762 .38436 m .23713 .40852 L .26417 .36596 L closepath p 0 0 .148 r
F P
0 g
s
.26417 .36596 m .23713 .40852 L .24717 .39188 L closepath p 0 0 0 r
F P
0 g
s
.2639 .49437 m .23762 .47635 L .25489 .47017 L closepath p .41 .386 .722 r
F P
0 g
s
.2639 .49437 m .24063 .49084 L .23762 .47635 L closepath p .438 .383 .699 r
F P
0 g
s
.24318 .32668 m .25509 .32411 L .23791 .33102 L closepath p .104 0 0 r
F P
0 g
s
.23791 .33102 m .25509 .32411 L .25963 .32584 L closepath p .412 0 .062 r
F P
0 g
s
.23791 .33102 m .25963 .32584 L .25425 .32963 L closepath p .378 0 .007 r
F P
0 g
s
.2384 .32648 m .23945 .31633 L p .23914 .32651 L .603 .409 .592 r
F P
0 g
s
.23895 .28085 m .24327 .28104 L .24328 .29962 L closepath p .393 .303 .634 r
F P
0 g
s
.23911 .38773 m .23946 .37325 L p .24296 .38561 L .603 .443 .637 r
F P
0 g
s
.23945 .31633 m .24288 .32667 L p .23914 .32651 L .603 .409 .592 r
F P
0 g
s
.24048 .64337 m .24808 .6426 L .23942 .64364 L p .109 .662 .727 r
F P
0 g
s
.24627 .31685 m .23945 .31633 L .24328 .29962 L closepath p .558 .407 .633 r
F P
0 g
s
.2449 .32631 m .23945 .31633 L .24503 .32628 L .558 .384 .601 r
F
.24318 .32668 m .23945 .31633 L .2449 .32631 L F
.23945 .31633 m .24627 .31685 L .24503 .32628 L p F P
0 g
s
.24288 .32667 m .23945 .31633 L p .24318 .32668 L .558 .384 .601 r
F P
0 g
s
.26031 .35703 m .23946 .37325 L .24032 .35671 L closepath p .603 .413 .598 r
F P
0 g
s
.23946 .37325 m .26031 .35703 L .25873 .37169 L p .607 .42 .603 r
F P
0 g
s
.25873 .37169 m .25815 .37264 L .23946 .37325 L .607 .42 .603 r
F
.25815 .37264 m .2569 .37463 L .23946 .37325 L F
.25579 .37642 m .23946 .37325 L p .2569 .37463 L F P
0 g
s
.23946 .37325 m .25579 .37642 L p .24296 .38561 L .603 .443 .637 r
F P
0 g
s
.26685 .33831 m .2637 .34282 L .24032 .35671 L closepath p .303 0 .137 r
F P
0 g
s
.26685 .33831 m .24032 .35671 L .24529 .33953 L closepath p .605 .344 .494 r
F P
0 g
s
.26031 .35703 m .24032 .35671 L .2637 .34282 L closepath p .594 .356 .526 r
F P
0 g
s
.28979 .63228 m .24048 .64337 L .28337 .63151 L closepath p .421 .86 .917 r
F P
0 g
s
.28979 .63228 m .24808 .6426 L .24048 .64337 L closepath p .304 .799 .83 r
F P
0 g
s
.26503 .5224 m .241 .50864 L .24063 .49084 L closepath p .515 .403 .664 r
F P
0 g
s
.26503 .5224 m .24063 .49084 L .2639 .49437 L closepath p .439 .381 .695 r
F P
0 g
s
.26142 .54642 m .241 .50864 L .26503 .5224 L closepath p .497 .41 .687 r
F P
0 g
s
.26142 .54642 m .24685 .54397 L .241 .50864 L closepath p .623 .463 .644 r
F P
0 g
s
.24373 .38518 m .24296 .38561 L .25579 .37642 L .603 .443 .637 r
F
.24923 .26835 m .25103 .2861 L .24327 .28104 L closepath p .691 .483 .6 r
F P
0 g
s
.24327 .28104 m .2462 .29828 L .24328 .29962 L closepath p .56 .438 .671 r
F P
0 g
s
.24327 .28104 m .25103 .2861 L .2462 .29828 L closepath p .655 .504 .666 r
F P
0 g
s
.24627 .31685 m .24328 .29962 L .2462 .29828 L closepath p .541 .395 .633 r
F P
0 g
s
.25529 .37722 m .24373 .38518 L .25579 .37642 L .603 .443 .637 r
F
.24566 .38412 m .24373 .38518 L .25529 .37722 L F
.24503 .32628 m .24627 .31685 L .25043 .32512 L p .695 .508 .63 r
F P
0 g
s
.27684 .33257 m .24529 .33953 L .25425 .32963 L closepath p .598 .228 .311 r
F P
0 g
s
.27684 .33257 m .26685 .33831 L .24529 .33953 L closepath p .437 0 .03 r
F P
0 g
s
.25461 .37832 m .24566 .38412 L .25529 .37722 L .603 .443 .637 r
F
.25152 .38089 m .24566 .38412 L .25461 .37832 L F
.2462 .29828 m .25103 .2861 L .2576 .28887 L closepath p .74 .599 .701 r
F P
0 g
s
.2462 .29828 m .2576 .28887 L .25571 .29835 L closepath p .733 .586 .692 r
F P
0 g
s
.25422 .3183 m .2462 .29828 L .25571 .29835 L closepath p .73 .543 .639 r
F P
0 g
s
.25422 .3183 m .24627 .31685 L .2462 .29828 L closepath p .714 .536 .646 r
F P
0 g
s
.25043 .32512 m .24627 .31685 L p .25132 .32493 L .716 .516 .618 r
F P
0 g
s
.25132 .32493 m .24627 .31685 L .25323 .32451 L .716 .516 .618 r
F
.25323 .32451 m .24627 .31685 L .25332 .32449 L F
.24627 .31685 m .25422 .3183 L .25424 .3243 L p F P
0 g
s
.25332 .32449 m .24627 .31685 L .25424 .3243 L .716 .516 .618 r
F
.26571 .5711 m .25032 .57091 L .24685 .54397 L closepath p .674 .533 .684 r
F P
0 g
s
.26571 .5711 m .24685 .54397 L .26142 .54642 L closepath p .611 .514 .72 r
F P
0 g
s
.24698 .26885 m .24958 .26224 L p .24923 .26835 L 0 0 .092 r
F P
0 g
s
.26417 .36596 m .24717 .39188 L .2691 .35503 L closepath p 0 0 0 r
F P
0 g
s
.2691 .35503 m .24717 .39188 L .25379 .37964 L closepath p 0 0 0 r
F P
0 g
s
.29889 .60248 m .24899 .59755 L .28667 .59773 L closepath p .446 .77 .997 r
F P
0 g
s
.29889 .60248 m .27927 .61222 L .24899 .59755 L closepath p .612 .733 .932 r
F P
0 g
s
.24923 .26835 m .25001 .25468 L .25879 .25757 L closepath p .529 .423 .678 r
F P
0 g
s
.24923 .26835 m .25879 .25757 L .25583 .27359 L closepath p .515 .404 .665 r
F P
0 g
s
.24923 .26835 m .25583 .27359 L .25103 .2861 L closepath p .572 .396 .604 r
F P
0 g
s
.25001 .25468 m .25159 .2447 L .26269 .24702 L closepath p .539 .368 .598 r
F P
0 g
s
.25001 .25468 m .26269 .24702 L .25879 .25757 L closepath p .55 .389 .615 r
F P
0 g
s
.28313 .58354 m .2663 .58675 L .25032 .57091 L closepath p .702 .572 .706 r
F P
0 g
s
.28313 .58354 m .25032 .57091 L .26571 .5711 L closepath p .67 .605 .774 r
F P
0 g
s
.28985 .61629 m .2509 .61748 L .27927 .61222 L closepath p .542 .815 .997 r
F P
0 g
s
.28985 .61629 m .2628 .6237 L .2509 .61748 L closepath p .608 .787 .967 r
F P
0 g
s
.25583 .27359 m .2576 .28887 L .25103 .2861 L closepath p .747 .589 .679 r
F P
0 g
s
.2758 .40646 m .25144 .44216 L .25762 .38436 L closepath p .379 .258 .586 r
F P
0 g
s
.27952 .45216 m .25144 .44216 L .2758 .40646 L closepath p .505 .417 .689 r
F P
0 g
s
.27952 .45216 m .25489 .47017 L .25144 .44216 L closepath p .509 .413 .682 r
F P
0 g
s
.25152 .38089 m .25299 .38 L p .25379 .37964 L 0 0 0 r
F P
0 g
s
.2539 .37946 m .25152 .38089 L p .25461 .37832 L .603 .443 .637 r
F P
0 g
s
.26269 .24702 m .25159 .2447 L .25568 .23696 L closepath p .533 .393 .636 r
F P
0 g
s
.25299 .38 m .2539 .37946 L p .25379 .37964 L 0 0 0 r
F P
0 g
s
.25422 .3183 m .25571 .29835 L .25613 .30085 L p .862 .662 .63 r
F P
0 g
s
.25587 .30887 m .25422 .3183 L .25613 .30085 L .862 .662 .63 r
F
.25579 .31157 m .25422 .3183 L .25587 .30887 L F
.25575 .31272 m .25422 .3183 L .25579 .31157 L F
.25503 .3184 m .25422 .3183 L p .25575 .31272 L F P
0 g
s
.25488 .31961 m .25422 .3183 L .25492 .31927 L .862 .648 .612 r
F
.2548 .32023 m .25422 .3183 L p .25488 .31961 L F P
0 g
s
.25492 .31927 m .25422 .3183 L .25493 .31919 L .862 .648 .612 r
F
.25422 .3183 m .25503 .3184 L p .25493 .31919 L F P
0 g
s
.25422 .3183 m .2548 .32023 L p .25469 .32106 L .89 .661 .586 r
F P
0 g
s
.25424 .3243 m .25422 .3183 L p .25469 .32106 L .89 .661 .586 r
F P
0 g
s
.25432 .32402 m .25424 .3243 L .25469 .32106 L .89 .661 .586 r
F
.25428 .32428 m .25424 .3243 L .25432 .32402 L F
.25428 .32429 m .25424 .3243 L .25428 .32428 L F
.27684 .33257 m .25425 .32963 L .25963 .32584 L closepath p .567 .136 .183 r
F P
0 g
s
.25428 .32429 m .25439 .32347 L p .25432 .32428 L 0 0 .203 r
F P
0 g
s
.25439 .32347 m .25433 .32428 L .25432 .32428 L 0 0 .203 r
F
.25433 .32428 m .25442 .3235 L .25433 .32428 L F
.25439 .32347 m .25442 .3235 L .25433 .32428 L F
.25433 .32428 m .25643 .31171 L .25509 .32411 L p F P
0 g
s
.25446 .32353 m .25433 .32428 L p .25442 .3235 L 0 0 .203 r
F P
0 g
s
.25439 .32347 m .25575 .31272 L p .25442 .3235 L 0 0 .203 r
F P
0 g
s
.25442 .3235 m .25575 .31272 L .25446 .32353 L 0 0 .203 r
F
.25575 .31272 m .25643 .31171 L .25446 .32353 L p F P
0 g
s
.28034 .47642 m .25489 .47017 L .27952 .45216 L closepath p .531 .453 .713 r
F P
0 g
s
.28034 .47642 m .2639 .49437 L .25489 .47017 L closepath p .53 .456 .717 r
F P
0 g
s
.25509 .32411 m .25643 .31171 L .25965 .31291 L closepath p .31 .197 .559 r
F P
0 g
s
.25509 .32411 m .25965 .31291 L .25963 .32584 L closepath p .31 .197 .559 r
F P
0 g
s
.26537 .23355 m .26269 .24702 L .25568 .23696 L closepath p .595 .411 .602 r
F P
0 g
s
.26525 .28703 m .25946 .30042 L .25571 .29835 L closepath p .808 .665 .706 r
F P
0 g
s
.26525 .28703 m .25571 .29835 L .2576 .28887 L closepath p .802 .656 .701 r
F P
0 g
s
.25571 .29835 m .25946 .30042 L .25942 .30311 L p .829 .641 .648 r
F P
0 g
s
.25852 .3036 m .25571 .29835 L .25942 .30311 L .829 .641 .648 r
F
.25675 .30456 m .25571 .29835 L p .25684 .30451 L F P
0 g
s
.25795 .30391 m .25571 .29835 L .25815 .3038 L .829 .641 .648 r
F
.25723 .3043 m .25571 .29835 L .25795 .30391 L F
.25815 .3038 m .25571 .29835 L .25828 .30373 L F
.25828 .30373 m .25571 .29835 L .25852 .3036 L F
.25684 .30451 m .25571 .29835 L .25696 .30445 L F
.257 .30443 m .25571 .29835 L .25723 .3043 L F
.25696 .30445 m .25571 .29835 L .257 .30443 L F
.25575 .31272 m .25664 .30462 L .25643 .31171 L closepath p 0 0 .255 r
F P
0 g
s
.25575 .31272 m .25598 .30561 L .25664 .30462 L closepath p 0 0 .255 r
F P
0 g
s
.25583 .27359 m .25879 .25757 L .27342 .26381 L closepath p .722 .601 .721 r
F P
0 g
s
.25583 .27359 m .27342 .26381 L .26406 .27544 L closepath p .722 .601 .72 r
F P
0 g
s
.25583 .27359 m .26406 .27544 L .2576 .28887 L closepath p .73 .577 .682 r
F P
0 g
s
.2566 .30468 m .25598 .30561 L .25613 .30085 L .862 .662 .63 r
F
.25613 .30085 m .25669 .30419 L p .2566 .30468 L F P
0 g
s
.25643 .31171 m .25664 .30462 L .25978 .30577 L closepath p .315 .191 .547 r
F P
0 g
s
.25643 .31171 m .25978 .30577 L .25965 .31291 L closepath p .315 .191 .547 r
F P
0 g
s
.25664 .30463 m .2566 .30468 L .25669 .30419 L .862 .662 .63 r
F
.25664 .30462 m .25664 .30463 L .25669 .30419 L F
.26336 .30097 m .25978 .30577 L .25664 .30462 L closepath p .402 0 .154 r
F P
0 g
s
.25667 .30461 m .25664 .30462 L .25669 .30419 L .862 .662 .63 r
F
.25671 .30459 m .25667 .30461 L .25669 .30419 L F
.25669 .30419 m .25675 .30456 L p .25671 .30459 L F P
0 g
s
.26989 .27716 m .2576 .28887 L .26406 .27544 L closepath p .79 .651 .709 r
F P
0 g
s
.26989 .27716 m .26525 .28703 L .2576 .28887 L closepath p .807 .683 .727 r
F P
0 g
s
.2758 .40646 m .25762 .38436 L .27757 .36441 L closepath p .425 .267 .562 r
F P
0 g
s
.27757 .36441 m .25762 .38436 L .26417 .36596 L closepath p .196 0 .312 r
F P
0 g
s
.25873 .37169 m .26031 .35703 L p .26155 .36717 L .639 .449 .609 r
F P
0 g
s
.25879 .25757 m .26269 .24702 L .27811 .25763 L closepath p .738 .593 .694 r
F P
0 g
s
.25879 .25757 m .27811 .25763 L .27342 .26381 L closepath p .738 .583 .682 r
F P
0 g
s
.25942 .30311 m .25946 .30042 L .25951 .30306 L .957 .823 .389 r
F
.26684 .29026 m .25946 .30042 L .26525 .28703 L closepath p .969 .917 .696 r
F P
0 g
s
.26684 .29026 m .26023 .30263 L .25946 .30042 L closepath p .955 .951 .686 r
F P
0 g
s
.25951 .30306 m .25946 .30042 L .25955 .30304 L .957 .823 .389 r
F
.25955 .30304 m .25946 .30042 L .25977 .30292 L F
.25946 .30042 m .25985 .30154 L p .25977 .30292 L F P
0 g
s
.25963 .32584 m .2663 .31227 L .26901 .32493 L closepath p .729 .533 .626 r
F P
0 g
s
.25963 .32584 m .25965 .31291 L .2663 .31227 L closepath p .729 .533 .626 r
F P
0 g
s
.27684 .33257 m .25963 .32584 L .26901 .32493 L closepath p .567 .13 .171 r
F P
0 g
s
.25965 .31291 m .26625 .30515 L .2663 .31227 L closepath p .727 .525 .617 r
F P
0 g
s
.25965 .31291 m .25978 .30577 L .26625 .30515 L closepath p .727 .525 .617 r
F P
0 g
s
.25977 .30292 m .25985 .30154 L .25983 .30289 L .957 .823 .389 r
F
.26336 .30097 m .26625 .30515 L .25978 .30577 L closepath p .628 .248 .304 r
F P
0 g
s
.25983 .30289 m .25985 .30154 L .25985 .30287 L .957 .823 .389 r
F
.25985 .30287 m .25985 .30154 L .25988 .30286 L F
.25985 .30154 m .26023 .30263 L .26022 .30267 L p F P
0 g
s
.25988 .30286 m .25985 .30154 L .26022 .30267 L .957 .823 .389 r
F
.26906 .3458 m .26031 .35703 L .2637 .34282 L closepath p .525 .291 .502 r
F P
0 g
s
.26906 .3458 m .26594 .35594 L .26031 .35703 L closepath p .633 .43 .59 r
F P
0 g
s
.26031 .35703 m .26594 .35594 L .26211 .36627 L p .639 .449 .609 r
F P
0 g
s
.26211 .36627 m .26155 .36717 L .26031 .35703 L .639 .449 .609 r
F
.2898 .53106 m .26142 .54642 L .26503 .5224 L closepath p .672 .572 .734 r
F P
0 g
s
.29086 .55281 m .26142 .54642 L .2898 .53106 L closepath p .7 .643 .789 r
F P
0 g
s
.29086 .55281 m .26571 .5711 L .26142 .54642 L closepath p .724 .592 .707 r
F P
0 g
s
.26211 .36627 m .26594 .35594 L p .26355 .36395 L .561 .366 .574 r
F P
0 g
s
.26537 .23355 m .27536 .23293 L .26269 .24702 L closepath p .645 .458 .616 r
F P
0 g
s
.28796 .23718 m .26269 .24702 L .27536 .23293 L closepath p .688 .521 .655 r
F P
0 g
s
.28796 .23718 m .28274 .24847 L .26269 .24702 L closepath p .706 .57 .699 r
F P
0 g
s
.27811 .25763 m .26269 .24702 L .28274 .24847 L closepath p .702 .606 .746 r
F P
0 g
s
.29765 .62159 m .2628 .6237 L .28985 .61629 L closepath p .619 .737 .931 r
F P
0 g
s
.29765 .62159 m .26834 .62736 L .2628 .6237 L closepath p .615 .743 .937 r
F P
0 g
s
.26336 .30097 m .27022 .30529 L .27149 .30611 L closepath p .387 0 0 r
F P
0 g
s
.26336 .30097 m .27149 .30611 L .26625 .30515 L closepath p .72 .261 .16 r
F P
0 g
s
.26594 .35594 m .26896 .35526 L p .26355 .36395 L .561 .366 .574 r
F P
0 g
s
.2637 .34282 m .26685 .33831 L .27481 .33699 L closepath p .395 .016 .197 r
F P
0 g
s
.2637 .34282 m .27481 .33699 L .26906 .3458 L closepath p .555 .259 .419 r
F P
0 g
s
.28705 .51313 m .2639 .49437 L .28034 .47642 L closepath p .565 .509 .751 r
F P
0 g
s
.28705 .51313 m .26503 .5224 L .2639 .49437 L closepath p .6 .508 .722 r
F P
0 g
s
.27932 .26588 m .26406 .27544 L .27342 .26381 L closepath p .805 .732 .784 r
F P
0 g
s
.27932 .26588 m .26989 .27716 L .26406 .27544 L closepath p .781 .669 .739 r
F P
0 g
s
.28547 .34509 m .26417 .36596 L .2691 .35503 L closepath p 0 0 .048 r
F P
0 g
s
.27757 .36441 m .26417 .36596 L .28547 .34509 L closepath p .186 0 .261 r
F P
0 g
s
.2898 .53106 m .26503 .5224 L .28705 .51313 L closepath p .633 .608 .807 r
F P
0 g
s
.26684 .29026 m .26525 .28703 L .26989 .27716 L closepath p .958 .923 .726 r
F P
0 g
s
.3084 .57036 m .26571 .5711 L .29086 .55281 L closepath p .705 .555 .681 r
F P
0 g
s
.3084 .57036 m .28313 .58354 L .26571 .5711 L closepath p .706 .598 .733 r
F P
0 g
s
.27716 .3458 m .26594 .35594 L .26906 .3458 L closepath p .66 .457 .6 r
F P
0 g
s
.26594 .35594 m .27716 .3458 L .27861 .34704 L p .516 .248 .447 r
F P
0 g
s
.27861 .34704 m .2717 .35285 L .26594 .35594 L .516 .248 .447 r
F
.2717 .35285 m .27028 .35404 L .26594 .35594 L F
.27028 .35404 m .26914 .355 L .26594 .35594 L F
.26896 .35526 m .26594 .35594 L p .2691 .35503 L F P
0 g
s
.26911 .355 m .2691 .35503 L .26594 .35594 L .516 .248 .447 r
F
.2663 .31227 m .26625 .30515 L .27149 .30611 L closepath p .813 .595 .607 r
F P
0 g
s
.2663 .31227 m .27149 .30611 L .27167 .31324 L closepath p .813 .595 .607 r
F P
0 g
s
.26901 .32493 m .2663 .31227 L .27167 .31324 L closepath p .813 .586 .593 r
F P
0 g
s
.29591 .59293 m .2663 .58675 L .28313 .58354 L closepath p .715 .705 .84 r
F P
0 g
s
.29591 .59293 m .28667 .59773 L .2663 .58675 L closepath p .691 .727 .877 r
F P
0 g
s
.26684 .29026 m .26989 .27716 L .2723 .27815 L closepath p .958 .924 .726 r
F P
0 g
s
.26685 .33831 m .27405 .33418 L p .27481 .33699 L .624 .206 .229 r
F P
0 g
s
.30354 .62675 m .26834 .62736 L .29765 .62159 L closepath p .614 .746 .94 r
F P
0 g
s
.30354 .62675 m .27612 .63146 L .26834 .62736 L closepath p .566 .782 .98 r
F P
0 g
s
.26901 .32493 m .27167 .31324 L .27659 .32629 L closepath p .813 .586 .593 r
F P
0 g
s
.26901 .32493 m .27659 .32629 L .28467 .32976 L closepath p .538 0 0 r
F P
0 g
s
.26901 .32493 m .28467 .32976 L .27684 .33257 L closepath p .623 .108 .002 r
F P
0 g
s
.26906 .3458 m .27481 .33699 L .27716 .3458 L closepath p .639 .357 .475 r
F P
0 g
s
.28547 .34509 m .2691 .35503 L .29665 .33189 L closepath p 0 0 0 r
F P
0 g
s
.26914 .355 m .2691 .35503 L .26911 .355 L .516 .248 .447 r
F
.27956 .27024 m .26989 .27716 L .27932 .26588 L closepath p .902 .937 .825 r
F P
0 g
s
.27956 .27024 m .2723 .27815 L .26989 .27716 L closepath p .889 .965 .835 r
F P
0 g
s
.27811 .25763 m .27932 .26588 L .27342 .26381 L closepath p .822 .712 .744 r
F P
0 g
s
.27405 .33418 m .27684 .33257 L p .27491 .33698 L .624 .206 .229 r
F P
0 g
s
.27405 .33418 m .27491 .33698 L .27481 .33699 L .624 .206 .229 r
F
.27716 .3458 m .27481 .33699 L .28484 .336 L closepath p .653 .366 .471 r
F P
0 g
s
.27491 .33698 m .27684 .33257 L .27768 .33671 L .624 .206 .229 r
F
.30169 .23966 m .28796 .23718 L .27536 .23293 L closepath p .506 0 0 r
F P
0 g
s
.31437 .37972 m .2758 .40646 L .27757 .36441 L closepath p .535 .358 .587 r
F P
0 g
s
.3144 .41687 m .2758 .40646 L .31437 .37972 L closepath p .573 .423 .639 r
F P
0 g
s
.3144 .41687 m .27952 .45216 L .2758 .40646 L closepath p .557 .456 .696 r
F P
0 g
s
.3061 .62631 m .28075 .63207 L .27612 .63146 L p .371 .83 .914 r
F P
0 g
s
.27643 .32585 m .28467 .32976 L .27659 .32629 L p .407 0 0 r
F P
0 g
s
.27684 .33257 m .28467 .32976 L .29079 .33143 L p .558 .025 0 r
F P
0 g
s
.28939 .33251 m .27684 .33257 L .29079 .33143 L .558 .025 0 r
F
.28916 .33269 m .27684 .33257 L .28916 .33268 L F
.28916 .33268 m .27684 .33257 L .28939 .33251 L F
.28842 .33325 m .27684 .33257 L p .28916 .33269 L F P
0 g
s
.27684 .33257 m .28842 .33325 L p .28709 .33427 L .584 .065 0 r
F P
0 g
s
.28508 .33582 m .27684 .33257 L p .28709 .33427 L .584 .065 0 r
F P
0 g
s
.27684 .33257 m .2847 .33567 L p .27768 .33671 L .624 .206 .229 r
F P
0 g
s
.27956 .27024 m .27892 .27215 L .27685 .27319 L p .653 .969 .889 r
F P
0 g
s
.27716 .3458 m .28484 .336 L .2846 .34201 L p .551 .239 .392 r
F P
0 g
s
.2846 .34201 m .28014 .34576 L .27716 .3458 L .551 .239 .392 r
F
.27861 .34704 m .27716 .3458 L p .27937 .3464 L F P
0 g
s
.28014 .34576 m .27937 .3464 L .27716 .3458 L .551 .239 .392 r
F
.3195 .3531 m .27757 .36441 L .28547 .34509 L closepath p .482 .206 .418 r
F P
0 g
s
.31437 .37972 m .27757 .36441 L .3195 .3531 L closepath p .542 .348 .567 r
F P
0 g
s
.27768 .33671 m .2847 .33567 L .2794 .33654 L .624 .206 .229 r
F
.28274 .24847 m .28095 .26094 L .27811 .25763 L closepath p .819 .757 .794 r
F P
0 g
s
.27811 .25763 m .28095 .26094 L .27932 .26588 L closepath p .886 .753 .709 r
F P
0 g
s
.28278 .2634 m .27892 .27215 L .27956 .27024 L closepath p .705 .817 .361 r
F P
0 g
s
.28278 .2634 m .28219 .26637 L .27892 .27215 L closepath p .86 .954 .606 r
F P
0 g
s
.31212 .6109 m .27927 .61222 L .29889 .60248 L closepath p .62 .793 .965 r
F P
0 g
s
.31212 .6109 m .28985 .61629 L .27927 .61222 L closepath p .562 .821 .994 r
F P
0 g
s
.27956 .27024 m .27932 .26588 L .28095 .26094 L closepath p .965 .918 .576 r
F P
0 g
s
.2794 .33654 m .2847 .33567 L .28484 .336 L .624 .206 .229 r
F
.31027 .44758 m .27952 .45216 L .3144 .41687 L closepath p .56 .46 .698 r
F P
0 g
s
.31027 .44758 m .28034 .47642 L .27952 .45216 L closepath p .563 .478 .717 r
F P
0 g
s
.27956 .27024 m .28095 .26094 L .28278 .2634 L closepath p .968 .925 .693 r
F P
0 g
s
.30726 .47316 m .28034 .47642 L .31027 .44758 L closepath p .578 .502 .733 r
F P
0 g
s
.30726 .47316 m .28705 .51313 L .28034 .47642 L closepath p .579 .519 .751 r
F P
0 g
s
.2889 .63022 m .28337 .63151 L .28075 .63207 L p .175 .705 .704 r
F P
0 g
s
.28274 .24847 m .29059 .25143 L .28095 .26094 L closepath p .858 .794 .787 r
F P
0 g
s
.29059 .25143 m .28278 .2634 L .28095 .26094 L closepath p .893 .953 .835 r
F P
0 g
s
.29073 .25514 m .28219 .26637 L .28278 .2634 L closepath p .853 .998 .8 r
F P
0 g
s
.29073 .25514 m .29176 .25724 L .28219 .26637 L closepath p .872 .969 .853 r
F P
0 g
s
.30169 .23966 m .28274 .24847 L .28796 .23718 L closepath p .845 .744 .752 r
F P
0 g
s
.30169 .23966 m .29059 .25143 L .28274 .24847 L closepath p .855 .798 .795 r
F P
0 g
s
.29059 .25143 m .29073 .25514 L .28278 .2634 L closepath p .893 .955 .834 r
F P
0 g
s
.31603 .58056 m .28313 .58354 L .3084 .57036 L closepath p .685 .548 .693 r
F P
0 g
s
.31603 .58056 m .29591 .59293 L .28313 .58354 L closepath p .68 .705 .866 r
F P
0 g
s
.31785 .62721 m .28337 .63151 L .2889 .63022 L p .502 .867 .985 r
F P
0 g
s
.31785 .62721 m .28979 .63228 L .28337 .63151 L closepath p .31 .796 .888 r
F P
0 g
s
.2846 .34201 m .28484 .336 L p .28618 .34068 L .388 .111 .368 r
F P
0 g
s
.2847 .33567 m .28508 .33582 L p .28484 .336 L .624 .206 .229 r
F P
0 g
s
.28484 .336 m .29834 .32563 L .29357 .33447 L p .388 .111 .368 r
F P
0 g
s
.29357 .33447 m .28663 .34031 L .28484 .336 L .388 .111 .368 r
F
.28663 .34031 m .28618 .34068 L .28484 .336 L F
.32759 .33317 m .28547 .34509 L .29665 .33189 L closepath p .411 .066 .266 r
F P
0 g
s
.3195 .3531 m .28547 .34509 L .32759 .33317 L closepath p .482 .205 .416 r
F P
0 g
s
.30876 .59989 m .28667 .59773 L .29591 .59293 L closepath p .703 .787 .916 r
F P
0 g
s
.30876 .59989 m .29889 .60248 L .28667 .59773 L closepath p .643 .815 .966 r
F P
0 g
s
.30512 .52119 m .28705 .51313 L .30726 .47316 L closepath p .643 .595 .786 r
F P
0 g
s
.30512 .52119 m .2898 .53106 L .28705 .51313 L closepath p .627 .603 .807 r
F P
0 g
s
.31785 .62721 m .2889 .63022 L .30646 .62623 L .502 .867 .985 r
F
.32303 .55208 m .2898 .53106 L .30512 .52119 L closepath p .631 .614 .814 r
F P
0 g
s
.32303 .55208 m .29086 .55281 L .2898 .53106 L closepath p .657 .61 .789 r
F P
0 g
s
.32023 .61677 m .28985 .61629 L .31212 .6109 L closepath p .602 .729 .934 r
F P
0 g
s
.32023 .61677 m .29765 .62159 L .28985 .61629 L closepath p .597 .733 .939 r
F P
0 g
s
.30251 .2467 m .29059 .25143 L .30169 .23966 L closepath p .883 .872 .822 r
F P
0 g
s
.30251 .2467 m .29073 .25514 L .29059 .25143 L closepath p .857 .938 .875 r
F P
0 g
s
.29176 .25724 m .29073 .25514 L .30251 .2467 L closepath p .845 .965 .881 r
F P
0 g
s
.34387 .56195 m .29086 .55281 L .32303 .55208 L closepath p .658 .494 .651 r
F P
0 g
s
.34387 .56195 m .3084 .57036 L .29086 .55281 L closepath p .643 .552 .737 r
F P
0 g
s
.29357 .33447 m .29834 .32563 L .29557 .3328 L p .21 0 .254 r
F P
0 g
s
.2977 .32349 m .29532 .32408 L .30243 .32204 L closepath p 0 0 0 r
F P
0 g
s
.29532 .32408 m .2977 .32349 L .2968 .32681 L p .238 .167 .57 r
F P
0 g
s
.29565 .32769 m .29532 .32408 L .2968 .32681 L .238 .167 .57 r
F
.29546 .32784 m .29532 .32408 L p .29565 .32769 L F P
0 g
s
.29557 .3328 m .29834 .32563 L p .29665 .33189 L .404 .112 .351 r
F P
0 g
s
.32784 .59261 m .29591 .59293 L .31603 .58056 L closepath p .686 .723 .878 r
F P
0 g
s
.32784 .59261 m .30876 .59989 L .29591 .59293 L closepath p .639 .782 .95 r
F P
0 g
s
.32759 .33317 m .29665 .33189 L .33365 .32143 L closepath p .393 .006 .18 r
F P
0 g
s
.33365 .32143 m .29665 .33189 L .30867 .3272 L closepath p 0 0 0 r
F P
0 g
s
.29674 .33186 m .29665 .33189 L .29676 .33175 L .404 .112 .351 r
F
.29676 .33175 m .29665 .33189 L .29834 .32563 L F
.29755 .33154 m .29674 .33186 L .29834 .32563 L F
.2968 .32681 m .2977 .32349 L .29783 .32603 L p .238 .167 .57 r
F P
0 g
s
.29759 .33152 m .29755 .33154 L .29834 .32563 L .404 .112 .351 r
F
.29775 .33146 m .29759 .33152 L .29834 .32563 L F
.32719 .62206 m .29765 .62159 L .32023 .61677 L closepath p .593 .75 .952 r
F P
0 g
s
.32719 .62206 m .30354 .62675 L .29765 .62159 L closepath p .602 .742 .943 r
F P
0 g
s
.30436 .32369 m .2977 .32349 L .30243 .32204 L closepath p .438 0 0 r
F P
0 g
s
.30031 .32553 m .2977 .32349 L .30033 .32553 L .663 .495 .646 r
F
.30033 .32553 m .2977 .32349 L .30049 .32552 L F
.2978 .32547 m .2977 .32349 L p .30031 .32553 L F P
0 g
s
.30049 .32552 m .2977 .32349 L .3028 .3254 L .663 .495 .646 r
F
.2977 .32349 m .30436 .32369 L .3028 .3254 L p F P
0 g
s
.30651 .32804 m .29775 .33146 L .29834 .32563 L .404 .112 .351 r
F
.2978 .32547 m .30031 .32553 L .29841 .32563 L .663 .495 .646 r
F
.2978 .32547 m .29835 .32563 L .29834 .32563 L F
.29783 .32603 m .2978 .32547 L p .29834 .32563 L F P
0 g
s
.2978 .32547 m .29841 .32563 L .29835 .32563 L .663 .495 .646 r
F
.29834 .32563 m .31046 .325 L .30805 .32744 L p .399 0 0 r
F P
0 g
s
.29834 .32563 m .30659 .32801 L p .30651 .32804 L .404 .112 .351 r
F P
0 g
s
.30659 .32801 m .29834 .32563 L p .30805 .32744 L .399 0 0 r
F P
0 g
s
.32411 .60791 m .29889 .60248 L .30876 .59989 L closepath p .657 .774 .936 r
F P
0 g
s
.32411 .60791 m .31212 .6109 L .29889 .60248 L closepath p .594 .787 .972 r
F P
0 g
s
.30436 .32369 m .30243 .32204 L .30724 .32517 L p .501 0 0 r
F P
0 g
s
.3028 .3254 m .30436 .32369 L p .30425 .32533 L .663 .495 .646 r
F P
0 g
s
.30706 .62609 m .3061 .62631 L p .30354 .62675 L .371 .83 .914 r
F P
0 g
s
.33203 .62143 m .30354 .62675 L .32719 .62206 L closepath p .185 .717 .749 r
F P
0 g
s
.30436 .32369 m .30447 .32531 L p .30425 .32533 L .663 .495 .646 r
F P
0 g
s
.30586 .32524 m .30436 .32369 L .30652 .32521 L .755 .564 .638 r
F
.30436 .32369 m .30699 .32518 L p .30652 .32521 L F P
0 g
s
.30514 .32528 m .30436 .32369 L .30518 .32528 L .755 .564 .638 r
F
.30518 .32528 m .30436 .32369 L .30586 .32524 L F
.30447 .32531 m .30436 .32369 L p .30514 .32528 L F P
0 g
s
.30699 .32518 m .30436 .32369 L p .30724 .32517 L .501 0 0 r
F P
0 g
s
.36074 .47009 m .30512 .52119 L .30726 .47316 L closepath p .636 .59 .785 r
F P
0 g
s
.35497 .52734 m .30512 .52119 L .36074 .47009 L closepath p .637 .591 .786 r
F P
0 g
s
.35497 .52734 m .32303 .55208 L .30512 .52119 L closepath p .627 .612 .815 r
F P
0 g
s
.31785 .62721 m .30646 .62623 L .31057 .62543 L .502 .867 .985 r
F
.30646 .62623 m .30706 .62609 L .31057 .62543 L F
.36074 .47009 m .30726 .47316 L .36455 .44095 L closepath p .637 .553 .742 r
F P
0 g
s
.36455 .44095 m .30726 .47316 L .31027 .44758 L closepath p .64 .559 .747 r
F P
0 g
s
.30805 .32744 m .31046 .325 L p .30823 .32737 L 0 0 0 r
F P
0 g
s
.30823 .32737 m .31853 .31816 L .31999 .32278 L p .24 0 .144 r
F P
0 g
s
.31046 .325 m .31853 .31816 L .30823 .32737 L p 0 0 0 r
F P
0 g
s
.35478 .57663 m .3084 .57036 L .34387 .56195 L closepath p .649 .582 .767 r
F P
0 g
s
.35478 .57663 m .31603 .58056 L .3084 .57036 L closepath p .66 .542 .71 r
F P
0 g
s
.31999 .32278 m .32022 .32352 L p .30867 .3272 L .24 0 .144 r
F P
0 g
s
.33365 .32143 m .30867 .3272 L .34008 .3172 L closepath p .045 0 0 r
F P
0 g
s
.34346 .60234 m .30876 .59989 L .32784 .59261 L closepath p .639 .755 .932 r
F P
0 g
s
.34346 .60234 m .32411 .60791 L .30876 .59989 L closepath p .615 .77 .955 r
F P
0 g
s
.36338 .42608 m .31027 .44758 L .3144 .41687 L closepath p .639 .533 .718 r
F P
0 g
s
.36455 .44095 m .31027 .44758 L .36338 .42608 L closepath p .636 .525 .711 r
F P
0 g
s
.31247 .62508 m .31785 .62721 L p .31057 .62543 L .502 .867 .985 r
F P
0 g
s
.33125 .61431 m .31212 .6109 L .32411 .60791 L closepath p .608 .735 .935 r
F P
0 g
s
.33125 .61431 m .32023 .61677 L .31212 .6109 L closepath p .616 .732 .929 r
F P
0 g
s
.34557 .62102 m .31247 .62508 L .31377 .62484 L .546 .887 .985 r
F
.34557 .62102 m .31785 .62721 L .31247 .62508 L p F P
0 g
s
.31671 .62429 m .34557 .62102 L p .31377 .62484 L .546 .887 .985 r
F P
0 g
s
.36093 .38243 m .31437 .37972 L .3195 .3531 L closepath p .637 .437 .595 r
F P
0 g
s
.36394 .40621 m .31437 .37972 L .36093 .38243 L closepath p .639 .487 .66 r
F P
0 g
s
.36394 .40621 m .3144 .41687 L .31437 .37972 L closepath p .646 .482 .647 r
F P
0 g
s
.36338 .42608 m .3144 .41687 L .36394 .40621 L closepath p .649 .496 .661 r
F P
0 g
s
.36129 .58866 m .31603 .58056 L .35478 .57663 L closepath p .648 .716 .898 r
F P
0 g
s
.36129 .58866 m .32784 .59261 L .31603 .58056 L closepath p .651 .714 .894 r
F P
0 g
s
.34557 .62102 m .31671 .62429 L p .32923 .62195 L .342 .815 .902 r
F P
0 g
s
.35042 .62305 m .31785 .62721 L .34557 .62102 L closepath p .493 .881 .965 r
F P
0 g
s
.31841 .31826 m .32055 .31515 L p .31841 .31826 L 0 0 0 r
F P
0 g
s
.32055 .31515 m .31853 .31816 L p .31841 .31826 L 0 0 0 r
F P
0 g
s
.31853 .31816 m .32541 .31023 L .32137 .32316 L p .029 0 .188 r
F P
0 g
s
.32541 .31023 m .31853 .31816 L .32055 .31515 L closepath p 0 0 0 r
F P
0 g
s
.32022 .32352 m .31853 .31816 L p .32061 .3234 L .029 0 .188 r
F P
0 g
s
.32137 .32316 m .32061 .3234 L .31853 .31816 L .029 0 .188 r
F
.36495 .35629 m .3195 .3531 L .32759 .33317 L closepath p .633 .361 .489 r
F P
0 g
s
.36093 .38243 m .3195 .3531 L .36495 .35629 L closepath p .643 .434 .584 r
F P
0 g
s
.33973 .61876 m .32023 .61677 L .33125 .61431 L closepath p .606 .769 .958 r
F P
0 g
s
.33973 .61876 m .32719 .62206 L .32023 .61677 L closepath p .627 .759 .941 r
F P
0 g
s
.32137 .32316 m .32541 .31023 L p .32194 .32297 L .485 .194 .396 r
F P
0 g
s
.32305 .32262 m .32194 .32297 L .32541 .31023 L .485 .194 .396 r
F
.38279 .55579 m .32303 .55208 L .35497 .52734 L closepath p .615 .588 .8 r
F P
0 g
s
.38279 .55579 m .34387 .56195 L .32303 .55208 L closepath p .629 .52 .711 r
F P
0 g
s
.32541 .31023 m .33697 .30726 L .32305 .32262 L p .485 .194 .396 r
F P
0 g
s
.32305 .32262 m .33697 .30726 L p .32346 .32249 L .284 0 .22 r
F P
0 g
s
.32432 .32222 m .32346 .32249 L .33697 .30726 L .284 0 .22 r
F
.35411 .60721 m .32411 .60791 L .34346 .60234 L closepath p .617 .756 .945 r
F P
0 g
s
.35411 .60721 m .33125 .61431 L .32411 .60791 L closepath p .63 .742 .928 r
F P
0 g
s
.33697 .30726 m .34441 .30565 L .32432 .32222 L p .284 0 .22 r
F P
0 g
s
.32432 .32222 m .34441 .30565 L p .32604 .32167 L .438 .158 .39 r
F P
0 g
s
.34441 .30565 m .33136 .31998 L .32604 .32167 L .438 .158 .39 r
F
.34297 .61897 m .32719 .62206 L .33973 .61876 L closepath p .332 .815 .849 r
F P
0 g
s
.34297 .61897 m .33203 .62143 L .32719 .62206 L closepath p .22 .742 .766 r
F P
0 g
s
.36571 .33404 m .32759 .33317 L .33365 .32143 L closepath p .57 .184 .27 r
F P
0 g
s
.36495 .35629 m .32759 .33317 L .36571 .33404 L closepath p .627 .367 .504 r
F P
0 g
s
.3739 .5958 m .32784 .59261 L .36129 .58866 L closepath p .645 .735 .914 r
F P
0 g
s
.3739 .5958 m .34346 .60234 L .32784 .59261 L closepath p .623 .752 .939 r
F P
0 g
s
.32923 .62195 m .33203 .62143 L .33902 .6205 L .342 .815 .902 r
F
.33902 .6205 m .34557 .62102 L p .32923 .62195 L F P
0 g
s
.36277 .61313 m .33125 .61431 L .35411 .60721 L closepath p .628 .757 .939 r
F P
0 g
s
.36277 .61313 m .33973 .61876 L .33125 .61431 L closepath p .616 .77 .954 r
F P
0 g
s
.34441 .30565 m .34882 .31144 L p .33136 .31998 L .438 .158 .39 r
F P
0 g
s
.34882 .31144 m .33315 .31941 L .33136 .31998 L .438 .158 .39 r
F
.33902 .6205 m .33203 .62143 L .33668 .62038 L .342 .815 .902 r
F
.34882 .31144 m .34008 .3172 L .33315 .31941 L .438 .158 .39 r
F
.37058 .32157 m .33365 .32143 L .34008 .3172 L closepath p .43 0 0 r
F P
0 g
s
.36571 .33404 m .33365 .32143 L .37058 .32157 L closepath p .567 .218 .336 r
F P
0 g
s
.33691 .62033 m .33902 .6205 L p .33668 .62038 L .342 .815 .902 r
F P
0 g
s
.35395 .61799 m .34557 .62102 L .33691 .62033 L p .492 .893 .939 r
F P
0 g
s
.33929 .6198 m .35395 .61799 L p .33691 .62033 L .492 .893 .939 r
F P
0 g
s
.34493 .6191 m .33929 .6198 L p .3406 .6195 L .28 .783 .837 r
F P
0 g
s
.36733 .61258 m .33973 .61876 L .36277 .61313 L closepath p .241 .755 .77 r
F P
0 g
s
.36733 .61258 m .34297 .61897 L .33973 .61876 L closepath p .338 .819 .851 r
F P
0 g
s
.3772 .3168 m .34008 .3172 L .35042 .31354 L closepath p .269 0 0 r
F P
0 g
s
.37058 .32157 m .34008 .3172 L .3772 .3168 L closepath p .439 0 0 r
F P
0 g
s
.34882 .31144 m .35042 .31354 L p .34008 .3172 L .438 .158 .39 r
F P
0 g
s
.34493 .6191 m .3406 .6195 L .34297 .61897 L .28 .783 .837 r
F
.369 .61272 m .34297 .61897 L .36733 .61258 L closepath p .263 .77 .777 r
F P
0 g
s
.34493 .6191 m .34297 .61897 L .34791 .61779 L .28 .783 .837 r
F
.37741 .60181 m .34346 .60234 L .3739 .5958 L closepath p .634 .674 .871 r
F P
0 g
s
.37741 .60181 m .35411 .60721 L .34346 .60234 L closepath p .578 .753 .961 r
F P
0 g
s
.39045 .57264 m .34387 .56195 L .38279 .55579 L closepath p .639 .61 .805 r
F P
0 g
s
.39045 .57264 m .35478 .57663 L .34387 .56195 L closepath p .656 .584 .762 r
F P
0 g
s
.35359 .30282 m .35042 .31354 L .34441 .30565 L closepath p .488 .168 .349 r
F P
0 g
s
.35395 .61799 m .34493 .6191 L p .34791 .61779 L .28 .783 .837 r
F P
0 g
s
.35973 .61855 m .34557 .62102 L .35395 .61799 L closepath p .447 .879 .899 r
F P
0 g
s
.35973 .61855 m .35042 .62305 L .34557 .62102 L closepath p .641 .944 .956 r
F P
0 g
s
.34862 .61762 m .35395 .61799 L p .34791 .61779 L .28 .783 .837 r
F P
0 g
s
.35145 .61694 m .37512 .61191 L p .34862 .61762 L .369 .838 .857 r
F P
0 g
s
.37512 .61191 m .35395 .61799 L .34862 .61762 L p .369 .838 .857 r
F P
0 g
s
.38936 .30459 m .35042 .31354 L .36302 .30352 L closepath p .574 .282 .434 r
F P
0 g
s
.36302 .30352 m .35042 .31354 L .35359 .30282 L closepath p .5 .179 .354 r
F P
0 g
s
.38936 .30459 m .38763 .31132 L .35042 .31354 L closepath p .56 .252 .403 r
F P
0 g
s
.3772 .3168 m .35042 .31354 L .38763 .31132 L closepath p .321 0 0 r
F P
0 g
s
.369 .61272 m .37512 .61191 L .35145 .61694 L p .217 .738 .749 r
F P
0 g
s
.35973 .61855 m .35395 .61799 L p .36585 .61457 L .579 .94 .909 r
F P
0 g
s
.38636 .60903 m .35411 .60721 L .37741 .60181 L closepath p .578 .753 .961 r
F P
0 g
s
.38636 .60903 m .36277 .61313 L .35411 .60721 L closepath p .587 .748 .954 r
F P
0 g
s
.392 .58416 m .35478 .57663 L .39045 .57264 L closepath p .647 .699 .884 r
F P
0 g
s
.392 .58416 m .36129 .58866 L .35478 .57663 L closepath p .633 .708 .901 r
F P
0 g
s
.4049 .49838 m .35497 .52734 L .36074 .47009 L closepath p .612 .568 .781 r
F P
0 g
s
.41076 .53427 m .35497 .52734 L .4049 .49838 L closepath p .617 .579 .79 r
F P
0 g
s
.41076 .53427 m .38279 .55579 L .35497 .52734 L closepath p .613 .588 .801 r
F P
0 g
s
.37196 .61282 m .35973 .61855 L p .36585 .61457 L .579 .94 .909 r
F P
0 g
s
.4149 .45553 m .36074 .47009 L .36455 .44095 L closepath p .648 .563 .745 r
F P
0 g
s
.4049 .49838 m .36074 .47009 L .4149 .45553 L closepath p .648 .56 .742 r
F P
0 g
s
.41117 .38998 m .36093 .38243 L .36495 .35629 L closepath p .678 .466 .591 r
F P
0 g
s
.41038 .40695 m .36093 .38243 L .41117 .38998 L closepath p .677 .494 .631 r
F P
0 g
s
.41038 .40695 m .36394 .40621 L .36093 .38243 L closepath p .666 .506 .659 r
F P
0 g
s
.39584 .59183 m .36129 .58866 L .392 .58416 L closepath p .634 .704 .897 r
F P
0 g
s
.39584 .59183 m .3739 .5958 L .36129 .58866 L closepath p .602 .731 .936 r
F P
0 g
s
.39347 .60845 m .36277 .61313 L .38636 .60903 L closepath p .191 .722 .777 r
F P
0 g
s
.39347 .60845 m .36733 .61258 L .36277 .61313 L closepath p .134 .68 .725 r
F P
0 g
s
.40892 .42174 m .36338 .42608 L .36394 .40621 L closepath p .668 .511 .663 r
F P
0 g
s
.4121 .43659 m .36338 .42608 L .40892 .42174 L closepath p .669 .521 .675 r
F P
0 g
s
.4121 .43659 m .36455 .44095 L .36338 .42608 L closepath p .662 .544 .711 r
F P
0 g
s
.40892 .42174 m .36394 .40621 L .41038 .40695 L closepath p .666 .513 .668 r
F P
0 g
s
.4149 .45553 m .36455 .44095 L .4121 .43659 L closepath p .661 .542 .708 r
F P
0 g
s
.413 .37036 m .36495 .35629 L .36571 .33404 L closepath p .68 .411 .508 r
F P
0 g
s
.41117 .38998 m .36495 .35629 L .413 .37036 L closepath p .67 .47 .605 r
F P
0 g
s
.41475 .35331 m .36571 .33404 L .37058 .32157 L closepath p .681 .335 .383 r
F P
0 g
s
.413 .37036 m .36571 .33404 L .41475 .35331 L closepath p .668 .418 .533 r
F P
0 g
s
.369 .61272 m .36733 .61258 L p .36904 .61231 L .25 .762 .775 r
F P
0 g
s
.37512 .61191 m .369 .61272 L p .37324 .61164 L .267 .772 .766 r
F P
0 g
s
.37324 .61164 m .369 .61272 L p .36904 .61231 L .25 .762 .775 r
F P
0 g
s
.42298 .33446 m .37058 .32157 L .3772 .3168 L closepath p .606 .15 .143 r
F P
0 g
s
.41475 .35331 m .37058 .32157 L .42298 .33446 L closepath p .651 .365 .473 r
F P
0 g
s
.37789 .61091 m .37512 .61191 L .37324 .61164 L .267 .772 .766 r
F
.40418 .59847 m .3739 .5958 L .39584 .59183 L closepath p .612 .682 .892 r
F P
0 g
s
.40418 .59847 m .37741 .60181 L .3739 .5958 L closepath p .622 .669 .874 r
F P
0 g
s
.37915 .61071 m .37512 .61191 L p .37789 .61091 L .267 .772 .766 r
F P
0 g
s
.42914 .32362 m .3772 .3168 L .38763 .31132 L closepath p .528 .031 .019 r
F P
0 g
s
.42298 .33446 m .3772 .3168 L .42914 .32362 L closepath p .617 .247 .319 r
F P
0 g
s
.4101 .60529 m .37741 .60181 L .40418 .59847 L closepath p .599 .742 .944 r
F P
0 g
s
.4101 .60529 m .38636 .60903 L .37741 .60181 L closepath p .576 .753 .961 r
F P
0 g
s
.42367 .55131 m .38279 .55579 L .41076 .53427 L closepath p .632 .626 .826 r
F P
0 g
s
.42367 .55131 m .39045 .57264 L .38279 .55579 L closepath p .633 .608 .806 r
F P
0 g
s
.39347 .60845 m .38636 .60903 L p .4051 .60608 L .208 .734 .785 r
F P
0 g
s
.43678 .32 m .38763 .31132 L .38936 .30459 L closepath p .623 .312 .423 r
F P
0 g
s
.42914 .32362 m .38763 .31132 L .43678 .32 L closepath p .52 .01 0 r
F P
0 g
s
.38936 .30459 m .39865 .30425 L .43705 .31342 L closepath p .304 0 0 r
F P
0 g
s
.43678 .32 m .38936 .30459 L .43705 .31342 L closepath p .621 .239 .299 r
F P
0 g
s
.4309 .57213 m .39045 .57264 L .42367 .55131 L closepath p .64 .622 .817 r
F P
0 g
s
.4309 .57213 m .392 .58416 L .39045 .57264 L closepath p .621 .68 .884 r
F P
0 g
s
.43298 .58373 m .392 .58416 L .4309 .57213 L closepath p .619 .656 .865 r
F P
0 g
s
.43298 .58373 m .39584 .59183 L .392 .58416 L closepath p .604 .688 .902 r
F P
0 g
s
.40937 .60541 m .39347 .60845 L p .4051 .60608 L .208 .734 .785 r
F P
0 g
s
.43736 .5937 m .39584 .59183 L .43298 .58373 L closepath p .605 .647 .865 r
F P
0 g
s
.43736 .5937 m .40418 .59847 L .39584 .59183 L closepath p .587 .677 .903 r
F P
0 g
s
.44768 .31152 m .43705 .31342 L .39865 .30425 L closepath p .362 0 0 r
F P
0 g
s
.44497 .60042 m .40418 .59847 L .43736 .5937 L closepath p .566 .739 .957 r
F P
0 g
s
.44497 .60042 m .4101 .60529 L .40418 .59847 L closepath p .576 .732 .949 r
F P
0 g
s
.4488 .45449 m .4049 .49838 L .4149 .45553 L closepath p .629 .542 .737 r
F P
0 g
s
.45534 .49079 m .4049 .49838 L .4488 .45449 L closepath p .636 .553 .743 r
F P
0 g
s
.45534 .49079 m .41076 .53427 L .4049 .49838 L closepath p .64 .596 .789 r
F P
0 g
s
.44547 .41152 m .40892 .42174 L .41038 .40695 L closepath p .639 .489 .663 r
F P
0 g
s
.44672 .42408 m .40892 .42174 L .44547 .41152 L closepath p .639 .49 .665 r
F P
0 g
s
.44672 .42408 m .4121 .43659 L .40892 .42174 L closepath p .639 .502 .679 r
F P
0 g
s
.44818 .39454 m .41038 .40695 L .41117 .38998 L closepath p .654 .474 .629 r
F P
0 g
s
.44547 .41152 m .41038 .40695 L .44818 .39454 L closepath p .642 .444 .6 r
F P
0 g
s
.4557 .51373 m .41076 .53427 L .45534 .49079 L closepath p .608 .542 .756 r
F P
0 g
s
.4557 .51373 m .42367 .55131 L .41076 .53427 L closepath p .637 .628 .825 r
F P
0 g
s
.45375 .37848 m .41117 .38998 L .413 .37036 L closepath p .69 .487 .608 r
F P
0 g
s
.44818 .39454 m .41117 .38998 L .45375 .37848 L closepath p .65 .387 .507 r
F P
0 g
s
.44593 .43784 m .4121 .43659 L .44672 .42408 L closepath p .639 .501 .679 r
F P
0 g
s
.44593 .43784 m .4149 .45553 L .4121 .43659 L closepath p .638 .525 .709 r
F P
0 g
s
.45339 .3702 m .413 .37036 L .41475 .35331 L closepath p .695 .442 .537 r
F P
0 g
s
.45375 .37848 m .413 .37036 L .45339 .3702 L closepath p .691 .426 .517 r
F P
0 g
s
.45632 .35092 m .41475 .35331 L .42298 .33446 L closepath p .676 .393 .485 r
F P
0 g
s
.45339 .3702 m .41475 .35331 L .45632 .35092 L closepath p .693 .449 .55 r
F P
0 g
s
.4488 .45449 m .4149 .45553 L .44593 .43784 L closepath p .629 .505 .693 r
F P
0 g
s
.46179 .3385 m .42298 .33446 L .42914 .32362 L closepath p .645 .278 .334 r
F P
0 g
s
.45632 .35092 m .42298 .33446 L .46179 .3385 L closepath p .675 .398 .495 r
F P
0 g
s
.4556 .5371 m .42367 .55131 L .4557 .51373 L closepath p .599 .571 .793 r
F P
0 g
s
.4556 .5371 m .4309 .57213 L .42367 .55131 L closepath p .61 .605 .821 r
F P
0 g
s
.46363 .32896 m .42914 .32362 L .43678 .32 L closepath p .524 .015 0 r
F P
0 g
s
.46179 .3385 m .42914 .32362 L .46363 .32896 L closepath p .637 .322 .421 r
F P
0 g
s
.46302 .56545 m .4309 .57213 L .4556 .5371 L closepath p .589 .575 .806 r
F P
0 g
s
.46302 .56545 m .43298 .58373 L .4309 .57213 L closepath p .593 .638 .865 r
F P
0 g
s
.47153 .5795 m .43298 .58373 L .46302 .56545 L closepath p .593 .639 .866 r
F P
0 g
s
.47153 .5795 m .43736 .5937 L .43298 .58373 L closepath p .593 .641 .867 r
F P
0 g
s
.46734 .32504 m .43678 .32 L .43705 .31342 L closepath p .586 .21 .297 r
F P
0 g
s
.46363 .32896 m .43678 .32 L .46734 .32504 L closepath p .547 .093 .127 r
F P
0 g
s
.47293 .31586 m .43705 .31342 L .44768 .31152 L closepath p .253 0 0 r
F P
0 g
s
.46734 .32504 m .43705 .31342 L .47293 .31586 L closepath p .586 .205 .286 r
F P
0 g
s
.483 .59639 m .43736 .5937 L .47153 .5795 L closepath p .604 .692 .905 r
F P
0 g
s
.483 .59639 m .44497 .60042 L .43736 .5937 L closepath p .554 .734 .959 r
F P
0 g
s
.44547 .41152 m .44818 .39454 L .50222 .37554 L closepath p .526 .338 .569 r
F P
0 g
s
.44547 .41152 m .50222 .37554 L .50214 .39509 L closepath p .587 .449 .66 r
F P
0 g
s
.44672 .42408 m .44547 .41152 L .50214 .39509 L closepath p .589 .453 .664 r
F P
0 g
s
.44593 .43784 m .44672 .42408 L .50274 .41349 L closepath p .588 .458 .671 r
F P
0 g
s
.44593 .43784 m .50274 .41349 L .50332 .43199 L closepath p .589 .461 .674 r
F P
0 g
s
.4488 .45449 m .44593 .43784 L .50332 .43199 L closepath p .592 .48 .695 r
F P
0 g
s
.44672 .42408 m .50214 .39509 L .50274 .41349 L closepath p .584 .444 .656 r
F P
0 g
s
.47293 .31586 m .44768 .31152 L .47478 .31542 L p .179 0 0 r
F P
0 g
s
.44818 .39454 m .48609 .37085 L .50222 .37554 L closepath p .346 0 .211 r
F P
0 g
s
.44818 .39454 m .45375 .37848 L .48609 .37085 L closepath p .539 .273 .461 r
F P
0 g
s
.4488 .45449 m .50332 .43199 L .50352 .44993 L closepath p .586 .466 .682 r
F P
0 g
s
.45534 .49079 m .4488 .45449 L .50352 .44993 L closepath p .591 .522 .745 r
F P
0 g
s
.45339 .3702 m .45632 .35092 L .48117 .34441 L closepath p .551 .321 .519 r
F P
0 g
s
.45339 .3702 m .48117 .34441 L .48609 .37085 L closepath p .562 .336 .53 r
F P
0 g
s
.45375 .37848 m .45339 .3702 L .48609 .37085 L closepath p .56 .325 .516 r
F P
0 g
s
.45534 .49079 m .50352 .44993 L .50419 .48301 L closepath p .553 .457 .699 r
F P
0 g
s
.4557 .51373 m .45534 .49079 L .50419 .48301 L closepath p .56 .504 .749 r
F P
0 g
s
.46302 .56545 m .4556 .5371 L .50664 .49926 L closepath p .58 .569 .806 r
F P
0 g
s
.4556 .5371 m .4557 .51373 L .50664 .49926 L closepath p .567 .543 .788 r
F P
0 g
s
.4557 .51373 m .50419 .48301 L .50664 .49926 L closepath p .554 .491 .739 r
F P
0 g
s
.45632 .35092 m .4839 .33429 L .48117 .34441 L closepath p .477 .158 .347 r
F P
0 g
s
.45632 .35092 m .46179 .3385 L .4839 .33429 L closepath p .537 .25 .427 r
F P
0 g
s
.46179 .3385 m .48604 .32517 L .4839 .33429 L closepath p .539 .255 .432 r
F P
0 g
s
.46179 .3385 m .46363 .32896 L .48604 .32517 L closepath p .508 .205 .388 r
F P
0 g
s
.46302 .56545 m .50664 .49926 L .51292 .53919 L closepath p .521 .491 .764 r
F P
0 g
s
.47153 .5795 m .46302 .56545 L .51292 .53919 L closepath p .574 .63 .869 r
F P
0 g
s
.46363 .32896 m .48994 .31983 L .48604 .32517 L closepath p .323 0 0 r
F P
0 g
s
.46363 .32896 m .46734 .32504 L .48994 .31983 L closepath p .319 0 0 r
F P
0 g
s
.46734 .32504 m .47293 .31586 L .49358 .31184 L p .407 .017 .183 r
F P
0 g
s
.46734 .32504 m .49358 .31184 L .49358 .31186 L .407 .017 .183 r
F
.49341 .31249 m .46734 .32504 L p .49358 .31186 L F P
0 g
s
.48994 .31983 m .46734 .32504 L .49341 .31249 L p .505 .163 .321 r
F P
0 g
s
.47153 .5795 m .51292 .53919 L .51847 .55528 L closepath p .535 .563 .83 r
F P
0 g
s
.47153 .5795 m .51847 .55528 L .52557 .57065 L closepath p .552 .61 .865 r
F P
0 g
s
.483 .59639 m .47153 .5795 L .52557 .57065 L closepath p .546 .665 .915 r
F P
0 g
s
.49358 .31184 m .47293 .31586 L .49383 .31089 L p .028 0 0 r
F P
0 g
s
.48928 .32849 m .48117 .34441 L .4839 .33429 L closepath p .071 0 .188 r
F P
0 g
s
.48928 .32849 m .48692 .34139 L .48117 .34441 L closepath p .335 .072 .359 r
F P
0 g
s
.48692 .34139 m .48609 .37085 L .48117 .34441 L closepath p .44 .256 .534 r
F P
0 g
s
.483 .59639 m .52557 .57065 L .52919 .57971 L closepath p .519 .594 .87 r
F P
0 g
s
.483 .59639 m .52919 .57971 L .53027 .5853 L closepath p .515 .578 .857 r
F P
0 g
s
.483 .59639 m .53027 .5853 L .53892 .593 L closepath p .498 .751 .986 r
F P
0 g
s
.53892 .593 m .54835 .59361 L .483 .59639 L closepath p .145 .687 .819 r
F P
0 g
s
.49125 .32085 m .4839 .33429 L .49127 .32079 L .276 .022 .336 r
F
.4839 .33429 m .48604 .32517 L .49127 .32079 L p F P
0 g
s
.49114 .32128 m .4839 .33429 L .49107 .32117 L .276 .022 .336 r
F
.49087 .32232 m .4839 .33429 L .49114 .32128 L F
.49022 .32483 m .4839 .33429 L p .49087 .32232 L F P
0 g
s
.48928 .32849 m .4839 .33429 L .49022 .32483 L p .016 0 .128 r
F P
0 g
s
.49187 .31844 m .48604 .32517 L .49192 .31828 L 0 0 0 r
F
.48604 .32517 m .48994 .31983 L .49192 .31828 L p F P
0 g
s
.49154 .31973 m .48604 .32517 L p .49187 .31844 L 0 0 0 r
F P
0 g
s
.49127 .32079 m .48604 .32517 L .49154 .31973 L p 0 0 0 r
F P
0 g
s
.48692 .34139 m .48925 .33399 L .48609 .37085 L closepath p 0 0 .221 r
F P
0 g
s
.48925 .33399 m .50103 .35468 L .48609 .37085 L closepath p .329 .219 .573 r
F P
0 g
s
.50103 .35468 m .50222 .37554 L .48609 .37085 L closepath p .289 .164 .53 r
F P
0 g
s
.48928 .32849 m .48925 .33399 L .48692 .34139 L closepath p 0 0 0 r
F P
0 g
s
.49149 .32126 m .49542 .33134 L .48925 .33399 L closepath p .303 .125 .466 r
F P
0 g
s
.48928 .32849 m .49149 .32126 L .48925 .33399 L closepath p 0 0 0 r
F P
0 g
s
.48925 .33399 m .49542 .33134 L .50293 .34009 L closepath p .395 .33 .665 r
F P
0 g
s
.48925 .33399 m .50293 .34009 L .50103 .35468 L closepath p .462 .27 .534 r
F P
0 g
s
.49395 .31043 m .49149 .32126 L .48928 .32849 L closepath p 0 0 0 r
F P
0 g
s
.49341 .31249 m .49335 .31274 L .48994 .31983 L .505 .163 .321 r
F
.49251 .316 m .48994 .31983 L p .49335 .31274 L F P
0 g
s
.48994 .31983 m .49251 .316 L p .4925 .31603 L 0 0 0 r
F P
0 g
s
.49246 .31616 m .48994 .31983 L .4925 .31603 L 0 0 0 r
F
.4923 .31678 m .48994 .31983 L p .49246 .31616 L F P
0 g
s
.49192 .31828 m .48994 .31983 L .4923 .31678 L p 0 0 .132 r
F P
0 g
s
.49117 .32115 m .49107 .32117 L .49125 .32085 L .276 .022 .336 r
F
.49114 .32128 m .49107 .32117 L .49116 .32121 L F
.49116 .32121 m .49107 .32117 L .49117 .32115 L F
.49647 .30942 m .49149 .32126 L .49395 .31043 L closepath p 0 0 .205 r
F P
0 g
s
.49647 .30942 m .49632 .32089 L .49149 .32126 L closepath p .324 .137 .467 r
F P
0 g
s
.49149 .32126 m .49632 .32089 L .49542 .33134 L closepath p .323 .134 .463 r
F P
0 g
s
.49877 .29552 m .49647 .30942 L .49395 .31043 L closepath p 0 0 .342 r
F P
0 g
s
.49632 .32089 m .50319 .33176 L .49542 .33134 L closepath p .569 .338 .526 r
F P
0 g
s
.49542 .33134 m .50319 .33176 L .50293 .34009 L closepath p .572 .363 .559 r
F P
0 g
s
.50236 .30839 m .49632 .32089 L .49647 .30942 L closepath p .315 .131 .464 r
F P
0 g
s
.50236 .30839 m .50448 .32162 L .49632 .32089 L closepath p .474 .303 .568 r
F P
0 g
s
.49632 .32089 m .50448 .32162 L .50319 .33176 L closepath p .474 .307 .574 r
F P
0 g
s
.5049 .29068 m .50236 .30839 L .49647 .30942 L closepath p .322 .153 .491 r
F P
0 g
s
.5049 .29068 m .49647 .30942 L .49877 .29552 L closepath p .386 .222 .532 r
F P
0 g
s
.49856 .29617 m .50073 .28346 L .49877 .29552 L p 0 0 0 r
F P
0 g
s
.50294 .28076 m .5049 .29068 L .49877 .29552 L closepath p .303 .095 .423 r
F P
0 g
s
.50073 .28346 m .50294 .28076 L .49877 .29552 L closepath p 0 0 .117 r
F P
0 g
s
.50366 .26967 m .50294 .28076 L .50073 .28346 L closepath p 0 0 .235 r
F P
0 g
s
.53277 .35212 m .50103 .35468 L .50293 .34009 L closepath p .606 .397 .572 r
F P
0 g
s
.53277 .35212 m .50222 .37554 L .50103 .35468 L closepath p .607 .401 .577 r
F P
0 g
s
.53764 .3752 m .53913 .38893 L .50214 .39509 L closepath p .556 .377 .594 r
F P
0 g
s
.53764 .3752 m .50214 .39509 L .50222 .37554 L closepath p .597 .457 .661 r
F P
0 g
s
.53913 .38893 m .50274 .41349 L .50214 .39509 L closepath p .571 .433 .655 r
F P
0 g
s
.53277 .35212 m .53764 .3752 L .50222 .37554 L closepath p .586 .368 .551 r
F P
0 g
s
.51843 .28897 m .50236 .30839 L .5049 .29068 L closepath p .503 .31 .552 r
F P
0 g
s
.51843 .28897 m .51726 .30448 L .50236 .30839 L closepath p .51 .319 .558 r
F P
0 g
s
.51726 .30448 m .50448 .32162 L .50236 .30839 L closepath p .513 .329 .569 r
F P
0 g
s
.53913 .38893 m .5389 .39982 L .50274 .41349 L closepath p .517 .339 .579 r
F P
0 g
s
.5389 .39982 m .50332 .43199 L .50274 .41349 L closepath p .557 .436 .67 r
F P
0 g
s
.53277 .35212 m .50293 .34009 L .50319 .33176 L closepath p .608 .393 .563 r
F P
0 g
s
.50366 .26967 m .5065 .26586 L .50294 .28076 L closepath p 0 0 .277 r
F P
0 g
s
.5065 .26586 m .50973 .27804 L .50294 .28076 L closepath p .384 .203 .509 r
F P
0 g
s
.50294 .28076 m .50973 .27804 L .5049 .29068 L closepath p .344 .12 .424 r
F P
0 g
s
.52053 .33183 m .50319 .33176 L .50448 .32162 L closepath p .538 .363 .592 r
F P
0 g
s
.53277 .35212 m .50319 .33176 L .52053 .33183 L closepath p .542 .417 .66 r
F P
0 g
s
.5389 .39982 m .53574 .43061 L .50332 .43199 L closepath p .537 .404 .647 r
F P
0 g
s
.53574 .43061 m .50352 .44993 L .50332 .43199 L closepath p .539 .428 .676 r
F P
0 g
s
.53574 .43061 m .53599 .45521 L .50352 .44993 L closepath p .577 .505 .737 r
F P
0 g
s
.53599 .45521 m .50419 .48301 L .50352 .44993 L closepath p .585 .482 .704 r
F P
0 g
s
.50714 .24731 m .5065 .26586 L .50366 .26967 L closepath p 0 0 .345 r
F P
0 g
s
.53599 .45521 m .53758 .4844 L .50419 .48301 L closepath p .609 .523 .732 r
F P
0 g
s
.53758 .4844 m .50664 .49926 L .50419 .48301 L closepath p .608 .53 .741 r
F P
0 g
s
.51726 .30448 m .51845 .31832 L .50448 .32162 L closepath p .502 .315 .559 r
F P
0 g
s
.52053 .33183 m .50448 .32162 L .51845 .31832 L closepath p .521 .371 .618 r
F P
0 g
s
.51843 .28897 m .5049 .29068 L .50973 .27804 L closepath p .49 .267 .502 r
F P
0 g
s
.50714 .24731 m .51139 .24432 L .5065 .26586 L closepath p .065 0 .372 r
F P
0 g
s
.51139 .24432 m .51691 .26483 L .5065 .26586 L closepath p .523 .349 .586 r
F P
0 g
s
.5065 .26586 m .51691 .26483 L .50973 .27804 L closepath p .504 .274 .499 r
F P
0 g
s
.53758 .4844 m .51292 .53919 L .50664 .49926 L closepath p .621 .563 .768 r
F P
0 g
s
.51175 .23179 m .51139 .24432 L .50714 .24731 L closepath p .051 0 .346 r
F P
0 g
s
.52403 .26723 m .50973 .27804 L .51691 .26483 L closepath p .53 .304 .516 r
F P
0 g
s
.52403 .26723 m .5203 .28026 L .50973 .27804 L closepath p .502 .26 .48 r
F P
0 g
s
.51843 .28897 m .50973 .27804 L .5203 .28026 L closepath p .501 .268 .494 r
F P
0 g
s
.51175 .23179 m .51458 .22907 L .51139 .24432 L closepath p .034 0 .34 r
F P
0 g
s
.51458 .22907 m .51841 .24124 L .51139 .24432 L closepath p .51 .317 .555 r
F P
0 g
s
.51139 .24432 m .51841 .24124 L .51691 .26483 L closepath p .526 .35 .586 r
F P
0 g
s
.51458 .22907 m .51175 .23179 L .51761 .21437 L p 0 0 .258 r
F P
0 g
s
.53758 .4844 m .54281 .51414 L .51292 .53919 L closepath p .584 .52 .749 r
F P
0 g
s
.54281 .51414 m .51847 .55528 L .51292 .53919 L closepath p .644 .629 .82 r
F P
0 g
s
.51838 .2106 m .51832 .22815 L .51458 .22907 L closepath p .378 .2 .509 r
F P
0 g
s
.51458 .22907 m .51832 .22815 L .51841 .24124 L closepath p .396 .252 .566 r
F P
0 g
s
.52654 .23849 m .52403 .26723 L .51691 .26483 L closepath p .511 .346 .593 r
F P
0 g
s
.52654 .23849 m .51691 .26483 L .51841 .24124 L closepath p .473 .306 .574 r
F P
0 g
s
.53441 .28781 m .53043 .30592 L .51726 .30448 L closepath p .594 .354 .522 r
F P
0 g
s
.53441 .28781 m .51726 .30448 L .51843 .28897 L closepath p .658 .445 .585 r
F P
0 g
s
.53043 .30592 m .51845 .31832 L .51726 .30448 L closepath p .596 .384 .564 r
F P
0 g
s
.52674 .15689 m .52486 .15994 L .51796 .15848 L closepath p .642 .352 .463 r
F P
0 g
s
.51796 .15848 m .52486 .15994 L .51927 .16609 L closepath p .634 .471 .645 r
F P
0 g
s
.5259 .20985 m .51832 .22815 L .52182 .21131 L closepath p .412 .229 .52 r
F P
0 g
s
.5259 .20985 m .52606 .22277 L .51832 .22815 L closepath p .483 .303 .56 r
F P
0 g
s
.51838 .2106 m .52182 .21131 L .51832 .22815 L closepath p .392 .211 .512 r
F P
0 g
s
.52606 .22277 m .51841 .24124 L .51832 .22815 L closepath p .502 .336 .587 r
F P
0 g
s
.52012 .18783 m .52182 .21131 L .51838 .2106 L closepath p .377 .27 .604 r
F P
0 g
s
.52606 .22277 m .52654 .23849 L .51841 .24124 L closepath p .471 .303 .57 r
F P
0 g
s
.53441 .28781 m .51843 .28897 L .5203 .28026 L closepath p .648 .406 .539 r
F P
0 g
s
.53043 .30592 m .53428 .31757 L .51845 .31832 L closepath p .611 .406 .578 r
F P
0 g
s
.53735 .32944 m .51845 .31832 L .53428 .31757 L closepath p .62 .457 .64 r
F P
0 g
s
.53735 .32944 m .52053 .33183 L .51845 .31832 L closepath p .634 .449 .615 r
F P
0 g
s
.54281 .51414 m .54459 .54769 L .51847 .55528 L closepath p .579 .545 .781 r
F P
0 g
s
.54459 .54769 m .52557 .57065 L .51847 .55528 L closepath p .595 .634 .86 r
F P
0 g
s
.51927 .16609 m .52431 .17061 L .51869 .17032 L closepath p .58 .366 .554 r
F P
0 g
s
.52431 .17061 m .52012 .18783 L .51869 .17032 L closepath p .582 .488 .713 r
F P
0 g
s
.52486 .15994 m .52431 .17061 L .51927 .16609 L closepath p .56 .368 .577 r
F P
0 g
s
.52431 .17061 m .52322 .18689 L .52012 .18783 L closepath p .392 .301 .632 r
F P
0 g
s
.52012 .18783 m .52322 .18689 L .52182 .21131 L closepath p .383 .274 .605 r
F P
0 g
s
.53468 .26597 m .5203 .28026 L .52403 .26723 L closepath p .608 .363 .521 r
F P
0 g
s
.53468 .26597 m .53316 .2764 L .5203 .28026 L closepath p .67 .453 .582 r
F P
0 g
s
.53441 .28781 m .5203 .28026 L .53316 .2764 L closepath p .649 .404 .533 r
F P
0 g
s
.53277 .35212 m .52053 .33183 L .53735 .32944 L closepath p .634 .45 .616 r
F P
0 g
s
.52953 .18986 m .52182 .21131 L .52322 .18689 L closepath p .72 .561 .673 r
F P
0 g
s
.52953 .18986 m .5259 .20985 L .52182 .21131 L closepath p .418 .243 .535 r
F P
0 g
s
.52703 .17143 m .52322 .18689 L .52431 .17061 L closepath p .44 .342 .648 r
F P
0 g
s
.52703 .17143 m .52953 .18986 L .52322 .18689 L closepath p .689 .589 .739 r
F P
0 g
s
.5326 .23937 m .52403 .26723 L .52654 .23849 L closepath p .563 .391 .605 r
F P
0 g
s
.5326 .23937 m .53468 .26597 L .52403 .26723 L closepath p .634 .463 .634 r
F P
0 g
s
.52431 .17061 m .52794 .15896 L .53096 .15928 L closepath p .506 .35 .603 r
F P
0 g
s
.52486 .15994 m .52794 .15896 L .52431 .17061 L closepath p .352 .197 .527 r
F P
0 g
s
.52431 .17061 m .53096 .15928 L .52703 .17143 L closepath p .463 .299 .573 r
F P
0 g
s
.52674 .15689 m .52794 .15896 L .52486 .15994 L closepath p .212 0 .185 r
F P
0 g
s
.54459 .54769 m .55109 .56713 L .52557 .57065 L closepath p .595 .634 .86 r
F P
0 g
s
.55109 .56713 m .52919 .57971 L .52557 .57065 L closepath p .594 .64 .866 r
F P
0 g
s
.52584 .13492 m .52784 .14794 L .52571 .14516 L closepath p .106 .096 .562 r
F P
0 g
s
.52571 .14516 m .52794 .15896 L .52674 .15689 L closepath p 0 0 .373 r
F P
0 g
s
.52571 .14516 m .52784 .14794 L .52794 .15896 L closepath p .056 .084 .573 r
F P
0 g
s
.52967 .12236 m .52584 .13492 L .52771 .12167 L closepath p 0 0 .38 r
F P
0 g
s
.52967 .12236 m .52899 .1381 L .52584 .13492 L closepath p .163 .16 .607 r
F P
0 g
s
.52584 .13492 m .52899 .1381 L .52784 .14794 L closepath p .203 .165 .59 r
F P
0 g
s
.52953 .18986 m .53089 .2102 L .5259 .20985 L closepath p .588 .398 .592 r
F P
0 g
s
.53089 .2102 m .52606 .22277 L .5259 .20985 L closepath p .588 .385 .574 r
F P
0 g
s
.53089 .2102 m .53305 .22417 L .52606 .22277 L closepath p .593 .391 .577 r
F P
0 g
s
.53305 .22417 m .52654 .23849 L .52606 .22277 L closepath p .593 .396 .585 r
F P
0 g
s
.53305 .22417 m .5326 .23937 L .52654 .23849 L closepath p .565 .366 .57 r
F P
0 g
s
.53384 .17434 m .52703 .17143 L .53096 .15928 L closepath p .721 .567 .68 r
F P
0 g
s
.53394 .18971 m .52703 .17143 L .53384 .17434 L closepath p .702 .589 .727 r
F P
0 g
s
.53394 .18971 m .52953 .18986 L .52703 .17143 L closepath p .652 .563 .742 r
F P
0 g
s
.52771 .12167 m .53207 .11157 L .52967 .12236 L closepath p 0 0 .323 r
F P
0 g
s
.53359 .15018 m .52784 .14794 L .52899 .1381 L closepath p .455 .392 .698 r
F P
0 g
s
.52784 .14794 m .53359 .15018 L .53096 .15928 L closepath p .433 .409 .733 r
F P
0 g
s
.52784 .14794 m .53096 .15928 L .52794 .15896 L closepath p .491 .447 .735 r
F P
0 g
s
.53642 .13857 m .52899 .1381 L .52967 .12236 L closepath p .506 .456 .736 r
F P
0 g
s
.53359 .15018 m .52899 .1381 L .53642 .13857 L closepath p .511 .424 .693 r
F P
0 g
s
.55109 .56713 m .5552 .57847 L .52919 .57971 L closepath p .597 .648 .871 r
F P
0 g
s
.5552 .57847 m .53027 .5853 L .52919 .57971 L closepath p .6 .639 .862 r
F P
0 g
s
.52953 .18986 m .53394 .18971 L .53089 .2102 L closepath p .641 .438 .593 r
F P
0 g
s
.54051 .12239 m .52967 .12236 L .53207 .11157 L closepath p .494 .378 .65 r
F P
0 g
s
.53642 .13857 m .52967 .12236 L .54051 .12239 L closepath p .492 .449 .737 r
F P
0 g
s
.5552 .57847 m .55924 .58357 L .53027 .5853 L closepath p .602 .683 .899 r
F P
0 g
s
.55924 .58357 m .53892 .593 L .53027 .5853 L closepath p .536 .769 .983 r
F P
0 g
s
.54463 .29301 m .53043 .30592 L .53441 .28781 L closepath p .744 .499 .56 r
F P
0 g
s
.54463 .29301 m .54489 .30392 L .53043 .30592 L closepath p .703 .437 .518 r
F P
0 g
s
.54489 .30392 m .53428 .31757 L .53043 .30592 L closepath p .711 .46 .543 r
F P
0 g
s
.53394 .18971 m .53771 .21086 L .53089 .2102 L closepath p .715 .507 .606 r
F P
0 g
s
.53089 .2102 m .53771 .21086 L .54212 .22997 L closepath p .715 .507 .606 r
F P
0 g
s
.53089 .2102 m .54212 .22997 L .53305 .22417 L closepath p .789 .527 .54 r
F P
0 g
s
.53359 .15018 m .53852 .16085 L .53096 .15928 L closepath p .622 .606 .813 r
F P
0 g
s
.53384 .17434 m .53096 .15928 L .53852 .16085 L closepath p .662 .528 .69 r
F P
0 g
s
.54051 .12239 m .53207 .11157 L .5486 .10748 L closepath p .529 .387 .632 r
F P
0 g
s
.54212 .22997 m .5326 .23937 L .53305 .22417 L closepath p .778 .544 .581 r
F P
0 g
s
.54212 .22997 m .54363 .24304 L .5326 .23937 L closepath p .771 .533 .575 r
F P
0 g
s
.54363 .24304 m .53468 .26597 L .5326 .23937 L closepath p .765 .562 .623 r
F P
0 g
s
.53277 .35212 m .54733 .32725 L .55642 .32875 L closepath p .813 .665 .699 r
F P
0 g
s
.53277 .35212 m .53735 .32944 L .54733 .32725 L closepath p .694 .508 .632 r
F P
0 g
s
.53277 .35212 m .55642 .32875 L .55909 .33674 L closepath p .748 .554 .633 r
F P
0 g
s
.53277 .35212 m .55909 .33674 L .56643 .34503 L closepath p .653 .378 .49 r
F P
0 g
s
.53764 .3752 m .53277 .35212 L .56643 .34503 L closepath p .671 .422 .537 r
F P
0 g
s
.54439 .26742 m .53316 .2764 L .53468 .26597 L closepath p .776 .553 .597 r
F P
0 g
s
.54439 .26742 m .54336 .28042 L .53316 .2764 L closepath p .732 .479 .545 r
F P
0 g
s
.54463 .29301 m .53441 .28781 L .53316 .2764 L closepath p .752 .479 .518 r
F P
0 g
s
.54463 .29301 m .53316 .2764 L .54336 .28042 L closepath p .733 .476 .54 r
F P
0 g
s
.53642 .13857 m .54428 .14738 L .53359 .15018 L closepath p .594 .507 .725 r
F P
0 g
s
.53359 .15018 m .54428 .14738 L .53852 .16085 L closepath p .614 .602 .815 r
F P
0 g
s
.53384 .17434 m .53852 .16085 L .5478 .16057 L closepath p .623 .486 .674 r
F P
0 g
s
.53384 .17434 m .5478 .16057 L .5425 .17395 L closepath p .588 .433 .638 r
F P
0 g
s
.5425 .17395 m .53394 .18971 L .53384 .17434 L closepath p .594 .502 .719 r
F P
0 g
s
.5425 .17395 m .53967 .18987 L .53394 .18971 L closepath p .528 .422 .678 r
F P
0 g
s
.53394 .18971 m .53967 .18987 L .53771 .21086 L closepath p .528 .38 .625 r
F P
0 g
s
.54489 .30392 m .54844 .31618 L .53428 .31757 L closepath p .737 .495 .563 r
F P
0 g
s
.54733 .32725 m .53428 .31757 L .54844 .31618 L closepath p .732 .478 .544 r
F P
0 g
s
.54733 .32725 m .53735 .32944 L .53428 .31757 L closepath p .692 .501 .625 r
F P
0 g
s
.54363 .24304 m .54439 .26742 L .53468 .26597 L closepath p .776 .574 .627 r
F P
0 g
s
.53574 .43061 m .5389 .39982 L .57288 .39993 L closepath p .668 .521 .675 r
F P
0 g
s
.53574 .43061 m .57288 .39993 L .57424 .42556 L closepath p .663 .512 .669 r
F P
0 g
s
.53599 .45521 m .53574 .43061 L .57424 .42556 L closepath p .674 .584 .746 r
F P
0 g
s
.53599 .45521 m .57424 .42556 L .57458 .45249 L closepath p .653 .545 .719 r
F P
0 g
s
.53758 .4844 m .53599 .45521 L .57458 .45249 L closepath p .653 .558 .734 r
F P
0 g
s
.54051 .12239 m .54697 .14059 L .53642 .13857 L closepath p .62 .578 .785 r
F P
0 g
s
.53642 .13857 m .54697 .14059 L .54428 .14738 L closepath p .647 .515 .688 r
F P
0 g
s
.53758 .4844 m .57458 .45249 L .57576 .47768 L closepath p .664 .577 .747 r
F P
0 g
s
.54281 .51414 m .53758 .4844 L .57576 .47768 L closepath p .664 .575 .745 r
F P
0 g
s
.53764 .3752 m .56643 .34503 L .56839 .35681 L closepath p .687 .446 .552 r
F P
0 g
s
.53913 .38893 m .53764 .3752 L .56839 .35681 L closepath p .71 .489 .587 r
F P
0 g
s
.53967 .18987 m .54209 .21237 L .53771 .21086 L closepath p .867 .684 .652 r
F P
0 g
s
.54847 .23373 m .54212 .22997 L .53771 .21086 L closepath p .864 .597 .533 r
F P
0 g
s
.54847 .23373 m .53771 .21086 L .54209 .21237 L closepath p .876 .599 .516 r
F P
0 g
s
.54428 .14738 m .55256 .15015 L .53852 .16085 L closepath p .724 .731 .856 r
F P
0 g
s
.55256 .15015 m .5478 .16057 L .53852 .16085 L closepath p .625 .528 .725 r
F P
0 g
s
.5389 .39982 m .57017 .38332 L .57288 .39993 L closepath p .659 .438 .573 r
F P
0 g
s
.5389 .39982 m .53913 .38893 L .57017 .38332 L closepath p .675 .468 .598 r
F P
0 g
s
.55924 .58357 m .5654 .58974 L .53892 .593 L closepath p .536 .771 .984 r
F P
0 g
s
.54835 .59361 m .53892 .593 L .5654 .58974 L closepath p .237 .75 .863 r
F P
0 g
s
.53913 .38893 m .56839 .35681 L .56934 .3707 L closepath p .576 .304 .467 r
F P
0 g
s
.53913 .38893 m .56934 .3707 L .57017 .38332 L closepath p .679 .484 .615 r
F P
0 g
s
.54936 .17579 m .53967 .18987 L .5425 .17395 L closepath p .855 .738 .732 r
F P
0 g
s
.54936 .17579 m .54533 .19147 L .53967 .18987 L closepath p .839 .71 .722 r
F P
0 g
s
.53967 .18987 m .54533 .19147 L .54209 .21237 L closepath p .851 .673 .661 r
F P
0 g
s
.5486 .10748 m .55255 .12768 L .54051 .12239 L closepath p .705 .6 .737 r
F P
0 g
s
.54051 .12239 m .55255 .12768 L .54697 .14059 L closepath p .685 .615 .771 r
F P
0 g
s
.54533 .19147 m .54552 .21668 L .54209 .21237 L closepath p .968 .798 .618 r
F P
0 g
s
.54552 .21668 m .54847 .23373 L .54209 .21237 L closepath p .953 .61 .226 r
F P
0 g
s
.55021 .24525 m .54212 .22997 L .54847 .23373 L closepath p .861 .609 .557 r
F P
0 g
s
.55021 .24525 m .54363 .24304 L .54212 .22997 L closepath p .877 .611 .532 r
F P
0 g
s
.5478 .16057 m .54936 .17579 L .5425 .17395 L closepath p .854 .74 .737 r
F P
0 g
s
.54281 .51414 m .57576 .47768 L .57817 .50631 L closepath p .676 .595 .757 r
F P
0 g
s
.54459 .54769 m .54281 .51414 L .57817 .50631 L closepath p .681 .625 .786 r
F P
0 g
s
.54963 .28436 m .54336 .28042 L .54439 .26742 L closepath p .918 .655 .524 r
F P
0 g
s
.55225 .29803 m .54336 .28042 L .54963 .28436 L closepath p .918 .595 .415 r
F P
0 g
s
.55225 .29803 m .54463 .29301 L .54336 .28042 L closepath p .892 .598 .481 r
F P
0 g
s
.55039 .27066 m .54439 .26742 L .54363 .24304 L closepath p .875 .654 .6 r
F P
0 g
s
.55039 .27066 m .54363 .24304 L .55021 .24525 L closepath p .874 .654 .6 r
F P
0 g
s
.55256 .15015 m .54428 .14738 L .54697 .14059 L closepath p .786 .674 .741 r
F P
0 g
s
.54963 .28436 m .54439 .26742 L .55039 .27066 L closepath p .88 .639 .57 r
F P
0 g
s
.54459 .54769 m .57817 .50631 L .57568 .52722 L closepath p .632 .552 .747 r
F P
0 g
s
.55109 .56713 m .54459 .54769 L .57568 .52722 L closepath p .694 .693 .845 r
F P
0 g
s
.55225 .29803 m .54489 .30392 L .54463 .29301 L closepath p .893 .596 .478 r
F P
0 g
s
.55225 .29803 m .55285 .30838 L .54489 .30392 L closepath p .888 .586 .471 r
F P
0 g
s
.55285 .30838 m .54844 .31618 L .54489 .30392 L closepath p .888 .577 .455 r
F P
0 g
s
.54831 .18213 m .54533 .19147 L .54936 .17579 L closepath p .838 .822 .342 r
F P
0 g
s
.54831 .18213 m .54575 .19194 L .54533 .19147 L closepath p .88 .842 .38 r
F P
0 g
s
.54533 .19147 m .54575 .19194 L .54552 .21668 L closepath p .871 .71 .188 r
F P
0 g
s
.55428 .14167 m .54697 .14059 L .55255 .12768 L closepath p .785 .734 .806 r
F P
0 g
s
.55256 .15015 m .54697 .14059 L .55428 .14167 L closepath p .806 .68 .725 r
F P
0 g
s
.55642 .32875 m .54733 .32725 L .54844 .31618 L closepath p .816 .556 .546 r
F P
0 g
s
.55981 .15023 m .5478 .16057 L .55256 .15015 L closepath p .828 .776 .803 r
F P
0 g
s
.55981 .15023 m .55334 .1646 L .5478 .16057 L closepath p .819 .754 .791 r
F P
0 g
s
.5478 .16057 m .55334 .1646 L .54936 .17579 L closepath p .854 .74 .737 r
F P
0 g
s
.553 .16643 m .54831 .18213 L .54936 .17579 L closepath p .882 .875 .436 r
F P
0 g
s
.54835 .59361 m .5654 .58974 L .57353 .59025 L closepath p .24 .751 .868 r
F P
0 g
s
.55285 .30838 m .55442 .31601 L .54844 .31618 L closepath p .844 .514 .429 r
F P
0 g
s
.55642 .32875 m .54844 .31618 L .55442 .31601 L closepath p .858 .556 .476 r
F P
0 g
s
.5486 .10748 m .55716 .11073 L .55255 .12768 L closepath p .645 .56 .745 r
F P
0 g
s
.57173 .11425 m .55716 .11073 L .5486 .10748 L closepath p .535 0 0 r
F P
0 g
s
.553 .16643 m .54936 .17579 L .55334 .1646 L closepath p .759 .81 .333 r
F P
0 g
s
.55253 .27514 m .54963 .28436 L .55039 .27066 L closepath p .934 .753 .277 r
F P
0 g
s
.54963 .28436 m .55253 .27514 L .55207 .28741 L p .909 .762 .272 r
F P
0 g
s
.55943 .2966 m .55225 .29803 L .54963 .28436 L closepath p .957 .622 .325 r
F P
0 g
s
.55034 .26465 m .55253 .27514 L .55039 .27066 L p .831 .551 .004 r
F P
0 g
s
.55109 .56713 m .57568 .52722 L .57565 .54802 L closepath p .595 .561 .786 r
F P
0 g
s
.5552 .57847 m .55109 .56713 L .57565 .54802 L closepath p .656 .683 .864 r
F P
0 g
s
.55943 .2966 m .56051 .30619 L .55225 .29803 L closepath p .947 .594 .298 r
F P
0 g
s
.56051 .30619 m .55285 .30838 L .55225 .29803 L closepath p .958 .652 .407 r
F P
0 g
s
.55716 .11073 m .5585 .12998 L .55255 .12768 L closepath p .819 .743 .78 r
F P
0 g
s
.55428 .14167 m .55255 .12768 L .5585 .12998 L closepath p .81 .751 .798 r
F P
0 g
s
.55428 .14167 m .55981 .15023 L .55256 .15015 L closepath p .84 .715 .725 r
F P
0 g
s
.56051 .30619 m .56055 .31422 L .55285 .30838 L closepath p .943 .611 .369 r
F P
0 g
s
.56055 .31422 m .55442 .31601 L .55285 .30838 L closepath p .936 .579 .315 r
F P
0 g
s
.56102 .15278 m .553 .16643 L .55334 .1646 L closepath p .791 .931 .564 r
F P
0 g
s
.56102 .15278 m .55334 .1646 L .55981 .15023 L closepath p .942 .962 .726 r
F P
0 g
s
.55409 .28993 m .56088 .29057 L .55943 .2966 L p .664 .264 .278 r
F P
0 g
s
.56112 .14466 m .55428 .14167 L .5585 .12998 L closepath p .819 .762 .799 r
F P
0 g
s
.55428 .14167 m .56112 .14466 L .55981 .15023 L closepath p .861 .717 .702 r
F P
0 g
s
.56055 .31422 m .5608 .31958 L .55442 .31601 L closepath p .95 .615 .348 r
F P
0 g
s
.56312 .32945 m .55642 .32875 L .55442 .31601 L closepath p .933 .61 .406 r
F P
0 g
s
.56312 .32945 m .55442 .31601 L .5608 .31958 L closepath p .936 .571 .285 r
F P
0 g
s
.5552 .57847 m .57565 .54802 L .57608 .57095 L closepath p .566 .559 .805 r
F P
0 g
s
.55924 .58357 m .5552 .57847 L .57608 .57095 L closepath p .59 .678 .902 r
F P
0 g
s
.55909 .33674 m .55642 .32875 L .56312 .32945 L closepath p .94 .636 .437 r
F P
0 g
s
.55716 .11073 m .56293 .11549 L .5585 .12998 L closepath p .665 .628 .803 r
F P
0 g
s
.57173 .11425 m .56293 .11549 L .55716 .11073 L closepath p .836 .574 .543 r
F P
0 g
s
.57173 .11425 m .5585 .12998 L .56293 .11549 L closepath p .855 .836 .826 r
F P
0 g
s
.57173 .11425 m .56545 .1306 L .5585 .12998 L closepath p .818 .76 .799 r
F P
0 g
s
.56112 .14466 m .5585 .12998 L .56545 .1306 L closepath p .817 .761 .8 r
F P
0 g
s
.55909 .33674 m .56312 .32945 L .56739 .33829 L closepath p .832 .471 .373 r
F P
0 g
s
.57847 .34566 m .55909 .33674 L .56739 .33829 L closepath p .747 .276 .127 r
F P
0 g
s
.57847 .34566 m .56643 .34503 L .55909 .33674 L closepath p .772 .345 .238 r
F P
0 g
s
.55924 .58357 m .57608 .57095 L .57855 .57771 L closepath p .605 .711 .92 r
F P
0 g
s
.5654 .58974 m .55924 .58357 L .57855 .57771 L closepath p .593 .797 .977 r
F P
0 g
s
.55943 .2966 m .56088 .29057 L .58104 .29083 L closepath p .634 .235 .271 r
F P
0 g
s
.57764 .29581 m .55943 .2966 L .57799 .29164 L p .622 .214 .25 r
F P
0 g
s
.56051 .30619 m .55943 .2966 L .57764 .29581 L closepath p .692 .383 .447 r
F P
0 g
s
.56102 .15278 m .55981 .15023 L .56112 .14466 L closepath p .976 .892 .533 r
F P
0 g
s
.56051 .30619 m .57764 .29581 L .5782 .29946 L closepath p .635 .293 .374 r
F P
0 g
s
.5782 .29946 m .57754 .30775 L .56051 .30619 L closepath p .605 .241 .325 r
F P
0 g
s
.56051 .30619 m .57754 .30775 L .57797 .3141 L closepath p .618 .287 .387 r
F P
0 g
s
.56051 .30619 m .57797 .3141 L .56055 .31422 L closepath p .611 .312 .439 r
F P
0 g
s
.56055 .31422 m .57797 .3141 L .57952 .31951 L closepath p .623 .357 .495 r
F P
0 g
s
.56055 .31422 m .57952 .31951 L .5608 .31958 L closepath p .625 .333 .454 r
F P
0 g
s
.56312 .32945 m .5608 .31958 L .57952 .31951 L closepath p .637 .38 .513 r
F P
0 g
s
.56102 .15278 m .56112 .14466 L .56378 .14475 L closepath p .949 .905 .744 r
F P
0 g
s
.56378 .14475 m .56112 .14466 L .56545 .1306 L closepath p .944 .913 .756 r
F P
0 g
s
.56312 .32945 m .57952 .31951 L .57953 .32998 L closepath p .651 .404 .532 r
F P
0 g
s
.57953 .32998 m .56739 .33829 L .56312 .32945 L closepath p .651 .407 .537 r
F P
0 g
s
.56378 .14475 m .56545 .1306 L .57029 .13597 L closepath p .902 .879 .802 r
F P
0 g
s
.5654 .58974 m .57855 .57771 L .58496 .5863 L closepath p .579 .767 .968 r
F P
0 g
s
.58496 .5863 m .57353 .59025 L .5654 .58974 L closepath p .321 .801 .901 r
F P
0 g
s
.57554 .12373 m .56545 .1306 L .57173 .11425 L closepath p .936 .918 .772 r
F P
0 g
s
.57554 .12373 m .57029 .13597 L .56545 .1306 L closepath p .939 .868 .742 r
F P
0 g
s
.57847 .34566 m .56839 .35681 L .56643 .34503 L closepath p .854 .561 .491 r
F P
0 g
s
.57953 .32998 m .57847 .34566 L .56739 .33829 L closepath p .691 .475 .589 r
F P
0 g
s
.57847 .34566 m .58539 .36104 L .56839 .35681 L closepath p .72 .364 .374 r
F P
0 g
s
.58539 .36104 m .56934 .3707 L .56839 .35681 L closepath p .729 .42 .457 r
F P
0 g
s
.58539 .36104 m .58781 .36995 L .56934 .3707 L closepath p .726 .416 .453 r
F P
0 g
s
.58781 .36995 m .57017 .38332 L .56934 .3707 L closepath p .762 .547 .607 r
F P
0 g
s
.58781 .36995 m .58932 .38833 L .57017 .38332 L closepath p .777 .574 .626 r
F P
0 g
s
.58932 .38833 m .57288 .39993 L .57017 .38332 L closepath p .781 .521 .544 r
F P
0 g
s
.58932 .38833 m .59086 .4072 L .57288 .39993 L closepath p .781 .522 .544 r
F P
0 g
s
.59086 .4072 m .57424 .42556 L .57288 .39993 L closepath p .759 .586 .662 r
F P
0 g
s
.59237 .58643 m .57353 .59025 L p .59179 .58639 L .303 .796 .858 r
F P
0 g
s
.57353 .59025 m .57861 .5885 L p .59179 .58639 L .303 .796 .858 r
F P
0 g
s
.59086 .4072 m .59301 .42227 L .57424 .42556 L closepath p .726 .535 .632 r
F P
0 g
s
.59301 .42227 m .57458 .45249 L .57424 .42556 L closepath p .743 .618 .719 r
F P
0 g
s
.57432 .29074 m .57436 .29053 L p .57434 .29074 L .129 .016 .445 r
F P
0 g
s
.57436 .29053 m .57436 .29074 L p .57434 .29074 L .129 .016 .445 r
F P
0 g
s
.57813 .28716 m .57436 .29053 L .57856 .27504 L closepath p 0 0 .291 r
F P
0 g
s
.57436 .29053 m .57813 .28716 L .5777 .29078 L p 0 0 .399 r
F P
0 g
s
.57447 .29074 m .57436 .29053 L .5777 .29078 L 0 0 .399 r
F
.57438 .29074 m .57436 .29053 L .57441 .29074 L F
.57441 .29074 m .57436 .29053 L .57443 .29074 L F
.57446 .29074 m .57436 .29053 L .57446 .29074 L F
.57446 .29074 m .57436 .29053 L .57447 .29074 L F
.57443 .29074 m .57436 .29053 L .57446 .29074 L F
.57436 .29074 m .57436 .29053 L p .57437 .29074 L F P
0 g
s
.57437 .29074 m .57436 .29053 L .57438 .29074 L 0 0 .399 r
F
.57438 .29074 m .57436 .29053 L .57438 .29074 L F
.57438 .29074 m .57436 .29053 L .57438 .29074 L F
.57438 .29074 m .57436 .29053 L .57438 .29074 L F
.57438 .29074 m .57436 .29053 L .57438 .29074 L F
.59301 .42227 m .59112 .44506 L .57458 .45249 L closepath p .722 .589 .707 r
F P
0 g
s
.59112 .44506 m .57576 .47768 L .57458 .45249 L closepath p .738 .634 .744 r
F P
0 g
s
.58879 .52442 m .57565 .54802 L .57568 .52722 L closepath p .728 .67 .794 r
F P
0 g
s
.58879 .52442 m .58338 .54004 L .57565 .54802 L closepath p .568 .469 .702 r
F P
0 g
s
.58338 .54004 m .57608 .57095 L .57565 .54802 L closepath p .685 .656 .816 r
F P
0 g
s
.59131 .5 m .57568 .52722 L .57817 .50631 L closepath p .74 .651 .761 r
F P
0 g
s
.59131 .5 m .58879 .52442 L .57568 .52722 L closepath p .723 .628 .752 r
F P
0 g
s
.59112 .44506 m .59297 .47374 L .57576 .47768 L closepath p .72 .613 .736 r
F P
0 g
s
.59297 .47374 m .57817 .50631 L .57576 .47768 L closepath p .723 .63 .753 r
F P
0 g
s
.58338 .54004 m .58236 .56763 L .57608 .57095 L closepath p .612 .583 .797 r
F P
0 g
s
.58236 .56763 m .57855 .57771 L .57608 .57095 L closepath p .664 .747 .912 r
F P
0 g
s
.59364 .30035 m .57754 .30775 L .5782 .29946 L closepath p .575 .215 .321 r
F P
0 g
s
.59364 .30035 m .5926 .30821 L .57754 .30775 L closepath p .607 .267 .368 r
F P
0 g
s
.57797 .3141 m .57754 .30775 L .5926 .30821 L closepath p .612 .282 .388 r
F P
0 g
s
.57764 .29581 m .59462 .28906 L .59261 .29286 L closepath p .357 0 0 r
F P
0 g
s
.57764 .29581 m .58104 .29083 L .59462 .28906 L closepath p .483 .063 .165 r
F P
0 g
s
.58039 .29178 m .57764 .29581 L p .57799 .29164 L .622 .214 .25 r
F P
0 g
s
.5782 .29946 m .57764 .29581 L .59261 .29286 L closepath p .549 .233 .386 r
F P
0 g
s
.5777 .29078 m .57813 .28716 L p .57792 .29079 L .372 .238 .566 r
F P
0 g
s
.57795 .29079 m .57792 .29079 L .57813 .28716 L .372 .238 .566 r
F
.57796 .29079 m .57795 .29079 L .57813 .28716 L F
.57799 .29079 m .57796 .29079 L .57813 .28716 L F
.57797 .3141 m .5926 .30821 L .59264 .31633 L closepath p .667 .384 .482 r
F P
0 g
s
.59264 .31633 m .57952 .31951 L .57797 .3141 L closepath p .667 .383 .481 r
F P
0 g
s
.57799 .29164 m .58104 .29083 L .58039 .29178 L p .622 .214 .25 r
F P
0 g
s
.57805 .29079 m .57799 .29079 L .57813 .28716 L .372 .238 .566 r
F
.57815 .29079 m .57805 .29079 L .57813 .28716 L F
.58607 .27209 m .57813 .28716 L .58084 .27405 L closepath p .329 .182 .524 r
F P
0 g
s
.58607 .27209 m .58406 .28524 L .57813 .28716 L closepath p .367 .224 .552 r
F P
0 g
s
.57813 .28716 m .57856 .27504 L .58084 .27405 L closepath p 0 0 .249 r
F P
0 g
s
.57813 .28716 m .5792 .28682 L p .57888 .2908 L .372 .238 .566 r
F P
0 g
s
.57888 .2908 m .57823 .29079 L .57813 .28716 L .372 .238 .566 r
F
.57823 .29079 m .57815 .29079 L .57813 .28716 L F
.59297 .47374 m .59131 .5 L .57817 .50631 L closepath p .739 .649 .76 r
F P
0 g
s
.5782 .29946 m .59261 .29286 L .59364 .30035 L closepath p .606 .333 .477 r
F P
0 g
s
.59113 .33179 m .57847 .34566 L .57953 .32998 L closepath p .712 .493 .591 r
F P
0 g
s
.59113 .33179 m .59239 .35041 L .57847 .34566 L closepath p .735 .527 .612 r
F P
0 g
s
.58539 .36104 m .57847 .34566 L .59239 .35041 L closepath p .74 .373 .356 r
F P
0 g
s
.58236 .56763 m .5849 .5801 L .57855 .57771 L closepath p .559 .629 .877 r
F P
0 g
s
.5849 .5801 m .58496 .5863 L .57855 .57771 L closepath p .342 .633 .96 r
F P
0 g
s
.58084 .27405 m .57856 .27504 L .58317 .25699 L closepath p 0 0 .222 r
F P
0 g
s
.57861 .5885 m .58496 .5863 L p .58949 .58636 L .303 .796 .858 r
F P
0 g
s
.58949 .58636 m .59179 .58639 L .57861 .5885 L .303 .796 .858 r
F
.57932 .2908 m .57888 .2908 L .5792 .28682 L .372 .238 .566 r
F
.5792 .28682 m .58406 .28524 L p .58237 .29065 L F P
0 g
s
.57958 .29081 m .57932 .2908 L .5792 .28682 L .372 .238 .566 r
F
.58143 .29077 m .57942 .2892 L .5792 .28682 L F
.58237 .29065 m .58193 .29071 L .5792 .28682 L F
.58193 .29071 m .58143 .29077 L .5792 .28682 L F
.58134 .29079 m .57942 .2892 L .58143 .29077 L F
.58104 .29083 m .57958 .29081 L .57942 .2892 L F
.58104 .29083 m .57942 .2892 L .58134 .29079 L F
.59264 .31633 m .59184 .32236 L .57952 .31951 L closepath p .69 .439 .538 r
F P
0 g
s
.57953 .32998 m .57952 .31951 L .59184 .32236 L closepath p .69 .436 .533 r
F P
0 g
s
.57953 .32998 m .59184 .32236 L .59113 .33179 L closepath p .711 .473 .563 r
F P
0 g
s
.59115 .25245 m .58084 .27405 L .58694 .2567 L closepath p .289 .12 .47 r
F P
0 g
s
.59115 .25245 m .58607 .27209 L .58084 .27405 L closepath p .313 .145 .487 r
F P
0 g
s
.58084 .27405 m .58317 .25699 L .58694 .2567 L closepath p .086 0 .346 r
F P
0 g
s
.60591 .54781 m .58236 .56763 L .58338 .54004 L closepath p .643 .609 .801 r
F P
0 g
s
.60591 .54781 m .6003 .56451 L .58236 .56763 L closepath p .601 .533 .75 r
F P
0 g
s
.6003 .56451 m .5849 .5801 L .58236 .56763 L closepath p .606 .662 .878 r
F P
0 g
s
.58406 .28524 m .58241 .29065 L p .58237 .29065 L .372 .238 .566 r
F P
0 g
s
.58241 .29065 m .58406 .28524 L p .58252 .29063 L .533 .399 .644 r
F P
0 g
s
.58406 .28524 m .58313 .29056 L .58252 .29063 L .533 .399 .644 r
F
.58406 .28524 m .58364 .29049 L .58313 .29056 L F
.5932 .23429 m .58317 .25699 L .5864 .23431 L closepath p .485 .375 .654 r
F P
0 g
s
.5932 .23429 m .58694 .2567 L .58317 .25699 L closepath p .09 0 .384 r
F P
0 g
s
.60494 .52808 m .58338 .54004 L .58879 .52442 L closepath p .681 .591 .748 r
F P
0 g
s
.60494 .52808 m .60591 .54781 L .58338 .54004 L closepath p .675 .579 .739 r
F P
0 g
s
.58406 .28524 m .58417 .29042 L p .58364 .29049 L .533 .399 .644 r
F P
0 g
s
.59918 .27706 m .58406 .28524 L .58607 .27209 L closepath p .662 .491 .642 r
F P
0 g
s
.59918 .27706 m .60032 .28756 L .58406 .28524 L closepath p .63 .428 .59 r
F P
0 g
s
.58406 .28524 m .59047 .28615 L p .59396 .28915 L .627 .473 .654 r
F P
0 g
s
.59396 .28915 m .59372 .28918 L .58406 .28524 L .627 .473 .654 r
F
.59372 .28918 m .593 .28927 L .58406 .28524 L F
.58673 .29009 m .58658 .29011 L .58406 .28524 L F
.58706 .29005 m .58673 .29009 L .58406 .28524 L F
.58658 .29011 m .58654 .29011 L .58406 .28524 L F
.58658 .29011 m .58658 .29011 L .58406 .28524 L F
.58654 .29011 m .58653 .29011 L .58406 .28524 L F
.58653 .29011 m .58609 .29017 L .58406 .28524 L F
.58609 .29017 m .58524 .29028 L .58406 .28524 L F
.58524 .29028 m .58506 .2903 L .58406 .28524 L F
.58505 .29031 m .58483 .29033 L .58406 .28524 L F
.58506 .2903 m .58505 .29031 L .58406 .28524 L F
.58483 .29033 m .58465 .29036 L .58406 .28524 L F
.58465 .29036 m .58433 .2904 L .58406 .28524 L F
.58417 .29042 m .58406 .28524 L p .58433 .2904 L F P
0 g
s
.58708 .29004 m .58706 .29005 L .58406 .28524 L .627 .473 .654 r
F
.58709 .29004 m .58708 .29004 L .58406 .28524 L F
.59201 .2894 m .58993 .28967 L .58406 .28524 L F
.58993 .28967 m .58709 .29004 L .58406 .28524 L F
.593 .28927 m .59201 .2894 L .58406 .28524 L F
.6003 .56451 m .60228 .57824 L .5849 .5801 L closepath p .561 .586 .836 r
F P
0 g
s
.60228 .57824 m .59179 .58639 L .5849 .5801 L closepath p .496 .751 .986 r
F P
0 g
s
.5849 .5801 m .59179 .58639 L .58496 .5863 L closepath p .444 .726 .986 r
F P
0 g
s
.58516 .20421 m .59807 .2013 L .5928 .2146 L closepath p .601 .377 .549 r
F P
0 g
s
.58516 .20421 m .5928 .2146 L .58808 .21442 L closepath p .396 .322 .655 r
F P
0 g
s
.58752 .21658 m .58516 .20421 L .58808 .21442 L closepath p 0 0 .37 r
F P
0 g
s
.58539 .36104 m .59239 .35041 L .59666 .36316 L closepath p .81 .464 .401 r
F P
0 g
s
.58781 .36995 m .58539 .36104 L .59666 .36316 L closepath p .811 .466 .405 r
F P
0 g
s
.58607 .27209 m .59115 .25245 L .60234 .25502 L closepath p .697 .516 .64 r
F P
0 g
s
.58607 .27209 m .60234 .25502 L .59918 .27706 L closepath p .67 .476 .614 r
F P
0 g
s
.58808 .21442 m .5932 .23429 L .5864 .23431 L closepath p .485 .364 .64 r
F P
0 g
s
.59115 .25245 m .58694 .2567 L .5932 .23429 L closepath p .331 .178 .518 r
F P
0 g
s
.58729 .22386 m .58752 .21658 L .58808 .21442 L p 0 0 .061 r
F P
0 g
s
.58781 .36995 m .59666 .36316 L .60206 .37729 L closepath p .847 .529 .449 r
F P
0 g
s
.58932 .38833 m .58781 .36995 L .60206 .37729 L closepath p .83 .615 .611 r
F P
0 g
s
.58808 .21442 m .5928 .2146 L .5932 .23429 L closepath p .397 .309 .639 r
F P
0 g
s
.58879 .52442 m .59131 .5 L .59495 .49946 L closepath p .773 .674 .753 r
F P
0 g
s
.59828 .50066 m .58879 .52442 L .59495 .49946 L closepath p .725 .624 .745 r
F P
0 g
s
.60865 .50249 m .58879 .52442 L .59828 .50066 L closepath p .715 .613 .741 r
F P
0 g
s
.60865 .50249 m .60494 .52808 L .58879 .52442 L closepath p .689 .572 .718 r
F P
0 g
s
.58932 .38833 m .60206 .37729 L .60414 .39228 L closepath p .839 .631 .621 r
F P
0 g
s
.59086 .4072 m .58932 .38833 L .60414 .39228 L closepath p .84 .567 .526 r
F P
0 g
s
.59047 .28615 m .60032 .28756 L .6 .28817 L p .627 .473 .654 r
F P
0 g
s
.6 .28817 m .5962 .2888 L .59047 .28615 L .627 .473 .654 r
F
.59462 .28906 m .59396 .28915 L .59047 .28615 L F
.5955 .28892 m .59462 .28906 L .59047 .28615 L F
.5962 .2888 m .5955 .28892 L .59047 .28615 L F
.59086 .4072 m .60414 .39228 L .60411 .41047 L closepath p .839 .566 .525 r
F P
0 g
s
.59301 .42227 m .59086 .4072 L .60411 .41047 L closepath p .84 .614 .596 r
F P
0 g
s
.59112 .44506 m .59301 .42227 L .60304 .42543 L closepath p .829 .685 .704 r
F P
0 g
s
.59112 .44506 m .60304 .42543 L .60291 .44178 L closepath p .81 .657 .693 r
F P
0 g
s
.59297 .47374 m .59112 .44506 L .60291 .44178 L closepath p .816 .685 .719 r
F P
0 g
s
.59113 .33179 m .59184 .32236 L .61019 .33634 L closepath p .794 .547 .564 r
F P
0 g
s
.59113 .33179 m .61019 .33634 L .60469 .34848 L closepath p .794 .56 .583 r
F P
0 g
s
.60469 .34848 m .59239 .35041 L .59113 .33179 L closepath p .778 .561 .605 r
F P
0 g
s
.59115 .25245 m .5932 .23429 L .59621 .23288 L closepath p .507 .329 .574 r
F P
0 g
s
.59115 .25245 m .59621 .23288 L .60549 .23311 L closepath p .743 .564 .652 r
F P
0 g
s
.59115 .25245 m .60549 .23311 L .60234 .25502 L closepath p .699 .503 .619 r
F P
0 g
s
.59131 .5 m .59297 .47374 L .59814 .47382 L closepath p .833 .73 .751 r
F P
0 g
s
.59131 .5 m .59814 .47382 L .59495 .49946 L closepath p .773 .666 .745 r
F P
0 g
s
.60228 .57824 m .60864 .58589 L .59179 .58639 L closepath p .488 .734 .981 r
F P
0 g
s
.59924 .58695 m .59179 .58639 L .60864 .58589 L closepath p .103 .656 .798 r
F P
0 g
s
.6106 .32175 m .59184 .32236 L .59264 .31633 L closepath p .758 .502 .546 r
F P
0 g
s
.6106 .32175 m .61019 .33634 L .59184 .32236 L closepath p .771 .565 .62 r
F P
0 g
s
.60433 .36079 m .59239 .35041 L .60469 .34848 L closepath p .77 .534 .576 r
F P
0 g
s
.59666 .36316 m .59239 .35041 L .60433 .36079 L closepath p .827 .472 .385 r
F P
0 g
s
.60116 .30467 m .5926 .30821 L .59364 .30035 L closepath p .676 .329 .379 r
F P
0 g
s
.60726 .31242 m .5926 .30821 L .60116 .30467 L closepath p .755 .477 .512 r
F P
0 g
s
.6106 .32175 m .59264 .31633 L .5926 .30821 L closepath p .759 .46 .478 r
F P
0 g
s
.6106 .32175 m .5926 .30821 L .60726 .31242 L closepath p .755 .464 .492 r
F P
0 g
s
.59261 .29286 m .60641 .2871 L .60035 .29217 L closepath p .469 .055 .171 r
F P
0 g
s
.59261 .29286 m .59462 .28906 L .60641 .2871 L closepath p .33 0 0 r
F P
0 g
s
.59261 .29286 m .60035 .29217 L .59364 .30035 L closepath p .577 .313 .479 r
F P
0 g
s
.59807 .2013 m .60073 .22004 L .5928 .2146 L closepath p .699 .484 .593 r
F P
0 g
s
.5928 .2146 m .60073 .22004 L .59621 .23288 L closepath p .658 .505 .664 r
F P
0 g
s
.5928 .2146 m .59621 .23288 L .5932 .23429 L closepath p .558 .438 .671 r
F P
0 g
s
.59297 .47374 m .60291 .44178 L .59814 .47382 L closepath p .836 .707 .722 r
F P
0 g
s
.59301 .42227 m .60411 .41047 L .60304 .42543 L closepath p .845 .623 .601 r
F P
0 g
s
.59364 .30035 m .60035 .29217 L .61226 .29579 L closepath p .609 .354 .506 r
F P
0 g
s
.59364 .30035 m .61226 .29579 L .60116 .30467 L closepath p .628 .402 .556 r
F P
0 g
s
.60615 .47041 m .59495 .49946 L .59814 .47382 L closepath p .768 .662 .745 r
F P
0 g
s
.60615 .47041 m .59828 .50066 L .59495 .49946 L closepath p .73 .618 .733 r
F P
0 g
s
.59501 .20199 m .59795 .193 L .59807 .2013 L p 0 0 .189 r
F P
0 g
s
.59621 .23288 m .60073 .22004 L .60717 .22308 L closepath p .752 .608 .698 r
F P
0 g
s
.59621 .23288 m .60717 .22308 L .60549 .23311 L closepath p .744 .594 .689 r
F P
0 g
s
.60433 .36079 m .60754 .36814 L .59666 .36316 L closepath p .852 .525 .433 r
F P
0 g
s
.60754 .36814 m .60206 .37729 L .59666 .36316 L closepath p .852 .531 .443 r
F P
0 g
s
.59795 .193 m .59865 .1868 L .59807 .2013 L closepath p 0 0 .04 r
F P
0 g
s
.59807 .2013 m .59865 .1868 L .6079 .1899 L closepath p .527 .422 .679 r
F P
0 g
s
.59807 .2013 m .6079 .1899 L .60511 .20687 L closepath p .513 .403 .666 r
F P
0 g
s
.59807 .2013 m .60511 .20687 L .60073 .22004 L closepath p .571 .395 .604 r
F P
0 g
s
.61211 .44492 m .59814 .47382 L .60291 .44178 L closepath p .861 .732 .719 r
F P
0 g
s
.61211 .44492 m .60615 .47041 L .59814 .47382 L closepath p .737 .572 .669 r
F P
0 g
s
.6159 .47405 m .59828 .50066 L .60615 .47041 L closepath p .785 .675 .743 r
F P
0 g
s
.6159 .47405 m .60865 .50249 L .59828 .50066 L closepath p .723 .587 .703 r
F P
0 g
s
.59865 .1868 m .59984 .17629 L .61152 .17878 L closepath p .538 .368 .599 r
F P
0 g
s
.59865 .1868 m .61152 .17878 L .6079 .1899 L closepath p .549 .388 .615 r
F P
0 g
s
.6065 .2559 m .59918 .27706 L .60234 .25502 L closepath p .95 .753 .602 r
F P
0 g
s
.6065 .2559 m .61136 .27562 L .59918 .27706 L closepath p .766 .522 .565 r
F P
0 g
s
.61136 .27562 m .60032 .28756 L .59918 .27706 L closepath p .768 .529 .572 r
F P
0 g
s
.61182 .5863 m .59924 .58695 L .60864 .58589 L p .096 .652 .771 r
F P
0 g
s
.61152 .17878 m .59984 .17629 L .60387 .16812 L closepath p .532 .393 .637 r
F P
0 g
s
.6 .28817 m .60032 .28756 L p .60002 .28816 L .556 .39 .612 r
F P
0 g
s
.60032 .28756 m .60006 .28816 L .60002 .28816 L .556 .39 .612 r
F
.60032 .28756 m .60014 .28814 L .60006 .28816 L F
.60032 .28756 m .60018 .28814 L .60014 .28814 L F
.60032 .28756 m .60023 .28813 L .60018 .28814 L F
.60032 .28756 m .60028 .28812 L .60023 .28813 L F
.60032 .28756 m .60028 .28812 L .60028 .28812 L F
.60032 .28756 m .6004 .2881 L p .60028 .28812 L F P
0 g
s
.62949 .55103 m .6003 .56451 L .60591 .54781 L closepath p .632 .566 .763 r
F P
0 g
s
.62949 .55103 m .62891 .56555 L .6003 .56451 L closepath p .616 .524 .728 r
F P
0 g
s
.62891 .56555 m .60228 .57824 L .6003 .56451 L closepath p .596 .612 .838 r
F P
0 g
s
.60032 .28756 m .61136 .27562 L .6125 .282 L p .797 .574 .599 r
F P
0 g
s
.60032 .28756 m .60489 .28735 L p .60111 .28798 L .793 .555 .577 r
F P
0 g
s
.60089 .28802 m .60089 .28802 L .60032 .28756 L .793 .555 .577 r
F
.60111 .28798 m .60106 .28799 L .60032 .28756 L F
.60106 .28799 m .60089 .28802 L .60032 .28756 L F
.60089 .28802 m .60076 .28804 L .60032 .28756 L F
.60076 .28804 m .6007 .28805 L .60032 .28756 L F
.6007 .28805 m .6007 .28805 L .60032 .28756 L F
.6007 .28805 m .60067 .28805 L .60032 .28756 L F
.6004 .2881 m .60032 .28756 L p .60067 .28805 L F P
0 g
s
.6125 .282 m .60641 .2871 L .60032 .28756 L .797 .574 .599 r
F
.60489 .28735 m .60032 .28756 L p .60641 .2871 L F P
0 g
s
.60035 .29217 m .60641 .2871 L .62126 .29131 L closepath p .631 .26 .321 r
F P
0 g
s
.60035 .29217 m .62126 .29131 L .61226 .29579 L closepath p .607 .206 .258 r
F P
0 g
s
.60511 .20687 m .60717 .22308 L .60073 .22004 L closepath p .761 .597 .675 r
F P
0 g
s
.60116 .30467 m .61226 .29579 L .61909 .30352 L closepath p .666 .463 .601 r
F P
0 g
s
.60116 .30467 m .61909 .30352 L .60726 .31242 L closepath p .668 .471 .61 r
F P
0 g
s
.60754 .36814 m .6121 .38239 L .60206 .37729 L closepath p .903 .606 .474 r
F P
0 g
s
.6121 .38239 m .60414 .39228 L .60206 .37729 L closepath p .897 .67 .587 r
F P
0 g
s
.62891 .56555 m .63228 .58057 L .60228 .57824 L closepath p .584 .575 .809 r
F P
0 g
s
.63228 .58057 m .60864 .58589 L .60228 .57824 L closepath p .444 .71 .979 r
F P
0 g
s
.6065 .2559 m .60234 .25502 L .60549 .23311 L closepath p .95 .753 .602 r
F P
0 g
s
.61474 .4287 m .60291 .44178 L .60304 .42543 L closepath p .874 .71 .676 r
F P
0 g
s
.61474 .4287 m .61211 .44492 L .60291 .44178 L closepath p .872 .707 .674 r
F P
0 g
s
.61458 .41369 m .60304 .42543 L .60411 .41047 L closepath p .873 .649 .596 r
F P
0 g
s
.61458 .41369 m .61474 .4287 L .60304 .42543 L closepath p .881 .665 .605 r
F P
0 g
s
.61362 .1646 m .61152 .17878 L .60387 .16812 L closepath p .595 .41 .601 r
F P
0 g
s
.61061 .39935 m .60411 .41047 L .60414 .39228 L closepath p .883 .604 .511 r
F P
0 g
s
.61061 .39935 m .61458 .41369 L .60411 .41047 L closepath p .872 .589 .505 r
F P
0 g
s
.6121 .38239 m .61061 .39935 L .60414 .39228 L closepath p .866 .617 .56 r
F P
0 g
s
.62321 .36501 m .60433 .36079 L .60469 .34848 L closepath p .727 .497 .577 r
F P
0 g
s
.62321 .36501 m .60754 .36814 L .60433 .36079 L closepath p .728 .478 .55 r
F P
0 g
s
.61455 .34485 m .60469 .34848 L .61019 .33634 L closepath p .789 .554 .58 r
F P
0 g
s
.61455 .34485 m .62321 .36501 L .60469 .34848 L closepath p .758 .483 .516 r
F P
0 g
s
.62974 .50851 m .60494 .52808 L .60865 .50249 L closepath p .659 .543 .712 r
F P
0 g
s
.62974 .50851 m .63013 .53532 L .60494 .52808 L closepath p .655 .537 .708 r
F P
0 g
s
.63013 .53532 m .60591 .54781 L .60494 .52808 L closepath p .644 .554 .738 r
F P
0 g
s
.60511 .20687 m .6079 .1899 L .62228 .19676 L closepath p .732 .609 .72 r
F P
0 g
s
.60511 .20687 m .62228 .19676 L .61319 .20897 L closepath p .732 .608 .719 r
F P
0 g
s
.60511 .20687 m .61319 .20897 L .60717 .22308 L closepath p .742 .583 .679 r
F P
0 g
s
.61439 .2213 m .60906 .2354 L .60549 .23311 L closepath p .83 .681 .699 r
F P
0 g
s
.61439 .2213 m .60549 .23311 L .60717 .22308 L closepath p .823 .671 .694 r
F P
0 g
s
.6065 .2559 m .60549 .23311 L .60906 .2354 L closepath p .842 .67 .668 r
F P
0 g
s
.63013 .53532 m .62949 .55103 L .60591 .54781 L closepath p .639 .542 .728 r
F P
0 g
s
.61911 .45135 m .60615 .47041 L .61211 .44492 L closepath p .739 .574 .669 r
F P
0 g
s
.61911 .45135 m .6159 .47405 L .60615 .47041 L closepath p .797 .657 .708 r
F P
0 g
s
.6125 .282 m .60697 .28725 L .60641 .2871 L .797 .574 .599 r
F
.60714 .25773 m .6065 .2559 L .60906 .2354 L closepath p .95 .824 .384 r
F P
0 g
s
.6065 .2559 m .60714 .25773 L .61048 .27207 L p .765 .486 0 r
F P
0 g
s
.60697 .28725 m .61338 .28696 L .61249 .28882 L p .793 .555 .577 r
F P
0 g
s
.6125 .282 m .61338 .28696 L .60697 .28725 L p .797 .574 .599 r
F P
0 g
s
.60714 .25773 m .60906 .2354 L .60938 .23783 L closepath p .87 .76 .253 r
F P
0 g
s
.61876 .21092 m .60717 .22308 L .61319 .20897 L closepath p .809 .665 .704 r
F P
0 g
s
.61876 .21092 m .61439 .2213 L .60717 .22308 L closepath p .828 .699 .722 r
F P
0 g
s
.61909 .30352 m .62515 .31026 L .60726 .31242 L closepath p .684 .5 .631 r
F P
0 g
s
.60726 .31242 m .62515 .31026 L .63019 .32328 L closepath p .69 .528 .662 r
F P
0 g
s
.60726 .31242 m .63019 .32328 L .62039 .32307 L closepath p .727 .466 .532 r
F P
0 g
s
.6106 .32175 m .60726 .31242 L .62039 .32307 L closepath p .74 .457 .501 r
F P
0 g
s
.60754 .36814 m .62321 .36501 L .62894 .37132 L closepath p .742 .52 .593 r
F P
0 g
s
.62894 .37132 m .6121 .38239 L .60754 .36814 L closepath p .742 .528 .604 r
F P
0 g
s
.6079 .1899 m .61152 .17878 L .62671 .1903 L closepath p .75 .601 .691 r
F P
0 g
s
.6079 .1899 m .62671 .1903 L .62228 .19676 L closepath p .75 .59 .678 r
F P
0 g
s
.63228 .58057 m .63956 .58985 L .60864 .58589 L closepath p .456 .642 .937 r
F P
0 g
s
.63619 .48003 m .60865 .50249 L .6159 .47405 L closepath p .684 .547 .693 r
F P
0 g
s
.63619 .48003 m .62974 .50851 L .60865 .50249 L closepath p .669 .522 .676 r
F P
0 g
s
.61543 .22487 m .60906 .2354 L .61439 .2213 L closepath p .972 .926 .623 r
F P
0 g
s
.61543 .22487 m .60938 .23783 L .60906 .2354 L closepath p .932 .94 .593 r
F P
0 g
s
.61019 .33634 m .6106 .32175 L .63164 .33792 L closepath p .754 .55 .621 r
F P
0 g
s
.61019 .33634 m .63164 .33792 L .61455 .34485 L closepath p .753 .542 .61 r
F P
0 g
s
.62039 .32307 m .63164 .33792 L .6106 .32175 L closepath p .747 .554 .633 r
F P
0 g
s
.61061 .39935 m .6121 .38239 L .62339 .39285 L closepath p .786 .541 .567 r
F P
0 g
s
.62339 .39285 m .62802 .41729 L .61061 .39935 L closepath p .761 .492 .527 r
F P
0 g
s
.62802 .41729 m .61458 .41369 L .61061 .39935 L closepath p .716 .501 .598 r
F P
0 g
s
.61338 .28696 m .61136 .27562 L .61429 .28933 L p .808 .549 0 r
F P
0 g
s
.61362 .1646 m .62365 .16406 L .61152 .17878 L closepath p .648 .458 .613 r
F P
0 g
s
.63628 .16873 m .61152 .17878 L .62365 .16406 L closepath p .695 .524 .652 r
F P
0 g
s
.63628 .16873 m .63131 .18061 L .61152 .17878 L closepath p .715 .576 .698 r
F P
0 g
s
.62671 .1903 m .61152 .17878 L .63131 .18061 L closepath p .709 .612 .746 r
F P
0 g
s
.62339 .39285 m .6121 .38239 L .62894 .37132 L closepath p .755 .553 .623 r
F P
0 g
s
.62618 .43343 m .61211 .44492 L .61474 .4287 L closepath p .708 .546 .667 r
F P
0 g
s
.62618 .43343 m .61911 .45135 L .61211 .44492 L closepath p .724 .576 .688 r
F P
0 g
s
.61226 .29579 m .62126 .29131 L .62983 .29947 L closepath p .733 .408 .431 r
F P
0 g
s
.62983 .29947 m .61909 .30352 L .61226 .29579 L closepath p .74 .46 .506 r
F P
0 g
s
.61249 .28882 m .61338 .28696 L p .61312 .289 L .872 .659 .611 r
F P
0 g
s
.61338 .28696 m .61348 .2891 L p .61312 .289 L .872 .659 .611 r
F P
0 g
s
.6278 .19911 m .61319 .20897 L .62228 .19676 L closepath p .824 .752 .783 r
F P
0 g
s
.6278 .19911 m .61876 .21092 L .61319 .20897 L closepath p .799 .683 .737 r
F P
0 g
s
.61355 .28912 m .61338 .28696 L p .61355 .28912 L .89 .687 .171 r
F P
0 g
s
.61338 .28696 m .6136 .28913 L .61359 .28913 L .89 .687 .171 r
F
.61338 .28696 m .61424 .28932 L p .61399 .28925 L F P
0 g
s
.61338 .28696 m .61399 .28925 L .61369 .28916 L .89 .687 .171 r
F
.61338 .28696 m .61369 .28916 L .61362 .28914 L F
.61338 .28696 m .61362 .28914 L .6136 .28913 L F
.61338 .28696 m .61359 .28913 L .61355 .28912 L F
.61348 .2891 m .61338 .28696 L .61355 .28912 L p .894 .69 .177 r
F P
0 g
s
.61424 .28932 m .61338 .28696 L p .61429 .28933 L .808 .549 0 r
F P
0 g
s
.61543 .22487 m .61439 .2213 L .61876 .21092 L closepath p .965 .939 .665 r
F P
0 g
s
.63164 .33792 m .63468 .36083 L .61455 .34485 L closepath p .735 .499 .572 r
F P
0 g
s
.61455 .34485 m .63468 .36083 L .62321 .36501 L closepath p .758 .483 .516 r
F P
0 g
s
.62802 .41729 m .61474 .4287 L .61458 .41369 L closepath p .711 .525 .636 r
F P
0 g
s
.62802 .41729 m .62618 .43343 L .61474 .4287 L closepath p .716 .533 .641 r
F P
0 g
s
.61543 .22487 m .61876 .21092 L .6206 .21211 L closepath p .965 .939 .664 r
F P
0 g
s
.64252 .44398 m .6159 .47405 L .61911 .45135 L closepath p .685 .551 .697 r
F P
0 g
s
.64252 .44398 m .63619 .48003 L .6159 .47405 L closepath p .683 .549 .696 r
F P
0 g
s
.62756 .20384 m .61876 .21092 L .6278 .19911 L closepath p .915 .964 .796 r
F P
0 g
s
.62756 .20384 m .6206 .21211 L .61876 .21092 L closepath p .893 .985 .798 r
F P
0 g
s
.61909 .30352 m .62983 .29947 L .63629 .30785 L closepath p .758 .498 .54 r
F P
0 g
s
.61909 .30352 m .63629 .30785 L .62515 .31026 L closepath p .758 .496 .537 r
F P
0 g
s
.64252 .44398 m .61911 .45135 L .62618 .43343 L closepath p .675 .52 .668 r
F P
0 g
s
.63164 .33792 m .62039 .32307 L .63019 .32328 L closepath p .741 .553 .64 r
F P
0 g
s
.62126 .29131 m .64314 .30645 L .62983 .29947 L closepath p .679 .142 0 r
F P
0 g
s
.62126 .29131 m .63404 .29999 L .64314 .30645 L closepath p .426 0 0 r
F P
0 g
s
.62671 .1903 m .6278 .19911 L .62228 .19676 L closepath p .844 .731 .739 r
F P
0 g
s
.62894 .37132 m .62321 .36501 L .63468 .36083 L closepath p .773 .516 .546 r
F P
0 g
s
.62894 .37132 m .64111 .37894 L .62339 .39285 L closepath p .759 .557 .623 r
F P
0 g
s
.63522 .40152 m .62339 .39285 L .64111 .37894 L closepath p .724 .496 .581 r
F P
0 g
s
.62339 .39285 m .63522 .40152 L .62802 .41729 L closepath p .744 .481 .533 r
F P
0 g
s
.64893 .17175 m .63628 .16873 L .62365 .16406 L closepath p .456 0 0 r
F P
0 g
s
.62515 .31026 m .63629 .30785 L .64014 .32412 L closepath p .783 .569 .611 r
F P
0 g
s
.62515 .31026 m .64014 .32412 L .63019 .32328 L closepath p .78 .57 .617 r
F P
0 g
s
.62756 .20384 m .62634 .20597 L .62529 .20653 L p .629 .962 .849 r
F P
0 g
s
.64966 .42437 m .62618 .43343 L .62802 .41729 L closepath p .691 .511 .638 r
F P
0 g
s
.64966 .42437 m .64252 .44398 L .62618 .43343 L closepath p .691 .511 .638 r
F P
0 g
s
.63131 .18061 m .62927 .19391 L .62671 .1903 L closepath p .84 .779 .793 r
F P
0 g
s
.62671 .1903 m .62927 .19391 L .6278 .19911 L closepath p .917 .778 .691 r
F P
0 g
s
.62756 .20384 m .62927 .19391 L .6305 .19666 L closepath p .966 .931 .616 r
F P
0 g
s
.62756 .20384 m .6278 .19911 L .62927 .19391 L closepath p .91 .875 .444 r
F P
0 g
s
.65508 .41304 m .62802 .41729 L .63522 .40152 L closepath p .702 .432 .513 r
F P
0 g
s
.65508 .41304 m .64966 .42437 L .62802 .41729 L closepath p .707 .447 .529 r
F P
0 g
s
.6305 .19666 m .62944 .19989 L .62812 .20248 L p .775 .884 .466 r
F P
0 g
s
.65548 .55619 m .62891 .56555 L .62949 .55103 L closepath p .573 .487 .72 r
F P
0 g
s
.65548 .55619 m .65499 .56963 L .62891 .56555 L closepath p .565 .465 .7 r
F P
0 g
s
.65499 .56963 m .63228 .58057 L .62891 .56555 L closepath p .533 .54 .808 r
F P
0 g
s
.63468 .36083 m .64111 .37894 L .62894 .37132 L closepath p .783 .528 .551 r
F P
0 g
s
.63131 .18061 m .63837 .18402 L .62927 .19391 L closepath p .884 .821 .781 r
F P
0 g
s
.63837 .18402 m .6305 .19666 L .62927 .19391 L closepath p .902 .977 .804 r
F P
0 g
s
.63804 .18804 m .63859 .19039 L .62944 .19989 L closepath p .875 .989 .82 r
F P
0 g
s
.63804 .18804 m .62944 .19989 L .6305 .19666 L closepath p .825 .99 .731 r
F P
0 g
s
.65602 .54272 m .62949 .55103 L .63013 .53532 L closepath p .583 .494 .72 r
F P
0 g
s
.65602 .54272 m .65548 .55619 L .62949 .55103 L closepath p .577 .475 .702 r
F P
0 g
s
.65658 .4883 m .62974 .50851 L .63619 .48003 L closepath p .588 .442 .651 r
F P
0 g
s
.65658 .4883 m .65632 .51578 L .62974 .50851 L closepath p .599 .462 .666 r
F P
0 g
s
.65632 .51578 m .63013 .53532 L .62974 .50851 L closepath p .588 .483 .702 r
F P
0 g
s
.62983 .29947 m .64918 .31469 L .63629 .30785 L closepath p .832 .361 .005 r
F P
0 g
s
.62983 .29947 m .64314 .30645 L .64918 .31469 L closepath p .818 .338 0 r
F P
0 g
s
.65632 .51578 m .65602 .54272 L .63013 .53532 L closepath p .589 .485 .703 r
F P
0 g
s
.63019 .32328 m .64014 .32412 L .64389 .33824 L closepath p .78 .576 .624 r
F P
0 g
s
.63019 .32328 m .64389 .33824 L .63164 .33792 L closepath p .773 .577 .633 r
F P
0 g
s
.63837 .18402 m .63804 .18804 L .6305 .19666 L closepath p .901 .979 .801 r
F P
0 g
s
.64893 .17175 m .63837 .18402 L .63131 .18061 L closepath p .88 .825 .789 r
F P
0 g
s
.64893 .17175 m .63131 .18061 L .63628 .16873 L closepath p .87 .767 .746 r
F P
0 g
s
.6446 .356 m .63468 .36083 L .63164 .33792 L closepath p .743 .505 .57 r
F P
0 g
s
.63164 .33792 m .65183 .35228 L .6446 .356 L closepath p .769 .508 .54 r
F P
0 g
s
.63164 .33792 m .64389 .33824 L .65183 .35228 L closepath p .764 .513 .555 r
F P
0 g
s
.65499 .56963 m .65836 .58468 L .63228 .58057 L closepath p .533 .541 .809 r
F P
0 g
s
.65836 .58468 m .63956 .58985 L .63228 .58057 L closepath p .435 .631 .937 r
F P
0 g
s
.63404 .29999 m .64231 .30532 L p .64314 .30645 L .648 .111 0 r
F P
0 g
s
.63468 .36083 m .6446 .356 L .65014 .36392 L closepath p .737 .493 .56 r
F P
0 g
s
.63468 .36083 m .65014 .36392 L .64111 .37894 L closepath p .737 .506 .578 r
F P
0 g
s
.66051 .39945 m .63522 .40152 L .64111 .37894 L closepath p .696 .468 .573 r
F P
0 g
s
.66051 .39945 m .65508 .41304 L .63522 .40152 L closepath p .691 .448 .55 r
F P
0 g
s
.66346 .45281 m .63619 .48003 L .64252 .44398 L closepath p .645 .513 .687 r
F P
0 g
s
.66346 .45281 m .65658 .4883 L .63619 .48003 L closepath p .594 .434 .635 r
F P
0 g
s
.63629 .30785 m .64918 .31469 L .65224 .32996 L closepath p .965 .665 .408 r
F P
0 g
s
.63629 .30785 m .65224 .32996 L .64014 .32412 L closepath p .965 .667 .415 r
F P
0 g
s
.63859 .19039 m .63804 .18804 L .6492 .17935 L closepath p .849 .986 .857 r
F P
0 g
s
.6492 .17935 m .63804 .18804 L .63837 .18402 L closepath p .868 .965 .859 r
F P
0 g
s
.6492 .17935 m .63837 .18402 L .64893 .17175 L closepath p .906 .902 .81 r
F P
0 g
s
.65836 .58468 m .66148 .59261 L .63956 .58985 L closepath p .433 .604 .918 r
F P
0 g
s
.66148 .59261 m .66089 .59447 L .63956 .58985 L closepath p .445 .594 .906 r
F P
0 g
s
.64014 .32412 m .65224 .32996 L .65557 .34357 L closepath p .959 .642 .379 r
F P
0 g
s
.64014 .32412 m .65557 .34357 L .64389 .33824 L closepath p .959 .664 .431 r
F P
0 g
s
.67128 .38828 m .64111 .37894 L .65014 .36392 L closepath p .709 .47 .561 r
F P
0 g
s
.67128 .38828 m .66051 .39945 L .64111 .37894 L closepath p .71 .467 .555 r
F P
0 g
s
.64314 .30645 m .64231 .30532 L .64364 .30685 L .648 .111 0 r
F
.64364 .30685 m .64231 .30532 L .64475 .30773 L F
.64231 .30532 m .64883 .30952 L p .64475 .30773 L F P
0 g
s
.66776 .43198 m .64252 .44398 L .64966 .42437 L closepath p .678 .497 .633 r
F P
0 g
s
.66776 .43198 m .66346 .45281 L .64252 .44398 L closepath p .67 .48 .618 r
F P
0 g
s
.66144 .32437 m .64918 .31469 L .64314 .30645 L closepath p .771 .285 0 r
F P
0 g
s
.66144 .32437 m .64314 .30645 L .65676 .31722 L closepath p .792 .312 0 r
F P
0 g
s
.65557 .34357 m .65183 .35228 L .64389 .33824 L closepath p .903 .494 .198 r
F P
0 g
s
.65183 .35228 m .65014 .36392 L .6446 .356 L closepath p .763 .496 .53 r
F P
0 g
s
.6455 .30832 m .64475 .30773 L .64883 .30952 L .648 .111 0 r
F
.65264 .31397 m .6455 .30832 L .64883 .30952 L F
.64883 .30952 m .65602 .3165 L p .65264 .31397 L F P
0 g
s
.66377 .33945 m .64918 .31469 L .66144 .32437 L closepath p .972 .67 .386 r
F P
0 g
s
.66377 .33945 m .65224 .32996 L .64918 .31469 L closepath p .971 .667 .378 r
F P
0 g
s
.67275 .42402 m .64966 .42437 L .65508 .41304 L closepath p .69 .427 .52 r
F P
0 g
s
.67275 .42402 m .66776 .43198 L .64966 .42437 L closepath p .697 .458 .558 r
F P
0 g
s
.65183 .35228 m .65677 .35786 L .65014 .36392 L closepath p .677 .415 .518 r
F P
0 g
s
.6785 .38221 m .65014 .36392 L .65677 .35786 L closepath p .719 .478 .561 r
F P
0 g
s
.6785 .38221 m .67128 .38828 L .65014 .36392 L closepath p .726 .47 .539 r
F P
0 g
s
.65557 .34357 m .66333 .35659 L .65183 .35228 L closepath p .882 .458 .185 r
F P
0 g
s
.66764 .36301 m .65677 .35786 L .65183 .35228 L closepath p .764 .253 0 r
F P
0 g
s
.66764 .36301 m .65183 .35228 L .66333 .35659 L closepath p .861 .416 .132 r
F P
0 g
s
.66546 .35207 m .65557 .34357 L .65224 .32996 L closepath p .967 .642 .315 r
F P
0 g
s
.66546 .35207 m .65224 .32996 L .66377 .33945 L closepath p .972 .668 .38 r
F P
0 g
s
.65308 .31431 m .65264 .31397 L .65602 .3165 L .648 .111 0 r
F
.65317 .31438 m .65308 .31431 L .65602 .3165 L F
.65574 .31641 m .65317 .31438 L .65602 .3165 L F
.67311 .57532 m .65499 .56963 L .65548 .55619 L closepath p .559 .46 .699 r
F P
0 g
s
.67704 .59272 m .65836 .58468 L .65499 .56963 L closepath p .5 .518 .806 r
F P
0 g
s
.67704 .59272 m .65499 .56963 L .67311 .57532 L closepath p .502 .518 .806 r
F P
0 g
s
.67633 .41723 m .65508 .41304 L .66051 .39945 L closepath p .697 .454 .552 r
F P
0 g
s
.67633 .41723 m .67275 .42402 L .65508 .41304 L closepath p .696 .421 .504 r
F P
0 g
s
.67619 .56364 m .65548 .55619 L .65602 .54272 L closepath p .55 .452 .696 r
F P
0 g
s
.67311 .57532 m .65548 .55619 L .67619 .56364 L closepath p .544 .458 .708 r
F P
0 g
s
.66546 .35207 m .66333 .35659 L .65557 .34357 L closepath p .83 .378 0 r
F P
0 g
s
.65602 .3165 m .65676 .31722 L p .65574 .31641 L .648 .111 0 r
F P
0 g
s
.67686 .5478 m .65602 .54272 L .65632 .51578 L closepath p .534 .44 .694 r
F P
0 g
s
.67619 .56364 m .65602 .54272 L .67686 .5478 L closepath p .53 .449 .709 r
F P
0 g
s
.68235 .52612 m .65632 .51578 L .65658 .4883 L closepath p .572 .44 .662 r
F P
0 g
s
.67686 .5478 m .65632 .51578 L .68235 .52612 L closepath p .562 .45 .683 r
F P
0 g
s
.68505 .4979 m .65658 .4883 L .66346 .45281 L closepath p .585 .426 .632 r
F P
0 g
s
.68235 .52612 m .65658 .4883 L .68505 .4979 L closepath p .577 .441 .66 r
F P
0 g
s
.68554 .38782 m .65677 .35786 L .66764 .36301 L closepath p .836 .365 .033 r
F P
0 g
s
.68554 .38782 m .6785 .38221 L .65677 .35786 L closepath p .678 .17 0 r
F P
0 g
s
.67851 .59725 m .66148 .59261 L .65836 .58468 L closepath p .342 .54 .903 r
F P
0 g
s
.67851 .59725 m .65836 .58468 L .67704 .59272 L closepath p .36 .544 .9 r
F P
0 g
s
.68471 .40538 m .66051 .39945 L .67128 .38828 L closepath p .706 .461 .551 r
F P
0 g
s
.68471 .40538 m .67633 .41723 L .66051 .39945 L closepath p .707 .453 .539 r
F P
0 g
s
.68241 .59973 m .66089 .59447 L .66148 .59261 L closepath p .436 .584 .902 r
F P
0 g
s
.68241 .59973 m .66148 .59261 L .67851 .59725 L closepath p .087 .541 .906 r
F P
0 g
s
.66546 .35207 m .67162 .36486 L .66333 .35659 L closepath p .849 .423 0 r
F P
0 g
s
.67482 .36852 m .66333 .35659 L .67162 .36486 L p .571 .074 0 r
F P
0 g
s
.66333 .35659 m .67674 .37052 L .66764 .36301 L closepath p .79 .318 0 r
F P
0 g
s
.68842 .46747 m .66346 .45281 L .66776 .43198 L closepath p .637 .447 .61 r
F P
0 g
s
.68505 .4979 m .66346 .45281 L .68842 .46747 L closepath p .635 .448 .613 r
F P
0 g
s
.66764 .36301 m .67674 .37052 L .69359 .39169 L closepath p .692 .19 0 r
F P
0 g
s
.66764 .36301 m .69359 .39169 L .68554 .38782 L closepath p .833 .362 .022 r
F P
0 g
s
.69529 .44774 m .66776 .43198 L .67275 .42402 L closepath p .681 .438 .547 r
F P
0 g
s
.68842 .46747 m .66776 .43198 L .69529 .44774 L closepath p .663 .456 .594 r
F P
0 g
s
.69344 .40268 m .67128 .38828 L .6785 .38221 L closepath p .72 .461 .533 r
F P
0 g
s
.69344 .40268 m .68471 .40538 L .67128 .38828 L closepath p .719 .462 .536 r
F P
0 g
s
.67162 .36486 m .67674 .37052 L .67482 .36852 L p .571 .074 0 r
F P
0 g
s
.69807 .44256 m .67275 .42402 L .67633 .41723 L closepath p .697 .423 .504 r
F P
0 g
s
.69529 .44774 m .67275 .42402 L .69807 .44256 L closepath p .674 .438 .555 r
F P
0 g
s
.67311 .57532 m .67619 .56364 L .69141 .56938 L closepath p .472 .385 .677 r
F P
0 g
s
.67311 .57532 m .69141 .56938 L .68917 .58283 L closepath p .486 .427 .717 r
F P
0 g
s
.67704 .59272 m .67311 .57532 L .68917 .58283 L closepath p .415 .457 .795 r
F P
0 g
s
.67619 .56364 m .67686 .5478 L .69426 .55523 L closepath p .521 .441 .707 r
F P
0 g
s
.67619 .56364 m .69426 .55523 L .69141 .56938 L closepath p .488 .368 .642 r
F P
0 g
s
.70275 .43745 m .67633 .41723 L .68471 .40538 L closepath p .688 .428 .525 r
F P
0 g
s
.69807 .44256 m .67633 .41723 L .70275 .43745 L closepath p .696 .423 .505 r
F P
0 g
s
.67686 .5478 m .68235 .52612 L .6963 .53222 L closepath p .513 .4 .663 r
F P
0 g
s
.67686 .5478 m .6963 .53222 L .69426 .55523 L closepath p .532 .432 .687 r
F P
0 g
s
.67704 .59272 m .68917 .58283 L .69306 .59676 L closepath p .441 .503 .827 r
F P
0 g
s
.67851 .59725 m .67704 .59272 L .69306 .59676 L closepath p .366 .548 .901 r
F P
0 g
s
.69809 .40573 m .69344 .40268 L .6785 .38221 L closepath p .749 .267 0 r
F P
0 g
s
.69809 .40573 m .6785 .38221 L .68554 .38782 L closepath p .779 .3 0 r
F P
0 g
s
.67851 .59725 m .69306 .59676 L .69596 .60252 L closepath p .372 .528 .882 r
F P
0 g
s
.68241 .59973 m .67851 .59725 L .69596 .60252 L closepath p .06 .523 .894 r
F P
0 g
s
.68235 .52612 m .68505 .4979 L .7023 .50569 L closepath p .521 .393 .646 r
F P
0 g
s
.68235 .52612 m .7023 .50569 L .6963 .53222 L closepath p .521 .393 .646 r
F P
0 g
s
.68241 .59973 m .69596 .60252 L .69963 .6061 L closepath p .11 .521 .911 r
F P
0 g
s
.70725 .43236 m .68471 .40538 L .69344 .40268 L closepath p .711 .444 .519 r
F P
0 g
s
.70275 .43745 m .68471 .40538 L .70725 .43236 L closepath p .737 .443 .483 r
F P
0 g
s
.68505 .4979 m .68842 .46747 L .71013 .47691 L closepath p .577 .397 .601 r
F P
0 g
s
.68505 .4979 m .71013 .47691 L .7023 .50569 L closepath p .553 .36 .574 r
F P
0 g
s
.68554 .38782 m .70744 .41137 L .69809 .40573 L closepath p .846 .384 .066 r
F P
0 g
s
.68554 .38782 m .69359 .39169 L .70744 .41137 L closepath p .834 .362 .023 r
F P
0 g
s
.68842 .46747 m .69529 .44774 L .71475 .46177 L closepath p .62 .41 .576 r
F P
0 g
s
.68842 .46747 m .71475 .46177 L .71013 .47691 L closepath p .601 .359 .522 r
F P
0 g
s
.68917 .58283 m .69141 .56938 L .69597 .57081 L p .471 .413 .711 r
F P
0 g
s
.68917 .58283 m .69597 .57081 L .69608 .57141 L .471 .413 .711 r
F
.68917 .58283 m .69608 .57141 L .69499 .57728 L F
.68917 .58283 m .69499 .57731 L .69496 .57746 L F
.68917 .58283 m .69499 .57729 L .69499 .57731 L F
.68917 .58283 m .69499 .57729 L .69499 .57729 L F
.68917 .58283 m .69499 .57728 L .69499 .57729 L F
.68917 .58283 m .69496 .57746 L .6949 .57777 L F
.68917 .58283 m .6949 .57777 L .69475 .5786 L F
.68917 .58283 m .69469 .57889 L .69465 .57912 L F
.68917 .58283 m .69473 .5787 L .69469 .57889 L F
.68917 .58283 m .69474 .57863 L .69473 .5787 L F
.68917 .58283 m .69474 .57862 L .69474 .57863 L F
.68917 .58283 m .69475 .5786 L .69474 .57862 L F
.68917 .58283 m .69461 .57933 L .6946 .57941 L F
.68917 .58283 m .69465 .57912 L .69461 .57933 L F
.68917 .58283 m .6946 .57941 L .69446 .58015 L F
.68917 .58283 m .69446 .58015 L .6944 .58046 L F
.69425 .58131 m .68917 .58283 L p .69439 .58054 L F P
0 g
s
.68917 .58283 m .6944 .58046 L .69439 .58053 L .471 .413 .711 r
F
.68917 .58283 m .69439 .58053 L .69439 .58054 L F
.68917 .58283 m .69425 .58131 L p .69385 .58343 L .475 .427 .725 r
F P
0 g
s
.69182 .5923 m .68917 .58283 L p .69134 .58472 L .323 .42 .806 r
F P
0 g
s
.68917 .58283 m .69385 .58343 L .69358 .58491 L .475 .427 .725 r
F
.68917 .58283 m .69339 .58594 L p .6933 .58643 L .323 .42 .806 r
F P
0 g
s
.69339 .58594 m .68917 .58283 L p .69358 .58491 L .475 .427 .725 r
F P
0 g
s
.69182 .5923 m .69134 .58472 L .69733 .5921 L .323 .42 .806 r
F
.69319 .587 m .69134 .58472 L .6933 .58643 L F
.69141 .56938 m .69386 .55722 L p .70103 .5604 L .5 .378 .646 r
F P
0 g
s
.69872 .56554 m .69141 .56938 L p .70103 .5604 L .5 .378 .646 r
F P
0 g
s
.69141 .56938 m .69872 .56554 L p .6983 .56648 L .498 .375 .643 r
F P
0 g
s
.69141 .56938 m .6983 .56648 L .69796 .56722 L .498 .375 .643 r
F
.69141 .56938 m .69796 .56722 L .69766 .56789 L F
.69141 .56938 m .69766 .56789 L .6976 .56804 L F
.69141 .56938 m .6976 .56804 L .69746 .56833 L F
.69141 .56938 m .69746 .56833 L .69727 .56876 L F
.69141 .56938 m .69727 .56876 L .69682 .56977 L F
.6963 .57091 m .69141 .56938 L p .69665 .57013 L F P
0 g
s
.69141 .56938 m .69682 .56977 L .69665 .57013 L .498 .375 .643 r
F
.69306 .59676 m .69182 .5923 L p .70717 .59851 L .323 .42 .806 r
F P
0 g
s
.69182 .5923 m .69847 .5935 L .70028 .59573 L .323 .42 .806 r
F
.69182 .5923 m .69735 .59212 L .69847 .5935 L F
.69182 .5923 m .69733 .5921 L .69735 .59212 L F
.71001 .59966 m .69306 .59676 L p .70717 .59851 L F P
0 g
s
.70108 .60058 m .69306 .59676 L p .71215 .60053 L .308 .44 .832 r
F P
0 g
s
.69306 .59676 m .71001 .59966 L p .7102 .59974 L .308 .44 .832 r
F P
0 g
s
.69306 .59676 m .7102 .59974 L .71215 .60053 L .308 .44 .832 r
F
.69596 .60252 m .69306 .59676 L .71472 .60707 L closepath p .245 .446 .859 r
F P
0 g
s
.69845 .58401 m .69319 .587 L .69608 .57141 L closepath p .653 .487 .647 r
F P
0 g
s
.69845 .58401 m .70028 .59573 L .69319 .587 L closepath p .707 .604 .74 r
F P
0 g
s
.70725 .43236 m .69344 .40268 L .71249 .42615 L closepath p .713 .445 .519 r
F P
0 g
s
.71249 .42615 m .69344 .40268 L .69809 .40573 L closepath p .705 .206 0 r
F P
0 g
s
.69386 .55722 m .69426 .55523 L p .70143 .55985 L .5 .378 .646 r
F P
0 g
s
.69386 .55722 m .70128 .56005 L .70103 .5604 L .5 .378 .646 r
F
.69386 .55722 m .70143 .55985 L .70128 .56005 L F
.69426 .55523 m .6963 .53222 L .71505 .53668 L closepath p .525 .427 .685 r
F P
0 g
s
.69426 .55523 m .70131 .54894 L p .70788 .55076 L .499 .386 .656 r
F P
0 g
s
.69426 .55523 m .70788 .55076 L .70675 .55235 L .499 .386 .656 r
F
.69426 .55523 m .70675 .55235 L .70612 .55324 L F
.69426 .55523 m .70612 .55324 L .70575 .55376 L F
.70434 .55574 m .69426 .55523 L p .70575 .55376 L F P
0 g
s
.69426 .55523 m .70434 .55574 L p .70232 .5586 L .5 .378 .646 r
F P
0 g
s
.70143 .55985 m .69426 .55523 L .70232 .5586 L .5 .378 .646 r
F
.69529 .44774 m .69807 .44256 L .71842 .45427 L closepath p .674 .437 .555 r
F P
0 g
s
.69529 .44774 m .71842 .45427 L .71475 .46177 L closepath p .677 .373 .451 r
F P
0 g
s
.69596 .60252 m .71472 .60707 L .71675 .61277 L closepath p .238 .451 .865 r
F P
0 g
s
.69963 .6061 m .69596 .60252 L .71675 .61277 L closepath p 0 .391 .829 r
F P
0 g
s
.69608 .57141 m .69597 .57081 L .69619 .57116 L .471 .413 .711 r
F
.69597 .57081 m .6963 .57091 L p .69619 .57116 L F P
0 g
s
.69608 .57141 m .70103 .5604 L .70555 .56601 L closepath p .907 .676 .578 r
F P
0 g
s
.69608 .57141 m .70555 .56601 L .70201 .57457 L closepath p .892 .64 .549 r
F P
0 g
s
.69608 .57141 m .70201 .57457 L .70071 .57755 L closepath p .896 .609 .493 r
F P
0 g
s
.70071 .57755 m .69845 .58401 L .69608 .57141 L closepath p .861 .623 .578 r
F P
0 g
s
.6963 .53222 m .7023 .50569 L .72181 .51242 L closepath p .539 .41 .653 r
F P
0 g
s
.6963 .53222 m .72181 .51242 L .71505 .53668 L closepath p .535 .403 .647 r
F P
0 g
s
.69807 .44256 m .70275 .43745 L .71842 .45427 L closepath p .691 .415 .5 r
F P
0 g
s
.71865 .43064 m .69809 .40573 L .70744 .41137 L closepath p .795 .3 0 r
F P
0 g
s
.71249 .42615 m .69809 .40573 L .71865 .43064 L closepath p .72 .215 0 r
F P
0 g
s
.70071 .57755 m .70231 .5841 L .69845 .58401 L closepath p .792 .541 .559 r
F P
0 g
s
.69845 .58401 m .7051 .59177 L .70028 .59573 L closepath p .688 .592 .742 r
F P
0 g
s
.7051 .59177 m .69845 .58401 L .70231 .5841 L closepath p .8 .598 .627 r
F P
0 g
s
.69963 .6061 m .71675 .61277 L .72159 .6188 L closepath p 0 .394 .846 r
F P
0 g
s
.7051 .59177 m .71243 .60064 L .70028 .59573 L closepath p .759 .736 .831 r
F P
0 g
s
.70201 .57457 m .70455 .58352 L .70071 .57755 L closepath p .881 .588 .487 r
F P
0 g
s
.70455 .58352 m .70231 .5841 L .70071 .57755 L closepath p .874 .59 .503 r
F P
0 g
s
.70103 .5604 m .71036 .54726 L .71085 .56104 L closepath p .916 .718 .62 r
F P
0 g
s
.70103 .5604 m .71085 .56104 L .70555 .56601 L closepath p .912 .673 .565 r
F P
0 g
s
.71226 .60057 m .71229 .60059 L .70108 .60058 L .308 .44 .832 r
F
.71273 .60064 m .71472 .60707 L .70108 .60058 L p F P
0 g
s
.71229 .60059 m .71243 .60064 L .70108 .60058 L .308 .44 .832 r
F
.71215 .60053 m .71226 .60057 L .70108 .60058 L F
.70131 .54894 m .70987 .5413 L p .71224 .54691 L .499 .386 .656 r
F P
0 g
s
.70131 .54894 m .71036 .54726 L .70788 .55076 L .499 .386 .656 r
F
.70555 .56601 m .7075 .58349 L .70201 .57457 L closepath p .902 .652 .552 r
F P
0 g
s
.70201 .57457 m .7075 .58349 L .70455 .58352 L closepath p .945 .616 .377 r
F P
0 g
s
.7023 .50569 m .71013 .47691 L .72735 .48993 L closepath p .56 .367 .576 r
F P
0 g
s
.7023 .50569 m .72735 .48993 L .72181 .51242 L closepath p .562 .37 .579 r
F P
0 g
s
.70231 .5841 m .70455 .58352 L .71038 .58583 L closepath p .89 .635 .547 r
F P
0 g
s
.70231 .5841 m .71038 .58583 L .70597 .5878 L closepath p .89 .643 .558 r
F P
0 g
s
.7051 .59177 m .70231 .5841 L .70597 .5878 L closepath p .896 .636 .537 r
F P
0 g
s
.70275 .43745 m .70725 .43236 L .72296 .44798 L closepath p .704 .398 .456 r
F P
0 g
s
.70275 .43745 m .72296 .44798 L .71842 .45427 L closepath p .698 .414 .489 r
F P
0 g
s
.70455 .58352 m .7075 .58349 L .71038 .58583 L closepath p .767 .267 .022 r
F P
0 g
s
.70597 .5878 m .71438 .59435 L .7051 .59177 L closepath p .88 .617 .537 r
F P
0 g
s
.7051 .59177 m .71438 .59435 L .71243 .60064 L closepath p .859 .744 .735 r
F P
0 g
s
.71085 .56104 m .70798 .57472 L .70555 .56601 L closepath p .919 .688 .573 r
F P
0 g
s
.70555 .56601 m .70798 .57472 L .7075 .58349 L closepath p .97 .702 .468 r
F P
0 g
s
.70597 .5878 m .71038 .58583 L .71725 .58756 L closepath p .907 .691 .599 r
F P
0 g
s
.70597 .5878 m .71725 .58756 L .7141 .59008 L closepath p .823 .429 .305 r
F P
0 g
s
.70597 .5878 m .7141 .59008 L .71438 .59435 L closepath p .856 .65 .624 r
F P
0 g
s
.70725 .43236 m .71249 .42615 L .73018 .44459 L closepath p .676 .395 .488 r
F P
0 g
s
.70725 .43236 m .73018 .44459 L .72296 .44798 L closepath p .67 .406 .513 r
F P
0 g
s
.71865 .43064 m .70744 .41137 L .72561 .43419 L closepath p .793 .297 0 r
F P
0 g
s
.71142 .58088 m .7075 .58349 L .70798 .57472 L closepath p .917 .644 .507 r
F P
0 g
s
.71142 .58088 m .71038 .58583 L .7075 .58349 L closepath p .788 .405 .33 r
F P
0 g
s
.71085 .56104 m .71068 .57161 L .70798 .57472 L closepath p .999 .824 .494 r
F P
0 g
s
.71068 .57161 m .71142 .58088 L .70798 .57472 L closepath p .946 .621 .382 r
F P
0 g
s
.70987 .5413 m .71505 .53668 L .71569 .54737 L p .499 .386 .656 r
F P
0 g
s
.70987 .5413 m .71569 .54737 L .71346 .54669 L .499 .386 .656 r
F
.70987 .5413 m .71346 .54669 L .71224 .54691 L F
.71013 .47691 m .71475 .46177 L .73269 .47058 L closepath p .599 .357 .521 r
F P
0 g
s
.71013 .47691 m .73269 .47058 L .72735 .48993 L closepath p .598 .353 .518 r
F P
0 g
s
.71036 .54726 m .71346 .54669 L .71567 .55775 L closepath p .977 .695 .304 r
F P
0 g
s
.71036 .54726 m .71567 .55775 L .71085 .56104 L closepath p .995 .78 .502 r
F P
0 g
s
.71142 .58088 m .71725 .58756 L .71038 .58583 L closepath p .874 .505 .342 r
F P
0 g
s
.71567 .55775 m .71068 .57161 L .71085 .56104 L closepath p .999 .822 .535 r
F P
0 g
s
.71567 .55775 m .71486 .56892 L .71068 .57161 L closepath p .999 .824 .534 r
F P
0 g
s
.71486 .56892 m .71493 .57851 L .71068 .57161 L closepath p .93 .581 .346 r
F P
0 g
s
.71068 .57161 m .71493 .57851 L .71142 .58088 L closepath p .921 .597 .411 r
F P
0 g
s
.71493 .57851 m .71725 .58756 L .71142 .58088 L closepath p .874 .505 .343 r
F P
0 g
s
.72299 .6007 m .71243 .60064 L .71438 .59435 L closepath p .874 .762 .735 r
F P
0 g
s
.71249 .42615 m .71865 .43064 L .73528 .44757 L closepath p .627 .101 0 r
F P
0 g
s
.71249 .42615 m .73528 .44757 L .73018 .44459 L closepath p .643 .117 0 r
F P
0 g
s
.71293 .60064 m .7221 .61411 L p .71273 .60064 L .21 .368 .804 r
F P
0 g
s
.7221 .61411 m .71472 .60707 L .71273 .60064 L p .21 .368 .804 r
F P
0 g
s
.71333 .60065 m .72039 .60569 L p .71293 .60064 L .084 .299 .781 r
F P
0 g
s
.72039 .60569 m .7221 .61411 L .71293 .60064 L p .084 .299 .781 r
F P
0 g
s
.72039 .60569 m .71333 .60065 L p .71751 .60067 L .236 .312 .739 r
F P
0 g
s
.71346 .54669 m .72286 .5608 L .71567 .55775 L closepath p .941 .698 .543 r
F P
0 g
s
.71346 .54669 m .72208 .54935 L .72286 .5608 L closepath p .947 .692 .519 r
F P
0 g
s
.72025 .59277 m .7141 .59008 L .71725 .58756 L closepath p .926 .614 .43 r
F P
0 g
s
.72025 .59277 m .71438 .59435 L .7141 .59008 L closepath p .931 .708 .579 r
F P
0 g
s
.72025 .59277 m .72269 .59417 L .71438 .59435 L closepath p .675 .193 .099 r
F P
0 g
s
.72299 .6007 m .71438 .59435 L .72269 .59417 L closepath p .865 .767 .753 r
F P
0 g
s
.7221 .61411 m .71675 .61277 L .71472 .60707 L closepath p .065 .316 .799 r
F P
0 g
s
.71475 .46177 m .71842 .45427 L .73927 .46133 L closepath p .625 .314 .424 r
F P
0 g
s
.71475 .46177 m .73927 .46133 L .73269 .47058 L closepath p .624 .311 .42 r
F P
0 g
s
.72286 .5608 m .71486 .56892 L .71567 .55775 L closepath p .933 .755 .638 r
F P
0 g
s
.72286 .5608 m .72079 .57026 L .71486 .56892 L closepath p .952 .809 .666 r
F P
0 g
s
.71486 .56892 m .71993 .57767 L .71493 .57851 L closepath p .946 .601 .33 r
F P
0 g
s
.72079 .57026 m .71993 .57767 L .71486 .56892 L closepath p .94 .624 .411 r
F P
0 g
s
.71993 .57767 m .72355 .58811 L .71493 .57851 L closepath p .894 .486 .22 r
F P
0 g
s
.71493 .57851 m .72355 .58811 L .71725 .58756 L closepath p .901 .523 .304 r
F P
0 g
s
.72975 .5104 m .71505 .53668 L .72181 .51242 L closepath p .571 .439 .662 r
F P
0 g
s
.72975 .5104 m .72918 .53601 L .71505 .53668 L closepath p .521 .379 .629 r
F P
0 g
s
.71505 .53668 m .72918 .53601 L .72091 .54899 L p .524 .405 .659 r
F P
0 g
s
.72091 .54899 m .71701 .54778 L .71505 .53668 L .524 .405 .659 r
F
.71569 .54737 m .71505 .53668 L p .71701 .54778 L F P
0 g
s
.7221 .61411 m .72159 .6188 L .71675 .61277 L closepath p 0 .39 .843 r
F P
0 g
s
.72355 .58811 m .72025 .59277 L .71725 .58756 L closepath p .932 .608 .401 r
F P
0 g
s
.71807 .60067 m .72039 .60569 L p .71751 .60067 L .236 .312 .739 r
F P
0 g
s
.71857 .60068 m .72039 .60569 L .71807 .60067 L p 0 0 .415 r
F P
0 g
s
.71842 .45427 m .72296 .44798 L .74471 .45975 L closepath p .698 .413 .489 r
F P
0 g
s
.71842 .45427 m .74471 .45975 L .73927 .46133 L closepath p .557 .025 0 r
F P
0 g
s
.71998 .60068 m .72797 .60962 L p .71857 .60068 L .213 .465 .881 r
F P
0 g
s
.72797 .60962 m .72039 .60569 L .71857 .60068 L p .213 .465 .881 r
F P
0 g
s
.74473 .45345 m .73528 .44757 L .71865 .43064 L closepath p .706 .177 0 r
F P
0 g
s
.74473 .45345 m .71865 .43064 L .72561 .43419 L closepath p .719 .192 0 r
F P
0 g
s
.72079 .57026 m .72377 .57543 L .71993 .57767 L closepath p .93 .611 .415 r
F P
0 g
s
.72377 .57543 m .72773 .58476 L .71993 .57767 L closepath p .815 .397 .251 r
F P
0 g
s
.71993 .57767 m .72773 .58476 L .72355 .58811 L closepath p .804 .44 .367 r
F P
0 g
s
.72224 .6007 m .72797 .60962 L .71998 .60068 L p .031 .361 .833 r
F P
0 g
s
.72355 .58811 m .72881 .59069 L .72025 .59277 L closepath p .623 .184 .186 r
F P
0 g
s
.72881 .59069 m .72269 .59417 L .72025 .59277 L closepath p .681 .28 .28 r
F P
0 g
s
.7221 .61411 m .72039 .60569 L .72797 .60962 L closepath p .287 .47 .866 r
F P
0 g
s
.72286 .5608 m .72738 .57558 L .72079 .57026 L closepath p .903 .752 .684 r
F P
0 g
s
.72738 .57558 m .72377 .57543 L .72079 .57026 L closepath p .938 .595 .349 r
F P
0 g
s
.72091 .54899 m .72132 .54833 L p .72208 .54935 L .491 .364 .635 r
F P
0 g
s
.72132 .54833 m .72918 .53601 L p .72509 .5534 L .491 .364 .635 r
F P
0 g
s
.7221 .61411 m .72765 .62279 L .72159 .6188 L closepath p 0 .351 .821 r
F P
0 g
s
.73553 .48597 m .72975 .5104 L .72181 .51242 L closepath p .541 .352 .573 r
F P
0 g
s
.73553 .48597 m .72181 .51242 L .72735 .48993 L closepath p .535 .345 .569 r
F P
0 g
s
.72208 .54935 m .7361 .56823 L .72286 .5608 L closepath p .843 .611 .586 r
F P
0 g
s
.7221 .61411 m .72797 .60962 L .73293 .61626 L p .265 .429 .838 r
F P
0 g
s
.72765 .62279 m .7221 .61411 L .72977 .61922 L p .184 .537 .925 r
F P
0 g
s
.7221 .61411 m .73289 .61764 L .73287 .61821 L .084 .487 .896 r
F
.7221 .61411 m .7329 .61704 L p .73289 .61764 L F P
0 g
s
.7221 .61411 m .73287 .61821 L .73283 .61944 L .084 .487 .896 r
F
.73277 .62121 m .7221 .61411 L p .73278 .6209 L F P
0 g
s
.7221 .61411 m .73279 .62054 L .73279 .62056 L .084 .487 .896 r
F
.7221 .61411 m .7328 .62033 L .73279 .62054 L F
.7221 .61411 m .73279 .62056 L .73278 .6209 L F
.7221 .61411 m .73281 .62015 L .73281 .62017 L F
.7221 .61411 m .73281 .62017 L .7328 .62033 L F
.7221 .61411 m .73281 .61996 L .73281 .62015 L F
.7221 .61411 m .73282 .61986 L .73281 .61996 L F
.7221 .61411 m .73282 .61984 L .73282 .61986 L F
.7221 .61411 m .73283 .61944 L .73282 .61984 L F
.7329 .61704 m .7221 .61411 L p .73293 .61626 L .265 .429 .838 r
F P
0 g
s
.72238 .6007 m .73228 .60947 L p .72224 .6007 L .371 .591 .93 r
F P
0 g
s
.73228 .60947 m .72797 .60962 L .72224 .6007 L p .371 .591 .93 r
F P
0 g
s
.73228 .60947 m .72238 .6007 L p .72293 .6007 L .552 .655 .904 r
F P
0 g
s
.72881 .59069 m .73133 .59292 L .72269 .59417 L closepath p .743 .376 .357 r
F P
0 g
s
.73133 .59292 m .72299 .6007 L .72269 .59417 L closepath p .848 .755 .76 r
F P
0 g
s
.72286 .5608 m .73873 .57181 L .72738 .57558 L closepath p .661 .626 .804 r
F P
0 g
s
.7361 .56823 m .73873 .57181 L .72286 .5608 L closepath p .802 .329 .1 r
F P
0 g
s
.72299 .6007 m .73233 .60759 L .72293 .6007 L .552 .655 .904 r
F
.73233 .60759 m .73228 .60947 L p .72293 .6007 L F P
0 g
s
.72296 .44798 m .73018 .44459 L .74471 .45975 L closepath p .685 .435 .539 r
F P
0 g
s
.72348 .60062 m .73249 .60199 L p .72299 .6007 L .552 .655 .904 r
F P
0 g
s
.73249 .60199 m .73233 .60759 L p .72299 .6007 L .552 .655 .904 r
F P
0 g
s
.73133 .59292 m .72791 .59994 L .72299 .6007 L closepath p .803 .663 .708 r
F P
0 g
s
.72729 .6012 m .72348 .60062 L p .72791 .59994 L .467 .695 .967 r
F P
0 g
s
.72773 .58476 m .72881 .59069 L .72355 .58811 L closepath p .623 .186 .189 r
F P
0 g
s
.72738 .57558 m .72911 .58112 L .72377 .57543 L closepath p .848 .4 .142 r
F P
0 g
s
.72377 .57543 m .72911 .58112 L .72773 .58476 L closepath p .852 .413 .172 r
F P
0 g
s
.72509 .5534 m .73061 .55156 L p .73048 .55954 L .529 .475 .74 r
F P
0 g
s
.73048 .55954 m .73046 .56057 L .72509 .5534 L .529 .475 .74 r
F
.73046 .56057 m .73046 .5606 L .72509 .5534 L F
.73046 .5606 m .73046 .56064 L .72509 .5534 L F
.72918 .53601 m .73061 .55156 L .72509 .5534 L p .491 .364 .635 r
F P
0 g
s
.73249 .60199 m .72729 .6012 L p .72998 .59956 L .467 .695 .967 r
F P
0 g
s
.72791 .59994 m .72998 .59956 L .72729 .6012 L .467 .695 .967 r
F
.74499 .47337 m .72735 .48993 L .73269 .47058 L closepath p .579 .335 .511 r
F P
0 g
s
.74499 .47337 m .73553 .48597 L .72735 .48993 L closepath p .38 .072 .311 r
F P
0 g
s
.73873 .57181 m .73912 .57657 L .72738 .57558 L closepath p .573 .363 .558 r
F P
0 g
s
.72738 .57558 m .7369 .58091 L .72911 .58112 L closepath p .519 .194 .357 r
F P
0 g
s
.72738 .57558 m .73912 .57657 L .7369 .58091 L closepath p .535 .169 .291 r
F P
0 g
s
.72977 .61922 m .73385 .6232 L .72765 .62279 L .184 .537 .925 r
F
.7354 .62567 m .72765 .62279 L p .73532 .62555 L F P
0 g
s
.73489 .62486 m .73532 .62555 L .72765 .62279 L .184 .537 .925 r
F
.73471 .62457 m .73489 .62486 L .72765 .62279 L F
.73385 .6232 m .73471 .62457 L .72765 .62279 L F
.72911 .58112 m .73513 .58593 L .72773 .58476 L closepath p .57 .083 .061 r
F P
0 g
s
.72773 .58476 m .73513 .58593 L .72881 .59069 L closepath p .612 .178 .192 r
F P
0 g
s
.74168 .59017 m .73816 .59805 L .72791 .59994 L closepath p .895 .926 .832 r
F P
0 g
s
.74168 .59017 m .72791 .59994 L .73133 .59292 L closepath p .868 .75 .731 r
F P
0 g
s
.72797 .60962 m .73228 .60947 L .73308 .61392 L p .402 .495 .839 r
F P
0 g
s
.72797 .60962 m .73306 .61396 L .73303 .61401 L .402 .495 .839 r
F
.72797 .60962 m .73308 .61392 L .73308 .61392 L F
.72797 .60962 m .73308 .61392 L .73306 .61396 L F
.73118 .61392 m .72797 .60962 L p .73303 .61401 L F P
0 g
s
.72881 .59069 m .73513 .58593 L .74168 .59017 L closepath p .702 .298 .28 r
F P
0 g
s
.72881 .59069 m .74168 .59017 L .73133 .59292 L closepath p .739 .38 .373 r
F P
0 g
s
.72911 .58112 m .7369 .58091 L .73513 .58593 L closepath p .495 .13 .275 r
F P
0 g
s
.72975 .5104 m .7318 .53804 L .72918 .53601 L closepath p .094 .035 .492 r
F P
0 g
s
.72918 .53601 m .73496 .551 L .73061 .55156 L closepath p .304 .225 .6 r
F P
0 g
s
.72918 .53601 m .7318 .53804 L .73496 .551 L closepath p 0 .054 .567 r
F P
0 g
s
.73553 .48597 m .73481 .51333 L .72975 .5104 L closepath p .281 .11 .463 r
F P
0 g
s
.72975 .5104 m .73481 .51333 L .7318 .53804 L closepath p .232 .135 .534 r
F P
0 g
s
.72977 .61922 m .73247 .62101 L p .73385 .6232 L .184 .537 .925 r
F P
0 g
s
.72998 .59956 m .73053 .59945 L .73249 .60199 L .467 .695 .967 r
F
.74471 .45975 m .73018 .44459 L .74676 .46013 L p .66 .136 0 r
F P
0 g
s
.73018 .44459 m .73528 .44757 L .74742 .46026 L p .65 .126 0 r
F P
0 g
s
.73018 .44459 m .74742 .46026 L .74702 .46018 L .65 .126 0 r
F
.74676 .46013 m .73018 .44459 L p .74702 .46018 L F P
0 g
s
.73046 .56064 m .73061 .55156 L p .73105 .56143 L .279 .106 .459 r
F P
0 g
s
.73223 .59914 m .73249 .60199 L p .73053 .59945 L .467 .695 .967 r
F P
0 g
s
.73061 .55156 m .73496 .551 L .73105 .56143 L p .279 .106 .459 r
F P
0 g
s
.73105 .56143 m .73496 .551 L p .73147 .562 L .537 .369 .602 r
F P
0 g
s
.73302 .61403 m .733 .61407 L .73118 .61392 L .402 .495 .839 r
F
.73293 .61626 m .73118 .61392 L p .733 .61407 L F P
0 g
s
.73303 .61401 m .73302 .61403 L .73118 .61392 L .402 .495 .839 r
F
.73303 .61401 m .73303 .61401 L .73118 .61392 L F
.73496 .551 m .74232 .55069 L .73147 .562 L p .537 .369 .602 r
F P
0 g
s
.73147 .562 m .74232 .55069 L p .73235 .56318 L .658 .55 .721 r
F P
0 g
s
.73481 .51333 m .74014 .5409 L .7318 .53804 L closepath p .547 .412 .648 r
F P
0 g
s
.7318 .53804 m .74232 .55069 L .73496 .551 L closepath p .543 .416 .658 r
F P
0 g
s
.7318 .53804 m .74014 .5409 L .74232 .55069 L closepath p .544 .416 .657 r
F P
0 g
s
.73502 .6039 m .73249 .60199 L .73223 .59914 L p 0 .262 .767 r
F P
0 g
s
.73229 .59913 m .73502 .6039 L .73223 .59914 L 0 .262 .767 r
F
.73228 .60947 m .73249 .60199 L .73502 .6039 L closepath p .148 .285 .749 r
F P
0 g
s
.73228 .60947 m .73502 .6039 L .73398 .61177 L closepath p .141 .276 .743 r
F P
0 g
s
.73403 .61201 m .73228 .60947 L .73398 .61177 L p 0 0 .481 r
F P
0 g
s
.73228 .60947 m .73403 .61201 L p .73402 .61211 L .618 .655 .864 r
F P
0 g
s
.73228 .60947 m .73402 .61211 L .73396 .61223 L .618 .655 .864 r
F
.73228 .60947 m .73396 .61223 L .7337 .61273 L F
.73228 .60947 m .7337 .61273 L .73352 .61307 L F
.73308 .61392 m .73228 .60947 L p .73352 .61307 L F P
0 g
s
.73258 .59908 m .73502 .6039 L p .73229 .59913 L 0 .262 .767 r
F P
0 g
s
.74232 .55069 m .74382 .56371 L .73235 .56318 L p .658 .55 .721 r
F P
0 g
s
.73235 .56318 m .74382 .56371 L .7361 .56823 L p .385 0 0 r
F P
0 g
s
.73247 .62101 m .73277 .62139 L .73276 .62148 L .184 .537 .925 r
F
.73247 .62101 m .73277 .62123 L .73277 .62128 L F
.73247 .62101 m .73277 .62121 L p .73277 .62123 L F P
0 g
s
.73247 .62101 m .73277 .62123 L .73277 .62123 L .184 .537 .925 r
F
.73247 .62101 m .73277 .62128 L .73277 .62139 L F
.74106 .59889 m .73502 .6039 L .73258 .59908 L p .554 .621 .874 r
F P
0 g
s
.73328 .59895 m .74106 .59889 L .73258 .59908 L .554 .621 .874 r
F
.75763 .46404 m .73269 .47058 L .73927 .46133 L closepath p .582 .262 .391 r
F P
0 g
s
.75763 .46404 m .74499 .47337 L .73269 .47058 L closepath p .579 .254 .383 r
F P
0 g
s
.73572 .61704 m .73276 .62148 L .733 .61407 L closepath p 0 0 .238 r
F P
0 g
s
.73572 .61704 m .73543 .62475 L .73276 .62148 L closepath p 0 0 .272 r
F P
0 g
s
.73543 .62475 m .73745 .62893 L .73276 .62148 L closepath p 0 0 .437 r
F P
0 g
s
.73689 .6109 m .733 .61407 L .73649 .60738 L closepath p 0 0 0 r
F P
0 g
s
.73689 .6109 m .73572 .61704 L .733 .61407 L closepath p 0 0 .194 r
F P
0 g
s
.73574 .59849 m .74106 .59889 L p .73328 .59895 L .554 .621 .874 r
F P
0 g
s
.73398 .61177 m .73502 .6039 L .73431 .61155 L p 0 0 .034 r
F P
0 g
s
.73409 .61199 m .73398 .61177 L p .73431 .61155 L 0 0 .034 r
F P
0 g
s
.73398 .61177 m .73409 .61199 L p .73404 .61207 L 0 0 .174 r
F P
0 g
s
.73404 .61207 m .73398 .61177 L p .73404 .61207 L 0 0 .174 r
F P
0 g
s
.73403 .61201 m .73402 .61211 L .73402 .61211 L .618 .655 .864 r
F
.73403 .61201 m .73402 .61211 L .73402 .61211 L F
.73403 .61201 m .73402 .61211 L .73402 .61211 L F
.73403 .61201 m .73402 .61211 L .73402 .61211 L F
.73403 .61201 m .73403 .61201 L .73402 .61211 L F
.73403 .61201 m .73403 .61209 L .73402 .61211 L F
.73403 .61201 m .73404 .61207 L .73404 .61208 L F
.73403 .61201 m .73404 .61208 L .73403 .61209 L F
.73431 .61155 m .73443 .6103 L p .73649 .60738 L 0 0 .502 r
F P
0 g
s
.73443 .6103 m .73502 .6039 L .73846 .6046 L p 0 0 .502 r
F P
0 g
s
.73553 .48597 m .74518 .48931 L .73481 .51333 L closepath p .465 .254 .507 r
F P
0 g
s
.74518 .48931 m .73963 .51446 L .73481 .51333 L closepath p .402 .188 .471 r
F P
0 g
s
.73481 .51333 m .73963 .51446 L .74014 .5409 L closepath p .368 .291 .637 r
F P
0 g
s
.7503 .59528 m .73846 .6046 L .73502 .6039 L closepath p .075 0 0 r
F P
0 g
s
.74106 .59889 m .7503 .59528 L .73502 .6039 L closepath p .67 .958 .944 r
F P
0 g
s
.73513 .58593 m .7369 .58091 L .74561 .58346 L closepath p .649 .281 .333 r
F P
0 g
s
.73513 .58593 m .74561 .58346 L .74168 .59017 L closepath p .683 .349 .402 r
F P
0 g
s
.7502 .46079 m .73528 .44757 L p .75055 .46085 L .723 .199 0 r
F P
0 g
s
.75055 .46085 m .73528 .44757 L .75069 .46088 L .723 .199 0 r
F
.75084 .46091 m .73528 .44757 L .75112 .46096 L F
.75112 .46096 m .73528 .44757 L .75162 .46105 L F
.73528 .44757 m .74473 .45345 L .7519 .46111 L p F P
0 g
s
.75162 .46105 m .73528 .44757 L .7519 .46111 L .723 .199 0 r
F
.75069 .46088 m .73528 .44757 L .75084 .46091 L F
.74742 .46026 m .73528 .44757 L .7502 .46079 L p .716 .19 0 r
F P
0 g
s
.74114 .61766 m .73543 .62475 L .73572 .61704 L closepath p .184 .114 .538 r
F P
0 g
s
.74114 .61766 m .74086 .62427 L .73543 .62475 L closepath p .104 .005 .447 r
F P
0 g
s
.74086 .62427 m .74071 .62991 L .73543 .62475 L closepath p .133 .422 .865 r
F P
0 g
s
.73543 .62475 m .74071 .62991 L .73745 .62893 L closepath p 0 .182 .691 r
F P
0 g
s
.74499 .47337 m .75039 .47666 L .73553 .48597 L closepath p .486 .193 .392 r
F P
0 g
s
.75039 .47666 m .74518 .48931 L .73553 .48597 L closepath p .487 .193 .393 r
F P
0 g
s
.74191 .61169 m .73572 .61704 L .73689 .6109 L closepath p .205 .096 .501 r
F P
0 g
s
.74191 .61169 m .74114 .61766 L .73572 .61704 L closepath p .19 .074 .481 r
F P
0 g
s
.73789 .59865 m .73574 .59849 L p .73816 .59805 L .544 .635 .891 r
F P
0 g
s
.7361 .56823 m .74382 .56371 L .74775 .56975 L closepath p .858 .438 .225 r
F P
0 g
s
.7361 .56823 m .74775 .56975 L .74535 .57193 L closepath p .436 0 0 r
F P
0 g
s
.7361 .56823 m .74535 .57193 L .73873 .57181 L closepath p .593 .339 .502 r
F P
0 g
s
.73846 .6046 m .7468 .60336 L .73649 .60738 L closepath p .2 0 .004 r
F P
0 g
s
.73649 .60738 m .7468 .60336 L .74775 .60481 L closepath p 0 0 0 r
F P
0 g
s
.73649 .60738 m .74775 .60481 L .73689 .6109 L closepath p .046 0 0 r
F P
0 g
s
.73689 .6109 m .74775 .60481 L .75088 .60681 L closepath p .297 0 .162 r
F P
0 g
s
.73689 .6109 m .75088 .60681 L .74191 .61169 L closepath p .227 0 .022 r
F P
0 g
s
.73912 .57657 m .74561 .58346 L .7369 .58091 L closepath p .65 .292 .351 r
F P
0 g
s
.73816 .59805 m .73945 .596 L .73789 .59865 L .544 .635 .891 r
F
.73945 .596 m .74106 .59889 L .73789 .59865 L p F P
0 g
s
.73922 .59607 m .73945 .596 L p .73816 .59805 L .544 .635 .891 r
F P
0 g
s
.73816 .59805 m .74265 .58891 L .73952 .59549 L closepath p .678 .791 .325 r
F P
0 g
s
.74168 .59017 m .74265 .58891 L .73816 .59805 L closepath p .675 .79 .324 r
F P
0 g
s
.7503 .59528 m .75009 .60039 L .73846 .6046 L closepath p .534 .22 .381 r
F P
0 g
s
.73846 .6046 m .75009 .60039 L .7468 .60336 L closepath p .095 0 0 r
F P
0 g
s
.73873 .57181 m .74535 .57193 L .74449 .57497 L closepath p .581 .288 .436 r
F P
0 g
s
.73873 .57181 m .74449 .57497 L .73912 .57657 L closepath p .535 .335 .556 r
F P
0 g
s
.73912 .57657 m .74449 .57497 L .74561 .58346 L closepath p .519 .297 .517 r
F P
0 g
s
.73945 .596 m .73922 .59607 L p .73924 .59603 L .514 .823 1 r
F P
0 g
s
.73927 .59598 m .73945 .596 L p .73924 .59603 L .514 .823 1 r
F P
0 g
s
.73927 .46133 m .74471 .45975 L .76469 .46353 L p .45 0 0 r
F P
0 g
s
.75763 .46404 m .73927 .46133 L .74928 .46223 L p .303 0 0 r
F P
0 g
s
.76423 .46356 m .73927 .46133 L p .76469 .46353 L .45 0 0 r
F P
0 g
s
.7394 .59573 m .73945 .596 L .73927 .59598 L p .546 .827 .998 r
F P
0 g
s
.73947 .59559 m .7503 .59528 L .7394 .59573 L .636 .888 .986 r
F
.7503 .59528 m .73945 .596 L .7394 .59573 L p F P
0 g
s
.73945 .596 m .7503 .59528 L .74106 .59889 L closepath p .789 .745 .812 r
F P
0 g
s
.73952 .59549 m .73948 .59559 L .73947 .59559 L .636 .888 .986 r
F
.73999 .59451 m .7503 .59528 L p .73948 .59559 L F P
0 g
s
.74518 .48931 m .74987 .49123 L .73963 .51446 L closepath p .209 .018 .388 r
F P
0 g
s
.74987 .49123 m .7463 .52073 L .73963 .51446 L closepath p .554 .379 .599 r
F P
0 g
s
.73963 .51446 m .7463 .52073 L .74613 .54095 L closepath p .503 .379 .643 r
F P
0 g
s
.73963 .51446 m .74613 .54095 L .74014 .5409 L closepath p .373 .295 .639 r
F P
0 g
s
.74265 .58891 m .7503 .59528 L .73999 .59451 L p .835 .794 .813 r
F P
0 g
s
.74014 .5409 m .74677 .55211 L .74232 .55069 L closepath p .398 .32 .652 r
F P
0 g
s
.74014 .5409 m .74613 .54095 L .74677 .55211 L closepath p .373 .31 .657 r
F P
0 g
s
.74086 .62427 m .74492 .63002 L .74071 .62991 L closepath p .104 .391 .848 r
F P
0 g
s
.74889 .61826 m .74086 .62427 L .74114 .61766 L closepath p .352 .202 .534 r
F P
0 g
s
.74889 .61826 m .74935 .62508 L .74086 .62427 L closepath p .425 .322 .634 r
F P
0 g
s
.74935 .62508 m .74492 .63002 L .74086 .62427 L closepath p .341 .536 .899 r
F P
0 g
s
.75267 .61131 m .74114 .61766 L .74191 .61169 L closepath p .452 .3 .583 r
F P
0 g
s
.75267 .61131 m .74889 .61826 L .74114 .61766 L closepath p .35 .118 .414 r
F P
0 g
s
.74168 .59017 m .74561 .58346 L .74265 .58891 L closepath p .632 .803 .371 r
F P
0 g
s
.75267 .61131 m .74191 .61169 L .75088 .60681 L closepath p .383 .024 .226 r
F P
0 g
s
.74677 .55211 m .74382 .56371 L .74232 .55069 L closepath p .38 .343 .692 r
F P
0 g
s
.7503 .59528 m .74265 .58891 L p .74292 .58841 L .951 .67 .468 r
F P
0 g
s
.7447 .63137 m .74291 .62997 L p .74307 .62997 L .308 .494 .878 r
F P
0 g
s
.74456 .58539 m .7503 .59528 L .74292 .58841 L .951 .67 .468 r
F
.74361 .62999 m .7447 .63137 L p .74307 .62997 L .308 .494 .878 r
F P
0 g
s
.7447 .63137 m .74361 .62999 L p .74445 .63001 L .175 .422 .858 r
F P
0 g
s
.74677 .55211 m .74853 .56246 L .74382 .56371 L closepath p .398 .361 .701 r
F P
0 g
s
.74382 .56371 m .74853 .56246 L .74775 .56975 L closepath p .392 .343 .683 r
F P
0 g
s
.74448 .63001 m .7447 .63137 L p .74445 .63001 L .175 .422 .858 r
F P
0 g
s
.74455 .63043 m .74448 .63001 L p .74492 .63002 L .574 .753 .962 r
F P
0 g
s
.74449 .57497 m .74775 .56975 L .745 .57885 L p .823 .805 .315 r
F P
0 g
s
.74775 .56975 m .74449 .57497 L .74535 .57193 L closepath p .84 .799 .306 r
F P
0 g
s
.7565 .63359 m .74455 .63043 L .74492 .63002 L .574 .753 .962 r
F
.7565 .63359 m .7447 .63137 L .74455 .63043 L p F P
0 g
s
.74534 .58395 m .7503 .59528 L p .74456 .58539 L .951 .67 .468 r
F P
0 g
s
.76184 .63758 m .75249 .63552 L .7447 .63137 L closepath p .042 .56 .871 r
F P
0 g
s
.76184 .63758 m .7447 .63137 L .7565 .63359 L closepath p .428 .753 .995 r
F P
0 g
s
.74935 .62508 m .75213 .62915 L .74492 .63002 L closepath p .376 .59 .928 r
F P
0 g
s
.75217 .62978 m .74492 .63002 L .75213 .62915 L closepath p .161 .698 .827 r
F P
0 g
s
.7565 .63359 m .74492 .63002 L .74657 .62997 L .574 .753 .962 r
F
.75039 .47666 m .74499 .47337 L .75763 .46404 L closepath p .515 .166 .313 r
F P
0 g
s
.745 .57885 m .74775 .56975 L p .74524 .58063 L .791 .526 .537 r
F P
0 g
s
.75039 .47666 m .75222 .47898 L .74518 .48931 L closepath p .409 .117 .354 r
F P
0 g
s
.75222 .47898 m .74987 .49123 L .74518 .48931 L closepath p .257 0 .243 r
F P
0 g
s
.74775 .56975 m .75018 .57684 L .74524 .58063 L p .791 .526 .537 r
F P
0 g
s
.74524 .58063 m .75018 .57684 L p .7454 .58187 L .829 .595 .584 r
F P
0 g
s
.74535 .58393 m .75112 .58419 L .74534 .58395 L .812 .644 .675 r
F
.75112 .58419 m .7503 .59528 L .74534 .58395 L p F P
0 g
s
.74553 .5836 m .75112 .58419 L .74535 .58393 L .812 .644 .675 r
F
.7454 .58187 m .74796 .58291 L p .74561 .58346 L F P
0 g
s
.75018 .57684 m .75112 .58419 L .7454 .58187 L p .829 .595 .584 r
F P
0 g
s
.74796 .58291 m .75112 .58419 L p .74553 .5836 L .812 .644 .675 r
F P
0 g
s
.74796 .58291 m .74553 .5836 L .74554 .5836 L .812 .644 .675 r
F
.74796 .58291 m .74554 .5836 L .74561 .58346 L F
.7463 .52073 m .75478 .53852 L .74613 .54095 L closepath p .654 .502 .664 r
F P
0 g
s
.74613 .54095 m .75478 .53852 L .75426 .54886 L closepath p .662 .527 .689 r
F P
0 g
s
.74613 .54095 m .75426 .54886 L .74677 .55211 L closepath p .653 .527 .698 r
F P
0 g
s
.74987 .49123 m .7565 .50144 L .7463 .52073 L closepath p .553 .378 .598 r
F P
0 g
s
.7565 .50144 m .75593 .52639 L .7463 .52073 L closepath p .631 .469 .645 r
F P
0 g
s
.7463 .52073 m .75593 .52639 L .75478 .53852 L closepath p .609 .481 .681 r
F P
0 g
s
.74675 .62996 m .7565 .63359 L p .74657 .62997 L .574 .753 .962 r
F P
0 g
s
.7565 .63359 m .74675 .62996 L p .74977 .63083 L .168 .624 .936 r
F P
0 g
s
.74675 .62996 m .74682 .62996 L .74977 .63083 L .168 .624 .936 r
F
.74677 .55211 m .75426 .54886 L .75373 .55805 L closepath p .626 .462 .641 r
F P
0 g
s
.74677 .55211 m .75373 .55805 L .74853 .56246 L closepath p .545 .466 .717 r
F P
0 g
s
.7468 .60336 m .75009 .60039 L .7628 .60114 L closepath p .482 0 0 r
F P
0 g
s
.7468 .60336 m .7628 .60114 L .75914 .60431 L closepath p .552 .072 .071 r
F P
0 g
s
.74775 .60481 m .7468 .60336 L .75914 .60431 L closepath p .265 0 0 r
F P
0 g
s
.74682 .62996 m .74693 .62996 L .74977 .63083 L .168 .624 .936 r
F
.7538 .63208 m .74693 .62996 L p .74706 .62995 L 0 .467 .787 r
F P
0 g
s
.7538 .63208 m .74706 .62995 L .74719 .62995 L 0 .467 .787 r
F
.7476 .62993 m .7538 .63208 L p .74719 .62995 L F P
0 g
s
.75089 .63056 m .7538 .63208 L .7476 .62993 L p .032 .549 .868 r
F P
0 g
s
.7476 .62993 m .74766 .62993 L .75089 .63056 L .032 .549 .868 r
F
.75089 .63056 m .74766 .62993 L p .75097 .62982 L .551 .925 .924 r
F P
0 g
s
.74775 .60481 m .75914 .60431 L .75756 .60764 L closepath p .596 .064 0 r
F P
0 g
s
.75088 .60681 m .74775 .60481 L .75756 .60764 L closepath p .431 0 0 r
F P
0 g
s
.75373 .55805 m .74775 .56975 L .74853 .56246 L closepath p .563 .496 .738 r
F P
0 g
s
.75373 .55805 m .75018 .57684 L .74775 .56975 L closepath p .413 .32 .641 r
F P
0 g
s
.76684 .61033 m .74889 .61826 L .75267 .61131 L closepath p .555 .343 .548 r
F P
0 g
s
.76684 .61033 m .76156 .61724 L .74889 .61826 L closepath p .49 .22 .432 r
F P
0 g
s
.76156 .61724 m .74935 .62508 L .74889 .61826 L closepath p .533 .404 .651 r
F P
0 g
s
.74928 .46223 m .76423 .46356 L p .75763 .46404 L .303 0 0 r
F P
0 g
s
.76156 .61724 m .7621 .6238 L .74935 .62508 L closepath p .482 .314 .576 r
F P
0 g
s
.74935 .62508 m .7621 .6238 L .7634 .62864 L closepath p .503 .563 .849 r
F P
0 g
s
.74935 .62508 m .7634 .62864 L .75213 .62915 L closepath p .406 .607 .93 r
F P
0 g
s
.75565 .63272 m .7565 .63359 L p .74977 .63083 L .168 .624 .936 r
F P
0 g
s
.74977 .63083 m .7538 .63208 L .75565 .63272 L .168 .624 .936 r
F
.75222 .47898 m .76249 .48135 L .74987 .49123 L closepath p .365 .04 .275 r
F P
0 g
s
.74987 .49123 m .76249 .48135 L .76653 .49023 L closepath p .525 .26 .456 r
F P
0 g
s
.74987 .49123 m .76653 .49023 L .7565 .50144 L closepath p .553 .378 .598 r
F P
0 g
s
.7503 .59528 m .7628 .60114 L .75009 .60039 L closepath p .632 .299 .39 r
F P
0 g
s
.75915 .55844 m .75018 .57684 L .75373 .55805 L closepath p .656 .551 .724 r
F P
0 g
s
.75915 .55844 m .76006 .57185 L .75018 .57684 L closepath p .528 .401 .651 r
F P
0 g
s
.76006 .57185 m .75433 .58167 L .75018 .57684 L closepath p .575 .507 .741 r
F P
0 g
s
.75018 .57684 m .75433 .58167 L .75112 .58419 L closepath p .736 .529 .612 r
F P
0 g
s
.76012 .58907 m .7503 .59528 L .75112 .58419 L closepath p .741 .581 .677 r
F P
0 g
s
.76012 .58907 m .76862 .60093 L .7503 .59528 L closepath p .698 .497 .612 r
F P
0 g
s
.7503 .59528 m .76862 .60093 L .7628 .60114 L closepath p .599 .066 0 r
F P
0 g
s
.75039 .47666 m .75763 .46404 L .76088 .47131 L closepath p .484 .133 .295 r
F P
0 g
s
.75039 .47666 m .76088 .47131 L .75999 .47585 L closepath p .448 .079 .245 r
F P
0 g
s
.75039 .47666 m .75999 .47585 L .75222 .47898 L closepath p .467 .126 .304 r
F P
0 g
s
.75088 .60681 m .75756 .60764 L .75909 .6087 L p .612 .056 0 r
F P
0 g
s
.75267 .61131 m .75088 .60681 L .76081 .6091 L closepath p .573 .121 .141 r
F P
0 g
s
.75255 .62977 m .75089 .63056 L .75182 .6298 L p .549 .928 .905 r
F P
0 g
s
.7538 .63208 m .75089 .63056 L .75259 .62975 L p .673 .894 .977 r
F P
0 g
s
.75182 .6298 m .75089 .63056 L p .75097 .62982 L .551 .925 .924 r
F P
0 g
s
.75433 .58167 m .76012 .58907 L .75112 .58419 L closepath p .757 .565 .637 r
F P
0 g
s
.75859 .62928 m .75213 .62915 L .7634 .62864 L closepath p .093 .649 .795 r
F P
0 g
s
.75217 .62978 m .75213 .62915 L .75859 .62928 L closepath p .139 .681 .828 r
F P
0 g
s
.75256 .62975 m .75255 .62977 L .75217 .62978 L .549 .928 .905 r
F
.75999 .47585 m .76249 .48135 L .75222 .47898 L closepath p .365 0 .143 r
F P
0 g
s
.76644 .6386 m .76092 .63937 L .75249 .63552 L p .205 .682 .935 r
F P
0 g
s
.75259 .62975 m .75255 .62977 L p .75257 .62975 L .549 .928 .905 r
F P
0 g
s
.75257 .62975 m .75255 .62977 L .75256 .62975 L .549 .928 .905 r
F
.7538 .62966 m .7538 .63208 L .75259 .62975 L .673 .894 .977 r
F
.75267 .61131 m .75978 .60938 L p .76684 .61033 L .397 0 0 r
F P
0 g
s
.75426 .54886 m .75915 .55844 L .75373 .55805 L closepath p .66 .491 .645 r
F P
0 g
s
.75677 .62943 m .7538 .63208 L p .7538 .62966 L .673 .894 .977 r
F P
0 g
s
.75484 .63189 m .75565 .63272 L p .7538 .63208 L .168 .624 .936 r
F P
0 g
s
.76484 .62999 m .7538 .63208 L .75663 .62955 L p .609 .743 .94 r
F P
0 g
s
.75478 .53852 m .76084 .54894 L .75426 .54886 L closepath p .611 .484 .683 r
F P
0 g
s
.75426 .54886 m .76084 .54894 L .75915 .55844 L closepath p .611 .471 .666 r
F P
0 g
s
.76006 .57185 m .76012 .58907 L .75433 .58167 L closepath p .597 .534 .755 r
F P
0 g
s
.75478 .53852 m .75593 .52639 L .76894 .53208 L closepath p .652 .52 .689 r
F P
0 g
s
.75478 .53852 m .76894 .53208 L .76182 .54037 L closepath p .64 .491 .664 r
F P
0 g
s
.75478 .53852 m .76182 .54037 L .76084 .54894 L closepath p .639 .494 .67 r
F P
0 g
s
.75587 .63169 m .76657 .63071 L p .75484 .63189 L .454 .819 .994 r
F P
0 g
s
.76657 .63071 m .7565 .63359 L .75484 .63189 L p .454 .819 .994 r
F P
0 g
s
.76484 .62999 m .76657 .63071 L .75587 .63169 L p .326 .791 .937 r
F P
0 g
s
.7565 .50144 m .76315 .5129 L .75593 .52639 L closepath p .571 .42 .637 r
F P
0 g
s
.75593 .52639 m .76315 .5129 L .77147 .52329 L closepath p .654 .52 .687 r
F P
0 g
s
.75593 .52639 m .77147 .52329 L .76894 .53208 L closepath p .654 .518 .685 r
F P
0 g
s
.76657 .63071 m .76945 .63114 L .7565 .63359 L closepath p .408 .833 .961 r
F P
0 g
s
.76945 .63114 m .76184 .63758 L .7565 .63359 L closepath p .545 .808 .994 r
F P
0 g
s
.7739 .50191 m .7565 .50144 L .76653 .49023 L closepath p .614 .464 .655 r
F P
0 g
s
.7739 .50191 m .76315 .5129 L .7565 .50144 L closepath p .613 .435 .618 r
F P
0 g
s
.75859 .62928 m .76484 .62999 L .75663 .62955 L .609 .743 .94 r
F
.75663 .62955 m .75677 .62943 L p .75859 .62928 L F P
0 g
s
.75756 .60764 m .75914 .60431 L .75925 .60874 L p .928 .721 .236 r
F P
0 g
s
.75756 .60764 m .75925 .60874 L .75909 .6087 L .928 .721 .236 r
F
.75763 .46404 m .78919 .46178 L .76088 .47131 L closepath p .467 .125 .303 r
F P
0 g
s
.75859 .62928 m .7634 .62864 L .7683 .63009 L closepath p .088 .615 .873 r
F P
0 g
s
.76353 .62969 m .76484 .62999 L .75859 .62928 L .609 .743 .94 r
F
.75914 .60431 m .7628 .60114 L .76044 .60902 L p .978 .744 .514 r
F P
0 g
s
.76044 .60902 m .76016 .60895 L .75914 .60431 L .978 .744 .514 r
F
.75925 .60874 m .75914 .60431 L p .76016 .60895 L F P
0 g
s
.76084 .54894 m .76796 .56723 L .75915 .55844 L closepath p .55 .414 .648 r
F P
0 g
s
.75915 .55844 m .76796 .56723 L .7719 .58058 L closepath p .483 .406 .694 r
F P
0 g
s
.75915 .55844 m .7719 .58058 L .76006 .57185 L closepath p .595 .452 .657 r
F P
0 g
s
.75978 .60938 m .76081 .6091 L p .76847 .61049 L .397 0 0 r
F P
0 g
s
.75978 .60938 m .76847 .61049 L .76684 .61033 L .397 0 0 r
F
.78919 .46178 m .78825 .4695 L .75999 .47585 L closepath p .533 .224 .389 r
F P
0 g
s
.78919 .46178 m .75999 .47585 L .76088 .47131 L closepath p .432 .065 .241 r
F P
0 g
s
.75999 .47585 m .78825 .4695 L .78669 .47379 L closepath p .402 0 .12 r
F P
0 g
s
.75999 .47585 m .78669 .47379 L .76249 .48135 L closepath p .404 0 .125 r
F P
0 g
s
.76006 .57185 m .7719 .58058 L .76012 .58907 L closepath p .511 .463 .74 r
F P
0 g
s
.76012 .58907 m .7719 .58058 L .78149 .59305 L closepath p .541 .52 .781 r
F P
0 g
s
.76012 .58907 m .78149 .59305 L .76862 .60093 L closepath p .566 .463 .696 r
F P
0 g
s
.76044 .60902 m .7628 .60114 L p .76081 .6091 L .978 .744 .514 r
F P
0 g
s
.76081 .6091 m .7706 .6107 L p .76847 .61049 L .397 0 0 r
F P
0 g
s
.7628 .60114 m .76333 .60951 L .76081 .6091 L .978 .744 .514 r
F
.76084 .54894 m .76182 .54037 L .77044 .5489 L closepath p .647 .501 .671 r
F P
0 g
s
.76084 .54894 m .77044 .5489 L .76931 .55383 L closepath p .639 .432 .586 r
F P
0 g
s
.76084 .54894 m .76931 .55383 L .76796 .56723 L closepath p .617 .449 .632 r
F P
0 g
s
.76092 .63937 m .76656 .63858 L .76961 .63945 L p .194 .688 .919 r
F P
0 g
s
.76932 .63977 m .76092 .63937 L p .76961 .63945 L .194 .688 .919 r
F P
0 g
s
.77616 .61266 m .76156 .61724 L .76684 .61033 L closepath p .619 .381 .534 r
F P
0 g
s
.77616 .61266 m .77302 .61758 L .76156 .61724 L closepath p .631 .407 .56 r
F P
0 g
s
.76156 .61724 m .77386 .62197 L .7621 .6238 L closepath p .599 .4 .584 r
F P
0 g
s
.76156 .61724 m .77302 .61758 L .77386 .62197 L closepath p .619 .348 .484 r
F P
0 g
s
.76182 .54037 m .76894 .53208 L .77418 .53489 L closepath p .625 .469 .651 r
F P
0 g
s
.76182 .54037 m .77418 .53489 L .77044 .5489 L closepath p .639 .501 .679 r
F P
0 g
s
.76656 .63858 m .76644 .6386 L p .76184 .63758 L .205 .682 .935 r
F P
0 g
s
.76945 .63114 m .77258 .63469 L .76184 .63758 L closepath p .568 .86 .996 r
F P
0 g
s
.76656 .63858 m .76184 .63758 L .77258 .63469 L closepath p .451 .868 .947 r
F P
0 g
s
.7621 .6238 m .77386 .62197 L .77371 .62737 L closepath p .634 .586 .782 r
F P
0 g
s
.7621 .6238 m .77371 .62737 L .7634 .62864 L closepath p .589 .621 .851 r
F P
0 g
s
.76249 .48135 m .78669 .47379 L .78417 .48181 L closepath p .564 .256 .406 r
F P
0 g
s
.76249 .48135 m .78417 .48181 L .76653 .49023 L closepath p .57 .28 .437 r
F P
0 g
s
.7628 .60114 m .76862 .60093 L .76521 .60982 L p .896 .7 .63 r
F P
0 g
s
.76424 .60966 m .7628 .60114 L p .76521 .60982 L .896 .7 .63 r
F P
0 g
s
.7628 .60114 m .76424 .60966 L p .76333 .60951 L .978 .744 .514 r
F P
0 g
s
.76315 .5129 m .7739 .50191 L .77869 .51586 L closepath p .663 .51 .667 r
F P
0 g
s
.76315 .5129 m .77869 .51586 L .77147 .52329 L closepath p .66 .521 .683 r
F P
0 g
s
.7634 .62864 m .77371 .62737 L .77511 .62913 L closepath p .44 .807 .995 r
F P
0 g
s
.7634 .62864 m .77511 .62913 L .7683 .63009 L closepath p .42 .803 .991 r
F P
0 g
s
.76462 .62979 m .76484 .62999 L .76353 .62969 L .609 .743 .94 r
F
.76463 .62979 m .76484 .62999 L p .76462 .62979 L F P
0 g
s
.76488 .62981 m .76484 .62999 L .76463 .62979 L p .782 .783 .854 r
F P
0 g
s
.76494 .62981 m .76484 .62999 L .76488 .62981 L p .828 .829 .848 r
F P
0 g
s
.76657 .63071 m .76484 .62999 L .76494 .62981 L p .906 .975 .798 r
F P
0 g
s
.76587 .62989 m .76657 .63071 L .76494 .62981 L .906 .975 .798 r
F
.76663 .63964 m .77334 .64177 L p .76503 .63957 L .058 .527 .893 r
F P
0 g
s
.77334 .64177 m .76536 .63982 L .76503 .63957 L p .058 .527 .893 r
F P
0 g
s
.76521 .60982 m .76862 .60093 L p .76606 .60996 L .927 .743 .634 r
F P
0 g
s
.77587 .64294 m .76536 .63982 L .77334 .64177 L closepath p 0 .482 .799 r
F P
0 g
s
.77587 .64294 m .76975 .64184 L .76536 .63982 L closepath p .013 .531 .86 r
F P
0 g
s
.76615 .62991 m .76657 .63071 L .76587 .62989 L .906 .975 .798 r
F
.76862 .60093 m .7661 .60997 L .76606 .60996 L .927 .743 .634 r
F
.76862 .60093 m .76612 .60997 L .7661 .60997 L F
.76862 .60093 m .76656 .61004 L .76612 .60997 L F
.76682 .62997 m .76657 .63071 L p .76615 .62991 L .906 .975 .798 r
F P
0 g
s
.76653 .49023 m .78417 .48181 L .7827 .49627 L closepath p .665 .459 .595 r
F P
0 g
s
.76653 .49023 m .7827 .49627 L .7739 .50191 L closepath p .656 .478 .631 r
F P
0 g
s
.76862 .60093 m .77042 .61067 L .76656 .61004 L .927 .743 .634 r
F
.76656 .63858 m .77258 .63469 L .77977 .63632 L closepath p .443 .837 .981 r
F P
0 g
s
.77097 .63984 m .76656 .63858 L .77977 .63632 L closepath p .481 .837 .995 r
F P
0 g
s
.76705 .62999 m .76657 .63071 L .76682 .62997 L p .699 .739 .882 r
F P
0 g
s
.76945 .63114 m .76657 .63071 L .76684 .6303 L p .811 .917 .909 r
F P
0 g
s
.77097 .63984 m .77334 .64177 L .76663 .63964 L p .191 .563 .936 r
F P
0 g
s
.76684 .6303 m .76705 .62999 L p .7683 .63009 L .811 .917 .909 r
F P
0 g
s
.76868 .63004 m .76945 .63114 L .76684 .6303 L .811 .917 .909 r
F
.77739 .61176 m .77616 .61266 L .76684 .61033 L closepath p .47 0 0 r
F P
0 g
s
.77739 .61176 m .76684 .61033 L .7706 .6107 L closepath p .185 0 0 r
F P
0 g
s
.76931 .55383 m .78044 .5702 L .76796 .56723 L closepath p .524 .367 .61 r
F P
0 g
s
.76796 .56723 m .78044 .5702 L .7719 .58058 L closepath p .503 .419 .693 r
F P
0 g
s
.78149 .59305 m .78473 .60927 L .76862 .60093 L closepath p .678 .726 .886 r
F P
0 g
s
.77209 .61093 m .76862 .60093 L p .77211 .61094 L .788 .688 .754 r
F P
0 g
s
.76862 .60093 m .77209 .61093 L p .77157 .61085 L .927 .743 .634 r
F P
0 g
s
.76862 .60093 m .77155 .61079 L .77042 .61067 L .927 .743 .634 r
F
.76862 .60093 m .78473 .60927 L .77211 .61094 L p .788 .688 .754 r
F P
0 g
s
.76989 .62987 m .76945 .63114 L p .76868 .63004 L .811 .917 .909 r
F P
0 g
s
.76894 .53208 m .77147 .52329 L .77418 .53489 L closepath p .614 .477 .671 r
F P
0 g
s
.77044 .5489 m .78044 .5702 L .76931 .55383 L closepath p .592 .387 .571 r
F P
0 g
s
.76961 .63945 m .77097 .63984 L .76932 .63977 L p .194 .688 .919 r
F P
0 g
s
.77258 .63469 m .76945 .63114 L .77194 .62958 L p .755 .939 .944 r
F P
0 g
s
.77194 .62958 m .76945 .63114 L .76989 .62987 L p .765 .867 .924 r
F P
0 g
s
.78277 .64769 m .76975 .64184 L .77587 .64294 L closepath p .084 .541 .905 r
F P
0 g
s
.77155 .61079 m .77052 .61069 L .77042 .61067 L .927 .743 .634 r
F
.77044 .5489 m .77418 .53489 L .78158 .54553 L closepath p .55 .412 .646 r
F P
0 g
s
.77044 .5489 m .78158 .54553 L .78076 .55349 L closepath p .503 .293 .527 r
F P
0 g
s
.77044 .5489 m .78076 .55349 L .78044 .5702 L closepath p .469 .33 .609 r
F P
0 g
s
.77155 .61079 m .77053 .61069 L .77052 .61069 L .927 .743 .634 r
F
.77155 .61079 m .77053 .61069 L .77053 .61069 L F
.77155 .61079 m .7706 .6107 L .77053 .61069 L F
.77155 .61079 m .77157 .61085 L .7706 .6107 L F
.77097 .63984 m .77977 .63632 L .78091 .63789 L closepath p .495 .797 .999 r
F P
0 g
s
.77334 .64177 m .77097 .63984 L .78091 .63789 L closepath p .54 .731 .963 r
F P
0 g
s
.77147 .52329 m .77869 .51586 L .78217 .52126 L closepath p .612 .448 .636 r
F P
0 g
s
.77147 .52329 m .78217 .52126 L .77418 .53489 L closepath p .62 .481 .67 r
F P
0 g
s
.7719 .58058 m .78044 .5702 L .7874 .58261 L closepath p .546 .48 .732 r
F P
0 g
s
.7719 .58058 m .7874 .58261 L .78149 .59305 L closepath p .532 .517 .784 r
F P
0 g
s
.77279 .62945 m .77258 .63469 L .77194 .62958 L .755 .939 .944 r
F
.77211 .61094 m .77774 .61019 L p .7768 .61166 L .787 .742 .812 r
F P
0 g
s
.77977 .63632 m .77258 .63469 L .78263 .62833 L p .749 .913 .948 r
F P
0 g
s
.78247 .62843 m .77258 .63469 L p .78191 .6284 L .746 .924 .95 r
F P
0 g
s
.77258 .63469 m .77643 .62898 L p .78191 .6284 L .746 .924 .95 r
F P
0 g
s
.7763 .62918 m .77258 .63469 L p .77279 .62945 L .755 .939 .944 r
F P
0 g
s
.7763 .62918 m .77279 .62945 L .77511 .62913 L .755 .939 .944 r
F
.78802 .61421 m .77302 .61758 L .77616 .61266 L closepath p .655 .438 .576 r
F P
0 g
s
.78802 .61421 m .78452 .6182 L .77302 .61758 L closepath p .622 .351 .486 r
F P
0 g
s
.78452 .6182 m .77386 .62197 L .77302 .61758 L closepath p .621 .349 .483 r
F P
0 g
s
.77334 .64177 m .78091 .63789 L .78389 .63953 L closepath p .551 .819 .996 r
F P
0 g
s
.77587 .64294 m .77334 .64177 L .78389 .63953 L closepath p .479 .847 .989 r
F P
0 g
s
.78311 .62756 m .77371 .62737 L .77386 .62197 L closepath p .668 .614 .785 r
F P
0 g
s
.77511 .62913 m .77371 .62737 L .78311 .62756 L closepath p .405 .784 .992 r
F P
0 g
s
.78452 .6182 m .78498 .62359 L .77386 .62197 L closepath p .663 .437 .568 r
F P
0 g
s
.78311 .62756 m .77386 .62197 L .78498 .62359 L closepath p .616 .623 .834 r
F P
0 g
s
.7739 .50191 m .7827 .49627 L .78623 .50853 L closepath p .672 .508 .655 r
F P
0 g
s
.7739 .50191 m .78623 .50853 L .77869 .51586 L closepath p .667 .513 .666 r
F P
0 g
s
.77418 .53489 m .78217 .52126 L .78562 .54058 L closepath p .546 .391 .622 r
F P
0 g
s
.77418 .53489 m .78562 .54058 L .78158 .54553 L closepath p .527 .405 .657 r
F P
0 g
s
.77612 .62902 m .7763 .62918 L .77511 .62913 L .755 .939 .944 r
F
.77511 .62913 m .78311 .62756 L .78254 .62833 L closepath p .327 .79 .938 r
F P
0 g
s
.78597 .63865 m .78277 .64769 L .77587 .64294 L p .695 .835 .95 r
F P
0 g
s
.77643 .62898 m .7763 .62918 L p .77612 .62902 L .755 .939 .944 r
F P
0 g
s
.77616 .61266 m .77739 .61176 L .78661 .61317 L closepath p .396 0 0 r
F P
0 g
s
.77616 .61266 m .78661 .61317 L .78802 .61421 L closepath p .467 0 0 r
F P
0 g
s
.7773 .61174 m .7768 .61166 L .77774 .61019 L .787 .742 .812 r
F
.7773 .61174 m .7773 .61174 L .77774 .61019 L F
.77812 .61122 m .7773 .61174 L .77774 .61019 L F
.77739 .61176 m .7773 .61174 L .77812 .61122 L F
.77742 .64529 m .78329 .65151 L .77734 .64525 L p 0 0 .5 r
F P
0 g
s
.77838 .61191 m .77739 .61176 L .77812 .61122 L .787 .742 .812 r
F
.78983 .65144 m .78329 .65151 L .77742 .64529 L p .439 .632 .936 r
F P
0 g
s
.78277 .64769 m .78983 .65144 L .77742 .64529 L .439 .632 .936 r
F
.77774 .61019 m .78473 .60927 L p .78011 .61217 L .787 .742 .812 r
F P
0 g
s
.77902 .612 m .77838 .61191 L .77774 .61019 L .787 .742 .812 r
F
.78011 .61217 m .77902 .612 L .77774 .61019 L F
.77869 .51586 m .78623 .50853 L .78217 .52126 L closepath p .629 .454 .627 r
F P
0 g
s
.78363 .62831 m .77977 .63632 L p .78263 .62833 L .749 .913 .948 r
F P
0 g
s
.78857 .62997 m .77977 .63632 L .78363 .62831 L p .76 .926 .943 r
F P
0 g
s
.78091 .63789 m .77977 .63632 L .78857 .62997 L closepath p .76 .926 .943 r
F P
0 g
s
.78473 .60927 m .78057 .61224 L p .78011 .61217 L .787 .742 .812 r
F P
0 g
s
.78405 .55439 m .78044 .5702 L .78076 .55349 L closepath p .433 .301 .6 r
F P
0 g
s
.78405 .55439 m .78357 .56862 L .78044 .5702 L closepath p .498 .363 .627 r
F P
0 g
s
.78044 .5702 m .78357 .56862 L .79227 .57839 L closepath p .561 .5 .745 r
F P
0 g
s
.78044 .5702 m .79227 .57839 L .7874 .58261 L closepath p .586 .497 .72 r
F P
0 g
s
.78057 .61224 m .78473 .60927 L .78164 .61241 L p .787 .743 .812 r
F P
0 g
s
.78158 .54553 m .78405 .55439 L .78076 .55349 L closepath p .453 .25 .513 r
F P
0 g
s
.78091 .63789 m .78857 .62997 L .79253 .6314 L closepath p .765 .979 .914 r
F P
0 g
s
.78389 .63953 m .78091 .63789 L .79253 .6314 L closepath p .828 .857 .869 r
F P
0 g
s
.7874 .58261 m .79418 .5972 L .78149 .59305 L closepath p .671 .694 .863 r
F P
0 g
s
.78149 .59305 m .79384 .60906 L .78473 .60927 L closepath p .834 .824 .838 r
F P
0 g
s
.78149 .59305 m .79418 .5972 L .79384 .60906 L closepath p .548 .723 .955 r
F P
0 g
s
.78158 .54553 m .78562 .54058 L .79084 .54631 L closepath p .496 .363 .629 r
F P
0 g
s
.78158 .54553 m .78596 .55428 L .78405 .55439 L closepath p .654 .368 .474 r
F P
0 g
s
.78158 .54553 m .79084 .54631 L .78596 .55428 L closepath p .497 .304 .548 r
F P
0 g
s
.78164 .61241 m .78473 .60927 L p .78201 .61246 L .839 .863 .861 r
F P
0 g
s
.78254 .62833 m .78247 .62843 L .78191 .6284 L .746 .924 .95 r
F
.78473 .60927 m .78344 .61268 L .78201 .61246 L .839 .863 .861 r
F
.78217 .52126 m .80492 .49639 L .78969 .51799 L closepath p .589 .381 .568 r
F P
0 g
s
.78217 .52126 m .78623 .50853 L .80492 .49639 L closepath p .668 .495 .642 r
F P
0 g
s
.78217 .52126 m .78969 .51799 L .79578 .53623 L closepath p .616 .441 .622 r
F P
0 g
s
.78217 .52126 m .79578 .53623 L .78562 .54058 L closepath p .621 .441 .618 r
F P
0 g
s
.78263 .62833 m .78247 .62843 L p .78254 .62833 L .746 .924 .95 r
F P
0 g
s
.8115 .62776 m .78254 .62833 L .78311 .62756 L closepath p .299 .764 .945 r
F P
0 g
s
.78417 .48181 m .80788 .4894 L .7827 .49627 L closepath p .649 .445 .593 r
F P
0 g
s
.7827 .49627 m .80788 .4894 L .80492 .49639 L closepath p .678 .529 .676 r
F P
0 g
s
.7827 .49627 m .80492 .49639 L .78623 .50853 L closepath p .677 .511 .654 r
F P
0 g
s
.7908 .65143 m .78983 .65144 L p .78277 .64769 L .439 .632 .936 r
F P
0 g
s
.78277 .64769 m .78597 .63863 L .79191 .64271 L closepath p .685 .824 .95 r
F P
0 g
s
.7908 .65143 m .78277 .64769 L .79191 .64271 L closepath p .672 .747 .907 r
F P
0 g
s
.8115 .62776 m .78311 .62756 L .78498 .62359 L closepath p .673 .692 .859 r
F P
0 g
s
.78329 .65151 m .7908 .65143 L .80209 .65564 L closepath p .34 .703 .983 r
F P
0 g
s
.78329 .65151 m .80209 .65564 L .79638 .65736 L closepath p .515 .712 .961 r
F P
0 g
s
.78473 .60927 m .7835 .61269 L p .78344 .61268 L .839 .863 .861 r
F P
0 g
s
.7835 .61269 m .78473 .60927 L p .78351 .61269 L .882 .915 .844 r
F P
0 g
s
.78473 .60927 m .78459 .61286 L p .78351 .61269 L .882 .915 .844 r
F P
0 g
s
.78596 .55428 m .78357 .56862 L .78405 .55439 L closepath p .691 .524 .656 r
F P
0 g
s
.78596 .55428 m .78478 .56829 L .78357 .56862 L closepath p .143 .029 .453 r
F P
0 g
s
.78357 .56862 m .78478 .56829 L .79062 .57577 L closepath p .243 .395 .816 r
F P
0 g
s
.78357 .56862 m .79062 .57577 L .79227 .57839 L closepath p .222 .387 .817 r
F P
0 g
s
.78556 .62827 m .78857 .62997 L p .78363 .62831 L .76 .926 .943 r
F P
0 g
s
.79245 .63147 m .78597 .63863 L .78389 .63953 L p .7 .959 .703 r
F P
0 g
s
.78597 .63863 m .78597 .63865 L p .78389 .63953 L .695 .835 .95 r
F P
0 g
s
.78669 .47379 m .81092 .48159 L .78417 .48181 L closepath p .638 .33 .433 r
F P
0 g
s
.78417 .48181 m .81092 .48159 L .80788 .4894 L closepath p .659 .406 .526 r
F P
0 g
s
.8115 .62776 m .78452 .6182 L .78802 .61421 L closepath p .728 .497 .578 r
F P
0 g
s
.8115 .62776 m .78498 .62359 L .78452 .6182 L closepath p .73 .488 .561 r
F P
0 g
s
.78459 .61286 m .78473 .60927 L p .78592 .61307 L .819 .87 .884 r
F P
0 g
s
.78473 .60927 m .79384 .60906 L .79174 .61618 L p .805 .865 .893 r
F P
0 g
s
.78636 .61302 m .78473 .60927 L p .79174 .61618 L .805 .865 .893 r
F P
0 g
s
.78473 .60927 m .78641 .61314 L p .78592 .61307 L .819 .87 .884 r
F P
0 g
s
.78596 .55428 m .78888 .56735 L .78478 .56829 L closepath p .722 .518 .613 r
F P
0 g
s
.78888 .56735 m .79062 .57577 L .78478 .56829 L closepath p .736 .562 .658 r
F P
0 g
s
.7889 .62821 m .78857 .62997 L .78556 .62827 L p .832 .93 .896 r
F P
0 g
s
.78562 .54058 m .79578 .53623 L .79585 .54082 L closepath p .561 .318 .504 r
F P
0 g
s
.78562 .54058 m .79585 .54082 L .79084 .54631 L closepath p .569 .372 .574 r
F P
0 g
s
.79084 .54631 m .79406 .54756 L .78596 .55428 L closepath p .59 .414 .612 r
F P
0 g
s
.79406 .54756 m .79438 .55058 L .78596 .55428 L closepath p .493 .267 .499 r
F P
0 g
s
.79438 .55058 m .79114 .55922 L .78596 .55428 L closepath p .541 .363 .589 r
F P
0 g
s
.78596 .55428 m .79114 .55922 L .79463 .56387 L closepath p .251 .326 .746 r
F P
0 g
s
.78596 .55428 m .79463 .56387 L .79247 .56554 L closepath p .661 .403 .519 r
F P
0 g
s
.78888 .56735 m .78596 .55428 L .79247 .56554 L closepath p .384 .299 .636 r
F P
0 g
s
.79302 .63229 m .78597 .63863 L .79245 .63147 L p .805 .994 .868 r
F P
0 g
s
.79191 .64271 m .78597 .63863 L .79302 .63229 L closepath p .813 .832 .865 r
F P
0 g
s
.78636 .61302 m .78661 .61317 L .78659 .61317 L .805 .865 .893 r
F
.78641 .61314 m .78636 .61302 L p .78659 .61317 L F P
0 g
s
.78661 .61317 m .8115 .62776 L .78802 .61421 L closepath p .462 0 0 r
F P
0 g
s
.78825 .4695 m .81338 .47238 L .78669 .47379 L closepath p .516 .085 .16 r
F P
0 g
s
.78669 .47379 m .81338 .47238 L .81092 .48159 L closepath p .633 .374 .508 r
F P
0 g
s
.79227 .57839 m .79842 .58825 L .7874 .58261 L closepath p .656 .62 .802 r
F P
0 g
s
.7874 .58261 m .79842 .58825 L .79418 .5972 L closepath p .566 .638 .881 r
F P
0 g
s
.78919 .46178 m .81401 .47015 L .78825 .4695 L closepath p .601 .282 .402 r
F P
0 g
s
.78825 .4695 m .81401 .47015 L .81338 .47238 L closepath p .475 0 .024 r
F P
0 g
s
.79034 .62818 m .78857 .62997 L .7889 .62821 L p .851 .946 .88 r
F P
0 g
s
.79253 .6314 m .78857 .62997 L .791 .62816 L p .856 .982 .857 r
F P
0 g
s
.791 .62816 m .78857 .62997 L .79034 .62818 L p .828 .996 .849 r
F P
0 g
s
.78888 .56735 m .79247 .56554 L .79348 .57314 L closepath p .466 .488 .797 r
F P
0 g
s
.78888 .56735 m .79348 .57314 L .79062 .57577 L closepath p .727 .556 .66 r
F P
0 g
s
.78919 .46178 m .81538 .46303 L .81401 .47015 L closepath p .601 .289 .414 r
F P
0 g
s
.78969 .51799 m .80492 .49639 L .82552 .4959 L closepath p .683 .503 .638 r
F P
0 g
s
.78969 .51799 m .82552 .4959 L .80242 .51868 L closepath p .663 .467 .609 r
F P
0 g
s
.78969 .51799 m .80242 .51868 L .79578 .53623 L closepath p .663 .467 .608 r
F P
0 g
s
.79348 .57314 m .79892 .57966 L .79062 .57577 L closepath p .717 .54 .649 r
F P
0 g
s
.79227 .57839 m .79062 .57577 L .79892 .57966 L closepath p .652 .594 .777 r
F P
0 g
s
.7908 .65143 m .79191 .64271 L .79779 .64749 L closepath p .738 .807 .906 r
F P
0 g
s
.80209 .65564 m .7908 .65143 L .79779 .64749 L closepath p .723 .736 .862 r
F P
0 g
s
.79084 .54631 m .79585 .54082 L .79934 .54544 L closepath p .623 .447 .624 r
F P
0 g
s
.79084 .54631 m .79934 .54544 L .79406 .54756 L closepath p .607 .382 .549 r
F P
0 g
s
.79543 .62808 m .79253 .6314 L p .791 .62816 L .856 .982 .857 r
F P
0 g
s
.79438 .55058 m .79418 .55418 L .79114 .55922 L closepath p .884 .753 .712 r
F P
0 g
s
.79114 .55922 m .79418 .55418 L .80024 .55573 L closepath p .556 .321 .515 r
F P
0 g
s
.79114 .55922 m .80024 .55573 L .79739 .55837 L closepath p .52 .25 .446 r
F P
0 g
s
.79463 .56387 m .79114 .55922 L .79739 .55837 L closepath p .574 .436 .655 r
F P
0 g
s
.79174 .61618 m .79384 .60906 L p .79432 .61769 L .822 .883 .888 r
F P
0 g
s
.79191 .64271 m .79302 .63229 L .79768 .63757 L closepath p .835 .851 .858 r
F P
0 g
s
.79779 .64749 m .79191 .64271 L .79768 .63757 L closepath p .821 .81 .84 r
F P
0 g
s
.79227 .57839 m .80028 .58302 L .79842 .58825 L closepath p .551 .576 .832 r
F P
0 g
s
.79227 .57839 m .79892 .57966 L .80028 .58302 L closepath p .674 .54 .694 r
F P
0 g
s
.79245 .63147 m .79251 .63142 L .79302 .63229 L .805 .994 .868 r
F
.8024 .5651 m .79247 .56554 L .79463 .56387 L closepath p .695 .458 .56 r
F P
0 g
s
.8024 .5651 m .803 .56859 L .79247 .56554 L closepath p .714 .583 .707 r
F P
0 g
s
.803 .56859 m .79348 .57314 L .79247 .56554 L closepath p .672 .641 .811 r
F P
0 g
s
.79253 .6314 m .793 .63114 L .79251 .63142 L .805 .994 .868 r
F
.793 .63114 m .79302 .63229 L p .79251 .63142 L F P
0 g
s
.79299 .63088 m .79253 .6314 L .79543 .62808 L p .535 .816 .467 r
F P
0 g
s
.79299 .63088 m .793 .63114 L p .79253 .6314 L .805 .994 .868 r
F P
0 g
s
.79302 .63229 m .79299 .63088 L p .79522 .62832 L .87 .988 .723 r
F P
0 g
s
.79543 .62808 m .79299 .63088 L p .79543 .62808 L .535 .816 .467 r
F P
0 g
s
.79768 .63757 m .79302 .63229 L .79601 .62807 L p .883 .852 .807 r
F P
0 g
s
.79601 .62807 m .79302 .63229 L p .79543 .62808 L .87 .988 .723 r
F P
0 g
s
.79543 .62808 m .79302 .63229 L .79522 .62832 L .87 .988 .723 r
F
.803 .56859 m .79963 .57375 L .79348 .57314 L closepath p .685 .688 .847 r
F P
0 g
s
.79348 .57314 m .79963 .57375 L .79892 .57966 L closepath p .722 .541 .644 r
F P
0 g
s
.79418 .5972 m .80839 .60438 L .79384 .60906 L closepath p .799 .922 .919 r
F P
0 g
s
.79384 .60906 m .80839 .60438 L .79432 .61769 L p .822 .883 .888 r
F P
0 g
s
.79406 .54756 m .79934 .54544 L .80309 .54932 L closepath p .654 .487 .645 r
F P
0 g
s
.79406 .54756 m .80309 .54932 L .79438 .55058 L closepath p .659 .386 .495 r
F P
0 g
s
.80551 .58749 m .80527 .5946 L .79418 .5972 L closepath p .752 .885 .939 r
F P
0 g
s
.80551 .58749 m .79418 .5972 L .79842 .58825 L closepath p .743 .853 .931 r
F P
0 g
s
.79418 .5972 m .80527 .5946 L .80839 .60438 L closepath p .716 .913 .962 r
F P
0 g
s
.79418 .55418 m .79438 .55058 L .80309 .54932 L closepath p .715 .607 .734 r
F P
0 g
s
.79418 .55418 m .80309 .54932 L .8013 .55112 L closepath p .646 .456 .612 r
F P
0 g
s
.79418 .55418 m .8013 .55112 L .80292 .55469 L closepath p .56 .284 .454 r
F P
0 g
s
.79418 .55418 m .80292 .55469 L .80024 .55573 L closepath p .56 .285 .456 r
F P
0 g
s
.79432 .61769 m .80839 .60438 L p .79592 .61863 L .828 .971 .891 r
F P
0 g
s
.80014 .55901 m .79463 .56387 L .79739 .55837 L closepath p .65 .526 .699 r
F P
0 g
s
.80014 .55901 m .8024 .5651 L .79463 .56387 L closepath p .681 .579 .734 r
F P
0 g
s
.80242 .51868 m .80149 .53591 L .79578 .53623 L closepath p .688 .493 .618 r
F P
0 g
s
.79578 .53623 m .80149 .53591 L .80273 .54049 L closepath p .671 .421 .533 r
F P
0 g
s
.79578 .53623 m .80273 .54049 L .79585 .54082 L closepath p .679 .411 .509 r
F P
0 g
s
.79585 .54082 m .80273 .54049 L .80536 .54437 L closepath p .694 .468 .576 r
F P
0 g
s
.79585 .54082 m .80536 .54437 L .79934 .54544 L closepath p .696 .46 .562 r
F P
0 g
s
.80839 .60438 m .804 .61439 L .79592 .61863 L p .828 .971 .891 r
F P
0 g
s
.79592 .61863 m .804 .61439 L .79837 .62007 L p .851 .946 .88 r
F P
0 g
s
.79768 .63757 m .79601 .62807 L .79825 .62802 L .883 .852 .807 r
F
.80962 .65832 m .79638 .65736 L .80209 .65564 L closepath p .515 .714 .962 r
F P
0 g
s
.80962 .65832 m .80577 .6601 L .79638 .65736 L closepath p .487 .726 .977 r
F P
0 g
s
.80024 .55573 m .80014 .55901 L .79739 .55837 L closepath p .667 .462 .598 r
F P
0 g
s
.80081 .63012 m .79768 .63757 L .79828 .62802 L p .971 .914 .692 r
F P
0 g
s
.79768 .63757 m .79825 .62802 L .79826 .62802 L .883 .852 .807 r
F
.79828 .62802 m .79768 .63757 L p .79826 .62802 L F P
0 g
s
.80322 .64217 m .79768 .63757 L .80081 .63012 L closepath p .898 .782 .725 r
F P
0 g
s
.79779 .64749 m .79768 .63757 L .80322 .64217 L closepath p .814 .804 .842 r
F P
0 g
s
.80642 .65331 m .79779 .64749 L .80322 .64217 L closepath p .767 .707 .795 r
F P
0 g
s
.80209 .65564 m .79779 .64749 L .80642 .65331 L closepath p .678 .716 .877 r
F P
0 g
s
.79828 .62802 m .79992 .62799 L .80081 .63012 L .971 .914 .692 r
F
.79837 .62007 m .804 .61439 L p .79855 .62017 L .828 .996 .849 r
F P
0 g
s
.80028 .58302 m .80551 .58749 L .79842 .58825 L closepath p .769 .817 .891 r
F P
0 g
s
.79855 .62017 m .804 .61439 L .79861 .6202 L .828 .996 .849 r
F
.79861 .6202 m .804 .61439 L .79865 .62023 L F
.79865 .62023 m .804 .61439 L .80034 .62122 L F
.79963 .57375 m .80466 .57934 L .79892 .57966 L closepath p .692 .513 .64 r
F P
0 g
s
.79892 .57966 m .80466 .57934 L .80668 .58206 L closepath p .697 .653 .803 r
F P
0 g
s
.79892 .57966 m .80668 .58206 L .80028 .58302 L closepath p .747 .575 .662 r
F P
0 g
s
.79934 .54544 m .80536 .54437 L .80916 .55026 L closepath p .728 .573 .681 r
F P
0 g
s
.79934 .54544 m .80916 .55026 L .80309 .54932 L closepath p .777 .467 .462 r
F P
0 g
s
.803 .56859 m .80306 .5726 L .79963 .57375 L closepath p .696 .703 .853 r
F P
0 g
s
.79963 .57375 m .80306 .5726 L .80466 .57934 L closepath p .648 .51 .681 r
F P
0 g
s
.79992 .62799 m .79992 .62799 L .80081 .63012 L .971 .914 .692 r
F
.79993 .62799 m .80081 .63012 L p .79992 .62799 L F P
0 g
s
.80106 .62797 m .80081 .63012 L .79993 .62799 L p .979 .788 .374 r
F P
0 g
s
.80024 .55573 m .80356 .55901 L .80014 .55901 L closepath p .76 .541 .6 r
F P
0 g
s
.80014 .55901 m .80356 .55901 L .8075 .56364 L closepath p .763 .686 .777 r
F P
0 g
s
.80014 .55901 m .8075 .56364 L .8024 .5651 L closepath p .824 .649 .665 r
F P
0 g
s
.80024 .55573 m .80292 .55469 L .80528 .55635 L closepath p .699 .612 .756 r
F P
0 g
s
.80024 .55573 m .80528 .55635 L .80356 .55901 L closepath p .711 .548 .666 r
F P
0 g
s
.80028 .58302 m .80668 .58206 L .80551 .58749 L closepath p .737 .813 .912 r
F P
0 g
s
.804 .61439 m .80336 .61898 L .80034 .62122 L p .828 .996 .849 r
F P
0 g
s
.80034 .62122 m .80336 .61898 L .80105 .62163 L p .856 .982 .857 r
F P
0 g
s
.80518 .62805 m .80081 .63012 L .80106 .62797 L p .988 .753 .489 r
F P
0 g
s
.80518 .62805 m .80543 .63617 L .80081 .63012 L closepath p .988 .753 .489 r
F P
0 g
s
.80322 .64217 m .80081 .63012 L .80543 .63617 L closepath p .963 .81 .643 r
F P
0 g
s
.80105 .62163 m .80336 .61898 L .80105 .62163 L p .535 .816 .467 r
F P
0 g
s
.80135 .62181 m .80105 .62163 L .80131 .62133 L .952 .852 .43 r
F
.80509 .62789 m .80518 .62805 L p .80106 .62797 L .988 .753 .489 r
F P
0 g
s
.80309 .54932 m .80787 .55266 L .8013 .55112 L closepath p .737 .598 .703 r
F P
0 g
s
.8013 .55112 m .80787 .55266 L .80292 .55469 L closepath p .732 .359 .345 r
F P
0 g
s
.8018 .62138 m .80131 .62133 L .80131 .62133 L .952 .852 .43 r
F
.8018 .62138 m .80135 .62181 L p .80131 .62133 L F P
0 g
s
.8017 .62089 m .8018 .62138 L p .80131 .62133 L .952 .852 .43 r
F P
0 g
s
.80135 .62181 m .8018 .62138 L p .80135 .62181 L .979 .788 .374 r
F P
0 g
s
.8018 .62138 m .80135 .62181 L .80135 .62181 L .979 .788 .374 r
F
.8018 .62138 m .80173 .62203 L p .80135 .62181 L F P
0 g
s
.80242 .51868 m .80792 .51989 L .80149 .53591 L closepath p .665 .474 .617 r
F P
0 g
s
.80149 .53591 m .80792 .51989 L .82132 .53097 L closepath p .714 .528 .637 r
F P
0 g
s
.80149 .53591 m .82132 .53097 L .81258 .54125 L closepath p .711 .521 .63 r
F P
0 g
s
.80273 .54049 m .80149 .53591 L .81258 .54125 L closepath p .741 .462 .508 r
F P
0 g
s
.80336 .61898 m .8017 .62089 L .8017 .62089 L .812 .71 .752 r
F
.80336 .61898 m .8017 .62089 L .8017 .62089 L F
.8017 .62089 m .80336 .61898 L .80174 .62109 L F
.80336 .61898 m .80336 .61898 L .8017 .62089 L F
.80173 .62203 m .8018 .62138 L .8023 .62237 L p .988 .753 .489 r
F P
0 g
s
.80336 .61898 m .80356 .61889 L .80174 .62109 L .812 .71 .752 r
F
.80174 .62109 m .80356 .61889 L .80175 .62113 L F
.80336 .61898 m .80336 .61898 L .80174 .62109 L F
.80357 .61889 m .80361 .61888 L .80175 .62113 L F
.80175 .62113 m .80361 .61888 L .80425 .6186 L F
.80175 .62113 m .80175 .62113 L .80425 .6186 L F
.80356 .61889 m .80357 .61889 L .80175 .62113 L F
.8018 .62138 m .80175 .62113 L .80184 .62104 L F
.80175 .62113 m .80425 .6186 L .80425 .6186 L F
.80175 .62113 m .80425 .6186 L .80425 .6186 L F
.8018 .62138 m .80184 .62104 L .80241 .62243 L F
.8018 .62138 m .80241 .62243 L .8023 .62237 L F
.80297 .62276 m .80184 .62104 L .80425 .6186 L F
.80241 .62243 m .80184 .62104 L .80297 .62276 L F
.80962 .65832 m .80209 .65564 L .80642 .65331 L closepath p .67 .688 .858 r
F P
0 g
s
.8075 .56364 m .803 .56859 L .8024 .5651 L closepath p .826 .657 .673 r
F P
0 g
s
.80242 .51868 m .82552 .4959 L .83258 .50785 L closepath p .69 .507 .635 r
F P
0 g
s
.80242 .51868 m .83258 .50785 L .80792 .51989 L closepath p .668 .454 .586 r
F P
0 g
s
.80536 .54437 m .80273 .54049 L .80932 .54293 L closepath p .767 .479 .499 r
F P
0 g
s
.80273 .54049 m .81258 .54125 L .80932 .54293 L closepath p .753 .551 .623 r
F P
0 g
s
.80292 .55469 m .80787 .55266 L .81148 .55719 L closepath p .875 .66 .608 r
F P
0 g
s
.80292 .55469 m .81148 .55719 L .80528 .55635 L closepath p .829 .418 .264 r
F P
0 g
s
.80297 .62276 m .8046 .61845 L .8058 .61793 L .812 .71 .752 r
F
.8048 .62383 m .80297 .62276 L .8058 .61793 L F
.80297 .62276 m .80425 .6186 L .8046 .61845 L F
.8075 .56364 m .80953 .56519 L .803 .56859 L closepath p .732 .507 .587 r
F P
0 g
s
.803 .56859 m .80953 .56519 L .80819 .56993 L closepath p .833 .783 .804 r
F P
0 g
s
.803 .56859 m .80819 .56993 L .80306 .5726 L closepath p .818 .798 .833 r
F P
0 g
s
.80306 .5726 m .80819 .56993 L .81274 .57489 L closepath p .739 .557 .647 r
F P
0 g
s
.80306 .5726 m .81274 .57489 L .80466 .57934 L closepath p .737 .565 .66 r
F P
0 g
s
.80309 .54932 m .80916 .55026 L .80787 .55266 L closepath p .789 .564 .596 r
F P
0 g
s
.81094 .64726 m .80322 .64217 L .80543 .63617 L closepath p .827 .624 .629 r
F P
0 g
s
.80642 .65331 m .80322 .64217 L .81094 .64726 L closepath p .718 .679 .811 r
F P
0 g
s
.80336 .61898 m .804 .61439 L .82088 .61142 L closepath p .747 .952 .943 r
F P
0 g
s
.80356 .55901 m .80528 .55635 L .81148 .55719 L closepath p .884 .806 .766 r
F P
0 g
s
.80356 .55901 m .81148 .55719 L .8075 .56364 L closepath p .872 .682 .642 r
F P
0 g
s
.804 .61439 m .80839 .60438 L .82566 .60558 L closepath p .779 .913 .931 r
F P
0 g
s
.804 .61439 m .82566 .60558 L .82088 .61142 L closepath p .727 .955 .947 r
F P
0 g
s
.80466 .57934 m .81274 .57489 L .81647 .57498 L closepath p .796 .716 .776 r
F P
0 g
s
.80466 .57934 m .81647 .57498 L .80668 .58206 L closepath p .786 .673 .739 r
F P
0 g
s
.80657 .6176 m .8048 .62383 L p .80611 .61779 L .812 .71 .752 r
F P
0 g
s
.8048 .62383 m .8058 .61793 L .80611 .61779 L .812 .71 .752 r
F
.8048 .62383 m .80502 .62306 L p .80505 .62398 L .832 .731 .753 r
F P
0 g
s
.80492 .49639 m .80788 .4894 L .8343 .48535 L closepath p .699 .553 .685 r
F P
0 g
s
.80492 .49639 m .8343 .48535 L .82552 .4959 L closepath p .684 .513 .65 r
F P
0 g
s
.80502 .62306 m .80657 .6176 L p .80585 .62445 L .832 .731 .753 r
F P
0 g
s
.80502 .62306 m .80507 .62399 L .80505 .62398 L .832 .731 .753 r
F
.80502 .62306 m .80507 .62399 L .80507 .62399 L F
.80502 .62306 m .80508 .624 L .80507 .62399 L F
.80502 .62306 m .8051 .62401 L .80509 .624 L F
.80502 .62306 m .80585 .62445 L .8051 .62401 L F
.80502 .62306 m .80509 .624 L .80508 .624 L F
.80509 .62789 m .80517 .62789 L .80518 .62805 L F
.80517 .62789 m .80529 .62817 L .80518 .62805 L F
.80528 .62789 m .80529 .62817 L .80517 .62789 L F
.81205 .64108 m .80543 .63617 L .80518 .62805 L closepath p .838 .628 .618 r
F P
0 g
s
.81205 .64108 m .80518 .62805 L .81254 .63584 L closepath p .934 .627 .435 r
F P
0 g
s
.80551 .58749 m .819 .58381 L .80527 .5946 L closepath p .844 .947 .886 r
F P
0 g
s
.80527 .5946 m .819 .58381 L .82205 .59031 L closepath p .832 .974 .886 r
F P
0 g
s
.80527 .5946 m .82205 .59031 L .80839 .60438 L closepath p .848 .967 .877 r
F P
0 g
s
.80529 .62788 m .80531 .62819 L .80528 .62789 L .832 .731 .753 r
F
.80528 .62789 m .80531 .62819 L .80529 .62817 L F
.80529 .62788 m .80535 .62788 L .80531 .62819 L F
.80531 .62819 m .80535 .62788 L .80553 .62843 L F
.80535 .62788 m .8056 .62788 L .80553 .62843 L F
.80536 .54437 m .80932 .54293 L .80885 .54724 L closepath p .843 .681 .681 r
F P
0 g
s
.80536 .54437 m .80885 .54724 L .80916 .55026 L closepath p .905 .577 .414 r
F P
0 g
s
.81094 .64726 m .80543 .63617 L .81205 .64108 L closepath p .839 .627 .616 r
F P
0 g
s
.80551 .58749 m .81679 .58135 L .819 .58381 L closepath p .771 .976 .917 r
F P
0 g
s
.80668 .58206 m .81679 .58135 L .80551 .58749 L closepath p .826 .898 .891 r
F P
0 g
s
.80553 .62843 m .8056 .62788 L .80565 .62855 L .832 .731 .753 r
F
.8056 .62788 m .80572 .62788 L .80565 .62855 L F
.80565 .62855 m .80641 .62786 L p .80663 .62959 L .88 .965 .846 r
F P
0 g
s
.80641 .62786 m .80565 .62855 L p .80572 .62788 L .832 .731 .753 r
F P
0 g
s
.81856 .65921 m .80577 .6601 L .80962 .65832 L closepath p .49 .752 .987 r
F P
0 g
s
.81856 .65921 m .81683 .66131 L .80577 .6601 L closepath p .469 .764 .994 r
F P
0 g
s
.8067 .61754 m .80685 .61748 L .80585 .62445 L .832 .731 .753 r
F
.80585 .62445 m .80685 .61748 L .80601 .62454 L F
.80657 .6176 m .8067 .61754 L .80585 .62445 L F
.80686 .61747 m .80688 .61746 L .80601 .62454 L F
.80601 .62454 m .80688 .61746 L .80609 .62459 L F
.80685 .61748 m .80686 .61747 L .80601 .62454 L F
.80689 .61746 m .8069 .61745 L .80609 .62459 L F
.80609 .62459 m .8069 .61745 L .80658 .62488 L F
.80688 .61746 m .80689 .61746 L .80609 .62459 L F
.80663 .62959 m .80641 .62786 L .80772 .62784 L .88 .965 .846 r
F
.80642 .65331 m .81094 .64726 L .81664 .65066 L closepath p .696 .646 .796 r
F P
0 g
s
.80642 .65331 m .81664 .65066 L .81257 .65638 L closepath p .689 .606 .758 r
F P
0 g
s
.80962 .65832 m .80642 .65331 L .81257 .65638 L closepath p .564 .641 .886 r
F P
0 g
s
.80695 .61743 m .807 .61741 L .80658 .62488 L .832 .731 .753 r
F
.80658 .62488 m .807 .61741 L .80669 .62494 L F
.8069 .61745 m .80695 .61743 L .80658 .62488 L F
.80778 .6308 m .80663 .62959 L .80772 .62784 L .88 .965 .846 r
F
.80668 .58206 m .81647 .57498 L .81679 .58135 L closepath p .849 .872 .857 r
F P
0 g
s
.80701 .61741 m .80702 .6174 L .80669 .62494 L .832 .731 .753 r
F
.80669 .62494 m .80702 .6174 L .80676 .62499 L F
.807 .61741 m .80701 .61741 L .80669 .62494 L F
.80703 .6174 m .80704 .6174 L .80676 .62499 L F
.80676 .62499 m .80704 .6174 L .8085 .626 L F
.80702 .6174 m .80703 .6174 L .80676 .62499 L F
.80704 .6174 m .80717 .61734 L .8085 .626 L F
.80717 .61734 m .80731 .61728 L .8085 .626 L F
.80731 .61728 m .80738 .61725 L .80932 .62528 L F
.80932 .62528 m .8085 .626 L p .80731 .61728 L F P
0 g
s
.80745 .61722 m .80932 .62528 L p .80738 .61725 L .832 .731 .753 r
F P
0 g
s
.8075 .56364 m .81148 .55719 L .80953 .56519 L closepath p .742 .5 .564 r
F P
0 g
s
.80778 .6308 m .80772 .62784 L .80861 .62782 L .88 .965 .846 r
F
.80837 .63143 m .80778 .6308 L .80861 .62782 L F
.80916 .55026 m .81148 .55719 L .80787 .55266 L closepath p .862 .663 .632 r
F P
0 g
s
.80788 .4894 m .81092 .48159 L .83332 .47812 L closepath p .626 .37 .511 r
F P
0 g
s
.80788 .4894 m .83332 .47812 L .8343 .48535 L closepath p .685 .492 .62 r
F P
0 g
s
.80792 .51989 m .83258 .50785 L .83456 .51464 L closepath p .723 .576 .688 r
F P
0 g
s
.80792 .51989 m .83456 .51464 L .82132 .53097 L closepath p .712 .529 .64 r
F P
0 g
s
.80953 .56519 m .81344 .56287 L .80819 .56993 L closepath p .893 .849 .794 r
F P
0 g
s
.80819 .56993 m .81344 .56287 L .8178 .56528 L closepath p .722 .567 .679 r
F P
0 g
s
.80819 .56993 m .8178 .56528 L .81274 .57489 L closepath p .718 .557 .671 r
F P
0 g
s
.80893 .62781 m .80837 .63143 L p .80861 .62782 L .88 .965 .846 r
F P
0 g
s
.80839 .60438 m .82205 .59031 L .83503 .59264 L closepath p .833 .873 .873 r
F P
0 g
s
.80839 .60438 m .83503 .59264 L .82566 .60558 L closepath p .83 .889 .884 r
F P
0 g
s
.8085 .626 m .80932 .62528 L p .80877 .62616 L .88 .965 .846 r
F P
0 g
s
.80903 .62631 m .80877 .62616 L .80932 .62528 L .88 .965 .846 r
F
.80885 .54724 m .80932 .54293 L .81535 .54996 L closepath p .794 .635 .684 r
F P
0 g
s
.80885 .54724 m .81535 .54996 L .81305 .55405 L closepath p .807 .609 .634 r
F P
0 g
s
.80885 .54724 m .81305 .55405 L .80916 .55026 L closepath p .924 .593 .396 r
F P
0 g
s
.80932 .62528 m .80915 .62638 L p .80903 .62631 L .88 .965 .846 r
F P
0 g
s
.80916 .55026 m .81305 .55405 L .81148 .55719 L closepath p .888 .673 .607 r
F P
0 g
s
.80932 .54293 m .81258 .54125 L .8185 .54659 L closepath p .793 .647 .7 r
F P
0 g
s
.80932 .54293 m .8185 .54659 L .81535 .54996 L closepath p .8 .635 .678 r
F P
0 g
s
.81305 .55405 m .81344 .56287 L .80953 .56519 L closepath p .789 .546 .57 r
F P
0 g
s
.81148 .55719 m .81305 .55405 L .80953 .56519 L closepath p .828 .59 .58 r
F P
0 g
s
.81856 .65921 m .80962 .65832 L .81257 .65638 L closepath p .592 .712 .927 r
F P
0 g
s
.81338 .47238 m .83332 .47812 L .81092 .48159 L closepath p .624 .365 .505 r
F P
0 g
s
.81664 .65066 m .81094 .64726 L .81205 .64108 L closepath p .795 .583 .614 r
F P
0 g
s
.81698 .64378 m .81205 .64108 L .81254 .63584 L closepath p .842 .527 .456 r
F P
0 g
s
.81664 .65066 m .81205 .64108 L .81698 .64378 L closepath p .827 .593 .585 r
F P
0 g
s
.81698 .64378 m .81254 .63584 L .81886 .63991 L closepath p .69 .489 .61 r
F P
0 g
s
.81257 .65638 m .81664 .65066 L .82248 .65347 L closepath p .671 .581 .745 r
F P
0 g
s
.81257 .65638 m .82248 .65347 L .82107 .65763 L closepath p .682 .633 .794 r
F P
0 g
s
.81856 .65921 m .81257 .65638 L .82107 .65763 L closepath p .617 .711 .913 r
F P
0 g
s
.81258 .54125 m .82132 .53097 L .82776 .54094 L closepath p .812 .677 .715 r
F P
0 g
s
.81258 .54125 m .82776 .54094 L .8185 .54659 L closepath p .811 .641 .672 r
F P
0 g
s
.8178 .56528 m .81647 .57498 L .81274 .57489 L closepath p .803 .666 .712 r
F P
0 g
s
.81305 .55405 m .81535 .54996 L .81941 .55603 L closepath p .669 .434 .556 r
F P
0 g
s
.81344 .56287 m .81305 .55405 L .81941 .55603 L closepath p .665 .449 .582 r
F P
0 g
s
.81338 .47238 m .81401 .47015 L .8321 .46999 L closepath p .458 0 .02 r
F P
0 g
s
.81338 .47238 m .8321 .46999 L .83236 .47291 L closepath p .517 .071 .129 r
F P
0 g
s
.81338 .47238 m .83236 .47291 L .83332 .47812 L closepath p .623 .372 .517 r
F P
0 g
s
.81344 .56287 m .81941 .55603 L .8178 .56528 L closepath p .737 .554 .645 r
F P
0 g
s
.81538 .46303 m .8321 .46999 L .81401 .47015 L closepath p .598 .287 .413 r
F P
0 g
s
.81535 .54996 m .8185 .54659 L .8215 .55552 L closepath p .728 .521 .611 r
F P
0 g
s
.81535 .54996 m .8215 .55552 L .81941 .55603 L closepath p .819 .45 .356 r
F P
0 g
s
.81538 .46303 m .83106 .46437 L .8321 .46999 L closepath p .606 .193 .233 r
F P
0 g
s
.8178 .56528 m .81979 .56804 L .81647 .57498 L closepath p .815 .678 .712 r
F P
0 g
s
.81647 .57498 m .81979 .56804 L .82339 .57271 L closepath p .939 .86 .736 r
F P
0 g
s
.81647 .57498 m .82339 .57271 L .81679 .58135 L closepath p .925 .915 .789 r
F P
0 g
s
.82248 .65347 m .81664 .65066 L .81698 .64378 L closepath p .735 .512 .59 r
F P
0 g
s
.81679 .58135 m .82339 .57271 L .82794 .57373 L closepath p .916 .884 .786 r
F P
0 g
s
.81679 .58135 m .82794 .57373 L .819 .58381 L closepath p .871 .987 .832 r
F P
0 g
s
.83063 .66137 m .81683 .66131 L .81856 .65921 L closepath p .424 .698 .978 r
F P
0 g
s
.83063 .66137 m .8318 .66291 L .81683 .66131 L closepath p .285 .72 .966 r
F P
0 g
s
.82407 .64463 m .81698 .64378 L .81886 .63991 L closepath p .676 .472 .602 r
F P
0 g
s
.82248 .65347 m .81698 .64378 L .82407 .64463 L closepath p .675 .493 .631 r
F P
0 g
s
.81941 .55603 m .81979 .56804 L .8178 .56528 L closepath p .863 .679 .651 r
F P
0 g
s
.8185 .54659 m .82776 .54094 L .83148 .55206 L closepath p .802 .622 .658 r
F P
0 g
s
.8185 .54659 m .83148 .55206 L .82535 .55436 L closepath p .807 .612 .639 r
F P
0 g
s
.8185 .54659 m .82535 .55436 L .8215 .55552 L closepath p .869 .59 .512 r
F P
0 g
s
.83063 .66137 m .81856 .65921 L .82107 .65763 L closepath p .586 .63 .862 r
F P
0 g
s
.82407 .64463 m .81886 .63991 L .82485 .64138 L closepath p .654 .475 .63 r
F P
0 g
s
.819 .58381 m .82794 .57373 L .83344 .57718 L closepath p .878 .971 .843 r
F P
0 g
s
.819 .58381 m .83344 .57718 L .82205 .59031 L closepath p .867 .979 .849 r
F P
0 g
s
.8215 .55552 m .82279 .56213 L .81941 .55603 L closepath p .9 .647 .547 r
F P
0 g
s
.81941 .55603 m .82279 .56213 L .81979 .56804 L closepath p .808 .635 .668 r
F P
0 g
s
.81979 .56804 m .82279 .56213 L .82553 .56297 L closepath p .961 .878 .707 r
F P
0 g
s
.81979 .56804 m .82553 .56297 L .82339 .57271 L closepath p .958 .851 .693 r
F P
0 g
s
.83285 .60684 m .82088 .61142 L .82566 .60558 L closepath p .732 .967 .935 r
F P
0 g
s
.83019 .65874 m .82107 .65763 L .82248 .65347 L closepath p .612 .557 .768 r
F P
0 g
s
.83063 .66137 m .82107 .65763 L .83019 .65874 L closepath p .568 .636 .879 r
F P
0 g
s
.82132 .53097 m .83456 .51464 L .82776 .54094 L closepath p .809 .677 .718 r
F P
0 g
s
.8215 .55552 m .82535 .55436 L .82553 .56297 L closepath p .91 .725 .64 r
F P
0 g
s
.8215 .55552 m .82553 .56297 L .82279 .56213 L closepath p .973 .687 .424 r
F P
0 g
s
.82205 .59031 m .83344 .57718 L .84279 .58206 L closepath p .85 .872 .857 r
F P
0 g
s
.82205 .59031 m .84279 .58206 L .83503 .59264 L closepath p .852 .86 .848 r
F P
0 g
s
.83009 .65398 m .82248 .65347 L .82407 .64463 L closepath p .642 .462 .624 r
F P
0 g
s
.83019 .65874 m .82248 .65347 L .83009 .65398 L closepath p .637 .553 .743 r
F P
0 g
s
.82339 .57271 m .82553 .56297 L .83232 .564 L closepath p .907 .791 .721 r
F P
0 g
s
.82339 .57271 m .83232 .564 L .82794 .57373 L closepath p .929 .866 .756 r
F P
0 g
s
.82361 .64108 m .83275 .64293 L .82485 .64138 L p .157 0 0 r
F P
0 g
s
.83167 .64671 m .82407 .64463 L .82485 .64138 L closepath p .64 .462 .626 r
F P
0 g
s
.83009 .65398 m .82407 .64463 L .83167 .64671 L closepath p .64 .461 .625 r
F P
0 g
s
.83167 .64671 m .82485 .64138 L .83275 .64293 L closepath p .67 .452 .58 r
F P
0 g
s
.82553 .56297 m .82535 .55436 L .83148 .55206 L closepath p .82 .653 .676 r
F P
0 g
s
.8343 .48535 m .83654 .49143 L .82552 .4959 L closepath p .752 .615 .707 r
F P
0 g
s
.82552 .4959 m .83654 .49143 L .83988 .50324 L closepath p .7 .482 .589 r
F P
0 g
s
.82552 .4959 m .83988 .50324 L .83258 .50785 L closepath p .681 .504 .64 r
F P
0 g
s
.82553 .56297 m .83148 .55206 L .83232 .564 L closepath p .909 .785 .712 r
F P
0 g
s
.82566 .60558 m .83503 .59264 L .84135 .59489 L closepath p .817 .864 .882 r
F P
0 g
s
.82566 .60558 m .84135 .59489 L .83544 .60374 L closepath p .817 .928 .907 r
F P
0 g
s
.83285 .60684 m .82566 .60558 L .83544 .60374 L closepath p .677 .956 .948 r
F P
0 g
s
.83456 .51464 m .8367 .53404 L .82776 .54094 L closepath p .784 .651 .715 r
F P
0 g
s
.83148 .55206 m .82776 .54094 L .8367 .53404 L closepath p .759 .6 .681 r
F P
0 g
s
.82794 .57373 m .83232 .564 L .83811 .56554 L closepath p .871 .778 .756 r
F P
0 g
s
.82794 .57373 m .83811 .56554 L .83344 .57718 L closepath p .882 .971 .839 r
F P
0 g
s
.83782 .6534 m .83009 .65398 L .83167 .64671 L closepath p .579 .402 .606 r
F P
0 g
s
.83745 .65753 m .83009 .65398 L .83782 .6534 L closepath p .593 .573 .801 r
F P
0 g
s
.83745 .65753 m .83019 .65874 L .83009 .65398 L closepath p .636 .552 .743 r
F P
0 g
s
.83787 .6608 m .83019 .65874 L .83745 .65753 L closepath p .642 .646 .838 r
F P
0 g
s
.83787 .6608 m .83063 .66137 L .83019 .65874 L closepath p .611 .667 .88 r
F P
0 g
s
.83063 .66137 m .83787 .6608 L .84182 .66179 L closepath p .382 .799 .976 r
F P
0 g
s
.83063 .66137 m .84182 .66179 L .8318 .66291 L closepath p .409 .808 .984 r
F P
0 g
s
.83106 .46437 m .84191 .46597 L .84227 .46851 L closepath p .51 0 0 r
F P
0 g
s
.83106 .46437 m .84227 .46851 L .8321 .46999 L closepath p .542 .149 .246 r
F P
0 g
s
.83148 .55206 m .8367 .53404 L .84024 .5507 L closepath p .818 .665 .693 r
F P
0 g
s
.83232 .564 m .83148 .55206 L .84024 .5507 L closepath p .823 .724 .756 r
F P
0 g
s
.83628 .64683 m .83167 .64671 L .83275 .64293 L closepath p .572 .354 .545 r
F P
0 g
s
.83782 .6534 m .83167 .64671 L .83628 .64683 L closepath p .578 .402 .607 r
F P
0 g
s
.8321 .46999 m .84227 .46851 L .84593 .47223 L closepath p .473 .026 .106 r
F P
0 g
s
.8321 .46999 m .84593 .47223 L .83236 .47291 L closepath p .478 .042 .132 r
F P
0 g
s
.83232 .564 m .84024 .5507 L .83811 .56554 L closepath p .865 .788 .773 r
F P
0 g
s
.83236 .47291 m .84593 .47223 L .84646 .47769 L closepath p .614 .404 .574 r
F P
0 g
s
.83236 .47291 m .84646 .47769 L .83332 .47812 L closepath p .624 .372 .516 r
F P
0 g
s
.83258 .50785 m .83988 .50324 L .84426 .51054 L closepath p .693 .526 .657 r
F P
0 g
s
.83258 .50785 m .84426 .51054 L .83456 .51464 L closepath p .685 .553 .7 r
F P
0 g
s
.83628 .64683 m .83275 .64293 L .83767 .64347 L closepath p .619 .356 .496 r
F P
0 g
s
.83285 .60684 m .83544 .60374 L .84149 .60111 L closepath p .685 .978 .887 r
F P
0 g
s
.83285 .60684 m .84149 .60111 L .83828 .6037 L closepath p .556 .923 .763 r
F P
0 g
s
.83332 .47812 m .84646 .47769 L .84739 .48411 L closepath p .641 .458 .619 r
F P
0 g
s
.83332 .47812 m .84739 .48411 L .8343 .48535 L closepath p .64 .46 .624 r
F P
0 g
s
.83811 .56554 m .84412 .56678 L .83344 .57718 L closepath p .834 .909 .889 r
F P
0 g
s
.83344 .57718 m .84412 .56678 L .84659 .57037 L closepath p .838 .945 .892 r
F P
0 g
s
.83344 .57718 m .84659 .57037 L .84726 .5725 L closepath p .757 .987 .897 r
F P
0 g
s
.83344 .57718 m .84726 .5725 L .84279 .58206 L closepath p .913 .838 .757 r
F P
0 g
s
.8343 .48535 m .84739 .48411 L .84856 .48901 L closepath p .644 .481 .648 r
F P
0 g
s
.8343 .48535 m .84856 .48901 L .83654 .49143 L closepath p .615 .542 .75 r
F P
0 g
s
.83456 .51464 m .84426 .51054 L .85109 .52436 L closepath p .698 .591 .732 r
F P
0 g
s
.83456 .51464 m .85109 .52436 L .8367 .53404 L closepath p .7 .59 .729 r
F P
0 g
s
.83503 .59264 m .84279 .58206 L .84536 .58501 L closepath p .88 .928 .85 r
F P
0 g
s
.83503 .59264 m .84536 .58501 L .84135 .59489 L closepath p .873 .833 .805 r
F P
0 g
s
.83544 .60374 m .84135 .59489 L .84723 .59207 L closepath p .839 .989 .861 r
F P
0 g
s
.83544 .60374 m .84723 .59207 L .84149 .60111 L closepath p .853 .933 .878 r
F P
0 g
s
.84346 .64697 m .83628 .64683 L .83767 .64347 L closepath p .501 .232 .439 r
F P
0 g
s
.84375 .65326 m .83628 .64683 L .84346 .64697 L closepath p .52 .384 .635 r
F P
0 g
s
.84375 .65326 m .83782 .6534 L .83628 .64683 L closepath p .546 .382 .61 r
F P
0 g
s
.83654 .49143 m .84856 .48901 L .85185 .49712 L closepath p .6 .471 .677 r
F P
0 g
s
.83654 .49143 m .85185 .49712 L .83988 .50324 L closepath p .622 .436 .61 r
F P
0 g
s
.8367 .53404 m .85109 .52436 L .85377 .53224 L closepath p .702 .595 .733 r
F P
0 g
s
.8367 .53404 m .85377 .53224 L .84024 .5507 L closepath p .702 .592 .73 r
F P
0 g
s
.84387 .65704 m .83745 .65753 L .83782 .6534 L closepath p .526 .514 .785 r
F P
0 g
s
.83787 .6608 m .83745 .65753 L .84387 .65704 L closepath p .524 .557 .831 r
F P
0 g
s
.84346 .64697 m .83767 .64347 L .84515 .64386 L closepath p .526 .209 .373 r
F P
0 g
s
.84387 .65704 m .83782 .6534 L .84375 .65326 L closepath p .551 .509 .763 r
F P
0 g
s
.83787 .6608 m .84387 .65704 L .84457 .65982 L closepath p .572 .675 .909 r
F P
0 g
s
.84457 .65982 m .84182 .66179 L .83787 .6608 L closepath p .412 .813 .983 r
F P
0 g
s
.83811 .56554 m .84024 .5507 L .85187 .55048 L closepath p .743 .674 .783 r
F P
0 g
s
.83811 .56554 m .85187 .55048 L .84412 .56678 L closepath p .852 .901 .87 r
F P
0 g
s
.83988 .50324 m .85185 .49712 L .85667 .50417 L closepath p .644 .482 .649 r
F P
0 g
s
.83988 .50324 m .85667 .50417 L .84426 .51054 L closepath p .644 .51 .685 r
F P
0 g
s
.85377 .53224 m .85187 .55048 L .84024 .5507 L closepath p .745 .655 .761 r
F P
0 g
s
.84135 .59489 m .84976 .58487 L .84723 .59207 L closepath p .919 .836 .748 r
F P
0 g
s
.84135 .59489 m .84536 .58501 L .84976 .58487 L closepath p .939 .933 .767 r
F P
0 g
s
.84457 .65982 m .84839 .66075 L .84182 .66179 L closepath p .388 .768 .991 r
F P
0 g
s
.85042 .46666 m .84191 .46597 L .85135 .46542 L closepath p .286 0 0 r
F P
0 g
s
.84191 .46597 m .85042 .46666 L .84227 .46851 L closepath p .382 0 0 r
F P
0 g
s
.84227 .46851 m .85042 .46666 L .84593 .47223 L closepath p .476 .026 .101 r
F P
0 g
s
.84279 .58206 m .84726 .5725 L .84883 .57618 L closepath p .954 .952 .693 r
F P
0 g
s
.84279 .58206 m .84883 .57618 L .84536 .58501 L closepath p .978 .889 .67 r
F P
0 g
s
.84346 .64697 m .84515 .64386 L .8667 .65636 L closepath p .615 .306 .424 r
F P
0 g
s
.8667 .65636 m .84375 .65326 L .84346 .64697 L closepath p .562 .416 .64 r
F P
0 g
s
.8667 .65636 m .84387 .65704 L .84375 .65326 L closepath p .541 .502 .761 r
F P
0 g
s
.8667 .65636 m .84457 .65982 L .84387 .65704 L closepath p .512 .633 .906 r
F P
0 g
s
.84412 .56678 m .8609 .54987 L .86025 .55539 L closepath p .838 .981 .874 r
F P
0 g
s
.85187 .55048 m .8609 .54987 L .84412 .56678 L closepath p .828 .868 .875 r
F P
0 g
s
.8533 .5603 m .84659 .57037 L .84412 .56678 L p .962 .93 .712 r
F P
0 g
s
.84426 .51054 m .85667 .50417 L .86408 .51341 L closepath p .669 .57 .734 r
F P
0 g
s
.84426 .51054 m .86408 .51341 L .85109 .52436 L closepath p .667 .577 .745 r
F P
0 g
s
.84839 .66075 m .84457 .65982 L .8667 .65636 L closepath p .423 .78 .996 r
F P
0 g
s
.84515 .64386 m .849 .64556 L .8667 .65636 L closepath p .49 0 0 r
F P
0 g
s
.84536 .58501 m .84883 .57618 L .85124 .57826 L closepath p .963 .855 .687 r
F P
0 g
s
.84536 .58501 m .85124 .57826 L .84976 .58487 L closepath p .967 .882 .696 r
F P
0 g
s
.85042 .46666 m .85165 .47072 L .84593 .47223 L closepath p .3 0 0 r
F P
0 g
s
.84593 .47223 m .85165 .47072 L .85064 .47641 L closepath p .589 .392 .583 r
F P
0 g
s
.84593 .47223 m .85064 .47641 L .84646 .47769 L closepath p .596 .391 .574 r
F P
0 g
s
.84646 .47769 m .85064 .47641 L .84739 .48411 L closepath p .615 .439 .621 r
F P
0 g
s
.85174 .56376 m .84659 .57037 L .8533 .5603 L p .708 .881 .488 r
F P
0 g
s
.84659 .57037 m .85174 .56376 L .84726 .5725 L closepath p .915 .921 .535 r
F P
0 g
s
.84726 .5725 m .85174 .56376 L .84992 .56844 L closepath p .75 .851 .409 r
F P
0 g
s
.84726 .5725 m .84992 .56844 L .84883 .57618 L closepath p .962 .816 .385 r
F P
0 g
s
.85064 .47641 m .85482 .48201 L .84739 .48411 L closepath p .513 .328 .568 r
F P
0 g
s
.84739 .48411 m .85482 .48201 L .86427 .4867 L closepath p .564 .482 .72 r
F P
0 g
s
.84739 .48411 m .86427 .4867 L .84856 .48901 L closepath p .575 .437 .656 r
F P
0 g
s
.84856 .48901 m .86427 .4867 L .86413 .49315 L closepath p .59 .529 .754 r
F P
0 g
s
.84856 .48901 m .86413 .49315 L .85185 .49712 L closepath p .618 .48 .672 r
F P
0 g
s
.85124 .57826 m .84883 .57618 L .84992 .56844 L closepath p .989 .791 .545 r
F P
0 g
s
.85124 .57826 m .84992 .56844 L .85044 .57204 L closepath p .719 .511 0 r
F P
0 g
s
.85042 .46666 m .86332 .46374 L .86409 .46641 L closepath p .372 0 .046 r
F P
0 g
s
.85042 .46666 m .85135 .46542 L .86332 .46374 L closepath p .117 0 0 r
F P
0 g
s
.85042 .46666 m .86409 .46641 L .85606 .4679 L closepath p .193 0 0 r
F P
0 g
s
.85042 .46666 m .85606 .4679 L .85165 .47072 L closepath p .185 0 .026 r
F P
0 g
s
.85087 .57396 m .85124 .57826 L .85044 .57204 L closepath p .802 .569 .006 r
F P
0 g
s
.85165 .47072 m .85404 .47322 L .85064 .47641 L closepath p .391 .214 .517 r
F P
0 g
s
.85064 .47641 m .85404 .47322 L .85482 .48201 L closepath p .462 .316 .596 r
F P
0 g
s
.85109 .52436 m .86408 .51341 L .86967 .51951 L closepath p .689 .618 .771 r
F P
0 g
s
.85109 .52436 m .86967 .51951 L .85377 .53224 L closepath p .682 .584 .739 r
F P
0 g
s
.85165 .47072 m .85606 .4679 L .86002 .47012 L closepath p .502 .221 .42 r
F P
0 g
s
.85165 .47072 m .86002 .47012 L .85404 .47322 L closepath p .501 .219 .417 r
F P
0 g
s
.85365 .56006 m .85174 .56376 L p .8533 .5603 L .708 .881 .488 r
F P
0 g
s
.85185 .49712 m .86413 .49315 L .86829 .49996 L closepath p .595 .418 .612 r
F P
0 g
s
.85185 .49712 m .86829 .49996 L .85667 .50417 L closepath p .586 .465 .682 r
F P
0 g
s
.85377 .53224 m .8601 .53936 L .85187 .55048 L closepath p .738 .649 .761 r
F P
0 g
s
.85187 .55048 m .8601 .53936 L .86472 .54511 L closepath p .846 .826 .828 r
F P
0 g
s
.85187 .55048 m .86472 .54511 L .8609 .54987 L closepath p .846 .838 .838 r
F P
0 g
s
.85377 .53224 m .86967 .51951 L .87587 .52497 L closepath p .727 .673 .796 r
F P
0 g
s
.85377 .53224 m .87587 .52497 L .8601 .53936 L closepath p .722 .647 .773 r
F P
0 g
s
.85404 .47322 m .86002 .47012 L .86374 .47108 L closepath p .538 .28 .473 r
F P
0 g
s
.85404 .47322 m .86374 .47108 L .86305 .47488 L closepath p .536 .277 .469 r
F P
0 g
s
.86427 .47857 m .85404 .47322 L .86305 .47488 L closepath p .534 .322 .538 r
F P
0 g
s
.85404 .47322 m .86427 .47857 L .85482 .48201 L closepath p .504 .347 .6 r
F P
0 g
s
.85482 .48201 m .86427 .47857 L .86427 .4867 L closepath p .557 .486 .731 r
F P
0 g
s
.85606 .4679 m .86409 .46641 L .86668 .46939 L closepath p .734 .446 .491 r
F P
0 g
s
.85606 .4679 m .86668 .46939 L .86002 .47012 L closepath p .56 .03 0 r
F P
0 g
s
.85667 .50417 m .86829 .49996 L .87403 .50776 L closepath p .623 .58 .785 r
F P
0 g
s
.85667 .50417 m .87403 .50776 L .86408 .51341 L closepath p .634 .561 .756 r
F P
0 g
s
.86002 .47012 m .86668 .46939 L .86374 .47108 L closepath p .426 0 0 r
F P
0 g
s
.8601 .53936 m .87587 .52497 L .87847 .53004 L closepath p .822 .877 .885 r
F P
0 g
s
.8601 .53936 m .87847 .53004 L .86472 .54511 L closepath p .822 .823 .85 r
F P
0 g
s
.8609 .54987 m .8687 .5445 L .86025 .55539 L closepath p .877 .993 .795 r
F P
0 g
s
.8609 .54987 m .86472 .54511 L .8687 .5445 L closepath p .859 .997 .794 r
F P
0 g
s
.86374 .47108 m .87244 .47469 L .86305 .47488 L closepath p .6 .334 .487 r
F P
0 g
s
.86305 .47488 m .87244 .47469 L .87283 .47765 L closepath p .602 .343 .497 r
F P
0 g
s
.86305 .47488 m .87283 .47765 L .86427 .47857 L closepath p .6 .358 .522 r
F P
0 g
s
.86332 .46374 m .87175 .462 L .87286 .46479 L closepath p .414 0 .129 r
F P
0 g
s
.86332 .46374 m .87286 .46479 L .86409 .46641 L closepath p .394 0 .04 r
F P
0 g
s
.86668 .46939 m .87201 .47089 L .86374 .47108 L closepath p .238 0 0 r
F P
0 g
s
.86374 .47108 m .87201 .47089 L .87244 .47469 L closepath p .589 .359 .535 r
F P
0 g
s
.86967 .51951 m .86408 .51341 L .87403 .50776 L closepath p .655 .612 .794 r
F P
0 g
s
.86409 .46641 m .87286 .46479 L .876 .46835 L closepath p .622 .393 .549 r
F P
0 g
s
.86409 .46641 m .876 .46835 L .86668 .46939 L closepath p .619 .444 .624 r
F P
0 g
s
.86427 .4867 m .87402 .4956 L .86413 .49315 L closepath p .499 .452 .735 r
F P
0 g
s
.86413 .49315 m .87402 .4956 L .87829 .50326 L closepath p .478 .48 .781 r
F P
0 g
s
.86413 .49315 m .87829 .50326 L .86829 .49996 L closepath p .667 .441 .568 r
F P
0 g
s
.86427 .47857 m .87209 .48469 L .86427 .4867 L closepath p .598 .519 .737 r
F P
0 g
s
.86427 .4867 m .87209 .48469 L .87402 .4956 L closepath p .581 .457 .676 r
F P
0 g
s
.86427 .47857 m .87283 .47765 L .87209 .48469 L closepath p .635 .515 .7 r
F P
0 g
s
.86472 .54511 m .88371 .52979 L .8687 .5445 L closepath p .779 .998 .826 r
F P
0 g
s
.86472 .54511 m .87847 .53004 L .88371 .52979 L closepath p .861 .964 .867 r
F P
0 g
s
.86668 .46939 m .876 .46835 L .88041 .47033 L closepath p .33 0 0 r
F P
0 g
s
.86668 .46939 m .88041 .47033 L .87201 .47089 L closepath p .238 0 0 r
F P
0 g
s
.87403 .50776 m .86829 .49996 L .87829 .50326 L closepath p .584 .567 .8 r
F P
0 g
s
.86967 .51951 m .87403 .50776 L .87994 .51399 L closepath p .676 .635 .802 r
F P
0 g
s
.86967 .51951 m .87994 .51399 L .88296 .51991 L closepath p .7 .709 .855 r
F P
0 g
s
.86967 .51951 m .88296 .51991 L .87587 .52497 L closepath p .714 .672 .808 r
F P
0 g
s
.87175 .462 m .88309 .46526 L .87286 .46479 L closepath p .5 .039 .09 r
F P
0 g
s
.87175 .462 m .88168 .46121 L .88309 .46526 L closepath p .507 .073 .151 r
F P
0 g
s
.88041 .47033 m .87244 .47469 L .87201 .47089 L closepath p .659 .41 .531 r
F P
0 g
s
.88422 .47617 m .88221 .48965 L .87209 .48469 L closepath p .653 .484 .642 r
F P
0 g
s
.87283 .47765 m .88422 .47617 L .87209 .48469 L closepath p .702 .575 .71 r
F P
0 g
s
.87209 .48469 m .88214 .49492 L .87402 .4956 L closepath p .661 .511 .67 r
F P
0 g
s
.87209 .48469 m .88221 .48965 L .88214 .49492 L closepath p .624 .509 .703 r
F P
0 g
s
.88041 .47033 m .88028 .47471 L .87244 .47469 L closepath p .605 .316 .452 r
F P
0 g
s
.87244 .47469 m .88028 .47471 L .87283 .47765 L closepath p .614 .352 .496 r
F P
0 g
s
.87283 .47765 m .88028 .47471 L .88422 .47617 L closepath p .694 .527 .658 r
F P
0 g
s
.87286 .46479 m .88379 .46836 L .876 .46835 L closepath p .624 .393 .547 r
F P
0 g
s
.87286 .46479 m .88309 .46526 L .88379 .46836 L closepath p .618 .411 .579 r
F P
0 g
s
.87402 .4956 m .88214 .49492 L .88484 .50077 L closepath p .664 .539 .701 r
F P
0 g
s
.87402 .4956 m .88484 .50077 L .87829 .50326 L closepath p .642 .554 .74 r
F P
0 g
s
.87829 .50326 m .87994 .51399 L .87403 .50776 L closepath p .62 .624 .833 r
F P
0 g
s
.87847 .53004 m .87587 .52497 L .88296 .51991 L closepath p .776 .86 .912 r
F P
0 g
s
.876 .46835 m .89176 .47004 L .88041 .47033 L closepath p .344 0 0 r
F P
0 g
s
.876 .46835 m .88379 .46836 L .89176 .47004 L closepath p .461 0 .021 r
F P
0 g
s
.87829 .50326 m .88484 .50077 L .89076 .50905 L closepath p .668 .644 .817 r
F P
0 g
s
.87829 .50326 m .8831 .51152 L .87994 .51399 L closepath p .583 .597 .832 r
F P
0 g
s
.87829 .50326 m .89076 .50905 L .8831 .51152 L closepath p .679 .639 .803 r
F P
0 g
s
.87847 .53004 m .88296 .51991 L .88617 .52327 L closepath p .815 .908 .904 r
F P
0 g
s
.88617 .52327 m .88371 .52979 L .87847 .53004 L closepath p .807 .977 .899 r
F P
0 g
s
.87994 .51399 m .8831 .51152 L .88712 .51576 L closepath p .656 .753 .921 r
F P
0 g
s
.87994 .51399 m .88712 .51576 L .88296 .51991 L closepath p .718 .718 .849 r
F P
0 g
s
.88028 .47471 m .89176 .47004 L .89392 .47525 L closepath p .72 .44 .502 r
F P
0 g
s
.88028 .47471 m .88041 .47033 L .89176 .47004 L closepath p .694 .39 .457 r
F P
0 g
s
.88028 .47471 m .89392 .47525 L .88422 .47617 L closepath p .721 .445 .508 r
F P
0 g
s
.88309 .46526 m .88168 .46121 L .89174 .46408 L closepath p .652 .164 .079 r
F P
0 g
s
.88214 .49492 m .88221 .48965 L .89276 .49151 L closepath p .853 .699 .691 r
F P
0 g
s
.88214 .49492 m .89276 .49151 L .88886 .49493 L closepath p .865 .835 .815 r
F P
0 g
s
.88484 .50077 m .88214 .49492 L .88886 .49493 L closepath p .875 .617 .545 r
F P
0 g
s
.88422 .47617 m .88817 .48168 L .88221 .48965 L closepath p .661 .491 .644 r
F P
0 g
s
.89367 .48898 m .88221 .48965 L .88817 .48168 L closepath p .851 .778 .781 r
F P
0 g
s
.89367 .48898 m .89276 .49151 L .88221 .48965 L closepath p .855 .685 .671 r
F P
0 g
s
.88296 .51991 m .88712 .51576 L .89164 .52009 L closepath p .786 .863 .907 r
F P
0 g
s
.88296 .51991 m .89164 .52009 L .88617 .52327 L closepath p .734 .889 .951 r
F P
0 g
s
.88309 .46526 m .89174 .46408 L .89547 .46703 L closepath p .618 .112 .027 r
F P
0 g
s
.88379 .46836 m .88309 .46526 L .89547 .46703 L closepath p .765 .504 .539 r
F P
0 g
s
.8831 .51152 m .89076 .50905 L .89555 .5112 L closepath p .69 .779 .919 r
F P
0 g
s
.8831 .51152 m .89555 .5112 L .88712 .51576 L closepath p .7 .766 .903 r
F P
0 g
s
.88617 .52327 m .89057 .52464 L .88371 .52979 L closepath p .776 .947 .933 r
F P
0 g
s
.88379 .46836 m .88424 .46831 L p .89176 .47004 L .722 .386 .41 r
F P
0 g
s
.89392 .47525 m .88817 .48168 L .88422 .47617 L closepath p .739 .506 .576 r
F P
0 g
s
.88424 .46831 m .89547 .46703 L .89726 .4713 L p .722 .386 .41 r
F P
0 g
s
.89116 .49783 m .88484 .50077 L .88886 .49493 L closepath p .426 .062 .242 r
F P
0 g
s
.88484 .50077 m .89116 .49783 L .89788 .50599 L closepath p .732 .653 .771 r
F P
0 g
s
.88484 .50077 m .89788 .50599 L .89076 .50905 L closepath p .722 .66 .788 r
F P
0 g
s
.88617 .52327 m .89164 .52009 L .89567 .52112 L closepath p .682 .962 .94 r
F P
0 g
s
.88617 .52327 m .89567 .52112 L .89057 .52464 L closepath p .799 .947 .919 r
F P
0 g
s
.88712 .51576 m .89555 .5112 L .89944 .51486 L closepath p .706 .89 .964 r
F P
0 g
s
.88712 .51576 m .89944 .51486 L .89164 .52009 L closepath p .759 .858 .923 r
F P
0 g
s
.89392 .47525 m .89634 .47981 L .88817 .48168 L closepath p .696 .445 .539 r
F P
0 g
s
.88817 .48168 m .89634 .47981 L .89822 .48782 L closepath p .762 .741 .834 r
F P
0 g
s
.88817 .48168 m .89822 .48782 L .89367 .48898 L closepath p .71 .748 .882 r
F P
0 g
s
.89276 .49151 m .89116 .49783 L .88886 .49493 L closepath p .382 .055 .281 r
F P
0 g
s
.89057 .52464 m .89567 .52112 L .89748 .52053 L closepath p .539 .906 .719 r
F P
0 g
s
.89076 .50905 m .89788 .50599 L .90156 .50612 L closepath p .747 .808 .9 r
F P
0 g
s
.89076 .50905 m .90156 .50612 L .89555 .5112 L closepath p .752 .768 .865 r
F P
0 g
s
.89116 .49783 m .89977 .49082 L .90224 .50026 L closepath p .75 .599 .689 r
F P
0 g
s
.89276 .49151 m .89977 .49082 L .89116 .49783 L closepath p .392 .063 .284 r
F P
0 g
s
.89116 .49783 m .90224 .50026 L .89788 .50599 L closepath p .728 .652 .774 r
F P
0 g
s
.89164 .52009 m .89944 .51486 L .90446 .51684 L closepath p .756 .937 .944 r
F P
0 g
s
.89164 .52009 m .90446 .51684 L .89567 .52112 L closepath p .635 .948 .948 r
F P
0 g
s
.89176 .47004 m .89726 .4713 L .90049 .47365 L closepath p .577 0 0 r
F P
0 g
s
.89176 .47004 m .90049 .47365 L .90228 .47795 L closepath p .806 .426 .337 r
F P
0 g
s
.89176 .47004 m .90228 .47795 L .89392 .47525 L closepath p .792 .472 .447 r
F P
0 g
s
.89367 .48898 m .89977 .49082 L .89276 .49151 L closepath p .451 .248 .512 r
F P
0 g
s
.89367 .48898 m .89822 .48782 L .90929 .49228 L closepath p .686 .553 .699 r
F P
0 g
s
.89367 .48898 m .90929 .49228 L .89977 .49082 L closepath p .309 0 0 r
F P
0 g
s
.89392 .47525 m .90292 .48427 L .89634 .47981 L closepath p .853 .484 .352 r
F P
0 g
s
.89392 .47525 m .90228 .47795 L .90292 .48427 L closepath p .793 .544 .561 r
F P
0 g
s
.89542 .46699 m .89546 .467 L p .89542 .46699 L .443 0 0 r
F P
0 g
s
.89546 .467 m .89543 .467 L .89542 .46699 L .443 0 0 r
F
.89543 .467 m .89543 .467 L .89546 .467 L F
.89544 .467 m .89543 .467 L .89546 .467 L F
.89547 .46703 m .89544 .467 L .89546 .467 L F
.89546 .467 m .89544 .467 L .89546 .467 L F
.89546 .467 m .89546 .467 L .89546 .467 L .45 .691 .969 r
F
.89546 .467 m .89546 .467 L p .89546 .467 L .443 0 0 r
F P
0 g
s
.89546 .467 m .89666 .46735 L .89559 .46723 L p .443 0 0 r
F P
0 g
s
.89559 .46723 m .89547 .46703 L .89546 .467 L .443 0 0 r
F
.89726 .4713 m .89547 .46703 L .89775 .47106 L closepath p .727 .351 0 r
F P
0 g
s
.89555 .5112 m .90156 .50612 L .90567 .50976 L closepath p .796 .918 .92 r
F P
0 g
s
.89555 .5112 m .90567 .50976 L .89944 .51486 L closepath p .802 .914 .915 r
F P
0 g
s
.89559 .46723 m .89666 .46735 L p .89679 .46936 L .739 .235 .028 r
F P
0 g
s
.89567 .52112 m .90446 .51684 L .90409 .51748 L closepath p .431 .865 .768 r
F P
0 g
s
.89567 .52112 m .90409 .51748 L .89748 .52053 L closepath p .343 .806 .7 r
F P
0 g
s
.89634 .47981 m .90292 .48427 L .89822 .48782 L closepath p .64 .663 .857 r
F P
0 g
s
.89666 .46735 m .89775 .47106 L p .89679 .46936 L .739 .235 .028 r
F P
0 g
s
.89775 .47106 m .89666 .46735 L .90263 .47178 L closepath p .761 .258 .01 r
F P
0 g
s
.89726 .4713 m .89775 .47106 L .89943 .47287 L p .552 .022 0 r
F P
0 g
s
.89943 .47287 m .89775 .47106 L .9002 .47343 L .861 .505 .375 r
F
.89775 .47106 m .90026 .47309 L p .90035 .47355 L F P
0 g
s
.9002 .47343 m .89775 .47106 L .90035 .47355 L .861 .505 .375 r
F
.9061 .47779 m .89775 .47106 L .90263 .47178 L closepath p .844 .4 .165 r
F P
0 g
s
.89788 .50599 m .90224 .50026 L .90738 .50051 L closepath p .796 .76 .82 r
F P
0 g
s
.89788 .50599 m .90738 .50051 L .90156 .50612 L closepath p .789 .737 .804 r
F P
0 g
s
.89822 .48782 m .90292 .48427 L .90929 .49228 L closepath p .613 .609 .823 r
F P
0 g
s
.89944 .51486 m .90567 .50976 L .91084 .51205 L closepath p .802 .935 .918 r
F P
0 g
s
.89944 .51486 m .91084 .51205 L .90446 .51684 L closepath p .795 .94 .923 r
F P
0 g
s
.89977 .49082 m .90929 .49228 L .90224 .50026 L closepath p .832 .645 .65 r
F P
0 g
s
.90035 .47355 m .90026 .47309 L .90049 .47365 L .861 .505 .375 r
F
.90026 .47309 m .9061 .47779 L .9039 .4776 L p F P
0 g
s
.90049 .47365 m .90026 .47309 L .9039 .4776 L .861 .505 .375 r
F
.90228 .47795 m .90049 .47365 L .90413 .47787 L closepath p .864 .436 .01 r
F P
0 g
s
.90156 .50612 m .90738 .50051 L .91011 .50184 L closepath p .823 .998 .845 r
F P
0 g
s
.90156 .50612 m .91011 .50184 L .90567 .50976 L closepath p .927 .91 .784 r
F P
0 g
s
.90224 .50026 m .90929 .49228 L .90738 .50051 L closepath p .816 .619 .636 r
F P
0 g
s
.90228 .47795 m .90413 .47787 L .90625 .48226 L closepath p .857 .421 0 r
F P
0 g
s
.90292 .48427 m .90228 .47795 L .90625 .48226 L closepath p .966 .678 .432 r
F P
0 g
s
.9061 .47779 m .90263 .47178 L .90597 .47706 L closepath p .707 .27 0 r
F P
0 g
s
.90292 .48427 m .90625 .48226 L .90929 .49228 L closepath p .924 .575 .353 r
F P
0 g
s
.9039 .4776 m .90599 .47778 L p .90413 .47787 L .797 .359 .206 r
F P
0 g
s
.90446 .51684 m .9096 .51346 L .90409 .51748 L closepath p .5 .879 .682 r
F P
0 g
s
.90413 .47787 m .90599 .47778 L .90413 .47788 L .797 .359 .206 r
F
.90599 .47778 m .9061 .47779 L p .90413 .47788 L F P
0 g
s
.9061 .47779 m .90507 .47983 L p .90413 .47788 L .797 .359 .206 r
F P
0 g
s
.90446 .51684 m .91084 .51205 L .9096 .51346 L closepath p .618 .948 .769 r
F P
0 g
s
.90507 .47983 m .9061 .47779 L p .90566 .48105 L .788 .347 .201 r
F P
0 g
s
.9061 .47779 m .90934 .48132 L .90566 .48105 L p .788 .347 .201 r
F P
0 g
s
.90566 .48105 m .90934 .48132 L p .90625 .48226 L .881 .578 .47 r
F P
0 g
s
.90567 .50976 m .91011 .50184 L .91586 .50599 L closepath p .883 .826 .786 r
F P
0 g
s
.90567 .50976 m .91586 .50599 L .91084 .51205 L closepath p .856 .93 .875 r
F P
0 g
s
.90934 .48132 m .9061 .47779 L .90597 .47706 L closepath p .658 .137 0 r
F P
0 g
s
.90934 .48132 m .90597 .47706 L .91086 .48051 L closepath p .711 .177 0 r
F P
0 g
s
.90654 .48325 m .90625 .48226 L .90802 .48358 L .881 .578 .47 r
F
.90934 .48132 m .90802 .48358 L p .90625 .48226 L F P
0 g
s
.90802 .48358 m .90712 .48513 L p .90654 .48325 L .881 .578 .47 r
F P
0 g
s
.90712 .48513 m .90934 .48132 L p .90751 .48644 L .855 .54 .454 r
F P
0 g
s
.90738 .50051 m .90929 .49228 L .90987 .49509 L closepath p .889 .827 .356 r
F P
0 g
s
.90738 .50051 m .90987 .49509 L .91011 .50184 L closepath p .985 .755 .359 r
F P
0 g
s
.90934 .48132 m .91235 .48629 L .90751 .48644 L p .855 .54 .454 r
F P
0 g
s
.90751 .48644 m .91235 .48629 L p .908 .48805 L .869 .586 .506 r
F P
0 g
s
.90807 .48828 m .908 .48805 L .91235 .48629 L .869 .586 .506 r
F
.91235 .48629 m .91654 .49701 L .90807 .48828 L p F P
0 g
s
.90807 .48828 m .9098 .49006 L p .90895 .49116 L .888 .554 .41 r
F P
0 g
s
.9098 .49006 m .90929 .49228 L .90895 .49116 L .888 .554 .41 r
F
.90952 .4934 m .90929 .49228 L .9098 .49006 L F
.91235 .48629 m .90934 .48132 L .91086 .48051 L closepath p .887 .458 .097 r
F P
0 g
s
.9098 .49006 m .91654 .49701 L p .90952 .4934 L .888 .554 .41 r
F P
0 g
s
.91654 .49701 m .90987 .49509 L p .90952 .4934 L .888 .554 .41 r
F P
0 g
s
.91084 .51205 m .91437 .50807 L .9096 .51346 L closepath p .531 .816 .469 r
F P
0 g
s
.91011 .50184 m .90987 .49509 L .91654 .49701 L closepath p .903 .678 .588 r
F P
0 g
s
.91011 .50184 m .91654 .49701 L .91586 .50599 L closepath p .94 .779 .655 r
F P
0 g
s
.91084 .51205 m .91586 .50599 L .91437 .50807 L closepath p .617 .873 .532 r
F P
0 g
s
.91235 .48629 m .91086 .48051 L .91416 .4858 L closepath p .887 .461 .086 r
F P
0 g
s
.91654 .49701 m .91235 .48629 L .91416 .4858 L closepath p .933 .562 .174 r
F P
0 g
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[44]:=",
  ImageSize->{315.438, 241.688},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004k0000l@P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7
?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7
?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7
?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00?l7?0L00:X7
102=1`00Y`L30004R`00ZP@0R@L00:H70`001X/0ZP2A[`D0Q`L00:H700L0T@2A0:ZA00@000?:0000
Q`L00:D71@000jnZ[`060003bP0708D7002U1`8000>AS0000P001I40[`2_008000?:00L0Q@L00:D7
0`00194009420003a`0000@0QPL00:D710000jX0[`060003[`0708H7002U1`8000NA0940ZP2_0080
00@70000Q`L00:D71@001Y40[jl0[`8000>_00L0QPL00:D70P000h`000050004[`0008P7002U1`03
078000:A0`000jl0000208T7002U1`050940ZY4010001<H000291`00Y@L60003/00000<0RPL00:D7
10000i40[`0200031`0008X7002T1`<000>_00001P2;1`00XPL;0003[`0708/7002S1`@000BZT02;
100D1`<0M@L00:<710000i40T00401471@001:X0001c1`00X`L30003Z@00008000?600003`L20008
T02Z0:Z@ZY030747002S1`<000>_00000P000jl0000>1`0908/0R`2@0:VZ00<000C:[`00K`L003D7
101Z1`X03`L20003R`0000@000BZZ@2_0P000lX01`1]1`00<PL30004R`00[`@0IPL50003ZP000080
3`L00`2;00040003ZZnZ00@000<7[`00KPL003870P001X^A0:X0ZPH0H`L20003ZP0000<000?600L0
3PL00`2@00030003ZP00008000?:[`000P1_1`00<@L2000;R`00[jX0T@2A0:l00`1R1`8000>Z0000
1@0?1`8000BZ0:nZ0P000jnZ00040707000a1`D000NAZP00[`2_008000?:00L0H@L70003ZP0700h7
0P000i4000020005ZJX00:l00`1a1`00<@L200:A0`000jl0a`020004[`0006870P001:T0094300l7
0P001H/0002Y00L0LPL003470`0017<00:l40004[`0006<71`000jT01`0>1`<000>;0:X00P000jX0
000207<7000a1`0709400940T@0600031`0706871`000lH01`0>1`@000BZ[jZ_101d1`00<@L01`2<
T@2Z0:l010001:l0001R1`0308/0008000>Y00000P0>1`<000BA0:ZZ1@1e1`00<@L50006/0370<N_
0`1S1`030780008000>Y00000P0;1`<000>Z00001`000`L01`1d1`00<0L30003T@0000D000B_0000
H`L00`2;00020003ZZT000802`L20005T8/009000`001:nZ001f1`00<0L40004ZP00ZPH0I0L:00`7
0`001JT0002Y00@0M`L003070P000g80T@0806D700<0R`000`001:T0000<1`<000>Y00000P000jT0
001h1`00<@L50004/:l0[`<0I@L40003ZP2Z00<030L00`1b00020003Z@0000<0N0L003470`000i40
000506D70`000g80R`020003Z@0000`70`000jX0ZP0407T7000^1`D000BA002Z0P000jl01`1U1`L0
00C1000030L:07T7000^1`X000>_0000IPL30003R`2;00H02`L40003Z@0000<0N@L002h700<0R`00
0P000i40T00406H70`000h/000020005aZT0Z@002`L00`2;00020004ZP0007X7000^1`@000>Y08/0
0P000lH0000H1`H0B0L30003R`00008000>Y00000P0;1`T0NPL002h710001Zl00036[`805PL20006
ZP2Z0:ZY10141`H000BZ002Y0`0;1`0508/00:l00`1k1`00;PL00`2A000801L70`001:X0ZY030003
[lH00080@0L200:;00<0LP000`001:T0000;1`<000>Z0:X00P1k1`00;PL80003aP0001H700<0T@00
1@001JZ_0<J_008000<700L0?PL20003R`00008000>Z0<H00P000jT01`0:1`@000BZZ@00N`L002l7
00<0T@001`0F1`H000FZT:X0aP030003bZl003l70`000h/000030003ZJX000802PL50004T00007/7
000_1`060:l0[`2A100F1`030940008000>_00000`000`N_000204070P000h^A00040003aP000080
2PL40004ZJX007`7000_1`T05PL40008ZJX0096Z0<H404470`000h/0R`020003ZP0000<02PL40004
ZY0007`7000_1`<000>Z09400`0F1`8000JZ0:nZ[jX30004[`00048700<0T@0010001JZ@00360080
2PL30003ZP000080O0L002l70`000h/0R`0301H700<0T@002`121`8000>;00000`001H^Z002Y00<0
2@L30003T:X00080O0L002l710000h/0000201H700D0R`00[`020003Z@0000<0@`L20004TH/0R`80
00FZT02YaP0300T70P001:VZ0:X307/7000_1`@000>Y0:X00P0E1`<000>;0:X00ZX404D71@0018/0
ZX/30004Z@0000P70`001jX0ZX/0Z@00N`L002l70P001G80R`2Y00805@L20005R`2Z[jX010161`80
00>;08/01P001JT00<H000P710001ZZ@003107/7000_1`H000>Z00004PL70003ZP00008000>_00L0
A@L40004R`00T0800ZT400L70P001G80ZX^Z008000>Y0000NPL002l70P000g80T00401870P001:X0
08/40004aP0004H700<0K@000`000jT0ZP050003aP0700H71P001900ZL4207X7000_1`8000>;0000
100C1`8000JYR`00ZZT304L70P0018/008D20007Z@000:T0a@0200L70`000i2ZR`0607X7000_1`80
00=b0000100C1`8000FYZP00Z@0304P70`000h/000030003Z@00008000?000001`L30009ZP00R`2Y
0:T007X7000_1`<000>;00000`0C1`D000>Y00000P181`<000E]R`00R`020003Z@0000@01PL60004
RjT000:Y0P1i1`00;PL01@1]002Z00D04`L20004T:X0[`<0B@L200>;0P000jV;00030004ZP0000H7
00<0LX`02`1i1`00;PL0201b002;ZP2Z0P0B1`030940008000>Z00000P181`X000>;00002`001QR;
K@2;ZP@000>5X`000`1g1`00;PL00`1]0007018700<0R`001P191`<000>;S8/00P001jZ@002ZZJX0
1@001j<0002S08D02@001:<00:<207D7000^1`D000>Y00000P0B1`030940008000BZ0000B@L300:;
0P001JT008/000:Y3`000h/00002R`D000BY0000M0L002h700D0K@00ZP020004Z@00014710000jl0
000204T70P0018b;08/20003ZH^Y00<000>Y00000`000jVTZ@0;0005ZP2Z0:T01@1c1`00;PL40003
R`0000@04@L00`2;00020004ZP35Q@806@LB01/70`000h/0R`0300:;0P001h/0002X0:P01@000hD0
Q@020004Y000R`8000>Y00000P02Z@@0L@L002d700<0LP000`000h/0000300`70P031`0308/00080
00>YaP00100C1`D000N;08/0Y8ZT00@000>:08X02@0F1`P000>;00000`000jT000020003a@0000L0
00BTRP2:10000jX0ZP080707000]1`0304/000<000B@002Y0`0:1`D000H7002;0:T701070`0016a[
Rf`8000:RhF;Q@2:Y8ZTRP/040L500:;0P001i2;002YY0000ZT30003a@2T008000B:08ZT10000j@0
R`020003ZP0000H000>400L0KPL002d700<0LP000P001:X00:T20003a@0700P72@000jX0[`0700l7
00<0RhD00V`00h^5R`0@0005RZB:Y8X00`000hZS000900D720001I6;08^<008000V;ZP2Z0:ZY0:P0
0`000h^T00040003ZH/000L000JZZJX0aZT606d7000]1`D000B;002Y0P000jT000081`<000>;0000
1P000jX000020003Z@0000d70P03K003R`0000h00X@<0003Q@25008000>:Q@003@000h/00003R`80
00FYR`00Z@090005Y8^Y0:T02P000i2ZT0020003ZP3500<000>M00L0J`L002d700H0L`2A0:X800P7
1P001Vd00:X0ZP@000BY0:T03@L01025K8D20003Q6^400D00h@00`2SQ00<0003Q00000D000ZS0025
XhFS08E[0P001X/006d0R`8000><00001`000jZY000<0003Z@2Y00@000>Y00000P001ZT00<H0a@@0
K0L002d70`001g800:T0ZJ@00P000l0000071`P000F@002YZP020003ZJX000:Y00<01`L00`L30087
0`000f`0000?0003XhB:00<000FSQJ>5X`0700:41P000hZ5000=0003TH/0008000J;0000Y:T30005
a@2YRjT00`000jT0000@0003_`2o00<0J`L002d71@001JZ;002Z00@000>Y00L010L40006ZP35R`2;
0`000jT0000600@72`000f/0Q@030007QJB:Y8FTRP07000=Q:<008ZSRZB:XhX0RP0C0003K@0000<0
00F;T@2ZT0030007Z@000:T0_`040004ZH/0Z@@000?6ZLH00`001:nY0:T50003_`000080J`L002`7
00D0LP00R`020004Z@00Y08000G00<2X00041`8000>;00000P000h@0000>008700<0QF`010002XE[
QF/0QHZ4RXD2000<RXF:Q@2:Y8X0RZV:1`0028FSQ@2:XhZS0P000hZTRP060003R`00008000=/08/0
0P000i2;00020003RjX0014000BT0:VY0P002ZT0ZJX0ZP2Z0:X806/7000/1`0:06d0096;08/0R`<0
00>X00002`000f/000060004XhZSRPT000=/00003@000hFS0002000=Y02YRZV:08^YRjV;Z@050008
Y02TRZB:Y8X20006ZHZY08^50P001H^:001b00/000NZT02YZP36008000Jo002YRjX20003Z@000080
00W6Z@00aZV_0:l00P000jT0Z@020003a@000080JPL002`71@001I40002@00@000NS0:<0Q02:00T0
00=[Q@000P000hX00004000<Q:<006b608E[QF^510001f`0002;ZH/020001:B:Y8X20008ZHZYRZV:
ZHX80003XhZS008000F:0000Q@060003S02@00<000ZARi40ZX^Z0:Z@0P001JZY003000<000>Y0:T0
0`000jT000050006ZP000:ZY0P000lD0000306X7000/1`D000F;0000R`020007Z@300<2X`0030007
R`00Q8D0Q@0<0003QJ<0008000F;08E/Q@03000ARhF;Q@2;ZH/0RjV;ZH^YRjT010003hZYRZT0ZH^Y
RjV;ZH/0R`020003ZH^Y00X000>:00001`002h^Z0:Z;092ZT02Y008000>_ZJl010003:VZ0:X0ZP36
0<JYaP8000>Y[`001`000kl0_`0306X7000/1`0308`000H000B@R`2;0P000jT000020004QHX0J`L0
00>T000018/40003RP0000H001J;K8]/Rfb;0:V;ZHX0RjV;ZH^YR`2;1@003XZT002:ZHZYRZT0ZHZY
0P002X^YRjV;Z@2;QH/20003K01/008000>;LP000P002H^Y002;08/0T0020005ZZT00:T01`003:ZY
ZP2Y0:T0Z@2ZZ@8000BY00000ZT606X7000]1`0308`000<000>;08/010001/00002XX`8000=/0000
10001h^Y002;ZH/02`004HF;QH^5RhF;QH/0ZH^YRjV;008000bYRjV;Z@2YRjT0ZH/2000>ZHX0RjV;
Z@2YRjV;ZH/20006ZH^YR`2:0P000h^500030006S02;R`2;0`000h/0R`040006[`00RjV;1`001ZVZ
Z@2YaP8000NY0000ZJZY008000G10<40_`020003_`0006T7000S1`0300L700L700<0LP000`0018`0
08/20003Z@2Y00T000R;08/0Y02;00:;0P001H/0002Y00@000=/00000P007H]/Rfb;K02;ZHZYRjV:
Z@2YRjT0ZH^YRjT0ZH^Y00@000b:ZHX0RZV:ZHZYRZT20003R`0000L000A]06eb0P001H^YRi2;00d0
00J;00300:T30004ZP2YZ@<000>T00000jT0102Y0:T40004XP0006T7000S1`<020L30003B`000080
00^Z002;0:P0X`2S0002R`8000V;08/0R`2;08/00P001JV;002Z00@000J5RhF;QH/4000LQ@2;ZH^Y
RjV;ZH^Y0:V;08^YRjV;Z@2YRjV;Z@@000J;ZH^YRjT9000:ZH^Y0:T0R`2<LP8000FZRjZ;ZP070008
ZX^YRjZ;09020005ZJ<0XjT010001JT00<2Y008000jZZJX0ZP31Z@2W_`2n06T7000S1`03TF^400H0
0PL2000;LP000780R`2;08/00P001:VS0:<20003R`0000:;0P0018X00003R`<000j5001/08]/Rfb;
K8]/R`T000FYRjV:Z@030007ZH^YRjV;Z@020005RjV;ZH/020001jV;ZH^YR`0018/00`2;R`020004
LVebK@H000>@RjT00P002X^YRjV;ZH/0RjT20005XjVS0:<00`002:T0ZJT0Z@2Y0P001JT0Z@2Y00@0
00Bo0;h0J0L002<700<0T@002P002Y40L`2<LP2;08/30003XjVS01L000LH08F;QH^500@000>ZR`00
0P0018^ZT:X9000@ZH^YRjV;ZH/0RjV;ZH/0Z@8000bZZJZYZZVZZJX0ZH/20008ZH^YR`1bS7820006
ZX^ZRjZ;0P003:Z;08^YRjZ;092ZZ@<000JSZ@2Y0<030003ZZVZ008000>Y`:T00P001ZX00:T0R`<0
00>R00L0I`L002@71P001J@00:>400L000Eb002@R`020003ZJ>Y00<000BA096;0P002ZV;08^@Ri00
ZH/50003K8/000<01H/03`2;ZH/0RjV;ZH^YRjV;Z@06000>RjV;ZH/0RjV;Z@2YRjT20007RjV;ZH^Y
0002R`0308^;00B;01P0LVd0K@2;ZH^@R`2;T8^@0:V;ZH/0RjT30003ZJ<000@000>S0:<00`02Z@80
00>TZ@000`001JT0002T00D0I`L002@700<0L`001@001JX0002<008000=;08/00`001H^Z002S0080
00NYX`2;08^Z008000R;002ZRjX0ZP@000<H00000P003ZZ;ZX^ZRjZ;ZP2ZRjZ@0P003ZZ@ZX^ZT:Z;
092Z0:Z@10001jV;002Y0:X00jT00jZY0002000TZJZY08/0RjV;ZH^YR`1b06eb0:Z;ZX/0RjZ;ZX/0
RjZ;Z@2Z0P001H/0Y:VT00<000NX0000Z02Y00P000>ZZ@000P000jVZZ@0406L7000T1`H000F;002Y
R`0?0004XjVSZ@8000F;002@R`030005R`00T8/01`000h/00008R`0C08^YRjV;Z@2YRjV;ZH^Y0:V;
Z@020004RjV;Z@H000V;ZH^YRjV;ZH/00P000jT00008R`80019]LVd0Ri2;08^@Ri2;T02YRjT40004
RX/0Z@8000>SZ@000P000l00000400:Y00D0Z@2Y0002Z@060:@00:VT101V1`009@L30003ZP0000@0
00>;04/00`02R`<000>ZR`000P001Z>YY:T0ZP8000B;ZP2Z0P0018^ZRjX30007K02<096;T@02000S
RjZ;ZX^Z0:Z@ZX^ZT:X0ZY2ZRjX0ZY2ZT:Z@092ZT:Z@ZY006P001G9]LP2Z008000Z;ZX^ZRjZ;ZP2Y
0P001JT0RjB;00P000S0Z@2Y0:VZZ@@000S1Z@2YZP2ZZ@8000>ZZ@000P1V1`009@L40003T@000080
00>;Z@000P002QP0R`2;0900T8/40003ZJ<0008000>;T@001@000h^@00020005K02;K@000h/0192;
0003R`0:0:V;ZH^YRjV;Z@8000^;ZH/0RjV;ZH^YR`020009ZH^YRjV;08^Z00@000N;ZH^YRjT000F;
10001F]/001]008000Z@Ri2;T8^@Ri2;0`0018/0RX/30004ZZVYZ@8000O6002X0:P000BY0`002:T0
Z@2YRjVT1`1U1`009PL30003R`0000P000^;096;08/0RjX0ZP040003Y000008000JZ0:Z@ZY05000<
601bS02ARi6;TH^A0P009h/0T:Z;ZY2ZRjZ@ZX/0T:X0ZY2ZT:Z@ZY2Z0:Z@ZY2Z0:ZYZZVZZ@040004
ZJ@0R`@000R;08E[QF^5J`h000^ZZJX0Y02ZT:X0aP020005Z@2YaZT00P001<D0ZZT30003ZLH000<0
00NYZZVZZ@2Y00D0I0L002H710001h/0R`2;ZH/00`000fd0R`02R`0:08/0R`2;ZH/0X`8000FT002;
T@020006Ri6;08/H0P001Vf;K@2;T0>;00B@Rh^;2@001H^YRjV;00<000Z;ZH^YRjV;ZH^Y0P003X^Y
R`2;ZH^ZRjV;ZX^Y10000hZ;Q@020004K6]/J`D000V;TH^ARi6;TH/00`000h^Y00020004ZX^ZR`80
00NZZJZY0:W600H000NYZP2ZZJZY008000RYY:V;Z@2YY0D0H`L002L70P001h/00:Z@0:T00P002QP0
08^ARi40ZX/200:Y00FZZ@00Z@040004R`00ZPH000ebS7:<096;TH^ARi6;00D000F@ZY2ZT00;000C
T:Z@ZY2ZT:Z@092Z0:ZYZZVZZ@04000:ZH/0RjB;06^5J`<000b;092Z0:Z@ZY2ZT:X20004RjX0ZP80
00B@ZY2Z10004ZW6ZLH0aP360<JY0:V_Z@2YaPL000BZZJX00ZT01:X0001S1`009`L20003TH/000H0
00<H000018/0102;T8/20003ZH^Y008000>:08X01P000f`000020006KH]]R`2@0h/02Y2;Rh/0Ri40
ZX/2000:ZX^ZRjZ;ZX^ZR`<000>@00001P000jV;00020003RjX000<000R;ZH^YR`2:R`8000A[K000
0X/20009Ri6;08^ARi6;008000B@Ri2;2`001JT0002Y008000C6Z@2Y0`003:VZ0:X0Z@2YR`2;Z@D0
00JY0:T0_P1R1`009`L20005RjZ;08/01P002I40002A0:Z;ZP020004Z@00Z@8000BS08^Y0`000f`0
0004000:LXab08^ARi6;T@8000R;ZX/0T:Z@ZP8000^@ZY2ZT:Z@ZP2ZT0020007ZY2ZT:Z@ZP07000=
ZH^YRjV;ZH^Y0:@0Q@03000=TH^A0:Z@ZY2Z0:Z@ZP020004RjZ;ZPD000>ZT:X00P001i2Z0<6Y`JT0
0`001jW6002Y0:T010003:ZY0:VZZ@2YZZVZZ@800ZT40647000W1`<000>;T@001@000aP0R`03R`P0
00B:Z@2T100018X0Jf`60005K@1]R`000X/0492;R`2ARi6;T@2ZRjZ;ZX/2000;ZX^ZRjZ;092ZT:X0
0P00392ZT:Z@ZY2ZT02;ZPH000>;Z@001004R`8000RARi6;TH/0R`/000>;ZH/00P001ZV;ZH^Y00:Y
0`001ZT0002Y00>Y00<0ZJT00ZT0302YRjV;08^YY:T0Z@8000>T00000P1P1`009`L30003TH/000D0
00I;096;TH/20004ZH^ZR`8000VYR`2;08^Y0:T01`002X`0S780LP2;TH/2000?ZX^ZRjZ;092ZT:Z@
ZY2Z008000^@ZY2ZT02@ZY2ZT002000@ZY2ZT:Z@ZP2ZT:Z@ZY2ZT0@000FARi6;T@020007T:Z@ZY2Z
T005000:RjZ;ZX/0Ri40ZP8000N@ZY2ZT:Z@00<000FY`@35Z@040003aP00008000^YZZT0ZJZYZZT0
Z@020003ZZVZ008000FY002n001O1`00:0L20006Rh`0T@1b1003R`0908/0092;T02Y008000J:Y8X0
RZT2009/0P007aP0LP1]R`2;092;08^ARi6;TH^A0:Z;ZX^ZRjZ;ZX/00P002jZ;ZP2ZT:Z@ZY2Z0080
00b@ZY2ZT02;ZX^ZRjX30003Z@2;00B;00H0TH^ARi460008Ri6;TH^AR`02R`d01:T01P2Y0:T0Z@80
0jT40004RjV;Z@D000JYRjV;0:@505l7000X1`0800NAR`2;06d40003TH/000:;0P0048^YR`2;Z@2Y
RjT0Z@2508D20008601b07:<LX`4000FRjZ;ZX^ZRjZ;092ZT:Z@ZY2ZT:Z@ZP8000^@092ZT:Z@ZY2Z
T0020008ZY2Z0:Z@ZY020008ZY2ZT02;TH/20003ZY2Z00<000^Z002@ZP2ZRjZ;ZP020003R`000080
00VZT:Z@ZY2ZT:X010001ZW10<D0a@@000BY`:T00ZT50003ZZVZ00H000BY0:VY101N1`00:0L0101c
08`30003LP0000800X/00`2;0002R`0508^@090010004XX0RP1/06`0601bR`2;KH]]R`8001V;TH^A
Ri6;TH^A0:Z;ZX^ZRjZ;ZX^ZRjZ;00T000B@ZY2Z0P000i00R`030007RjV;ZH^Y0002R`0308^A00<0
00nAR`00TH^ARi40TH^A08/00P004H/0R`2@ZY2ZT:Z@ZP2Z0:ZY00@000G0Z@00`0020005Y:T0ZJ@0
0`002h^YRjT0ZH/0RjV;00@000>Y00000P1M1`00:0L00`1bT@040003LP00008000bA08^;08^A0:X0
ZH/30006R`0006b:0P02600808ab07:<LX`2000FT@0008^ZRjZ;ZX/0T:Z@ZY2ZT:Z@ZP@000>@0900
0`000i2ZT00=0005T:Z@ZY001P000jZ@ZP02000:T:Z@ZY2ZT02;ZP<000bZT02;ZP2ZT:Z@ZY020007
ZY00ZLJYaP02000OZLD0aJT0Z@2Y0:T0Y:VT0:VZZ@2YZZT0ZJZY0:@0Z@0405d7000Y1`0300N<00H0
00<H00000h/03P2;R`2;08/0Q@2:08ZT0`0038/0602;LX/0Rff;K@8000BARi6;0`002h^ARi40ZX^Z
RjZ;00@000ZZT:Z@ZP2ZT:Z@0P003ZZ@ZY2ZT:Z@092ZRjZ@1`001H^ARi6;008000VARi6;TH^ARi40
10005X^AR`2;TH/0T:Z@ZY00T:Z@ZP2ZZJX20003Z@00008000K0Z@2Y0:@20004ZH^YR`<000Z;ZH^Y
0:V;08^Y0P000j@0000205`7000Y1`@000J@000H06d30006S8/0Ri6;0P002:X0R`2Y0:V;0P02R`80
01:<LXab07:<LP2<096ZTJZAZY430004R`2@ZP@0012@ZY2ZT:Z@ZY00T:Z@ZY2Z0`002jZ@ZY00T:Z@
ZY2Z00@000J@092ZT:X20009T:Z@ZY2ZT:Z@00@000>ZRjX00P005h^ZR`2@ZY2ZT02@ZY2ZT02YaP2Z
Z@2Y008000?600000P001jT00:VTZJ@01`002ZVZ0:T0ZJBY0:T30003^@0705X7000Y1`<000>;0000
1@000g:<000300:;0`001h^508ZT0:T00P006F`0601bRg:;08]]R`2A096;TH^ARi6;TH/010005jZ@
ZY2ZT:Z@ZY2ZT:X0ZY2ZT:Z@ZY2Z00<000BZ0:Z;10002:Z@ZX^Z096;0P001I6;TH^A00L001fA096;
T@2ARi6;T@2ZT:Z@092ZT:Z@ZP2Z0:ZYZP020004Z@2Ya@8000>Y00001@000h^YR`030003R`000080
00BYRjV;0P000jT0000305X7000Z1`<000>@00000P001Fd0TH^A008000B;002A0P02R`070:V;08/0
R`03000FLXabS780LX`0TH`0TJZAZY6ZTJZAZP8000>AZP0010004jZ@ZY2ZT:Z@092ZT:Z@ZY2ZT:X0
2@001ZZ@ZY2ZT0<000>@000010000jZ@0002000IT:Z@08^Z096;ZX^AR`2@ZY00T:Z@ZY2ZT0020003
ZZT000<000S60000aP2ZZ@8000>Z00000P001JT0Y:VT008000RY0:T0ZJBYY0<000>Y00000P1I1`00
:PL60008600H07:<08`20003Ri0000H000FYRP00R`03000ORg:;LX/0R`2ARi40TH^ARi6;TH^A0940
ZY6ZTJZAZP060006T:X0ZY2Z1`002JZ@ZP2ZT:Z@ZP0:0006Ri6;TH^A0P004X^ARi6;T@2A096;TH^A
Ri40ZP8000J@ZY2ZT:X60004ZP00aP<000?5Z@000P001jT0T:Z@ZY002@0028^Y002;Z@2Y0`000j80
001I1`00:PL30004L`00B`8000F;T@2AR`020006TH^A096;1`00:H/008abS7:<LP1b08bAS02AZY6Z
TJZAZP2ZT@2AZY6ZTJZAZY6ZTJZA00H001:ZT:Z@ZY2ZT:Z@092ZT:Z@ZY030008T02@ZY2ZT:X30008
ZY2ZT:Z@ZY020008TH^ZRi6;ZX/20003ZY0000D000N_002Z[jZ_008000B_00360`001jZY0:VZZJX0
0`002jZ@002ZZ@2@0:BY008000>TZ@000P000jT0000205P7000Z1`L000U]08`0S780LP000X/01@2;
R`2;008000>ZR`000P009X/008]bRg80K@2;TH^AR`2;TH^ARi40TH^A0:ZAZY6ZTJZAZY6Z0`001:X0
ZY45000<T:Z@ZY2ZT02@ZY2Z0P00292ZT:X0TH^A0`001i6;TH^ARi404P00496ZTJZAZP2Z0:nZ[`2Z
0<H20003ZLH000@000^Z0000ZY00T:Z@ZP020003RjV;008000JY002;0:<20003Z@000080E`L002/7
10000aP0B`02000>TH^A096;08/0Ri40ZX/2000WZY000940S02<LX`0S02AS96<T@2ZTJZAZP2ZTJX0
ZY6ZTJZAZY6Z00<000bZTJZA096_TJnA[i450004T:X0ZP<000RZT:Z@ZY00T08000>Z00001`003:X0
0900RjZ;TH^ZR`<000NA[i6_TJnA008000J_T@2Z[jX2000:aP360<H0aP2_Z@8000VZZJZY092ZT:X0
1@000jZ@00020005Y0000:T00P000jT0000205H7000[1`@000<H06d00`003H]b07:;0900R`2ARi40
0P0028^Z002;07:<0P006g80096;TH^AR`2;TH^A096;TH/0TJZAZY6ZT@02000?ZY6ZTJZAZP2ZTJZA
ZY6Z01L000RARi6;TH^AR`8000VARi40TH^ARi400P02T@0G0:ZAZY6ZT@2AZY6Z0:nZ0:T0Z@2Y0:h0
0P000jZY00030006ZY2Z0:Z@0P001JZ;002Y008000J;0000RjT20003RP0000D0E@L002`71`000h^A
00020004TH/0R`80012ZRjZ;0900T02;08/0LXab0`005HbAS96<T@2ZTJX0ZY6ZTJX0ZY6ZT@030005
TJZAZY4030004:nA[i6_TJl0ZY6ZTJZAZY43000:TJZAZY6ZTJZAZP80022@ZY2ZT02;TH^Z0:ZAZY40
TJnA[i40TJnA[i40ZP2Z[`8000NZaP00[P2Y008000>ZT:X00`003:X0092ZT:X0ZP2ZT0H000JY002T
08/505@7000/1`<000LH0780S7:A00<000j;0000TH^ARi40ZX/0R`8000JA08abS782000BTH^ARi6;
08^A096;TH^AR`2A2`000i6Z0002000EZP000:ZAZY2ZTJZ@ZY40T:ZAZY2Z008000^@ZP2ARi6;TH^A
R`02000:TH^ARi6;T@2AR`801I404@2ZTJZA096ZTJZAZP2_0:ZY008000F_0000aP020006Z@2@ZY2Z
0`003JZ;ZX^Z0:Z;08^YRjT00P0018/0RjT40003RjT000<0D`L002d70P001aP0K@2;TH/01`0038^Z
RjZ;092Z0:X0ZP8000=bS0000P003i4008bAS940ZP2ZTJZAZP070008[i6_TJnA[i420007[i40TJnA
[`03000I[i6_TJnA[i40TJZAZP2ZTJZAZP2ZTJZAZP020009T:Z@ZY2ZT:Z@00<001>AZY6ZTJZA096_
T@2A[i6_TJnA008000>_ZP000`000lH0aP020005ZZl0ZY0010004:X0092ZT02@ZY00T:Z@ZY040003
Z@2Y008000>;0:T00`1B1`00;@L2000<6000095b0780LP1b0P002I6;TH^AR`2;ZP020003R`000080
00J<LP2;TH/30004R`00T@D000fA002AZY6ZTJZAZY6Z008000bAZY6Z0:ZAZY6ZTJX40007TJZ@ZP2Z
T002000:ZY2ZT:Z@08^AR`8000NARi6;TH^A00L01i40302Z002AZY6ZTJZAZP@000BZ0:Z_0P000jo6
0002000:ZP00T02@092ZT08000^Z0:Z;ZX/0RjV;Z@070006ZH/00:V;0`1A1`00;PL00`0H0002000:
Ri40T@2A0940ZP@000F;ZP2ZT002000>ZP00LY5bT@2ARi6;TH/4000@ZY6ZTJX0ZP2_TJnA[i6_T@80
01:_TJnA[i40TJnA[i6_TJnA[i43000>T@00096ZTJZAZY6Z0:X20003T00000H000>ZTJX00P000i6Z
00040009ZY400:nA[i6_00L000J_ZP00aZl20004aP2_Z@8000>_TJl00P001Y2ZTJZ@ZP8000n@ZY2Z
0:Z@ZP2ZR`2;ZH/010001:T00:T30507000]1`0401P0K@8000N<08ab07:<008000F;TH^AR`08000<
T@00LXabS02ARi6;0`001hbAS96<T@000Y407`2AZY6ZTJZA096ZTJZAZY40TJZAZY6ZTJZAZY6ZTJX0
4P0:T@0:0940TH^ARi6;T@T000BZTJZA0P000jnZ[`020007ZZl00:l0Z@02000<ZY2Z0:X0ZY2ZT:Z@
10002ZZ;ZP2YR`2;ZH/20004ZHX0RPL000>M00L0CPL002d700@0B`1c0P001h/0Ri40TH/00P001:Z;
ZX/20005ZP00TJX00P003940LP1bTG80Ri6;00:A0100ZY6ZTJX0ZY40TJnA[i6_0P004i6_TJnA[i6_
0:nA[i6_TJnA[i4010003:nA096_TJnA[i6_T@8000bZTJZAZY6ZTJZAZY42000;ZY40TJZAZY6ZTJX0
0P004Y6_TJnA[i6_T@2A0:Z_0:l0aP8000B_002_0P003I6_0:nA096ZT:ZA08/00P003JX0ZY00T:X0
ZX^ZRjX00P0018ZTRZ@20003Q00000<0CPL002d700D0601]LP020009LXab07:<LP2;00<00i40102A
09420004T@2<LP8000j<096;08bAS02<THbA00>A00D0ZY6ZT@02000AZY6ZTJZAZY6ZT@2AZY6ZTJX0
10004Y6ZTJZAZP2ZTJZAZY6ZT@2A00^A0240Ri6;T@2ARi6;TH^A0:X0ZY6ZTJZAZY40TJX0[jX0ZZl0
0`005lH0002_0:X0ZY2Z0:Z@ZY00R`2;T@2Z008000V;08/0RjV;ZH/04P031`l02`L902@7000]1`05
04/0Lfd00P001Y6;T@2AR`@000JZTJZAZY420003[`2Z008000AbTG:A1005T@0;096Z0:ZAZY40TJl0
0P003i6_TJnA[i6_TJnA[`2_T@040012[i6_TJnA[i6_0:nA[i6_TJnA096_T@2AZY6ZTJZAZY40TJZA
ZY40TJZAZY6Z0:ZA096_TJnA[i40TJnA[`2_0:nZ1P005/H0[`2_TJnA092ZT@2;ZP2ZR`2@ZY030003
RjX000<000b;0000RZB:Y02TRZ@=0003Y8X000<000BT0:B:600Q1`00;@L=0004LP2;T@@00Y401@2Z
002;008000B<LXab0`002HbAS96<T@2A0005T@<000JZTJZAZY4;000KTJZAZY6ZTJZAZY6ZT@2AZY6Z
TJZA096ZTJX000RA00l0Ri6;TH^A096;TH^A0:X00Y401P2ZTJZAZP8000FAZY6ZT@020003[jZ_00H0
01RZ0900T:Z@ZP2ZT02;TH/0Ri6;08^ZRjX300:;00J@Rh/0T8/20004Y8X0Q@D000=/06`0:P0P1`00
;@L02@1;07=]08^AR`030009Ri40ZY6ZTJZA00L000>A07800`000g<0T@0400:A0P000jZAZP0<0007
TK2A/96`T@04000^[i6_TJnA[i6_TJnA[`2_TJnA[i40TJnA[i40TJZAZY6ZT@2AZY6ZTJZA096ZT@80
016ZTJZAZP2_TJl0[i6_TJnA[`020007ZZnZ0:l0Z@02000JaP00TJnA[`2ZT@2;ZY2Z0:Z;ZP2ZT:Z@
08/20009ZH^Y0:V;ZH^Y00P000>5RP001@001V^508Z4RR407`L002d700P06000LP2ALP8000ZALP1b
08^ARi6;0P001940ZY45000:Li5cTG<0S96<T@L00i40;`2ZTJZAZY6ZTJZAZY6ZT@2A096Z0:ZAZY6Z
TJZAZY6ZT@2AZY6ZT@2AZY6ZTJX000JA00l0Ri6;TH^ARi40TH/0TJX00i404JZA096Z0:ZAZY6ZTJZA
0:Z_00L001f_0000ZY2ZT02@08^ARi6;08^ARi40ZX^Z092;T0020003R`2;00@000J;08^;06`A0005
RjV;08/00`000h/0R`020003Z@0000h000>X00007PL002d70P0046d0096;08/0Ri40T@2ZTJX20003
TJZA008000B_002A0`001G>ALi4000FA0P001Y6ZTJZAZP8001BA/96`TK2A/96`TK2A096`0;2A/080
00jA[i6_TJnA[i40TJnA[`8002bA[i6_TJnA096ZTJZA096ZTJZAZY6ZT@2A096ZTJZAZY6Z0:l0[i6_
TJnA[`D000B_002_0`001:X00:l3000H[`00RjZ@ZX^Z0:Z;ZX/0T:Z@08^ZRjZ;0`002X^YRjT0ZH/0
QH/30004Q@25J`D000>:Z@000`001H/0RjV;00/000>;000020000jT0000201h7000/1`0404/0LP80
00Qb07:<095bT@8000>;000019401P2A002AS08000jALP1cTG>A096<T@2A00JA0P00;ZZAZY6ZTJZA
ZY6ZT@2AZY40TJZAZY40TJZAZY6ZTJX0ZY6Z0:ZAZY6ZTJZAZP03T@8000Z;TH^ARi6;TH^A0P000i6Z
T@02T@04ZY6AT@8000>ZT@001@001ZX00:Z_ZPD000RZ[`00T:Z@ZP<001FARi6;T@2ARi6;TH/0RjX0
T8^@Ri000`04R`0308/000D000>500004P0018/00:X30003ZP2;008000><00000`000h/0R`0801`7
000/1`0501P0KG<00P001X^A096;T@H000JAZY6Z0:l20003TJZA008000NAS000TG<000:A02<0ZY6Z
0:ZAZY40T@2A/96`TK2A/96`T@2A/940TK2A/96`T@02000Q[i6_TJnA096_0:nA[i6_TJnA[i6_0:X0
ZY6ZTJZAZY6Z018000fA[i6_TJl0/02_ZP2Z00@000S6002Z[i6_ZP8000>Z00000P003i00RjZ;ZX^Z
0:X0ZX^ZR`030006RjX008^Y10001HD008E/008000BTRjB:0`002Y000:V;ZH/0RjT20003T00000X0
00FZ002@ZP020003[`00008000C5Z@006`L002`700<0B`002P001Xab08^ZR`H000FZ096<T@02000:
LY5bT@2A096<00>A00<0TI400Y40702Z0:ZAZY6ZTJZAZY40TJZAZP2ZTJZAZY6ZTJX2000CTJZAZP2Z
0:ZAZY6ZTJZAZY6ZT@020003TH/000L000FA0000ZP03T@04ZY6AT@<000jAZY6ZTJZAZP2ZT@2Z[`L0
00jZ0:Z@ZY00T02;TH^AR`<000J;TH^ARi42000;Ri2;08^@092;T8/01@000fd000030004QH/0R`80
00J;Z@2YRjT20003R`2;00P000>Z00001`000i40000<01/7000/1`<000EbL`2AR`03000:Ri6;T@2Z
T:Z@ZPh000><T@000`00=i6ZTJX0ZY6Z0;2A096`TK2A/96`T@2A/96`T@2A/96`TK2A/96`T@2A[`2_
0:nA[i6_TJnA[i401`001i6ZTJZAZY400P002JZA096ZTJZAZP02000BT@00[i6_TJnA[`2`TJl0[`2`
1P003ZX0TJnZ0900T:Z@ZY2Z1P000h^ZR`02000>ZX/0RjZ;08^ZRjZ;08/20003R`0000L000ZZT:Z@
092ZT:Z@1`000jT0ZP030003T02@008000>[00000P000i00000;01X7000/1`0804/006d0LY42000?
R`1bTG80RjZ;ZP2Z0:ZA00<000>;TH/00P000i5b000400FA02@0TI6A096Z0:ZAZY6ZTJZA096ZTJZA
ZP2ZTJZAZY6ZTJZAZY440003ZY6Z00P000BAZP002I400`2<T@02T@05S02AZY400P000jZAZP02000?
TJZAZY6Z0:ZAZY40ZP2Z00D000BYZP2Z0P001i2Z096;TH/00P001Y4008^AR`<000b;0000Ri2;08^@
Ri060003K@0H008000B@R`2;10000jT0Z@050007ZP00T:Z@ZP030007ZP00R`2;ZP0=0003aP0000<0
6PL002`70P0047<0L`2A08`0TH/0R`2@ZY020007ZY40TJX0ZP02000:T:X0T@2AS96<T@8002ZAZY6Z
0:ZA096`T@2A/96`TK2A096`TK2A/940TK2A/96`TK2A/96`TK05000C/96`TK2A/96`0:l0ZY6ZTJZA
ZP020008TJZAZY6Z0:X2000BTK2A/02Z0:nA[i6_0;2A[i6`0P001:Z`0<H3000H[`2_0:Z@ZY00T:Z@
092ZT02@092ZT:Z@0P003JZ;ZX^Z0:Z;ZP2Z0940200028/0T:Z@092Z3`001ZZYZP2ZT0@000FZR`00
TP020003T@1/008000>_0000100J1`00;0L2000HK@00T@2;K@2;T@2<0:Z;092ZT02AZY6Z0P003X/0
R`2;07:ALY5b09400i40902AT@2ZTJX0ZY6ZTJZA096ZTJZAZY6Z0:ZAZY6ZTJZAZY6ZT@<000fAZY6Z
TJZAZY6ZTJX000:A00<0TI400i420003THbA00:A00><T@000P001jZAZY6ZT@000Y402`2ZTJX0ZY6Z
TJZA008000>_ZP0010006JT00:Z@ZY2Z096;08^ARi40TH/0Ri6;TH/00`001H^@R`2;008000N@R`00
R`2;00@000>;00001P003:X0ZY2ZT02@ZY2ZT08000JZ002;08/20003ZP2A00H000FA0000ZP0801X7
000[1`040780L`<000><LP000`000i4000020003T@0000P000jA0:X0TH/0S96<T@2ZT@8002>ZT@00
/96`T@2A/96`T@2A/96`TK2A/940TK2A/96`TK2A/002000MTK00/96`TK2A/96`TK00[i6`0:ZAZY40
TJZAZY401P005k2A/96`T@2AZY40TJl0/96_TK2A[`2`00<000C6002Z0`001Y2ZT:X0ZP80016@ZY2Z
T02@ZP2ZT:Z@ZP2AR`050004RjZ;ZPD000>;0:X010000i2ZR`030004T:X0ZP<000BZT:Z@0P001:ZY
090:0003/00000/000<700L06PL002/700D0B`00L`070003S00000@000>;00000P001:Z@ZX/50008
Ri6;07:ALP04T@0Q0940096ZTJX0ZY6ZT@2AZY6ZTJZAZY6Z0:ZAZY6ZTJZA0080012ZTJZAZP2ZTJZA
ZY6ZTJX00Y400jZA0002T@800Y4400:A10001i6ZTJZAZP000i402@2Z0:ZAZY6ZT@040003[`000080
00FZ[`00T@03000HT000TH^ARi6;T@2ARi40TH^AR`2;TH^A0P001h^@Ri2;08/01P002Y00002ZR`2@
09060003ZY2Z00<000>ZT:X00`001:X0ZY020003ZP2Z00P000FZ0000[`0301/7000[1`<000Ic0000
S7820007T@00R`2@ZP020008T:X0ZP2ZT:X2000@R`00TH^AR`2<T@2ZTJZAZP8003BA/96`T@2A/940
TK2A/96`TK2A/940TK2A/96`T@2A/96`TK2A096`TK2A/96`0:nA/96_1`005;2A/96`0;00TY40TK2A
/02ZTJZA0P000k0000040006T@00/02_0`002:l0[`2_TJnA10004jZ@ZY2ZT02@ZY2Z0:Z@ZP2ARi40
0P001H/0002;00D000=]08/00P000jX0ZP020006T:Z@092Z0P000h/0T0020004ZP2ZZ@<000>Z0000
10001:nA094?01`7000[1`0>04/0KG<0S780B`2;TH/4000:ZY2Z0:Z;ZP2ZT08000>AR`000P001X^A
R`1b00FA0P006jZAZY6Z0:ZA096ZTJZAZY6ZTJZA096ZTJZAZP02000ATJZAZY6ZTJX0ZY6ZTJZAZP00
0Y401:ZATI4300RA00<0T@000i401P2ZTJZA00BA0`002:nA[i6_T@2A0P003JnZ002_0:ZA096ZTJX0
0P000h^A0003000>Ri40TH^AR`2;T@2ARi420005Ri6;TH/020001Y0008/0R`<000>@ZP000P002JX0
0900RjT0ZP030003ZP00008000BA096Z10001Y40TJl0/0<000Be002_0P0L1`00:`L0201c001b07:<
0P001H^ARi6;008000nZT:Z@092ZT02@ZP2A09400P002X/00940T@2ZTJX2000KT@00096`T@2A096`
TK2A/96`TK2A/02`TK2A008001N`TK2A/96`TK2A096`TK2A/02_TK2A[`030008/96`TK2A/942000?
/02BTI:A096`T@2AZY6Z008000RA/96`TK2A/08000BZ0:Z_0P004JX0ZP2A[i6_0:X0ZY6ZTJZA00@0
016ZT:Z@0900Ri40T@2ARjZ;T@020005T00007800`001i2Z002A09000`0019000:X20003T:X000H0
00bZ0000[jX0TJnA0:X40003/00000L07@L002/700d0B`00L`2<06d0TH^<00<0016Z0000ZP2ZRjZ;
08/0R`2;T@020003Ri4000<000VA002ZTJl0ZY400P004jX0096ZTJZAZY6ZTJZAZY40TJX00P004i6Z
TJZAZY6ZTJZAZP2ZTJZAZP000Y400jZA00020003T@2A00JA00D0TI6A0002T@07S940ZY6Z0003T@<0
00V_TJnA[i6_0:X00P000jZ_ZP030005TJX0ZY400P002ZZA08^ARi6;TH/50003Ri6;008000RA096;
08^AR`<000>;08/00P001X`0T@2ZR`8000JZT02@08/;0005ZJZYZZT00P002JX0002ZT@2A[`0<01l7
000[1`@0011b0780B`2;T@2AR`2@ZY2Z1@003ZZ@002AR`2;0900Ri6<0`002Y6`TK2A/02`TK090007
/96`TK2A/003000G/96`TK2A/96`TK2A/940TK2A/02_TK000P007940TK00/96`TK2A096`T@2ATY6B
T@2A/02ZTJX2000;T@2A/96`TK00[i400P000jnZ[`02000@ZZnZ0940TJZA[i40TJZAZP8000FA002Z
T0060009TH^A0:Z;08^A008000>;T@0020000jZ@00060005T02@ZY000`001JVZZJZY00<000>Z[jX0
1`000jX0ZP060247000[1`D000Ab04]]0`001Y6;092ZT08000FZ002;ZP04000=Ri40T@2Z096;TH/0
T@020007[i6ZT@2AZP020008T@2A[i6_TJl:000CTJZAZY6ZTJZAZY6ZTJZA096Z0002T@8000B_TK00
0Y400`2AT@02T@800i40102ATH`3T@050:ZA09400P003Jl0T@2_TJnA[`2ZTJX020006I6Z002@ZY6Z
T:X0TH^A096;T@2A096;TH/00P002Y6;T@2AR`2;TH/20003TH/000H000R;0000T:X0ZP<000BZR`2;
0P002:ZY002ZZ@2Z0P001:ZAZY420005ZP00ZZl010000jl01`0R1`00:`L50004S01bK@<000B;T@2Z
0P002Y2ZT02@092ZT:X2000>Ri40092Z096;T@2_TJl3000H/0000;2A/02`TK2A/96`TK2A/940T@2A
1P003K2A/96`TK2A/02`0:l00P006Y6`TK00/96`0;2A/02`TK2A/02BTI:ATY6B0P004940TK00/940
TK2A/940TJl20007Z[2Z096_T@04000FZY6_TJZA096Z0:ZAZY40T@2@ZY2ZT0<000b;0000Ri6;T@2A
Ri460003T000008000>Z0:X00`001JX0002Z008000RYZZVZ0:X0aP8000FZ[`00ZP070004[`0002@7
000[1`<000A]001]0`000i6;0003000<ZX^ZRjX0TH/0RjZ;0`0048/00:Z;08^AR`2AZY6ZTJX2000@
TJnA[i6_0:nA[i6_TJnA[`8000ZAZY40TJZAZY6Z1P000i6ZT@040006/96_TK001940102AT@05T@03
096<00>A00><TI400`00196_T@03T@06096_T@2A10001ZnZ[`2ZT@8000>Z0:X010000i6Z0002000@
TH^ARi40TH/0Ri6;08^AR`@000>ARi400P000h^AR`060005ZP00T:X00P001h/0Ri40ZX/00`000jVZ
Z@040003[jX000<000>Z00000P000jZ_000402H7000[1`0404/0L`<000Ic002;TH/4000=ZY2ZT:X0
ZY00T:X0ZP02000=R`2@ZY00S02A[i6_T@02000T/0000;2A/940TK2A/96`TK00/96`TK2A096`TK2A
/96`TK2A1@00396`TK2A/02`TK2A/0<000n`TK2A/02BTI:ATY6BTI800P005Y6`T@2A/96`0;2A0940
[`2A0:Z`0:l20009TJnA096_TJnA00@000NZTJZAZY4000>A00L0T:X0TH^A008000>;T@002@0028/0
LP2@ZP2Z0P002I000:Z;092ZT0050004ZJX0aP8000>Z00000P001Jl0002`00D09`L002/700P0B`1]
04]bB`8000FAR`2;0002T@0<0:Z;ZX/0Ri40ZX^Z0P001g80002;ZX/00P000jZAZP020006TJnA096_
0P007i6_T@2A[i6_TJl0ZY6ZTJZAZY40TJZAZY6ZTJZAZY400P002:l0/96_TK001I4200NA00<0S940
0Y402hbATI6<096_TJl000BA00/0T@2A096`0940[`03000;ZY6ZTJX0ZY6ZTJX00`000i400002000C
Ri40TH^A096;08^A0940TH^AR`020003TH^A00<000FA08`0ZP030003ZP0000D000>Z000010001JVZ
0:nZ00@000BZTJZA0P001:l00:l302L7000[1`0304/0008000=cKG<00P0028^A0:X0ZY6Z0P003I2Z
0:Z@ZP2ZT02@08/00P000jZ@ZP02000iT@00/:Z`Z[00/96`T@2A/02`TK2A/02`TK2A/96`TK2A096`
TK2A/96`T@2A/96`0;2A/02`TK2A00@000n`TK2A/940TI:ATY6BTI800`002k2A/02`TK2A/02`0080
00F_/02`T@04000:/000TJnA096_T@8000R_T@2AZY6ZT@@00i400`2Z00020006TH/0S96;0`0018^A
09470003ZP00008000F@08/0R`0?0005Z@00[jX00P000k001`020003]00702L7000[1`<000A]04]b
0P001X^A096;00FA0`002h^ARi40ZP2A095b00P000b_TJnA[i40TJnA[i420004[`2_T@8001>ZTJZA
ZY6ZTJZAZY40TJZAZY6Z008000^A[i6`T@2A/940T@030005TJX0T@0019400`2<T@02T@03S96A00<0
00JATP2_T@05T@8000F`TK2A0002T@8000FA096_T@020003ZP2Z008000BAZY4019430003T@0000H0
00N;TH^A096;008000JA096;08/>0003T@0000d000R_002Y0940ZP8000<7[`000P0Y1`00:`L01`1;
07<0Lfd00P003I40ZY6Z0:ZAZY6ZTJX01000292Z002@ZP2A0P001Y6ZT@2Z[`@000ZZ/02`TK2A/96`
0P008I40TK2A/96`TK2A/96`TK2A096`TK00/96`TK2A/940T@03000BTK2A/02`T@2A/940TI:ATY6B
0P003I40TINA0940TK2A/9400P002[2_/:l0TK2A0:X20005/96`TK001@001JnA[`2Z00<000NZTJZA
094000:A0`002I6;TH`0R`2;T@020006R`2;T@1c10000i4000060004R`00R``000C60<JZ0P000k00
000402X7000[1`D000=;04/00P001I6;TH/000JA0`000i6;0003000@TH/0LP2@096Z0:ZYZZVZZAH0
00ZZTJZAZY6ZT@2A0P001;2A[i480006ZY6ZTJX00i401@2ATI4000>A00^<002GT@2ATY6BT@020003
T@0000@000FA/96`0002T@8000R_002AZY6_T@8000>ZTJX01P003X^ARi6;T@2AR`2<T@2A0`001I40
08^A00D000>ALP001P000h/0R`0B0003ZP2_00`0:`L002/71@000g<0LP02000=TJZAZP2ZTJZAZP2Z
T@020004ZY6ZT@8001BZT000ZP2ZT@2Z[jZ_ZZnZ0:T0T@H000>ZT:X00P001Y6ZTJZAZQ0000BA/96`
0`004K2A/96`0;2A/940TK00TY6B008000ZA/940TINAUi6G0`003;2A/02`T@2_/02`T@<000NA002`
TK2A008000B`TK2A0P001:X0ZY440006ZY40TJX00Y400`2AT@02T@D000C600350`0016fZRjl40003
T02;008000>Z09404@001<L00:X200031`0002`7000[1`D000Y;04/0Ri6;TH/019406P2ARi40T@2A
Ri40ZP2A0940ZP2ZT@2YZZVZ0P002Y2ZT02AZY6ZTJX70005ZY6ZTJX01P005Y6_0:nAZY6_TJZA[`2_
TJl0ZY6ZT@05T@800Y42000:Ui6GT@2ATY6BT@<000^AUi6G09NAU`2`T@040003T@2_008000RA[`00
T@2AZP8000ZAZY40S96<THbA100028^A0940TH^A0`0019470:h50007S6d00940LP040003T@000080
00=b00004@001:X00;D202d7000[1`H000=;00000P001ZZAZP2ZT@8000nZTJZAZP2Z0:ZA096Z0:X0
0P006i2Z0:X0[jZ_0:nY[jV_Z@2A[i6_TJnA[i6_T@06000@/96`TK2A/96`0;2A/96`T@8000n`TK2A
096`T@2A/96`TK0010002Y6`TK2A096GTIL2000=TK2A096`T@2A/940[`020005/02`TK000`001[00
ZY40T@8000RZTJZAZP2ZT@@000>ZT@001P001i40096;08/01@002/D00<H0[i6_09470003T@0000H0
00>A00001P000lX000020004[`0002h7000[1`0304/000H001R<T@2AR`2A08^ARi6;TH/0Ri40TH/0
TJX30003ZY2Z008001>YZP2ZT:Z@ZY2ZT02AZY6ZTJZA00@000>ZT@001000396_TJnA[i40TJnAZP80
00ZA[i6_TJl0ZY401Y43000:Ui6GTINA096BT@8000jGTINAU`2GT@2AUi6GT@8000>`T@000i401@2A
002A00h000^;TH^AR`2;T@2AS0030007LY5b08/0Z@0E0003R`0000<000=b00001`000lD000030004
[`0002l7000[1`0404/0C0D00i44000BTJZAZY6ZTJX0ZY6Z0:X0ZY6Z0P000i6ZT004000=[jV_ZJnY
[jV_Z@2A[`04000;TK2A/96`0;2A/9400`00696`TK00/96`0;2A/96`TK2A0940TK2A/0@000ZA096`
TK2A096G0P00296`TK2A/96`0P000i6`T@040009/940TK2A/02`00<000V`002A/96`TK000`000iL0
0002000;TJX0ZY6Z0:ZA08/00P000i4000060005a@00/9400`001jP0002;09000`000h/000040003
/00000X000>d00000P0_1`00:`L30004B`00LP8000F<THbAS0060009TH^ARi40TH^A008000B@TI6Z
5`004:ZAZY6ZTJZAZP2ZTJZAZY44000>[`000:nA[i6_TJnA[i46000<U`2GTIL0Ui6GT@2A0P0029NA
Ui6GTINA1`001Y40002A00BA10002I6`09:A096GT@0200:A00<0TI400i44000;TH/0R`1bTG:A0940
2@001LH0[i6_00D000>A09400P001H/0002A00P000>Z0<X01@000`L0000202l7000[1`0304/00080
00Ac094019420005/96`TK003000196ZTJX4000<T@00[jZ_ZZnZ[jZ_0P001:X0TJl70009/940TK2A
/96`01@000>A/9400P002;2A/96`0;2A0`00196`TK090003/96`008000>A/9402P001k00/96`0;00
0P002I6`TK2A096GT@020003U`0000D000FA0000T@0<0003T@0000P000JZ002<094F02l7000[1`D0
00Y;07<0S96<09402I401`2A002BTI800`02T@04ZP2ZT08000RZ0:ZAZY6ZT@8000^ZT:X0ZY2ZTJZ@
ZP0;000=/96`TK00/96`TK2A/004000?T@2AUi40TINAUi6GT@2A00P000BGTINA0P000i:ATP030003
Ui4000@000BAU`2G0P001I4009:A00L01940102A0004T@800Y430005TG>A09401`000kl000060003
Z@2A008000>A00000P0018bA094<0005]0000:l00`0_1`00:`L50004L`2A00:A01L0TK2A096`TK2A
/940TK00/02`T@2A/0030005ZP2_ZZl010001ZZ_002A[`/000R`U[2F/9H0UPH000J`U[00/9440006
/96`TK2A0`00196`TK080008T[2B/9:`T[020004TINAU`P000>A/0001@000iL000030004U`2G00>G
20002iNAUi40T@2ATP2B008000B<002Z40001:T00003T@<000>AZP003@001:l00<X302l7000[1`03
04/000<000JAL`2<T@03T@03096A00BA00h0TI6A[`2A09:A096`T@@000BZT:Z@20001k2A002`0;00
10003YJ`U[2F/02`TK2F/96`5`004iL0002GT@2AUi6GTINAU`2BTI800`000i80U`0900:G00<0UiL0
10001iL0OYL0Ugh010001IMh07RG00D00Y401P2A07>AL`<000>;00001@000j8000030004RP2Y[`<0
00JAS02A09430003TJl000T000>_1jl01@0_1`00:`L6000>Li40T@2`TK2A096`TK020008TK2A/940
TK020006T@2A/96`1P004jn`[`2F/9J`UP2_/9J`0;2F/9H01@001IJ`0;2F00D000ZF/02`U[2F/9J`
10001i40TK2A/9400P003;2G/9J`0;2B/9:`TP8000JG0000Ui47000:UkD009L0U`2G]@l000NGTYNB
09:G008000Yi0000T@2A07RB3@000j<000040003Z@0000<000NAZi40T@2F00<000>_00000`000k@0
000400041`0002l7000[1`H000BAL`001I401P2ATI40[`>A00F_TI6A0003T@080940TK2A/94:000@
/000U[2F/9J`0;2F/9J`UPX000R`TK2A/02GT@@000RGTINAUi40T@<000bAUi6GTINAU`2GTIL:00>G
0P02U`8000>G0000200<1`P000EhTWP0N0060005TWQh0:X01@000lD000050005aJP0ZJX01002T@06
0940/02G10000jl0]0030003[`N_00@0<0L002`710000i5c0002000GTK2A/96`0;00/96`TK2A/96`
0;2A/9400P001k2A/96`TK000`000k2_/00C0004/9N`U`8000>`Uk006@002;2F/9N`UP2B1`02Ua00
6PL70003U`00008000BATW=i3P000jo500020004ZP2ZT@D000>F/0000P000kD0000300031jl00080
00B_0000<0L002`720000i40000200:A0P001JnATI6_00>A00F_TI40TP02T@P000>A/9400`0019H0
0;020003U[2F00<000RF/000U[2F/0`000>A/9400P000iNAU`050004UiJG00>G00>FU`004@0/1a80
00>_000010000`L000030007ZY2Z096;T@0;00051`07[`L00P001:l0000a1`00;0L01P1;002BT@80
00F`TK2A/00C0003/9N`008000>G00004`000iNeU`020003/9N`00@000NA/940TK2A01P0@@L60006
N000NGQi1`000k0000080008T@00/0070;030006]@07]02_0`000jl000020347000/1`@000Ec07>A
0004T@8000^`09NAUi6GTINAU`020003UP0000<000>FUiH00P001k2F/9J`U[001P001iJ`0;2F/9H0
500029NFUiJG09NF1P1?1`L000=i00004@000k2F/0020004UP2_]@H000>_1`00100b1`00;0L60005
TW<0TK000P005I6`T@2A/96`TK2A/02`Uk2G09N`U`040009/000Uk2G/9N`008000>G]@001@000iNe
00030006Uk0009N`10001YN`002G/0P0E`L;0003T@1;00L000C50:o70P000iH0[`0200031kD000X0
=0L002`700<0L`000P000i5cT@03000</96`TK2A096GTINA0P003K2F/9J`UP2FUiJGUYL01@002K2F
002`Uk2G/00200:G6P1N1`@000=@00002P001XD0aJk6ZP:_1`001:l0[kD503H7000]1`X001jA/96`
TK2A096`T@2G/9N`Uk2G/9L0Uk2G/9N`Uk0@0005U`00/9L03P1V1`H000=i00001`000hA[Q0030003
ajl0008000B_1`071P0h1`00;@L40006TG<0N02A1P000k00U`02000?U[2F/9J`U[2F/9H0UYNF00l0
00FFUiH0UP0@06/710000e2B00020004R`00R`H000>YR`000`000kD01`0503X7000]1`03098000<0
00ZAU`2`TK2A/96`60000iNe00020006/9L0Uk2G2`000gd0000506l76@000k@0000m1`00;@L00`1c
T@030009TWRB09NA/96G008000^F/02`U[2F/9J`UP040005/9L00;000YL30003/00000d000=nO@00
1`1b1`T000F;0000R`040003Z@2Y008000Fd[`36000l1`00;PL60005Ui40TK000`002K2G/9N`0;2G
/0040003UkD000l000FG/9L0U`020003]@0000X0M`L00gT000090003/00000P000>_00L0>`L002l7
00D0LZT0U`0G0003/9J`00<000>`000030000iNLU`0707d710000i8000050003]@0000<000G:0<G5
1`0203/7000`1`0308^Z00@000JG002`Uk040009UkD0]INeUkFG00T000>G00002P281`0305EW0080
00>G/9L00P0010L00;D70003]00703/7000`1`<000>Y00002P04U`030;2F01<0T@L90003[`0000P0
?0L003471@000iL000020005OP2GW9L020000kFG000<09H730000jl0[`020003aP0703X7000a1`D0
19LA0003/00000D0W0L60003]@0000`0=`L0038720002Y:A0940TK2A/9430003U`0000H0XPL30009
/02`0:l0[`2Y00`0<@L003<71P00296<096`0;2A202/1`X000BZ002@2@0_1`00=0L>0004/0000;07
10001K00002`00@000>@08/010001:X0000]1`00=PL=0;@71@001IH0002A00l0:`L003L72@2i1`@0
00>F00001@000i40000902X7003l1a80;@L00?l72`0a1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00
o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00
o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00
o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00o`Ll1`00\
\>"],
  ImageRangeCache->{{{0, 314.438}, {240.688, 0}} -> {0.0262315, 0.0968626, 
  0.0030986, 0.0030986}},
  CellTags->"DXFFiles"]
}, Open  ]],

Cell[BoxData[
    \(MVClear[]; 
    MVLoadDXF["\</export/home/kuska/MathView3D/triceratops.dxf\>"]; \n
    MVRotate3D[{1, 0, 0}, \(-\[Pi]\)/2]; \nMVRotate3D[{0, 0, 1}, \[Pi]/8]; \n
    grTric = Map[Prepend[#, SurfaceColor[RGBColor[0.25, 1, 0.25]]]&, 
        MVGetGraphics3D[], {1, 1}]; \)], "Input",
  CellLabel->"In[36]:=",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(tricera = 
      Show[MVRotate3D[MVRotate3D[grTric, {1, 0, 0}, \(-\[Pi]\)/2], {0, 0, 1}, 
          \[Pi]/8], Boxed -> False, RenderAll -> False, 
        ViewPoint -> {1.540, \ \(-2.844\), \ 0.995}]; \)\)], "Input",
  CellLabel->"In[59]:=",
  CellTags->"DXFFiles"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .63639 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0107323 1.01073 -0.154181 1.01073 [
[ 0 0 0 0 ]
[ 1 .63639 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .63639 L
0 .63639 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.24768 .49113 m .23999 .48758 L .24006 .48319 L closepath p .049 .183 .154 r
F P
0 g
s
.24768 .49113 m .24873 .49774 L .23999 .48758 L closepath p .034 .217 .171 r
F P
0 g
s
.24903 .48009 m .24006 .48319 L .24048 .47979 L closepath p .06 .133 .132 r
F P
0 g
s
.24903 .48009 m .24775 .48672 L .24006 .48319 L closepath p .061 .147 .135 r
F P
0 g
s
.24768 .49113 m .24006 .48319 L .24775 .48672 L closepath p .049 .183 .154 r
F P
0 g
s
.24945 .47303 m .24945 .47667 L .24048 .47979 L closepath p .057 .111 .126 r
F P
0 g
s
.24945 .47303 m .24048 .47979 L .24137 .47788 L closepath p 0 0 .029 r
F P
0 g
s
.24903 .48009 m .24048 .47979 L .24945 .47667 L closepath p .06 .133 .132 r
F P
0 g
s
.25033 .47112 m .24137 .47788 L .25107 .47047 L p 0 0 0 r
F P
0 g
s
.24945 .47303 m .24137 .47788 L .25033 .47112 L closepath p 0 0 .029 r
F P
0 g
s
.24768 .49113 m .25905 .48444 L .25985 .49136 L closepath p .093 .359 .179 r
F P
0 g
s
.24768 .49113 m .24775 .48672 L .25905 .48444 L closepath p .087 .311 .169 r
F P
0 g
s
.25985 .49136 m .24873 .49774 L .24768 .49113 L closepath p .091 .393 .189 r
F P
0 g
s
.26034 .47777 m .25905 .48444 L .24775 .48672 L closepath p .081 .221 .145 r
F P
0 g
s
.26034 .47777 m .24775 .48672 L .24903 .48009 L closepath p .081 .221 .145 r
F P
0 g
s
.25985 .49136 m .26089 .49802 L .24873 .49774 L closepath p .091 .393 .189 r
F P
0 g
s
.26089 .49802 m .25069 .49811 L .24873 .49774 L closepath p .022 .626 .215 r
F P
0 g
s
.24903 .48009 m .26056 .46807 L .26328 .47104 L closepath p .075 .181 .137 r
F P
0 g
s
.24903 .48009 m .24945 .47667 L .26056 .46807 L closepath p .076 .192 .139 r
F P
0 g
s
.24903 .48009 m .26328 .47104 L .26034 .47777 L closepath p .082 .237 .149 r
F P
0 g
s
.24945 .47303 m .25033 .47112 L .26236 .46055 L closepath p .006 0 .036 r
F P
0 g
s
.24945 .47303 m .26236 .46055 L .26055 .46442 L closepath p .001 0 .03 r
F P
0 g
s
.26055 .46442 m .24945 .47667 L .24945 .47303 L closepath p .07 .15 .13 r
F P
0 g
s
.26055 .46442 m .26056 .46807 L .24945 .47667 L closepath p .07 .15 .13 r
F P
0 g
s
.26089 .49802 m .26471 .49874 L .25069 .49811 L closepath p .022 .626 .215 r
F P
0 g
s
.27671 .47288 m .27716 .47971 L .25905 .48444 L closepath p .096 .313 .163 r
F P
0 g
s
.27671 .47288 m .25905 .48444 L .26034 .47777 L closepath p .085 .236 .147 r
F P
0 g
s
.27796 .4867 m .25905 .48444 L .27716 .47971 L closepath p .101 .383 .181 r
F P
0 g
s
.27796 .4867 m .25985 .49136 L .25905 .48444 L closepath p .101 .383 .181 r
F P
0 g
s
.27796 .4867 m .26089 .49802 L .25985 .49136 L closepath p .103 .429 .192 r
F P
0 g
s
.27671 .47288 m .26034 .47777 L .26328 .47104 L closepath p .087 .258 .152 r
F P
0 g
s
.27686 .44898 m .26056 .46807 L .26055 .46442 L closepath p .077 .171 .132 r
F P
0 g
s
.27686 .44898 m .26055 .46442 L .26236 .46055 L closepath p 0 0 .023 r
F P
0 g
s
.27686 .44898 m .27699 .45541 L .26056 .46807 L closepath p .064 .101 .118 r
F P
0 g
s
.26056 .46807 m .27699 .45541 L .27978 .45837 L closepath p .076 .182 .136 r
F P
0 g
s
.26056 .46807 m .27978 .45837 L .26328 .47104 L closepath p .076 .182 .136 r
F P
0 g
s
.27796 .4867 m .28092 .49436 L .26089 .49802 L closepath p .112 .501 .205 r
F P
0 g
s
.26089 .49802 m .28092 .49436 L .28475 .49509 L closepath p .089 .772 .245 r
F P
0 g
s
.26089 .49802 m .28475 .49509 L .26471 .49874 L closepath p .089 .772 .245 r
F P
0 g
s
.27686 .44898 m .26236 .46055 L .27866 .4451 L closepath p 0 0 .023 r
F P
0 g
s
.2797 .46609 m .26328 .47104 L .27978 .45837 L closepath p .086 .245 .149 r
F P
0 g
s
.27671 .47288 m .26328 .47104 L .2797 .46609 L closepath p .087 .258 .152 r
F P
0 g
s
.27671 .47288 m .29481 .45813 L .29593 .46761 L closepath p .103 .346 .168 r
F P
0 g
s
.27671 .47288 m .2797 .46609 L .29481 .45813 L closepath p .095 .293 .158 r
F P
0 g
s
.29593 .46761 m .27716 .47971 L .27671 .47288 L closepath p .102 .333 .165 r
F P
0 g
s
.27686 .44898 m .29791 .43026 L .29516 .43544 L closepath p .003 0 .02 r
F P
0 g
s
.27686 .44898 m .27866 .4451 L .29791 .43026 L closepath p .021 0 .042 r
F P
0 g
s
.29516 .43544 m .27699 .45541 L .27686 .44898 L closepath p .072 .127 .12 r
F P
0 g
s
.29516 .43544 m .29533 .4419 L .27699 .45541 L closepath p .072 .127 .12 r
F P
0 g
s
.27978 .45837 m .27699 .45541 L .29533 .4419 L closepath p .081 .184 .133 r
F P
0 g
s
.29593 .46761 m .2964 .4745 L .27716 .47971 L closepath p .102 .333 .165 r
F P
0 g
s
.27796 .4867 m .27716 .47971 L .2964 .4745 L closepath p .107 .403 .182 r
F P
0 g
s
.27796 .4867 m .2964 .4745 L .29798 .48368 L closepath p .108 .414 .184 r
F P
0 g
s
.29798 .48368 m .28092 .49436 L .27796 .4867 L closepath p .11 .496 .205 r
F P
0 g
s
.29488 .45037 m .29481 .45813 L .2797 .46609 L closepath p .091 .262 .151 r
F P
0 g
s
.29488 .45037 m .2797 .46609 L .27978 .45837 L closepath p .091 .262 .151 r
F P
0 g
s
.27978 .45837 m .29533 .4419 L .29488 .45037 L closepath p .082 .187 .133 r
F P
0 g
s
.29798 .48368 m .30095 .49141 L .28092 .49436 L closepath p .11 .496 .205 r
F P
0 g
s
.30095 .49141 m .28475 .49509 L .28092 .49436 L closepath p .08 .756 .242 r
F P
0 g
s
.30095 .49141 m .30465 .49213 L .28475 .49509 L closepath p .08 .756 .242 r
F P
0 g
s
.31365 .43742 m .31477 .45077 L .29481 .45813 L closepath p .1 .311 .16 r
F P
0 g
s
.31365 .43742 m .29481 .45813 L .29488 .45037 L closepath p .091 .263 .151 r
F P
0 g
s
.31592 .46034 m .29481 .45813 L .31477 .45077 L closepath p .103 .347 .168 r
F P
0 g
s
.31592 .46034 m .29593 .46761 L .29481 .45813 L closepath p .103 .347 .168 r
F P
0 g
s
.31365 .43742 m .29488 .45037 L .29533 .4419 L closepath p .079 .178 .132 r
F P
0 g
s
.3171 .42245 m .29516 .43544 L .29791 .43026 L closepath p .033 0 .038 r
F P
0 g
s
.3171 .42245 m .29533 .4419 L .29516 .43544 L closepath p .086 .167 .123 r
F P
0 g
s
.3171 .42245 m .31409 .42889 L .29533 .4419 L closepath p .021 0 .045 r
F P
0 g
s
.31365 .43742 m .29533 .4419 L .31409 .42889 L closepath p .079 .178 .132 r
F P
0 g
s
.31592 .46034 m .2964 .4745 L .29593 .46761 L closepath p .102 .332 .165 r
F P
0 g
s
.31592 .46034 m .31895 .47042 L .2964 .4745 L closepath p .121 .466 .191 r
F P
0 g
s
.32053 .47968 m .2964 .4745 L .31895 .47042 L closepath p .12 .447 .186 r
F P
0 g
s
.32053 .47968 m .29798 .48368 L .2964 .4745 L closepath p .12 .447 .186 r
F P
0 g
s
.29791 .43026 m .30261 .4269 L .32323 .4127 L closepath p 0 0 0 r
F P
0 g
s
.29791 .43026 m .32323 .4127 L .31983 .41724 L closepath p 0 0 0 r
F P
0 g
s
.3171 .42245 m .29791 .43026 L .31983 .41724 L closepath p .033 0 .038 r
F P
0 g
s
.32053 .47968 m .30095 .49141 L .29798 .48368 L closepath p .122 .525 .205 r
F P
0 g
s
.32053 .47968 m .3204 .49187 L .30095 .49141 L closepath p .102 .356 .172 r
F P
0 g
s
.30095 .49141 m .32411 .49259 L .30465 .49213 L closepath p .026 .636 .217 r
F P
0 g
s
.30095 .49141 m .3204 .49187 L .32411 .49259 L closepath p .026 .636 .217 r
F P
0 g
s
.31365 .43742 m .31409 .42889 L .33545 .41909 L closepath p .096 .233 .138 r
F P
0 g
s
.31365 .43742 m .33545 .41909 L .33266 .43153 L closepath p .078 .126 .115 r
F P
0 g
s
.33266 .43153 m .31477 .45077 L .31365 .43742 L closepath p .094 .293 .159 r
F P
0 g
s
.33844 .41261 m .31409 .42889 L .3171 .42245 L closepath p .059 0 .067 r
F P
0 g
s
.33844 .41261 m .33545 .41909 L .31409 .42889 L closepath p .059 0 .067 r
F P
0 g
s
.33266 .43153 m .33381 .445 L .31477 .45077 L closepath p .094 .293 .159 r
F P
0 g
s
.31592 .46034 m .31477 .45077 L .33381 .445 L closepath p .097 .327 .167 r
F P
0 g
s
.31592 .46034 m .33381 .445 L .33607 .45867 L closepath p .101 .355 .172 r
F P
0 g
s
.33607 .45867 m .31895 .47042 L .31592 .46034 L closepath p .102 .417 .19 r
F P
0 g
s
.3171 .42245 m .31983 .41724 L .34202 .40634 L closepath p .05 0 .049 r
F P
0 g
s
.3171 .42245 m .34202 .40634 L .33844 .41261 L closepath p .045 0 .042 r
F P
0 g
s
.32132 .22265 m .31816 .19269 L .3193 .1881 L closepath p 0 0 .096 r
F P
0 g
s
.31816 .19269 m .32132 .22265 L .32194 .22954 L closepath p 0 0 .023 r
F P
0 g
s
.33607 .45867 m .33912 .46885 L .31895 .47042 L closepath p .102 .417 .19 r
F P
0 g
s
.32053 .47968 m .31895 .47042 L .33912 .46885 L closepath p .102 .394 .183 r
F P
0 g
s
.32247 .21816 m .3193 .1881 L .32102 .18459 L closepath p 0 0 .115 r
F P
0 g
s
.32132 .22265 m .3193 .1881 L .32247 .21816 L closepath p 0 0 .096 r
F P
0 g
s
.31983 .41724 m .3454 .40178 L .34202 .40634 L closepath p .02 0 0 r
F P
0 g
s
.31983 .41724 m .32323 .4127 L .3454 .40178 L closepath p .02 0 0 r
F P
0 g
s
.34128 .4807 m .3204 .49187 L .32053 .47968 L closepath p .104 .362 .173 r
F P
0 g
s
.34128 .4807 m .34113 .493 L .3204 .49187 L closepath p .104 .362 .173 r
F P
0 g
s
.34113 .493 m .32411 .49259 L .3204 .49187 L closepath p .017 .61 .211 r
F P
0 g
s
.32053 .47968 m .33912 .46885 L .34128 .4807 L closepath p .102 .401 .185 r
F P
0 g
s
.32423 .18062 m .33151 .21442 L .32102 .18459 L closepath p .107 .412 .185 r
F P
0 g
s
.32247 .21816 m .32102 .18459 L .33151 .21442 L closepath p .137 .484 .184 r
F P
0 g
s
.33055 .25235 m .32132 .22265 L .32247 .21816 L closepath p 0 0 .124 r
F P
0 g
s
.33055 .25235 m .32707 .259 L .32132 .22265 L closepath p .023 .105 .145 r
F P
0 g
s
.32707 .259 m .32194 .22954 L .32132 .22265 L closepath p 0 0 .046 r
F P
0 g
s
.32707 .259 m .32767 .26575 L .32194 .22954 L closepath p 0 0 .046 r
F P
0 g
s
.33964 .24869 m .32247 .21816 L .33151 .21442 L closepath p .13 .412 .168 r
F P
0 g
s
.33964 .24869 m .33055 .25235 L .32247 .21816 L closepath p .13 .412 .168 r
F P
0 g
s
.32974 .1786 m .32295 .17005 L .32964 .16821 L closepath p .137 .46 .177 r
F P
0 g
s
.32295 .17005 m .33101 .18919 L .32423 .18062 L closepath p .125 .447 .182 r
F P
0 g
s
.32295 .17005 m .32974 .1786 L .33101 .18919 L closepath p .125 .447 .182 r
F P
0 g
s
.34346 .49353 m .32411 .49259 L .34113 .493 L p .017 .61 .211 r
F P
0 g
s
.32423 .18062 m .33473 .21051 L .33151 .21442 L closepath p .107 .412 .185 r
F P
0 g
s
.32423 .18062 m .33101 .18919 L .33473 .21051 L closepath p .118 .439 .185 r
F P
0 g
s
.33055 .25235 m .33747 .29962 L .32707 .259 L closepath p .027 .125 .149 r
F P
0 g
s
.32707 .259 m .33747 .29962 L .33893 .3079 L closepath p 0 0 .058 r
F P
0 g
s
.32707 .259 m .33893 .3079 L .32767 .26575 L closepath p 0 0 .09 r
F P
0 g
s
.32974 .1786 m .32964 .16821 L .33245 .16699 L closepath p .105 .358 .17 r
F P
0 g
s
.32974 .1786 m .33245 .16699 L .33254 .1774 L closepath p .105 .358 .17 r
F P
0 g
s
.32974 .1786 m .33254 .1774 L .33502 .18128 L closepath p .117 .486 .198 r
F P
0 g
s
.32974 .1786 m .33544 .192 L .33101 .18919 L closepath p .149 .519 .184 r
F P
0 g
s
.32974 .1786 m .33502 .18128 L .33544 .192 L closepath p .128 .473 .188 r
F P
0 g
s
.3481 .28557 m .33055 .25235 L .33964 .24869 L closepath p .126 .373 .159 r
F P
0 g
s
.3481 .28557 m .34101 .2931 L .33055 .25235 L closepath p .119 .36 .161 r
F P
0 g
s
.33055 .25235 m .34101 .2931 L .33747 .29962 L closepath p .027 .125 .149 r
F P
0 g
s
.33544 .192 m .33473 .21051 L .33101 .18919 L closepath p .147 .521 .187 r
F P
0 g
s
.33151 .21442 m .33473 .21051 L .33916 .21334 L closepath p .128 .531 .203 r
F P
0 g
s
.33151 .21442 m .33916 .21334 L .34613 .21439 L closepath p .099 .778 .248 r
F P
0 g
s
.33151 .21442 m .34613 .21439 L .34781 .24227 L closepath p .16 .489 .166 r
F P
0 g
s
.33151 .21442 m .34781 .24227 L .33964 .24869 L closepath p .162 .492 .165 r
F P
0 g
s
.33358 .16489 m .33254 .1774 L .33245 .16699 L closepath p .003 .024 .131 r
F P
0 g
s
.33358 .16489 m .33363 .17018 L .33254 .1774 L closepath p .003 .024 .131 r
F P
0 g
s
.33611 .17406 m .33254 .1774 L .33363 .17018 L closepath p .132 .511 .195 r
F P
0 g
s
.33611 .17406 m .33502 .18128 L .33254 .1774 L closepath p .132 .511 .195 r
F P
0 g
s
.35272 .42708 m .33266 .43153 L .33545 .41909 L closepath p .085 .15 .118 r
F P
0 g
s
.35272 .42708 m .33381 .445 L .33266 .43153 L closepath p .096 .3 .159 r
F P
0 g
s
.33639 .16093 m .33592 .16252 L .33358 .16489 L closepath p .012 0 .047 r
F P
0 g
s
.33592 .16252 m .33363 .17018 L .33358 .16489 L closepath p .128 .431 .176 r
F P
0 g
s
.33592 .16252 m .33597 .16782 L .33363 .17018 L closepath p .128 .431 .176 r
F P
0 g
s
.33363 .17018 m .33597 .16782 L .33611 .17406 L closepath p .144 .531 .192 r
F P
0 g
s
.35272 .42708 m .35111 .44666 L .33381 .445 L closepath p .081 .208 .141 r
F P
0 g
s
.35339 .46048 m .33607 .45867 L .33381 .445 L closepath p .078 .288 .168 r
F P
0 g
s
.35339 .46048 m .33381 .445 L .35111 .44666 L closepath p .078 .288 .168 r
F P
0 g
s
.33544 .192 m .33916 .21334 L .33473 .21051 L closepath p .147 .521 .187 r
F P
0 g
s
.33611 .17406 m .34151 .18115 L .33502 .18128 L closepath p .168 .647 .204 r
F P
0 g
s
.33502 .18128 m .34151 .18115 L .34191 .19188 L closepath p .169 .606 .192 r
F P
0 g
s
.33502 .18128 m .34191 .19188 L .33544 .192 L closepath p .169 .606 .192 r
F P
0 g
s
.33544 .192 m .34191 .19188 L .33916 .21334 L closepath p .17 .583 .185 r
F P
0 g
s
.35774 .40598 m .33545 .41909 L .35773 .40593 L .09 .075 .085 r
F
.33545 .41909 m .33844 .41261 L .35773 .40593 L p F P
0 g
s
.35787 .40692 m .33545 .41909 L p .35774 .40598 L .09 .075 .085 r
F P
0 g
s
.3555 .41453 m .33545 .41909 L .35606 .4079 L p .055 0 .033 r
F P
0 g
s
.35272 .42708 m .33545 .41909 L .3555 .41453 L closepath p .085 .15 .118 r
F P
0 g
s
.33976 .16249 m .33592 .16252 L .33639 .16093 L closepath p .166 .351 .109 r
F P
0 g
s
.33927 .16775 m .33592 .16252 L .33976 .16249 L closepath p .176 .576 .177 r
F P
0 g
s
.33927 .16775 m .33597 .16782 L .33592 .16252 L closepath p .17 .568 .181 r
F P
0 g
s
.33597 .16782 m .33927 .16775 L .33941 .17399 L closepath p .169 .613 .194 r
F P
0 g
s
.33597 .16782 m .33941 .17399 L .33611 .17406 L closepath p .169 .613 .194 r
F P
0 g
s
.35339 .46048 m .33912 .46885 L .33607 .45867 L closepath p .074 .344 .186 r
F P
0 g
s
.33611 .17406 m .33941 .17399 L .34151 .18115 L closepath p .168 .647 .204 r
F P
0 g
s
.33976 .16249 m .33639 .16093 L .34022 .16091 L closepath p .166 .351 .109 r
F P
0 g
s
.34536 .31837 m .33747 .29962 L .34101 .2931 L closepath p .019 .082 .14 r
F P
0 g
s
.34536 .31837 m .34562 .32823 L .33747 .29962 L closepath p 0 0 .122 r
F P
0 g
s
.33893 .3079 m .33747 .29962 L .34562 .32823 L closepath p 0 0 .07 r
F P
0 g
s
.35756 .40475 m .33844 .41261 L .35755 .40467 L .079 0 .063 r
F
.33844 .41261 m .34202 .40634 L .35755 .40467 L p F P
0 g
s
.35773 .40593 m .33844 .41261 L p .35756 .40475 L .079 0 .063 r
F P
0 g
s
.34136 .31644 m .33893 .3079 L .34562 .32823 L p 0 0 .07 r
F P
0 g
s
.35339 .46048 m .3599 .47217 L .33912 .46885 L closepath p .085 .437 .204 r
F P
0 g
s
.36207 .48413 m .34128 .4807 L .33912 .46885 L closepath p .093 .375 .183 r
F P
0 g
s
.36207 .48413 m .33912 .46885 L .3599 .47217 L closepath p .093 .375 .183 r
F P
0 g
s
.34889 .19295 m .33916 .21334 L .34191 .19188 L closepath p .193 .667 .187 r
F P
0 g
s
.34889 .19295 m .34613 .21439 L .33916 .21334 L closepath p .193 .667 .187 r
F P
0 g
s
.33927 .16775 m .33976 .16249 L .34271 .16479 L closepath p .204 .676 .178 r
F P
0 g
s
.33927 .16775 m .34271 .16479 L .34222 .17004 L closepath p .204 .676 .178 r
F P
0 g
s
.33941 .17399 m .33927 .16775 L .34222 .17004 L closepath p .193 .69 .193 r
F P
0 g
s
.34222 .17004 m .34151 .18115 L .33941 .17399 L closepath p .197 .715 .195 r
F P
0 g
s
.3481 .28557 m .33964 .24869 L .34781 .24227 L closepath p .155 .447 .157 r
F P
0 g
s
.34271 .16479 m .33976 .16249 L .34022 .16091 L closepath p .222 .598 .124 r
F P
0 g
s
.34022 .16091 m .34527 .16687 L .34271 .16479 L closepath p .229 .515 .047 r
F P
0 g
s
.34022 .16091 m .34279 .16299 L .34527 .16687 L closepath p .229 .515 .047 r
F P
0 g
s
.35248 .31087 m .34101 .2931 L .3481 .28557 L closepath p .119 .361 .161 r
F P
0 g
s
.35248 .31087 m .34536 .31837 L .34101 .2931 L closepath p .119 .361 .161 r
F P
0 g
s
.36207 .48413 m .34113 .493 L .34128 .4807 L closepath p .097 .338 .17 r
F P
0 g
s
.36207 .48413 m .36137 .49758 L .34113 .493 L closepath p .097 .339 .17 r
F P
0 g
s
.34702 .33629 m .34136 .31644 L p .34562 .32823 L 0 0 .07 r
F P
0 g
s
.34222 .17004 m .34431 .1772 L .34151 .18115 L closepath p .197 .715 .195 r
F P
0 g
s
.34749 .17741 m .34151 .18115 L .34431 .1772 L closepath p .187 .65 .188 r
F P
0 g
s
.34749 .17741 m .34191 .19188 L .34151 .18115 L closepath p .188 .664 .191 r
F P
0 g
s
.34749 .17741 m .34889 .19295 L .34191 .19188 L closepath p .191 .68 .192 r
F P
0 g
s
.34202 .40634 m .35657 .39778 L .3566 .39797 L .037 0 .008 r
F
.34202 .40634 m .3454 .40178 L .35657 .39778 L p F P
0 g
s
.34202 .40634 m .35661 .39805 L .35665 .39833 L .037 0 .008 r
F
.34202 .40634 m .35661 .39803 L .35661 .39805 L F
.34202 .40634 m .3566 .39797 L .35661 .39803 L F
.34202 .40634 m .35665 .39833 L .35675 .39903 L F
.35676 .39908 m .34202 .40634 L p .35675 .39903 L F P
0 g
s
.35705 .40111 m .34202 .40634 L .35676 .39908 L p .109 .23 .126 r
F P
0 g
s
.35755 .40467 m .34202 .40634 L .35705 .40111 L p .079 0 .063 r
F P
0 g
s
.34222 .17004 m .34271 .16479 L .34527 .16687 L closepath p .231 .776 .17 r
F P
0 g
s
.34222 .17004 m .34527 .16687 L .34431 .1772 L closepath p .231 .776 .17 r
F P
0 g
s
.34527 .16687 m .34279 .16299 L .34537 .16687 L p .154 .19 0 r
F P
0 g
s
.34431 .1772 m .34527 .16687 L .34845 .16708 L closepath p .188 .619 .178 r
F P
0 g
s
.34431 .1772 m .34845 .16708 L .34749 .17741 L closepath p .188 .619 .178 r
F P
0 g
s
.34769 .33179 m .34536 .31837 L .35248 .31087 L closepath p .124 .39 .166 r
F P
0 g
s
.34536 .31837 m .34794 .3416 L .34562 .32823 L closepath p 0 0 .08 r
F P
0 g
s
.34536 .31837 m .34769 .33179 L .34794 .3416 L closepath p 0 0 .08 r
F P
0 g
s
.35397 .19325 m .34613 .21439 L .34889 .19295 L closepath p .203 .708 .187 r
F P
0 g
s
.35397 .19325 m .35343 .21128 L .34613 .21439 L closepath p .193 .66 .185 r
F P
0 g
s
.35513 .23913 m .34613 .21439 L .35343 .21128 L closepath p .185 .567 .165 r
F P
0 g
s
.35513 .23913 m .34781 .24227 L .34613 .21439 L closepath p .185 .567 .165 r
F P
0 g
s
.35226 .16541 m .35177 .17066 L .34749 .17741 L closepath p .139 .444 .171 r
F P
0 g
s
.35226 .16541 m .34749 .17741 L .34845 .16708 L closepath p .139 .444 .171 r
F P
0 g
s
.35333 .17467 m .34749 .17741 L .35177 .17066 L closepath p .172 .616 .193 r
F P
0 g
s
.35333 .17467 m .34905 .18142 L .34749 .17741 L closepath p .172 .616 .193 r
F P
0 g
s
.34749 .17741 m .34905 .18142 L .34889 .19295 L closepath p .166 .603 .195 r
F P
0 g
s
.34769 .33179 m .35248 .31087 L .35687 .31055 L closepath p .119 .369 .164 r
F P
0 g
s
.34769 .33179 m .35687 .31055 L .35207 .3315 L closepath p .119 .369 .164 r
F P
0 g
s
.34794 .3416 m .34769 .33179 L .35207 .3315 L closepath p .117 .578 .221 r
F P
0 g
s
.35513 .23913 m .35632 .27912 L .34781 .24227 L closepath p .179 .507 .152 r
F P
0 g
s
.3481 .28557 m .34781 .24227 L .35632 .27912 L closepath p .155 .447 .157 r
F P
0 g
s
.34794 .3416 m .35207 .3315 L .35537 .3445 L closepath p .012 .09 .146 r
F P
0 g
s
.34794 .3416 m .35537 .3445 L .35067 .35817 L closepath p .014 .081 .142 r
F P
0 g
s
.3481 .28557 m .35632 .27912 L .36456 .29475 L closepath p .156 .456 .158 r
F P
0 g
s
.3481 .28557 m .36456 .29475 L .36732 .30083 L closepath p .166 .418 .134 r
F P
0 g
s
.3481 .28557 m .36732 .30083 L .35687 .31055 L closepath p .153 .436 .155 r
F P
0 g
s
.3481 .28557 m .35687 .31055 L .35248 .31087 L closepath p .118 .359 .161 r
F P
0 g
s
.35226 .16541 m .34845 .16708 L .35141 .16369 L closepath p .092 .072 .082 r
F P
0 g
s
.34905 .18142 m .35397 .19325 L .34889 .19295 L closepath p .202 .724 .193 r
F P
0 g
s
.35333 .17467 m .35413 .18172 L .34905 .18142 L closepath p .198 .768 .207 r
F P
0 g
s
.34905 .18142 m .35413 .18172 L .35397 .19325 L closepath p .202 .724 .193 r
F P
0 g
s
.35412 .37813 m .35067 .35817 L .35537 .3445 L closepath p .006 .043 .135 r
F P
0 g
s
.35111 .44666 m .35272 .42708 L .36037 .42595 L p .043 .082 .126 r
F P
0 g
s
.36039 .42604 m .35111 .44666 L .36037 .42595 L .043 .082 .126 r
F
.36078 .42892 m .35111 .44666 L .36039 .42604 L F
.36082 .42922 m .35111 .44666 L p .36078 .42892 L F P
0 g
s
.36123 .44532 m .35111 .44666 L .36082 .42922 L p .043 .082 .126 r
F P
0 g
s
.35339 .46048 m .35111 .44666 L .36123 .44532 L closepath p .048 .202 .16 r
F P
0 g
s
.35226 .16541 m .35141 .16369 L .35522 .16202 L closepath p .092 .072 .082 r
F P
0 g
s
.3558 .16351 m .35532 .16877 L .35177 .17066 L closepath p .17 .557 .177 r
F P
0 g
s
.3558 .16351 m .35177 .17066 L .35226 .16541 L closepath p .17 .557 .177 r
F P
0 g
s
.35177 .17066 m .35532 .16877 L .35333 .17467 L closepath p .178 .629 .191 r
F P
0 g
s
.35207 .3315 m .35687 .31055 L .36294 .31715 L closepath p .115 .355 .162 r
F P
0 g
s
.35207 .3315 m .36294 .31715 L .36626 .33017 L closepath p .116 .359 .163 r
F P
0 g
s
.35207 .3315 m .36626 .33017 L .35537 .3445 L closepath p .116 .359 .163 r
F P
0 g
s
.35522 .16202 m .3558 .16351 L .35226 .16541 L closepath p .138 .3 .122 r
F P
0 g
s
.35272 .42708 m .3555 .41453 L .35967 .42081 L p .038 0 .096 r
F P
0 g
s
.36037 .42595 m .35272 .42708 L p .35971 .42113 L .038 0 .096 r
F P
0 g
s
.35971 .42113 m .35272 .42708 L .3597 .42102 L .038 0 .096 r
F
.3597 .42102 m .35272 .42708 L .35967 .42081 L F
.35532 .16877 m .35591 .17483 L .35333 .17467 L closepath p .202 .735 .196 r
F P
0 g
s
.35333 .17467 m .35591 .17483 L .35413 .18172 L closepath p .198 .768 .207 r
F P
0 g
s
.35339 .46048 m .36123 .44532 L .36258 .46024 L closepath p .041 .168 .154 r
F P
0 g
s
.36258 .46024 m .3599 .47217 L .35339 .46048 L closepath p .04 .333 .197 r
F P
0 g
s
.35397 .19325 m .36145 .19419 L .35343 .21128 L closepath p .205 .7 .183 r
F P
0 g
s
.36886 .19768 m .35343 .21128 L .36145 .19419 L closepath p .217 .762 .186 r
F P
0 g
s
.36886 .19768 m .36087 .21472 L .35343 .21128 L closepath p .217 .762 .186 r
F P
0 g
s
.36552 .23904 m .35343 .21128 L .36087 .21472 L closepath p .227 .666 .141 r
F P
0 g
s
.36552 .23904 m .35513 .23913 L .35343 .21128 L closepath p .213 .653 .158 r
F P
0 g
s
.35397 .19325 m .35413 .18172 L .35708 .17833 L closepath p .214 .763 .189 r
F P
0 g
s
.35397 .19325 m .35708 .17833 L .35873 .17877 L closepath p .2 .704 .19 r
F P
0 g
s
.35397 .19325 m .35873 .17877 L .36052 .18264 L closepath p .194 .676 .189 r
F P
0 g
s
.36052 .18264 m .36145 .19419 L .35397 .19325 L closepath p .203 .73 .193 r
F P
0 g
s
.35412 .37813 m .35537 .3445 L .35885 .36455 L closepath p .006 .043 .135 r
F P
0 g
s
.36586 .38444 m .35412 .37813 L .35885 .36455 L closepath p .034 .172 .159 r
F P
0 g
s
.35772 .40349 m .35484 .3856 L .35412 .37813 L closepath p 0 0 .032 r
F P
0 g
s
.36586 .38444 m .35772 .40349 L .35412 .37813 L closepath p .041 .16 .152 r
F P
0 g
s
.35886 .17145 m .35413 .18172 L .35591 .17483 L closepath p .227 .89 .2 r
F P
0 g
s
.35886 .17145 m .35708 .17833 L .35413 .18172 L closepath p .227 .89 .2 r
F P
0 g
s
.35772 .40349 m .35841 .41074 L .35484 .3856 L closepath p 0 0 .032 r
F P
0 g
s
.36552 .23904 m .36547 .26773 L .35513 .23913 L closepath p .21 .576 .135 r
F P
0 g
s
.35513 .23913 m .36547 .26773 L .35632 .27912 L closepath p .188 .535 .151 r
F P
0 g
s
.3593 .1642 m .3558 .16351 L .35522 .16202 L closepath p .196 .414 .089 r
F P
0 g
s
.3593 .1642 m .35522 .16202 L .35871 .16271 L closepath p .196 .414 .089 r
F P
0 g
s
.3579 .16893 m .35532 .16877 L .3558 .16351 L closepath p .204 .679 .178 r
F P
0 g
s
.35532 .16877 m .3579 .16893 L .35591 .17483 L closepath p .202 .735 .196 r
F P
0 g
s
.36626 .33017 m .35885 .36455 L .35537 .3445 L closepath p .112 .339 .159 r
F P
0 g
s
.35821 .41013 m .3555 .41453 L .35606 .4079 L .055 0 .033 r
F
.35823 .41024 m .3555 .41453 L .35821 .41013 L F
.35829 .4107 m .3555 .41453 L p .35823 .41024 L F P
0 g
s
.3555 .41453 m .35829 .4107 L p .35831 .41087 L .007 0 0 r
F P
0 g
s
.35838 .41137 m .3555 .41453 L .35831 .41087 L .007 0 0 r
F
.35844 .41179 m .3555 .41453 L .35838 .41137 L F
.35848 .41212 m .3555 .41453 L p .35844 .41179 L F P
0 g
s
.35861 .41306 m .3555 .41453 L .35848 .41212 L p .007 0 0 r
F P
0 g
s
.35967 .42081 m .3555 .41453 L .35861 .41306 L p .052 0 .09 r
F P
0 g
s
.3579 .16893 m .3558 .16351 L .3593 .1642 L closepath p .201 .673 .18 r
F P
0 g
s
.35591 .17483 m .3579 .16893 L .35886 .17145 L closepath p .223 .835 .196 r
F P
0 g
s
.35606 .4079 m .35778 .40697 L p .35821 .41013 L .055 0 .033 r
F P
0 g
s
.37372 .28329 m .35632 .27912 L .36547 .26773 L closepath p .18 .492 .145 r
F P
0 g
s
.37372 .28329 m .36456 .29475 L .35632 .27912 L closepath p .18 .492 .145 r
F P
0 g
s
.36294 .31715 m .35687 .31055 L .36732 .30083 L closepath p .142 .368 .143 r
F P
0 g
s
.35886 .17145 m .35983 .16905 L .35708 .17833 L closepath p .2 .811 .082 r
F P
0 g
s
.36147 .16949 m .35708 .17833 L .35983 .16905 L closepath p .203 .682 .18 r
F P
0 g
s
.36147 .16949 m .35873 .17877 L .35708 .17833 L closepath p .203 .682 .18 r
F P
0 g
s
.36586 .38444 m .3678 .43868 L .35772 .40349 L closepath p .055 .227 .163 r
F P
0 g
s
.35841 .41074 m .35772 .40349 L .3678 .43868 L closepath p 0 0 .074 r
F P
0 g
s
.35816 .40892 m .35816 .40902 L .35778 .40697 L .055 0 .033 r
F
.35778 .40697 m .35787 .40692 L p .35816 .40892 L F P
0 g
s
.35816 .40902 m .35816 .40919 L .35778 .40697 L .055 0 .033 r
F
.35816 .40919 m .35817 .40964 L .35778 .40697 L F
.35817 .40964 m .35817 .40987 L .35778 .40697 L F
.3579 .16893 m .3593 .1642 L .36026 .16672 L closepath p .237 .845 .177 r
F P
0 g
s
.3579 .16893 m .36026 .16672 L .35886 .17145 L closepath p .237 .845 .177 r
F P
0 g
s
.35841 .41074 m .35817 .40987 L .35816 .40892 L p 0 0 .01 r
F P
0 g
s
.3623 .44005 m .35817 .40987 L .35841 .41074 L closepath p 0 0 .027 r
F P
0 g
s
.35841 .41074 m .3678 .43868 L .36257 .44113 L closepath p 0 0 .098 r
F P
0 g
s
.3623 .44005 m .35841 .41074 L .36257 .44113 L closepath p 0 0 .027 r
F P
0 g
s
.36519 .16805 m .3638 .17277 L .35873 .17877 L closepath p .135 .397 .159 r
F P
0 g
s
.36519 .16805 m .35873 .17877 L .36147 .16949 L closepath p .135 .397 .159 r
F P
0 g
s
.36559 .17664 m .35873 .17877 L .3638 .17277 L closepath p .178 .65 .196 r
F P
0 g
s
.36559 .17664 m .36052 .18264 L .35873 .17877 L closepath p .178 .65 .196 r
F P
0 g
s
.36626 .33017 m .37528 .34387 L .35885 .36455 L closepath p .125 .39 .165 r
F P
0 g
s
.38236 .36379 m .35885 .36455 L .37528 .34387 L closepath p .124 .382 .164 r
F P
0 g
s
.38236 .36379 m .36586 .38444 L .35885 .36455 L closepath p .124 .382 .164 r
F P
0 g
s
.35886 .17145 m .36026 .16672 L .35983 .16905 L closepath p .2 .811 .082 r
F P
0 g
s
.35974 .16535 m .36135 .1657 L p .36026 .16672 L .201 .673 .179 r
F P
0 g
s
.36509 .1665 m .35983 .16905 L p .35991 .16858 L .201 .673 .179 r
F P
0 g
s
.36509 .1665 m .36147 .16949 L .35983 .16905 L closepath p .202 .686 .182 r
F P
0 g
s
.36258 .46024 m .36913 .47202 L .3599 .47217 L closepath p .04 .333 .197 r
F P
0 g
s
.36207 .48413 m .3599 .47217 L .36913 .47202 L closepath p .042 .222 .169 r
F P
0 g
s
.36509 .1665 m .35991 .16858 L .36012 .16747 L .201 .673 .179 r
F
.36135 .1657 m .36509 .1665 L p .36012 .16747 L F P
0 g
s
.36012 .16747 m .36014 .16734 L .36135 .1657 L .201 .673 .179 r
F
.36014 .16734 m .36026 .16672 L .36135 .1657 L F
.36559 .17664 m .36799 .1836 L .36052 .18264 L closepath p .197 .775 .21 r
F P
0 g
s
.36052 .18264 m .36799 .1836 L .36145 .19419 L closepath p .203 .73 .193 r
F P
0 g
s
.36087 .42955 m .36123 .44532 L .36082 .42922 L .043 .082 .126 r
F
.36183 .43664 m .36123 .44532 L .36087 .42955 L F
.37064 .20243 m .36087 .21472 L .36886 .19768 L closepath p .238 .901 .185 r
F P
0 g
s
.37064 .20243 m .37007 .21773 L .36087 .21472 L closepath p .23 .799 .178 r
F P
0 g
s
.37472 .24195 m .36087 .21472 L .37007 .21773 L closepath p .234 .683 .132 r
F P
0 g
s
.37472 .24195 m .36552 .23904 L .36087 .21472 L closepath p .234 .683 .132 r
F P
0 g
s
.3619 .43712 m .36123 .44532 L p .36183 .43664 L .043 .082 .126 r
F P
0 g
s
.36225 .446 m .36123 .44532 L .36179 .43844 L p 0 0 .035 r
F P
0 g
s
.36446 .46159 m .36123 .44532 L .3631 .44656 L closepath p 0 0 .062 r
F P
0 g
s
.36446 .46159 m .36258 .46024 L .36123 .44532 L closepath p 0 0 .062 r
F P
0 g
s
.3817 .49378 m .36137 .49758 L .36207 .48413 L closepath p .083 .29 .165 r
F P
0 g
s
.3817 .49378 m .38097 .50736 L .36137 .49758 L closepath p .083 .29 .165 r
F P
0 g
s
.37384 .18159 m .36886 .19768 L .36145 .19419 L closepath p .212 .779 .196 r
F P
0 g
s
.37384 .18159 m .36145 .19419 L .36799 .1836 L closepath p .214 .796 .198 r
F P
0 g
s
.36519 .16805 m .36147 .16949 L .36509 .1665 L closepath p .074 0 .04 r
F P
0 g
s
.36179 .43844 m .3619 .43712 L p .3623 .44005 L 0 0 .035 r
F P
0 g
s
.3623 .44005 m .36225 .446 L .36179 .43844 L 0 0 .035 r
F
.36207 .48413 m .36913 .47202 L .37698 .48215 L closepath p .045 .238 .172 r
F P
0 g
s
.36207 .48413 m .37698 .48215 L .3817 .49378 L closepath p .047 .34 .196 r
F P
0 g
s
.36235 .44044 m .36225 .446 L .3623 .44005 L 0 0 .035 r
F
.36253 .44618 m .36225 .446 L p .36238 .44068 L F P
0 g
s
.36238 .44068 m .36225 .446 L .36235 .44044 L 0 0 .035 r
F
.36235 .44044 m .3623 .44005 L p .36251 .4409 L 0 0 .106 r
F P
0 g
s
.36257 .44113 m .36235 .44044 L .36251 .4409 L 0 0 .106 r
F
.37235 .47176 m .3631 .44656 L .36235 .44044 L p F P
0 g
s
.36239 .44075 m .36253 .44618 L .36238 .44068 L 0 0 .035 r
F
.3624 .44085 m .36253 .44618 L .36239 .44075 L F
.36256 .44218 m .36253 .44618 L .3624 .44085 L F
.36259 .44239 m .36253 .44618 L .36256 .44218 L F
.3631 .44656 m .36253 .44618 L p .36259 .44239 L F P
0 g
s
.37792 .4655 m .36257 .44113 L .3678 .43868 L closepath p 0 0 .132 r
F P
0 g
s
.37235 .47176 m .36257 .44113 L .37255 .46758 L closepath p 0 0 .105 r
F P
0 g
s
.37792 .4655 m .37255 .46758 L .36257 .44113 L closepath p 0 0 .132 r
F P
0 g
s
.36446 .46159 m .36913 .47202 L .36258 .46024 L closepath p 0 0 .112 r
F P
0 g
s
.36294 .31715 m .36732 .30083 L .37341 .30741 L closepath p .142 .368 .143 r
F P
0 g
s
.37341 .30741 m .36626 .33017 L .36294 .31715 L closepath p .157 .463 .16 r
F P
0 g
s
.37235 .47176 m .36446 .46159 L .3631 .44656 L closepath p .077 .28 .166 r
F P
0 g
s
.36922 .16649 m .3638 .17277 L .36519 .16805 L closepath p .175 .561 .174 r
F P
0 g
s
.36922 .16649 m .36782 .1712 L .3638 .17277 L closepath p .175 .561 .174 r
F P
0 g
s
.3638 .17277 m .36782 .1712 L .36559 .17664 L closepath p .182 .657 .194 r
F P
0 g
s
.37235 .47176 m .36913 .47202 L .36446 .46159 L closepath p 0 0 .14 r
F P
0 g
s
.36456 .29475 m .37372 .28329 L .37964 .28639 L closepath p .132 .243 .106 r
F P
0 g
s
.36456 .29475 m .37964 .28639 L .38243 .29246 L closepath p .155 .398 .14 r
F P
0 g
s
.36456 .29475 m .38243 .29246 L .36732 .30083 L closepath p .155 .398 .14 r
F P
0 g
s
.36519 .16805 m .36509 .1665 L .36881 .16506 L closepath p .074 0 .04 r
F P
0 g
s
.36881 .16506 m .36922 .16649 L .36519 .16805 L closepath p .149 .325 .119 r
F P
0 g
s
.36552 .23904 m .37587 .26756 L .36547 .26773 L closepath p .21 .576 .135 r
F P
0 g
s
.37587 .26756 m .37372 .28329 L .36547 .26773 L closepath p .206 .521 .119 r
F P
0 g
s
.39165 .27855 m .37587 .26756 L .36552 .23904 L closepath p .222 .596 .121 r
F P
0 g
s
.39165 .27855 m .36552 .23904 L .37472 .24195 L closepath p .229 .574 .094 r
F P
0 g
s
.36782 .1712 m .36939 .17713 L .36559 .17664 L closepath p .201 .747 .2 r
F P
0 g
s
.36559 .17664 m .36939 .17713 L .36799 .1836 L closepath p .197 .775 .21 r
F P
0 g
s
.36586 .38444 m .38236 .36379 L .38789 .37428 L closepath p .11 .31 .151 r
F P
0 g
s
.36586 .38444 m .38789 .37428 L .37605 .41985 L closepath p .119 .389 .169 r
F P
0 g
s
.36586 .38444 m .37605 .41985 L .3678 .43868 L closepath p .055 .227 .163 r
F P
0 g
s
.37341 .30741 m .38298 .3158 L .36626 .33017 L closepath p .157 .462 .16 r
F P
0 g
s
.38298 .3158 m .37528 .34387 L .36626 .33017 L closepath p .151 .423 .153 r
F P
0 g
s
.38243 .29246 m .37341 .30741 L .36732 .30083 L closepath p .151 .37 .134 r
F P
0 g
s
.37792 .4655 m .3678 .43868 L .37605 .41985 L closepath p .09 .389 .189 r
F P
0 g
s
.36922 .16649 m .37162 .1717 L .36782 .1712 L closepath p .206 .694 .181 r
F P
0 g
s
.36782 .1712 m .37162 .1717 L .36939 .17713 L closepath p .201 .747 .2 r
F P
0 g
s
.37299 .17414 m .3716 .18061 L .36799 .1836 L closepath p .218 .861 .206 r
F P
0 g
s
.37299 .17414 m .36799 .1836 L .36939 .17713 L closepath p .218 .861 .206 r
F P
0 g
s
.37384 .18159 m .36799 .1836 L .3716 .18061 L closepath p .214 .795 .198 r
F P
0 g
s
.37269 .16717 m .36922 .16649 L .36881 .16506 L closepath p .198 .441 .098 r
F P
0 g
s
.37269 .16717 m .36881 .16506 L .37228 .16574 L closepath p .198 .441 .098 r
F P
0 g
s
.37384 .18159 m .37372 .18675 L .36886 .19768 L closepath p .224 .836 .195 r
F P
0 g
s
.37372 .18675 m .37064 .20243 L .36886 .19768 L closepath p .234 .901 .191 r
F P
0 g
s
.37235 .47176 m .37698 .48215 L .36913 .47202 L closepath p 0 0 .14 r
F P
0 g
s
.36922 .16649 m .37269 .16717 L .37162 .1717 L closepath p .2 .686 .185 r
F P
0 g
s
.36939 .17713 m .37162 .1717 L .37299 .17414 L closepath p .216 .822 .201 r
F P
0 g
s
.37289 .19925 m .37007 .21773 L .37064 .20243 L closepath p .249 .858 .132 r
F P
0 g
s
.37289 .19925 m .37232 .21453 L .37007 .21773 L closepath p .249 .858 .132 r
F P
0 g
s
.37472 .24195 m .37007 .21773 L .37232 .21453 L closepath p .242 .699 .117 r
F P
0 g
s
.37372 .18675 m .37549 .19153 L .37064 .20243 L closepath p .234 .901 .191 r
F P
0 g
s
.37612 .19067 m .37064 .20243 L .37549 .19153 L p .224 .946 .147 r
F P
0 g
s
.3731 .19715 m .37289 .19925 L .37064 .20243 L p .246 .881 .133 r
F P
0 g
s
.37299 .17414 m .3737 .17171 L .3716 .18061 L closepath p .19 .755 .06 r
F P
0 g
s
.3716 .18061 m .3737 .17171 L .37595 .1727 L closepath p .22 .771 .185 r
F P
0 g
s
.3716 .18061 m .37595 .1727 L .37384 .18159 L closepath p .22 .771 .185 r
F P
0 g
s
.37162 .1717 m .37269 .16717 L .37406 .16961 L closepath p .232 .827 .181 r
F P
0 g
s
.37162 .1717 m .37406 .16961 L .37299 .17414 L closepath p .232 .827 .181 r
F P
0 g
s
.37406 .16961 m .37269 .16717 L .37228 .16574 L closepath p .214 .491 0 r
F P
0 g
s
.37255 .46758 m .39427 .5011 L .37235 .47176 L closepath p 0 .067 .157 r
F P
0 g
s
.37235 .47176 m .39427 .5011 L .37698 .48215 L closepath p 0 .027 .147 r
F P
0 g
s
.37792 .4655 m .39463 .49712 L .37255 .46758 L closepath p 0 .094 .163 r
F P
0 g
s
.37255 .46758 m .39463 .49712 L .39427 .5011 L closepath p 0 .067 .157 r
F P
0 g
s
.37299 .17414 m .37406 .16961 L .3737 .17171 L closepath p .19 .755 .06 r
F P
0 g
s
.38243 .29246 m .38963 .30047 L .37341 .30741 L closepath p .184 .533 .155 r
F P
0 g
s
.39922 .30884 m .37341 .30741 L .38963 .30047 L closepath p .174 .447 .136 r
F P
0 g
s
.39922 .30884 m .38298 .3158 L .37341 .30741 L closepath p .174 .447 .136 r
F P
0 g
s
.37393 .17039 m .37431 .17197 L .3737 .17171 L .155 .386 .135 r
F
.37587 .26756 m .38206 .28316 L .37372 .28329 L closepath p .203 .508 .119 r
F P
0 g
s
.37964 .28639 m .37372 .28329 L .38206 .28316 L closepath p .141 .034 0 r
F P
0 g
s
.3796 .18228 m .37372 .18675 L .37384 .18159 L closepath p .203 .77 .203 r
F P
0 g
s
.3796 .18228 m .37549 .19153 L .37372 .18675 L closepath p .21 .878 .217 r
F P
0 g
s
.3808 .1726 m .37384 .18159 L .37595 .1727 L closepath p .19 .643 .184 r
F P
0 g
s
.3808 .1726 m .37973 .17712 L .37384 .18159 L closepath p .19 .643 .184 r
F P
0 g
s
.3796 .18228 m .37384 .18159 L .37973 .17712 L closepath p .203 .77 .203 r
F P
0 g
s
.37393 .17039 m .37745 .16999 L p .37431 .17197 L .155 .386 .135 r
F P
0 g
s
.37431 .17197 m .37745 .16999 L .37435 .17199 L .155 .386 .135 r
F
.37435 .17199 m .37745 .16999 L .37437 .172 L F
.37745 .16999 m .37595 .1727 L p .37437 .172 L F P
0 g
s
.37454 .23986 m .37566 .24398 L p .37472 .24195 L .185 .451 0 r
F P
0 g
s
.38298 .3158 m .39206 .32947 L .37528 .34387 L closepath p .151 .423 .153 r
F P
0 g
s
.39636 .34491 m .37528 .34387 L .39206 .32947 L closepath p .128 .285 .127 r
F P
0 g
s
.39636 .34491 m .38236 .36379 L .37528 .34387 L closepath p .129 .395 .163 r
F P
0 g
s
.3796 .18228 m .3805 .18471 L .37549 .19153 L closepath p .21 .878 .217 r
F P
0 g
s
.37587 .26756 m .39548 .29223 L .39238 .28975 L .223 .537 .093 r
F
.37587 .26756 m .39787 .29413 L p .39548 .29223 L F P
0 g
s
.37874 .27481 m .37587 .26756 L p .39238 .28975 L .223 .537 .093 r
F P
0 g
s
.37587 .26756 m .39165 .27855 L .39787 .29413 L closepath p .223 .537 .093 r
F P
0 g
s
.3808 .1726 m .37595 .1727 L .37745 .16999 L closepath p .188 .547 .155 r
F P
0 g
s
.37605 .41985 m .38789 .37428 L .40179 .39062 L closepath p .11 .352 .165 r
F P
0 g
s
.37605 .41985 m .40179 .39062 L .38983 .43637 L closepath p .11 .352 .165 r
F P
0 g
s
.37792 .4655 m .37605 .41985 L .38983 .43637 L closepath p .063 .298 .179 r
F P
0 g
s
.39902 .51285 m .37698 .48215 L .39427 .5011 L closepath p 0 .086 .161 r
F P
0 g
s
.39902 .51285 m .3817 .49378 L .37698 .48215 L closepath p 0 .086 .161 r
F P
0 g
s
.3808 .1726 m .37745 .16999 L .3823 .16989 L closepath p .188 .547 .155 r
F P
0 g
s
.37792 .4655 m .38983 .43637 L .3933 .47285 L closepath p .067 .317 .182 r
F P
0 g
s
.3933 .47285 m .39463 .49712 L .37792 .4655 L closepath p .025 .357 .206 r
F P
0 g
s
.39208 .28952 m .39173 .28923 L .37874 .27481 L .223 .537 .093 r
F
.39238 .28975 m .39208 .28952 L .37874 .27481 L F
.3811 .28076 m .37874 .27481 L p .39173 .28923 L F P
0 g
s
.37973 .17712 m .38288 .1768 L .3796 .18228 L closepath p .207 .782 .202 r
F P
0 g
s
.38288 .1768 m .3805 .18471 L .3796 .18228 L closepath p .223 .895 .205 r
F P
0 g
s
.37964 .28639 m .38206 .28316 L .38799 .28625 L closepath p .141 .034 0 r
F P
0 g
s
.38799 .28625 m .38243 .29246 L .37964 .28639 L closepath p .2 .471 .109 r
F P
0 g
s
.38394 .17229 m .37973 .17712 L .3808 .1726 L closepath p .208 .721 .185 r
F P
0 g
s
.38394 .17229 m .38288 .1768 L .37973 .17712 L closepath p .208 .721 .185 r
F P
0 g
s
.38288 .1768 m .38376 .17924 L .3805 .18471 L closepath p .223 .895 .205 r
F P
0 g
s
.3823 .16989 m .38394 .17229 L .3808 .1726 L closepath p .207 .647 .165 r
F P
0 g
s
.39902 .51285 m .38097 .50736 L .3817 .49378 L closepath p .047 .168 .151 r
F P
0 g
s
.39902 .51285 m .40111 .51955 L .38097 .50736 L closepath p .011 .324 .202 r
F P
0 g
s
.38759 .28604 m .38757 .28604 L .3811 .28076 L .223 .537 .093 r
F
.38759 .28605 m .38759 .28604 L .3811 .28076 L F
.38799 .28625 m .38759 .28605 L .3811 .28076 L F
.38206 .28316 m .3811 .28076 L p .38757 .28604 L F P
0 g
s
.38489 .17315 m .3823 .16989 L .38324 .17076 L closepath p .211 .407 0 r
F P
0 g
s
.38489 .17315 m .38394 .17229 L .3823 .16989 L closepath p .211 .407 0 r
F P
0 g
s
.39636 .34491 m .38789 .37428 L .38236 .36379 L closepath p .114 .317 .15 r
F P
0 g
s
.38799 .28625 m .39237 .28995 L .38243 .29246 L closepath p .184 .381 .095 r
F P
0 g
s
.39237 .28995 m .38963 .30047 L .38243 .29246 L closepath p .2 .522 .13 r
F P
0 g
s
.38376 .17924 m .38288 .1768 L .38394 .17229 L closepath p .245 .893 .167 r
F P
0 g
s
.4152 .31418 m .38298 .3158 L .39922 .30884 L closepath p .172 .428 .132 r
F P
0 g
s
.4152 .31418 m .39206 .32947 L .38298 .3158 L closepath p .173 .446 .137 r
F P
0 g
s
.38376 .17924 m .38394 .17229 L .38489 .17315 L closepath p .236 .871 .184 r
F P
0 g
s
.39636 .34491 m .40195 .35537 L .38789 .37428 L closepath p .114 .317 .15 r
F P
0 g
s
.40729 .3638 m .38789 .37428 L .40195 .35537 L closepath p .118 .335 .153 r
F P
0 g
s
.40729 .3638 m .40179 .39062 L .38789 .37428 L closepath p .121 .362 .159 r
F P
0 g
s
.38799 .28625 m .40223 .29781 L .39237 .28995 L closepath p .121 0 0 r
F P
0 g
s
.38799 .28625 m .39787 .29413 L .40223 .29781 L closepath p .121 0 0 r
F P
0 g
s
.39237 .28995 m .39956 .29793 L .38963 .30047 L closepath p .2 .522 .13 r
F P
0 g
s
.38963 .30047 m .39956 .29793 L .41018 .30359 L closepath p .202 .547 .136 r
F P
0 g
s
.38963 .30047 m .41018 .30359 L .39922 .30884 L closepath p .192 .39 .084 r
F P
0 g
s
.40179 .39062 m .41939 .44129 L .38983 .43637 L closepath p .112 .362 .166 r
F P
0 g
s
.42272 .47813 m .38983 .43637 L .41939 .44129 L closepath p .105 .437 .193 r
F P
0 g
s
.42272 .47813 m .3933 .47285 L .38983 .43637 L closepath p .105 .437 .193 r
F P
0 g
s
.43031 .318 m .39206 .32947 L .4152 .31418 L closepath p .16 .363 .12 r
F P
0 g
s
.43031 .318 m .40717 .33334 L .39206 .32947 L closepath p .16 .363 .12 r
F P
0 g
s
.39636 .34491 m .39206 .32947 L .40717 .33334 L closepath p .16 .361 .12 r
F P
0 g
s
.40223 .29781 m .39956 .29793 L .39237 .28995 L closepath p .178 .208 0 r
F P
0 g
s
.3933 .47285 m .41 .50465 L .39463 .49712 L closepath p .025 .357 .206 r
F P
0 g
s
.42272 .47813 m .41 .50465 L .3933 .47285 L closepath p .087 .521 .222 r
F P
0 g
s
.39902 .51285 m .39427 .5011 L .39463 .49712 L closepath p 0 0 .12 r
F P
0 g
s
.39902 .51285 m .39463 .49712 L .41 .50465 L closepath p .021 .358 .207 r
F P
0 g
s
.39519 .28744 m .41278 .30342 L p .39565 .28859 L .176 .173 0 r
F P
0 g
s
.41278 .30342 m .40223 .29781 L .39565 .28859 L p .176 .173 0 r
F P
0 g
s
.39565 .28859 m .40223 .29781 L .39787 .29413 L p .177 .173 0 r
F P
0 g
s
.40717 .33334 m .41989 .34336 L .39636 .34491 L closepath p .133 .218 .095 r
F P
0 g
s
.39636 .34491 m .41989 .34336 L .40195 .35537 L closepath p .144 .365 .14 r
F P
0 g
s
.39902 .51285 m .41 .50465 L .41724 .52471 L closepath p .01 .277 .193 r
F P
0 g
s
.41724 .52471 m .40111 .51955 L .39902 .51285 L closepath p 0 .273 .194 r
F P
0 g
s
.4152 .31418 m .39922 .30884 L .41018 .30359 L closepath p .167 .232 .051 r
F P
0 g
s
.39956 .29793 m .40223 .29781 L .41278 .30342 L closepath p .123 0 0 r
F P
0 g
s
.39956 .29793 m .41278 .30342 L .41018 .30359 L closepath p .123 0 0 r
F P
0 g
s
.41724 .52471 m .41928 .53144 L .40111 .51955 L closepath p 0 .273 .194 r
F P
0 g
s
.40729 .3638 m .41225 .40563 L .40179 .39062 L closepath p .116 .341 .157 r
F P
0 g
s
.40179 .39062 m .41225 .40563 L .41939 .44129 L closepath p .086 .301 .166 r
F P
0 g
s
.40195 .35537 m .41989 .34336 L .42529 .35176 L closepath p .144 .37 .141 r
F P
0 g
s
.40195 .35537 m .42529 .35176 L .40729 .3638 L closepath p .144 .37 .141 r
F P
0 g
s
.43031 .318 m .43068 .33174 L .40717 .33334 L closepath p .128 .172 .081 r
F P
0 g
s
.40717 .33334 m .43068 .33174 L .41989 .34336 L closepath p .133 .218 .095 r
F P
0 g
s
.40729 .3638 m .42529 .35176 L .41777 .37868 L closepath p .145 .374 .141 r
F P
0 g
s
.40729 .3638 m .41777 .37868 L .41225 .40563 L closepath p .116 .341 .157 r
F P
0 g
s
.42272 .47813 m .42419 .51062 L .41 .50465 L closepath p .042 .3 .189 r
F P
0 g
s
.4314 .53081 m .41 .50465 L .42419 .51062 L closepath p .026 .327 .2 r
F P
0 g
s
.4314 .53081 m .41724 .52471 L .41 .50465 L closepath p .026 .327 .2 r
F P
0 g
s
.41278 .30342 m .42608 .3089 L .41018 .30359 L closepath p .08 0 0 r
F P
0 g
s
.4152 .31418 m .41018 .30359 L .42608 .3089 L closepath p .167 .232 .051 r
F P
0 g
s
.433 .40548 m .41225 .40563 L .41777 .37868 L closepath p .106 .304 .153 r
F P
0 g
s
.433 .40548 m .41939 .44129 L .41225 .40563 L closepath p .106 .356 .169 r
F P
0 g
s
.44213 .30987 m .4152 .31418 L .42608 .3089 L closepath p .136 .064 .016 r
F P
0 g
s
.44213 .30987 m .43031 .318 L .4152 .31418 L closepath p .151 .158 .041 r
F P
0 g
s
.4314 .53081 m .41928 .53144 L .41724 .52471 L closepath p .015 .338 .204 r
F P
0 g
s
.41777 .37868 m .42529 .35176 L .4395 .34758 L closepath p .102 .208 .124 r
F P
0 g
s
.41777 .37868 m .4395 .34758 L .43201 .37462 L closepath p .102 .208 .124 r
F P
0 g
s
.433 .40548 m .41777 .37868 L .43201 .37462 L closepath p .112 .314 .151 r
F P
0 g
s
.4314 .53081 m .44027 .54127 L .41928 .53144 L closepath p .015 .384 .212 r
F P
0 g
s
.433 .40548 m .44025 .44143 L .41939 .44129 L closepath p .106 .356 .169 r
F P
0 g
s
.43846 .48563 m .41939 .44129 L .44025 .44143 L closepath p .105 .434 .192 r
F P
0 g
s
.43846 .48563 m .42272 .47813 L .41939 .44129 L closepath p .101 .426 .192 r
F P
0 g
s
.44709 .32817 m .42529 .35176 L .41989 .34336 L closepath p .131 .354 .148 r
F P
0 g
s
.44709 .32817 m .41989 .34336 L .43068 .33174 L closepath p .049 0 .024 r
F P
0 g
s
.43846 .48563 m .42419 .51062 L .42272 .47813 L closepath p .083 .444 .206 r
F P
0 g
s
.43846 .48563 m .43975 .5182 L .42419 .51062 L closepath p .083 .444 .206 r
F P
0 g
s
.4314 .53081 m .42419 .51062 L .43975 .5182 L closepath p .067 .45 .214 r
F P
0 g
s
.44709 .32817 m .4395 .34758 L .42529 .35176 L closepath p .099 .181 .116 r
F P
0 g
s
.44213 .30987 m .44222 .31593 L .43031 .318 L closepath p .096 0 0 r
F P
0 g
s
.43031 .318 m .44222 .31593 L .43068 .33174 L closepath p .144 .219 .082 r
F P
0 g
s
.46171 .30779 m .45028 .3236 L .43068 .33174 L closepath p .068 0 .03 r
F P
0 g
s
.46171 .30779 m .43068 .33174 L .44222 .31593 L closepath p .068 0 .03 r
F P
0 g
s
.44709 .32817 m .43068 .33174 L .45028 .3236 L closepath p .027 0 0 r
F P
0 g
s
.4314 .53081 m .43975 .5182 L .44739 .53419 L closepath p .065 .44 .212 r
F P
0 g
s
.44739 .53419 m .44027 .54127 L .4314 .53081 L closepath p .053 .475 .222 r
F P
0 g
s
.45385 .36731 m .43201 .37462 L .4395 .34758 L closepath p .082 .132 .114 r
F P
0 g
s
.45494 .39843 m .43201 .37462 L .45385 .36731 L closepath p .095 .261 .147 r
F P
0 g
s
.45494 .39843 m .433 .40548 L .43201 .37462 L closepath p .095 .261 .147 r
F P
0 g
s
.433 .40548 m .45494 .39843 L .46159 .42813 L closepath p .104 .365 .173 r
F P
0 g
s
.433 .40548 m .46159 .42813 L .44025 .44143 L closepath p .11 .365 .169 r
F P
0 g
s
.43846 .48563 m .44025 .44143 L .45059 .47343 L closepath p .101 .42 .191 r
F P
0 g
s
.45059 .47343 m .45291 .51606 L .43846 .48563 L closepath p .121 .55 .212 r
F P
0 g
s
.43846 .48563 m .45291 .51606 L .43975 .5182 L closepath p .097 .494 .211 r
F P
0 g
s
.44709 .32817 m .45594 .33638 L .4395 .34758 L closepath p .062 .034 .095 r
F P
0 g
s
.45385 .36731 m .4395 .34758 L .45594 .33638 L closepath p .076 .127 .117 r
F P
0 g
s
.46055 .53216 m .43975 .5182 L .45291 .51606 L closepath p .098 .523 .218 r
F P
0 g
s
.46055 .53216 m .44739 .53419 L .43975 .5182 L closepath p .098 .523 .218 r
F P
0 g
s
.47198 .46042 m .44025 .44143 L .46159 .42813 L closepath p .124 .479 .192 r
F P
0 g
s
.47198 .46042 m .45059 .47343 L .44025 .44143 L closepath p .124 .479 .192 r
F P
0 g
s
.44739 .53419 m .45629 .54472 L .44027 .54127 L closepath p .053 .475 .222 r
F P
0 g
s
.44213 .30987 m .45395 .30781 L .44222 .31593 L closepath p .096 0 0 r
F P
0 g
s
.45395 .30781 m .44213 .30987 L .45051 .30808 L p 0 0 0 r
F P
0 g
s
.44222 .31593 m .47309 .30125 L .46171 .30779 L closepath p 0 0 0 r
F P
0 g
s
.44222 .31593 m .45395 .30781 L .47309 .30125 L closepath p .003 0 0 r
F P
0 g
s
.46353 .31688 m .44709 .32817 L .47088 .30729 L closepath p 0 0 0 r
F P
0 g
s
.47088 .30729 m .44709 .32817 L .45028 .3236 L closepath p 0 0 0 r
F P
0 g
s
.44709 .32817 m .46353 .31688 L .45594 .33638 L closepath p .062 .034 .095 r
F P
0 g
s
.46055 .53216 m .45629 .54472 L .44739 .53419 L closepath p .097 .567 .227 r
F P
0 g
s
.45028 .3236 m .46171 .30779 L .4746 .30224 L p .032 0 .006 r
F P
0 g
s
.47088 .30729 m .45028 .3236 L .47397 .30279 L closepath p 0 0 0 r
F P
0 g
s
.45051 .30808 m .45753 .30658 L p .45395 .30781 L 0 0 0 r
F P
0 g
s
.47198 .46042 m .46508 .50415 L .45059 .47343 L closepath p .135 .583 .211 r
F P
0 g
s
.45059 .47343 m .46508 .50415 L .45291 .51606 L closepath p .121 .55 .212 r
F P
0 g
s
.46055 .53216 m .45291 .51606 L .46508 .50415 L closepath p .128 .597 .219 r
F P
0 g
s
.4748 .32695 m .47283 .35802 L .45385 .36731 L closepath p .089 .166 .12 r
F P
0 g
s
.4748 .32695 m .45385 .36731 L .45594 .33638 L closepath p .089 .166 .12 r
F P
0 g
s
.47283 .35802 m .45494 .39843 L .45385 .36731 L closepath p .104 .29 .15 r
F P
0 g
s
.47283 .35802 m .47587 .38518 L .45494 .39843 L closepath p .099 .267 .146 r
F P
0 g
s
.46159 .42813 m .45494 .39843 L .47587 .38518 L closepath p .117 .401 .174 r
F P
0 g
s
.45594 .33638 m .46353 .31688 L .47454 .31127 L closepath p .077 .092 .104 r
F P
0 g
s
.45594 .33638 m .47454 .31127 L .4748 .32695 L closepath p .082 .117 .108 r
F P
0 g
s
.46055 .53216 m .47181 .54421 L .45629 .54472 L closepath p .096 .563 .226 r
F P
0 g
s
.45629 .54472 m .47181 .54421 L .48296 .54631 L closepath p .043 .68 .226 r
F P
0 g
s
.45629 .54472 m .48296 .54631 L .46742 .5468 L closepath p .043 .68 .226 r
F P
0 g
s
.46055 .53216 m .46508 .50415 L .47843 .5256 L closepath p .119 .563 .216 r
F P
0 g
s
.47843 .5256 m .47181 .54421 L .46055 .53216 L closepath p .122 .606 .224 r
F P
0 g
s
.46159 .42813 m .47587 .38518 L .4826 .41507 L closepath p .117 .401 .174 r
F P
0 g
s
.47198 .46042 m .46159 .42813 L .4826 .41507 L closepath p .128 .489 .192 r
F P
0 g
s
.46171 .30779 m .47309 .30125 L .47889 .29995 L p 0 0 0 r
F P
0 g
s
.47806 .30043 m .46171 .30779 L .47889 .29995 L 0 0 0 r
F
.47589 .30168 m .46171 .30779 L p .47806 .30043 L F P
0 g
s
.46353 .31688 m .47088 .30729 L .48188 .30166 L closepath p .026 0 .016 r
F P
0 g
s
.46353 .31688 m .48188 .30166 L .47454 .31127 L closepath p .026 0 .016 r
F P
0 g
s
.47198 .46042 m .4848 .49199 L .46508 .50415 L closepath p .121 .526 .206 r
F P
0 g
s
.4982 .51368 m .46508 .50415 L .4848 .49199 L closepath p .126 .577 .215 r
F P
0 g
s
.4982 .51368 m .47843 .5256 L .46508 .50415 L closepath p .126 .577 .215 r
F P
0 g
s
.46812 .30295 m .47104 .28967 L p .46874 .30274 L .013 0 .08 r
F P
0 g
s
.4693 .30255 m .46874 .30274 L .47104 .28967 L .013 0 .08 r
F
.46933 .30254 m .4693 .30255 L .47104 .28967 L F
.46963 .30243 m .46933 .30254 L .47104 .28967 L F
.47015 .30225 m .46963 .30243 L .47104 .28967 L F
.47058 .30211 m .47015 .30225 L .47104 .28967 L F
.47069 .30207 m .47058 .30211 L .47104 .28967 L F
.47104 .28967 m .47696 .28209 L .47069 .30207 L p F P
0 g
s
.47069 .30207 m .47696 .28209 L p .47227 .30153 L .013 0 .08 r
F P
0 g
s
.47397 .30279 m .48188 .30166 L .47088 .30729 L closepath p 0 0 0 r
F P
0 g
s
.48247 .25447 m .47696 .28209 L .47104 .28967 L closepath p .017 0 .085 r
F P
0 g
s
.48247 .25447 m .47104 .28967 L .47894 .25878 L closepath p 0 0 .066 r
F P
0 g
s
.47843 .5256 m .48974 .53776 L .47181 .54421 L closepath p .122 .606 .224 r
F P
0 g
s
.48974 .53776 m .48296 .54631 L .47181 .54421 L closepath p .12 .823 .25 r
F P
0 g
s
.47198 .46042 m .4826 .41507 L .49351 .44303 L closepath p .12 .459 .189 r
F P
0 g
s
.49351 .44303 m .4848 .49199 L .47198 .46042 L closepath p .134 .558 .206 r
F P
0 g
s
.47696 .28209 m .47273 .30137 L p .47227 .30153 L .013 0 .08 r
F P
0 g
s
.47273 .30137 m .47696 .28209 L p .47309 .30125 L .121 .271 .128 r
F P
0 g
s
.49147 .32154 m .47283 .35802 L .4748 .32695 L closepath p .095 .185 .122 r
F P
0 g
s
.49147 .32154 m .49343 .34458 L .47283 .35802 L closepath p .103 .22 .127 r
F P
0 g
s
.47283 .35802 m .49343 .34458 L .49658 .37181 L closepath p .117 .319 .148 r
F P
0 g
s
.47283 .35802 m .49658 .37181 L .47587 .38518 L closepath p .117 .319 .148 r
F P
0 g
s
.47309 .30125 m .47696 .28209 L .47694 .29942 L .121 .271 .128 r
F
.47309 .30125 m .47694 .29942 L .47363 .30102 L F
.47889 .29995 m .47309 .30125 L .48708 .29521 L p 0 0 0 r
F P
0 g
s
.47363 .30102 m .47694 .29942 L .47694 .29959 L .121 .271 .128 r
F
.47397 .30279 m .49161 .2926 L .49048 .29626 L closepath p 0 0 0 r
F P
0 g
s
.47397 .30279 m .49048 .29626 L .48188 .30166 L closepath p 0 0 0 r
F P
0 g
s
.47436 .30257 m .47397 .30279 L .4746 .30224 L .032 0 .006 r
F
.47585 .3017 m .47436 .30257 L .4746 .30224 L F
.48188 .30166 m .49111 .30579 L .47454 .31127 L closepath p .058 0 .039 r
F P
0 g
s
.49147 .32154 m .47454 .31127 L .49111 .30579 L closepath p .09 .139 .11 r
F P
0 g
s
.49147 .32154 m .4748 .32695 L .47454 .31127 L closepath p .09 .139 .11 r
F P
0 g
s
.4746 .30224 m .47589 .30168 L p .47585 .3017 L .032 0 .006 r
F P
0 g
s
.4826 .41507 m .47587 .38518 L .49658 .37181 L closepath p .135 .448 .175 r
F P
0 g
s
.47694 .29959 m .47696 .28209 L .47695 .29958 L .121 .271 .128 r
F
.47695 .29958 m .47696 .28209 L .47719 .29948 L F
.48247 .25447 m .48714 .27824 L .47696 .28209 L closepath p .122 .278 .13 r
F P
0 g
s
.47696 .28209 m .48714 .27824 L .47785 .2992 L p .121 .271 .128 r
F P
0 g
s
.47741 .29939 m .47696 .28209 L .47785 .2992 L .121 .271 .128 r
F
.47719 .29948 m .47696 .28209 L .47723 .29946 L F
.47723 .29946 m .47696 .28209 L .47741 .29939 L F
.47785 .2992 m .48714 .27824 L p .47833 .29899 L .104 .2 .118 r
F P
0 g
s
.47833 .29899 m .48714 .27824 L .47837 .29897 L .104 .2 .118 r
F
.47837 .29897 m .48714 .27824 L .47929 .29857 L F
.4982 .51368 m .48974 .53776 L .47843 .5256 L closepath p .13 .619 .222 r
F P
0 g
s
.47894 .25878 m .48747 .23727 L .48763 .23708 L p 0 0 .027 r
F P
0 g
s
.48761 .23779 m .47894 .25878 L .48763 .23708 L 0 0 .027 r
F
.48761 .23785 m .47894 .25878 L .48761 .23779 L F
.48761 .2382 m .47894 .25878 L .48761 .23785 L F
.4876 .23861 m .47894 .25878 L .48761 .2382 L F
.48759 .23922 m .47894 .25878 L .4876 .23861 L F
.48758 .23969 m .47894 .25878 L .48759 .23922 L F
.48757 .24029 m .47894 .25878 L p .48758 .23969 L F P
0 g
s
.48247 .25447 m .47894 .25878 L .48736 .24073 L p 0 0 .027 r
F P
0 g
s
.47929 .29857 m .48714 .27824 L .47947 .2985 L .104 .2 .118 r
F
.47947 .2985 m .48714 .27824 L .48002 .29826 L F
.48002 .29826 m .48714 .27824 L .48027 .29815 L F
.48027 .29815 m .48714 .27824 L .48408 .29651 L F
.49111 .30579 m .48188 .30166 L .49048 .29626 L closepath p .052 0 .057 r
F P
0 g
s
.48755 .24118 m .48247 .25447 L .48736 .24073 L 0 0 .027 r
F
.48763 .24143 m .48247 .25447 L p .48755 .24118 L F P
0 g
s
.48838 .24371 m .48247 .25447 L .48763 .24143 L p .027 0 .073 r
F P
0 g
s
.48247 .25447 m .48838 .24371 L p .49 .2486 L .131 .353 .148 r
F P
0 g
s
.48247 .25447 m .49 .2486 L .49007 .24882 L .131 .353 .148 r
F
.48247 .25447 m .49004 .24884 L .49014 .24902 L F
.49113 .25114 m .48247 .25447 L p .49014 .24902 L F P
0 g
s
.48714 .27824 m .48247 .25447 L .49113 .25114 L p .122 .278 .13 r
F P
0 g
s
.4826 .41507 m .49658 .37181 L .5023 .39613 L closepath p .105 .322 .16 r
F P
0 g
s
.51328 .4243 m .4826 .41507 L .5023 .39613 L closepath p .13 .481 .188 r
F P
0 g
s
.51328 .4243 m .49351 .44303 L .4826 .41507 L closepath p .13 .481 .188 r
F P
0 g
s
.48974 .53776 m .50299 .54028 L .48296 .54631 L closepath p .12 .823 .25 r
F P
0 g
s
.48714 .27824 m .48625 .29503 L p .48408 .29651 L .104 .2 .118 r
F P
0 g
s
.48408 .29651 m .48625 .29503 L .48469 .29625 L .104 .2 .118 r
F
.48452 .16155 m .48516 .13715 L .48768 .14233 L closepath p .029 .126 .148 r
F P
0 g
s
.48452 .16155 m .48768 .14233 L .48702 .16677 L closepath p .029 .126 .148 r
F P
0 g
s
.48952 .17359 m .48452 .16155 L .48702 .16677 L closepath p 0 0 .131 r
F P
0 g
s
.48469 .29625 m .48625 .29503 L .4848 .2962 L .104 .2 .118 r
F
.49351 .44303 m .50634 .47496 L .4848 .49199 L closepath p .134 .558 .206 r
F P
0 g
s
.4982 .51368 m .4848 .49199 L .50634 .47496 L closepath p .14 .603 .213 r
F P
0 g
s
.4848 .2962 m .48625 .29503 L .48495 .29613 L .104 .2 .118 r
F
.48495 .29613 m .48625 .29503 L .48508 .29608 L F
.48508 .29608 m .48625 .29503 L .48521 .29602 L F
.48708 .1248 m .48516 .13715 L p .48712 .12427 L 0 0 0 r
F P
0 g
s
.48516 .13715 m .48588 .1316 L .48712 .12427 L p 0 0 0 r
F P
0 g
s
.48696 .12655 m .48516 .13715 L .48708 .1248 L p 0 0 0 r
F P
0 g
s
.48696 .12655 m .48768 .14233 L .48516 .13715 L closepath p .057 .178 .148 r
F P
0 g
s
.48625 .29503 m .48622 .29559 L p .48521 .29602 L .104 .2 .118 r
F P
0 g
s
.48545 .19233 m .48628 .18186 L .49445 .17799 L closepath p .05 .075 .119 r
F P
0 g
s
.49502 .19069 m .48545 .19233 L .49445 .17799 L closepath p .077 .207 .143 r
F P
0 g
s
.49502 .19069 m .486 .205 L .48545 .19233 L closepath p .077 .207 .143 r
F P
0 g
s
.49502 .19069 m .48786 .22523 L .486 .205 L closepath p .103 .33 .164 r
F P
0 g
s
.48622 .29559 m .48714 .27824 L p .48658 .29543 L .161 .385 .128 r
F P
0 g
s
.48952 .17359 m .49445 .17799 L .48628 .18186 L closepath p .052 .094 .124 r
F P
0 g
s
.48708 .29521 m .48658 .29543 L .48714 .27824 L .161 .385 .128 r
F
.48696 .12655 m .48751 .11864 L .493 .12282 L closepath p .105 .231 .129 r
F P
0 g
s
.49596 .1292 m .48696 .12655 L .493 .12282 L closepath p .114 .295 .143 r
F P
0 g
s
.48696 .12655 m .48991 .13292 L .48768 .14233 L closepath p .062 .195 .15 r
F P
0 g
s
.49596 .1292 m .48991 .13292 L .48696 .12655 L closepath p .114 .295 .143 r
F P
0 g
s
.48702 .16677 m .48768 .14233 L .49595 .15232 L closepath p .121 .44 .183 r
F P
0 g
s
.49847 .15916 m .48702 .16677 L .49595 .15232 L closepath p .102 .342 .168 r
F P
0 g
s
.49847 .15916 m .48952 .17359 L .48702 .16677 L closepath p .102 .342 .168 r
F P
0 g
s
.48764 .29489 m .48708 .29521 L .48714 .27824 L .161 .385 .128 r
F
.49113 .25114 m .49122 .25133 L .48714 .27824 L .122 .278 .13 r
F
.49209 .25317 m .48714 .27824 L p .49122 .25133 L F P
0 g
s
.49757 .26081 m .48714 .27824 L .49713 .25994 L .157 .408 .141 r
F
.48714 .27824 m .4895 .2663 L p .49705 .25978 L F P
0 g
s
.49713 .25994 m .48714 .27824 L .49705 .25978 L .157 .408 .141 r
F
.49828 .2622 m .48714 .27824 L p .49792 .26149 L F P
0 g
s
.49792 .26149 m .48714 .27824 L .49757 .26081 L .157 .408 .141 r
F
.50171 .27481 m .48714 .27824 L .498 .26261 L p .169 .466 .149 r
F P
0 g
s
.48714 .27824 m .48926 .27774 L p .48855 .29436 L .161 .385 .128 r
F P
0 g
s
.48779 .29481 m .48775 .29483 L .48714 .27824 L .161 .385 .128 r
F
.48781 .29479 m .48779 .29481 L .48714 .27824 L F
.48775 .29483 m .48775 .29483 L .48714 .27824 L F
.48775 .29483 m .48774 .29483 L .48714 .27824 L F
.48774 .29483 m .48774 .29484 L .48714 .27824 L F
.48774 .29484 m .48774 .29484 L .48714 .27824 L F
.48774 .29484 m .48774 .29484 L .48714 .27824 L F
.48774 .29484 m .48773 .29484 L .48714 .27824 L F
.48773 .29484 m .48773 .29484 L .48714 .27824 L F
.48773 .29484 m .48764 .29489 L .48714 .27824 L F
.48855 .29436 m .4885 .2944 L .48714 .27824 L F
.4885 .2944 m .48822 .29456 L .48714 .27824 L F
.48822 .29456 m .48781 .29479 L .48714 .27824 L F
.48736 .24073 m .48757 .24029 L p .48757 .24032 L 0 0 .027 r
F P
0 g
s
.48757 .24032 m .48755 .24114 L .48736 .24073 L 0 0 .027 r
F
.48755 .24114 m .48755 .24118 L .48736 .24073 L F
.49067 .11549 m .493 .12282 L .48751 .11864 L closepath p .09 .242 .145 r
F P
0 g
s
.49018 .23285 m .48755 .24118 L .48786 .22523 L closepath p 0 0 .092 r
F P
0 g
s
.49014 .24902 m .48755 .24118 L .48834 .23868 L p .028 .148 .155 r
F P
0 g
s
.49595 .15232 m .48768 .14233 L .48991 .13292 L closepath p .126 .445 .181 r
F P
0 g
s
.49502 .19069 m .49049 .21685 L .48786 .22523 L closepath p 0 0 .027 r
F P
0 g
s
.49018 .23285 m .48786 .22523 L .49049 .21685 L closepath p 0 0 .092 r
F P
0 g
s
.48806 .17732 m .48838 .17085 L p .48847 .17626 L 0 0 .036 r
F P
0 g
s
.48834 .23868 m .49018 .23285 L p .48913 .24142 L .028 .148 .155 r
F P
0 g
s
.48913 .24142 m .489 .2425 L .48834 .23868 L .028 .148 .155 r
F
.48847 .17626 m .48838 .17085 L .4884 .1709 L 0 0 .036 r
F
.48847 .17626 m .4884 .1709 L .48854 .17122 L F
.48936 .174 m .48847 .17626 L .48854 .17122 L F
.48936 .174 m .48854 .17122 L .48878 .1718 L F
.49029 .29336 m .48855 .29436 L .48926 .27774 L .161 .385 .128 r
F
.48936 .174 m .48878 .1718 L .48925 .17294 L 0 0 .036 r
F
.489 .2425 m .49018 .23285 L .4921 .22843 L closepath p 0 0 .077 r
F P
0 g
s
.489 .2425 m .4921 .22843 L .49092 .2381 L closepath p 0 0 .077 r
F P
0 g
s
.489 .2425 m .49092 .2381 L .49225 .24133 L closepath p .038 .157 .152 r
F P
0 g
s
.489 .2425 m .49225 .24133 L .49137 .24751 L closepath p .038 .151 .151 r
F P
0 g
s
.49014 .24902 m .489 .2425 L .49137 .24751 L closepath p 0 0 .12 r
F P
0 g
s
.48944 .1734 m .48936 .174 L .48925 .17294 L 0 0 .036 r
F
.48926 .27774 m .50171 .27481 L .49386 .29147 L p .161 .385 .128 r
F P
0 g
s
.49099 .29296 m .49029 .29336 L .48926 .27774 L .161 .385 .128 r
F
.49386 .29147 m .49379 .2915 L .48926 .27774 L F
.49379 .2915 m .49346 .29167 L .48926 .27774 L F
.49346 .29167 m .49098 .29289 L .48926 .27774 L F
.48937 .17398 m .48936 .174 L .48944 .1734 L 0 0 .036 r
F
.48937 .17398 m .48944 .1734 L .48944 .1734 L F
.48946 .17374 m .48937 .17398 L .48944 .1734 L F
.48937 .17398 m .48944 .1734 L .48944 .1734 L F
.48946 .17374 m .48944 .1734 L .48945 .17341 L F
.48946 .17374 m .48945 .17341 L .48947 .17346 L F
.48952 .17359 m .48946 .17374 L .48947 .17346 L F
.4895 .2663 m .49179 .25469 L p .49705 .25978 L .157 .408 .141 r
F P
0 g
s
.49772 .16967 m .48952 .17359 L .49847 .15916 L closepath p .054 .112 .128 r
F P
0 g
s
.48952 .17359 m .49772 .16967 L .49445 .17799 L closepath p .052 .094 .124 r
F P
0 g
s
.4982 .51368 m .5132 .52588 L .48974 .53776 L closepath p .141 .664 .226 r
F P
0 g
s
.48974 .53776 m .5132 .52588 L .52649 .52846 L closepath p .148 .845 .248 r
F P
0 g
s
.48974 .53776 m .52649 .52846 L .50299 .54028 L closepath p .148 .845 .248 r
F P
0 g
s
.4982 .14291 m .48991 .13292 L .49596 .1292 L closepath p .136 .456 .177 r
F P
0 g
s
.49595 .15232 m .48991 .13292 L .4982 .14291 L closepath p .126 .445 .181 r
F P
0 g
s
.49004 .24884 m .49007 .24882 L .49013 .24901 L .131 .353 .148 r
F
.49004 .24884 m .49013 .24901 L .49014 .24902 L F
.49014 .24902 m .49137 .24751 L .4925 .25403 L closepath p 0 0 .12 r
F P
0 g
s
.49018 .23285 m .494 .21646 L .4921 .22843 L closepath p 0 0 .069 r
F P
0 g
s
.49018 .23285 m .49049 .21685 L .494 .21646 L closepath p .061 .167 .142 r
F P
0 g
s
.50364 .2918 m .49048 .29626 L p .50377 .29127 L .009 0 0 r
F P
0 g
s
.49048 .29626 m .49161 .2926 L .50377 .29127 L p .009 0 0 r
F P
0 g
s
.49982 .30035 m .49048 .29626 L .50202 .29235 L p .056 0 .045 r
F P
0 g
s
.49111 .30579 m .49048 .29626 L .49982 .30035 L closepath p .052 0 .057 r
F P
0 g
s
.49049 .21685 m .49502 .19069 L .49854 .19024 L closepath p .059 .127 .13 r
F P
0 g
s
.49049 .21685 m .49854 .19024 L .494 .21646 L closepath p .059 .127 .13 r
F P
0 g
s
.49067 .11549 m .49852 .11273 L .49618 .11967 L closepath p .106 .231 .128 r
F P
0 g
s
.49067 .11549 m .49618 .11967 L .493 .12282 L closepath p .09 .242 .145 r
F P
0 g
s
.49613 .23123 m .49092 .2381 L .4921 .22843 L closepath p .14 .411 .159 r
F P
0 g
s
.49092 .2381 m .49613 .23123 L .49747 .23446 L closepath p .127 .343 .148 r
F P
0 g
s
.49092 .2381 m .49747 .23446 L .49225 .24133 L closepath p .127 .343 .148 r
F P
0 g
s
.49161 .2926 m .49098 .29289 L .49346 .29167 L .161 .385 .128 r
F
.49124 .29281 m .49098 .29289 L .49161 .2926 L F
.49124 .29281 m .49099 .29296 L .49098 .29289 L F
.49147 .32154 m .49111 .30579 L .49982 .30035 L closepath p .086 .13 .109 r
F P
0 g
s
.49225 .24133 m .49655 .24754 L .49137 .24751 L closepath p .112 .431 .187 r
F P
0 g
s
.49137 .24751 m .49655 .24754 L .49539 .25653 L closepath p .111 .458 .195 r
F P
0 g
s
.49137 .24751 m .49539 .25653 L .4925 .25403 L closepath p 0 .079 .159 r
F P
0 g
s
.49147 .32154 m .49982 .30035 L .50187 .31414 L closepath p .061 .031 .094 r
F P
0 g
s
.50187 .31414 m .49343 .34458 L .49147 .32154 L closepath p .082 .161 .124 r
F P
0 g
s
.50358 .28853 m .49161 .2926 L .50514 .28579 L p .011 0 0 r
F P
0 g
s
.50377 .29127 m .49161 .2926 L .50454 .2882 L p .009 0 0 r
F P
0 g
s
.49179 .25469 m .49203 .2535 L p .49273 .25437 L .157 .408 .141 r
F P
0 g
s
.49433 .2568 m .49179 .25469 L .49273 .25437 L .157 .408 .141 r
F
.49705 .25978 m .49179 .25469 L .49564 .25802 L F
.49448 .25701 m .49179 .25469 L .49433 .2568 L F
.49203 .2535 m .49209 .25317 L p .49224 .25348 L F P
0 g
s
.49273 .25437 m .49203 .2535 L .49256 .25413 L .157 .408 .141 r
F
.49241 .25384 m .49203 .2535 L .4924 .25382 L F
.49229 .25359 m .49203 .2535 L .49224 .25348 L F
.4924 .25382 m .49203 .2535 L .49229 .25359 L F
.4925 .25403 m .49203 .2535 L .49241 .25384 L F
.49256 .25413 m .49203 .2535 L .4925 .25403 L F
.49613 .23123 m .4921 .22843 L .494 .21646 L closepath p .139 .413 .161 r
F P
0 g
s
.49744 .24135 m .49225 .24133 L .49747 .23446 L closepath p .112 .264 .135 r
F P
0 g
s
.49225 .24133 m .49744 .24135 L .49655 .24754 L closepath p .112 .431 .187 r
F P
0 g
s
.49448 .25701 m .4925 .25403 L .4963 .25831 L p 0 .069 .15 r
F P
0 g
s
.4963 .25831 m .4925 .25403 L .49539 .25653 L p 0 .069 .15 r
F P
0 g
s
.493 .12282 m .49618 .11967 L .50193 .12004 L closepath p .136 .519 .194 r
F P
0 g
s
.493 .12282 m .50193 .12004 L .50217 .13099 L closepath p .129 .423 .172 r
F P
0 g
s
.493 .12282 m .50217 .13099 L .49596 .1292 L closepath p .173 .399 .119 r
F P
0 g
s
.50187 .31414 m .5039 .33723 L .49343 .34458 L closepath p .082 .161 .124 r
F P
0 g
s
.49658 .37181 m .49343 .34458 L .5039 .33723 L closepath p .097 .26 .146 r
F P
0 g
s
.51328 .4243 m .50634 .47496 L .49351 .44303 L closepath p .146 .584 .205 r
F P
0 g
s
.49386 .29147 m .49738 .284 L p .49548 .29065 L .161 .385 .128 r
F P
0 g
s
.49854 .19024 m .49804 .21925 L .494 .21646 L closepath p .138 .414 .162 r
F P
0 g
s
.49613 .23123 m .494 .21646 L .49804 .21925 L closepath p .139 .413 .161 r
F P
0 g
s
.50212 .18305 m .49445 .17799 L .49772 .16967 L closepath p .097 .28 .152 r
F P
0 g
s
.50212 .18305 m .49502 .19069 L .49445 .17799 L closepath p .1 .276 .148 r
F P
0 g
s
.49655 .25881 m .49448 .25701 L p .4963 .25831 L 0 .069 .15 r
F P
0 g
s
.50212 .18305 m .49854 .19024 L .49502 .19069 L closepath p .054 .024 .098 r
F P
0 g
s
.5027 .25405 m .49539 .25653 L .49655 .24754 L closepath p .064 .277 .173 r
F P
0 g
s
.5027 .25405 m .50152 .26305 L .49539 .25653 L closepath p .064 .277 .173 r
F P
0 g
s
.50105 .26765 m .49539 .25653 L .50152 .26305 L closepath p 0 .148 .173 r
F P
0 g
s
.49738 .284 m .49583 .29047 L .49548 .29065 L .161 .385 .128 r
F
.49705 .25978 m .49564 .25802 L .49655 .25881 L .157 .408 .141 r
F
.49738 .284 m .49659 .2901 L .49583 .29047 L .161 .385 .128 r
F
.49595 .15232 m .4982 .14291 L .50441 .1447 L closepath p .158 .576 .193 r
F P
0 g
s
.49595 .15232 m .50441 .1447 L .50614 .15581 L closepath p .143 .472 .176 r
F P
0 g
s
.49595 .15232 m .50614 .15581 L .49847 .15916 L closepath p .148 .445 .163 r
F P
0 g
s
.4982 .14291 m .49596 .1292 L .50217 .13099 L closepath p .165 .537 .176 r
F P
0 g
s
.50603 .23737 m .49613 .23123 L .49804 .21925 L closepath p .122 .352 .156 r
F P
0 g
s
.50603 .23737 m .49747 .23446 L .49613 .23123 L closepath p .127 .343 .148 r
F P
0 g
s
.49618 .11967 m .49852 .11273 L .50427 .11308 L closepath p .139 .366 .144 r
F P
0 g
s
.49618 .11967 m .50427 .11308 L .50193 .12004 L closepath p .139 .366 .144 r
F P
0 g
s
.5027 .25405 m .49655 .24754 L .49744 .24135 L closepath p .074 .286 .17 r
F P
0 g
s
.49658 .37181 m .5039 .33723 L .50608 .35824 L closepath p .08 .196 .138 r
F P
0 g
s
.50608 .35824 m .5023 .39613 L .49658 .37181 L closepath p .105 .322 .16 r
F P
0 g
s
.49738 .284 m .49677 .29001 L .49659 .2901 L .161 .385 .128 r
F
.49738 .284 m .49807 .28935 L .49677 .29001 L F
.49738 .284 m .50171 .27481 L .50107 .28784 L p F P
0 g
s
.50107 .28784 m .50098 .28789 L .49738 .284 L .161 .385 .128 r
F
.50098 .28789 m .49858 .28909 L .49738 .284 L F
.49858 .28909 m .49846 .28915 L .49738 .284 L F
.4983 .28923 m .49817 .2893 L .49738 .284 L F
.49817 .2893 m .49807 .28935 L .49738 .284 L F
.49846 .28915 m .4983 .28923 L .49738 .284 L F
.50603 .23737 m .49744 .24135 L .49747 .23446 L closepath p .13 .321 .138 r
F P
0 g
s
.50603 .23737 m .50602 .24428 L .49744 .24135 L closepath p .13 .321 .138 r
F P
0 g
s
.5027 .25405 m .49744 .24135 L .50602 .24428 L closepath p .12 .386 .168 r
F P
0 g
s
.49772 .16967 m .49847 .15916 L .50614 .15581 L closepath p .144 .413 .156 r
F P
0 g
s
.49772 .16967 m .50614 .15581 L .5054 .16633 L closepath p .144 .413 .156 r
F P
0 g
s
.50212 .18305 m .49772 .16967 L .5054 .16633 L closepath p .14 .379 .148 r
F P
0 g
s
.498 .26261 m .49828 .2622 L p .50043 .26642 L .169 .466 .149 r
F P
0 g
s
.50105 .26765 m .50104 .26787 L .498 .26261 L .169 .466 .149 r
F
.50082 .26719 m .50105 .26765 L .498 .26261 L F
.5005 .26657 m .50082 .26719 L .498 .26261 L F
.50043 .26642 m .5005 .26657 L .498 .26261 L F
.50104 .26787 m .50075 .27165 L .498 .26261 L F
.50212 .18305 m .49804 .21925 L .49854 .19024 L closepath p .155 .471 .164 r
F P
0 g
s
.50212 .18305 m .50162 .21204 L .49804 .21925 L closepath p .155 .471 .164 r
F P
0 g
s
.50603 .23737 m .49804 .21925 L .50162 .21204 L closepath p .135 .378 .152 r
F P
0 g
s
.4982 .51368 m .50634 .47496 L .52318 .49638 L closepath p .137 .592 .212 r
F P
0 g
s
.52318 .49638 m .5132 .52588 L .4982 .51368 L closepath p .146 .669 .224 r
F P
0 g
s
.4982 .14291 m .50217 .13099 L .50441 .1447 L closepath p .165 .537 .176 r
F P
0 g
s
.49982 .30035 m .50202 .29235 L .50311 .29391 L .056 0 .045 r
F
.50519 .30768 m .49982 .30035 L .50311 .29391 L closepath p 0 0 .086 r
F P
0 g
s
.50519 .30768 m .50187 .31414 L .49982 .30035 L closepath p 0 0 .086 r
F P
0 g
s
.5008 .27112 m .50075 .27165 L .50104 .26787 L .169 .466 .149 r
F
.5008 .27112 m .50171 .27481 L .50075 .27165 L F
.50794 .28421 m .5008 .27112 L .50152 .26305 L closepath p .055 .247 .169 r
F P
0 g
s
.50105 .26765 m .50152 .26305 L .5008 .27112 L closepath p 0 0 0 r
F P
0 g
s
.50161 .27373 m .50171 .27481 L .5008 .27112 L .169 .466 .149 r
F
.50794 .28421 m .5056 .28652 L .5008 .27112 L closepath p .032 .2 .168 r
F P
0 g
s
.50107 .28784 m .50134 .28223 L p .50113 .28781 L .201 .524 .129 r
F P
0 g
s
.50132 .28772 m .50113 .28781 L .50134 .28223 L .201 .524 .129 r
F
.50211 .28732 m .50132 .28772 L .50134 .28223 L F
.50134 .28223 m .50171 .27481 L p .50268 .28703 L F P
0 g
s
.50211 .28732 m .50134 .28223 L .50215 .2873 L .201 .524 .129 r
F
.50215 .2873 m .50134 .28223 L .50216 .28729 L F
.50216 .28729 m .50134 .28223 L .5023 .28723 L F
.50231 .28722 m .50134 .28223 L .50234 .2872 L F
.50258 .28708 m .50134 .28223 L .50268 .28703 L F
.50258 .28708 m .50134 .28223 L .50258 .28708 L F
.50234 .2872 m .50134 .28223 L .50258 .28708 L F
.50231 .28722 m .50134 .28223 L .50231 .28722 L F
.5023 .28723 m .50134 .28223 L .50231 .28722 L F
.5027 .25405 m .50867 .27614 L .50152 .26305 L closepath p .056 .248 .169 r
F P
0 g
s
.50794 .28421 m .50152 .26305 L .50867 .27614 L closepath p .055 .247 .169 r
F P
0 g
s
.50162 .27375 m .50171 .27481 L .50161 .27373 L .169 .466 .149 r
F
.50162 .21204 m .50212 .18305 L .50831 .18767 L closepath p .166 .506 .165 r
F P
0 g
s
.50162 .21204 m .50831 .18767 L .51652 .19069 L closepath p .183 .577 .17 r
F P
0 g
s
.50162 .21204 m .51652 .19069 L .50984 .21506 L closepath p .183 .577 .17 r
F P
0 g
s
.50603 .23737 m .50162 .21204 L .50984 .21506 L closepath p .188 .522 .145 r
F P
0 g
s
.50178 .27427 m .50171 .27481 L p .50162 .27375 L .169 .466 .149 r
F P
0 g
s
.50171 .27481 m .50178 .27427 L p .50183 .27444 L .194 .56 .152 r
F P
0 g
s
.50171 .27481 m .50196 .27486 L p .50202 .27505 L .194 .517 .136 r
F P
0 g
s
.5023 .27594 m .50171 .27481 L p .5021 .27528 L .194 .517 .136 r
F P
0 g
s
.50207 .27519 m .50171 .27481 L .50206 .27517 L .194 .517 .136 r
F
.5021 .27528 m .50171 .27481 L .50207 .27519 L F
.50206 .27517 m .50171 .27481 L .50206 .27516 L F
.50206 .27516 m .50171 .27481 L .50202 .27505 L F
.50196 .27486 m .50171 .27481 L p .5019 .27466 L .194 .56 .152 r
F P
0 g
s
.50186 .27454 m .50187 .27456 L .50171 .27481 L .194 .56 .152 r
F
.50187 .27456 m .5019 .27466 L .50171 .27481 L F
.50183 .27444 m .50186 .27454 L .50171 .27481 L F
.50288 .28693 m .50268 .28703 L .50171 .27481 L .201 .524 .129 r
F
.50334 .2867 m .50288 .28693 L .50171 .27481 L F
.50339 .28667 m .50334 .2867 L .50171 .27481 L F
.50342 .28666 m .50339 .28667 L .50171 .27481 L F
.50426 .28624 m .50356 .28659 L .50171 .27481 L F
.50356 .28659 m .50342 .28666 L .50171 .27481 L F
.50459 .28607 m .50426 .28624 L .50171 .27481 L F
.50171 .27481 m .5023 .27594 L p .50474 .286 L F P
0 g
s
.50474 .286 m .50467 .28603 L .50171 .27481 L .201 .524 .129 r
F
.50467 .28603 m .50464 .28605 L .50171 .27481 L F
.50464 .28605 m .50463 .28605 L .50171 .27481 L F
.50463 .28605 m .50462 .28606 L .50171 .27481 L F
.50462 .28606 m .50462 .28606 L .50171 .27481 L F
.50462 .28606 m .50459 .28607 L .50171 .27481 L F
.50519 .30768 m .5039 .33723 L .50187 .31414 L closepath p .025 .003 .112 r
F P
0 g
s
.50193 .12004 m .51004 .11223 L .51026 .1214 L closepath p .166 .511 .166 r
F P
0 g
s
.50193 .12004 m .50427 .11308 L .51004 .11223 L closepath p .149 .408 .149 r
F P
0 g
s
.5105 .13235 m .50217 .13099 L .50193 .12004 L closepath p .165 .538 .176 r
F P
0 g
s
.5105 .13235 m .50193 .12004 L .51026 .1214 L closepath p .165 .538 .176 r
F P
0 g
s
.50202 .29235 m .50364 .2918 L p .50311 .29391 L .056 0 .045 r
F P
0 g
s
.50212 .18305 m .5054 .16633 L .50888 .17362 L closepath p .18 .533 .159 r
F P
0 g
s
.50831 .18767 m .50212 .18305 L .50888 .17362 L closepath p .173 .495 .154 r
F P
0 g
s
.50441 .1447 m .50217 .13099 L .5105 .13235 L closepath p .165 .537 .176 r
F P
0 g
s
.50608 .35824 m .51185 .3826 L .5023 .39613 L closepath p .105 .322 .16 r
F P
0 g
s
.51328 .4243 m .5023 .39613 L .51185 .3826 L closepath p .138 .499 .187 r
F P
0 g
s
.50447 .2829 m .50474 .286 L .5023 .27594 L .201 .524 .129 r
F
.50602 .24428 m .5121 .26461 L .5027 .25405 L closepath p .105 .324 .16 r
F P
0 g
s
.5027 .25405 m .5121 .26461 L .50867 .27614 L closepath p .074 .295 .172 r
F P
0 g
s
.50878 .30207 m .50311 .29391 L .50525 .28538 L p 0 0 .092 r
F P
0 g
s
.50878 .30207 m .50519 .30768 L .50311 .29391 L closepath p 0 0 .086 r
F P
0 g
s
.50514 .28579 m .50502 .28628 L .50358 .28853 L .011 0 0 r
F
.50502 .28628 m .50495 .28656 L .50358 .28853 L F
.50495 .28656 m .50486 .28693 L .50358 .28853 L F
.50486 .28693 m .50481 .28712 L .50358 .28853 L F
.50475 .28737 m .50473 .28744 L .50358 .28853 L F
.50475 .28736 m .50475 .28737 L .50358 .28853 L F
.50481 .28712 m .50481 .28714 L .50358 .28853 L F
.50481 .28714 m .50475 .28736 L .50358 .28853 L F
.50454 .2882 m .50358 .28853 L p .50473 .28744 L F P
0 g
s
.50519 .30768 m .50479 .33315 L .5039 .33723 L closepath p 0 0 .032 r
F P
0 g
s
.50698 .35419 m .5039 .33723 L .50479 .33315 L closepath p 0 0 .079 r
F P
0 g
s
.50698 .35419 m .50608 .35824 L .5039 .33723 L closepath p 0 0 .079 r
F P
0 g
s
.50441 .1447 m .5105 .13235 L .51133 .1463 L closepath p .194 .681 .19 r
F P
0 g
s
.51306 .1574 m .50441 .1447 L .51133 .1463 L closepath p .199 .627 .168 r
F P
0 g
s
.51306 .1574 m .50614 .15581 L .50441 .1447 L closepath p .199 .627 .168 r
F P
0 g
s
.50482 .28596 m .50447 .2829 L .50454 .2831 L .201 .524 .129 r
F
.50482 .28596 m .50474 .286 L .50447 .2829 L F
.50483 .28595 m .50454 .2831 L .50454 .28311 L F
.50483 .28595 m .50482 .28596 L .50454 .2831 L F
.50488 .28593 m .50454 .28311 L .50459 .28327 L F
.50488 .28593 m .50483 .28595 L .50454 .28311 L F
.50497 .28588 m .50459 .28327 L .50468 .28357 L F
.50497 .28588 m .50488 .28593 L .50459 .28327 L F
.50514 .28579 m .50468 .28357 L .50492 .28432 L F
.50514 .28579 m .50497 .28588 L .50468 .28357 L F
.50878 .30207 m .50479 .33315 L .50519 .30768 L closepath p .001 0 .091 r
F P
0 g
s
.50878 .30207 m .5084 .32756 L .50479 .33315 L closepath p .001 0 .091 r
F P
0 g
s
.51586 .3517 m .50479 .33315 L .5084 .32756 L closepath p .06 .19 .15 r
F P
0 g
s
.51586 .3517 m .50698 .35419 L .50479 .33315 L closepath p .098 .249 .141 r
F P
0 g
s
.50525 .28538 m .50514 .28579 L .50492 .28432 L .201 .524 .129 r
F
.50878 .30207 m .50525 .28538 L .5056 .28652 L 0 0 .092 r
F
.51765 .16367 m .5054 .16633 L .50614 .15581 L closepath p .186 .557 .161 r
F P
0 g
s
.52113 .17095 m .5054 .16633 L .51765 .16367 L closepath p .184 .539 .156 r
F P
0 g
s
.52113 .17095 m .50888 .17362 L .5054 .16633 L closepath p .184 .539 .156 r
F P
0 g
s
.51113 .29979 m .5056 .28652 L .50794 .28421 L closepath p 0 0 .12 r
F P
0 g
s
.51113 .29979 m .50878 .30207 L .5056 .28652 L closepath p 0 0 .12 r
F P
0 g
s
.50603 .23737 m .51544 .25481 L .50602 .24428 L closepath p .137 .343 .139 r
F P
0 g
s
.50602 .24428 m .51544 .25481 L .5121 .26461 L closepath p .105 .324 .16 r
F P
0 g
s
.50603 .23737 m .50984 .21506 L .51568 .22826 L closepath p .164 .429 .141 r
F P
0 g
s
.51568 .22826 m .52412 .24319 L .50603 .23737 L closepath p .137 .276 .114 r
F P
0 g
s
.50603 .23737 m .52412 .24319 L .51544 .25481 L closepath p .133 .336 .141 r
F P
0 g
s
.50698 .35419 m .51185 .3826 L .50608 .35824 L closepath p 0 0 .124 r
F P
0 g
s
.51765 .16367 m .50614 .15581 L .51306 .1574 L closepath p .199 .494 .119 r
F P
0 g
s
.51328 .4243 m .53168 .44989 L .50634 .47496 L closepath p .131 .53 .201 r
F P
0 g
s
.54864 .47161 m .50634 .47496 L .53168 .44989 L closepath p .142 .599 .211 r
F P
0 g
s
.54864 .47161 m .52318 .49638 L .50634 .47496 L closepath p .142 .599 .211 r
F P
0 g
s
.51586 .3517 m .51185 .3826 L .50698 .35419 L closepath p .108 .367 .171 r
F P
0 g
s
.50794 .28421 m .50867 .27614 L .51675 .28999 L closepath p .029 .151 .155 r
F P
0 g
s
.50794 .28421 m .51675 .28999 L .51113 .29979 L closepath p .054 .124 .132 r
F P
0 g
s
.50831 .18767 m .50888 .17362 L .51507 .17823 L closepath p .173 .495 .154 r
F P
0 g
s
.51507 .17823 m .51652 .19069 L .50831 .18767 L closepath p .185 .559 .162 r
F P
0 g
s
.5163 .31794 m .5084 .32756 L .50878 .30207 L closepath p .052 .048 .108 r
F P
0 g
s
.52382 .3421 m .5084 .32756 L .5163 .31794 L closepath p .095 .275 .152 r
F P
0 g
s
.52382 .3421 m .51586 .3517 L .5084 .32756 L closepath p .095 .275 .152 r
F P
0 g
s
.52023 .27844 m .50867 .27614 L .5121 .26461 L closepath p .046 .18 .155 r
F P
0 g
s
.52023 .27844 m .51675 .28999 L .50867 .27614 L closepath p .046 .18 .155 r
F P
0 g
s
.5163 .31794 m .50878 .30207 L .51113 .29979 L closepath p 0 0 .114 r
F P
0 g
s
.50888 .17362 m .52113 .17095 L .52503 .17866 L closepath p .188 .596 .171 r
F P
0 g
s
.50888 .17362 m .52503 .17866 L .51507 .17823 L closepath p .192 .428 .103 r
F P
0 g
s
.51652 .19069 m .5289 .20428 L .50984 .21506 L closepath p .18 .564 .169 r
F P
0 g
s
.53479 .21744 m .50984 .21506 L .5289 .20428 L closepath p .163 .427 .141 r
F P
0 g
s
.53479 .21744 m .51568 .22826 L .50984 .21506 L closepath p .163 .427 .141 r
F P
0 g
s
.51472 .10946 m .51004 .11223 L .51392 .10922 L p .068 0 .066 r
F P
0 g
s
.51026 .1214 m .51004 .11223 L .51472 .10946 L closepath p .132 .4 .163 r
F P
0 g
s
.51026 .1214 m .51472 .10946 L .51495 .11863 L closepath p .132 .4 .163 r
F P
0 g
s
.5105 .13235 m .51026 .1214 L .51495 .11863 L closepath p .138 .452 .174 r
F P
0 g
s
.5105 .13235 m .51495 .11863 L .51793 .12122 L closepath p .168 .578 .185 r
F P
0 g
s
.5105 .13235 m .51793 .12122 L .51673 .13 L closepath p .184 .67 .196 r
F P
0 g
s
.51133 .1463 m .5105 .13235 L .51673 .13 L closepath p .183 .647 .191 r
F P
0 g
s
.5163 .31794 m .51113 .29979 L .51675 .28999 L closepath p .039 .056 .12 r
F P
0 g
s
.51133 .1463 m .51673 .13 L .51756 .14395 L closepath p .183 .647 .191 r
F P
0 g
s
.51133 .1463 m .51756 .14395 L .5245 .14697 L closepath p .189 .795 .222 r
F P
0 g
s
.51133 .1463 m .52304 .16047 L .51306 .1574 L closepath p .215 .669 .161 r
F P
0 g
s
.51133 .1463 m .5245 .14697 L .52304 .16047 L closepath p .201 .673 .18 r
F P
0 g
s
.51586 .3517 m .52077 .3802 L .51185 .3826 L closepath p .108 .367 .171 r
F P
0 g
s
.51328 .4243 m .51185 .3826 L .52077 .3802 L closepath p .111 .411 .182 r
F P
0 g
s
.52023 .27844 m .5121 .26461 L .51544 .25481 L closepath p .085 .244 .149 r
F P
0 g
s
.51765 .16367 m .51306 .1574 L .52304 .16047 L closepath p .211 .482 .091 r
F P
0 g
s
.52318 .49638 m .53826 .50875 L .5132 .52588 L closepath p .146 .669 .224 r
F P
0 g
s
.53826 .50875 m .52649 .52846 L .5132 .52588 L closepath p .161 .848 .245 r
F P
0 g
s
.51328 .4243 m .52077 .3802 L .53597 .40632 L closepath p .128 .475 .188 r
F P
0 g
s
.53597 .40632 m .53168 .44989 L .51328 .4243 L closepath p .137 .538 .199 r
F P
0 g
s
.51392 .10922 m .51591 .10768 L p .51472 .10946 L .068 0 .066 r
F P
0 g
s
.5171 .10591 m .51495 .11863 L .51472 .10946 L closepath p .049 .141 .141 r
F P
0 g
s
.5171 .10591 m .51722 .11057 L .51495 .11863 L closepath p .049 .141 .141 r
F P
0 g
s
.5202 .11316 m .51495 .11863 L .51722 .11057 L closepath p .156 .577 .195 r
F P
0 g
s
.5202 .11316 m .51793 .12122 L .51495 .11863 L closepath p .156 .577 .195 r
F P
0 g
s
.5265 .1911 m .51507 .17823 L .52503 .17866 L closepath p .202 .607 .158 r
F P
0 g
s
.5265 .1911 m .51652 .19069 L .51507 .17823 L closepath p .202 .607 .158 r
F P
0 g
s
.5378 .25219 m .51544 .25481 L .52412 .24319 L closepath p .129 .32 .138 r
F P
0 g
s
.5378 .25219 m .52908 .26387 L .51544 .25481 L closepath p .129 .32 .138 r
F P
0 g
s
.52023 .27844 m .51544 .25481 L .52908 .26387 L closepath p .12 .335 .152 r
F P
0 g
s
.53479 .21744 m .5338 .23403 L .51568 .22826 L closepath p .138 .257 .105 r
F P
0 g
s
.51568 .22826 m .5338 .23403 L .52412 .24319 L closepath p .137 .276 .114 r
F P
0 g
s
.54424 .36633 m .51586 .3517 L .52382 .3421 L closepath p .12 .41 .175 r
F P
0 g
s
.54424 .36633 m .52077 .3802 L .51586 .3517 L closepath p .122 .406 .173 r
F P
0 g
s
.5163 .31794 m .51675 .28999 L .52198 .30812 L closepath p .039 .056 .12 r
F P
0 g
s
.5163 .31794 m .52198 .30812 L .53634 .32671 L closepath p .096 .325 .167 r
F P
0 g
s
.5163 .31794 m .53634 .32671 L .52382 .3421 L closepath p .105 .297 .152 r
F P
0 g
s
.51652 .19069 m .5265 .1911 L .5289 .20428 L closepath p .2 .557 .142 r
F P
0 g
s
.51793 .12122 m .52439 .12769 L .51673 .13 L closepath p .193 .7 .196 r
F P
0 g
s
.51756 .14395 m .51673 .13 L .52439 .12769 L closepath p .191 .673 .19 r
F P
0 g
s
.52023 .27844 m .52198 .30812 L .51675 .28999 L closepath p 0 0 .114 r
F P
0 g
s
.52067 .10229 m .52047 .10265 L .5171 .10591 L closepath p 0 0 0 r
F P
0 g
s
.52047 .10265 m .51722 .11057 L .5171 .10591 L closepath p .131 .399 .163 r
F P
0 g
s
.52047 .10265 m .52059 .10732 L .51722 .11057 L closepath p .131 .399 .163 r
F P
0 g
s
.51722 .11057 m .52059 .10732 L .5202 .11316 L closepath p .159 .578 .193 r
F P
0 g
s
.51756 .14395 m .52439 .12769 L .53133 .13074 L closepath p .208 .757 .195 r
F P
0 g
s
.51756 .14395 m .53133 .13074 L .5245 .14697 L closepath p .208 .757 .195 r
F P
0 g
s
.51765 .16367 m .52304 .16047 L .52762 .16673 L closepath p .211 .482 .091 r
F P
0 g
s
.52113 .17095 m .51765 .16367 L .52762 .16673 L closepath p .216 .579 .127 r
F P
0 g
s
.5202 .11316 m .52559 .11891 L .51793 .12122 L closepath p .193 .733 .204 r
F P
0 g
s
.51793 .12122 m .52559 .11891 L .52439 .12769 L closepath p .193 .7 .196 r
F P
0 g
s
.52059 .10732 m .5241 .11199 L .5202 .11316 L closepath p .192 .692 .195 r
F P
0 g
s
.5202 .11316 m .5241 .11199 L .52559 .11891 L closepath p .193 .733 .204 r
F P
0 g
s
.52908 .26387 m .53093 .29219 L .52023 .27844 L closepath p .107 .274 .142 r
F P
0 g
s
.52023 .27844 m .53093 .29219 L .52198 .30812 L closepath p .096 .263 .147 r
F P
0 g
s
.5254 .10244 m .52047 .10265 L .52067 .10229 L closepath p .069 0 0 r
F P
0 g
s
.52449 .10615 m .52047 .10265 L .5254 .10244 L closepath p .171 .577 .182 r
F P
0 g
s
.52449 .10615 m .52059 .10732 L .52047 .10265 L closepath p .185 .569 .166 r
F P
0 g
s
.52059 .10732 m .52449 .10615 L .5241 .11199 L closepath p .192 .692 .195 r
F P
0 g
s
.5254 .10244 m .52067 .10229 L .52559 .10208 L closepath p .069 0 0 r
F P
0 g
s
.55958 .39262 m .52077 .3802 L .54424 .36633 L closepath p .131 .48 .188 r
F P
0 g
s
.55958 .39262 m .53597 .40632 L .52077 .3802 L closepath p .131 .48 .188 r
F P
0 g
s
.52113 .17095 m .52762 .16673 L .53093 .17307 L closepath p .214 .567 .124 r
F P
0 g
s
.52503 .17866 m .52113 .17095 L .53093 .17307 L closepath p .216 .628 .146 r
F P
0 g
s
.52198 .30812 m .53093 .29219 L .54537 .31073 L closepath p .113 .345 .16 r
F P
0 g
s
.52198 .30812 m .54537 .31073 L .53634 .32671 L closepath p .113 .345 .16 r
F P
0 g
s
.53362 .1673 m .52304 .16047 L .5245 .14697 L closepath p .223 .756 .176 r
F P
0 g
s
.53767 .17312 m .52304 .16047 L .53362 .1673 L closepath p .216 .418 .029 r
F P
0 g
s
.53767 .17312 m .52762 .16673 L .52304 .16047 L closepath p .215 .407 .024 r
F P
0 g
s
.54864 .47161 m .53826 .50875 L .52318 .49638 L closepath p .153 .674 .222 r
F P
0 g
s
.54424 .36633 m .52382 .3421 L .53634 .32671 L closepath p .127 .417 .172 r
F P
0 g
s
.5241 .11199 m .52449 .10615 L .52766 .10837 L closepath p .198 .713 .194 r
F P
0 g
s
.52766 .10837 m .52559 .11891 L .5241 .11199 L closepath p .204 .758 .2 r
F P
0 g
s
.5493 .24131 m .52412 .24319 L .5338 .23403 L closepath p .144 .313 .12 r
F P
0 g
s
.5493 .24131 m .5378 .25219 L .52412 .24319 L closepath p .142 .291 .114 r
F P
0 g
s
.53272 .11629 m .52439 .12769 L .52559 .11891 L closepath p .196 .714 .196 r
F P
0 g
s
.53272 .11629 m .53133 .13074 L .52439 .12769 L closepath p .204 .768 .201 r
F P
0 g
s
.52449 .10615 m .5254 .10244 L .52857 .10467 L closepath p .202 .707 .188 r
F P
0 g
s
.52449 .10615 m .52857 .10467 L .52766 .10837 L closepath p .202 .707 .188 r
F P
0 g
s
.54033 .13519 m .5245 .14697 L .53133 .13074 L closepath p .214 .784 .195 r
F P
0 g
s
.54033 .13519 m .53508 .15386 L .5245 .14697 L closepath p .213 .782 .195 r
F P
0 g
s
.53362 .1673 m .5245 .14697 L .53508 .15386 L closepath p .223 .756 .176 r
F P
0 g
s
.52503 .17866 m .53093 .17307 L .53482 .18075 L closepath p .216 .628 .146 r
F P
0 g
s
.5265 .1911 m .52503 .17866 L .53482 .18075 L closepath p .216 .646 .152 r
F P
0 g
s
.52857 .10467 m .5254 .10244 L .52559 .10208 L closepath p .188 .234 0 r
F P
0 g
s
.52766 .10837 m .52916 .1153 L .52559 .11891 L closepath p .204 .758 .2 r
F P
0 g
s
.53272 .11629 m .52559 .11891 L .52916 .1153 L closepath p .195 .689 .191 r
F P
0 g
s
.53826 .50875 m .55509 .51212 L .52649 .52846 L closepath p .161 .848 .245 r
F P
0 g
s
.54084 .52208 m .52649 .52846 L .55509 .51212 L closepath p .112 .871 .188 r
F P
0 g
s
.5265 .1911 m .53482 .18075 L .53784 .19027 L closepath p .208 .599 .146 r
F P
0 g
s
.5289 .20428 m .5265 .1911 L .53784 .19027 L closepath p .207 .574 .139 r
F P
0 g
s
.54097 .17944 m .52762 .16673 L .53767 .17312 L closepath p .23 .556 .08 r
F P
0 g
s
.54097 .17944 m .53093 .17307 L .52762 .16673 L closepath p .23 .556 .08 r
F P
0 g
s
.52766 .10837 m .52857 .10467 L .53095 .10802 L closepath p .224 .802 .187 r
F P
0 g
s
.52766 .10837 m .53095 .10802 L .52916 .1153 L closepath p .224 .802 .187 r
F P
0 g
s
.5289 .20428 m .53784 .19027 L .54027 .20342 L closepath p .207 .574 .139 r
F P
0 g
s
.53479 .21744 m .5289 .20428 L .54027 .20342 L closepath p .2 .489 .115 r
F P
0 g
s
.5378 .25219 m .53985 .27754 L .52908 .26387 L closepath p .123 .288 .133 r
F P
0 g
s
.52908 .26387 m .53985 .27754 L .53093 .29219 L closepath p .107 .274 .142 r
F P
0 g
s
.52916 .1153 m .53095 .10802 L .53451 .10901 L closepath p .196 .681 .187 r
F P
0 g
s
.52916 .1153 m .53451 .10901 L .53272 .11629 L closepath p .196 .681 .187 r
F P
0 g
s
.53067 .10763 m .53111 .10799 L p .53083 .10785 L .129 0 0 r
F P
0 g
s
.53101 .10801 m .53083 .10785 L .53111 .10799 L .129 0 0 r
F
.53095 .10802 m .53083 .10785 L .53101 .10801 L F
.53482 .18075 m .53093 .17307 L .54097 .17944 L closepath p .234 .626 .108 r
F P
0 g
s
.55599 .2903 m .53093 .29219 L .53985 .27754 L closepath p .132 .396 .161 r
F P
0 g
s
.55599 .2903 m .54537 .31073 L .53093 .29219 L closepath p .131 .362 .151 r
F P
0 g
s
.53095 .10802 m .53686 .10708 L .53451 .10901 L closepath p .202 .616 .161 r
F P
0 g
s
.53272 .11629 m .53375 .11988 L .53133 .13074 L closepath p .189 .713 .203 r
F P
0 g
s
.53375 .11988 m .53584 .13297 L .53133 .13074 L closepath p .209 .795 .203 r
F P
0 g
s
.54033 .13519 m .53133 .13074 L .53584 .13297 L closepath p .217 .775 .19 r
F P
0 g
s
.53597 .40632 m .5545 .43213 L .53168 .44989 L closepath p .137 .538 .199 r
F P
0 g
s
.54864 .47161 m .53168 .44989 L .5545 .43213 L closepath p .145 .603 .21 r
F P
0 g
s
.53928 .10766 m .53272 .11629 L .53451 .10901 L closepath p .168 .564 .182 r
F P
0 g
s
.53928 .10766 m .53837 .11136 L .53272 .11629 L closepath p .168 .564 .182 r
F P
0 g
s
.53939 .11495 m .53272 .11629 L .53837 .11136 L closepath p .187 .708 .204 r
F P
0 g
s
.53939 .11495 m .53375 .11988 L .53272 .11629 L closepath p .187 .708 .204 r
F P
0 g
s
.53362 .1673 m .53508 .15386 L .53657 .17155 L p .238 .816 .167 r
F P
0 g
s
.53939 .11495 m .53827 .12213 L .53375 .11988 L closepath p .198 .812 .217 r
F P
0 g
s
.53375 .11988 m .53827 .12213 L .53584 .13297 L closepath p .209 .795 .203 r
F P
0 g
s
.53479 .21744 m .54459 .22685 L .5338 .23403 L closepath p .168 .358 .108 r
F P
0 g
s
.5601 .23408 m .5338 .23403 L .54459 .22685 L closepath p .15 .25 .087 r
F P
0 g
s
.5601 .23408 m .5493 .24131 L .5338 .23403 L closepath p .15 .25 .087 r
F P
0 g
s
.53928 .10766 m .53451 .10901 L .53686 .10708 L closepath p .122 .134 .073 r
F P
0 g
s
.53479 .21744 m .54027 .20342 L .54677 .21334 L closepath p .198 .48 .115 r
F P
0 g
s
.54459 .22685 m .53479 .21744 L .54677 .21334 L closepath p .179 .336 .082 r
F P
0 g
s
.53482 .18075 m .54097 .17944 L .54165 .18497 L closepath p .24 .711 .129 r
F P
0 g
s
.54467 .19446 m .53482 .18075 L .54165 .18497 L closepath p .239 .644 .101 r
F P
0 g
s
.54467 .19446 m .53784 .19027 L .53482 .18075 L closepath p .239 .644 .101 r
F P
0 g
s
.53508 .15386 m .54033 .13519 L .54548 .13908 L closepath p .224 .83 .193 r
F P
0 g
s
.53508 .15386 m .54548 .13908 L .54023 .1577 L closepath p .224 .83 .193 r
F P
0 g
s
.53508 .15386 m .54023 .1577 L .53714 .16239 L closepath p .17 .832 .239 r
F P
0 g
s
.53657 .17155 m .53508 .15386 L p .53708 .17228 L .224 .709 .05 r
F P
0 g
s
.53508 .15386 m .53714 .16239 L .53708 .17228 L p .224 .709 .05 r
F P
0 g
s
.53584 .13297 m .54061 .12019 L .54176 .12058 L closepath p .215 .805 .199 r
F P
0 g
s
.53584 .13297 m .53827 .12213 L .54061 .12019 L closepath p .226 .866 .197 r
F P
0 g
s
.53584 .13297 m .54176 .12058 L .54281 .1242 L closepath p .2 .728 .196 r
F P
0 g
s
.54281 .1242 m .54033 .13519 L .53584 .13297 L closepath p .21 .794 .202 r
F P
0 g
s
.55958 .39262 m .5545 .43213 L .53597 .40632 L closepath p .137 .54 .199 r
F P
0 g
s
.53634 .32671 m .54537 .31073 L .56664 .32891 L closepath p .134 .443 .174 r
F P
0 g
s
.53634 .32671 m .56664 .32891 L .55689 .35094 L closepath p .134 .423 .168 r
F P
0 g
s
.54424 .36633 m .53634 .32671 L .55689 .35094 L closepath p .127 .417 .172 r
F P
0 g
s
.53928 .10766 m .53686 .10708 L .54163 .10573 L closepath p .122 .134 .073 r
F P
0 g
s
.5493 .24131 m .5542 .25834 L .5378 .25219 L closepath p .135 .254 .107 r
F P
0 g
s
.5378 .25219 m .5542 .25834 L .53985 .27754 L closepath p .13 .308 .133 r
F P
0 g
s
.53784 .19027 m .54467 .19446 L .5498 .20617 L closepath p .236 .607 .087 r
F P
0 g
s
.53784 .19027 m .5498 .20617 L .54027 .20342 L closepath p .235 .647 .114 r
F P
0 g
s
.54864 .47161 m .57001 .48803 L .53826 .50875 L closepath p .148 .656 .22 r
F P
0 g
s
.53826 .50875 m .57001 .48803 L .58688 .49153 L closepath p .163 .848 .244 r
F P
0 g
s
.53826 .50875 m .58688 .49153 L .55509 .51212 L closepath p .163 .848 .244 r
F P
0 g
s
.54403 .11417 m .53827 .12213 L .54168 .1161 L closepath p .222 .978 .206 r
F P
0 g
s
.54403 .11417 m .54061 .12019 L .53827 .12213 L closepath p .222 .978 .206 r
F P
0 g
s
.53939 .11495 m .54168 .1161 L .53827 .12213 L closepath p .198 .812 .217 r
F P
0 g
s
.54302 .10665 m .53837 .11136 L .53928 .10766 L closepath p .19 .655 .187 r
F P
0 g
s
.54302 .10665 m .54211 .11035 L .53837 .11136 L closepath p .19 .655 .187 r
F P
0 g
s
.53837 .11136 m .54211 .11035 L .53939 .11495 L closepath p .192 .72 .202 r
F P
0 g
s
.54163 .10573 m .54302 .10665 L .53928 .10766 L closepath p .167 .384 .12 r
F P
0 g
s
.54211 .11035 m .54168 .1161 L .53939 .11495 L closepath p .206 .801 .207 r
F P
0 g
s
.5704 .27099 m .53985 .27754 L .5542 .25834 L closepath p .145 .387 .146 r
F P
0 g
s
.5704 .27099 m .55599 .2903 L .53985 .27754 L closepath p .145 .387 .146 r
F P
0 g
s
.54552 .14633 m .54023 .1577 L .54548 .13908 L closepath p .24 .934 .153 r
F P
0 g
s
.54027 .20342 m .5498 .20617 L .55628 .21604 L closepath p .218 .448 .05 r
F P
0 g
s
.54027 .20342 m .55628 .21604 L .54677 .21334 L closepath p .218 .448 .05 r
F P
0 g
s
.54281 .1242 m .54729 .12644 L .54033 .13519 L closepath p .21 .794 .202 r
F P
0 g
s
.55084 .12569 m .54033 .13519 L .54729 .12644 L closepath p .223 .926 .209 r
F P
0 g
s
.55084 .12569 m .54548 .13908 L .54033 .13519 L closepath p .219 .837 .201 r
F P
0 g
s
.54061 .12019 m .54182 .11806 L .54179 .11925 L .221 .773 .184 r
F
.54176 .12058 m .54061 .12019 L p .54179 .11925 L F P
0 g
s
.54084 .52208 m .55509 .51212 L .56943 .50576 L closepath p .112 .871 .188 r
F P
0 g
s
.54527 .10775 m .54302 .10665 L .54163 .10573 L closepath p .131 0 0 r
F P
0 g
s
.54527 .10775 m .54163 .10573 L .54388 .10683 L closepath p .131 0 0 r
F P
0 g
s
.54211 .11035 m .5444 .1115 L .54168 .1161 L closepath p .206 .801 .207 r
F P
0 g
s
.54168 .1161 m .5444 .1115 L .54403 .11417 L closepath p .226 .947 .205 r
F P
0 g
s
.54834 .1116 m .54747 .11534 L .54176 .12058 L closepath p .166 .545 .178 r
F P
0 g
s
.54834 .1116 m .54176 .12058 L .54271 .1165 L p .166 .545 .178 r
F P
0 g
s
.54852 .11897 m .54176 .12058 L .54747 .11534 L closepath p .187 .699 .201 r
F P
0 g
s
.54271 .1165 m .54176 .12058 L p .54182 .11806 L .221 .773 .184 r
F P
0 g
s
.54852 .11897 m .54281 .1242 L .54176 .12058 L closepath p .187 .699 .201 r
F P
0 g
s
.5444 .1115 m .54211 .11035 L .54302 .10665 L closepath p .217 .772 .188 r
F P
0 g
s
.54834 .1116 m .54271 .1165 L .54332 .11542 L .166 .545 .178 r
F
.54852 .11897 m .54729 .12644 L .54281 .1242 L closepath p .2 .81 .215 r
F P
0 g
s
.5444 .1115 m .54302 .10665 L .54527 .10775 L closepath p .223 .781 .183 r
F P
0 g
s
.54511 .11267 m .54834 .1116 L p .54332 .11542 L .166 .545 .178 r
F P
0 g
s
.54332 .11542 m .54403 .11417 L .54511 .11267 L .166 .545 .178 r
F
.54431 .11294 m .54511 .11267 L p .54403 .11417 L F P
0 g
s
.5444 .1115 m .5449 .11042 L .54403 .11417 L closepath p .211 .819 .085 r
F P
0 g
s
.55958 .39262 m .54424 .36633 L .55689 .35094 L closepath p .141 .498 .185 r
F P
0 g
s
.54578 .11125 m .54834 .1116 L .54431 .11294 L p .109 .065 .059 r
F P
0 g
s
.54444 .11239 m .54578 .11125 L p .54431 .11294 L .109 .065 .059 r
F P
0 g
s
.5444 .1115 m .54527 .10775 L .5449 .11042 L closepath p .211 .819 .085 r
F P
0 g
s
.54461 .11167 m .54578 .11125 L .54444 .11239 L p .056 0 .01 r
F P
0 g
s
.54459 .22685 m .54677 .21334 L .55657 .2227 L closepath p .179 .336 .082 r
F P
0 g
s
.54459 .22685 m .55657 .2227 L .57164 .22868 L closepath p .199 .484 .115 r
F P
0 g
s
.54459 .22685 m .57164 .22868 L .57884 .23161 L closepath p .091 0 0 r
F P
0 g
s
.54459 .22685 m .57884 .23161 L .5601 .23408 L closepath p .15 .111 .017 r
F P
0 g
s
.55599 .2903 m .56664 .32891 L .54537 .31073 L closepath p .148 .44 .161 r
F P
0 g
s
.55084 .12569 m .55023 .13125 L .54548 .13908 L closepath p .226 .872 .199 r
F P
0 g
s
.55023 .13125 m .54552 .14633 L .54548 .13908 L closepath p .232 .951 .195 r
F P
0 g
s
.55023 .13125 m .55025 .13852 L .54552 .14633 L closepath p .232 .951 .195 r
F P
0 g
s
.54834 .1116 m .54578 .11125 L .55064 .10963 L closepath p .109 .065 .059 r
F P
0 g
s
.56496 .22144 m .55657 .2227 L .54677 .21334 L closepath p .186 .238 .002 r
F P
0 g
s
.55941 .21897 m .54677 .21334 L .55628 .21604 L p .149 .021 0 r
F P
0 g
s
.55309 .11814 m .54959 .12447 L .54729 .12644 L closepath p .224 .98 .203 r
F P
0 g
s
.55309 .11814 m .54729 .12644 L .55079 .12011 L closepath p .224 .98 .203 r
F P
0 g
s
.54852 .11897 m .55079 .12011 L .54729 .12644 L closepath p .2 .81 .215 r
F P
0 g
s
.55084 .12569 m .54729 .12644 L .54959 .12447 L closepath p .229 .881 .196 r
F P
0 g
s
.55211 .11045 m .54747 .11534 L .54834 .1116 L closepath p .189 .64 .183 r
F P
0 g
s
.55211 .11045 m .55125 .11419 L .54747 .11534 L closepath p .189 .64 .183 r
F P
0 g
s
.54747 .11534 m .55125 .11419 L .54852 .11897 L closepath p .193 .711 .199 r
F P
0 g
s
.55064 .10963 m .55211 .11045 L .54834 .1116 L closepath p .16 .335 .11 r
F P
0 g
s
.55125 .11419 m .55079 .12011 L .54852 .11897 L closepath p .208 .798 .205 r
F P
0 g
s
.54864 .47161 m .5545 .43213 L .57409 .45171 L closepath p .145 .604 .21 r
F P
0 g
s
.57409 .45171 m .57001 .48803 L .54864 .47161 L closepath p .152 .658 .218 r
F P
0 g
s
.57784 .24185 m .5493 .24131 L .5601 .23408 L closepath p .156 .282 .093 r
F P
0 g
s
.5493 .24131 m .56574 .24738 L .5542 .25834 L closepath p .135 .254 .107 r
F P
0 g
s
.57784 .24185 m .56574 .24738 L .5493 .24131 L closepath p .137 .109 .039 r
F P
0 g
s
.5521 .11993 m .55124 .12388 L p .54959 .12447 L .241 .852 .17 r
F P
0 g
s
.55124 .12388 m .55084 .12569 L .54959 .12447 L p .241 .852 .17 r
F P
0 g
s
.55622 .12739 m .55023 .13125 L .55084 .12569 L closepath p .211 .823 .207 r
F P
0 g
s
.55622 .12739 m .55025 .13852 L .55023 .13125 L closepath p .213 .911 .219 r
F P
0 g
s
.55622 .12739 m .55621 .13109 L .55025 .13852 L closepath p .213 .911 .219 r
F P
0 g
s
.55435 .11153 m .55211 .11045 L .55064 .10963 L closepath p .099 0 0 r
F P
0 g
s
.55435 .11153 m .55064 .10963 L .55288 .11072 L closepath p .099 0 0 r
F P
0 g
s
.55125 .11419 m .55352 .11533 L .55079 .12011 L closepath p .208 .798 .205 r
F P
0 g
s
.55079 .12011 m .55352 .11533 L .55309 .11814 L closepath p .228 .948 .202 r
F P
0 g
s
.55766 .11805 m .55084 .12569 L .5521 .11993 L p .199 .666 .18 r
F P
0 g
s
.55766 .11805 m .55683 .12183 L .55084 .12569 L closepath p .199 .666 .18 r
F P
0 g
s
.55622 .12739 m .55084 .12569 L .55683 .12183 L closepath p .211 .823 .207 r
F P
0 g
s
.55352 .11533 m .55125 .11419 L .55211 .11045 L closepath p .219 .765 .184 r
F P
0 g
s
.55766 .11805 m .5521 .11993 L .55243 .11934 L .199 .666 .18 r
F
.55352 .11533 m .55211 .11045 L .55435 .11153 L closepath p .224 .774 .179 r
F P
0 g
s
.55305 .11822 m .55766 .11805 L p .55243 .11934 L .199 .666 .18 r
F P
0 g
s
.55308 .11822 m .55305 .11822 L p .55309 .11814 L .185 .406 .104 r
F P
0 g
s
.55308 .11822 m .55309 .11814 L .55365 .11645 L .185 .406 .104 r
F
.55365 .11645 m .55766 .11805 L .55308 .11822 L p F P
0 g
s
.55337 .11687 m .55365 .11645 L p .55309 .11814 L .185 .406 .104 r
F P
0 g
s
.55352 .11533 m .55392 .11435 L .55309 .11814 L closepath p .202 .781 .069 r
F P
0 g
s
.55365 .11645 m .55337 .11687 L p .55342 .11665 L .246 .814 .147 r
F P
0 g
s
.55345 .11649 m .55365 .11645 L p .55342 .11665 L .246 .814 .147 r
F P
0 g
s
.55346 .11643 m .55365 .11645 L .55345 .11649 L p .106 0 0 r
F P
0 g
s
.55352 .11533 m .55435 .11153 L .55392 .11435 L closepath p .202 .781 .069 r
F P
0 g
s
.55766 .11805 m .55365 .11645 L .55883 .11626 L closepath p .185 .406 .104 r
F P
0 g
s
.5704 .27099 m .5542 .25834 L .56574 .24738 L closepath p .157 .372 .128 r
F P
0 g
s
.55958 .39262 m .5767 .41854 L .5545 .43213 L closepath p .14 .549 .2 r
F P
0 g
s
.59638 .43828 m .5545 .43213 L .5767 .41854 L closepath p .146 .605 .21 r
F P
0 g
s
.59638 .43828 m .57409 .45171 L .5545 .43213 L closepath p .146 .605 .21 r
F P
0 g
s
.58688 .49153 m .56943 .50576 L .55509 .51212 L closepath p .13 .912 .198 r
F P
0 g
s
.57731 .30838 m .55599 .2903 L .5704 .27099 L closepath p .155 .437 .153 r
F P
0 g
s
.55599 .2903 m .57731 .30838 L .56664 .32891 L closepath p .148 .44 .161 r
F P
0 g
s
.55981 .12208 m .55621 .13109 L .55622 .12739 L closepath p .227 .943 .204 r
F P
0 g
s
.55981 .12208 m .5598 .12579 L .55621 .13109 L closepath p .227 .943 .204 r
F P
0 g
s
.55683 .12183 m .55981 .12208 L .55622 .12739 L closepath p .216 .84 .205 r
F P
0 g
s
.55628 .21604 m .56496 .22144 L .55941 .21897 L p .149 .021 0 r
F P
0 g
s
.57992 .22737 m .55657 .2227 L .56496 .22144 L closepath p .164 .111 0 r
F P
0 g
s
.57992 .22737 m .57164 .22868 L .55657 .2227 L closepath p .164 .111 0 r
F P
0 g
s
.56063 .11831 m .55683 .12183 L .55766 .11805 L closepath p .223 .769 .179 r
F P
0 g
s
.56063 .11831 m .55981 .12208 L .55683 .12183 L closepath p .223 .769 .179 r
F P
0 g
s
.58264 .3374 m .55689 .35094 L .56664 .32891 L closepath p .141 .454 .172 r
F P
0 g
s
.58264 .3374 m .57284 .35948 L .55689 .35094 L closepath p .141 .454 .172 r
F P
0 g
s
.55958 .39262 m .55689 .35094 L .57284 .35948 L closepath p .132 .47 .184 r
F P
0 g
s
.55883 .11626 m .56063 .11831 L .55766 .11805 L closepath p .222 .613 .129 r
F P
0 g
s
.55958 .39262 m .57284 .35948 L .57716 .36394 L closepath p .095 .308 .162 r
F P
0 g
s
.55958 .39262 m .57716 .36394 L .57766 .37077 L closepath p .049 .085 .123 r
F P
0 g
s
.55958 .39262 m .57766 .37077 L .57879 .37673 L closepath p .113 .435 .187 r
F P
0 g
s
.55958 .39262 m .57879 .37673 L .58088 .3854 L closepath p .133 .576 .211 r
F P
0 g
s
.5767 .41854 m .55958 .39262 L .58088 .3854 L closepath p .13 .532 .202 r
F P
0 g
s
.5598 .12579 m .56063 .11831 L .56138 .11967 L closepath p .243 .925 .171 r
F P
0 g
s
.5598 .12579 m .55981 .12208 L .56063 .11831 L closepath p .233 .871 .111 r
F P
0 g
s
.57784 .24185 m .5601 .23408 L .57884 .23161 L closepath p .153 .135 .023 r
F P
0 g
s
.59469 .25126 m .56574 .24738 L .57784 .24185 L closepath p .167 .285 .077 r
F P
0 g
s
.5704 .27099 m .56574 .24738 L .58261 .25686 L closepath p .157 .373 .128 r
F P
0 g
s
.59469 .25126 m .58261 .25686 L .56574 .24738 L closepath p .167 .285 .077 r
F P
0 g
s
.58264 .3374 m .56664 .32891 L .57731 .30838 L closepath p .147 .439 .161 r
F P
0 g
s
.58688 .49153 m .60338 .48314 L .56943 .50576 L closepath p .127 .896 .185 r
F P
0 g
s
.57409 .45171 m .59558 .46833 L .57001 .48803 L closepath p .152 .658 .218 r
F P
0 g
s
.59558 .46833 m .58688 .49153 L .57001 .48803 L closepath p .17 .846 .24 r
F P
0 g
s
.58653 .28396 m .5704 .27099 L .58261 .25686 L closepath p .171 .45 .14 r
F P
0 g
s
.57731 .30838 m .5704 .27099 L .58653 .28396 L closepath p .164 .46 .152 r
F P
0 g
s
.60056 .23177 m .57164 .22868 L .57992 .22737 L closepath p .108 0 0 r
F P
0 g
s
.60785 .2347 m .57884 .23161 L .57164 .22868 L closepath p .088 0 0 r
F P
0 g
s
.60785 .2347 m .57164 .22868 L .60056 .23177 L closepath p .088 0 0 r
F P
0 g
s
.57284 .35948 m .58264 .3374 L .59409 .34194 L closepath p .157 .528 .18 r
F P
0 g
s
.57284 .35948 m .59409 .34194 L .59847 .34637 L closepath p .126 .322 .142 r
F P
0 g
s
.57284 .35948 m .59847 .34637 L .57716 .36394 L closepath p .126 .322 .142 r
F P
0 g
s
.59638 .43828 m .59558 .46833 L .57409 .45171 L closepath p .151 .658 .219 r
F P
0 g
s
.5767 .41854 m .57808 .40762 L p .59655 .4114 L .13 .532 .202 r
F P
0 g
s
.59663 .41199 m .5767 .41854 L p .59655 .4114 L .13 .532 .202 r
F P
0 g
s
.59638 .43828 m .5767 .41854 L .58656 .4153 L p .133 .592 .215 r
F P
0 g
s
.60014 .35518 m .57716 .36394 L .59847 .34637 L closepath p .08 .058 .088 r
F P
0 g
s
.57716 .36394 m .60014 .35518 L .60068 .36206 L closepath p .102 .236 .134 r
F P
0 g
s
.57716 .36394 m .60068 .36206 L .57766 .37077 L closepath p .102 .236 .134 r
F P
0 g
s
.59822 .29729 m .57731 .30838 L .58653 .28396 L closepath p .162 .453 .151 r
F P
0 g
s
.59822 .29729 m .589 .32181 L .57731 .30838 L closepath p .162 .453 .151 r
F P
0 g
s
.58264 .3374 m .57731 .30838 L .589 .32181 L closepath p .151 .448 .161 r
F P
0 g
s
.60068 .36206 m .60061 .37382 L .57766 .37077 L closepath p .117 .39 .171 r
F P
0 g
s
.57766 .37077 m .60061 .37382 L .60174 .37982 L closepath p .114 .438 .188 r
F P
0 g
s
.57766 .37077 m .60174 .37982 L .57879 .37673 L closepath p .114 .438 .188 r
F P
0 g
s
.57784 .24185 m .57884 .23161 L .59657 .23933 L closepath p .153 .135 .023 r
F P
0 g
s
.57784 .24185 m .59657 .23933 L .6098 .24673 L closepath p .159 .173 .033 r
F P
0 g
s
.57784 .24185 m .6098 .24673 L .59469 .25126 L closepath p .168 .252 .059 r
F P
0 g
s
.57808 .40762 m .58088 .3854 L p .59635 .40952 L .13 .532 .202 r
F P
0 g
s
.57808 .40762 m .59635 .40952 L .59639 .4099 L .13 .532 .202 r
F
.57808 .40762 m .59639 .4099 L .59644 .41032 L F
.57808 .40762 m .59644 .41032 L .59655 .4114 L F
.60046 .38438 m .57879 .37673 L p .60065 .38374 L .103 .516 .213 r
F P
0 g
s
.60171 .38029 m .60142 .38126 L .57879 .37673 L .103 .516 .213 r
F
.60142 .38126 m .60065 .38374 L .57879 .37673 L F
.57879 .37673 m .60174 .37982 L .60182 .37996 L p F P
0 g
s
.60182 .37996 m .60171 .38029 L .57879 .37673 L .103 .516 .213 r
F
.59873 .38937 m .57879 .37673 L p .59881 .38916 L .108 .507 .209 r
F P
0 g
s
.58088 .3854 m .57879 .37673 L .59873 .38937 L p .108 .507 .209 r
F P
0 g
s
.59967 .38687 m .57879 .37673 L .59968 .38684 L .108 .507 .209 r
F
.59969 .38682 m .57879 .37673 L .59969 .38634 L F
.59968 .38684 m .57879 .37673 L .59969 .38682 L F
.59881 .38916 m .57879 .37673 L .59953 .3868 L F
.57879 .37673 m .60046 .38438 L p .59983 .3864 L F P
0 g
s
.61921 .23895 m .59657 .23933 L .57884 .23161 L closepath p .148 .07 0 r
F P
0 g
s
.61921 .23895 m .57884 .23161 L .60785 .2347 L closepath p .1 0 0 r
F P
0 g
s
.59528 .40725 m .58088 .3854 L .59558 .39062 L p .112 .5 .205 r
F P
0 g
s
.58088 .3854 m .59627 .40874 L p .59635 .40952 L .13 .532 .202 r
F P
0 g
s
.59795 .39146 m .58088 .3854 L p .5981 .39105 L .108 .507 .209 r
F P
0 g
s
.5981 .39105 m .58088 .3854 L .59832 .39047 L .108 .507 .209 r
F
.59832 .39047 m .58088 .3854 L .59834 .39042 L F
.59834 .39042 m .58088 .3854 L .59834 .39041 L F
.59834 .39041 m .58088 .3854 L .59836 .39036 L F
.59836 .39036 m .58088 .3854 L .59841 .39024 L F
.59841 .39024 m .58088 .3854 L .59845 .39013 L F
.59845 .39013 m .58088 .3854 L .5986 .38972 L F
.5986 .38972 m .58088 .3854 L .5986 .38972 L F
.5986 .38972 m .58088 .3854 L .59863 .38964 L F
.59863 .38964 m .58088 .3854 L .59864 .38961 L F
.59864 .38961 m .58088 .3854 L .59868 .38951 L F
.59868 .38951 m .58088 .3854 L .59868 .3895 L F
.59868 .3895 m .58088 .3854 L .59869 .38948 L F
.59869 .38948 m .58088 .3854 L .59869 .38947 L F
.59869 .38947 m .58088 .3854 L .5987 .38946 L F
.5987 .38946 m .58088 .3854 L .59872 .3894 L F
.59872 .3894 m .58088 .3854 L .59873 .38937 L F
.59873 .38937 m .58088 .3854 L .59873 .38937 L F
.61174 .26232 m .58261 .25686 L .59469 .25126 L closepath p .178 .354 .092 r
F P
0 g
s
.58653 .28396 m .58261 .25686 L .59874 .26975 L closepath p .171 .45 .14 r
F P
0 g
s
.61174 .26232 m .59874 .26975 L .58261 .25686 L closepath p .181 .424 .117 r
F P
0 g
s
.59409 .34194 m .58264 .3374 L .589 .32181 L closepath p .163 .502 .167 r
F P
0 g
s
.59822 .29729 m .58653 .28396 L .59874 .26975 L closepath p .172 .455 .141 r
F P
0 g
s
.58656 .4153 m .59662 .41199 L p .59719 .41893 L .133 .592 .215 r
F P
0 g
s
.59687 .4215 m .58656 .4153 L .59719 .41893 L .133 .592 .215 r
F
.59638 .43828 m .58656 .4153 L .59687 .4215 L F
.59558 .46833 m .61378 .47223 L .58688 .49153 L closepath p .17 .846 .24 r
F P
0 g
s
.60338 .48314 m .58688 .49153 L .61378 .47223 L closepath p .135 .914 .189 r
F P
0 g
s
.59822 .29729 m .60045 .32633 L .589 .32181 L closepath p .167 .475 .153 r
F P
0 g
s
.59409 .34194 m .589 .32181 L .60045 .32633 L closepath p .163 .502 .167 r
F P
0 g
s
.60045 .32633 m .59847 .34637 L .59409 .34194 L closepath p .121 .321 .146 r
F P
0 g
s
.61174 .26232 m .59469 .25126 L .6098 .24673 L closepath p .18 .329 .078 r
F P
0 g
s
.596 .40624 m .59528 .40725 L .59585 .40492 L .112 .5 .205 r
F
.59585 .40492 m .59528 .40725 L .59533 .40481 L F
.59602 .40644 m .59528 .40725 L .596 .40627 L F
.596 .40627 m .59528 .40725 L .596 .40624 L F
.59627 .40874 m .59528 .40725 L p .59602 .40648 L F P
0 g
s
.59602 .40648 m .59528 .40725 L .59602 .40644 L .112 .5 .205 r
F
.59605 .40031 m .59533 .40481 L .59546 .397 L F
.5959 .40434 m .59533 .40481 L .59605 .40031 L F
.59589 .40466 m .59533 .40481 L .5959 .40434 L F
.59588 .40492 m .59533 .40481 L .59589 .40466 L F
.59624 .39818 m .59546 .397 L .59558 .39062 L F
.59605 .40031 m .59546 .397 L .59611 .39876 L F
.59614 .39862 m .59546 .397 L .59615 .39858 L F
.59611 .39877 m .59546 .397 L .59614 .39862 L F
.59615 .39858 m .59546 .397 L .59623 .39824 L F
.59623 .39824 m .59546 .397 L .59624 .39818 L F
.59663 .39631 m .59661 .3964 L .59558 .39062 L F
.59661 .3964 m .59626 .39809 L .59558 .39062 L F
.59626 .39809 m .59624 .39818 L .59558 .39062 L F
.5972 .39359 m .59663 .39631 L .59558 .39062 L F
.59558 .39062 m .5975 .3913 L p .59726 .39331 L F P
0 g
s
.59726 .39331 m .5972 .39359 L .59558 .39062 L .112 .5 .205 r
F
.59558 .46833 m .59612 .44791 L p .60773 .44858 L .164 .702 .22 r
F P
0 g
s
.61112 .45435 m .59558 .46833 L p .60773 .44858 L .164 .702 .22 r
F P
0 g
s
.59558 .46833 m .61144 .4549 L .6116 .45516 L .18 .84 .235 r
F
.59558 .46833 m .61112 .45435 L p .61144 .4549 L F P
0 g
s
.61488 .46075 m .59558 .46833 L p .6116 .45516 L .18 .84 .235 r
F P
0 g
s
.61378 .47223 m .59558 .46833 L .61456 .46088 L p .18 .84 .235 r
F P
0 g
s
.59588 .40492 m .59587 .40509 L .59585 .40492 L .112 .5 .205 r
F
.59679 .40511 m .59587 .40509 L .59611 .39879 L closepath p 0 0 .088 r
F P
0 g
s
.59679 .40511 m .59655 .4114 L .59587 .40509 L closepath p 0 0 .088 r
F P
0 g
s
.59735 .39351 m .59611 .39879 L .59726 .39331 L closepath p 0 0 0 r
F P
0 g
s
.59735 .39351 m .5962 .399 L .59611 .39879 L closepath p 0 0 0 r
F P
0 g
s
.59611 .39879 m .59611 .39876 L .59611 .39878 L .112 .5 .205 r
F
.59611 .39879 m .5962 .399 L .59679 .40511 L closepath p 0 0 .031 r
F P
0 g
s
.59611 .39878 m .59611 .39876 L .59611 .39877 L .112 .5 .205 r
F
.59612 .44791 m .59638 .43828 L p .60702 .44388 L .164 .702 .22 r
F P
0 g
s
.60707 .44397 m .60771 .4485 L .59612 .44791 L .164 .702 .22 r
F
.60702 .44389 m .60707 .44397 L .59612 .44791 L F
.60702 .44388 m .60702 .44389 L .59612 .44791 L F
.60771 .4485 m .60773 .44858 L .59612 .44791 L F
.5962 .399 m .59735 .39351 L .59812 .39386 L closepath p .114 .338 .158 r
F P
0 g
s
.5962 .399 m .59812 .39386 L .59798 .39936 L closepath p .13 .409 .167 r
F P
0 g
s
.5962 .399 m .59858 .40547 L .59679 .40511 L closepath p .124 .464 .187 r
F P
0 g
s
.5962 .399 m .59798 .39936 L .59858 .40547 L closepath p .124 .464 .187 r
F P
0 g
s
.59735 .42304 m .59638 .43828 L p .59687 .4215 L .133 .592 .215 r
F P
0 g
s
.59638 .43828 m .597 .42853 L p .60004 .43152 L .122 .551 .212 r
F P
0 g
s
.60008 .43166 m .59638 .43828 L p .60004 .43152 L .122 .551 .212 r
F P
0 g
s
.59638 .43828 m .60008 .43166 L p .60064 .43341 L .155 .72 .229 r
F P
0 g
s
.60067 .4335 m .59638 .43828 L p .60064 .43341 L .155 .72 .229 r
F P
0 g
s
.59638 .43828 m .60067 .4335 L p .6009 .43577 L .159 .746 .232 r
F P
0 g
s
.59638 .43828 m .60141 .43548 L .60144 .43552 L .159 .746 .232 r
F
.60149 .43558 m .59638 .43828 L p .60144 .43552 L F P
0 g
s
.59638 .43828 m .60149 .43558 L p .60304 .43753 L .156 .705 .226 r
F P
0 g
s
.60423 .43904 m .59638 .43828 L .60307 .43757 L .156 .705 .226 r
F
.60307 .43757 m .59638 .43828 L .60305 .43754 L F
.60305 .43754 m .59638 .43828 L .60304 .43754 L F
.60304 .43754 m .59638 .43828 L .60304 .43754 L F
.60304 .43754 m .59638 .43828 L .60304 .43753 L F
.59638 .43828 m .60664 .44321 L p .60702 .44388 L .164 .702 .22 r
F P
0 g
s
.60446 .43933 m .59638 .43828 L .60367 .43899 L .156 .705 .226 r
F
.60481 .43996 m .59638 .43828 L .60446 .43933 L F
.60603 .44213 m .59638 .43828 L .60481 .43996 L F
.60664 .44321 m .59638 .43828 L p .60609 .44224 L F P
0 g
s
.60609 .44224 m .59638 .43828 L .60603 .44213 L .156 .705 .226 r
F
.59679 .40511 m .60248 .41261 L .59655 .4114 L closepath p .117 .505 .203 r
F P
0 g
s
.59655 .4114 m .60248 .41261 L .60331 .41865 L closepath p .128 .433 .176 r
F P
0 g
s
.59655 .4114 m .60331 .41865 L .59738 .41743 L closepath p .128 .433 .176 r
F P
0 g
s
.63244 .24631 m .59657 .23933 L .61921 .23895 L closepath p .158 .128 .009 r
F P
0 g
s
.63244 .24631 m .6098 .24673 L .59657 .23933 L closepath p .158 .128 .009 r
F P
0 g
s
.59719 .41893 m .59662 .41199 L .59738 .41743 L .133 .592 .215 r
F
.59738 .41743 m .59662 .41199 L .59672 .41259 L F
.59672 .41259 m .59662 .41199 L .59669 .4124 L F
.59667 .41227 m .59662 .41199 L .59667 .41224 L F
.59664 .41205 m .59662 .41199 L .59664 .41203 L F
.59667 .41222 m .59662 .41199 L .59664 .41205 L F
.59667 .41224 m .59662 .41199 L .59667 .41222 L F
.59669 .4124 m .59662 .41199 L .59667 .41227 L F
.59662 .41199 m .59663 .41199 L p .59664 .41203 L F P
0 g
s
.59679 .40511 m .59858 .40547 L .60248 .41261 L closepath p .117 .505 .203 r
F P
0 g
s
.59876 .4273 m .59687 .4215 L .59738 .41743 L closepath p .024 .076 .136 r
F P
0 g
s
.59876 .4273 m .60122 .43524 L .59687 .4215 L closepath p 0 0 .059 r
F P
0 g
s
.597 .42853 m .59735 .42304 L p .59981 .43078 L .122 .551 .212 r
F P
0 g
s
.597 .42853 m .59981 .43078 L .59984 .43088 L .122 .551 .212 r
F
.597 .42853 m .59984 .43088 L .60004 .43152 L F
.59934 .38775 m .59735 .39351 L .59726 .39331 L closepath p 0 0 0 r
F P
0 g
s
.5975 .3913 m .59795 .39146 L p .59726 .39331 L .112 .5 .205 r
F P
0 g
s
.59812 .39386 m .59735 .39351 L .59934 .38775 L closepath p .124 .3 .136 r
F P
0 g
s
.60106 .4236 m .59738 .41743 L .60331 .41865 L closepath p .135 .326 .135 r
F P
0 g
s
.60106 .4236 m .59927 .42323 L .59738 .41743 L closepath p .135 .326 .135 r
F P
0 g
s
.59876 .4273 m .59738 .41743 L .59927 .42323 L closepath p .024 .076 .136 r
F P
0 g
s
.59798 .39936 m .59812 .39386 L .60079 .39442 L closepath p .186 .595 .173 r
F P
0 g
s
.59798 .39936 m .60079 .39442 L .60066 .39991 L closepath p .186 .595 .173 r
F P
0 g
s
.59858 .40547 m .59798 .39936 L .60066 .39991 L closepath p .182 .638 .189 r
F P
0 g
s
.59812 .39386 m .59934 .38775 L .60011 .3881 L closepath p .124 .3 .136 r
F P
0 g
s
.60079 .39442 m .59812 .39386 L .60011 .3881 L closepath p .188 .569 .163 r
F P
0 g
s
.59822 .29729 m .59874 .26975 L .61442 .27669 L closepath p .159 .412 .14 r
F P
0 g
s
.59822 .29729 m .61442 .27669 L .61285 .29281 L closepath p .149 .357 .132 r
F P
0 g
s
.60045 .32633 m .59822 .29729 L .61285 .29281 L closepath p .158 .446 .154 r
F P
0 g
s
.60045 .32633 m .611 .33932 L .59847 .34637 L closepath p .134 .363 .149 r
F P
0 g
s
.60014 .35518 m .59847 .34637 L .611 .33932 L closepath p .094 .091 .087 r
F P
0 g
s
.60066 .39991 m .60248 .41261 L .59858 .40547 L closepath p .177 .616 .188 r
F P
0 g
s
.59874 .26975 m .61174 .26232 L .62741 .26922 L closepath p .166 .324 .096 r
F P
0 g
s
.59874 .26975 m .62741 .26922 L .61442 .27669 L closepath p .166 .324 .096 r
F P
0 g
s
.59927 .42323 m .59981 .4278 L .59876 .4273 L closepath p .021 .066 .134 r
F P
0 g
s
.59876 .4273 m .60007 .43144 L .60122 .43524 L closepath p 0 0 .059 r
F P
0 g
s
.59981 .4278 m .60007 .43144 L .59876 .4273 L closepath p .002 .124 .16 r
F P
0 g
s
.59881 .38916 m .59953 .3868 L .59934 .38775 L .108 .507 .209 r
F
.6016 .42818 m .59927 .42323 L .60106 .4236 L closepath p .132 .387 .158 r
F P
0 g
s
.6016 .42818 m .59981 .4278 L .59927 .42323 L closepath p .132 .387 .158 r
F P
0 g
s
.60011 .3881 m .59934 .38775 L .60176 .38133 L closepath p .13 .263 .116 r
F P
0 g
s
.59934 .38775 m .59953 .3868 L .59938 .38765 L .108 .507 .209 r
F
.59938 .38765 m .59953 .3868 L .59967 .38687 L F
.59969 .38682 m .59969 .38634 L .59983 .3864 L F
.59969 .38682 m .59969 .38634 L .59969 .38682 L F
.59969 .38682 m .59983 .3864 L .59979 .38656 L F
.59979 .38656 m .60266 .3772 L p .60176 .38133 L 0 0 0 r
F P
0 g
s
.60281 .43157 m .59981 .4278 L .6016 .42818 L closepath p .12 .491 .198 r
F P
0 g
s
.60281 .43157 m .60112 .43194 L .59981 .4278 L closepath p .166 .544 .177 r
F P
0 g
s
.59981 .4278 m .60112 .43194 L .60007 .43144 L closepath p .002 .124 .16 r
F P
0 g
s
.60007 .43144 m .60112 .43194 L .60331 .43552 L closepath p 0 .202 .183 r
F P
0 g
s
.60331 .43552 m .60122 .43524 L .60007 .43144 L closepath p .018 .309 .198 r
F P
0 g
s
.60011 .3881 m .60176 .38133 L .60433 .38247 L closepath p .13 .263 .116 r
F P
0 g
s
.60011 .3881 m .60433 .38247 L .60502 .38879 L closepath p .19 .576 .162 r
F P
0 g
s
.60011 .3881 m .60502 .38879 L .60079 .39442 L closepath p .19 .576 .162 r
F P
0 g
s
.60014 .35518 m .611 .33932 L .61124 .34054 L p .094 .091 .087 r
F P
0 g
s
.60014 .35518 m .61124 .34054 L .61005 .34441 L .094 .091 .087 r
F
.60014 .35518 m .61005 .34441 L .60983 .34513 L F
.60853 .35048 m .60014 .35518 L p .60855 .34645 L F P
0 g
s
.60853 .35116 m .60014 .35518 L .60853 .35048 L p 0 0 0 r
F P
0 g
s
.60018 .35562 m .60014 .35518 L .60853 .35116 L p .061 .121 .126 r
F P
0 g
s
.60853 .35131 m .60018 .35562 L .60853 .3512 L .061 .121 .126 r
F
.60853 .3512 m .60018 .35562 L .60853 .35116 L F
.60068 .36206 m .60018 .35562 L p .60853 .35131 L F P
0 g
s
.60045 .32633 m .61285 .29281 L .61514 .32186 L closepath p .158 .446 .154 r
F P
0 g
s
.60045 .32633 m .61514 .32186 L .611 .33932 L closepath p .15 .374 .136 r
F P
0 g
s
.60056 .23177 m .63197 .24137 L .60785 .2347 L closepath p .077 0 0 r
F P
0 g
s
.60056 .23177 m .62019 .2372 L .63197 .24137 L closepath p .077 0 0 r
F P
0 g
s
.60061 .37382 m .60068 .36206 L .60169 .36192 L p .023 .078 .137 r
F P
0 g
s
.60236 .36576 m .60061 .37382 L .60169 .36192 L .023 .078 .137 r
F
.60248 .36892 m .60061 .37382 L .60236 .36576 L F
.60253 .37018 m .60061 .37382 L p .60248 .36892 L F P
0 g
s
.60061 .37382 m .60253 .37018 L p .60257 .37128 L .023 .078 .137 r
F P
0 g
s
.60257 .37128 m .60258 .37163 L .60061 .37382 L .023 .078 .137 r
F
.60258 .37163 m .60259 .37191 L .60061 .37382 L F
.60259 .37191 m .6026 .37202 L .60061 .37382 L F
.6026 .37204 m .6026 .37211 L .60061 .37382 L F
.6026 .37202 m .6026 .37204 L .60061 .37382 L F
.6026 .37222 m .6026 .37225 L .60061 .37382 L F
.6026 .37211 m .6026 .37213 L .60061 .37382 L F
.6026 .37213 m .6026 .37217 L .60061 .37382 L F
.6026 .37217 m .6026 .37222 L .60061 .37382 L F
.6025 .37233 m .60261 .37242 L .60061 .37382 L F
.60261 .37242 m .60263 .37261 L .60061 .37382 L F
.60263 .37261 m .60266 .37298 L .60061 .37382 L F
.6027 .37355 m .60061 .37382 L p .60266 .37298 L F P
0 g
s
.60061 .37382 m .6027 .37355 L p .6029 .37609 L .026 .18 .165 r
F P
0 g
s
.60087 .37522 m .60061 .37382 L p .6029 .37609 L .026 .18 .165 r
F P
0 g
s
.60066 .39991 m .60502 .38879 L .60456 .40706 L closepath p .202 .649 .171 r
F P
0 g
s
.60066 .39991 m .60079 .39442 L .60502 .38879 L closepath p .202 .649 .171 r
F P
0 g
s
.60066 .39991 m .60456 .40706 L .60248 .41261 L closepath p .177 .616 .188 r
F P
0 g
s
.60067 .4335 m .6007 .43359 L .6009 .43577 L .159 .746 .232 r
F
.60853 .35195 m .60068 .36206 L .60853 .35131 L .061 .121 .126 r
F
.60852 .35319 m .60068 .36206 L .60853 .35195 L F
.60852 .35355 m .60068 .36206 L .60833 .35341 L F
.60841 .35388 m .60068 .36206 L p .60852 .35355 L F P
0 g
s
.60068 .36206 m .60841 .35388 L p .60833 .35411 L .013 0 .069 r
F P
0 g
s
.60833 .35411 m .60805 .35493 L .60068 .36206 L .013 0 .069 r
F
.60447 .36153 m .60068 .36206 L p .60805 .35493 L F P
0 g
s
.6007 .43359 m .60122 .43524 L .6009 .43577 L .159 .746 .232 r
F
.6029 .37612 m .60087 .37522 L .6029 .37609 L .026 .18 .165 r
F
.60174 .37982 m .60087 .37522 L p .6029 .37612 L F P
0 g
s
.6029 .37609 m .60087 .37522 L .6029 .37609 L .026 .18 .165 r
F
.6029 .37609 m .60087 .37522 L .6029 .37609 L F
.60122 .43524 m .60122 .43525 L .6009 .43577 L .159 .746 .232 r
F
.6009 .43577 m .60122 .43525 L .60141 .43548 L F
.60106 .4236 m .60331 .41865 L .60655 .42273 L closepath p .186 .556 .16 r
F P
0 g
s
.60106 .4236 m .60655 .42273 L .6043 .42768 L closepath p .186 .556 .16 r
F P
0 g
s
.6043 .42768 m .6016 .42818 L .60106 .4236 L closepath p .191 .557 .154 r
F P
0 g
s
.60112 .43194 m .60281 .43157 L .60499 .43514 L closepath p .171 .676 .209 r
F P
0 g
s
.60112 .43194 m .60499 .43514 L .60331 .43552 L closepath p .171 .676 .209 r
F P
0 g
s
.60331 .43552 m .60445 .43931 L .60122 .43524 L closepath p .018 .309 .198 r
F P
0 g
s
.60281 .43157 m .6016 .42818 L .6043 .42768 L closepath p .197 .664 .182 r
F P
0 g
s
.60169 .36192 m .60428 .36156 L p .60424 .36195 L .023 .078 .137 r
F P
0 g
s
.60417 .36212 m .60397 .36263 L .60169 .36192 L .023 .078 .137 r
F
.60424 .36195 m .60417 .36212 L .60169 .36192 L F
.60309 .36236 m .60296 .36515 L .60169 .36192 L F
.60296 .36515 m .60281 .36609 L .60169 .36192 L F
.60281 .36609 m .60262 .36727 L .60169 .36192 L F
.60277 .37671 m .60174 .37982 L .6029 .37612 L .026 .18 .165 r
F
.60274 .37684 m .60174 .37982 L .60277 .37671 L F
.60268 .37713 m .60174 .37982 L .60274 .37684 L F
.60266 .3772 m .60174 .37982 L .60268 .37713 L F
.60255 .37757 m .60174 .37982 L .60266 .3772 L F
.60252 .37767 m .60174 .37982 L .60255 .37757 L F
.60194 .37956 m .60174 .37982 L p .60252 .37767 L F P
0 g
s
.60174 .37982 m .60194 .37956 L p .60192 .37962 L 0 0 .117 r
F P
0 g
s
.60183 .37991 m .60174 .37982 L p .60183 .3799 L 0 0 .117 r
F P
0 g
s
.60183 .3799 m .60183 .3799 L .60174 .37982 L 0 0 .117 r
F
.60192 .37962 m .60183 .3799 L .60174 .37982 L F
.60182 .37996 m .60174 .37982 L .60183 .37991 L p 0 0 .141 r
F P
0 g
s
.6029 .37612 m .60433 .38247 L .60176 .38133 L closepath p .114 .34 .159 r
F P
0 g
s
.60247 .36823 m .60245 .36831 L .60236 .36576 L .023 .078 .137 r
F
.60262 .3673 m .60261 .36731 L .60236 .36576 L F
.60262 .36727 m .60262 .3673 L .60236 .36576 L F
.60261 .36731 m .60246 .36801 L .60236 .36576 L F
.60328 .36499 m .60245 .36831 L .60296 .36515 L closepath p 0 0 .029 r
F P
0 g
s
.60328 .36499 m .60277 .36815 L .60245 .36831 L closepath p 0 0 .029 r
F P
0 g
s
.60245 .36831 m .60277 .36815 L .60275 .3721 L closepath p 0 0 .07 r
F P
0 g
s
.60275 .3721 m .60261 .37234 L .60245 .36831 L closepath p 0 0 .028 r
F P
0 g
s
.6025 .36803 m .60246 .36801 L .60261 .36731 L .023 .078 .137 r
F
.6025 .36803 m .60247 .36823 L .60246 .36801 L F
.6025 .36803 m .60246 .36801 L .6025 .36803 L F
.6025 .36803 m .60246 .36801 L .6025 .36803 L F
.61402 .40455 m .60248 .41261 L .60456 .40706 L closepath p .19 .675 .192 r
F P
0 g
s
.61402 .40455 m .60331 .41865 L .60248 .41261 L closepath p .179 .58 .175 r
F P
0 g
s
.6026 .37225 m .60261 .37234 L .6025 .37233 L .023 .078 .137 r
F
.60261 .37234 m .60261 .37242 L .6025 .37233 L F
.60275 .3721 m .6029 .37612 L .60261 .37234 L closepath p 0 0 .028 r
F P
0 g
s
.60277 .36815 m .60352 .37244 L .60275 .3721 L closepath p .093 .391 .188 r
F P
0 g
s
.60275 .3721 m .60352 .37244 L .60547 .37726 L closepath p .069 .421 .207 r
F P
0 g
s
.60275 .3721 m .60547 .37726 L .6029 .37612 L closepath p .069 .421 .207 r
F P
0 g
s
.60355 .36849 m .60277 .36815 L .60328 .36499 L closepath p .111 .346 .162 r
F P
0 g
s
.60277 .36815 m .60355 .36849 L .60352 .37244 L closepath p .093 .391 .188 r
F P
0 g
s
.60281 .43157 m .6043 .42768 L .6055 .43107 L closepath p .197 .664 .182 r
F P
0 g
s
.60499 .43514 m .60281 .43157 L .6055 .43107 L closepath p .198 .712 .194 r
F P
0 g
s
.6029 .37612 m .60547 .37726 L .60433 .38247 L closepath p .113 .34 .159 r
F P
0 g
s
.60296 .36515 m .60696 .35812 L .60564 .36143 L closepath p 0 0 0 r
F P
0 g
s
.60296 .36515 m .60428 .36184 L .60696 .35812 L closepath p 0 0 0 r
F P
0 g
s
.60564 .36143 m .60328 .36499 L .60296 .36515 L closepath p 0 0 0 r
F P
0 g
s
.60397 .36263 m .60296 .36515 L .60309 .36236 L .023 .078 .137 r
F
.60526 .36471 m .60328 .36499 L .60564 .36143 L closepath p .129 .277 .122 r
F P
0 g
s
.60355 .36849 m .60328 .36499 L .60526 .36471 L closepath p .14 .433 .166 r
F P
0 g
s
.60331 .43552 m .60499 .43514 L .61004 .43808 L closepath p .17 .755 .227 r
F P
0 g
s
.60331 .43552 m .61004 .43808 L .60445 .43931 L closepath p .17 .755 .227 r
F P
0 g
s
.61402 .40455 m .60655 .42273 L .60331 .41865 L closepath p .174 .55 .171 r
F P
0 g
s
.60338 .48314 m .61378 .47223 L .62198 .46808 L p .135 .914 .189 r
F P
0 g
s
.60338 .48314 m .62198 .46808 L .62273 .46882 L .135 .914 .189 r
F
.62295 .46912 m .60338 .48314 L p .62273 .46882 L F P
0 g
s
.60352 .37244 m .60355 .36849 L .6062 .36872 L closepath p .184 .701 .204 r
F P
0 g
s
.6062 .36872 m .60547 .37726 L .60352 .37244 L closepath p .184 .696 .203 r
F P
0 g
s
.60355 .36849 m .60526 .36471 L .60792 .36494 L closepath p .188 .658 .189 r
F P
0 g
s
.60355 .36849 m .60792 .36494 L .6062 .36872 L closepath p .188 .658 .189 r
F P
0 g
s
.60446 .43933 m .60367 .43899 L .60445 .43931 L .156 .705 .226 r
F
.60445 .43931 m .60367 .43899 L .60423 .43904 L F
.60428 .36184 m .60424 .36195 L .60428 .36156 L .023 .078 .137 r
F
.60696 .35812 m .60428 .36184 L .60754 .35644 L p 0 0 0 r
F P
0 g
s
.60443 .3616 m .60428 .36184 L .60428 .36156 L .023 .078 .137 r
F
.60428 .36156 m .60447 .36153 L p .60443 .3616 L F P
0 g
s
.6055 .43107 m .6043 .42768 L .60655 .42273 L closepath p .213 .694 .171 r
F P
0 g
s
.61701 .37314 m .60433 .38247 L .60547 .37726 L closepath p .179 .594 .18 r
F P
0 g
s
.61701 .37314 m .60502 .38879 L .60433 .38247 L closepath p .168 .512 .165 r
F P
0 g
s
.60445 .43931 m .61004 .43808 L .61266 .44274 L closepath p .169 .602 .191 r
F P
0 g
s
.60445 .43931 m .61266 .44274 L .60707 .44397 L closepath p .169 .602 .191 r
F P
0 g
s
.60754 .35644 m .60447 .36153 L .60805 .35493 L .013 0 .069 r
F
.61447 .38629 m .60456 .40706 L .60502 .38879 L closepath p .186 .595 .173 r
F P
0 g
s
.61447 .38629 m .61402 .40455 L .60456 .40706 L closepath p .186 .595 .173 r
F P
0 g
s
.6055 .43107 m .61004 .43808 L .60499 .43514 L closepath p .202 .729 .194 r
F P
0 g
s
.61701 .37314 m .61447 .38629 L .60502 .38879 L closepath p .188 .628 .181 r
F P
0 g
s
.60526 .36471 m .60564 .36143 L .60762 .36115 L closepath p .129 .277 .122 r
F P
0 g
s
.60792 .36494 m .60526 .36471 L .60762 .36115 L closepath p .19 .585 .165 r
F P
0 g
s
.6062 .36872 m .60815 .37355 L .60547 .37726 L closepath p .184 .696 .203 r
F P
0 g
s
.61701 .37314 m .60547 .37726 L .60815 .37355 L closepath p .186 .708 .205 r
F P
0 g
s
.6055 .43107 m .60655 .42273 L .61055 .43401 L closepath p .213 .694 .171 r
F P
0 g
s
.6055 .43107 m .61055 .43401 L .61004 .43808 L closepath p .202 .729 .194 r
F P
0 g
s
.60762 .36115 m .60564 .36143 L .60696 .35812 L closepath p .113 .108 .073 r
F P
0 g
s
.6077 .19181 m .61023 .207 L .60593 .21449 L closepath p .044 .152 .148 r
F P
0 g
s
.6077 .19181 m .60593 .21449 L .60623 .19446 L closepath p 0 0 .065 r
F P
0 g
s
.61141 .22966 m .60593 .21449 L .61023 .207 L closepath p .08 .329 .178 r
F P
0 g
s
.6112 .23239 m .60593 .21449 L .61141 .22966 L closepath p 0 0 .103 r
F P
0 g
s
.6062 .36872 m .60792 .36494 L .61384 .36097 L closepath p .203 .731 .194 r
F P
0 g
s
.6062 .36872 m .61384 .36097 L .60815 .37355 L closepath p .203 .731 .194 r
F P
0 g
s
.60979 .16606 m .6077 .19181 L .60623 .19446 L closepath p 0 0 .045 r
F P
0 g
s
.60979 .16606 m .60623 .19446 L .60832 .16878 L closepath p 0 0 .045 r
F P
0 g
s
.61402 .40455 m .6155 .41756 L .60655 .42273 L closepath p .194 .645 .179 r
F P
0 g
s
.6155 .41756 m .61055 .43401 L .60655 .42273 L closepath p .196 .663 .182 r
F P
0 g
s
.60852 .35355 m .61355 .35718 L .60696 .35812 L closepath p .14 .428 .165 r
F P
0 g
s
.60762 .36115 m .60696 .35812 L .61355 .35718 L closepath p .113 .107 .073 r
F P
0 g
s
.61286 .44914 m .60707 .44397 L .61266 .44274 L closepath p .162 .475 .159 r
F P
0 g
s
.61183 .45412 m .60773 .44858 L .60707 .44397 L closepath p .08 .369 .189 r
F P
0 g
s
.61183 .45412 m .60707 .44397 L .61117 .44951 L closepath p .08 .369 .189 r
F P
0 g
s
.61286 .44914 m .61117 .44951 L .60707 .44397 L closepath p .162 .475 .159 r
F P
0 g
s
.60762 .36115 m .61355 .35718 L .61384 .36097 L closepath p .185 .569 .166 r
F P
0 g
s
.60762 .36115 m .61384 .36097 L .60792 .36494 L closepath p .185 .569 .166 r
F P
0 g
s
.6159 .17262 m .6077 .19181 L .60979 .16606 L closepath p .1 .265 .145 r
F P
0 g
s
.6159 .17262 m .61871 .18607 L .6077 .19181 L closepath p .167 .564 .182 r
F P
0 g
s
.6077 .19181 m .61871 .18607 L .62126 .20126 L closepath p .156 .464 .161 r
F P
0 g
s
.6077 .19181 m .62126 .20126 L .61023 .207 L closepath p .156 .464 .161 r
F P
0 g
s
.61183 .45412 m .61531 .46148 L .60773 .44858 L closepath p 0 .09 .162 r
F P
0 g
s
.61921 .23895 m .60785 .2347 L .63197 .24137 L closepath p .077 0 0 r
F P
0 g
s
.6227 .36057 m .61701 .37314 L .60815 .37355 L closepath p .187 .655 .189 r
F P
0 g
s
.6227 .36057 m .60815 .37355 L .61384 .36097 L closepath p .187 .655 .189 r
F P
0 g
s
.6126 .15375 m .60979 .16606 L .60832 .16878 L closepath p 0 0 .011 r
F P
0 g
s
.60852 .35355 m .60833 .35341 L .60852 .35319 L .061 .121 .126 r
F
.61033 .34771 m .61511 .35261 L .60852 .35355 L closepath p .144 .59 .207 r
F P
0 g
s
.61033 .34771 m .60852 .35355 L .60854 .34929 L closepath p 0 0 .102 r
F P
0 g
s
.60852 .35355 m .61511 .35261 L .61355 .35718 L closepath p .14 .428 .165 r
F P
0 g
s
.61033 .34771 m .60854 .34929 L .61035 .34344 L closepath p 0 0 .102 r
F P
0 g
s
.60855 .34645 m .60903 .34768 L .60854 .34929 L .094 .091 .087 r
F
.60855 .34645 m .60983 .34513 L .60919 .3472 L F
.60855 .34645 m .60919 .3472 L .60903 .34768 L F
.6126 .15375 m .6159 .17262 L .60979 .16606 L closepath p .094 .261 .149 r
F P
0 g
s
.63924 .2529 m .6098 .24673 L .63244 .24631 L closepath p .174 .233 .038 r
F P
0 g
s
.61174 .26232 m .6098 .24673 L .62681 .25772 L closepath p .18 .329 .078 r
F P
0 g
s
.63924 .2529 m .62681 .25772 L .6098 .24673 L closepath p .18 .297 .059 r
F P
0 g
s
.6195 .42885 m .61004 .43808 L .61055 .43401 L closepath p .202 .726 .194 r
F P
0 g
s
.6195 .42885 m .61266 .44274 L .61004 .43808 L closepath p .189 .63 .181 r
F P
0 g
s
.62186 .206 m .61023 .207 L .62126 .20126 L closepath p .124 .228 .109 r
F P
0 g
s
.62186 .206 m .61572 .22223 L .61023 .207 L closepath p .137 .464 .178 r
F P
0 g
s
.61141 .22966 m .61023 .207 L .61572 .22223 L closepath p .08 .329 .178 r
F P
0 g
s
.61096 .343 m .61232 .34742 L .61033 .34771 L closepath p .143 .527 .191 r
F P
0 g
s
.61035 .34344 m .61096 .343 L .61033 .34771 L closepath p 0 0 .079 r
F P
0 g
s
.61033 .34771 m .61232 .34742 L .61511 .35261 L closepath p .144 .59 .207 r
F P
0 g
s
.61296 .33876 m .61096 .343 L .61035 .34344 L closepath p 0 0 0 r
F P
0 g
s
.61296 .33876 m .61035 .34344 L .61234 .33921 L closepath p 0 0 0 r
F P
0 g
s
.61035 .34344 m .61124 .34054 L .6107 .3427 L .094 .091 .087 r
F
.6155 .41756 m .6195 .42885 L .61055 .43401 L closepath p .196 .663 .182 r
F P
0 g
s
.61124 .34054 m .61138 .34124 L p .6107 .3427 L .094 .091 .087 r
F P
0 g
s
.61295 .34271 m .61096 .343 L .61296 .33876 L closepath p .138 .393 .155 r
F P
0 g
s
.61096 .343 m .61295 .34271 L .61232 .34742 L closepath p .143 .527 .191 r
F P
0 g
s
.611 .33932 m .6121 .3347 L p .61516 .33523 L .13 .298 .129 r
F P
0 g
s
.61408 .33676 m .611 .33932 L .61455 .33583 L .13 .298 .129 r
F
.61381 .33713 m .611 .33932 L .61408 .33676 L F
.61316 .33805 m .611 .33932 L .61381 .33713 L F
.61278 .33859 m .611 .33932 L .61316 .33805 L F
.6127 .3387 m .611 .33932 L .61259 .33867 L F
.61258 .33887 m .611 .33932 L .6127 .3387 L F
.61251 .33897 m .611 .33932 L .61258 .33887 L F
.61249 .339 m .611 .33932 L p .61251 .33897 L F P
0 g
s
.611 .33932 m .61249 .339 L p .61247 .33903 L .114 .14 .086 r
F P
0 g
s
.6124 .33904 m .61234 .33921 L .611 .33932 L .114 .14 .086 r
F
.61138 .34124 m .611 .33932 L p .61234 .33921 L F P
0 g
s
.61111 .23469 m .6112 .23239 L .61114 .2347 L p 0 0 0 r
F P
0 g
s
.61114 .2347 m .6112 .23239 L p .61115 .2347 L 0 0 0 r
F P
0 g
s
.6112 .23239 m .61116 .2347 L .61115 .2347 L 0 0 0 r
F
.6112 .23239 m .61116 .2347 L .61116 .2347 L F
.6112 .23239 m .61116 .2347 L .61116 .2347 L F
.6112 .23239 m .61117 .2347 L .61116 .2347 L F
.6112 .23239 m .61117 .2347 L .61117 .2347 L F
.6112 .23239 m .61117 .23471 L .61117 .2347 L F
.6112 .23239 m .61117 .23471 L .61117 .23471 L F
.6112 .23239 m .61117 .23471 L .61117 .23471 L F
.6112 .23239 m .61118 .23471 L .61117 .23471 L F
.61498 .45392 m .61117 .44951 L .61286 .44914 L closepath p .166 .541 .176 r
F P
0 g
s
.61498 .45392 m .6133 .45429 L .61117 .44951 L closepath p .166 .541 .176 r
F P
0 g
s
.61117 .44951 m .6133 .45429 L .61183 .45412 L closepath p .067 .326 .185 r
F P
0 g
s
.6112 .23239 m .61119 .23471 L .61118 .23471 L 0 0 0 r
F
.6112 .23239 m .6112 .23471 L .61119 .23471 L F
.6112 .23239 m .61121 .23472 L .6112 .23471 L F
.6112 .23239 m .61141 .22966 L .61121 .23472 L p F P
0 g
s
.61121 .23472 m .61141 .22966 L p .61127 .23473 L 0 0 .071 r
F P
0 g
s
.61141 .22966 m .61166 .23484 L .61127 .23473 L 0 0 .071 r
F
.61141 .22966 m .61572 .22223 L .6138 .23335 L closepath p 0 0 .06 r
F P
0 g
s
.61141 .22966 m .61196 .23492 L .61166 .23484 L 0 0 .071 r
F
.61141 .22966 m .61197 .23492 L .61196 .23492 L F
.61141 .22966 m .6138 .23335 L p .61197 .23492 L F P
0 g
s
.61174 .26232 m .62681 .25772 L .64022 .26533 L closepath p .156 .173 .041 r
F P
0 g
s
.61174 .26232 m .64022 .26533 L .62741 .26922 L closepath p .167 .281 .076 r
F P
0 g
s
.61183 .45412 m .61411 .458 L .61531 .46148 L closepath p 0 .09 .162 r
F P
0 g
s
.6133 .45429 m .61411 .458 L .61183 .45412 L closepath p .058 .407 .208 r
F P
0 g
s
.61196 .23492 m .6138 .23335 L p .61232 .23502 L 0 0 0 r
F P
0 g
s
.6121 .3347 m .61217 .33441 L p .61525 .33513 L .13 .298 .129 r
F P
0 g
s
.61516 .33523 m .6121 .3347 L .61523 .33515 L .13 .298 .129 r
F
.61523 .33515 m .6121 .3347 L .61525 .33513 L F
.61217 .33441 m .61514 .32186 L p .6183 .33177 L F P
0 g
s
.6182 .33188 m .61217 .33441 L .6183 .33177 L .13 .298 .129 r
F
.61537 .335 m .61217 .33441 L .6182 .33188 L F
.61537 .335 m .61525 .33513 L .61217 .33441 L F
.61232 .34742 m .61295 .34271 L .6161 .34326 L closepath p .171 .629 .197 r
F P
0 g
s
.6161 .34326 m .61511 .35261 L .61232 .34742 L closepath p .173 .644 .199 r
F P
0 g
s
.61239 .23504 m .61232 .23502 L .6138 .23335 L 0 0 0 r
F
.61234 .33921 m .61516 .33523 L .61893 .33107 L closepath p F P
0 g
s
.61234 .33921 m .61893 .33107 L .61612 .33504 L closepath p 0 0 0 r
F P
0 g
s
.61612 .33504 m .61296 .33876 L .61234 .33921 L closepath p 0 0 0 r
F P
0 g
s
.61247 .33903 m .61234 .33921 L .6124 .33904 L .114 .14 .086 r
F
.61268 .23512 m .61239 .23504 L .6138 .23335 L 0 0 0 r
F
.6127 .3387 m .61259 .33867 L .61278 .33859 L .13 .298 .129 r
F
.61871 .16027 m .6126 .15375 L .62463 .14935 L closepath p .108 .269 .139 r
F P
0 g
s
.6126 .15375 m .61871 .16027 L .6159 .17262 L closepath p .094 .261 .149 r
F P
0 g
s
.6195 .42885 m .61655 .44589 L .61266 .44274 L closepath p .187 .624 .18 r
F P
0 g
s
.61286 .44914 m .61266 .44274 L .61655 .44589 L closepath p .202 .601 .156 r
F P
0 g
s
.6138 .23335 m .61738 .23167 L .61268 .23512 L p 0 0 0 r
F P
0 g
s
.61268 .23512 m .61738 .23167 L .61547 .23589 L p .028 0 .027 r
F P
0 g
s
.6416 .29182 m .61285 .29281 L .61442 .27669 L closepath p .16 .397 .134 r
F P
0 g
s
.6416 .29182 m .61514 .32186 L .61285 .29281 L closepath p .164 .464 .154 r
F P
0 g
s
.61286 .44914 m .61655 .44589 L .61675 .45229 L closepath p .202 .601 .156 r
F P
0 g
s
.61675 .45229 m .61498 .45392 L .61286 .44914 L closepath p .202 .601 .156 r
F P
0 g
s
.61295 .34271 m .61296 .33876 L .61446 .33914 L closepath p .106 .291 .149 r
F P
0 g
s
.61295 .34271 m .61446 .33914 L .61761 .33969 L closepath p .175 .598 .185 r
F P
0 g
s
.61295 .34271 m .61761 .33969 L .6161 .34326 L closepath p .175 .598 .185 r
F P
0 g
s
.61446 .33914 m .61296 .33876 L .61612 .33504 L closepath p .113 .179 .102 r
F P
0 g
s
.61498 .45392 m .61559 .45818 L .6133 .45429 L closepath p .17 .639 .2 r
F P
0 g
s
.6133 .45429 m .61559 .45818 L .61411 .458 L closepath p .058 .407 .208 r
F P
0 g
s
.62728 .349 m .61355 .35718 L .61511 .35261 L closepath p .179 .595 .18 r
F P
0 g
s
.62728 .349 m .61384 .36097 L .61355 .35718 L closepath p .171 .526 .167 r
F P
0 g
s
.61456 .46088 m .61666 .46257 L .61378 .47223 L .18 .84 .235 r
F
.61666 .46257 m .61742 .46319 L .61378 .47223 L F
.61742 .46319 m .61806 .4637 L .61378 .47223 L F
.61806 .4637 m .6183 .4639 L .61378 .47223 L F
.6183 .4639 m .61951 .46499 L .61378 .47223 L F
.61951 .46499 m .61965 .46516 L .61378 .47223 L F
.62053 .4662 m .61378 .47223 L p .61965 .46516 L F P
0 g
s
.61378 .47223 m .62053 .4662 L p .62059 .46627 L .167 .972 .203 r
F P
0 g
s
.62059 .46627 m .62079 .46652 L .61378 .47223 L .167 .972 .203 r
F
.62079 .46652 m .62098 .46675 L .61378 .47223 L F
.62098 .46675 m .62118 .46698 L .61378 .47223 L F
.62207 .46804 m .61378 .47223 L p .62118 .46698 L F P
0 g
s
.61738 .23167 m .6138 .23335 L .61956 .22223 L closepath p 0 0 .034 r
F P
0 g
s
.61572 .22223 m .61956 .22223 L .6138 .23335 L closepath p 0 0 .077 r
F P
0 g
s
.62728 .349 m .6227 .36057 L .61384 .36097 L closepath p .187 .644 .186 r
F P
0 g
s
.62894 .3945 m .61402 .40455 L .61447 .38629 L closepath p .184 .588 .173 r
F P
0 g
s
.61402 .40455 m .62894 .3945 L .63041 .40751 L closepath p .189 .629 .18 r
F P
0 g
s
.61402 .40455 m .63041 .40751 L .6155 .41756 L closepath p .189 .629 .18 r
F P
0 g
s
.61408 .33676 m .61455 .33583 L .61516 .33523 L .13 .298 .129 r
F
.618 .46114 m .61531 .46148 L .61411 .458 L closepath p .061 .52 .228 r
F P
0 g
s
.61411 .458 m .61559 .45818 L .618 .46114 L closepath p .043 .487 .225 r
F P
0 g
s
.64307 .27568 m .61442 .27669 L .62741 .26922 L closepath p .14 .168 .064 r
F P
0 g
s
.6416 .29182 m .61442 .27669 L .64307 .27568 L closepath p .16 .397 .134 r
F P
0 g
s
.61446 .33914 m .61612 .33504 L .61762 .33542 L closepath p .113 .179 .102 r
F P
0 g
s
.61761 .33969 m .61446 .33914 L .61762 .33542 L closepath p .178 .525 .158 r
F P
0 g
s
.61701 .37314 m .62923 .38481 L .61447 .38629 L closepath p .179 .594 .179 r
F P
0 g
s
.62894 .3945 m .61447 .38629 L .62923 .38481 L closepath p .179 .6 .181 r
F P
0 g
s
.61456 .46088 m .61488 .46075 L p .61499 .46094 L .18 .84 .235 r
F P
0 g
s
.61456 .46088 m .61499 .46094 L .61531 .46148 L .18 .84 .235 r
F
.6169 .4572 m .61498 .45392 L .61675 .45229 L closepath p .213 .682 .168 r
F P
0 g
s
.61498 .45392 m .6169 .4572 L .61559 .45818 L closepath p .165 .625 .2 r
F P
0 g
s
.6161 .34326 m .61889 .34845 L .61511 .35261 L closepath p .173 .644 .199 r
F P
0 g
s
.62728 .349 m .61511 .35261 L .61889 .34845 L closepath p .184 .718 .208 r
F P
0 g
s
.61514 .32186 m .63345 .30107 L p .6319 .30634 L .128 .27 .121 r
F P
0 g
s
.6319 .30634 m .62587 .31506 L .61514 .32186 L .128 .27 .121 r
F
.62587 .31506 m .62587 .31507 L .61514 .32186 L F
.62587 .31507 m .62556 .31547 L .61514 .32186 L F
.62172 .31783 m .62329 .31876 L .61514 .32186 L F
.62335 .31874 m .62324 .31925 L .61514 .32186 L F
.62324 .31925 m .6232 .31943 L .61514 .32186 L F
.6232 .31943 m .62312 .31977 L .61514 .32186 L F
.62312 .31977 m .62302 .32007 L .61514 .32186 L F
.623 .32013 m .61514 .32186 L p .62302 .32007 L F P
0 g
s
.61514 .32186 m .62152 .32046 L p .62241 .32194 L .13 .298 .129 r
F P
0 g
s
.62212 .32283 m .61514 .32186 L .62232 .32223 L .13 .298 .129 r
F
.62232 .32223 m .61514 .32186 L .62241 .32194 L F
.61652 .32256 m .61514 .32186 L .62211 .32283 L F
.61514 .32186 m .62049 .32799 L p .61893 .33107 L F P
0 g
s
.61893 .33107 m .6183 .33177 L .61514 .32186 L .13 .298 .129 r
F
.62049 .32799 m .61514 .32186 L p .62093 .32669 L F P
0 g
s
.62122 .32574 m .61514 .32186 L .62125 .32564 L .13 .298 .129 r
F
.62125 .32564 m .61514 .32186 L .62128 .32556 L F
.62128 .32556 m .61514 .32186 L .6214 .32518 L F
.6214 .32518 m .61514 .32186 L .62143 .32506 L F
.62093 .32669 m .61514 .32186 L .62122 .32574 L F
.618 .46114 m .61919 .46461 L .61531 .46148 L closepath p .061 .52 .228 r
F P
0 g
s
.61547 .23589 m .61738 .23167 L p .616 .23604 L .147 .24 .087 r
F P
0 g
s
.63305 .42027 m .6155 .41756 L .63041 .40751 L closepath p .191 .652 .184 r
F P
0 g
s
.63305 .42027 m .6195 .42885 L .6155 .41756 L closepath p .191 .654 .185 r
F P
0 g
s
.61559 .45818 m .6169 .4572 L .61931 .46016 L closepath p .18 .76 .221 r
F P
0 g
s
.61559 .45818 m .61931 .46016 L .618 .46114 L closepath p .18 .76 .221 r
F P
0 g
s
.62186 .206 m .62148 .2111 L .61572 .22223 L closepath p .03 .016 .112 r
F P
0 g
s
.61572 .22223 m .62148 .2111 L .61956 .22223 L closepath p 0 0 .077 r
F P
0 g
s
.62494 .16074 m .6159 .17262 L .61871 .16027 L closepath p .123 .374 .162 r
F P
0 g
s
.62494 .16074 m .62355 .16936 L .6159 .17262 L closepath p .155 .546 .188 r
F P
0 g
s
.61871 .18607 m .6159 .17262 L .62355 .16936 L closepath p .153 .525 .183 r
F P
0 g
s
.61738 .23167 m .61767 .2365 L .616 .23604 L .147 .24 .087 r
F
.6161 .34326 m .62392 .33659 L .61889 .34845 L closepath p .19 .669 .191 r
F P
0 g
s
.6161 .34326 m .61761 .33969 L .62392 .33659 L closepath p .19 .669 .191 r
F P
0 g
s
.61762 .33542 m .61612 .33504 L .61893 .33107 L closepath p .114 .092 .064 r
F P
0 g
s
.61652 .32256 m .62211 .32283 L .62202 .32312 L .13 .298 .129 r
F
.62143 .32506 m .61652 .32256 L .62202 .32312 L F
.6195 .42885 m .62423 .43901 L .61655 .44589 L closepath p .202 .682 .181 r
F P
0 g
s
.62423 .43901 m .62291 .45677 L .61655 .44589 L closepath p .201 .673 .18 r
F P
0 g
s
.61867 .45556 m .61655 .44589 L .62291 .45677 L closepath p .227 .684 .147 r
F P
0 g
s
.61867 .45556 m .61675 .45229 L .61655 .44589 L closepath p .227 .684 .147 r
F P
0 g
s
.6169 .4572 m .61675 .45229 L .61867 .45556 L closepath p .213 .682 .168 r
F P
0 g
s
.61931 .46016 m .6169 .4572 L .61867 .45556 L closepath p .224 .78 .182 r
F P
0 g
s
.63177 .37166 m .61701 .37314 L .6227 .36057 L closepath p .18 .618 .186 r
F P
0 g
s
.61701 .37314 m .63177 .37166 L .62923 .38481 L closepath p .179 .594 .179 r
F P
0 g
s
.61738 .23167 m .61956 .22223 L .63091 .21882 L closepath p .148 .334 .123 r
F P
0 g
s
.61738 .23167 m .63091 .21882 L .62876 .22826 L closepath p .148 .334 .123 r
F P
0 g
s
.61738 .23167 m .62876 .22826 L .63041 .23663 L closepath p .144 .298 .114 r
F P
0 g
s
.61738 .23167 m .63041 .23663 L .6204 .23726 L p .147 .24 .087 r
F P
0 g
s
.61738 .23167 m .6204 .23726 L .61767 .2365 L .147 .24 .087 r
F
.61762 .33542 m .62392 .33659 L .61761 .33969 L closepath p .174 .511 .158 r
F P
0 g
s
.61762 .33542 m .61893 .33107 L .62393 .33233 L closepath p .114 .092 .064 r
F P
0 g
s
.61762 .33542 m .62393 .33233 L .62392 .33659 L closepath p .174 .511 .158 r
F P
0 g
s
.618 .46114 m .61931 .46016 L .62356 .46136 L closepath p .188 .868 .233 r
F P
0 g
s
.618 .46114 m .62356 .46136 L .61919 .46461 L closepath p .188 .868 .233 r
F P
0 g
s
.61919 .46461 m .61951 .46499 L .6183 .4639 L .18 .84 .235 r
F
.61867 .45556 m .62356 .46136 L .61931 .46016 L closepath p .222 .776 .183 r
F P
0 g
s
.61867 .45556 m .62291 .45677 L .62356 .46136 L closepath p .222 .776 .183 r
F P
0 g
s
.61871 .18607 m .62355 .16936 L .62635 .18282 L closepath p .153 .525 .183 r
F P
0 g
s
.62635 .18282 m .62126 .20126 L .61871 .18607 L closepath p .143 .428 .162 r
F P
0 g
s
.61871 .16027 m .62786 .14757 L .62751 .14821 L .063 .07 .107 r
F
.61871 .16027 m .62463 .14935 L .62786 .14757 L p F P
0 g
s
.62655 .14998 m .61871 .16027 L p .62751 .14821 L .063 .07 .107 r
F P
0 g
s
.62285 .15801 m .61871 .16027 L .62653 .15001 L p .063 .07 .107 r
F P
0 g
s
.62494 .16074 m .61871 .16027 L .62285 .15801 L closepath p .116 .502 .203 r
F P
0 g
s
.63231 .33715 m .61889 .34845 L .62392 .33659 L closepath p .189 .665 .19 r
F P
0 g
s
.63231 .33715 m .62728 .349 L .61889 .34845 L closepath p .189 .665 .19 r
F P
0 g
s
.62108 .32681 m .62393 .33233 L .61893 .33107 L closepath p .108 .266 .138 r
F P
0 g
s
.6271 .46557 m .61919 .46461 L .62356 .46136 L closepath p .172 .677 .209 r
F P
0 g
s
.6271 .46557 m .62273 .46882 L .61919 .46461 L closepath p .172 .677 .209 r
F P
0 g
s
.61921 .23895 m .63197 .24137 L .64337 .24563 L closepath p .077 0 0 r
F P
0 g
s
.61921 .23895 m .65062 .25131 L .63244 .24631 L closepath p .138 0 0 r
F P
0 g
s
.61921 .23895 m .64337 .24563 L .65062 .25131 L closepath p .138 0 0 r
F P
0 g
s
.6195 .42885 m .63305 .42027 L .63776 .43043 L closepath p .193 .667 .187 r
F P
0 g
s
.6195 .42885 m .63776 .43043 L .62423 .43901 L closepath p .193 .667 .187 r
F P
0 g
s
.62148 .2111 m .63091 .21882 L .61956 .22223 L closepath p .158 .419 .144 r
F P
0 g
s
.6204 .23726 m .62036 .23726 L p .62019 .2372 L .147 .24 .087 r
F P
0 g
s
.62036 .23726 m .63041 .23663 L p .62199 .23784 L .119 .012 .018 r
F P
0 g
s
.62049 .32799 m .62093 .32669 L .62061 .32775 L .13 .298 .129 r
F
.62061 .32775 m .62207 .32297 L .62108 .32681 L p 0 0 0 r
F P
0 g
s
.62259 .32224 m .62108 .32681 L .62207 .32297 L closepath p 0 0 0 r
F P
0 g
s
.62259 .32224 m .62608 .32806 L .62108 .32681 L closepath p .089 .405 .194 r
F P
0 g
s
.62108 .32681 m .62608 .32806 L .62393 .33233 L closepath p .108 .266 .138 r
F P
0 g
s
.62635 .18282 m .6321 .19574 L .62126 .20126 L closepath p .191 .626 .177 r
F P
0 g
s
.62186 .206 m .62126 .20126 L .6321 .19574 L closepath p .15 .3 .108 r
F P
0 g
s
.62148 .2111 m .62186 .206 L .63218 .20554 L closepath p .165 .463 .152 r
F P
0 g
s
.62148 .2111 m .63218 .20554 L .63181 .21064 L closepath p .165 .463 .152 r
F P
0 g
s
.62148 .2111 m .63181 .21064 L .63091 .21882 L closepath p .162 .414 .137 r
F P
0 g
s
.62152 .32046 m .623 .32013 L p .62262 .32128 L .13 .298 .129 r
F P
0 g
s
.62255 .32151 m .62152 .32046 L .62262 .32128 L .13 .298 .129 r
F
.62241 .32194 m .62152 .32046 L .6225 .32167 L F
.62251 .32163 m .62152 .32046 L .62251 .32163 L F
.62251 .32163 m .62152 .32046 L .62251 .32163 L F
.62251 .32163 m .62152 .32046 L .62255 .32151 L F
.6225 .32167 m .62152 .32046 L .62251 .32163 L F
.62556 .31547 m .62378 .31782 L .62172 .31783 L .128 .27 .121 r
F
.62378 .31782 m .62345 .31825 L .62172 .31783 L F
.62345 .31825 m .62329 .31876 L .62172 .31783 L F
.62186 .206 m .6321 .19574 L .63273 .20046 L closepath p .15 .3 .108 r
F P
0 g
s
.63273 .20046 m .63218 .20554 L .62186 .206 L closepath p .158 .355 .12 r
F P
0 g
s
.62241 .46844 m .62198 .46808 L .62207 .46804 L .135 .914 .189 r
F
.62198 .46808 m .62207 .46804 L p .62207 .46804 L F P
0 g
s
.62259 .46866 m .62198 .46808 L .62241 .46844 L .135 .914 .189 r
F
.62273 .46882 m .62198 .46808 L .62259 .46866 L F
.63041 .23663 m .62411 .23859 L p .62199 .23784 L .119 .012 .018 r
F P
0 g
s
.62259 .32224 m .62207 .32297 L .62312 .31977 L p 0 0 0 r
F P
0 g
s
.62207 .32297 m .62211 .32283 L .62209 .3229 L .13 .298 .129 r
F
.62209 .3229 m .62211 .32283 L .62212 .32283 L F
.62212 .32283 m .62211 .32283 L .62212 .32283 L F
.62345 .31825 m .62409 .32261 L .62259 .32224 L closepath p .099 .35 .172 r
F P
0 g
s
.62259 .32224 m .62409 .32261 L .62608 .32806 L closepath p .089 .405 .194 r
F P
0 g
s
.62728 .349 m .63551 .35959 L .6227 .36057 L closepath p .185 .631 .185 r
F P
0 g
s
.63177 .37166 m .6227 .36057 L .63551 .35959 L closepath p .185 .621 .182 r
F P
0 g
s
.62994 .47279 m .62273 .46882 L .6271 .46557 L closepath p .156 .549 .187 r
F P
0 g
s
.63133 .47733 m .62273 .46882 L .62862 .47377 L closepath p .017 .392 .213 r
F P
0 g
s
.63133 .47733 m .62544 .47239 L .62273 .46882 L closepath p .017 .392 .213 r
F P
0 g
s
.62994 .47279 m .62862 .47377 L .62273 .46882 L closepath p .156 .549 .187 r
F P
0 g
s
.62861 .1528 m .62285 .15801 L .62651 .15006 L closepath p .149 .549 .193 r
F P
0 g
s
.62861 .1528 m .62494 .16074 L .62285 .15801 L closepath p .149 .549 .193 r
F P
0 g
s
.62423 .43901 m .63058 .44989 L .62291 .45677 L closepath p .201 .673 .18 r
F P
0 g
s
.63058 .44989 m .62356 .46136 L .62291 .45677 L closepath p .21 .746 .19 r
F P
0 g
s
.62336 .31868 m .62329 .31876 L .62345 .31825 L .128 .27 .121 r
F
.62336 .31868 m .62335 .31874 L .62329 .31876 L F
.62649 .31589 m .62345 .31825 L .62587 .31506 L closepath p .127 .351 .151 r
F P
0 g
s
.62649 .31589 m .62496 .31862 L .62345 .31825 L closepath p .11 .241 .128 r
F P
0 g
s
.62345 .31825 m .62496 .31862 L .62409 .32261 L closepath p .099 .35 .172 r
F P
0 g
s
.63502 .16568 m .62355 .16936 L .62494 .16074 L closepath p .176 .626 .192 r
F P
0 g
s
.62355 .16936 m .63502 .16568 L .62635 .18282 L closepath p .173 .578 .181 r
F P
0 g
s
.63058 .44989 m .6271 .46557 L .62356 .46136 L closepath p .202 .697 .186 r
F P
0 g
s
.63643 .32717 m .62392 .33659 L .62393 .33233 L closepath p .173 .507 .158 r
F P
0 g
s
.63643 .32717 m .63231 .33715 L .62392 .33659 L closepath p .19 .636 .181 r
F P
0 g
s
.63643 .32717 m .62393 .33233 L .62608 .32806 L closepath p .183 .615 .182 r
F P
0 g
s
.62409 .32261 m .62496 .31862 L .62738 .32006 L closepath p .163 .604 .197 r
F P
0 g
s
.62738 .32006 m .62608 .32806 L .62409 .32261 L closepath p .16 .576 .192 r
F P
0 g
s
.62411 .23859 m .62433 .23852 L p .62768 .23985 L .119 .012 .018 r
F P
0 g
s
.62411 .23859 m .62768 .23985 L .62767 .23985 L .119 .012 .018 r
F
.64366 .43616 m .62423 .43901 L .63776 .43043 L closepath p .197 .715 .195 r
F P
0 g
s
.64366 .43616 m .63058 .44989 L .62423 .43901 L closepath p .197 .668 .183 r
F P
0 g
s
.62433 .23852 m .63008 .23673 L p .62902 .23991 L .119 .012 .018 r
F P
0 g
s
.62768 .23985 m .62433 .23852 L .62792 .23986 L .119 .012 .018 r
F
.62433 .23852 m .62902 .23991 L .62792 .23986 L F
.62861 .1528 m .63643 .15706 L .62494 .16074 L closepath p .179 .689 .206 r
F P
0 g
s
.63502 .16568 m .62494 .16074 L .63643 .15706 L closepath p .176 .626 .192 r
F P
0 g
s
.62496 .31862 m .62649 .31589 L .62891 .31733 L closepath p .18 .575 .173 r
F P
0 g
s
.62496 .31862 m .62891 .31733 L .62738 .32006 L closepath p .18 .575 .173 r
F P
0 g
s
.6319 .30634 m .62588 .31506 L .62587 .31506 L .128 .27 .121 r
F
.62649 .31589 m .62587 .31506 L .62928 .31265 L closepath p .159 .379 .128 r
F P
0 g
s
.6319 .30634 m .62589 .31505 L .62588 .31506 L .128 .27 .121 r
F
.6319 .30634 m .63042 .31133 L .62589 .31505 L F
.63042 .31133 m .62928 .31265 L .62589 .31505 L F
.62738 .32006 m .62937 .32551 L .62608 .32806 L closepath p .16 .576 .192 r
F P
0 g
s
.63643 .32717 m .62608 .32806 L .62937 .32551 L closepath p .18 .744 .218 r
F P
0 g
s
.62635 .18282 m .63502 .16568 L .64478 .16627 L closepath p .192 .673 .19 r
F P
0 g
s
.62635 .18282 m .64478 .16627 L .63611 .1834 L closepath p .192 .673 .19 r
F P
0 g
s
.62635 .18282 m .63611 .1834 L .6321 .19574 L closepath p .193 .629 .176 r
F P
0 g
s
.62649 .31589 m .62928 .31265 L .6299 .31348 L closepath p .159 .379 .128 r
F P
0 g
s
.62891 .31733 m .62649 .31589 L .6299 .31348 L closepath p .185 .559 .162 r
F P
0 g
s
.62651 .15006 m .63302 .1414 L .63 .14697 L closepath p .113 .29 .143 r
F P
0 g
s
.62651 .15006 m .62952 .1445 L .63302 .1414 L closepath p .113 .29 .143 r
F P
0 g
s
.63 .14697 m .62861 .1528 L .62651 .15006 L closepath p .154 .554 .19 r
F P
0 g
s
.62653 .15001 m .62654 .15001 L .62651 .15006 L .063 .07 .107 r
F
.62653 .15001 m .62655 .14998 L p .62654 .15001 L F P
0 g
s
.62681 .25772 m .63924 .2529 L .65262 .26043 L closepath p .155 .149 .027 r
F P
0 g
s
.62681 .25772 m .65262 .26043 L .64022 .26533 L closepath p .155 .149 .027 r
F P
0 g
s
.63058 .44989 m .6325 .4678 L .6271 .46557 L closepath p .192 .659 .185 r
F P
0 g
s
.62994 .47279 m .6271 .46557 L .6325 .4678 L closepath p .197 .631 .171 r
F P
0 g
s
.64393 .33824 m .62728 .349 L .63231 .33715 L closepath p .192 .68 .192 r
F P
0 g
s
.64393 .33824 m .64009 .34802 L .62728 .349 L closepath p .184 .61 .179 r
F P
0 g
s
.62728 .349 m .64009 .34802 L .63551 .35959 L closepath p .185 .631 .185 r
F P
0 g
s
.62738 .32006 m .63446 .31643 L .62937 .32551 L closepath p .194 .648 .18 r
F P
0 g
s
.62738 .32006 m .62891 .31733 L .63446 .31643 L closepath p .194 .648 .18 r
F P
0 g
s
.64022 .26533 m .64307 .27568 L .62741 .26922 L closepath p .14 .109 .035 r
F P
0 g
s
.62767 .23985 m .62792 .23986 L p .62787 .23992 L .163 .103 0 r
F P
0 g
s
.62792 .23986 m .62789 .23993 L .62787 .23992 L .163 .103 0 r
F
.62792 .23986 m .62791 .23993 L .62789 .23993 L F
.62792 .23986 m .62793 .23994 L .62791 .23993 L F
.62792 .23986 m .63208 .24003 L .63638 .24302 L p F P
0 g
s
.63197 .24137 m .62903 .24033 L .62792 .23986 L .163 .103 0 r
F
.62792 .23986 m .62903 .24033 L .62793 .23994 L F
.63444 .15093 m .62861 .1528 L .63 .14697 L closepath p .177 .645 .196 r
F P
0 g
s
.62861 .1528 m .63444 .15093 L .63643 .15706 L closepath p .179 .689 .206 r
F P
0 g
s
.634 .47717 m .62862 .47377 L .62994 .47279 L closepath p .164 .616 .199 r
F P
0 g
s
.62862 .47377 m .63268 .47814 L .63133 .47733 L closepath p 0 .271 .194 r
F P
0 g
s
.634 .47717 m .63268 .47814 L .62862 .47377 L closepath p .164 .616 .199 r
F P
0 g
s
.63091 .21882 m .64489 .23271 L .62876 .22826 L closepath p .198 .528 .135 r
F P
0 g
s
.64659 .24107 m .63041 .23663 L .62876 .22826 L closepath p .195 .434 .1 r
F P
0 g
s
.64659 .24107 m .62876 .22826 L .64489 .23271 L closepath p .195 .434 .1 r
F P
0 g
s
.6299 .31348 m .63446 .31643 L .62891 .31733 L closepath p .191 .581 .163 r
F P
0 g
s
.62894 .3945 m .62923 .38481 L .64568 .37268 L closepath p .179 .6 .181 r
F P
0 g
s
.62894 .3945 m .64568 .37268 L .64539 .38238 L closepath p .179 .6 .181 r
F P
0 g
s
.62894 .3945 m .64539 .38238 L .64795 .39972 L closepath p .185 .649 .189 r
F P
0 g
s
.62894 .3945 m .64795 .39972 L .63041 .40751 L closepath p .188 .628 .18 r
F P
0 g
s
.62902 .23991 m .63008 .23673 L .63208 .24003 L .119 .012 .018 r
F
.63177 .37166 m .64568 .37268 L .62923 .38481 L closepath p .178 .589 .179 r
F P
0 g
s
.6299 .31348 m .62928 .31265 L .63339 .30983 L closepath p .175 .386 .111 r
F P
0 g
s
.63042 .31133 m .62974 .31234 L .62928 .31265 L .128 .27 .121 r
F
.64151 .3181 m .62937 .32551 L .63446 .31643 L closepath p .196 .659 .181 r
F P
0 g
s
.64151 .3181 m .63643 .32717 L .62937 .32551 L closepath p .196 .659 .181 r
F P
0 g
s
.63042 .31133 m .63022 .31201 L .62974 .31234 L .128 .27 .121 r
F
.6299 .31348 m .63339 .30983 L .63545 .31258 L closepath p .175 .386 .111 r
F P
0 g
s
.6299 .31348 m .63545 .31258 L .63446 .31643 L closepath p .191 .581 .163 r
F P
0 g
s
.62994 .47279 m .6325 .4678 L .63534 .47501 L closepath p .197 .631 .171 r
F P
0 g
s
.63534 .47501 m .634 .47717 L .62994 .47279 L closepath p .198 .625 .168 r
F P
0 g
s
.63585 .14509 m .63 .14697 L .63302 .1414 L closepath p .172 .573 .18 r
F P
0 g
s
.63444 .15093 m .63 .14697 L .63585 .14509 L closepath p .177 .645 .196 r
F P
0 g
s
.63008 .23673 m .63041 .23663 L .63208 .24003 L p .119 .012 .018 r
F P
0 g
s
.63123 .31131 m .63022 .31201 L .63345 .30107 L .128 .27 .121 r
F
.65212 .24908 m .63208 .24003 L .63041 .23663 L closepath p .167 .129 0 r
F P
0 g
s
.65212 .24908 m .63041 .23663 L .64659 .24107 L closepath p .181 .265 .04 r
F P
0 g
s
.63305 .42027 m .63041 .40751 L .64795 .39972 L closepath p .19 .649 .185 r
F P
0 g
s
.63058 .44989 m .64366 .43616 L .65105 .44445 L closepath p .203 .716 .189 r
F P
0 g
s
.63058 .44989 m .65105 .44445 L .638 .45816 L closepath p .203 .716 .189 r
F P
0 g
s
.63058 .44989 m .638 .45816 L .6325 .4678 L closepath p .212 .713 .179 r
F P
0 g
s
.63181 .21064 m .64322 .22011 L .63091 .21882 L closepath p .193 .528 .141 r
F P
0 g
s
.63091 .21882 m .64322 .22011 L .64489 .23271 L closepath p .194 .534 .143 r
F P
0 g
s
.63174 .31097 m .63123 .31131 L .63345 .30107 L .128 .27 .121 r
F
.63268 .47814 m .63513 .47965 L .63133 .47733 L closepath p 0 .201 .153 r
F P
0 g
s
.63178 .31094 m .63174 .31097 L .63345 .30107 L .128 .27 .121 r
F
.63177 .37166 m .63551 .35959 L .64942 .36061 L closepath p .178 .593 .18 r
F P
0 g
s
.63177 .37166 m .64942 .36061 L .64568 .37268 L closepath p .178 .593 .18 r
F P
0 g
s
.6416 .29182 m .63339 .30983 L .63178 .31094 L .128 .27 .121 r
F
.63345 .30107 m .6416 .29182 L p .63178 .31094 L F P
0 g
s
.64681 .20634 m .63181 .21064 L .63218 .20554 L closepath p .2 .587 .153 r
F P
0 g
s
.64681 .20634 m .6441 .21194 L .63181 .21064 L closepath p .193 .51 .135 r
F P
0 g
s
.63181 .21064 m .6441 .21194 L .64322 .22011 L closepath p .193 .528 .141 r
F P
0 g
s
.63208 .24003 m .64161 .24497 L .64081 .24467 L .163 .103 0 r
F
.63208 .24003 m .6408 .24397 L p .64161 .24497 L F P
0 g
s
.63638 .24302 m .63208 .24003 L p .64081 .24467 L .163 .103 0 r
F P
0 g
s
.63611 .1834 m .64701 .19639 L .6321 .19574 L closepath p .201 .666 .177 r
F P
0 g
s
.64734 .20126 m .6321 .19574 L .64701 .19639 L closepath p .19 .412 .099 r
F P
0 g
s
.64734 .20126 m .63273 .20046 L .6321 .19574 L closepath p .19 .416 .1 r
F P
0 g
s
.63273 .20046 m .64681 .20634 L .63218 .20554 L closepath p .196 .492 .124 r
F P
0 g
s
.63643 .32717 m .64393 .33824 L .63231 .33715 L closepath p .193 .653 .182 r
F P
0 g
s
.63924 .2529 m .63244 .24631 L .65062 .25131 L closepath p .177 .188 0 r
F P
0 g
s
.638 .45816 m .6447 .47784 L .6325 .4678 L closepath p .208 .69 .177 r
F P
0 g
s
.63901 .47803 m .6325 .4678 L .6447 .47784 L closepath p .218 .663 .155 r
F P
0 g
s
.63901 .47803 m .63534 .47501 L .6325 .4678 L closepath p .218 .663 .155 r
F P
0 g
s
.634 .47717 m .63649 .48047 L .63268 .47814 L closepath p .177 .732 .217 r
F P
0 g
s
.63268 .47814 m .63649 .48047 L .63513 .47965 L closepath p 0 .201 .153 r
F P
0 g
s
.63273 .20046 m .64734 .20126 L .64681 .20634 L closepath p .196 .492 .124 r
F P
0 g
s
.63452 .14038 m .63675 .14142 L .63302 .1414 L closepath p .156 .232 .07 r
F P
0 g
s
.63585 .14509 m .63302 .1414 L .63675 .14142 L closepath p .177 .577 .176 r
F P
0 g
s
.63305 .42027 m .64795 .39972 L .65057 .41247 L closepath p .19 .649 .185 r
F P
0 g
s
.63305 .42027 m .65057 .41247 L .63776 .43043 L closepath p .191 .665 .188 r
F P
0 g
s
.6416 .29182 m .63514 .30856 L p .63339 .30983 L .128 .27 .121 r
F P
0 g
s
.63973 .30913 m .63545 .31258 L .63339 .30983 L closepath p .189 .386 .089 r
F P
0 g
s
.63514 .30856 m .63346 .30982 L .63339 .30983 L .128 .27 .121 r
F
.63514 .30856 m .63372 .3098 L .63346 .30982 L F
.63514 .30856 m .63471 .30969 L p .63372 .3098 L F P
0 g
s
.63767 .48019 m .634 .47717 L .63534 .47501 L closepath p .209 .688 .175 r
F P
0 g
s
.634 .47717 m .63767 .48019 L .63649 .48047 L closepath p .149 .69 .227 r
F P
0 g
s
.63945 .14681 m .63444 .15093 L .63585 .14509 L closepath p .19 .699 .198 r
F P
0 g
s
.63945 .14681 m .63643 .15706 L .63444 .15093 L closepath p .192 .719 .201 r
F P
0 g
s
.64179 .31188 m .64151 .3181 L .63446 .31643 L closepath p .2 .571 .147 r
F P
0 g
s
.64179 .31188 m .63446 .31643 L .63545 .31258 L closepath p .212 .67 .166 r
F P
0 g
s
.63452 .14038 m .63825 .1404 L .63675 .14142 L closepath p .156 .232 .07 r
F P
0 g
s
.63471 .30969 m .6416 .29182 L p .63592 .30955 L .129 .276 .122 r
F P
0 g
s
.64142 .15295 m .64478 .16627 L .63502 .16568 L closepath p .191 .689 .195 r
F P
0 g
s
.64142 .15295 m .63502 .16568 L .63643 .15706 L closepath p .186 .665 .193 r
F P
0 g
s
.63767 .48019 m .63534 .47501 L .63901 .47803 L closepath p .209 .688 .175 r
F P
0 g
s
.63973 .30913 m .64179 .31188 L .63545 .31258 L closepath p .189 .386 .089 r
F P
0 g
s
.65399 .34244 m .63551 .35959 L .64009 .34802 L closepath p .189 .649 .186 r
F P
0 g
s
.65399 .34244 m .64942 .36061 L .63551 .35959 L closepath p .178 .577 .175 r
F P
0 g
s
.64035 .14314 m .63585 .14509 L .63675 .14142 L closepath p .2 .669 .18 r
F P
0 g
s
.64035 .14314 m .63945 .14681 L .63585 .14509 L closepath p .2 .669 .18 r
F P
0 g
s
.63636 .30951 m .63592 .30955 L .6416 .29182 L .129 .276 .122 r
F
.651 .18408 m .63611 .1834 L .64478 .16627 L closepath p .199 .714 .193 r
F P
0 g
s
.63611 .1834 m .651 .18408 L .64701 .19639 L closepath p .201 .666 .177 r
F P
0 g
s
.63654 .30949 m .63636 .30951 L .6416 .29182 L .129 .276 .122 r
F
.63945 .14681 m .64142 .15295 L .63643 .15706 L closepath p .192 .719 .201 r
F P
0 g
s
.63643 .32717 m .64151 .3181 L .65053 .31912 L closepath p .201 .685 .184 r
F P
0 g
s
.63643 .32717 m .65053 .31912 L .64804 .32827 L closepath p .194 .621 .171 r
F P
0 g
s
.63643 .32717 m .64804 .32827 L .64393 .33824 L closepath p .193 .653 .182 r
F P
0 g
s
.63649 .48047 m .63767 .48019 L .6417 .48159 L closepath p .135 .826 .25 r
F P
0 g
s
.63649 .48047 m .6417 .48159 L .64052 .48186 L closepath p .135 .826 .25 r
F P
0 g
s
.63676 .30946 m .63654 .30949 L .6416 .29182 L .129 .276 .122 r
F
.63675 .14142 m .63825 .1404 L .64035 .14314 L closepath p .21 .582 .136 r
F P
0 g
s
.63722 .30941 m .63676 .30946 L .6416 .29182 L .129 .276 .122 r
F
.63825 .3093 m .63722 .30941 L .6416 .29182 L F
.6417 .48159 m .63767 .48019 L .63901 .47803 L closepath p .221 .76 .181 r
F P
0 g
s
.64366 .43616 m .63776 .43043 L .65057 .41247 L closepath p .199 .715 .193 r
F P
0 g
s
.6562 .45089 m .638 .45816 L .65105 .44445 L closepath p .214 .806 .2 r
F P
0 g
s
.6562 .45089 m .65017 .46823 L .638 .45816 L closepath p .209 .688 .175 r
F P
0 g
s
.638 .45816 m .65017 .46823 L .6447 .47784 L closepath p .208 .69 .177 r
F P
0 g
s
.63825 .1404 m .6411 .14299 L .64319 .14572 L closepath p .209 .375 0 r
F P
0 g
s
.63825 .1404 m .64319 .14572 L .64035 .14314 L closepath p .209 .376 0 r
F P
0 g
s
.63849 .30927 m .63825 .3093 L .6416 .29182 L .129 .276 .122 r
F
.6416 .29182 m .65548 .28767 L .63849 .30927 L p F P
0 g
s
.63849 .30927 m .63858 .30915 L p .63926 .30919 L .129 .276 .122 r
F P
0 g
s
.63858 .30915 m .65548 .28767 L p .64532 .30904 L .129 .276 .122 r
F P
0 g
s
.63973 .30913 m .63926 .30919 L .63858 .30915 L .129 .276 .122 r
F
.63901 .47803 m .6447 .47784 L .64737 .48139 L closepath p .219 .76 .183 r
F P
0 g
s
.63901 .47803 m .64737 .48139 L .6417 .48159 L closepath p .219 .76 .183 r
F P
0 g
s
.63924 .2529 m .65062 .25131 L .65745 .25789 L closepath p .177 .188 0 r
F P
0 g
s
.63924 .2529 m .65745 .25789 L .66628 .26416 L closepath p .135 0 0 r
F P
0 g
s
.63924 .2529 m .66628 .26416 L .65262 .26043 L closepath p .135 0 0 r
F P
0 g
s
.63945 .14681 m .64035 .14314 L .64319 .14572 L closepath p .225 .778 .18 r
F P
0 g
s
.63945 .14681 m .64319 .14572 L .64142 .15295 L closepath p .225 .778 .18 r
F P
0 g
s
.64179 .31188 m .63973 .30913 L .6464 .30902 L closepath p .201 .371 .055 r
F P
0 g
s
.65783 .33266 m .64009 .34802 L .64393 .33824 L closepath p .186 .619 .18 r
F P
0 g
s
.65783 .33266 m .65399 .34244 L .64009 .34802 L closepath p .186 .619 .18 r
F P
0 g
s
.65262 .26043 m .6559 .27175 L .64022 .26533 L closepath p .137 .049 .006 r
F P
0 g
s
.64022 .26533 m .6559 .27175 L .64307 .27568 L closepath p .14 .109 .035 r
F P
0 g
s
.64052 .48186 m .6417 .48159 L .64737 .48139 L closepath p .093 .831 .229 r
F P
0 g
s
.64052 .48186 m .64737 .48139 L .64346 .48231 L closepath p .093 .831 .229 r
F P
0 g
s
.6408 .24397 m .65212 .24908 L .64898 .24897 L p .163 .103 0 r
F P
0 g
s
.6408 .24397 m .64898 .24897 L .64337 .24563 L .163 .103 0 r
F
.6408 .24397 m .64337 .24563 L .64161 .24497 L F
.6411 .14299 m .64645 .14588 L .64319 .14572 L closepath p .186 .582 .168 r
F P
0 g
s
.6411 .14299 m .64947 .14358 L .64645 .14588 L closepath p .188 .577 .165 r
F P
0 g
s
.64142 .15295 m .64319 .14572 L .64645 .14588 L closepath p .186 .612 .178 r
F P
0 g
s
.64142 .15295 m .64645 .14588 L .64468 .1531 L closepath p .186 .612 .178 r
F P
0 g
s
.64142 .15295 m .64623 .15616 L .64478 .16627 L closepath p .179 .66 .199 r
F P
0 g
s
.64142 .15295 m .64468 .1531 L .64623 .15616 L closepath p .185 .653 .191 r
F P
0 g
s
.64179 .31188 m .65053 .31912 L .64151 .3181 L closepath p .202 .577 .147 r
F P
0 g
s
.67187 .26991 m .6416 .29182 L .64307 .27568 L closepath p .153 .374 .133 r
F P
0 g
s
.65548 .28767 m .6416 .29182 L .66149 .27743 L p .109 .113 .079 r
F P
0 g
s
.64179 .31188 m .6464 .30902 L .6508 .31291 L closepath p .165 .168 .018 r
F P
0 g
s
.64179 .31188 m .6508 .31291 L .65053 .31912 L closepath p .202 .577 .147 r
F P
0 g
s
.67187 .26991 m .64307 .27568 L .6559 .27175 L closepath p .025 0 0 r
F P
0 g
s
.65333 .21525 m .64322 .22011 L .6441 .21194 L closepath p .216 .615 .14 r
F P
0 g
s
.65333 .21525 m .65247 .22663 L .64322 .22011 L closepath p .221 .668 .151 r
F P
0 g
s
.64489 .23271 m .64322 .22011 L .65247 .22663 L closepath p .226 .626 .127 r
F P
0 g
s
.64337 .24563 m .6638 .25779 L .67112 .26351 L closepath p .123 0 0 r
F P
0 g
s
.64337 .24563 m .67112 .26351 L .65062 .25131 L closepath p .123 0 0 r
F P
0 g
s
.65197 .48431 m .64346 .48231 L .64737 .48139 L closepath p .146 .758 .24 r
F P
0 g
s
.65197 .48431 m .64806 .48522 L .64346 .48231 L closepath p .146 .758 .24 r
F P
0 g
s
.64366 .43616 m .65057 .41247 L .65645 .41823 L closepath p .199 .715 .193 r
F P
0 g
s
.64366 .43616 m .65645 .41823 L .66717 .42902 L closepath p .206 .758 .197 r
F P
0 g
s
.64366 .43616 m .66717 .42902 L .65105 .44445 L closepath p .205 .716 .187 r
F P
0 g
s
.65783 .33266 m .64393 .33824 L .64804 .32827 L closepath p .186 .619 .18 r
F P
0 g
s
.6441 .21194 m .64681 .20634 L .65602 .20965 L closepath p .216 .63 .145 r
F P
0 g
s
.6441 .21194 m .65602 .20965 L .65333 .21525 L closepath p .216 .63 .145 r
F P
0 g
s
.65012 .1436 m .64468 .1531 L .64645 .14588 L closepath p .123 .36 .157 r
F P
0 g
s
.65012 .1436 m .64923 .14727 L .64468 .1531 L closepath p .123 .36 .157 r
F P
0 g
s
.65077 .15034 m .64468 .1531 L .64923 .14727 L closepath p .171 .63 .197 r
F P
0 g
s
.65077 .15034 m .64623 .15616 L .64468 .1531 L closepath p .171 .63 .197 r
F P
0 g
s
.65017 .46823 m .64737 .48139 L .6447 .47784 L closepath p .219 .76 .182 r
F P
0 g
s
.65064 .16716 m .64478 .16627 L .64623 .15616 L closepath p .197 .729 .199 r
F P
0 g
s
.651 .18408 m .64478 .16627 L .65064 .16716 L closepath p .199 .713 .193 r
F P
0 g
s
.64489 .23271 m .65247 .22663 L .65416 .23921 L closepath p .226 .626 .127 r
F P
0 g
s
.64489 .23271 m .65416 .23921 L .65803 .25144 L closepath p .225 .53 .082 r
F P
0 g
s
.64489 .23271 m .65803 .25144 L .64659 .24107 L closepath p .225 .523 .075 r
F P
0 g
s
.65548 .28767 m .6464 .30902 L .64532 .30904 L .129 .276 .122 r
F
.66732 .35558 m .66509 .37015 L .64539 .38238 L closepath p .166 .527 .172 r
F P
0 g
s
.66732 .35558 m .64539 .38238 L .64568 .37268 L closepath p .177 .593 .181 r
F P
0 g
s
.66762 .38752 m .64539 .38238 L .66509 .37015 L closepath p .178 .628 .191 r
F P
0 g
s
.66762 .38752 m .64795 .39972 L .64539 .38238 L closepath p .178 .628 .191 r
F P
0 g
s
.66732 .35558 m .64568 .37268 L .64942 .36061 L closepath p .177 .587 .18 r
F P
0 g
s
.65077 .15034 m .65209 .15706 L .64623 .15616 L closepath p .192 .761 .212 r
F P
0 g
s
.65064 .16716 m .64623 .15616 L .65209 .15706 L closepath p .197 .729 .199 r
F P
0 g
s
.65548 .28767 m .64744 .30866 L p .6464 .30902 L .129 .276 .122 r
F P
0 g
s
.65561 .30873 m .6508 .31291 L .6464 .30902 L closepath p .161 .19 .039 r
F P
0 g
s
.64744 .30866 m .64682 .30901 L .6464 .30902 L .129 .276 .122 r
F
.65012 .1436 m .64645 .14588 L .64947 .14358 L closepath p 0 0 0 r
F P
0 g
s
.65212 .24908 m .64659 .24107 L .65803 .25144 L closepath p .184 .226 0 r
F P
0 g
s
.64734 .20126 m .65602 .20965 L .64681 .20634 L closepath p .216 .572 .123 r
F P
0 g
s
.64744 .30866 m .64731 .30899 L p .64682 .30901 L .129 .276 .122 r
F P
0 g
s
.6585 .18533 m .64701 .19639 L .651 .18408 L closepath p .208 .696 .179 r
F P
0 g
s
.6585 .18533 m .65631 .19891 L .64701 .19639 L closepath p .211 .72 .182 r
F P
0 g
s
.64734 .20126 m .64701 .19639 L .65631 .19891 L closepath p .21 .481 .093 r
F P
0 g
s
.64731 .30899 m .65548 .28767 L p .64789 .30897 L .003 0 .052 r
F P
0 g
s
.64734 .20126 m .65631 .19891 L .65666 .20378 L closepath p .21 .481 .093 r
F P
0 g
s
.64734 .20126 m .65666 .20378 L .65602 .20965 L closepath p .215 .578 .127 r
F P
0 g
s
.65017 .46823 m .65197 .48431 L .64737 .48139 L closepath p .215 .741 .183 r
F P
0 g
s
.64789 .30897 m .65548 .28767 L .64835 .30896 L .003 0 .052 r
F
.65645 .41823 m .64795 .39972 L .66762 .38752 L closepath p .181 .655 .195 r
F P
0 g
s
.65645 .41823 m .65057 .41247 L .64795 .39972 L closepath p .213 .705 .174 r
F P
0 g
s
.66031 .32351 m .64804 .32827 L .65053 .31912 L closepath p .19 .602 .17 r
F P
0 g
s
.66031 .32351 m .65783 .33266 L .64804 .32827 L closepath p .19 .602 .17 r
F P
0 g
s
.6555 .49034 m .64806 .48522 L .65197 .48431 L closepath p .149 .632 .214 r
F P
0 g
s
.6555 .49034 m .65432 .49061 L .64806 .48522 L closepath p .149 .632 .214 r
F P
0 g
s
.65548 .28767 m .65828 .29188 L .64835 .30896 L p .003 0 .052 r
F P
0 g
s
.64835 .30896 m .65828 .29188 L p .64891 .30894 L 0 0 .048 r
F P
0 g
s
.64929 .30893 m .64891 .30894 L .65828 .29188 L 0 0 .048 r
F
.64898 .24897 m .65212 .24908 L p .65451 .25205 L .226 .527 .078 r
F P
0 g
s
.653 .25133 m .64898 .24897 L .65451 .25205 L .226 .527 .078 r
F
.6528 .25124 m .64898 .24897 L .653 .25133 L F
.64923 .14727 m .65372 .14183 L .65283 .1455 L closepath p .165 .528 .173 r
F P
0 g
s
.64923 .14727 m .65012 .1436 L .65372 .14183 L closepath p .165 .528 .173 r
F P
0 g
s
.65283 .1455 m .65077 .15034 L .64923 .14727 L closepath p .176 .639 .195 r
F P
0 g
s
.65828 .29188 m .65899 .29611 L .64929 .30893 L p 0 0 .048 r
F P
0 g
s
.64929 .30893 m .65899 .29611 L .65158 .30886 L p 0 0 .048 r
F P
0 g
s
.65399 .34244 m .66732 .35558 L .64942 .36061 L closepath p .175 .563 .174 r
F P
0 g
s
.65012 .1436 m .64947 .14358 L .65315 .1413 L closepath p 0 0 0 r
F P
0 g
s
.65372 .14183 m .65012 .1436 L .65315 .1413 L closepath p .092 .006 .053 r
F P
0 g
s
.65017 .46823 m .6562 .45089 L .66682 .45677 L closepath p .217 .725 .176 r
F P
0 g
s
.65017 .46823 m .66682 .45677 L .66083 .47407 L closepath p .217 .725 .176 r
F P
0 g
s
.65017 .46823 m .65765 .48566 L .65197 .48431 L closepath p .202 .706 .188 r
F P
0 g
s
.65017 .46823 m .66083 .47407 L .65765 .48566 L closepath p .216 .727 .177 r
F P
0 g
s
.66031 .32351 m .65053 .31912 L .6508 .31291 L closepath p .196 .556 .148 r
F P
0 g
s
.65511 .25563 m .65062 .25131 L .66687 .26442 L p .155 .073 0 r
F P
0 g
s
.66918 .26544 m .65062 .25131 L .67164 .2662 L .181 .192 0 r
F
.66915 .26543 m .65062 .25131 L .66918 .26544 L F
.66915 .26543 m .65062 .25131 L .66915 .26543 L F
.66687 .26442 m .65062 .25131 L p .66915 .26543 L F P
0 g
s
.67164 .2662 m .65062 .25131 L .67241 .26611 L .181 .192 0 r
F
.65062 .25131 m .67112 .26351 L .67241 .26611 L p F P
0 g
s
.65797 .15814 m .65064 .16716 L .65209 .15706 L closepath p .196 .726 .199 r
F P
0 g
s
.65648 .16805 m .65064 .16716 L .65797 .15814 L closepath p .197 .731 .2 r
F P
0 g
s
.651 .18408 m .65064 .16716 L .65648 .16805 L closepath p .199 .713 .193 r
F P
0 g
s
.65375 .1508 m .65077 .15034 L .65283 .1455 L closepath p .196 .736 .202 r
F P
0 g
s
.65077 .15034 m .65375 .1508 L .65209 .15706 L closepath p .192 .761 .212 r
F P
0 g
s
.65561 .30873 m .66002 .31261 L .6508 .31291 L closepath p .161 .19 .039 r
F P
0 g
s
.66031 .32351 m .6508 .31291 L .66002 .31261 L closepath p .198 .554 .144 r
F P
0 g
s
.651 .18408 m .65648 .16805 L .66398 .16933 L closepath p .204 .737 .194 r
F P
0 g
s
.651 .18408 m .66398 .16933 L .6585 .18533 L closepath p .204 .737 .194 r
F P
0 g
s
.6562 .45089 m .65105 .44445 L .66717 .42902 L closepath p .215 .806 .199 r
F P
0 g
s
.65158 .30886 m .65899 .29611 L p .65287 .30882 L 0 0 .044 r
F P
0 g
s
.6555 .49034 m .65197 .48431 L .65765 .48566 L closepath p .201 .713 .191 r
F P
0 g
s
.65678 .1485 m .65511 .15476 L .65209 .15706 L closepath p .216 .862 .209 r
F P
0 g
s
.65678 .1485 m .65209 .15706 L .65375 .1508 L closepath p .216 .862 .209 r
F P
0 g
s
.65797 .15814 m .65209 .15706 L .65511 .15476 L closepath p .201 .684 .183 r
F P
0 g
s
.65212 .24908 m .66129 .25614 L .66032 .2555 L .184 .226 0 r
F
.65212 .24908 m .65803 .25144 L p .65983 .25502 L F P
0 g
s
.65212 .24908 m .66032 .2555 L .65793 .25393 L .184 .226 0 r
F
.65614 .25272 m .65212 .24908 L p .65793 .25393 L F P
0 g
s
.65212 .24908 m .65637 .25293 L p .65451 .25205 L .226 .527 .078 r
F P
0 g
s
.65333 .21525 m .6587 .2307 L .65247 .22663 L closepath p .231 .706 .147 r
F P
0 g
s
.65247 .22663 m .65473 .24101 L .65465 .24074 L .225 .75 .172 r
F
.65247 .22663 m .65465 .24074 L .65462 .24065 L F
.65247 .22663 m .65561 .24379 L .65533 .24291 L F
.65247 .22663 m .65533 .24291 L .65485 .24138 L F
.65247 .22663 m .65714 .24862 L .65561 .24379 L F
.65247 .22663 m .65485 .24138 L .65484 .24136 L F
.65247 .22663 m .65484 .24136 L .65473 .24101 L F
.65416 .23921 m .65247 .22663 L .65433 .23975 L p .181 .51 0 r
F P
0 g
s
.65247 .22663 m .65462 .24065 L .6546 .2406 L .225 .75 .172 r
F
.65433 .23975 m .65247 .22663 L p .6546 .2406 L F P
0 g
s
.65247 .22663 m .6587 .2307 L .65714 .24862 L p .225 .75 .172 r
F P
0 g
s
.66628 .26416 m .66689 .26675 L .65262 .26043 L closepath p .096 0 0 r
F P
0 g
s
.65262 .26043 m .66689 .26675 L .6559 .27175 L closepath p .128 .026 .009 r
F P
0 g
s
.6528 .25124 m .6565 .25299 L p .65629 .25332 L .105 0 0 r
F P
0 g
s
.65581 .14596 m .65283 .1455 L .65372 .14183 L closepath p .202 .678 .18 r
F P
0 g
s
.65375 .1508 m .65283 .1455 L .65581 .14596 L closepath p .196 .736 .202 r
F P
0 g
s
.65287 .30882 m .65899 .29611 L .65289 .30882 L 0 0 .044 r
F
.65289 .30882 m .65899 .29611 L .65315 .30881 L F
.65315 .1413 m .6569 .1424 L .65372 .14183 L closepath p .137 0 0 r
F P
0 g
s
.65315 .1413 m .65633 .14187 L .6569 .1424 L closepath p .137 0 0 r
F P
0 g
s
.65315 .30881 m .65899 .29611 L .65391 .30879 L 0 0 .044 r
F
.65602 .20965 m .65955 .21934 L .65333 .21525 L closepath p .23 .714 .151 r
F P
0 g
s
.65333 .21525 m .65955 .21934 L .6587 .2307 L closepath p .231 .706 .147 r
F P
0 g
s
.65581 .14596 m .65372 .14183 L .6569 .1424 L closepath p .198 .674 .183 r
F P
0 g
s
.65678 .1485 m .65375 .1508 L .65581 .14596 L closepath p .214 .826 .204 r
F P
0 g
s
.65391 .30879 m .65571 .30821 L p .65463 .30876 L 0 0 0 r
F P
0 g
s
.65391 .30879 m .65899 .29611 L .65484 .30849 L 0 0 .044 r
F
.66142 .32353 m .67454 .32371 L .65399 .34244 L p .18 .576 .173 r
F P
0 g
s
.65399 .34244 m .67454 .32371 L .67189 .33741 L closepath p .172 .52 .164 r
F P
0 g
s
.65399 .34244 m .67189 .33741 L .66732 .35558 L closepath p .175 .563 .174 r
F P
0 g
s
.66002 .49398 m .65432 .49061 L .6555 .49034 L closepath p .149 .701 .229 r
F P
0 g
s
.66002 .49398 m .65884 .49425 L .65432 .49061 L closepath p .149 .701 .229 r
F P
0 g
s
.65571 .30821 m .66265 .30647 L .65463 .30876 L p 0 0 0 r
F P
0 g
s
.65463 .30876 m .65985 .30727 L p .65561 .30873 L 0 0 0 r
F P
0 g
s
.65484 .30849 m .65899 .29611 L .65485 .30849 L 0 0 .044 r
F
.65899 .29611 m .65829 .30461 L .65485 .30849 L p F P
0 g
s
.65485 .30849 m .65829 .30461 L p .65505 .30842 L 0 0 0 r
F P
0 g
s
.65505 .30842 m .65829 .30461 L .65506 .30842 L 0 0 0 r
F
.65829 .30461 m .65571 .30821 L p .65506 .30842 L F P
0 g
s
.66541 .26355 m .65511 .25563 L .66628 .26416 L .155 .073 0 r
F
.66628 .26416 m .65511 .25563 L .66636 .2642 L F
.66636 .2642 m .65511 .25563 L .66644 .26423 L F
.66644 .26423 m .65511 .25563 L .66687 .26442 L F
.65745 .25789 m .65511 .25563 L p .66541 .26355 L F P
0 g
s
.65678 .1485 m .65727 .14776 L .65511 .15476 L closepath p .2 .821 .086 r
F P
0 g
s
.65511 .15476 m .65727 .14776 L .65873 .14817 L closepath p .2 .68 .183 r
F P
0 g
s
.65511 .15476 m .65873 .14817 L .65658 .15516 L closepath p .2 .68 .183 r
F P
0 g
s
.65797 .15814 m .65511 .15476 L .65658 .15516 L closepath p .199 .684 .185 r
F P
0 g
s
.66194 .27809 m .65548 .28767 L .66149 .27743 L .109 .113 .079 r
F
.66194 .27809 m .65828 .29188 L .65548 .28767 L closepath p .015 0 .045 r
F P
0 g
s
.6555 .49034 m .65765 .48566 L .66117 .49168 L closepath p .201 .713 .191 r
F P
0 g
s
.66117 .49168 m .66002 .49398 L .6555 .49034 L closepath p .202 .712 .19 r
F P
0 g
s
.66002 .31261 m .65561 .30873 L .66265 .30647 L closepath p .153 .215 .067 r
F P
0 g
s
.65985 .30727 m .66265 .30647 L .65561 .30873 L p 0 0 0 r
F P
0 g
s
.67112 .30155 m .65571 .30821 L .65829 .30461 L closepath p .039 0 .001 r
F P
0 g
s
.67112 .30155 m .66265 .30647 L .65571 .30821 L closepath p .024 0 0 r
F P
0 g
s
.65787 .14494 m .65581 .14596 L .6569 .1424 L closepath p .235 .847 .181 r
F P
0 g
s
.65787 .14494 m .65678 .1485 L .65581 .14596 L closepath p .235 .847 .181 r
F P
0 g
s
.6559 .27175 m .66689 .26675 L .67944 .26529 L p .036 0 0 r
F P
0 g
s
.67874 .26572 m .6559 .27175 L .67944 .26529 L .036 0 0 r
F
.67187 .26991 m .6559 .27175 L .67812 .26609 L p F P
0 g
s
.67827 .266 m .6559 .27175 L .67867 .26576 L .036 0 0 r
F
.67867 .26576 m .6559 .27175 L .67874 .26572 L F
.67812 .26609 m .6559 .27175 L p .67827 .266 L F P
0 g
s
.65666 .20378 m .6598 .21668 L .65602 .20965 L closepath p .249 .751 .103 r
F P
0 g
s
.65602 .20965 m .6598 .21668 L .65955 .21934 L closepath p .23 .593 .031 r
F P
0 g
s
.65793 .25393 m .6565 .25299 L .65614 .25272 L .184 .226 0 r
F
.65637 .25293 m .65614 .25272 L p .6565 .25299 L F P
0 g
s
.6562 .45089 m .66717 .42902 L .67228 .43548 L closepath p .215 .806 .199 r
F P
0 g
s
.6562 .45089 m .67228 .43548 L .66682 .45677 L closepath p .209 .751 .192 r
F P
0 g
s
.6565 .25299 m .6638 .25779 L p .65629 .25332 L .105 0 0 r
F P
0 g
s
.6585 .18533 m .65894 .18948 L .65631 .19891 L closepath p .246 .883 .133 r
F P
0 g
s
.65894 .18948 m .66035 .20048 L .65631 .19891 L closepath p .232 .752 .161 r
F P
0 g
s
.66043 .21081 m .65666 .20378 L .65631 .19891 L closepath p .235 .624 .044 r
F P
0 g
s
.66043 .21081 m .65631 .19891 L .66035 .20048 L closepath p .233 .723 .149 r
F P
0 g
s
.65645 .41823 m .66762 .38752 L .67574 .39682 L closepath p .209 .786 .201 r
F P
0 g
s
.68638 .40769 m .65645 .41823 L .67574 .39682 L closepath p .205 .758 .198 r
F P
0 g
s
.68638 .40769 m .66717 .42902 L .65645 .41823 L closepath p .205 .758 .198 r
F P
0 g
s
.65648 .16805 m .65797 .15814 L .66381 .15903 L closepath p .197 .731 .2 r
F P
0 g
s
.669 .16203 m .65648 .16805 L .66381 .15903 L closepath p .207 .803 .207 r
F P
0 g
s
.669 .16203 m .66398 .16933 L .65648 .16805 L closepath p .204 .739 .195 r
F P
0 g
s
.66226 .14615 m .66117 .1497 L .65658 .15516 L closepath p .122 .353 .156 r
F P
0 g
s
.66226 .14615 m .65658 .15516 L .65873 .14817 L closepath p .122 .353 .156 r
F P
0 g
s
.66256 .15268 m .65658 .15516 L .66117 .1497 L closepath p .171 .636 .199 r
F P
0 g
s
.66256 .15268 m .65797 .15814 L .65658 .15516 L closepath p .171 .636 .199 r
F P
0 g
s
.65666 .20378 m .66043 .21081 L .6598 .21668 L closepath p .249 .751 .103 r
F P
0 g
s
.65678 .1485 m .65787 .14494 L .65727 .14776 L closepath p .2 .821 .086 r
F P
0 g
s
.65714 .24862 m .6587 .2307 L p .65725 .24897 L .249 .846 .141 r
F P
0 g
s
.65725 .24897 m .6587 .2307 L .65728 .24906 L .249 .846 .141 r
F
.65873 .14817 m .65727 .14776 L p .65732 .14753 L .2 .676 .182 r
F P
0 g
s
.65728 .24906 m .6587 .2307 L .65728 .24906 L .249 .846 .141 r
F
.65728 .24906 m .6587 .2307 L .65746 .24963 L F
.65749 .14674 m .65873 .14817 L p .65732 .14753 L .2 .676 .182 r
F P
0 g
s
.65746 .24963 m .6587 .2307 L .65758 .25001 L .249 .846 .141 r
F
.65749 .14674 m .65749 .14673 L .65749 .14674 L .136 .24 .101 r
F
.65749 .14674 m .65749 .14673 L .65816 .14751 L F
.65749 .14673 m .66189 .14595 L p .65816 .14751 L F P
0 g
s
.6587 .2307 m .66128 .23075 L .65758 .25001 L p .249 .846 .141 r
F P
0 g
s
.65758 .25001 m .66128 .23075 L p .6579 .25103 L .249 .846 .141 r
F P
0 g
s
.66083 .47407 m .67113 .4907 L .65765 .48566 L closepath p .217 .729 .177 r
F P
0 g
s
.66523 .49319 m .66117 .49168 L .65765 .48566 L closepath p .218 .725 .175 r
F P
0 g
s
.66523 .49319 m .65765 .48566 L .67113 .4907 L closepath p .218 .725 .175 r
F P
0 g
s
.66031 .32351 m .66142 .32353 L p .65783 .33266 L .18 .576 .173 r
F P
0 g
s
.66128 .23075 m .65932 .2533 L p .6579 .25103 L .249 .846 .141 r
F P
0 g
s
.6579 .25103 m .65932 .2533 L .65803 .25144 L .249 .846 .141 r
F
.66256 .15268 m .66381 .15903 L .65797 .15814 L closepath p .191 .766 .214 r
F P
0 g
s
.65803 .25144 m .66129 .25614 L p .65983 .25502 L .184 .226 0 r
F P
0 g
s
.6582 .14755 m .65816 .14751 L .66189 .14595 L .136 .24 .101 r
F
.65822 .14758 m .6582 .14755 L .66169 .14604 L F
.65873 .14817 m .65822 .14758 L .66169 .14604 L F
.66194 .27809 m .66477 .28225 L .65828 .29188 L closepath p .015 0 .045 r
F P
0 g
s
.66899 .27994 m .65828 .29188 L .66477 .28225 L closepath p 0 0 .018 r
F P
0 g
s
.66899 .27994 m .65899 .29611 L .65828 .29188 L closepath p .01 0 .048 r
F P
0 g
s
.65829 .30461 m .66714 .28652 L .66647 .295 L closepath p .001 0 .046 r
F P
0 g
s
.65829 .30461 m .65899 .29611 L .66714 .28652 L closepath p .001 0 .046 r
F P
0 g
s
.67112 .30155 m .65829 .30461 L .66647 .295 L closepath p .087 .209 .137 r
F P
0 g
s
.6585 .18533 m .66398 .16933 L .66597 .17434 L closepath p .225 .838 .194 r
F P
0 g
s
.6585 .18533 m .66597 .17434 L .66639 .17851 L closepath p .232 .905 .195 r
F P
0 g
s
.6585 .18533 m .66639 .17851 L .65894 .18948 L closepath p .232 .905 .195 r
F P
0 g
s
.65955 .21934 m .66128 .23075 L .6587 .2307 L closepath p .249 .79 .124 r
F P
0 g
s
.66189 .14595 m .65873 .14817 L p .66169 .14604 L .136 .24 .101 r
F P
0 g
s
.66226 .14615 m .65873 .14817 L .66189 .14595 L closepath p 0 0 0 r
F P
0 g
s
.66305 .49591 m .65884 .49425 L .66002 .49398 L closepath p .139 .81 .248 r
F P
0 g
s
.66638 .17956 m .65894 .18948 L .66639 .17851 L closepath p .184 .911 .135 r
F P
0 g
s
.66638 .17956 m .66296 .19106 L .65894 .18948 L closepath p .232 .754 .162 r
F P
0 g
s
.65894 .18948 m .66296 .19106 L .66035 .20048 L closepath p .232 .752 .161 r
F P
0 g
s
.66899 .27994 m .66714 .28652 L .65899 .29611 L closepath p 0 0 .024 r
F P
0 g
s
.65932 .2533 m .66128 .23075 L p .65932 .2533 L .178 .63 .017 r
F P
0 g
s
.65932 .2533 m .66128 .23075 L .65932 .2533 L .178 .63 .017 r
F
.66128 .23075 m .66153 .22807 L .65932 .2533 L p F P
0 g
s
.65955 .21934 m .6598 .21668 L .66153 .22807 L closepath p .217 .637 .026 r
F P
0 g
s
.65955 .21934 m .66153 .22807 L .66128 .23075 L closepath p .217 .637 .026 r
F P
0 g
s
.66407 .49549 m .66002 .49398 L .66117 .49168 L closepath p .211 .764 .193 r
F P
0 g
s
.66305 .49591 m .66002 .49398 L .66407 .49549 L closepath p .137 .807 .248 r
F P
0 g
s
.66002 .31261 m .66265 .30647 L .67422 .31282 L closepath p .159 .24 .069 r
F P
0 g
s
.67454 .32371 m .66002 .31261 L .67422 .31282 L closepath p .178 .489 .146 r
F P
0 g
s
.67454 .32371 m .66031 .32351 L .66002 .31261 L closepath p .178 .489 .146 r
F P
0 g
s
.66316 .19114 m .66035 .20048 L .66296 .19106 L closepath p .19 .766 .065 r
F P
0 g
s
.67853 .45867 m .66083 .47407 L .66682 .45677 L closepath p .215 .717 .176 r
F P
0 g
s
.67853 .45867 m .67429 .47914 L .66083 .47407 L closepath p .217 .732 .178 r
F P
0 g
s
.66083 .47407 m .67429 .47914 L .67113 .4907 L closepath p .217 .729 .177 r
F P
0 g
s
.66117 .1497 m .66583 .14454 L .66474 .14809 L closepath p .166 .535 .174 r
F P
0 g
s
.66117 .1497 m .66226 .14615 L .66583 .14454 L closepath p .166 .535 .174 r
F P
0 g
s
.66474 .14809 m .66256 .15268 L .66117 .1497 L closepath p .176 .645 .197 r
F P
0 g
s
.66407 .49549 m .66117 .49168 L .66523 .49319 L closepath p .211 .764 .193 r
F P
0 g
s
.66149 .27743 m .67187 .26991 L p .66194 .27809 L .109 .113 .079 r
F P
0 g
s
.66226 .14615 m .66189 .14595 L .66543 .14393 L closepath p 0 0 0 r
F P
0 g
s
.67187 .26991 m .66799 .27312 L .66194 .27809 L closepath p 0 0 0 r
F P
0 g
s
.67127 .27568 m .66194 .27809 L .66799 .27312 L closepath p .044 0 .012 r
F P
0 g
s
.67127 .27568 m .66477 .28225 L .66194 .27809 L closepath p .049 0 .024 r
F P
0 g
s
.66583 .14454 m .66226 .14615 L .66543 .14393 L closepath p .097 .024 .055 r
F P
0 g
s
.66553 .15314 m .66256 .15268 L .66474 .14809 L closepath p .195 .742 .204 r
F P
0 g
s
.66256 .15268 m .66553 .15314 L .66381 .15903 L closepath p .191 .766 .214 r
F P
0 g
s
.67112 .30155 m .67422 .31282 L .66265 .30647 L closepath p .143 .353 .136 r
F P
0 g
s
.66296 .19106 m .66638 .17956 L .66657 .17966 L closepath p .189 .773 .068 r
F P
0 g
s
.66296 .19106 m .66657 .17966 L .66316 .19114 L closepath p .189 .773 .068 r
F P
0 g
s
.66305 .49591 m .6685 .49528 L .66748 .4957 L closepath p .117 .882 .234 r
F P
0 g
s
.66305 .49591 m .66407 .49549 L .6685 .49528 L closepath p .117 .882 .234 r
F P
0 g
s
.6638 .25779 m .67106 .26339 L .67112 .26351 L .141 .02 0 r
F
.6638 .25779 m .67457 .26586 L p .67441 .26587 L F P
0 g
s
.6638 .25779 m .67441 .26587 L .67436 .26588 L .141 .02 0 r
F
.6638 .25779 m .67436 .26588 L .67435 .26588 L F
.6638 .25779 m .67435 .26588 L .67433 .26588 L F
.6638 .25779 m .67433 .26588 L .67431 .26589 L F
.66869 .15093 m .66381 .15903 L .66553 .15314 L closepath p .213 .862 .212 r
F P
0 g
s
.66869 .15093 m .66697 .15682 L .66381 .15903 L closepath p .213 .862 .212 r
F P
0 g
s
.669 .16203 m .66381 .15903 L .66697 .15682 L closepath p .21 .798 .203 r
F P
0 g
s
.66597 .17434 m .66398 .16933 L .669 .16203 L closepath p .217 .829 .201 r
F P
0 g
s
.6685 .49528 m .66407 .49549 L .66523 .49319 L closepath p .223 .836 .195 r
F P
0 g
s
.66771 .14855 m .66474 .14809 L .66583 .14454 L closepath p .201 .687 .183 r
F P
0 g
s
.66553 .15314 m .66474 .14809 L .66771 .14855 L closepath p .195 .742 .204 r
F P
0 g
s
.6755 .27335 m .66477 .28225 L .67127 .27568 L closepath p 0 0 0 r
F P
0 g
s
.6755 .27335 m .66899 .27994 L .66477 .28225 L closepath p 0 0 0 r
F P
0 g
s
.68737 .34859 m .66509 .37015 L .66732 .35558 L closepath p .156 .49 .169 r
F P
0 g
s
.68737 .34859 m .68514 .36319 L .66509 .37015 L closepath p .156 .49 .169 r
F P
0 g
s
.68862 .37248 m .66509 .37015 L .68514 .36319 L closepath p .168 .652 .206 r
F P
0 g
s
.68862 .37248 m .66762 .38752 L .66509 .37015 L closepath p .172 .61 .191 r
F P
0 g
s
.66523 .49319 m .67113 .4907 L .67439 .49279 L closepath p .221 .84 .199 r
F P
0 g
s
.66523 .49319 m .67439 .49279 L .6685 .49528 L closepath p .221 .84 .199 r
F P
0 g
s
.66543 .14393 m .669 .14511 L .66583 .14454 L closepath p .157 .104 0 r
F P
0 g
s
.66543 .14393 m .66859 .1445 L .669 .14511 L closepath p .157 .104 0 r
F P
0 g
s
.66869 .15093 m .66553 .15314 L .66771 .14855 L closepath p .212 .827 .207 r
F P
0 g
s
.66771 .14855 m .66583 .14454 L .669 .14511 L closepath p .198 .682 .186 r
F P
0 g
s
.66597 .17434 m .669 .16203 L .67099 .16706 L closepath p .217 .829 .201 r
F P
0 g
s
.67207 .16576 m .66639 .17851 L .66597 .17434 L closepath p .234 .908 .193 r
F P
0 g
s
.67207 .16576 m .66597 .17434 L .67099 .16706 L closepath p .172 .885 .127 r
F P
0 g
s
.66628 .26416 m .67104 .26627 L p .66964 .26643 L .096 0 0 r
F P
0 g
s
.66689 .26675 m .66628 .26416 L p .66964 .26643 L .096 0 0 r
F P
0 g
s
.66729 .1765 m .66638 .17956 L .66707 .17698 L p .189 .775 .069 r
F P
0 g
s
.66707 .17698 m .66638 .17956 L .66639 .17851 L p .189 .775 .069 r
F P
0 g
s
.66657 .17966 m .66638 .17956 L p .66642 .17944 L .166 .267 0 r
F P
0 g
s
.66643 .17939 m .66657 .17966 L p .66642 .17944 L .166 .267 0 r
F P
0 g
s
.67425 .28084 m .67517 .28646 L .66647 .295 L closepath p .023 0 .06 r
F P
0 g
s
.67425 .28084 m .66647 .295 L .66714 .28652 L closepath p .005 0 .047 r
F P
0 g
s
.67985 .29301 m .66647 .295 L .67517 .28646 L closepath p .088 .21 .136 r
F P
0 g
s
.67985 .29301 m .67112 .30155 L .66647 .295 L closepath p .088 .21 .136 r
F P
0 g
s
.67853 .45867 m .66682 .45677 L .67228 .43548 L closepath p .211 .761 .192 r
F P
0 g
s
.66869 .15093 m .66951 .15005 L .66697 .15682 L closepath p .217 .9 .121 r
F P
0 g
s
.66697 .15682 m .66951 .15005 L .67176 .15081 L closepath p .223 .804 .189 r
F P
0 g
s
.66697 .15682 m .67176 .15081 L .66921 .15758 L closepath p .223 .804 .189 r
F P
0 g
s
.669 .16203 m .66697 .15682 L .66921 .15758 L closepath p .221 .814 .194 r
F P
0 g
s
.66899 .27994 m .67425 .28084 L .66714 .28652 L closepath p 0 0 .029 r
F P
0 g
s
.68638 .40769 m .67228 .43548 L .66717 .42902 L closepath p .211 .805 .203 r
F P
0 g
s
.68693 .33396 m .66732 .35558 L .67189 .33741 L closepath p .164 .518 .171 r
F P
0 g
s
.68693 .33396 m .68737 .34859 L .66732 .35558 L closepath p .153 .453 .161 r
F P
0 g
s
.68862 .37248 m .67574 .39682 L .66762 .38752 L closepath p .189 .775 .217 r
F P
0 g
s
.66998 .14748 m .66771 .14855 L .669 .14511 L closepath p .23 .847 .188 r
F P
0 g
s
.66998 .14748 m .66869 .15093 L .66771 .14855 L closepath p .23 .847 .188 r
F P
0 g
s
.67848 .26728 m .67796 .26779 L .66799 .27312 L 0 0 0 r
F
.66799 .27312 m .67187 .26991 L .67904 .26673 L p F P
0 g
s
.67904 .26673 m .67848 .26728 L .66799 .27312 L 0 0 0 r
F
.67754 .2682 m .66799 .27312 L p .67796 .26779 L F P
0 g
s
.67635 .27094 m .66799 .27312 L p .67719 .26855 L .039 0 .001 r
F P
0 g
s
.66799 .27312 m .67722 .26837 L p .67719 .26855 L .039 0 .001 r
F P
0 g
s
.67127 .27568 m .66799 .27312 L .67635 .27094 L p .044 0 .012 r
F P
0 g
s
.66869 .15093 m .66998 .14748 L .66951 .15005 L closepath p .217 .9 .121 r
F P
0 g
s
.67609 .27425 m .66899 .27994 L .6755 .27335 L closepath p 0 0 0 r
F P
0 g
s
.66899 .27994 m .67609 .27425 L .67425 .28084 L closepath p 0 0 .029 r
F P
0 g
s
.67521 .15768 m .669 .16203 L .66921 .15758 L closepath p .204 .759 .2 r
F P
0 g
s
.67521 .15768 m .67099 .16706 L .669 .16203 L closepath p .208 .816 .208 r
F P
0 g
s
.66918 .26544 m .67164 .2662 L .67104 .26627 L .181 .192 0 r
F
.67672 .1498 m .66921 .15758 L .67176 .15081 L closepath p .195 .67 .185 r
F P
0 g
s
.67672 .1498 m .67543 .15323 L .66921 .15758 L closepath p .195 .67 .185 r
F P
0 g
s
.67521 .15768 m .66921 .15758 L .67543 .15323 L closepath p .204 .759 .2 r
F P
0 g
s
.67176 .15081 m .66951 .15005 L p .66958 .14965 L .231 .716 .151 r
F P
0 g
s
.66975 .14875 m .67176 .15081 L p .66958 .14965 L .231 .716 .151 r
F P
0 g
s
.67037 .14939 m .67284 .14952 L p .67091 .14994 L .166 .235 .054 r
F P
0 g
s
.67095 .14998 m .67091 .14994 L .67284 .14952 L .166 .235 .054 r
F
.67098 .15001 m .67095 .14998 L .67284 .14952 L F
.67284 .14952 m .67176 .15081 L p .67098 .15001 L F P
0 g
s
.67521 .15768 m .67621 .16024 L .67099 .16706 L closepath p .208 .816 .208 r
F P
0 g
s
.67729 .15894 m .67099 .16706 L .67621 .16024 L closepath p .151 .846 .119 r
F P
0 g
s
.67729 .15894 m .67207 .16576 L .67099 .16706 L closepath p .151 .846 .119 r
F P
0 g
s
.67104 .4942 m .6798 .49393 L p .67103 .49421 L .135 .84 .25 r
F P
0 g
s
.6798 .49393 m .67641 .49535 L .67103 .49421 L p .135 .84 .25 r
F P
0 g
s
.67439 .49279 m .6798 .49393 L .67104 .4942 L p .135 .84 .25 r
F P
0 g
s
.67106 .26339 m .67394 .26593 L .67114 .26352 L .141 .02 0 r
F
.67106 .26339 m .67431 .26589 L .67394 .26593 L F
.67106 .26339 m .67114 .26352 L .67112 .26351 L F
.67112 .30155 m .67985 .29301 L .6867 .29874 L closepath p .136 .498 .189 r
F P
0 g
s
.67112 .30155 m .6867 .29874 L .68272 .30789 L closepath p .13 .39 .161 r
F P
0 g
s
.67112 .30155 m .68272 .30789 L .67422 .31282 L closepath p .143 .353 .136 r
F P
0 g
s
.67241 .26611 m .67112 .26351 L p .67256 .26609 L .183 .217 0 r
F P
0 g
s
.67112 .26351 m .67394 .26593 L p .67256 .26609 L .183 .217 0 r
F P
0 g
s
.67429 .47914 m .67439 .49279 L .67113 .4907 L closepath p .233 .805 .174 r
F P
0 g
s
.67394 .26593 m .67394 .26593 L .67114 .26352 L .141 .02 0 r
F
.67127 .27568 m .67635 .27094 L .67622 .27131 L .044 0 .012 r
F
.67607 .27173 m .67127 .27568 L p .67622 .27131 L F P
0 g
s
.67606 .27176 m .67127 .27568 L .67607 .27173 L p 0 0 0 r
F P
0 g
s
.6755 .27335 m .67127 .27568 L .67606 .27176 L p 0 0 0 r
F P
0 g
s
.67672 .1498 m .67176 .15081 L .67284 .14952 L closepath p .13 .067 .029 r
F P
0 g
s
.67187 .26991 m .68227 .26356 L p .68043 .26537 L 0 0 0 r
F P
0 g
s
.68043 .26537 m .68018 .26561 L .67187 .26991 L 0 0 0 r
F
.67904 .26673 m .67187 .26991 L .68013 .26566 L p F P
0 g
s
.68013 .26566 m .67187 .26991 L p .68018 .26561 L 0 0 0 r
F P
0 g
s
.68957 .32025 m .67189 .33741 L .67454 .32371 L closepath p .161 .48 .161 r
F P
0 g
s
.68957 .32025 m .68693 .33396 L .67189 .33741 L closepath p .161 .48 .161 r
F P
0 g
s
.68638 .40769 m .68398 .43745 L .67228 .43548 L closepath p .208 .787 .202 r
F P
0 g
s
.67853 .45867 m .67228 .43548 L .68398 .43745 L closepath p .211 .761 .192 r
F P
0 g
s
.67672 .1498 m .67284 .14952 L .6778 .14851 L closepath p .13 .067 .029 r
F P
0 g
s
.68955 .3125 m .67422 .31282 L .68272 .30789 L closepath p .149 .399 .146 r
F P
0 g
s
.68955 .3125 m .67454 .32371 L .67422 .31282 L closepath p .149 .398 .146 r
F P
0 g
s
.68504 .26805 m .67425 .28084 L .67609 .27425 L closepath p .051 0 .052 r
F P
0 g
s
.68504 .26805 m .68498 .27341 L .67425 .28084 L closepath p .078 .027 .077 r
F P
0 g
s
.68498 .27341 m .67517 .28646 L .67425 .28084 L closepath p .065 0 .06 r
F P
0 g
s
.67429 .47914 m .67853 .45867 L .69138 .46152 L closepath p .214 .723 .178 r
F P
0 g
s
.67429 .47914 m .69138 .46152 L .68715 .48192 L closepath p .214 .723 .178 r
F P
0 g
s
.67429 .47914 m .68715 .48192 L .68578 .49381 L closepath p .215 .719 .177 r
F P
0 g
s
.67429 .47914 m .68578 .49381 L .6798 .49393 L closepath p .197 .715 .196 r
F P
0 g
s
.67429 .47914 m .6798 .49393 L .67439 .49279 L closepath p .213 .743 .186 r
F P
0 g
s
.68955 .3125 m .68957 .32025 L .67454 .32371 L closepath p .161 .476 .16 r
F P
0 g
s
.68498 .27341 m .68594 .27901 L .67517 .28646 L closepath p .065 0 .06 r
F P
0 g
s
.68594 .27901 m .67985 .29301 L .67517 .28646 L closepath p .114 .238 .123 r
F P
0 g
s
.67621 .16024 m .67521 .15768 L .67851 .15295 L closepath p .22 .833 .199 r
F P
0 g
s
.67851 .15295 m .67521 .15768 L .67543 .15323 L closepath p .214 .793 .197 r
F P
0 g
s
.67543 .15323 m .67672 .1498 L .6798 .14952 L closepath p .215 .767 .189 r
F P
0 g
s
.67543 .15323 m .6798 .14952 L .67851 .15295 L closepath p .215 .767 .189 r
F P
0 g
s
.67609 .27425 m .6755 .27335 L .67719 .26855 L closepath p 0 0 0 r
F P
0 g
s
.68862 .37248 m .69668 .38185 L .67574 .39682 L closepath p .189 .775 .217 r
F P
0 g
s
.69668 .38185 m .68638 .40769 L .67574 .39682 L closepath p .187 .753 .214 r
F P
0 g
s
.68612 .26234 m .68504 .26805 L .67609 .27425 L closepath p .073 .025 .082 r
F P
0 g
s
.68612 .26234 m .67609 .27425 L .67719 .26855 L closepath p .073 .025 .082 r
F P
0 g
s
.67621 .16024 m .67851 .15295 L .6795 .15552 L closepath p .22 .833 .199 r
F P
0 g
s
.67729 .15894 m .67621 .16024 L .6795 .15552 L closepath p .171 .883 .127 r
F P
0 g
s
.68473 .4987 m .67641 .49535 L .6798 .49393 L closepath p .135 .666 .23 r
F P
0 g
s
.68473 .4987 m .68371 .49913 L .67641 .49535 L closepath p .135 .666 .23 r
F P
0 g
s
.6798 .14952 m .67672 .1498 L .6778 .14851 L closepath p .183 .318 .066 r
F P
0 g
s
.67722 .26837 m .67754 .2682 L p .67719 .26855 L .039 0 .001 r
F P
0 g
s
.68612 .26234 m .67719 .26855 L .6841 .26176 L closepath p 0 0 0 r
F P
0 g
s
.6795 .15552 m .67851 .15295 L .6798 .14952 L closepath p .23 .843 .188 r
F P
0 g
s
.67853 .45867 m .68398 .43745 L .7005 .44121 L closepath p .211 .758 .192 r
F P
0 g
s
.67853 .45867 m .7005 .44121 L .69138 .46152 L closepath p .211 .758 .192 r
F P
0 g
s
.6795 .15552 m .6798 .14952 L .67996 .15122 L closepath p .211 .737 .054 r
F P
0 g
s
.68473 .4987 m .6798 .49393 L .68578 .49381 L closepath p .197 .715 .196 r
F P
0 g
s
.68594 .27901 m .6902 .28375 L .67985 .29301 L closepath p .125 .286 .13 r
F P
0 g
s
.6902 .28375 m .6867 .29874 L .67985 .29301 L closepath p .16 .496 .168 r
F P
0 g
s
.69354 .30334 m .68272 .30789 L .6867 .29874 L closepath p .137 .418 .164 r
F P
0 g
s
.69354 .30334 m .68955 .3125 L .68272 .30789 L closepath p .137 .418 .164 r
F P
0 g
s
.68371 .49913 m .68473 .4987 L .69015 .50098 L closepath p .138 .753 .243 r
F P
0 g
s
.68371 .49913 m .69015 .50098 L .68913 .5014 L closepath p .138 .753 .243 r
F P
0 g
s
.68638 .40769 m .69535 .41332 L .68398 .43745 L closepath p .208 .786 .202 r
F P
0 g
s
.68398 .43745 m .69535 .41332 L .71185 .4172 L closepath p .207 .782 .202 r
F P
0 g
s
.68398 .43745 m .71185 .4172 L .7005 .44121 L closepath p .207 .782 .202 r
F P
0 g
s
.69439 .25591 m .6841 .26176 L .69235 .25535 L closepath p 0 0 0 r
F P
0 g
s
.69439 .25591 m .68612 .26234 L .6841 .26176 L closepath p 0 0 0 r
F P
0 g
s
.68473 .4987 m .68578 .49381 L .6907 .49858 L closepath p .197 .715 .196 r
F P
0 g
s
.69015 .50098 m .68473 .4987 L .6907 .49858 L closepath p .197 .706 .193 r
F P
0 g
s
.70222 .26799 m .68498 .27341 L .6958 .26895 L closepath p .085 0 .037 r
F P
0 g
s
.70222 .26799 m .68594 .27901 L .68498 .27341 L closepath p .096 .027 .058 r
F P
0 g
s
.68504 .26805 m .6958 .26895 L .68498 .27341 L closepath p .112 .126 .081 r
F P
0 g
s
.69584 .26359 m .68504 .26805 L .68612 .26234 L closepath p .122 .193 .098 r
F P
0 g
s
.68504 .26805 m .69584 .26359 L .6958 .26895 L closepath p .112 .126 .081 r
F P
0 g
s
.70076 .35792 m .68514 .36319 L .68737 .34859 L closepath p .133 .406 .163 r
F P
0 g
s
.70424 .36725 m .68514 .36319 L .70076 .35792 L closepath p .148 .607 .209 r
F P
0 g
s
.70424 .36725 m .68862 .37248 L .68514 .36319 L closepath p .148 .607 .209 r
F P
0 g
s
.70072 .49484 m .68578 .49381 L .68715 .48192 L closepath p .212 .71 .177 r
F P
0 g
s
.69519 .49888 m .68578 .49381 L .70072 .49484 L closepath p .213 .705 .175 r
F P
0 g
s
.69519 .49888 m .6907 .49858 L .68578 .49381 L closepath p .213 .705 .175 r
F P
0 g
s
.68594 .27901 m .70222 .26799 L .70651 .27269 L closepath p .142 .29 .113 r
F P
0 g
s
.68594 .27901 m .70651 .27269 L .6902 .28375 L closepath p .142 .29 .113 r
F P
0 g
s
.69584 .26359 m .68612 .26234 L .69439 .25591 L closepath p .122 .202 .101 r
F P
0 g
s
.69668 .38185 m .70432 .38771 L .68638 .40769 L closepath p .2 .816 .216 r
F P
0 g
s
.69535 .41332 m .68638 .40769 L .70432 .38771 L closepath p .198 .796 .214 r
F P
0 g
s
.6902 .28375 m .69706 .28948 L .6867 .29874 L closepath p .16 .496 .168 r
F P
0 g
s
.69706 .28948 m .69354 .30334 L .6867 .29874 L closepath p .145 .406 .152 r
F P
0 g
s
.69819 .32045 m .68693 .33396 L .68957 .32025 L closepath p .13 .362 .152 r
F P
0 g
s
.69819 .32045 m .70052 .32547 L .68693 .33396 L closepath p .137 .399 .158 r
F P
0 g
s
.68693 .33396 m .70052 .32547 L .70718 .32995 L closepath p .135 .386 .155 r
F P
0 g
s
.68693 .33396 m .70718 .32995 L .69354 .33845 L closepath p .135 .386 .155 r
F P
0 g
s
.68693 .33396 m .69354 .33845 L .68737 .34859 L closepath p .132 .39 .159 r
F P
0 g
s
.68715 .48192 m .69138 .46152 L .7027 .46471 L closepath p .214 .723 .178 r
F P
0 g
s
.68715 .48192 m .7027 .46471 L .70209 .48299 L closepath p .213 .709 .176 r
F P
0 g
s
.70072 .49484 m .68715 .48192 L .70209 .48299 L closepath p .212 .71 .177 r
F P
0 g
s
.69354 .33845 m .70559 .34758 L .68737 .34859 L closepath p .139 .42 .163 r
F P
0 g
s
.70076 .35792 m .68737 .34859 L .70559 .34758 L closepath p .138 .4 .157 r
F P
0 g
s
.69668 .38185 m .68862 .37248 L .70424 .36725 L closepath p .152 .728 .233 r
F P
0 g
s
.6941 .50168 m .68913 .5014 L .69015 .50098 L closepath p .131 .866 .248 r
F P
0 g
s
.68955 .3125 m .69354 .30334 L .69816 .31268 L closepath p .13 .389 .161 r
F P
0 g
s
.68955 .3125 m .69816 .31268 L .69819 .32045 L closepath p .13 .377 .157 r
F P
0 g
s
.68955 .3125 m .69819 .32045 L .68957 .32025 L closepath p .13 .377 .157 r
F P
0 g
s
.6907 .49858 m .69464 .50128 L .69015 .50098 L closepath p .21 .758 .193 r
F P
0 g
s
.6941 .50168 m .69015 .50098 L .69464 .50128 L closepath p .133 .87 .248 r
F P
0 g
s
.70936 .28124 m .6902 .28375 L .70651 .27269 L closepath p .157 .382 .132 r
F P
0 g
s
.70936 .28124 m .69706 .28948 L .6902 .28375 L closepath p .164 .493 .163 r
F P
0 g
s
.6907 .49858 m .69519 .49888 L .69464 .50128 L closepath p .21 .758 .193 r
F P
0 g
s
.7005 .44121 m .7027 .46471 L .69138 .46152 L closepath p .21 .755 .192 r
F P
0 g
s
.70623 .24602 m .69235 .25535 L .70192 .24773 L closepath p 0 0 0 r
F P
0 g
s
.70623 .24602 m .69439 .25591 L .69235 .25535 L closepath p 0 0 0 r
F P
0 g
s
.69706 .28948 m .70085 .29692 L .69354 .30334 L closepath p .146 .408 .152 r
F P
0 g
s
.70085 .29692 m .69816 .31268 L .69354 .30334 L closepath p .148 .421 .155 r
F P
0 g
s
.71588 .33017 m .69354 .33845 L .70718 .32995 L closepath p .13 .348 .147 r
F P
0 g
s
.71588 .33017 m .71179 .33741 L .69354 .33845 L closepath p .139 .445 .17 r
F P
0 g
s
.69354 .33845 m .71179 .33741 L .70559 .34758 L closepath p .139 .42 .163 r
F P
0 g
s
.69945 .49964 m .6941 .50168 L .69464 .50128 L closepath p .129 .916 .212 r
F P
0 g
s
.69945 .49964 m .6989 .50004 L .6941 .50168 L closepath p .129 .916 .212 r
F P
0 g
s
.70623 .24602 m .70021 .25821 L .69439 .25591 L closepath p .125 .244 .115 r
F P
0 g
s
.69584 .26359 m .69439 .25591 L .70021 .25821 L closepath p .128 .216 .1 r
F P
0 g
s
.69464 .50128 m .69519 .49888 L .69945 .49964 L closepath p .222 .807 .191 r
F P
0 g
s
.69519 .49888 m .70072 .49484 L .70496 .49561 L closepath p .221 .809 .192 r
F P
0 g
s
.69519 .49888 m .70496 .49561 L .69945 .49964 L closepath p .221 .809 .192 r
F P
0 g
s
.69535 .41332 m .70432 .38771 L .71325 .39341 L closepath p .198 .796 .214 r
F P
0 g
s
.69535 .41332 m .71325 .39341 L .72176 .39547 L closepath p .199 .81 .215 r
F P
0 g
s
.69535 .41332 m .72176 .39547 L .71185 .4172 L closepath p .199 .81 .215 r
F P
0 g
s
.69584 .26359 m .70418 .26439 L .6958 .26895 L closepath p .126 .167 .083 r
F P
0 g
s
.70222 .26799 m .6958 .26895 L .70418 .26439 L closepath p .079 0 .022 r
F P
0 g
s
.69584 .26359 m .70786 .25753 L .7035 .26292 L closepath p .136 .255 .107 r
F P
0 g
s
.69584 .26359 m .70021 .25821 L .70786 .25753 L closepath p .136 .255 .107 r
F P
0 g
s
.69584 .26359 m .7035 .26292 L .70418 .26439 L closepath p .123 .126 .068 r
F P
0 g
s
.69668 .38185 m .70424 .36725 L .71707 .37011 L closepath p .172 .831 .238 r
F P
0 g
s
.69668 .38185 m .71707 .37011 L .72467 .37602 L closepath p .171 .804 .235 r
F P
0 g
s
.69668 .38185 m .72467 .37602 L .70432 .38771 L closepath p .171 .804 .235 r
F P
0 g
s
.69706 .28948 m .70936 .28124 L .71317 .28868 L closepath p .154 .422 .149 r
F P
0 g
s
.69706 .28948 m .71317 .28868 L .70085 .29692 L closepath p .154 .422 .149 r
F P
0 g
s
.70085 .29692 m .70548 .30626 L .69816 .31268 L closepath p .148 .421 .155 r
F P
0 g
s
.70548 .30626 m .69819 .32045 L .69816 .31268 L closepath p .152 .448 .16 r
F P
0 g
s
.70548 .30626 m .70528 .31682 L .69819 .32045 L closepath p .12 .302 .14 r
F P
0 g
s
.70052 .32547 m .69819 .32045 L .70528 .31682 L closepath p .131 .389 .159 r
F P
0 g
s
.70623 .24602 m .71207 .2483 L .70021 .25821 L closepath p .125 .244 .115 r
F P
0 g
s
.71207 .2483 m .70786 .25753 L .70021 .25821 L closepath p .143 .353 .136 r
F P
0 g
s
.7005 .44121 m .72392 .41654 L .72221 .44161 L closepath p .205 .767 .201 r
F P
0 g
s
.7005 .44121 m .71185 .4172 L .72392 .41654 L closepath p .206 .777 .202 r
F P
0 g
s
.7005 .44121 m .71179 .44447 L .7027 .46471 L closepath p .21 .755 .192 r
F P
0 g
s
.7005 .44121 m .72221 .44161 L .71179 .44447 L closepath p .202 .789 .208 r
F P
0 g
s
.70052 .32547 m .70528 .31682 L .70762 .32185 L closepath p .131 .389 .159 r
F P
0 g
s
.70052 .32547 m .70762 .32185 L .71219 .32268 L closepath p .131 .389 .159 r
F P
0 g
s
.70052 .32547 m .71219 .32268 L .70718 .32995 L closepath p .132 .391 .16 r
F P
0 g
s
.70072 .49484 m .71423 .48215 L .71502 .49292 L closepath p .215 .731 .179 r
F P
0 g
s
.70072 .49484 m .70209 .48299 L .71423 .48215 L closepath p .213 .713 .177 r
F P
0 g
s
.70072 .49484 m .71502 .49292 L .71051 .49488 L closepath p .213 .829 .206 r
F P
0 g
s
.70072 .49484 m .71051 .49488 L .70496 .49561 L closepath p .201 .883 .226 r
F P
0 g
s
.70076 .35792 m .71649 .34601 L .71165 .35637 L closepath p .127 .352 .151 r
F P
0 g
s
.70076 .35792 m .70559 .34758 L .71649 .34601 L closepath p .127 .352 .151 r
F P
0 g
s
.7153 .36163 m .70076 .35792 L .71165 .35637 L closepath p .134 .577 .21 r
F P
0 g
s
.7153 .36163 m .70424 .36725 L .70076 .35792 L closepath p .135 .575 .21 r
F P
0 g
s
.70548 .30626 m .70085 .29692 L .71317 .28868 L closepath p .156 .435 .151 r
F P
0 g
s
.72088 .23566 m .70623 .24602 L .70192 .24773 L closepath p 0 0 0 r
F P
0 g
s
.72088 .23566 m .70192 .24773 L .71653 .2374 L closepath p 0 0 0 r
F P
0 g
s
.70209 .48299 m .7027 .46471 L .71484 .46392 L closepath p .213 .71 .176 r
F P
0 g
s
.70209 .48299 m .71484 .46392 L .71423 .48215 L closepath p .213 .71 .176 r
F P
0 g
s
.70222 .26799 m .71972 .2549 L .72413 .25941 L closepath p .109 .157 .097 r
F P
0 g
s
.70222 .26799 m .72413 .25941 L .70651 .27269 L closepath p .126 .287 .13 r
F P
0 g
s
.70222 .26799 m .70418 .26439 L .70775 .26385 L p .079 0 .022 r
F P
0 g
s
.71179 .44447 m .71484 .46392 L .7027 .46471 L closepath p .212 .766 .192 r
F P
0 g
s
.70769 .49636 m .70315 .49693 L p .70867 .49621 L .035 .684 .184 r
F P
0 g
s
.70315 .49693 m .70867 .49621 L p .70315 .49693 L .035 .684 .184 r
F P
0 g
s
.71051 .49488 m .70867 .49621 L .70315 .49693 L p .135 .911 .238 r
F P
0 g
s
.70315 .49693 m .71051 .49488 L p .70315 .49693 L .135 .911 .238 r
F P
0 g
s
.70496 .49561 m .71051 .49488 L .70315 .49693 L p .135 .911 .238 r
F P
0 g
s
.7115 .26035 m .7035 .26292 L .70786 .25753 L closepath p .169 .424 .133 r
F P
0 g
s
.7035 .26292 m .7115 .26035 L .71174 .26086 L p .131 .14 .063 r
F P
0 g
s
.71012 .26208 m .7035 .26292 L p .71174 .26086 L .131 .14 .063 r
F P
0 g
s
.70418 .26439 m .7035 .26292 L .71012 .26208 L p .131 .14 .063 r
F P
0 g
s
.70917 .26279 m .70418 .26439 L p .71012 .26208 L .131 .14 .063 r
F P
0 g
s
.70775 .26385 m .70418 .26439 L .70917 .26279 L p .036 0 0 r
F P
0 g
s
.7153 .36163 m .71562 .36297 L .70424 .36725 L closepath p .143 .902 .245 r
F P
0 g
s
.71707 .37011 m .70424 .36725 L .71562 .36297 L closepath p .159 .835 .244 r
F P
0 g
s
.72467 .37602 m .71325 .39341 L .70432 .38771 L closepath p .171 .796 .233 r
F P
0 g
s
.70548 .30626 m .70828 .31078 L .70528 .31682 L closepath p .13 .334 .142 r
F P
0 g
s
.70828 .31078 m .71102 .31469 L .70528 .31682 L closepath p .107 .231 .127 r
F P
0 g
s
.70762 .32185 m .70528 .31682 L .71102 .31469 L closepath p .122 .371 .162 r
F P
0 g
s
.70548 .30626 m .71317 .28868 L .71401 .29483 L closepath p .139 .359 .142 r
F P
0 g
s
.70548 .30626 m .71401 .29483 L .71682 .29934 L closepath p .135 .34 .139 r
F P
0 g
s
.70548 .30626 m .71682 .29934 L .70828 .31078 L closepath p .135 .34 .139 r
F P
0 g
s
.71179 .33741 m .71649 .34601 L .70559 .34758 L closepath p .128 .369 .156 r
F P
0 g
s
.71207 .2483 m .70623 .24602 L .72088 .23566 L closepath p .124 .249 .117 r
F P
0 g
s
.70936 .28124 m .70651 .27269 L .72413 .25941 L closepath p .132 .329 .138 r
F P
0 g
s
.71588 .33017 m .70718 .32995 L .71219 .32268 L closepath p .13 .384 .159 r
F P
0 g
s
.70762 .32185 m .71102 .31469 L .7156 .31552 L closepath p .13 .408 .167 r
F P
0 g
s
.70762 .32185 m .7156 .31552 L .71219 .32268 L closepath p .13 .408 .167 r
F P
0 g
s
.70876 .49623 m .70769 .49636 L p .70867 .49621 L .035 .684 .184 r
F P
0 g
s
.70769 .49636 m .70876 .49623 L p .70769 .49636 L .214 .574 .127 r
F P
0 g
s
.71207 .2483 m .71636 .24963 L .70786 .25753 L closepath p .112 .218 .118 r
F P
0 g
s
.7115 .26035 m .70786 .25753 L .71636 .24963 L closepath p .151 .448 .16 r
F P
0 g
s
.71403 .30863 m .70828 .31078 L .71682 .29934 L closepath p .098 .16 .109 r
F P
0 g
s
.70828 .31078 m .71403 .30863 L .71102 .31469 L closepath p .107 .231 .127 r
F P
0 g
s
.71608 .49745 m .70867 .49621 L .71051 .49488 L closepath p .123 .751 .247 r
F P
0 g
s
.71608 .49745 m .71553 .49785 L .70867 .49621 L closepath p .123 .751 .247 r
F P
0 g
s
.70936 .28124 m .72413 .25941 L .72702 .26794 L closepath p .132 .329 .138 r
F P
0 g
s
.70936 .28124 m .72702 .26794 L .72201 .28313 L closepath p .145 .412 .154 r
F P
0 g
s
.70936 .28124 m .72201 .28313 L .71317 .28868 L closepath p .145 .407 .153 r
F P
0 g
s
.71608 .49745 m .71051 .49488 L .71502 .49292 L closepath p .208 .715 .183 r
F P
0 g
s
.71403 .30863 m .7156 .31552 L .71102 .31469 L closepath p .133 .348 .144 r
F P
0 g
s
.7115 .26035 m .71636 .24963 L .72 .25244 L closepath p .151 .448 .16 r
F P
0 g
s
.7115 .26035 m .72 .25244 L .72114 .25322 L p .046 0 .048 r
F P
0 g
s
.72071 .25373 m .7115 .26035 L p .72114 .25322 L .046 0 .048 r
F P
0 g
s
.7115 .26035 m .72071 .25373 L p .71908 .25537 L .076 .045 .087 r
F P
0 g
s
.71174 .26086 m .7115 .26035 L p .71908 .25537 L .076 .045 .087 r
F P
0 g
s
.71749 .35562 m .71165 .35637 L .71649 .34601 L closepath p .115 .297 .143 r
F P
0 g
s
.72114 .36089 m .71165 .35637 L .71749 .35562 L closepath p .122 .555 .213 r
F P
0 g
s
.72114 .36089 m .7153 .36163 L .71165 .35637 L closepath p .122 .555 .213 r
F P
0 g
s
.71179 .33741 m .71588 .33017 L .72471 .33119 L closepath p .132 .407 .165 r
F P
0 g
s
.71179 .33741 m .72471 .33119 L .72062 .33844 L closepath p .132 .407 .165 r
F P
0 g
s
.71179 .33741 m .72062 .33844 L .71649 .34601 L closepath p .133 .376 .154 r
F P
0 g
s
.71179 .44447 m .72221 .44161 L .72522 .46105 L closepath p .202 .741 .197 r
F P
0 g
s
.71179 .44447 m .72522 .46105 L .71484 .46392 L closepath p .202 .741 .197 r
F P
0 g
s
.72176 .39547 m .72392 .41654 L .71185 .4172 L closepath p .202 .822 .216 r
F P
0 g
s
.71207 .2483 m .72088 .23566 L .72497 .2372 L closepath p .078 .039 .082 r
F P
0 g
s
.71207 .2483 m .72497 .2372 L .72929 .23851 L closepath p .111 .222 .12 r
F P
0 g
s
.71207 .2483 m .72929 .23851 L .71636 .24963 L closepath p .111 .222 .12 r
F P
0 g
s
.7156 .31552 m .72091 .32289 L .71219 .32268 L closepath p .13 .409 .167 r
F P
0 g
s
.71588 .33017 m .71219 .32268 L .72091 .32289 L closepath p .13 .384 .159 r
F P
0 g
s
.72201 .28313 m .71401 .29483 L .71317 .28868 L closepath p .138 .357 .142 r
F P
0 g
s
.72467 .37602 m .72539 .3816 L .71325 .39341 L closepath p .19 .911 .235 r
F P
0 g
s
.71325 .39341 m .72539 .3816 L .73391 .38369 L closepath p .184 .838 .232 r
F P
0 g
s
.71325 .39341 m .73391 .38369 L .72176 .39547 L closepath p .184 .838 .232 r
F P
0 g
s
.72201 .28313 m .72287 .28928 L .71401 .29483 L closepath p .138 .357 .142 r
F P
0 g
s
.72287 .28928 m .71682 .29934 L .71401 .29483 L closepath p .136 .341 .139 r
F P
0 g
s
.71403 .30863 m .71682 .29934 L .71679 .30401 L closepath p .139 .316 .127 r
F P
0 g
s
.71403 .30863 m .71679 .30401 L .72084 .3048 L closepath p .127 .263 .119 r
F P
0 g
s
.71403 .30863 m .72084 .3048 L .71808 .30943 L closepath p .127 .263 .119 r
F P
0 g
s
.71403 .30863 m .71808 .30943 L .7156 .31552 L closepath p .126 .328 .144 r
F P
0 g
s
.71423 .48215 m .72522 .46105 L .72622 .47751 L closepath p .224 .807 .188 r
F P
0 g
s
.71423 .48215 m .71484 .46392 L .72522 .46105 L closepath p .199 .664 .179 r
F P
0 g
s
.72699 .48824 m .71423 .48215 L .72622 .47751 L closepath p .223 .751 .176 r
F P
0 g
s
.72699 .48824 m .71502 .49292 L .71423 .48215 L closepath p .223 .751 .176 r
F P
0 g
s
.72418 .49408 m .71502 .49292 L .72699 .48824 L closepath p .216 .658 .155 r
F P
0 g
s
.71608 .49745 m .71502 .49292 L .72058 .49549 L closepath p .208 .715 .183 r
F P
0 g
s
.72418 .49408 m .72058 .49549 L .71502 .49292 L closepath p .216 .658 .155 r
F P
0 g
s
.71839 .36157 m .7153 .36163 L .72114 .36089 L p .07 .759 .232 r
F P
0 g
s
.72871 .36309 m .7153 .36163 L .72593 .36141 L p .116 .805 .25 r
F P
0 g
s
.7153 .36163 m .72901 .36442 L .71562 .36297 L closepath p .105 .794 .248 r
F P
0 g
s
.7153 .36163 m .72871 .36309 L .72901 .36442 L closepath p .105 .794 .248 r
F P
0 g
s
.71553 .49785 m .71608 .49745 L .72095 .49796 L closepath p .132 .851 .249 r
F P
0 g
s
.71553 .49785 m .72095 .49796 L .7204 .49836 L closepath p .132 .851 .249 r
F P
0 g
s
.71982 .31674 m .7156 .31552 L .71808 .30943 L closepath p .146 .415 .155 r
F P
0 g
s
.7156 .31552 m .71982 .31674 L .72091 .32289 L closepath p .145 .425 .159 r
F P
0 g
s
.71707 .37011 m .71562 .36297 L .72844 .36585 L closepath p .159 .835 .244 r
F P
0 g
s
.72844 .36585 m .71562 .36297 L .72901 .36442 L closepath p .112 .801 .249 r
F P
0 g
s
.71588 .33017 m .72091 .32289 L .72586 .32036 L closepath p .143 .448 .168 r
F P
0 g
s
.71588 .33017 m .72586 .32036 L .73137 .3217 L closepath p .177 .673 .203 r
F P
0 g
s
.71588 .33017 m .73137 .3217 L .73142 .32826 L closepath p .159 .495 .169 r
F P
0 g
s
.71588 .33017 m .73142 .32826 L .72471 .33119 L closepath p .124 .107 .058 r
F P
0 g
s
.71608 .49745 m .72058 .49549 L .72095 .49796 L closepath p .207 .703 .181 r
F P
0 g
s
.73289 .23717 m .71636 .24963 L .72929 .23851 L closepath p .049 0 .049 r
F P
0 g
s
.73289 .23717 m .72 .25244 L .71636 .24963 L closepath p .144 .454 .169 r
F P
0 g
s
.72062 .33844 m .72023 .34681 L .71649 .34601 L closepath p .092 .19 .126 r
F P
0 g
s
.71749 .35562 m .71649 .34601 L .72023 .34681 L closepath p .09 .225 .14 r
F P
0 g
s
.72311 .29375 m .71679 .30401 L .71682 .29934 L closepath p .121 .259 .124 r
F P
0 g
s
.72311 .29375 m .72309 .29841 L .71679 .30401 L closepath p .121 .259 .124 r
F P
0 g
s
.72309 .29841 m .72084 .3048 L .71679 .30401 L closepath p .127 .293 .131 r
F P
0 g
s
.72287 .28928 m .72311 .29375 L .71682 .29934 L closepath p .132 .323 .136 r
F P
0 g
s
.72844 .36585 m .72467 .37602 L .71707 .37011 L closepath p .161 .794 .239 r
F P
0 g
s
.71749 .35562 m .72023 .34681 L .73721 .35086 L closepath p .111 .309 .151 r
F P
0 g
s
.71749 .35562 m .73721 .35086 L .73446 .35972 L closepath p .111 .309 .151 r
F P
0 g
s
.71749 .35562 m .73545 .36245 L .72114 .36089 L closepath p .08 .472 .214 r
F P
0 g
s
.71749 .35562 m .73446 .35972 L .73545 .36245 L closepath p .08 .472 .214 r
F P
0 g
s
.7223 .31064 m .71808 .30943 L .72084 .3048 L closepath p .15 .365 .134 r
F P
0 g
s
.71982 .31674 m .71808 .30943 L .7223 .31064 L closepath p .146 .415 .155 r
F P
0 g
s
.72593 .36141 m .71839 .36157 L p .72114 .36089 L .07 .759 .232 r
F P
0 g
s
.7202 .25433 m .71908 .25537 L p .72071 .25373 L .076 .045 .087 r
F P
0 g
s
.71908 .25537 m .7202 .25433 L p .71921 .25528 L 0 0 0 r
F P
0 g
s
.72015 .25439 m .71921 .25528 L p .7202 .25433 L 0 0 0 r
F P
0 g
s
.71921 .25528 m .72015 .25439 L p .71972 .2549 L 0 0 0 r
F P
0 g
s
.73812 .24292 m .71972 .2549 L .73366 .23846 L closepath p .093 .155 .113 r
F P
0 g
s
.73812 .24292 m .72413 .25941 L .71972 .2549 L closepath p .093 .155 .113 r
F P
0 g
s
.71982 .31674 m .7223 .31064 L .72477 .3142 L closepath p .131 .349 .147 r
F P
0 g
s
.72477 .3142 m .72091 .32289 L .71982 .31674 L closepath p .138 .406 .159 r
F P
0 g
s
.73299 .23925 m .72 .25244 L p .73366 .23846 L .034 0 .065 r
F P
0 g
s
.72 .25244 m .72779 .24321 L p .73366 .23846 L .034 0 .065 r
F P
0 g
s
.72114 .25322 m .72 .25244 L .73299 .23925 L p .034 0 .065 r
F P
0 g
s
.72062 .33844 m .72436 .33923 L .72023 .34681 L closepath p .092 .19 .126 r
F P
0 g
s
.72436 .33923 m .73721 .35086 L .72023 .34681 L closepath p .113 .284 .141 r
F P
0 g
s
.72095 .49796 m .72384 .49762 L .7204 .49836 L closepath p .133 .92 .226 r
F P
0 g
s
.72058 .49549 m .72418 .49408 L .72455 .49655 L closepath p .222 .74 .173 r
F P
0 g
s
.72058 .49549 m .72455 .49655 L .72095 .49796 L closepath p .222 .74 .173 r
F P
0 g
s
.72436 .33923 m .72062 .33844 L .72471 .33119 L closepath p .09 .214 .136 r
F P
0 g
s
.72309 .29841 m .72423 .30346 L .72084 .3048 L closepath p .137 .336 .136 r
F P
0 g
s
.7223 .31064 m .72084 .3048 L .72423 .30346 L closepath p .137 .335 .135 r
F P
0 g
s
.73163 .22618 m .72497 .2372 L .72088 .23566 L closepath p .078 .042 .084 r
F P
0 g
s
.72477 .3142 m .72586 .32036 L .72091 .32289 L closepath p .138 .406 .159 r
F P
0 g
s
.72095 .49796 m .72455 .49655 L .72384 .49762 L closepath p .206 .976 .223 r
F P
0 g
s
.73391 .38369 m .73386 .39487 L .72176 .39547 L closepath p .188 .88 .235 r
F P
0 g
s
.72176 .39547 m .73386 .39487 L .72392 .41654 L closepath p .202 .822 .216 r
F P
0 g
s
.72702 .26794 m .7334 .2843 L .72201 .28313 L closepath p .127 .34 .146 r
F P
0 g
s
.72287 .28928 m .72201 .28313 L .7334 .2843 L closepath p .128 .327 .142 r
F P
0 g
s
.73948 .41199 m .72221 .44161 L .72392 .41654 L closepath p .211 .789 .199 r
F P
0 g
s
.73948 .41199 m .73771 .437 L .72221 .44161 L closepath p .211 .789 .199 r
F P
0 g
s
.73771 .437 m .72522 .46105 L .72221 .44161 L closepath p .211 .763 .193 r
F P
0 g
s
.7223 .31064 m .72423 .30346 L .72494 .30806 L closepath p .121 .271 .129 r
F P
0 g
s
.72741 .31162 m .7223 .31064 L .72494 .30806 L closepath p .136 .355 .144 r
F P
0 g
s
.72741 .31162 m .72477 .3142 L .7223 .31064 L closepath p .136 .355 .144 r
F P
0 g
s
.72287 .28928 m .7334 .2843 L .73366 .28876 L closepath p .124 .298 .136 r
F P
0 g
s
.72287 .28928 m .73366 .28876 L .72311 .29375 L closepath p .124 .298 .136 r
F P
0 g
s
.72309 .29841 m .72311 .29375 L .73366 .28876 L closepath p .117 .247 .124 r
F P
0 g
s
.72309 .29841 m .73366 .28876 L .72649 .29707 L closepath p .147 .425 .157 r
F P
0 g
s
.72309 .29841 m .72649 .29707 L .72423 .30346 L closepath p .137 .336 .136 r
F P
0 g
s
.7273 .49372 m .72659 .49479 L .72384 .49762 L closepath p .184 .942 .155 r
F P
0 g
s
.7273 .49372 m .72384 .49762 L .72455 .49655 L closepath p .184 .942 .155 r
F P
0 g
s
.72392 .41654 m .74617 .3939 L .73948 .41199 L closepath p .207 .89 .222 r
F P
0 g
s
.72392 .41654 m .73386 .39487 L .74617 .3939 L closepath p .198 .805 .215 r
F P
0 g
s
.73812 .24292 m .72702 .26794 L .72413 .25941 L closepath p .123 .308 .14 r
F P
0 g
s
.72418 .49408 m .72699 .48824 L .7301 .48789 L closepath p .238 .79 .162 r
F P
0 g
s
.72418 .49408 m .7301 .48789 L .7273 .49372 L closepath p .238 .79 .162 r
F P
0 g
s
.72455 .49655 m .72418 .49408 L .7273 .49372 L closepath p .238 .775 .158 r
F P
0 g
s
.72919 .2998 m .72423 .30346 L .72649 .29707 L closepath p .126 .29 .13 r
F P
0 g
s
.72919 .2998 m .72494 .30806 L .72423 .30346 L closepath p .125 .283 .129 r
F P
0 g
s
.72436 .33923 m .72471 .33119 L .72923 .33536 L closepath p .159 .439 .15 r
F P
0 g
s
.72436 .33923 m .72923 .33536 L .74033 .34093 L closepath p .128 .252 .114 r
F P
0 g
s
.72436 .33923 m .73544 .34481 L .73721 .35086 L closepath p .121 .283 .133 r
F P
0 g
s
.72436 .33923 m .74033 .34093 L .73544 .34481 L closepath p .128 .252 .114 r
F P
0 g
s
.73654 .36362 m .72539 .3816 L .72467 .37602 L closepath p .202 .938 .228 r
F P
0 g
s
.72844 .36585 m .73654 .36362 L .72467 .37602 L closepath p .203 .969 .226 r
F P
0 g
s
.72923 .33536 m .72471 .33119 L .73142 .32826 L closepath p .172 .43 .132 r
F P
0 g
s
.72741 .31162 m .72586 .32036 L .72477 .3142 L closepath p .15 .439 .159 r
F P
0 g
s
.72919 .2998 m .7299 .3044 L .72494 .30806 L closepath p .125 .283 .129 r
F P
0 g
s
.72741 .31162 m .72494 .30806 L .7299 .3044 L closepath p .136 .356 .144 r
F P
0 g
s
.73163 .22618 m .73576 .22769 L .72497 .2372 L closepath p .078 .042 .084 r
F P
0 g
s
.73576 .22769 m .72929 .23851 L .72497 .2372 L closepath p .111 .225 .121 r
F P
0 g
s
.73771 .437 m .7343 .4566 L .72522 .46105 L closepath p .221 .823 .195 r
F P
0 g
s
.72622 .47751 m .72522 .46105 L .7343 .4566 L closepath p .22 .796 .19 r
F P
0 g
s
.73654 .36362 m .73721 .36926 L .72539 .3816 L closepath p .202 .938 .228 r
F P
0 g
s
.73721 .36926 m .73391 .38369 L .72539 .3816 L closepath p .193 .826 .224 r
F P
0 g
s
.72741 .31162 m .73003 .31138 L .72586 .32036 L closepath p .121 .332 .15 r
F P
0 g
s
.73554 .31273 m .72586 .32036 L .73003 .31138 L closepath p .185 .626 .183 r
F P
0 g
s
.73554 .31273 m .73137 .3217 L .72586 .32036 L closepath p .185 .626 .183 r
F P
0 g
s
.72955 .3618 m .72871 .36309 L p .72593 .36141 L .116 .805 .25 r
F P
0 g
s
.72622 .47751 m .7343 .4566 L .73527 .47303 L closepath p .22 .796 .19 r
F P
0 g
s
.72699 .48824 m .72622 .47751 L .73527 .47303 L closepath p .217 .736 .179 r
F P
0 g
s
.73637 .29149 m .72649 .29707 L .73366 .28876 L closepath p .122 .29 .134 r
F P
0 g
s
.73637 .29149 m .72919 .2998 L .72649 .29707 L closepath p .122 .29 .134 r
F P
0 g
s
.72699 .48824 m .73527 .47303 L .73725 .48178 L closepath p .225 .785 .181 r
F P
0 g
s
.72699 .48824 m .73725 .48178 L .7341 .48542 L closepath p .228 .915 .202 r
F P
0 g
s
.72699 .48824 m .7341 .48542 L .7301 .48789 L closepath p .224 .941 .208 r
F P
0 g
s
.73812 .24292 m .73842 .26906 L .72702 .26794 L closepath p .128 .329 .143 r
F P
0 g
s
.72702 .26794 m .73842 .26906 L .7334 .2843 L closepath p .127 .34 .146 r
F P
0 g
s
.73003 .31138 m .72741 .31162 L .7299 .3044 L closepath p .118 .282 .135 r
F P
0 g
s
.72779 .24321 m .73289 .23717 L .73408 .23795 L p .034 0 .065 r
F P
0 g
s
.73366 .23846 m .72779 .24321 L .73408 .23795 L .034 0 .065 r
F
.73398 .36228 m .72844 .36585 L p .73377 .36226 L .163 .969 .222 r
F P
0 g
s
.72844 .36585 m .72901 .36442 L .73341 .36222 L p .163 .969 .222 r
F P
0 g
s
.73348 .36223 m .72844 .36585 L .73341 .36222 L .163 .969 .222 r
F
.73377 .36226 m .72844 .36585 L .73348 .36223 L F
.73528 .36243 m .72844 .36585 L .73398 .36228 L p F P
0 g
s
.73654 .36362 m .72844 .36585 L .73528 .36243 L p .178 .977 .229 r
F P
0 g
s
.7341 .48542 m .73176 .48898 L .72865 .4909 L p .206 .996 .212 r
F P
0 g
s
.72892 .49035 m .7341 .48542 L p .72865 .4909 L .206 .996 .212 r
F P
0 g
s
.72871 .36309 m .72955 .3618 L p .7308 .36194 L .148 .938 .234 r
F P
0 g
s
.73113 .36198 m .72871 .36309 L p .73084 .36194 L .148 .938 .234 r
F P
0 g
s
.72871 .36309 m .7308 .36194 L .73084 .36194 L .148 .938 .234 r
F
.72871 .36309 m .73084 .36194 L .73084 .36194 L F
.73201 .36207 m .72871 .36309 L .73113 .36198 L p F P
0 g
s
.72901 .36442 m .72871 .36309 L .73201 .36207 L p .179 .977 .229 r
F P
0 g
s
.73201 .36207 m .72871 .36309 L .73201 .36207 L p .052 .722 .169 r
F P
0 g
s
.7301 .48789 m .7341 .48542 L .72892 .49035 L p .206 .996 .212 r
F P
0 g
s
.73201 .36207 m .73231 .3621 L .72901 .36442 L .179 .977 .229 r
F
.73231 .3621 m .7331 .36219 L .72901 .36442 L F
.73341 .36222 m .72901 .36442 L p .7331 .36219 L F P
0 g
s
.72919 .2998 m .73637 .29149 L .73252 .30416 L closepath p .117 .261 .129 r
F P
0 g
s
.72919 .2998 m .73252 .30416 L .7299 .3044 L closepath p .117 .261 .129 r
F P
0 g
s
.72923 .33536 m .73142 .32826 L .73593 .33243 L closepath p .172 .43 .132 r
F P
0 g
s
.73593 .33243 m .74033 .34093 L .72923 .33536 L closepath p .137 .187 .077 r
F P
0 g
s
.73576 .22769 m .73899 .22784 L .72929 .23851 L closepath p .065 .022 .088 r
F P
0 g
s
.73289 .23717 m .72929 .23851 L .73899 .22784 L closepath p .068 .038 .091 r
F P
0 g
s
.73003 .31138 m .7299 .3044 L .73252 .30416 L closepath p .118 .282 .135 r
F P
0 g
s
.73554 .31273 m .73003 .31138 L .73252 .30416 L closepath p .189 .578 .164 r
F P
0 g
s
.73554 .31273 m .74169 .32059 L .73137 .3217 L closepath p .202 .715 .19 r
F P
0 g
s
.73137 .3217 m .74169 .32059 L .74174 .32714 L closepath p .201 .633 .167 r
F P
0 g
s
.73137 .3217 m .74174 .32714 L .73142 .32826 L closepath p .201 .633 .167 r
F P
0 g
s
.73593 .33243 m .73142 .32826 L .74174 .32714 L closepath p .184 .407 .105 r
F P
0 g
s
.73576 .22769 m .73163 .22618 L .74047 .21875 L closepath p .077 .052 .089 r
F P
0 g
s
.7393 .48628 m .73176 .48898 L .7341 .48542 L closepath p .182 .812 .229 r
F P
0 g
s
.7393 .48628 m .73859 .48735 L .73176 .48898 L closepath p .182 .812 .229 r
F P
0 g
s
.74084 .29164 m .73252 .30416 L .73637 .29149 L closepath p .169 .469 .15 r
F P
0 g
s
.74084 .29164 m .737 .30433 L .73252 .30416 L closepath p .169 .469 .15 r
F P
0 g
s
.73554 .31273 m .73252 .30416 L .737 .30433 L closepath p .17 .538 .172 r
F P
0 g
s
.73289 .23717 m .74262 .22647 L .73558 .23613 L p .038 0 .06 r
F P
0 g
s
.73289 .23717 m .73899 .22784 L .74262 .22647 L closepath p .068 .038 .091 r
F P
0 g
s
.73408 .23795 m .73289 .23717 L p .73558 .23613 L .038 0 .06 r
F P
0 g
s
.73842 .26906 m .73366 .28876 L .7334 .2843 L closepath p .098 .222 .132 r
F P
0 g
s
.73842 .26906 m .73637 .29149 L .73366 .28876 L closepath p .116 .289 .139 r
F P
0 g
s
.73812 .24292 m .73366 .23846 L .74334 .2267 L closepath p .09 .155 .115 r
F P
0 g
s
.73386 .39487 m .74622 .38134 L .75858 .3804 L closepath p .178 .887 .24 r
F P
0 g
s
.73386 .39487 m .75858 .3804 L .74617 .3939 L closepath p .178 .887 .24 r
F P
0 g
s
.73391 .38369 m .74622 .38134 L .73386 .39487 L closepath p .17 .826 .238 r
F P
0 g
s
.73721 .36926 m .74302 .37071 L .73391 .38369 L closepath p .193 .826 .224 r
F P
0 g
s
.74622 .38134 m .73391 .38369 L .74302 .37071 L closepath p .178 .734 .217 r
F P
0 g
s
.7393 .48628 m .7341 .48542 L .73725 .48178 L closepath p .205 .7 .183 r
F P
0 g
s
.73771 .437 m .74095 .45382 L .7343 .4566 L closepath p .221 .823 .195 r
F P
0 g
s
.73527 .47303 m .7343 .4566 L .74095 .45382 L closepath p .221 .797 .19 r
F P
0 g
s
.76717 .36219 m .73446 .35972 L .73721 .35086 L closepath p .114 .32 .152 r
F P
0 g
s
.73446 .35972 m .76374 .36491 L .76473 .36766 L closepath p .085 .487 .215 r
F P
0 g
s
.73446 .35972 m .76473 .36766 L .73545 .36245 L closepath p .085 .487 .215 r
F P
0 g
s
.76717 .36219 m .76374 .36491 L .73446 .35972 L closepath p .112 .201 .111 r
F P
0 g
s
.73527 .47303 m .74095 .45382 L .7426 .46547 L closepath p .235 .865 .185 r
F P
0 g
s
.74456 .47421 m .73527 .47303 L .7426 .46547 L closepath p .229 .793 .177 r
F P
0 g
s
.74456 .47421 m .73725 .48178 L .73527 .47303 L closepath p .229 .793 .177 r
F P
0 g
s
.73528 .36243 m .73531 .36243 L .73654 .36362 L .178 .977 .229 r
F
.73601 .36254 m .73654 .36362 L p .73531 .36243 L F P
0 g
s
.73545 .36245 m .73601 .36254 L .73531 .36243 L .178 .977 .229 r
F
.73544 .34481 m .74033 .34093 L .76532 .35606 L closepath p .126 .237 .111 r
F P
0 g
s
.73544 .34481 m .76717 .36219 L .73721 .35086 L closepath p .12 .281 .133 r
F P
0 g
s
.73544 .34481 m .76532 .35606 L .76717 .36219 L closepath p .12 .281 .133 r
F P
0 g
s
.74353 .31328 m .73554 .31273 L .737 .30433 L closepath p .198 .648 .176 r
F P
0 g
s
.73554 .31273 m .74353 .31328 L .74169 .32059 L closepath p .195 .711 .197 r
F P
0 g
s
.73558 .23613 m .74235 .22683 L p .74334 .2267 L 0 0 .011 r
F P
0 g
s
.73576 .22769 m .74047 .21875 L .74243 .21856 L closepath p 0 0 .021 r
F P
0 g
s
.73576 .22769 m .74243 .21856 L .74568 .2187 L closepath p .066 .063 .102 r
F P
0 g
s
.73576 .22769 m .74568 .2187 L .73899 .22784 L closepath p .066 .063 .102 r
F P
0 g
s
.73593 .33243 m .74174 .32714 L .74342 .33294 L closepath p .207 .549 .129 r
F P
0 g
s
.73593 .33243 m .74841 .34167 L .74033 .34093 L closepath p .143 .197 .073 r
F P
0 g
s
.73593 .33243 m .75592 .34217 L .74841 .34167 L closepath p .156 .103 0 r
F P
0 g
s
.73593 .33243 m .74342 .33294 L .75592 .34217 L closepath p .156 .103 0 r
F P
0 g
s
.73654 .36362 m .73601 .36254 L p .73676 .36268 L .219 .989 .183 r
F P
0 g
s
.73842 .26906 m .74084 .29164 L .73637 .29149 L closepath p .169 .471 .15 r
F P
0 g
s
.73711 .36274 m .73654 .36362 L p .73679 .36268 L .219 .989 .183 r
F P
0 g
s
.73679 .36268 m .73654 .36362 L .73676 .36268 L .219 .989 .183 r
F
.73721 .36926 m .73654 .36362 L .73711 .36274 L p .227 .973 .2 r
F P
0 g
s
.74084 .29164 m .74498 .30489 L .737 .30433 L closepath p .198 .595 .158 r
F P
0 g
s
.74353 .31328 m .737 .30433 L .74498 .30489 L closepath p .198 .648 .176 r
F P
0 g
s
.7381 .36292 m .73721 .36926 L .73711 .36274 L .227 .973 .2 r
F
.73943 .36315 m .73721 .36926 L p .7381 .36292 L F P
0 g
s
.74031 .36331 m .73721 .36926 L .73943 .36315 L p .228 .982 .183 r
F P
0 g
s
.74253 .36371 m .73721 .36926 L .74031 .36331 L p .205 .791 .206 r
F P
0 g
s
.74302 .37071 m .73721 .36926 L .74253 .36371 L p .205 .791 .206 r
F P
0 g
s
.74464 .48037 m .73725 .48178 L .74456 .47421 L closepath p .208 .62 .154 r
F P
0 g
s
.74464 .48037 m .74244 .48265 L .73725 .48178 L closepath p .208 .62 .154 r
F P
0 g
s
.7393 .48628 m .73725 .48178 L .74244 .48265 L closepath p .205 .7 .183 r
F P
0 g
s
.75044 .42385 m .73771 .437 L .73948 .41199 L closepath p .231 .853 .189 r
F P
0 g
s
.75044 .42385 m .74436 .43425 L .73771 .437 L closepath p .217 .722 .174 r
F P
0 g
s
.73771 .437 m .74436 .43425 L .74095 .45382 L closepath p .221 .823 .195 r
F P
0 g
s
.73812 .24292 m .74334 .2267 L .74484 .22649 L closepath p 0 0 .056 r
F P
0 g
s
.74234 .24285 m .73812 .24292 L .74484 .22649 L closepath p .073 .115 .116 r
F P
0 g
s
.73812 .24292 m .74234 .24285 L .74669 .24889 L closepath p .075 .288 .169 r
F P
0 g
s
.73812 .24292 m .74669 .24889 L .73842 .26906 L closepath p .107 .264 .139 r
F P
0 g
s
.73842 .26906 m .74669 .24889 L .75847 .248 L closepath p .131 .366 .152 r
F P
0 g
s
.73842 .26906 m .75847 .248 L .75018 .26823 L closepath p .131 .366 .152 r
F P
0 g
s
.73842 .26906 m .75018 .26823 L .75338 .28004 L closepath p .134 .515 .195 r
F P
0 g
s
.73842 .26906 m .74652 .29239 L .74084 .29164 L closepath p .213 .6 .14 r
F P
0 g
s
.73842 .26906 m .75338 .28004 L .74652 .29239 L closepath p .163 .5 .166 r
F P
0 g
s
.73859 .48735 m .74354 .48484 L .74283 .48591 L closepath p .199 .918 .23 r
F P
0 g
s
.73859 .48735 m .7393 .48628 L .74354 .48484 L closepath p .199 .918 .23 r
F P
0 g
s
.73899 .22784 m .74568 .2187 L .74845 .21921 L p .112 .289 .143 r
F P
0 g
s
.74762 .22042 m .73899 .22784 L p .74842 .21925 L .112 .289 .143 r
F P
0 g
s
.74842 .21925 m .73899 .22784 L .74845 .21921 L .112 .289 .143 r
F
.74262 .22647 m .73899 .22784 L .74762 .22042 L p .068 .033 .09 r
F P
0 g
s
.7393 .48628 m .74244 .48265 L .74354 .48484 L closepath p .204 .693 .183 r
F P
0 g
s
.75358 .40815 m .73948 .41199 L .74617 .3939 L closepath p .202 .869 .223 r
F P
0 g
s
.75044 .42385 m .73948 .41199 L .75358 .40815 L closepath p .203 .859 .221 r
F P
0 g
s
.76532 .35606 m .74033 .34093 L .74841 .34167 L closepath p .141 .166 .062 r
F P
0 g
s
.74243 .21856 m .74047 .21875 L .74827 .21036 L closepath p 0 0 0 r
F P
0 g
s
.74498 .30489 m .74084 .29164 L .74652 .29239 L closepath p .213 .625 .148 r
F P
0 g
s
.74436 .43425 m .74747 .44648 L .74095 .45382 L closepath p .234 .874 .187 r
F P
0 g
s
.7426 .46547 m .74095 .45382 L .74747 .44648 L closepath p .234 .863 .187 r
F P
0 g
s
.74169 .32059 m .74353 .31328 L .75165 .31635 L closepath p .209 .771 .197 r
F P
0 g
s
.74169 .32059 m .75165 .31635 L .74981 .32363 L closepath p .209 .771 .197 r
F P
0 g
s
.74169 .32059 m .74574 .33222 L .74174 .32714 L closepath p .221 .699 .162 r
F P
0 g
s
.74169 .32059 m .74604 .33143 L .74574 .33222 L closepath p .222 .562 .016 r
F P
0 g
s
.74169 .32059 m .74981 .32363 L .74604 .33143 L closepath p .214 .75 .186 r
F P
0 g
s
.74342 .33294 m .74174 .32714 L .74574 .33222 L closepath p .236 .604 .084 r
F P
0 g
s
.74234 .24285 m .74484 .22649 L .74908 .22638 L closepath p .073 .115 .116 r
F P
0 g
s
.74234 .24285 m .74908 .22638 L .7516 .22653 L closepath p .063 .078 .11 r
F P
0 g
s
.74234 .24285 m .7516 .22653 L .7516 .23213 L closepath p .119 .33 .151 r
F P
0 g
s
.74669 .24889 m .74234 .24285 L .7516 .23213 L closepath p .122 .349 .154 r
F P
0 g
s
.74235 .22683 m .74262 .22647 L p .74339 .22663 L 0 0 .011 r
F P
0 g
s
.74235 .22683 m .74335 .22669 L .74334 .2267 L 0 0 .011 r
F
.74235 .22683 m .74338 .22665 L .74335 .22669 L F
.74235 .22683 m .74339 .22663 L .74338 .22665 L F
.74243 .21856 m .74827 .21036 L .75025 .21015 L closepath p 0 0 0 r
F P
0 g
s
.75025 .21015 m .74568 .2187 L .74243 .21856 L closepath p .065 .008 .083 r
F P
0 g
s
.74244 .48265 m .74464 .48037 L .74573 .48256 L closepath p .22 .71 .167 r
F P
0 g
s
.74244 .48265 m .74573 .48256 L .74354 .48484 L closepath p .22 .71 .167 r
F P
0 g
s
.74253 .36371 m .74377 .36393 L .74302 .37071 L .205 .791 .206 r
F
.7426 .46547 m .74747 .44648 L .74909 .45813 L closepath p .234 .863 .187 r
F P
0 g
s
.74456 .47421 m .7426 .46547 L .74909 .45813 L closepath p .227 .789 .179 r
F P
0 g
s
.74746 .22066 m .74262 .22647 L .74762 .22042 L .068 .033 .09 r
F
.74652 .22204 m .74262 .22647 L p .74746 .22066 L F P
0 g
s
.74262 .22647 m .74652 .22204 L p .74556 .22344 L 0 0 .002 r
F P
0 g
s
.74262 .22647 m .74556 .22344 L .74546 .2236 L 0 0 .002 r
F
.7436 .22632 m .74262 .22647 L .74524 .22392 L p F P
0 g
s
.74524 .22392 m .74262 .22647 L p .74546 .2236 L 0 0 .002 r
F P
0 g
s
.74262 .22647 m .7436 .22632 L p .74339 .22663 L 0 0 .011 r
F P
0 g
s
.74354 .48484 m .74556 .48324 L .74283 .48591 L closepath p .194 .977 .176 r
F P
0 g
s
.74632 .36438 m .74302 .37071 L p .74377 .36393 L .205 .791 .206 r
F P
0 g
s
.75538 .36835 m .74302 .37071 L .74632 .36438 L p .178 .646 .196 r
F P
0 g
s
.74622 .38134 m .74302 .37071 L .75538 .36835 L closepath p .178 .734 .217 r
F P
0 g
s
.75034 .21843 m .74484 .22649 L .74334 .2267 L closepath p 0 0 .03 r
F P
0 g
s
.75034 .21843 m .74334 .2267 L .74883 .21865 L closepath p 0 0 .03 r
F P
0 g
s
.74342 .33294 m .74574 .33222 L .74664 .33532 L p .236 .604 .084 r
F P
0 g
s
.74353 .31328 m .74498 .30489 L .75066 .30562 L closepath p .213 .706 .176 r
F P
0 g
s
.74353 .31328 m .75066 .30562 L .75165 .31635 L closepath p .216 .736 .179 r
F P
0 g
s
.74354 .48484 m .74573 .48256 L .74556 .48324 L closepath p .211 .988 .18 r
F P
0 g
s
.74436 .43425 m .75044 .42385 L .75351 .4361 L closepath p .238 .877 .181 r
F P
0 g
s
.74436 .43425 m .75351 .4361 L .74747 .44648 L closepath p .238 .877 .181 r
F P
0 g
s
.74456 .47421 m .74909 .45813 L .75127 .46554 L closepath p .236 .833 .177 r
F P
0 g
s
.74456 .47421 m .75127 .46554 L .74913 .46909 L closepath p .22 .974 .167 r
F P
0 g
s
.74456 .47421 m .74913 .46909 L .74674 .47249 L closepath p .243 .904 .171 r
F P
0 g
s
.74464 .48037 m .74456 .47421 L .74674 .47249 L closepath p .235 .713 .141 r
F P
0 g
s
.74464 .48037 m .74674 .47249 L .74682 .47864 L closepath p .235 .713 .141 r
F P
0 g
s
.74573 .48256 m .74464 .48037 L .74682 .47864 L closepath p .235 .712 .141 r
F P
0 g
s
.75404 .2177 m .74484 .22649 L .75034 .21843 L closepath p .085 .158 .121 r
F P
0 g
s
.75404 .2177 m .74908 .22638 L .74484 .22649 L closepath p .072 .086 .106 r
F P
0 g
s
.74498 .30489 m .74652 .29239 L .75066 .30562 L closepath p .213 .625 .148 r
F P
0 g
s
.74682 .47864 m .74666 .47932 L .74556 .48324 L closepath p .176 .727 .052 r
F P
0 g
s
.74682 .47864 m .74556 .48324 L .74573 .48256 L closepath p .176 .727 .052 r
F P
0 g
s
.75025 .21015 m .75138 .21038 L .74568 .2187 L closepath p 0 0 .02 r
F P
0 g
s
.74568 .2187 m .75138 .21038 L .74934 .21783 L p .113 .277 .138 r
F P
0 g
s
.74932 .21785 m .74568 .2187 L .74934 .21783 L .113 .277 .138 r
F
.74845 .21921 m .74568 .2187 L p .74925 .21797 L F P
0 g
s
.74925 .21797 m .74568 .2187 L .74932 .21785 L .113 .277 .138 r
F
.74574 .33222 m .74604 .33143 L .74601 .33318 L p .212 .693 .04 r
F P
0 g
s
.74601 .33318 m .74604 .33143 L .74606 .33333 L p .212 .693 .04 r
F P
0 g
s
.74981 .32363 m .74901 .33662 L .74604 .33143 L closepath p .229 .836 .188 r
F P
0 g
s
.74604 .33143 m .74901 .33662 L .74886 .33696 L p .235 .834 .178 r
F P
0 g
s
.74604 .33143 m .74886 .33696 L .74742 .33589 L .235 .834 .178 r
F
.74604 .33143 m .74695 .33555 L .74606 .33306 L F
.74604 .33143 m .74722 .33574 L .74695 .33555 L F
.74604 .33143 m .74742 .33589 L .74722 .33574 L F
.74673 .33538 m .74606 .33306 L .74606 .33306 L F
.74695 .33555 m .74673 .33538 L .74606 .33306 L F
.74673 .33538 m .74606 .33306 L .74606 .33306 L F
.74673 .33538 m .74606 .33306 L .74673 .33538 L F
.74673 .33538 m .74606 .33306 L .74606 .33306 L F
.74673 .33538 m .74673 .33538 L .74606 .33306 L F
.74673 .33538 m .74606 .33306 L .74673 .33538 L F
.74673 .33538 m .74606 .33307 L .74606 .33306 L F
.74673 .33538 m .74606 .33306 L .74673 .33538 L F
.74606 .33307 m .74669 .33535 L .74606 .33333 L F
.74672 .33538 m .74669 .33535 L .74606 .33307 L F
.74672 .33538 m .74606 .33307 L .74673 .33538 L F
.74669 .33535 m .74664 .33532 L .74606 .33333 L F
.76034 .39008 m .74617 .3939 L .75858 .3804 L closepath p .183 .941 .238 r
F P
0 g
s
.75358 .40815 m .74617 .3939 L .76034 .39008 L closepath p .202 .869 .223 r
F P
0 g
s
.7639 .37471 m .74622 .38134 L .75538 .36835 L closepath p .174 .709 .215 r
F P
0 g
s
.7639 .37471 m .75858 .3804 L .74622 .38134 L closepath p .165 .9 .245 r
F P
0 g
s
.75026 .36508 m .75538 .36835 L .74632 .36438 L .178 .646 .196 r
F
.75338 .28004 m .75969 .30722 L .74652 .29239 L closepath p .196 .672 .185 r
F P
0 g
s
.75969 .30722 m .75066 .30562 L .74652 .29239 L closepath p .224 .644 .137 r
F P
0 g
s
.74669 .24889 m .7516 .23213 L .75597 .23816 L closepath p .122 .349 .154 r
F P
0 g
s
.74669 .24889 m .75597 .23816 L .76092 .23768 L closepath p .136 .425 .167 r
F P
0 g
s
.74669 .24889 m .76092 .23768 L .75847 .248 L closepath p .132 .398 .162 r
F P
0 g
s
.74674 .47249 m .74913 .46909 L .74676 .47365 L p .209 .945 .147 r
F P
0 g
s
.74676 .47365 m .74913 .46909 L p .74676 .47393 L .209 .945 .147 r
F P
0 g
s
.74913 .46909 m .74857 .47135 L .74676 .47393 L p .209 .945 .147 r
F P
0 g
s
.74909 .45813 m .74747 .44648 L .75351 .4361 L closepath p .238 .87 .181 r
F P
0 g
s
.75571 .20211 m .75025 .21015 L .74827 .21036 L closepath p 0 0 .027 r
F P
0 g
s
.76532 .35606 m .74841 .34167 L .77352 .35682 L closepath p .141 .166 .062 r
F P
0 g
s
.74841 .34167 m .75592 .34217 L .77526 .35441 L closepath p .132 0 0 r
F P
0 g
s
.74841 .34167 m .77526 .35441 L .77352 .35682 L closepath p .14 .17 .064 r
F P
0 g
s
.74845 .21921 m .74925 .21797 L .74915 .21812 L .113 .277 .138 r
F
.74845 .21921 m .74915 .21812 L .74883 .21865 L F
.75396 .46843 m .74857 .47135 L .74913 .46909 L closepath p .184 .944 .237 r
F P
0 g
s
.75396 .46843 m .75379 .46911 L .74857 .47135 L closepath p .184 .944 .237 r
F P
0 g
s
.74883 .21865 m .75636 .20967 L .75034 .21843 L closepath p 0 0 .024 r
F P
0 g
s
.75416 .2101 m .75636 .20967 L .74883 .21865 L p .012 0 .057 r
F P
0 g
s
.74883 .21865 m .75409 .21012 L .75416 .2101 L p .012 0 .057 r
F P
0 g
s
.74883 .21865 m .74883 .21865 L p .74883 .21865 L .23 .972 .193 r
F P
0 g
s
.74886 .33696 m .74901 .33662 L p .74903 .33708 L .239 .863 .177 r
F P
0 g
s
.74981 .32363 m .75278 .32884 L .74901 .33662 L closepath p .229 .836 .188 r
F P
0 g
s
.75278 .32884 m .75136 .33769 L .74901 .33662 L closepath p .235 .881 .187 r
F P
0 g
s
.75073 .33834 m .74922 .33722 L .74901 .33662 L .239 .863 .177 r
F
.74922 .33722 m .74903 .33708 L .74901 .33662 L F
.74901 .33662 m .75136 .33769 L .75073 .33834 L p F P
0 g
s
.74908 .22638 m .75404 .2177 L .75657 .21784 L closepath p .063 .049 .1 r
F P
0 g
s
.74908 .22638 m .75657 .21784 L .7516 .22653 L closepath p .063 .049 .1 r
F P
0 g
s
.74909 .45813 m .75351 .4361 L .75514 .4495 L closepath p .237 .866 .182 r
F P
0 g
s
.7573 .45692 m .74909 .45813 L .75514 .4495 L closepath p .233 .831 .18 r
F P
0 g
s
.7573 .45692 m .75127 .46554 L .74909 .45813 L closepath p .233 .831 .18 r
F P
0 g
s
.75396 .46843 m .74913 .46909 L .75127 .46554 L closepath p .218 .805 .196 r
F P
0 g
s
.74934 .21783 m .75138 .21038 L p .74978 .2171 L .113 .278 .138 r
F P
0 g
s
.74978 .2171 m .75138 .21038 L .75198 .21354 L .113 .278 .138 r
F
.75646 .32909 m .74981 .32363 L .75165 .31635 L closepath p .216 .801 .196 r
F P
0 g
s
.75646 .32909 m .75278 .32884 L .74981 .32363 L closepath p .197 .81 .217 r
F P
0 g
s
.75847 .248 m .76529 .25281 L .75018 .26823 L closepath p .156 .478 .165 r
F P
0 g
s
.76848 .26464 m .75018 .26823 L .76529 .25281 L closepath p .181 .632 .189 r
F P
0 g
s
.76848 .26464 m .75338 .28004 L .75018 .26823 L closepath p .181 .632 .189 r
F P
0 g
s
.75025 .21015 m .75571 .20211 L .75685 .20232 L closepath p 0 0 .024 r
F P
0 g
s
.75025 .21015 m .75685 .20232 L .75138 .21038 L closepath p 0 0 .024 r
F P
0 g
s
.75471 .36587 m .75538 .36835 L p .75026 .36508 L .178 .646 .196 r
F P
0 g
s
.75404 .2177 m .75034 .21843 L .75636 .20967 L closepath p .082 .126 .112 r
F P
0 g
s
.76263 .394 m .75044 .42385 L .75358 .40815 L closepath p .235 .966 .179 r
F P
0 g
s
.76263 .394 m .75945 .40975 L .75044 .42385 L closepath p .235 .966 .179 r
F P
0 g
s
.75945 .40975 m .75351 .4361 L .75044 .42385 L closepath p .234 .874 .188 r
F P
0 g
s
.75066 .30562 m .76067 .31792 L .75165 .31635 L closepath p .223 .753 .176 r
F P
0 g
s
.75066 .30562 m .75969 .30722 L .76067 .31792 L closepath p .223 .753 .176 r
F P
0 g
s
.75073 .33834 m .75136 .33769 L p .75079 .33839 L .239 .863 .177 r
F P
0 g
s
.75136 .33769 m .75097 .33852 L p .75079 .33839 L .239 .863 .177 r
F P
0 g
s
.75097 .33852 m .75136 .33769 L p .7511 .33861 L .209 .68 .173 r
F P
0 g
s
.7511 .33861 m .75136 .33769 L .75112 .33863 L .209 .68 .173 r
F
.75112 .33863 m .75136 .33769 L .75119 .33868 L F
.75119 .33868 m .75136 .33769 L .75127 .33874 L F
.75127 .33874 m .75136 .33769 L .75196 .33925 L F
.75791 .4623 m .75127 .46554 L .7573 .45692 L closepath p .218 .713 .17 r
F P
0 g
s
.75791 .4623 m .7561 .46489 L .75127 .46554 L closepath p .218 .713 .17 r
F P
0 g
s
.75396 .46843 m .75127 .46554 L .7561 .46489 L closepath p .218 .805 .196 r
F P
0 g
s
.75278 .32884 m .75504 .33793 L .75136 .33769 L closepath p .204 .764 .201 r
F P
0 g
s
.75136 .33769 m .75504 .33793 L .75294 .33997 L p .209 .68 .173 r
F P
0 g
s
.75196 .33925 m .75136 .33769 L .75294 .33997 L .209 .68 .173 r
F
.75759 .203 m .75138 .21038 L .75799 .20218 L .164 .557 .182 r
F
.75138 .21038 m .75685 .20232 L .75799 .20218 L p F P
0 g
s
.75755 .20306 m .75138 .21038 L .75759 .203 L .164 .557 .182 r
F
.75697 .20425 m .75138 .21038 L p .75719 .20381 L F P
0 g
s
.75719 .20381 m .75138 .21038 L .75755 .20306 L .164 .557 .182 r
F
.75138 .21038 m .75697 .20425 L p .7564 .20541 L .113 .265 .134 r
F P
0 g
s
.75367 .21079 m .75138 .21038 L p .7564 .20541 L .113 .265 .134 r
F P
0 g
s
.75138 .21038 m .75367 .21079 L p .75198 .21354 L .113 .278 .138 r
F P
0 g
s
.75409 .23154 m .7516 .23213 L .7516 .22653 L closepath p .099 .266 .146 r
F P
0 g
s
.7516 .23213 m .75409 .23154 L .75684 .2332 L closepath p .104 .343 .166 r
F P
0 g
s
.7516 .23213 m .75684 .2332 L .75597 .23816 L closepath p .106 .329 .161 r
F P
0 g
s
.75657 .21784 m .7541 .22592 L .7516 .22653 L closepath p .09 .168 .12 r
F P
0 g
s
.75409 .23154 m .7516 .22653 L .7541 .22592 L closepath p .099 .266 .146 r
F P
0 g
s
.75646 .32909 m .75165 .31635 L .76067 .31792 L closepath p .218 .803 .195 r
F P
0 g
s
.75278 .32884 m .75646 .32909 L .75504 .33793 L closepath p .204 .764 .201 r
F P
0 g
s
.75294 .33997 m .75504 .33793 L .75335 .34027 L p .205 .995 .19 r
F P
0 g
s
.75335 .34027 m .75504 .33793 L p .75339 .34031 L .197 .816 .219 r
F P
0 g
s
.76653 .2949 m .75338 .28004 L .76848 .26464 L closepath p .186 .669 .194 r
F P
0 g
s
.75338 .28004 m .76653 .2949 L .75969 .30722 L closepath p .196 .672 .185 r
F P
0 g
s
.75339 .34031 m .75504 .33793 L .75399 .34075 L .197 .816 .219 r
F
.75945 .40975 m .75817 .42496 L .75351 .4361 L closepath p .243 .926 .157 r
F P
0 g
s
.75514 .4495 m .75351 .4361 L .75817 .42496 L closepath p .243 .874 .17 r
F P
0 g
s
.76691 .38516 m .75358 .40815 L .76034 .39008 L p .222 .967 .208 r
F P
0 g
s
.76263 .394 m .75358 .40815 L .76691 .38516 L p .18 .884 .123 r
F P
0 g
s
.75415 .21 m .75367 .21079 L .7564 .20541 L .113 .265 .134 r
F
.75409 .21012 m .75367 .21079 L .75415 .21 L F
.75379 .46911 m .75718 .46619 L .75701 .46687 L closepath p .205 .998 .212 r
F P
0 g
s
.75379 .46911 m .75396 .46843 L .75718 .46619 L closepath p .205 .998 .212 r
F P
0 g
s
.75718 .46619 m .75396 .46843 L .7561 .46489 L closepath p .222 .835 .197 r
F P
0 g
s
.75399 .34075 m .75504 .33793 L .75456 .34117 L .197 .816 .219 r
F
.75404 .2177 m .76007 .20893 L .76015 .21337 L closepath p .063 .04 .096 r
F P
0 g
s
.75404 .2177 m .75636 .20967 L .76007 .20893 L closepath p .082 .126 .112 r
F P
0 g
s
.75404 .2177 m .76015 .21337 L .75657 .21784 L closepath p .063 .038 .096 r
F P
0 g
s
.76125 .19971 m .75409 .21012 L .75897 .20018 L closepath p .015 0 .069 r
F P
0 g
s
.76125 .19971 m .75636 .20967 L .75409 .21012 L closepath p .015 0 .069 r
F P
0 g
s
.75409 .21012 m .75414 .21014 L p .75416 .21011 L .238 .949 .179 r
F P
0 g
s
.75409 .23154 m .7541 .22592 L .75666 .22657 L closepath p .094 .25 .145 r
F P
0 g
s
.75684 .2332 m .75409 .23154 L .75666 .22657 L closepath p .112 .333 .158 r
F P
0 g
s
.75657 .21784 m .75625 .21985 L .7541 .22592 L closepath p 0 0 0 r
F P
0 g
s
.7541 .22592 m .75625 .21985 L .75881 .22049 L closepath p .093 .257 .148 r
F P
0 g
s
.7541 .22592 m .75881 .22049 L .75666 .22657 L closepath p .093 .257 .148 r
F P
0 g
s
.75416 .2101 m .75416 .21011 L .75414 .21014 L .238 .949 .179 r
F
.75456 .34117 m .75504 .33793 L .75479 .34134 L .197 .816 .219 r
F
.75611 .36612 m .75538 .36835 L .75471 .36587 L p .173 .633 .197 r
F P
0 g
s
.75504 .33793 m .7592 .33901 L .75479 .34134 L p .197 .816 .219 r
F P
0 g
s
.75479 .34134 m .7592 .33901 L p .75517 .34162 L .197 .816 .219 r
F P
0 g
s
.75646 .32909 m .7592 .33901 L .75504 .33793 L closepath p .208 .779 .2 r
F P
0 g
s
.75514 .4495 m .75817 .42496 L .75978 .43838 L closepath p .243 .874 .17 r
F P
0 g
s
.7573 .45692 m .75514 .4495 L .75978 .43838 L closepath p .238 .835 .173 r
F P
0 g
s
.75517 .34162 m .7592 .33901 L .75592 .34217 L .197 .816 .219 r
F
.75745 .36636 m .75538 .36835 L .75611 .36612 L p .172 .63 .197 r
F P
0 g
s
.76212 .36877 m .75538 .36835 L .75745 .36636 L p .172 .63 .197 r
F P
0 g
s
.7639 .37471 m .75538 .36835 L .76212 .36877 L closepath p .163 .706 .222 r
F P
0 g
s
.75685 .20232 m .75571 .20211 L .75961 .19903 L p 0 0 0 r
F P
0 g
s
.7592 .33901 m .75687 .34223 L p .75592 .34217 L .197 .816 .219 r
F P
0 g
s
.75687 .34223 m .75661 .34261 L p .75592 .34217 L .197 .816 .219 r
F P
0 g
s
.7618 .23271 m .75597 .23816 L .75684 .2332 L closepath p .137 .445 .173 r
F P
0 g
s
.7618 .23271 m .76092 .23768 L .75597 .23816 L closepath p .137 .445 .173 r
F P
0 g
s
.7561 .46489 m .75791 .4623 L .75899 .4636 L closepath p .233 .826 .18 r
F P
0 g
s
.7561 .46489 m .75899 .4636 L .75718 .46619 L closepath p .233 .826 .18 r
F P
0 g
s
.75625 .21985 m .76015 .21337 L .75983 .21537 L closepath p 0 0 .016 r
F P
0 g
s
.75625 .21985 m .75657 .21784 L .76015 .21337 L closepath p 0 0 .016 r
F P
0 g
s
.75625 .21985 m .75983 .21537 L .75881 .22049 L closepath p .097 .207 .127 r
F P
0 g
s
.76125 .19971 m .76007 .20893 L .75636 .20967 L closepath p .084 .149 .118 r
F P
0 g
s
.75646 .32909 m .76067 .31792 L .76932 .32021 L closepath p .21 .766 .195 r
F P
0 g
s
.75646 .32909 m .76932 .32021 L .76511 .33136 L closepath p .21 .766 .195 r
F P
0 g
s
.75646 .32909 m .76511 .33136 L .7592 .33901 L closepath p .208 .779 .2 r
F P
0 g
s
.75661 .34261 m .7592 .33901 L p .7574 .34311 L .198 .819 .219 r
F P
0 g
s
.76143 .21954 m .75666 .22657 L .75881 .22049 L closepath p .129 .408 .168 r
F P
0 g
s
.76143 .21954 m .75941 .22824 L .75666 .22657 L closepath p .114 .331 .156 r
F P
0 g
s
.75684 .2332 m .75666 .22657 L .75941 .22824 L closepath p .112 .333 .158 r
F P
0 g
s
.7618 .23271 m .75684 .2332 L .75941 .22824 L closepath p .137 .45 .174 r
F P
0 g
s
.75961 .19903 m .75923 .19972 L .75685 .20232 L 0 0 0 r
F
.75908 .19999 m .75685 .20232 L p .75923 .19972 L F P
0 g
s
.75799 .20218 m .75685 .20232 L .75905 .20001 L p .149 .307 .113 r
F P
0 g
s
.75701 .46687 m .75718 .46619 L .75899 .4636 L closepath p .209 .956 .153 r
F P
0 g
s
.7573 .45692 m .75978 .43838 L .76189 .44338 L closepath p .246 .871 .159 r
F P
0 g
s
.75821 .45422 m .75791 .4623 L .7573 .45692 L p .25 .781 .112 r
F P
0 g
s
.7592 .33901 m .76195 .34043 L .7574 .34311 L p .198 .819 .219 r
F P
0 g
s
.7574 .34311 m .76195 .34043 L .76235 .34624 L p .182 .617 .184 r
F P
0 g
s
.75906 .36665 m .76212 .36877 L .75745 .36636 L .172 .63 .197 r
F
.75884 .4593 m .75791 .4623 L .75821 .45422 L p .25 .781 .112 r
F P
0 g
s
.75899 .4636 m .75791 .4623 L .75884 .4593 L closepath p .247 .734 .115 r
F P
0 g
s
.75945 .40975 m .7613 .41627 L .75817 .42496 L closepath p .243 .924 .153 r
F P
0 g
s
.75978 .43838 m .75817 .42496 L .7613 .41627 L closepath p .242 .873 .172 r
F P
0 g
s
.75825 .45412 m .75884 .4593 L p .75821 .45422 L .25 .781 .112 r
F P
0 g
s
.76092 .23768 m .76456 .2447 L .75847 .248 L closepath p .155 .487 .17 r
F P
0 g
s
.75847 .248 m .76456 .2447 L .76529 .25281 L closepath p .154 .481 .169 r
F P
0 g
s
.7639 .37471 m .76929 .38337 L .75858 .3804 L closepath p .16 .858 .246 r
F P
0 g
s
.76034 .39008 m .75858 .3804 L .76929 .38337 L closepath p .146 .85 .249 r
F P
0 g
s
.76143 .21954 m .75881 .22049 L .75983 .21537 L closepath p .114 .269 .134 r
F P
0 g
s
.75897 .20018 m .75905 .20001 L .75906 .20001 L .149 .307 .113 r
F
.76125 .19971 m .75897 .20018 L .76317 .19263 L closepath p .009 0 .04 r
F P
0 g
s
.75905 .20001 m .75908 .19999 L p .75906 .20001 L .149 .307 .113 r
F P
0 g
s
.76261 .36728 m .76212 .36877 L p .75906 .36665 L .172 .63 .197 r
F P
0 g
s
.76195 .34043 m .7592 .33901 L .76511 .33136 L closepath p .211 .8 .202 r
F P
0 g
s
.76143 .21954 m .76249 .2243 L .75941 .22824 L closepath p .132 .401 .163 r
F P
0 g
s
.76488 .22877 m .75941 .22824 L .76249 .2243 L closepath p .143 .46 .172 r
F P
0 g
s
.76488 .22877 m .7618 .23271 L .75941 .22824 L closepath p .143 .46 .172 r
F P
0 g
s
.75945 .40975 m .76589 .38927 L .76259 .40104 L closepath p .244 .922 .156 r
F P
0 g
s
.75945 .40975 m .76259 .40104 L .7613 .41627 L closepath p .243 .924 .153 r
F P
0 g
s
.75945 .40975 m .75945 .40975 L p .75945 .40975 L .239 .92 .141 r
F P
0 g
s
.75945 .40975 m .75945 .40973 L .75945 .40975 L .239 .92 .141 r
F
.76589 .38927 m .75945 .40975 L p .75945 .40973 L F P
0 g
s
.75945 .40973 m .75945 .40973 L .75945 .40975 L .239 .92 .141 r
F
.75945 .40973 m .76261 .39411 L .76589 .38927 L F
.76653 .2949 m .76067 .31792 L .75969 .30722 L closepath p .199 .685 .186 r
F P
0 g
s
.75978 .43838 m .7613 .41627 L .76322 .42153 L closepath p .244 .869 .121 r
F P
0 g
s
.76531 .42657 m .75978 .43838 L .76322 .42153 L closepath p .249 .858 .147 r
F P
0 g
s
.76531 .42657 m .76189 .44338 L .75978 .43838 L closepath p .249 .858 .147 r
F P
0 g
s
.76475 .2112 m .76245 .21442 L .75983 .21537 L closepath p .07 0 .045 r
F P
0 g
s
.76475 .2112 m .75983 .21537 L .76015 .21337 L closepath p .059 0 .034 r
F P
0 g
s
.76143 .21954 m .75983 .21537 L .76245 .21442 L closepath p .114 .269 .134 r
F P
0 g
s
.76125 .19971 m .76466 .20676 L .76007 .20893 L closepath p .116 .257 .128 r
F P
0 g
s
.76007 .20893 m .76466 .20676 L .76475 .2112 L closepath p .101 .148 .102 r
F P
0 g
s
.76007 .20893 m .76475 .2112 L .76015 .21337 L closepath p .101 .148 .102 r
F P
0 g
s
.76189 .44338 m .76339 .44352 L .76042 .44771 L p .236 .957 .168 r
F P
0 g
s
.76653 .2949 m .76932 .32021 L .76067 .31792 L closepath p .213 .748 .187 r
F P
0 g
s
.76702 .23437 m .76092 .23768 L .7618 .23271 L closepath p .16 .533 .179 r
F P
0 g
s
.76092 .23768 m .76702 .23437 L .76456 .2447 L closepath p .155 .487 .17 r
F P
0 g
s
.76577 .19851 m .76125 .19971 L .76441 .19173 L closepath p .094 .109 .094 r
F P
0 g
s
.76125 .19971 m .76317 .19263 L .76441 .19173 L closepath p 0 0 .023 r
F P
0 g
s
.76125 .19971 m .76577 .19851 L .76466 .20676 L closepath p .109 .244 .13 r
F P
0 g
s
.76259 .40104 m .76482 .40385 L .7613 .41627 L closepath p .242 .914 .142 r
F P
0 g
s
.76322 .42153 m .7613 .41627 L .76482 .40385 L closepath p .244 .915 .147 r
F P
0 g
s
.76585 .21713 m .76143 .21954 L .76245 .21442 L closepath p .15 .4 .146 r
F P
0 g
s
.76585 .21713 m .76249 .2243 L .76143 .21954 L closepath p .159 .472 .161 r
F P
0 g
s
.76702 .23437 m .7618 .23271 L .76488 .22877 L closepath p .158 .545 .185 r
F P
0 g
s
.76656 .4332 m .76189 .44338 L .76531 .42657 L closepath p .232 .766 .165 r
F P
0 g
s
.76656 .4332 m .76553 .43825 L .76189 .44338 L closepath p .232 .766 .165 r
F P
0 g
s
.76339 .44352 m .76189 .44338 L .76553 .43825 L closepath p .236 .957 .168 r
F P
0 g
s
.76195 .34043 m .76511 .33136 L .76786 .33279 L closepath p .211 .8 .202 r
F P
0 g
s
.76195 .34043 m .76786 .33279 L .767 .34129 L closepath p .191 .663 .188 r
F P
0 g
s
.76195 .34043 m .767 .34129 L .76341 .34691 L p .192 .648 .183 r
F P
0 g
s
.76235 .34624 m .76195 .34043 L p .76341 .34691 L .192 .648 .183 r
F P
0 g
s
.76734 .36852 m .76212 .36877 L .76261 .36728 L p .163 .591 .194 r
F P
0 g
s
.76911 .37446 m .76212 .36877 L .76734 .36852 L closepath p .158 .693 .222 r
F P
0 g
s
.76911 .37446 m .7639 .37471 L .76212 .36877 L closepath p .158 .693 .222 r
F P
0 g
s
.76815 .2139 m .76245 .21442 L .76475 .2112 L closepath p .168 .37 .113 r
F P
0 g
s
.76815 .2139 m .76585 .21713 L .76245 .21442 L closepath p .168 .37 .113 r
F P
0 g
s
.76585 .21713 m .76691 .22188 L .76249 .2243 L closepath p .159 .472 .161 r
F P
0 g
s
.76691 .22188 m .76488 .22877 L .76249 .2243 L closepath p .161 .491 .165 r
F P
0 g
s
.76259 .40104 m .76589 .38927 L .76812 .39213 L closepath p .241 .921 .146 r
F P
0 g
s
.76259 .40104 m .76812 .39213 L .76482 .40385 L closepath p .241 .921 .146 r
F P
0 g
s
.76734 .36852 m .76261 .36728 L .76376 .36749 L .163 .591 .194 r
F
.76261 .39411 m .76275 .39377 L .76589 .38927 L .239 .92 .141 r
F
.76263 .394 m .76275 .39377 L .76261 .39411 L F
.76275 .39377 m .76275 .39377 L .76261 .39411 L F
.76711 .38501 m .76263 .394 L p .767 .38509 L .18 .884 .123 r
F P
0 g
s
.76691 .38516 m .767 .38509 L .76263 .394 L .18 .884 .123 r
F
.76275 .39377 m .76509 .38906 L .76589 .38927 L .239 .92 .141 r
F
.77048 .18399 m .76441 .19173 L .76317 .19263 L closepath p 0 0 0 r
F P
0 g
s
.77048 .18399 m .76317 .19263 L .76924 .1849 L closepath p 0 0 0 r
F P
0 g
s
.76322 .42153 m .76482 .40385 L .76672 .40916 L closepath p .244 .915 .147 r
F P
0 g
s
.76531 .42657 m .76322 .42153 L .76672 .40916 L closepath p .243 .88 .171 r
F P
0 g
s
.76637 .4385 m .76339 .44352 L .76553 .43825 L closepath p .238 .94 .183 r
F P
0 g
s
.76341 .34691 m .767 .34129 L p .76683 .34907 L .192 .648 .183 r
F P
0 g
s
.76374 .36491 m .76717 .36219 L .78844 .3688 L closepath p .13 .304 .132 r
F P
0 g
s
.76374 .36491 m .78175 .3717 L .76473 .36766 L closepath p .08 .474 .214 r
F P
0 g
s
.76374 .36491 m .78844 .3688 L .78499 .37153 L closepath p .13 .304 .132 r
F P
0 g
s
.76374 .36491 m .78499 .37153 L .78175 .3717 L closepath p .12 .384 .167 r
F P
0 g
s
.76734 .36852 m .76376 .36749 L .76691 .36817 L .163 .591 .194 r
F
.76376 .36749 m .76473 .36766 L .76691 .36817 L F
.7639 .37471 m .76911 .37446 L .77463 .37772 L closepath p .11 .806 .249 r
F P
0 g
s
.7639 .37471 m .77463 .37772 L .76929 .38337 L closepath p .16 .858 .246 r
F P
0 g
s
.77048 .18399 m .76894 .19052 L .76441 .19173 L closepath p .061 0 .017 r
F P
0 g
s
.76577 .19851 m .76441 .19173 L .76894 .19052 L closepath p .094 .109 .094 r
F P
0 g
s
.76702 .23437 m .77149 .2391 L .76456 .2447 L closepath p .153 .479 .169 r
F P
0 g
s
.76529 .25281 m .76456 .2447 L .77149 .2391 L closepath p .152 .476 .168 r
F P
0 g
s
.76466 .20676 m .76577 .19851 L .77519 .19855 L closepath p .155 .409 .143 r
F P
0 g
s
.76466 .20676 m .77519 .19855 L .77409 .2068 L closepath p .155 .409 .143 r
F P
0 g
s
.76466 .20676 m .76815 .2139 L .76475 .2112 L closepath p .171 .362 .106 r
F P
0 g
s
.76466 .20676 m .77409 .2068 L .76815 .2139 L closepath p .151 .331 .119 r
F P
0 g
s
.76672 .40916 m .76482 .40385 L .76812 .39213 L closepath p .243 .923 .153 r
F P
0 g
s
.76691 .22188 m .76805 .23035 L .76488 .22877 L closepath p .181 .572 .171 r
F P
0 g
s
.76702 .23437 m .76488 .22877 L .76805 .23035 L closepath p .177 .585 .179 r
F P
0 g
s
.76581 .38762 m .76589 .38927 L p .76509 .38906 L .239 .92 .141 r
F P
0 g
s
.77588 .33048 m .76511 .33136 L .76932 .32021 L closepath p .204 .735 .194 r
F P
0 g
s
.77588 .33048 m .76786 .33279 L .76511 .33136 L closepath p .198 .818 .218 r
F P
0 g
s
.76529 .25281 m .77149 .2391 L .77223 .24721 L closepath p .152 .476 .168 r
F P
0 g
s
.77582 .25261 m .76529 .25281 L .77223 .24721 L closepath p .179 .671 .201 r
F P
0 g
s
.77582 .25261 m .76848 .26464 L .76529 .25281 L closepath p .18 .631 .189 r
F P
0 g
s
.76531 .42657 m .76672 .40916 L .76921 .41219 L closepath p .244 .883 .169 r
F P
0 g
s
.76599 .42406 m .76656 .4332 L .76531 .42657 L p .248 .786 .13 r
F P
0 g
s
.77352 .35682 m .76717 .36219 L .76532 .35606 L closepath p .153 .355 .126 r
F P
0 g
s
.76553 .43825 m .76656 .4332 L .7674 .43346 L closepath p .249 .858 .147 r
F P
0 g
s
.76553 .43825 m .7674 .43346 L .76637 .4385 L closepath p .249 .858 .147 r
F P
0 g
s
.76577 .19851 m .76894 .19052 L .77438 .18968 L closepath p .154 .35 .123 r
F P
0 g
s
.76577 .19851 m .77438 .18968 L .77519 .19855 L closepath p .152 .337 .12 r
F P
0 g
s
.76616 .38691 m .76589 .38927 L .76581 .38762 L p .233 .862 .108 r
F P
0 g
s
.77009 .22346 m .76585 .21713 L .76815 .2139 L closepath p .202 .55 .137 r
F P
0 g
s
.77009 .22346 m .76691 .22188 L .76585 .21713 L closepath p .187 .542 .154 r
F P
0 g
s
.76812 .39213 m .76589 .38927 L .77048 .38234 L p .24 .933 .152 r
F P
0 g
s
.76972 .38349 m .76589 .38927 L p .76851 .38396 L .247 .883 .155 r
F P
0 g
s
.76589 .38927 m .7665 .38622 L p .76851 .38396 L .247 .883 .155 r
F P
0 g
s
.7665 .38622 m .76589 .38927 L .76616 .38691 L p .233 .862 .108 r
F P
0 g
s
.76704 .42742 m .76656 .4332 L .76599 .42406 L p .248 .786 .13 r
F P
0 g
s
.76599 .42406 m .76609 .4237 L .76704 .42742 L .248 .786 .13 r
F
.76624 .42316 m .76704 .42742 L p .76609 .4237 L F P
0 g
s
.7665 .38622 m .76711 .38501 L .76851 .38396 L .247 .883 .155 r
F
.76653 .2949 m .76848 .26464 L .77561 .26863 L closepath p .21 .751 .192 r
F P
0 g
s
.76653 .2949 m .77561 .26863 L .78578 .27148 L closepath p .212 .759 .192 r
F P
0 g
s
.76653 .2949 m .78578 .27148 L .77671 .29767 L closepath p .212 .759 .192 r
F P
0 g
s
.76653 .2949 m .7747 .31125 L .76932 .32021 L closepath p .221 .767 .182 r
F P
0 g
s
.76653 .2949 m .77671 .29767 L .7747 .31125 L closepath p .212 .758 .191 r
F P
0 g
s
.7674 .43346 m .76656 .4332 L .76704 .42742 L closepath p .25 .795 .124 r
F P
0 g
s
.76672 .40916 m .76812 .39213 L .77093 .39678 L closepath p .239 .922 .184 r
F P
0 g
s
.77341 .39984 m .76672 .40916 L .77093 .39678 L closepath p .236 .904 .188 r
F P
0 g
s
.77341 .39984 m .76921 .41219 L .76672 .40916 L closepath p .236 .904 .188 r
F P
0 g
s
.76683 .34907 m .76754 .3492 L .76748 .34949 L p .191 .664 .188 r
F P
0 g
s
.767 .34129 m .76754 .3492 L .76683 .34907 L p .192 .648 .183 r
F P
0 g
s
.76734 .36852 m .76691 .36817 L .76702 .3682 L .163 .591 .194 r
F
.76691 .22188 m .77009 .22346 L .76805 .23035 L closepath p .181 .572 .171 r
F P
0 g
s
.77502 .33898 m .767 .34129 L .76786 .33279 L closepath p .197 .687 .188 r
F P
0 g
s
.767 .34129 m .77502 .33898 L .76988 .34945 L closepath p .193 .618 .172 r
F P
0 g
s
.767 .34129 m .76988 .34945 L .76754 .3492 L closepath p .156 .537 .184 r
F P
0 g
s
.76702 .23437 m .76805 .23035 L .77201 .23049 L closepath p .155 .497 .172 r
F P
0 g
s
.76702 .23437 m .77201 .23049 L .77098 .23451 L closepath p .155 .497 .172 r
F P
0 g
s
.76702 .23437 m .77098 .23451 L .77149 .2391 L closepath p .155 .48 .167 r
F P
0 g
s
.76729 .36827 m .76734 .36852 L p .76702 .3682 L .163 .591 .194 r
F P
0 g
s
.77352 .35682 m .78844 .3688 L .76717 .36219 L closepath p .134 .246 .106 r
F P
0 g
s
.76735 .36828 m .76734 .36852 L .76729 .36827 L p .153 .566 .194 r
F P
0 g
s
.76745 .3683 m .76734 .36852 L .76735 .36828 L p .154 .567 .194 r
F P
0 g
s
.77157 .36933 m .76734 .36852 L .76745 .3683 L p .154 .567 .194 r
F P
0 g
s
.76734 .36852 m .77157 .36933 L .7744 .37454 L closepath p .125 .666 .234 r
F P
0 g
s
.76734 .36852 m .7744 .37454 L .76911 .37446 L closepath p .163 .707 .222 r
F P
0 g
s
.76745 .3683 m .77148 .36926 L .77157 .36933 L .154 .567 .194 r
F
.76748 .34949 m .76754 .3492 L p .7675 .3495 L .156 .526 .181 r
F P
0 g
s
.76754 .3492 m .76755 .34953 L p .7675 .3495 L .156 .526 .181 r
F P
0 g
s
.76754 .3492 m .76765 .34959 L .76758 .34955 L .156 .526 .181 r
F
.76754 .3492 m .76803 .34984 L .7679 .34975 L F
.76754 .3492 m .7679 .34975 L .76765 .3496 L F
.76754 .3492 m .76765 .3496 L .76765 .34959 L F
.76754 .3492 m .76765 .34959 L .76765 .34959 L F
.76755 .34953 m .76754 .3492 L p .76758 .34955 L F P
0 g
s
.76754 .3492 m .76957 .35081 L .76803 .34984 L .156 .526 .181 r
F
.76754 .3492 m .76988 .34945 L .76957 .35081 L p F P
0 g
s
.77502 .33898 m .76786 .33279 L .77588 .33048 L closepath p .197 .687 .188 r
F P
0 g
s
.77009 .22346 m .77201 .23049 L .76805 .23035 L closepath p .155 .466 .163 r
F P
0 g
s
.77093 .39678 m .76812 .39213 L .77127 .38166 L p .242 .915 .176 r
F P
0 g
s
.77127 .38166 m .76812 .39213 L p .77091 .38197 L .24 .933 .152 r
F P
0 g
s
.77048 .38234 m .77091 .38197 L .76812 .39213 L .24 .933 .152 r
F
.77409 .2068 m .77199 .21866 L .76815 .2139 L closepath p .173 .444 .136 r
F P
0 g
s
.77009 .22346 m .76815 .2139 L .77199 .21866 L closepath p .159 .438 .15 r
F P
0 g
s
.77052 .41329 m .76848 .41489 L p .76921 .41219 L .221 .932 .211 r
F P
0 g
s
.77582 .25261 m .77561 .26863 L .76848 .26464 L closepath p .204 .764 .201 r
F P
0 g
s
.76851 .38396 m .76929 .38337 L .76972 .38349 L .247 .883 .155 r
F
.76894 .19052 m .77589 .18316 L .77438 .18968 L closepath p .115 .023 .03 r
F P
0 g
s
.76894 .19052 m .77048 .18399 L .77589 .18316 L closepath p .115 .023 .03 r
F P
0 g
s
.77463 .37772 m .76911 .37446 L .7744 .37454 L closepath p .105 .795 .248 r
F P
0 g
s
.77455 .40529 m .76921 .41219 L .77341 .39984 L closepath p .217 .802 .196 r
F P
0 g
s
.77455 .40529 m .77329 .409 L .76921 .41219 L closepath p .217 .802 .196 r
F P
0 g
s
.77129 .41269 m .76921 .41219 L .77329 .409 L closepath p .221 .932 .211 r
F P
0 g
s
.76921 .41219 m .77129 .41269 L .77052 .41329 L p .221 .932 .211 r
F P
0 g
s
.77048 .38234 m .76972 .38349 L p .76929 .38337 L .247 .883 .155 r
F P
0 g
s
.77582 .37773 m .76929 .38337 L .77463 .37772 L p .147 .903 .16 r
F P
0 g
s
.7797 .32378 m .76932 .32021 L .7747 .31125 L closepath p .209 .697 .177 r
F P
0 g
s
.77588 .33048 m .76932 .32021 L .7797 .32378 L closepath p .203 .734 .194 r
F P
0 g
s
.76957 .35081 m .76988 .34945 L p .76983 .35098 L .162 .551 .182 r
F P
0 g
s
.77009 .35114 m .76983 .35098 L .76988 .34945 L .162 .551 .182 r
F
.76988 .34945 m .77502 .33898 L .77872 .3408 L closepath p .173 .517 .162 r
F P
0 g
s
.76988 .34945 m .77872 .3408 L .77358 .35128 L closepath p .173 .517 .162 r
F P
0 g
s
.76988 .34945 m .77358 .35128 L .77252 .35268 L p .162 .551 .182 r
F P
0 g
s
.77012 .35116 m .7701 .35115 L .76988 .34945 L .162 .551 .182 r
F
.7701 .35115 m .77009 .35114 L .76988 .34945 L F
.77172 .35217 m .77147 .35201 L .76988 .34945 L F
.77147 .35201 m .77012 .35116 L .76988 .34945 L F
.77252 .35268 m .77172 .35217 L .76988 .34945 L F
.77009 .22346 m .77199 .21866 L .77626 .22231 L closepath p .163 .458 .152 r
F P
0 g
s
.77009 .22346 m .77626 .22231 L .77436 .22712 L closepath p .163 .458 .152 r
F P
0 g
s
.77009 .22346 m .77436 .22712 L .77201 .23049 L closepath p .154 .464 .163 r
F P
0 g
s
.77436 .38542 m .77093 .39678 L .77152 .38144 L p .242 .915 .176 r
F P
0 g
s
.77152 .38144 m .77093 .39678 L p .7714 .38155 L .242 .915 .176 r
F P
0 g
s
.77093 .39678 m .77127 .38166 L .7714 .38155 L .242 .915 .176 r
F
.77341 .39984 m .77093 .39678 L .77436 .38542 L closepath p .239 .898 .181 r
F P
0 g
s
.77201 .23049 m .77436 .23416 L .77098 .23451 L closepath p .132 .403 .164 r
F P
0 g
s
.77488 .23876 m .77098 .23451 L .77436 .23416 L closepath p .132 .412 .166 r
F P
0 g
s
.77488 .23876 m .77149 .2391 L .77098 .23451 L closepath p .132 .412 .166 r
F P
0 g
s
.77422 .40962 m .77129 .41269 L .77329 .409 L closepath p .22 .926 .212 r
F P
0 g
s
.77148 .36926 m .7715 .36927 L .77157 .36933 L .154 .567 .194 r
F
.77488 .23876 m .77223 .24721 L .77149 .2391 L closepath p .132 .415 .167 r
F P
0 g
s
.7715 .36927 m .77152 .36927 L .77157 .36933 L .154 .567 .194 r
F
.77152 .38144 m .77167 .38131 L .77436 .38542 L .242 .915 .176 r
F
.77152 .36927 m .77156 .36928 L .77157 .36933 L .154 .567 .194 r
F
.77157 .36928 m .77157 .36933 L p .77156 .36928 L F P
0 g
s
.77173 .36932 m .77157 .36933 L .77157 .36928 L p .159 .585 .195 r
F P
0 g
s
.7804 .37403 m .77157 .36933 L .77173 .36932 L p .152 .726 .232 r
F P
0 g
s
.77157 .36933 m .7804 .37403 L .7744 .37454 L closepath p .152 .726 .232 r
F P
0 g
s
.77182 .38118 m .77436 .38542 L p .77167 .38131 L .242 .915 .176 r
F P
0 g
s
.77173 .36932 m .7753 .37017 L .7804 .37403 L .152 .726 .232 r
F
.77436 .38542 m .77182 .38118 L p .77587 .37773 L .238 .923 .186 r
F P
0 g
s
.77586 .37773 m .77587 .37773 L .77182 .38118 L .238 .923 .186 r
F
.77582 .37773 m .77586 .37773 L .77182 .38118 L F
.77409 .2068 m .77792 .21155 L .77199 .21866 L closepath p .173 .444 .136 r
F P
0 g
s
.77792 .21155 m .77626 .22231 L .77199 .21866 L closepath p .173 .447 .137 r
F P
0 g
s
.77436 .22712 m .77436 .23416 L .77201 .23049 L closepath p .146 .423 .157 r
F P
0 g
s
.77488 .23876 m .77534 .2439 L .77223 .24721 L closepath p .133 .42 .168 r
F P
0 g
s
.77894 .24931 m .77223 .24721 L .77534 .2439 L closepath p .172 .662 .205 r
F P
0 g
s
.77894 .24931 m .77582 .25261 L .77223 .24721 L closepath p .172 .662 .205 r
F P
0 g
s
.77252 .35268 m .77358 .35128 L .77324 .35313 L p .158 .528 .179 r
F P
0 g
s
.77324 .35313 m .77358 .35128 L p .77327 .35315 L .153 .507 .178 r
F P
0 g
s
.77358 .35128 m .77328 .35316 L .77327 .35315 L .153 .507 .178 r
F
.77358 .35128 m .77329 .35317 L .77328 .35316 L F
.77329 .409 m .77455 .40529 L .77548 .40592 L closepath p .236 .9 .189 r
F P
0 g
s
.77329 .409 m .77548 .40592 L .77422 .40962 L closepath p .236 .9 .189 r
F P
0 g
s
.77358 .35128 m .77392 .35357 L p .77329 .35317 L .153 .507 .178 r
F P
0 g
s
.77341 .39984 m .77436 .38542 L .77735 .38763 L closepath p .236 .889 .186 r
F P
0 g
s
.77395 .39815 m .77455 .40529 L .77341 .39984 L p .25 .82 .122 r
F P
0 g
s
.77352 .35682 m .77526 .35441 L .79028 .36339 L closepath p .141 .171 .064 r
F P
0 g
s
.77352 .35682 m .79028 .36339 L .78855 .36583 L closepath p .141 .171 .064 r
F P
0 g
s
.77352 .35682 m .78855 .36583 L .78844 .3688 L closepath p .118 .262 .129 r
F P
0 g
s
.77979 .3464 m .77358 .35128 L .77872 .3408 L closepath p .14 .366 .144 r
F P
0 g
s
.77979 .3464 m .77698 .35505 L .77358 .35128 L closepath p .162 .512 .171 r
F P
0 g
s
.77358 .35128 m .77479 .35412 L .77462 .35401 L .153 .507 .178 r
F
.77392 .35357 m .77358 .35128 L p .77462 .35401 L F P
0 g
s
.77358 .35128 m .77577 .35371 L p .77526 .35441 L .153 .507 .178 r
F P
0 g
s
.77358 .35128 m .77526 .35441 L .77479 .35412 L .153 .507 .178 r
F
.77395 .39815 m .77402 .39796 L .77495 .40183 L .25 .82 .122 r
F
.77495 .40183 m .77455 .40529 L .77395 .39815 L p F P
0 g
s
.77409 .39774 m .77495 .40183 L p .77402 .39796 L .25 .82 .122 r
F P
0 g
s
.7881 .20989 m .77409 .2068 L .77519 .19855 L closepath p .183 .512 .148 r
F P
0 g
s
.7881 .20989 m .77792 .21155 L .77409 .2068 L closepath p .183 .445 .123 r
F P
0 g
s
.77735 .38763 m .77436 .38542 L .78037 .37768 L p .234 .896 .191 r
F P
0 g
s
.77587 .37773 m .777 .37775 L .77436 .38542 L .238 .923 .186 r
F
.77717 .37775 m .77436 .38542 L p .777 .37775 L F P
0 g
s
.78016 .37795 m .77436 .38542 L .77717 .37775 L p .234 .896 .191 r
F P
0 g
s
.77436 .22712 m .7767 .23079 L .77436 .23416 L closepath p .146 .423 .157 r
F P
0 g
s
.78386 .23551 m .77436 .23416 L .7767 .23079 L closepath p .153 .457 .162 r
F P
0 g
s
.78386 .23551 m .77488 .23876 L .77436 .23416 L closepath p .153 .472 .167 r
F P
0 g
s
.7767 .23079 m .77436 .22712 L .77626 .22231 L closepath p .159 .438 .15 r
F P
0 g
s
.78615 .18839 m .77438 .18968 L .77589 .18316 L closepath p .135 .096 .035 r
F P
0 g
s
.77438 .18968 m .78615 .18839 L .787 .19725 L closepath p .171 .393 .119 r
F P
0 g
s
.77438 .18968 m .787 .19725 L .77519 .19855 L closepath p .171 .393 .119 r
F P
0 g
s
.7744 .37454 m .7804 .37403 L .78442 .37676 L closepath p .127 .807 .25 r
F P
0 g
s
.7744 .37454 m .78442 .37676 L .77989 .37779 L closepath p .089 .761 .246 r
F P
0 g
s
.77463 .37772 m .7744 .37454 L .77623 .37562 L p .105 .795 .248 r
F P
0 g
s
.77548 .40592 m .77455 .40529 L .77495 .40183 L closepath p .248 .8 .137 r
F P
0 g
s
.77623 .37562 m .77989 .37779 L .77463 .37772 L p .105 .795 .248 r
F P
0 g
s
.78712 .3034 m .7747 .31125 L .77671 .29767 L closepath p .233 .839 .183 r
F P
0 g
s
.7747 .31125 m .78712 .3034 L .78506 .31483 L closepath p .214 .616 .143 r
F P
0 g
s
.7797 .32378 m .7747 .31125 L .78506 .31483 L closepath p .209 .697 .177 r
F P
0 g
s
.78386 .23551 m .7825 .23914 L .77488 .23876 L closepath p .151 .459 .164 r
F P
0 g
s
.7825 .23914 m .77534 .2439 L .77488 .23876 L closepath p .151 .474 .169 r
F P
0 g
s
.77502 .33898 m .77588 .33048 L .78449 .33185 L closepath p .213 .743 .186 r
F P
0 g
s
.77502 .33898 m .78449 .33185 L .78353 .33806 L closepath p .197 .606 .163 r
F P
0 g
s
.77502 .33898 m .78353 .33806 L .77872 .3408 L closepath p .184 .424 .113 r
F P
0 g
s
.7881 .20989 m .77519 .19855 L .787 .19725 L closepath p .18 .516 .153 r
F P
0 g
s
.77577 .35371 m .77605 .35488 L .77526 .35441 L .153 .507 .178 r
F
.77875 .37099 m .7804 .37403 L p .7753 .37017 L .152 .726 .232 r
F P
0 g
s
.7825 .23914 m .78297 .24429 L .77534 .2439 L closepath p .151 .474 .169 r
F P
0 g
s
.78297 .24429 m .77894 .24931 L .77534 .2439 L closepath p .143 .616 .214 r
F P
0 g
s
.77582 .25261 m .78294 .25663 L .77561 .26863 L closepath p .204 .764 .201 r
F P
0 g
s
.78294 .25663 m .78578 .27148 L .77561 .26863 L closepath p .207 .783 .203 r
F P
0 g
s
.77577 .35371 m .77698 .35505 L p .77605 .35488 L .153 .507 .178 r
F P
0 g
s
.78294 .25663 m .77582 .25261 L .77894 .24931 L closepath p .189 .781 .219 r
F P
0 g
s
.77588 .33048 m .7797 .32378 L .78831 .32517 L closepath p .21 .772 .197 r
F P
0 g
s
.77588 .33048 m .78831 .32517 L .78449 .33185 L closepath p .21 .772 .197 r
F P
0 g
s
.78615 .18839 m .77589 .18316 L .78062 .18459 L closepath p .108 0 0 r
F P
0 g
s
.77698 .35505 m .77691 .3554 L p .77605 .35488 L .153 .507 .178 r
F P
0 g
s
.77792 .21155 m .78195 .2191 L .77626 .22231 L closepath p .18 .474 .138 r
F P
0 g
s
.78195 .2191 m .78125 .22503 L .77626 .22231 L closepath p .154 .296 .101 r
F P
0 g
s
.7767 .23079 m .77626 .22231 L .78125 .22503 L closepath p .138 .374 .149 r
F P
0 g
s
.77735 .38763 m .77855 .3885 L .7764 .39058 L p .22 .993 .193 r
F P
0 g
s
.78841 .22973 m .7767 .23079 L .78125 .22503 L closepath p .153 .457 .162 r
F P
0 g
s
.78841 .22973 m .78386 .23551 L .7767 .23079 L closepath p .153 .457 .162 r
F P
0 g
s
.78914 .28988 m .77671 .29767 L .78578 .27148 L closepath p .233 .87 .188 r
F P
0 g
s
.78712 .3034 m .77671 .29767 L .78914 .28988 L closepath p .233 .839 .183 r
F P
0 g
s
.77691 .3554 m .77698 .35505 L .77955 .35698 L p .123 .394 .168 r
F P
0 g
s
.77698 .35505 m .77979 .3464 L .79185 .35539 L closepath p .132 .388 .158 r
F P
0 g
s
.77698 .35505 m .78942 .36288 L .78534 .36044 L .132 .388 .158 r
F
.77698 .35505 m .79185 .35539 L .78942 .36288 L p F P
0 g
s
.77955 .35698 m .77698 .35505 L p .78534 .36044 L .132 .388 .158 r
F P
0 g
s
.78016 .37795 m .77717 .37775 L .77989 .37779 L .234 .896 .191 r
F
.78106 .37752 m .77735 .38763 L p .7805 .37765 L F P
0 g
s
.78217 .38075 m .77735 .38763 L .78106 .37752 L p .233 .889 .192 r
F P
0 g
s
.7805 .37765 m .77735 .38763 L .78037 .37768 L .234 .896 .191 r
F
.78217 .38075 m .7809 .3842 L .77735 .38763 L closepath p .233 .889 .192 r
F P
0 g
s
.77855 .3885 m .77735 .38763 L .7809 .3842 L closepath p .22 .993 .193 r
F P
0 g
s
.77792 .21155 m .7881 .20989 L .79214 .21743 L closepath p .187 .482 .133 r
F P
0 g
s
.77792 .21155 m .79214 .21743 L .78195 .2191 L closepath p .187 .482 .133 r
F P
0 g
s
.78136 .38454 m .77855 .3885 L .7809 .3842 L closepath p .216 .989 .18 r
F P
0 g
s
.77979 .3464 m .77872 .3408 L .78353 .33806 L closepath p .197 .515 .132 r
F P
0 g
s
.78448 .37272 m .77875 .37099 L .78114 .37155 L .078 .547 .229 r
F
.7804 .37403 m .77875 .37099 L p .78448 .37272 L F P
0 g
s
.78297 .24429 m .78514 .24745 L .77894 .24931 L closepath p .148 .643 .218 r
F P
0 g
s
.78514 .24745 m .78496 .25361 L .77894 .24931 L closepath p .147 .766 .241 r
F P
0 g
s
.78294 .25663 m .77894 .24931 L .78496 .25361 L closepath p .197 .793 .214 r
F P
0 g
s
.78506 .31483 m .78831 .32517 L .7797 .32378 L closepath p .214 .725 .179 r
F P
0 g
s
.77979 .3464 m .78353 .33806 L .7846 .34365 L closepath p .197 .515 .132 r
F P
0 g
s
.79185 .35539 m .77979 .3464 L .7846 .34365 L closepath p .168 .318 .092 r
F P
0 g
s
.78016 .37795 m .77989 .37779 L .78008 .37775 L .234 .896 .191 r
F
.78037 .37768 m .78016 .37795 L p .78008 .37775 L F P
0 g
s
.79148 .37535 m .7804 .37403 L p .78448 .37272 L .078 .547 .229 r
F P
0 g
s
.78442 .37676 m .7804 .37403 L .79148 .37535 L p .037 .6 .232 r
F P
0 g
s
.78615 .18839 m .78062 .18459 L .79088 .18982 L closepath p .108 0 0 r
F P
0 g
s
.7809 .3842 m .78217 .38075 L .78263 .38109 L closepath p .232 .944 .149 r
F P
0 g
s
.7809 .3842 m .78263 .38109 L .78136 .38454 L closepath p .232 .944 .149 r
F P
0 g
s
.78217 .38075 m .78106 .37752 L .78134 .37746 L .233 .889 .192 r
F
.78448 .37272 m .78114 .37155 L .78175 .3717 L .078 .547 .229 r
F
.78195 .2191 m .78695 .22181 L .78125 .22503 L closepath p .154 .296 .101 r
F P
0 g
s
.78695 .22181 m .78841 .22973 L .78125 .22503 L closepath p .171 .408 .124 r
F P
0 g
s
.78173 .37737 m .78217 .38075 L p .78134 .37746 L .233 .889 .192 r
F P
0 g
s
.78189 .37733 m .78217 .38075 L .78173 .37737 L p .213 .667 .033 r
F P
0 g
s
.78499 .37153 m .79454 .3765 L .78175 .3717 L closepath p .117 .567 .218 r
F P
0 g
s
.78207 .37729 m .78217 .38075 L .78189 .37733 L p .213 .667 .033 r
F P
0 g
s
.79307 .2212 m .78195 .2191 L .79214 .21743 L closepath p .188 .498 .137 r
F P
0 g
s
.79307 .2212 m .78695 .22181 L .78195 .2191 L closepath p .158 .13 .011 r
F P
0 g
s
.78213 .37728 m .78263 .38109 L p .78207 .37729 L .235 .731 .065 r
F P
0 g
s
.78263 .38109 m .78217 .38075 L .78207 .37729 L p .235 .731 .065 r
F P
0 g
s
.78386 .23551 m .80079 .24563 L .7825 .23914 L closepath p .157 .485 .167 r
F P
0 g
s
.80079 .24563 m .78297 .24429 L .7825 .23914 L closepath p .156 .489 .169 r
F P
0 g
s
.78294 .25663 m .78496 .25361 L .7923 .2557 L closepath p .199 .808 .215 r
F P
0 g
s
.78294 .25663 m .7923 .2557 L .79028 .25871 L closepath p .199 .808 .215 r
F P
0 g
s
.78294 .25663 m .79028 .25871 L .78578 .27148 L closepath p .207 .783 .203 r
F P
0 g
s
.78297 .24429 m .80079 .24563 L .80295 .24879 L closepath p .145 .639 .218 r
F P
0 g
s
.78297 .24429 m .80295 .24879 L .78514 .24745 L closepath p .145 .639 .218 r
F P
0 g
s
.78907 .34327 m .78353 .33806 L .78449 .33185 L closepath p .237 .768 .157 r
F P
0 g
s
.78907 .34327 m .7846 .34365 L .78353 .33806 L closepath p .24 .63 .085 r
F P
0 g
s
.78386 .23551 m .78841 .22973 L .80215 .24198 L closepath p .158 .482 .165 r
F P
0 g
s
.78386 .23551 m .80215 .24198 L .80079 .24563 L closepath p .157 .485 .167 r
F P
0 g
s
.79205 .37556 m .78442 .37676 L .79148 .37535 L .037 .6 .232 r
F
.79333 .37604 m .78442 .37676 L p .79205 .37556 L F P
0 g
s
.79824 .37842 m .78442 .37676 L .79333 .37604 L p .037 .6 .232 r
F P
0 g
s
.78831 .32517 m .79003 .33708 L .78449 .33185 L closepath p .216 .813 .199 r
F P
0 g
s
.78907 .34327 m .78449 .33185 L .79003 .33708 L closepath p .237 .768 .157 r
F P
0 g
s
.78907 .34327 m .79033 .34654 L .7846 .34365 L closepath p .205 .346 .016 r
F P
0 g
s
.80364 .35591 m .79791 .35302 L .7846 .34365 L closepath p .127 0 0 r
F P
0 g
s
.80364 .35591 m .7846 .34365 L .79033 .34654 L closepath p .127 0 0 r
F P
0 g
s
.79185 .35539 m .7846 .34365 L .79791 .35302 L closepath p .129 .273 .121 r
F P
0 g
s
.78514 .24745 m .79116 .25176 L .78496 .25361 L closepath p .147 .766 .241 r
F P
0 g
s
.79116 .25176 m .7923 .2557 L .78496 .25361 L closepath p .141 .829 .249 r
F P
0 g
s
.7978 .37634 m .78499 .37153 L .78844 .3688 L closepath p .155 .463 .162 r
F P
0 g
s
.78499 .37153 m .7978 .37634 L .79454 .3765 L closepath p .117 .567 .218 r
F P
0 g
s
.78506 .31483 m .78712 .3034 L .79191 .30549 L closepath p .225 .663 .142 r
F P
0 g
s
.78506 .31483 m .79191 .30549 L .78986 .31691 L closepath p .225 .663 .142 r
F P
0 g
s
.78506 .31483 m .78986 .31691 L .78831 .32517 L closepath p .22 .735 .175 r
F P
0 g
s
.79116 .25176 m .78514 .24745 L .80295 .24879 L closepath p .117 .724 .245 r
F P
0 g
s
.79372 .27172 m .78578 .27148 L .79028 .25871 L closepath p .21 .801 .203 r
F P
0 g
s
.78914 .28988 m .78578 .27148 L .79372 .27172 L closepath p .208 .821 .21 r
F P
0 g
s
.78615 .18839 m .79617 .2 L .787 .19725 L closepath p .217 .538 .106 r
F P
0 g
s
.78615 .18839 m .79088 .18982 L .79617 .2 L closepath p .217 .538 .106 r
F P
0 g
s
.78695 .22181 m .79307 .2212 L .79809 .22698 L closepath p .195 .387 .078 r
F P
0 g
s
.78695 .22181 m .79197 .2276 L .78841 .22973 L closepath p .172 .41 .124 r
F P
0 g
s
.78695 .22181 m .79809 .22698 L .79197 .2276 L closepath p .195 .387 .078 r
F P
0 g
s
.7881 .20989 m .787 .19725 L .79617 .2 L closepath p .22 .636 .141 r
F P
0 g
s
.78712 .3034 m .78914 .28988 L .79174 .30542 L p .234 .846 .182 r
F P
0 g
s
.7881 .20989 m .79617 .2 L .79728 .21262 L closepath p .22 .636 .141 r
F P
0 g
s
.7881 .20989 m .79728 .21262 L .79885 .21941 L closepath p .215 .505 .094 r
F P
0 g
s
.7881 .20989 m .79885 .21941 L .79214 .21743 L closepath p .215 .505 .094 r
F P
0 g
s
.78986 .31691 m .79238 .32779 L .78831 .32517 L closepath p .237 .809 .169 r
F P
0 g
s
.78831 .32517 m .79238 .32779 L .79003 .33708 L closepath p .229 .842 .189 r
F P
0 g
s
.79197 .2276 m .80215 .24198 L .78841 .22973 L closepath p .179 .464 .135 r
F P
0 g
s
.80054 .3737 m .78844 .3688 L .78855 .36583 L closepath p .162 .407 .135 r
F P
0 g
s
.7978 .37634 m .78844 .3688 L .80054 .3737 L closepath p .154 .464 .163 r
F P
0 g
s
.79028 .36339 m .80064 .37072 L .78855 .36583 L closepath p .167 .292 .081 r
F P
0 g
s
.80054 .3737 m .78855 .36583 L .80064 .37072 L closepath p .162 .407 .135 r
F P
0 g
s
.78907 .34327 m .79003 .33708 L .79188 .33767 L closepath p .249 .836 .139 r
F P
0 g
s
.79997 .26264 m .79537 .2808 L .78914 .28988 L closepath p .238 .95 .181 r
F P
0 g
s
.79997 .26264 m .78914 .28988 L .79372 .27172 L closepath p .238 .95 .181 r
F P
0 g
s
.79174 .30542 m .78914 .28988 L p .79176 .30543 L .234 .846 .182 r
F P
0 g
s
.79176 .30543 m .78914 .28988 L .79177 .30543 L .234 .846 .182 r
F
.79177 .30543 m .78914 .28988 L .79178 .30543 L F
.78914 .28988 m .78958 .28924 L p .79178 .30543 L F P
0 g
s
.78942 .36288 m .78945 .3628 L p .79028 .36339 L .123 .349 .154 r
F P
0 g
s
.78945 .3628 m .79185 .35539 L p .79202 .36463 L .123 .349 .154 r
F P
0 g
s
.78958 .28924 m .79537 .2808 L .79199 .30607 L p .234 .846 .182 r
F P
0 g
s
.79178 .30543 m .78958 .28924 L .79191 .30549 L .234 .846 .182 r
F
.79191 .30549 m .79392 .31954 L .78986 .31691 L closepath p .242 .745 .137 r
F P
0 g
s
.78986 .31691 m .79392 .31954 L .79238 .32779 L closepath p .237 .809 .169 r
F P
0 g
s
.79423 .32839 m .79003 .33708 L .79238 .32779 L closepath p .243 .915 .171 r
F P
0 g
s
.79423 .32839 m .79188 .33767 L .79003 .33708 L closepath p .243 .915 .171 r
F P
0 g
s
.79823 .25898 m .79028 .25871 L .7923 .2557 L closepath p .204 .841 .217 r
F P
0 g
s
.79372 .27172 m .79028 .25871 L .79823 .25898 L closepath p .21 .801 .203 r
F P
0 g
s
.80732 .2502 m .79116 .25176 L .80295 .24879 L p .108 .791 .249 r
F P
0 g
s
.79592 .25312 m .79116 .25176 L .80732 .2502 L p .108 .791 .249 r
F P
0 g
s
.79116 .25176 m .79592 .25312 L .7923 .2557 L closepath p .141 .829 .249 r
F P
0 g
s
.79791 .35302 m .8189 .36955 L .79185 .35539 L closepath p .134 .312 .131 r
F P
0 g
s
.79185 .35539 m .8189 .36955 L .79202 .36463 L p .123 .349 .154 r
F P
0 g
s
.80571 .23984 m .79197 .2276 L .79809 .22698 L closepath p .197 .403 .082 r
F P
0 g
s
.79197 .2276 m .80571 .23984 L .80215 .24198 L closepath p .179 .464 .135 r
F P
0 g
s
.79202 .36463 m .8189 .36955 L p .7935 .36567 L .117 .418 .179 r
F P
0 g
s
.79307 .2212 m .79214 .21743 L .79885 .21941 L closepath p .219 .572 .118 r
F P
0 g
s
.80055 .25548 m .7923 .2557 L .79592 .25312 L closepath p .145 .896 .247 r
F P
0 g
s
.79823 .25898 m .7923 .2557 L .80055 .25548 L closepath p .197 .845 .223 r
F P
0 g
s
.79423 .32839 m .79238 .32779 L .79392 .31954 L closepath p .248 .869 .152 r
F P
0 g
s
.79307 .2212 m .79885 .21941 L .79978 .22317 L closepath p .219 .572 .118 r
F P
0 g
s
.79307 .2212 m .79978 .22317 L .80219 .22818 L closepath p .201 .354 .044 r
F P
0 g
s
.79307 .2212 m .80219 .22818 L .79809 .22698 L closepath p .201 .354 .044 r
F P
0 g
s
.79454 .3765 m .79824 .37842 L .79333 .37604 L .037 .6 .232 r
F
.7935 .36567 m .81657 .37389 L p .79924 .36973 L .093 .478 .209 r
F P
0 g
s
.8189 .36955 m .81657 .37389 L .7935 .36567 L p .117 .418 .179 r
F P
0 g
s
.79997 .26264 m .79372 .27172 L .79823 .25898 L closepath p .238 .953 .18 r
F P
0 g
s
.79625 .37641 m .79824 .37842 L .79454 .3765 L .037 .6 .232 r
F
.80743 .25023 m .79592 .25312 L p .80732 .2502 L .108 .791 .249 r
F P
0 g
s
.80418 .25291 m .79592 .25312 L .80743 .25023 L p .083 .805 .229 r
F P
0 g
s
.80055 .25548 m .79592 .25312 L .80418 .25291 L closepath p .145 .896 .247 r
F P
0 g
s
.80036 .21461 m .79728 .21262 L .79617 .2 L closepath p .241 .708 .068 r
F P
0 g
s
.80036 .21461 m .79617 .2 L .79856 .20555 L closepath p .22 .603 .021 r
F P
0 g
s
.79971 .37859 m .79824 .37842 L p .79625 .37641 L .037 .6 .232 r
F P
0 g
s
.7965 .3764 m .79971 .37859 L p .79625 .37641 L .037 .6 .232 r
F P
0 g
s
.8257 .38302 m .79971 .37859 L .7965 .3764 L p .017 .544 .224 r
F P
0 g
s
.79775 .37634 m .8257 .38302 L p .7965 .3764 L .017 .544 .224 r
F P
0 g
s
.80222 .22321 m .79728 .21262 L p .8029 .22398 L .231 .559 .032 r
F P
0 g
s
.79728 .21262 m .80165 .21857 L .8029 .22398 L p .231 .559 .032 r
F P
0 g
s
.79885 .21941 m .79728 .21262 L .80222 .22321 L p .231 .559 .032 r
F P
0 g
s
.80036 .21461 m .80165 .21857 L .79728 .21262 L closepath p .212 .446 0 r
F P
0 g
s
.79793 .37638 m .79775 .37634 L p .7978 .37634 L .061 .561 .234 r
F P
0 g
s
.7978 .37634 m .82111 .38043 L .79793 .37638 L .061 .561 .234 r
F
.7979 .37625 m .82111 .38043 L p .7978 .37634 L F P
0 g
s
.82111 .38043 m .7979 .37625 L p .7981 .37606 L .094 .474 .208 r
F P
0 g
s
.79791 .35302 m .80105 .35461 L p .80167 .35498 L .104 0 0 r
F P
0 g
s
.79791 .35302 m .82503 .36717 L .8189 .36955 L closepath p .134 .312 .131 r
F P
0 g
s
.82111 .38043 m .8257 .38302 L .79793 .37638 L p .061 .561 .234 r
F P
0 g
s
.80571 .23984 m .79809 .22698 L .80219 .22818 L closepath p .206 .404 .06 r
F P
0 g
s
.80383 .37404 m .7981 .37606 L .79811 .37604 L .094 .474 .208 r
F
.80383 .37404 m .82111 .38043 L p .7981 .37606 L F P
0 g
s
.80383 .37404 m .79811 .37604 L .79814 .37601 L .094 .474 .208 r
F
.80383 .37404 m .79814 .37601 L .79881 .37537 L F
.80269 .25592 m .79823 .25898 L .80441 .253 L p .213 .995 .208 r
F P
0 g
s
.80441 .253 m .79823 .25898 L .80055 .25548 L p .213 .995 .208 r
F P
0 g
s
.79997 .26264 m .79823 .25898 L .80269 .25592 L closepath p .223 .966 .207 r
F P
0 g
s
.8012 .37307 m .80383 .37404 L p .79881 .37537 L .094 .474 .208 r
F P
0 g
s
.80593 .22739 m .79978 .22317 L .79885 .21941 L closepath p .225 .585 .111 r
F P
0 g
s
.79924 .36973 m .80209 .37131 L .80064 .37072 L .093 .478 .209 r
F
.81657 .37389 m .80209 .37131 L p .79924 .36973 L F P
0 g
s
.80835 .2324 m .79978 .22317 L .80593 .22739 L closepath p .208 .365 .018 r
F P
0 g
s
.80835 .2324 m .80219 .22818 L .79978 .22317 L closepath p .208 .365 .018 r
F P
0 g
s
.80351 .22432 m .80165 .21857 L .80036 .21461 L closepath p .168 .372 0 r
F P
0 g
s
.80064 .37072 m .80128 .3728 L .80054 .3737 L closepath p .245 .715 .113 r
F P
0 g
s
.8012 .37307 m .80054 .3737 L .80059 .37364 L .094 .474 .208 r
F
.80055 .25548 m .80418 .25291 L .80715 .25124 L closepath p .118 .874 .177 r
F P
0 g
s
.8011 .37303 m .8012 .37307 L p .80059 .37364 L .094 .474 .208 r
F P
0 g
s
.80209 .37131 m .80075 .37107 L p .80064 .37072 L .093 .478 .209 r
F P
0 g
s
.80075 .37107 m .80266 .37141 L p .80079 .37122 L .094 .474 .208 r
F P
0 g
s
.80978 .24783 m .80079 .24563 L .80215 .24198 L closepath p .143 .425 .16 r
F P
0 g
s
.80912 .24959 m .80295 .24879 L .80079 .24563 L closepath p .079 .504 .221 r
F P
0 g
s
.80978 .24783 m .80912 .24959 L .80079 .24563 L closepath p .143 .425 .16 r
F P
0 g
s
.80079 .37122 m .80266 .37141 L .80107 .3721 L .094 .474 .208 r
F
.80105 .35461 m .8045 .35646 L .80167 .35498 L .104 0 0 r
F
.80105 .35461 m .80364 .35591 L p .8045 .35646 L F P
0 g
s
.80107 .3721 m .80266 .37141 L .80128 .3728 L .094 .474 .208 r
F
.81657 .37389 m .82111 .38043 L .8011 .37303 L p F P
0 g
s
.8011 .37302 m .81657 .37389 L .8011 .37303 L .094 .474 .208 r
F
.80266 .37141 m .81657 .37389 L p .8011 .37302 L F P
0 g
s
.80266 .37141 m .8011 .37302 L .80128 .3728 L .094 .474 .208 r
F
.8029 .22398 m .80165 .21857 L .80369 .22486 L p .203 .502 0 r
F P
0 g
s
.80369 .22486 m .80165 .21857 L p .80369 .22486 L .168 .372 0 r
F P
0 g
s
.80369 .22487 m .80369 .22486 L .80165 .21857 L .168 .372 0 r
F
.80165 .21857 m .80351 .22432 L .80369 .22487 L p F P
0 g
s
.80369 .22486 m .80369 .22486 L .80165 .21857 L .168 .372 0 r
F
.81149 .2468 m .80215 .24198 L .80571 .23984 L closepath p .168 .387 .12 r
F P
0 g
s
.81149 .2468 m .80978 .24783 L .80215 .24198 L closepath p .168 .387 .12 r
F P
0 g
s
.80571 .23984 m .80219 .22818 L .80981 .24103 L closepath p .206 .404 .06 r
F P
0 g
s
.80981 .24103 m .80219 .22818 L .80835 .2324 L closepath p .212 .39 .028 r
F P
0 g
s
.80479 .25276 m .80269 .25592 L p .80441 .253 L .213 .995 .208 r
F P
0 g
s
.80912 .24959 m .81016 .25111 L .80295 .24879 L closepath p .079 .504 .221 r
F P
0 g
s
.80369 .22487 m .80351 .22432 L .80398 .22519 L p .211 .445 0 r
F P
0 g
s
.80364 .35591 m .80476 .35659 L p .8045 .35646 L .104 0 0 r
F P
0 g
s
.8079 .25039 m .80418 .25291 L .80756 .25028 L p .106 .842 .169 r
F P
0 g
s
.80715 .25124 m .80418 .25291 L .8079 .25039 L p .106 .842 .169 r
F P
0 g
s
.80756 .25028 m .80418 .25291 L p .80743 .25023 L .083 .805 .229 r
F P
0 g
s
.81347 .24737 m .81149 .2468 L .80571 .23984 L closepath p .2 .344 .041 r
F P
0 g
s
.81347 .24737 m .80571 .23984 L .80981 .24103 L closepath p .2 .344 .041 r
F P
0 g
s
.80815 .25047 m .80715 .25124 L p .80795 .2504 L .106 .842 .169 r
F P
0 g
s
.80794 .2504 m .80795 .2504 L .80715 .25124 L .106 .842 .169 r
F
.8079 .25039 m .80794 .2504 L .80715 .25124 L F
.80994 .23602 m .80981 .24103 L .80835 .2324 L closepath p .209 .491 0 r
F P
0 g
s
.8139 .25097 m .80912 .24959 L .80978 .24783 L closepath p .214 .746 .185 r
F P
0 g
s
.8139 .25097 m .81016 .25111 L .80912 .24959 L closepath p .217 .721 .174 r
F P
0 g
s
.8139 .25097 m .80978 .24783 L .81149 .2468 L closepath p .211 .751 .19 r
F P
0 g
s
.80981 .24103 m .80994 .23602 L .8112 .24343 L p .209 .491 0 r
F P
0 g
s
.8139 .25097 m .81245 .25177 L .81016 .25111 L closepath p .217 .721 .174 r
F P
0 g
s
.8139 .25097 m .81149 .2468 L .81347 .24737 L closepath p .213 .753 .188 r
F P
0 g
s
.8139 .25097 m .81347 .24737 L .81423 .24912 L closepath p .22 .77 .184 r
F P
0 g
s
.83866 .37761 m .81657 .37389 L .8189 .36955 L closepath p .102 .343 .168 r
F P
0 g
s
.83582 .38389 m .81657 .37389 L .8363 .38197 L closepath p .061 .419 .209 r
F P
0 g
s
.83582 .38389 m .82111 .38043 L .81657 .37389 L closepath p .066 .418 .207 r
F P
0 g
s
.83866 .37761 m .8363 .38197 L .81657 .37389 L closepath p .102 .343 .168 r
F P
0 g
s
.83955 .37595 m .83866 .37761 L .8189 .36955 L closepath p .124 .23 .11 r
F P
0 g
s
.83955 .37595 m .8189 .36955 L .82503 .36717 L closepath p .124 .232 .111 r
F P
0 g
s
.83582 .38389 m .8257 .38302 L .82111 .38043 L closepath p .022 .497 .225 r
F P
0 g
s
.8315 .38438 m .8257 .38302 L .83582 .38389 L p .022 .497 .225 r
F P
0 g
s
.83941 .38624 m .8315 .38438 L p .83582 .38389 L .022 .497 .225 r
F P
0 g
s
.8363 .38197 m .8432 .38874 L .83582 .38389 L closepath p .015 .215 .178 r
F P
0 g
s
.84452 .38374 m .8363 .38197 L .83866 .37761 L closepath p .058 .137 .134 r
F P
0 g
s
.8363 .38197 m .8437 .38682 L .8432 .38874 L closepath p .015 .215 .178 r
F P
0 g
s
.84452 .38374 m .8437 .38682 L .8363 .38197 L closepath p .052 .192 .155 r
F P
0 g
s
.83955 .37595 m .84452 .38374 L .83866 .37761 L closepath p .106 .151 .099 r
F P
0 g
s
.83955 .37595 m .84541 .38207 L .84452 .38374 L closepath p .106 .151 .099 r
F P
0 g
s
.8437 .38682 m .84566 .38839 L .8432 .38874 L closepath p .089 .542 .225 r
F P
0 g
s
.8437 .38682 m .84452 .38374 L .84648 .38531 L closepath p .15 .594 .204 r
F P
0 g
s
.8437 .38682 m .84648 .38531 L .84566 .38839 L closepath p .15 .594 .204 r
F P
0 g
s
.84541 .38207 m .84648 .38531 L .84452 .38374 L closepath p .196 .563 .151 r
F P
0 g
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[59]:=",
  ImageSize->{485.562, 308.938},
  ImageMargins->{{1, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40007U0001=0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7
iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7
iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7
iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7
iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?l7
iPL00?l7iPL00?l7iPL00?l7iPL00?l7iPL00?`70`3V1`00n`L60>@7003j1`H000>000L01@3M1`00
n@L40003N`0000T0g0L00?P70`000e<000020004P000D`P0f@L00?073P000eT000030003DP0000H0
f0L00>d70`000g40MP020003C@00008000=k07/00P000g/0N`0;0=P7003/1`P000==05801@001E80
001B00h0ePL00>`700@0C@1=0P001gH0059f07H00P02Na<0e@L00>`710000e80000:0007N`1k07/0
N`080004DP000=D7003[1`03075=00<000ICMP00DgH5009k00@0N`1k1P000h0000050=D7003[1`<0
00E=04efC@020003MU8000<000=k00000P0015T007/40005F@0005T00`3E1`00j`L30009L@0007H0
MP1f008000Af07]k0P02N`0307/000<000Ek0000N`030=H7003[1`8000=a000010000e9f00020003
N`00008000EI0000F@020003N`1k00L0ePL00>/70P001Dd0001C00<000=C00000`02N`0607]kN`1k
0`000g/0P0020003N`000=L7003[1`<000M=MP1fDP1B00<000QB001k07]IN`@000EI001kF@040=P7
003[1`@000=C05<00P001GIC001k008000Ak07]k0P03N`8000>0Nh000`3I1`00j`L20004MP00C@<0
00=BMP000P0017/005T20005N`1kFG/0203I1`00j`L20004D`00MP8000AC001k0P0017H00002N`09
07]kN`1k081k00<0:0L30:l7003[1`8000=f07H00P001E9f001B008000Qk001INeT0F@8000Ak07]I
0`0X1`<000=f00L010L30:H7003[1`<000=f07H00P001U<0Ng/0N`8017/200=k00<0Nh000`0V1`l0
0`L20:47003[1`8000If001=058:0009N`00FG]IN`1k00<09`L20005DP0005T020000`L000030:07
003[1`8000IC0000Dg/30003N`00009k1005N`8000>000L09PL40003MP0000P000=a000010021`80
W0L00>X70`000gI=0003000;MP00DP1B059kDW/00`001G]I001k00809@L40003N`0000H000=B0000
1@000eT0C@0509/7003Z1`<000=CMU<00P000gICMP020004M@1e00=k0P000g/000020003N`070287
1@000e<000020003N`0000X000=k00001P2K1`00jPL30005MTd007H00P000dd0DP02000>N`1kDP1I
05UkF@1IN`0Q1`8000==CTd00`000e80DP020005N`0007/030001580002K1`00jPL30007DgH0MP1f
DP050005MG/0N`000W/00`1kN`0302870P000g5=L@03009k1@0017H007/20004N`00N`T0V`L00>X7
00T0CP1f04i=04d020002G/005Uk07]INP0202470P000dh0CP030003Ne9k00@000=B05801`000eT0
000609`7002S1`80A@L30004D`1=MP8000A=001k0`0017H00002N`0507]kNP008PL0101a07470003
NgH0009k00<0N`000P0017/007/30003N`000080W@L00:871@051`80?0L01@1>001=00<000U=001f
DWIB04d00P0017D0NeT20003NP07024700H0C@1=04d40006Ne8007]B0P002W/0001k07/0NeT20003
N`0709d7002Q1`L00PL500<700<01`L0=PL40004C@00MP8000=eNgD01@000gD00005028700H0L@1a
07H20003MW/0008017/00`1kN`020005N`00P7/01@2N1`00X@LA0003C@0003H70P000di=00030003
MP00008000EBMU80DP0802470P002Td007H0MP1kDW/;0003N`0000@000=I00L0WPL009h72`000e80
000:03H700D0CGH0L@090003MP1k008000=e00000`0Q1`<000M=001fD`1f00<017/2009k00<0N`00
10000h00002P1`00W0L20003C@0000D000=I00002P000g/0000403870`001Dh004ef008000ABMU9f
1@000gH0000302470P001Dd0001B00<000IBMP1kDW/20008DP1B059k07/20003F@000:07002L1`80
00=f00002@000g/000080003C@0000<0<@L60006MTd0M@1e0`0017Ek07/602470P000g5=00020005
D`1fN`000g/2009k00H0N`1k07/20003N`000:47002L1`/000=k00001P000e80F@0803470`000aP0
CP02000>DWH0059fDWIB059k07/402070P001De>07IB008000AfDP1B0P002W/0Ne80DP1B05T20003
N`000:47002M1`<000=f000030000gH000050004MP00034710001G5=001f00L000QkM@00NgD0M@80
80L20007L@00DgH0MP020005N`1kN`0017/30004N`00P0<0X@L009d710000gH0N`080003F@0000<0
00=I00001P0a1`0601P004e>0`0017H0MU880004Ne80DP807`L00`1>C@020004MP1fDP<000QBNe80
DW]BN`@000EI0020002R1`00W@L90003N`0000L00W/00`1k00020003N`00034700<0L@0010001W40
07D0M@<000IeNgD0MG/40003NP0701h700D0C@1CN`020004DgH000=k00<0Ng/00g/3009k102R1`00
W0L50006MP00DP1B0`00158007/80004F@0003470P001AP004e>008000iBMP00DP1B059f07]BN`D0
7PL20004CP1fDP8000=fDP0020000e9kDP020003N`0000<0=0L206`7002K1`0307H000H000=k0000
0W/6009k00<0N`001@0b1`070740L@1aC@06000;NgD0M@1eN`1k07/00P000gD01`0M1`<000EkDg/0
MP040003MW]f008000=kMP001P000g/0000b1`P0J@L009X70P000deC00040006MP0007]B1P001G]B
001k008000=I00L0<PL5000DC@00059f07IBMP1fDP1B0580DWH301h70`000e9f0002000@MP1k07]B
0580DW]BNe80DPD0<@L60003J`000080I`L009X70`001dh0D`1C05<00P03N`@00W/20003N`0000@0
<`L20003CG40008000Af07IB0`001GD0MG]e00<000=eNgD00`0N1`8001MC001f07IB07Ik07/0NgH0
MW]f07/0N`020003N`0702l73P1U1`00VPL30003D`0000L000AkDP1B0P000g]B00020005F@00P000
=0L20003CP00008000M=0580DWIB00`000=j00007@L3000BCGI=04ef07]B0580DW]B059k0P000e9k
0003030720001Tef001B06D7002J1`<000Y>0000D`1CN`1k1P02N`800W/303D71@002G40001f07]e
N`030007NgEk07]eN`0501d70`004WIBMP1f07]fNgH0MW]fNgH0MP8000Ik0000N`0_1`<000Uf07H0
MTd0CGH00P000gD01`1S1`00VPL20004C@00D`<000EfDP1BN`020003DP1B00@000=k00000P0e1`80
00=>000010003e80DWIB059f07IB059fDP020003N`0001d70P001GH0MTef008000ABNe9k10000e9k
00020003Ne8000<0;`L50006LDd0CGI=0`000e80001T1`00VPL20003MP1f00D000Ak07]f1002N`03
07]k008000=k00L0=@L00`1a00060004MP1kM@8000ekMG/0NgD0M@1eN`1k008070L20005DWH00580
0P000g]f00020003L00000D00W/0101k000_1`@000MfCGH0MTef00@000=e0000H`L009X70P000def
00050003DP1B008000Qk001BN`1kF@@0=PL40003CP0000<000]fDP1B059fDWH0MP020005DWIB07D0
0P0L1`<000==MTd00P002G/0001`07E=M@020003C@1=00<000=I00L0;PL<0004MU80DP<0H`L009X7
0`0015<007H30003N`1k00<000Ak07/00W/20003P00003H70P000dd00005000:M@1eN`1kMG]eN`<0
00IkMG/0NgD201/70`0017IBMU820003MP1f00/000Af0000;PL50003CGI=00<000MeMWEf07]e0080
H`L009X70`0017H004d40003N`1k008000EB0580F@020004P7X003H70P000dh00005000;MP1f07IB
MU9fDWH020000gD01`0J1`<000==MP001P000g40MP040004DP00MP@0;@L6000>MP0007IB059fDP1B
05820687002J1`<000ACMP1f10000gH0MP020008N`00N`1k07/303H700<0MTd0200027]e07EkMG]e
0`0017EfMGH20004M@0001X700<0CP000P000g40000=0004N`1kM@80;0L50009C@00MTd0MGIe0080
00Mf07/0NgD00687002K1`8000EfC@1=D`040005DW/0Ne800`001ET0NP1I0080=PL01@1=CP1>00D0
015fDWH0MU9fDP1=MTd0DWH0MP030003MDd001T71P001GI=04ef00<000=f00000P000e80000502/7
10000dh0L@080003DWH0008000AkDW/0HPL009/70P001e=f001>05<00P001G/007Ik00800W/20004
Nh0003H700<0MP000P000gH000020006MG]e07Ek0P003e9fDWH0MWD0MGH0MP1f000I1`<000AfDgIC
2@001WEfM@1eN`<000=k00L0:@L20003LH0000D000If0000MWD60004NgD00647002K1`8000=fCGH0
0P00258007]B059k0`001W/005UkF@80=PL00`1=0002000MC@1=059fDP1BMP1fCGI=MTef07H0MTd0
CGD0M@0060L40003MTef00@000E=MP00DP060003Ne800080:@L20003L@1a00H000A=001f0`001GI=
MP1e00@0H@L009/70P000e=fD`0:0003N`1k00<000>0Nh000P0f1`0607H0L@1f1@004G/0001fDWIB
MU9fDP1e07Ef00H060L01P1f001CMP@000U=MP1fM@1eMWD00`000gD0NP0202P70`000dd000080003
MP0000<000IfM@00MWD20647002K1`8000=f00000P002E9fDWIB059fDP020003N`00008000=I0000
=`L40006C@00059f1`0027I=MTefC@1=0P000g00000401L700H0C@1=04d3000=C@00059fDP1B059k
DP020003N`0002L710000g4000090005MP00CGH00P0014d007D30607002K1`<000UfC@00MU9fDWH0
0P002GEk001eN`1kNP0303L70P003g40001fC@1eMP1fDWIBMP080005L7I`07001@0G1`8000=f07H0
0P000defC@040005MWD007/00P000gD0000202H70`000dd000050003C@1e008000Mf07Ie07Ef00P0
H0L009/70`000defC@020004MU9fDP8000IfDP1B05820004Ne8003L710001dd0CGH0MU800P0027I=
MTefCGI=10002G1=04e`CG00L00301D700@0CP1>0`000gH000030003MP0000<000=kDW/00P000e80
1`0T1`D000==CP002`000dd000020003MU9f00D0H0L009/70`000gI=MP02000<DWIBMP1k07/0N`1k
0P0017Yk000g1`@000mf001=07D0DP1BMU9fDWH00P002dea07H0MP1fL7I`00L04`L30003CP000080
00Uf07H0MP1fMGH00`001G]e001k00808`L40003L@0000L000Af001B0`000dd0C@060004N`0005l7
002K1`@000mfC@1B059fDP1B0580DW/00P001E80DW/003L70`003WH0001f07H0MP1fCGI=0P001g1=
L4d0CG000P001dd0001=MTd01P0B1`H000ea0740MU9f07IB059k00<000Ak00008PL=0003C@0000<0
00==04d00`000dd0000705l7002K1`@0015=MP1f07IBMP1kM@1e07D0M@020003NgX003L70`0014d0
CGH20006M@1BMP1f0P002TeaCG5=04efCGH70003MWEf00H040L6000AC@1=07EfMGH0MWD0MG]e07X0
0`0Q1`l000EC0000DP040003MP1f008000=`MP00101O1`00W0L2000AC@1=MP1f07IBMP1f07H0Ne80
0P000g/0N`0203H70`0017H0LDd20004MP1fC@8000e`CG1=L4e`071=L01`01<030L50003C@1=00<0
00]BMU9f07H0Ne80F@04020720000dh000080003C@0000<000E=MP00C@020003MDd000<0GPL009`7
0P004GH0MTd0DWH0MU80M@1e07Ek008000=j07X00P0f1`D000==MTd00`000e80C@040006LDeaC@1=
0P00175=LDd20003L01`014020L70003C@0000@000=f00000P0017EkNg/201l70P000f`000050003
L@00008000=C000010000gH000060004M@1`MPD0G@L009`70`0037H0MP1fDP1BMP1fDP8000IkDW/0
Ne8303H7100014d007420008C@1=L@1aCG43000;L000CG1=L4e`CG000`001G00L01`01400`L50009
LDeaC@1=04ee00<000QkDW/0NeUk01h730000g4000040003DP0000<000Q=04ef07H0MP8000I=M@00
C@1M1`00W0L3000<C@1=059fDP1B07Ek0P000gEkM@020004N`0003H70`000def0005000:CG5=L@1a
CG5=LA<000=`070050000dea00020004MP1fM@8000MfM@00NgD0009k00=j00L070L30004C@0060D0
00=f000010000e<000050008DP1B059f07H805`7002L1`<001EfC@1=059f07H0MU9f07]BNe80DW/0
0P0f1`H000==00000P003G5=LDd0CG5=LDd0CG400P001de`CG1=L4d00P003G00001`0700L01`0700
1@000f/000060003CG0000H000IeCGD0MDd40004NeUk01X71P001G40001>00H000==00002@000dd0
00020003MTef00<000Af001=0P1L1`00W0L30008CGH0MP1fDWH20003MG]e00H000AkNP00=PL40003
L@1f00<000Uf001=LDd0CG400P001DefCGI=00<000U=L@00CG5=LDd00P002g00L01`0700L01`0140
00=[000020001WD0Ng]j01P71P001Td001P0L@D000Af07I=0`00159f07H;0005MU8007H0101L1`00
W@L30007CGH0059fDP020008MP0007IBMU830003DP0703@70P000dh000030006LDd0075=10001g40
LDeaCG403@000g1=L0050003L00000<000=`07001`000g00L0050003J`00008000E=0000DP020004
NgX001L72`000dh000030004L@1aC@H000If001=04d20003MP0000H000==MP000P1L1`00W@L30008
MTd007IBMU820007MWD007]eN`020004M@1e03@71@001GH0001f008000E=MTefC@080005LDeaCG40
0P001Td00700L0X000U=L4e`CG1=L4d00P000g0000030003L00000P00g020004MW^001H72@001U80
0740L@<000AfC@1=0P000gI=0004000:DP00059f07IBMP<000Ef0000MP0205`7002M1`<000==MTd0
0P001gIB001fDWH00P000e9fDP030003DP0003@71P000dh0MP020008C@1=LDea0742000>CG5=LDea
CG40L01`CG030003J`1[008000==L4d02P002G00001`0700L00=0007L00007]j000D1`D000=a0000
0P0014d004h50003CP1a00<000UfC@1BMP1f07H0501K1`00W@L01P1>07I=MP@000EB07EfM@020003
NgEk008000Ae07D0<`L20003MP00008000==04d010006GI=MTd0CGI=04eaCG5=L@1aCG40LDeaCG40
1@001G5=LDea008000I=LDeaCG4K0003P007018710000dea00030004DP00L@D000=a00000`000gI=
MP050003MU80008000=C00001P000g]C000505/7002M1`@000MfCGH0MTef008000UBMP1f07IB0580
0P000gD01`0b1`03075=008000Mf0000MP1>00<001QaC@1=LDeaC@1=LDea071=L4d0CG1=L4d40003
L00000L000=`CG0010000dd000060003L01`00`000=j00L04@L40003C@00008000=fC@001@001W40
04e>C@L000E=MP00C@020005N`0007/00`000g/0000905X7002M1`0504h004d00`000e9fDP02000>
MWD0M@1eN`1k07/0N`0a1`8000==MP000`001Dh007I=00<000U=04efCGI=MTd00P006G40LDeaCG40
LDeaC@1=04ef075=LDeaCG401@0037I`MW1f07I`071fL0D000=`07000g0701470`001G40L@1a00<0
019f0000L@1f07H0LDea07I=MTd70005MW/007/00`000f@000060003P5T00080FPL009d70P002AP0
04ef07I=MP02000:DWH0MU80DP1BMP8000=B00L0<0L01@1=001a00<000M>C@1=CP1>00<000AaCG5=
1`0044e`CG1=04e`C@1=MTd0CGH20006CG1=L4e`0P002dd0071=L4d0CG1=008000I`CG1=L4d60003
L01`00@040L30003L@1a00@000==MP000P003Th004d0C@1=CP1fCGI=0P000gI=MP060004FG/0N`80
00=T00001P1N1`00W@L50006MTef07IB0P003gIe07Ef07/0NgD0M@1e000`1`8000A=LDef1P0014ef
07H60004CGI=MP8001==001aCG5=04d0CGI=MP1fCGI=008001=aC@00MW1fL7H0MW1fL01`MW1f0080
00=`MP000P000g0000050003L00000803PL20003CG5=00@000=f07H00`004g4004d0CGH0LDd0CGI=
04d0DWH00P000gIkMP040004N`20N`L0H0L009h710005DefC@1=MP1f07H0MU80DP1BMP1fDP0202l7
0P0014eaCG420003CDi=00<000==04d00`0034eaCG5=L@1a075=L@<000m`001=MTefC@1=MTefCGH0
10001Te`CG1=L08000Q=L01`CG1=L0D000E`0700L0070003L00000`72@000gI=0006000FCG40LDd0
C@1=MTd0C@1=MP1kDP1BN`8000=IN`0010000g/000030607002N1`D000e=MP1fDP1B07D0MGIe0080
00QkMG/0N`1k02h70`0024ea001=MP1a0P0014h007H2000@C@00MTefC@1=04efCGI=MP8000a=MTef
CGH0MTefCGH30003MWD000H000E`0000L0040005L4d0070040091`<000A=L@1a0`000gH00003000L
L@00CGI=04ef0740MTd0CGH0MP1kMW/0N`1k009k0P000g/000050647002N1`0304h0008000efC@1=
MP1f07H0MU9f008000IBMU80DWH202h70`001W5=04efC@@000U>0000CP1aCG400P0074d0CG40LDea
CG5=04d0CGI=MTd0CGI=MP1f07H20007CGI=MTefC@080005L01`07000`000g00L00=00H710000g00
00020003C@00008000==00000P0074d0C@1=LDd0C@1=04ef07H0Ne9kDP1B05T0FG/20006FG/0NeUk
0P1Q1`00WPL2000@CP1f07I=0580DP1e07EfM@8000MkMG/0NgEk0080;0L20007C@1=L@1fC@020005
CP00CW400P005Dd0CGI=MTd0CG5=04efCGI=04efC@02000FMTef07I=MP1fDP1BMU80MGIeMWEfM@80
00Af001=1`0014d0070=0003L00700<74@002Td007H0MP1fCGH20005C@1=MTd00P001g]fNgIkMP00
0W/03@1k07/0N`200800P000H@L009h71P005Def07H0MP1f07IBMU80DWIBMP1fDP0202`70P002G40
LDd0CGH0MP040009CDh0CP1aCG5=00<002U=LDd0CG5=L@1fCGI=MP1fCGH0MTd0CGI=04efC@1=MTef
C@1=L4e`C@020006L4e`CG1=1P001f/0J`1[06/01@000g00000;0003L000008000==LDd00P00275=
L@1aCG5=0`000defC@020003Ne9k00D000MkF@00NeUk008000=I05T00`1Q1`00WPL2000;CP00C@1=
0580DWH00P0017EfMGH60003MG]e0080:`L20006CG40L@1f0P000e80C@020009LDh00740MTef00<0
01iaCG40MTef07I=MTefCGH0MTd0C@1BMU9f07IBMU82000;MP00CGI=MTefCGH00P000dea00080007
L01`06/0J`0=0003L@1a00D000=aCG400`0017I=MTd30006CGH004ef1`000g]fN`0600=k0P000h00
P0020687002O1`@000Qf07H0MP1fC@8000UfDWH0MP1fDWH00`000gH0000[1`8000AaC@1=0`0014ef
04h2000LC@0004d0CG40CP1>075=04ea07I=MTefCGI=MP@000a=MTefC@1=MTefCGH?0009J`1[06/0
J`1[01<000=`000010002dd0CG40LDd0CG5=008000=aC@001@003WIB059kDW]BN`1k07]I0`001UUk
07]IN`<0HPL009l7100024ef001B059f0P002GEf07Ie07EfM@040003N`0002T70`002dea001=059f
DP1=008000Aa04i=0P005WI=04d0CG40L@1f07I=MTefCGI=MTd3000<DWIBMU9f07IBMU9f0P002Td0
07I=MTefCGH3000;L@000700L01`07003`000f`0L@03000=K000CG40L@1fC@1=MP030005MP1fCGH0
0`003GICMP1kMW]fN`1kMP000W/00`1kN`090687002O1`D000==04d00P003GI=001f07IBMU80DWH0
0`0017D0MP0Y1`<000=aC@000P001defCGH0MTd00P000dh0CP020008C@1=CP1>CDh20007C@1=MTef
C@05000DC@0004efCGI=04efCGH0MP1fCGH20003CG1=008000E`0700L00=0003MTef00l000AaCG5=
0P003G5=L@1a07I=04d0CG400`0027H0MU9f07]B0P002W]BN`1k07]INeT20003NeUk00<0HPL009l7
0`0014h0MTd3000BDWIB07D0MGIeMP1f07H0MP1f0`0X1`8000AaCG5=0P000gIB00030003L@000080
00==CP000P001W5=04eaC@<000==00000`002DefCGI=0580DP02000@MU9f07IBMP1fM@1=MTefCA80
00MfCGI=MTef00/000I/06`0LDd5000=CGH0MP1fC@1=MTd0C@03000@DWH0MU<0MP1fNgIkMP1k00=k
0`000g^0N`0206<7002P1`0604h004ef0`000gI=MP020006DWIBMU9f0P00259f07D0ME80:0L20006
CG5=04ef0`000gH0MP090008CTe>C@1=CTd3000PCGI=MTefCGI=MP1fC@1=MP1fCGH0MTd0CGH0MTef
CGH4000CCGI=MTefCGI=MTd0CGI=MTefC@0C0005C@1=MTd010002GI=MP1aCG40L@030003MU9f00@0
00aBNe9kDW/0N`1kFG/20003F@1I00<0H`L00:0710000gI=MP020003DWIB00<000EeMWEfM@020008
MWD0MGH0MP0W1`0401P0L@<000UfCGI=MTd0MDh00`0014i=CTd2000^LDea075=04d0CGI=MTefCGI=
MP1fDWH0MU9f07IB0580MGIe04efCGH0MP1fC@80011fCGI=MTefCGH0MTefCGI=10000f/000030003
K01/00<000=aCG400P0014efCGH2000;CGI=MTd0CGI=04d00P0047/0MU9fDP1fNgIkMW/0N`03N`06
07/0P0200`1S1`00X0L30004CP1fC@8000IfC@1=0584000<MU80DWIB0580DWD09PL2000:L@1aCDi=
MTefC@8000MfDP00CP1>00@000U=CTd0CDh0MTd00P004GI=MTefCGH0MTefC@1=MTef00@001m=MTef
07I=MTd0C@1=MTef07I=MTefCGI=04efCGI=00l000==04d00P000gI=00060009CGI=04eaCG5=0080
00If001BMU83000ADW]BNe80F@1IN`1k081I05T00`1S1`00X@L40003CGI=00<000QBMP1fMGIeMPL0
00EfM@1e000V1`8000m=LDd0CGI=MTefCGIeMTd00`00=Te>001=04ea0740MTefCGH0MTefCGH0MU9f
DWH0MU9fDWIB07EfMGH0MTefC@1=04efCGI=MP8000Y=MTefC@1=MTef2`0021P0001a075=1@001e9f
001B04d02`001U80M@1BMP8000]k07IkMW/0N`1k0002N`0507^008000`1S1`00X@L2000DCP1fCGH0
MTefCGH0MU9fDWH0L4d40005M@00DP009@L20003LDea008000Q=MTef07IBMPT001m>C@1=04d0CGI=
MTef07I=MTd0CGI=MTd0CGI=MTef008001a=MTd0CGI=04ea07I=MTefCGI=04efCGH0MTef0`000dd0
00080005CG40LDd00`002e9fDWH0MTef07I=00<000==LDd010003WH0MP1f07]B0580DW]B0P002G/0
NeUk081IP0020003F@070687002Q1`@000==MP000`002e9fDP1eMWEf075=00<000AeMP1f0`0T1`80
00eaCG40LDefCGI=MTef00X000==LDd00`0024efCGI=MTef0P009Dd0DWIBMU9f07IBMU9f07IB07Ef
07I=04efC@1=MTefCGI=MTd00P001gI=04ef074030000dd0C@040004MU80DP8000=fDP001@000gI=
00060006DP1e05800W/01P1kN`1fN`<017/0101kP7/306@7002Q1`8000E>0000MP02000:C@00MTd0
DWH0L08000==MTd00P001WD0001`02<70P0034eaCDh0MTefC@1=MPD000U>04h0601>CDh00`001GI=
MTef00H002I=MTefCGI=MP1fCGI=04efC@1=04d0CG5=L@1fCGI=MTefCGI=MQ<000Y>0000MU80DWIB
1@001dd0001=04d01P004740MP1fDP1BNe80DW]B0583000;FG]I05V0FH00N`00I0L00:870P0024h0
MTef07IB10001GH0001f00L000=fM@00100Q1`0:601aC@1=LDefC@H000DH01P0L@040004CG5=L@L0
00AfCGI=0`00659fDWIBMU80DWIB059fDWH0MP1f07I=MQ4000EaCG4060070003LDd000<000IfDWH0
MU830005DWIB05802P000g40000500=k00<0Ng/00W/300Ak00<0P7/00P000h1k001T1`00XPL50006
MTd007I=20002gI=001f07IB059f00@07`L30007CG40CP1fC@0>000C601>CDh0L@1fCGI=MTefCGH0
MP02000@CGI=MTef07I=MP1fCGI=MPL001Af07I=MTefCGI=MTefC@1=LDeaC@`000I>07H0MU840003
MP1f008000==MP000P001G40L@1a00L000akDP00Ne9k07]BNe82000<NeUkF@1I05UkFG/0I0L00:87
1@003Tef001BMU9f07H0MTef0P002Dd0MGH0MP1fM@02009`00<01`L070L0200HCG40LDea1@000f/0
J`0:000TCG5=04ef07I=MTefCGH0MU80DP1BMU9fDP1BMP1fDWIBMU9f0P00159fDWH2000;DWIBMU9f
DWIBMU800P00175=LDd20003600H00<000Ua075=L@1fDgH00`002GI=04ef07IBMP020003C@1=00D0
00Q=001f07/0N`<00W/0101kNg/5000:N`0007^0081k06@7002S1`<000Y=04d0C@1=MTef0P002Dd0
07I=059fDP020003MU8000L070L20003LDe>014002HH0000CTd0CG5=04efCGI=MTd0CGH0MTd0CGI=
MP1fC@1=MTefC@<000E=MTefC@02000BMTefCGI=MTefC@1=LDeaC@1=1P0024h0CP1>07IB10001GI=
MP1f008000==MTd00P00175=04d40003L@00008000aB001k07/0N`1kDW/30003NeT0008000MINeT0
FG/006@7002S1`0304h0008001=f07H0MU9fDWIB04efCGH0MWEf00H000=eN`00100J1`8000@HC@1=
10001F`0K01/00L000=>00000P005Dd0CGI=MTd0CGI=MTd0DWIB059fDP020009MU80DWH0MU9f0080
01iB001fDWIB0580DWIBMU9fDWIB04d0CG5=04ea01P40003CG5=00D000==MTd01P0017H005820003
MP0000@000Q=001a07H0N`8000Ie07D0N`02N`<000Ik07]k07/20004P02006D7002S1`8000<H07H0
0P001dd0CGI=MTd00`001Def07IB008000IfC@1=MDd801T75`007QP0001>075=LDeaC@1=MTef07I=
MTd0CGI=MP1fC@8000MfCGH0MTef008001Q=MTef07H0MTefCGI=MTd0CG40LDd0CG4=000<MTefC@1=
MTd0CGI=0P001GH0L@1a00/000YfDWH0Ne80DP1B0P002g/005T0FG/0NeUk008000=I00L0I0L00:<7
10002Dd0C@1=059fDP020004MTd0C@<000EeMWEfM@020004MW1fL08000=a00000P0E1``000E/0000
L@0=000IMTefCGI=MTd0CGH0MU9fDWH0MU9fDWH0MP020007DP00MU9fDP02000IMU9fDP1B059fDWIB
MU80CG5=04ea075=60080004CGH0MP8000A=MP1f100014d0CGH30003MP0000@000==00002`001GD0
001k00800W/20003P7^000D0I@L00:@700D0601fC@02000:L@1fC@1=04efC@H000MfCGI=MP1e00<0
00E`0000L00300l74`000dd000070003600H008001Q=LDeaCG5=L@1fC@1=MTefC@1=MTefCGH4000O
CGI=MTd0C@1=MTd0CGI=04efCGI=04eaC@1=L@1aC@080008LDeaCG40LDd4000;MTd007H0MTd0CG40
1P000dd000050004DW/0N`H000IB05T0FG/20005FG/0Ne800`1U1`00Y0L30006CGH004ef0P001Dd0
CGI=00<000AeMWEf20002TeaCG5=04eaCG4500L74`000g40L@030003K00000d000EfCGI=MP02000/
DWIBMU9f07IBMU9fDWH0MU9fDWIBMP1f07IBMP1fDWH0MU9fDP1=LDea07430003L@0000@0019=0000
CGI=04efC@1=MTd0CGH20005C@0004d010000gH0L@030006N`00MP1f0P0027/0N`1k07/00W/00`1k
N`030004NgD0M@80I@L00:@70`000gI=MP040005CGH0MTd00P002GI=04efCGI=MP020003CGI=0080
00M`0700L01`00<000=[06/05`000dh0CP030003C@0000X000I=04eaCG47000=MTefCGI=MP1fCGI=
MP03000JMTefCGI=04ef07I=04efC@1=MTd0CG5=LDd60005L01`07000P000de`C@020003L@1a0080
00M=001f07I=00D000E=0000C@080004MP1kDP<000AIN`1k0P001e80FG/0NeT00P000g]BN`0406D7
002U1`<000Af001=0`0034d0001B059fDP1eMP<000MfMGIeMWEf00<000Ea0000L00=0003J`1[00/0
00Ma0740L@1a00<000=/00000`007W40001fC@1=MTefCGI=MP1f07IBMU9fDP1BMU9fDP8001Uf07IB
MU80DWIB0580DWIBMP1fDWH0LDea00L000=k000060002E80001=04d0C@060003N`0000@00g/0101k
07/2009k00D0N`00N`020006M@0007X0I0L00:D710001GH0MP1a00<000QfCGH0MTefCA@000E`0700
L00E0007CP0004h0CP0=000dCTe>075=LDeaCG5=04ef07I=MTef07I=MTd0CGH0MP1fCGH0MTef07H0
MTefC@1=MP1aC@8000Mf0000MP1f008000EB001fDP03000:DWIBMP1fCGI=MP@000E=MP00C@060003
MP1a00<000MkDP1B059k00<000Ak07]I0P002W]IN`1k07/0Ne820004N`1k06@7002U1`0301P00080
00if001=MP1fDWH0MU9fDP<000aeMWEfMGIeMWEfMGH20004CGI=MPL000Ma06/0J`1[00L000Ma0740
L@1a00<000Ea0000K007000ZL@00CGI=MP1fCGI=MTd0DgICMP1fDWIB059fDWH0MU80DWH0MU80DWIB
0P0037IBMU9f07H0LDd0MPT001Ue001k07/0MWD0MGIe04efCGH0MU80C@1=008000Mf0000MP1f00T0
0W/200=k1P03N`800W/00`1kN`040003NP0706<7002U1`<000==04d00P002W4004efC@1=MTd2000C
MP1fCGI=MTefCGI=MP1fCGI=MP050003L01`00l000U=04d0CP1>04h01@000dh00009000OCTe>075=
LDd0CGI=MTef07I=MTd0CGH0MTef07I=MP020004CGI=MP8000Y=MTefC@1=L@1a1@001W/0001kDP<0
00eB0580DWIB059f07I=008000IfCGH0MTd40003MP00008000A=MP1f0`001G/0Ne9k008000QI001k
07]IN`<000]k001BN`1k07/0N`0306@7002V1`8001af0000MTd0C@1B059fDP1B059f07IeMWEfMGIe
0P000gI=MP030008MP1aC@1=LDd40003J`1[00<000M/0740L@1a00<000Ea0740L@030003L@0000H0
00A=L@1f0P003Tef07I=MP1fDgICMU=f0P003U9f07IB059fDP1BMU9f0P000e9fDP020004MU9fDP80
00=fCGH00`001W/0N`1k00=k0P000gD000020004MWEfM@D000=BMP001`002Dd0C@1=04efC@020003
MP0000P00W/01@20080000=k1002N`800W/20003NP0706<7002V1`8000a=0000C@1=L@1f07H2000<
CGH0MTef07I=MTef0P000def00020009MTef0700L01`00l000U=04d0C@1>04h01@000dh0CP070006
CTd0CDi=0`007Dea07I=MTefCGI=MTd0C@1=04efCGI=04efC@1B008000=f07H010001TeaCG5=L@D0
00AkDP1B0P001G/0MU9f00d000QfC@00MTd0C@`000=k00000`0017]B05T30006F@1I05Uk10000e80
00020004Ne80NP80I0L00:H7100027H004efC@1B0P003gIBMU80DWIBMP1fMGH0MP030004MTefC@80
00IaCG5=04d40003J`0000<000=/00000`000g40L@060003CG40008000=a00001@002G5=L@1fCGI=
MP02000=C@1CMU=fDgICMU=fD`02000:MP1fDWIBMP1fDP8000=fDWH00`002WI=MTef07I=04h40003
NgH0008000AeN`1k0`001g]fNgIkMW/00P001E9fDWIB008000EfC@1=MP030006MP1f07I=10001GIB
07Ek00@000=k00000P001X00N`1k009k00<0N`000W/30003P00006@7002W1`<000M=04eaCG5=0080
00afCGI=04efCGI=04d40006MTefCGI=0P001G00L01`014000==04d01@001dh0CP1>04h00`0044h0
001>CDi=04e>CDi=CTd80006MTefCGI=0P003gI=MTefC@1=MP1f07IBMP020008CGI=MP1aCG4?000F
MU9fDWIBMU9f07H0MTefC@1=04d0C@8000Qf001=04d0C@<000=B000020000g/000050007NeT0FG/0
N`0906@7002W1`8000E=001fC@04000<MU9fDWH0MU9fDWIB0P0047I=MTefCGI=04eaCG5=LDd:0003
K01/00<000=a074010001deaCG5=04d00P000g40L@05000:LDeaC@1=MTefC@<000IBMU9fDWH8000>
DWIBMU9f07H0MU80DWH30009MTefC@1=MP1a008000=f000010000g]fN`03000=NgEkMW]fN`1fM@1B
MP020003DWH000<000==MTd040000gH000040003N`0000800W/00`1kN`0400=k0P000h00001T1`00
Y`L70003CG40008000YfCGI=04efCGI=0P000gH000020003CGH0008000=`00004`001Dd0001>00<0
00=>04h00`000dh0CP030003CP0000<001Y>CDh0CTe>C@1=04d0CGI=MTefCGI=MTd0C@8000afCGI=
04d0DWH0MU820008MP1fCGH0LDd5000<DP1B0580DW/0Ne9k0P003U9fDWH0MU80CGH0MTef3P000def
C@030004C@1=MP8000ABMP1k0`000g/0N`080006F@0005Uk0`1T1`00Y`L20007D`1=MTefC@030008
DWIB059fDWH20007CG40MTefC@07000@CG5=L@1[06/0MP1a06`0K0D000=a0740100024eaCG40LDea
0`008g4004ea0740LDd0CGI=04ifCP1>059fDWIBMU9fDWH0MU=f008000=BMU800P003WIB0580C@1=
MP1fCGI=1`003gIk07]f07H0MW]fN`1kMP020009MWEf07H0MU9f00@000==MTd010001WIBMP1fC@d0
00Ak001k0P02N`@00W/00`1kN`030003N`0006D7002X1`@000==LDd00P007GI=04ef07I=MP1aCG5=
04efCGH0L01`0700L01`00D000=aC@002@003dh0CP1>04h0CP1>04h0CP030003CP1>00@001m=CP1>
04i=CP1fCGH0MTd0CGI=MTefCGH0MTefCGI=008000Mf07IBMP1f008000A=MTef0P000dd000050006
DW/0Ne9k0P000e8000020008MU80DWIBMU8:0003MP00008000I=MTef07H20005CGH0MTd00P000g00
00030006N`00F@1I1P001W]B05T0DPD0I@L00:P70P009Dh0MTd0C@1=MP1fDP1BMP1aCG5=L@1fCGH0
LDea075=LDea06/00P001DefC@1=008000=/06`00`001g40L@1a07400P0024eaC@1=LDea0`009g40
001a075=L@1a07H0MTifCP1>MTh0DWIBMU9fDP1CMU=fDgICMP030003MU9f00<000AfCGI=20000g]f
0008000:MW]f07H0MGIeMP@000QBMU80DP1=MQ8000I=001a07430005P00007/01@000g/000020004
M@00N`<0I@L00:P700h06000C@1=L@1fCGH0MP<000maCG5=LDd0CGH0L01`07000`000g0000020003
CG5=008000=>00000`001dh0CP1>04h00`000dh0CP030003CP1>00<000]>0000CP1>04i=CP03000H
MTefCGH0MTef07I=MTefC@1=MTefCGI=0P000gH0MP020009DP1=04efCGI=00L000MBN`00DW]B00<0
00MBNe80DP1B00@001UfDP1BMU9f07H0MP1f07I=MTefC@1=04ef00H000==L@000P000g0000020003
FG/000<000AB059k2P1U1`00Z0L2000:CP0007I=04efC@<000U=LDeaCG5=LDd00P002W5=LDeaCG5=
04d20008MTefC@1=LDd40007L@000740L@03000;L@00CG40LDeaCG400`008g4004d0C@1=LDea07I>
MTifCP1>MTif07IBMU9f07ICMU=f008000MCMP1fDgIC008000Ef07I=MP080005MW/0N`000W/01@1f
N`1k0080025f0000MW]fNgIk07IBMP1fDP1=04ef07I=MTef07I=04d00P000g40000<0003N`0000L0
00=e00000P000g/0001V1`00Z0L00`0H00020008LDd0CGI=MTdG0005CG5=LDd00P000dh0CP030007
CP1>04h0CP03000;CP0004h0CP1>04h00P000de>C@03000MC@00CTe>07H0MTefCGI=04efCGH0MTef
07I=MTd00P002gI=MTd0DWIBMP1a008000==00001P001g/0N`1k07/00P001U9k07/0N`@000iBMU9f
DWH0MU9f07I=MP8000A=MTef0P001Tef07I=MP@000E=LDd0C@030004C@00L0D000=kDP000P000e9k
00030003DP000080IPL00:T700/0CP00CGH0MTefC@020009MTefCGI=MTef008000Q=MTefCGI=MP80
00Y=04efC@1=LDea0`000g40L@03000?L@1a0740L@1aC@1=LDea00<000=aCG400`006g4004ea07I>
MP1fCWI>MP1fCWI>MP1fDP1CMP02000>DgICMU<0DgICMP1fCGH30003L@1a00L000QkMP00NgIkMPD0
015f07H0MW]fNgH0DWH0MU9fDP02000;MTefCGI=04d0CGH01P001DeaC@1=00@000=a00001@000f/0
N`050004MP00C@<0IPL00:T710009g5=04efC@1=MP1fCGI=MTefC@1=04efCGI=MP1fCGI=04ea04h0
CP03000=CP1>04h0CP1>04h0CP05000:CP1>075=04e>C@<001I=CP00C@1=MTef07I=MTd0CGI=MTef
0P004Dd007I=MTefCGH0MU9f075=00/000=B00002`004e80DP1BMU80DWIBMP1f07I=MTd03P001W40
001aC@L000E=0000C@0;0003DP00008000=`07400`1T1`00Z@L05@1>001=MP1fCGH0MTd0CGI=MTef
C@020006MP1fCGI=0P001gI=MTef07H02@000g40L@03000JL@000740L@1a075=L@1fCGH0LDd0CG40
LDd3000BCWI>MTh0CWI>MP1fCWI>MTif0P003U=fDgICMU=f07ICMP1f0`001G40001a00`001=fNgIk
07]f07Ik07]fNgH0MW]f008000=fDWH01P002DefCGI=MP1fC@050006CGI=MTef10000dea000:0003
NgD0008000M=MP1f075=0080H`L00:T710007g40MTef07I=MTd0CGI=MTd0C@1=04efC@1=MTefCGH0
1@002Dh0CP1>04h0CP030007CP0004h0CP03000=CP00075=LDd0CDh0CP02000FCDh004efCGI=MTd0
CGI=MP1fCGI=MP@000U=MTefCGI=05800P000g40L@070003CP1f00<001af0000N`1kDW]B0580DWIB
059fDWIB0580DP1=0P0067H0MP1fC@1=MTef075=04d0C@1=LDeaC@<000==04d00P00174004d4000;
L00007H0MP1fCGH00`00175=L@1R1`00ZPL5000CCGH0MTefCGH0MTefC@1=MP1fC@030003C@0000L0
00A=LDea0`001g40L@1a07400`001G40L@1a008000i=04efCGI=MP1aC@1=L@<0029f0000MTifCWI>
04ifCP1>MTif07H0MU<0DgICMU<0D`1=0P001GH0L@1a00<000Ea1`00MP04000EMP1fN`1kMP1f07Ik
MW/0NgIkMW]f00D001IBMP1fDP1=MTd0CGI=04d0CGH0MP1f100024efC@1=MTef0P0014d007430003
L@0000L000AfC@1=0P000gH0001P1`00ZPL00`0H0002000CMP1fCGI=MTd0CGI=MP1a075=L@040007
CGI=MTd0C@02000;CP1>04h0CP1>04h00`001dh0CP1>04h0100034eaCG5=LDd0CDh0CP8000I=Ddd0
CGH2000CCGI=MTd0CGH0MTefC@1B04efC@020003MTef00@000==0000100014h0CPL2000=MP00DP1B
N`1kDW/0N`02000;DWIBMP1fDWIBMU80100037IBMP1fDP1=MTefC@`000iaCG5=LDea075=LDeaC@@0
00=aC@0020000def00040003LDea00<0GPL00:X71P006GI=MTefCGI=04ef07I=04ef07H0MTefCGH0
0P0034ef075=L@1a0740L@<000Ua0740L@1a07400`002WI=MTefCGI=MTd20007LDea07I=MP020003
CWI>008001=fCWI>04h0CWI>05=f07ICMU=f00D000EfC@00L@020003CG5=008000<H000010001gIk
07]fNgH00`0027IkMW]f07Ik1`003WIBMP1fDWH0MTefCGI=10000g]CN`08000=CGH0MTefCGI=MTef
C@030006C@1=MP1a0`001gI=MTefCGH00P000defC@0305`7002Z1`0S01P004d0CGI=MTefCGH0MP1a
CG40LDd0CG40MTef07I=MTd00P000dh000030003CP1>00L000=>04h00P0034d0001=LDeaCG5=L@<0
00I>0000Ddd20005MTefCGH00P000defC@02000BMTd0DWIB04efCGI=MTd0CG5=10001Di=CTe>01/0
00=fDWH00P003580DWIB059fDP1=MP@000=BMP001@000e9f000>0005LDeaCG401@000g1=L0040007
CGH0LDeaC@020004LDd0C@80FPL00:X70P003AP004efCGI=MTefCGH00P002defC@1=04efC@1=0080
015fCGI=04eaC@1=L@1a0740L@050006L@00075=0P002WI=001fCGI=MTd20003L@1f008000]=MP00
CWI>MTifCP020003MP1f008000aCMU=fD`1CMU=fDgH20006C@0004ea30003gIkMP1f07IkMW/0NgIk
MP030004MW]fN`8000UCMP1fDP1BMU8010002g]CNe=kD`1CNe=k00/00X0?0004CG5=L@<000Ua001=
MTefCGH00P0014d0CGH205P7002Z1a0001A=LDd0C@1=LDeaC@1=MTefCGH0CP<000U>0000CP1>04h0
1@001Di=CTe>008000A=LDea0P0014eaCG430006CTd005==0P002GI=MTefCGI=MP05000>MU9fDWH0
MTefCGH0CTd60007CP0007H06005000DNe8007]B059kDP1BMU80DP1BMU820009MP1f07IB059f0080
00AB059f1@0017IB05820003N`1k00T000>0000030001dd0001=L4d00`000dd000020004LDeaC@@0
00AaC@00EPL00:X70P005aP0MTefCGI=MTefC@1=04ef07H0MTef00@000]=MTefC@1=04eaC@020005
L@1a074010003deaCG5=L@1a07I=MTefC@020007MTd0C@1=MP02000?CGH0MP1fCWI>MTifCWI>0080
00ifDgICMU<0DgICMU<0C@8000AaCG5=0P0017I=MTd6000=N`0007]fNgH0MP1fN`02000?MW/0NgH0
MW/0MU<0DP1B008000]kDg]C05=kDg]CN`040008I81TP020I8050009P000N`1fNgIk018000A=001f
2@0017I=001E1`00Z`L2000@CP1fCGI=MTefC@1=L@1aC@8000IaC@00MTd2000AMTefCGH0CP1>04h0
CP1>04h00P0054d0CDi=CTe>C@1=04eaCG5=LDea100014e>04h3000<Ddd0CGI=MTefCGI=0P003gH0
MU9fDWH0MTefC@1=CP02000<C@0004efCGI=MP0H0P002e8007]B059kDW]B00@000efDWIB0580DWH0
MU9f00@000aBNe9kDW/0MU9fDWH30003Nf@0008000=T06@00P000g/0P0020007F@1IN`1kDP080003
MU80010000I=MTefCGH30003L@0005<7002[1`80019=04efCGI=MTef07I=MP1f07H20003CGI=0080
00af07I=MP1a075=LDd2000SL@00075=L@1aCG5=LDd0CG40MTefCGI=MTefC@1=MTd0CGH00P007Tef
07I>MTifCWI>05=fD`1CMU=fD`1CMU=f07H0L@H000I=MTefCDh50007NgH0MW]fN`060003MW]f00<0
00MfN`1fDgIC00H000MkDg]CN`1k00<000QkI01kP7^0N`<000ATN`200P000h0000050006N`1k06Ak
10000e=k1`050003DP0000X000Af07I=0`000dd0001A1`00Z`L2000>CP00CGI=MTef075=LDd40005
MTefCGH00P001Dea07I=008000=>04h01@006Di=CTe>C@1=CTe>CDh0CTe>075=LDeaCG400`001Ti=
CP1>C@8000UCC@1=MTefCGH00P004DefCGI=059fDWH0MTefC@1=00<000E=0000C@0B0006MU9f07IB
10001gIBMP1fDWH00`001G]BNe9k00L000YkDW]B06AkI7]T0P0017/0Nf@30004F@00N`8000MBNe9k
07]B00<000=B00000`000e80DP060007MP1CCE==D`060003CDh000D0C`L00:/70P003Te>07I=MTef
07I=MTef0P004defCGI=MTd0C@1=MP1f075=LDd00`0094eaCG5=LDea075=LDea075=LDd0CGI=MTef
07H0MP1fC@1=MP8000M=MP1fCWI>0080019fDgICMU=f07ICMU<0DgIC04d30005D`1CMU<00`000e=f
D`040003N`0000<000]kMW]fN`1kMW]fN`020005MW/0MU<00P002g]f07IkMW]fNgIk008000Yk06Ak
I7/0P7^00P0016B0I8030009P000N`1k07/0009k00<0Ng/00W/B0003MU=f008000ECMU=fD`0;04h7
002[1`8000a>001=MTef075=LDd2000>MP00CGI=MTef0740LDd7000WC@00CTe>CDi=CTd0CDi=CTd0
CDi=CP1aCG5=L@1fC@1=04e>04i=008001]CCE<0MTd0CGI=MTefCGI=MP1fDWH0MTef04h00P000def
00050006C@0004e>0`003WH0001f07IBMU80DWIB0P007WH0MU80DP1BNe9k07]BNe9kDW/0N`1kF@1B
N`1kI08000AkI7]T10001H0005Uk008000QBN`1k07]BN`<000=k07/010000eT000070005CE==Ddd0
0`0014eC05<40003CDh000<0C@L00:/700`0CTe>C@1=MP1fCGH2000ACGH0MTd0CGI=MP1fC@1=MTd0
0P007WI=MTef0740LDeaCG5=L@1aCG5=04eaCG5=04efC@80011fDgH0MP1fC@1=MTd0CGI=0`004e=f
DgICMU=fDgICMP1fD`1CMU<00`001E=f001C008000MfDgICMU=f00L0011kMW/0NgIk07]f07H0MP1f
0`003gIkMW]fN`1kMW]fN`1k0004N`0306@000P000AT00200P0017^0N`04N`D000R0I01TP020I0:0
00@0N`000`L5009C00=fDgH00`0017H0MTh40003MP1a0080C0L00:/710001Th0MTd0C@8001=fCGI=
04ea07I=MTd0CG40LDea008002m=LDeaC@1=CP1>CDi=CTd0CDi=CP1>CDi=CP1aC@1=MTef04i=04e>
04i=CP1CC@03000CCGI=MTefCGI=MTefC@1BMP1fC@030008CGH0MU80DU<30004De9CDP<0019B0580
DWIBMP1fDP1B0580DWH3000<Ne9kDW]BNe80DW]B0P0027]I05UkFG]I10001G]TNfAk008000=I06@0
0P001W]I001kDP/000ET06@0NP0500X72`000e<0000604/7002[1`0304h0008000=>00000P005Def
CGI=MP1fCGH0MTef07I=04efC@02000JMTefCGH0LDea075=LDeaC@1=L@1aCG5=LDd3000?DgICMU<0
CGI=04d0CGI=00<000=>00000`0045=fDgICMU=fDgIC05=f07H3000;MP1kDg/0MU=fDgH00P003U<0
001BMP1k07]fNgIk10001GH0MW]f00<000afNgIkMW]fN`1kMP03N`0307]k00=k0`001FB0I81T0080
00MkI01k07^000H000IT07^0Nh030003P6@0008000Ak00200P0A1`T000=>00000P1;1`00[0L0101>
04h30014MTefCGI=MTd0CG5=L@1fC@1=L@1aC@1=CP1aCG5=04e>CDh0CTe>CDh0CTd0CDi=CTe>07I=
MTefCGH0CTe>04h0CTd20003MTef008000==MTd00`002defCGI=MTefC@1B00H000=fDWH00`001U=B
De9CDP@000Af07]B20001GIB059f00<000=kDP001@000g]B00020008FG]IN`1kFG/30005NeT007/0
0P002EUkF@1T05UkF@030007I7]T06B0I0030005I800Nf@00P000h1j000301H72@1;1`00[0L2000C
CP1>07I=MTefCGI=04efCGI=MP02000QCGI=04ef07I=04efCGH0LDeaCG40LDeaC@1=04eaCG5=00<0
00]CMU=fDgH0MTefC@02000@MTd0D`1>MP1fD`1CMU=fD`@000MfDgICMP1f00@000=CNe<00P002WH0
05<0DgICMU<20003N`0000D001EkMW]f07H0MW/0NgH0MW]fNgIkMW/00`05N`0507]kN`000W/02`1k
P7^007]TNfAk008000=kP0000P002h00P6@0Nh1k07/000>010000h1kP00301`7101;1`00[0L0101>
04h3000>MTefCGI=MP1aCG5=LDd2000OLDea075=04e>C@1=LDd0CDi=CTe>04i=CP1>04i=CP02000>
C@00MTefCGI=04e>CDh2000DC@1BMP1fC@1=MP1fCGI=MTefCGH30003MTef00@000MB0000DWIB0080
00UfDP1BDe9C07H00`001GH0059f008000ABMP1f0`0037]BN`1kDW]BNe9kDP@000mkFG]I05Uk07]I
05UkFG/00`000g]IN`02000<F@00N`20I01TN`200P001F@0I81T008000F00800P00200030P0701d7
00<01`L0BPL00:`70P000dh00002000YCGI=MTefC@1=MTefCGI=04efC@1=MP1fCGH0MTef075=LDea
CG40LDd00P00;G5=L@1fCWH0MP1fDgICMP1fCGI=MP1f07IC04if07ICMP1fDgICMU=fDgICMP08000<
DgICMU=f07H0MU=f0`001W/0001kMP8000AkMW]f1@0037IkMW]fN`1kMW]fN`8000Ef07]k0002N`04
07]k00=k00@0P7^00P02N`8000aT0000Nh00P7^008030003P7/000@000=k00000`0017/0001/1`00
[0L00`1>0004000>MTefCGI=MP1aCG5=LDd3000HCG40L@1>CDi=CP1aC@1=CTe>CDi=CP1>0P005Dd0
07I=MTd0C@1=MTef04i=CTe>C@02000IMU9f07I=04efCGH0MTefCGI=MTefCGI=MP030007MU9fDWIB
MP020007DWH0MU9fDP05000:DP1B059fDWIBMP8001IBNe9kDW]B059kDW/0Ne9k07]I05Uk0`003W]I
NeT0FG/0N`1kFG]I0P0027]IN`20I81T0`000f@000020004P5V0F@H000>00000K@L00:`70P000dh0
L@02000=CGI=MTef07I=MTefC@020005MP1fCGH00P004TefCGI=04ef075=LDeaCG5=L@@002i>MTif
CP1>MP1fDgIC04efCGI=MP1fDgIC04if07ICMU<0DgICMU=fDgICMU=f10001e=fDgICMU<00P001gIC
05=fDgH00`002W/0N`1k07]fNgH3000;MP0007IkMW/0NgH00P0017]fN`04N`<01G/01P1kP01k00Ak
00<0Nh00300027/0081kP7^00`000h00000306d7002[1`@000=>00000P003defCGI=04eaCG5=04d0
C@02000FLDd004i=CTe>C@1=04e>CDi=CTe>C@8001EfCGI=MTef07I=04efCGH0CTe>CDh00P005e9f
DWH0MP1fCGI=MTd0CGI=MTefCGI=008001=fDP00MU9fDWIBMP1fDWIB059f00@0015B001fDP1BMU80
DW/0Ne9kDP03000SDP00059kDW]B05UkFG]I05UkFG]IN`1k07]I05UkF@1INeT00`002EUkFG/0NeUk
F@020005P020FH002@1^1`00Z@L6000ECG40L@1fCGI=MP1fCGI=MP1f07I=0080019f001=MTefCGH0
MP1aCG5=LDd3000^CP1>MTifCWI>04ifCP1CMP1fCGI=MP1fD`1CMU<0CP1CMU=fDgH0MU=fDgICMP80
01ECMU=f07/0MU=fDgIC05=fDgH0MU<00`005GD0MP1fN`1kMP1fNgH0MW]fNgIkMP020008NgIkMW]f
N`04N`0307/000Ek00<0N`000W/02@1kN`1kP7/0N`02000@P7^0N`1kP7/0N`1k07^0N`8000>00000
1@1]1`00YPL50005CP0004h010004TefCGH0LDeaCG40LDd0CGH0CP8000Q=CTe>CDi=CP8000A=CTe>
0P00;deCCE<0MTefCGI=04efCGH0MTd0CDi=CP1fCGH0MU9f07H0MTefCGI=MP1fCGI=008001MfDWIB
MP1fDWH0MU9fDWH0MU9fDWH0MP060003DWIB00<000]BNe9kDP1BNe9kDP03000NDW]BNe9k07]INeT0
FG]I05UkF@1I05Uk07]I05Uk0P004eT0F@1INeT0FG]I05V0081I05T02`1]1`00Y0L:000HLDea075=
04efC@1=MTef07I=04efC@1=0P0027I=MTefCGI=0P00<g40001fCWI>MTif07I>MTif07I>MTh0DgH0
MTef07ICMU<0DgIC04h0DgICMU=fD`1CMP02000HDgICMU=fD`1CNe<0DgICMU<0DgICMU=f1`000g]f
0002000:MW]fNgIk07]fN`8000Yf07H0MW]fNgH00g/00`1kN`03N`0307]k00800g/07@1k07^0081k
P020081kP020N`1kP020Nh00P02000L000>000L0K0L00:873@0064h0001>07I=MP1aCG40LDea07I=
MTd0C@`001YCCE==DdeCCE==04efCGI=04efCGH0MP1>C@8001AfCGI=MTd0DWH0MP1fCGI=MTefC@80
01YfDWIBMU9fDWH0MU9fDP1BMU9f07IBMU9fDP<000AB001f10000e8000020004Ne80DP8001IkDP1B
N`1kDW]B05UkF@1INeUkFG]I0`007UUkFG/0N`1k07]IN`1kFG/0N`1kF@1IP5T0FH1IP08000=o0000
1@000eT01`1/1`00X0L60003K01/00H0029=L@1aC@1=MTd0CGH0MTefC@1=MTef07H0MTefCGI=MTef
0P00;difCWI>MTifCWH0MTifCP1>MTifCP1C04d0DgICMU=fDgH0MU<0CP1CMU=fDgIC00<001]CMU=f
DgICMU<0Dg]CNe<0DgIC05=fDgICMU<01`0027H007]fNgIk10001gIkMP1fNgH00P0017]fN`02N`03
07]k00Ik0P04N`0307/0008001F0Nh1k07^0N`1k07/0Nh1kP020Nh000P001800080606h7002N1a40
00E>0000CP02000JC@1=L@1aCG5=L@1fCGI=04d0CDi=CTe>CDh2000GC@00DdeCCE==Ddd0CGI=MP1f
CGI=MTd00`0034efCGI=MTefCGH0MP8000I=MTefCGH2000EDP00MU9fDWIBMU9f07IBMU9f07IB0080
00QfDWIBMU9f1`D000UB0000DW]BNe800`005U9kDP1BNe9kDP1BNe80F@1INeUkFG/60003FG]I00<0
00YINeUkF@1INeUk0P0025T0FH1IP0200`001G/0P02000L0KPL009`72P001G40001[00@003Y=LDd0
CG40MP1f07I=MTef07I=MTefC@1=04efCGI=MTd0C@1=MP1fCWI>MTif07I>MTh0CWI>MTif0P003E=f
DgICMU=fDgIC05<00P007GICMU=f07ICMP1f07ICMU=fDgIC05=kDg]CN`1f008000MCMU=fDgIC00@0
01IkMP00NgH0MW]f07Ik07]fN`1kMW]f0`0017H0N`04N`<00W/30003N`0000<000]kP7^0Nh00P7^0
N`030004Nh1kP0<000B0Nh1k2P1_1`00VPLE0005CP1>04h00P00<dd0C@1=LDeaC@1=MTefCGH0CP1>
CDi=CTe>04h0MTef05==DdeCCE<0MTd0CGI=MTefC@02000=MTefCGI=MTefCGH0MP02000LCGI=MP1f
DWIB059f07IBMU9fDP1BMU9fDWIBMP8000EBMU9fDP03000HDWH00580DP1BN`1kDP1BNe9k07/0Ne9k
1@001G/005Uk00<000AkFG]I0P001G/0NeUk00@000YINeUk07]INeUk0P000eT000050004FG]IN`X0
K`L009T72@000f`000050003L@0000<000MaC@1=L@1f008000i=MTefCGH0MTefCGI=MP8001i=MTef
CGH0MTd0CGI=MP1fCWI>MTh0CWH0MTifCWH3000=MP0007ICMU=fDgICMP03000LMU=f07ICMU=f07IC
05=fDgICMP1kDg]CNe=kD`8000=f00000`001G]CNe=f00@000=fNgH00`001GIkMW]f008000EkMP00
N`020003N`0000@01W/01@1k07/000=k00@0Ng]k3@04N`8000Ik07^008070747002H1`d000M>04h0
CP1>00D000E>04h0CP02000?CG5=LDea07I=MTefCGI=008000m>CDi=CTd0CDh0MTefCGH00P003deC
CE<0MTd0CGI=MP1fDP020003MTef00<000IfCGI=MTd3000MCGH0MU9fDWH0MU9fDP1BMU9fDP1BMU9f
DWIBMU800`001e9fDWIBMU80100017H0Ne850004DW]BN`@000=BNe800P0077]B001kFG]INeUkF@1I
05Uk07]I05UkFG]INeT20004NeUkF@8000akFG]IN`1k07]I05T90747002F1a0000Ma0740L@0H00<0
045aC@1=LDd0CGI=MTefC@1=MTefCGI=MP1f07I=MTef07I=MP1fCGI=MTd0CWI>04h0CWI>MP1fD`1C
05=fDgICMP030003MU=f008001aC001fDgICMU<0DgICMU<0DgICMP1kDg]CNe=k0P002U<0Dg]CNe=k
Dg/20004C@00N`8000Ef07IkMP06000;NgIkMP1k07/0N`0027/00`200002N`0407]k00Ek0`04N`03
07/000Ak00P0Ng/0Nh00P08000=k0000101a1`00U@L<0005CP1>04h00`000dh000050005CP1>04h0
0P0044eaCG5=L@1fCGI=MTefCGH2000ICDi=CP1>CDi=04efCGI=MTd0CE<0MP1fC@02000?MU9f07I=
04efCGI=MTef00<002Yf001BMU9fDWIBMP1fDWIBMU80DWIB059fDWIBMP1fDWIB059fDWIBMU83000<
DP00Ne80DP1BNe9k0`00=g]BNe80DP1BN`1kF@1INeUkFG/0N`1kF@1IN`1kFG]IN`1kF@1INeT0FG/0
NeUkF@1IN`1kFG/00`000h0000040747002C1`L000e/0000L@1a0740L@1a00<000=a01P0100014d0
CG42000ACP1=MTefC@1=MTefCGI=MTd00P004WI=MTd0CGI=MP1fCGI=MTefC@8001Uf001>05=fDgIC
05=f07ICMU=fDgICMU=f008001UCMU=fDgICMU<0DgICMU=f07ICMP1kDg]C008000ekDg]CNe<0Dg]C
Ne=k00D0015kMW/0N`1kMW/0NgIk07]f0002N`0307/000Ak0P03N`04081k009k00@0Ng/017/0101k
N`03N`0507]kN`000g/03@1kN`20Nh1kP02008001@1a1`00TPL;0003CP1>00<000E>04h0CP090003
CP1>00<0015>075=LDd0CGI=MTefCGI=MP02000DCDi=04e>CDi=04efCGI=MTefCGH4000EDWIBMU9f
07I=MP1fCGI=MTefCGI=00<001IBMU9fDWIBMP1fDWIBMU9f07H0MU9f0P003e9fDWIBMU9fDP1BMU9f
DP060005N`1kDW/00`002W]BNe80DW/0Ne82000FNeUkFG]IN`1k07]I05UkF@1I05UkF@8001MkFG/0
NeT0FG]IN`1kF@1INeT0FG]IN`020003F@00008000=k0000L@L009472@002G5=001a0740L@060004
C@00CPH001I=04h0CWH0MTef07I=MTefCGI=MTef0P00@4ef07I=MTefC@1=MTefCGI=MTef07ICMU=f
DgH0MU=fD`1CMU=fDgICMU<0D`1C05=fDgICMU<0DgICMU=fDgH2000DDg/0Ne=kDg]CNe=kDg]C05=k
Dg/60005MP1fNgH0100017]fNgH2009k00<0N`001g/20004P7^000=k00<0N`000W/00`1kN`02N`04
07]k00Ak00D0Ng]k0003N`04081kP0<000>000001@1a1`00T0L90007CP0004h0CP040005CDi=CTd0
10000aP00003000FCP1C04i=CP1aC@1=MTefCGI=MTefC@80019>C@1=CTe>CDh0MTefCGI=MTd3000[
DWIBMU9fDP1=MTef07I=MTefCGI=059f07H0MU9fDWIBMP1fDWIBMU9fDP04000AMU9fDWIBMU9fDWIB
MP1fDWH01`000g/0N`030003Ne9k00D000=kDW/00P001EUkFG]I00@000Qk07/0NeUkF@8001]kF@1I
NeUkF@1IN`1kFG]I05Uk07]INeT0FG/00P000gZ000070747002?1`<000Qa0000LDd0C@8000=a0000
0`003gI=MTefC@1=L@1>04h06003000WMTh0CWI>MP1f07I=MTefCGI=MTefC@1=04efCGI=MTd0CGI=
MTef008002mCMU<0DgICMU=f07ICMU=f07ICMU=fD`1CN`1fDgH0MU=fDgIC05=fDgICMU=fD`04000@
Ne=kDg]CNe=kDg]CN`1kD`@000=k00001@001W/0NgIkMP8000Ek07]k0003N`<00W/20004N`1k00Ak
1006N`0307/000Ek00@0Ng/017/01`2007/0P7/01@000gX0001a1`00S@L40007CP0004h0CP05000:
CTd0C@1=CTe>C@8000=>000020001deC04i=CTd00P00CGI=MTefCGI=MTefCGH0CP1>CDi=CTe>07I=
MTef07I=MTef07IBMU9fDP1=MTefC@1=MTefC@1BMU80DWIBMP1fDWIBMP1fDWIBMU9f00@000=BMP00
1@002U9fDWIBMU9f07H20003F@0000<000iB0000DP1BNe80DW]BN`H000ABNe9k0P000eUkF@030003
FG]I008001QkFG]INeUk07/0NeUkFG]I05T0FG]INeT50004NX00P0@000=I00L0L0L008`710001W40
04eaC@@000]aCG5=04ea07I=MP020003CG5=00@000=>01P00P002DifCP1>MTifCP04000:MTefCGI=
MTefC@<000]=MTefCGI=MP1fC@02000lMU=fDgIC05=fDgIC05=fDgICMU<0DgIC05=kDg/0MU=fDgH0
MU=fD`1CMU=fDgH0Ne<0D`1CNe=kDg]C1`000e=k00030003I00000<000=f00000W/3009k1P06N`05
0800P7/00P02N`D01g/200Mk0P05N`<000=k07/00`0017X008030747002;1`h001E>CDi=04e>C@1=
CP1>04h0CP0H01P01@002E==D`1>CDi=CP020003CE<000L000e=MTd0CDi=CTe>CDi=00<004E=MTef
CGI=MTd0DWIBMP1fCGI=MTef07I=059fDWIB059fDWIB059fDP1BMU9fDWH0MU9f07IB059fDWIBMU9f
DWIBMU801@000e9k00030006DP1B059k1@004W]BN`1k07]BNe9k07/0NeT0F@8000IkFG]I05T4000<
NeT007]INeUkFG]I0P000g/000040008FG/007Z008020004NP00074700291`D000Ma0000LDea0080
01==04eaC@1=LDea07H0LDeaCG5=008000=>04h0100024ifCP1>MTif0P002e<0DgICMU=fDgIC00h0
01UfDgICMU=fDgICMP1fDgIC05=fDgICMU=f008001ECNe=kDg/0MU=fDgIC05=f07ICMU<00P005W]C
N`1kDg]C05=kDg]CNe=kDg]CNe<30004I7]TN`8000AC00000W/30006NgIkMW/00g/03P1kNg/0P7/0
N`1k07/00g/00`1kN`04N`l017/20005P02008000P000gX00003078700281`D000E>0000CP04000;
CDh0CTe>04i=CTd00P001dh0CP1>04h00`001AP0000H008000M=Ddd0CDi=008001QC05==DdeCCE==
DdeCC@1=DdeCCE==Ddd3000iCGI=MTefCGI=059fDWH0MTefCGI=MTd0DWIBMU9fDP1BMU9fDWIB0580
DWIB059fDWIB059fDWIB008000UfDWIBMU9fDWH00P001U9k07]BN`D000Ek0000N`02000RDW]BN`1k
DW]BN`1k07]I05T0FG/0N`1kFG/0NeUkFG]IN`8000II001kFG/20007F@1INeUkF@030007F@1I07X0
NP030003NP07078700271`L0019a075=04d0CGI=04ea075=LDd60004LDeaC@8000=>00000`002AP0
MTif07I>MP02000>DgH0MU=fDgICMU=fDgH20008DgICMU=fDgH2000FD`00MU=fDgICMU<0DgIC05=f
DgICMP<0011kDg]CNe=k07ICMU=fDgIC0P0097IC05=kDg]CN`1kDg]CNe=k07]CNe=kDg/0Ne=kD`1T
NfAkI0D00g/01P1k07Ik00Ek0`000h1kP00200=k00<0N`000g/00`1kN`03N`@00g/300=k00<0Ng/0
0g/20005N`00P7/00P000h00000307<700261`<000E>0000CP04000;C@1=CTe>04h0CTd01`000de>
00060003CP0000<001PH05==04e>04i=04eC05==DdeCCE==Ddd20008DdeCCE==Ddd2000HMTef07I=
MTefCGH0MU80CGI=MTef07IB0P003gIBMU9fDP1BMU9fDWIBMP04000EDWIBMU9fDP1BMU9fDWIBMP1f
DWIB008000]kDW]BNe9kDW]BN`060008DW/0N`1k07/50005N`1k07/00P0065UkFG/0NeT0FG/0NeUk
F@1I05T0FG]IN`8000aINeUk07]IN`1k07/20007FH0007X0NP0207@700251`<0011a0000LDea075=
04efCGI=1@004Def07H0LDd0CG5=LDd0CG5=00L0021>MP1fCP1>MP1fDgH0MU=fDgICMU=f07H0MU=f
DgICMP8000ACMU=f0P00>U=fDgIC05=f07ICMU=f07]C05<0Dg]CNe<0DgICMU=fDgICMP1kDg]CNe=k
Dg/0Ne=kDg]CNe=k07/2000=Dg]CNe=kD`1TNfAkI0030005M@00N`000g/300Ek00H0P7/0N`05N`06
07]k07/017/01P1kN`1k00=k0P000g/0N`02N`0;07]kN`1k07/0Nh000`001H0007X007@700241`<0
00E>0000CP040007C@1=CTe>C@03000BCDi=CP1>C@1=CTd0CDi=CP1>1`00:QP0Ddd0C@1=CTd0CE==
D`1CCE==DdeCC@1=04eCCE==Ddd0C@1=MTefC@8000YfCGI=0580CGI=0P00?7IBMP1fDP1BMU9f07IB
MU9fDWIBMP1f07IBMU9fDWIB059fDWIBMU9fDWH0Ne9kDW]BNe9k07]BNe9kDPL000=BN`000P002G]B
Ne9k07]IN`02000XFG]INeUk07]I05T0FG]I05Uk07]I05UkF@1I05T0FG/0NeT0FG/0N`X0M@L00887
1P000g40L@040005C@1=MTd00`001Dd0CGI=008000QfCWI>04eaC@8000MaCG40L@1>00<006<H01P0
MP1f07I>MP1fDgIC05=fDgICMU=f07H0MU=fDgICMP1f07ICMU=fDgH0MU=f07H0MU<0Dg]CNe<0Dg/0
Ne=kD`1CMU=fDgICMP1kD`1CNe=kDg]C05=kDg]CNe=kDg/00`003g]CNe=kDg]C06AkI7]TN`070006
N`0007/00g/00`20N`020004N`0000=k00L0Ng]k07/0009k00D0Ng]k0003N`0307/0009k00H0Ng/0
N`02N`080800P7/0Nh030004P7/007D700211`@000<H04h00`000dh0CP06000GCDi=CTd0C@1=CTe>
CDh0CTe>CDh0CTd00P000dh00003000]6000C@1=04e>C@1=DdeCC@1=DdeCCE==04eC05==DdeCC@1=
04efCGI=MTef0080035=001fC@1BMU9fDWH0MU80DWIBMP1fDWIBMU9f07IBMU80DWIBMU9f07IBMU9f
DWIB008000=k07/00P003E9kDW]BN`1kDW]BNe801@000e8000030004DW/0N`@000IBN`1kDW/20006
F@1INeUk10005eUkF@1INeUk07/0NeT0FG/0N`1k07]I00@000>0F@000P0015Uk001e1`00P0L40003
L@1a008000E=L@1aC@060005MTefCGH00P003TifCWI>MTh0CG5=LDea1@0014h004d4000/MP00CWI>
MP1fDgICMU<0DgICMU=f07IC05=fDgICMP1fCP1CMU=fDgICMU<50012Dg]CNe=kD`1CNe<0Dg]C05=f
DgICMP1kDg]CNe<0Dg]CNe<0Dg]CNe=kD`1CNe<0Dg]C05=kDg]C06AkI7]T07/70P0017/007/300=k
0P05N`0307]k00=k1002N`801G/00`1kN`02N`800g/01P1kNg/0N`8000UkP020N`1k07/01P1e1`00
O`L20003CP0000@000A=CP1>0`001Di=CTe>01d000=>01P01@00L4i=CTd0CE==DdeCC@1=DdeCCE<0
Ddd0CE==Ddd0CE<0MTefCGI=MTefCGH0MU9fDWIBMU80DWIBMP1fDWH0MU9fDWH0MU9fDWIBMU80DWIB
MP1fDWIBMU80DW]BN`1kDW]B059kDW/0Ne9kDW]TNf@60005N`1k07/00`002W]BNe80DW]BNe830003
DW/000@000UkFG/0NeUkFG/00P0025UkFG/0NeUk0P001eT0001I07X010001G]I058007D7001n1`80
00=a00000`000g40L@040009C@1=MTefCGI=00@000IfCWI>MTh20007MP1fCWI>MP020004D`1>D`80
00A=000H1000:DifCWH0MU=fDgICMU<0DgICMU<0DgH0MU=fDgH0MTh0DgICMU=fDgIC00<002YCNe=k
Dg]CN`1kDg]CN`1kD`1CMU=f07]CNe=kDg]CN`1kDg]C05=kDg/2000IDg]CN`1kDg]CNe<0Dg]C06Ak
081kP7/0N`060006NgH0N`1k0P03N`0407]kN`8000=k07/00g/2009k1P03N`0307]k00=k00<0Ng/0
10001W/00800P0<000J0001jN`1e1`00O@L20003CP00008000==CTd00`002Dd0CDi=04e>C@04000@
MTd0CE==Ddd0CE==04e>C@8000ICCE<0CTd30004C@00608002a=CTe>CDh0DdeCCE==D`1CCE==D`1C
CE<0DdeC05==D`1fCGI=MTefC@1BMP8003MBMU9fDWIB059fDWIB059f07IBMP1fDWIBMU9fDWIBMP1f
DWH0MU9f07]BNe9kDP1BNe9kDW]B008000Ik07/0Nf@30003I7]T008000EkDP1BN`020009DW/0N`1k
DP1B008000ikDW/0Ne9k07/0Ne9kDP8000=kDP002`001G/005Uk008000=I07X00`000gX0000307D7
001l1`D000E=04eaC@020009L@00CGI=MP1f008000Y=MP1fDgIC04if0P001DifCWI>00<000ICMU=f
DgH20003CU<000L002]>MTifCP1CMU=fDgICMP1fDgIC05=fD`1CMU<0CWI>MP1fDgICMU<0Dg]C0080
03UkDg]CNe=k07]CNe=k07/0MU=f07]CNe=kDg]CNe=kDg/0N`1kDg/0Ne=kDg]CN`1kDg]CNe=kDg/0
0P002g/0001kP7^0Nh1k00D000=f00000W/00`1kN`0400Ak00L0Ng]k07/000Ak00<0N`0017/200Ak
0P000gIkMP020004N`1kN`8000F007/0P00407H7001k1`8000=>04h01P001dd004i=CTd010002Di=
001fCGI=MP030003DdeC00L000ICCE==Ddd20003CTd000D002LH0000CTe>05==DdeCCE==D`1CCE<0
DdeC05==D`1CCE==04efCGH00P00559fDP1BMP1fDWIBMP1fDWIBMU9f0P004U9f07IBMU9fDWIBMU9f
DWIBMP8001IB001kDW]BNe9kDP1BNe9kDW]BNe9k0`002W]TNfAkI7]TP0L2000DMP0007]BNe80DW/0
Ne9kDW]B0582000DNe80DW]B0580DW]BN`1k07]BNe82000;Ne80F@1I05Uk08000P0015T007/207L7
001i1`<000=a074010000deaC@020003MP0000D000QfC@1CMU=fD`P000=fCP002@001WIC04iCCP@0
00E>001>D`02000gCP1CMU=fDgICMU=f07IC05=fDgH0MU<0CWI>MP1fDgH0Ne=kDg/0MU<0Dg]CNe<0
Dg]CNe=kD`02000BMP1kDg]CNe=kDg]CNe=kDg]C0`0055=kDg]CNe=kDg/0Ne=kDg]CNe=k0P000fB0
00030004Nh1kP0T00g/00`1kN`0300Ak0P04N`0407]k009k00D0Ng]k0002N`0>07]kN`1k07Ik0800
N`02N`800W/607L7001h1`P000==CTd01P001Ti=CP1>C@T001A=MP1CCE==Ddd0CE<0Ddd0CE==D`/0
00HH001=CTd4000IDdeCCE==DdeC05<0DdeCC@1=D`1CCE==D`03000HMU9fDWIB059f07IBMU9f07IB
MU9fDWIB0P002GIBMU9fDWIBMP0:000DDW]BNe9kDW]BNe80DW]BNe9kDW/20007I7]TNfAkI0060003
1`00008000MB001kDP1B00<000AB001k0P007E9kDW]B0580DW]B059k07]BN`1kDP1BN`1k07]I0080
00=kFG/00P000eUk000307T7001g1`H000I=04eaCG440003CGI=008000=fC@000`003GICMU=fDgIC
04ifCWH00P004TifCWH0MTh0CWI>MTh0CDh0CPD000I>05=>Ddh20003MP000080031CMU=fD`1C05=f
DgIC05<0CWI>MTh0Dg]CNe=kDg/0MU=f07]CNe<0Dg]CNe=kDg/;0008Ne=kDg]CNe<4000@Ne=kDg]C
Ne=k07]CNe=kD`8000V0081TP6B0I8000P000f@01`031`T000Ak00000g/300Ak00<0N`0017/0101k
N`03N`0;07]k07/0N`20Nh000P001G/0001k00@0N`L007H71P0014e>04h40007CDh0CTd0C@050003
C@00008001AfCGI=MTd0CE<0DdeCCE==04eCC@8000ECCE<0CP020003CDh000@000DH04i=CP020005
DU=BDe801@003dd005==DdeC05<0DdeCC@02000IMU9fDWIBMU80DWIB059fDWH0MU9fDWIBMP03000@
MU9fDWIBMU9fDWIBMU9fDP8001ekDW/0Ne9kDW]BNe80DW]BNe80I7]T06AkI7]TN`0400H71@000g]B
00020006F@00Ne9k1@006W/0059kDW]B0580DW]BN`1k07]B0580FG]I0`001ET007]I00<0N`L007D7
1P000dd000020006L@00CGI=1@002Th0CWH0MU=fDgH40003DgH00080015fCWI>MTif07I>MTif07I>
MP0200=>1`003diCCP1>05=fDgICMU=fD`03000=DgICMU<0D`1>MTh0D`02000FNe=kDg]CN`1fDgH0
Ne=k07]CNe=kD`80031k001CNe=kDg]CNe=kDg]CNe<0I01CNe=k07]CNe=kDg/0Ne=kD`1TP6@0I81T
P6@400T70P000g/000040003N`00009k00<0N`000g/300Ak00<0N`001G/200=k0P000g^0N`02009k
00<0Ng/00P000gD0001k1`00M0L;0005CP1>CDh010004DeC001=DdeC07I=MTefCGI=00l000M=DdeC
05==008000A>CDi=1`0044e>04h0De9CDU=BDe9CDU<30004DdeCC@8000ECC@1BMP02000bDWIBMU9f
DP1BMU9f07IB059fDWIB059f07IB059fDWIBMU9fDWIBMU80DW/0Ne9kDW/2000EDW]BN`1kDW]B06Ak
I7/0NfAkI01I00802`L30003N`00008000aI0580DP1BN`1kDW/50009N`00DW]BNe9k008000=BNe80
1@001eT0FG/0NeT01@1k1`00M0L50004CG40L@L000mfCP1>MP1fCWI>MTh0DgH02@001gI>MTifCWH0
0`000gI=000700=>2@003e=>04iC07ICMU=fDgICMP020006DgH005=f0P00D4h0Dg]C05<0Dg]CNe<0
DgICMU<0Dg/0Ne=kD`1CNe<0Dg]C05=kDg]CNe=kDg]C06AkI01CNe=kDg]C05=kD`1CNe<0I81TP020
I81T06@040L40007I000N`1k0003N`0307]k00<00W/@00Ak00<0N`000g/507/7001d1`<000=>0000
200014eCCE<30003DdeC00L000AfDWIB0P0024i=04eCCE==0P002U==04e>CDi=CTd60004CTe>C@@0
010H04i=04d0DU=BDe9CDU=B0P0015==Ddd20013D`00DWIBMP1f07IBMU9f07IBMU9fDP1B059fDP1B
MU9f07IBMU80DWIBMU9fDWIB059kDP1BNe9kDW]BNe80DW/0N`020008I7]TNf@0I7/401@71@0027/0
Ne9kDP1B10000g]BN`02000:DW]BNe9k07]BN`8000II001kFG/20007F@00N`1k001l1`00M0L00`1=
00070004MTifCPX000=fD`000P0015=fDgH2000ECWI>MP1fCWI>04ifCWH0MTefCGI=008000=f0740
0P0014ifCWH6000FCU<0Ddh0DgICMU=fD`1>05=fDgICMP8003eCNe=kD`1CN`1kDg]C05=fDgICMP1k
07]C05=kDg]C05=kDg/0Ne=kDg]CNe<0I7]TN`1kDg]CNe=kDg]C00@000V0I81TP6B008000`0H1`03
05<000<017/30003N`00009k00<0N`001G/01P1kN`1k00=k0P001G/0001k00@0O@L007D71P001Ti=
CP1CC@D000eB059fDWIB059fDP1B008000U>CDi=CTd0CE<00P002DeCCE==04e>C@030007CDi=04e>
C@030003CDh000D001E=04e>05=BDe9CDU<0CP1CCE==Ddd00`004e9fDWH0MU80DWIBMP1fDWIBMU80
0P00:GH0059fDWIBMP1fDWIBMP1fDWIBMU80DW]BNe80DW]BNe9kDW]BNe9k00@000ATNfAk1@0J1`<0
00=B00001P0027/0N`1k07]B0P003W]B059k07]BN`1kFG]I1P201`00MPL;000:DgH005=f07ICMP80
00=CMP001`0014ifCWH30009MTifCWI>MP1f008000e=MTefCGH0LDeaCG5=00P001EC05=>D`1fDgIC
MP1CCP1CMU=fDgH00`0057]CNe<0Dg]C05=kD`1CMU=fDgIC0P009W]CNe=kDg/0Ne=kDg]CN`1kDg]C
06AkI7]T05=kDg]CNe=kDg]C0P00181TP6@701d71P02N`<000Ak07]k0P04N`0507/0N`0017/00`1k
00030003P0070887001f1`0604h0DdeC0`002E==D`1fDWIBMP0P000?CDi=CTe>C@1=CTe>CDi=00T0
05U=CTd0DU=BDe80CDh0DdeCCE<0D`1C07IBMP1fDWH0MU9f07IBMU9fDWH0MU9fDWIBMU80DWIBMU9f
DWH0MU80DW]BNe9k07]BNe9kDW]BNe80I7]TNfAkI00501l70`000eT000050003F@1B008000EkDW]B
N`030003Ne8000@000=BNe800P241`00M`L2000=CP00MTh0DgH0MU=fD`0901T71@001Def075=00D0
00==L@001P00:5=>Ddh0DgIC04iCCU<0MU=fD`1C05<0Dg]CN`1kDg/0Ne<0DgICMU<2000UN`1kDg]C
Ne=k07]CNe=kDg]C05<0I7]TNfAkI01CNe=kDg]CN`020005I81TP6@01P0S1a`000=k00000P251`00
N0L30003DdeC00X09@L50007CDi=CTe>C@07001=CDi=CP1CDU<0CTe>C@1=Ddd0CE<0De80DWIB059f
DP1BMP1fDWIB059fDP1BMU9fDWIB059fDWIBMU9fDP1BNe9kDW]BN`1kDW]BNe800P0017]TNf@30004
I00002H70`051`@000EB001kDP0;0003DP0008H7001i1`X0;@L20009LDeaCG5=L@1a00L001iCCU<0
MP1CCU=>D`1fD`1CMU<0DgH0Ne=k07]CNe<20003MU=f008001ACNe=kD`1CNe=kD`1CNe=kDg]CN`@0
019TNfAkI7/0Ne=kDg]C06B0I8020003I80000<0<@L900P70`271`00/PL20008CDi=CP1>04h5000O
CP00C@1B04e>CDi=04eC05==D`1CDU<0MU80DWIBMP02000XDWH0MU9fDWIBMU80DWIBMP1fDWIBMU9f
07/0Ne80DW]BNe80DW]BN`8000=TN`000P000g]TN`0303D71`2B1`00]0L20009LDd00740L@1>0080
00=>Ddh00`006TiCCU=>D`1f07ICMU=f07IC05=k07]CNe=k0`0057]CNe=kDg]CNe<0Dg]C05=kDg]C
0P001W]C05=kD`8000UkI7/0Ne=k08000`00181TP6@40<l7002f1`<000e>04h0CP0H04i=CTe>0080
01U=CTe>C@1=04eCCE==059CDP1B059fDWIB008002AfDWIBMU9fDWIBMP1fDWH0MU9fDP1BNe9k07]B
Ne9k07]B05840006NfAkI7]T0`3A1`00^0L20005L@1a07400P0014iCCU<50003DdiC0080015CMU=f
DgH0MU=f07/0Ne=kD`07000>Dg]CNe=kDg/0Ne<0Dg/2000<Dg]CN`1kDg]CNe=k1P001VB0I81TP0@0
dPL00;T710000dh000020005C@00CTd00P0014h004d2000;DdeCCE==059CDU<00P001E9fDP1B0080
00AfDWIB10001GIBMU9f0080019BMP1kDW]BNe80DW]BNe9kDW/40005I7]TNf@0103D1`00^`L800A>
00D0D`1CCP03000;DgICMU=f07ICMU<00P003g/005=k07/0Ne=kDg]CN`040003D`00008000ikDg]C
Ne<0Dg]CNe=kD`8000B0I81T1`3F1`00_@L00`1=CP02000=CDi=CTe>04h0CTd0C@02000:DdeCC@1B
De9CDP80019fDWH0MU80DWIBMU9fDWIBMU83000HDW]BNe9kDW/0Ne9kDW]B06@0I7]TNfAk103G1`00
_PL20004D`0000A>00T0CU<0Ddh0DgH00`001GH0MU=f00<001=kDg]C05=kD`1CNe=kDg]CNe=k00@0
00aCNe=kDg]C05=kDg/20007I01TP6B0I0040=T700301`<0015>001=CTd0CDh0CP1CCE==D`05000D
De9C07IBMP1fDWIB059fDWIBMU82000HN`1kDP1BNe9kDP1BNe9k07]TN`1kI7]T103J1`00`PL2009>
0P0034h0Ddh0CU<0MU=fD`@001afDgIC05=kD`1CNe=k07]CNe=kD`1CNe<0Dg]C0P001W]CN`1kD`80
00N0I800P6B000@0g0L00<<720007di=04d0CE==D`1C05<0De9CDP1BMP1fDWIBMP1fDWH00P00659k
DW/0Ne9kDW/0Ne80DP1TNfAkI01TN`<0gPL00<D700<0CTh0100075=>04h0DgH0MU<0DgH0MU=f07]C
05=kDg]C05<2000=Ne=kDg/0Ne=kDg]CN`040007I81TP6@0I0030>0700361`8000A=CP1>0P0034d0
05<0DdeCC@1BD`8000YBD`00DWIBMU9f0P00459kDW]BNe80DW]BNe9kDW/40004I7]TN`<0hPL00<P7
00H0CTh0Ddh5000;DgICMP1fDgIC05<00P0017]CNe<2000@N`00Dg]CNe=k07]CNe=kD`8000B0I81T
1@3T1`00b@L30003CDi=00@000YCCE==059CDU=B0`005580001BMP1fDP1BNe9k07]BNe9k0P0016Ak
I7/50>H7003;1`@000=>Ddh01@001E<0DgIC00D0015CNe=kD`1CNe<0Dg]C05=kD`020003P6B000<0
00>00000j0L00<d70P0025==Ddd0CE==1@000e9CDP02000?MU9fDWH0MU9fDP1BN`1k008000=T0000
0P0017]T003Z1`00c`L3000JD`1fDgH0MP1kDg]C05=k07]CN`1kDg]CNe<70004I81TP080k0L00=87
0`005e<0DdeC05=BD`1CDU<0MU80DWIBMU9f00@000ETNfAkI0020>h7003D1`<0019CMU=f07]C05=k
D`1CN`1kDg/40003I81T00D0l0L00=L7100025<0059CDU=B0P002WH0001kI01TNf@30?<7003K1``0
00ATP6B0103f1`00h0L<0?T7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7
003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7
003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7
003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7
003o1nH7003o1nH7003o1nH7003o1nH7003o1nH7003o1nH70000\
\>"],
  ImageRangeCache->{{{0, 484.562}, {307.938, 0}} -> {0.00991744, 0.152541, 
  0.0020447, 0.0020447}},
  CellTags->"DXFFiles"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reading and writing Geomviews OFF files", "Section"],

Cell[TextData[{
  "The OpenGL viewer is somewhat limited in its interactive features. \
Geomview from the Geometry Center is a much more extended sophisticated \
system. One may save the graphics in Geomviews OFF format. They can be reaqd \
by Geomview or directly by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "  for later processing.  Here are two of the sample graphics that come \
with Geomview."
}], "Text"],

Cell[BoxData[
    \(MVClear[]; \ MVLoadOFF["\</export/home/kuska/MathView3D/dragon.off\>"]; 
    \)], "Input",
  CellLabel->"In[40]:="],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVRotate3D[{1, 0, 0}, \(-\[Pi]\)/2]\)], "Input",
  CellLabel->"In[41]:="],

Cell[BoxData[
    \({{1.`, 0.`, 0.`}, \(-1.57079632679489655`\)}\)], "Output",
  CellLabel->"Out[41]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
      MVGetGraphics3D[]\  /. 
        \ {x_?NumericQ, y_?NumericQ, z_?NumericQ} :> \ {x, z, y}, 
      ViewPoint -> {\(-2.338\), \ 1.408, \ 2.000}, Boxed -> False]; \)\)], 
  "Input",
  CellLabel->"In[31]:="],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .95874 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
5.55112e-17 1.01702 -0.00788418 1.01702 [
[ 0 0 0 0 ]
[ 1 .95874 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .95874 L
0 .95874 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.59662 .42098 m .59236 .45411 L .56776 .44162 L closepath p .803 .599 .626 r
F P
0 g
s
.57617 .41353 m .59662 .42098 L .56776 .44162 L closepath p .778 .551 .592 r
F P
0 g
s
.59662 .42098 m .61671 .46826 L .59236 .45411 L closepath p .798 .595 .626 r
F P
0 g
s
.57631 .43322 m .60098 .44579 L .59662 .42098 L closepath p .827 .502 .439 r
F P
0 g
s
.57617 .41353 m .57631 .43322 L .59662 .42098 L closepath p .779 .419 .375 r
F P
0 g
s
.59662 .42098 m .60098 .44579 L .6254 .46001 L closepath p .831 .505 .436 r
F P
0 g
s
.57617 .41353 m .54554 .43409 L .51952 .38819 L closepath p .85 .668 .655 r
F P
0 g
s
.57617 .41353 m .56776 .44162 L .54554 .43409 L closepath p .815 .593 .601 r
F P
0 g
s
.51952 .38819 m .55404 .42563 L .57617 .41353 L closepath p .867 .554 .455 r
F P
0 g
s
.61671 .46826 m .59843 .47386 L .59236 .45411 L closepath p .819 .557 .542 r
F P
0 g
s
.61671 .46826 m .6254 .46001 L .59843 .47386 L closepath p .742 .973 .926 r
F P
0 g
s
.59236 .45411 m .60098 .44579 L .56776 .44162 L closepath p .641 .691 .881 r
F P
0 g
s
.60711 .46558 m .60098 .44579 L .59236 .45411 L closepath p .419 .328 .646 r
F P
0 g
s
.59843 .47386 m .60711 .46558 L .59236 .45411 L closepath p .419 .328 .646 r
F P
0 g
s
.6254 .46001 m .60711 .46558 L .59843 .47386 L closepath p .742 .973 .926 r
F P
0 g
s
.60098 .44579 m .60711 .46558 L .6254 .46001 L closepath p .82 .556 .54 r
F P
0 g
s
.57555 .468 m .58295 .45405 L .59465 .47963 L closepath p .803 .551 .558 r
F P
0 g
s
.57555 .468 m .56889 .44888 L .58295 .45405 L closepath p .818 .571 .566 r
F P
0 g
s
.59465 .47963 m .58295 .45405 L .59189 .46164 L closepath p .012 .145 .658 r
F P
0 g
s
.60098 .44579 m .57631 .43322 L .56776 .44162 L closepath p .641 .691 .881 r
F P
0 g
s
.57746 .44049 m .59189 .46164 L .58295 .45405 L closepath p 0 .129 .678 r
F P
0 g
s
.58295 .45405 m .56889 .44888 L .57746 .44049 L closepath p .666 .738 .904 r
F P
0 g
s
.56776 .44162 m .55542 .45039 L .54554 .43409 L closepath p .819 .557 .542 r
F P
0 g
s
.56776 .44162 m .57631 .43322 L .55542 .45039 L closepath p .639 .944 .718 r
F P
0 g
s
.57555 .468 m .55783 .46621 L .56889 .44888 L closepath p .835 .579 .552 r
F P
0 g
s
.56642 .45783 m .57746 .44049 L .56889 .44888 L closepath p 0 0 0 r
F P
0 g
s
.56889 .44888 m .55783 .46621 L .56642 .45783 L closepath p 0 0 0 r
F P
0 g
s
.57555 .468 m .59465 .47963 L .58355 .48598 L closepath p .809 .537 .526 r
F P
0 g
s
.59189 .46164 m .60333 .47136 L .59465 .47963 L closepath p .586 .596 .83 r
F P
0 g
s
.58355 .48598 m .59465 .47963 L .60333 .47136 L closepath p .681 .975 .806 r
F P
0 g
s
.57631 .43322 m .56398 .44197 L .55542 .45039 L closepath p .639 .944 .718 r
F P
0 g
s
.55404 .42563 m .56398 .44197 L .57631 .43322 L closepath p .82 .556 .54 r
F P
0 g
s
.54554 .43409 m .55404 .42563 L .51509 .43844 L closepath p .752 .967 .931 r
F P
0 g
s
.56398 .44197 m .55404 .42563 L .54554 .43409 L closepath p .498 .454 .739 r
F P
0 g
s
.55542 .45039 m .56398 .44197 L .54554 .43409 L closepath p .498 .454 .739 r
F P
0 g
s
.57555 .468 m .58355 .48598 L .55783 .46621 L closepath p .829 .543 .506 r
F P
0 g
s
.57746 .44049 m .56642 .45783 L .59189 .46164 L closepath p .82 .556 .54 r
F P
0 g
s
.60333 .47136 m .56642 .45783 L .59223 .4777 L closepath p .82 .556 .54 r
F P
0 g
s
.59189 .46164 m .56642 .45783 L .60333 .47136 L closepath p .82 .556 .54 r
F P
0 g
s
.60333 .47136 m .59223 .4777 L .58355 .48598 L closepath p .681 .975 .806 r
F P
0 g
s
.58355 .48598 m .59889 .50686 L .52968 .48192 L closepath p .819 .557 .542 r
F P
0 g
s
.52968 .48192 m .55783 .46621 L .58355 .48598 L closepath p .819 .557 .542 r
F P
0 g
s
.59889 .50686 m .58355 .48598 L .59223 .4777 L closepath p .524 .495 .766 r
F P
0 g
s
.51509 .43844 m .55404 .42563 L .52415 .42932 L closepath p .752 .967 .931 r
F P
0 g
s
.51952 .38819 m .52415 .42932 L .55404 .42563 L closepath p .87 .55 .443 r
F P
0 g
s
.59889 .50686 m .64345 .55076 L .55668 .55932 L closepath p .798 .491 .469 r
F P
0 g
s
.52968 .48192 m .59889 .50686 L .55668 .55932 L closepath p .82 .523 .487 r
F P
0 g
s
.59889 .50686 m .63255 .47523 L .64345 .55076 L closepath p .531 .508 .775 r
F P
0 g
s
.59223 .4777 m .60764 .49864 L .59889 .50686 L closepath p .524 .495 .766 r
F P
0 g
s
.53825 .4735 m .56642 .45783 L .55783 .46621 L closepath p .696 .981 .819 r
F P
0 g
s
.55783 .46621 m .52968 .48192 L .53825 .4735 L closepath p .696 .981 .819 r
F P
0 g
s
.53825 .4735 m .60764 .49864 L .59223 .4777 L closepath p .82 .556 .54 r
F P
0 g
s
.59223 .4777 m .56642 .45783 L .53825 .4735 L closepath p .82 .556 .54 r
F P
0 g
s
.7209 .35477 m .73738 .37673 L .6656 .34863 L closepath p .83 .551 .517 r
F P
0 g
s
.6656 .34863 m .69385 .33355 L .7209 .35477 L closepath p .83 .551 .517 r
F P
0 g
s
.69385 .33355 m .6656 .34863 L .67528 .33915 L closepath p .696 .981 .819 r
F P
0 g
s
.52968 .48192 m .59889 .50686 L .67293 .3443 L p .663 .742 .908 r
F P
0 g
s
.67052 .34381 m .52968 .48192 L p .67293 .3443 L .663 .742 .908 r
F P
0 g
s
.64666 .31637 m .5151 .44826 L .52968 .48192 L closepath p .519 .493 .768 r
F P
0 g
s
.52968 .48192 m .67528 .33915 L .64666 .31637 L closepath p .519 .493 .768 r
F P
0 g
s
.51952 .38819 m .44359 .40094 L .38161 .33522 L closepath p .828 .666 .682 r
F P
0 g
s
.51952 .38819 m .51509 .43844 L .44359 .40094 L closepath p .81 .584 .596 r
F P
0 g
s
.51509 .43844 m .52415 .42932 L .44359 .40094 L closepath p .636 .687 .881 r
F P
0 g
s
.38423 .4237 m .37442 .42322 L .38335 .42306 L p .1 0 0 r
F P
0 g
s
.38335 .42306 m .44324 .42202 L .38423 .4237 L p .1 0 0 r
F P
0 g
s
.44324 .42202 m .3726 .41594 L .37284 .40933 L closepath p .843 .577 .536 r
F P
0 g
s
.44324 .42202 m .37442 .42322 L .3726 .41594 L closepath p .844 .736 .745 r
F P
0 g
s
.44324 .42202 m .45143 .40371 L p .38423 .4237 L .751 .926 .947 r
F P
0 g
s
.40386 .40253 m .38423 .4237 L p .45143 .40371 L .751 .926 .947 r
F P
0 g
s
.38423 .4237 m .41471 .42387 L .4364 .46501 L closepath p .805 .42 .325 r
F P
0 g
s
.51952 .38819 m .45873 .38491 L .48634 .40366 L p .868 .548 .444 r
F P
0 g
s
.38161 .33522 m .45873 .38491 L .51952 .38819 L closepath p .831 .448 .327 r
F P
0 g
s
.44359 .40094 m .52415 .42932 L .45873 .38491 L closepath p .636 .687 .881 r
F P
0 g
s
.38161 .33522 m .44359 .40094 L .37083 .37488 L closepath p .822 .667 .691 r
F P
0 g
s
.44359 .40094 m .45873 .38491 L .40838 .37916 L p .682 .768 .917 r
F P
0 g
s
.40838 .37916 m .38084 .37602 L p .41463 .39057 L .682 .768 .917 r
F P
0 g
s
.3826 .36756 m .40862 .37052 L .40509 .37316 L p .705 .348 .368 r
F P
0 g
s
.37083 .37488 m .45873 .38491 L .43578 .37513 L p .682 .768 .917 r
F P
0 g
s
.37416 .37118 m .37083 .37488 L p .43578 .37513 L .682 .768 .917 r
F P
0 g
s
.40509 .37316 m .39678 .37938 L .38801 .37207 L p .705 .348 .368 r
F P
0 g
s
.38801 .37207 m .3826 .36756 L p .40509 .37316 L .705 .348 .368 r
F P
0 g
s
.3826 .36756 m .38801 .37207 L p .40212 .35797 L .541 .406 .645 r
F P
0 g
s
.40192 .35773 m .3826 .36756 L p .40212 .35797 L .541 .406 .645 r
F P
0 g
s
.38871 .35505 m .37416 .37118 L p .38801 .37207 L .682 .768 .917 r
F P
0 g
s
.40862 .37052 m .40328 .35838 L p .40067 .3651 L .945 .84 .708 r
F P
0 g
s
.40249 .35842 m .40067 .3651 L p .40328 .35838 L .945 .84 .708 r
F P
0 g
s
.40574 .36231 m .45873 .38491 L .38498 .33739 L p .816 .376 .197 r
F P
0 g
s
.40328 .35838 m .4028 .35729 L .40249 .35842 L p .945 .84 .708 r
F P
0 g
s
.38161 .33522 m .38871 .35505 L .39713 .35864 L p .816 .376 .197 r
F P
0 g
s
.40067 .3651 m .4028 .35729 L p .38801 .37207 L .541 .406 .645 r
F P
0 g
s
.39713 .35864 m .40574 .36231 L p .40227 .35842 L .816 .376 .197 r
F P
0 g
s
.38161 .33522 m .39713 .35864 L .40227 .35842 L .816 .376 .197 r
F
.40509 .37316 m .40862 .37052 L p .40067 .3651 L .945 .84 .708 r
F P
0 g
s
.43578 .37513 m .40574 .36231 L p .38801 .37207 L .682 .768 .917 r
F P
0 g
s
.40574 .36231 m .38871 .35505 L p .38801 .37207 L .682 .768 .917 r
F P
0 g
s
.38498 .33739 m .38161 .33522 L p .40574 .36231 L .816 .376 .197 r
F P
0 g
s
.4028 .35729 m .40192 .35773 L p .40212 .35797 L .541 .406 .645 r
F P
0 g
s
.38801 .37207 m .39678 .37938 L .40067 .3651 L p .541 .406 .645 r
F P
0 g
s
.38084 .37602 m .37083 .37488 L .37815 .3775 L p .682 .768 .917 r
F P
0 g
s
.37815 .3775 m .41463 .39057 L p .38084 .37602 L .682 .768 .917 r
F P
0 g
s
.41463 .39057 m .44359 .40094 L p .40838 .37916 L .682 .768 .917 r
F P
0 g
s
.40067 .3651 m .39678 .37938 L .40509 .37316 L p .945 .84 .708 r
F P
0 g
s
.45143 .40371 m .49676 .30235 L .40386 .40253 L p .751 .926 .947 r
F P
0 g
s
.49676 .30235 m .38423 .4237 L .41471 .42387 L closepath p .747 .916 .949 r
F P
0 g
s
.38423 .4237 m .41543 .4043 L .39205 .42935 L p 0 0 0 r
F P
0 g
s
.38423 .4237 m .4364 .46501 L .40299 .4467 L p .738 .214 0 r
F P
0 g
s
.40299 .4467 m .38036 .4343 L .38124 .43191 L p .738 .214 0 r
F P
0 g
s
.38124 .43191 m .38423 .4237 L p .40299 .4467 L .738 .214 0 r
F P
0 g
s
.4364 .46501 m .40556 .48834 L .38037 .43432 L p .883 .606 .512 r
F P
0 g
s
.38038 .43431 m .4364 .46501 L p .38037 .43432 L .883 .606 .512 r
F P
0 g
s
.37773 .4447 m .38036 .4343 L .40556 .48834 L closepath p .832 .441 0 r
F P
0 g
s
.39205 .42935 m .37773 .4447 L .38141 .43282 L p 0 0 0 r
F P
0 g
s
.38141 .43282 m .38423 .4237 L p .38652 .42535 L 0 0 0 r
F P
0 g
s
.39205 .42935 m .38141 .43282 L .38652 .42535 L 0 0 0 r
F
.38037 .43432 m .38036 .4343 L .38038 .43431 L p .883 .606 .512 r
F P
0 g
s
.4364 .46501 m .41471 .42387 L .45152 .45408 L p .801 .419 .333 r
F P
0 g
s
.4364 .46501 m .4419 .45933 L .41931 .47737 L p .62 .922 .664 r
F P
0 g
s
.41931 .47737 m .40556 .48834 L .4364 .46501 L p .62 .922 .664 r
F P
0 g
s
.45152 .45408 m .46528 .46537 L .4364 .46501 L p .801 .419 .333 r
F P
0 g
s
.43301 .48746 m .40556 .48834 L .43254 .4668 L p .821 .556 .537 r
F P
0 g
s
.44324 .42202 m .5151 .44826 L .46528 .46537 L closepath p .805 .42 .327 r
F P
0 g
s
.41471 .42387 m .44324 .42202 L .46528 .46537 L closepath p .8 .419 .333 r
F P
0 g
s
.46528 .46537 m .47358 .45692 L .4364 .46501 L closepath p .737 .908 .953 r
F P
0 g
s
.47358 .45692 m .4419 .45933 L .4364 .46501 L closepath p .737 .908 .953 r
F P
0 g
s
.43254 .4668 m .4419 .45933 L .4383 .47072 L p .821 .556 .537 r
F P
0 g
s
.4383 .47072 m .43301 .48746 L p .43254 .4668 L .821 .556 .537 r
F P
0 g
s
.49676 .30235 m .41543 .4043 L .38423 .4237 L closepath p 0 0 0 r
F P
0 g
s
.38423 .4237 m .45856 .35015 L p .3878 .4263 L 0 0 0 r
F P
0 g
s
.45856 .35015 m .45872 .35 L p .45848 .35025 L 0 0 0 r
F P
0 g
s
.45848 .35025 m .3878 .4263 L .45856 .35015 L 0 0 0 r
F
.45843 .35036 m .3878 .4263 L p .45872 .35 L F P
0 g
s
.41543 .4043 m .49676 .30235 L p .45843 .35036 L .878 .819 .787 r
F P
0 g
s
.38423 .4237 m .37773 .4447 L p .38119 .43341 L .276 0 0 r
F P
0 g
s
.38119 .43341 m .38036 .4343 L .38423 .4237 L p 0 0 0 r
F P
0 g
s
.38417 .4237 m .38423 .4237 L p .38119 .43341 L .276 0 0 r
F P
0 g
s
.3878 .4263 m .38119 .43341 L p .38423 .4237 L 0 0 0 r
F P
0 g
s
.41543 .4043 m .38112 .43366 L p .45843 .35036 L 0 0 0 r
F P
0 g
s
.38112 .43366 m .38036 .4343 L .38119 .43341 L p 0 0 0 r
F P
0 g
s
.38119 .43341 m .45843 .35036 L p .38112 .43366 L 0 0 0 r
F P
0 g
s
.41543 .4043 m .37773 .4447 L .38036 .4343 L closepath p 0 0 0 r
F P
0 g
s
.37773 .4447 m .37442 .42322 L .38417 .4237 L p .276 0 0 r
F P
0 g
s
.41594 .40358 m .41543 .4043 L p .45843 .35036 L .878 .819 .787 r
F P
0 g
s
.45843 .35036 m .45932 .3494 L .41543 .4043 L p 0 0 0 r
F P
0 g
s
.49676 .30235 m .45932 .3494 L .38423 .4237 L closepath p 0 0 0 r
F P
0 g
s
.50248 .29495 m .49364 .30449 L .49299 .2802 L p .344 .25 .603 r
F P
0 g
s
.49299 .2802 m .49295 .27876 L .49408 .28067 L p .344 .25 .603 r
F P
0 g
s
.49408 .28067 m .50248 .29495 L p .49299 .2802 L .344 .25 .603 r
F P
0 g
s
.49364 .30449 m .48398 .29426 L .49295 .27876 L closepath p .941 .725 .582 r
F P
0 g
s
.48398 .29426 m .50248 .29495 L .49295 .27876 L closepath p .686 .371 .436 r
F P
0 g
s
.49114 .28694 m .49864 .28859 L .51169 .28828 L p .45 0 0 r
F P
0 g
s
.55493 .28305 m .63955 .26481 L .63814 .28653 L p .783 .551 .584 r
F P
0 g
s
.63814 .28653 m .63619 .31649 L .59086 .29783 L p .783 .551 .584 r
F P
0 g
s
.58912 .29712 m .55493 .28305 L p .62961 .28617 L .783 .551 .584 r
F P
0 g
s
.59086 .29783 m .58912 .29712 L p .63814 .28653 L .783 .551 .584 r
F P
0 g
s
.63814 .28653 m .58912 .29712 L .62961 .28617 L .783 .551 .584 r
F
.63955 .26481 m .68009 .3002 L .63619 .31649 L closepath p .781 .549 .584 r
F P
0 g
s
.63619 .31649 m .64629 .30632 L .55493 .28305 L closepath p .636 .687 .881 r
F P
0 g
s
.63955 .26481 m .55493 .28305 L .47742 .22765 L closepath p .799 .602 .636 r
F P
0 g
s
.55493 .28305 m .64629 .30632 L .57171 .26528 L closepath p .636 .687 .881 r
F P
0 g
s
.57208 .28686 m .49676 .30235 L .48775 .32248 L p .751 .926 .947 r
F P
0 g
s
.48775 .32248 m .44324 .42202 L .55299 .30689 L p .751 .926 .947 r
F P
0 g
s
.55299 .30689 m .57208 .28686 L p .48775 .32248 L .751 .926 .947 r
F P
0 g
s
.57208 .28686 m .53552 .29551 L .52089 .31556 L p .754 .937 .945 r
F P
0 g
s
.52089 .31556 m .44324 .42202 L .55299 .30689 L p .754 .937 .945 r
F P
0 g
s
.55299 .30689 m .57208 .28686 L p .52089 .31556 L .754 .937 .945 r
F P
0 g
s
.51169 .28828 m .57208 .28686 L .50856 .28693 L p .45 0 0 r
F P
0 g
s
.50856 .28693 m .49114 .28694 L p .51169 .28828 L .45 0 0 r
F P
0 g
s
.49864 .28859 m .41199 .3975 L .41541 .40426 L p .886 .837 .792 r
F P
0 g
s
.41541 .40426 m .41543 .4043 L .41594 .40358 L p .886 .837 .792 r
F P
0 g
s
.41594 .40358 m .49864 .28859 L p .41541 .40426 L .886 .837 .792 r
F P
0 g
s
.49676 .30235 m .49864 .28859 L .41594 .40358 L p .878 .819 .787 r
F P
0 g
s
.45872 .35 m .45932 .3494 L .45843 .35036 L p 0 0 0 r
F P
0 g
s
.38335 .42306 m .38335 .42306 L p .38423 .4237 L .1 0 0 r
F P
0 g
s
.41471 .42387 m .44324 .42202 L .4564 .40397 L p .754 .937 .945 r
F P
0 g
s
.4564 .40397 m .53552 .29551 L .4332 .40422 L p .754 .937 .945 r
F P
0 g
s
.4332 .40422 m .41471 .42387 L p .4564 .40397 L .754 .937 .945 r
F P
0 g
s
.48634 .40366 m .48774 .40461 L p .48965 .40362 L .868 .548 .444 r
F P
0 g
s
.51952 .38819 m .48634 .40366 L .48965 .40362 L .868 .548 .444 r
F
.48774 .40461 m .52415 .42932 L .52146 .40544 L p F P
0 g
s
.52146 .40544 m .51952 .38819 L p .48774 .40461 L .868 .548 .444 r
F P
0 g
s
.71836 .27897 m .68009 .3002 L .63955 .26481 L closepath p .786 .546 .573 r
F P
0 g
s
.68009 .3002 m .68965 .29068 L .63619 .31649 L closepath p .752 .967 .931 r
F P
0 g
s
.63619 .31649 m .68965 .29068 L .64629 .30632 L closepath p .752 .967 .931 r
F P
0 g
s
.5151 .44826 m .64666 .31637 L .57208 .28686 L closepath p .664 .74 .906 r
F P
0 g
s
.57137 .35738 m .62992 .39345 L .64197 .33795 L p .77 .525 .565 r
F P
0 g
s
.64197 .33795 m .64666 .31637 L .61659 .33275 L p .77 .525 .565 r
F P
0 g
s
.61659 .33275 m .57137 .35738 L p .64197 .33795 L .77 .525 .565 r
F P
0 g
s
.67293 .3443 m .67528 .33915 L .67052 .34381 L p .663 .742 .908 r
F P
0 g
s
.69385 .33355 m .71909 .32203 L .73201 .34867 L closepath p .83 .551 .517 r
F P
0 g
s
.73115 .28935 m .72871 .32244 L .70309 .30873 L closepath p .816 .591 .597 r
F P
0 g
s
.73115 .28935 m .75413 .33782 L .72871 .32244 L closepath p .812 .587 .597 r
F P
0 g
s
.75413 .33782 m .73566 .34289 L .72871 .32244 L closepath p .83 .551 .517 r
F P
0 g
s
.72871 .32244 m .7384 .31308 L .70309 .30873 L closepath p .641 .691 .881 r
F P
0 g
s
.72871 .32244 m .74542 .33358 L .7384 .31308 L closepath p .419 .328 .646 r
F P
0 g
s
.73566 .34289 m .74542 .33358 L .72871 .32244 L closepath p .419 .328 .646 r
F P
0 g
s
.73201 .34867 m .71909 .32203 L .72877 .31265 L closepath p .464 .398 .698 r
F P
0 g
s
.73201 .34867 m .7209 .35477 L .69385 .33355 L closepath p .83 .551 .517 r
F P
0 g
s
.75413 .33782 m .76389 .32855 L .73566 .34289 L closepath p .742 .973 .926 r
F P
0 g
s
.76389 .32855 m .74542 .33358 L .74505 .33393 L p .742 .973 .926 r
F P
0 g
s
.74505 .33393 m .73566 .34289 L .75984 .33061 L p .742 .973 .926 r
F P
0 g
s
.75984 .33061 m .76389 .32855 L p .74505 .33393 L .742 .973 .926 r
F P
0 g
s
.72877 .31265 m .74178 .33936 L .73201 .34867 L closepath p .464 .398 .698 r
F P
0 g
s
.72877 .31265 m .72313 .32534 L .73697 .33574 L p .849 .555 .49 r
F P
0 g
s
.73697 .33574 m .74178 .33936 L .73972 .33513 L p .849 .555 .49 r
F P
0 g
s
.73972 .33513 m .72877 .31265 L p .73697 .33574 L .849 .555 .49 r
F P
0 g
s
.7209 .35477 m .73201 .34867 L .74178 .33936 L closepath p .681 .975 .806 r
F P
0 g
s
.74178 .33936 m .73067 .34544 L .7209 .35477 L closepath p .681 .975 .806 r
F P
0 g
s
.73738 .37673 m .7209 .35477 L .73067 .34544 L closepath p .524 .495 .766 r
F P
0 g
s
.73067 .34544 m .74723 .36748 L .73738 .37673 L closepath p .524 .495 .766 r
F P
0 g
s
.59889 .50686 m .74723 .36748 L .67528 .33915 L closepath p .663 .742 .908 r
F P
0 g
s
.64345 .55076 m .61872 .61079 L .55668 .55932 L closepath p .784 .455 .431 r
F P
0 g
s
.64345 .55076 m .63255 .47523 L .67165 .52531 L closepath p .531 .508 .775 r
F P
0 g
s
.67165 .52531 m .61872 .61079 L .64345 .55076 L closepath p 0 0 .04 r
F P
0 g
s
.52968 .48192 m .55668 .55932 L .50943 .51109 L closepath p .894 .555 .398 r
F P
0 g
s
.5151 .44826 m .52968 .48192 L .50943 .51109 L closepath p .952 .66 .446 r
F P
0 g
s
.50943 .51109 m .46528 .46537 L .50586 .45143 L p .883 .582 .472 r
F P
0 g
s
.51337 .4674 m .50943 .51109 L p .50586 .45143 L .883 .582 .472 r
F P
0 g
s
.55668 .55932 m .50562 .48175 L .50943 .51109 L closepath p .604 .616 .836 r
F P
0 g
s
.51751 .50803 m .50562 .48175 L .55659 .55918 L p .917 .524 .137 r
F P
0 g
s
.55659 .55918 m .55668 .55932 L .55661 .55923 L p .917 .524 .137 r
F P
0 g
s
.55661 .55923 m .51751 .50803 L p .55659 .55918 L .917 .524 .137 r
F P
0 g
s
.50586 .45143 m .5151 .44826 L .51337 .4674 L p .883 .582 .472 r
F P
0 g
s
.57208 .28686 m .44324 .42202 L .5151 .44826 L closepath p .664 .74 .906 r
F P
0 g
s
.60345 .64453 m .61872 .61079 L .67165 .52531 L closepath p 0 0 0 r
F P
0 g
s
.61452 .63482 m .60345 .64453 L .67165 .52531 L closepath p 0 0 .022 r
F P
0 g
s
.55668 .55932 m .61872 .61079 L .57963 .6124 L closepath p .803 .406 .3 r
F P
0 g
s
.61872 .61079 m .60345 .64453 L .57963 .6124 L closepath p .861 .546 .453 r
F P
0 g
s
.61452 .63482 m .67165 .52531 L .63255 .47523 L closepath p .822 .555 .536 r
F P
0 g
s
.63255 .47523 m .59056 .60255 L .61452 .63482 L closepath p .822 .555 .536 r
F P
0 g
s
.67025 .53077 m .59056 .60255 L .63255 .47523 L closepath p .057 0 .209 r
F P
0 g
s
.63255 .47523 m .71016 .40231 L .67025 .53077 L closepath p .057 0 .209 r
F P
0 g
s
.52939 .59524 m .57963 .6124 L .58002 .63447 L p .821 .556 .538 r
F P
0 g
s
.58002 .63447 m .57146 .64631 L .52939 .59524 L .821 .556 .538 r
F
.57338 .65385 m .52939 .59524 L p .58041 .65718 L F P
0 g
s
.58002 .63447 m .58041 .65718 L p .57146 .64631 L .821 .556 .538 r
F P
0 g
s
.58041 .65718 m .58052 .66336 L .57338 .65385 L p .821 .556 .538 r
F P
0 g
s
.52939 .59524 m .58052 .66336 L .56167 .62687 L p .509 .493 .775 r
F P
0 g
s
.57963 .6124 m .52939 .59524 L p .50669 .573 L .812 .456 .383 r
F P
0 g
s
.49161 .55949 m .57963 .6124 L p .50401 .57156 L .812 .456 .383 r
F P
0 g
s
.55668 .55932 m .57963 .6124 L .49161 .55949 L p .809 .408 .292 r
F P
0 g
s
.56167 .62687 m .54016 .58523 L .52939 .59524 L p .509 .493 .775 r
F P
0 g
s
.66254 .86959 m .50392 .61691 L .5278 .5947 L p .51 .495 .776 r
F P
0 g
s
.52939 .59524 m .50401 .57156 L p .50669 .573 L .812 .456 .383 r
F P
0 g
s
.49014 .55861 m .50392 .61691 L .42005 .58566 L closepath p .792 .679 .74 r
F P
0 g
s
.49014 .55861 m .52939 .59524 L .46715 .57399 L closepath p .812 .456 .383 r
F P
0 g
s
.50401 .57156 m .49014 .55861 L .49161 .55949 L p .812 .456 .383 r
F P
0 g
s
.49161 .55949 m .49014 .55861 L .49064 .55861 L p .809 .408 .292 r
F P
0 g
s
.49064 .55861 m .55668 .55932 L p .49161 .55949 L .809 .408 .292 r
F P
0 g
s
.55668 .55932 m .49014 .55861 L .49608 .53175 L closepath p .893 .783 .733 r
F P
0 g
s
.51751 .50803 m .55668 .55932 L .49608 .53175 L closepath p .919 .731 .632 r
F P
0 g
s
.50943 .51109 m .50562 .48175 L .46528 .46537 L closepath p .597 .611 .836 r
F P
0 g
s
.46528 .46537 m .50562 .48175 L .47358 .45692 L closepath p .587 .615 .846 r
F P
0 g
s
.47358 .45692 m .50562 .48175 L .49161 .46248 L closepath p 0 .309 .776 r
F P
0 g
s
.50562 .48175 m .51751 .50803 L .49161 .46248 L closepath p .778 .392 0 r
F P
0 g
s
.51751 .50803 m .49608 .53175 L .49161 .46248 L closepath p .949 .806 .667 r
F P
0 g
s
.42005 .58566 m .43882 .53714 L .49014 .55861 L closepath p .795 .69 .749 r
F P
0 g
s
.49014 .55861 m .46715 .57399 L .43882 .53714 L closepath p .841 .506 .42 r
F P
0 g
s
.45644 .52091 m .49014 .55861 L .43882 .53714 L closepath p .802 .682 .732 r
F P
0 g
s
.49608 .53175 m .49014 .55861 L .45644 .52091 L closepath p .797 .682 .738 r
F P
0 g
s
.49161 .46248 m .49608 .53175 L .45644 .52091 L closepath p .791 .693 .757 r
F P
0 g
s
.43882 .53714 m .42005 .58566 L .40954 .55433 L closepath p .94 .881 .749 r
F P
0 g
s
.43882 .53714 m .46715 .57399 L .40954 .55433 L closepath p .82 .514 .472 r
F P
0 g
s
.53109 .46527 m .43285 .56228 L .52939 .59524 L closepath p .654 .753 .923 r
F P
0 g
s
.52939 .59524 m .63044 .50135 L .53109 .46527 L closepath p .654 .753 .923 r
F P
0 g
s
.5278 .5947 m .52909 .5935 L p .53134 .59814 L .51 .495 .776 r
F P
0 g
s
.53134 .59814 m .66254 .86959 L .5278 .5947 L .51 .495 .776 r
F
.52909 .5935 m .53848 .58477 L .55089 .61328 L p F P
0 g
s
.55089 .61328 m .66254 .86959 L p .52909 .5935 L .51 .495 .776 r
F P
0 g
s
.53848 .58477 m .52503 .60254 L p .53933 .5883 L .882 .65 .583 r
F P
0 g
s
.53848 .58477 m .67431 .86066 L .66254 .86959 L closepath p .51 .495 .776 r
F P
0 g
s
.56789 .65124 m .53848 .58477 L p .55782 .66516 L .882 .65 .583 r
F P
0 g
s
.57221 .65329 m .53848 .58477 L .56789 .65124 L p .795 .622 .667 r
F P
0 g
s
.61876 .5122 m .54016 .58523 L .59056 .60255 L closepath p .654 .753 .923 r
F P
0 g
s
.60345 .64453 m .58052 .66336 L .57963 .6124 L closepath p .821 .556 .538 r
F P
0 g
s
.54016 .58523 m .59159 .65364 L .59056 .60255 L closepath p .822 .555 .536 r
F P
0 g
s
.59159 .65364 m .54016 .58523 L .58052 .66336 L closepath p .509 .493 .775 r
F P
0 g
s
.57717 .54878 m .52503 .60254 L .53848 .58477 L closepath p 0 0 0 r
F P
0 g
s
.50392 .61691 m .65169 .87303 L .48806 .63683 L closepath p .816 .594 .602 r
F P
0 g
s
.50392 .61691 m .66254 .86959 L .65169 .87303 L closepath p .806 .592 .613 r
F P
0 g
s
.42005 .58566 m .50392 .61691 L .48806 .63683 L closepath p .825 .607 .609 r
F P
0 g
s
.47863 .65218 m .48806 .63683 L .51657 .67955 L p .914 .546 .319 r
F P
0 g
s
.51657 .67955 m .6467 .87449 L .47863 .65218 L p .914 .546 .319 r
F P
0 g
s
.48806 .63683 m .65169 .87303 L .6467 .87449 L closepath p .915 .533 .276 r
F P
0 g
s
.42005 .58566 m .48806 .63683 L .41725 .6103 L closepath p .828 .605 .6 r
F P
0 g
s
.41725 .6103 m .52503 .60254 L .47863 .65218 L closepath p .686 .565 .713 r
F P
0 g
s
.50677 .62072 m .47863 .65218 L .52503 .60254 L p .586 .872 .555 r
F P
0 g
s
.49331 .63647 m .453 .63368 L p .47863 .65218 L .65 .758 .928 r
F P
0 g
s
.47863 .65218 m .6467 .87449 L .52503 .60254 L closepath p .706 .597 .732 r
F P
0 g
s
.47863 .65218 m .49546 .63662 L .49331 .63647 L p .65 .758 .928 r
F P
0 g
s
.41725 .6103 m .47863 .65218 L .42593 .63181 L closepath p .795 .356 .205 r
F P
0 g
s
.453 .63368 m .42593 .63181 L .47863 .65218 L p .65 .758 .928 r
F P
0 g
s
.52503 .60254 m .54491 .57808 L .50677 .62072 L p .586 .872 .555 r
F P
0 g
s
.52503 .60254 m .55782 .66516 L p .53933 .5883 L .882 .65 .583 r
F P
0 g
s
.55782 .66516 m .66241 .86488 L .56789 .65124 L p .882 .65 .583 r
F P
0 g
s
.66241 .86488 m .67431 .86066 L .57221 .65329 L p .795 .622 .667 r
F P
0 g
s
.56789 .65124 m .66241 .86488 L p .57221 .65329 L .795 .622 .667 r
F P
0 g
s
.38161 .33522 m .26792 .26808 L .24239 .2114 L closepath p .82 .685 .716 r
F P
0 g
s
.38161 .33522 m .37083 .37488 L .26792 .26808 L closepath p .83 .676 .692 r
F P
0 g
s
.35277 .33975 m .36152 .32981 L .35933 .33466 L p 0 0 .136 r
F P
0 g
s
.27801 .25774 m .27717 .24558 L p .27579 .25063 L .979 .889 .534 r
F P
0 g
s
.26792 .26808 m .27635 .27415 L p .26977 .25785 L .603 .655 .874 r
F P
0 g
s
.24239 .2114 m .26792 .26808 L .2245 .22297 L closepath p .805 .68 .726 r
F P
0 g
s
.25993 .25142 m .27801 .25774 L .27604 .26131 L p .811 .485 .438 r
F P
0 g
s
.25993 .25142 m .27604 .26131 L .26466 .25503 L .811 .485 .438 r
F
.26792 .26808 m .26466 .25503 L p .26051 .25268 L .596 .65 .874 r
F P
0 g
s
.25658 .23802 m .25956 .22897 L p .25801 .23325 L .695 .755 .256 r
F P
0 g
s
.24431 .24355 m .25199 .25153 L p .25301 .24845 L .596 .65 .874 r
F P
0 g
s
.24116 .22881 m .25658 .23802 L .25501 .24236 L p .881 .552 .423 r
F P
0 g
s
.24219 .23169 m .24116 .22881 L p .25501 .24236 L .881 .552 .423 r
F P
0 g
s
.23192 .23068 m .24431 .24355 L p .25464 .23291 L .596 .65 .874 r
F P
0 g
s
.24116 .22881 m .24219 .23169 L p .25933 .22946 L .648 .44 .588 r
F P
0 g
s
.25801 .23325 m .25933 .22946 L p .24219 .23169 L .648 .44 .588 r
F P
0 g
s
.25678 .22662 m .24116 .22881 L p .25933 .22946 L .648 .44 .588 r
F P
0 g
s
.25956 .22897 m .25943 .22924 L .25801 .23325 L .695 .755 .256 r
F
.2483 .21719 m .25933 .22946 L .25979 .22846 L .705 .189 0 r
F
.28258 .24714 m .26519 .23168 L p .27724 .24556 L F P
0 g
s
.26519 .23168 m .24239 .2114 L p .27684 .24517 L .705 .189 0 r
F P
0 g
s
.27684 .24517 m .27724 .24556 L .26519 .23168 L .705 .189 0 r
F
.26051 .25268 m .25993 .25142 L p .26466 .25503 L .811 .485 .438 r
F P
0 g
s
.25993 .25142 m .26051 .25268 L p .2771 .24552 L .573 .387 .592 r
F P
0 g
s
.2771 .24552 m .25993 .25142 L .2771 .24552 L .573 .387 .592 r
F
.27713 .2455 m .25993 .25142 L p .27713 .2455 L F P
0 g
s
.25914 .23001 m .25801 .23325 L p .25915 .23003 L .695 .755 .256 r
F P
0 g
s
.25993 .22785 m .25914 .23001 L .25915 .23003 L .695 .755 .256 r
F
.25962 .22864 m .25914 .23001 L p .25993 .22785 L F P
0 g
s
.26466 .25503 m .26977 .25785 L .26732 .25177 L .603 .655 .874 r
F
.27712 .24555 m .26051 .25268 L .26397 .25225 L .573 .387 .592 r
F
.27724 .24556 m .277 .24535 L p .27714 .24559 L .862 .411 .08 r
F P
0 g
s
.27716 .24553 m .27712 .24555 L .27712 .24555 L .573 .387 .592 r
F
.27712 .24555 m .26051 .25268 L .27712 .24555 L F
.37875 .35998 m .38215 .34657 L p .37932 .35273 L .883 .929 .545 r
F P
0 g
s
.38212 .34658 m .37944 .35245 L p .38215 .34657 L .883 .929 .545 r
F P
0 g
s
.35933 .33466 m .35932 .33469 L p .35927 .33472 L 0 0 .136 r
F P
0 g
s
.35277 .33975 m .35933 .33466 L .35927 .33472 L 0 0 .136 r
F
.35234 .34261 m .35277 .33975 L p .35932 .33469 L F P
0 g
s
.35999 .35073 m .37875 .35998 L .3765 .36277 L p .848 .47 .335 r
F P
0 g
s
.36118 .35238 m .35999 .35073 L p .3765 .36277 L .848 .47 .335 r
F P
0 g
s
.35999 .35073 m .36118 .35238 L p .38207 .3465 L .64 .466 .632 r
F P
0 g
s
.38206 .34649 m .35999 .35073 L p .38207 .3465 L .64 .466 .632 r
F P
0 g
s
.37935 .35265 m .38212 .34658 L p .36118 .35238 L .64 .466 .632 r
F P
0 g
s
.38207 .3465 m .36118 .35238 L .38212 .34658 L .64 .466 .632 r
F
.38502 .34475 m .38161 .33522 L p .38089 .34737 L .862 .411 .08 r
F P
0 g
s
.38212 .34658 m .38217 .34647 L p .38207 .3465 L .64 .466 .632 r
F P
0 g
s
.38215 .34657 m .37944 .35245 L .37942 .35253 L .883 .929 .545 r
F
.38871 .35505 m .38502 .34475 L p .38212 .34658 L .862 .411 .08 r
F P
0 g
s
.38212 .34658 m .38089 .34737 L .38627 .35266 L .862 .411 .08 r
F
.38212 .34658 m .38627 .35266 L .38871 .35505 L F
.32382 .2946 m .31694 .28329 L .31601 .28969 L p .971 .786 .596 r
F P
0 g
s
.30093 .29337 m .32382 .2946 L .32068 .30092 L p .701 .407 .475 r
F P
0 g
s
.32374 .30827 m .31032 .29861 L p .32393 .32247 L .615 .656 .866 r
F P
0 g
s
.3024 .29684 m .32393 .32247 L .31032 .29861 L .615 .656 .866 r
F
.29608 .29543 m .29924 .2985 L .3024 .29684 L F
.29924 .2985 m .32393 .32247 L .3024 .29684 L F
.30093 .29337 m .32068 .30092 L .31032 .29861 L .701 .407 .475 r
F
.3024 .29684 m .30234 .2967 L p .31032 .29861 L F P
0 g
s
.30234 .2967 m .30093 .29337 L p .31032 .29861 L .701 .407 .475 r
F P
0 g
s
.31694 .28329 m .30093 .29337 L .30234 .2967 L p .456 .291 .568 r
F P
0 g
s
.30234 .2967 m .3024 .29684 L p .30276 .2965 L .456 .291 .568 r
F P
0 g
s
.31694 .28329 m .30234 .2967 L .30276 .2965 L .456 .291 .568 r
F
.31601 .28969 m .31694 .28329 L p .30276 .2965 L F P
0 g
s
.30276 .2965 m .3024 .29684 L .30556 .29518 L .456 .291 .568 r
F
.30276 .2965 m .30556 .29518 L .31601 .28969 L F
.32765 .31792 m .34397 .31138 L .34063 .31738 L p .455 .18 .406 r
F P
0 g
s
.32853 .32107 m .32765 .31792 L p .34063 .31738 L .455 .18 .406 r
F P
0 g
s
.31601 .28969 m .31592 .29027 L p .32068 .30092 L .971 .786 .596 r
F P
0 g
s
.31032 .29861 m .32374 .30827 L p .32061 .30107 L .603 .655 .874 r
F P
0 g
s
.32779 .32622 m .34632 .34944 L .3488 .34663 L .615 .656 .866 r
F
.38103 .34498 m .30128 .26859 L .38089 .34737 L .862 .411 .08 r
F
.2927 .27775 m .30122 .26751 L p .29857 .27387 L 0 0 .126 r
F P
0 g
s
.30122 .2676 m .29857 .27387 L p .30122 .26751 L 0 0 .126 r
F P
0 g
s
.28753 .28713 m .3226 .32483 L .32779 .32622 L .615 .656 .866 r
F
.28695 .28783 m .3226 .32483 L p .28753 .28713 L F P
0 g
s
.29857 .27387 m .29831 .27448 L p .29203 .28173 L 0 0 .126 r
F P
0 g
s
.31457 .29956 m .31175 .31891 L .3024 .29684 L p .701 .407 .475 r
F P
0 g
s
.32061 .30107 m .32053 .30089 L p .32068 .30092 L .701 .407 .475 r
F P
0 g
s
.32061 .30107 m .32068 .30092 L p .31592 .29027 L .971 .786 .596 r
F P
0 g
s
.32068 .30092 m .32382 .2946 L p .31601 .28969 L .971 .786 .596 r
F P
0 g
s
.31032 .29861 m .32984 .30327 L .31685 .28984 L .603 .655 .874 r
F
.31032 .29861 m .31685 .28984 L .30741 .2943 L F
.30556 .29518 m .31032 .29861 L p .30741 .2943 L F P
0 g
s
.26792 .26808 m .29608 .29543 L .30452 .29731 L .615 .656 .866 r
F
.31032 .29861 m .29906 .2905 L p .30452 .29731 L F P
0 g
s
.29906 .2905 m .26792 .26808 L p .30452 .29731 L .615 .656 .866 r
F P
0 g
s
.2814 .26666 m .2927 .27775 L .29203 .28173 L p .945 .609 .358 r
F P
0 g
s
.28229 .26958 m .2814 .26666 L p .29203 .28173 L .945 .609 .358 r
F P
0 g
s
.29203 .28173 m .29201 .28182 L p .28229 .26958 L .945 .609 .358 r
F P
0 g
s
.28241 .26999 m .28229 .26958 L p .29201 .28182 L .945 .609 .358 r
F P
0 g
s
.2814 .26666 m .28241 .26999 L p .30122 .2676 L .712 .496 .594 r
F P
0 g
s
.29203 .28173 m .2927 .27775 L p .29857 .27387 L 0 0 .126 r
F P
0 g
s
.29831 .27448 m .30122 .2676 L p .28241 .26999 L .712 .496 .594 r
F P
0 g
s
.29201 .28182 m .29203 .28173 L p .29831 .27448 L 0 0 .126 r
F P
0 g
s
.30108 .2674 m .2814 .26666 L p .30122 .2676 L .712 .496 .594 r
F P
0 g
s
.27579 .25055 m .27716 .24553 L p .26397 .25225 L .573 .387 .592 r
F P
0 g
s
.27604 .26131 m .27801 .25774 L p .27579 .25063 L .979 .889 .534 r
F P
0 g
s
.26792 .26808 m .28695 .28783 L p .28753 .28713 L .615 .656 .866 r
F P
0 g
s
.25389 .2535 m .26792 .26808 L p .26051 .25268 L .596 .65 .874 r
F P
0 g
s
.26466 .25503 m .26792 .26808 L p .26977 .25785 L .603 .655 .874 r
F P
0 g
s
.27713 .2455 m .26051 .25268 L .27716 .24553 L .573 .387 .592 r
F
.26732 .25177 m .27571 .2473 L .25822 .22922 L .603 .655 .874 r
F
.27635 .27415 m .27579 .25055 L .26732 .25177 L F
.27579 .25055 m .27571 .2473 L .26732 .25177 L F
.26477 .25508 m .26974 .27274 L .26054 .25275 L p .811 .485 .438 r
F P
0 g
s
.26974 .27274 m .27327 .25978 L p .27604 .26131 L .811 .485 .438 r
F P
0 g
s
.26397 .25225 m .26466 .25503 L p .26732 .25177 L .603 .655 .874 r
F P
0 g
s
.26974 .27274 m .27572 .2508 L p .26397 .25225 L .573 .387 .592 r
F P
0 g
s
.25501 .24236 m .25658 .23802 L p .25801 .23325 L .695 .755 .256 r
F P
0 g
s
.26051 .25268 m .25822 .22922 L .24431 .24355 L .596 .65 .874 r
F
.26466 .25503 m .26397 .25225 L p .26051 .25268 L F P
0 g
s
.26397 .25225 m .25822 .22922 L p .26051 .25268 L .596 .65 .874 r
F P
0 g
s
.26054 .25275 m .26051 .25268 L p .26466 .25503 L .811 .485 .438 r
F P
0 g
s
.26051 .25268 m .26054 .25275 L p .26397 .25225 L .573 .387 .592 r
F P
0 g
s
.26054 .25275 m .26974 .27274 L p .26397 .25225 L .573 .387 .592 r
F P
0 g
s
.27717 .24558 m .27716 .24558 L .27579 .25059 L .979 .889 .534 r
F
.27579 .25063 m .27579 .25059 L .26974 .27274 L F
.27717 .24558 m .27579 .25059 L .27579 .25063 L F
.34189 .32909 m .35277 .33975 L .35234 .34261 L p .943 .596 .33 r
F P
0 g
s
.3284 .31386 m .33984 .33792 L .35104 .34099 L .615 .656 .866 r
F
.3278 .30529 m .32765 .31792 L .32846 .32082 L p 0 0 .315 r
F P
0 g
s
.32846 .32082 m .32853 .32107 L p .32844 .31922 L 0 0 .315 r
F P
0 g
s
.3278 .30529 m .32846 .32082 L .32844 .31922 L 0 0 .315 r
F
.34397 .31138 m .3278 .30529 L .32834 .30893 L p .814 .603 .618 r
F P
0 g
s
.33463 .31333 m .33765 .31829 L .33532 .31381 L .603 .655 .874 r
F
.34063 .31738 m .34057 .31748 L p .33765 .31829 L .455 .18 .406 r
F P
0 g
s
.34057 .31748 m .34397 .31138 L p .32834 .30893 L .814 .603 .618 r
F P
0 g
s
.32834 .30887 m .33765 .31829 L .32827 .3029 L .603 .655 .874 r
F
.32846 .31696 m .33984 .33792 L .3284 .31386 L .615 .656 .866 r
F
.32832 .30878 m .3278 .30529 L p .32844 .31922 L 0 0 .315 r
F P
0 g
s
.32844 .31922 m .32853 .32107 L .32836 .3116 L 0 0 .315 r
F
.32844 .31922 m .32836 .3116 L .32832 .30878 L F
.32374 .30827 m .32836 .3116 L p .32061 .30107 L .603 .655 .874 r
F P
0 g
s
.32836 .3116 m .33765 .31829 L p .32834 .30887 L .603 .655 .874 r
F P
0 g
s
.34225 .3301 m .34189 .32909 L p .35234 .34261 L .943 .596 .33 r
F P
0 g
s
.36152 .32981 m .34189 .32909 L .34225 .3301 L p .713 .502 .603 r
F P
0 g
s
.38036 .34904 m .36108 .33516 L p .35104 .34099 L .615 .656 .866 r
F P
0 g
s
.35932 .33469 m .36152 .32981 L p .34225 .3301 L .713 .502 .603 r
F P
0 g
s
.36108 .33516 m .32836 .3116 L p .34274 .33023 L .615 .656 .866 r
F P
0 g
s
.34274 .33023 m .35104 .34099 L .36108 .33516 L .615 .656 .866 r
F
.32836 .3116 m .32587 .3098 L p .32839 .31293 L F P
0 g
s
.32839 .31293 m .35104 .34099 L .32836 .3116 L .615 .656 .866 r
F
.32587 .3098 m .32374 .30827 L p .35104 .34099 L F P
0 g
s
.3226 .32483 m .34632 .34944 L p .32779 .32622 L .615 .656 .866 r
F P
0 g
s
.32393 .32247 m .33352 .33178 L .329 .32092 L .615 .656 .866 r
F
.33352 .33178 m .33984 .33792 L .32374 .30827 L F
.33187 .33311 m .32853 .32107 L p .33765 .31829 L .455 .18 .406 r
F P
0 g
s
.32393 .32247 m .329 .32092 L .32374 .30827 L .615 .656 .866 r
F
.32853 .32107 m .33187 .33311 L .32832 .30878 L p 0 0 .315 r
F P
0 g
s
.32836 .3116 m .32834 .30887 L .32061 .30107 L .603 .655 .874 r
F
.33765 .31829 m .35975 .3342 L p .34295 .31683 L F P
0 g
s
.33765 .31829 m .34295 .31683 L .32984 .30327 L .603 .655 .874 r
F
.34057 .31748 m .33187 .33311 L p .33765 .31829 L .455 .18 .406 r
F P
0 g
s
.32834 .30893 m .33187 .33311 L .34057 .31748 L p .814 .603 .618 r
F P
0 g
s
.32827 .3029 m .33463 .31333 L .32984 .30327 L .603 .655 .874 r
F
.33463 .31333 m .33532 .31381 L .32984 .30327 L F
.32061 .30107 m .32834 .30887 L .32827 .3029 L F
.32374 .30827 m .32846 .31696 L .3284 .31386 L .615 .656 .866 r
F
.33984 .33792 m .35463 .35228 L .36118 .35238 L F
.33984 .33792 m .3765 .36277 L .38036 .34904 L F
.37083 .37488 m .37413 .37122 L p .3488 .34663 L F P
0 g
s
.34632 .34944 m .37083 .37488 L p .3488 .34663 L .615 .656 .866 r
F P
0 g
s
.35234 .34261 m .35065 .35395 L .34225 .3301 L p .943 .596 .33 r
F P
0 g
s
.27716 .24553 m .26974 .27274 L p .27717 .24554 L .979 .889 .534 r
F P
0 g
s
.36318 .33043 m .27568 .24606 L p .35975 .3342 L .603 .655 .874 r
F P
0 g
s
.27568 .24606 m .25822 .22922 L p .27571 .2473 L .603 .655 .874 r
F P
0 g
s
.35975 .3342 m .27568 .24606 L .27571 .2473 L .603 .655 .874 r
F
.27606 .26217 m .30556 .29518 L .27571 .2473 L F
.27635 .27415 m .29906 .2905 L p .27606 .26217 L F P
0 g
s
.29906 .2905 m .30556 .29518 L p .27606 .26217 L .603 .655 .874 r
F P
0 g
s
.29201 .28182 m .28986 .29444 L .28241 .26999 L p .945 .609 .358 r
F P
0 g
s
.3024 .29684 m .31175 .31891 L .31601 .28969 L p .456 .291 .568 r
F P
0 g
s
.30741 .2943 m .30098 .27523 L .29592 .27726 L .603 .655 .874 r
F
.30098 .27523 m .30004 .27246 L .29831 .27448 L F
.30098 .27523 m .29831 .27448 L .29592 .27726 L F
.30556 .29518 m .30741 .2943 L .27571 .2473 L F
.29831 .27448 m .28986 .29444 L .29201 .28182 L p 0 0 .126 r
F P
0 g
s
.29592 .27726 m .30004 .27246 L .27571 .2473 L .603 .655 .874 r
F
.28241 .26999 m .28986 .29444 L .29831 .27448 L p .712 .496 .594 r
F P
0 g
s
.30741 .2943 m .31685 .28984 L .30004 .27246 L .603 .655 .874 r
F
.29866 .26396 m .30323 .27046 L .38103 .34498 L .862 .411 .08 r
F
.30122 .26751 m .3013 .26741 L p .30122 .2676 L 0 0 .126 r
F P
0 g
s
.30122 .2676 m .3013 .26741 L .30108 .2674 L p .712 .496 .594 r
F P
0 g
s
.36408 .32016 m .29866 .26396 L p .38103 .34498 L .862 .411 .08 r
F P
0 g
s
.29866 .26396 m .27724 .24556 L p .30323 .27046 L .862 .411 .08 r
F P
0 g
s
.30128 .26859 m .27724 .24556 L .30131 .26921 L .862 .411 .08 r
F
.30128 .26859 m .30131 .26921 L .38089 .34737 L F
.35932 .33469 m .35065 .35395 L .35234 .34261 L p 0 0 .136 r
F P
0 g
s
.34225 .3301 m .35065 .35395 L .35932 .33469 L p .713 .502 .603 r
F P
0 g
s
.31592 .29027 m .31175 .31891 L .32061 .30107 L p .971 .786 .596 r
F P
0 g
s
.38161 .33522 m .36408 .32016 L p .38103 .34498 L .862 .411 .08 r
F P
0 g
s
.38217 .34647 m .38206 .34649 L p .38207 .3465 L .64 .466 .632 r
F P
0 g
s
.36118 .35238 m .37413 .37122 L .3765 .36277 L .615 .656 .866 r
F
.3765 .36277 m .37237 .3679 L .36118 .35238 L p .848 .47 .335 r
F P
0 g
s
.36118 .35238 m .37237 .3679 L .37935 .35265 L p .64 .466 .632 r
F P
0 g
s
.35463 .35228 m .37413 .37122 L .36118 .35238 L .615 .656 .866 r
F
.37944 .35245 m .37237 .3679 L p .37942 .35253 L .883 .929 .545 r
F P
0 g
s
.35975 .3342 m .38036 .34904 L p .36318 .33043 L .603 .655 .874 r
F P
0 g
s
.38036 .34904 m .38871 .35505 L .3808 .34742 L p .603 .655 .874 r
F P
0 g
s
.3808 .34742 m .36318 .33043 L p .38036 .34904 L .603 .655 .874 r
F P
0 g
s
.27724 .24556 m .27714 .24559 L .38871 .35505 L .862 .411 .08 r
F
.27568 .24606 m .27975 .24998 L p .27714 .24559 L F P
0 g
s
.27975 .24998 m .38871 .35505 L p .27714 .24559 L .862 .411 .08 r
F P
0 g
s
.25822 .22922 m .26397 .25225 L p .26732 .25177 L .603 .655 .874 r
F P
0 g
s
.27572 .2508 m .27579 .25055 L p .26397 .25225 L .573 .387 .592 r
F P
0 g
s
.27716 .24553 m .27717 .24549 L p .27713 .2455 L .573 .387 .592 r
F P
0 g
s
.2591 .22997 m .277 .24535 L p .25979 .22846 L .705 .189 0 r
F P
0 g
s
.277 .24535 m .27724 .24556 L p .27684 .24517 L .705 .189 0 r
F P
0 g
s
.25979 .22846 m .277 .24535 L .27684 .24517 L .705 .189 0 r
F
.27714 .24559 m .2591 .22997 L .27568 .24606 L .862 .411 .08 r
F
.277 .24535 m .2591 .22997 L p .27714 .24559 L F P
0 g
s
.27713 .2455 m .2771 .24552 L .2771 .24552 L .573 .387 .592 r
F
.27717 .24549 m .27713 .2455 L p .27713 .2455 L F P
0 g
s
.25501 .24236 m .25049 .25486 L .24219 .23169 L p .881 .552 .423 r
F P
0 g
s
.2245 .22297 m .23192 .23068 L p .24042 .23021 L .596 .65 .874 r
F P
0 g
s
.25822 .22922 m .23379 .22469 L p .24042 .23021 L .596 .65 .874 r
F P
0 g
s
.23379 .22469 m .2245 .22297 L p .24042 .23021 L .596 .65 .874 r
F P
0 g
s
.24219 .23169 m .25049 .25486 L .25801 .23325 L p .648 .44 .588 r
F P
0 g
s
.25464 .23291 m .25822 .22922 L .23192 .23068 L .596 .65 .874 r
F
.24239 .2114 m .2591 .22997 L .25933 .22946 L .705 .189 0 r
F
.24239 .2114 m .25933 .22946 L .2483 .21719 L F
.25933 .22946 m .2605 .2261 L .25678 .22662 L p .648 .44 .588 r
F P
0 g
s
.25956 .22897 m .25993 .22785 L p .25943 .22924 L .695 .755 .256 r
F P
0 g
s
.24239 .2114 m .25822 .22922 L .2591 .22997 L p .705 .189 0 r
F P
0 g
s
.2591 .22997 m .25822 .22922 L .25907 .23004 L p .862 .411 .08 r
F P
0 g
s
.25907 .23004 m .27568 .24606 L p .2591 .22997 L .862 .411 .08 r
F P
0 g
s
.25993 .22785 m .2605 .2261 L .25962 .22864 L p .695 .755 .256 r
F P
0 g
s
.25801 .23325 m .25049 .25486 L .25501 .24236 L p .695 .755 .256 r
F P
0 g
s
.25389 .2535 m .26051 .25268 L .25301 .24845 L .596 .65 .874 r
F
.25199 .25153 m .25389 .2535 L p .25301 .24845 L F P
0 g
s
.26974 .27274 m .27604 .26131 L p .27579 .25063 L .979 .889 .534 r
F P
0 g
s
.3765 .36277 m .37875 .35998 L p .37932 .35273 L .883 .929 .545 r
F P
0 g
s
.37413 .37122 m .38871 .35505 L .38036 .34904 L p .615 .656 .866 r
F P
0 g
s
.37237 .3679 m .3765 .36277 L p .37932 .35273 L .883 .929 .545 r
F P
0 g
s
.38215 .34657 m .38217 .34647 L .38212 .34658 L p .883 .929 .545 r
F P
0 g
s
.27717 .24558 m .27717 .24554 L p .27716 .24558 L .979 .889 .534 r
F P
0 g
s
.27717 .24554 m .27717 .24549 L .27716 .24553 L p .979 .889 .534 r
F P
0 g
s
.27724 .24556 m .38161 .33522 L .28258 .24714 L p .705 .189 0 r
F P
0 g
s
.53552 .29551 m .49676 .30235 L .41471 .42387 L closepath p .747 .916 .949 r
F P
0 g
s
.7384 .31308 m .76389 .32855 L .73115 .28935 L closepath p .836 .453 .327 r
F P
0 g
s
.71836 .27897 m .73115 .28935 L .70309 .30873 L closepath p .847 .655 .643 r
F P
0 g
s
.71271 .29928 m .7384 .31308 L .73115 .28935 L closepath p .83 .45 .333 r
F P
0 g
s
.71836 .27897 m .71271 .29928 L .73115 .28935 L closepath p .894 .57 .427 r
F P
0 g
s
.40184 .40354 m .39379 .41222 L .39334 .38745 L closepath p .344 .25 .603 r
F P
0 g
s
.39379 .41222 m .38417 .40291 L .39334 .38745 L closepath p .927 .717 .599 r
F P
0 g
s
.38417 .40291 m .40184 .40354 L .39334 .38745 L closepath p .69 .39 .461 r
F P
0 g
s
.37284 .40933 m .3726 .41594 L .20995 .28175 L closepath p .893 .624 .522 r
F P
0 g
s
.35942 .39882 m .35308 .39386 L p .35893 .39218 L .642 .705 .892 r
F P
0 g
s
.35792 .37865 m .36456 .36082 L .35893 .39218 L p .966 .849 .438 r
F P
0 g
s
.35893 .39218 m .36456 .36082 L .3757 .38737 L p .572 .396 .605 r
F P
0 g
s
.3757 .38737 m .36456 .36082 L .35792 .37865 L p .831 .512 .45 r
F P
0 g
s
.34547 .3879 m .33413 .37901 L p .34267 .37923 L .642 .705 .892 r
F P
0 g
s
.33768 .36379 m .34452 .35405 L .34148 .37556 L p .992 .838 .589 r
F P
0 g
s
.34148 .37556 m .34452 .35405 L .35436 .36952 L p .523 .381 .629 r
F P
0 g
s
.35436 .36952 m .34452 .35405 L .33768 .36379 L p .781 .447 .422 r
F P
0 g
s
.33413 .37901 m .31019 .36026 L p .35436 .36952 L .642 .705 .892 r
F P
0 g
s
.31019 .36026 m .29679 .34976 L p .31384 .35561 L .642 .705 .892 r
F P
0 g
s
.31384 .35561 m .35436 .36952 L .31019 .36026 L .642 .705 .892 r
F
.31437 .35493 m .3231 .33437 L .33114 .35677 L p .706 .509 .62 r
F P
0 g
s
.32195 .34525 m .3231 .33437 L .31437 .35493 L p 0 0 .12 r
F P
0 g
s
.29679 .34976 m .29001 .34445 L p .32903 .35412 L .642 .705 .892 r
F P
0 g
s
.29001 .34445 m .28169 .33794 L p .29099 .34112 L .642 .705 .892 r
F P
0 g
s
.29099 .34112 m .32903 .35412 L .29001 .34445 L .642 .705 .892 r
F
.29573 .32496 m .29987 .31255 L .29216 .33713 L p .706 .747 .24 r
F P
0 g
s
.29216 .33713 m .29987 .31255 L .30972 .33576 L p .645 .457 .615 r
F P
0 g
s
.28172 .32598 m .30006 .33651 L .28192 .33792 L .642 .705 .892 r
F
.24733 .30446 m .26435 .31481 L .24573 .30468 L F
.24604 .30393 m .24573 .30468 L .23939 .30123 L F
.24733 .30446 m .24573 .30468 L .24604 .30393 L F
.24733 .30446 m .23939 .30123 L .24075 .3012 L F
.26435 .31481 m .26307 .31527 L .23232 .29762 L F
.23217 .2973 m .23232 .29762 L .22444 .29309 L F
.26435 .31481 m .23232 .29762 L .23217 .2973 L F
.24075 .3012 m .23939 .30123 L .22444 .29309 L F
.28169 .33794 m .22444 .29309 L p .28172 .32598 L F P
0 g
s
.22444 .29309 m .22264 .29168 L p .24733 .30446 L .642 .705 .892 r
F P
0 g
s
.22264 .29168 m .20995 .28175 L p .22443 .29055 L .642 .705 .892 r
F P
0 g
s
.22443 .29055 m .24733 .30446 L .22264 .29168 L .642 .705 .892 r
F
.28165 .32151 m .27812 .2903 L .28144 .30859 L p 0 0 .216 r
F P
0 g
s
.26655 .31401 m .27812 .2903 L .28165 .32151 L p .816 .597 .605 r
F P
0 g
s
.28144 .30859 m .27812 .2903 L .26655 .31401 L p .455 .185 .414 r
F P
0 g
s
.33114 .35677 m .3231 .33437 L .32195 .34525 L p .949 .594 .269 r
F P
0 g
s
.20995 .28175 m .30233 .33005 L .33388 .35706 L .642 .705 .892 r
F
.30972 .33576 m .29987 .31255 L .29573 .32496 L p .9 .556 .384 r
F P
0 g
s
.30233 .33005 m .35928 .35984 L .33792 .35751 L .642 .705 .892 r
F
.30233 .33005 m .33792 .35751 L .33388 .35706 L F
.26435 .31481 m .30006 .33651 L .26307 .31527 L F
.30006 .33651 m .32903 .35412 L .28169 .33794 L F
.22785 .28838 m .23147 .25217 L .24286 .27888 L p .449 .301 .588 r
F P
0 g
s
.22007 .27231 m .23147 .25217 L .22785 .28838 L p .982 .792 .574 r
F P
0 g
s
.24722 .30105 m .25574 .26838 L .26329 .3007 L p .66 .447 .585 r
F P
0 g
s
.25169 .29015 m .25574 .26838 L .24722 .30105 L p 0 0 0 r
F P
0 g
s
.49117 .28018 m .47489 .29795 L p .4623 .28604 L .642 .705 .892 r
F P
0 g
s
.47489 .29795 m .46253 .31144 L p .46235 .292 L .642 .705 .892 r
F P
0 g
s
.4623 .28604 m .47489 .29795 L .46235 .292 L .642 .705 .892 r
F
.47742 .22765 m .47307 .24645 L p .45628 .24202 L .734 .558 .653 r
F P
0 g
s
.44919 .22313 m .47742 .22765 L p .45628 .24202 L .734 .558 .653 r
F P
0 g
s
.47307 .24645 m .46841 .26663 L .45699 .26229 L p .734 .558 .653 r
F P
0 g
s
.46198 .25722 m .462 .2573 L .46202 .25733 L p 0 0 .33 r
F P
0 g
s
.46202 .25733 m .462 .2573 L .46213 .25723 L p .232 0 .166 r
F P
0 g
s
.35341 .22298 m .34419 .21948 L p .35264 .22125 L .734 .558 .653 r
F P
0 g
s
.29056 .19913 m .28539 .19717 L p .29038 .19847 L .734 .558 .653 r
F P
0 g
s
.3107 .20369 m .28539 .19717 L .28547 .19705 L .734 .558 .653 r
F
.34714 .21329 m .31338 .20447 L .31337 .2044 L F
.34419 .21948 m .29056 .19913 L p .34714 .21329 L F P
0 g
s
.34714 .21329 m .29056 .19913 L .29038 .19847 L .734 .558 .653 r
F
.31287 .20153 m .31288 .20159 L .31289 .20164 L p 0 0 .214 r
F P
0 g
s
.31289 .20164 m .31288 .20159 L .313 .20155 L p .431 .125 .343 r
F P
0 g
s
.34805 .21088 m .34818 .21102 L .34816 .21113 L p .946 .6 .321 r
F P
0 g
s
.34816 .21113 m .34818 .21102 L .34827 .21091 L p 0 0 .126 r
F P
0 g
s
.28423 .19672 m .44919 .22313 L p .45628 .24202 L .734 .558 .653 r
F P
0 g
s
.28423 .19672 m .27676 .16778 L .28167 .16924 L p .731 .58 .686 r
F P
0 g
s
.28737 .18832 m .28788 .1893 L p .28792 .19016 L .937 .617 .169 r
F P
0 g
s
.28788 .1893 m .28869 .19087 L .2887 .19231 L p .937 .617 .169 r
F P
0 g
s
.2887 .19278 m .28423 .19672 L .28167 .16924 L .731 .58 .686 r
F
.29016 .19767 m .28555 .19694 L p .2887 .19278 L F P
0 g
s
.28555 .19694 m .28423 .19672 L p .2887 .19278 L .731 .58 .686 r
F P
0 g
s
.2887 .19278 m .2887 .19231 L p .28962 .19157 L .227 0 .276 r
F P
0 g
s
.2887 .19231 m .2887 .19278 L p .28788 .1893 L .937 .617 .169 r
F P
0 g
s
.31063 .20376 m .28539 .19717 L .3107 .20369 L .734 .558 .653 r
F
.28788 .1893 m .2887 .19278 L .28792 .19016 L .937 .617 .169 r
F
.2887 .19231 m .28869 .19087 L .29057 .18957 L p .227 0 .276 r
F P
0 g
s
.47742 .22765 m .29016 .19767 L p .2887 .19278 L .731 .58 .686 r
F P
0 g
s
.29057 .18957 m .29177 .18873 L p .2887 .19231 L .227 0 .276 r
F P
0 g
s
.29177 .18873 m .28962 .19157 L .2887 .19231 L .227 0 .276 r
F
.28167 .16924 m .30197 .1753 L p .2887 .19278 L .731 .58 .686 r
F P
0 g
s
.30197 .1753 m .47742 .22765 L p .2887 .19278 L .731 .58 .686 r
F P
0 g
s
.29177 .18873 m .30159 .18193 L p .28962 .19157 L .227 0 .276 r
F P
0 g
s
.32902 .19539 m .32591 .20042 L p .31289 .20164 L .431 .125 .343 r
F P
0 g
s
.31289 .20164 m .32591 .20042 L .32477 .20078 L .431 .125 .343 r
F
.35727 .20007 m .35433 .20713 L p .34816 .21113 L 0 0 .126 r
F P
0 g
s
.34816 .21113 m .35433 .20713 L .35017 .21229 L 0 0 .126 r
F
.45699 .26229 m .35341 .22298 L p .47307 .24645 L .734 .558 .653 r
F P
0 g
s
.47307 .24645 m .35341 .22298 L .35264 .22125 L .734 .558 .653 r
F
.47307 .24645 m .34419 .21948 L .34714 .21329 L F
.43622 .24489 m .44571 .23411 L .44339 .23936 L p 0 0 .136 r
F P
0 g
s
.44169 .25531 m .4335 .25058 L .43269 .2515 L .617 .658 .868 r
F
.46841 .26663 m .44053 .25381 L .44169 .25531 L F
.44053 .25381 m .4335 .25058 L .44169 .25531 L F
.4253 .23333 m .43622 .24489 L .43592 .24784 L p .942 .578 .27 r
F P
0 g
s
.36965 .2208 m .38826 .21883 L .38496 .22562 L p .61 .297 .414 r
F P
0 g
s
.37101 .22531 m .36965 .2208 L p .38496 .22562 L .61 .297 .414 r
F P
0 g
s
.37481 .20546 m .36965 .2208 L .37101 .22531 L p 0 0 .439 r
F P
0 g
s
.37325 .21774 m .38065 .21949 L p .37365 .21623 L .626 .651 .855 r
F P
0 g
s
.37481 .20546 m .37101 .22531 L .37325 .21774 L 0 0 .439 r
F
.37504 .21101 m .37481 .20546 L p .37325 .21774 L F P
0 g
s
.34827 .21091 m .35727 .20007 L p .34816 .21113 L 0 0 .126 r
F P
0 g
s
.33682 .19928 m .34805 .21088 L p .34745 .21534 L .946 .6 .321 r
F P
0 g
s
.3464 .21683 m .34754 .21546 L .31484 .20395 L .617 .658 .868 r
F
.34391 .21082 m .3375 .2093 L p .34754 .21546 L F P
0 g
s
.3375 .2093 m .31484 .20395 L p .34446 .21437 L .617 .658 .868 r
F P
0 g
s
.34754 .21546 m .3375 .2093 L .34446 .21437 L .617 .658 .868 r
F
.31354 .20434 m .31289 .20164 L p .32477 .20078 L .431 .125 .343 r
F P
0 g
s
.34391 .21082 m .3341 .1978 L .32591 .20042 L .626 .651 .855 r
F
.31484 .20395 m .3375 .2093 L p .32591 .20042 L F P
0 g
s
.3375 .2093 m .34391 .21082 L p .32591 .20042 L .626 .651 .855 r
F P
0 g
s
.32477 .20078 m .31484 .20395 L .31354 .20434 L .431 .125 .343 r
F
.27676 .16778 m .28423 .19672 L .24239 .2114 L closepath p .432 .391 .713 r
F P
0 g
s
.24687 .14442 m .27676 .16778 L .24239 .2114 L closepath p .589 .615 .845 r
F P
0 g
s
.28179 .17757 m .28737 .18832 L p .28769 .19259 L .937 .617 .169 r
F P
0 g
s
.28179 .17757 m .28769 .19259 L .28768 .19256 L .937 .617 .169 r
F
.27854 .17165 m .26772 .16786 L p .28848 .1951 L .626 .651 .854 r
F P
0 g
s
.2892 .19604 m .28871 .19494 L .28848 .1951 L .626 .651 .854 r
F
.28871 .19494 m .27854 .17165 L .28848 .1951 L F
.28871 .19494 m .2887 .19278 L p .29825 .18852 L .227 0 .276 r
F P
0 g
s
.28257 .18089 m .28257 .18086 L p .28769 .19259 L .937 .617 .169 r
F P
0 g
s
.28257 .18086 m .28179 .17757 L p .28768 .19256 L .937 .617 .169 r
F P
0 g
s
.28768 .19256 m .28769 .19259 L .28257 .18086 L .937 .617 .169 r
F
.30159 .18193 m .28179 .17757 L .28257 .18086 L p .772 .545 .59 r
F P
0 g
s
.30159 .18193 m .29827 .1885 L p .2887 .19278 L .227 0 .276 r
F P
0 g
s
.2892 .19604 m .30801 .18195 L .28871 .19494 L .626 .651 .854 r
F
.29827 .1885 m .29827 .18851 L p .29825 .18852 L .227 0 .276 r
F P
0 g
s
.2887 .19278 m .29827 .1885 L .29825 .18852 L .227 0 .276 r
F
.31337 .2044 m .31063 .20376 L .3107 .20369 L .734 .558 .653 r
F
.31338 .20447 m .31063 .20376 L .31337 .2044 L F
.30946 .20555 m .31354 .20434 L .2901 .19811 L .617 .658 .868 r
F
.31484 .20395 m .31345 .20362 L p .31354 .20434 L F P
0 g
s
.313 .20155 m .32902 .19539 L p .31289 .20164 L .431 .125 .343 r
F P
0 g
s
.31135 .1878 m .31287 .20153 L p .31353 .20429 L 0 0 .214 r
F P
0 g
s
.31135 .1878 m .31353 .20429 L .31349 .204 L 0 0 .214 r
F
.31135 .1878 m .31354 .20431 L .31351 .20413 L F
.31164 .18947 m .31135 .1878 L p .31351 .20413 L F P
0 g
s
.31484 .20395 m .31266 .19747 L .31344 .20356 L .626 .651 .855 r
F
.31351 .20413 m .31208 .19288 L .31164 .18947 L 0 0 .214 r
F
.28737 .18832 m .28792 .19016 L .28809 .19351 L .937 .617 .169 r
F
.28559 .19704 m .2901 .19811 L p .28599 .19677 L .626 .651 .855 r
F P
0 g
s
.28995 .19776 m .28599 .19677 L .2901 .19811 L .626 .651 .855 r
F
.31289 .20164 m .31354 .20434 L p .31287 .20153 L 0 0 .214 r
F P
0 g
s
.34595 .21737 m .31484 .20395 L .31354 .20434 L .617 .658 .868 r
F
.34816 .21113 m .34754 .21546 L p .34805 .21088 L .946 .6 .321 r
F P
0 g
s
.35044 .21918 m .3464 .21683 L .34595 .21737 L .617 .658 .868 r
F
.34964 .21815 m .3464 .21683 L .35044 .21918 L F
.39518 .22658 m .41171 .2331 L .41008 .23733 L p .821 .463 .378 r
F P
0 g
s
.39684 .22975 m .39518 .22658 L p .41008 .23733 L .821 .463 .378 r
F P
0 g
s
.41234 .21952 m .39518 .22658 L .39684 .22975 L p .546 .368 .59 r
F P
0 g
s
.44661 .25825 m .3845 .23308 L .38252 .23362 L .617 .658 .868 r
F
.44431 .25693 m .31354 .20434 L .44661 .25825 L F
.46841 .26663 m .34964 .21815 L .35044 .21918 L F
.40975 .24179 m .38919 .23161 L .39063 .23366 L F
.40975 .24179 m .37101 .22531 L .37099 .22539 L F
.40983 .24064 m .35017 .21229 L .40975 .24179 L F
.4335 .25058 m .35017 .21229 L .43269 .2515 L F
.46841 .26663 m .34754 .21546 L .34964 .21815 L F
.41234 .21952 m .39684 .22975 L .40956 .22632 L .546 .368 .59 r
F
.41171 .2331 m .41234 .21952 L .41093 .2259 L p .902 .834 .372 r
F P
0 g
s
.55493 .28305 m .48935 .2706 L p .50028 .25861 L .787 .611 .663 r
F P
0 g
s
.48935 .2706 m .47877 .26859 L p .48803 .2643 L .787 .611 .663 r
F P
0 g
s
.50028 .25861 m .48935 .2706 L .48803 .2643 L .787 .611 .663 r
F
.48526 .23325 m .55493 .28305 L p .47307 .24645 L F P
0 g
s
.50955 .26037 m .50559 .26343 L p .48803 .2643 L .479 .068 .181 r
F P
0 g
s
.50955 .26037 m .4924 .24605 L .49388 .25546 L p .893 .709 .646 r
F P
0 g
s
.50949 .26609 m .49063 .26634 L p .52172 .27442 L .682 .768 .917 r
F P
0 g
s
.55493 .28305 m .49063 .26634 L .49062 .26637 L .682 .768 .917 r
F
.47877 .26859 m .48935 .2706 L p .49048 .26634 L F P
0 g
s
.48935 .2706 m .55493 .28305 L p .49062 .26637 L .682 .768 .917 r
F P
0 g
s
.49062 .26637 m .49048 .26634 L .48935 .2706 L .682 .768 .917 r
F
.49064 .26631 m .48803 .2643 L p .50559 .26343 L .479 .068 .181 r
F P
0 g
s
.48803 .2643 m .48759 .26396 L .48939 .26366 L p .479 .068 .181 r
F P
0 g
s
.48939 .26366 m .50955 .26037 L p .48803 .2643 L .479 .068 .181 r
F P
0 g
s
.48779 .26319 m .48759 .26396 L .48803 .2643 L p .105 .149 .624 r
F P
0 g
s
.47877 .26859 m .46841 .26663 L p .50028 .25861 L .787 .611 .663 r
F P
0 g
s
.48803 .2643 m .49064 .26631 L p .48779 .26319 L .105 .149 .624 r
F P
0 g
s
.48779 .26319 m .49064 .26631 L .49078 .26582 L .105 .149 .624 r
F
.46841 .26663 m .47307 .24645 L p .50028 .25861 L .787 .611 .663 r
F P
0 g
s
.50559 .26343 m .50955 .26037 L p .49388 .25546 L .893 .709 .646 r
F P
0 g
s
.4924 .24605 m .48779 .26319 L p .49078 .26582 L .105 .149 .624 r
F P
0 g
s
.49388 .25546 m .4924 .24605 L p .49078 .26582 L .105 .149 .624 r
F P
0 g
s
.46841 .26663 m .57171 .26528 L .48805 .24485 L closepath p .682 .768 .917 r
F P
0 g
s
.49626 .27063 m .49121 .2662 L p .50559 .26343 L .479 .068 .181 r
F P
0 g
s
.49064 .26631 m .49626 .27063 L p .49078 .26582 L .105 .149 .624 r
F P
0 g
s
.49388 .25546 m .49626 .27063 L .50559 .26343 L p .893 .709 .646 r
F P
0 g
s
.55493 .28305 m .57171 .26528 L .50949 .26609 L p .682 .768 .917 r
F P
0 g
s
.55493 .28305 m .50949 .26609 L .52172 .27442 L .682 .768 .917 r
F
.49063 .26634 m .49048 .26634 L p .49062 .26637 L F P
0 g
s
.49626 .27063 m .49388 .25546 L p .49078 .26582 L .105 .149 .624 r
F P
0 g
s
.49048 .26634 m .46841 .26663 L .47877 .26859 L p .682 .768 .917 r
F P
0 g
s
.47307 .24645 m .47742 .22765 L .48526 .23325 L p .787 .611 .663 r
F P
0 g
s
.47954 .24787 m .47456 .25371 L p .46237 .25804 L .232 0 .166 r
F P
0 g
s
.46302 .25935 m .46237 .25804 L p .47456 .25371 L .232 0 .166 r
F P
0 g
s
.46237 .25804 m .46202 .25733 L p .46401 .25626 L .232 0 .166 r
F P
0 g
s
.46401 .25626 m .47954 .24787 L .46237 .25804 L .232 0 .166 r
F
.41171 .2331 m .41093 .2259 L .41092 .22599 L .902 .834 .372 r
F
.37504 .21101 m .37365 .21623 L .38065 .21949 L .626 .651 .855 r
F
.38496 .22562 m .38826 .21883 L p .37507 .21156 L .934 .705 .568 r
F P
0 g
s
.41008 .23733 m .41171 .2331 L p .41092 .22595 L .902 .834 .372 r
F P
0 g
s
.42585 .23482 m .4253 .23333 L p .43592 .24784 L .942 .578 .27 r
F P
0 g
s
.43592 .24784 m .43622 .24489 L p .44339 .23936 L 0 0 .136 r
F P
0 g
s
.46213 .25723 m .47954 .24787 L p .46202 .25733 L .232 0 .166 r
F P
0 g
s
.45694 .24146 m .46198 .25722 L p .463 .25928 L 0 0 .33 r
F P
0 g
s
.47954 .24787 m .45694 .24146 L .45786 .24392 L p .788 .588 .63 r
F P
0 g
s
.48805 .24485 m .45561 .23719 L p .45786 .24392 L .617 .658 .868 r
F P
0 g
s
.46291 .25902 m .47346 .25307 L .45786 .24392 L .617 .658 .868 r
F
.47456 .25371 m .47954 .24787 L p .45786 .24392 L .788 .588 .63 r
F P
0 g
s
.47346 .25307 m .48805 .24485 L .45786 .24392 L .617 .658 .868 r
F
.46841 .26663 m .48437 .24892 L p .46302 .25935 L F P
0 g
s
.46492 .26502 m .46302 .25935 L .45795 .26182 L .617 .658 .868 r
F
.46841 .26663 m .46302 .25935 L .46492 .26502 L F
.46202 .25733 m .46302 .25935 L p .46198 .25722 L 0 0 .33 r
F P
0 g
s
.46198 .25722 m .46302 .25935 L .463 .25928 L 0 0 .33 r
F
.45786 .24392 m .45694 .24146 L p .463 .25928 L F P
0 g
s
.48437 .24892 m .48805 .24485 L p .45795 .26182 L .617 .658 .868 r
F P
0 g
s
.47456 .25371 m .4655 .26435 L .46302 .25935 L p .232 0 .166 r
F P
0 g
s
.46302 .25935 m .4655 .26435 L p .463 .25928 L 0 0 .33 r
F P
0 g
s
.45786 .24392 m .4655 .26435 L .47456 .25371 L p .788 .588 .63 r
F P
0 g
s
.4655 .26435 m .45786 .24392 L p .463 .25928 L 0 0 .33 r
F P
0 g
s
.45795 .26182 m .46291 .25902 L .35017 .21229 L .617 .658 .868 r
F
.44571 .23411 m .4253 .23333 L .42585 .23482 L p .713 .498 .596 r
F P
0 g
s
.46291 .25902 m .45756 .24303 L .36912 .22015 L .617 .658 .868 r
F
.43269 .2515 m .40983 .24064 L .40975 .24179 L F
.29056 .19913 m .46841 .26663 L p .2901 .19811 L F P
0 g
s
.43592 .24784 m .43477 .25883 L .42585 .23482 L p .942 .578 .27 r
F P
0 g
s
.44339 .23936 m .43477 .25883 L .43592 .24784 L p 0 0 .136 r
F P
0 g
s
.42585 .23482 m .43477 .25883 L .44339 .23936 L p .713 .498 .596 r
F P
0 g
s
.44339 .23936 m .44571 .23411 L p .42585 .23482 L .713 .498 .596 r
F P
0 g
s
.45561 .23719 m .35017 .21229 L p .36912 .22015 L .617 .658 .868 r
F P
0 g
s
.45756 .24303 m .45561 .23719 L .36912 .22015 L .617 .658 .868 r
F
.48805 .24485 m .41847 .22358 L .41093 .2259 L .626 .651 .855 r
F
.40956 .22632 m .41087 .22663 L p .41093 .2259 L F P
0 g
s
.41087 .22663 m .48805 .24485 L p .41093 .2259 L .626 .651 .855 r
F P
0 g
s
.41093 .2259 m .41092 .22595 L p .41092 .22599 L .902 .834 .372 r
F P
0 g
s
.41092 .22595 m .41093 .2259 L p .40956 .22632 L .546 .368 .59 r
F P
0 g
s
.41093 .2259 m .41234 .21952 L p .40956 .22632 L .546 .368 .59 r
F P
0 g
s
.38826 .21883 m .37481 .20546 L .37504 .21101 L p .934 .705 .568 r
F P
0 g
s
.38826 .21883 m .37504 .21101 L .37561 .21186 L .934 .705 .568 r
F
.38065 .21949 m .40956 .22632 L p .37447 .21014 L .626 .651 .855 r
F P
0 g
s
.41847 .22358 m .37447 .21014 L .40956 .22632 L .626 .651 .855 r
F
.41008 .23733 m .40613 .24757 L .39684 .22975 L p .821 .463 .378 r
F P
0 g
s
.39684 .22975 m .40613 .24757 L .41092 .22595 L p .546 .368 .59 r
F P
0 g
s
.38252 .23362 m .31354 .20434 L .3139 .20725 L .617 .658 .868 r
F
.3845 .23308 m .31354 .20434 L .38252 .23362 L F
.46841 .26663 m .44431 .25693 L .44661 .25825 L F
.34964 .21815 m .34754 .21546 L .3464 .21683 L F
.34754 .21546 m .3458 .22752 L p .34745 .21534 L .946 .6 .321 r
F P
0 g
s
.34805 .21088 m .34754 .21546 L .34745 .21534 L .946 .6 .321 r
F
.3464 .21683 m .31484 .20395 L .34595 .21737 L .617 .658 .868 r
F
.37081 .22531 m .35017 .21229 L .34754 .21546 L F
.3458 .22752 m .34752 .2156 L p .35017 .21229 L 0 0 .126 r
F P
0 g
s
.37419 .2122 m .35386 .20701 L .37365 .21623 L .626 .651 .855 r
F
.33682 .19928 m .34745 .21534 L .34677 .21448 L .946 .6 .321 r
F
.33799 .20295 m .33682 .19928 L p .34743 .21548 L F P
0 g
s
.35727 .20007 m .33682 .19928 L .33799 .20295 L p .712 .492 .589 r
F P
0 g
s
.35433 .20713 m .35727 .20007 L p .33799 .20295 L .712 .492 .589 r
F P
0 g
s
.34754 .21546 m .34816 .21113 L p .35017 .21229 L 0 0 .126 r
F P
0 g
s
.34391 .21082 m .37325 .21774 L p .3341 .1978 L .626 .651 .855 r
F P
0 g
s
.37365 .21623 m .3341 .1978 L .37325 .21774 L .626 .651 .855 r
F
.35433 .20713 m .35417 .20749 L p .35017 .21229 L 0 0 .126 r
F P
0 g
s
.37672 .22544 m .35017 .21229 L .37081 .22531 L .617 .658 .868 r
F
.35417 .20749 m .3458 .22752 L p .35017 .21229 L 0 0 .126 r
F P
0 g
s
.31631 .21593 m .31354 .20434 L p .32573 .20071 L .431 .125 .343 r
F P
0 g
s
.31354 .20434 m .31631 .21593 L p .31353 .20429 L 0 0 .214 r
F P
0 g
s
.31287 .20153 m .31354 .20434 L .31353 .20429 L 0 0 .214 r
F
.3139 .20725 m .31354 .20434 L .30946 .20555 L .617 .658 .868 r
F
.31345 .2036 m .28599 .19677 L .31345 .20362 L .626 .651 .855 r
F
.28423 .19672 m .2892 .19604 L p .27471 .17704 L .626 .651 .854 r
F P
0 g
s
.28995 .19775 m .28956 .19686 L .28599 .19677 L .626 .651 .855 r
F
.31547 .19242 m .30834 .1934 L p .32241 .19773 L F P
0 g
s
.30834 .1934 m .28423 .19672 L p .31697 .19759 L .626 .651 .855 r
F P
0 g
s
.32241 .19773 m .30834 .1934 L .31697 .19759 L .626 .651 .855 r
F
.28737 .18832 m .28809 .19351 L .28769 .19259 L .937 .617 .169 r
F
.28962 .19157 m .30159 .18193 L .2887 .19278 L .227 0 .276 r
F
.29827 .1885 m .30159 .18193 L p .28257 .18086 L .772 .545 .59 r
F P
0 g
s
.30801 .18195 m .27854 .17165 L p .28257 .18086 L .626 .651 .854 r
F P
0 g
s
.2887 .19278 m .28872 .19504 L p .28792 .19016 L .937 .617 .169 r
F P
0 g
s
.28792 .19016 m .28872 .19504 L .2881 .19367 L .937 .617 .169 r
F
.28871 .19494 m .29827 .1885 L .28257 .18086 L .626 .651 .854 r
F
.29827 .1885 m .30801 .18195 L .28257 .18086 L F
.28257 .18086 m .28257 .18089 L p .29827 .1885 L .772 .545 .59 r
F P
0 g
s
.28539 .19717 m .28423 .19672 L p .28547 .19705 L .734 .558 .653 r
F P
0 g
s
.2901 .19811 m .28423 .19672 L .29056 .19913 L p .617 .658 .868 r
F P
0 g
s
.28769 .19259 m .28809 .19351 L .28879 .20731 L .937 .617 .169 r
F
.2901 .19811 m .31345 .20362 L p .28995 .19776 L .626 .651 .855 r
F P
0 g
s
.31345 .2036 m .28956 .19686 L .28995 .19775 L .626 .651 .855 r
F
.31344 .20356 m .28956 .19686 L .31345 .2036 L F
.31345 .20362 m .2901 .19811 L p .31354 .20434 L .617 .658 .868 r
F P
0 g
s
.31344 .20356 m .31266 .19747 L .28956 .19686 L .626 .651 .855 r
F
.2892 .19604 m .31208 .19288 L p .30489 .18429 L .626 .651 .854 r
F P
0 g
s
.28871 .19494 m .28871 .19494 L p .29827 .18851 L .227 0 .276 r
F P
0 g
s
.29825 .18852 m .29827 .18851 L .28871 .19494 L .227 0 .276 r
F
.28768 .19256 m .28769 .19259 L .28769 .19259 L .937 .617 .169 r
F
.28879 .20731 m .28257 .18089 L p .28769 .19259 L F P
0 g
s
.29827 .18851 m .28879 .20731 L .28871 .19494 L p .227 0 .276 r
F P
0 g
s
.28872 .19504 m .28879 .20731 L p .2881 .19367 L .937 .617 .169 r
F P
0 g
s
.28257 .18089 m .28879 .20731 L .29827 .18851 L p .772 .545 .59 r
F P
0 g
s
.29827 .18851 m .29827 .1885 L p .28257 .18089 L .772 .545 .59 r
F P
0 g
s
.31345 .20362 m .31484 .20395 L p .31345 .2036 L .626 .651 .855 r
F P
0 g
s
.31484 .20395 m .31344 .20356 L .31345 .2036 L .626 .651 .855 r
F
.31484 .20395 m .32241 .19773 L .31266 .19747 L F
.31208 .19288 m .31547 .19242 L p .31164 .18947 L .626 .651 .854 r
F P
0 g
s
.31631 .21593 m .31165 .18948 L p .31354 .20432 L 0 0 .214 r
F P
0 g
s
.31165 .18948 m .31164 .18947 L p .31208 .19288 L 0 0 .214 r
F P
0 g
s
.31208 .19288 m .31354 .20432 L .31165 .18948 L 0 0 .214 r
F
.31351 .20413 m .31354 .20432 L .31208 .19288 L F
.31349 .204 m .31353 .20429 L .31354 .20431 L F
.31351 .20413 m .31354 .20431 L .31354 .20432 L F
.31208 .19288 m .31164 .18947 L .30489 .18429 L .626 .651 .854 r
F
.32591 .20042 m .32902 .19539 L p .31379 .19112 L .829 .604 .598 r
F P
0 g
s
.32902 .19539 m .31135 .1878 L .31165 .18948 L p .829 .604 .598 r
F P
0 g
s
.32902 .19539 m .31165 .18948 L .31379 .19112 L .829 .604 .598 r
F
.34417 .19459 m .33158 .19019 L p .31484 .20395 L .626 .651 .855 r
F P
0 g
s
.32591 .20042 m .32573 .20071 L p .32477 .20078 L .431 .125 .343 r
F P
0 g
s
.32573 .20071 m .32591 .20042 L p .31184 .19058 L .829 .604 .598 r
F P
0 g
s
.32591 .20042 m .32591 .20042 L p .32571 .20028 L .829 .604 .598 r
F P
0 g
s
.32571 .20028 m .31184 .19058 L .32591 .20042 L .829 .604 .598 r
F
.33158 .19019 m .30801 .18195 L p .30489 .18429 L .626 .651 .854 r
F P
0 g
s
.31547 .19242 m .33158 .19019 L p .30489 .18429 L .626 .651 .854 r
F P
0 g
s
.26772 .16786 m .27229 .17586 L p .27471 .17704 L .626 .651 .854 r
F P
0 g
s
.27229 .17586 m .28423 .19672 L p .27471 .17704 L .626 .651 .854 r
F P
0 g
s
.28423 .19672 m .28559 .19704 L p .28599 .19677 L .626 .651 .855 r
F P
0 g
s
.31379 .19112 m .31184 .19058 L .32591 .20042 L .829 .604 .598 r
F
.32477 .20078 m .32573 .20071 L .31484 .20395 L .431 .125 .343 r
F
.33158 .19019 m .31547 .19242 L p .32241 .19773 L .626 .651 .855 r
F P
0 g
s
.31165 .18948 m .31184 .19058 L p .31379 .19112 L .829 .604 .598 r
F P
0 g
s
.31184 .19058 m .31631 .21593 L .32573 .20071 L p .829 .604 .598 r
F P
0 g
s
.34677 .21448 m .34745 .21534 L .34743 .21548 L .946 .6 .321 r
F
.35017 .21229 m .34391 .21082 L p .34754 .21546 L .617 .658 .868 r
F P
0 g
s
.3458 .22752 m .33813 .20341 L p .34743 .21548 L .946 .6 .321 r
F P
0 g
s
.33813 .20341 m .33799 .20295 L p .34391 .21082 L .946 .6 .321 r
F P
0 g
s
.34391 .21082 m .34743 .21548 L .33813 .20341 L .946 .6 .321 r
F
.34752 .2156 m .34754 .21546 L p .35017 .21229 L 0 0 .126 r
F P
0 g
s
.33799 .20295 m .3458 .22752 L .35433 .20713 L p .712 .492 .589 r
F P
0 g
s
.37447 .21014 m .3341 .1978 L .35386 .20701 L .626 .651 .855 r
F
.37447 .21014 m .35386 .20701 L .37419 .2122 L F
.37447 .21014 m .37365 .21623 L .37504 .21101 L F
.37507 .21156 m .37504 .21101 L p .37325 .21774 L 0 0 .439 r
F P
0 g
s
.39063 .23366 m .37672 .22544 L .37101 .22531 L .617 .658 .868 r
F
.38919 .23161 m .37672 .22544 L .39063 .23366 L F
.38496 .22562 m .37641 .2432 L .37101 .22531 L p .61 .297 .414 r
F P
0 g
s
.37641 .2432 m .37507 .21156 L p .37325 .21774 L 0 0 .439 r
F P
0 g
s
.37101 .22531 m .37113 .22569 L p .37325 .21774 L 0 0 .439 r
F P
0 g
s
.37113 .22569 m .37641 .2432 L p .37325 .21774 L 0 0 .439 r
F P
0 g
s
.37504 .21101 m .37507 .21156 L p .37561 .21186 L .934 .705 .568 r
F P
0 g
s
.37507 .21156 m .37641 .2432 L .38496 .22562 L p .934 .705 .568 r
F P
0 g
s
.37099 .22539 m .37101 .22531 L .37081 .22531 L .617 .658 .868 r
F
.41092 .22595 m .40613 .24757 L .41008 .23733 L p .902 .834 .372 r
F P
0 g
s
.35635 .2046 m .34417 .19459 L .3341 .1978 L .626 .651 .855 r
F
.48805 .24485 m .35996 .20011 L p .35635 .2046 L F P
0 g
s
.35996 .20011 m .34417 .19459 L p .35635 .2046 L .626 .651 .855 r
F P
0 g
s
.44363 .28311 m .45552 .25897 L .46235 .292 L p .814 .583 .589 r
F P
0 g
s
.46235 .292 m .45552 .25897 L .46222 .2778 L p 0 0 .39 r
F P
0 g
s
.28644 .32174 m .4623 .28604 L .3991 .38066 L .642 .705 .892 r
F
.46253 .31144 m .3991 .38066 L p .4623 .28604 L F P
0 g
s
.26329 .3007 m .25574 .26838 L .25169 .29015 L p .924 .608 .426 r
F P
0 g
s
.24225 .28157 m .25201 .28151 L p .28644 .32174 L .642 .705 .892 r
F P
0 g
s
.45315 .28039 m .49117 .28018 L p .42813 .29298 L .642 .705 .892 r
F P
0 g
s
.42951 .28715 m .42813 .29298 L .36262 .30628 L .642 .705 .892 r
F
.43263 .28626 m .42813 .29298 L .42951 .28715 L F
.41586 .2752 m .42578 .26187 L .43011 .2846 L p .85 .635 .612 r
F P
0 g
s
.43011 .2846 m .42578 .26187 L .43294 .27266 L p .035 .079 .577 r
F P
0 g
s
.46222 .2778 m .45552 .25897 L .44363 .28311 L p .496 .219 .423 r
F P
0 g
s
.25201 .28151 m .45315 .28039 L p .34813 .30922 L .642 .705 .892 r
F P
0 g
s
.45315 .28039 m .36262 .30628 L .34813 .30922 L .642 .705 .892 r
F
.24286 .27888 m .23147 .25217 L .22007 .27231 L p .763 .448 .453 r
F P
0 g
s
.24239 .2114 m .28423 .19672 L .25822 .22922 L closepath p .432 .391 .713 r
F P
0 g
s
.25921 .2094 m .26784 .19853 L .26374 .20108 L p .788 .995 .871 r
F P
0 g
s
.25013 .1919 m .25921 .2094 L .25665 .21003 L p .945 .701 .537 r
F P
0 g
s
.24864 .19763 m .25013 .1919 L p .25665 .21003 L .945 .701 .537 r
F P
0 g
s
.25809 .20821 m .2245 .22297 L .24357 .22651 L p .606 .641 .859 r
F P
0 g
s
.25665 .21003 m .25921 .2094 L p .26374 .20108 L .788 .995 .871 r
F P
0 g
s
.24239 .2114 m .2245 .22297 L .21297 .18853 L closepath p .768 .597 .666 r
F P
0 g
s
.24239 .2114 m .21297 .18853 L .24687 .14442 L closepath p .589 .615 .845 r
F P
0 g
s
.26784 .19853 m .25013 .1919 L .24864 .19763 L p .691 .413 .498 r
F P
0 g
s
.2245 .22297 m .25809 .20821 L p .25999 .20022 L .606 .641 .859 r
F P
0 g
s
.24502 .1968 m .2245 .22297 L p .25999 .20022 L .606 .641 .859 r
F P
0 g
s
.26374 .20108 m .26784 .19853 L p .24864 .19763 L .691 .413 .498 r
F P
0 g
s
.27709 .18425 m .26772 .16786 L p .25809 .20821 L .606 .641 .859 r
F P
0 g
s
.25665 .21003 m .24463 .21296 L .24864 .19763 L p .945 .701 .537 r
F P
0 g
s
.24864 .19763 m .24463 .21296 L .26374 .20108 L p .691 .413 .498 r
F P
0 g
s
.26772 .16786 m .24502 .1968 L p .25999 .20022 L .606 .641 .859 r
F P
0 g
s
.25809 .20821 m .28423 .19672 L .27709 .18425 L p .606 .641 .859 r
F P
0 g
s
.28423 .19672 m .25809 .20821 L p .25811 .2123 L .606 .641 .859 r
F P
0 g
s
.25822 .22922 m .26421 .22174 L p .25811 .2123 L .606 .641 .859 r
F P
0 g
s
.26421 .22174 m .28423 .19672 L p .25811 .2123 L .606 .641 .859 r
F P
0 g
s
.24357 .22651 m .25822 .22922 L p .25809 .20821 L .606 .641 .859 r
F P
0 g
s
.26374 .20108 m .24463 .21296 L .25665 .21003 L p .788 .995 .871 r
F P
0 g
s
.43294 .27266 m .42578 .26187 L .41586 .2752 L p .552 .239 .391 r
F P
0 g
s
.38538 .27221 m .39453 .25179 L .40403 .27365 L p .695 .487 .601 r
F P
0 g
s
.38442 .27322 m .39476 .26241 L .40293 .27231 L .642 .705 .892 r
F
.45107 .26996 m .49117 .28018 L .37808 .28081 L p F P
0 g
s
.40403 .27365 m .39453 .25179 L .39476 .26241 L p .92 .527 .158 r
F P
0 g
s
.37808 .28081 m .20995 .28175 L p .45107 .26996 L .642 .705 .892 r
F P
0 g
s
.40564 .27736 m .38294 .27766 L .38538 .27221 L p .695 .487 .601 r
F P
0 g
s
.40403 .27365 m .40564 .27736 L p .38538 .27221 L .695 .487 .601 r
F P
0 g
s
.39476 .26241 m .38953 .26295 L .38538 .27221 L p 0 0 .19 r
F P
0 g
s
.39476 .26241 m .39453 .25179 L p .39153 .25849 L 0 0 .19 r
F P
0 g
s
.33901 .26824 m .35349 .26384 L .36132 .2659 L .642 .705 .892 r
F
.34873 .26258 m .35978 .2418 L .36821 .26772 L p .755 .536 .6 r
F P
0 g
s
.3404 .26743 m .36506 .26032 L .33901 .26824 L .642 .705 .892 r
F
.36259 .2545 m .35978 .2418 L .34873 .26258 L p .296 .027 .325 r
F P
0 g
s
.36821 .26772 m .35978 .2418 L .36259 .2545 L p .804 .469 0 r
F P
0 g
s
.32215 .26735 m .38766 .2538 L .33677 .26847 L .642 .705 .892 r
F
.30829 .26628 m .32223 .26409 L .32215 .26735 L F
.32223 .26409 m .38766 .2538 L .32215 .26735 L F
.32264 .24666 m .32853 .22952 L .32233 .25976 L p .912 .828 .366 r
F P
0 g
s
.26967 .2458 m .26609 .20871 L .27443 .23069 L p 0 0 .49 r
F P
0 g
s
.32233 .25976 m .32853 .22952 L .33844 .25408 L p .536 .349 .574 r
F P
0 g
s
.2515 .2728 m .29832 .23874 L .31031 .26013 L .642 .705 .892 r
F
.30576 .25462 m .28507 .25252 L .28704 .24695 L p .734 .501 .575 r
F P
0 g
s
.28704 .24695 m .29772 .2167 L .30486 .25041 L p .734 .501 .575 r
F P
0 g
s
.30486 .25041 m .30576 .25462 L p .28704 .24695 L .734 .501 .575 r
F P
0 g
s
.29832 .23874 m .29772 .2167 L .28704 .24695 L p .25 .017 .352 r
F P
0 g
s
.28951 .23995 m .29567 .23401 L .29832 .23874 L .642 .705 .892 r
F
.28305 .24617 m .28951 .23995 L .29832 .23874 L F
.20995 .28175 m .27175 .24733 L .27529 .24995 L F
.20995 .28175 m .29832 .23874 L .2515 .2728 L F
.2518 .23259 m .26609 .20871 L .26967 .2458 L p .87 .644 .593 r
F P
0 g
s
.20995 .28175 m .27266 .23084 L .29567 .23401 L .642 .705 .892 r
F
.27443 .23069 m .26609 .20871 L .2518 .23259 L p .597 .295 .428 r
F P
0 g
s
.27266 .23084 m .27861 .22601 L .28645 .22968 L .642 .705 .892 r
F
.27266 .23084 m .28645 .22968 L .29567 .23401 L F
.2585 .22088 m .27611 .22537 L p .27429 .22692 L F P
0 g
s
.2245 .22297 m .26772 .16786 L .21297 .18853 L closepath p .447 .414 .729 r
F P
0 g
s
.2585 .22088 m .20995 .28175 L .17894 .24629 L p .54 .552 .816 r
F P
0 g
s
.17894 .24629 m .13883 .20044 L .1901 .2092 L p .54 .552 .816 r
F P
0 g
s
.18812 .20494 m .21769 .19908 L .19666 .22336 L p .862 .696 .674 r
F P
0 g
s
.1901 .2092 m .2024 .2113 L p .1957 .22129 L .54 .552 .816 r
F P
0 g
s
.1901 .2092 m .1957 .22129 L .17894 .24629 L .54 .552 .816 r
F
.2024 .2113 m .20507 .21249 L .17894 .24629 L F
.20507 .21249 m .21769 .19908 L .18812 .20494 L p .693 .275 .248 r
F P
0 g
s
.19666 .22336 m .21769 .19908 L .20507 .21249 L p .625 .876 .532 r
F P
0 g
s
.21297 .18853 m .26772 .16786 L .24687 .14442 L closepath p .447 .414 .729 r
F P
0 g
s
.13883 .20044 m .17795 .14771 L .2585 .22088 L closepath p .54 .552 .816 r
F P
0 g
s
.19208 .22866 m .18221 .20611 L .18812 .20494 L p .862 .696 .674 r
F P
0 g
s
.19666 .22336 m .19208 .22866 L p .18812 .20494 L .862 .696 .674 r
F P
0 g
s
.18221 .20611 m .2023 .21543 L .20507 .21249 L p .693 .275 .248 r
F P
0 g
s
.18812 .20494 m .18221 .20611 L p .20507 .21249 L .693 .275 .248 r
F P
0 g
s
.2024 .2113 m .20557 .21184 L p .20507 .21249 L .54 .552 .816 r
F P
0 g
s
.20557 .21184 m .2585 .22088 L p .17894 .24629 L .54 .552 .816 r
F P
0 g
s
.2023 .21543 m .19208 .22866 L .19666 .22336 L p .625 .876 .532 r
F P
0 g
s
.20507 .21249 m .2023 .21543 L p .19666 .22336 L .625 .876 .532 r
F P
0 g
s
.17323 .14956 m .17795 .14771 L .13883 .20044 L closepath p .801 .984 .894 r
F P
0 g
s
.17323 .14956 m .17575 .15912 L .24319 .21911 L p .486 .539 .837 r
F P
0 g
s
.17795 .14771 m .17323 .14956 L .24555 .21558 L p .813 .524 .499 r
F P
0 g
s
.22958 .26503 m .25712 .22261 L .26787 .22531 L .642 .705 .892 r
F
.23329 .25248 m .25712 .22261 L p .22958 .26503 L F P
0 g
s
.24555 .21558 m .25808 .22702 L .24729 .21634 L p .813 .524 .499 r
F P
0 g
s
.24729 .21634 m .17795 .14771 L p .24555 .21558 L .813 .524 .499 r
F P
0 g
s
.27141 .22685 m .26787 .22531 L .27197 .22634 L .642 .705 .892 r
F
.22958 .26503 m .26787 .22531 L .27141 .22685 L F
.27401 .22716 m .27197 .22634 L .27429 .22692 L F
.22958 .26503 m .27197 .22634 L .27401 .22716 L F
.2585 .22088 m .17795 .14771 L .24729 .21634 L p .803 .53 .526 r
F P
0 g
s
.25712 .22261 m .2585 .22088 L p .27429 .22692 L .642 .705 .892 r
F P
0 g
s
.24729 .21634 m .25808 .22702 L .2585 .22088 L p .803 .53 .526 r
F P
0 g
s
.24319 .21911 m .25559 .23014 L .24344 .21825 L p .486 .539 .837 r
F P
0 g
s
.24344 .21825 m .17323 .14956 L p .24319 .21911 L .486 .539 .837 r
F P
0 g
s
.20995 .28175 m .23329 .25248 L p .22958 .26503 L .642 .705 .892 r
F P
0 g
s
.17262 .23752 m .25559 .23014 L .24319 .21911 L p .749 .547 .622 r
F P
0 g
s
.24319 .21911 m .17575 .15912 L .17405 .20173 L p .749 .547 .622 r
F P
0 g
s
.17405 .20173 m .17262 .23752 L p .24319 .21911 L .749 .547 .622 r
F P
0 g
s
.20995 .28175 m .27611 .22537 L .27498 .22896 L .642 .705 .892 r
F
.27031 .2524 m .24792 .23907 L .2518 .23259 L p .87 .644 .593 r
F P
0 g
s
.26967 .2458 m .27031 .2524 L p .2518 .23259 L .87 .644 .593 r
F P
0 g
s
.27443 .23069 m .27612 .23513 L p .26967 .2458 L 0 0 .49 r
F P
0 g
s
.27622 .24484 m .29567 .23401 L .28305 .24617 L .642 .705 .892 r
F
.28507 .25252 m .29842 .24252 L .29832 .23874 L p .25 .017 .352 r
F P
0 g
s
.28704 .24695 m .28507 .25252 L p .29832 .23874 L .25 .017 .352 r
F P
0 g
s
.30486 .25041 m .29772 .2167 L .29832 .23874 L p .956 .632 .226 r
F P
0 g
s
.29832 .23874 m .29842 .24252 L p .30486 .25041 L .956 .632 .226 r
F P
0 g
s
.31031 .26013 m .29567 .23401 L .33844 .25408 L .642 .705 .892 r
F
.33844 .25408 m .32853 .22952 L .32264 .24666 L p .824 .484 .413 r
F P
0 g
s
.27861 .22601 m .38766 .2538 L p .35036 .25967 L .642 .705 .892 r
F P
0 g
s
.30029 .26754 m .33844 .25408 L .35036 .25967 L .642 .705 .892 r
F
.20995 .28175 m .33844 .25408 L .30029 .26754 L F
.32111 .26574 m .32124 .25075 L .32264 .24666 L p .912 .828 .366 r
F P
0 g
s
.32233 .25976 m .32111 .26574 L p .32264 .24666 L .912 .828 .366 r
F P
0 g
s
.32264 .24666 m .32124 .25075 L p .33844 .25408 L .824 .484 .413 r
F P
0 g
s
.20995 .28175 m .39476 .26241 L .38442 .27322 L .642 .705 .892 r
F
.3698 .27261 m .34545 .26873 L .34873 .26258 L p .755 .536 .6 r
F P
0 g
s
.36821 .26772 m .3698 .27261 L p .34873 .26258 L .755 .536 .6 r
F P
0 g
s
.35349 .26384 m .36506 .26032 L .36132 .2659 L .642 .705 .892 r
F
.34545 .26873 m .36333 .25786 L .36259 .2545 L p .296 .027 .325 r
F P
0 g
s
.34873 .26258 m .34545 .26873 L p .36259 .2545 L .296 .027 .325 r
F P
0 g
s
.36132 .2659 m .36506 .26032 L .36718 .26529 L .642 .705 .892 r
F
.36333 .25786 m .3698 .27261 L .36821 .26772 L p .804 .469 0 r
F P
0 g
s
.36259 .2545 m .36333 .25786 L p .36821 .26772 L .804 .469 0 r
F P
0 g
s
.36718 .26529 m .38766 .2538 L .39476 .26241 L .642 .705 .892 r
F
.36506 .26032 m .38766 .2538 L .36718 .26529 L F
.38294 .27766 m .39482 .2652 L .39476 .26241 L p 0 0 .19 r
F P
0 g
s
.38538 .27221 m .38294 .27766 L p .39476 .26241 L 0 0 .19 r
F P
0 g
s
.39482 .2652 m .40564 .27736 L .40403 .27365 L p .92 .527 .158 r
F P
0 g
s
.39476 .26241 m .39482 .2652 L p .40403 .27365 L .92 .527 .158 r
F P
0 g
s
.38766 .2538 m .45107 .26996 L p .40293 .27231 L .642 .705 .892 r
F P
0 g
s
.43086 .28851 m .41284 .27925 L .41586 .2752 L p .85 .635 .612 r
F P
0 g
s
.43011 .2846 m .43086 .28851 L p .41586 .2752 L .85 .635 .612 r
F P
0 g
s
.43294 .27266 m .43448 .27498 L p .43011 .2846 L .035 .079 .577 r
F P
0 g
s
.41284 .27925 m .43448 .27498 L .43294 .27266 L p .552 .239 .391 r
F P
0 g
s
.41586 .2752 m .41284 .27925 L p .43294 .27266 L .552 .239 .391 r
F P
0 g
s
.45315 .28039 m .43646 .28879 L .43263 .28626 L .642 .705 .892 r
F
.43448 .27498 m .43086 .28851 L .43011 .2846 L p .035 .079 .577 r
F P
0 g
s
.46311 .29566 m .44127 .28793 L .44363 .28311 L p .814 .583 .589 r
F P
0 g
s
.46235 .292 m .46311 .29566 L p .44363 .28311 L .814 .583 .589 r
F P
0 g
s
.46222 .2778 m .46324 .28067 L p .46235 .292 L 0 0 .39 r
F P
0 g
s
.44127 .28793 m .46324 .28067 L .46222 .2778 L p .496 .219 .423 r
F P
0 g
s
.44363 .28311 m .44127 .28793 L p .46222 .2778 L .496 .219 .423 r
F P
0 g
s
.46324 .28067 m .46311 .29566 L .46235 .292 L p 0 0 .39 r
F P
0 g
s
.48769 .2907 m .49114 .28694 L .278 .23554 L p .641 .704 .892 r
F P
0 g
s
.27612 .23513 m .27443 .23069 L p .25246 .23763 L .597 .295 .428 r
F P
0 g
s
.26635 .2365 m .27612 .23513 L p .25246 .23763 L .597 .295 .428 r
F P
0 g
s
.2518 .23259 m .24859 .23795 L p .27443 .23069 L .597 .295 .428 r
F P
0 g
s
.24859 .23795 m .24792 .23907 L p .25246 .23763 L .597 .295 .428 r
F P
0 g
s
.25246 .23763 m .27443 .23069 L .24859 .23795 L .597 .295 .428 r
F
.25808 .22702 m .25559 .23014 L .278 .23554 L p .641 .704 .892 r
F P
0 g
s
.278 .23554 m .49114 .28694 L .28803 .23472 L p .641 .704 .892 r
F P
0 g
s
.28803 .23472 m .25808 .22702 L p .278 .23554 L .641 .704 .892 r
F P
0 g
s
.27175 .24733 m .27622 .24484 L .27529 .24995 L .642 .705 .892 r
F
.27612 .23513 m .27031 .2524 L .26967 .2458 L p 0 0 .49 r
F P
0 g
s
.278 .23554 m .25559 .23014 L .27671 .23565 L p .641 .704 .892 r
F P
0 g
s
.27671 .23565 m .48769 .2907 L p .278 .23554 L .641 .704 .892 r
F P
0 g
s
.24792 .23907 m .26635 .2365 L p .25246 .23763 L .597 .295 .428 r
F P
0 g
s
.27529 .24995 m .27622 .24484 L .28305 .24617 L .642 .705 .892 r
F
.29842 .24252 m .30576 .25462 L .30486 .25041 L p .956 .632 .226 r
F P
0 g
s
.34002 .25799 m .32111 .26574 L .32233 .25976 L p .536 .349 .574 r
F P
0 g
s
.33844 .25408 m .34002 .25799 L p .32233 .25976 L .536 .349 .574 r
F P
0 g
s
.27611 .22537 m .27861 .22601 L p .27498 .22896 L .642 .705 .892 r
F P
0 g
s
.32124 .25075 m .34002 .25799 L .33844 .25408 L p .824 .484 .413 r
F P
0 g
s
.33677 .26847 m .3404 .26743 L .33901 .26824 L .642 .705 .892 r
F
.20995 .28175 m .30829 .26628 L .33677 .26847 L F
.43646 .28879 m .42813 .29298 L .43263 .28626 L F
.25201 .28151 m .34813 .30922 L .28644 .32174 L F
.20995 .28175 m .24225 .28157 L p .28644 .32174 L F P
0 g
s
.24286 .27888 m .2449 .28365 L p .22785 .28838 L .449 .301 .588 r
F P
0 g
s
.22713 .29565 m .21698 .27778 L .22007 .27231 L p .982 .792 .574 r
F P
0 g
s
.22785 .28838 m .22713 .29565 L p .22007 .27231 L .982 .792 .574 r
F P
0 g
s
.21698 .27778 m .2449 .28365 L .24286 .27888 L p .763 .448 .453 r
F P
0 g
s
.22007 .27231 m .21698 .27778 L p .24286 .27888 L .763 .448 .453 r
F P
0 g
s
.26329 .3007 m .26423 .30473 L p .24722 .30105 L .66 .447 .585 r
F P
0 g
s
.25094 .29418 m .26423 .30473 L .26329 .3007 L p .924 .608 .426 r
F P
0 g
s
.25169 .29015 m .25094 .29418 L p .26329 .3007 L .924 .608 .426 r
F P
0 g
s
.29014 .34357 m .2945 .32866 L .29573 .32496 L p .706 .747 .24 r
F P
0 g
s
.29216 .33713 m .29014 .34357 L p .29573 .32496 L .706 .747 .24 r
F P
0 g
s
.30972 .33576 m .31159 .34014 L p .29216 .33713 L .645 .457 .615 r
F P
0 g
s
.2945 .32866 m .31159 .34014 L .30972 .33576 L p .9 .556 .384 r
F P
0 g
s
.29573 .32496 m .2945 .32866 L p .30972 .33576 L .9 .556 .384 r
F P
0 g
s
.2449 .28365 m .22713 .29565 L .22785 .28838 L p .449 .301 .588 r
F P
0 g
s
.26423 .30473 m .24577 .30663 L .24722 .30105 L p .66 .447 .585 r
F P
0 g
s
.24577 .30663 m .25094 .29418 L .25169 .29015 L p 0 0 0 r
F P
0 g
s
.24722 .30105 m .24577 .30663 L p .25169 .29015 L 0 0 0 r
F P
0 g
s
.28144 .30859 m .2821 .31223 L p .28165 .32151 L 0 0 .216 r
F P
0 g
s
.28172 .32598 m .28192 .33792 L .28169 .33794 L .642 .705 .892 r
F
.28221 .32646 m .2638 .31963 L .26655 .31401 L p .816 .597 .605 r
F P
0 g
s
.28165 .32151 m .28221 .32646 L p .26655 .31401 L .816 .597 .605 r
F P
0 g
s
.2638 .31963 m .2821 .31223 L .28144 .30859 L p .455 .185 .414 r
F P
0 g
s
.26655 .31401 m .2638 .31963 L p .28144 .30859 L .455 .185 .414 r
F P
0 g
s
.2821 .31223 m .28221 .32646 L .28165 .32151 L p 0 0 .216 r
F P
0 g
s
.31159 .34014 m .29014 .34357 L .29216 .33713 L p .645 .457 .615 r
F P
0 g
s
.32903 .35412 m .35436 .36952 L .3364 .36335 L .642 .705 .892 r
F
.32903 .35412 m .3364 .36335 L .29679 .34976 L F
.33114 .35677 m .33251 .36056 L p .31437 .35493 L .706 .509 .62 r
F P
0 g
s
.32165 .34816 m .33251 .36056 L .33114 .35677 L p .949 .594 .269 r
F P
0 g
s
.32195 .34525 m .32165 .34816 L p .33114 .35677 L .949 .594 .269 r
F P
0 g
s
.33251 .36056 m .31207 .36033 L .31437 .35493 L p .706 .509 .62 r
F P
0 g
s
.31207 .36033 m .32165 .34816 L .32195 .34525 L p 0 0 .12 r
F P
0 g
s
.31437 .35493 m .31207 .36033 L p .32195 .34525 L 0 0 .12 r
F P
0 g
s
.33388 .35706 m .36671 .36372 L .35436 .36952 L .642 .705 .892 r
F
.35928 .35984 m .36671 .36372 L .33388 .35706 L F
.36671 .36372 m .3991 .38066 L .33413 .37901 L F
.34083 .38017 m .33559 .36676 L .33768 .36379 L p .992 .838 .589 r
F P
0 g
s
.34148 .37556 m .34083 .38017 L p .33768 .36379 L .992 .838 .589 r
F P
0 g
s
.35436 .36952 m .35606 .37218 L p .34148 .37556 L .523 .381 .629 r
F P
0 g
s
.33559 .36676 m .35606 .37218 L .35436 .36952 L p .781 .447 .422 r
F P
0 g
s
.33768 .36379 m .33559 .36676 L p .35436 .36952 L .781 .447 .422 r
F P
0 g
s
.35308 .39386 m .34547 .3879 L p .3991 .38066 L .642 .705 .892 r
F P
0 g
s
.34267 .37923 m .3991 .38066 L .34547 .3879 L .642 .705 .892 r
F
.35606 .37218 m .34083 .38017 L .34148 .37556 L p .523 .381 .629 r
F P
0 g
s
.35807 .39694 m .35659 .38221 L .35792 .37865 L p .966 .849 .438 r
F P
0 g
s
.35893 .39218 m .35807 .39694 L p .35792 .37865 L .966 .849 .438 r
F P
0 g
s
.37284 .40933 m .35942 .39882 L p .3757 .38737 L .642 .705 .892 r
F P
0 g
s
.35893 .39218 m .3757 .38737 L .35942 .39882 L .642 .705 .892 r
F
.3991 .38066 m .3876 .39322 L p .3757 .38737 L F P
0 g
s
.3876 .39322 m .37284 .40933 L p .3757 .38737 L .642 .705 .892 r
F P
0 g
s
.3757 .38737 m .37693 .39032 L p .35893 .39218 L .572 .396 .605 r
F P
0 g
s
.35659 .38221 m .37693 .39032 L .3757 .38737 L p .831 .512 .45 r
F P
0 g
s
.35792 .37865 m .35659 .38221 L p .3757 .38737 L .831 .512 .45 r
F P
0 g
s
.37693 .39032 m .35807 .39694 L .35893 .39218 L p .572 .396 .605 r
F P
0 g
s
.71836 .27897 m .70309 .30873 L .68009 .3002 L closepath p .82 .618 .63 r
F P
0 g
s
.68965 .29068 m .71271 .29928 L .71836 .27897 L closepath p .826 .487 .414 r
F P
0 g
s
.63955 .26481 m .68965 .29068 L .71836 .27897 L closepath p .737 .329 .278 r
F P
0 g
s
.37773 .4447 m .36685 .44609 L .37442 .42322 L closepath p .246 0 0 r
F P
0 g
s
.3726 .41594 m .37442 .42322 L .3543 .42203 L closepath p .938 .675 .202 r
F P
0 g
s
.3726 .41594 m .20947 .28783 L .20995 .28175 L closepath p .893 .624 .522 r
F P
0 g
s
.3726 .41594 m .3757 .41258 L .20947 .28783 L closepath p .641 .704 .892 r
F P
0 g
s
.3757 .41258 m .3726 .41594 L .3543 .42203 L closepath p .63 .92 .646 r
F P
0 g
s
.4419 .45933 m .47358 .45692 L .49161 .46248 L closepath p .217 0 0 r
F P
0 g
s
.3757 .41258 m .3543 .42203 L .24524 .3262 L closepath p .76 .247 0 r
F P
0 g
s
.3757 .41258 m .21184 .28487 L .20947 .28783 L closepath p .641 .704 .892 r
F P
0 g
s
.3757 .41258 m .24524 .3262 L .21184 .28487 L closepath p .854 .402 .109 r
F P
0 g
s
.7384 .31308 m .74542 .33358 L .76389 .32855 L closepath p .83 .55 .515 r
F P
0 g
s
.71016 .40231 m .74723 .36748 L .75079 .45388 L closepath p .531 .508 .775 r
F P
0 g
s
.71016 .40231 m .69522 .56403 L .67025 .53077 L closepath p .828 .552 .522 r
F P
0 g
s
.71016 .40231 m .75079 .45388 L .69522 .56403 L closepath p .828 .552 .522 r
F P
0 g
s
.69385 .33355 m .70452 .31621 L .71909 .32203 L closepath p .83 .551 .517 r
F P
0 g
s
.71909 .32203 m .70452 .31621 L .71505 .32312 L closepath p .474 0 0 r
F P
0 g
s
.71505 .32312 m .72877 .31265 L .71909 .32203 L closepath p .598 .914 .663 r
F P
0 g
s
.70309 .30873 m .6908 .31725 L .68009 .3002 L closepath p .83 .551 .517 r
F P
0 g
s
.70309 .30873 m .71271 .29928 L .6908 .31725 L closepath p .639 .944 .718 r
F P
0 g
s
.70309 .30873 m .7384 .31308 L .71271 .29928 L closepath p .641 .691 .881 r
F P
0 g
s
.65169 .87303 m .75956 .93486 L .6467 .87449 L closepath p .761 .266 .046 r
F P
0 g
s
.66254 .86959 m .75956 .93486 L .65169 .87303 L closepath p .795 .562 .585 r
F P
0 g
s
.67431 .86066 m .75956 .93486 L .66254 .86959 L closepath p .578 .627 .864 r
F P
0 g
s
.66241 .86488 m .6467 .87449 L .75956 .93486 L closepath p .748 .648 .75 r
F P
0 g
s
.67431 .86066 m .66241 .86488 L .75956 .93486 L closepath p .796 .627 .672 r
F P
0 g
s
.75079 .45388 m .74723 .36748 L .78167 .42601 L closepath p .531 .508 .775 r
F P
0 g
s
.69522 .56403 m .75079 .45388 L .71308 .54009 L p 0 0 .022 r
F P
0 g
s
.73562 .50722 m .7072 .55352 L .75079 .45388 L closepath p .004 0 .063 r
F P
0 g
s
.69343 .43161 m .68208 .52011 L .73417 .50756 L p .811 .627 .653 r
F P
0 g
s
.73417 .50756 m .73562 .50722 L .73505 .50621 L p .811 .627 .653 r
F P
0 g
s
.73505 .50621 m .69343 .43161 L p .73417 .50756 L .811 .627 .653 r
F P
0 g
s
.71308 .54009 m .7072 .55352 L .70063 .55929 L p 0 0 .022 r
F P
0 g
s
.70063 .55929 m .69522 .56403 L p .71308 .54009 L 0 0 .022 r
F P
0 g
s
.78167 .42601 m .73562 .50722 L .75079 .45388 L closepath p 0 0 .04 r
F P
0 g
s
.70352 .32412 m .71505 .32312 L .70452 .31621 L closepath p .509 0 0 r
F P
0 g
s
.70452 .31621 m .69385 .33355 L .70352 .32412 L closepath p 0 0 0 r
F P
0 g
s
.72313 .32534 m .72877 .31265 L .71505 .32312 L closepath p .788 .478 .464 r
F P
0 g
s
.63955 .26481 m .64629 .30632 L .68965 .29068 L closepath p .73 .427 .466 r
F P
0 g
s
.57171 .26528 m .64629 .30632 L .63955 .26481 L closepath p .835 .502 .423 r
F P
0 g
s
.47742 .22765 m .57171 .26528 L .63955 .26481 L closepath p .717 .253 .147 r
F P
0 g
s
.43301 .48746 m .4419 .45933 L .49161 .46248 L closepath p .543 .264 .442 r
F P
0 g
s
.6908 .31725 m .71271 .29928 L .70043 .30779 L closepath p .639 .944 .718 r
F P
0 g
s
.68965 .29068 m .70043 .30779 L .71271 .29928 L closepath p .83 .55 .515 r
F P
0 g
s
.68009 .3002 m .70043 .30779 L .68965 .29068 L closepath p .498 .454 .739 r
F P
0 g
s
.6908 .31725 m .70043 .30779 L .68009 .3002 L closepath p .498 .454 .739 r
F P
0 g
s
.59159 .65364 m .58052 .66336 L .60345 .64453 L closepath p .498 .846 .579 r
F P
0 g
s
.61452 .63482 m .59159 .65364 L .60345 .64453 L closepath p .498 .846 .579 r
F P
0 g
s
.72313 .32534 m .73067 .34544 L .74178 .33936 L closepath p .841 .572 .532 r
F P
0 g
s
.67025 .53077 m .61876 .5122 L .59056 .60255 L closepath p .654 .753 .923 r
F P
0 g
s
.59159 .65364 m .61452 .63482 L .59056 .60255 L closepath p .822 .555 .536 r
F P
0 g
s
.72313 .32534 m .71505 .32312 L .70352 .32412 L closepath p .775 .406 .358 r
F P
0 g
s
.72313 .32534 m .70352 .32412 L .73067 .34544 L closepath p .82 .563 .551 r
F P
0 g
s
.67528 .33915 m .70352 .32412 L .69385 .33355 L closepath p .696 .981 .819 r
F P
0 g
s
.67528 .33915 m .74723 .36748 L .73067 .34544 L closepath p .83 .55 .515 r
F P
0 g
s
.73067 .34544 m .70352 .32412 L .67528 .33915 L closepath p .83 .55 .515 r
F P
0 g
s
.6467 .87449 m .66241 .86488 L .52503 .60254 L closepath p .731 .609 .721 r
F P
0 g
s
.45644 .52091 m .43301 .48746 L .49161 .46248 L closepath p .671 .534 .688 r
F P
0 g
s
.43301 .48746 m .43882 .53714 L .40556 .48834 L closepath p .836 .625 .618 r
F P
0 g
s
.45644 .52091 m .43882 .53714 L .43301 .48746 L closepath p .725 .546 .648 r
F P
0 g
s
.69343 .43161 m .73562 .50722 L .78167 .42601 L closepath p .85 .635 .61 r
F P
0 g
s
.69343 .43161 m .78167 .42601 L .74723 .36748 L closepath p .846 .611 .584 r
F P
0 g
s
.69343 .43161 m .74723 .36748 L .67528 .33915 L closepath p .827 .582 .568 r
F P
0 g
s
.40954 .55433 m .42005 .58566 L .41725 .6103 L closepath p .994 .809 .439 r
F P
0 g
s
.41308 .49938 m .43882 .53714 L .41504 .522 L p .923 .593 .398 r
F P
0 g
s
.43882 .53714 m .40954 .55433 L .39411 .50867 L closepath p .909 .713 .625 r
F P
0 g
s
.49581 .46785 m .41993 .54391 L .44332 .55194 L closepath p .654 .753 .923 r
F P
0 g
s
.40954 .55433 m .41725 .6103 L .40899 .57688 L closepath p .928 .722 .236 r
F P
0 g
s
.43285 .56228 m .40899 .57688 L .40954 .55433 L closepath p .821 .556 .538 r
F P
0 g
s
.41993 .54391 m .41944 .56651 L .44332 .55194 L closepath p .822 .555 .536 r
F P
0 g
s
.4569 .52936 m .4572 .50656 L .41717 .54668 L p .211 .063 .452 r
F P
0 g
s
.41717 .54668 m .40954 .55433 L .41747 .55015 L p .211 .063 .452 r
F P
0 g
s
.41747 .55015 m .4569 .52936 L p .41717 .54668 L .211 .063 .452 r
F P
0 g
s
.39411 .50867 m .40556 .48834 L .41308 .49938 L p .923 .593 .398 r
F P
0 g
s
.41504 .522 m .39411 .50867 L p .41308 .49938 L .923 .593 .398 r
F P
0 g
s
.40556 .48834 m .39411 .50867 L .37773 .4447 L closepath p .99 .741 .449 r
F P
0 g
s
.41725 .6103 m .42593 .63181 L .40899 .57688 L closepath p 0 0 .33 r
F P
0 g
s
.62992 .39345 m .67528 .33915 L .64666 .31637 L closepath p .707 .463 .553 r
F P
0 g
s
.57137 .35738 m .64666 .31637 L .57208 .28686 L closepath p .815 .624 .645 r
F P
0 g
s
.54491 .57808 m .52503 .60254 L .56138 .56506 L p .6 .894 .601 r
F P
0 g
s
.50145 .58698 m .49546 .63662 L .51411 .61455 L p .703 .754 .891 r
F P
0 g
s
.51411 .61455 m .54491 .57808 L .50145 .58698 L p .703 .754 .891 r
F P
0 g
s
.50145 .58698 m .44012 .61822 L .49546 .63662 L closepath p .694 .746 .891 r
F P
0 g
s
.48893 .55876 m .44012 .61822 L .50145 .58698 L closepath p .692 .735 .884 r
F P
0 g
s
.50145 .58698 m .51665 .70195 L .48925 .56042 L p .922 .612 .138 r
F P
0 g
s
.48925 .56042 m .48893 .55876 L .50145 .58698 L p .922 .612 .138 r
F P
0 g
s
.54491 .57808 m .51665 .70195 L .50145 .58698 L closepath p .626 .353 .485 r
F P
0 g
s
.54491 .57808 m .51665 .70195 L .52815 .59133 L p .998 .791 .49 r
F P
0 g
s
.52815 .59133 m .53155 .55853 L .54491 .57808 L p .998 .791 .49 r
F P
0 g
s
.56138 .56506 m .57717 .54878 L .54491 .57808 L p .6 .894 .601 r
F P
0 g
s
.62992 .39345 m .69343 .43161 L .67528 .33915 L closepath p .76 .538 .597 r
F P
0 g
s
.41944 .56651 m .40899 .57688 L .43285 .56228 L closepath p .667 .967 .779 r
F P
0 g
s
.43285 .56228 m .44332 .55194 L .41944 .56651 L closepath p .667 .967 .779 r
F P
0 g
s
.47742 .22765 m .48805 .24485 L .57171 .26528 L closepath p .698 .207 .075 r
F P
0 g
s
.33158 .19019 m .48805 .24485 L .47742 .22765 L closepath p .687 .203 .094 r
F P
0 g
s
.27676 .16778 m .33158 .19019 L .47742 .22765 L closepath p .597 .048 0 r
F P
0 g
s
.49546 .63662 m .44012 .61822 L .42593 .63181 L closepath p .65 .758 .928 r
F P
0 g
s
.40954 .55433 m .41993 .54391 L .39411 .50867 L closepath p .506 .5 .784 r
F P
0 g
s
.40954 .55433 m .40899 .57688 L .41993 .54391 L closepath p .211 .063 .452 r
F P
0 g
s
.40954 .55433 m .40899 .57688 L .4569 .52936 L closepath p .211 .063 .452 r
F P
0 g
s
.40899 .57688 m .42593 .63181 L .44012 .61822 L closepath p .498 .49 .779 r
F P
0 g
s
.44332 .55194 m .51974 .47648 L .49581 .46785 L closepath p .654 .753 .923 r
F P
0 g
s
.5217 .39674 m .53042 .4234 L .53025 .39267 L p .973 .846 .444 r
F P
0 g
s
.53025 .39267 m .53011 .3683 L .5217 .39674 L p .973 .846 .444 r
F P
0 g
s
.5217 .39674 m .50377 .38658 L .51469 .40007 L p .864 .416 .062 r
F P
0 g
s
.5623 .36662 m .53011 .3683 L .53042 .4234 L closepath p .846 .787 .794 r
F P
0 g
s
.59548 .39266 m .5623 .36662 L .53042 .4234 L closepath p .72 .613 .737 r
F P
0 g
s
.6082 .41969 m .59548 .39266 L .53042 .4234 L closepath p .67 .492 .636 r
F P
0 g
s
.58695 .44335 m .6082 .41969 L .53042 .4234 L closepath p .592 .229 .322 r
F P
0 g
s
.53042 .4234 m .54633 .43137 L p .54712 .4315 L .309 0 0 r
F P
0 g
s
.5614 .43433 m .53042 .4234 L p .54767 .43177 L .309 0 0 r
F P
0 g
s
.5217 .39674 m .5386 .4275 L .53042 .4234 L p 0 .087 .644 r
F P
0 g
s
.54633 .43137 m .5217 .39674 L .54018 .43037 L p .856 .405 .024 r
F P
0 g
s
.54018 .43037 m .54222 .43408 L .54633 .43137 L p .856 .405 .024 r
F P
0 g
s
.54633 .43137 m .54767 .43177 L p .54712 .4315 L .309 0 0 r
F P
0 g
s
.54767 .43177 m .58695 .44335 L .5614 .43433 L p .309 0 0 r
F P
0 g
s
.51469 .40007 m .53904 .43016 L p .5217 .39674 L .864 .416 .062 r
F P
0 g
s
.53904 .43016 m .54222 .43408 L .54018 .43037 L p .864 .416 .062 r
F P
0 g
s
.54018 .43037 m .5217 .39674 L p .53904 .43016 L .864 .416 .062 r
F P
0 g
s
.59548 .39266 m .6082 .41969 L .69343 .43161 L closepath p .699 .504 .621 r
F P
0 g
s
.62992 .39345 m .59548 .39266 L .69343 .43161 L closepath p .569 .612 .856 r
F P
0 g
s
.5623 .36662 m .59548 .39266 L .57137 .35738 L closepath p .587 .615 .846 r
F P
0 g
s
.57137 .35738 m .59548 .39266 L .62992 .39345 L closepath p .571 .608 .851 r
F P
0 g
s
.49676 .30235 m .57208 .28686 L .49864 .28859 L closepath p .834 .779 .8 r
F P
0 g
s
.49117 .28018 m .49114 .28694 L .57208 .28686 L closepath p .805 .53 .521 r
F P
0 g
s
.57137 .35738 m .57208 .28686 L .53552 .29551 L closepath p .819 .628 .644 r
F P
0 g
s
.27676 .16778 m .26772 .16786 L .33158 .19019 L closepath p .518 0 0 r
F P
0 g
s
.27676 .16778 m .24687 .14442 L .26772 .16786 L closepath p .815 .342 .083 r
F P
0 g
s
.54491 .57808 m .60395 .52932 L .57717 .54878 L closepath p 0 0 0 r
F P
0 g
s
.57717 .54878 m .60395 .52932 L .61725 .51151 L closepath p 0 0 0 r
F P
0 g
s
.20995 .28175 m .20947 .28783 L .13883 .20044 L closepath p .846 .577 .532 r
F P
0 g
s
.39411 .50867 m .41993 .54391 L .37773 .4447 L closepath p .555 .518 .769 r
F P
0 g
s
.53612 .3621 m .53011 .3683 L .5623 .36662 L closepath p .737 .908 .953 r
F P
0 g
s
.53612 .3621 m .5623 .36662 L .57137 .35738 L closepath p .737 .908 .953 r
F P
0 g
s
.20947 .28783 m .13415 .20234 L .13883 .20044 L closepath p .834 .581 .556 r
F P
0 g
s
.20947 .28783 m .21184 .28487 L .13415 .20234 L closepath p .54 .552 .816 r
F P
0 g
s
.41993 .54391 m .40953 .44216 L .38749 .43423 L closepath p .793 .663 .72 r
F P
0 g
s
.41993 .54391 m .44829 .49469 L .40953 .44216 L closepath p .595 .518 .737 r
F P
0 g
s
.37773 .4447 m .41993 .54391 L .38749 .43423 L closepath p .497 .486 .775 r
F P
0 g
s
.40899 .57688 m .41944 .56651 L .41993 .54391 L closepath p .211 .063 .452 r
F P
0 g
s
.68208 .52011 m .7072 .55352 L .73562 .50722 L closepath p .789 .554 .581 r
F P
0 g
s
.13415 .20234 m .17323 .14956 L .13883 .20044 L closepath p .801 .984 .894 r
F P
0 g
s
.4569 .52936 m .40899 .57688 L .48893 .55876 L closepath p .551 .576 .833 r
F P
0 g
s
.40899 .57688 m .44012 .61822 L .48893 .55876 L closepath p .54 .505 .765 r
F P
0 g
s
.58695 .44335 m .69343 .43161 L .6082 .41969 L closepath p .679 .339 .392 r
F P
0 g
s
.21184 .28487 m .14322 .20282 L .13415 .20234 L closepath p .584 .563 .797 r
F P
0 g
s
.21184 .28487 m .15734 .25743 L .14322 .20282 L closepath p .898 .518 .303 r
F P
0 g
s
.21184 .28487 m .2047 .29315 L .18988 .28197 L p .87 .445 .198 r
F P
0 g
s
.18507 .27139 m .21184 .28487 L p .18988 .28197 L .87 .445 .198 r
F P
0 g
s
.21184 .28487 m .24524 .3262 L .2047 .29315 L closepath p .863 .433 .191 r
F P
0 g
s
.24524 .3262 m .23422 .25621 L .2047 .29315 L closepath p .605 .652 .87 r
F P
0 g
s
.19395 .28965 m .18091 .27818 L p .18988 .28197 L .847 .39 .107 r
F P
0 g
s
.17823 .27318 m .18988 .28197 L p .18091 .27818 L .847 .39 .107 r
F P
0 g
s
.18988 .28197 m .2047 .29315 L .19395 .28965 L p .847 .39 .107 r
F P
0 g
s
.19395 .28965 m .2047 .29315 L .20811 .28304 L p .764 .92 .941 r
F P
0 g
s
.20811 .28304 m .20989 .27777 L p .19823 .28424 L .764 .92 .941 r
F P
0 g
s
.20626 .27412 m .19823 .28424 L p .20989 .27777 L .764 .92 .941 r
F P
0 g
s
.19823 .28424 m .19395 .28965 L p .20811 .28304 L .764 .92 .941 r
F P
0 g
s
.23422 .25621 m .2147 .26348 L .21295 .26867 L p .764 .92 .941 r
F P
0 g
s
.21295 .26867 m .2047 .29315 L .22545 .2672 L p .764 .92 .941 r
F P
0 g
s
.22545 .2672 m .23422 .25621 L p .21295 .26867 L .764 .92 .941 r
F P
0 g
s
.19823 .28424 m .20626 .27412 L p .20425 .30908 L .274 .19 .575 r
F P
0 g
s
.20867 .28001 m .20425 .30908 L p .20626 .27412 L .274 .19 .575 r
F P
0 g
s
.21589 .29768 m .23875 .33439 L .20368 .31284 L closepath p .608 .662 .877 r
F P
0 g
s
.23069 .32016 m .20873 .28017 L .21589 .29768 L .843 .432 0 r
F
.23948 .33351 m .23875 .33439 L .21589 .29768 L p .608 .662 .877 r
F P
0 g
s
.23844 .33265 m .20873 .28017 L p .23681 .3313 L .843 .432 0 r
F P
0 g
s
.25254 .33698 m .24524 .3262 L .2605 .33961 L p .753 .237 0 r
F P
0 g
s
.2605 .33961 m .27083 .34869 L p .25254 .33698 L .753 .237 0 r
F P
0 g
s
.27083 .34869 m .34802 .41651 L p .26452 .34696 L .753 .237 0 r
F P
0 g
s
.27083 .34869 m .26452 .34696 L .25254 .33698 L .753 .237 0 r
F
.25254 .33698 m .26263 .34774 L .24646 .34681 L closepath p .883 .468 .213 r
F P
0 g
s
.24646 .34681 m .26263 .34774 L .26352 .34668 L p .772 .98 .91 r
F P
0 g
s
.3194 .39286 m .25254 .33698 L p .34802 .41651 L .753 .237 0 r
F P
0 g
s
.26352 .34668 m .29088 .31424 L .25123 .34331 L p .772 .98 .91 r
F P
0 g
s
.25123 .34331 m .24646 .34681 L p .26352 .34668 L .772 .98 .91 r
F P
0 g
s
.28746 .33894 m .29088 .31424 L .26352 .34668 L p .589 .63 .86 r
F P
0 g
s
.26352 .34668 m .26263 .34774 L .26415 .3472 L p .589 .63 .86 r
F P
0 g
s
.26415 .3472 m .28746 .33894 L p .26352 .34668 L .589 .63 .86 r
F P
0 g
s
.34802 .41651 m .3543 .42203 L .3194 .39286 L p .753 .237 0 r
F P
0 g
s
.25254 .33698 m .24109 .33483 L p .24196 .33349 L .773 .967 .925 r
F P
0 g
s
.25254 .33698 m .28294 .28067 L .24524 .3262 L closepath p .237 .131 .525 r
F P
0 g
s
.25115 .31942 m .25254 .33698 L p .24196 .33349 L .773 .967 .925 r
F P
0 g
s
.28294 .28067 m .23422 .25621 L .24524 .3262 L closepath p .605 .652 .87 r
F P
0 g
s
.21589 .29768 m .25115 .31942 L .23948 .33351 L p .608 .662 .877 r
F P
0 g
s
.21589 .29768 m .23681 .3313 L p .23069 .32016 L .843 .432 0 r
F P
0 g
s
.24109 .33483 m .23875 .33439 L .23948 .33351 L p .773 .967 .925 r
F P
0 g
s
.23875 .33439 m .25254 .33698 L .24646 .34681 L closepath p .907 .509 .233 r
F P
0 g
s
.23875 .33439 m .24646 .34681 L .19349 .31964 L closepath p .824 .554 .531 r
F P
0 g
s
.19349 .31964 m .20368 .31284 L .23875 .33439 L closepath p .824 .554 .531 r
F P
0 g
s
.23681 .3313 m .24646 .34681 L .23844 .33265 L p .843 .432 0 r
F P
0 g
s
.23948 .33351 m .25115 .31942 L p .24109 .33483 L .773 .967 .925 r
F P
0 g
s
.20989 .27777 m .2147 .26348 L .20626 .27412 L p .764 .92 .941 r
F P
0 g
s
.21034 .26898 m .21589 .29768 L .20368 .31284 L closepath p .274 .19 .575 r
F P
0 g
s
.20569 .30446 m .21589 .29768 L .20212 .26995 L p .825 .554 .528 r
F P
0 g
s
.20212 .26995 m .19618 .25799 L .19871 .27036 L p .825 .554 .528 r
F P
0 g
s
.19871 .27036 m .20569 .30446 L p .20212 .26995 L .825 .554 .528 r
F P
0 g
s
.20626 .27412 m .21034 .26898 L .20867 .28001 L p .274 .19 .575 r
F P
0 g
s
.20873 .28017 m .20405 .2719 L p .21589 .29768 L .843 .432 0 r
F P
0 g
s
.20405 .2719 m .19618 .25799 L .20212 .26995 L p .843 .432 0 r
F P
0 g
s
.20212 .26995 m .21589 .29768 L p .20405 .2719 L .843 .432 0 r
F P
0 g
s
.18091 .27818 m .15734 .25743 L .17823 .27318 L p .847 .39 .107 r
F P
0 g
s
.18988 .28197 m .15734 .25743 L .18507 .27139 L p .87 .445 .198 r
F P
0 g
s
.19395 .28965 m .18415 .27334 L .15734 .25743 L closepath p .61 .629 .845 r
F P
0 g
s
.19395 .28965 m .19823 .28424 L .18415 .27334 L closepath p .579 .615 .852 r
F P
0 g
s
.19823 .28424 m .20368 .31284 L .19003 .28524 L p .824 .554 .531 r
F P
0 g
s
.19003 .28524 m .18415 .27334 L .19823 .28424 L p .824 .554 .531 r
F P
0 g
s
.20425 .30908 m .20239 .30558 L .19823 .28424 L .274 .19 .575 r
F
.20351 .31197 m .19823 .28424 L p .20373 .31246 L F P
0 g
s
.20368 .31284 m .19349 .31964 L .18415 .27334 L closepath p .824 .554 .531 r
F P
0 g
s
.20569 .30446 m .18415 .27334 L .18765 .29071 L p .278 .197 .582 r
F P
0 g
s
.18765 .29071 m .19349 .31964 L .19932 .31239 L p .278 .197 .582 r
F P
0 g
s
.20425 .30908 m .20373 .31246 L p .20239 .30558 L .274 .19 .575 r
F P
0 g
s
.19932 .31239 m .20569 .30446 L p .18765 .29071 L .278 .197 .582 r
F P
0 g
s
.20373 .31246 m .20368 .31284 L .20351 .31197 L p .274 .19 .575 r
F P
0 g
s
.17575 .15912 m .17323 .14956 L .13415 .20234 L closepath p .572 .597 .84 r
F P
0 g
s
.14322 .20282 m .17575 .15912 L .13415 .20234 L closepath p .572 .597 .84 r
F P
0 g
s
.68208 .52011 m .69343 .43161 L .60186 .45153 L p .806 .622 .653 r
F P
0 g
s
.60251 .45281 m .68208 .52011 L p .60186 .45153 L .806 .622 .653 r
F P
0 g
s
.67025 .53077 m .67235 .58275 L .61876 .5122 L closepath p .828 .552 .522 r
F P
0 g
s
.69522 .56403 m .67235 .58275 L .67025 .53077 L closepath p .828 .552 .522 r
F P
0 g
s
.68435 .57222 m .69522 .56403 L .69969 .56012 L p .498 .846 .579 r
F P
0 g
s
.69969 .56012 m .7072 .55352 L .68906 .56836 L p .498 .846 .579 r
F P
0 g
s
.68906 .56836 m .68435 .57222 L p .69969 .56012 L .498 .846 .579 r
F P
0 g
s
.68435 .57222 m .67235 .58275 L .69522 .56403 L closepath p .498 .846 .579 r
F P
0 g
s
.7072 .55352 m .68208 .52011 L .68407 .5659 L p .828 .552 .52 r
F P
0 g
s
.68407 .5659 m .68435 .57222 L .68906 .56836 L p .828 .552 .52 r
F P
0 g
s
.68906 .56836 m .7072 .55352 L p .68407 .5659 L .828 .552 .52 r
F P
0 g
s
.68208 .52011 m .63044 .50135 L .67286 .55712 L p .828 .552 .52 r
F P
0 g
s
.67286 .55712 m .68435 .57222 L .68306 .54267 L p .828 .552 .52 r
F P
0 g
s
.68306 .54267 m .68208 .52011 L p .67286 .55712 L .828 .552 .52 r
F P
0 g
s
.60395 .52932 m .54491 .57808 L .58454 .54862 L p 0 0 0 r
F P
0 g
s
.60395 .52932 m .66126 .57352 L p .61916 .5142 L .755 .449 .468 r
F P
0 g
s
.63044 .50135 m .61876 .5122 L .67235 .58275 L closepath p .509 .493 .775 r
F P
0 g
s
.68435 .57222 m .63044 .50135 L .67235 .58275 L closepath p .509 .493 .775 r
F P
0 g
s
.66126 .57352 m .85204 .72066 L .67115 .55953 L p .755 .449 .468 r
F P
0 g
s
.67115 .55953 m .61725 .51151 L p .66126 .57352 L .755 .449 .468 r
F P
0 g
s
.85204 .72066 m .85966 .72007 L .68604 .5707 L p .797 .35 .18 r
F P
0 g
s
.68479 .57168 m .85204 .72066 L p .68604 .5707 L .797 .35 .18 r
F P
0 g
s
.61564 .5255 m .85072 .71911 L .60395 .52932 L closepath p 0 .308 .792 r
F P
0 g
s
.85072 .71911 m .85204 .72066 L .64953 .56447 L p 0 .138 .603 r
F P
0 g
s
.64953 .56447 m .60395 .52932 L .64715 .56255 L p 0 .138 .603 r
F P
0 g
s
.64715 .56255 m .85072 .71911 L p .64953 .56447 L 0 .138 .603 r
F P
0 g
s
.58454 .54862 m .61564 .5255 L .60395 .52932 L p 0 0 0 r
F P
0 g
s
.5969 .54283 m .55571 .50753 L .59307 .51873 L p .65 .758 .928 r
F P
0 g
s
.61725 .51151 m .60395 .52932 L p .61916 .5142 L .755 .449 .468 r
F P
0 g
s
.60117 .45168 m .56641 .47809 L .62881 .50076 L p .82 .615 .626 r
F P
0 g
s
.62881 .50076 m .63044 .50135 L .61155 .4693 L p .82 .615 .626 r
F P
0 g
s
.61155 .4693 m .60117 .45168 L p .62881 .50076 L .82 .615 .626 r
F P
0 g
s
.60117 .45168 m .63044 .50135 L .68208 .52011 L closepath p .82 .615 .626 r
F P
0 g
s
.87351 .70957 m .65553 .47591 L .62881 .50076 L p .51 .495 .776 r
F P
0 g
s
.62881 .50076 m .6274 .50207 L p .63136 .50541 L .51 .495 .776 r
F P
0 g
s
.63136 .50541 m .87351 .70957 L .62881 .50076 L .51 .495 .776 r
F
.54222 .43408 m .56641 .47809 L .60117 .45168 L closepath p .802 .582 .603 r
F P
0 g
s
.56591 .46581 m .60395 .52932 L .61564 .5255 L closepath p 0 .224 .743 r
F P
0 g
s
.59307 .51873 m .61564 .5255 L .609 .53164 L p .65 .758 .928 r
F P
0 g
s
.609 .53164 m .5969 .54283 L p .59307 .51873 L .65 .758 .928 r
F P
0 g
s
.68604 .5707 m .67561 .56173 L p .68479 .57168 L .797 .35 .18 r
F P
0 g
s
.67561 .56173 m .61725 .51151 L .67115 .55953 L p .797 .35 .18 r
F P
0 g
s
.67115 .55953 m .68479 .57168 L p .67561 .56173 L .797 .35 .18 r
F P
0 g
s
.6274 .50207 m .61725 .51151 L .64173 .53043 L p .51 .495 .776 r
F P
0 g
s
.64173 .53043 m .87351 .70957 L p .6274 .50207 L .51 .495 .776 r
F P
0 g
s
.60186 .45153 m .60117 .45168 L .60251 .45281 L p .806 .622 .653 r
F P
0 g
s
.58695 .44335 m .60117 .45168 L .69343 .43161 L closepath p .534 .064 .085 r
F P
0 g
s
.36685 .44609 m .41199 .3975 L .3543 .42203 L closepath p .51 .506 .788 r
F P
0 g
s
.41543 .4043 m .41199 .3975 L .36685 .44609 L closepath p .658 .749 .917 r
F P
0 g
s
.37773 .4447 m .41543 .4043 L .36685 .44609 L closepath p .658 .749 .917 r
F P
0 g
s
.53612 .3621 m .57137 .35738 L .53552 .29551 L closepath p .819 .628 .644 r
F P
0 g
s
.14322 .20282 m .15998 .24183 L .17575 .15912 L closepath p .138 .019 .442 r
F P
0 g
s
.14322 .20282 m .15734 .25743 L .15998 .24183 L closepath p 0 0 .445 r
F P
0 g
s
.3543 .42203 m .31527 .36869 L .26263 .34774 L closepath p .59 .627 .856 r
F P
0 g
s
.3543 .42203 m .36024 .3934 L .31527 .36869 L closepath p .572 .62 .861 r
F P
0 g
s
.39158 .38121 m .36024 .3934 L .3543 .42203 L closepath p .598 .645 .868 r
F P
0 g
s
.3543 .42203 m .41199 .3975 L .39158 .38121 L closepath p .51 .506 .788 r
F P
0 g
s
.40953 .44216 m .42757 .42311 L .4054 .41505 L closepath p .658 .749 .917 r
F P
0 g
s
.38749 .43423 m .40953 .44216 L .4054 .41505 L closepath p .658 .749 .917 r
F P
0 g
s
.40953 .44216 m .44829 .49469 L .42757 .42311 L closepath p .498 .485 .773 r
F P
0 g
s
.53612 .3621 m .53552 .29551 L .49676 .30235 L closepath p .815 .624 .645 r
F P
0 g
s
.44829 .49469 m .46672 .47604 L .42757 .42311 L closepath p .498 .485 .773 r
F P
0 g
s
.50377 .38658 m .5217 .39674 L .53011 .3683 L closepath p .975 .746 .53 r
F P
0 g
s
.50377 .38658 m .53011 .3683 L .53612 .3621 L closepath p .62 .922 .664 r
F P
0 g
s
.53612 .3621 m .49676 .30235 L .45932 .3494 L closepath p .763 .615 .695 r
F P
0 g
s
.50377 .38658 m .53612 .3621 L .45932 .3494 L closepath p .734 .357 .337 r
F P
0 g
s
.31527 .36869 m .28746 .33894 L .26263 .34774 L closepath p .588 .627 .858 r
F P
0 g
s
.54633 .43137 m .60117 .45168 L .58695 .44335 L closepath p .491 0 0 r
F P
0 g
s
.49114 .28694 m .49117 .28018 L .2585 .22088 L closepath p .754 .486 .527 r
F P
0 g
s
.28294 .28067 m .39158 .38121 L .48769 .2907 L closepath p .746 .593 .686 r
F P
0 g
s
.47809 .29023 m .48769 .2907 L .47933 .28852 L p .746 .571 .658 r
F P
0 g
s
.49114 .28694 m .48769 .2907 L .39158 .38121 L closepath p .63 .92 .646 r
F P
0 g
s
.49864 .28859 m .49114 .28694 L .45708 .3192 L p .612 .392 .559 r
F P
0 g
s
.45708 .3192 m .39158 .38121 L .44665 .33356 L p .612 .392 .559 r
F P
0 g
s
.44665 .33356 m .49864 .28859 L p .45708 .3192 L .612 .392 .559 r
F P
0 g
s
.25559 .23014 m .28294 .28067 L .47809 .29023 L p .746 .571 .658 r
F P
0 g
s
.47933 .28852 m .25559 .23014 L p .47809 .29023 L .746 .571 .658 r
F P
0 g
s
.27922 .30504 m .28294 .28067 L .25254 .33698 L closepath p .237 .131 .525 r
F P
0 g
s
.25115 .31942 m .20569 .30446 L .24046 .32238 L p .825 .554 .528 r
F P
0 g
s
.27203 .31364 m .25254 .33698 L .25115 .31942 L p .773 .967 .925 r
F P
0 g
s
.24046 .32238 m .25892 .3319 L .25115 .31942 L p .825 .554 .528 r
F P
0 g
s
.25115 .31942 m .27922 .30504 L .27203 .31364 L p .773 .967 .925 r
F P
0 g
s
.53155 .55853 m .5969 .54283 L .54491 .57808 L closepath p .582 .773 .97 r
F P
0 g
s
.46672 .47604 m .49581 .46785 L .42757 .42311 L closepath p .313 .409 .801 r
F P
0 g
s
.42757 .42311 m .49581 .46785 L .4054 .41505 L closepath p .623 .062 0 r
F P
0 g
s
.36024 .3934 m .39158 .38121 L .3253 .35723 L closepath p .593 .622 .85 r
F P
0 g
s
.36024 .3934 m .31533 .46387 L .31527 .36869 L closepath p .776 .487 .498 r
F P
0 g
s
.31533 .46387 m .36024 .3934 L .3253 .35723 L closepath p .869 .619 .558 r
F P
0 g
s
.31527 .36869 m .31533 .46387 L .28746 .33894 L closepath p .904 .602 .464 r
F P
0 g
s
.25808 .22702 m .49114 .28694 L .2585 .22088 L closepath p .754 .486 .527 r
F P
0 g
s
.61725 .51151 m .85966 .72007 L .87351 .70957 L closepath p .51 .495 .776 r
F P
0 g
s
.4054 .41505 m .49581 .46785 L .41543 .4043 L closepath p .497 .486 .775 r
F P
0 g
s
.49581 .46785 m .51974 .47648 L .49722 .48995 L p .828 .552 .522 r
F P
0 g
s
.49722 .48995 m .49574 .49084 L .49574 .48801 L p .828 .552 .522 r
F P
0 g
s
.49574 .48801 m .49581 .46785 L p .49722 .48995 L .828 .552 .522 r
F P
0 g
s
.50708 .47959 m .49574 .49084 L .51974 .47648 L closepath p .667 .967 .779 r
F P
0 g
s
.50708 .47959 m .53992 .52265 L .52837 .53083 L p .447 .471 .791 r
F P
0 g
s
.52837 .53083 m .48893 .55876 L .50297 .4975 L p .447 .471 .791 r
F P
0 g
s
.50297 .4975 m .50708 .47959 L p .52837 .53083 L .447 .471 .791 r
F P
0 g
s
.50709 .45655 m .4569 .52936 L .4572 .50656 L closepath p .211 .063 .452 r
F P
0 g
s
.50377 .38658 m .45932 .3494 L .41543 .4043 L closepath p .62 .421 .591 r
F P
0 g
s
.3253 .35723 m .29088 .31424 L .28746 .33894 L closepath p .586 .627 .859 r
F P
0 g
s
.31533 .46387 m .3253 .35723 L .28746 .33894 L closepath p .752 .509 .564 r
F P
0 g
s
.3253 .35723 m .39158 .38121 L .29088 .31424 L closepath p .583 .626 .86 r
F P
0 g
s
.54222 .43408 m .60117 .45168 L .54633 .43137 L closepath p .541 0 0 r
F P
0 g
s
.60117 .45168 m .65553 .47591 L .56294 .44529 L closepath p .466 0 0 r
F P
0 g
s
.56294 .44529 m .54222 .43408 L .60117 .45168 L closepath p .398 0 0 r
F P
0 g
s
.15998 .24183 m .17262 .23752 L .17575 .15912 L closepath p .433 .284 .578 r
F P
0 g
s
.41543 .4043 m .46676 .43369 L .50377 .38658 L closepath p .618 .417 .587 r
F P
0 g
s
.41543 .4043 m .49581 .46785 L .46676 .43369 L closepath p .447 .481 .802 r
F P
0 g
s
.18415 .27334 m .19014 .26569 L .15734 .25743 L closepath p .589 .63 .86 r
F P
0 g
s
.18415 .27334 m .20569 .30446 L .19618 .25799 L closepath p .278 .197 .582 r
F P
0 g
s
.27922 .30504 m .39158 .38121 L .28294 .28067 L closepath p .745 .593 .687 r
F P
0 g
s
.4569 .52936 m .50708 .47959 L .48893 .55876 L closepath p .551 .576 .833 r
F P
0 g
s
.53155 .55853 m .51665 .70195 L .48893 .55876 L closepath p .646 .436 .585 r
F P
0 g
s
.48893 .55876 m .53992 .52265 L .53155 .55853 L closepath p .574 .749 .96 r
F P
0 g
s
.15998 .24183 m .15734 .25743 L .17262 .23752 L closepath p .515 .523 .802 r
F P
0 g
s
.19014 .26569 m .17262 .23752 L .15734 .25743 L closepath p .589 .63 .86 r
F P
0 g
s
.25559 .23014 m .25808 .22702 L .17323 .14956 L closepath p .54 .552 .816 r
F P
0 g
s
.50709 .45655 m .50708 .47959 L .4569 .52936 L closepath p .211 .063 .452 r
F P
0 g
s
.17262 .23752 m .23422 .25621 L .25559 .23014 L closepath p .753 .569 .647 r
F P
0 g
s
.23422 .25621 m .28294 .28067 L .25559 .23014 L closepath p .757 .575 .65 r
F P
0 g
s
.19349 .31964 m .24646 .34681 L .25892 .3319 L closepath p .628 .695 .893 r
F P
0 g
s
.24646 .34681 m .29088 .31424 L .25892 .3319 L closepath p .772 .98 .91 r
F P
0 g
s
.19014 .26569 m .19618 .25799 L .17262 .23752 L closepath p .589 .63 .86 r
F P
0 g
s
.53155 .55853 m .53992 .52265 L .5969 .54283 L closepath p .586 .761 .962 r
F P
0 g
s
.25892 .3319 m .29088 .31424 L .27922 .30504 L closepath p .747 .559 .641 r
F P
0 g
s
.25115 .31942 m .25892 .3319 L .27922 .30504 L closepath p .731 .537 .629 r
F P
0 g
s
.21034 .26898 m .2147 .26348 L .19618 .25799 L closepath p .579 .615 .852 r
F P
0 g
s
.17262 .23752 m .2147 .26348 L .23422 .25621 L closepath p .75 .577 .661 r
F P
0 g
s
.19618 .25799 m .2147 .26348 L .17262 .23752 L closepath p .561 .625 .872 r
F P
0 g
s
.50377 .38658 m .46676 .43369 L .54222 .43408 L closepath p .744 .592 .687 r
F P
0 g
s
.46676 .43369 m .50709 .45655 L .54222 .43408 L closepath p .695 .359 .402 r
F P
0 g
s
.50709 .45655 m .56641 .47809 L .54222 .43408 L closepath p .825 .587 .579 r
F P
0 g
s
.54222 .43408 m .56294 .44529 L .50709 .45655 L closepath p .384 0 .06 r
F P
0 g
s
.25115 .31942 m .21589 .29768 L .20569 .30446 L closepath p .825 .554 .528 r
F P
0 g
s
.19349 .31964 m .25892 .3319 L .20569 .30446 L closepath p .628 .695 .893 r
F P
0 g
s
.51974 .47648 m .53109 .46527 L .50708 .47959 L closepath p .667 .967 .779 r
F P
0 g
s
.65553 .47591 m .86469 .71103 L .64504 .49121 L closepath p .825 .487 .417 r
F P
0 g
s
.65553 .47591 m .87351 .70957 L .86469 .71103 L closepath p .834 .479 .382 r
F P
0 g
s
.56294 .44529 m .65553 .47591 L .64504 .49121 L closepath p .811 .468 .407 r
F P
0 g
s
.46676 .43369 m .49581 .46785 L .50709 .45655 L closepath p .506 .5 .784 r
F P
0 g
s
.49581 .46785 m .49574 .49084 L .50709 .45655 L closepath p .211 .063 .452 r
F P
0 g
s
.50709 .45655 m .50708 .47959 L .53109 .46527 L closepath p .828 .552 .52 r
F P
0 g
s
.5969 .54283 m .53992 .52265 L .55571 .50753 L closepath p .65 .758 .928 r
F P
0 g
s
.49574 .49084 m .50708 .47959 L .50709 .45655 L closepath p .211 .063 .452 r
F P
0 g
s
.50709 .45655 m .56591 .46581 L .50708 .47959 L closepath p .579 .346 .527 r
F P
0 g
s
.50709 .45655 m .56294 .44529 L .56591 .46581 L closepath p .578 .313 .478 r
F P
0 g
s
.64504 .49121 m .85072 .71911 L .61564 .5255 L closepath p .737 .53 .613 r
F P
0 g
s
.64504 .49121 m .86469 .71103 L .85072 .71911 L closepath p .731 .531 .621 r
F P
0 g
s
.56294 .44529 m .64504 .49121 L .56591 .46581 L closepath p .81 .478 .426 r
F P
0 g
s
.56591 .46581 m .55571 .50753 L .50708 .47959 L closepath p .614 .445 .63 r
F P
0 g
s
.50708 .47959 m .55571 .50753 L .53992 .52265 L closepath p .498 .49 .779 r
F P
0 g
s
.85966 .72007 m .97051 .78691 L .87351 .70957 L closepath p .578 .627 .864 r
F P
0 g
s
.85966 .72007 m .85204 .72066 L .97051 .78691 L closepath p .788 .333 .163 r
F P
0 g
s
.86469 .71103 m .97051 .78691 L .85072 .71911 L closepath p .779 .636 .703 r
F P
0 g
s
.87351 .70957 m .97051 .78691 L .86469 .71103 L closepath p .857 .533 .436 r
F P
0 g
s
.85204 .72066 m .85072 .71911 L .97051 .78691 L closepath p 0 .293 .698 r
F P
0 g
s
.56591 .46581 m .61564 .5255 L .55571 .50753 L closepath p .797 .631 .675 r
F P
0 g
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[31]:=",
  ImageSize->{230.5, 220.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40003V0000g1000`40O003h00OiWoo003VOol00>Io
o`00iWoo003VOol00>Ioo`00iWoo003VOol00>Ioo`00iWoo003VOol00>Ioo`00iWoo003VOol00>Io
o`00iWoo003VOol00>Ioo`00iWoo003VOol00>Ioo`00iWoo003VOol00>Ioo`00iWoo003VOol00>Io
o`00iWoo003VOol00>Ioo`00iWoo003VOol00>Ioo`00:7oo00<007ooOol03Goo00<007ooOol0Zgoo
000WOol00`00IPl0000=Ool01000=IH00000Zgoo000VOol01@00AUX006H?00002goo00@003FF=IHe
UP800:Yoo`009Goo00<004IJAUX00P0000=V3`00Ool02Goo00<003FF=IH00SFF0P00ZGoo000UOol0
0`00AUX0000300000fH?00000007Ool00`00=IHeUP04=IH3002WOol002Aoo`060016FP0030d005hc
1@001Goo00<003FF=IH01CFF1P00Y7oo000SOol00`00AUX000030002GS<50003Ool00`00=IHeUP04
=IH300000dZJ00000004002ROol0029oo`050016FP00000<3@020003GS<500000goo000eUP03=IH3
00001CFF000004ZJ00000TjJ1`00WWoo000ROol3000230d20004GS<6000013FF0000000013FF00@0
04ZJBYY:VP<000=>VPH009]oo`008Goo0`000``=0P001Ehc1P001SFF00<004ZJBYX00TZJ0P0000=R
4TjJCYX014jJ1P00V7oo000POol3000430d20006GS<50005=IH00`00BYY:VP04BYX20002HQ801@00
CYY>VP00JU800TjJ1`00U7oo000OOol200000dIJ000<3@0230d00`00=A400008GS<40003=IH20000
0dZJ=II:VP03BYX200000f8B001>VP060003CYX7002AOol001moo`030016FP0000@<3@8000YN<`@0
008eUP03001:VTZJ00E:VP800003HQ800000009>VP05001ZDTjJCYX00005CYX7002>Ool001ioo`<0
00D<3@8000]N<`D00004BYX005F?EHl300001DZJ001oo`00FS@00`0000@01P00JU80009>VP04001I
lUWbFO82CYX8002;Ool001moo`8000D<3@8000aN<`<00003IQ0005F?00@00005BYX007ooOol00003
Ool00`0000H000020002CYX01P00FO9IlUWb001JE@=>VPT008Moo`0087oo0P000``=00<003DA0000
3Ehc0`0000AV4000001:VP<000Uoo`800003CY]J=0000080009IlP80009>VP04001>VTjJCYX30000
0eBR000000030024Ool0025oo`07000<3@`=30d003DA00003Uhc0`0000IV4000BYX004T=0005Ool0
1@00Oomoogoo00000Woo00P004jKCY/005Wb001>V`000TjJ0P0000=fd@00CYX00TjJ0P000eBR1000
PGoo000QOol2000230d00`00=A40000?GS<300001FH@000004T=00001Goo00<007ooOol01goo00<0
05Wb00000djK00<0000=00000`0000A>VUXdCYY>VP<000=DXPD007eoo`008Woo00H000`=30d003DA
000@GS<400000eg`000000020002Ool2000:Ool200002DjK001J=5XdCY/007KA001>V`050002CYX3
0004E:84001jOol0029oo`80000430d003DA000>GS<80003GO040006Ool01@00Oomoogoo00000goo
00<004jK00000P000djK00<0045b@G800`0000=>VUXdCYX00`001EBR1000MWoo000ROol30002=A40
0`00GS=N<`02GS<:0003GS@00`00GS@000060002GO050002Ool00`00Oomoo`07Ool30002FS@2CY/0
0`00@G800002CY/5000014jJ0000000015BR1P00LWoo000ROol0100000d003DA1P0035hd00@005hd
GS@000AN=0H0009Ml0D000Uoo`03001oo`0000800004FS@000000002CY/01000FO9>VdjK100000=>
VP0000000eBR00<005C1E<401`00KGoo000SOol400000dYk00000003000:GS@00`00GSAN=002GS@0
0`00GSAN=005GS@600000eg`000000030002Ool01@00Oomoogoo00000goo100014jK00@005WbFO80
00A>V`H000AD`@<0009Hi0D006Qoo`008goo0P0000<n6@00BW/0100000AN=000000000MN=003001N
=5hd009N=003001N=5hd00QN=0L00003GO00000000<000Uoo`<0009>V`03001IlP0000A>V`03001>
VdjK009>V`D000AD`@<000AHi0@006Aoo`008goo0`000dYk10000ehd10000Uhd00<005hdGS@015hd
00<005hdGS@02ehd3@000Woo00<007ooOol00goo0`0000=IlP00CY/024jK00=RM000Ea`00ULL1@00
0UC10`001EST1@00Ggoo000TOol20003BW/00`00BW]:N`030005GS@20007GS@00`00GSAN=00?GS@=
0006Ool00djK00000005CY/01@00HU<004jK00000ULL00<005LLEa`015LL100000=D`@00000025ST
1@00FWoo000TOol20004BW/00`00BW]:N`040002GS@50006GS@00`00GSAN=00BGS@=0003Ool00`00
Ool0000300001DjK000004jK00000eLL00<005LLEa`025LL1P002UST1000EWoo000UOol600000dYk
0000000500000enM00000002BY/20004GS@00`00GSAN=00EGS@?0002Ool01P00?<d00000CY/000=G
700529<007;DL]@00003Ea`M001AOol002Ioo`<00004BW/000008;8200001D8i000005nM00000TZK
00@004ZKBY]:V`8000=N=003001N=5hd01QN=28000IG7004001>^`00000<JNd7001=Ool002Moo`04
001>NP0000038;8300000enM001:V`02BY/00`00BY]:V`03BY/200000ehd001N=00MGS@>0003Ool0
0`00Ea`00002000=Ea`00`00C[]>^`02C[/2000:JNd00`00FJiI[P020002GDP50018Ool002Qoo`@0
008P/P<00003Gid004ZK009:V`03001:VdZK00E:V`<0021N=0h000Eoo`<000UG7003001>^djk00E>
^`03001YkFW]00QYk@04001I[UV^FJh20005GDP40014Ool002Yoo`80008P/P@000A:V`03001:VdZK
00M:V``001UN=0@00003D]`0000000h0009oo`D000Q>^`8000QYk@03001I[UV^00=I[P8000MMB0<0
045oo`00:goo0P0000<P/P00IS000P0014ZK00<004ZKBY/01TZK0`002UiE5P001ehd100000=Bg000
00005@001Tjk0P001VW]00<005V^FJh01EV^0P000ee80`0000AUk000000003moo`00:goo00D0022b
000006H`008000E:V`03001:VdZK00A:V`<00003GUD005iE01iNE@h00003CI40000000P0009[6@@0
009VM003001VM000008000E>^`8000AYk@03001I[UV^00MI[P<000=Uk005001b;G8]LRd0000nOol0
02]oo`<P/P800003IS0004ZK00A:V`03001:VdZK00A:V`03000LL000009NE@03001NEEiE02QNE@03
001=TDfA008000I[6@@000IVM004001VM6IdIW@50003C[/01@00JNeYkFW]00001eV^0`0000Lme`00
000006G/001b;@04000mOol002]oo`03000P/R2b00800003IS0004ZK00A:V`03001:VdZK00=:V`80
0003770005iE009NE@03001NEEiE02INE@03001=TDfA00<0009[6@@00003IW@0000000MVM003001V
M6Id00AVM0D00003C[/00000009Yk@04001I[UV^FJh50002?ML00`00JRmZ;`030003JJX2000lOol0
02]oo`04000P/R2b8;8200000fH`00000003BY/00`00BY]:V`02BY/2000277000`00GUENE@02GUD0
0`00GUENE@0TGUD00`00CI400002000016lh000000001FId0P001fId00<006IdIW@01VId0P0000=>
^`000000100000=I[P0000000`000Woo0P0000Dme`00000005:k0080009YZP03001YZ@0003]oo`00
:goo0P0000<P/P00C[`00`0000EV<000BY]:V`0000=:V`04000LL00077020005GUD2000SGUD02@00
CI4007oo001_>000GWD00006IW@00`00IW@00006IW@00`00IWAVM008IW@30002C[/50005Ool00`00
?ML0000500001e:k000006VZ001YZ@0003Yoo`00;7oo0P0014jl1P000TZK00<001a`00000`0000=N
=00000001EiE00<005iEGUD07eiE00L004fA001oo`00KcP00002GWD00`00IWAVM004IW@00`00IW@0
0006IW@00`00IWAVM00;IW@50008Ool3000017oo000005:k1@0000=YZ@00Ool0>Goo000/Ool20006
C[`00`00IS00000200001TZK00000000FQ<000=N=08000ENE@03001NEEiE01eNE@04001=T@00Ool2
00000flh001NM@02GWD00`00IWAVM003IW@01000IWAVM0001FId00<006IdIW@03FId0P002Goo00<0
07ooOol00P0000=oo`00D[/01`00>Goo000/Ool30006C[`2000016H`000000000UXC00<005hdGS@0
1P0015iE00<005iEGUD06eiE0TfA00H007oo000006lh0005GWD00`00IWAVM002IW@01@00IWAVM6Id
000016Id00<006IdIW@036Id10002Woo1`0000EZ;`00<C@006X_008003Qoo`00;Goo1`000djl0P00
00AV<000FQ=J4`@00003BW/004Yj00<000ANE@03001NEEiE01UNE@03CI400000009oo`03001_>000
00INM@03001VM6Id009VM005001VM6IdIW@00004IW@00`00IWAVM00:IW@20002HQ400`00FM400009
Ool60002JRl30002JRl2000gOol002moo`8000=>_0L00003FQ<0000000800006BW/004YjBWX004Yk
0`000eiE00<005iEGUD05eiE0P000goo00<006lh000025ie00<006IdIW@00VId00D006IdIWAVM000
00=VM003001VM6Id00UVM005001R4F8AHQ400003FM420005Ool20002JRl00`00IS5V<@03000016X_
000a=0000VX_0P00=Woo000aOol20006C[`5000016?L001:N`000dYj00<004YkBW/00P000eiE00<0
05iEGUD055iE0`000goo0`002Uie00D006IdIWAVM00000=VM005001VM6IdIW@00009IW@20004HQ40
0`00FM5Id@03FM400`00Oomoo`020005JRl01@00IS5V<@00JS001`00=goo000cOol20005C[`20002
Hm`01000BW/0000014Yj00<004YkBW/00`000UiE00<005iEGUD04UiE0`000goo0`002eie00D006Id
IWAVM00000AVM004001VM6Id0007IW@20006HQ400`00FM5Id@04FM440006JRl30003JS04000hOol0
03Eoo`80009>_080009Sg0<00004BWX000000003BWX01000BW]:NdYk0P000UiE00<005iEGUD045iE
0`000goo0P0000=_>000GWD02eie00@006IdIW@000EVM003001VM00000EVM08000QR4@03001IdEWA
00=Id@800003IU40000000H0009Z;`D003]oo`00=goo0P000f?L0P000dYj00<005gaGO400P000TYj
00@004YkBW]:N`80009NE@03001NEEiE00iNE@<000Aoo`03001_>00000eNM@04001VM6Id0005IW@0
0`00IW@00004IW@00`00HQ5R4@07HQ400`00FM5Id@03FM400`00GQ800003IU430003JRl5000mOol0
03eoo`<0009:NP04001MlEgaGO430005BW/200000eiE001NE@0=GUD30004Ool5000=GWD00`00IW@0
0006IW@20002IW@2000:HQ400`00FM5Id@03FM401000GQ9N4P001FIA0P001FX_00<007ooOol0>goo
000nOol50005GO420006BW/01000GUD000002eiE0`0000AYZ@00001oo`<00003CI8005U:00H000QN
M@8000IVM0@000aR4@03001IdEWA009Id@05001N4UhBGQ800007IU420004JRl00`00Oomoo`0jOol0
03moo`800003BWX005ga00IMl@8000E:N`800003GUD005iE00QNE@<000AYZ@<000==TP8000EIBPH0
009NM@8000AVM0@000iR4@05001IdEWAFM400004GQ800`00IU5VD@07IU4300000fX_001oo`0kOol0
041oo`<000MMl@D000A:N`03001NE@0000MNE@<00007IFD00000JJT006mH00001TfB00<005U:FDX0
25U:20000UnM00@004Yj000000aR4@04001IdEWA0005GQ800`00IU5VD@0:IU43000kOol0045oo`<0
00IMl@03001^L@00009:NP<0009:N`<000ENE@D0009UI@@000M=TP03001IBUU:00QIBPT0009:NP04
0016FP000009HQ401@00FM5IdEWA00001UhB00<006IAIU402fIA0P00>Woo0013Ool00`00GO5Ml@04
GO401000KW5^L@0014Yj0`0000=:N`0000000eiE00<005;L00000eLL1@002DfB00<005U:FDX01UU:
0P001V/H00<005Yg00000TYj00@004IJAUX000QR4@04001IdEWA0007GQ800`00IU5VD@0;IU400`00
JW80000iOol004=oo`03001MlEga00AMl@05001^LFiaKW400006BWX4000015iE001Bg0001ELL0`00
2dfB0P0015U:10001V/H00@005YgFWL0009:NP030016FTIJ008000IR4@03001Id@0000QN4P03001V
DFIA00YVD@04001ZLVYb000iOol004=oo`03001MlEga00AMl@03001^LFia009^L@03001:NTYj00I:
NP800003D]`0000000D00003KeP004fB00e=TP04001IBP000002N_020005JaP00`00FWMJM`02FWL3
0004AUX00`00HQ5R4@03HQ420009GQ800`00IU5VD@0:IU401@00JW9ZLVYb0000>7oo0014Ool00`00
GO5Ml@02GO400`00KW5^L@04KW400`00BWY:NP03BWX300000cgh000000020002Ea`00`00KeP0000>
CI830004N_020005JaP00`00FWMJM`04FWL20004AUX20002HQ400`00FM40000:GQ800`00IU5VD@08
IU400`00JW9ZLP02JW800`00Oomoo`0fOol004Aoo`03001MlEga009Ml@03001^LFia00E^L@03001:
NTYj00800003BY/003gh008mn0L000Y=TPD00004CI8005iE0003N_020004JaP00`00FWMJM`06FWL0
0`00AUY6FP04AUX4000;GQ800`00IU5VD@07IU400`00JW9ZLP04JW800`00Oomoo`0eOol004Aoo`03
001MlEga009Ml@03001^LFia00I^L@80009:V`03000mn3gh00@mn0D000E=TPD000I=TP03001NEEiE
00<00003Ndd006/H009[6003001JMeYg00MJM`X000aN4P03001VDFIA00IVD@03001ZLVYb00IZLP03
001oogoo03Aoo`00A7oo00<005gaGO400Uga00<006iaKW401Fia0P000TZK00<003gh?OP01Cgh0P00
00=_F000CI801@002TfB00@005iEGUENE@<00003Ndd006/H009[6003001JMeYg00EJM`80009Al`<0
00E6NP8000YN4P03001VDFIA00EVD@03001ZLVYb00MZLP03001oogoo03Aoo`00AGoo00D005gaGO5M
l@0000M^L@80009:V`03000mn3gh00<mn0<00004D_`00000000>CI800`00GUENE@03GUD40002JaP0
0`00FWMJM`04FWL20004DO<00`00EO=El`030004AWX20008GQ800`00IU5VD@05IU400`00JW9ZLP08
JW800`00Oomoo`0cOol004Eoo`05001MlEgaGO400006KW420002BY/01000?OPmn3gh0`0000Dmn000
D_`003gh008000e=TP03001NEEiE00ENE@<00003JaP005Yg00AJM`8000MAl`03001EleGc009El`80
00A6NP03001N4UhB00IN4P03001VDFIA00=VD@03001ZLVYb00YZLP03001oogoo039oo`00AGoo00D0
05gaGO5Ml@0000E^L@<00003BY/003gh00<000<mn004001Bo0000002?OP3000;CI800`00GUENE@06
GUD200000f/H001JM`02FWL20009DO<00`00EO=El`04EO<30002AWX20006GQ800`00IU5VD@02IU40
0`00JW9ZLP0;JW800`00Oomoo`0bOol004Eoo`05001MlEgaGO400005KW460005?OP01000D_`00000
13gh0`002DfB00<005iEGUD025iE0P000UYg0P00357c00<005GcEO<01UGc0`0000=6NP00000015hB
00D006IAIU5VD@0000mZLP03001oogoo035oo`00AWoo00@005gaGO4000A^L@03001oo`00008000Hm
n003001Bo00000@000@mn0<000I=TP03001NEEiE00YNE@<000IAl``000QEl`800003AWX005hB009N
4P04001VDFIA000@JW800`00Oomoo`0aOol004Ioo`04001MlEga0004KW401@00Oomoogoo00001Sgh
0`0000ABo000D_aBo08000<mn004000ef@000003CI800`00GUENE@0;GUD00`00JH0000060007B<T0
0`00EDaEC002ED`60004EO<40002GQ800`00IU40000=JW85000bOol004Ioo`04001MlEga0003KW40
0`00Oomoo`02Ool00`00?OPmn003?OP20003D_`01000D_aBo5;l0P000cgh00<003GI=MT00P0000==
TP00GUD03EiE10002TS900<005E<ED`02EE<300016Yb2P000fYc00<007ooOol0<Goo0016Ool01000
GO5Ml@000Via00<007ooOol017oo00@003gh?OPmn0<000=Bo004001Bo0000002?OP20002?OP01000
=MTefCGI5`001TS900<005E<ED`03UE<30003FYc00<007ooOol0<Goo0017Ool20003KW400`00Oomo
o`04Ool01`00?OPmn000D_`006:F00800003D_`0000000@mn0@00003?OP00000008ef@8000aEC080
0003;040000000H00098b@03001EC5E<00=EC100009VM080011ZL`03001oogoo039oo`00Agoo0P00
0Via00<007ooOol01Woo0P0000ABo000?OP0009RUP8000Dmn005000mn0009IT00002?OP00`00=MTe
f@030009ED`00`00;04/0@03;04>0005@2860007IW@2000@JW<00`00Oomoo`0bOol004Moo`800003
KW4007oo00Qoo`<0008mn004001RUV:FHYH20004?OP01@00?OPmn0009IT00P0000@mn000=MT0008m
n08000EEC08000D/0@/000=08PH000eVM003001VM00000iZL`03001oogoo03=oo`00Agoo0P0000=^
L@00Ool02Woo00@003gh?OP000=RUP03001:5`0000@mn004000mn3gh00029IT300000cGI000mn002
?OP2000015E<000002`11P000dD^30004FId00@006IdIW@000iZL`03001oogoo03=oo`00B7oo0P00
3Goo00@003gh?OP0009RUP03001:5`0000@mn004000mn3gh00049IT200000cGI000mn003?OP40009
ABh00`00ILeUc@050002>002000AIW@01@00IWAVM6Id000036Yc00<007ooOol0=7oo0018Ool2000=
Ool01000?OPmn0000f:F00<004XG00000cgh00D003gh?OPmn00000@UV@@0008mn08000=9D0H000=5
;P03001UcFG=009Uc@800005IL`00000000h0002000?IW@01@00IWAVM6Id00002fYc00<007ooOol0
=Goo0018Ool00`00Oomoo`0<Ool200000cgh001RUP02HYH01000BQM:5`000Sgh00<003gh?OP00Sgh
00<002FI9IT00bFI0P0000@mn000000000U9D0D000EUc@80009Uc0D000eVM003001VM6Id009VM003
001ZLfYc00QZL`03001oogoo03Eoo`00Egoo0P000Sgh00@006:FHYH000=:5`04000mn3gh0004?OP0
0`009ITUV@039IT200000aP]00000008BE0300000dfc000000030004ILd20003IL`4000;IW@00`00
IWAVM002IW@00`00JW=ZL`07JW<00`00Oomoo`0fOol005Qoo`800005?OP006:FHYH00003BQL01000
?OPmn00013gh00<002FI9IT01P0000<H;@00BE000dU@100014fc00<006Ie00000VIc0`000fG=0P00
16G<0`002FId00<006IdIW@00fId00<006YcJW<01VYc00<007ooOol0=Woo001HOol200001Cgh001R
UV:F000014XG00<003gh00001Cgh0P0000=B5@0000000`0000<H;@00BE000`001dfc00D006IeIWEV
M@0000AVL`@00003ILd0000000EUc08000IVM003001VM6Id00AVM003001ZLfYc00EZL`03001oogoo
03Moo`00F7oo00L004HH000mn000HYH00005BQL20004?OP01P00DQEB5E8E000H;@<0000362d00000
00Y=/`04001VMFIeIWD20007IW<50003IL`00`00IN`00005IW@00`00IWAVM005IW@00`00JW=ZL`03
JW<00`00Oomoo`0hOol005Uoo`800004?OP006:F0006BQL200000cgh00000003DQD0100062d001P]
1P002Dfc00<006IeIWD00VIe10001FIc00<006Ic00000`0000EUc000INaUk00000AVM003001VM6Id
00EVM003001ZLfYc00=ZL`03001oogoo03Qoo`00FGoo00H004HH000006:F0006BQL30004DQD00`00
62d0000262d200000dHj000ef0030007CK<00`00IWEVM@03IWD00`00D_d000030002IW<01000IW=V
LfIc0`0016G/00D006IdIWAVM00000QVM005001ZLfYcJW<0000kOol005Uoo`04001660000002HYH0
0`00BQM:5`04BQL00`00DQEB5@02DQD0100062dH;@000QP]00<004Hj00000SGH00<003gh?OP00P00
1Dfc00<006IeIWD016Ie00<005;mD_d01P0000IVL`00?OP005hC0004IN`01000IWAVM00026Id00D0
06YcJW=ZL`0003]oo`00FGoo00L004HH000mn000HYH00005BQL00`00DQEB5@02DQD0100062dH;@00
0QP]00@004HjASX000<ef004000mn3gh?OP20002CK<00`00IWEVM@06IWD00`00D_eBo@02D_d40000
0cgh001N4`02GQ<01000INaUk6G/0P0000=VM000IW@026Id00<006Yc0000?7oo001JOol01P00AQP0
0000HYH000E:5`80009B5@04000H;AP]000262d01@00ASY6>THj000013GH00<003gh?OP00Sgh0`00
2FIe00<005;mD_d00e;m00<003gh00001EhC00@006G/INaUk0<000UVM003001ZL`0003aoo`00FWoo
00H004HH000006:F0004BQL01@00FQ@0058E00000aP]0`0014Hj00<003GH=MP00cGH00<003gh?OP0
0cgh0P002FIe00L005;mD_eBo@00?OP00007GQ<01000INaUk6G/0P002FId0P00?Goo001JOol20000
14HH0000000014XG00<005XDFQ@00P000QP]0P001DHj00<003GH=MP013GH00<003gh?OP00Sgh00<0
05;m000010001FIe00@005;mD_d0008mn003001N4ehC00EN4`8000=Uk08000AVM0D003ioo`00Fgoo
1@000dXG0P0000EJ500062d001P]008000I6>P03000ef3GH00Def004000mn0000003D_d00`00D_eB
o@040003IWD200000ef^00000009GQ<20003IN`50004HS800`00Oomoo`0lOol005]oo`800003AQP0
000000=:5`<0008H;@<000M6>P03000ef3GH00Hef08000EBo@03001BoE;m009Bo@H00003GJh005hC
00YN4`8000=Uk003001R<V8b00AR<P03001oogoo03eoo`00Fgoo00<003gh00000P000TXG00<006Yc
00000aP]0P0024Hj00<003GH=MP01CGH0P0000=:oP00000015;m0P001U;m0`003EhC0P000fG/00<0
068bHS800f8b00<007ooOol0?Goo001LOol00`00?OP0000200000dXG0000000462d00`00ASY6>P07
ASX00`00=MPef003=MP200000dKn001:oP02B_h30003D_d00`00D_eBo@04D_d00`000CP0000=GQ<2
0003IN`01@00HS9R<V8b0000@7oo001LOol00`00?OPmn004000362d200000dHj0016>P06ASX00`00
=MPef003=MP01000A_i6oP001d[n0`0000=Bo@00D_d00e;m0P000P4h00<005hCGQ<02ehC0P000fG/
00@0068bHS80041oo`00GGoo00<003gh00000P000QP]0`000dHj00<004HjASX014Hj00<003GH=MP0
0SGH00D004KnA_i6oP0000Y:oP<00003D_d0000000<00081>08000aN4`80009Uk0<0045oo`00GGoo
00D003gh?OP001P]0080000362d004Hj00A6>P030016>THj00=6>P05000ef3GH=MP00004A_h00`00
B_i:oP0:B_h20002D_d01000D_eBo@000`4h00<005hCGQ<02UhC0P000VG/0P00@Goo001MOol00`00
?OP0000300000aP]0016>P06ASX20003ASX01000=MPef0001DKn00<004[nB_h01T[n1@000e;m00@0
05;mD_eBo@800081>003001N4ehC00YN4`80009Uk080041oo`00GWoo0`000QP]00<004HjASX024Hj
00@004Hj000ef08000E6oP03001:oT[n00=:oP@0009;7P03001BoE;m009Bo@03001BoE;m00=Bo@80
0081>003001N4ehC00UN4`03001J4`00009Uk08003moo`00GWoo0`0000<H;@00ASX02dHj0`001dKn
00<004[n00000`001D/N00<005;mD_d00e;m00<005;mD_d015;m0P000P4h00<005hCGQ<02EhC00<0
05XC00000VG/0P00?Woo001NOol0100062d0000034Hj3@001d/N0P001U;m00<005;mD_d01E;m0`00
00<1>00000002UhC00<005XC00000VG/0P00?Goo001NOol00`0062d00008ASX500000d7g00000008
DK820006Bah00`00D_eBo@06D_d00`00D_eBo@04D_d00`00IS1V<00200020CP00`00GQ=N4`08GQ<0
0`00FQ<00002IN`2000lOol005eoo`<000E6>P@000E1m`03001Fn`0000QA/P03001o3`0000A;7P03
001BoE;m00IBo@03001BoE;m00=Bo@8000AV<0800081>003001N4ehC00QN4`03001J4`00009Uk080
03]oo`00GGoo0P0000=6>P0000000`00247g00D005KkE_]Fn`0000MA/P05001o3gl?001;7P020009
D_d00`00D_eBo@02D_d20004IS000`00Ool0000200000`4h001N4`09GQ<00`00FQ<00002IN`2000j
Ool005eoo`<000e1m`05001FneKkE_/00007DK801000O`mo3gl?0P002U;m00@005;mD_eBo@8000EV
<004001oogooOol200000`4h001N4`09GQ<01`00FQ<006G/001Yb`0003Uoo`00GGoo0P003D7g00<0
05KkE_/00UKk00<0056bDK801E6b00D007l?O`mo3`0000YBo@07001BoE;mD_d006H`000016H`00<0
07ooOol00goo0P0000<1>00000002EhC00L005XC001Uk000JL/0000hOol005ioo`8000]1m`03001F
neKk00=Fn`03001A/U6b00AA/P8000=o3`03001BoE;m00QBo@04001BoE;m0002IS000`00IS1V<002
IS000`00Oomoo`04Ool300000`4h001N4`08GQ<01`00FQ<006G/001Yb`0003Moo`00GWoo00<003gh
00002D7g00<005KkE_/015Kk00<0056bDK80156b0P000Wl?100025;m00<005;m00000fH`00D006H`
IS1V<00000Uoo`8000030CP005hC00MN4`08001J4eXC001Uk000JL/003Ioo`00GWoo00@003gh?OP0
00M1m`03001FneKk00=Fn`<000IA/P80009o3`80009V<0<000EBo@8000AV<005001V<6H`IS00000:
Ool200000`4h001N4`07GQ<01@00FQ=J4`00IN`00P00=Woo001NOol200000cgh0011m`05@OL00`00
E_]Fn`02E_/20003E_`00`00DK9A/P03DK820002O`l20005IS0300000e;m00000005IS001000IS1V
<00037oo0P0000<1>000GQ<01ehC00D005XCFQ<006G/008003Eoo`00Ggoo00@003gh?OP000E1m`03
001FneKk008000EFo003001A/U6b00=A/P80009o3`<000MV<08000IV<004001V<6H`000>Ool40007
GQ<00`00FQ<00002IN`2000dOol005moo`05000mn3gh?OP00003@OL01000E_/000001eKl00<0056b
DK800e6b0P0000=o3`00KY800`001fH`0P001FH`00<006H`000047oo0P0000<1>000GQ<01UhC00<0
05XC00000VG/0P00<goo001OOol20003?OP00`00@OL000020009E_`00`00DK9A/P03DK802@00O`mo
3`00?OQ^TP00?OP00006IS000`00IS000004IS000`00IS00000AOol200000`4h001N4`06GQ<00`00
FQ<00002IN`2000bOol0061oo`03000mn3gh008mn08000]Fo003001A/U6b00=A/P04001o3gl?0002
KY830006IS001000IS1V<0000VH`00<006H`000057oo00<0004h00001ehC00L005XC001Uk000JL/0
000aOol0061oo`05000mn3gh?OP00002D_d20009E_`20004DK801000O`mo3`000fjB00<003gh0000
1FH`00@006H`IS00009V<003001V<00001Eoo`<000MN4`07001J4`00IN`006W;0000<7oo001QOol3
0005D_d20006E_`00`00E_`00004DK801`00O`l00000KY8mn6jB00800003?OP006H`00=V<006001V
<6H`IS0006H`0P005goo10001EhC00P005XCFQ<006G/001Yb`00;goo001QOol40006D_d30003E_`0
0`00E_`00004DK801000O`l0000016jB0P0000<mn000IS000VH`00<006H`IS000VH`0`0067oo0P00
00<1>000GQ<015hC00P005XCFQ<006G/001Yb`00;Woo001SOol:000015;m000005Kl0`000e6b0P00
00Ao3`00000000=^TP800003?OP006H`009V<003001V<6H`00=V<003001oogoo01Uoo`<000EN4`08
001J4eXC001Uk000JL/002eoo`00IGoo0P0000=Bo@0000000U;m1P0000=JEP00DK800U6b0P0000=o
3`00FUH00P000fjB0P0000Dmn000IS1V<00000AV<003001oogoo01]oo`<000EN4`08001J4eXC001U
k000JL/002aoo`00Igoo100015;m0P000UYF00@0056bDK9A/P<0009JEP8000=^TP80008mn003001V
<00000=V<003001oogoo01eoo`<000EN4`08001J4eXC001Uk000JL/002]oo`00JWoo0`0000=Bo@00
000015YF00H0056bDK8007l?0003FUH20003KY820002?OP200000fH`0000000ROol30004GQ<02000
FQ=J4`00IN`006W;000ZOol006aoo`8000IJEP06001A/U6b001o3`000eYF0P000fjB0P000Sgh0`00
9Goo0`0015hC00P005XCFQ<006G/001Yb`00:Goo001^Ool00`00FUIJEP03FUH01P00DK9A/P00O`l0
00=JEP<0009^TP<0008mn003001oogoo02Eoo`<000AN4`03001J4eXC0080009Yb`03001oogoo02Io
o`00Kgoo00<005YFFUH00UYF00<0056bDK800P001EYF0P000fjB0P0000<mn000Ool0:Goo0P0015hC
00<005XCFQ<00P000VW;00<007ooOol09Goo001_Ool00`00FUIJEP02FUH00`00DK9A/P020005FUH3
0002KY83000/Ool20003GQ<01000FQ=J4eXC0P000VW;00<007ooOol097oo001`Ool00`00FUIJEP02
FUH01000DK8000001UYF0`000VjB0P00;Goo0`000UhC00@005XCFQ=J4`80009Yb`03001oogoo02=o
o`00LGoo00H005YFFUIJEP00DK820006FUH00`00FUH00002KY82000^Ool30002GQ<01000FQ=J4eXC
0P0000=Yb`00Ool097oo001bOol01P00FUIJEP00DK8000QJEP<0009^TP8002moo`80009N4`04001J
4eXCFQ<200000fW;001oo`0SOol0079oo`03001JEUYF00<000QJEP03001JEP00009^TP80031oo`80
009N4`04001J4eXCFQ<200000fW;001oo`0ROol007=oo`03001JEP00008000UJEP<0009^TP80031o
o`80009N4`04001J4eXCFQ<200000fW;001oo`0QOol007Aoo`@000UJEP03001JEP00009^TP80039o
o`800003GQ<005XC009J4`800003JL/007oo021oo`00MGoo0P002UYF00P005YF000006jB001RM000
<Woo0P0000=N4`00FQ<00UXC00@006W;JL/0021oo`00MGoo0P002eYF00<005YF00000VjB0P00<goo
0P0000=N4`00FQ<00UXC00@006W;JL/001moo`00MWoo00<005YFFUH02EYF00P005YF000006jB001R
M000=7oo0P0015XC00@006W;JL/001ioo`00Mgoo00<005YFFUH02EYF00<005YF00000VjB0P00=Goo
0P0015XC00@006W;JL/001eoo`00N7oo00<005YFFUH025YF00P005YF000006jB001RM000=Goo0P00
15XC0P0000=Z3@00Ool06goo001hOol00`00FUIJEP09FUH00`00FUH00002KY82000gOol01@00FQ<0
0000HWD00P0000=Z3@0000006Woo001iOol00`00FUIJEP08FUH02000FUH00000KY80069d000gOol0
0`00HWERM@03HWD01000JPeZ3@006Goo001jOol00`00FUIJEP08FUH00`00FUH00002KY82000hOol2
0004HWD01000JPeZ3@0067oo001kOol00`00FUIJEP07FUH02000FUH00000KY80069d000iOol20003
HWD200000fX=001oo`0FOol007aoo`03001JEUYF00MJEP03001JEP00009^TP8003]oo`8000=RM@03
001Z3@0001Ioo`00O7oo00<005YFFUH01eYF00P005YFFUH006jB001RM000?7oo0P000V9e100057oo
001mOol00`00FUIJEP06FUH02000FUIJEP00KY80069d000nOol01`00HWERMF9e001Z3@0001=oo`00
OWoo00<005YFFUH01UYF00P005YFFUH006jB001RM000?Woo0P000V9e00<006X=00004Woo001oOol0
0`00FUIJEP05FUH02000FUIJEP00KY80069d0010Ool20000169e000000004Goo001oOol00`00FUIJ
EP06FUH02000FUIJEP00KY80069d0011Ool20000169e000000003goo0020Ool00`00FUIJEP05FUH0
2000FUIJEP00KY80069d0013Ool5000>Ool0085oo`03001JEUYF00EJEP05001JEUYF001^TP020015
Ool01000HWD000003Goo0022Ool00`00FUIJEP04FUH00`00FUIJEP0300000f9d001oo`14Ool4000<
Ool0089oo`03001JEUYF00EJEP05001JEUYF001^TP020017Ool3000;Ool008=oo`03001JEUYF00AJ
EP03001JEUYF00<00003HW@007oo04Moo`<000Uoo`00Q7oo00<005YFFUH015YF00<005YFFUH00P00
00=RM000Ool0BGoo0P0027oo0025Ool00`00FUIJEP03FUH00`00FUIJEP0300000f9d001oo`1:Ool0
0`00Oomoo`05Ool008Eoo`03001JEUYF00=JEP04001JEUYFFUH200000f9d001oo`1BOol008Ioo`03
001JEUYF00=JEP03001JEUYF00<00003HW@007oo055oo`00Qgoo00<005YFFUH00UYF00@005YFFUIJ
EP800003HW@007oo055oo`00R7oo00<005YFFUH00UYF00@005YFFUIJEP800003HW@007oo051oo`00
RGoo00D005YFFUIJEP0000=JEP800003HW@007oo051oo`00RGoo00<005YFFUH00UYF00@005YFFUIJ
EP800003HW@007oo04moo`00RWoo00D005YFFUIJEP0000=JEP800003HW@007oo04moo`00Rgoo00D0
05YFFUIJEP0000=JEP800003HW@007oo04ioo`00S7oo00@005YFFUH000=JEP800003HW@007oo04io
o`00S7oo00@005YFFUH000AJEP04001RM69d001>Ool008eoo`04001JEUYF0003FUH200000f9d001o
o`1=Ool008ioo`03001JEP0000AJEP04001RM69d001=Ool008moo`03001JEP0000=JEP800003HW@0
07oo04aoo`00Sgoo00<005YF000015YF00@0069dHW@004aoo`00T7oo00<005YF00000eYF0P0000=R
M000Ool0Bgoo002AOol20004FUH01000HWARM000Bgoo002BOol20003FUH200000f9d001oo`1:Ool0
099oo`8000AJEP800003HWD007oo04Uoo`00Tgoo0P0000AJEP00001NU`800003HWD007oo04Qoo`00
U7oo0P001EjG00@0069e000004Moo`00UGoo0P001EjG00@0069eHWD004Ioo`00Ugoo0P0015jG0P00
00=RM@00Ool0A7oo002IOol20004GYL00`00HWD00014Ool009]oo`8000=NU`03001RM@0004=oo`00
WGoo0P000UjG0`00@Woo002OOol20002GYL30010Ool00:5oo`03001NUejG00<003moo`00XWoo0P00
0UjG0P00?Woo002TOol200000ejG0000000mOol00:Ioo`@003aoo`00Z7oo1000>Woo002ZOol3000i
Ool00:aoo`8003Qoo`00[Woo00<007ooOol0=Goo003VOol00>Ioo`00iWoo003VOol00>Ioo`00\
\>"],
  ImageRangeCache->{{{0, 229.5}, {219.938, 0}} -> {-0.000213542, 0.00774752, 
  0.00428624, 0.00428624}}]
}, Open  ]],

Cell[BoxData[
    \(Needs["\<Graphics`Shapes`\>"]\)], "Input",
  CellLabel->"In[4]:="],

Cell[BoxData[
    \(MVClear[]; MVLoadOFF["\</export/home/kuska/MathView3D/head.off\>"]; \n
    grHead = MVGetGraphics3D[]; \)], "Input",
  CellLabel->"In[42]:="],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[RotateShape[grHead, 0, \(-Pi\)/2, \(-Pi\)/2], Axes -> True]; 
    \)\)], "Input",
  CellLabel->"In[5]:="],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.47973 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.229746 1.51917 1.11022e-016 1.51917 [
[.52852 1.48939 -6.00138 0 ]
[.52852 1.48939 5.99862 9 ]
[.63156 1.46525 -5.89261 0 ]
[.63156 1.46525 6.10739 9 ]
[.73716 1.4405 -2.89008 0 ]
[.73716 1.4405 3.10992 9 ]
[.84543 1.41512 -2.83192 0 ]
[.84543 1.41512 3.16808 9 ]
[.95647 1.38909 -2.77173 0 ]
[.95647 1.38909 3.22827 9 ]
[.09985 1.1949 -6.47664 0 ]
[.09985 1.1949 5.52336 9 ]
[.27488 1.33167 -3.13783 0 ]
[.27488 1.33167 2.86217 9 ]
[.42989 1.45279 -3.05288 0 ]
[.42989 1.45279 2.94712 9 ]
[.07001 .225 -6 -2.67033 ]
[.07001 .225 0 6.32967 ]
[.04831 .4407 -6 -2.78868 ]
[.04831 .4407 0 6.21132 ]
[.02439 .67848 -12 -2.9199 ]
[.02439 .67848 0 6.0801 ]
[-0.0021 .9419 -12 -3.06622 ]
[-0.0021 .9419 0 5.93378 ]
[ 0 0 0 0 ]
[ 1 1.47973 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
.4888 1.47973 m
1 1.35958 L
s
.52853 1.4704 m
.52853 1.4609 L
s
[(-4)] .52852 1.48939 .00023 -1 Mshowa
.63122 1.44626 m
.63105 1.43677 L
s
[(-2)] .63156 1.46525 -0.0179 -1 Mshowa
.73646 1.42153 m
.73612 1.41204 L
s
[(0)] .73716 1.4405 -0.03664 -1 Mshowa
.84437 1.39616 m
.84384 1.38668 L
s
[(2)] .84543 1.41512 -0.05603 -1 Mshowa
.95503 1.37016 m
.95431 1.36069 L
s
[(4)] .95647 1.38909 -0.07609 -1 Mshowa
.125 Mabswid
.55397 1.46442 m
.55394 1.45872 L
s
.57956 1.4584 m
.57951 1.45271 L
s
.60531 1.45235 m
.60523 1.44666 L
s
.65729 1.44014 m
.65716 1.43444 L
s
.68352 1.43397 m
.68336 1.42828 L
s
.70991 1.42777 m
.70973 1.42207 L
s
.76319 1.41524 m
.76295 1.40955 L
s
.79008 1.40892 m
.78981 1.40323 L
s
.81714 1.40256 m
.81685 1.39687 L
s
.87177 1.38972 m
.87142 1.38404 L
s
.89935 1.38324 m
.89897 1.37756 L
s
.9271 1.37672 m
.92669 1.37104 L
s
.50325 1.47634 m
.50327 1.47064 L
s
.98313 1.36355 m
.98267 1.35787 L
s
.25 Mabswid
0 1.09651 m
.4888 1.47973 L
s
.10135 1.17597 m
.1021 1.16651 L
s
[(-5)] .09985 1.1949 .07944 -1 Mshowa
.27575 1.3127 m
.27619 1.30322 L
s
[(0)] .27488 1.33167 .04594 -1 Mshowa
.43022 1.43381 m
.43039 1.42431 L
s
[(5)] .42989 1.45279 .01763 -1 Mshowa
.125 Mabswid
.13801 1.20471 m
.13842 1.19903 L
s
.17374 1.23273 m
.17412 1.22704 L
s
.20859 1.26005 m
.20893 1.25436 L
s
.24259 1.2867 m
.24288 1.28101 L
s
.30813 1.33808 m
.30836 1.33239 L
s
.33974 1.36286 m
.33993 1.35717 L
s
.3706 1.38707 m
.37077 1.38137 L
s
.40076 1.41071 m
.40089 1.40501 L
s
.06373 1.14648 m
.06423 1.1408 L
s
.02512 1.11621 m
.02566 1.11054 L
s
.45902 1.45639 m
.4591 1.45069 L
s
.48718 1.47846 m
.48722 1.47277 L
s
.25 Mabswid
.0896 .19777 m
0 1.09651 L
s
.0876 .21785 m
.0964 .21427 L
s
[(0)] .07001 .225 1 -0.40659 Mshowa
.06606 .43395 m
.07493 .43058 L
s
[(5)] .04831 .4407 1 -0.38029 Mshowa
.04231 .67218 m
.05126 .66904 L
s
[(10)] .02439 .67848 1 -0.35113 Mshowa
.01599 .93614 m
.02504 .93326 L
s
[(15)] -0.0021 .9419 1 -0.31862 Mshowa
.125 Mabswid
.08345 .25945 m
.08874 .25733 L
s
.07923 .30184 m
.08452 .29974 L
s
.07492 .34503 m
.08023 .34296 L
s
.07053 .38906 m
.07585 .38701 L
s
.06149 .47972 m
.06683 .47772 L
s
.05684 .5264 m
.06218 .52443 L
s
.05209 .57402 m
.05745 .57208 L
s
.04725 .62261 m
.05261 .62069 L
s
.03726 .72279 m
.04265 .72093 L
s
.03211 .77445 m
.0375 .77263 L
s
.02685 .82721 m
.03226 .82541 L
s
.02148 .88109 m
.02689 .87933 L
s
.01038 .99239 m
.01582 .99069 L
s
.00465 1.04987 m
.0101 1.04821 L
s
.25 Mabswid
.0896 .19777 m
0 1.09651 L
s
0 1.09651 m
.4888 1.47973 L
s
.4888 1.47973 m
.49477 .6513 L
s
.49477 .6513 m
.0896 .19777 L
s
.57312 0 m
.92808 .50635 L
s
.92808 .50635 m
1 1.35958 L
s
1 1.35958 m
.58259 .92246 L
s
.58259 .92246 m
.57312 0 L
s
.0896 .19777 m
0 1.09651 L
s
0 1.09651 m
.58259 .92246 L
s
.58259 .92246 m
.57312 0 L
s
.57312 0 m
.0896 .19777 L
s
.49477 .6513 m
.92808 .50635 L
s
.92808 .50635 m
1 1.35958 L
s
1 1.35958 m
.4888 1.47973 L
s
.4888 1.47973 m
.49477 .6513 L
s
0 0 m
1 0 L
1 1.47973 L
0 1.47973 L
closepath
clip
newpath
.5 Mabswid
.47513 .55131 m .42896 .52096 L .47275 .53152 L closepath p .995 .748 .802 r
F P
0 g
s
.51794 .55217 m .47513 .55131 L .47275 .53152 L closepath p .942 .71 .813 r
F P
0 g
s
.51794 .55217 m .47275 .53152 L .5175 .53401 L closepath p .944 .704 .802 r
F P
0 g
s
.54856 .55021 m .51794 .55217 L .5175 .53401 L closepath p .887 .659 .802 r
F P
0 g
s
.54856 .55021 m .5175 .53401 L .54671 .53103 L closepath p .887 .658 .801 r
F P
0 g
s
.43013 .54289 m .37673 .49789 L .42896 .52096 L closepath p 1 .801 .8 r
F P
0 g
s
.47513 .55131 m .43013 .54289 L .42896 .52096 L closepath p .991 .754 .816 r
F P
0 g
s
.56536 .54635 m .54856 .55021 L .54671 .53103 L closepath p .844 .626 .801 r
F P
0 g
s
.56536 .54635 m .54671 .53103 L .5653 .52682 L closepath p .845 .626 .799 r
F P
0 g
s
.58316 .54165 m .56536 .54635 L .5653 .52682 L closepath p .836 .618 .798 r
F P
0 g
s
.58316 .54165 m .5653 .52682 L .58166 .52257 L closepath p .834 .619 .8 r
F P
0 g
s
.47906 .58834 m .43013 .54289 L .47513 .55131 L closepath p .992 .754 .817 r
F P
0 g
s
.51963 .5901 m .47906 .58834 L .47513 .55131 L closepath p .936 .714 .827 r
F P
0 g
s
.51963 .5901 m .47513 .55131 L .51794 .55217 L closepath p .942 .711 .815 r
F P
0 g
s
.5517 .58702 m .51963 .5901 L .51794 .55217 L closepath p .888 .669 .816 r
F P
0 g
s
.5517 .58702 m .51794 .55217 L .54856 .55021 L closepath p .89 .669 .813 r
F P
0 g
s
.61518 .53391 m .58316 .54165 L .58166 .52257 L closepath p .843 .626 .801 r
F P
0 g
s
.61518 .53391 m .58166 .52257 L .61517 .51482 L closepath p .844 .623 .795 r
F P
0 g
s
.43255 .57606 m .37861 .51789 L .43013 .54289 L closepath p 1 .786 .763 r
F P
0 g
s
.47906 .58834 m .43255 .57606 L .43013 .54289 L closepath p 1 .745 .782 r
F P
0 g
s
.43013 .54289 m .37861 .51789 L .37673 .49789 L closepath p 1 .788 .767 r
F P
0 g
s
.56978 .58252 m .5517 .58702 L .54856 .55021 L closepath p .847 .637 .813 r
F P
0 g
s
.56978 .58252 m .54856 .55021 L .56536 .54635 L closepath p .848 .637 .813 r
F P
0 g
s
.58637 .57814 m .56978 .58252 L .56536 .54635 L closepath p .839 .631 .813 r
F P
0 g
s
.58637 .57814 m .56536 .54635 L .58316 .54165 L closepath p .841 .631 .811 r
F P
0 g
s
.37861 .51789 m .35164 .48134 L .37673 .49789 L closepath p 1 .821 .737 r
F P
0 g
s
.61517 .51482 m .66198 .4944 L .66181 .51425 L closepath p .769 .568 .792 r
F P
0 g
s
.61517 .51482 m .66181 .51425 L .61518 .53391 L closepath p .767 .561 .783 r
F P
0 g
s
.61703 .57086 m .58637 .57814 L .58316 .54165 L closepath p .85 .638 .812 r
F P
0 g
s
.61703 .57086 m .58316 .54165 L .61518 .53391 L closepath p .848 .638 .813 r
F P
0 g
s
.48191 .62441 m .43255 .57606 L .47906 .58834 L closepath p 1 .763 .808 r
F P
0 g
s
.52446 .62431 m .48191 .62441 L .47906 .58834 L closepath p .946 .717 .818 r
F P
0 g
s
.52446 .62431 m .47906 .58834 L .51963 .5901 L closepath p .937 .722 .837 r
F P
0 g
s
.55462 .61864 m .52446 .62431 L .51963 .5901 L closepath p .883 .684 .842 r
F P
0 g
s
.55462 .61864 m .51963 .5901 L .5517 .58702 L closepath p .892 .681 .829 r
F P
0 g
s
.38147 .55441 m .3516 .50254 L .37861 .51789 L closepath p 1 .828 .766 r
F P
0 g
s
.43255 .57606 m .38147 .55441 L .37861 .51789 L closepath p 1 .794 .791 r
F P
0 g
s
.37861 .51789 m .3516 .50254 L .35164 .48134 L closepath p 1 .829 .769 r
F P
0 g
s
.61518 .53391 m .66181 .51425 L .6623 .55203 L closepath p .779 .583 .805 r
F P
0 g
s
.61518 .53391 m .6623 .55203 L .61703 .57086 L closepath p .779 .584 .806 r
F P
0 g
s
.3516 .50254 m .35227 .44667 L .35164 .48134 L closepath p .009 .009 .027 r
F P
0 g
s
.43319 .61438 m .38147 .55441 L .43255 .57606 L closepath p 1 .798 .798 r
F P
0 g
s
.48191 .62441 m .43319 .61438 L .43255 .57606 L closepath p 1 .762 .807 r
F P
0 g
s
.57329 .61557 m .55462 .61864 L .5517 .58702 L closepath p .876 .669 .83 r
F P
0 g
s
.57329 .61557 m .5517 .58702 L .56978 .58252 L closepath p .859 .666 .843 r
F P
0 g
s
.59018 .61149 m .57329 .61557 L .56978 .58252 L closepath p .858 .665 .843 r
F P
0 g
s
.59018 .61149 m .56978 .58252 L .58637 .57814 L closepath p .853 .664 .846 r
F P
0 g
s
.66198 .4944 m .67704 .48761 L .67807 .50707 L closepath p .764 .562 .788 r
F P
0 g
s
.66198 .4944 m .67807 .50707 L .66181 .51425 L closepath p .762 .562 .79 r
F P
0 g
s
.62098 .60387 m .59018 .61149 L .58637 .57814 L closepath p .86 .669 .846 r
F P
0 g
s
.62098 .60387 m .58637 .57814 L .61703 .57086 L closepath p .862 .669 .844 r
F P
0 g
s
.35115 .53459 m .3516 .50254 L .35196 .50317 L p 1 .825 .744 r
F P
0 g
s
.35196 .50317 m .38147 .55441 L .35115 .53459 L p 1 .825 .744 r
F P
0 g
s
.38376 .59683 m .35 .57638 L .35115 .53459 L closepath p 1 .836 .766 r
F P
0 g
s
.35115 .53459 m .35 .57638 L .35018 .57256 L p .02 .02 .046 r
F P
0 g
s
.38376 .59683 m .35115 .53459 L .38147 .55441 L closepath p 1 .836 .765 r
F P
0 g
s
.35018 .57256 m .35337 .50273 L .35115 .53459 L p .02 .02 .046 r
F P
0 g
s
.35115 .53459 m .35268 .46964 L .3516 .50254 L closepath p .009 .009 .018 r
F P
0 g
s
.35115 .53459 m .35337 .50273 L .35268 .46964 L closepath p .042 .042 .151 r
F P
0 g
s
.35268 .46964 m .35337 .50273 L .35328 .50227 L p 1 .83 .376 r
F P
0 g
s
.35227 .44709 m .3516 .50254 L .35268 .46964 L p .024 .024 .091 r
F P
0 g
s
.35328 .50227 m .34069 .43538 L .35268 .46964 L p 1 .83 .376 r
F P
0 g
s
.35268 .46964 m .35227 .44667 L .35227 .44709 L p .024 .024 .091 r
F P
0 g
s
.52721 .67162 m .48191 .62441 L .52446 .62431 L closepath p .95 .734 .839 r
F P
0 g
s
.5592 .66866 m .52721 .67162 L .52446 .62431 L closepath p .9 .695 .841 r
F P
0 g
s
.5592 .66866 m .52446 .62431 L .55462 .61864 L closepath p .886 .693 .852 r
F P
0 g
s
.67704 .48761 m .69362 .47944 L .69314 .49979 L closepath p .75 .553 .788 r
F P
0 g
s
.67704 .48761 m .69314 .49979 L .67807 .50707 L closepath p .751 .552 .786 r
F P
0 g
s
.48564 .66923 m .43319 .61438 L .48191 .62441 L closepath p 1 .772 .82 r
F P
0 g
s
.52721 .67162 m .48564 .66923 L .48191 .62441 L closepath p .956 .733 .83 r
F P
0 g
s
.43319 .61438 m .38376 .59683 L .38147 .55441 L closepath p 1 .796 .792 r
F P
0 g
s
.66181 .51425 m .67807 .50707 L .67757 .54544 L closepath p .769 .577 .804 r
F P
0 g
s
.66181 .51425 m .67757 .54544 L .6623 .55203 L closepath p .772 .578 .803 r
F P
0 g
s
.57763 .66414 m .5592 .66866 L .55462 .61864 L closepath p .875 .685 .852 r
F P
0 g
s
.57763 .66414 m .55462 .61864 L .57329 .61557 L closepath p .882 .688 .85 r
F P
0 g
s
.61703 .57086 m .6623 .55203 L .66532 .58596 L closepath p .798 .624 .843 r
F P
0 g
s
.61703 .57086 m .66532 .58596 L .62098 .60387 L closepath p .799 .623 .841 r
F P
0 g
s
.69362 .47944 m .7158 .4642 L .71639 .48274 L closepath p .676 .493 .77 r
F P
0 g
s
.69362 .47944 m .71639 .48274 L .69314 .49979 L closepath p .676 .493 .769 r
F P
0 g
s
.59569 .65972 m .57763 .66414 L .57329 .61557 L closepath p .859 .671 .849 r
F P
0 g
s
.59569 .65972 m .57329 .61557 L .59018 .61149 L closepath p .861 .671 .849 r
F P
0 g
s
.4359 .66029 m .38376 .59683 L .43319 .61438 L closepath p 1 .808 .811 r
F P
0 g
s
.48564 .66923 m .4359 .66029 L .43319 .61438 L closepath p 1 .772 .825 r
F P
0 g
s
.67807 .50707 m .69314 .49979 L .69387 .53809 L closepath p .761 .57 .803 r
F P
0 g
s
.67807 .50707 m .69387 .53809 L .67757 .54544 L closepath p .762 .571 .803 r
F P
0 g
s
.62658 .65217 m .59569 .65972 L .59018 .61149 L closepath p .859 .67 .849 r
F P
0 g
s
.62658 .65217 m .59018 .61149 L .62098 .60387 L closepath p .86 .67 .847 r
F P
0 g
s
.35268 .46964 m .33918 .40868 L .35227 .44667 L closepath p 1 .846 .409 r
F P
0 g
s
.69314 .49979 m .71639 .48274 L .71856 .52144 L closepath p .693 .519 .791 r
F P
0 g
s
.69314 .49979 m .71856 .52144 L .69387 .53809 L closepath p .695 .518 .789 r
F P
0 g
s
.62098 .60387 m .66532 .58596 L .66999 .63493 L closepath p .801 .627 .846 r
F P
0 g
s
.62098 .60387 m .66999 .63493 L .62658 .65217 L closepath p .801 .627 .845 r
F P
0 g
s
.7158 .4642 m .74132 .43637 L .74148 .45762 L closepath p .54 .399 .749 r
F P
0 g
s
.7158 .4642 m .74148 .45762 L .71639 .48274 L closepath p .533 .383 .731 r
F P
0 g
s
.52915 .71239 m .48564 .66923 L .52721 .67162 L closepath p .959 .749 .851 r
F P
0 g
s
.5636 .70907 m .52915 .71239 L .52721 .67162 L closepath p .906 .708 .853 r
F P
0 g
s
.5636 .70907 m .52721 .67162 L .5592 .66866 L closepath p .903 .708 .855 r
F P
0 g
s
.4873 .70858 m .4359 .66029 L .48564 .66923 L closepath p 1 .772 .824 r
F P
0 g
s
.52915 .71239 m .4873 .70858 L .48564 .66923 L closepath p .972 .746 .829 r
F P
0 g
s
.6623 .55203 m .67757 .54544 L .68043 .57916 L closepath p .791 .615 .838 r
F P
0 g
s
.6623 .55203 m .68043 .57916 L .66532 .58596 L closepath p .784 .613 .841 r
F P
0 g
s
.5835 .70398 m .5636 .70907 L .5592 .66866 L closepath p .87 .684 .856 r
F P
0 g
s
.5835 .70398 m .5592 .66866 L .57763 .66414 L closepath p .876 .685 .852 r
F P
0 g
s
.38568 .64212 m .35 .57638 L .38376 .59683 L closepath p 1 .844 .781 r
F P
0 g
s
.4359 .66029 m .38568 .64212 L .38376 .59683 L closepath p 1 .807 .804 r
F P
0 g
s
.35268 .46964 m .34069 .43538 L .33918 .40868 L closepath p 1 .791 .316 r
F P
0 g
s
.5988 .70014 m .5835 .70398 L .57763 .66414 L closepath p .865 .678 .853 r
F P
0 g
s
.5988 .70014 m .57763 .66414 L .59569 .65972 L closepath p .861 .676 .855 r
F P
0 g
s
.67757 .54544 m .69387 .53809 L .69578 .57062 L closepath p .773 .591 .822 r
F P
0 g
s
.67757 .54544 m .69578 .57062 L .68043 .57916 L closepath p .751 .585 .833 r
F P
0 g
s
.71639 .48274 m .74148 .45762 L .74345 .49658 L closepath p .575 .44 .778 r
F P
0 g
s
.71639 .48274 m .74345 .49658 L .71856 .52144 L closepath p .584 .435 .764 r
F P
0 g
s
.43718 .70038 m .38568 .64212 L .4359 .66029 L closepath p 1 .812 .811 r
F P
0 g
s
.4873 .70858 m .43718 .70038 L .4359 .66029 L closepath p 1 .772 .824 r
F P
0 g
s
.63049 .69237 m .5988 .70014 L .59569 .65972 L closepath p .863 .678 .855 r
F P
0 g
s
.63049 .69237 m .59569 .65972 L .62658 .65217 L closepath p .862 .678 .856 r
F P
0 g
s
.53193 .75382 m .4873 .70858 L .52915 .71239 L closepath p .974 .755 .841 r
F P
0 g
s
.56655 .74937 m .53193 .75382 L .52915 .71239 L closepath p .908 .705 .845 r
F P
0 g
s
.56655 .74937 m .52915 .71239 L .5636 .70907 L closepath p .905 .705 .85 r
F P
0 g
s
.69387 .53809 m .71856 .52144 L .72059 .55747 L closepath p .724 .564 .828 r
F P
0 g
s
.69387 .53809 m .72059 .55747 L .69578 .57062 L closepath p .734 .562 .816 r
F P
0 g
s
.66532 .58596 m .68043 .57916 L .68628 .62787 L closepath p .786 .617 .845 r
F P
0 g
s
.66532 .58596 m .68628 .62787 L .66999 .63493 L closepath p .788 .617 .844 r
F P
0 g
s
.49108 .75227 m .43718 .70038 L .4873 .70858 L closepath p 1 .778 .833 r
F P
0 g
s
.53193 .75382 m .49108 .75227 L .4873 .70858 L closepath p .975 .754 .839 r
F P
0 g
s
.62658 .65217 m .66999 .63493 L .67594 .67479 L closepath p .805 .635 .852 r
F P
0 g
s
.62658 .65217 m .67594 .67479 L .63049 .69237 L closepath p .804 .635 .853 r
F P
0 g
s
.58774 .74489 m .56655 .74937 L .5636 .70907 L closepath p .87 .679 .849 r
F P
0 g
s
.58774 .74489 m .5636 .70907 L .5835 .70398 L closepath p .868 .678 .85 r
F P
0 g
s
.74132 .43637 m .75789 .40337 L .75948 .42605 L closepath p .315 .238 .682 r
F P
0 g
s
.74132 .43637 m .75948 .42605 L .74148 .45762 L closepath p .3 .21 .654 r
F P
0 g
s
.68043 .57916 m .69578 .57062 L .70215 .62091 L closepath p .758 .599 .845 r
F P
0 g
s
.68043 .57916 m .70215 .62091 L .68628 .62787 L closepath p .767 .603 .843 r
F P
0 g
s
.38568 .64212 m .3512 .62034 L .35 .57638 L closepath p 1 .842 .758 r
F P
0 g
s
.3512 .62034 m .353 .54532 L .35 .57638 L closepath p .058 .058 .208 r
F P
0 g
s
.35 .57638 m .353 .54532 L .35337 .50273 L closepath p .04 .04 .123 r
F P
0 g
s
.60307 .74168 m .58774 .74489 L .5835 .70398 L closepath p .87 .68 .85 r
F P
0 g
s
.60307 .74168 m .5835 .70398 L .5988 .70014 L closepath p .865 .678 .853 r
F P
0 g
s
.353 .54532 m .34286 .46786 L .35337 .50273 L closepath p 1 .842 .384 r
F P
0 g
s
.35337 .50273 m .34286 .46786 L .34069 .43538 L closepath p 1 .756 .255 r
F P
0 g
s
.38761 .67919 m .3512 .62034 L .38568 .64212 L closepath p 1 .843 .76 r
F P
0 g
s
.43718 .70038 m .38761 .67919 L .38568 .64212 L closepath p 1 .806 .788 r
F P
0 g
s
.34069 .43538 m .33831 .3697 L .33918 .40868 L closepath p .016 .016 .092 r
F P
0 g
s
.63765 .73314 m .60307 .74168 L .5988 .70014 L closepath p .863 .676 .853 r
F P
0 g
s
.63765 .73314 m .5988 .70014 L .63049 .69237 L closepath p .863 .676 .853 r
F P
0 g
s
.43791 .74302 m .38761 .67919 L .43718 .70038 L closepath p 1 .819 .81 r
F P
0 g
s
.49108 .75227 m .43791 .74302 L .43718 .70038 L closepath p 1 .776 .823 r
F P
0 g
s
.71856 .52144 m .74345 .49658 L .74519 .53422 L closepath p .591 .444 .772 r
F P
0 g
s
.71856 .52144 m .74519 .53422 L .72059 .55747 L closepath p .579 .453 .792 r
F P
0 g
s
.69578 .57062 m .72059 .55747 L .72624 .60495 L closepath p .743 .577 .829 r
F P
0 g
s
.69578 .57062 m .72624 .60495 L .70215 .62091 L closepath p .724 .574 .841 r
F P
0 g
s
.74148 .45762 m .75948 .42605 L .75872 .46172 L closepath p .239 .184 .596 r
F P
0 g
s
.74148 .45762 m .75872 .46172 L .74345 .49658 L closepath p .226 .184 .644 r
F P
0 g
s
.53835 .79825 m .49108 .75227 L .53193 .75382 L closepath p .975 .759 .845 r
F P
0 g
s
.57062 .79475 m .53835 .79825 L .53193 .75382 L closepath p .92 .72 .854 r
F P
0 g
s
.57062 .79475 m .53193 .75382 L .56655 .74937 L closepath p .913 .72 .862 r
F P
0 g
s
.63049 .69237 m .67594 .67479 L .68089 .71651 L closepath p .803 .632 .85 r
F P
0 g
s
.63049 .69237 m .68089 .71651 L .63765 .73314 L closepath p .801 .634 .854 r
F P
0 g
s
.59164 .78926 m .57062 .79475 L .56655 .74937 L closepath p .87 .688 .862 r
F P
0 g
s
.59164 .78926 m .56655 .74937 L .58774 .74489 L closepath p .874 .689 .86 r
F P
0 g
s
.66999 .63493 m .68628 .62787 L .68955 .66912 L closepath p .79 .623 .849 r
F P
0 g
s
.66999 .63493 m .68955 .66912 L .67594 .67479 L closepath p .785 .621 .851 r
F P
0 g
s
.4945 .79471 m .43791 .74302 L .49108 .75227 L closepath p 1 .777 .825 r
F P
0 g
s
.53835 .79825 m .4945 .79471 L .49108 .75227 L closepath p .983 .756 .831 r
F P
0 g
s
.60823 .7858 m .59164 .78926 L .58774 .74489 L closepath p .864 .682 .86 r
F P
0 g
s
.60823 .7858 m .58774 .74489 L .60307 .74168 L closepath p .872 .686 .857 r
F P
0 g
s
.34286 .46786 m .34182 .40251 L .34069 .43538 L closepath p .041 .041 .179 r
F P
0 g
s
.34069 .43538 m .34182 .40251 L .33831 .3697 L closepath p .058 .058 .261 r
F P
0 g
s
.68628 .62787 m .70215 .62091 L .70719 .66184 L closepath p .771 .611 .851 r
F P
0 g
s
.68628 .62787 m .70719 .66184 L .68955 .66912 L closepath p .779 .614 .848 r
F P
0 g
s
.64181 .77813 m .60823 .7858 L .60307 .74168 L closepath p .864 .68 .857 r
F P
0 g
s
.64181 .77813 m .60307 .74168 L .63765 .73314 L closepath p .864 .68 .857 r
F P
0 g
s
.72059 .55747 m .74519 .53422 L .74886 .58054 L closepath p .573 .444 .785 r
F P
0 g
s
.72059 .55747 m .74886 .58054 L .72624 .60495 L closepath p .561 .444 .793 r
F P
0 g
s
.39105 .72085 m .35274 .65702 L .38761 .67919 L closepath p 1 .849 .747 r
F P
0 g
s
.43791 .74302 m .39105 .72085 L .38761 .67919 L closepath p 1 .814 .783 r
F P
0 g
s
.38761 .67919 m .35274 .65702 L .3512 .62034 L closepath p 1 .835 .72 r
F P
0 g
s
.7636 .37925 m .75789 .40337 L .76342 .36093 L p .062 .062 .199 r
F P
0 g
s
.75789 .40337 m .7636 .37925 L .75948 .42605 L closepath p .083 .083 .323 r
F P
0 g
s
.75948 .42605 m .7636 .37925 L .76513 .41803 L closepath p .069 .069 .257 r
F P
0 g
s
.7636 .37925 m .75941 .38793 L .76392 .41169 L p .962 .748 .228 r
F P
0 g
s
.76392 .41169 m .76513 .41803 L .7636 .37925 L p .962 .748 .228 r
F P
0 g
s
.76342 .36093 m .76386 .3575 L .7636 .37925 L p .062 .062 .199 r
F P
0 g
s
.35274 .65702 m .35262 .58809 L .3512 .62034 L closepath p .039 .039 .153 r
F P
0 g
s
.35262 .58809 m .35607 .62796 L .35358 .57639 L p .038 .038 .184 r
F P
0 g
s
.35262 .58809 m .353 .54532 L .35171 .59923 L p .016 .016 .046 r
F P
0 g
s
.35262 .58809 m .34777 .50963 L .353 .54532 L closepath p .939 .752 .226 r
F P
0 g
s
.35358 .57639 m .35255 .55509 L .35262 .58809 L p .038 .038 .184 r
F P
0 g
s
.35171 .59923 m .3512 .62034 L .35262 .58809 L p .016 .016 .046 r
F P
0 g
s
.43986 .78567 m .39105 .72085 L .43791 .74302 L closepath p 1 .827 .807 r
F P
0 g
s
.4945 .79471 m .43986 .78567 L .43791 .74302 L closepath p 1 .778 .827 r
F P
0 g
s
.74345 .49658 m .75872 .46172 L .76135 .50089 L closepath p .203 .18 .623 r
F P
0 g
s
.74345 .49658 m .76135 .50089 L .74519 .53422 L closepath p .203 .18 .624 r
F P
0 g
s
.54075 .82166 m .4945 .79471 L .53835 .79825 L closepath p .983 .759 .835 r
F P
0 g
s
.57351 .81896 m .54075 .82166 L .53835 .79825 L closepath p .935 .723 .842 r
F P
0 g
s
.57351 .81896 m .53835 .79825 L .57062 .79475 L closepath p .923 .735 .871 r
F P
0 g
s
.70215 .62091 m .72624 .60495 L .73287 .64517 L closepath p .724 .575 .842 r
F P
0 g
s
.70215 .62091 m .73287 .64517 L .70719 .66184 L closepath p .721 .575 .844 r
F P
0 g
s
.63765 .73314 m .68089 .71651 L .6861 .76104 L closepath p .801 .634 .854 r
F P
0 g
s
.63765 .73314 m .6861 .76104 L .64181 .77813 L closepath p .802 .633 .853 r
F P
0 g
s
.59527 .81422 m .57351 .81896 L .57062 .79475 L closepath p .879 .702 .873 r
F P
0 g
s
.59527 .81422 m .57062 .79475 L .59164 .78926 L closepath p .875 .703 .877 r
F P
0 g
s
.34777 .50963 m .35255 .55509 L .35271 .55166 L p .143 .143 .53 r
F P
0 g
s
.34298 .46876 m .353 .54532 L .34777 .50963 L p .751 .542 .013 r
F P
0 g
s
.35271 .55166 m .35623 .47557 L .34777 .50963 L p .143 .143 .53 r
F P
0 g
s
.34777 .50963 m .34286 .46786 L .34298 .46876 L p .751 .542 .013 r
F P
0 g
s
.67594 .67479 m .68955 .66912 L .69412 .71045 L closepath p .783 .615 .846 r
F P
0 g
s
.67594 .67479 m .69412 .71045 L .68089 .71651 L closepath p .776 .612 .848 r
F P
0 g
s
.49696 .81691 m .43986 .78567 L .4945 .79471 L closepath p 1 .768 .813 r
F P
0 g
s
.54075 .82166 m .49696 .81691 L .4945 .79471 L closepath p .987 .752 .819 r
F P
0 g
s
.6134 .81123 m .59527 .81422 L .59164 .78926 L closepath p .893 .715 .877 r
F P
0 g
s
.6134 .81123 m .59164 .78926 L .60823 .7858 L closepath p .875 .715 .894 r
F P
0 g
s
.75948 .42605 m .76513 .41803 L .75872 .46172 L closepath p .063 .063 .194 r
F P
0 g
s
.68955 .66912 m .70719 .66184 L .71258 .70206 L closepath p .777 .61 .844 r
F P
0 g
s
.68955 .66912 m .71258 .70206 L .69412 .71045 L closepath p .774 .609 .845 r
F P
0 g
s
.54346 .83543 m .49696 .81691 L .54075 .82166 L closepath p .989 .762 .832 r
F P
0 g
s
.57618 .83293 m .54346 .83543 L .54075 .82166 L closepath p .945 .734 .845 r
F P
0 g
s
.57618 .83293 m .54075 .82166 L .57351 .81896 L closepath p .941 .75 .873 r
F P
0 g
s
.59765 .83017 m .57618 .83293 L .57351 .81896 L closepath p .894 .719 .88 r
F P
0 g
s
.59765 .83017 m .57351 .81896 L .59527 .81422 L closepath p .887 .727 .899 r
F P
0 g
s
.64683 .80376 m .6134 .81123 L .60823 .7858 L closepath p .874 .715 .894 r
F P
0 g
s
.64683 .80376 m .60823 .7858 L .64181 .77813 L closepath p .876 .713 .891 r
F P
0 g
s
.34182 .40251 m .34832 .33501 L .33831 .3697 L closepath p .151 .151 .553 r
F P
0 g
s
.34777 .50963 m .3469 .43808 L .34286 .46786 L closepath p .096 .096 .389 r
F P
0 g
s
.34286 .46786 m .3469 .43808 L .34182 .40251 L closepath p .108 .108 .453 r
F P
0 g
s
.60416 .86177 m .58111 .86619 L .57618 .83293 L closepath p .912 .773 .933 r
F P
0 g
s
.60416 .86177 m .57618 .83293 L .59765 .83017 L closepath p .906 .772 .939 r
F P
0 g
s
.58111 .86619 m .54346 .83543 L .57618 .83293 L closepath p .96 .806 .928 r
F P
0 g
s
.61612 .82659 m .59765 .83017 L .59527 .81422 L closepath p .888 .728 .899 r
F P
0 g
s
.61612 .82659 m .59527 .81422 L .6134 .81123 L closepath p .895 .724 .885 r
F P
0 g
s
.54627 .86802 m .49719 .83482 L .54346 .83543 L closepath p .988 .815 .908 r
F P
0 g
s
.58111 .86619 m .54627 .86802 L .54346 .83543 L closepath p .971 .803 .911 r
F P
0 g
s
.54346 .83543 m .49719 .83482 L .49696 .81691 L closepath p .984 .782 .868 r
F P
0 g
s
.49719 .83482 m .44031 .80779 L .49696 .81691 L closepath p 1 .804 .865 r
F P
0 g
s
.49696 .81691 m .44031 .80779 L .43986 .78567 L closepath p 1 .772 .821 r
F P
0 g
s
.72624 .60495 m .74886 .58054 L .75303 .62034 L closepath p .546 .424 .777 r
F P
0 g
s
.72624 .60495 m .75303 .62034 L .73287 .64517 L closepath p .523 .432 .804 r
F P
0 g
s
.44031 .80779 m .39126 .7648 L .43986 .78567 L closepath p 1 .826 .803 r
F P
0 g
s
.43986 .78567 m .39126 .7648 L .39105 .72085 L closepath p 1 .827 .806 r
F P
0 g
s
.39126 .7648 m .3545 .69261 L .39105 .72085 L closepath p 1 .874 .773 r
F P
0 g
s
.39126 .7648 m .35162 .7386 L .3545 .69261 L closepath p 1 .877 .799 r
F P
0 g
s
.3545 .69261 m .35162 .7386 L .3527 .72303 L p .009 .009 .009 r
F P
0 g
s
.35625 .66288 m .39105 .72085 L .3545 .69261 L p 1 .839 .704 r
F P
0 g
s
.3527 .72303 m .3569 .66287 L .3545 .69261 L p .009 .009 .009 r
F P
0 g
s
.3545 .69261 m .35274 .65702 L .35625 .66288 L p 1 .839 .704 r
F P
0 g
s
.74519 .53422 m .76135 .50089 L .76604 .54801 L closepath p .263 .214 .68 r
F P
0 g
s
.74519 .53422 m .76604 .54801 L .74886 .58054 L closepath p .269 .213 .676 r
F P
0 g
s
.70719 .66184 m .73287 .64517 L .73814 .68685 L closepath p .712 .559 .831 r
F P
0 g
s
.70719 .66184 m .73814 .68685 L .71258 .70206 L closepath p .707 .559 .835 r
F P
0 g
s
.64984 .81919 m .61612 .82659 L .6134 .81123 L closepath p .876 .711 .887 r
F P
0 g
s
.64984 .81919 m .6134 .81123 L .64683 .80376 L closepath p .875 .717 .897 r
F P
0 g
s
.61748 .92428 m .5921 .92744 L .58111 .86619 L closepath p .936 .788 .929 r
F P
0 g
s
.61748 .92428 m .58111 .86619 L .60416 .86177 L closepath p .914 .781 .942 r
F P
0 g
s
.5921 .92744 m .54627 .86802 L .58111 .86619 L closepath p .973 .813 .922 r
F P
0 g
s
.62314 .85711 m .60416 .86177 L .59765 .83017 L closepath p .888 .761 .94 r
F P
0 g
s
.62314 .85711 m .59765 .83017 L .61612 .82659 L closepath p .897 .762 .933 r
F P
0 g
s
.64181 .77813 m .6861 .76104 L .69018 .78706 L closepath p .822 .679 .895 r
F P
0 g
s
.64181 .77813 m .69018 .78706 L .64683 .80376 L closepath p .822 .677 .893 r
F P
0 g
s
.3545 .69261 m .35607 .62796 L .35274 .65702 L closepath p .069 .069 .254 r
F P
0 g
s
.35274 .65702 m .35607 .62796 L .35262 .58809 L closepath p .087 .087 .343 r
F P
0 g
s
.68089 .71651 m .69412 .71045 L .70092 .75449 L closepath p .78 .621 .856 r
F P
0 g
s
.68089 .71651 m .70092 .75449 L .6861 .76104 L closepath p .791 .626 .853 r
F P
0 g
s
.5525 .92869 m .49738 .86881 L .54627 .86802 L closepath p .985 .84 .944 r
F P
0 g
s
.5921 .92744 m .5525 .92869 L .54627 .86802 L closepath p .941 .808 .949 r
F P
0 g
s
.54627 .86802 m .49738 .86881 L .49719 .83482 L closepath p .983 .818 .917 r
F P
0 g
s
.49738 .86881 m .44003 .82866 L .49719 .83482 L closepath p 1 .839 .915 r
F P
0 g
s
.49719 .83482 m .44003 .82866 L .44031 .80779 L closepath p 1 .807 .872 r
F P
0 g
s
.63993 .91927 m .61748 .92428 L .60416 .86177 L closepath p .889 .765 .945 r
F P
0 g
s
.63993 .91927 m .60416 .86177 L .62314 .85711 L closepath p .89 .765 .944 r
F P
0 g
s
.66101 .84545 m .62314 .85711 L .61612 .82659 L closepath p .869 .743 .936 r
F P
0 g
s
.66101 .84545 m .61612 .82659 L .64984 .81919 L closepath p .882 .728 .905 r
F P
0 g
s
.55738 1.00036 m .48938 .92357 L .5525 .92869 L closepath p .992 .834 .928 r
F P
0 g
s
.60662 .99713 m .55738 1.00036 L .5525 .92869 L closepath p .948 .801 .933 r
F P
0 g
s
.60662 .99713 m .5525 .92869 L .5921 .92744 L closepath p .94 .799 .939 r
F P
0 g
s
.5525 .92869 m .48938 .92357 L .49738 .86881 L closepath p .992 .84 .936 r
F P
0 g
s
.6343 .99293 m .60662 .99713 L .5921 .92744 L closepath p .93 .793 .941 r
F P
0 g
s
.6343 .99293 m .5921 .92744 L .61748 .92428 L closepath p .937 .795 .937 r
F P
0 g
s
.69412 .71045 m .71258 .70206 L .71981 .74726 L closepath p .782 .624 .859 r
F P
0 g
s
.69412 .71045 m .71981 .74726 L .70092 .75449 L closepath p .786 .626 .857 r
F P
0 g
s
.64683 .80376 m .69018 .78706 L .69264 .80265 L closepath p .82 .672 .888 r
F P
0 g
s
.64683 .80376 m .69264 .80265 L .64984 .81919 L closepath p .821 .681 .899 r
F P
0 g
s
.48938 .92357 m .43797 .85676 L .49738 .86881 L closepath p 1 .867 .938 r
F P
0 g
s
.49738 .86881 m .43797 .85676 L .44003 .82866 L closepath p 1 .826 .877 r
F P
0 g
s
.68447 .90821 m .63993 .91927 L .62314 .85711 L closepath p .882 .76 .945 r
F P
0 g
s
.68447 .90821 m .62314 .85711 L .66101 .84545 L closepath p .874 .76 .952 r
F P
0 g
s
.44003 .82866 m .39229 .78733 L .44031 .80779 L closepath p 1 .856 .861 r
F P
0 g
s
.44031 .80779 m .39229 .78733 L .39126 .7648 L closepath p 1 .831 .819 r
F P
0 g
s
.35262 .58809 m .35255 .55509 L .34777 .50963 L closepath p .046 .046 .223 r
F P
0 g
s
.75872 .46172 m .76513 .41803 L .76726 .46256 L closepath p .096 .096 .348 r
F P
0 g
s
.75872 .46172 m .76726 .46256 L .76135 .50089 L closepath p .096 .096 .357 r
F P
0 g
s
.65563 .98828 m .6343 .99293 L .61748 .92428 L closepath p .889 .764 .944 r
F P
0 g
s
.65563 .98828 m .61748 .92428 L .63993 .91927 L closepath p .889 .764 .944 r
F P
0 g
s
.3469 .43808 m .3536 .36863 L .34182 .40251 L closepath p .175 .175 .658 r
F P
0 g
s
.34182 .40251 m .3536 .36863 L .34832 .33501 L closepath p .173 .173 .645 r
F P
0 g
s
.61603 1.06052 m .55778 1.06362 L .55738 1.00036 L closepath p .951 .776 .898 \
r
F P
0 g
s
.61603 1.06052 m .55738 1.00036 L .60662 .99713 L closepath p .945 .777 .905 \
r
F P
0 g
s
.55778 1.06362 m .48371 .99758 L .55738 1.00036 L closepath p 1 .821 .895 r
F P
0 g
s
.55738 1.00036 m .48371 .99758 L .48938 .92357 L closepath p 1 .833 .911 r
F P
0 g
s
.64984 .81919 m .69264 .80265 L .70084 .83311 L closepath p .839 .724 .938 r
F P
0 g
s
.64984 .81919 m .70084 .83311 L .66101 .84545 L closepath p .846 .71 .915 r
F P
0 g
s
.64574 1.05642 m .61603 1.06052 L .60662 .99713 L closepath p .919 .759 .908 \
r
F P
0 g
s
.64574 1.05642 m .60662 .99713 L .6343 .99293 L closepath p .923 .76 .906 r
F P
0 g
s
.48371 .99758 m .43022 .90755 L .48938 .92357 L closepath p 1 .875 .907 r
F P
0 g
s
.48938 .92357 m .43022 .90755 L .43797 .85676 L closepath p 1 .867 .894 r
F P
0 g
s
.76386 .3575 m .75914 .33226 L .75996 .35423 L closepath p .949 .733 .207 r
F P
0 g
s
.76386 .3575 m .75996 .35423 L .7636 .37925 L closepath p .962 .775 .259 r
F P
0 g
s
.70095 .97753 m .65563 .98828 L .63993 .91927 L closepath p .883 .76 .945 r
F P
0 g
s
.70095 .97753 m .63993 .91927 L .68447 .90821 L closepath p .882 .76 .946 r
F P
0 g
s
.73287 .64517 m .75303 .62034 L .75806 .66508 L closepath p .502 .404 .782 r
F P
0 g
s
.73287 .64517 m .75806 .66508 L .73814 .68685 L closepath p .5 .404 .783 r
F P
0 g
s
.43797 .85676 m .39238 .80849 L .44003 .82866 L closepath p 1 .873 .869 r
F P
0 g
s
.44003 .82866 m .39238 .80849 L .39229 .78733 L closepath p 1 .836 .805 r
F P
0 g
s
.6674 1.05117 m .64574 1.05642 L .6343 .99293 L closepath p .861 .718 .911 r
F P
0 g
s
.6674 1.05117 m .6343 .99293 L .65563 .98828 L closepath p .878 .725 .904 r
F P
0 g
s
.71258 .70206 m .73814 .68685 L .74261 .73306 L closepath p .712 .568 .843 r
F P
0 g
s
.71258 .70206 m .74261 .73306 L .71981 .74726 L closepath p .7 .566 .85 r
F P
0 g
s
.34777 .50963 m .35623 .47557 L .3469 .43808 L closepath p .174 .19 .699 r
F P
0 g
s
.6861 .76104 m .70092 .75449 L .70486 .7797 L closepath p .803 .652 .877 r
F P
0 g
s
.6861 .76104 m .70486 .7797 L .69018 .78706 L closepath p .781 .649 .893 r
F P
0 g
s
.39229 .78733 m .35162 .7386 L .39126 .7648 L closepath p 1 .871 .788 r
F P
0 g
s
.66101 .84545 m .70084 .83311 L .71768 .89588 L closepath p .857 .745 .948 r
F P
0 g
s
.66101 .84545 m .71768 .89588 L .68447 .90821 L closepath p .846 .744 .957 r
F P
0 g
s
.74886 .58054 m .76604 .54801 L .77046 .58869 L closepath p .279 .225 .685 r
F P
0 g
s
.74886 .58054 m .77046 .58869 L .75303 .62034 L closepath p .279 .225 .686 r
F P
0 g
s
.71157 1.04146 m .6674 1.05117 L .65563 .98828 L closepath p .874 .722 .904 r
F P
0 g
s
.71157 1.04146 m .65563 .98828 L .70095 .97753 L closepath p .872 .722 .906 r
F P
0 g
s
.3545 .69261 m .3569 .66287 L .35607 .62796 L closepath p .05 .05 .179 r
F P
0 g
s
.43022 .90755 m .39183 .8359 L .43797 .85676 L closepath p 1 .901 .89 r
F P
0 g
s
.43797 .85676 m .39183 .8359 L .39238 .80849 L closepath p 1 .858 .814 r
F P
0 g
s
.3536 .36863 m .35664 .31222 L .34832 .33501 L closepath p .174 .174 .649 r
F P
0 g
s
.70092 .75449 m .71981 .74726 L .72385 .77156 L closepath p .793 .641 .871 r
F P
0 g
s
.70092 .75449 m .72385 .77156 L .70486 .7797 L closepath p .789 .642 .876 r
F P
0 g
s
.69018 .78706 m .70486 .7797 L .70839 .7955 L closepath p .785 .658 .9 r
F P
0 g
s
.69018 .78706 m .70839 .7955 L .69264 .80265 L closepath p .792 .652 .886 r
F P
0 g
s
.7636 .37925 m .75996 .35423 L .75941 .38793 L closepath p .949 .76 .238 r
F P
0 g
s
.61817 1.12525 m .55169 1.12278 L .55778 1.06362 L closepath p .946 .735 .845 \
r
F P
0 g
s
.61817 1.12525 m .55778 1.06362 L .61603 1.06052 L closepath p .943 .735 .85 \
r
F P
0 g
s
.55169 1.12278 m .48646 1.05503 L .55778 1.06362 L closepath p 1 .796 .849 r
F P
0 g
s
.55778 1.06362 m .48646 1.05503 L .48371 .99758 L closepath p 1 .816 .876 r
F P
0 g
s
.64959 1.12345 m .61817 1.12525 L .61603 1.06052 L closepath p .901 .702 .85 \
r
F P
0 g
s
.64959 1.12345 m .61603 1.06052 L .64574 1.05642 L closepath p .905 .704 .848 \
r
F P
0 g
s
.68447 .90821 m .71768 .89588 L .73866 .96404 L closepath p .842 .735 .949 r
F P
0 g
s
.68447 .90821 m .73866 .96404 L .70095 .97753 L closepath p .843 .735 .948 r
F P
0 g
s
.48646 1.05503 m .42323 .97943 L .48371 .99758 L closepath p 1 .859 .859 r
F P
0 g
s
.48371 .99758 m .42323 .97943 L .43022 .90755 L closepath p 1 .877 .887 r
F P
0 g
s
.39238 .80849 m .34945 .76015 L .39229 .78733 L closepath p 1 .865 .786 r
F P
0 g
s
.39229 .78733 m .34945 .76015 L .35162 .7386 L closepath p 1 .879 .814 r
F P
0 g
s
.35623 .47557 m .36104 .40961 L .3469 .43808 L closepath p .28 .271 .743 r
F P
0 g
s
.3469 .43808 m .36104 .40961 L .3536 .36863 L closepath p .238 .222 .703 r
F P
0 g
s
.67356 1.11916 m .64959 1.12345 L .64574 1.05642 L closepath p .861 .67 .846 \
r
F P
0 g
s
.67356 1.11916 m .64574 1.05642 L .6674 1.05117 L closepath p .842 .661 .853 \
r
F P
0 g
s
.69264 .80265 m .70839 .7955 L .71815 .82655 L closepath p .823 .719 .946 r
F P
0 g
s
.69264 .80265 m .71815 .82655 L .70084 .83311 L closepath p .833 .72 .939 r
F P
0 g
s
.3569 .66287 m .3558 .59288 L .35607 .62796 L closepath p .008 .008 .042 r
F P
0 g
s
.35362 .57723 m .35607 .62796 L .3558 .59288 L p .039 .039 .19 r
F P
0 g
s
.35367 .62745 m .3558 .59288 L .35614 .61463 L p .979 .817 .313 r
F P
0 g
s
.35614 .61463 m .3569 .66287 L .35367 .62745 L p .979 .817 .313 r
F P
0 g
s
.35215 .70922 m .3435 .66823 L .35367 .62745 L closepath p 1 1 .77 r
F P
0 g
s
.35367 .62745 m .3435 .66823 L .35095 .6416 L p .035 .035 .035 r
F P
0 g
s
.35215 .70922 m .35367 .62745 L .3569 .66287 L closepath p 1 .868 .39 r
F P
0 g
s
.35095 .6416 m .36519 .59066 L .35367 .62745 L p .035 .035 .035 r
F P
0 g
s
.3558 .59288 m .35367 .62745 L .35774 .60679 L p .124 .124 .379 r
F P
0 g
s
.35774 .60679 m .36668 .56142 L .3558 .59288 L p .124 .124 .379 r
F P
0 g
s
.3558 .59288 m .35255 .55509 L .35362 .57723 L p .039 .039 .19 r
F P
0 g
s
.76135 .50089 m .76726 .46256 L .77125 .50931 L closepath p .105 .105 .402 r
F P
0 g
s
.76135 .50089 m .77125 .50931 L .76604 .54801 L closepath p .103 .103 .41 r
F P
0 g
s
.70095 .97753 m .73866 .96404 L .75154 1.02727 L closepath p .825 .689 .906 r
F P
0 g
s
.70095 .97753 m .75154 1.02727 L .71157 1.04146 L closepath p .825 .689 .906 \
r
F P
0 g
s
.73814 .68685 m .75806 .66508 L .76219 .70871 L closepath p .484 .382 .765 r
F P
0 g
s
.73814 .68685 m .76219 .70871 L .74261 .73306 L closepath p .462 .379 .776 r
F P
0 g
s
.70486 .7797 m .72385 .77156 L .72571 .78605 L closepath p .792 .647 .88 r
F P
0 g
s
.70486 .7797 m .72571 .78605 L .70839 .7955 L closepath p .786 .658 .9 r
F P
0 g
s
.71981 .74726 m .74261 .73306 L .74489 .75711 L closepath p .683 .536 .824 r
F P
0 g
s
.71981 .74726 m .74489 .75711 L .72385 .77156 L closepath p .668 .552 .857 r
F P
0 g
s
.70084 .83311 m .71815 .82655 L .73804 .88799 L closepath p .837 .732 .949 r
F P
0 g
s
.70084 .83311 m .73804 .88799 L .71768 .89588 L closepath p .838 .732 .949 r
F P
0 g
s
.75914 .33226 m .71672 .31179 L .71669 .33597 L closepath p 1 .795 .776 r
F P
0 g
s
.75914 .33226 m .71669 .33597 L .75996 .35423 L closepath p 1 .789 .769 r
F P
0 g
s
.71883 1.10917 m .67356 1.11916 L .6674 1.05117 L closepath p .851 .668 .853 \
r
F P
0 g
s
.71883 1.10917 m .6674 1.05117 L .71157 1.04146 L closepath p .856 .668 .849 \
r
F P
0 g
s
.42323 .97943 m .38681 .88552 L .43022 .90755 L closepath p 1 .908 .878 r
F P
0 g
s
.43022 .90755 m .38681 .88552 L .39183 .8359 L closepath p 1 .902 .867 r
F P
0 g
s
.70839 .7955 m .72571 .78605 L .73696 .81799 L closepath p .81 .707 .941 r
F P
0 g
s
.70839 .7955 m .73696 .81799 L .71815 .82655 L closepath p .803 .707 .947 r
F P
0 g
s
.39183 .8359 m .34728 .78381 L .39238 .80849 L closepath p 1 .868 .809 r
F P
0 g
s
.39238 .80849 m .34728 .78381 L .34945 .76015 L closepath p 1 .877 .827 r
F P
0 g
s
.75303 .62034 m .77046 .58869 L .77537 .63171 L closepath p .277 .223 .684 r
F P
0 g
s
.75303 .62034 m .77537 .63171 L .75806 .66508 L closepath p .267 .228 .696 r
F P
0 g
s
.71768 .89588 m .73804 .88799 L .75808 .95677 L closepath p .838 .732 .949 r
F P
0 g
s
.71768 .89588 m .75808 .95677 L .73866 .96404 L closepath p .838 .732 .949 r
F P
0 g
s
.71815 .82655 m .73696 .81799 L .75607 .878 L closepath p .8 .7 .941 r
F P
0 g
s
.71815 .82655 m .75607 .878 L .73804 .88799 L closepath p .765 .687 .953 r
F P
0 g
s
.3558 .59288 m .36313 .52439 L .35255 .55509 L closepath p .157 .157 .569 r
F P
0 g
s
.35255 .55509 m .36313 .52439 L .35623 .47557 L closepath p .168 .168 .634 r
F P
0 g
s
.72385 .77156 m .74489 .75711 L .74563 .77186 L closepath p .644 .512 .823 r
F P
0 g
s
.72385 .77156 m .74563 .77186 L .72571 .78605 L closepath p .645 .539 .857 r
F P
0 g
s
.36104 .40961 m .35879 .34766 L .3536 .36863 L closepath p .161 .161 .634 r
F P
0 g
s
.3536 .36863 m .35879 .34766 L .35664 .31222 L closepath p .131 .131 .477 r
F P
0 g
s
.49818 1.11487 m .4243 1.03537 L .48646 1.05503 L closepath p 1 .844 .826 r
F P
0 g
s
.55169 1.12278 m .49818 1.11487 L .48646 1.05503 L closepath p 1 .799 .866 r
F P
0 g
s
.48646 1.05503 m .4243 1.03537 L .42323 .97943 L closepath p 1 .857 .845 r
F P
0 g
s
.76513 .41803 m .75941 .38793 L .76099 .43084 L closepath p .946 .729 .202 r
F P
0 g
s
.76513 .41803 m .76099 .43084 L .76726 .46256 L closepath p .946 .729 .202 r
F P
0 g
s
.34945 .76015 m .35215 .70922 L .35162 .7386 L closepath p .851 .744 .206 r
F P
0 g
s
.35162 .7386 m .35215 .70922 L .3569 .66287 L closepath p .852 .744 .207 r
F P
0 g
s
.35879 .34766 m .35921 .29463 L .35664 .31222 L closepath p .096 .096 .362 r
F P
0 g
s
.72571 .78605 m .74563 .77186 L .75625 .80284 L closepath p .691 .616 .921 r
F P
0 g
s
.72571 .78605 m .75625 .80284 L .73696 .81799 L closepath p .67 .622 .941 r
F P
0 g
s
.71157 1.04146 m .75154 1.02727 L .7605 1.09518 L closepath p .8 .631 .85 r
F P
0 g
s
.71157 1.04146 m .7605 1.09518 L .71883 1.10917 L closepath p .805 .631 .847 \
r
F P
0 g
s
.75996 .35423 m .71669 .33597 L .71648 .37088 L closepath p 1 .794 .777 r
F P
0 g
s
.75996 .35423 m .71648 .37088 L .75941 .38793 L closepath p 1 .781 .765 r
F P
0 g
s
.71672 .31179 m .67471 .28272 L .67381 .31144 L closepath p 1 .855 .766 r
F P
0 g
s
.71672 .31179 m .67381 .31144 L .71669 .33597 L closepath p 1 .841 .75 r
F P
0 g
s
.36313 .52439 m .36786 .44641 L .35623 .47557 L closepath p .169 .169 .635 r
F P
0 g
s
.35623 .47557 m .36786 .44641 L .36104 .40961 L closepath p .175 .175 .67 r
F P
0 g
s
.73866 .96404 m .75808 .95677 L .77278 1.02032 L closepath p .823 .692 .912 r
F P
0 g
s
.73866 .96404 m .77278 1.02032 L .75154 1.02727 L closepath p .841 .699 .905 \
r
F P
0 g
s
.73804 .88799 m .75607 .878 L .77894 .9467 L closepath p .766 .689 .954 r
F P
0 g
s
.73804 .88799 m .77894 .9467 L .75808 .95677 L closepath p .778 .694 .951 r
F P
0 g
s
.38681 .88552 m .34558 .81074 L .39183 .8359 L closepath p 1 .895 .869 r
F P
0 g
s
.39183 .8359 m .34558 .81074 L .34728 .78381 L closepath p 1 .872 .826 r
F P
0 g
s
.61688 1.15563 m .54891 1.15311 L .55169 1.12278 L closepath p .937 .69 .788 \
r
F P
0 g
s
.61688 1.15563 m .55169 1.12278 L .61817 1.12525 L closepath p .937 .69 .789 \
r
F P
0 g
s
.54891 1.15311 m .49818 1.11487 L .55169 1.12278 L closepath p .999 .745 .792 \
r
F P
0 g
s
.74261 .73306 m .76219 .70871 L .76337 .73206 L closepath p .411 .314 .723 r
F P
0 g
s
.74261 .73306 m .76337 .73206 L .74489 .75711 L closepath p .414 .334 .747 r
F P
0 g
s
.64842 1.15331 m .61688 1.15563 L .61817 1.12525 L closepath p .886 .647 .785 \
r
F P
0 g
s
.64842 1.15331 m .61817 1.12525 L .64959 1.12345 L closepath p .885 .647 .785 \
r
F P
0 g
s
.76604 .54801 m .77125 .50931 L .77317 .54819 L closepath p .082 .082 .303 r
F P
0 g
s
.76604 .54801 m .77317 .54819 L .77046 .58869 L closepath p .081 .081 .343 r
F P
0 g
s
.73696 .81799 m .75625 .80284 L .78246 .85905 L closepath p .692 .659 .97 r
F P
0 g
s
.73696 .81799 m .78246 .85905 L .75607 .878 L closepath p .751 .669 .941 r
F P
0 g
s
.34945 .76015 m .34697 .73313 L .35215 .70922 L closepath p 1 .96 .534 r
F P
0 g
s
.34697 .73313 m .34945 .76015 L .34728 .78381 L closepath p .978 .835 .335 r
F P
0 g
s
.4243 1.03537 m .3774 .95635 L .42323 .97943 L closepath p 1 .898 .815 r
F P
0 g
s
.42323 .97943 m .3774 .95635 L .38681 .88552 L closepath p 1 .912 .84 r
F P
0 g
s
.67409 1.14936 m .64842 1.15331 L .64959 1.12345 L closepath p .849 .616 .78 \
r
F P
0 g
s
.67409 1.14936 m .64959 1.12345 L .67356 1.11916 L closepath p .842 .616 .788 \
r
F P
0 g
s
.75808 .95677 m .77894 .9467 L .79549 1.00942 L closepath p .758 .651 .913 r
F P
0 g
s
.75808 .95677 m .79549 1.00942 L .77278 1.02032 L closepath p .764 .653 .911 \
r
F P
0 g
s
.75806 .66508 m .77537 .63171 L .7803 .67518 L closepath p .25 .208 .68 r
F P
0 g
s
.75806 .66508 m .7803 .67518 L .76219 .70871 L closepath p .251 .207 .677 r
F P
0 g
s
.51162 1.1488 m .42924 1.08787 L .49818 1.11487 L closepath p 1 .792 .739 r
F P
0 g
s
.54891 1.15311 m .51162 1.1488 L .49818 1.11487 L closepath p .986 .759 .831 \
r
F P
0 g
s
.49818 1.11487 m .42924 1.08787 L .4243 1.03537 L closepath p 1 .843 .821 r
F P
0 g
s
.74489 .75711 m .76337 .73206 L .76763 .74928 L closepath p .493 .435 .827 r
F P
0 g
s
.74489 .75711 m .76763 .74928 L .74563 .77186 L closepath p .47 .376 .767 r
F P
0 g
s
.3558 .59288 m .36668 .56142 L .36313 .52439 L closepath p .168 .168 .615 r
F P
0 g
s
.36122 .3258 m .3893 .2508 L .35921 .29463 L closepath p .335 .223 .649 r
F P
0 g
s
.35879 .34766 m .36122 .3258 L .35921 .29463 L closepath p .073 .073 .274 r
F P
0 g
s
.36556 .39155 m .36122 .3258 L .35879 .34766 L closepath p .114 .114 .473 r
F P
0 g
s
.36104 .40961 m .36556 .39155 L .35879 .34766 L closepath p .138 .138 .546 r
F P
0 g
s
.71669 .33597 m .67381 .31144 L .67281 .34647 L closepath p 1 .854 .772 r
F P
0 g
s
.71669 .33597 m .67281 .34647 L .71648 .37088 L closepath p 1 .837 .756 r
F P
0 g
s
.71974 1.13921 m .67409 1.14936 L .67356 1.11916 L closepath p .827 .604 .787 \
r
F P
0 g
s
.71974 1.13921 m .67356 1.11916 L .71883 1.10917 L closepath p .827 .604 .786 \
r
F P
0 g
s
.75607 .878 m .78246 .85905 L .81142 .92568 L closepath p .755 .676 .947 r
F P
0 g
s
.75607 .878 m .81142 .92568 L .77894 .9467 L closepath p .742 .675 .954 r
F P
0 g
s
.36786 .44641 m .36556 .39155 L .36104 .40961 L closepath p .144 .144 .583 r
F P
0 g
s
.75154 1.02727 m .77278 1.02032 L .78217 1.08685 L closepath p .815 .635 .843 \
r
F P
0 g
s
.75154 1.02727 m .78217 1.08685 L .7605 1.09518 L closepath p .793 .625 .85 r
F P
0 g
s
.3774 .95635 m .33982 .85626 L .38681 .88552 L closepath p 1 .905 .843 r
F P
0 g
s
.38681 .88552 m .33982 .85626 L .34558 .81074 L closepath p 1 .892 .818 r
F P
0 g
s
.74563 .77186 m .76763 .74928 L .78109 .78025 L closepath p .611 .577 .926 r
F P
0 g
s
.74563 .77186 m .78109 .78025 L .75625 .80284 L closepath p .616 .569 .914 r
F P
0 g
s
.34153 .75585 m .34728 .78381 L .34133 .76984 L closepath p 1 .823 .334 r
F P
0 g
s
.34558 .81074 m .34133 .76984 L .34728 .78381 L closepath p 1 .87 .401 r
F P
0 g
s
.34697 .73313 m .34728 .78381 L .34153 .75585 L closepath p 1 1 .666 r
F P
0 g
s
.76726 .46256 m .76099 .43084 L .7628 .47685 L closepath p .952 .736 .212 r
F P
0 g
s
.76726 .46256 m .7628 .47685 L .77125 .50931 L closepath p .969 .726 .209 r
F P
0 g
s
.61255 1.18785 m .53877 1.18904 L .54891 1.15311 L closepath p .923 .655 .752 \
r
F P
0 g
s
.61255 1.18785 m .54891 1.15311 L .61688 1.15563 L closepath p .925 .647 .735 \
r
F P
0 g
s
.54891 1.15311 m .53877 1.18904 L .51162 1.1488 L closepath p .978 .711 .768 \
r
F P
0 g
s
.75941 .38793 m .71648 .37088 L .71715 .41502 L closepath p 1 .783 .768 r
F P
0 g
s
.75941 .38793 m .71715 .41502 L .76099 .43084 L closepath p 1 .775 .762 r
F P
0 g
s
.64563 1.18464 m .61255 1.18785 L .61688 1.15563 L closepath p .876 .602 .726 \
r
F P
0 g
s
.64563 1.18464 m .61688 1.15563 L .64842 1.15331 L closepath p .871 .603 .735 \
r
F P
0 g
s
.77894 .9467 m .81142 .92568 L .83249 .98496 L closepath p .706 .609 .901 r
F P
0 g
s
.77894 .9467 m .83249 .98496 L .79549 1.00942 L closepath p .696 .61 .909 r
F P
0 g
s
.77278 1.02032 m .79549 1.00942 L .80628 1.07317 L closepath p .724 .572 .839 \
r
F P
0 g
s
.77278 1.02032 m .80628 1.07317 L .78217 1.08685 L closepath p .73 .575 .836 \
r
F P
0 g
s
.77046 .58869 m .77317 .54819 L .77528 .59232 L closepath p .056 .056 .22 r
F P
0 g
s
.77046 .58869 m .77528 .59232 L .77537 .63171 L closepath p .053 .053 .255 r
F P
0 g
s
.35215 .70922 m .34179 .68212 L .34205 .67779 L closepath p 1 .914 .519 r
F P
0 g
s
.35215 .70922 m .34205 .67779 L .34383 .67466 L closepath p 1 1 .923 r
F P
0 g
s
.35215 .70922 m .34383 .67466 L .3435 .66823 L closepath p 1 .769 .27 r
F P
0 g
s
.34697 .73313 m .34179 .68212 L .35215 .70922 L closepath p 1 1 .705 r
F P
0 g
s
.53877 1.18904 m .43387 1.12229 L .51162 1.1488 L closepath p 1 .697 .617 r
F P
0 g
s
.51162 1.1488 m .43387 1.12229 L .42924 1.08787 L closepath p 1 .817 .795 r
F P
0 g
s
.36668 .56142 m .37396 .49322 L .36313 .52439 L closepath p .158 .158 .586 r
F P
0 g
s
.36313 .52439 m .37396 .49322 L .36786 .44641 L closepath p .162 .162 .609 r
F P
0 g
s
.75625 .80284 m .78109 .78025 L .81046 .82538 L closepath p .652 .626 .957 r
F P
0 g
s
.75625 .80284 m .81046 .82538 L .78246 .85905 L closepath p .642 .63 .968 r
F P
0 g
s
.67471 .28272 m .62525 .26058 L .6244 .29624 L closepath p 1 .82 .807 r
F P
0 g
s
.67471 .28272 m .6244 .29624 L .67381 .31144 L closepath p 1 .803 .79 r
F P
0 g
s
.42924 1.08787 m .37375 1.00469 L .4243 1.03537 L closepath p 1 .878 .781 r
F P
0 g
s
.4243 1.03537 m .37375 1.00469 L .3774 .95635 L closepath p 1 .899 .822 r
F P
0 g
s
.67247 1.1802 m .64563 1.18464 L .64842 1.15331 L closepath p .825 .564 .726 \
r
F P
0 g
s
.67247 1.1802 m .64842 1.15331 L .67409 1.14936 L closepath p .828 .564 .723 \
r
F P
0 g
s
.71883 1.10917 m .7605 1.09518 L .76077 1.12577 L closepath p .774 .562 .779 \
r
F P
0 g
s
.71883 1.10917 m .76077 1.12577 L .71974 1.13921 L closepath p .773 .562 .779 \
r
F P
0 g
s
.34558 .81074 m .33706 .78194 L .34133 .76984 L closepath p 1 1 .783 r
F P
0 g
s
.33706 .78194 m .34558 .81074 L .33982 .85626 L closepath p 1 .961 .584 r
F P
0 g
s
.35367 .62745 m .36519 .59066 L .36668 .56142 L closepath p .123 .123 .375 r
F P
0 g
s
.76219 .70871 m .7803 .67518 L .78266 .69761 L closepath p .236 .189 .663 r
F P
0 g
s
.76219 .70871 m .78266 .69761 L .76337 .73206 L closepath p .228 .184 .646 r
F P
0 g
s
.71708 1.17034 m .67247 1.1802 L .67409 1.14936 L closepath p .801 .542 .719 \
r
F P
0 g
s
.71708 1.17034 m .67409 1.14936 L .71974 1.13921 L closepath p .801 .544 .723 \
r
F P
0 g
s
.37396 .49322 m .3688 .42924 L .36786 .44641 L closepath p .07 .07 .32 r
F P
0 g
s
.36786 .44641 m .3688 .42924 L .36556 .39155 L closepath p .054 .054 .238 r
F P
0 g
s
.36556 .39155 m .36345 .3724 L .36122 .3258 L closepath p .794 .614 .045 r
F P
0 g
s
.36345 .3724 m .3889 .28352 L .36122 .3258 L closepath p .329 .23 .663 r
F P
0 g
s
.36122 .3258 m .3889 .28352 L .3893 .2508 L closepath p .252 .185 .586 r
F P
0 g
s
.71648 .37088 m .67281 .34647 L .67139 .39232 L closepath p 1 .859 .794 r
F P
0 g
s
.71648 .37088 m .67139 .39232 L .71715 .41502 L closepath p 1 .811 .754 r
F P
0 g
s
.78246 .85905 m .81046 .82538 L .84176 .89126 L closepath p .588 .556 .915 r
F P
0 g
s
.78246 .85905 m .84176 .89126 L .81142 .92568 L closepath p .557 .561 .938 r
F P
0 g
s
.34153 .75585 m .32764 .70106 L .34697 .73313 L closepath p 1 .98 .824 r
F P
0 g
s
.34697 .73313 m .32764 .70106 L .34179 .68212 L closepath p 1 1 .985 r
F P
0 g
s
.3688 .42924 m .36345 .3724 L .36556 .39155 L closepath p .741 .553 .01 r
F P
0 g
s
.79549 1.00942 m .83249 .98496 L .84658 1.0404 L closepath p .633 .502 .817 r
F P
0 g
s
.79549 1.00942 m .84658 1.0404 L .80628 1.07317 L closepath p .626 .504 .824 \
r
F P
0 g
s
.76337 .73206 m .78266 .69761 L .78835 .71746 L closepath p .409 .388 .817 r
F P
0 g
s
.76337 .73206 m .78835 .71746 L .76763 .74928 L closepath p .401 .371 .802 r
F P
0 g
s
.36519 .59066 m .3733 .52715 L .36668 .56142 L closepath p .065 .065 .179 r
F P
0 g
s
.36668 .56142 m .3733 .52715 L .37396 .49322 L closepath p .08 .08 .253 r
F P
0 g
s
.67381 .31144 m .6244 .29624 L .62543 .33463 L closepath p 1 .825 .823 r
F P
0 g
s
.67381 .31144 m .62543 .33463 L .67281 .34647 L closepath p 1 .794 .797 r
F P
0 g
s
.77125 .50931 m .7628 .47685 L .7617 .51634 L closepath p 1 .855 .399 r
F P
0 g
s
.77125 .50931 m .7617 .51634 L .77317 .54819 L closepath p 1 .824 .371 r
F P
0 g
s
.7605 1.09518 m .78217 1.08685 L .78317 1.1166 L closepath p .75 .536 .764 r
F P
0 g
s
.7605 1.09518 m .78317 1.1166 L .76077 1.12577 L closepath p .741 .537 .773 r
F P
0 g
s
.62525 .26058 m .57508 .24791 L .57354 .28512 L closepath p 1 .79 .831 r
F P
0 g
s
.62525 .26058 m .57354 .28512 L .6244 .29624 L closepath p 1 .773 .817 r
F P
0 g
s
.37375 1.00469 m .33732 .93192 L .3774 .95635 L closepath p 1 .919 .807 r
F P
0 g
s
.3774 .95635 m .33732 .93192 L .33982 .85626 L closepath p 1 .906 .784 r
F P
0 g
s
.43387 1.12229 m .37759 1.0557 L .42924 1.08787 L closepath p 1 .857 .736 r
F P
0 g
s
.42924 1.08787 m .37759 1.0557 L .37375 1.00469 L closepath p 1 .874 .766 r
F P
0 g
s
.71974 1.13921 m .76077 1.12577 L .75733 1.15742 L closepath p .735 .486 .703 \
r
F P
0 g
s
.71974 1.13921 m .75733 1.15742 L .71708 1.17034 L closepath p .734 .488 .706 \
r
F P
0 g
s
.60294 1.21696 m .51914 1.21708 L .53877 1.18904 L closepath p .896 .537 .584 \
r
F P
0 g
s
.60294 1.21696 m .53877 1.18904 L .61255 1.18785 L closepath p .899 .587 .671 \
r
F P
0 g
s
.51914 1.21708 m .44711 1.16864 L .53877 1.18904 L closepath p 1 .71 .678 r
F P
0 g
s
.53877 1.18904 m .44711 1.16864 L .43387 1.12229 L closepath p 1 .775 .768 r
F P
0 g
s
.63975 1.21297 m .60294 1.21696 L .61255 1.18785 L closepath p .84 .527 .648 \
r
F P
0 g
s
.63975 1.21297 m .61255 1.18785 L .64563 1.18464 L closepath p .844 .525 .638 \
r
F P
0 g
s
.78039 .63676 m .77537 .63171 L .77529 .59546 L p .053 .053 .253 r
F P
0 g
s
.77537 .63171 m .78039 .63676 L .7803 .67518 L closepath p .053 .053 .254 r
F P
0 g
s
.7803 .67518 m .78039 .63676 L .7837 .65964 L closepath p .067 .067 .323 r
F P
0 g
s
.78039 .63676 m .77116 .62042 L .7829 .65713 L p .956 .688 .165 r
F P
0 g
s
.7829 .65713 m .7837 .65964 L .78039 .63676 L p .956 .688 .165 r
F P
0 g
s
.77529 .59546 m .77528 .59232 L .78039 .63676 L p .053 .053 .253 r
F P
0 g
s
.34133 .76984 m .31995 .73123 L .34153 .75585 L closepath p 1 .928 .718 r
F P
0 g
s
.34153 .75585 m .31995 .73123 L .32764 .70106 L closepath p 1 .98 .839 r
F P
0 g
s
.76763 .74928 m .78835 .71746 L .79719 .74328 L closepath p .434 .413 .834 r
F P
0 g
s
.76763 .74928 m .79719 .74328 L .78109 .78025 L closepath p .449 .479 .902 r
F P
0 g
s
.76099 .43084 m .71715 .41502 L .71828 .45918 L closepath p 1 .779 .77 r
F P
0 g
s
.76099 .43084 m .71828 .45918 L .7628 .47685 L closepath p 1 .795 .781 r
F P
0 g
s
.34383 .67466 m .33806 .62678 L .3435 .66823 L closepath p .882 .678 .129 r
F P
0 g
s
.3435 .66823 m .33806 .62678 L .36519 .59066 L closepath p 1 1 .984 r
F P
0 g
s
.66523 1.20855 m .63975 1.21297 L .64563 1.18464 L closepath p .785 .471 .621 \
r
F P
0 g
s
.66523 1.20855 m .64563 1.18464 L .67247 1.1802 L closepath p .783 .471 .623 \
r
F P
0 g
s
.81142 .92568 m .84176 .89126 L .85707 .94761 L closepath p .469 .436 .843 r
F P
0 g
s
.81142 .92568 m .85707 .94761 L .83249 .98496 L closepath p .444 .445 .866 r
F P
0 g
s
.32473 .85663 m .33982 .85626 L .32395 .87512 L closepath p 1 .932 .588 r
F P
0 g
s
.33732 .93192 m .32395 .87512 L .33982 .85626 L closepath p 1 .936 .59 r
F P
0 g
s
.3213 .82702 m .33982 .85626 L .32473 .85663 L closepath p 1 .845 .493 r
F P
0 g
s
.33706 .78194 m .33982 .85626 L .3213 .82702 L closepath p 1 1 .856 r
F P
0 g
s
.34205 .67779 m .3368 .63736 L .34383 .67466 L closepath p 1 1 .921 r
F P
0 g
s
.34383 .67466 m .3368 .63736 L .33806 .62678 L closepath p 1 .893 .452 r
F P
0 g
s
.34179 .68212 m .32964 .64358 L .34205 .67779 L closepath p 1 .925 .558 r
F P
0 g
s
.34205 .67779 m .32964 .64358 L .3368 .63736 L closepath p 1 1 1 r
F P
0 g
s
.33706 .78194 m .31782 .74285 L .34133 .76984 L closepath p 1 .975 .855 r
F P
0 g
s
.34133 .76984 m .31782 .74285 L .31995 .73123 L closepath p 1 .952 .794 r
F P
0 g
s
.78217 1.08685 m .80628 1.07317 L .80731 1.1036 L closepath p .673 .476 .748 \
r
F P
0 g
s
.78217 1.08685 m .80731 1.1036 L .78317 1.1166 L closepath p .672 .476 .749 r
F P
0 g
s
.3889 .28352 m .39491 .24588 L .3893 .2508 L closepath p .539 .363 .699 r
F P
0 g
s
.34179 .68212 m .3274 .65069 L .32964 .64358 L closepath p 1 1 .73 r
F P
0 g
s
.32764 .70106 m .3274 .65069 L .34179 .68212 L closepath p 1 1 .997 r
F P
0 g
s
.57508 .24791 m .54053 .24114 L .54344 .27712 L closepath p 1 .769 .812 r
F P
0 g
s
.57508 .24791 m .54344 .27712 L .57354 .28512 L closepath p 1 .796 .83 r
F P
0 g
s
.71167 1.19855 m .66523 1.20855 L .67247 1.1802 L closepath p .752 .442 .612 \
r
F P
0 g
s
.71167 1.19855 m .67247 1.1802 L .71708 1.17034 L closepath p .752 .441 .611 \
r
F P
0 g
s
.37132 .47265 m .36631 .41244 L .3688 .42924 L closepath p .813 .622 .055 r
F P
0 g
s
.37396 .49322 m .37132 .47265 L .3688 .42924 L closepath p .809 .619 .051 r
F P
0 g
s
.368 .45379 m .36631 .41244 L .36701 .42082 L p .893 .692 .148 r
F P
0 g
s
.36701 .42082 m .37132 .47265 L .368 .45379 L p .893 .692 .148 r
F P
0 g
s
.36778 .48513 m .368 .45379 L .36948 .4829 L p .93 .748 .219 r
F P
0 g
s
.36948 .4829 m .37059 .50484 L .36778 .48513 L p .93 .748 .219 r
F P
0 g
s
.36778 .48513 m .3611 .50447 L .36599 .4875 L p .802 .833 .345 r
F P
0 g
s
.368 .45379 m .36778 .48513 L .3688 .48046 L p .132 .132 .435 r
F P
0 g
s
.37059 .50484 m .368 .45379 L .37132 .47265 L closepath p .964 .779 .264 r
F P
0 g
s
.3688 .48046 m .38415 .40994 L .368 .45379 L p .132 .132 .435 r
F P
0 g
s
.38147 .43375 m .38415 .40994 L .38185 .42049 L p .064 .064 .127 r
F P
0 g
s
.38185 .42049 m .36778 .48513 L .38147 .43375 L p .064 .064 .127 r
F P
0 g
s
.36599 .4875 m .38147 .43375 L .36778 .48513 L p .802 .833 .345 r
F P
0 g
s
.37509 .44691 m .38147 .43375 L .37767 .44694 L p .923 .95 .51 r
F P
0 g
s
.37767 .44694 m .3611 .50447 L .37509 .44691 L p .923 .95 .51 r
F P
0 g
s
.36631 .41244 m .368 .45379 L .36883 .44962 L p .152 .152 .523 r
F P
0 g
s
.368 .45379 m .38415 .40994 L .38425 .37205 L closepath p .131 .131 .427 r
F P
0 g
s
.37509 .44691 m .35446 .45871 L .37367 .4469 L p .539 .929 .758 r
F P
0 g
s
.38147 .43375 m .37509 .44691 L .38125 .44129 L p .616 .271 .475 r
F P
0 g
s
.38851 .43778 m .39504 .42869 L .37851 .44379 L p .097 .097 .097 r
F P
0 g
s
.37851 .44379 m .37509 .44691 L .38851 .43778 L p .097 .097 .097 r
F P
0 g
s
.38125 .44129 m .39504 .42869 L .38147 .43375 L p .616 .271 .475 r
F P
0 g
s
.38851 .43778 m .39993 .42539 L .39504 .42869 L closepath p .089 .089 .089 r
F P
0 g
s
.38851 .43778 m .39436 .43449 L .39993 .42539 L closepath p .137 .129 .129 r
F P
0 g
s
.38851 .43778 m .36984 .44567 L .39282 .43519 L p .511 .93 .815 r
F P
0 g
s
.39282 .43519 m .39436 .43449 L .38851 .43778 L p .511 .93 .815 r
F P
0 g
s
.37367 .4469 m .38851 .43778 L .37509 .44691 L p .539 .929 .758 r
F P
0 g
s
.36399 .37813 m .3688 .42924 L .36631 .41244 L p .799 .605 .035 r
F P
0 g
s
.36883 .44962 m .38425 .37205 L .36631 .41244 L p .152 .152 .523 r
F P
0 g
s
.36631 .41244 m .36345 .3724 L .36399 .37813 L p .799 .605 .035 r
F P
0 g
s
.33806 .62678 m .36851 .55164 L .36519 .59066 L closepath p 1 1 .975 r
F P
0 g
s
.36519 .59066 m .36851 .55164 L .3733 .52715 L closepath p .903 .832 .326 r
F P
0 g
s
.78109 .78025 m .79719 .74328 L .81741 .78753 L closepath p .392 .414 .856 r
F P
0 g
s
.78109 .78025 m .81741 .78753 L .81046 .82538 L closepath p .351 .469 .93 r
F P
0 g
s
.3733 .52715 m .37132 .47265 L .37396 .49322 L closepath p .922 .75 .22 r
F P
0 g
s
.39658 .27887 m .39965 .24363 L .39491 .24588 L closepath p .742 .539 .775 r
F P
0 g
s
.3889 .28352 m .39658 .27887 L .39491 .24588 L closepath p .706 .511 .769 r
F P
0 g
s
.3213 .82702 m .31541 .75884 L .33706 .78194 L closepath p 1 .993 .865 r
F P
0 g
s
.33706 .78194 m .31541 .75884 L .31782 .74285 L closepath p 1 .96 .786 r
F P
0 g
s
.44711 1.16864 m .38042 1.09557 L .43387 1.12229 L closepath p 1 .806 .661 r
F P
0 g
s
.43387 1.12229 m .38042 1.09557 L .37759 1.0557 L closepath p 1 .862 .749 r
F P
0 g
s
.39885 .27625 m .40707 .24039 L .39965 .24363 L closepath p .75 .533 .759 r
F P
0 g
s
.39658 .27887 m .39885 .27625 L .39965 .24363 L closepath p .432 .278 .66 r
F P
0 g
s
.36631 .41244 m .38602 .33405 L .36345 .3724 L closepath p .202 .18 .628 r
F P
0 g
s
.36345 .3724 m .38602 .33405 L .3889 .28352 L closepath p .151 .151 .444 r
F P
0 g
s
.67281 .34647 m .62543 .33463 L .62236 .37807 L closepath p 1 .833 .853 r
F P
0 g
s
.67281 .34647 m .62236 .37807 L .67139 .39232 L closepath p 1 .787 .818 r
F P
0 g
s
.31995 .73123 m .3058 .66796 L .32764 .70106 L closepath p 1 .983 .838 r
F P
0 g
s
.32764 .70106 m .3058 .66796 L .3274 .65069 L closepath p 1 1 1 r
F P
0 g
s
.71715 .41502 m .67139 .39232 L .67155 .44132 L closepath p 1 .837 .799 r
F P
0 g
s
.71715 .41502 m .67155 .44132 L .71828 .45918 L closepath p 1 .792 .764 r
F P
0 g
s
.76077 1.12577 m .78317 1.1166 L .78099 1.14825 L closepath p .698 .457 .696 \
r
F P
0 g
s
.76077 1.12577 m .78099 1.14825 L .75733 1.15742 L closepath p .702 .457 .693 \
r
F P
0 g
s
.77317 .54819 m .7617 .51634 L .76023 .55486 L closepath p 1 .893 .482 r
F P
0 g
s
.77317 .54819 m .76023 .55486 L .77528 .59232 L closepath p 1 .863 .454 r
F P
0 g
s
.6244 .29624 m .57354 .28512 L .57313 .33023 L closepath p 1 .792 .844 r
F P
0 g
s
.6244 .29624 m .57313 .33023 L .62543 .33463 L closepath p 1 .782 .836 r
F P
0 g
s
.54053 .24114 m .51654 .23349 L .51379 .26712 L closepath p 1 .787 .794 r
F P
0 g
s
.54053 .24114 m .51379 .26712 L .54344 .27712 L closepath p 1 .797 .8 r
F P
0 g
s
.40661 .27233 m .41951 .23508 L .40707 .24039 L closepath p .756 .535 .756 r
F P
0 g
s
.39885 .27625 m .40661 .27233 L .40707 .24039 L closepath p .731 .515 .751 r
F P
0 g
s
.37759 1.0557 m .33544 .98236 L .37375 1.00469 L closepath p 1 .891 .728 r
F P
0 g
s
.37375 1.00469 m .33544 .98236 L .33732 .93192 L closepath p 1 .909 .759 r
F P
0 g
s
.7803 .67518 m .7837 .65964 L .78266 .69761 L closepath p .059 .059 .268 r
F P
0 g
s
.80628 1.07317 m .84658 1.0404 L .84852 1.07047 L closepath p .537 .372 .712 \
r
F P
0 g
s
.80628 1.07317 m .84852 1.07047 L .80731 1.1036 L closepath p .537 .372 .713 \
r
F P
0 g
s
.42313 .23408 m .41951 .23508 L .41789 .26716 L closepath p .797 .568 .763 r
F P
0 g
s
.40661 .27233 m .41789 .26716 L .41951 .23508 L closepath p .749 .528 .752 r
F P
0 g
s
.71708 1.17034 m .75733 1.15742 L .752 1.18552 L closepath p .668 .367 .581 r
F P
0 g
s
.71708 1.17034 m .752 1.18552 L .71167 1.19855 L closepath p .668 .367 .582 r
F P
0 g
s
.42313 .23408 m .41789 .26716 L .42266 .26609 L closepath p .801 .571 .764 r
F P
0 g
s
.43094 .23133 m .42313 .23408 L .42266 .26609 L closepath p .782 .556 .761 r
F P
0 g
s
.51654 .23349 m .44525 .22883 L .44484 .26178 L closepath p .958 .701 .779 r
F P
0 g
s
.51654 .23349 m .44484 .26178 L .51379 .26712 L closepath p .967 .719 .796 r
F P
0 g
s
.3368 .63736 m .33222 .58637 L .33806 .62678 L closepath p 1 .879 .421 r
F P
0 g
s
.33806 .62678 m .33222 .58637 L .36851 .55164 L closepath p 1 1 1 r
F P
0 g
s
.43094 .23133 m .42266 .26609 L .43117 .26368 L closepath p .798 .571 .767 r
F P
0 g
s
.43621 .22966 m .43094 .23133 L .43117 .26368 L closepath p .801 .574 .768 r
F P
0 g
s
.83249 .98496 m .85707 .94761 L .8671 1.01078 L closepath p .364 .343 .788 r
F P
0 g
s
.83249 .98496 m .8671 1.01078 L .84658 1.0404 L closepath p .393 .343 .772 r
F P
0 g
s
.57354 .28512 m .54344 .27712 L .54304 .32476 L closepath p 1 .798 .833 r
F P
0 g
s
.57354 .28512 m .54304 .32476 L .57313 .33023 L closepath p 1 .811 .84 r
F P
0 g
s
.78317 1.1166 m .80731 1.1036 L .80527 1.13591 L closepath p .613 .379 .658 r
F P
0 g
s
.78317 1.1166 m .80527 1.13591 L .78099 1.14825 L closepath p .604 .38 .667 r
F P
0 g
s
.7628 .47685 m .71828 .45918 L .71873 .50195 L closepath p 1 .809 .803 r
F P
0 g
s
.7628 .47685 m .71873 .50195 L .7617 .51634 L closepath p 1 .789 .788 r
F P
0 g
s
.43621 .22966 m .43117 .26368 L .43547 .26375 L closepath p .933 .694 .8 r
F P
0 g
s
.44525 .22883 m .43621 .22966 L .43547 .26375 L closepath p .904 .67 .798 r
F P
0 g
s
.44525 .22883 m .43547 .26375 L .44484 .26178 L closepath p .829 .595 .77 r
F P
0 g
s
.31782 .74285 m .2958 .70241 L .31995 .73123 L closepath p 1 .958 .788 r
F P
0 g
s
.31995 .73123 m .2958 .70241 L .3058 .66796 L closepath p 1 .983 .852 r
F P
0 g
s
.38602 .33405 m .39658 .27887 L .3889 .28352 L closepath p .668 .45 .714 r
F P
0 g
s
.36851 .55164 m .37059 .50484 L .3733 .52715 L closepath p 1 .949 .518 r
F P
0 g
s
.3733 .52715 m .37059 .50484 L .37132 .47265 L closepath p .914 .743 .21 r
F P
0 g
s
.61678 1.25098 m .57583 1.25597 L .60294 1.21696 L closepath p .774 .376 .459 \
r
F P
0 g
s
.61678 1.25098 m .60294 1.21696 L .63975 1.21297 L closepath p .759 .369 .467 \
r
F P
0 g
s
.57583 1.25597 m .51914 1.21708 L .60294 1.21696 L closepath p .872 .487 .524 \
r
F P
0 g
s
.81046 .82538 m .81741 .78753 L .84414 .85257 L closepath p .169 .249 .773 r
F P
0 g
s
.81046 .82538 m .84414 .85257 L .84176 .89126 L closepath p .156 .241 .781 r
F P
0 g
s
.39567 .32836 m .39885 .27625 L .39658 .27887 L closepath p .394 .23 .619 r
F P
0 g
s
.38602 .33405 m .39567 .32836 L .39658 .27887 L closepath p .62 .407 .694 r
F P
0 g
s
.62543 .33463 m .57313 .33023 L .57197 .37101 L closepath p 1 .787 .843 r
F P
0 g
s
.62543 .33463 m .57197 .37101 L .62236 .37807 L closepath p 1 .805 .856 r
F P
0 g
s
.78633 .68099 m .78266 .69761 L .7836 .6636 L p .066 .066 .302 r
F P
0 g
s
.78266 .69761 m .78633 .68099 L .78835 .71746 L closepath p .094 .094 .502 r
F P
0 g
s
.78835 .71746 m .78633 .68099 L .79096 .70807 L closepath p .048 .048 .276 r
F P
0 g
s
.78633 .68099 m .78182 .66886 L .78984 .70325 L p .792 .544 .027 r
F P
0 g
s
.78984 .70325 m .79096 .70807 L .78633 .68099 L p .792 .544 .027 r
F P
0 g
s
.7836 .6636 m .7837 .65964 L .78633 .68099 L p .066 .066 .302 r
F P
0 g
s
.64708 1.24569 m .61678 1.25098 L .63975 1.21297 L closepath p .726 .334 .447 \
r
F P
0 g
s
.64708 1.24569 m .63975 1.21297 L .66523 1.20855 L closepath p .702 .319 .452 \
r
F P
0 g
s
.32964 .64358 m .32898 .59676 L .3368 .63736 L closepath p 1 1 1 r
F P
0 g
s
.3368 .63736 m .32898 .59676 L .33222 .58637 L closepath p 1 .992 .647 r
F P
0 g
s
.54344 .27712 m .51379 .26712 L .51434 .31525 L closepath p 1 .791 .792 r
F P
0 g
s
.54344 .27712 m .51434 .31525 L .54304 .32476 L closepath p 1 .843 .818 r
F P
0 g
s
.31541 .75884 m .29324 .71246 L .31782 .74285 L closepath p 1 .957 .789 r
F P
0 g
s
.31782 .74285 m .29324 .71246 L .2958 .70241 L closepath p 1 .961 .798 r
F P
0 g
s
.32473 .85663 m .30925 .80319 L .3213 .82702 L closepath p 1 .806 .378 r
F P
0 g
s
.3213 .82702 m .30925 .80319 L .31541 .75884 L closepath p 1 .986 .7 r
F P
0 g
s
.39781 .32782 m .40661 .27233 L .39885 .27625 L closepath p .72 .496 .734 r
F P
0 g
s
.39567 .32836 m .39781 .32782 L .39885 .27625 L closepath p .755 .525 .741 r
F P
0 g
s
.4658 1.20833 m .38851 1.13988 L .44711 1.16864 L closepath p 1 .721 .546 r
F P
0 g
s
.51914 1.21708 m .4658 1.20833 L .44711 1.16864 L closepath p 1 .727 .712 r
F P
0 g
s
.44711 1.16864 m .38851 1.13988 L .38042 1.09557 L closepath p 1 .819 .693 r
F P
0 g
s
.36631 .41244 m .38425 .37205 L .38602 .33405 L closepath p .127 .127 .38 r
F P
0 g
s
.32964 .64358 m .31596 .60717 L .32898 .59676 L closepath p 1 1 1 r
F P
0 g
s
.30671 .63159 m .31596 .60717 L .32964 .64358 L closepath p 1 1 1 r
F P
0 g
s
.30671 .63159 m .32964 .64358 L .3274 .65069 L closepath p 1 .916 .782 r
F P
0 g
s
.33544 .98236 m .31788 .90166 L .33732 .93192 L closepath p 1 .939 .647 r
F P
0 g
s
.33732 .93192 m .31788 .90166 L .32395 .87512 L closepath p 1 .998 .775 r
F P
0 g
s
.52438 1.25417 m .48814 1.23274 L .51914 1.21708 L closepath p .881 .449 .424 \
r
F P
0 g
s
.52438 1.25417 m .51914 1.21708 L .57583 1.25597 L closepath p .873 .443 .426 \
r
F P
0 g
s
.48814 1.23274 m .4658 1.20833 L .51914 1.21708 L closepath p .985 .603 .553 \
r
F P
0 g
s
.77528 .59232 m .76023 .55486 L .76727 .5998 L closepath p 1 .717 .228 r
F P
0 g
s
.77528 .59232 m .76727 .5998 L .78039 .63676 L closepath p 1 .744 .251 r
F P
0 g
s
.67139 .39232 m .62236 .37807 L .62295 .43086 L closepath p 1 .795 .831 r
F P
0 g
s
.67139 .39232 m .62295 .43086 L .67155 .44132 L closepath p 1 .778 .819 r
F P
0 g
s
.4028 .32482 m .41789 .26716 L .40661 .27233 L closepath p .736 .506 .732 r
F P
0 g
s
.39781 .32782 m .4028 .32482 L .40661 .27233 L closepath p .639 .424 .703 r
F P
0 g
s
.69762 1.23644 m .64708 1.24569 L .66523 1.20855 L closepath p .68 .297 .44 r
F P
0 g
s
.69762 1.23644 m .66523 1.20855 L .71167 1.19855 L closepath p .673 .301 .457 \
r
F P
0 g
s
.38042 1.09557 m .33581 1.02301 L .37759 1.0557 L closepath p 1 .879 .715 r
F P
0 g
s
.37759 1.0557 m .33581 1.02301 L .33544 .98236 L closepath p 1 .897 .749 r
F P
0 g
s
.75733 1.15742 m .78099 1.14825 L .77406 1.1773 L closepath p .627 .329 .563 \
r
F P
0 g
s
.75733 1.15742 m .77406 1.1773 L .752 1.18552 L closepath p .624 .328 .565 r
F P
0 g
s
.33222 .58637 m .36501 .52656 L .36851 .55164 L closepath p 1 1 1 r
F P
0 g
s
.36851 .55164 m .36501 .52656 L .37059 .50484 L closepath p 1 1 .648 r
F P
0 g
s
.80731 1.1036 m .84852 1.07047 L .8514 1.10408 L closepath p .482 .293 .642 r
F P
0 g
s
.80731 1.1036 m .8514 1.10408 L .80527 1.13591 L closepath p .478 .271 .612 r
F P
0 g
s
.3274 .65069 m .3058 .66796 L .291 .63101 L closepath p 1 1 1 r
F P
0 g
s
.3274 .65069 m .291 .63101 L .30671 .63159 L closepath p 1 .958 .885 r
F P
0 g
s
.78835 .71746 m .79096 .70807 L .79719 .74328 L closepath p .064 .064 .437 r
F P
0 g
s
.30925 .80319 m .29125 .72937 L .31541 .75884 L closepath p 1 .957 .788 r
F P
0 g
s
.31541 .75884 m .29125 .72937 L .29324 .71246 L closepath p 1 .957 .787 r
F P
0 g
s
.42266 .26609 m .41789 .26716 L .41642 .31975 L closepath p .794 .555 .747 r
F P
0 g
s
.4028 .32482 m .41642 .31975 L .41789 .26716 L closepath p .753 .522 .739 r
F P
0 g
s
.71828 .45918 m .67155 .44132 L .6732 .48122 L closepath p 1 .815 .801 r
F P
0 g
s
.71828 .45918 m .6732 .48122 L .71873 .50195 L closepath p 1 .815 .801 r
F P
0 g
s
.3058 .66796 m .2958 .70241 L .28823 .6822 L closepath p 1 1 .836 r
F P
0 g
s
.28823 .6822 m .291 .63101 L .3058 .66796 L closepath p 1 1 .854 r
F P
0 g
s
.51379 .26712 m .44484 .26178 L .44365 .31201 L closepath p .961 .691 .76 r
F P
0 g
s
.51379 .26712 m .44365 .31201 L .51434 .31525 L closepath p .993 .746 .802 r
F P
0 g
s
.42266 .26609 m .41642 .31975 L .42218 .3174 L closepath p .759 .524 .737 r
F P
0 g
s
.43117 .26368 m .42266 .26609 L .42218 .3174 L closepath p .787 .547 .742 r
F P
0 g
s
.57313 .33023 m .54304 .32476 L .54423 .36406 L closepath p 1 .79 .811 r
F P
0 g
s
.57313 .33023 m .54423 .36406 L .57197 .37101 L closepath p 1 .833 .834 r
F P
0 g
s
.43117 .26368 m .42218 .3174 L .42859 .31604 L closepath p .839 .594 .758 r
F P
0 g
s
.43547 .26375 m .43117 .26368 L .42859 .31604 L closepath p .926 .668 .768 r
F P
0 g
s
.43547 .26375 m .42859 .31604 L .43078 .31492 L closepath p .747 .508 .725 r
F P
0 g
s
.44484 .26178 m .43547 .26375 L .43078 .31492 L closepath p .818 .569 .742 r
F P
0 g
s
.62236 .37807 m .57197 .37101 L .57255 .42106 L closepath p 1 .785 .829 r
F P
0 g
s
.62236 .37807 m .57255 .42106 L .62295 .43086 L closepath p 1 .789 .832 r
F P
0 g
s
.44484 .26178 m .43078 .31492 L .44365 .31201 L closepath p .845 .595 .752 r
F P
0 g
s
.7617 .51634 m .71873 .50195 L .7198 .53836 L closepath p 1 .786 .783 r
F P
0 g
s
.7617 .51634 m .7198 .53836 L .76023 .55486 L closepath p 1 .794 .789 r
F P
0 g
s
.32395 .87512 m .30464 .83818 L .32473 .85663 L closepath p 1 .928 .715 r
F P
0 g
s
.32473 .85663 m .30464 .83818 L .30925 .80319 L closepath p 1 .957 .771 r
F P
0 g
s
.78099 1.14825 m .80527 1.13591 L .80188 1.16398 L closepath p .517 .242 .531 \
r
F P
0 g
s
.78099 1.14825 m .80188 1.16398 L .77406 1.1773 L closepath p .524 .239 .519 \
r
F P
0 g
s
.71167 1.19855 m .752 1.18552 L .73653 1.2223 L closepath p .545 .197 .37 r
F P
0 g
s
.71167 1.19855 m .73653 1.2223 L .69762 1.23644 L closepath p .531 .198 .391 \
r
F P
0 g
s
.32898 .59676 m .32853 .55779 L .33222 .58637 L closepath p 1 .989 .612 r
F P
0 g
s
.33222 .58637 m .32853 .55779 L .36501 .52656 L closepath p 1 1 1 r
F P
0 g
s
.36501 .52656 m .36778 .48513 L .37059 .50484 L closepath p 1 1 .626 r
F P
0 g
s
.54304 .32476 m .51434 .31525 L .51542 .35672 L closepath p 1 .824 .789 r
F P
0 g
s
.54304 .32476 m .51542 .35672 L .54423 .36406 L closepath p 1 .831 .792 r
F P
0 g
s
.79719 .74328 m .79096 .70807 L .80757 .75326 L closepath p .071 .071 .473 r
F P
0 g
s
.79719 .74328 m .80757 .75326 L .81741 .78753 L closepath p .061 .061 .49 r
F P
0 g
s
.84176 .89126 m .84414 .85257 L .85744 .90781 L closepath p .106 .106 .511 r
F P
0 g
s
.84176 .89126 m .85744 .90781 L .85707 .94761 L closepath p .1 .1 .517 r
F P
0 g
s
.84658 1.0404 m .8671 1.01078 L .86544 1.05059 L closepath p .315 .245 .692 r
F P
0 g
s
.84658 1.0404 m .86544 1.05059 L .84852 1.07047 L closepath p .347 .23 .651 r
F P
0 g
s
.43873 1.23483 m .40161 1.18293 L .4658 1.20833 L closepath p 1 .584 .389 r
F P
0 g
s
.43873 1.23483 m .4658 1.20833 L .48814 1.23274 L closepath p .985 .531 .356 \
r
F P
0 g
s
.4658 1.20833 m .40161 1.18293 L .38851 1.13988 L closepath p 1 .777 .655 r
F P
0 g
s
.38851 1.13988 m .33953 1.06827 L .38042 1.09557 L closepath p 1 .831 .628 r
F P
0 g
s
.38042 1.09557 m .33953 1.06827 L .33581 1.02301 L closepath p 1 .873 .697 r
F P
0 g
s
.27995 .68919 m .28643 .68271 L .2958 .70241 L closepath p 1 1 .799 r
F P
0 g
s
.27995 .68919 m .2958 .70241 L .29324 .71246 L closepath p 1 .994 .766 r
F P
0 g
s
.2958 .70241 m .28643 .68271 L .28823 .6822 L closepath p 1 .914 .573 r
F P
0 g
s
.31788 .90166 m .30296 .85337 L .32395 .87512 L closepath p 1 .97 .808 r
F P
0 g
s
.32395 .87512 m .30296 .85337 L .30464 .83818 L closepath p 1 .933 .731 r
F P
0 g
s
.30671 .63159 m .31345 .61584 L .31596 .60717 L closepath p .054 .054 .054 r
F P
0 g
s
.30671 .63159 m .291 .63101 L .31345 .61584 L closepath p 1 1 1 r
F P
0 g
s
.33544 .98236 m .31987 .94578 L .31788 .90166 L closepath p 1 .85 .434 r
F P
0 g
s
.31958 .97052 m .33544 .98236 L .33581 1.02301 L closepath p 1 .9 .521 r
F P
0 g
s
.31958 .97052 m .31987 .94578 L .33544 .98236 L closepath p 1 .901 .522 r
F P
0 g
s
.78039 .63676 m .76727 .5998 L .77116 .62042 L closepath p .915 .623 .087 r
F P
0 g
s
.67155 .44132 m .62295 .43086 L .6257 .46941 L closepath p 1 .781 .823 r
F P
0 g
s
.67155 .44132 m .6257 .46941 L .6732 .48122 L closepath p 1 .773 .817 r
F P
0 g
s
.57197 .37101 m .54423 .36406 L .54383 .40951 L closepath p 1 .803 .791 r
F P
0 g
s
.57197 .37101 m .54383 .40951 L .57255 .42106 L closepath p 1 .84 .809 r
F P
0 g
s
.30464 .83818 m .28362 .77243 L .30925 .80319 L closepath p 1 .953 .777 r
F P
0 g
s
.30925 .80319 m .28362 .77243 L .29125 .72937 L closepath p 1 .957 .788 r
F P
0 g
s
.31596 .60717 m .32661 .56807 L .32898 .59676 L closepath p 1 1 1 r
F P
0 g
s
.32898 .59676 m .32661 .56807 L .32853 .55779 L closepath p 1 .886 .418 r
F P
0 g
s
.38425 .37205 m .39567 .32836 L .38602 .33405 L closepath p .64 .439 .723 r
F P
0 g
s
.71873 .50195 m .6732 .48122 L .67868 .51447 L closepath p 1 .778 .738 r
F P
0 g
s
.71873 .50195 m .67868 .51447 L .7198 .53836 L closepath p 1 .812 .77 r
F P
0 g
s
.32853 .55779 m .3611 .50447 L .36501 .52656 L closepath p 1 1 1 r
F P
0 g
s
.36501 .52656 m .3611 .50447 L .36778 .48513 L closepath p 1 1 .745 r
F P
0 g
s
.29324 .71246 m .29125 .72937 L .27166 .69577 L closepath p 1 .979 .733 r
F P
0 g
s
.29324 .71246 m .27166 .69577 L .27995 .68919 L closepath p 1 .984 .741 r
F P
0 g
s
.31345 .61584 m .30985 .58018 L .31596 .60717 L closepath p 1 .987 .634 r
F P
0 g
s
.31596 .60717 m .30985 .58018 L .32661 .56807 L closepath p 1 1 1 r
F P
0 g
s
.80527 1.13591 m .8514 1.10408 L .84572 1.12922 L closepath p .206 .159 .238 \
r
F P
0 g
s
.80527 1.13591 m .84572 1.12922 L .80188 1.16398 L closepath p .259 .179 .433 \
r
F P
0 g
s
.39202 .36821 m .39781 .32782 L .39567 .32836 L closepath p .765 .547 .765 r
F P
0 g
s
.38425 .37205 m .39202 .36821 L .39567 .32836 L closepath p .658 .456 .732 r
F P
0 g
s
.752 1.18552 m .77406 1.1773 L .75977 1.21355 L closepath p .465 .186 .306 r
F P
0 g
s
.752 1.18552 m .75977 1.21355 L .73653 1.2223 L closepath p .428 .184 .312 r
F P
0 g
s
.51434 .31525 m .44365 .31201 L .44424 .34939 L closepath p .985 .717 .767 r
F P
0 g
s
.51434 .31525 m .44424 .34939 L .51542 .35672 L closepath p 1 .768 .81 r
F P
0 g
s
.39698 .36554 m .4028 .32482 L .39781 .32782 L closepath p .653 .448 .724 r
F P
0 g
s
.39202 .36821 m .39698 .36554 L .39781 .32782 L closepath p .643 .439 .721 r
F P
0 g
s
.28823 .6822 m .28643 .68271 L .27884 .66893 L closepath p 1 .917 .576 r
F P
0 g
s
.28823 .6822 m .27884 .66893 L .2809 .65845 L closepath p 1 .998 .725 r
F P
0 g
s
.28823 .6822 m .2809 .65845 L .291 .63101 L closepath p 1 1 .872 r
F P
0 g
s
.7837 .65964 m .77116 .62042 L .77619 .64291 L closepath p .829 .541 .04 r
F P
0 g
s
.7837 .65964 m .77619 .64291 L .78633 .68099 L closepath p .903 .655 .111 r
F P
0 g
s
.25906 .73931 m .29125 .72937 L .28362 .77243 L closepath p 1 .974 .765 r
F P
0 g
s
.25906 .73931 m .27166 .69577 L .29125 .72937 L closepath p 1 1 .794 r
F P
0 g
s
.62295 .43086 m .57255 .42106 L .57637 .45421 L closepath p 1 .754 .782 r
F P
0 g
s
.62295 .43086 m .57637 .45421 L .6257 .46941 L closepath p 1 .796 .817 r
F P
0 g
s
.76023 .55486 m .7198 .53836 L .72701 .58268 L closepath p 1 .756 .728 r
F P
0 g
s
.76023 .55486 m .72701 .58268 L .76727 .5998 L closepath p 1 .775 .741 r
F P
0 g
s
.29833 .89632 m .31788 .90166 L .31987 .94578 L closepath p 1 .885 .638 r
F P
0 g
s
.29916 .87962 m .31788 .90166 L .29833 .89632 L closepath p 1 .916 .678 r
F P
0 g
s
.31788 .90166 m .29916 .87962 L .30296 .85337 L closepath p 1 .967 .786 r
F P
0 g
s
.40258 .36298 m .41642 .31975 L .4028 .32482 L closepath p .762 .538 .755 r
F P
0 g
s
.39698 .36554 m .40258 .36298 L .4028 .32482 L closepath p .703 .491 .743 r
F P
0 g
s
.31345 .61584 m .30804 .5893 L .30985 .58018 L closepath p 1 .923 .501 r
F P
0 g
s
.31345 .61584 m .291 .63101 L .28286 .60445 L closepath p 1 1 1 r
F P
0 g
s
.31345 .61584 m .28286 .60445 L .30804 .5893 L closepath p 1 1 1 r
F P
0 g
s
.54423 .36406 m .51542 .35672 L .51708 .39885 L closepath p 1 .819 .775 r
F P
0 g
s
.54423 .36406 m .51708 .39885 L .54383 .40951 L closepath p 1 .829 .78 r
F P
0 g
s
.32661 .56807 m .3291 .5306 L .32853 .55779 L closepath p .933 .827 .319 r
F P
0 g
s
.32853 .55779 m .3291 .5306 L .3611 .50447 L closepath p 1 1 1 r
F P
0 g
s
.291 .63101 m .2809 .65845 L .27396 .64025 L closepath p 1 1 .872 r
F P
0 g
s
.291 .63101 m .27396 .64025 L .27336 .62972 L closepath p 1 1 .913 r
F P
0 g
s
.291 .63101 m .27336 .62972 L .28286 .60445 L closepath p 1 1 .845 r
F P
0 g
s
.27704 .67942 m .27884 .66893 L .28643 .68271 L closepath p 1 1 .749 r
F P
0 g
s
.27704 .67942 m .28643 .68271 L .27995 .68919 L closepath p 1 1 .812 r
F P
0 g
s
.45414 1.24556 m .43873 1.23483 L .48814 1.23274 L closepath p .845 .324 .166 \
r
F P
0 g
s
.45414 1.24556 m .48814 1.23274 L .52438 1.25417 L closepath p .804 .271 .161 \
r
F P
0 g
s
.85707 .94761 m .85744 .90781 L .86605 .96213 L closepath p .063 .063 .325 r
F P
0 g
s
.85707 .94761 m .86605 .96213 L .8671 1.01078 L closepath p .062 .062 .327 r
F P
0 g
s
.42218 .3174 m .41642 .31975 L .41445 .35858 L closepath p .768 .541 .754 r
F P
0 g
s
.40258 .36298 m .41445 .35858 L .41642 .31975 L closepath p .749 .526 .749 r
F P
0 g
s
.30296 .85337 m .27789 .80754 L .30464 .83818 L closepath p 1 .933 .732 r
F P
0 g
s
.30464 .83818 m .27789 .80754 L .28362 .77243 L closepath p 1 .952 .774 r
F P
0 g
s
.81741 .78753 m .82711 .81876 L .84414 .85257 L closepath p .76 .398 .03 r
F P
0 g
s
.81741 .78753 m .80757 .75326 L .82711 .81876 L closepath p .004 .004 .215 r
F P
0 g
s
.86544 1.05059 m .8514 1.10408 L .84852 1.07047 L closepath p .289 .19 .586 r
F P
0 g
s
.3218 1.03475 m .31934 1.0011 L .33581 1.02301 L closepath p 1 .824 .399 r
F P
0 g
s
.3218 1.03475 m .33581 1.02301 L .33953 1.06827 L closepath p 1 .831 .405 r
F P
0 g
s
.33581 1.02301 m .31934 1.0011 L .32085 .98806 L closepath p 1 .962 .611 r
F P
0 g
s
.33581 1.02301 m .32085 .98806 L .31958 .97052 L closepath p 1 .775 .304 r
F P
0 g
s
.42218 .3174 m .41445 .35858 L .42107 .35599 L closepath p .787 .558 .76 r
F P
0 g
s
.42859 .31604 m .42218 .3174 L .42107 .35599 L closepath p .844 .605 .77 r
F P
0 g
s
.6732 .48122 m .6257 .46941 L .6286 .49802 L closepath p 1 .741 .769 r
F P
0 g
s
.6732 .48122 m .6286 .49802 L .67868 .51447 L closepath p 1 .741 .769 r
F P
0 g
s
.77406 1.1773 m .80188 1.16398 L .7872 1.20049 L closepath p .318 .165 .207 r
F P
0 g
s
.77406 1.1773 m .7872 1.20049 L .75977 1.21355 L closepath p .288 .165 .222 r
F P
0 g
s
.42859 .31604 m .42107 .35599 L .42762 .35396 L closepath p .813 .577 .76 r
F P
0 g
s
.43078 .31492 m .42859 .31604 L .42762 .35396 L closepath p .764 .537 .751 r
F P
0 g
s
.43078 .31492 m .42762 .35396 L .43417 .35228 L closepath p .85 .61 .77 r
F P
0 g
s
.44365 .31201 m .43078 .31492 L .43417 .35228 L closepath p .852 .612 .77 r
F P
0 g
s
.40161 1.18293 m .34519 1.11545 L .38851 1.13988 L closepath p 1 .778 .566 r
F P
0 g
s
.38851 1.13988 m .34519 1.11545 L .33953 1.06827 L closepath p 1 .848 .673 r
F P
0 g
s
.78633 .68099 m .77619 .64291 L .78182 .66886 L closepath p .764 .496 .02 r
F P
0 g
s
.57255 .42106 m .54383 .40951 L .54574 .44143 L closepath p 1 .802 .749 r
F P
0 g
s
.57255 .42106 m .54574 .44143 L .57637 .45421 L closepath p 1 .801 .748 r
F P
0 g
s
.30985 .58018 m .32367 .53867 L .32661 .56807 L closepath p 1 1 1 r
F P
0 g
s
.32661 .56807 m .32367 .53867 L .3291 .5306 L closepath p 1 1 .757 r
F P
0 g
s
.44365 .31201 m .43417 .35228 L .44424 .34939 L closepath p .843 .603 .767 r
F P
0 g
s
.7198 .53836 m .67868 .51447 L .69008 .54936 L closepath p 1 .712 .595 r
F P
0 g
s
.7198 .53836 m .69008 .54936 L .72701 .58268 L closepath p 1 .796 .669 r
F P
0 g
s
.3291 .5306 m .37509 .44691 L .3611 .50447 L closepath p 1 1 1 r
F P
0 g
s
.38415 .40994 m .39202 .36821 L .38425 .37205 L closepath p .678 .491 .765 r
F P
0 g
s
.29916 .87962 m .27719 .82669 L .30296 .85337 L closepath p 1 .962 .791 r
F P
0 g
s
.30296 .85337 m .27719 .82669 L .27789 .80754 L closepath p 1 .93 .724 r
F P
0 g
s
.2809 .65845 m .27884 .66893 L .26887 .64449 L closepath p 1 .999 .715 r
F P
0 g
s
.2809 .65845 m .26887 .64449 L .27396 .64025 L closepath p 1 .96 .644 r
F P
0 g
s
.30804 .5893 m .30659 .54766 L .30985 .58018 L closepath p 1 .899 .441 r
F P
0 g
s
.30985 .58018 m .30659 .54766 L .32367 .53867 L closepath p 1 1 1 r
F P
0 g
s
.55592 1.26589 m .52438 1.25417 L .57583 1.25597 L closepath p .754 .251 .202 \
r
F P
0 g
s
.58757 1.26982 m .55592 1.26589 L .57583 1.25597 L closepath p .567 .153 .153 \
r
F P
0 g
s
.58757 1.26982 m .57583 1.25597 L .61678 1.25098 L closepath p .542 .159 .159 \
r
F P
0 g
s
.61504 1.2699 m .58757 1.26982 L .61678 1.25098 L closepath p .51 .155 .155 r
F P
0 g
s
.61504 1.2699 m .61678 1.25098 L .64708 1.24569 L closepath p .474 .153 .153 \
r
F P
0 g
s
.27995 .68919 m .27166 .69577 L .27704 .67942 L closepath p 1 1 .769 r
F P
0 g
s
.6257 .46941 m .57637 .45421 L .581 .47591 L closepath p 1 .714 .693 r
F P
0 g
s
.6257 .46941 m .581 .47591 L .6286 .49802 L closepath p 1 .77 .753 r
F P
0 g
s
.51542 .35672 m .44424 .34939 L .44619 .38678 L closepath p 1 .748 .785 r
F P
0 g
s
.51542 .35672 m .44619 .38678 L .51708 .39885 L closepath p 1 .764 .799 r
F P
0 g
s
.39249 .40413 m .39698 .36554 L .39202 .36821 L closepath p .664 .476 .756 r
F P
0 g
s
.38415 .40994 m .39249 .40413 L .39202 .36821 L closepath p .644 .46 .75 r
F P
0 g
s
.67363 1.26082 m .61504 1.2699 L .64708 1.24569 L closepath p .404 .143 .143 \
r
F P
0 g
s
.67363 1.26082 m .64708 1.24569 L .69762 1.23644 L closepath p .404 .143 .143 \
r
F P
0 g
s
.26739 .65465 m .26887 .64449 L .27884 .66893 L closepath p 1 .984 .678 r
F P
0 g
s
.26739 .65465 m .27884 .66893 L .27704 .67942 L closepath p 1 1 .717 r
F P
0 g
s
.25906 .73931 m .28362 .77243 L .25332 .76837 L closepath p 1 .961 .734 r
F P
0 g
s
.28362 .77243 m .27789 .80754 L .25332 .76837 L closepath p 1 .97 .746 r
F P
0 g
s
.39787 .40053 m .40258 .36298 L .39698 .36554 L closepath p .716 .515 .766 r
F P
0 g
s
.39249 .40413 m .39787 .40053 L .39698 .36554 L closepath p .653 .466 .751 r
F P
0 g
s
.8671 1.01078 m .86605 .96213 L .86859 .99906 L closepath p .02 .02 .117 r
F P
0 g
s
.8671 1.01078 m .86859 .99906 L .86544 1.05059 L closepath p .009 .009 .009 r
F P
0 g
s
.32367 .53867 m .35446 .45871 L .3291 .5306 L closepath p .931 1 .612 r
F P
0 g
s
.3291 .5306 m .35446 .45871 L .37509 .44691 L closepath p 1 1 1 r
F P
0 g
s
.30804 .5893 m .302 .55512 L .30659 .54766 L closepath p 1 1 .691 r
F P
0 g
s
.30804 .5893 m .28286 .60445 L .27543 .57277 L closepath p 1 1 1 r
F P
0 g
s
.30804 .5893 m .27543 .57277 L .302 .55512 L closepath p 1 1 1 r
F P
0 g
s
.79096 .70807 m .78182 .66886 L .79091 .712 L closepath p .724 .469 .009 r
F P
0 g
s
.79096 .70807 m .79091 .712 L .80757 .75326 L closepath p .715 .382 .014 r
F P
0 g
s
.27704 .67942 m .27166 .69577 L .26262 .66995 L closepath p 1 1 .763 r
F P
0 g
s
.27704 .67942 m .26262 .66995 L .26739 .65465 L closepath p 1 1 .725 r
F P
0 g
s
.28286 .60445 m .27336 .62972 L .26681 .59215 L closepath p 1 1 .802 r
F P
0 g
s
.28286 .60445 m .26681 .59215 L .27543 .57277 L closepath p 1 1 .804 r
F P
0 g
s
.31987 .94578 m .31958 .97052 L .30968 .97833 L closepath p 1 .916 .662 r
F P
0 g
s
.31987 .94578 m .30968 .97833 L .29943 .96555 L closepath p 1 .909 .665 r
F P
0 g
s
.31987 .94578 m .29943 .96555 L .29913 .94108 L closepath p 1 .881 .649 r
F P
0 g
s
.29913 .94108 m .29976 .9244 L .31987 .94578 L closepath p 1 .911 .687 r
F P
0 g
s
.29833 .89632 m .31987 .94578 L .29976 .9244 L closepath p 1 .886 .64 r
F P
0 g
s
.40241 .39776 m .41445 .35858 L .40258 .36298 L closepath p .759 .544 .767 r
F P
0 g
s
.39787 .40053 m .40241 .39776 L .40258 .36298 L closepath p .643 .451 .739 r
F P
0 g
s
.29916 .87962 m .29833 .89632 L .2737 .89626 L closepath p 1 .916 .678 r
F P
0 g
s
.29916 .87962 m .2737 .89626 L .27425 .85114 L closepath p 1 .917 .679 r
F P
0 g
s
.29916 .87962 m .27425 .85114 L .27719 .82669 L closepath p 1 .952 .749 r
F P
0 g
s
.54383 .40951 m .51708 .39885 L .51986 .42968 L closepath p 1 .81 .75 r
F P
0 g
s
.54383 .40951 m .51986 .42968 L .54574 .44143 L closepath p 1 .805 .747 r
F P
0 g
s
.76727 .5998 m .72701 .58268 L .73253 .60405 L closepath p 1 .721 .656 r
F P
0 g
s
.76727 .5998 m .73253 .60405 L .77116 .62042 L closepath p 1 .772 .709 r
F P
0 g
s
.80188 1.16398 m .84572 1.12922 L .7872 1.20049 L closepath p .119 .119 .119 \
r
F P
0 g
s
.67868 .51447 m .6286 .49802 L .65138 .51762 L closepath p .822 .299 .167 r
F P
0 g
s
.67868 .51447 m .65138 .51762 L .69008 .54936 L closepath p 1 .716 .552 r
F P
0 g
s
.50999 1.26749 m .45414 1.24556 L .52438 1.25417 L closepath p .592 .112 .112 \
r
F P
0 g
s
.50999 1.26749 m .52438 1.25417 L .55592 1.26589 L closepath p .582 .111 .111 \
r
F P
0 g
s
.34519 1.11545 m .3218 1.03475 L .33953 1.06827 L closepath p 1 .785 .337 r
F P
0 g
s
.27396 .64025 m .26887 .64449 L .27336 .62972 L closepath p 1 1 .693 r
F P
0 g
s
.57637 .45421 m .54574 .44143 L .55056 .45984 L closepath p 1 .713 .612 r
F P
0 g
s
.57637 .45421 m .55056 .45984 L .581 .47591 L closepath p 1 .745 .644 r
F P
0 g
s
.42107 .35599 m .41445 .35858 L .41247 .39319 L closepath p .792 .569 .77 r
F P
0 g
s
.40241 .39776 m .41247 .39319 L .41445 .35858 L closepath p .741 .526 .758 r
F P
0 g
s
.86544 1.05059 m .86859 .99906 L .86582 1.05252 L closepath p .004 .004 .004 \
r
F P
0 g
s
.86544 1.05059 m .86582 1.05252 L .8514 1.10408 L closepath p .007 .007 .007 \
r
F P
0 g
s
.42107 .35599 m .41247 .39319 L .42122 .39073 L closepath p .802 .578 .774 r
F P
0 g
s
.42762 .35396 m .42107 .35599 L .42122 .39073 L closepath p .821 .594 .777 r
F P
0 g
s
.69762 1.23644 m .73653 1.2223 L .7144 1.24411 L closepath p .133 .104 .104 r
F P
0 g
s
.69762 1.23644 m .7144 1.24411 L .67363 1.26082 L closepath p .128 .105 .105 \
r
F P
0 g
s
.39911 1.22022 m .35848 1.16239 L .40161 1.18293 L closepath p .991 .519 .228 \
r
F P
0 g
s
.39911 1.22022 m .40161 1.18293 L .43873 1.23483 L closepath p .991 .52 .227 \
r
F P
0 g
s
.40161 1.18293 m .35848 1.16239 L .34519 1.11545 L closepath p 1 .779 .568 r
F P
0 g
s
.42762 .35396 m .42122 .39073 L .42736 .38951 L closepath p .864 .634 .79 r
F P
0 g
s
.43417 .35228 m .42762 .35396 L .42736 .38951 L closepath p .858 .628 .789 r
F P
0 g
s
.30659 .54766 m .34432 .45942 L .32367 .53867 L closepath p 1 1 1 r
F P
0 g
s
.32367 .53867 m .34432 .45942 L .35446 .45871 L closepath p 1 1 .984 r
F P
0 g
s
.6286 .49802 m .581 .47591 L .60982 .48819 L closepath p .016 .338 .704 r
F P
0 g
s
.6286 .49802 m .60982 .48819 L .65138 .51762 L closepath p .592 .047 .047 r
F P
0 g
s
.43417 .35228 m .42736 .38951 L .43462 .38839 L closepath p .864 .634 .79 r
F P
0 g
s
.44424 .34939 m .43417 .35228 L .43462 .38839 L closepath p .853 .625 .79 r
F P
0 g
s
.25906 .73931 m .2576 .69183 L .27166 .69577 L closepath p 1 1 .738 r
F P
0 g
s
.27166 .69577 m .2576 .69183 L .26262 .66995 L closepath p 1 1 .718 r
F P
0 g
s
.38147 .43375 m .39249 .40413 L .38415 .40994 L closepath p .587 .375 .675 r
F P
0 g
s
.29976 .9244 m .2737 .89626 L .29833 .89632 L closepath p 1 .886 .643 r
F P
0 g
s
.44424 .34939 m .43462 .38839 L .44619 .38678 L closepath p .87 .642 .796 r
F P
0 g
s
.27336 .62972 m .26887 .64449 L .26545 .60777 L closepath p 1 1 .67 r
F P
0 g
s
.27336 .62972 m .26545 .60777 L .26681 .59215 L closepath p 1 .88 .423 r
F P
0 g
s
.84414 .85257 m .82711 .81876 L .83822 .8653 L closepath p .952 .63 .116 r
F P
0 g
s
.84414 .85257 m .83822 .8653 L .85744 .90781 L closepath p .952 .637 .12 r
F P
0 g
s
.8514 1.10408 m .86582 1.05252 L .85566 1.10235 L closepath p .03 .03 .03 r
F P
0 g
s
.8514 1.10408 m .85566 1.10235 L .84572 1.12922 L closepath p .028 .028 .028 \
r
F P
0 g
s
.32085 .98806 m .30968 .97833 L .31958 .97052 L closepath p 1 .836 .608 r
F P
0 g
s
.73877 .62388 m .77619 .64291 L .77116 .62042 L closepath p 1 .782 .692 r
F P
0 g
s
.73877 .62388 m .77116 .62042 L .73253 .60405 L closepath p 1 .714 .621 r
F P
0 g
s
.24821 .7925 m .25332 .76837 L .27789 .80754 L closepath p 1 .966 .737 r
F P
0 g
s
.24821 .7925 m .27789 .80754 L .27719 .82669 L closepath p 1 .938 .694 r
F P
0 g
s
.39504 .42869 m .39787 .40053 L .39249 .40413 L closepath p .671 .494 .776 r
F P
0 g
s
.38147 .43375 m .39504 .42869 L .39249 .40413 L closepath p .798 .587 .792 r
F P
0 g
s
.69162 .55828 m .69198 .56259 L .72701 .58268 L closepath p 1 .817 .703 r
F P
0 g
s
.69017 .55614 m .69162 .55828 L .72701 .58268 L closepath p .922 .418 .14 r
F P
0 g
s
.69008 .54936 m .69017 .55614 L .72701 .58268 L closepath p 1 .84 .768 r
F P
0 g
s
.71989 .57859 m .69198 .56259 L .70529 .57281 L p .597 .059 .059 r
F P
0 g
s
.69008 .54936 m .65138 .51762 L .65101 .52309 L p 1 .797 .711 r
F P
0 g
s
.65101 .52309 m .65065 .52842 L .66801 .53764 L p 1 .797 .711 r
F P
0 g
s
.66801 .53764 m .69008 .54936 L p .65101 .52309 L 1 .797 .711 r
F P
0 g
s
.69008 .54936 m .65065 .52842 L .69017 .55614 L closepath p 1 .833 .773 r
F P
0 g
s
.65065 .52842 m .65543 .53064 L .67412 .54436 L p .066 .379 .822 r
F P
0 g
s
.67412 .54436 m .69017 .55614 L .67351 .54445 L p .066 .379 .822 r
F P
0 g
s
.67351 .54445 m .65065 .52842 L p .67412 .54436 L .066 .379 .822 r
F P
0 g
s
.69162 .55828 m .69093 .55726 L p .67554 .54572 L .901 .396 .126 r
F P
0 g
s
.65543 .53064 m .67698 .5471 L p .67554 .54572 L .901 .396 .126 r
F P
0 g
s
.67698 .5471 m .69162 .55828 L p .67554 .54572 L .901 .396 .126 r
F P
0 g
s
.69093 .55726 m .69017 .55614 L .68636 .55335 L p .901 .396 .126 r
F P
0 g
s
.68636 .55335 m .65543 .53064 L p .69093 .55726 L .901 .396 .126 r
F P
0 g
s
.69162 .55828 m .65543 .53064 L .65573 .53713 L closepath p 1 .755 .572 r
F P
0 g
s
.69162 .55828 m .65573 .53713 L .69198 .56259 L closepath p 1 .827 .687 r
F P
0 g
s
.70529 .57281 m .6981 .56729 L p .67857 .54372 L .077 .269 .779 r
F P
0 g
s
.6981 .56729 m .69198 .56259 L .68944 .55901 L p .077 .269 .779 r
F P
0 g
s
.68944 .55901 m .67857 .54372 L p .6981 .56729 L .077 .269 .779 r
F P
0 g
s
.67857 .54372 m .67616 .54031 L .67913 .54364 L p .077 .269 .779 r
F P
0 g
s
.67913 .54364 m .70529 .57281 L p .67857 .54372 L .077 .269 .779 r
F P
0 g
s
.69198 .56259 m .65573 .53713 L .67616 .54031 L closepath p .419 .765 1 r
F P
0 g
s
.70529 .57281 m .67616 .54031 L .68543 .56494 L p 1 .635 .265 r
F P
0 g
s
.68543 .56494 m .687 .5691 L .70529 .57281 L p 1 .635 .265 r
F P
0 g
s
.70529 .57281 m .72701 .58268 L .71989 .57859 L p .597 .059 .059 r
F P
0 g
s
.72701 .58268 m .70529 .57281 L .73253 .60405 L closepath p 1 .737 .427 r
F P
0 g
s
.51708 .39885 m .44619 .38678 L .44658 .40963 L closepath p 1 .718 .736 r
F P
0 g
s
.51708 .39885 m .44658 .40963 L .51986 .42968 L closepath p 1 .762 .782 r
F P
0 g
s
.302 .55512 m .32179 .4614 L .30659 .54766 L closepath p 1 1 .607 r
F P
0 g
s
.30659 .54766 m .32179 .4614 L .34432 .45942 L closepath p 1 1 1 r
F P
0 g
s
.39993 .42539 m .40241 .39776 L .39787 .40053 L closepath p .651 .465 .752 r
F P
0 g
s
.39504 .42869 m .39993 .42539 L .39787 .40053 L closepath p .545 .386 .727 r
F P
0 g
s
.26562 .64177 m .26887 .64449 L .26739 .65465 L closepath p 1 .913 .459 r
F P
0 g
s
.26521 .61884 m .26545 .60777 L .26887 .64449 L closepath p .878 .71 .165 r
F P
0 g
s
.26521 .61884 m .26887 .64449 L .26562 .64177 L closepath p 1 .969 .541 r
F P
0 g
s
.54574 .44143 m .51986 .42968 L .52433 .44529 L closepath p 1 .675 .546 r
F P
0 g
s
.52433 .44529 m .44824 .42415 L .4738 .43138 L p .225 .008 .008 r
F P
0 g
s
.52433 .44529 m .55056 .45984 L .54849 .45194 L p 1 .72 .59 r
F P
0 g
s
.55056 .45984 m .52433 .44529 L .54681 .45207 L closepath p .081 .52 .912 r
F P
0 g
s
.4738 .43138 m .54681 .45207 L .52433 .44529 L p .225 .008 .008 r
F P
0 g
s
.54849 .45194 m .54574 .44143 L .52433 .44529 L p 1 .72 .59 r
F P
0 g
s
.40445 .42203 m .41247 .39319 L .40241 .39776 L closepath p .741 .528 .76 r
F P
0 g
s
.39993 .42539 m .40445 .42203 L .40241 .39776 L closepath p .611 .431 .737 r
F P
0 g
s
.581 .47591 m .55056 .45984 L .57863 .46841 L closepath p .08 .529 .913 r
F P
0 g
s
.581 .47591 m .57863 .46841 L .60982 .48819 L closepath p .074 .441 .865 r
F P
0 g
s
.80757 .75326 m .79091 .712 L .80457 .78684 L closepath p .968 .673 .16 r
F P
0 g
s
.80757 .75326 m .80457 .78684 L .82711 .81876 L closepath p 1 .65 .188 r
F P
0 g
s
.27719 .82669 m .27425 .85114 L .24592 .81003 L closepath p 1 .964 .718 r
F P
0 g
s
.27719 .82669 m .24592 .81003 L .24821 .7925 L closepath p 1 .904 .629 r
F P
0 g
s
.26739 .65465 m .26262 .66995 L .26323 .6374 L closepath p 1 .953 .51 r
F P
0 g
s
.26739 .65465 m .26323 .6374 L .26562 .64177 L closepath p .869 .684 .133 r
F P
0 g
s
.54072 1.27072 m .50999 1.26749 L .55592 1.26589 L closepath p .217 .045 .045 \
r
F P
0 g
s
.54072 1.27072 m .55592 1.26589 L .58757 1.26982 L closepath p .237 .046 .046 \
r
F P
0 g
s
.42122 .39073 m .41247 .39319 L .41149 .41799 L closepath p .796 .566 .762 r
F P
0 g
s
.40445 .42203 m .41149 .41799 L .41247 .39319 L closepath p .696 .484 .738 r
F P
0 g
s
.32085 .98806 m .31934 1.0011 L .30817 1.02358 L closepath p 1 .96 .603 r
F P
0 g
s
.32085 .98806 m .30817 1.02358 L .30087 1.01588 L closepath p 1 .895 .654 r
F P
0 g
s
.32085 .98806 m .30087 1.01588 L .29785 .98499 L closepath p 1 .843 .633 r
F P
0 g
s
.32085 .98806 m .29785 .98499 L .30968 .97833 L closepath p 1 .865 .66 r
F P
0 g
s
.29976 .9244 m .29913 .94108 L .27383 .93764 L closepath p 1 .915 .671 r
F P
0 g
s
.29976 .9244 m .27383 .93764 L .2737 .89626 L closepath p 1 .888 .649 r
F P
0 g
s
.73653 1.2223 m .75977 1.21355 L .74295 1.22935 L closepath p .083 .083 .083 \
r
F P
0 g
s
.73653 1.2223 m .74295 1.22935 L .7144 1.24411 L closepath p .084 .084 .084 r
F P
0 g
s
.74096 .63777 m .77619 .64291 L .73877 .62388 L closepath p 1 .809 .737 r
F P
0 g
s
.77619 .64291 m .74096 .63777 L .74445 .64785 L closepath p 1 .662 .554 r
F P
0 g
s
.77619 .64291 m .74445 .64785 L .78182 .66886 L closepath p 1 .798 .691 r
F P
0 g
s
.42122 .39073 m .41149 .41799 L .41873 .41674 L closepath p .831 .602 .778 r
F P
0 g
s
.42736 .38951 m .42122 .39073 L .41873 .41674 L closepath p .863 .629 .785 r
F P
0 g
s
.302 .55512 m .31293 .46226 L .32179 .4614 L closepath p 1 1 .932 r
F P
0 g
s
.302 .55512 m .27543 .57277 L .28606 .47246 L closepath p 1 1 1 r
F P
0 g
s
.302 .55512 m .28606 .47246 L .31293 .46226 L closepath p 1 1 1 r
F P
0 g
s
.42736 .38951 m .41873 .41674 L .42531 .41599 L closepath p .882 .649 .793 r
F P
0 g
s
.43462 .38839 m .42736 .38951 L .42531 .41599 L closepath p .863 .633 .79 r
F P
0 g
s
.35848 1.16239 m .32379 1.08708 L .34519 1.11545 L closepath p 1 .655 .186 r
F P
0 g
s
.34519 1.11545 m .32379 1.08708 L .3218 1.03475 L closepath p 1 .853 .46 r
F P
0 g
s
.43462 .38839 m .42531 .41599 L .43345 .41446 L closepath p .917 .689 .812 r
F P
0 g
s
.44619 .38678 m .43462 .38839 L .43345 .41446 L closepath p .874 .653 .807 r
F P
0 g
s
.73253 .60405 m .70529 .57281 L .71439 .59864 L closepath p 1 .668 .316 r
F P
0 g
s
.73253 .60405 m .71439 .59864 L .73877 .62388 L closepath p 1 .692 .343 r
F P
0 g
s
.27543 .57277 m .26681 .59215 L .26968 .51941 L closepath p 1 1 .714 r
F P
0 g
s
.27543 .57277 m .26968 .51941 L .28606 .47246 L closepath p 1 1 .911 r
F P
0 g
s
.3218 1.03475 m .30817 1.02358 L .31934 1.0011 L closepath p 1 .866 .611 r
F P
0 g
s
.44619 .38678 m .43345 .41446 L .44658 .40963 L closepath p .734 .501 .727 r
F P
0 g
s
.51986 .42968 m .44658 .40963 L .44438 .42135 L closepath p .985 .607 .562 r
F P
0 g
s
.51986 .42968 m .44438 .42135 L .52433 .44529 L closepath p 1 .648 .612 r
F P
0 g
s
.24174 .74621 m .24721 .7294 L .25906 .73931 L closepath p 1 1 .698 r
F P
0 g
s
.25906 .73931 m .25217 .71281 L .2576 .69183 L closepath p 1 .999 .651 r
F P
0 g
s
.24721 .7294 m .25217 .71281 L .25906 .73931 L closepath p 1 1 .702 r
F P
0 g
s
.25332 .76837 m .24174 .74621 L .25906 .73931 L closepath p 1 .98 .659 r
F P
0 g
s
.39436 .43449 m .40445 .42203 L .39993 .42539 L closepath p .104 .104 .104 r
F P
0 g
s
.55056 .45984 m .54681 .45207 L .57863 .46841 L closepath p .082 .532 .919 r
F P
0 g
s
.29943 .96555 m .27383 .93764 L .29913 .94108 L closepath p 1 .884 .634 r
F P
0 g
s
.27425 .85114 m .25262 .87252 L .23675 .8468 L closepath p 1 .97 .714 r
F P
0 g
s
.27425 .85114 m .23675 .8468 L .24592 .81003 L closepath p 1 .951 .691 r
F P
0 g
s
.2737 .89626 m .25262 .87252 L .27425 .85114 L closepath p 1 .916 .686 r
F P
0 g
s
.26262 .66995 m .2576 .69183 L .25365 .67971 L closepath p .956 .887 .406 r
F P
0 g
s
.26262 .66995 m .25365 .67971 L .26323 .6374 L closepath p .965 .883 .4 r
F P
0 g
s
.85744 .90781 m .83822 .8653 L .84477 .91851 L closepath p 1 .787 .329 r
F P
0 g
s
.85744 .90781 m .84477 .91851 L .86605 .96213 L closepath p 1 .774 .318 r
F P
0 g
s
.30968 .97833 m .29785 .98499 L .29943 .96555 L closepath p 1 .932 .713 r
F P
0 g
s
.39842 .4318 m .41149 .41799 L .40445 .42203 L closepath p .207 .142 .142 r
F P
0 g
s
.39842 .4318 m .40445 .42203 L .39489 .43384 L p .125 .125 .125 r
F P
0 g
s
.40412 .42872 m .41149 .41799 L .40034 .42977 L p .256 .151 .151 r
F P
0 g
s
.40034 .42977 m .39842 .4318 L .40412 .42872 L p .256 .151 .151 r
F P
0 g
s
.41873 .41674 m .41149 .41799 L .40412 .42872 L closepath p .63 .215 .348 r
F P
0 g
s
.41873 .41674 m .40412 .42872 L .41209 .42675 L closepath p .539 .185 .264 r
F P
0 g
s
.40412 .42872 m .39451 .43034 L .40891 .4274 L p .244 .763 .766 r
F P
0 g
s
.40891 .4274 m .41209 .42675 L .40412 .42872 L p .244 .763 .766 r
F P
0 g
s
.39842 .4318 m .38569 .43665 L .4027 .42934 L p .491 .916 .798 r
F P
0 g
s
.4027 .42934 m .40412 .42872 L .39842 .4318 L p .491 .916 .798 r
F P
0 g
s
.39489 .43384 m .39436 .43449 L .39842 .4318 L p .125 .125 .125 r
F P
0 g
s
.26323 .6374 m .25365 .67971 L .25949 .65395 L p 0 0 0 r
F P
0 g
s
.26537 .62791 m .26562 .64177 L .26323 .6374 L p .905 .784 .26 r
F P
0 g
s
.25949 .65395 m .26551 .62746 L .26323 .6374 L p 0 0 0 r
F P
0 g
s
.26323 .6374 m .26521 .61884 L .26537 .62791 L p .905 .784 .26 r
F P
0 g
s
.65138 .51762 m .60982 .48819 L .60922 .49595 L p 1 .748 .625 r
F P
0 g
s
.61614 .5013 m .65138 .51762 L p .60922 .49595 L 1 .748 .625 r
F P
0 g
s
.65138 .51762 m .60906 .49803 L .65065 .52842 L closepath p 1 .802 .709 r
F P
0 g
s
.61888 .50061 m .65543 .53064 L .65196 .52903 L p .086 .405 .866 r
F P
0 g
s
.65196 .52903 m .65065 .52842 L .64668 .52494 L p .086 .405 .866 r
F P
0 g
s
.64668 .52494 m .61888 .50061 L p .65196 .52903 L .086 .405 .866 r
F P
0 g
s
.60922 .49595 m .60906 .49803 L .61614 .5013 L p 1 .748 .625 r
F P
0 g
s
.7597 .68489 m .79091 .712 L .78182 .66886 L closepath p 1 .793 .565 r
F P
0 g
s
.7597 .68489 m .78182 .66886 L .74445 .64785 L closepath p 1 .654 .467 r
F P
0 g
s
.56638 1.27136 m .54072 1.27072 L .58757 1.26982 L closepath p .065 .015 .015 \
r
F P
0 g
s
.56638 1.27136 m .58757 1.26982 L .61504 1.2699 L closepath p .04 .011 .011 r
F P
0 g
s
.75977 1.21355 m .7872 1.20049 L .77511 1.20848 L closepath p .029 .029 .029 \
r
F P
0 g
s
.75977 1.21355 m .77511 1.20848 L .74295 1.22935 L closepath p .047 .047 .047 \
r
F P
0 g
s
.42531 .41599 m .41873 .41674 L .41209 .42675 L closepath p .694 .267 .359 r
F P
0 g
s
.43345 .41446 m .42531 .41599 L .41755 .42592 L closepath p .681 .248 .338 r
F P
0 g
s
.43345 .41446 m .41755 .42592 L .4253 .42419 L closepath p .552 .18 .216 r
F P
0 g
s
.44658 .40963 m .43345 .41446 L .4253 .42419 L closepath p .319 .149 .149 r
F P
0 g
s
.42531 .41599 m .41209 .42675 L .41755 .42592 L closepath p .639 .203 .309 r
F P
0 g
s
.26681 .59215 m .26545 .60777 L .26617 .57236 L closepath p .812 .705 .156 r
F P
0 g
s
.26681 .59215 m .26617 .57236 L .26968 .51941 L closepath p .833 .735 .195 r
F P
0 g
s
.25217 .71281 m .25365 .67971 L .2576 .69183 L closepath p .958 .893 .414 r
F P
0 g
s
.73877 .62388 m .71439 .59864 L .71665 .60931 L closepath p 1 .765 .456 r
F P
0 g
s
.73877 .62388 m .71665 .60931 L .74096 .63777 L closepath p 1 .841 .562 r
F P
0 g
s
.40571 1.2283 m .39911 1.22022 L .43873 1.23483 L closepath p .944 .449 .15 r
F P
0 g
s
.40571 1.2283 m .43873 1.23483 L .45414 1.24556 L closepath p .711 .141 .105 \
r
F P
0 g
s
.52433 .44529 m .44438 .42135 L .44824 .42415 L closepath p .368 .042 .042 r
F P
0 g
s
.35446 .45871 m .36984 .44567 L .38851 .43778 L closepath p .801 1 .978 r
F P
0 g
s
.44658 .40963 m .4253 .42419 L .44438 .42135 L closepath p .758 .392 .514 r
F P
0 g
s
.65065 .52842 m .60906 .49803 L .61888 .50061 L closepath p .224 .618 1 r
F P
0 g
s
.29943 .96555 m .29785 .98499 L .28121 .99359 L closepath p 1 .936 .573 r
F P
0 g
s
.29943 .96555 m .28121 .99359 L .27383 .93764 L closepath p 1 .833 .542 r
F P
0 g
s
.30817 1.02358 m .3218 1.03475 L .32379 1.08708 L closepath p 1 .87 .618 r
F P
0 g
s
.25332 .76837 m .23719 .76802 L .24174 .74621 L closepath p 1 .981 .66 r
F P
0 g
s
.24821 .7925 m .23719 .76802 L .25332 .76837 L closepath p 1 .985 .666 r
F P
0 g
s
.2655 .60443 m .26521 .61884 L .26323 .6374 L closepath p .826 .727 .184 r
F P
0 g
s
.2655 .60443 m .26323 .6374 L .26518 .61892 L p .033 .033 .085 r
F P
0 g
s
.26577 .59168 m .26545 .60777 L .2655 .60443 L p .709 .591 .017 r
F P
0 g
s
.26545 .60777 m .26521 .61884 L .26548 .60514 L p .709 .591 .017 r
F P
0 g
s
.26548 .60514 m .2655 .60443 L .2655 .60461 L p .709 .591 .017 r
F P
0 g
s
.2655 .60461 m .26545 .60777 L p .26548 .60514 L .709 .591 .017 r
F P
0 g
s
.26551 .62746 m .26821 .59031 L .26519 .61889 L p .031 .031 .076 r
F P
0 g
s
.26519 .61889 m .26323 .6374 L .26495 .62988 L p .031 .031 .076 r
F P
0 g
s
.26495 .62988 m .26551 .62746 L p .26519 .61889 L .031 .031 .076 r
F P
0 g
s
.26518 .61892 m .26821 .59031 L .2655 .60443 L p .033 .033 .085 r
F P
0 g
s
.2655 .60443 m .26617 .57236 L .26577 .59168 L p .709 .591 .017 r
F P
0 g
s
.37822 .44041 m .39842 .4318 L .39436 .43449 L closepath p .672 1 .923 r
F P
0 g
s
.36984 .44567 m .37822 .44041 L .39436 .43449 L closepath p .643 1 .908 r
F P
0 g
s
.74096 .63777 m .71665 .60931 L .72016 .62409 L closepath p 1 .759 .424 r
F P
0 g
s
.74096 .63777 m .72016 .62409 L .74445 .64785 L closepath p 1 .631 .26 r
F P
0 g
s
.70529 .57281 m .687 .5691 L .71439 .59864 L closepath p 1 .66 .291 r
F P
0 g
s
.37822 .44041 m .38569 .43665 L .39842 .4318 L closepath p .439 .881 .771 r
F P
0 g
s
.34432 .45942 m .36984 .44567 L .35446 .45871 L closepath p .847 1 1 r
F P
0 g
s
.7872 1.20049 m .84572 1.12922 L .83824 1.13887 L closepath p .005 .005 .005 \
r
F P
0 g
s
.84572 1.12922 m .85566 1.10235 L .83824 1.13887 L closepath p .775 .882 .439 \
r
F P
0 g
s
.63543 1.25896 m .56638 1.27136 L .61504 1.2699 L closepath p .3 .813 .884 r
F P
0 g
s
.63543 1.25896 m .61504 1.2699 L .67363 1.26082 L closepath p .296 .811 .896 \
r
F P
0 g
s
.60982 .48819 m .57863 .46841 L .577 .47889 L closepath p 1 .779 .694 r
F P
0 g
s
.60982 .48819 m .577 .47889 L .60906 .49803 L closepath p 1 .732 .63 r
F P
0 g
s
.86605 .96213 m .84477 .91851 L .84788 .96064 L closepath p 1 .842 .42 r
F P
0 g
s
.86605 .96213 m .84788 .96064 L .86859 .99906 L closepath p 1 .855 .435 r
F P
0 g
s
.3661 1.19097 m .34217 1.13133 L .35848 1.16239 L closepath p .746 .384 .027 \
r
F P
0 g
s
.3661 1.19097 m .35848 1.16239 L .39911 1.22022 L closepath p .793 .312 .069 \
r
F P
0 g
s
.35848 1.16239 m .34217 1.13133 L .32379 1.08708 L closepath p .733 .369 .024 \
r
F P
0 g
s
.65573 .53713 m .65543 .53064 L .61888 .50061 L closepath p 1 .759 .566 r
F P
0 g
s
.37908 .43126 m .39451 .43034 L .39047 .43376 L closepath p .965 1 1 r
F P
0 g
s
.39451 .43034 m .38436 .43179 L .38722 .43151 L p .063 .028 .028 r
F P
0 g
s
.40378 .42885 m .39047 .43376 L .39451 .43034 L p .842 1 .967 r
F P
0 g
s
.38569 .43665 m .39047 .43376 L .40412 .42872 L closepath p .606 .991 .87 r
F P
0 g
s
.38722 .43151 m .40408 .42984 L .39451 .43034 L p .063 .028 .028 r
F P
0 g
s
.39451 .43034 m .40412 .42872 L .40378 .42885 L p .842 1 .967 r
F P
0 g
s
.2737 .89626 m .25052 .90168 L .24374 .88845 L closepath p 1 .959 .652 r
F P
0 g
s
.2737 .89626 m .24374 .88845 L .25262 .87252 L closepath p 1 .746 .395 r
F P
0 g
s
.27383 .93764 m .25052 .90168 L .2737 .89626 L closepath p 1 .893 .586 r
F P
0 g
s
.473 1.2563 m .40571 1.2283 L .45414 1.24556 L closepath p .343 .022 .022 r
F P
0 g
s
.473 1.2563 m .45414 1.24556 L .50999 1.26749 L closepath p .341 .026 .026 r
F P
0 g
s
.41209 .42675 m .39451 .43034 L .40408 .42984 L closepath p .209 .07 .07 r
F P
0 g
s
.41755 .42592 m .41209 .42675 L .40408 .42984 L closepath p .086 .052 .052 r
F P
0 g
s
.24775 .7131 m .24174 .74621 L .2387 .75063 L closepath p .767 .752 .228 r
F P
0 g
s
.24775 .7131 m .2387 .75063 L .24198 .74176 L p .09 .09 .122 r
F P
0 g
s
.25341 .68498 m .25217 .71281 L .24775 .7131 L p .873 .792 .271 r
F P
0 g
s
.24775 .7131 m .25217 .71281 L .24721 .7294 L closepath p .852 .844 .351 r
F P
0 g
s
.24775 .7131 m .24721 .7294 L .24174 .74621 L closepath p .859 .858 .371 r
F P
0 g
s
.24198 .74176 m .25718 .7007 L .24775 .7131 L p .09 .09 .122 r
F P
0 g
s
.24775 .7131 m .25365 .67971 L .25341 .68498 L p .873 .792 .271 r
F P
0 g
s
.30087 1.01588 m .28121 .99359 L .29785 .98499 L closepath p 1 .838 .487 r
F P
0 g
s
.41281 .42763 m .41755 .42592 L .40594 .4293 L p .014 .014 .014 r
F P
0 g
s
.4253 .42419 m .41755 .42592 L .41281 .42763 L closepath p .018 .018 .018 r
F P
0 g
s
.42354 .42583 m .4253 .42419 L .41507 .42701 L p .036 .036 .036 r
F P
0 g
s
.41507 .42701 m .41281 .42763 L .42354 .42583 L p .036 .036 .036 r
F P
0 g
s
.44438 .42135 m .4253 .42419 L .42354 .42583 L closepath p .037 .037 .037 r
F P
0 g
s
.44438 .42135 m .42354 .42583 L .44824 .42415 L closepath p .259 .082 .082 r
F P
0 g
s
.42354 .42583 m .41179 .42439 L .43502 .42424 L p .305 .815 .944 r
F P
0 g
s
.43502 .42424 m .44824 .42415 L .4329 .42519 L p .305 .815 .944 r
F P
0 g
s
.4329 .42519 m .42835 .4255 L p .43334 .42455 L .305 .815 .944 r
F P
0 g
s
.4329 .42519 m .43334 .42455 L .43502 .42424 L .305 .815 .944 r
F
.57863 .46841 m .54681 .45207 L .5444 .45921 L closepath p 1 .627 .501 r
F P
0 g
s
.54681 .45207 m .44824 .42415 L .43939 .42475 L p .628 .116 .116 r
F P
0 g
s
.43939 .42475 m .43807 .42483 L .44669 .42699 L p .628 .116 .116 r
F P
0 g
s
.44669 .42699 m .54681 .45207 L p .43939 .42475 L .628 .116 .116 r
F P
0 g
s
.54681 .45207 m .43807 .42483 L .5444 .45921 L closepath p .966 .553 .482 r
F P
0 g
s
.55216 .45935 m .55184 .45935 L p .43642 .42005 L .119 .633 .953 r
F P
0 g
s
.55184 .45935 m .5444 .45921 L .50521 .445 L p .119 .633 .953 r
F P
0 g
s
.42835 .4255 m .42898 .42507 L .43502 .42424 L .305 .815 .944 r
F
.43939 .42475 m .44824 .42415 L .43355 .42425 L p .122 .029 .029 r
F P
0 g
s
.43355 .42425 m .41179 .42439 L p .43939 .42475 L .122 .029 .029 r
F P
0 g
s
.43807 .42483 m .43563 .42058 L p .45001 .42569 L .2 .689 1 r
F P
0 g
s
.5444 .45921 m .43807 .42483 L p .45001 .42569 L .2 .689 1 r
F P
0 g
s
.43807 .42483 m .43939 .42475 L p .43628 .42471 L .122 .029 .029 r
F P
0 g
s
.41179 .42439 m .43807 .42483 L p .43628 .42471 L .122 .029 .029 r
F P
0 g
s
.50521 .445 m .45247 .42587 L p .55184 .45935 L .119 .633 .953 r
F P
0 g
s
.45247 .42587 m .43642 .42005 L p .45387 .42599 L .119 .633 .953 r
F P
0 g
s
.45247 .42587 m .45387 .42599 L .55184 .45935 L .119 .633 .953 r
F
.43642 .42005 m .43503 .41954 L .43644 .42002 L p F P
0 g
s
.43644 .42002 m .55216 .45935 L p .43642 .42005 L .119 .633 .953 r
F P
0 g
s
.43563 .42058 m .43503 .41954 L .43642 .42005 L p .2 .689 1 r
F P
0 g
s
.43642 .42005 m .5444 .45921 L p .43563 .42058 L .2 .689 1 r
F P
0 g
s
.42835 .4255 m .42354 .42583 L p .42898 .42507 L .305 .815 .944 r
F P
0 g
s
.41281 .42763 m .39936 .42648 L .41977 .42593 L p .435 .905 1 r
F P
0 g
s
.41977 .42593 m .42354 .42583 L .41281 .42763 L p .435 .905 1 r
F P
0 g
s
.40594 .4293 m .40408 .42984 L .41281 .42763 L p .014 .014 .014 r
F P
0 g
s
.82711 .81876 m .80457 .78684 L .8118 .8374 L closepath p 1 .803 .405 r
F P
0 g
s
.82711 .81876 m .8118 .8374 L .83822 .8653 L closepath p 1 .759 .379 r
F P
0 g
s
.74445 .64785 m .72016 .62409 L .73017 .66527 L closepath p 1 .768 .464 r
F P
0 g
s
.74445 .64785 m .73017 .66527 L .7597 .68489 L closepath p 1 .645 .397 r
F P
0 g
s
.64115 .51482 m .67616 .54031 L .65573 .53713 L closepath p .542 .768 1 r
F P
0 g
s
.60906 .49803 m .577 .47889 L .58913 .48091 L closepath p .266 .672 1 r
F P
0 g
s
.60906 .49803 m .58913 .48091 L .61888 .50061 L closepath p .173 .605 1 r
F P
0 g
s
.61888 .50061 m .58763 .49023 L .61834 .50888 L closepath p 1 .754 .633 r
F P
0 g
s
.61834 .50888 m .58763 .49023 L .59655 .49386 L p .439 .78 1 r
F P
0 g
s
.64757 .52903 m .61888 .50061 L .61834 .50888 L p 1 .785 .616 r
F P
0 g
s
.59655 .49386 m .63653 .51011 L .61834 .50888 L p .439 .78 1 r
F P
0 g
s
.61834 .50888 m .65573 .53713 L .64757 .52903 L p 1 .785 .616 r
F P
0 g
s
.65573 .53713 m .61834 .50888 L .64115 .51482 L closepath p .831 .94 1 r
F P
0 g
s
.7872 1.20049 m .83824 1.13887 L .80555 1.17984 L closepath p .026 .026 .026 \
r
F P
0 g
s
.7872 1.20049 m .80555 1.17984 L .77511 1.20848 L closepath p .645 .95 .678 r
F P
0 g
s
.35879 .44401 m .37822 .44041 L .36984 .44567 L closepath p .833 1 1 r
F P
0 g
s
.34432 .45942 m .35879 .44401 L .36984 .44567 L closepath p .956 1 1 r
F P
0 g
s
.23276 .78494 m .23719 .76802 L .24821 .7925 L closepath p 1 .987 .669 r
F P
0 g
s
.23276 .78494 m .24821 .7925 L .24592 .81003 L closepath p 1 .915 .568 r
F P
0 g
s
.57863 .46841 m .5444 .45921 L .577 .47889 L closepath p 1 .773 .695 r
F P
0 g
s
.76216 .73572 m .79091 .712 L .75975 .71726 L closepath p 1 .826 .684 r
F P
0 g
s
.79091 .712 m .76216 .73572 L .76827 .75977 L closepath p 1 .729 .619 r
F P
0 g
s
.79091 .712 m .76827 .75977 L .80457 .78684 L closepath p 1 .828 .646 r
F P
0 g
s
.75975 .71726 m .79091 .712 L .7597 .68489 L closepath p 1 .895 .754 r
F P
0 g
s
.30817 1.02358 m .32379 1.08708 L .30312 1.07676 L closepath p 1 .87 .616 r
F P
0 g
s
.30817 1.02358 m .30312 1.07676 L .30178 1.03254 L closepath p 1 .868 .62 r
F P
0 g
s
.30817 1.02358 m .30178 1.03254 L .30137 1.02448 L closepath p 1 .891 .629 r
F P
0 g
s
.30817 1.02358 m .30137 1.02448 L .30087 1.01588 L closepath p 1 .868 .605 r
F P
0 g
s
.86859 .99906 m .84788 .96064 L .84579 1.01052 L closepath p 1 .933 .563 r
F P
0 g
s
.86859 .99906 m .84579 1.01052 L .86582 1.05252 L closepath p 1 .951 .579 r
F P
0 g
s
.36766 .43822 m .38569 .43665 L .37822 .44041 L closepath p .76 1 1 r
F P
0 g
s
.35879 .44401 m .36766 .43822 L .37822 .44041 L closepath p .818 1 1 r
F P
0 g
s
.71439 .59864 m .687 .5691 L .68871 .57814 L closepath p 1 .749 .427 r
F P
0 g
s
.71439 .59864 m .68871 .57814 L .71665 .60931 L closepath p 1 .764 .449 r
F P
0 g
s
.32179 .4614 m .35879 .44401 L .34432 .45942 L closepath p .972 1 1 r
F P
0 g
s
.37477 .4337 m .39047 .43376 L .38569 .43665 L closepath p .861 1 1 r
F P
0 g
s
.36766 .43822 m .37477 .4337 L .38569 .43665 L closepath p .849 1 1 r
F P
0 g
s
.37477 .4337 m .37908 .43126 L .39047 .43376 L closepath p .82 1 1 r
F P
0 g
s
.26821 .59031 m .27505 .55458 L .2655 .60443 L closepath p 0 0 0 r
F P
0 g
s
.27505 .55458 m .26617 .57236 L .2655 .60443 L closepath p .138 .138 .558 r
F P
0 g
s
.577 .47889 m .5444 .45921 L .55216 .45935 L closepath p .184 .62 1 r
F P
0 g
s
.577 .47889 m .55216 .45935 L .58913 .48091 L closepath p .254 .67 1 r
F P
0 g
s
.67363 1.26082 m .7144 1.24411 L .68126 1.24139 L closepath p .621 1 1 r
F P
0 g
s
.67363 1.26082 m .68126 1.24139 L .63543 1.25896 L closepath p .616 1 1 r
F P
0 g
s
.37908 .43126 m .38436 .43179 L .39451 .43034 L closepath p .477 .099 .099 r
F P
0 g
s
.61888 .50061 m .58913 .48091 L .58763 .49023 L closepath p 1 .793 .687 r
F P
0 g
s
.30087 1.01588 m .30137 1.02448 L .28227 1.0286 L closepath p 1 .868 .563 r
F P
0 g
s
.30087 1.01588 m .28227 1.0286 L .28121 .99359 L closepath p 1 .9 .587 r
F P
0 g
s
.6698 .55108 m .687 .5691 L .67616 .54031 L closepath p 1 .665 .363 r
F P
0 g
s
.64115 .51482 m .63653 .51011 L .67616 .54031 L closepath p .038 .257 .714 r
F P
0 g
s
.61834 .50888 m .63653 .51011 L .64115 .51482 L closepath p .257 .736 1 r
F P
0 g
s
.67616 .54031 m .63653 .51011 L .6698 .55108 L closepath p 1 .613 .341 r
F P
0 g
s
.71665 .60931 m .68871 .57814 L .69275 .59382 L closepath p 1 .766 .452 r
F P
0 g
s
.71665 .60931 m .69275 .59382 L .72016 .62409 L closepath p 1 .765 .45 r
F P
0 g
s
.24592 .81003 m .23675 .8468 L .22901 .824 L closepath p 1 .97 .604 r
F P
0 g
s
.24592 .81003 m .22901 .824 L .23276 .78494 L closepath p 1 .912 .562 r
F P
0 g
s
.40408 .42984 m .38436 .43179 L .3902 .42798 L closepath p .795 1 1 r
F P
0 g
s
.41281 .42763 m .40408 .42984 L .3902 .42798 L closepath p .604 1 1 r
F P
0 g
s
.2387 .75063 m .24174 .74621 L .23719 .76802 L closepath p .776 .744 .214 r
F P
0 g
s
.27383 .93764 m .25237 .92741 L .24539 .90286 L closepath p 1 .933 .672 r
F P
0 g
s
.27383 .93764 m .24539 .90286 L .25052 .90168 L closepath p 1 .971 .749 r
F P
0 g
s
.28121 .99359 m .25237 .92741 L .27383 .93764 L closepath p 1 .832 .533 r
F P
0 g
s
.32379 1.08708 m .34217 1.13133 L .31618 1.12053 L closepath p 1 .655 .4 r
F P
0 g
s
.32379 1.08708 m .31618 1.12053 L .30312 1.07676 L closepath p 1 .705 .396 r
F P
0 g
s
.41281 .42763 m .3902 .42798 L .39936 .42648 L closepath p .43 .902 .998 r
F P
0 g
s
.28948 1.06636 m .28227 1.0286 L .30137 1.02448 L closepath p 1 .79 .488 r
F P
0 g
s
.28948 1.06636 m .30137 1.02448 L .30178 1.03254 L closepath p 1 .857 .451 r
F P
0 g
s
.25262 .87252 m .24374 .88845 L .23833 .86073 L closepath p 1 .848 .43 r
F P
0 g
s
.25262 .87252 m .23833 .86073 L .23675 .8468 L closepath p 1 .65 .185 r
F P
0 g
s
.42354 .42583 m .39936 .42648 L .41179 .42439 L closepath p .467 .925 1 r
F P
0 g
s
.33568 .44302 m .36766 .43822 L .35879 .44401 L closepath p .865 1 1 r
F P
0 g
s
.32179 .4614 m .33568 .44302 L .35879 .44401 L closepath p .98 1 1 r
F P
0 g
s
.83822 .8653 m .8118 .8374 L .81697 .87902 L closepath p 1 .847 .509 r
F P
0 g
s
.83822 .8653 m .81697 .87902 L .84477 .91851 L closepath p 1 .861 .521 r
F P
0 g
s
.86582 1.05252 m .84579 1.01052 L .83953 1.06157 L closepath p 1 .987 .647 r
F P
0 g
s
.86582 1.05252 m .83953 1.06157 L .85566 1.10235 L closepath p 1 1 .691 r
F P
0 g
s
.27505 .55458 m .2793 .51616 L .26617 .57236 L closepath p .118 .118 .423 r
F P
0 g
s
.26968 .51941 m .26617 .57236 L .2793 .51616 L closepath p .152 .152 .6 r
F P
0 g
s
.30312 1.07676 m .28948 1.06636 L .30178 1.03254 L closepath p 1 .851 .453 r
F P
0 g
s
.72016 .62409 m .69275 .59382 L .70123 .63356 L closepath p 1 .766 .452 r
F P
0 g
s
.72016 .62409 m .70123 .63356 L .73017 .66527 L closepath p 1 .766 .452 r
F P
0 g
s
.7144 1.24411 m .74295 1.22935 L .70859 1.22677 L closepath p .751 1 1 r
F P
0 g
s
.7144 1.24411 m .70859 1.22677 L .68126 1.24139 L closepath p .76 1 1 r
F P
0 g
s
.30786 .45466 m .33568 .44302 L .32179 .4614 L closepath p .982 1 1 r
F P
0 g
s
.30786 .45466 m .32179 .4614 L .31293 .46226 L closepath p 1 .966 .716 r
F P
0 g
s
.58913 .48091 m .55216 .45935 L .54838 .4667 L closepath p 1 .699 .573 r
F P
0 g
s
.58913 .48091 m .54838 .4667 L .58763 .49023 L closepath p 1 .782 .689 r
F P
0 g
s
.34485 .43465 m .37477 .4337 L .36766 .43822 L closepath p .893 1 1 r
F P
0 g
s
.33568 .44302 m .34485 .43465 L .36766 .43822 L closepath p .927 1 1 r
F P
0 g
s
.7597 .68489 m .73017 .66527 L .73728 .69823 L closepath p 1 .785 .613 r
F P
0 g
s
.7597 .68489 m .73728 .69823 L .75975 .71726 L closepath p 1 .914 .712 r
F P
0 g
s
.5 1.26236 m .473 1.2563 L .50999 1.26749 L closepath p .024 .499 .775 r
F P
0 g
s
.5 1.26236 m .50999 1.26749 L .54072 1.27072 L closepath p .039 .605 .826 r
F P
0 g
s
.23276 .78494 m .2387 .75063 L .23719 .76802 L closepath p .8 .786 .271 r
F P
0 g
s
.25718 .7007 m .25365 .67971 L .24775 .7131 L closepath p .103 .103 .209 r
F P
0 g
s
.26551 .62746 m .25365 .67971 L .26387 .61985 L closepath p .831 .818 .314 r
F P
0 g
s
.26387 .61985 m .25365 .67971 L .25718 .7007 L closepath p .059 .059 .176 r
F P
0 g
s
.35461 .42939 m .37908 .43126 L .37477 .4337 L closepath p .863 1 1 r
F P
0 g
s
.34485 .43465 m .35461 .42939 L .37477 .4337 L closepath p .794 1 1 r
F P
0 g
s
.3902 .42798 m .38436 .43179 L .36518 .42254 L closepath p .829 1 1 r
F P
0 g
s
.3595 .42538 m .36518 .42254 L .38436 .43179 L closepath p .831 1 1 r
F P
0 g
s
.3595 .42538 m .38436 .43179 L .37908 .43126 L closepath p .352 .834 1 r
F P
0 g
s
.74295 1.22935 m .77511 1.20848 L .74024 1.20689 L closepath p .841 1 1 r
F P
0 g
s
.74295 1.22935 m .74024 1.20689 L .70859 1.22677 L closepath p .835 1 1 r
F P
0 g
s
.35461 .42939 m .3595 .42538 L .37908 .43126 L closepath p .934 1 1 r
F P
0 g
s
.83824 1.13887 m .85566 1.10235 L .83953 1.06157 L closepath p 1 1 .915 r
F P
0 g
s
.54838 .4667 m .60946 .48897 L .58763 .49023 L closepath p .617 .866 1 r
F P
0 g
s
.58763 .49023 m .60946 .48897 L .63653 .51011 L closepath p .645 .85 1 r
F P
0 g
s
.3661 1.19097 m .39911 1.22022 L .40571 1.2283 L closepath p .703 .205 .049 r
F P
0 g
s
.77511 1.20848 m .80555 1.17984 L .79529 1.15424 L closepath p .994 1 1 r
F P
0 g
s
.77511 1.20848 m .79529 1.15424 L .74024 1.20689 L closepath p .994 1 1 r
F P
0 g
s
.3902 .42798 m .36518 .42254 L .37419 .4221 L closepath p .485 .922 1 r
F P
0 g
s
.39936 .42648 m .3902 .42798 L .37419 .4221 L closepath p .595 .988 1 r
F P
0 g
s
.68871 .57814 m .687 .5691 L .6663 .55217 L closepath p 1 .75 .531 r
F P
0 g
s
.6698 .55108 m .6663 .55217 L .687 .5691 L closepath p 1 .845 .699 r
F P
0 g
s
.80555 1.17984 m .83824 1.13887 L .80939 1.13494 L closepath p 1 1 .963 r
F P
0 g
s
.83824 1.13887 m .83953 1.06157 L .80939 1.13494 L closepath p 1 1 .836 r
F P
0 g
s
.80555 1.17984 m .80939 1.13494 L .79529 1.15424 L closepath p 1 1 .922 r
F P
0 g
s
.39936 .42648 m .37419 .4221 L .38747 .4188 L closepath p .74 1 1 r
F P
0 g
s
.41179 .42439 m .39936 .42648 L .38747 .4188 L closepath p .669 1 1 r
F P
0 g
s
.63653 .51011 m .64742 .5313 L .6698 .55108 L closepath p 1 .576 .27 r
F P
0 g
s
.62917 .5161 m .64742 .5313 L .63653 .51011 L closepath p 1 .577 .273 r
F P
0 g
s
.60946 .48897 m .62917 .5161 L .63653 .51011 L closepath p 1 .548 .253 r
F P
0 g
s
.28606 .47246 m .26968 .51941 L .27686 .48995 L closepath p .985 1 .629 r
F P
0 g
s
.26968 .51941 m .2793 .51616 L .27686 .48995 L closepath p .152 .152 .593 r
F P
0 g
s
.43807 .42483 m .41179 .42439 L .40354 .41782 L closepath p .464 .9 1 r
F P
0 g
s
.43807 .42483 m .40354 .41782 L .43503 .41954 L closepath p .419 .873 1 r
F P
0 g
s
.55216 .45935 m .43503 .41954 L .42471 .42091 L closepath p .808 .27 .155 r
F P
0 g
s
.55216 .45935 m .42471 .42091 L .54838 .4667 L closepath p 1 .625 .544 r
F P
0 g
s
.41179 .42439 m .38747 .4188 L .40354 .41782 L closepath p .548 .957 1 r
F P
0 g
s
.26821 .59031 m .26551 .62746 L .26387 .61985 L closepath p .804 .709 .162 r
F P
0 g
s
.80457 .78684 m .76827 .75977 L .76917 .80673 L closepath p 1 .874 .727 r
F P
0 g
s
.80457 .78684 m .76917 .80673 L .8118 .8374 L closepath p 1 .824 .689 r
F P
0 g
s
.42545 1.22846 m .3661 1.19097 L .40571 1.2283 L closepath p .064 .323 .791 r
F P
0 g
s
.42545 1.22846 m .40571 1.2283 L .473 1.2563 L closepath p .045 .425 .804 r
F P
0 g
s
.2387 .75063 m .23276 .78494 L .22901 .824 L closepath p .027 .027 .027 r
F P
0 g
s
.84477 .91851 m .81697 .87902 L .81934 .92745 L closepath p 1 .897 .583 r
F P
0 g
s
.84477 .91851 m .81934 .92745 L .84788 .96064 L closepath p 1 .886 .573 r
F P
0 g
s
.75975 .71726 m .73728 .69823 L .7388 .71319 L closepath p 1 .832 .571 r
F P
0 g
s
.75975 .71726 m .7388 .71319 L .76216 .73572 L closepath p 1 .843 .584 r
F P
0 g
s
.30786 .45466 m .32377 .44169 L .33568 .44302 L closepath p .851 1 1 r
F P
0 g
s
.34485 .43465 m .33568 .44302 L .32489 .43869 L closepath p .9 1 1 r
F P
0 g
s
.32377 .44169 m .32489 .43869 L .33568 .44302 L closepath p 1 1 1 r
F P
0 g
s
.31293 .46226 m .28606 .47246 L .29029 .46191 L closepath p .993 1 1 r
F P
0 g
s
.31293 .46226 m .29029 .46191 L .30786 .45466 L closepath p 1 1 .986 r
F P
0 g
s
.66302 .55775 m .69275 .59382 L .68871 .57814 L closepath p 1 .782 .572 r
F P
0 g
s
.68871 .57814 m .6663 .55217 L .66302 .55775 L closepath p 1 .718 .468 r
F P
0 g
s
.23675 .8468 m .23833 .86073 L .22901 .824 L closepath p .882 .598 .055 r
F P
0 g
s
.52894 1.26046 m .5 1.26236 L .54072 1.27072 L closepath p .353 .842 1 r
F P
0 g
s
.52894 1.26046 m .54072 1.27072 L .56638 1.27136 L closepath p .252 .763 .982 \
r
F P
0 g
s
.64742 .5313 m .6663 .55217 L .6698 .55108 L closepath p 1 .808 .666 r
F P
0 g
s
.27359 1.02809 m .26205 .99923 L .28121 .99359 L closepath p 1 .843 .468 r
F P
0 g
s
.27359 1.02809 m .28121 .99359 L .28227 1.0286 L closepath p 1 .856 .428 r
F P
0 g
s
.28121 .99359 m .26205 .99923 L .25541 .9625 L closepath p 1 .809 .436 r
F P
0 g
s
.28121 .99359 m .25541 .9625 L .25299 .94653 L closepath p 1 .825 .46 r
F P
0 g
s
.25237 .92741 m .28121 .99359 L .25111 .91936 L closepath p .056 .056 .496 r
F P
0 g
s
.25111 .91936 m .24539 .90286 L .25237 .92741 L closepath p .028 .028 .305 r
F P
0 g
s
.28121 .99359 m .25299 .94653 L .25111 .91936 L closepath p 1 .917 .632 r
F P
0 g
s
.30312 1.07676 m .31618 1.12053 L .30685 1.10563 L closepath p .869 .524 .059 \
r
F P
0 g
s
.30312 1.07676 m .30685 1.10563 L .28948 1.06636 L closepath p .932 .507 .097 \
r
F P
0 g
s
.27938 .56249 m .27505 .55458 L .26821 .59031 L closepath p .132 .132 .435 r
F P
0 g
s
.26821 .59031 m .26387 .61985 L .27938 .56249 L closepath p .058 .058 .058 r
F P
0 g
s
.73017 .66527 m .70123 .63356 L .70724 .66655 L closepath p 1 .784 .482 r
F P
0 g
s
.73017 .66527 m .70724 .66655 L .73728 .69823 L closepath p 1 .782 .479 r
F P
0 g
s
.25052 .90168 m .24539 .90286 L .24374 .88845 L closepath p 1 1 .804 r
F P
0 g
s
.66302 .55775 m .6663 .55217 L .64955 .54078 L closepath p 1 .694 .458 r
F P
0 g
s
.64742 .5313 m .64955 .54078 L .6663 .55217 L closepath p 1 .761 .562 r
F P
0 g
s
.42471 .42091 m .56966 .46747 L .54838 .4667 L closepath p .376 .789 1 r
F P
0 g
s
.54838 .4667 m .56966 .46747 L .60946 .48897 L closepath p .436 .798 1 r
F P
0 g
s
.33284 .43357 m .35461 .42939 L .34485 .43465 L closepath p .711 1 1 r
F P
0 g
s
.34485 .43465 m .32489 .43869 L .33284 .43357 L closepath p .797 1 1 r
F P
0 g
s
.30786 .45466 m .29029 .46191 L .30155 .44938 L closepath p .936 1 1 r
F P
0 g
s
.30786 .45466 m .30155 .44938 L .31546 .44331 L closepath p .958 1 1 r
F P
0 g
s
.30786 .45466 m .31546 .44331 L .32377 .44169 L closepath p .895 1 1 r
F P
0 g
s
.60946 .48897 m .61783 .50886 L .62917 .5161 L closepath p 1 .637 .474 r
F P
0 g
s
.56966 .46747 m .61783 .50886 L .60946 .48897 L closepath p 1 .636 .456 r
F P
0 g
s
.34217 1.13133 m .3661 1.19097 L .34355 1.1618 L closepath p .986 .567 .131 r
F P
0 g
s
.34217 1.13133 m .34355 1.1618 L .32506 1.13118 L closepath p .977 .499 .184 \
r
F P
0 g
s
.32506 1.13118 m .34355 1.1618 L .33908 1.15418 L p .006 .006 .384 r
F P
0 g
s
.31938 1.12185 m .34217 1.13133 L .32506 1.13118 L p .91 .4 .142 r
F P
0 g
s
.33908 1.15418 m .32222 1.12547 L .32506 1.13118 L p .006 .006 .384 r
F P
0 g
s
.32506 1.13118 m .31618 1.12053 L .31938 1.12185 L p .91 .4 .142 r
F P
0 g
s
.76216 .73572 m .7388 .71319 L .74515 .73756 L closepath p 1 .758 .447 r
F P
0 g
s
.76216 .73572 m .74515 .73756 L .76827 .75977 L closepath p 1 .735 .424 r
F P
0 g
s
.70123 .63356 m .69275 .59382 L .66474 .6023 L closepath p 1 .773 .518 r
F P
0 g
s
.69275 .59382 m .65559 .56423 L .66474 .6023 L closepath p 1 .723 .473 r
F P
0 g
s
.66302 .55775 m .65559 .56423 L .69275 .59382 L closepath p 1 .777 .56 r
F P
0 g
s
.28606 .47246 m .27686 .48995 L .28162 .47696 L closepath p .852 .94 .514 r
F P
0 g
s
.28606 .47246 m .28162 .47696 L .29029 .46191 L closepath p .817 .934 .517 r
F P
0 g
s
.62917 .5161 m .64955 .54078 L .64742 .5313 L closepath p 1 .761 .565 r
F P
0 g
s
.34389 .42733 m .3595 .42538 L .35461 .42939 L closepath p .894 1 1 r
F P
0 g
s
.33284 .43357 m .34389 .42733 L .35461 .42939 L closepath p .791 1 1 r
F P
0 g
s
.60603 1.24674 m .52894 1.26046 L .56638 1.27136 L closepath p .588 .981 1 r
F P
0 g
s
.60603 1.24674 m .56638 1.27136 L .63543 1.25896 L closepath p .59 .979 1 r
F P
0 g
s
.24778 .90662 m .24539 .90286 L .25111 .91936 L closepath p .031 .031 .314 r
F P
0 g
s
.24539 .90286 m .24778 .90662 L .2468 .89557 L closepath p .047 .047 .265 r
F P
0 g
s
.24539 .90286 m .2468 .89557 L .24374 .88845 L closepath p .049 .049 .283 r
F P
0 g
s
.27938 .56249 m .28232 .53977 L .27505 .55458 L closepath p .143 .143 .429 r
F P
0 g
s
.28232 .53977 m .2793 .51616 L .27505 .55458 L closepath p .35 .338 .789 r
F P
0 g
s
.84788 .96064 m .81934 .92745 L .81863 .97936 L closepath p 1 .929 .647 r
F P
0 g
s
.84788 .96064 m .81863 .97936 L .84579 1.01052 L closepath p 1 .95 .663 r
F P
0 g
s
.34814 .42433 m .36518 .42254 L .3595 .42538 L closepath p .707 1 1 r
F P
0 g
s
.34389 .42733 m .34814 .42433 L .3595 .42538 L closepath p .786 1 1 r
F P
0 g
s
.66302 .55775 m .64955 .54078 L .65239 .55013 L closepath p 1 .75 .588 r
F P
0 g
s
.65239 .55013 m .64955 .54078 L .62152 .52135 L closepath p 1 .751 .57 r
F P
0 g
s
.62917 .5161 m .62152 .52135 L .64955 .54078 L closepath p 1 .791 .637 r
F P
0 g
s
.65239 .55013 m .65559 .56423 L .66302 .55775 L closepath p 1 .706 .514 r
F P
0 g
s
.24374 .88845 m .2468 .89557 L .24939 .87815 L closepath p .086 .086 .262 r
F P
0 g
s
.24374 .88845 m .24939 .87815 L .24177 .8694 L closepath p .143 .19 .745 r
F P
0 g
s
.24374 .88845 m .24177 .8694 L .23833 .86073 L closepath p .066 .066 .388 r
F P
0 g
s
.31546 .44331 m .30155 .44938 L .30561 .44539 L closepath p .771 1 .959 r
F P
0 g
s
.31546 .44331 m .30561 .44539 L .32377 .44169 L closepath p .005 .005 .005 r
F P
0 g
s
.37419 .4221 m .36518 .42254 L .35382 .42221 L closepath p .133 .687 .804 r
F P
0 g
s
.34814 .42433 m .35382 .42221 L .36518 .42254 L closepath p .591 1 1 r
F P
0 g
s
.27686 .48995 m .2793 .51616 L .28395 .50538 L closepath p .241 .226 .706 r
F P
0 g
s
.27686 .48995 m .28395 .50538 L .28162 .47696 L closepath p .223 .216 .703 r
F P
0 g
s
.37419 .4221 m .35382 .42221 L .36061 .42109 L closepath p .462 .923 1 r
F P
0 g
s
.38747 .4188 m .37419 .4221 L .36061 .42109 L closepath p .576 .992 1 r
F P
0 g
s
.28948 1.06636 m .27359 1.02809 L .28227 1.0286 L closepath p 1 .734 .298 r
F P
0 g
s
.23833 .86073 m .24177 .8694 L .24539 .85663 L closepath p .133 .133 .483 r
F P
0 g
s
.23833 .86073 m .24539 .85663 L .22901 .824 L closepath p .146 .165 .72 r
F P
0 g
s
.32377 .44169 m .30561 .44539 L .31299 .44021 L closepath p .816 1 1 r
F P
0 g
s
.32377 .44169 m .31299 .44021 L .32489 .43869 L closepath p 1 1 1 r
F P
0 g
s
.29029 .46191 m .28162 .47696 L .293 .46354 L closepath p .086 .086 .086 r
F P
0 g
s
.29029 .46191 m .293 .46354 L .30155 .44938 L closepath p .085 .085 .085 r
F P
0 g
s
.42471 .42091 m .43503 .41954 L .40354 .41782 L closepath p .312 .071 .071 r
F P
0 g
s
.61783 .50886 m .62152 .52135 L .62917 .5161 L closepath p 1 .691 .564 r
F P
0 g
s
.38747 .4188 m .36061 .42109 L .374 .41824 L closepath p .435 .907 .975 r
F P
0 g
s
.374 .41824 m .40354 .41782 L .38747 .4188 L closepath p .194 .734 .85 r
F P
0 g
s
.81697 .87902 m .8118 .8374 L .77111 .83217 L closepath p 1 .839 .715 r
F P
0 g
s
.8118 .8374 m .76917 .80673 L .77111 .83217 L closepath p 1 .838 .715 r
F P
0 g
s
.40354 .41782 m .374 .41824 L .3801 .41719 L p .49 .938 1 r
F P
0 g
s
.3801 .41719 m .38482 .41638 L .3849 .41639 L p .49 .938 1 r
F P
0 g
s
.39003 .41678 m .40354 .41782 L p .38887 .41743 L .49 .938 1 r
F P
0 g
s
.42471 .42091 m .40354 .41782 L .38976 .41714 L closepath p .181 .027 .027 r
F P
0 g
s
.3849 .41639 m .39003 .41678 L p .3801 .41719 L .49 .938 1 r
F P
0 g
s
.3801 .41719 m .39003 .41678 L .38887 .41743 L .49 .938 1 r
F
.40354 .41782 m .38482 .41638 L .38976 .41714 L closepath p .306 .049 .049 r
F P
0 g
s
.42471 .42091 m .38976 .41714 L .39327 .41499 L p .887 .995 1 r
F P
0 g
s
.39327 .41499 m .39591 .41337 L .39865 .41409 L p .887 .995 1 r
F P
0 g
s
.39865 .41409 m .42471 .42091 L p .39327 .41499 L .887 .995 1 r
F P
0 g
s
.38976 .41714 m .39491 .41463 L .39569 .41365 L p .02 .02 .02 r
F P
0 g
s
.39569 .41365 m .39591 .41337 L .39327 .41499 L p .02 .02 .02 r
F P
0 g
s
.39327 .41499 m .38976 .41714 L p .39569 .41365 L .02 .02 .02 r
F P
0 g
s
.32489 .43869 m .31299 .44021 L .32201 .43422 L closepath p .755 1 1 r
F P
0 g
s
.32489 .43869 m .32201 .43422 L .33284 .43357 L closepath p .745 1 1 r
F P
0 g
s
.62349 .53073 m .65559 .56423 L .65239 .55013 L closepath p 1 .707 .508 r
F P
0 g
s
.65239 .55013 m .62152 .52135 L .62349 .53073 L closepath p 1 .777 .621 r
F P
0 g
s
.73728 .69823 m .70724 .66655 L .71012 .68554 L closepath p 1 .837 .573 r
F P
0 g
s
.73728 .69823 m .71012 .68554 L .7388 .71319 L closepath p 1 .832 .566 r
F P
0 g
s
.57413 .4828 m .62152 .52135 L .61783 .50886 L closepath p 1 .695 .547 r
F P
0 g
s
.56966 .46747 m .57413 .4828 L .61783 .50886 L closepath p 1 .634 .451 r
F P
0 g
s
.22901 .824 m .24539 .85663 L .25866 .82709 L closepath p .398 .391 .826 r
F P
0 g
s
.22901 .824 m .25866 .82709 L .24111 .79275 L closepath p .39 .419 .863 r
F P
0 g
s
.22901 .824 m .24111 .79275 L .2387 .75063 L closepath p .149 .149 .52 r
F P
0 g
s
.74675 .78454 m .76917 .80673 L .76827 .75977 L closepath p 1 .895 .677 r
F P
0 g
s
.74675 .78454 m .76827 .75977 L .74592 .75467 L closepath p 1 .894 .677 r
F P
0 g
s
.74592 .75467 m .76827 .75977 L .74515 .73756 L closepath p 1 .862 .636 r
F P
0 g
s
.42471 .42091 m .43508 .41733 L .56966 .46747 L closepath p .383 .792 1 r
F P
0 g
s
.43508 .41733 m .57413 .4828 L .56966 .46747 L closepath p 1 .612 .524 r
F P
0 g
s
.28742 .52806 m .2868 .5152 L .2793 .51616 L closepath p .391 .323 .747 r
F P
0 g
s
.28742 .52806 m .2793 .51616 L .28232 .53977 L closepath p .265 .272 .752 r
F P
0 g
s
.28395 .50538 m .2793 .51616 L .2868 .5152 L closepath p .406 .412 .847 r
F P
0 g
s
.33086 .43142 m .33284 .43357 L .32405 .4341 L p .462 .922 .931 r
F P
0 g
s
.33284 .43357 m .33086 .43142 L .34389 .42733 L closepath p .598 .994 .901 r
F P
0 g
s
.34389 .42733 m .33086 .43142 L .33767 .42783 L closepath p .526 .949 .864 r
F P
0 g
s
.33086 .43142 m .33246 .43848 L .33711 .42896 L p .33 .172 .266 r
F P
0 g
s
.33711 .42896 m .33767 .42783 L .33086 .43142 L p .33 .172 .266 r
F P
0 g
s
.32405 .4341 m .32201 .43422 L .33086 .43142 L p .462 .922 .931 r
F P
0 g
s
.28709 .54971 m .28742 .52806 L .28232 .53977 L closepath p .153 .153 .529 r
F P
0 g
s
.28709 .54971 m .28232 .53977 L .27938 .56249 L closepath p .17 .17 .533 r
F P
0 g
s
.28162 .47696 m .28395 .50538 L .29255 .49044 L closepath p .371 .332 .77 r
F P
0 g
s
.28162 .47696 m .29255 .49044 L .293 .46354 L closepath p .478 .365 .747 r
F P
0 g
s
.30155 .44938 m .293 .46354 L .29596 .45839 L closepath p .637 .816 .389 r
F P
0 g
s
.30155 .44938 m .29596 .45839 L .30561 .44539 L closepath p .055 .055 .055 r
F P
0 g
s
.34389 .42733 m .33767 .42783 L .34814 .42433 L closepath p .637 1 .881 r
F P
0 g
s
.25111 .91936 m .25299 .94653 L .25025 .93134 L closepath p .899 .69 .146 r
F P
0 g
s
.25111 .91936 m .25025 .93134 L .24972 .91934 L closepath p .81 .66 .099 r
F P
0 g
s
.25111 .91936 m .2487 .90842 L .24778 .90662 L closepath p .025 .025 .285 r
F P
0 g
s
.25111 .91936 m .24972 .91934 L .2487 .90842 L closepath p .8 .571 .028 r
F P
0 g
s
.7388 .71319 m .71012 .68554 L .71416 .70901 L closepath p 1 .804 .519 r
F P
0 g
s
.7388 .71319 m .71416 .70901 L .74515 .73756 L closepath p 1 .763 .473 r
F P
0 g
s
.66726 .62998 m .70123 .63356 L .66474 .6023 L closepath p 1 .803 .567 r
F P
0 g
s
.66726 .62998 m .70724 .66655 L .70123 .63356 L closepath p 1 .788 .55 r
F P
0 g
s
.62349 .53073 m .62152 .52135 L .60493 .52312 L closepath p 1 .756 .679 r
F P
0 g
s
.60493 .52312 m .62152 .52135 L .57524 .49812 L closepath p 1 .754 .677 r
F P
0 g
s
.57413 .4828 m .57524 .49812 L .62152 .52135 L closepath p 1 .781 .721 r
F P
0 g
s
.84579 1.01052 m .81863 .97936 L .81113 1.03513 L closepath p 1 .993 .743 r
F P
0 g
s
.84579 1.01052 m .81113 1.03513 L .83953 1.06157 L closepath p 1 .993 .744 r
F P
0 g
s
.27267 .71772 m .2387 .75063 L .27118 .76041 L closepath p .461 .346 .733 r
F P
0 g
s
.24111 .79275 m .27118 .76041 L .2387 .75063 L closepath p .465 .338 .719 r
F P
0 g
s
.27267 .71772 m .25718 .7007 L .2387 .75063 L closepath p .398 .259 .657 r
F P
0 g
s
.24177 .8694 m .24939 .87815 L .2516 .86697 L closepath p .292 .286 .755 r
F P
0 g
s
.24177 .8694 m .2516 .86697 L .24539 .85663 L closepath p .313 .364 .836 r
F P
0 g
s
.65559 .56423 m .62942 .56402 L .6313 .5834 L closepath p 1 .758 .594 r
F P
0 g
s
.65559 .56423 m .6313 .5834 L .66474 .6023 L closepath p 1 .712 .562 r
F P
0 g
s
.62455 .54463 m .62942 .56402 L .65559 .56423 L closepath p 1 .738 .572 r
F P
0 g
s
.62349 .53073 m .62455 .54463 L .65559 .56423 L closepath p 1 .741 .577 r
F P
0 g
s
.27938 .56249 m .26387 .61985 L .28134 .58422 L closepath p .168 .168 .469 r
F P
0 g
s
.27938 .56249 m .28134 .58422 L .28709 .54971 L closepath p .155 .155 .442 r
F P
0 g
s
.29981 .51319 m .29255 .49044 L .28395 .50538 L closepath p .52 .512 .903 r
F P
0 g
s
.29981 .51319 m .28395 .50538 L .2868 .5152 L closepath p .544 .505 .882 r
F P
0 g
s
.34814 .42433 m .33767 .42783 L .34347 .42553 L closepath p .322 .804 .727 r
F P
0 g
s
.34814 .42433 m .34347 .42553 L .35382 .42221 L closepath p .3 .789 .722 r
F P
0 g
s
.32506 1.13118 m .30685 1.10563 L .31618 1.12053 L closepath p .03 .03 .546 r
F P
0 g
s
.30561 .44539 m .29596 .45839 L .30509 .45211 L closepath p .138 .138 .138 r
F P
0 g
s
.30561 .44539 m .30509 .45211 L .31299 .44021 L closepath p .137 .137 .137 r
F P
0 g
s
.45062 1.23007 m .42545 1.22846 L .473 1.2563 L closepath p .405 .831 1 r
F P
0 g
s
.45062 1.23007 m .473 1.2563 L .5 1.26236 L closepath p .24 .699 1 r
F P
0 g
s
.41639 .41438 m .43508 .41733 L .42471 .42091 L closepath p .383 .741 1 r
F P
0 g
s
.42471 .42091 m .39591 .41337 L .41639 .41438 L closepath p .909 .993 1 r
F P
0 g
s
.27267 .71772 m .2739 .68979 L .25718 .7007 L closepath p .331 .247 .684 r
F P
0 g
s
.26387 .61985 m .25718 .7007 L .2739 .68979 L closepath p .391 .353 .784 r
F P
0 g
s
.35382 .42221 m .34347 .42553 L .34562 .42486 L closepath p .003 .003 .003 r
F P
0 g
s
.35382 .42221 m .34562 .42486 L .35465 .42255 L closepath p .018 .018 .018 r
F P
0 g
s
.35382 .42221 m .35465 .42255 L .36061 .42109 L closepath p .016 .016 .016 r
F P
0 g
s
.2868 .5152 m .28742 .52806 L .29981 .51319 L closepath p .527 .431 .799 r
F P
0 g
s
.36061 .42109 m .35465 .42255 L .36283 .419 L closepath p .385 .85 .767 r
F P
0 g
s
.36061 .42109 m .36283 .419 L .374 .41824 L closepath p .307 .816 .857 r
F P
0 g
s
.2468 .89557 m .24778 .90662 L .2487 .90842 L closepath p .043 .043 .248 r
F P
0 g
s
.60493 .52312 m .62455 .54463 L .62349 .53073 L closepath p 1 .71 .609 r
F P
0 g
s
.81934 .92745 m .81697 .87902 L .77086 .86447 L closepath p 1 .864 .758 r
F P
0 g
s
.76868 .85185 m .77086 .86447 L .81697 .87902 L closepath p 1 .792 .657 r
F P
0 g
s
.77111 .83217 m .76868 .85185 L .81697 .87902 L closepath p 1 .865 .781 r
F P
0 g
s
.31299 .44021 m .30509 .45211 L .31213 .44684 L closepath p .132 .132 .132 r
F P
0 g
s
.31299 .44021 m .31213 .44684 L .32201 .43422 L closepath p .142 .138 .138 r
F P
0 g
s
.63543 1.25896 m .68126 1.24139 L .64879 1.22888 L closepath p .806 1 1 r
F P
0 g
s
.63543 1.25896 m .64879 1.22888 L .60603 1.24674 L closepath p .806 1 1 r
F P
0 g
s
.2504 .94174 m .25025 .93134 L .25299 .94653 L closepath p .825 .606 .04 r
F P
0 g
s
.2504 .94174 m .25299 .94653 L .25541 .9625 L closepath p .807 .566 .03 r
F P
0 g
s
.48098 .44232 m .57524 .49812 L .57413 .4828 L closepath p 1 .764 .733 r
F P
0 g
s
.43508 .41733 m .48098 .44232 L .57413 .4828 L closepath p .75 .187 .124 r
F P
0 g
s
.29863 .53691 m .29981 .51319 L .28742 .52806 L closepath p .484 .375 .755 r
F P
0 g
s
.29863 .53691 m .28742 .52806 L .28709 .54971 L closepath p .486 .375 .754 r
F P
0 g
s
.374 .41824 m .36283 .419 L .37504 .41708 L closepath p .215 .748 .807 r
F P
0 g
s
.374 .41824 m .37504 .41708 L .38482 .41638 L closepath p .223 .754 .803 r
F P
0 g
s
.293 .46354 m .29255 .49044 L .29419 .48656 L closepath p .165 .165 .566 r
F P
0 g
s
.293 .46354 m .29419 .48656 L .29703 .48188 L closepath p .248 .186 .632 r
F P
0 g
s
.293 .46354 m .29703 .48188 L .29596 .45839 L closepath p .175 .175 .608 r
F P
0 g
s
.30685 1.10563 m .28531 1.05081 L .28948 1.06636 L closepath p .019 .019 .345 \
r
F P
0 g
s
.27359 1.02809 m .28948 1.06636 L .28531 1.05081 L closepath p .021 .021 .354 \
r
F P
0 g
s
.38482 .41638 m .38557 .41561 L .38976 .41714 L closepath p .026 .572 .796 r
F P
0 g
s
.38976 .41714 m .38557 .41561 L .39491 .41463 L closepath p .292 .808 .898 r
F P
0 g
s
.38482 .41638 m .37504 .41708 L .38557 .41561 L closepath p .17 .713 .774 r
F P
0 g
s
.25262 .90024 m .24939 .87815 L .2468 .89557 L closepath p .149 .149 .542 r
F P
0 g
s
.2468 .89557 m .2487 .90842 L .25262 .90024 L closepath p .142 .142 .567 r
F P
0 g
s
.41639 .41438 m .41628 .41516 L .43508 .41733 L closepath p .179 .024 .024 r
F P
0 g
s
.39591 .41337 m .41628 .41516 L .41639 .41438 L closepath p .129 .024 .024 r
F P
0 g
s
.32506 1.13118 m .32222 1.12547 L .30685 1.10563 L closepath p .015 .015 .478 \
r
F P
0 g
s
.32201 .43422 m .31213 .44684 L .3246 .44161 L closepath p .477 .189 .331 r
F P
0 g
s
.32201 .43422 m .3246 .44161 L .33086 .43142 L closepath p .536 .191 .314 r
F P
0 g
s
.39591 .41337 m .39491 .41463 L .41628 .41516 L closepath p .192 .047 .047 r
F P
0 g
s
.24539 .85663 m .2516 .86697 L .2609 .83085 L closepath p .444 .431 .85 r
F P
0 g
s
.24539 .85663 m .2609 .83085 L .25866 .82709 L closepath p .568 .585 .959 r
F P
0 g
s
.74761 .80167 m .77111 .83217 L .76917 .80673 L closepath p 1 .867 .626 r
F P
0 g
s
.76917 .80673 m .74675 .78454 L .74761 .80167 L closepath p 1 .867 .626 r
F P
0 g
s
.2739 .68979 m .28803 .67475 L .26387 .61985 L closepath p .241 .239 .723 r
F P
0 g
s
.28005 .6468 m .26387 .61985 L .28803 .67475 L closepath p .132 .132 .625 r
F P
0 g
s
.27589 .6323 m .26387 .61985 L .28005 .6468 L closepath p .145 .145 .64 r
F P
0 g
s
.26387 .61985 m .27589 .6323 L .2754 .62436 L closepath p .17 .17 .605 r
F P
0 g
s
.26387 .61985 m .2754 .62436 L .27565 .61784 L closepath p .173 .173 .533 r
F P
0 g
s
.26387 .61985 m .27565 .61784 L .27762 .61559 L closepath p .12 .12 .12 r
F P
0 g
s
.26387 .61985 m .27762 .61559 L .28025 .60888 L closepath p .182 .165 .344 r
F P
0 g
s
.26387 .61985 m .28025 .60888 L .282 .60267 L closepath p .27 .183 .483 r
F P
0 g
s
.26387 .61985 m .282 .60267 L .28196 .60041 L closepath p .338 .195 .567 r
F P
0 g
s
.26387 .61985 m .28196 .60041 L .28206 .59535 L closepath p .495 .371 .743 r
F P
0 g
s
.26387 .61985 m .28206 .59535 L .28163 .58815 L closepath p .595 .495 .837 r
F P
0 g
s
.26387 .61985 m .28163 .58815 L .28134 .58422 L closepath p .574 .472 .82 r
F P
0 g
s
.71416 .70901 m .74592 .75467 L .74515 .73756 L closepath p 1 .862 .638 r
F P
0 g
s
.57435 .50985 m .57652 .52765 L .60493 .52312 L closepath p 1 .764 .717 r
F P
0 g
s
.60493 .52312 m .57652 .52765 L .62455 .54463 L closepath p 1 .738 .691 r
F P
0 g
s
.60493 .52312 m .57524 .49812 L .57435 .50985 L closepath p 1 .79 .76 r
F P
0 g
s
.28709 .54971 m .28134 .58422 L .28872 .5647 L closepath p .106 .106 .259 r
F P
0 g
s
.28709 .54971 m .28872 .5647 L .29863 .53691 L closepath p .157 .157 .335 r
F P
0 g
s
.29596 .45839 m .29703 .48188 L .30593 .47392 L closepath p .599 .452 .776 r
F P
0 g
s
.29596 .45839 m .30593 .47392 L .30509 .45211 L closepath p .64 .468 .765 r
F P
0 g
s
.25541 .9625 m .26205 .99923 L .2504 .94174 L closepath p .771 .523 .021 r
F P
0 g
s
.33086 .43142 m .3246 .44161 L .33246 .43848 L closepath p .436 .179 .254 r
F P
0 g
s
.2487 .90842 m .24972 .91934 L .25401 .90884 L closepath p .127 .127 .347 r
F P
0 g
s
.2487 .90842 m .25401 .90884 L .25262 .90024 L closepath p .128 .128 .471 r
F P
0 g
s
.25491 .88447 m .2516 .86697 L .24939 .87815 L closepath p .203 .2 .693 r
F P
0 g
s
.25262 .90024 m .25491 .88447 L .24939 .87815 L closepath p .315 .226 .666 r
F P
0 g
s
.80939 1.13494 m .83953 1.06157 L .81113 1.03513 L closepath p 1 1 .918 r
F P
0 g
s
.29255 .49044 m .29981 .51319 L .30113 .51085 L closepath p .436 .456 .882 r
F P
0 g
s
.29255 .49044 m .30113 .51085 L .29419 .48656 L closepath p .484 .485 .892 r
F P
0 g
s
.57435 .50985 m .57524 .49812 L .52571 .47935 L closepath p 1 .784 .761 r
F P
0 g
s
.48098 .44232 m .52571 .47935 L .57524 .49812 L closepath p 1 .752 .711 r
F P
0 g
s
.57652 .52765 m .62942 .56402 L .62455 .54463 L closepath p 1 .712 .652 r
F P
0 g
s
.47939 1.22866 m .45062 1.23007 L .5 1.26236 L closepath p .579 .923 1 r
F P
0 g
s
.47939 1.22866 m .5 1.26236 L .52894 1.26046 L closepath p .614 .947 1 r
F P
0 g
s
.68126 1.24139 m .70859 1.22677 L .67806 1.21532 L closepath p .892 1 1 r
F P
0 g
s
.68126 1.24139 m .67806 1.21532 L .64879 1.22888 L closepath p .853 1 1 r
F P
0 g
s
.27359 1.02809 m .28531 1.05081 L .26836 1.02528 L closepath p .738 .264 .05 \
r
F P
0 g
s
.26836 1.02528 m .26205 .99923 L .27359 1.02809 L closepath p .944 .609 .095 \
r
F P
0 g
s
.33767 .42783 m .33246 .43848 L .33976 .43627 L closepath p .562 .2 .396 r
F P
0 g
s
.33767 .42783 m .33976 .43627 L .34347 .42553 L closepath p .528 .199 .4 r
F P
0 g
s
.29419 .48656 m .30113 .51085 L .30546 .50642 L closepath p .697 .629 .932 r
F P
0 g
s
.29419 .48656 m .30546 .50642 L .29703 .48188 L closepath p .575 .567 .936 r
F P
0 g
s
.3661 1.19097 m .42545 1.22846 L .40695 1.19812 L closepath p .196 .501 .978 \
r
F P
0 g
s
.3661 1.19097 m .40695 1.19812 L .39443 1.19215 L closepath p .142 .526 .982 \
r
F P
0 g
s
.3661 1.19097 m .39443 1.19215 L .34355 1.1618 L closepath p .158 .5 .974 r
F P
0 g
s
.30509 .45211 m .30593 .47392 L .31214 .46778 L closepath p .551 .399 .74 r
F P
0 g
s
.30509 .45211 m .31214 .46778 L .31213 .44684 L closepath p .62 .434 .733 r
F P
0 g
s
.29981 .51319 m .29863 .53691 L .3019 .53363 L closepath p .494 .383 .759 r
F P
0 g
s
.29981 .51319 m .3019 .53363 L .30099 .51192 L closepath p .167 .167 .608 r
F P
0 g
s
.29981 .51319 m .30099 .51192 L .30113 .51085 L closepath p .145 .145 .432 r
F P
0 g
s
.71012 .68554 m .70724 .66655 L .66726 .62998 L closepath p 1 .835 .634 r
F P
0 g
s
.25025 .93134 m .2504 .94174 L .2518 .92851 L closepath p .03 .03 .053 r
F P
0 g
s
.25025 .93134 m .2518 .92851 L .24972 .91934 L closepath p .041 .041 .22 r
F P
0 g
s
.25539 .87479 m .2609 .83085 L .2516 .86697 L closepath p .224 .209 .694 r
F P
0 g
s
.25491 .88447 m .25539 .87479 L .2516 .86697 L closepath p .195 .194 .689 r
F P
0 g
s
.27118 .76041 m .24111 .79275 L .25866 .82709 L closepath p .615 .542 .881 r
F P
0 g
s
.34347 .42553 m .33976 .43627 L .34198 .43494 L closepath p .49 .195 .383 r
F P
0 g
s
.34347 .42553 m .34198 .43494 L .34562 .42486 L closepath p .622 .246 .421 r
F P
0 g
s
.34562 .42486 m .34198 .43494 L .35015 .43211 L closepath p .563 .2 .394 r
F P
0 g
s
.34562 .42486 m .35015 .43211 L .35465 .42255 L closepath p .622 .211 .351 r
F P
0 g
s
.70859 1.22677 m .74024 1.20689 L .70519 1.1946 L closepath p .956 1 1 r
F P
0 g
s
.70859 1.22677 m .70519 1.1946 L .67806 1.21532 L closepath p .991 1 1 r
F P
0 g
s
.35465 .42255 m .35015 .43211 L .36333 .4278 L closepath p .509 .187 .294 r
F P
0 g
s
.35465 .42255 m .36333 .4278 L .36283 .419 L closepath p .487 .193 .368 r
F P
0 g
s
.71416 .70901 m .7184 .75509 L .74592 .75467 L closepath p 1 .861 .635 r
F P
0 g
s
.74592 .75467 m .7184 .75509 L .74675 .78454 L closepath p 1 .895 .673 r
F P
0 g
s
.30113 .51085 m .30099 .51192 L .30441 .51265 L closepath p .528 .379 .729 r
F P
0 g
s
.30113 .51085 m .30441 .51265 L .30546 .50642 L closepath p .521 .382 .74 r
F P
0 g
s
.30099 .51192 m .3019 .53363 L .30617 .52912 L closepath p .516 .417 .79 r
F P
0 g
s
.30099 .51192 m .30617 .52912 L .30441 .51265 L closepath p .516 .418 .79 r
F P
0 g
s
.57652 .52765 m .57435 .50985 L .55118 .5032 L closepath p 1 .758 .723 r
F P
0 g
s
.57435 .50985 m .52571 .47935 L .55118 .5032 L closepath p 1 .784 .761 r
F P
0 g
s
.31213 .44684 m .31214 .46778 L .32323 .46222 L closepath p .72 .513 .759 r
F P
0 g
s
.31213 .44684 m .32323 .46222 L .3246 .44161 L closepath p .738 .518 .749 r
F P
0 g
s
.29703 .48188 m .30546 .50642 L .3141 .49899 L closepath p .691 .642 .951 r
F P
0 g
s
.29703 .48188 m .3141 .49899 L .30593 .47392 L closepath p .733 .652 .933 r
F P
0 g
s
.74761 .80167 m .7597 .84071 L .77111 .83217 L closepath p 1 .888 .669 r
F P
0 g
s
.77111 .83217 m .7597 .84071 L .76868 .85185 L closepath p 1 .943 .707 r
F P
0 g
s
.28872 .5647 m .28875 .56113 L .29863 .53691 L closepath p 1 .995 .579 r
F P
0 g
s
.29863 .53691 m .28875 .56113 L .29245 .55535 L closepath p .076 .076 .076 r
F P
0 g
s
.29863 .53691 m .29245 .55535 L .3019 .53363 L closepath p .125 .125 .129 r
F P
0 g
s
.43856 .43041 m .48098 .44232 L .43508 .41733 L closepath p .967 .56 .495 r
F P
0 g
s
.41628 .41516 m .43856 .43041 L .43508 .41733 L closepath p .937 .541 .518 r
F P
0 g
s
.36283 .419 m .36333 .4278 L .37321 .4268 L closepath p .694 .276 .377 r
F P
0 g
s
.36283 .419 m .37321 .4268 L .37504 .41708 L closepath p .686 .285 .407 r
F P
0 g
s
.28134 .58422 m .28163 .58815 L .28803 .5725 L closepath p .172 .172 .642 r
F P
0 g
s
.28803 .5725 m .29016 .57796 L .28924 .57392 L p .102 .102 .5 r
F P
0 g
s
.28803 .5725 m .28872 .5647 L .28646 .57067 L p .14 .14 .432 r
F P
0 g
s
.28872 .5647 m .28803 .5725 L .28816 .56913 L closepath p .782 .687 .134 r
F P
0 g
s
.28924 .57392 m .28816 .56913 L .28803 .5725 L p .102 .102 .5 r
F P
0 g
s
.28646 .57067 m .28134 .58422 L .28803 .5725 L p .14 .14 .432 r
F P
0 g
s
.30546 .50642 m .30441 .51265 L .31922 .50947 L closepath p .739 .58 .836 r
F P
0 g
s
.30546 .50642 m .31922 .50947 L .3141 .49899 L closepath p .703 .661 .965 r
F P
0 g
s
.81934 .92745 m .76985 .90109 L .76708 .95775 L closepath p 1 .86 .756 r
F P
0 g
s
.81934 .92745 m .76708 .95775 L .81863 .97936 L closepath p 1 .884 .774 r
F P
0 g
s
.81934 .92745 m .77086 .86447 L .76985 .90109 L closepath p 1 .867 .766 r
F P
0 g
s
.28163 .58815 m .28206 .59535 L .29067 .58135 L closepath p .281 .248 .714 r
F P
0 g
s
.28163 .58815 m .29067 .58135 L .28803 .5725 L closepath p .355 .375 .83 r
F P
0 g
s
.74024 1.20689 m .79529 1.15424 L .77083 1.1292 L closepath p 1 1 1 r
F P
0 g
s
.74024 1.20689 m .77083 1.1292 L .70519 1.1946 L closepath p 1 1 1 r
F P
0 g
s
.30441 .51265 m .30617 .52912 L .31792 .52106 L closepath p .652 .519 .825 r
F P
0 g
s
.30441 .51265 m .31792 .52106 L .31922 .50947 L closepath p .707 .496 .746 r
F P
0 g
s
.3246 .44161 m .32323 .46222 L .33145 .45895 L closepath p .723 .506 .745 r
F P
0 g
s
.3246 .44161 m .33145 .45895 L .33246 .43848 L closepath p .739 .514 .742 r
F P
0 g
s
.37504 .41708 m .37321 .4268 L .38618 .4249 L closepath p .7 .296 .41 r
F P
0 g
s
.37504 .41708 m .38618 .4249 L .38557 .41561 L closepath p .701 .294 .405 r
F P
0 g
s
.25548 .88469 m .25539 .87479 L .25491 .88447 L closepath p .449 .417 .83 r
F P
0 g
s
.25548 .88469 m .25491 .88447 L .25262 .90024 L closepath p .49 .382 .761 r
F P
0 g
s
.28206 .59535 m .28196 .60041 L .29227 .58697 L closepath p .297 .211 .656 r
F P
0 g
s
.28206 .59535 m .29227 .58697 L .29067 .58135 L closepath p .46 .47 .887 r
F P
0 g
s
.39491 .41463 m .39574 .42545 L .4193 .42607 L closepath p .842 .446 .491 r
F P
0 g
s
.39491 .41463 m .4193 .42607 L .41628 .41516 L closepath p .845 .467 .528 r
F P
0 g
s
.38557 .41561 m .39574 .42545 L .39491 .41463 L closepath p .83 .437 .492 r
F P
0 g
s
.3019 .53363 m .29245 .55535 L .29849 .5473 L closepath p .081 .081 .081 r
F P
0 g
s
.3019 .53363 m .29849 .5473 L .30617 .52912 L closepath p .104 .104 .104 r
F P
0 g
s
.24972 .91934 m .2518 .92851 L .25443 .91932 L closepath p .12 .12 .323 r
F P
0 g
s
.24972 .91934 m .25443 .91932 L .25401 .90884 L closepath p .12 .12 .309 r
F P
0 g
s
.25548 .88469 m .25668 .87487 L .25539 .87479 L closepath p .582 .527 .885 r
F P
0 g
s
.25673 .88735 m .25668 .87487 L .25548 .88469 L closepath p .13 .13 .45 r
F P
0 g
s
.25673 .88735 m .25548 .88469 L .25262 .90024 L closepath p .147 .147 .453 r
F P
0 g
s
.79529 1.15424 m .80939 1.13494 L .77643 1.11735 L closepath p 1 1 1 r
F P
0 g
s
.80939 1.13494 m .81113 1.03513 L .77643 1.11735 L closepath p 1 1 .919 r
F P
0 g
s
.25539 .87479 m .26251 .83666 L .2609 .83085 L closepath p .163 .163 .601 r
F P
0 g
s
.25668 .87487 m .26251 .83666 L .25539 .87479 L closepath p .577 .556 .922 r
F P
0 g
s
.25262 .90024 m .25401 .90884 L .25709 .90349 L closepath p .225 .183 .608 r
F P
0 g
s
.25262 .90024 m .25709 .90349 L .25926 .89504 L closepath p .164 .164 .69 r
F P
0 g
s
.25262 .90024 m .25926 .89504 L .25876 .89252 L closepath p .144 .144 .475 r
F P
0 g
s
.25876 .89252 m .25673 .88735 L .25262 .90024 L closepath p .131 .131 .375 r
F P
0 g
s
.71012 .68554 m .66726 .62998 L .66942 .65286 L closepath p 1 .798 .554 r
F P
0 g
s
.66942 .65286 m .71416 .70901 L .71012 .68554 L closepath p 1 .805 .566 r
F P
0 g
s
.27128 1.01783 m .2642 .99079 L .26205 .99923 L closepath p .048 .048 .368 r
F P
0 g
s
.27128 1.01783 m .26205 .99923 L .26836 1.02528 L closepath p .073 .073 .386 \
r
F P
0 g
s
.26205 .99923 m .2642 .99079 L .26166 .95627 L closepath p .031 .031 .174 r
F P
0 g
s
.2504 .94174 m .26205 .99923 L .26166 .95627 L closepath p .097 .097 .523 r
F P
0 g
s
.38557 .41561 m .38618 .4249 L .39574 .42545 L closepath p .851 .41 .394 r
F P
0 g
s
.41628 .41516 m .4193 .42607 L .43856 .43041 L closepath p .934 .521 .477 r
F P
0 g
s
.30593 .47392 m .3141 .49899 L .31988 .49406 L closepath p .68 .618 .93 r
F P
0 g
s
.30593 .47392 m .31988 .49406 L .31214 .46778 L closepath p .717 .633 .922 r
F P
0 g
s
.79529 1.15424 m .77643 1.11735 L .77083 1.1292 L closepath p 1 1 .924 r
F P
0 g
s
.5434 1.21652 m .47939 1.22866 L .52894 1.26046 L closepath p .732 1 1 r
F P
0 g
s
.5434 1.21652 m .52894 1.26046 L .60603 1.24674 L closepath p .733 1 1 r
F P
0 g
s
.28196 .60041 m .282 .60267 L .29322 .59315 L closepath p .288 .188 .549 r
F P
0 g
s
.28196 .60041 m .29322 .59315 L .29227 .58697 L closepath p .458 .418 .828 r
F P
0 g
s
.25668 .87487 m .26369 .84528 L .26251 .83666 L closepath p .154 .154 .572 r
F P
0 g
s
.25872 .88089 m .26369 .84528 L .25668 .87487 L closepath p .147 .147 .536 r
F P
0 g
s
.25673 .88735 m .25872 .88089 L .25668 .87487 L closepath p .16 .16 .557 r
F P
0 g
s
.66474 .6023 m .63379 .59816 L .63518 .61339 L closepath p 1 .723 .579 r
F P
0 g
s
.66474 .6023 m .6313 .5834 L .63379 .59816 L closepath p 1 .72 .575 r
F P
0 g
s
.66474 .6023 m .63518 .61339 L .66726 .62998 L closepath p 1 .779 .628 r
F P
0 g
s
.33246 .43848 m .33145 .45895 L .3386 .45518 L closepath p .706 .487 .733 r
F P
0 g
s
.33246 .43848 m .3386 .45518 L .33976 .43627 L closepath p .74 .505 .728 r
F P
0 g
s
.57652 .52765 m .58042 .56683 L .62942 .56402 L closepath p 1 .719 .666 r
F P
0 g
s
.62942 .56402 m .58042 .56683 L .6313 .5834 L closepath p 1 .706 .651 r
F P
0 g
s
.29466 .70644 m .28803 .67475 L .2739 .68979 L closepath p .299 .326 .799 r
F P
0 g
s
.29466 .70644 m .2739 .68979 L .27267 .71772 L closepath p .435 .333 .733 r
F P
0 g
s
.57652 .52765 m .55118 .5032 L .4766 .482 L closepath p 1 .782 .798 r
F P
0 g
s
.55118 .5032 m .52571 .47935 L .44458 .45587 L closepath p 1 .814 .86 r
F P
0 g
s
.55118 .5032 m .44458 .45587 L .4766 .482 L closepath p 1 .766 .774 r
F P
0 g
s
.282 .60267 m .28025 .60888 L .28836 .60205 L closepath p .153 .153 .36 r
F P
0 g
s
.282 .60267 m .28836 .60205 L .29322 .59315 L closepath p .142 .142 .225 r
F P
0 g
s
.43856 .43041 m .44458 .45587 L .52571 .47935 L closepath p 1 .798 .835 r
F P
0 g
s
.43856 .43041 m .52571 .47935 L .48098 .44232 L closepath p 1 .802 .843 r
F P
0 g
s
.30617 .52912 m .29849 .5473 L .30552 .54188 L closepath p .136 .136 .136 r
F P
0 g
s
.30617 .52912 m .30552 .54188 L .31792 .52106 L closepath p .15 .144 .144 r
F P
0 g
s
.53775 .53177 m .58042 .56683 L .57652 .52765 L closepath p 1 .696 .68 r
F P
0 g
s
.57652 .52765 m .4766 .482 L .50805 .50493 L closepath p 1 .679 .625 r
F P
0 g
s
.50805 .50493 m .53775 .53177 L .57652 .52765 L closepath p 1 .77 .76 r
F P
0 g
s
.33976 .43627 m .3386 .45518 L .34181 .45513 L closepath p .846 .594 .75 r
F P
0 g
s
.33976 .43627 m .34181 .45513 L .34198 .43494 L closepath p .767 .537 .748 r
F P
0 g
s
.34198 .43494 m .34181 .45513 L .35079 .45204 L closepath p .782 .549 .75 r
F P
0 g
s
.34198 .43494 m .35079 .45204 L .35015 .43211 L closepath p .77 .544 .755 r
F P
0 g
s
.31214 .46778 m .31988 .49406 L .32951 .48865 L closepath p .778 .672 .924 r
F P
0 g
s
.31214 .46778 m .32951 .48865 L .32323 .46222 L closepath p .809 .68 .908 r
F P
0 g
s
.3141 .49899 m .31922 .50947 L .32375 .50941 L closepath p .545 .571 .955 r
F P
0 g
s
.3141 .49899 m .32375 .50941 L .31988 .49406 L closepath p .677 .613 .926 r
F P
0 g
s
.28025 .60888 m .27762 .61559 L .2822 .61136 L closepath p .088 .088 .088 r
F P
0 g
s
.28025 .60888 m .2822 .61136 L .28836 .60205 L closepath p .094 .094 .094 r
F P
0 g
s
.26181 .89068 m .25872 .88089 L .25673 .88735 L closepath p .137 .137 .438 r
F P
0 g
s
.26181 .89068 m .25673 .88735 L .25876 .89252 L closepath p .143 .143 .374 r
F P
0 g
s
.7597 .84071 m .77086 .86447 L .76868 .85185 L closepath p 1 .791 .437 r
F P
0 g
s
.7184 .75509 m .74761 .80167 L .74675 .78454 L closepath p 1 .868 .622 r
F P
0 g
s
.31922 .50947 m .31792 .52106 L .32652 .51567 L closepath p .659 .454 .729 r
F P
0 g
s
.31922 .50947 m .32652 .51567 L .32375 .50941 L closepath p .547 .455 .817 r
F P
0 g
s
.25401 .90884 m .25443 .91932 L .25926 .90936 L closepath p .152 .152 .382 r
F P
0 g
s
.25401 .90884 m .25926 .90936 L .25709 .90349 L closepath p .153 .153 .431 r
F P
0 g
s
.35015 .43211 m .35079 .45204 L .36329 .44889 L closepath p .796 .564 .759 r
F P
0 g
s
.35015 .43211 m .36329 .44889 L .36333 .4278 L closepath p .781 .561 .771 r
F P
0 g
s
.27267 .71772 m .29867 .73716 L .29466 .70644 L closepath p .481 .423 .819 r
F P
0 g
s
.27118 .76041 m .29867 .73716 L .27267 .71772 L closepath p .551 .42 .769 r
F P
0 g
s
.30685 1.10563 m .30646 1.0948 L .28531 1.05081 L closepath p .103 .123 .696 \
r
F P
0 g
s
.32222 1.12547 m .30646 1.0948 L .30685 1.10563 L closepath p .105 .143 .713 \
r
F P
0 g
s
.25872 .88089 m .26545 .85506 L .26369 .84528 L closepath p .159 .159 .59 r
F P
0 g
s
.26181 .89068 m .26545 .85506 L .25872 .88089 L closepath p .127 .127 .426 r
F P
0 g
s
.27762 .61559 m .27565 .61784 L .28098 .6154 L closepath p .078 .078 .078 r
F P
0 g
s
.27762 .61559 m .28098 .6154 L .2822 .61136 L closepath p .103 .103 .158 r
F P
0 g
s
.25866 .82709 m .2609 .83085 L .28458 .80553 L closepath p .707 .655 .955 r
F P
0 g
s
.25866 .82709 m .28458 .80553 L .28331 .80126 L closepath p .687 .623 .931 r
F P
0 g
s
.27118 .76041 m .25866 .82709 L .28331 .80126 L closepath p .662 .586 .904 r
F P
0 g
s
.32323 .46222 m .32951 .48865 L .33644 .48447 L closepath p .778 .659 .908 r
F P
0 g
s
.32323 .46222 m .33644 .48447 L .33145 .45895 L closepath p .802 .669 .9 r
F P
0 g
s
.31988 .49406 m .32375 .50941 L .33424 .50635 L closepath p .75 .662 .933 r
F P
0 g
s
.31988 .49406 m .33424 .50635 L .32951 .48865 L closepath p .774 .664 .917 r
F P
0 g
s
.28872 .5647 m .28816 .56913 L .28875 .56113 L closepath p .806 .699 .148 r
F P
0 g
s
.26251 .83666 m .28458 .80553 L .2609 .83085 L closepath p .455 .309 .686 r
F P
0 g
s
.34355 1.1618 m .39443 1.19215 L .40492 1.19099 L closepath p .495 .595 1 r
F P
0 g
s
.34355 1.1618 m .40492 1.19099 L .38335 1.15902 L closepath p .362 .585 1 r
F P
0 g
s
.34355 1.1618 m .38335 1.15902 L .32222 1.12547 L closepath p .362 .586 1 r
F P
0 g
s
.27565 .61784 m .2754 .62436 L .28015 .62503 L closepath p .171 .171 .528 r
F P
0 g
s
.27565 .61784 m .28015 .62503 L .28098 .6154 L closepath p .16 .16 .539 r
F P
0 g
s
.4193 .42607 m .44458 .45587 L .43856 .43041 L closepath p 1 .786 .845 r
F P
0 g
s
.25709 .90349 m .25926 .90936 L .26097 .89731 L closepath p .122 .122 .413 r
F P
0 g
s
.25709 .90349 m .26097 .89731 L .25926 .89504 L closepath p .144 .144 .581 r
F P
0 g
s
.36333 .4278 m .36329 .44889 L .37378 .44889 L closepath p .903 .66 .785 r
F P
0 g
s
.36333 .4278 m .37378 .44889 L .37321 .4268 L closepath p .869 .646 .804 r
F P
0 g
s
.32375 .50941 m .32652 .51567 L .33485 .51008 L closepath p .695 .533 .81 r
F P
0 g
s
.32375 .50941 m .33485 .51008 L .33424 .50635 L closepath p .693 .513 .783 r
F P
0 g
s
.2504 .94174 m .26166 .95627 L .26507 .94323 L closepath p .155 .155 .47 r
F P
0 g
s
.2504 .94174 m .26507 .94323 L .2518 .92851 L closepath p .155 .155 .607 r
F P
0 g
s
.31792 .52106 m .30552 .54188 L .31235 .53589 L closepath p .11 .11 .11 r
F P
0 g
s
.31792 .52106 m .31235 .53589 L .32652 .51567 L closepath p .193 .145 .145 r
F P
0 g
s
.28803 .5725 m .29067 .58135 L .29016 .57796 L closepath p .059 .059 .405 r
F P
0 g
s
.33145 .45895 m .33644 .48447 L .34325 .48079 L closepath p .783 .656 .899 r
F P
0 g
s
.33145 .45895 m .34325 .48079 L .3386 .45518 L closepath p .806 .666 .893 r
F P
0 g
s
.37321 .4268 m .37378 .44889 L .3882 .44831 L closepath p .899 .67 .805 r
F P
0 g
s
.37321 .4268 m .3882 .44831 L .38618 .4249 L closepath p .878 .666 .823 r
F P
0 g
s
.77086 .86447 m .75489 .88947 L .76985 .90109 L closepath p 1 .912 .702 r
F P
0 g
s
.7597 .84071 m .75489 .88947 L .77086 .86447 L closepath p 1 .896 .697 r
F P
0 g
s
.66688 .67137 m .69341 .73091 L .71416 .70901 L closepath p 1 .81 .599 r
F P
0 g
s
.71416 .70901 m .69341 .73091 L .7184 .75509 L closepath p 1 .862 .628 r
F P
0 g
s
.66942 .65286 m .66688 .67137 L .71416 .70901 L closepath p 1 .861 .689 r
F P
0 g
s
.39574 .42545 m .4019 .45027 L .42736 .4529 L closepath p .983 .743 .81 r
F P
0 g
s
.39574 .42545 m .42736 .4529 L .4193 .42607 L closepath p .95 .758 .874 r
F P
0 g
s
.3961 .44908 m .4019 .45027 L .39574 .42545 L closepath p .997 .751 .804 r
F P
0 g
s
.38618 .4249 m .3961 .44908 L .39574 .42545 L closepath p .992 .746 .805 r
F P
0 g
s
.4193 .42607 m .42736 .4529 L .44458 .45587 L closepath p 1 .787 .85 r
F P
0 g
s
.28875 .56113 m .28816 .56913 L .2916 .56429 L closepath p .172 .172 .565 r
F P
0 g
s
.28875 .56113 m .2916 .56429 L .29245 .55535 L closepath p .205 .178 .545 r
F P
0 g
s
.2754 .62436 m .27589 .6323 L .28187 .63359 L closepath p .374 .288 .712 r
F P
0 g
s
.2754 .62436 m .28187 .63359 L .28015 .62503 L closepath p .168 .193 .717 r
F P
0 g
s
.29067 .58135 m .29227 .58697 L .29233 .58299 L closepath p .116 .116 .599 r
F P
0 g
s
.29067 .58135 m .29233 .58299 L .29016 .57796 L closepath p .099 .099 .604 r
F P
0 g
s
.32951 .48865 m .33424 .50635 L .3411 .50691 L closepath p .741 .643 .916 r
F P
0 g
s
.32951 .48865 m .3411 .50691 L .33644 .48447 L closepath p .777 .657 .906 r
F P
0 g
s
.38618 .4249 m .3882 .44831 L .3961 .44908 L closepath p .978 .741 .814 r
F P
0 g
s
.26251 .83666 m .28656 .81635 L .28458 .80553 L closepath p .635 .528 .85 r
F P
0 g
s
.26369 .84528 m .28656 .81635 L .26251 .83666 L closepath p .564 .423 .763 r
F P
0 g
s
.3386 .45518 m .34325 .48079 L .34641 .48082 L closepath p .929 .75 .886 r
F P
0 g
s
.3386 .45518 m .34641 .48082 L .34181 .45513 L closepath p .872 .717 .9 r
F P
0 g
s
.29322 .59315 m .28836 .60205 L .28973 .59971 L closepath p .006 .006 .006 r
F P
0 g
s
.29322 .59315 m .28973 .59971 L .29265 .59029 L closepath p 1 .978 .55 r
F P
0 g
s
.29265 .59029 m .29154 .5958 L .29275 .59195 L p .044 .044 .044 r
F P
0 g
s
.29257 .58895 m .29322 .59315 L .29265 .59029 L p .741 .517 .011 r
F P
0 g
s
.29275 .59195 m .29444 .58655 L .29265 .59029 L p .044 .044 .044 r
F P
0 g
s
.29265 .59029 m .29227 .58697 L .29257 .58895 L p .741 .517 .011 r
F P
0 g
s
.29227 .58697 m .29265 .59029 L .29233 .58299 L closepath p .023 .023 .109 r
F P
0 g
s
.7184 .75509 m .71796 .79895 L .74761 .80167 L closepath p 1 .89 .675 r
F P
0 g
s
.74761 .80167 m .71796 .79895 L .74701 .82778 L closepath p 1 .895 .681 r
F P
0 g
s
.74761 .80167 m .74701 .82778 L .7597 .84071 L closepath p 1 .891 .691 r
F P
0 g
s
.32652 .51567 m .31235 .53589 L .32224 .53081 L closepath p .305 .162 .181 r
F P
0 g
s
.32652 .51567 m .32224 .53081 L .33485 .51008 L closepath p .189 .148 .148 r
F P
0 g
s
.34181 .45513 m .34641 .48082 L .35513 .47927 L closepath p .875 .72 .9 r
F P
0 g
s
.34181 .45513 m .35513 .47927 L .35079 .45204 L closepath p .855 .711 .908 r
F P
0 g
s
.28816 .56913 m .29016 .57796 L .29254 .57247 L closepath p .289 .303 .776 r
F P
0 g
s
.28816 .56913 m .29254 .57247 L .2916 .56429 L closepath p .33 .304 .757 r
F P
0 g
s
.63379 .59816 m .6313 .5834 L .58042 .56683 L closepath p 1 .687 .624 r
F P
0 g
s
.25876 .89252 m .25926 .89504 L .26181 .89068 L closepath p .155 .155 .496 r
F P
0 g
s
.2518 .92851 m .26507 .94323 L .26644 .9343 L closepath p .173 .173 .573 r
F P
0 g
s
.2518 .92851 m .26644 .9343 L .25443 .91932 L closepath p .173 .173 .571 r
F P
0 g
s
.28531 1.05081 m .27128 1.01783 L .26836 1.02528 L closepath p .1 .1 .649 r
F P
0 g
s
.28005 .6468 m .28187 .63359 L .27589 .6323 L closepath p .354 .335 .783 r
F P
0 g
s
.74701 .82778 m .74897 .84721 L .7597 .84071 L closepath p 1 .84 .594 r
F P
0 g
s
.7597 .84071 m .74897 .84721 L .74663 .88151 L closepath p 1 .94 .672 r
F P
0 g
s
.7597 .84071 m .74663 .88151 L .75489 .88947 L closepath p 1 .9 .69 r
F P
0 g
s
.81863 .97936 m .76708 .95775 L .74626 1.01593 L closepath p 1 .954 .888 r
F P
0 g
s
.81863 .97936 m .74626 1.01593 L .81113 1.03513 L closepath p 1 .904 .849 r
F P
0 g
s
.33644 .48447 m .3411 .50691 L .34755 .50459 L closepath p .766 .65 .906 r
F P
0 g
s
.33644 .48447 m .34755 .50459 L .34325 .48079 L closepath p .784 .658 .901 r
F P
0 g
s
.4766 .482 m .44458 .45587 L .4515 .47858 L closepath p 1 .798 .871 r
F P
0 g
s
.4515 .47858 m .44458 .45587 L .42839 .46358 L closepath p .993 .795 .875 r
F P
0 g
s
.42839 .46358 m .44458 .45587 L .42736 .4529 L closepath p 1 .818 .895 r
F P
0 g
s
.33424 .50635 m .33485 .51008 L .3411 .50691 L closepath p .742 .547 .787 r
F P
0 g
s
.35079 .45204 m .35513 .47927 L .36745 .47667 L closepath p .884 .731 .907 r
F P
0 g
s
.35079 .45204 m .36745 .47667 L .36329 .44889 L closepath p .857 .723 .921 r
F P
0 g
s
.29245 .55535 m .2916 .56429 L .29656 .55504 L closepath p .149 .149 .44 r
F P
0 g
s
.29245 .55535 m .29656 .55504 L .29849 .5473 L closepath p .141 .141 .307 r
F P
0 g
s
.29016 .57796 m .29233 .58299 L .29425 .57903 L closepath p .264 .351 .84 r
F P
0 g
s
.29016 .57796 m .29425 .57903 L .29254 .57247 L closepath p .296 .311 .783 r
F P
0 g
s
.25443 .91932 m .26644 .9343 L .27232 .91899 L closepath p .299 .188 .521 r
F P
0 g
s
.25443 .91932 m .27232 .91899 L .25926 .90936 L closepath p .308 .193 .601 r
F P
0 g
s
.28836 .60205 m .2822 .61136 L .28228 .60983 L closepath p .775 .941 .554 r
F P
0 g
s
.28836 .60205 m .28228 .60983 L .28973 .59971 L closepath p .652 .87 .488 r
F P
0 g
s
.26369 .84528 m .28868 .82591 L .28656 .81635 L closepath p .654 .536 .847 r
F P
0 g
s
.26545 .85506 m .28868 .82591 L .26369 .84528 L closepath p .577 .419 .747 r
F P
0 g
s
.50805 .50493 m .4766 .482 L .4515 .47858 L closepath p .997 .767 .829 r
F P
0 g
s
.25926 .89504 m .26097 .89731 L .26181 .89068 L closepath p .165 .165 .564 r
F P
0 g
s
.29233 .58299 m .29265 .59029 L .29444 .58655 L closepath p .136 .136 .49 r
F P
0 g
s
.29233 .58299 m .29444 .58655 L .29425 .57903 L closepath p .137 .137 .49 r
F P
0 g
s
.29867 .73716 m .27118 .76041 L .30572 .78022 L closepath p .664 .592 .909 r
F P
0 g
s
.28331 .80126 m .30572 .78022 L .27118 .76041 L closepath p .669 .591 .905 r
F P
0 g
s
.33485 .51008 m .32224 .53081 L .33017 .52767 L closepath p .522 .194 .351 r
F P
0 g
s
.33485 .51008 m .33017 .52767 L .3411 .50691 L closepath p .422 .184 .313 r
F P
0 g
s
.63518 .61339 m .63669 .63417 L .66726 .62998 L closepath p 1 .778 .626 r
F P
0 g
s
.66726 .62998 m .63669 .63417 L .66942 .65286 L closepath p 1 .773 .621 r
F P
0 g
s
.42839 .46358 m .42736 .4529 L .4019 .45027 L closepath p .99 .809 .898 r
F P
0 g
s
.34325 .48079 m .34755 .50459 L .35068 .50343 L closepath p .952 .773 .893 r
F P
0 g
s
.34325 .48079 m .35068 .50343 L .34641 .48082 L closepath p .93 .761 .9 r
F P
0 g
s
.34641 .48082 m .35068 .50343 L .35804 .4977 L closepath p .898 .74 .904 r
F P
0 g
s
.34641 .48082 m .35804 .4977 L .35513 .47927 L closepath p .879 .735 .915 r
F P
0 g
s
.36329 .44889 m .36745 .47667 L .37765 .47577 L closepath p .943 .783 .916 r
F P
0 g
s
.36329 .44889 m .37765 .47577 L .37378 .44889 L closepath p .923 .776 .926 r
F P
0 g
s
.2916 .56429 m .29254 .57247 L .29777 .56334 L closepath p .296 .277 .742 r
F P
0 g
s
.2916 .56429 m .29777 .56334 L .29656 .55504 L closepath p .292 .256 .718 r
F P
0 g
s
.25926 .90936 m .27232 .91899 L .27459 .90673 L closepath p .268 .189 .634 r
F P
0 g
s
.25926 .90936 m .27459 .90673 L .26097 .89731 L closepath p .287 .259 .725 r
F P
0 g
s
.50805 .50493 m .4515 .47858 L .45864 .49814 L closepath p 1 .726 .753 r
F P
0 g
s
.45864 .49814 m .53775 .53177 L .50805 .50493 L closepath p 1 .797 .848 r
F P
0 g
s
.4515 .47858 m .42839 .46358 L .43141 .4758 L closepath p 1 .775 .827 r
F P
0 g
s
.42839 .46358 m .4019 .45027 L .40555 .47468 L closepath p .984 .817 .916 r
F P
0 g
s
.40555 .47468 m .43141 .4758 L .42839 .46358 L closepath p .948 .74 .851 r
F P
0 g
s
.3411 .50691 m .33017 .52767 L .33943 .5226 L closepath p .394 .18 .297 r
F P
0 g
s
.3411 .50691 m .33943 .5226 L .34755 .50459 L closepath p .446 .185 .309 r
F P
0 g
s
.29265 .59029 m .28973 .59971 L .29154 .5958 L closepath p .039 .039 .039 r
F P
0 g
s
.2822 .61136 m .28098 .6154 L .28212 .61442 L closepath p .055 .055 .055 r
F P
0 g
s
.2822 .61136 m .28212 .61442 L .28228 .60983 L closepath p .002 .002 .002 r
F P
0 g
s
.45864 .49814 m .4515 .47858 L .43141 .4758 L closepath p .999 .723 .758 r
F P
0 g
s
.37378 .44889 m .37765 .47577 L .39324 .4744 L closepath p .947 .792 .924 r
F P
0 g
s
.37378 .44889 m .39324 .4744 L .3882 .44831 L closepath p .921 .787 .943 r
F P
0 g
s
.3961 .44908 m .40555 .47468 L .4019 .45027 L closepath p .996 .826 .913 r
F P
0 g
s
.35513 .47927 m .35804 .4977 L .37109 .4922 L closepath p .904 .752 .914 r
F P
0 g
s
.35513 .47927 m .37109 .4922 L .36745 .47667 L closepath p .89 .754 .93 r
F P
0 g
s
.28005 .6468 m .28503 .64888 L .28187 .63359 L closepath p .551 .519 .894 r
F P
0 g
s
.28803 .67475 m .28503 .64888 L .28005 .6468 L closepath p .57 .508 .869 r
F P
0 g
s
.48536 .54177 m .58042 .56683 L .53775 .53177 L closepath p 1 .683 .65 r
F P
0 g
s
.45864 .49814 m .48536 .54177 L .53775 .53177 L closepath p 1 .674 .64 r
F P
0 g
s
.3961 .44908 m .40114 .4744 L .40555 .47468 L closepath p 1 .83 .908 r
F P
0 g
s
.3882 .44831 m .40114 .4744 L .3961 .44908 L closepath p .985 .817 .914 r
F P
0 g
s
.29254 .57247 m .29425 .57903 L .30096 .57066 L closepath p .467 .429 .836 r
F P
0 g
s
.29254 .57247 m .30096 .57066 L .29777 .56334 L closepath p .486 .509 .917 r
F P
0 g
s
.3882 .44831 m .39324 .4744 L .40114 .4744 L closepath p .938 .798 .941 r
F P
0 g
s
.28098 .6154 m .28015 .62503 L .282 .6247 L closepath p .11 .11 .348 r
F P
0 g
s
.28098 .6154 m .282 .6247 L .28212 .61442 L closepath p .078 .078 .272 r
F P
0 g
s
.26097 .89731 m .27633 .90282 L .26181 .89068 L closepath p .352 .28 .714 r
F P
0 g
s
.26181 .89068 m .26876 .86804 L .26545 .85506 L closepath p .351 .291 .73 r
F P
0 g
s
.26097 .89731 m .27459 .90673 L .27633 .90282 L closepath p .381 .244 .647 r
F P
0 g
s
.26545 .85506 m .29079 .83851 L .28868 .82591 L closepath p .65 .511 .816 r
F P
0 g
s
.26876 .86804 m .29079 .83851 L .26545 .85506 L closepath p .624 .469 .779 r
F P
0 g
s
.36745 .47667 m .37109 .4922 L .37847 .48791 L closepath p .981 .811 .911 r
F P
0 g
s
.36745 .47667 m .37847 .48791 L .37765 .47577 L closepath p .947 .813 .949 r
F P
0 g
s
.35068 .50343 m .35764 .5018 L .35804 .4977 L closepath p .811 .601 .798 r
F P
0 g
s
.34755 .50459 m .34453 .52249 L .35068 .50343 L closepath p .691 .308 .445 r
F P
0 g
s
.35068 .50343 m .34453 .52249 L .35218 .52 L closepath p .54 .198 .383 r
F P
0 g
s
.35068 .50343 m .35218 .52 L .35764 .5018 L closepath p .609 .219 .385 r
F P
0 g
s
.60603 1.24674 m .64879 1.22888 L .59468 1.19766 L closepath p .878 1 1 r
F P
0 g
s
.60603 1.24674 m .59468 1.19766 L .5434 1.21652 L closepath p .863 1 1 r
F P
0 g
s
.37765 .47577 m .37847 .48791 L .39577 .47938 L closepath p .912 .786 .95 r
F P
0 g
s
.37765 .47577 m .39577 .47938 L .39324 .4744 L closepath p .923 .693 .811 r
F P
0 g
s
.28015 .62503 m .28187 .63359 L .2861 .63676 L closepath p .164 .174 .704 r
F P
0 g
s
.28015 .62503 m .2861 .63676 L .282 .6247 L closepath p .113 .113 .608 r
F P
0 g
s
.34755 .50459 m .33943 .5226 L .34453 .52249 L closepath p .738 .348 .457 r
F P
0 g
s
.43666 .49583 m .43141 .4758 L .40555 .47468 L closepath p .945 .722 .828 r
F P
0 g
s
.43666 .49583 m .40555 .47468 L .41289 .49563 L closepath p .944 .722 .83 r
F P
0 g
s
.40146 .47531 m .41289 .49563 L .40555 .47468 L closepath p 1 .75 .799 r
F P
0 g
s
.40114 .4744 m .40146 .47531 L .40555 .47468 L closepath p .987 .716 .763 r
F P
0 g
s
.29425 .57903 m .29444 .58655 L .30242 .57959 L closepath p .438 .341 .741 r
F P
0 g
s
.29425 .57903 m .30242 .57959 L .30096 .57066 L closepath p .435 .387 .802 r
F P
0 g
s
.35804 .4977 m .35764 .5018 L .36875 .50035 L closepath p .861 .645 .81 r
F P
0 g
s
.35804 .4977 m .36875 .50035 L .37109 .4922 L closepath p .86 .657 .828 r
F P
0 g
s
.28973 .59971 m .28228 .60983 L .28425 .60625 L closepath p .703 .908 .53 r
F P
0 g
s
.28973 .59971 m .28425 .60625 L .29154 .5958 L closepath p .771 .944 .561 r
F P
0 g
s
.29849 .5473 m .29656 .55504 L .30329 .54765 L closepath p .166 .166 .401 r
F P
0 g
s
.29849 .5473 m .30329 .54765 L .30552 .54188 L closepath p .162 .155 .25 r
F P
0 g
s
.39324 .4744 m .39577 .47938 L .40146 .47531 L closepath p .921 .692 .812 r
F P
0 g
s
.39324 .4744 m .40146 .47531 L .40114 .4744 L closepath p .92 .683 .8 r
F P
0 g
s
.37847 .48791 m .37964 .49869 L .3976 .49603 L closepath p .876 .627 .765 r
F P
0 g
s
.37847 .48791 m .3976 .49603 L .39577 .47938 L closepath p .857 .671 .851 r
F P
0 g
s
.37109 .4922 m .37964 .49869 L .37847 .48791 L closepath p .884 .632 .765 r
F P
0 g
s
.45864 .49814 m .43141 .4758 L .43666 .49583 L closepath p .979 .742 .815 r
F P
0 g
s
.40146 .47531 m .40631 .49588 L .41289 .49563 L closepath p .946 .736 .847 r
F P
0 g
s
.39577 .47938 m .40631 .49588 L .40146 .47531 L closepath p .948 .737 .846 r
F P
0 g
s
.29128 .65596 m .2861 .63676 L .28187 .63359 L closepath p .16 .175 .712 r
F P
0 g
s
.29128 .65596 m .28187 .63359 L .28503 .64888 L closepath p .169 .207 .73 r
F P
0 g
s
.37109 .4922 m .36875 .50035 L .37964 .49869 L closepath p .852 .648 .824 r
F P
0 g
s
.32222 1.12547 m .36472 1.13538 L .3459 1.10986 L closepath p .362 .566 1 r
F P
0 g
s
.38335 1.15902 m .36472 1.13538 L .32222 1.12547 L closepath p .328 .577 1 r
F P
0 g
s
.32222 1.12547 m .3459 1.10986 L .31174 1.08104 L closepath p .346 .531 .987 \
r
F P
0 g
s
.32222 1.12547 m .31174 1.08104 L .30646 1.0948 L closepath p .227 .43 .925 r
F P
0 g
s
.35764 .5018 m .35218 .52 L .36407 .51723 L closepath p .631 .238 .393 r
F P
0 g
s
.35764 .5018 m .36407 .51723 L .36875 .50035 L closepath p .674 .259 .372 r
F P
0 g
s
.63518 .61339 m .63379 .59816 L .58035 .59953 L closepath p 1 .665 .627 r
F P
0 g
s
.63379 .59816 m .58042 .56683 L .58035 .59953 L closepath p 1 .728 .699 r
F P
0 g
s
.29444 .58655 m .29154 .5958 L .29781 .5899 L closepath p .137 .137 .271 r
F P
0 g
s
.29444 .58655 m .29781 .5899 L .30242 .57959 L closepath p .141 .141 .216 r
F P
0 g
s
.39577 .47938 m .3976 .49603 L .40631 .49588 L closepath p .948 .737 .846 r
F P
0 g
s
.71796 .79895 m .74897 .84721 L .74701 .82778 L closepath p 1 .841 .582 r
F P
0 g
s
.28503 .64888 m .28803 .67475 L .29464 .66554 L closepath p .339 .33 .784 r
F P
0 g
s
.28503 .64888 m .29464 .66554 L .29128 .65596 L closepath p .155 .183 .728 r
F P
0 g
s
.74664 .89937 m .76708 .95775 L .76985 .90109 L closepath p 1 .909 .695 r
F P
0 g
s
.74664 .89937 m .76985 .90109 L .75489 .88947 L closepath p 1 .907 .693 r
F P
0 g
s
.28331 .80126 m .28458 .80553 L .30587 .79794 L closepath p .666 .609 .928 r
F P
0 g
s
.28331 .80126 m .30587 .79794 L .30572 .78022 L closepath p .66 .578 .895 r
F P
0 g
s
.36875 .50035 m .36407 .51723 L .37629 .51638 L closepath p .757 .335 .4 r
F P
0 g
s
.36875 .50035 m .37629 .51638 L .37964 .49869 L closepath p .692 .307 .443 r
F P
0 g
s
.28656 .81635 m .30587 .79794 L .28458 .80553 L closepath p .634 .527 .85 r
F P
0 g
s
.29656 .55504 m .29777 .56334 L .30514 .55679 L closepath p .557 .45 .804 r
F P
0 g
s
.29656 .55504 m .30514 .55679 L .30329 .54765 L closepath p .548 .473 .84 r
F P
0 g
s
.28228 .60983 m .28212 .61442 L .2833 .61179 L closepath p .066 .066 .215 r
F P
0 g
s
.28228 .60983 m .2833 .61179 L .28425 .60625 L closepath p .082 .082 .166 r
F P
0 g
s
.29805 .70011 m .28803 .67475 L .29466 .70644 L closepath p .115 .115 .549 r
F P
0 g
s
.29464 .66554 m .28803 .67475 L .29805 .70011 L closepath p .243 .205 .679 r
F P
0 g
s
.43666 .49583 m .48536 .54177 L .45864 .49814 L closepath p .964 .672 .724 r
F P
0 g
s
.28212 .61442 m .282 .6247 L .28722 .62213 L closepath p .214 .182 .646 r
F P
0 g
s
.28212 .61442 m .28722 .62213 L .2833 .61179 L closepath p .112 .112 .603 r
F P
0 g
s
.37964 .49869 m .37629 .51638 L .3887 .51482 L closepath p .773 .378 .464 r
F P
0 g
s
.37964 .49869 m .3887 .51482 L .3976 .49603 L closepath p .752 .371 .483 r
F P
0 g
s
.58035 .59953 m .58042 .56683 L .53942 .604 L closepath p .981 .679 .709 r
F P
0 g
s
.53942 .604 m .58042 .56683 L .48536 .54177 L closepath p 1 .766 .768 r
F P
0 g
s
.30552 .54188 m .30329 .54765 L .30937 .54243 L closepath p .141 .141 .196 r
F P
0 g
s
.30552 .54188 m .30937 .54243 L .31235 .53589 L closepath p .139 .139 .171 r
F P
0 g
s
.30646 1.0948 m .31174 1.08104 L .28531 1.05081 L closepath p .227 .43 .925 r
F P
0 g
s
.29154 .5958 m .28425 .60625 L .29015 .60175 L closepath p .082 .082 .082 r
F P
0 g
s
.29154 .5958 m .29015 .60175 L .29781 .5899 L closepath p .076 .076 .076 r
F P
0 g
s
.26876 .86804 m .29076 .85227 L .29079 .83851 L closepath p .575 .409 .735 r
F P
0 g
s
.29777 .56334 m .30096 .57066 L .30942 .56413 L closepath p .683 .627 .938 r
F P
0 g
s
.29777 .56334 m .30942 .56413 L .30514 .55679 L closepath p .681 .638 .953 r
F P
0 g
s
.28656 .81635 m .31067 .81159 L .30587 .79794 L closepath p .757 .713 .987 r
F P
0 g
s
.28868 .82591 m .31067 .81159 L .28656 .81635 L closepath p .726 .585 .854 r
F P
0 g
s
.41289 .49563 m .41849 .51821 L .4454 .52579 L closepath p .977 .628 .623 r
F P
0 g
s
.41289 .49563 m .4454 .52579 L .43666 .49583 L closepath p .931 .67 .766 r
F P
0 g
s
.40631 .49588 m .41284 .51662 L .41289 .49563 L closepath p .871 .525 .599 r
F P
0 g
s
.41289 .49563 m .41284 .51662 L .41849 .51821 L closepath p 1 .658 .581 r
F P
0 g
s
.3976 .49603 m .40359 .51542 L .40631 .49588 L closepath p .864 .503 .568 r
F P
0 g
s
.40631 .49588 m .40359 .51542 L .41284 .51662 L closepath p .914 .549 .576 r
F P
0 g
s
.43666 .49583 m .4454 .52579 L .48536 .54177 L closepath p .95 .681 .757 r
F P
0 g
s
.3976 .49603 m .3887 .51482 L .40359 .51542 L closepath p .891 .518 .554 r
F P
0 g
s
.282 .6247 m .2861 .63676 L .29443 .6362 L closepath p .399 .425 .866 r
F P
0 g
s
.282 .6247 m .29443 .6362 L .28722 .62213 L closepath p .32 .404 .876 r
F P
0 g
s
.69341 .73091 m .68042 .76511 L .7184 .75509 L closepath p 1 .931 .761 r
F P
0 g
s
.7184 .75509 m .68042 .76511 L .71796 .79895 L closepath p 1 .875 .707 r
F P
0 g
s
.77643 1.11735 m .81113 1.03513 L .74626 1.01593 L closepath p 1 .966 .945 r
F P
0 g
s
.30096 .57066 m .30242 .57959 L .31118 .5739 L closepath p .622 .523 .853 r
F P
0 g
s
.30096 .57066 m .31118 .5739 L .30942 .56413 L closepath p .621 .524 .855 r
F P
0 g
s
.66688 .67137 m .66942 .65286 L .63669 .63417 L closepath p 1 .825 .706 r
F P
0 g
s
.29466 .70644 m .29867 .73716 L .29805 .70011 L closepath p .105 .105 .456 r
F P
0 g
s
.2861 .63676 m .29128 .65596 L .29684 .6627 L closepath p .184 .25 .76 r
F P
0 g
s
.2861 .63676 m .29684 .6627 L .30047 .65765 L closepath p .349 .352 .805 r
F P
0 g
s
.2861 .63676 m .30047 .65765 L .29443 .6362 L closepath p .395 .371 .805 r
F P
0 g
s
.28868 .82591 m .31349 .81976 L .31067 .81159 L closepath p .804 .726 .968 r
F P
0 g
s
.29079 .83851 m .31349 .81976 L .28868 .82591 L closepath p .76 .588 .829 r
F P
0 g
s
.30329 .54765 m .30514 .55679 L .31141 .55078 L closepath p .574 .491 .846 r
F P
0 g
s
.30329 .54765 m .31141 .55078 L .30937 .54243 L closepath p .57 .495 .853 r
F P
0 g
s
.30242 .57959 m .29781 .5899 L .30635 .58248 L closepath p .149 .149 .249 r
F P
0 g
s
.30242 .57959 m .30635 .58248 L .31118 .5739 L closepath p .154 .148 .171 r
F P
0 g
s
.31235 .53589 m .30937 .54243 L .31898 .53595 L closepath p .328 .177 .327 r
F P
0 g
s
.31235 .53589 m .31898 .53595 L .32224 .53081 L closepath p .296 .16 .17 r
F P
0 g
s
.28425 .60625 m .2833 .61179 L .2896 .6064 L closepath p .166 .166 .474 r
F P
0 g
s
.28425 .60625 m .2896 .6064 L .29015 .60175 L closepath p .168 .168 .508 r
F P
0 g
s
.64879 1.22888 m .67806 1.21532 L .62538 1.18293 L closepath p .919 1 1 r
F P
0 g
s
.64879 1.22888 m .62538 1.18293 L .59468 1.19766 L closepath p .936 1 1 r
F P
0 g
s
.74897 .84721 m .71294 .86192 L .74663 .88151 L closepath p 1 .901 .759 r
F P
0 g
s
.71796 .79895 m .71294 .86192 L .74897 .84721 L closepath p 1 .913 .77 r
F P
0 g
s
.45062 1.23007 m .40695 1.19812 L .42545 1.22846 L closepath p .704 .851 1 r
F P
0 g
s
.30514 .55679 m .30942 .56413 L .31512 .56014 L closepath p .638 .617 .955 r
F P
0 g
s
.30514 .55679 m .31512 .56014 L .31141 .55078 L closepath p .652 .606 .934 r
F P
0 g
s
.2833 .61179 m .28722 .62213 L .29333 .61931 L closepath p .4 .402 .839 r
F P
0 g
s
.2833 .61179 m .29333 .61931 L .2896 .6064 L closepath p .387 .401 .844 r
F P
0 g
s
.63669 .63417 m .63518 .61339 L .57916 .63584 L closepath p 1 .711 .684 r
F P
0 g
s
.57916 .63584 m .63518 .61339 L .58035 .59953 L closepath p 1 .697 .672 r
F P
0 g
s
.29684 .6627 m .29128 .65596 L .29464 .66554 L closepath p .165 .225 .756 r
F P
0 g
s
.29079 .83851 m .31628 .83238 L .31349 .81976 L closepath p .841 .715 .925 r
F P
0 g
s
.29076 .85227 m .31628 .83238 L .29079 .83851 L closepath p .794 .583 .789 r
F P
0 g
s
.30942 .56413 m .31118 .5739 L .31962 .56836 L closepath p .665 .556 .864 r
F P
0 g
s
.30942 .56413 m .31962 .56836 L .31512 .56014 L closepath p .62 .585 .93 r
F P
0 g
s
.32224 .53081 m .31898 .53595 L .32796 .53361 L closepath p .531 .19 .315 r
F P
0 g
s
.32224 .53081 m .32796 .53361 L .33017 .52767 L closepath p .525 .194 .356 r
F P
0 g
s
.31174 1.08104 m .30771 1.02543 L .28531 1.05081 L closepath p .434 .53 .961 \
r
F P
0 g
s
.30771 1.02543 m .27128 1.01783 L .28531 1.05081 L closepath p .447 .551 .976 \
r
F P
0 g
s
.31905 .75615 m .30737 .72472 L .29867 .73716 L closepath p .173 .233 .751 r
F P
0 g
s
.30572 .78022 m .31905 .75615 L .29867 .73716 L closepath p .198 .24 .743 r
F P
0 g
s
.29805 .70011 m .29867 .73716 L .30737 .72472 L closepath p .171 .217 .736 r
F P
0 g
s
.30937 .54243 m .31141 .55078 L .32037 .54426 L closepath p .687 .574 .87 r
F P
0 g
s
.30937 .54243 m .32037 .54426 L .31898 .53595 L closepath p .694 .538 .817 r
F P
0 g
s
.29781 .5899 m .29015 .60175 L .29947 .59424 L closepath p .105 .105 .105 r
F P
0 g
s
.29781 .5899 m .29947 .59424 L .30635 .58248 L closepath p .106 .106 .106 r
F P
0 g
s
.29785 .67078 m .29684 .6627 L .29464 .66554 L closepath p .104 .104 .245 r
F P
0 g
s
.29785 .67078 m .29464 .66554 L .29563 .67559 L p .074 .074 .316 r
F P
0 g
s
.30047 .65765 m .29684 .6627 L .29785 .67078 L closepath p .125 .125 .273 r
F P
0 g
s
.30293 .65667 m .30047 .65765 L .29785 .67078 L closepath p .393 .199 .545 r
F P
0 g
s
.31277 .6532 m .30293 .65667 L .29785 .67078 L closepath p .147 .147 .316 r
F P
0 g
s
.29785 .67078 m .28829 .6766 L .30061 .66483 L p .7 .978 .696 r
F P
0 g
s
.30061 .66483 m .31277 .6532 L .29785 .67078 L p .7 .978 .696 r
F P
0 g
s
.29563 .67559 m .29805 .70011 L .29785 .67078 L p .074 .074 .316 r
F P
0 g
s
.33017 .52767 m .32796 .53361 L .33421 .52938 L closepath p .259 .163 .231 r
F P
0 g
s
.33017 .52767 m .33421 .52938 L .33943 .5226 L closepath p .28 .156 .156 r
F P
0 g
s
.28722 .62213 m .29443 .6362 L .29909 .63419 L closepath p .401 .453 .895 r
F P
0 g
s
.28722 .62213 m .29909 .63419 L .29333 .61931 L closepath p .425 .46 .893 r
F P
0 g
s
.31118 .5739 m .30635 .58248 L .31239 .57802 L closepath p .118 .118 .118 r
F P
0 g
s
.31118 .5739 m .31239 .57802 L .31962 .56836 L closepath p .126 .126 .126 r
F P
0 g
s
.74663 .88151 m .74664 .89937 L .75489 .88947 L closepath p 1 .9 .69 r
F P
0 g
s
.67806 1.21532 m .70519 1.1946 L .65125 1.16733 L closepath p 1 1 1 r
F P
0 g
s
.67806 1.21532 m .65125 1.16733 L .62538 1.18293 L closepath p .979 1 1 r
F P
0 g
s
.31141 .55078 m .31512 .56014 L .32352 .55433 L closepath p .727 .651 .936 r
F P
0 g
s
.31141 .55078 m .32352 .55433 L .32037 .54426 L closepath p .731 .646 .929 r
F P
0 g
s
.30771 1.02543 m .29612 .98783 L .27128 1.01783 L closepath p .482 .515 .926 \
r
F P
0 g
s
.29612 .98783 m .2642 .99079 L .27128 1.01783 L closepath p .512 .559 .958 r
F P
0 g
s
.30572 .78022 m .30587 .79794 L .33428 .80895 L closepath p .662 .579 .895 r
F P
0 g
s
.30572 .78022 m .33428 .80895 L .32804 .79588 L closepath p .174 .402 .909 r
F P
0 g
s
.32804 .79588 m .31905 .75615 L .30572 .78022 L closepath p .354 .437 .898 r
F P
0 g
s
.33943 .5226 m .33421 .52938 L .34106 .52889 L closepath p .756 .375 .484 r
F P
0 g
s
.33943 .5226 m .34106 .52889 L .34453 .52249 L closepath p .749 .371 .485 r
F P
0 g
s
.31067 .81159 m .33981 .81959 L .33428 .80895 L closepath p .593 .721 1 r
F P
0 g
s
.31067 .81159 m .33428 .80895 L .30587 .79794 L closepath p .602 .613 .971 r
F P
0 g
s
.31349 .81976 m .33981 .81959 L .31067 .81159 L closepath p .702 .663 .967 r
F P
0 g
s
.31512 .56014 m .31962 .56836 L .3264 .5644 L closepath p .768 .658 .914 r
F P
0 g
s
.31512 .56014 m .3264 .5644 L .32352 .55433 L closepath p .743 .679 .959 r
F P
0 g
s
.29076 .85227 m .31572 .84716 L .31628 .83238 L closepath p .792 .58 .787 r
F P
0 g
s
.31349 .81976 m .3421 .82837 L .33981 .81959 L closepath p .697 .699 1 r
F P
0 g
s
.31628 .83238 m .3421 .82837 L .31349 .81976 L closepath p .748 .652 .922 r
F P
0 g
s
.31898 .53595 m .32037 .54426 L .32885 .54182 L closepath p .814 .622 .825 r
F P
0 g
s
.31898 .53595 m .32885 .54182 L .32796 .53361 L closepath p .798 .632 .854 r
F P
0 g
s
.70519 1.1946 m .77083 1.1292 L .71804 1.09335 L closepath p 1 1 1 r
F P
0 g
s
.70519 1.1946 m .71804 1.09335 L .65125 1.16733 L closepath p 1 1 1 r
F P
0 g
s
.63677 .65037 m .66413 .70784 L .66688 .67137 L closepath p 1 .803 .634 r
F P
0 g
s
.66688 .67137 m .66413 .70784 L .69341 .73091 L closepath p 1 .816 .624 r
F P
0 g
s
.66688 .67137 m .63669 .63417 L .63677 .65037 L closepath p 1 .785 .604 r
F P
0 g
s
.34453 .52249 m .34106 .52889 L .35012 .52472 L closepath p .475 .188 .326 r
F P
0 g
s
.34453 .52249 m .35012 .52472 L .35218 .52 L closepath p .479 .184 .282 r
F P
0 g
s
.29443 .6362 m .30047 .65765 L .30293 .65667 L closepath p .531 .461 .836 r
F P
0 g
s
.29443 .6362 m .30293 .65667 L .29909 .63419 L closepath p .365 .364 .812 r
F P
0 g
s
.29015 .60175 m .2896 .6064 L .29931 .60131 L closepath p .344 .206 .62 r
F P
0 g
s
.29015 .60175 m .29931 .60131 L .29947 .59424 L closepath p .352 .265 .695 r
F P
0 g
s
.41849 .51821 m .44624 .52961 L .4454 .52579 L closepath p .997 .774 .839 r
F P
0 g
s
.4454 .52579 m .44624 .52961 L .48536 .54177 L closepath p .938 .737 .858 r
F P
0 g
s
.77083 1.1292 m .77643 1.11735 L .71804 1.09335 L closepath p 1 1 .965 r
F P
0 g
s
.77643 1.11735 m .74626 1.01593 L .71804 1.09335 L closepath p 1 .978 .883 r
F P
0 g
s
.31628 .83238 m .34176 .84353 L .3421 .82837 L closepath p .758 .713 .987 r
F P
0 g
s
.31572 .84716 m .34176 .84353 L .31628 .83238 L closepath p .849 .627 .797 r
F P
0 g
s
.40695 1.19812 m .40492 1.19099 L .39443 1.19215 L closepath p .495 .6 1 r
F P
0 g
s
.29612 .98783 m .2928 .96642 L .2642 .99079 L closepath p .507 .473 .864 r
F P
0 g
s
.2928 .96642 m .26166 .95627 L .2642 .99079 L closepath p .529 .506 .892 r
F P
0 g
s
.27232 .91899 m .29744 .9023 L .27459 .90673 L closepath p .712 .59 .872 r
F P
0 g
s
.27459 .90673 m .29744 .9023 L .2976 .89886 L closepath p .603 .316 .567 r
F P
0 g
s
.27459 .90673 m .2976 .89886 L .27633 .90282 L closepath p .773 .664 .918 r
F P
0 g
s
.31962 .56836 m .31239 .57802 L .31915 .5723 L closepath p .083 .083 .083 r
F P
0 g
s
.31962 .56836 m .31915 .5723 L .3264 .5644 L closepath p .101 .101 .101 r
F P
0 g
s
.26644 .9343 m .28512 .91731 L .27232 .91899 L closepath p .657 .469 .752 r
F P
0 g
s
.27232 .91899 m .28512 .91731 L .29744 .9023 L closepath p .674 .526 .818 r
F P
0 g
s
.47939 1.22866 m .43135 1.20167 L .45062 1.23007 L closepath p .736 .893 1 r
F P
0 g
s
.45062 1.23007 m .43135 1.20167 L .40695 1.19812 L closepath p .594 .807 1 r
F P
0 g
s
.42039 .5252 m .44817 .53642 L .44624 .52961 L closepath p 1 .782 .836 r
F P
0 g
s
.41849 .51821 m .42039 .5252 L .44624 .52961 L closepath p 1 .734 .769 r
F P
0 g
s
.44624 .52961 m .44817 .53642 L .48536 .54177 L closepath p .937 .742 .866 r
F P
0 g
s
.2896 .6064 m .29333 .61931 L .30605 .6152 L closepath p .576 .53 .891 r
F P
0 g
s
.2896 .6064 m .30605 .6152 L .29931 .60131 L closepath p .513 .54 .939 r
F P
0 g
s
.32796 .53361 m .32885 .54182 L .33544 .53826 L closepath p .725 .578 .845 r
F P
0 g
s
.32796 .53361 m .33544 .53826 L .33421 .52938 L closepath p .724 .578 .847 r
F P
0 g
s
.32037 .54426 m .32352 .55433 L .33112 .55125 L closepath p .823 .703 .927 r
F P
0 g
s
.32037 .54426 m .33112 .55125 L .32885 .54182 L closepath p .844 .697 .899 r
F P
0 g
s
.35218 .52 m .35012 .52472 L .36312 .52307 L closepath p .68 .266 .378 r
F P
0 g
s
.35218 .52 m .36312 .52307 L .36407 .51723 L closepath p .698 .35 .515 r
F P
0 g
s
.41284 .51662 m .42039 .5252 L .41849 .51821 L closepath p 1 .763 .727 r
F P
0 g
s
.30635 .58248 m .29947 .59424 L .30558 .58996 L closepath p .126 .126 .126 r
F P
0 g
s
.30635 .58248 m .30558 .58996 L .31239 .57802 L closepath p .12 .12 .12 r
F P
0 g
s
.2928 .96642 m .28949 .94836 L .26166 .95627 L closepath p .545 .492 .865 r
F P
0 g
s
.26166 .95627 m .28949 .94836 L .28992 .93284 L closepath p .508 .389 .759 r
F P
0 g
s
.26166 .95627 m .28992 .93284 L .2677 .94362 L closepath p .566 .478 .834 r
F P
0 g
s
.26166 .95627 m .2677 .94362 L .26507 .94323 L closepath p .435 .334 .734 r
F P
0 g
s
.4211 .5349 m .44919 .54119 L .44817 .53642 L closepath p .988 .741 .802 r
F P
0 g
s
.42039 .5252 m .4211 .5349 L .44817 .53642 L closepath p .999 .815 .895 r
F P
0 g
s
.44817 .53642 m .44919 .54119 L .48536 .54177 L closepath p .935 .708 .819 r
F P
0 g
s
.71294 .86192 m .74628 .89085 L .74663 .88151 L closepath p 1 .889 .738 r
F P
0 g
s
.74663 .88151 m .74628 .89085 L .74664 .89937 L closepath p 1 .794 .793 r
F P
0 g
s
.26507 .94323 m .27124 .93396 L .26644 .9343 L closepath p .469 .366 .753 r
F P
0 g
s
.26644 .9343 m .27124 .93396 L .28512 .91731 L closepath p .415 .197 .482 r
F P
0 g
s
.32352 .55433 m .3264 .5644 L .33415 .56173 L closepath p .82 .729 .96 r
F P
0 g
s
.32352 .55433 m .33415 .56173 L .33112 .55125 L closepath p .832 .728 .948 r
F P
0 g
s
.26507 .94323 m .2677 .94362 L .27124 .93396 L closepath p .436 .326 .723 r
F P
0 g
s
.40359 .51542 m .41497 .52376 L .41284 .51662 L closepath p .964 .719 .799 r
F P
0 g
s
.41284 .51662 m .41497 .52376 L .42039 .5252 L closepath p 1 .765 .737 r
F P
0 g
s
.3264 .5644 m .31915 .5723 L .32732 .56972 L closepath p .38 .158 .158 r
F P
0 g
s
.3264 .5644 m .32732 .56972 L .33415 .56173 L closepath p .348 .157 .157 r
F P
0 g
s
.3887 .51482 m .40475 .52376 L .40359 .51542 L closepath p .957 .693 .768 r
F P
0 g
s
.40359 .51542 m .40475 .52376 L .41497 .52376 L closepath p .975 .705 .763 r
F P
0 g
s
.29333 .61931 m .29909 .63419 L .31078 .63137 L closepath p .534 .53 .916 r
F P
0 g
s
.29333 .61931 m .31078 .63137 L .30605 .6152 L closepath p .577 .531 .893 r
F P
0 g
s
.32885 .54182 m .33112 .55125 L .33784 .54862 L closepath p .815 .678 .901 r
F P
0 g
s
.32885 .54182 m .33784 .54862 L .33544 .53826 L closepath p .779 .682 .935 r
F P
0 g
s
.58035 .59953 m .53942 .604 L .57916 .63584 L closepath p .968 .65 .678 r
F P
0 g
s
.36407 .51723 m .36312 .52307 L .37496 .52187 L closepath p .836 .471 .55 r
F P
0 g
s
.36407 .51723 m .37496 .52187 L .37629 .51638 L closepath p .837 .482 .57 r
F P
0 g
s
.63669 .63417 m .57916 .63584 L .63677 .65037 L closepath p 1 .691 .662 r
F P
0 g
s
.33421 .52938 m .33544 .53826 L .34212 .53705 L closepath p .885 .694 .855 r
F P
0 g
s
.33421 .52938 m .34212 .53705 L .34106 .52889 L closepath p .884 .694 .856 r
F P
0 g
s
.42233 .54076 m .44853 .54375 L .44919 .54119 L closepath p .958 .685 .753 r
F P
0 g
s
.4211 .5349 m .42233 .54076 L .44919 .54119 L closepath p .985 .818 .916 r
F P
0 g
s
.44919 .54119 m .44853 .54375 L .48536 .54177 L closepath p .918 .645 .741 r
F P
0 g
s
.37629 .51638 m .37496 .52187 L .3891 .5218 L closepath p .876 .519 .581 r
F P
0 g
s
.37629 .51638 m .3891 .5218 L .3887 .51482 L closepath p .877 .581 .69 r
F P
0 g
s
.41497 .52376 m .4211 .5349 L .42039 .5252 L closepath p 1 .861 .877 r
F P
0 g
s
.3887 .51482 m .3891 .5218 L .40475 .52376 L closepath p .966 .65 .681 r
F P
0 g
s
.30737 .72472 m .30814 .71695 L .29805 .70011 L closepath p .079 .079 .495 r
F P
0 g
s
.29805 .70011 m .30814 .71695 L .28798 .6932 L closepath p .73 .243 .053 r
F P
0 g
s
.29805 .70011 m .29153 .68325 L .29785 .67078 L closepath p 1 1 .817 r
F P
0 g
s
.28798 .6932 m .29153 .68325 L .29805 .70011 L closepath p 1 .87 .436 r
F P
0 g
s
.74358 .94736 m .74626 1.01593 L .76708 .95775 L closepath p 1 1 .872 r
F P
0 g
s
.74664 .89937 m .74358 .94736 L .76708 .95775 L closepath p 1 .91 .699 r
F P
0 g
s
.34106 .52889 m .34212 .53705 L .35001 .53361 L closepath p .785 .623 .854 r
F P
0 g
s
.34106 .52889 m .35001 .53361 L .35012 .52472 L closepath p .799 .612 .826 r
F P
0 g
s
.48536 .54177 m .45212 .61821 L .49776 .61294 L closepath p 1 .794 .856 r
F P
0 g
s
.49776 .61294 m .53942 .604 L .48536 .54177 L closepath p .986 .78 .863 r
F P
0 g
s
.45212 .61821 m .48536 .54177 L .4508 .58731 L closepath p 1 .851 .874 r
F P
0 g
s
.4508 .58731 m .48536 .54177 L .45214 .57054 L closepath p 1 .835 .866 r
F P
0 g
s
.45136 .56252 m .45214 .57054 L .48536 .54177 L closepath p 1 .762 .817 r
F P
0 g
s
.45136 .55596 m .45136 .56252 L .48536 .54177 L closepath p .967 .703 .77 r
F P
0 g
s
.45097 .55449 m .45136 .55596 L .48536 .54177 L closepath p .678 .252 .351 r
F P
0 g
s
.44989 .54898 m .45097 .55449 L .48536 .54177 L closepath p .874 .549 .637 r
F P
0 g
s
.44853 .54375 m .44989 .54898 L .48536 .54177 L closepath p .907 .614 .706 r
F P
0 g
s
.68042 .76511 m .68848 .7903 L .71796 .79895 L closepath p 1 .829 .621 r
F P
0 g
s
.71796 .79895 m .68848 .7903 L .71294 .86192 L closepath p 1 .927 .752 r
F P
0 g
s
.33112 .55125 m .33415 .56173 L .34229 .55863 L closepath p .876 .755 .944 r
F P
0 g
s
.33112 .55125 m .34229 .55863 L .33784 .54862 L closepath p .848 .761 .976 r
F P
0 g
s
.42376 .54687 m .44989 .54898 L .44853 .54375 L closepath p .915 .619 .703 r
F P
0 g
s
.42233 .54076 m .42376 .54687 L .44853 .54375 L closepath p .972 .781 .881 r
F P
0 g
s
.31572 .84716 m .33828 .84881 L .34176 .84353 L closepath p .871 .696 .873 r
F P
0 g
s
.41598 .5331 m .42233 .54076 L .4211 .5349 L closepath p 1 .877 .867 r
F P
0 g
s
.41497 .52376 m .41598 .5331 L .4211 .5349 L closepath p 1 .863 .844 r
F P
0 g
s
.40475 .52376 m .41598 .5331 L .41497 .52376 L closepath p .997 .801 .878 r
F P
0 g
s
.29909 .63419 m .30293 .65667 L .31497 .64457 L closepath p .456 .432 .845 r
F P
0 g
s
.29909 .63419 m .31497 .64457 L .31078 .63137 L closepath p .505 .427 .81 r
F P
0 g
s
.33415 .56173 m .32732 .56972 L .33466 .56636 L closepath p .247 .142 .142 r
F P
0 g
s
.33415 .56173 m .33466 .56636 L .34229 .55863 L closepath p .316 .145 .145 r
F P
0 g
s
.41576 .53907 m .42376 .54687 L .42233 .54076 L closepath p 1 .836 .834 r
F P
0 g
s
.41598 .5331 m .41576 .53907 L .42233 .54076 L closepath p 1 .883 .905 r
F P
0 g
s
.74664 .89937 m .74628 .89085 L .71529 .89845 L closepath p 1 .879 .709 r
F P
0 g
s
.71294 .86192 m .71529 .89845 L .74628 .89085 L closepath p 1 .869 .7 r
F P
0 g
s
.66413 .70784 m .66467 .75391 L .69341 .73091 L closepath p 1 .9 .779 r
F P
0 g
s
.69341 .73091 m .66467 .75391 L .68042 .76511 L closepath p 1 .874 .762 r
F P
0 g
s
.33544 .53826 m .33784 .54862 L .34344 .54684 L closepath p .855 .732 .935 r
F P
0 g
s
.33544 .53826 m .34344 .54684 L .34212 .53705 L closepath p .898 .737 .901 r
F P
0 g
s
.2677 .94362 m .28992 .93284 L .29176 .92617 L closepath p .357 .19 .455 r
F P
0 g
s
.2677 .94362 m .29176 .92617 L .27124 .93396 L closepath p .665 .563 .873 r
F P
0 g
s
.35012 .52472 m .35001 .53361 L .36359 .53204 L closepath p .898 .693 .84 r
F P
0 g
s
.35012 .52472 m .36359 .53204 L .36312 .52307 L closepath p .882 .71 .88 r
F P
0 g
s
.3891 .5218 m .40657 .53175 L .40475 .52376 L closepath p .994 .784 .858 r
F P
0 g
s
.40475 .52376 m .40657 .53175 L .41598 .5331 L closepath p 1 .793 .851 r
F P
0 g
s
.34212 .53705 m .34344 .54684 L .35158 .54429 L closepath p .843 .698 .901 r
F P
0 g
s
.34212 .53705 m .35158 .54429 L .35001 .53361 L closepath p .821 .701 .924 r
F P
0 g
s
.41672 .54625 m .42809 .55264 L .42376 .54687 L closepath p .946 .502 .399 r
F P
0 g
s
.41576 .53907 m .41672 .54625 L .42376 .54687 L closepath p 1 .817 .775 r
F P
0 g
s
.42376 .54687 m .42809 .55264 L .44989 .54898 L closepath p .883 .528 .588 r
F P
0 g
s
.33784 .54862 m .34229 .55863 L .34604 .55749 L closepath p .788 .729 .984 r
F P
0 g
s
.33784 .54862 m .34604 .55749 L .34344 .54684 L closepath p .859 .745 .947 r
F P
0 g
s
.40657 .53175 m .41576 .53907 L .41598 .5331 L closepath p 1 .838 .913 r
F P
0 g
s
.34229 .55863 m .33466 .56636 L .34216 .56443 L closepath p .436 .163 .163 r
F P
0 g
s
.34229 .55863 m .34216 .56443 L .34604 .55749 L closepath p .349 .157 .157 r
F P
0 g
s
.31174 1.08104 m .32555 1.06654 L .30771 1.02543 L closepath p .449 .544 .969 \
r
F P
0 g
s
.3459 1.10986 m .32555 1.06654 L .31174 1.08104 L closepath p .47 .575 .992 r
F P
0 g
s
.27124 .93396 m .29176 .92617 L .29776 .91088 L closepath p .479 .204 .5 r
F P
0 g
s
.27124 .93396 m .29776 .91088 L .28512 .91731 L closepath p .691 .48 .737 r
F P
0 g
s
.31239 .57802 m .30558 .58996 L .31358 .58265 L closepath p .122 .122 .122 r
F P
0 g
s
.31239 .57802 m .31358 .58265 L .31915 .5723 L closepath p .123 .123 .123 r
F P
0 g
s
.4069 .5389 m .41672 .54625 L .41576 .53907 L closepath p 1 .774 .808 r
F P
0 g
s
.40657 .53175 m .4069 .5389 L .41576 .53907 L closepath p 1 .827 .873 r
F P
0 g
s
.42809 .55264 m .45097 .55449 L .44989 .54898 L closepath p .903 .568 .629 r
F P
0 g
s
.28512 .91731 m .29776 .91088 L .29744 .9023 L closepath p .814 .713 .946 r
F P
0 g
s
.37496 .52187 m .38957 .53047 L .3891 .5218 L closepath p .952 .742 .848 r
F P
0 g
s
.3891 .5218 m .38957 .53047 L .40657 .53175 L closepath p .998 .777 .843 r
F P
0 g
s
.34344 .54684 m .34604 .55749 L .35449 .55438 L closepath p .874 .754 .945 r
F P
0 g
s
.34344 .54684 m .35449 .55438 L .35158 .54429 L closepath p .864 .757 .957 r
F P
0 g
s
.31277 .6532 m .31497 .64457 L .30293 .65667 L closepath p .148 .148 .32 r
F P
0 g
s
.36312 .52307 m .37583 .53145 L .37496 .52187 L closepath p .952 .76 .874 r
F P
0 g
s
.37496 .52187 m .37583 .53145 L .38957 .53047 L closepath p .942 .753 .875 r
F P
0 g
s
.38957 .53047 m .4069 .5389 L .40657 .53175 L closepath p 1 .805 .878 r
F P
0 g
s
.36312 .52307 m .36359 .53204 L .37583 .53145 L closepath p .949 .762 .88 r
F P
0 g
s
.43135 1.20167 m .40492 1.19099 L .40695 1.19812 L closepath p .707 .755 1 r
F P
0 g
s
.74664 .89937 m .71529 .89845 L .70864 .94246 L closepath p 1 .947 .789 r
F P
0 g
s
.74664 .89937 m .70864 .94246 L .74358 .94736 L closepath p 1 .866 .751 r
F P
0 g
s
.34604 .55749 m .34216 .56443 L .35029 .56122 L closepath p .344 .157 .157 r
F P
0 g
s
.34604 .55749 m .35029 .56122 L .35449 .55438 L closepath p .36 .155 .155 r
F P
0 g
s
.35001 .53361 m .35158 .54429 L .36441 .54251 L closepath p .922 .772 .923 r
F P
0 g
s
.35001 .53361 m .36441 .54251 L .36359 .53204 L closepath p .918 .775 .93 r
F P
0 g
s
.34176 .84353 m .3558 .8531 L .35432 .84169 L closepath p .532 .611 1 r
F P
0 g
s
.34176 .84353 m .35432 .84169 L .3421 .82837 L closepath p .523 .51 .899 r
F P
0 g
s
.33828 .84881 m .3558 .8531 L .34176 .84353 L closepath p .794 .602 .816 r
F P
0 g
s
.38988 .53891 m .4083 .5461 L .4069 .5389 L closepath p .972 .743 .825 r
F P
0 g
s
.38957 .53047 m .38988 .53891 L .4069 .5389 L closepath p .986 .828 .928 r
F P
0 g
s
.4069 .5389 m .4083 .5461 L .41672 .54625 L closepath p 1 .773 .806 r
F P
0 g
s
.29744 .9023 m .30809 .8923 L .2976 .89886 L closepath p .548 .274 .553 r
F P
0 g
s
.33828 .84881 m .34824 .85651 L .3558 .8531 L closepath p .746 .708 .989 r
F P
0 g
s
.4083 .5461 m .41936 .55383 L .41672 .54625 L closepath p .978 .648 .657 r
F P
0 g
s
.41672 .54625 m .41936 .55383 L .42809 .55264 L closepath p .979 .649 .656 r
F P
0 g
s
.35158 .54429 m .35449 .55438 L .36328 .5528 L closepath p .895 .776 .953 r
F P
0 g
s
.35158 .54429 m .36328 .5528 L .36441 .54251 L closepath p .921 .769 .92 r
F P
0 g
s
.36359 .53204 m .37516 .54063 L .37583 .53145 L closepath p .957 .814 .941 r
F P
0 g
s
.37583 .53145 m .37516 .54063 L .38988 .53891 L closepath p .946 .805 .94 r
F P
0 g
s
.37583 .53145 m .38988 .53891 L .38957 .53047 L closepath p .951 .801 .93 r
F P
0 g
s
.29947 .59424 m .29931 .60131 L .30382 .59683 L closepath p .174 .174 .599 r
F P
0 g
s
.29947 .59424 m .30382 .59683 L .30558 .58996 L closepath p .277 .181 .442 r
F P
0 g
s
.36359 .53204 m .36441 .54251 L .37516 .54063 L closepath p .966 .811 .927 r
F P
0 g
s
.35449 .55438 m .35029 .56122 L .36267 .56014 L closepath p .676 .206 .253 r
F P
0 g
s
.35449 .55438 m .36267 .56014 L .36328 .5528 L closepath p .655 .248 .377 r
F P
0 g
s
.36441 .54251 m .37641 .55067 L .37516 .54063 L closepath p .953 .76 .874 r
F P
0 g
s
.37516 .54063 m .37641 .55067 L .39115 .54802 L closepath p .933 .746 .876 r
F P
0 g
s
.37516 .54063 m .39115 .54802 L .38988 .53891 L closepath p .933 .745 .874 r
F P
0 g
s
.42022 1.18737 m .38335 1.15902 L .40492 1.19099 L closepath p .539 .697 1 r
F P
0 g
s
.43135 1.20167 m .42022 1.18737 L .40492 1.19099 L closepath p .537 .749 1 r
F P
0 g
s
.38988 .53891 m .39115 .54802 L .4083 .5461 L closepath p .966 .767 .869 r
F P
0 g
s
.36441 .54251 m .36328 .5528 L .37641 .55067 L closepath p .927 .775 .921 r
F P
0 g
s
.47939 1.22866 m .5434 1.21652 L .51723 1.19162 L closepath p .834 .945 1 r
F P
0 g
s
.47939 1.22866 m .51723 1.19162 L .45842 1.20174 L closepath p .836 .949 1 r
F P
0 g
s
.47939 1.22866 m .45842 1.20174 L .43135 1.20167 L closepath p .714 .887 1 r
F P
0 g
s
.41936 .55383 m .4326 .55831 L .42809 .55264 L closepath p .871 .45 .446 r
F P
0 g
s
.42809 .55264 m .4326 .55831 L .45097 .55449 L closepath p .855 .455 .487 r
F P
0 g
s
.29931 .60131 m .30605 .6152 L .30983 .61224 L closepath p .327 .426 .896 r
F P
0 g
s
.29931 .60131 m .30983 .61224 L .30382 .59683 L closepath p .383 .446 .896 r
F P
0 g
s
.39115 .54802 m .40977 .55391 L .4083 .5461 L closepath p .903 .594 .676 r
F P
0 g
s
.4083 .5461 m .40977 .55391 L .41936 .55383 L closepath p .979 .644 .649 r
F P
0 g
s
.3214 .88032 m .34824 .85651 L .31869 .88225 L closepath p .063 .063 .063 r
F P
0 g
s
.30917 .89258 m .3214 .88032 L .31869 .88225 L closepath p .092 .092 .092 r
F P
0 g
s
.30917 .89258 m .31869 .88225 L .30809 .8923 L closepath p .242 .158 .192 r
F P
0 g
s
.36328 .5528 m .36267 .56014 L .37549 .55792 L closepath p .747 .323 .391 r
F P
0 g
s
.36328 .5528 m .37549 .55792 L .37641 .55067 L closepath p .755 .376 .487 r
F P
0 g
s
.4326 .55831 m .45136 .55596 L .45097 .55449 L closepath p .752 .298 .326 r
F P
0 g
s
.3421 .82837 m .35432 .84169 L .35062 .83285 L closepath p .152 .332 .86 r
F P
0 g
s
.3421 .82837 m .35062 .83285 L .33981 .81959 L closepath p .192 .322 .833 r
F P
0 g
s
.29153 .68325 m .28829 .6766 L .29785 .67078 L closepath p 1 1 .688 r
F P
0 g
s
.30588 .90049 m .30917 .89258 L .29744 .9023 L closepath p .858 .745 .947 r
F P
0 g
s
.29776 .91088 m .30588 .90049 L .29744 .9023 L closepath p .859 .749 .952 r
F P
0 g
s
.29744 .9023 m .30917 .89258 L .30809 .8923 L closepath p .238 .161 .23 r
F P
0 g
s
.29176 .92617 m .30588 .90049 L .29776 .91088 L closepath p .166 .161 .312 r
F P
0 g
s
.31915 .5723 m .31358 .58265 L .32248 .57882 L closepath p .511 .197 .398 r
F P
0 g
s
.31915 .5723 m .32248 .57882 L .32732 .56972 L closepath p .554 .198 .378 r
F P
0 g
s
.37641 .55067 m .37549 .55792 L .39248 .55632 L closepath p .816 .429 .499 r
F P
0 g
s
.37641 .55067 m .39248 .55632 L .39115 .54802 L closepath p .827 .495 .608 r
F P
0 g
s
.39115 .54802 m .39248 .55632 L .40977 .55391 L closepath p .894 .541 .594 r
F P
0 g
s
.3214 .88032 m .3558 .8531 L .34824 .85651 L closepath p .36 .178 .306 r
F P
0 g
s
.33981 .81959 m .35062 .83285 L .34462 .82161 L closepath p .084 .103 .673 r
F P
0 g
s
.33981 .81959 m .34462 .82161 L .33428 .80895 L closepath p .086 .096 .669 r
F P
0 g
s
.57916 .63584 m .53942 .604 L .49776 .61294 L closepath p .947 .678 .758 r
F P
0 g
s
.63322 .69128 m .63677 .65037 L .57916 .63584 L closepath p 1 .743 .735 r
F P
0 g
s
.63677 .65037 m .63322 .69128 L .66413 .70784 L closepath p 1 .815 .718 r
F P
0 g
s
.43345 .56169 m .45136 .56252 L .45136 .55596 L closepath p .983 .716 .769 r
F P
0 g
s
.4326 .55831 m .43345 .56169 L .45136 .55596 L closepath p .936 .623 .68 r
F P
0 g
s
.30605 .6152 m .31078 .63137 L .31738 .62796 L closepath p .554 .516 .889 r
F P
0 g
s
.30605 .6152 m .31738 .62796 L .30983 .61224 L closepath p .332 .434 .902 r
F P
0 g
s
.32889 .88015 m .3558 .8531 L .3214 .88032 L closepath p .827 .643 .842 r
F P
0 g
s
.31613 .89294 m .32889 .88015 L .3214 .88032 L closepath p .829 .652 .853 r
F P
0 g
s
.31613 .89294 m .3214 .88032 L .30917 .89258 L closepath p .933 .767 .905 r
F P
0 g
s
.40977 .55391 m .42311 .56013 L .41936 .55383 L closepath p .854 .405 .375 r
F P
0 g
s
.41936 .55383 m .42311 .56013 L .4326 .55831 L closepath p .852 .404 .378 r
F P
0 g
s
.30925 .91545 m .31613 .89294 L .30588 .90049 L closepath p .694 .446 .682 r
F P
0 g
s
.29176 .92617 m .30925 .91545 L .30588 .90049 L closepath p .67 .43 .682 r
F P
0 g
s
.30588 .90049 m .31613 .89294 L .30917 .89258 L closepath p .931 .827 .982 r
F P
0 g
s
.30814 .71695 m .30737 .72472 L .31905 .75615 L closepath p .079 .079 .498 r
F P
0 g
s
.32732 .56972 m .32248 .57882 L .33125 .57493 L closepath p .447 .185 .315 r
F P
0 g
s
.32732 .56972 m .33125 .57493 L .33466 .56636 L closepath p .445 .186 .318 r
F P
0 g
s
.30558 .58996 m .30382 .59683 L .31315 .59214 L closepath p .502 .249 .557 r
F P
0 g
s
.30558 .58996 m .31315 .59214 L .31358 .58265 L closepath p .484 .313 .67 r
F P
0 g
s
.31881 .90761 m .32889 .88015 L .31613 .89294 L closepath p .687 .459 .71 r
F P
0 g
s
.30925 .91545 m .31881 .90761 L .31613 .89294 L closepath p .755 .505 .711 r
F P
0 g
s
.33795 .89191 m .3558 .8531 L .32889 .88015 L closepath p .749 .535 .763 r
F P
0 g
s
.31881 .90761 m .33795 .89191 L .32889 .88015 L closepath p .766 .539 .751 r
F P
0 g
s
.39248 .55632 m .41288 .56081 L .40977 .55391 L closepath p .823 .415 .458 r
F P
0 g
s
.40977 .55391 m .41288 .56081 L .42311 .56013 L closepath p .863 .431 .42 r
F P
0 g
s
.4374 .56982 m .45214 .57054 L .45136 .56252 L closepath p .999 .761 .817 r
F P
0 g
s
.43345 .56169 m .4374 .56982 L .45136 .56252 L closepath p .99 .744 .803 r
F P
0 g
s
.32804 .79588 m .33428 .80895 L .34462 .82161 L closepath p .103 .134 .705 r
F P
0 g
s
.31078 .63137 m .31497 .64457 L .32027 .63854 L closepath p .509 .43 .811 r
F P
0 g
s
.31078 .63137 m .32027 .63854 L .31738 .62796 L closepath p .511 .43 .81 r
F P
0 g
s
.42311 .56013 m .43345 .56169 L .4326 .55831 L closepath p 1 .664 .64 r
F P
0 g
s
.33466 .56636 m .33125 .57493 L .33906 .57294 L closepath p .609 .225 .396 r
F P
0 g
s
.33466 .56636 m .33906 .57294 L .34216 .56443 L closepath p .607 .224 .398 r
F P
0 g
s
.35433 .86102 m .35661 .85259 L .3558 .8531 L closepath p .038 .038 .038 r
F P
0 g
s
.33795 .89191 m .35433 .86102 L .3558 .8531 L closepath p .126 .126 .194 r
F P
0 g
s
.3558 .8531 m .35661 .85259 L .35574 .84683 L closepath p .051 .051 .282 r
F P
0 g
s
.3558 .8531 m .35574 .84683 L .35432 .84169 L closepath p .076 .076 .421 r
F P
0 g
s
.28798 .6932 m .28829 .6766 L .29153 .68325 L closepath p 1 .873 .395 r
F P
0 g
s
.34216 .56443 m .33906 .57294 L .34867 .5702 L closepath p .587 .206 .389 r
F P
0 g
s
.34216 .56443 m .34867 .5702 L .35029 .56122 L closepath p .561 .212 .433 r
F P
0 g
s
.42461 .56422 m .4374 .56982 L .43345 .56169 L closepath p 1 .755 .764 r
F P
0 g
s
.42311 .56013 m .42461 .56422 L .43345 .56169 L closepath p 1 .771 .779 r
F P
0 g
s
.30382 .59683 m .30983 .61224 L .31995 .60841 L closepath p .79 .701 .95 r
F P
0 g
s
.30382 .59683 m .31995 .60841 L .31315 .59214 L closepath p .749 .702 .981 r
F P
0 g
s
.45214 .57054 m .43947 .57844 L .44382 .59364 L closepath p 1 .823 .892 r
F P
0 g
s
.45214 .57054 m .44382 .59364 L .4508 .58731 L closepath p .935 .739 .864 r
F P
0 g
s
.4374 .56982 m .43947 .57844 L .45214 .57054 L closepath p 1 .815 .885 r
F P
0 g
s
.40743 1.16778 m .36472 1.13538 L .38335 1.15902 L closepath p .467 .662 1 r
F P
0 g
s
.42022 1.18737 m .40743 1.16778 L .38335 1.15902 L closepath p .45 .662 1 r
F P
0 g
s
.50628 .63833 m .49776 .61294 L .45212 .61821 L closepath p .993 .747 .804 r
F P
0 g
s
.57916 .63584 m .49776 .61294 L .50628 .63833 L closepath p .946 .723 .828 r
F P
0 g
s
.37549 .55792 m .39599 .56432 L .39248 .55632 L closepath p .82 .437 .508 r
F P
0 g
s
.39248 .55632 m .39599 .56432 L .41288 .56081 L closepath p .834 .443 .498 r
F P
0 g
s
.45842 1.20174 m .42022 1.18737 L .43135 1.20167 L closepath p .745 .865 1 r
F P
0 g
s
.36267 .56014 m .37403 .56784 L .37549 .55792 L closepath p .799 .407 .482 r
F P
0 g
s
.37549 .55792 m .37403 .56784 L .39599 .56432 L closepath p .816 .422 .485 r
F P
0 g
s
.28992 .93284 m .30925 .91545 L .29176 .92617 L closepath p .586 .306 .568 r
F P
0 g
s
.33341 .80421 m .34462 .82161 L .33695 .80902 L p .02 .02 .45 r
F P
0 g
s
.35062 .83285 m .35148 .83819 L .34462 .82161 L closepath p .788 .4 .042 r
F P
0 g
s
.33695 .80902 m .34462 .82161 L .35148 .83819 L closepath p .859 .448 .069 r
F P
0 g
s
.33695 .80902 m .34758 .83013 L .34446 .81922 L p .085 .085 .514 r
F P
0 g
s
.34446 .81922 m .32963 .76734 L .33695 .80902 L p .085 .085 .514 r
F P
0 g
s
.33695 .80902 m .32804 .79588 L .33341 .80421 L p .02 .02 .45 r
F P
0 g
s
.32804 .79588 m .33695 .80902 L .32869 .7632 L closepath p .138 .138 .699 r
F P
0 g
s
.32804 .79588 m .32869 .7632 L .31905 .75615 L closepath p .131 .131 .662 r
F P
0 g
s
.31358 .58265 m .31315 .59214 L .32207 .58872 L closepath p .719 .503 .745 r
F P
0 g
s
.31358 .58265 m .32207 .58872 L .32248 .57882 L closepath p .723 .502 .739 r
F P
0 g
s
.71826 1.01865 m .71804 1.09335 L .74626 1.01593 L closepath p 1 1 .869 r
F P
0 g
s
.72752 1.00212 m .71826 1.01865 L .74626 1.01593 L closepath p 1 1 .957 r
F P
0 g
s
.74358 .94736 m .72752 1.00212 L .74626 1.01593 L closepath p 1 1 .807 r
F P
0 g
s
.30983 .61224 m .31738 .62796 L .32452 .62405 L closepath p .696 .652 .959 r
F P
0 g
s
.30983 .61224 m .32452 .62405 L .31995 .60841 L closepath p .772 .656 .909 r
F P
0 g
s
.35029 .56122 m .34867 .5702 L .3623 .56985 L closepath p .809 .426 .503 r
F P
0 g
s
.35029 .56122 m .3623 .56985 L .36267 .56014 L closepath p .809 .426 .504 r
F P
0 g
s
.36267 .56014 m .3623 .56985 L .37403 .56784 L closepath p .797 .416 .503 r
F P
0 g
s
.42776 .57353 m .43947 .57844 L .4374 .56982 L closepath p 1 .843 .858 r
F P
0 g
s
.42461 .56422 m .42776 .57353 L .4374 .56982 L closepath p 1 .773 .796 r
F P
0 g
s
.41288 .56081 m .42461 .56422 L .42311 .56013 L closepath p 1 .765 .791 r
F P
0 g
s
.35432 .84169 m .35574 .84683 L .35533 .84182 L closepath p .042 .042 .32 r
F P
0 g
s
.35432 .84169 m .35533 .84182 L .35062 .83285 L closepath p .039 .039 .434 r
F P
0 g
s
.63322 .69128 m .65172 .72707 L .66413 .70784 L closepath p 1 .876 .824 r
F P
0 g
s
.66413 .70784 m .65172 .72707 L .66467 .75391 L closepath p 1 .859 .817 r
F P
0 g
s
.31754 .64694 m .31497 .64457 L .31366 .6497 L p .123 .123 .23 r
F P
0 g
s
.31497 .64457 m .31754 .64694 L .32027 .63854 L closepath p .126 .126 .172 r
F P
0 g
s
.32474 .64208 m .32027 .63854 L .31789 .64588 L p .381 .188 .399 r
F P
0 g
s
.31789 .64588 m .31754 .64694 L .32474 .64208 L p .381 .188 .399 r
F P
0 g
s
.32027 .63854 m .32474 .64208 L .32696 .63526 L closepath p .386 .188 .392 r
F P
0 g
s
.33378 .6377 m .32696 .63526 L .32513 .64089 L p .532 .211 .464 r
F P
0 g
s
.32513 .64089 m .32474 .64208 L .33378 .6377 L p .532 .211 .464 r
F P
0 g
s
.32696 .63526 m .33378 .6377 L .33465 .63197 L closepath p .533 .207 .455 r
F P
0 g
s
.33465 .63197 m .33378 .6377 L .33893 .6359 L closepath p .57 .237 .467 r
F P
0 g
s
.33378 .6377 m .32677 .6355 L .33869 .63589 L p .546 .976 1 r
F P
0 g
s
.33869 .63589 m .33893 .6359 L .33378 .6377 L p .546 .976 1 r
F P
0 g
s
.32474 .64208 m .31885 .63947 L .33231 .63787 L p .655 1 1 r
F P
0 g
s
.33231 .63787 m .33378 .6377 L .32474 .64208 L p .655 1 1 r
F P
0 g
s
.31754 .64694 m .31445 .64867 L .32119 .64435 L p .422 .826 .618 r
F P
0 g
s
.32119 .64435 m .32474 .64208 L .31754 .64694 L p .422 .826 .618 r
F P
0 g
s
.31366 .6497 m .31277 .6532 L .31754 .64694 L p .123 .123 .23 r
F P
0 g
s
.44382 .59364 m .43947 .57844 L .42886 .58688 L closepath p 1 .837 .877 r
F P
0 g
s
.42776 .57353 m .42886 .58688 L .43947 .57844 L closepath p 1 .863 .894 r
F P
0 g
s
.41408 .56567 m .42776 .57353 L .42461 .56422 L closepath p 1 .78 .784 r
F P
0 g
s
.41288 .56081 m .41408 .56567 L .42461 .56422 L closepath p 1 .713 .714 r
F P
0 g
s
.4508 .58731 m .44382 .59364 L .45212 .61821 L closepath p .96 .778 .89 r
F P
0 g
s
.32248 .57882 m .32207 .58872 L .33092 .58575 L closepath p .747 .521 .745 r
F P
0 g
s
.32248 .57882 m .33092 .58575 L .33125 .57493 L closepath p .732 .525 .764 r
F P
0 g
s
.31738 .62796 m .32027 .63854 L .32696 .63526 L closepath p .647 .518 .829 r
F P
0 g
s
.31738 .62796 m .32696 .63526 L .32452 .62405 L closepath p .628 .519 .844 r
F P
0 g
s
.35574 .84683 m .35661 .85259 L .3552 .85247 L closepath p .851 .608 .049 r
F P
0 g
s
.35574 .84683 m .3552 .85247 L .35533 .84182 L closepath p .777 .628 .06 r
F P
0 g
s
.28949 .94836 m .31141 .9184 L .28992 .93284 L closepath p .709 .555 .828 r
F P
0 g
s
.28992 .93284 m .31141 .9184 L .30925 .91545 L closepath p .239 .159 .205 r
F P
0 g
s
.43132 .6074 m .45212 .61821 L .44382 .59364 L closepath p 1 .806 .862 r
F P
0 g
s
.44382 .59364 m .42886 .58688 L .43132 .6074 L closepath p 1 .84 .882 r
F P
0 g
s
.39036 1.13416 m .3459 1.10986 L .36472 1.13538 L closepath p .588 .695 1 r
F P
0 g
s
.40743 1.16778 m .39036 1.13416 L .36472 1.13538 L closepath p .585 .705 1 r
F P
0 g
s
.31315 .59214 m .31995 .60841 L .32728 .60516 L closepath p .777 .718 .98 r
F P
0 g
s
.31315 .59214 m .32728 .60516 L .32207 .58872 L closepath p .819 .724 .955 r
F P
0 g
s
.39599 .56432 m .41408 .56567 L .41288 .56081 L closepath p .96 .68 .744 r
F P
0 g
s
.31661 .91328 m .33795 .89191 L .31881 .90761 L closepath p .121 .121 .121 r
F P
0 g
s
.31141 .9184 m .31661 .91328 L .31881 .90761 L closepath p .097 .097 .097 r
F P
0 g
s
.31141 .9184 m .31881 .90761 L .30925 .91545 L closepath p .268 .159 .185 r
F P
0 g
s
.35062 .83285 m .35533 .84182 L .35148 .83819 L closepath p .798 .355 .058 r
F P
0 g
s
.31995 .60841 m .32452 .62405 L .33296 .62155 L closepath p .815 .683 .907 r
F P
0 g
s
.31995 .60841 m .33296 .62155 L .32728 .60516 L closepath p .759 .674 .941 r
F P
0 g
s
.3495 .87541 m .35433 .86102 L .34764 .87572 L p .082 .082 .082 r
F P
0 g
s
.34764 .87572 m .33933 .894 L .3495 .87541 L p .082 .082 .082 r
F P
0 g
s
.3495 .87541 m .34913 .88783 L .35132 .8751 L p .093 .093 .205 r
F P
0 g
s
.33933 .894 m .35433 .86102 L .33795 .89191 L closepath p .08 .08 .08 r
F P
0 g
s
.3495 .87541 m .35661 .85259 L .35433 .86102 L closepath p .795 .804 .301 r
F P
0 g
s
.3552 .85247 m .35661 .85259 L .35437 .85977 L p .842 .835 .338 r
F P
0 g
s
.35437 .85977 m .3495 .87541 L .3552 .85247 L p .842 .835 .338 r
F P
0 g
s
.3552 .85247 m .35582 .85463 L .3556 .84895 L p .022 .022 .051 r
F P
0 g
s
.3556 .84895 m .35533 .84182 L .3552 .85247 L p .022 .022 .051 r
F P
0 g
s
.35132 .8751 m .3552 .85247 L .3495 .87541 L p .093 .093 .205 r
F P
0 g
s
.31661 .91328 m .33933 .894 L .33795 .89191 L closepath p .089 .089 .089 r
F P
0 g
s
.413 .57796 m .42886 .58688 L .42776 .57353 L closepath p 1 .843 .903 r
F P
0 g
s
.41408 .56567 m .413 .57796 L .42776 .57353 L closepath p 1 .83 .892 r
F P
0 g
s
.71529 .89845 m .71294 .86192 L .66732 .87562 L closepath p 1 .854 .728 r
F P
0 g
s
.66733 .83486 m .66732 .87562 L .71294 .86192 L closepath p 1 .909 .779 r
F P
0 g
s
.66737 .81726 m .66733 .83486 L .71294 .86192 L closepath p 1 .903 .769 r
F P
0 g
s
.66216 .7897 m .66737 .81726 L .71294 .86192 L closepath p 1 .858 .674 r
F P
0 g
s
.66216 .7897 m .71294 .86192 L .68848 .7903 L closepath p 1 .92 .909 r
F P
0 g
s
.33125 .57493 m .33092 .58575 L .33892 .58275 L closepath p .772 .558 .774 r
F P
0 g
s
.33125 .57493 m .33892 .58275 L .33906 .57294 L closepath p .792 .558 .753 r
F P
0 g
s
.51677 .67825 m .57916 .63584 L .50628 .63833 L closepath p .945 .716 .819 r
F P
0 g
s
.57916 .63584 m .51677 .67825 L .57964 .68043 L closepath p .937 .703 .809 r
F P
0 g
s
.57916 .63584 m .57964 .68043 L .63322 .69128 L closepath p 1 .761 .804 r
F P
0 g
s
.37526 1.09511 m .32555 1.06654 L .3459 1.10986 L closepath p .639 .683 1 r
F P
0 g
s
.39036 1.13416 m .37526 1.09511 L .3459 1.10986 L closepath p .645 .7 1 r
F P
0 g
s
.32869 .7632 m .31707 .74173 L .31905 .75615 L closepath p .091 .112 .684 r
F P
0 g
s
.31905 .75615 m .31707 .74173 L .30814 .71695 L closepath p .036 .036 .333 r
F P
0 g
s
.39691 .56851 m .413 .57796 L .41408 .56567 L closepath p 1 .817 .892 r
F P
0 g
s
.39599 .56432 m .39691 .56851 L .41408 .56567 L closepath p .964 .698 .767 r
F P
0 g
s
.44418 1.17974 m .40743 1.16778 L .42022 1.18737 L closepath p .748 .841 1 r
F P
0 g
s
.45842 1.20174 m .44418 1.17974 L .42022 1.18737 L closepath p .748 .865 1 r
F P
0 g
s
.66467 .75391 m .67537 .76624 L .68042 .76511 L closepath p 1 .695 .461 r
F P
0 g
s
.68042 .76511 m .67537 .76624 L .68848 .7903 L closepath p 1 .822 .58 r
F P
0 g
s
.41115 .59545 m .43132 .6074 L .42886 .58688 L closepath p 1 .815 .893 r
F P
0 g
s
.413 .57796 m .41115 .59545 L .42886 .58688 L closepath p 1 .851 .922 r
F P
0 g
s
.31277 .6532 m .28829 .6766 L .30952 .65721 L closepath p .016 .016 .016 r
F P
0 g
s
.31754 .64694 m .31277 .6532 L .30952 .65721 L closepath p .572 .82 .443 r
F P
0 g
s
.45971 .65027 m .46636 .68077 L .50628 .63833 L closepath p .971 .773 .87 r
F P
0 g
s
.50628 .63833 m .46636 .68077 L .51677 .67825 L closepath p .924 .704 .826 r
F P
0 g
s
.50628 .63833 m .45212 .61821 L .45971 .65027 L closepath p .981 .801 .898 r
F P
0 g
s
.32207 .58872 m .32728 .60516 L .33599 .60245 L closepath p .822 .726 .955 r
F P
0 g
s
.32207 .58872 m .33599 .60245 L .33092 .58575 L closepath p .836 .728 .946 r
F P
0 g
s
.32256 .93758 m .31859 .92417 L .28949 .94836 L closepath p .646 .509 .817 r
F P
0 g
s
.32256 .93758 m .28949 .94836 L .2928 .96642 L closepath p .678 .588 .895 r
F P
0 g
s
.28949 .94836 m .31859 .92417 L .31141 .9184 L closepath p .521 .329 .664 r
F P
0 g
s
.33906 .57294 m .33892 .58275 L .34879 .57953 L closepath p .787 .553 .752 r
F P
0 g
s
.33906 .57294 m .34879 .57953 L .34867 .5702 L closepath p .783 .555 .759 r
F P
0 g
s
.37403 .56784 m .39691 .56851 L .39599 .56432 L closepath p .95 .689 .772 r
F P
0 g
s
.32452 .62405 m .32696 .63526 L .33465 .63197 L closepath p .768 .614 .856 r
F P
0 g
s
.32452 .62405 m .33465 .63197 L .33296 .62155 L closepath p .787 .611 .835 r
F P
0 g
s
.70864 .94246 m .72752 1.00212 L .74358 .94736 L closepath p 1 .947 .855 r
F P
0 g
s
.45971 .65027 m .45212 .61821 L .43496 .65342 L closepath p 1 .845 .861 r
F P
0 g
s
.43496 .65342 m .45212 .61821 L .43167 .6182 L closepath p .994 .766 .832 r
F P
0 g
s
.43132 .6074 m .43167 .6182 L .45212 .61821 L closepath p 1 .825 .903 r
F P
0 g
s
.32728 .60516 m .33296 .62155 L .33922 .61949 L closepath p .799 .698 .939 r
F P
0 g
s
.32728 .60516 m .33922 .61949 L .33599 .60245 L closepath p .814 .702 .932 r
F P
0 g
s
.39847 .58238 m .41115 .59545 L .413 .57796 L closepath p 1 .846 .922 r
F P
0 g
s
.39691 .56851 m .39847 .58238 L .413 .57796 L closepath p 1 .819 .897 r
F P
0 g
s
.56921 1.1715 m .51723 1.19162 L .5434 1.21652 L closepath p .896 .967 1 r
F P
0 g
s
.56921 1.1715 m .5434 1.21652 L .59468 1.19766 L closepath p .902 .976 1 r
F P
0 g
s
.3758 .57429 m .39847 .58238 L .39691 .56851 L closepath p .989 .81 .9 r
F P
0 g
s
.37403 .56784 m .3758 .57429 L .39691 .56851 L closepath p .96 .73 .822 r
F P
0 g
s
.41115 .59545 m .42166 .61955 L .43132 .6074 L closepath p 1 .805 .868 r
F P
0 g
s
.43167 .6182 m .43132 .6074 L .42166 .61955 L closepath p 1 .856 .895 r
F P
0 g
s
.45842 1.20174 m .51723 1.19162 L .48985 1.15764 L closepath p .848 .923 1 r
F P
0 g
s
.45842 1.20174 m .48985 1.15764 L .44418 1.17974 L closepath p .842 .915 1 r
F P
0 g
s
.34867 .5702 m .34879 .57953 L .36214 .57621 L closepath p .862 .618 .768 r
F P
0 g
s
.34867 .5702 m .36214 .57621 L .3623 .56985 L closepath p .873 .567 .671 r
F P
0 g
s
.33092 .58575 m .33599 .60245 L .34291 .59966 L closepath p .873 .751 .941 r
F P
0 g
s
.33092 .58575 m .34291 .59966 L .33892 .58275 L closepath p .856 .747 .952 r
F P
0 g
s
.29612 .98783 m .32256 .93758 L .2928 .96642 L closepath p .693 .611 .912 r
F P
0 g
s
.3623 .56985 m .3758 .57429 L .37403 .56784 L closepath p .941 .719 .829 r
F P
0 g
s
.3606 1.04098 m .30771 1.02543 L .32555 1.06654 L closepath p .68 .692 1 r
F P
0 g
s
.37526 1.09511 m .3606 1.04098 L .32555 1.06654 L closepath p .676 .683 1 r
F P
0 g
s
.39974 .59904 m .41035 .61962 L .41115 .59545 L closepath p .998 .798 .873 r
F P
0 g
s
.41115 .59545 m .41035 .61962 L .42166 .61955 L closepath p 1 .802 .873 r
F P
0 g
s
.39847 .58238 m .39974 .59904 L .41115 .59545 L closepath p 1 .845 .917 r
F P
0 g
s
.33296 .62155 m .33465 .63197 L .34169 .62926 L closepath p .814 .63 .836 r
F P
0 g
s
.33296 .62155 m .34169 .62926 L .33922 .61949 L closepath p .774 .631 .875 r
F P
0 g
s
.3623 .56985 m .36214 .57621 L .3758 .57429 L closepath p .94 .623 .673 r
F P
0 g
s
.43496 .65342 m .43167 .6182 L .42166 .61955 L closepath p 1 .79 .824 r
F P
0 g
s
.30814 .71695 m .29927 .70387 L .28798 .6932 L closepath p .025 .025 .551 r
F P
0 g
s
.29927 .70387 m .30814 .71695 L .31707 .74173 L closepath p .79 .421 .04 r
F P
0 g
s
.37624 .59055 m .39974 .59904 L .39847 .58238 L closepath p .974 .815 .924 r
F P
0 g
s
.3758 .57429 m .37624 .59055 L .39847 .58238 L closepath p .98 .831 .939 r
F P
0 g
s
.33599 .60245 m .33922 .61949 L .34552 .61716 L closepath p .88 .748 .931 r
F P
0 g
s
.33599 .60245 m .34552 .61716 L .34291 .59966 L closepath p .871 .744 .936 r
F P
0 g
s
.33892 .58275 m .34291 .59966 L .35243 .59699 L closepath p .91 .782 .947 r
F P
0 g
s
.33892 .58275 m .35243 .59699 L .34879 .57953 L closepath p .879 .778 .97 r
F P
0 g
s
.68848 .7903 m .67537 .76624 L .66216 .7897 L closepath p 1 .885 .866 r
F P
0 g
s
.35533 .84182 m .35582 .85463 L .35148 .83819 L closepath p .949 .571 .091 r
F P
0 g
s
.35036 .98921 m .33144 .95498 L .29612 .98783 L closepath p .691 .614 .917 r
F P
0 g
s
.35036 .98921 m .29612 .98783 L .30771 1.02543 L closepath p .686 .655 .968 r
F P
0 g
s
.29612 .98783 m .33144 .95498 L .32256 .93758 L closepath p .674 .588 .898 r
F P
0 g
s
.29069 .66277 m .29337 .65261 L .28829 .6766 L closepath p .732 .74 .219 r
F P
0 g
s
.28829 .6766 m .29337 .65261 L .30952 .65721 L closepath p .92 1 1 r
F P
0 g
s
.28798 .6932 m .29927 .70387 L .29069 .66277 L closepath p .108 .108 .581 r
F P
0 g
s
.29069 .66277 m .28829 .6766 L .28798 .6932 L closepath p .027 .027 .094 r
F P
0 g
s
.36214 .57621 m .37624 .59055 L .3758 .57429 L closepath p 1 .865 .933 r
F P
0 g
s
.43496 .65342 m .46636 .68077 L .45971 .65027 L closepath p 1 .821 .834 r
F P
0 g
s
.37817 .60785 m .40573 .62108 L .39974 .59904 L closepath p .974 .792 .893 r
F P
0 g
s
.39974 .59904 m .40573 .62108 L .41035 .61962 L closepath p .969 .789 .895 r
F P
0 g
s
.37624 .59055 m .37817 .60785 L .39974 .59904 L closepath p .979 .803 .903 r
F P
0 g
s
.3606 1.04098 m .35036 .98921 L .30771 1.02543 L closepath p .7 .678 .985 r
F P
0 g
s
.34291 .59966 m .34552 .61716 L .35418 .61368 L closepath p .855 .733 .935 r
F P
0 g
s
.34291 .59966 m .35418 .61368 L .35243 .59699 L closepath p .897 .742 .908 r
F P
0 g
s
.31859 .92417 m .31661 .91328 L .31141 .9184 L closepath p .548 .325 .635 r
F P
0 g
s
.33922 .61949 m .34169 .62926 L .34576 .62811 L closepath p .783 .637 .875 r
F P
0 g
s
.33922 .61949 m .34576 .62811 L .34552 .61716 L closepath p .84 .652 .842 r
F P
0 g
s
.36448 .59304 m .37817 .60785 L .37624 .59055 L closepath p 1 .835 .892 r
F P
0 g
s
.36214 .57621 m .36448 .59304 L .37624 .59055 L closepath p 1 .863 .921 r
F P
0 g
s
.31754 .64694 m .30952 .65721 L .31445 .64867 L closepath p .769 .936 .547 r
F P
0 g
s
.33695 .80902 m .35148 .83819 L .34758 .83013 L closepath p .003 .003 .325 r
F P
0 g
s
.32963 .76734 m .33269 .77119 L .32882 .76224 L p .064 .064 .504 r
F P
0 g
s
.32963 .76734 m .32869 .7632 L .33103 .77615 L p .007 .007 .157 r
F P
0 g
s
.32869 .7632 m .3172 .73537 L .31707 .74173 L closepath p .13 .271 .816 r
F P
0 g
s
.32869 .7632 m .32963 .76734 L .3172 .73537 L closepath p .761 .422 .027 r
F P
0 g
s
.32882 .76224 m .3172 .73537 L .32963 .76734 L p .064 .064 .504 r
F P
0 g
s
.33103 .77615 m .33695 .80902 L .32963 .76734 L p .007 .007 .157 r
F P
0 g
s
.66216 .7897 m .67537 .76624 L .65147 .77857 L closepath p 1 .883 .865 r
F P
0 g
s
.66467 .75391 m .65147 .77857 L .67537 .76624 L closepath p 1 .838 .829 r
F P
0 g
s
.63322 .69128 m .63943 .74131 L .65172 .72707 L closepath p 1 .875 .863 r
F P
0 g
s
.65172 .72707 m .63943 .74131 L .66467 .75391 L closepath p 1 .856 .853 r
F P
0 g
s
.43496 .65342 m .42166 .61955 L .41054 .63195 L closepath p 1 .802 .789 r
F P
0 g
s
.41054 .63195 m .42166 .61955 L .41035 .61962 L closepath p .963 .659 .703 r
F P
0 g
s
.35243 .59699 m .35418 .61368 L .36552 .61115 L closepath p .936 .771 .907 r
F P
0 g
s
.35243 .59699 m .36552 .61115 L .36448 .59304 L closepath p .892 .767 .945 r
F P
0 g
s
.34879 .57953 m .35243 .59699 L .36448 .59304 L closepath p .9 .792 .968 r
F P
0 g
s
.33465 .63197 m .33893 .6359 L .34169 .62926 L closepath p .586 .234 .443 r
F P
0 g
s
.34879 .57953 m .36448 .59304 L .36214 .57621 L closepath p .925 .79 .942 r
F P
0 g
s
.34552 .61716 m .34576 .62811 L .35452 .62568 L closepath p .836 .648 .841 r
F P
0 g
s
.34552 .61716 m .35452 .62568 L .35418 .61368 L closepath p .819 .649 .858 r
F P
0 g
s
.31859 .92417 m .33933 .894 L .31661 .91328 L closepath p .628 .377 .641 r
F P
0 g
s
.43726 1.16712 m .39036 1.13416 L .40743 1.16778 L closepath p .762 .814 1 r
F P
0 g
s
.44418 1.17974 m .43726 1.16712 L .40743 1.16778 L closepath p .747 .841 1 r
F P
0 g
s
.66467 .75391 m .64137 .76985 L .65147 .77857 L closepath p 1 .834 .828 r
F P
0 g
s
.63943 .74131 m .64137 .76985 L .66467 .75391 L closepath p 1 .847 .837 r
F P
0 g
s
.57964 .68043 m .61853 .7359 L .63322 .69128 L closepath p 1 .82 .889 r
F P
0 g
s
.63322 .69128 m .61853 .7359 L .63943 .74131 L closepath p 1 .832 .892 r
F P
0 g
s
.36552 .61115 m .37587 .62342 L .37817 .60785 L closepath p .957 .754 .86 r
F P
0 g
s
.37817 .60785 m .37587 .62342 L .40573 .62108 L closepath p .982 .778 .864 r
F P
0 g
s
.36448 .59304 m .36552 .61115 L .37817 .60785 L closepath p .984 .838 .943 r
F P
0 g
s
.41054 .63195 m .41035 .61962 L .40573 .62108 L closepath p .88 .592 .706 r
F P
0 g
s
.34169 .62926 m .33893 .6359 L .34255 .63466 L closepath p .638 .283 .469 r
F P
0 g
s
.34169 .62926 m .34255 .63466 L .34576 .62811 L closepath p .592 .253 .47 r
F P
0 g
s
.41054 .63195 m .40573 .62108 L .37587 .62342 L closepath p .925 .611 .675 r
F P
0 g
s
.35418 .61368 m .35452 .62568 L .36446 .62318 L closepath p .911 .722 .867 r
F P
0 g
s
.35418 .61368 m .36446 .62318 L .36552 .61115 L closepath p .92 .721 .856 r
F P
0 g
s
.36552 .61115 m .36446 .62318 L .37587 .62342 L closepath p .958 .754 .86 r
F P
0 g
s
.32474 .64208 m .31445 .64867 L .31885 .63947 L closepath p 1 1 .859 r
F P
0 g
s
.34576 .62811 m .34255 .63466 L .35265 .63239 L closepath p .687 .348 .525 r
F P
0 g
s
.34576 .62811 m .35265 .63239 L .35452 .62568 L closepath p .686 .35 .532 r
F P
0 g
s
.34361 .90011 m .3495 .87541 L .33933 .894 L closepath p .182 .169 .409 r
F P
0 g
s
.31859 .92417 m .34361 .90011 L .33933 .894 L closepath p .285 .176 .361 r
F P
0 g
s
.71529 .89845 m .68729 .90359 L .70864 .94246 L closepath p 1 .913 .813 r
F P
0 g
s
.71529 .89845 m .66732 .87562 L .68729 .90359 L closepath p 1 .897 .797 r
F P
0 g
s
.46636 .68077 m .4293 .68356 L .42443 .72105 L closepath p 1 .838 .885 r
F P
0 g
s
.46636 .68077 m .42443 .72105 L .47959 .72273 L closepath p 1 .821 .874 r
F P
0 g
s
.47959 .72273 m .51677 .67825 L .46636 .68077 L closepath p .939 .776 .911 r
F P
0 g
s
.4293 .68356 m .46636 .68077 L .43496 .65342 L closepath p 1 .834 .88 r
F P
0 g
s
.34361 .90011 m .34913 .88783 L .3495 .87541 L closepath p .156 .156 .36 r
F P
0 g
s
.35452 .62568 m .35265 .63239 L .36242 .63138 L closepath p .777 .435 .567 r
F P
0 g
s
.35452 .62568 m .36242 .63138 L .36446 .62318 L closepath p .776 .437 .574 r
F P
0 g
s
.36446 .62318 m .37585 .63137 L .37587 .62342 L closepath p .917 .605 .677 r
F P
0 g
s
.41054 .63195 m .37587 .62342 L .37585 .63137 L closepath p .925 .612 .677 r
F P
0 g
s
.36446 .62318 m .36242 .63138 L .37585 .63137 L closepath p .926 .582 .62 r
F P
0 g
s
.33378 .6377 m .31885 .63947 L .32677 .6355 L closepath p .677 1 1 r
F P
0 g
s
.35582 .85463 m .36044 .85124 L .35774 .84681 L p .356 .44 .9 r
F P
0 g
s
.35582 .85463 m .3552 .85247 L .35253 .86803 L p .031 .031 .031 r
F P
0 g
s
.35582 .85463 m .34758 .83013 L .35148 .83819 L closepath p .811 .487 .039 r
F P
0 g
s
.35774 .84681 m .34758 .83013 L .35582 .85463 L p .356 .44 .9 r
F P
0 g
s
.35253 .86803 m .34913 .88783 L .35582 .85463 L p .031 .031 .031 r
F P
0 g
s
.51677 .67825 m .47959 .72273 L .52117 .72302 L closepath p .935 .771 .907 r
F P
0 g
s
.52117 .72302 m .57964 .68043 L .51677 .67825 L closepath p .948 .78 .906 r
F P
0 g
s
.40196 .65418 m .40458 .68432 L .43496 .65342 L closepath p 1 1 .946 r
F P
0 g
s
.43496 .65342 m .40458 .68432 L .4293 .68356 L closepath p 1 .882 .88 r
F P
0 g
s
.41054 .63195 m .41235 .64176 L .43496 .65342 L closepath p 1 .638 .44 r
F P
0 g
s
.43496 .65342 m .41235 .64176 L .40196 .65418 L closepath p 1 .829 .731 r
F P
0 g
s
.33893 .6359 m .32677 .6355 L .33319 .63448 L closepath p .307 .818 .903 r
F P
0 g
s
.33893 .6359 m .33319 .63448 L .34255 .63466 L closepath p .507 .951 .947 r
F P
0 g
s
.44418 1.17974 m .48985 1.15764 L .43726 1.16712 L closepath p .837 .892 1 r
F P
0 g
s
.63943 .74131 m .62495 .75771 L .64137 .76985 L closepath p 1 .791 .861 r
F P
0 g
s
.61853 .7359 m .62495 .75771 L .63943 .74131 L closepath p 1 .824 .88 r
F P
0 g
s
.62538 1.18293 m .56921 1.1715 L .59468 1.19766 L closepath p .956 .992 1 r
F P
0 g
s
.34255 .63466 m .33319 .63448 L .34456 .63823 L closepath p .491 .074 .074 r
F P
0 g
s
.34255 .63466 m .34456 .63823 L .35265 .63239 L closepath p .213 .102 .102 r
F P
0 g
s
.52117 .72302 m .5697 .72547 L .57964 .68043 L closepath p .937 .761 .892 r
F P
0 g
s
.57964 .68043 m .5697 .72547 L .61853 .7359 L closepath p .995 .818 .905 r
F P
0 g
s
.37585 .63137 m .39103 .64123 L .41054 .63195 L closepath p .676 .171 .171 r
F P
0 g
s
.41054 .63195 m .39103 .64123 L .41235 .64176 L closepath p 1 .597 .503 r
F P
0 g
s
.31859 .92417 m .32984 .92905 L .35414 .90523 L closepath p .656 .486 .777 r
F P
0 g
s
.31859 .92417 m .35414 .90523 L .34361 .90011 L closepath p .637 .453 .746 r
F P
0 g
s
.32256 .93758 m .32984 .92905 L .31859 .92417 L closepath p .638 .504 .816 r
F P
0 g
s
.63849 1.12473 m .65125 1.16733 L .71804 1.09335 L closepath p 1 .985 1 r
F P
0 g
s
.69516 1.06274 m .63849 1.12473 L .71804 1.09335 L closepath p 1 .996 1 r
F P
0 g
s
.71826 1.01865 m .69516 1.06274 L .71804 1.09335 L closepath p 1 .989 .932 r
F P
0 g
s
.64741 .80664 m .66737 .81726 L .66216 .7897 L closepath p 1 .827 .778 r
F P
0 g
s
.64216 .78373 m .64741 .80664 L .66216 .7897 L closepath p 1 .825 .777 r
F P
0 g
s
.65147 .77857 m .64216 .78373 L .66216 .7897 L closepath p 1 .882 .862 r
F P
0 g
s
.47959 .72273 m .48802 .73812 L .522 .72757 L closepath p .937 .766 .899 r
F P
0 g
s
.47959 .72273 m .522 .72757 L .52117 .72302 L closepath p .936 .786 .926 r
F P
0 g
s
.44687 .75693 m .48802 .73812 L .47959 .72273 L closepath p .971 .777 .877 r
F P
0 g
s
.47959 .72273 m .42443 .72105 L .44687 .75693 L closepath p 1 .866 .93 r
F P
0 g
s
.35514 .64143 m .39103 .64123 L .37585 .63137 L closepath p .678 .197 .225 r
F P
0 g
s
.35514 .64143 m .37585 .63137 L .36242 .63138 L closepath p .823 .382 .385 r
F P
0 g
s
.35265 .63239 m .34456 .63823 L .35514 .64143 L closepath p .796 .284 .188 r
F P
0 g
s
.35265 .63239 m .35514 .64143 L .36242 .63138 L closepath p .626 .188 .261 r
F P
0 g
s
.61853 .7359 m .6126 .75072 L .62495 .75771 L closepath p .994 .809 .894 r
F P
0 g
s
.5697 .72547 m .6126 .75072 L .61853 .7359 L closepath p .996 .812 .894 r
F P
0 g
s
.41907 1.12733 m .37526 1.09511 L .39036 1.13416 L closepath p .762 .775 1 r
F P
0 g
s
.43726 1.16712 m .41907 1.12733 L .39036 1.13416 L closepath p .764 .814 1 r
F P
0 g
s
.55159 1.15008 m .48985 1.15764 L .51723 1.19162 L closepath p .899 .944 1 r
F P
0 g
s
.55159 1.15008 m .51723 1.19162 L .56921 1.1715 L closepath p .895 .938 1 r
F P
0 g
s
.72752 1.00212 m .68913 .99476 L .67024 1.02558 L closepath p 1 .982 .956 r
F P
0 g
s
.72752 1.00212 m .67024 1.02558 L .71826 1.01865 L closepath p 1 .994 .967 r
F P
0 g
s
.70864 .94246 m .68913 .99476 L .72752 1.00212 L closepath p 1 .941 .895 r
F P
0 g
s
.46128 .78019 m .45409 .80966 L .46603 .78831 L closepath p 1 .872 .935 r
F P
0 g
s
.49243 .75223 m .46128 .78019 L .46603 .78831 L closepath p .97 .855 .978 r
F P
0 g
s
.49243 .75223 m .46603 .78831 L .49458 .76083 L closepath p .937 .808 .954 r
F P
0 g
s
.46603 .78831 m .45409 .80966 L .46448 .81836 L closepath p 1 .881 .939 r
F P
0 g
s
.46603 .78831 m .46448 .81836 L .47714 .80003 L closepath p .96 .817 .941 r
F P
0 g
s
.49458 .76083 m .46603 .78831 L .47714 .80003 L closepath p .942 .816 .959 r
F P
0 g
s
.49458 .76083 m .47714 .80003 L .49759 .77368 L closepath p .87 .734 .923 r
F P
0 g
s
.47714 .80003 m .46448 .81836 L .47554 .81584 L closepath p .921 .764 .913 r
F P
0 g
s
.47714 .80003 m .47554 .81584 L .48126 .80294 L closepath p .75 .612 .87 r
F P
0 g
s
.49759 .77368 m .47714 .80003 L .48126 .80294 L closepath p .774 .607 .842 r
F P
0 g
s
.45435 .76982 m .44813 .80142 L .46128 .78019 L closepath p 1 .946 .945 r
F P
0 g
s
.48802 .73812 m .45435 .76982 L .46128 .78019 L closepath p 1 .94 1 r
F P
0 g
s
.48802 .73812 m .46128 .78019 L .49243 .75223 L closepath p .957 .832 .963 r
F P
0 g
s
.46128 .78019 m .44813 .80142 L .45409 .80966 L closepath p 1 .926 .938 r
F P
0 g
s
.52117 .72302 m .522 .72757 L .56931 .73026 L closepath p .94 .788 .925 r
F P
0 g
s
.52117 .72302 m .56931 .73026 L .5697 .72547 L closepath p .938 .801 .943 r
F P
0 g
s
.62055 .78331 m .64216 .78373 L .65147 .77857 L closepath p 1 .707 .721 r
F P
0 g
s
.62055 .78331 m .65147 .77857 L .64137 .76985 L closepath p 1 .849 .871 r
F P
0 g
s
.62538 1.18293 m .59671 1.15639 L .56921 1.1715 L closepath p .963 .988 1 r
F P
0 g
s
.65125 1.16733 m .59671 1.15639 L .62538 1.18293 L closepath p .978 .991 1 r
F P
0 g
s
.522 .72757 m .48802 .73812 L .49243 .75223 L closepath p .959 .833 .962 r
F P
0 g
s
.522 .72757 m .49243 .75223 L .52305 .73904 L closepath p .882 .704 .872 r
F P
0 g
s
.52305 .73904 m .49243 .75223 L .49458 .76083 L closepath p .923 .8 .956 r
F P
0 g
s
.56931 .73026 m .522 .72757 L .52305 .73904 L closepath p .938 .745 .87 r
F P
0 g
s
.52305 .73904 m .49458 .76083 L .52324 .74952 L closepath p .847 .67 .861 r
F P
0 g
s
.52324 .74952 m .49458 .76083 L .49759 .77368 L closepath p .876 .738 .922 r
F P
0 g
s
.46448 .81836 m .45405 .82145 L .45957 .82748 L closepath p 1 .687 .684 r
F P
0 g
s
.49383 .82782 m .46448 .81836 L .45957 .82748 L closepath p .792 .445 .564 r
F P
0 g
s
.5007 .82339 m .46448 .81836 L .49383 .82782 L closepath p .778 .606 .838 r
F P
0 g
s
.50623 .82272 m .47554 .81584 L .46448 .81836 L closepath p .754 .398 .532 r
F P
0 g
s
.50623 .82272 m .46448 .81836 L .5007 .82339 L closepath p .623 .175 .175 r
F P
0 g
s
.453 .81568 m .45405 .82145 L .46448 .81836 L closepath p 1 .837 .827 r
F P
0 g
s
.45409 .80966 m .453 .81568 L .46448 .81836 L closepath p 1 .855 .853 r
F P
0 g
s
.44687 .75693 m .45435 .76982 L .48802 .73812 L closepath p 1 .922 .998 r
F P
0 g
s
.43726 1.16712 m .48985 1.15764 L .46645 1.11388 L closepath p .842 .864 1 r
F P
0 g
s
.43726 1.16712 m .46645 1.11388 L .41907 1.12733 L closepath p .835 .855 1 r
F P
0 g
s
.30952 .65721 m .29337 .65261 L .31445 .64867 L closepath p 1 1 1 r
F P
0 g
s
.62495 .75771 m .6126 .75072 L .60567 .76239 L closepath p 1 .788 .847 r
F P
0 g
s
.62495 .75771 m .60567 .76239 L .61226 .77023 L closepath p 1 .789 .848 r
F P
0 g
s
.64137 .76985 m .62495 .75771 L .61226 .77023 L closepath p 1 .786 .846 r
F P
0 g
s
.5697 .72547 m .56931 .73026 L .6126 .75072 L closepath p .973 .831 .946 r
F P
0 g
s
.6126 .75072 m .56931 .73026 L .56622 .73785 L closepath p 1 .771 .823 r
F P
0 g
s
.6126 .75072 m .56622 .73785 L .60567 .76239 L closepath p 1 .785 .845 r
F P
0 g
s
.42443 .72105 m .4293 .68356 L .40458 .68432 L closepath p 1 .883 .881 r
F P
0 g
s
.64137 .76985 m .61226 .77023 L .62055 .78331 L closepath p .992 .734 .784 r
F P
0 g
s
.71826 1.01865 m .67024 1.02558 L .69516 1.06274 L closepath p 1 .959 .928 r
F P
0 g
s
.49759 .77368 m .48126 .80294 L .49875 .78824 L closepath p .667 .486 .767 r
F P
0 g
s
.52324 .74952 m .49759 .77368 L .52323 .7632 L closepath p .787 .606 .828 r
F P
0 g
s
.52323 .7632 m .49759 .77368 L .49875 .78824 L closepath p .763 .557 .782 r
F P
0 g
s
.33144 .95498 m .32984 .92905 L .32256 .93758 L closepath p .718 .611 .894 r
F P
0 g
s
.44813 .80142 m .453 .81568 L .45409 .80966 L closepath p 1 .913 .844 r
F P
0 g
s
.56931 .73026 m .52305 .73904 L .56622 .73785 L closepath p .907 .66 .78 r
F P
0 g
s
.56622 .73785 m .52305 .73904 L .52324 .74952 L closepath p .925 .732 .866 r
F P
0 g
s
.43227 .77796 m .44011 .79067 L .45435 .76982 L closepath p 1 .964 .948 r
F P
0 g
s
.43227 .77796 m .45435 .76982 L .44687 .75693 L closepath p 1 .935 .922 r
F P
0 g
s
.45435 .76982 m .44011 .79067 L .44813 .80142 L closepath p 1 .951 .944 r
F P
0 g
s
.31445 .64867 m .29337 .65261 L .3015 .64416 L closepath p .947 1 1 r
F P
0 g
s
.31445 .64867 m .3015 .64416 L .31885 .63947 L closepath p 1 1 .995 r
F P
0 g
s
.65125 1.16733 m .62309 1.13866 L .59671 1.15639 L closepath p .999 .976 1 r
F P
0 g
s
.63849 1.12473 m .62309 1.13866 L .65125 1.16733 L closepath p 1 .979 1 r
F P
0 g
s
.34913 .88783 m .35414 .90523 L .35582 .85463 L closepath p .32 .261 .709 r
F P
0 g
s
.34361 .90011 m .35414 .90523 L .34913 .88783 L closepath p .633 .456 .753 r
F P
0 g
s
.44083 .81437 m .45405 .82145 L .453 .81568 L closepath p 1 .843 .82 r
F P
0 g
s
.44813 .80142 m .44083 .81437 L .453 .81568 L closepath p 1 .901 .896 r
F P
0 g
s
.66733 .83486 m .66737 .81726 L .64741 .80664 L closepath p 1 .848 .816 r
F P
0 g
s
.44011 .79067 m .44083 .81437 L .44813 .80142 L closepath p 1 .947 .908 r
F P
0 g
s
.41235 .64176 m .39103 .64123 L .40196 .65418 L closepath p 1 .768 .698 r
F P
0 g
s
.64216 .78373 m .63284 .79846 L .64741 .80664 L closepath p 1 .814 .795 r
F P
0 g
s
.62055 .78331 m .63284 .79846 L .64216 .78373 L closepath p 1 .74 .761 r
F P
0 g
s
.56622 .73785 m .52324 .74952 L .56555 .74741 L closepath p .905 .683 .817 r
F P
0 g
s
.56555 .74741 m .52324 .74952 L .52323 .7632 L closepath p .911 .706 .844 r
F P
0 g
s
.45957 .82748 m .45405 .82145 L .43665 .82644 L closepath p 1 .673 .634 r
F P
0 g
s
.44083 .81437 m .43665 .82644 L .45405 .82145 L closepath p 1 .815 .768 r
F P
0 g
s
.34943 .93733 m .35414 .90523 L .32984 .92905 L closepath p .721 .576 .846 r
F P
0 g
s
.33144 .95498 m .34943 .93733 L .32984 .92905 L closepath p .699 .596 .889 r
F P
0 g
s
.60567 .76239 m .56622 .73785 L .56555 .74741 L closepath p 1 .776 .823 r
F P
0 g
s
.60567 .76239 m .56555 .74741 L .60356 .76358 L closepath p .989 .845 .946 r
F P
0 g
s
.61226 .77023 m .60567 .76239 L .60356 .76358 L closepath p .964 .791 .903 r
F P
0 g
s
.61226 .77023 m .60356 .76358 L .61121 .77099 L closepath p .979 .783 .875 r
F P
0 g
s
.60356 .76358 m .56555 .74741 L .56352 .7588 L closepath p 1 .781 .825 r
F P
0 g
s
.60356 .76358 m .56352 .7588 L .59783 .77238 L closepath p 1 .773 .814 r
F P
0 g
s
.61121 .77099 m .60356 .76358 L .59783 .77238 L closepath p 1 .77 .813 r
F P
0 g
s
.70864 .94246 m .63577 .94649 L .65773 .99243 L closepath p 1 .934 .913 r
F P
0 g
s
.70864 .94246 m .65773 .99243 L .68913 .99476 L closepath p 1 .894 .891 r
F P
0 g
s
.65991 .90722 m .63577 .94649 L .70864 .94246 L closepath p 1 .952 .934 r
F P
0 g
s
.65991 .90722 m .70864 .94246 L .68729 .90359 L closepath p 1 .918 .851 r
F P
0 g
s
.42443 .72105 m .43227 .77796 L .44687 .75693 L closepath p 1 .87 .896 r
F P
0 g
s
.40402 1.08137 m .3606 1.04098 L .37526 1.09511 L closepath p .763 .743 1 r
F P
0 g
s
.41907 1.12733 m .40402 1.08137 L .37526 1.09511 L closepath p .775 .777 1 r
F P
0 g
s
.62055 .78331 m .61226 .77023 L .61121 .77099 L closepath p .946 .726 .832 r
F P
0 g
s
.61121 .77099 m .59783 .77238 L .6061 .77582 L closepath p 1 .743 .783 r
F P
0 g
s
.62055 .78331 m .61121 .77099 L .6061 .77582 L closepath p .992 .726 .772 r
F P
0 g
s
.3172 .73537 m .29927 .70387 L .31707 .74173 L closepath p .125 .227 .782 r
F P
0 g
s
.45957 .82748 m .43665 .82644 L .45254 .84274 L closepath p 1 .683 .645 r
F P
0 g
s
.48379 .8459 m .45957 .82748 L .45254 .84274 L closepath p .929 .575 .602 r
F P
0 g
s
.49383 .82782 m .45957 .82748 L .48379 .8459 L closepath p .859 .647 .815 r
F P
0 g
s
.31885 .63947 m .3015 .64416 L .31298 .63809 L closepath p .57 .978 .894 r
F P
0 g
s
.31885 .63947 m .31298 .63809 L .32677 .6355 L closepath p .553 .969 .894 r
F P
0 g
s
.52323 .7632 m .49875 .78824 L .52391 .77713 L closepath p .766 .561 .785 r
F P
0 g
s
.35036 .98921 m .34943 .93733 L .33144 .95498 L closepath p .722 .629 .913 r
F P
0 g
s
.56555 .74741 m .52323 .7632 L .56352 .7588 L closepath p .889 .66 .801 r
F P
0 g
s
.42443 .72105 m .38545 .71882 L .39527 .7411 L closepath p 1 .944 .942 r
F P
0 g
s
.42443 .72105 m .39527 .7411 L .43227 .77796 L closepath p 1 .874 .893 r
F P
0 g
s
.40458 .68432 m .38545 .71882 L .42443 .72105 L closepath p 1 .884 .865 r
F P
0 g
s
.41069 .80976 m .42918 .8132 L .44011 .79067 L closepath p 1 .926 .886 r
F P
0 g
s
.41069 .80976 m .44011 .79067 L .43227 .77796 L closepath p 1 .962 .91 r
F P
0 g
s
.44011 .79067 m .42918 .8132 L .44083 .81437 L closepath p 1 .963 .891 r
F P
0 g
s
.35414 .90523 m .36044 .85124 L .35582 .85463 L closepath p .198 .18 .644 r
F P
0 g
s
.56352 .7588 m .52323 .7632 L .52391 .77713 L closepath p .888 .655 .795 r
F P
0 g
s
.34952 .82497 m .32963 .76734 L .34758 .83013 L closepath p .139 .144 .706 r
F P
0 g
s
.36044 .85124 m .34952 .82497 L .34758 .83013 L closepath p .152 .274 .812 r
F P
0 g
s
.42918 .8132 m .43665 .82644 L .44083 .81437 L closepath p 1 .872 .775 r
F P
0 g
s
.62055 .78331 m .6061 .77582 L .59554 .78525 L closepath p .994 .718 .756 r
F P
0 g
s
.59554 .78525 m .6061 .77582 L .59783 .77238 L closepath p 1 .734 .767 r
F P
0 g
s
.63149 .82258 m .66732 .87562 L .66733 .83486 L closepath p 1 .86 .824 r
F P
0 g
s
.66733 .83486 m .64741 .80664 L .63149 .82258 L closepath p 1 .846 .803 r
F P
0 g
s
.59783 .77238 m .56352 .7588 L .56039 .77437 L closepath p 1 .79 .844 r
F P
0 g
s
.59783 .77238 m .56039 .77437 L .58461 .78783 L closepath p .993 .726 .771 r
F P
0 g
s
.59554 .78525 m .59783 .77238 L .58461 .78783 L closepath p .985 .715 .764 r
F P
0 g
s
.32677 .6355 m .31298 .63809 L .32129 .63727 L closepath p .024 .024 .024 r
F P
0 g
s
.32677 .6355 m .32129 .63727 L .33319 .63448 L closepath p .018 .018 .018 r
F P
0 g
s
.62055 .78331 m .6175 .79373 L .63284 .79846 L closepath p .992 .745 .802 r
F P
0 g
s
.60832 .79208 m .6175 .79373 L .62055 .78331 L closepath p .977 .729 .798 r
F P
0 g
s
.59554 .78525 m .60832 .79208 L .62055 .78331 L closepath p 1 .805 .854 r
F P
0 g
s
.63284 .79846 m .63149 .82258 L .64741 .80664 L closepath p 1 .805 .778 r
F P
0 g
s
.38545 .71882 m .40458 .68432 L .36288 .68221 L closepath p 1 .931 .879 r
F P
0 g
s
.40196 .65418 m .36288 .68221 L .40458 .68432 L closepath p 1 1 .967 r
F P
0 g
s
.33319 .63448 m .32129 .63727 L .33492 .63819 L closepath p .312 .073 .073 r
F P
0 g
s
.33319 .63448 m .33492 .63819 L .34456 .63823 L closepath p .446 .061 .061 r
F P
0 g
s
.34456 .63823 m .33492 .63819 L .35514 .64143 L closepath p .405 .056 .056 r
F P
0 g
s
.56352 .7588 m .52391 .77713 L .56039 .77437 L closepath p .903 .684 .821 r
F P
0 g
s
.34952 .82497 m .33269 .77119 L .32963 .76734 L closepath p .078 .078 .509 r
F P
0 g
s
.59671 1.15639 m .55159 1.15008 L .56921 1.1715 L closepath p .957 .959 1 r
F P
0 g
s
.35514 .64143 m .33492 .63819 L .39103 .64123 L closepath p .255 .776 .94 r
F P
0 g
s
.41907 1.12733 m .46645 1.11388 L .44632 1.06319 L closepath p .829 .807 1 r
F P
0 g
s
.41907 1.12733 m .44632 1.06319 L .40402 1.08137 L closepath p .836 .815 1 r
F P
0 g
s
.67024 1.02558 m .63849 1.12473 L .69516 1.06274 L closepath p 1 .956 1 r
F P
0 g
s
.68729 .90359 m .66732 .87562 L .65991 .90722 L closepath p 1 .908 .841 r
F P
0 g
s
.39527 .7411 m .41069 .80976 L .43227 .77796 L closepath p 1 .87 .876 r
F P
0 g
s
.38759 1.02294 m .35036 .98921 L .3606 1.04098 L closepath p .768 .726 .995 r
F P
0 g
s
.40402 1.08137 m .38759 1.02294 L .3606 1.04098 L closepath p .778 .746 1 r
F P
0 g
s
.34943 .93733 m .36448 .89242 L .35414 .90523 L closepath p .62 .485 .805 r
F P
0 g
s
.35414 .90523 m .36448 .89242 L .36044 .85124 L closepath p .747 .652 .923 r
F P
0 g
s
.63284 .79846 m .61844 .81328 L .63149 .82258 L closepath p 1 .775 .785 r
F P
0 g
s
.6175 .79373 m .61844 .81328 L .63284 .79846 L closepath p .99 .706 .741 r
F P
0 g
s
.6175 .79373 m .6097 .80729 L .61844 .81328 L closepath p .995 .71 .74 r
F P
0 g
s
.60832 .79208 m .6097 .80729 L .6175 .79373 L closepath p .968 .676 .725 r
F P
0 g
s
.50827 .83219 m .49383 .82782 L .48379 .8459 L closepath p .798 .577 .776 r
F P
0 g
s
.51243 .82825 m .5007 .82339 L .49383 .82782 L closepath p .734 .534 .777 r
F P
0 g
s
.51243 .82825 m .49383 .82782 L .50827 .83219 L closepath p .74 .689 .972 r
F P
0 g
s
.60832 .79208 m .5985 .80203 L .6097 .80729 L closepath p .952 .665 .728 r
F P
0 g
s
.59375 .78861 m .5985 .80203 L .60832 .79208 L closepath p 1 .779 .801 r
F P
0 g
s
.59554 .78525 m .59375 .78861 L .60832 .79208 L closepath p 1 .779 .802 r
F P
0 g
s
.45254 .84274 m .43665 .82644 L .41874 .83453 L closepath p 1 .697 .692 r
F P
0 g
s
.42918 .8132 m .41874 .83453 L .43665 .82644 L closepath p 1 .879 .848 r
F P
0 g
s
.58461 .78783 m .59375 .78861 L .59554 .78525 L closepath p .993 .733 .782 r
F P
0 g
s
.51557 .82661 m .50623 .82272 L .5007 .82339 L closepath p .72 .273 .327 r
F P
0 g
s
.40196 .65418 m .39103 .64123 L .33492 .63819 L closepath p 1 .793 .795 r
F P
0 g
s
.51557 .82661 m .5007 .82339 L .51243 .82825 L closepath p .763 .384 .491 r
F P
0 g
s
.3678 .84892 m .34952 .82497 L .36044 .85124 L closepath p .731 .724 1 r
F P
0 g
s
.36448 .89242 m .3678 .84892 L .36044 .85124 L closepath p .698 .614 .913 r
F P
0 g
s
.51588 1.0999 m .46645 1.11388 L .48985 1.15764 L closepath p .91 .898 1 r
F P
0 g
s
.51588 1.0999 m .48985 1.15764 L .542 1.13802 L closepath p .902 .888 1 r
F P
0 g
s
.542 1.13802 m .48985 1.15764 L .55159 1.15008 L closepath p .907 .923 1 r
F P
0 g
s
.42918 .8132 m .41069 .80976 L .41874 .83453 L closepath p 1 .903 .853 r
F P
0 g
s
.33521 .64737 m .36288 .68221 L .40196 .65418 L closepath p 1 .872 .882 r
F P
0 g
s
.40196 .65418 m .33492 .63819 L .33521 .64737 L closepath p 1 .695 .658 r
F P
0 g
s
.33269 .77119 m .31777 .73049 L .3172 .73537 L closepath p .107 .107 .643 r
F P
0 g
s
.31777 .73049 m .29926 .69983 L .3172 .73537 L closepath p .128 .244 .795 r
F P
0 g
s
.3172 .73537 m .29926 .69983 L .29927 .70387 L closepath p .119 .203 .762 r
F P
0 g
s
.29069 .66277 m .29927 .70387 L .29926 .69983 L closepath p .062 .062 .398 r
F P
0 g
s
.37617 .94483 m .37526 .92018 L .34943 .93733 L closepath p .752 .616 .873 r
F P
0 g
s
.35036 .98921 m .37617 .94483 L .34943 .93733 L closepath p .737 .641 .917 r
F P
0 g
s
.34943 .93733 m .37526 .92018 L .37553 .89002 L closepath p .738 .593 .854 r
F P
0 g
s
.34943 .93733 m .37553 .89002 L .36448 .89242 L closepath p .834 .705 .919 r
F P
0 g
s
.59375 .78861 m .58848 .79322 L .5985 .80203 L closepath p 1 .77 .817 r
F P
0 g
s
.58461 .78783 m .58848 .79322 L .59375 .78861 L closepath p .996 .749 .802 r
F P
0 g
s
.59671 1.15639 m .58697 1.14136 L .55159 1.15008 L closepath p .956 .96 1 r
F P
0 g
s
.62309 1.13866 m .58697 1.14136 L .59671 1.15639 L closepath p .998 .974 1 r
F P
0 g
s
.3554 .82005 m .33269 .77119 L .34952 .82497 L closepath p .64 .649 .99 r
F P
0 g
s
.3678 .84892 m .3554 .82005 L .34952 .82497 L closepath p .667 .696 1 r
F P
0 g
s
.67024 1.02558 m .61768 1.13226 L .63849 1.12473 L closepath p 1 .926 1 r
F P
0 g
s
.63849 1.12473 m .61768 1.13226 L .62309 1.13866 L closepath p 1 .9 .982 r
F P
0 g
s
.48379 .8459 m .45254 .84274 L .46881 .8532 L closepath p .928 .573 .6 r
F P
0 g
s
.50304 .84971 m .48379 .8459 L .46881 .8532 L closepath p .679 .226 .294 r
F P
0 g
s
.50827 .83219 m .48379 .8459 L .50304 .84971 L closepath p .758 .511 .718 r
F P
0 g
s
.58461 .78783 m .5543 .81483 L .55807 .81712 L closepath p .938 .692 .791 r
F P
0 g
s
.58461 .78783 m .55807 .81712 L .58848 .79322 L closepath p .977 .748 .826 r
F P
0 g
s
.38759 1.02294 m .37749 .97039 L .35036 .98921 L closepath p .791 .731 .983 r
F P
0 g
s
.35036 .98921 m .37749 .97039 L .37617 .94483 L closepath p .752 .66 .928 r
F P
0 g
s
.36448 .89242 m .37553 .89002 L .3678 .84892 L closepath p .844 .742 .957 r
F P
0 g
s
.62309 1.13866 m .61768 1.13226 L .58697 1.14136 L closepath p 1 .899 .988 r
F P
0 g
s
.58848 .79322 m .55807 .81712 L .56547 .82675 L closepath p .965 .728 .812 r
F P
0 g
s
.58848 .79322 m .56547 .82675 L .5985 .80203 L closepath p 1 .775 .837 r
F P
0 g
s
.68913 .99476 m .65773 .99243 L .67024 1.02558 L closepath p 1 .934 .943 r
F P
0 g
s
.65991 .90722 m .66732 .87562 L .60802 .88559 L closepath p 1 .856 .844 r
F P
0 g
s
.589 .85466 m .60802 .88559 L .66732 .87562 L closepath p 1 .819 .806 r
F P
0 g
s
.589 .85466 m .66732 .87562 L .63149 .82258 L closepath p 1 .864 .873 r
F P
0 g
s
.45254 .84274 m .41874 .83453 L .44459 .85771 L closepath p 1 .728 .735 r
F P
0 g
s
.46881 .8532 m .45254 .84274 L .44459 .85771 L closepath p .92 .61 .681 r
F P
0 g
s
.38545 .71882 m .38432 .74922 L .39527 .7411 L closepath p 1 .95 .892 r
F P
0 g
s
.39527 .7411 m .38432 .74922 L .41069 .80976 L closepath p 1 .887 .85 r
F P
0 g
s
.38442 .84427 m .3554 .82005 L .3678 .84892 L closepath p .801 .776 1 r
F P
0 g
s
.39648 .87552 m .38442 .84427 L .3678 .84892 L closepath p .791 .732 .985 r
F P
0 g
s
.39648 .87552 m .3678 .84892 L .37553 .89002 L closepath p .838 .738 .956 r
F P
0 g
s
.5985 .80203 m .56547 .82675 L .57414 .838 L closepath p .913 .633 .731 r
F P
0 g
s
.5985 .80203 m .57414 .838 L .6097 .80729 L closepath p .95 .68 .757 r
F P
0 g
s
.51923 .83314 m .50827 .83219 L .50304 .84971 L closepath p .772 .524 .724 r
F P
0 g
s
.52739 .8297 m .51243 .82825 L .50827 .83219 L closepath p .809 .8 1 r
F P
0 g
s
.52739 .8297 m .50827 .83219 L .51923 .83314 L closepath p .782 .615 .845 r
F P
0 g
s
.61768 1.13226 m .60243 1.09752 L .57415 1.1213 L closepath p 1 .948 1 r
F P
0 g
s
.61768 1.13226 m .57415 1.1213 L .58697 1.14136 L closepath p 1 .959 1 r
F P
0 g
s
.67024 1.02558 m .60243 1.09752 L .61768 1.13226 L closepath p 1 .955 1 r
F P
0 g
s
.37526 .92018 m .39648 .87552 L .37553 .89002 L closepath p .769 .618 .862 r
F P
0 g
s
.29956 .65546 m .3015 .64416 L .29337 .65261 L closepath p .366 .196 .539 r
F P
0 g
s
.29956 .65546 m .29337 .65261 L .29069 .66277 L closepath p .38 .195 .505 r
F P
0 g
s
.33911 .77034 m .31777 .73049 L .33269 .77119 L closepath p .624 .656 1 r
F P
0 g
s
.3554 .82005 m .33911 .77034 L .33269 .77119 L closepath p .623 .64 .988 r
F P
0 g
s
.40402 1.08137 m .44632 1.06319 L .42613 1.00365 L closepath p .82 .769 1 r
F P
0 g
s
.40402 1.08137 m .42613 1.00365 L .38759 1.02294 L closepath p .834 .784 1 r
F P
0 g
s
.6097 .80729 m .57414 .838 L .57901 .84531 L closepath p .984 .733 .794 r
F P
0 g
s
.6097 .80729 m .57901 .84531 L .61844 .81328 L closepath p .983 .732 .793 r
F P
0 g
s
.52891 .82573 m .51557 .82661 L .51243 .82825 L closepath p .691 .286 .403 r
F P
0 g
s
.61844 .81328 m .589 .85466 L .63149 .82258 L closepath p 1 .79 .821 r
F P
0 g
s
.58697 1.14136 m .542 1.13802 L .55159 1.15008 L closepath p .958 .939 1 r
F P
0 g
s
.52891 .82573 m .51243 .82825 L .52739 .8297 L closepath p .811 .507 .65 r
F P
0 g
s
.37363 .81236 m .33911 .77034 L .3554 .82005 L closepath p .825 .77 1 r
F P
0 g
s
.38442 .84427 m .37363 .81236 L .3554 .82005 L closepath p .828 .78 1 r
F P
0 g
s
.61844 .81328 m .57901 .84531 L .589 .85466 L closepath p .975 .719 .784 r
F P
0 g
s
.31777 .73049 m .30171 .69507 L .29926 .69983 L closepath p .454 .56 .983 r
F P
0 g
s
.29926 .69983 m .30171 .69507 L .29069 .66277 L closepath p .322 .397 .868 r
F P
0 g
s
.58697 1.14136 m .57415 1.1213 L .542 1.13802 L closepath p .958 .939 1 r
F P
0 g
s
.38545 .71882 m .36939 .71785 L .38432 .74922 L closepath p 1 .909 .915 r
F P
0 g
s
.38545 .71882 m .36288 .68221 L .35247 .68169 L closepath p 1 .934 .943 r
F P
0 g
s
.38545 .71882 m .35247 .68169 L .36939 .71785 L closepath p 1 .934 .948 r
F P
0 g
s
.51923 .83314 m .50304 .84971 L .5119 .85189 L closepath p .726 .466 .68 r
F P
0 g
s
.53159 .83731 m .51923 .83314 L .5119 .85189 L closepath p .879 .623 .757 r
F P
0 g
s
.52739 .8297 m .51923 .83314 L .53159 .83731 L closepath p .835 .734 .954 r
F P
0 g
s
.5543 .81483 m .52891 .82573 L .52739 .8297 L closepath p .7 .397 .596 r
F P
0 g
s
.41874 .83453 m .43842 .87884 L .44459 .85771 L closepath p .983 .757 .833 r
F P
0 g
s
.41874 .83453 m .41069 .80976 L .39562 .81007 L closepath p 1 .883 .908 r
F P
0 g
s
.41874 .83453 m .39562 .81007 L .40517 .83695 L closepath p 1 .889 .954 r
F P
0 g
s
.41874 .83453 m .40517 .83695 L .42058 .87479 L closepath p 1 .805 .867 r
F P
0 g
s
.41874 .83453 m .42058 .87479 L .43842 .87884 L closepath p .908 .725 .875 r
F P
0 g
s
.41069 .80976 m .38432 .74922 L .37288 .7579 L closepath p 1 .882 .881 r
F P
0 g
s
.41069 .80976 m .37288 .7579 L .39562 .81007 L closepath p 1 .883 .909 r
F P
0 g
s
.33911 .77034 m .3212 .72683 L .31777 .73049 L closepath p .478 .569 .983 r
F P
0 g
s
.3212 .72683 m .30171 .69507 L .31777 .73049 L closepath p .486 .582 .992 r
F P
0 g
s
.46881 .8532 m .44459 .85771 L .45295 .86575 L closepath p .928 .627 .699 r
F P
0 g
s
.49465 .85743 m .46881 .8532 L .45295 .86575 L closepath p .767 .407 .53 r
F P
0 g
s
.50304 .84971 m .46881 .8532 L .49465 .85743 L closepath p .727 .302 .381 r
F P
0 g
s
.31045 .64893 m .31298 .63809 L .3015 .64416 L closepath p .571 .282 .543 r
F P
0 g
s
.31045 .64893 m .3015 .64416 L .29956 .65546 L closepath p .542 .311 .619 r
F P
0 g
s
.58943 1.0397 m .59444 1.07007 L .67024 1.02558 L closepath p 1 .942 1 r
F P
0 g
s
.67024 1.02558 m .59444 1.07007 L .60243 1.09752 L closepath p 1 .965 1 r
F P
0 g
s
.65773 .99243 m .58943 1.0397 L .67024 1.02558 L closepath p 1 .917 .975 r
F P
0 g
s
.29069 .66277 m .30171 .69507 L .30366 .68599 L closepath p .27 .361 .849 r
F P
0 g
s
.29069 .66277 m .30366 .68599 L .29956 .65546 L closepath p .681 .578 .88 r
F P
0 g
s
.37526 .92018 m .40043 .90305 L .39648 .87552 L closepath p .852 .711 .911 r
F P
0 g
s
.37617 .94483 m .40043 .90305 L .37526 .92018 L closepath p .835 .681 .887 r
F P
0 g
s
.35423 .76549 m .3212 .72683 L .33911 .77034 L closepath p .821 .786 1 r
F P
0 g
s
.37363 .81236 m .35423 .76549 L .33911 .77034 L closepath p .817 .768 1 r
F P
0 g
s
.38432 .74922 m .36939 .71785 L .35555 .71778 L closepath p 1 .908 .917 r
F P
0 g
s
.38432 .74922 m .35555 .71778 L .37288 .7579 L closepath p 1 .904 .896 r
F P
0 g
s
.3949 .84026 m .37363 .81236 L .38442 .84427 L closepath p .956 .854 .991 r
F P
0 g
s
.40987 .87923 m .3949 .84026 L .38442 .84427 L closepath p .934 .791 .935 r
F P
0 g
s
.40987 .87923 m .38442 .84427 L .39648 .87552 L closepath p .855 .769 .979 r
F P
0 g
s
.5119 .85189 m .50304 .84971 L .49465 .85743 L closepath p .683 .253 .345 r
F P
0 g
s
.3949 .84026 m .38339 .81099 L .37363 .81236 L closepath p .932 .849 1 r
F P
0 g
s
.40517 .83695 m .38339 .81099 L .3949 .84026 L closepath p .952 .859 .999 r
F P
0 g
s
.40517 .83695 m .3949 .84026 L .40987 .87923 L closepath p .932 .79 .936 r
F P
0 g
s
.52739 .8297 m .53159 .83731 L .5475 .83483 L closepath p .871 .75 .942 r
F P
0 g
s
.55807 .81712 m .52739 .8297 L .5475 .83483 L closepath p .887 .699 .86 r
F P
0 g
s
.5543 .81483 m .52739 .8297 L .55807 .81712 L closepath p .9 .729 .888 r
F P
0 g
s
.40987 .87923 m .39648 .87552 L .40043 .90305 L closepath p .877 .729 .911 r
F P
0 g
s
.57415 1.1213 m .51588 1.0999 L .542 1.13802 L closepath p .949 .906 1 r
F P
0 g
s
.38759 1.02294 m .42613 1.00365 L .40966 .94986 L closepath p .808 .722 .961 \
r
F P
0 g
s
.38759 1.02294 m .40966 .94986 L .37749 .97039 L closepath p .869 .785 .986 r
F P
0 g
s
.40517 .83695 m .39562 .81007 L .38339 .81099 L closepath p .976 .863 .98 r
F P
0 g
s
.40517 .83695 m .40987 .87923 L .42058 .87479 L closepath p .84 .723 .935 r
F P
0 g
s
.53159 .83731 m .5119 .85189 L .53261 .85078 L closepath p .765 .459 .625 r
F P
0 g
s
.5475 .83483 m .53159 .83731 L .53261 .85078 L closepath p .767 .461 .625 r
F P
0 g
s
.3212 .72683 m .30366 .68599 L .30171 .69507 L closepath p .474 .575 .99 r
F P
0 g
s
.44459 .85771 m .43842 .87884 L .45295 .86575 L closepath p .874 .645 .796 r
F P
0 g
s
.32108 .648 m .32129 .63727 L .31298 .63809 L closepath p .804 .534 .702 r
F P
0 g
s
.32108 .648 m .31298 .63809 L .31045 .64893 L closepath p .85 .536 .648 r
F P
0 g
s
.55807 .81712 m .5475 .83483 L .56547 .82675 L closepath p .906 .722 .873 r
F P
0 g
s
.65991 .90722 m .62221 .91173 L .63577 .94649 L closepath p 1 .914 .922 r
F P
0 g
s
.65991 .90722 m .60802 .88559 L .62221 .91173 L closepath p 1 .877 .881 r
F P
0 g
s
.48928 1.05668 m .44632 1.06319 L .46645 1.11388 L closepath p .903 .85 1 r
F P
0 g
s
.48928 1.05668 m .46645 1.11388 L .51588 1.0999 L closepath p .904 .851 1 r
F P
0 g
s
.38339 .81099 m .35423 .76549 L .37363 .81236 L closepath p .931 .829 .985 r
F P
0 g
s
.39562 .81007 m .37288 .7579 L .36313 .7624 L closepath p .98 .852 .964 r
F P
0 g
s
.39562 .81007 m .36313 .7624 L .38339 .81099 L closepath p .976 .851 .967 r
F P
0 g
s
.37617 .94483 m .40966 .94986 L .40043 .90305 L closepath p .847 .696 .895 r
F P
0 g
s
.37749 .97039 m .40966 .94986 L .37617 .94483 L closepath p .84 .729 .944 r
F P
0 g
s
.40987 .87923 m .40043 .90305 L .41836 .90679 L closepath p .801 .647 .872 r
F P
0 g
s
.43842 .87884 m .42058 .87479 L .40987 .87923 L closepath p .874 .813 1 r
F P
0 g
s
.43842 .87884 m .40987 .87923 L .41836 .90679 L closepath p .869 .688 .863 r
F P
0 g
s
.38339 .81099 m .36313 .7624 L .35423 .76549 L closepath p .931 .829 .984 r
F P
0 g
s
.36939 .71785 m .35247 .68169 L .34226 .68123 L closepath p 1 .915 .995 r
F P
0 g
s
.36939 .71785 m .34226 .68123 L .35555 .71778 L closepath p 1 .918 .929 r
F P
0 g
s
.35423 .76549 m .33374 .7185 L .3212 .72683 L closepath p .814 .784 1 r
F P
0 g
s
.33374 .7185 m .30366 .68599 L .3212 .72683 L closepath p .817 .791 1 r
F P
0 g
s
.60243 1.09752 m .59444 1.07007 L .55353 1.08789 L closepath p 1 .944 1 r
F P
0 g
s
.60243 1.09752 m .55353 1.08789 L .57415 1.1213 L closepath p 1 .93 1 r
F P
0 g
s
.41836 .90679 m .40043 .90305 L .40966 .94986 L closepath p .797 .662 .895 r
F P
0 g
s
.33521 .64737 m .33492 .63819 L .32129 .63727 L closepath p .88 .576 .678 r
F P
0 g
s
.33521 .64737 m .32129 .63727 L .32108 .648 L closepath p .87 .587 .71 r
F P
0 g
s
.29956 .65546 m .30366 .68599 L .31848 .67963 L closepath p .798 .664 .897 r
F P
0 g
s
.29956 .65546 m .31848 .67963 L .31045 .64893 L closepath p .743 .648 .921 r
F P
0 g
s
.63577 .94649 m .61824 .99126 L .65773 .99243 L closepath p 1 .922 .968 r
F P
0 g
s
.65773 .99243 m .61824 .99126 L .58943 1.0397 L closepath p 1 .94 .99 r
F P
0 g
s
.5475 .83483 m .53261 .85078 L .56464 .84217 L closepath p .833 .543 .684 r
F P
0 g
s
.56547 .82675 m .5475 .83483 L .56464 .84217 L closepath p .826 .571 .735 r
F P
0 g
s
.56547 .82675 m .56464 .84217 L .57414 .838 L closepath p .9 .633 .746 r
F P
0 g
s
.36313 .7624 m .33374 .7185 L .35423 .76549 L closepath p .936 .849 1 r
F P
0 g
s
.57415 1.1213 m .55353 1.08789 L .51588 1.0999 L closepath p .948 .906 1 r
F P
0 g
s
.37288 .7579 m .35555 .71778 L .34536 .7187 L closepath p 1 .88 .957 r
F P
0 g
s
.37288 .7579 m .34536 .7187 L .36313 .7624 L closepath p .988 .875 .982 r
F P
0 g
s
.5119 .85189 m .49465 .85743 L .50185 .85875 L closepath p .823 .462 .553 r
F P
0 g
s
.53261 .85078 m .5119 .85189 L .50185 .85875 L closepath p .672 .266 .39 r
F P
0 g
s
.36313 .7624 m .34536 .7187 L .33374 .7185 L closepath p .935 .848 1 r
F P
0 g
s
.45295 .86575 m .43842 .87884 L .47381 .88426 L closepath p .874 .645 .796 r
F P
0 g
s
.45295 .86575 m .47381 .88426 L .48871 .87106 L closepath p .868 .646 .804 r
F P
0 g
s
.49465 .85743 m .45295 .86575 L .48871 .87106 L closepath p .861 .59 .725 r
F P
0 g
s
.35247 .68169 m .36288 .68221 L .33521 .64737 L closepath p 1 .869 .861 r
F P
0 g
s
.50185 .85875 m .49465 .85743 L .48871 .87106 L closepath p .852 .58 .721 r
F P
0 g
s
.35555 .71778 m .34226 .68123 L .32811 .67762 L closepath p 1 .894 .97 r
F P
0 g
s
.35555 .71778 m .32811 .67762 L .34536 .7187 L closepath p 1 .894 .972 r
F P
0 g
s
.33374 .7185 m .31848 .67963 L .30366 .68599 L closepath p .852 .8 1 r
F P
0 g
s
.47381 .88426 m .43842 .87884 L .45249 .90545 L closepath p .875 .689 .858 r
F P
0 g
s
.43842 .87884 m .41836 .90679 L .45249 .90545 L closepath p .868 .686 .861 r
F P
0 g
s
.31045 .64893 m .31848 .67963 L .32811 .67762 L closepath p .866 .727 .918 r
F P
0 g
s
.31045 .64893 m .32811 .67762 L .32108 .648 L closepath p .922 .746 .888 r
F P
0 g
s
.34536 .7187 m .31848 .67963 L .33374 .7185 L closepath p .935 .846 .999 r
F P
0 g
s
.32108 .648 m .32811 .67762 L .34226 .68123 L closepath p 1 .814 .846 r
F P
0 g
s
.32108 .648 m .34226 .68123 L .33521 .64737 L closepath p .919 .788 .946 r
F P
0 g
s
.34536 .7187 m .32811 .67762 L .31848 .67963 L closepath p .886 .829 1 r
F P
0 g
s
.53208 .85723 m .50185 .85875 L .48871 .87106 L closepath p .837 .56 .706 r
F P
0 g
s
.53261 .85078 m .50185 .85875 L .53208 .85723 L closepath p .816 .507 .643 r
F P
0 g
s
.33521 .64737 m .34226 .68123 L .35247 .68169 L closepath p 1 .856 .918 r
F P
0 g
s
.41836 .90679 m .40966 .94986 L .42525 .92975 L closepath p .779 .645 .888 r
F P
0 g
s
.45249 .90545 m .41836 .90679 L .42525 .92975 L closepath p .87 .7 .878 r
F P
0 g
s
.57414 .838 m .56464 .84217 L .57368 .84823 L closepath p .947 .722 .826 r
F P
0 g
s
.57414 .838 m .57368 .84823 L .57901 .84531 L closepath p .955 .729 .826 r
F P
0 g
s
.46103 1.0011 m .42613 1.00365 L .44632 1.06319 L closepath p .895 .814 .998 \
r
F P
0 g
s
.46103 1.0011 m .44632 1.06319 L .48928 1.05668 L closepath p .901 .82 1 r
F P
0 g
s
.56464 .84217 m .53261 .85078 L .53208 .85723 L closepath p .811 .503 .642 r
F P
0 g
s
.55353 1.08789 m .48928 1.05668 L .51588 1.0999 L closepath p .941 .865 1 r
F P
0 g
s
.59444 1.07007 m .58943 1.0397 L .53198 1.04984 L closepath p .985 .905 1 r
F P
0 g
s
.59444 1.07007 m .53198 1.04984 L .55353 1.08789 L closepath p .996 .891 .994 \
r
F P
0 g
s
.56464 .84217 m .53208 .85723 L .57368 .84823 L closepath p .939 .73 .847 r
F P
0 g
s
.43656 .95242 m .40966 .94986 L .42613 1.00365 L closepath p .842 .743 .959 r
F P
0 g
s
.42525 .92975 m .40966 .94986 L .43656 .95242 L closepath p .844 .732 .943 r
F P
0 g
s
.63577 .94649 m .55906 .95002 L .5732 .99367 L closepath p 1 .886 .961 r
F P
0 g
s
.63577 .94649 m .5732 .99367 L .61824 .99126 L closepath p 1 .882 .959 r
F P
0 g
s
.63577 .94649 m .62221 .91173 L .54687 .91132 L closepath p 1 .882 .978 r
F P
0 g
s
.63577 .94649 m .54687 .91132 L .55906 .95002 L closepath p 1 .855 .925 r
F P
0 g
s
.57901 .84531 m .57368 .84823 L .589 .85466 L closepath p .954 .728 .825 r
F P
0 g
s
.57368 .84823 m .53208 .85723 L .53608 .87184 L closepath p .933 .713 .829 r
F P
0 g
s
.57368 .84823 m .53608 .87184 L .589 .85466 L closepath p .949 .743 .853 r
F P
0 g
s
.55353 1.08789 m .53198 1.04984 L .48928 1.05668 L closepath p .934 .868 1 r
F P
0 g
s
.48871 .87106 m .47381 .88426 L .48631 .88676 L closepath p .907 .704 .845 r
F P
0 g
s
.54014 .89026 m .53608 .87184 L .48871 .87106 L closepath p .909 .747 .903 r
F P
0 g
s
.54014 .89026 m .48871 .87106 L .48631 .88676 L closepath p .921 .717 .849 r
F P
0 g
s
.53208 .85723 m .48871 .87106 L .53608 .87184 L closepath p .902 .694 .838 r
F P
0 g
s
.43656 .95242 m .42613 1.00365 L .46103 1.0011 L closepath p .895 .794 .975 r
F P
0 g
s
.61824 .99126 m .5732 .99367 L .58943 1.0397 L closepath p 1 .896 .975 r
F P
0 g
s
.60802 .88559 m .589 .85466 L .53608 .87184 L closepath p .974 .806 .913 r
F P
0 g
s
.45249 .90545 m .42525 .92975 L .43656 .95242 L closepath p .909 .761 .921 r
F P
0 g
s
.48631 .88676 m .47381 .88426 L .45249 .90545 L closepath p .904 .732 .887 r
F P
0 g
s
.62221 .91173 m .60802 .88559 L .54014 .89026 L closepath p .984 .854 .962 r
F P
0 g
s
.62221 .91173 m .54014 .89026 L .54687 .91132 L closepath p .997 .806 .884 r
F P
0 g
s
.45249 .90545 m .43656 .95242 L .47639 .95103 L closepath p .9 .752 .918 r
F P
0 g
s
.48278 .90839 m .45249 .90545 L .47639 .95103 L closepath p .911 .756 .913 r
F P
0 g
s
.48631 .88676 m .45249 .90545 L .48278 .90839 L closepath p .911 .746 .899 r
F P
0 g
s
.53198 1.04984 m .46103 1.0011 L .48928 1.05668 L closepath p .932 .834 .989 \
r
F P
0 g
s
.58943 1.0397 m .5732 .99367 L .50331 .99872 L closepath p .966 .869 .997 r
F P
0 g
s
.58943 1.0397 m .50331 .99872 L .53198 1.04984 L closepath p .98 .858 .97 r
F P
0 g
s
.60802 .88559 m .53608 .87184 L .54014 .89026 L closepath p .976 .789 .888 r
F P
0 g
s
.54014 .89026 m .48631 .88676 L .48278 .90839 L closepath p .925 .759 .902 r
F P
0 g
s
.50331 .99872 m .47639 .95103 L .43656 .95242 L closepath p .903 .81 .987 r
F P
0 g
s
.50331 .99872 m .43656 .95242 L .46103 1.0011 L closepath p .922 .806 .966 r
F P
0 g
s
.53198 1.04984 m .50331 .99872 L .46103 1.0011 L closepath p .922 .836 1 r
F P
0 g
s
.55906 .95002 m .54687 .91132 L .48278 .90839 L closepath p .928 .806 .959 r
F P
0 g
s
.55906 .95002 m .48278 .90839 L .47639 .95103 L closepath p .948 .791 .92 r
F P
0 g
s
.54014 .89026 m .48278 .90839 L .54687 .91132 L closepath p .928 .768 .911 r
F P
0 g
s
.5732 .99367 m .55906 .95002 L .47639 .95103 L closepath p .949 .848 .989 r
F P
0 g
s
.5732 .99367 m .47639 .95103 L .50331 .99872 L closepath p .965 .834 .958 r
F P
0 g
s
.25 Mabswid
.57312 0 m
.92808 .50635 L
s
.92808 .50635 m
1 1.35958 L
s
1 1.35958 m
.58259 .92246 L
s
.58259 .92246 m
.57312 0 L
s
.0896 .19777 m
0 1.09651 L
s
0 1.09651 m
.58259 .92246 L
s
.58259 .92246 m
.57312 0 L
s
.57312 0 m
.0896 .19777 L
s
0 1.09651 m
.4888 1.47973 L
s
.10135 1.17597 m
.1021 1.16651 L
s
[(-5)] .09985 1.1949 .07944 -1 Mshowa
.27575 1.3127 m
.27619 1.30322 L
s
[(0)] .27488 1.33167 .04594 -1 Mshowa
.43022 1.43381 m
.43039 1.42431 L
s
[(5)] .42989 1.45279 .01763 -1 Mshowa
.125 Mabswid
.13801 1.20471 m
.13842 1.19903 L
s
.17374 1.23273 m
.17412 1.22704 L
s
.20859 1.26005 m
.20893 1.25436 L
s
.24259 1.2867 m
.24288 1.28101 L
s
.30813 1.33808 m
.30836 1.33239 L
s
.33974 1.36286 m
.33993 1.35717 L
s
.3706 1.38707 m
.37077 1.38137 L
s
.40076 1.41071 m
.40089 1.40501 L
s
.06373 1.14648 m
.06423 1.1408 L
s
.02512 1.11621 m
.02566 1.11054 L
s
.45902 1.45639 m
.4591 1.45069 L
s
.48718 1.47846 m
.48722 1.47277 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[5]:=",
  ImageSize->{194.625, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`400032000181000`40O003h00O`Woo0032Ool00<9o
o`00`Woo0032Ool007=oo`8004eoo`00LGoo0P0000=oo`000000C7oo001^Ool30003Ool2001<Ool0
06aoo`8000Ioo`03001oo`0004]oo`00JGoo0`0027oo00@007ooOol004Yoo`00Igoo0P002goo00D0
07ooOomoo`0004Uoo`00I7oo0`003Goo00D007ooOomoo`0004Uoo`00HWoo0P0047oo00<007ooOol0
0Woo00<007ooOol0AWoo001OOol3000BOol00`00Oomoo`03Ool00`00Oomoo`15Ool005eoo`8001Eo
o`03001oogoo00=oo`03001oogoo04Eoo`00FWoo0`005goo00<007ooOol017oo00<007ooOol0A7oo
001HOol2000JOol00`00Oomoo`05Ool00`00Oomoo`13Ool005Eoo`<001aoo`03001oogoo00Eoo`03
001oogoo04=oo`00Dgoo0P007goo00<007ooOol01Woo00<007ooOol0@Woo001@Ool3000QOol00`00
Oomoo`07Ool00`00Oomoo`11Ool004ioo`8002Aoo`03001oogoo00Qoo`03001oogoo041oo`00Bgoo
0`009Woo00<007ooOol027oo00<007ooOol0@7oo0019Ool2000YOol00`00Oomoo`09Ool00`00Oomo
o`0oOol004Ioo`<002]oo`03001oogoo00Yoo`03001oogoo03ioo`00@goo0`00;Woo00<007ooOol0
2Woo00<007ooOol0?Woo0011Ool2000aOol00`00Oomoo`0;Ool00`00Oomoo`0mOol003ioo`<003=o
o`03001oogoo00aoo`03001oogoo03aoo`00?7oo0P00=Woo00<007ooOol037oo00<007ooOol0?7oo
000iOol3000hOol00`00Oomoo`0=Ool00`00Oomoo`0kOol003Moo`8003]oo`03001oogoo00ioo`03
001oogoo03Yoo`00=7oo0`00?Goo00<007ooOol03goo00<007ooOol0>Goo000bOol20010Ool00`00
Oomoo`0?Ool00`00Oomoo`0iOol002moo`<0049oo`03001oogoo011oo`03001oogoo03Qoo`00;Goo
0P00AGoo00<007ooOol04Goo00<007ooOol0=goo000ZOol30017Ool00`00Oomoo`0AOol00`00Oomo
o`0gOol002Qoo`8004Yoo`03001oogoo019oo`03001oogoo03Ioo`009Goo0`00C7oo00<007ooOol0
4goo00<007ooOol0=Goo000SOol2001?Ool00`00Oomoo`0DOol00`00Oomoo`0dOol0021oo`<0055o
o`03001oogoo01Aoo`03001oogoo03Aoo`007Woo0P00E7oo00<007ooOol05Goo00<007ooOol0<goo
000LOol2001FOol00`00Oomoo`0FOol00`00Oomoo`0bOol001aoo`8005Ioo`03001oogoo01Ioo`03
001oogoo039oo`0077oo00<007oo0000EGoo00<007ooOol05goo00<007ooOol0<Goo000LOol20000
0goo001oo`1COol00`00Oomoo`0HOol00`00Oomoo`0`Ool001aoo`05001oogooOol0001COol00`00
Oomoo`0HOol00`00Oomoo`0`Ool001aoo`05001oogooOol0001COol00`00Oomoo`0IOol00`00Oomo
o`0_Ool001Ioo`03001oogoo009oo`03001oogoo00=oo`03001oogoo03Ioo`X0011oo`03001oogoo
01Yoo`03001oogoo02ioo`005Goo00<007oo00000goo00<007ooOol017oo00<007ooOol0<goo1000
00Qb^@00N]Qf]g[HM[Qjf7Jg200027oo00<007ooOol06goo00<007ooOol0;Goo000EOol00`00Ool0
0003Ool00`00Oomoo`05Ool00`00Oomoo`0`Ool800000gZhN]Qf^002N[P017[HM[Qj^7Zh0P0000=j
f000OaT01@000goo00<007ooOol06goo00<007ooOol0;Goo000EOol00`00Ool00003Ool00`00Oomo
o`06Ool00`00Oomoo`0/Ool;00001W[HM[Qjf7JgN]Qf^0<00006N]QjfGKH001o67lI0P0000=nnGlI
000010007Goo00<007ooOol0;7oo000EOol00`00Ool00003Ool00`00Oomoo`07Ool00`00Oomoo`0Z
Ool<00000gJhN[Qj^00300000g[HN]Ujf002N]P02P00OaQo6000OcT007lIO_Uo6Gki0P0000Ao6P00
000001]oo`03001oogoo02]oo`005Woo00<007ooOol00Woo0P002Goo00<007ooOol0:7oo2`0000=Z
F000N]P00P0001Qjf7[IM]QjfG[HN]Uff7[I001o6@00OcUo6000O_Qo6Gki001o6@00OaUo>WlIOcX3
000HOol00`00Oomoo`0[Ool001]oo`03001oogoo00Uoo`03001oogoo02Moo`P00003ISP0000000L0
0006N]Ujf7[HN]QjfG[H0P00015o>GlHOcT007lIO_T007lIOcX007lJOaUo6WlIOaYo6GlJ00<001Io
o`03001oogoo02Yoo`006goo00<007ooOol02Woo00<007ooOol09Goo0`0000EEm`00000005hG00@0
0007ISL007ZhM[Mj^7JgN[P0200000mo67liOaP007ki001o6GljOaT007lIOcYo6GljOaT00P0000=o
6@00OeT00P005Goo00<007ooOol0:Goo000KOol00`00Oomoo`0:Ool00`00Oomoo`0TOol01`00:=@0
0000FOP005hG00<00003HSL00000008000Uj]`<00004OaP00000OcT2000027lIOcYo6GlJ001o6WlI
OaX2000017kiOaUo6@000gli0P004goo00<007ooOol0:Goo000KOol00`00Oomoo`0;Ool00`00Oomo
o`0SOol200000aba001Em`0:00001gZhM[Mj^7JgN[Qf]gZh00800006N]Qnn@00OcUo67li100000Eo
6GljOaT007lI0080000;OaYnnGlIO_Uo6P00OcUo>7mIOcQo>@02000BOol00`00Oomoo`0XOol001Yo
o`03001oogoo00eoo`03001oogoo025oo`04000Xe0007;8300000eWf000000060006N[L01P00O_Un
nGkiO_P000=o60800003OcYo6@0000D00003OaUnnGlI009o6@03O_T007li00Eo>@<0015oo`03001o
ogoo02Moo`006Woo00<007ooOol03Woo00<007ooOol087oo0P0000@L/P00001A]@T00005M[Mj^7Jg
N[Qf]`0200002GkiN]QnnG[HO_T007liOaQo>@02000017lIOcYo6Glj0P0000=o6Wlj00000`0000En
nGlI001oFGlh00@00003OcUo60000080011oo`03001oogoo02Ioo`006Woo00<007ooOol03goo00<0
07ooOol07Woo00@002CD000L/P`000Aj]`03001nn7ki009nn@04O_QnnGki0002OaP02@00OeT007lj
OaYo>P00OcX00006OaX50002OcP027lHOcQo>7lh001oF7mG000?Ool00`00Oomoo`0VOol001Yoo`80
015oo`03001oogoo01eoo`03000Xe00000d0000NM[Mj^000001jf7kiN]QnnG[HO_Ujf7ki001o>@00
OeYoF@00OaUo>P00OcYo6@00OaYo>WlJOcYo6Wlj0`0000aoFP00000007liOaQo>GlH001oEgmhOeL2
000>Ool00`00Oomoo`0UOol001Yoo`03001oogoo015oo`03001oogoo01aoo`<000039=@0000000H0
0003JUL0000000<00003O_Unn7ki009nn@05O_QnnGkiO_Unn0020003OeT01000OcYo6@000Wlj00<0
07lJOaX00glJ00H007lJOaX007liOeT2OcT300001glh001oF7mGOeQoEgmH008000eoo`03001oogoo
02Aoo`006Woo00<007ooOol04Woo00<007ooOol06goo0`0000<Xm00000001P0000=VE`00JUL01`00
00Ijf7kiN]QnnG[HO_T20002OeT01Gli001o6@00OaT00Wlj00@007lJOcYo6P80000;OaYo6GlJ001o
FWliOeYo>GmJOcUoFP0:000<Ool00`00Oomoo`0TOol001Yoo`03001oogoo01=oo`03001oogoo01Uo
o`<00003:=@0012=00@00003GOH0000000@00006O]QjegkHN]Mnf7[G2@000WmI0P0017lj00@007lJ
OaX000Eo6P06001o>GliOcUoFGli0P0000Ao>@00000000EoE`<000]oo`03001oogoo02=oo`006Woo
00<007ooOol057oo00<007ooOol067oo0`0000DTe00036d004VE00D0000=JUL006YG001jegkHN[Mn
f7[GO]Qj]gkHN]L00P0000Ao6@00OeQo>0D0009o>P0>OaT00000001o6GlJOaUo6WlIOaX007mJOcUo
FP800006OaQo>GlH001oEgmh0P0000Io=WmGOeL007lGOcP2000:Ool00`00Oomoo`0ROol001Yoo`03
001oogoo01Aoo`03001oogoo01Moo`09000@U0009=@Xe00048d004fE00L00009JWP007kHN]Mnf7[G
O]QjegkH00800004O_Uo6Gki0003OcP20002OaT80004OaX03000OcT00000OaUo67liOaQo6@00OeQo
E`9oF0800003OeL007lh00=o608000Qoo`03001oogoo029oo`006Woo00<007ooOol05Goo00<007oo
Ool05Woo00L001Bd000XmBCD000<K@0;00001GZgO]QjegkHN[L00P0000mo6GkhOaUnnGlI001oFGlh
001o>000OaUo>GlHOcT00P0000Ao6GljOaYo6PL0000=OcUo67liOaQo>GlH001oEgmhOeMoN7mGOgP0
2@0027oo00<007ooOol08Goo000IOol00`00Oomoo`0GOol00`00Oomoo`0EOol300000bCD000@S@09
00001VYG000007kHN]Mnf080000=O_Uo6GkiOaUnnGlIO_T007lh001o>7mH000017lI0P001glJ0P00
00=o>WmJ00000`0000=o67li00000WmH100000UoE`00000007lHOaMo67lG00002Goo00<007ooOol0
87oo000IOol2000IOol00`00Oomoo`0COol400000bSd000@SP09000016Yh000007[G0P0001Eo6Gki
OaUnn7lIO_Uo6GkhOaT007mH001oFGlhOeT007liOaT007li00000WlJ00=o6GljOaX00P0000Qo>`00
Oe]o>WmKOcYoFglj1@0000Qo=gmGOcMoF7lg001o5glh0P0000Ao5glhOaL000Uoo`03001oogoo01mo
o`006Goo00<007ooOol06Goo00<007ooOol04Woo300000EN5`00000006YH00P00006OaUnnGlIO_Uo
6Gki0P0000MoF7lhOeQo>000OaT00002OeX00`00OaYo6P02OaX01@00OcYo>glj00000WmJ00Ao>WmJ
OeYoFPD000EoE`03001o>7lH009o60800003OaP007oo00Qoo`03001oogoo01moo`006Goo00<007oo
Ool06Woo00<007ooOol04Goo300000=N5`000000100000Inf7khO_Qnn7kHO_P8000017lhOeUo>7mI
0P0001=o>GmJOcT007lIOcYo6P00OaYo>glJOc/007mKOcYoFgljOe/00002OaT03`00OeQoN@00OeQo
=gmGOcL007lGOcQo5glhOaMo>002000:Ool00`00Oomoo`0NOol001Uoo`03001oogoo01]oo`03001o
ogoo011oo`H000035;<0000000X000Ynn0<00005OcP00000001o>00200001GmJOcUoFP00OaX00P00
00=o>WlkOcX00Wlj00@007ljOeYoFP8000=o6@04001oF7mHOgT20002OeL01@00OaQo67lhOaP00P00
00=o6000Ool02Woo00<007ooOol07Goo000IOol00`00Oomoo`0LOol00`00Oomoo`0>Ool200000a2C
0000000300000`a[0000000200000f8h000000040002O_P01WkHO_Qnn7khO]Qnn0800007O_Qo6@00
OcQo=gmHOcL0100000moFWli000007lkOaYo>glJOc]o6Wlk001oFglj00000glI00Qo>WlI001oF7mi
OeQoNGmH0`0000Qo5`00000007kgOaQo5`002goo00<007ooOol07Goo000IOol00`00Oomoo`0MOol0
0`00Oomoo`0=Ool200001aBd000000006;<000a[00X00003KWP007kh00Ann0800005O_Unn7lIO_P0
0004OcP200000glHOcP000060002OcX00glkOcX000020006OaT00`00OgUoF002OeP00gmiOeQoF002
0003OaL017lHOaMo5`0037oo00<007ooOol077oo000IOol00`00Oomoo`0MOol00`00Oomoo`0=Ool6
00000aBC0000000:00000fjI001nn002O_P2000047khOaUnn7lIO_Qo6Gkh001o=glh001oF000OcUo
67li0P0000Ao6GlJOaUo6PH000=o6@0:OcYo6GlIOaT007mHOgUoF7miOeP3000027lG000007lHOaMo
67kg000=Ool00`00Oomoo`0KOol001Uoo`03001oogoo01ioo`03001oogoo00aoo`L0000336/00000
00L00005JWP006ih001nn00200003glIO_QnnGkhOaUnn7kiO_Qo6@00OcP007mHOcP00002OcP02P00
OeT007lJOaUo6WlIOaYo6GlJ0P0000Ao6@00000000Ao6@04001oF7mHOgT20002OcL01000OaQo5glH
0`0000=o6000Ool03Goo00<007ooOol06Woo000IOol2000POol00`00Oomoo`0;OolH00008GkhOaUn
n7lIO_Qo6GkhOaUnn000OcL007lh001o67li001oFGlh001o6GlJOaUo6WlIOaX007lj001o>WlIOcYo
6@0300000glI001oF00200002glgOaMo>7lg001o5glHOaMo67lGOaP00P003Woo00<007ooOol06Woo
000HOol00`00Oomoo`0QOol00`00Oomoo`0:Ool200000aBd0000000800001EhG00000000JWP00`00
00=neWkGO]H01P0000=o6GkhO_T00P0000Io>7mH001o>7lH0002OeT04@00OaYo6GlJOaUo6P00OaYo
6@00OaUo>WlIOaYo6GljOaT00`001Wlg00H007lHOaMo67lGOaP3000?Ool00`00Oomoo`0IOol001Qo
o`03001oogoo029oo`03001oogoo00Uoo`80000349@00000008000034900000000/00009O]MneWkG
O[InegkFO]Mn]WkG00@00007OcQoF7lg001o6000OcP00WmI00@007lIOaYo6@80000;OcYo6GlJ001o
>WlIOcYo6GljOaUo>P0400003GlGOcQo=glgOaL007lGOaQo5glH001o67kg008000moo`03001oogoo
01Qoo`0067oo00<007ooOol08goo00<007ooOol027oo0`0000DHe@000000012@00/00005FOH007kF
O]MneP060002O_P01GlH00000000OeP00P0017mI00l007lJOaT007lIOaYo6GlJOaT007lIOaYo6Glj
OaT00P0000Ao6@00OeUoF08000=o=`03001o67lG0080009o5`03OaP0000000moo`03001oogoo01Qo
o`0067oo00<007ooOol097oo00<007ooOol01goo0`0000@LeP00000<K`L00003IUL0000000L0000<
O_Qo67khOaQnn7lHO_Qo67kh001o5Wlg0`000gmI01Qo>000OaT007lIOcYo6GlJOaUo>P00OcYo6Glj
OaT007kiOaYo6@00OcQoFGlhOeT200002Glg00000000O_Mo67kgOaQnm`02000@Ool00`00Oomoo`0G
Ool001Qoo`03001oogoo02Eoo`03001oogoo00Ioo`<000036=D0000000h00003OU5jDGia00@000=n
n004OaQnn7kh0002OcL200001Glg00000000OeT00P0000ao6GlJOaUo6WlIOaYo6@00OaUo>WlI0004
OaT01`00OeUoF7mIOeQoFGmH00<000=o5`03OaQo5glG0080015oo`03001oogoo01Ioo`0067oo00<0
07ooOol09Woo00<007ooOol01Woo3@0000AYd000LPmf<0P00003NU40000000800003OaQnn0000080
0006OaH007lGOcMo5Wlg100000Qo6GlJOaUo>WlIOaX007lj0P0000]nnGlJOaUo6Gki001o>7mIOcQo
FGlh00<00007OaL00000OaQnmglHO_L00P004Goo00<007ooOol05Woo000HOol00`00Oomoo`0VOol0
0`00Oomoo`05Ool300000cbX0000000=0002NRl=0004OcL01000OcMo=glG0P000Wkh100000Ao6WlI
000000Mo6@09001oFGmHOeUoF000OaMo600000=o60<00003OaP00000019oo`03001oogoo01Eoo`00
67oo0P00:7oo00@007ooOomoo`L00003HSL0000000H00003Na]fnP0000@00003NRij;`0000H00004
OY=nLWjCOW83000017lGOcL007lg0P0000=o5P00O_P00Wkh00=o67khO_P01@0000Uo6WlIOaUnnGlJ
OaT007lhOeT00P0000Yo5glhOaL007kgOaQo5glHO_Mo60<001=oo`03001oogoo01Aoo`0067oo00<0
07ooOol0:7oo00@007oo000hZ@H00003HSP0000000L00003N_/007lJ01000003OY8007j@00@000=o
=`03001nn7kh00=nn0800004OaT007lIOaT40002OaT30002OaL01WlHOaMo5`00OaQo5`=o60@001Ao
o`03001oogoo01=oo`004Woo00<007ooOol00goo00<007ooOol0:7oo2@0000QV>000001bV000KYP0
07[I0P0000=nn@00Na/00P0000=o6P00OaT00P0000=o>P0000003P0000Mo5WlgOaL007khOaQnn002
00002WljOaUo>@00OaYo6GlJO_Uo6WlI100000io>7lGOcQo5glhOaL007lGOaQnmglH001o>7lG0P00
57oo00<007ooOol04goo000AOol00`00Ool00003Ool00`00Oomoo`0YOol300000eWg0000000>0000
17[i000007lJ0P0000=o>WlIOcX00P0000=o6Wlj0000100000=nL0000000100000IndP00001o=`00
O_P200000glIOcUo6@02OaT00`00OaUo6@04OaT300000gli00000003OaL017lH001o67lH0P000WlH
00=o>00000005Goo00<007ooOol04Woo000COol01@00Oomoogoo0000:Goo1P0000=ImP0000000`00
00=Zg000L_`03P0000Ao>WlIOcYo6@800003OcX0000000@00003OTmnL00000@0000@O]<00000001o
>GlIOcYo6GliOaUo>P00OaYnnGlJOaT200001WlJO_T007lhOcUo>0<00009OaL00000001o67lhOaMo
>7lH008001Ioo`03001oogoo015oo`004Goo0P0017oo00<007ooOol09Woo00<000Q200000P0000=I
m`0000001@0000=Zg00000001P0000=k?7ll00000`0000=o6`0000000`0000Eo6Wlj000007lj0080
0006O_H00000001nL7i@1P0000Ane@000000009o6@03OcUo6@00009o6@80009o6@05O_Uo6@00OcUo
>003OcT017lh000000000glH00=o>7lHOaP00P005Woo00<007ooOol04Goo000AOol00`00Oomoo`03
Ool00`00Oomoo`0VOolA00000gLL0000000200000gmMNcao?@0200001G/KOc/00000O_L02`0000Mn
mP00000007i?OW1nC`0400000gkFO[Ene@03000047li00000000OaUnnGlJO_Uo6Gki001o>7liOcQo
>Glh0`0000Mo>000001o>7lHOcQo5`02000GOol00`00Oomoo`0@Ool0015oo`<000=oo`03001oogoo
02Eoo`L00003I[`0000000H00003Ma`007LM00@00004Nc`00000M_X300001Glk00000000OaP00P00
00=o>Wlk00001P0000Io5`00000007i`OU040003O]D60002OaT01GkiOaUo6GlI00000gli0P000Wli
00@007liOcUo>@<00003OcP0000001Qoo`03001oogoo00moo`005goo0P009Woo00<003eg00007P00
00Ao>`00OaQnm`800004Oc]o>WlkOcX500001GkfOaL00000OTl01P0000Io5gkgOaL007khO]L40000
17lJO_T007lh0P0000]o>GlhOcUo>000OaQo>GlhOcUo67li00<001Qoo`03001oogoo00moo`005goo
00<007ooOol097oo1@0000ANW@00001RW0X00003Lad0000000@0009ff@80009ne`H00003O_Qo6000
00800003Oc]o>Wlk00@000=nmP@00003OVh00000009o5`08O_L007kgO_QnegkhO_Mnn0@00009OcQo
>GliOcUo>7li001o>Glh00=o>@@001Uoo`03001oogoo00ioo`005goo00<007ooOol097oo0P0000<]
F0000000100000=RW0000000200000=^^fZJ00000P0000Aff@00001jfP<00003O_P007kG00L00005
OaP00000OcYo>`0500000glGO_H0000400002Wi]001nmglGO_L007kGO_Qnegkh0P0000mne`000000
07lhOcUo>7liOcP007lhOcUo67li00000WlH0`006Goo00<007ooOol03Goo000GOol00`00Oomoo`0T
Ool@00000g;l001^M`0<00000g[i00000003O]L:00000glk0000000800000gi]001nK@0200000glG
001nn00300001WkhO_Mnn000OaMo5P8000=o>@03001o>Gli008000=o60<001Yoo`03001oogoo00ao
o`005goo00<007ooOol097oo1P0000=FWP0000001@0000=bo00000001P0000IYl@00000007:GM[P9
00000gkg0000000300000glI0000000300000glH001o600300000glFO_H0000700002gkhO]Mnn7kG
O_Qnegkh001o5glFOaL00P0000Ao>000000000=o6003OcUo6000008001Yoo`03001oogoo00aoo`00
5goo00<007ooOol08goo00D0012B000mo47l00<00003EYh0000000H00009D>/005D]001MKP00KO00
078A00<00006MYMf^000001jUWZf0`0000=o>P0000001`0000=o6GlH0000200000=o5P0000000P00
00=n9`0000000`0000Ynn7kGO_Qnm`00OaIo5glFOaMo5P<000Io6003001oC00001Yoo`8000eoo`00
5Woo00<007ooOol097oo00<0012B00003@0000Y<b`00GB/005e]001]l78A001f<P80000:NW8007:G
001fU`00MYH007jCO[@2000017lJOc]o6Wlj1P0000=o67lI0000100000=o>00000001@0000En9`00
000006TU00<00003O_P007lG00@00003OaL007lH00@0009o6003001oB`0001Moo`<00003Ool007oo
00aoo`005Woo00<007ooOol097oo00<0012B00002`0000=0Z`00C>`0300000]jL`00001j]gJg001j
UP00OY=n/gjC00<00003OaYo>P00009o5`@00003OaT007kf00/00003ORP0000000L00003OcQo=glh
009o>003001o67lH00=o60D001Aoo`<000Aoo`03001oogoo00]oo`005Woo00<007ooOol097oo1`00
00=B6P0000005`0000Yj]gJG001fUP00OY=n]7jCO[AnT`800006OcX007lHO_Mo67kg1P0000=o5Wke
00001@0000=o=00000000`0000=ndGk@O]400`0000Mo>7lGOcQo=`00OaQo>003OaP300000glZ001o
o`0@Ool30007Ool00`00Oomoo`0;Ool001Ioo`8002Eoo`@00003>GL0000000T00003K]/0000000D0
0003OaX0000000L0000;N[Mf]`00001n]7jCO[=nTgjdOY=n/`030003OaL00gkg0000000500000gkE
O_Ene@0400000glD000000030003O]401`00O]800000OcMo>00000=o6080009o60<000moo`8000Yo
o`03001oogoo00]oo`005Woo00<007ooOol097oo2`0000=JFP0000005P0000YfU`00001n]7jCO[An
TgjdOY=n]0H00003O[AneGjd00D00003O]EnmGkE00800003Oa<0000000800009OTX00000001nd000
O[9ndWjb00<00007OaP00000OcUo67lhOaP00`0037oo0`0037oo00<007ooOol02goo000FOol00`00
Oomoo`0TOol400000ceg0000000700000fJm0000000300000gLm0000000500000gll000000080000
0gjCO[AnT`040003O[@00`00O[En]@03O[D200000gkD001ne00;00000gi:OVX000020004O]830000
0glHOcQo6002OaP017lh000000002Goo0`0047oo00<007ooOol02Woo000FOol00`00Oomoo`0TOolE
00000gKk0000000B000027jCO[AnTgjdOY=n]7jCO[@200000gjdO]En]00200001WkEO]@007jcO]An
/`800003O_80000000T00006O]800000Oa@007kf0P0000Mo>7lHOcUo6000Ofh00006Ool3000COol0
0`00Oomoo`0:Ool001Ioo`03001oogoo02=oo`X00003FYd0000001d00003O_L00000008000In]005
001n]GjeO[D00003O]@20002O]@400000gk?0000000200000gj:0000000200001GleOaAo57lD0000
0Wkf0`0000Ao6000Ohl000=oo`<001Ioo`03001oogoo00Yoo`005Woo00<007ooOol08goo200000=B
O@0000002P0000=USP0000004@0000IjnP00O_Mo67kgOaP400004GjCO[AnT`00O[Ane@00O]Ene7kE
O]Ane@00O]An/`00O]40100000En[P00000007j80080000:OaAo=GlDOcD007kgO_Io5gkfO_L7000I
Ool00`00Oomoo`0:Ool001Ioo`03001oogoo02=oo`L00003EWd0000000`0009U[a40000:Oa]k6P00
O_Mo67kgOaQnmglHO_L70006O]@300000gkaO]40000300001Gk>00000000OVP00P0000Ao57leOa@0
00EnmP800003Ofh007oo01]oo`03001oogoo00Yoo`005Goo00<007ooOol097oo2@0000A:?000000L
Q1400003O_P0000000<0000COe/00000OaYo>`00Oa]jnP00O_Mo67kgOaQnmglHO_Mo67kgOaP04@00
00=o4@0000000P0000=nIgj800000P0000Qo=@00OaMnmWkgO_H007kf0`0077oo00<007ooOol02Woo
000EOol2000UOolE00000g[I0000000200000glk0000000600005GmK001oF`00Oc]o6P00NaYo6g/J
001nmglHO_Mo67kgOaQnmglHO_Mo600600001GjdO]An]7kD00000Wkb100000=o4@0000001`0000=o
=@00O_H00P0000=o5WlG00000P0077oo00<007ooOol02Woo000EOol00`00Oomoo`0TOol600000cFI
0000000700000fih0000000400000g/k001o>`0400000gml00000003000067mK001o>glJOc/007lK
NaYo6`00OaQnmglHO_Mo67kgOaQnmglH001n]P00O_Ine@D0000?O[<00000001n/7kA001ncP00Oa40
07lB001nRGj800H00005OaMnmWlGO_Io5`02000LOol00`00Oomoo`0:Ool001Eoo`03001oogoo02Ao
o`@00003:GT0000000X00005Mad00000001kG@0400000gmL0000000700000gmK001o6P02OaX03P00
NaYo6g/JOa/007lHO_Mo67kgOaQnmglH0002N[H00`00O]Ene@02O]D00`00O[=n/`020004O]400`00
O/incP0200001WlA000007j8001nZ@@00005OaIo5glGOaMo5P02000LOol00`00Oomoo`0:Ool001Eo
o`03001oogoo02Aoo`0300048@0001800003Ncd0000000<0000TOeaoO@00001o>`00M]X007mJOg/0
07lkOaYo>`00Oa]jnWlKNaYo6`00OaQnmglHO_Mo6000O[IjUGjf001nmWkEO_IneGkf0P0000Qn/`00
000007jaO]5n/0000gk>0P0000Eo3P00OXT007jY00800006O]@00000OaMnmWlG0P0077oo00<007oo
Ool02Woo000EOol00`00Oomoo`0TOolA00000g<N00000002000017/lNcd007mL0`0000=o?0000000
0`00019oF`00OaYo6WlkOaX007/JOa]k6WlKNaYo6`00OaQnmglH0002N[H00gjfNYD00007O]D30000
1GkA001ndGkA00000gk>0P0000QnkGl>000007j9001nc0000WkD0P0000=o5`00000077oo00<007oo
Ool02Woo000EOol00`00Oomoo`0TOolE00000g/K0000000400000gllOc/0000200005GKk001oF`00
OaYo>glJOc]o6P00NaYo6g[jOa]k6WlK001o6000O[IjU@0C000017k>000007l@1P0000=nR@000000
0gkD100077oo00<007ooOol02Woo000EOol00`00Oomoo`0TOol700000bfl0000000;00000g/l0000
000800001Wlj00000000OaYo>`=o6P08Oc/007lKNaYo6g/JOa]k6PD00005NYAjUGZE001n]0030005
OY<00`00O]5ndP04000017k_Oa1o3gl@2`0000Ao5Wke000001eoo`03001oogoo00Uoo`005Goo00<0
07ooOol097oo4@0000=fo00000000P0000=k?7/KOc`00`0000Eo>glJ001o6Wlk00<00009Oc]o6Wlk
OaYo>glJ001jnWlK00@0000=MYMj^000NYEjM7ZEMWAjU@00O]En]7kEO[@0100000=n/`00O_800P00
00]nl`00000007k_Oa1nk`00O^d007k]00800003O/X007kC00800004O_Eo5Wke000MOol00`00Oomo
o`09Ool001Eoo`03001oogoo02Aoo`H00003AQ/0000000D00003K_d0000000@00004Na/00000Na/2
00000glk0000000300000glk001o>`03OaX60003M[L01GJGM[Mf]gJg00000gZE00YjU000O[An]Gjd
O[En]7jeO[An]@<000=o4`0:001o<Gla001nkgl@001o3P00O^d3000017jZ001ndgkC0P0000=o5P00
000077oo00<007ooOol02Goo000DOol2000VOolE00000g[jOa/0000?000017[hO_Ujf7ki1@0000Ej
^7JGN[QfUgZh00800009NWAjU@00O]En]7kEO[AneGjd00800004O]D000000003Oa<02000Oc9o4@00
O^l007k]O`h200001gk/00000000O[9ndgjb00@001aoo`03001oogoo00Uoo`0057oo00<007ooOol0
9Goo300000=RFP000000200000=fMG9DMWD00P0000Eo>000001o6Wki00800004Nc`000000003O_T0
17kh001ffGKJ0WKI10000WJg00MfU`00NYD007jdO[En]0030003O]D00`00Oa=o4`0200000glC001o
4@02Oc4300001Wk]000007k]001oD@@0009o5@8001aoo`03001oogoo00Uoo`0057oo00<007ooOol0
9Goo2@0000<mV@0000000`0000=AC`0000002@0000Io67li000007[iOaX2000017/KOcak6gll1000
00Ujf000M]UjfWKIN_YffG[JM]T01P0000Ene@00001n]GkE01800003O^d0000000<00003OaH00000
009o5@03OaH007oo01]oo`03001oogoo00Uoo`0057oo00<007ooOol097oo5@0000A@YP00001DY@80
0004OTl007lHOcP3000017mKOeX007/K0W/l00=o?7/KNc`05@0017jd0`000WlC00Mnl`00001o4Gl@
Oa5o400200000gl=O^d0000800000glE001oo`0KOol00`00Oomoo`09Ool001Aoo`03001oogoo02Ao
o`03000aN00000L00003JQ<0000000800003O]@0000001000005OeYoF`00Ocak6`0400001G/K0000
07JiN]X00WKI00If^G[JM]UffGJiN]X400000gkF00000003000017jdO[D007lC0P0000mnlWlC001o
<@00Oa5o47la001o3`00O`h007l>O^d00P0000=nTGj@00000`0077oo00<007ooOol02Goo000DOol0
0`00Oomoo`0TOol00`00;EP0000500000d4c0000000=00001GlH001o5@00OeH0100000=oFWmK0000
0`0000Qjng/KNa]k6`00M]Ub^@0027KI00/007JiM[T007kh001nmWkgO_Inmgkf00<0009o4P@00008
Oa000000001o47lA001o3`800003O^io3Wk^00@00004OcEo5@00000KOol00`00Oomoo`09Ool001Ao
o`03001oogoo02Aoo`03000aN00000X00005Mch00000Oe]oO002Og/?00003gmJ001k6g[jNa]jnW/K
001ffW:iM]X007KIM[UjfP02M]T047Ji000007KIL[Qff@00OaUnn000O]InmgkFO_MneWkg100000]o
=GmF001o<Gl@001o<`00Oa0007l>00800003O`ink@0000<00003OcEo5Gle008001]oo`03001oogoo
00Uoo`0057oo00<007ooOol097oo00@002eH000LmPD00005GW`00000001c7@030007Og/30004Ood3
0002OcX00`00N_]k6`02Na/01P00M]Ub^GKJL[T000Aff@8000=f^@03M]Uf^@00009nn@07O_P007kf
O_MnmWkgO_H00P0000Ao<`00001o=@800004Oc5o4000Oe<20002O`h30002OX`01`00OXh00000OaEo
=@0001]oo`03001oogoo00Uoo`0057oo00<007ooOol097oo00@0035h000Lm`@00003HY/0000000<0
0006Lad00000001oFgml0Wmk01YoF`00OoiooP00OoiooGon001oW000Oe/007mJOcT007[jNa/007KJ
L[UffW:iM]X007[J0WKI018007KIM[UffG:hM]Uf^GKI001o6GkhOaUnn000O]InmgkFO_L200000glC
Oc@0000200001Wme000007l@001o<`D00005OV]nS7i[001nK@02OXh3000KOol00`00Oomoo`09Ool0
01Aoo`8002Aoo`800004;EP001SF7?H800000glj000000020004Og/01@00OoiooWon00000Wom00<0
07mk00000gmK00`007lj001k6`00M]Ub^GKJL[UffG:iM]X300001GKIM[Uf^GJiM]T00gJi00<007kh
O_T00Wki00Unn000O_Inmgkf001o=P00Oa<00Wlc00H007mE001o<@00Oa030002O`h30002OX`00`00
OXinSP02OXh00`00Oomoo`0JOol00`00Oomoo`08Ool001=oo`03001oogoo02Aoo`800004<GP]F000
6=H400000fJl00000004000017mL000007mk0P000Won00aooGon001ooP00OiaoN`00Oc]oG7lkOe/3
000017KJL[UffW:i2P0001EffGJiM]Ub^7KI001o6GkhOaUnn7lIO_P007kFO_L007lf001o=7lCOc@0
0P0000=oE@00Oa400P0000eoE000001o3Wl?001o4`00OX`007j>OVenSP02000LOol00`00Oomoo`08
Ool000Ioo`D0009oo`03001oogoo00=oo`03001oogoo02Aoo`800003;EPaF00000D00003HY/00000
00/000=ooP800003Og/0000000`000=g708000Eg70P00008O_QnnGkiO_Unn000O_H0009o=P03001o
<glC00<00003OeD007la00<00006Oe4007l>001o4WlC1@0000=o4@00Ool06goo00<007ooOol027oo
0008Ool01`00Oomoogoo001oo`0000Aoo`03001oogoo02Aoo`04000]F35h;EP300001@QA00000000
Jal00`0000=oOP0000000P0000Ao_P00Ok]oV`H00007Oi]oN`00Oc]oG7lkOe`00P00019fo7LMM_ag
77KlMaefo000L_`007LLNaafng/LMaak77KkNa`400000glk00000003000027kh000007lfOcD007lC
Oc@200001Wlb001oE@00Oa5o<@80000;Oc400000Oa=o4WlCOa8007lAOc5o4@02000KOol00`00Oomo
o`08Ool000Qoo`07001oogooOol007oo000017oo00<007ooOol097oo00P0035H;EPaF0009CT000ab
0`0000=[?f/O00000P0000=oOP00Oil00`0017nk00<007mk00000gmk00D007mKOe]oF`00009o>`03
001g77LL00Ag7003001g700000Mg7003001fo0000080009o>`04Ocao>glkOc/30003OcH20002Oc<0
0`00Oc80000200000gla0000000400000glC00000002000017lAOc5o4@006goo00<007ooOol027oo
0008Ool01`00Oomoogoo001oo`0000Aoo`03001oogoo02=oo`800003;EP00000008000039ET00000
00<00003Kcl0000000D0000VOkh00000Oi]o^gnK001oNP00Og]oW7mk001o>gmL001oFgljOe/007LL
M_ag7GKlMa`007LLM_`007KkNaag77/LM_]k7000Maabn`800009Oc`007mLOc]o?7lkOeao>gll00<0
0003OcH00000009o<`0:001o<glbOc8007la001o<000Oe43000027nfOiD007mg001o4Gla000KOol0
0`00Oomoo`08Ool000Qoo`07001oogooOol007oo000017oo00<007ooOol08goo0`0000@TmBSe000e
E`D00003Ial0000000@00003Oi`000000080009o^`07001oNWmkOg/007mk00000WmK00<007lkOc/0
0glk00T007LLMaag7000Maafo7LL000017LL0P0000=fo7KkM_`00P000Wll00@007lkOc]o?09o>`<0
0004OgT000000003Oc<01000Oc9o<Wlb0P0000Ao;gl`001oD@80009oUP06001oUgmg001o4@006goo
00<007ooOol027oo0007Ool20004Ool00`00Oomoo`03Ool00`00Oomoo`0SOol300001BSe000eEcDf
=EL01`0000QgG`00Oih007nmOi`007nm0P0000Io_000Oi]oNWnKOgX200006WmL001oFgljOe]o>gmK
OcX007KlMa`007LLM_ag77;l001g77/LM_/007KkMaabngLLM_/2000027mLNcao?000Oeao>gllOc/2
00001Wmk001oNWmi001o<P9o<`03001o4Wlc009o<P<0000:Oc000000Oc4007nE001oMgnGOgL2000K
Ool00`00Oomoo`08Ool001=oo`03001oogoo02=oo`03000/m33d00800003=CH0000000D00003Ial0
000000@00004Okd00000Okd20003Og/00gmjOg/000020007Oc/01@00Maag7000M_`00gLL00`007LL
Ma`007KkM_afngKlM_/007ll0003Oc`02P00Oc]o>glk001o>`00OgX007mi10000Wlc00<007lbOc80
0`0000=nkP0000000P0000=o<@0000000WnG0`006goo00<007ooOol027oo000COol2000TOol01000
<A@/m34D0P0000<iVCfj00001@0000=W7`000000500000EoFgljOe]o>gmK00800005M_ag77;lMaaf
o00300005GKkMaabngLLM_]g7000Oe`007llNc]oG7/l001o>gmL001oG7lk001oNP0300003GmhOiP0
07lc001o<P00Og<007md001o3gk^00L00004OcD00000000JOol00`00Oomoo`08Ool001=oo`03001o
ogoo029oo`80008`m0@00004>ITi^@005=L400000fLO001gG`030002Oi`01000Oi`0000017nm0`00
00=oOGmMOgd060001Gll00X007lk001o>gmL001oN`00OgT000=oV0@00003Oe<007mC00<00006O^h0
0000Oe8007mE1@006Woo00<007ooOol027oo000BOol00`00Oomoo`0SOol02000;?@a5000>EL004Fh
@IL800001FLO00000000Ogh00P0000=oW000Oi`00P000Wnm0P0000ioO000001oG7mmOeaoOGmLOgeo
G000N_/007;k100000Ebf`00M_]ffWKk00/00006Oe`00000Oc]oG7lk0`0000moF@00OgMoV7mhOiQo
M`00Oe=oLgmC001oD`00Oe<00P0000Qo3P00000007kc001oE7lc0P006Goo00<007ooOol027oo000B
Ool00`00Oomoo`0SOol00`00<?@00002>EH02000@IM5V4FGAKP001CG4=L400000fLO001gG`020000
0gnK0000000200002Gnl000007mkOgaoN`00Og/00002Ogd017mMOgeoO@000W/K0P0000Eg?W]N001^
ffjk00`00003Ocao>gll00@00005Oc]oG000001oFP020003OiP200002Gng001oDgmc001oLgmC001o
D`040002O]<00`00Oc=o=002000JOol00`00Oomoo`07Ool0019oo`03001oogoo02=oo`800007>ELe
ESUG0015^46GAKP00`0000DDm`00000006/o00H0000COg]oW000001o^`00Og]oO7mKOg`007ml001o
OGmLOgeoG000N_/00`0000Ag?G]N001bf`<00003M_]bfP0000@00003O_T0000000<00005Nc]o?7/K
001o>`0300004GmJOg/00000OgQoV000OgP007nhOiL007mC001oDgmdOe<00`0000Mnc000O]8007lc
OeAo<`02000JOol00`00Oomoo`07Ool0019oo`03001oogoo029oo`80008iEP04>ELiEP00AIL20000
0e9L000Dm`0600000fLO0000000200000gnLOi]oW004000017mkOgaoN`000Wml00<007mmOgd01@00
00=g?P0000003@0000=o6@0000001@0000=o>`00Oc/0100000IoNP00000007mh0004OkL200000gmc
Oe=oL`0200000gmC001nc0020002Oc@017mD001jY0006Goo00<007ooOol01goo000BOol00`00Oomo
o`0ROol300000cUG=EHiE`020002DU`0158l000DeQSg100000AW7`00001gG`H0000;OgX007mK001o
NgmlOe/007mK001oG00200001Gno000007LmNeh01`0000=bfW:jM]X00P0000=^V000OaT00P0000Ao
6P00OaYo6P800007Oc`00000OaX007mIOeX00P0000QoEgmhOgL007nGOkQoUgnh200000QoD`00001o
E7lcOeAo<gmD0P006Goo00<007ooOol01goo000BOol00`00Oomoo`0ROol00`00>EHiEP040005DU`0
0`005?LDm`030002Ial50003Og/02000OgX007mk001oNgml0002Og`00`00Oimo_`02000017LnNeh0
07Ln0P0000E^g000HUT007:j00h00006OaX00000OaYo6`000WmJ1@0000=o]`0000000P0000QoMP00
001oU7nEOg@007me0`0000=o=7mDOc<00Wld0P006Goo00<007ooOol01goo000BOol00`00Oomoo`0R
Ool01@00>EHi=SUF=CH0100000EFG59LDU`001Sg00D00003Jal000000080000?Ned007mkOg/007nK
001oN`00OgaoN`00Oe/007nl00800005Okl007]N001g?P0200000fkL001RF@0700000g:i00000009
00000glJ0000000400000gmJOcT0000200001GmeOgIoEGmfOgD00`0000EoM7nE001oUGmd00@00003
OeAo<gmD00<001Qoo`03001oogoo00Moo`004Woo00<007ooOol08Goo0P000cUF00H004XK000005IL
0002DU`200000bVl0000000300000fLO00000004000017mk001oNP000Wmk00P007ml000007nlOkd0
07no0`0000=g?P0000001P0000=^V00000000W:i00@007lJOaYo6P800003OaT007lI00L0009oF@03
001oF@0000<00007OgIoMGmfOgEoMP00OiD00P0000IoU000Og@007nE0002OgH00`00OeAo<`03000H
Ool00`00Oomoo`07Ool0019oo`80029oo`0;000<K`00=CH004XK001B>`00DUaFG00400000bVK0000
0003000016KoIal007MO0`0000MoW7mk000007mkOe]oN`0800001G]N00000000J]`05@0000=je`00
N]P00P0000=oF`0000000`0000YoV`00000007mfOgEoMWmE001oM080000<OiD007nE001oU@00OgMo
EWmg001oE7lc0P0067oo00<007ooOol01goo000BOol00`00Oomoo`0QOol02@00490003UF001>6`00
DU`00002DU`00eILDU`00002:K`00bVk0000000300000fKo0000000400000gmk00000004Og/20000
27oM00000000Neh007LmMch70002N_X01P00L]YbfP00L]]bfP<00005O]L00000001ne`080002Og/0
1@00OgYoN`00Oe<00P0000EoMGmf001oMGmd00@00003OiEoM00000AoMP<00003OXD007oo01Ioo`03
001oogoo00Moo`004Goo00<007ooOol08Woo00`000a_000004hkBQ/0058l001B>eILDU`300000bVk
;K`0000300000fKo001Vo`0200002gmn00000000Oi`007mkOe]oN`00Om`01P0000=g?GLn00001000
00eo6W[iO_X007lJ001ffg;JM]/007;K001^^`0=00001Wmk000007mjOi]oNP80000GOe=o<P00OgD0
07mdOiD007mc001oUGmdOiD007mgOgIoMgmFOgMoMP00OVD0000GOol00`00Oomoo`07Ool0015oo`03
001oogoo029oo`03000@T00000=>6`09001BG59L001BG000DS`0058k00H00003JamVo`0000<00003
Neh007ml00<00003Og/007oL009og0H00003Mch0000000800005O_YjnG[jN_UnnP02000017;JM]Yb
fW;J0P0000E^^P00N_T007lH00P00003OcP0000000P00006Oc<00000OiEoM7me0`000WnE00<007mf
OgH00Wmf1@005goo00<007ooOol01goo000AOol00`00Oomoo`0ROol200001c6k001>?4XK001B>eIL
00800005DS/00000001B>`0200001a2F00000000Jal006/O00800003NeeoOP0000<00003Oe/007nl
01d00003O_T0000000<00006N[H00000OcT007mH0`0000Io_7nLOk`007ngOiL500002Gmd000007mc
001oU@00OgMoEP0200001WmEOcD00000OXD001Ioo`03001oogoo00Moo`004Goo00<007ooOol08Woo
0P0000D]VP00BQ]>6`00009BG0D00003BUh0000000D00003Iam[7`000080009kGP03001oO00000@0
0003Oi/0000000800003Ogd007/m00800003O_Uo6@0000800004O_YjnG[jN_X300001W;KM_/00000
N_T0009o6P03001jUP0000800003O]L0000000<00003OcP0000000=oW003Ok`007ng009o]`800003
OiD0000000800003Og<00000008000Ao=@04001o;000000GOol00`00Oomoo`06Ool0015oo`03001o
ogoo025oo`800003;IXa^bfJ00800003EUaB?00000<00005G]mN_P00BUi>OP0500000f/OI_m[7`03
00000g]N00000003000027nlOi/007nk001oN`00Ogd400001GlIO_T007[iO_X00P0000EjnGlJN_T0
07;I00800003M[T0000000<00004N[P00000N[H200001glH001o>000OeQo=gmH00800007Oiao_7nL
001oUgngOiL00`0000=oM7nE0000100000UoEGleOeEo=GmEOcD007l;Ob`00P005Woo00<007ooOol0
1Woo000AOol00`00Oomoo`0QOol200000c6k;IX]V`020000159L000004hJ1`000T[O0P0000ABW`00
IamW7`@00005Ngh00000001oVP0?000017kiN_UnnG[i0P0000MjnP00M]X007Jh001jf@0200001G9f
001f]`00N]P00P0000Eo6000OcP007mI00800004OcP007mD0002Oi`01P00OkMoUgng001oV0800005
OiD00000001oL`030005OcD01000O``000005Woo00<007ooOol01Woo000AOol00`00Oomoo`0QOol0
1P00<K/]VS6k;IXa^`<00004DQ/00000FSX=00001fLO00000000Neh007mm00800005Oi/007nK001o
>00E00000g[i0000000200001WJG000007ki001o>P<00003OeUoF7mI00<0000=OgD007nlOiao_000
OkMoU`00OiMoV000OiD00P0000eoM000Oe<007le001o=GmEOcEoE@00Obao2`02000EOol00`00Oomo
o`06Ool0015oo`80029oo`h00004GUX00000GY`200001EJo000005:o00000fLO0`0000MkGP000000
07nJ001oNP0200000gjcOY<00002HK4?000017;K001bMW9f100000=nn@0000000WlI0P000WmI00=o
F000OeT00P0000EoE@00OkaoW000009o]`03001oV7nH00<00003OgD007md0080009oEP80009o=@03
001o;000008001Eoo`03001oogoo00Ioo`004Goo00<007ooOol08Goo0P0000<]NC6J;GT00`0000Q5
U`00000005hiHUX006[M2P0000EVof/O000007MO00800003Oih0000000800003O[D0000000l00003
K[Ybffjj00@00003MYL007ki00@00004OaT00000Og/7000037nH001oM@00Ok`007ng001oV7nGOkQo
U`800004OiD007nEOg@400000gmF0000000200000gk[O`/0000EOol00`00Oomoo`06Ool0015oo`03
001oogoo029oo`04000aVRei<IT200001DFG00000000DOT01@0000AZN000001Ng`X00003Mel00000
00@0009n]@800008MU800000JYU^VFYi001R>0X00003ISL0079g00@00005N]P007lI001oF`03Og/0
1000OgUoFGmi0P0000AoUgnH001oE@800005OiL007nHOiQoU`02OiP70002OeH40003O^/2000DOol0
0`00Oomoo`06Ool0015oo`03001oogoo029oo`800003<IX0000000@00003COUB6@0000800004IUP0
06iiJWP5000015[O001RofLO1@0000Eo>glJ000007kg00H00003LS80000000@00003HQH000000080
0004GK8006HEJSD600003W[H001o6@00Oe]oNgmJOg]oF`00OeUoNWmIOgX3000017nH000007nL0P00
00Eo^7nGOiQoUgnh0080000:Oc800000OgQoE`00OcIoEWleOeH2000017k[O``0000057oo00<007oo
Ool01Woo000@Ool00`00Oomoo`0SOol200000c5i0000000300000e7i0000000200000f9H00000008
00000e[O00000003000017LM000007lk0`0000AnegkF001bE@L00003GOH0000000800003GM<00000
00@00004M]X000000002N]P200001GmKOg]oNgmkOe/00Wmk00D007mIOgUoF@00009oV003001oV000
00<00005OiMoV7nHOiQoU`050002OgP00gmH001oEP02OeH01`00Odl007k[001fH`0001=oo`03001o
ogoo00Ioo`0047oo00<007ooOol08goo00<000bF00001`0000E9U`00000007ng00<00003Odl00000
00H00003H]h0000000<00008OaX00000O_X007kg001fMAD00003N_/000000080000@Oe]oNgmJOg]o
FgmkOeYoNgmK001oFGmj001oV7nGOiP500000gnHOiMo^00200000gmg00000002000037mhOeQoN7mG
001o=WmF001oD7m?001nj`8001=oo`03001oogoo00Ioo`0047oo00<007ooOol097oo0P0000@e^P00
0000009=f080009of0<00005OkH00000001oN@0300001EIl001Zg@00H]h00P0000Mg7@00000007;K
001nnP0500000f6@GI00000;00000gKJM]T00002M_X20002Nc`50003Og/01WmKOg]oN`00OeT000Eo
V0<00003OiP0000000=oU`<00003OiD00000009oN005001oEWmFOeH00002Odl3000COol00`00Oomo
o`06Ool0011oo`03001oogoo02Aoo`/00003OmT007oI00T00004FWd006[nI]d700000g;J001jn@07
00000fYFKUH0000500001Fji001b^GKJL[T0100000Uk?000Ocek7000OcQoFGlhOcT00`0000IoNgmJ
Og/007mj0003OiP00gnGOiQoV003000017mgOiQoMgnH0P0000=oM@00OgD0100000QoN000OeIo=P00
OdmoD7m?0P004goo00<007ooOol01Woo000@Ool00`00Oomoo`0TOol800001gnE00000000OmP007oI
00@00003GQL0000000800004J_h006[nJ]d400000f[M001bf`0200001FjH001^N000KYT0200017:j
100017/l00D007/lNcak?00000Io>@H0009oV0@00004OeH000000003OiL300000gme000000060000
17mF001oCgm?0P0000=jP`00Ool04Woo00<007ooOol01Goo000@Ool00`00Oomoo`0TOol700000gne
OiAo]@02000017nh001of7oI0P0000=N5`0000000`0000EZg@00I]eZoVKM00<00004I]`00000L]/2
00001VihKYUZN000JWQ^V@800004L]X00000L_`800004W/LNcak77lm001k?7/LOcd007liOcQoFGlh
OcUo>7mIOcQo>@D00007OgMoEWmgOeIoM`00OgH00`0000EoV000OiD007nF00<00003OgQoE`0000H0
0004Odl00000000COol00`00Oomoo`05Ool0011oo`8002Ioo`H0009oU@06OiAoU@00OmP007oI0P00
0UhG0`0000INW@00J_h006[nJ]d<000016jIKWP006jI0P000W;J00AffP000000009bo004001k6g/L
Na/60003Nc`00`00OcUo>@0600000gmjOeX000030005OeH00`00OgH0000300002GnE000007nF001o
N000OgMoE`02OgL200000gm`Oe1oL003000COol00`00Oomoo`05Ool0011oo`03001oogoo02Eoo`H0
0003OkEoU00000`00003FW`006[M00L00008J[Y^^fZj001Z^P00J[Y^^`X0000<L[Yff`00001bf`00
Na]k77/KOcak6g/L200000aoFWmjOeYoNWmJOgYoFP00OiT007mhOiT200001gmFOgMoEP00OeEoMWmE
00H00008OgQoEgmh001oN7mGOgQoE`800005Og1oC`00O`T0000COol00`00Oomoo`05Ool0011oo`03
001oogoo02Eoo`P0009oU@07001oU@00001E]@00GQL0100000MNW@000000069k001Zg0020000269k
001^^Vjk001^^`00K[/3K[X4000017KlL_afo7;l3@000g]N00=oGW]N00000P000gmJ00AoNWmJOeX0
009oVP04001oV7nIOgP2000017mF001oMGme0`0000QoU@00OiD007mgOgQoE`000Wmg00MoEgmg0000
07m?Og1oD002000COol00`00Oomoo`05Ool0011oo`03001oogoo02Eoo`l00005GQL00000001=]`05
00003eiJ001Z_000HW]VV`00K[/006jkJYY^^`00K[]Z^P0200001g;l001g77;kMaabngLL00@00003
L_`007lK00@0000:Ogh007mnNeeoOW]MOgh007mlOg/200001GmjOeYoNWmJ00000WnJ00D007nIOgQo
VGmh00D00005OgEoMP00001oU@0200003GmhOeMoN7mG001oEgmhOeL007ma001oCgm`00<0019oo`03
001oogoo00Eoo`003goo00<007ooOol09Woo3@0000MMmP000000022[001=e`02000016Zl00000000
0V[L0P000VJK0P0000E^^VjkK[X006jj0080009bo003001bo00000<00007Maafo7LLL_ag77Kl0000
17lk10000g]N00=oGP00Og/00Wmk0P0000AoFP00OiYoVPH0009oV@03001o^7nh00L00006OeMoN7mg
OgQoE`000Wmg00P007nBOi8007m?001o2@004Woo00<007ooOol01Goo000?Ool00`00Oomoo`0VOol;
00000e6D001Mm@0400000e7G00000002000015YJ000006kM0P0000=RV`0000004@0000mbo7LML_ag
77;lMad007llOa]o?7lKOc`007llOa/01@0000AoNgmlOg]oN`D00005Ok]oVWnj001o^@02OiT01@00
OmUo^7oHOkP00P000WnE00ioUP00000007mhOeMoN000OgQoE`00Og5oTWma0P0000=o:P00Ool04Goo
00<007ooOol01Goo000?Ool00`00Oomoo`0WOol200000bBb000000020000158I0000056e0P0000A=
L`00001Ae`@00003FUYNFP0000@00005IYaRVfJL001^f`03K[/30003L_`300000g<mMcig?P0H0000
0gnM001o_@020003OiX01@00OiUoVGnI00000gnh00@007mG0000009oU@03001oEgmg00800008OgMo
N000OgL007nBOg5oTP@0019oo`03001oogoo00Eoo`003goo00<007ooOol09goo300000=1<`000000
0`00011J>P00001NFeXj001VV`00J[]Zg000IY`006:L001^f`<00003L_d007LM009bo003001k?GLL
00800009Mcic7GLnLaeg?W<MMch007mm0080000POgioW`00OgmoOWnOOgioOgmnOil007nMOiao_GnL
001o_7oMOkd007nJOk]oVP00OiUo^@00OmQo^000OeMoM`80009oU@05001oN7mGOgMoE`0200001Wmh
000007maOi9oL@800003O`T007oo015oo`03001oogoo00Eoo`003goo00<007ooOol09goo0`0000<P
[`00DML01@0000Lm=0000000057H001J>P02000035iJ00000000J[/006ZlJ[/006:L001RN`800006
L_ebo000L_abo@000W;l00D007LLMaag77LM0080009g?P07Lceg?WLnMch007mMOe`00`0027mn00H0
07nMOi`007nmOmd3Okd2000027nJ001oVGnI001o^000OeL2OgL02`00OiD007nE001oEgmgOeMoMgmG
OgL00`0000AoTP00001oL@800003O`T007oo011oo`03001oogoo00Eoo`003goo00<007ooOol09goo
2P0000<m=0000000100000IJFP00000005ikGU/200007V[LJ[]Zg000I[`006:L001c7G;l001^o7;m
K_`007;l001g77/mMaak?GLLNceg7000Laeg?W<MMcic7@@00003OgeoG0000080000BOgioWgmnOgmo
OP00OiaoW@00Omeo_GoMOkaogGnmOmd007nk0P0001=o^000OeMoMgmGOgL007nF001oUP00OgMoEgmh
OeMoMgmGOgQoE`0200000gnBOg5oTP0200000gkY001oo`0@Ool00`00Oomoo`05Ool000moo`8002Qo
o`03000@TP0000D00004FSX00000DOT400001eIL00000000GW]NFeik00800005J[]Z_6Zl001RV`02
00001F[m001bo000L_d00g;l0P000WLL00Ig7GLLMaag77LMMa`6000027mMOeaoGGmLOeeoG000Oe`2
0003Ogh01000Oid007oM0gnm00EogGnmOkeo_GoM00@0000@OeMoMgmgOgMoE`00OiD007nE001oEgmg
OeMoMgmGOgL40003Og401000Od`000004Woo00<007ooOol017oo0002Ool50002Ool00`00Oomoo`03
Ool00`00Oomoo`0WOol@000015IK000005jM0P0000MNFP00J[]Zg6ZkJ]aZ^`0200002F[nJ]d007;l
001^o7;mK_abo@0200001gLLNceg77/mMaak?GLL00D0000<Lad007mLOgeoG7mmOeaoOGmL001oG7mm
0`0000QoO`00001o_GoMOkaogGnm2P0000IoMgmGOgL007nFOiD2000017mhOeMoMgmG0P0000QoV7mg
001oT`00Og5oTWma0`004Woo00<007ooOol017oo0004Ool01`00Oomoogoo001oo`0000Aoo`03001o
ogoo02Moo`80000348h0000000@0009FG0800009DU`00000001JO000FWaNW5Yl00@00003J[aZ^fZl
008000=Zg@05001bo7;lL_abo@02L_`01000Maeg77LL100000=g77LMMa`00WLL0`0000MoG7mMOeao
GGmLOeeoG00=0005Ok`20003OkX40002OiD2000017mGOgL0000017mh00@007ndOk@0009oL@03001o
C@00019oo`03001oogoo00Aoo`0017oo00<007ooOol00goo00<007ooOol00Woo00<007ooOol09goo
1P0000U1fP00DS]FG000EW`005ILDS/00P0000EJW@00GYeJO5jM00800003I]eZoP0002D00005Ogd0
07mmOiioO@0300001WnLOkeo_7nlOiao_@9o_00:Oi`00000Ok`007oKOkYofgnjOm/300004WnF0000
0000OiQoMgnHOgMoV7mg001oTgndOi<007ma001oC080015oo`03001oogoo00Aoo`0017oo00@007oo
Oomoo`8000Aoo`03001oogoo02Uoo`<00006AO]1fT7k@O/0059K0P0000AFG59KEU`0009JO004001J
O5jLFW`20003J]d30003J_h01F[m000007<oLch01`0000EgGWMOMemgGgMN00=gG`04MeigGgMO0009
Ogh400000gnn001o_007Ok`01@00Og/007nl000017nj0P000Wok0`001gmh00<007ndOi<00Wnd0P00
0Wm<00<007ooOol03goo00<007ooOol017oo0004Ool01@00Oomoogoo00001Goo00<007ooOol0:Goo
0P0000E27000000004Gk0080000?EU`0058kEU`005Im001JO000GYeJO000GY`006[M00800007Kal0
06KN001ZoV[MJ_h0100000M_7W<oLcic?`00Neik?@0200004g]oMeikGgMNNgmgGW]OMeikOgMN001o
OGnNOgeoWWmmOiioOGnN00800008Ogd007nMOkh007nlOiao_@9o_00;Oi`007mkOi`007nl001ofgnj
Om]o^P020003Oo/03`00Om`00000OgMoV7mgOiQoM`00Oi=o]7nCOkAoT`04000AOol00`00Oomoo`04
Ool000=oo`8000=oo`<000=oo`03001oogoo02Uoo`800004@O]1o47k@O`2000035ILEW`005IL001F
O5IL001JO000FW`0009NW@800007Kai_7flN001VgP00J_d0100000Qg?W<n001c?W<oLcic?`000g]N
0P000WMO00MgGWMOMemgGgMNMel00007Ogh02000OgioO@00Oiio_WnN0005Ok`02000Og]oW000Okao
^gnl0003OkX20002Oo/01@00Omao^goKOk/00P000gmh00@007ndOkAo]0<0009oC@03001oogoo00mo
o`03001oogoo00Aoo`003Woo00<007ooOol0:Goo0`0000<mnd8L@O/00P0000QJO@00EU`005ImEUaF
OEIL0P0000=NW@00GYd00P0000A[7VlOJai_7`@0000AI]d006lNLch007LnLah006lNLcmc?P00Mcek
GW/mNeig?@03000047MNNgmgGW]OMeikO`00OiioOGnNOgeoWP00OiioOGmn0P0000AoWWnnOid0009o
_00HOi`007mkOiaoN`00Ok]o_7nk001o^WoK001on@00Oo]onP00Ok]og7nkOmao^goL0P0000=oM`00
Oi<00P0000Ioe7nd000007m]000AOol00`00Oomoo`04Ool000ioo`03001oogoo02Uoo`<00003@O`0
0000008mf`<00003EWaFG5Il009FO0D00003H]h006lO00<000=VgP03001ZoP00009c7P04001c?WLn
Lch2000017<oLch007/m0g]N00=k?G]NNeh00P000gMO00=gGP00Ogh00gmn00<007mnOgd00Wmn00H0
07nN001oWWnn0002Ok`01P00Og]oW7mkOi`000=o_006Ok/007nj001on@000Wok00T007oKOk]og7nk
Om]o^goLOk/00`000gnd00Eoe000001oC@00015oo`03001oogoo00Aoo`003Woo00<007ooOol0:Goo
100000@mfcgJ@O/mfP800007EWeFG5ImEUaFOEILEWd00P0000=RgVKo00000P0001e_7f/N001RgFKN
H]d006[M001_7W<oKah007<MMcic7WLn001c?`00Neig?G]NNcekGWLmNeik?G]N00800005MeikG`00
OiioO@0200002gnNOgeoOWmm001oWWnnOid007nNOkh00P0001UoNgnLOg]oW7mk001o^gnlOk]o_000
Om/007oi001on`00Omao^goLOk]og7nkOmao^`0400001WoDOkAoeGnd001oT@80015oo`03001oogoo
00Aoo`003Woo0P00:goo500000IRof;N001_7VlO0003I]h200003g<NLci_7W<N001g?W<nMcic?WLn
001c?`00Neik?@03Neh01G/mNeikGW]NNcd00P0000=gG`0000000Wmn00=oO@0000002`0000AoNgnL
Og/0009o_00@Ok]o_7nl001o^P00OoT007ok001o^goLOk]ofgnkOm`20002OiX01P00OoL007oDOk@0
009o/@80015oo`03001oogoo00Aoo`003Woo00<007ooOol0:Woo0P0000@n7D8n?Se2?P80009FWP03
DWeFWUJN00D0000HH]iVof;NI_l006lOJah006;MI]iRgFKN001c?flNLci_7W<o001g?W<MMcic7WLn
0P0000Mk?G]NMcekGW/mNeig?@0;00002GmnOil007noOkio_gnnOkmo_P0200000gnmOid000060000
1WnkOkao^`00OoQon@800005Ok]og7nkOmao^`0200003WnJOgYoVgnJ001oeWog001oe000Ok5oTGna
000AOol00`00Oomoo`04Ool000ioo`03001oogoo02Yoo`800006@Shn?Chn000n7CWl1@0000IFWP00
EYh005kO0003H]h00`00Kah00005I]hE0006Ngl00`00OiioOP03Oih01Wmn001o_gnoOkio_`<00003
Oih007nm009o_@05001ooWoMOmiog@040002OoT200000goLOk]of`020006OiX01P00OoMom`00Ok@0
00=o/@03001oogoo011oo`03001oogoo00=oo`003Woo00<007ooOol0:Woo0`0000=2?P00?Qd00`00
0eJN0P0000EFWUZoEYh005kN00800007I_l006lO001VgV;MI]h00`0000Ec?floLcm_?g<o00<00006
Memc?gMOLcmgGg<o0`0001QgGg]oNemkOgMONgmkGg]oMemkO`00OiioOWnOOgh007nnOkmo_Wno001o
_gnNOkh20002Okd01@00OoiogGonOmd0100000=ooP00OoT00P0000Ao^`00001oV`9oVP09OgYoVgnJ
OiX007ogOmIomgoF00800004Ok400000000BOol00`00Oomoo`03Ool000eoo`03001oogoo02aoo`D0
008io@0?001FWUJN001J_eJNE[mF_UZo001RgekOH]l006;N00800009I]h00000Jal006LO001c?flo
009c?`800008Mel007MOLemgGg=OMel0009kGP8000QkO`0:001oWWnNOgioWP00Okmo_Wno0005Oih0
1P00Okeo_@00OmeogP8000=ooP03001ooGon00@000UoVP03001omgog00=om`800003Om9o/@00019o
o`03001oogoo00=oo`003Goo00<007ooOol0:Woo1`0000<io3hM>O`00P0000]J_eJNF[mFWUZoEYh0
05kNH_mNgV;O00@0000>Iam[7fLO001Vo`00Kcmc?flo001_?g<o001gG`800008LcmgGg<oMel007mn
NeioOP80000=NemkOgMONgmkGg]oMel007mnOih007noOkh00P0000Mo_gnNOkioWWnoOiio_P0@0000
1WmjOi]oVWnJOgYoV`9oVP06OgYoV`00OoMoeWog100000=o/GoB00004Woo00<007ooOol00goo000=
Ool00`00Oomoo`0WOol30003Ool20002=Sh500003EJNE[mF_UZoEYiF_`00H]mNgf;OG]mRg`0000=[
7`05Iam[7`00Ial00003Lcl01@00Lcmc?`00Mch00WMO00D007=OMemcG`0000=kGP03NgikGP000080
00AkO`06001oWWnN001o_P001gnN1`0000=o_7oMOmd00P0017ok00<007oM00000`001WnJ00@007og
OoL0009om`800003Ok9o/@00019oo`03001oogoo00=oo`003Goo00<007ooOol08goo10001Woo0P00
00@f7SXo=Qhj?P80000;D[l00000001FWUZoEYiJ_`00H]mNgP0300007fLOJamW7f/OIal006KoIal0
07<oKcl006loLem_?`00LcigGg<n001c?gMOLcl007]NOgikGWmnNeioOW]N0080000;NgmkG`00Ogh0
07nn001oWWnoOiio_P0900001GoMOkaogGnlOmd00P0000IonWokOoX007oK0002Omd00gnmOmh00003
0002OiX3000027ogOmIom`00Oo@007oB000BOol00`00Oomoo`03Ool000eoo`03001oogoo021oo`<0
00Yoo`<00004=Qhf?P00=Qh20003E[l3000015JN000000000f;o00`006LOJam[7`00I_mW7fKo001_
?g<o0004Lcl01000McigGgMO0P0000=gG`00Ngh00g]N00EkOW]NNeikGW]n00800003Ngl0000000H0
009o_`@00004Oml007oM0003Omd00`00Ood00003Oo/00`00Om/00008Omd30005OoL01@00Om@007na
00004Woo00<007ooOol00goo000=Ool00`00Oomoo`0MOol3000>Ool2000013Xn=Qh003HM0P0000IF
ge:oE[mB_eKOD[l200009f;oH]mRoekOH_l006/OIam[7`00IamVof/OI_l006lo001_?g<oKcmcG`00
Memc?WMOLcigG`00Mel007mnNeioOW]NOgikGWmnNeioOP0500007GoOOkmo_gnoOmmo_`00Okl007no
001o_goO001og@00Omeo_000Om`007ojOo]onP00Om/007nmOmh00WoM00Io_GoNOmeogGnmOmh20000
1goFOoMoeP00Om=om7oC0080019oo`03001oogoo00=oo`003Goo00<007ooOol06Goo10004Goo1000
0SHN00<003Hn00001EJo0`0016;o00`006/OJamW7`00I_mW7fKoIal007<o0005Lcl00`00Memg?P03
Mel200000g]NNgikGP0600000gnONil00006Okl200000goO001o_`0200001GoOOkmog`00Omd00P00
0Wom00@007okOo/000=of`03001ogGoM00Iog@800008Oo`007ogOoL007odOm=om080019oo`03001o
ogoo00=oo`003Goo0P005goo0`005Woo0P0000df7CHN<Od003HN001B_eKOD[mF_`00E[mBW`020000
36;oH]mRo`00JamW7f/O001[7fKoIamVo`80000;Lcm_?g=OKcmc?flo001c?WMOLcigG`0600002W]o
OimkWgnONgmoWg^OOil007oO0gno00Mog`00Ommo_`00Okmog`02000057noOmmo_goO001og@00Ooeo
g000OoYon`00Ooaofgol001ogGnmOmh2Omd00gnmOmiog@02000027omOo`007oF001odgodOm<2000B
Ool00`00Oomoo`03Ool000eoo`03001oogoo01=oo`<001Uoo`<0008f7P03000f7P00009F_`800006
E[mB_eJo001Ro`000f;o00X006/O001W7fKoIamVofLOI_l000Ac?`L000AgO`D00007NimoWg^OOimk
WgnO00000Wno0P000WoO00<007noOml00P0000Eo_goOOkmoggno00800006Ood007om001on`0017oK
00<007oMOmd01GoM00<007ol00000gol0P0000Eom000001om@00019oo`03001oogoo00=oo`003Goo
00<007ooOol03goo10007Goo0`0000@f7@00=Qhj?P80000DDYmF_e:OE[mBW`00G]mRo`00H_mRg`00
Ial006KoIamVof/OI_mW7`D00004LgmcG`00Kel200002WMoLemgOgMOMgl007nONgioWg]n100000=o
Wg^O00000P000WoO00Eo_goO001oggno00800009Ommo_goOOkmoggno001og7om00800008OoX007oK
OoaofgolOm/0009og@0;OkeogWoMOmd007om001ooGolOoeoo0020003OoD00`00Oomoo`0@Ool00`00
Oomoo`03Ool000aoo`03001oogoo00eoo`<0025oo`<00005=Qh003Hn=Qh00002E[l04E:oE[mF_eJo
D[l005koH_l006;o001[7`00IamVofLOI_l00`001W=O00@007=O001gO`8000AgO`08001oWgmoOimo
OgnOOgmoW`<000Eog`0>001oggno001oW`00OkmoggnoOmmo_`00Ool0009oo@03001on`0000Eof`03
001ogGoM00Aog@05001oo000Ooaoo@02Oo`01@00OoEomGoe000057oo00<007ooOol00Woo000<Ool0
0`00Oomoo`09Ool4000TOol7000033Xn000005JoDYmF_e:OE[l006;oG]mRo`800004Ial006KoJal2
00001VLo001W7`00KemcO`9cG`07001cOg=OLel007MoMel00P0000YgOgMO001kOWnONgioWg]nOimk
OP<0009og`06Okmog`00Ommo_goO0P0000]o_goOOkmoggno001oggoo001ooGoL00800009Ooaofgol
Om]oo7oK001o_GoN009og@08Okd007olOod007omOoaoo@<0009om@03001oogoo019oo`03001oogoo
009oo`0037oo00<007ooOol01Woo0`00:Goo0P0000LI_@00>Slj?cYO001>_`02000015:oE[mF_`00
0eko00ARo`00001Vo`8000=[?`03001[?`0000AcG`06001cGg=OLel007MO0gMo0P0000IgO`00Ogmo
WgmoOil20002Okl04000OmmoggoO001oggnoOml007no001oggnoOmmo_`0017oo00<007om00000P00
1WoK00<007oMOmd00WoM00@007olOo`000=oo0<0009om@03001oogoo019oo`03001oogoo009oo`00
37oo00<007ooOol00goo0`00;7oo0P0000TIW000000jGcXn>Ul005:oCYl00P0000IF_`00H_mNgf;o
G]l3000026LoJcmW?f/o001[?fLO0002Lel03FmO001cGg=OKemcOg=O001gGgMoMemgOgMO00800004
OimkOP000002Okl05goOOkmo_`00Oml007oOOkmoggno001oW`00Okmoggno001oggooOmmoogoOOol0
0`0001Uoo000OoaofgolOm]oo000Omeo_GoN001ooGolOod007omOo`007oi001oo000OoD0000DOol0
0`00Oomoo`02Ool000aoo`03001oo`0000<0031oo`<00005:Ql003Xo>Sl00003C[l01E:o00000000
H_l00P0000=S7`00Jcl016/o00D006LoJcmW?`00009cG`03001cGg=O00=cG`03001gOgMO00=gO`03
Mel0000000@000Ao_`80000;Ommo_goOOkmog`00Okl007oOOkl00007Ool20003Om/00`00Om]of`02
Om/03`00OmeogGoM001oo7omOo`007olOod007oj001oo002000EOol00`00Oomoo`02Ool000aoo`80
03Aoo`8000037Kd002XO00800007>Sh004jOD[m>We:oCYl00P0001=S7ekoHamNo`00Icm[?fLo001W
7f/oIam[?`00Lgl007=OKemcO`02Lel016mOLgl007Mo100000IgOg^O001o_gnOOkl500004goOOkmo
ggnoOml007no001oggnoOml007ooOmmoogoOOomoggoo00<00009Ooaofgol001oo7oKOoaof`00009o
g@0=001ooGolOod007omOo`007oi001oo7om00005Goo00<007ooOol00Woo000=Ool00`00Oomoo`0c
Ool300000bXN:QlZ7P0200000e:oC[m>_`040004Hal03000Jcm[?f/o001[?fLoJcmW?`00Lel000Ec
G`@000IgW`03001o_gno00Ao_`80000;OkmoggnoOmmo_`00Oil007noOml00005Ool900000goK001o
f`02Om/01000Omd007om0gol00<007ol00000Woj00@007olOo`001Eoo`03001oogoo009oo`003Woo
0P00<goo0`0000@Z7bXN:QlYoP@00003D]m>_e;O00<0000?HamNof<O001[?fLo001W7f/oIam[?fLO
001_Gg=o00@00003LgmcW`0000@0000JNimgWg^OMgl007nOOkmoWgno001oggno001o_goOOkl007nO
Okl007oO001oogoOOomog`800005Oomog`00Ool00002Ool500004WolOm/007nm001oo7omOoaoo@00
Ood007ojOoT007olOod001Eoo`03001oogoo009oo`0047oo0P00<Woo0P0000TZ7RXO:QhZ7bXN001B
_e;OC[l00U:o00@005lOGal0009S7`08001[?`00Icm[?fLoJcmW?`@000=cW`05LgmcW`00Lil00003
Mil20003Mil01000Okmo_`000gno00D007oOOkmog`00009o_`80009oo`<00004OmmoogoO0002Ool0
0`00Oomoo`02Ool00`00Ooh0000200000goK001oo002Oo`01GomOo`007ol00000Woj00<007ol0000
5Woo00<007ooOol00Woo000@Ool2000bOol300001BWn:Ql00000:Sl00P0000mBgdjoD]m>_`00G_mS
7`00Hal006/o001[?fLOJcl00`00011[O`00LimcOg>OLgmcWg=o001_Og>O001kWgNONimgO`@0000<
Okl007oOOkmoggnoOml007oO001o_gnO1@0000IoggooOmmoogoO0003Ool00`00Oomoo`03Ool20000
2Gon00000000Ooeoo7omOoaoo@0200001WoiOoYon@00Oo`001Ioo`03001oogoo009oo`004Woo00<0
07ooOol0<7oo100000@^?bXo;SlZ?`<00004C[mB_`00Hal2Gal3000016/o000000000V]O00<006]O
00000W>O00AcOg>OLil000=cO`03Lil007NO00EgW`<000Eo_`03001o_`00009o_`800008OmmoogoO
OomoggooOml000=oo`03001oogoo00Ioo`H000=oo08000=onP8001Moo`03001oogoo009oo`004goo
0P00<Goo100000H^?bXo;Sl004[OA[l2000015;O001S7eko100001Q[OfMOJgmWG`00Jem[OfMO001c
Og>OLgmcW`00Lim_Og>OKgmcW`00NimgWg^OMgl200000g^o00000003000017oOOkl007nO0`0000Qo
ggooOmmoogoOOomog`0017oo00<007ooOol027oo0`0000Uoo`00001oo@00OoYonGojOoT00P005goo
00<007ooOol00Woo000EOol00`00Oomoo`0_Ool200000bYO0000000:0002Gcl00`00JemWG`03Jel0
1P00Jgm[Of]oJel0009cW`04Lgl007=oLil3Lgl00`00MimgW`0200001GNoNkmk_`00Okl00goO1@00
0WoO00P007oOOomoggooOmmoo`0017oo00<007ooOol01goo10000goo10000Woj00<007ooOol05goo
00<007ooOol00Goo000FOol00`00Ool0000^Ool09P00;Ul000009SlZGbHo:Ul004[OA[m:gdJo001O
?f<oGcmS?`00JgmWGf]o001[OfMOJgm[Gf]o001cWg=o001_Og>OKgmcWfmoLilA00002WooOmmoo`00
OomoggooOmmoo`001Goo00<007ooOol01Woo0P000Woo0P000Woo0P0000=ooWom00000P006Goo00<0
07ooOol00Goo000GOol00`00Oomoo`0^Ool00`00;UlZG`0200029Sl300000dZoB]l0000200000f<o
Gcl00002Jel01FMO001[Of]oJel00f]o00@007>O001cW`9cO`<00005LkmcW`00Mkmk_`0300000g_O
Mml00003Oml200002WoO001oggoO001oggooOmmoo`001Woo00<007ooOol01Goo00<007oo00000Woo
0P0000Aoo`00Oo`000=ooP03001oogoo01Qoo`03001oogoo005oo`0067oo0P00;Woo0P0000<^GbXo
;Ul00P0001DV?`00=Yl004JoB]l006<o001S?elo001WG`00Iem[Of]OJgmWGf]oJel00P0000EcW`00
000007>o00T00005Nmmg_g_O001og`020003Oml027oo001oo`00Oomoggoo0007Ool00`00Oomoo`04
Ool00`00Ool00004Ool300001Gom001ooWom00006goo00<007ooOol00Goo000JOol00`00Oomoo`0/
Ool01P00:Ul^GbYO;UlZG`80008jW`03001:_`00009O?`07001S?`00Jel006]oJel00f]o00=[G`00
00001P0017>o00D007>o001g_g_O00L000Eog`05001og`00Ool00008Ool00`00Oomoo`03Ool01000
Oomoo`001Goo0`000Won0P006goo00<007ooOol00Goo000KOol00`00Oomoo`0/Ool=00000eloHcl0
000200000f]oJel0000600004F^O001__g?OKkmc_fnoLmm__`00Kkmc_fno001g_g_OMkl00P0000=k
o`00Ool01@0000Aoo`00000000Uoo`03001oogoo009oo`04001oogoo0005Ool400000gomOoh0000L
Ool00`00Oomoo`01Ool001aoo`03001oo`0002aoo`<00004:Ul^GbXo;Ul<00002V]oJil00000Jil0
06^OJkm[W`001G>o00L007>oLkm__g>o001kg`0200001g_ONomkgg_o001kog_O00=ko`<000Yoo`05
001oogooOol00003Ool00`00Oomoo`02Ool200000goo00000002000MOol00`00Oomoo`01Ool001eo
o`8002eoo`@00003;UlZ?biO00800003=Ylj_`0000<00003Gel006Mo0080000BJkmWWf^o001[_f^O
Kkm[W`00KkmcgfnoLkl007>oKkmc_fno100000YkogOONomggg_o001oog_OOomkg`800003Ool00000
00Moo`05001oogooOol00004Ool02@00Oomoogoo001oo`00Oomko`03000MOol00`00Oomoo`01Ool0
01ioo`8002aoo`03000AoP0000<00006:Sl004:o000fW`000emO0`0016^O00L006^oJim[_f^OJkl0
0003Lkl00`00Lkm__`0400000g?O001ko`0500000g_o001ko`020003Nol01000Ool000001Goo00@0
07ooOol000Aoo`07001oogooOol007oo00000goo00@007ooOol001aoo`03001oogoo005oo`0087oo
00<007ooOol0:Woo00D0017n4Mh003YO00<00005@[l00000001OG`0200003emo001[_fNOJkl006^o
Jim__f^OJkm[Wfno001c_`04000036noLmm_g`00Kmmco`00Nomggg_oMmmko`<00007OomkogooNomo
og_o00000Woo0P000goo00<007oo00001Goo00X007oo000007ooNol007_oOol000Aoo`03001oogoo
01Uoo`03001oogoo005oo`008Goo00<007ooOol0:Woo1@0000<nG`00000010000Umo00aOGemo001[
W`00Jkm[Wf^oJim[_f^OJkl400001VoOLmmcgg?OKml000Acg`05001kogOONomgo`0200000gOo0000
0004Nol00`00Oomoo`02Ool200000goo00000005Ool00`00Ool00004Ool01000Ool000001Goo00<0
07ooOol067oo00<007ooOol00Goo000ROol2000[Ool600000cYO0000000700000f^O001[W`060000
4F_O001[g`00Lmm__g?OKml006oOLom_gg?OKmmco`00Nol00P000WOo00=co`00Lol00P0000Aoog_o
Ool000Eoo`<000Eoo`800006Nomoog_oOol007oo0P001goo00<007ooOol05goo00<007ooOol00Goo
000TOol2000ZOol200025Oh>0006Jkl02@00JkmW_f^o001cgfoOLml00006Lml20006Mol01000Lomg
o`000W_o00<007ooOol00goo0P0000Eko`00000007oo008000Eoo`<000Uoo`03001oogoo01Moo`40
005oo`5oo`009Goo00<007ooOol0:Goo0`0000<IoaGn6Ol00P0001];7d/o000005MoFimGO`00Ikm[
gfNoJmmW_f_O001[gfNoJkmW_`00Kmmcg`00Lom_gg?OKml00`0000E[o`00000007?o00=go`04001g
og?oMol2000017_o001oogoo0P0000Iko`00MomkogOoNol400000g_oOomko`03000;Ool00`00Oomo
o`0FOol10001Ool1Ool002Ioo`8002Yoo`080005>000000EoP00Eel0009;?`<00003FimKO`0000A[
_`03001[_fNo00=[_`03Ikl007?O00D00007Jom_of_oKom[o`00Kol00P000gOo00D007?oLomgog?o
00800003Ool007oo009oo`04001kogOoNol30002Jol4000>Ool00`00Oomoo`0EOol10001Ool1Ool0
021oo`03001oogoo00Eoo`03001oogoo02Qoo`070005F@Dh000EoP00Eel00`0000=KOenO00000P00
00Y[gfNoJml006_OIkm[_fNoJmmW_`@0000:Kom[ofooJom_of_oKom[ofoo0003Kol200001gOo001c
ogOoLomgog?o00<0000=Nomoog_o001gog_o001[ofOoJol006_oIol00P0047oo00<007ooOol057oo
0@000Goo0Goo000OOol00`00Ool00007Ool00`00Oomoo`0XOol00`001CP5>00200000e]OEemKG`02
00000e]oGimKO`0200000f^o001[_`06000015_oGol005_o0P0000Q[ofooJom_of_oKom[o`001Foo
0`0000Mco`00000006_o001_o`02000017oo000000000f_o00AWo`00000001=oo`03001oogoo01=o
o`40005oo`5oo`006Woo0`0017oo00<007ooOol01Woo0P00:Goo4@0000aOoe_oGomKoeooFomOo`00
GomKoeooFol200001VooJom_of_oKol000I_o`<00008Jom_of_oKol007?oKomco`T001Ioo`03001o
ogoo019oo`40005oo`5oo`007goo0P002Woo0P00:Woo0P0000H5^@00000c?c=O<cl200000d^oAim;
_`0200000e_oGomKo`02Fol01EooFol005_oGol00e_o00QOo`00000006ooJom_o`000foo0P0000AW
o`00000000=[o`03001_og?o00<000AOo`8001Qoo`03001oogoo015oo`40005oo`5oo`007goo00<0
07ooOol02goo00<007ooOol0:Goo1@0001TcGc<o000004^oAil004^o001KoeooFomOoe_oGol005oo
FomOoe_oGomKoeooFomOo`0200000f_o001_o`0200001VOoHol006?oIomSo`800009Kol00000001S
oeooHol005oo008001]oo`03001oogoo011oo`40005oo`5oo`007goo0`0037oo00<007ooOol0:goo
1@0000HN_`000017W`00Bkl200001EooFomKoe_o00000e_o00EOoe_oFomKoeoo009Ko`@000AWo`03
001WofOo00=Wo`T001ioo`03001oogoo00moo`40005oo`5oo`00;goo00<007ooOol0;Goo0`00018N
gaZo0017W`00Bkm?g`00GomKo`00FomOoe_oGomKoeooFol600002T_o000006?oIomSo`00HomWof?o
0P0000A?o`00Com;o`80029oo`03001oogoo00ioo`40005oo`5oo`00<7oo0`00;Woo1P0000=?_d^o
Ckl00`000U_o1@000d_O00<004_OBml00d_O0`0000QWo`00000004_oCol004oo0P009Goo00<007oo
Ool03Goo0@000Goo0Goo000bOol00`00Oomoo`0_Ool;00003TOOBom7gd_oAmm;odOO001;gd_oAmm;
od_OBol;000XOol00`00Oomoo`0<Ool10001Ool1Ool003=oo`03001oogoo02moo`03001oo`0000X0
009;g`800003Bml0000000/002eoo`03001oogoo00aoo`40005oo`00=7oo00<007ooOol0;Woo00<0
07ooOol00goo4`00=7oo00<007ooOol02goo0@000Goo000eOol2000]Ool00`00Oomoo`1<Ool00`00
Oomoo`0:Ool10001Ool003Moo`8002]oo`03001oogoo04eoo`03001oogoo00Uoo`40005oo`00>7oo
00<007ooOol0:Goo00<007ooOol0CGoo00<007ooOol02Goo0@000Goo000iOol00`00Oomoo`0XOol0
0`00Oomoo`1>Ool00`00Oomoo`08Ool10001Ool003Yoo`8002Qoo`03001oogoo04moo`03001oogoo
00Moo`40005oo`00?7oo00<007oo00009Goo00<007ooOol0D7oo00<007ooOol01Woo0@000Goo000m
Ool2000UOol00`00Oomoo`1AOol00`00Oomoo`05Ool10001Ool003ioo`03001oogoo02=oo`03001o
ogoo059oo`03001oogoo00Aoo`40005oo`00?goo0P008goo00<007ooOol0Dgoo00<007ooOol00goo
0@000Goo0011Ool00`00Oomoo`0POol00`00Oomoo`1DOol00`00Oomoo`02Ool10001Ool0049oo`03
001oo`0001moo`03001oogoo05Eoo`03001oogoo009oo`400000@goo0P007goo00<007ooOol0EWoo
00@007ooOomoo`400000A7oo0P007Woo00<007ooOol0Egoo00<007ooOol00@000016Ool00`00Oomo
o`0KOol00`00Oomoo`1HOol10001Ool1000003eoo`03001oogoo00Moo`03001oogoo01Yoo`03001o
ogoo05Uoo`800000?7oo00<007oo00002Goo0P006Woo00<007ooOol0DWoo00<007ooOol00Woo1000
000lOol00`00Ool0000:Ool2000IOol00`00Oomoo`1BOol00`00Ool000030003Ool003aoo`03001o
o`0000aoo`03001oogoo01Ioo`03001oogoo04eoo`03001oogoo00@000Moo`00?7oo00<007oo0000
3Goo00<007ooOol05Goo00<007ooOol0B7oo00@007ooOomoo`@000]oo`00?Goo00<007ooOol03Goo
00<007oo000057oo00<007ooOol0@goo00@007ooOomoo`D000moo`00CWoo0P0057oo00<007ooOol0
?Woo00<007ooOol00Woo100057oo001@Ool00`00Oomoo`0AOol00`00Oomoo`0nOol5000HOol0055o
o`03001oogoo011oo`03001oogoo03Uoo`03001oo`0000<001aoo`00DWoo0P0000=oo`00Ool03Goo
00<007ooOol0=7oo00<007ooOol0100087oo001DOol2000>Ool00`00Oomoo`0ZOol00`00Oomoo`02
Ool00`00Oomoo`05000KOol30006Ool005Eoo`03001oogoo00aoo`03001oogoo02Yoo`04001oogoo
Ool4000QOol00`00Oomoo`05Ool005Ioo`03001oogoo00]oo`03001oogoo02Ioo`04001oogooOol4
000SOol40006Ool005Moo`800003Ool007oo00Qoo`03001oogoo025oo`03001oogoo009oo`@002Mo
o`03001oo`0000Moo`00FGoo0P002Goo00<007ooOol08Goo1@00;7oo0P001goo001JOol00`00Oomo
o`07Ool00`00Oomoo`0GOol00`00Oomoo`02Ool5000LOol3000COol00`00Oomoo`05Ool005]oo`03
001oogoo00Ioo`03001oogoo01Moo`03001oo`0000<0021oo`03001oogoo01]oo`00G7oo0P0000=o
o`00Ool00goo00<007ooOol04goo00<007oo00000`009Goo00<007ooOol06Woo001NOol20004Ool0
0`00Oomoo`0>Ool00`00Oomoo`04000YOol00`00Oomoo`0JOol005moo`03001oogoo009oo`03001o
ogoo00Eoo`05001oogooOol00002Ool5000/Ool00`00Ool0000KOol0061oo`05001oogooOol00007
Ool00`00Oomoo`04000NOol00`00Oomoo`0AOol00`00Oomoo`0JOol0065oo`800005Ool007ooOol0
0003Ool4000QOol00`00Ool0000_Ool005Uoo`03001oogoo00Moo`80009oo`@002Eoo`03001oo`00
02moo`00F7oo00<007oo00002Goo0`00:Goo00<007oo0000;goo001JOol00`00Oomoo`0ROol3000>
Ool00`00Ool0000_Ool005Qoo`8002Eoo`03001oogoo00moo`03001oogoo02ioo`00F7oo00<007oo
Ool07goo0`000goo00<007ooOol0?goo001HOol3000UOol00`00Oomoo`0oOol006ioo`<000ioo`03
001oo`00041oo`00Kgoo00<007ooOol03Woo00<007ooOol0?goo001XOol30002Ool4001AOol006eo
o`03001oo`00059oo`00KWoo0P00DWoo001_Ool00`00Oomoo`1@Ool00<9oo`00`Woo0032Ool00<9o
o`00\
\>"],
  ImageRangeCache->{{{0, 193.625}, {287, 0}} -> {0.106263, -0.0146349, 
  0.0036319, 0.0036319}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(astroida = 
      ParametricPlot3D[
        Evaluate[{\((a*Cos[u]*Cos[v])\)^3, \n\t\t\((b*Sin[u]*Cos[v])\)^3, \n
              \t\t\t\((c*Sin[v])\)^3}\  /. \ {a -> 1, b -> 1, c -> 1}], {u, 
          \(-Pi\), Pi}, {v, \(-Pi\), Pi}, PlotRange -> All, 
        PlotPoints -> 20]; \)\)], "Input",
  CellLabel->"In[43]:="],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.09054 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0588087 1.17088 1.11022e-16 1.17088 [
[ 0 0 0 0 ]
[ 1 1.09054 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.08856 .24464 m
0 .80661 L
s
0 .80661 m
.40259 1.09054 L
s
.40259 1.09054 m
.41994 .58321 L
s
.41994 .58321 m
.08856 .24464 L
s
.6704 0 m
.92721 .4172 L
s
.92721 .4172 m
1 .95369 L
s
1 .95369 m
.70381 .58915 L
s
.70381 .58915 m
.6704 0 L
s
.08856 .24464 m
0 .80661 L
s
0 .80661 m
.70381 .58915 L
s
.70381 .58915 m
.6704 0 L
s
.6704 0 m
.08856 .24464 L
s
.41994 .58321 m
.92721 .4172 L
s
.92721 .4172 m
1 .95369 L
s
1 .95369 m
.40259 1.09054 L
s
.40259 1.09054 m
.41994 .58321 L
s
0 0 m
1 0 L
1 1.09054 L
0 1.09054 L
closepath
clip
newpath
.63197 .7248 m .64266 .72488 L .62416 .72 L .802 .727 .782 r
F
.66651 .74601 m .65172 .73291 L .63197 .7248 L F
.65172 .73291 m .64266 .72488 L .63197 .7248 L F
.5 Mabswid
.62249 .7209 m .66651 .74601 L p .62416 .72 L F P
0 g
s
.6019 .66885 m .6382 .71239 L .59084 .70061 L .791 .695 .758 r
F
.49406 .68198 m .48431 .6792 L .46658 .67775 L .713 .677 .814 r
F
.34337 .66764 m .48371 .681 L p .46658 .67775 L F P
0 g
s
.48371 .681 m .49406 .68198 L p .46658 .67775 L .713 .677 .814 r
F P
0 g
s
.24898 .6753 m .34337 .66764 L p .34051 .66687 L .686 .622 .779 r
F P
0 g
s
.33206 .6661 m .36073 .66629 L p .34065 .66688 L .696 .455 .554 r
F P
0 g
s
.32632 .63391 m .36305 .59475 L .3156 .6308 L .682 .543 .69 r
F
.50926 .37466 m .51022 .37251 L .50988 .37324 L .907 .757 .684 r
F
.52091 .33688 m .52094 .33688 L p .50989 .37291 L .624 .404 .563 r
F P
0 g
s
.52114 .33692 m .52115 .33693 L p .52126 .33777 L .911 .651 .532 r
F P
0 g
s
.52114 .33698 m .52306 .35552 L .52114 .33703 L .844 .604 .575 r
F
.52306 .35552 m .52356 .36043 L .52305 .35548 L F
.52356 .36043 m .52356 .36041 L .52305 .35548 L F
.52771 .38061 m .52661 .37391 L .52584 .37199 L .88 .643 .575 r
F
.53824 .43191 m .53831 .4322 L .53806 .43373 L .904 .666 .569 r
F
.54603 .46365 m .54811 .47208 L .54535 .46157 L F
.54811 .47208 m .55117 .48552 L .54078 .44455 L F
.55105 .48531 m .54078 .44455 L p .55117 .48552 L F P
0 g
s
.54229 .45213 m .54799 .4737 L .54465 .46404 L .835 .63 .625 r
F
.54415 .45792 m .55105 .48531 L p .54229 .45213 L F P
0 g
s
.56433 .52278 m .55645 .49537 L .55105 .48531 L .869 .664 .623 r
F
.57718 .55905 m .55799 .50009 L .56433 .52278 L F
.55799 .50009 m .55645 .49537 L .56433 .52278 L F
.54078 .44455 m .54415 .45792 L p .54229 .45213 L .835 .63 .625 r
F P
0 g
s
.54811 .47208 m .54078 .44455 L .54043 .4428 L .904 .666 .569 r
F
.53831 .4322 m .54603 .46365 L .54043 .4428 L F
.54603 .46365 m .54535 .46157 L .54043 .4428 L F
.54465 .46404 m .55105 .48531 L .54949 .48848 L .835 .63 .625 r
F
.54799 .4737 m .55105 .48531 L .54465 .46404 L F
.54078 .44455 m .54949 .48848 L .54472 .49814 L F
.52101 .33688 m .52109 .3369 L p .52094 .33821 L .771 .546 .593 r
F P
0 g
s
.52094 .33821 m .51963 .34954 L .52094 .33752 L .771 .546 .593 r
F
.52094 .33821 m .52094 .33752 L .52101 .33688 L F
.51963 .34954 m .52019 .34411 L .52101 .33688 L F
.51738 .36872 m .52028 .34329 L p .51741 .36879 L F P
0 g
s
.52028 .34329 m .52101 .33688 L p .52019 .34411 L .771 .546 .593 r
F P
0 g
s
.51741 .36879 m .52028 .34329 L .52019 .34411 L .771 .546 .593 r
F
.52115 .33693 m .52115 .33694 L p .52115 .33695 L .987 .767 .373 r
F P
0 g
s
.52306 .35552 m .52584 .37199 L .52514 .37576 L .844 .604 .575 r
F
.52504 .36597 m .52584 .37199 L .52306 .35552 L F
.52115 .33697 m .52584 .37199 L p .52114 .33698 L F P
0 g
s
.5003 .41468 m .51319 .37105 L .51018 .37424 L .667 .459 .595 r
F
.51768 .35339 m .5176 .35763 L .51635 .36 L .701 .481 .587 r
F
.50649 .3859 m .51018 .37424 L .50886 .37564 L .667 .459 .595 r
F
.50239 .41245 m .48098 .4501 L .49146 .42407 L .706 .517 .631 r
F
.50295 .41392 m .47907 .45485 L .50239 .41245 L F
.47907 .45485 m .48098 .4501 L .50239 .41245 L F
.50256 .4154 m .46947 .47163 L .50295 .41392 L F
.38695 .56735 m .38821 .57096 L .37192 .58367 L .642 .464 .627 r
F
.39594 .55962 m .38821 .57096 L .38695 .56735 L F
.35334 .59624 m .37192 .58367 L .34021 .60842 L F
.38623 .56574 m .38057 .57284 L .35334 .59624 L F
.38057 .57284 m .37192 .58367 L .35334 .59624 L F
.40129 .55178 m .37192 .58367 L .38623 .56574 L F
.38582 .57197 m .38623 .57003 L .37974 .57695 L .682 .543 .69 r
F
.41043 .55363 m .38991 .57868 L .34484 .60564 L .642 .464 .627 r
F
.41681 .5416 m .41205 .54577 L .41037 .5539 L .674 .495 .635 r
F
.41681 .5416 m .41037 .5539 L .41315 .55196 L F
.41092 .55325 m .34484 .60564 L .34021 .60842 L .642 .464 .627 r
F
.37801 .60877 m .38359 .58246 L .32632 .63391 L .682 .543 .69 r
F
.38359 .58246 m .38582 .57197 L .36514 .5935 L F
.38359 .58246 m .36514 .5935 L .33158 .62844 L F
.41681 .5416 m .41383 .53716 L .41205 .54577 L .674 .495 .635 r
F
.32828 .61904 m .33507 .61339 L .33031 .61717 L .798 .77 .828 r
F
.32617 .62118 m .33507 .61339 L .32255 .62381 L F
.38549 .56928 m .38623 .57003 L p .38969 .5654 L .854 .768 .766 r
F P
0 g
s
.40129 .55178 m .39594 .55962 L .38695 .56735 L .642 .464 .627 r
F
.3158 .6306 m .34484 .60564 L .31845 .62794 L .798 .77 .828 r
F
.24827 .67339 m .32255 .62381 L p .29996 .64655 L .656 .521 .688 r
F P
0 g
s
.29332 .64997 m .34484 .60564 L p .29321 .65001 L .798 .77 .828 r
F P
0 g
s
.32774 .63264 m .33158 .62844 L .32632 .63391 L .682 .543 .69 r
F
.38359 .58246 m .33158 .62844 L .32774 .63264 L F
.38582 .57197 m .37974 .57695 L .32632 .63391 L F
.34484 .60564 m .3156 .6308 L p .36305 .59475 L F P
0 g
s
.46658 .67775 m .39679 .6645 L .35953 .66669 L .713 .677 .814 r
F
.35174 .66832 m .35953 .66669 L .34337 .66764 L F
.46658 .67775 m .35953 .66669 L .35174 .66832 L F
.51813 .64381 m .47669 .67729 L .44554 .66085 L .743 .664 .773 r
F
.6019 .66885 m .59084 .70061 L .57267 .68716 L .791 .695 .758 r
F
.45688 .64379 m .44028 .58932 L .39116 .61958 L .731 .615 .729 r
F
.4614 .53651 m .4635 .57773 L .39638 .60221 L .709 .559 .683 r
F
.34284 .6387 m .37801 .60877 L .32632 .63391 L .682 .543 .69 r
F
.3743 .62625 m .37801 .60877 L .34284 .6387 L F
.40402 .63083 m .33206 .6661 L .31976 .6577 L .707 .617 .755 r
F
.3743 .62625 m .39856 .61014 L .38623 .57003 L .682 .543 .69 r
F
.4635 .57773 m .39856 .61014 L .39638 .60221 L .709 .559 .683 r
F
.48201 .636 m .43318 .566 L .45688 .64379 L .731 .615 .729 r
F
.33206 .6661 m .30069 .65617 L .28454 .6618 L .686 .622 .779 r
F
.32843 .66362 m .30069 .65617 L .33206 .6661 L F
.29996 .64655 m .29317 .65339 L .24827 .67339 L .656 .521 .688 r
F
.40115 .62832 m .31897 .65716 L .31531 .6496 L .707 .617 .755 r
F
.31183 .65228 m .31531 .6496 L p .31897 .65716 L F P
0 g
s
.33405 .66746 m .29333 .66296 L .24898 .6753 L .686 .622 .779 r
F
.29333 .66296 m .28414 .66194 L .26075 .67082 L F
.29333 .66296 m .26075 .67082 L .24898 .6753 L F
.25453 .67228 m .24898 .6753 L p .28414 .66194 L F P
0 g
s
.24817 .67459 m .24823 .67385 L p .25211 .67317 L .62 .389 .546 r
F P
0 g
s
.24823 .67385 m .24827 .67339 L .25404 .67245 L p .62 .389 .546 r
F P
0 g
s
.25404 .67245 m .25423 .67241 L p .24823 .67385 L .62 .389 .546 r
F P
0 g
s
.25237 .67328 m .25322 .67278 L .24866 .67501 L .731 .765 .881 r
F
.25322 .67278 m .25134 .67345 L .24866 .67501 L F
.25211 .67317 m .24823 .67385 L .25423 .67241 L .62 .389 .546 r
F
.34284 .6387 m .32632 .63391 L .31655 .64911 L .682 .543 .69 r
F
.57053 .67086 m .52786 .60496 L .50125 .63004 L .778 .658 .73 r
F
.51784 .53793 m .54486 .59422 L .48964 .60702 L .76 .617 .7 r
F
.5315 .60266 m .52786 .60496 L .57053 .67086 L .778 .658 .73 r
F
.61277 .71684 m .55915 .68979 L .54471 .68798 L .742 .705 .816 r
F
.52574 .65002 m .47669 .67729 L .51813 .64381 L .743 .664 .773 r
F
.60839 .71359 m .53996 .67971 L .61277 .71684 L .742 .705 .816 r
F
.53996 .67971 m .53935 .6798 L .61277 .71684 L F
.63496 .70337 m .59232 .63137 L .59264 .65852 L .85 .691 .685 r
F
.58229 .6643 m .5315 .60266 L .57053 .67086 L .778 .658 .73 r
F
.56435 .55482 m .59526 .62094 L .5735 .62837 L .82 .658 .682 r
F
.57789 .57705 m .59526 .62094 L .56435 .55482 L F
.54078 .44455 m .54472 .49814 L .53732 .48539 L .835 .63 .625 r
F
.63653 .70014 m .5927 .63 L .63496 .70337 L .85 .691 .685 r
F
.59022 .59894 m .595 .60896 L .60033 .62602 L .893 .686 .616 r
F
.5927 .63 m .59232 .63137 L .63496 .70337 L .85 .691 .685 r
F
.63566 .70193 m .65297 .72681 L .64134 .71546 L .867 .715 .691 r
F
.62478 .71966 m .60589 .71175 L p .6382 .71239 L .791 .695 .758 r
F P
0 g
s
.52574 .65002 m .55894 .6771 L .47669 .67729 L .743 .664 .773 r
F
.55896 .67712 m .5592 .67719 L p .50125 .63004 L .778 .658 .73 r
F P
0 g
s
.51064 .68372 m .47669 .67729 L p .55908 .6771 L .743 .664 .773 r
F P
0 g
s
.60839 .71359 m .60589 .71175 L .49803 .65895 L .742 .705 .816 r
F
.5592 .67719 m .5015 .65999 L p .60589 .71175 L F P
0 g
s
.5015 .65999 m .49803 .65895 L p .60589 .71175 L .742 .705 .816 r
F P
0 g
s
.47669 .67729 m .51064 .68372 L p .48704 .67988 L .768 .516 .554 r
F P
0 g
s
.40769 .63404 m .33206 .6661 L .40402 .63083 L .707 .617 .755 r
F
.40769 .63404 m .42844 .6522 L .33206 .6661 L F
.36141 .6663 m .33206 .6661 L p .42844 .6522 L F P
0 g
s
.42848 .65223 m .36141 .6663 L .42844 .6522 L .707 .617 .755 r
F
.47957 .67881 m .41956 .66335 L .39877 .6644 L .713 .677 .814 r
F
.29941 .67099 m .3326 .66846 L .24898 .6753 L .115 .668 .763 r
F
.34337 .66764 m .33527 .66829 L p .3326 .66846 L F P
0 g
s
.33527 .66829 m .24898 .6753 L p .3326 .66846 L .115 .668 .763 r
F P
0 g
s
.31065 .66982 m .29401 .66825 L .33192 .6685 L p .246 0 0 r
F P
0 g
s
.34051 .66687 m .33206 .6661 L .33405 .66746 L .686 .622 .779 r
F
.57053 .67086 m .50125 .63004 L .5592 .67719 L .778 .658 .73 r
F
.55063 .63618 m .44554 .66085 L .42955 .65241 L .743 .664 .773 r
F
.50824 .62787 m .55063 .63618 L p .42955 .65241 L F P
0 g
s
.42939 .65232 m .36998 .64658 L p .4756 .67672 L .713 .677 .814 r
F P
0 g
s
.42955 .65241 m .42879 .65251 L .42939 .65232 L .743 .664 .773 r
F
.42955 .65241 m .42939 .65232 L .50824 .62787 L F
.49411 .68166 m .48704 .67988 L .51064 .68372 L .768 .516 .554 r
F
.48431 .6792 m .42939 .65232 L .4756 .67672 L .713 .677 .814 r
F
.49406 .68198 m .49243 .68124 L .48431 .6792 L F
.49243 .68124 m .42939 .65232 L .48431 .6792 L F
.49406 .68198 m .50337 .68472 L p .50258 .68272 L .742 .705 .816 r
F P
0 g
s
.49654 .66756 m .49411 .68166 L p .50258 .68272 L .742 .705 .816 r
F P
0 g
s
.49411 .68166 m .49406 .68198 L p .50258 .68272 L .742 .705 .816 r
F P
0 g
s
.50337 .68472 m .61277 .71684 L p .56396 .69614 L .742 .705 .816 r
F P
0 g
s
.56396 .69614 m .49654 .66756 L .50337 .68472 L .742 .705 .816 r
F
.49406 .68198 m .49411 .68192 L p .50105 .68368 L .294 .7 .974 r
F P
0 g
s
.50671 .68536 m .61261 .71677 L .50337 .68472 L .294 .7 .974 r
F
.61277 .71684 m .6181 .7184 L p .60839 .71359 L .742 .705 .816 r
F P
0 g
s
.6181 .7184 m .61277 .71684 L p .61261 .71677 L .294 .7 .974 r
F P
0 g
s
.61277 .71684 m .50337 .68472 L p .61261 .71677 L .294 .7 .974 r
F P
0 g
s
.59555 .70995 m .56117 .69572 L .60558 .71396 L .409 0 0 r
F
.5625 .69592 m .56247 .69354 L .60589 .71175 L p .843 .579 .54 r
F P
0 g
s
.62478 .71966 m .61215 .71485 L .60716 .71238 L .843 .579 .54 r
F
.60589 .71175 m .60639 .71195 L p .60716 .71238 L F P
0 g
s
.60639 .71195 m .62478 .71966 L p .60716 .71238 L .843 .579 .54 r
F P
0 g
s
.6382 .71239 m .60589 .71175 L .59084 .70061 L .791 .695 .758 r
F
.56247 .69354 m .5591 .67724 L p .55922 .67721 L F P
0 g
s
.6359 .70144 m .64811 .71983 L .63566 .70193 L .867 .715 .691 r
F
.63759 .69796 m .64669 .71687 L .6359 .70144 L F
.64669 .71687 m .64811 .71983 L .6359 .70144 L F
.66651 .74601 m .66517 .74331 L p .66121 .73764 L .005 .313 .804 r
F P
0 g
s
.59942 .66608 m .57267 .68716 L .55922 .67721 L .791 .695 .758 r
F
.6019 .66885 m .57267 .68716 L .59942 .66608 L F
.48201 .636 m .50224 .62115 L .43318 .566 L .731 .615 .729 r
F
.45502 .56822 m .43318 .566 L p .50224 .62115 L F P
0 g
s
.49554 .62538 m .44945 .61634 L p .54486 .59422 L .76 .617 .7 r
F P
0 g
s
.66651 .74601 m .65507 .72886 L p .60388 .69051 L .802 .727 .782 r
F P
0 g
s
.63566 .70193 m .64134 .71546 L .63422 .7049 L .867 .715 .691 r
F
.63303 .70735 m .60388 .69051 L p .65507 .72886 L .802 .727 .782 r
F P
0 g
s
.64702 .72898 m .6382 .71239 L p .64134 .71546 L .867 .715 .691 r
F P
0 g
s
.6382 .71239 m .63422 .7049 L p .64134 .71546 L .867 .715 .691 r
F P
0 g
s
.64004 .7114 m .64555 .71458 L p .63496 .70337 L .85 .691 .685 r
F P
0 g
s
.63496 .70337 m .59264 .65852 L .63422 .7049 L .85 .691 .685 r
F
.63496 .70337 m .63422 .7049 L .64004 .7114 L F
.64555 .71458 m .63543 .69347 L p .59304 .6288 L F P
0 g
s
.64555 .71458 m .59304 .6288 L .5927 .63 L .85 .691 .685 r
F
.59526 .62094 m .58821 .64593 L .5735 .62837 L .82 .658 .682 r
F
.62338 .6734 m .60845 .63713 L .5869 .61268 L .85 .691 .685 r
F
.63543 .69347 m .61133 .64315 L p .62338 .6734 L F P
0 g
s
.61133 .64315 m .60845 .63713 L p .62338 .6734 L .85 .691 .685 r
F P
0 g
s
.63543 .69347 m .5869 .61268 L .56965 .59311 L .85 .691 .685 r
F
.62519 .6722 m .61161 .64375 L .63538 .69336 L .68 .309 0 r
F
.62519 .6722 m .63538 .69336 L .63465 .69186 L F
.61133 .64315 m .63759 .69796 L p .61161 .64375 L F P
0 g
s
.65833 .73353 m .63465 .69186 L p .65876 .73363 L .026 .284 .78 r
F P
0 g
s
.63465 .69186 m .62232 .67016 L p .63394 .69041 L .026 .284 .78 r
F P
0 g
s
.62232 .67016 m .6222 .66981 L p .63394 .69041 L .026 .284 .78 r
F P
0 g
s
.6222 .66981 m .61873 .65924 L p .62512 .67234 L .026 .284 .78 r
F P
0 g
s
.62512 .67234 m .63394 .69041 L .6222 .66981 L .026 .284 .78 r
F
.63465 .69186 m .63538 .69336 L .63759 .69796 L .68 .309 0 r
F
.62232 .67016 m .63465 .69186 L p .63566 .70193 L .867 .715 .691 r
F P
0 g
s
.63465 .69186 m .63783 .69746 L p .63759 .69796 L .867 .715 .691 r
F P
0 g
s
.63465 .69186 m .63759 .69796 L .63566 .70193 L .867 .715 .691 r
F
.62232 .67016 m .61161 .64375 L p .61292 .6466 L 0 0 .414 r
F P
0 g
s
.58294 .57944 m .60033 .62602 L .59794 .62276 L .893 .686 .616 r
F
.58258 .58475 m .59564 .6196 L p .57921 .57922 L .82 .658 .682 r
F P
0 g
s
.58294 .57944 m .59308 .60279 L .595 .60896 L .893 .686 .616 r
F
.58294 .57944 m .595 .60896 L .59022 .59894 L F
.60845 .63713 m .61133 .64315 L p .61161 .64375 L .68 .309 0 r
F P
0 g
s
.60845 .63713 m .59564 .6196 L .595 .62187 L .85 .691 .685 r
F
.58331 .57397 m .62232 .67016 L p .58294 .57944 L .893 .686 .616 r
F P
0 g
s
.58258 .58475 m .5499 .4934 L .55792 .54426 L .869 .664 .623 r
F
.56217 .53732 m .5735 .62837 L .55423 .60539 L .82 .658 .682 r
F
.5186 .41491 m .5305 .47363 L .51121 .48994 L .768 .58 .644 r
F
.52109 .3369 m .52114 .33692 L p .52258 .36998 L .844 .604 .575 r
F P
0 g
s
.52664 .42991 m .52181 .37108 L .51814 .41405 L .804 .59 .613 r
F
.51875 .41338 m .51121 .48994 L .50056 .4709 L .768 .58 .644 r
F
.51728 .415 m .51819 .37047 L .50743 .42576 L .737 .533 .616 r
F
.52181 .37108 m .5212 .37204 L .51814 .41405 L .804 .59 .613 r
F
.52109 .3369 m .52112 .33764 L p .51989 .37409 L .771 .546 .593 r
F P
0 g
s
.52112 .33764 m .52256 .36992 L p .52013 .37372 L .771 .546 .593 r
F P
0 g
s
.52112 .33764 m .52013 .37372 L .51989 .37409 L .771 .546 .593 r
F
.52109 .3369 m .51989 .37409 L .51741 .36879 L F
.52256 .36992 m .52109 .3369 L p .52258 .36998 L .844 .604 .575 r
F P
0 g
s
.51974 .348 m .51738 .36872 L p .51744 .36572 L .701 .481 .587 r
F P
0 g
s
.5197 .3481 m .51744 .36572 L .51768 .35339 L .701 .481 .587 r
F
.51974 .348 m .51744 .36572 L .5197 .3481 L F
.52115 .33693 m .52114 .33692 L p .52112 .33703 L .918 .624 .472 r
F P
0 g
s
.521 .33691 m .51995 .34618 L .51768 .35339 L .701 .481 .587 r
F
.51995 .34618 m .5197 .3481 L .51768 .35339 L F
.521 .33692 m .51995 .34618 L p .521 .33691 L F P
0 g
s
.52109 .33727 m .52108 .33736 L .52109 .33727 L 0 0 0 r
F
.52106 .33755 m .52108 .33736 L .52109 .33727 L F
.52105 .3376 m .52106 .33755 L .52106 .33755 L F
.52106 .33755 m .52108 .33736 L .52106 .33755 L F
.52101 .33688 m .521 .33692 L p .521 .33691 L .701 .481 .587 r
F P
0 g
s
.52106 .3375 m .52094 .33688 L .52091 .33688 L p .61 .364 .52 r
F P
0 g
s
.51319 .37105 m .51273 .3689 L .50886 .37564 L .667 .459 .595 r
F
.51027 .37241 m .50886 .37564 L p .51273 .3689 L F P
0 g
s
.50424 .39563 m .50899 .37801 L .50649 .3859 L .667 .459 .595 r
F
.46947 .47163 m .46944 .47137 L .50295 .41392 L .706 .517 .631 r
F
.50743 .42576 m .51819 .37047 L .50239 .41245 L .737 .533 .616 r
F
.51784 .53793 m .48964 .60702 L .47343 .58885 L .76 .617 .7 r
F
.5027 .53413 m .49476 .46943 L .48196 .54896 L .737 .57 .667 r
F
.5305 .47363 m .51338 .49381 L .51121 .48994 L .768 .58 .644 r
F
.51654 .45071 m .51086 .52829 L p .54478 .56569 L .794 .621 .667 r
F P
0 g
s
.44945 .61634 m .47117 .55668 L p .51322 .53155 L .76 .617 .7 r
F P
0 g
s
.55389 .55084 m .51654 .45071 L .54478 .56569 L .794 .621 .667 r
F
.59681 .65619 m .61176 .66266 L p .5591 .67724 L .791 .695 .758 r
F P
0 g
s
.58229 .6643 m .58821 .64593 L .50754 .57357 L .778 .658 .73 r
F
.60388 .69051 m .59264 .65852 L p .58229 .6643 L F P
0 g
s
.59264 .65852 m .58821 .64593 L p .58229 .6643 L .778 .658 .73 r
F P
0 g
s
.53472 .58212 m .50754 .57357 L p .58821 .64593 L .778 .658 .73 r
F P
0 g
s
.55063 .63618 m .54164 .57082 L p .56217 .53732 L .82 .658 .682 r
F P
0 g
s
.55427 .58827 m .56965 .59311 L p .54478 .56569 L .794 .621 .667 r
F P
0 g
s
.54478 .56569 m .51086 .52829 L .54164 .57082 L .794 .621 .667 r
F
.54478 .56569 m .54164 .57082 L .55427 .58827 L F
.54833 .50437 m .55792 .54426 L p .5499 .4934 L .869 .664 .623 r
F P
0 g
s
.55389 .55084 m .54833 .50437 L .53028 .47389 L .794 .621 .667 r
F
.55389 .55084 m .53028 .47389 L .52674 .47806 L F
.55792 .54426 m .54833 .50437 L p .55389 .55084 L F P
0 g
s
.53449 .51884 m .5628 .53193 L p .56217 .53732 L .82 .658 .682 r
F P
0 g
s
.54164 .57082 m .53449 .51884 L p .56217 .53732 L .82 .658 .682 r
F P
0 g
s
.53449 .51884 m .51784 .53793 L .51322 .53155 L .76 .617 .7 r
F
.51784 .53793 m .44945 .61634 L .51322 .53155 L F
.52644 .51715 m .53449 .51884 L p .47117 .55668 L F P
0 g
s
.43884 .53396 m .46919 .46948 L .44508 .49871 L .674 .495 .635 r
F
.44735 .528 m .46919 .46948 L .43884 .53396 L F
.50745 .40708 m .49606 .46834 L .48649 .47634 L .706 .517 .631 r
F
.50256 .4154 m .48649 .47634 L .4717 .4887 L F
.50256 .4154 m .4717 .4887 L .46947 .47163 L F
.48196 .54896 m .49476 .46943 L .45477 .52281 L .737 .57 .667 r
F
.47788 .57055 m .50754 .57357 L p .45477 .52281 L F P
0 g
s
.48846 .50918 m .39638 .60221 L .39256 .58831 L .709 .559 .683 r
F
.48846 .50918 m .39256 .58831 L .41037 .5539 L F
.47437 .50908 m .48846 .50918 L p .41037 .5539 L F P
0 g
s
.43318 .566 m .47173 .48888 L p .46919 .46948 L .674 .495 .635 r
F P
0 g
s
.47907 .45485 m .44755 .50888 L .48098 .4501 L .706 .517 .631 r
F
.46919 .46948 m .46891 .46735 L .43721 .50206 L .674 .495 .635 r
F
.45027 .48465 m .43721 .50206 L p .46891 .46735 L F P
0 g
s
.46919 .46948 m .43721 .50206 L .42456 .52072 L .674 .495 .635 r
F
.46919 .46948 m .42456 .52072 L .41399 .53638 L F
.42456 .52072 m .41495 .53174 L .41402 .53624 L F
.42456 .52072 m .41402 .53624 L .41399 .53638 L F
.4997 .41531 m .50899 .37801 L .50424 .39563 L .667 .459 .595 r
F
.50012 .41486 m .50899 .37801 L .4997 .41531 L F
.49214 .44805 m .51018 .37424 L .50899 .37801 L F
.45098 .48222 m .47769 .45028 L .44828 .4865 L .632 .43 .591 r
F
.48528 .43553 m .47769 .45028 L p .47755 .45045 L F P
0 g
s
.48137 .44085 m .47755 .45045 L .46949 .46009 L .632 .43 .591 r
F
.48137 .44085 m .46949 .46009 L .4738 .45116 L F
.48528 .43553 m .47755 .45045 L .48137 .44085 L F
.47643 .4492 m .47666 .44947 L p .47774 .44699 L .907 .757 .684 r
F P
0 g
s
.50886 .37564 m .49135 .41578 L p .4915 .42185 L .667 .459 .595 r
F P
0 g
s
.49214 .44805 m .50886 .37564 L .4915 .42185 L .667 .459 .595 r
F
.41551 .52906 m .41983 .52458 L p .41495 .53174 L .642 .464 .627 r
F P
0 g
s
.40129 .55178 m .41551 .52906 L .41495 .53174 L .642 .464 .627 r
F
.40129 .55178 m .33348 .61466 L .41551 .52906 L F
.33031 .61717 m .41551 .52906 L p .33348 .61466 L F P
0 g
s
.33715 .61017 m .33031 .61717 L .34385 .60469 L .317 .031 .311 r
F
.41551 .52906 m .33031 .61717 L p .41438 .53114 L F P
0 g
s
.32632 .63391 m .3156 .6308 L .31547 .63919 L .682 .543 .69 r
F
.31845 .62794 m .34484 .60564 L .33507 .61339 L .798 .77 .828 r
F
.31845 .62794 m .32617 .62118 L .32255 .62381 L F
.32255 .62381 m .32634 .62128 L p .29996 .64655 L .656 .521 .688 r
F P
0 g
s
.32634 .62128 m .33031 .61717 L p .33348 .61466 L .642 .464 .627 r
F P
0 g
s
.3248 .62154 m .33644 .61152 L .329 .61731 L .798 .77 .828 r
F
.38549 .56928 m .34385 .60469 L p .33644 .61152 L F P
0 g
s
.34385 .60469 m .329 .61731 L p .33644 .61152 L .798 .77 .828 r
F P
0 g
s
.41639 .52964 m .41816 .52704 L .38549 .56928 L .854 .768 .766 r
F
.41438 .53114 m .38549 .56928 L p .42301 .52077 L F P
0 g
s
.38549 .56928 m .41438 .53114 L p .41268 .53427 L .148 0 .202 r
F P
0 g
s
.41857 .54345 m .44508 .49871 L .41399 .53638 L .674 .495 .635 r
F
.43318 .566 m .44508 .49871 L .41857 .54345 L F
.44214 .49514 m .44918 .48601 L .44557 .49054 L .854 .768 .766 r
F
.41639 .52964 m .43888 .49972 L .44214 .49514 L F
.43888 .49972 m .44918 .48601 L .44214 .49514 L F
.49155 .42397 m .49135 .41578 L .49083 .42474 L .667 .459 .595 r
F
.50596 .39531 m .48528 .43553 L p .45098 .48222 L .632 .43 .591 r
F P
0 g
s
.4738 .45116 m .46949 .46009 L .45098 .48222 L .632 .43 .591 r
F
.45351 .47877 m .45098 .48272 L .4543 .4777 L .388 .147 .423 r
F
.45351 .47877 m .4543 .4777 L .47838 .44137 L F
.46279 .46436 m .45351 .47877 L .47838 .44137 L F
.43721 .50206 m .45027 .48465 L p .43816 .50073 L .854 .768 .766 r
F P
0 g
s
.45027 .48465 m .44918 .48601 L .41639 .52964 L .854 .768 .766 r
F
.44755 .50888 m .43242 .51157 L p .44629 .48964 L .632 .43 .591 r
F P
0 g
s
.44828 .4865 m .44825 .48718 L .44629 .48964 L .632 .43 .591 r
F
.44825 .48718 m .44755 .50888 L .44629 .48964 L F
.47769 .45028 m .45027 .48465 L .44828 .4865 L F
.49147 .41548 m .50988 .37324 L .49358 .41018 L .907 .757 .684 r
F
.50987 .37298 m .50979 .37344 L .50961 .37383 L .624 .404 .563 r
F
.49135 .41578 m .51027 .37241 L p .49147 .41548 L .907 .757 .684 r
F P
0 g
s
.51471 .36119 m .51273 .3689 L p .50989 .37291 L .624 .404 .563 r
F P
0 g
s
.52086 .33745 m .52091 .33699 L .52094 .33688 L .701 .481 .587 r
F
.52093 .33697 m .52091 .33691 L p .52094 .33691 L .464 .275 .539 r
F P
0 g
s
.52091 .33691 m .52091 .33688 L p .52092 .33689 L .464 .275 .539 r
F P
0 g
s
.52091 .33691 m .52092 .33689 L .52094 .33691 L .464 .275 .539 r
F
.52094 .33688 m .52101 .33688 L p .52086 .33745 L .701 .481 .587 r
F P
0 g
s
.52086 .33745 m .51381 .36483 L .52091 .33699 L .701 .481 .587 r
F
.50596 .39531 m .51374 .36497 L p .51381 .36483 L F P
0 g
s
.51374 .36497 m .52094 .33688 L p .51381 .36483 L .701 .481 .587 r
F P
0 g
s
.51027 .37241 m .51242 .36748 L p .51022 .37251 L .907 .757 .684 r
F P
0 g
s
.51483 .36086 m .5146 .36332 L .51381 .36483 L .701 .481 .587 r
F
.51768 .35339 m .51635 .36 L .51483 .36086 L F
.51635 .36 m .51368 .37329 L .5146 .36332 L F
.51635 .36 m .5146 .36332 L .51483 .36086 L F
.51032 .37231 m .51027 .37241 L .50987 .37298 L .624 .404 .563 r
F
.51027 .37241 m .51022 .37251 L .49147 .41548 L .907 .757 .684 r
F
.5088 .37556 m .49358 .41018 L p .50988 .37324 L F P
0 g
s
.50892 .37549 m .51022 .37251 L .50926 .37466 L .907 .757 .684 r
F
.51273 .3689 m .50596 .39531 L p .50987 .37298 L .624 .404 .563 r
F P
0 g
s
.50989 .37291 m .51032 .37231 L .50987 .37298 L .624 .404 .563 r
F
.52094 .33688 m .51471 .36119 L p .50989 .37291 L F P
0 g
s
.50989 .37291 m .50989 .37292 L .52091 .33688 L .624 .404 .563 r
F
.50958 .37389 m .52091 .33688 L p .50961 .37383 L F P
0 g
s
.52091 .33688 m .50985 .37301 L p .51338 .36464 L .425 .209 .479 r
F P
0 g
s
.50989 .37291 m .50961 .37383 L .50989 .37292 L .624 .404 .563 r
F
.50985 .37301 m .50958 .37389 L p .51022 .37234 L .425 .209 .479 r
F P
0 g
s
.50985 .37301 m .51022 .37234 L .51338 .36464 L .425 .209 .479 r
F
.49358 .41064 m .50892 .37549 L .50926 .37466 L .907 .757 .684 r
F
.49361 .41058 m .50892 .37549 L .49358 .41064 L F
.5088 .37556 m .51022 .37234 L p .5032 .38919 L .278 .05 .378 r
F P
0 g
s
.51201 .36826 m .5088 .37556 L p .51226 .36778 L .907 .757 .684 r
F P
0 g
s
.51022 .37234 m .51201 .36826 L p .51054 .37175 L .278 .05 .378 r
F P
0 g
s
.50988 .37324 m .51226 .36778 L .5088 .37556 L .907 .757 .684 r
F
.50961 .37383 m .5074 .38683 L .50284 .3959 L .624 .404 .563 r
F
.50596 .39531 m .50284 .3959 L p .5074 .38683 L F P
0 g
s
.47774 .44699 m .48611 .42777 L .47643 .4492 L .907 .757 .684 r
F
.48853 .42287 m .484 .43003 L .49139 .41521 L .632 .43 .591 r
F
.48853 .42287 m .49139 .41521 L .49135 .41578 L F
.48611 .42777 m .49147 .41548 L .48519 .42956 L .907 .757 .684 r
F
.48611 .42777 m .48519 .42956 L .47643 .4492 L F
.47643 .4492 m .46779 .46061 L p .46715 .46163 L .854 .768 .766 r
F P
0 g
s
.44557 .49054 m .44649 .48931 L p .44629 .48964 L .632 .43 .591 r
F P
0 g
s
.49147 .41548 m .49361 .41058 L .49358 .41064 L .907 .757 .684 r
F
.44546 .49015 m .50284 .3959 L p .44799 .48682 L .632 .43 .591 r
F P
0 g
s
.44691 .48777 m .44409 .4924 L p .44817 .48627 L .388 .147 .423 r
F P
0 g
s
.46891 .46735 m .44755 .50888 L p .44828 .4865 L .632 .43 .591 r
F P
0 g
s
.484 .43003 m .44629 .48964 L .491 .41553 L .632 .43 .591 r
F
.44409 .4924 m .45098 .48272 L .46279 .46436 L .388 .147 .423 r
F
.44557 .49054 m .46715 .46163 L .43676 .50158 L .854 .768 .766 r
F
.46779 .46061 m .43676 .50158 L p .46715 .46163 L F P
0 g
s
.4499 .48424 m .46779 .46061 L .45112 .48279 L .148 0 .202 r
F
.46779 .46061 m .47643 .4492 L .46776 .46071 L F
.46779 .46061 m .46776 .46071 L .45112 .48279 L F
.45027 .48465 m .4652 .46474 L p .44918 .48601 L .854 .768 .766 r
F P
0 g
s
.4652 .46474 m .47666 .44947 L p .46534 .46452 L .854 .768 .766 r
F P
0 g
s
.4652 .46474 m .46534 .46452 L .44918 .48601 L .854 .768 .766 r
F
.44755 .50888 m .48528 .43553 L p .4862 .43422 L .706 .517 .631 r
F P
0 g
s
.48528 .43553 m .49083 .42474 L p .4862 .43422 L .706 .517 .631 r
F P
0 g
s
.49135 .41578 m .49139 .41521 L .49233 .41332 L .632 .43 .591 r
F
.484 .43003 m .491 .41553 L .49233 .41332 L F
.49147 .41548 m .49251 .41286 L .48433 .43122 L .907 .757 .684 r
F
.4912 .4156 m .48433 .43122 L p .49251 .41286 L F P
0 g
s
.49147 .41548 m .48433 .43122 L .48519 .42956 L .907 .757 .684 r
F
.49146 .42407 m .4862 .43422 L .49083 .42474 L .706 .517 .631 r
F
.48098 .4501 m .44755 .50888 L .4862 .43422 L F
.48893 .43035 m .4862 .43422 L .49146 .42407 L F
.48098 .4501 m .4862 .43422 L .48893 .43035 L F
.47666 .44947 m .48945 .44829 L p .49062 .42825 L .667 .459 .595 r
F P
0 g
s
.50284 .3959 m .50596 .39531 L p .48853 .42287 L .632 .43 .591 r
F P
0 g
s
.48853 .42287 m .49233 .41332 L .50284 .3959 L .632 .43 .591 r
F
.49135 .41578 m .47774 .44698 L p .49062 .42825 L .667 .459 .595 r
F P
0 g
s
.47774 .44698 m .47666 .44947 L p .49062 .42825 L .667 .459 .595 r
F P
0 g
s
.47769 .45028 m .46891 .46735 L p .45027 .48465 L .632 .43 .591 r
F P
0 g
s
.46807 .46092 m .45027 .48465 L p .46891 .46735 L .674 .495 .635 r
F P
0 g
s
.46649 .46463 m .44409 .50513 L p .46588 .46651 L 0 0 .107 r
F P
0 g
s
.48945 .44829 m .49214 .44805 L p .49155 .42397 L .667 .459 .595 r
F P
0 g
s
.49155 .42397 m .49083 .42474 L .48945 .44829 L .667 .459 .595 r
F
.51635 .36 m .51738 .36872 L .51368 .37329 L .701 .481 .587 r
F
.5159 .36557 m .50827 .39207 L p .51618 .37486 L .737 .533 .616 r
F P
0 g
s
.51709 .35672 m .51592 .35936 L p .51591 .35941 L .907 .757 .684 r
F P
0 g
s
.51819 .37047 m .50827 .39207 L .50526 .40485 L .737 .533 .616 r
F
.51242 .36748 m .51711 .35674 L p .51241 .3675 L .907 .757 .684 r
F P
0 g
s
.51381 .36483 m .51368 .37329 L .51275 .36897 L .701 .481 .587 r
F
.5146 .36332 m .51368 .37329 L .51381 .36483 L F
.51711 .35674 m .51242 .36748 L p .51319 .37105 L .667 .459 .595 r
F P
0 g
s
.50166 .41323 m .51319 .37105 L .5003 .41468 L .667 .459 .595 r
F
.5176 .35763 m .51738 .36872 L .51635 .36 L .701 .481 .587 r
F
.52098 .33784 m .52108 .33723 L .52106 .3375 L 0 0 .286 r
F
.52106 .3375 m .52112 .33707 L .52106 .3375 L 0 0 0 r
F
.52106 .3375 m .52114 .33692 L .52114 .33692 L p .853 .553 .481 r
F P
0 g
s
.52112 .33707 m .52112 .33705 L .52111 .33713 L 0 0 0 r
F
.52112 .33707 m .52111 .33713 L .52106 .3375 L F
.52106 .33764 m .52093 .33697 L .52093 .33698 L .61 .364 .52 r
F
.52106 .33762 m .52114 .33691 L .52114 .33699 L .987 .767 .373 r
F
.52106 .33761 m .52114 .33691 L .52106 .33762 L F
.52114 .33692 m .52115 .33697 L p .52114 .33698 L .844 .604 .575 r
F P
0 g
s
.52106 .3375 m .52115 .33693 L p .52112 .33703 L .918 .624 .472 r
F P
0 g
s
.52104 .33769 m .52105 .33759 L .52105 .3376 L 0 0 0 r
F
.52105 .33759 m .52108 .33736 L .52105 .3376 L F
.52103 .33777 m .52106 .3375 L .52104 .33769 L F
.51995 .34618 m .51974 .348 L p .5197 .3481 L .701 .481 .587 r
F P
0 g
s
.51963 .34954 m .51741 .36879 L .52019 .34411 L .771 .546 .593 r
F
.51861 .35662 m .5202 .35681 L p .52106 .3375 L .736 .504 .577 r
F P
0 g
s
.52038 .34279 m .52106 .3375 L .52103 .33807 L .66 .434 .566 r
F
.52106 .33754 m .52106 .33753 L .52106 .33754 L .987 .767 .373 r
F
.52103 .33777 m .52106 .3375 L .52103 .33777 L 0 0 0 r
F
.52106 .3375 m .51848 .35661 L .51861 .35662 L p .736 .504 .577 r
F P
0 g
s
.51848 .35661 m .52038 .34279 L .52047 .34075 L .66 .434 .566 r
F
.52038 .34279 m .52103 .33807 L .52092 .33861 L F
.52038 .34279 m .52092 .33861 L .52047 .34075 L F
.52114 .33692 m .52106 .3375 L p .52112 .33703 L .918 .624 .472 r
F P
0 g
s
.52113 .337 m .52106 .3375 L p .52112 .33705 L 0 0 0 r
F P
0 g
s
.52089 .3384 m .52106 .3375 L .52088 .33852 L 0 0 .286 r
F
.52089 .3384 m .52098 .33784 L .52106 .3375 L F
.52083 .33879 m .52089 .3384 L .52088 .33852 L F
.52106 .3375 m .52101 .33688 L p .52095 .33701 L .687 .415 .505 r
F P
0 g
s
.52106 .33753 m .52114 .33691 L .52106 .33758 L .987 .767 .373 r
F
.52106 .3375 m .52114 .33691 L p .52106 .33753 L F P
0 g
s
.52106 .3375 m .52106 .33753 L .52106 .33753 L .987 .767 .373 r
F
.52105 .33756 m .52106 .3375 L .52105 .33756 L 0 0 0 r
F
.52112 .33705 m .52106 .3375 L .52106 .33753 L F
.52114 .33692 m .52111 .33691 L p .52106 .3375 L .853 .553 .481 r
F P
0 g
s
.52106 .3375 m .52108 .33725 L .52108 .33726 L 0 0 .286 r
F
.52106 .3375 m .52108 .33726 L .52107 .33731 L F
.52111 .33691 m .52109 .3369 L p .52106 .3375 L .853 .553 .481 r
F P
0 g
s
.52112 .33703 m .52111 .33711 L .52111 .33712 L .918 .624 .472 r
F
.52111 .33712 m .52106 .3375 L .52107 .33746 L F
.52107 .33742 m .52108 .33735 L .52106 .33748 L F
.52108 .33723 m .52109 .33721 L .52106 .3375 L 0 0 .286 r
F
.52108 .33735 m .52107 .33746 L .52106 .33748 L .918 .624 .472 r
F
.52106 .3375 m .52106 .3375 L .52106 .33751 L .945 .734 .586 r
F
.52089 .3384 m .52106 .3375 L .52107 .33731 L 0 0 .286 r
F
.52083 .33879 m .52106 .3375 L .52089 .3384 L F
.52108 .33735 m .52106 .33748 L .52106 .33748 L .918 .624 .472 r
F
.52108 .33735 m .52112 .33703 L .52107 .33746 L F
.52109 .33721 m .52109 .33721 L .52106 .3375 L 0 0 .286 r
F
.52108 .33736 m .52108 .33736 L .52109 .33727 L 0 0 0 r
F
.52108 .33736 m .52115 .33685 L .52109 .33727 L F
.52114 .33691 m .52114 .33691 L p .52106 .33753 L .987 .767 .373 r
F P
0 g
s
.52106 .3375 m .52115 .33685 L p .52104 .33769 L 0 0 0 r
F P
0 g
s
.52114 .33692 m .52114 .33698 L .52114 .33703 L .844 .604 .575 r
F
.52106 .3375 m .52118 .3368 L p .52108 .33725 L 0 0 .286 r
F P
0 g
s
.52106 .3375 m .52106 .33751 L .52096 .33802 L .945 .734 .586 r
F
.52105 .33756 m .52106 .33751 L .52106 .33751 L F
.52106 .33753 m .52106 .3375 L .52105 .33756 L 0 0 0 r
F
.52103 .33807 m .52106 .3375 L .52047 .34075 L .66 .434 .566 r
F
.52099 .33815 m .52107 .33757 L .52103 .33783 L .987 .767 .373 r
F
.52107 .33757 m .52115 .33695 L .52106 .3375 L F
.52106 .3375 m .52109 .3369 L p .52105 .33739 L .769 .477 .491 r
F P
0 g
s
.52106 .33708 m .52106 .3375 L .52109 .3369 L .853 .553 .481 r
F
.52105 .33737 m .52106 .33723 L .52106 .3375 L F
.52109 .3369 m .52106 .3375 L p .52106 .33723 L F P
0 g
s
.52106 .3375 m .52107 .33737 L .52106 .33738 L .336 .137 .454 r
F
.52106 .3375 m .52106 .33739 L .52104 .33749 L F
.52106 .33758 m .52106 .3375 L .52093 .33697 L .61 .364 .52 r
F
.52107 .33757 m .52106 .33751 L .52106 .33752 L .987 .767 .373 r
F
.52107 .33757 m .52106 .3375 L .52106 .33751 L F
.52106 .33764 m .52115 .33693 L .52106 .33764 L F
.52056 .33996 m .52106 .33751 L .52105 .33756 L .945 .734 .586 r
F
.52096 .33802 m .52106 .3375 L p .52096 .33802 L F P
0 g
s
.52056 .33997 m .52096 .33802 L p .52056 .33996 L .945 .734 .586 r
F P
0 g
s
.52096 .33802 m .52096 .33802 L .52056 .33996 L .945 .734 .586 r
F
.51711 .35674 m .52056 .33997 L p .52056 .33996 L F P
0 g
s
.52096 .33802 m .52106 .3375 L .52096 .33802 L .945 .734 .586 r
F
.52106 .3375 m .52096 .33802 L p .52106 .3375 L F P
0 g
s
.52127 .33776 m .52667 .37407 L .52661 .37391 L .911 .651 .532 r
F
.52127 .33776 m .52667 .37407 L p .52127 .33776 L F P
0 g
s
.52505 .36604 m .52661 .37391 L .52585 .372 L .911 .651 .532 r
F
.52115 .33693 m .52661 .37391 L .52505 .36604 L F
.53742 .42551 m .52675 .37461 L .52771 .38061 L .88 .643 .575 r
F
.53842 .43013 m .53754 .42597 L .53742 .42551 L F
.53754 .42597 m .52675 .37461 L .53742 .42551 L F
.51737 .45099 m .51654 .45071 L p .54833 .50437 L .794 .621 .667 r
F P
0 g
s
.54554 .52395 m .53449 .51884 L p .55105 .48531 L .835 .63 .625 r
F P
0 g
s
.58304 .57795 m .53931 .46341 L .58258 .58475 L .869 .664 .623 r
F
.53135 .4192 m .54078 .44455 L .53732 .48539 L .835 .63 .625 r
F
.53931 .46341 m .53712 .45767 L .5389 .46264 L .869 .664 .623 r
F
.53931 .46341 m .5389 .46264 L .58258 .58475 L F
.55815 .51276 m .55105 .48531 L .55013 .49176 L F
.58304 .57795 m .57718 .55905 L .55815 .51276 L F
.56871 .53844 m .55105 .48531 L .55815 .51276 L F
.57718 .55905 m .56433 .52278 L .56871 .53844 L F
.56433 .52278 m .55105 .48531 L .56871 .53844 L F
.55161 .48634 m .56779 .53568 L .55778 .51149 L .904 .666 .569 r
F
.5632 .52195 m .56779 .53568 L p .56289 .52074 L .773 .477 0 r
F P
0 g
s
.55376 .49036 m .56779 .53568 L .55161 .48634 L .904 .666 .569 r
F
.53846 .42972 m .52675 .37461 L p .53842 .43013 L .88 .643 .575 r
F P
0 g
s
.53808 .42794 m .53846 .42972 L p .53794 .42747 L .955 .657 .223 r
F P
0 g
s
.57917 .56389 m .55645 .49537 L .57718 .55905 L .869 .664 .623 r
F
.58326 .57468 m .55645 .49537 L p .58318 .57597 L F P
0 g
s
.57031 .53758 m .55783 .49956 L .58165 .56993 L .889 .557 .061 r
F
.57031 .53758 m .58165 .56993 L .58116 .56867 L F
.5577 .49908 m .58167 .56998 L p .55783 .49956 L F P
0 g
s
.58167 .56998 m .58326 .57468 L p .58165 .56993 L .889 .557 .061 r
F P
0 g
s
.55783 .49956 m .58167 .56998 L .58165 .56993 L .889 .557 .061 r
F
.577 .55795 m .57911 .56338 L .56779 .53568 L 0 0 .414 r
F
.5946 .60823 m .58318 .57597 L .58304 .57795 L .869 .664 .623 r
F
.58318 .57597 m .57718 .55905 L .58304 .57795 L F
.61176 .66266 m .58258 .58475 L p .58294 .57944 L .893 .686 .616 r
F P
0 g
s
.59077 .59819 m .58304 .57795 L .58258 .58475 L .869 .664 .623 r
F
.59564 .6196 m .59526 .62094 L .57789 .57705 L .82 .658 .682 r
F
.59564 .6196 m .57789 .57705 L .57921 .57922 L F
.64 .70345 m .61681 .65496 L p .61524 .65164 L .68 .309 0 r
F P
0 g
s
.64 .70345 m .61524 .65164 L .61351 .64787 L .68 .309 0 r
F
.62232 .67016 m .61471 .66476 L p .59308 .60279 L .893 .686 .616 r
F P
0 g
s
.61471 .66476 m .61176 .66266 L p .59308 .60279 L .893 .686 .616 r
F P
0 g
s
.60839 .63705 m .5946 .60823 L .58154 .60031 L p .85 .691 .685 r
F P
0 g
s
.5946 .60823 m .60845 .63713 L p .6119 .64439 L .68 .309 0 r
F P
0 g
s
.62059 .66265 m .61273 .6462 L .6119 .64439 L .68 .309 0 r
F
.61273 .6462 m .5946 .60823 L .6119 .64439 L F
.62021 .66373 m .62232 .67016 L p .61292 .6466 L 0 0 .414 r
F P
0 g
s
.61524 .65164 m .62021 .66373 L .61292 .6466 L 0 0 .414 r
F
.61639 .65393 m .61351 .64787 L .62059 .66265 L .68 .309 0 r
F
.61351 .64787 m .5946 .60823 L .61273 .6462 L F
.61351 .64787 m .61273 .6462 L .62059 .66265 L F
.6161 .6535 m .62021 .66373 L .61524 .65164 L 0 0 .414 r
F
.5946 .60823 m .58971 .59794 L .58675 .59722 L .869 .664 .623 r
F
.5915 .60635 m .5946 .60823 L p .58675 .59722 L F P
0 g
s
.61681 .65496 m .59112 .60122 L .5946 .60823 L p .68 .309 0 r
F P
0 g
s
.64 .70345 m .61639 .65393 L .62059 .66265 L .68 .309 0 r
F
.64239 .70825 m .64 .70345 L p .60845 .63713 L F P
0 g
s
.61548 .65216 m .61681 .65496 L p .59934 .6098 L .157 .321 .779 r
F P
0 g
s
.5946 .60823 m .59077 .59819 L .58971 .59794 L .869 .664 .623 r
F
.55792 .54426 m .56965 .59311 L p .59077 .59819 L F P
0 g
s
.577 .55795 m .57031 .53758 L .57975 .56502 L .889 .557 .061 r
F
.58331 .57397 m .58116 .56867 L p .60481 .62628 L 0 0 .414 r
F P
0 g
s
.58116 .56867 m .56779 .53568 L p .57911 .56338 L 0 0 .414 r
F P
0 g
s
.57031 .53758 m .58116 .56867 L .57975 .56502 L .889 .557 .061 r
F
.60481 .62628 m .58116 .56867 L .57911 .56338 L 0 0 .414 r
F
.5946 .60823 m .58326 .57468 L p .58318 .57597 L .869 .664 .623 r
F P
0 g
s
.58116 .56867 m .58165 .56993 L .58326 .57468 L .889 .557 .061 r
F
.58318 .57597 m .57917 .56389 L .57718 .55905 L .869 .664 .623 r
F
.56779 .53568 m .58116 .56867 L p .58294 .57944 L .893 .686 .616 r
F P
0 g
s
.58116 .56867 m .58331 .57397 L p .58326 .57468 L .893 .686 .616 r
F P
0 g
s
.58116 .56867 m .58326 .57468 L .58294 .57944 L .893 .686 .616 r
F
.55645 .49537 m .5577 .49908 L p .55783 .49956 L .889 .557 .061 r
F P
0 g
s
.53794 .42747 m .56779 .53568 L p .54811 .47208 L .904 .666 .569 r
F P
0 g
s
.56224 .51565 m .56328 .52112 L .56779 .53568 L .773 .477 0 r
F
.56328 .52112 m .56419 .52471 L .56779 .53568 L F
.56779 .53568 m .55783 .49956 L p .55828 .50137 L F P
0 g
s
.56779 .53568 m .5644 .52553 L p .57507 .553 L 0 0 .414 r
F P
0 g
s
.57507 .553 m .577 .55795 L .57392 .5505 L 0 0 .414 r
F
.57507 .553 m .57392 .5505 L .56779 .53568 L F
.5946 .60823 m .59137 .60172 L p .577 .55795 L .889 .557 .061 r
F P
0 g
s
.577 .55795 m .57975 .56502 L .5946 .60823 L .889 .557 .061 r
F
.59282 .5962 m .57911 .56338 L .577 .55795 L 0 0 .414 r
F
.59112 .60122 m .57554 .55401 L .57594 .55487 L .889 .557 .061 r
F
.55828 .50137 m .55953 .50628 L .56224 .51565 L .773 .477 0 r
F
.53146 .41896 m .54078 .44455 L .53135 .4192 L .835 .63 .625 r
F
.53806 .43373 m .54078 .44455 L p .53146 .41896 L F P
0 g
s
.53763 .43805 m .53736 .43217 L .53703 .43617 L .88 .643 .575 r
F
.53736 .43217 m .53699 .42394 L .53702 .4314 L F
.53736 .43217 m .53702 .4314 L .53703 .43617 L F
.53839 .43037 m .53699 .42394 L .53763 .43805 L F
.53125 .40289 m .52795 .39076 L .5342 .42143 L F
.52795 .39076 m .52663 .3859 L .52831 .3938 L F
.52795 .39076 m .52831 .3938 L .5342 .42143 L F
.52115 .33694 m .52106 .3375 L p .52115 .33695 L .987 .767 .373 r
F P
0 g
s
.51848 .35661 m .52106 .3375 L p .52038 .34279 L .66 .434 .566 r
F P
0 g
s
.51886 .35466 m .52106 .3375 L p .52103 .33777 L 0 0 0 r
F P
0 g
s
.52118 .33799 m .52514 .37576 L .52258 .36998 L .844 .604 .575 r
F
.52323 .35802 m .52109 .33747 L .52114 .33827 L .886 .627 .54 r
F
.52307 .35734 m .52114 .33827 L .5222 .35356 L F
.52323 .35802 m .52114 .33827 L .52307 .35734 L F
.52105 .3376 m .52114 .33688 L .52104 .33769 L 0 0 0 r
F
.52091 .33688 m .52106 .3375 L p .52105 .33759 L .61 .364 .52 r
F P
0 g
s
.52106 .3375 m .51711 .35674 L p .52047 .34075 L .66 .434 .566 r
F P
0 g
s
.51711 .35674 m .51756 .3567 L p .51805 .35405 L .66 .434 .566 r
F P
0 g
s
.51805 .35405 m .52047 .34075 L .51718 .35638 L .66 .434 .566 r
F
.51805 .35405 m .51718 .35638 L .51711 .35674 L F
.51756 .3567 m .51848 .35661 L p .52047 .34075 L F P
0 g
s
.51319 .37105 m .51815 .35752 L .51711 .35674 L .667 .459 .595 r
F
.51825 .35663 m .51711 .35674 L p .51815 .35752 L F P
0 g
s
.51709 .35672 m .518 .35601 L p .51546 .36171 L .278 .05 .378 r
F P
0 g
s
.51573 .36035 m .51709 .35672 L .51546 .36171 L .278 .05 .378 r
F
.51032 .37231 m .51273 .3689 L .51027 .37241 L .624 .404 .563 r
F
.51242 .36748 m .51027 .37241 L p .51273 .3689 L .667 .459 .595 r
F P
0 g
s
.51235 .36868 m .51683 .35734 L .51399 .365 L .278 .05 .378 r
F
.50979 .37344 m .50596 .39531 L .50961 .37383 L .624 .404 .563 r
F
.51236 .36808 m .51683 .35734 L .51266 .3679 L .278 .05 .378 r
F
.51236 .36845 m .51266 .3679 L .51235 .36868 L F
.51236 .36808 m .51266 .3679 L .51236 .36845 L F
.51226 .36778 m .51591 .35941 L .51238 .36742 L .907 .757 .684 r
F
.51592 .35936 m .51238 .36742 L p .51591 .35941 L F P
0 g
s
.51241 .3675 m .51711 .35674 L .51232 .36766 L .907 .757 .684 r
F
.51711 .35674 m .51709 .35672 L p .51226 .36778 L F P
0 g
s
.51232 .36766 m .51711 .35674 L .51226 .36778 L .907 .757 .684 r
F
.51238 .36742 m .51709 .35672 L p .51237 .36774 L .278 .05 .378 r
F P
0 g
s
.52101 .33688 m .51768 .35339 L .51483 .36086 L .701 .481 .587 r
F
.51106 .3708 m .51237 .36774 L .51176 .36954 L .278 .05 .378 r
F
.51237 .36774 m .51709 .35672 L .51236 .36808 L F
.51237 .36774 m .51236 .36808 L .51176 .36954 L F
.51683 .35734 m .51709 .35672 L .51399 .365 L F
.52093 .33698 m .52093 .33697 L .52091 .33691 L .61 .364 .52 r
F
.52091 .33691 m .52091 .33691 L .52091 .33688 L F
.52093 .33698 m .52091 .33691 L .52091 .33691 L F
.52093 .33687 m .52094 .33691 L p .52094 .33691 L .464 .275 .539 r
F P
0 g
s
.52094 .33691 m .52093 .33687 L .52094 .33686 L p .311 .178 .533 r
F P
0 g
s
.52094 .33686 m .52097 .33683 L p .52094 .33691 L .311 .178 .533 r
F P
0 g
s
.52097 .33683 m .52093 .33687 L p .52066 .33813 L .141 0 .344 r
F P
0 g
s
.49214 .44805 m .51368 .37329 L p .51319 .37105 L .667 .459 .595 r
F P
0 g
s
.51368 .37329 m .51848 .35661 L p .51319 .37105 L .667 .459 .595 r
F P
0 g
s
.51368 .37329 m .51434 .39537 L .51069 .37697 L .701 .481 .587 r
F
.51345 .37226 m .51368 .37329 L .51308 .37402 L F
.51345 .37226 m .51308 .37402 L .51069 .37697 L F
.51275 .36897 m .51345 .37226 L .51069 .37697 L F
.518 .35601 m .51493 .36438 L p .51543 .36185 L .278 .05 .378 r
F P
0 g
s
.50961 .37384 m .50958 .37389 L p .50961 .37383 L .624 .404 .563 r
F P
0 g
s
.50958 .37389 m .50284 .3959 L p .51169 .36967 L .425 .209 .479 r
F P
0 g
s
.51238 .36742 m .51201 .36826 L p .51222 .36786 L .907 .757 .684 r
F P
0 g
s
.51242 .36748 m .51241 .3675 L .51022 .37251 L .907 .757 .684 r
F
.51241 .3675 m .51232 .36766 L .50988 .37324 L F
.51201 .36826 m .51238 .36742 L p .51087 .37126 L .278 .05 .378 r
F P
0 g
s
.51226 .36778 m .51238 .36742 L .51222 .36786 L .907 .757 .684 r
F
.51238 .36742 m .51237 .36774 L .51087 .37126 L .278 .05 .378 r
F
.51022 .37251 m .51241 .3675 L .50988 .37324 L .907 .757 .684 r
F
.50958 .37389 m .51169 .36967 L .51022 .37234 L .425 .209 .479 r
F
.50988 .37324 m .51232 .36766 L .51226 .36778 L .907 .757 .684 r
F
.51201 .36826 m .51087 .37126 L .5032 .38919 L .278 .05 .378 r
F
.51022 .37234 m .51169 .36967 L .51338 .36464 L .425 .209 .479 r
F
.50284 .3959 m .50528 .39449 L p .51234 .36848 L F P
0 g
s
.51169 .36967 m .50284 .3959 L .51234 .36848 L .425 .209 .479 r
F
.51236 .36845 m .51018 .37645 L p .51233 .36974 L .141 0 .344 r
F P
0 g
s
.51018 .37645 m .50528 .39449 L p .51087 .37487 L .141 0 .344 r
F P
0 g
s
.51018 .37645 m .51087 .37487 L .51233 .36974 L .141 0 .344 r
F
.50827 .39207 m .50239 .41245 L p .50526 .40485 L .737 .533 .616 r
F P
0 g
s
.51434 .39537 m .50596 .39531 L p .51069 .37697 L .701 .481 .587 r
F P
0 g
s
.51087 .37487 m .50528 .39449 L .51224 .37352 L .141 0 .344 r
F
.51228 .37159 m .51224 .37352 L .51484 .3657 L F
.51087 .37487 m .51224 .37352 L .51228 .37159 L F
.50528 .39449 m .51236 .36845 L p .51234 .36848 L .425 .209 .479 r
F P
0 g
s
.51233 .36974 m .51087 .37487 L .51232 .37013 L .141 0 .344 r
F
.51232 .37013 m .51087 .37487 L .51228 .37159 L F
.49083 .42474 m .50596 .39531 L p .50745 .40708 L .706 .517 .631 r
F P
0 g
s
.49692 .40327 m .51229 .37143 L .50912 .3801 L .278 .05 .378 r
F
.49316 .41725 m .49878 .40834 L .48937 .43406 L F
.48759 .43793 m .48937 .43406 L .48659 .44164 L F
.49316 .41725 m .48937 .43406 L .48759 .43793 L F
.49888 .408 m .49887 .40805 L .50046 .40523 L 0 0 .107 r
F
.49612 .41256 m .49729 .41071 L .49888 .40795 L F
.49889 .40787 m .49888 .40795 L .50046 .40523 L F
.49612 .41256 m .49888 .40795 L .49889 .40787 L F
.49729 .41071 m .49771 .41004 L .50046 .40523 L F
.49884 .40824 m .4894 .43399 L p .49878 .40834 L .278 .05 .378 r
F P
0 g
s
.4894 .43399 m .48659 .44164 L p .48937 .43406 L .278 .05 .378 r
F P
0 g
s
.49878 .40834 m .4894 .43399 L .48937 .43406 L .278 .05 .378 r
F
.4896 .43354 m .49884 .40824 L .49876 .40885 L 0 0 .051 r
F
.48659 .44164 m .4894 .43399 L p .4896 .43354 L F P
0 g
s
.4894 .43399 m .49884 .40824 L p .4896 .43354 L 0 0 .051 r
F P
0 g
s
.49621 .4269 m .48659 .44164 L p .4896 .43354 L 0 0 .051 r
F P
0 g
s
.4896 .43354 m .49876 .40885 L .49793 .41472 L 0 0 .051 r
F
.4896 .43354 m .49793 .41472 L .49621 .4269 L F
.50596 .39531 m .51434 .39537 L p .50745 .40708 L .706 .517 .631 r
F P
0 g
s
.49793 .41472 m .49641 .4262 L .49621 .4269 L 0 0 .051 r
F
.49876 .40885 m .49798 .4146 L .49793 .41472 L F
.49798 .4146 m .49641 .4262 L .49793 .41472 L F
.49214 .44805 m .47666 .44947 L p .46919 .46948 L .674 .495 .635 r
F P
0 g
s
.47666 .44947 m .46807 .46092 L p .46919 .46948 L .674 .495 .635 r
F P
0 g
s
.48334 .44406 m .48659 .44164 L p .46559 .4674 L .148 0 .202 r
F P
0 g
s
.47666 .44947 m .49135 .41578 L p .49147 .41548 L .907 .757 .684 r
F P
0 g
s
.47643 .4492 m .48473 .43059 L .48598 .438 L .278 .05 .378 r
F
.48659 .44164 m .48482 .44296 L p .48598 .438 L F P
0 g
s
.48482 .44296 m .47643 .4492 L p .48598 .438 L .278 .05 .378 r
F P
0 g
s
.48433 .43122 m .47643 .4492 L p .48519 .42956 L .907 .757 .684 r
F P
0 g
s
.48473 .43059 m .49316 .41725 L .48659 .44164 L .278 .05 .378 r
F
.48473 .43059 m .49116 .41576 L .49096 .41662 L F
.48433 .43122 m .4912 .4156 L p .48473 .43059 L F P
0 g
s
.49251 .41286 m .49142 .41466 L p .4912 .4156 L .388 .147 .423 r
F P
0 g
s
.49358 .41018 m .4912 .4156 L p .49251 .41286 L .907 .757 .684 r
F P
0 g
s
.4912 .4156 m .49116 .41576 L .48473 .43059 L .278 .05 .378 r
F
.48473 .43059 m .49096 .41662 L .48945 .42312 L F
.50284 .3959 m .49251 .41286 L p .48249 .43373 L .388 .147 .423 r
F P
0 g
s
.4912 .4156 m .49358 .41018 L p .49116 .41576 L .278 .05 .378 r
F P
0 g
s
.4912 .4156 m .44691 .48777 L .48249 .43373 L .388 .147 .423 r
F
.49142 .41466 m .44691 .48777 L p .4912 .4156 L F P
0 g
s
.48249 .43373 m .48108 .43666 L .48985 .42138 L .388 .147 .423 r
F
.48988 .42125 m .48985 .42138 L .49126 .41893 L F
.48249 .43373 m .48985 .42138 L .48988 .42125 L F
.49096 .41662 m .49358 .41018 L .49692 .40327 L .278 .05 .378 r
F
.49116 .41576 m .49358 .41018 L .49096 .41662 L F
.47643 .4492 m .48433 .43122 L p .48473 .43059 L F P
0 g
s
.49771 .41004 m .49887 .40805 L .49888 .408 L 0 0 .107 r
F
.47252 .45408 m .49793 .40969 L .49771 .41004 L F
.49793 .40969 m .49887 .40805 L .49771 .41004 L F
.48364 .43363 m .48945 .42312 L p .47838 .44137 L .388 .147 .423 r
F P
0 g
s
.48945 .42312 m .50528 .39449 L p .48985 .42138 L .388 .147 .423 r
F P
0 g
s
.47838 .44137 m .48945 .42312 L .48985 .42138 L .388 .147 .423 r
F
.44409 .50513 m .44433 .50488 L p .46551 .46765 L 0 0 .107 r
F P
0 g
s
.46551 .46765 m .44433 .50488 L .45304 .49179 L 0 0 .107 r
F
.46588 .46651 m .44409 .50513 L .46551 .46765 L F
.43787 .50309 m .43242 .51157 L .43849 .50229 L .388 .147 .423 r
F
.44557 .49054 m .45005 .48462 L .45051 .48392 L .854 .768 .766 r
F
.46571 .46394 m .47666 .44947 L .46584 .46372 L F
.47666 .44947 m .47643 .4492 L p .45051 .48392 L F P
0 g
s
.44918 .48601 m .47666 .44947 L .44557 .49054 L .854 .768 .766 r
F
.46649 .46463 m .47643 .4492 L .46918 .46144 L .148 0 .202 r
F
.45005 .48462 m .46571 .46394 L .45051 .48392 L .854 .768 .766 r
F
.46571 .46394 m .46584 .46372 L .45051 .48392 L F
.45112 .48279 m .46776 .46071 L .46649 .46463 L .148 0 .202 r
F
.46776 .46071 m .47643 .4492 L .46649 .46463 L F
.47252 .45408 m .46848 .4619 L .48215 .43767 L 0 0 .107 r
F
.4794 .44206 m .48215 .43767 L .49805 .40949 L F
.47252 .45408 m .48215 .43767 L .4794 .44206 L F
.43787 .50309 m .43849 .50229 L .45174 .48206 L .388 .147 .423 r
F
.45033 .48373 m .43787 .50309 L .45174 .48206 L F
.4648 .46663 m .43242 .51157 L .4479 .49804 L F
.47213 .453 m .45519 .47798 L .4648 .46663 L F
.45519 .47798 m .43242 .51157 L .4648 .46663 L F
.47838 .44137 m .4648 .46663 L .48364 .43363 L F
.44409 .50513 m .46649 .46463 L p .4648 .46663 L F P
0 g
s
.46649 .46463 m .48364 .43363 L p .4648 .46663 L .388 .147 .423 r
F P
0 g
s
.45512 .4769 m .43242 .51157 L .47213 .453 L .388 .147 .423 r
F
.44827 .488 m .47422 .45549 L .44494 .49273 L .148 0 .202 r
F
.47422 .45549 m .48334 .44406 L .47575 .45369 L F
.47422 .45549 m .47575 .45369 L .44494 .49273 L F
.45535 .47953 m .47575 .45369 L .45348 .48225 L F
.47575 .45369 m .48334 .44406 L .47666 .45261 L F
.47575 .45369 m .47666 .45261 L .45348 .48225 L F
.44214 .49514 m .44557 .49054 L .4434 .49336 L .854 .768 .766 r
F
.41983 .52458 m .41551 .52906 L p .41438 .53114 L .317 .031 .311 r
F P
0 g
s
.43676 .50158 m .41438 .53114 L p .44557 .49054 L .854 .768 .766 r
F P
0 g
s
.43242 .51157 m .44409 .4924 L p .44629 .48964 L .632 .43 .591 r
F P
0 g
s
.44409 .4924 m .43242 .51157 L p .45098 .48272 L .388 .147 .423 r
F P
0 g
s
.43527 .50643 m .44494 .49273 L .46559 .4674 L .148 0 .202 r
F
.46517 .46868 m .46559 .4674 L .48659 .44164 L F
.44845 .50068 m .45796 .49096 L p .46211 .47816 L 0 0 .107 r
F P
0 g
s
.46551 .46765 m .45304 .49179 L .46211 .47816 L 0 0 .107 r
F
.45304 .49179 m .44845 .50068 L .46211 .47816 L F
.43242 .51157 m .44409 .50513 L p .4479 .49804 L .388 .147 .423 r
F P
0 g
s
.43318 .566 m .41399 .53638 L .41383 .53716 L .674 .495 .635 r
F
.43242 .51157 m .44755 .50888 L p .40129 .55178 L .642 .464 .627 r
F P
0 g
s
.41495 .53174 m .43721 .50206 L p .41639 .52964 L .854 .768 .766 r
F P
0 g
s
.43721 .50206 m .43816 .50073 L .41639 .52964 L .854 .768 .766 r
F
.43721 .50206 m .41495 .53174 L p .42456 .52072 L .674 .495 .635 r
F P
0 g
s
.41639 .52964 m .42301 .52077 L .41816 .52704 L .854 .768 .766 r
F
.41983 .52458 m .43242 .51157 L p .40129 .55178 L .642 .464 .627 r
F P
0 g
s
.44755 .50888 m .38821 .57096 L .40129 .55178 L .642 .464 .627 r
F
.41495 .53174 m .38623 .57003 L p .41399 .53638 L .674 .495 .635 r
F P
0 g
s
.44409 .50513 m .43242 .51157 L p .41157 .53632 L .317 .031 .311 r
F P
0 g
s
.38991 .57868 m .41037 .5539 L p .39256 .58831 L .709 .559 .683 r
F P
0 g
s
.41037 .5539 m .38991 .57868 L p .41043 .55363 L .642 .464 .627 r
F P
0 g
s
.41199 .54609 m .41043 .55363 L .38821 .57096 L .642 .464 .627 r
F
.3914 .56959 m .40741 .56821 L p .41181 .54599 L .674 .495 .635 r
F P
0 g
s
.41181 .54599 m .41362 .53684 L .40094 .55552 L .674 .495 .635 r
F
.41181 .54599 m .40094 .55552 L .3914 .56959 L F
.38969 .5654 m .41639 .52964 L .38549 .56928 L .854 .768 .766 r
F
.38623 .57003 m .3914 .56959 L p .40094 .55552 L .674 .495 .635 r
F P
0 g
s
.40094 .55552 m .41362 .53684 L .39102 .56424 L .674 .495 .635 r
F
.40094 .55552 m .39102 .56424 L .38623 .57003 L F
.35932 .61573 m .32634 .62128 L p .33348 .61466 L .642 .464 .627 r
F P
0 g
s
.34021 .60842 m .34097 .60871 L .33348 .61466 L .642 .464 .627 r
F
.34097 .60871 m .35932 .61573 L .33348 .61466 L F
.3875 .56992 m .38623 .57003 L p .39856 .61014 L .682 .543 .69 r
F P
0 g
s
.38991 .57868 m .35932 .61573 L p .34021 .60842 L .642 .464 .627 r
F P
0 g
s
.33507 .61339 m .38549 .56928 L .3248 .62154 L .798 .77 .828 r
F
.33031 .61717 m .32634 .62128 L p .34676 .6023 L .317 .031 .311 r
F P
0 g
s
.3463 .60268 m .37767 .57595 L .34676 .6023 L 0 0 0 r
F
.34676 .6023 m .37767 .57595 L .35988 .59156 L F
.33031 .61717 m .34676 .6023 L .34385 .60469 L .317 .031 .311 r
F
.3154 .64349 m .26168 .66894 L .24827 .67339 L .656 .521 .688 r
F
.25404 .67245 m .24827 .67339 L p .26168 .66894 L F P
0 g
s
.28959 .65318 m .28824 .65196 L p .25237 .67328 L .731 .765 .881 r
F P
0 g
s
.24898 .6753 m .25453 .67228 L p .25237 .67328 L .731 .765 .881 r
F P
0 g
s
.25453 .67228 m .25404 .67245 L .26168 .66894 L .656 .521 .688 r
F
.33206 .6661 m .28454 .6618 L .33405 .66746 L .686 .622 .779 r
F
.28454 .6618 m .28414 .66194 L .33405 .66746 L F
.31183 .65228 m .25453 .67228 L .26168 .66894 L .656 .521 .688 r
F
.3154 .64388 m .26168 .66894 L .3154 .64349 L F
.36022 .66633 m .3408 .66689 L .36073 .66629 L .696 .455 .554 r
F
.34337 .66764 m .34394 .66718 L p .34051 .66687 L .686 .622 .779 r
F P
0 g
s
.3408 .66689 m .29401 .66825 L .34065 .66688 L .696 .455 .554 r
F
.3408 .66689 m .34065 .66688 L .36073 .66629 L F
.40402 .63083 m .31976 .6577 L .40115 .62832 L .707 .617 .755 r
F
.3154 .64388 m .31183 .65228 L .26168 .66894 L .656 .521 .688 r
F
.31545 .64071 m .3154 .64349 L .29317 .65339 L F
.32632 .63391 m .31547 .63919 L .31531 .6496 L .682 .543 .69 r
F
.32632 .63391 m .31531 .6496 L .31655 .64911 L F
.35932 .61573 m .29317 .65339 L .29996 .64655 L .656 .521 .688 r
F
.31062 .65146 m .28959 .65318 L p .32843 .66362 L .686 .622 .779 r
F P
0 g
s
.31976 .6577 m .29419 .66586 L .31897 .65716 L .707 .617 .755 r
F
.31976 .6577 m .31897 .65716 L .40115 .62832 L F
.31531 .6496 m .31183 .65228 L p .3154 .64388 L .656 .521 .688 r
F P
0 g
s
.36998 .64658 m .31062 .65146 L p .34051 .66687 L .686 .622 .779 r
F P
0 g
s
.34051 .66687 m .31062 .65146 L .32843 .66362 L .686 .622 .779 r
F
.33204 .66778 m .29401 .66825 L .33558 .66705 L .696 .455 .554 r
F
.34044 .66736 m .33558 .66705 L .36141 .6663 L F
.33204 .66778 m .33558 .66705 L .34044 .66736 L F
.47345 .67908 m .47665 .67968 L p .42256 .66948 L .15 0 0 r
F P
0 g
s
.50686 .6854 m .5068 .68538 L .50671 .68536 L .294 .7 .974 r
F
.6181 .7184 m .61515 .71633 L .60839 .71359 L .742 .705 .816 r
F
.61689 .7172 m .61215 .71485 L .62478 .71966 L .843 .579 .54 r
F
.61515 .71633 m .5592 .67719 L .60589 .71175 L .742 .705 .816 r
F
.61515 .71633 m .60589 .71175 L .60839 .71359 L F
.61853 .71784 m .61689 .7172 L .62478 .71966 L .843 .579 .54 r
F
.60388 .69051 m .5592 .67719 L p .61515 .71633 L .742 .705 .816 r
F P
0 g
s
.6181 .7184 m .61769 .71759 L p .61515 .71633 L .742 .705 .816 r
F P
0 g
s
.6181 .7184 m .62249 .7209 L p .62416 .72 L .802 .727 .782 r
F P
0 g
s
.61769 .71759 m .6181 .7184 L p .64266 .72488 L .802 .727 .782 r
F P
0 g
s
.64702 .72898 m .63615 .72675 L .61689 .7172 L .843 .579 .54 r
F
.61769 .71759 m .60388 .69051 L p .61515 .71633 L .742 .705 .816 r
F P
0 g
s
.64702 .72898 m .64711 .73133 L .63615 .72675 L p .843 .579 .54 r
F P
0 g
s
.56373 .69968 m .5068 .68538 L .50686 .6854 L .294 .7 .974 r
F
.61689 .7172 m .62108 .71927 L .5625 .69592 L .843 .579 .54 r
F
.60595 .71411 m .5625 .69592 L p .62108 .71927 L F P
0 g
s
.61689 .7172 m .5625 .69592 L .61215 .71485 L .843 .579 .54 r
F
.58909 .70606 m .5625 .69592 L .59297 .70703 L F
.58909 .70606 m .59297 .70703 L .60589 .71175 L F
.61215 .71485 m .58909 .70606 L .60589 .71175 L F
.60589 .71175 m .56247 .69354 L p .59084 .70061 L .791 .695 .758 r
F P
0 g
s
.59411 .70937 m .56117 .69572 L .59555 .70995 L .409 0 0 r
F
.6069 .71053 m .6181 .7184 L p .61261 .71677 L .294 .7 .974 r
F P
0 g
s
.58033 .70386 m .61261 .71677 L .5068 .68538 L .294 .7 .974 r
F
.59411 .70937 m .55938 .69548 L .56117 .69572 L .409 0 0 r
F
.55926 .69543 m .55929 .69543 L p .55938 .69548 L F P
0 g
s
.55929 .69543 m .56117 .69572 L p .55938 .69548 L .409 0 0 r
F P
0 g
s
.56375 .69969 m .5068 .68538 L .56373 .69968 L .294 .7 .974 r
F
.57988 .70368 m .50671 .68536 L .51062 .68611 L F
.56117 .69572 m .5625 .69592 L .60595 .71411 L p .409 0 0 r
F P
0 g
s
.60558 .71396 m .56117 .69572 L .60595 .71411 L .409 0 0 r
F
.6069 .71053 m .58032 .70385 L .57988 .70368 L .294 .7 .974 r
F
.58032 .70385 m .50671 .68536 L .57988 .70368 L F
.60595 .71411 m .60645 .71432 L p .60558 .71396 L .409 0 0 r
F P
0 g
s
.51062 .68611 m .51571 .68707 L p .51122 .68622 L .15 0 0 r
F P
0 g
s
.51571 .68707 m .51428 .68681 L .51122 .68622 L .15 0 0 r
F
.56247 .69354 m .5625 .69592 L .51062 .68611 L p .768 .516 .554 r
F P
0 g
s
.56247 .69354 m .51062 .68611 L .49976 .68335 L .768 .516 .554 r
F
.60432 .70872 m .6069 .71053 L p .51062 .68611 L .294 .7 .974 r
F P
0 g
s
.51274 .68651 m .51062 .68611 L .50105 .68368 L .294 .7 .974 r
F
.60432 .70872 m .51062 .68611 L .51274 .68651 L F
.55641 .69479 m .55525 .69466 L .5419 .6921 L .15 0 0 r
F
.50105 .68368 m .50671 .68536 L .49406 .68198 L .294 .7 .974 r
F
.50337 .68472 m .49406 .68198 L p .50671 .68536 L F P
0 g
s
.49406 .68198 m .49411 .68166 L p .49243 .68124 L .713 .677 .814 r
F P
0 g
s
.49778 .68239 m .47345 .67908 L .49206 .6814 L .768 .516 .554 r
F
.47957 .67881 m .35689 .65694 L .46658 .67775 L .713 .677 .814 r
F
.42069 .66668 m .42061 .66908 L .36128 .66869 L p .696 .455 .554 r
F P
0 g
s
.42033 .67248 m .47761 .67986 L .35192 .66841 L .347 .765 .978 r
F
.35918 .66863 m .35192 .66841 L .34337 .66764 L F
.42033 .67248 m .35192 .66841 L .35918 .66863 L F
.48371 .681 m .34337 .66764 L p .47761 .67986 L F P
0 g
s
.35689 .65694 m .34337 .66764 L p .39301 .66379 L .713 .677 .814 r
F P
0 g
s
.34337 .66764 m .34392 .66758 L p .34401 .66759 L .347 .765 .978 r
F P
0 g
s
.42033 .67248 m .35918 .66863 L .36128 .66869 L .347 .765 .978 r
F
.34392 .66758 m .34337 .66764 L p .3326 .66846 L .115 .668 .763 r
F P
0 g
s
.33192 .6685 m .3326 .66846 L .29941 .67099 L .115 .668 .763 r
F
.34392 .66758 m .3326 .66846 L .33192 .6685 L F
.39679 .6645 m .35925 .6664 L .34337 .66764 L .713 .677 .814 r
F
.36128 .66869 m .34337 .66764 L .34401 .66759 L .347 .765 .978 r
F
.35918 .66863 m .34337 .66764 L .36128 .66869 L F
.36128 .66869 m .31945 .66842 L p .34909 .66791 L .696 .455 .554 r
F P
0 g
s
.42069 .66668 m .31945 .66842 L .33204 .66778 L .696 .455 .554 r
F
.42848 .65223 m .49803 .65895 L p .39116 .61958 L .731 .615 .729 r
F P
0 g
s
.36617 .6496 m .35689 .65694 L p .39877 .6644 L .713 .677 .814 r
F P
0 g
s
.41956 .66335 m .36617 .6496 L .39877 .6644 L .713 .677 .814 r
F
.36073 .66629 m .36141 .6663 L p .36022 .66633 L .696 .455 .554 r
F P
0 g
s
.39679 .6645 m .39301 .66379 L .35925 .6664 L .713 .677 .814 r
F
.47669 .67729 m .36617 .6496 L .47957 .67881 L F
.36998 .64658 m .36617 .6496 L p .48431 .6792 L F P
0 g
s
.42069 .66668 m .36141 .6663 L p .42848 .65223 L .707 .617 .755 r
F P
0 g
s
.42901 .65212 m .42879 .65251 L p .42848 .65223 L .707 .617 .755 r
F P
0 g
s
.42879 .65251 m .42069 .66668 L p .42848 .65223 L .707 .617 .755 r
F P
0 g
s
.36141 .6663 m .42069 .66668 L p .33204 .66778 L .696 .455 .554 r
F P
0 g
s
.36907 .66505 m .34392 .66758 L p .33192 .6685 L .115 .668 .763 r
F P
0 g
s
.31992 .66924 m .33192 .6685 L .29941 .67099 L .115 .668 .763 r
F
.36907 .66505 m .33192 .6685 L .31992 .66924 L F
.34394 .66718 m .36998 .64658 L p .34051 .66687 L .686 .622 .779 r
F P
0 g
s
.36907 .66505 m .29301 .67154 L .29077 .67157 L .115 .668 .763 r
F
.29401 .66825 m .30029 .67073 L .25431 .6748 L .48 .074 .193 r
F
.24866 .67501 m .24898 .6753 L p .25237 .67328 L .731 .765 .881 r
F P
0 g
s
.24898 .6753 m .24866 .67501 L p .25403 .67481 L .536 .715 .954 r
F P
0 g
s
.25431 .6748 m .25415 .67481 L p .25403 .67481 L .536 .715 .954 r
F P
0 g
s
.25415 .67481 m .24898 .6753 L p .25403 .67481 L .536 .715 .954 r
F P
0 g
s
.24817 .67459 m .24866 .67501 L p .25134 .67345 L .731 .765 .881 r
F P
0 g
s
.29419 .66586 m .25281 .67476 L .24817 .67459 L .62 .389 .546 r
F
.25403 .67481 m .29401 .66825 L p .25431 .6748 L .48 .074 .193 r
F P
0 g
s
.28824 .65196 m .25294 .67088 L p .24817 .67459 L .731 .765 .881 r
F P
0 g
s
.24827 .67339 m .25254 .67113 L p .31442 .6297 L .063 0 0 r
F P
0 g
s
.25254 .67113 m .25268 .67105 L p .25278 .67098 L .063 0 0 r
F P
0 g
s
.31442 .6297 m .25254 .67113 L .25278 .67098 L .063 0 0 r
F
.32255 .62381 m .31564 .62842 L p .31442 .6297 L F P
0 g
s
.329 .61731 m .28824 .65196 L p .31845 .62794 L .798 .77 .828 r
F P
0 g
s
.29416 .64711 m .31373 .63043 L .31102 .63329 L 0 0 0 r
F
.31325 .63069 m .31442 .6297 L p .29416 .64711 L F P
0 g
s
.31442 .6297 m .31373 .63043 L .29416 .64711 L 0 0 0 r
F
.32634 .62128 m .32255 .62381 L p .31269 .63152 L .063 0 0 r
F P
0 g
s
.32634 .62128 m .31269 .63152 L .31162 .63266 L .063 0 0 r
F
.29321 .65001 m .31845 .62794 L .28824 .65196 L .798 .77 .828 r
F
.29416 .64711 m .29682 .64493 L .28988 .65079 L 0 0 0 r
F
.32634 .62128 m .35932 .61573 L p .29996 .64655 L .656 .521 .688 r
F P
0 g
s
.29321 .65001 m .3158 .6306 L .31845 .62794 L .798 .77 .828 r
F
.35125 .60794 m .32634 .62128 L p .31162 .63266 L .063 0 0 r
F P
0 g
s
.29682 .64493 m .31102 .63329 L .28988 .65079 L 0 0 0 r
F
.31102 .63329 m .31066 .63367 L .28988 .65079 L F
.3156 .6308 m .28959 .65318 L .31528 .65107 L p .682 .543 .69 r
F P
0 g
s
.30477 .63988 m .35125 .60794 L p .31162 .63266 L .063 0 0 r
F P
0 g
s
.331 .61691 m .32897 .61869 L .28988 .65079 L 0 0 0 r
F
.38623 .57003 m .38549 .56928 L p .33507 .61339 L .798 .77 .828 r
F P
0 g
s
.34484 .60564 m .38623 .57003 L p .33507 .61339 L .798 .77 .828 r
F P
0 g
s
.38623 .57003 m .34484 .60564 L p .36305 .59475 L .682 .543 .69 r
F P
0 g
s
.35125 .60794 m .38349 .57223 L p .37651 .57795 L .317 .031 .311 r
F P
0 g
s
.39856 .61014 m .43318 .566 L .3875 .56992 L p .682 .543 .69 r
F P
0 g
s
.32634 .62128 m .35125 .60794 L p .37651 .57795 L .317 .031 .311 r
F P
0 g
s
.35125 .60794 m .36482 .59416 L p .38297 .57289 L 0 0 0 r
F P
0 g
s
.36482 .59416 m .37137 .5875 L p .3733 .58507 L 0 0 0 r
F P
0 g
s
.38297 .57289 m .36482 .59416 L .3733 .58507 L 0 0 0 r
F
.35625 .59705 m .34772 .6061 L p .3071 .63804 L F P
0 g
s
.35125 .60794 m .30477 .63988 L p .34967 .60456 L .735 .986 .896 r
F P
0 g
s
.30353 .64101 m .28988 .65079 L p .32897 .61869 L 0 0 0 r
F P
0 g
s
.28824 .65196 m .31325 .63069 L p .29416 .64711 L 0 0 0 r
F P
0 g
s
.28988 .65079 m .28824 .65196 L p .29416 .64711 L 0 0 0 r
F P
0 g
s
.28824 .65196 m .30113 .64273 L p .27845 .65579 L .673 .96 .941 r
F P
0 g
s
.34676 .6023 m .32634 .62128 L .37651 .57795 L .317 .031 .311 r
F
.38349 .57223 m .35125 .60794 L p .38297 .57289 L 0 0 0 r
F P
0 g
s
.32897 .61869 m .37964 .57709 L .3067 .63835 L 0 0 0 r
F
.40741 .56821 m .43318 .566 L p .41681 .5416 L .674 .495 .635 r
F P
0 g
s
.41383 .53716 m .41362 .53684 L .40741 .56821 L .674 .495 .635 r
F
.44755 .50888 m .41092 .55325 L p .41043 .55363 L .642 .464 .627 r
F P
0 g
s
.44755 .50888 m .41043 .55363 L .41199 .54609 L .642 .464 .627 r
F
.37964 .57709 m .41552 .54763 L .37855 .57846 L 0 0 0 r
F
.37964 .57709 m .37855 .57846 L .30353 .64101 L F
.38623 .57003 m .41495 .53174 L p .41639 .52964 L .854 .768 .766 r
F P
0 g
s
.39896 .55957 m .38629 .56871 L p .40503 .54838 L .148 0 .202 r
F P
0 g
s
.42503 .52623 m .40503 .54838 L p .44147 .50825 L 0 0 0 r
F P
0 g
s
.40503 .54838 m .40949 .54354 L .44147 .50825 L 0 0 0 r
F
.40949 .54354 m .4117 .54115 L .44147 .50825 L F
.44147 .50825 m .4117 .54115 L .43596 .51483 L F
.41157 .53632 m .4115 .53645 L .4117 .5362 L .317 .031 .311 r
F
.41383 .53716 m .41399 .53638 L .41362 .53684 L .674 .495 .635 r
F
.4115 .53645 m .40776 .54335 L .41469 .53605 L .317 .031 .311 r
F
.42783 .52073 m .42269 .52764 L .44409 .50513 L F
.38549 .56928 m .4115 .53645 L .40503 .54838 L .148 0 .202 r
F
.38549 .56928 m .41268 .53427 L .4115 .53645 L F
.4056 .54775 m .44409 .50513 L p .38924 .56283 L .317 .031 .311 r
F P
0 g
s
.43242 .51157 m .41983 .52458 L p .38924 .56283 L .317 .031 .311 r
F P
0 g
s
.4115 .53645 m .48659 .44164 L .41662 .53581 L .148 0 .202 r
F
.40503 .54838 m .41662 .53581 L .39896 .55957 L F
.32255 .62381 m .27845 .65579 L .29172 .64794 L .063 0 0 r
F
.31442 .6297 m .34385 .60469 L p .31373 .63043 L 0 0 0 r
F P
0 g
s
.34385 .60469 m .38549 .56928 L p .3463 .60268 L 0 0 0 r
F P
0 g
s
.34385 .60469 m .3463 .60268 L .31373 .63043 L 0 0 0 r
F
.36308 .58894 m .38549 .56928 L .38321 .5714 L F
.3618 .58999 m .38549 .56928 L .36308 .58894 L F
.331 .61691 m .38629 .56871 L .37975 .57695 L F
.37767 .57595 m .38549 .56928 L .331 .61691 L F
.38924 .56283 m .37651 .57795 L .4056 .54775 L .317 .031 .311 r
F
.38629 .56871 m .38549 .56928 L p .40503 .54838 L .148 0 .202 r
F P
0 g
s
.38549 .56928 m .38629 .56871 L p .38321 .5714 L 0 0 0 r
F P
0 g
s
.38629 .56871 m .41552 .54763 L p .37975 .57695 L 0 0 0 r
F P
0 g
s
.38349 .57223 m .4056 .54775 L p .37651 .57795 L .317 .031 .311 r
F P
0 g
s
.4115 .53645 m .41662 .53581 L .40503 .54838 L .148 0 .202 r
F
.38349 .57223 m .38297 .57289 L .4117 .54115 L 0 0 0 r
F
.38349 .57223 m .4117 .54115 L .40949 .54354 L F
.40503 .54838 m .38349 .57223 L p .40949 .54354 L F P
0 g
s
.41552 .54763 m .39896 .55957 L p .41662 .53581 L .148 0 .202 r
F P
0 g
s
.44409 .50513 m .42503 .52623 L p .44147 .50825 L 0 0 0 r
F P
0 g
s
.44166 .50864 m .41552 .54763 L p .41662 .53581 L .148 0 .202 r
F P
0 g
s
.3733 .58507 m .38044 .57688 L p .41552 .54763 L 0 0 0 r
F P
0 g
s
.331 .61691 m .41552 .54763 L .32897 .61869 L 0 0 0 r
F
.3733 .58507 m .41552 .54763 L .42363 .53516 L p .81 1 .83 r
F P
0 g
s
.42363 .53516 m .43228 .52189 L p .42147 .53347 L .81 1 .83 r
F P
0 g
s
.3733 .58507 m .42363 .53516 L .42147 .53347 L .81 1 .83 r
F
.44409 .50513 m .4214 .5322 L .4381 .51367 L 0 0 0 r
F
.41552 .54763 m .44012 .51082 L .43504 .51798 L .611 .819 .414 r
F
.43228 .52189 m .42363 .53516 L p .43504 .51798 L F P
0 g
s
.42363 .53516 m .41552 .54763 L p .43504 .51798 L .611 .819 .414 r
F P
0 g
s
.4214 .5322 m .39631 .56216 L .42147 .53347 L 0 0 0 r
F
.44433 .50488 m .44409 .50513 L p .4381 .51367 L F P
0 g
s
.48659 .44164 m .44166 .50864 L p .41662 .53581 L .148 0 .202 r
F P
0 g
s
.46594 .46632 m .48334 .44406 L .46559 .4674 L .148 0 .202 r
F
.42301 .52077 m .44214 .49514 L .41438 .53114 L .854 .768 .766 r
F
.44214 .49514 m .4434 .49336 L .41438 .53114 L F
.42898 .51218 m .45112 .48279 L .46836 .46283 L .148 0 .202 r
F
.45112 .48279 m .46918 .46144 L .46836 .46283 L F
.46848 .4619 m .46551 .46765 L .46545 .47313 L 0 0 .107 r
F
.46211 .47816 m .46545 .47318 L p .46545 .47313 L .611 .819 .414 r
F P
0 g
s
.46545 .47318 m .48413 .44532 L p .46545 .47313 L .611 .819 .414 r
F P
0 g
s
.46545 .47313 m .48413 .44532 L .48208 .44814 L .611 .819 .414 r
F
.46947 .47163 m .4717 .4887 L .44755 .50888 L .706 .517 .631 r
F
.46947 .47163 m .44755 .50888 L .46944 .47137 L F
.46723 .47045 m .46545 .47313 L .46533 .48344 L 0 0 .107 r
F
.46848 .4619 m .46545 .47313 L .46723 .47045 L F
.45796 .49096 m .46533 .48344 L p .46551 .46765 L F P
0 g
s
.47643 .4492 m .48334 .44406 L p .46836 .46283 L .148 0 .202 r
F P
0 g
s
.44802 .50889 m .44755 .50888 L p .49606 .46834 L .706 .517 .631 r
F P
0 g
s
.47173 .48888 m .49214 .44805 L p .46919 .46948 L .674 .495 .635 r
F P
0 g
s
.50528 .39449 m .49873 .40908 L .5066 .39644 L 0 0 .107 r
F
.49798 .4146 m .5066 .39644 L p .49873 .40908 L F P
0 g
s
.50837 .39126 m .50528 .39449 L p .5066 .39644 L 0 0 .107 r
F P
0 g
s
.5066 .39644 m .49798 .4146 L p .50647 .39661 L .754 .844 .394 r
F P
0 g
s
.51228 .37159 m .51493 .36438 L p .51382 .36805 L 0 0 .051 r
F P
0 g
s
.51493 .36438 m .518 .35601 L p .5149 .36494 L 0 0 .051 r
F P
0 g
s
.51493 .36438 m .5149 .36494 L .51382 .36805 L 0 0 .051 r
F
.51287 .36751 m .518 .35601 L .51323 .36685 L .278 .05 .378 r
F
.51486 .36564 m .51269 .37589 L p .51483 .36627 L 0 0 .016 r
F P
0 g
s
.51269 .37589 m .51213 .37854 L p .51285 .37526 L 0 0 .016 r
F P
0 g
s
.51269 .37589 m .51285 .37526 L .51483 .36627 L 0 0 .016 r
F
.51484 .3657 m .51266 .376 L .51485 .36566 L .141 0 .344 r
F
.51266 .376 m .51213 .37851 L .51267 .37598 L F
.51266 .376 m .51267 .37598 L .51485 .36566 L F
.51213 .37854 m .51486 .36564 L p .51213 .37851 L F P
0 g
s
.51486 .36564 m .51734 .35876 L .51608 .36282 L 0 0 .051 r
F
.51436 .36681 m .51487 .36541 L .51486 .36564 L F
.51487 .36541 m .51734 .35876 L .51486 .36564 L F
.51734 .35876 m .51886 .35466 L .51602 .36304 L F
.51886 .35466 m .51702 .36055 L p .51602 .36304 L F P
0 g
s
.51294 .3749 m .51327 .37362 L .51481 .36651 L 0 0 .016 r
F
.51213 .37854 m .51481 .36651 L .51483 .36627 L F
.51213 .37851 m .51486 .36564 L .51485 .36566 L .141 0 .344 r
F
.50528 .39449 m .50974 .38982 L p .51484 .3657 L F P
0 g
s
.51268 .37132 m .51485 .36595 L .50512 .39848 L 0 0 .051 r
F
.51485 .36595 m .51636 .36218 L .51477 .36732 L F
.51819 .37047 m .51697 .36786 L .51618 .37486 L .737 .533 .616 r
F
.51697 .36786 m .5159 .36557 L .51618 .37486 L F
.51738 .36872 m .51576 .38295 L p .51368 .37329 L .701 .481 .587 r
F P
0 g
s
.51576 .38295 m .51434 .39537 L p .51368 .37329 L .701 .481 .587 r
F P
0 g
s
.51481 .36651 m .51147 .38186 L .5146 .37048 L 0 0 .016 r
F
.51485 .36595 m .51477 .36732 L .50512 .39848 L 0 0 .051 r
F
.5146 .37048 m .51147 .38186 L .51215 .3834 L 0 0 .016 r
F
.51702 .36055 m .51688 .36097 L p .51636 .36218 L 0 0 .051 r
F P
0 g
s
.51688 .36097 m .51475 .36779 L p .51636 .36218 L 0 0 .051 r
F P
0 g
s
.51475 .36779 m .50512 .39848 L p .51477 .36732 L 0 0 .051 r
F P
0 g
s
.51636 .36218 m .51475 .36779 L .51477 .36732 L 0 0 .051 r
F
.51567 .36417 m .51688 .36097 L .51694 .36076 L .863 .83 .356 r
F
.51567 .36417 m .51694 .36076 L .51645 .36164 L F
.51475 .36779 m .51688 .36097 L p .51567 .36417 L F P
0 g
s
.51268 .37132 m .51436 .36681 L .51602 .36304 L 0 0 .051 r
F
.51436 .36681 m .51608 .36282 L .51602 .36304 L F
.518 .35601 m .51886 .35466 L p .51436 .36681 L F P
0 g
s
.51619 .3621 m .51831 .35563 L .5167 .3601 L .863 .83 .356 r
F
.51567 .36417 m .51645 .36164 L .51619 .3621 L F
.51645 .36164 m .51831 .35563 L .51619 .3621 L F
.51709 .35672 m .51711 .35674 L p .52096 .33802 L .945 .734 .586 r
F P
0 g
s
.52096 .33802 m .51709 .35672 L p .52096 .33802 L .945 .734 .586 r
F P
0 g
s
.518 .35601 m .51709 .35672 L p .51922 .34632 L .336 .137 .454 r
F P
0 g
s
.51709 .35672 m .51848 .34998 L p .51859 .34938 L .336 .137 .454 r
F P
0 g
s
.51848 .34998 m .52106 .3375 L p .51859 .34938 L .336 .137 .454 r
F P
0 g
s
.51859 .34938 m .52106 .3375 L .52104 .33749 L .336 .137 .454 r
F
.51863 .35442 m .51696 .35909 L p .51831 .35563 L .863 .83 .356 r
F P
0 g
s
.51382 .36805 m .518 .35601 L .51436 .36681 L 0 0 .051 r
F
.51971 .34347 m .51885 .34802 L .51976 .34328 L 0 0 .016 r
F
.51885 .34802 m .51617 .36217 L .51882 .34821 L F
.51885 .34802 m .51882 .34821 L .51976 .34328 L F
.51886 .35466 m .52105 .3376 L .52104 .33769 L 0 0 0 r
F
.52106 .3375 m .51994 .34625 L p .52088 .33852 L 0 0 .286 r
F P
0 g
s
.51994 .34625 m .51886 .35466 L p .51995 .34598 L 0 0 .286 r
F P
0 g
s
.52088 .33852 m .51994 .34625 L .51995 .34598 L 0 0 .286 r
F
.51905 .35327 m .521 .33807 L .52099 .33815 L .987 .767 .373 r
F
.52106 .3375 m .51905 .35327 L p .52103 .33779 L 0 0 0 r
F P
0 g
s
.52105 .33756 m .52106 .3375 L .52103 .33779 L 0 0 0 r
F
.52106 .3375 m .52093 .33697 L p .52097 .33711 L .464 .275 .539 r
F P
0 g
s
.52106 .33762 m .52093 .33697 L .52106 .33764 L .61 .364 .52 r
F
.52102 .33781 m .52106 .3375 L .52095 .33701 L .687 .415 .505 r
F
.52106 .3375 m .52106 .3375 L .52106 .3375 L .945 .734 .586 r
F
.52106 .3375 m .52247 .35751 L p .52127 .34013 L .886 .627 .54 r
F P
0 g
s
.52109 .33747 m .52106 .3375 L .52127 .34013 L .886 .627 .54 r
F
.51922 .34632 m .52105 .33745 L .52096 .33799 L .336 .137 .454 r
F
.52314 .35796 m .52323 .35802 L p .5222 .35356 L .886 .627 .54 r
F P
0 g
s
.52103 .33784 m .52106 .3375 L .52102 .33781 L .687 .415 .505 r
F
.52103 .33783 m .52106 .33752 L .52096 .33825 L .987 .767 .373 r
F
.52107 .33757 m .52106 .33752 L .52103 .33783 L F
.52105 .33745 m .52105 .33746 L .52096 .33799 L .336 .137 .454 r
F
.518 .35601 m .52106 .3375 L p .52083 .33879 L 0 0 .286 r
F P
0 g
s
.51886 .35466 m .518 .35601 L p .52083 .33879 L 0 0 .286 r
F P
0 g
s
.52083 .33879 m .52088 .33852 L .51886 .35466 L 0 0 .286 r
F
.52106 .3375 m .518 .35601 L p .52105 .33742 L .336 .137 .454 r
F P
0 g
s
.521 .33754 m .52106 .33754 L .52104 .33773 L .687 .415 .505 r
F
.52106 .33752 m .52106 .33752 L .52103 .33783 L .987 .767 .373 r
F
.52099 .33815 m .52106 .33764 L .52106 .33764 L F
.521 .33807 m .52115 .33693 L .52099 .33815 L F
.52323 .35802 m .52312 .35778 L p .52134 .34029 L .977 .7 .306 r
F P
0 g
s
.52126 .33947 m .52111 .33794 L .52064 .35686 L .808 .565 .571 r
F
.52112 .33739 m .52111 .33741 L .5211 .33789 L F
.52111 .33794 m .5211 .33789 L .52064 .35686 L F
.52106 .33737 m .52107 .33735 L .52106 .33737 L .336 .137 .454 r
F
.52106 .33737 m .52106 .33737 L .52105 .33746 L F
.521 .33754 m .52105 .33752 L .52104 .33754 L .687 .415 .505 r
F
.521 .33754 m .52105 .33751 L .52105 .33752 L F
.52095 .33701 m .52106 .3375 L .52106 .33754 L F
.521 .33754 m .52095 .33701 L .52105 .33751 L F
.52094 .33688 m .52106 .3375 L p .52095 .33701 L F P
0 g
s
.52106 .3375 m .52097 .33711 L .52102 .33727 L .464 .275 .539 r
F
.521 .33715 m .521 .33712 L .52106 .3375 L .769 .477 .491 r
F
.52107 .33735 m .52106 .3375 L .52105 .33742 L .336 .137 .454 r
F
.52104 .33754 m .52105 .33743 L .52098 .33765 L .769 .477 .491 r
F
.52105 .33743 m .52109 .3369 L .52103 .33732 L F
.52105 .33743 m .52103 .33732 L .52098 .33765 L F
.52094 .33691 m .52106 .3375 L .52105 .33745 L .464 .275 .539 r
F
.52107 .33745 m .52097 .33711 L .52094 .33691 L F
.52094 .33691 m .52106 .3375 L p .52094 .33691 L F P
0 g
s
.52105 .33745 m .52106 .33737 L .52105 .33746 L .336 .137 .454 r
F
.52107 .33746 m .52106 .3375 L .52106 .33748 L .918 .624 .472 r
F
.52106 .3375 m .52106 .33738 L .52106 .33739 L .336 .137 .454 r
F
.52103 .33732 m .52109 .3369 L .521 .33715 L .769 .477 .491 r
F
.52106 .33748 m .52106 .33748 L .52106 .33749 L .918 .624 .472 r
F
.52106 .33744 m .52107 .33742 L .52106 .33749 L F
.52107 .33734 m .52106 .3375 L .52107 .33735 L .336 .137 .454 r
F
.52107 .33742 m .52106 .33748 L .52106 .33749 L .918 .624 .472 r
F
.52104 .3373 m .52106 .3375 L .52106 .33714 L .853 .553 .481 r
F
.52106 .33746 m .52106 .33744 L .52106 .33746 L .918 .624 .472 r
F
.52105 .33737 m .52094 .33691 L .52106 .33749 L .464 .275 .539 r
F
.52101 .33688 m .52106 .3375 L p .521 .33712 L .769 .477 .491 r
F P
0 g
s
.52106 .33753 m .52106 .33758 L .52106 .33762 L .987 .767 .373 r
F
.52106 .33754 m .52106 .33762 L .52104 .33775 L F
.52106 .33753 m .52106 .33762 L .52106 .33754 L F
.52106 .33757 m .52106 .33754 L .52104 .33775 L F
.52106 .33754 m .52106 .33754 L .52104 .33775 L F
.52104 .33773 m .52106 .33754 L .52106 .33757 L .687 .415 .505 r
F
.52104 .33754 m .52105 .33752 L .52106 .33754 L F
.52106 .33752 m .52106 .33751 L .52106 .33752 L .987 .767 .373 r
F
.52106 .33751 m .52106 .33751 L .52106 .33752 L F
.52106 .33752 m .52106 .33751 L .52106 .33752 L F
.52106 .3375 m .52105 .33739 L .52092 .3392 L .769 .477 .491 r
F
.52105 .33745 m .52106 .3375 L .52106 .33749 L .464 .275 .539 r
F
.52098 .33765 m .521 .33715 L .52106 .3375 L .769 .477 .491 r
F
.52106 .3375 m .52102 .33727 L .52104 .33735 L .464 .275 .539 r
F
.521 .33756 m .52104 .3373 L .52106 .3375 L .853 .553 .481 r
F
.52104 .3373 m .52104 .33724 L .52106 .3375 L F
.52107 .33735 m .52107 .33734 L .52107 .33735 L .336 .137 .454 r
F
.52107 .33734 m .5211 .3372 L .52106 .3375 L F
.52104 .33769 m .52114 .33688 L .52112 .33705 L 0 0 0 r
F
.52103 .33784 m .521 .33754 L .52104 .33773 L .687 .415 .505 r
F
.52095 .33833 m .52098 .33765 L .52106 .3375 L .769 .477 .491 r
F
.52106 .33752 m .52106 .3375 L .52105 .33759 L .987 .767 .373 r
F
.52106 .33746 m .52106 .3375 L .52112 .33703 L .918 .624 .472 r
F
.52101 .33753 m .52106 .3375 L .52104 .3373 L .853 .553 .481 r
F
.52106 .33749 m .52106 .3375 L .52106 .3375 L .918 .624 .472 r
F
.52106 .33714 m .52106 .3375 L .52106 .33708 L .853 .553 .481 r
F
.52106 .33748 m .52106 .3375 L .52106 .33749 L .918 .624 .472 r
F
.52439 .34869 m .52545 .35264 L .52434 .34858 L .415 .257 .557 r
F
.52545 .35264 m .52841 .36374 L .52541 .35256 L F
.52545 .35264 m .52541 .35256 L .52434 .34858 L F
.62409 .49133 m .56835 .443 L .59268 .48912 L .436 .359 .671 r
F
.62068 .47839 m .56835 .443 L .62409 .49133 L F
.62945 .65167 m .59934 .6098 L .61681 .65496 L .157 .321 .779 r
F
.6252 .63682 m .59934 .6098 L .62945 .65167 L F
.53824 .43191 m .53846 .42972 L .54811 .47208 L .904 .666 .569 r
F
.55645 .49537 m .54555 .46313 L .55587 .4943 L .869 .664 .623 r
F
.54811 .47208 m .55376 .49036 L .55161 .48634 L .904 .666 .569 r
F
.5628 .53193 m .55172 .48794 L p .54811 .47208 L F P
0 g
s
.55577 .49411 m .54555 .46313 L p .54536 .46314 L .869 .664 .623 r
F P
0 g
s
.55577 .49411 m .54536 .46314 L .54982 .48301 L .869 .664 .623 r
F
.57554 .55401 m .54555 .46313 L .57594 .55487 L .889 .557 .061 r
F
.53699 .42394 m .53125 .40289 L .5342 .42143 L .88 .643 .575 r
F
.53699 .42394 m .53146 .41896 L .53702 .4314 L F
.53699 .42394 m .52584 .37199 L .52577 .3724 L F
.53699 .42394 m .52577 .3724 L .52582 .37281 L F
.53742 .42551 m .52584 .37199 L .53699 .42394 L F
.54555 .46313 m .53742 .42551 L .53699 .42394 L F
.5307 .40122 m .53846 .42972 L .53824 .43191 L .904 .666 .569 r
F
.5307 .40122 m .53794 .42747 L p .53846 .42972 L F P
0 g
s
.54555 .46313 m .53846 .42972 L p .53842 .43013 L .88 .643 .575 r
F P
0 g
s
.53788 .42724 m .53794 .42747 L .54555 .46313 L .955 .657 .223 r
F
.53846 .42972 m .54555 .46313 L p .53794 .42747 L F P
0 g
s
.54078 .44455 m .53806 .43373 L p .54043 .4428 L .904 .666 .569 r
F P
0 g
s
.53831 .4322 m .54043 .4428 L .53806 .43373 L .904 .666 .569 r
F
.54555 .46313 m .53839 .43037 L .53763 .43805 L .88 .643 .575 r
F
.54555 .46313 m .53703 .43617 L .53705 .44129 L F
.53712 .45767 m .5445 .46245 L p .53705 .44129 L F P
0 g
s
.5445 .46245 m .54555 .46313 L p .53705 .44129 L .88 .643 .575 r
F P
0 g
s
.54093 .44408 m .53633 .42183 L .53641 .42213 L .955 .657 .223 r
F
.54432 .46059 m .54085 .44376 L p .54093 .44408 L F P
0 g
s
.54085 .44376 m .53633 .42183 L p .54093 .44408 L .955 .657 .223 r
F P
0 g
s
.54432 .46059 m .53641 .42213 L .54555 .46313 L .955 .657 .223 r
F
.53641 .42213 m .53718 .42479 L .54555 .46313 L F
.53795 .42843 m .53679 .4237 L .54062 .44245 L 0 0 .299 r
F
.5396 .43678 m .54062 .44245 L .54432 .46059 L F
.53795 .42843 m .54062 .44245 L .5396 .43678 L F
.53384 .41315 m .53602 .42151 L .53573 .42001 L .773 .477 0 r
F
.5335 .40797 m .53767 .42629 L .53788 .42724 L .955 .657 .223 r
F
.53718 .42479 m .53788 .42724 L .54555 .46313 L F
.53794 .42747 m .53583 .41981 L p .53764 .4264 L .773 .477 0 r
F P
0 g
s
.54126 .43961 m .53583 .41981 L .53597 .42038 L .773 .477 0 r
F
.54605 .45721 m .54126 .43961 L .53597 .42038 L F
.5335 .40797 m .53788 .42724 L .53718 .42479 L .955 .657 .223 r
F
.53849 .42944 m .53794 .42747 L p .54126 .43961 L .773 .477 0 r
F P
0 g
s
.54126 .43961 m .53794 .42747 L .53764 .4264 L .773 .477 0 r
F
.53583 .41981 m .5307 .40122 L p .53597 .42038 L F P
0 g
s
.53703 .42469 m .53794 .4284 L .54812 .46479 L .773 .477 0 r
F
.54605 .45721 m .54812 .46479 L .55184 .4781 L F
.53679 .4237 m .53659 .42292 L .54072 .44304 L 0 0 .299 r
F
.54062 .44245 m .54072 .44304 L .54432 .46059 L F
.53679 .4237 m .54072 .44304 L .54062 .44245 L F
.54397 .45689 m .54336 .45046 L p .5382 .42944 L F P
0 g
s
.54555 .46313 m .55645 .49537 L p .57594 .55487 L .889 .557 .061 r
F P
0 g
s
.55781 .50029 m .55037 .47838 L .54555 .46313 L .889 .557 .061 r
F
.54982 .48301 m .54536 .46314 L .53931 .46341 L .869 .664 .623 r
F
.54555 .46313 m .53712 .45767 L p .53931 .46341 L F P
0 g
s
.56653 .52732 m .54432 .46059 L p .55781 .50029 L .889 .557 .061 r
F P
0 g
s
.55092 .47656 m .55076 .47617 L .55271 .4858 L 0 .005 .571 r
F
.55271 .4858 m .55953 .50628 L p .55772 .50004 L F P
0 g
s
.55092 .47656 m .55271 .4858 L .55772 .50004 L 0 .005 .571 r
F
.54432 .46059 m .54397 .45689 L p .53795 .42843 L 0 0 .299 r
F P
0 g
s
.53794 .4284 m .5513 .47786 L .55184 .4781 L .773 .477 0 r
F
.54432 .46059 m .54555 .46313 L p .55037 .47838 L .889 .557 .061 r
F P
0 g
s
.54336 .45046 m .54432 .46059 L p .55092 .47656 L 0 .005 .571 r
F P
0 g
s
.53855 .43086 m .5632 .52195 L p .53794 .4284 L .773 .477 0 r
F P
0 g
s
.55085 .47633 m .53855 .43086 L p .5515 .47742 L .009 .089 .6 r
F P
0 g
s
.5515 .47742 m .53855 .43086 L .53844 .43027 L .009 .089 .6 r
F
.54501 .45344 m .54626 .45796 L .54605 .45721 L .773 .477 0 r
F
.52663 .3859 m .53146 .41896 L p .5342 .42143 L .88 .643 .575 r
F P
0 g
s
.53699 .42394 m .52651 .37599 L .53056 .40033 L .88 .643 .575 r
F
.52651 .37599 m .52538 .37082 L .53056 .40033 L F
.53056 .40033 m .52538 .37082 L .52663 .3859 L F
.5307 .40122 m .52966 .40039 L p .52585 .372 L .911 .651 .532 r
F P
0 g
s
.53633 .42183 m .53723 .42499 L .53597 .42038 L .773 .477 0 r
F
.53641 .42213 m .52763 .37965 L .52704 .37656 L .955 .657 .223 r
F
.53633 .42183 m .52763 .37965 L p .53641 .42213 L F P
0 g
s
.53316 .40647 m .53633 .42183 L p .53063 .39383 L 0 0 .299 r
F P
0 g
s
.53633 .42183 m .54432 .46059 L p .53659 .42292 L 0 0 .299 r
F P
0 g
s
.53641 .42213 m .5335 .40797 L .53718 .42479 L .955 .657 .223 r
F
.52675 .37461 m .52661 .37391 L p .52771 .38061 L .88 .643 .575 r
F P
0 g
s
.5335 .40797 m .52704 .37656 L .53767 .42629 L .955 .657 .223 r
F
.53063 .39383 m .53633 .42183 L .53659 .42292 L 0 0 .299 r
F
.53692 .42425 m .54626 .45796 L .54501 .45344 L .773 .477 0 r
F
.52966 .40039 m .5307 .40122 L p .53824 .43191 L .904 .666 .569 r
F P
0 g
s
.5297 .39817 m .53693 .4243 L .53692 .42425 L .773 .477 0 r
F
.53693 .4243 m .54626 .45796 L .53692 .42425 L F
.5297 .39817 m .54812 .46479 L .54626 .45796 L F
.5569 .49622 m .53849 .42944 L p .55184 .4781 L F P
0 g
s
.55184 .4781 m .53849 .42944 L .54126 .43961 L .773 .477 0 r
F
.52817 .38185 m .52676 .37471 L .53808 .42794 L .955 .657 .223 r
F
.52661 .37391 m .52675 .37461 L p .52676 .37471 L F P
0 g
s
.52675 .37461 m .53808 .42794 L p .52676 .37471 L .955 .657 .223 r
F P
0 g
s
.52667 .37407 m .5307 .40122 L p .52585 .372 L .911 .651 .532 r
F P
0 g
s
.52677 .37474 m .52702 .37674 L .52724 .37791 L .904 .636 .132 r
F
.52681 .37528 m .52696 .37645 L .52677 .37474 L F
.52696 .37645 m .5297 .39817 L .52702 .37674 L F
.52696 .37645 m .52702 .37674 L .52677 .37474 L F
.52123 .33785 m .52514 .37576 L .52118 .33799 L .844 .604 .575 r
F
.52126 .33777 m .52585 .372 L .52125 .33777 L .911 .651 .532 r
F
.52584 .37199 m .52125 .33777 L p .52585 .372 L F P
0 g
s
.52661 .37391 m .52323 .35802 L p .52538 .37082 L .88 .643 .575 r
F P
0 g
s
.52684 .37557 m .52537 .36854 L .52529 .36805 L .955 .657 .223 r
F
.52356 .36043 m .52514 .37576 L .52356 .36041 L .844 .604 .575 r
F
.52323 .35802 m .52247 .35751 L p .52538 .37082 L .88 .643 .575 r
F P
0 g
s
.52684 .3756 m .52312 .35778 L .52684 .37557 L .955 .657 .223 r
F
.52537 .36854 m .52312 .35778 L .52529 .36805 L F
.5307 .40122 m .52676 .37471 L p .52677 .37474 L .904 .636 .132 r
F P
0 g
s
.52737 .38148 m .5297 .39817 L p .52681 .37528 L .904 .636 .132 r
F P
0 g
s
.52323 .35802 m .52601 .3711 L p .52704 .37656 L .955 .657 .223 r
F P
0 g
s
.52554 .36893 m .52374 .36081 L .52568 .36991 L 0 0 .299 r
F
.52582 .3704 m .52568 .36991 L .5261 .37187 L F
.52554 .36893 m .52568 .36991 L .52582 .3704 L F
.5261 .37187 m .52568 .36991 L .52688 .37602 L F
.52763 .37965 m .52312 .35778 L p .52704 .37656 L .955 .657 .223 r
F P
0 g
s
.52374 .36081 m .52546 .36915 L p .52568 .36991 L 0 0 .299 r
F P
0 g
s
.52546 .36915 m .52688 .37602 L p .52568 .36991 L 0 0 .299 r
F P
0 g
s
.52312 .35778 m .52323 .35802 L p .52529 .36805 L .955 .657 .223 r
F P
0 g
s
.52312 .35778 m .52374 .36081 L p .52554 .36893 L 0 0 .299 r
F P
0 g
s
.5239 .3567 m .52582 .3704 L p .52664 .37388 L .904 .636 .132 r
F P
0 g
s
.52601 .3711 m .52661 .37391 L p .52704 .37656 L .955 .657 .223 r
F P
0 g
s
.52582 .3704 m .52677 .37541 L .52664 .37388 L .904 .636 .132 r
F
.52713 .3798 m .52732 .38095 L .52664 .37388 L F
.52582 .3704 m .5239 .3567 L p .52582 .37041 L 0 0 .394 r
F P
0 g
s
.52398 .35719 m .52583 .37044 L .5239 .3567 L 0 0 .394 r
F
.52583 .37044 m .52582 .37041 L .5239 .3567 L F
.52679 .37505 m .52661 .37391 L .52676 .37471 L .955 .657 .223 r
F
.52737 .38148 m .52681 .37528 L .52664 .37388 L .904 .636 .132 r
F
.52817 .38185 m .53808 .42794 L .53788 .42724 L .955 .657 .223 r
F
.53692 .42425 m .54605 .45721 L .53723 .42499 L .773 .477 0 r
F
.54605 .45721 m .55184 .4781 L .54126 .43961 L F
.53704 .42244 m .5515 .47742 L .53844 .43027 L .009 .089 .6 r
F
.52114 .33691 m .5239 .3567 L p .52664 .37388 L .904 .636 .132 r
F P
0 g
s
.52385 .35623 m .52467 .36062 L .5255 .36624 L 0 0 .394 r
F
.52667 .37407 m .52115 .33693 L p .52664 .37388 L .904 .636 .132 r
F P
0 g
s
.52467 .36062 m .52649 .37043 L .5255 .36624 L 0 0 .394 r
F
.5255 .36624 m .52649 .37043 L .5267 .37351 L F
.52125 .33768 m .52664 .37388 L .52127 .33776 L .904 .636 .132 r
F
.53561 .41444 m .52621 .37178 L .52603 .37116 L 0 0 .299 r
F
.53554 .41403 m .52621 .37178 L .53561 .41444 L F
.5513 .47786 m .55586 .49361 L .55184 .4781 L .773 .477 0 r
F
.59137 .60172 m .59112 .60122 L p .55645 .49537 L .889 .557 .061 r
F P
0 g
s
.55184 .4781 m .55828 .50137 L .5569 .49622 L .773 .477 0 r
F
.55184 .4781 m .56289 .52074 L .56419 .52471 L F
.55866 .50285 m .55873 .50313 L .55781 .50029 L .889 .557 .061 r
F
.55184 .4781 m .56419 .52471 L .55828 .50137 L .773 .477 0 r
F
.57426 .53305 m .55806 .48856 L .55092 .47656 L 0 .005 .571 r
F
.57232 .52552 m .55806 .48856 L .57426 .53305 L F
.63001 .61982 m .64563 .6825 L .6387 .68408 L .434 .524 .851 r
F
.63242 .61923 m .64563 .6825 L .63001 .61982 L F
.57783 .52695 m .6014 .60385 L .58374 .5697 L .235 .32 .747 r
F
.65557 .59167 m .60862 .57495 L .62098 .62201 L .424 .463 .797 r
F
.61561 .61303 m .62098 .62201 L p .60862 .57495 L F P
0 g
s
.65512 .58086 m .6638 .60875 L .65608 .60407 L .522 .544 .819 r
F
.65608 .61348 m .65645 .61277 L p .62098 .62201 L .424 .463 .797 r
F P
0 g
s
.65557 .59167 m .62098 .62201 L .65645 .61277 L .424 .463 .797 r
F
.65604 .61299 m .62536 .58544 L p .6638 .60875 L .522 .544 .819 r
F P
0 g
s
.58326 .57468 m .5946 .60823 L p .57975 .56502 L .889 .557 .061 r
F P
0 g
s
.6252 .63682 m .61561 .61303 L .59718 .60414 L .157 .321 .779 r
F
.57539 .55381 m .56653 .52732 L .56412 .51986 L .889 .557 .061 r
F
.57507 .553 m .56653 .52732 L p .57539 .55381 L F P
0 g
s
.59112 .60122 m .56412 .51986 L .56684 .52765 L .889 .557 .061 r
F
.56653 .52732 m .57507 .553 L p .57639 .55585 L 0 .005 .571 r
F P
0 g
s
.57878 .56101 m .61407 .64846 L .60481 .62628 L 0 0 .414 r
F
.57783 .52695 m .58374 .5697 L .57193 .52403 L .235 .32 .747 r
F
.58144 .56079 m .56653 .52732 L .57639 .55585 L 0 .005 .571 r
F
.58864 .56784 m .57842 .5491 L .58144 .56079 L F
.57842 .5491 m .56653 .52732 L .58144 .56079 L F
.6252 .63682 m .59718 .60414 L .59934 .6098 L .157 .321 .779 r
F
.57974 .56307 m .61407 .64846 L .57878 .56101 L 0 0 .414 r
F
.58053 .56478 m .59566 .60014 L .6091 .63542 L F
.61681 .65496 m .62512 .67234 L p .6255 .67164 L .157 .321 .779 r
F P
0 g
s
.61904 .61524 m .62536 .58544 L p .6387 .68408 L .434 .524 .851 r
F P
0 g
s
.63743 .64959 m .61681 .65496 L .6255 .67164 L .157 .321 .779 r
F
.61407 .64846 m .61873 .65924 L .6161 .6535 L 0 0 .414 r
F
.58144 .56079 m .57639 .55585 L .58493 .5743 L 0 .005 .571 r
F
.5632 .52195 m .6091 .63542 L .61407 .64846 L 0 0 .414 r
F
.6091 .63542 m .61873 .65924 L .61407 .64846 L F
.59282 .5962 m .60481 .62628 L .57911 .56338 L F
.60951 .63859 m .58331 .57397 L p .59832 .61048 L F P
0 g
s
.59081 .59701 m .59112 .60122 L p .59718 .60414 L .157 .321 .779 r
F P
0 g
s
.61561 .61303 m .59566 .60014 L .59718 .60414 L .157 .321 .779 r
F
.59566 .60014 m .59081 .59701 L .59718 .60414 L F
.58374 .5697 m .59623 .60051 L .58493 .5743 L .235 .32 .747 r
F
.61325 .64252 m .59566 .60014 L p .59623 .60051 L F P
0 g
s
.59566 .60014 m .58493 .5743 L p .59623 .60051 L .235 .32 .747 r
F P
0 g
s
.6014 .60385 m .59623 .60051 L .58374 .5697 L .235 .32 .747 r
F
.58992 .58507 m .58864 .56784 L p .58144 .56079 L 0 .005 .571 r
F P
0 g
s
.58144 .56079 m .58493 .5743 L .58992 .58507 L 0 .005 .571 r
F
.5632 .52195 m .58493 .5743 L p .58053 .56478 L 0 0 .414 r
F P
0 g
s
.58493 .5743 m .59566 .60014 L p .58053 .56478 L 0 0 .414 r
F P
0 g
s
.56684 .52765 m .56412 .51986 L .55873 .50313 L .889 .557 .061 r
F
.56684 .52765 m .55873 .50313 L .55866 .50285 L F
.58864 .56784 m .55092 .47656 L .56653 .52732 L 0 .005 .571 r
F
.55783 .49956 m .5569 .49622 L p .55828 .50137 L .773 .477 0 r
F P
0 g
s
.55953 .50628 m .56653 .52732 L p .55092 .47656 L 0 .005 .571 r
F P
0 g
s
.5632 .52195 m .56355 .51987 L p .58374 .5697 L .235 .32 .747 r
F P
0 g
s
.58493 .5743 m .5632 .52195 L p .58374 .5697 L .235 .32 .747 r
F P
0 g
s
.5632 .52195 m .57974 .56307 L .57878 .56101 L 0 0 .414 r
F
.57523 .55334 m .59282 .5962 L .577 .55795 L F
.5632 .52195 m .57878 .56101 L .57523 .55334 L F
.57878 .56101 m .59282 .5962 L .57523 .55334 L F
.58116 .56867 m .58326 .57468 L .57975 .56502 L .889 .557 .061 r
F
.59086 .59768 m .58992 .58507 L p .57639 .55585 L 0 .005 .571 r
F P
0 g
s
.59934 .6098 m .59718 .60414 L .59112 .60122 L .157 .321 .779 r
F
.61407 .64846 m .6161 .6535 L .61478 .65032 L 0 0 .414 r
F
.61407 .64846 m .61478 .65032 L .60481 .62628 L F
.71774 .52633 m .64343 .51939 L .6642 .55233 L .504 .467 .749 r
F
.71613 .5141 m .64343 .51939 L .71774 .52633 L F
.65086 .4664 m .64319 .46251 L .63729 .46254 L .505 .35 .604 r
F
.65086 .4664 m .63729 .46254 L .64776 .46606 L F
.7522 .49667 m .69403 .49101 L .71473 .50345 L .536 .412 .659 r
F
.77727 .49214 m .75886 .49134 L .75561 .49394 L F
.77727 .49214 m .75561 .49394 L .77018 .49342 L F
.75886 .49134 m .73181 .49095 L .74395 .49436 L F
.75561 .49394 m .74395 .49436 L .7522 .49667 L F
.75886 .49134 m .74395 .49436 L .75561 .49394 L F
.77018 .49342 m .75561 .49394 L .7522 .49667 L F
.73181 .49095 m .71813 .49076 L .7522 .49667 L F
.70911 .48919 m .71987 .49353 L .7522 .49667 L F
.70293 .58737 m .67546 .61346 L .67082 .62881 L .551 .618 .873 r
F
.67546 .61346 m .66873 .61986 L .66852 .62087 L F
.67546 .61346 m .66852 .62087 L .66022 .6605 L F
.66045 .66305 m .66022 .6605 L .658 .67113 L F
.67546 .61346 m .66022 .6605 L .66045 .66305 L F
.63955 .68547 m .6387 .68408 L p .64563 .6825 L .434 .524 .851 r
F P
0 g
s
.62519 .67221 m .63759 .69796 L .62519 .67221 L .68 .309 0 r
F
.63759 .69796 m .64555 .71458 L p .62519 .67221 L F P
0 g
s
.63098 .68434 m .62519 .6722 L .63347 .68944 L .68 .309 0 r
F
.6391 .68442 m .63743 .64959 L p .62512 .67234 L .157 .321 .779 r
F P
0 g
s
.66157 .73861 m .66121 .73764 L .65833 .73353 L .005 .313 .804 r
F
.65815 .73349 m .66157 .73861 L p .65833 .73353 L F P
0 g
s
.61325 .64252 m .63955 .68547 L p .65308 .71856 L .026 .284 .78 r
F P
0 g
s
.61325 .64252 m .65308 .71856 L .65358 .72245 L .026 .284 .78 r
F
.61325 .64252 m .65929 .73376 L .65876 .73363 L F
.65819 .67638 m .65442 .68199 L .65269 .71559 L .455 .585 .894 r
F
.65442 .68199 m .64792 .70465 L .65269 .71559 L F
.66121 .73764 m .6598 .73388 L .65833 .73353 L .005 .313 .804 r
F
.63955 .68547 m .65586 .72286 L .66141 .73894 L .026 .284 .78 r
F
.65358 .72245 m .65308 .71856 L .65929 .73376 L F
.65442 .68199 m .64449 .69678 L .64792 .70465 L .455 .585 .894 r
F
.66018 .73276 m .65819 .67638 L p .65324 .71166 L F P
0 g
s
.65324 .71166 m .65269 .71559 L .65749 .72659 L .455 .585 .894 r
F
.65324 .71166 m .65749 .72659 L .66018 .73276 L F
.65537 .71991 m .66141 .73894 L p .64738 .69825 L .026 .284 .78 r
F P
0 g
s
.65876 .73363 m .63465 .69186 L .63394 .69041 L .026 .284 .78 r
F
.64555 .71458 m .64239 .70825 L p .63098 .68434 L .68 .309 0 r
F P
0 g
s
.63098 .68434 m .63347 .68944 L .64555 .71458 L .68 .309 0 r
F
.61325 .64252 m .65876 .73363 L .63394 .69041 L .026 .284 .78 r
F
.63783 .69746 m .64067 .70246 L p .64035 .70294 L .867 .715 .691 r
F P
0 g
s
.65507 .72886 m .65815 .73349 L p .65833 .73353 L .005 .313 .804 r
F P
0 g
s
.64067 .70246 m .66141 .73894 L p .64035 .70294 L .867 .715 .691 r
F P
0 g
s
.64035 .70294 m .66141 .73894 L .64811 .71983 L .867 .715 .691 r
F
.63783 .69746 m .64811 .71983 L .63759 .69796 L F
.64449 .69678 m .64 .70345 L .65859 .73044 L .455 .585 .894 r
F
.6598 .73388 m .65749 .72659 L p .65881 .73121 L .934 .695 .216 r
F P
0 g
s
.65749 .72659 m .65859 .73044 L .66018 .73276 L .455 .585 .894 r
F
.64449 .69678 m .65859 .73044 L .65749 .72659 L F
.6602 .73327 m .66018 .73276 L p .65859 .73044 L F P
0 g
s
.64 .70345 m .65881 .73121 L p .65859 .73044 L .455 .585 .894 r
F P
0 g
s
.65881 .73121 m .6602 .73327 L p .65859 .73044 L .455 .585 .894 r
F P
0 g
s
.65876 .73363 m .65929 .73376 L .66141 .73894 L .026 .284 .78 r
F
.65833 .73353 m .6602 .73327 L .65507 .72886 L .005 .313 .804 r
F
.6606 .73407 m .6602 .73327 L p .65833 .73353 L F P
0 g
s
.66141 .73894 m .65833 .73353 L p .65876 .73363 L .026 .284 .78 r
F P
0 g
s
.66141 .73894 m .64702 .72898 L p .64134 .71546 L .867 .715 .691 r
F P
0 g
s
.65297 .72681 m .66141 .73894 L .64134 .71546 L .867 .715 .691 r
F
.60388 .69051 m .61769 .71759 L p .64266 .72488 L .802 .727 .782 r
F P
0 g
s
.6382 .71239 m .64702 .72898 L .62478 .71966 L p .791 .695 .758 r
F P
0 g
s
.64064 .72159 m .65886 .73247 L p .6599 .73623 L .9 .845 .781 r
F P
0 g
s
.64711 .73133 m .6615 .74128 L p .65486 .73837 L .819 .337 .015 r
F P
0 g
s
.64711 .73133 m .64702 .72898 L .66141 .73894 L p .91 .636 .51 r
F P
0 g
s
.66141 .73894 m .6615 .74128 L .64711 .73133 L p .91 .636 .51 r
F P
0 g
s
.63615 .72675 m .62485 .72202 L p .62108 .71927 L .843 .579 .54 r
F P
0 g
s
.66165 .74295 m .62249 .7209 L .62594 .72261 L 0 .394 .79 r
F
.66391 .74453 m .62249 .7209 L p .66165 .74295 L F P
0 g
s
.6069 .71053 m .61261 .71677 L .58033 .70386 L .294 .7 .974 r
F
.6181 .7184 m .6177 .71812 L p .61972 .71922 L 0 .394 .79 r
F P
0 g
s
.63876 .72858 m .64711 .73133 L p .65258 .7363 L .819 .337 .015 r
F P
0 g
s
.65486 .73837 m .65059 .73578 L .65258 .7363 L .819 .337 .015 r
F
.65059 .73578 m .63876 .72858 L .65258 .7363 L F
.62249 .7209 m .63048 .72467 L .64621 .73264 L 0 .394 .79 r
F
.62485 .72202 m .60595 .71411 L p .6177 .71812 L .843 .579 .54 r
F P
0 g
s
.62108 .71927 m .62485 .72202 L .6177 .71812 L .843 .579 .54 r
F
.65152 .73656 m .62594 .72261 L .64209 .7306 L 0 .394 .79 r
F
.66165 .74295 m .62594 .72261 L .65152 .73656 L F
.49406 .68198 m .48371 .681 L p .47761 .67986 L .347 .765 .978 r
F P
0 g
s
.49411 .68192 m .49406 .68198 L p .47761 .67986 L .347 .765 .978 r
F P
0 g
s
.47665 .67968 m .47761 .67986 L .4205 .67249 L .347 .765 .978 r
F
.49411 .68192 m .47761 .67986 L .47665 .67968 L F
.4205 .67249 m .47761 .67986 L .42033 .67248 L F
.47089 .67861 m .42883 .67076 L .42256 .66948 L .15 0 0 r
F
.42844 .67069 m .42823 .67064 L p .42883 .67076 L F P
0 g
s
.42823 .67064 m .42256 .66948 L p .42883 .67076 L .15 0 0 r
F P
0 g
s
.36128 .66869 m .42061 .66908 L p .38287 .66933 L .246 0 0 r
F P
0 g
s
.42061 .66908 m .40786 .67007 L .38287 .66933 L .246 0 0 r
F
.42069 .66668 m .36128 .66869 L .34909 .66791 L .696 .455 .554 r
F
.42883 .67076 m .42033 .67248 L .4027 .66992 L .347 .765 .978 r
F
.42061 .66908 m .42823 .67064 L p .40786 .67007 L .246 0 0 r
F P
0 g
s
.42823 .67064 m .42844 .67069 L p .42791 .67067 L .246 0 0 r
F P
0 g
s
.42823 .67064 m .42791 .67067 L .40786 .67007 L .246 0 0 r
F
.42791 .67067 m .42883 .67076 L p .4027 .66992 L .347 .765 .978 r
F P
0 g
s
.44561 .67406 m .47132 .679 L .4205 .67249 L .347 .765 .978 r
F
.47089 .67861 m .42256 .66948 L .47131 .67868 L .15 0 0 r
F
.47131 .67868 m .42256 .66948 L .47665 .67968 L F
.47345 .67908 m .42061 .66908 L p .49206 .6814 L .768 .516 .554 r
F P
0 g
s
.48431 .6792 m .47669 .67729 L .47957 .67881 L .713 .677 .814 r
F
.49206 .6814 m .42061 .66908 L .48704 .67988 L .768 .516 .554 r
F
.49803 .65895 m .42939 .65232 L p .49243 .68124 L .713 .677 .814 r
F P
0 g
s
.49411 .68166 m .49803 .65895 L p .49243 .68124 L .713 .677 .814 r
F P
0 g
s
.42256 .66948 m .42061 .66908 L .47345 .67908 L p .15 0 0 r
F P
0 g
s
.47669 .67729 m .42069 .66668 L p .44554 .66085 L .743 .664 .773 r
F P
0 g
s
.47349 .67669 m .47669 .67729 L p .48442 .67946 L .768 .516 .554 r
F P
0 g
s
.48704 .67988 m .48442 .67946 L .47669 .67729 L .768 .516 .554 r
F
.42061 .66908 m .42069 .66668 L .47349 .67669 L p F P
0 g
s
.48442 .67946 m .42061 .66908 L .47349 .67669 L .768 .516 .554 r
F
.49786 .67735 m .47132 .679 L .44561 .67406 L .347 .765 .978 r
F
.47913 .68 m .47345 .67908 L .49778 .68239 L .768 .516 .554 r
F
.51062 .68611 m .50671 .68536 L .50105 .68368 L .294 .7 .974 r
F
.51062 .68611 m .47345 .67908 L p .49976 .68335 L .768 .516 .554 r
F P
0 g
s
.49536 .68197 m .49206 .6814 L .49411 .68166 L .768 .516 .554 r
F
.49536 .68197 m .49411 .68166 L .51064 .68372 L F
.49206 .6814 m .48704 .67988 L .49411 .68166 L F
.49786 .67735 m .49411 .68192 L p .47132 .679 L .347 .765 .978 r
F P
0 g
s
.44703 .67415 m .42033 .67248 L .42883 .67076 L .347 .765 .978 r
F
.42883 .67076 m .42791 .67067 L p .50081 .68466 L .524 0 0 r
F P
0 g
s
.36907 .66505 m .42791 .67067 L p .4027 .66992 L .347 .765 .978 r
F P
0 g
s
.4205 .67249 m .4205 .67249 L .42033 .67248 L .347 .765 .978 r
F
.34392 .66758 m .36907 .66505 L p .42033 .67248 L F P
0 g
s
.50081 .68466 m .42791 .67067 L .41438 .67179 L .524 0 0 r
F
.31406 .66952 m .36907 .66505 L p .29077 .67157 L .115 .668 .763 r
F P
0 g
s
.24898 .6753 m .25415 .67481 L p .29301 .67154 L .115 .668 .763 r
F P
0 g
s
.25415 .67481 m .27002 .67332 L p .28688 .67201 L .115 .668 .763 r
F P
0 g
s
.29301 .67154 m .25415 .67481 L .28688 .67201 L .115 .668 .763 r
F
.24799 .67397 m .24736 .67388 L .24768 .67416 L p .731 .765 .881 r
F P
0 g
s
.31442 .6297 m .25278 .67098 L .28624 .65015 L .063 0 0 r
F
.25278 .67098 m .25268 .67105 L .25294 .67088 L F
.25278 .67098 m .25294 .67088 L .28624 .65015 L F
.31564 .62842 m .24827 .67339 L p .31442 .6297 L F P
0 g
s
.24736 .67388 m .25294 .67088 L p .25778 .66726 L .673 .96 .941 r
F P
0 g
s
.25294 .67088 m .27852 .65717 L p .26502 .66374 L .673 .96 .941 r
F P
0 g
s
.26631 .66298 m .28824 .65196 L .27845 .65579 L .673 .96 .941 r
F
.24817 .67459 m .25294 .67088 L .25268 .67105 L .731 .765 .881 r
F
.31183 .65228 m .29419 .66586 L .25423 .67241 L p .656 .521 .688 r
F P
0 g
s
.25423 .67241 m .25404 .67245 L p .25453 .67228 L .656 .521 .688 r
F P
0 g
s
.31183 .65228 m .25423 .67241 L .25453 .67228 L .656 .521 .688 r
F
.25453 .67228 m .28959 .65318 L p .25237 .67328 L .731 .765 .881 r
F P
0 g
s
.29419 .66586 m .24817 .67459 L .25211 .67317 L .62 .389 .546 r
F
.24866 .67501 m .24817 .67459 L p .25403 .67481 L .536 .715 .954 r
F P
0 g
s
.29419 .66586 m .29401 .66825 L .24807 .67578 L p .62 .389 .546 r
F P
0 g
s
.25431 .6748 m .25403 .67481 L .24964 .6742 L .536 .715 .954 r
F
.28454 .6618 m .2753 .66096 L .28414 .66194 L .686 .622 .779 r
F
.25322 .67278 m .28824 .65196 L .25134 .67345 L .731 .765 .881 r
F
.25134 .67345 m .28824 .65196 L .24817 .67459 L F
.2753 .66096 m .25453 .67228 L p .28414 .66194 L .686 .622 .779 r
F P
0 g
s
.28959 .65318 m .2753 .66096 L p .28454 .6618 L .686 .622 .779 r
F P
0 g
s
.25174 .67178 m .24949 .67424 L .28223 .65764 L .435 .233 .505 r
F
.25423 .67241 m .29419 .66586 L p .25211 .67317 L .62 .389 .546 r
F P
0 g
s
.25247 .67344 m .28223 .65764 L p .24949 .67424 L .435 .233 .505 r
F P
0 g
s
.24827 .67339 m .24817 .67459 L p .25174 .67178 L .435 .233 .505 r
F P
0 g
s
.25403 .67481 m .24817 .67459 L .24964 .6742 L .536 .715 .954 r
F
.25281 .67476 m .25155 .67503 L .24817 .67459 L .62 .389 .546 r
F
.24807 .67578 m .24811 .67533 L p .25155 .67503 L F P
0 g
s
.24811 .67533 m .24817 .67459 L p .25155 .67503 L .62 .389 .546 r
F P
0 g
s
.24817 .67459 m .24807 .67578 L p .24949 .67424 L .435 .233 .505 r
F P
0 g
s
.24817 .67459 m .24949 .67424 L .25174 .67178 L .435 .233 .505 r
F
.24768 .67416 m .24817 .67459 L p .25268 .67105 L .731 .765 .881 r
F P
0 g
s
.24817 .67459 m .24736 .67388 L p .24964 .6742 L .536 .715 .954 r
F P
0 g
s
.24968 .67419 m .24964 .6742 L .24736 .67388 L .536 .715 .954 r
F
.25431 .6748 m .24964 .6742 L .24968 .67419 L F
.25268 .67105 m .24827 .67339 L p .25174 .67178 L .435 .233 .505 r
F P
0 g
s
.24736 .67388 m .248 .67396 L p .25268 .67105 L .731 .765 .881 r
F P
0 g
s
.28822 .67161 m .25431 .6748 L p .24853 .67403 L .536 .715 .954 r
F P
0 g
s
.25087 .67368 m .24853 .67403 L .24736 .67388 L .536 .715 .954 r
F
.28822 .67161 m .24853 .67403 L .25087 .67368 L F
.29301 .67154 m .28298 .67232 L .28822 .67161 L .115 .668 .763 r
F
.27002 .67332 m .28822 .67161 L p .28298 .67232 L F P
0 g
s
.28822 .67161 m .28686 .6704 L p .29029 .67134 L .228 .566 .938 r
F P
0 g
s
.30029 .67073 m .31289 .6757 L .30041 .67552 L .48 .074 .193 r
F
.28417 .67216 m .30041 .67552 L .25384 .67484 L F
.30029 .67073 m .30041 .67552 L .28417 .67216 L F
.31945 .66842 m .29401 .66825 L p .33204 .66778 L .696 .455 .554 r
F P
0 g
s
.33206 .6661 m .29419 .66586 L p .31976 .6577 L .707 .617 .755 r
F P
0 g
s
.34051 .66687 m .32843 .66362 L .33206 .6661 L .686 .622 .779 r
F
.29401 .66825 m .29419 .66586 L .33206 .6661 L p .696 .455 .554 r
F P
0 g
s
.34065 .66688 m .29401 .66825 L .33206 .6661 L .696 .455 .554 r
F
.28223 .65764 m .25247 .67344 L p .25273 .67341 L .547 .451 .698 r
F P
0 g
s
.30069 .65617 m .28959 .65318 L .28454 .6618 L .686 .622 .779 r
F
.29984 .66 m .28223 .65764 L p .25273 .67341 L .547 .451 .698 r
F P
0 g
s
.25268 .67105 m .25294 .67088 L .25277 .67098 L .731 .765 .881 r
F
.25268 .67105 m .25277 .67098 L .24736 .67388 L F
.28223 .65764 m .28241 .65523 L .25268 .67105 L p .435 .233 .505 r
F P
0 g
s
.25174 .67178 m .28223 .65764 L .25268 .67105 L .435 .233 .505 r
F
.28624 .65015 m .25294 .67088 L .27845 .65579 L .063 0 0 r
F
.27852 .65717 m .28824 .65196 L p .25778 .66726 L .673 .96 .941 r
F P
0 g
s
.25268 .67105 m .27046 .66159 L p .27845 .65579 L .063 0 0 r
F P
0 g
s
.25294 .67088 m .25268 .67105 L .27845 .65579 L .063 0 0 r
F
.25294 .67088 m .26502 .66374 L .25778 .66726 L .673 .96 .941 r
F
.27845 .65579 m .30113 .64273 L .30109 .6424 L F
.29172 .64794 m .27958 .65574 L .28241 .65523 L .063 0 0 r
F
.27046 .66159 m .28241 .65523 L p .27958 .65574 L F P
0 g
s
.28241 .65523 m .28223 .65764 L .28502 .65477 L p .185 .032 .427 r
F P
0 g
s
.27845 .65579 m .27046 .66159 L .29172 .64794 L .063 0 0 r
F
.28241 .65523 m .30477 .63988 L p .31162 .63266 L F P
0 g
s
.25778 .66726 m .26631 .66298 L .28134 .6532 L .673 .96 .941 r
F
.32897 .61869 m .3067 .63835 L .30539 .63938 L 0 0 0 r
F
.31102 .63329 m .331 .61691 L .31066 .63367 L F
.32255 .62381 m .32828 .61904 L .3248 .62154 L .798 .77 .828 r
F
.32828 .61904 m .33031 .61717 L .3248 .62154 L F
.31102 .63329 m .3463 .60268 L .331 .61691 L 0 0 0 r
F
.31373 .63043 m .3463 .60268 L .31102 .63329 L F
.31162 .63266 m .31269 .63152 L .28241 .65523 L .063 0 0 r
F
.34385 .60469 m .34676 .6023 L .38924 .56283 L .317 .031 .311 r
F
.3463 .60268 m .35988 .59156 L .331 .61691 L 0 0 0 r
F
.41438 .53114 m .33715 .61017 L .38394 .56772 L .317 .031 .311 r
F
.33715 .61017 m .34385 .60469 L .38394 .56772 L F
.331 .61691 m .3618 .58999 L .36308 .58894 L 0 0 0 r
F
.38924 .56283 m .34676 .6023 L .37651 .57795 L .317 .031 .311 r
F
.34967 .60456 m .30093 .64129 L .32705 .61124 L .735 .986 .896 r
F
.34967 .60456 m .32705 .61124 L .34533 .5953 L F
.43596 .51483 m .38297 .57289 L .39631 .56216 L 0 0 0 r
F
.4117 .54115 m .38297 .57289 L .43596 .51483 L F
.43527 .50643 m .46559 .4674 L .46517 .46868 L .148 0 .202 r
F
.44494 .49273 m .45535 .47953 L .45348 .48225 L F
.42426 .51809 m .44827 .488 L .44494 .49273 L F
.43676 .50158 m .4499 .48424 L p .42898 .51218 L F P
0 g
s
.4499 .48424 m .45112 .48279 L .42898 .51218 L .148 0 .202 r
F
.4115 .53645 m .44494 .49273 L .43527 .50643 L F
.41268 .53427 m .44494 .49273 L .4115 .53645 L F
.41983 .52458 m .38394 .56772 L .38924 .56283 L .317 .031 .311 r
F
.41268 .53427 m .42426 .51809 L .44494 .49273 L .148 0 .202 r
F
.41438 .53114 m .42426 .51809 L p .41268 .53427 L F P
0 g
s
.42426 .51809 m .43676 .50158 L p .42898 .51218 L .148 0 .202 r
F P
0 g
s
.45646 .48574 m .42147 .53347 L .42036 .51117 L .885 .99 .745 r
F
.42672 .52631 m .42147 .53347 L .43504 .51798 L 0 0 0 r
F
.42672 .52631 m .43504 .51798 L .4381 .51367 L F
.43504 .51798 m .42147 .53347 L p .45646 .48574 L .885 .99 .745 r
F P
0 g
s
.39087 .54595 m .35918 .58297 L .36985 .52281 L .826 .93 .901 r
F
.38763 .4966 m .38765 .49641 L p .36378 .53176 L .786 .858 .904 r
F P
0 g
s
.39087 .54595 m .36985 .52281 L .37871 .50975 L .826 .93 .901 r
F
.37104 .51201 m .38765 .49641 L .3812 .48435 L .786 .858 .904 r
F
.36378 .53176 m .37197 .51962 L .37104 .51201 L F
.37197 .51962 m .38765 .49641 L .37104 .51201 L F
.42147 .53347 m .41119 .52543 L .42036 .51117 L .885 .99 .745 r
F
.41157 .53632 m .38924 .56283 L .40776 .54335 L .317 .031 .311 r
F
.37137 .5875 m .39631 .56216 L p .38297 .57289 L 0 0 0 r
F P
0 g
s
.36261 .59031 m .3733 .58507 L .44372 .5043 L .81 1 .83 r
F
.37137 .5875 m .35125 .60794 L p .34967 .60456 L .735 .986 .896 r
F P
0 g
s
.36422 .59312 m .3733 .58507 L p .36261 .59031 L .81 1 .83 r
F P
0 g
s
.31442 .62224 m .35877 .58359 L .3439 .58862 L .753 .889 .94 r
F
.34967 .60456 m .34378 .59199 L .36189 .58904 L .735 .986 .896 r
F
.36261 .59031 m .44372 .5043 L .35877 .58359 L .81 1 .83 r
F
.36422 .59312 m .36189 .58904 L .37137 .5875 L .735 .986 .896 r
F
.34967 .60456 m .36189 .58904 L .36422 .59312 L F
.40197 .4726 m .39487 .48441 L .36992 .43976 L .879 .885 .834 r
F
.40197 .4726 m .36559 .43532 L .36878 .42978 L F
.47252 .45408 m .49805 .40949 L .49793 .40969 L 0 0 .107 r
F
.46649 .46463 m .46848 .4619 L .47252 .45408 L F
.47569 .44799 m .46649 .46463 L p .4793 .44223 L F P
0 g
s
.49537 .41657 m .46848 .4619 L .46533 .48344 L 0 0 .107 r
F
.48215 .43767 m .46848 .4619 L .49537 .41657 L F
.45098 .48272 m .45512 .4769 L .47838 .44137 L .388 .147 .423 r
F
.47838 .44137 m .45512 .4769 L .47213 .453 L F
.44691 .48777 m .46279 .46436 L .48249 .43373 L F
.48249 .43373 m .46279 .46436 L .47838 .44137 L F
.45646 .48574 m .42036 .51117 L .43276 .49188 L .885 .99 .745 r
F
.46533 .48344 m .49798 .4146 L p .49873 .40908 L 0 0 .107 r
F P
0 g
s
.49798 .4146 m .49858 .41327 L .50645 .39665 L .754 .844 .394 r
F
.5066 .39644 m .50974 .38982 L .50837 .39126 L p 0 0 .107 r
F P
0 g
s
.50647 .39661 m .50645 .39665 L .50915 .39094 L .754 .844 .394 r
F
.49798 .4146 m .50645 .39665 L .50647 .39661 L F
.51885 .34802 m .51481 .36651 L .5146 .37048 L 0 0 .016 r
F
.5189 .34774 m .51887 .3479 L .51885 .34802 L F
.51887 .3479 m .51481 .36651 L .51885 .34802 L F
.5202 .35681 m .52064 .35686 L p .52106 .3375 L .736 .504 .577 r
F P
0 g
s
.51434 .39537 m .51303 .40116 L p .50745 .40708 L .706 .517 .631 r
F P
0 g
s
.50754 .57357 m .51338 .49381 L p .49815 .46658 L .737 .57 .667 r
F P
0 g
s
.49815 .46658 m .49606 .46834 L .50745 .40708 L .706 .517 .631 r
F
.49476 .46943 m .49815 .46658 L .49337 .45803 L .737 .57 .667 r
F
.50754 .57357 m .49815 .46658 L .49476 .46943 L F
.51303 .40116 m .49815 .46658 L p .50745 .40708 L .706 .517 .631 r
F P
0 g
s
.49815 .46658 m .48846 .50918 L p .49606 .46834 L .706 .517 .631 r
F P
0 g
s
.51826 .41137 m .51814 .41405 L p .51832 .41161 L .804 .59 .613 r
F P
0 g
s
.52153 .3676 m .51826 .41137 L .51832 .41161 L .804 .59 .613 r
F
.5203 .36482 m .51826 .41137 L p .52153 .3676 L F P
0 g
s
.51728 .415 m .51967 .37362 L .51819 .37047 L .737 .533 .616 r
F
.51814 .41405 m .51989 .37409 L p .51728 .415 L F P
0 g
s
.51989 .37409 m .52064 .35686 L p .51967 .37362 L .737 .533 .616 r
F P
0 g
s
.51989 .37409 m .51967 .37362 L .51728 .415 L .737 .533 .616 r
F
.51989 .37409 m .52378 .39749 L .51612 .37999 L .771 .546 .593 r
F
.53097 .42006 m .52291 .37072 L .52664 .42991 L .804 .59 .613 r
F
.52291 .37072 m .52064 .35686 L .52153 .3676 L F
.52291 .37072 m .52153 .3676 L .52664 .42991 L F
.52386 .36697 m .52663 .3859 L p .52538 .37082 L .88 .643 .575 r
F P
0 g
s
.53097 .42006 m .52514 .37576 L .52291 .37072 L .804 .59 .613 r
F
.53135 .4192 m .53732 .48539 L .5305 .42113 L .835 .63 .625 r
F
.53712 .45767 m .53146 .41896 L p .53097 .42006 L .804 .59 .613 r
F P
0 g
s
.53146 .41896 m .52514 .37576 L p .53097 .42006 L .804 .59 .613 r
F P
0 g
s
.5305 .42113 m .53732 .48539 L .53281 .47761 L .835 .63 .625 r
F
.52966 .40039 m .53135 .4192 L .5305 .42113 L F
.52514 .37576 m .52966 .40039 L .52327 .38582 L .844 .604 .575 r
F
.52582 .37281 m .52577 .3724 L .52575 .3725 L .88 .643 .575 r
F
.52584 .37199 m .52688 .37967 L p .52514 .37576 L .844 .604 .575 r
F P
0 g
s
.52688 .37967 m .52966 .40039 L p .52514 .37576 L .844 .604 .575 r
F P
0 g
s
.52256 .36992 m .52304 .38081 L p .51989 .37409 L .771 .546 .593 r
F P
0 g
s
.52304 .38081 m .52378 .39749 L p .51989 .37409 L .771 .546 .593 r
F P
0 g
s
.51434 .39537 m .52378 .39749 L p .51875 .41338 L .768 .58 .644 r
F P
0 g
s
.50743 .42576 m .51434 .39537 L p .51875 .41338 L .768 .58 .644 r
F P
0 g
s
.51848 .35661 m .5159 .36557 L p .51819 .37047 L .737 .533 .616 r
F P
0 g
s
.51741 .36879 m .51989 .37409 L .51612 .37999 L .771 .546 .593 r
F
.52378 .39749 m .51434 .39537 L p .51612 .37999 L F P
0 g
s
.52064 .35686 m .51848 .35661 L p .51825 .36765 L .737 .533 .616 r
F P
0 g
s
.51819 .37047 m .51883 .36693 L .51825 .36765 L .737 .533 .616 r
F
.51883 .36693 m .52064 .35686 L .51825 .36765 L F
.51434 .39537 m .51738 .36872 L p .51741 .36879 L .771 .546 .593 r
F P
0 g
s
.51967 .37362 m .52064 .35686 L .51819 .37047 L .737 .533 .616 r
F
.51451 .37971 m .52268 .38258 L p .52229 .37187 L .81 .496 .459 r
F P
0 g
s
.52039 .35343 m .51915 .36423 L p .51857 .3628 L .852 .512 .409 r
F P
0 g
s
.52181 .37108 m .52153 .3676 L .5212 .37204 L .804 .59 .613 r
F
.51915 .36423 m .51451 .37971 L .52229 .37187 L .81 .496 .459 r
F
.51747 .36013 m .51647 .36673 L p .51915 .36423 L F P
0 g
s
.51647 .36673 m .51451 .37971 L p .51915 .36423 L .81 .496 .459 r
F P
0 g
s
.5253 .36767 m .52312 .35778 L .52554 .36893 L 0 0 .299 r
F
.52106 .3375 m .52108 .33732 L .52107 .33737 L .336 .137 .454 r
F
.52109 .33727 m .52115 .33685 L .52114 .33688 L 0 0 0 r
F
.52114 .33688 m .52114 .33691 L p .52112 .33705 L F P
0 g
s
.52114 .33691 m .52113 .337 L p .52112 .33705 L 0 0 0 r
F P
0 g
s
.52115 .33685 m .52106 .3375 L p .52109 .33721 L 0 0 .286 r
F P
0 g
s
.52115 .33695 m .52106 .3375 L .52106 .3375 L .987 .767 .373 r
F
.52109 .33747 m .52106 .3375 L .52108 .33738 L .886 .627 .54 r
F
.52109 .33747 m .52108 .33738 L .52106 .3375 L F
.52112 .33732 m .52106 .3375 L .52111 .33722 L .808 .565 .571 r
F
.52112 .33732 m .52111 .33722 L .52106 .3375 L F
.52106 .3375 m .52107 .33746 L .52106 .3375 L .945 .734 .586 r
F
.52106 .3375 m .52107 .33749 L .52107 .33749 L F
.52115 .3371 m .52115 .33703 L .52106 .3375 L .736 .504 .577 r
F
.52106 .33764 m .52115 .33693 L .52115 .33695 L .987 .767 .373 r
F
.52106 .3375 m .52117 .33696 L .52109 .33729 L .336 .137 .454 r
F
.52112 .33717 m .52117 .33696 L .52117 .33687 L F
.52104 .33749 m .52112 .33717 L .52117 .33687 L F
.5211 .3372 m .52117 .33687 L .52106 .3375 L F
.52118 .3368 m .52106 .3375 L p .52117 .33687 L F P
0 g
s
.52114 .33691 m .52115 .33693 L p .52114 .33699 L .987 .767 .373 r
F P
0 g
s
.52115 .33693 m .52115 .33693 L p .52115 .33695 L .987 .767 .373 r
F P
0 g
s
.52115 .33703 m .52106 .3375 L .52116 .33678 L .736 .504 .577 r
F
.52115 .3371 m .52106 .3375 L .52115 .33703 L F
.52118 .3368 m .52121 .33677 L p .52439 .34869 L .415 .257 .557 r
F P
0 g
s
.52117 .33696 m .52121 .33677 L .5212 .3368 L .336 .137 .454 r
F
.52385 .35623 m .5267 .37351 L .52583 .37044 L 0 0 .394 r
F
.52664 .37388 m .52677 .37474 L .52667 .37407 L .904 .636 .132 r
F
.52676 .37471 m .52667 .37407 L p .52677 .37474 L F P
0 g
s
.52667 .37409 m .52677 .37482 L .52677 .37474 L .904 .636 .132 r
F
.52664 .37388 m .52677 .37482 L .52667 .37409 L F
.52664 .37388 m .52681 .37528 L .52677 .37482 L F
.53615 .41745 m .52601 .37107 L .52979 .38924 L 0 0 .299 r
F
.52601 .37107 m .52593 .3708 L .52979 .38924 L F
.53561 .41444 m .52603 .37116 L .53615 .41745 L F
.52603 .37116 m .52554 .36893 L .52601 .37107 L F
.52603 .37116 m .52601 .37107 L .53615 .41745 L F
.52114 .33703 m .52123 .33785 L .52118 .33799 L .844 .604 .575 r
F
.52106 .3375 m .5212 .33682 L p .52117 .33696 L .336 .137 .454 r
F P
0 g
s
.5212 .33682 m .52121 .33677 L p .5212 .33681 L .336 .137 .454 r
F P
0 g
s
.53592 .41356 m .55236 .47889 L .53704 .42244 L .009 .089 .6 r
F
.55236 .47889 m .5515 .47742 L .53704 .42244 L F
.53592 .41356 m .53704 .42244 L .53532 .41279 L F
.5267 .37351 m .53532 .41279 L p .52647 .36719 L .127 .121 .58 r
F P
0 g
s
.57232 .52552 m .55903 .49022 L .55806 .48856 L 0 .005 .571 r
F
.55806 .48856 m .55085 .47633 L .55092 .47656 L F
.55953 .50628 m .56328 .52112 L .56224 .51565 L .773 .477 0 r
F
.53592 .41356 m .55454 .46068 L .5632 .52195 L .009 .089 .6 r
F
.61655 .46265 m .65269 .50335 L .62429 .4921 L .499 .399 .672 r
F
.6405 .49363 m .64165 .49256 L p .59268 .48912 L .436 .359 .671 r
F P
0 g
s
.59494 .48934 m .63235 .55074 L .61471 .55177 L .426 .407 .735 r
F
.58776 .46664 m .60005 .48982 L .59494 .48934 L F
.60005 .48982 m .63235 .55074 L .59494 .48934 L F
.59181 .52315 m .55462 .46121 L .57294 .52452 L .287 .302 .701 r
F
.58871 .49857 m .55462 .46121 L .59181 .52315 L F
.65473 .57154 m .60751 .56889 L .65557 .59167 L .424 .463 .797 r
F
.60751 .56889 m .60862 .57495 L .65557 .59167 L F
.56785 .49439 m .57296 .48727 L p .59171 .52239 L .426 .407 .735 r
F P
0 g
s
.65568 .6776 m .65442 .68199 L .65819 .67638 L .455 .585 .894 r
F
.65392 .67846 m .64 .70345 L .65442 .68199 L F
.65764 .67075 m .65702 .67293 L .66012 .65936 L 0 0 .267 r
F
.62536 .58544 m .65604 .61299 L p .6513 .67974 L .434 .524 .851 r
F P
0 g
s
.62536 .58544 m .6513 .67974 L .64563 .6825 L .434 .524 .851 r
F
.65568 .6776 m .66012 .65936 L p .65702 .67293 L 0 0 .267 r
F P
0 g
s
.66012 .65936 m .65568 .6776 L p .65604 .61299 L .434 .524 .851 r
F P
0 g
s
.65604 .61299 m .65568 .6776 L .6513 .67974 L .434 .524 .851 r
F
.65764 .67075 m .658 .67113 L p .66873 .61986 L .551 .618 .873 r
F P
0 g
s
.66012 .65936 m .66866 .62433 L p .65816 .66896 L 0 0 .267 r
F P
0 g
s
.658 .67113 m .65764 .67075 L p .6574 .67159 L .455 .585 .894 r
F P
0 g
s
.65704 .67286 m .63916 .68566 L .65392 .67846 L .455 .585 .894 r
F
.658 .67113 m .6574 .67159 L .65704 .67286 L F
.6574 .67159 m .63916 .68566 L .65704 .67286 L F
.69674 .59074 m .68621 .60198 L .70078 .58854 L .551 .618 .873 r
F
.67393 .61333 m .63743 .64959 L .68049 .60809 L F
.68621 .60198 m .68049 .60809 L .70078 .58854 L F
.67393 .61333 m .68049 .60809 L .68621 .60198 L F
.78894 .51113 m .73612 .54025 L .73034 .54817 L .58 .601 .837 r
F
.78894 .51113 m .73034 .54817 L .71335 .57146 L F
.71258 .57585 m .78894 .51113 L p .72472 .56239 L F P
0 g
s
.72472 .56239 m .71335 .57146 L .71258 .57585 L .58 .601 .837 r
F
.70078 .58854 m .63743 .64959 L .70293 .58737 L .551 .618 .873 r
F
.7062 .58125 m .71258 .57585 L p .71335 .57146 L .58 .601 .837 r
F P
0 g
s
.70576 .58112 m .7062 .58125 L p .73612 .54025 L .58 .601 .837 r
F P
0 g
s
.69674 .59074 m .71347 .57074 L p .7019 .58525 L .176 .06 .473 r
F P
0 g
s
.69654 .59086 m .67393 .61333 L .68621 .60198 L .551 .618 .873 r
F
.69654 .59086 m .68621 .60198 L .69674 .59074 L F
.71347 .57074 m .69674 .59074 L p .71951 .53634 L .522 .544 .819 r
F P
0 g
s
.65645 .61277 m .63743 .64959 L p .69654 .59086 L .551 .618 .873 r
F P
0 g
s
.69674 .59074 m .66866 .62433 L p .68406 .59768 L .522 .544 .819 r
F P
0 g
s
.71951 .53634 m .69674 .59074 L .68406 .59768 L .522 .544 .819 r
F
.70576 .58112 m .7019 .58525 L .71347 .57074 L .176 .06 .473 r
F
.71774 .52633 m .6642 .55233 L .71904 .53623 L .504 .467 .749 r
F
.67258 .52575 m .71951 .53634 L p .68406 .59768 L .522 .544 .819 r
F P
0 g
s
.67161 .55015 m .68406 .59768 L .6694 .60569 L .522 .544 .819 r
F
.67258 .52575 m .67626 .54879 L .67161 .55015 L F
.67626 .54879 m .68406 .59768 L .67161 .55015 L F
.6469 .54988 m .65405 .55531 L p .64343 .51939 L .504 .467 .749 r
F P
0 g
s
.67161 .55015 m .67109 .56326 L .6642 .55233 L .522 .544 .819 r
F
.63743 .64959 m .65764 .67075 L p .66873 .61986 L .551 .618 .873 r
F P
0 g
s
.67109 .56326 m .66866 .62433 L .6642 .55233 L .522 .544 .819 r
F
.57783 .52695 m .60448 .55237 L .61325 .64252 L .235 .32 .747 r
F
.61742 .62288 m .61325 .64252 L p .60448 .55237 L F P
0 g
s
.65473 .57154 m .6469 .54988 L .58864 .56784 L .424 .463 .797 r
F
.6642 .55233 m .66866 .62433 L .65512 .58086 L .522 .544 .819 r
F
.6642 .55233 m .65512 .58086 L .65405 .55531 L F
.67709 .57279 m .65405 .55531 L p .65473 .57154 L .424 .463 .797 r
F P
0 g
s
.65405 .55531 m .6469 .54988 L p .65473 .57154 L .424 .463 .797 r
F P
0 g
s
.65405 .55531 m .65512 .58086 L .64773 .55716 L .522 .544 .819 r
F
.65405 .55531 m .67709 .57279 L p .6642 .55233 L .504 .467 .749 r
F P
0 g
s
.64343 .51939 m .65405 .55531 L .6642 .55233 L .504 .467 .749 r
F
.78486 .51261 m .67709 .57279 L .78894 .51113 L .58 .601 .837 r
F
.72218 .49305 m .72274 .49305 L p .76108 .52124 L .554 .503 .754 r
F P
0 g
s
.74763 .49683 m .79306 .4936 L .78311 .5022 L .554 .503 .754 r
F
.7768 .49348 m .79306 .4936 L p .74763 .49683 L F P
0 g
s
.79282 .4925 m .78532 .49288 L .79306 .4936 L .585 .525 .754 r
F
.79463 .49162 m .78532 .49288 L .81934 .49112 L F
.81934 .49112 m .79282 .4925 L .80644 .49484 L F
.79282 .4925 m .79306 .4936 L .80644 .49484 L F
.78968 .49329 m .78825 .49281 L .80827 .49367 L .488 .67 .945 r
F
.78068 .4931 m .78546 .4929 L .78968 .49329 L F
.77972 .49308 m .78148 .49303 L .78068 .4931 L F
.78148 .49303 m .78546 .4929 L .78068 .4931 L F
.78605 .49252 m .78532 .49288 L p .79463 .49162 L .585 .525 .754 r
F P
0 g
s
.83825 .48567 m .79264 .49166 L p .79463 .49162 L .585 .525 .754 r
F P
0 g
s
.79264 .49166 m .78605 .49252 L p .79463 .49162 L .585 .525 .754 r
F P
0 g
s
.78605 .49252 m .79264 .49166 L p .79147 .49207 L .138 0 0 r
F P
0 g
s
.78605 .49252 m .77727 .49214 L .77563 .49323 L .536 .412 .659 r
F
.72274 .49305 m .7768 .49348 L p .74763 .49683 L .554 .503 .754 r
F P
0 g
s
.72274 .49305 m .78311 .5022 L .76108 .52124 L .554 .503 .754 r
F
.72218 .49305 m .76108 .52124 L .73517 .53065 L F
.8085 .49206 m .81101 .49206 L p .78148 .49303 L .488 .67 .945 r
F P
0 g
s
.8085 .49206 m .78148 .49303 L .78157 .49302 L .488 .67 .945 r
F
.79192 .49202 m .8085 .49206 L p .78157 .49302 L F P
0 g
s
.78546 .4929 m .78825 .49281 L .78968 .49329 L .488 .67 .945 r
F
.78486 .51261 m .77966 .5145 L .67709 .57279 L .58 .601 .837 r
F
.71908 .5365 m .68308 .56761 L p .77966 .5145 L F P
0 g
s
.68308 .56761 m .67709 .57279 L p .77966 .5145 L .58 .601 .837 r
F P
0 g
s
.7388 .54314 m .73863 .54066 L .77966 .5145 L p .4 .221 .52 r
F P
0 g
s
.79586 .50672 m .83825 .48567 L p .82082 .4907 L .585 .525 .754 r
F P
0 g
s
.77966 .5145 m .79538 .50447 L p .78902 .50935 L .4 .221 .52 r
F P
0 g
s
.79059 .5083 m .78902 .50935 L .79538 .50447 L .4 .221 .52 r
F
.79538 .50447 m .77966 .5145 L p .79306 .4936 L .554 .503 .754 r
F P
0 g
s
.77966 .5145 m .73863 .54066 L p .76108 .52124 L .554 .503 .754 r
F P
0 g
s
.79306 .4936 m .77966 .5145 L .76108 .52124 L .554 .503 .754 r
F
.72152 .50222 m .73517 .53065 L .71904 .53623 L F
.72513 .50157 m .73517 .53065 L .72152 .50222 L F
.71951 .53634 m .71904 .53623 L p .73517 .53065 L F P
0 g
s
.82082 .4907 m .80805 .49438 L .8075 .49494 L .585 .525 .754 r
F
.81336 .49549 m .8075 .49494 L .79586 .50672 L F
.82082 .4907 m .8075 .49494 L .81336 .49549 L F
.78068 .4931 m .78968 .49329 L .7779 .49337 L .488 .67 .945 r
F
.79306 .4936 m .77018 .49342 L p .78968 .49329 L F P
0 g
s
.78532 .49288 m .74764 .5118 L p .80644 .49484 L .585 .525 .754 r
F P
0 g
s
.8122 .49375 m .79306 .4936 L p .78968 .49329 L .488 .67 .945 r
F P
0 g
s
.7917 .50879 m .79586 .50672 L p .80805 .49438 L .585 .525 .754 r
F P
0 g
s
.7898 .50892 m .7917 .50879 L p .80805 .49438 L .585 .525 .754 r
F P
0 g
s
.8122 .49375 m .81007 .49777 L .78902 .50935 L .4 .221 .52 r
F
.74764 .5118 m .7898 .50892 L p .80805 .49438 L .585 .525 .754 r
F P
0 g
s
.79538 .50447 m .8122 .49375 L p .79059 .5083 L .4 .221 .52 r
F P
0 g
s
.79306 .4936 m .8122 .49375 L .79538 .50447 L p .554 .503 .754 r
F P
0 g
s
.81101 .49206 m .8122 .49375 L p .80827 .49367 L .488 .67 .945 r
F P
0 g
s
.78825 .49281 m .81101 .49206 L .80827 .49367 L .488 .67 .945 r
F
.81125 .49458 m .81101 .49206 L .7879 .4941 L p .65 .401 .53 r
F P
0 g
s
.78221 .49522 m .77976 .49568 L .78437 .49336 L .138 0 0 r
F
.78015 .49563 m .77976 .49568 L p .78221 .49522 L F P
0 g
s
.77972 .49308 m .78068 .4931 L .77018 .49342 L .488 .67 .945 r
F
.78068 .4931 m .7779 .49337 L .77018 .49342 L F
.78157 .49302 m .78148 .49303 L .77972 .49308 L F
.74764 .5118 m .78532 .49288 L p .77563 .49323 L .536 .412 .659 r
F P
0 g
s
.71407 .4984 m .72218 .49305 L p .71904 .53623 L .554 .503 .754 r
F P
0 g
s
.71473 .50345 m .74764 .5118 L p .77727 .49214 L .536 .412 .659 r
F P
0 g
s
.77976 .49568 m .73924 .50063 L p .78437 .49336 L .138 0 0 r
F P
0 g
s
.78532 .49288 m .78605 .49252 L p .77563 .49323 L .536 .412 .659 r
F P
0 g
s
.73924 .50063 m .78517 .49267 L p .78437 .49336 L .138 0 0 r
F P
0 g
s
.78015 .49563 m .78739 .49538 L .7879 .4941 L .65 .401 .53 r
F
.74764 .5118 m .71908 .5365 L p .78628 .51086 L .58 .601 .837 r
F P
0 g
s
.7898 .50892 m .74764 .5118 L p .78628 .51086 L .58 .601 .837 r
F P
0 g
s
.71955 .53609 m .74764 .5118 L p .71774 .52633 L .504 .467 .749 r
F P
0 g
s
.58776 .46664 m .64171 .50424 L p .63235 .55074 L .426 .407 .735 r
F P
0 g
s
.59555 .55289 m .58864 .56784 L p .6469 .54988 L .424 .463 .797 r
F P
0 g
s
.64773 .55716 m .62536 .58544 L p .64171 .50424 L .426 .407 .735 r
F P
0 g
s
.71613 .5141 m .69948 .49958 L .60993 .52183 L .504 .467 .749 r
F
.74764 .5118 m .71473 .50345 L p .71613 .5141 L F P
0 g
s
.71473 .50345 m .69948 .49958 L p .71613 .5141 L .504 .467 .749 r
F P
0 g
s
.62778 .50536 m .60993 .52183 L p .69948 .49958 L .504 .467 .749 r
F P
0 g
s
.67258 .52575 m .64171 .50424 L p .65269 .50335 L .499 .399 .672 r
F P
0 g
s
.61925 .46263 m .58948 .44127 L .61655 .46265 L .505 .35 .604 r
F
.72851 .49003 m .70911 .48919 L .71813 .49076 L .536 .412 .659 r
F
.7213 .49188 m .72218 .49305 L p .71754 .49135 L .211 .509 .91 r
F P
0 g
s
.64776 .46606 m .63729 .46254 L .61655 .46265 L .505 .35 .604 r
F
.61783 .46245 m .64427 .46928 L .67258 .52575 L .499 .399 .672 r
F
.61806 .47012 m .56835 .443 L .62068 .47839 L .436 .359 .671 r
F
.61866 .4707 m .62068 .47839 L .64165 .49256 L F
.61806 .47012 m .62068 .47839 L .61866 .4707 L F
.57828 .43215 m .58477 .42847 L .58578 .44139 L .456 .323 .611 r
F
.53943 .40246 m .5681 .44054 L .56865 .44581 L .328 .278 .649 r
F
.59494 .48934 m .62536 .58544 L .59171 .52239 L .426 .407 .735 r
F
.59494 .48934 m .59171 .52239 L .58746 .48863 L F
.57049 .46357 m .60993 .52183 L p .56835 .443 L .436 .359 .671 r
F P
0 g
s
.53943 .40246 m .56865 .44581 L .55631 .45708 L .328 .278 .649 r
F
.56865 .44581 m .58515 .47028 L .57049 .46357 L F
.56865 .44581 m .57049 .46357 L .55631 .45708 L F
.54046 .41936 m .55381 .45593 L p .53592 .41356 L .009 .089 .6 r
F P
0 g
s
.58871 .49857 m .55381 .45593 L .55462 .46121 L .287 .302 .701 r
F
.60993 .52183 m .58846 .49662 L .58871 .49857 L F
.58846 .49662 m .55381 .45593 L .58871 .49857 L F
.58776 .46664 m .59494 .48934 L .58746 .48863 L .426 .407 .735 r
F
.60993 .52183 m .57049 .46357 L p .55381 .45593 L .287 .302 .701 r
F P
0 g
s
.57296 .48727 m .58515 .47028 L p .58746 .48863 L .426 .407 .735 r
F P
0 g
s
.58515 .47028 m .58776 .46664 L p .58746 .48863 L .426 .407 .735 r
F P
0 g
s
.55381 .45593 m .56785 .49439 L p .55454 .46068 L .009 .089 .6 r
F P
0 g
s
.55381 .45593 m .55454 .46068 L .53592 .41356 L .009 .089 .6 r
F
.56355 .51987 m .5632 .52195 L p .55454 .46068 L F P
0 g
s
.55462 .46121 m .55381 .45593 L .54692 .45277 L .287 .302 .701 r
F
.52647 .36719 m .53532 .41279 L .53594 .41358 L .127 .121 .58 r
F
.52371 .34834 m .52838 .36372 L .52832 .36933 L .217 .138 .547 r
F
.52538 .35251 m .52838 .36372 L p .52371 .34834 L F P
0 g
s
.52218 .34273 m .52832 .36933 L .52628 .36696 L .217 .138 .547 r
F
.52118 .3368 m .52292 .34473 L .52218 .34273 L F
.52292 .34473 m .52832 .36933 L .52218 .34273 L F
.52838 .36372 m .52432 .34854 L p .52841 .36374 L .415 .257 .557 r
F P
0 g
s
.52432 .34854 m .52118 .3368 L p .52434 .34858 L .415 .257 .557 r
F P
0 g
s
.52841 .36374 m .52432 .34854 L .52434 .34858 L .415 .257 .557 r
F
.53942 .40237 m .56809 .4405 L .53943 .40246 L .328 .278 .649 r
F
.56809 .4405 m .5681 .44054 L .53943 .40246 L F
.61815 .46282 m .67258 .52575 L .61655 .46265 L .499 .399 .672 r
F
.64817 .48655 m .58995 .44254 L p .56835 .443 L .436 .359 .671 r
F P
0 g
s
.58995 .44254 m .56766 .43636 L .56835 .443 L .436 .359 .671 r
F
.54836 .40728 m .58776 .46664 L p .53942 .40237 L .328 .278 .649 r
F P
0 g
s
.5212 .33694 m .52628 .36696 L .5212 .33717 L .217 .138 .547 r
F
.52625 .36694 m .52115 .33685 L p .52628 .36696 L F P
0 g
s
.54372 .39021 m .5414 .38618 L .54166 .38744 L .369 .26 .597 r
F
.54836 .40728 m .53942 .40237 L .53942 .40234 L .328 .278 .649 r
F
.54372 .39021 m .54166 .38744 L .55078 .43168 L .369 .26 .597 r
F
.54728 .39639 m .54372 .39021 L .55078 .43168 L F
.53938 .38426 m .53561 .37612 L .53336 .37855 L F
.53729 .37904 m .53561 .37612 L .53938 .38426 L F
.53561 .37612 m .52854 .36385 L .53025 .36905 L F
.53561 .37612 m .53025 .36905 L .53336 .37855 L F
.5414 .38618 m .53336 .37855 L .55078 .43168 L F
.53851 .3832 m .53842 .38309 L p .5392 .38384 L 0 .077 .608 r
F P
0 g
s
.5414 .38618 m .53729 .37904 L .53938 .38426 L .369 .26 .597 r
F
.54967 .4328 m .55078 .43168 L p .53025 .36905 L F P
0 g
s
.53025 .36905 m .52854 .36385 L .53006 .3688 L .369 .26 .597 r
F
.53025 .36905 m .53006 .3688 L .54967 .4328 L F
.52854 .36385 m .52838 .36372 L .54046 .40292 L F
.52854 .36385 m .54046 .40292 L .54052 .40295 L F
.52832 .36933 m .53942 .40234 L p .52838 .36372 L F P
0 g
s
.53942 .40234 m .52836 .36947 L p .53951 .40379 L .127 .121 .58 r
F P
0 g
s
.52836 .36947 m .52824 .36925 L .53951 .40379 L .127 .121 .58 r
F
.5266 .36424 m .52656 .36415 L .53951 .40379 L F
.52593 .36225 m .52595 .36247 L .52642 .36391 L F
.52656 .36415 m .52642 .36391 L .53951 .40379 L F
.52593 .36225 m .52642 .36391 L .52656 .36415 L F
.56063 .42068 m .57398 .43812 L .57032 .4371 L .456 .323 .611 r
F
.52838 .36372 m .53942 .40234 L .54046 .40292 L .369 .26 .597 r
F
.52841 .36374 m .53818 .3823 L .53016 .37029 L .415 .257 .557 r
F
.54748 .39839 m .57398 .43812 L .56063 .42068 L .456 .323 .611 r
F
.54733 .39684 m .57398 .43812 L .54748 .39839 L F
.53729 .37904 m .53083 .36576 L .52854 .36385 L .369 .26 .597 r
F
.53588 .37551 m .53083 .36576 L .53729 .37904 L F
.54221 .38695 m .53083 .36576 L .5414 .38618 L F
.53113 .36602 m .53457 .37225 L p .53094 .36342 L .481 .306 .566 r
F P
0 g
s
.54728 .39639 m .54221 .38695 L .5414 .38618 L .369 .26 .597 r
F
.52121 .33677 m .53089 .36273 L p .52841 .36374 L .415 .257 .557 r
F P
0 g
s
.53089 .36273 m .53113 .36602 L .52841 .36374 L .415 .257 .557 r
F
.53094 .36342 m .53457 .37225 L .53173 .36491 L .481 .306 .566 r
F
.55564 .39951 m .53255 .36577 L .5448 .38944 L F
.55564 .39951 m .5448 .38944 L .55464 .39899 L F
.53255 .36577 m .53143 .36414 L .53691 .37469 L F
.53143 .36414 m .53094 .36342 L .53173 .36491 L F
.53143 .36414 m .53173 .36491 L .53691 .37469 L F
.53027 .36109 m .53088 .36272 L .53027 .36108 L .891 .481 .065 r
F
.53088 .36272 m .53088 .36273 L .53027 .36108 L F
.56019 .40442 m .55464 .39899 L p .56 .40423 L 0 .077 .608 r
F P
0 g
s
.57933 .42314 m .55464 .39899 L .55483 .39916 L 0 .077 .608 r
F
.58411 .42782 m .57931 .42312 L .57933 .42314 L F
.57931 .42312 m .55464 .39899 L .57933 .42314 L F
.64427 .46928 m .68138 .47887 L .67957 .4885 L .499 .399 .672 r
F
.64427 .46928 m .67957 .4885 L .67258 .52575 L F
.6153 .45513 m .6074 .45089 L .60982 .45329 L .505 .35 .604 r
F
.64319 .46251 m .61438 .45082 L .62917 .46258 L F
.59654 .46046 m .58776 .46664 L p .58477 .42847 L .456 .323 .611 r
F P
0 g
s
.61438 .45082 m .60243 .44597 L .62917 .46258 L .505 .35 .604 r
F
.60243 .44597 m .58528 .43901 L .6074 .45089 L F
.60243 .44597 m .6074 .45089 L .62917 .46258 L F
.54719 .39546 m .54291 .38761 L .54728 .39639 L .369 .26 .597 r
F
.54291 .38761 m .53068 .36518 L .5423 .38704 L F
.54291 .38761 m .5423 .38704 L .54728 .39639 L F
.52513 .3538 m .52458 .35412 L p .52864 .36005 L .481 .306 .566 r
F P
0 g
s
.52806 .3586 m .52513 .3538 L .52864 .36005 L .481 .306 .566 r
F
.53125 .36368 m .52797 .35838 L .53143 .36414 L F
.52797 .35838 m .52513 .3538 L .52806 .3586 L F
.52797 .35838 m .52806 .3586 L .53143 .36414 L F
.52118 .3368 m .52538 .35251 L p .52371 .34834 L .217 .138 .547 r
F P
0 g
s
.5212 .33694 m .5212 .33717 L .52115 .33685 L .217 .138 .547 r
F
.52114 .33713 m .5212 .33682 L .5212 .33682 L .945 .734 .586 r
F
.52106 .3375 m .52109 .33729 L .52108 .33732 L .336 .137 .454 r
F
.52108 .33738 m .52106 .3375 L .52107 .33729 L .886 .627 .54 r
F
.52108 .33738 m .52107 .33729 L .52106 .3375 L F
.52118 .3368 m .52115 .33685 L p .52107 .33731 L 0 0 .286 r
F P
0 g
s
.52111 .33722 m .52106 .3375 L .52109 .33693 L .808 .565 .571 r
F
.52106 .3375 m .52109 .33693 L .52109 .33693 L F
.52111 .33722 m .52109 .33693 L .52106 .3375 L F
.5211 .33677 m .5211 .33677 L p .52106 .3375 L .736 .504 .577 r
F P
0 g
s
.5211 .33677 m .52116 .33678 L p .52106 .3375 L .736 .504 .577 r
F P
0 g
s
.52109 .33693 m .52109 .33692 L .52106 .3375 L .808 .565 .571 r
F
.52106 .3375 m .52106 .33742 L .52107 .33743 L .464 .275 .539 r
F
.52108 .33677 m .52106 .3375 L p .52109 .33692 L .808 .565 .571 r
F P
0 g
s
.52106 .3375 m .52107 .33743 L .52107 .33743 L .464 .275 .539 r
F
.52107 .33749 m .52107 .33746 L .52107 .33745 L .945 .734 .586 r
F
.52106 .3375 m .52109 .33738 L .52106 .3375 L F
.52106 .3375 m .52107 .33743 L .52107 .33744 L .464 .275 .539 r
F
.52116 .33678 m .52106 .3375 L p .52115 .33703 L .736 .504 .577 r
F P
0 g
s
.5212 .33682 m .5212 .33681 L .52117 .33696 L .336 .137 .454 r
F
.52115 .33685 m .52118 .3368 L p .5212 .33694 L .217 .138 .547 r
F P
0 g
s
.52117 .33696 m .5212 .3368 L .52117 .33687 L .336 .137 .454 r
F
.52115 .33693 m .52127 .33776 L p .52127 .33776 L .911 .651 .532 r
F P
0 g
s
.52115 .33693 m .52114 .33691 L p .52125 .33768 L .904 .636 .132 r
F P
0 g
s
.52125 .33768 m .52127 .33776 L .52115 .33693 L .904 .636 .132 r
F
.52114 .33691 m .52115 .33685 L p .52121 .33734 L 0 0 .394 r
F P
0 g
s
.52121 .33734 m .52385 .35623 L .52123 .33757 L 0 0 .394 r
F
.52121 .33734 m .52123 .33757 L .52114 .33691 L F
.52114 .33698 m .52504 .36597 L .52306 .35552 L .844 .604 .575 r
F
.52114 .33713 m .52118 .33693 L .52114 .33713 L .945 .734 .586 r
F
.52118 .33693 m .52121 .33677 L .52118 .33693 L F
.52115 .33685 m .52248 .34472 L p .52385 .35623 L 0 0 .394 r
F P
0 g
s
.52118 .33693 m .52118 .33693 L .52114 .33713 L .945 .734 .586 r
F
.5212 .33677 m .52121 .33677 L p .52106 .3375 L .66 .434 .566 r
F P
0 g
s
.52106 .3375 m .52107 .33744 L .52107 .33745 L .464 .275 .539 r
F
.52116 .33678 m .52108 .33677 L p .52113 .33696 L .724 .426 .472 r
F P
0 g
s
.5239 .3567 m .52114 .33691 L p .52385 .35623 L 0 0 .394 r
F P
0 g
s
.52121 .33677 m .52118 .3368 L p .52117 .33687 L .336 .137 .454 r
F P
0 g
s
.52106 .3375 m .52116 .33678 L .5212 .33677 L p .66 .434 .566 r
F P
0 g
s
.52113 .33696 m .52365 .34751 L .52118 .33686 L .724 .426 .472 r
F
.52113 .33696 m .52118 .33686 L .52116 .33678 L F
.53089 .36273 m .52121 .33677 L p .53088 .36273 L .891 .481 .065 r
F P
0 g
s
.52121 .33677 m .53022 .36102 L p .53088 .36272 L .891 .481 .065 r
F P
0 g
s
.53096 .36252 m .52513 .3538 L .53255 .36577 L .481 .306 .566 r
F
.55712 .40138 m .55597 .39968 L p .55654 .40082 L F P
0 g
s
.53086 .36231 m .52513 .3538 L p .53096 .36252 L .481 .306 .566 r
F P
0 g
s
.53143 .36414 m .52806 .3586 L .52864 .36005 L .481 .306 .566 r
F
.5309 .3629 m .53248 .36714 L .53113 .36602 L .415 .257 .557 r
F
.53089 .36273 m .53248 .36714 L .5309 .3629 L F
.53089 .36273 m .53255 .3672 L p .53248 .36714 L F P
0 g
s
.54474 .3907 m .53068 .36518 L p .54719 .39546 L .369 .26 .597 r
F P
0 g
s
.53068 .36518 m .52458 .35412 L p .53141 .36684 L .369 .26 .597 r
F P
0 g
s
.53068 .36518 m .53141 .36684 L .54728 .39639 L .369 .26 .597 r
F
.52458 .35412 m .53113 .36602 L p .53094 .36342 L .481 .306 .566 r
F P
0 g
s
.53255 .3672 m .53818 .3823 L p .52841 .36374 L .415 .257 .557 r
F P
0 g
s
.53083 .36576 m .52458 .35412 L .52854 .36385 L .369 .26 .597 r
F
.52323 .35802 m .52106 .3375 L p .52109 .33747 L .886 .627 .54 r
F P
0 g
s
.52106 .3375 m .52323 .35802 L p .52112 .33812 L .977 .7 .306 r
F P
0 g
s
.52134 .34029 m .52112 .33812 L .52323 .35802 L .977 .7 .306 r
F
.52356 .3552 m .52301 .35667 L p .52404 .35961 L .127 .121 .58 r
F P
0 g
s
.52248 .34472 m .52625 .36694 L p .52385 .35623 L 0 0 .394 r
F P
0 g
s
.52625 .36694 m .52642 .36942 L .52385 .35623 L 0 0 .394 r
F
.52625 .36694 m .5267 .37351 L .52647 .36719 L .127 .121 .58 r
F
.52836 .37934 m .52775 .37721 L .52649 .37043 L 0 0 .394 r
F
.52642 .36942 m .52649 .37043 L .52385 .35623 L F
.52836 .37934 m .52649 .37043 L .52642 .36942 L F
.53951 .40379 m .52642 .36391 L .52599 .36314 L .127 .121 .58 r
F
.52617 .3657 m .52599 .36314 L .52404 .35961 L F
.53951 .40379 m .52599 .36314 L .52617 .3657 L F
.52385 .35623 m .52398 .35719 L .5239 .3567 L 0 0 .394 r
F
.52642 .36391 m .52356 .3552 L .52404 .35961 L .127 .121 .58 r
F
.54568 .39429 m .58578 .44139 L .57398 .43812 L .456 .323 .611 r
F
.54715 .39506 m .54474 .3907 L p .54719 .39546 L .369 .26 .597 r
F P
0 g
s
.54889 .39598 m .58578 .44139 L .54568 .39429 L .456 .323 .611 r
F
.56145 .421 m .58528 .43901 L p .58924 .44062 L .505 .35 .604 r
F P
0 g
s
.55464 .39899 m .5448 .38944 L .54792 .39547 L .481 .306 .566 r
F
.54701 .39366 m .54807 .39255 L .57828 .43215 L .456 .323 .611 r
F
.54807 .39255 m .58477 .42847 L p .57828 .43215 L F P
0 g
s
.58675 .43385 m .56145 .421 L .58924 .44062 L .505 .35 .604 r
F
.53457 .37225 m .56145 .421 L p .53173 .36491 L .481 .306 .566 r
F P
0 g
s
.55642 .40071 m .57858 .42234 L .58411 .42782 L 0 .077 .608 r
F
.55887 .40308 m .57858 .42234 L .55642 .40071 L F
.55495 .39928 m .54677 .39128 L .54682 .39133 L F
.58411 .42782 m .55483 .39916 L .55495 .39928 L F
.55483 .39916 m .54677 .39128 L .55495 .39928 L F
.55564 .39951 m .55464 .39899 L .5558 .4001 L .481 .306 .566 r
F
.55654 .40082 m .55564 .39951 L .5558 .4001 L F
.54743 .3919 m .55805 .4023 L .54682 .39133 L 0 .077 .608 r
F
.55805 .4023 m .58411 .42782 L .55971 .40394 L F
.55805 .4023 m .55971 .40394 L .54682 .39133 L F
.5558 .4001 m .54792 .39547 L .56145 .421 L .481 .306 .566 r
F
.55464 .39899 m .54792 .39547 L .5558 .4001 L F
.56789 .41741 m .55654 .40082 L .55862 .40937 L F
.56145 .421 m .56789 .41741 L p .55862 .40937 L F P
0 g
s
.64319 .46251 m .58675 .43385 L .58924 .44062 L .505 .35 .604 r
F
.60828 .45137 m .6087 .4519 L p .60627 .44951 L 0 .077 .608 r
F P
0 g
s
.58477 .42847 m .56019 .40442 L p .58411 .42782 L 0 .077 .608 r
F P
0 g
s
.58411 .42782 m .56019 .40442 L .56 .40423 L 0 .077 .608 r
F
.56789 .41741 m .56145 .421 L p .60416 .44269 L .505 .35 .604 r
F P
0 g
s
.78605 .49252 m .67903 .47707 L p .70911 .48919 L .536 .412 .659 r
F P
0 g
s
.68581 .47809 m .67917 .47711 L .68578 .47804 L .527 .173 .309 r
F
.67903 .47707 m .68578 .47804 L p .67917 .47711 L F P
0 g
s
.68138 .47887 m .68728 .48039 L .68698 .48131 L .499 .399 .672 r
F
.68138 .47887 m .68698 .48131 L .67258 .52575 L F
.67222 .47493 m .68728 .48039 L p .62009 .46303 L F P
0 g
s
.62064 .46309 m .62009 .46303 L .61783 .46245 L .499 .399 .672 r
F
.67222 .47493 m .62009 .46303 L .62064 .46309 L F
.65086 .4664 m .58414 .42676 L .60416 .44269 L .505 .35 .604 r
F
.70016 .48616 m .70697 .4891 L p .70911 .48919 L .536 .412 .659 r
F P
0 g
s
.70326 .50552 m .67258 .52575 L p .68728 .48039 L .499 .399 .672 r
F P
0 g
s
.66902 .47562 m .64817 .48655 L p .70684 .48909 L .536 .412 .659 r
F P
0 g
s
.67532 .4759 m .68129 .47798 L .68423 .47916 L .211 .509 .91 r
F
.68168 .4779 m .7213 .49188 L .67532 .4759 L F
.65602 .46884 m .68423 .47916 L .64776 .46606 L F
.68728 .48039 m .64776 .46606 L p .68423 .47916 L F P
0 g
s
.67903 .47707 m .66902 .47562 L p .70911 .48919 L .536 .412 .659 r
F P
0 g
s
.70911 .48919 m .64817 .48655 L .71987 .49353 L .536 .412 .659 r
F
.68713 .48055 m .66902 .47562 L .67917 .47711 L .527 .173 .309 r
F
.68815 .48244 m .66902 .47562 L .68713 .48055 L F
.72851 .49003 m .71813 .49076 L .73181 .49095 L .536 .412 .659 r
F
.75886 .49134 m .72851 .49003 L .73181 .49095 L F
.68653 .47942 m .67917 .47711 L .68581 .47809 L .527 .173 .309 r
F
.67732 .47636 m .68168 .4779 L .67915 .47711 L .211 .509 .91 r
F
.67732 .47636 m .67915 .47711 L .67532 .4759 L F
.65602 .46884 m .67732 .47636 L .67532 .4759 L F
.69948 .49958 m .71473 .50345 L p .69403 .49101 L .536 .412 .659 r
F P
0 g
s
.72218 .49305 m .70326 .50552 L p .69365 .49041 L .499 .399 .672 r
F P
0 g
s
.68728 .48039 m .71747 .49134 L p .69365 .49041 L .499 .399 .672 r
F P
0 g
s
.71747 .49134 m .72218 .49305 L p .69365 .49041 L .499 .399 .672 r
F P
0 g
s
.782 .4954 m .78316 .4944 L .78343 .49438 L .138 0 0 r
F
.78316 .4944 m .78517 .49267 L .78343 .49438 L F
.78343 .49438 m .78517 .49267 L .78594 .4926 L F
.78706 .49453 m .782 .4954 L p .79147 .49207 L F P
0 g
s
.79147 .49207 m .782 .4954 L .78594 .4926 L .138 0 0 r
F
.72218 .49305 m .68728 .48039 L p .68423 .47916 L .211 .509 .91 r
F P
0 g
s
.72218 .49305 m .7213 .49188 L p .77972 .49308 L .488 .67 .945 r
F P
0 g
s
.77018 .49342 m .72218 .49305 L p .77972 .49308 L .488 .67 .945 r
F P
0 g
s
.72218 .49305 m .72513 .50157 L .72152 .50222 L .554 .503 .754 r
F
.7213 .49188 m .70986 .49289 L p .73032 .49316 L .27 0 0 r
F P
0 g
s
.64817 .48655 m .69948 .49958 L p .69403 .49101 L .536 .412 .659 r
F P
0 g
s
.70986 .49289 m .69453 .49425 L p .70738 .49386 L .27 0 0 r
F P
0 g
s
.70738 .49386 m .73032 .49316 L .70986 .49289 L .27 0 0 r
F
.62917 .46258 m .6153 .45513 L .61925 .46263 L .505 .35 .604 r
F
.6153 .45513 m .60982 .45329 L .61925 .46263 L F
.6074 .45089 m .58948 .44127 L .61925 .46263 L F
.6087 .4519 m .67222 .47493 L p .61783 .46245 L .499 .399 .672 r
F P
0 g
s
.64817 .48655 m .66902 .47562 L p .65086 .4664 L .505 .35 .604 r
F P
0 g
s
.61655 .46265 m .64817 .48655 L p .65086 .4664 L .505 .35 .604 r
F P
0 g
s
.65599 .48294 m .69263 .49073 L p .64225 .48086 L .436 0 0 r
F P
0 g
s
.64225 .48086 m .69263 .49073 L .69005 .4904 L .436 0 0 r
F
.64171 .50424 m .62429 .4921 L p .65269 .50335 L .499 .399 .672 r
F P
0 g
s
.69948 .49958 m .64817 .48655 L .62778 .50536 L p .504 .467 .749 r
F P
0 g
s
.64165 .49256 m .62068 .47839 L .62409 .49133 L .436 .359 .671 r
F
.64165 .49256 m .64817 .48655 L p .61806 .47012 L F P
0 g
s
.62377 .48205 m .65047 .49815 L .69453 .49425 L p .568 .188 .282 r
F P
0 g
s
.69453 .49425 m .70986 .49289 L p .69263 .49073 L .568 .188 .282 r
F P
0 g
s
.62377 .48205 m .69453 .49425 L .69263 .49073 L .568 .188 .282 r
F
.64311 .4802 m .66902 .47562 L p .68815 .48244 L .527 .173 .309 r
F P
0 g
s
.69263 .49073 m .64311 .4802 L p .68815 .48244 L .527 .173 .309 r
F P
0 g
s
.64225 .48086 m .69005 .4904 L .68746 .49007 L .436 0 0 r
F
.69453 .49425 m .65047 .49815 L p .70344 .49539 L .27 0 0 r
F P
0 g
s
.70344 .49539 m .70738 .49386 L .69453 .49425 L .27 0 0 r
F
.60993 .52183 m .6405 .49363 L p .59268 .48912 L .436 .359 .671 r
F P
0 g
s
.53951 .40379 m .52684 .36762 L .5419 .44276 L .127 .121 .58 r
F
.52684 .36762 m .52647 .36719 L .5419 .44276 L F
.56785 .49439 m .53295 .39877 L p .53938 .40232 L .328 .278 .649 r
F P
0 g
s
.56809 .4405 m .58776 .46664 L .5681 .44054 L .328 .278 .649 r
F
.54967 .4328 m .5432 .44869 L p .54692 .45277 L .287 .302 .701 r
F P
0 g
s
.57049 .46357 m .54967 .4328 L p .54692 .45277 L .287 .302 .701 r
F P
0 g
s
.57032 .4371 m .58578 .44139 L .58776 .46664 L .456 .323 .611 r
F
.58995 .44254 m .61655 .46265 L p .58948 .44127 L .505 .35 .604 r
F P
0 g
s
.58776 .46664 m .6087 .4519 L p .61591 .46023 L .499 .399 .672 r
F P
0 g
s
.62429 .4921 m .58776 .46664 L p .61591 .46023 L .499 .399 .672 r
F P
0 g
s
.54967 .4328 m .57049 .46357 L p .56835 .443 L .436 .359 .671 r
F P
0 g
s
.58776 .46664 m .58515 .47028 L p .56865 .44581 L .328 .278 .649 r
F P
0 g
s
.5681 .44054 m .58776 .46664 L .56865 .44581 L .328 .278 .649 r
F
.58515 .47028 m .56785 .49439 L p .55631 .45708 L F P
0 g
s
.53442 .38627 m .53851 .3832 L p .57032 .4371 L .456 .323 .611 r
F P
0 g
s
.58776 .46664 m .56766 .43636 L p .57032 .4371 L .456 .323 .611 r
F P
0 g
s
.56766 .43636 m .53442 .38627 L p .57032 .4371 L .456 .323 .611 r
F P
0 g
s
.56835 .443 m .56614 .43594 L .54967 .4328 L .436 .359 .671 r
F
.56145 .421 m .55078 .43168 L p .56614 .43594 L F P
0 g
s
.55078 .43168 m .54967 .4328 L p .56614 .43594 L .436 .359 .671 r
F P
0 g
s
.57476 .47971 m .56983 .45977 L .60878 .47302 L .631 .248 .299 r
F
.62963 .48011 m .60701 .47994 L p .60878 .47302 L F P
0 g
s
.60701 .47994 m .57476 .47971 L p .60878 .47302 L .631 .248 .299 r
F P
0 g
s
.6087 .4519 m .59654 .46046 L p .58924 .44062 L .456 .323 .611 r
F P
0 g
s
.58414 .42676 m .56789 .41741 L p .60416 .44269 L .505 .35 .604 r
F P
0 g
s
.58477 .42847 m .60625 .4495 L p .58924 .44062 L .456 .323 .611 r
F P
0 g
s
.60625 .4495 m .6087 .4519 L p .58924 .44062 L .456 .323 .611 r
F P
0 g
s
.66902 .47562 m .64311 .4802 L p .61757 .45888 L .58 .289 .439 r
F P
0 g
s
.65497 .46849 m .6535 .46669 L .65086 .4664 L .505 .35 .604 r
F
.65896 .46983 m .6535 .46669 L p .65497 .46849 L F P
0 g
s
.60828 .45137 m .64718 .46532 L p .65602 .46884 L .211 .509 .91 r
F P
0 g
s
.6087 .4519 m .60828 .45137 L p .65602 .46884 L .211 .509 .91 r
F P
0 g
s
.64776 .46606 m .6087 .4519 L p .65602 .46884 L .211 .509 .91 r
F P
0 g
s
.64718 .46532 m .60828 .45137 L p .6172 .4588 L .568 .188 .282 r
F P
0 g
s
.61757 .45888 m .60879 .45156 L .63852 .46344 L .58 .289 .439 r
F
.61757 .45888 m .63852 .46344 L .66902 .47562 L F
.61594 .45845 m .6172 .4588 L .60828 .45137 L .568 .188 .282 r
F
.6724 .47436 m .68653 .47942 L p .65602 .46884 L .211 .509 .91 r
F P
0 g
s
.68653 .47942 m .7213 .49188 L p .68168 .4779 L .211 .509 .91 r
F P
0 g
s
.68653 .47942 m .68168 .4779 L .65602 .46884 L .211 .509 .91 r
F
.68423 .47916 m .7213 .49188 L .71754 .49135 L F
.68863 .48332 m .69164 .48889 L .7213 .49188 L .568 .188 .282 r
F
.55462 .46121 m .54692 .45277 L .55798 .48327 L .287 .302 .701 r
F
.55806 .48856 m .54336 .45046 L .55085 .47633 L 0 .005 .571 r
F
.5432 .44869 m .54336 .45046 L p .55806 .48856 L F P
0 g
s
.55903 .49022 m .5432 .44869 L p .55806 .48856 L 0 .005 .571 r
F P
0 g
s
.5382 .42944 m .53795 .42843 L .54397 .45689 L 0 0 .299 r
F
.54429 .45157 m .58864 .56784 L p .54692 .45277 L .287 .302 .701 r
F P
0 g
s
.53146 .41896 m .53712 .45767 L p .53702 .4314 L .88 .643 .575 r
F P
0 g
s
.51931 .41623 m .5305 .47363 L .5186 .41491 L .768 .58 .644 r
F
.52462 .40698 m .5305 .47363 L p .51875 .41338 L F P
0 g
s
.53449 .51884 m .52664 .42991 L p .5305 .42113 L .835 .63 .625 r
F P
0 g
s
.53712 .45767 m .53519 .44585 L .51761 .45043 L .804 .59 .613 r
F
.51654 .45071 m .51737 .45099 L p .51761 .45043 L F P
0 g
s
.51737 .45099 m .53712 .45767 L p .51761 .45043 L .804 .59 .613 r
F P
0 g
s
.52378 .39749 m .52966 .40039 L p .5305 .42113 L .835 .63 .625 r
F P
0 g
s
.52664 .42991 m .52378 .39749 L p .5305 .42113 L .835 .63 .625 r
F P
0 g
s
.54336 .45046 m .5432 .44869 L p .5382 .42944 L 0 0 .299 r
F P
0 g
s
.5432 .44869 m .54429 .45157 L p .54692 .45277 L .287 .302 .701 r
F P
0 g
s
.60878 .47302 m .57751 .45417 L .5713 .45349 L p .592 .095 .044 r
F P
0 g
s
.5713 .45349 m .5419 .45027 L p .56092 .45674 L .592 .095 .044 r
F P
0 g
s
.56092 .45674 m .60878 .47302 L .5713 .45349 L .592 .095 .044 r
F
.55462 .46121 m .55798 .48327 L .57294 .52452 L .287 .302 .701 r
F
.57193 .52403 m .55903 .49022 L p .57232 .52552 L 0 .005 .571 r
F P
0 g
s
.56785 .49439 m .56355 .51987 L p .55454 .46068 L .009 .089 .6 r
F P
0 g
s
.54432 .46059 m .54788 .47128 L p .55076 .47617 L 0 .005 .571 r
F P
0 g
s
.54788 .47128 m .55271 .4858 L p .55076 .47617 L 0 .005 .571 r
F P
0 g
s
.53712 .45767 m .54833 .50437 L p .5499 .4934 L .869 .664 .623 r
F P
0 g
s
.52674 .47806 m .53028 .47389 L .51654 .45071 L .794 .621 .667 r
F
.54833 .50437 m .53712 .45767 L .51737 .45099 L p F P
0 g
s
.55013 .49176 m .55105 .48531 L .53931 .46341 L .869 .664 .623 r
F
.56092 .45674 m .57476 .47971 L .54665 .47364 L p .712 .271 .202 r
F P
0 g
s
.54665 .47364 m .52316 .46858 L p .53415 .46514 L .712 .271 .202 r
F P
0 g
s
.53415 .46514 m .56092 .45674 L .54665 .47364 L .712 .271 .202 r
F
.50745 .40708 m .48649 .47634 L .50256 .4154 L .706 .517 .631 r
F
.48846 .50918 m .50743 .42576 L p .51875 .41338 L .768 .58 .644 r
F P
0 g
s
.51654 .45071 m .49214 .44805 L p .49337 .45803 L .737 .57 .667 r
F P
0 g
s
.51338 .49381 m .51654 .45071 L p .49337 .45803 L .737 .57 .667 r
F P
0 g
s
.49214 .44805 m .50012 .41486 L .4997 .41531 L .667 .459 .595 r
F
.51656 .44986 m .51728 .415 L .50743 .42576 L .737 .533 .616 r
F
.51814 .41405 m .51654 .45071 L p .53519 .44585 L .804 .59 .613 r
F P
0 g
s
.49214 .44805 m .51654 .45071 L p .51656 .44986 L .737 .533 .616 r
F P
0 g
s
.50314 .41441 m .50256 .4154 L .50295 .41392 L .706 .517 .631 r
F
.50239 .41245 m .49214 .44805 L p .51656 .44986 L .737 .533 .616 r
F P
0 g
s
.49214 .44805 m .50166 .41323 L .5003 .41468 L .667 .459 .595 r
F
.51654 .45071 m .51814 .41405 L p .51728 .415 L .737 .533 .616 r
F P
0 g
s
.53415 .46514 m .52938 .4489 L .51704 .44472 L p .693 .176 0 r
F P
0 g
s
.51704 .44472 m .49137 .43601 L p .50351 .44427 L .693 .176 0 r
F P
0 g
s
.50351 .44427 m .53415 .46514 L .51704 .44472 L .693 .176 0 r
F
.49873 .40908 m .48215 .43767 L .49537 .41657 L 0 0 .107 r
F
.50046 .40523 m .49887 .40805 L .49873 .40908 L F
.49887 .40805 m .48215 .43767 L .49873 .40908 L F
.50528 .39449 m .50284 .3959 L p .49126 .41893 L .388 .147 .423 r
F P
0 g
s
.50528 .39449 m .48364 .43363 L p .50046 .40523 L 0 0 .107 r
F P
0 g
s
.49876 .40885 m .51268 .37132 L .50364 .40182 L 0 0 .051 r
F
.50915 .39094 m .49858 .41327 L .50469 .39944 L .754 .844 .394 r
F
.49798 .4146 m .50364 .40182 L .49641 .4262 L 0 0 .051 r
F
.49876 .40885 m .50364 .40182 L .49798 .4146 L F
.49798 .4146 m .48208 .44814 L p .49858 .41327 L .754 .844 .394 r
F P
0 g
s
.46848 .4619 m .46588 .46651 L .46551 .46765 L 0 0 .107 r
F
.45348 .48225 m .46594 .46632 L .46559 .4674 L .148 0 .202 r
F
.40065 .52124 m .43533 .47494 L .39459 .50978 L .905 .943 .821 r
F
.4612 .4745 m .47795 .45383 L .47432 .45211 L .948 .949 .642 r
F
.48208 .44814 m .48201 .44822 L .49858 .41327 L .754 .844 .394 r
F
.47795 .45383 m .47812 .45358 L p .47432 .45211 L .948 .949 .642 r
F P
0 g
s
.47812 .45358 m .48344 .44161 L .47432 .45211 L .948 .949 .642 r
F
.4634 .47529 m .47795 .45383 L p .4612 .4745 L F P
0 g
s
.51268 .37132 m .50512 .39848 L .50366 .40178 L 0 0 .051 r
F
.50364 .40182 m .49641 .42627 L .49641 .4262 L F
.51268 .37132 m .50366 .40178 L .50364 .40182 L F
.50366 .40178 m .49641 .42627 L .50364 .40182 L F
.50974 .38982 m .51213 .37854 L p .51213 .37851 L .141 0 .344 r
F P
0 g
s
.51213 .37854 m .50974 .38982 L p .51147 .38186 L 0 0 .016 r
F P
0 g
s
.50974 .38982 m .51007 .38893 L p .51147 .38186 L 0 0 .016 r
F P
0 g
s
.50915 .39094 m .50469 .39944 L .51007 .38893 L .754 .844 .394 r
F
.47432 .45211 m .48344 .44161 L .49014 .42653 L .948 .949 .642 r
F
.47569 .44799 m .4793 .44223 L .49771 .41004 L 0 0 .107 r
F
.47569 .44799 m .49771 .41004 L .49729 .41071 L F
.50284 .3959 m .49657 .40756 L .49692 .40768 L .388 .147 .423 r
F
.49657 .40756 m .4904 .41903 L .49692 .40768 L F
.49692 .40768 m .4904 .41903 L .49126 .41893 L F
.49584 .40729 m .50912 .3801 L .49878 .40834 L .278 .05 .378 r
F
.49793 .40969 m .49805 .40949 L .49887 .40805 L 0 0 .107 r
F
.49584 .40729 m .49878 .40834 L .49316 .41725 L .278 .05 .378 r
F
.49692 .40327 m .50912 .3801 L .49584 .40729 L F
.4904 .41903 m .48988 .42125 L .49126 .41893 L .388 .147 .423 r
F
.48945 .42312 m .49692 .40327 L .49316 .41725 L .278 .05 .378 r
F
.49096 .41662 m .49692 .40327 L .48945 .42312 L F
.4904 .41903 m .48249 .43373 L .48988 .42125 L .388 .147 .423 r
F
.48364 .43363 m .47569 .44799 L p .49729 .41071 L 0 0 .107 r
F P
0 g
s
.48364 .43363 m .49729 .41071 L .49612 .41256 L 0 0 .107 r
F
.49692 .40327 m .5032 .38919 L .50024 .39638 L .278 .05 .378 r
F
.49692 .40327 m .5088 .37556 L .5032 .38919 L F
.49358 .41018 m .5088 .37556 L p .49692 .40327 L F P
0 g
s
.50144 .39604 m .49014 .42653 L .48154 .42686 L .987 .882 .54 r
F
.50495 .39871 m .51475 .36779 L .51567 .36417 L .863 .83 .356 r
F
.50512 .39848 m .51475 .36779 L p .50495 .39871 L F P
0 g
s
.51228 .37159 m .49884 .40824 L p .51229 .37143 L .278 .05 .378 r
F P
0 g
s
.51229 .37143 m .49884 .40824 L .49878 .40834 L .278 .05 .378 r
F
.49876 .40885 m .50931 .37969 L .51268 .37132 L 0 0 .051 r
F
.49884 .40824 m .50931 .37969 L p .49876 .40885 L F P
0 g
s
.50535 .39229 m .49014 .42653 L .50144 .39604 L .987 .882 .54 r
F
.50288 .40157 m .49014 .42653 L p .50535 .39229 L F P
0 g
s
.51483 .36627 m .51481 .36651 L .5189 .34774 L 0 0 .016 r
F
.50495 .39871 m .51567 .36417 L .50575 .39079 L .863 .83 .356 r
F
.50569 .39103 m .50535 .39229 L .50144 .39604 L .987 .882 .54 r
F
.50495 .39871 m .50575 .39079 L .50288 .40157 L .863 .83 .356 r
F
.51617 .36217 m .51618 .36213 L .51943 .34499 L 0 0 .016 r
F
.51943 .34499 m .51943 .34499 L .51976 .34328 L F
.51617 .36217 m .51943 .34499 L .51943 .34499 L F
.51486 .36564 m .52097 .33683 L p .51485 .36566 L .141 0 .344 r
F P
0 g
s
.52066 .33813 m .51484 .3657 L .51977 .34249 L .141 0 .344 r
F
.52097 .33683 m .51974 .34262 L p .5189 .34774 L 0 0 .016 r
F P
0 g
s
.51976 .34328 m .51943 .345 L p .51999 .34123 L .941 .796 .337 r
F P
0 g
s
.5146 .3705 m .51943 .345 L p .5146 .37048 L 0 0 .016 r
F P
0 g
s
.51943 .345 m .51976 .34328 L p .51943 .34499 L 0 0 .016 r
F P
0 g
s
.5146 .37048 m .51943 .345 L .51943 .34499 L 0 0 .016 r
F
.51922 .34632 m .52096 .33799 L .518 .35601 L .336 .137 .454 r
F
.52107 .33735 m .52107 .33735 L .52105 .33746 L F
.52105 .33746 m .52105 .33746 L .518 .35601 L F
.52107 .33735 m .52105 .33746 L .52105 .33746 L F
.52247 .35751 m .52314 .35796 L p .5222 .35356 L .886 .627 .54 r
F P
0 g
s
.52312 .35778 m .52106 .3375 L p .52134 .34029 L .977 .7 .306 r
F P
0 g
s
.52126 .33947 m .52106 .3375 L .52112 .33732 L .808 .565 .571 r
F
.52247 .35751 m .52106 .3375 L p .52126 .33947 L F P
0 g
s
.5211 .33783 m .52109 .33775 L .5211 .33789 L .808 .565 .571 r
F
.52112 .33739 m .5211 .33789 L .52111 .33794 L F
.52109 .33775 m .52106 .33762 L .5211 .33789 L F
.52108 .3377 m .52106 .3375 L .52112 .33812 L .977 .7 .306 r
F
.52108 .33775 m .52108 .3377 L .52112 .33812 L F
.52112 .33732 m .52106 .33762 L .5211 .33783 L .808 .565 .571 r
F
.52112 .33819 m .52106 .3375 L .52114 .33745 L .736 .504 .577 r
F
.52106 .33762 m .52064 .35686 L p .5211 .33789 L .808 .565 .571 r
F P
0 g
s
.52064 .35686 m .52106 .3375 L p .52112 .33819 L .736 .504 .577 r
F P
0 g
s
.52064 .35686 m .52112 .33819 L .52106 .3375 L .736 .504 .577 r
F
.52103 .33807 m .52108 .33774 L .52106 .3375 L .66 .434 .566 r
F
.52107 .3376 m .52106 .3375 L .52108 .3377 L .977 .7 .306 r
F
.52121 .33677 m .52121 .33677 L p .5212 .33682 L .945 .734 .586 r
F P
0 g
s
.5212 .33682 m .52121 .33677 L .5212 .33682 L .945 .734 .586 r
F
.52107 .33749 m .52114 .33713 L .52107 .33746 L F
.52107 .33749 m .52107 .33746 L .52107 .33749 L F
.52112 .33739 m .52112 .33732 L .52111 .33741 L .808 .565 .571 r
F
.52106 .3375 m .52106 .33762 L p .52112 .33732 L F P
0 g
s
.52112 .33819 m .52115 .3371 L .52106 .3375 L .736 .504 .577 r
F
.52108 .33775 m .52107 .33761 L .52108 .3377 L .977 .7 .306 r
F
.52107 .33761 m .52107 .3376 L .52108 .3377 L F
.52103 .33807 m .52106 .3375 L .52108 .33774 L .66 .434 .566 r
F
.52105 .33759 m .52106 .3375 L .52107 .33764 L .61 .364 .52 r
F
.52104 .33775 m .52106 .33761 L .52106 .33762 L .987 .767 .373 r
F
.52108 .33838 m .52106 .3375 L .52103 .33784 L .687 .415 .505 r
F
.52107 .33761 m .52106 .33755 L .52107 .3376 L .977 .7 .306 r
F
.52107 .33763 m .52106 .3375 L .52106 .33758 L .61 .364 .52 r
F
.52106 .33755 m .52106 .3375 L .52107 .3376 L .977 .7 .306 r
F
.52107 .33767 m .52106 .3375 L .52107 .33763 L .61 .364 .52 r
F
.52106 .33754 m .52106 .3375 L .52106 .33755 L .977 .7 .306 r
F
.52106 .33754 m .52106 .3375 L .52106 .33757 L .687 .415 .505 r
F
.52103 .33784 m .52107 .33788 L .52107 .33805 L F
.52064 .35686 m .52247 .35751 L p .52126 .33947 L .808 .565 .571 r
F P
0 g
s
.52104 .33773 m .52106 .33757 L .52107 .33788 L .687 .415 .505 r
F
.52106 .3375 m .52301 .35667 L p .52109 .33763 L 0 0 .452 r
F P
0 g
s
.52134 .34029 m .52108 .33775 L .52112 .33812 L .977 .7 .306 r
F
.52306 .35552 m .52305 .35548 L .52114 .33703 L .844 .604 .575 r
F
.52106 .3375 m .52458 .35412 L p .52106 .3375 L .464 .275 .539 r
F P
0 g
s
.52106 .3375 m .52356 .3552 L p .52114 .33785 L .311 .178 .533 r
F P
0 g
s
.52106 .3375 m .5212 .33682 L .52106 .3375 L .945 .734 .586 r
F
.52121 .33677 m .5212 .33682 L p .5212 .33682 L F P
0 g
s
.5212 .33682 m .52106 .3375 L p .5212 .33682 L .945 .734 .586 r
F P
0 g
s
.52458 .35412 m .52513 .3538 L p .52107 .33744 L .464 .275 .539 r
F P
0 g
s
.52107 .33744 m .52106 .3375 L .52458 .35412 L .464 .275 .539 r
F
.52106 .3375 m .52513 .3538 L p .5214 .33895 L .61 .364 .52 r
F P
0 g
s
.52134 .34029 m .52106 .3375 L .52106 .33754 L .977 .7 .306 r
F
.52108 .33774 m .52106 .3375 L .52109 .3377 L .66 .434 .566 r
F
.52356 .3552 m .52458 .35412 L p .52155 .3396 L .311 .178 .533 r
F P
0 g
s
.52513 .3538 m .52134 .3386 L p .52122 .33806 L .464 .275 .539 r
F P
0 g
s
.52134 .3386 m .52106 .3375 L p .52122 .33806 L .464 .275 .539 r
F P
0 g
s
.52126 .33947 m .52112 .33739 L .52111 .33794 L .808 .565 .571 r
F
.52122 .33806 m .52106 .3375 L .52107 .33745 L .464 .275 .539 r
F
.52155 .3396 m .52108 .33737 L .52115 .33789 L .311 .178 .533 r
F
.52155 .3396 m .52115 .33789 L .52356 .3552 L F
.52155 .3396 m .52106 .3375 L .52111 .33752 L F
.52458 .35412 m .52106 .3375 L p .52155 .3396 L F P
0 g
s
.52123 .33846 m .52114 .33792 L .5214 .33895 L .61 .364 .52 r
F
.52176 .34055 m .5214 .33895 L .52513 .3538 L F
.52123 .33846 m .5214 .33895 L .52176 .34055 L F
.52442 .35393 m .52123 .33846 L .52176 .34055 L F
.52513 .3538 m .52483 .35386 L p .52176 .34055 L F P
0 g
s
.52483 .35386 m .52442 .35393 L p .52176 .34055 L .61 .364 .52 r
F P
0 g
s
.52107 .33764 m .52106 .3375 L .5214 .33895 L .61 .364 .52 r
F
.52108 .33737 m .52108 .33738 L .52356 .3552 L .311 .178 .533 r
F
.52442 .35393 m .52106 .3375 L p .52121 .33834 L .61 .364 .52 r
F P
0 g
s
.52108 .33774 m .52109 .3377 L .52106 .3375 L .66 .434 .566 r
F
.52114 .33745 m .52106 .3375 L .52115 .3371 L .736 .504 .577 r
F
.52107 .33745 m .52114 .33713 L .52114 .33713 L .945 .734 .586 r
F
.52107 .33746 m .52114 .33713 L .52107 .33745 L F
.52106 .3375 m .52121 .33677 L p .52109 .33738 L F P
0 g
s
.52121 .33677 m .52106 .3375 L p .52109 .3377 L .66 .434 .566 r
F P
0 g
s
.52109 .3377 m .52106 .3375 L .52121 .33677 L .66 .434 .566 r
F
.52108 .33738 m .52106 .3375 L .52114 .33785 L .311 .178 .533 r
F
.52111 .33764 m .52108 .33754 L .52109 .33763 L 0 0 .452 r
F
.52114 .33791 m .52109 .33763 L .52301 .35667 L F
.52111 .33764 m .52109 .33763 L .52114 .33791 L F
.52118 .33817 m .52114 .33791 L .52301 .35667 L F
.52107 .33805 m .52107 .33788 L .52154 .34013 L .687 .415 .505 r
F
.52172 .34113 m .52154 .34013 L .52442 .35393 L F
.52356 .3552 m .52106 .3375 L p .52119 .33821 L 0 0 .452 r
F P
0 g
s
.52121 .33834 m .52106 .3375 L .52107 .33767 L .61 .364 .52 r
F
.52109 .33763 m .52106 .3375 L .52109 .33763 L 0 0 .452 r
F
.52106 .3375 m .52442 .35393 L p .52107 .33762 L .687 .415 .505 r
F P
0 g
s
.52109 .33774 m .52106 .3375 L .52108 .33765 L 0 0 .073 r
F
.52301 .35667 m .52106 .3375 L p .52109 .33774 L F P
0 g
s
.52116 .33849 m .52154 .34013 L .52172 .34113 L .687 .415 .505 r
F
.52107 .33805 m .52116 .33844 L .52116 .33849 L F
.52301 .35667 m .52356 .3552 L p .52119 .33821 L 0 0 .452 r
F P
0 g
s
.52116 .33844 m .52154 .34013 L .52116 .33849 L .687 .415 .505 r
F
.52119 .33821 m .52118 .33817 L .52119 .33819 L 0 0 .452 r
F
.52119 .33821 m .52119 .33819 L .52301 .35667 L F
.52113 .33696 m .52116 .33756 L .52124 .33743 L .724 .426 .472 r
F
.51905 .35327 m .52106 .33757 L .52104 .33775 L .987 .767 .373 r
F
.52253 .35392 m .52103 .33784 L .52116 .33849 L .687 .415 .505 r
F
.52219 .34862 m .52116 .33849 L .52172 .34113 L F
.52301 .35667 m .52109 .33774 L .52171 .34389 L 0 0 .073 r
F
.52109 .33774 m .52108 .33768 L .52171 .34389 L F
.52106 .3375 m .52171 .34389 L p .52106 .3375 L F P
0 g
s
.52253 .35392 m .52116 .33849 L .52219 .34862 L .687 .415 .505 r
F
.52123 .33846 m .52105 .33759 L .52114 .33792 L .61 .364 .52 r
F
.52123 .33846 m .52106 .33762 L .52106 .33764 L F
.52107 .33788 m .52106 .33757 L .52117 .33807 L .687 .415 .505 r
F
.52109 .33774 m .52106 .3375 L .52108 .33768 L 0 0 .073 r
F
.52253 .35392 m .52106 .3375 L p .52108 .33838 L .687 .415 .505 r
F P
0 g
s
.52442 .35393 m .52121 .33834 L .52123 .33846 L .61 .364 .52 r
F
.52154 .34013 m .52117 .33807 L .52442 .35393 L .687 .415 .505 r
F
.52121 .33834 m .52106 .33762 L .52123 .33846 L .61 .364 .52 r
F
.52107 .33788 m .52117 .33807 L .52154 .34013 L .687 .415 .505 r
F
.52301 .35667 m .52373 .35996 L p .52357 .35827 L .127 .121 .58 r
F P
0 g
s
.52373 .35996 m .52457 .3638 L p .52357 .35827 L .127 .121 .58 r
F P
0 g
s
.52379 .35889 m .52357 .35827 L .52457 .3638 L .127 .121 .58 r
F
.52171 .34389 m .52312 .35778 L .52309 .3575 L p 0 0 .073 r
F P
0 g
s
.52309 .3575 m .52301 .35667 L p .52171 .34389 L 0 0 .073 r
F P
0 g
s
.52247 .35751 m .52386 .36697 L p .52538 .37082 L .88 .643 .575 r
F P
0 g
s
.52247 .35751 m .52064 .35686 L p .52291 .37072 L .804 .59 .613 r
F P
0 g
s
.52514 .37576 m .52247 .35751 L p .52291 .37072 L .804 .59 .613 r
F P
0 g
s
.52301 .35667 m .52312 .35778 L p .52372 .36051 L 0 0 .299 r
F P
0 g
s
.53442 .38627 m .53938 .40232 L .53177 .39358 L .328 .278 .649 r
F
.52628 .36696 m .52839 .37947 L .52625 .36694 L .217 .138 .547 r
F
.53554 .41403 m .52489 .36528 L .52375 .36064 L 0 0 .299 r
F
.53534 .41291 m .52492 .36538 L .53554 .41403 L F
.52492 .36538 m .52422 .3622 L .52489 .36528 L F
.52492 .36538 m .52489 .36528 L .53554 .41403 L F
.53442 .38627 m .53125 .39108 L p .52821 .37838 L .217 .138 .547 r
F P
0 g
s
.5419 .44276 m .52422 .3622 L p .54202 .44334 L 0 0 .299 r
F P
0 g
s
.53125 .39108 m .53054 .39217 L p .52839 .37947 L .217 .138 .547 r
F P
0 g
s
.52821 .37838 m .53125 .39108 L .52839 .37947 L .217 .138 .547 r
F
.53054 .39217 m .52836 .37934 L p .52839 .37947 L F P
0 g
s
.52836 .37934 m .52625 .36694 L p .52839 .37947 L .217 .138 .547 r
F P
0 g
s
.52375 .36064 m .52422 .3622 L .52372 .36051 L 0 0 .299 r
F
.52489 .36528 m .52422 .3622 L .52375 .36064 L F
.54202 .44334 m .53534 .41291 L .53554 .41403 L F
.53433 .40722 m .54046 .41936 L .53592 .41356 L .009 .089 .6 r
F
.53054 .39217 m .53269 .39807 L p .53433 .40722 L F P
0 g
s
.53269 .39807 m .54046 .41936 L p .53433 .40722 L .009 .089 .6 r
F P
0 g
s
.53054 .39217 m .53125 .39108 L p .53177 .39358 L .328 .278 .649 r
F P
0 g
s
.53125 .39108 m .53442 .38627 L p .53177 .39358 L .328 .278 .649 r
F P
0 g
s
.5432 .44869 m .5419 .44276 L p .54202 .44334 L 0 0 .299 r
F P
0 g
s
.52422 .3622 m .52301 .35667 L p .52372 .36051 L 0 0 .299 r
F P
0 g
s
.53942 .40234 m .54967 .4328 L p .54046 .40292 L .369 .26 .597 r
F P
0 g
s
.54967 .4328 m .53942 .40234 L p .53951 .40379 L .127 .121 .58 r
F P
0 g
s
.55078 .43168 m .56145 .421 L p .55602 .41157 L .369 .26 .597 r
F P
0 g
s
.55602 .41157 m .54728 .39639 L .548 .40368 L .369 .26 .597 r
F
.55602 .41157 m .548 .40368 L .55078 .43168 L F
.54052 .40295 m .54046 .40292 L .54967 .4328 L F
.56153 .42615 m .54559 .42113 L .55975 .41891 L .631 .248 .299 r
F
.54859 .42973 m .53951 .40379 L .54082 .41982 L .127 .121 .58 r
F
.5419 .44276 m .5432 .44869 L p .53951 .40379 L F P
0 g
s
.54011 .41876 m .55975 .41891 L p .54559 .42113 L .631 .248 .299 r
F P
0 g
s
.61591 .46023 m .61815 .46282 L .61655 .46265 L .499 .399 .672 r
F
.61089 .46477 m .61221 .4596 L .64311 .4802 L .631 .248 .299 r
F
.55654 .40082 m .5558 .4001 L .56145 .421 L .481 .306 .566 r
F
.56145 .421 m .54715 .39506 L p .54728 .39639 L .369 .26 .597 r
F P
0 g
s
.55975 .41891 m .56789 .41741 L p .58558 .43275 L .58 .289 .439 r
F P
0 g
s
.58995 .44254 m .56646 .42479 L p .56766 .43636 L .436 .359 .671 r
F P
0 g
s
.6087 .4519 m .58477 .42847 L p .58411 .42782 L 0 .077 .608 r
F P
0 g
s
.59365 .45271 m .60828 .45137 L p .60127 .44513 L .622 .315 .429 r
F P
0 g
s
.58509 .43071 m .58857 .43966 L .60127 .44513 L .622 .315 .429 r
F
.58528 .43901 m .58995 .44254 L p .58948 .44127 L .505 .35 .604 r
F P
0 g
s
.58857 .43966 m .59365 .45271 L .60127 .44513 L .622 .315 .429 r
F
.56646 .42479 m .56145 .421 L p .56835 .443 L .436 .359 .671 r
F P
0 g
s
.58261 .43572 m .5871 .43902 L p .56099 .42395 L .631 .248 .299 r
F P
0 g
s
.56099 .42395 m .5871 .43902 L .576 .43424 L .631 .248 .299 r
F
.58617 .4352 m .55975 .41891 L .58558 .43275 L .58 .289 .439 r
F
.53255 .36577 m .53691 .37469 L .54792 .39547 L .481 .306 .566 r
F
.53851 .3832 m .54889 .39598 L .54568 .39429 L .456 .323 .611 r
F
.56789 .41741 m .55975 .41891 L p .54648 .39396 L .6 .339 .493 r
F P
0 g
s
.53016 .37029 m .53851 .3832 L .53194 .37697 L .415 .257 .557 r
F
.53818 .3823 m .53851 .3832 L p .53016 .37029 L F P
0 g
s
.53842 .38309 m .53851 .3832 L p .53255 .36719 L .891 .481 .065 r
F P
0 g
s
.5414 .38618 m .53588 .37551 L .53729 .37904 L .369 .26 .597 r
F
.5385 .38316 m .53255 .3672 L p .5385 .38316 L .891 .481 .065 r
F P
0 g
s
.53255 .36719 m .5385 .38316 L .53255 .3672 L .891 .481 .065 r
F
.52832 .36933 m .53442 .38627 L .52821 .37838 L .217 .138 .547 r
F
.52513 .3538 m .52762 .3575 L p .52765 .35772 L .6 .339 .493 r
F P
0 g
s
.53075 .36437 m .52513 .3538 L .53084 .36267 L .6 .339 .493 r
F
.52458 .35412 m .52356 .3552 L p .52854 .36385 L .369 .26 .597 r
F P
0 g
s
.52838 .36372 m .53059 .37198 L p .52832 .36933 L .217 .138 .547 r
F P
0 g
s
.53059 .37198 m .53442 .38627 L p .52832 .36933 L .217 .138 .547 r
F P
0 g
s
.52782 .35886 m .52565 .35572 L .52525 .35403 L .6 .339 .493 r
F
.52565 .35572 m .52442 .35393 L .52525 .35403 L F
.52442 .35393 m .52513 .3538 L p .52525 .35403 L F P
0 g
s
.52789 .35928 m .52442 .35393 L .52782 .35886 L .6 .339 .493 r
F
.52996 .36385 m .52442 .35393 L .52879 .36068 L F
.52566 .36144 m .52832 .36933 L p .52838 .36372 L .369 .26 .597 r
F P
0 g
s
.52356 .3552 m .52566 .36144 L p .52854 .36385 L .369 .26 .597 r
F P
0 g
s
.52719 .36232 m .53026 .37514 L .52721 .36228 L .724 .426 .472 r
F
.52442 .35393 m .5281 .36069 L p .52722 .36227 L .673 .365 .443 r
F P
0 g
s
.52646 .36363 m .52593 .36109 L .52722 .36227 L .673 .365 .443 r
F
.52593 .36109 m .52442 .35393 L .52722 .36227 L F
.53442 .38627 m .53151 .37539 L p .5315 .37534 L .415 .257 .557 r
F P
0 g
s
.53151 .37539 m .52838 .36372 L p .5315 .37534 L .415 .257 .557 r
F P
0 g
s
.52841 .36374 m .5315 .37534 L .52838 .36372 L .415 .257 .557 r
F
.52722 .36227 m .5323 .38367 L p .53055 .36534 L .642 .374 .498 r
F P
0 g
s
.53023 .37498 m .52722 .36227 L p .53026 .37514 L .724 .426 .472 r
F P
0 g
s
.53247 .3671 m .53069 .36572 L .53842 .38309 L .642 .374 .498 r
F
.53084 .36273 m .5307 .36548 L .53247 .3671 L F
.5307 .36548 m .53069 .36572 L .53247 .3671 L F
.53842 .38309 m .54665 .39112 L p .54722 .3917 L 0 .077 .608 r
F P
0 g
s
.53842 .38309 m .53211 .3661 L p .53204 .36596 L .642 .374 .498 r
F P
0 g
s
.53233 .36664 m .53842 .38309 L .53255 .36719 L .891 .481 .065 r
F
.53211 .3661 m .53249 .36713 L p .53233 .36664 L F P
0 g
s
.53249 .36713 m .53842 .38309 L p .53233 .36664 L .891 .481 .065 r
F P
0 g
s
.5307 .36548 m .52442 .35393 L p .53071 .3652 L .6 .339 .493 r
F P
0 g
s
.53084 .36267 m .52765 .35772 L .53054 .36185 L .6 .339 .493 r
F
.52762 .3575 m .53054 .36185 L p .52765 .35772 L F P
0 g
s
.54079 .38326 m .52996 .36385 L .52879 .36068 L .6 .339 .493 r
F
.53012 .3608 m .53056 .36543 L .53069 .36572 L .642 .374 .498 r
F
.53012 .3608 m .53069 .36572 L .53084 .36273 L F
.53851 .3832 m .53442 .38627 L p .53194 .37697 L .415 .257 .557 r
F P
0 g
s
.53639 .38328 m .53842 .38309 L p .53056 .36543 L .642 .374 .498 r
F P
0 g
s
.5341 .37424 m .53052 .36531 L .53408 .37179 L .673 .365 .443 r
F
.53557 .37616 m .53408 .37179 L .53884 .38045 L F
.5341 .37424 m .53408 .37179 L .53557 .37616 L F
.54677 .39128 m .53851 .3832 L p .54682 .39133 L 0 .077 .608 r
F P
0 g
s
.5392 .38384 m .54722 .3917 L .53851 .3832 L 0 .077 .608 r
F
.54722 .3917 m .54724 .39172 L .53851 .3832 L F
.54648 .39396 m .55035 .39629 L .56789 .41741 L .6 .339 .493 r
F
.53851 .3832 m .54733 .39684 L .54748 .39839 L .456 .323 .611 r
F
.54743 .3919 m .54682 .39133 L .53851 .3832 L 0 .077 .608 r
F
.53851 .3832 m .54677 .39128 L p .54701 .39366 L .456 .323 .611 r
F P
0 g
s
.54677 .39128 m .54807 .39255 L p .54701 .39366 L .456 .323 .611 r
F P
0 g
s
.55173 .40099 m .55858 .40643 L .54637 .39229 L .622 .315 .429 r
F
.5461 .39342 m .54637 .39229 L .53842 .38309 L F
.54653 .39399 m .54637 .39229 L .5461 .39342 L F
.55975 .41891 m .54604 .39369 L p .54648 .39396 L .6 .339 .493 r
F P
0 g
s
.54604 .39369 m .55975 .41891 L p .54597 .39397 L .673 .365 .443 r
F P
0 g
s
.54605 .39408 m .55173 .40099 L .5461 .39342 L .622 .315 .429 r
F
.54597 .39397 m .5461 .39342 L .5412 .38682 L F
.54605 .39408 m .5461 .39342 L .54597 .39397 L F
.53442 .38627 m .54836 .40728 L p .53938 .40232 L .328 .278 .649 r
F P
0 g
s
.55836 .4189 m .54011 .41876 L p .54407 .40202 L .673 .365 .443 r
F P
0 g
s
.54407 .40202 m .54597 .39397 L .55331 .40874 L .673 .365 .443 r
F
.54407 .40202 m .55331 .40874 L .55836 .4189 L F
.53248 .39013 m .53739 .39717 L .54032 .39618 L .699 .333 .35 r
F
.53739 .39717 m .54352 .40162 L .54032 .39618 L F
.53835 .38906 m .53793 .39001 L .54549 .39602 L .622 .315 .429 r
F
.55975 .41891 m .55836 .4189 L p .54597 .39397 L .673 .365 .443 r
F P
0 g
s
.5456 .39553 m .54549 .39602 L .55173 .40099 L .622 .315 .429 r
F
.53835 .38906 m .54549 .39602 L .5456 .39553 L F
.55975 .41891 m .58261 .43572 L p .56099 .42395 L .631 .248 .299 r
F P
0 g
s
.58888 .44641 m .59365 .45271 L .58509 .43071 L .622 .315 .429 r
F
.58261 .43572 m .55975 .41891 L p .58617 .4352 L .58 .289 .439 r
F P
0 g
s
.58673 .43752 m .58617 .4352 L .59775 .44233 L .58 .289 .439 r
F
.58261 .43572 m .58617 .4352 L .58673 .43752 L F
.57084 .43201 m .56099 .42395 L .576 .43424 L .631 .248 .299 r
F
.58571 .45343 m .59047 .45299 L p .55968 .41448 L .622 .315 .429 r
F P
0 g
s
.59047 .45299 m .59365 .45271 L p .58918 .44768 L .622 .315 .429 r
F P
0 g
s
.55968 .41448 m .59047 .45299 L .58918 .44768 L .622 .315 .429 r
F
.54549 .39602 m .58918 .44768 L .55173 .40099 L F
.53793 .39001 m .54502 .39799 L .54549 .39602 L F
.54502 .39799 m .58918 .44768 L .54549 .39602 L F
.54352 .40162 m .54395 .40194 L .54032 .39618 L .699 .333 .35 r
F
.54011 .41876 m .53804 .41875 L p .54352 .40162 L .673 .365 .443 r
F P
0 g
s
.5323 .38367 m .55061 .40759 L .54574 .3988 L .622 .315 .429 r
F
.5432 .44869 m .54967 .4328 L p .54859 .42973 L .127 .121 .58 r
F P
0 g
s
.54859 .42973 m .54082 .41982 L .5432 .44869 L .127 .121 .58 r
F
.55059 .43959 m .53804 .41875 L p .54559 .42113 L .631 .248 .299 r
F P
0 g
s
.56042 .42752 m .54559 .42113 L .56153 .42615 L .631 .248 .299 r
F
.55059 .43959 m .54559 .42113 L .56042 .42752 L F
.57751 .45417 m .58571 .45343 L p .56856 .43103 L .622 .315 .429 r
F P
0 g
s
.56204 .42822 m .56042 .42752 L .56153 .42615 L .631 .248 .299 r
F
.56204 .42822 m .56099 .42395 L .57084 .43201 L F
.55061 .40759 m .56268 .4285 L .56856 .43103 L .622 .315 .429 r
F
.56268 .4285 m .57751 .45417 L .56856 .43103 L F
.5323 .38367 m .56042 .42752 L p .55061 .40759 L F P
0 g
s
.56042 .42752 m .56945 .4416 L p .56204 .42822 L .622 .315 .429 r
F P
0 g
s
.56042 .42752 m .56204 .42822 L .55061 .40759 L .622 .315 .429 r
F
.54352 .40162 m .54407 .40202 L .54011 .41876 L .673 .365 .443 r
F
.53804 .41875 m .54163 .40025 L .54352 .40162 L F
.53295 .39877 m .53054 .39217 L p .53938 .40232 L .328 .278 .649 r
F P
0 g
s
.53378 .39454 m .53739 .39717 L .53248 .39013 L .699 .333 .35 r
F
.53026 .37514 m .52722 .36227 L .52721 .36228 L .724 .426 .472 r
F
.5323 .38367 m .53023 .37498 L p .53026 .37514 L F P
0 g
s
.5323 .38367 m .53639 .38328 L p .53056 .36543 L .642 .374 .498 r
F P
0 g
s
.5323 .38367 m .53564 .38743 L .53835 .38906 L .622 .315 .429 r
F
.53564 .38743 m .53793 .39001 L .53835 .38906 L F
.52365 .34751 m .53152 .38358 L .53026 .37514 L .724 .426 .472 r
F
.52327 .34656 m .53152 .38358 L .52365 .34751 L F
.52442 .35393 m .52404 .35393 L p .52172 .34113 L .687 .415 .505 r
F P
0 g
s
.52836 .36947 m .52832 .36933 L p .52824 .36925 L .127 .121 .58 r
F P
0 g
s
.52832 .36933 m .5266 .36424 L p .52823 .36924 L .127 .121 .58 r
F P
0 g
s
.52832 .36933 m .52823 .36924 L .52824 .36925 L .127 .121 .58 r
F
.5266 .36424 m .52593 .36225 L p .52656 .36415 L F P
0 g
s
.52593 .36225 m .52356 .3552 L p .52595 .36247 L .127 .121 .58 r
F P
0 g
s
.52625 .36694 m .52836 .37934 L p .52642 .36942 L 0 0 .394 r
F P
0 g
s
.52628 .36696 m .52832 .36933 L .52821 .37838 L .217 .138 .547 r
F
.52684 .36762 m .52617 .3657 L .52647 .36719 L .127 .121 .58 r
F
.52617 .3657 m .52625 .36694 L .52647 .36719 L F
.52836 .37934 m .53054 .39217 L p .52775 .37721 L 0 0 .394 r
F P
0 g
s
.52704 .37656 m .52684 .3756 L .52684 .37557 L .955 .657 .223 r
F
.52704 .37656 m .52679 .37505 L .52817 .38185 L F
.52679 .37505 m .52676 .37471 L .52817 .38185 L F
.52554 .36893 m .5261 .37187 L .52979 .38924 L 0 0 .299 r
F
.52979 .38924 m .5261 .37187 L .52763 .37965 L F
.52688 .37602 m .52763 .37965 L p .5261 .37187 L F P
0 g
s
.52724 .37791 m .52817 .38599 L .5307 .40122 L .904 .636 .132 r
F
.52702 .37674 m .52817 .38599 L .52724 .37791 L F
.53692 .42425 m .53723 .42499 L .53633 .42183 L .773 .477 0 r
F
.52907 .38602 m .53063 .39383 L .53795 .42843 L 0 0 .299 r
F
.52907 .38602 m .53795 .42843 L .5382 .42944 L F
.5432 .44869 m .52907 .38602 L .5382 .42944 L F
.52907 .38602 m .52763 .37965 L .53063 .39383 L F
.52763 .37965 m .53316 .40647 L p .53063 .39383 L F P
0 g
s
.5267 .37351 m .53054 .39217 L .52713 .37974 L 0 0 .394 r
F
.52649 .37043 m .52775 .37721 L .5267 .37351 L F
.52775 .37721 m .53054 .39217 L .5267 .37351 L F
.52601 .37107 m .52554 .36893 L .52593 .3708 L 0 0 .299 r
F
.52582 .3704 m .52713 .3798 L p .52677 .37541 L .904 .636 .132 r
F P
0 g
s
.52583 .37044 m .5267 .37351 L .52713 .37974 L 0 0 .394 r
F
.52713 .3798 m .52737 .38148 L p .52732 .38095 L .904 .636 .132 r
F P
0 g
s
.52737 .38148 m .52582 .3704 L p .52741 .38174 L 0 0 .394 r
F P
0 g
s
.52741 .38174 m .52582 .3704 L .52582 .37041 L 0 0 .394 r
F
.52583 .37044 m .52741 .38174 L .52582 .37041 L F
.5297 .39817 m .52737 .38148 L p .52741 .38174 L F P
0 g
s
.52817 .38599 m .5297 .39817 L .5307 .40122 L .904 .636 .132 r
F
.53054 .39217 m .5297 .39817 L p .52713 .37974 L 0 0 .394 r
F P
0 g
s
.52258 .36998 m .52514 .37576 L .52327 .38582 L .844 .604 .575 r
F
.52966 .40039 m .52584 .37199 L p .52585 .372 L .911 .651 .532 r
F P
0 g
s
.52373 .3827 m .53152 .38358 L p .52832 .3692 L .724 .426 .472 r
F P
0 g
s
.52617 .3657 m .52379 .35889 L .5267 .37351 L .127 .121 .58 r
F
.52457 .3638 m .5267 .37351 L p .52379 .35889 L F P
0 g
s
.52393 .35977 m .52534 .36566 L p .52195 .36162 L .76 .42 .409 r
F P
0 g
s
.52253 .35392 m .52442 .35393 L p .52646 .36363 L .673 .365 .443 r
F P
0 g
s
.5269 .36284 m .52534 .36566 L .52832 .3692 L .724 .426 .472 r
F
.52603 .37116 m .5253 .36767 L .52554 .36893 L 0 0 .299 r
F
.52621 .37178 m .5253 .36767 L .52603 .37116 L F
.52575 .3725 m .52584 .37199 L .52538 .37082 L .88 .643 .575 r
F
.52395 .36401 m .52373 .3827 L .52832 .3692 L .724 .426 .472 r
F
.52378 .39749 m .52256 .36992 L p .52258 .36998 L .844 .604 .575 r
F P
0 g
s
.52268 .38258 m .5294 .39136 L .53248 .39013 L .699 .333 .35 r
F
.53229 .40169 m .53375 .40903 L .53659 .42292 L 0 0 .299 r
F
.53229 .40169 m .53659 .42292 L .53679 .4237 L F
.53407 .41397 m .53692 .42425 L .53532 .4183 L .773 .477 0 r
F
.53407 .41397 m .53532 .4183 L .5307 .40122 L F
.53795 .42843 m .53229 .40169 L .53679 .4237 L 0 0 .299 r
F
.53692 .42425 m .53597 .42038 L .5307 .40122 L .773 .477 0 r
F
.53229 .40169 m .53063 .39383 L .53375 .40903 L 0 0 .299 r
F
.53011 .39521 m .53054 .39217 L p .53602 .42151 L .009 .089 .6 r
F P
0 g
s
.5294 .39136 m .53192 .39319 L .53248 .39013 L .699 .333 .35 r
F
.53192 .39319 m .53378 .39454 L .53307 .39215 L F
.53192 .39319 m .53307 .39215 L .53248 .39013 L F
.5323 .38367 m .52648 .38301 L p .53462 .39076 L F P
0 g
s
.54032 .39618 m .53603 .38949 L p .53462 .39076 L .699 .333 .35 r
F P
0 g
s
.53603 .38949 m .5323 .38367 L p .53462 .39076 L .699 .333 .35 r
F P
0 g
s
.52257 .36237 m .52373 .3827 L .52261 .36242 L .724 .426 .472 r
F
.52253 .35392 m .52393 .35977 L p .52195 .36162 L .76 .42 .409 r
F P
0 g
s
.52145 .36103 m .52253 .35392 L .52195 .36162 L .76 .42 .409 r
F
.52106 .33753 m .52106 .3375 L .52106 .33754 L .977 .7 .306 r
F
.52106 .33757 m .52106 .3375 L .52107 .33762 L .687 .415 .505 r
F
.52106 .3375 m .52253 .35392 L .52116 .3536 L p .769 .477 .491 r
F P
0 g
s
.52116 .3536 m .52114 .3536 L p .52106 .3375 L .769 .477 .491 r
F P
0 g
s
.52116 .33678 m .52116 .33678 L p .52106 .3375 L .736 .504 .577 r
F P
0 g
s
.52039 .35343 m .52253 .35392 L p .52145 .36103 L .76 .42 .409 r
F P
0 g
s
.52109 .33702 m .52256 .36235 L .52113 .33696 L .724 .426 .472 r
F
.52256 .36235 m .52373 .3827 L .52257 .36237 L F
.52256 .36235 m .52257 .36237 L .52113 .33696 L F
.51748 .35329 m .51486 .36564 L p .5189 .34774 L 0 0 .016 r
F P
0 g
s
.51486 .36564 m .51483 .36627 L .5189 .34774 L 0 0 .016 r
F
.51882 .34824 m .51619 .3621 L p .51849 .34949 L .941 .796 .337 r
F P
0 g
s
.51974 .34262 m .51755 .353 L .5189 .34774 L 0 0 .016 r
F
.51619 .3621 m .51607 .36231 L .51889 .34715 L .941 .796 .337 r
F
.51897 .34683 m .51889 .34715 L .51999 .34123 L F
.51619 .3621 m .51889 .34715 L .51897 .34683 L F
.51373 .36705 m .51543 .36185 L .51421 .36568 L .278 .05 .378 r
F
.51373 .36705 m .51421 .36568 L .51252 .37095 L F
.51229 .37143 m .51373 .36705 L .51252 .37095 L F
.51399 .365 m .51573 .36035 L .51546 .36171 L F
.51235 .36886 m .51543 .36185 L .51229 .37143 L F
.51493 .36438 m .51252 .37095 L p .51543 .36185 L F P
0 g
s
.51743 .35343 m .51232 .37013 L .51484 .3657 L .141 0 .344 r
F
.51252 .37095 m .51228 .37159 L p .51229 .37143 L .278 .05 .378 r
F P
0 g
s
.51232 .37013 m .51228 .37159 L .51484 .3657 L .141 0 .344 r
F
.50024 .39638 m .51235 .36886 L .51229 .37143 L .278 .05 .378 r
F
.51056 .37608 m .51538 .36312 L .51555 .36186 L .999 .806 .468 r
F
.51484 .3657 m .51485 .36566 L .51977 .34249 L .141 0 .344 r
F
.5181 .35028 m .51607 .36231 L .5153 .36369 L .941 .796 .337 r
F
.51889 .34715 m .51607 .36231 L .5181 .35028 L F
.51442 .36084 m .51236 .36845 L p .51877 .34711 L .141 0 .344 r
F P
0 g
s
.5032 .38919 m .51287 .36751 L .50024 .39638 L .278 .05 .378 r
F
.51287 .36751 m .51323 .36685 L .50024 .39638 L F
.51236 .36845 m .51233 .36974 L .51877 .34711 L .141 0 .344 r
F
.51525 .35948 m .51233 .36974 L .51743 .35343 L F
.51233 .36974 m .51232 .37013 L .51743 .35343 L F
.51877 .34711 m .51525 .35948 L .51743 .35343 L F
.52093 .33687 m .51442 .36084 L p .51877 .34711 L F P
0 g
s
.51999 .34123 m .5153 .36369 L .51583 .35976 L .941 .796 .337 r
F
.51213 .37854 m .51147 .38186 L .51327 .37362 L 0 0 .016 r
F
.51007 .38893 m .51215 .3834 L p .51147 .38186 L F P
0 g
s
.51162 .3785 m .51215 .3834 L p .5088 .38736 L .987 .882 .54 r
F P
0 g
s
.51087 .37126 m .51106 .3708 L .50891 .37639 L .278 .05 .378 r
F
.51106 .3708 m .51176 .36954 L .50891 .37639 L F
.51338 .36464 m .51169 .36967 L .51234 .36848 L .425 .209 .479 r
F
.51438 .36094 m .51234 .36848 L .51442 .36084 L F
.51236 .36845 m .51442 .36084 L p .51234 .36848 L F P
0 g
s
.50891 .37639 m .51236 .36808 L .51235 .36868 L .278 .05 .378 r
F
.51022 .37234 m .51054 .37175 L .5032 .38919 L F
.5032 .38919 m .51087 .37126 L .50891 .37639 L F
.52028 .33995 m .51583 .35976 L .52029 .33982 L .941 .796 .337 r
F
.51584 .35969 m .52029 .33982 L p .51583 .35976 L F P
0 g
s
.52098 .33673 m .51673 .35573 L p .51532 .36011 L .995 .756 .431 r
F P
0 g
s
.51673 .35573 m .51584 .35969 L p .51535 .36008 L .995 .756 .431 r
F P
0 g
s
.51673 .35573 m .51535 .36008 L .51532 .36011 L .995 .756 .431 r
F
.5041 .38989 m .50926 .37619 L .50286 .38753 L .999 .806 .468 r
F
.50926 .37619 m .50448 .38427 L .50286 .38753 L F
.46069 .45325 m .4406 .47969 L .42078 .45414 L .966 .911 .706 r
F
.50531 .39806 m .505 .39864 L .50945 .38869 L .754 .844 .394 r
F
.505 .39864 m .48839 .43045 L .50314 .40121 L F
.505 .39864 m .50314 .40121 L .50945 .38869 L F
.50272 .38837 m .50024 .39113 L .50295 .38705 L .995 .783 .504 r
F
.49014 .42653 m .50945 .38869 L p .48839 .43045 L .754 .844 .394 r
F P
0 g
s
.39372 .54262 m .37939 .50875 L .3876 .49664 L .826 .93 .901 r
F
.36992 .43976 m .38766 .49639 L .36559 .43532 L .879 .885 .834 r
F
.39487 .48441 m .38766 .49639 L .36992 .43976 L F
.38765 .49641 m .38763 .4966 L p .40997 .45932 L .905 .943 .821 r
F P
0 g
s
.39459 .50978 m .40997 .45932 L .38763 .4966 L .905 .943 .821 r
F
.3877 .49651 m .36696 .4577 L .36335 .4392 L .879 .885 .834 r
F
.37918 .42452 m .37246 .4234 L .36914 .41928 L .941 .831 .707 r
F
.39623 .44488 m .43202 .43655 L .40381 .42599 L .956 .86 .705 r
F
.38353 .42421 m .39875 .42555 L .3824 .41797 L F
.39875 .42555 m .40381 .42599 L .38793 .42004 L F
.39875 .42555 m .38793 .42004 L .3824 .41797 L F
.39459 .50978 m .43533 .47494 L .40997 .45932 L .905 .943 .821 r
F
.41957 .4953 m .36878 .42978 L .37839 .42612 L .879 .885 .834 r
F
.41433 .46391 m .41957 .4953 L p .37839 .42612 L F P
0 g
s
.39006 .46024 m .40381 .42599 L .38353 .42421 L .956 .86 .705 r
F
.43202 .43655 m .4307 .43552 L .3824 .41797 L F
.4697 .42086 m .48612 .41418 L .47093 .41874 L .994 .841 .584 r
F
.4599 .4377 m .48214 .41775 L .4697 .42086 L F
.48214 .41775 m .48612 .41418 L .4697 .42086 L F
.4642 .43936 m .48612 .41418 L .4599 .4377 L F
.4642 .43936 m .48154 .42686 L .48612 .41418 L F
.47084 .42375 m .47353 .42898 L p .46648 .42895 L .983 .809 .589 r
F P
0 g
s
.40358 .45261 m .43202 .43655 L .39623 .44488 L .956 .86 .705 r
F
.44012 .51082 m .46211 .47816 L .48208 .44814 L .611 .819 .414 r
F
.50469 .39944 m .48151 .44892 L .48032 .45055 L .754 .844 .394 r
F
.44012 .51082 m .48208 .44814 L .4813 .44921 L .611 .819 .414 r
F
.49858 .41327 m .48201 .44822 L .50469 .39944 L .754 .844 .394 r
F
.48201 .44822 m .46533 .48344 L .48151 .44892 L F
.48201 .44822 m .48151 .44892 L .50469 .39944 L F
.44691 .48777 m .44817 .48627 L .46279 .46436 L .388 .147 .423 r
F
.45098 .48272 m .45033 .48373 L .45512 .4769 L F
.45033 .48373 m .45174 .48206 L .45512 .4769 L F
.46077 .47924 m .43276 .49188 L .45202 .47121 L .885 .99 .745 r
F
.44532 .50251 m .47994 .45107 L .46063 .47945 L .611 .819 .414 r
F
.45646 .48574 m .43276 .49188 L .46077 .47924 L .885 .99 .745 r
F
.47994 .45107 m .47821 .45346 L .46324 .47553 L .611 .819 .414 r
F
.47994 .45107 m .46324 .47553 L .46063 .47945 L F
.44166 .50864 m .46211 .47816 L p .44012 .51082 L F P
0 g
s
.44012 .51082 m .4813 .44921 L .44532 .50251 L .611 .819 .414 r
F
.4813 .44921 m .47994 .45107 L .44532 .50251 L F
.50469 .39944 m .48032 .45055 L .48814 .43103 L .754 .844 .394 r
F
.46077 .47924 m .45202 .47121 L .4634 .47529 L .885 .99 .745 r
F
.47812 .45358 m .44372 .5043 L p .47821 .45346 L .611 .819 .414 r
F P
0 g
s
.48032 .45055 m .46778 .47686 L .47812 .45358 L .754 .844 .394 r
F
.48032 .45055 m .47812 .45358 L .48814 .43103 L F
.49606 .42712 m .47812 .45358 L p .4956 .42781 L .611 .819 .414 r
F P
0 g
s
.4956 .42781 m .47812 .45358 L .47821 .45346 L .611 .819 .414 r
F
.4813 .44921 m .4956 .42781 L .47994 .45107 L F
.48814 .43103 m .47812 .45358 L .47667 .4529 L .754 .844 .394 r
F
.46778 .47686 m .47465 .45834 L p .47812 .45358 L F P
0 g
s
.47465 .45834 m .47667 .4529 L p .47812 .45358 L .754 .844 .394 r
F P
0 g
s
.46778 .47686 m .46533 .48344 L p .46077 .47924 L .885 .99 .745 r
F P
0 g
s
.46077 .47924 m .4634 .47529 L .46778 .47686 L .885 .99 .745 r
F
.43504 .51798 m .44532 .50251 L .43228 .52189 L .611 .819 .414 r
F
.44012 .51082 m .44532 .50251 L .43504 .51798 L F
.46063 .47945 m .46324 .47553 L .44372 .5043 L F
.4553 .47238 m .4966 .41367 L .4669 .4404 L .948 .949 .642 r
F
.49481 .41459 m .4669 .4404 L p .4966 .41367 L F P
0 g
s
.48249 .43373 m .47838 .44137 L .48108 .43666 L .388 .147 .423 r
F
.46649 .46463 m .46588 .46651 L .46848 .4619 L 0 0 .107 r
F
.4612 .4745 m .4981 .41154 L .4553 .47238 L .948 .949 .642 r
F
.4457 .48428 m .44372 .5043 L p .4669 .4404 L F P
0 g
s
.47432 .45211 m .49014 .42653 L .4893 .42656 L .948 .949 .642 r
F
.47667 .4529 m .42751 .44395 L .43998 .4353 L .966 .911 .706 r
F
.39006 .46024 m .41759 .45897 L p .43226 .43675 L .956 .86 .705 r
F P
0 g
s
.44914 .43357 m .46673 .4216 L .45197 .42699 L .994 .841 .584 r
F
.46673 .4216 m .47093 .41874 L .46118 .42298 L F
.46673 .4216 m .46118 .42298 L .45197 .42699 L F
.4504 .43406 m .46728 .42146 L .44914 .43357 L F
.46728 .42146 m .47093 .41874 L .46673 .4216 L F
.46728 .42146 m .46673 .4216 L .44914 .43357 L F
.43888 .43074 m .43226 .43675 L .42693 .43255 L .983 .809 .589 r
F
.45115 .42889 m .44467 .43159 L .43888 .43074 L F
.44467 .43159 m .43226 .43675 L .43888 .43074 L F
.44914 .43357 m .43226 .43675 L p .44467 .43159 L F P
0 g
s
.44968 .43232 m .44467 .43159 L .45115 .42889 L .983 .809 .589 r
F
.44914 .43357 m .44467 .43159 L .44968 .43232 L F
.44067 .42884 m .42693 .43255 L .42214 .42877 L F
.45115 .42889 m .42693 .43255 L .44067 .42884 L F
.43998 .4353 m .44914 .43357 L p .4543 .43719 L .966 .911 .706 r
F P
0 g
s
.47667 .4529 m .43998 .4353 L .4543 .43719 L .966 .911 .706 r
F
.44177 .4405 m .44385 .44588 L p .44779 .4367 L .994 .841 .584 r
F P
0 g
s
.44779 .4367 m .44823 .43568 L .44698 .43661 L .994 .841 .584 r
F
.44779 .4367 m .44698 .43661 L .44177 .4405 L F
.47546 .44368 m .47667 .4529 L p .44914 .43357 L .966 .911 .706 r
F P
0 g
s
.46214 .45026 m .42078 .45414 L .42751 .44395 L .966 .911 .706 r
F
.46069 .45325 m .42078 .45414 L .46214 .45026 L F
.44824 .43675 m .47093 .41874 L .44823 .43568 L .994 .841 .584 r
F
.44835 .45756 m .45976 .43795 L .44824 .43675 L F
.45976 .43795 m .47093 .41874 L .44824 .43675 L F
.44385 .44588 m .44835 .45756 L p .44823 .43568 L F P
0 g
s
.39006 .46024 m .43226 .43675 L .43202 .43655 L .956 .86 .705 r
F
.45255 .42564 m .42214 .42877 L .45256 .42563 L .983 .809 .589 r
F
.45115 .42889 m .44067 .42884 L .45255 .42564 L F
.44067 .42884 m .42214 .42877 L .45255 .42564 L F
.48612 .41418 m .487 .41174 L .47093 .41874 L .994 .841 .584 r
F
.50597 .38997 m .48866 .41112 L .49741 .39789 L .987 .882 .54 r
F
.51567 .36417 m .50618 .3892 L .50575 .39079 L .863 .83 .356 r
F
.5088 .38736 m .50569 .39103 L .50597 .38997 L .987 .882 .54 r
F
.50569 .39103 m .48866 .41112 L .50597 .38997 L F
.51567 .36417 m .5161 .36245 L .50618 .3892 L .863 .83 .356 r
F
.50931 .37969 m .51228 .37159 L p .51268 .37132 L 0 0 .051 r
F P
0 g
s
.51228 .37159 m .51382 .36805 L .51268 .37132 L 0 0 .051 r
F
.51268 .37132 m .51382 .36805 L .51436 .36681 L F
.5161 .36245 m .51395 .3677 L .50618 .3892 L .863 .83 .356 r
F
.51558 .36319 m .50631 .3887 L .50618 .3892 L F
.5167 .3601 m .51557 .36321 L .51558 .36319 L F
.51557 .36321 m .50631 .3887 L .51558 .36319 L F
.51555 .36186 m .51584 .35969 L .50292 .38719 L .999 .806 .468 r
F
.51555 .36186 m .50292 .38719 L .50286 .38753 L F
.50608 .38882 m .51555 .36186 L .51179 .36947 L F
.51583 .35976 m .51471 .36474 L .51584 .35969 L .941 .796 .337 r
F
.5161 .36245 m .5167 .3601 L .51395 .3677 L .863 .83 .356 r
F
.50754 .38528 m .5153 .36369 L p .51056 .37608 L .999 .806 .468 r
F P
0 g
s
.5153 .36369 m .51538 .36312 L .51056 .37608 L .999 .806 .468 r
F
.51696 .35909 m .5153 .36369 L p .5167 .3601 L .863 .83 .356 r
F P
0 g
s
.5153 .36369 m .50631 .3887 L p .51557 .36321 L .863 .83 .356 r
F P
0 g
s
.5167 .3601 m .5153 .36369 L .51557 .36321 L .863 .83 .356 r
F
.5088 .38736 m .49741 .39789 L .50602 .38885 L .987 .882 .54 r
F
.50378 .39609 m .50283 .39846 L .50054 .40479 L .863 .83 .356 r
F
.50575 .39079 m .50283 .39846 L .50378 .39609 L F
.50575 .39079 m .50618 .3892 L .50283 .39846 L F
.50051 .40483 m .4981 .41154 L p .50054 .40479 L F P
0 g
s
.50144 .39604 m .48154 .42686 L .48594 .41469 L .987 .882 .54 r
F
.48616 .41407 m .47563 .42709 L .47824 .42342 L F
.50144 .39604 m .48594 .41469 L .48616 .41407 L F
.48594 .41469 m .47563 .42709 L .48616 .41407 L F
.48342 .40974 m .45255 .42564 L p .48703 .41167 L .994 .841 .584 r
F P
0 g
s
.48866 .41112 m .48616 .41407 L .49552 .40061 L .987 .882 .54 r
F
.48866 .41112 m .49552 .40061 L .49741 .39789 L F
.48616 .41407 m .48713 .41138 L .49604 .39933 L F
.49741 .39789 m .49604 .39933 L .5023 .39086 L F
.48616 .41407 m .49604 .39933 L .49741 .39789 L F
.48743 .41055 m .48342 .40974 L .48703 .41167 L .994 .841 .584 r
F
.487 .41174 m .45255 .42564 L .45197 .42699 L F
.489 .41087 m .48703 .41167 L .487 .41174 L F
.48703 .41167 m .45255 .42564 L .487 .41174 L F
.48666 .41157 m .48743 .41055 L p .47824 .42342 L .987 .882 .54 r
F P
0 g
s
.48743 .41055 m .49038 .40664 L p .48731 .41088 L .987 .882 .54 r
F P
0 g
s
.47824 .42342 m .48743 .41055 L .48731 .41088 L .987 .882 .54 r
F
.47353 .42898 m .47084 .42375 L p .48358 .4096 L .995 .783 .504 r
F P
0 g
s
.4929 .40203 m .47353 .42898 L .48358 .4096 L .995 .783 .504 r
F
.48713 .41138 m .49038 .40664 L .49124 .40582 L .987 .882 .54 r
F
.47824 .42342 m .48731 .41088 L .48713 .41138 L F
.48731 .41088 m .49038 .40664 L .48713 .41138 L F
.49038 .40664 m .49539 .4 L p .49124 .40582 L F P
0 g
s
.49124 .40582 m .49539 .4 L .49604 .39933 L .987 .882 .54 r
F
.5023 .39086 m .49539 .4 L p .50272 .38837 L .995 .783 .504 r
F P
0 g
s
.49539 .4 m .47353 .42898 L p .4929 .40203 L .995 .783 .504 r
F P
0 g
s
.49539 .4 m .49311 .40186 L .50272 .38837 L .995 .783 .504 r
F
.49311 .40186 m .49417 .40027 L .50272 .38837 L F
.50024 .39113 m .4929 .40203 L .48358 .4096 L F
.50272 .38837 m .49417 .40027 L .50024 .39113 L F
.49417 .40027 m .4929 .40203 L .50024 .39113 L F
.5028 .38793 m .50286 .38753 L .50222 .38883 L .999 .806 .468 r
F
.5028 .38793 m .50222 .38883 L .50083 .39164 L F
.5023 .39086 m .5028 .38793 L .50083 .39164 L F
.5041 .38989 m .50286 .38753 L .5023 .39086 L F
.50106 .39132 m .50331 .38651 L .50375 .38607 L .919 .605 .433 r
F
.50608 .38882 m .51179 .36947 L .5041 .38989 L .999 .806 .468 r
F
.4981 .41154 m .50754 .38528 L p .51056 .37608 L F P
0 g
s
.51162 .3785 m .5023 .39086 L p .5026 .38908 L .995 .783 .504 r
F P
0 g
s
.50352 .38749 m .5026 .38908 L .50272 .38837 L .995 .783 .504 r
F
.51162 .3785 m .5026 .38908 L .50352 .38749 L F
.49533 .4036 m .49568 .4046 L p .50608 .38882 L .999 .806 .468 r
F P
0 g
s
.49568 .4046 m .4981 .41154 L p .50377 .3954 L .999 .806 .468 r
F P
0 g
s
.50377 .3954 m .50608 .38882 L .49568 .4046 L .999 .806 .468 r
F
.50602 .38885 m .49741 .39789 L .5023 .39086 L .987 .882 .54 r
F
.50631 .3887 m .50051 .40483 L p .50618 .3892 L .863 .83 .356 r
F P
0 g
s
.50618 .3892 m .50051 .40483 L .50054 .40479 L .863 .83 .356 r
F
.5023 .39086 m .51162 .3785 L p .5088 .38736 L .987 .882 .54 r
F P
0 g
s
.49539 .4 m .5023 .39086 L p .49604 .39933 L .987 .882 .54 r
F P
0 g
s
.50083 .39164 m .49533 .4036 L p .50608 .38882 L .999 .806 .468 r
F P
0 g
s
.50288 .40157 m .50575 .39079 L .50054 .40479 L .863 .83 .356 r
F
.51215 .3834 m .50978 .38805 L p .50535 .39229 L .987 .882 .54 r
F P
0 g
s
.51215 .3834 m .50535 .39229 L .50569 .39103 L .987 .882 .54 r
F
.50978 .38805 m .50288 .40157 L p .50535 .39229 L F P
0 g
s
.50469 .39944 m .47667 .4529 L .50531 .39806 L .754 .844 .394 r
F
.50469 .39944 m .50531 .39806 L .51007 .38893 L F
.50512 .39848 m .49641 .42627 L p .50366 .40178 L 0 0 .051 r
F P
0 g
s
.49621 .4269 m .50512 .39848 L p .50495 .39871 L .863 .83 .356 r
F P
0 g
s
.51215 .3834 m .51007 .38893 L p .50531 .39806 L .754 .844 .394 r
F P
0 g
s
.50945 .38869 m .51215 .3834 L p .50531 .39806 L .754 .844 .394 r
F P
0 g
s
.4981 .41154 m .49622 .42682 L p .50495 .39871 L .863 .83 .356 r
F P
0 g
s
.49638 .42634 m .50495 .39871 L .49622 .42682 L .863 .83 .356 r
F
.50495 .39871 m .50054 .40479 L .4981 .41154 L F
.5075 .39453 m .5066 .39644 L p .50915 .39094 L .754 .844 .394 r
F P
0 g
s
.5066 .39644 m .50647 .39661 L .50915 .39094 L .754 .844 .394 r
F
.50745 .40708 m .50314 .41441 L .50239 .41245 L .706 .517 .631 r
F
.50916 .41117 m .49812 .40579 L p .50197 .39674 L .852 .512 .409 r
F P
0 g
s
.50197 .39674 m .50634 .38646 L .50722 .39415 L .852 .512 .409 r
F
.50197 .39674 m .50722 .39415 L .50916 .41117 L F
.51451 .37971 m .51218 .38461 L p .51661 .38951 L .795 .41 .325 r
F P
0 g
s
.50974 .38982 m .5075 .39453 L p .50915 .39094 L .754 .844 .394 r
F P
0 g
s
.51218 .38461 m .50776 .39388 L p .51624 .3897 L .795 .41 .325 r
F P
0 g
s
.51218 .38461 m .51624 .3897 L .51661 .38951 L .795 .41 .325 r
F
.52064 .35686 m .5203 .36482 L p .52153 .3676 L .804 .59 .613 r
F P
0 g
s
.52192 .36165 m .51915 .36423 L .52229 .37187 L .81 .496 .459 r
F
.51354 .41331 m .51624 .3897 L p .51697 .39061 L .76 .42 .409 r
F P
0 g
s
.53519 .44585 m .53097 .42006 L .52664 .42991 L .804 .59 .613 r
F
.52938 .4489 m .52683 .42363 L p .52415 .42457 L .795 .41 .325 r
F P
0 g
s
.50634 .38646 m .50975 .40431 L .50916 .41117 L .852 .512 .409 r
F
.49814 .41408 m .49812 .40579 L .50293 .41339 L .828 .407 .241 r
F
.4981 .41154 m .49481 .41459 L p .4966 .41367 L .948 .949 .642 r
F P
0 g
s
.4981 .41154 m .49533 .4036 L p .48612 .41418 L .994 .841 .584 r
F P
0 g
s
.48154 .42686 m .4981 .41154 L p .48612 .41418 L .994 .841 .584 r
F P
0 g
s
.49621 .4269 m .4981 .41154 L p .4927 .42075 L .948 .949 .642 r
F P
0 g
s
.48579 .41585 m .49473 .40855 L .49814 .41408 L .828 .407 .241 r
F
.49473 .40855 m .49812 .40579 L .49814 .41408 L F
.50776 .39388 m .51079 .39239 L .50975 .40431 L .852 .512 .409 r
F
.51162 .3785 m .51252 .37449 L p .51253 .37447 L .941 .796 .337 r
F P
0 g
s
.51346 .36933 m .51162 .3785 L .51112 .37744 L .995 .756 .431 r
F
.51112 .37744 m .51162 .3785 L p .50352 .38749 L .995 .783 .504 r
F P
0 g
s
.50352 .38749 m .50272 .38837 L .50388 .38686 L .995 .783 .504 r
F
.50352 .38749 m .50388 .38686 L .51112 .37744 L F
.51535 .36008 m .50083 .39164 L p .51584 .35969 L .999 .806 .468 r
F P
0 g
s
.50926 .37619 m .51179 .36947 L .50448 .38427 L .999 .806 .468 r
F
.50331 .38651 m .51476 .36195 L .50901 .37791 L .919 .605 .433 r
F
.51523 .36048 m .51472 .36256 L .51521 .36052 L .995 .756 .431 r
F
.51112 .37744 m .51471 .36258 L p .51472 .36256 L F P
0 g
s
.51471 .36258 m .51521 .36052 L p .51472 .36256 L .995 .756 .431 r
F P
0 g
s
.51471 .36474 m .51253 .37447 L .5147 .36475 L .941 .796 .337 r
F
.51252 .37449 m .5147 .36475 L p .51253 .37447 L F P
0 g
s
.5147 .36475 m .51584 .35969 L p .51471 .36474 L .941 .796 .337 r
F P
0 g
s
.51584 .35969 m .51162 .3785 L p .51498 .3618 L .995 .756 .431 r
F P
0 g
s
.50295 .38705 m .51535 .36008 L p .50317 .38671 L .919 .605 .433 r
F P
0 g
s
.51597 .35875 m .51535 .36008 L p .51584 .35969 L .999 .806 .468 r
F P
0 g
s
.51535 .36008 m .51597 .35875 L p .51535 .3601 L .919 .605 .433 r
F P
0 g
s
.51535 .36008 m .51535 .3601 L .50317 .38671 L .919 .605 .433 r
F
.51563 .35986 m .51498 .3618 L .51527 .36035 L .995 .756 .431 r
F
.5152 .36042 m .51535 .3601 L .51521 .36052 L .919 .605 .433 r
F
.51535 .36008 m .51527 .36035 L .51532 .36011 L .995 .756 .431 r
F
.51563 .35986 m .51527 .36035 L .51535 .36008 L F
.51584 .35969 m .51498 .3618 L .51563 .35986 L F
.51583 .35976 m .5153 .36369 L .51471 .36474 L .941 .796 .337 r
F
.51542 .36195 m .51397 .36716 L .51548 .36237 L .894 .587 .457 r
F
.51397 .36716 m .51112 .37744 L .51351 .36918 L F
.51213 .37854 m .51327 .37362 L .51294 .3749 L 0 0 .016 r
F
.51215 .3834 m .51162 .3785 L p .51471 .36474 L .941 .796 .337 r
F P
0 g
s
.51607 .36231 m .51215 .3834 L .51471 .36474 L .941 .796 .337 r
F
.51397 .36716 m .51351 .36918 L .51548 .36237 L .894 .587 .457 r
F
.51521 .36052 m .51471 .36258 L p .51542 .36195 L F P
0 g
s
.51471 .36258 m .51112 .37744 L p .51542 .36195 L .894 .587 .457 r
F P
0 g
s
.50331 .38651 m .5152 .36042 L .51476 .36195 L .919 .605 .433 r
F
.50317 .38671 m .5152 .36042 L .50331 .38651 L F
.50299 .38677 m .50578 .38357 L p .50272 .38837 L .995 .783 .504 r
F P
0 g
s
.50578 .38357 m .51112 .37744 L p .50388 .38686 L .995 .783 .504 r
F P
0 g
s
.50272 .38837 m .50578 .38357 L .50388 .38686 L .995 .783 .504 r
F
.51203 .37805 m .51154 .37772 L p .50815 .38888 L .893 .542 .373 r
F P
0 g
s
.50815 .38888 m .50732 .39161 L .51203 .37805 L .893 .542 .373 r
F
.51617 .36217 m .5146 .37048 L .51618 .36213 L 0 0 .016 r
F
.51215 .3834 m .5146 .3705 L p .5146 .37048 L F P
0 g
s
.51619 .3621 m .51215 .3834 L p .51607 .36231 L .941 .796 .337 r
F P
0 g
s
.51451 .37971 m .51203 .37805 L p .50732 .39161 L .893 .542 .373 r
F P
0 g
s
.51007 .38893 m .50974 .38982 L p .50915 .39094 L .754 .844 .394 r
F P
0 g
s
.50776 .39388 m .51451 .37971 L p .50732 .39161 L .893 .542 .373 r
F P
0 g
s
.49812 .40579 m .50932 .41125 L p .50293 .41339 L .828 .407 .241 r
F P
0 g
s
.50724 .3912 m .49647 .41432 L .50466 .40039 L .893 .542 .373 r
F
.51077 .37867 m .50676 .38867 L .50724 .3912 L F
.50676 .38867 m .49647 .41432 L .50724 .3912 L F
.51112 .37744 m .51048 .37903 L .51077 .37867 L F
.51048 .37903 m .49642 .41432 L .51077 .37867 L F
.48344 .44161 m .49621 .4269 L .4927 .42075 L .948 .949 .642 r
F
.49642 .41432 m .4882 .43494 L .49647 .41432 L .893 .542 .373 r
F
.49642 .41432 m .49647 .41432 L .51077 .37867 L F
.47994 .45107 m .4956 .42781 L .47821 .45346 L .611 .819 .414 r
F
.4813 .44921 m .48208 .44814 L .4956 .42781 L F
.47812 .45358 m .49621 .4269 L p .48344 .44161 L .948 .949 .642 r
F P
0 g
s
.49641 .42627 m .49621 .4269 L p .49641 .4262 L 0 0 .051 r
F P
0 g
s
.49621 .4269 m .49606 .42712 L p .4956 .42781 L .611 .819 .414 r
F P
0 g
s
.48208 .44814 m .48659 .44164 L .4956 .42781 L .611 .819 .414 r
F
.49622 .42682 m .49621 .4269 L p .49638 .42634 L .863 .83 .356 r
F P
0 g
s
.48659 .44164 m .49621 .4269 L p .4956 .42781 L .611 .819 .414 r
F P
0 g
s
.4882 .43494 m .49137 .43601 L p .49968 .41658 L .795 .41 .325 r
F P
0 g
s
.49137 .43601 m .51704 .44472 L p .50807 .43018 L .795 .41 .325 r
F P
0 g
s
.49968 .41658 m .49137 .43601 L .50807 .43018 L .795 .41 .325 r
F
.53532 .41279 m .5419 .44276 L p .53594 .41358 L .127 .121 .58 r
F P
0 g
s
.53532 .41279 m .53704 .42244 L .5341 .41123 L .009 .089 .6 r
F
.5432 .44869 m .53561 .41444 L .53615 .41745 L 0 0 .299 r
F
.5432 .44869 m .53554 .41403 L .53561 .41444 L F
.5341 .41123 m .53704 .42244 L .53602 .42151 L .009 .089 .6 r
F
.53228 .40769 m .53602 .42151 L p .53384 .41315 L .773 .477 0 r
F P
0 g
s
.5297 .39817 m .53228 .40769 L p .53384 .41315 L .773 .477 0 r
F P
0 g
s
.5307 .40122 m .5297 .39817 L p .53407 .41397 L .773 .477 0 r
F P
0 g
s
.53602 .42151 m .5297 .39817 L .53011 .39521 L p .009 .089 .6 r
F P
0 g
s
.52966 .40039 m .53329 .4148 L p .53146 .41896 L .835 .63 .625 r
F P
0 g
s
.52966 .40039 m .53146 .41896 L .5313 .4186 L .835 .63 .625 r
F
.53329 .4148 m .53806 .43373 L p .53146 .41896 L F P
0 g
s
.51875 .41338 m .51931 .41623 L .5186 .41491 L .768 .58 .644 r
F
.53806 .43373 m .52966 .40039 L p .53824 .43191 L .904 .666 .569 r
F P
0 g
s
.51354 .41331 m .53804 .41875 L p .53673 .42018 L .712 .271 .202 r
F P
0 g
s
.5313 .4186 m .53146 .41896 L .53135 .4192 L .835 .63 .625 r
F
.52683 .42363 m .51354 .41331 L .53673 .42018 L .712 .271 .202 r
F
.50807 .43018 m .51316 .41446 L p .52683 .42363 L F P
0 g
s
.51316 .41446 m .51354 .41331 L p .52683 .42363 L .712 .271 .202 r
F P
0 g
s
.52683 .42363 m .52938 .4489 L .5713 .45349 L p .699 .333 .35 r
F P
0 g
s
.49981 .41679 m .50293 .41339 L .51354 .41331 L .828 .407 .241 r
F
.49814 .41408 m .4882 .43494 L p .49968 .41658 L .795 .41 .325 r
F P
0 g
s
.49812 .40579 m .50615 .38545 L p .50634 .38646 L .852 .512 .409 r
F P
0 g
s
.50106 .39132 m .5024 .3887 L .502 .38999 L .919 .605 .433 r
F
.5024 .3887 m .50375 .38607 L .502 .38999 L F
.502 .38999 m .50375 .38607 L .50503 .38574 L F
.50286 .38753 m .50292 .38719 L .50083 .39164 L .999 .806 .468 r
F
.50083 .39164 m .50295 .38705 L p .50106 .39132 L .919 .605 .433 r
F P
0 g
s
.50295 .38705 m .50317 .38671 L .50106 .39132 L .919 .605 .433 r
F
.49533 .4036 m .48871 .40701 L p .489 .41087 L .994 .841 .584 r
F P
0 g
s
.45445 .42467 m .48342 .40974 L p .45534 .42489 L .916 .564 .354 r
F P
0 g
s
.48343 .40977 m .47084 .42375 L .48227 .41068 L .995 .783 .504 r
F
.48203 .41049 m .48342 .40974 L .48203 .41051 L .916 .564 .354 r
F
.48342 .40974 m .48212 .41079 L .45711 .42441 L F
.48343 .40977 m .48227 .41068 L .50295 .38705 L .995 .783 .504 r
F
.50024 .39113 m .48343 .40977 L .50295 .38705 L F
.48359 .40965 m .48342 .40974 L p .48347 .40975 L .994 .841 .584 r
F P
0 g
s
.48743 .41055 m .48359 .40965 L .48347 .40975 L .994 .841 .584 r
F
.48871 .40701 m .48359 .40965 L p .48743 .41055 L F P
0 g
s
.489 .41087 m .48871 .40701 L .48743 .41055 L .994 .841 .584 r
F
.48229 .41065 m .47084 .42375 L .48212 .41079 L .995 .783 .504 r
F
.47084 .42375 m .48206 .41084 L p .48212 .41079 L F P
0 g
s
.48342 .40974 m .49533 .4036 L p .48212 .41079 L .916 .564 .354 r
F P
0 g
s
.50295 .38705 m .48314 .40968 L .50299 .38677 L .995 .783 .504 r
F
.48314 .40968 m .48229 .41065 L .48314 .40968 L F
.48314 .40968 m .48314 .40968 L .50299 .38677 L F
.45711 .42441 m .48212 .41079 L .47582 .41591 L .916 .564 .354 r
F
.48206 .41084 m .50299 .38677 L p .48229 .41065 L .995 .783 .504 r
F P
0 g
s
.48308 .40967 m .48206 .41084 L p .49448 .40116 L .893 .542 .373 r
F P
0 g
s
.48206 .41084 m .47084 .42375 L p .48222 .41287 L .893 .542 .373 r
F P
0 g
s
.48212 .41079 m .48833 .40741 L .48212 .41163 L .916 .564 .354 r
F
.48212 .41079 m .48212 .41163 L .48206 .41084 L F
.48833 .40741 m .48688 .40894 L .47582 .41591 L F
.48616 .41407 m .47824 .42342 L .48713 .41138 L .987 .882 .54 r
F
.47353 .42898 m .48666 .41157 L p .47824 .42342 L F P
0 g
s
.48594 .41469 m .48154 .42686 L .47563 .42709 L .987 .882 .54 r
F
.48222 .41287 m .47084 .42375 L .48284 .42083 L .893 .542 .373 r
F
.47582 .41591 m .48688 .40894 L .48359 .41239 L .916 .564 .354 r
F
.48206 .41084 m .48222 .41287 L .49448 .40116 L .893 .542 .373 r
F
.45534 .42489 m .48203 .41049 L .45711 .42441 L .916 .564 .354 r
F
.48203 .41049 m .48203 .41051 L .45711 .42441 L F
.45255 .42564 m .45445 .42467 L p .45534 .42489 L F P
0 g
s
.45279 .42508 m .45534 .42489 L p .45256 .42563 L .983 .809 .589 r
F P
0 g
s
.45534 .42489 m .47084 .42375 L p .4526 .42563 L .983 .809 .589 r
F P
0 g
s
.45256 .42563 m .45534 .42489 L .4526 .42563 L .983 .809 .589 r
F
.48757 .41058 m .48743 .41055 L .48739 .41066 L .994 .841 .584 r
F
.48757 .41058 m .48739 .41066 L .48703 .41167 L F
.489 .41087 m .48757 .41058 L .48703 .41167 L F
.48612 .41418 m .489 .41087 L .487 .41174 L F
.48223 .413 m .49812 .40579 L .48247 .41605 L .916 .564 .354 r
F
.49533 .4036 m .49734 .40518 L p .49121 .40668 L F P
0 g
s
.49734 .40518 m .49812 .40579 L p .49121 .40668 L .916 .564 .354 r
F P
0 g
s
.48359 .41239 m .49121 .40668 L .49812 .40579 L .916 .564 .354 r
F
.48833 .40741 m .49533 .4036 L .48688 .40894 L F
.48688 .40894 m .49533 .4036 L .48359 .41239 L F
.48249 .41632 m .48957 .41155 L p .48579 .41585 L .828 .407 .241 r
F P
0 g
s
.48957 .41155 m .49812 .40579 L p .48579 .41585 L .828 .407 .241 r
F P
0 g
s
.49812 .40579 m .47508 .42131 L p .47614 .4202 L .916 .564 .354 r
F P
0 g
s
.50106 .39132 m .50317 .38671 L .50331 .38651 L .919 .605 .433 r
F
.49533 .4036 m .50083 .39164 L p .50106 .39132 L F P
0 g
s
.49812 .40579 m .49533 .4036 L p .50106 .39132 L .919 .605 .433 r
F P
0 g
s
.50893 .37994 m .50578 .38357 L p .50674 .3853 L .893 .542 .373 r
F P
0 g
s
.50578 .38357 m .50357 .38612 L p .50615 .38545 L .893 .542 .373 r
F P
0 g
s
.50578 .38357 m .50615 .38545 L .50674 .3853 L .893 .542 .373 r
F
.50728 .38257 m .49812 .40579 L p .50106 .39132 L .919 .605 .433 r
F P
0 g
s
.4893 .42656 m .49014 .42653 L .4927 .42075 L .948 .949 .642 r
F
.4669 .4404 m .48154 .42686 L p .4642 .43936 L .994 .841 .584 r
F P
0 g
s
.49014 .42653 m .47667 .4529 L p .47511 .42711 L .987 .882 .54 r
F P
0 g
s
.47667 .4529 m .49014 .42653 L p .48839 .43045 L .754 .844 .394 r
F P
0 g
s
.50299 .38677 m .48284 .42083 L .4882 .43494 L .893 .542 .373 r
F
.47828 .42854 m .48372 .43205 L p .48284 .42083 L F P
0 g
s
.48372 .43205 m .4882 .43494 L p .48284 .42083 L .893 .542 .373 r
F P
0 g
s
.47582 .41591 m .48223 .413 L .47851 .41864 L .916 .564 .354 r
F
.48223 .413 m .48247 .41605 L .47851 .41864 L F
.47084 .42375 m .46492 .42418 L p .46734 .42792 L .863 .435 .195 r
F P
0 g
s
.46492 .42418 m .45534 .42489 L p .46584 .42754 L .863 .435 .195 r
F P
0 g
s
.46492 .42418 m .46584 .42754 L .46734 .42792 L .863 .435 .195 r
F
.47824 .42342 m .47563 .42709 L .47511 .42711 L .987 .882 .54 r
F
.47084 .42375 m .47828 .42854 L p .48284 .42083 L .893 .542 .373 r
F P
0 g
s
.4882 .43494 m .49814 .41408 L p .49647 .41432 L .893 .542 .373 r
F P
0 g
s
.51112 .37744 m .50893 .37994 L p .4882 .43494 L .893 .542 .373 r
F P
0 g
s
.50893 .37994 m .4936 .41737 L .4882 .43494 L .893 .542 .373 r
F
.47353 .42898 m .44914 .43357 L p .45115 .42889 L .983 .809 .589 r
F P
0 g
s
.42751 .44395 m .43156 .43782 L .43998 .4353 L .966 .911 .706 r
F
.41048 .44085 m .42338 .43842 L p .41731 .4421 L F P
0 g
s
.42338 .43842 m .43452 .43633 L p .42547 .43936 L .966 .911 .706 r
F P
0 g
s
.41731 .4421 m .42338 .43842 L .42547 .43936 L .966 .911 .706 r
F
.4214 .42819 m .3824 .41797 L p .4307 .43552 L .956 .86 .705 r
F P
0 g
s
.43226 .43675 m .41048 .44085 L p .42214 .42877 L .983 .809 .589 r
F P
0 g
s
.42214 .42877 m .41624 .43488 L .40376 .42869 L .983 .809 .589 r
F
.41048 .44085 m .40376 .42869 L p .41624 .43488 L F P
0 g
s
.37698 .42899 m .39006 .46024 L p .3824 .41797 L .956 .86 .705 r
F P
0 g
s
.40376 .42869 m .41048 .44085 L p .36923 .413 L .941 .831 .707 r
F P
0 g
s
.38709 .41975 m .40376 .42869 L p .36923 .413 L .941 .831 .707 r
F P
0 g
s
.38497 .42363 m .37422 .4174 L .36923 .413 L .941 .831 .707 r
F
.37422 .4174 m .3618 .4102 L .36756 .41237 L F
.37422 .4174 m .36756 .41237 L .36923 .413 L F
.37422 .4174 m .36914 .41928 L .3618 .4102 L F
.38497 .42363 m .38106 .42256 L .37422 .4174 L F
.38106 .42256 m .36914 .41928 L .37422 .4174 L F
.3824 .41797 m .37113 .41502 L p .38039 .42205 L .956 .86 .705 r
F P
0 g
s
.37113 .41502 m .3824 .41797 L .38859 .42062 L .881 .453 .167 r
F
.36923 .413 m .3618 .4102 L .35787 .40534 L .941 .831 .707 r
F
.34015 .39176 m .35274 .40122 L p .34866 .39812 L 0 .148 .597 r
F P
0 g
s
.35274 .40122 m .35731 .40465 L p .35325 .40159 L 0 .148 .597 r
F P
0 g
s
.35274 .40122 m .35325 .40159 L .34866 .39812 L 0 .148 .597 r
F
.3824 .41797 m .3854 .41876 L p .38709 .41975 L .881 .453 .167 r
F P
0 g
s
.35731 .40465 m .34015 .39176 L p .36949 .41973 L .854 .843 .833 r
F P
0 g
s
.34867 .39913 m .38709 .41975 L p .36923 .413 L .941 .831 .707 r
F P
0 g
s
.38269 .41738 m .35325 .40159 L p .35712 .40439 L .56 0 0 r
F P
0 g
s
.40376 .42869 m .38709 .41975 L p .39495 .42432 L .56 0 0 r
F P
0 g
s
.41048 .44085 m .36914 .41928 L .38497 .42363 L .941 .831 .707 r
F
.38819 .42767 m .3809 .42227 L p .39986 .42718 L .881 .453 .167 r
F P
0 g
s
.40726 .43149 m .38819 .42767 L p .39986 .42718 L .881 .453 .167 r
F P
0 g
s
.40104 .43183 m .40909 .43301 L .41191 .43535 L .631 .067 0 r
F
.42751 .44395 m .41731 .4421 L .43156 .43782 L .966 .911 .706 r
F
.43452 .43633 m .43998 .4353 L p .41731 .4421 L F P
0 g
s
.41278 .45464 m .41048 .44085 L p .4406 .47969 L .966 .911 .706 r
F P
0 g
s
.41759 .45897 m .44835 .45756 L p .43262 .43622 L .956 .86 .705 r
F P
0 g
s
.42751 .44395 m .42078 .45414 L .41048 .44085 L .966 .911 .706 r
F
.44501 .45459 m .409 .43198 L p .40922 .43311 L .56 0 0 r
F P
0 g
s
.42098 .44289 m .44501 .45459 L p .40922 .43311 L .56 0 0 r
F P
0 g
s
.44835 .45756 m .43999 .43589 L p .43262 .43622 L .956 .86 .705 r
F P
0 g
s
.43999 .43589 m .43878 .43274 L p .43515 .43479 L .956 .86 .705 r
F P
0 g
s
.43262 .43622 m .43999 .43589 L .43515 .43479 L .956 .86 .705 r
F
.44914 .43357 m .47353 .42898 L .47546 .44368 L p .966 .911 .706 r
F P
0 g
s
.44835 .45756 m .4642 .43936 L .4599 .4377 L .994 .841 .584 r
F
.4406 .47969 m .45202 .47121 L p .43276 .49188 L .885 .99 .745 r
F P
0 g
s
.45202 .47121 m .4406 .47969 L p .46069 .45325 L .966 .911 .706 r
F P
0 g
s
.4669 .4404 m .44835 .45756 L .4457 .48428 L p .948 .949 .642 r
F P
0 g
s
.44835 .45756 m .4669 .4404 L p .4642 .43936 L .994 .841 .584 r
F P
0 g
s
.47667 .4529 m .47353 .42898 L p .47511 .42711 L .987 .882 .54 r
F P
0 g
s
.4882 .43494 m .47348 .42545 L p .46706 .43203 L .863 .435 .195 r
F P
0 g
s
.46706 .43203 m .4694 .44063 L .4882 .43494 L .863 .435 .195 r
F
.45532 .44025 m .46773 .4345 L .46962 .44144 L .828 .407 .241 r
F
.47226 .45114 m .45433 .44266 L p .47009 .44318 L F P
0 g
s
.44501 .45459 m .4608 .44741 L p .45415 .44524 L .863 .435 .195 r
F P
0 g
s
.45213 .44459 m .438 .43999 L .44501 .45459 L .863 .435 .195 r
F
.43202 .43655 m .43226 .43675 L .43262 .43622 L .956 .86 .705 r
F
.43878 .43274 m .44177 .4405 L p .44823 .43568 L .994 .841 .584 r
F P
0 g
s
.45442 .4427 m .47226 .45114 L p .45975 .44664 L .631 .067 0 r
F P
0 g
s
.45202 .47121 m .47667 .4529 L .47465 .45834 L p .885 .99 .745 r
F P
0 g
s
.47465 .45834 m .46778 .47686 L p .4634 .47529 L .885 .99 .745 r
F P
0 g
s
.4634 .47529 m .45202 .47121 L .47465 .45834 L .885 .99 .745 r
F
.46533 .48344 m .46778 .47686 L p .48032 .45055 L .754 .844 .394 r
F P
0 g
s
.48151 .44892 m .46533 .48344 L .48032 .45055 L .754 .844 .394 r
F
.47667 .4529 m .46069 .45325 L .46214 .45026 L .966 .911 .706 r
F
.47667 .4529 m .45202 .47121 L p .46069 .45325 L F P
0 g
s
.48208 .44814 m .46533 .48344 L p .48201 .44822 L .754 .844 .394 r
F P
0 g
s
.49476 .46943 m .49337 .45803 L .47828 .48335 L .737 .57 .667 r
F
.49008 .45216 m .45477 .52281 L p .49337 .45803 L F P
0 g
s
.4608 .44741 m .44501 .45459 L .47896 .46589 L p .693 .176 0 r
F P
0 g
s
.47896 .46589 m .49184 .47018 L p .4873 .46685 L .693 .176 0 r
F P
0 g
s
.4608 .44741 m .47896 .46589 L .4873 .46685 L .693 .176 0 r
F
.4873 .46685 m .49743 .46303 L .52316 .46858 L p .401 0 0 r
F P
0 g
s
.52316 .46858 m .54665 .47364 L p .5363 .47246 L .401 0 0 r
F P
0 g
s
.4873 .46685 m .52316 .46858 L .5363 .47246 L .401 0 0 r
F
.58746 .48863 m .59171 .52239 L .57296 .48727 L .426 .407 .735 r
F
.5513 .47786 m .5632 .52195 L .55586 .49361 L .773 .477 0 r
F
.55236 .47889 m .5632 .52195 L .5515 .47742 L .009 .089 .6 r
F
.5632 .52195 m .55085 .47633 L p .5515 .47742 L F P
0 g
s
.55105 .48531 m .5628 .53193 L .54554 .52395 L p .835 .63 .625 r
F P
0 g
s
.55172 .48794 m .55105 .48531 L p .55117 .48552 L .904 .666 .569 r
F P
0 g
s
.5363 .47246 m .54038 .48633 L .59368 .49205 L p .592 .095 .044 r
F P
0 g
s
.59368 .49205 m .6078 .49357 L p .59791 .49065 L .592 .095 .044 r
F P
0 g
s
.5363 .47246 m .59368 .49205 L .59791 .49065 L .592 .095 .044 r
F
.59791 .49065 m .57476 .47971 L .62963 .48011 L p .436 0 0 r
F P
0 g
s
.47226 .45114 m .49743 .46303 L p .46351 .44939 L .631 .067 0 r
F P
0 g
s
.45312 .44491 m .45213 .44459 L .44501 .45459 L .863 .435 .195 r
F
.43057 .43615 m .45437 .44269 L .45975 .44664 L .631 .067 0 r
F
.45437 .44269 m .45442 .4427 L .45975 .44664 L F
.43057 .43615 m .47226 .45114 L .4491 .44361 L F
.4479 .43962 m .46584 .42754 L p .46704 .43195 L .828 .407 .241 r
F P
0 g
s
.46638 .42907 m .46628 .42918 L .47348 .42545 L .863 .435 .195 r
F
.46177 .43406 m .46704 .43195 L .45532 .44025 L .828 .407 .241 r
F
.46704 .43195 m .46773 .4345 L .45532 .44025 L F
.4608 .44741 m .47226 .45114 L .46351 .44939 L .631 .067 0 r
F
.47254 .45127 m .47226 .45114 L p .47717 .44341 L .828 .407 .241 r
F P
0 g
s
.47009 .44318 m .47717 .44341 L .47226 .45114 L .828 .407 .241 r
F
.45433 .44266 m .4479 .43962 L p .46962 .44144 L F P
0 g
s
.46962 .44144 m .47009 .44318 L .45433 .44266 L .828 .407 .241 r
F
.46628 .42918 m .44501 .45459 L .46706 .43203 L .863 .435 .195 r
F
.42048 .42746 m .45279 .42508 L p .42214 .42877 L .983 .809 .589 r
F P
0 g
s
.42214 .42877 m .45279 .42508 L .45256 .42563 L .983 .809 .589 r
F
.43548 .43188 m .4214 .42819 L p .43262 .43622 L .956 .86 .705 r
F P
0 g
s
.43262 .43622 m .4214 .42819 L .4307 .43552 L .956 .86 .705 r
F
.43202 .43655 m .43262 .43622 L .4307 .43552 L F
.43878 .43274 m .43548 .43188 L p .43262 .43622 L F P
0 g
s
.4479 .43962 m .41919 .43388 L p .41879 .43208 L .881 .453 .167 r
F P
0 g
s
.42261 .43499 m .43057 .43615 L .4491 .44361 L .631 .067 0 r
F
.43878 .43274 m .4479 .43962 L p .41835 .43016 L .881 .453 .167 r
F P
0 g
s
.39399 .42235 m .4214 .42819 L .4045 .42572 L .881 .453 .167 r
F
.39069 .4211 m .4214 .42819 L .39399 .42235 L F
.3854 .41876 m .4214 .42819 L p .38859 .42062 L F P
0 g
s
.40376 .42869 m .42048 .42746 L p .42214 .42877 L .983 .809 .589 r
F P
0 g
s
.41779 .42766 m .40376 .42869 L p .41816 .42932 L .863 .435 .195 r
F P
0 g
s
.41797 .42848 m .43878 .43274 L .41835 .43016 L .881 .453 .167 r
F
.4214 .42819 m .43878 .43274 L p .41797 .42848 L F P
0 g
s
.41779 .42766 m .41797 .42848 L .4045 .42572 L .881 .453 .167 r
F
.4214 .42819 m .41797 .42848 L .41779 .42766 L F
.40376 .42869 m .40431 .42904 L p .4182 .42949 L .863 .435 .195 r
F P
0 g
s
.43319 .42997 m .41816 .42932 L .4182 .42949 L .863 .435 .195 r
F
.41816 .42932 m .40376 .42869 L .4182 .42949 L F
.40431 .42904 m .4136 .43487 L p .41848 .43365 L F P
0 g
s
.4182 .42949 m .40431 .42904 L .4191 .43349 L .863 .435 .195 r
F
.39495 .42432 m .39739 .42574 L .3809 .42227 L .881 .453 .167 r
F
.40376 .42869 m .39495 .42432 L .40468 .42999 L .56 0 0 r
F
.41835 .43016 m .41879 .43208 L .3809 .42227 L .881 .453 .167 r
F
.438 .43999 m .41984 .43409 L .44501 .45459 L .863 .435 .195 r
F
.4136 .43487 m .42038 .43913 L p .41919 .43388 L F P
0 g
s
.4191 .43349 m .41848 .43365 L .41919 .43388 L .863 .435 .195 r
F
.41848 .43365 m .4136 .43487 L .41919 .43388 L F
.40376 .42869 m .40475 .43007 L .40651 .43087 L .56 0 0 r
F
.409 .43198 m .40431 .42904 L p .40651 .43087 L F P
0 g
s
.40431 .42904 m .40376 .42869 L p .40651 .43087 L .56 0 0 r
F P
0 g
s
.39954 .42994 m .40726 .43149 L p .39221 .42897 L .631 .067 0 r
F P
0 g
s
.40726 .43149 m .4076 .43177 L .39221 .42897 L .631 .067 0 r
F
.4076 .43177 m .40773 .43188 L .39221 .42897 L F
.40651 .43087 m .40922 .43311 L .409 .43198 L .56 0 0 r
F
.41919 .43388 m .43057 .43615 L p .42026 .43422 L .631 .067 0 r
F P
0 g
s
.42038 .43913 m .44501 .45459 L p .4191 .43349 L .863 .435 .195 r
F P
0 g
s
.41553 .43334 m .42026 .43422 L .40884 .4328 L .631 .067 0 r
F
.40862 .43262 m .40884 .4328 L .40104 .43183 L F
.41553 .43334 m .40884 .4328 L .40862 .43262 L F
.42026 .43422 m .42261 .43499 L .40104 .43183 L F
.39593 .4313 m .41373 .43686 L .42098 .44289 L F
.42138 .44322 m .41925 .44222 L p .42098 .44289 L F P
0 g
s
.41925 .44222 m .39593 .4313 L p .42098 .44289 L .631 .067 0 r
F P
0 g
s
.44501 .45459 m .42098 .44289 L p .46886 .46133 L 0 .515 .828 r
F P
0 g
s
.44501 .45459 m .46886 .46133 L .48522 .46763 L 0 .515 .828 r
F
.49184 .47018 m .47896 .46589 L p .48522 .46763 L F P
0 g
s
.47896 .46589 m .44501 .45459 L p .48522 .46763 L 0 .515 .828 r
F P
0 g
s
.53281 .47761 m .53732 .48539 L .53449 .51884 L .835 .63 .625 r
F
.5305 .47363 m .53449 .51884 L .49855 .5113 L p .768 .58 .644 r
F P
0 g
s
.52498 .47116 m .50459 .46883 L .54038 .48633 L .693 .176 0 r
F
.49743 .46303 m .4873 .46685 L p .47365 .45684 L .631 .067 0 r
F P
0 g
s
.41373 .43686 m .49743 .46303 L .42138 .44322 L .631 .067 0 r
F
.41191 .43535 m .42261 .43499 L .49743 .46303 L F
.42261 .43499 m .4491 .44361 L .49743 .46303 L F
.40104 .43183 m .49743 .46303 L .39593 .4313 L F
.49476 .46943 m .47828 .48335 L .45477 .52281 L .737 .57 .667 r
F
.48846 .50918 m .44802 .50889 L p .49606 .46834 L .706 .517 .631 r
F P
0 g
s
.50056 .4709 m .51121 .48994 L .48846 .50918 L .768 .58 .644 r
F
.49184 .47018 m .54038 .48633 L p .4608 .44741 L .693 .176 0 r
F P
0 g
s
.52415 .42457 m .51118 .4291 L .52938 .4489 L .795 .41 .325 r
F
.46628 .42918 m .46706 .43203 L .47348 .42545 L .863 .435 .195 r
F
.46704 .43195 m .51354 .41331 L .46773 .4345 L .828 .407 .241 r
F
.51704 .44472 m .52938 .4489 L p .51118 .4291 L .795 .41 .325 r
F P
0 g
s
.49743 .46303 m .50351 .44427 L p .51036 .44894 L .712 .271 .202 r
F P
0 g
s
.50807 .43018 m .51704 .44472 L .51118 .4291 L .795 .41 .325 r
F
.55369 .47759 m .54704 .47563 L .57476 .47971 L .401 0 0 r
F
.54038 .48633 m .5363 .47246 L p .52498 .47116 L .693 .176 0 r
F P
0 g
s
.54665 .47364 m .57476 .47971 L p .4873 .46685 L .401 0 0 r
F P
0 g
s
.53602 .42151 m .53855 .43086 L p .53794 .4284 L .773 .477 0 r
F P
0 g
s
.53855 .43086 m .53602 .42151 L p .53844 .43027 L .009 .089 .6 r
F P
0 g
s
.5362 .42168 m .53602 .42151 L .53794 .4284 L .773 .477 0 r
F
.5362 .42168 m .53794 .4284 L .53703 .42469 L F
.53704 .42244 m .53844 .43027 L .53602 .42151 L .009 .089 .6 r
F
.53099 .42643 m .53673 .42018 L .54876 .43836 L .712 .271 .202 r
F
.51036 .44894 m .50807 .43018 L .53673 .42018 L F
.53804 .41875 m .53852 .41955 L p .53673 .42018 L F P
0 g
s
.53852 .41955 m .56092 .45674 L p .53673 .42018 L .712 .271 .202 r
F P
0 g
s
.53192 .39319 m .53804 .41875 L p .53106 .39397 L .76 .42 .409 r
F P
0 g
s
.52648 .38301 m .52268 .38258 L p .53401 .39131 L .699 .333 .35 r
F P
0 g
s
.53401 .39131 m .54032 .39618 L .53462 .39076 L .699 .333 .35 r
F
.53401 .39131 m .53462 .39076 L .52648 .38301 L F
.52378 .39749 m .52462 .40698 L p .51875 .41338 L .768 .58 .644 r
F P
0 g
s
.53804 .41875 m .51354 .41331 L p .51415 .40923 L .76 .42 .409 r
F P
0 g
s
.51415 .40923 m .52443 .39996 L .53804 .41875 L .76 .42 .409 r
F
.53054 .39217 m .53433 .40722 L .5341 .41123 L .009 .089 .6 r
F
.53433 .40722 m .53592 .41356 L .53532 .41279 L F
.53433 .40722 m .53532 .41279 L .5341 .41123 L F
.52966 .40039 m .52378 .39749 L p .52327 .38582 L .844 .604 .575 r
F P
0 g
s
.52443 .39996 m .53076 .39424 L .53721 .41596 L .76 .42 .409 r
F
.52443 .39996 m .53721 .41596 L .53804 .41875 L F
.53804 .41875 m .53192 .39319 L p .5328 .39383 L .673 .365 .443 r
F P
0 g
s
.53076 .39424 m .53106 .39397 L .53804 .41875 L .76 .42 .409 r
F
.53942 .40234 m .53942 .40237 L .53938 .40232 L .328 .278 .649 r
F
.53938 .40232 m .53942 .40237 L .53943 .40246 L F
.53804 .41875 m .5328 .39383 L .54163 .40025 L .673 .365 .443 r
F
.53804 .41875 m .54011 .41876 L p .54559 .42113 L .631 .248 .299 r
F P
0 g
s
.53126 .39379 m .52683 .42363 L .57751 .45417 L .699 .333 .35 r
F
.5294 .39136 m .52905 .39579 L .53126 .39379 L F
.52905 .39579 m .52683 .42363 L .53126 .39379 L F
.52268 .38258 m .5231 .38678 L p .5294 .39136 L F P
0 g
s
.5231 .38678 m .52683 .42363 L p .5294 .39136 L .699 .333 .35 r
F P
0 g
s
.52268 .38258 m .51451 .37971 L p .51661 .38951 L .795 .41 .325 r
F P
0 g
s
.51112 .37744 m .51451 .37971 L p .51745 .36014 L .894 .587 .457 r
F P
0 g
s
.51881 .35256 m .51597 .35875 L p .51584 .35969 L .999 .806 .468 r
F P
0 g
s
.51555 .36186 m .5165 .35915 L .51584 .35969 L .999 .806 .468 r
F
.5165 .35915 m .51881 .35256 L .51584 .35969 L F
.51753 .3566 m .51469 .36379 L p .51768 .35569 L .919 .605 .433 r
F P
0 g
s
.51768 .35569 m .51469 .36379 L .51497 .3627 L .919 .605 .433 r
F
.51905 .35327 m .51886 .35466 L p .52106 .3375 L 0 0 0 r
F P
0 g
s
.5153 .36369 m .51905 .35327 L p .51538 .36312 L .999 .806 .468 r
F P
0 g
s
.51563 .3601 m .51785 .35465 L .51768 .35569 L .919 .605 .433 r
F
.51597 .35875 m .51785 .35465 L p .51563 .3601 L F P
0 g
s
.51905 .35327 m .51881 .35256 L p .51555 .36186 L .999 .806 .468 r
F P
0 g
s
.51538 .36312 m .51905 .35327 L .51555 .36186 L .999 .806 .468 r
F
.51768 .35569 m .51844 .35359 L .51833 .35429 L .919 .605 .433 r
F
.51844 .35359 m .51881 .35256 L .51833 .35429 L F
.51881 .35256 m .51905 .35275 L p .51833 .35429 L F P
0 g
s
.51785 .35465 m .51881 .35256 L p .51768 .35569 L .919 .605 .433 r
F P
0 g
s
.51645 .36164 m .51694 .36076 L .51831 .35563 L .863 .83 .356 r
F
.51688 .36097 m .51722 .35978 L .51831 .35563 L F
.51567 .36173 m .51112 .37744 L .51745 .36014 L .894 .587 .457 r
F
.51831 .35563 m .51696 .35909 L .5167 .3601 L .863 .83 .356 r
F
.51257 .37181 m .51753 .3566 L .51668 .36214 L .852 .512 .409 r
F
.51469 .36379 m .51542 .36195 L p .51257 .37181 L F P
0 g
s
.51542 .36195 m .51753 .3566 L p .51594 .36149 L .852 .512 .409 r
F P
0 g
s
.51542 .36195 m .51548 .36237 L .51567 .36173 L .894 .587 .457 r
F
.51542 .36195 m .51594 .36149 L .51257 .37181 L .852 .512 .409 r
F
.51476 .36195 m .51563 .3601 L .51376 .36512 L .919 .605 .433 r
F
.51476 .36195 m .51597 .35875 L .51563 .3601 L F
.51522 .36058 m .51522 .36053 L p .51532 .36011 L .995 .756 .431 r
F P
0 g
s
.51535 .3601 m .51597 .35875 L .51476 .36195 L .919 .605 .433 r
F
.51563 .3601 m .51768 .35569 L .51497 .3627 L F
.51563 .3601 m .51497 .3627 L .50728 .38257 L F
.51521 .36052 m .51542 .36195 L .51889 .35062 L .894 .587 .457 r
F
.51542 .36195 m .51567 .36173 L .51889 .35062 L F
.51905 .35275 m .51753 .3566 L p .51768 .35569 L .919 .605 .433 r
F P
0 g
s
.52106 .33751 m .52106 .3375 L .52105 .33756 L .945 .734 .586 r
F
.52106 .3375 m .52107 .33749 L .52105 .33756 L F
.52065 .34035 m .52106 .3375 L .521 .33756 L .853 .553 .481 r
F
.51884 .35265 m .51905 .35327 L p .51945 .3501 L .987 .767 .373 r
F P
0 g
s
.52106 .3375 m .52106 .3375 L .51905 .35327 L 0 0 0 r
F
.51945 .3501 m .52099 .33815 L .51884 .35265 L .987 .767 .373 r
F
.51905 .35275 m .52106 .3375 L p .52065 .34035 L .853 .553 .481 r
F P
0 g
s
.51884 .35265 m .52099 .33815 L .52021 .3433 L .987 .767 .373 r
F
.51881 .35256 m .51884 .35265 L p .52021 .3433 L F P
0 g
s
.51848 .35661 m .51825 .35663 L p .51815 .35752 L .667 .459 .595 r
F P
0 g
s
.51905 .35327 m .51863 .35442 L p .51831 .35563 L .863 .83 .356 r
F P
0 g
s
.51905 .35327 m .52106 .3375 L p .52106 .33753 L .987 .767 .373 r
F P
0 g
s
.51886 .35466 m .51905 .35327 L p .51831 .35563 L .863 .83 .356 r
F P
0 g
s
.51702 .36055 m .51886 .35466 L p .51831 .35563 L .863 .83 .356 r
F P
0 g
s
.51905 .35275 m .51983 .35315 L p .51735 .36023 L .852 .512 .409 r
F P
0 g
s
.51711 .36045 m .51905 .35275 L .51735 .36023 L .852 .512 .409 r
F
.52098 .33672 m .52098 .33673 L p .51532 .36011 L .995 .756 .431 r
F P
0 g
s
.52106 .3375 m .52104 .33735 L .52105 .33739 L .464 .275 .539 r
F
.52106 .33746 m .52106 .33746 L .52106 .33746 L .918 .624 .472 r
F
.52106 .33746 m .52106 .33746 L .52106 .33746 L F
.52106 .3375 m .52105 .33739 L .52106 .33742 L .464 .275 .539 r
F
.52106 .33749 m .52106 .3375 L .52106 .3375 L F
.52106 .33741 m .52105 .33737 L .52106 .33749 L F
.52104 .33724 m .52109 .3369 L .52105 .33737 L .853 .553 .481 r
F
.52107 .33729 m .52105 .33711 L .52106 .3375 L .886 .627 .54 r
F
.52105 .33737 m .52106 .33749 L .52106 .3375 L .464 .275 .539 r
F
.52109 .3369 m .52101 .33688 L p .521 .33715 L .769 .477 .491 r
F P
0 g
s
.52101 .33674 m .52106 .3375 L p .52105 .33711 L .886 .627 .54 r
F P
0 g
s
.52106 .3375 m .52103 .33722 L .52105 .33728 L .311 .178 .533 r
F
.52106 .33741 m .52106 .33749 L .52106 .3375 L .464 .275 .539 r
F
.52106 .3375 m .52104 .33682 L .52109 .33693 L .808 .565 .571 r
F
.52107 .33744 m .52106 .33741 L .52106 .3375 L .464 .275 .539 r
F
.52106 .3375 m .52108 .33677 L .5211 .33677 L p .736 .504 .577 r
F P
0 g
s
.52106 .3375 m .52105 .33728 L .52108 .33737 L .311 .178 .533 r
F
.52099 .33678 m .52098 .33673 L p .52099 .33678 L 0 0 .073 r
F P
0 g
s
.52098 .33672 m .52101 .33674 L p .52107 .33729 L .886 .627 .54 r
F P
0 g
s
.52106 .3375 m .52101 .33674 L p .52104 .33682 L .808 .565 .571 r
F P
0 g
s
.52105 .33724 m .52102 .33708 L .52106 .3375 L .311 .178 .533 r
F
.52099 .33678 m .52106 .3375 L p .52103 .33705 L 0 0 .452 r
F P
0 g
s
.52103 .33705 m .52103 .33713 L .52099 .33678 L 0 0 .452 r
F
.52104 .33721 m .52103 .33713 L .52103 .33705 L F
.521 .33683 m .52099 .33678 L .52099 .33678 L 0 0 .073 r
F
.52106 .3375 m .52099 .33678 L p .521 .33683 L F P
0 g
s
.52097 .33683 m .52099 .33678 L p .52103 .33713 L 0 0 .452 r
F P
0 g
s
.52066 .33813 m .51977 .34249 L .52097 .33683 L .141 0 .344 r
F
.52098 .33673 m .52099 .33678 L p .52028 .33995 L .941 .796 .337 r
F P
0 g
s
.52098 .33673 m .52098 .33672 L p .52099 .33677 L .977 .7 .306 r
F P
0 g
s
.52106 .33748 m .52099 .33674 L .52099 .33677 L .977 .7 .306 r
F
.52099 .33674 m .52098 .33673 L .52099 .33677 L F
.52106 .33749 m .52106 .3375 L .52106 .33749 L .464 .275 .539 r
F
.52097 .33683 m .52106 .3375 L p .52102 .33708 L .311 .178 .533 r
F P
0 g
s
.52105 .33783 m .52106 .3375 L .52107 .33832 L .769 .477 .491 r
F
.52105 .33783 m .52107 .33832 L .52114 .3536 L F
.52114 .3536 m .52039 .35343 L p .52105 .33783 L F P
0 g
s
.52103 .33784 m .52104 .33773 L .52107 .33788 L .687 .415 .505 r
F
.51881 .35256 m .52106 .3375 L p .52105 .3375 L .918 .624 .472 r
F P
0 g
s
.52106 .3375 m .52096 .33699 L .52103 .33722 L .311 .178 .533 r
F
.52106 .3375 m .52094 .33691 L p .52096 .33699 L F P
0 g
s
.521 .33756 m .52106 .3375 L .52101 .33753 L .853 .553 .481 r
F
.52106 .3375 m .51881 .35256 L p .52105 .33759 L .987 .767 .373 r
F P
0 g
s
.51905 .35275 m .52065 .34035 L .52106 .3375 L .853 .553 .481 r
F
.52065 .34035 m .521 .33756 L .52106 .3375 L F
.52096 .33825 m .52106 .33752 L .51881 .35256 L .987 .767 .373 r
F
.52099 .33815 m .52103 .33783 L .51881 .35256 L F
.51905 .35275 m .51881 .35256 L p .52105 .3375 L .918 .624 .472 r
F P
0 g
s
.52105 .3375 m .52105 .3375 L .51905 .35275 L .918 .624 .472 r
F
.52106 .3375 m .52106 .3375 L .52106 .33752 L .987 .767 .373 r
F
.52039 .35343 m .52095 .33833 L .52106 .3375 L .769 .477 .491 r
F
.52098 .33765 m .52103 .33732 L .521 .33715 L F
.52106 .3375 m .51905 .35275 L p .52106 .3375 L .918 .624 .472 r
F P
0 g
s
.51983 .35315 m .51911 .35278 L p .52106 .3375 L .853 .553 .481 r
F P
0 g
s
.52092 .3392 m .52104 .33754 L .52098 .33765 L .769 .477 .491 r
F
.51911 .35278 m .51905 .35275 L p .52106 .3375 L .853 .553 .481 r
F P
0 g
s
.52106 .33746 m .52106 .33747 L .51905 .35275 L .918 .624 .472 r
F
.52106 .33744 m .52106 .33749 L .51905 .35275 L F
.52039 .35343 m .52106 .3375 L p .52092 .3392 L .769 .477 .491 r
F P
0 g
s
.52106 .3375 m .52039 .35343 L .51983 .35315 L p .853 .553 .481 r
F P
0 g
s
.52101 .33688 m .52094 .33688 L p .52095 .33701 L .687 .415 .505 r
F P
0 g
s
.52108 .33737 m .52094 .33691 L .52097 .33683 L .311 .178 .533 r
F
.52101 .33674 m .52097 .33728 L .52192 .36165 L .81 .496 .459 r
F
.51758 .36003 m .52039 .35343 L .51943 .35838 L .852 .512 .409 r
F
.52127 .36226 m .51975 .35901 L .52145 .36103 L .76 .42 .409 r
F
.51915 .36423 m .51975 .35901 L p .52127 .36226 L F P
0 g
s
.51975 .35901 m .52039 .35343 L p .52145 .36103 L .76 .42 .409 r
F P
0 g
s
.51983 .35315 m .52039 .35343 L p .5198 .3548 L .852 .512 .409 r
F P
0 g
s
.51451 .37971 m .51747 .36013 L p .51745 .36014 L .894 .587 .457 r
F P
0 g
s
.51668 .36214 m .51983 .35315 L .5198 .3548 L .852 .512 .409 r
F
.52097 .33728 m .51915 .36423 L .52037 .36309 L .81 .496 .459 r
F
.52127 .36226 m .52145 .36103 L .52195 .36162 L .76 .42 .409 r
F
.52097 .33728 m .52037 .36309 L .52192 .36165 L .81 .496 .459 r
F
.52106 .3375 m .52105 .33719 L .52107 .33729 L .886 .627 .54 r
F
.52106 .33725 m .52105 .33724 L .52106 .3375 L .311 .178 .533 r
F
.52107 .3373 m .52106 .33725 L .52106 .3375 L F
.52107 .33735 m .52106 .3375 L .52108 .33738 L F
.52107 .3373 m .52106 .3375 L .52107 .33735 L F
.52106 .3375 m .52097 .33683 L p .52108 .3374 L 0 0 .452 r
F P
0 g
s
.52106 .3375 m .52098 .33673 L p .52106 .33748 L .977 .7 .306 r
F P
0 g
s
.52106 .3375 m .52099 .33679 L .52102 .33701 L .886 .627 .54 r
F
.52106 .3375 m .521 .33683 L .52106 .3375 L 0 0 .073 r
F
.52098 .33673 m .52106 .3375 L p .521 .33683 L F P
0 g
s
.52106 .3375 m .52098 .33672 L p .52099 .33679 L .886 .627 .54 r
F P
0 g
s
.52106 .3375 m .52102 .33701 L .52105 .33719 L .886 .627 .54 r
F
.52108 .3374 m .52104 .33721 L .52103 .33705 L 0 0 .452 r
F
.51688 .36097 m .51702 .36055 L p .51722 .35978 L .863 .83 .356 r
F P
0 g
s
.51753 .3566 m .51905 .35275 L p .51668 .36214 L .852 .512 .409 r
F P
0 g
s
.52097 .33711 m .52093 .33697 L .52093 .33695 L .464 .275 .539 r
F
.52097 .33711 m .52093 .33695 L .52094 .33691 L F
.52098 .33672 m .51745 .36014 L .521 .33684 L .894 .587 .457 r
F
.52101 .33674 m .52098 .33673 L p .521 .33684 L F P
0 g
s
.52098 .33673 m .52098 .33672 L p .521 .33684 L .894 .587 .457 r
F P
0 g
s
.51747 .36013 m .52101 .33674 L p .51745 .36014 L .894 .587 .457 r
F P
0 g
s
.51668 .36214 m .51711 .36045 L .51735 .36023 L .852 .512 .409 r
F
.51889 .35062 m .51567 .36173 L .51745 .36014 L .894 .587 .457 r
F
.51668 .36214 m .51758 .36003 L .51303 .39128 L .852 .512 .409 r
F
.51758 .36003 m .51943 .35838 L .51303 .39128 L F
.51214 .39172 m .51303 .39128 L .50916 .41117 L F
.51668 .36214 m .51303 .39128 L .51214 .39172 L F
.51661 .38951 m .51681 .39041 L .51798 .38884 L .795 .41 .325 r
F
.51661 .38951 m .51798 .38884 L .52268 .38258 L F
.51415 .40923 m .51697 .39061 L .52443 .39996 L .76 .42 .409 r
F
.51681 .39041 m .52415 .42457 L .52323 .39845 L .795 .41 .325 r
F
.51681 .39041 m .52323 .39845 L .52268 .38258 L F
.53192 .39319 m .535 .39867 L .53378 .39454 L .699 .333 .35 r
F
.5294 .39136 m .53126 .39379 L .53192 .39319 L F
.53126 .39379 m .535 .39867 L .53192 .39319 L F
.52534 .36566 m .52253 .35392 L p .52646 .36363 L .673 .365 .443 r
F P
0 g
s
.52404 .35393 m .52253 .35392 L p .52172 .34113 L .687 .415 .505 r
F P
0 g
s
.52164 .34475 m .52373 .3827 L .52411 .35035 L .724 .426 .472 r
F
.52118 .33817 m .52111 .33764 L .52114 .33791 L 0 0 .452 r
F
.52108 .33677 m .52268 .38258 L .52373 .3827 L p .724 .426 .472 r
F P
0 g
s
.52106 .33754 m .52106 .3375 L .52106 .33752 L .977 .7 .306 r
F
.52108 .33754 m .52106 .3375 L .52109 .33763 L 0 0 .452 r
F
.52103 .33705 m .52106 .3375 L .52108 .33754 L F
.52105 .3375 m .52106 .3375 L .52106 .33746 L .918 .624 .472 r
F
.52101 .33674 m .52108 .33677 L p .52109 .33693 L .808 .565 .571 r
F P
0 g
s
.52098 .33672 m .52106 .3375 L p .52106 .33753 L .977 .7 .306 r
F P
0 g
s
.52111 .33752 m .52106 .3375 L .52108 .33737 L .311 .178 .533 r
F
.52111 .33764 m .52103 .33705 L .52108 .33754 L 0 0 .452 r
F
.52108 .33765 m .52106 .3375 L .521 .33683 L 0 0 .073 r
F
.52105 .3375 m .52106 .33746 L .52105 .3375 L .918 .624 .472 r
F
.52108 .33737 m .52107 .3373 L .52108 .33738 L .311 .178 .533 r
F
.52106 .33752 m .52106 .3375 L .52106 .33748 L .977 .7 .306 r
F
.52119 .33821 m .52106 .3375 L .52108 .3374 L 0 0 .452 r
F
.52268 .38258 m .52108 .33677 L .52101 .33674 L p .81 .496 .459 r
F P
0 g
s
.52443 .39996 m .52127 .36226 L .53076 .39424 L .76 .42 .409 r
F
.52683 .42363 m .52268 .38258 L p .52415 .42457 L .795 .41 .325 r
F P
0 g
s
.51697 .39061 m .52127 .36226 L .52443 .39996 L .76 .42 .409 r
F
.51354 .41331 m .50807 .43018 L p .50216 .4206 L .828 .407 .241 r
F P
0 g
s
.47009 .44318 m .4882 .43494 L p .4694 .44063 L .863 .435 .195 r
F P
0 g
s
.46773 .4345 m .51354 .41331 L .46962 .44144 L .828 .407 .241 r
F
.47348 .42545 m .47084 .42375 L p .46734 .42792 L .863 .435 .195 r
F P
0 g
s
.46734 .42792 m .46638 .42907 L .46821 .42814 L .863 .435 .195 r
F
.46734 .42792 m .46821 .42814 L .47348 .42545 L F
.4479 .43962 m .44726 .43914 L p .46519 .42738 L .916 .564 .354 r
F P
0 g
s
.46519 .42738 m .47614 .4202 L .46557 .42747 L .916 .564 .354 r
F
.46519 .42738 m .46557 .42747 L .4479 .43962 L F
.46704 .43195 m .4709 .42413 L .51354 .41331 L .828 .407 .241 r
F
.51126 .4122 m .51354 .41331 L p .50293 .41339 L F P
0 g
s
.51354 .41331 m .50916 .41117 L p .51857 .3628 L .852 .512 .409 r
F P
0 g
s
.51915 .36423 m .51354 .41331 L p .51857 .3628 L .852 .512 .409 r
F P
0 g
s
.51681 .39041 m .51118 .4291 L .52415 .42457 L .795 .41 .325 r
F
.51661 .38951 m .51654 .39007 L .51681 .39041 L F
.51654 .39007 m .51118 .4291 L .51681 .39041 L F
.5328 .39383 m .52534 .36566 L .52646 .36363 L .673 .365 .443 r
F
.53192 .39319 m .52534 .36566 L p .5328 .39383 L F P
0 g
s
.5713 .45349 m .57751 .45417 L p .52683 .42363 L .699 .333 .35 r
F P
0 g
s
.535 .39867 m .55871 .42962 L .53739 .39717 L .699 .333 .35 r
F
.55871 .42962 m .574 .44958 L .55943 .42873 L F
.55871 .42962 m .55943 .42873 L .53739 .39717 L F
.55943 .42873 m .574 .44958 L .56023 .42775 L F
.56945 .4416 m .54032 .39618 L p .574 .44958 L F P
0 g
s
.56922 .45956 m .55059 .43959 L .56042 .42752 L .631 .248 .299 r
F
.57476 .47971 m .56092 .45674 L p .56922 .45956 L F P
0 g
s
.56092 .45674 m .55059 .43959 L p .56922 .45956 L .631 .248 .299 r
F P
0 g
s
.56945 .4416 m .57751 .45417 L p .55061 .40759 L .622 .315 .429 r
F P
0 g
s
.57476 .47971 m .56042 .42752 L .56204 .42822 L .631 .248 .299 r
F
.57476 .47971 m .60244 .49279 L p .546 .48449 L .401 0 0 r
F P
0 g
s
.50271 .47248 m .57476 .47971 L .546 .48449 L .401 0 0 r
F
.554 .45891 m .5363 .47246 L .65047 .49815 L .592 .095 .044 r
F
.6078 .49357 m .65047 .49815 L p .5363 .47246 L F P
0 g
s
.65047 .49815 m .60878 .47302 L p .5581 .45763 L .592 .095 .044 r
F P
0 g
s
.554 .45891 m .65047 .49815 L .5581 .45763 L .592 .095 .044 r
F
.624 .4946 m .59791 .49065 L .64311 .4802 L .436 0 0 r
F
.65047 .49815 m .6078 .49357 L p .64489 .49635 L .047 .605 .81 r
F P
0 g
s
.6078 .49357 m .54038 .48633 L p .58516 .49027 L .047 .605 .81 r
F P
0 g
s
.60244 .49279 m .59791 .49065 L p .624 .4946 L .436 0 0 r
F P
0 g
s
.54728 .48477 m .546 .48449 L p .60244 .49279 L .401 0 0 r
F P
0 g
s
.65593 .49728 m .63824 .48387 L .70691 .49404 L .436 0 0 r
F
.624 .4946 m .63824 .48387 L .65593 .49728 L F
.71904 .53623 m .67258 .52575 L .71407 .4984 L p .554 .503 .754 r
F P
0 g
s
.73924 .50063 m .69087 .50028 L p .70691 .49404 L .436 0 0 r
F P
0 g
s
.65047 .49815 m .691 .50023 L p .70011 .49668 L .27 0 0 r
F P
0 g
s
.62963 .48011 m .64311 .4802 L p .59791 .49065 L .436 0 0 r
F P
0 g
s
.63824 .48387 m .64225 .48086 L .68746 .49007 L .436 0 0 r
F
.70986 .49289 m .7213 .49188 L p .62377 .48205 L .568 .188 .282 r
F P
0 g
s
.76912 .49198 m .7785 .49169 L .78517 .49267 L .527 .173 .309 r
F
.79192 .49202 m .79147 .49207 L .79264 .49166 L .138 0 0 r
F
.78605 .49252 m .78517 .49267 L p .7785 .49169 L .527 .173 .309 r
F P
0 g
s
.7213 .49188 m .76912 .49198 L p .77972 .49308 L .488 .67 .945 r
F P
0 g
s
.78 .49165 m .78605 .49252 L p .7785 .49169 L .527 .173 .309 r
F P
0 g
s
.76912 .49198 m .79192 .49202 L p .7847 .49265 L .488 .67 .945 r
F P
0 g
s
.76912 .49198 m .7847 .49265 L .77972 .49308 L .488 .67 .945 r
F
.76912 .49198 m .7213 .49188 L p .76873 .49199 L .27 0 0 r
F P
0 g
s
.68129 .47798 m .7213 .49188 L .68423 .47916 L .211 .509 .91 r
F
.73032 .49316 m .76873 .49199 L .7213 .49188 L .27 0 0 r
F
.7213 .49188 m .68653 .47942 L p .70518 .48766 L .568 .188 .282 r
F P
0 g
s
.73924 .50063 m .70747 .48932 L .71975 .48985 L .527 .173 .309 r
F
.78517 .49267 m .78605 .49252 L p .78594 .4926 L .138 0 0 r
F P
0 g
s
.79147 .49207 m .78594 .4926 L .78605 .49252 L .138 0 0 r
F
.79192 .49202 m .78157 .49302 L .77972 .49308 L .488 .67 .945 r
F
.78517 .49267 m .74883 .49897 L p .71975 .48985 L .527 .173 .309 r
F P
0 g
s
.74883 .49897 m .73924 .50063 L p .71975 .48985 L .527 .173 .309 r
F P
0 g
s
.73924 .50063 m .70738 .49386 L p .71922 .4935 L .527 .173 .309 r
F P
0 g
s
.70738 .49386 m .73924 .50063 L p .70691 .49404 L .436 0 0 r
F P
0 g
s
.65047 .49815 m .76541 .49476 L .77289 .49352 L .27 0 0 r
F
.76541 .49476 m .78781 .4941 L .77516 .49361 L F
.76541 .49476 m .77516 .49361 L .77289 .49352 L F
.74764 .5118 m .71613 .5141 L .71774 .52633 L .504 .467 .749 r
F
.78022 .49645 m .74059 .51985 L .69273 .51951 L p .7 .66 .808 r
F P
0 g
s
.64171 .50424 m .67258 .52575 L .64773 .55716 L p .426 .407 .735 r
F P
0 g
s
.70835 .50362 m .69244 .51935 L p .7119 .51387 L .683 .578 .731 r
F P
0 g
s
.70835 .50362 m .69909 .50262 L p .68751 .51662 L .726 .65 .774 r
F P
0 g
s
.69244 .51935 m .70835 .50362 L p .68751 .51662 L .726 .65 .774 r
F P
0 g
s
.62852 .51905 m .67845 .51161 L .69244 .51935 L .7 .66 .808 r
F
.69191 .5195 m .62852 .51905 L p .69244 .51935 L F P
0 g
s
.59171 .52239 m .60761 .55219 L .59325 .53459 L .426 .407 .735 r
F
.60201 .53894 m .60993 .52183 L p .59181 .52315 L .287 .302 .701 r
F P
0 g
s
.61774 .51897 m .67651 .51054 L .67692 .51077 L .7 .66 .808 r
F
.66871 .50623 m .68751 .51662 L .69909 .50262 L .726 .65 .774 r
F
.77331 .49698 m .77914 .49661 L .78015 .49563 L .65 .401 .53 r
F
.77914 .49661 m .78022 .49645 L .69273 .51951 L .7 .66 .808 r
F
.69244 .51935 m .77914 .49661 L .69273 .51951 L F
.76628 .49853 m .77914 .49661 L .69244 .51935 L F
.7917 .50879 m .7898 .50892 L p .78628 .51086 L .58 .601 .837 r
F P
0 g
s
.7917 .50879 m .71908 .5365 L .78486 .51261 L .58 .601 .837 r
F
.78894 .51113 m .7917 .50879 L p .78486 .51261 L F P
0 g
s
.78484 .51216 m .78809 .5106 L .7388 .54314 L .4 .221 .52 r
F
.78988 .50974 m .7917 .50879 L p .78582 .51268 L .714 .94 .959 r
F P
0 g
s
.7917 .50879 m .78988 .50974 L p .79144 .50899 L 0 0 0 r
F P
0 g
s
.7917 .50879 m .78894 .51113 L p .78582 .51268 L .714 .94 .959 r
F P
0 g
s
.71734 .56876 m .71362 .5732 L p .70603 .58084 L .176 .06 .473 r
F P
0 g
s
.70847 .57859 m .71299 .5748 L .7062 .58125 L .714 .94 .959 r
F
.78894 .51113 m .71258 .57585 L p .71299 .5748 L F P
0 g
s
.71258 .57585 m .7062 .58125 L p .71299 .5748 L .714 .94 .959 r
F P
0 g
s
.74706 .53822 m .78894 .51113 L .71299 .5748 L .714 .94 .959 r
F
.71362 .5732 m .71299 .5748 L .70847 .57859 L F
.74706 .53822 m .71299 .5748 L .71362 .5732 L F
.67709 .57279 m .65645 .61277 L p .70481 .58213 L .551 .618 .873 r
F P
0 g
s
.7062 .58125 m .70576 .58112 L p .70481 .58213 L .551 .618 .873 r
F P
0 g
s
.70576 .58112 m .67709 .57279 L p .70481 .58213 L .551 .618 .873 r
F P
0 g
s
.71347 .57074 m .73863 .54066 L p .70603 .58084 L .176 .06 .473 r
F P
0 g
s
.70603 .58084 m .70537 .58164 L .70576 .58112 L .176 .06 .473 r
F
.70603 .58084 m .70576 .58112 L .71347 .57074 L F
.7062 .58125 m .65645 .61277 L .69654 .59086 L .551 .618 .873 r
F
.70293 .58737 m .7062 .58125 L p .698 .58941 L F P
0 g
s
.698 .58941 m .69654 .59086 L .69674 .59074 L .551 .618 .873 r
F
.698 .58941 m .69674 .59074 L .70293 .58737 L F
.70537 .58164 m .66877 .62676 L .7019 .58525 L .176 .06 .473 r
F
.66877 .62676 m .66866 .62433 L .69674 .59074 L p F P
0 g
s
.7019 .58525 m .66877 .62676 L .69674 .59074 L .176 .06 .473 r
F
.69612 .5941 m .69688 .5932 L p .67033 .6227 L .766 .985 .904 r
F P
0 g
s
.7062 .58125 m .70293 .58737 L p .69776 .59234 L .763 .892 .935 r
F P
0 g
s
.69214 .59776 m .67033 .6227 L .69688 .5932 L .766 .985 .904 r
F
.70436 .58289 m .70578 .58141 L .69427 .59533 L .176 .06 .473 r
F
.70156 .58699 m .69427 .59533 L p .70578 .58141 L F P
0 g
s
.69612 .5941 m .66877 .62676 L p .6876 .60355 L .176 .06 .473 r
F P
0 g
s
.70603 .58084 m .71362 .5732 L .70436 .58289 L .176 .06 .473 r
F
.70537 .58164 m .7019 .58525 L .70576 .58112 L F
.67801 .59166 m .70578 .58141 L p .69688 .5932 L .763 .892 .935 r
F P
0 g
s
.69167 .59822 m .67033 .6227 L .69214 .59776 L .766 .985 .904 r
F
.6694 .60569 m .68406 .59768 L .66866 .62433 L .522 .544 .819 r
F
.67801 .59166 m .69688 .5932 L .67145 .61771 L .763 .892 .935 r
F
.71362 .5732 m .71734 .56876 L p .71375 .57288 L .766 .985 .904 r
F P
0 g
s
.73863 .54066 m .71734 .56876 L .70603 .58084 L .176 .06 .473 r
F
.71734 .56876 m .7144 .57123 L .71375 .57288 L .766 .985 .904 r
F
.7388 .54314 m .71734 .56876 L p .73506 .54537 L .176 .06 .473 r
F P
0 g
s
.71734 .56876 m .7388 .54314 L p .7334 .54804 L .766 .985 .904 r
F P
0 g
s
.67709 .57279 m .70576 .58112 L p .73612 .54025 L .58 .601 .837 r
F P
0 g
s
.71951 .53634 m .73863 .54066 L .71347 .57074 L p .522 .544 .819 r
F P
0 g
s
.7334 .54804 m .71806 .56195 L .71734 .56876 L .766 .985 .904 r
F
.71806 .56195 m .7144 .57123 L .71734 .56876 L F
.7377 .54427 m .74706 .53822 L .71362 .5732 L .714 .94 .959 r
F
.71411 .57197 m .71362 .5732 L .70847 .57859 L F
.7377 .54427 m .71362 .5732 L .71411 .57197 L F
.65604 .61299 m .66866 .62433 L .66012 .65936 L p .434 .524 .851 r
F P
0 g
s
.66866 .62433 m .65647 .61338 L p .6638 .60875 L .522 .544 .819 r
F P
0 g
s
.63743 .64959 m .62098 .62201 L p .6252 .63682 L .157 .321 .779 r
F P
0 g
s
.62098 .62201 m .63743 .64959 L .65608 .61348 L p .424 .463 .797 r
F P
0 g
s
.63743 .64959 m .6252 .63682 L .62945 .65167 L .157 .321 .779 r
F
.62098 .62201 m .61561 .61303 L p .6252 .63682 L F P
0 g
s
.6387 .68408 m .61325 .64252 L .61904 .61524 L p .434 .524 .851 r
F P
0 g
s
.66877 .62676 m .69612 .5941 L p .67033 .6227 L .766 .985 .904 r
F P
0 g
s
.66866 .62433 m .66877 .62676 L p .66718 .6309 L 0 0 .267 r
F P
0 g
s
.66866 .62433 m .66025 .66165 L .65816 .66896 L 0 0 .267 r
F
.67271 .61649 m .65722 .63143 L .66877 .62676 L .766 .985 .904 r
F
.65681 .63182 m .65722 .63164 L p .65722 .63143 L F P
0 g
s
.65722 .63164 m .66877 .62676 L p .65722 .63143 L .766 .985 .904 r
F P
0 g
s
.65645 .61277 m .67709 .57279 L p .65557 .59167 L .424 .463 .797 r
F P
0 g
s
.65647 .61338 m .65604 .61299 L p .65645 .61277 L .522 .544 .819 r
F P
0 g
s
.6638 .60875 m .65647 .61338 L .65645 .61277 L .522 .544 .819 r
F
.65635 .61106 m .67801 .59166 L .65722 .63143 L p .835 .999 .804 r
F P
0 g
s
.65722 .63143 m .67801 .59166 L p .67145 .61771 L .763 .892 .935 r
F P
0 g
s
.62723 .64433 m .65681 .63182 L p .67854 .6013 L .766 .985 .904 r
F P
0 g
s
.73517 .53065 m .76108 .52124 L .73863 .54066 L .554 .503 .754 r
F
.73863 .54066 m .7388 .54314 L p .73506 .54537 L .176 .06 .473 r
F P
0 g
s
.7062 .58125 m .67801 .59166 L .72027 .55555 L p .714 .94 .959 r
F P
0 g
s
.7388 .54314 m .72069 .55527 L p .71806 .56195 L .766 .985 .904 r
F P
0 g
s
.7377 .54427 m .7062 .58125 L .72027 .55555 L .714 .94 .959 r
F
.65006 .61652 m .62723 .64433 L p .70377 .56988 L .766 .985 .904 r
F P
0 g
s
.71806 .56195 m .70377 .56988 L .62723 .64433 L .766 .985 .904 r
F
.71806 .56195 m .65663 .61767 L .7144 .57123 L F
.65663 .61767 m .62723 .64433 L .65671 .61961 L F
.65663 .61767 m .65671 .61961 L .7144 .57123 L F
.63743 .64959 m .63916 .68566 L p .6574 .67159 L .455 .585 .894 r
F P
0 g
s
.65764 .67075 m .63743 .64959 L p .6574 .67159 L .455 .585 .894 r
F P
0 g
s
.6161 .6535 m .61524 .65164 L .60481 .62628 L 0 0 .414 r
F
.60481 .62628 m .61524 .65164 L .61292 .6466 L F
.61351 .64787 m .61524 .65164 L .5946 .60823 L .68 .309 0 r
F
.60481 .62628 m .61292 .6466 L .60951 .63859 L 0 0 .414 r
F
.61161 .64375 m .60951 .63859 L p .61292 .6466 L F P
0 g
s
.60481 .62628 m .60951 .63859 L .59832 .61048 L 0 0 .414 r
F
.59112 .60122 m .61548 .65216 L p .59934 .6098 L .157 .321 .779 r
F P
0 g
s
.61325 .64252 m .61873 .65924 L p .6091 .63542 L 0 0 .414 r
F P
0 g
s
.61873 .65924 m .61325 .64252 L p .63394 .69041 L .026 .284 .78 r
F P
0 g
s
.55938 .69548 m .60432 .70872 L p .57426 .70143 L .294 .7 .974 r
F P
0 g
s
.56247 .69354 m .49976 .68335 L .49778 .68239 L .768 .516 .554 r
F
.49976 .68335 m .47913 .68 L .49411 .68192 L F
.49976 .68335 m .49411 .68192 L .49778 .68239 L F
.51064 .68372 m .56247 .69354 L p .49536 .68197 L F P
0 g
s
.49803 .65895 m .49654 .66756 L p .54471 .68798 L .742 .705 .816 r
F P
0 g
s
.55915 .68979 m .49803 .65895 L .54471 .68798 L .742 .705 .816 r
F
.55908 .6771 m .56247 .69354 L .51064 .68372 L p .743 .664 .773 r
F P
0 g
s
.5625 .69592 m .5618 .69581 L p .55641 .69479 L .15 0 0 r
F P
0 g
s
.55926 .69543 m .55525 .69466 L .55641 .69479 L .15 0 0 r
F
.51571 .68707 m .5625 .69592 L p .55641 .69479 L F P
0 g
s
.57426 .70143 m .51274 .68651 L .55938 .69548 L .294 .7 .974 r
F
.55641 .69479 m .5419 .6921 L .51571 .68707 L .15 0 0 r
F
.5419 .6921 m .51428 .68681 L .51571 .68707 L F
.51274 .68651 m .50105 .68368 L .55704 .695 L .294 .7 .974 r
F
.51274 .68651 m .55704 .695 L .55938 .69548 L F
.5419 .6921 m .44802 .6747 L .51428 .68681 L .15 0 0 r
F
.58032 .70385 m .56375 .69969 L .56373 .69968 L .294 .7 .974 r
F
.58033 .70386 m .56375 .69969 L .58032 .70385 L F
.55938 .69548 m .55914 .69541 L p .59374 .71177 L .615 .062 0 r
F P
0 g
s
.5618 .69581 m .55929 .69543 L p .55641 .69479 L .15 0 0 r
F P
0 g
s
.55929 .69543 m .55924 .69542 L .55641 .69479 L .15 0 0 r
F
.55914 .69541 m .55938 .69548 L p .55704 .695 L .294 .7 .974 r
F P
0 g
s
.59374 .71177 m .55914 .69541 L .53658 .69125 L .615 .062 0 r
F
.55926 .69543 m .55107 .69419 L p .55525 .69466 L .15 0 0 r
F P
0 g
s
.49786 .67735 m .55914 .69541 L p .50105 .68368 L .294 .7 .974 r
F P
0 g
s
.55704 .695 m .50105 .68368 L .55914 .69541 L .294 .7 .974 r
F
.5419 .6921 m .54484 .69277 L .44802 .6747 L .15 0 0 r
F
.49496 .68358 m .44802 .6747 L p .54484 .69277 L F P
0 g
s
.55525 .69466 m .55107 .69419 L .5419 .6921 L .15 0 0 r
F
.55914 .69541 m .49786 .67735 L p .53658 .69125 L .615 .062 0 r
F P
0 g
s
.55929 .69543 m .55926 .69543 L p .55924 .69542 L .15 0 0 r
F P
0 g
s
.55107 .69419 m .55926 .69543 L p .62279 .72128 L .409 0 0 r
F P
0 g
s
.62227 .72119 m .55107 .69419 L .62279 .72128 L .409 0 0 r
F
.56102 .71019 m .59374 .71177 L .53658 .69125 L .615 .062 0 r
F
.62645 .72277 m .62227 .72119 L .62279 .72128 L .409 0 0 r
F
.62645 .72277 m .55926 .69543 L .59411 .70937 L F
.6409 .72876 m .59411 .70937 L .63615 .72675 L F
.59411 .70937 m .59555 .70995 L .63615 .72675 L F
.6069 .71053 m .58033 .70386 L .58032 .70385 L .294 .7 .974 r
F
.60645 .71432 m .62485 .72202 L p .59555 .70995 L .409 0 0 r
F P
0 g
s
.62577 .72245 m .59555 .70995 L .62485 .72202 L .409 0 0 r
F
.62485 .72202 m .63615 .72675 L p .62577 .72245 L F P
0 g
s
.64711 .73133 m .63316 .72594 L .62645 .72277 L .409 0 0 r
F
.62478 .71966 m .64702 .72898 L p .61853 .71784 L .843 .579 .54 r
F P
0 g
s
.59916 .72639 m .62797 .73022 L .62773 .7246 L .827 .365 .115 r
F
.64067 .72948 m .64621 .73264 L p .63048 .72467 L 0 .394 .79 r
F P
0 g
s
.6444 .73028 m .63496 .72733 L .55107 .69419 L .409 0 0 r
F
.64711 .73133 m .64107 .72934 L p .6444 .73028 L F P
0 g
s
.64107 .72934 m .63496 .72733 L p .6444 .73028 L .409 0 0 r
F P
0 g
s
.59916 .72639 m .62773 .7246 L .60246 .71506 L .827 .365 .115 r
F
.62797 .73022 m .63437 .7291 L .62773 .7246 L F
.63205 .72839 m .6247 .72761 L .61329 .72018 L .819 .337 .015 r
F
.62393 .72747 m .61329 .72018 L p .6247 .72761 L F P
0 g
s
.58766 .70947 m .55107 .69419 L p .63496 .72733 L .409 0 0 r
F P
0 g
s
.60246 .71506 m .62234 .72256 L .60258 .71467 L .827 .365 .115 r
F
.62234 .72256 m .62773 .7246 L .60258 .71467 L F
.62002 .71937 m .61972 .71922 L .60432 .70872 L 0 .394 .79 r
F
.60432 .70872 m .64067 .72948 L p .62002 .71937 L F P
0 g
s
.63048 .72467 m .61972 .71922 L .64067 .72948 L 0 .394 .79 r
F
.6177 .71812 m .60432 .70872 L p .61972 .71922 L F P
0 g
s
.59555 .70995 m .60558 .71396 L .60645 .71432 L .409 0 0 r
F
.62249 .7209 m .6181 .7184 L p .63048 .72467 L 0 .394 .79 r
F P
0 g
s
.63316 .72594 m .55107 .69419 L .62645 .72277 L .409 0 0 r
F
.60258 .71467 m .62773 .7246 L .60432 .70872 L .827 .365 .115 r
F
.58951 .71299 m .59374 .71177 L p .56102 .71019 L .615 .062 0 r
F P
0 g
s
.63363 .72546 m .60432 .70872 L p .63798 .73155 L .827 .365 .115 r
F P
0 g
s
.59134 .71645 m .55251 .72996 L .54553 .70867 L .803 .361 .196 r
F
.60432 .70872 m .58951 .71299 L p .59803 .72913 L .827 .365 .115 r
F P
0 g
s
.61005 .73249 m .60742 .73342 L .59803 .72913 L .906 .541 .331 r
F
.59833 .72922 m .60432 .70872 L .59803 .72913 L .827 .365 .115 r
F
.59808 .71383 m .61329 .72018 L p .54553 .70867 L .803 .361 .196 r
F P
0 g
s
.55606 .70971 m .53658 .69125 L .51606 .68747 L .615 .062 0 r
F
.57137 .71822 m .56102 .71019 L .55606 .70971 L F
.56102 .71019 m .53658 .69125 L .55606 .70971 L F
.47717 .69056 m .50081 .68466 L .41438 .67179 L .524 0 0 r
F
.51533 .69369 m .49446 .71716 L .46117 .69293 L .683 .221 .149 r
F
.44802 .6747 m .42844 .67069 L p .51122 .68622 L .15 0 0 r
F P
0 g
s
.51122 .68622 m .42844 .67069 L .47089 .67861 L .15 0 0 r
F
.51122 .68622 m .47089 .67861 L .51062 .68611 L F
.47089 .67861 m .47131 .67868 L .51062 .68611 L F
.47665 .67968 m .51062 .68611 L p .47131 .67868 L F P
0 g
s
.51428 .68681 m .44802 .6747 L .51122 .68622 L .15 0 0 r
F
.49411 .68192 m .49786 .67735 L p .50105 .68368 L .294 .7 .974 r
F P
0 g
s
.49786 .67735 m .50173 .68694 L p .57137 .71822 L .615 .062 0 r
F P
0 g
s
.55538 .71356 m .57137 .71822 L p .50173 .68694 L .615 .062 0 r
F P
0 g
s
.5079 .68602 m .55107 .69419 L p .51533 .69369 L .683 .221 .149 r
F P
0 g
s
.51533 .69369 m .46117 .69293 L .50173 .68694 L .683 .221 .149 r
F
.51533 .69369 m .50173 .68694 L .5079 .68602 L F
.50677 .69941 m .41438 .67179 L .40066 .67293 L .524 0 0 r
F
.38972 .6853 m .38882 .67392 L .3155 .67224 L .54 .127 .206 r
F
.31693 .67277 m .3155 .67224 L .31356 .67219 L F
.38972 .6853 m .3155 .67224 L .31693 .67277 L F
.46493 .75198 m .45014 .71973 L .40863 .70558 L .617 .3 .409 r
F
.46493 .75198 m .40863 .70558 L .41403 .71159 L F
.38901 .68626 m .31356 .67219 L .31298 .67521 L .54 .127 .206 r
F
.38972 .6853 m .31356 .67219 L .38901 .68626 L F
.4463 .68932 m .44301 .71724 L .41871 .68262 L .615 .212 .255 r
F
.44301 .71724 m .43555 .71464 L .41871 .68262 L F
.41871 .68262 m .43555 .71464 L .39798 .69764 L F
.38892 .69066 m .39798 .69764 L .38495 .69174 L F
.41871 .68262 m .39798 .69764 L .38892 .69066 L F
.28686 .6704 m .25273 .67341 L p .29061 .66262 L .576 .571 .81 r
F P
0 g
s
.31289 .6757 m .31245 .67802 L .31036 .67758 L .48 .074 .193 r
F
.31289 .6757 m .30897 .67861 L .28209 .67525 L F
.24807 .67578 m .25384 .67484 L p .3123 .67884 L F P
0 g
s
.38648 .68887 m .31285 .67594 L .32614 .68164 L .54 .127 .206 r
F
.31245 .67802 m .32614 .68164 L p .31285 .67594 L F P
0 g
s
.32614 .68164 m .34023 .68536 L p .34983 .68609 L .54 .127 .206 r
F P
0 g
s
.38648 .68887 m .32614 .68164 L .34983 .68609 L .54 .127 .206 r
F
.37948 .68487 m .31298 .67521 L .38648 .68887 L F
.29401 .66825 m .31395 .67013 L p .31289 .6757 L .48 .074 .193 r
F P
0 g
s
.31298 .67521 m .28822 .67161 L .31285 .67594 L .54 .127 .206 r
F
.31298 .67521 m .31285 .67594 L .38648 .68887 L F
.35622 .6741 m .32264 .67955 L p .31289 .6757 L .48 .074 .193 r
F P
0 g
s
.28822 .67161 m .31406 .66952 L p .29077 .67157 L .115 .668 .763 r
F P
0 g
s
.30937 .6699 m .28822 .67161 L p .38882 .67392 L .54 .127 .206 r
F P
0 g
s
.31395 .67013 m .35622 .6741 L p .31289 .6757 L .48 .074 .193 r
F P
0 g
s
.40786 .67007 m .41926 .67139 L .37581 .67326 L .246 0 0 r
F
.40786 .67007 m .37581 .67326 L .37049 .67299 L F
.38439 .67429 m .35622 .6741 L p .41926 .67139 L F P
0 g
s
.40786 .67007 m .35622 .6741 L .38287 .66933 L .246 0 0 r
F
.50677 .69941 m .40066 .67293 L .39583 .67706 L .524 0 0 r
F
.38287 .66933 m .31065 .66982 L .36128 .66869 L .246 0 0 r
F
.35622 .6741 m .31065 .66982 L p .38287 .66933 L F P
0 g
s
.36907 .66505 m .38882 .67392 L p .40066 .67293 L .524 0 0 r
F P
0 g
s
.38882 .67392 m .39583 .67706 L p .40066 .67293 L .524 0 0 r
F P
0 g
s
.47591 .69651 m .50677 .69941 L p .39583 .67706 L .524 0 0 r
F P
0 g
s
.39783 .67437 m .44802 .6747 L p .41871 .68262 L .615 .212 .255 r
F P
0 g
s
.41871 .68262 m .38495 .69174 L .39583 .67706 L .615 .212 .255 r
F
.41871 .68262 m .39583 .67706 L .39783 .67437 L F
.46854 .70315 m .46117 .69293 L p .49446 .71716 L .683 .221 .149 r
F P
0 g
s
.37702 .69061 m .38778 .69914 L .38204 .69742 L .576 .273 .418 r
F
.37318 .69006 m .38204 .69742 L .34349 .68582 L F
.37702 .69061 m .38204 .69742 L .37318 .69006 L F
.47262 .74578 m .49575 .80179 L .48034 .78154 L .668 .348 .423 r
F
.46653 .75453 m .48034 .78154 L .4633 .75916 L F
.47262 .74578 m .48034 .78154 L .46653 .75453 L F
.48636 .79412 m .4633 .75916 L p .49575 .80179 L F P
0 g
s
.5029 .8302 m .49917 .8046 L .48636 .79412 L .64 .363 .484 r
F
.51152 .84406 m .49917 .8046 L .50287 .83004 L F
.51152 .84406 m .50287 .83004 L .50297 .83032 L F
.51158 .8438 m .49917 .8046 L p .51152 .84406 L F P
0 g
s
.49146 .75768 m .49995 .80312 L .47262 .74578 L .668 .348 .423 r
F
.45014 .71973 m .40473 .70388 L .40482 .70428 L .617 .3 .409 r
F
.44802 .6747 m .46854 .70315 L p .44189 .7267 L .615 .212 .255 r
F P
0 g
s
.49875 .75734 m .50984 .72548 L .46458 .70672 L .687 .306 .32 r
F
.49875 .75734 m .46458 .70672 L .45453 .71577 L F
.46713 .74231 m .45014 .71973 L p .46493 .75198 L .617 .3 .409 r
F P
0 g
s
.4424 .7267 m .44189 .7267 L p .46854 .70315 L .615 .212 .255 r
F P
0 g
s
.45212 .69426 m .42938 .69212 L p .50984 .72548 L .687 .306 .32 r
F P
0 g
s
.5338 .70285 m .48573 .72697 L .51533 .69369 L .683 .221 .149 r
F
.5347 .76467 m .5599 .73939 L .53326 .71935 L .778 .379 .296 r
F
.54286 .76083 m .5599 .73939 L .5347 .76467 L F
.55014 .75741 m .53005 .80532 L .53088 .76647 L .842 .49 .388 r
F
.52776 .76306 m .51629 .80554 L .51429 .74834 L .75 .396 .384 r
F
.52269 .84896 m .5321 .81012 L .52645 .79753 L .803 .476 .436 r
F
.5249 .84612 m .5321 .81012 L p .52269 .84896 L F P
0 g
s
.52299 .86174 m .52188 .85001 L .5249 .84612 L .849 .544 .472 r
F
.52236 .86688 m .52229 .8643 L .52299 .86174 L F
.52229 .8643 m .52188 .85001 L .52299 .86174 L F
.52917 .82763 m .53208 .82526 L .53253 .80975 L .889 .573 .442 r
F
.52816 .83414 m .53208 .82526 L .52917 .82763 L F
.52531 .85329 m .5249 .84612 L .5257 .85089 L .849 .544 .472 r
F
.5247 .85551 m .5249 .84612 L .52531 .85329 L F
.52236 .86688 m .5249 .84612 L .52466 .85728 L F
.52647 .84499 m .52976 .83584 L .52801 .83508 L .889 .573 .442 r
F
.52976 .83584 m .53812 .81254 L .53192 .82601 L F
.52976 .83584 m .53192 .82601 L .52778 .83499 L F
.53812 .81254 m .5399 .80757 L .52778 .83499 L F
.53208 .82526 m .5399 .80757 L .5357 .80881 L F
.53208 .82526 m .5357 .80881 L .53253 .80975 L F
.5399 .80757 m .53948 .8037 L .53253 .80975 L F
.53971 .80576 m .53261 .81934 L .53948 .8037 L .914 .593 .422 r
F
.53895 .80809 m .53261 .81934 L .53971 .80576 L F
.53312 .81733 m .53948 .8037 L p .53261 .81934 L F P
0 g
s
.52527 .84833 m .52647 .84499 L .52778 .83499 L .889 .573 .442 r
F
.52647 .84499 m .52801 .83508 L .52778 .83499 L F
.55107 .69419 m .54235 .71697 L p .48573 .72697 L .683 .221 .149 r
F P
0 g
s
.49655 .72896 m .48573 .72697 L p .54235 .71697 L .683 .221 .149 r
F P
0 g
s
.53504 .70764 m .50677 .69941 L p .53326 .71935 L .778 .379 .296 r
F P
0 g
s
.5599 .73939 m .54235 .71697 L .53326 .71935 L .778 .379 .296 r
F
.54235 .71697 m .53504 .70764 L .53326 .71935 L F
.53489 .77086 m .53326 .71935 L .51636 .72378 L F
.51203 .74635 m .50984 .72548 L .49875 .75734 L .687 .306 .32 r
F
.51429 .74834 m .51629 .80554 L .49784 .75809 L .75 .396 .384 r
F
.50677 .69941 m .51228 .74615 L p .53489 .77086 L .778 .379 .296 r
F P
0 g
s
.52642 .73445 m .53505 .73603 L p .51429 .74834 L .75 .396 .384 r
F P
0 g
s
.51429 .74834 m .49784 .75809 L .51228 .74615 L .75 .396 .384 r
F
.51429 .74834 m .51228 .74615 L .52642 .73445 L F
.53505 .73603 m .52664 .79735 L p .51701 .80631 L F P
0 g
s
.51701 .80631 m .51527 .80444 L .51502 .84697 L .719 .407 .45 r
F
.51505 .84722 m .51502 .84697 L .51501 .84798 L F
.51701 .80631 m .51502 .84697 L .51505 .84722 L F
.5175 .86445 m .51158 .8438 L .5158 .84895 L .685 .407 .496 r
F
.51839 .87255 m .51453 .85882 L .5175 .86445 L F
.516 .86627 m .51453 .85882 L .51839 .87255 L F
.51453 .85882 m .51158 .8438 L .5175 .86445 L F
.47859 .78612 m .48656 .79903 L .48256 .79444 L .599 .332 .483 r
F
.49995 .80312 m .4979 .807 L .47262 .74578 L .668 .348 .423 r
F
.50505 .79349 m .49536 .76014 L p .51629 .80554 L .75 .396 .384 r
F P
0 g
s
.49875 .75734 m .42938 .69212 L .49536 .76014 L .687 .306 .32 r
F
.49536 .76014 m .48573 .72697 L p .49784 .75809 L .75 .396 .384 r
F P
0 g
s
.51629 .80554 m .49536 .76014 L .49784 .75809 L .75 .396 .384 r
F
.42938 .69212 m .46713 .74231 L p .49536 .76014 L .687 .306 .32 r
F P
0 g
s
.51476 .77239 m .49875 .75734 L .49536 .76014 L .687 .306 .32 r
F
.51333 .84308 m .51453 .80365 L .50005 .80354 L .719 .407 .45 r
F
.50103 .80731 m .50005 .80354 L .49973 .80353 L F
.51333 .84308 m .50005 .80354 L .50103 .80731 L F
.51453 .80365 m .50505 .79349 L .49973 .80353 L F
.52076 .84794 m .52645 .79753 L .51776 .80561 L .803 .476 .436 r
F
.51993 .88586 m .51986 .88318 L .51998 .88122 L .766 .467 .48 r
F
.51993 .88586 m .51998 .88122 L .52004 .88382 L F
.51986 .88318 m .51585 .849 L .5188 .84433 L F
.51993 .88586 m .51985 .88513 L p .51986 .88318 L F P
0 g
s
.51985 .88513 m .51585 .849 L p .51986 .88318 L .766 .467 .48 r
F P
0 g
s
.52019 .88488 m .52236 .86688 L .52466 .85728 L .849 .544 .472 r
F
.52003 .88587 m .52016 .88477 L .52019 .88488 L F
.52016 .88477 m .52236 .86688 L .52019 .88488 L F
.51967 .88421 m .516 .86627 L .51839 .87255 L .685 .407 .496 r
F
.51985 .88586 m .51981 .88565 L .51967 .88421 L F
.51981 .88565 m .516 .86627 L .51967 .88421 L F
.54241 .80195 m .5398 .80659 L .54775 .78846 L .914 .593 .422 r
F
.53129 .82454 m .53076 .82421 L p .53895 .80809 L F P
0 g
s
.53076 .82421 m .53009 .8238 L p .53132 .82162 L .914 .593 .422 r
F P
0 g
s
.53895 .80809 m .53076 .82421 L .53132 .82162 L .914 .593 .422 r
F
.53865 .80901 m .53129 .82454 L .53895 .80809 L F
.52572 .85097 m .53997 .80814 L .52527 .84833 L .889 .573 .442 r
F
.51985 .88586 m .51967 .88421 L .51993 .88582 L .685 .407 .496 r
F
.51967 .88421 m .5158 .84895 L .51915 .87882 L F
.51967 .88421 m .51915 .87882 L .51993 .88582 L F
.52003 .88587 m .52064 .87402 L p .52236 .86688 L .849 .544 .472 r
F P
0 g
s
.52064 .87402 m .52188 .85001 L p .52236 .86688 L .849 .544 .472 r
F P
0 g
s
.52253 .8512 m .52645 .79753 L .52076 .84794 L .803 .476 .436 r
F
.52003 .88587 m .51993 .88586 L p .52184 .84993 L .766 .467 .48 r
F P
0 g
s
.52337 .82124 m .52188 .85001 L p .52184 .84993 L .766 .467 .48 r
F P
0 g
s
.52188 .85001 m .52003 .88587 L p .52184 .84993 L .766 .467 .48 r
F P
0 g
s
.52004 .88382 m .51998 .88122 L .52184 .84993 L .766 .467 .48 r
F
.51986 .88318 m .5188 .84433 L .52 .84653 L F
.51986 .88318 m .52 .84653 L .52184 .84993 L F
.5196 .86133 m .52175 .86189 L p .52076 .84794 L .803 .476 .436 r
F P
0 g
s
.53505 .73603 m .51701 .80631 L .51629 .80554 L .75 .396 .384 r
F
.51545 .77304 m .5188 .84433 L p .52076 .84794 L .803 .476 .436 r
F P
0 g
s
.5188 .84433 m .5196 .86133 L p .52076 .84794 L .803 .476 .436 r
F P
0 g
s
.52024 .88506 m .52477 .85685 L .52667 .84986 L .915 .619 .467 r
F
.52023 .88503 m .52477 .85685 L p .52024 .88506 L F P
0 g
s
.51985 .88577 m .51992 .88548 L .51985 .88572 L .62 .394 .553 r
F
.51985 .88586 m .51986 .8858 L .51985 .88577 L F
.52527 .84833 m .53253 .80975 L .5249 .84612 L .889 .573 .442 r
F
.5266 .84994 m .53997 .80814 L p .52572 .85097 L F P
0 g
s
.53997 .80814 m .5399 .80757 L .52527 .84833 L .889 .573 .442 r
F
.53728 .81323 m .53129 .82454 L .53865 .80901 L .914 .593 .422 r
F
.5394 .80957 m .53706 .81389 L p .53997 .80814 L F P
0 g
s
.53706 .81389 m .53129 .82454 L p .53728 .81323 L .914 .593 .422 r
F P
0 g
s
.53706 .81389 m .53728 .81323 L .53997 .80814 L .914 .593 .422 r
F
.52572 .85097 m .53009 .8238 L p .52699 .84742 L .915 .619 .467 r
F P
0 g
s
.53946 .80951 m .53947 .80951 L .5266 .84994 L .997 .768 .444 r
F
.53997 .80814 m .5266 .84994 L p .53964 .80897 L F P
0 g
s
.52679 .84925 m .52678 .84939 L .5266 .84994 L .997 .768 .444 r
F
.53009 .8238 m .53129 .82454 L p .52677 .84931 L .915 .619 .467 r
F P
0 g
s
.52667 .84986 m .52671 .84952 L .52677 .84931 L .915 .619 .467 r
F
.52671 .84952 m .53009 .8238 L .52677 .84931 L F
.52738 .847 m .53481 .82117 L .53964 .80897 L .997 .768 .444 r
F
.52692 .84722 m .53129 .82454 L .52696 .84663 L .986 .776 .379 r
F
.5201 .8859 m .52009 .8859 L p .52677 .84931 L .915 .619 .467 r
F P
0 g
s
.53129 .82454 m .52679 .84925 L p .52677 .84931 L .915 .619 .467 r
F P
0 g
s
.52679 .84925 m .5201 .8859 L p .52677 .84931 L .915 .619 .467 r
F P
0 g
s
.52679 .84925 m .52734 .84618 L p .52692 .84722 L .986 .776 .379 r
F P
0 g
s
.5261 .85052 m .52699 .84742 L .52667 .84986 L .915 .619 .467 r
F
.52779 .84428 m .53081 .83407 L p .52641 .85 L .997 .768 .444 r
F P
0 g
s
.53129 .82454 m .53706 .81389 L p .5373 .81139 L .998 .825 .484 r
F P
0 g
s
.53997 .80814 m .53728 .81323 L .5398 .80659 L .914 .593 .422 r
F
.53728 .81323 m .53865 .80901 L .5398 .80659 L F
.53706 .81389 m .5394 .80957 L p .5376 .81112 L .998 .825 .484 r
F P
0 g
s
.5373 .81139 m .53706 .81389 L .5376 .81112 L .998 .825 .484 r
F
.52734 .84618 m .53129 .82454 L p .52692 .84722 L .986 .776 .379 r
F P
0 g
s
.55266 .77021 m .57012 .75668 L .57072 .7382 L .882 .52 .354 r
F
.59524 .72384 m .57486 .72797 L .59134 .71645 L .803 .361 .196 r
F
.57486 .72797 m .53505 .73603 L .57873 .72084 L F
.57486 .72797 m .57873 .72084 L .59134 .71645 L F
.5398 .80659 m .53971 .80576 L .54817 .7868 L .914 .593 .422 r
F
.54859 .78624 m .54817 .7868 L .554 .77375 L F
.5398 .80659 m .54817 .7868 L .54859 .78624 L F
.54889 .78586 m .554 .77375 L .55422 .77369 L F
.5398 .80659 m .54859 .78624 L .54889 .78586 L F
.54859 .78624 m .554 .77375 L .54889 .78586 L F
.53123 .83178 m .53474 .82012 L .53408 .82301 L .935 .819 .365 r
F
.55006 .77487 m .54657 .78516 L .54142 .79204 L .842 .49 .388 r
F
.55006 .77487 m .53467 .80548 L .53591 .79944 L F
.55339 .77392 m .56647 .76209 L .55266 .77021 L .882 .52 .354 r
F
.56647 .76209 m .57012 .75668 L .55266 .77021 L F
.53971 .80576 m .55339 .77392 L .554 .77375 L .914 .593 .422 r
F
.55912 .77232 m .54 .80846 L .5398 .80659 L F
.55917 .7723 m .56049 .77094 L p .55339 .77392 L .882 .52 .354 r
F P
0 g
s
.58479 .74573 m .58635 .74411 L p .58394 .74629 L .882 .52 .354 r
F P
0 g
s
.56647 .76209 m .58635 .74411 L .57012 .75668 L .882 .52 .354 r
F
.57967 .74255 m .58278 .73795 L .58254 .73754 L F
.59524 .72384 m .58531 .7366 L .53505 .73603 L .803 .361 .196 r
F
.57167 .7544 m .58635 .74411 L .58278 .73795 L .882 .52 .354 r
F
.57012 .75668 m .5812 .7481 L .57167 .7544 L F
.5812 .7481 m .58635 .74411 L .57167 .7544 L F
.59833 .72922 m .63675 .73267 L .59916 .72639 L .827 .365 .115 r
F
.59824 .72953 m .59995 .72967 L .59833 .72922 L F
.59995 .72967 m .63675 .73267 L .59833 .72922 L F
.54688 .74082 m .53505 .73603 L p .58531 .7366 L .803 .361 .196 r
F P
0 g
s
.58372 .72514 m .57137 .71822 L p .58635 .74411 L .882 .52 .354 r
F P
0 g
s
.58219 .73843 m .59803 .72913 L p .60346 .73161 L .906 .541 .331 r
F P
0 g
s
.60742 .73342 m .59859 .73517 L .60346 .73161 L .906 .541 .331 r
F
.59859 .73517 m .58219 .73843 L .60346 .73161 L F
.57012 .75668 m .57967 .74255 L .57137 .71822 L .882 .52 .354 r
F
.57068 .75181 m .56535 .7483 L p .57708 .74415 L .906 .541 .331 r
F P
0 g
s
.58418 .74164 m .57068 .75181 L .57708 .74415 L .906 .541 .331 r
F
.57638 .74902 m .57068 .75181 L .58418 .74164 L F
.56049 .77094 m .56647 .76209 L .55339 .77392 L .882 .52 .354 r
F
.56535 .7483 m .57068 .75181 L p .54241 .80195 L .914 .593 .422 r
F P
0 g
s
.56043 .7641 m .56042 .76412 L .57068 .75181 L .998 .825 .484 r
F
.55339 .77392 m .56535 .7483 L p .554 .77375 L .914 .593 .422 r
F P
0 g
s
.55625 .76719 m .5461 .78954 L .5442 .79218 L .842 .49 .388 r
F
.56535 .7483 m .55408 .77244 L p .55625 .76719 L F P
0 g
s
.55408 .77244 m .5461 .78954 L p .55625 .76719 L .842 .49 .388 r
F P
0 g
s
.56535 .7483 m .55625 .76719 L .56073 .75633 L .842 .49 .388 r
F
.57068 .75181 m .57008 .75021 L p .56043 .7641 L .998 .825 .484 r
F P
0 g
s
.57967 .74255 m .58254 .73754 L .5822 .73694 L .882 .52 .354 r
F
.57967 .74255 m .5822 .73694 L .57792 .73741 L F
.58154 .73881 m .56535 .7483 L p .58357 .73679 L .803 .361 .196 r
F P
0 g
s
.56535 .7483 m .58219 .73843 L p .58547 .73778 L .906 .541 .331 r
F P
0 g
s
.58592 .74102 m .56535 .7483 L .58547 .73778 L .906 .541 .331 r
F
.57008 .75021 m .56861 .74635 L p .56408 .75618 L .998 .825 .484 r
F P
0 g
s
.56408 .75618 m .56043 .7641 L .57008 .75021 L .998 .825 .484 r
F
.55266 .77021 m .57072 .7382 L .56391 .73895 L .882 .52 .354 r
F
.5399 .77632 m .5599 .73939 L p .53489 .77086 L .778 .379 .296 r
F P
0 g
s
.55268 .77032 m .54881 .78217 L .55006 .77487 L .842 .49 .388 r
F
.54881 .78217 m .5442 .79218 L .54657 .78516 L F
.54881 .78217 m .54657 .78516 L .55006 .77487 L F
.54142 .79204 m .5442 .79218 L .53467 .80548 L F
.5321 .81012 m .5249 .84612 L p .53253 .80975 L .889 .573 .442 r
F P
0 g
s
.52778 .83499 m .52816 .83414 L .52917 .82763 L .889 .573 .442 r
F
.52337 .82124 m .53009 .8238 L p .52689 .84356 L .849 .544 .472 r
F P
0 g
s
.52009 .8859 m .52003 .88587 L p .52009 .88552 L .849 .544 .472 r
F P
0 g
s
.53009 .8238 m .5201 .88585 L p .52009 .88552 L .849 .544 .472 r
F P
0 g
s
.5201 .88585 m .52009 .8859 L p .52009 .88552 L .849 .544 .472 r
F P
0 g
s
.52413 .84998 m .52175 .86189 L .52253 .86228 L p .889 .573 .442 r
F P
0 g
s
.52 .88465 m .52011 .88363 L .52 .88486 L 0 0 .181 r
F
.51993 .8854 m .52 .88465 L .52 .88486 L F
.51998 .88514 m .51999 .88507 L .52 .88492 L .883 .733 .222 r
F
.52 .88486 m .52009 .8859 L p .52001 .88482 L .914 .646 .517 r
F P
0 g
s
.52253 .86228 m .52195 .86199 L p .52 .88486 L .891 .597 .483 r
F P
0 g
s
.52 .88487 m .52 .88486 L .52 .88487 L .883 .733 .222 r
F
.52061 .87937 m .52 .88486 L .52 .88487 L F
.51999 .88499 m .52 .88486 L .51995 .88525 L .698 .467 .569 r
F
.51999 .88507 m .51999 .88501 L .52 .88488 L .883 .733 .222 r
F
.51986 .88565 m .51997 .88483 L .51989 .88549 L .212 .036 .411 r
F
.52002 .88481 m .52048 .88056 L .52 .88485 L .883 .733 .222 r
F
.52002 .88481 m .52 .88485 L .52 .88487 L F
.52048 .88056 m .52017 .88336 L .52 .88487 L F
.52 .88486 m .52264 .86233 L .52253 .86228 L p .891 .597 .483 r
F P
0 g
s
.52011 .88363 m .52 .88486 L .52011 .88357 L 0 0 .181 r
F
.51997 .88497 m .52 .88486 L .52011 .88363 L F
.51997 .88483 m .52002 .88441 L .51997 .88484 L .212 .036 .411 r
F
.52 .88486 m .52175 .86189 L p .52002 .88437 L .809 .517 .495 r
F P
0 g
s
.52002 .88441 m .52002 .88437 L .52002 .88442 L .212 .036 .411 r
F
.52002 .88441 m .52002 .88442 L .51997 .88484 L F
.51997 .88483 m .51997 .88484 L .51989 .88549 L F
.51993 .88539 m .52 .88486 L .51997 .88497 L 0 0 .181 r
F
.52 .88486 m .51986 .8858 L p .51995 .88525 L .698 .467 .569 r
F P
0 g
s
.51985 .88577 m .52 .88485 L .51999 .88467 L .212 .036 .411 r
F
.51453 .85882 m .51043 .84873 L .51158 .8438 L .685 .407 .496 r
F
.51985 .88586 m .51157 .8532 L p .51453 .85882 L F P
0 g
s
.51157 .8532 m .51043 .84873 L p .51453 .85882 L .685 .407 .496 r
F P
0 g
s
.51251 .85118 m .51181 .84451 L .5029 .8302 L .64 .363 .484 r
F
.51706 .86111 m .51476 .85384 L p .51251 .85118 L F P
0 g
s
.51476 .85384 m .51181 .84451 L p .51251 .85118 L .64 .363 .484 r
F P
0 g
s
.49317 .81458 m .48636 .79412 L .48667 .80193 L .64 .363 .484 r
F
.50142 .82784 m .48636 .79412 L .49317 .81458 L F
.51251 .85118 m .48636 .79412 L .51111 .84953 L F
.46402 .76591 m .47859 .78612 L .47052 .77744 L .599 .332 .483 r
F
.45889 .75683 m .46628 .76781 L .46402 .76591 L F
.46628 .76781 m .47859 .78612 L .46402 .76591 L F
.47262 .74578 m .4633 .75916 L .46713 .74231 L .668 .348 .423 r
F
.48841 .77061 m .40482 .70428 L .40707 .71436 L .617 .3 .409 r
F
.40257 .69425 m .42938 .69212 L p .35461 .68645 L .54 .127 .206 r
F P
0 g
s
.35461 .68645 m .35823 .68793 L .40257 .69425 L .54 .127 .206 r
F
.37702 .69061 m .40473 .70388 L .4071 .71447 L .576 .273 .418 r
F
.38495 .69174 m .40473 .70388 L p .37702 .69061 L F P
0 g
s
.48841 .77061 m .40707 .71436 L .42198 .72975 L .617 .3 .409 r
F
.47259 .76406 m .42286 .73059 L .44359 .74876 L F
.41782 .7272 m .43028 .73762 L p .42286 .73059 L F P
0 g
s
.43028 .73762 m .44359 .74876 L p .42286 .73059 L .617 .3 .409 r
F P
0 g
s
.38778 .69914 m .4071 .71447 L .39808 .71074 L .576 .273 .418 r
F
.38778 .69914 m .39808 .71074 L .38658 .70119 L F
.38778 .69914 m .38432 .70505 L .36546 .69267 L F
.35461 .68645 m .34023 .68536 L .35823 .68793 L .54 .127 .206 r
F
.38778 .69914 m .32835 .68191 L .32264 .67955 L .576 .273 .418 r
F
.34023 .68536 m .32614 .68164 L p .32322 .68058 L .228 .566 .938 r
F P
0 g
s
.34023 .68536 m .32322 .68058 L .32631 .68132 L .228 .566 .938 r
F
.40855 .72096 m .32631 .68132 L p .4071 .71447 L .576 .273 .418 r
F P
0 g
s
.32631 .68132 m .32264 .67955 L p .32835 .68191 L .576 .273 .418 r
F P
0 g
s
.4071 .71447 m .32631 .68132 L .32835 .68191 L .576 .273 .418 r
F
.38658 .70119 m .39808 .71074 L .38432 .70505 L F
.36546 .69267 m .38432 .70505 L .32835 .68191 L F
.32264 .67955 m .32631 .68132 L p .32911 .682 L .504 .363 .621 r
F P
0 g
s
.35622 .6741 m .37702 .69061 L .32264 .67955 L .576 .273 .418 r
F
.37702 .69061 m .34349 .68582 L .32264 .67955 L F
.33192 .6685 m .36128 .66869 L p .31065 .66982 L .246 0 0 r
F P
0 g
s
.38882 .67392 m .36907 .66505 L .30937 .6699 L p .54 .127 .206 r
F P
0 g
s
.34799 .66645 m .32264 .67955 L p .32911 .682 L .504 .363 .621 r
F P
0 g
s
.38186 .6959 m .38054 .69601 L p .40482 .70428 L .617 .3 .409 r
F P
0 g
s
.40473 .70388 m .38186 .6959 L .40482 .70428 L .617 .3 .409 r
F
.42938 .69212 m .37948 .68487 L .38648 .68887 L .54 .127 .206 r
F
.40473 .70388 m .38495 .69174 L p .43555 .71464 L .615 .212 .255 r
F P
0 g
s
.38054 .69601 m .40257 .69425 L p .37318 .69006 L .54 .127 .206 r
F P
0 g
s
.34023 .68536 m .38054 .69601 L p .37318 .69006 L .54 .127 .206 r
F P
0 g
s
.38054 .69601 m .41782 .7272 L p .42198 .72975 L .617 .3 .409 r
F P
0 g
s
.40211 .71093 m .38054 .69601 L .42198 .72975 L .617 .3 .409 r
F
.38368 .69856 m .38307 .69805 L .38368 .69857 L 0 .103 .592 r
F
.40707 .71436 m .40743 .71597 L .42198 .72975 L .617 .3 .409 r
F
.41629 .72591 m .40813 .71909 L .4129 .72306 L 0 .103 .592 r
F
.48841 .77061 m .42198 .72975 L .47259 .76406 L .617 .3 .409 r
F
.42198 .72975 m .42286 .73059 L .47259 .76406 L F
.44359 .74876 m .43028 .73762 L p .43725 .74343 L 0 .103 .592 r
F P
0 g
s
.43725 .74343 m .42569 .72922 L p .44035 .74603 L .599 .332 .483 r
F P
0 g
s
.41403 .71159 m .40863 .70558 L .40482 .70428 L .617 .3 .409 r
F
.40473 .70388 m .44189 .7267 L p .4071 .71447 L .576 .273 .418 r
F P
0 g
s
.44189 .7267 m .42569 .72922 L p .4071 .71447 L .576 .273 .418 r
F P
0 g
s
.42569 .72922 m .44 .72699 L p .4681 .77344 L .599 .332 .483 r
F P
0 g
s
.4681 .77344 m .47052 .77744 L .45724 .76316 L .599 .332 .483 r
F
.4681 .77344 m .45724 .76316 L .42569 .72922 L F
.47859 .78612 m .48256 .79444 L .446 .75251 L F
.45724 .76316 m .446 .75251 L .42569 .72922 L F
.47859 .78612 m .446 .75251 L .45724 .76316 L F
.42569 .72922 m .43935 .74518 L .43853 .73854 L .48 .312 .576 r
F
.46128 .75809 m .4724 .77296 L .46836 .76956 L .599 .332 .483 r
F
.44189 .7267 m .4633 .75916 L p .46128 .75809 L F P
0 g
s
.4633 .75916 m .4724 .77296 L p .46128 .75809 L .599 .332 .483 r
F P
0 g
s
.46128 .75809 m .47859 .78612 L .45889 .75683 L .599 .332 .483 r
F
.44 .72699 m .44189 .7267 L p .46845 .77377 L F P
0 g
s
.46845 .77377 m .47052 .77744 L .4681 .77344 L .599 .332 .483 r
F
.46845 .77377 m .4681 .77344 L .44 .72699 L F
.42198 .72975 m .41782 .7272 L .42286 .73059 L .617 .3 .409 r
F
.43725 .74343 m .40813 .71909 L .43512 .74164 L 0 .103 .592 r
F
.43028 .73762 m .40813 .71909 L p .43725 .74343 L F P
0 g
s
.41298 .72309 m .40855 .72096 L p .40812 .71902 L .576 .273 .418 r
F P
0 g
s
.4071 .71447 m .41298 .72309 L .40812 .71902 L .576 .273 .418 r
F
.42569 .72922 m .41298 .72309 L p .4071 .71447 L F P
0 g
s
.4129 .72306 m .42569 .72922 L p .40627 .71557 L .504 .363 .621 r
F P
0 g
s
.4129 .72306 m .40627 .71557 L .40752 .71852 L .504 .363 .621 r
F
.40752 .71852 m .3831 .69808 L p .406 .71493 L .452 .336 .631 r
F P
0 g
s
.38368 .69856 m .38368 .69857 L .40757 .71857 L 0 .103 .592 r
F
.40758 .71857 m .40757 .71857 L .41727 .72668 L F
.38368 .69856 m .40757 .71857 L .40758 .71857 L F
.42938 .69212 m .38972 .6853 L .38901 .68626 L .54 .127 .206 r
F
.45014 .71973 m .42938 .69212 L .38186 .6959 L p .617 .3 .409 r
F P
0 g
s
.42938 .69212 m .39583 .67706 L p .38972 .6853 L .54 .127 .206 r
F P
0 g
s
.3806 .69473 m .39375 .68593 L p .39911 .6986 L .452 .336 .631 r
F P
0 g
s
.44189 .7267 m .40473 .70388 L p .43555 .71464 L .615 .212 .255 r
F P
0 g
s
.39911 .6986 m .4189 .71105 L p .43632 .70762 L .432 .384 .704 r
F P
0 g
s
.43615 .70635 m .43632 .70762 L .43937 .70702 L .432 .384 .704 r
F
.39911 .6986 m .43632 .70762 L .43615 .70635 L F
.39583 .67706 m .42938 .69212 L .47591 .69651 L p .524 0 0 r
F P
0 g
s
.44802 .6747 m .4463 .68932 L .41871 .68262 L .615 .212 .255 r
F
.39583 .67706 m .38882 .67392 L p .38972 .6853 L .54 .127 .206 r
F P
0 g
s
.44802 .6747 m .38439 .67429 L p .41926 .67139 L .246 0 0 r
F P
0 g
s
.43604 .70558 m .41095 .67442 L .43515 .69914 L .314 .297 .679 r
F
.43817 .72317 m .39967 .69895 L p .39921 .69884 L .504 .363 .621 r
F P
0 g
s
.40707 .71436 m .40482 .70428 L .38054 .69601 L .617 .3 .409 r
F
.402 .70546 m .43817 .72317 L .39921 .69884 L .504 .363 .621 r
F
.44301 .71724 m .44189 .7267 L .43555 .71464 L .615 .212 .255 r
F
.4189 .71105 m .43817 .72317 L .44201 .71972 L p .432 .384 .704 r
F P
0 g
s
.45453 .71577 m .46458 .70672 L .42938 .69212 L .687 .306 .32 r
F
.44201 .71972 m .45299 .70987 L p .43632 .70762 L .432 .384 .704 r
F P
0 g
s
.4189 .71105 m .43777 .71813 L .43632 .70762 L .432 .384 .704 r
F
.43777 .71813 m .44201 .71972 L .43632 .70762 L F
.46653 .76086 m .46327 .75914 L .47112 .77159 L .617 .3 .409 r
F
.46327 .75914 m .44359 .74876 L .47112 .77159 L F
.44359 .74876 m .47126 .7719 L p .47112 .77159 L F P
0 g
s
.47259 .76406 m .46653 .76086 L .47126 .7719 L .617 .3 .409 r
F
.48294 .77102 m .48543 .77083 L p .47259 .76406 L F P
0 g
s
.48543 .77083 m .48841 .77061 L p .47259 .76406 L .617 .3 .409 r
F P
0 g
s
.47126 .7719 m .48294 .77102 L p .47259 .76406 L .617 .3 .409 r
F P
0 g
s
.48841 .77061 m .46713 .74231 L p .46493 .75198 L .617 .3 .409 r
F P
0 g
s
.4633 .75916 m .44189 .7267 L .47064 .72688 L p .668 .348 .423 r
F P
0 g
s
.48225 .76477 m .471 .77166 L p .46024 .75423 L .452 .336 .631 r
F P
0 g
s
.471 .77166 m .46183 .76398 L p .46219 .75739 L .452 .336 .631 r
F P
0 g
s
.44189 .7267 m .46128 .75809 L .45889 .75683 L .599 .332 .483 r
F
.46183 .76398 m .44318 .74838 L p .46063 .75665 L .452 .336 .631 r
F P
0 g
s
.46183 .76398 m .46063 .75665 L .46219 .75739 L .452 .336 .631 r
F
.43817 .72317 m .46063 .75665 L p .46048 .75571 L .362 .274 .62 r
F P
0 g
s
.43935 .74518 m .44318 .74838 L .43671 .72388 L .48 .312 .576 r
F
.44318 .74838 m .4435 .74863 L .44346 .74851 L F
.44318 .74838 m .44346 .74851 L .43671 .72388 L F
.44318 .74838 m .46183 .76398 L p .44172 .74716 L 0 .103 .592 r
F P
0 g
s
.46183 .76398 m .471 .77166 L p .46169 .76387 L 0 .103 .592 r
F P
0 g
s
.44172 .74716 m .46183 .76398 L .46169 .76387 L 0 .103 .592 r
F
.4698 .77031 m .47018 .77087 L p .46754 .76601 L .362 .274 .62 r
F P
0 g
s
.4724 .77296 m .48636 .79412 L p .47576 .78019 L .599 .332 .483 r
F P
0 g
s
.48636 .79412 m .48642 .79551 L .47576 .78019 L .599 .332 .483 r
F
.47224 .77282 m .47576 .78019 L .46836 .76956 L F
.4724 .77296 m .47576 .78019 L .47224 .77282 L F
.51544 .86122 m .51706 .86111 L p .51245 .85214 L .64 .363 .484 r
F P
0 g
s
.51245 .85214 m .5124 .85204 L .51251 .85235 L .64 .363 .484 r
F
.51245 .85214 m .51251 .85235 L .51544 .86122 L F
.51585 .849 m .51993 .88586 L p .5158 .84895 L .685 .407 .496 r
F P
0 g
s
.51985 .8844 m .51544 .86122 L .51976 .88358 L .649 .393 .518 r
F
.52 .88486 m .51982 .88582 L p .51985 .88572 L .62 .394 .553 r
F P
0 g
s
.51992 .88548 m .52 .88486 L .51985 .88572 L .62 .394 .553 r
F
.51985 .88579 m .52 .88486 L p .51985 .88577 L .212 .036 .411 r
F P
0 g
s
.51998 .88514 m .52001 .88483 L .52002 .88481 L .883 .733 .222 r
F
.51992 .88585 m .51999 .88499 L .51986 .8858 L .698 .467 .569 r
F
.51998 .88516 m .51998 .88513 L .51998 .88514 L .883 .733 .222 r
F
.51998 .88513 m .52001 .88483 L .51998 .88514 L F
.52238 .86116 m .52251 .86213 L p .52002 .88481 L F P
0 g
s
.52001 .88483 m .52238 .86116 L .52002 .88481 L .883 .733 .222 r
F
.51999 .88504 m .52004 .88451 L .51998 .88516 L F
.52004 .88451 m .52238 .86116 L .52006 .8843 L F
.51999 .88434 m .52002 .88437 L .5196 .86133 L .809 .517 .495 r
F
.52175 .86189 m .52079 .86164 L p .52002 .88437 L F P
0 g
s
.52079 .86164 m .5196 .86133 L p .52002 .88437 L .809 .517 .495 r
F P
0 g
s
.52002 .88437 m .52002 .88447 L .51997 .88484 L .212 .036 .411 r
F
.52 .88486 m .52 .88483 L .51995 .88517 L F
.51998 .88513 m .51998 .88513 L .51998 .88516 L .883 .733 .222 r
F
.52004 .88451 m .52006 .8843 L .51998 .88513 L F
.52264 .86233 m .52 .88486 L p .52 .88491 L .993 .746 .397 r
F P
0 g
s
.52 .88486 m .52238 .86116 L p .51999 .88504 L .883 .733 .222 r
F P
0 g
s
.52001 .88481 m .52 .88486 L .52001 .88481 L .914 .646 .517 r
F
.52048 .88056 m .52061 .87937 L .52017 .88336 L .883 .733 .222 r
F
.52251 .86213 m .52264 .86233 L p .52 .88491 L .993 .746 .397 r
F P
0 g
s
.5208 .87763 m .52 .88486 L p .52061 .87937 L .883 .733 .222 r
F P
0 g
s
.52061 .87937 m .52229 .8641 L .5208 .87763 L .883 .733 .222 r
F
.52251 .86213 m .52229 .86414 L p .52229 .8641 L F P
0 g
s
.52229 .86414 m .5208 .87763 L p .52229 .8641 L .883 .733 .222 r
F P
0 g
s
.52 .88486 m .52251 .86213 L p .52 .88491 L .993 .746 .397 r
F P
0 g
s
.51993 .88586 m .51999 .88587 L p .52 .88498 L .773 .533 .571 r
F P
0 g
s
.52004 .88451 m .51998 .88513 L .51998 .88513 L .883 .733 .222 r
F
.51999 .88501 m .52 .88487 L .52 .88488 L F
.52 .88487 m .52 .88486 L .52 .88489 L F
.51993 .88586 m .52 .88486 L p .51999 .88546 L .698 .467 .569 r
F P
0 g
s
.51999 .88501 m .52002 .88481 L .52 .88487 L .883 .733 .222 r
F
.52 .88486 m .51993 .88586 L p .52002 .88475 L .773 .533 .571 r
F P
0 g
s
.51999 .88467 m .52001 .88449 L .52002 .88441 L .212 .036 .411 r
F
.51992 .88548 m .51998 .88515 L .52 .88486 L .62 .394 .553 r
F
.51992 .88548 m .52 .88486 L .51998 .88515 L F
.51993 .8854 m .52 .88486 L .51993 .88539 L 0 0 .181 r
F
.51985 .88586 m .52 .88486 L p .51992 .88548 L .62 .394 .553 r
F P
0 g
s
.51989 .88582 m .52 .88486 L p .51993 .8854 L 0 0 .181 r
F P
0 g
s
.51991 .88554 m .51993 .8854 L .52 .88486 L 0 0 .181 r
F
.51989 .88582 m .51991 .88554 L .52 .88486 L F
.52 .88486 m .51989 .88582 L p .52 .88463 L .212 .036 .411 r
F P
0 g
s
.51997 .88502 m .52 .88486 L p .51998 .88506 L .649 .393 .518 r
F P
0 g
s
.52 .88484 m .52 .88483 L .52 .8848 L .212 .036 .411 r
F
.52 .88484 m .52 .8848 L .51999 .88467 L F
.52 .88485 m .52 .88484 L .51999 .88467 L F
.52 .88486 m .51993 .88335 L .51995 .88455 L .728 .45 .506 r
F
.52 .88486 m .51999 .88504 L .51998 .88514 L .883 .733 .222 r
F
.51998 .88514 m .52 .88492 L .52 .8849 L F
.51997 .88525 m .52 .8849 L .52 .88492 L F
.52175 .86189 m .52 .88486 L p .52004 .88392 L .891 .597 .483 r
F P
0 g
s
.52 .88483 m .52 .88486 L .52 .88482 L .891 .597 .483 r
F
.52 .88483 m .52 .88482 L .52175 .86189 L F
.52004 .88392 m .52 .88483 L .52175 .86189 L F
.52 .88483 m .52001 .88478 L .52002 .88447 L .212 .036 .411 r
F
.52264 .86233 m .52527 .84833 L .52413 .84998 L .889 .573 .442 r
F
.52253 .86228 m .52264 .86233 L p .52413 .84998 L F P
0 g
s
.52477 .85685 m .52572 .85097 L p .5266 .84994 L .915 .619 .467 r
F P
0 g
s
.52466 .85728 m .5247 .85551 L .52531 .85329 L .849 .544 .472 r
F
.52279 .86153 m .52591 .85211 L .52527 .84833 L .889 .573 .442 r
F
.52264 .86233 m .52276 .86196 L p .52279 .86153 L F P
0 g
s
.52276 .86196 m .52591 .85211 L p .52279 .86153 L .889 .573 .442 r
F P
0 g
s
.52503 .85429 m .52575 .85117 L .52585 .85176 L .997 .768 .444 r
F
.53129 .82454 m .53115 .8242 L p .52715 .84091 L .986 .776 .379 r
F P
0 g
s
.53731 .80267 m .53209 .81796 L .53241 .8164 L .842 .49 .388 r
F
.53731 .80267 m .52981 .82369 L .53175 .81097 L F
.52527 .8528 m .52575 .85117 L p .52503 .85429 L .997 .768 .444 r
F P
0 g
s
.5257 .85089 m .5249 .84612 L .52689 .84356 L .849 .544 .472 r
F
.52677 .84758 m .52894 .83653 L .52569 .85021 L .986 .776 .379 r
F
.52894 .83653 m .53015 .82979 L .52569 .85021 L F
.5249 .84612 m .52337 .82124 L .52689 .84356 L .849 .544 .472 r
F
.52569 .85021 m .53129 .82454 L .52715 .84091 L .986 .776 .379 r
F
.53009 .8238 m .52981 .82369 L p .53241 .8164 L .842 .49 .388 r
F P
0 g
s
.53014 .82184 m .52568 .85024 L p .52569 .85021 L .986 .776 .379 r
F P
0 g
s
.52568 .85024 m .52721 .84052 L p .52583 .84935 L 0 0 0 r
F P
0 g
s
.52251 .86213 m .52238 .86116 L p .52571 .85011 L .935 .819 .365 r
F P
0 g
s
.52571 .85011 m .53122 .83179 L .52537 .85215 L .935 .819 .365 r
F
.52571 .85011 m .52537 .85215 L .52251 .86213 L F
.52721 .84052 m .52717 .84079 L .52588 .84901 L 0 0 0 r
F
.52571 .85011 m .52733 .84373 L .53122 .83179 L .935 .819 .365 r
F
.52573 .84997 m .52733 .84373 L .52571 .85011 L F
.52721 .84052 m .52588 .84901 L .52583 .84935 L 0 0 0 r
F
.52502 .85186 m .52595 .84858 L .52562 .8504 L .935 .819 .365 r
F
.52595 .84858 m .52733 .84373 L .52573 .84997 L F
.52595 .84858 m .52573 .84997 L .52562 .8504 L F
.52667 .84986 m .52679 .84925 L p .52346 .86448 L .986 .776 .379 r
F P
0 g
s
.53081 .83407 m .52833 .84343 L .52641 .85 L .997 .768 .444 r
F
.52641 .85 m .52833 .84343 L .52738 .847 L F
.52572 .85097 m .52699 .84742 L .5261 .85052 L .915 .619 .467 r
F
.52477 .85685 m .5266 .84994 L .52667 .84986 L F
.52641 .85 m .52738 .847 L .52591 .85211 L .997 .768 .444 r
F
.5266 .84994 m .52591 .85211 L p .52738 .847 L F P
0 g
s
.52591 .85211 m .5266 .84994 L p .52572 .85097 L .889 .573 .442 r
F P
0 g
s
.52575 .85117 m .52779 .84428 L p .52585 .85176 L .997 .768 .444 r
F P
0 g
s
.52585 .85176 m .52779 .84428 L .52641 .85 L .997 .768 .444 r
F
.52264 .86233 m .52251 .86213 L p .52641 .85 L F P
0 g
s
.52591 .85211 m .52264 .86233 L p .52641 .85 L .997 .768 .444 r
F P
0 g
s
.52679 .84925 m .52692 .84722 L .52346 .86448 L .986 .776 .379 r
F
.52692 .84722 m .52696 .84663 L .52677 .84758 L F
.53209 .81796 m .53009 .8238 L .53241 .8164 L .842 .49 .388 r
F
.53334 .81654 m .53009 .8238 L .53731 .80267 L F
.54246 .79732 m .53009 .8238 L p .54407 .79256 L F P
0 g
s
.53009 .8238 m .53312 .81733 L p .53261 .81934 L .914 .593 .422 r
F P
0 g
s
.53115 .8242 m .53014 .82184 L p .52811 .83483 L .986 .776 .379 r
F P
0 g
s
.5461 .78954 m .54246 .79732 L p .54407 .79256 L .842 .49 .388 r
F P
0 g
s
.55625 .76719 m .5442 .79218 L .56073 .75633 L .842 .49 .388 r
F
.54246 .79732 m .55339 .77392 L p .53971 .80576 L .914 .593 .422 r
F P
0 g
s
.51637 .77329 m .51545 .77304 L p .5321 .81012 L .803 .476 .436 r
F P
0 g
s
.52527 .8528 m .52295 .86062 L p .52537 .85218 L .935 .819 .365 r
F P
0 g
s
.52649 .84828 m .52295 .86062 L .52298 .86048 L .935 .819 .365 r
F
.52251 .86213 m .52527 .8528 L p .52503 .85429 L .997 .768 .444 r
F P
0 g
s
.52303 .85989 m .52422 .85897 L p .52438 .85837 L .144 0 .315 r
F P
0 g
s
.5442 .79218 m .5256 .82016 L .52567 .81803 L .842 .49 .388 r
F
.54407 .79256 m .5256 .82017 L .5442 .79218 L F
.5256 .82017 m .52458 .8217 L .5256 .82016 L F
.5256 .82017 m .5256 .82016 L .5442 .79218 L F
.5461 .78954 m .54407 .79256 L .5442 .79218 L F
.54407 .79256 m .53334 .81654 L .53731 .80267 L F
.54657 .78516 m .5442 .79218 L .54142 .79204 L F
.53842 .80462 m .54641 .7863 L .53253 .80975 L .889 .573 .442 r
F
.53948 .8037 m .54246 .79732 L p .53971 .80576 L .914 .593 .422 r
F P
0 g
s
.54212 .8014 m .54641 .7863 L .53842 .80462 L .889 .573 .442 r
F
.55228 .7683 m .52742 .81419 L .55014 .75741 L .842 .49 .388 r
F
.55006 .77487 m .52742 .81419 L .55228 .7683 L F
.52742 .81419 m .52569 .8157 L .55014 .75741 L F
.53825 .77937 m .54641 .78371 L p .5512 .77113 L .882 .52 .354 r
F P
0 g
s
.5399 .80757 m .54212 .8014 L .54064 .80268 L .889 .573 .442 r
F
.53971 .80576 m .54064 .80268 L .53948 .8037 L F
.5399 .80757 m .54064 .80268 L .53971 .80576 L F
.53591 .79944 m .53467 .80548 L .52337 .82124 L .842 .49 .388 r
F
.54758 .78434 m .53825 .77937 L p .53253 .80975 L .889 .573 .442 r
F P
0 g
s
.54584 .77866 m .54622 .78203 L p .54239 .79475 L .935 .819 .365 r
F P
0 g
s
.5379 .80965 m .54117 .79686 L .54239 .79475 L .935 .819 .365 r
F
.54117 .79686 m .54584 .77866 L .54239 .79475 L F
.5311 .81653 m .53014 .82184 L p .53632 .80533 L .826 .841 .375 r
F P
0 g
s
.53889 .80462 m .54584 .77866 L .5379 .80965 L .935 .819 .365 r
F
.53847 .80672 m .53621 .81233 L .53815 .80838 L .998 .825 .484 r
F
.53621 .81233 m .53129 .82454 L .5373 .81139 L F
.53621 .81233 m .5373 .81139 L .53775 .8104 L F
.53621 .81233 m .53775 .8104 L .53815 .80838 L F
.53938 .80215 m .54584 .77866 L .53889 .80462 L .935 .819 .365 r
F
.54212 .8014 m .54758 .78434 L p .54641 .7863 L .889 .573 .442 r
F P
0 g
s
.53014 .82184 m .53129 .82454 L p .53847 .80672 L .998 .825 .484 r
F P
0 g
s
.5379 .80965 m .54306 .7925 L .53763 .81102 L .935 .819 .365 r
F
.54622 .78203 m .54331 .79185 L p .54306 .7925 L F P
0 g
s
.54331 .79185 m .5376 .81112 L p .54306 .7925 L .935 .819 .365 r
F P
0 g
s
.5376 .81112 m .53634 .8154 L p .53763 .81102 L .935 .819 .365 r
F P
0 g
s
.54306 .7925 m .5376 .81112 L .53763 .81102 L .935 .819 .365 r
F
.55268 .77032 m .55006 .77487 L .55228 .7683 L .842 .49 .388 r
F
.54641 .78371 m .54758 .78434 L p .5512 .77113 L .882 .52 .354 r
F P
0 g
s
.54622 .78203 m .54758 .78434 L p .54192 .80188 L .997 .768 .444 r
F P
0 g
s
.53997 .80814 m .54212 .8014 L p .5399 .80757 L .889 .573 .442 r
F P
0 g
s
.53964 .80897 m .54057 .80516 L .54192 .80188 L .997 .768 .444 r
F
.54057 .80516 m .54622 .78203 L .54192 .80188 L F
.53964 .80897 m .54192 .80188 L .53997 .80814 L F
.54758 .78434 m .54212 .8014 L p .54192 .80188 L F P
0 g
s
.54212 .8014 m .53997 .80814 L p .54192 .80188 L .997 .768 .444 r
F P
0 g
s
.53795 .80996 m .54622 .78203 L p .53831 .80917 L .997 .768 .444 r
F P
0 g
s
.53795 .80996 m .53831 .80917 L .53779 .81097 L .997 .768 .444 r
F
.5376 .81112 m .53795 .80996 L p .53779 .81097 L F P
0 g
s
.5398 .80659 m .53865 .80901 L .53895 .80809 L .914 .593 .422 r
F
.54241 .80195 m .54775 .78846 L .56535 .7483 L F
.5398 .80659 m .53895 .80809 L .53971 .80576 L F
.53481 .82117 m .54622 .78203 L .53964 .80897 L .997 .768 .444 r
F
.54593 .79249 m .53775 .8104 L .54 .80846 L .998 .825 .484 r
F
.53775 .8104 m .5373 .81139 L .53762 .81104 L F
.53775 .8104 m .53762 .81104 L .54 .80846 L F
.54758 .78434 m .55266 .77021 L .5512 .77113 L .882 .52 .354 r
F
.57137 .71822 m .55014 .75741 L p .55266 .77021 L F P
0 g
s
.55014 .75741 m .53825 .77937 L p .55266 .77021 L .882 .52 .354 r
F P
0 g
s
.56073 .75633 m .55316 .77275 L .55268 .77032 L .842 .49 .388 r
F
.55316 .77275 m .54881 .78217 L .55268 .77032 L F
.56535 .7483 m .54688 .74082 L p .5599 .73939 L .803 .361 .196 r
F P
0 g
s
.58357 .73679 m .56637 .74766 L .5599 .73939 L .803 .361 .196 r
F
.56637 .74766 m .56535 .7483 L .5599 .73939 L F
.55299 .77403 m .55339 .77392 L .55333 .77363 L .882 .52 .354 r
F
.55299 .77403 m .55333 .77363 L .55266 .77021 L F
.55109 .77456 m .55299 .77403 L .55266 .77021 L F
.54758 .78434 m .55339 .77392 L .55109 .77456 L F
.55422 .77369 m .554 .77375 L .56535 .7483 L .914 .593 .422 r
F
.55004 .77752 m .54622 .78203 L p .55018 .77711 L .973 .908 .565 r
F P
0 g
s
.56381 .75098 m .55268 .77032 L .55014 .75741 L .842 .49 .388 r
F
.55361 .76703 m .55335 .76779 L .55412 .76609 L .826 .841 .375 r
F
.53505 .73603 m .56535 .7483 L p .56381 .75098 L .842 .49 .388 r
F P
0 g
s
.53088 .76647 m .53505 .73603 L p .56381 .75098 L .842 .49 .388 r
F P
0 g
s
.5658 .73895 m .5532 .76823 L p .55423 .76765 L .998 .825 .484 r
F P
0 g
s
.56268 .76141 m .56486 .76159 L .57068 .75181 L .998 .825 .484 r
F
.56042 .76412 m .5623 .76305 L .56268 .76141 L F
.5623 .76305 m .56486 .76159 L .56268 .76141 L F
.54622 .78203 m .5623 .76305 L p .56222 .76341 L .997 .8 .518 r
F P
0 g
s
.54622 .78203 m .54533 .77419 L p .55031 .76988 L .973 .908 .565 r
F P
0 g
s
.55031 .76988 m .5532 .76823 L .54622 .78203 L .973 .908 .565 r
F
.5532 .76823 m .56085 .76388 L .55018 .77711 L F
.5532 .76823 m .55018 .77711 L .54622 .78203 L F
.55966 .77217 m .54 .80846 L p .55912 .77232 L .914 .593 .422 r
F P
0 g
s
.5541 .77756 m .55917 .7723 L p .55339 .77392 L .882 .52 .354 r
F P
0 g
s
.55195 .7765 m .55359 .77492 L .55339 .77392 L .882 .52 .354 r
F
.54758 .78434 m .55359 .77492 L .55195 .7765 L F
.54758 .78434 m .5541 .77756 L p .55359 .77492 L F P
0 g
s
.54622 .78203 m .56222 .76341 L .56128 .7675 L .997 .8 .518 r
F
.5623 .76305 m .55004 .77752 L p .56085 .76388 L .973 .908 .565 r
F P
0 g
s
.56085 .76388 m .55004 .77752 L .55018 .77711 L .973 .908 .565 r
F
.5658 .73895 m .55539 .76698 L .56398 .75577 L .998 .825 .484 r
F
.55539 .76698 m .55966 .76455 L .56398 .75577 L F
.5532 .76823 m .55412 .76609 L p .55335 .76779 L .826 .841 .375 r
F P
0 g
s
.54758 .78434 m .54622 .78203 L p .56128 .7675 L .997 .8 .518 r
F P
0 g
s
.56128 .7675 m .56095 .76895 L .54758 .78434 L .997 .8 .518 r
F
.55917 .7723 m .54758 .78434 L p .56095 .76895 L F P
0 g
s
.56049 .77094 m .58479 .74573 L p .56647 .76209 L .882 .52 .354 r
F P
0 g
s
.58479 .74573 m .58394 .74629 L .56647 .76209 L .882 .52 .354 r
F
.56061 .77041 m .56051 .77086 L .55917 .7723 L .997 .8 .518 r
F
.58294 .7466 m .58427 .74563 L .55917 .7723 L F
.58635 .74411 m .58479 .74573 L p .58427 .74563 L F P
0 g
s
.58479 .74573 m .55917 .7723 L p .58427 .74563 L .997 .8 .518 r
F P
0 g
s
.59495 .73783 m .57638 .74902 L .58418 .74164 L .906 .541 .331 r
F
.58631 .74414 m .57638 .74902 L .59495 .73783 L F
.61091 .73273 m .63675 .73267 L p .59995 .72967 L .827 .365 .115 r
F P
0 g
s
.61372 .73272 m .61091 .73273 L p .61383 .73246 L .98 .833 .626 r
F P
0 g
s
.64641 .73467 m .63873 .73259 L .6374 .73073 L .819 .337 .015 r
F
.65236 .73728 m .63942 .73283 L .64641 .73467 L F
.63942 .73283 m .63873 .73259 L .64641 .73467 L F
.65546 .73874 m .66651 .74601 L p .64209 .7306 L 0 .394 .79 r
F P
0 g
s
.6615 .74128 m .6586 .74022 L p .65236 .73728 L .819 .337 .015 r
F P
0 g
s
.6586 .74022 m .6615 .74128 L .6599 .73623 L p .927 .88 .768 r
F P
0 g
s
.63615 .72675 m .64711 .73133 L p .6409 .72876 L .409 0 0 r
F P
0 g
s
.64621 .73264 m .64064 .72159 L p .65902 .73894 L .9 .845 .781 r
F P
0 g
s
.66651 .74601 m .65844 .7407 L p .65902 .73894 L .9 .845 .781 r
F P
0 g
s
.65844 .7407 m .64621 .73264 L p .65902 .73894 L .9 .845 .781 r
F P
0 g
s
.6586 .74022 m .63905 .73305 L p .65236 .73728 L .819 .337 .015 r
F P
0 g
s
.63675 .73267 m .63878 .73266 L p .59916 .72639 L .827 .365 .115 r
F P
0 g
s
.63878 .73266 m .63798 .73155 L .59916 .72639 L .827 .365 .115 r
F
.63183 .73083 m .63599 .73193 L .61372 .73272 L .98 .833 .626 r
F
.63878 .73266 m .63675 .73267 L p .63599 .73193 L F P
0 g
s
.63675 .73267 m .61372 .73272 L p .63599 .73193 L .98 .833 .626 r
F P
0 g
s
.62797 .73022 m .63798 .73155 L .63437 .7291 L .827 .365 .115 r
F
.63905 .73305 m .62403 .72754 L p .63873 .73259 L .819 .337 .015 r
F P
0 g
s
.63905 .73305 m .63873 .73259 L .65236 .73728 L .819 .337 .015 r
F
.61329 .72018 m .63876 .72858 L p .63979 .72921 L F P
0 g
s
.65486 .73837 m .63205 .72839 L .63979 .72921 L .819 .337 .015 r
F
.63205 .72839 m .61329 .72018 L .63979 .72921 L F
.62403 .72754 m .62393 .72747 L p .6247 .72761 L F P
0 g
s
.6374 .73073 m .62403 .72754 L .63205 .72839 L .819 .337 .015 r
F
.63873 .73259 m .62403 .72754 L .6374 .73073 L F
.62403 .72754 m .63599 .73193 L p .63037 .72648 L .927 .88 .768 r
F P
0 g
s
.62403 .72754 m .61397 .73212 L p .61005 .73249 L .906 .541 .331 r
F P
0 g
s
.63183 .73083 m .61091 .73273 L .61602 .73171 L .98 .833 .626 r
F
.61397 .73212 m .61198 .73302 L p .61091 .73273 L .906 .541 .331 r
F P
0 g
s
.61397 .73212 m .61091 .73273 L .61005 .73249 L .906 .541 .331 r
F
.61329 .72018 m .62403 .72754 L p .61005 .73249 L F P
0 g
s
.63496 .72733 m .61329 .72018 L .58766 .70947 L p .409 0 0 r
F P
0 g
s
.59803 .72913 m .61329 .72018 L p .61005 .73249 L .906 .541 .331 r
F P
0 g
s
.60834 .72463 m .61747 .72377 L p .61624 .72672 L .98 .833 .626 r
F P
0 g
s
.62403 .72754 m .6229 .72419 L p .61635 .72646 L .99 .868 .614 r
F P
0 g
s
.61635 .72646 m .61575 .72667 L .61624 .72672 L .99 .868 .614 r
F
.61635 .72646 m .61624 .72672 L .62403 .72754 L F
.60621 .72162 m .5974 .72793 L .59524 .72384 L .803 .361 .196 r
F
.5974 .72793 m .58531 .7366 L .59524 .72384 L F
.61329 .72018 m .59638 .7301 L p .60621 .72162 L F P
0 g
s
.59638 .7301 m .58531 .7366 L p .60621 .72162 L .803 .361 .196 r
F P
0 g
s
.58951 .71299 m .57137 .71822 L p .5973 .72914 L .827 .365 .115 r
F P
0 g
s
.59824 .72953 m .59787 .72882 L .5973 .72914 L .827 .365 .115 r
F
.59787 .72882 m .58951 .71299 L .5973 .72914 L F
.57792 .73741 m .5822 .73694 L .57137 .71822 L .882 .52 .354 r
F
.60742 .73342 m .58592 .74102 L .58547 .73778 L .906 .541 .331 r
F
.58635 .74411 m .5973 .73276 L .58372 .72514 L p .882 .52 .354 r
F P
0 g
s
.61329 .72018 m .59524 .72384 L .59134 .71645 L .803 .361 .196 r
F
.5973 .73276 m .59824 .72953 L .5973 .72914 L .827 .365 .115 r
F
.57137 .71822 m .59407 .73095 L p .5973 .72914 L F P
0 g
s
.59407 .73095 m .5973 .73276 L p .5973 .72914 L .827 .365 .115 r
F P
0 g
s
.5658 .73895 m .57068 .75181 L p .57911 .74201 L .99 .868 .614 r
F P
0 g
s
.58083 .73795 m .59342 .72261 L .58446 .73457 L .973 .908 .565 r
F
.5811 .7373 m .59342 .72261 L .58083 .73795 L F
.56391 .73895 m .57072 .7382 L .57137 .71822 L .882 .52 .354 r
F
.5877 .72173 m .5658 .73895 L p .57911 .74201 L .99 .868 .614 r
F P
0 g
s
.5737 .7424 m .59342 .72261 L .57961 .73908 L .973 .908 .565 r
F
.57137 .71822 m .58951 .71299 L p .56102 .71019 L .615 .062 0 r
F P
0 g
s
.54553 .70867 m .55107 .69419 L .59808 .71383 L p .803 .361 .196 r
F P
0 g
s
.5658 .73895 m .5877 .72173 L p .57726 .7213 L .912 .973 .681 r
F P
0 g
s
.5737 .7424 m .58123 .72146 L .59342 .72261 L .973 .908 .565 r
F
.59374 .71177 m .60432 .70872 L .55938 .69548 L p .615 .062 0 r
F P
0 g
s
.57267 .68716 m .59084 .70061 L .56247 .69354 L .791 .695 .758 r
F
.59116 .69389 m .59337 .72197 L p .58123 .72146 L .973 .908 .565 r
F P
0 g
s
.59337 .72197 m .59342 .72261 L p .58123 .72146 L .973 .908 .565 r
F P
0 g
s
.56861 .74635 m .56707 .74231 L p .56398 .75577 L .998 .825 .484 r
F P
0 g
s
.56707 .74231 m .5658 .73895 L p .56398 .75577 L .998 .825 .484 r
F P
0 g
s
.54235 .71697 m .54553 .70867 L p .55251 .72996 L .803 .361 .196 r
F P
0 g
s
.5599 .73939 m .57137 .71822 L .53504 .70764 L p .778 .379 .296 r
F P
0 g
s
.5658 .73895 m .56618 .737 L p .55658 .75831 L .826 .841 .375 r
F P
0 g
s
.56618 .737 m .56929 .72097 L p .56174 .74314 L .826 .841 .375 r
F P
0 g
s
.56174 .74314 m .55658 .75831 L .56618 .737 L .826 .841 .375 r
F
.54499 .7712 m .55218 .75671 L p .55359 .75444 L .685 .836 .407 r
F P
0 g
s
.55359 .75444 m .55649 .75042 L .54499 .7712 L .685 .836 .407 r
F
.56174 .74314 m .54617 .76921 L p .55649 .75042 L F P
0 g
s
.54617 .76921 m .54499 .7712 L p .55649 .75042 L .685 .836 .407 r
F P
0 g
s
.57083 .71308 m .55759 .74847 L .56233 .74007 L .826 .841 .375 r
F
.55638 .75213 m .56174 .74314 L p .55923 .75052 L .973 .908 .565 r
F P
0 g
s
.56398 .75577 m .56369 .75703 L .56861 .74635 L .998 .825 .484 r
F
.55412 .76609 m .5658 .73895 L p .55658 .75831 L .826 .841 .375 r
F P
0 g
s
.55361 .76703 m .55412 .76609 L .55658 .75831 L .826 .841 .375 r
F
.5658 .73895 m .54168 .79878 L .54593 .79249 L .998 .825 .484 r
F
.55552 .76143 m .5737 .7424 L .57961 .73908 L .973 .908 .565 r
F
.54499 .7712 m .55593 .76022 L .55552 .76143 L F
.55593 .76022 m .5737 .7424 L .55552 .76143 L F
.54499 .7712 m .55638 .75213 L p .5737 .7424 L F P
0 g
s
.54593 .79249 m .55042 .78585 L .56369 .75703 L .998 .825 .484 r
F
.54593 .79249 m .56369 .75703 L .56398 .75577 L F
.54499 .7712 m .56904 .75371 L .55031 .76988 L .973 .908 .565 r
F
.55031 .76988 m .56904 .75371 L .56085 .76388 L F
.54499 .7712 m .58446 .73457 L .56904 .75371 L F
.54016 .79864 m .54584 .77866 L .54052 .79801 L .935 .819 .365 r
F
.54003 .79887 m .53014 .82184 L p .53847 .80672 L .998 .825 .484 r
F P
0 g
s
.53123 .83178 m .54584 .77866 L .53123 .83178 L .935 .819 .365 r
F
.53362 .81376 m .54003 .79887 L p .53648 .80506 L .826 .841 .375 r
F P
0 g
s
.52733 .84373 m .54384 .77592 L .54584 .77866 L .935 .819 .365 r
F
.54499 .7712 m .54506 .77185 L p .54384 .77592 L F P
0 g
s
.54506 .77185 m .54584 .77866 L p .54384 .77592 L .935 .819 .365 r
F P
0 g
s
.54412 .78449 m .55361 .76703 L .55658 .75831 L .826 .841 .375 r
F
.54533 .77419 m .54517 .77281 L p .55031 .76988 L .973 .908 .565 r
F P
0 g
s
.54517 .77281 m .54499 .7712 L p .55031 .76988 L .973 .908 .565 r
F P
0 g
s
.55759 .74847 m .55726 .74936 L .56233 .74007 L .826 .841 .375 r
F
.55726 .74936 m .54412 .78449 L .56018 .74531 L F
.55726 .74936 m .56018 .74531 L .56233 .74007 L F
.56233 .74007 m .54412 .78449 L .55658 .75831 L F
.56929 .72097 m .57083 .71308 L p .55658 .75831 L F P
0 g
s
.52894 .83653 m .53129 .82454 L .53015 .82979 L .986 .776 .379 r
F
.52678 .84363 m .54361 .77668 L p .52821 .83943 L .935 .819 .365 r
F P
0 g
s
.54361 .77668 m .54499 .7712 L p .54377 .77614 L .935 .819 .365 r
F P
0 g
s
.54361 .77668 m .54377 .77614 L .52821 .83943 L .935 .819 .365 r
F
.52657 .84606 m .54499 .7712 L .52733 .84373 L F
.53699 .80349 m .54222 .78737 L .54359 .77686 L 0 0 0 r
F
.55097 .76093 m .55218 .75671 L .54222 .78737 L F
.54219 .78767 m .54222 .78737 L .53699 .80349 L F
.55097 .76093 m .54222 .78737 L .54219 .78767 L F
.57083 .71308 m .54219 .78767 L .53687 .80385 L .826 .841 .375 r
F
.53656 .8048 m .53673 .8045 L .53678 .80432 L F
.53673 .8045 m .54003 .79887 L .53678 .80432 L F
.54003 .79887 m .5532 .76823 L p .53678 .80432 L F P
0 g
s
.53648 .80506 m .54003 .79887 L .53656 .8048 L .826 .841 .375 r
F
.54499 .7712 m .53648 .80506 L p .53699 .80349 L 0 0 0 r
F P
0 g
s
.53632 .80533 m .5532 .76823 L .55335 .76779 L .826 .841 .375 r
F
.53696 .80363 m .53687 .80385 L .53693 .80372 L F
.53696 .80363 m .53693 .80372 L .55335 .76779 L F
.54412 .78449 m .53696 .80363 L .55335 .76779 L F
.53687 .80385 m .53656 .80479 L .55335 .76779 L F
.52692 .84722 m .52527 .8528 L p .52649 .84828 L .935 .819 .365 r
F P
0 g
s
.52649 .84828 m .52527 .8528 L .52537 .85218 L .935 .819 .365 r
F
.52692 .84722 m .52677 .84758 L .52346 .86448 L .986 .776 .379 r
F
.52615 .84735 m .52657 .84606 L .52595 .84858 L .935 .819 .365 r
F
.52671 .8439 m .52657 .84606 L .52615 .84735 L F
.52595 .84858 m .52657 .84606 L .52733 .84373 L F
.52346 .86448 m .52677 .84758 L .52569 .85021 L .986 .776 .379 r
F
.52117 .87907 m .52667 .84986 L .52346 .86448 L F
.5201 .8859 m .52112 .88029 L p .52117 .87907 L F P
0 g
s
.52112 .88029 m .52667 .84986 L p .52117 .87907 L .986 .776 .379 r
F P
0 g
s
.52183 .8749 m .52597 .84848 L .52671 .8439 L 0 0 0 r
F
.52176 .87539 m .52182 .87498 L .52183 .8749 L F
.52182 .87498 m .52597 .84848 L .52183 .8749 L F
.52671 .8439 m .52678 .84363 L p .52657 .84606 L .935 .819 .365 r
F P
0 g
s
.52678 .84363 m .52821 .83943 L .52657 .84606 L .935 .819 .365 r
F
.52668 .8441 m .53699 .80349 L .52678 .84363 L 0 0 0 r
F
.52632 .84659 m .53699 .80349 L .52668 .8441 L F
.53648 .80506 m .52678 .84363 L p .53699 .80349 L F P
0 g
s
.54219 .78767 m .55649 .75042 L p .5421 .78831 L 0 0 .028 r
F P
0 g
s
.53988 .79962 m .54016 .79864 L .53938 .80215 L .935 .819 .365 r
F
.54016 .79864 m .54052 .79801 L .53938 .80215 L F
.55423 .76765 m .54003 .79887 L .53847 .80672 L .998 .825 .484 r
F
.5532 .76823 m .54003 .79887 L p .55423 .76765 L F P
0 g
s
.5304 .83293 m .53988 .79962 L .5329 .82554 L .935 .819 .365 r
F
.53988 .79962 m .53938 .80215 L .53123 .83178 L F
.52619 .84746 m .54125 .79491 L .52632 .84659 L 0 0 0 r
F
.52633 .84704 m .54125 .79491 L .52619 .84746 L F
.54125 .79491 m .54129 .79456 L .52632 .84659 L F
.54412 .78449 m .55335 .76779 L .55361 .76703 L .826 .841 .375 r
F
.54123 .79508 m .5421 .78831 L .55649 .75042 L 0 0 .028 r
F
.55102 .76077 m .55097 .76093 L .5413 .79453 L 0 0 0 r
F
.54129 .79456 m .5413 .79453 L .52632 .84659 L F
.55102 .76077 m .5413 .79453 L .54129 .79456 L F
.61176 .66266 m .6019 .66885 L .59942 .66608 L .791 .695 .758 r
F
.59794 .62276 m .60033 .62602 L .61176 .66266 L .893 .686 .616 r
F
.59564 .6196 m .61176 .66266 L p .59526 .62094 L .82 .658 .682 r
F P
0 g
s
.58821 .64593 m .59264 .65852 L p .59232 .63137 L .85 .691 .685 r
F P
0 g
s
.61176 .66266 m .58657 .65175 L p .59526 .62094 L .82 .658 .682 r
F P
0 g
s
.5592 .67719 m .60388 .69051 L p .57053 .67086 L .778 .658 .73 r
F P
0 g
s
.60388 .69051 m .58229 .6643 L .57053 .67086 L .778 .658 .73 r
F
.63422 .7049 m .6382 .71239 L p .6019 .66885 L .791 .695 .758 r
F P
0 g
s
.59264 .65852 m .60388 .69051 L .64004 .7114 L p .85 .691 .685 r
F P
0 g
s
.61176 .66266 m .63422 .7049 L p .6019 .66885 L .791 .695 .758 r
F P
0 g
s
.57971 .71234 m .59644 .68186 L .59116 .69389 L .685 .836 .407 r
F
.60513 .66604 m .59838 .6795 L p .59644 .68186 L F P
0 g
s
.59838 .6795 m .59116 .69389 L p .59644 .68186 L .685 .836 .407 r
F P
0 g
s
.59644 .68186 m .57083 .71308 L .56929 .72097 L p .912 .973 .681 r
F P
0 g
s
.56929 .72097 m .56618 .737 L p .57485 .7212 L .912 .973 .681 r
F P
0 g
s
.59644 .68186 m .56929 .72097 L .57485 .7212 L .912 .973 .681 r
F
.57485 .7212 m .59116 .69389 L p .57574 .72123 L .973 .908 .565 r
F P
0 g
s
.55649 .75042 m .57083 .71308 L .57632 .70639 L p 0 0 .028 r
F P
0 g
s
.57632 .70639 m .58943 .69042 L p .57422 .71813 L 0 0 .028 r
F P
0 g
s
.57422 .71813 m .55649 .75042 L .57632 .70639 L 0 0 .028 r
F
.59116 .69389 m .57485 .7212 L p .57971 .71234 L .685 .836 .407 r
F P
0 g
s
.56618 .737 m .5658 .73895 L p .57532 .72122 L .912 .973 .681 r
F P
0 g
s
.59644 .68186 m .57485 .7212 L .57532 .72122 L .912 .973 .681 r
F
.57485 .7212 m .56618 .737 L .57532 .72122 L F
.57574 .72123 m .59116 .69389 L .58123 .72146 L .973 .908 .565 r
F
.58531 .7366 m .58154 .73881 L p .58357 .73679 L .803 .361 .196 r
F P
0 g
s
.59342 .72261 m .59369 .72602 L p .58446 .73457 L .973 .908 .565 r
F P
0 g
s
.58427 .74563 m .57068 .75181 L p .58631 .74414 L .906 .541 .331 r
F P
0 g
s
.57068 .75181 m .55966 .77217 L .55912 .77232 L .914 .593 .422 r
F
.56061 .77041 m .57068 .75181 L p .56124 .76767 L .998 .825 .484 r
F P
0 g
s
.57068 .75181 m .57597 .74941 L p .57808 .74443 L .99 .868 .614 r
F P
0 g
s
.57911 .74201 m .57717 .74426 L .57808 .74443 L .99 .868 .614 r
F
.57717 .74426 m .57068 .75181 L .57808 .74443 L F
.57597 .74941 m .57635 .74924 L p .57628 .74866 L F P
0 g
s
.57635 .74924 m .58427 .74563 L p .57628 .74866 L .99 .868 .614 r
F P
0 g
s
.57808 .74443 m .57628 .74866 L .58427 .74563 L .99 .868 .614 r
F
.57094 .75537 m .57944 .7447 L .57338 .75359 L .997 .8 .518 r
F
.57944 .7447 m .59394 .7265 L .58004 .74481 L F
.57944 .7447 m .58004 .74481 L .57338 .75359 L F
.56949 .75457 m .56534 .75946 L p .58446 .73457 L .973 .908 .565 r
F P
0 g
s
.58446 .73457 m .56534 .75946 L .5614 .76319 L .973 .908 .565 r
F
.59369 .72602 m .57808 .74443 L p .58446 .73457 L F P
0 g
s
.58446 .73457 m .57808 .74443 L .57938 .74136 L .973 .908 .565 r
F
.60834 .72463 m .59369 .72602 L .59336 .72191 L p .942 .938 .76 r
F P
0 g
s
.59336 .72191 m .59246 .71049 L p .59739 .71487 L .942 .938 .76 r
F P
0 g
s
.59739 .71487 m .60834 .72463 L .5934 .72192 L .942 .938 .76 r
F
.59739 .71487 m .5934 .72192 L .59336 .72191 L F
.60379 .71651 m .60834 .72463 L .59246 .71049 L F
.59472 .71621 m .59816 .71351 L p .58205 .71392 L .912 .973 .681 r
F P
0 g
s
.59816 .71351 m .61483 .70041 L p .60276 .70538 L .912 .973 .681 r
F P
0 g
s
.61483 .70041 m .59816 .71351 L p .60188 .71549 L .99 .868 .614 r
F P
0 g
s
.6229 .72419 m .61483 .70041 L p .60188 .71549 L .99 .868 .614 r
F P
0 g
s
.60188 .71549 m .61297 .72138 L .6229 .72419 L .99 .868 .614 r
F
.59246 .71049 m .59116 .69389 L p .60015 .70999 L .942 .938 .76 r
F P
0 g
s
.60015 .70999 m .60379 .71651 L .59816 .71351 L .942 .938 .76 r
F
.60015 .70999 m .59816 .71351 L .59246 .71049 L F
.61789 .72277 m .61728 .72423 L .6229 .72419 L .99 .868 .614 r
F
.61747 .72377 m .60834 .72463 L p .60379 .71651 L .942 .938 .76 r
F P
0 g
s
.59998 .7177 m .61728 .72423 L .61789 .72277 L .99 .868 .614 r
F
.64064 .72159 m .62668 .7229 L p .63019 .71746 L .942 .938 .76 r
F P
0 g
s
.63422 .7049 m .61176 .66266 L p .63566 .70193 L .867 .715 .691 r
F P
0 g
s
.65507 .72886 m .64555 .71458 L .63303 .70735 L p .802 .727 .782 r
F P
0 g
s
.64555 .71458 m .63653 .70014 L .63496 .70337 L .85 .691 .685 r
F
.63937 .70323 m .61483 .70041 L .62039 .71682 L p .927 .88 .768 r
F P
0 g
s
.62039 .71682 m .62391 .72719 L p .62668 .7229 L .927 .88 .768 r
F P
0 g
s
.63937 .70323 m .62039 .71682 L .62668 .7229 L .927 .88 .768 r
F
.61176 .66266 m .62232 .67016 L p .63566 .70193 L .867 .715 .691 r
F P
0 g
s
.61483 .70041 m .61909 .68112 L p .60276 .70538 L .912 .973 .681 r
F P
0 g
s
.61909 .68112 m .62075 .67362 L p .61749 .67939 L .912 .973 .681 r
F P
0 g
s
.61749 .67939 m .60276 .70538 L .61909 .68112 L .912 .973 .681 r
F
.61749 .67939 m .59116 .69389 L .59838 .6795 L p .835 .999 .804 r
F P
0 g
s
.59838 .6795 m .60513 .66604 L p .6074 .6685 L .835 .999 .804 r
F P
0 g
s
.6074 .6685 m .61749 .67939 L .59838 .6795 L .835 .999 .804 r
F
.62723 .64433 m .6272 .64438 L .62722 .64435 L .912 .973 .681 r
F
.59116 .69389 m .60491 .70159 L .60218 .70641 L .942 .938 .76 r
F
.59116 .69389 m .61747 .72377 L .60379 .71651 L F
.61873 .65924 m .62021 .66373 L p .6161 .6535 L 0 0 .414 r
F P
0 g
s
.61483 .70041 m .63937 .70323 L p .63766 .70051 L .852 .941 .88 r
F P
0 g
s
.62512 .67234 m .64 .70345 L .6391 .68442 L p .157 .321 .779 r
F P
0 g
s
.62261 .67656 m .63808 .66804 L .63979 .7039 L p .942 .938 .76 r
F P
0 g
s
.65816 .66896 m .66022 .66176 L .65767 .671 L 0 0 .267 r
F
.65816 .66896 m .65767 .671 L .65764 .67075 L F
.65819 .67638 m .658 .67113 L p .65392 .67846 L .455 .585 .894 r
F P
0 g
s
.64971 .70206 m .65537 .71991 L p .64911 .70293 L .026 .284 .78 r
F P
0 g
s
.64738 .69825 m .64971 .70206 L p .64911 .70293 L .026 .284 .78 r
F P
0 g
s
.65765 .67107 m .658 .67113 L p .65617 .6794 L .816 .871 .887 r
F P
0 g
s
.64971 .70206 m .65712 .72668 L .65765 .7281 L .934 .695 .216 r
F
.65537 .71991 m .64971 .70206 L p .65765 .7281 L F P
0 g
s
.65577 .67999 m .65617 .6794 L .65054 .70478 L .816 .871 .887 r
F
.65765 .67107 m .65617 .6794 L .65577 .67999 L F
.66022 .66176 m .658 .67113 L .65798 .66991 L .763 .892 .935 r
F
.66177 .65615 m .66026 .66213 L .66022 .66176 L F
.66026 .66213 m .658 .67113 L .66022 .66176 L F
.658 .67113 m .65765 .67107 L p .65798 .66991 L F P
0 g
s
.65767 .671 m .66022 .66176 L .65546 .68046 L 0 0 .267 r
F
.66022 .66176 m .65075 .70057 L p .65546 .68046 L F P
0 g
s
.65767 .671 m .65075 .70057 L .6507 .70065 L 0 0 .267 r
F
.65765 .67107 m .65649 .67601 L .65726 .67251 L F
.65763 .6712 m .65649 .67601 L .65765 .67107 L F
.65767 .671 m .65726 .67251 L .65764 .67075 L F
.63916 .68566 m .64 .70345 L p .65392 .67846 L .455 .585 .894 r
F P
0 g
s
.65392 .67846 m .65442 .68199 L .65568 .6776 L .455 .585 .894 r
F
.65568 .6776 m .64971 .70206 L p .6513 .67974 L .434 .524 .851 r
F P
0 g
s
.64971 .70206 m .65006 .7033 L .65007 .70329 L .934 .695 .216 r
F
.6551 .7211 m .65436 .71851 L .65586 .72286 L .026 .284 .78 r
F
.63955 .68547 m .65436 .71851 L .6551 .7211 L F
.63955 .68547 m .64738 .69825 L p .65436 .71851 L F P
0 g
s
.65006 .7033 m .6551 .7211 L .65007 .70329 L .934 .695 .216 r
F
.65007 .70329 m .6551 .7211 L .65521 .72135 L F
.64981 .70443 m .6551 .7211 L p .65123 .70743 L F P
0 g
s
.64971 .70206 m .64981 .70443 L p .65123 .70743 L .934 .695 .216 r
F P
0 g
s
.64971 .70206 m .65547 .72227 L .65712 .72668 L .934 .695 .216 r
F
.65649 .67601 m .64981 .70443 L .65702 .67293 L 0 0 .267 r
F
.65075 .70057 m .64981 .70443 L p .6507 .70065 L F P
0 g
s
.64563 .6825 m .6513 .67974 L .64971 .70206 L .434 .524 .851 r
F
.64981 .70443 m .64971 .70206 L .65568 .6776 L p 0 0 .267 r
F P
0 g
s
.65702 .67293 m .64981 .70443 L .65568 .6776 L 0 0 .267 r
F
.64971 .70206 m .63955 .68547 L p .64563 .6825 L .434 .524 .851 r
F P
0 g
s
.64981 .70443 m .65075 .70057 L p .64796 .70056 L .852 .941 .88 r
F P
0 g
s
.6598 .73507 m .6615 .74128 L .6608 .73822 L .934 .695 .216 r
F
.64555 .71458 m .65507 .72886 L p .6602 .73327 L .005 .313 .804 r
F P
0 g
s
.6602 .73327 m .64 .70345 L .64555 .71458 L p .005 .313 .804 r
F P
0 g
s
.65712 .72668 m .65884 .73243 L .6598 .73507 L .934 .695 .216 r
F
.65886 .73247 m .64064 .72159 L p .65801 .73012 L .816 .871 .887 r
F P
0 g
s
.65884 .73243 m .6615 .74128 L .6598 .73507 L .934 .695 .216 r
F
.64064 .72159 m .63979 .7039 L p .65801 .73012 L .816 .871 .887 r
F P
0 g
s
.65881 .73121 m .66032 .73652 L .6598 .73388 L .934 .695 .216 r
F
.66032 .73652 m .6608 .73822 L .66058 .73724 L F
.66032 .73652 m .66058 .73724 L .6598 .73388 L F
.63808 .66804 m .64672 .69353 L .64018 .70332 L .816 .871 .887 r
F
.63979 .7039 m .63973 .70267 L p .64018 .70332 L F P
0 g
s
.63973 .70267 m .63808 .66804 L p .64018 .70332 L .816 .871 .887 r
F P
0 g
s
.64981 .70443 m .64018 .70332 L p .64429 .70937 L .927 .88 .768 r
F P
0 g
s
.64933 .68963 m .64205 .70053 L .65075 .70057 L .852 .941 .88 r
F
.64796 .70056 m .64205 .70053 L .64981 .70443 L F
.64205 .70053 m .64128 .70168 L .64981 .70443 L F
.63937 .70323 m .64981 .70443 L p .63766 .70051 L F P
0 g
s
.64018 .70332 m .65801 .73012 L .63979 .7039 L .816 .871 .887 r
F
.65547 .72227 m .6533 .72263 L .6563 .72705 L .927 .88 .768 r
F
.65759 .72894 m .65801 .73012 L .64018 .70332 L .816 .871 .887 r
F
.65902 .73894 m .64064 .72159 L .6599 .73623 L .9 .845 .781 r
F
.6599 .73623 m .65759 .72894 L p .6586 .74022 L .927 .88 .768 r
F P
0 g
s
.65759 .72894 m .65547 .72227 L p .6563 .72705 L .927 .88 .768 r
F P
0 g
s
.6563 .72705 m .6586 .74022 L .65759 .72894 L .927 .88 .768 r
F
.6551 .7211 m .65547 .72227 L p .65521 .72135 L .934 .695 .216 r
F P
0 g
s
.65547 .72227 m .6615 .74128 L p .65712 .72668 L .934 .695 .216 r
F P
0 g
s
.65547 .72227 m .64981 .70443 L p .6483 .70578 L .927 .88 .768 r
F P
0 g
s
.65649 .67601 m .65702 .67293 L .65764 .67075 L 0 0 .267 r
F
.63808 .66804 m .65765 .67107 L p .65577 .67999 L .816 .871 .887 r
F P
0 g
s
.63808 .66804 m .65577 .67999 L .64672 .69353 L .816 .871 .887 r
F
.65816 .66896 m .65764 .67075 L .66012 .65936 L 0 0 .267 r
F
.63808 .66804 m .65685 .63213 L p .65798 .66991 L .763 .892 .935 r
F P
0 g
s
.65731 .63619 m .65798 .66991 L .65685 .63213 L .763 .892 .935 r
F
.66025 .66165 m .66022 .66176 L p .65816 .66896 L 0 0 .267 r
F P
0 g
s
.66022 .66176 m .65798 .66991 L .65731 .63619 L .763 .892 .935 r
F
.65765 .67107 m .63808 .66804 L p .65798 .66991 L F P
0 g
s
.62723 .64433 m .61909 .68112 L p .64895 .64238 L .852 .941 .88 r
F P
0 g
s
.63808 .66804 m .62261 .67656 L p .63244 .66381 L .835 .999 .804 r
F P
0 g
s
.59116 .69389 m .61749 .67939 L p .60491 .70159 L .942 .938 .76 r
F P
0 g
s
.6272 .64438 m .58205 .71392 L .62722 .64435 L .912 .973 .681 r
F
.65681 .63182 m .62723 .64433 L p .64895 .64238 L .852 .941 .88 r
F P
0 g
s
.62722 .64435 m .58205 .71392 L .60276 .70538 L .912 .973 .681 r
F
.62075 .67362 m .62723 .64433 L p .60276 .70538 L F P
0 g
s
.62723 .64433 m .59644 .68186 L p .57532 .72122 L .912 .973 .681 r
F P
0 g
s
.62723 .64433 m .57532 .72122 L .57726 .7213 L .912 .973 .681 r
F
.62723 .64433 m .65006 .61652 L p .61396 .6605 L 0 0 0 r
F P
0 g
s
.67801 .59166 m .65635 .61106 L p .67186 .59759 L 0 0 0 r
F P
0 g
s
.64953 .61717 m .61396 .6605 L .65006 .61652 L 0 0 0 r
F
.62331 .64912 m .65635 .61106 L .64442 .64826 L .835 .999 .804 r
F
.6074 .6685 m .62723 .64433 L p .61396 .6605 L 0 0 0 r
F P
0 g
s
.61341 .64953 m .64953 .61717 L p .62331 .64912 L .835 .999 .804 r
F P
0 g
s
.64953 .61717 m .65635 .61106 L p .62331 .64912 L .835 .999 .804 r
F P
0 g
s
.57083 .71308 m .59644 .68186 L p .59279 .68632 L 0 0 0 r
F P
0 g
s
.61341 .64953 m .60513 .66604 L p .59486 .6838 L .685 .836 .407 r
F P
0 g
s
.55962 .74472 m .59224 .687 L .59486 .6838 L .685 .836 .407 r
F
.59279 .68632 m .59147 .68794 L .57083 .71308 L 0 0 0 r
F
.58943 .69042 m .57632 .70639 L p .59147 .68794 L F P
0 g
s
.57632 .70639 m .57083 .71308 L p .59147 .68794 L 0 0 0 r
F P
0 g
s
.57726 .7213 m .57532 .72122 L .5658 .73895 L .912 .973 .681 r
F
.55923 .75052 m .57574 .72123 L .5737 .7424 L .973 .908 .565 r
F
.57574 .72123 m .58123 .72146 L .5737 .7424 L F
.57316 .72402 m .57485 .7212 L p .57307 .72454 L F P
0 g
s
.57316 .72402 m .57307 .72454 L .55923 .75052 L .973 .908 .565 r
F
.56174 .74314 m .57316 .72402 L p .55923 .75052 L F P
0 g
s
.55359 .75444 m .57971 .71234 L .56174 .74314 L .685 .836 .407 r
F
.57485 .7212 m .56174 .74314 L p .57971 .71234 L F P
0 g
s
.57485 .7212 m .57574 .72123 L .55923 .75052 L .973 .908 .565 r
F
.55962 .74472 m .60513 .66604 L .57971 .71234 L .685 .836 .407 r
F
.57232 .72159 m .61341 .64953 L .55962 .74472 L F
.55305 .75518 m .61341 .64953 L p .55359 .75444 L F P
0 g
s
.58205 .71392 m .57726 .7213 L .59472 .71621 L .912 .973 .681 r
F
.5877 .72173 m .59472 .71621 L p .57726 .7213 L F P
0 g
s
.57175 .73195 m .59853 .67933 L .57422 .71813 L 0 0 .083 r
F
.5732 .73006 m .59853 .67933 L .57175 .73195 L F
.59147 .68794 m .57422 .71813 L p .59853 .67933 L F P
0 g
s
.60513 .66604 m .61341 .64953 L p .61749 .67939 L .835 .999 .804 r
F P
0 g
s
.58205 .71392 m .59816 .71351 L .60276 .70538 L .912 .973 .681 r
F
.59225 .68699 m .54162 .79491 L .54123 .79508 L 0 0 .028 r
F
.54162 .79491 m .54113 .79581 L .54123 .79508 L F
.52779 .84428 m .52692 .84722 L p .53122 .83179 L .935 .819 .365 r
F P
0 g
s
.53122 .83179 m .52692 .84722 L .52649 .84828 L .935 .819 .365 r
F
.53122 .83179 m .53123 .83178 L .52779 .84428 L F
.53408 .82301 m .52779 .84428 L p .53123 .83178 L F P
0 g
s
.52738 .847 m .53946 .80951 L .52678 .84939 L .997 .768 .444 r
F
.52738 .847 m .52678 .84939 L .52679 .84925 L F
.52348 .86818 m .52648 .84659 L .52859 .8402 L 0 0 .274 r
F
.52265 .87147 m .52648 .84659 L .52348 .86818 L F
.53946 .80951 m .53964 .80897 L .53947 .80951 L .997 .768 .444 r
F
.54 .80846 m .5394 .80957 L p .53997 .80814 L .914 .593 .422 r
F P
0 g
s
.5394 .80957 m .54 .80846 L p .5373 .81139 L .998 .825 .484 r
F P
0 g
s
.53474 .82012 m .5379 .80965 L .53634 .8154 L .935 .819 .365 r
F
.53311 .82494 m .53938 .80215 L .53474 .82012 L F
.53938 .80215 m .53889 .80462 L .53474 .82012 L F
.54168 .79878 m .53847 .80672 L .54593 .79249 L .998 .825 .484 r
F
.53847 .80672 m .53815 .80838 L .54593 .79249 L F
.53474 .82012 m .53889 .80462 L .5379 .80965 L .935 .819 .365 r
F
.53123 .83178 m .53311 .82494 L .53474 .82012 L F
.52733 .84373 m .5304 .83293 L .53123 .83178 L F
.5304 .83293 m .5329 .82554 L .53123 .83178 L F
.53634 .8154 m .53408 .82301 L p .53474 .82012 L F P
0 g
s
.53081 .83407 m .5376 .81112 L p .52833 .84343 L .997 .768 .444 r
F P
0 g
s
.5376 .81112 m .53779 .81097 L .52833 .84343 L .997 .768 .444 r
F
.52833 .84343 m .53481 .82117 L .52738 .847 L F
.52733 .84373 m .53123 .83178 L .53122 .83179 L .935 .819 .365 r
F
.53122 .83179 m .53123 .83178 L .53123 .83178 L F
.52864 .8428 m .54186 .79445 L .52859 .84047 L .144 0 .315 r
F
.52505 .86195 m .52859 .8402 L .52867 .84448 L 0 0 .274 r
F
.52143 .87941 m .52348 .86818 L .52505 .86195 L F
.52348 .86818 m .52859 .8402 L .52505 .86195 L F
.52187 .87469 m .52671 .8439 L .5262 .84742 L 0 0 0 r
F
.52176 .87539 m .52183 .8749 L .52187 .87469 L F
.52183 .8749 m .52671 .8439 L .52187 .87469 L F
.5262 .84742 m .52638 .84621 L .5264 .84613 L F
.52638 .84621 m .52671 .8439 L .5264 .84613 L F
.52715 .84091 m .53115 .8242 L .52811 .83483 L .986 .776 .379 r
F
.5264 .84613 m .52671 .8439 L .52862 .83218 L 0 0 0 r
F
.52671 .8439 m .52819 .83429 L .52862 .83218 L F
.52597 .84848 m .52711 .8412 L .52671 .8439 L F
.52711 .8412 m .52819 .83429 L .52671 .8439 L F
.5281 .83486 m .52597 .84848 L .52588 .84901 L F
.52422 .85381 m .52671 .8439 L p .52502 .85186 L .935 .819 .365 r
F P
0 g
s
.52542 .85023 m .52668 .8441 L .52482 .85145 L 0 0 0 r
F
.52678 .84363 m .52671 .8439 L p .52668 .8441 L F P
0 g
s
.52671 .8439 m .52482 .85145 L p .52668 .8441 L 0 0 0 r
F P
0 g
s
.52502 .85186 m .52671 .8439 L .52595 .84858 L .935 .819 .365 r
F
.52542 .85023 m .52632 .84659 L .52668 .8441 L 0 0 0 r
F
.52632 .84659 m .55097 .76093 L .53699 .80349 L F
.52715 .84093 m .52711 .8412 L .52597 .84848 L F
.52619 .84746 m .52637 .84692 L .52209 .87507 L 0 0 .274 r
F
.52206 .87521 m .52209 .87507 L .52143 .87941 L F
.52619 .84746 m .52209 .87507 L .52206 .87521 L F
.54186 .79445 m .52859 .84022 L .52887 .83953 L .144 0 .315 r
F
.52859 .84022 m .52846 .84067 L .52859 .8403 L F
.52859 .84022 m .52859 .8403 L .52887 .83953 L F
.59225 .68699 m .54123 .79508 L .55649 .75042 L 0 0 .028 r
F
.52859 .8402 m .55112 .76045 L p .52633 .84703 L 0 0 0 r
F P
0 g
s
.54113 .79581 m .52859 .8402 L p .54186 .79445 L .144 0 .315 r
F P
0 g
s
.52859 .8402 m .52846 .84067 L p .52859 .84022 L .144 0 .315 r
F P
0 g
s
.54186 .79445 m .52859 .8402 L .52859 .84022 L .144 0 .315 r
F
.52009 .8859 m .5201 .8859 L p .52001 .88481 L .914 .646 .517 r
F P
0 g
s
.5201 .88588 m .5201 .8859 L .5201 .8859 L p .943 .682 .216 r
F P
0 g
s
.5201 .8859 m .52009 .88588 L p .5201 .88588 L .943 .682 .216 r
F P
0 g
s
.52009 .8859 m .52023 .88503 L p .52024 .88506 L .915 .619 .467 r
F P
0 g
s
.52 .88486 m .52303 .85989 L p .52034 .88169 L 0 0 .181 r
F P
0 g
s
.52 .88486 m .52034 .88169 L .52011 .88357 L 0 0 .181 r
F
.51999 .88479 m .52 .88475 L .52 .88482 L .44 .23 .497 r
F
.52 .88483 m .52002 .88447 L .51999 .88467 L .212 .036 .411 r
F
.52 .88475 m .52 .88477 L .52 .88482 L .44 .23 .497 r
F
.52 .88486 m .5196 .86133 L p .51993 .88335 L .728 .45 .506 r
F P
0 g
s
.5196 .86133 m .52 .88486 L p .51988 .88182 L .809 .517 .495 r
F P
0 g
s
.52195 .86199 m .52175 .86189 L p .52 .88486 L .891 .597 .483 r
F P
0 g
s
.52 .88463 m .5201 .8838 L .52 .88486 L .212 .036 .411 r
F
.52001 .8848 m .52001 .88479 L .52001 .88479 L .914 .646 .517 r
F
.52238 .86116 m .52 .88486 L p .52011 .88363 L 0 0 .181 r
F P
0 g
s
.52303 .85989 m .52 .88486 L p .5201 .8838 L .212 .036 .411 r
F P
0 g
s
.52 .88462 m .52002 .88437 L .52001 .88436 L .809 .517 .495 r
F
.52 .88462 m .52001 .88436 L .51991 .88423 L F
.51999 .88476 m .52 .88462 L .51991 .88423 L F
.52 .88486 m .52002 .88437 L .52 .88468 L F
.52 .88447 m .52004 .88392 L .52 .88486 L .891 .597 .483 r
F
.5201 .8859 m .52 .88486 L p .52001 .88481 L .914 .646 .517 r
F P
0 g
s
.52 .88486 m .52009 .88579 L p .52008 .8857 L .943 .682 .216 r
F P
0 g
s
.52001 .88481 m .52 .88486 L .52001 .88482 L .914 .646 .517 r
F
.52009 .88575 m .52 .88486 L .52005 .88449 L .85 .596 .554 r
F
.52009 .8859 m .52 .88486 L p .52009 .88575 L F P
0 g
s
.52003 .88587 m .52009 .8859 L p .52004 .88419 L .85 .596 .554 r
F P
0 g
s
.52009 .88588 m .52001 .885 L .52008 .8857 L .943 .682 .216 r
F
.52003 .8852 m .52003 .88512 L .52003 .88587 L .85 .596 .554 r
F
.52003 .88587 m .52 .88486 L p .52006 .88547 L .773 .533 .571 r
F P
0 g
s
.52 .88481 m .52 .88478 L .52 .88486 L .44 .23 .497 r
F
.52 .88447 m .52 .88486 L .52004 .88392 L .891 .597 .483 r
F
.52 .88478 m .52001 .88473 L .52 .88486 L .44 .23 .497 r
F
.52 .88482 m .52 .88481 L .52 .88486 L F
.52 .88486 m .51706 .86111 L p .51998 .88507 L .649 .393 .518 r
F P
0 g
s
.52 .88486 m .52003 .88587 L p .52003 .88512 L .85 .596 .554 r
F P
0 g
s
.52 .88486 m .52001 .88482 L .52001 .88482 L .212 .036 .411 r
F
.52 .88486 m .52001 .88479 L .52001 .88481 L .914 .646 .517 r
F
.52 .88473 m .52002 .88463 L .52 .88477 L .44 .23 .497 r
F
.52002 .88463 m .52001 .88473 L .52 .88482 L F
.52002 .88468 m .52 .88486 L .52003 .88512 L .85 .596 .554 r
F
.52 .88486 m .52006 .8845 L .52001 .88478 L .212 .036 .411 r
F
.52 .88486 m .52001 .88478 L .52001 .88482 L F
.52001 .88481 m .52001 .88479 L .52 .88486 L .914 .646 .517 r
F
.52 .88486 m .52 .88485 L .52 .88485 L .44 .23 .497 r
F
.52 .88486 m .52 .88485 L .52 .88485 L F
.52 .88475 m .52 .88473 L .52 .88477 L F
.52 .88473 m .52 .88477 L .52 .88482 L F
.52 .88475 m .52 .88479 L .52 .88475 L F
.52 .88479 m .52 .88485 L .52 .88475 L F
.52 .88485 m .52 .88485 L .52 .88475 L F
.51986 .8858 m .51985 .88586 L .51986 .88586 L p .698 .467 .569 r
F P
0 g
s
.51986 .8858 m .51992 .88548 L .51985 .88577 L .62 .394 .553 r
F
.51989 .88582 m .51985 .88584 L p .52 .88463 L .212 .036 .411 r
F P
0 g
s
.51998 .88465 m .52 .88447 L .52 .88486 L .891 .597 .483 r
F
.51706 .86111 m .51544 .86122 L p .51998 .88508 L .649 .393 .518 r
F P
0 g
s
.5194 .88174 m .51927 .8811 L .51542 .8612 L .95 .6 .297 r
F
.51544 .86122 m .51542 .8612 L p .51927 .8811 L F P
0 g
s
.51999 .88478 m .51941 .88179 L .5194 .88174 L .95 .6 .297 r
F
.51999 .88479 m .51941 .88179 L .51999 .88478 L F
.52 .88486 m .51941 .88179 L p .51999 .88479 L F P
0 g
s
.51995 .88516 m .51978 .88374 L .52 .88486 L .728 .45 .506 r
F
.51996 .88527 m .51995 .8846 L .51978 .88374 L F
.51999 .88507 m .52 .88488 L .52 .8849 L .883 .733 .222 r
F
.52 .88493 m .52 .88486 L .51999 .88499 L .698 .467 .569 r
F
.52 .88486 m .52 .88491 L .52 .88491 L .993 .746 .397 r
F
.52 .88486 m .52 .88491 L .52 .88492 L F
.51993 .88586 m .52003 .88522 L .51999 .88499 L .698 .467 .569 r
F
.52003 .8846 m .52002 .88467 L .52003 .88512 L .85 .596 .554 r
F
.52004 .88419 m .52003 .88512 L .52003 .8852 L F
.52006 .88547 m .52 .88486 L .52017 .88373 L .773 .533 .571 r
F
.52005 .88542 m .52 .88486 L .52008 .8857 L .943 .682 .216 r
F
.52006 .88547 m .52001 .88493 L .52 .88498 L .773 .533 .571 r
F
.52003 .88522 m .52001 .88491 L .51999 .88499 L .698 .467 .569 r
F
.52003 .88522 m .52 .88486 L .52005 .88542 L .943 .682 .216 r
F
.52006 .88547 m .52017 .8838 L .52001 .88493 L .773 .533 .571 r
F
.52001 .88478 m .52006 .8845 L .52005 .88449 L .212 .036 .411 r
F
.52006 .8845 m .52422 .85897 L .52074 .88012 L F
.52 .88486 m .52422 .85897 L p .52006 .8845 L F P
0 g
s
.52422 .85897 m .52 .88486 L p .52001 .88473 L .44 .23 .497 r
F P
0 g
s
.52004 .88452 m .52005 .88449 L .52004 .88449 L .212 .036 .411 r
F
.52004 .88452 m .52004 .88449 L .52002 .88447 L F
.52001 .88478 m .52004 .88452 L .52002 .88447 L F
.52006 .8845 m .52074 .88012 L .52002 .88447 L F
.52017 .8838 m .52017 .88373 L .52002 .88475 L .773 .533 .571 r
F
.52002 .88463 m .52023 .88334 L .52001 .88473 L .44 .23 .497 r
F
.52023 .88334 m .52013 .88397 L .52001 .88473 L F
.52013 .8835 m .52303 .85989 L .5201 .88376 L .212 .036 .411 r
F
.52295 .86062 m .52251 .86213 L p .52298 .86048 L .935 .819 .365 r
F P
0 g
s
.52303 .85989 m .52238 .86116 L p .52011 .88357 L 0 0 .181 r
F P
0 g
s
.52009 .88421 m .52059 .88152 L .52422 .85897 L .44 .23 .497 r
F
.52009 .88588 m .5201 .8859 L p .52346 .86448 L .986 .776 .379 r
F P
0 g
s
.52172 .87563 m .52619 .84746 L p .52206 .87521 L 0 0 .274 r
F P
0 g
s
.52161 .87705 m .52206 .87521 L .52143 .87941 L 0 0 .274 r
F
.52172 .87563 m .52206 .87521 L .52161 .87705 L F
.52015 .88558 m .52867 .84448 L .52011 .88582 L F
.52014 .88579 m .52011 .88582 L p .52867 .84448 L F P
0 g
s
.52867 .84454 m .52014 .88579 L p .52867 .84448 L 0 0 .274 r
F P
0 g
s
.55112 .76045 m .55102 .76077 L .52633 .84704 L 0 0 0 r
F
.55112 .76045 m .52633 .84704 L .52633 .84703 L F
.52845 .83323 m .52846 .83371 L .52648 .84659 L 0 0 .274 r
F
.52637 .84692 m .52648 .84659 L .52143 .87941 L F
.52845 .83323 m .52648 .84659 L .52637 .84692 L F
.52633 .84703 m .52633 .84704 L .52617 .8476 L 0 0 0 r
F
.52588 .84901 m .52597 .84848 L .52176 .87539 L F
.52583 .84935 m .52588 .84901 L .52176 .87539 L F
.52346 .86448 m .52569 .85021 L .52343 .86463 L .986 .776 .379 r
F
.52568 .85024 m .52343 .86463 L p .52569 .85021 L F P
0 g
s
.52343 .86463 m .52568 .85024 L p .52176 .87539 L 0 0 0 r
F P
0 g
s
.52568 .85024 m .52583 .84935 L .52176 .87539 L 0 0 0 r
F
.52846 .84067 m .52582 .85002 L p .52859 .84047 L .144 0 .315 r
F P
0 g
s
.52187 .87469 m .52847 .83309 L .52191 .87439 L 0 0 0 r
F
.52847 .83309 m .52847 .8331 L .52191 .87439 L F
.52191 .87439 m .52191 .87439 L .52172 .87563 L F
.52847 .83309 m .52191 .87439 L .52191 .87439 L F
.52981 .82369 m .52458 .8217 L p .53175 .81097 L .842 .49 .388 r
F P
0 g
s
.52188 .85001 m .52275 .83324 L p .5249 .84612 L .849 .544 .472 r
F P
0 g
s
.52275 .83324 m .52337 .82124 L p .5249 .84612 L .849 .544 .472 r
F P
0 g
s
.52175 .86189 m .52253 .8512 L .52076 .84794 L .803 .476 .436 r
F
.53005 .80532 m .52743 .79919 L .53088 .76647 L .842 .49 .388 r
F
.52392 .81724 m .53088 .76647 L p .52569 .8157 L F P
0 g
s
.51545 .77304 m .53825 .77937 L p .53489 .77086 L .778 .379 .296 r
F P
0 g
s
.52175 .86189 m .5249 .84612 L p .52269 .84896 L .803 .476 .436 r
F P
0 g
s
.53005 .80532 m .52569 .8157 L .52743 .79919 L .842 .49 .388 r
F
.5321 .81012 m .53825 .77937 L .51637 .77329 L p .803 .476 .436 r
F P
0 g
s
.53489 .77086 m .54286 .76083 L .5347 .76467 L .778 .379 .296 r
F
.53825 .77937 m .5399 .77632 L p .53489 .77086 L F P
0 g
s
.53929 .74387 m .54303 .74389 L p .53555 .78439 L .703 .494 .602 r
F P
0 g
s
.53264 .78763 m .53929 .74387 L p .53555 .78439 L .703 .494 .602 r
F P
0 g
s
.55218 .75671 m .54499 .7712 L p .54359 .77686 L 0 0 0 r
F P
0 g
s
.54222 .78737 m .55218 .75671 L .54359 .77686 L 0 0 0 r
F
.53825 .77937 m .5321 .81012 L p .53253 .80975 L .889 .573 .442 r
F P
0 g
s
.53312 .809 m .53734 .79729 L p .53258 .78769 L .74 .548 .635 r
F P
0 g
s
.53734 .79729 m .54627 .77249 L p .53264 .78763 L .74 .548 .635 r
F P
0 g
s
.53258 .78769 m .53734 .79729 L .53264 .78763 L .74 .548 .635 r
F
.52742 .81419 m .52337 .82124 L .52569 .8157 L .842 .49 .388 r
F
.52184 .84993 m .5188 .84433 L .52245 .83856 L .766 .467 .48 r
F
.5192 .86018 m .51921 .8531 L .51501 .84798 L .719 .407 .45 r
F
.5196 .86133 m .51933 .85553 L p .5192 .86018 L F P
0 g
s
.51933 .85553 m .51921 .8531 L p .5192 .86018 L .719 .407 .45 r
F P
0 g
s
.51759 .86807 m .51756 .86799 L .51739 .86675 L .391 .225 .533 r
F
.51941 .88179 m .51544 .86122 L p .5194 .88174 L .95 .6 .297 r
F P
0 g
s
.51756 .86799 m .51542 .8612 L .51739 .86675 L .391 .225 .533 r
F
.51542 .8612 m .51649 .8606 L p .51739 .86675 L F P
0 g
s
.51905 .87805 m .5155 .86163 L .51759 .86807 L .391 .225 .533 r
F
.51988 .88421 m .52 .88486 L p .51988 .88421 L .95 .6 .297 r
F P
0 g
s
.51995 .8846 m .51993 .88335 L .51978 .88374 L .728 .45 .506 r
F
.51991 .88423 m .5199 .88336 L .52 .88486 L .809 .517 .495 r
F
.52 .88486 m .51542 .8612 L p .51971 .88323 L .391 .225 .533 r
F P
0 g
s
.51971 .88323 m .51542 .8612 L .5195 .88205 L .391 .225 .533 r
F
.52 .88486 m .51995 .8846 L .51995 .88482 L .728 .45 .506 r
F
.51982 .88582 m .52 .88485 L .52 .88475 L .44 .23 .497 r
F
.51983 .88586 m .51985 .88586 L p .51985 .88572 L .62 .394 .553 r
F P
0 g
s
.51984 .88585 m .51985 .88579 L p .51985 .88577 L .212 .036 .411 r
F P
0 g
s
.51984 .88585 m .51983 .88586 L p .52 .88482 L .44 .23 .497 r
F P
0 g
s
.52 .88486 m .51995 .88482 L .51996 .88527 L .728 .45 .506 r
F
.52003 .88587 m .52006 .88547 L .52 .88498 L .773 .533 .571 r
F
.52 .88491 m .52 .88491 L .52 .88486 L .993 .746 .397 r
F
.51999 .88546 m .52 .88486 L .52003 .88522 L .698 .467 .569 r
F
.52 .88486 m .52 .88502 L .51999 .88509 L .62 .394 .553 r
F
.52 .88486 m .51999 .88509 L .51998 .88515 L F
.51997 .88525 m .51998 .88514 L .52 .8849 L .883 .733 .222 r
F
.51998 .88507 m .52 .88486 L .51999 .88508 L .649 .393 .518 r
F
.51997 .88525 m .52 .88492 L .51998 .88505 L .883 .733 .222 r
F
.52 .88486 m .51998 .88507 L .51998 .88508 L .649 .393 .518 r
F
.51998 .88509 m .51999 .88498 L .52 .88486 L .993 .746 .397 r
F
.51998 .88515 m .52 .88505 L .52 .88486 L .62 .394 .553 r
F
.52 .88492 m .52 .88491 L .52 .88486 L .993 .746 .397 r
F
.52003 .88522 m .52 .88486 L .5206 .88148 L .698 .467 .569 r
F
.52003 .88522 m .5206 .88148 L .52001 .88491 L F
.5206 .88148 m .52078 .88028 L .52023 .88356 L F
.5206 .88148 m .52023 .88356 L .52001 .88491 L F
.52007 .88435 m .52047 .8822 L .52009 .88421 L .44 .23 .497 r
F
.52187 .87469 m .52191 .87439 L .52172 .87563 L 0 0 0 r
F
.56042 .76412 m .55042 .78585 L .56124 .76767 L .998 .825 .484 r
F
.5623 .76305 m .56124 .76767 L .56486 .76159 L F
.56042 .76412 m .56124 .76767 L .5623 .76305 L F
.55539 .76698 m .54593 .79249 L .55966 .76455 L F
.56222 .76341 m .59394 .7265 L .56128 .7675 L .997 .8 .518 r
F
.56222 .76341 m .56949 .75457 L .56833 .75629 L F
.5623 .76305 m .56949 .75457 L p .56222 .76341 L F P
0 g
s
.56534 .75946 m .5623 .76305 L p .5614 .76319 L .973 .908 .565 r
F P
0 g
s
.5614 .76319 m .5623 .76305 L .56085 .76388 L .973 .908 .565 r
F
.5379 .80965 m .53763 .81102 L .53634 .8154 L .935 .819 .365 r
F
.53815 .80838 m .53775 .8104 L .54593 .79249 L .998 .825 .484 r
F
.54593 .79249 m .54 .80846 L .55042 .78585 L F
.54 .80846 m .55966 .77217 L p .55042 .78585 L F P
0 g
s
.61341 .64953 m .63244 .66381 L .61749 .67939 L .835 .999 .804 r
F
.59226 .71149 m .61776 .6714 L .59545 .70124 L .391 .139 .408 r
F
.61341 .64953 m .64442 .64826 L .63244 .66381 L .835 .999 .804 r
F
.53297 .82728 m .54271 .80476 L .54223 .79376 L .144 0 .315 r
F
.52915 .8411 m .54016 .813 L .53297 .82728 L F
.54016 .813 m .54304 .80565 L .54123 .80953 L F
.54016 .813 m .54123 .80953 L .53297 .82728 L F
.54271 .80476 m .54716 .79036 L .54223 .79376 L F
.54968 .78863 m .54986 .78812 L .54223 .79376 L F
.54992 .78808 m .55109 .77723 L .54223 .79376 L F
.57175 .73195 m .55109 .77723 L p .57211 .72991 L 0 0 .028 r
F P
0 g
s
.55109 .77723 m .54054 .80037 L p .54997 .77933 L 0 0 .028 r
F P
0 g
s
.55109 .77723 m .54997 .77933 L .57211 .72991 L 0 0 .028 r
F
.5528 .7809 m .55017 .78581 L .55081 .7842 L .284 .041 .358 r
F
.5528 .7809 m .55081 .7842 L .57175 .73195 L F
.59334 .70801 m .55404 .77883 L .59545 .70124 L F
.55404 .77883 m .55017 .78581 L .5528 .7809 L F
.55404 .77883 m .5528 .7809 L .59545 .70124 L F
.52845 .83323 m .52172 .87563 L p .52847 .8331 L 0 0 0 r
F P
0 g
s
.52749 .84701 m .54223 .79376 L .54186 .79445 L .144 0 .315 r
F
.52438 .85837 m .52756 .84685 L .52749 .84701 L F
.52756 .84685 m .54223 .79376 L .52749 .84701 L F
.56128 .7675 m .57094 .75537 L .56704 .76194 L .997 .8 .518 r
F
.56704 .76194 m .58294 .7466 L .56051 .77086 L F
.56704 .76194 m .56051 .77086 L .56061 .77041 L F
.56704 .76194 m .56061 .77041 L .56095 .76895 L F
.56128 .7675 m .56704 .76194 L .56095 .76895 L F
.57808 .74443 m .56949 .75457 L p .57938 .74136 L .973 .908 .565 r
F P
0 g
s
.59369 .72602 m .59394 .7265 L p .56833 .75629 L .997 .8 .518 r
F P
0 g
s
.56949 .75457 m .59369 .72602 L p .56833 .75629 L .997 .8 .518 r
F P
0 g
s
.54024 .80858 m .53555 .81131 L p .54224 .79964 L .284 .041 .358 r
F P
0 g
s
.52749 .85577 m .52867 .84454 L .53187 .83791 L .339 .132 .446 r
F
.5239 .86759 m .52867 .84454 L p .52749 .85577 L F P
0 g
s
.54553 .80031 m .5499 .78826 L p .54548 .8004 L .144 0 .315 r
F P
0 g
s
.5499 .78826 m .54991 .7882 L .54778 .79406 L .144 0 .315 r
F
.51585 .849 m .51701 .80631 L .51505 .84722 L .719 .407 .45 r
F
.5188 .84433 m .51701 .80631 L p .51585 .849 L F P
0 g
s
.51502 .84611 m .51545 .77383 L .51333 .84308 L .719 .407 .45 r
F
.51256 .81936 m .52337 .82124 L p .52245 .83856 L .766 .467 .48 r
F P
0 g
s
.51585 .849 m .51256 .81936 L p .52245 .83856 L .766 .467 .48 r
F P
0 g
s
.51629 .80554 m .51701 .80631 L .51578 .80746 L .75 .396 .384 r
F
.52337 .82124 m .51256 .81936 L p .51578 .80746 L F P
0 g
s
.52664 .79735 m .52337 .82124 L p .51578 .80746 L .75 .396 .384 r
F P
0 g
s
.51256 .81936 m .50505 .79349 L p .51629 .80554 L .75 .396 .384 r
F P
0 g
s
.5221 .80803 m .51274 .80549 L p .51488 .82914 L .811 .574 .581 r
F P
0 g
s
.52181 .85879 m .5193 .85837 L p .52178 .84417 L .774 .561 .611 r
F P
0 g
s
.5201 .88583 m .52172 .87563 L p .52143 .87941 L 0 0 .274 r
F P
0 g
s
.5201 .88583 m .52009 .88588 L p .52176 .87539 L 0 0 0 r
F P
0 g
s
.52172 .87563 m .5201 .88583 L p .52176 .87539 L 0 0 0 r
F P
0 g
s
.52343 .86463 m .52009 .88588 L p .52346 .86448 L .986 .776 .379 r
F P
0 g
s
.52009 .88588 m .52343 .86463 L p .52176 .87539 L 0 0 0 r
F P
0 g
s
.52487 .85868 m .52 .88486 L p .52 .88505 L .62 .394 .553 r
F P
0 g
s
.51999 .88498 m .52 .88492 L .52 .88486 L .993 .746 .397 r
F
.51999 .88587 m .52003 .88587 L p .52 .88498 L .773 .533 .571 r
F P
0 g
s
.51999 .88508 m .52 .88486 L .52002 .88519 L .649 .393 .518 r
F
.51998 .88506 m .52 .88486 L .51998 .88507 L F
.5199 .88578 m .51989 .88582 L p .52 .88486 L 0 0 .181 r
F P
0 g
s
.51998 .88506 m .52 .88486 L .51998 .88506 L .649 .393 .518 r
F
.52 .88486 m .51991 .88577 L .5199 .88578 L p 0 0 .181 r
F P
0 g
s
.51998 .88505 m .52 .88492 L .51999 .88499 L .883 .733 .222 r
F
.52 .88486 m .51998 .88514 L .51995 .8854 L F
.51997 .88525 m .51999 .88499 L .51995 .8854 L F
.51996 .8851 m .51997 .88522 L .51997 .88524 L .728 .45 .506 r
F
.52 .88486 m .51991 .88573 L p .51999 .88499 L .883 .733 .222 r
F P
0 g
s
.52 .88489 m .52 .88486 L .51999 .88499 L .883 .733 .222 r
F
.51991 .88577 m .52 .88486 L p .51995 .8854 L F P
0 g
s
.51997 .8852 m .52 .88486 L .51997 .88522 L .728 .45 .506 r
F
.51996 .8851 m .52 .88486 L .51997 .8852 L F
.51995 .88516 m .51996 .8851 L .51997 .88524 L F
.51997 .88519 m .51996 .88513 L .51993 .8855 L .809 .517 .495 r
F
.52 .88486 m .51996 .88513 L .51997 .88519 L F
.52 .88486 m .51995 .88517 L .51985 .88577 L .212 .036 .411 r
F
.51985 .88577 m .51986 .88565 L .51985 .88584 L F
.51986 .88565 m .51989 .88549 L .51985 .88584 L F
.52 .88486 m .51984 .88585 L p .52 .88482 L .44 .23 .497 r
F P
0 g
s
.52 .88486 m .51999 .88479 L .51996 .88513 L .809 .517 .495 r
F
.52 .88486 m .52 .88485 L .51982 .88582 L .44 .23 .497 r
F
.51999 .88479 m .52 .88482 L .51983 .88586 L F
.51993 .8855 m .51992 .88451 L .52 .88486 L .809 .517 .495 r
F
.52 .88485 m .52 .88486 L .52 .88486 L .95 .6 .297 r
F
.51999 .88481 m .52 .88486 L .52 .88485 L F
.52 .88486 m .51999 .88479 L .52 .88484 L F
.52 .88485 m .51999 .88479 L .51999 .88478 L F
.51988 .88421 m .52 .88486 L .51999 .88481 L F
.52 .88484 m .51988 .88421 L .52 .88485 L F
.52 .88486 m .51998 .88473 L .51997 .88469 L .728 .45 .506 r
F
.51998 .88473 m .51995 .88455 L .51997 .88469 L F
.51997 .88469 m .51995 .88455 L .51995 .8846 L F
.51706 .86111 m .52 .88486 L p .51971 .88323 L F P
0 g
s
.52 .88486 m .52 .88468 L .51999 .88476 L .809 .517 .495 r
F
.51978 .88374 m .51971 .88323 L .52 .88486 L .728 .45 .506 r
F
.52 .88486 m .51999 .88476 L .51999 .88479 L .809 .517 .495 r
F
.51988 .88421 m .51941 .8818 L .51988 .88421 L .95 .6 .297 r
F
.51542 .8612 m .51941 .88179 L p .51941 .8818 L F P
0 g
s
.51941 .88179 m .51988 .88421 L p .51941 .8818 L .95 .6 .297 r
F P
0 g
s
.5195 .88205 m .5155 .86163 L .51764 .87154 L .391 .225 .533 r
F
.5195 .88205 m .51764 .87154 L .51905 .87805 L F
.51991 .88577 m .51849 .87466 L p .51677 .86805 L 0 0 .465 r
F P
0 g
s
.51991 .88577 m .51991 .88573 L p .51973 .88434 L .714 .508 0 r
F P
0 g
s
.51973 .88434 m .51851 .87481 L .51971 .88419 L .714 .508 0 r
F
.51973 .88434 m .51971 .88419 L .51991 .88577 L F
.51978 .88374 m .51993 .88335 L .51706 .86111 L .728 .45 .506 r
F
.52 .88486 m .5195 .88205 L .51987 .88391 L .391 .225 .533 r
F
.51998 .88465 m .52 .88486 L .52 .88447 L .891 .597 .483 r
F
.51999 .88479 m .51999 .88476 L .51992 .88451 L .809 .517 .495 r
F
.52 .88484 m .52 .88485 L .52 .88487 L .95 .6 .297 r
F
.51992 .88451 m .51991 .88423 L .52 .88486 L .809 .517 .495 r
F
.51997 .88478 m .52 .88486 L .51998 .88465 L .891 .597 .483 r
F
.52 .88486 m .51987 .88391 L .51996 .88436 L .391 .225 .533 r
F
.52 .88486 m .52001 .88482 L .52 .88483 L .212 .036 .411 r
F
.52 .88475 m .52 .88485 L .52003 .88456 L .44 .23 .497 r
F
.52001 .88479 m .52 .88471 L .52 .88486 L .914 .646 .517 r
F
.5199 .88336 m .51988 .88182 L .52 .88486 L .809 .517 .495 r
F
.52 .88486 m .51996 .88436 L .52 .88456 L .391 .225 .533 r
F
.51991 .88423 m .51995 .88428 L .51991 .88373 L .809 .517 .495 r
F
.52 .88453 m .51987 .88391 L .51905 .87805 L .391 .225 .533 r
F
.52 .88456 m .52 .88456 L .52 .88453 L F
.52 .88459 m .52 .88456 L .52 .88456 L F
.5196 .86133 m .51887 .86126 L p .51993 .88335 L .728 .45 .506 r
F P
0 g
s
.51887 .86126 m .51706 .86111 L p .51993 .88335 L .728 .45 .506 r
F P
0 g
s
.51991 .88373 m .51999 .88434 L .5196 .86133 L .809 .517 .495 r
F
.52 .88486 m .52 .88472 L .52001 .88479 L .914 .646 .517 r
F
.52001 .88479 m .52 .88472 L .52 .88471 L F
.52002 .8835 m .52 .88486 L .52002 .88468 L .85 .596 .554 r
F
.52005 .88449 m .52 .88486 L .52002 .8835 L F
.52002 .8835 m .52 .88283 L .52 .88486 L F
.52 .88486 m .52 .88329 L .52002 .8835 L F
.52 .88458 m .52 .88456 L .52 .88486 L .391 .225 .533 r
F
.52 .88456 m .52 .88453 L .52 .88486 L F
.52001 .88479 m .51999 .88459 L .52001 .88479 L .914 .646 .517 r
F
.52 .88453 m .51993 .88409 L .52 .88486 L .391 .225 .533 r
F
.52 .88456 m .51987 .88391 L .52 .88453 L F
.51995 .88428 m .51999 .88434 L .51991 .88373 L .809 .517 .495 r
F
.51706 .86111 m .51918 .86129 L p .51501 .84798 L .719 .407 .45 r
F P
0 g
s
.51918 .86129 m .5196 .86133 L p .5192 .86018 L .719 .407 .45 r
F P
0 g
s
.5192 .86018 m .51501 .84798 L .51918 .86129 L .719 .407 .45 r
F
.51991 .88573 m .5199 .88572 L p .51877 .87691 L .976 .694 .301 r
F P
0 g
s
.51706 .86111 m .51251 .85118 L .51111 .84953 L .64 .363 .484 r
F
.51649 .8606 m .52 .88486 L p .51905 .87805 L .391 .225 .533 r
F P
0 g
s
.51501 .84798 m .51201 .84432 L .51706 .86111 L .719 .407 .45 r
F
.49917 .8046 m .50555 .82476 L p .49973 .80353 L F P
0 g
s
.50555 .82476 m .50878 .83496 L p .50203 .81118 L .719 .407 .45 r
F P
0 g
s
.50878 .83496 m .51158 .8438 L p .50922 .83575 L .719 .407 .45 r
F P
0 g
s
.50203 .81118 m .50878 .83496 L .50922 .83575 L .719 .407 .45 r
F
.50203 .81118 m .51158 .8438 L .51201 .84432 L F
.49973 .80353 m .50555 .82476 L .50203 .81118 L F
.50298 .81931 m .4994 .81983 L p .49173 .80542 L .599 .332 .483 r
F P
0 g
s
.48642 .79551 m .50298 .81931 L .48656 .79903 L .599 .332 .483 r
F
.46836 .76956 m .48642 .79551 L .48656 .79903 L F
.48636 .79412 m .48566 .77658 L .47999 .78891 L .64 .363 .484 r
F
.49917 .8046 m .48566 .77658 L .48636 .79412 L F
.48636 .79412 m .50298 .81931 L p .48642 .79551 L .599 .332 .483 r
F P
0 g
s
.51501 .84798 m .51333 .84308 L .51201 .84432 L .719 .407 .45 r
F
.51333 .84308 m .49973 .80353 L .50966 .83653 L F
.51333 .84308 m .50966 .83653 L .51201 .84432 L F
.51256 .81936 m .50513 .81932 L p .5011 .80094 L .668 .348 .423 r
F P
0 g
s
.50513 .81932 m .50492 .81932 L p .50457 .81764 L .668 .348 .423 r
F P
0 g
s
.5011 .80094 m .50513 .81932 L .50457 .81764 L .668 .348 .423 r
F
.51158 .8438 m .50298 .81931 L .51407 .8331 L .685 .407 .496 r
F
.48573 .72697 m .50505 .79349 L p .5011 .80094 L .668 .348 .423 r
F P
0 g
s
.50505 .79349 m .51256 .81936 L p .5011 .80094 L .668 .348 .423 r
F P
0 g
s
.51407 .8331 m .50567 .82266 L .51256 .81936 L .685 .407 .496 r
F
.50298 .81931 m .50492 .81932 L p .50567 .82266 L F P
0 g
s
.50492 .81932 m .51256 .81936 L p .50567 .82266 L .685 .407 .496 r
F P
0 g
s
.51052 .83272 m .5101 .81153 L .50728 .81445 L 0 0 .465 r
F
.50976 .80935 m .50728 .81445 L p .5101 .81153 L F P
0 g
s
.512 .83668 m .51035 .8132 L .51052 .83272 L 0 0 .465 r
F
.51035 .8132 m .51004 .80877 L .5101 .81153 L F
.51035 .8132 m .5101 .81153 L .51052 .83272 L F
.51256 .81936 m .51585 .849 L p .5158 .84895 L .685 .407 .496 r
F P
0 g
s
.51585 .849 m .51505 .84722 L .51517 .84818 L .719 .407 .45 r
F
.51501 .84798 m .51502 .84611 L .51333 .84308 L F
.51517 .84818 m .51505 .84722 L .51501 .84798 L F
.5158 .84895 m .51158 .8438 L .51407 .8331 L .685 .407 .496 r
F
.51629 .85284 m .51497 .8471 L .51457 .8435 L 0 0 .426 r
F
.51497 .8471 m .51345 .84053 L .51446 .8432 L F
.51497 .8471 m .51446 .8432 L .51457 .8435 L F
.51282 .83884 m .51004 .80877 L .512 .83668 L 0 0 .465 r
F
.51374 .84128 m .51274 .83249 L .51282 .83884 L F
.51274 .83249 m .51004 .80877 L .51282 .83884 L F
.51158 .8438 m .51706 .86111 L p .51201 .84432 L .719 .407 .45 r
F P
0 g
s
.51649 .8606 m .51288 .84389 L .51584 .85383 L .213 .158 .575 r
F
.51669 .86037 m .51649 .8606 L p .51584 .85383 L F P
0 g
s
.51282 .83884 m .51751 .85943 L p .51288 .84389 L .213 .158 .575 r
F P
0 g
s
.51539 .85013 m .51533 .84987 L p .51629 .85284 L 0 0 .426 r
F P
0 g
s
.50683 .81482 m .50728 .81445 L p .50497 .81195 L .362 .274 .62 r
F P
0 g
s
.50492 .81932 m .50298 .81931 L p .49997 .8032 L .668 .348 .423 r
F P
0 g
s
.48573 .72697 m .50492 .81932 L .49997 .8032 L .668 .348 .423 r
F
.48677 .8043 m .44007 .7469 L p .48266 .79466 L .599 .332 .483 r
F P
0 g
s
.48656 .79903 m .48672 .80316 L .48266 .79466 L .599 .332 .483 r
F
.48672 .80316 m .48677 .8043 L .48266 .79466 L F
.50643 .84307 m .48898 .8079 L .50594 .8421 L .852 .404 .001 r
F
.47126 .7719 m .47279 .77499 L p .47415 .77753 L .64 .363 .484 r
F P
0 g
s
.48656 .79903 m .44007 .7469 L .44055 .74715 L .599 .332 .483 r
F
.48656 .79903 m .43725 .74343 L .44035 .74603 L F
.51982 .88586 m .50713 .84482 L p .51262 .85741 L .609 .361 .516 r
F P
0 g
s
.51111 .84953 m .47415 .77753 L .50761 .84539 L .64 .363 .484 r
F
.51111 .84953 m .50142 .82784 L .49317 .81458 L F
.50298 .81931 m .51985 .88586 L .51982 .88586 L p .609 .361 .516 r
F P
0 g
s
.51982 .88582 m .51982 .88586 L p .51982 .88583 L .62 .394 .553 r
F P
0 g
s
.51982 .88586 m .51982 .88582 L p .51982 .88582 L .44 .23 .497 r
F P
0 g
s
.51982 .88586 m .51983 .88586 L p .51985 .88572 L .62 .394 .553 r
F P
0 g
s
.51983 .88586 m .51982 .88586 L p .51982 .88582 L .44 .23 .497 r
F P
0 g
s
.51982 .88586 m .51984 .88585 L p .50717 .84492 L .463 .278 .544 r
F P
0 g
s
.50602 .84216 m .50761 .84539 L p .50599 .8421 L .64 .363 .484 r
F P
0 g
s
.50736 .84509 m .50683 .84384 L .50761 .84539 L .609 .361 .516 r
F
.50713 .84482 m .50683 .84384 L p .51262 .85741 L F P
0 g
s
.50736 .84509 m .50761 .84539 L .51153 .85313 L .609 .361 .516 r
F
.51006 .85034 m .50602 .84216 L p .50978 .8498 L .852 .404 .001 r
F P
0 g
s
.50602 .84216 m .50668 .84354 L .50978 .8498 L .852 .404 .001 r
F
.50715 .84484 m .50686 .8439 L .50713 .84482 L .463 .278 .544 r
F
.50683 .84384 m .50713 .84482 L p .50686 .8439 L F P
0 g
s
.51262 .85741 m .50736 .84509 L .51153 .85313 L .609 .361 .516 r
F
.51984 .88585 m .51118 .85285 L p .50686 .8439 L .463 .278 .544 r
F P
0 g
s
.50978 .8498 m .48692 .80358 L .48898 .8079 L .852 .404 .001 r
F
.50297 .83032 m .50287 .83004 L .5029 .8302 L .64 .363 .484 r
F
.50683 .84384 m .4994 .81983 L p .50955 .84768 L .609 .361 .516 r
F P
0 g
s
.50683 .84384 m .50955 .84768 L .50761 .84539 L .609 .361 .516 r
F
.4994 .81983 m .50683 .84384 L p .50686 .8439 L .463 .278 .544 r
F P
0 g
s
.48672 .80316 m .48692 .80358 L .50672 .84362 L .852 .404 .001 r
F
.50668 .84354 m .50672 .84362 L .50978 .8498 L F
.48672 .80316 m .50672 .84362 L .50668 .84354 L F
.47279 .77499 m .50602 .84216 L p .47415 .77753 L .64 .363 .484 r
F P
0 g
s
.47415 .77753 m .50602 .84216 L .50599 .8421 L .64 .363 .484 r
F
.50761 .84539 m .50955 .84768 L .51153 .85313 L .609 .361 .516 r
F
.51153 .85313 m .50955 .84768 L .51041 .8487 L F
.50761 .84539 m .51544 .86122 L p .51111 .84953 L .64 .363 .484 r
F P
0 g
s
.50602 .84216 m .48672 .80316 L p .50668 .84354 L .852 .404 .001 r
F P
0 g
s
.4994 .81983 m .50298 .81931 L p .51041 .8487 L .609 .361 .516 r
F P
0 g
s
.50955 .84768 m .4994 .81983 L .51041 .8487 L .609 .361 .516 r
F
.50686 .8439 m .50554 .83136 L .4994 .81983 L .463 .278 .544 r
F
.50978 .8498 m .50643 .84307 L .50959 .84945 L .852 .404 .001 r
F
.50959 .84945 m .50755 .84533 L p .50843 .84642 L .415 .269 .572 r
F P
0 g
s
.51542 .8612 m .51544 .86122 L p .50987 .84998 L .852 .404 .001 r
F P
0 g
s
.50987 .84998 m .50978 .8498 L .50987 .84999 L .852 .404 .001 r
F
.50987 .84998 m .50987 .84999 L .51542 .8612 L F
.50959 .84945 m .51001 .85029 L p .50978 .8498 L F P
0 g
s
.51001 .85029 m .51542 .8612 L p .50987 .84999 L .852 .404 .001 r
F P
0 g
s
.50978 .8498 m .51001 .85029 L .50987 .84999 L .852 .404 .001 r
F
.50728 .81445 m .50572 .81572 L p .50839 .82518 L .306 .184 .544 r
F P
0 g
s
.51043 .84873 m .50811 .83958 L p .51158 .8438 L .685 .407 .496 r
F P
0 g
s
.50811 .83958 m .50298 .81931 L p .51158 .8438 L .685 .407 .496 r
F P
0 g
s
.5124 .85204 m .51111 .84953 L .51488 .85971 L .64 .363 .484 r
F
.50755 .84533 m .50554 .83136 L .50686 .8439 L .463 .278 .544 r
F
.51544 .86122 m .51006 .85034 L p .50983 .8499 L .852 .404 .001 r
F P
0 g
s
.50987 .84998 m .50983 .8499 L .50978 .8498 L .852 .404 .001 r
F
.51544 .86122 m .50983 .8499 L .50987 .84998 L F
.51742 .87185 m .51586 .86304 L p .51246 .85759 L .306 .184 .544 r
F P
0 g
s
.51542 .8612 m .51001 .85029 L p .51 .84833 L .415 .269 .572 r
F P
0 g
s
.51542 .8612 m .51 .84833 L .51172 .85043 L .415 .269 .572 r
F
.5155 .86163 m .51542 .8612 L .51759 .86807 L .391 .225 .533 r
F
.5124 .85204 m .51488 .85971 L .51544 .86122 L .64 .363 .484 r
F
.51181 .84451 m .51158 .8438 L p .51152 .84406 L F P
0 g
s
.51291 .84632 m .50989 .82918 L p .51118 .84397 L .306 .184 .544 r
F P
0 g
s
.51649 .8606 m .51297 .85073 L p .51361 .84727 L .213 .158 .575 r
F P
0 g
s
.51297 .85073 m .50996 .84232 L p .51291 .84632 L .213 .158 .575 r
F P
0 g
s
.51297 .85073 m .51291 .84632 L .51361 .84727 L .213 .158 .575 r
F
.50989 .82918 m .50728 .81445 L p .50993 .8365 L .306 .184 .544 r
F P
0 g
s
.50993 .8365 m .50728 .81445 L .50995 .84038 L .306 .184 .544 r
F
.51531 .84973 m .51285 .83892 L .51497 .8471 L 0 0 .426 r
F
.50728 .81445 m .5105 .83268 L p .51052 .83272 L 0 0 .465 r
F P
0 g
s
.48841 .77061 m .51545 .77304 L p .51476 .77239 L .687 .306 .32 r
F P
0 g
s
.46713 .74231 m .48841 .77061 L p .51476 .77239 L .687 .306 .32 r
F P
0 g
s
.49917 .8046 m .48841 .77061 L p .48566 .77658 L .64 .363 .484 r
F P
0 g
s
.48841 .77061 m .48543 .77083 L p .48566 .77658 L .64 .363 .484 r
F P
0 g
s
.48543 .77083 m .47126 .7719 L p .48566 .77658 L .64 .363 .484 r
F P
0 g
s
.49447 .80031 m .49347 .79952 L .48225 .76477 L .415 .269 .572 r
F
.47999 .78891 m .48566 .77658 L .47126 .7719 L .64 .363 .484 r
F
.4994 .81983 m .48797 .80577 L p .48656 .79903 L .599 .332 .483 r
F P
0 g
s
.48797 .80577 m .48677 .8043 L p .48672 .80316 L .599 .332 .483 r
F P
0 g
s
.48656 .79903 m .48797 .80577 L .48672 .80316 L .599 .332 .483 r
F
.48797 .80577 m .48493 .79984 L .47826 .79063 L .48 .312 .576 r
F
.48655 .803 m .48912 .80818 L p .48669 .80328 L .852 .404 .001 r
F P
0 g
s
.48912 .80818 m .50594 .8421 L p .48901 .80796 L .852 .404 .001 r
F P
0 g
s
.48669 .80328 m .48912 .80818 L .48901 .80796 L .852 .404 .001 r
F
.48672 .80316 m .47126 .7719 L p .48692 .80358 L F P
0 g
s
.47126 .7719 m .471 .77166 L p .47136 .77215 L .852 .404 .001 r
F P
0 g
s
.471 .77166 m .47126 .7719 L p .46895 .76997 L 0 .103 .592 r
F P
0 g
s
.46895 .76997 m .43866 .74461 L .471 .77166 L 0 .103 .592 r
F
.43866 .74461 m .43945 .74526 L .471 .77166 L F
.42569 .72922 m .43725 .74343 L p .43935 .74518 L .48 .312 .576 r
F P
0 g
s
.43512 .74164 m .43222 .7392 L .44172 .74716 L 0 .103 .592 r
F
.43945 .74526 m .44172 .74716 L .471 .77166 L F
.43512 .74164 m .44172 .74716 L .43945 .74526 L F
.47929 .78835 m .47126 .7719 L .47213 .77392 L .852 .404 .001 r
F
.47126 .7719 m .44359 .74876 L p .43725 .74343 L 0 .103 .592 r
F P
0 g
s
.43866 .74461 m .43725 .74343 L .43512 .74164 L 0 .103 .592 r
F
.47126 .7719 m .43725 .74343 L .43866 .74461 L F
.46801 .76907 m .46334 .76143 L .43817 .72317 L .48 .312 .576 r
F
.46893 .77049 m .46811 .76915 L .46801 .76907 L F
.46811 .76915 m .46334 .76143 L .46801 .76907 L F
.48655 .80299 m .48655 .803 L p .48669 .80328 L .852 .404 .001 r
F P
0 g
s
.48677 .8043 m .48797 .80577 L p .48668 .80376 L .48 .312 .576 r
F P
0 g
s
.48669 .80328 m .48111 .79182 L .47929 .78835 L .852 .404 .001 r
F
.48898 .8079 m .48736 .80459 L .48669 .80328 L F
.48736 .80459 m .48111 .79182 L .48669 .80328 L F
.47826 .79063 m .48312 .7963 L .43671 .72388 L .48 .312 .576 r
F
.48692 .80358 m .48111 .79182 L .48898 .8079 L .852 .404 .001 r
F
.48111 .79182 m .47126 .7719 L .47929 .78835 L F
.48667 .80193 m .48636 .79412 L .47999 .78891 L .64 .363 .484 r
F
.48655 .803 m .471 .77166 L p .48597 .79958 L .415 .269 .572 r
F P
0 g
s
.48656 .79903 m .49173 .80542 L .4994 .81983 L .599 .332 .483 r
F
.48598 .79965 m .50219 .81858 L .48497 .79368 L .48 .312 .576 r
F
.48797 .80577 m .4994 .81983 L p .48635 .80179 L F P
0 g
s
.48655 .803 m .48635 .80179 L .48493 .79984 L .48 .312 .576 r
F
.48797 .80577 m .48635 .80179 L .48655 .803 L F
.48932 .79624 m .471 .77166 L .48406 .78244 L .415 .269 .572 r
F
.47213 .77392 m .47136 .77215 L .471 .77166 L .852 .404 .001 r
F
.4804 .78919 m .471 .77166 L .48426 .78945 L .415 .269 .572 r
F
.4848 .79267 m .48426 .78945 L .48932 .79624 L F
.4804 .78919 m .48426 .78945 L .4848 .79267 L F
.48597 .79958 m .4804 .78919 L .4848 .79267 L F
.47648 .7814 m .48497 .79368 L .47857 .78338 L .48 .312 .576 r
F
.50219 .81858 m .4848 .79267 L p .48497 .79368 L F P
0 g
s
.4848 .79267 m .47857 .78338 L p .48497 .79368 L .48 .312 .576 r
F P
0 g
s
.49995 .80312 m .50298 .81931 L .4979 .807 L .668 .348 .423 r
F
.50298 .81931 m .48636 .79412 L p .49575 .80179 L F P
0 g
s
.50219 .81858 m .50683 .81482 L p .49518 .80186 L .362 .274 .62 r
F P
0 g
s
.47826 .76721 m .48225 .76477 L p .49347 .79952 L .415 .269 .572 r
F P
0 g
s
.49212 .79845 m .49544 .80807 L .49518 .80186 L .362 .274 .62 r
F
.48632 .79387 m .49544 .80807 L .49212 .79845 L F
.49544 .80807 m .50219 .81858 L .49518 .80186 L F
.51 .84833 m .50554 .83136 L p .50755 .84533 L .463 .278 .544 r
F P
0 g
s
.49231 .7986 m .49518 .8018 L .49353 .79956 L .362 .274 .62 r
F
.49518 .8018 m .50517 .81296 L .49516 .80145 L F
.49518 .8018 m .49516 .80145 L .49353 .79956 L F
.48573 .72697 m .49146 .75768 L .47262 .74578 L .668 .348 .423 r
F
.48225 .76477 m .49288 .75132 L p .49308 .75227 L .213 .158 .575 r
F P
0 g
s
.48841 .77061 m .49917 .8046 L p .49973 .80353 L .719 .407 .45 r
F P
0 g
s
.49972 .80446 m .49514 .80084 L .50728 .81445 L .362 .274 .62 r
F
.49514 .80084 m .49353 .79956 L .49516 .80133 L F
.49514 .80084 m .49516 .80133 L .50728 .81445 L F
.51291 .84623 m .51052 .83272 L .51291 .84629 L 0 0 .465 r
F
.5105 .83268 m .51291 .84632 L p .51052 .83272 L F P
0 g
s
.51052 .83272 m .51291 .84632 L .51291 .84629 L 0 0 .465 r
F
.50989 .82918 m .50993 .8365 L .51076 .8434 L .306 .184 .544 r
F
.51172 .85043 m .51 .84833 L .50998 .84539 L .415 .269 .572 r
F
.51649 .8606 m .51542 .8612 L p .50998 .84539 L F P
0 g
s
.51 .84823 m .50839 .82518 L .50219 .81858 L .306 .184 .544 r
F
.50995 .84038 m .50839 .82518 L .51 .84823 L F
.49353 .79956 m .50517 .81296 L .50497 .81195 L .362 .274 .62 r
F
.48932 .79624 m .47826 .76721 L .49347 .79952 L .415 .269 .572 r
F
.48213 .78734 m .49212 .79845 L .48198 .78707 L .362 .274 .62 r
F
.49212 .79845 m .50683 .81482 L .49231 .7986 L F
.49212 .79845 m .49231 .7986 L .48198 .78707 L F
.48667 .79545 m .50219 .81858 L p .48213 .78734 L F P
0 g
s
.50554 .83136 m .50219 .81858 L .4994 .81983 L p .463 .278 .544 r
F P
0 g
s
.4994 .81983 m .50219 .81858 L p .48312 .7963 L .48 .312 .576 r
F P
0 g
s
.50219 .81858 m .51118 .85285 L p .51125 .85299 L .306 .184 .544 r
F P
0 g
s
.50998 .84539 m .49447 .80031 L .50996 .84232 L .415 .269 .572 r
F
.48225 .76477 m .49514 .80084 L p .49447 .80031 L F P
0 g
s
.49514 .80084 m .50996 .84232 L p .49447 .80031 L .415 .269 .572 r
F P
0 g
s
.50996 .84232 m .51649 .8606 L p .50998 .84539 L .415 .269 .572 r
F P
0 g
s
.51125 .85299 m .50995 .84038 L .51 .84823 L .306 .184 .544 r
F
.51118 .84397 m .51076 .8434 L .51214 .85493 L F
.50989 .82918 m .51076 .8434 L .51118 .84397 L F
.51214 .85493 m .50993 .8365 L .51125 .85299 L F
.50993 .8365 m .50995 .84038 L .51125 .85299 L F
.51291 .84632 m .51291 .84583 L .51361 .84727 L .213 .158 .575 r
F
.51291 .84623 m .5129 .84526 L .51252 .84403 L 0 0 .465 r
F
.51361 .84727 m .50662 .83298 L .51338 .8462 L .213 .158 .575 r
F
.51677 .86805 m .51494 .8519 L .5129 .84526 L 0 0 .465 r
F
.51677 .86805 m .5129 .84526 L .51291 .84623 L F
.49514 .80084 m .48225 .76477 L p .49502 .79819 L .213 .158 .575 r
F P
0 g
s
.471 .77166 m .47826 .76721 L p .48406 .78244 L .415 .269 .572 r
F P
0 g
s
.46902 .76392 m .49353 .79956 L .49972 .80446 L .362 .274 .62 r
F
.5129 .84526 m .51287 .84323 L .51252 .84403 L 0 0 .465 r
F
.50572 .81572 m .50219 .81858 L p .50839 .82518 L .306 .184 .544 r
F P
0 g
s
.51287 .84323 m .512 .83668 L .51252 .84403 L 0 0 .465 r
F
.51285 .83892 m .51345 .84053 L .51497 .8471 L 0 0 .426 r
F
.51524 .84948 m .51282 .83884 L p .51531 .84973 L F P
0 g
s
.51545 .77304 m .48841 .77061 L p .49973 .80353 L .719 .407 .45 r
F P
0 g
s
.51545 .77383 m .51545 .77304 L .49973 .80353 L .719 .407 .45 r
F
.4944 .78345 m .48225 .76477 L .49308 .75227 L .213 .158 .575 r
F
.49502 .79819 m .48551 .77331 L .4944 .78345 L F
.48551 .77331 m .48225 .76477 L .4944 .78345 L F
.46024 .75423 m .45732 .73634 L .48225 .76477 L .452 .336 .631 r
F
.46854 .70315 m .48573 .72697 L .4424 .7267 L p .615 .212 .255 r
F P
0 g
s
.43817 .72317 m .45984 .75179 L .45666 .73235 L .362 .274 .62 r
F
.45299 .70987 m .43843 .72294 L p .45666 .73235 L F P
0 g
s
.43843 .72294 m .43817 .72317 L p .45666 .73235 L .362 .274 .62 r
F P
0 g
s
.47064 .72688 m .48573 .72697 L p .47262 .74578 L .668 .348 .423 r
F P
0 g
s
.47262 .74578 m .46713 .74231 L .47064 .72688 L .668 .348 .423 r
F
.4813 .76178 m .45826 .73317 L .47494 .74178 L .314 .297 .679 r
F
.45666 .73235 m .48225 .76477 L p .45732 .73634 L .452 .336 .631 r
F P
0 g
s
.48225 .76477 m .47771 .75902 L p .4813 .76178 L .314 .297 .679 r
F P
0 g
s
.46088 .70279 m .45299 .70987 L p .4673 .73784 L .362 .274 .62 r
F P
0 g
s
.47771 .75902 m .45666 .73235 L p .4813 .76178 L .314 .297 .679 r
F P
0 g
s
.45666 .73235 m .45826 .73317 L .4813 .76178 L .314 .297 .679 r
F
.4673 .73784 m .47715 .74292 L .46088 .70279 L .362 .274 .62 r
F
.49288 .75132 m .48225 .76477 L p .47883 .75401 L .314 .297 .679 r
F P
0 g
s
.47883 .75401 m .47767 .74324 L .49288 .75132 L .314 .297 .679 r
F
.50984 .72548 m .50677 .69941 L .45212 .69426 L p .687 .306 .32 r
F P
0 g
s
.48573 .72697 m .46854 .70315 L p .49446 .71716 L .683 .221 .149 r
F P
0 g
s
.50173 .68694 m .50081 .68466 L p .47717 .69056 L .524 0 0 r
F P
0 g
s
.46117 .69293 m .44802 .6747 L .5079 .68602 L p .683 .221 .149 r
F P
0 g
s
.46088 .70279 m .47708 .73772 L .47398 .70871 L .107 .137 .609 r
F
.47082 .67917 m .46348 .6966 L p .47398 .70871 L F P
0 g
s
.46348 .6966 m .46088 .70279 L p .47398 .70871 L .107 .137 .609 r
F P
0 g
s
.51542 .77281 m .51228 .74615 L p .51476 .77239 L .687 .306 .32 r
F P
0 g
s
.51228 .74615 m .50984 .72548 L p .51203 .74635 L .687 .306 .32 r
F P
0 g
s
.51228 .74615 m .51203 .74635 L .51476 .77239 L .687 .306 .32 r
F
.48573 .72697 m .52642 .73445 L p .49784 .75809 L .75 .396 .384 r
F P
0 g
s
.49288 .75132 m .47398 .70871 L p .4837 .71993 L 0 .106 .629 r
F P
0 g
s
.47767 .74319 m .46088 .70279 L p .47715 .74292 L .362 .274 .62 r
F P
0 g
s
.42883 .67076 m .49786 .67735 L p .44703 .67415 L .347 .765 .978 r
F P
0 g
s
.50081 .68466 m .49786 .67735 L .42883 .67076 L p .524 0 0 r
F P
0 g
s
.50125 .63004 m .50224 .62115 L p .48201 .636 L .731 .615 .729 r
F P
0 g
s
.50224 .62115 m .50125 .63004 L p .52786 .60496 L .778 .658 .73 r
F P
0 g
s
.42879 .65251 m .44945 .61634 L .50824 .62787 L p .743 .664 .773 r
F P
0 g
s
.44022 .63261 m .4592 .67538 L p .4532 .67343 L .314 .297 .679 r
F P
0 g
s
.45675 .66326 m .45837 .67129 L .47297 .67405 L .276 .334 .742 r
F
.42844 .67069 m .44802 .6747 L p .40786 .67007 L .246 0 0 r
F P
0 g
s
.4306 .66606 m .46088 .70279 L .46348 .6966 L p .276 .334 .742 r
F P
0 g
s
.46348 .6966 m .47082 .67917 L p .4592 .67538 L .276 .334 .742 r
F P
0 g
s
.4306 .66606 m .46348 .6966 L .4592 .67538 L .276 .334 .742 r
F
.47398 .70871 m .49272 .75096 L p .47767 .74319 L .314 .297 .679 r
F P
0 g
s
.49272 .75096 m .49288 .75132 L p .47767 .74324 L .314 .297 .679 r
F P
0 g
s
.47494 .74178 m .47767 .74319 L .47767 .74324 L .314 .297 .679 r
F
.47767 .74319 m .49272 .75096 L .47767 .74324 L F
.50223 .79239 m .49308 .75221 L p .50223 .79234 L 0 0 .426 r
F P
0 g
s
.5038 .80606 m .47767 .74319 L p .50482 .8112 L .362 .274 .62 r
F P
0 g
s
.47883 .75401 m .47494 .74178 L .47767 .74324 L .314 .297 .679 r
F
.50482 .8112 m .47767 .74319 L .47715 .74292 L .362 .274 .62 r
F
.4673 .73784 m .50482 .8112 L .47715 .74292 L F
.45984 .75179 m .50482 .8112 L .45299 .70987 L F
.51252 .84403 m .512 .83668 L .51052 .83272 L 0 0 .465 r
F
.51288 .84389 m .49502 .79819 L .4944 .78345 L .213 .158 .575 r
F
.50159 .79166 m .4944 .78345 L .49308 .75227 L F
.50223 .79239 m .50728 .81445 L .50269 .79291 L .107 .137 .609 r
F
.49564 .78487 m .50372 .8054 L .50223 .79239 L F
.50372 .8054 m .50728 .81445 L .50223 .79239 L F
.50728 .81445 m .50976 .80935 L p .50269 .79291 L F P
0 g
s
.50976 .80935 m .51004 .80877 L p .50931 .80718 L .107 .137 .609 r
F P
0 g
s
.50269 .79291 m .50976 .80935 L .50931 .80718 L .107 .137 .609 r
F
.50269 .79291 m .50931 .80718 L .50705 .79789 L F
.51004 .80877 m .50836 .80264 L p .50931 .80718 L F P
0 g
s
.50836 .80264 m .50705 .79789 L p .50931 .80718 L .107 .137 .609 r
F P
0 g
s
.50223 .79234 m .49308 .75221 L .50212 .79142 L 0 0 .426 r
F
.4944 .78345 m .50728 .81445 L p .49564 .78487 L .107 .137 .609 r
F P
0 g
s
.50728 .81445 m .5038 .80606 L p .50482 .8112 L .362 .274 .62 r
F P
0 g
s
.51282 .83884 m .512 .83668 L .51287 .84323 L 0 0 .465 r
F
.50517 .81296 m .50683 .81482 L .50497 .81195 L .362 .274 .62 r
F
.51288 .84389 m .50159 .79166 L .51282 .83884 L .213 .158 .575 r
F
.50159 .79166 m .49308 .75227 L .50223 .79238 L F
.50159 .79166 m .50223 .79238 L .51282 .83884 L F
.47767 .74319 m .4944 .78345 L p .47708 .73772 L .107 .137 .609 r
F P
0 g
s
.4944 .78345 m .49564 .78487 L .47708 .73772 L .107 .137 .609 r
F
.49288 .75132 m .49308 .75221 L p .49308 .75227 L .213 .158 .575 r
F P
0 g
s
.49308 .75221 m .51282 .83884 L p .49308 .75227 L .213 .158 .575 r
F P
0 g
s
.49308 .75221 m .49288 .75132 L p .49306 .75213 L 0 0 .426 r
F P
0 g
s
.51453 .80365 m .51545 .77383 L .50505 .79349 L .719 .407 .45 r
F
.50705 .79789 m .49988 .77184 L p .50269 .79291 L .107 .137 .609 r
F P
0 g
s
.49308 .75221 m .49306 .75213 L .50212 .7914 L 0 0 .426 r
F
.49308 .75221 m .50212 .7914 L .50212 .79142 L F
.50269 .79291 m .49988 .77184 L .50196 .79001 L .107 .137 .609 r
F
.49752 .7512 m .49314 .75131 L .49491 .73757 L 0 0 .426 r
F
.49596 .74276 m .49491 .73757 L .49495 .73728 L F
.49752 .7512 m .49491 .73757 L .49596 .74276 L F
.49296 .75076 m .49495 .73728 L p .49314 .75131 L F P
0 g
s
.50284 .78083 m .50434 .7862 L .50357 .78307 L 0 0 .316 r
F
.51446 .8432 m .51004 .80877 L p .51374 .84128 L 0 0 .465 r
F P
0 g
s
.51004 .80877 m .50976 .80935 L p .5101 .81153 L 0 0 .465 r
F P
0 g
s
.50216 .79178 m .50223 .79239 L .50269 .79291 L .107 .137 .609 r
F
.50216 .79178 m .50269 .79291 L .50196 .79001 L F
.51004 .80877 m .50943 .80442 L p .50434 .7862 L 0 0 .316 r
F P
0 g
s
.50284 .78083 m .50645 .79483 L .50434 .7862 L 0 0 .316 r
F
.50645 .79483 m .51004 .80877 L .50434 .7862 L F
.51445 .84313 m .51252 .82819 L .51268 .82935 L .714 .508 0 r
F
.49988 .77184 m .50836 .80264 L p .50284 .78083 L 0 0 .316 r
F P
0 g
s
.50836 .80264 m .51004 .80877 L p .50645 .79483 L 0 0 .316 r
F P
0 g
s
.50836 .80264 m .50645 .79483 L .50284 .78083 L 0 0 .316 r
F
.51394 .83929 m .51288 .83098 L .51445 .84313 L .714 .508 0 r
F
.51288 .83098 m .5098 .80706 L .51252 .82819 L F
.51288 .83098 m .51252 .82819 L .51445 .84313 L F
.51346 .83558 m .5132 .83351 L .51394 .83929 L F
.51394 .83929 m .5132 .83351 L .51288 .83098 L F
.51445 .84313 m .5119 .82322 L .51446 .8432 L F
.51004 .80877 m .51189 .82317 L p .5119 .82322 L F P
0 g
s
.51189 .82317 m .51446 .8432 L p .5119 .82322 L .714 .508 0 r
F P
0 g
s
.51432 .84216 m .51352 .83781 L .51346 .83558 L 0 0 .426 r
F
.50943 .80442 m .50885 .80348 L p .50608 .79278 L .952 .641 .212 r
F P
0 g
s
.51352 .83781 m .49596 .74276 L .51111 .82312 L 0 0 .426 r
F
.49625 .74416 m .51111 .82312 L p .49596 .74276 L F P
0 g
s
.51111 .82312 m .51346 .83558 L p .51352 .83781 L 0 0 .426 r
F P
0 g
s
.5032 .78193 m .49988 .77184 L .49752 .7512 L 0 0 .426 r
F
.51352 .83781 m .49988 .77184 L .5032 .78193 L F
.51457 .8435 m .51397 .8406 L .51352 .83781 L F
.51397 .8406 m .49988 .77184 L .51352 .83781 L F
.50943 .80442 m .50616 .79188 L p .50357 .78307 L 0 0 .316 r
F P
0 g
s
.50616 .79188 m .50499 .78739 L p .50482 .78687 L 0 0 .316 r
F P
0 g
s
.50357 .78307 m .50616 .79188 L .50482 .78687 L 0 0 .316 r
F
.50434 .7862 m .50943 .80442 L .50357 .78307 L F
.50499 .78739 m .50943 .80442 L p .50608 .79278 L .952 .641 .212 r
F P
0 g
s
.50885 .80348 m .51274 .80549 L p .50392 .77509 L .884 .656 .588 r
F P
0 g
s
.50369 .7766 m .50885 .80348 L .50392 .77509 L .884 .656 .588 r
F
.51476 .77239 m .51203 .74635 L .49875 .75734 L .687 .306 .32 r
F
.49496 .73201 m .50957 .73788 L p .50536 .76557 L .847 .618 .591 r
F P
0 g
s
.50339 .77858 m .4986 .75211 L p .50536 .76557 L .847 .618 .591 r
F P
0 g
s
.4986 .75211 m .49496 .73201 L p .50536 .76557 L .847 .618 .591 r
F P
0 g
s
.51397 .83963 m .51117 .82575 L .51363 .83849 L .901 .613 .111 r
F
.51397 .83963 m .51363 .83849 L .51405 .84026 L F
.50194 .78986 m .49288 .75132 L .49752 .7512 L 0 0 .426 r
F
.50284 .78083 m .50357 .78307 L .49588 .75246 L 0 0 .316 r
F
.49918 .76513 m .49588 .75246 L p .50357 .78307 L F P
0 g
s
.50677 .69941 m .50173 .68694 L p .47717 .69056 L .524 0 0 r
F P
0 g
s
.47297 .67405 m .48648 .72314 L p .48534 .71826 L 0 0 .316 r
F P
0 g
s
.47297 .67405 m .48534 .71826 L .47727 .68417 L 0 0 .316 r
F
.47165 .66535 m .47297 .67405 L p .47727 .68417 L F P
0 g
s
.49495 .73728 m .49296 .75076 L p .48648 .72314 L 0 .106 .629 r
F P
0 g
s
.47761 .68528 m .49258 .73018 L p .48648 .72314 L 0 .106 .629 r
F P
0 g
s
.49258 .73018 m .49495 .73728 L p .48648 .72314 L 0 .106 .629 r
F P
0 g
s
.48648 .72314 m .47297 .67405 L p .48474 .72113 L .107 .137 .609 r
F P
0 g
s
.47727 .68417 m .47111 .65757 L .46954 .65147 L 0 0 .316 r
F
.48534 .71826 m .47761 .68528 L .47727 .68417 L F
.47761 .68528 m .47111 .65757 L .47727 .68417 L F
.50284 .78083 m .49588 .75246 L .48459 .70994 L F
.49186 .73001 m .49495 .73728 L p .49269 .73073 L .781 .455 0 r
F P
0 g
s
.49269 .73073 m .4815 .69832 L .4857 .71117 L .781 .455 0 r
F
.49269 .73073 m .4857 .71117 L .49186 .73001 L F
.49269 .73073 m .47761 .68528 L .4815 .69832 L F
.49384 .73393 m .47761 .68528 L p .49269 .73073 L F P
0 g
s
.47973 .69445 m .47761 .68528 L .47981 .69469 L 0 .106 .629 r
F
.47973 .69445 m .47981 .69469 L .49296 .75076 L F
.49296 .75076 m .49288 .75132 L p .47973 .69445 L F P
0 g
s
.48482 .70941 m .49186 .73001 L p .4815 .69832 L .781 .455 0 r
F P
0 g
s
.47981 .69469 m .48286 .70321 L .47111 .65757 L 0 0 .316 r
F
.48534 .71826 m .48286 .70321 L .47981 .69469 L F
.48534 .71826 m .50284 .78083 L .48286 .70321 L F
.47297 .67405 m .47082 .67917 L p .48143 .71731 L .107 .137 .609 r
F P
0 g
s
.49288 .75132 m .4837 .71993 L .48632 .72295 L 0 .106 .629 r
F
.48143 .71731 m .50216 .79178 L .48461 .72098 L .107 .137 .609 r
F
.48143 .71731 m .48461 .72098 L .47297 .67405 L F
.4592 .67538 m .44022 .63261 L p .45761 .66753 L 0 .106 .629 r
F P
0 g
s
.46088 .70279 m .47767 .74319 L p .47708 .73772 L .107 .137 .609 r
F P
0 g
s
.45299 .70987 m .46088 .70279 L p .43937 .70702 L .432 .384 .704 r
F P
0 g
s
.43632 .70762 m .45299 .70987 L .43937 .70702 L .432 .384 .704 r
F
.43542 .70106 m .43937 .70702 L .46088 .70279 L F
.43615 .70635 m .41095 .67442 L p .43604 .70558 L .314 .297 .679 r
F P
0 g
s
.41095 .67442 m .43615 .70635 L p .42986 .70503 L .452 .336 .631 r
F P
0 g
s
.42986 .70503 m .40332 .69948 L .41095 .67442 L .452 .336 .631 r
F
.39967 .69895 m .39911 .6986 L p .39921 .69884 L .504 .363 .621 r
F P
0 g
s
.39375 .68593 m .41095 .67442 L p .40171 .70477 L .452 .336 .631 r
F P
0 g
s
.39911 .6986 m .43615 .70635 L .43542 .70106 L .432 .384 .704 r
F
.43615 .70635 m .43937 .70702 L .43542 .70106 L F
.45666 .73235 m .43615 .70635 L p .45826 .73317 L .314 .297 .679 r
F P
0 g
s
.43615 .70635 m .43604 .70558 L .45826 .73317 L .314 .297 .679 r
F
.40066 .67293 m .41438 .67179 L .36907 .66505 L .524 0 0 r
F
.42791 .67067 m .36907 .66505 L p .41438 .67179 L F P
0 g
s
.42069 .66668 m .42879 .65251 L p .42955 .65241 L .743 .664 .773 r
F P
0 g
s
.44554 .66085 m .42069 .66668 L .42955 .65241 L .743 .664 .773 r
F
.42901 .65212 m .42848 .65223 L .42844 .6522 L .707 .617 .755 r
F
.40017 .61601 m .44945 .61634 L p .31531 .6496 L F P
0 g
s
.39116 .61958 m .36998 .64658 L .42848 .65223 L p .731 .615 .729 r
F P
0 g
s
.44945 .61634 m .42901 .65212 L p .42844 .6522 L .707 .617 .755 r
F P
0 g
s
.44945 .61634 m .42844 .6522 L .40769 .63404 L .707 .617 .755 r
F
.41095 .67442 m .42834 .64959 L p .43515 .69914 L .314 .297 .679 r
F P
0 g
s
.38495 .69174 m .35622 .6741 L .39783 .67437 L p .615 .212 .255 r
F P
0 g
s
.37049 .67299 m .37581 .67326 L .35622 .6741 L .246 0 0 r
F
.36998 .65249 m .41095 .67442 L .39375 .68593 L p .503 .43 .707 r
F P
0 g
s
.35622 .6741 m .38495 .69174 L p .37702 .69061 L .576 .273 .418 r
F P
0 g
s
.39375 .68593 m .3806 .69473 L p .37828 .6855 L .503 .43 .707 r
F P
0 g
s
.37828 .6855 m .36998 .65249 L .39375 .68593 L .503 .43 .707 r
F
.32911 .682 m .33921 .68442 L .34799 .66645 L .504 .363 .621 r
F
.33921 .68442 m .3781 .6855 L .34799 .66645 L F
.34799 .66645 m .37828 .6855 L p .37573 .67538 L .432 .384 .704 r
F P
0 g
s
.40171 .70477 m .3806 .69473 L .39911 .6986 L .452 .336 .631 r
F
.39911 .6986 m .37828 .6855 L p .39921 .69884 L .504 .363 .621 r
F P
0 g
s
.37828 .6855 m .34799 .66645 L p .3781 .6855 L .504 .363 .621 r
F P
0 g
s
.37752 .68548 m .37828 .6855 L .3806 .69473 L .503 .43 .707 r
F
.39921 .69884 m .37828 .6855 L .3781 .6855 L .504 .363 .621 r
F
.3806 .69473 m .37976 .69529 L p .37752 .68548 L .503 .43 .707 r
F P
0 g
s
.31528 .65107 m .36998 .64658 L p .34284 .6387 L .682 .543 .69 r
F P
0 g
s
.36998 .64658 m .39116 .61958 L p .3743 .62625 L .682 .543 .69 r
F P
0 g
s
.36998 .64658 m .3743 .62625 L .34284 .6387 L .682 .543 .69 r
F
.41095 .67442 m .36998 .65249 L p .40522 .65074 L .431 .452 .781 r
F P
0 g
s
.41747 .65013 m .39382 .62144 L .36178 .6529 L p .432 .384 .704 r
F P
0 g
s
.44945 .61634 m .40402 .63083 L .40115 .62832 L .707 .617 .755 r
F
.44945 .61634 m .40769 .63404 L .40402 .63083 L F
.39382 .62144 m .41747 .65013 L p .41564 .63829 L .276 .334 .742 r
F P
0 g
s
.39116 .61958 m .39856 .61014 L p .3743 .62625 L .682 .543 .69 r
F P
0 g
s
.39856 .61014 m .39116 .61958 L p .44028 .58932 L .731 .615 .729 r
F P
0 g
s
.4635 .57773 m .44945 .61634 L .38698 .61592 L p .709 .559 .683 r
F P
0 g
s
.43318 .566 m .39856 .61014 L p .44028 .58932 L .731 .615 .729 r
F P
0 g
s
.40864 .59303 m .44022 .63261 L .41937 .6624 L p .431 .452 .781 r
F P
0 g
s
.4532 .67343 m .4375 .66831 L .44022 .63261 L .314 .297 .679 r
F
.44022 .63261 m .44225 .61326 L p .45761 .66753 L 0 .106 .629 r
F P
0 g
s
.44225 .61326 m .44375 .59892 L p .44829 .62142 L 0 .106 .629 r
F P
0 g
s
.45761 .66753 m .44225 .61326 L .44829 .62142 L 0 .106 .629 r
F
.47082 .67917 m .47297 .67405 L p .45837 .67129 L .276 .334 .742 r
F P
0 g
s
.46088 .70279 m .4306 .66606 L p .43449 .6943 L .432 .384 .704 r
F P
0 g
s
.37573 .67538 m .39911 .6986 L .46088 .70279 L .432 .384 .704 r
F
.42834 .64959 m .44022 .63261 L p .43515 .69914 L .314 .297 .679 r
F P
0 g
s
.41523 .63567 m .45837 .67129 L .45675 .66326 L .276 .334 .742 r
F
.39382 .62144 m .41564 .63829 L .41523 .63567 L F
.41564 .63829 m .45837 .67129 L .41523 .63567 L F
.43228 .52189 m .44372 .5043 L p .3733 .58507 L .81 1 .83 r
F P
0 g
s
.39631 .56216 m .37137 .5875 L p .36189 .58904 L .735 .986 .896 r
F P
0 g
s
.44372 .5043 m .40499 .52946 L p .35877 .58359 L .81 1 .83 r
F P
0 g
s
.35909 .58307 m .35877 .58359 L p .40499 .52946 L .81 1 .83 r
F P
0 g
s
.35872 .5835 m .39631 .56216 L p .36189 .58904 L .735 .986 .896 r
F P
0 g
s
.39631 .56216 m .35912 .58328 L p .35918 .58297 L .826 .93 .901 r
F P
0 g
s
.39631 .56216 m .35918 .58297 L .39087 .54595 L .826 .93 .901 r
F
.31531 .6496 m .35932 .61573 L .40017 .61601 L p .707 .617 .755 r
F P
0 g
s
.38698 .61592 m .35932 .61573 L p .39638 .60221 L .709 .559 .683 r
F P
0 g
s
.39856 .61014 m .38698 .61592 L .39638 .60221 L .709 .559 .683 r
F
.39256 .58831 m .39638 .60221 L .35932 .61573 L F
.35932 .61573 m .38991 .57868 L p .39256 .58831 L F P
0 g
s
.38367 .56175 m .40276 .58566 L p .39201 .58956 L .431 .452 .781 r
F P
0 g
s
.40428 .5923 m .40725 .58403 L p .40276 .58566 L .454 .533 .849 r
F P
0 g
s
.39087 .54595 m .37871 .50975 L .39372 .54262 L .826 .93 .901 r
F
.37871 .50975 m .37939 .50875 L .39372 .54262 L F
.40662 .53254 m .39631 .56216 L p .38759 .49651 L F P
0 g
s
.38763 .4966 m .38455 .54273 L p .39459 .50978 L .905 .943 .821 r
F P
0 g
s
.38455 .54273 m .41119 .52543 L p .44835 .45756 L .905 .943 .821 r
F P
0 g
s
.39631 .56216 m .39087 .54595 L .39372 .54262 L .826 .93 .901 r
F
.36356 .45133 m .35903 .53876 L p .36744 .51748 L .786 .858 .904 r
F P
0 g
s
.36175 .44198 m .36311 .43807 L .35761 .42698 L .854 .843 .833 r
F
.36311 .43807 m .36949 .41973 L .35844 .40919 L F
.36311 .43807 m .35844 .40919 L .35066 .40177 L F
.35761 .42698 m .35066 .40177 L .34015 .39176 L F
.36311 .43807 m .35066 .40177 L .35761 .42698 L F
.36701 .43285 m .37113 .41502 L .36311 .43807 L F
.3633 .43928 m .36311 .43807 L .36175 .44198 L F
.36701 .43285 m .36311 .43807 L .3633 .43928 L F
.36744 .51748 m .37543 .49727 L .37461 .49347 L .786 .858 .904 r
F
.36744 .51748 m .37461 .49347 L .36356 .45133 L F
.37461 .49347 m .37872 .4797 L .36679 .45738 L F
.37461 .49347 m .36679 .45738 L .36356 .45133 L F
.3877 .49651 m .36335 .4392 L .36559 .43532 L .879 .885 .834 r
F
.36378 .53176 m .39006 .46024 L .36744 .51748 L .786 .858 .904 r
F
.36271 .53334 m .36744 .51748 L .35903 .53876 L F
.36378 .53176 m .36744 .51748 L .36271 .53334 L F
.35877 .58359 m .35909 .58307 L p .3439 .58862 L .753 .889 .94 r
F P
0 g
s
.34378 .59199 m .35872 .5835 L p .36189 .58904 L .735 .986 .896 r
F P
0 g
s
.34399 .58852 m .35909 .58307 L .35771 .57303 L .753 .889 .94 r
F
.3439 .58862 m .34396 .58859 L .34399 .58852 L F
.34396 .58859 m .35909 .58307 L .34399 .58852 L F
.35536 .5561 m .36383 .52987 L p .35912 .58328 L .826 .93 .901 r
F P
0 g
s
.36383 .52987 m .37443 .49701 L p .37225 .50928 L .826 .93 .901 r
F P
0 g
s
.35912 .58328 m .36383 .52987 L .37225 .50928 L .826 .93 .901 r
F
.36271 .53334 m .35536 .5561 L p .35904 .54562 L 0 0 .179 r
F P
0 g
s
.38455 .54273 m .35904 .54562 L p .34399 .58852 L .753 .889 .94 r
F P
0 g
s
.34378 .59199 m .35544 .55667 L p .35912 .58328 L .826 .93 .901 r
F P
0 g
s
.30477 .63988 m .30109 .6424 L p .34967 .60456 L .735 .986 .896 r
F P
0 g
s
.30109 .6424 m .30093 .64129 L .34967 .60456 L .735 .986 .896 r
F
.30539 .63938 m .3067 .63835 L .30353 .64101 L 0 0 0 r
F
.30029 .63681 m .3141 .62258 L p .30093 .64129 L .735 .986 .896 r
F P
0 g
s
.30093 .64129 m .3141 .62258 L .31437 .62232 L .735 .986 .896 r
F
.30093 .64129 m .31442 .62224 L .32705 .61124 L F
.33817 .61623 m .30353 .64101 L p .3071 .63804 L 0 0 0 r
F P
0 g
s
.34772 .6061 m .33817 .61623 L p .3071 .63804 L 0 0 0 r
F P
0 g
s
.304 .64068 m .33817 .61623 L p .28134 .6532 L .673 .96 .941 r
F P
0 g
s
.27537 .65708 m .27975 .65457 L .28134 .6532 L .673 .96 .941 r
F
.27975 .65457 m .304 .64068 L .28134 .6532 L F
.3141 .62258 m .31442 .62224 L p .31437 .62232 L .735 .986 .896 r
F P
0 g
s
.31442 .62224 m .34378 .59199 L p .32705 .61124 L .735 .986 .896 r
F P
0 g
s
.34533 .5953 m .33997 .59998 L .34378 .59199 L .735 .986 .896 r
F
.33997 .59998 m .32705 .61124 L .34378 .59199 L F
.34378 .59199 m .34364 .59444 L .34578 .5878 L p 0 0 .179 r
F P
0 g
s
.34578 .5878 m .35455 .56068 L p .34378 .59199 L 0 0 .179 r
F P
0 g
s
.31442 .62224 m .30029 .63681 L p .30983 .62749 L .185 .032 .427 r
F P
0 g
s
.30724 .63043 m .35877 .58359 L .31442 .62224 L .753 .889 .94 r
F
.33817 .61623 m .35877 .58359 L p .30724 .63043 L F P
0 g
s
.30628 .63153 m .30694 .63078 L .30029 .63681 L .185 .032 .427 r
F
.30554 .63227 m .30594 .63192 L .30628 .63153 L F
.30522 .63227 m .30594 .63192 L p .30724 .63043 L .753 .889 .94 r
F P
0 g
s
.30594 .63192 m .33817 .61623 L p .30724 .63043 L .753 .889 .94 r
F P
0 g
s
.30554 .63227 m .30628 .63153 L .30029 .63681 L .185 .032 .427 r
F
.30628 .63153 m .30807 .62975 L .30694 .63078 L F
.30807 .62975 m .34364 .59444 L .31066 .62667 L F
.30807 .62975 m .31066 .62667 L .30029 .63681 L F
.27313 .65894 m .304 .64068 L .27537 .65708 L .673 .96 .941 r
F
.28523 .64984 m .28134 .6532 L .30082 .64053 L F
.28523 .64984 m .30082 .64053 L .30074 .63999 L F
.33817 .61623 m .30522 .63227 L .30016 .63591 L p F P
0 g
s
.30016 .63591 m .2654 .6609 L p .27961 .65218 L .673 .96 .941 r
F P
0 g
s
.27961 .65218 m .33817 .61623 L .30016 .63591 L .673 .96 .941 r
F
.30109 .6424 m .28241 .65523 L p .30093 .64129 L .735 .986 .896 r
F P
0 g
s
.26978 .65919 m .33817 .61623 L .2654 .6609 L .673 .96 .941 r
F
.25778 .66726 m .33817 .61623 L .25401 .6691 L F
.29045 .64918 m .30057 .63877 L p .30093 .64129 L .735 .986 .896 r
F P
0 g
s
.28824 .65196 m .28959 .65318 L p .29321 .65001 L .798 .77 .828 r
F P
0 g
s
.28959 .65318 m .29332 .64997 L p .29321 .65001 L .798 .77 .828 r
F P
0 g
s
.28223 .65764 m .29995 .65189 L .30008 .64149 L .521 .524 .799 r
F
.30074 .63999 m .30082 .64053 L .33817 .61623 L .673 .96 .941 r
F
.30006 .64298 m .30012 .63923 L .30423 .63433 L .576 .571 .81 r
F
.30522 .63227 m .30724 .63043 L .31304 .62345 L .753 .889 .94 r
F
.30724 .63043 m .31442 .62224 L .31304 .62345 L F
.30983 .62749 m .34364 .59444 L .31442 .62224 L .185 .032 .427 r
F
.3079 .62793 m .30522 .63227 L p .31304 .62345 L .753 .889 .94 r
F P
0 g
s
.30628 .63153 m .31336 .6255 L .34364 .59444 L .185 .032 .427 r
F
.31426 .62467 m .34364 .59444 L p .31336 .6255 L F P
0 g
s
.30522 .63227 m .3079 .62793 L p .31426 .62467 L .554 .615 .868 r
F P
0 g
s
.31454 .62453 m .31322 .62562 L .31426 .62467 L .554 .615 .868 r
F
.31322 .62562 m .30522 .63227 L .31426 .62467 L F
.31393 .62501 m .31426 .62467 L p .31336 .6255 L .185 .032 .427 r
F P
0 g
s
.34895 .59729 m .34364 .59444 L .31426 .62467 L p .521 .524 .799 r
F P
0 g
s
.31426 .62467 m .31393 .62501 L p .31454 .62453 L .521 .524 .799 r
F P
0 g
s
.31454 .62453 m .34895 .59729 L .31426 .62467 L .521 .524 .799 r
F
.28241 .65523 m .30029 .63681 L p .30037 .6374 L .735 .986 .896 r
F P
0 g
s
.30029 .63681 m .29436 .64292 L p .30554 .63227 L .185 .032 .427 r
F P
0 g
s
.29436 .64292 m .28241 .65523 L p .30027 .6375 L .185 .032 .427 r
F P
0 g
s
.30027 .6375 m .30554 .63227 L .29436 .64292 L .185 .032 .427 r
F
.32138 .63359 m .30591 .63233 L p .31454 .62453 L .554 .615 .868 r
F P
0 g
s
.30594 .63192 m .31393 .62501 L .30628 .63153 L .185 .032 .427 r
F
.30591 .63233 m .31322 .62562 L .31454 .62453 L .554 .615 .868 r
F
.28241 .65523 m .3043 .63425 L .30554 .63227 L .185 .032 .427 r
F
.35932 .61573 m .33404 .63519 L p .3154 .64349 L .656 .521 .688 r
F P
0 g
s
.35932 .61573 m .3154 .64349 L .31545 .64071 L .656 .521 .688 r
F
.30591 .63233 m .33705 .63487 L p .30423 .63433 L .576 .571 .81 r
F P
0 g
s
.3043 .63425 m .31393 .62501 L .30708 .63094 L .185 .032 .427 r
F
.3043 .63425 m .30708 .63094 L .30554 .63227 L F
.28686 .6704 m .29061 .66262 L .29984 .66 L .576 .571 .81 r
F
.30271 .65918 m .29982 .66122 L p .29984 .66 L F P
0 g
s
.29982 .66122 m .28686 .6704 L p .29984 .66 L .576 .571 .81 r
F P
0 g
s
.30113 .64273 m .304 .64068 L p .30109 .6424 L .673 .96 .941 r
F P
0 g
s
.31655 .64911 m .31531 .6496 L .31528 .65107 L .682 .543 .69 r
F
.33404 .63519 m .3154 .64388 L .3154 .64349 L .656 .521 .688 r
F
.34284 .6387 m .31655 .64911 L .31528 .65107 L .682 .543 .69 r
F
.33404 .63519 m .31531 .6496 L p .3154 .64388 L .656 .521 .688 r
F P
0 g
s
.29419 .66586 m .31183 .65228 L p .31897 .65716 L .707 .617 .755 r
F P
0 g
s
.33705 .63487 m .30271 .65918 L p .30423 .63433 L .576 .571 .81 r
F P
0 g
s
.28223 .65764 m .29984 .66 L p .29995 .65189 L .521 .524 .799 r
F P
0 g
s
.29061 .66262 m .30423 .63433 L .30271 .65918 L .576 .571 .81 r
F
.36014 .63236 m .36138 .64789 L .39382 .62144 L .521 .524 .799 r
F
.36178 .6529 m .36844 .64636 L p .36138 .64789 L F P
0 g
s
.36844 .64636 m .39382 .62144 L p .36138 .64789 L .521 .524 .799 r
F P
0 g
s
.34799 .66645 m .36178 .6529 L p .33989 .65254 L .521 .524 .799 r
F P
0 g
s
.31786 .65732 m .34799 .66645 L .33989 .65254 L .521 .524 .799 r
F
.28686 .6704 m .29982 .66122 L p .30446 .66283 L .503 .43 .707 r
F P
0 g
s
.31298 .67521 m .31356 .67219 L .28822 .67161 L .54 .127 .206 r
F
.31022 .67479 m .28686 .6704 L .30446 .66283 L .503 .43 .707 r
F
.24736 .67388 m .28686 .6704 L .28822 .67161 L p .536 .715 .954 r
F P
0 g
s
.24807 .67578 m .3059 .66581 L .25273 .67341 L .547 .451 .698 r
F
.3059 .66581 m .32222 .663 L .30573 .66546 L F
.3059 .66581 m .30573 .66546 L .25273 .67341 L F
.32264 .67955 m .34799 .66645 L p .30801 .67018 L F P
0 g
s
.28822 .67161 m .31245 .67802 L p .31285 .67594 L .54 .127 .206 r
F P
0 g
s
.32614 .68164 m .31245 .67802 L p .31896 .67937 L .228 .566 .938 r
F P
0 g
s
.31289 .6757 m .32264 .67955 L .3123 .67884 L .48 .074 .193 r
F
.31093 .67625 m .32264 .67955 L .30801 .67018 L .547 .451 .698 r
F
.31126 .67693 m .28686 .6704 L p .31022 .67479 L .503 .43 .707 r
F P
0 g
s
.32264 .67955 m .31226 .67903 L p .3123 .67884 L .48 .074 .193 r
F P
0 g
s
.31896 .67937 m .32264 .67955 L p .29402 .67149 L .547 .451 .698 r
F P
0 g
s
.32503 .68062 m .31126 .67693 L p .3737 .68671 L .503 .43 .707 r
F P
0 g
s
.31126 .67693 m .31022 .67479 L .3737 .68671 L .503 .43 .707 r
F
.28686 .6704 m .31016 .67664 L p .31789 .67888 L .228 .566 .938 r
F P
0 g
s
.31226 .67903 m .24807 .67578 L p .3123 .67884 L .48 .074 .193 r
F P
0 g
s
.31036 .67758 m .3123 .67884 L .30897 .67861 L .48 .074 .193 r
F
.28209 .67525 m .30897 .67861 L .25384 .67484 L F
.31441 .67793 m .30287 .67515 L .30021 .67405 L .228 .566 .938 r
F
.30287 .67515 m .28822 .67161 L .30021 .67405 L F
.30021 .67405 m .28822 .67161 L .29029 .67134 L F
.31718 .67881 m .28822 .67161 L .31441 .67793 L F
.31245 .67802 m .28822 .67161 L p .32187 .67997 L F P
0 g
s
.24807 .67578 m .31226 .67903 L p .29402 .67149 L .547 .451 .698 r
F P
0 g
s
.31016 .67664 m .32503 .68062 L p .31789 .67888 L .228 .566 .938 r
F P
0 g
s
.31245 .67802 m .3123 .67884 L .31036 .67758 L .48 .074 .193 r
F
.31226 .67903 m .31896 .67937 L p .31126 .67693 L .547 .451 .698 r
F P
0 g
s
.31226 .67903 m .31126 .67693 L .29402 .67149 L .547 .451 .698 r
F
.29402 .67149 m .31093 .67625 L .30801 .67018 L F
.33989 .65254 m .36178 .6529 L .36138 .64789 L .521 .524 .799 r
F
.34826 .65073 m .30333 .66046 L .33705 .63487 L .503 .43 .707 r
F
.29982 .66122 m .30271 .65918 L p .30333 .66046 L F P
0 g
s
.30271 .65918 m .33705 .63487 L p .30333 .66046 L .503 .43 .707 r
F P
0 g
s
.30446 .66283 m .29982 .66122 L .30333 .66046 L .503 .43 .707 r
F
.25273 .67341 m .30573 .66546 L .29984 .66 L .547 .451 .698 r
F
.32222 .663 m .30333 .66046 L p .30573 .66546 L F P
0 g
s
.30333 .66046 m .29984 .66 L p .30573 .66546 L .547 .451 .698 r
F P
0 g
s
.29984 .66 m .34799 .66645 L p .29995 .65189 L .521 .524 .799 r
F P
0 g
s
.41095 .67442 m .40522 .65074 L .41027 .65049 L .431 .452 .781 r
F
.44022 .63261 m .40864 .59303 L p .42811 .5963 L .25 .38 .8 r
F P
0 g
s
.49803 .65895 m .48201 .636 L .45688 .64379 L .731 .615 .729 r
F
.43723 .59783 m .41744 .55563 L .40428 .5923 L p .276 .334 .742 r
F P
0 g
s
.44496 .5874 m .44375 .59892 L p .42241 .55228 L .25 .38 .8 r
F P
0 g
s
.41552 .54763 m .44166 .50864 L p .44012 .51082 L .611 .819 .414 r
F P
0 g
s
.42147 .53347 m .39631 .56216 L .40954 .52414 L p .885 .99 .745 r
F P
0 g
s
.47117 .55668 m .4635 .57773 L p .4614 .53651 L .709 .559 .683 r
F P
0 g
s
.45477 .52281 m .43318 .566 L .47788 .57055 L p .737 .57 .667 r
F P
0 g
s
.41037 .5539 m .44755 .50888 L .47437 .50908 L p .709 .559 .683 r
F P
0 g
s
.43318 .566 m .44735 .528 L .43884 .53396 L .674 .495 .635 r
F
.41502 .52957 m .44496 .5874 L p .42241 .55228 L .25 .38 .8 r
F P
0 g
s
.44022 .63261 m .42811 .5963 L .43723 .59783 L .25 .38 .8 r
F
.44375 .59892 m .44225 .61326 L p .43723 .59783 L F P
0 g
s
.44225 .61326 m .44022 .63261 L p .43723 .59783 L .25 .38 .8 r
F P
0 g
s
.46678 .64639 m .47001 .66082 L .46954 .65147 L 0 .043 .608 r
F
.47297 .67405 m .47165 .66535 L p .47001 .66082 L F P
0 g
s
.47165 .66535 m .46954 .65147 L p .47001 .66082 L 0 .043 .608 r
F P
0 g
s
.44496 .5874 m .4232 .543 L .43011 .55501 L 0 0 .495 r
F
.38759 .49651 m .41957 .4953 L .40662 .53254 L p .826 .93 .901 r
F P
0 g
s
.44717 .46947 m .44835 .45756 L p .41119 .52543 L .905 .943 .821 r
F P
0 g
s
.40954 .52414 m .41957 .4953 L p .42036 .51117 L .885 .99 .745 r
F P
0 g
s
.41119 .52543 m .40954 .52414 L .42036 .51117 L .885 .99 .745 r
F
.41119 .52543 m .44372 .5043 L .44717 .46947 L p .905 .943 .821 r
F P
0 g
s
.42004 .5378 m .42791 .55073 L .4081 .50196 L 0 .043 .608 r
F
.43164 .55686 m .43491 .56335 L p .43011 .55501 L 0 0 .495 r
F P
0 g
s
.4232 .543 m .42861 .55188 L .43011 .55501 L 0 0 .495 r
F
.42861 .55188 m .43164 .55686 L .43011 .55501 L F
.4081 .50196 m .41046 .51556 L p .42004 .5378 L 0 .043 .608 r
F P
0 g
s
.41744 .55563 m .43253 .5834 L .41502 .52957 L 0 .043 .608 r
F
.41046 .51556 m .41205 .5247 L p .41502 .52957 L F P
0 g
s
.41205 .5247 m .41744 .55563 L p .41502 .52957 L 0 .043 .608 r
F P
0 g
s
.3633 .43928 m .3667 .45691 L p .36434 .44569 L .912 .724 .226 r
F P
0 g
s
.3667 .45691 m .36679 .45738 L p .36663 .45659 L .912 .724 .226 r
F P
0 g
s
.36663 .45659 m .36434 .44569 L .3667 .45691 L .912 .724 .226 r
F
.34015 .39176 m .36175 .44198 L p .35761 .42698 L .854 .843 .833 r
F P
0 g
s
.36175 .44198 m .3638 .44675 L p .36422 .44493 L .854 .843 .833 r
F P
0 g
s
.3667 .45691 m .35906 .41724 L p .36696 .4577 L .879 .885 .834 r
F P
0 g
s
.36422 .44493 m .36701 .43285 L .3633 .43928 L .854 .843 .833 r
F
.36422 .44493 m .3633 .43928 L .36175 .44198 L F
.37871 .50975 m .37827 .5104 L .37443 .49701 L .826 .93 .901 r
F
.37827 .5104 m .36985 .52281 L .37443 .49701 L F
.3743 .49952 m .37405 .49819 L .3723 .5036 L p 0 0 .179 r
F P
0 g
s
.37443 .49701 m .3743 .49952 L .37256 .4905 L p .912 .724 .226 r
F P
0 g
s
.37246 .4234 m .37839 .42612 L p .36878 .42978 L .879 .885 .834 r
F P
0 g
s
.37839 .42612 m .37246 .4234 L p .37918 .42452 L .941 .831 .707 r
F P
0 g
s
.39006 .46024 m .37372 .42121 L p .3643 .44545 L .854 .843 .833 r
F P
0 g
s
.3638 .44675 m .36458 .44715 L p .3643 .44545 L .854 .843 .833 r
F P
0 g
s
.36457 .4471 m .36434 .44569 L .36663 .45659 L .912 .724 .226 r
F
.36458 .44715 m .36663 .45659 L .36655 .45629 L F
.3638 .44675 m .3643 .44545 L .36422 .44493 L .854 .843 .833 r
F
.36223 .43631 m .36457 .4471 L .36458 .44715 L .912 .724 .226 r
F
.36457 .4471 m .36663 .45659 L .36458 .44715 L F
.36422 .44493 m .37372 .42121 L .36701 .43285 L .854 .843 .833 r
F
.3643 .44545 m .37372 .42121 L .36422 .44493 L F
.36458 .44715 m .39006 .46024 L p .3643 .44545 L F P
0 g
s
.36223 .43631 m .35891 .4198 L .36434 .44569 L .912 .724 .226 r
F
.36223 .43631 m .36434 .44569 L .36457 .4471 L F
.36655 .45629 m .36663 .45659 L .36679 .45738 L F
.3638 .44675 m .36356 .45133 L p .37963 .45673 L .786 .858 .904 r
F P
0 g
s
.39006 .46024 m .37817 .45413 L p .37963 .45673 L .786 .858 .904 r
F P
0 g
s
.37817 .45413 m .3638 .44675 L p .37963 .45673 L .786 .858 .904 r
F P
0 g
s
.35891 .4198 m .35906 .41724 L .35996 .42194 L p .912 .724 .226 r
F P
0 g
s
.35996 .42194 m .3633 .43928 L p .35891 .4198 L .912 .724 .226 r
F P
0 g
s
.36559 .43532 m .36335 .4392 L .35906 .41724 L .879 .885 .834 r
F
.36701 .43285 m .37246 .4234 L .36992 .42026 L .854 .843 .833 r
F
.36878 .42978 m .36653 .43369 L .35906 .41724 L .879 .885 .834 r
F
.36653 .43369 m .36559 .43532 L .3622 .42594 L F
.36653 .43369 m .3622 .42594 L .35906 .41724 L F
.35978 .42125 m .37137 .4406 L .36882 .43624 L 0 0 .433 r
F
.3701 .43462 m .37271 .44078 L p .36611 .43158 L .238 .477 .885 r
F P
0 g
s
.37271 .44078 m .37599 .44853 L p .37155 .44091 L .238 .477 .885 r
F P
0 g
s
.37155 .44091 m .36611 .43158 L .37271 .44078 L .238 .477 .885 r
F
.37155 .44091 m .35987 .42141 L p .37016 .44106 L .24 .431 .849 r
F P
0 g
s
.35987 .42141 m .37155 .44091 L p .35978 .42125 L 0 0 .433 r
F P
0 g
s
.36679 .45738 m .3667 .45691 L p .36696 .4577 L .879 .885 .834 r
F P
0 g
s
.38082 .48363 m .39006 .46024 L .36356 .45133 L .786 .858 .904 r
F
.41957 .4953 m .3877 .49651 L p .38766 .49639 L .879 .885 .834 r
F P
0 g
s
.41957 .4953 m .38766 .49639 L .40197 .4726 L .879 .885 .834 r
F
.40997 .45932 m .39006 .46024 L .38765 .49641 L p .905 .943 .821 r
F P
0 g
s
.38765 .49641 m .39006 .46024 L p .3812 .48435 L .786 .858 .904 r
F P
0 g
s
.38918 .47965 m .37599 .44853 L p .40298 .50536 L 0 0 .495 r
F P
0 g
s
.37599 .44853 m .37271 .44078 L p .37718 .45032 L 0 0 .495 r
F P
0 g
s
.40298 .50536 m .37599 .44853 L .37718 .45032 L 0 0 .495 r
F
.37155 .44091 m .37718 .45032 L p .37137 .4406 L 0 0 .433 r
F P
0 g
s
.37718 .45032 m .4081 .50196 L p .37722 .45038 L 0 0 .433 r
F P
0 g
s
.37271 .44078 m .3701 .43462 L p .3779 .4514 L 0 0 .495 r
F P
0 g
s
.37774 .45116 m .4081 .50196 L .38184 .45789 L 0 0 .433 r
F
.37812 .45173 m .38184 .45789 L .37863 .45249 L F
.37774 .45116 m .38184 .45789 L .37812 .45173 L F
.4081 .50196 m .40122 .48997 L p .37863 .45249 L F P
0 g
s
.37718 .45032 m .37271 .44078 L .3779 .4514 L 0 0 .495 r
F
.37137 .4406 m .37722 .45038 L .36882 .43624 L 0 0 .433 r
F
.37722 .45038 m .4081 .50196 L .37774 .45116 L F
.37722 .45038 m .37774 .45116 L .36882 .43624 L F
.4081 .50196 m .37718 .45032 L p .39592 .49029 L .24 .431 .849 r
F P
0 g
s
.44496 .5874 m .41502 .52957 L p .4232 .543 L 0 0 .495 r
F P
0 g
s
.47343 .58885 m .48964 .60702 L .44945 .61634 L .76 .617 .7 r
F
.44496 .5874 m .46288 .64112 L p .46028 .63761 L 0 .106 .629 r
F P
0 g
s
.45132 .62551 m .47297 .67405 L .47001 .66082 L 0 .043 .608 r
F
.46288 .64112 m .47001 .66082 L .46678 .64639 L F
.45132 .62551 m .47001 .66082 L .46288 .64112 L F
.41744 .55563 m .43723 .59783 L p .43253 .5834 L F P
0 g
s
.33705 .63487 m .35775 .6024 L p .36078 .64039 L .431 .452 .781 r
F P
0 g
s
.32138 .63359 m .31454 .62453 L .32788 .6177 L .554 .615 .868 r
F
.35877 .58359 m .33817 .61623 L p .36261 .59031 L .81 1 .83 r
F P
0 g
s
.33817 .61623 m .36422 .59312 L p .36261 .59031 L .81 1 .83 r
F P
0 g
s
.33705 .63487 m .33248 .6345 L p .33926 .61187 L .554 .615 .868 r
F P
0 g
s
.33235 .61231 m .33704 .613 L .34895 .59729 L .521 .524 .799 r
F
.35811 .60222 m .34895 .59729 L p .33704 .613 L F P
0 g
s
.32121 .62111 m .34895 .59729 L .31454 .62453 L .521 .524 .799 r
F
.30008 .64149 m .32444 .61946 L .30012 .63923 L F
.32444 .61946 m .34895 .59729 L .32121 .62111 L F
.32444 .61946 m .32121 .62111 L .30012 .63923 L F
.34364 .59444 m .35772 .60201 L p .35807 .60174 L .454 .533 .849 r
F P
0 g
s
.34364 .59444 m .35807 .60174 L .35712 .58735 L .454 .533 .849 r
F
.39054 .5901 m .37364 .59624 L .38367 .56175 L .431 .452 .781 r
F
.40499 .52946 m .38455 .54273 L .35909 .58307 L p .81 1 .83 r
F P
0 g
s
.35867 .54566 m .38455 .54273 L .38763 .4966 L p .786 .858 .904 r
F P
0 g
s
.379 .53391 m .37712 .52381 L .37728 .52367 L .495 .61 .896 r
F
.35909 .58307 m .38455 .54273 L p .35771 .57303 L .753 .889 .94 r
F P
0 g
s
.38367 .56175 m .3789 .53412 L .379 .53391 L .495 .61 .896 r
F
.3789 .53412 m .37712 .52381 L .379 .53391 L F
.37871 .50975 m .37443 .49701 L .37939 .50875 L .826 .93 .901 r
F
.37543 .49727 m .38082 .48363 L .37461 .49347 L .786 .858 .904 r
F
.37939 .50875 m .37443 .49701 L .3876 .49664 L .826 .93 .901 r
F
.35867 .54566 m .38763 .4966 L .36378 .53176 L .786 .858 .904 r
F
.38681 .51579 m .3743 .49952 L .36257 .53583 L p .454 .533 .849 r
F P
0 g
s
.37256 .4905 m .36666 .45991 L p .37443 .49701 L .912 .724 .226 r
F P
0 g
s
.37443 .49701 m .36666 .45991 L .36662 .4597 L .912 .724 .226 r
F
.35903 .53876 m .35867 .54566 L p .36378 .53176 L .786 .858 .904 r
F P
0 g
s
.36369 .53236 m .3743 .49952 L p .35982 .54338 L 0 0 .179 r
F P
0 g
s
.3723 .5036 m .36271 .53334 L p .3743 .49952 L 0 0 .179 r
F P
0 g
s
.36653 .45958 m .3638 .44675 L .36525 .4519 L .495 .61 .896 r
F
.3586 .54709 m .3593 .54497 L .36271 .53334 L 0 0 .179 r
F
.3593 .54497 m .3743 .49952 L .36334 .53149 L F
.3593 .54497 m .36334 .53149 L .36271 .53334 L F
.38367 .56175 m .35808 .60188 L p .35866 .54712 L .554 .615 .868 r
F P
0 g
s
.35904 .54562 m .35867 .54566 L p .34399 .58852 L .753 .889 .94 r
F P
0 g
s
.35792 .54903 m .3586 .54709 L .35904 .54562 L 0 0 .179 r
F
.34378 .59199 m .35859 .54712 L .35792 .54903 L F
.35859 .54712 m .3586 .54709 L .35792 .54903 L F
.35867 .54566 m .3638 .44675 L p .37373 .5042 L .495 .61 .896 r
F P
0 g
s
.38367 .56175 m .35903 .54589 L p .37373 .5042 L .495 .61 .896 r
F P
0 g
s
.35903 .54589 m .38367 .56175 L p .35866 .54712 L .554 .615 .868 r
F P
0 g
s
.35529 .5596 m .35523 .55858 L .35688 .55305 L .554 .615 .868 r
F
.35455 .56068 m .35523 .55858 L p .3586 .54709 L 0 0 .179 r
F P
0 g
s
.35529 .5596 m .35688 .55305 L .35808 .60188 L .554 .615 .868 r
F
.35688 .55305 m .35866 .54712 L .35808 .60188 L F
.35523 .55858 m .36369 .53236 L p .35909 .54569 L 0 0 .179 r
F P
0 g
s
.3586 .54709 m .35523 .55858 L .35909 .54569 L 0 0 .179 r
F
.35543 .55659 m .34378 .59199 L .35536 .5561 L p .826 .93 .901 r
F P
0 g
s
.35536 .5561 m .34378 .59199 L p .35904 .54562 L 0 0 .179 r
F P
0 g
s
.34378 .59199 m .35455 .56068 L .3586 .54709 L 0 0 .179 r
F
.35688 .55305 m .35867 .54566 L .35866 .54712 L .554 .615 .868 r
F
.3586 .54709 m .36369 .53236 L .35982 .54338 L 0 0 .179 r
F
.35903 .54589 m .35867 .54566 L p .37373 .5042 L .495 .61 .896 r
F P
0 g
s
.35867 .54566 m .35903 .54589 L p .35866 .54712 L .554 .615 .868 r
F P
0 g
s
.37728 .52367 m .36559 .45695 L .36653 .45958 L .495 .61 .896 r
F
.36559 .45695 m .3638 .44675 L .36653 .45958 L F
.36568 .45396 m .36536 .45219 L p .36551 .45291 L .912 .724 .226 r
F P
0 g
s
.3638 .44675 m .36455 .44772 L p .36525 .4519 L .495 .61 .896 r
F P
0 g
s
.38384 .55923 m .38367 .56175 L p .36653 .45958 L .495 .61 .896 r
F P
0 g
s
.36653 .45958 m .36455 .44772 L .36666 .45991 L .495 .61 .896 r
F
.36653 .45958 m .36666 .45991 L .38384 .55923 L F
.37443 .49701 m .36657 .45944 L .36551 .45291 L .912 .724 .226 r
F
.36455 .44772 m .3638 .44675 L p .36278 .44187 L .238 .477 .885 r
F P
0 g
s
.36316 .44183 m .36278 .44187 L .3596 .42662 L .238 .477 .885 r
F
.36455 .44772 m .36278 .44187 L .36316 .44183 L F
.37443 .49701 m .36536 .45219 L .3649 .4496 L .912 .724 .226 r
F
.36536 .45219 m .36102 .43074 L .3649 .4496 L F
.3649 .4496 m .36102 .43074 L .36223 .43631 L F
.36527 .45144 m .36223 .43631 L .36458 .44715 L F
.36527 .45144 m .36458 .44715 L .36679 .45738 L F
.37461 .49347 m .38082 .48363 L .37872 .4797 L .786 .858 .904 r
F
.37443 .49701 m .38759 .49651 L p .3876 .49664 L .826 .93 .901 r
F P
0 g
s
.3877 .49651 m .37443 .49701 L .36679 .45738 L p .879 .885 .834 r
F P
0 g
s
.36455 .44772 m .38918 .47965 L .38676 .51566 L p .495 .61 .896 r
F P
0 g
s
.3743 .49952 m .36758 .46311 L .36682 .46037 L .24 .431 .849 r
F
.36666 .45991 m .3743 .49952 L p .36682 .46037 L F P
0 g
s
.36758 .46311 m .36733 .46179 L .36682 .46037 L .24 .431 .849 r
F
.36679 .45738 m .37443 .49701 L p .36527 .45144 L .912 .724 .226 r
F P
0 g
s
.36666 .45991 m .36455 .44772 L .38676 .51566 L .495 .61 .896 r
F
.38455 .54273 m .40065 .52124 L .39459 .50978 L .905 .943 .821 r
F
.41744 .55563 m .38681 .51579 L p .38676 .51589 L .454 .533 .849 r
F P
0 g
s
.36758 .46311 m .36793 .46345 L .36733 .46179 L .24 .431 .849 r
F
.3743 .49952 m .37769 .49976 L .37643 .49519 L F
.3743 .49952 m .37643 .49519 L .36758 .46311 L F
.40118 .5156 m .38918 .47965 L .40562 .51558 L .25 .38 .8 r
F
.40778 .51557 m .38918 .47965 L p .40298 .50536 L 0 0 .495 r
F P
0 g
s
.38918 .47965 m .40778 .51557 L p .40562 .51558 L .25 .38 .8 r
F P
0 g
s
.40118 .5156 m .39379 .51563 L .38918 .47965 L .25 .38 .8 r
F
.38671 .51566 m .41744 .55563 L .41046 .51556 L p .24 .431 .849 r
F P
0 g
s
.41315 .55196 m .41037 .5539 L .40741 .56821 L .674 .495 .635 r
F
.41092 .55325 m .41037 .5539 L p .41043 .55363 L .642 .464 .627 r
F P
0 g
s
.39819 .56575 m .41744 .55563 L .38676 .51589 L .454 .533 .849 r
F
.38675 .51584 m .38384 .55923 L p .37764 .52338 L .495 .61 .896 r
F P
0 g
s
.40276 .58566 m .38367 .56175 L .38675 .51584 L p .25 .38 .8 r
F P
0 g
s
.36257 .53583 m .35523 .55858 L p .38676 .51587 L .454 .533 .849 r
F P
0 g
s
.38676 .51587 m .38681 .51579 L .38675 .51584 L .454 .533 .849 r
F
.38676 .51587 m .38675 .51584 L .36257 .53583 L F
.35523 .55858 m .34364 .59444 L p .35609 .57175 L F P
0 g
s
.35609 .57175 m .38681 .51579 L .35523 .55858 L .454 .533 .849 r
F
.39382 .62144 m .35811 .60222 L p .35775 .6024 L .521 .524 .799 r
F P
0 g
s
.34364 .59444 m .35712 .58735 L .35609 .57175 L .454 .533 .849 r
F
.33248 .6345 m .32138 .63359 L p .32853 .61737 L .554 .615 .868 r
F P
0 g
s
.32444 .61946 m .329 .61712 L .33235 .61231 L .521 .524 .799 r
F
.33248 .6345 m .32853 .61737 L .33906 .61197 L .554 .615 .868 r
F
.33248 .6345 m .33906 .61197 L .33926 .61187 L F
.33705 .63487 m .34298 .60996 L .3431 .6099 L F
.33705 .63487 m .33926 .61187 L .34298 .60996 L F
.34364 .59444 m .34378 .59199 L .31442 .62224 L p .185 .032 .427 r
F P
0 g
s
.35867 .54566 m .3079 .62793 L p .34399 .58852 L .753 .889 .94 r
F P
0 g
s
.31304 .62345 m .34399 .58852 L .3079 .62793 L .753 .889 .94 r
F
.31304 .62345 m .31442 .62224 L .3439 .58862 L F
.33853 .59467 m .32138 .63359 L .35867 .54566 L .554 .615 .868 r
F
.35479 .55195 m .35867 .54566 L p .32138 .63359 L F P
0 g
s
.35772 .60201 m .35811 .60222 L p .35808 .60188 L .454 .533 .849 r
F P
0 g
s
.35808 .60188 m .38367 .56175 L p .37364 .59624 L .431 .452 .781 r
F P
0 g
s
.47297 .67405 m .43723 .59783 L p .41154 .6118 L .276 .334 .742 r
F P
0 g
s
.39382 .62144 m .41523 .63567 L .41154 .6118 L .276 .334 .742 r
F
.41523 .63567 m .47297 .67405 L .41154 .6118 L F
.40428 .5923 m .39382 .62144 L p .43723 .59783 L F P
0 g
s
.39382 .62144 m .40428 .5923 L p .35808 .60188 L .454 .533 .849 r
F P
0 g
s
.35808 .60188 m .40428 .5923 L .40276 .58566 L .454 .533 .849 r
F
.40522 .65074 m .39054 .5901 L .41027 .65049 L .431 .452 .781 r
F
.39054 .5901 m .38367 .56175 L .39201 .58956 L F
.39054 .5901 m .39201 .58956 L .41027 .65049 L F
.40725 .58403 m .41744 .55563 L p .35772 .60201 L .454 .533 .849 r
F P
0 g
s
.35808 .60188 m .33705 .63487 L p .35529 .5596 L .554 .615 .868 r
F P
0 g
s
.31416 .63848 m .39382 .62144 L .33704 .613 L .521 .524 .799 r
F
.329 .61712 m .33704 .613 L .33235 .61231 L F
.31416 .63848 m .33704 .613 L .329 .61712 L F
.33926 .61187 m .35512 .55704 L .35523 .55858 L .554 .615 .868 r
F
.33906 .61197 m .35512 .55704 L .33926 .61187 L F
.33906 .61197 m .32853 .61737 L .35512 .55704 L F
.30008 .64149 m .31416 .63848 L .329 .61712 L .521 .524 .799 r
F
.30008 .64149 m .329 .61712 L .32444 .61946 L F
.33704 .613 m .35886 .61624 L .35775 .6024 L F
.35886 .61624 m .39382 .62144 L .35775 .6024 L F
.34826 .65073 m .33705 .63487 L .34952 .65046 L .503 .43 .707 r
F
.33705 .63487 m .36138 .64789 L p .34952 .65046 L F P
0 g
s
.36138 .64789 m .33705 .63487 L p .36078 .64039 L .431 .452 .781 r
F P
0 g
s
.29995 .65189 m .36014 .63236 L .30008 .64149 L .521 .524 .799 r
F
.36014 .63236 m .39382 .62144 L .35985 .6287 L F
.36014 .63236 m .35985 .6287 L .30008 .64149 L F
.36998 .65249 m .36138 .64789 L p .40522 .65074 L .431 .452 .781 r
F P
0 g
s
.36138 .64789 m .36078 .64039 L .40522 .65074 L .431 .452 .781 r
F
.35811 .60222 m .39382 .62144 L p .35808 .60188 L .454 .533 .849 r
F P
0 g
s
.35775 .6024 m .35808 .60188 L p .36078 .64039 L .431 .452 .781 r
F P
0 g
s
.36078 .64039 m .35808 .60188 L .37364 .59624 L .431 .452 .781 r
F
.35609 .57175 m .39819 .56575 L .38676 .51589 L .454 .533 .849 r
F
.35712 .58735 m .39819 .56575 L .35609 .57175 L F
.35523 .55858 m .35512 .55704 L .35688 .55305 L .554 .615 .868 r
F
.39819 .56575 m .39913 .56985 L .41744 .55563 L .454 .533 .849 r
F
.38675 .51584 m .38676 .51566 L p .40276 .58566 L .25 .38 .8 r
F P
0 g
s
.38676 .51566 m .38918 .47965 L p .39379 .51563 L .25 .38 .8 r
F P
0 g
s
.40276 .58566 m .38676 .51566 L .39379 .51563 L .25 .38 .8 r
F
.3743 .49952 m .38671 .51566 L p .4081 .50196 L .24 .431 .849 r
F P
0 g
s
.41046 .51556 m .4081 .50196 L p .38671 .51566 L .24 .431 .849 r
F P
0 g
s
.37769 .49976 m .40116 .50146 L .36758 .46311 L .24 .431 .849 r
F
.40116 .50146 m .39592 .49029 L .36758 .46311 L F
.38676 .51566 m .38675 .51584 L p .36666 .45991 L .495 .61 .896 r
F P
0 g
s
.37764 .52338 m .36666 .45991 L .38675 .51584 L .495 .61 .896 r
F
.42811 .5963 m .40276 .58566 L .39379 .51563 L .25 .38 .8 r
F
.42811 .5963 m .39379 .51563 L .40118 .5156 L F
.42241 .55228 m .40562 .51558 L .41502 .52957 L F
.40778 .51557 m .41502 .52957 L p .40562 .51558 L F P
0 g
s
.40116 .50146 m .4081 .50196 L .39592 .49029 L .24 .431 .849 r
F
.42811 .5963 m .40118 .5156 L .44375 .59892 L .25 .38 .8 r
F
.40118 .5156 m .40562 .51558 L .44375 .59892 L F
.46288 .64112 m .44496 .5874 L p .46044 .6344 L .781 .455 0 r
F P
0 g
s
.43723 .59783 m .44829 .62142 L p .43253 .5834 L 0 .043 .608 r
F P
0 g
s
.44829 .62142 m .47297 .67405 L p .45132 .62551 L 0 .043 .608 r
F P
0 g
s
.4592 .67538 m .47082 .67917 L .45837 .67129 L .276 .334 .742 r
F
.47708 .73772 m .50216 .79178 L .47082 .67917 L .107 .137 .609 r
F
.4837 .71993 m .45761 .66753 L .44496 .5874 L 0 .106 .629 r
F
.44375 .59892 m .44496 .5874 L p .45761 .66753 L F P
0 g
s
.44829 .62142 m .45132 .62551 L .43253 .5834 L 0 .043 .608 r
F
.51345 .84053 m .49288 .75132 L .51457 .8435 L 0 0 .426 r
F
.50216 .79178 m .50196 .79001 L .47297 .67405 L .107 .137 .609 r
F
.4837 .71993 m .45921 .63617 L .48632 .72295 L 0 .106 .629 r
F
.45921 .63617 m .44496 .5874 L .46028 .63761 L F
.45921 .63617 m .46028 .63761 L .48632 .72295 L F
.46044 .6344 m .44496 .5874 L .44154 .58037 L .781 .455 0 r
F
.46044 .6344 m .44154 .58037 L .44501 .59178 L F
.43253 .5834 m .46081 .63541 L .42004 .5378 L 0 .043 .608 r
F
.46081 .63541 m .45422 .61722 L .42004 .5378 L F
.47883 .69058 m .46561 .65229 L .47761 .68528 L 0 .106 .629 r
F
.48632 .72295 m .46561 .65229 L .47883 .69058 L F
.48632 .72295 m .46028 .63761 L .46561 .65229 L F
.41502 .52957 m .42004 .5378 L .41046 .51556 L 0 .043 .608 r
F
.3795 .45381 m .3779 .4514 L .3701 .43462 L 0 0 .495 r
F
.35906 .41724 m .35891 .4198 L .34852 .40169 L p .991 .733 .392 r
F P
0 g
s
.35891 .4198 m .36882 .43624 L .35059 .40516 L 0 0 .433 r
F
.36882 .43624 m .35788 .41748 L .35059 .40516 L F
.37599 .44853 m .38918 .47965 L p .37063 .44101 L .238 .477 .885 r
F P
0 g
s
.36434 .44569 m .35891 .4198 L .3633 .43928 L .912 .724 .226 r
F
.36364 .44177 m .37016 .44106 L .35987 .42141 L .24 .431 .849 r
F
.36351 .44179 m .35987 .42141 L .36346 .44179 L F
.36364 .44177 m .35987 .42141 L .36351 .44179 L F
.37155 .44091 m .37137 .4406 L .35978 .42125 L 0 0 .433 r
F
.36102 .43074 m .35891 .4198 L p .36223 .43631 L .912 .724 .226 r
F P
0 g
s
.36657 .45944 m .36102 .43074 L p .3648 .4494 L .912 .724 .226 r
F P
0 g
s
.36102 .43074 m .36316 .44183 L p .35987 .42141 L .24 .431 .849 r
F P
0 g
s
.38918 .47965 m .36455 .44772 L p .36316 .44183 L .238 .477 .885 r
F P
0 g
s
.38918 .47965 m .36316 .44183 L .3678 .44132 L .238 .477 .885 r
F
.36316 .44183 m .36657 .45944 L p .36346 .44179 L .24 .431 .849 r
F P
0 g
s
.36346 .44179 m .35987 .42141 L .36316 .44183 L .24 .431 .849 r
F
.35987 .42141 m .35891 .4198 L .36102 .43074 L p F P
0 g
s
.3678 .44132 m .36316 .44183 L .3596 .42662 L .238 .477 .885 r
F
.37063 .44101 m .3596 .42662 L .36611 .43158 L F
.38918 .47965 m .3678 .44132 L .37063 .44101 L F
.3678 .44132 m .3596 .42662 L .37063 .44101 L F
.36733 .46179 m .36364 .44177 L .36682 .46037 L .24 .431 .849 r
F
.36364 .44177 m .36351 .44179 L .36682 .46037 L F
.36733 .46179 m .39592 .49029 L .37016 .44106 L F
.36733 .46179 m .37016 .44106 L .36364 .44177 L F
.37063 .44101 m .36611 .43158 L .37155 .44091 L .238 .477 .885 r
F
.37063 .44101 m .37155 .44091 L .37599 .44853 L F
.3596 .42662 m .35816 .4205 L .36294 .42615 L F
.36611 .43158 m .36294 .42615 L .3701 .43462 L F
.3596 .42662 m .36294 .42615 L .36611 .43158 L F
.35788 .41748 m .35669 .41543 L .35059 .40516 L 0 0 .433 r
F
.3809 .42227 m .37113 .41502 L p .4045 .42572 L .881 .453 .167 r
F P
0 g
s
.47871 .47008 m .42138 .44322 L p .4228 .44359 L .631 .067 0 r
F P
0 g
s
.4228 .44359 m .47365 .45684 L .47649 .46427 L .631 .067 0 r
F
.4228 .44359 m .47649 .46427 L .47871 .47008 L F
.39593 .4313 m .38819 .42767 L p .40104 .43183 L F P
0 g
s
.39221 .42897 m .41553 .43334 L .40104 .43183 L .631 .067 0 r
F
.38819 .42767 m .39954 .42994 L p .39221 .42897 L F P
0 g
s
.50039 .47184 m .4873 .46685 L .54704 .47563 L .401 0 0 r
F
.4873 .46685 m .47871 .47008 L p .47365 .45684 L .631 .067 0 r
F P
0 g
s
.54038 .48633 m .49184 .47018 L p .5115 .47492 L 0 .515 .828 r
F P
0 g
s
.50271 .47248 m .54704 .47563 L .55369 .47759 L .401 0 0 r
F
.48522 .46763 m .5115 .47492 L .49184 .47018 L 0 .515 .828 r
F
.47871 .47008 m .50039 .47184 L .50271 .47248 L .401 0 0 r
F
.50039 .47184 m .54704 .47563 L .50271 .47248 L F
.48522 .46763 m .4873 .46685 L p .50039 .47184 L F P
0 g
s
.46886 .46133 m .52841 .47815 L .5115 .47492 L 0 .515 .828 r
F
.52841 .47815 m .53556 .48073 L .5115 .47492 L F
.46886 .46133 m .5115 .47492 L .48522 .46763 L F
.46886 .46133 m .42754 .44463 L .50534 .46982 L F
.546 .48449 m .54728 .48477 L p .52451 .47381 L .419 .758 .995 r
F P
0 g
s
.78706 .49453 m .79147 .49207 L .79192 .49202 L .138 0 0 r
F
.81039 .49209 m .81071 .49209 L .81041 .49208 L .27 0 0 r
F
.65047 .49815 m .74831 .49445 L .74838 .49444 L F
.75792 .49285 m .81041 .49208 L .77506 .492 L F
.65047 .49815 m .74838 .49444 L .75792 .49285 L F
.74838 .49444 m .81041 .49208 L .75792 .49285 L F
.68917 .50013 m .67185 .49924 L p .66939 .49894 L .047 .605 .81 r
F P
0 g
s
.67185 .49924 m .65047 .49815 L p .66939 .49894 L .047 .605 .81 r
F P
0 g
s
.65047 .49815 m .77506 .492 L .76437 .49222 L .27 0 0 r
F
.64913 .49771 m .66939 .49894 L .65047 .49815 L .047 .605 .81 r
F
.64489 .49635 m .68917 .50013 L .64913 .49771 L F
.691 .50023 m .68917 .50013 L p .64489 .49635 L F P
0 g
s
.69087 .50028 m .61716 .49975 L p .65593 .49728 L .436 0 0 r
F P
0 g
s
.65593 .49728 m .70691 .49404 L .691 .50023 L .436 0 0 r
F
.70011 .49668 m .69141 .50025 L .78781 .4941 L .27 0 0 r
F
.691 .50023 m .69141 .50025 L p .70011 .49668 L F P
0 g
s
.54038 .48633 m .54657 .4847 L p .66394 .49662 L .047 .605 .81 r
F P
0 g
s
.76609 .49602 m .78781 .4941 L p .69141 .50025 L .27 0 0 r
F P
0 g
s
.73924 .50063 m .78015 .49563 L p .69545 .49796 L .389 .798 .98 r
F P
0 g
s
.54038 .48633 m .66394 .49662 L .67744 .4984 L .047 .605 .81 r
F
.65593 .49728 m .691 .50023 L .69087 .50028 L .436 0 0 r
F
.6153 .49112 m .6532 .4952 L p .54657 .4847 L .047 .605 .81 r
F P
0 g
s
.69168 .50028 m .73924 .50063 L p .69545 .49796 L .389 .798 .98 r
F P
0 g
s
.65588 .49555 m .67779 .49845 L .69545 .49796 L .389 .798 .98 r
F
.67779 .49845 m .69168 .50028 L .69545 .49796 L F
.7082 .50111 m .77331 .49698 L .78015 .49563 L .65 .401 .53 r
F
.7879 .4941 m .76609 .49602 L p .78015 .49563 L F P
0 g
s
.76609 .49602 m .7082 .50111 L p .78015 .49563 L .65 .401 .53 r
F P
0 g
s
.7082 .50111 m .69673 .49995 L p .69141 .50025 L .27 0 0 r
F P
0 g
s
.69715 .49992 m .7082 .50111 L .76609 .49602 L p .27 0 0 r
F P
0 g
s
.65588 .49555 m .69545 .49796 L .6645 .4957 L .389 .798 .98 r
F
.65588 .49555 m .6645 .4957 L .6532 .4952 L F
.6532 .4952 m .7082 .50111 L p .66394 .49662 L .047 .605 .81 r
F P
0 g
s
.54657 .4847 m .6532 .4952 L .66394 .49662 L .047 .605 .81 r
F
.58516 .49027 m .7082 .50111 L .64489 .49635 L F
.60244 .49279 m .64489 .49635 L .6078 .49357 L F
.58516 .49027 m .64489 .49635 L .60244 .49279 L F
.61716 .49975 m .624 .4946 L .65593 .49728 L .436 0 0 r
F
.7082 .50111 m .691 .50023 L p .64489 .49635 L .047 .605 .81 r
F P
0 g
s
.61716 .49975 m .69168 .50028 L p .64178 .49469 L .389 .798 .98 r
F P
0 g
s
.64178 .49469 m .69168 .50028 L .65588 .49555 L .389 .798 .98 r
F
.67744 .4984 m .66394 .49662 L .7082 .50111 L .047 .605 .81 r
F
.54689 .48007 m .54728 .48477 L .57187 .48707 L .419 .758 .995 r
F
.52451 .47381 m .53279 .4778 L .54689 .48007 L F
.53279 .4778 m .54728 .48477 L .54689 .48007 L F
.6423 .49548 m .69909 .50262 L .69959 .50254 L .734 .466 .523 r
F
.51213 .47157 m .546 .48449 L .52451 .47381 L .419 .758 .995 r
F
.52938 .4489 m .53415 .46514 L p .554 .45891 L .592 .095 .044 r
F P
0 g
s
.52938 .4489 m .554 .45891 L .5581 .45763 L .592 .095 .044 r
F
.5419 .45027 m .52938 .4489 L p .60878 .47302 L F P
0 g
s
.49743 .46303 m .53099 .42643 L .56092 .45674 L .712 .271 .202 r
F
.49214 .44805 m .49008 .45216 L p .49337 .45803 L .737 .57 .667 r
F P
0 g
s
.50351 .44427 m .49743 .46303 L .47254 .45127 L p .828 .407 .241 r
F P
0 g
s
.48413 .44532 m .48659 .44164 L p .48208 .44814 L .611 .819 .414 r
F P
0 g
s
.47717 .44341 m .50351 .44427 L .47254 .45127 L .828 .407 .241 r
F
.52316 .46858 m .49743 .46303 L p .56092 .45674 L .712 .271 .202 r
F P
0 g
s
.49254 .44392 m .4608 .44741 L .50459 .46883 L .693 .176 0 r
F
.4882 .43494 m .50459 .46883 L .52498 .47116 L F
.48906 .46394 m .546 .48449 L .50006 .46696 L .419 .758 .995 r
F
.52841 .47815 m .50534 .46982 L .54919 .48402 L 0 .515 .828 r
F
.46886 .46133 m .50534 .46982 L .52841 .47815 L F
.54666 .48468 m .54038 .48633 L p .5115 .47492 L F P
0 g
s
.546 .48449 m .47871 .47008 L p .50271 .47248 L .401 0 0 r
F P
0 g
s
.54919 .48402 m .54666 .48468 L p .5115 .47492 L 0 .515 .828 r
F P
0 g
s
.47871 .47008 m .48522 .46763 L p .50039 .47184 L .401 0 0 r
F P
0 g
s
.47871 .47008 m .546 .48449 L p .47275 .45947 L .419 .758 .995 r
F P
0 g
s
.45612 .45295 m .48906 .46394 L p .42754 .44463 L 0 .515 .828 r
F P
0 g
s
.48906 .46394 m .54919 .48402 L p .50534 .46982 L 0 .515 .828 r
F P
0 g
s
.42754 .44463 m .48906 .46394 L .50534 .46982 L 0 .515 .828 r
F
.6532 .4952 m .6153 .49112 L p .64124 .49467 L .734 .466 .523 r
F P
0 g
s
.60244 .49279 m .61716 .49975 L p .58258 .4899 L .401 0 0 r
F P
0 g
s
.61716 .49975 m .54728 .48477 L p .58258 .4899 L .401 0 0 r
F P
0 g
s
.61716 .49975 m .60244 .49279 L p .624 .4946 L .436 0 0 r
F P
0 g
s
.61716 .49975 m .63576 .49443 L p .54959 .48051 L .419 .758 .995 r
F P
0 g
s
.6153 .49112 m .61356 .49094 L p .6171 .49141 L .734 .466 .523 r
F P
0 g
s
.6171 .49141 m .64124 .49467 L .6153 .49112 L .734 .466 .523 r
F
.48831 .46367 m .61716 .49975 L .48034 .4608 L .419 .758 .995 r
F
.58049 .48787 m .54919 .48402 L .6153 .49112 L p .047 .605 .81 r
F P
0 g
s
.63345 .49433 m .64124 .49467 L .61356 .49094 L .734 .466 .523 r
F
.63605 .49434 m .64178 .49469 L .64764 .49446 L .389 .798 .98 r
F
.64178 .49469 m .65588 .49555 L .6532 .4952 L F
.64178 .49469 m .6532 .4952 L .64764 .49446 L F
.63576 .49443 m .61716 .49975 L p .64178 .49469 L F P
0 g
s
.54728 .48477 m .61716 .49975 L p .57187 .48707 L .419 .758 .995 r
F P
0 g
s
.59181 .52315 m .57294 .52452 L .59325 .53459 L .287 .302 .701 r
F
.59325 .53459 m .56785 .49439 L p .59171 .52239 L .426 .407 .735 r
F P
0 g
s
.60993 .52183 m .58871 .49857 L .59181 .52315 L .287 .302 .701 r
F
.56785 .49439 m .59325 .53459 L p .57783 .52695 L .235 .32 .747 r
F P
0 g
s
.56785 .49439 m .57783 .52695 L .57193 .52403 L .235 .32 .747 r
F
.56355 .51987 m .56644 .50278 L p .57193 .52403 L F P
0 g
s
.56644 .50278 m .56785 .49439 L p .57193 .52403 L .235 .32 .747 r
F P
0 g
s
.61774 .51897 m .54741 .50407 L p .61823 .49379 L .726 .7 .825 r
F P
0 g
s
.61504 .49369 m .61823 .49379 L .54741 .50407 L .726 .7 .825 r
F
.63577 .49472 m .61774 .51897 L p .61823 .49379 L F P
0 g
s
.62389 .5107 m .63577 .49472 L p .6533 .49771 L .7 .66 .808 r
F P
0 g
s
.65688 .49969 m .66871 .50623 L .68988 .50372 L .726 .65 .774 r
F
.62389 .5107 m .6533 .49771 L .66871 .50623 L .7 .66 .808 r
F
.66158 .4986 m .6533 .49771 L p .68988 .50372 L .726 .65 .774 r
F P
0 g
s
.61538 .49364 m .54921 .48653 L p .61183 .49418 L .726 .65 .774 r
F P
0 g
s
.54921 .48653 m .56959 .50066 L .61183 .49418 L .726 .65 .774 r
F
.524 .4702 m .52498 .47116 L .52583 .4706 L .693 .176 0 r
F
.524 .4702 m .52583 .4706 L .53415 .46514 L F
.4882 .43494 m .524 .4702 L .53415 .46514 L F
.5363 .47246 m .53415 .46514 L p .52498 .47116 L F P
0 g
s
.63605 .49434 m .48034 .4608 L p .54959 .48051 L .419 .758 .995 r
F P
0 g
s
.63576 .49443 m .63605 .49434 L p .54959 .48051 L .419 .758 .995 r
F P
0 g
s
.47063 .45888 m .47876 .46112 L .45612 .45295 L .818 .536 .513 r
F
.47422 .45948 m .47033 .45864 L p .47732 .46059 L .419 .758 .995 r
F P
0 g
s
.48906 .46394 m .45612 .45295 L p .47876 .46112 L .818 .536 .513 r
F P
0 g
s
.47732 .46059 m .48831 .46367 L .47422 .45948 L .419 .758 .995 r
F
.48034 .4608 m .47422 .45948 L p .48831 .46367 L F P
0 g
s
.47048 .45884 m .47871 .47008 L p .47275 .45947 L .419 .758 .995 r
F P
0 g
s
.40475 .43007 m .4087 .43563 L .40922 .43311 L .56 0 0 r
F
.38819 .42767 m .41925 .44222 L p .39543 .42803 L .265 .651 .966 r
F P
0 g
s
.4087 .43563 m .40984 .43631 L .40922 .43311 L .56 0 0 r
F
.3629 .40865 m .41925 .44222 L .42015 .44248 L .265 .651 .966 r
F
.40468 .42999 m .38624 .42 L .40202 .43164 L .56 0 0 r
F
.40468 .42999 m .40202 .43164 L .4087 .43563 L F
.40909 .43301 m .42261 .43499 L .41191 .43535 L .631 .067 0 r
F
.3629 .40865 m .42015 .44248 L .36986 .41239 L .265 .651 .966 r
F
.40922 .43311 m .40984 .43631 L .42098 .44289 L .56 0 0 r
F
.41008 .43758 m .42015 .44248 L p .40984 .43631 L F P
0 g
s
.42015 .44248 m .42098 .44289 L p .40984 .43631 L .56 0 0 r
F P
0 g
s
.42098 .44289 m .42015 .44248 L p .46886 .46133 L 0 .515 .828 r
F P
0 g
s
.42015 .44248 m .42754 .44463 L .46886 .46133 L 0 .515 .828 r
F
.43246 .44606 m .42015 .44248 L .46622 .46422 L .265 .651 .966 r
F
.41925 .44222 m .42138 .44322 L p .42015 .44248 L F P
0 g
s
.42138 .44322 m .46622 .46422 L p .42015 .44248 L .265 .651 .966 r
F P
0 g
s
.3629 .40865 m .38054 .42191 L p .39543 .42803 L .265 .651 .966 r
F P
0 g
s
.38054 .42191 m .38819 .42767 L p .39543 .42803 L .265 .651 .966 r
F P
0 g
s
.36277 .40856 m .37212 .4148 L .35742 .40461 L .898 .611 .494 r
F
.37212 .4148 m .40998 .44012 L .37378 .41566 L F
.37212 .4148 m .37378 .41566 L .35742 .40461 L F
.35906 .41724 m .37246 .4234 L p .36878 .42978 L .879 .885 .834 r
F P
0 g
s
.35906 .41724 m .34867 .39913 L p .35787 .40534 L .941 .831 .707 r
F P
0 g
s
.37246 .4234 m .35906 .41724 L p .35787 .40534 L .941 .831 .707 r
F P
0 g
s
.34852 .40169 m .34867 .39913 L .35906 .41724 L p .991 .733 .392 r
F P
0 g
s
.3701 .43462 m .35669 .41543 L p .3602 .40867 L .845 .741 .75 r
F P
0 g
s
.36097 .4072 m .35967 .40624 L p .35996 .40645 L 0 .148 .597 r
F P
0 g
s
.38819 .42767 m .3629 .40865 L p .36199 .40798 L 0 .148 .597 r
F P
0 g
s
.38092 .42195 m .40998 .44012 L p .36277 .40856 L .898 .611 .494 r
F P
0 g
s
.36986 .41239 m .35742 .40461 L p .36539 .40999 L .898 .611 .494 r
F P
0 g
s
.36097 .4072 m .3629 .40865 L p .36909 .41194 L .265 .651 .966 r
F P
0 g
s
.38269 .41738 m .35712 .40439 L .36986 .41239 L .56 0 0 r
F
.35742 .40461 m .36986 .41239 L p .35712 .40439 L F P
0 g
s
.3824 .41797 m .38709 .41975 L .38859 .42062 L .881 .453 .167 r
F
.38269 .41738 m .36986 .41239 L .40202 .43164 L .56 0 0 r
F
.46622 .46422 m .47871 .47008 L p .39412 .42521 L .265 .651 .966 r
F P
0 g
s
.3629 .40865 m .39186 .42423 L .39412 .42521 L .265 .651 .966 r
F
.39186 .42423 m .46622 .46422 L .39412 .42521 L F
.43529 .44689 m .43463 .4467 L .47871 .47008 L F
.44835 .45756 m .41759 .45897 L p .43533 .47494 L .905 .943 .821 r
F P
0 g
s
.40998 .44012 m .4334 .45924 L .45238 .45609 L .768 .69 .776 r
F
.47871 .47008 m .47048 .45884 L p .41511 .43611 L .265 .651 .966 r
F P
0 g
s
.44372 .5043 m .4612 .4745 L .4553 .47238 L .948 .949 .642 r
F
.46533 .48344 m .45646 .48574 L .46077 .47924 L .885 .99 .745 r
F
.44532 .50251 m .46063 .47945 L .44372 .5043 L .611 .819 .414 r
F
.44372 .5043 m .4634 .47529 L p .4612 .4745 L .948 .949 .642 r
F P
0 g
s
.41759 .45897 m .40997 .45932 L p .43533 .47494 L .905 .943 .821 r
F P
0 g
s
.4406 .47969 m .41957 .4953 L .41278 .45464 L p .966 .911 .706 r
F P
0 g
s
.4784 .48944 m .42069 .4627 L p .46606 .45805 L .768 .725 .812 r
F P
0 g
s
.42069 .4627 m .41855 .4617 L p .41955 .46163 L .768 .725 .812 r
F P
0 g
s
.41955 .46163 m .46606 .45805 L .42069 .4627 L .768 .725 .812 r
F
.35799 .40502 m .38819 .42767 L .35967 .40624 L 0 .148 .597 r
F
.37113 .41502 m .38819 .42767 L p .36092 .40729 L F P
0 g
s
.37113 .41502 m .37698 .42899 L p .38039 .42205 L .956 .86 .705 r
F P
0 g
s
.37113 .41502 m .35731 .40465 L p .36949 .41973 L .854 .843 .833 r
F P
0 g
s
.37839 .42612 m .41048 .44085 L .41433 .46391 L p .879 .885 .834 r
F P
0 g
s
.41048 .44085 m .37839 .42612 L p .37918 .42452 L .941 .831 .707 r
F P
0 g
s
.37372 .42121 m .37113 .41502 L p .36992 .42026 L .854 .843 .833 r
F P
0 g
s
.41947 .46155 m .41217 .44507 L p .36972 .41495 L .84 .718 .729 r
F P
0 g
s
.41217 .44507 m .40998 .44012 L .38092 .42195 L p .84 .718 .729 r
F P
0 g
s
.36069 .4071 m .38819 .42767 L .35553 .40317 L 0 .148 .597 r
F
.34866 .39812 m .38819 .42767 L .34463 .39508 L F
.38092 .42195 m .36972 .41495 L p .41217 .44507 L .84 .718 .729 r
F P
0 g
s
.35512 .40287 m .35553 .40317 L .35466 .40252 L 0 .148 .597 r
F
.36069 .4071 m .35553 .40317 L .35512 .40287 L F
.39006 .46024 m .40358 .45261 L .39623 .44488 L .956 .86 .705 r
F
.45212 .45614 m .46606 .45805 L .41855 .4617 L .768 .725 .812 r
F
.44372 .5043 m .43228 .52189 L p .44532 .50251 L .611 .819 .414 r
F P
0 g
s
.41957 .4953 m .4406 .47969 L p .43276 .49188 L .885 .99 .745 r
F P
0 g
s
.4381 .51367 m .43504 .51798 L .46533 .48344 L 0 0 0 r
F
.46533 .48344 m .44433 .50488 L p .4381 .51367 L F P
0 g
s
.44433 .50488 m .44845 .50068 L p .45304 .49179 L 0 0 .107 r
F P
0 g
s
.46533 .48344 m .43504 .51798 L p .45646 .48574 L .885 .99 .745 r
F P
0 g
s
.49855 .5113 m .48846 .50918 L p .51121 .48994 L .768 .58 .644 r
F P
0 g
s
.47117 .55668 m .48846 .50918 L .52644 .51715 L p .76 .617 .7 r
F P
0 g
s
.50754 .57357 m .5027 .53413 L .48196 .54896 L .737 .57 .667 r
F
.54164 .57082 m .54486 .59422 L p .51784 .53793 L .76 .617 .7 r
F P
0 g
s
.51086 .52829 m .50754 .57357 L .55427 .58827 L p .794 .621 .667 r
F P
0 g
s
.53449 .51884 m .54164 .57082 L p .51784 .53793 L .76 .617 .7 r
F P
0 g
s
.48846 .50918 m .47117 .55668 L p .4614 .53651 L .709 .559 .683 r
F P
0 g
s
.44335 .51551 m .478 .52665 L p .45202 .4936 L .768 .69 .776 r
F P
0 g
s
.4784 .48944 m .48443 .52449 L p .46959 .51035 L .818 .684 .716 r
F P
0 g
s
.51338 .49381 m .49855 .5113 L .51121 .48994 L .768 .58 .644 r
F
.53732 .48539 m .54472 .49814 L .53449 .51884 L .835 .63 .625 r
F
.48443 .52449 m .4784 .48944 L .54612 .50379 L p .758 .645 .736 r
F P
0 g
s
.46959 .51035 m .42026 .46333 L .4784 .48944 L .818 .684 .716 r
F
.41855 .4617 m .42069 .4627 L p .42026 .46333 L F P
0 g
s
.42069 .4627 m .4784 .48944 L p .42026 .46333 L .818 .684 .716 r
F P
0 g
s
.43276 .49188 m .42036 .51117 L .41957 .4953 L .885 .99 .745 r
F
.44335 .51551 m .42026 .46333 L p .41106 .47676 L .84 .718 .729 r
F P
0 g
s
.45819 .56557 m .44335 .51551 L .4182 .50026 L p .872 .683 .644 r
F P
0 g
s
.4182 .50026 m .40178 .49031 L p .41644 .50987 L .872 .683 .644 r
F P
0 g
s
.41644 .50987 m .45819 .56557 L .4182 .50026 L .872 .683 .644 r
F
.44445 .57002 m .43491 .56335 L p .41537 .51572 L .892 .705 .641 r
F P
0 g
s
.41644 .50987 m .44445 .57002 L p .41537 .51572 L .892 .705 .641 r
F P
0 g
s
.44445 .57002 m .41644 .50987 L p .43645 .53655 L .818 .684 .716 r
F P
0 g
s
.45202 .4936 m .478 .52665 L .48443 .52449 L .768 .69 .776 r
F
.42026 .46333 m .44335 .51551 L p .45202 .4936 L F P
0 g
s
.54065 .54679 m .49113 .53087 L p .49517 .5455 L .796 .645 .694 r
F P
0 g
s
.58821 .64593 m .56965 .59311 L .53472 .58212 L p .778 .658 .73 r
F P
0 g
s
.53668 .59108 m .54065 .54679 L p .49517 .5455 L .796 .645 .694 r
F P
0 g
s
.49113 .53087 m .44335 .51551 L p .46641 .5569 L .796 .645 .694 r
F P
0 g
s
.50224 .62115 m .50754 .57357 L .45502 .56822 L p .731 .615 .729 r
F P
0 g
s
.5315 .60266 m .50754 .57357 L .52786 .60496 L .778 .658 .73 r
F
.50754 .57357 m .50224 .62115 L p .52786 .60496 L F P
0 g
s
.49617 .5927 m .48639 .53587 L p .52321 .57629 L .758 .645 .736 r
F P
0 g
s
.49617 .5927 m .45116 .57297 L p .48 .5426 L .818 .684 .716 r
F P
0 g
s
.478 .52665 m .54065 .54679 L p .50491 .51762 L .768 .69 .776 r
F P
0 g
s
.48443 .52449 m .49617 .5927 L p .46959 .51035 L .818 .684 .716 r
F P
0 g
s
.54486 .59422 m .55063 .63618 L .49554 .62538 L p .76 .617 .7 r
F P
0 g
s
.55063 .63618 m .52574 .65002 L .51813 .64381 L .743 .664 .773 r
F
.50125 .63004 m .49803 .65895 L .55896 .67712 L p .778 .658 .73 r
F P
0 g
s
.49617 .5927 m .52321 .57629 L .53668 .59108 L .758 .645 .736 r
F
.54751 .60297 m .52301 .59806 L p .53668 .59108 L F P
0 g
s
.52301 .59806 m .49617 .5927 L p .53668 .59108 L .758 .645 .736 r
F P
0 g
s
.59181 .52315 m .59325 .53459 L .60201 .53894 L .287 .302 .701 r
F
.60448 .55237 m .59325 .53459 L p .60761 .55219 L .426 .407 .735 r
F P
0 g
s
.59325 .53459 m .60448 .55237 L p .57783 .52695 L .235 .32 .747 r
F P
0 g
s
.6469 .54988 m .60993 .52183 L .59555 .55289 L p .424 .463 .797 r
F P
0 g
s
.58864 .56784 m .59555 .55289 L p .57294 .52452 L .287 .302 .701 r
F P
0 g
s
.59555 .55289 m .60201 .53894 L p .59325 .53459 L .287 .302 .701 r
F P
0 g
s
.59325 .53459 m .57294 .52452 L .59555 .55289 L .287 .302 .701 r
F
.60751 .56889 m .58864 .56784 L .60862 .57495 L .424 .463 .797 r
F
.58864 .56784 m .61561 .61303 L p .60862 .57495 L F P
0 g
s
.61561 .61303 m .58864 .56784 L .59081 .59701 L p .157 .321 .779 r
F P
0 g
s
.57193 .52403 m .58374 .5697 L .56355 .51987 L .235 .32 .747 r
F
.58864 .56784 m .57193 .52403 L p .57232 .52552 L 0 .005 .571 r
F P
0 g
s
.58864 .56784 m .57232 .52552 L .57426 .53305 L 0 .005 .571 r
F
.60993 .52183 m .6469 .54988 L p .64343 .51939 L .504 .467 .749 r
F P
0 g
s
.59112 .60122 m .59086 .59768 L p .57639 .55585 L 0 .005 .571 r
F P
0 g
s
.59112 .60122 m .57507 .553 L p .57539 .55381 L .889 .557 .061 r
F P
0 g
s
.5644 .52553 m .5632 .52195 L p .577 .55795 L 0 0 .414 r
F P
0 g
s
.57507 .553 m .59112 .60122 L p .57639 .55585 L 0 .005 .571 r
F P
0 g
s
.56965 .59311 m .5915 .60635 L p .58675 .59722 L .869 .664 .623 r
F P
0 g
s
.5628 .53193 m .56779 .53568 L p .58294 .57944 L .893 .686 .616 r
F P
0 g
s
.58258 .58475 m .5628 .53193 L p .58294 .57944 L .893 .686 .616 r
F P
0 g
s
.56965 .59311 m .55389 .55084 L .54478 .56569 L .794 .621 .667 r
F
.57921 .57922 m .57789 .57705 L .56217 .53732 L .82 .658 .682 r
F
.58154 .60031 m .56965 .59311 L p .595 .62187 L .85 .691 .685 r
F P
0 g
s
.5927 .63 m .59304 .6288 L .56965 .59311 L .85 .691 .685 r
F
.56382 .53464 m .58258 .58475 L p .56217 .53732 L .82 .658 .682 r
F P
0 g
s
.58258 .58475 m .57921 .57922 L .56217 .53732 L .82 .658 .682 r
F
.5927 .63 m .56965 .59311 L .59232 .63137 L .85 .691 .685 r
F
.56217 .53732 m .57789 .57705 L .56435 .55482 L .82 .658 .682 r
F
.56965 .59311 m .58821 .64593 L p .59232 .63137 L .85 .691 .685 r
F P
0 g
s
.56965 .59311 m .55792 .54426 L p .55389 .55084 L .794 .621 .667 r
F P
0 g
s
.57682 .60883 m .54751 .60297 L p .57293 .57256 L .758 .645 .736 r
F P
0 g
s
.62536 .58544 m .60448 .55237 L p .60761 .55219 L .426 .407 .735 r
F P
0 g
s
.60448 .55237 m .62536 .58544 L .61742 .62288 L p .235 .32 .747 r
F P
0 g
s
.54065 .54679 m .55216 .58317 L .59952 .55898 L .738 .593 .694 r
F
.60942 .55392 m .59502 .55243 L p .59952 .55898 L F P
0 g
s
.59502 .55243 m .54065 .54679 L p .59952 .55898 L .738 .593 .694 r
F P
0 g
s
.59564 .6196 m .58154 .60031 L .595 .62187 L .85 .691 .685 r
F
.57682 .60883 m .59952 .55898 L p .61829 .5863 L .709 .586 .715 r
F P
0 g
s
.6014 .60385 m .61325 .64252 L .59623 .60051 L .235 .32 .747 r
F
.61471 .55177 m .63235 .55074 L .62536 .58544 L .426 .407 .735 r
F
.62536 .58544 m .64773 .55716 L p .65512 .58086 L .522 .544 .819 r
F P
0 g
s
.65512 .58086 m .65608 .60407 L .62536 .58544 L .522 .544 .819 r
F
.62536 .58544 m .63242 .61923 L .63001 .61982 L .434 .524 .851 r
F
.59566 .60014 m .61325 .64252 L p .6091 .63542 L 0 0 .414 r
F P
0 g
s
.58821 .64593 m .58657 .65175 L .5735 .62837 L .82 .658 .682 r
F
.5591 .67724 m .55063 .63618 L .59681 .65619 L p .791 .695 .758 r
F P
0 g
s
.65293 .55843 m .60942 .55392 L p .60604 .56847 L .738 .593 .694 r
F P
0 g
s
.57682 .60883 m .61829 .5863 L .62756 .59979 L .709 .586 .715 r
F
.63404 .60922 m .61045 .60906 L p .62756 .59979 L F P
0 g
s
.61045 .60906 m .57682 .60883 L p .62756 .59979 L .709 .586 .715 r
F P
0 g
s
.64773 .55716 m .67258 .52575 L p .67161 .55015 L .522 .544 .819 r
F P
0 g
s
.67709 .57279 m .71955 .53609 L p .6642 .55233 L .504 .467 .749 r
F P
0 g
s
.65293 .55843 m .69788 .55459 L p .69599 .52589 L .683 .578 .731 r
F P
0 g
s
.69788 .55459 m .71351 .55326 L p .71066 .5488 L .683 .578 .731 r
F P
0 g
s
.69599 .52589 m .69788 .55459 L .71066 .5488 L .683 .578 .731 r
F
.73849 .54334 m .7388 .54314 L p .74706 .53822 L 0 0 0 r
F P
0 g
s
.72069 .55527 m .73849 .54334 L p .74706 .53822 L 0 0 0 r
F P
0 g
s
.77987 .517 m .7388 .54314 L p .79478 .50618 L .4 .221 .52 r
F P
0 g
s
.78902 .50935 m .7388 .54314 L .77966 .5145 L .4 .221 .52 r
F
.76789 .52475 m .74706 .53822 L .74356 .54031 L 0 0 0 r
F
.7388 .54314 m .77885 .51764 L p .74356 .54031 L F P
0 g
s
.77885 .51764 m .7956 .50698 L p .74356 .54031 L 0 0 0 r
F P
0 g
s
.81556 .49701 m .79586 .50672 L .7956 .50698 L 0 0 0 r
F
.81768 .49589 m .79586 .50672 L p .81556 .49701 L F P
0 g
s
.78582 .51268 m .78894 .51113 L .77987 .517 L .714 .94 .959 r
F
.7956 .50698 m .78397 .51438 L p .79047 .50945 L .4 .221 .52 r
F P
0 g
s
.79549 .5071 m .74356 .54031 L .7956 .50698 L 0 0 0 r
F
.74903 .53098 m .78582 .51268 L .7377 .54427 L .714 .94 .959 r
F
.78582 .51268 m .77987 .517 L .7377 .54427 L F
.78397 .51438 m .77987 .517 L p .78686 .51193 L .4 .221 .52 r
F P
0 g
s
.79478 .50618 m .78397 .51438 L .78686 .51193 L .4 .221 .52 r
F
.72069 .55527 m .72025 .55557 L p .71806 .56195 L .766 .985 .904 r
F P
0 g
s
.72025 .55557 m .70377 .56988 L .71806 .56195 L .766 .985 .904 r
F
.72027 .55555 m .67801 .59166 L p .71812 .55826 L 0 0 0 r
F P
0 g
s
.72025 .55557 m .67543 .58561 L p .70377 .56988 L .766 .985 .904 r
F P
0 g
s
.65827 .61071 m .71812 .55826 L .69458 .57787 L 0 0 0 r
F
.67186 .59759 m .69458 .57787 L .67801 .59166 L F
.65827 .61071 m .69458 .57787 L .67186 .59759 L F
.65681 .63182 m .64895 .64238 L .65685 .63213 L .852 .941 .88 r
F
.65722 .63164 m .65681 .63182 L p .65685 .63213 L F P
0 g
s
.67854 .6013 m .65681 .63182 L .67271 .61649 L .766 .985 .904 r
F
.65722 .63143 m .65685 .63213 L p .65635 .61106 L .835 .999 .804 r
F P
0 g
s
.65685 .63213 m .63808 .66804 L p .64442 .64826 L .835 .999 .804 r
F P
0 g
s
.65635 .61106 m .65685 .63213 L .64442 .64826 L .835 .999 .804 r
F
.66877 .62676 m .65722 .63164 L p .65792 .64157 L .852 .941 .88 r
F P
0 g
s
.65722 .63164 m .65685 .63213 L .65792 .64157 L .852 .941 .88 r
F
.70255 .57283 m .62498 .65401 L .65447 .61403 L .153 0 .16 r
F
.69815 .5803 m .62498 .65401 L .70255 .57283 L F
.67543 .58561 m .65722 .61031 L p .70377 .56988 L .766 .985 .904 r
F P
0 g
s
.67543 .58561 m .65006 .61652 L p .65009 .61649 L .766 .985 .904 r
F P
0 g
s
.65635 .61106 m .64953 .61717 L p .67186 .59759 L 0 0 0 r
F P
0 g
s
.67186 .59759 m .64953 .61717 L .64874 .61814 L 0 0 0 r
F
.59545 .70124 m .5528 .7809 L .57175 .73195 L .284 .041 .358 r
F
.59667 .6978 m .6225 .66116 L .57967 .71981 L 0 0 .083 r
F
.57118 .7351 m .6225 .66116 L .57146 .73355 L F
.57067 .73794 m .59059 .70843 L p .57118 .7351 L F P
0 g
s
.59059 .70843 m .6225 .66116 L p .57118 .7351 L 0 0 .083 r
F P
0 g
s
.57967 .71981 m .6225 .66116 L .61808 .65551 L 0 0 .083 r
F
.57146 .73355 m .61808 .65551 L .59853 .67933 L F
.6463 .62112 m .62709 .64453 L .62349 .64891 L 0 0 0 r
F
.6225 .66116 m .62913 .64204 L .61808 .65551 L 0 0 .083 r
F
.64753 .61962 m .67186 .59759 L .64874 .61814 L 0 0 0 r
F
.62913 .64204 m .64614 .62131 L p .62498 .65401 L .153 0 .16 r
F P
0 g
s
.62498 .65401 m .64614 .62131 L .65447 .61403 L .153 0 .16 r
F
.75871 .53128 m .71895 .55722 L .72069 .55527 L 0 0 0 r
F
.73863 .54066 m .71951 .53634 L p .73517 .53065 L .554 .503 .754 r
F P
0 g
s
.74903 .53098 m .7377 .54427 L .72069 .55527 L .714 .94 .959 r
F
.75871 .53128 m .72069 .55527 L .74706 .53822 L 0 0 0 r
F
.79412 .50847 m .81599 .49717 L p .7956 .50698 L F P
0 g
s
.81599 .49717 m .83825 .48567 L p .7956 .50698 L 0 0 0 r
F P
0 g
s
.81192 .49661 m .77987 .5175 L .81007 .49777 L 0 0 0 r
F
.75302 .53502 m .67543 .58561 L .75371 .53454 L F
.67709 .57279 m .65473 .57154 L .65557 .59167 L .424 .463 .797 r
F
.6481 .60931 m .63404 .60922 L p .67752 .57274 L .709 .586 .715 r
F P
0 g
s
.70286 .55417 m .65293 .55843 L p .64423 .59077 L .676 .511 .656 r
F P
0 g
s
.64423 .59077 m .6893 .56637 L .70286 .55417 L .676 .511 .656 r
F
.67752 .57274 m .69002 .56598 L .6481 .60931 L .709 .586 .715 r
F
.77987 .5175 m .75999 .5308 L .67543 .58561 L 0 0 0 r
F
.70676 .5657 m .67543 .58561 L p .7442 .54104 L F P
0 g
s
.77253 .52268 m .73838 .54501 L .7442 .54104 L 0 0 0 r
F
.73838 .54501 m .70676 .5657 L .7442 .54104 L F
.69589 .58412 m .72859 .55142 L .70355 .57113 L .299 0 .256 r
F
.71812 .55826 m .70355 .57113 L p .72859 .55142 L F P
0 g
s
.67436 .60503 m .59225 .68699 L .67952 .59907 L .153 0 .16 r
F
.64709 .62014 m .61535 .65884 L .59279 .68632 L 0 0 0 r
F
.6463 .62112 m .58943 .69042 L .64709 .62014 L F
.64614 .62131 m .62913 .64204 L p .6463 .62112 L F P
0 g
s
.62913 .64204 m .62709 .64453 L .6463 .62112 L 0 0 0 r
F
.64861 .6183 m .59279 .68632 L .61396 .6605 L F
.65827 .61071 m .67186 .59759 L .64753 .61962 L F
.65827 .61071 m .64753 .61962 L .64614 .62131 L F
.67543 .58561 m .64709 .62014 L .64861 .6183 L F
.64709 .62014 m .59279 .68632 L .64861 .6183 L F
.65626 .60897 m .67543 .58561 L p .61396 .6605 L F P
0 g
s
.64614 .62131 m .64753 .61962 L .63204 .63363 L 0 0 0 r
F
.64534 .62229 m .63204 .63363 L p .62913 .64204 L 0 0 .083 r
F P
0 g
s
.61535 .65884 m .58943 .69042 L .59147 .68794 L 0 0 0 r
F
.61535 .65884 m .59147 .68794 L .59279 .68632 L F
.63204 .63363 m .64534 .62229 L p .64614 .62131 L F P
0 g
s
.63204 .63363 m .61341 .64953 L p .61808 .65551 L 0 0 .083 r
F P
0 g
s
.62913 .64204 m .63204 .63363 L .61808 .65551 L 0 0 .083 r
F
.62913 .64204 m .59225 .68699 L p .62709 .64453 L 0 0 0 r
F P
0 g
s
.59225 .68699 m .58943 .69042 L p .62349 .64891 L 0 0 0 r
F P
0 g
s
.62709 .64453 m .59225 .68699 L .62349 .64891 L 0 0 0 r
F
.61341 .64953 m .63204 .63363 L p .64753 .61962 L F P
0 g
s
.64753 .61962 m .64874 .61814 L .61341 .64953 L 0 0 0 r
F
.59225 .68699 m .62913 .64204 L p .62498 .65401 L .153 0 .16 r
F P
0 g
s
.64953 .61717 m .61341 .64953 L p .64874 .61814 L 0 0 0 r
F P
0 g
s
.6481 .60931 m .62222 .65047 L p .61813 .65508 L .707 .533 .651 r
F P
0 g
s
.61692 .64787 m .6481 .60931 L .61813 .65508 L .707 .533 .651 r
F
.59644 .68186 m .6074 .6685 L p .59279 .68632 L 0 0 0 r
F P
0 g
s
.59279 .68632 m .6074 .6685 L .61396 .6605 L 0 0 0 r
F
.59853 .67933 m .61808 .65551 L .61341 .64953 L 0 0 .083 r
F
.63244 .66381 m .64442 .64826 L .63808 .66804 L .835 .999 .804 r
F
.61909 .68112 m .61483 .70041 L p .64895 .64238 L .852 .941 .88 r
F P
0 g
s
.61749 .67939 m .62261 .67656 L p .60491 .70159 L .942 .938 .76 r
F P
0 g
s
.62261 .67656 m .61749 .67939 L p .63244 .66381 L .835 .999 .804 r
F P
0 g
s
.71375 .57288 m .67854 .6013 L .67271 .61649 L .766 .985 .904 r
F
.7144 .57123 m .67854 .6013 L .71375 .57288 L F
.69671 .58273 m .67849 .60459 L .59225 .68699 L .153 0 .16 r
F
.71637 .56364 m .72212 .55944 L .74903 .53098 L .299 0 .256 r
F
.81138 .49696 m .77875 .51856 L .75999 .5308 L 0 0 0 r
F
.81007 .49783 m .77253 .52268 L p .81138 .49696 L F P
0 g
s
.81193 .49659 m .75999 .5308 L .77987 .5175 L 0 0 0 r
F
.81244 .49626 m .81138 .49696 L .81193 .49659 L F
.81138 .49696 m .75999 .5308 L .81193 .49659 L F
.70728 .57515 m .73088 .55655 L .73385 .5494 L .299 0 .256 r
F
.77987 .5175 m .7816 .51635 L .81007 .49777 L 0 0 0 r
F
.81007 .49777 m .75871 .53128 L .74706 .53822 L F
.81007 .49777 m .74706 .53822 L .76789 .52475 L F
.78697 .51214 m .7834 .51475 L .79434 .50824 L .769 .792 .872 r
F
.77875 .51856 m .77253 .52268 L .76883 .52508 L 0 0 0 r
F
.77875 .51856 m .76883 .52508 L .75999 .5308 L F
.77987 .5175 m .75302 .53502 L .7816 .51635 L F
.75302 .53502 m .75371 .53454 L .7816 .51635 L F
.77253 .52268 m .77886 .51848 L p .7834 .51475 L .769 .792 .872 r
F P
0 g
s
.77886 .51848 m .79434 .50824 L p .7834 .51475 L .769 .792 .872 r
F P
0 g
s
.73088 .55655 m .7877 .51178 L .73385 .5494 L .299 0 .256 r
F
.8122 .49375 m .81244 .49626 L .81007 .49777 L p .4 .221 .52 r
F P
0 g
s
.81244 .49626 m .81125 .49458 L p .80762 .49708 L .769 .792 .872 r
F P
0 g
s
.78697 .51214 m .75691 .52706 L p .78217 .51564 L .299 0 .256 r
F P
0 g
s
.74903 .53098 m .78988 .50974 L p .78582 .51268 L .714 .94 .959 r
F P
0 g
s
.74903 .53098 m .78697 .51214 L p .79144 .50899 L 0 0 0 r
F P
0 g
s
.78697 .51214 m .7877 .51178 L p .78781 .5117 L 0 0 0 r
F P
0 g
s
.78781 .5117 m .79144 .50899 L .78697 .51214 L 0 0 0 r
F
.73385 .5494 m .78217 .51564 L .75691 .52706 L .299 0 .256 r
F
.75691 .52706 m .74903 .53098 L p .70728 .57515 L F P
0 g
s
.70355 .57113 m .71812 .55826 L p .65827 .61071 L 0 0 0 r
F P
0 g
s
.77253 .52268 m .72608 .55341 L .70676 .5657 L p 0 0 0 r
F P
0 g
s
.74903 .53098 m .71812 .55826 L p .72859 .55142 L .299 0 .256 r
F P
0 g
s
.72518 .55226 m .72608 .55341 L p .78115 .51537 L .769 .792 .872 r
F P
0 g
s
.78063 .51568 m .72518 .55226 L .78115 .51537 L .769 .792 .872 r
F
.78988 .50974 m .74903 .53098 L p .79144 .50899 L 0 0 0 r
F P
0 g
s
.74059 .51985 m .78681 .51193 L p .77701 .51788 L .644 .482 .649 r
F P
0 g
s
.77146 .52124 m .72518 .55226 L p .78063 .51568 L .769 .792 .872 r
F P
0 g
s
.72608 .55341 m .77253 .52268 L p .78302 .51423 L .769 .792 .872 r
F P
0 g
s
.78302 .51423 m .77253 .52268 L .78697 .51214 L .769 .792 .872 r
F
.72025 .55557 m .72069 .55527 L p .71895 .55722 L 0 0 0 r
F P
0 g
s
.72027 .55555 m .72075 .55514 L p .72069 .55527 L .714 .94 .959 r
F P
0 g
s
.72075 .55514 m .74903 .53098 L p .72069 .55527 L .714 .94 .959 r
F P
0 g
s
.71812 .55826 m .74903 .53098 L .72027 .55555 L p 0 0 0 r
F P
0 g
s
.78115 .51537 m .72608 .55341 L .78302 .51423 L .769 .792 .872 r
F
.71066 .5488 m .74059 .51985 L p .77913 .51776 L .644 .482 .649 r
F P
0 g
s
.77778 .51701 m .77146 .52124 L p .78204 .51475 L .769 .792 .872 r
F P
0 g
s
.77146 .52124 m .78063 .51568 L .78204 .51475 L .769 .792 .872 r
F
.78902 .50935 m .78809 .5106 L .78484 .51216 L .4 .221 .52 r
F
.78681 .51193 m .74059 .51985 L p .79096 .50222 L .662 .568 .739 r
F P
0 g
s
.79321 .50667 m .81125 .49458 L .78811 .49661 L p .683 .578 .731 r
F P
0 g
s
.80425 .49757 m .79321 .50667 L .79096 .50222 L .662 .568 .739 r
F
.78902 .50935 m .79478 .50618 L .78809 .5106 L .4 .221 .52 r
F
.79321 .50667 m .78681 .51193 L .79096 .50222 L .662 .568 .739 r
F
.74838 .49444 m .81015 .49211 L .81039 .49209 L .27 0 0 r
F
.74831 .49445 m .81015 .49211 L .74838 .49444 L F
.78022 .49645 m .8058 .49492 L .78702 .49544 L .65 .401 .53 r
F
.74059 .51985 m .78022 .49645 L p .79684 .49678 L .662 .568 .739 r
F P
0 g
s
.77914 .49661 m .78133 .49647 L .78022 .49645 L .65 .401 .53 r
F
.78133 .49647 m .8058 .49492 L .78022 .49645 L F
.77914 .49661 m .78702 .49544 L .78015 .49563 L F
.78781 .4941 m .81071 .49209 L p .77289 .49352 L .27 0 0 r
F P
0 g
s
.78739 .49538 m .81125 .49458 L .79504 .49424 L .65 .401 .53 r
F
.78739 .49538 m .79504 .49424 L .7879 .4941 L F
.8058 .49492 m .81125 .49458 L .78015 .49563 L F
.8085 .49206 m .79192 .49202 L p .81041 .49208 L .27 0 0 r
F P
0 g
s
.7082 .50111 m .70835 .50362 L .78811 .49661 L p .65 .401 .53 r
F P
0 g
s
.78811 .49661 m .81125 .49458 L p .78133 .49647 L .65 .401 .53 r
F P
0 g
s
.7082 .50111 m .78811 .49661 L .78133 .49647 L .65 .401 .53 r
F
.80425 .49757 m .74059 .51985 L .79684 .49678 L .662 .568 .739 r
F
.81015 .49211 m .81071 .49209 L .81039 .49209 L .27 0 0 r
F
.81041 .49208 m .79192 .49202 L .7919 .49204 L F
.81041 .49208 m .77572 .49199 L .77506 .492 L F
.81071 .49209 m .81101 .49206 L p .81041 .49208 L F P
0 g
s
.83825 .48567 m .79463 .49162 L .81934 .49112 L .585 .525 .754 r
F
.81101 .49206 m .8085 .49206 L p .81041 .49208 L .27 0 0 r
F P
0 g
s
.805 .49695 m .79684 .49678 L .81349 .48995 L .662 .568 .739 r
F
.80425 .49757 m .79684 .49678 L .805 .49695 L F
.782 .4954 m .78706 .49453 L p .79684 .49678 L F P
0 g
s
.78706 .49453 m .81349 .48995 L p .8056 .49319 L .662 .568 .739 r
F P
0 g
s
.8056 .49319 m .79684 .49678 L .78706 .49453 L .662 .568 .739 r
F
.78022 .49645 m .782 .4954 L p .79684 .49678 L F P
0 g
s
.81125 .49458 m .79321 .50667 L p .79041 .50897 L .769 .792 .872 r
F P
0 g
s
.81244 .49626 m .8122 .49375 L .81101 .49206 L p .974 .666 .318 r
F P
0 g
s
.81101 .49206 m .81125 .49458 L .81244 .49626 L p .974 .666 .318 r
F P
0 g
s
.81349 .48995 m .83825 .48567 L p .80425 .49757 L .662 .568 .739 r
F P
0 g
s
.79041 .50897 m .77778 .51701 L .78204 .51475 L .769 .792 .872 r
F
.79321 .50667 m .77778 .51701 L p .79041 .50897 L F P
0 g
s
.72518 .55226 m .79321 .50667 L p .76628 .49853 L .683 .578 .731 r
F P
0 g
s
.78811 .49661 m .76628 .49853 L p .79321 .50667 L .683 .578 .731 r
F P
0 g
s
.73373 .54855 m .74459 .53754 L .76102 .53042 L .644 .482 .649 r
F
.74459 .53754 m .75192 .5301 L .75658 .53026 L F
.74459 .53754 m .75658 .53026 L .76102 .53042 L F
.76102 .53042 m .75658 .53026 L .77913 .51776 L F
.75658 .53026 m .75192 .5301 L .77146 .52124 L F
.75658 .53026 m .77146 .52124 L .77913 .51776 L F
.78204 .51475 m .78063 .51568 L .78115 .51537 L .769 .792 .872 r
F
.79478 .50618 m .79626 .50537 L .79493 .50633 L .4 .221 .52 r
F
.79626 .50537 m .81007 .49777 L .79493 .50633 L F
.81007 .49777 m .7956 .50698 L p .79493 .50633 L F P
0 g
s
.79586 .50672 m .7917 .50879 L p .79191 .50876 L 0 0 0 r
F P
0 g
s
.7956 .50698 m .79191 .50876 L .79412 .50847 L 0 0 0 r
F
.79586 .50672 m .79191 .50876 L .7956 .50698 L F
.79478 .50618 m .7956 .50698 L .79047 .50945 L .4 .221 .52 r
F
.80762 .49708 m .78204 .51475 L .79017 .50981 L .769 .792 .872 r
F
.7877 .51178 m .79412 .50847 L p .7917 .50879 L 0 0 0 r
F P
0 g
s
.80762 .49708 m .79017 .50981 L .81244 .49626 L .769 .792 .872 r
F
.7877 .51178 m .78697 .51214 L p .78217 .51564 L .299 0 .256 r
F P
0 g
s
.69589 .58412 m .70728 .57515 L .72212 .55944 L .299 0 .256 r
F
.69589 .58412 m .72212 .55944 L .71637 .56364 L F
.69389 .58751 m .67849 .60459 L .69671 .58273 L .153 0 .16 r
F
.70578 .58141 m .7062 .58125 L p .69776 .59234 L .763 .892 .935 r
F P
0 g
s
.70578 .58141 m .69776 .59234 L .69688 .5932 L .763 .892 .935 r
F
.71362 .5732 m .70156 .58699 L p .70578 .58141 L .176 .06 .473 r
F P
0 g
s
.71375 .57288 m .67271 .61649 L .69167 .59822 L .766 .985 .904 r
F
.71375 .57288 m .69167 .59822 L .71362 .5732 L F
.69176 .59113 m .7877 .51178 L p .70728 .57515 L .299 0 .256 r
F P
0 g
s
.69306 .58892 m .70728 .57515 L .69589 .58412 L .299 0 .256 r
F
.69176 .59113 m .70728 .57515 L .69306 .58892 L F
.72608 .55341 m .70043 .58027 L .71167 .5678 L .153 0 .16 r
F
.72608 .55341 m .67934 .60367 L p .70043 .58027 L F P
0 g
s
.67513 .60488 m .68508 .59665 L p .69589 .58412 L .299 0 .256 r
F P
0 g
s
.67849 .60459 m .6544 .63049 L p .62672 .6571 L .153 0 .16 r
F P
0 g
s
.67849 .60459 m .62672 .6571 L .59225 .68699 L .153 0 .16 r
F
.60169 .68117 m .59225 .68699 L p .62672 .6571 L F P
0 g
s
.65336 .61545 m .63386 .64434 L p .62565 .65813 L .391 .139 .408 r
F P
0 g
s
.61776 .6714 m .62707 .65677 L .62565 .65813 L .391 .139 .408 r
F
.62707 .65677 m .65336 .61545 L .62565 .65813 L F
.63386 .64434 m .59545 .70124 L p .61776 .6714 L F P
0 g
s
.60503 .67911 m .60169 .68117 L p .62672 .6571 L .153 0 .16 r
F P
0 g
s
.61478 .6731 m .60503 .67911 L p .6255 .65834 L .153 0 .16 r
F P
0 g
s
.59225 .68699 m .61478 .6731 L p .60788 .68513 L .284 .041 .358 r
F P
0 g
s
.61478 .6731 m .61433 .67261 L p .61192 .67703 L .909 .796 .724 r
F P
0 g
s
.6544 .63049 m .67849 .60459 L p .67785 .60519 L .846 .805 .811 r
F P
0 g
s
.67011 .61254 m .66548 .61783 L .67785 .60519 L .846 .805 .811 r
F
.66548 .61783 m .6544 .63049 L .67785 .60519 L F
.55966 .77217 m .56061 .77041 L p .55042 .78585 L .998 .825 .484 r
F P
0 g
s
.56061 .77041 m .55917 .7723 L .56095 .76895 L .997 .8 .518 r
F
.55042 .78585 m .56061 .77041 L .56124 .76767 L .998 .825 .484 r
F
.57877 .73701 m .55824 .77586 L .5499 .78826 L .284 .041 .358 r
F
.56676 .76011 m .55824 .77586 L .57877 .73701 L F
.60188 .71549 m .5877 .72173 L .57911 .74201 L .99 .868 .614 r
F
.59816 .71351 m .5877 .72173 L p .60188 .71549 L F P
0 g
s
.59043 .71735 m .55824 .77586 L p .59116 .71503 L .284 .041 .358 r
F P
0 g
s
.5474 .79513 m .54778 .79406 L .53171 .83822 L .144 0 .315 r
F
.5474 .79513 m .53171 .83822 L .5317 .83826 L F
.53171 .83823 m .5317 .83826 L .53168 .83833 L F
.5474 .79513 m .5317 .83826 L .53171 .83823 L F
.52438 .85853 m .54553 .80031 L p .52438 .85837 L F P
0 g
s
.54553 .80031 m .54548 .8004 L .52438 .85837 L .144 0 .315 r
F
.53156 .84238 m .55971 .77368 L p .54374 .80604 L .424 .201 .468 r
F P
0 g
s
.55971 .77368 m .55888 .77492 L .54374 .80604 L .424 .201 .468 r
F
.56095 .77066 m .55966 .77315 L .56265 .76771 L .909 .796 .724 r
F
.5499 .78826 m .54778 .79406 L .5474 .79513 L .144 0 .315 r
F
.59226 .71149 m .59045 .71729 L .61551 .67493 L .391 .139 .408 r
F
.61478 .6731 m .5499 .78826 L .55017 .78581 L .284 .041 .358 r
F
.5849 .72521 m .55017 .78581 L .60132 .69363 L F
.56095 .77066 m .57243 .74265 L p .55173 .78838 L .424 .201 .468 r
F P
0 g
s
.55017 .78581 m .55094 .77869 L .57175 .73195 L .284 .041 .358 r
F
.55982 .76096 m .5598 .761 L .54991 .78818 L .144 0 .315 r
F
.54991 .7882 m .54991 .78818 L .52438 .85837 L F
.55982 .76096 m .54991 .78818 L .54991 .7882 L F
.54271 .80476 m .54968 .78863 L .54716 .79036 L F
.54304 .80565 m .54992 .78808 L .53297 .82728 L F
.54091 .81279 m .54992 .78808 L .52915 .8411 L F
.5246 .85777 m .5598 .761 L .52467 .85733 L F
.5499 .78826 m .54024 .80858 L .54224 .79964 L .284 .041 .358 r
F
.55109 .77723 m .57175 .73195 L p .55094 .77869 L F P
0 g
s
.57211 .72991 m .54997 .77933 L .54162 .79491 L 0 0 .028 r
F
.5598 .761 m .56528 .74593 L .55135 .77483 L .144 0 .315 r
F
.55109 .77723 m .55135 .77483 L .54223 .79376 L F
.5598 .761 m .55135 .77483 L .55109 .77723 L F
.54054 .80037 m .55109 .77723 L p .54037 .80241 L .284 .041 .358 r
F P
0 g
s
.54037 .80241 m .55109 .77723 L .55094 .77869 L .284 .041 .358 r
F
.53941 .80458 m .53974 .80336 L .54037 .80241 L F
.53974 .80336 m .54054 .80037 L .54037 .80241 L F
.57243 .74265 m .56528 .74593 L p .55035 .7876 L .424 .201 .468 r
F P
0 g
s
.54638 .79867 m .55417 .78192 L .55035 .7876 L .424 .201 .468 r
F
.55417 .78192 m .57243 .74265 L .55035 .7876 L F
.54374 .80604 m .55173 .78838 L .54638 .79867 L F
.55173 .78838 m .57243 .74265 L .54638 .79867 L F
.53076 .84499 m .5316 .84178 L .53081 .84483 L .339 .132 .446 r
F
.53176 .83947 m .53428 .83243 L .53185 .83811 L .424 .201 .468 r
F
.53428 .83243 m .55035 .7876 L .5355 .82816 L F
.53428 .83243 m .5355 .82816 L .53185 .83811 L F
.53076 .84499 m .53187 .83791 L .5316 .84178 L .339 .132 .446 r
F
.52014 .88579 m .52749 .85577 L .53076 .84499 L F
.52749 .85577 m .53187 .83791 L .53076 .84499 L F
.53428 .83243 m .53156 .84238 L .54374 .80604 L .424 .201 .468 r
F
.53176 .83947 m .53159 .84186 L .53428 .83243 L F
.53159 .84186 m .53156 .84238 L .53428 .83243 L F
.5499 .78826 m .53187 .83791 L p .55035 .7876 L F P
0 g
s
.53187 .83791 m .52422 .85897 L p .53185 .83811 L .424 .201 .468 r
F P
0 g
s
.55035 .7876 m .53187 .83791 L .53185 .83811 L .424 .201 .468 r
F
.53162 .84144 m .53187 .83791 L .53238 .83879 L .339 .132 .446 r
F
.53238 .83879 m .53187 .83791 L .53348 .83456 L F
.5335 .83451 m .52018 .88577 L p .53348 .83456 L F P
0 g
s
.5316 .84178 m .5316 .84181 L .53081 .84483 L .339 .132 .446 r
F
.52018 .88577 m .52014 .88579 L p .53076 .84499 L F P
0 g
s
.53076 .84499 m .53081 .84483 L .52018 .88577 L .339 .132 .446 r
F
.52018 .88577 m .53159 .84186 L p .53083 .84454 L .951 .759 .609 r
F P
0 g
s
.52487 .85868 m .5312 .84326 L p .53176 .83947 L .424 .201 .468 r
F P
0 g
s
.5312 .84326 m .53156 .84238 L p .53159 .84186 L .424 .201 .468 r
F P
0 g
s
.5312 .84326 m .53159 .84186 L .53176 .83947 L .424 .201 .468 r
F
.52018 .88577 m .52018 .88577 L p .53083 .84454 L .951 .759 .609 r
F P
0 g
s
.53083 .84454 m .53153 .84183 L .53083 .84454 L .951 .759 .609 r
F
.53083 .84454 m .53083 .84454 L .52018 .88577 L F
.53156 .84238 m .5312 .84326 L p .53164 .84211 L .625 .411 .571 r
F P
0 g
s
.5312 .84326 m .52779 .85157 L p .53123 .84311 L .625 .411 .571 r
F P
0 g
s
.53164 .84211 m .5312 .84326 L .53123 .84311 L .625 .411 .571 r
F
.533 .83614 m .53153 .84181 L p .533 .83613 L .951 .759 .609 r
F P
0 g
s
.53153 .84182 m .53123 .84311 L .53114 .84332 L .625 .411 .571 r
F
.53153 .84182 m .53114 .84332 L .52779 .85157 L F
.53153 .84181 m .533 .83614 L p .53148 .84046 L .661 .439 .572 r
F P
0 g
s
.53151 .84111 m .53153 .84182 L .52779 .85157 L .625 .411 .571 r
F
.53207 .83953 m .53153 .84181 L .53151 .84111 L F
.53153 .84181 m .53153 .84182 L .53151 .84111 L F
.53153 .84181 m .53082 .84456 L p .53153 .84182 L .951 .759 .609 r
F P
0 g
s
.533 .83613 m .53153 .84181 L .53153 .84182 L .951 .759 .609 r
F
.53082 .84456 m .52018 .88577 L p .53083 .84455 L F P
0 g
s
.533 .83613 m .53082 .84456 L .53083 .84455 L .951 .759 .609 r
F
.5316 .84178 m .53162 .84144 L .53238 .83879 L .339 .132 .446 r
F
.53137 .84253 m .5335 .83451 L .53153 .84183 L .951 .759 .609 r
F
.5316 .84178 m .53348 .83456 L .5316 .84181 L .339 .132 .446 r
F
.53159 .84186 m .5335 .83451 L p .53083 .84454 L .951 .759 .609 r
F P
0 g
s
.53153 .84183 m .533 .83613 L .53153 .84183 L .951 .759 .609 r
F
.53207 .83953 m .53336 .83643 L .53123 .84311 L .625 .411 .571 r
F
.53336 .83643 m .55866 .77571 L .53164 .84211 L F
.53336 .83643 m .53164 .84211 L .53123 .84311 L F
.53153 .84183 m .5335 .83451 L .533 .83613 L .951 .759 .609 r
F
.52467 .85733 m .54091 .81279 L .52915 .8411 L .144 0 .315 r
F
.53207 .83953 m .54208 .80769 L .53336 .83643 L .625 .411 .571 r
F
.52487 .85868 m .53176 .83947 L .52751 .85045 L .424 .201 .468 r
F
.52422 .85897 m .52487 .85868 L p .52751 .85045 L F P
0 g
s
.52047 .8822 m .52487 .85868 L .52059 .88152 L .44 .23 .497 r
F
.52047 .8822 m .52059 .88152 L .52009 .88421 L F
.52176 .87539 m .52187 .87469 L .52172 .87563 L 0 0 0 r
F
.52487 .85868 m .52459 .8588 L p .52059 .88152 L .44 .23 .497 r
F P
0 g
s
.52459 .8588 m .52422 .85897 L p .52059 .88152 L .44 .23 .497 r
F P
0 g
s
.52023 .88334 m .52422 .85897 L .52013 .88397 L .44 .23 .497 r
F
.52017 .8838 m .52002 .88475 L .52 .88498 L .773 .533 .571 r
F
.52422 .85897 m .52303 .85989 L p .52016 .88352 L .212 .036 .411 r
F P
0 g
s
.52001 .88449 m .52422 .85897 L .52002 .88441 L .212 .036 .411 r
F
.52017 .88373 m .52002 .88474 L .52002 .88475 L .773 .533 .571 r
F
.52002 .88474 m .52 .88486 L .52002 .88475 L F
.52 .88485 m .52001 .88483 L .52003 .88456 L .44 .23 .497 r
F
.52001 .88483 m .52002 .88474 L .52003 .88456 L F
.52002 .88477 m .52047 .8822 L .52004 .88449 L F
.52002 .88466 m .52004 .88449 L .52003 .88456 L F
.52003 .8845 m .52002 .88467 L .52003 .8846 L .85 .596 .554 r
F
.52002 .88477 m .52004 .88449 L .52002 .88466 L .44 .23 .497 r
F
.52003 .8845 m .52003 .88463 L .52003 .88463 L .85 .596 .554 r
F
.52004 .88419 m .52003 .8845 L .52003 .8846 L F
.52 .88486 m .52001 .88483 L .52 .88485 L .44 .23 .497 r
F
.52013 .8835 m .5201 .88376 L .52002 .88447 L .212 .036 .411 r
F
.52057 .87771 m .52 .88486 L .52002 .88474 L .773 .533 .571 r
F
.52001 .88483 m .52002 .88477 L .52002 .88474 L .44 .23 .497 r
F
.52002 .88477 m .52487 .85868 L .52047 .8822 L F
.52001 .88491 m .52 .88486 L .52 .88493 L .698 .467 .569 r
F
.52001 .88488 m .52 .88486 L .52001 .88491 L F
.52002 .88527 m .52 .88486 L .51998 .88508 L .649 .393 .518 r
F
.52004 .88538 m .52 .88486 L .52001 .88502 L .943 .682 .216 r
F
.52009 .88588 m .52 .88486 L p .52004 .88538 L F P
0 g
s
.52 .88496 m .52 .88486 L .52003 .88522 L .943 .682 .216 r
F
.52001 .88502 m .52001 .88501 L .52003 .88522 L F
.52003 .88522 m .52003 .88522 L .52005 .88542 L F
.52001 .88502 m .52003 .88522 L .52003 .88522 L F
.52001 .88502 m .52005 .88542 L .52005 .88544 L F
.52011 .88582 m .5201 .88583 L p .52015 .88558 L 0 0 .274 r
F P
0 g
s
.52009 .8858 m .52008 .8857 L .52009 .88579 L .943 .682 .216 r
F
.52009 .88588 m .52008 .8857 L .52009 .8858 L F
.52009 .88579 m .5201 .88588 L p .52009 .88588 L F P
0 g
s
.52009 .88588 m .52009 .88588 L p .52009 .88579 L .943 .682 .216 r
F P
0 g
s
.52009 .88588 m .5201 .88583 L p .52005 .88523 L 0 0 .265 r
F P
0 g
s
.52 .88486 m .52014 .88579 L p .52008 .88525 L .391 .225 .533 r
F P
0 g
s
.52002 .88466 m .52 .88486 L .52008 .88525 L .391 .225 .533 r
F
.52 .88486 m .5201 .88583 L p .52005 .88517 L .141 .057 .492 r
F P
0 g
s
.52078 .88028 m .52 .88486 L .52023 .88356 L .698 .467 .569 r
F
.5206 .88148 m .52018 .88383 L .52332 .86351 L F
.52018 .88383 m .52 .88486 L .52332 .86351 L F
.52332 .86351 m .52 .88486 L .52404 .8588 L F
.52023 .88356 m .52 .88486 L .52001 .88488 L F
.52 .88486 m .52487 .85868 L p .52001 .88483 L .44 .23 .497 r
F P
0 g
s
.52404 .8588 m .52 .88486 L p .52078 .88028 L .698 .467 .569 r
F P
0 g
s
.51999 .88509 m .5233 .86357 L .52487 .85868 L .62 .394 .553 r
F
.52018 .88577 m .52 .88486 L p .52002 .88466 L .391 .225 .533 r
F P
0 g
s
.52002 .88458 m .52 .88486 L .52005 .88517 L .141 .057 .492 r
F
.52014 .88579 m .52018 .88577 L p .52002 .88466 L .391 .225 .533 r
F P
0 g
s
.52 .88486 m .52018 .88577 L p .52012 .88558 L .649 .393 .518 r
F P
0 g
s
.5201 .88583 m .52014 .88579 L p .5201 .88545 L .141 .057 .492 r
F P
0 g
s
.5201 .88545 m .52005 .88517 L .5201 .88583 L .141 .057 .492 r
F
.5201 .88569 m .52007 .88542 L .52013 .88562 L .649 .393 .518 r
F
.52014 .88579 m .5239 .86759 L p .52749 .85577 L .339 .132 .446 r
F P
0 g
s
.52015 .88574 m .52013 .88562 L .52018 .88577 L .649 .393 .518 r
F
.5201 .88569 m .52013 .88562 L .52015 .88574 L F
.52 .88486 m .52017 .88571 L p .52001 .88489 L .95 .6 .297 r
F P
0 g
s
.52014 .88579 m .52 .88486 L p .5201 .88545 L .141 .057 .492 r
F P
0 g
s
.52006 .88534 m .52 .88486 L .52012 .88558 L .649 .393 .518 r
F
.52 .88486 m .52404 .8588 L p .52 .88502 L .62 .394 .553 r
F P
0 g
s
.52 .88488 m .52 .88487 L .52001 .88489 L .95 .6 .297 r
F
.52001 .88492 m .52001 .88489 L .52018 .88577 L F
.5201 .88545 m .52 .88486 L .52009 .88536 L .141 .057 .492 r
F
.52 .88488 m .52001 .88489 L .52001 .88492 L .95 .6 .297 r
F
.52018 .88577 m .52 .88486 L p .52 .88488 L F P
0 g
s
.52009 .88536 m .52 .88486 L .52003 .88475 L .141 .057 .492 r
F
.52012 .88577 m .52002 .88576 L p .51995 .88516 L .728 .45 .506 r
F P
0 g
s
.52002 .88576 m .52012 .88577 L p .52012 .88557 L .738 .512 .586 r
F P
0 g
s
.52001 .88489 m .52017 .88571 L .52017 .88571 L .95 .6 .297 r
F
.5201 .88545 m .52009 .88536 L .52005 .88517 L .141 .057 .492 r
F
.52009 .88536 m .52002 .88458 L .52005 .88513 L F
.52009 .88536 m .52005 .88513 L .52005 .88517 L F
.52005 .88544 m .52005 .88542 L .52008 .8857 L .943 .682 .216 r
F
.52 .88486 m .52009 .88588 L p .52003 .8851 L 0 0 .265 r
F P
0 g
s
.52002 .88527 m .52002 .88519 L .52018 .88577 L .649 .393 .518 r
F
.52005 .88523 m .52003 .8851 L .52009 .88588 L 0 0 .265 r
F
.5201 .88583 m .52 .88486 L p .52005 .88523 L F P
0 g
s
.5201 .88569 m .52 .88486 L .52002 .88527 L .649 .393 .518 r
F
.52012 .88577 m .52 .88486 L p .5201 .88569 L F P
0 g
s
.52 .88487 m .52 .88486 L .52001 .88489 L .95 .6 .297 r
F
.52002 .88519 m .52 .88486 L .52006 .88534 L .649 .393 .518 r
F
.52 .88486 m .52012 .88577 L p .52011 .88573 L .728 .45 .506 r
F P
0 g
s
.52003 .8851 m .52 .88486 L .52003 .8851 L 0 0 .265 r
F
.5201 .88569 m .52002 .88527 L .52007 .88542 L .649 .393 .518 r
F
.52001 .885 m .52001 .885 L .52001 .88501 L .943 .682 .216 r
F
.52001 .885 m .52001 .88501 L .52001 .88502 L F
.52003 .88545 m .52 .88486 L .52011 .88573 L .728 .45 .506 r
F
.52003 .88505 m .52 .88486 L .52003 .8851 L 0 0 .265 r
F
.52 .88486 m .52 .88486 L .52 .88487 L .95 .6 .297 r
F
.52003 .88509 m .52 .88486 L .52 .88473 L 0 0 .265 r
F
.52005 .88523 m .52 .88486 L .52003 .88509 L F
.51773 .85778 m .5193 .85837 L p .52002 .8835 L .85 .596 .554 r
F P
0 g
s
.52004 .88419 m .52002 .88447 L .52002 .88467 L .85 .596 .554 r
F
.52004 .88419 m .52002 .8835 L .52002 .88447 L F
.52002 .88437 m .52013 .8835 L .52002 .88447 L .212 .036 .411 r
F
.52 .88486 m .5201 .87056 L .52057 .87771 L .773 .533 .571 r
F
.52003 .88456 m .52007 .88435 L .52009 .88421 L .44 .23 .497 r
F
.52002 .88441 m .52016 .88352 L .52303 .85989 L .212 .036 .411 r
F
.52017 .88373 m .52057 .87771 L .52002 .88474 L .773 .533 .571 r
F
.51999 .88507 m .52404 .8588 L .51999 .88509 L .62 .394 .553 r
F
.52017 .88373 m .52 .88486 L .52057 .87771 L .773 .533 .571 r
F
.5193 .85837 m .52181 .85879 L p .52057 .87771 L F P
0 g
s
.52181 .85879 m .52 .88486 L p .52057 .87771 L .773 .533 .571 r
F P
0 g
s
.52 .88486 m .52181 .85879 L .52205 .85879 L p .698 .467 .569 r
F P
0 g
s
.5245 .85468 m .52303 .85989 L p .52438 .85837 L .144 0 .315 r
F P
0 g
s
.52238 .86116 m .52422 .85381 L p .52502 .85186 L .935 .819 .365 r
F P
0 g
s
.52238 .86116 m .52502 .85186 L .52562 .8504 L .935 .819 .365 r
F
.52303 .85989 m .5245 .85468 L p .52542 .85023 L 0 0 0 r
F P
0 g
s
.52303 .85989 m .52542 .85023 L .52482 .85145 L 0 0 0 r
F
.52482 .85145 m .52238 .86116 L .52294 .86006 L p F P
0 g
s
.52294 .86006 m .52303 .85989 L p .52482 .85145 L 0 0 0 r
F P
0 g
s
.52205 .85879 m .52404 .8588 L p .52 .88486 L .698 .467 .569 r
F P
0 g
s
.52018 .88577 m .52018 .88577 L p .52 .88488 L .95 .6 .297 r
F P
0 g
s
.5221 .80803 m .52117 .84284 L p .52121 .84277 L .738 .512 .586 r
F P
0 g
s
.53505 .73603 m .52776 .76306 L .51429 .74834 L .75 .396 .384 r
F
.51209 .7389 m .53929 .74387 L p .53853 .74827 L .774 .561 .611 r
F P
0 g
s
.51228 .74615 m .51545 .77304 L p .53489 .77086 L .778 .379 .296 r
F P
0 g
s
.52416 .77019 m .51209 .7389 L .53853 .74827 L .774 .561 .611 r
F
.51776 .80561 m .52645 .79753 L .51545 .77304 L .803 .476 .436 r
F
.5164 .78203 m .51291 .74708 L .52416 .77019 L .774 .561 .611 r
F
.51291 .74708 m .51209 .7389 L .52416 .77019 L F
.51594 .78273 m .51776 .79021 L .51895 .77813 L .807 .607 .632 r
F
.5221 .80803 m .52255 .79974 L p .51776 .79021 L F P
0 g
s
.54235 .71697 m .53505 .73603 L .49655 .72896 L p .683 .221 .149 r
F P
0 g
s
.51636 .72378 m .53326 .71935 L .50677 .69941 L .778 .379 .296 r
F
.50173 .68694 m .50677 .69941 L .55538 .71356 L p .615 .062 0 r
F P
0 g
s
.55107 .69419 m .5338 .70285 L .51533 .69369 L .683 .221 .149 r
F
.51606 .68747 m .53658 .69125 L .49786 .67735 L .615 .062 0 r
F
.55107 .69419 m .49496 .68358 L p .54484 .69277 L .15 0 0 r
F P
0 g
s
.53996 .67971 m .49803 .65895 L .53935 .6798 L .742 .705 .816 r
F
.53929 .74387 m .51209 .7389 L p .51591 .73123 L .77 .599 .667 r
F P
0 g
s
.51591 .73123 m .52372 .71554 L .52665 .72087 L .77 .599 .667 r
F
.51591 .73123 m .52665 .72087 L .53929 .74387 L F
.61883 .65929 m .60963 .6632 L .61433 .67261 L .676 .511 .656 r
F
.59224 .687 m .61341 .64953 L .59486 .6838 L .685 .836 .407 r
F
.61341 .64953 m .59147 .68794 L p .59853 .67933 L 0 0 .083 r
F P
0 g
s
.59519 .67413 m .61433 .67261 L p .60963 .6632 L .676 .511 .656 r
F P
0 g
s
.61397 .63018 m .58127 .67524 L .60963 .6632 L .676 .511 .656 r
F
.61838 .65658 m .60963 .6632 L .61883 .65929 L F
.61397 .63018 m .60963 .6632 L .61838 .65658 L F
.61967 .6556 m .6481 .60931 L .62121 .65444 L .633 .438 .601 r
F
.62222 .65047 m .61838 .65658 L p .61818 .65539 L .707 .533 .651 r
F P
0 g
s
.55992 .71404 m .56208 .71948 L .58127 .67524 L .74 .548 .635 r
F
.53505 .73603 m .54235 .71697 L p .55251 .72996 L .803 .361 .196 r
F P
0 g
s
.53929 .74387 m .56173 .7186 L .55739 .72948 L .707 .533 .651 r
F
.56173 .7186 m .56323 .72236 L .55739 .72948 L F
.55739 .72948 m .56323 .72236 L .5638 .72379 L F
.55843 .72865 m .5638 .72379 L .56398 .72424 L F
.55739 .72948 m .5638 .72379 L .55843 .72865 L F
.58127 .67524 m .5638 .72379 L p .56725 .72067 L .668 .47 .608 r
F P
0 g
s
.59225 .68699 m .57175 .73195 L p .57211 .72991 L 0 0 .028 r
F P
0 g
s
.57146 .73355 m .59667 .6978 L .57967 .71981 L 0 0 .083 r
F
.57175 .73195 m .59225 .68699 L p .60788 .68513 L .284 .041 .358 r
F P
0 g
s
.58943 .69042 m .59225 .68699 L p .55649 .75042 L 0 0 .028 r
F P
0 g
s
.56398 .72424 m .56564 .72213 L .58769 .70537 L .707 .533 .651 r
F
.58127 .67524 m .56725 .72067 L .5697 .71845 L .668 .47 .608 r
F
.56564 .72213 m .59205 .69824 L .58769 .70537 L .707 .533 .651 r
F
.61433 .67261 m .59861 .67386 L p .59386 .68529 L .668 .47 .608 r
F P
0 g
s
.59386 .68529 m .59252 .6966 L .61433 .67261 L .668 .47 .608 r
F
.56528 .74593 m .57067 .73794 L p .57146 .73355 L 0 0 .083 r
F P
0 g
s
.60788 .68513 m .5849 .72521 L .60132 .69363 L .284 .041 .358 r
F
.59545 .70124 m .56528 .74593 L p .59226 .71149 L .391 .139 .408 r
F P
0 g
s
.57146 .73355 m .5732 .73006 L .57175 .73195 L 0 0 .083 r
F
.55218 .75671 m .56528 .74593 L p .57146 .73355 L F P
0 g
s
.55359 .75444 m .57232 .72159 L .55962 .74472 L .685 .836 .407 r
F
.55218 .75671 m .55305 .75518 L p .55359 .75444 L F P
0 g
s
.57422 .71813 m .55218 .75671 L p .57146 .73355 L 0 0 .083 r
F P
0 g
s
.56339 .74402 m .5918 .70262 L .59096 .70968 L .633 .438 .601 r
F
.55824 .77586 m .54953 .7917 L p .5499 .78826 L .284 .041 .358 r
F P
0 g
s
.56528 .74593 m .5499 .78826 L p .55035 .7876 L .424 .201 .468 r
F P
0 g
s
.55982 .76096 m .56528 .74593 L p .5598 .761 L .144 0 .315 r
F P
0 g
s
.56027 .77184 m .5616 .75996 L .57243 .74265 L .625 .411 .571 r
F
.56528 .74593 m .57243 .74265 L p .59061 .71359 L .391 .139 .408 r
F P
0 g
s
.56528 .74593 m .55255 .7564 L p .54223 .79376 L .144 0 .315 r
F P
0 g
s
.56339 .74402 m .56855 .74324 L p .5616 .75996 L .625 .411 .571 r
F P
0 g
s
.56855 .74324 m .57243 .74265 L p .56307 .75762 L .625 .411 .571 r
F P
0 g
s
.5616 .75996 m .56855 .74324 L .56307 .75762 L .625 .411 .571 r
F
.57243 .74265 m .59226 .71149 L .59061 .71359 L .391 .139 .408 r
F
.55887 .77426 m .56339 .74402 L .56027 .77184 L .625 .411 .571 r
F
.55779 .77654 m .55824 .77586 L .55966 .77315 L .909 .796 .724 r
F
.56014 .77206 m .55082 .78894 L .55005 .79058 L F
.55831 .77524 m .55082 .78894 L .56014 .77206 L F
.5499 .78826 m .55982 .76096 L p .54991 .7882 L .144 0 .315 r
F P
0 g
s
.55017 .78581 m .53941 .80458 L .55094 .77869 L .284 .041 .358 r
F
.54992 .78808 m .5598 .761 L .55109 .77723 L .144 0 .315 r
F
.54923 .78632 m .56339 .74402 L .55887 .77426 L .625 .411 .571 r
F
.56095 .77066 m .55005 .79058 L .55966 .77315 L .909 .796 .724 r
F
.57243 .74265 m .56095 .77066 L p .56027 .77184 L .625 .411 .571 r
F P
0 g
s
.5619 .76883 m .56014 .77206 L .56095 .77066 L .909 .796 .724 r
F
.56014 .77206 m .55005 .79058 L .56095 .77066 L F
.55806 .77568 m .55082 .78894 L p .55831 .77524 L F P
0 g
s
.55082 .78894 m .55806 .77568 L p .55459 .77987 L .668 .47 .608 r
F P
0 g
s
.54923 .78632 m .55887 .77426 L .55806 .77568 L .625 .411 .571 r
F
.54953 .7917 m .54024 .80858 L p .5499 .78826 L .284 .041 .358 r
F P
0 g
s
.56339 .74402 m .55436 .74396 L p .56632 .73012 L .707 .533 .651 r
F P
0 g
s
.58769 .70537 m .56857 .73579 L p .56632 .73012 L .707 .533 .651 r
F P
0 g
s
.56857 .73579 m .56339 .74402 L p .56632 .73012 L .707 .533 .651 r
F P
0 g
s
.55436 .74396 m .53929 .74387 L p .56398 .72424 L .707 .533 .651 r
F P
0 g
s
.53929 .74387 m .53264 .78763 L p .53191 .78614 L .774 .561 .611 r
F P
0 g
s
.54303 .74389 m .56339 .74402 L p .55766 .75985 L .703 .494 .602 r
F P
0 g
s
.54997 .77933 m .54054 .80037 L .54162 .79491 L 0 0 .028 r
F
.54186 .79445 m .55255 .7564 L .55127 .76003 L .144 0 .315 r
F
.54223 .79376 m .55255 .7564 L .54186 .79445 L F
.55218 .75671 m .54113 .79581 L p .54186 .79445 L F P
0 g
s
.54125 .79491 m .55102 .76077 L .54129 .79456 L 0 0 0 r
F
.54838 .78544 m .54627 .77249 L .55766 .75985 L .703 .494 .602 r
F
.55102 .76077 m .55218 .75671 L .55097 .76093 L 0 0 0 r
F
.55112 .76045 m .55218 .75671 L p .55102 .76077 L F P
0 g
s
.55255 .7564 m .55218 .75671 L p .55127 .76003 L .144 0 .315 r
F P
0 g
s
.54627 .77249 m .54303 .74389 L .55766 .75985 L .703 .494 .602 r
F
.53668 .78314 m .54231 .74831 L .54627 .77249 L F
.54231 .74831 m .54303 .74389 L .54627 .77249 L F
.56398 .72424 m .58769 .70537 L .56632 .73012 L .707 .533 .651 r
F
.56398 .72424 m .56632 .73012 L .55436 .74396 L F
.56725 .72067 m .54799 .78306 L .5697 .71845 L .668 .47 .608 r
F
.54968 .78863 m .54992 .78808 L .54986 .78812 L .144 0 .315 r
F
.55017 .78581 m .5499 .78826 L .54224 .79964 L .284 .041 .358 r
F
.54799 .78306 m .54014 .80849 L .54816 .78407 L .668 .47 .608 r
F
.54799 .78306 m .54816 .78407 L .5697 .71845 L F
.54014 .80849 m .54939 .79156 L p .54816 .78407 L F P
0 g
s
.53176 .83947 m .53185 .83811 L .52422 .85897 L .424 .201 .468 r
F
.54024 .80858 m .5335 .83451 L p .53348 .83456 L .339 .132 .446 r
F P
0 g
s
.53168 .83833 m .53171 .83822 L .52438 .85837 L .144 0 .315 r
F
.54014 .80849 m .54024 .80858 L p .55005 .79058 L .909 .796 .724 r
F P
0 g
s
.55016 .79035 m .54014 .80849 L .55005 .79058 L .909 .796 .724 r
F
.55082 .78894 m .54939 .79156 L p .55016 .79035 L F P
0 g
s
.54939 .79156 m .54014 .80849 L p .55016 .79035 L .909 .796 .724 r
F P
0 g
s
.54162 .79491 m .54054 .80037 L .54111 .79599 L 0 0 .028 r
F
.53014 .82184 m .53362 .81376 L p .53632 .80533 L .826 .841 .375 r
F P
0 g
s
.53632 .80533 m .53362 .81376 L .53648 .80506 L .826 .841 .375 r
F
.54111 .79599 m .53192 .81557 L .53959 .79869 L 0 0 .028 r
F
.54113 .79581 m .53959 .79869 L .54123 .79508 L F
.54111 .79599 m .53959 .79869 L .54113 .79581 L F
.54162 .79491 m .54111 .79599 L .54113 .79581 L F
.53555 .81131 m .54054 .80037 L p .53941 .80458 L .284 .041 .358 r
F P
0 g
s
.54123 .79508 m .53192 .81557 L .53256 .81344 L 0 0 .028 r
F
.53555 .81131 m .53192 .81557 L p .54111 .79599 L F P
0 g
s
.54054 .80037 m .53555 .81131 L p .54111 .79599 L 0 0 .028 r
F P
0 g
s
.53687 .80385 m .54219 .78767 L .53632 .80533 L .826 .841 .375 r
F
.54219 .78767 m .53383 .80943 L .53632 .80533 L F
.53383 .80943 m .53354 .81017 L p .5336 .81007 L F P
0 g
s
.53383 .80943 m .5336 .81007 L .53632 .80533 L .826 .841 .375 r
F
.53354 .81017 m .5311 .81653 L p .53632 .80533 L F P
0 g
s
.53354 .81017 m .54219 .78767 L p .53256 .81344 L 0 0 .028 r
F P
0 g
s
.53256 .81344 m .54219 .78767 L .5421 .78831 L 0 0 .028 r
F
.53632 .80533 m .53648 .80506 L .53656 .8048 L .826 .841 .375 r
F
.53687 .80385 m .53632 .80533 L .53656 .80479 L F
.54123 .79508 m .53256 .81344 L .5421 .78831 L 0 0 .028 r
F
.53312 .809 m .54014 .80849 L p .54799 .78306 L .668 .47 .608 r
F P
0 g
s
.56725 .72067 m .54655 .77424 L .54799 .78306 L .668 .47 .608 r
F
.54655 .77424 m .53312 .809 L .54799 .78306 L F
.53555 .81131 m .54024 .80858 L p .53348 .83456 L .339 .132 .446 r
F P
0 g
s
.52867 .84454 m .53555 .81131 L p .53348 .83456 L .339 .132 .446 r
F P
0 g
s
.53555 .81131 m .52867 .84454 L p .52867 .84448 L 0 0 .274 r
F P
0 g
s
.52859 .8402 m .52897 .8304 L .5312 .8323 L 0 0 .274 r
F
.52648 .84659 m .52846 .83371 L .52859 .8402 L F
.52846 .83371 m .52897 .8304 L .52859 .8402 L F
.53097 .8334 m .52966 .83697 L .5312 .8323 L F
.52867 .84448 m .52966 .83697 L .53097 .8334 L F
.52887 .83953 m .52859 .8403 L .52859 .84047 L .144 0 .315 r
F
.52867 .84448 m .52859 .8402 L .52966 .83697 L 0 0 .274 r
F
.5312 .8323 m .52897 .8304 L .53555 .81131 L F
.53192 .81557 m .5311 .81653 L p .53256 .81344 L 0 0 .028 r
F P
0 g
s
.5311 .81653 m .53354 .81017 L p .53256 .81344 L 0 0 .028 r
F P
0 g
s
.53192 .81557 m .53555 .81131 L p .52897 .8304 L 0 0 .274 r
F P
0 g
s
.52735 .84324 m .53408 .82496 L .53312 .809 L .661 .439 .572 r
F
.54014 .80849 m .53781 .80866 L p .53408 .82496 L F P
0 g
s
.53781 .80866 m .53312 .809 L p .53408 .82496 L .661 .439 .572 r
F P
0 g
s
.53115 .83775 m .53128 .83755 L .53086 .83892 L .703 .494 .602 r
F
.53144 .83733 m .54014 .80849 L .53118 .83283 L .661 .439 .572 r
F
.52859 .84047 m .52582 .85002 L .52578 .85022 L .144 0 .315 r
F
.52582 .85002 m .52859 .8402 L p .52617 .8476 L 0 0 0 r
F P
0 g
s
.52859 .8402 m .52633 .84703 L .52617 .8476 L 0 0 0 r
F
.52859 .8403 m .52846 .84067 L .52859 .84047 L .144 0 .315 r
F
.52864 .8428 m .52859 .84047 L .5245 .85468 L F
.53136 .83744 m .53118 .83283 L .52735 .84324 L .661 .439 .572 r
F
.53144 .83733 m .53118 .83283 L .53136 .83744 L F
.53086 .83892 m .53136 .83744 L p .52881 .84734 L .625 .411 .571 r
F P
0 g
s
.53136 .83744 m .53144 .83936 L .52881 .84734 L .625 .411 .571 r
F
.533 .83613 m .53344 .83445 L .533 .83614 L .951 .759 .609 r
F
.54014 .80849 m .53344 .83444 L p .53344 .83445 L F P
0 g
s
.53344 .83444 m .533 .83614 L p .53344 .83445 L .951 .759 .609 r
F P
0 g
s
.54024 .80858 m .54014 .80849 L p .53344 .83445 L .951 .759 .609 r
F P
0 g
s
.533 .83613 m .5334 .8346 L .53344 .83445 L .951 .759 .609 r
F
.5334 .8346 m .54024 .80858 L .53344 .83445 L F
.533 .83614 m .54014 .80849 L p .53137 .83756 L .661 .439 .572 r
F P
0 g
s
.53144 .83936 m .53144 .83949 L .52881 .84734 L .625 .411 .571 r
F
.533 .83614 m .53137 .83756 L .53148 .84046 L .661 .439 .572 r
F
.52881 .84734 m .53357 .83314 L .53207 .83953 L .625 .411 .571 r
F
.52881 .84734 m .5315 .8409 L .52779 .85157 L F
.52438 .85837 m .5246 .85777 L .52467 .85733 L .144 0 .315 r
F
.52438 .85837 m .52467 .85731 L .5245 .85468 L F
.52422 .85897 m .52438 .85853 L p .52438 .85837 L F P
0 g
s
.52487 .85868 m .52404 .8588 L p .52803 .84921 L .625 .411 .571 r
F P
0 g
s
.52803 .84921 m .52881 .84734 L .52798 .84974 L .625 .411 .571 r
F
.52803 .84921 m .52798 .84974 L .52487 .85868 L F
.52619 .84746 m .52845 .83323 L p .52637 .84692 L 0 0 .274 r
F P
0 g
s
.5281 .83486 m .5279 .83615 L .52711 .8412 L 0 0 0 r
F
.5281 .83486 m .52711 .8412 L .52715 .84093 L F
.52721 .84052 m .5281 .83486 L p .52717 .84079 L F P
0 g
s
.52731 .84329 m .53079 .82279 L p .52683 .8419 L .738 .512 .586 r
F P
0 g
s
.5311 .81653 m .53192 .81557 L p .52897 .8304 L 0 0 .274 r
F P
0 g
s
.53079 .82279 m .53312 .809 L p .53124 .8185 L .738 .512 .586 r
F P
0 g
s
.52863 .83211 m .52845 .83323 L p .52862 .83218 L 0 0 0 r
F P
0 g
s
.52862 .83218 m .52845 .83323 L .52847 .8331 L 0 0 0 r
F
.52847 .83309 m .52862 .83218 L .52847 .8331 L F
.52845 .83323 m .5311 .81653 L p .52846 .83371 L 0 0 .274 r
F P
0 g
s
.52862 .83218 m .53021 .82199 L .53023 .82204 L 0 0 0 r
F
.52862 .83218 m .52819 .83429 L .53021 .82199 L F
.5281 .83486 m .53014 .82184 L p .5279 .83615 L F P
0 g
s
.53014 .82184 m .53028 .82111 L p .52819 .83429 L 0 0 0 r
F P
0 g
s
.53028 .82111 m .5311 .81653 L p .52819 .83429 L 0 0 0 r
F P
0 g
s
.5311 .81653 m .52863 .83211 L p .52862 .83218 L 0 0 0 r
F P
0 g
s
.53079 .82279 m .53124 .8185 L .52667 .84144 L .738 .512 .586 r
F
.5249 .84612 m .52413 .84998 L p .52527 .84833 L .889 .573 .442 r
F P
0 g
s
.52181 .85879 m .52559 .83833 L .52656 .84568 L .703 .494 .602 r
F
.52559 .83833 m .52706 .84257 L .52656 .84568 L F
.52647 .84621 m .52644 .84641 L .52197 .85879 L F
.52199 .85828 m .52656 .84568 L .52197 .85879 L F
.52656 .84568 m .52647 .84621 L .52197 .85879 L F
.52582 .85002 m .5245 .85468 L p .52578 .85022 L .144 0 .315 r
F P
0 g
s
.52467 .85731 m .52556 .85162 L .5245 .85468 L .144 0 .315 r
F
.52617 .8476 m .52633 .84704 L .52619 .84746 L 0 0 0 r
F
.52562 .8504 m .52573 .84997 L .52571 .85011 L .935 .819 .365 r
F
.5245 .85468 m .52582 .85002 L p .52542 .85023 L 0 0 0 r
F P
0 g
s
.52582 .85002 m .52617 .8476 L .52542 .85023 L 0 0 0 r
F
.52542 .85023 m .52617 .8476 L .52632 .84659 L F
.52404 .8588 m .52197 .85879 L p .52644 .84641 L .703 .494 .602 r
F P
0 g
s
.52683 .8419 m .52667 .84144 L .52045 .87272 L .738 .512 .586 r
F
.53079 .82279 m .52667 .84144 L .52683 .8419 L F
.52458 .8217 m .52337 .82124 L p .52567 .81803 L .842 .49 .388 r
F P
0 g
s
.5256 .82016 m .52458 .8217 L .52567 .81803 L .842 .49 .388 r
F
.52337 .82124 m .52392 .81724 L p .52569 .8157 L F P
0 g
s
.53312 .809 m .5221 .80803 L p .52162 .82691 L .738 .512 .586 r
F P
0 g
s
.52121 .84277 m .52419 .83432 L .52162 .82691 L .738 .512 .586 r
F
.52419 .83432 m .53312 .809 L .52162 .82691 L F
.52181 .85879 m .52513 .837 L p .52559 .83833 L .703 .494 .602 r
F P
0 g
s
.51921 .8531 m .5188 .84433 L p .51585 .849 L .719 .407 .45 r
F P
0 g
s
.5193 .85837 m .51859 .84664 L p .5223 .84117 L .774 .561 .611 r
F P
0 g
s
.52178 .84417 m .5223 .84117 L .52219 .84507 L .774 .561 .611 r
F
.52178 .84417 m .52219 .84507 L .52181 .85879 L F
.5223 .84117 m .52477 .83753 L .52181 .85879 L F
.52667 .84144 m .52856 .83192 L .52513 .837 L .738 .512 .586 r
F
.53023 .82204 m .53021 .82199 L .5311 .81653 L 0 0 0 r
F
.52856 .83192 m .53312 .809 L .52121 .84277 L .738 .512 .586 r
F
.5221 .80803 m .53312 .809 L p .53258 .78769 L .74 .548 .635 r
F P
0 g
s
.52559 .83833 m .53555 .78439 L .52706 .84257 L .703 .494 .602 r
F
.53555 .78439 m .54303 .74389 L .53668 .78314 L F
.53555 .78439 m .53668 .78314 L .52706 .84257 L F
.52958 .67028 m .52684 .72068 L p .53077 .71689 L .74 .548 .635 r
F P
0 g
s
.57267 .68716 m .56247 .69354 L .55922 .67721 L .791 .695 .758 r
F
.55063 .63618 m .55908 .6771 L p .55894 .6771 L .743 .664 .773 r
F P
0 g
s
.55063 .63618 m .55894 .6771 L .52574 .65002 L .743 .664 .773 r
F
.52958 .67028 m .53077 .71689 L .55266 .69577 L .74 .548 .635 r
F
.5746 .6746 m .54303 .67157 L p .55266 .69577 L F P
0 g
s
.54303 .67157 m .52958 .67028 L p .55266 .69577 L .74 .548 .635 r
F P
0 g
s
.56208 .71948 m .5621 .71952 L .58127 .67524 L .74 .548 .635 r
F
.5638 .72379 m .58127 .67524 L p .5621 .71952 L F P
0 g
s
.55266 .69577 m .55992 .71404 L .5746 .6746 L .74 .548 .635 r
F
.58127 .67524 m .5746 .6746 L p .55992 .71404 L F P
0 g
s
.58127 .67524 m .59519 .67413 L p .60963 .6632 L .676 .511 .656 r
F P
0 g
s
.61692 .64787 m .61289 .6237 L .6481 .60931 L .707 .533 .651 r
F
.61045 .60906 m .63404 .60922 L p .61289 .6237 L F P
0 g
s
.63404 .60922 m .6481 .60931 L p .61289 .6237 L .707 .533 .651 r
F P
0 g
s
.58127 .67524 m .61289 .6237 L p .58952 .63972 L .738 .593 .694 r
F P
0 g
s
.57682 .60883 m .61045 .60906 L p .59957 .63283 L .707 .533 .651 r
F P
0 g
s
.58657 .65175 m .55063 .63618 L p .5735 .62837 L .82 .658 .682 r
F P
0 g
s
.57351 .65069 m .58127 .67524 L .58952 .63972 L .738 .593 .694 r
F
.57041 .65281 m .57682 .60883 L .59957 .63283 L .707 .533 .651 r
F
.54065 .54679 m .53668 .59108 L p .55216 .58317 L .738 .593 .694 r
F P
0 g
s
.56779 .53568 m .5628 .53193 L p .55778 .51149 L .904 .666 .569 r
F P
0 g
s
.5628 .53193 m .56382 .53464 L p .56217 .53732 L .82 .658 .682 r
F P
0 g
s
.54679 .50357 m .54065 .54679 L .59502 .55243 L p .726 .65 .774 r
F P
0 g
s
.59502 .55243 m .60942 .55392 L p .60386 .54945 L .726 .65 .774 r
F P
0 g
s
.54679 .50357 m .59502 .55243 L .60386 .54945 L .726 .65 .774 r
F
.52321 .57629 m .58822 .53688 L .54751 .60297 L .758 .645 .736 r
F
.58822 .53688 m .59692 .54387 L .54751 .60297 L F
.57351 .65069 m .56323 .65773 L .58127 .67524 L .738 .593 .694 r
F
.54303 .67157 m .5746 .6746 L p .56323 .65773 L F P
0 g
s
.5746 .6746 m .58127 .67524 L p .56323 .65773 L .738 .593 .694 r
F P
0 g
s
.55423 .60539 m .5735 .62837 L .55063 .63618 L .82 .658 .682 r
F
.57682 .60883 m .56323 .65773 L p .55705 .64856 L .77 .599 .667 r
F P
0 g
s
.55266 .69577 m .56323 .65773 L p .56107 .71694 L .707 .533 .651 r
F P
0 g
s
.56323 .65773 m .57682 .60883 L p .57041 .65281 L .707 .533 .651 r
F P
0 g
s
.56323 .65773 m .57041 .65281 L .56107 .71694 L .707 .533 .651 r
F
.51209 .7389 m .50646 .68722 L p .51501 .70125 L .77 .599 .667 r
F P
0 g
s
.52958 .67028 m .51633 .66633 L p .52403 .71605 L .807 .607 .632 r
F P
0 g
s
.51633 .66633 m .48874 .65811 L p .50646 .68722 L .807 .607 .632 r
F P
0 g
s
.50646 .68722 m .52403 .71605 L .51633 .66633 L .807 .607 .632 r
F
.48533 .6571 m .48379 .65623 L p .4939 .70834 L .884 .656 .588 r
F P
0 g
s
.49496 .73201 m .49514 .71121 L .49724 .71464 L .839 .649 .645 r
F
.50957 .73788 m .49496 .73201 L p .49724 .71464 L F P
0 g
s
.49527 .71092 m .48533 .6571 L p .4939 .70834 L .884 .656 .588 r
F P
0 g
s
.51209 .7389 m .50957 .73788 L p .49724 .71464 L .839 .649 .645 r
F P
0 g
s
.50646 .68722 m .51209 .7389 L p .49514 .71121 L .839 .649 .645 r
F P
0 g
s
.54695 .63359 m .54303 .67157 L .56323 .65773 L .738 .593 .694 r
F
.55705 .64856 m .53452 .61515 L .57682 .60883 L .77 .599 .667 r
F
.52301 .59806 m .54751 .60297 L p .53452 .61515 L F P
0 g
s
.54751 .60297 m .57682 .60883 L p .53452 .61515 L .77 .599 .667 r
F P
0 g
s
.52958 .67028 m .53452 .61515 L p .51852 .63017 L .796 .645 .694 r
F P
0 g
s
.52958 .67028 m .54303 .67157 L p .54695 .63359 L .738 .593 .694 r
F P
0 g
s
.55216 .58317 m .54127 .62516 L .54695 .63359 L .738 .593 .694 r
F
.54127 .62516 m .52958 .67028 L .54695 .63359 L F
.52684 .72068 m .52958 .67028 L p .52492 .71753 L .807 .607 .632 r
F P
0 g
s
.53982 .74198 m .53929 .74387 L p .52372 .71554 L .77 .599 .667 r
F P
0 g
s
.51794 .8358 m .5221 .80803 L .51488 .82914 L .811 .574 .581 r
F
.52117 .84284 m .52163 .82559 L p .51794 .8358 L F P
0 g
s
.52163 .82559 m .5221 .80803 L p .51794 .8358 L .811 .574 .581 r
F P
0 g
s
.51397 .83963 m .51396 .83957 L .51396 .83958 L .901 .613 .111 r
F
.51396 .83957 m .50336 .7822 L .51396 .83958 L F
.48459 .70994 m .49355 .7437 L .46954 .65147 L 0 0 .316 r
F
.49588 .75246 m .49181 .73685 L p .49355 .7437 L F P
0 g
s
.49181 .73685 m .46954 .65147 L p .49355 .7437 L 0 0 .316 r
F P
0 g
s
.49496 .73201 m .49186 .73001 L p .47908 .68689 L .907 .678 .581 r
F P
0 g
s
.44179 .58489 m .44932 .60231 L .4081 .50196 L 0 .043 .608 r
F
.44976 .60333 m .44894 .60135 L p .44932 .60231 L F P
0 g
s
.44894 .60135 m .4081 .50196 L p .44932 .60231 L 0 .043 .608 r
F P
0 g
s
.46678 .64639 m .44976 .60333 L .44179 .58489 L 0 .043 .608 r
F
.40122 .48997 m .38506 .46289 L p .37863 .45249 L 0 0 .433 r
F P
0 g
s
.38057 .45542 m .3701 .43462 L p .3857 .46361 L .892 .705 .641 r
F P
0 g
s
.38506 .46289 m .40122 .48997 L .40178 .49031 L p .84 .718 .729 r
F P
0 g
s
.40178 .49031 m .4182 .50026 L p .40509 .48548 L .84 .718 .729 r
F P
0 g
s
.38506 .46289 m .40178 .49031 L .40509 .48548 L .84 .718 .729 r
F
.38727 .44722 m .41855 .4617 L p .41928 .46476 L .818 .684 .716 r
F P
0 g
s
.40689 .48285 m .38727 .44722 L .41928 .46476 L .818 .684 .716 r
F
.38057 .45542 m .3857 .46361 L .38506 .46289 L .892 .705 .641 r
F
.40122 .48997 m .4157 .51393 L .4164 .51011 L .872 .683 .644 r
F
.49803 .65895 m .50125 .63004 L p .48201 .636 L .731 .615 .729 r
F P
0 g
s
.49578 .63689 m .46719 .59593 L .49617 .5927 L .839 .649 .645 r
F
.45116 .57297 m .46989 .58118 L p .46719 .59593 L F P
0 g
s
.46989 .58118 m .49617 .5927 L p .46719 .59593 L .839 .649 .645 r
F P
0 g
s
.44445 .57002 m .4506 .55543 L .45527 .56167 L .818 .684 .716 r
F
.46907 .60225 m .46719 .59593 L p .46677 .59824 L .872 .683 .644 r
F P
0 g
s
.37863 .45249 m .38506 .46289 L .38057 .45542 L 0 0 .433 r
F
.41434 .52137 m .38057 .45542 L .40437 .49833 L .892 .705 .641 r
F
.4178 .52937 m .41419 .5222 L p .41434 .52137 L F P
0 g
s
.41419 .5222 m .38057 .45542 L p .41434 .52137 L .892 .705 .641 r
F P
0 g
s
.4815 .69832 m .47317 .67419 L .47937 .69347 L .781 .455 0 r
F
.4232 .543 m .38658 .46826 L .4178 .52937 L 0 0 .495 r
F
.38658 .46826 m .3795 .45381 L .41167 .51727 L F
.38658 .46826 m .41167 .51727 L .4178 .52937 L F
.40122 .48997 m .4081 .50196 L p .42668 .54786 L .88 .525 .034 r
F P
0 g
s
.42668 .54786 m .4178 .52937 L .40122 .48997 L .88 .525 .034 r
F
.41689 .52747 m .41434 .52137 L p .4178 .52937 L F P
0 g
s
.41434 .52137 m .40122 .48997 L p .4178 .52937 L .88 .525 .034 r
F P
0 g
s
.40122 .48997 m .40149 .4906 L p .40179 .4911 L .872 .683 .644 r
F P
0 g
s
.40149 .4906 m .41434 .52137 L p .40179 .4911 L .872 .683 .644 r
F P
0 g
s
.4081 .50196 m .41182 .51103 L p .41154 .51047 L .88 .525 .034 r
F P
0 g
s
.41182 .51103 m .42861 .55188 L p .41154 .51047 L .88 .525 .034 r
F P
0 g
s
.41154 .51047 m .42861 .55188 L .42668 .54786 L .88 .525 .034 r
F
.40179 .4911 m .41434 .52137 L .41506 .51746 L .872 .683 .644 r
F
.48533 .6571 m .46907 .60225 L p .46674 .5984 L F P
0 g
s
.46382 .61437 m .48533 .6571 L .46674 .5984 L .872 .683 .644 r
F
.46332 .61712 m .45116 .57297 L .46719 .59593 L .839 .649 .645 r
F
.46829 .64844 m .44976 .60333 L p .46678 .64639 L 0 .043 .608 r
F P
0 g
s
.46081 .63541 m .46678 .64639 L .45422 .61722 L 0 .043 .608 r
F
.4232 .543 m .4178 .52937 L .42861 .55188 L 0 0 .495 r
F
.42004 .5378 m .46678 .64639 L .42791 .55073 L 0 .043 .608 r
F
.42668 .54786 m .44546 .59325 L .44666 .59719 L .88 .525 .034 r
F
.38057 .45542 m .4178 .52937 L p .3795 .45381 L 0 0 .495 r
F P
0 g
s
.4178 .52937 m .43609 .57235 L .41689 .52747 L .88 .525 .034 r
F
.43609 .57235 m .44666 .59719 L .43679 .57409 L F
.43609 .57235 m .43679 .57409 L .41689 .52747 L F
.43954 .57961 m .44666 .59719 L .4178 .52937 L F
.44836 .60279 m .41689 .52747 L p .44666 .59719 L F P
0 g
s
.46288 .64112 m .47761 .68528 L p .46028 .63761 L 0 .106 .629 r
F P
0 g
s
.43253 .5834 m .45132 .62551 L .46678 .64639 L 0 .043 .608 r
F
.47317 .67419 m .46044 .6344 L .44501 .59178 L .781 .455 0 r
F
.42861 .55188 m .44501 .59178 L p .42668 .54786 L .88 .525 .034 r
F P
0 g
s
.44501 .59178 m .44546 .59325 L .42668 .54786 L .88 .525 .034 r
F
.44712 .59872 m .44501 .59178 L .43897 .5751 L .781 .455 0 r
F
.47317 .67419 m .44501 .59178 L .44712 .59872 L F
.46954 .65147 m .46829 .64844 L p .46678 .64639 L 0 .043 .608 r
F P
0 g
s
.44501 .59178 m .46954 .65147 L p .44546 .59325 L .88 .525 .034 r
F P
0 g
s
.4815 .69832 m .48048 .6952 L .47317 .67419 L .781 .455 0 r
F
.48048 .6952 m .46044 .6344 L .47317 .67419 L F
.48308 .70384 m .48989 .73016 L .47926 .69315 L .952 .641 .212 r
F
.48308 .70384 m .49945 .76702 L .48726 .72 L F
.48308 .70384 m .49918 .76513 L .49945 .76702 L F
.48351 .70505 m .49918 .76513 L p .48308 .70384 L F P
0 g
s
.47926 .69315 m .49948 .76727 L .47905 .69254 L .952 .641 .212 r
F
.48989 .73016 m .49948 .76727 L .47926 .69315 L F
.49948 .76727 m .49979 .76946 L .47905 .69254 L F
.48735 .72036 m .49945 .76702 L .49948 .76727 L F
.48734 .72032 m .49945 .76702 L .48735 .72036 L F
.48534 .71826 m .49988 .77184 L .50284 .78083 L 0 0 .316 r
F
.51396 .83958 m .50336 .7822 L .50114 .77314 L .901 .613 .111 r
F
.51007 .82 m .50114 .77314 L .49918 .76513 L F
.51396 .83958 m .50114 .77314 L .51007 .82 L F
.51437 .84253 m .50194 .78986 L .49988 .77184 L 0 0 .426 r
F
.51441 .84281 m .50194 .78986 L .51437 .84253 L F
.50212 .79142 m .50212 .7914 L .51497 .8471 L F
.51282 .83884 m .51285 .83892 L .51531 .84973 L F
.51649 .8606 m .51751 .85943 L p .51938 .87816 L .141 .057 .492 r
F P
0 g
s
.51978 .88295 m .51649 .8606 L .51938 .87816 L .141 .057 .492 r
F
.51999 .88459 m .51999 .88459 L .52001 .88479 L .914 .646 .517 r
F
.5193 .85837 m .52 .88486 L p .52 .88283 L .85 .596 .554 r
F P
0 g
s
.51993 .88409 m .51905 .87805 L .52 .88486 L .391 .225 .533 r
F
.52 .88486 m .51649 .8606 L p .51997 .88426 L .141 .057 .492 r
F P
0 g
s
.5155 .84645 m .51772 .85823 L .51761 .85885 L p 0 0 .426 r
F P
0 g
s
.51761 .85885 m .51751 .85943 L p .5155 .84645 L 0 0 .426 r
F P
0 g
s
.51995 .88391 m .51984 .88278 L .51998 .88452 L .141 .057 .492 r
F
.51984 .88278 m .51751 .85943 L .51998 .88452 L F
.51751 .85943 m .52 .88486 L p .51998 .88452 L F P
0 g
s
.52 .88486 m .5193 .85837 L p .5201 .87056 L .773 .533 .571 r
F P
0 g
s
.51995 .88393 m .51978 .88295 L .51938 .87816 L .141 .057 .492 r
F
.52 .88471 m .51999 .88457 L .51998 .88452 L .914 .646 .517 r
F
.52 .88486 m .51982 .88123 L .52 .88329 L .85 .596 .554 r
F
.52002 .88458 m .51997 .88426 L .51995 .88393 L .141 .057 .492 r
F
.51997 .88426 m .51978 .88295 L .51995 .88392 L F
.51751 .85943 m .51772 .85823 L p .51978 .8822 L 0 0 .265 r
F P
0 g
s
.51998 .88452 m .51992 .8838 L .51969 .88109 L .914 .646 .517 r
F
.52 .88471 m .51969 .88109 L .52 .88486 L F
.52 .88486 m .51993 .88403 L .51999 .88459 L F
.52 .88486 m .51999 .88459 L .52 .88472 L F
.52 .88486 m .51751 .85943 L p .52 .88473 L 0 0 .265 r
F P
0 g
s
.51773 .85778 m .52 .88486 L p .51969 .88109 L .914 .646 .517 r
F P
0 g
s
.51979 .88228 m .51994 .88407 L .51978 .8822 L 0 0 .265 r
F
.51994 .88408 m .51994 .88407 L .51979 .88228 L F
.51994 .88407 m .51751 .85943 L .51978 .8822 L F
.52 .88488 m .51992 .88393 L .51745 .85761 L .943 .682 .216 r
F
.51772 .85823 m .51745 .85761 L p .51992 .88393 L F P
0 g
s
.51997 .88478 m .51998 .88465 L .52 .88486 L .891 .597 .483 r
F
.51993 .8855 m .51999 .88479 L .51992 .88451 L .809 .517 .495 r
F
.5199 .88572 m .51997 .88478 L .52 .88486 L .891 .597 .483 r
F
.5199 .88572 m .52 .88486 L p .51997 .88478 L F P
0 g
s
.52 .88486 m .52 .88486 L .52 .88486 L .95 .6 .297 r
F
.52 .88486 m .51772 .85823 L p .52 .88488 L .943 .682 .216 r
F P
0 g
s
.51999 .88457 m .51745 .85761 L .51998 .88452 L .914 .646 .517 r
F
.52 .88472 m .51745 .85761 L .52 .88471 L F
.52 .88486 m .51773 .85778 L p .51849 .86637 L .85 .596 .554 r
F P
0 g
s
.51997 .88426 m .51997 .88426 L .51995 .88393 L .141 .057 .492 r
F
.51997 .88426 m .51995 .88392 L .51995 .88393 L F
.52 .88486 m .51997 .88426 L .51997 .88429 L F
.52 .88486 m .51849 .86637 L .51982 .88123 L .85 .596 .554 r
F
.52 .88473 m .51994 .88408 L .51979 .88228 L 0 0 .265 r
F
.52 .88486 m .51745 .85761 L p .51993 .88403 L .914 .646 .517 r
F P
0 g
s
.52 .88473 m .51978 .8822 L .52 .88486 L 0 0 .265 r
F
.51772 .85823 m .52 .88486 L p .51978 .8822 L F P
0 g
s
.51745 .85761 m .51773 .85778 L p .51969 .88109 L .914 .646 .517 r
F P
0 g
s
.51992 .8838 m .51745 .85761 L .51969 .88109 L .914 .646 .517 r
F
.52002 .88466 m .52 .88458 L .52 .88486 L .391 .225 .533 r
F
.52002 .88466 m .52 .88459 L .52 .88456 L F
.52 .88486 m .52 .88456 L .52002 .88466 L F
.52 .88486 m .51997 .88429 L .52002 .88458 L .141 .057 .492 r
F
.51849 .87466 m .51446 .8432 L p .51677 .86805 L 0 0 .465 r
F P
0 g
s
.51446 .8432 m .51374 .84128 L .51677 .86805 L 0 0 .465 r
F
.5199 .88572 m .51876 .87722 L p .51877 .87691 L .976 .694 .301 r
F P
0 g
s
.51396 .83958 m .49918 .76513 L .49974 .76908 L .901 .613 .111 r
F
.5013 .77683 m .49974 .76908 L .50105 .77566 L F
.49984 .76984 m .51349 .8379 L p .49974 .76908 L F P
0 g
s
.51877 .87691 m .5135 .83793 L .51405 .84025 L .976 .694 .301 r
F
.51349 .8379 m .51419 .84137 L p .51351 .83796 L .901 .613 .111 r
F P
0 g
s
.49974 .76908 m .51349 .8379 L .51351 .83796 L .901 .613 .111 r
F
.51876 .87722 m .51345 .8377 L p .51877 .87691 L .976 .694 .301 r
F P
0 g
s
.51117 .82575 m .51324 .83662 L .51351 .83798 L .901 .613 .111 r
F
.51117 .82575 m .51351 .83798 L .51363 .83849 L F
.51117 .82575 m .5013 .77683 L .51324 .83662 L F
.5013 .77683 m .50105 .77566 L .51351 .83796 L F
.51351 .83798 m .51351 .83796 L .51419 .84137 L F
.5013 .77683 m .51351 .83796 L .51351 .83798 L F
.51349 .8379 m .51332 .83702 L p .51334 .83688 L .913 .659 .541 r
F P
0 g
s
.51334 .83688 m .51338 .83714 L .51346 .83776 L .976 .694 .301 r
F
.51345 .83771 m .51346 .83776 L .51877 .87691 L F
.51334 .83688 m .51346 .83776 L .51345 .83771 L F
.51368 .83736 m .50336 .7822 L .51397 .83963 L .901 .613 .111 r
F
.5034 .78237 m .50336 .7822 L .51368 .83736 L F
.5034 .78237 m .51368 .83736 L .51368 .83733 L F
.50357 .78307 m .5034 .78237 L .51368 .83733 L F
.50357 .78307 m .51368 .83733 L .51366 .83721 L F
.51346 .83558 m .50357 .78307 L p .51366 .83721 L F P
0 g
s
.51334 .83688 m .49984 .76984 L .51091 .82112 L .913 .659 .541 r
F
.51332 .83702 m .49984 .76984 L p .51334 .83688 L F P
0 g
s
.51278 .83276 m .51292 .83361 L .51877 .87691 L .976 .694 .301 r
F
.51349 .83794 m .51334 .83683 L .51345 .8377 L .889 .633 .546 r
F
.51334 .83688 m .51278 .83276 L p .51338 .83714 L .976 .694 .301 r
F P
0 g
s
.51278 .83276 m .50885 .80348 L p .51292 .83361 L .976 .694 .301 r
F P
0 g
s
.50885 .80348 m .51278 .83276 L p .5128 .8328 L .889 .633 .546 r
F P
0 g
s
.50885 .80348 m .50943 .80442 L p .51405 .84025 L .976 .694 .301 r
F P
0 g
s
.51381 .83923 m .50885 .80348 L .51405 .84025 L .976 .694 .301 r
F
.5135 .83793 m .51292 .83361 L .51381 .83923 L F
.51292 .83361 m .50885 .80348 L .51381 .83923 L F
.51515 .84352 m .51359 .83494 L p .51595 .8413 L .847 .618 .591 r
F P
0 g
s
.51405 .84025 m .51278 .83045 L .51405 .84026 L .976 .694 .301 r
F
.51877 .87689 m .51504 .84797 L p .51851 .87481 L .714 .508 0 r
F P
0 g
s
.5144 .84296 m .51279 .83049 L p .51505 .848 L .714 .508 0 r
F P
0 g
s
.51279 .83049 m .50943 .80442 L p .51295 .83174 L .714 .508 0 r
F P
0 g
s
.51279 .83049 m .51295 .83174 L .51695 .86273 L .714 .508 0 r
F
.51656 .85973 m .51695 .86273 L .51851 .87481 L F
.51279 .83049 m .51695 .86273 L .51656 .85973 L F
.51252 .82819 m .5098 .80706 L .51217 .82536 L F
.51252 .82819 m .51217 .82536 L .51268 .82935 L F
.51295 .83174 m .50943 .80442 L .5132 .83351 L F
.51295 .83174 m .5132 .83351 L .51346 .83558 L F
.50943 .80442 m .5098 .80706 L p .51288 .83098 L F P
0 g
s
.5132 .83351 m .50943 .80442 L .51288 .83098 L .714 .508 0 r
F
.51788 .86996 m .51295 .83174 L .51394 .83929 L F
.51851 .87481 m .51695 .86273 L .51788 .86996 L F
.51695 .86273 m .51295 .83174 L .51788 .86996 L F
.51405 .84025 m .51405 .84026 L .51877 .87689 L .976 .694 .301 r
F
.51457 .8435 m .51432 .84216 L .51533 .84553 L 0 0 .426 r
F
.51346 .83558 m .51533 .84553 L p .51432 .84216 L F P
0 g
s
.51851 .87481 m .51394 .83929 L .51445 .84313 L .714 .508 0 r
F
.51496 .84455 m .51397 .83963 L .51485 .84424 L .901 .613 .111 r
F
.51751 .85943 m .51457 .8435 L .5155 .84645 L 0 0 .426 r
F
.51496 .84455 m .51485 .84424 L .51634 .85211 L .901 .613 .111 r
F
.51588 .84911 m .51496 .84455 L .51634 .85211 L F
.51533 .84553 m .5155 .84645 L p .51457 .8435 L 0 0 .426 r
F P
0 g
s
.51405 .84026 m .51877 .87689 L p .51405 .84026 L .976 .694 .301 r
F P
0 g
s
.51419 .84137 m .51634 .85211 L p .51426 .84117 L .901 .613 .111 r
F P
0 g
s
.51397 .83963 m .51405 .84026 L .51426 .84117 L .901 .613 .111 r
F
.51405 .84026 m .51419 .84137 L .51426 .84117 L F
.51279 .83049 m .51405 .84026 L p .51278 .83045 L .976 .694 .301 r
F P
0 g
s
.51278 .83045 m .51405 .84026 L .51405 .84026 L .976 .694 .301 r
F
.51588 .84911 m .51368 .83736 L .51397 .83963 L .901 .613 .111 r
F
.51368 .83733 m .51368 .83736 L .51588 .84911 L F
.51366 .83721 m .51368 .83733 L .51588 .84911 L F
.51533 .84553 m .51346 .83558 L p .51588 .84911 L F P
0 g
s
.51346 .83558 m .51366 .83721 L .51588 .84911 L .901 .613 .111 r
F
.51745 .85761 m .51772 .85823 L p .51588 .84911 L F P
0 g
s
.51634 .85211 m .51745 .85761 L p .51588 .84911 L .901 .613 .111 r
F P
0 g
s
.51772 .85823 m .51533 .84553 L p .51588 .84911 L .901 .613 .111 r
F P
0 g
s
.51745 .85761 m .51523 .84644 L .51519 .84501 L .913 .659 .541 r
F
.51745 .85761 m .51419 .84137 L p .51412 .84083 L F P
0 g
s
.51773 .85778 m .51515 .84352 L p .51595 .8413 L .847 .618 .591 r
F P
0 g
s
.51709 .84439 m .51595 .8413 L .51683 .83886 L .847 .618 .591 r
F
.51773 .85778 m .51595 .8413 L .51709 .84439 L F
.51519 .84501 m .51412 .84083 L .51351 .8356 L .913 .659 .541 r
F
.51523 .84644 m .51412 .84083 L .51519 .84501 L F
.51412 .84083 m .51334 .83688 L .51347 .83584 L F
.51412 .84083 m .51347 .83584 L .51351 .8356 L F
.51405 .84026 m .51363 .83849 L .51419 .84137 L .901 .613 .111 r
F
.51412 .84083 m .51349 .8379 L .51334 .83688 L .913 .659 .541 r
F
.51363 .83849 m .51351 .83798 L .51419 .84137 L .901 .613 .111 r
F
.51419 .84137 m .51349 .8379 L p .51412 .84083 L .913 .659 .541 r
F P
0 g
s
.51877 .87691 m .51405 .84025 L .51877 .87689 L .976 .694 .301 r
F
.51457 .8435 m .51441 .84281 L .51437 .84253 L 0 0 .426 r
F
.5098 .80706 m .51004 .80877 L p .51268 .82935 L .714 .508 0 r
F P
0 g
s
.50943 .80442 m .51279 .83049 L p .51278 .83045 L .976 .694 .301 r
F P
0 g
s
.51334 .83683 m .51421 .82195 L .50885 .80348 L .889 .633 .546 r
F
.51364 .83456 m .51421 .82195 L .51334 .83683 L F
.51277 .83042 m .50678 .79732 L p .51106 .81716 L .847 .618 .591 r
F P
0 g
s
.50642 .79445 m .50294 .78152 L p .50608 .79278 L .952 .641 .212 r
F P
0 g
s
.50608 .79278 m .50294 .78152 L .50256 .77995 L .952 .641 .212 r
F
.50608 .79278 m .50052 .7725 L .49925 .76748 L F
.50608 .79278 m .4789 .69215 L .47905 .69254 L F
.49948 .76727 m .50608 .79278 L .49979 .76946 L F
.50336 .7822 m .49731 .74986 L .49779 .75943 L .901 .613 .111 r
F
.49744 .75052 m .49731 .74986 L .50336 .7822 L F
.49744 .75052 m .50336 .7822 L .5034 .78237 L F
.50357 .78307 m .49744 .75052 L p .5034 .78237 L F P
0 g
s
.50499 .78739 m .49918 .76513 L p .50357 .78307 L 0 0 .316 r
F P
0 g
s
.49948 .76727 m .50499 .78739 L .50608 .79278 L .952 .641 .212 r
F
.49918 .76513 m .50499 .78739 L p .49948 .76727 L F P
0 g
s
.50885 .80348 m .50642 .79445 L p .50608 .79278 L .952 .641 .212 r
F P
0 g
s
.50294 .78152 m .50642 .79445 L p .50405 .77999 L .884 .656 .588 r
F P
0 g
s
.50339 .77858 m .50405 .77999 L .5036 .77721 L .884 .656 .588 r
F
.50294 .78152 m .50405 .77999 L .50339 .77858 L F
.49496 .73201 m .50339 .77858 L p .50324 .77827 L .913 .659 .541 r
F P
0 g
s
.49495 .73728 m .49625 .74416 L p .49596 .74276 L 0 0 .426 r
F P
0 g
s
.49732 .7501 m .49731 .74986 L .49721 .74952 L .901 .613 .111 r
F
.49732 .7501 m .49721 .74952 L .49186 .73001 L F
.49779 .75943 m .49732 .7501 L .49186 .73001 L F
.49744 .75052 m .49495 .73728 L p .49731 .74986 L F P
0 g
s
.49495 .73728 m .4944 .73598 L p .49731 .74986 L .901 .613 .111 r
F P
0 g
s
.4944 .73598 m .49186 .73001 L p .49731 .74986 L .901 .613 .111 r
F P
0 g
s
.50324 .77827 m .49186 .73001 L .49776 .74765 L .913 .659 .541 r
F
.50188 .77539 m .49186 .73001 L .50324 .77827 L F
.50274 .7772 m .5036 .77721 L .50238 .76978 L .884 .656 .588 r
F
.50294 .78152 m .50339 .77858 L .50274 .7772 L F
.50339 .77858 m .5036 .77721 L .50274 .7772 L F
.49974 .76908 m .49918 .76513 L .49779 .75943 L .901 .613 .111 r
F
.49974 .76908 m .49779 .75943 L .49984 .76984 L F
.49186 .73001 m .4976 .75865 L p .49779 .75943 L F P
0 g
s
.4976 .75865 m .49984 .76984 L p .49779 .75943 L .901 .613 .111 r
F P
0 g
s
.49984 .76984 m .49186 .73001 L p .50188 .77539 L .913 .659 .541 r
F P
0 g
s
.49984 .76984 m .50188 .77539 L .50052 .7725 L .913 .659 .541 r
F
.49282 .73683 m .50125 .77406 L .50238 .76978 L .884 .656 .588 r
F
.50125 .77406 m .50294 .78152 L .50274 .7772 L F
.50125 .77406 m .50274 .7772 L .50238 .76978 L F
.50294 .78152 m .50052 .7725 L p .50256 .77995 L .952 .641 .212 r
F P
0 g
s
.50052 .7725 m .50294 .78152 L p .50125 .77406 L .884 .656 .588 r
F P
0 g
s
.50052 .7725 m .50188 .77539 L .50633 .79554 L .913 .659 .541 r
F
.50633 .79554 m .50188 .77539 L .50324 .77827 L F
.50633 .79554 m .50324 .77827 L .50678 .79732 L F
.50339 .77858 m .50678 .79732 L p .50324 .77827 L F P
0 g
s
.50642 .79445 m .50885 .80348 L p .50238 .76978 L .884 .656 .588 r
F P
0 g
s
.50678 .79732 m .50461 .78534 L p .50833 .80371 L .847 .618 .591 r
F P
0 g
s
.50678 .79732 m .50833 .80371 L .51106 .81716 L .847 .618 .591 r
F
.50034 .77212 m .50002 .77067 L .50052 .7725 L .913 .659 .541 r
F
.50002 .77067 m .49984 .76984 L .50052 .7725 L F
.50052 .7725 m .4789 .69215 L p .49925 .76748 L .952 .641 .212 r
F P
0 g
s
.51773 .85778 m .51745 .85761 L p .50633 .79554 L .913 .659 .541 r
F P
0 g
s
.50678 .79732 m .51773 .85778 L p .50633 .79554 L .913 .659 .541 r
F P
0 g
s
.51756 .85931 m .51595 .8413 L p .51349 .83794 L .889 .633 .546 r
F P
0 g
s
.51595 .8413 m .51421 .82195 L p .51362 .83503 L .889 .633 .546 r
F P
0 g
s
.51595 .8413 m .51362 .83503 L .51349 .83794 L .889 .633 .546 r
F
.51999 .88459 m .51745 .85761 L .51999 .88459 L .914 .646 .517 r
F
.51987 .88391 m .5195 .88205 L .51905 .87805 L .391 .225 .533 r
F
.51745 .85761 m .52 .88486 L p .52 .88488 L .943 .682 .216 r
F P
0 g
s
.51996 .88507 m .51465 .82655 L .51992 .88556 L .811 .574 .581 r
F
.51274 .80549 m .51466 .8269 L p .51465 .82655 L F P
0 g
s
.51466 .8269 m .51992 .88556 L p .51465 .82655 L .811 .574 .581 r
F P
0 g
s
.51328 .83437 m .51333 .83677 L .51218 .82819 L .913 .659 .541 r
F
.51334 .83683 m .5128 .8328 L .51334 .83688 L .889 .633 .546 r
F
.51278 .83276 m .51334 .83688 L p .5128 .8328 L F P
0 g
s
.51333 .83677 m .51334 .83688 L .51218 .82819 L .913 .659 .541 r
F
.51218 .82819 m .51334 .83688 L .51091 .82112 L F
.51351 .8356 m .51347 .83584 L .51328 .83437 L F
.51347 .83584 m .51334 .83688 L .51328 .83437 L F
.51359 .83494 m .51277 .83042 L p .51595 .8413 L .847 .618 .591 r
F P
0 g
s
.51277 .83042 m .51106 .81716 L .51595 .8413 L .847 .618 .591 r
F
.51504 .84797 m .5144 .84296 L p .51656 .85973 L .714 .508 0 r
F P
0 g
s
.5144 .84296 m .51505 .848 L .51656 .85973 L .714 .508 0 r
F
.51504 .84797 m .51656 .85973 L .51851 .87481 L F
.51349 .83794 m .51364 .83456 L .51334 .83683 L .889 .633 .546 r
F
.51701 .80631 m .51545 .77304 L p .51527 .80444 L .719 .407 .45 r
F P
0 g
s
.51545 .77304 m .51542 .77281 L p .51476 .77239 L .687 .306 .32 r
F P
0 g
s
.51419 .78312 m .50598 .76679 L .51209 .7389 L .847 .618 .591 r
F
.51592 .78275 m .51209 .7389 L .5164 .78203 L .774 .561 .611 r
F
.51274 .80549 m .5221 .80803 L p .51776 .79021 L .807 .607 .632 r
F P
0 g
s
.51594 .78273 m .51551 .78575 L .51776 .79021 L .807 .607 .632 r
F
.51551 .78575 m .51274 .80549 L .51776 .79021 L F
.51016 .79154 m .51274 .80549 L p .51551 .78575 L F P
0 g
s
.51484 .78441 m .51551 .78575 L .51594 .78273 L .807 .607 .632 r
F
.51016 .79154 m .51551 .78575 L .51484 .78441 L F
.51209 .7389 m .51484 .78441 L p .51419 .78312 L .847 .618 .591 r
F P
0 g
s
.51794 .8358 m .51484 .78441 L p .52117 .84284 L .774 .561 .611 r
F P
0 g
s
.51484 .78441 m .51209 .7389 L p .51592 .78275 L .774 .561 .611 r
F P
0 g
s
.51484 .78441 m .51592 .78275 L .52117 .84284 L .774 .561 .611 r
F
.48533 .6571 m .49527 .71092 L p .49736 .70649 L .807 .607 .632 r
F P
0 g
s
.46954 .65147 m .46663 .64653 L p .44666 .59719 L .88 .525 .034 r
F P
0 g
s
.44546 .59325 m .46954 .65147 L .44666 .59719 L .88 .525 .034 r
F
.46663 .64653 m .48533 .6571 L p .46382 .61437 L .872 .683 .644 r
F P
0 g
s
.46078 .631 m .46663 .64653 L .46527 .63097 L .872 .683 .644 r
F
.46177 .62559 m .46527 .63097 L .46382 .61437 L F
.46078 .631 m .46527 .63097 L .46177 .62559 L F
.44445 .57002 m .46177 .62559 L p .4602 .62316 L .907 .678 .581 r
F P
0 g
s
.43491 .56335 m .44445 .57002 L p .4602 .62316 L .907 .678 .581 r
F P
0 g
s
.44501 .59178 m .44154 .58037 L .43897 .5751 L .781 .455 0 r
F
.43491 .56335 m .44023 .57608 L p .43011 .55501 L 0 0 .495 r
F P
0 g
s
.44023 .57608 m .44496 .5874 L p .43011 .55501 L 0 0 .495 r
F P
0 g
s
.44496 .5874 m .43491 .56335 L p .43897 .5751 L .781 .455 0 r
F P
0 g
s
.45389 .61347 m .43491 .56335 L .4602 .62316 L .907 .678 .581 r
F
.44445 .57002 m .45116 .57297 L p .46332 .61712 L .839 .649 .645 r
F P
0 g
s
.46205 .62405 m .44445 .57002 L .46332 .61712 L .839 .649 .645 r
F
.47573 .66603 m .47778 .67232 L .47904 .67312 L .884 .656 .588 r
F
.46954 .65147 m .47165 .66535 L p .47727 .68417 L 0 0 .316 r
F P
0 g
s
.46663 .64653 m .46954 .65147 L p .47029 .65436 L .952 .641 .212 r
F P
0 g
s
.4815 .69832 m .47937 .69347 L .48351 .70505 L .781 .455 0 r
F
.47761 .68528 m .46288 .64112 L p .4815 .69832 L F P
0 g
s
.46288 .64112 m .46044 .6344 L .4815 .69832 L .781 .455 0 r
F
.46663 .64653 m .47069 .6559 L p .48308 .70384 L .952 .641 .212 r
F P
0 g
s
.46954 .65147 m .48351 .70505 L p .48308 .70384 L .952 .641 .212 r
F P
0 g
s
.46663 .64653 m .47778 .67232 L .47573 .66603 L .884 .656 .588 r
F
.46663 .64653 m .47437 .67016 L .47778 .67232 L F
.47329 .66948 m .47855 .6794 L .47904 .67312 L F
.49496 .73201 m .4914 .71911 L .49514 .71121 L .839 .649 .645 r
F
.47904 .67312 m .48223 .68635 L .49062 .69793 L .884 .656 .588 r
F
.4914 .71911 m .48254 .68692 L .49514 .71121 L .839 .649 .645 r
F
.48254 .68692 m .49464 .70974 L .49514 .71121 L F
.49496 .73201 m .47855 .6794 L p .48254 .68692 L F P
0 g
s
.47329 .66948 m .47805 .68588 L .47855 .6794 L .884 .656 .588 r
F
.47908 .68689 m .47823 .68355 L .49496 .73201 L .907 .678 .581 r
F
.46899 .64786 m .46663 .64653 L p .47904 .67312 L .884 .656 .588 r
F P
0 g
s
.471 .65517 m .47855 .6794 L p .47823 .68355 L .907 .678 .581 r
F P
0 g
s
.47855 .6794 m .49496 .73201 L p .47823 .68355 L .907 .678 .581 r
F P
0 g
s
.47904 .67312 m .49062 .69793 L .47917 .67143 L .884 .656 .588 r
F
.47904 .67312 m .47917 .67143 L .46899 .64786 L F
.49495 .73728 m .49384 .73393 L p .49269 .73073 L .781 .455 0 r
F P
0 g
s
.49186 .73001 m .49496 .73201 L p .49776 .74765 L .913 .659 .541 r
F P
0 g
s
.48223 .68635 m .50238 .76978 L .49162 .70404 L .884 .656 .588 r
F
.48223 .68635 m .49162 .70404 L .49062 .69793 L F
.47937 .69347 m .47926 .69315 L .48482 .70941 L .781 .455 0 r
F
.47317 .67419 m .45176 .61257 L .47937 .69347 L F
.48351 .70505 m .47937 .69347 L .48482 .70941 L F
.44814 .60206 m .47926 .69315 L p .44787 .60116 L F P
0 g
s
.47734 .68751 m .47926 .69315 L .47937 .69347 L .781 .455 0 r
F
.44787 .60116 m .4777 .68856 L .47734 .68751 L F
.4777 .68856 m .47926 .69315 L .47734 .68751 L F
.47317 .67419 m .44712 .59872 L .45176 .61257 L F
.44712 .59872 m .44787 .60116 L .45176 .61257 L F
.47926 .69315 m .44814 .60206 L p .46515 .64401 L .907 .678 .581 r
F P
0 g
s
.48308 .70384 m .47926 .69315 L .46663 .64653 L .952 .641 .212 r
F
.47926 .69315 m .47905 .69254 L .47838 .69009 L F
.47926 .69315 m .47838 .69009 L .46663 .64653 L F
.4789 .69215 m .47786 .68827 L p .47905 .69254 L F P
0 g
s
.47786 .68827 m .46663 .64653 L p .47838 .69009 L .952 .641 .212 r
F P
0 g
s
.47905 .69254 m .47786 .68827 L .47838 .69009 L .952 .641 .212 r
F
.46663 .64653 m .44836 .60279 L p .44666 .59719 L .88 .525 .034 r
F P
0 g
s
.45091 .60889 m .46663 .64653 L p .45268 .61161 L .872 .683 .644 r
F P
0 g
s
.43823 .57307 m .44814 .60206 L p .43897 .5751 L .781 .455 0 r
F P
0 g
s
.43897 .5751 m .44814 .60206 L .44787 .60116 L .781 .455 0 r
F
.44712 .59872 m .43897 .5751 L .44787 .60116 L F
.43491 .56335 m .43823 .57307 L p .43897 .5751 L F P
0 g
s
.44814 .60206 m .43491 .56335 L p .45389 .61347 L .907 .678 .581 r
F P
0 g
s
.44814 .60206 m .45389 .61347 L .4511 .60918 L .907 .678 .581 r
F
.43491 .56335 m .43164 .55686 L p .42064 .52856 L .892 .705 .641 r
F P
0 g
s
.44305 .58751 m .45268 .61161 L .42685 .54089 L .872 .683 .644 r
F
.45268 .61161 m .46663 .64653 L .45459 .61454 L F
.45268 .61161 m .45459 .61454 L .42685 .54089 L F
.4511 .60918 m .45389 .61347 L .46421 .64073 L .907 .678 .581 r
F
.46663 .64653 m .46719 .64859 L p .46738 .64908 L .884 .656 .588 r
F P
0 g
s
.46719 .64859 m .47786 .68827 L p .46738 .64908 L .884 .656 .588 r
F P
0 g
s
.46738 .64908 m .47786 .68827 L .47805 .68588 L .884 .656 .588 r
F
.46663 .64653 m .47329 .66948 L .47437 .67016 L F
.46515 .64401 m .46421 .64073 L .46968 .65518 L .907 .678 .581 r
F
.44814 .60206 m .46421 .64073 L .46515 .64401 L F
.49186 .73001 m .47926 .69315 L p .4779 .68775 L F P
0 g
s
.49186 .73001 m .4779 .68775 L .4782 .68392 L .907 .678 .581 r
F
.47908 .68689 m .4782 .68392 L .47823 .68355 L F
.47805 .68588 m .50238 .76978 L .48223 .68635 L .884 .656 .588 r
F
.47855 .6794 m .48223 .68635 L .47904 .67312 L F
.47805 .68588 m .48223 .68635 L .47855 .6794 L F
.48379 .65623 m .46899 .64786 L p .49062 .69793 L F P
0 g
s
.47855 .6794 m .44445 .57002 L p .48254 .68692 L .839 .649 .645 r
F P
0 g
s
.42685 .54089 m .45459 .61454 L .46382 .61437 L .872 .683 .644 r
F
.45459 .61454 m .46078 .631 L .46177 .62559 L F
.45459 .61454 m .46177 .62559 L .46382 .61437 L F
.46968 .65518 m .46421 .64073 L .4602 .62316 L .907 .678 .581 r
F
.46421 .64073 m .45389 .61347 L .45835 .62033 L F
.46421 .64073 m .45835 .62033 L .4602 .62316 L F
.4782 .68392 m .4779 .68775 L .46968 .65518 L F
.47926 .69315 m .46515 .64401 L .46968 .65518 L F
.47805 .68588 m .4789 .69215 L .48582 .71268 L .884 .656 .588 r
F
.47786 .68827 m .4789 .69215 L p .47805 .68588 L F P
0 g
s
.50238 .76978 m .51274 .80549 L .4939 .70834 L .884 .656 .588 r
F
.49162 .70404 m .4939 .70834 L .48379 .65623 L F
.50238 .76978 m .4939 .70834 L .49162 .70404 L F
.48533 .6571 m .48874 .65811 L p .47041 .60055 L .796 .645 .694 r
F P
0 g
s
.46719 .59593 m .48533 .6571 L p .47041 .60055 L .796 .645 .694 r
F P
0 g
s
.48874 .65811 m .51633 .66633 L p .50194 .64572 L .796 .645 .694 r
F P
0 g
s
.49617 .5927 m .50194 .64572 L p .49578 .63689 L .839 .649 .645 r
F P
0 g
s
.51111 .63712 m .50194 .64572 L .52958 .67028 L .796 .645 .694 r
F
.51633 .66633 m .52958 .67028 L p .50194 .64572 L F P
0 g
s
.49617 .5927 m .52301 .59806 L p .52068 .62814 L .77 .599 .667 r
F P
0 g
s
.51111 .63712 m .52958 .67028 L .51852 .63017 L .796 .645 .694 r
F
.51501 .70125 m .50489 .64295 L .52068 .62814 L .77 .599 .667 r
F
.50489 .64295 m .49617 .5927 L .52068 .62814 L F
.50646 .68722 m .49617 .5927 L p .51501 .70125 L F P
0 g
s
.46641 .5569 m .51111 .63712 L .49517 .5455 L .796 .645 .694 r
F
.51111 .63712 m .51852 .63017 L .49517 .5455 L F
.44335 .51551 m .48874 .65811 L .51111 .63712 L F
.51209 .7389 m .51501 .70125 L .52372 .71554 L .77 .599 .667 r
F
.50536 .76557 m .51016 .79154 L p .51484 .78441 L .807 .607 .632 r
F P
0 g
s
.50957 .73788 m .51209 .7389 L p .50598 .76679 L .847 .618 .591 r
F P
0 g
s
.48533 .6571 m .49736 .70649 L .50589 .68843 L .807 .607 .632 r
F
.48874 .65811 m .48533 .6571 L p .52403 .71605 L F P
0 g
s
.49514 .71121 m .49578 .63689 L .50646 .68722 L .839 .649 .645 r
F
.49514 .71121 m .49464 .70974 L .49514 .71068 L F
.4939 .70834 m .5052 .76663 L .49527 .71092 L .884 .656 .588 r
F
.51274 .80549 m .50536 .76557 L p .5052 .76663 L F P
0 g
s
.50536 .76557 m .49527 .71092 L p .5052 .76663 L .884 .656 .588 r
F P
0 g
s
.50339 .77858 m .50536 .76557 L .50598 .76679 L .847 .618 .591 r
F
.50536 .76557 m .50957 .73788 L .50598 .76679 L F
.50589 .68843 m .51594 .78273 L .52403 .71605 L .807 .607 .632 r
F
.49736 .70649 m .51594 .78273 L .50589 .68843 L F
.50194 .64572 m .50646 .68722 L p .49578 .63689 L .839 .649 .645 r
F P
0 g
s
.49514 .71068 m .46332 .61712 L .49578 .63689 L .839 .649 .645 r
F
.49464 .70974 m .46332 .61712 L .49514 .71068 L F
.48254 .68692 m .46332 .61712 L .49464 .70974 L F
.46295 .63245 m .4602 .62316 L .4668 .6417 L .907 .678 .581 r
F
.46177 .62559 m .4668 .6417 L p .4602 .62316 L F P
0 g
s
.49736 .70649 m .50536 .76557 L .51594 .78273 L .807 .607 .632 r
F
.49527 .71092 m .50536 .76557 L p .49736 .70649 L F P
0 g
s
.50536 .76557 m .51484 .78441 L .51594 .78273 L .807 .607 .632 r
F
.5193 .85837 m .51773 .85778 L p .51683 .83886 L .847 .618 .591 r
F P
0 g
s
.51683 .83886 m .51779 .8362 L .5193 .85837 L .847 .618 .591 r
F
.51598 .84122 m .51794 .8358 L .51488 .82914 L .811 .574 .581 r
F
.51484 .78441 m .5193 .85837 L p .51419 .78312 L .847 .618 .591 r
F P
0 g
s
.51859 .84664 m .51794 .8358 L p .52117 .84284 L .774 .561 .611 r
F P
0 g
s
.52005 .88523 m .52 .88473 L .52003 .8851 L 0 0 .265 r
F
.52002 .88527 m .51998 .88506 L .52002 .88519 L .649 .393 .518 r
F
.51998 .8853 m .52 .88486 L .52003 .88545 L .728 .45 .506 r
F
.51998 .88515 m .51999 .88507 L .51999 .88509 L .62 .394 .553 r
F
.51995 .8854 m .51999 .88504 L .51998 .88516 L .883 .733 .222 r
F
.52002 .88576 m .52 .88486 L p .51996 .88527 L .728 .45 .506 r
F P
0 g
s
.52 .88486 m .52002 .88576 L .51997 .88575 L p .809 .517 .495 r
F P
0 g
s
.52001 .88502 m .52 .88486 L .52 .88488 L .943 .682 .216 r
F
.52 .88488 m .52 .88485 L .52 .88487 L .95 .6 .297 r
F
.52002 .88458 m .51995 .88391 L .52 .88486 L .141 .057 .492 r
F
.52 .88486 m .52 .88484 L .52 .88488 L .95 .6 .297 r
F
.52001 .885 m .52 .88496 L .52001 .885 L .943 .682 .216 r
F
.52 .88473 m .52 .88486 L .52003 .88505 L 0 0 .265 r
F
.51995 .8854 m .51999 .88499 L .51993 .88547 L .883 .733 .222 r
F
.51993 .88557 m .52 .88486 L .52 .88492 L .993 .746 .397 r
F
.51993 .88551 m .51993 .8855 L .52 .88486 L .809 .517 .495 r
F
.51991 .88571 m .51993 .88547 L .51991 .88573 L .883 .733 .222 r
F
.51993 .88551 m .52 .88486 L .51993 .8855 L .809 .517 .495 r
F
.51991 .88573 m .51993 .88557 L .52 .88486 L .993 .746 .397 r
F
.51993 .88557 m .51998 .88509 L .52 .88486 L F
.51997 .88525 m .51993 .88547 L .51991 .88571 L .883 .733 .222 r
F
.51994 .88574 m .52 .88486 L p .51993 .88551 L .809 .517 .495 r
F P
0 g
s
.51993 .88565 m .51993 .88551 L .52 .88486 L .809 .517 .495 r
F
.51991 .88572 m .52 .88486 L p .51993 .88557 L .993 .746 .397 r
F P
0 g
s
.51989 .88436 m .51794 .8358 L .51598 .84122 L .811 .574 .581 r
F
.51992 .88451 m .51999 .88476 L .51991 .88423 L .809 .517 .495 r
F
.51997 .88524 m .52 .88486 L .51998 .8853 L .728 .45 .506 r
F
.52002 .88576 m .51986 .88205 L .51989 .88436 L .811 .574 .581 r
F
.51986 .88205 m .51794 .8358 L .51989 .88436 L F
.51993 .88586 m .51992 .88586 L p .51993 .88582 L .685 .407 .496 r
F P
0 g
s
.51993 .88586 m .51999 .88499 L .51992 .88585 L .698 .467 .569 r
F
.51997 .88575 m .51994 .88574 L p .52 .88486 L .809 .517 .495 r
F P
0 g
s
.52 .88496 m .52 .88488 L .52 .88486 L .943 .682 .216 r
F
.52003 .88475 m .52 .88486 L .52002 .88458 L .141 .057 .492 r
F
.51982 .88582 m .52 .88486 L p .51982 .88582 L .44 .23 .497 r
F P
0 g
s
.51985 .88572 m .51982 .88582 L .51982 .88583 L .62 .394 .553 r
F
.51982 .88582 m .51999 .88479 L .51987 .88554 L .44 .23 .497 r
F
.51982 .88582 m .51987 .88554 L .51983 .88586 L F
.51982 .88582 m .51987 .88554 L .51982 .88582 L F
.52002 .88576 m .52117 .84284 L p .51794 .8358 L .811 .574 .581 r
F P
0 g
s
.51683 .83886 m .51419 .78312 L .51779 .8362 L .847 .618 .591 r
F
.51421 .82195 m .51274 .80549 L .50933 .80373 L p .889 .633 .546 r
F P
0 g
s
.5164 .82973 m .50339 .77858 L .50598 .76679 L .847 .618 .591 r
F
.5164 .82973 m .50598 .76679 L .51419 .78312 L F
.52117 .84284 m .51592 .78275 L .5223 .84117 L .774 .561 .611 r
F
.51592 .78275 m .5164 .78203 L .5223 .84117 L F
.51595 .8413 m .50766 .80039 L .51683 .83886 L .847 .618 .591 r
F
.50766 .80039 m .50461 .78534 L .50679 .79487 L F
.50766 .80039 m .50679 .79487 L .51683 .83886 L F
.51345 .8377 m .51334 .83688 L p .51345 .83771 L .976 .694 .301 r
F P
0 g
s
.51334 .83688 m .51345 .8377 L p .51334 .83683 L .889 .633 .546 r
F P
0 g
s
.51345 .8377 m .51345 .83771 L .51877 .87691 L .976 .694 .301 r
F
.49989 .77019 m .50052 .7725 L p .49282 .73683 L .884 .656 .588 r
F P
0 g
s
.50052 .7725 m .50125 .77406 L .49282 .73683 L .884 .656 .588 r
F
.51091 .82112 m .50034 .77212 L .50633 .79554 L .913 .659 .541 r
F
.50034 .77212 m .50052 .7725 L .50633 .79554 L F
.50933 .80373 m .50885 .80348 L p .51421 .82195 L .889 .633 .546 r
F P
0 g
s
.51683 .83886 m .50679 .79487 L .5164 .82973 L .847 .618 .591 r
F
.50679 .79487 m .50461 .78534 L .50626 .79157 L F
.50679 .79487 m .50626 .79157 L .5164 .82973 L F
.5199 .88572 m .51994 .88574 L p .51349 .83794 L .889 .633 .546 r
F P
0 g
s
.51345 .8377 m .5199 .88572 L p .51349 .83794 L .889 .633 .546 r
F P
0 g
s
.48582 .71268 m .49989 .77019 L .49282 .73683 L .884 .656 .588 r
F
.48726 .72 m .48734 .72032 L .48735 .72036 L .952 .641 .212 r
F
.4789 .69215 m .49989 .77019 L p .48582 .71268 L .884 .656 .588 r
F P
0 g
s
.44836 .60279 m .45091 .60889 L p .44305 .58751 L .872 .683 .644 r
F P
0 g
s
.45091 .60889 m .45268 .61161 L .44305 .58751 L .872 .683 .644 r
F
.4178 .52937 m .41434 .52137 L .42064 .52856 L .892 .705 .641 r
F
.41434 .52137 m .40437 .49833 L .41496 .518 L F
.41434 .52137 m .41496 .518 L .42064 .52856 L F
.42668 .54786 m .43954 .57961 L .4178 .52937 L .88 .525 .034 r
F
.43164 .55686 m .4178 .52937 L p .42064 .52856 L .892 .705 .641 r
F P
0 g
s
.4178 .52937 m .43164 .55686 L p .42861 .55188 L 0 0 .495 r
F P
0 g
s
.41689 .52747 m .44836 .60279 L p .42677 .54676 L .872 .683 .644 r
F P
0 g
s
.44836 .60279 m .44305 .58751 L .42677 .54676 L .872 .683 .644 r
F
.48254 .68692 m .46205 .62405 L .46332 .61712 L .839 .649 .645 r
F
.45901 .56471 m .46719 .59593 L .47041 .60055 L .796 .645 .694 r
F
.44335 .51551 m .45819 .56557 L p .45901 .56471 L F P
0 g
s
.45819 .56557 m .46719 .59593 L p .45901 .56471 L .796 .645 .694 r
F P
0 g
s
.45116 .57297 m .44445 .57002 L p .48 .5426 L .818 .684 .716 r
F P
0 g
s
.3701 .43462 m .38727 .44722 L p .40115 .48103 L .892 .705 .641 r
F P
0 g
s
.44445 .57002 m .43645 .53655 L .4506 .55543 L .818 .684 .716 r
F
.40178 .49031 m .40122 .48997 L p .45819 .56557 L .872 .683 .644 r
F P
0 g
s
.41235 .50835 m .38857 .46685 L .40115 .48103 L .892 .705 .641 r
F
.38857 .46685 m .3701 .43462 L .40115 .48103 L F
.39681 .48246 m .3701 .43462 L .41235 .50835 L F
.42064 .52856 m .41496 .518 L .41537 .51572 L F
.40122 .48997 m .46382 .61437 L .46674 .5984 L .872 .683 .644 r
F
.41496 .518 m .3701 .43462 L .41537 .51572 L .892 .705 .641 r
F
.38727 .44722 m .41644 .50987 L p .41537 .51572 L F P
0 g
s
.41644 .50987 m .40509 .48548 L p .43645 .53655 L .818 .684 .716 r
F P
0 g
s
.40509 .48548 m .38727 .44722 L p .40689 .48285 L .818 .684 .716 r
F P
0 g
s
.40509 .48548 m .4182 .50026 L .407 .4827 L .84 .718 .729 r
F
.40509 .48548 m .40689 .48285 L .43645 .53655 L .818 .684 .716 r
F
.4182 .50026 m .44335 .51551 L p .407 .4827 L .84 .718 .729 r
F P
0 g
s
.38506 .46289 m .40509 .48548 L .407 .4827 L .84 .718 .729 r
F
.3701 .43462 m .3602 .40867 L .36172 .40946 L .845 .741 .75 r
F
.35669 .41543 m .35282 .4089 L .3602 .40867 L F
.34852 .40169 m .35282 .4089 L p .35352 .40863 L .84 .718 .729 r
F P
0 g
s
.35282 .4089 m .35728 .40717 L .3602 .40867 L .845 .741 .75 r
F
.38727 .44722 m .3701 .43462 L p .36172 .40946 L F P
0 g
s
.34867 .39913 m .34863 .39988 L p .36542 .41136 L .898 .611 .494 r
F P
0 g
s
.34867 .39913 m .36542 .41136 L .36737 .41236 L .898 .611 .494 r
F
.36283 .41003 m .38727 .44722 L p .36172 .40946 L .845 .741 .75 r
F P
0 g
s
.40998 .44012 m .45202 .4936 L .50491 .51762 L .768 .69 .776 r
F
.40998 .44012 m .42026 .46333 L p .45202 .4936 L F P
0 g
s
.37246 .4234 m .37372 .42121 L .36992 .42026 L .854 .843 .833 r
F
.38727 .44722 m .36283 .41003 L p .42147 .45192 L .768 .725 .812 r
F P
0 g
s
.34863 .39988 m .34852 .40169 L .35728 .40717 L p .898 .611 .494 r
F P
0 g
s
.35728 .40717 m .36972 .41495 L p .36283 .41003 L .898 .611 .494 r
F P
0 g
s
.34863 .39988 m .35728 .40717 L .36283 .41003 L .898 .611 .494 r
F
.34852 .40169 m .35352 .40863 L .35411 .4084 L .84 .718 .729 r
F
.41949 .46445 m .41855 .4617 L .42026 .46333 L .818 .684 .716 r
F
.41106 .47676 m .407 .4827 L .44335 .51551 L .84 .718 .729 r
F
.43645 .53655 m .41855 .4617 L .4506 .55543 L .818 .684 .716 r
F
.41855 .4617 m .38727 .44722 L p .40707 .44994 L .768 .725 .812 r
F P
0 g
s
.40707 .44994 m .45212 .45614 L .41947 .46155 L .768 .725 .812 r
F
.40707 .44994 m .41947 .46155 L .41855 .4617 L F
.34852 .40169 m .41106 .47676 L .41947 .46155 L .84 .718 .729 r
F
.36972 .41495 m .34852 .40169 L p .41947 .46155 L F P
0 g
s
.34463 .39508 m .34015 .39176 L p .34866 .39812 L 0 .148 .597 r
F P
0 g
s
.35059 .40516 m .34852 .40169 L .35351 .41038 L p 0 0 .433 r
F P
0 g
s
.34015 .39176 m .34463 .39508 L p .34939 .40311 L .845 .741 .75 r
F P
0 g
s
.35282 .4089 m .35059 .40516 L p .35669 .41543 L 0 0 .433 r
F P
0 g
s
.35351 .41038 m .35891 .4198 L p .35059 .40516 L 0 0 .433 r
F P
0 g
s
.35669 .41543 m .34735 .40206 L p .34939 .40311 L .845 .741 .75 r
F P
0 g
s
.34735 .40206 m .34015 .39176 L p .34939 .40311 L .845 .741 .75 r
F P
0 g
s
.36972 .41495 m .38092 .42195 L p .34863 .39988 L .898 .611 .494 r
F P
0 g
s
.35036 .39939 m .3609 .40733 L .36092 .40729 L 0 .148 .597 r
F
.34979 .39897 m .3609 .40733 L .35036 .39939 L F
.3609 .40733 m .37113 .41502 L .36092 .40729 L F
.35731 .40465 m .37113 .41502 L p .34866 .39812 L F P
0 g
s
.34867 .39913 m .38092 .42195 L .36277 .40856 L .898 .611 .494 r
F
.35996 .40645 m .36199 .40798 L .36097 .4072 L 0 .148 .597 r
F
.35466 .40252 m .34463 .39508 L p .36069 .4071 L F P
0 g
s
.35742 .40461 m .34867 .39913 L p .36277 .40856 L .898 .611 .494 r
F P
0 g
s
.36097 .4072 m .36283 .41003 L p .36172 .40946 L .845 .741 .75 r
F P
0 g
s
.36737 .41236 m .36542 .41136 L .38092 .42195 L .898 .611 .494 r
F
.35411 .4084 m .407 .4827 L .41106 .47676 L .84 .718 .729 r
F
.35352 .40863 m .407 .4827 L .35411 .4084 L F
.35712 .40439 m .34867 .39913 L .35688 .40426 L .56 0 0 r
F
.35325 .40159 m .34867 .39913 L p .35712 .40439 L F P
0 g
s
.34939 .40311 m .34463 .39508 L .35811 .40685 L .845 .741 .75 r
F
.35728 .40717 m .35811 .40685 L .3602 .40867 L F
.34939 .40311 m .35811 .40685 L .35728 .40717 L F
.35282 .4089 m .34939 .40311 L .35728 .40717 L F
.35669 .41543 m .3701 .43462 L p .36294 .42615 L .238 .477 .885 r
F P
0 g
s
.3701 .43462 m .38057 .45542 L p .3795 .45381 L 0 0 .495 r
F P
0 g
s
.35352 .40863 m .38506 .46289 L .407 .4827 L .84 .718 .729 r
F
.35282 .4089 m .38506 .46289 L p .35352 .40863 L F P
0 g
s
.4506 .55543 m .41949 .46445 L .46959 .51035 L .818 .684 .716 r
F
.41949 .46445 m .42026 .46333 L .46959 .51035 L F
.4157 .51393 m .46907 .60225 L .4164 .51011 L .872 .683 .644 r
F
.46332 .61712 m .46719 .59593 L .49578 .63689 L .839 .649 .645 r
F
.41537 .51572 m .39994 .48807 L .41235 .50835 L .892 .705 .641 r
F
.39994 .48807 m .39681 .48246 L .39933 .48666 L F
.39994 .48807 m .39933 .48666 L .41235 .50835 L F
.41434 .52137 m .41689 .52747 L p .41506 .51746 L .872 .683 .644 r
F P
0 g
s
.41689 .52747 m .42677 .54676 L .41506 .51746 L .872 .683 .644 r
F
.41506 .51746 m .44305 .58751 L .42685 .54089 L F
.4164 .51011 m .46677 .59824 L .45819 .56557 L F
.46719 .59593 m .45819 .56557 L p .46677 .59824 L F P
0 g
s
.45527 .56167 m .46959 .51035 L .48 .5426 L .818 .684 .716 r
F
.4506 .55543 m .46959 .51035 L .45527 .56167 L F
.43645 .53655 m .40689 .48285 L .41928 .46476 L F
.38506 .46289 m .3857 .46361 L .38749 .46694 L .892 .705 .641 r
F
.38506 .46289 m .38082 .45579 L p .38057 .45542 L 0 0 .433 r
F P
0 g
s
.38057 .45542 m .38749 .46694 L .40437 .49833 L .892 .705 .641 r
F
.38082 .45579 m .35546 .41332 L p .36636 .43176 L 0 0 .433 r
F P
0 g
s
.37863 .45249 m .38057 .45542 L .36636 .43176 L 0 0 .433 r
F
.38057 .45542 m .38082 .45579 L .36636 .43176 L F
.51594 .78273 m .51895 .77813 L .52492 .71753 L .807 .607 .632 r
F
.52403 .71605 m .52492 .71753 L .52958 .67028 L F
.51594 .78273 m .52492 .71753 L .52403 .71605 L F
.53668 .59108 m .52958 .67028 L p .55216 .58317 L .738 .593 .694 r
F P
0 g
s
.52372 .71554 m .52318 .6258 L .55705 .64856 L .77 .599 .667 r
F
.52318 .6258 m .52301 .59806 L .53452 .61515 L F
.52318 .6258 m .53452 .61515 L .55705 .64856 L F
.47823 .68355 m .46295 .63245 L .471 .65517 L .907 .678 .581 r
F
.4782 .68392 m .46295 .63245 L .47823 .68355 L F
.4668 .6417 m .471 .65517 L p .46295 .63245 L F P
0 g
s
.51501 .70125 m .52301 .59806 L .52372 .71554 L .77 .599 .667 r
F
.49517 .5455 m .51852 .63017 L .53668 .59108 L .796 .645 .694 r
F
.53452 .61515 m .53668 .59108 L p .51852 .63017 L F P
0 g
s
.55216 .58317 m .54695 .63359 L .57351 .65069 L .738 .593 .694 r
F
.54695 .63359 m .56323 .65773 L .57351 .65069 L F
.52372 .71554 m .55705 .64856 L .55266 .69577 L .77 .599 .667 r
F
.56323 .65773 m .55266 .69577 L p .55705 .64856 L F P
0 g
s
.59952 .55898 m .57682 .60883 L p .57293 .57256 L .758 .645 .736 r
F P
0 g
s
.55216 .58317 m .57351 .65069 L .60942 .55392 L .738 .593 .694 r
F
.57351 .65069 m .58952 .63972 L .60942 .55392 L F
.62756 .59979 m .65293 .55843 L p .60604 .56847 L F P
0 g
s
.59952 .55898 m .60386 .54945 L p .61829 .5863 L .709 .586 .715 r
F P
0 g
s
.60386 .54945 m .61774 .51897 L p .61795 .5447 L .709 .586 .715 r
F P
0 g
s
.58822 .53688 m .61774 .51897 L .59692 .54387 L .758 .645 .736 r
F
.60942 .55392 m .65293 .55843 L p .62305 .54299 L .726 .65 .774 r
F P
0 g
s
.61774 .51897 m .69191 .5195 L p .61795 .5447 L .709 .586 .715 r
F P
0 g
s
.69191 .5195 m .69273 .51951 L p .69206 .51974 L .709 .586 .715 r
F P
0 g
s
.61795 .5447 m .69191 .5195 L .69206 .51974 L .709 .586 .715 r
F
.69206 .51974 m .65293 .55843 L p .69599 .52589 L .683 .578 .731 r
F P
0 g
s
.60386 .54945 m .60942 .55392 L .62305 .54299 L .726 .65 .774 r
F
.61774 .51897 m .59952 .55898 L p .57293 .57256 L .758 .645 .736 r
F P
0 g
s
.48639 .53587 m .49517 .5455 L .49113 .53087 L .796 .645 .694 r
F
.47047 .45918 m .4784 .48944 L p .46606 .45805 L .768 .725 .812 r
F P
0 g
s
.45212 .45614 m .45606 .45548 L .46606 .45805 L .768 .725 .812 r
F
.4784 .48944 m .47047 .45918 L p .48902 .46647 L .726 .7 .825 r
F P
0 g
s
.4784 .48944 m .48902 .46647 L .48978 .46697 L .726 .7 .825 r
F
.5467 .50417 m .5484 .49225 L p .50137 .47455 L .768 .69 .776 r
F P
0 g
s
.48922 .46654 m .48902 .46647 L p .5484 .49225 L .768 .69 .776 r
F P
0 g
s
.4974 .47195 m .5484 .49225 L .48902 .46647 L .768 .69 .776 r
F
.4784 .48944 m .48978 .46697 L .52545 .49028 L .726 .7 .825 r
F
.45238 .45609 m .50137 .47455 L .45606 .45548 L .768 .69 .776 r
F
.50137 .47455 m .5484 .49225 L .4974 .47195 L F
.50137 .47455 m .4974 .47195 L .45606 .45548 L F
.54921 .48653 m .54679 .50357 L p .54767 .50403 L .726 .65 .774 r
F P
0 g
s
.4784 .48944 m .52545 .49028 L .52645 .49093 L .726 .7 .825 r
F
.54921 .48653 m .54767 .50403 L .56959 .50066 L .726 .65 .774 r
F
.58822 .53688 m .56849 .52101 L .61774 .51897 L .758 .645 .736 r
F
.54065 .54679 m .5467 .50417 L p .54602 .50383 L .768 .69 .776 r
F P
0 g
s
.51045 .51576 m .54065 .54679 L .54602 .50383 L .768 .69 .776 r
F
.54612 .50379 m .4784 .48944 L p .52645 .49093 L .726 .7 .825 r
F P
0 g
s
.45238 .45609 m .5467 .50417 L .50137 .47455 L .768 .69 .776 r
F
.4334 .45924 m .54065 .54679 L .45238 .45609 L F
.54612 .50379 m .61774 .51897 L p .48443 .52449 L .758 .645 .736 r
F P
0 g
s
.54767 .50403 m .60386 .54945 L .62305 .54299 L .726 .65 .774 r
F
.54679 .50357 m .54741 .50407 L .54767 .50403 L F
.54741 .50407 m .60386 .54945 L .54767 .50403 L F
.54741 .50407 m .54612 .50379 L p .61504 .49369 L .726 .7 .825 r
F P
0 g
s
.5484 .49225 m .54921 .48653 L .48922 .46654 L p .768 .69 .776 r
F P
0 g
s
.52645 .49093 m .61504 .49369 L .54612 .50379 L .726 .7 .825 r
F
.45238 .45609 m .51045 .51576 L .54602 .50383 L .768 .69 .776 r
F
.52321 .57629 m .48443 .52449 L .56849 .52101 L .758 .645 .736 r
F
.52321 .57629 m .56849 .52101 L .58822 .53688 L F
.56959 .50066 m .54767 .50403 L .65293 .55843 L .726 .65 .774 r
F
.63637 .49475 m .64473 .49588 L .6423 .49548 L .734 .466 .523 r
F
.54919 .48402 m .63517 .49469 L .63637 .49475 L F
.63517 .49469 m .64473 .49588 L .63637 .49475 L F
.52645 .49093 m .57504 .4915 L .61504 .49369 L .726 .7 .825 r
F
.54657 .4847 m .54919 .48402 L .57404 .48727 L .047 .605 .81 r
F
.54919 .48402 m .54921 .48653 L .61363 .49345 L p .734 .466 .523 r
F P
0 g
s
.61363 .49345 m .61538 .49364 L p .63157 .49424 L .734 .466 .523 r
F P
0 g
s
.61504 .49369 m .61538 .49364 L .61823 .49379 L .726 .7 .825 r
F
.61538 .49364 m .69909 .50262 L p .63517 .49469 L .734 .466 .523 r
F P
0 g
s
.63157 .49424 m .61538 .49364 L .63517 .49469 L .734 .466 .523 r
F
.54919 .48402 m .5795 .48778 L p .57404 .48727 L .047 .605 .81 r
F P
0 g
s
.54919 .48402 m .61363 .49345 L .63157 .49424 L .734 .466 .523 r
F
.61356 .49094 m .54919 .48402 L p .63489 .49457 L F P
0 g
s
.62554 .49208 m .63605 .49434 L p .61823 .49379 L .726 .7 .825 r
F P
0 g
s
.61538 .49364 m .62554 .49208 L .61823 .49379 L .726 .7 .825 r
F
.57504 .4915 m .62554 .49208 L .61538 .49364 L F
.57504 .4915 m .61538 .49364 L .61504 .49369 L F
.52545 .49028 m .62554 .49208 L .52645 .49093 L F
.52545 .49028 m .48987 .4668 L .54698 .49067 L F
.48978 .46697 m .48987 .4668 L .52545 .49028 L F
.48902 .46647 m .45606 .45548 L p .4974 .47195 L .768 .69 .776 r
F P
0 g
s
.48978 .46697 m .48902 .46647 L .48987 .4668 L .726 .7 .825 r
F
.69909 .50262 m .70835 .50362 L p .69959 .50254 L .734 .466 .523 r
F P
0 g
s
.70835 .50362 m .7082 .50111 L p .69959 .50254 L .734 .466 .523 r
F P
0 g
s
.76628 .49853 m .70835 .50362 L p .7119 .51387 L .683 .578 .731 r
F P
0 g
s
.782 .4954 m .78015 .49563 L p .78221 .49522 L .138 0 0 r
F P
0 g
s
.78015 .49563 m .782 .4954 L p .69545 .49796 L .389 .798 .98 r
F P
0 g
s
.61774 .51897 m .62389 .5107 L p .782 .4954 L .7 .66 .808 r
F P
0 g
s
.62852 .51905 m .61774 .51897 L p .78022 .49645 L .7 .66 .808 r
F P
0 g
s
.59092 .51545 m .65293 .55843 L .68751 .51662 L .726 .65 .774 r
F
.69273 .51951 m .69191 .5195 L p .69244 .51935 L .7 .66 .808 r
F P
0 g
s
.65293 .55843 m .69244 .51935 L p .68751 .51662 L .726 .65 .774 r
F P
0 g
s
.69273 .51951 m .74059 .51985 L p .61795 .5447 L .709 .586 .715 r
F P
0 g
s
.69244 .51935 m .69206 .51974 L p .7119 .51387 L .683 .578 .731 r
F P
0 g
s
.69206 .51974 m .69599 .52589 L .7119 .51387 L .683 .578 .731 r
F
.54838 .78544 m .56339 .74402 L .5484 .78555 L .703 .494 .602 r
F
.56339 .74402 m .54445 .79927 L p .54217 .80284 L F P
0 g
s
.54864 .78703 m .56339 .74402 L p .54923 .78632 L .625 .411 .571 r
F P
0 g
s
.57243 .74265 m .56339 .74402 L p .59096 .70968 L .633 .438 .601 r
F P
0 g
s
.56339 .74402 m .58806 .7048 L p .5822 .71661 L .633 .438 .601 r
F P
0 g
s
.58127 .67524 m .55082 .78894 L .59386 .68529 L .668 .47 .608 r
F
.59861 .67386 m .59519 .67413 L p .59386 .68529 L F P
0 g
s
.59519 .67413 m .58127 .67524 L p .59386 .68529 L .668 .47 .608 r
F P
0 g
s
.59116 .71503 m .56676 .76011 L .57877 .73701 L .284 .041 .358 r
F
.57243 .74265 m .59045 .71729 L .59226 .71149 L .391 .139 .408 r
F
.59096 .70968 m .59025 .71571 L .57243 .74265 L .633 .438 .601 r
F
.57243 .74265 m .58705 .72305 L p .59045 .71729 L .391 .139 .408 r
F P
0 g
s
.60843 .68342 m .61433 .67261 L p .59252 .6966 L .668 .47 .608 r
F P
0 g
s
.58635 .70339 m .59236 .69795 L .59252 .6966 L .668 .47 .608 r
F
.59236 .69795 m .60843 .68342 L .59252 .6966 L F
.58806 .7048 m .59236 .69795 L p .5822 .71661 L .633 .438 .601 r
F P
0 g
s
.59236 .69795 m .59205 .70058 L .5822 .71661 L .633 .438 .601 r
F
.60132 .69363 m .59334 .70801 L .59545 .70124 L .284 .041 .358 r
F
.59236 .69795 m .59021 .71605 L .60843 .68342 L .668 .47 .608 r
F
.55824 .77586 m .58426 .72857 L p .55966 .77315 L .909 .796 .724 r
F P
0 g
s
.58705 .72305 m .61597 .67441 L .61551 .67493 L .391 .139 .408 r
F
.61478 .6731 m .59043 .71735 L p .59116 .71503 L .284 .041 .358 r
F P
0 g
s
.58426 .72857 m .61478 .6731 L p .58475 .72761 L .909 .796 .724 r
F P
0 g
s
.55966 .77315 m .58426 .72857 L .58475 .72761 L .909 .796 .724 r
F
.59025 .71571 m .5901 .71699 L .57243 .74265 L .633 .438 .601 r
F
.58705 .72305 m .57243 .74265 L p .5901 .71699 L F P
0 g
s
.58948 .71821 m .58395 .72824 L .5847 .727 L .909 .796 .724 r
F
.61433 .67261 m .5901 .71699 L p .58948 .71821 L F P
0 g
s
.5901 .71699 m .58395 .72824 L p .58948 .71821 L .909 .796 .724 r
F P
0 g
s
.58705 .72305 m .58475 .72761 L .58995 .71816 L .909 .796 .724 r
F
.58705 .72305 m .66548 .61783 L p .61597 .67441 L .391 .139 .408 r
F P
0 g
s
.61192 .67703 m .58948 .71821 L .61478 .6731 L .909 .796 .724 r
F
.58948 .71821 m .58705 .72305 L .61478 .6731 L F
.58705 .72305 m .58995 .71816 L .61478 .6731 L F
.60947 .68665 m .58982 .71932 L .5901 .71699 L .633 .438 .601 r
F
.5901 .71699 m .60843 .68342 L p .59021 .71605 L .668 .47 .608 r
F P
0 g
s
.59233 .71256 m .5901 .71699 L .59025 .71571 L .633 .438 .601 r
F
.60947 .68665 m .5901 .71699 L .59233 .71256 L F
.58982 .71932 m .58705 .72305 L p .5901 .71699 L F P
0 g
s
.55806 .77568 m .5901 .71699 L p .55459 .77987 L .668 .47 .608 r
F P
0 g
s
.55459 .77987 m .5901 .71699 L .59021 .71605 L .668 .47 .608 r
F
.5822 .71661 m .59212 .70046 L .59701 .69504 L .633 .438 .601 r
F
.58635 .70339 m .55459 .77987 L .59021 .71605 L .668 .47 .608 r
F
.61838 .65658 m .59236 .69795 L p .61818 .65539 L .707 .533 .651 r
F P
0 g
s
.59236 .69795 m .58769 .70537 L p .59205 .69824 L .707 .533 .651 r
F P
0 g
s
.61818 .65539 m .59236 .69795 L .59205 .69824 L .707 .533 .651 r
F
.58635 .70339 m .59021 .71605 L .59236 .69795 L .668 .47 .608 r
F
.62672 .6571 m .61478 .6731 L .6255 .65834 L .153 0 .16 r
F
.61551 .67493 m .64854 .63612 L .62707 .65677 L .391 .139 .408 r
F
.66548 .61783 m .67513 .60488 L p .61551 .67493 L F P
0 g
s
.6544 .63049 m .61478 .6731 L p .62672 .6571 L .153 0 .16 r
F P
0 g
s
.6225 .66116 m .65336 .61545 L .64534 .62229 L p 0 0 .083 r
F P
0 g
s
.6225 .66116 m .64534 .62229 L .62913 .64204 L 0 0 .083 r
F
.62859 .64884 m .61883 .65929 L .62545 .66054 L .676 .511 .656 r
F
.61838 .65658 m .6481 .60931 L p .61967 .6556 L .633 .438 .601 r
F P
0 g
s
.56173 .7186 m .61813 .65508 L .56323 .72236 L .707 .533 .651 r
F
.61813 .65508 m .62222 .65047 L .61818 .65539 L F
.61813 .65508 m .61818 .65539 L .56323 .72236 L F
.62121 .65444 m .6481 .60931 L .62884 .64866 L .633 .438 .601 r
F
.65293 .55843 m .6415 .59225 L .64423 .59077 L .676 .511 .656 r
F
.65293 .55843 m .62756 .59979 L p .6415 .59225 L F P
0 g
s
.71774 .52633 m .71904 .53623 L .71955 .53609 L .504 .467 .749 r
F
.61829 .5863 m .74059 .51985 L .67752 .57274 L .709 .586 .715 r
F
.62756 .59979 m .67752 .57274 L .63404 .60922 L F
.61829 .5863 m .67752 .57274 L .62756 .59979 L F
.64423 .59077 m .62022 .65518 L .62859 .64884 L .676 .511 .656 r
F
.6415 .59225 m .62022 .65518 L .64423 .59077 L F
.62022 .65518 m .61883 .65929 L .62859 .64884 L F
.59236 .69795 m .61838 .65658 L p .59205 .70058 L .633 .438 .601 r
F P
0 g
s
.6415 .59225 m .61397 .63018 L .61838 .65658 L .676 .511 .656 r
F
.62022 .65518 m .61838 .65658 L .61883 .65929 L F
.6415 .59225 m .61838 .65658 L .62022 .65518 L F
.61838 .65658 m .61967 .6556 L .59205 .70058 L .633 .438 .601 r
F
.59096 .70968 m .60947 .68665 L .59025 .71571 L F
.61829 .5863 m .61795 .5447 L .74059 .51985 L .709 .586 .715 r
F
.60386 .54945 m .61795 .5447 L .61829 .5863 L F
.61289 .6237 m .58127 .67524 L p .61397 .63018 L .676 .511 .656 r
F P
0 g
s
.56107 .71694 m .59957 .63283 L .61692 .64787 L .707 .533 .651 r
F
.59957 .63283 m .61045 .60906 L .61289 .6237 L F
.59957 .63283 m .61289 .6237 L .61692 .64787 L F
.56107 .71694 m .57041 .65281 L .59957 .63283 L F
.60604 .56847 m .58952 .63972 L .62756 .59979 L .738 .593 .694 r
F
.61289 .6237 m .62756 .59979 L p .58952 .63972 L F P
0 g
s
.62756 .59979 m .61289 .6237 L p .6415 .59225 L .676 .511 .656 r
F P
0 g
s
.61289 .6237 m .61397 .63018 L .6415 .59225 L .676 .511 .656 r
F
.53929 .74387 m .55266 .69577 L p .56107 .71694 L .707 .533 .651 r
F P
0 g
s
.55266 .69577 m .53982 .74198 L p .52684 .72068 L .77 .599 .667 r
F P
0 g
s
.52372 .71554 m .55266 .69577 L .52684 .72068 L .77 .599 .667 r
F
.53258 .7875 m .5226 .79877 L .53258 .78769 L .74 .548 .635 r
F
.52523 .77241 m .5226 .79877 L .53258 .7875 L F
.53191 .78614 m .52416 .77019 L .53853 .74827 L .774 .561 .611 r
F
.52684 .72068 m .5226 .79877 L p .53077 .71689 L .74 .548 .635 r
F P
0 g
s
.5226 .79877 m .5221 .80803 L p .53258 .78769 L .74 .548 .635 r
F P
0 g
s
.52255 .79974 m .52684 .72068 L p .51776 .79021 L .807 .607 .632 r
F P
0 g
s
.51776 .79021 m .52684 .72068 L .52492 .71753 L .807 .607 .632 r
F
.5223 .84117 m .5164 .78203 L .52416 .77019 L .774 .561 .611 r
F
.5223 .84117 m .52416 .77019 L .53191 .78614 L F
.53077 .71689 m .52523 .77241 L .53258 .7875 L .74 .548 .635 r
F
.54627 .77249 m .5638 .72379 L p .53262 .7876 L F P
0 g
s
.53258 .78769 m .53264 .78763 L .53262 .7876 L .74 .548 .635 r
F
.53264 .78763 m .54627 .77249 L .53262 .7876 L F
.53258 .78769 m .5638 .72379 L .5621 .71952 L F
.53258 .7875 m .53258 .78769 L .53262 .78759 L F
.53264 .78763 m .52513 .837 L p .53191 .78614 L .774 .561 .611 r
F P
0 g
s
.52513 .837 m .52181 .85879 L p .52477 .83753 L .774 .561 .611 r
F P
0 g
s
.53191 .78614 m .52513 .837 L .52477 .83753 L .774 .561 .611 r
F
.52197 .85879 m .52181 .85879 L p .52199 .85828 L .703 .494 .602 r
F P
0 g
s
.52045 .87272 m .52667 .84144 L .52121 .84277 L .738 .512 .586 r
F
.52667 .84144 m .52513 .837 L .52121 .84277 L F
.52559 .83833 m .53264 .78763 L .53555 .78439 L .703 .494 .602 r
F
.52513 .837 m .53264 .78763 L p .52559 .83833 L F P
0 g
s
.5638 .72379 m .56564 .72213 L .56398 .72424 L .707 .533 .651 r
F
.5638 .72379 m .53312 .809 L p .56725 .72067 L .668 .47 .608 r
F P
0 g
s
.52656 .84568 m .52713 .84355 L .52755 .84295 L .703 .494 .602 r
F
.52404 .8588 m .52487 .85868 L p .5233 .86357 L .62 .394 .553 r
F P
0 g
s
.53508 .82217 m .53482 .82319 L .52644 .84641 L .703 .494 .602 r
F
.53508 .82217 m .52644 .84641 L .52647 .84621 L F
.52467 .85731 m .52864 .8428 L .52556 .85162 L .144 0 .315 r
F
.52404 .8588 m .53086 .83892 L p .52881 .84734 L .625 .411 .571 r
F P
0 g
s
.52656 .84568 m .53508 .82217 L .52647 .84621 L .703 .494 .602 r
F
.52644 .84641 m .53482 .82319 L .52479 .8566 L F
.53086 .83892 m .52479 .8566 L p .53482 .82319 L F P
0 g
s
.52443 .86053 m .52735 .84324 L .52435 .86076 L .661 .439 .572 r
F
.53312 .809 m .52731 .84329 L p .52735 .84324 L F P
0 g
s
.52731 .84329 m .52435 .86076 L p .52735 .84324 L .661 .439 .572 r
F P
0 g
s
.52755 .84295 m .54627 .77249 L .54838 .78544 L .703 .494 .602 r
F
.52713 .84355 m .54627 .77249 L .52755 .84295 L F
.52725 .84311 m .53668 .78314 L .54627 .77249 L F
.52656 .84568 m .52706 .84257 L .52725 .84311 L F
.52706 .84257 m .53668 .78314 L .52725 .84311 L F
.52479 .8566 m .52404 .8588 L p .52644 .84641 L F P
0 g
s
.52435 .86076 m .52731 .84329 L p .52045 .87272 L .738 .512 .586 r
F P
0 g
s
.52731 .84329 m .52683 .8419 L .52045 .87272 L .738 .512 .586 r
F
.52779 .85157 m .52487 .85868 L p .52881 .84734 L .625 .411 .571 r
F P
0 g
s
.53148 .84046 m .53137 .83756 L .52443 .86053 L .661 .439 .572 r
F
.54939 .79156 m .55082 .78894 L p .54816 .78407 L .668 .47 .608 r
F P
0 g
s
.5335 .83451 m .54024 .80858 L p .5334 .8346 L .951 .759 .609 r
F P
0 g
s
.53325 .83534 m .5334 .8346 L .533 .83613 L .951 .759 .609 r
F
.5335 .83451 m .5334 .8346 L .53325 .83534 L F
.53482 .82319 m .53727 .8164 L .53128 .83755 L .703 .494 .602 r
F
.54024 .80858 m .54953 .7917 L p .55005 .79058 L .909 .796 .724 r
F P
0 g
s
.53727 .8164 m .54217 .80284 L .53592 .82277 L .703 .494 .602 r
F
.53727 .8164 m .53592 .82277 L .53128 .83755 L F
.54864 .78703 m .54923 .78632 L .53144 .83949 L .625 .411 .571 r
F
.54864 .78703 m .53144 .83949 L .53144 .83936 L F
.53592 .82277 m .53136 .83744 L .53128 .83755 L .703 .494 .602 r
F
.54217 .80284 m .53586 .82302 L .53592 .82277 L F
.53586 .82302 m .53136 .83744 L .53592 .82277 L F
.54445 .79927 m .53514 .82641 L p .54217 .80284 L F P
0 g
s
.53514 .82641 m .53136 .83744 L p .53586 .82302 L .703 .494 .602 r
F P
0 g
s
.54217 .80284 m .53514 .82641 L .53586 .82302 L .703 .494 .602 r
F
.53136 .83744 m .53086 .83892 L p .53128 .83755 L F P
0 g
s
.52443 .86053 m .53144 .83733 L .52735 .84324 L .661 .439 .572 r
F
.53136 .83744 m .54864 .78703 L p .53144 .83936 L .625 .411 .571 r
F P
0 g
s
.53482 .82319 m .53128 .83755 L .53115 .83775 L .703 .494 .602 r
F
.53508 .82217 m .54838 .78544 L .53482 .82319 L F
.54838 .78544 m .5484 .78555 L .53482 .82319 L F
.5697 .71845 m .54816 .78407 L .54933 .78621 L .668 .47 .608 r
F
.55261 .78225 m .54933 .78621 L .55082 .78894 L F
.5697 .71845 m .54933 .78621 L .55261 .78225 L F
.53258 .7875 m .53262 .78759 L .5621 .71952 L .74 .548 .635 r
F
.53258 .7875 m .5621 .71952 L .56208 .71948 L F
.54953 .7917 m .55824 .77586 L p .55779 .77654 L .909 .796 .724 r
F P
0 g
s
.55005 .79058 m .54953 .7917 L .55779 .77654 L .909 .796 .724 r
F
.55628 .77965 m .55173 .78838 L .54942 .79348 L .424 .201 .468 r
F
.55628 .77965 m .54942 .79348 L .54374 .80604 L F
.56095 .77066 m .55628 .77965 L .54374 .80604 L F
.54217 .80744 m .54553 .7974 L .55866 .77571 L .625 .411 .571 r
F
.54553 .7974 m .54923 .78632 L .55689 .7777 L F
.54553 .7974 m .55689 .7777 L .55866 .77571 L F
.55806 .77568 m .55887 .77426 L .5588 .77479 L F
.58395 .72824 m .55806 .77568 L p .5847 .727 L .909 .796 .724 r
F P
0 g
s
.55806 .77568 m .55831 .77524 L .5847 .727 L .909 .796 .724 r
F
.54923 .78632 m .5588 .77479 L .55866 .77571 L .625 .411 .571 r
F
.59096 .70968 m .61265 .67224 L .59852 .70027 L .633 .438 .601 r
F
.5918 .70262 m .61265 .67224 L .59096 .70968 L F
.59386 .68529 m .58635 .70339 L .59252 .6966 L .668 .47 .608 r
F
.56323 .72236 m .61818 .65539 L .56564 .72213 L .707 .533 .651 r
F
.56323 .72236 m .56564 .72213 L .5638 .72379 L F
.56564 .72213 m .61818 .65539 L .59205 .69824 L F
.53929 .74387 m .61692 .64787 L .61813 .65508 L F
.53077 .71689 m .53258 .7875 L .55992 .71404 L .74 .548 .635 r
F
.53077 .71689 m .55992 .71404 L .55266 .69577 L F
.59212 .70046 m .61967 .6556 L .59701 .69504 L .633 .438 .601 r
F
.61967 .6556 m .62121 .65444 L .59701 .69504 L F
.55992 .71404 m .53258 .7875 L .56208 .71948 L .74 .548 .635 r
F
.52012 .88557 m .52121 .84277 L .52003 .88529 L .738 .512 .586 r
F
.52117 .84284 m .52003 .88529 L p .52121 .84277 L F P
0 g
s
.52003 .88529 m .52002 .88576 L p .52012 .88557 L .738 .512 .586 r
F P
0 g
s
.52017 .88571 m .52018 .88577 L p .52017 .88571 L .95 .6 .297 r
F P
0 g
s
.52012 .88577 m .5201 .88569 L .52015 .88574 L .649 .393 .518 r
F
.52018 .88577 m .52016 .88577 L p .52015 .88574 L F P
0 g
s
.52016 .88577 m .52012 .88577 L p .52015 .88574 L .649 .393 .518 r
F P
0 g
s
.53357 .83314 m .54208 .80769 L .53207 .83953 L .625 .411 .571 r
F
.53083 .84454 m .53137 .84253 L .53153 .84183 L .951 .759 .609 r
F
.53336 .83643 m .54217 .80744 L .55866 .77571 L .625 .411 .571 r
F
.54208 .80769 m .54217 .80744 L .53336 .83643 L F
.5847 .727 m .55831 .77524 L .56014 .77206 L .909 .796 .724 r
F
.55866 .77571 m .55887 .77426 L .56027 .77184 L .625 .411 .571 r
F
.64957 .63671 m .6564 .62821 L .66019 .62492 L .633 .438 .601 r
F
.67513 .60488 m .66548 .61783 L p .66133 .62207 L F P
0 g
s
.61478 .6731 m .6544 .63049 L p .66548 .61783 L .846 .805 .811 r
F P
0 g
s
.67011 .61254 m .66242 .62096 L .66548 .61783 L .846 .805 .811 r
F
.66242 .62096 m .61478 .6731 L .66548 .61783 L F
.61433 .67261 m .62002 .66643 L p .61883 .65929 L .676 .511 .656 r
F P
0 g
s
.62002 .66643 m .62545 .66054 L p .61883 .65929 L .676 .511 .656 r
F P
0 g
s
.61433 .67261 m .61478 .6731 L p .62104 .66624 L .846 .805 .811 r
F P
0 g
s
.62545 .66054 m .61433 .67261 L p .62104 .66624 L .846 .805 .811 r
F P
0 g
s
.59852 .70027 m .62348 .66072 L .60947 .68665 L .633 .438 .601 r
F
.67513 .60488 m .65336 .61545 L p .62707 .65677 L .391 .139 .408 r
F P
0 g
s
.64854 .63612 m .67513 .60488 L .62707 .65677 L .391 .139 .408 r
F
.67543 .58561 m .6463 .62112 L .64709 .62014 L 0 0 0 r
F
.71779 .55868 m .72269 .55557 L p .65447 .61403 L .153 0 .16 r
F P
0 g
s
.65336 .61545 m .65827 .61071 L .64614 .62131 L 0 0 0 r
F
.64534 .62229 m .65336 .61545 L p .64614 .62131 L F P
0 g
s
.67543 .58561 m .64614 .62131 L p .6463 .62112 L 0 0 0 r
F P
0 g
s
.64614 .62131 m .67543 .58561 L p .65447 .61403 L .153 0 .16 r
F P
0 g
s
.65006 .61652 m .65626 .60897 L p .64953 .61717 L 0 0 0 r
F P
0 g
s
.63315 .64539 m .65682 .60788 L .6352 .64384 L .633 .438 .601 r
F
.70213 .57732 m .67436 .60503 L .71547 .56285 L .153 0 .16 r
F
.67436 .60503 m .67952 .59907 L .71547 .56285 L F
.65336 .61545 m .67513 .60488 L p .69589 .58412 L .299 0 .256 r
F P
0 g
s
.70355 .57113 m .65336 .61545 L p .69589 .58412 L .299 0 .256 r
F P
0 g
s
.67543 .58561 m .71779 .55868 L p .65447 .61403 L .153 0 .16 r
F P
0 g
s
.65336 .61545 m .70355 .57113 L p .65827 .61071 L 0 0 0 r
F P
0 g
s
.67513 .60488 m .6481 .60931 L p .69409 .57718 L .644 .482 .649 r
F P
0 g
s
.71167 .5678 m .69389 .58751 L .69671 .58273 L .153 0 .16 r
F
.72269 .55557 m .69815 .5803 L .70255 .57283 L F
.69228 .58798 m .7208 .55702 L p .69453 .5746 L .676 .511 .656 r
F P
0 g
s
.69409 .57718 m .71115 .56527 L .69269 .58557 L .644 .482 .649 r
F
.69409 .57718 m .69269 .58557 L .67513 .60488 L F
.68508 .59665 m .67513 .60488 L p .68626 .59524 L F P
0 g
s
.69269 .58557 m .68626 .59524 L .67513 .60488 L .644 .482 .649 r
F
.71115 .56527 m .69228 .58798 L .69269 .58557 L F
.69228 .58798 m .68626 .59524 L .69269 .58557 L F
.67543 .58561 m .72025 .55557 L p .71895 .55722 L 0 0 0 r
F P
0 g
s
.6481 .60931 m .69656 .56243 L p .69466 .57379 L .644 .482 .649 r
F P
0 g
s
.69656 .56243 m .69702 .56219 L .70286 .55417 L .676 .511 .656 r
F
.72518 .55226 m .70286 .55417 L p .69648 .56293 L F P
0 g
s
.6481 .60931 m .69466 .57379 L .69409 .57718 L .644 .482 .649 r
F
.69648 .56293 m .68179 .58312 L .69463 .57399 L .676 .511 .656 r
F
.69648 .56293 m .69463 .57399 L .72518 .55226 L F
.67934 .60367 m .67849 .60459 L p .71167 .5678 L .153 0 .16 r
F P
0 g
s
.68508 .59665 m .69306 .58892 L .69589 .58412 L .299 0 .256 r
F
.70213 .57732 m .71547 .56285 L .72269 .55557 L .153 0 .16 r
F
.71167 .5678 m .70213 .57732 L .72269 .55557 L F
.72269 .55557 m .72608 .55341 L p .71167 .5678 L F P
0 g
s
.72089 .55696 m .68963 .59117 L .72608 .55341 L .846 .805 .811 r
F
.68963 .59117 m .68111 .6005 L .68706 .59427 L F
.68963 .59117 m .68706 .59427 L .72608 .55341 L F
.69306 .58892 m .72608 .55341 L p .68111 .6005 L F P
0 g
s
.72089 .55696 m .69206 .58825 L .68963 .59117 L .846 .805 .811 r
F
.72518 .55226 m .7208 .55702 L p .72089 .55696 L F P
0 g
s
.65682 .60788 m .67513 .60488 L p .66133 .62207 L .633 .438 .601 r
F P
0 g
s
.65378 .62978 m .65682 .60788 L .66133 .62207 L .633 .438 .601 r
F
.62545 .66054 m .65378 .62978 L p .64054 .6398 L .676 .511 .656 r
F P
0 g
s
.62859 .64884 m .62545 .66054 L .64054 .6398 L .676 .511 .656 r
F
.62755 .65656 m .6352 .64384 L .65378 .62978 L .633 .438 .601 r
F
.6352 .64384 m .65682 .60788 L .65378 .62978 L F
.68179 .58312 m .64054 .6398 L .69228 .58798 L .676 .511 .656 r
F
.65378 .62978 m .69228 .58798 L p .64054 .6398 L F P
0 g
s
.7208 .55702 m .65378 .62978 L p .72089 .55696 L .846 .805 .811 r
F P
0 g
s
.65378 .62978 m .62545 .66054 L p .65396 .6296 L .846 .805 .811 r
F P
0 g
s
.65378 .62978 m .65396 .6296 L .72089 .55696 L .846 .805 .811 r
F
.68111 .6005 m .62545 .66054 L .62104 .66624 L F
.67795 .60356 m .62545 .66054 L .68111 .6005 L F
.60947 .68665 m .62755 .65656 L .66133 .62207 L .633 .438 .601 r
F
.72608 .55341 m .72518 .55226 L p .72089 .55696 L .846 .805 .811 r
F P
0 g
s
.7208 .55702 m .72518 .55226 L p .68179 .58312 L .676 .511 .656 r
F P
0 g
s
.71115 .56527 m .75192 .5301 L .73077 .55156 L .644 .482 .649 r
F
.69409 .57718 m .7162 .55918 L .71115 .56527 L F
.74059 .51985 m .71066 .5488 L p .70905 .5463 L .709 .586 .715 r
F P
0 g
s
.7119 .51387 m .69599 .52589 L .72518 .55226 L .683 .578 .731 r
F
.71351 .55326 m .72518 .55226 L p .69599 .52589 L F P
0 g
s
.7119 .51387 m .72518 .55226 L .76628 .49853 L .683 .578 .731 r
F
.7162 .55918 m .75192 .5301 L .71115 .56527 L .644 .482 .649 r
F
.69466 .57379 m .75192 .5301 L .69409 .57718 L F
.71066 .5488 m .69656 .56243 L p .70905 .5463 L .709 .586 .715 r
F P
0 g
s
.69656 .56243 m .6481 .60931 L p .69002 .56598 L .709 .586 .715 r
F P
0 g
s
.69708 .55937 m .69656 .56243 L .70286 .55417 L .676 .511 .656 r
F
.63849 .61208 m .69648 .56293 L .69708 .55937 L F
.70905 .5463 m .69656 .56243 L .69002 .56598 L .709 .586 .715 r
F
.69656 .56243 m .69648 .56293 L .69702 .56219 L .676 .511 .656 r
F
.63849 .61208 m .68179 .58312 L .69648 .56293 L F
.69466 .57379 m .71066 .5488 L .75192 .5301 L .644 .482 .649 r
F
.69656 .56243 m .71066 .5488 L p .69466 .57379 L F P
0 g
s
.68179 .58312 m .69228 .58798 L .69453 .5746 L .676 .511 .656 r
F
.6481 .60931 m .65682 .60788 L p .62884 .64866 L .633 .438 .601 r
F P
0 g
s
.61265 .67224 m .65682 .60788 L .59852 .70027 L .633 .438 .601 r
F
.63849 .61208 m .62859 .64884 L .68179 .58312 L .676 .511 .656 r
F
.62859 .64884 m .64054 .6398 L .68179 .58312 L F
.62348 .66072 m .63315 .64539 L .62755 .65656 L .633 .438 .601 r
F
.63315 .64539 m .6352 .64384 L .62755 .65656 L F
.62348 .66072 m .62755 .65656 L .60947 .68665 L F
.6564 .62821 m .66133 .62207 L .66019 .62492 L F
.66548 .61783 m .66019 .62492 L p .66133 .62207 L F P
0 g
s
.68111 .6005 m .67011 .61254 L .67849 .60459 L .846 .805 .811 r
F
.67849 .60459 m .69306 .58892 L p .68508 .59665 L F P
0 g
s
.68111 .6005 m .67849 .60459 L .68508 .59665 L .846 .805 .811 r
F
.68508 .59665 m .69176 .59113 L p .69306 .58892 L .299 0 .256 r
F P
0 g
s
.78208 .51643 m .68508 .59665 L p .76102 .53042 L .644 .482 .649 r
F P
0 g
s
.76102 .53042 m .68508 .59665 L .68626 .59524 L .644 .482 .649 r
F
.68963 .59117 m .67795 .60356 L .68111 .6005 L .846 .805 .811 r
F
.69206 .58825 m .67795 .60356 L .68963 .59117 L F
.71115 .56527 m .76102 .53042 L .68626 .59524 L .644 .482 .649 r
F
.7877 .51178 m .78208 .51643 L p .76102 .53042 L F P
0 g
s
.81244 .49626 m .81007 .49783 L p .81138 .49696 L 0 0 0 r
F P
0 g
s
.7956 .50698 m .81007 .49777 L p .79549 .5071 L 0 0 0 r
F P
0 g
s
.81007 .49777 m .81244 .49626 L p .81192 .49661 L 0 0 0 r
F P
0 g
s
.79434 .50824 m .81244 .49626 L p .78204 .51475 L .769 .792 .872 r
F P
0 g
s
.7877 .51178 m .78681 .51193 L p .81807 .49273 L .662 .568 .739 r
F P
0 g
s
.81807 .49273 m .78681 .51193 L .80425 .49757 L .662 .568 .739 r
F
.83825 .48567 m .81768 .49589 L p .81556 .49701 L 0 0 0 r
F P
0 g
s
.83825 .48567 m .79412 .50847 L p .81807 .49273 L .662 .568 .739 r
F P
0 g
s
.79412 .50847 m .7877 .51178 L p .81807 .49273 L .662 .568 .739 r
F P
0 g
s
.73077 .55156 m .73373 .54855 L .76102 .53042 L .644 .482 .649 r
F
.60947 .68665 m .64957 .63671 L .58982 .71932 L .633 .438 .601 r
F
.66019 .62492 m .58982 .71932 L p .64957 .63671 L F P
0 g
s
.59701 .69504 m .62121 .65444 L .62884 .64866 L .633 .438 .601 r
F
.67752 .57274 m .70905 .5463 L .69002 .56598 L .709 .586 .715 r
F
.64423 .59077 m .63849 .61208 L .6893 .56637 L .676 .511 .656 r
F
.78204 .51475 m .78115 .51537 L .78302 .51423 L .769 .792 .872 r
F
.77913 .51776 m .77701 .51788 L .7877 .51178 L .644 .482 .649 r
F
.78681 .51193 m .7877 .51178 L p .77701 .51788 L F P
0 g
s
.46641 .5569 m .49517 .5455 L .48639 .53587 L .796 .645 .694 r
F
.48639 .53587 m .48443 .52449 L p .52321 .57629 L .758 .645 .736 r
F P
0 g
s
.42026 .46333 m .41947 .46155 L p .41106 .47676 L .84 .718 .729 r
F P
0 g
s
.45202 .4936 m .48443 .52449 L .50491 .51762 L .768 .69 .776 r
F
.45606 .45548 m .40998 .44012 L p .45238 .45609 L F P
0 g
s
.46179 .45577 m .47047 .45918 L .4705 .45919 L .818 .536 .513 r
F
.46179 .45577 m .48922 .46654 L .47028 .45876 L F
.47028 .45876 m .47063 .45888 L .45612 .45295 L F
.47048 .45884 m .47033 .45864 L p .41511 .43611 L .265 .651 .966 r
F P
0 g
s
.36986 .41239 m .42015 .44248 L .43246 .44606 L .265 .651 .966 r
F
.41042 .43361 m .40821 .43268 L .43463 .4467 L F
.41042 .43361 m .43463 .4467 L .43529 .44689 L F
.41008 .43758 m .45612 .45295 L p .42754 .44463 L 0 .515 .828 r
F P
0 g
s
.41008 .43758 m .46592 .45801 L .46179 .45577 L .818 .536 .513 r
F
.46592 .45801 m .48902 .46647 L .47047 .45918 L F
.46592 .45801 m .47047 .45918 L .46179 .45577 L F
.40475 .43007 m .40922 .43311 L .40651 .43087 L .56 0 0 r
F
.41919 .43388 m .42026 .43422 L .40773 .43188 L .631 .067 0 r
F
.41919 .43388 m .40773 .43188 L .4076 .43177 L F
.40726 .43149 m .41919 .43388 L p .4076 .43177 L F P
0 g
s
.39495 .42432 m .38624 .42 L .40468 .42999 L .56 0 0 r
F
.4045 .42572 m .41835 .43016 L .39739 .42574 L .881 .453 .167 r
F
.4045 .42572 m .39739 .42574 L .39495 .42432 L F
.4191 .43349 m .40431 .42904 L .41848 .43365 L .863 .435 .195 r
F
.39986 .42718 m .41879 .43208 L .40726 .43149 L .881 .453 .167 r
F
.41919 .43388 m .40726 .43149 L p .41879 .43208 L F P
0 g
s
.38859 .42062 m .39069 .4211 L .39399 .42235 L .881 .453 .167 r
F
.39495 .42432 m .38269 .41738 L .38624 .42 L .56 0 0 r
F
.38709 .41975 m .38269 .41738 L p .39495 .42432 L F P
0 g
s
.47033 .45864 m .36097 .4072 L p .41511 .43611 L .265 .651 .966 r
F P
0 g
s
.42015 .44248 m .41008 .43758 L p .42754 .44463 L 0 .515 .828 r
F P
0 g
s
.41008 .43758 m .40998 .44012 L .45606 .45548 L p .818 .536 .513 r
F P
0 g
s
.41008 .43758 m .45606 .45548 L .46592 .45801 L .818 .536 .513 r
F
.45612 .45295 m .41008 .43758 L p .47028 .45876 L F P
0 g
s
.36097 .4072 m .46165 .45456 L p .44929 .45375 L .768 .725 .812 r
F P
0 g
s
.42152 .45193 m .4105 .44378 L .44929 .45375 L .768 .725 .812 r
F
.36986 .41239 m .41008 .43758 L p .40202 .43164 L .56 0 0 r
F P
0 g
s
.36305 .40832 m .36909 .41194 L .40821 .43268 L .265 .651 .966 r
F
.36097 .4072 m .36909 .41194 L .36305 .40832 L F
.4087 .43563 m .40202 .43164 L .41008 .43758 L .56 0 0 r
F
.4087 .43563 m .41008 .43758 L .40984 .43631 L F
.36288 .40822 m .40821 .43268 L .41042 .43361 L .265 .651 .966 r
F
.36097 .4072 m .36305 .40832 L .36288 .40822 L F
.36305 .40832 m .40821 .43268 L .36288 .40822 L F
.38104 .4194 m .36986 .41239 L p .37378 .41566 L .898 .611 .494 r
F P
0 g
s
.36539 .40999 m .37378 .41566 L .36986 .41239 L .898 .611 .494 r
F
.40998 .44012 m .41008 .43758 L .38104 .4194 L p F P
0 g
s
.37378 .41566 m .40998 .44012 L .38104 .4194 L .898 .611 .494 r
F
.4105 .44378 m .36097 .4072 L .44929 .45375 L .768 .725 .812 r
F
.47033 .45864 m .47047 .45918 L p .46606 .45805 L F P
0 g
s
.48902 .46647 m .48922 .46654 L p .4705 .45919 L .818 .536 .513 r
F P
0 g
s
.47047 .45918 m .48902 .46647 L .4705 .45919 L .818 .536 .513 r
F
.48922 .46654 m .54921 .48653 L p .47028 .45876 L F P
0 g
s
.46165 .45456 m .47033 .45864 L p .46606 .45805 L .768 .725 .812 r
F P
0 g
s
.45606 .45548 m .46165 .45456 L .46606 .45805 L .768 .725 .812 r
F
.45606 .45548 m .48902 .46647 L p .46592 .45801 L .818 .536 .513 r
F P
0 g
s
.42152 .45193 m .46165 .45456 L .45212 .45614 L .768 .725 .812 r
F
.42147 .45192 m .36097 .4072 L .42152 .45193 L F
.3629 .40865 m .36097 .4072 L p .36199 .40798 L 0 .148 .597 r
F P
0 g
s
.36283 .41003 m .36097 .4072 L p .42147 .45192 L .768 .725 .812 r
F P
0 g
s
.36172 .40946 m .3602 .40867 L .36097 .4072 L .845 .741 .75 r
F
.35967 .40624 m .35466 .40252 L p .35799 .40502 L 0 .148 .597 r
F P
0 g
s
.35688 .40426 m .35742 .40461 L p .35712 .40439 L .56 0 0 r
F P
0 g
s
.54698 .49067 m .48987 .4668 L .55079 .49074 L .726 .7 .825 r
F
.47033 .45864 m .47164 .45916 L .47275 .45947 L .419 .758 .995 r
F
.47033 .45864 m .47048 .45884 L p .47164 .45916 L F P
0 g
s
.47033 .45864 m .546 .48449 L .48831 .46367 L .419 .758 .995 r
F
.47063 .45888 m .54921 .48653 L .47876 .46112 L .818 .536 .513 r
F
.52841 .47815 m .54919 .48402 L .53556 .48073 L 0 .515 .828 r
F
.54919 .48402 m .48906 .46394 L p .53892 .48267 L .818 .536 .513 r
F P
0 g
s
.54921 .48653 m .54919 .48402 L p .53892 .48267 L .818 .536 .513 r
F P
0 g
s
.50006 .46696 m .51213 .47157 L .52451 .47381 L .419 .758 .995 r
F
.48831 .46367 m .48906 .46394 L .50006 .46696 L F
.47876 .46112 m .54921 .48653 L .48906 .46394 L .818 .536 .513 r
F
.47033 .45864 m .62554 .49208 L p .55079 .49074 L .726 .7 .825 r
F P
0 g
s
.47047 .45918 m .47033 .45864 L p .55079 .49074 L .726 .7 .825 r
F P
0 g
s
.6533 .49771 m .61538 .49364 L p .65688 .49969 L .726 .65 .774 r
F P
0 g
s
.61183 .49418 m .65688 .49969 L .61538 .49364 L .726 .65 .774 r
F
.65688 .49969 m .68988 .50372 L .6533 .49771 L F
.63637 .49475 m .69959 .50254 L .64543 .49523 L .734 .466 .523 r
F
.64473 .49588 m .69909 .50262 L .6423 .49548 L F
.7082 .50111 m .6532 .4952 L p .69959 .50254 L F P
0 g
s
.64124 .49467 m .69959 .50254 L .6532 .4952 L .734 .466 .523 r
F
.70344 .49539 m .76873 .49199 L .70738 .49386 L .27 0 0 r
F
.64764 .49446 m .69545 .49796 L .77134 .49572 L .389 .798 .98 r
F
.69909 .50262 m .66158 .4986 L p .68988 .50372 L .726 .65 .774 r
F P
0 g
s
.782 .4954 m .77748 .49537 L p .77134 .49572 L .389 .798 .98 r
F P
0 g
s
.77748 .49537 m .63605 .49434 L p .77134 .49572 L .389 .798 .98 r
F P
0 g
s
.69051 .50405 m .65253 .49728 L .77748 .49537 L .7 .66 .808 r
F
.63489 .49457 m .63637 .49475 L .64124 .49467 L .734 .466 .523 r
F
.63637 .49475 m .64543 .49523 L .64124 .49467 L F
.63489 .49457 m .64124 .49467 L .63345 .49433 L F
.63605 .49434 m .63576 .49443 L p .64178 .49469 L .389 .798 .98 r
F P
0 g
s
.63605 .49434 m .64737 .49443 L p .65253 .49728 L .7 .66 .808 r
F P
0 g
s
.6532 .4952 m .6645 .4957 L .64764 .49446 L .389 .798 .98 r
F
.64737 .49443 m .77748 .49537 L p .65253 .49728 L .7 .66 .808 r
F P
0 g
s
.66871 .50623 m .6533 .49771 L .69051 .50405 L .7 .66 .808 r
F
.77748 .49537 m .782 .4954 L p .62389 .5107 L F P
0 g
s
.61183 .49418 m .59092 .51545 L .66871 .50623 L .726 .65 .774 r
F
.61183 .49418 m .66871 .50623 L .65688 .49969 L F
.56959 .50066 m .59092 .51545 L .61183 .49418 L F
.63605 .49434 m .63577 .49472 L p .61823 .49379 L .726 .7 .825 r
F P
0 g
s
.63577 .49472 m .63605 .49434 L p .69051 .50405 L .7 .66 .808 r
F P
0 g
s
.34866 .39812 m .34979 .39897 L .35036 .39939 L 0 .148 .597 r
F
.34463 .39508 m .36097 .4072 L p .3602 .40867 L .845 .741 .75 r
F P
0 g
s
.35546 .41332 m .35282 .4089 L p .35669 .41543 L 0 0 .433 r
F P
0 g
s
.34955 .40521 m .35669 .41543 L p .35233 .41359 L .238 .477 .885 r
F P
0 g
s
.35891 .4198 m .35987 .42141 L p .35978 .42125 L 0 0 .433 r
F P
0 g
s
.35233 .41359 m .35669 .41543 L .36294 .42615 L .238 .477 .885 r
F
.37863 .45249 m .35546 .41332 L .35788 .41748 L 0 0 .433 r
F
.36882 .43624 m .37863 .45249 L .35788 .41748 L F
.43253 .5834 m .42004 .5378 L .41502 .52957 L 0 .043 .608 r
F
.41502 .52957 m .40778 .51557 L p .4232 .543 L 0 0 .495 r
F P
0 g
s
.40778 .51557 m .40298 .50536 L .4232 .543 L 0 0 .495 r
F
.36793 .46345 m .39592 .49029 L .36733 .46179 L .24 .431 .849 r
F
.36682 .46037 m .36351 .44179 L .36657 .45944 L F
.36351 .44179 m .36346 .44179 L .36657 .45944 L F
.36666 .45991 m .36657 .45944 L p .36662 .4597 L .912 .724 .226 r
F P
0 g
s
.36657 .45944 m .36666 .45991 L p .36682 .46037 L .24 .431 .849 r
F P
0 g
s
.37712 .52381 m .3638 .44675 L .37728 .52367 L .495 .61 .896 r
F
.3431 .6099 m .35523 .55858 L .35529 .5596 L .554 .615 .868 r
F
.34298 .60996 m .35523 .55858 L .3431 .6099 L F
.34298 .60996 m .33926 .61187 L .35523 .55858 L F
.40522 .65074 m .36078 .64039 L .37364 .59624 L .431 .452 .781 r
F
.40522 .65074 m .37364 .59624 L .39054 .5901 L F
.35712 .58735 m .35807 .60174 L .39913 .56985 L .454 .533 .849 r
F
.35712 .58735 m .39913 .56985 L .39819 .56575 L F
.4232 .543 m .40298 .50536 L .3779 .4514 L 0 0 .495 r
F
.4232 .543 m .3779 .4514 L .3795 .45381 L F
.40298 .50536 m .37718 .45032 L .3779 .4514 L F
.36882 .43624 m .37774 .45116 L .37863 .45249 L 0 0 .433 r
F
.37718 .45032 m .37722 .45038 L .37137 .4406 L F
.37718 .45032 m .37155 .44091 L p .39592 .49029 L .24 .431 .849 r
F P
0 g
s
.39592 .49029 m .37155 .44091 L .37016 .44106 L .24 .431 .849 r
F
.3596 .42662 m .35233 .41359 L .35816 .4205 L .238 .477 .885 r
F
.3638 .44675 m .34015 .39176 L p .3596 .42662 L F P
0 g
s
.40864 .59303 m .40276 .58566 L p .42811 .5963 L .25 .38 .8 r
F P
0 g
s
.41937 .6624 m .41095 .67442 L p .40864 .59303 L .431 .452 .781 r
F P
0 g
s
.34799 .66645 m .43449 .6943 L .41747 .65013 L .432 .384 .704 r
F
.36178 .6529 m .34799 .66645 L p .41747 .65013 L F P
0 g
s
.34799 .66645 m .32222 .663 L p .28658 .66914 L .547 .451 .698 r
F P
0 g
s
.29402 .67149 m .34799 .66645 L .28658 .66914 L .547 .451 .698 r
F
.3737 .68671 m .30446 .66283 L .34826 .65073 L .503 .43 .707 r
F
.30446 .66283 m .30333 .66046 L .34826 .65073 L F
.3737 .68671 m .31022 .67479 L .30446 .66283 L F
.29995 .65189 m .33989 .65254 L .36014 .63236 L .521 .524 .799 r
F
.33989 .65254 m .36138 .64789 L .36014 .63236 L F
.29995 .65189 m .31786 .65732 L .33989 .65254 L F
.37274 .68535 m .34826 .65073 L .37752 .68548 L .503 .43 .707 r
F
.34826 .65073 m .34952 .65046 L .37752 .68548 L F
.37752 .68548 m .34952 .65046 L .36998 .65249 L F
.36138 .64789 m .36998 .65249 L p .34952 .65046 L F P
0 g
s
.25247 .67344 m .25233 .67352 L p .25273 .67341 L .547 .451 .698 r
F P
0 g
s
.25233 .67352 m .24807 .67578 L p .25273 .67341 L .547 .451 .698 r
F P
0 g
s
.28502 .65477 m .31393 .62501 L p .28241 .65523 L .185 .032 .427 r
F P
0 g
s
.30522 .63227 m .30591 .63233 L p .30423 .63433 L .576 .571 .81 r
F P
0 g
s
.30591 .63233 m .30522 .63227 L p .31322 .62562 L .554 .615 .868 r
F P
0 g
s
.32788 .6177 m .31454 .62453 L .33052 .61125 L .554 .615 .868 r
F
.33052 .61125 m .31454 .62453 L .35479 .55195 L F
.3079 .62793 m .35479 .55195 L p .31454 .62453 L F P
0 g
s
.31393 .62501 m .30012 .63923 L p .32121 .62111 L .521 .524 .799 r
F P
0 g
s
.32121 .62111 m .31454 .62453 L .31393 .62501 L .521 .524 .799 r
F
.30012 .63923 m .28223 .65764 L p .30008 .64149 L F P
0 g
s
.27764 .666 m .26537 .66919 L .30013 .63801 L .576 .571 .81 r
F
.30012 .63923 m .30013 .63801 L .30423 .63433 L F
.27764 .666 m .30013 .63801 L .30012 .63923 L F
.29061 .66262 m .27764 .666 L .30012 .63923 L F
.29061 .66262 m .30012 .63923 L .30006 .64298 L F
.30423 .63433 m .26537 .66919 L .30522 .63227 L F
.26631 .66298 m .27313 .65894 L .27537 .65708 L .673 .96 .941 r
F
.25778 .66726 m .28134 .6532 L .28523 .64984 L F
.2654 .6609 m .25401 .6691 L p .26978 .65919 L F P
0 g
s
.25401 .6691 m .24736 .67388 L p .25778 .66726 L .673 .96 .941 r
F P
0 g
s
.24736 .67388 m .30016 .63591 L p .26537 .66919 L .576 .571 .81 r
F P
0 g
s
.30016 .63591 m .30522 .63227 L p .26537 .66919 L .576 .571 .81 r
F P
0 g
s
.35688 .55305 m .33853 .59467 L .35867 .54566 L .554 .615 .868 r
F
.41027 .65049 m .39201 .58956 L .40864 .59303 L .431 .452 .781 r
F
.40276 .58566 m .40864 .59303 L p .39201 .58956 L F P
0 g
s
.4306 .66606 m .41747 .65013 L p .43449 .6943 L .432 .384 .704 r
F P
0 g
s
.4306 .66606 m .4592 .67538 L .45837 .67129 L .276 .334 .742 r
F
.47399 .7388 m .43515 .69914 L .4375 .66831 L .314 .297 .679 r
F
.47399 .7388 m .4375 .66831 L .4532 .67343 L F
.41564 .63829 m .4306 .66606 L .45837 .67129 L .276 .334 .742 r
F
.41747 .65013 m .4306 .66606 L p .41564 .63829 L F P
0 g
s
.37752 .68548 m .36998 .65249 L .37828 .6855 L .503 .43 .707 r
F
.37976 .69529 m .37274 .68535 L .37752 .68548 L F
.37976 .69529 m .3806 .69473 L p .40171 .70477 L .452 .336 .631 r
F P
0 g
s
.40707 .71436 m .40211 .71093 L .40743 .71597 L .617 .3 .409 r
F
.406 .71493 m .3831 .69808 L .40218 .70587 L .452 .336 .631 r
F
.4157 .72538 m .38368 .69856 L .41727 .72668 L 0 .103 .592 r
F
.37976 .69529 m .3831 .69808 L p .38307 .69805 L F P
0 g
s
.3831 .69808 m .41727 .72668 L p .38307 .69805 L 0 .103 .592 r
F P
0 g
s
.38054 .69601 m .34023 .68536 L p .32725 .68155 L .228 .566 .938 r
F P
0 g
s
.33053 .68234 m .32725 .68155 L .32187 .67997 L .228 .566 .938 r
F
.38054 .69601 m .32725 .68155 L .33053 .68234 L F
.33921 .68442 m .37976 .69529 L p .33366 .68309 L F P
0 g
s
.3831 .69808 m .37976 .69529 L p .40171 .70477 L .452 .336 .631 r
F P
0 g
s
.40404 .71029 m .43817 .72317 L .402 .70546 L .504 .363 .621 r
F
.40218 .70587 m .40171 .70477 L .43988 .72126 L .452 .336 .631 r
F
.3831 .69808 m .40171 .70477 L .40218 .70587 L F
.37976 .69529 m .32503 .68062 L p .3737 .68671 L .503 .43 .707 r
F P
0 g
s
.39262 .70597 m .40404 .71029 L .37814 .69378 L .504 .363 .621 r
F
.40404 .71029 m .402 .70546 L .37814 .69378 L F
.38054 .69601 m .37976 .69529 L p .4157 .72538 L 0 .103 .592 r
F P
0 g
s
.40813 .71909 m .38054 .69601 L p .4157 .72538 L 0 .103 .592 r
F P
0 g
s
.32939 .68202 m .32984 .68208 L .37976 .69529 L .228 .566 .938 r
F
.37976 .69529 m .38054 .69601 L p .32939 .68202 L F P
0 g
s
.33921 .68442 m .39262 .70597 L .37814 .69378 L .504 .363 .621 r
F
.34023 .68536 m .32108 .6798 L .32187 .67997 L .228 .566 .938 r
F
.32911 .682 m .33525 .68431 L .33921 .68442 L .504 .363 .621 r
F
.33525 .68431 m .39262 .70597 L .33921 .68442 L F
.33921 .68442 m .39921 .69884 L .3781 .6855 L F
.37828 .6855 m .39911 .6986 L p .37573 .67538 L .432 .384 .704 r
F P
0 g
s
.49288 .75132 m .49296 .75076 L p .49314 .75131 L 0 0 .426 r
F P
0 g
s
.49988 .77184 m .48648 .72314 L p .50196 .79001 L .107 .137 .609 r
F P
0 g
s
.50196 .79001 m .48648 .72314 L .48474 .72113 L .107 .137 .609 r
F
.48648 .72314 m .49988 .77184 L p .48534 .71826 L 0 0 .316 r
F P
0 g
s
.50461 .78534 m .50339 .77858 L p .5164 .82973 L .847 .618 .591 r
F P
0 g
s
.51533 .84987 m .51524 .84948 L p .51531 .84973 L 0 0 .426 r
F P
0 g
s
.51533 .84987 m .51497 .8471 L .51629 .85284 L 0 0 .426 r
F
.51751 .85943 m .51669 .86037 L p .51584 .85383 L .213 .158 .575 r
F P
0 g
s
.51851 .87481 m .51445 .84313 L .51849 .87466 L .714 .508 0 r
F
.51446 .8432 m .51849 .87466 L p .51445 .84313 L F P
0 g
s
.51751 .85943 m .51539 .85013 L p .51629 .85284 L 0 0 .426 r
F P
0 g
s
.51877 .87689 m .51991 .88573 L p .51877 .87691 L .976 .694 .301 r
F P
0 g
s
.51849 .87466 m .51991 .88577 L p .51851 .87481 L .714 .508 0 r
F P
0 g
s
.51991 .88573 m .51877 .87689 L p .51851 .87481 L .714 .508 0 r
F P
0 g
s
.51994 .88574 m .51756 .85931 L p .51349 .83794 L .889 .633 .546 r
F P
0 g
s
.50238 .76978 m .50369 .7766 L .50392 .77509 L .884 .656 .588 r
F
.47708 .73772 m .49564 .78487 L .50223 .79239 L .107 .137 .609 r
F
.47708 .73772 m .50223 .79239 L .50216 .79178 L F
.51285 .83892 m .50223 .79234 L .51345 .84053 L 0 0 .426 r
F
.50223 .79234 m .50212 .79142 L .51345 .84053 L F
.51288 .84389 m .4944 .78345 L .50159 .79166 L .213 .158 .575 r
F
.47398 .70871 m .4592 .67538 L p .4837 .71993 L 0 .106 .629 r
F P
0 g
s
.4592 .67538 m .45761 .66753 L .4837 .71993 L 0 .106 .629 r
F
.47494 .74178 m .4532 .67343 L .47398 .70871 L .314 .297 .679 r
F
.4592 .67538 m .47398 .70871 L p .4532 .67343 L F P
0 g
s
.47494 .74178 m .47398 .70871 L .47767 .74319 L .314 .297 .679 r
F
.47494 .74178 m .45328 .72698 L .47399 .7388 L F
.45826 .73317 m .45328 .72698 L .47494 .74178 L F
.43615 .70635 m .45666 .73235 L p .42986 .70503 L .452 .336 .631 r
F P
0 g
s
.42986 .70503 m .45666 .73235 L .45732 .73634 L .452 .336 .631 r
F
.51282 .83884 m .50223 .79239 L p .51285 .83892 L 0 0 .426 r
F P
0 g
s
.50223 .79239 m .50223 .79234 L .51285 .83892 L 0 0 .426 r
F
.51374 .84128 m .51282 .83884 L .51287 .84323 L 0 0 .465 r
F
.51338 .8462 m .51289 .84493 L .51288 .84389 L .213 .158 .575 r
F
.51289 .84493 m .49502 .79819 L .51288 .84389 L F
.49514 .80084 m .49502 .79819 L .51289 .84493 L F
.5129 .84499 m .51289 .84493 L .51338 .8462 L F
.49514 .80084 m .51289 .84493 L .5129 .84499 L F
.50996 .84232 m .50662 .83298 L p .51291 .84583 L F P
0 g
s
.50996 .84232 m .51291 .84583 L .51291 .84632 L .213 .158 .575 r
F
.51291 .84632 m .5167 .86775 L p .51291 .84629 L 0 0 .465 r
F P
0 g
s
.51677 .86805 m .51291 .84623 L .5167 .86775 L 0 0 .465 r
F
.51291 .84623 m .51291 .84629 L .5167 .86775 L F
.50662 .83298 m .49514 .80084 L p .5129 .84499 L .213 .158 .575 r
F P
0 g
s
.5129 .84526 m .5129 .84499 L .51338 .8462 L .213 .158 .575 r
F
.50662 .83298 m .5129 .84499 L .5129 .84526 L F
.48493 .79984 m .4994 .81983 L .48312 .7963 L .48 .312 .576 r
F
.50959 .84945 m .48932 .79624 L .49347 .79952 L .415 .269 .572 r
F
.48213 .78734 m .48632 .79387 L .49212 .79845 L .362 .274 .62 r
F
.50194 .82938 m .48597 .79958 L .48932 .79624 L .415 .269 .572 r
F
.47857 .78338 m .48667 .79545 L p .48213 .78734 L .362 .274 .62 r
F P
0 g
s
.48597 .79958 m .4848 .79267 L .48932 .79624 L .415 .269 .572 r
F
.48312 .7963 m .48598 .79965 L .47263 .77583 L .48 .312 .576 r
F
.48598 .79965 m .48497 .79368 L .47263 .77583 L F
.48312 .7963 m .47263 .77583 L .43671 .72388 L F
.46893 .77049 m .47648 .7814 L .4688 .76972 L F
.46811 .76915 m .4688 .76972 L .46334 .76143 L F
.46893 .77049 m .4688 .76972 L .46811 .76915 L F
.44405 .75067 m .46681 .77224 L .46141 .76357 L F
.4435 .74863 m .46141 .76357 L .43671 .72388 L F
.44405 .75067 m .46141 .76357 L .4435 .74863 L F
.43866 .74461 m .43512 .74164 L .43945 .74526 L 0 .103 .592 r
F
.46902 .76392 m .48198 .78707 L .49353 .79956 L .362 .274 .62 r
F
.50998 .84539 m .50959 .84945 L .49347 .79952 L .415 .269 .572 r
F
.50998 .84539 m .49347 .79952 L .49447 .80031 L F
.51494 .8519 m .51374 .84128 L .51287 .84323 L 0 0 .465 r
F
.51494 .8519 m .51287 .84323 L .5129 .84526 L F
.51586 .86304 m .51291 .84632 L p .51214 .85493 L .306 .184 .544 r
F P
0 g
s
.51291 .84632 m .51118 .84397 L .51214 .85493 L .306 .184 .544 r
F
.51586 .86304 m .51214 .85493 L .51246 .85759 L F
.51001 .85029 m .50959 .84945 L p .50998 .84539 L .415 .269 .572 r
F P
0 g
s
.51118 .85285 m .51 .84833 L p .50755 .84533 L .463 .278 .544 r
F P
0 g
s
.51246 .85759 m .51214 .85493 L .51125 .85299 L .306 .184 .544 r
F
.50843 .84642 m .50594 .8421 L .50194 .82938 L .415 .269 .572 r
F
.50594 .8421 m .50959 .84945 L p .50643 .84307 L .852 .404 .001 r
F P
0 g
s
.50755 .84533 m .50594 .8421 L p .50843 .84642 L .415 .269 .572 r
F P
0 g
s
.50717 .84492 m .50715 .84484 L .50713 .84482 L .463 .278 .544 r
F
.50713 .84482 m .51982 .88586 L p .50717 .84492 L F P
0 g
s
.48898 .8079 m .48669 .80328 L .48901 .80796 L .852 .404 .001 r
F
.48898 .8079 m .48901 .80796 L .50594 .8421 L F
.48912 .80818 m .48655 .803 L p .50194 .82938 L .415 .269 .572 r
F P
0 g
s
.48655 .803 m .48597 .79958 L .50194 .82938 L .415 .269 .572 r
F
.48677 .8043 m .48668 .80376 L .47826 .79063 L .48 .312 .576 r
F
.48198 .78707 m .49231 .7986 L .49353 .79956 L .362 .274 .62 r
F
.47857 .78338 m .48213 .78734 L .48198 .78707 L F
.5171 .8699 m .51677 .86805 L .51742 .87185 L 0 0 .465 r
F
.5167 .86775 m .51742 .87185 L p .51677 .86805 L F P
0 g
s
.5223 .84117 m .53191 .78614 L .52477 .83753 L .774 .561 .611 r
F
.52012 .88577 m .52435 .86076 L p .52045 .87272 L .738 .512 .586 r
F P
0 g
s
.52012 .88577 m .52018 .88577 L p .52443 .86053 L .661 .439 .572 r
F P
0 g
s
.52435 .86076 m .52012 .88577 L p .52443 .86053 L .661 .439 .572 r
F P
0 g
s
.55971 .77368 m .56095 .77066 L p .55888 .77492 L .424 .201 .468 r
F P
0 g
s
.5619 .76883 m .56095 .77066 L .58475 .72761 L .909 .796 .724 r
F
.56095 .77066 m .56265 .76771 L .58475 .72761 L F
.55866 .77571 m .56027 .77184 L .55971 .77368 L .625 .411 .571 r
F
.56095 .77066 m .55971 .77368 L p .56027 .77184 L F P
0 g
s
.58948 .71821 m .5619 .76883 L .58475 .72761 L .909 .796 .724 r
F
.55971 .77368 m .53156 .84238 L p .55866 .77571 L .625 .411 .571 r
F P
0 g
s
.55866 .77571 m .53156 .84238 L .53164 .84211 L .625 .411 .571 r
F
.52338 .86676 m .53153 .84181 L .52443 .86053 L .661 .439 .572 r
F
.52143 .87941 m .52265 .87147 L .52348 .86818 L 0 0 .274 r
F
.52018 .88577 m .53082 .84456 L p .52338 .86676 L .661 .439 .572 r
F P
0 g
s
.53082 .84456 m .53153 .84181 L p .52338 .86676 L .661 .439 .572 r
F P
0 g
s
.5315 .8409 m .53151 .84111 L .52779 .85157 L .625 .411 .571 r
F
.53153 .84181 m .53148 .84046 L .52443 .86053 L .661 .439 .572 r
F
.5315 .8409 m .53207 .83953 L .53151 .84111 L .625 .411 .571 r
F
.53153 .84183 m .53153 .84183 L .52018 .88577 L .951 .759 .609 r
F
.65685 .63213 m .64895 .64238 L .65792 .64157 L .852 .941 .88 r
F
.65792 .64157 m .64895 .64238 L .61483 .70041 L F
.64533 .65876 m .62708 .68368 L .64205 .70053 L F
.62708 .68368 m .61483 .70041 L .63766 .70051 L F
.62708 .68368 m .63766 .70051 L .64205 .70053 L F
.60491 .70159 m .62261 .67656 L .64064 .72159 L .942 .938 .76 r
F
.66025 .66165 m .66021 .66164 L .66022 .66173 L .852 .941 .88 r
F
.66022 .66176 m .66025 .66165 L p .66022 .66173 L F P
0 g
s
.66877 .62676 m .66025 .66165 L p .66718 .6309 L 0 0 .267 r
F P
0 g
s
.66025 .66165 m .66877 .62676 L p .66428 .63289 L .852 .941 .88 r
F P
0 g
s
.66428 .63289 m .64533 .65876 L .66025 .66165 L .852 .941 .88 r
F
.66177 .65615 m .66022 .66176 L .65722 .63143 L .763 .892 .935 r
F
.66022 .66176 m .65731 .63619 L .65722 .63143 L F
.65685 .63213 m .65722 .63143 L p .65731 .63619 L F P
0 g
s
.66022 .66173 m .64533 .65876 L .64933 .68963 L .852 .941 .88 r
F
.66021 .66164 m .64533 .65876 L .66022 .66173 L F
.63979 .7039 m .64064 .72159 L p .62261 .67656 L .942 .938 .76 r
F P
0 g
s
.63101 .7162 m .64064 .72159 L .63019 .71746 L .942 .938 .76 r
F
.64621 .73264 m .65546 .73874 L p .65059 .73578 L 0 .394 .79 r
F P
0 g
s
.65059 .73578 m .64209 .7306 L .64621 .73264 L 0 .394 .79 r
F
.63578 .72669 m .63363 .72546 L p .63505 .72745 L .827 .365 .115 r
F P
0 g
s
.63191 .73074 m .63663 .7258 L .64621 .73264 L .98 .833 .626 r
F
.63663 .7258 m .64064 .72159 L .64621 .73264 L F
.63075 .72685 m .64064 .72159 L .6345 .72803 L F
.62805 .72423 m .64064 .72159 L .63075 .72685 L F
.62789 .72408 m .62403 .72754 L .63037 .72648 L .927 .88 .768 r
F
.59369 .72602 m .60834 .72463 L p .61624 .72672 L .98 .833 .626 r
F P
0 g
s
.59423 .72439 m .61575 .72667 L .6229 .72419 L .99 .868 .614 r
F
.61747 .72377 m .62668 .7229 L p .61624 .72672 L .98 .833 .626 r
F P
0 g
s
.62668 .7229 m .64064 .72159 L p .62805 .72423 L .98 .833 .626 r
F P
0 g
s
.62789 .72408 m .62685 .72307 L .62403 .72754 L .927 .88 .768 r
F
.62391 .72719 m .62403 .72754 L p .62685 .72307 L F P
0 g
s
.63937 .70323 m .62668 .7229 L .62685 .72307 L .927 .88 .768 r
F
.62668 .7229 m .62391 .72719 L .62685 .72307 L F
.63076 .72686 m .6345 .72803 L .63448 .72805 L .98 .833 .626 r
F
.62472 .72494 m .63075 .72685 L .63076 .72686 L F
.63075 .72685 m .6345 .72803 L .63076 .72686 L F
.62668 .7229 m .62805 .72423 L .61624 .72672 L F
.64429 .70937 m .63937 .70323 L .62685 .72307 L .927 .88 .768 r
F
.64429 .70937 m .62685 .72307 L .62789 .72408 L F
.6483 .70578 m .62789 .72408 L .65547 .72227 L F
.62789 .72408 m .63037 .72648 L .65547 .72227 L F
.62472 .72494 m .62805 .72423 L .63075 .72685 L .98 .833 .626 r
F
.64672 .69353 m .6577 .7291 L .64549 .69537 L .816 .871 .887 r
F
.6577 .7291 m .65801 .73012 L .65759 .72894 L F
.6577 .7291 m .65759 .72894 L .64549 .69537 L F
.64549 .69537 m .65759 .72894 L .64018 .70332 L F
.6563 .72705 m .63664 .72543 L .6586 .74022 L .927 .88 .768 r
F
.6533 .72263 m .63664 .72543 L .6563 .72705 L F
.64018 .70332 m .63937 .70323 L p .64429 .70937 L F P
0 g
s
.64533 .65876 m .64205 .70053 L .64933 .68963 L .852 .941 .88 r
F
.65712 .72668 m .6608 .73822 L .65881 .73121 L .934 .695 .216 r
F
.65537 .71991 m .65798 .729 L .65881 .73121 L F
.65537 .71991 m .65765 .7281 L .65798 .729 L F
.65749 .72659 m .65537 .71991 L p .65881 .73121 L F P
0 g
s
.64672 .69353 m .6602 .73327 L .65801 .73012 L .816 .871 .887 r
F
.6602 .73327 m .65886 .73247 L p .65801 .73012 L F P
0 g
s
.64205 .70053 m .63766 .70051 L .64128 .70168 L .852 .941 .88 r
F
.65075 .70057 m .65577 .67999 L p .64933 .68963 L F P
0 g
s
.64672 .69353 m .65577 .67999 L .65054 .70478 L .816 .871 .887 r
F
.65767 .671 m .65763 .6712 L .65765 .67107 L 0 0 .267 r
F
.65577 .67999 m .66022 .66176 L p .64933 .68963 L .852 .941 .88 r
F P
0 g
s
.66022 .66173 m .64933 .68963 L .66022 .66176 L .852 .941 .88 r
F
.658 .67113 m .6602 .73327 L p .65054 .70478 L .816 .871 .887 r
F P
0 g
s
.67145 .61771 m .66177 .65615 L .65722 .63143 L .763 .892 .935 r
F
.60491 .70159 m .63101 .7162 L .60218 .70641 L .942 .938 .76 r
F
.63101 .7162 m .63019 .71746 L .60218 .70641 L F
.61624 .72672 m .62472 .72494 L .63076 .72686 L .98 .833 .626 r
F
.63178 .72785 m .63076 .72686 L .63448 .72805 L F
.61624 .72672 m .63076 .72686 L .63178 .72785 L F
.60218 .70641 m .63019 .71746 L .61747 .72377 L .942 .938 .76 r
F
.62668 .7229 m .61747 .72377 L p .63019 .71746 L F P
0 g
s
.57961 .73908 m .5811 .7373 L .58083 .73795 L .973 .908 .565 r
F
.62232 .72832 m .62403 .72754 L p .62166 .72815 L .99 .868 .614 r
F P
0 g
s
.61491 .72988 m .62232 .72832 L .62166 .72815 L .99 .868 .614 r
F
.61397 .73212 m .62232 .72832 L p .61491 .72988 L F P
0 g
s
.59369 .72602 m .61624 .72672 L .61538 .72875 L .98 .833 .626 r
F
.61575 .72667 m .61136 .72818 L .62403 .72754 L .99 .868 .614 r
F
.61091 .73273 m .60874 .73316 L .61005 .73249 L .906 .541 .331 r
F
.58427 .74563 m .58631 .74414 L .61005 .73249 L F
.61005 .73249 m .58631 .74414 L .59495 .73783 L F
.59994 .73275 m .61091 .73273 L p .59824 .72953 L .827 .365 .115 r
F P
0 g
s
.61091 .73273 m .59995 .72967 L .59824 .72953 L .827 .365 .115 r
F
.5973 .73276 m .59994 .73275 L p .59824 .72953 L F P
0 g
s
.61091 .73273 m .5973 .73276 L p .6141 .73182 L .98 .833 .626 r
F P
0 g
s
.61602 .73171 m .61397 .73212 L .6141 .73182 L .98 .833 .626 r
F
.61397 .73212 m .61091 .73273 L .6141 .73182 L F
.61572 .72796 m .61491 .72988 L .62403 .72754 L .99 .868 .614 r
F
.59824 .72953 m .59833 .72922 L .59803 .72913 L .827 .365 .115 r
F
.5973 .73276 m .59369 .72602 L p .62488 .72995 L .98 .833 .626 r
F P
0 g
s
.62488 .72995 m .61602 .73171 L .5973 .73276 L .98 .833 .626 r
F
.61136 .72818 m .60492 .73244 L .61491 .72988 L .99 .868 .614 r
F
.61136 .72818 m .61491 .72988 L .61572 .72796 L F
.61538 .72875 m .61624 .72672 L .63183 .73083 L .98 .833 .626 r
F
.62695 .72954 m .63448 .72805 L .63183 .73083 L F
.61624 .72672 m .63448 .72805 L .62695 .72954 L F
.59998 .7177 m .59423 .72439 L .61728 .72423 L .99 .868 .614 r
F
.5973 .73276 m .58635 .74411 L p .58427 .74563 L .997 .8 .518 r
F P
0 g
s
.58294 .7466 m .584 .74558 L .58427 .74563 L .997 .8 .518 r
F
.584 .74558 m .5973 .73276 L .58427 .74563 L F
.59394 .7265 m .5973 .73276 L p .58294 .7466 L F P
0 g
s
.57338 .75359 m .59394 .7265 L .58294 .7466 L .997 .8 .518 r
F
.59423 .72439 m .58236 .73822 L .61575 .72667 L .99 .868 .614 r
F
.63878 .73266 m .64621 .73264 L p .63798 .73155 L .827 .365 .115 r
F P
0 g
s
.63183 .73083 m .63191 .73074 L .64621 .73264 L .98 .833 .626 r
F
.64621 .73264 m .63878 .73266 L p .63183 .73083 L F P
0 g
s
.63664 .72543 m .63037 .72648 L .63905 .73305 L .927 .88 .768 r
F
.63599 .73193 m .63905 .73305 L p .63037 .72648 L F P
0 g
s
.61198 .73302 m .58427 .74563 L p .60874 .73316 L .906 .541 .331 r
F P
0 g
s
.61198 .73302 m .60874 .73316 L .61091 .73273 L .906 .541 .331 r
F
.61198 .73302 m .61397 .73212 L p .58236 .73822 L .99 .868 .614 r
F P
0 g
s
.58236 .73822 m .61397 .73212 L .61491 .72988 L .99 .868 .614 r
F
.63183 .73083 m .61372 .73272 L .61383 .73246 L .98 .833 .626 r
F
.57094 .75537 m .57338 .75359 L .56704 .76194 L .997 .8 .518 r
F
.57911 .74201 m .57808 .74443 L .58642 .74465 L .99 .868 .614 r
F
.58427 .74563 m .58642 .74465 L p .57808 .74443 L F P
0 g
s
.52 .88486 m .5199 .88572 L .51991 .88573 L p .993 .746 .397 r
F P
0 g
s
.51742 .87185 m .51846 .87774 L p .51748 .8721 L 0 0 .465 r
F P
0 g
s
.5171 .8699 m .51742 .87185 L .51748 .8721 L 0 0 .465 r
F
.50594 .8421 m .48912 .80818 L p .49643 .82026 L .415 .269 .572 r
F P
0 g
s
.50594 .8421 m .49643 .82026 L .50194 .82938 L .415 .269 .572 r
F
.47018 .77087 m .47857 .78338 L p .47112 .77165 L .362 .274 .62 r
F P
0 g
s
.47018 .77087 m .47112 .77165 L .46754 .76601 L .362 .274 .62 r
F
.43817 .72317 m .43671 .72388 L p .46893 .77049 L .48 .312 .576 r
F P
0 g
s
.47857 .78338 m .43817 .72317 L p .47648 .7814 L .48 .312 .576 r
F P
0 g
s
.45984 .75179 m .48198 .78707 L .46902 .76392 L .362 .274 .62 r
F
.43817 .72317 m .46048 .75571 L .45984 .75179 L F
.46219 .75739 m .46059 .75639 L .46092 .75534 L .452 .336 .631 r
F
.46059 .75639 m .46037 .75501 L .46092 .75534 L F
.42569 .72922 m .43817 .72317 L p .39262 .70597 L .504 .363 .621 r
F P
0 g
s
.422 .72979 m .40739 .71595 L .46092 .75534 L .452 .336 .631 r
F
.44318 .74838 m .42979 .73716 L p .46219 .75739 L F P
0 g
s
.42979 .73716 m .422 .72979 L .46092 .75534 L .452 .336 .631 r
F
.43671 .72388 m .42569 .72922 L p .43853 .73854 L .48 .312 .576 r
F P
0 g
s
.46092 .75534 m .40739 .71595 L .43988 .72126 L .452 .336 .631 r
F
.46048 .75571 m .48198 .78707 L .45984 .75179 L .362 .274 .62 r
F
.46059 .75639 m .42979 .73716 L .46037 .75501 L .452 .336 .631 r
F
.46048 .75571 m .4698 .77031 L .46754 .76601 L .362 .274 .62 r
F
.46063 .75665 m .4698 .77031 L p .46048 .75571 L F P
0 g
s
.46024 .75423 m .42986 .70503 L .45732 .73634 L .452 .336 .631 r
F
.43988 .72126 m .40171 .70477 L .40332 .69948 L F
.43988 .72126 m .40332 .69948 L .42986 .70503 L F
.45328 .72698 m .43604 .70558 L .47399 .7388 L .314 .297 .679 r
F
.43604 .70558 m .43515 .69914 L .47399 .7388 L F
.37814 .69378 m .402 .70546 L .39921 .69884 L .504 .363 .621 r
F
.37819 .70052 m .40813 .71996 L .39262 .70597 L F
.40813 .71996 m .4129 .72306 L .40752 .71852 L F
.40813 .71996 m .40752 .71852 L .39262 .70597 L F
.41727 .72668 m .40752 .71852 L p .40739 .71595 L .452 .336 .631 r
F P
0 g
s
.40752 .71852 m .406 .71493 L .40739 .71595 L .452 .336 .631 r
F
.41727 .72668 m .40739 .71595 L .422 .72979 L F
.42979 .73716 m .41727 .72668 L p .422 .72979 L F P
0 g
s
.40739 .71595 m .406 .71493 L .40639 .715 L .452 .336 .631 r
F
.40739 .71595 m .40639 .715 L .43988 .72126 L F
.42979 .73716 m .44318 .74838 L p .43222 .7392 L 0 .103 .592 r
F P
0 g
s
.43222 .7392 m .44318 .74838 L .44172 .74716 L 0 .103 .592 r
F
.43935 .74518 m .44405 .75067 L .4435 .74863 L .48 .312 .576 r
F
.44405 .75067 m .47826 .79063 L .46681 .77224 L F
.43935 .74518 m .44007 .7469 L .44405 .75067 L F
.44007 .7469 m .43725 .74343 L p .44055 .74715 L .599 .332 .483 r
F P
0 g
s
.43725 .74343 m .44007 .7469 L p .43935 .74518 L .48 .312 .576 r
F P
0 g
s
.40752 .71852 m .40627 .71557 L .39262 .70597 L .504 .363 .621 r
F
.406 .71493 m .40218 .70587 L .43988 .72126 L .452 .336 .631 r
F
.32911 .682 m .40855 .72096 L .37819 .70052 L .504 .363 .621 r
F
.32631 .68132 m .40855 .72096 L p .32911 .682 L F P
0 g
s
.32503 .68062 m .33921 .68442 L p .32187 .67997 L .228 .566 .938 r
F P
0 g
s
.32187 .67997 m .33921 .68442 L .33366 .68309 L .228 .566 .938 r
F
.32187 .67997 m .31789 .67888 L .32503 .68062 L F
.41727 .72668 m .42979 .73716 L p .4157 .72538 L 0 .103 .592 r
F P
0 g
s
.4157 .72538 m .42979 .73716 L .43222 .7392 L 0 .103 .592 r
F
.43512 .74164 m .4157 .72538 L .43222 .7392 L F
.44405 .75067 m .48677 .8043 L .47826 .79063 L .48 .312 .576 r
F
.44007 .7469 m .48677 .8043 L p .44405 .75067 L F P
0 g
s
.40855 .72096 m .4129 .72306 L p .37819 .70052 L .504 .363 .621 r
F P
0 g
s
.43512 .74164 m .41629 .72591 L .4157 .72538 L 0 .103 .592 r
F
.41629 .72591 m .4129 .72306 L .4157 .72538 L F
.47826 .79063 m .48493 .79984 L .48312 .7963 L .48 .312 .576 r
F
.51971 .88514 m .51125 .85299 L .5198 .88571 L .306 .184 .544 r
F
.51118 .85285 m .5198 .88571 L p .51125 .85299 L F P
0 g
s
.51985 .88584 m .51984 .88585 L p .51985 .88577 L .212 .036 .411 r
F P
0 g
s
.51984 .88585 m .51989 .88582 L p .51971 .88514 L .306 .184 .544 r
F P
0 g
s
.5198 .88571 m .51984 .88585 L p .51971 .88514 L .306 .184 .544 r
F P
0 g
s
.51994 .88574 m .51993 .88565 L .52 .88486 L .809 .517 .495 r
F
.52 .88486 m .51994 .88574 L .51993 .88574 L p .891 .597 .483 r
F P
0 g
s
.51991 .88577 m .51997 .88525 L .51991 .88573 L .883 .733 .222 r
F
.51993 .88574 m .5199 .88572 L p .52 .88486 L .891 .597 .483 r
F P
0 g
s
.61136 .72818 m .58236 .73822 L .60492 .73244 L .99 .868 .614 r
F
.56704 .76194 m .57338 .75359 L .58294 .7466 L .997 .8 .518 r
F
.58236 .73822 m .57911 .74201 L .58642 .74465 L .99 .868 .614 r
F
.58642 .74465 m .61198 .73302 L p .60116 .73492 L F P
0 g
s
.58236 .73822 m .58642 .74465 L .60116 .73492 L .99 .868 .614 r
F
.51992 .88586 m .51985 .88586 L p .51993 .88582 L .685 .407 .496 r
F P
0 g
s
.51986 .88586 m .51993 .88586 L p .51986 .8858 L .698 .467 .569 r
F P
0 g
s
.51989 .88582 m .51991 .88577 L p .51962 .88412 L 0 0 .465 r
F P
0 g
s
.51962 .88412 m .5171 .8699 L .51974 .885 L 0 0 .465 r
F
.51962 .88412 m .51974 .885 L .51989 .88582 L F
.51846 .87774 m .51989 .88582 L p .5171 .8699 L F P
0 g
s
.51994 .88574 m .52002 .88576 L p .51996 .88507 L .811 .574 .581 r
F P
0 g
s
.51992 .88556 m .51994 .88574 L p .51996 .88507 L .811 .574 .581 r
F P
0 g
s
.51991 .88573 m .51991 .88573 L p .51991 .88573 L .883 .733 .222 r
F P
0 g
s
.51991 .88573 m .51991 .88577 L p .51991 .88573 L .883 .733 .222 r
F P
0 g
s
.46754 .76601 m .47857 .78338 L .48198 .78707 L .362 .274 .62 r
F
.51989 .88582 m .51742 .87185 L p .51246 .85759 L .306 .184 .544 r
F P
0 g
s
.32631 .68132 m .32322 .68058 L .32108 .6798 L .228 .566 .938 r
F
.32614 .68164 m .31896 .67937 L .32108 .6798 L F
.32939 .68202 m .32187 .67997 L .32984 .68208 L F
.27254 .67156 m .29402 .67149 L .28658 .66914 L .547 .451 .698 r
F
.25384 .67484 m .25403 .67481 L p .25431 .6748 L .48 .074 .193 r
F P
0 g
s
.24807 .67578 m .27941 .67285 L .27254 .67156 L .547 .451 .698 r
F
.27941 .67285 m .29402 .67149 L .27254 .67156 L F
.62488 .72995 m .63183 .73083 L .61602 .73171 L .98 .833 .626 r
F
.60188 .71549 m .59998 .7177 L .61297 .72138 L .99 .868 .614 r
F
.69167 .59822 m .67271 .61649 L .67033 .6227 L .766 .985 .904 r
F
.69688 .5932 m .69776 .59234 L .66799 .63146 L .763 .892 .935 r
F
.67145 .61771 m .69688 .5932 L .66799 .63146 L F
.69167 .59822 m .69214 .59776 L .71362 .5732 L .766 .985 .904 r
F
.69688 .5932 m .71362 .5732 L p .69214 .59776 L F P
0 g
s
.658 .67113 m .70293 .58737 L p .67082 .62881 L .551 .618 .873 r
F P
0 g
s
.66799 .63146 m .69776 .59234 L .658 .67113 L .763 .892 .935 r
F
.70293 .58737 m .658 .67113 L p .69776 .59234 L F P
0 g
s
.67845 .51161 m .76628 .49853 L .69244 .51935 L .7 .66 .808 r
F
.59092 .51545 m .68751 .51662 L .66871 .50623 L .726 .65 .774 r
F
.782 .4954 m .78022 .49645 L p .67692 .51077 L .7 .66 .808 r
F P
0 g
s
.67651 .51054 m .782 .4954 L .67692 .51077 L .7 .66 .808 r
F
.53415 .46514 m .5363 .47246 L p .554 .45891 L .592 .095 .044 r
F P
0 g
s
.53099 .42643 m .54876 .43836 L .56092 .45674 L .712 .271 .202 r
F
.4882 .43494 m .47009 .44318 L p .49254 .44392 L .693 .176 0 r
F P
0 g
s
.49137 .43601 m .4882 .43494 L p .53415 .46514 L .693 .176 0 r
F P
0 g
s
.46962 .44144 m .50216 .4206 L .50351 .44427 L .828 .407 .241 r
F
.50807 .43018 m .50351 .44427 L p .50216 .4206 L F P
0 g
s
.49861 .41484 m .50293 .41339 L .49981 .41679 L .828 .407 .241 r
F
.49968 .41658 m .50807 .43018 L .51118 .4291 L .795 .41 .325 r
F
.50351 .44427 m .50807 .43018 L p .51036 .44894 L .712 .271 .202 r
F P
0 g
s
.41835 .43016 m .4479 .43962 L .41879 .43208 L .881 .453 .167 r
F
.43319 .42997 m .4182 .42949 L .4191 .43349 L .863 .435 .195 r
F
.42026 .43422 m .43057 .43615 L .42261 .43499 L .631 .067 0 r
F
.4045 .42572 m .41797 .42848 L .41835 .43016 L .881 .453 .167 r
F
.42048 .42746 m .41779 .42766 L p .43319 .42997 L .863 .435 .195 r
F P
0 g
s
.41779 .42766 m .41816 .42932 L .43319 .42997 L .863 .435 .195 r
F
.43319 .42997 m .4191 .43349 L .41984 .43409 L F
.43057 .43615 m .4479 .43962 L .45442 .4427 L p .631 .067 0 r
F P
0 g
s
.43319 .42997 m .41984 .43409 L .438 .43999 L .863 .435 .195 r
F
.47009 .44318 m .4608 .44741 L p .49254 .44392 L .693 .176 0 r
F P
0 g
s
.46962 .44144 m .50351 .44427 L .47009 .44318 L .828 .407 .241 r
F
.4491 .44361 m .4608 .44741 L .46351 .44939 L .631 .067 0 r
F
.4479 .43962 m .46177 .43406 L .45532 .44025 L .828 .407 .241 r
F
.46448 .43134 m .46166 .43121 L .45213 .44459 L .863 .435 .195 r
F
.46448 .43134 m .45213 .44459 L .45312 .44491 L F
.4608 .44741 m .47009 .44318 L p .45415 .44524 L F P
0 g
s
.45415 .44524 m .47009 .44318 L .4694 .44063 L .863 .435 .195 r
F
.46706 .43203 m .45415 .44524 L .4694 .44063 L F
.45362 .42552 m .45998 .42892 L p .45115 .42889 L .983 .809 .589 r
F P
0 g
s
.46946 .42108 m .47851 .41864 L .44726 .43914 L .916 .564 .354 r
F
.45255 .42564 m .43878 .43274 L p .45197 .42699 L .994 .841 .584 r
F P
0 g
s
.45115 .42889 m .45256 .42563 L .45362 .42552 L .983 .809 .589 r
F
.43878 .43274 m .45255 .42564 L p .45711 .42441 L .916 .564 .354 r
F P
0 g
s
.44823 .43568 m .4504 .43406 L .44914 .43357 L .994 .841 .584 r
F
.44823 .43568 m .45197 .42699 L .43878 .43274 L F
.44726 .43914 m .43878 .43274 L p .45711 .42441 L .916 .564 .354 r
F P
0 g
s
.45711 .42441 m .46946 .42108 L .44726 .43914 L .916 .564 .354 r
F
.47508 .42131 m .4479 .43962 L p .47614 .4202 L F P
0 g
s
.46734 .42792 m .45279 .42508 L .46448 .43134 L .863 .435 .195 r
F
.45534 .42489 m .45279 .42508 L p .46734 .42792 L F P
0 g
s
.46584 .42754 m .4709 .42413 L p .46704 .43195 L .828 .407 .241 r
F P
0 g
s
.47582 .41591 m .47851 .41864 L .46946 .42108 L .916 .564 .354 r
F
.50631 .38634 m .50299 .38677 L .4936 .41737 L .893 .542 .373 r
F
.49448 .40116 m .48222 .41287 L .48284 .42083 L F
.4709 .42413 m .49861 .41484 L .49981 .41679 L .828 .407 .241 r
F
.51257 .37182 m .51615 .36561 L .51001 .40564 L .852 .512 .409 r
F
.51257 .37182 m .51001 .40564 L .50975 .40431 L F
.50776 .39388 m .49814 .41408 L p .51661 .38951 L .795 .41 .325 r
F P
0 g
s
.4728 .42349 m .48249 .41632 L .50293 .41339 L .828 .407 .241 r
F
.49814 .41408 m .49968 .41658 L .51661 .38951 L .795 .41 .325 r
F
.50932 .41125 m .51126 .4122 L p .50293 .41339 L .828 .407 .241 r
F P
0 g
s
.50975 .40431 m .51001 .40564 L .50916 .41117 L .852 .512 .409 r
F
.51661 .38951 m .49968 .41658 L .51118 .4291 L .795 .41 .325 r
F
.51154 .37772 m .51112 .37744 L p .50466 .40039 L .893 .542 .373 r
F P
0 g
s
.50659 .3868 m .51257 .37182 L .51079 .39239 L .852 .512 .409 r
F
.50634 .38646 m .50808 .38237 L .50659 .3868 L F
.50808 .38237 m .51257 .37182 L .50659 .3868 L F
.5103 .37574 m .51469 .36379 L .51102 .37449 L F
.50928 .37751 m .51212 .37031 L p .5103 .37574 L F P
0 g
s
.51212 .37031 m .51469 .36379 L p .5103 .37574 L .852 .512 .409 r
F P
0 g
s
.51476 .36195 m .51376 .36512 L .50901 .37791 L .919 .605 .433 r
F
.51532 .36011 m .51523 .36048 L .51531 .36011 L .995 .756 .431 r
F
.51521 .36052 m .51531 .36011 L p .51523 .36048 L F P
0 g
s
.50331 .38651 m .50901 .37791 L .50728 .38257 L .919 .605 .433 r
F
.50357 .38612 m .50299 .38677 L p .50674 .3853 L .893 .542 .373 r
F P
0 g
s
.50634 .38646 m .50728 .38257 L .50932 .37944 L .852 .512 .409 r
F
.50674 .3853 m .50616 .38553 L .50631 .38634 L .893 .542 .373 r
F
.50615 .38545 m .50728 .38257 L p .50634 .38646 L .852 .512 .409 r
F P
0 g
s
.50728 .38257 m .50928 .37751 L p .50932 .37944 L .852 .512 .409 r
F P
0 g
s
.50928 .37751 m .51102 .37449 L .50932 .37944 L .852 .512 .409 r
F
.50466 .40039 m .50776 .39388 L p .50732 .39161 L .893 .542 .373 r
F P
0 g
s
.50634 .38646 m .51079 .39239 L .50776 .39388 L .852 .512 .409 r
F
.51697 .39061 m .51915 .36423 L .52127 .36226 L .76 .42 .409 r
F
.51624 .3897 m .51915 .36423 L p .51697 .39061 L F P
0 g
s
.50616 .38553 m .50299 .38677 L .50631 .38634 L .893 .542 .373 r
F
.51257 .37182 m .51668 .36214 L .51615 .36561 L .852 .512 .409 r
F
.50503 .38574 m .50375 .38607 L .50728 .38257 L .919 .605 .433 r
F
.49311 .40186 m .4929 .40203 L .49417 .40027 L .995 .783 .504 r
F
.48206 .41084 m .48212 .41163 L .47582 .41591 L .916 .564 .354 r
F
.4709 .42413 m .47508 .42131 L p .4728 .42349 L .828 .407 .241 r
F P
0 g
s
.47508 .42131 m .48249 .41632 L p .4728 .42349 L .828 .407 .241 r
F P
0 g
s
.49814 .41408 m .50466 .40039 L p .49647 .41432 L .893 .542 .373 r
F P
0 g
s
.46166 .43121 m .4361 .43603 L .438 .43999 L .863 .435 .195 r
F
.46166 .43121 m .438 .43999 L .45213 .44459 L F
.45711 .42441 m .47582 .41591 L .46946 .42108 L .916 .564 .354 r
F
.43731 .42894 m .43319 .42997 L .46448 .43134 L .863 .435 .195 r
F
.42048 .42746 m .43319 .42997 L .43731 .42894 L F
.46166 .43121 m .43319 .42997 L .4361 .43603 L F
.45279 .42508 m .42048 .42746 L p .43731 .42894 L F P
0 g
s
.45279 .42508 m .43731 .42894 L .46448 .43134 L .863 .435 .195 r
F
.34015 .39176 m .34955 .40521 L p .35233 .41359 L .238 .477 .885 r
F P
0 g
s
.51531 .36011 m .52098 .33672 L p .51532 .36011 L .995 .756 .431 r
F P
0 g
s
.50932 .37944 m .51469 .36379 L .51257 .37181 L .852 .512 .409 r
F
.50299 .38677 m .48308 .40967 L p .49448 .40116 L .893 .542 .373 r
F P
0 g
s
.5152 .36042 m .51521 .36052 L .51476 .36195 L .919 .605 .433 r
F
.51531 .36011 m .51521 .36052 L p .51889 .35062 L .894 .587 .457 r
F P
0 g
s
.51974 .34262 m .51748 .35329 L p .51755 .353 L 0 0 .016 r
F P
0 g
s
.51999 .34123 m .51889 .34715 L .5181 .35028 L .941 .796 .337 r
F
.51943 .345 m .51882 .34824 L p .51897 .34683 L F P
0 g
s
.51882 .34824 m .51849 .34949 L .51897 .34683 L .941 .796 .337 r
F
.51943 .345 m .51897 .34683 L .51999 .34123 L F
.52029 .33982 m .52098 .33673 L p .52028 .33995 L F P
0 g
s
.52099 .33678 m .52097 .33683 L p .51971 .34347 L 0 0 .016 r
F P
0 g
s
.51976 .34328 m .52099 .33678 L p .51971 .34347 L 0 0 .016 r
F P
0 g
s
.52099 .33678 m .51976 .34328 L p .51999 .34123 L .941 .796 .337 r
F P
0 g
s
.52098 .33672 m .51531 .36011 L p .51889 .35062 L .894 .587 .457 r
F P
0 g
s
.51469 .36379 m .50728 .38257 L p .51497 .3627 L .919 .605 .433 r
F P
0 g
s
.52038 .34609 m .51747 .36013 L .51915 .36423 L .81 .496 .459 r
F
.52101 .33674 m .51892 .35053 L p .52038 .34609 L F P
0 g
s
.51892 .35053 m .51747 .36013 L p .52038 .34609 L .81 .496 .459 r
F P
0 g
s
.52534 .36566 m .53192 .39319 L p .52195 .36162 L .76 .42 .409 r
F P
0 g
s
.52195 .36162 m .53192 .39319 L .53106 .39397 L .76 .42 .409 r
F
.52127 .36226 m .52195 .36162 L .53106 .39397 L F
.52127 .36226 m .53106 .39397 L .53076 .39424 L F
.52411 .35035 m .52395 .36401 L .52534 .36566 L .724 .426 .472 r
F
.52411 .35035 m .52534 .36566 L .5269 .36284 L F
.53842 .38309 m .53711 .38321 L p .54605 .39408 L .622 .315 .429 r
F P
0 g
s
.54605 .39408 m .5412 .38682 L .53842 .38309 L .622 .315 .429 r
F
.53711 .38321 m .5323 .38367 L p .5423 .39258 L F P
0 g
s
.5423 .39258 m .55173 .40099 L .53711 .38321 L .622 .315 .429 r
F
.53152 .38358 m .5323 .38367 L p .53026 .37514 L .724 .426 .472 r
F P
0 g
s
.54163 .40025 m .5284 .37284 L .54373 .38945 L .673 .365 .443 r
F
.5328 .39383 m .5284 .37284 L .54163 .40025 L F
.54352 .40162 m .54597 .39397 L .54407 .40202 L F
.535 .39867 m .53739 .39717 L .53378 .39454 L .699 .333 .35 r
F
.55943 .42873 m .56023 .42775 L .54395 .40194 L F
.55943 .42873 m .54395 .40194 L .54352 .40162 L F
.54352 .40162 m .54373 .38945 L .54597 .39397 L .673 .365 .443 r
F
.54163 .40025 m .54373 .38945 L .54352 .40162 L F
.53739 .39717 m .55943 .42873 L .54352 .40162 L .699 .333 .35 r
F
.54722 .3917 m .54793 .39238 L .54743 .3919 L 0 .077 .608 r
F
.54665 .39112 m .54793 .39238 L p .54722 .3917 L F P
0 g
s
.55597 .39968 m .55519 .39881 L .55654 .40082 L .481 .306 .566 r
F
.55654 .40082 m .53096 .36252 L .53255 .36577 L F
.55597 .39968 m .53086 .36231 L p .55519 .39881 L F P
0 g
s
.53086 .36231 m .53096 .36252 L .55519 .39881 L .481 .306 .566 r
F
.53255 .36577 m .53125 .36368 L .53143 .36414 L F
.53086 .36231 m .55597 .39968 L p .54079 .38326 L .6 .339 .493 r
F P
0 g
s
.55597 .39968 m .55712 .40138 L p .55512 .39917 L .6 .339 .493 r
F P
0 g
s
.54079 .38326 m .55597 .39968 L .55512 .39917 L .6 .339 .493 r
F
.56 .40418 m .57858 .42234 L p .55887 .40308 L 0 .077 .608 r
F P
0 g
s
.54079 .38326 m .55712 .40138 L .54818 .39257 L .6 .339 .493 r
F
.56 .40418 m .55887 .40308 L .54743 .3919 L 0 .077 .608 r
F
.56789 .41741 m .55712 .40138 L p .55654 .40082 L .481 .306 .566 r
F P
0 g
s
.54743 .3919 m .55887 .40308 L .55642 .40071 L 0 .077 .608 r
F
.55448 .39878 m .56 .40418 L p .55446 .39877 L F P
0 g
s
.54793 .39238 m .54814 .3926 L p .54743 .3919 L 0 .077 .608 r
F P
0 g
s
.55464 .39899 m .54677 .39128 L p .55483 .39916 L 0 .077 .608 r
F P
0 g
s
.55448 .39878 m .54793 .39238 L p .55351 .3982 L .622 .315 .429 r
F P
0 g
s
.55712 .40138 m .56789 .41741 L p .55185 .3972 L .6 .339 .493 r
F P
0 g
s
.55136 .3969 m .55351 .3982 L .54793 .39238 L .622 .315 .429 r
F
.54818 .39257 m .55448 .39878 L .55185 .3972 L .6 .339 .493 r
F
.55448 .39878 m .55712 .40138 L .55185 .3972 L F
.55035 .39629 m .55185 .3972 L .56789 .41741 L F
.55858 .40643 m .57053 .41594 L .55351 .3982 L .622 .315 .429 r
F
.55858 .40643 m .55351 .3982 L .55136 .3969 L F
.56 .40418 m .55448 .39878 L p .57053 .41594 L F P
0 g
s
.57053 .41594 m .55448 .39878 L .55351 .3982 L .622 .315 .429 r
F
.54722 .3917 m .54743 .3919 L .54724 .39172 L 0 .077 .608 r
F
.54793 .39238 m .53842 .38309 L p .54926 .39563 L .622 .315 .429 r
F P
0 g
s
.55858 .40643 m .55136 .3969 L .54926 .39563 L .622 .315 .429 r
F
.55136 .3969 m .54793 .39238 L .54926 .39563 L F
.54306 .38752 m .54818 .39257 L .55185 .3972 L .6 .339 .493 r
F
.54306 .38752 m .55185 .3972 L .55035 .39629 L F
.54648 .39396 m .54373 .38945 L .54079 .38326 L F
.54604 .39369 m .54373 .38945 L p .54648 .39396 L F P
0 g
s
.55173 .40099 m .54794 .39484 L .54653 .39399 L .622 .315 .429 r
F
.54794 .39484 m .54637 .39229 L .54653 .39399 L F
.54373 .38945 m .54604 .39369 L p .54597 .39397 L .673 .365 .443 r
F P
0 g
s
.54373 .38945 m .5307 .36548 L p .54079 .38326 L .6 .339 .493 r
F P
0 g
s
.5307 .36548 m .53071 .3652 L .54079 .38326 L .6 .339 .493 r
F
.53026 .36467 m .5307 .36548 L p .53052 .36531 L .673 .365 .443 r
F P
0 g
s
.5307 .36548 m .53884 .38045 L p .53052 .36531 L .673 .365 .443 r
F P
0 g
s
.52722 .36227 m .5238 .34789 L p .52721 .36228 L .724 .426 .472 r
F P
0 g
s
.5281 .36069 m .53026 .36467 L p .52646 .36363 L .673 .365 .443 r
F P
0 g
s
.52879 .36068 m .52789 .35928 L .52782 .35886 L .6 .339 .493 r
F
.5238 .34789 m .52722 .36227 L p .53012 .3608 L .642 .374 .498 r
F P
0 g
s
.52663 .36445 m .52646 .36363 L .53026 .36467 L .673 .365 .443 r
F
.52663 .36445 m .53026 .36467 L .5341 .37424 L F
.52722 .36227 m .53055 .36534 L .53012 .3608 L .642 .374 .498 r
F
.52365 .34751 m .52719 .36232 L .5238 .34789 L .724 .426 .472 r
F
.52719 .36232 m .52721 .36228 L .5238 .34789 L F
.53284 .36829 m .53075 .36437 L .53084 .36267 L .6 .339 .493 r
F
.53884 .38045 m .54224 .38671 L p .52663 .36445 L .673 .365 .443 r
F P
0 g
s
.54224 .38671 m .54373 .38945 L p .54102 .38549 L .673 .365 .443 r
F P
0 g
s
.52663 .36445 m .54224 .38671 L .54102 .38549 L .673 .365 .443 r
F
.54648 .39396 m .54306 .38752 L .55035 .39629 L .6 .339 .493 r
F
.58918 .44768 m .58902 .44699 L .55173 .40099 L .622 .315 .429 r
F
.58012 .43601 m .59365 .45271 L .58143 .43657 L F
.60828 .45137 m .59502 .45258 L p .61273 .45755 L .568 .188 .282 r
F P
0 g
s
.61273 .45755 m .61594 .45845 L .61316 .45588 L .568 .188 .282 r
F
.61273 .45755 m .61316 .45588 L .60828 .45137 L F
.58541 .43204 m .58558 .43275 L .56789 .41741 L .58 .289 .439 r
F
.58411 .42782 m .60828 .45137 L .60627 .44951 L 0 .077 .608 r
F
.6535 .46669 m .58414 .42676 L p .65086 .4664 L .505 .35 .604 r
F P
0 g
s
.60716 .44998 m .58446 .42809 L .65896 .46983 L .58 .289 .439 r
F
.58152 .42526 m .58414 .42676 L p .58446 .42809 L F P
0 g
s
.58414 .42676 m .65896 .46983 L p .58446 .42809 L .58 .289 .439 r
F P
0 g
s
.58152 .42526 m .60828 .45137 L .58411 .42782 L 0 .077 .608 r
F
.57858 .42234 m .58446 .42809 L p .58152 .42526 L F P
0 g
s
.58446 .42809 m .60828 .45137 L p .58152 .42526 L 0 .077 .608 r
F P
0 g
s
.56789 .41741 m .58152 .42526 L p .58446 .42809 L .58 .289 .439 r
F P
0 g
s
.60828 .45137 m .58446 .42809 L p .58509 .43071 L .622 .315 .429 r
F P
0 g
s
.60879 .45156 m .58541 .43204 L .58446 .42809 L .58 .289 .439 r
F
.58541 .43204 m .56789 .41741 L .58446 .42809 L F
.55173 .40099 m .58509 .43071 L .57858 .42234 L .622 .315 .429 r
F
.58446 .42809 m .57858 .42234 L p .58509 .43071 L F P
0 g
s
.61757 .45888 m .58261 .43572 L .59775 .44233 L .58 .289 .439 r
F
.64311 .4802 m .61266 .45782 L p .61757 .45888 L F P
0 g
s
.59502 .45258 m .59047 .45299 L p .61266 .45782 L .568 .188 .282 r
F P
0 g
s
.61266 .45782 m .6172 .4588 L .61273 .45755 L .568 .188 .282 r
F
.61266 .45782 m .61273 .45755 L .59502 .45258 L F
.61266 .45782 m .64311 .4802 L p .61233 .45912 L .631 .248 .299 r
F P
0 g
s
.66902 .47562 m .66172 .47142 L p .65965 .47087 L .505 .35 .604 r
F P
0 g
s
.66172 .47142 m .65896 .46983 L p .65965 .47087 L .505 .35 .604 r
F P
0 g
s
.65965 .47087 m .65896 .46983 L .65497 .46849 L .505 .35 .604 r
F
.64718 .46532 m .6724 .47436 L p .65896 .46983 L .211 .509 .91 r
F P
0 g
s
.60879 .45156 m .60716 .44998 L .66902 .47562 L .58 .289 .439 r
F
.64718 .46532 m .65896 .46983 L .65602 .46884 L .211 .509 .91 r
F
.65896 .46983 m .66902 .47562 L p .60716 .44998 L .58 .289 .439 r
F P
0 g
s
.61255 .45826 m .67153 .47404 L .6172 .4588 L .568 .188 .282 r
F
.57751 .45417 m .61175 .46141 L .61255 .45826 L F
.61221 .4596 m .61233 .45912 L .64311 .4802 L .631 .248 .299 r
F
.61175 .46141 m .67153 .47404 L .61255 .45826 L .568 .188 .282 r
F
.59047 .45299 m .57751 .45417 L p .6172 .4588 L F P
0 g
s
.58918 .44768 m .59365 .45271 L .58902 .44699 L .622 .315 .429 r
F
.57751 .45417 m .61133 .46304 L .61175 .46141 L .568 .188 .282 r
F
.57751 .45417 m .61099 .46436 L .61133 .46304 L F
.56856 .43103 m .58571 .45343 L .56967 .43151 L .622 .315 .429 r
F
.61001 .46406 m .61089 .46477 L .61099 .46436 L .568 .188 .282 r
F
.57751 .45417 m .61089 .46477 L .61001 .46406 L F
.57751 .45417 m .60878 .47302 L p .61089 .46477 L F P
0 g
s
.54574 .3988 m .58571 .45343 L .55968 .41448 L .622 .315 .429 r
F
.56204 .42822 m .57084 .43201 L .61089 .46477 L .631 .248 .299 r
F
.61068 .46556 m .61089 .46477 L .62963 .48011 L F
.56204 .42822 m .61089 .46477 L .61068 .46556 L F
.57751 .45417 m .56945 .4416 L p .574 .44958 L .699 .333 .35 r
F P
0 g
s
.56983 .45977 m .56204 .42822 L .61068 .46556 L .631 .248 .299 r
F
.60878 .47302 m .61068 .46556 L .62963 .48011 L F
.56983 .45977 m .61068 .46556 L .60878 .47302 L F
.61266 .45782 m .58261 .43572 L p .61757 .45888 L .58 .289 .439 r
F P
0 g
s
.56011 .40845 m .58888 .44641 L .58509 .43071 L .622 .315 .429 r
F
.5871 .43902 m .61266 .45782 L p .576 .43424 L .631 .248 .299 r
F P
0 g
s
.576 .43424 m .61266 .45782 L .61233 .45912 L .631 .248 .299 r
F
.59391 .43997 m .58617 .4352 L .58558 .43275 L .58 .289 .439 r
F
.55173 .40099 m .58012 .43601 L .56011 .40845 L .622 .315 .429 r
F
.58012 .43601 m .58143 .43657 L .56011 .40845 L F
.57084 .43201 m .576 .43424 L .61233 .45912 L .631 .248 .299 r
F
.57084 .43201 m .61233 .45912 L .61221 .4596 L F
.66902 .47562 m .67903 .47707 L p .67917 .47711 L .527 .173 .309 r
F P
0 g
s
.61175 .46141 m .68863 .48332 L .67153 .47404 L .568 .188 .282 r
F
.61133 .46304 m .68863 .48332 L .61175 .46141 L F
.61133 .46304 m .66785 .48165 L .65326 .47404 L F
.61099 .46436 m .66785 .48165 L .61133 .46304 L F
.64311 .4802 m .62963 .48011 L p .61089 .46477 L .631 .248 .299 r
F P
0 g
s
.61089 .46477 m .57084 .43201 L .61221 .4596 L .631 .248 .299 r
F
.55061 .40759 m .56856 .43103 L .54574 .3988 L .622 .315 .429 r
F
.56856 .43103 m .56967 .43151 L .54574 .3988 L F
.52116 .33756 m .52257 .36237 L .52128 .33813 L .724 .426 .472 r
F
.52257 .36237 m .52261 .36242 L .52128 .33813 L F
.52125 .33777 m .52114 .33692 L p .52126 .33777 L .911 .651 .532 r
F P
0 g
s
.52116 .33756 m .52128 .33813 L .52124 .33743 L .724 .426 .472 r
F
.52108 .33677 m .52109 .33702 L .52113 .33696 L F
.63824 .48387 m .68746 .49007 L .70691 .49404 L .436 0 0 r
F
.67153 .47404 m .64718 .46532 L p .6172 .4588 L .568 .188 .282 r
F P
0 g
s
.65326 .47404 m .69164 .48889 L .68863 .48332 L .568 .188 .282 r
F
.66785 .48165 m .69164 .48889 L .65326 .47404 L F
.64311 .4802 m .65599 .48294 L p .64225 .48086 L .436 0 0 r
F P
0 g
s
.61099 .46436 m .62377 .48205 L .69164 .48889 L .568 .188 .282 r
F
.60878 .47302 m .62377 .48205 L p .61099 .46436 L F P
0 g
s
.68863 .48332 m .70518 .48766 L .687 .4803 L .568 .188 .282 r
F
.71922 .4935 m .69263 .49073 L .68815 .48244 L .527 .173 .309 r
F
.70738 .49386 m .69263 .49073 L p .71922 .4935 L F P
0 g
s
.68746 .49007 m .70738 .49386 L .70691 .49404 L .436 0 0 r
F
.69005 .4904 m .70738 .49386 L .68746 .49007 L F
.69263 .49073 m .70738 .49386 L p .69005 .4904 L F P
0 g
s
.70518 .48766 m .68699 .48027 L .687 .4803 L .568 .188 .282 r
F
.70518 .48766 m .68653 .47942 L .68699 .48027 L F
.70747 .48932 m .67917 .47711 L .71975 .48985 L .527 .173 .309 r
F
.6724 .47436 m .67153 .47404 L p .687 .4803 L .568 .188 .282 r
F P
0 g
s
.68699 .48027 m .6724 .47436 L .687 .4803 L .568 .188 .282 r
F
.71975 .48985 m .68653 .47942 L .7785 .49169 L .527 .173 .309 r
F
.68653 .47942 m .68581 .47809 L .7785 .49169 L F
.71975 .48985 m .7785 .49169 L .76912 .49198 L F
.77572 .49199 m .76912 .49198 L p .77506 .492 L .27 0 0 r
F P
0 g
s
.77506 .492 m .76912 .49198 L .76873 .49199 L .27 0 0 r
F
.68863 .48332 m .687 .4803 L .67153 .47404 L .568 .188 .282 r
F
.70747 .48932 m .68815 .48244 L .68713 .48055 L .527 .173 .309 r
F
.59775 .44233 m .59391 .43997 L .58558 .43275 L .58 .289 .439 r
F
.59775 .44233 m .58558 .43275 L .58541 .43204 L F
.54306 .38752 m .54079 .38326 L .54818 .39257 L .6 .339 .493 r
F
.52439 .34869 m .52434 .34858 L .52118 .3368 L .415 .257 .557 r
F
.52116 .33678 m .5238 .34789 L p .53012 .3608 L .642 .374 .498 r
F P
0 g
s
.57858 .42234 m .56 .40418 L p .57053 .41594 L .622 .315 .429 r
F P
0 g
s
.5238 .34789 m .52116 .33678 L p .52365 .34751 L .724 .426 .472 r
F P
0 g
s
.52124 .33743 m .52327 .34656 L .52365 .34751 L .724 .426 .472 r
F
.53255 .3672 m .53089 .36273 L p .53255 .36719 L .891 .481 .065 r
F P
0 g
s
.53255 .36719 m .53089 .36273 L .53088 .36273 L .891 .481 .065 r
F
.53211 .3661 m .53084 .36267 L p .53204 .36596 L .642 .374 .498 r
F P
0 g
s
.53204 .36596 m .53084 .36267 L .53084 .36269 L .642 .374 .498 r
F
.53088 .36272 m .53255 .36719 L .53088 .36273 L .891 .481 .065 r
F
.53088 .36272 m .53211 .3661 L .53255 .36719 L F
.53022 .36102 m .53211 .3661 L p .53088 .36272 L F P
0 g
s
.53284 .36829 m .53084 .36267 L .53086 .36231 L .6 .339 .493 r
F
.53054 .36185 m .53086 .36231 L p .53084 .36267 L F P
0 g
s
.53084 .36267 m .53022 .36102 L p .53084 .36269 L .642 .374 .498 r
F P
0 g
s
.53022 .36102 m .52121 .33677 L p .53015 .36088 L .642 .374 .498 r
F P
0 g
s
.53204 .36596 m .53022 .36102 L .53015 .36088 L .642 .374 .498 r
F
.52121 .33677 m .52116 .33678 L p .53012 .3608 L F P
0 g
s
.53012 .3608 m .53204 .36596 L .53015 .36088 L .642 .374 .498 r
F
.53012 .3608 m .53015 .36088 L .52121 .33677 L F
.52124 .33743 m .52164 .34475 L .52411 .35035 L .724 .426 .472 r
F
.5284 .37284 m .52663 .36445 L .54102 .38549 L .673 .365 .443 r
F
.5284 .37284 m .54102 .38549 L .54373 .38945 L F
.54079 .38326 m .53284 .36829 L .53086 .36231 L .6 .339 .493 r
F
.52121 .33677 m .52121 .33677 L p .53027 .36109 L .891 .481 .065 r
F P
0 g
s
.53027 .36109 m .53027 .36108 L .52121 .33677 L .891 .481 .065 r
F
.68653 .47942 m .6724 .47436 L p .68699 .48027 L .568 .188 .282 r
F P
0 g
s
.68627 .47815 m .68581 .47809 L .68578 .47804 L .527 .173 .309 r
F
.7785 .49169 m .68627 .47815 L .78 .49165 L F
.68578 .47804 m .78 .49165 L p .68627 .47815 L F P
0 g
s
.82546 .48788 m .81349 .48995 L p .79264 .49166 L .138 0 0 r
F P
0 g
s
.81349 .48995 m .78706 .49453 L p .79192 .49202 L .138 0 0 r
F P
0 g
s
.81349 .48995 m .79192 .49202 L .79264 .49166 L .138 0 0 r
F
.79192 .49202 m .77572 .49199 L p .7919 .49204 L .27 0 0 r
F P
0 g
s
.79264 .49166 m .83825 .48567 L .82546 .48788 L p .138 0 0 r
F P
0 g
s
.25273 .67341 m .24736 .67388 L p .29061 .66262 L .576 .571 .81 r
F P
0 g
s
.32187 .67997 m .31718 .67881 L .31441 .67793 L .228 .566 .938 r
F
.24807 .67578 m .25247 .67344 L p .24949 .67424 L .435 .233 .505 r
F P
0 g
s
.63905 .73305 m .65681 .73956 L p .65017 .7349 L .927 .88 .768 r
F P
0 g
s
.63664 .72543 m .63905 .73305 L .65017 .7349 L .927 .88 .768 r
F
.66651 .74601 m .66391 .74453 L p .66165 .74295 L 0 .394 .79 r
F P
0 g
s
.65681 .73956 m .6586 .74022 L p .63664 .72543 L .927 .88 .768 r
F P
0 g
s
.6637 .74121 m .6606 .73407 L .6598 .73388 L .005 .313 .804 r
F
.6637 .74121 m .6598 .73388 L .66121 .73764 L F
.66157 .73861 m .66651 .74601 L p .66121 .73764 L F P
0 g
s
.6615 .74128 m .66141 .73894 L .6598 .73388 L p .934 .695 .216 r
F P
0 g
s
.66376 .74264 m .66123 .74103 L .6606 .73407 L .9 .845 .781 r
F
.6642 .74384 m .66212 .74187 L .66376 .74264 L F
.66212 .74187 m .66123 .74103 L .66376 .74264 L F
.66123 .74103 m .65902 .73894 L .6599 .73623 L F
.66517 .74331 m .6606 .73407 L p .6637 .74121 L .005 .313 .804 r
F P
0 g
s
.76437 .49222 m .77506 .492 L .76873 .49199 L .27 0 0 r
F
.6602 .73327 m .6606 .73407 L p .66123 .74103 L .9 .845 .781 r
F P
0 g
s
.6606 .73407 m .66651 .74601 L p .6642 .74384 L .9 .845 .781 r
F P
0 g
s
.65886 .73247 m .6602 .73327 L p .66123 .74103 L .9 .845 .781 r
F P
0 g
s
.52091 .33688 m .52093 .33687 L p .52094 .33691 L .464 .275 .539 r
F P
0 g
s
.52093 .33687 m .52091 .33688 L p .51438 .36094 L .425 .209 .479 r
F P
0 g
s
.51442 .36084 m .52093 .33687 L p .51438 .36094 L .425 .209 .479 r
F P
0 g
s
.25 Mabswid
.6704 0 m
.92721 .4172 L
s
.92721 .4172 m
1 .95369 L
s
1 .95369 m
.70381 .58915 L
s
.70381 .58915 m
.6704 0 L
s
.08856 .24464 m
0 .80661 L
s
0 .80661 m
.70381 .58915 L
s
.70381 .58915 m
.6704 0 L
s
.6704 0 m
.08856 .24464 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[43]:=",
  ImageSize->{263.875, 287.75},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004700017`P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj168
NQdo_`00?g=cLda<C410Y/C4a000O`X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm000_bLW9`0o000o?`0oOb40<00o_`0oo`1o
001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?clo
Oclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0
Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?goo
Ogoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?kno
Okno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omo
oono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7iYUVoiVI09VI<iVIIYVIc9VIoiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mV
oonI0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c:l70P1F1`00[@L200031`0005D7002[1`80
0`L205D7002X1`<01@L00`07001D1`00YPL200P700<01`00E0L00:<70`0:1`0400L705<7002Q1`80
3@L01@071`L00587002O1`803`L01@071`L00587002L1`<04@L00`071`021`0300L704l7002J1`80
5@L00`071`021`0300L704h7002G1`<05`L00`071`021`0300L704h7002E1`806PL00`071`031`03
00L704d7002C1`8070L00`071`031`0300L704d7002@1`<07PL00`071`041`0300L704`7002>1`80
8@L00`071`051`0300L704/7002;1`<08`L00`071`051`0300L704/700291`809PL00`071`061`03
00L704X700271`80:0L00`071`061`0300L704X700241`<0:PL00`071`071`0300L704T700221`80
;@L00`071`081`0300L704P7001o1`<0;`L00`071`081`0300L704P7001m1`80<PL00`071`091`03
00L704L7001k1`80=0L00`071`0:1`0300L704H7001h1`<0=PL00`071`0:1`0300L704H7001f1`80
>@L00`071`0;1`0300L704D7001d1`80>`L00`071`0;1`0300L704D7001a1`<0?PL00`071`0;1`03
00L704@7001_1`80@@L00`071`0<1`0300L704<7001/1`<0@`L00`071`0<1`0300L704<7001Z1`80
APL00`071`0=1`0300L70487001X1`80B0L00`071`0=1`0300L70487001U1`<0BPL00`071`0>1`03
00L70447001S1`80C@L00`071`0?1`0300L70407001P1`<0C`L00`071`0?1`0300L70407001N1`80
DPL00`071`0@1`0300L703l7001L1`80E0L00`071`0A1`0300L703h7001I1`<0EPL00`071`0A1`03
00L703h7001G1`80F@L00`071`0B1`0300L703d7001D1`<0F`L00`071`0B1`0300L703d7001B1`80
GPL00`071`0C1`0300L703`7001@1`80H0L00`071`0D1`0300L703/7001=1`<0HPL00`071`0D1`03
00L703/7001;1`80I@L00`071`0E1`0300L703X700191`80I`L00`071`0E1`0300L703X700161`<0
JPL00`071`0E1`0300L703T700141`80K@L00`071`0F1`0300L703P700111`<0K`L00`071`0F1`03
00L703P7000o1`80LPL00`071`0G1`0300L703L7000m1`80M0L00`071`0H1`0300L703H7000j1`<0
MPL00`071`0H1`0300L703H7000h1`80N@L00`071`0I1`0300L703D7000e1`<0N`L00`071`0I1`03
00L703D7000c1`80OPL00`071`0J1`0300L703@7000a1`80P0L00`071`0K1`0300L703<7000^1`<0
PPL00`071`0K1`0300L703<7000/1`80Q@L00`071`0L1`0300L70387000Y1`<0Q`L00`071`0L1`03
00L70387000W1`80RPL00`071`0M1`0300L70347000U1`80S0L00`071`0N1`0300L70307000R1`<0
SPL00`071`0N1`0300L70307000P1`80T@L00`071`0O1`0300L702l7000M1`<0U0L00`071`0N1`03
00L702l7000K1`80U`L00`071`0O1`0300L702h7000I1`80V@L00`071`0P1`0300L702d7000G1`80
V`L00`071`0P1`0300L702d7000G1`80V`L00`071`0Q1`0300L702`7000G1`0300L009X700<01`L0
8PL00`071`0[1`005`L010071`2I1`0300L7028700<01`L0:`L001H700<01`L00PL00`071`2F1`03
00L702<700<01`L0:PL001H700<01`L00`L00`071`2E1`0300L702<700<01`L0:PL001H700<01`L0
10L00`071`2D1`0300L702@700<01`L0:@L001H700<01`L01@L00`071`2C1`0300L702D700<01`L0
:0L001H700<01`L01PL00`071`2B1`0300L702D700<01`L0:0L001H700<01`L01`L00`071`2A1`03
00L702H700<01`L09`L001D700<01`L02@L00`071`2@1`0300L702H700<01`L09`L001D700<01`L0
2PL00`071`2?1`0300L702L700<01`L09PL001D700<01`L02`L00`071`2>1`0300L702P700<01`L0
9@L001D700<01`L030L00`071`2=1`0300L702P700<01`L09@L001D700<01`L03@L00`071`2=1`03
00L702P700<01`L090L001D700<01`L03PL00`071`2<1`0300L702T700<01`L08`L001D700<01`L0
3`L00`071`2;1`0300L702T700<01`L08`L001@700<01`L04@L00`071`2:1`0300L702X700<01`L0
8PL001@700<01`L04PL00`071`291`0300L702X700<01`L08PL001@700<01`L04`L00`071`281`03
00L702/700<01`L08@L001@700<01`L050L00`071`271`0300L702`700<01`L080L001@700<01`L0
5@L00`071`261`0300L702`700<01`L080L001@700<01`L05PL00`071`251`0300L702d700<01`L0
7`L001<700<01`L05`L00`071`251`0300L702d700<01`L07`L001<700<01`L060L00`071`1G1`80
:`L00`071`0^1`0300L701h7000C1`0300L701T700<01`L0EPL202/700<01`L0;`L00`071`0M1`00
4`L00`071`0J1`0300L705@7100Z1`0300L702l700<01`L07@L001<700<01`L06`L00`071`1C1`@0
:PL00`071`0`1`0300L701`7000C1`0300L701`700<01`L0DPL402X700<01`L0<@L00`071`0K1`00
4PL00`071`0N1`0300L70547100Z1`0300L7034700<01`L06`L0018700<01`L07`L00`071`1?1`<0
00>;0000:@L00`071`0b1`0300L701X7000B1`0300L7020700<01`L0CPL20004RX/002X700<01`L0
<@L00`071`0J1`004PL00`071`0Q1`0300L704d70`0018Z;000Y1`0300L7038700<01`L06@L00187
00<01`L08PL00`071`1;1`8000JR08]/R`0Y1`0300L703<700<01`L060L0018700<01`L08`L00`07
1`1:1`8000F:08Z;QP0202P700<01`L0<`L00`071`0H1`004PL00`071`0T1`0300L704T710001X/0
Rfb602L700<01`L0=0L00`071`0G1`004@L00`071`0V1`0300L704L70P002J80R`2;QX^6000W1`03
00L703@700<01`L05`L0014700<01`L09`L00`071`161`8000R:08D0Q@1/R`809PL00`071`0e1`03
00L701H7000A1`0300L702P700<01`L0A0L0302WRZ80R`2;08^6R`809@L00`071`0f1`0300L701D7
000A1`0300L702T700<01`L06@L00`071`0V1`0=0:::XXF:Q@2506b;K00302@700<01`L0=PL00`07
1`0E1`004@L00`071`0Z1`0300L701P70`0U1`0A0:NQXP2RQJ80RhF;QX^6R`0090L00`071`0g1`03
00L701@7000A1`0300L702/700<01`L060L302<70P004HZR08Z5RXD0K8E/QF`0K8H002<700<01`L0
>0L00`071`0C1`0040L00`071`0]1`0300L701L71@0P1`0E0:NQYj40QJ:5XXD0QH^5R`2;KH/00287
00<01`L0>0L00`071`0C1`0040L00`071`0^1`0300L701L700=K0:P00P0N1`0G0:::XXX0QHZ5RXF:
08E/QFb5K8]/QP008@L00`071`0i1`0300L70187000@1`0300L702l700<01`L05PL30004Z00001X7
0P006:NQYj6W0::5XXFRQ@25RhF;QH]]06f;020700<01`L0>@L00`071`0B1`0040L00`071`0`1`03
00L701H70`000jRAT@0201H70P006hZR0:65X@2:QHZ5RXF:08E/QFb5K01/QP26000O1`0300L703T7
0P0C1`0040L00`071`0a1`0300L701D700P0FanAZ940T@@040L2000GYj6W0:65XHFQQJ:5XXFRQ@25
RhF;QH/00P001Fd0KH/001h700<01`L0=PL300031`070187000@1`0300L7038700<01`L050L0201K
FP2AZ96X0Y4600/700@0RZ::0P007:65XHFQQJ65RXF:QHX0QFb5K8E/Q@26K8I/QP0M1`0300L703<7
0`041`0300L70147000@1`0300L703<700<01`L050L0301K7jRAZ96X0:VAZ@l000nQRP00XHFQQJ65
XHFQQJ801HD03X^5RhF;QH/0Rff;KH/07@L00`071`0_1`<01`L00`071`0A1`003`L00`071`0e1`03
00L701<700T0FU/0Z96XTJP01I42000_XHBQQ:64XHBQQ:64XHFQQJ65XHFQQJ6508B5Q8F4QHB5K8E/
QFb5K8E/QVa]K0006`L00`071`0/1`<02`L00`071`0@1`003`L00`071`0f1`0300L701<702P0Fal0
Z96XTJRAZ96YTJT0XHZQRZ6:XHZQRZ6:XHFQQJ65XHFQQJ402XD038]/QVb6K8I/Rff6K@806@L00`07
1`0Y1`<03PL00`071`0@1`003`L00`071`0g1`0300L7018700`0FeXOTJRAZ96XTJP5T@<002aVXHBQ
Q:64XHBQQJ65XHFQQJ6508B5Q8F4QHB5Q8D0QFb5K8E/QFb506e/K@805`L00`071`0V1`<04@L00`07
1`0@1`003`L00`071`0h1`0300L7018701<0F`00TJRAZ96XTJRAZI6YTJVA00800164RZ6:XHZQ0:65
XHFQQJ65000<Q@0:08I/QVb6K8I/QP<000B6K@005@L00`071`0S1`<050L00`071`0@1`003`L00`07
1`0i1`0300L7014700/0EE/0Z96XTJRAZ00:T@8002f4XH@0IXAVQ6J4IXAVQ025Q8F4QHB5Q8F4QHB5
K8E/QFb5K8E/QF`0K6e/K@0050L00`071`0P1`<05`L00`071`0@1`003`L00`071`0j1`0300L70107
01L0FeYK0:RAZ96XTJRAZI6YTJVAZI6YT@02000?Q6J4OhAoQ7n4OhAoQ7n400^50140K8I/QVb6K8I/
QV`0KHI]QP02018700<01`L07@L301X700<01`L040L000h700<01`L0?0L00`071`0?1`0=05EKEE/0
Z96XTJRAZ00<T@8000eVOfJ4IXAVQ6J4IXAV00:401b5Q8F4QHB5Q8F4QFb5K8E/QFb5K8E/Q@1]K6e/
0P0@1`0300L701X70`0M1`0300L70107000>1`0300L703d700<01`L03PL05`1KFU]J096XTJRAZ96X
TJVAZI6YTJVA00P000b4OhAoQ7n4OhAoQ7l9Q@0DIXI/QVb6K8I/QVb6K8H0QVf6KHH200h700<01`L0
5`L3020700<01`L040L000h700<01`L0?PL00`071`0=1`8001UKEE]E096XTJRAZ96X096<THbAS96<
THbA00<000>;00000P008fJ4IXAVQ6J4IXF4QHAoIWmVOfJ5K8E/QFb5K8E/QFM/IfaW009/00A]K000
30L00`071`0D1`<090L00`071`0?1`003PL00`071`0o1`0300L700`701l0FeYKFU^WZ96XTJRAZ940
TJVAZI6YTJVAZI6YTJVA008000>BT@0010008gn4OhAo06J4IXAVQ6J4IXAVQ6b6K8I/QVb5K8E/QFb5
K8E/00<02PL00`071`0A1`<09`L00`071`0?1`003PL00`071`101`0300L700/70P009U]EFeEKRjRA
Z96XTJR<THbAS96<THbAS96<THbAS96<TXbBS9:<10008H@0OfIoIWmVOfIoIWmVOfIoIXE/QFM/IfaW
K6M/IfaWK00300P700<01`L03PL302X700<01`L03`L000h700<01`L0@@L00`071`0:1`0/07UdFeYK
FZRWZ96XTJRAZ96XTJVAZI6YTJVAZI6YTJVAZI6BT@2ATY6BTI85000OQ6J4IXAVQ6J4IXAVQ6J4IP1/
QFb5K8E]QFb5K8E/Q@0200H700<01`L02`L302d700<01`L03`L000h700<01`L0@PL00`071`091`0a
07=iEE]EFeFXRjRAZ8bAS96<THbAS96<THbAS96<THbAS96<THbBS9:<TXbBS9:<TP05000LOfIoIWmV
OfIoIWmV06M/IfaWK6M/IfeWK6M/I`L02PL3030700<01`L03`L000d700<01`L0A0L00`071`081`0f
07UdFeYKFU^WZ02XTJRAZ96XTJRAZI6YTJVAZI40TJVAZI6YTJVATY6BTI:ATY6BTI:ATY6B10006VJ4
IXAVQ6J4IP1/QFb5K8E/QFb5K6e/QFb50P000fM]000600041`0003<700<01`L03`L000d700<01`L0
A@L00`071`071`0j07<0Le]EFeEKRjR<THbAS96<THbAS96<THbAS96<08bAS96<THbAS9:<TXbBS9:<
TXbBS9:<TXbBS0D0021VOfIoIP1WK6M/IfaWK6M/IfaWK6M/IfaW06M/IfaWK0H05@L401P700<01`L0
3`L000d700<01`L0APL00`071`061`0o07T0NEYKFU/OZ:L0TJRAZ96XTJRAZ96YTJVAZI6YT@2AZI6Y
TJVAZI6YTI:ATY6BTI:ATY6BTI:ATY6BTI:A00@03FH00faV0002000@KFM]08EWKFM]IfeWKFM]I`L0
20L901X700<01`L03PL000d700<01`L0A`L00`071`051`1307=iLe]EFeEK7jR;08bAS96<THbAS96<
THbAS96<THbAS96<THbAS96<THbBS9:<TXbBS9:<TXbBS9:<TXbBS9:<TP04000?CFI=ITeVCFI=ITeV
CFI=00<0019WK000IfaWK6M/IfaWK6M/If`@01`700<01`L03PL000d700<01`L0B0L00`071`031`80
01miM7UdFeYKFQl0Z02XTJRAZ96XTJRAZ96YTJVAZI6Y00>A01^YTJVAZI6YTJVATY6BTI:ATY6BTI:A
TY6BTI807@000feWK@0E0003T@2A00<07PL00`071`0>1`003@L00`071`181`0300L700<706T0NG=i
LgUEFeEKEAn;Z02AS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<TXbBS9:<THbAS96<THbA
S96<THbAS96<THbAS95cTG>ALi5cTG>ALi40TG>ALi5cTG<0Li5cTG<01@001g><LhacS7<00`0P1`03
00L700h7000<1`0300L704X700<01`L00PL08`1dNG@0M5]JFeYK7`2W096XTJRAZ96XTJRAZ96YTJVA
ZI6Y00>A036YTJVAZI6YTJVAZI6BTI:ATY6BTI:ATY6BTI:ATY6BTI:ATY6BTI:ATY6BTI:ATY6B00H0
00FATY6BT@020003TY6B00:A00f<THbAS96<THbA0940028700<01`L03PL000`700<01`L0B`L03`07
1`L0NG=iLgUcFeEKE@027`1=Z8/0S96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<
THbAS96<THbAS96<THbAS96<THbAS96<THbALi5cTG>ALi400P005W>ALi5cTG>ALhacS7><LhacS7><
Lh`202@700<01`L03PL000`700<01`L0C0L0:0071`1dNGAiM7UJFeYKFQnWZ02XTJRAZ96XTJRAZ96X
TJVAZI6YTJT3T@03ZI6Y00:A03:<THbAS96<TY6BTI:ATY6BTI:ATY6BTI:ATY6BTI:ATY6BTI:ATY6B
TI:ATY40TI:ATP:A00n<THbAS96<THbAS96<TH`00`0U1`0300L700h7000<1`0300L704d707<01`1i
LgUcNG<0EE]EFeDORjR<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbA
S96<THbAS96<THbAS96<THbAS96<THbALi5cT@2ALhacS7><LhacS7><LhacS7><Lh`002P700<01`L0
3@L000`700<01`L0CPL2000lM7UdNGAiLe]EFeEK7jR@Z96XTJRAZ96XTJRAZ96XTJVAZI6YTJV<THbA
S96<THbAS96<THbBTI:ATY6B0i40;I:ATY6BTI:ATY6BTI:ATY6BTI:ATY40S96<08bAS96<THbAS96<
THbAS96<T@0302T700<01`L03@L000`700<01`L0C`L0EP1iLgUcM5YdEEYEFUEJ7jP0Z8bAS96<THbA
S96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbA
0X`05g><08acS7><LhacS7><LhacS7><LX`002X700<01`L03@L000`700<01`L0C`L0>01dNGAiLgUc
FeEKEE]E7i2XT:RAZ96XTJRAZ96XTJRAZ02AS96<THbAS96<THbAS96<THbAS96<0Y40;i:ATI6BTI:A
TY6BTI:ATY6BTI:ATY6BTI:<THbAS96<08bAS96<THbAS96<THbA00:<00=cS000:`L00`071`0=1`00
2`L00`071`1@1`1@07UcM5YdFWAEFUEJEEYE7h^XRi6<THbAS96<THbAS96<08bAS96<THbAS96<THbA
S96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS94:S00A08acS7><LhacS7><LhabS7800P0/1`03
00L700d7000;1`0300L705070P00>7UcNG=iLgUEFeEKEE]EZ900T:RAZ96XTJRAZ96XTJP0THbAS96<
THbAS96<THbAS96<THbAS96<0Y40;i:ATY6BTI:ATY6BTI:ATY6BS96<THbAS96<THbAS02<THbAS96<
THbALhacS7<002h700<01`L03@L000/700<01`L0D0L20006FWAJM5Yd0UX0@5EJEEYEFQnXRjR;THbA
S96<THbAS96<08bAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS94@S00@08acS7><
LhacS7:<LXab02l700<01`L03@L000/700<01`L0D0L2001WNG=iLgUcNEEKEE]EFeDhT:RAZ92XTJRA
Z96XTJRAZ8bAS96<THbAS96<THbAS96<THbAS96<THbAS96<TY6BTI:ATY6BS96<THbAS96<THbAS96<
THbAS02<THbAS95cS7><Lhac000`1`0300L700d7000;1`0300L7050700P0M5YdFWAJM09J03YEFUEJ
EEYE08^XRjR;THbAS96<THbARi6;THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbA5X`03P2<
LhacS7:<LXabS780<PL00`071`0<1`002`L00`071`1?1`80011cNG=iLgUcNEEKEE]EFeEK0P00DY2X
T:RAZ96XTJRAZ8bXS:R<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbA
S96<THbAS96<08`0LhacS7><Lh`203<700<01`L030L000X700<01`L0D0L0@WT0M5YdFWAJM5YdEEYE
FUEJECP0Z8^ARjR<THbARi6;TH^ARi6;THbAS96<THbAS96<THbAS96<THbAS96<THbA01V<0P002G:<
LXabS7:<000e1`0300L700`7000:1`0300L704l70P00FWUcNG=iLgUcNG=KEE]EFeEK7`2@Z92XT:RA
Z96XS:R<Z8bXS:R<THbAS96<THbAS96<THbAS96<THbAS96<T@2AS96<THbAS96<THbAS96<THbAS96<
THbAS08000R<LhacS7><03H700<01`L030L000X700<01`L0CPL20011M5YdFWAJM5YdFWAEFUEJEEYE
FQl0RjR<Z8^XS96;TH^ARi6;TH^ARi6<THbAS96<THbAS96<THbAS96<THbAS0006X`20007LXabS7:<
000g1`0300L700`7000:1`0300L704h706D0Ug=iLgUcNG=iLgUcFeEKEE]EFeDh0:R@Z92XT02<Z8bX
S:R<Z8bXS:R<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<T@2<
08acS7><000h1`0300L700`7000:1`0300L704d701L0NG=iFWAJM5YdFWAJM5EJEEYEFUEJ7`02000/
Z8^XRjP0TH^ARi6;TH^ARi6;TH^AS96<THbAS96<THbAS96<THbARi6;TH/GS007LXab07:<LP0203T7
00<01`L030L000X700<01`L0C@L0F@2GTW=iLgUcNG=iLgUcNEEKEE]EFeEK7`2@Z900T02<Z8bXS:R<
Z8bXS:R<Z8bAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbA00B<00AcS02<
0P0l1`0300L700/700091`0300L704d700l0LgUcM5YdFWAJM5YdFW@00UX0=eEJEEYEFUDh0:R;Z8^X
R`2;TH^ARi6;TH^ARi6;TH^AS96<THbAS96<THbARi6;TH^ARi6;TH/04H`01g:<LXabS7800`0m1`03
00L700/700091`0300L704`705L0Li9cTW=iLgUcNG=iLgUcNEEKEE]EFeEK7cR@Z92XT:P0Z8bXS:R<
Z8bXS:R<Z8bXS:R<THbAS96<THbAS96<THbAS96<THbAS96<THbAS96<THbAS94028`017><000n1`03
00L700/700091`0300L704`70100NG=iLgAJM5YdFWAJM5Yd0UX0>5EJEEYEFUDh7`2;Z8^XR`2;TH^A
Ri6;TH^ARi6;TH^ARi6<THbAS96;TH^ARi6;TH^ARi6;TH^A38`037:<LXabS7:<08ab03l700<01`L0
2`L000T700<01`L0B`L0E02BLi9cTW=iLgUcNG=iLgUcN@1KEE]EFeEKECP0Z92XT:P0Z8bXS:R<Z8bX
S:R<Z8bXS:R<Z8bAS96<THbAS96<THbAS96<THbAS96<THbAS96<T@Z<00<0S7<00P0o1`0300L700/7
00091`0300L704X70P00Bg>GLgUcNEYdFWAJM5YdEG=EL`1JEEYEFUEJ7cR;Z8^XRjP0TH^ARi6;TH^A
Ri6;TH^ARi6;08^ARi6;TH^ARi6;TH^ARi6;TH^AR`07S00=LXabS7:<LXabS02<LP0303l700<01`L0
2`L000T700<01`L0BPL0C`1cTW>BLi9cNG=iLgUcM7=dLgAcM01KEE]EFeDh7`2@Z92XT02<Z8bXS:R<
Z8bXS:R<Z8bXS:P0THbAS96<THbAS96<THbAS96<THbAS9403X`00`070002008700<01`L0?0L00`07
1`0;1`002@L00`071`191`0E07=iTGUcNG=iFWAJM5EcEG=ELeEc00ME041J03PO08^XRjR;TH^ARi6;
TH^ARi6;TH^ARi6;08^ARi6;TH^ARi6;TH^ARi6;TH^ARi40S7:<LXabS7:<LXab1P041`0300L703/7
00<01`L02`L000P700<01`L0B@L2001?TYNBLi9cTW=dLgAcM7=dLgAcM7=dEEYEFUEJE@0O>02XT:R@
Z92XS:R<Z8bXS:R<Z8bXS:R<Z02AS96<THbAS96<THbAS96<THbAS96<0009S0@000D700NG00041`03
00L703`700<01`L02PL000P700<01`L0B@L0201iLgUcNG=i0W<035EcEG=ELeEcEG=EL`UE01T07`2;
Z8^XRjR;TH^ARi6;TH^ARi6;TH^A008001f;TH^ARi6;TH^ARi6;TH^ARi6;TH/0LXabS7:<LP030003
LX`0008000>GT@001PL00`071`0k1`0300L700X700081`0300L704P703T0TW<0Li9cM7=dLgAcM7=d
LgAcM7=dLgAEFW=JEEYEFP0h7jR@Z92XT:R<Z8bXS:R<Z8bXS:R<Z8`00P005Y6<THbAS96<THbAS96<
THbAS96<T@04S0<01H`00`2A1`02U`0300L700D700<01`L0>PL00`071`0:1`0020L00`071`181`05
07=iLgT01W<03UEcEG=ELeEcEG=ELeF<25D0<00h0:R;Z8^XRjR;TH^ARi6;TH^ARi6;T@2ARi6;TH^A
Ri6;TH^ARi6;TH^ARi40S0@000ebS7:<LXab096GTIL000T700<01`L0>@L00`071`0:1`0020L00`07
1`171`1?07>BLgAcM7=dLgAcM7=dLgAcM7=dLgAcM5EJEEYEFUEJE@0O0:NXT:R@Z92XS:R<Z8bXS:R<
Z8bXS:R<THbAS96<THbAS96<THbAS96<T@0300Z<00<0TIL00iL00`071`081`0300L703P700<01`L0
2PL000P700<01`L0APL00`1cN@0:L`0<EG=ELeEcEG=ELeEc2eD0:CPO08^XRjR;Z8^XRi6;TH^ARi6;
TH^ARi6;TH^ARi6;TH^ARi6;TH^A00@0019bS7:<LXabS7:<LX`0Ui6GT@0;1`0300L703P700<01`L0
2PL000L700<01`L0A`L20018LgAcM7=dLgAcM7=dLgAcM7=dLgAcM7=dEGAEFUEJEEYEFP0h0:R@Z92X
T:R@Z8bXS:R<Z8bXS:R<Z8bAS96<THbAS96<THbA0`000i6<T@0<S00309NG00>G00<01`L02PL00`07
1`0g1`0300L700X700071`0300L704H70P0<L`0<EG=ELeEcEG=ELeEc35D08Al0002;Z8^XRjR;TH^A
Ri6;TH^ARi6;TH^ARi6;TH^AR`04000HTH^ARi6;S7:<LXabS7:<LX`0Ui6GTIL03PL00`071`0g1`03
00L700T700071`0300L704D70P009g=dLgAcM7=dLgAcM7=dLgAcM7=dLgAcM7=dEGAEFUEJEEYEFUDn
7`02000JZ92XT:R@Z92XS:R<Z8bXS:R<Z8bXS96<TH`3000:S96<THbAS96<T@Z<00=bUiL00iL00`07
1`0>1`0300L703H700<01`L02@L000L700<01`L0A0L2011c00YELeEcEG=ELeEc3ED06SP008^XRjR;
Z8^XRi6;TH^ARi6;TH^ARi6;0`007X^ARi6;TH^ARhb;S8^<LXabS7:<LXa]096GTINA014700<01`L0
=@L00`071`091`001`L00`071`131`0[L`1cM7=dLgAcM7=dLgAcM7=dLgAcM7=dLgAcM7=d07AEFUEJ
EEYEFUEJ7`03000CT:R@Z92XT:R<Z8bXS:R<Z8bXS003000@S96<THbAS96<THbAS96<T@R<00=b09H0
19L00`071`0A1`0300L703@700<01`L02@L000L700<01`L0@@L00`1cL`0BL`0<EG=ELeEcEG=ELeD0
2eD04cPO002XRjR;Z8^XRi40TH^ARi4010008h^ARi6;TH^ARi6;S8^<Rhb;S8^<LXabS7:<K@2A1i6G
09L001<700<01`L0=0L00`071`091`001`L00`071`101`0_07>BLgAcM7=dLgAcM7=dLgAcM7=d07Ac
M7=dLgAcM7=dKWAEFUEJEEYEFUEJ7ch00P00292XT:R@Z92X0P000hbX0002000GZ8bXS:R<Z8bAS96<
THbAS96<THbAS9401X`01W:<00NG1`:G00<01`L04`L00`071`0c1`0300L700T700061`0300L70407
00<0Lg<03W<01VicKW=^L`8000aELeEcEG=ELeEcE@0:E@0;7`0008^XRjR;Z8/01000:I6;TH^ARi6;
TH^ARi6;S8^<Rhb;S8^<Rhb;S7:<KHa]S007T@NAUi4001H700<01`L0<PL00`071`091`001PL00`07
1`0n1`8000acTW>BLgAcM7=dLg@>L`0GKW=^LficKW=^Lfh0EEYEFUEJEEYE?Ql00P001JR@Z92X00@0
01b<Z8bXS:R<Z8bXS:R<Z8bAS96<THbAS96<THbA0X`037:<LX`01iH7UPNF01P700<01`L0<@L00`07
1`091`001PL00`071`0m1`0307=c00Ec0280LficKW=^LficKW=^LficKW=^LeAcE7=DLeAcE7=DLe@0
25D01Alh0:N;00@002jXRjP0TH^ARi6;TH^ARi6;TH^A08b;S8^<Rhb;S8^<Rha]S6f<K@2A1i47TIL0
6@L00`071`0a1`0300L700P700061`0300L703`700X0TW>BLi9cS7><57<05ficKW=^LficKW=^Lfh0
EEYEFUEJEChO00<0036@Z92XT:R@08bXS:R<Z8bXS:R<Z8bXS02<THbAS96<THbAS95bS7:<LP2F09H7
UPL001X700<01`L0<@L00`071`081`001PL00`071`0j1`800g<0:VecKG=]LfecKG=^LficKW=^Lfic
KW=^LficE7=DLeAcE7=DLeAcE7=D00EE0`00=1l00:R;Z8^XRjP0TH^ARi6;TH^ARi6;TH/0Rhb;S8^<
Rhb;S8^<Rha]S6d0T@NA1i47T@0K1`0300L7030700<01`L020L000H700<01`L0>@L01`1cTW><Lh`0
0P000g><L`02L`03S7=c011c015^LficKW=^LficKW=^LficE@03000XEChO?P2XT:R@Z92XT02<Z8bX
S:R<Z8bXS:R<Z02AS96<THbAS96<T@:<00]bS780UPNF1iH7UP0201`700<01`L0;`L00`071`081`00
1@L00`071`0i1`8002ecKG=]Lfd0KG=]LfecKG=^LficKW=^LficKW=^LficKW=DLeAcE7=DLeAcE7<0
100045AEE5D0>1l0R`2;Z8^XRjP2000WRi6;TH^ARi6;TH^AR`2;S8^<Rhb;S8^<Rh`0S6f<00NX1i47
T@L001h700<01`L0;PL00`071`081`001@L00`071`121`L000=c00003g<02VicKW=^LficKW<3000<
LficEG=ELeD07chO0P00;ZR@Z92XT:P0Z8bXS:R<Z8bXS:R<Z02AS96<THbAS96<TH^<LX`01iH7UPNF
1iH201h700<01`L0;@L00`071`081`001@L00`071`1;1`0G04<007=^LficKW=^LficKW=DLeAc07<0
0`00?g=DLeAcE5EDEEAEE00O>02WRjR;Z8^XR`2;TH^ARi6;TH^ARi40S8^<Rhb;S8^<Rha]S6f<0:bA
1i47T@NA1`0201h700<01`L0;@L00`071`081`001@L00`071`1?1`@027<0101cKW<3000hL`1cKW=^
LficEG=ELeEc03hO0:NXT:R@Z92X0:R<Z8bXS:R<Z8bXS02<THbAS96<THbA08abS003[007UPNF1iH7
UP0201l700<01`L0;@L00`071`071`001@L00`071`1A1`8000QcKG=]LfecK@@004EcE7=DLeAcE7=D
LeAEE5EDEEAE03PO08^XRjR;Z8/0Ri6;TH^ARi6;TH^A08b;S8^<Rha]S8^<KHa][:R/Z:bA1i47TIL0
0P0P1`0300L702`700<01`L01`L000D700<01`L0D`L20003KG=]00<000=]Lfd00P00<g=^LficKW=^
LeEcEG=ELeEcE@0O?P2XYjR@Z92XRjR<Z8bXS:R<Z8bX096<THbAS96;T@02S004LXab00F/00H7UPNF
1i82024700<01`L0:`L00`071`071`001@L00`071`1E1`<004a]LfecKG=]L`1cE7=DLeAcE7=DEEAE
E5EDEE@07`2;Yh^XRjR;TH^ARi6;TH^ARi6;08^<Rhb;S6f<R`1]S6d0Z:bX[:R/Z0NAUi6G0P0R1`03
00L702X700<01`L01`L000@700<01`L0D`L300870`00=g=]LfecKG=]LfecKW=^LficEG=ELeEcEG<0
?Ql0YjR@Z8^XRjR<Z8bXS:R<Z8bARi6<THbAS9400X`017:<LX`9[0041i87TP808`L00`071`0Y1`03
00L700L700041`0300L705070`081`8004Y]LfecKFiDKUA^E6iDKUAEE5EDEEAEE5D0>02WRjR;TH^A
Ri6;TH^ARi6;S8^<Rhb;S8^<Rha]S6f<Z:bX[:R/Z:bX[02GTINA02<700<01`L0:@L00`071`071`00
10L00`071`1<1`@03@L2000aLfd0KG=]LfecKG=]LeEcEG=ELeEcE@0O0:NXYjR;Z8^XRjR<Z8bXS96;
TH^AS96<T@02S005RhabS0001:`00`2/[003[0061i87TPL090L00`071`0X1`0300L700L700041`03
00L704T70`0C1`8001I]05A^E6iDKUA^E6iDEEAEE5EDEE@h0P0078^W096;TH^ARi6;TH^<Rhb;S8^<
Rhb;S6d0KH`2000@Z:bX[:R/TJbX[96/TINA02D700<01`L0:0L00`071`061`0010L00`071`161`<0
60L0;@1]LfecKG=]LfecKG=ELeEcEG=ELaln0:RW08^XRjR;Z8bXRi6;TH^ARi6<T@02S004Rhb;S080
00>W[:`01J`02JP7Z0NX1i87000V1`0300L702L700<01`L01PL000@700<01`L0@`L301d70P00@UA^
E6iDKUAcE01DEEAEE5ED?Ql0RjL0TH^ARi6;T@2<Rhb;S8^<Rhb;S6f<08a]0:N/YjbX[:R/Z:bA[96/
TINA02L700<01`L09PL00`071`061`000`L00`071`101`@08PL04`1]LfecKG=]Lfd0EG=ELeEc7eD0
0P008JL0RjR;Z8^XR`2;TH^ARi6;TH/0Rhb;S8/0YjbW[:N/Y`030009Z0NX1jP7Z0L002L700<01`L0
9PL00`071`061`000`L00`071`0m1`<09`L04SQD05A^E6iDKUAEE5EDEE@n7`<0022;08^ARi6;T@2<
Rhb;S8^<Rh`0S6f<K@2W[:N/YjbW[0<000V/TJbW[96/T@00:0L00`071`0U1`0300L700H700031`03
00L703X70`0Z1`8002U]LfecKG=]Lfd0EG=ELamE7`2WZ02XRjR;Z8/0Ri6;TH^ARi40S8^<R`020006
[:N/YjbW0P003:bW[:N/Z0L01jP702T700<01`L090L00`071`061`000`L00`071`0g1`<0;0L05007
1`1DKUA^E6iDKUAEE5ED?Qln0P0078/0Ri6;TH^A08b;S8^<Rhb;06f<KH`0[:N/0:4200T70P000jbA
000Z1`0300L702<700<01`L01PL000<700<01`L0<`L402l701@01`L7?VecKG=]Lfe^E7=ELamE7`80
01FX0:R;Z8^XR`2;TH^ARi6;T@2<Rh`00P001JN/Yj`000d7100Z1`0300L7028700<01`L01PL000<7
00<01`L0<0L303<700<01`L00PL2000ZKUA^E6iDKUA^E00O?Ql0R`2;TH^ARi40S8^<Rhb;S02<KHa]
0:N/Yj`03`L00`2/000Z1`0300L702<700<01`L01@L0008700<01`L0;PL303H700<01`L00`L2000=
LfecKFiDKUA^05DOE@03000BZ8^XRjR;08^ARi6;TH/0Rha]0P000j/0000C1`80:PL00`071`0R1`03
00L700D700021`0300L702/70`0i1`0300L700D702D0KUA^E6iDKUA^7chO08/0Ri6;TH^A08b;S8^<
R`1]S6f<0:L0048700<01`L08@L00`071`051`000PL00`071`0W1`@0?0L00`071`061`0<07=]LeA^
E6iD01mE0P005JL0RjR;Z8/0Ri6;TH^A08b;S02[00141`0300L7020700<01`L01@L0008700<01`L0
90L3040700<01`L01PL03@1]KFd0E6iDKP0n7ch00`00496;TH^A08b;S8^<R`1]S6d204H700<01`L0
7`L00`071`051`000PL00`071`0Q1`<0@`L00`071`071`0306ec008001UDKUA^7eDO0:L0RjR;Z02X
Ri6;TH/0Rhb;0080A`L00`071`0O1`0300L700D700021`0300L701h70`161`0300L700P700/0KFd0
KUA^E00O?P03000@TH^AR`2;S8^<Rh`0S6f<04T700<01`L07PL00`071`051`000PL00`071`0J1`@0
B@L00`071`091`0G06d006iDKUAE7eD0Z8^XRjP0Z8^ARi400P000h^<001;1`0300L701h700<01`L0
10L000031`0701T70`1=1`0300L700X701/0K@1DKUA^7chO08^XRi6;S8^<Rhb;06f<CP00C0L00`07
1`0M1`0300L700@700000`L01`0F1`<0D0L00`071`0:1`0K07<005A^E00OEJL0RjR;Z8^XRi6;TH^<
Rh`004d700<01`L070L00`071`041`0000<700L04`L305<700<01`L02`L02@1]06iDKU@n7`02000>
Z8^<Rhb;S8^<LXa]R`1?1`0300L701/700<01`L010L000031`0700l7101F1`0300L700`70P005UA^
E6hOEAl0RjR;Z8^XRi6;S8^<K@1?1`0300L701/700<01`L010L000031`0700`70`1J1`0300L700d7
0P0055A^E00O?X^XRjR;S8^<08abS6d0D@L00`071`0J1`0300L700@700000`L01`091`<0G@L00`07
1`0>1`0607=DKUAE0P003H^XRjR;Z8/0Rhb;S000DPL00`071`0I1`0300L700@700000`071`061`<0
H0L00`071`0>1`0D06e]E6hOE02XRjR;S8^<08abR`1D1`0300L701P700<01`L010L0000300L70087
101S1`0300L700l701<0KFiD04l0RjR;Z8^XR`2;S6d005D700<01`L060L00`071`031`0000<01`00
0P1W1`0300L7010700D0E6d0E0020008RjR;S8/0LX`205H700<01`L060L00`071`031`000P1Z1`03
00L701470100KFh07`2XRjR;Z02<Rh`0E`L00`071`0G1`0300L700<700000`L0001Y1`0300L70147
0P003U@07`2;Z8^<R`1bS6d0F0L00`071`0F1`0300L700<700031`0300L706H700<01`L04PL0301^
05D0Z8/0RjP0S080FPL00`071`0E1`0300L700<700041`0300L706D700<01`L04PL20004K@00R`80
00J<R`1bR`1K1`0300L701@700<01`L00`L000D70P1U1`0300L701<700D0K@00Z0020005RjR;S000
G0L00`071`0D1`0300L700<700071`0300L7068700<01`L050L30006RjP0S02<0P1M1`0300L701@7
00<01`L00PL000P70P1R1`0300L701@700Y]KP00R`2;08/0G`L00`071`0C1`0300L70087000:1`03
00L705l700<01`L05@L30003Z02<00<0H0L00`071`0B1`0300L70087000;1`80GPL00`071`0F1`06
06d0RjR;0`1Q1`0300L7014700<01`L00PL000d700<01`L0F`L00`071`0G1`<000B;06d0HPL00`07
1`0A1`0300L70087000>1`0300L705X700<01`L05`L30004Z02<06<700<01`L040L00`071`021`00
3`L205X700<01`L05`L30004R`1]06@700<01`L03`L00`071`021`004@L00`071`1G1`0300L701P7
00<0XZP00`1U1`0300L700h700<01`L00PL001870P1G1`0300L701P71@1W1`0300L700h700<01`L0
0@L001@700<01`L0E0L00`071`0H1`<000>;00L0I`L00`071`0=1`0300L70047000E1`0300L705<7
00<01`L060L30003R`0706L700<01`L03@L00`071`011`005PL205<700<01`L06@L00e@0001Z1`03
00L700`700<01`L00@L001P700<01`L0D0L00`071`0I1`03E00006/700<01`L02`L00`071`011`00
6@L2050700<01`L06@L306`700<01`L02PL00`071`011`006`L00`071`1=1`0300L701T70`1]1`03
00L700T700<01`L00@L001`70P1=1`0300L701X700<01`L0K0L00`071`0:1`400@L11`007PL00`07
1`1:1`0300L701X700<01`L0K@L00`071`091`400@L11`007`L00`071`191`0300L708/700<01`L0
20L100470@L002070P191`0300L708`700<01`L01`L100470@L0028700<01`L0APL00`071`2=1`03
00L700H70@011`47000S1`80APL00`071`2=1`0300L700H70@011`47000U1`0300L704<700<01`L0
SPL00`071`051`400@L11`009PL204<700<01`L0S`L00`071`051`400@L002P700<01`L0@0L00`07
1`2@1`0300L700@70@011`00:@L00`071`0o1`0300L7094700<01`L00`L10047000Z1`80?`L00`07
1`2B1`0300L700870@011`00;0L00`071`0l1`0300L7098700<01`L00PL10047000]1`80?0L00`07
1`2C1`0500L71`000@L002l700<01`L0>@L00`071`2D1`0400L70047000`1`0300L703P700<01`L0
U@L00`0700011`00<@L203L700<01`L0U`L100470@0003<700<01`L0=0L00`071`2G1`400@L10000
=0L203@700<01`L0V0L20000=PL00`071`0a1`0300L709L70`0003L70P0a1`0300L709<710031`00
>@L00`071`0^1`0300L708l710071`00>PL00`071`0]1`0300L708X71@0;1`00>`L202d700<01`L0
QPL40107000m1`0300L702X700<01`L0PPL401@7000n1`80:PL00`071`1m1`D060L0040700<01`L0
9`L00`071`1i1`@07@L0044700<01`L09PL00`071`1d1`D08@L004870P0V1`0300L70707100V1`00
A0L00`071`0S1`0300L706`7100Z1`00A@L202<700<01`L0I`L502h700171`0300L7020700<01`L0
H`L403<700181`8080L00`071`1O1`@0=`L004X700<01`L07@L00`071`1J1`D0>`L004/700<01`L0
70L00`071`1F1`@0@0L004`70P0L1`0300L7058710141`00CPL00`071`0I1`0300L704d71@181`00
C`L201T700<01`L0B@L404d7001A1`0300L701H700<01`L0A0L50547001B1`805PL00`071`101`@0
EPL005@700<01`L04`L00`071`0l1`@0FPL005D700<01`L04PL00`071`0g1`D0GPL005H70P0A1`03
00L703@7101S1`00F0L00`071`0>1`0300L70307101W1`00F@L200h700<01`L0:`L506/7001K1`03
00L700/700<01`L09`L40707001L1`0300L700X700<01`L08`L407@7001M1`802PL00`071`0N1`D0
N0L005l700<01`L01`L00`071`0J1`@0O@L006070P071`0300L701D71@211`00HPL00`071`041`03
00L7014710261`00H`L200@700<01`L03@L408X7001U1`0500L71`002PL508h7001V1`0400L700H7
102C1`00I`L30087102G1`00J@L309/70000\
\>"],
  ImageRangeCache->{{{0, 262.875}, {286.75, 0}} -> {0.0502218, -0.000125351, 
  0.00324895, 0.00324895}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClear[]; 
    MVWriteOFF["\</export/home/kuska/MathView3D/astroid.off\>", 
      Graphics3D[astroida]]\)], "Input",
  CellLabel->"In[44]:="],

Cell[BoxData[
    \("/export/home/kuska/MathView3D/astroid.off"\)], "Output",
  CellLabel->"Out[44]="]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Writing QuickDraw3D Metafiles", "Section",
  CellTags->"QuickDraw3D"],

Cell[TextData[{
  "The OpenGL viewer supports the export of Apple's QuickDraw3D metafiles. \
The format can be used to import the 3d data into Corel Draw 8, Corel Dream \
3D 8 and Apple specific 3d programs and viewers. For more information visit ",
  
  ButtonBox["Apples QuickDraw3D site",
    ButtonData:>{
      URL[ "http://www.apple.com/quicktime/qd3d/index.html"], None},
    ButtonStyle->"Hyperlink"],
  ". There you will also find a free viewer for MS-Windows."
}], "Text",
  CellTags->"QuickDraw3D"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVWrite3DMF["\<astro.3dmf\>", astroida]\)], "Input",
  CellLabel->"In[10]:=",
  CellTags->"QuickDraw3D"],

Cell[BoxData[
    \("astro.3dmf"\)], "Output",
  CellLabel->"Out[10]=",
  CellTags->"QuickDraw3D"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Writing VRML 2.0 Files", "Section",
  CellTags->"VRML"],

Cell[TextData[{
  "The OpenGL viewer can also export its data  as VRML 2.0 files. All \
features including colors and textures are supported by ",
  StyleBox["MVWriteVRML[]",
    FontFamily->"Courier"],
  ". "
}], "Text",
  CellTags->"VRML"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[MVWriteVRML]\)], "Input",
  CellLabel->"In[2]:=",
  CellTags->"VRML"],

Cell[BoxData[
    \({VRMLHeadLight \[Rule] False, VRMLExtrusion \[Rule] True, 
      MVTextureFilePrefix \[Rule] "mv"}\)], "Output",
  CellLabel->"Out[2]=",
  CellTags->"VRML"]
}, Open  ]],

Cell[TextData[{
  "The expanded lines are translated to VRML extrusions. In the most cases \
the tubes are shown correctly in the VRML viewer. For some curves the \
extrusion by te VRML system may fail. The ",
  StyleBox["VRMLExtrusion",
    FontFamily->"Courier"],
  "\[Rule]",
  StyleBox["False",
    FontFamily->"Courier"],
  " option can be used to store polygons instead of extrusions for expanded \
lines. The textures in the scene are stored in PNG files with names starting \
with ",
  StyleBox["MVTextureFilePrefix",
    FontFamily->"Courier"],
  ". "
}], "Text",
  CellTags->"VRML"]
}, Closed]],

Cell[CellGroupData[{

Cell["Writing PostScript Files", "Section",
  CellTags->"PostScript"],

Cell[TextData[{
  "The OpenGL viewer can write smooth shaded PostScript files. This feature \
bases on Mark Kilgrad's ",
  StyleBox["rendereps",
    FontFamily->"Helvetica"],
  " program and uses the same Painter algorithm for hidden surface removal. \
The triangles generated by ",
  StyleBox["mathview3d",
    FontFamily->"Helvetica"],
  " are not broken during the sorting of the 3D objects and so some pictures \
will not rendered correctly. In contrast to the PostScript generated by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " itself the polygons are smoothly shaded if the graphics is displayed with \
smooth faces in the output window. "
}], "Text",
  CellTags->"PostScript"],

Cell[BoxData[
    \(MVClear[]; MVShow3D[grMaeder2]; \)], "Input",
  CellLabel->"In[9]:=",
  CellTags->"PostScript"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVWriteEPS["\<maeder2.eps\>"]\)], "Input",
  CellLabel->"In[39]:=",
  CellTags->"PostScript"],

Cell[BoxData[
    \("maeder2.eps"\)], "Output",
  CellLabel->"Out[39]=",
  CellTags->"PostScript"]
}, Open  ]],

Cell[TextData[{
  "When only a file name is given as ",
  StyleBox["MVWriteEPS[\"myfile.eps\"]",
    FontFamily->"Courier"],
  " than the current contents of the window is saved as PostScript. Using the \
``Save as PostScript file'' option in the ``Files'' menu will save the image \
as  ",
  StyleBox["mview3d.eps",
    FontFamily->"Courier"],
  ". The option ",
  StyleBox["EPSBitmap",
    FontFamily->"Courier"],
  " is used to write the bitmap image of the frame buffer or to write \
resolution independent PostScript. "
}], "Text",
  CellTags->"PostScript"]
}, Closed]],

Cell[CellGroupData[{

Cell["Writing Portable Network Graphics Bitmaps", "Section",
  CellTags->"PNGBitmaps"],

Cell[TextData[{
  "The ",
  ButtonBox["PNG format",
    ButtonData:>{
      URL[ "http://www.cdrom.com/pub/png"], None},
    ButtonStyle->"Hyperlink"],
  " is the only supported format for saving the bitmap images created by ",
  StyleBox["MathGL3d",
    FontFamily->"Times New Roman"],
  ". The format is supported by  most image processing programs (",
  ButtonBox["Corel PhotoPaint",
    ButtonData:>{
      URL[ "http://www.corel.com"], None},
    ButtonStyle->"Hyperlink"],
  ", ",
  ButtonBox["Picture Publishe",
    ButtonData:>{
      URL[ "http://www.micrografx.com"], None},
    ButtonStyle->"Hyperlink"],
  "r, ",
  ButtonBox["PhotoImpact",
    ButtonData:>{
      URL[ "http://www.ulead.com"], None},
    ButtonStyle->"Hyperlink"],
  ", xv 3.10a, Image Magic, ...) and (now) by the popular WEB browsers like \
Netscape. With  the ",
  StyleBox["MVWritePNG[\"filename\",options]",
    FontFamily->"Courier"],
  " the image in the viewer window is saved to a PNG file. Valid options are \
",
  StyleBox["MVDisplayGamma",
    FontFamily->"Courier"],
  "\[Rule]\[Gamma]_Real to add a \[Gamma]-correction and ",
  StyleBox["MVInterlacePNG",
    FontFamily->"Courier"],
  "\[Rule]bool_(True | False) to create an interlaced picture. As second \
argument you can still give a ",
  StyleBox["Graphics3D[]",
    FontFamily->"Courier"],
  " object that is copied to the viewer window and finaly saved to a PNG \
file. For all other bitmap formats an image conversion tool is needed or the \
command ",
  StyleBox["Display[\"file\",MVGetRasterImage[],format]",
    FontFamily->"Courier New"],
  " can be used."
}], "Text",
  CellTags->"PNGBitmaps"]
}, Closed]],

Cell[CellGroupData[{

Cell["Animations", "Section",
  CellTags->"Animation"],

Cell[TextData[{
  "The OpenGL Viewer can calculate and render simple animations of the \
objects. The ",
  StyleBox["MVTranslate3D[]",
    FontFamily->"Courier"],
  " and the ",
  StyleBox["MVRotate3D[]",
    FontFamily->"Courier"],
  " functions will move and rotate the objects inside the viewer.The Options \
of ",
  StyleBox["MVTranslate[]",
    FontFamily->"Courier"],
  " and ",
  StyleBox["MVRotate3D[]",
    FontFamily->"Courier"],
  " allow the motion of a single ",
  StyleBox["MVMesh3D[]",
    FontFamily->"Courier"],
  " object. Creating the objects of every animation frame with ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " and pass the scene to the OpenGL Viewer is also possible. A freeware \
utility like the ",
  StyleBox["convert",
    FontFamily->"Helvetica"],
  " program of the Image Magic suite or a commercial application like ULead's \
GIF animator can be used to glue the frames to an animation. More complicated \
animations are possible with POVRay."
}], "Text",
  CellTags->"Animation"],

Cell[BoxData[
    \(Needs["\<Graphics`Polyhedra`\>"]\)], "Input",
  CellLabel->"In[36]:=",
  CellTags->"Animation"],

Cell["At first some nice graphics is passed to the viewer window.", "Text",
  CellTags->"Animation"],

Cell[BoxData[
    \(MVClear[]; \n
    \[IndentingNewLine]MVShow3D[
      \[IndentingNewLine]Graphics3D[Stellate[Icosahedron[]], 
        \[IndentingNewLine]\ 
        ViewPoint \[Rule] \ {1.02486, \(-1.98794\), 1.63883}], 
      \[IndentingNewLine]MVPolygonShading \[Rule] MVFlat]; \)], "Input",
  CellLabel->"In[14]:=",
  CellTags->"Animation"],

Cell["The resolution of the image depends on the window size. ", "Text",
  CellTags->"Animation"],

Cell[BoxData[
    \(\(MVReshapeWindow[128, 128]; \)\)], "Input",
  CellLabel->"In[26]:=",
  CellTags->"Animation"],

Cell[TextData[{
  "Because we need the animation inside of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " the frames are saved with the ",
  StyleBox["MVGetRasterImage[]",
    FontFamily->"Courier"],
  " function. If an external program is used to collect the frames one has to \
save PNG bitmaps. We need only seven frames to get the impression of a full \
turn. "
}], "Text",
  CellTags->"Animation"],

Cell[BoxData[
    \(frameNo = 18; \n\[Alpha] = 2  \[Pi]/frameNo // N; \nimageList = {}; \n\n
    Do[\[IndentingNewLine]MVRotate3D[{0, 0, 1}, \[Alpha]]; 
      \[IndentingNewLine]AppendTo[imageList, MVGetRasterImage[]], 
      \[IndentingNewLine]{i, 0, frameNo/3}\[IndentingNewLine]]\)], "Input",
  CellLabel->"In[50]:=",
  CellTags->"Animation"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(Show[
          Graphics[#, AspectRatio \[Rule] Automatic, 
            ImageSize \[Rule] {128, 128}]]\ &\)\ /@\ imageList; \)\)], "Input",\

  CellLabel->"In[54]:=",
  CellTags->"Animation"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
%%ImageSize: 128 128 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00744048 0.0238095 0.00744048 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
384 string
128 128 8 [128 0 0 128 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B63637B6363000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008C7B63735A5A7B6363000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A4A638C735A8C7B5A7B63637B6363000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
524A634A4263947B638C735A8C7B637B635A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B6363
4A42634A42638C7B5A8C7B638C735A7B63637B6363000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A4263
524A6B4A42638C7B638C735A8C7B6384735A7B63637B635A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A42634A4263
4A42634A4A6B8C735A8C7B638C735A8C7B6384735A7B63637B6363000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A42634A4A634A4263
524A6B4A4263947B638C735A8C7B6384735A8C7B6384735A846B637B6363000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000007B63634A42634A42634A4263
4A42634A42638C7B638C7B638C735A8C7B5A8C735A8C735A84735A846363000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A4A634A42634A4A634A4263
524A6B4A42638C7B6384735A8C7B638C735A8C7B6384735A8C7B6384735A846B63000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A42634A42634A42634A42634A4263
4A42634A4A6B8C7B5A8C7B638C7B638C7B6384735A8C735A84735A8C735A84735A846363
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004A426342425A4A42634A42634A4A6B4A4263
4A4A6B4A42638C7B638C735A8C7B6384735A8C7B6384735A8C7B6384735A8C7B5A7B6363
846B6B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B63634A42634A42634A42634A42634A42634A4263
4A42634A426B4A42638C7B638C7B5A8C7B638C735A8C7B6384735A84735A84735A84735A
846363000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000042425A4A426342425A4A4A634A42634A4A634A4263
524A6B4A4263524A6B8C735A8C7B6384735A8C7B6384735A8C7B6384735A8C7B5A84735A
847B5A846363000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004A42634A42634A42634A42634A42634A42634A42634A4263
4A42634A4A6B4A42638C7B638C7B638C7B6384735A8C7B5A84735A8C7B5A84735A84735A
84735A84735A846363000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000042425A4A426342425A4A42634242634A42634A42634A4A6B4A4263
524A6B4A4263524A6B8C7B5A8C7B6384735A8C7B6384735A8C7B6384735A8C7B6384735A
84735A7B735284735A846363000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B636B4A42634242634A42634A42634A42634A42634A42634A42634A426B
4A42634A4A6B4A426B8C7B638C7B638C7B638C7B5A8C7B6384735A847B5A84735A84735A
84735A84735A84735A84735A525273000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A426342395A4A42634242634A42634242634A4A6B4A42634A4A6B4A4263
524A6B4A4263524A6B84735A8C7B63847B5A8C7B6384735A8C7B6384735A847B5A84735A
8C7B6384735A84735A7B6B5284735A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A426342425A4A42634A42634A42634A42634A42634A42634A426B4A42634A426B
4A426B4A4A6B4A426B8C7B638C7B638C7B63847B5A8C7B6384735A8C7B6384735A84735A
84735A84735A84735A84735A84735A846B6B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000312929000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000312942422929
000000000000000000000000000000000000000000000000000000000000000000000000
4A426342395A4A426342425A4A426342425A4A42634242634A4A6B4A42634A4A6B4A4263
4A4A6B4A4263524A6B8C7B638C7B6384735A8C7B6384735A8C7B6384735A8C7B6384735A
847B5A84735A847B5A7B735A84735A7B6B528C6B6B000000000000000000000000000000
0000000000000000000000000000000000003931313129295A5A42000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000292942312942
422929422929000000000000000000000000000000000000000000000000000000000000
42425A4A42634242634A42634242634A42634A42634A42634A42634A42634A42634A4A6B
4A42634A4A6B4A426B8C7B638C7B638C7B638C7B638C7B6384735A847B5A84735A84735A
84735A84735A84735A84735A7B735A84735A7B735A8C6B6B000000000000000000000000
0000000000000000000000003129293929295A5A395A5A425A5239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292942
31314A4229294A313142292900000000000000000000000000000000000000000042395A
4A426342425A4A426342395A4A42634242634A42634242634A4A6B4A42634A4A6B4A4263
524A6B4A4263524A6B847B5A8C7B63847B5A8C7B6384735A8C7B6384735A8C7B6384735A
847B6384735A84735A7B735284735A7B735284735A7B6B52000000000000000000000000
0000000000003931313129293931315A5239635A425A52395A5A42000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292942
29294231294A4A29294A29294A29294A313142292900000000000000000042395A42425A
42425A4A426342425A4A42634A42634A42634A42634A42634A42634A426B4A42634A426B
4A426B4A4A6B4A426B8C7B638C7B638C7B63847B5A8C7B63847B638C7B6384735A84735A
84735A84735A84735A84735A84735A84735A7B73527B735A7B7352000000000000000000
392929393131392929392931635A42635A425A52395A52395A5239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292942
31294A29294231314A4A29294A31314229294A31314229294A313142395A4A426342395A
4A426342395A4A426342425A4A42634242634A4A6B4242634A4A6B4A42634A4A6B4A4263
4A4A6B4A4263524A6B8C7B638C7B63847B5A8C7B6384735A8C7B6384735A8C7B6384735A
847B637B735A847B5A7B735A84735A7B735284735A7B6B527B735A7B6B52393131312929
393131392929393131635A42635A425A5239635A425A52395A5A42000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000312942
29294231294229294231294A31294A4A31314A29294A31314A292942425A42425A424263
42425A4242634242634A42634242634A42634A42634A42634A42634A426B4A42634A4A6B
4A426B4A4A6B4A426B8C7B638C7B638C7B638C7B638C7B63847B5A847B6384735A847B63
84735A84735A84735A84735A7B735A84735A7B735A7B735A7B73527B735A7B6B52393131
392929393131635A42635A42635A42635A425A5A395A5A425A5239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292942
31314A29294231294A29294231314A29294A5231314A29294A426342395A42426342395A
4A426342425A4A426342425A4A42634242634A42634242634A4A6B4A42634A4A6B4A4263
524A6B4A426B524A6B847B638C8463847B638C7B6384735A8C7B6384735A847B6384735A
847B6384735A847B5A7B735A847B5A7B735A84735A7B6B5284735A7B6B527B735A6B7B84
6B634A635A426B6342635A396363425A5A39635A425A52395A5A42000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292942
29294231294A29294231294A31294A31314A31294A42425A42425A42426342395A424263
4242634A42634242634A42634A42634A42634242634A42634A42634A4A6B4A42634A426B
4A426B4A4A6B4A426B8C7B638C7B638C7B63847B638C7B63847B63847B6384735A847B5A
84735A847B5A84735A84735A7B735A84735A7B735A7B735A7B735A7B735A7B6B52523131
523131523139522931635A42635A42635A425A52395A5A425A5239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292942
31294A29294231314A63423163423963423163423963423142426342395A4A426342395A
4A426342425A4A42634242634A42634242634A4A6B4242634A4A6B4A42634A4A6B4A4263
4A4A6B4A4263524A6B8C7B638C8463847B638C7B63847B5A8C7B6384735A8C7B6384735A
847B637B735A847B637B735A84735A7B735284735A7B73527B735A7B6B527B735A4A2931
5231394A29315231395229315231395229315A31394A4A6B5A5A42000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000634231
63423163423163423163423163423163423163423163423163423142425A42425A424263
4242634A42634242634A42634242634A42634A42634A42634A42634A426B4A426B4A4A6B
4A426B4A4A6B4A4A6B8C7B638C7B638C7B638C7B638C7B63847B63847B63847B5A847B63
84735A847B5A84735A84735A7B735A84735A7B735A7B735A7B735A7B735A523131523139
5231315231395231315231395231315231395231395A31395231395A31395A31395A3139
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B4A396342316B4A396342316B4A39634231
6B4A396342316B42396342316B4A396342316B42396342316B4A3963423142426342395A
4A42634242634A42634242634A4A6B4242634A426B4242634A4A6B4A42634A4A6B4A4263
524A6B4A426B524A6B847B638C846B847B638C7B6384735A8C7B63847B5A847B6384735A
847B6384735A847B637B735A847B637B735A84735A7B6B5284735A522931523139522931
5A31395229315231395229315A31395231315A31395229315A31395231315A3139522931
5A31395A31395A31395A2931633139000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000006B42316B42396B42316B4A396B42316B42316B42316B4A396342316B4231
6B42316B42396342316B42316342316B42396342316342316342316B423142425A424263
4242634A42634242634A42634A42634A42634A42634A42634A42634A4A6B4A42634A426B
4A4A6B4A4A6B4A426B8C7B638C7B638C7B63847B638C7B63847B638C7B63847B5A847B63
847B5A847B6384735A84735A84735A84735A7B735A7B735A523139523139522931523139
5231395231395231315231395231395A31395231395A31395A31395A31395A31395A3139
5A31395A31395A31395A31395A31396331395A3139633139633139633139000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006B4231734A396B4231
6B4A396B42316B4A396B42316B4A396B42316B4A396B42316B4A396342316B4A39634231
6B4A396342316B4A396342316B4A396342316B4A396342316B4A396342316B4A3942395A
4A426342425A4A42634242634A426B4242634A4A6B4242634A4A6B4242634A4A6B4A4263
4A4A6B4A426B524A738C7B638C8463847B638C7B63847B638C7B6384735A8C7B6384735A
847B637B735A847B637B735A847B5A7B735A847B5A5229315231395229315A3139522931
5A31395229315A31395229315A31395229315A31395231315A31395A31315A31395A3139
5A31395A31396331395A31396331395A31396331395A31396331395A3139633939848C8C
84948C000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000734A396B4A39734A396B4A396B4A396B4A396B4A39
6B42316B4A396B42396B4A396B42316B4A396B42316B4A396B42316B4A396B42316B4A39
6B42316B4A396B42316B4A396B42316B42396B42316B42396342316B42316342316B4239
4242634242634242634A42634242634A42634A42634A426B4A42634A426B4A426B4A4A6B
4A426B4A4A6B4A4A6B8C84638C7B638C7B63847B638C7B63847B63847B63847B63847B63
84735A847B5A84735A847B5A7B735A84735A523139523139523139523139523139523139
5231395A31395231395A31395A31395A31395A31395A31395A31395A31395A31395A3139
5A31395A31395A31396331395A313963313963313963313963313984948C84948C84948C
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000063738C63739463738C734A396B4231734A396B4231
734A396B4231734A396B4231734A396B42316B4A396B4231734A396B42316B4A396B4231
6B4A396B42316B4A396342316B4A396B42316B4A396342316B4A396342316B4A39634231
6B4A394242634A42634242634A4A6B4242634A426B4242634A4A6B4A42634A4A6B4A4263
524A6B4A426B524A6B847B638C846B847B638C7B63847B5A8C7B63847B5A847B6384735A
847B6384735A847B637B735A847B635229315231395229315A31395231395A3139522931
5A31395231395A31395229315A31395A31395A31395A31316331395A31395A31395A3131
6339425A31396331395A31396339425A313984948C7B8C8C8494947B8C8C84948C7B8C8C
947B738C7B6B947B6B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000063738C63739463739463739463738C734A39
734A39734A396B4A396B4A396B4A39734A396B42316B4A396B4A396B4A396B42316B4A39
6B4A396B4A396B42316B4A396B4A396B4A396B42316B4A396B42316B4A396B42316B4231
6B42316B4A394242634A42634A42634A426B4242634A426B4A426B4A4A6B4A426B4A4A6B
4A4A6B4A4A6B4A426B8C7B638C7B638C8463847B638C7B63847B63847B63847B5A847B63
847B63847B6384735A84735A5231395231395231395231395231395A31395231395A3139
5A31395A31395A31395A31395A31395A31395A31395A31395A31395A31395A31395A3139
5A31396331395A3139633139848C8C84948C7B8C8C84948C84948C84948C7B8C8C84948C
947B6B947B738C7B6B947B6B947B6B947B6B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006B5A6B7363736B5A6B63739463738C63739463738C6B7B9463738C
6B7B946B4231734A396B4231734A396B4231734A396B4231734A396B4231734A396B4231
734A396B4231734A396B4231734A396B4231734A396B42316B4A396B42316B4A396B4231
6B4A396B42316B4A394242634A426B4242634A4A6B4242634A4A6B4A42634A4A6B4A426B
4A4A6B4A426B524A738C7B638C846B847B638C846B847B638C7B63847B5A8C7B6384735A
847B637B735A847B635229315231395229315A31395229315A31395229315A3139523139
5A31395229315A31395A31395A31395A31396331395A31396331395A31396331395A3139
6331397B8C8C8494947B8C8C8494947B8C8C8494947B8C8C84948C7B8C8C84948C8C7B6B
947B738C7B6B947B738C7B6B947B738C7B6B947B738C7B6B947B6B8C736B000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
6B5A6B6B63736B636B7363736B637373637373637363739463738C63739463738C637394
63738C637394637394637B94734A39734A39734A39734A396B4A39734A396B4A39734A39
6B4A316B4A396B4A396B4A396B4A316B4A396B4A316B4A396B42316B4A396B42316B4A39
6B42316B4A396B42316B4A394242634A42634A42634A426B4A426B4A4A6B4A426B4A4A6B
4A426B4A4A6B4A4A6B8C846B8C7B638C7B638C7B638C7B63847B63847B63847B63847B63
847B5A847B635231395231395231395231395231395A31395231395A31395A31395A3139
5A31395A31395A31395A31395A31395A31395A31396331395A31396331395A31397B948C
7B8C8C7B948C7B8C8C84948C7B8C8C7B8C8C7B8C8C84948C7B8C8C84948C7B8C8C947B73
947B6B947B73947B6B947B6B947B6B947B6B947B6B947B6B8C7B6B947B6B8C7B6B947B6B
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006B5A6B6B63736B5A6B
7363736B5A6B7363736B5A6B7363736B636B7363735A738C637B9463738C63739463738C
6B7B9463738C637B9463738C6B7B9463738C6B7B946B4231734A396B4A31734A396B4231
734A396B4A31734A396B4231734A396B4231734A396B4231734A396B4231734A396B4231
734A396B42316B4A396B4231734A394242634A4A6B4242634A4A6B4A426B4A4A6B4A426B
524A734A426B524A73847B638C846B847B638C8463847B638C8463847B63847B6384735A
847B6384735A5A31395229315A31395231395A31395229315A31395231395A3139522939
5A31425A31395A31395A31396339425A31396331395A31398494947B8C8C8494947B8C8C
8494947B8C8C84948C7B8C8C8494947B8C8C84948C7B8C8C84948C7B8C8C84948C8C7B6B
9484738C7B6B947B738C7B6B9484738C7B6B947B738C7B6B9484738C7B6B947B6B000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006B636B6B63736B5A6B6B6373
6B63737363736B63737363737363737363736B637363738C63738C63739463738C637394
63738C63739463738C637394637394637B9463738C637B94637B946B7B94734A39734A39
734A39734A396B4A31734A39734A39734A396B4A31734A396B4A39734A396B4A316B4A39
6B4A39734A396B42316B4A396B4A316B4A394242634A426B4A426B4A4A6B4A426B4A4A6B
4A4A6B4A4A6B4A4A6B8C84638C7B638C8463847B63847B63847B63847B63847B63847B63
847B635231395231395231395231395A31395231395A31395A31395A31395A31395A3139
5A31395A31395A31395A31395A31396331397B8C8C7B8C8C7B8C8C7B8C947B8C8C7B8C8C
7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B948C7B8C8C7B8C8C7B8C8C7B948C947B6B947B73
947B73947B73947B6B947B73947B6B947B738C7B6B947B6B947B6B947B6B000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006B5A6B7363736B5A6B
7363736B5A6B7363736B636B7363736B636B7363736B63736373945A738C63739463738C
63739463738C637B9463738C6B7B9463738C6B7B9463738C6B7B9463738C6B7B9463738C
734A396B4A31734A396B4A31734A396B4231734A396B4A31734A396B4231734A396B4231
734A396B4231734A396B4231734A396B42314A4A6B4242634A4A6B4242634A4A6B4A426B
4A4A734A426B524A738C7B638C846B847B638C846B847B638C8463847B638C7B63847B5A
5A31395229395A31395231395A31395231395A31395231395A31395A31395A31395A3139
6331395A31396331427B8C8C7B8C947B8C8C7B94947B8C8C7B94947B8C8C7B94947B8C8C
7B948C7B8C8C7B948C7B8C8C7B948C7B8C8C84948C7B8C8C7B948C7B8C84948473947B6B
9484738C7B6B9484738C7B6B9484738C7B6B9484738C7B6B9484738C7B6B000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006B63736B6373
6B63737363736B63737363736B637373637373637373637363738C63738C63738C637394
63738C63739463738C63739463738C63739463738C637B9463738C637B946373946B7B94
637B946B7B94637B94734A39734A39734A39734A39734A39734A39734A39734A39734A39
6B4A31734A396B4A39734A396B4A31734A396B4A314A4A6B4A426B4A4A6B4A426B4A4A6B
4A426B4A4A6B4A4A6B8C846B8C7B638C8463847B638C7B63847B63847B63847B635A3139
5231395A31395231395A31395A31395A31395A31395A31395A31395A31395A31395A3139
5A31397B8C947B8C8C7B8C947B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C
7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C947B73947B73
947B73947B73947B73947B73947B6B947B73947B6B947B73947B6B000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007363736B5A6B
7363736B63737363736B5A6B7363736B63737363736B63736373945A738C6373945A6B8C
637B9463738C63739463738C637B9463738C637B9463738C6B7B9463738C6B7B9463738C
6B7B9463738C6B7B9463738C6B7B94637B94734A39734A31734A39734A31734A396B4231
734A39734A31734A396B4231734A396B4A31734A396B42314A4A6B4A426B4A4A6B4A426B
524A734A426B4A4A73847B638C846B847B638C846B847B638C846B847B635A3139522939
5A31425231395A31395229395A31425A31395A31395A31396339425A31397B8C94738C8C
7B8C947B8C8C7B8C94738C8C7B94947B8C8C7B8C94738C8C7B94947B8C8C7B8C8C738C8C
7B94947B8C8C7B8C8C738C847B948C7B8C8C7B8C8C738C847B948C947B73948473947B6B
948473947B6B9484738C7B6B948473947B6B9484738C7B6B000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006B6373
6B63737363736B637373637373637373637373637373637373637363738C5A738C63738C
63738C63739463738C63738C63738C63739463738C63738C63738C637B9463738C637B94
6373946B7B9463738C637B94637B946B7B94637B946B7B946B7B94734A39734A39734A39
734A39734A39734A39734A39734A39734A39734A31734A39734A394A4A6B4A426B4A4A6B
4A4A6B4A4A734A4A6B8C846B8C84638C846B847B638C7B63847B635A31395231395A3139
5A31395A31395A31395A31395A31395A31395A31395A31397B8C947B8C947B8C8C7B8C94
7B8C8C7B8C94738C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C
7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C847B8C8C7B8C8C948473947B73947B73
947B73948473947B6B947B73947B73948473947B6B000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7363736B63737363736B63737363736B6373736B7B6B6373736B7B5A6B8C6373945A738C
6373945A738C63739463738C63739463738C637B9463738C637B9463738C6B7B9463738C
6B7B9463738C6B7B9463738C6B7B9463738C6B7B94637B8C6B7B9463738C6B7B94734A31
735239734A31735239734A31734A396B4A31734A396B4A31734A396B4A314A4A734A426B
4A4A734A426B524A738C7B638C846B847B638C846B847B635A31395229395A3139523139
5A31425231395A31425A31395A314273848C7B8C9473848C7B8C9473848C7B8C94738C8C
7B8C9473848C7B8C94738C8C7B8C94738C8C7B8C94738C8C7B8C8C73848C7B8C8C738C8C
7B8C8C738C847B8C8C738C847B8C8C738C847B8C8C738C847B8C8C947B739C8473947B73
948473947B6B948473947B6B948473947B6B948473000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007363736B637373637373637373637373637373637373637363738C5A738C63738C
63738C63738C63738C63738C63738C63738C63738C63738C63738C63739463738C637B94
63738C637B9463738C637B94637B8C637B94637B946B7B94637B946B7B94637B946B7B94
6B7B946B7B94734A39734A39734A39734A39734A39734A39734A39734A39734A394A4A6B
4A426B4A4A6B4A4A6B8C846B8C7B638C846B847B635A31395231395A31395A31395A3139
5A31395A31395A31397B8C94738C8C7B8C94738C8C7B8C94738C8C7B8C94738C8C7B8C94
738C8C7B8C8C738C8C7B8C8C738C8C7B8C8C738C8C7B8C8C738C8C7B8C8C738C8C7B8C8C
738C8C7B8C8C7B8C8C7B8C8C738C8C7B8C8C7B8C8C7B8C8C7B8C84948473947B73948473
947B73948473947B73948473947B73948473000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007363736B6373736B7B6B637373637B6B6373736B7B73637363738C5A6B8C
6373945A738C63738C5A6B8C637B9463738C6373945A738C637B9463738C637B9463738C
6B7B9463738C637B9463738C6B7B9463738C6B7B9463738C6B7B94637B8C6B7B9463738C
6B7B94637B946B7B94637B8C6B8494734A39734A39734A317B5239734A31734A39734A31
524A734A426B524A73847B638C846B847B635A31395229395A31425231395A3139523139
7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C7B8C8C73848C7B8C9473848C7B8C8C73848C7B8C8C738C8C7B8C8C738484
7B8C8C738C847B8C8C7384847B8C8C738C847B8C8C7384849C847B947B73948473947B73
9C8473947B73948473947B6B9C8473000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000006B637373637373637373637B736373736373736373736B7B5A6B8C63738C
63738C63738C5A738C63738C63738C63738C63738C63738C63738C63738C63738C63738C
63738C637B9463738C637B8C637B8C637B9463738C637B94637B946B7B94637B8C6B7B94
637B946B7B94637B946B7B946B7B946B7B946B7B946B7B94734A39734A39734A39734A39
734A394A4A734A4A6B8C846B8C846B5A31395231395A31395A31395A3139738494738C94
738C94738C9473848C738C94738C8C7B8C9473848C738C8C738C8C7B8C9473848C738C8C
738C8C7B8C8C73848C738C8C738C8C7B8C8C738C8C738C8C738C8C7B8C8C738C8C738C8C
738C8C7B8C8C738C84738C8C738C8C7B8C8C738C847B8C8C948473948473947B73948473
948473948473947B73948473000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006B6373736B7B6B6373736B7B736373736B7B736373736B7B5A6B8C
63738C5A6B8C63738C5A738C63738C5A738C6373945A738C6373945A738C637B9463738C
637B9463738C637B9463738C637B9463738C6B7B9463738C6B7B9463738C6B7B9463738C
6B7B94637B8C6B7B94637B8C6B7B94637B8C6B8494637B946B8494637B8C6B8494734A31
7B5239734A31524A738C7B638C846B5231395A314273848C738C9473848C7B8C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C7B8C8C73848C7B8C8C73848C7B8C8C73848C7B8C8C7384847B8C8C738484
7B8C8C7384847B8C8C7384847B8C8C7384847B8C8C7384849C847B947B739C847B947B73
9C8473947B739C8473947B73000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000073637373637B736373736B7B736373736B7B73637B63738C
5A738C63738C5A738C63738C5A738C63738C63738C63738C63738C63738C63738C63738C
63738C63738C63738C637B8C63738C637B8C63738C637B9463738C637B94637B8C6B7B94
637B8C6B7B94637B946B7B94637B946B7B946B7B946B7B946B7B946B7B946B7B946B7B94
6B7B947B4A394A4A738C846B5A3139738C94738494738C94738494738C94738494738C94
73848C738C9473848C738C9473848C738C8C73848C738C8C73848C738C8C73848C738C8C
73848C738C8C73848C738C8C73848C738C8C738C8C738C8C73848C738C8C738C8C738C8C
738484738C8C738C84738C8C738C84738C8C738C849C8473948473948473948473948473
948473948473948473000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000736373736B7B6B6373736B7B736373736B7B5A6B84
63738C5A6B8C63738C5A6B8463738C5A738C63738C5A6B8C637B945A738C63738C5A738C
637B9463738C637B945A738C6B7B9463738C637B9463738C6B7B9463738C6B7B9463738C
6B7B94637B8C6B7B9463738C6B7B94637B8C6B7B94637B8C6B8494637B946B8494637B8C
6B84946B7B945231427B5239738C9473848C738C946B848C738C9473848C738C946B848C
738C9473848C738C9473848C738C9473848C738C9473848C7B8C9473848C738C8C73848C
7B8C8C73848C738C8C7384847B8C8C73848C738C8C7384847B8C8C738484738C8C738484
7B8C8C738484738C8C7384847B8C8C738484738C8C947B739C847B9484739C847B947B73
9C847B947B73000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000073637373637B73637B736B7B736373736B7B
5A738C63738C5A6B8C5A738C5A738C63738C5A738C63738C63738C63738C63738C63738C
63738C63738C63738C63738C63738C637B9463738C637B8C63738C637B9463738C637B8C
637B8C6B7B94637B8C637B94637B946B7B94637B8C6B7B946B7B946B7B94637B946B7B94
52314252314284847352527B7B5A42845A4273848C738494738494738C9473848C73848C
73848C738C9473848C73848C73848C738C8C73848C73848C73848C738C8C73848C738C8C
73848C738C8C73848C73848C73848C738C8C73848C738C8C73848C738C8C738484738C8C
738C8C738C8C738484738C84738484738C8C7384849C84739484739C847B948473948473
948473000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000736B7B736373736B7B736373736B7B736373
63738C5A6B8463738C5A6B8463738C5A6B8463738C5A6B8C63738C5A738C63738C5A738C
63738C5A738C637B945A738C637B9463738C637B9463738C637B9463738C6B7B9463738C
6B7B9463738C6B7B9463738C6B7B9463738C6B7B94637B8C6B7B94637B8C52314A523142
52314A8484738C8C7B4A527B525A847B5239845A427B5239738C946B848C738C946B848C
738C946B848C738C946B848C738C946B848C738C946B848C738C8C6B848C738C8C73848C
738C8C6B8484738C8C738484738C8C738484738C8C738484738C8C738484738C8C738484
738C8C738484738C8C738484738C8C7384849C847B9484739C847B9484739C847B948473
9C847B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000736B7B73637B736B7B736B7B736B7B
5A6B8C5A738C5A6B8C5A738C5A6B8C5A738C5A738C63738C5A738C63738C63738C63738C
5A738C63738C63738C63738C63738C63738C63738C637B8C63738C637B8C63738C637B8C
63738C637B8C637B8C637B94637B8C637B94637B8C6B7B94523142523142523142523142
848C738C8C73848C7352527B52527B52527B7B5A42845A427B5A397B5A427B5A3973848C
73848C73848C73848C73848C73848C73848C73848C73848C73848C73848C73848C738C8C
73848C73848C73848C73848C73848C73848C73848C738C8C73848473848C738484738C8C
738484738C84738484738C84738484738C849484739C847B9484739C847B9484739C847B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000736B7B73637B736B7B736373
7B6B7B5A6B845A738C5A6B8463738C5A6B8463738C5A6B8463738C5A738C63738C5A6B84
63738C5A738C63738C5A738C637B9463738C637B8C5A738C637B9463738C637B9463738C
6B7B9463738C637B9463738C6B7B94637B8C5231424A294252314A52314252314A4A3142
8C8C7B8484738C8C7B4A527B525A844A527B525A847B5239845A427B5A397B5A427B5239
7B5A426B848C73848C6B848C738C946B848C738C8C6B848C738C8C6B848C738C8C6B8484
738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484
738C8C6B8484738C8C6B8484738C8C7384849C847B9484739C847B9484739C847B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000736B7B73637B736B7B
736B7B5A738C5A6B845A738C5A738C5A738C5A6B8C5A738C5A738C63738C5A738C63738C
63738C63738C5A738C63738C63738C63738C63738C63738C63738C637B8C63738C637B8C
63738C637B8C63738C637B8C4A31425231424A314252314252314252314A5231428C8C7B
8C8C738C8C7B84847352527B52527B525A7B4A527B845A427B5A42845A427B5A397B5A42
7B5A397B5A427B523973848C73848C73848C6B848C73848C73848C73848C6B848C73848C
73848C73848C6B848473848C73848C73848C6B848473848C738484738C8C6B8484738484
738484738C846B8484738484738484738C849484739C847B9C847B9C847B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000073637B7B6B7B73637B
7B6B7B73637B5A738C5A6B8463738C5A6B8463738C5A6B8463738C5A6B8463738C5A6B84
63738C5A738463738C5A738C63738C5A738C637B8C5A738C637B8C5A738C637B9463738C
637B9463738C5231424A31425231424A314252314A4A314252314A4A31428C8C7B848473
8C8C7B8484738C8C7B4A527B525A7B4A527B525A844A527B845A427B5A39845A427B5A39
845A427B52397B5A427B52397B5A42735239738C8C6B848C73848C6B8484738C8C6B8484
738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484
738C8C6B8484738C8C6B8484738C849484739C8C7B9484739C847B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000736B7B736B7B
736B7B736B7B5A6B845A738C5A6B845A738C5A6B8C5A738C5A6B845A738C5A738C63738C
5A738C63738C5A738C63738C5A738C63738C63738C63738C63738C63738C63738C637B8C
4A31424A31424A31425231424A31425231425231425231425231428C8C7B848C7B8C8C7B
848C73848C7384847352527B4A527B52527B52527B525A8452527B845A427B5A42845A42
7B5A397B5A427B5A397B5A427B5A397B5A427B5A397B5A426B848C6B848C6B848C73848C
6B848473848C6B848473848C6B848473848C6B848473848C6B84847384846B8484738484
6B84847384846B84847384846B84849C847B9C847B9C847B9C847B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000073637B
7B6B84736B7B5A738C5A6B845A738C5A6B845A738C5A6B8463738C5A6B8463738C5A6B84
63738C5A738463738C5A6B8463738C5A738C63738C5A7384637B8C5A738C4A31424A2942
52314A4A31425231424A294252314A4A314252314A4A29428C8C7B848C738C8C7B848473
8C8C7B8484738C8C734A5273525A844A527B525A844A527B525A8452527B845A427B5A39
8463427B5A39845A427B5239845A427B52397B5A427352397B5A4273523973848C6B8484
73848C6B848473848C6B8484738C8C6B848473848C6B8484738C8C6B84847384846B8484
738C8C6B84847384846B8484738C849C847B9C8C7B948473000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
736B7B7B6B84736B7B5A6B845A6B845A738C5A6B845A738C5A738C5A738C5A6B845A738C
5A738C63738C5A73845A738C5A738C63738C5A738C63738C4A31424A31424A31424A3142
4A31425231424A314252314252314252314A5231425231428C8C7B8C8C7B848C73848C7B
848C73848C738484734A527B4A527B525A7B4A527B52527B52527B525A8452527B845A42
845A42845A427B5A397B5A427B5A397B5A427B5A397B5A397B5A397B5A427352397B5A39
6B84846B848C6B84846B84846B848473848C6B84846B84846B84846B84846B84846B8484
6B84847384846B84846B84849C847B9C8C7B9C847B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000736B7B7B6B845A6B845A738C5A6B845A738C5A6B845A738C5A6B8463738C5A6B84
63738C5A6B8463738C5A6B8463738C5A73844A31424A29425231424A29425231424A2942
52314A4A314252314A4A314252314A4A314252314A848C7B8C8C7B8484738C8C7B848473
8C8C7B8484738C8C7B4A527B525A7B4A527B525A844A527B525A844A527B525A847B5A39
8463427B5A398463427B5A39845A427B5A39845A427B52397B5A427B52397B5A42735239
7B5A427352397B5A426B848473848C6B848473848C6B84847384846B84847384846B8484
7384846B84847384846B84849C8C7B9C847B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007B6B84736B7B5A6B845A6B845A6B845A6B845A738C5A6B845A738C5A6B845A738C
5A6B845A738C5A73845A738C4A31424A31424A31424A31424A31424A31424A3142523142
4A314252314252314252314A52314252314A848C7B8C8C7B848C7B8C8C7B848C7B848C7B
848473848C7384847352527B4A527B52527B4A527B525A844A527B525A84525A84525A84
845A42845A42845A42845A427B5A397B5A427B5A397B5A427B5A397B5A427B5A397B5A42
7B5A397B5A397352397B5A397352396B84846B84846B84846B84846B84846B84846B8484
6B84846B84846B84846B84849C847B9C8C7B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B6B84736B7B5A738C5A6B845A738C526B845A738C5A6B845A738C5A6B84
63738C4A29424A31424229394A31424A29424A31424A294252314A4A314252314A4A2942
52314A4A314252314A4A314252314A848C7B8C8C7B8484738C8C7B8484738C8C7B848473
8C8C7B848473848C734A5273525A844A527B525A844A527B525A844A527B525A844A527B
525A84845A398463427B5A398463427B5A39845A427B5A39845A427B5A397B5A42735239
7B5A427B52397B5A427352397B5A427352397B5A396B7B846B848C6B84846B84846B7B7B
7384846B84846B84849C847BA58C84523929524231000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000002929297B6B845A6B845A6B845A6B845A6B845A6B845A738C5A6B844A3142
4A31424A31424A31424A31424A31424A31424A31424A31424A31425231424A3142523142
52314252314A52314252314A8C8C7B8C8C7B848C7B848C7B848C7B848C7B848473848C73
848C73848C738484734A527B4A527B52527B4A527B52527B52527B525A8452527B525A84
525A84525A84845A42845A42845A42845A427B5A397B5A427B5A427B5A427B5A397B5A42
7B5A397B5A427B5A397B5A397B5A397B5A39735239735A39735239735A396B84846B8484
6B84846B84846B84849C8C7B5A42315A4231523929524231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000002929293131312929297B7384526B845A738C526B845A738C4229394A3142422939
4A31424229424A31424A29424A31424A294252314A4A294252314A4A294252314A4A3142
52314A4A314252314A5231428C8C84848C7B8C8C7B848C7B8C8C7B8484738C8C7B848473
848C7B848473848C7B4A527B52527B4A527B525A844A527B525A844A527B525A844A527B
525A844A5284525A84845A398463427B5A398463427B5A398463427B5A398463427B5A39
7B5A427B5A397B5A427352397B5A427352397B5A427352397B5A427352397B5A42735239
6B8484637B7B6B84845A42296342315239295A4231523929524231000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
2929293129313129313131313129315A6B845A6B844A31424229424A31424A31424A3142
4A31424A31424A31424A31424A31424A31424A31425231424A31425231424A314252314A
52314252314A5231428C8C7B848C7B848C7B848C7B848C7B848C7B848C7B848C73848C7B
848473848C738484734A527B4A527B4A527B4A527B525A7B4A527B525A84525284525A84
525A84525A84525A84525A84845A42846342845A428463427B5A42845A427B5A427B5A42
7B5A397B5A427B5A397B5A427B5A397B5A397B5A397B5A39735A397B5A39735239735A39
7352397352396B84846342315A42315A42315A42315A4231523929524231000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292929
3131312929293131312929293131314229394A31424229394A31424229424A3142422942
4A31424A29424A31424A294252314A4A314252314A4A294252314A4A314252314A4A2942
52394A5231428C8C84848C7B8C8C84848C7B8C8C7B8484738C8C7B848473848C7B848473
848C7B848473848C734A5273525A7B4A527B525A7B4A527B525A844A527B525A844A527B
525A84525A84525A845252848C6342845A428463427B5A398463427B5A398463427B5A39
8463427B5A397B5A427B5A397B5A427B5A397B5A427352397B5A427352397B5A39735239
7B5A42735239A59C8C52736B634A315A42315A42315239295A4231523929524231000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000292929312931
312931313131312931313131425A734A31424229424A31424A31424A31424A31424A3142
4A31424A31424A31424A31424A314252314A4A31425231424A314252314A4A314252314A
52314A8C8C84848C7B848C7B848C7B848C7B848C7B848C7B848C7B848C7B848473848C73
848473848C737B84734A527B4A527B52527B4A527B52527B4A527B525A844A527B525A84
525A84525A84525A84525A84525A84846342845A42846342845A428463427B5A42845A42
7B5A42845A427B5A397B5A427B5A397B5A427B5A397B5A397B5A397B5A42735A39735A39
735A397B5A399C948C5273735273735A73735A42315A42315A42315A4231523929524231
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000292929313131292929
3131313129294A6373425A734A5A737B738C4A31424229424A31424229424A31424A2942
4A31424A294252314A4A294252314A4A314252314A4A314252314A4A314252314A4A3142
8C9484848C7B8C9484848C7B8C8C7B84847B8C8C7B848473848C7B848473848C7B848473
848C7B7B8473848C7B4A527352527B4A527B525A844A527B525A844A527B525A844A527B
525A844A527B525A84525A84525A8C525A848C6342845A398463427B5A398463427B5A39
8463427B5A398463427B5A398463427B5A397B63427B5A397B5A427352397B5A42735239
7B5A42735239A59C8C526B6B5A7373526B6B5A7373526B6B6342315A42295A4231523929
524231000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000292929312931312931313131
312931425A73425A73425A73425A737B7B8C4A31424A31424A31424A31424A31424A3142
4A31424A31424A31424A31424A314252314A4A314252314A4A314252314A52314A52314A
848C7B8C8C84848C7B848C7B848C7B848C7B848C7B848C7B848C7B848C7B848C73848C7B
8484738484738484738484734A527B4A527B4A527B52527B4A527B525A844A527B525A84
525A84525A84525A84525A84525A84525A84525A84846342846342846342846342846342
845A428463427B5A42845A427B5A397B5A427B5A397B5A427B5A397B5A397B5A397B5A39
735A397B5A399C9C8C527373526B6B52736B526B6B52736B526B6B5A42315A42315A4231
524229524231000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000029292931313129292931313139526B
425A73425A6B425A7342526B4A5A737B738C4A31424229424A314A4A29424A31424A2942
52314A4A314252314A4A294252314A4A314252314A4A294252314A4A314252314A848C7B
8C9484848C7B8C8C84848C7B8C8C84848C7B848C7B848473848C7B848473848C7B7B8473
848C7B7B8473848C737B8473525A7B4A527B525A7B4A527B525A844A527B525A844A527B
525A844A5284525A844A5284525A8C525A84525A8C525A848C6342845A428463427B5A39
8463427B5A398463427B5A398463427B5A398463427B5A398463427B5A397B5A42735A39
7B5A429C948CA59C8C526B6B5A7373526B6B52736B526B6B527373526B6B52736B524229
5A4231523929524231000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000029292931293131293131313142526B425A6B
425A6B425A73425A6B425A73425A737B7B944A31424A31424A31424A31424A31424A3142
4A31424A314A4A31424A314A4A314252314A4A314252314A52314A52314A848C84848C84
848C7B848C84848C7B848C7B848C7B848C7B848C7B848C7B848C7B848C7B848473848C73
848473848C737B84737B84734A527B4A527B4A527B4A527B4A527B525A844A527B525A84
525A84525A84525A84525A84525A84525A84525A84525A8C846342846342846342846342
846342846342845A42846342845A428463427B5A397B5A427B5A427B5A427B5A397B5A39
7B5A39A59C8C9C948C526B6B526B6B52736B526B6B526B6B526B6B526B6B526B6B526B6B
526B6B5A4231524229524231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000313131292929425A6B39526B425A7339526B
425A7339526B425A7342526B425A7373738C4A314A4A29424A314A4A29424A314A4A2942
52314A4A294252314A4A314252314A4A314252314A4A314252314A848C7B8C9484848C7B
8C8C84848C7B8C8C84848C7B848C7B84847B848C7B848473848C7B7B8473848C7B7B8473
848C7B7B8473848C7B7B84734A527B4A527B525A844A527B525A844A527B525A844A527B
525A844A527B525A844A5A84525A84525A8452638C525A8452638C8463428C6342845A42
846342845A398463427B5A398463427B5A398463427B5A398463427B5A398463427B5A39
7B5A429C948CA59C8C4A6B6B52736B4A6B6352736B4A6B63526B6B4A6B63526B6B4A6B63
526B6B4A6B635A42315239295A4231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031293139526B42526B39526B42526B42526B425A6B
42526B425A6B42526B425A6B42526B7B7B947B738C4A31424A31424A31424A31424A314A
4A31424A314A4A314252314A4A314252314A52314A52314A52314A8C8C84848C84848C84
848C7B848C7B848C7B848C7B848C7B848C7B848C7B848C7B84847B848C7B848473848C7B
7B84738484737B84737B84734A527B4A527B4A527B4A527B4A527B525A844A527B525A84
4A5284525A84525A84525A84525A84525A84525A84525A8C525A84525A8C846342846342
846342846342846342846342845A42846342845A428463427B5A397B5A427B5A397B5A42
7B5A399C9C8C9C9C8C526B6B526B6B526B6B526B6B526B6B4A6B63526B6B4A6B63526B6B
4A6B63526B634A6B63526B63524229524231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031313139526342526B395263425A6B39526B425A6B395263
425A6B39526B425A6B39526B425A737B738C7B7B9442294252314A4A31424A314A4A2942
52314A4A314252314A4A294252394A4A314252314A4A31428C9484848C7B8C8C84848C7B
8C9484848C7B848C8484847B848C84848C7B848C7B7B8473848C7B7B8473848C7B7B8473
848C7B7B8473848C737B846B525A7B4A527B52527B4A527B525A844A527B525A844A527B
525A844A5284525A844A5284525A8C525A84525A8C4A5A845A638C525A8452638C846342
8C6342846342846342845A39846342845A428463427B5A398463427B5A398463427B5A39
A59C949C9C8CA59C8C4A6363526B6B4A6B63526B6B4A6363526B6B4A6B63526B6B4A6363
526B6B4A6363526B634A6363526B6B4A6363524231000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000039526339526342526B39526342526B39526B42526B39526342526B
42526B425A6B39526B42526B42526B7B7B947B738C4A31424A31424A314A4A31424A314A
4A314A52314A4A314252314A52314A52314A52314A848C84848C84848C84848C7B848C84
848C7B848C84848C7B848C7B848C7B848C7B7B847B848C7B848C7B848C7B7B8473848473
7B8473848C737B84737B84734A527B4A527B4A527B4A527B4A527B525A844A527B4A5A84
525A84525A844A5A84525A84525A84525A84525A84525A84525A8C52638C525A8C525A8C
8463428C63428463428463428463428463428463428463428463428463427B5A397B5A42
9C9C8C9C9C8C9C9C8C526B6B4A6B63526B6B4A63634A6B634A6B63526B6B4A63634A6B63
4A6B63526B634A63634A6B634A63634A6B634A63634A6363000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000042526B394A6342526B394A6342526B39526342526B39526342526B395263
425A6B395263425A6B395263425A6B73738C7B7B944A294252314A4A314252314A4A3142
52314A4A314252314A4A314252314A4A31428C948C848C848C9484848C7B8C9484848C7B
848C84848C7B848C84848C7B848C7B7B847B848C7B7B8473848C7B7B8473848C7B7B8473
848C7B7B8473848C737B84734A527B4A527B525A7B4A527B525A844A527B525A844A527B
525A844A527B525A844A5A84525A844A5A84525A8C525A8452638C525A845A638C525A84
8C63428463428C63428463428C6342846339846342845A398463427B5A398463427B5A39
9C9C949C948CA59C8C4A6363526B6B4A6363526B6B4A6363526B6B4A6363526B6B4A6363
526B634A635A526B634A635A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000039526339526339526339526339526339526342526B
39526342526B39526B42526B3952637B7B947B738C4A314A4A31424A314A4A314A52314A
4A314A52314A52314A52314A52314A8C9484848C84849484848C84848C84848C84848C84
848C7B848C7B848C7B848C7B848C7B848C7B848C7B848C7B7B847B848C7B7B8473848C7B
7B84737B84737B84737B84734A527B4A527B4A527B4A527B4A527B4A527B4A527B525A84
4A5284525A84525A84525A84525A84525A84525A84525A8C525A84525A8C525A8C52638C
525A8C8C63428463428C6342846342846342846342846342846342846342846342846342
9C9C8C9C9C8C9C9C8C4A6B634A63634A6B634A63634A6B634A63634A63634A63634A6363
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000042526B394A63
42526B39526342526B395263425A6B73738C7B7B944A294252314A4A314252314A4A3142
52314A4A314A52314A4A314252394A848C848C9484848C7B8C9484848C7B848C84848C7B
849484848C7B848C847B847B848C847B847B848C7B7B8473848C7B7B8473848C7B7B8473
848C7B7B84737B84737B846B4A527B4A527B4A527B4A527B525A844A527B525A844A527B
525A844A527B525A844A527B525A8C4A5A84525A8C4A5A8452638C525A8452638C525A84
5A638C525A8C8C6B428463428C6B428463428C63428463398C634284634284634294948C
9C9C949C9C8C9C9C8C94948C526B6B4A63634A6B634A635A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000003952633952633952637B7B947B738C7B7B944A314A52314A4A314A52314A
52314A52314A52314A52314A848C84849484848C84848C84848C84848C84848C7B848C84
848C7B848C84848C7B848C7B848C7B848C7B7B847B7B8C7B7B8C7B848C7B7B84737B8473
7B84737B84737B84737B84734A527B4A527B4A527B4A527B4A527B4A5A844A527B4A5A84
4A5A84525A844A5284525A84525A84525A84525A84525A84525A84525A8C525A84525A8C
525A8C52638C525A8C8C63428C63428C63428463428463428463428463428463429C9C8C
9C9C8C9C9C8C9C9C8C9C9C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B9473738C7B7B9473738C52314A4A314252314A4A3142
52314A4A314A52394A848C848C948C848C848C9484848C84849484848C7B849484848C7B
848C847B8C7B848C847B847B848C7B7B847B848C7B7B8473848C7B7B8473848C7B7B8473
7B8C7B7B84737B8C737B84734A527B4A52734A5A7B4A527B525A844A527B525A844A527B
525A844A527B525A844A5284525A844A5A84525A8C4A5A84525A8C525A8452638C525A84
52638C525A8C5A638C525A8C8C6B428463428C6B428463428C63428463428C6342949C8C
9C9C94949C8C9C9C94949C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000073738C7B7B947B7B947B7B944A314A52314A4A314A52314A
52314A52314A84948484948C848C84848C84848C84848C84848C84848C84848C84848C84
848C7B848C7B848C7B848C7B7B8C7B848C7B7B8C7B7B8C7B7B84737B8C7B7B84737B8473
7B84737B84737B84737B84734A527B4A527B4A527B4A527B4A527B4A527B4A527B4A5A84
4A527B525A844A5A84525A844A5A84525A84525A84525A84525A84525A8C525A8C52638C
525A8C52638C52638C52638C52638C8C63428C63428C63428463428463428463429C9C8C
9C9C8C9C9C8C9C9C8C9C9C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B9473738C7B7B9473738C52394A4A314A52314A4A314A
52394A848C848C948C848C848C948C848C84849484848C7B849484848C7B848C847B8C7B
848C847B8C7B848C847B847B848C7B7B847B848C7B7B8473848C7B7B84737B8C7B7B8473
7B8C7B7B84737B847373846B4A527B4A527B4A527B4A527B525A844A527B4A5A844A527B
525A844A527B525A844A527B525A844A5A84525A844A5A8452638C525A8452638C525A84
5A638C525A8C52638C525A8C5A63948C63428C6B428463428C6B4A8463429C9C9494948C
9CA594949C8C9C9C8C949C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B73947B7B9473738C7B7B9452314A52314A4A314A52314A
52314A84948C848C84849484848C84849484848C84848C84848C84848C84848C7B848C84
848C7B848C847B8C7B848C7B7B8C7B848C7B7B847B7B8C7B7B847B7B8C7B7B84737B8473
7B84737B84737B84737B84734A527B4A527B4A527B4A527B4A527B4A5A7B4A527B4A527B
4A5A84525A844A5284525A84525A84525A844A5A84525A84525A84525A8C525A84525A8C
525A8C52638C525A8C52638C52638C5A63948C63428C6B428C63428C6B42949C8C9C9C8C
9C9C8C9C9C94949C8C9C9C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B9473738C7B7B9473738C52314A4A314A52394A4A314A
8C948C848C848C948C848C8484948C848C84849484848C848494847B8C7B848C847B8C7B
848C847B847B848C847B847B848C7B7B8473848C7B7B84737B8C7B7B84737B8C7B7B8473
7B8C7B7384737B84737384734A527B4252734A5A7B4A527B4A5A7B4A527B525A844A527B
525A844A527B525A844A5284525A844A5284525A8C4A5A84525A8C4A5A8452638C525A84
52638C525A845A638C525A8C5A6394525A8C5A63948C63428C6B428463429C9C94949C8C
9C9C94949C8C9CA594949C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000073738C7B7B947B7B947B7B947B7B9452314A52314A84948C
848C8484948C848C84849484848C84848C84848C84848C84848C84848C84848C7B848C84
7B8C7B848C7B7B8C7B848C7B7B847B7B8C7B7B847B7B8C7B7B84737B847B7B84737B8473
7B84737B84737B84737B84734A527B4A527B4A527B4A527B4A527B4A527B4A527B4A5A84
4A527B4A5A844A5A84525A844A5A84525A84525A84525A84525A84525A8C525A8C52638C
525A8C52638C525A8C52638C52638C52638C52638C5A63948C63428C6B42949C8C9C9C94
949C8C9C9C8C949C8C9C9C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B9473738C7B7B9473738C7B7B944A314A8C948C848C84
8C948C848C8484948C848C8484948C848C848494847B8C7B8494847B8C7B848C847B847B
848C847B8C7B848C7B7B847B848C7B7B847B7B8C7B7B84737B8C7B7B84737B8C7B738473
7B8C7B7384737B847373846B4A527B4A527B4A527B42527B4A5A844A527B4A5A844A527B
525A844A527B525A844A527B525A844A5A84525A844A5A8452638C4A5A8452638C4A5A84
52638C525A8C52638C525A8C5A6394525A8C5A6394525A8C5A6394949C8C9C9C94949C8C
9CA594949C8C9C9C94949C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000737B947B7B947373947B7B947B7B948C948C848C8C84948C
84948C84948C848C84848C84848C84849484848C84848C84848C84848C847B8C7B7B8C84
7B8C7B848C847B8C7B7B8C7B7B8C7B7B8C7B7B847B7B847B7B847B7B8C7B7B84737B8473
7B84737B84737384737B84734A527B4A527B4A527B4A527B4A527B4A527B4A527B4A527B
4A527B4A5A844A527B4A5A844A5A84525A844A5A84525A84525A84525A8C525A84525A8C
525A8C52638C525A8C52638C52638C52638C52638C526394949C949C9C94949C8C949C94
949C8C9C9C94949C8C949C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B9473738C7B7B947373947B7B947373947B7B94737B94
7B7B9C7B7B9484948C848C8484948C7B8C848494847B8C84848C847B8C7B848C847B8C7B
848C847B847B848C847B847B7B8C7B7B84737B8C7B7B84737B8C7B7384737B8C7B738473
7B847B7384737B84737384734A527B4252734A527B4A527B4A5A7B4A527B4A5A844A527B
525A844A527B525A844A527B525A844A5284525A8C4A5A84525A8C4A5A8452638C4A5A84
52638C949C949C9C94949C949C9C94949C8C9C9C94949C8C9C9C94949C8C9CA594949C8C
9CA594949C8C9CA594949C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007373947B7B94737B947B7B94737B947B7B947B7B947B7B94
7B7B947B7B947B7B947B7B9C7B7B947B7B9C7B7B947B849C7B7B9C7B849C7B8C7B7B8C84
7B8C7B7B8C7B7B8C7B7B8C7B7B847B7B8C7B7B847B7B8C7B7B84737B847B7B84737B8473
7384737B84737384737B84734A52734A527B4A527B4A527B4A527B4A527B4A527B4A5A84
4A527B4A5A844A5A844A5A844A5A84525A844A5A84949C9C949C94949C94949C94949C94
949C94949C94949C94949C94949C94949C94949C94949C94949C94949C94949C94949C94
949C8C949C8C949C8C949C8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B947373947B7B9473738C7B7B9C737B947B7B94737B94
7B7B9C7B7B947B7B9C737B9484849C7B7B947B849C7B7B9484849C7B7B9484849C7B7B94
84849C7B7B9C84849C7B7B9C8484A57B849C7B8C7B7384737B8C7B7384737B8C7B738473
7B8C7B7384737B8473737B6B4A527B4252734A527B42527B4A5A844A527B4A5A844A527B
949C9C949C9C949C9C8C9494949C9C949C94949C9C8C9494949C9C949C94949C9C949C94
94A59C949C94949C94949C949CA594949C94949C94949C8C9CA594949C8C949C94949C8C
9CA594949C8C94A594949C8C9CA594000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000737B947B7B947373947B7B947B7B947B7B94737B947B7B94
7B7B947B7B9C7B7B947B7B947B7B947B849C7B7B947B7B9C7B7B9C7B849C7B7B9C7B849C
7B849C84849C7B849C84849C84849C84849C7B849C84849C84849C8484A584849C8484A5
8484A5848CA57384737384734A527B4A527B949C9C949C9C949C9C949C9C949C9C949C9C
949C9C949C9C949C9C949C9C949C9C949C9C949C94949C9C949C94949C9C949C94949C94
949C94949C94949C94949C94949C94949C94949C94949C94949C94949C94949C8C949C94
949C9494A594949C8C949C8C949C8C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B9473738C7B7B947373947B7B947373947B7B9C737B94
7B7B9C737B947B849C7B7B947B849C7B7B9484849C7B7B947B849C7B7B9484849C7B7B9C
84849C7B7B9C84849C7B849C84849C7B849C8484A57B849C848CA57B849C848CA584849C
848CA584849C39426B3139636B7B6363735A949CA58C9C9C949CA58C949C949C9C8C9C9C
949C9C8C9C9C949C9C8C9C94949C9C8C9C94949C9C8C9C94949C9C8C9C9494A59C949C94
949C9C8C9C9494A59C949C9494A594949C9494A594949C9494A594949C8C94A594949C8C
94A594949C8C94A594949C8C94A594000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000737394737B94737B947B7B94737B947B7B947B7B947B7B94
7B7B947B7B947B7B947B7B9C7B7B947B7B9C7B7B9C7B849C7B7B9C7B849C7B7B9C7B849C
7B849C84849C7B849C84849C7B849C84849C84849C84849C84849C8484A584849C8484A5
31396331426331426331426B63736363736363735A63735A5A6B5A949C9C949C9C949C9C
949C9C949C9C949C9C949C9C949C9C949C9C949C9C949C9C949C94949C9C949C94949C9C
949C94949C94949C94949C94949C94949C94949C94949C94949C94949C94949C94949C94
949C8C949C94949C8C94A594949C8C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B947373947B7B947373947B7B9C737B947B7B9C737B94
7B849C737B947B7B9C737B947B849C7B7B947B849C7B7B9484849C7B7B9C84849C7B7B94
84849C7B7B9C84849C7B7B9C8484A57B849C8484A57B849C848CA531395A31426329395A
31426331396331426B313963637B6363735A6373635A6B5A6373635A6B5A636B5A8C9C9C
949C9C8C9C9C949C9C8C9C9494A59C8C9C94949C9C8C9C9494A59C8C9C94949C9C8C9C94
94A59C8C9C94949C948C9C9494A59C949C9494A5948C9C8C94A594949C8C94A5948C9C8C
94A594949C8C94A5948C9C8C94A594000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000737B947B7B94737394737B94737B947B7B94737B947B7B94
7B7B947B7B9C7B7B947B7B9C7B7B9C7B849C7B7B947B7B9C7B7B9C7B849C7B7B9C7B849C
7B849C84849C7B849C7B849C7B849C8484A57B849C31395A31395A31396331395A313963
3139633142633139633142636373636373635A6B5A63735A5A6B5A5A6B5A5A6B5A5A6B5A
5A6B525A6B5A8C9C9C949C9C949C9C949C9C8C9C94949C9C949C9C949C9C8C9C94949C94
949C94949C94949C94949C94949C94949C94949C94949C94949C9494A594949C94949C94
949C9494A594949C8C949C8C949C8C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B7B947373947B7B947373947B7B9C7373947B7B9C737B94
7B7B9C737B947B849C737B947B849C7B7B947B849C7B7B947B849C7B7B9484849C7B7B9C
84849C7B7B9C8484A57B849C8484A500000031395A29395A31395A29395A31426331395A
31426331396331426B3139636373635A6B5A6373635A6B5A63735A5A6B525A6B5A5A6B52
5A6B5A5263525A6B52000000949C9C8C9C9494A59C8C9C9494A59C8C9C9494A59C8C9C94
94A59C8C9C9494A59C8C9C9494A5948C9C9494A5948C9C9494A5948C9C8C94A5948C9C8C
94A5948C9C8C94A5948C9C8C94A594000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000737B94737394737B94737B947B7B94737B947B7B94737B947B7B9C
737B947B7B9C7B7B947B7B9C7B7B947B7B9C7B7B9C7B849C7B7B9C7B849C7B849C7B849C
7B849C7B849C00000000000000000000000000000031395A31395A31395A31395A313963
31396331426331396331426363735A63735A5A6B5A63735A5A6B5A5A6B5A5A6B5A5A6B5A
5A63525A6B52526352000000000000000000000000949C9C8C9C94949C9C8C9C94949C9C
8C9C94949C94949C94949C948C9C94949C94949C9494A5948C9C94949C94949C9494A594
949C8C94A594949C8C94A594949C8C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007373947B7B947373947B7B947373947B7B9C737B947B7B9C737B94
7B849C737B947B7B9C737B947B849C7B7B947B849C7B7B947B849C7B7B9C7B849C7B7B9C
00000000000000000000000000000000000000000029315231425A29395A31426329395A
3142633139633142633139636373635A6B5A63735A5A6B5A63735A5A6B525A6B5A526352
5A6B5A5263520000000000000000000000000000000000000000008C9C9494A59C8C9C94
94A59C8C9C9494A59C8C9C9494A59C8C9C9494A5948C9C9494A5948C9C9494A5948C9C8C
94A5948C9C8C94A5948C9C8C94A594000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000737B94737B94737B94737394737B94737B947B7B9C737B947B7B94
7B7B947B7B9C737B947B7B9C7B7B9C7B849C7B7B9C7B849C7B7B9C000000000000000000
00000000000000000000000000000000000000000000000031395A31395A31395A31395A
31396331426331396331426363735A63735A5A6B5A5A6B5A5A6B5A5A6B5A5A6B525A6B52
5A63525A6B52000000000000000000000000000000000000000000000000000000000000
8C9C9494A59C8C9C948C9C948C9C9494A5948C9C94949C948C9C9494A5948C9C94949C94
949C9494A5948C9C8C94A594949C8C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000737394737B947373947B7B9C7373947B7B9C737B947B7B9C737B94
7B7B9C737B947B849C737B947B849C737B947B849C000000000000000000000000000000
00000000000000000000000000000000000000000000000031395A29395A31426329395A
31426329395A3142633139636373635A6B5A63735A5A6B5A5A6B5A5263525A6B5A526352
5A6B5A000000000000000000000000000000000000000000000000000000000000000000
00000000000094A59C8C9C9494A59C8C9C9494A59C8C9C9494A5948C9C8C94A5948C9C8C
94A5948C9C8C94A5948C9C8C94A594000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000737B94737394737B94737B94737B94737B947B7B94737B947B7B9C
737B947B7B9C7B7B9C7B7B9C000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000031395A31395A31395A
31395A3139633139633142635A6B5A5A735A5A6B5A5A6B5A5A6B525A6B5A5A6B525A6B52
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008CA5948C9C9494A5948C9C948CA5948C9C9494A594
8C9C9494A5948CA59494A5948C9C8C94A594000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007373947B7B9C7373947B7B9C7373947B7B9C737B947B7B9C737B94
7B849C737B94000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000029395A31395A29395A
31426331395A31426329395A6373635A6B5A5A6B5A526B525A6B5A5263525A6B5A526352
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000094A5948C9C9494A5948C9C8C
94A5948C9C8C94A5948C9C8C94A5948C9C8C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000737B94737B94737B94737B94737B94737B947B7B9C737B94737B9C
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000029395A31395A
31395A31426331395A3139635A6B5A5A6B5A5A6B5A5A6B5A5A6B5A5A6B5A526352000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C9C948CA594
8CA59494A5948C9C8C8CA5948CA59494A594000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000737394737B9C7373947B7B9C7373947B7B9C000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000029395A
31426329395A31426331395A5A735A5A6B5A5A6B5A526B525A6B5A5263525A6B5A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C9C8C94A5948C9C8C94A5948C9C8C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000737B94737B94737B94737B94000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000031395A
31395A31395A31395A3142635A6B5A5A6B5A5A6B5A5A6B5A5263525A6B52000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008CA5948CA58C8CA594000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000737394000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31425A29395A31426329395A5A735A5A6B5A5A6B5A5263525A6B5A526352000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31395A31395A31395A31395A3139635A6B5A526B525A6B52526352000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029395A31426329395A3142635263525A6B5A526352000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031395A31395A31395A5A6B5A526B525A6B52000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031395A29395A3142635263525A6B5A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031395A31395A5A6B5A526352000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029395A314263526352000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031395A526B52000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[54]:=",
  ImageSize->{128, 128},
  AnimationDisplayTime->0.169,
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000200000P0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj2LW
9`00?g=cLda<C240<410Y/30`9T0IPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm?ooc020P800082006IVc3<cIP2009VIod:J
Yb8R8_mV<`00P00g??mVc5J0Zk75fBa1EkV>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7De3eDE:b/[:BT
Y>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEeMJ_0eLC4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7i/;2`W9bLP00
o`1o001oO`1oo`3o003oO`3oogl007l0Ogl0ogmo07moOgmoogoo07ooOgooool00?l0Ool0oomo0?mo
Oomooooo0?ooOoooo`0002XZ:UEEEGmoOjZZZ]GEeOoooh0700201`00P0L0080700031gT010L000<7
N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7>`000j21000k00@7
00031cX000>1VZ00?0041`000`Li0005Vh6iWj00>`041`000`Lh0007X862Wj2JX00j00@700031cP0
00R1Vh6iWj2JX3T010L000<7=`002X62PH:JX9nPVZ0h00@700031cH000b1PX6KPKVOX9nPWj0g00@7
00031cD000fJPX62PH:OX9nPVZ2J03L010L000<7=@003X:1PX6KPKVO^InPWj2O=P041`000`Ld000@
PX62PH:1PYnPWj2JX9ZPVSD010L000<7<`0048:1PX62PH:1^InPWj2OX9l2X3@010L000<7<P004j21
PX62PH:1PX6PWj2OX9ZPVZ00=0041`000`Lb000DPH:1PX62PI^1VinPWj2OX9nPWj0c00@700031c40
01J1PX62PH:1PX62PJ2OX9nPWj2JX9ZP<P041`000`L`000HPH:1PX62PH:1Vh6KWj2OX9nPWj2OX9nP
<@041`000`L_000JVX:1PX62PH:1PX62PJ2OX9nPWj2JX9ZPVX8`00@700031bl001Z2PH:1PX62PH:1
Vh6KWj2OX9nPWj2OX9nPWc0010L000<7;P0078:1PX62PH:1PX62PH:1X9nPWj2OX9ZPVZ2JX9X_00@7
00031b0000@IP@212@007h:1PX62PH:1PX62PH:1PYniWkVOX9nPWj2OX9nPWh401`001H40PH6O0240
10L000<78@001AV1Nh5k00H00222PH:1PX62PH:1PX62PH:1X9nPVZ2JX9ZPVZ2JX9ZPVPH000J109Z1
VX4Q00@700031b4000R16H5kPG^1N`800X4028:1PX62PH:11h804IniWkVO^InPWj2OX9nPWj2O0080
0h401WbOPIn1Wb4010L000<78@00;1V16H5kPG^1Nh5lPX62PH:1PX62PH:1PX62PJ2OX9ZPVZ2JX9ZP
VZ2JX9ZP0X402@2109n1VX6JP@0Q00@700031b4001B16H4IPGb1Nh5kPX62PH:1PX62P@N201:O^Ini
Wj2OX9nPWj2OX9nPWj03P@07Wh6OPIn1W`0Q00@700031b4003LIPAV16H4IPG^2O8:1PX62PH:1PX62
PH:1PX6PWj2JX9ZPVZ2JX9ZPVZ2JX9ZPPIn1Wh6OPIZ1024010L000<78@00584IPAV1O85lPX62PH:1
PX62PH:11h8079niWkVOX9nPWj2OX9nPWj2OX86JPIZ1Wh6OPIlQ00@700031b4003LIPAV1P9Z0VX22
O8:1PX62PH:1PX62PH:1PX6PWj2OX9ZPVZ2JX9ZPVZ2JX9Z1Nh5kPG^1Nh:1024010L000<78@0059Z1
VX6JPIZ1VX62PH:1PX62PH:11h8049niWkVOX9nPWj2OX9nPWj02W`0=PIZ1VX6JPIZ1VX6JP@0N00@7
00031a`004>JP9Z0VX2JP9Z0VX2JP9Z1PX62PH:1PX62PH:1PX62PJ2OX9ZPWj2JX9ZPVZ2JX9ZPNh5k
PG^1Nh5kPG^1Nh5kVW^J01X010L000<76@07VP0EPIZJVX6JPIZ1VX6JPH:1PX62PH:100N202RO^Ini
WkVOX9nPWj2OX9nPPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6J5@041`000`LD001AVX2JP9Z0VX2JP9Z0
VX2JP9Z0VX2JP9Z0PX62PH:1PX62PH:1PX62PJ2OX9nPVZ2JX9ZPVZ2JX7^1Nh5kPG^1Nh5kPG^JNiYk
VW^JNiYkVW^J00:P4P041`000`L@01ZJ00V1VYZ2PH:1PX402H80:YniWkVOX9nPWj2OX9nPPIZ1VX6J
PIZ1VX6JPIZ1VX6JPIZ1VX6JPJJPYQ<010L000<74000D:22X82JP9Z0VX2JP9Z0VX2JP9Z0VX2JP9Z0
VX2JP8:1PX62PH:1PX62PH:1X9nPWj2OX9ZPVZ2JX7^1NiYkPG^1Nh5kVW^JNiYkVWbJNiYk1j000inP
000?00@700031a4000FQQj67X@0IVP03PX6200Z2032O^IniWj2OX9nPWj21VX6JPIZ1VX6JPIZ1VX6J
PIZ1VX6JPJJPYZ2VX:JP^J2iX;T=00@700031`h004jPPJ22X8:PQj27VX2JP9Z0VX2JP9Z0VX2JP9Z0
VX2JP9Z0PX62PH:1PX62PH:1X9nPWj2OX9ZPVZ1kPG^JNiYkPG^1NiYkVW^JNiYkVW/9X00;Wj2OX9nP
Wj2OX9X02@041`000`L;000?X9^PVj2KX8NQQj67XHNQ01NJ00>2PH802880=IniWkVOX9nPWj21VX6J
PIZ1VX6JPIZ1VX6JPIZ1VZ2VX:JPYZ2VX:JP^J2iX;VP^J2iX;VP00P010L000<72002P@0kX86PPJ21
X86PQj22X8NPPZ22X82JP9Z0VX2JP9Z0VX2JP9Z0VX2JP8:1PX62PH:1PX6PWj2OX9nPVZ000YX047^J
NiYkVW^1O9YkVWbJNiX>X00<Wj2OX9nPWj2OX9n120041`000`L8000JPJ2KX9^PVj2KX8NQQj67XHNQ
Qj67XHNQVZ0AVP04PX62P@F200RO^InPWj2OX0:J00^1VX6JPIZ1VX6JP@02VP0MX9ZVX:JPYZ2VX:JP
YZ2VX:JP^J2iX;VP^J2iX;T02@041`000`L90017PJ21X86PPJ21X8:PPZ22X8:PPZ27X8NPQj27X82J
P9Z0VX2JP9Z0VX2JP8:1PX62PH:1X9nPWj2OX7^1Nh5kVW^JNiYkVW/05J002YnPWj2OX9nPWj0:00@7
00031`T001nPVj2KX9^PVj2KXHNQQj67XHNQQj6PXJ2QX:6PXJ2Q00fJ188019^2PYl5X00;PIZ1VX6J
PIZ1VX400Z007jJPYZ2VX:JPYZ2VX:JPYZ2VX:FP^J2iX;VP^J2iX;T02`041`000`L:0012X86PPJ21
X86PPJ22X8:PPZ22X8:PQj27X8NPQj27X8NPP9Z0VX2JP9Z0VX22PH:1PX6PWj2OX7^1NiYkVW^JNiYk
1J000hNPX00?X00:Wj2OX9nPWj2OX0`010L000<72`009J2KX9^PVj2KX8NQQj67XHNQQj67XJ2QX:6P
XJ2QX:6PXIZOVYl01iX4PP03VinP00:P02f1VX6JPIZ1VX6QX:6PYZ2VX:JPYZ2VX:JPYZ2VX:JPYZ2U
X;VP^J2iX;VP^J0030041`000`L<000kX86PPJ21X86PPZ22X8:PPZ22X8:PQj27X8NPQj27X8NPQj20
VX2JP9Z0VX22PH:1X9nPNh5kVW^JNiX01Z003hNPQj27X8NPQj27X8NPQ`03X00=Qj2PX9nPWj2OX9nP
W`0=00@700031`d002FPVj2KX9^PVj67XHNQQj67XHNQQj6PXHNQX:6PXJ2QQj67XHNQ00>J02^OVYZJ
Vh:KWj21VX6JPIZPXJ2QX:6PXJ2VX:JPYZ2VX:JPYZ2VX:JPYZ2V00>P00RiX;VP^J2iX0h010L000<7
3@00>h6PPJ21X86PPZ22X8:PPZ22X8:PPZ27X8NPQj27X8NPQj27X8NPQj20VX2JP8:1X7^1NiYkX8NP
Qj2700>P01F7X8NPQj27X8NPQj27X8NPQj27X8L00j001inPWj2OX9l03`041`000`L>001BVj2KX9^P
Vj27XHNQQj67XHNQQj67XHNQQj6PXJ2QQj67XHNQQj67XHNQVYnJVinJPJ6PXJ2QX:6PXJ2QX:6PYZ2V
X:JPYZ2VX:JPYZ2VX:JPYP>P00NiX;VP^J2i00l010L000<73`00FX6PPJ21X8:PPZ22X8:PPZ22X8:P
PZ27X8NPQj27X8NPQj27X8NPQj27X8NJPIn7X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NP
Qj2OX9nPWj2OX10010L000<74000F9^PVj2KX8NQQj67XHNQQj67XHNQQj6PXHNQX:67XHNQQj67XHNQ
Qj67XJ2JWj6PXJ2QX:6PXJ2QX:6PXJ2VX:JPYZ2VX:JPYZ2VX:JPYZ2VX;VP^J2iX;TA00@700031a40
0561X8:PPJ22X8:PPZ22X8:PPZ22X8:PQj27X8:PQj27X8NPQj27X8NPO86OViZPQj27X8NPQj27X8NP
Qj27X8NPQj27X8NPQj27X8NPQj27X9l00j000inP000A00@700031a4005JPVj2KX9^QQj67X8NQQj67
XHNQQj67XHNQQj67XHNQQj67XHNQQiZ1VZ2nPYnJWj2QX:6PXJ2QX:6PXJ2VX:JPYZ2VX:JPYZ2VX:JP
YZ2VX;VP^J2iX18010L000<74P00:Z22X8:PPZ22X8:PPZ22X8:PPZ22X8:PQj22X8:PQj27X8NPO85l
PInPW`:202:OVYnJWhNPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj271Z0C00@700031a<005:PVj2K
X8NQQj27X8NQQj27XHNQQj67XHNQQj67XHNQQiYlVX6JPKjP_X:7PYnJWiZOVZ6PXJ2QX:6PYZ2VX:JP
YZ2VX:JPYZ2VX:JPYZ2iX;VP50041`000`LD000XX8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ1l
PGb1O85lX9nPW`>200JJWiZOVYl2VP0IQj27X8NPQj27X8NPQj27X8NPQj27X8NPQ`02X003^J0001@0
10L000<75000Ci^PVj2KXHNPQj27X8NPQj27XHNQQj67XHNQQiZ1VX6JPIZ1_Z2nX;j2Qh:7VYnJWiZO
VYnJWj2QX:JPYZ2VX:JPYZ2VX:JPYZ2VX:FP^J005P041`000`LF000EX8:PPZ22X8:PPZ22X8:PPZ22
X8:P00:200MlPGb1O85l00:10j000inPW`05PP0MPInIWiVOVInIWh2PQj27X8NPQj27X8NPQj27X8L0
0Z000kVOP@0F00@700031aL00j0048NQQj67XHNQQj67XHNQQj47P@09VX6JPKjPYJ2U00:P01>2X8:P
PZ22WiZOVYnJWiZOVYnJ012P00>nX0005`041`000`LH000LX8:PPZ22X8:PPZ22X8:PPX5lPGb1O85l
PGb1O0RP0X8048N2Qh:7VInIWiVOVInIWiT4VP0<X8NPQj27X8NPQj270Z0I00@700031aP00Z0038NP
Qj67XHNQQj67X@V100VJPIZ1YJ2UX:D01:0068:PPZ22X8:PVYnJWiZOVYnJWiZOVYnJW`^P00>n0000
5`041`000`LI000IPZ22X8:PPZ22X7b1Nh5lPGb1O85lPGb1O009X00CWh:2Qh:7PXN2QiVOVInIWiVO
V@07VP0<VJ27X8NPQj27^IZ160041`000`LI0008VhNPQj27XHL<P@09VX6JPKjPYJ2U00>P022UX:22
X8:PPZ22X8:PVYnJWiZOVYnJWiZOVYnJWiZOVPJP00>JP9X05`041`000`LH000IP@2PPZ22X7^1Nh5l
PGb1O85lPGb1O85lP@0;X0:201>7PXN2Qh:7PXNIWiVOVInIWiZO00ZJ00R7X8NJNiYkPAH010L000<7
5`00285lPGbPQh5l2X403iZ1PH6JPIZPYJ2UX:FPY@04X00XPXN2X8:PPZ22X8:PVYnJWiZOVYnJWiZO
VYnJWiZOVYnJX86JPIZ0VQD010L000<75P006H40P@21Nh5kPGb1O85lPGb1O85lPGb1O8403:006Yn2
PXN2Qh:7PXN2Qh:OVYnIWiVOVYnJWiZO19X039VJVYZPQh2JNiYkPA@010L000<75@000h5lP@02P@03
PX5l00Z100>JPIX00h402IZPYJ2UX:FPY@06X00ZPZ22X8:PPZ22X8:PPYnJWiZOVYnJWiZOVYnJWiZO
VYnJ_hNPPIZ1VX2J4`041`000`LD0004P@2100:101>7Vh5lPGb1O85lPGb1O85lPGb100nP0X806HN2
Qh:7PXN2Qh:7PYnIWiVOVYnJWiZOVYl01YX02Z27PJ21VX2JNh4B00@700031a<000>1O8400X4018:7
PZ4;P@0AVX6JPIZ1YJ2UX:FPYJ2UX:D01:02Q`0ZPZ22X8:PPZ22X8:PPYnJWiZOVYnJWiZOVYnJWiZO
V[n7X8NPQiZ1VX2J4@041`000`LB000JP@2108:1Qh67Vh5lPGb1O85lPGb1O85lPG`?X00^Wj22Qh:7
PXN2Qh:7PXN2Qh:OVYnJWiZOVYnJWiZOVYnJ^J27PHN1Qh67P9YkPA0010L000<74@002h5lPH67PXN2
Qh:Q00^100BJPIZ10Z002JFPYJ2UX:FPY@06X0:702b2Qh:PPZ22X8:PPZ22XIZOVYnJWiZOVYnJWiZO
VYnP_hNPQj27X8NPQiZ0VPl010L000<74P02P@0GPX62PH:1Qh:1O85lPGb1O85lPGb1O8404Z00;X:7
PXN2Qh:7PXN2Qh:7PXNJWiZOVYnJWiZOVYnJWiZiX8N1Qh67PHN1Qh6JNiX>00@700031a0000elQh:7
PXN2Qh:7PZ6P00R100nJPIZ1VX6nX:FPYJ2UX:D02:02Q`0^PXN2X8:PPZ22X8:PPZ62X9ZOVYnJWiZO
VYnJWiZOX;n7X8NPQZ26X8JPQZ20VPd010L000<740006X:1PX62PH:1PVZ7Vj1lPGb1O85lPGb1O85l
4J000inPPP02PP0[Qh:7PXN2Qh:7PXN2X8:OVYnJWiZOVYnJWiZOX;jPQh67PHN1Qh67PHN1Q`02P@`0
10L000<73@02P@0>Qh67PHN1Qh67PHN2XJ0;P@0DVjFPYJ2UX:FPYJ2UX:FPYJ2UX:D2Q`0`PXN2Qh:7
PXN2Qh:7PZ22XH:hV[RJ^9ZOVYnJWiZoX;n1Qh67QXN1QhJ7PHMZPFZ12`041`000`L@000;PVZ2JX9Z
PVZ2JX800Z002Wb1O85lPGb1O84@X006Wj2OX9nP2H8088N2Qh:7PXN2Qh:OVYnJWiZOVYnJWj2nX8N1
Qh67PHN14`041`000`LE0008Qh67PHN2XJ0:P@0CYJ2UX:FPYJ2UX:FPYJ2UX:FPY@02X0:702B2Qh:7
PXN2Qh:7PXN2Qh:QPZ62^9ZhV[RJ^9ZOX;nP_j27QXLG00@700031aT000=ZPZ000Z00285lPGb1O85l
4J001YnPWj2OX0N201^7PX:2Qh:7PXN2Qh:7PYnJWiZOVYnJ_Z2nX;h06P041`000`LJ0004X:6PX@N1
01NUX:FPYJ2UX:FPYJ2UX:FPYJ2UX:FPY@02Q`0QPXN2Qh:7PXN2Qh:7PXN2XH:QPZ62^9ZhV[RJ^:2o
X;nP01X010L000<76P04X006PGb1O85l4J0029nPWj2OX9nP2H806HN2Qh:7PXN2Qh:7PXN2WiZOVYnJ
_Z2nX;h06P041`000`LJ0004X:6PX@F101NUX:FPYJ2UX:FPYJ2UX:FPYJ2UX:FPY@02X0:70262Qh:7
PXN2Qh:7PXN2Qh:QPZ62XH:QV[RJ^9ZoX;nP_j006P041`000`LJ00BP00F1O85lP@0DX006Wj2OX9nP
1h806hN2PX:7PXN2Qh:7PXN2Qh:PVYnJWj2nX;jP_P0J00@700031aX000BPXJ2Q18406:2UX:FPYJ2U
X:FPYJ2UX:FPYJ2UX:FPY@:P0XL08H:7PXN2Qh:7PXN2Qh:7PXR2XH:QPZ62XIZhV[nP_j2oX00J00@7
00031aX01J000gb1X00CX008Wj2OX9nPWj09PP0IQh:7PXN2Qh:7PXN2Qh:7PZ2JWj2nX;jP_P0J00@7
00031aX0022PXJ2QX86PYJ2UX:FPYJ2UX:FPYJ2UX:FPYJ2UX:FPY@:P0XL08H:7PXN2Qh:7PXN2Qh:7
PXR2XH:QPZ62XH:QX;nP_j2oX00J00@700031aX07J001YnPWj26X0^201N7PXN2Qh:7PXN2Qh:7PXNP
_Z2UX;jPY@0J00@700031aX001jPXJ2QX:6PXJ2QX:FPYJ2UX:FPYJ2UX:FPYJ2UX:D4X0:70262Qh:7
PXN2Qh:7PXN2Qh:8PZ6P_j2oX;nP_j2oX;nP_j006P041`000`LJ012P00>QX:002:0029nPWj26X8JP
2H806HN2Qh:7X:FPYJ2UX:FPYJ2UX:FPYJ2UX:D06P041`000`LJ000PX:6PXJ2QX:6PXJ2QX:6PXJ2Q
X:6PXJ2QX:FPYJ2UX:D2X00TPXN2Qh:7PXNP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j266@041`00
0`LJ012P03JQX:6PXJ2QX:6PXJ2QX:6PXHJPPZJPYZ2VX:JPYZ2VX:FPYJ2UX:FPYJ2UX:FPYJ2UX:FP
YH4I00@700031aX002>PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2VX:JPYZ22P@02X00Q_j2oX;nP
_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2601T010L000<76P000j27X00;X00HXJ2QX:6PXJ2QX:6P
XJ2QX:4IPVZ7PJ210Z007ZJPYZ2VX:JPYZ2UX:JPYJ2UX:FPYJ2UX:FPYJ2UPAT010L000<76P00:J2Q
X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QPZ21PVZ2PH:1X8JPQZ2600BP01VoX;nP_j2oX;nP_j2oX;nP
_j2oX;nP_j2601T010L000<76P00AZ22X8:QPZ67XHNQQj6PXJ2QX:6PXJ2QX00IPQV2JX9ZPVZ2PJ21
X86PPJ21W`2VX:JPYZ2UX:FPYJ2UX:FPYJ2UX:FPYH4I00@700031aT001J1X:6PXJ2QX:6PXJ2QX:6P
XJ2QX:6P1@004h62PH:1PX62PJ26X8JPQZ21X84010005KnP_j2oX;nP_j2oX;nP_j2oX;nPP@0I00@7
00031aT001@IX8:PPZ67XHNQQj67XJ2QX:6PX@L001:1JX9ZPVZ2JX:1X86PPJ21X847000CX:FPYJ2U
X:FPYJ2UX:FPYJ2UP@0I00@700031aT001:1X:6PXJ2QX:6PXJ2QX:6PXJ0:000APX62PH:1PX6PQZ26
X8JPPJ002@004KnP_j2oX;nP_j2oX;nP_j2101T010L000<76@0041VPPZ62XHNQQj67XHNQQj4<000@
6H9ZPVZ2JX:1X86PPJ21X0`000nPYJ2UX:FPYJ2UX:FPYH406@041`000`LI000=PJ2QX:6PXJ2QX:6P
X@0@000>PH:1PX62PJ26X8JPQZ0@000<X;nP_j2oX;nP_j2o6@041`000`LI000;6J62XH:QQj67XHL0
4P003X8IPVZ2JX:1X86PPJ214`002J2UX:FPYJ2UX00I00@700031aT000V1X:6PXJ2QX:405@0038:1
PX62PJ26X8JPPAH000NoX;nP_j2o01T010L000<76@001QVQPZ62XAT000^2JX9ZPX6PPJ21X00I0004
YJ2UX1T010L000<76@00186PXJ0K000:PH:1PX6PQZ21X1`000>P_`0060041`000`LI00036@0001d0
00UZPVZ2PJ21X840=`041`000`Li0008PX62PJ26X84h00@700031cX000J2JX:1X84i00@700031c/0
00F2PJ26X00i00@700031c/000@IPX6P>P041`000`Ll0003PJ2103X010L000<7?002PC/010L000<7
N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@7
00201`00P0L008070000\
\>"],
  ImageRangeCache->{{{0, 127}, {127, 0}} -> {-3.20067, -3.20067, 1.05828, 
  1.05828}},
  CellTags->"Animation"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
%%ImageSize: 128 128 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00744048 0.0238095 0.00744048 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
384 string
128 128 8 [128 0 0 128 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006B525A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000063525A6B5A63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000006B525A6B52636B525A6B52636B525A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000006B5A6363525A6B5A636B525A6B5A636B525A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313152947B6B000000000000000000000000000000000000000000000000000000
0000006B5A636B525A6B52636B52636B5A636B525A6B5A636B5A63000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A947B73947B6B947B738C736B000000000000000000000000000000000000
00000063525A6B5A636B525A6B5A636B525A735A636B525A735A636B525A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313152947B6B947B73947B6B947B6B947B6B947B6B000000000000000000000000
6B525A6B52636B52636B5A636B52636B5A636B5A636B5A636B5A636B5A636B5A63000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A3131528C7B6B948473947B6B947B738C736B947B738C7B6B000000000000
6B5A636B525A6B5A636B525A735A636B52636B5A636B525A735A636B5A63735A636B5263
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313152312952947B73947B6B947B73947B6B947B6B947B6B947B6B947B6B947B6B
6B52636B5A636B52636B5A636B5A636B5A636B52636B5A636B5A63735A636B5A63735A63
735A63000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A393152947B6B947B73947B6B947B73947B6B947B738C7B6B947B738C7B6B
947B738C736B6B5A636B52636B5A636B5263735A636B5263735A636B5263735A6B6B5A63
735A6B6B5A63000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000073634A6B5A4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313152313152947B73947B6B947B73947B6B947B73947B6B947B6B947B6B947B6B
947B6B947B6B947B6B947B6B6B52636B5A636B5A63735A636B5A63735A63735A63735A63
735A63735A63735A63000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000073634A73634A6B634A73634A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A31315231294A9C8473947B6B947B738C7B6B948473947B6B947B738C7B6B
947B738C7B6B947B6B8C736B947B738C7B6B735A636B5263735A6B6B5A63735A6B6B5A63
73636B735A63735A6B735A63000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000073634A7363526B5A4A7363526B5A4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313152312952313152947B73947B73947B6B947B73947B6B947B73947B6B947B6B
947B6B947B6B8C7B6B947B6B947B6B947B6B8C7B6B947B6B8C7B6B735A636B5A63735A63
735A63735A6B735A63735A6B735A63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
73634A73635273634A73634A73634A73634A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A393152312952948473947B6B948473947B6B9484738C7B6B9484738C7B6B
947B738C7B6B947B738C7B6B947B6B8C7B6B947B6B8C7B6B947B6B8C7363947B6B6B5A63
73636B735A6373636B735A637B636B735A63000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007B6B5273634A
7B6B5273634A7363526B634A7363526B5A4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000313152313152313152947B73947B73947B73947B73947B6B947B73947B6B947B73
947B6B947B6B947B6B947B6B947B6B947B6B947B6B947B6B8C7B6B947B6B8C7B6B947B6B
8C7B6B735A6B735A63735A6B735A6373636B735A6B000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007B6B52736352736352
73634A73635273634A73634A73634A73634A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A31315231294A39315A947B73948473947B6B948473947B6B947B738C7B6B
948473947B6B947B738C7B6B947B738C7B6B947B6B8C736B947B738C7B6B947B6B8C7363
947B6B8C7363947B6B735A637B636B735A637B636B735A63000000000000000000000000
00000000000000000000000000000000000000000073634A7B6B527363527B6B5273634A
7B6B5273634A7363526B634A7363526B634A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031315231294A313152313152948473947B73947B73947B73947B73947B6B947B73
947B6B947B73947B6B947B6B947B6B947B6B8C7B6B947B6B947B6B947B6B8C7B6B947B6B
8C7B6B947B6B8C7B638C7B6B8C7B6B7B636B735A6B7B636B7B636B000000000000000000
0000000000000000000000000000000000007B6B527B6B527B6B527B6B52736352736352
73635273635273634A73634A73634A73634A6B634A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A31315231294A393152947B739C8473947B6B948473947B6B948473947B6B
9484738C7B6B9484738C7B6B947B738C7B6B947B738C7B6B947B6B8C7B6B947B6B8C7B6B
947B6B8C7363947B6B8C7363947B6B8C7363947B6B735A6B7B636B735A6B000000000000
0000000000000000000000000000007B6B527B6B527B63527B6B5273634A7B6B5273634A
7B6B5273634A736B5273634A7363526B634A736352000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031314A31294A313152313152948473947B73948473947B73947B73947B73947B73
947B6B947B73947B6B947B73947B6B947B6B947B6B947B6B8C7B6B947B6B8C7B6B947B6B
8C7B6B8C7B6B8C7B6B8C7B6B8C7B6B8C7B6B8C7B6B8C7B6B8C73637B636B7B636B000000
000000000000000000846B5A7B6B527B6B527B6B527B6B527B6B527B6B52736B527B6B52
73634A73635273634A73635273634A73634A6B634A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A31315231294A393152312952948473947B6B9C8473947B739484738C7B6B
948473947B6B9484738C7B6B9484738C7B6B947B738C7B6B9484738C7B6B947B6B8C7B6B
947B6B8C7B6B947B6B8C7363947B6B8C7B63947B6B8C7363947B6B8C73638C7B6B8C7363
00000000000084735A7B6B5284735A7B6B527B6B527363527B6B527363527B6B5273634A
7B6B5273634A736B526B634A736B526B634A736352000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031315231294A313152313152313152947B73948473947B73948473947B73947B73
947B73948473947B6B947B73947B6B947B738C7B6B947B6B947B6B947B6B8C7B6B947B6B
8C7B6B947B6B8C7B6B8C7B6B8C7B6B947B6B8C7B6B8C7B6B8C7B6B8C7B6B8C73638C7B6B
8C7B63393131312931393131393131393131393131393131393131393131393131393131
3931318C73738C73738C73738C73738C73738C73738C73738C73739473738C73738C7373
8C73739473738C737394737394737373848C73848473848C73848C73848C73848473848C
73848C73848C73848473848C73848C7B848C738484000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A31315231294A3131523129529C8473947B73948473947B73948473947B6B
9484738C7B6B9484738C7B6B9484738C7B6B9484738C7B6B9484738C7B6B9484738C7B6B
947B6B8C7B6B947B6B8C7B6B947B6B8C7363947B6B8C7363947B6B8C7363947B6B8C7363
393131312929393131312931393131392931393131393131423139393131423939393131
8C949C8C73738C737B8C73738C737B8C737394737B8C737394737B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B947373947B7B947373000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031314A31294A313152312952313152948473948473947B73948473947B73948473
947B73948473947B73948473947B6B947B73947B6B947B738C7B6B947B6B8C7B6B947B6B
8C7B6B947B6B8C7B6B947B6B8C7B6B8C7B6B8C7B6B8C7B6B8C7B6B8C7B6B8C7B6B393131
31313139313139313139313139313139313139313139313139313139313184949484949C
8494948494948C73738C73738C73738C73738C73738C73738C73738C73738C7373947373
8C73739473738C73739473738C7373947B73947373947B73947373947B73947373947B73
947373947B73947B73947B73947373000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031294A31315229294A39315231295231315A947B739C847B947B73948473947B6B
948473947B739484738C7B6B9484738C7B6B9484738C7B6B9484738C7B6B9484738C7B6B
9484738C7B6B947B6B8C7B6B94846B8C7B6B947B6B8C7363947B6B8C7B63393131312929
393131393131393131312931393139393131393131393131423939848C948C949C848C94
8C949C848C948C737B8C7373947B7B8C737394737B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B947373947B7B8C7373
947B7B947373947B7B9473739C7B7B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031314A31294A313152313152313152313152948473948473948473947B73948473
948473948473947B73948473947B73948473947B6B947B73947B6B9484738C7B6B947B6B
947B6B947B6B8C7B6B8C7B6B8C7B6B947B6B8C7B6B8C7B6B8C7B6B393131312931393131
39313139313139313139313139313139313139313139313184949484949C848C94849494
8494948494948C73738C73738C737394737B8C73738C73738C7373947B7B8C73738C7373
947373947B7B8C7373947B73947373947B7B8C7373947B73947B73947B7B947373947B73
947B73947B7B947373947B73000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000422929
31314A29294A31315231294A31315231295239315A947B7394847B947B73948473947B73
948473947B73948473947B6B9484738C7B6B9484738C7B6B9484738C7B6B9484738C7B6B
9484738C7B6B9484738C7B6B947B6B8C7B6B947B6B8C7B6B947B6B312929393131312931
393131392931393139393131393139393131423939848C9484949C848C948C949C848C94
84949C848C94947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B947373
947B7B947373947B7B947373000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000422929
31294A31314A31294A313152312952313152313152948473948473948473948473948473
948473948473948473948473947B73948473947B73948473947B6B947B73947B6B947B73
8C7B6B947B6B8C7B6B947B6B8C7B6B947B6B8C7B6B947B6B313131393131393131393131
393131393131393131393131393131848C94848C94849494848C94849494848C94849494
848C94849494848C948C737B8C73738C737B8C7373947B7B8C737394737B8C7373947B7B
8C7373947B7B947373947B7B947373947B7B947B73947B7B947B73947B7B947B73947B7B
947B73947B7B947B73000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000422929
31314A31294A31315229294A3131523129523131523129529C847B948473948473947B73
9C847B947B739484738C7B6B948473947B739484738C7B6B9484738C7B6B9484738C7B6B
9484738C7B6B9484738C7B6B9484738C7B6B947B6B312929393131313131393131312931
39313939313139313139313184949C848C9484949C7B8C9484949C848C948494947B8C94
84949C848C948494948C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B947373947B7B8C7373947B7B947373947B7B947373
9C7B7B947B73423942000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31294A31314A31294A31315231315231315231295231315294847394847B948473948473
948473948473947B73948473948473948473947B73948473947B739484738C7B6B947B73
947B6B9484738C7B6B947B6B8C7B6B947B6B312931393131393131393131393131393131
393131393131393131848C94848C9484949C848C94848C94848C94849494848C94848C94
848C94849494848C948C737B8C737B947B7B8C73738C737B8C737B947B7B8C7373947B7B
947B7B947B7B8C7373947B7B947B7B947B7B947B73947B7B947B7B947B7B947B73947B7B
947B7B947B7B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31314A29294A31315231294A31315231294A3131523129529C847B947B739C847B947B73
94847B947B73948473947B73948473947B6B9484738C7B6B9484738C7B6B9484738C7B6B
9484738C7B6B9484738C7B6B948473312931393131312931393131392931393139393131
39313939313184949C7B8C9484949C7B8C9484949C7B8C9484949C7B8C9484949C7B8C94
8494947B8C948494948C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B947373947B7B947373947B7B947B73
947B7B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31294A31314A31294A313152312952313152313152313152948473948473948473948473
948473948473948473948473948473948473948473948473947B73948473947B73948473
947B6B948473947B6B948473313131393131393131393131393131393131393131393131
7B8C94848C947B8C94848C947B8C94848C947B8C94848C947B8C94848C94848C94848C94
848C94848C94848C948494948C737B947B7B8C737B947B7B8C737B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005A3931
31314A29294A31315229294A31315231295231315231295239315A94847394847B947B73
9C847B948473948473947B7394847B947B739484738C7B6B948473947B739484738C7B6B
9484738C7B6B9484738C7B6B3931313131313931313129313931393931313931397B8C94
84949C7B8C94848C9C7B8C9484949C7B8C94848C947B8C9484949C7B8C948494947B8C94
8494947B8C948494947B8C8C947B7B8C737B947B7B8C7373947B7B8C737B947B7B8C7373
947B7B947B7B947B7B8C7373947B7B947B7B947B7B8C73739C7B7B947B7B947B7B947B73
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005A42395A39395A3939
5A393931314A31294A31314A31315231315231295231315231315294847B948473948473
94847394847B948473948473948473948473947B73948473948473948473947B73948473
947B739484738C7B6B3931313931313931313931313931313931313931317B8C947B8C94
7B8C94848C947B8C947B8C947B8C94848C947B8C947B8C947B8C94848C947B8C94848C94
7B8C94848C947B8C94848C948C7B7B947B7B8C737B947B7B947B7B947B7B8C7B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006342395A39316342395A39316342395A3939
6342395A393131315229294A31315231294A31315231295231315A9484739C847B948473
9C847B9484739C847B948473948473947B73948473947B739484738C7B739484738C7B6B
9484738C7B6B3931313129313931313931313931393931313931397B8494848C9C7B8C94
848C9C7B8C94848C9C7B8C94848C947B8C94848C947B8C948494947B8C948494947B8C94
8494947B8C8C8494947B8C8C947B7B8C737B947B7B8C737B947B7B8C737B947B7B8C737B
947B7B8C737B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B9C847B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005A39395A39395A39395A42395A39395A42395A3939634239
5A39396342395A393931315231294A31315231315231315231315294847B94847394847B
948473948473948473948473948473948473948473948473948473948473948473948473
947B733931313931313931313931313931313931317B8C947B8C947B8C947B8C947B8C94
7B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C94
7B8C947B8C947B8C94848C947B8C94947B7B947B7B947B7B8C7B7B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A39396342395A39316342395A39396342395A39396342395A39396342395A3931
6342395A393963423929294A31315231294A31315231295239315A3131529C847B948473
9C8C7B94847394847B947B739C847B9484739484738C7B7394847B947B739484738C7B6B
3931393131313931313129313931393931317B8C947384947B8C9C7B8C947B8C947B8494
7B8C9C7B8C947B8C947B8494848C947B8C947B8C947B8C8C8494947B8C947B8C947B8C8C
8494947B8C94848C947B8C8C8494948C737B947B7B8C737B947B848C7B7B947B7B8C737B
947B84947B7B947B7B8C7B7B948484947B7B947B7B947B7B9C8484947B7B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005A4239
5A39396342395A39395A42395A42396342395A39396342396342396342395A3939634239
6342396342395A393963423931314A31315231295231315231315231315A94847394847B
94847394847B948473948473948473948473948473948473948473948473948473393131
3931313931313931313931313931317B8C947B84947B8C947B8C947B8C947B8C947B8C94
7B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C94
7B8C947B8C947B8C947B8C947B8C94947B7B8C7B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006342395A39396342425A3939
6342395A39396342395A39396342395A39396342425A39396342395A39396342395A3939
6342395A393963424263393963423929294A31315231295231315A3129529C8C7B948473
9C847B9484739C847B94847394847B94847394847B9484739484738C7B73948473312931
3931313131313931393931317B8C9C7384947B8C9C7384947B8C947384947B8C94738494
7B8C947384947B8C947B84947B8C947B848C7B8C947B8C947B8C947B8C8C7B8C947B8C8C
7B8C947B8C8C848C947B8C8C7B8C948C737B947B848C7B7B947B848C7B7B947B84947B7B
947B84947B7B948484947B7B948484947B7B9C8484947B7B9C8484000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005A39396342395A39396342395A4239634239
5A3939634239634239634239634239634239634239634239634239634239634239634239
63423963423963423963423963423963423931315231315231315231315A94847B94847B
94847394847B94847394847B948473948473948473948473948473948473393131393131
3931313931317384947B8C947384947B8C947B8C947B8C947B8C947B8C947B8C947B8C94
7B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C94
7B8C947B8C947B8C947B8C947B8C8C7B8C94947B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005A39396342425A39396342425A39396342425A39396342395A3939
6342426339396342425A39396342426339396342395A3939634242634239634242633939
6342426342396342396339396B424263423931315231295239315A31295231315A948473
9C8C7B9484739C847B9484739C8C7B94847394847B94847394847B313131393131312931
3931397384947B8C947384947B8C9C7384947B8C947384947B8C947384947B8C9473848C
7B8C947384947B8C9473848C7B8C9473848C7B8C9473848C7B8C947B8C8C7B8C9473848C
7B8C947B8C8C7B8C9473848C7B94947B8C8C947B848C7B7B948484947B7B947B848C7B7B
948484947B7B948484947B7B9C8484947B7B948484393139000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000006342395A3939634239634239634239634239634239634239634239633939634239
634239634239634239634239634239634239634239634239634239634239634239634239
6342396342396342396342396342396B423963423931315231315231315A31315294847B
94847B9C847B94847394847B94847394847B948473948473393131393131393131393131
7384947B8C947384947384947384947B8C947384947B8C94738C947B8C947384947B8C94
7B8C947B8C947384947B8C947B8C947B8C94738C8C7B8C947B8C947B8C947B8C8C7B8C94
7B8C947B8C947B8C8C7B8C947B8C947B8C94947B7B947B7B947B7B947B84947B7B947B7B
947B7B947B84947B7B947B7B947B7B948484947B7B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006342425A3939
6342425A39396342425A39396342425A39396342426339396342425A3939634242633939
6342426339396342426342396342426339396B42426342396B42426342396B4242634239
6B42426342396B42426342396B42426342396B424263423931315231295239315A948473
9C8C7B9484739C8C7B948473948C7B948473948C7B3129313931393131317B8C9C738494
7B8C947384947B8C947384947B8C947384947B8C947384947B8C9473848C7B8C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C7B8C94738C8C7B8C94738C8C948484947B7B9484848C7B7B948484947B7B
948484947B7B9C8484947B7B9C8484947B7B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000634239634239634239634239634242
634239634239634239634239634239634239634239634242634239634239634239634239
6342396342396342396342426342396342396342396342396342396342396342396B4242
6342396B42396342396B42396342396B42396342396B42423131523131523131529C8C7B
94847B94847B94847B94847B94847394847B393131393131393131738494738494738494
738494738494738494738C94738494738C947384947B8C94738494738C947384947B8C94
7384947B8C94738C947B8C947384947B8C94738C947B8C94738C8C7B8C947B8C8C7B8C94
738C8C7B8C947B8C8C7B8C947B8C8C7B8C947B8C8C947B84947B7B947B84947B7B947B84
947B7B947B84947B7B948484947B7B948484000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000005263846342425A39396342426342396342425A3939
6342426342396342426339396B42426342396342426339396B4242634239634242633939
6B42426342396B42426342396B42426342396B42426342396B4A426342396B4242634239
6B4A426342396B42426342396B4A426342396B42426342396B4A4231295231315A312952
9C8C7B94847B9C8C7B9484739C8C7B9484733931313129317B8C9C738494738494738494
7B8C94738494738C9473848C7B8C94738494738C9473848C7B8C947384947B8C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C7B8C9473848C7B8C94738C8C7B8C948C7B7B948484947B7B948484947B7B
9C8484947B7B948484947B7B9C8484000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000526384526B84526B845A6B84526384526B84
526B845A6B8C6342396342426342396342426342396342426342396B4242634239634242
6342396B42426342396342426342396B42426342396B42426B42396B42426342396B4242
6B42396B42426342396B42426B42396B42426B42396B42426B42396B424231315231315A
94847B9C8C7B94847B94847B94847B393131393131738494738494738494738494738494
738494738494738494738494738494738C94738494738494738494738C94738494738494
738494738C9473848C738C94738C947B8C9473848C738C94738C8C7B8C9473848C738C8C
738C8C7B8C94738C8C7B8C8C7B8C8C7B8C94738C8C947B84947B84948484947B7B947B84
947B84948484947B7B948484948484000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005263845A6B845263845A6B84526384
5A6B8C5263845A6B8C526B845A6B8C526B845A6B8C526B845A6B8C526B846B4242634239
6B42426342396B4A426342396B42426342396B4A426342396B4A426342396B4A42634239
6B4A426342396B4A426B42396B4A426342396B4A426B42396B4A426B42396B4A42313152
9C8C7B94847B9C8C7B9484733931396B84947384947384947384946B8494738C94738494
738C9473848C738C94738494738C9473848C738C9473848C738C9473848C7B8C9473848C
738C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C94947B7B948484947B7B9C8484947B7B
9C8484947B7B9C8484947B7B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000526384526B84526384526B84
526B84526B84526B845A6B84526B845A6B84526B845A6B8C5A6B845A6B8C5A6B845A6B8C
5A6B845A6B8C5A6B8C5A738C5A6B8C5A738C6B42426B42426B42396B42426B42426B4242
6B42396B42426B42426B42426B42396B42426B42426B4A426B42396B42426B424231315A
94847B948C7B94847B393131738494738494738494738494738494738494738494738494
738494738494738494738494738494738494738494738494738494738494738494738C94
73848C73849473848C738C9473848C738C9473848C738C9473848C738C9473848C738C94
73848C738C8C738C8C7B8C8C738C8C738C8C738C8C948484947B84948484947B84948484
947B84948484947B84634231634231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005263845A6B84526384
5A6B845263845A6B845263845A6B8C526B845A6B8C5263845A6B8C526B845A6B8C526B84
5A738C5A6B845A738C526B845A738C5A6B845A738C5A6B8463738C5A6B8C5A738C5A6B8C
63738C5A738C6B4A426B4239734A426B4239734A426B4239734A426B4239734A426B4239
39315A94847B3931396B7B947384946B84947384946B7B8C7384946B84947384946B848C
738C946B848C7384946B848C738C9473848C738C946B848C738C9473848C738C946B848C
738C9473848C738C9473848C7B8C9473848C738C9473848C7B8C9473848C738C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C8C73848C9C8484947B84948484947B7B
9C8484947B849C84845A4231634A315A4231634231000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000526384526B84
526384526B84526384526B84526B845A6B84526B845A6B84526B845A6B84526B845A6B84
5A6B845A6B8C5A6B845A6B8C5A6B8C5A738C5A6B845A6B8C5A738C5A738C5A6B8C5A738C
5A738C5A738C5A738C5A738C5A738C63738C5A738C63738C5A738C6373946B42396B4A42
6B4A423931316B84947384946B84947384946B84947384947384947384946B8494738494
7384947384946B848C73849473849473849473848C73849473849473849473848C738494
738494738C9473848C73849473848C738C9473848C73848C73848C738C9473848C738C8C
73848C738C8C73848C738C8C738C8C738C8C73848C738C8C947B84948484947B84948484
9484849C8484634231634231634231634231634231634231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000052637B
526B8452637B5A6B845263845A6B845263845A6B8C5263845A6B8C5263845A6B8C526B84
5A6B8C526B845A738C526B845A738C526B845A738C5A6B845A738C5A6B845A738C5A6B84
5A738C5A6B8463738C5A6B8C63738C5A738C6373945A738C6373945A738C637B945A738C
637B94313963845A527B524A7B5A4A6B7B8C7384946B848C7384946B848C7384946B848C
7384946B848C7384946B848C7384946B848C738C946B848C7384946B848C738C946B848C
738C946B848C738C9473848C738C946B848C738C9473848C738C9473848C738C9473848C
738C9473848C738C9473848C738C9473848C7B8C9473848C948484947B849C8484947B84
9C8484947B846B4A396342316B4A31634231634A31634231634A315A4231000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
526384526B84526384526B84526384526B84526B845A6B84526B845A6B84526B845A6B84
526B845A6B845A6B845A6B8C5A6B845A6B8C5A6B845A6B8C5A6B845A738C5A6B8C5A738C
5A6B8C5A738C5A738C5A738C5A738C5A738C5A738C5A738C5A738C5A738C5A738C63738C
948C84948C847B524A7B5A4A7B524A7B5A4A7B524A7B5A4A7B524A7B5A4A7B524A738494
6B84947384946B84947384946B848C73849473848C7384946B848C73849473848C738494
73848C73849473848C73849473848C73848C73848C73848C73848C73848C73848C738C8C
73848C738C8C73848C738C8C73848C738C8C73848C738C8C947B84948484948484948484
9484846B4A31634A316B4A31634231634A31634231634A31634231634231634231000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000052637B526B8452637B5A6B845263845A6B8452637B5A6B84526B845A6B84526384
5A6B8C526B845A6B8C526B845A738C526B845A738C526B845A738C5A6B845A738C526B84
5A738C5A6B845A738C5A6B8463738C5A6B8C5A738C5A6B846373945A738C63738C312931
9C948C948C8431396B7B524A845A527B524A845A4A7B524A845A527B524A845A4A7B524A
845A527B524A845A4A7B524A845A4A6B848C7384946B848C738C946B848C7384946B848C
738C946B848C7384946B848C738C946B848C738C946B848C738C9473848C738C8C6B848C
738C9473848C738C8C6B848C738C9473848C738C8C738484738C8C947B849C8484947B84
9C848C6B42316B4A316342316B4A396342316B4A316342316B4A31634231634A315A4229
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000052637B526384526384526B84526384526B84526B845A6B84526B84526B84
526B845A6B84526B845A6B845A6B845A6B8C526B845A6B845A6B845A738C5A6B845A6B8C
5A6B8C5A738C5A6B845A738C5A738C5A738C5A738C5A738C5A738C5A738C313131948C8C
948C8494948C31396331396B7B5A4A845A4A7B524A7B5A4A7B5A4A845A4A7B524A7B5A4A
7B5A4A845A4A7B524A7B5A4A7B5A4A845A4A7B524A7B5A4A7B5A4A845A4A7B524A73848C
73848C7384946B848C73848C73848C73848C6B848C73848C73848C73848C73848C73848C
73848C738C8C73848C73848C73848C738C8C73848C738C8C73848C9C8484948484948484
6B4A316B4A316B4A316B4A316B4A316B4A316342316B4A31634231634A31634231634231
634231634A31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000052637B526B8452637B5A6B8452637B5A6B8452637B5A6B84526384
5A6B845263845A6B84526B845A6B8C526B845A738C526B845A738C526B845A738C526B84
5A738C526B845A738C5A6B845A738C5A6B845A738C5A6B8463738C31293194948C948C84
94948C948C8494948C31396339426B7B524A845A527B524A845A527B524A845A527B524A
845A527B524A845A527B524A845A4A7B524A845A4A7B524A845A4A7B524A845A4A7B524A
845A4A7B5242845A4A7B524A845A4A6B848C738C946B848C738C8C6B848C738C8C6B848C
738C8C6B848C738C8C6B848C738C8C6B848C738C8C6B848C738C8C947B849C848C6B4A31
734A396B4A31734A396B42316B4A396342316B4A396342316B4A316342316B4A31634231
6B4A31634231634A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000052637B526384526384526B84526384526B84526B84526B84
526B84526B84526B845A6B84526B845A6B84526B845A6B84526B845A6B845A6B845A6B8C
5A6B845A6B8C5A6B845A738C5A6B845A738C5A738C5A738C313131313131948C8C94948C
948C84948C8C948C8431396B31396331396B845A4A845A4A7B5A4A845A4A845A4A845A4A
7B5A4A845A4A7B5A4A845A4A7B5A4A845A4A7B5A4A845A4A7B5A4A845A4A7B5A4A845A4A
7B5A4A845A4A7B5A4A845A4A7B5A4A845A4A7B5A4A845A4A7B5A4A7B5A4A7B5A4A845A4A
6B848C73848C73848C73848C73848C73848C73848C738C8C73848C948484948484734A39
6B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A31634A316B4A31
634231634A31634231634A31634231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000052637B526B8452637B5A6B8452637B526B8452637B
5A6B845263845A6B845263845A6B84526B845A6B84526B845A738C526B845A6B8C526B84
5A738C526B845A738C526B845A738C5A6B845A738C312931393139948C8494948C8C8C84
94948C948C8494948C8C8C8439426B31396339426B7B524A845A527B5A4A845A527B524A
845A527B5A4A845A527B524A845A527B5A4A845A4A7B524A845A527B5A4A845A4A7B524A
845A527B5A4A845A4A7B524A845A4A7B5A4A845A4A7B5242845A4A7B5A4A845A4A7B5242
845A4A7B5242845A4A7B5242845A4A7B5242738C8C6B8484738C8C6B84847352396B4A31
7352396B4A31734A396B4A31734A396B4A316B4A396B42316B4A396B4A316B4A31634231
6B4A396342316B4A316342316B4A39634231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000052637B526384526384526B8452637B526B84
526B84526B84526B84526B84526B845A6B84526B84526B84526B845A6B84526B845A6B84
5A6B845A6B8C526B845A6B845A6B845A738C313131313131948C8C94948C948C84948C8C
948C8C94948C948C84948C8431396331396B31396B31396B845A4A845A52845A4A845A4A
845A4A845A52845A4A845A4A845A4A845A52845A4A845A4A845A4A845A4A7B5A4A845A4A
845A4A845A4A7B5A4A845A4A845A4A845A4A7B5A4A845A4A845A4A845A4A7B5A4A845A4A
845A4A845A4A7B5A4A845A4A845A4A845A4A7B5A4A845A4A7B5A4A845A4A637B7B637B7B
734A31734A396B4A31734A316B4A31734A396B4A316B4A316B4A316B4A316B4A316B4A31
6B4A316B4A31634231634A31634A316B4A31634231000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000052637B526B8452637B526B8452637B
526B8452637B5A6B8452637B5A6B8452637B5A6B84526B845A6B84526B845A6B8C526B84
5A6B8C526B845A738C526B845A738C31293139313931313194948C8C8C8494948C8C8C84
94948C8C8C8494948C8C8C8431426B31396339426B31396B39426B845A4A846352845A4A
845A527B5A4A845A52845A4A845A527B5A4A845A52845A4A845A527B5A4A845A527B5A4A
845A527B5A4A845A527B5A4A845A4A7B5A4A845A4A7B5A4A845A4A7B5A4A845A4A7B5A4A
845A4A7B5A42845A4A7B5A4A845A4A7B5A42845A4A7B5A42845A4A94849463847B637B7B
63847B5A7B7B63847B5A7B73637B7B6B4A31734A396B4A31734A396B4A31734A396B4A31
6B4A396342316B4A396342316B4A316342316B4A31634231634A31000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000052638452637B52638452637B526B84
52637B526B84526B84526B84526B84526B84526B84526B84526B84526B84526B845A6B84
526B845A6B84526B845A6B8431313131313131313194948C948C8C94948C948C8C94948C
948C8494948C948C8494948C948C8431396B31396B31396B31396B31426B845A4A845A52
845A4A845A52845A4A845A52845A4A845A52845A4A845A52845A4A845A4A845A4A845A4A
845A4A845A4A845A4A845A4A845A4A845A4A845A4A845A4A845A4A845A4A845A4A845A4A
845A4A845A4A845A4A845A4A845A4A845A4A845A4A845A4A948C94948C94948C94637B7B
637B7B637B7B637B7B637B7B5A7B73637B7B637B73637B7B5A7B736B4A316B4A316B4A31
6B4A316B4A316B4A316B4A316B4A316B4A31634A316B4A31634231634A31000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000526B8452637B526B844A637B
526B8452637B526B8452637B5A6B8452637B526B8452637B5A6B84526B845A6B8452637B
5A7384526B845A6B84526B843131393131313131398C8C8494948C948C8494948C8C8C84
94948C8C8C8494948C8C8C8494948C31396331426B31396339426B31396B394273845A4A
8C6352845A4A846352845A4A8C6352845A4A845A52845A4A8C6352845A4A845A52845A4A
846352845A4A845A527B5A4A846352845A4A845A4A7B5A4A846352845A4A845A4A7B5A4A
846352845A4A845A4A7B5A4A84634A845A4A845A4A7B5A42948C9C948C94948C9C5A7B73
637B7B5A7B73637B7B5A7373637B7B5A7B73637B7B5A7373637B7B5A7373637B735A7373
637B7B5A73736B4A396B42316B4A396B4A316B4A316342316B4A396342316B4A31634231
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000052638452637B52637B
52637B526B8452637B526B84526B84526B8452637B526B84526B84526B84526B84526B84
526B845A6B84526B84313131313131313131948C8C94948C948C8C94948C948C84948C8C
948C8C94948C8C8C84948C84948C8431396B31396331396B31396B39426B31396B845A52
845A52846352845A4A845A52845A4A845A52845A4A845A52845A4A845A52845A4A845A4A
845A4A845A52845A4A845A4A845A4A845A52845A4A845A4A845A4A845A4A845A4A845A4A
845A4A845A4A845A4A845A4A845A4A845A4A845A4A948C94948C94948C9C948C94637B7B
5A7B7B637B7B5A7B735A7B735A7B73637B7B5A73735A7B735A7B73637B735A73735A7B73
5A7B735A7B735A73735A7B735A73735A7B735A736B6B4A316B4A316B4A31634A31634A31
634A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000526B844A637B
526B844A637B526B8452637B526B8452637B526B8452637B526B8452637B5A6B8452637B
5A6B84526B7B3131393129313131398C8C8C94948C8C8C8C94948C8C8C8494948C8C8C84
94948C8C8C8494948C8C8C8494948C8C8C8439426B31396339426B31396B39427331396B
8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A
8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A846352845A4A846352845A4A
846352845A4A846352845A4A84634A845A4A948C9C8C8494948C9C948C94948C9C5A7373
637B7B5A7373637B7B5A7373637B7B5A7373637B7B5A7373637B735A7373637B735A7373
637B735A736B637B735A736B5A7B735A736B5A7B735A736B5A7B735A736B5A7B735A736B
6B4A31634231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000052637B
52637B52637B52637B526B8452637B526B8452637B526B8452637B526B84526B84526B84
526B8431313131313131313131313194948C948C8C94948C948C8C94948C948C8C94948C
8C8C84948C8C8C8C8494948C8C8C84948C8431396331396B31396B31396B31396B31426B
31396B8C6352845A528C6352845A52846352845A52846352845A4A846352845A4A846352
845A4A845A52845A4A846352845A4A845A52845A4A846352845A4A845A4A845A4A845A4A
845A4A845A4A845A4A845A4A845A4A845A4A948C94948C9C948C94948C9C948C94948C9C
5A73735A7B735A73735A7B735A73735A7B735A73735A7B735A73735A73735A73735A7B73
5A73735A73735A73735A73735A736B5A73735A736B5A736B5A736B5A736B5A736B5A736B
5A736B5A736B5A736B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
526B844A637B526B844A637B526B8452637B526B844A637B5A6B8452637B526B8452637B
3131393129313131313129319494948C8C8C94948C8C8C8494948C8C8C8C94948C8C8C84
94948C8C8C8494948C8C8C8494948C8C8C8494948431396339426B31396B39426B31396B
39427331396B8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A
8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A
8C6352845A4A8C6352845A4A8C63528C8C94948C9C8C8494948C9C948C94948C9C8C8C94
637B7B5A73735A7B735A7373637B735A73735A7B735A736B637B735A73735A7B735A736B
5A7B735A736B5A737352736B5A7B735A736B5A737352736B5A737352736B5A736B526B6B
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000052637B4A637B52637B52637B526B8452637B52637B52637B526B8452637B313131
31313131313131313131313194948C94948C8C8C8C94948C948C8C94948C8C8C8C948C8C
948C8C94948C8C8C84948C848C8C8494948C8C8C8431396331396B31426B31396B31426B
31396B39427331396B8C63528C63528C6352845A528C63528463528C6352845A4A8C6352
8463528C6352845A4A8C6352845A528C6352845A4A84635284634A8C6352845A4A84634A
845A4A8C6352845A4A845A4A948C94948C9C8C8C94948C9C948C94948C9C948C94948C9C
5A73735A7B735A73735A73735A73735A73735A73735A73735A73735A73735A736B5A7373
5A736B5A73735A736B5A736B5A736B5A736B5A736B5A736B5A736B5A736B000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000526B844A637B526B844A637B526B844A637B526B8452637B313131312931
3131313129313931398C8C8C9494948C8C8C94948C8C8C8C94948C8C8C8494948C8C8C84
94948C8C8C8494948C8C8C8494948C8C8C8494948C31396339426B31396339426B31396B
39427331396B39427331396B8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A
8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A
8C6352845A4A8C6352845A4A948C9C8C8494948C9C8C8C94948C9C8C8C94948C9C8C8C94
94949C5A73735A7B735A73735A737352736B5A737352736B5A737352736B5A737352736B
5A737352736B5A737352736B5A7373526B6B5A736B526B6B000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000052637B4A637B52637B52637B52637B52637B313131313131313131
31313131313194948C94948C8C8C8C94948C948C8C94948C8C8C8C94948C8C8C8C94948C
8C8C8C94948C8C8C8494948C8C8C848C8C848C8C848C8C8431396331396B31396B31426B
31396B31426B31396B3942733142738C63528C63528C63528C63528C63528C63528C6352
8C63528C63528C63528C63528463528C63528C63528C635284634A8C63528C63528C6352
845A4A8C635284634A948C9C8C8C94948C9C948C94948C9C948C94948C9C948C9C948C9C
948C9C5A73735A73735A73735A73735A73735A736B5A73735A736B5A73735A736B5A736B
52736B5A736B52736B5A736B52736B5A736B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000526B7B4A637B526B7B4A637B313131312931313131292931
3931393131319494948C8C8C9494948C8C8C94948C8C8C8C94948C8C8C8C94948C8C8C84
94948C8C8C8494948C8C8C8494948C8C8C8494948C8C8C8439426B31396331426B313963
39427331396B39427331396B3942733139738C6352845A4A8C63528C63528C6352845A4A
8C6352845A528C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A8C6352845A4A
8C6352845A4A948C9C8C8494948C9C8C8C94948C9C8C8494948C9C8C8C94948C9C8C8C94
94949C52736B5A7373526B6B5A737352736B5A7373526B6B5A737352736B5A7373526B6B
5A7373526B6B5A736B526B63000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000052637B4A637B313131313131313131313131313131
3131319494948C8C8C94948C8C948C94948C8C8C8C8C948C8C8C8C94948C8C8C8C8C948C
8C8C8C94948C8C8C848C8C8C8C8C848C948C8C8C848C8C8431396331396B31396331396B
31396B39426B31396B3142733142733942733142738C63528C63528C63528C63528C6352
8C63528C63528C63528C63528C63528C63528C63528C63528C63528C63528C634A8C6352
8C63528C63528C84948C8C9C8C8C9C948C9C8C8C94948C9C948C9C948C9C8C8C9C948C9C
948C9C39426B52736B5A73735A736B5A737352736B5A736B52736B5A736B526B6B52736B
52736B5A736B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000313131292931313131292931313139312931
3131398C8C8C9494948C8C8C9494948C8C8C9494948C8C8C94948C8C8C8C94948C8C8C84
94948C8C8C8494948C8C8C8494948C8C8C8494948C8C8C848C948C31396339426B313963
39426B31396B39427331396B39427331396B3942738C63528C63528C63528C63528C6352
8C63528C63528C63528C63528C63528C634A8C63528C634A8C6352845A4A8C63528C634A
8C63528C8494948C9C8C8494948C9C8C8494948C9C8C8C94948C9C8C8C9494949C8C8C94
94949C8C8C945A7373526B6B5A7373526B6B5A7373526B6B5A736B526B6B5A736B526B6B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000312931313131313131313131313131313139
8C948C9494948C948C94948C8C8C8C94948C8C8C8C94948C8C8C8C8C948C8C8C8C8C948C
8C8C8C8C948C8C8C848C948C8C8C848C8C848C8C848C8C848C8C8431396331396331396B
31396B31426B31396B31427331396B3142733142733942738C63528C63528C63528C6352
8C63528C63528C63528C63528C63528C63528C63528C63528C63528C63528C63528C6352
8C8C948C8C9C8C8C9C8C8C9C8C8C9C948C9C8C8C9C948C9C8C8C9C948C9C948C9C948C9C
948C9C948C9C52736B5A736B526B6B52736B526B6B52736B526B6B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003131312929313131393131313131398C8C8C
9494948C8C8C9494948C8C8C9494948C8C8C94948C8C8C8C94948C8C8C8C94948C8C8C84
94948C8C8C848C948C8C8C8494948C8C8C848C948C8C8C8494948C8C8C8431426B313963
39426B31396B39426B31396B39427331396B39427331397339427B8C63528C63528C6352
946B5A8C63528C63528C6352946B528C63528C63528C634A946B528C63528C63528C634A
948C9C8C8494948C9C8C8494948C9C8C8C9C948C9C8C8C94948C9C8C8C9C948C9C8C8C94
94949C8C8C9C5A7373526B6B5A7373526B6B52736B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000313131312931313131313131313139313131949494
8C948C9494948C8C8C8C948C8C948C94948C8C8C8C8C948C8C8C8C8C948C8C8C8C8C8C8C
8C8C8C8C948C8C8C848C8C8C8C8C848C948C8C8C848C8C848C8C848C948431396331396B
31396B31426B31396B31426B31426B3942733142733142733142733942738C63528C6352
8C63529463528C63528C63528C63528C63528C63528C63528C63528C63528C63528C8C9C
8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C948C9C8C8C9C8C8C9C8C8C9C948C9C8C8C9C948C9C
948C9C94949C8C8C9C526B6B526B6B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000002929313131393129313131393131319494948C8C8C
9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C94948C8C8C8C94948C8C8C84
94948C8C8C8494948C8C8C848C948C8C8C848C948C8C8C848C948C8C8C8431426B313963
31426B31396339426B31396B39427331396B39427331397339427331397339427B8C6352
946B5A8C6352946B5A8C6352946B528C6352946B5A8C6352946B528C63528C8C9C8C8494
8C8C9C8C8494948C9C8C8494948C9C8C8C94948C9C8C8C9C948C9C8C8C9494949C8C8C9C
94949C8C8C9C94949C000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003952633131313131313131313131319494948C9494949494
8C948C8C94948C948C8C94948C8C8C8C948C8C948C8C948C8C8C8C8C948C8C8C8C8C948C
8C8C8C8C948C8C8C8C8C948C8C8C848C8C8C8C8C848C8C848C8C848C8C848C8C8431396B
31396331396B31396B31426B31396B31427331427339427331427339427331427339427B
8C63529463528C6352946B528C63529463528C63529463528C63529463528C8C9C8C8C9C
8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C948C9C8C8C9C8C8C9C8C8C9C948C9C
8C8C9C948C9C948C9C000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003131393129313131393129313931398C8C8C9494948C8C8C
9494948C8C8C9494948C8C8C9494948C8C8C8C948C8C8C8C9494948C8C8C8C948C8C8C84
94948C8C8C848C948C8C8C848C948C8C8C848C948C848C848C948C8C8C848C9484313963
39426B31396331426B31396B39427331396B39427331396B39427331427339427B314273
394A7B8C6352946B5A8C6352946B5A8C6352946B5A8C6352946B5A8C849C8C8C9C848494
948C9C8C8C9C8C8C9C8C8494948C9C8C8C9C948C9C8C8C9C94949C8C8C9C948C9C8C8C9C
9494A58C8C9C94949C8C8C9C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003131313131313131313131318C94949494948C948C8C9494
8C94948C94948C8C8C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C8C8C8C8C948C
8C8C8C8C948C8C8C848C8C8C8C8C8C8C948C8C8C848C8C848C8C848C94848C8C84313963
31396331426B31396B31396B31396B31426B31396B314273314273394273314273394273
31427339427B8C6352946B52946352946B5A8C6352946B528C8C9C8C8C9C8C849C8C8C9C
8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C948C9C8C8C9C8C8C9C
8C8C9C94949C8C8C9C948C9C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000394A5A3131393129313131393131319494948C8C8C9494948C8C8C
9494948C8C8C9494948C8C8C8C94948C8C8C9494948C8C8C8C948C8C8C8C8C948C8C8C8C
8C948C8C8C848C948C8C8C848C948C848C848C948C848C848C948C848C848C948C848C84
31426B31396339426B31396B39426B31396B39427331396B39427331397339427B314273
39427B314273394A7B8C6352946B5A8C6352946B5A8C63528C8C9C8484948C8C9C8C849C
8C8C9C8C849C8C8C9C8C8C9C8C8C9C8C8C9C948C9C8C8C9C94949C8C8C9C9494A58C8C9C
9494A58C8C9C9494A58C8C9C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003952633131313131313131318C94948C94948C94948C94948C9494
8C948C8C94948C948C8C94948C8C8C8C948C8C8C8C8C948C8C8C8C8C948C8C8C8C8C948C
8C8C8C8C948C8C8C8C8C948C8C8C848C8C8C8C8C848C8C8C8C8C848C8C848C8C848C8C84
31396331396B31396B31426B31396B31426B31396B314273314273314273314273394273
31427339427B31427B39427B946352946B5A946B528C8C9C8C849C8C8C9C8C8C9C8C8C9C
8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C948C9C
8C8C9C948C9C8C8C9C94949C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000002929313131393131319494948C8C8C9494948C94948C94948C8C8C
9494948C8C8C8C94948C8C8C9494948C8C8C8C94948C8C8C8C94948C8C8C8C948C848C84
8C948C8C8C8C8C948C848C848C948C848C848C948C848C848C948C848C848C948C848C84
8C948C31396331426B31396339426B31396B39427331396B39427331396B394273313973
39427B31427339427B314273946B5A9463528C8C9C8484948C8C9C84849C8C8C9C84849C
8C8C9C8C8C9C8C8C9C84849C948CA58C8C9C8C8C9C8C8C9C9494A58C8C9C94949C8C8C9C
9494A58C8C9C9494A58C8C9C9494A5000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000394A5A3131313131313131398C94948C94948C94948C94948C94948C9494
8C94948C94948C8C8C8C948C8C948C8C94948C8C8C8C948C8C8C8C8C948C8C8C8C8C948C
8C8C8C8C948C8C8C8C8C8C8C8C8C8C8C948C848C848C8C848C8C848C948C848C848C8C84
8C8C8431396331396331396B31396B31426B31396B31426B31426B394273314273314273
31427339427B31427339427B39427B946B5A84849C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C
8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C
8C8C9C9494A58C8C9C8C949C8C949C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003129313131398C8C949494948C8C949494948C8C8C9494948C8C8C
8C94948C8C8C8C94948C8C8C8C94948C8C8C8C94948C8C8C8C948C848C8C8C948C848C8C
8C948C848C848C948C848C848C948C848C848C948C848C848C948C848C848C948C848C84
8C948431396331426B31396331426B31396B39427331396B39427331396B394273313973
39427B31427339427B3142738C8C9C84849C8C8C9C84849C8C8C9C84849C8C8CA584849C
8C8C9C84849C8C8CA58C8C9C8C8CA58C8C9C8C94A58C8C9C8C94A58C8C9C9494A58C8C9C
9494A58C8C9C9494A58C8C9C9494A5000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000003131313131393131318C94948C94948C94948C94948C94948C94948C9494
8C948C8C94948C948C8C94948C8C8C8C948C8C948C8C948C8C8C8C8C948C8C8C8C8C948C
8C8C8C8C948C8C8C8C8C948C848C848C8C8C8C8C848C8C8C848C848C8C84848C848C8C84
848C848C8C8431396331396B31396B31426B31396B31426B31396B314273314273314273
31427339427384849C8C8C9C84849C8C8C9C84849C8C8C9C848C9C8C8C9C8C8C9C8C8C9C
8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C8C94A5
8C8C9C8C94A58C8C9C9494A58C8C9C8C94A5000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031313931293194949C8C94948C94948C8C949494948C8C948C94948C8C8C
8C94948C8C8C8C9494848C8C8C94948C8C8C8C9494848C8C8C9494848C8C8C948C848C8C
8C948C848C8C8C948C848C848C948C848C848C948C848C848C948C848C848C948C848C84
8C948C848C8431396B31396339426B31396331426B31396B39427331396B39427331396B
39427384849C8C8C9C84849C8C8CA584849C8C8C9C84849C8C8CA584849C8C8C9C84849C
8C8CA5848C9C8C8CA5848C9C8C94A58C8C9C8C8CA58C8C9C9494A58C8C9C8C94A58C8C9C
9494A58C8C9C9494A58C8C9C9494A58C8C9C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000003131318C94948C94948C94948C94948C94948C94948C94948C94948C9494
8C94948C94948C8C8C8C94948C948C8C94948C8C8C8C948C8C948C8C948C848C8C8C948C
8C8C8C8C948C848C8C8C8C8C8C8C8C8C948C848C848C8C84848C848C948C848C84848C84
848C848C8C8431396331396331396331426B31396B31396B31426B31427331396B84849C
84849C848C9C84849C84849C84849C8C8C9C84849C8C8C9C848C9C8C8C9C848C9C8C8C9C
8C8C9C8C8CA58C8C9C8C8C9C8C8C9C8C8CA58C8C9C8C8C9C8C8C9C8C94A58C8C9C8C8CA5
8C8C9C8C94A58C8C9C8C94A58C94A59494A5000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003129313131398C94948C949C8C8C948C949C8C8C948C94948C8C8C8C94948C8C8C
8C9494848C8C8C9494848C8C8C9494848C8C8C9494848C8C8C9494848C8C8C948C848C8C
8C948C848C848C948C848C848C948C848C848C948C848C848C948C848C848C948C848C84
8C948C848C848C948431396331426B31396339426B31396B39427384849C848C9C84849C
848C9C84849C8C8CA584849C8C8CA584849C8C8CA584849C8C8CA584849C8C8CA584849C
8C8CA5848C9C8C8CA5848C9C8C8CA5848C9C8C94A58C8C9C8C94A58C8C9C8C94A58C8C9C
8C94A58C8C9C9494A58C8C9C9494A58C8C9C42426B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003131398C94948C94948C94948C94948C94948C94948C94948C94948C94948C9494
8C94948C94948C94948C94948C8C8C8C948C8C948C8C948C8C8C8C8C948C8C8C8C8C948C
848C8C8C948C848C8C8C948C848C8C8C8C8C848C848C948C848C84848C84848C84848C84
848C84848C84848C8431396331396331396B31396B31426B84849C84849C84849C84849C
84849C84849C84849C848C9C84849C848C9C848C9C8C8C9C848C9C8C8C9C848C9C8C8C9C
848C9C8C8C9C8C8C9C8C8CA58C8C9C8C8CA58C8C9C8C8CA58C8C9C8C8CA58C8C9C8C94A5
8C8C9C8C94A58C8CA58C94A58C94A58C94A58C94A5000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003131318C949C8C8C948C949C8C94948C9494848C948C949C8C94948C9494848C8C
8C9494848C8C8C9494848C8C8C9494848C8C8C9494848C8C8C9494848C8C8C948C848C8C
8C948C848C8C8C948C848C848C948C848C848C948C848C848C948C848C848C948C848C84
8C948C848C84848C84848C7B31426B31396384849C7B849C848CA584849C848C9C84849C
8C8CA584849C848CA584849C8C8CA584849C8C8CA584849C8C8CA5848C9C8C8CA584849C
8C8CA5848C9C8C8CA5848C9C8C94A5848C9C8C94A5848C9C8C94A58C8C9C8C94A5848C9C
9494A58C8C9C8C94A58C8C9C9494A58C8C9C9494A5000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
3131318C949C8C94948C94948C94948C94948C94948C94948C94948C94948C94948C9494
8C94948C9494848C8C8C94948C948C8C9494848C8C8C948C8C948C8C948C848C8C8C948C
848C8C8C948C848C8C848C8C848C8C8C948C848C84848C8C848C848C948C848C84848C84
848C84848C84848C84848C8484849C84849C84849C84849C84849C84849C84849C84849C
84849C848CA584849C848C9C848C9C848CA584849C848C9C848C9C8C8CA5848C9C8C8C9C
848C9C8C8CA5848C9C8C8CA58C8CA58C8CA58C8C9C8C8CA58C8CA58C94A58C8C9C8C8CA5
8C94A58C94A58C8CA58C94A58C94A58C94A58C94A5000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C949C8C94948C949C8C94948C949C848C948C949C848C948C9494848C948C9494848C94
8C94947B5A4A6B6B8C636B846B6B8C636B846B738C6B6B846B738C6B6B846B738C6B6B84
73738C6B6B8C73738C6B6B8C73738C6B6B8C7373946B738C7373946B738C737B9473738C
737B9473738C737B9429294A31315A7384737B847B738473848CA57B849C848CA584849C
848CA584849C8C8CA584849C848CA584849C8C8CA584849C8C8CA584849C8C8CA5848C9C
8C8CA5848C9C8C8CA5848C9C8C94A5848C9C8C94A5848C9C8C94A5848C9C8C94A58C8C9C
8C94A58C8C9C8C94A58C8C9C8C94A58C8C9C9494A58C8C9C000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000636B84636B84636B84
636B84636B84636B846B6B84636B846B6B8C6B6B846B6B8C6B6B846B6B8C6B6B8C6B738C
6B6B8C6B738C6B738C73738C6B738C73738C6B738C73738C73738C73739473738C737394
73738C00000029294A3131523131527B84737384737B8473738473738473738473738473
84849C848C9C84849C848CA584849C848C9C848C9C848CA5848C9C848CA5848C9C8C8CA5
848C9C8C8CA5848C9C8C8CA5848C9C8C8CA58C8CA58C94A58C8CA58C8CA58C8CA58C94A5
8C8CA58C94A58C94A58C94A58C94A58C94A58C94A58C94A5000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000636384636B84636384
6B6B8C636B846B6B8C6363846B6B8C636B846B6B8C636B846B738C6B6B846B738C6B6B84
73738C6B6B8C73738C6B6B8C7373946B738C7373946B738C737B946B738C7373946B738C
00000000000031314A29294A3131523131527B8473737B6B7B8473737B6B738473737B6B
738473737B6B73846B6B7B6B73846B84849C848CA584849C8C8CA5848C9C8C8CA584849C
8C8CA5848C9C8C8CA5848C9C8C94A5848C9C8C94A5848C9C8C94A5848C9C8C94A5848C9C
8C94A58C8CA58C94A58C8C9C9494A58C94A58C94A58C8C9C000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000636B84636384636B84
636B846B6B84636B846B6B846B6B846B6B8C636B846B6B8C6B6B8C6B738C6B6B8C6B738C
6B6B8C6B738C6B6B8C6B738C6B738C73738C6B738C73738C73738C737394000000000000
00000000000000000031314A31314A313152738473738473738473738473737B6B738473
737B6B73846B737B6B737B6B737B6B737B6B6B7B636B7B6B6B7B63848CA5848C9C848CA5
848CA58C8CA5848C9C8C8CA5848CA58C8CA5848CA58C8CA58C8CA58C94A58C8CA58C94A5
8C94A58C94A58C8CA58C94A58C94A58C94A58C94A58C94A58C94A5000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006B6B84636384
6B6B846363846B6B8C636B846B6B8C636B846B738C636B846B738C636B846B738C6B6B84
6B738C6B6B8C73738C6B6B8C73738C6B6B8C7373946B738C737394000000000000000000
00000000000000000000000031314A29294A313152737B737B8473737B6B738473737B6B
7384736B7B6B7384736B7B6B737B6B6B7B63737B6B6B7363737B6B6B73636B7B6B6B7363
000000848C9C8C94A5848C9C8C94A5848C9C8C94A5848C9C8C94A5848C9C8C94A5848CA5
8C94A5848CA58C94A58C8CA58C94A58C8CA58C94A58C94A58C94A5000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000636384636B84
636B84636B84636B846B6B84636B846B6B8C6B6B846B6B8C6B6B846B6B8C6B6B8C6B738C
6B6B8C6B738C6B738C6B738C6B738C6B738C6B738C000000000000000000000000000000
00000000000000000000000000000031314A31314A738473738473738473737B6B738473
737B6B737B6B737B6B737B6B6B7B6B6B7B6B6B7B6B6B7B6B6B73636B7B636B7363000000
000000000000000000000000000000000000848CA58C94A5848CA58C94A58C8CA58C94A5
8C8CA58C94A58C94A58C94A58C94A58C94A58C94A58C94A58C94A5000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000636B84636384
6B6B8C6363846B6B8C6363846B6B8C636B846B6B8C636B846B738C636B846B738C636B84
73738C6B6B8C6B738C6B6B8C7373946B738C000000000000000000000000000000000000
00000000000000000000000000000029294A31315229294A7B8473737B6B7384736B7B6B
7384736B7B6B737B6B6B7B6B737B6B6B7B636B7B6B6B7363737B6B6B73636B7B63000000
0000000000000000000000000000000000000000000000000000000000008C94A5848CA5
8C94AD8C8CA58C94A5848CA58C94AD8C94A58C94A58C8CA5949CAD39396B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000636384636B84
636B84636B84636B84636B84636B846B6B8C636B846B6B8C6B6B8C6B738C6B6B8C6B6B8C
6B738C6B738C6B6B8C6B738C000000000000000000000000000000000000000000000000
00000000000000000000000000000000000029294A31314A738473738473737B6B737B6B
737B6B737B6B6B7B6B6B7B6B6B7B6B6B7B6B6B73636B7B636B7B636B7B6B6B7363000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008C94A58C94A58C94A58C94A58C94A58C94A58C94A58C94A5000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000636B84636384
6B6B846363846B6B8C636B846B6B8C636B846B738C636B846B738C636B846B738C6B6B84
6B738C6B6B8C73738C000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000029294A3131526B7B6B7384736B7B6B
7384736B7B6B73846B6B7B6B737B6B6B7363737B6B6B73636B7B6B637363000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008C94A58C94AD8C94A5000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000636B84636B84
636B84636B84636B84636B84636B846B6B8C636B846B6B8C636B846B6B8C6B6B8C6B738C
6B6B8C000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000029294A313152737B6B737B6B
6B7B6B737B6B6B7B6B6B7B6B6B7B6B6B7B6B6B7B636B7B6B6B73636B7B63000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000636B84636384
6B6B8C636B846B6B8C6363846B6B8C636B846B6B8C636B846B738C636B846B738C636B84
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031314A29294A7384736B7B6B
7384736B7B6B737B6B6B7363737B6B6B73636B7B6B6373636B7B6B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000636B84
636B84636B84636B84636B84636B846B6B8C636B846B6B8C6B6B8C6B738C000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000031314A29294A737B6B
6B7B6B737B6B6B7B6B6B7B6B6B7B6B6B7B6B6B73636B7B636B7363000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000636384
636B8C6363846B6B8C636B846B6B8C636B846B738C636B846B738C000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003131526B7B6B
737B6B6B7B6B737B6B6B73636B7B6B6B73636B7B6B637363000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000636B84
636B84636B84636B84636B8C636B84636B8C636B84000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000029294A313152
6B7B6B6B7B6B6B7B6B6B7B6B6B7B6B6B7B6B6B73636B7B63000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000636384
636B8C636384636B8C6363846B6B8C636B84000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000029294A
7384736B7B6B6B7B6B6B73636B7B6B6B73636B7B6B637363000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000636B84
636B84636B84636B84636B84000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
31314A6B7B6B6B7B6B6B7B6B6B7B6B6B7B6B637363000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000636384
636B8C636384636B8C000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029294A6B7B6B6B73636B7B6B6373636B7B6B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000636B84
636B84000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031314A6B7B6B6B7B6B6B73636B7B6B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005A6384
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031314A6373636B7B6B637363000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006B7B6B6B7B636B7B6B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006B7B6B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000637363000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[54]:=",
  ImageSize->{128, 128},
  AnimationDisplayTime->0.169,
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000200000P0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj2LW
9`00?g=cLda<C240<410Y/30`9T0IPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm?ooc020P800082006IVc3<cIP2009VIod:J
Yb8R8_mV<`00P00g??mVc5J0Zk75fBa1EkV>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7De3eDE:b/[:BT
Y>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEeMJ_0eLC4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7i/;2`W9bLP00
o`1o001oO`1oo`3o003oO`3oogl007l0Ogl0ogmo07moOgmoogoo07ooOgooool00?l0Ool0oomo0?mo
Oomooooo0?ooOoooo`0002XZ:UEEEGmoOjZZZ]GEeOoooh0700201`00P0L0080700031gT010L000<7
N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`L`0003
P@0004H010L000<7<0000h6P001600@700031bl000F1Vh6KP@1500@700031bl000JPPJ2JX9Y400@7
00031b<000<IX00020002:21Vh6PPJ21@`041`000`LS0005PJ2iX;T01P002H6PVZ2JX9ZPVP1200@7
00031b<000LIX9nPWj2O00<000^1Vh6PPJ21X86PP@1100@700031b<00X405;VP^J2iX840X9ZPVZ2J
X9ZPVZ2J@0041`000`LS000G6H6OX9nPWj2OX86PPJ21X86PPJ21X840?`041`000`LS00:100ViX;VP
^J2iX;T00Z002iZPVZ2JX9ZPVZ2J01P000>OVP008`041`000`LS000I6H6OX9nPWj2OX9nPWj21X86P
PJ21X86PP@0E0004PIn1Wb@010L000<78`03P@0<X;VP^J2iX;VP^J2i0Z002IZPVZ2JX9ZPO00C0005
VZ2JX9X090041`000`LS000J6H4IX9nPWj2OX9nPWj2OX9nPWj21X86PPJ0B0006PIn1Wh6O90041`00
0`LS00:101V2X;VP^J2iX;VP^J2iX;VP^IniVZ2JX9ZP00l000R1VZ2JX9ZPVR@010L000<78`0071V1
6J2OX9nPWj2OX9nPWj2OX9nPWj2OX86PPJ0>0008PIn1Wh6OPIlT00@700031b<018406KVP^J2iX;VP
^J2iX;VP^J2iWkVO^IZPVj002`02W`08X9ZPVZ2JX9XT00@700031b<001hIPAV1Wj2OX9nPWj2OX9nP
Wj2OX9nPWj2OX9ZPVZ09000<WiZOPIn1Wh6OPIn18`041`000`LS000KPGb1PKVP^J2iX;VP^J2iX;VP
^J2iX;VO^J2i00:P00>OX84010003h60X9nPWj2JWiZOVYnJW`0S00@700031b<0024IPAV16J2O^InP
Wj2OX9nPWj2OX9nPWj2OX9nPVZ2JX9X00P004:2JX9ZOPIn1WiZOPIn1Wh4S00@700031b<002B2O8:1
PZ2iX;VP^J2iX;VP^J2iX;VP^J2iX;VP^J2iWkVOPG`9P@0>X;VP^J2iX;VP^J2iX;T3X00<XJ2QX:6P
XJ2QX:6P30041`000`LS000/6H4IPAViWj2OX9nPWj2OX9nPWj2OX9nPWj2JX9ZPVZ2JP@210840P@21
0842X007Vj2KX9^PV`07X00=VZ2JX9ZPVZ2JX9ZPVP0=00@700031b<00221O89lPZ2iX;VP^J2iX;VP
^J2iX;VP^J2iX;VP^J2iX0Z10Z007:JP^J2iX;VP^J2iX;VP^J2iX;VP^J2iX;VP^J0>00@700031b<0
02XIPAV16H:O^InPWj2OX9nPWj2OX9nPWj2OX9nPVZ2OP@210840P@210844X005Vj2PX9/02J0039ZP
VZ2JX9ZPVZ2J^@h010L000<78`00885lPX62PKVP^J2iX;VP^J2iX;VP^J2iX;VP^J2iX85l2H407Z2V
X:JP^J2iX;VP^J2iX;VP^J2iX;VP^J2iX;VP^@l010L000<78@00;7^16H4IPAV2Wj2OX9nPWj2OX9nP
Wj2OX9nPWj2OX9nP0840P@210840PJ2Q4Z002iZPVZ2JX9ZPVZ2J00l010L000<78@007h5lPGb2O8:1
^J2iX;VP^J2iX;VP^J2iX;VP^J2iX;T02H408:JPYZ2VX:JPYZ2iX;VP^J2iX;VP^J2iX;VP^J2iX;VP
40041`000`LQ000;Nh4IPAV16H4I^Il00j007YnPWj2OX9nPWj2OX9nPWj00P@210840P@2QX:6PXA:P
00ZJX9ZPVZ2J^In140041`000`LR000MO85lPX62O8:P^J2iX;VP^J2iX;VP^J2iX;VP^G`028408JJP
YZ2VX:JPYZ2VX;VP^J2iX;VP^J2iX;VP^J2iX;VP^@0A00@700031b8000Z16H4IPAV16KVO0j007YnP
Wj2OX9nPWj2OX9nP0840P@210840XJ2QX:6PXA:P00RJX9ZPVZ2OX18010L000<78P006Wb1O89lPX62
X;VP^J2iX;VP^J2iX;VP^J2i28408j2VX:JPYZ2VX:JPYZ2VX;VP^J2iX;VP^J2iX;VP^J2iX;VP0180
10L000<78@002g^16H4IPAV16H:O00>P01ZOX:2PWj2OX9nPWj2OVP210840PJ2QX:6PX@>P00>QX:00
4Z001IZiX:2O01<010L000<77`000iZ1VP02P@0HO861PWb2PKVP^J2iX;VP^J2iX;VP^J2i1X409:2V
X:JPYZ2VX:JPYZ2VX:JP^J2iX;VP^J2iX;VP^J2iX;VP^A<010L000<7700049YkVW^JNiYkPAV16H4I
PYl5X00FWj2OX9nPWj2OX0210840PJ2QX:6PXAfP00>i00004P041`000`LJ000PPIZ1VX6JPIZ1VX62
O8:1PX6iX;VP^J2iX;VP^J2iX;T5P@0UYZ2VX:JPYZ2VX:JPYZ2VX:JPYZ2VX;VP^J2iX;VP^J2iX;VP
^@0E00@700031aP001BJNiYkVW^JNiYkVW^J6H4IPAV2P@NP00jOX:2PWj00P@210:27X@>P00>QX:00
6j000kVP000D00@700031a<001>10861VX6JPIZ1VX6JPIZ1VX6J00:100=lPX400Z003;VP^J2iX;VP
^J2iO0B102JVX:JPYZ2VX:JPYZ2VX:JPYZ2VX:JPYZ2iX;VP^J2iX;VP^J2iX1H010L000<74@006G^J
NiZ0VW^JP9Z0VX2JP9Z0VX2JP9Z0PAT00X404:2OX9nPWj2OX9nP0840PHLVX1L010L000<73P005h6J
PIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ100>J00j1PX62X;jP^J2nX;VP^@:102YlPJ2VX:6PXJ2QX:6PXJ2Q
X:6PXJ2VX:JPYZ2VX;VP^J2iX;VP^J2iX84G00@700031``0022JNiZ0VX2JP9YkVX2JP9Z0VX2JP9Z0
VX2JP9Z0VQV16@:P00jiWj2OX9nP0840PHNPQbFP60041`000`L9000BVX6JPIZ1VX6JPIZ1VX6JPIZ1
3iX02h62X;jP_Z2nX;il00:102VVX:JPYZ2QX:6PXJ2QX:6PXJ2QX:6PYZ2VX:JPYZ2iX;VP^J2iX;VP
^@0I00@700031`L002:0VX2JP9Z0VX2JP9Z0VX2JP9Z0VX2JP9Z0VX6JP9Z0VX2J0X400aViX003X00=
Wj00P@2PQj27X8NPQ`03X003Qj2P00^P00>7X:003Z0I00@700031`D001BPPIZ1VX6JPIZ1VX6JPIZ1
VX6JPA:J03F2O;jP_Z2nX85lYZ2QX:JPYZ2QX:6PXJ2QX:6PXJ2QX:6PYZ2VX:JPYZ2VX;VP^J2iX;VP
P@0I00@700031`H002V7PXN2Qh:PP9Z0VX2JP9Z0VX2JP9Z1VX6JPIZ1VX6JPIZ1VX6JO8:P^@03X00F
P@2PQj27X8NPQj27X8NPQj27X8NPQ`>P00N7X8NPQj2700>P00>7X:001j000h40000G00@700031`L0
01:PPZ22XH:QQj67XHNQQiZ1VX4CVP0cPKjP_Z21X:6PXJ2VX:JPYZ2QX:6PXJ2QX:6PXJ2QX:6PYZ2V
X:JPYZ2VX;VP^J2iX;VP01X010L000<720009HN2Qh:7PZ22X8:PPZ22X8:PPZ22VX6JPIZ1VX6JPIZ1
VX6JPH800j009X67X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj272:000iX0000H00@7
00031`P001aZX8:PPZ22XHNQPZ67XHNQQj67X8NPQj67XHNQ2IX0=8:PPJ2QX:6PXJ2QX:JPXJ2QX:6P
XJ2QX:6PXJ2QX:6PYZ2VX:JPYZ2VX;VP^J2iX;V1VX4H00@700031`T002B2Qh:7PXN2X8:PPZ22X8:P
PZ22X8:PPZ22X8:PPZ22X8:PPIX2P@0WQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NP
Qj2700JP00JiP9Z0VX0G00@700031`X005Z2X8:PPZ22XH:QPZ67XHNQQj27X8NPQj27X8NQQj67XHNQ
Qj61X9ZOX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PYZ2VX:JPYZ2VX;VP^J2iX9Z1VX6JPIXE00@7
00031`/00262Qh:7PXN2X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8800j009iZOVYnJWiZOVZ27X8NP
Qj27X8NPQj27X8NPQj27X8NPQj27X8NPQ`06X009VX2JP9Z0VX2J01@010L000<73000D8:PPZ22X8:P
Qj22XHNQQj27X8NPQj27X8NPQj67X8NQO;nPPYZPVYnJX9ZOVZ2JWiZQX:6PXJ2QX:6PXJ2QX:6PYZ2V
X:JPYZ2VX:JP^J2j0YX00h6JVP03VP03PIX0018010L000<73@007h:7PXN2Qh:PPXN2X8:PPXN2X8:P
PZ22X8:PPZ22X8400j003QV2VYnJWiZOVYnJWiZO0iX06InJVXNPQj27X8NPQj27X8NPQj27X8NPQkT0
0Z003H2JP9Z0VX2JP9Z0VX004@041`000`L>000MPZ22X8:PPZ22X8:PQj27X8NPQj27X8NPQj27X8L0
0Z00;KnP_h62VZ2JX9ZPVZ2JX9ZOVYnJWiZOVYnJWiZQX:6PYZ2VX:JPYZ2VX:JP^P0>VP03P@0000h0
10L000<740006HN2Qh:7PXN2Qh:7PXN2Qh:PPZ22X8:PPZ000X400inPX003X0036H:100:J01NOVYZJ
WiZOVYnJWiZOVYnJWiZOVYnJW`03VP09X9ZPVZ27X8NP00:J0120VX2JP9Z0VX2JP9Z0VX2J3P041`00
0`LA000GX8:PQj27X8NPQj27X8NPQj27XHNPQj400X401j2oX;nP_Z0018809YnJWiZOVYnJX9ZOVYnJ
WiZOVYnJWiZOVYnJWiZOVYnJWiZOVZ273IX01H6JPIZ100`010L000<74P0068N2Qh:7PXN2X8:PPZ22
X8:PPZ22X02100NP03iZPVZ2JYnJWiZOVYnJWiZOVYnJWiZOVYnJWiZOVYnJWiZOVYnJWiZPQj27X8NP
PIZ0VX2JP9Z0VX2JP9Z0VPX010L000<74P0058:PPZ27X8NPQj27X8NPQj27X8NP0h402KnP_j2nX;jP
_P05PP0/WiZOVYnJX9ZPVYnJWiZOVYnJWiZOVYnJWiZOVYnJWiZoX;n7X8NPQj27X8L9VP04PIZ1VPT0
10L000<74`005X:7PXN2Qh:PPXN2X8:PPZ22X8:10849X00UPVZ2JX9ZWiZOVYnJWiZOVYnJWiZOVYnJ
WiZOVYnJWiZOVYnJW`04X00FQj27X8JPPJ21X86PPIZ0VX2JP9Z0VPP010L000<750004H:PPZ27X8NP
PZ27X8NPQj2700>100ZP_j2oX;nP_Z2n1X80=IZPVYnJX9ZPVZ2JWiZPVYnJX9ZOVYnJWiZOVYnJWj2o
X;n7X8NPQj27X8NPQj27X8NPQj2700>J00>1VP001P041`000`LE000BPXN2Qh:7PXN2Qh:PPZ22P@21
3:008VZ2JX9ZPYZOVYnJWiZOVYnJWiZOVYnJWiZOVYnJWiZOVYl6X00HQj27X86PPJ21X86PPJ21X86P
PJ21X82J1P041`000`LF000=PZ22X8:PPZ22X8NPQ`04P@0<_j2oX;nP_Z2nX;jP1h8089ZPVZ2JX9ZP
VZ2JX9ZPVZ2JX9ZOVYnJWiZOV[nP_j2o0Z0068NPQj27X8NPQj27X8NPQj27X8NPQj27X0D010L000<7
5`003h:7PXN2Qh:PPXN2P@21000>X00PPVZ2JX9ZPYZOVYnJWiZOVYnJWiZOVYnJWiZOVYnJWiX7X00F
Qj26X86PPJ21X86PPJ21X86PPJ21Q`P010L000<760002H:7PZ22X8:PPP05P@0?X;nP_j2oX;nP_Z2n
X;j100N201jJ^IZiV[VJ^IZiV[VJ^IZPV[RJWiZOV[nP_j2oX;l2X00DQj27X8NPQj27X8NPQj27X8NP
Qj0900@700031aT000b2Qh:7PXN2P@21084?X00OPVZ2JX9ZPVZKVYnJWiZOVYnJWiZOVYnJWiZOVYnJ
W`0:X00AQZ21X86PPJ21X86PPJ21X8402`041`000`LI0006PH:PPZ221H404J2oX;nP_j2oX;nP_Z2n
X;jP00R202niV[VJ^IZiV[VJ^IZiV[VJ^IZOVYnP_j2oX;nP_j2oQj27X8NPQj27X8NPQj27X00=00@7
00031aX000Z7PXN2P@2108404J0076Z2JX9ZPVZ26InJWiZOVYnJWiZOVYnJWiZOVYl;X00>Qh67PJ21
X86PPJ21X84?00@700031a/000>PPX401H404[nP_j2oX;nP_j2oX;jP_Z2nX0Z202ViV[VJ^IZiV[VJ
^IZiV[RJ^IZjX;nP_j2oX;nP_h:PQj27X8NPQj27X00A00@700031a`000N10840P@2101>P01Z2JX9Z
PVZ26H:JWiZOVYnJWiZOVYnJWiZOVPfP00V1Qh6PPJ21X8404`041`000`LL0003O86100>101FP_j2o
X;nP_j2oX;nP_Z2nX;jP_X402H8079ZiV[VJ^IZiV[VJ^IZiV[nP_j2oX;nP_j2oX;l2X005Qj27X8L0
5P041`000`LL0005P@2108405Z0066Z2JX9ZPQV26H:JWiZPVYnJWiZOVYnJW`FP00FQX:2PX@03X005
XH67PHL060041`000`LK0003PGb100>101ZoX;nP_j2oX;nP_j2oX;nP_j2oX;nPPX62P@N201j1^IZi
V[VJ^IZiV[VP_j2oX;nP_j2oX;nP_j2oJX4I00@700031aX000F20840P@0IX00DPQV26H9ZPVZ2JX9[
WiZOVYnJWiX2W`0@XJ2QX:6PXJ2QX:6PXJ2QX1/010L000<76P05P@0L_j2oX;nP_j2oX;nP_j2oX;nP
_j2oX;nP_X62P@Z201ZiWkVO^IniWknP_j2oX;nP_j2oX;nP_j2oX1X010L000<76P001840P@0JX00W
PQV26H9ZPVZ2JX9ZPV^OVYnJWiZQX:6PXJ2QX:6PXJ2QX:6PXJ2Q01X010L000<76@05P@0MX;nP_j2o
X;nP_j2oX;nP_j2oX;nP_j2oX;nPPX4038806;VO^IniX;nP_j2oX;nP_j2oX;nP_j2oX1X010L000<7
6P000h40P@0LX00V6H8IPVZ2JX9ZPVZ2Jh9[WiZOX:6PXJ2QX:6PXJ2QX:6PXJ2QX:4J00@700031aT0
0h408J2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_h62P@0;PP0HWkVPXJ2oX;nP_j2oX;nP_j2o
X;nP_j226@041`000`LI00036H4001jP02J26H9ZPVZ2JX9ZPV^2Jh:JXJ2QX:6PXJ2QX:6PXJ2QX:6P
XJ2QPAT010L000<76@02P@0R_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_X62P@^201RPXJ2Q
X;nP_j2oX;nP_j2oX;nP_j2oX88I00@700031aX000>1X:007Z009AV2JX9ZPVZ2JX9ZXJ2QX:6PXJ2Q
X:6PXJ2QX:6PXJ2QX:6PXJ006@041`000`LI000TPJ2oX;nP_j2oX;nPYZ2oX:JP_j2oX;nP_j2oX;nP
_j2oX8:128806j6PXJ2QX:6P_j2oX;nP_j2oX;nP_j2oX;nP_`0I00@700031aT000>1X:007j009AV2
JX8IPVZ2X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ006@041`000`LH000UO:2P_j2oX;nP_j2V
X:JPYZ2VX;nP_j2oX;nP_j2oX;nP_j2nP@04PP0PXJ2QX:6PXJ2QX:6P_j2oX;nP_j2oX;nP_j2oX;nP
_h8H00@700031aP000>1X:008J009H8IPVZQX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:40
60041`000`LH000TPKnP_j2VX;nPYZ2VX:JPYZ2VX;nP_j2oX;nP_j2oX;nP_j2U0X808j2QX:6PXJ2Q
X:6PXJ2QX;nP_j2oX;nP_j2oX;nP_j2oX;nP01P010L000<760000j6PX00SX00SXJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:4060041`000`LG000=_j2oX:JPYZ2VX:JPW`07X00BXJ2QX:6P
XJ2QX:6PXJ2QX:4I1:008J6PXJ2QX:6PXJ2QX;nP_j2oX;nP_j2oX;nP_j2oX;nP_`0G00@700031al0
04>2X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8806H6OX9nPQZ26XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2Q
X:6PXJ2QX:6P01L010L000<77`001Z22X8NPQ`:P012QQj6PXJ2QX:6PXJ2QX:6P0X420003PAV200nP
01JVX;nP_j2oX;nP_j2oX;nP_j2oX;n15P041`000`LP000FX8:PPZ22X8:PPZ22X8:PPZ22X8:PPPH0
0QT09H66X8JPQZ26X8JPQZ21X86P0:6PXJ2QX:6PXJ2QX:6PXJ2QX:405P041`000`LP000DPZ27X8NP
Qj6PXJ2QX:6PXJ2QX:4900:13J000inP0006000>YZ2VX;nP_j2oX;nP_j0F00@700031b0001>PPZ22
X8:PPZ22X8:PPZ22X8:P00X0014IPAVPQZ26X8JPQZ26X86PP@0;000;X:6PXJ2QX:6PYQT05@041`00
0`LP000APZ27X8NPQj67XJ2QX:6PXJ003@000aV1X009X004Wj2OX0l000NoX;nP_j2o01D010L000<7
80004:22X8:PPZ22X8:PPZ22X88?000>6H66X8JPQZ26X86PPJ0D0003X:6P01D010L000<780003XNP
Qj27X8NQQj67XJ2Q4P000aV2X007X004Wj2OPB/010L000<780003J22X8:PPZ22X8:PPZ004`003H4I
X8JPQZ21X86PPJ00;0041`000`LQ000;X8NPQj27XHNQX8405@000h4IX006X003Wj2O02`010L000<7
8@002X:PPZ22X8:PPZ0G000:PHJPQZ21X86PPBd010L000<78@002:27X8NQQj676@000aV2X005X003
Wj0002`010L000<78@001h:PPZ22X8806`002AVPQZ21X86PP@0]00@700031b4000FPQj27X00N0003
PJ2P00>P00>60000;0041`000`LQ0004PZ22X20000HIX86PPJ0^00@700031b4000>PQ`008@001H6P
X9nP02l010L000<78@000h80000R00:100>PP@00;P041`000`M70003X9nP02l010L000<7B0000j00
000^00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT0
10L000<7N@041`00P0L0080700201`00\
\>"],
  ImageRangeCache->{{{0, 127}, {127, 0}} -> {-3.20067, -3.20067, 1.05828, 
  1.05828}},
  CellTags->"Animation"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
%%ImageSize: 128 128 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00744048 0.0238095 0.00744048 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
384 string
128 128 8 [128 0 0 128 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007B635A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000846B5A7B6352000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007B63527B6B5A7B635A7B6B5A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000524A5A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6352846B5A7B6352846B5A7B6352846B5A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000524252524A5A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B6B5A846B5A7B6B5A846B5A7B635A7B6B5A7B6352000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000524252524A5A5242525A4A5A000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007B635A846B5A7B635A846B5A7B6352846B5A7B6352846B5A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000524A5A52425252425A52425A524A5A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000846B5A846B5A846B5A846B5A7B6B5A7B6B5A7B6B5A846B5A7B63527B6B5A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000846B73000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000524252524A5A5242525A4A5A5242525A4A5A52425A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
846B637B6B5A846B637B6B5A846B5A7B635A846B5A7B635A846B5A7B6352846B5A7B6352
000000000000000000000000000000000000000000000000000000000000000000000000
8C6B7B846B738C6B7B846B73000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000052425A52425A524A5A52425A524A5A524A5A5A4A5A
524A5A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000846B5A
846B5A846B5A846B5A846B5A846B5A846B5A846B5A846B5A7B6B5A846B5A7B6B5A846B5A
7B6B5A000000000000000000000000000000000000000000846B73846B73846B73846B73
846B73846B73846B735A5A39000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000524A5A524252524A5A5242525A4A5A52425A5A4A5A52425A
5A4A63524A5A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000847363846B5A
847363846B5A846B637B6B5A8473637B6B5A846B5A7B635A846B5A7B6B5A846B5A7B6352
846B5A0000000000000000000000007B6373846B737B6373846B7B846B73846B7B846373
8C6B7B846B738C6B7B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000052425A524A5A52425A524A5A524A5A5A4A5A52425A5A4A5A
5A4A5A5A4A5A5A4A5A5A4A5A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000847363846B5A846B5A
846B5A846B63846B5A846B5A846B5A846B5A846B5A846B5A846B5A846B5A7B6B5A7B6B5A
7B6B5A846B737B6373846B73846B73846B73846373846B73846B73846B7B846B73846B73
846B73846B7B846B73000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000524A5A5242525A4A5A5242525A4A5A52425A5A4A5A52425A
5A4A6352425A5A4A635A4A5A5A4A63000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008C7363846B5A8C7363846B5A
847363846B5A847363846B5A8473637B6B5A8473637B6B5A846B5A7B6B5A846B737B6373
846B737B6373846B7B7B6373846B7B7B6373846B7B846373846B7B846373846B7B846B73
8C6B7B846B735A6342000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008C73736B7B8C000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000052425A524A5A52425A524A5A52425A524A5A524A5A5A4A5A
524A5A5A4A5A5A4A5A5A4A635A4A5A5A4A63000000000000000000000000000000000000
0000000000000000000000000000000000000000008C7363846B63847363846B63847363
846B5A846B63846B5A846B63846B5A846B5A7B63737B63737B63737B63737B63737B6B73
7B6373846B737B6B73846B73846B73846B73846B73846B7B846B73846B7B846B73846B7B
846B73846B7B525A39000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000947B7B8C73738C7B7B8C7373
947B7B8C7373947B7B8C7373947B7B000000000000000000000000000000000000000000
0000000000000000004A42525A4A5A52425A524A5A5242525A4A5A52425A5A4A5A52425A
5A4A63524A5A5A4A63524A5A5A52635A4A5A5A4A63313952000000000000000000000000
000000000000000000000000000000846B638C7363846B5A8C7363846B5A8C7363846B5A
8C7363846B5A84736373636B7B6B737B63737B6B737B6373846B7B7B6373846B737B6373
846B7B7B6373846B7B7B6373846B7B7B6B73846B7B7B6373846B7B846B73846B7B846B73
8C737B5A5A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008C73738C737B8C73738C7373
8C7373947B7B8C73738C73738C73738C7B738C73738C73738C7373947B738C73738C7373
6B7B8C00000000000052425A52425A524A5A52425A524A5A524A5A5A4A5A524A5A5A4A5A
5A4A5A5A4A635A4A5A5A4A635A4A635A4A635A4A5A5A4A635A4A63000000000000000000
0000000000000000000000008C73638C7363846B638C73638473638C7363846B5A7B6373
7B63737B63737B63737B63737B63737B63737B63737B63737B63737B6B737B63737B6B73
7B6B73846B737B6373846B73846B73846B7B846B73846B7B846B73846B7B846B73846B7B
846B7B5A5A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000008C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B738C7373947B738C7373947B738C7373947B738C7373947B7352425A5A4A63524A5A
5A4A63524A5A5A4A635A4A5A5A4A635A4A5A6352635A4A5A6352635A4A63000000000000
000000000000000000846B638C7363846B638C7363846B637B6373735A6B7B6373736373
7B63737363737B6B737363737B6B737B63737B6B7B7B6373846B7B7B6373846B7B7B6373
846B7B7B6373846B7B7B6373846B7B7B6373846B7B846B73846B7B846B738C6B7B846B73
8C737B525A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000008C737B8C73738C7B7B
8C73738C7B7B8C7373947B7B8C73738C73738C7373947B738C7373947B738C7373947B73
8C7373947B738C7373947B738C7373947B738C7373947B738C7373947B738C7373947B73
8C7373947B738C7373947B73737B8C5A4A635A4A635A4A635A4A636352635A4A63000000
0000000000008C73638C73637363737363737363737363737363737363737363737B6373
7363737B63737B63737B63737B63737B63737B63737B6B737B63737B6B737B6373846B7B
7B6373846B737B6B73846B7B846B73846B7B846B73846B7B846B73846B7B846B7B846B7B
5A5A39525A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B738C7373947B7B8C7373947B738C7373
947B738C7373947B738C736B947B738C7373947B738C736B848C9C6342316342315A3929
5A4231735A6B736373735A6B7B63737363737B6373735A6B7B63737363737B6373736373
7B6B7B7363737B6B737363737B6B7B7B63737B6B7B7B6373846B7B7B6373846B7B7B6373
846B7B7B6373846B7B7B6373846B7B846B73846B7B7B6B738C737B846B73846B7B846B73
5A6342000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C73738C7B7B
8C737B947B7B8C73738C7B7B8C7373947B7B8C7373947B738C7373947B7B8C7373947B73
947B73947B7B8C7373947B73947B73947B738C7373947B73947B73947B738C7373947B73
947B73947B738C7373947B73947B73947B738C73737B8C9C7B8C9C848C9C634231634231
5A3931736373735A737363737363737363737363737363737363737B63737363737B6373
7B63737B63737B63737B63737B63737B6B7B7B63737B6B737B6B73846B7B7B63737B6B73
7B6B73846B7B7B6B73846B7B846B7B846B7B846B73846B7B846B7B846B7B846B7B846B7B
525A39000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B738C7373947B7B8C7373947B738C7373947B737B849C848C9C7B8494848C9C634231
634231523931736373735A73736373735A737B63737363737B63737363737B6B7B736373
7B6B7B7363737B6B7B7B63737B6B7B7B6373846B7B7B6373846B7B7B6373846B7B7B6373
846B7B7B6373846B7B7B6B73846B7B7B6B73846B7B846B7384737B846B738C737B525A39
525A39000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000947B7B
8C737B947B7B8C7B7B947B7B8C7373947B7B8C7B73947B7B8C7373947B7B947B73947B7B
8C7B73947B7B947B73947B7B8C7B73947B73947B73947B73947B73947B73947B73947B73
947B73947B73947B73947B73947B73947B737B8C9C7B8C9C7B849C7B8C9C7B8C9C7B8C9C
6342316342315A39317363737363737363737363737B63737363737B63737363737B6373
7B63737B63737B63737B6B737B63737B6B7B7B63737B6B7B7B63737B6B7B7B6B73846B7B
7B6B73846B7B846B7B846B7B846B7B846B7B846B7B846B7B846B7B846B7B846B7B5A5A39
525239000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7B73947B7B8C7373947B7B8C73737B8C9C7B8494848C9C7B849C7B8C9C7B8494
848C9C6342316342315239297B637B7363737B63737363737B6B7B7363737B637B736373
7B6B7B7B63737B6B7B7363737B6B7B7B63737B6B7B7B6373846B7B7B6373846B7B7B6373
846B7B7B6B73846B7B7B6B7384737B846B7B846B7B846B738C737B846B7B8C737B525239
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
947B7B947B7B8C7B7B947B7B947B7B947B7B8C7373947B7B947B7B947B7B947B73947B7B
947B7B947B7B8C7B73947B7B947B73947B7B947B73947B73947B73947B7B947B73947B73
947B73947B7B947B73947B73947B737B8C9C7B84947B8C9C7B8C9C7B8C9C7B849C7B8C9C
7B8C9C6B42396342315A42315A39317363737363737363737363737B63737363737B6373
7B63737B6B7B7B63737B6B7B7B63737B6B7B7B63737B6B7B7B6B7B846B7B7B6B73846B7B
846B7B846B7B7B6B7B846B7B846B7B846B7B846B7B846B7B846B7B84737B525A39525A39
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7B73947B7B8C7373947B7B8C7373
947B7B8C7373947B7B947B73947B7B7384947B8C9C7B84947B8C9C7B84947B8C9C7B8494
7B8C9C7B84946B42395A39315A42317363737B6B7B7363737B6B7B7363737B6B7B736373
7B6B7B7363737B6B7B7B63737B6B7B7B6373846B7B7B6373846B7B7B6373846B7B7B6B73
846B7B7B6B73846B7B7B6B7B84737B846B7B84737B846B7B84737B846B7B5A63424A5239
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B73947B7B947B7B947B7B947B73947B7B947B73947B7B
947B73947B7B947B73947B7B7B84947B849C7B849C7B8C9C7B849C7B8C9C7B849C7B8C9C
7B84947B8C9C7B8C9C6342316342315A42317363737B637B7363737B63737B63737B6B7B
7B63737B6B7B7B63737B6B7B7B63737B6B7B7B6B7B7B6B7B7B6B7B846B7B7B6B7B846B7B
7B6B7B846B7B846B7B846B7B846B7B846B7B846B7B846B7B846B7B5A6342525A39525239
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000947B7B8C7373947B7B947B7B947B7B8C7373947B7B947B7B947B7B8C7373
947B7B947B73947B7B8C7373947B7B947B73947B7B8C7373947B7B947B73947B7B8C7373
947B7B947B73947B7B8C73737B8C9C7384947B8C9C7384947B8C9C7B84947B8C9C738494
7B8C9C7B84947B8C9C7B84946B42395A39315A42317363737B6B7B7363737B6B7B736373
7B6B7B7B63737B6B7B7B6373846B7B7B63737B6B7B7B6373846B7B7B6B7B846B7B7B6373
84737B7B6B7B846B7B7B6B7B847384846B7B84737B846B7B8C7384525A39525A39000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B73947B7B947B7B947B7B947B73947B7B
947B7B947B7B947B737B849C7B849C7B8C9C7384947B849C7B849C7B8C9C7B84947B8C9C
7B8C9C7B8C9C7B84947B8C9C7B8C9C6342315A39315A39317B63737B6B7B7363737B637B
7B637B7B6B7B7B63737B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B846B7B7B6B7B846B7B
846B7B846B7B846B7B846B7B846B7B84737B846B7B846B7B846B7B5A5A42525239000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000947B84947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B73947B7B947B73947B7B947B73947B7B947B73947B7B947B73947B7B947B73
947B7B947B73947B7B7384947B8C9C7384947B8C9C7384947B8C9C7384947B8C9C738494
7B8C9C7384947B8C9C7B84947B8C9C7B84946342395A39317B6B7B7363737B6B7B736373
7B6B7B7363737B6B7B7B63737B6B7B7B6373846B7B7B6B7B846B7B7B6B7B846B7B7B6B7B
846B7B7B6B7B847384846B7B847384846B7B847384846B7B5A6342525A39525A39000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B7B7384947B849C7384947B849C7384947B849C7384947B849C7B84947B8C9C
7B84947B8C9C7B84947B8C9C7B84947B8C9C7B8C946342315A39317B6B7B7B637B7B6B7B
7B637B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B846B7B7B6B7B846B7B
846B7B846B7B846B7B846B7B846B7B846B7B846B7B84737B525A39525A394A5239000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000947B7B948484947B7B947B7B947B7B9C8484947B7B947B7B947B7B
94847B947B7B947B7B947B739C847B947B7B947B7B947B739C847B947B7B947B7B947B73
9C847B947B737B849C7384947B8C9C7384947B849C7384947B8C9C7384947B8C9C738494
7B8C9C7384947B8C9C7384947B8C9C7B84947B8C9C7384946342395A39317B6B7B736373
7B6B7B7B637B7B6B7B7B6373846B7B7B6B7B846B7B7B637B846B847B6B7B846B7B7B6B7B
8473847B6B7B84737B7B6B7B847384846B7B847384846B7B5A63424A5239000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B7B849C7384947384947384947B849C7384947384947384947B8C9C7384947B8494
7B84947B8C9C7384947B84947B8C947B8C9C7B84947B8C946342316342315A39317B6B7B
7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B846B7B7B6B7B846B7B
846B7B846B7B846B7B846B7B846B7B847384846B7B5A6342525A39525A39000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000947B84947B7B948484947B7B948484947B7B9C8484947B7B
94847B947B7B9C847B947B7B9C847B947B7B9C847B947B7B9C847B947B7B9C847B947B7B
9C847B7384947B849C7384947B849C7384947B8C9C7384947B8C9C7384947B8C9C738494
7B8C9C7384947B8C9C7384947B8C9C7384947B8C9C7384947B8C9C63423163423173637B
7B6B7B7B637B7B6B7B7B6B7B846B7B7B6B7B846B7B7B6B7B846B847B6B7B8473847B6B7B
8473847B6B7B847384846B7B847384846B7B8C7384525A39525A394A5231000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000947B7B947B7B947B7B947B7B947B7B947B7B947B7B
947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B
73849473849473849473849C7384947384947384947384947384947384947384947B8494
7384947B84947384947B8C947384947B8C947B84947B8C947384947B8C94634231634231
7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B7B6B7B846B7B7B6B7B846B7B846B7B846B84
846B7B846B7B846B7B847384846B7B84738484737B525A424A52394A5239000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000947B7B948484947B7B9C8484947B7B948484947B7B
9C8484947B7B9C8484947B7B9C8484947B7B9C847B947B7B9C8484947B7B9C847B737B94
73849C73849473849C737B947B849C73849473849C7384947B8C9C7384947B849C738494
7B8C9C7384947B8C9C7384947B8C9C7384947B8C9C7384947B8C9C7384947B8C9C633931
6342397B6B7B7B6B7B7B637B846B847B6B7B846B847B6B7B8473847B6B7B8473847B6B7B
847384846B7B847384846B7B8C7384846B7B5A6342525239525A396339316342395A3931
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000947B7B947B7B947B7B948484947B7B947B7B
947B7B9C8484947B7B947B7B947B7B9C847B947B7B947B7B947B7B9C847B947B7B738494
738494738494738494738494738494738494738494738494738494738494738494738494
7384947B84947384947384947384947B8C947384947B84947384947B8C947384947B8C94
6342315A42317B6B7B7B6B7B7B6B7B846B7B7B6B7B7B6B7B7B6B7B846B847B6B7B846B7B
846B7B847384846B7B847384847384847384525A39525A39634239634239634231634239
634239634239000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000009C8484947B7B9C8484947B7B9C8484947B7B
9C8484947B7B9C8484947B7B9C8484947B7B9C8484947B7B9C8484947B7B73849C737B94
73849C6B7B9473849C73849473849C73849473849C73849473849C7384947B8C9C738494
7B8C9C7384947B8C9C7384947B8C9C7384947B8C9C7384947B8C9C7384947B8C9C738494
7B8C9C5A3931846B847B6B7B846B847B6B7B8473847B6B7B8473847B6B7B8473847B6B7B
8473847B6B7B847384846B7B847384525A395A63424A52396B42396342316B4239634231
6B42396342316B4239634231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000009C8484947B7B948484947B7B9C8484
947B7B948484947B7B9C8484947B7B94847B947B7B9C847B947B7B9C847B738494738494
738494738494738494738494738494738494738494738494738494738494738494738494
738494738494738494738494738494738494738494738C947384947B8C947384947B8C94
7384947B8C946342317B6B7B7B6B7B7B6B7B7B6B7B846B847B6B7B846B84846B7B847384
846B7B847384846B84847384846B845A6342525A396B42396342396B42396342396B4239
634239634239634239634239634239634239000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000947B7B9C8484947B7B9C8484947B7B
9C8484947B7B9C8484947B7B9C8484947B7B9C8484947B7B9C84846B7B947384946B7B94
73849C737B947384946B7B9473849C7384947384946B7B9473849C738494738494738494
7B8C9C7384947384947384947B8C9C7384947B8C947384947B8C9C7384947B8C94738494
7B8C9C7384947B8C945A39318473847B6B7B846B847B6B7B8473847B6B7B8473847B6B7B
847384846B7B847384846B7B8C7384525A39525A396342316B4A396342396B4239634231
6B4A396342396B42396342316B4A396342316B4239634231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004A31314A3139947B849C8484947B7B9C8484
947B849C8484947B7B9484849484849C8484947B7B9C848494847B7384946B7B94738494
738494738494738494738494738494738494738494738494738494738494738494738494
738494738494738494738494738494738494738494738494738494738C94738494738494
7384947B8C94738494738C945A4231846B847B6B7B846B84846B84847384846B7B847384
846B84847384846B848473845A6342525A426B42396B42396B42396B42396342396B4239
6B42396B42396342396B42396B42396B42396342396B42396B42396B4239000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000005231394A31315239394A31319C8484947B7B9C8484947B7B
9C8484947B7B9C8484947B7B9C8484947B7B9C8484947B7B7384946B7B947384946B7B94
7384946B7B947384946B7B947384946B7B947384946B7B947384946B8494738494738494
738494738494738494738494738C947384947B8C94738494738C947384947B8C94738494
7B8C947384947B8C947384946342397B6B7B8473847B6B7B8473847B6B7B8473847B6B7B
847384846B7B847384846B7B5A63424A5A396B4A396B42396B4A396342396B4A39634239
6B4A396342396B4A396342396B4A396342316B4A396342396B4A396342316B4A39634239
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005231394A31395231394A31395231395231399C84849484849C8484
947B849C84849484849C84849484849C84849C84849C84846B7B947384946B7B94738494
6B7B947384947384947384946B8494738494738494738494738494738494738494738494
738494738494738494738494738494738494738494738494738494738494738494738494
738494738C94738494738C947384946342317B6B848473847B6B84847384846B84847384
846B84847384847384847384525A396B42396B42396B4A396B42396B42396B42396B4239
6B42396B42396B42396B42396B42396B42396B42396B42396B42396B42396B42396B4239
6B42396B4239000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A31315231394A31315239394A31395239394A3131523939947B849C8484947B7B
9C8484947B7B9C8484947B7B9C8484947B7B9C84846B7B8C7384946B7B947384946B7B8C
7384946B7B947384946B7B8C7384946B7B947384946B7B8C7384946B84947384946B7B8C
7384947384947384946B848C738C9473849473849473848C7B8C94738494738C9473848C
7B8C94738494738C9473848C7B8C947384946342317B6B7B8473847B6B848473847B6B7B
847384846B84847384525A395263426B42396B4A396B4239734A426B42396B4A396B4239
734A426B42396B4A396B4239734A426B42396B4A396B4239734A396B42396B4A396B4231
6B4A396B42396B4A396B4231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A3139
4A31395231394A3139523139523139523139523139523139523139523939947B849C8484
9C84849C84849484849C84849C84849C848494847B6B7B946B7B947384946B7B946B8494
6B7B947384946B7B946B84946B84947384946B7B947384947384947384946B8494738494
738494738494738494738494738494738494738494738494738494738494738494738494
738494738C94738494738494738494738C94738494634231846B84847384846B84847384
8473848473848473845263426B4239734A396B42396B42396B4239734A396B42396B4A39
6B42396B4A396B42396B42396B42396B4A396B42396B42396B42396B4A396B42396B4239
6B42396B4A396B42396B42396B42396B4A39000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005239394A3131
5239394A31315239394A31395239394A31395239395231395239395231399C8484947B84
9C8484947B7B9C8484947B849C8484947B7B7384946B7B947384946B7B8C7384946B7B94
7384946B7B8C7384946B7B947384946B7B8C7384946B7B947384946B7B8C7384946B7B94
7384946B7B8C7384946B84947384946B848C7384946B848C7384946B848C738C94738494
738C9473848C738C9473848C738C9473848C7B8C9473848C6342317B6B84847384846B84
847384846B845A63426B4239734A426B4239734A426B4239734A426B4239734A426B4239
734A426B4239734A426B4239734A396B4239734A426B4239734A396B4239734A396B4239
734A396B4239734A396B4239734A396B4239734A396B4239000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A31395231394A3139523139
4A31395231395231395231395231395231395231395239395231395239395231399C8484
9C84849C84849C84849C84849C84849C84846B7B946B7B946B7B946B7B946B7B946B8494
6B7B946B84946B7B946B84946B7B946B84946B84947384946B84947384946B8494738494
6B8494738494738494738494738494738494738494738494738494738494738494738494
738494738494738494738494738494738494738494738C94738494847384847384847384
847384847384526342734A426B4239734A39734A39734A396B4239734A39734A39734A39
6B4239734A396B4A39734A396B4239734A396B4A39734A396B4239734A396B4239734A39
6B4239734A396B4A39734A396B42396B4A396B4239734A396B42396B4A39000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005231394A31315239394A31395231394A3131
5239395231395239394A31395239395231395239395231395A39425231395A3939947B84
9C84849484849C8484947B7B9C84846B7B946B7B946B7B8C7384946B7B8C6B84946B7B8C
7384946B7B947384946B7B8C7384946B7B8C7384946B7B8C7384946B7B947384946B7B8C
7384946B848C7384946B7B8C7384946B84947384946B848C738C946B848C7384946B848C
738C9473848C738C946B848C738C9473848C738C9473848C7B8C9473848C8473847B6B84
84738C847384734A426B4239734A42734239734A426B4239734A426B4239734A426B4239
734A426B4239734A426B4239734A426B4239734A396B4239734A426B4239734A396B4239
734A426B4239734A396B4239734A426B4239734A396B4239734A426B4239734A396B4239
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004A31395231394A31395231395231395231394A3139523139
5231395239395231395231395231395239395231395231395231395A39395231395A3939
9C84849C84849C84849C84849C84846B7B946B7B946B7B946B7B946B7B946B7B8C6B7B94
6B7B946B84946B7B946B84946B7B946B84946B7B8C6B84946B84947384946B7B946B8494
6B84947384946B848C7384947384947384946B84947384947384947384946B848C738494
738494738494738494738494738494738C9473848C738494738494738C94738494847384
8473845A6342734A39734A39734A39734A42734A39734A39734A39734A42734A39734A39
734A39734A42734A39734A39734A39734A39734A39734A39734A39734A42734A39734A39
734A39734A396B4A39734A39734A39734A396B4A39734A39734A39734A396B4239734A39
734A39734A39000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A31315239394A31395239394A31395239394A3139523939523139
5239395231395239425231395A39395231395A39425231395A39425231395A3942523139
9C84849484849C848C9C84846B7B94637B8C6B84946B7B8C6B7B94637B8C6B84946B7B8C
6B84946B7B8C7384946B7B8C6B84946B7B8C7384946B7B8C7384946B7B8C7384946B7B8C
7384946B7B8C7384946B7B8C7384946B7B8C7384946B848C7384946B848C7384946B848C
7384946B848C738C946B848C738C946B848C738C9473848C738C9473848C738C945A4231
84738C7342397B4A42734239734A427342397B4A42734A39734A42734239734A42734239
734A42734239734A42734A39734A42734239734A42734239734A42734239734A42734A39
734A426B4239734A42734239734A426B4239734A42734239734A426B4239734A426B4239
734A426B4239734A426B4239000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A31394A31395231394A3139523139523139523139523139523139523139523939
5231395239395231395239395231395239395231395A39395231395A39395A39395A3942
5A31399C84849C84849C84846B7B946B7B946B7B946B7B946B7B8C6B7B946B7B946B7B94
6B7B946B7B946B7B946B84946B7B8C6B84946B7B946B84946B7B946B84946B84946B8494
6B7B8C6B84946B84947384946B84947384946B84947384946B848C7384946B8494738494
6B849473849473849473849473848C73849473849473849473848C738494738494738C94
847384734A42734A39734A42734A39734A42734A39734A42734A39734A42734A396B8484
6B84846B84846B84846B84846B847B6B84846B847B6B84846B847B6B84846B847B6B847B
6B847B6B847B6B847B6B847B6B847B6B847B6B847B6B847B6B847B6B847B6B847B6B847B
6B84736B847B6B84736B847B000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A3131
5239394A31395239394A31395239394A31395239394A3139523942523139523939523139
5A39425231395A39395231395A39425231395A39425231395A39425231395A3942523139
5A39429C84849C848C63738C6B7B94637B8C6B7B94637B8C6B84946B7B8C6B7B94637B8C
6B84946B7B8C6B8494637B8C6B84946B7B8C6B84946B7B8C7384946B7B8C6B84946B7B8C
7384946B7B8C7384946B7B8C7384946B7B8C7384946B7B8C7384946B848C7384946B848C
7B5252734A4A7B5252734A4A7B5252734A4A7B5252734A4A7B52527B4A4A7B5252734A4A
847B946B84846B848C6B8484738C8C6B84846B84846B8484738C8C6B84846B84846B847B
6B84846B847B6B84846B7B7B6B84846B847B6B8484637B7B6B84846B847B6B8484637B7B
6B84846B847B6B847B637B7B6B847B637B7B6B847B637B736B847B637B736B847B637B73
6B847B637B736B847B000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A3139523139
5231395239394A3139523139523139523939523139523139523139523939523139523939
5231395A39395231395239395239395A39425231395A39395A39395A39425A31395A3939
5A39395A39429C84846B7B946B7B946B7B94637B8C6B7B946B7B946B7B946B7B8C6B7B94
6B7B946B7B946B7B8C6B7B946B7B946B8494734A4A734A4A734A4A735252734A4A734A4A
734A4A735252734A4A73524A734A4A7B5252734A4A73524A73524A7B5252734A4A7B5252
7B524A7B5252734A4A7B52527B524A7B52527B524A7B52527B524A7B52527B524A5A6B52
847B8C6B848C6B84846B84846B84846B84846B84846B84846B84846B84846B84846B8484
6B84846B84846B847B6B84846B847B6B84846B847B6B847B6B847B6B847B6B847B6B847B
6B847B6B847B6B7B7B6B847B6B847B6B847B637B7B6B847B6B847B6B847B637B736B847B
6B8473000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005239394A31395239394A3139
5239394A31395239425231395239395231395239425231395239425231395A3942523139
5A39425231395A39425231395A39425231395A39425231395A39425231395A39425A3139
5A3942526B84734A526B4A4A734A526B4A4A7352526B4A4A7352526B4A4A7352526B4A4A
7352526B4A4A7352526B4A4A7352526B4A4A735252734A4A735252734A4A735252734A4A
7B5252734A4A7B5252734A4A7B5252734A4A7B5252734A4A7B5252734A4A7B5252734A4A
7B5252734A4A7B5252734A4A7B5252734A4A7B52527B524A7B52527B4A4A5A73527B738C
847B947B7B8C6B848C6B84846B84846B7B846B84846B84846B8484637B7B6B84846B7B7B
6B8484637B7B6B8484637B7B6B8484637B7B6B8484637B7B6B8484637B7B6B847B637B7B
6B847B637B7B6B847B637B736B847B637B736B847B637B736B847B637B736B847B637B73
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004A31395231394A3139523139523139523139
5231395231395231395239395231395239395231395239395231395239395231395A3942
5231395A3939526384526B84526384526B84526B84526B84526B84526B84526B84526B84
526B84526B849C9494734A526B4A4A734A526B4A4A734A526B4A4A734A52734A4A734A52
734A4A734A52734A4A735252734A4A735252734A4A735252734A4A735252734A4A735252
734A4A735252734A4A7B5252734A4A7B525273524A7B525273524A7B52527B52527B5252
7B524A7B52527B52527B52527B524A7B52527B52527B52527B524A5A6B525A6B52847B94
847B8C847B945A42316B84846B84846B84846B84846B84846B84846B84846B84846B8484
6B847B6B84846B847B6B84846B847B6B847B6B847B6B847B637B7B6B847B637B7B6B847B
637B7B6B847B637B7B6B847B637B7B6B847B637B736B847B637B736B847B637B73000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000004A31395239394A31395239424A31395239394A3139
52394252313952637B4A5A7B5263844A637B5263844A637B526B844A637B5263844A637B
526B844A637B526B844A637B526B844A637B526B844A637B526B84526384526B84526384
526B8C9C948C9C94949C8C8C7352526B4A4A7352526B4A4A7352526B4A4A7352526B4A4A
735252734A4A735252734A4A735252734A4A735252734A4A7B5252734A4A7B5252734A4A
7B5252734A4A7B5252734A4A7B5252734A4A7B5252734A4A7B5252734A4A7B5252734A4A
7B52527B524A7B5252734A4A7B52527B524A7B52527B4A4A6373525A6B4A847B947B738C
847B947B7B8C847B94637B7B6B8484637B846B8484637B7B6B8484637B7B6B8484637B7B
6B8484637B7B6B8484637B7B6B8484637B7B6B847B637B7B6B8484637B7B6B847B637B73
6B847B637B736B847B637B736B847B637B736B847B637B736B847B637B73000000000000
000000000000000000000000000000000000000000000000
0000000000000000004A5A734A5A734A5A7B4A5A734A5A7B4A5A7B4A637B4A5A7B4A637B
4A637B4A637B4A5A7B4A637B4A637B52637B4A637B4A637B4A637B52637B4A637B52637B
52637B5263844A637B526384526384526B8452637B526B84526B84526B84526384526B84
526B849C94949C948C9C94943131527352526B4A4A734A52734A52735252734A4A734A52
734A52735252734A4A735252734A52735252734A4A7352527352527B5252734A52735252
7352527B5252734A4A7B52527B52527B52527352527B52527B52527B52527B52527B5252
7B52527B52527B52527B52527B52527B52527B52527B52525A6B52847B947B7B8C847B94
847B94847B94847B8C5A42316B84846B8484637B7B6B84846B847B6B8484637B7B6B847B
6B847B6B8484637B7B6B847B637B7B6B847B637B7B637B7B637B7B6B847B637B7B637B7B
637B7B6B847B637B73637B7B637B736B847B637B73637B73000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000425A734A637B425A734A637B4A5A734A637B4A5A73
4A637B4A5A7352637B4A5A7352637B4A5A7B52637B4A637B52637B4A637B5263844A637B
5263844A637B526B844A637B526B844A637B526B844A637B526B844A637B526B84526384
9C94949C948C9C94949C948C9C94946B4A4A7352526B4A4A7352526B4A4A735252734A4A
735252734A4A7B5252734A4A735252734A4A7B5252734A4A7B5252734A4A7B5252734A4A
7B5252734A4A7B5252734A4A7B5252734A4A7B525273524A7B5252734A4A7B52527B524A
7B52527B524A7B52527B52527B52527B524A8452525A6B525A6B527B738C847B947B7B8C
847B947B7B8C847B947B7B8C6B8484637B7B6B8484637B7B6B8484637B7B6B8484637B7B
6B8484637B7B6B8484637B7B6B847B637B7B6B847B637B7B6B847B637B736B847B637B73
6B847B637B736B847B637B736B847B637B736B847B000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A5A734A5A734A5A7B4A5A734A637B
4A5A734A637B4A5A7B4A637B4A637B4A637B4A637B4A637B4A637B4A637B4A637B4A637B
4A637B52637B4A637B5263844A637B52637B4A637B526B8452637B526B84526384526B84
9C94949C94949C94949C94949C948C9C9494734A52735252734A52735252734A52735252
734A52735252735252735252734A527352527352527B52527352527B52527352527B5252
7352527B52527B52527B52527B52527B52527B52527B52527B52527B52527B52527B5252
7B52527B52527B52527B52527B52527B52525A6B525A6B527B7B8C847B947B7B8C847B94
7B7B8C847B94847B94847B94847B946B8484637B7B6B8484637B7B637B7B637B7B63847B
637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B7B
637B73637B7B637B73637B73637B73637B73000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A637B4A5A734A637B425A73
4A637B4A5A734A637B4A5A7352637B4A5A734A637B4A5A7352637B4A637B52637B4A5A7B
5263844A637B52637B4A637B526B844A637B526B844A637B526B844A637B526B844A637B
9C9C949C948C9C949494948C9C94949C948C9C94946B4A4A735252734A52735252734A4A
7B5252734A527B5252734A4A7B5252734A527B5252734A4A7B5252734A527B5252734A4A
7B52527352527B5252734A4A7B52527B52527B5252734A4A7B52527B52527B52527B5252
845A527B52527B52527B5252845A525A6B525A735252634A847B947B738C847B947B738C
847B947B7B8C847B947B7B8C848494637B7B6B8484637B7B6B8484637B7B6B847B637B7B
6B8484637B7B63847B637B736B847B637B7B637B7B637B736B847B637B73637B7B637B73
6B847B637B73637B7B637B73000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A5A734A5A73
4A5A734A637B4A5A734A5A7B4A5A7B4A637B4A5A734A637B4A637B4A637B4A637B4A637B
4A637B52637B4A637B4A637B4A637B52637B4A637B52637B4A637B526B844A637B9C9494
9C94949C94949C94949C94949C94949C94949C948C313152735252735252734A52735252
7352527B5252734A527352527352527B52527352527B52527B52527B52527352527B5252
7B52527B52527B52527B52527B52527B52527B52527B52527B52527B52527B52527B5252
7B52527B52527B52527B52525A73525A73525A6B527B7B947B7B8C847B947B7B8C847B94
847B94847B947B7B94847B94847B94848494637B7B637B7B637B7B637B7B637B7B637B7B
637B7B637B7B637B7B637B7B637B7B637B7B637B73637B7B637B7B637B7B637B73637B73
637B73637B7B637B73000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
4A637B425A734A637B4A5A734A637B4A5A734A637B4A5A734A637B4A5A7352637B4A5A73
52637B4A5A7352637B4A637B52637B4A637B526B844A637B526B844A637B526B84949494
9C949494948C9C949494948C9C949494948C9C949494948C7B5252734A527B5252734A52
7B5252734A527B5252734A527B5252734A527B52527352527B5252734A527B5252735252
7B52527352527B525A7B52527B52527B52527B525A7B52527B52527B5252845A5A7B5252
845A527B5252845A5A5A6B525A7352526B525A6B527B738C847B947B738C847B947B7B8C
847B947B7B8C8484947B7B8C8484947B7B8C6B8484637B7B637B7B637B7B63847B637B7B
637B7B637B73637B7B637B73637B7B637B73637B7B637B73637B7B5A7B73637B7B637B73
637B7B5A7373000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A5A734A5A734A5A734A5A734A5A734A5A734A637B4A5A734A637B4A637B4A637B
4A637B4A637B4A637B4A637B4A637B4A637B4A637B52637B4A637B52637B9C94949C9494
9C94949C94949C94949C94949C94949C94949C94949C94949C948C7352527352527B5252
7352527B52527352527B52527352527B52527B52527B52527B52527B52527B52527B5252
7B52527B52527B52527B52527B52527B52527B52527B52527B52527B52527B5252845252
7B52528452525A73525A73525A6B525A6B527B7B8C7B7B947B7B8C7B7B947B7B94847B94
7B7B94847B94847B94847B94847B94847B94847B94637B7B637B7B637B7B637B7B637B7B
637B7B637B7B637B7B637B7B637B73637B7B637B73637B7B637B73637B73637B73637B73
637B73000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000425A734A637B4A5A734A637B425A734A637B4A5A734A637B4A5A73
52637B4A5A734A637B4A5A7352637B4A637B52637B4A5A73526B844A637B9C9494949494
9C9C949494949C949494948C9C9C9494948C9C949494948C9C9C9494948C7B5252734A52
7B525A7352527B525A734A527B525A7352527B5252734A527B525A7B52527B525A735252
7B5A5A7B52527B525A7B5252845A5A7B525284525A7B5252845A5A7B5252845A5A7B5252
845A5A5A73525A735A526B525A7352526B4A7B7B947B738C847B947B7B8C847B947B738C
847B947B7B8C847B947B7B8C848494847B948484945A7B73637B7B637B7B637B7B5A7373
637B7B637B73637B7B5A7373637B7B5A7B73637B7B5A7373637B7B5A7373637B73000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A5A734A5A734A5A734A5A734A637B4A5A734A6373
4A637B4A637B4A5A734A637B4A637B4A637B4A637B4A637B4A637B52637B9494949C9494
9C94949C94949494949C94949C94949C94949494949C94949C94949C94943131527B5252
7B52527B525A7352527B52527B52527B52527B52527B52527B52527B525A7B52527B5252
7B52527B525A7B52527B52527B525284525A7B52527B52527B5252845A5A7B5252845252
8452525A735A5A6B525A6B525A6B527B7B947B738C7B7B947B7B94847B947B7B8C7B7B94
7B7B94847B947B7B94847B94847B94848494847B94848494637B7B637B7B637B73637B7B
637B7B637B7B5A7B73637B73637B73637B7B5A7373637B73637B73637B73000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004A637B425A734A637B425A734A637B425A73
4A637B4A5A734A637B4A5A734A637B4A5A7352637B4A5A7352637B9494949C9C9C949494
9C9C949494949C9C949494949C9C9494948C9C9C9494948C9C949494948C9C9C94313152
7B525A7352527B525A7352527B525A7352527B525A7B52527B525A7B52527B5A5A7B5252
7B5A5A7B5252845A5A7B5252845A5A7B5252845A5A7B5252845A5A7B5252845A5A7B5252
63735A5A6B525A735A526B525A6B527B738C7B7B947B738C847B947B738C847B947B7B8C
847B947B7B8C8484947B7B948484947B7B948484947B7B94637B7B5A7373637B7B5A7373
637B7B5A7373637B7B5A7373637B7B5A7373637B7B5A7373637B73000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A5A734A5A734A5A734A5A73
4A5A734A63734A5A734A637B4A5A734A637B4A63734A637B4A637B9C94949494949C9494
9494949C94949494949C94949494949C94949494949C94949494949C94949494949C9494
7B52527B525A7B52527B525A7B52527B525A7B52527B525A7B52527B525A7B52527B525A
7B52527B525A7B52527B525A7B525284525A7B5252845A5A7B525284525A7B52525A735A
5A73525A73525A6B525A6B527B738C7B7B947B7B8C7B7B947B7B947B7B947B7B94847B94
7B7B94847B947B7B94847B947B7B94847B94847B94848494847B94637B7B5A7B73637B7B
5A7373637B735A7B73637B735A7373637B735A7373637B73000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A6373425A73
4A637B4A5A734A637B425A734A637B4A5A734A637B4A5A739C9C9C9494949C9C94949494
9C9C9C9494949C9C9494948C9C9C949494949C949494948C9C9C9494948C9C949494948C
3131527B52527B525A7352527B5A5A7B52527B525A7B5252845A5A7B52527B5A5A7B5252
845A5A7B5252845A5A7B5252845A5A7B5252845A5A7B5252845A5A7B525263735A5A6B52
5A735A526B525A6B5273738C7B7B947B738C7B7B947B738C847B947B7B8C847B947B738C
8484947B7B94847B947B7B8C8484947B7B948484947B7B9484849C5A7373637B7B5A7373
637B7B5A7373637B735A7373637B7B5A7373000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A5A73
4A5A734A63734A5A734A5A734A5A734A637B4A5A734A63739494949C9494949494949494
9494949C94949494949494949494949C94949494949494949494949C949494948C949494
9494943131527B52527B525A7B525A7B525A7B52527B525A7B525A7B5A5A7B52527B525A
7B525A845A5A7B525284525A84525A845A5A7B5252845A5A84525A63735A5A73525A7352
5A6B525A6B52526B527B7B947B7B947B7B947B738C7B7B947B7B947B7B947B7B947B7B94
7B7B94847B947B7B94847B94847B948484947B7B948484948484948484945A73735A7373
5A7373637B735A73735A73735A7373000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000425A734A6373425A734A637B425A734A637B4A5A739C9C9C9494949C9C9C949494
9C9C949494949C9C949494949C9C9494948C9C9C9494948C9C9C9494948C9C9C9494948C
9C9C9494948C39315A7B52527B5A5A7B5252845A5A7B5252845A5A7B5252845A5A7B5252
845A5A7B5252845A5A7B5252845A5A7B5252845A5A7B5252637B5A5A73525A735A526B52
5A735252634A7B7B9473738C7B7B9473738C7B7B947B738C7B7B947B738C847B947B7B8C
847B947B7B8C8484947B7B948484947B7B948484947B7B9484849C7B7B94637B7B5A7373
637B735A7373637B735A7373000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A5A734A5A734A5A734A5A73949494949494949494949494949494
949494949494949494949494949494949494949494949494949494949494949494949494
94949494949494949431315A7B525A7B525A7B525A845A5A7B525A845A5A7B525A845A5A
7B525A845A5A84525A845A5A84525A845A5A84525A637B5A5A735A5A735A5A6B525A6B52
526B52526B527B738C7B7B947B738C7B7B947B7B947B7B947B7B947B7B947B7B947B7B94
7B7B947B7B947B7B94847B947B7B94847B947B7B94848494847B948484948484945A7373
5A73735A73735A7373000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004A6373425A734A63739494949C9C9C9494949C9C9C949494
9C9C9C949494949C949494949C9C9C9494949C9C9494948C9C9C94949494949C9494948C
9C9C9494948C9C9C9431294A845A5A7B525A845A5A7B5252845A5A7B525A845A5A7B5252
845A5A7B525A845A5A7B5252845A5A84525A637B5A5A735263735A5A6B525A735A526B52
5A6B5273738C7B7B9473738C7B7B947B738C7B7B9473738C7B7B947B7B947B7B947B7B8C
8484947B7B94847B947B7B9484849C7B7B948484947B7B9484849C7B7B9484849C5A736B
637B73000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000949494949494949494949C9C949494949494
949494949C94949494949494949494949C94949494949494949494949494949494949494
94949494949494948C949494313152845A5A7B525A845A5A845A5A845A5A7B525A845A5A
845A5A845A5A84525A845A5A845A5A845A5A5A735A5A735A5A735A5A735A526B52526B52
526B527B7B9473738C7B7B947B7B947B7B947B738C7B7B947B7B947B7B947B7B947B7B94
7B7B947B7B947B7B947B7B947B7B948484947B7B94848494848494848494847B94848494
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000009C9C9C9494949C9C9C9494949C9C9C949494
949C9C9494949C9C9C9494949C9C949494949C9C94949494949C9494948C9C9C9494948C
9C9C9494948C9C9C9494948C949C94313152845A5A7B525A845A5A7B525A845A5A7B525A
845A5A7B525A845A5A84525A845A5A5A735A637B5A5A73525A735A526B525A6B5252634A
7B7B9473738C7B7B9473738C7B7B9473738C7B7B9473738C7B7B947B7B8C7B7B947B7B94
7B7B947B7B948484947B7B948484947B7B9484849C7B7B9484849C7B7B9484849C7B7B94
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000949494949C9C949494949C94949494949C94
949494949494949494949494949494949494949494949C94949494949494949494949494
94949494949494949494949494948C94949431315A845A5A845A5A845A5A845A5A845A5A
845A5A845A5A845A5A845A5A5A735A5A735A5A735A5A735A5A6B525A6B52526B52526B52
73738C7B7B947373947B7B9473738C7B7B947B7B947B7B947B7B947B7B947B7B947B7B94
7B7B947B7B947B7B947B7B947B7B948484947B7B948484947B7B94848494848494848494
848494000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000949494949C9C949494949C9C949494949C9C949494
9C9C9C949494949C9C949494949C9C949494949C949494949C9C94949494949C9494948C
949C9494948C949C9494948C9C9C9494948C31315231315A845A6384525A845A5A7B525A
845A63845A5A845A5A5A735A637B635A735A5A735A526B525A735A526B52526B5273738C
7B7B9473738C7B7B9473738C7B7B9473738C7B7B9473738C7B7B947B7B947B7B947B7B8C
8484947B7B947B84947B7B9484849C7B7B948484947B7B9484849C7B7B9484849C7B7B94
84849C000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000949C9C94949494949C949494949C9C949494949494
949494949C9C949494949494949494949C94949494949494949494949C94949494949494
949494949C94949494949494949494949C9494948C313152845A5A845A5A845A5A845A5A
845A5A845A5A5A735A5A735A5A735A5A735A5A6B525A6B52526B52526B5252634A737394
7373947B7B9473738C7B7B947B7B947B7B94737B947B7B947B7B947B7B947B7B947B7B94
7B7B947B7B947B7B947B7B947B7B948484947B7B947B84947B849484849C7B8494848494
84849484849C000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000949C9C949494949C9C949494949C9C8C9494949C9C949494
949C9C949494949C9C949494949C9C8C9494949C9C949494949C94949494949C94949494
949C948C948C949C9494948C949C9494948C949C9494948C39315A84525A845A63845A5A
845A635A735A637B635A735A5A735A526B525A735A526B525A6B5252634A7B7B9473738C
7B7B9473738C7B7B9473738C7B7B9473738C7B7B9473738C7B7B94737B8C7B7B947B7B94
7B7B947B7B9484849C7B7B9484849C7B7B9484849C7B7B9484849C7B7B9484849C7B7B94
84849C7B8494000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000094949C949C9C94949C949C9C949494949C9C949494949C9C
949494949C94949494949C94949494949C94949494949C94949494949C94949494949C94
949494949494949494949C94949494949494949494949C9494948C31315A845A5A845A63
637B63637B635A735A5A735A5A735A5A735A526B52526B52526B5273739473738C737394
73738C737B947373947B7B947373947B7B947B7B947B7B947B7B947B7B947B7B947B7B94
7B7B947B7B947B7B947B7B947B7B947B84947B7B948484947B7B948484947B849484849C
84849484849C848494000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000949C9C8C9494949C9C8C9494949C9C949494949C9C8C9494
949C9C8C9494949C9C8C9494949C9C949494949C948C9494949C9C8C9494949C948C948C
949C94949494949C948C948C949C948C948C949C948C948C949C9431314A39395A5A7B5A
637B635A735A5A735A526B525A735A526B525A6B52526352526B5273738C737B9473738C
7B7B9473738C7B7B9473738C7B7B9473738C7B7B9473738C7B7B947B7B947B7B94737B94
7B849C7B7B947B84947B7B9484849C7B7B9484849C7B7B9484849C7B7B9484849C7B7B94
84849C7B849484849C000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000949C9C94949C949C9C949494949C9C949C9C949C9C949494949C9C
949494949C9C949494949494949494949C94949494949C94949494949C94949494949494
949494949C94949494949494949494949C94949494949494949494949C94313152637B63
5A7B5A5A735A5A735A5A735A5A6B5A5A6B5A526B52526B5273738C73739473738C737394
7373947B7B94737394737B94737B947B7B94737B947B7B947B7B947B7B947B7B947B7B94
7B7B947B7B947B7B947B7B947B7B947B849C7B7B947B84947B849484849C7B8494848494
84849C84849C84849484849C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008C9494949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494
949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C948C9494949C948C9494
949C948C9494949C948C9494949C948C948C949C948C948C949C948C948C949C94313152
637B635A735A5A735A526B525A735A526B525A6B5252634A73739473738C737B9473738C
7B7B9473738C7B7B9473738C7B7B9473738C7B7B947373947B7B94737B947B7B947B7B94
7B7B947B7B947B849C7B7B947B849C7B7B9484849C7B7B9484849C7B7B9484849C7B7B94
84849C7B849484849C7B8494000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000949C9C949C9C8C949C949C9C94949C949C9C94949C949C9C949494949C9C
8C9494949C9C949494949C9C949494949C94949494949C948C9494949C94949494949C94
949494949C94949494949C948C9494949C94949494949C94949494949C94949494949C94
31315A5A735A5A735A5A735A526B52526B52526B5273739473738C73739473738C737B94
737394737B947373947B7B947373947B7B94737B947B7B94737B947B7B947B7B947B7B94
7B7B947B7B947B7B947B84947B7B947B84947B7B947B849C7B7B947B849C7B849484849C
7B849484849C84849C84849C84849C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000949C9C8C9494949C9C8C949C949C9C8C9494949C9C8C9494949C9C8C9494
949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C948C9494
949C9C8C9494949C948C948C949C948C9494949C948C948C949C948C948C949C948C948C
3931525A735A5A735A526B525A6B5A526352526B526B738C737B9473738C737B9473738C
7B7B9473738C7B7B9473738C7B7B947373947B7B9473738C7B7B94737B947B7B94737B94
7B849C7B7B947B849C7B7B9484849C7B7B947B849C7B7B9484849C7B7B9484849C7B7B94
84849C7B849484849C7B8494848C9C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008C949C949C9C949C9C949C9C8C949C949C9C8C9C9C949C9C8C9494949C9C
949C9C949C9C8C9494949C948C9C94949C9C8C9494949C94949C94949C948C9494949C94
8C9494949C948C9494949C94949494949C948C9494949C948C9494949C948C948C949494
949494313152526B52526B52526B52526B5273738C73739473739473739473738C737394
737394737B94737394737B94737B947B7B94737B947B7B947B7B947B7B94737B947B7B94
7B7B947B7B947B7B947B7B947B7B947B849C7B7B947B84947B849484849C7B84947B849C
7B849C84849C7B849484849C84849C84849C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C949C949C9C8C949C949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494
949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494
949C948C9494949C948C9494949C948C9494949C948C9494949C948C948C949C948C948C
949C948C948C5A735A526B52526B524A634A7373946B738C737B946B738C737B9473738C
737B9473738C7B7B9473738C7B7B947373947B7B94737B947B7B94737B947B7B9C737B94
7B849C737B947B849C7B7B947B849C7B7B9484849C7B7B9484849C7B7B9484849C7B8494
84849C7B849484849C7B8494848C9C7B8494000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000949C9C8C9C9C949C9C8C9C9C949C9C8C9C9C949C9C8C949C949C9C8C9C9C949C9C
8C9494949C9C8C9C94949C9C8C9494949C9C8C9494949C948C9494949C948C9494949C94
8C9494949C948C9494949C948C9494949C948C9494949C948C9494949C948C9494949C94
8C9494949C948C9494526B5252635273739473738C73739473738C737394737394737B94
737394737B94737394737B94737B94737B94737B947B7B94737B947B7B94737B947B7B94
7B7B947B7B947B7B947B84947B7B947B84947B7B947B849C7B84947B849C7B849484849C
7B849484849C7B849C84849C7B849C84849C84849C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
949CA58C949C949C9C8C949C949C9C8C949C949C9C8C9494949C9C8C9494949C9C8C9494
949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494
949C9C8C9494949C948C9494949C948C9494949C948C948C949C948C9494949C948C948C
949C948C948C949C944A634A526B526B738C7373946B738C737B9473738C737B946B738C
737B94737394737B9473738C7B7B947373947B7B947373947B7B9C737B947B7B94737B94
7B849C737B947B849C737B947B849C7B7B947B849C7B7B9484849C7B7B9484849C7B7B94
84849C7B849484849C7B8494848C9C7B849C848C9C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C9C9C949C9C8C9C9C8C9C9C8C9C9C949C9C8C949C8C9C9C8C9C9C949C9C8C949C8C9C9C
8C9C9C949C9C8C94948C9C9C8C9C9C949C9C8C94948C9C948C9C94949C9C8C94948C9C94
8C9C94949C948C94948C9C948C9C94949C948C94948C9C948C9C94949C948C94948C9C94
8C9494949C948C948C735252848C847373946B738C73739473739473739473738C737394
737394737B94737394737B94737B94737B94737B94737B94737B947B7B94737B947B7B94
7B7B947B7B947B7B947B7B947B7B947B849C7B7B947B849C7B84947B849C7B84947B849C
7B849C84849C7B849C84849C84849C84849C7B849C84849C000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
949CA58C949C949CA58C949C949C9C8C949C949C9C8C949C949C9C8C9494949C9C8C9494
949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9494
949C9C8C9494949C9C8C9494949C948C9494949C948C9494949C947352525A63845A5A84
5A63845A638463638C000000000000848C848C948C848C848C9484848C848C9484848C7B
848C84848C7B848C84737394737B947373947B7B94737B947B7B94737B947B849C737B94
7B849C737B947B849C737B947B849C7B7B947B849C7B7B947B849C7B7B9484849C7B8494
84849C7B8494848C9C7B8494848C9C7B8494848C9C7B849C000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008C9C9C
8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C
8C9C9C8C9C9C8C9C9C8C9C9C8C9C948C9C9C8C9C948C9C9C8C94948C9C948C9C948C9C94
8C9C948C9C948C9C947B5252525A7B5A5A7B5A5A7B5A63845A5A7B5A63845A63845A6384
5A63845A6384000000000000000000000000848C84848C84848C84848C84848C84848C84
848C84848C84848C84848C84848C7B848C7B848C7B848C7B848C7B848C7B7B847B848C7B
7B847B7B7B947B7B947B849C7B7B947B849C7B84947B849C7B84947B849C7B849C7B849C
7B849C7B849C7B849C84849C7B849C84849C84849C848C9C84849C000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008C949C
949CA58C9C9C8C9C9C8C949C949CA58C949C8C9C9C8C9494949C9C8C949C8C9C9C8C9494
949C9C8C94948C9C9C8C9494949C9C8C94948C9C9C8C9494949C9C735252525A7B525273
5A5A7B525A7B5A5A7B525A7B5A6384525A7B5A6384525A7B5A63845A5A7B5A63845A5A7B
6363840000000000000000000000000000000000000000008C948C848C84848C84848C7B
848C84848C7B848C847B847B848C847B847B848C847B847B848C847B847B848C7B7B8473
848C7B7B8473848C7B7B8473848C7B7B84737B847B7B84730000000000000000007B7B94
84849C7B849484849C7B8494848C9C7B849C848C9C7B849C848CA5000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C9C9C8C9C9C
8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C
8C9C9C8C9C9C8C9C94735252525273525A73525273525A73525A73525A7B525A73525A7B
525A7B5A5A7B525A7B525A7B525A7B5A6384525A7B5A63845A63845A63845A5A845A6384
5A6384000000000000000000000000000000000000000000000000848C84848C84848C84
848C84848C84848C7B848C84848C7B848C847B8C7B848C7B848C7B848C7B7B847B7B8C7B
7B847B848C7B7B84737B847B7B84737B847B7B8473000000000000000000000000000000
000000000000000000000000000000000000000000000000000000848C9C000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C9CA58C9C9C
8C9CA58C949C8C9CA58C949C8C9C9C8C949C8C9C9C8C949C8C9C9C8C949C000000000000
0000000000000000000000000000004A5273525A7B4A5273525A7B525273525A7B525A73
5A5A7B525A735A5A7B525A7B5A6384525A7B5A6384525A7B5A6384525A7B5A63845A5A84
000000000000000000000000000000000000000000000000000000000000849484848C7B
848C847B8C7B848C847B847B848C847B847B848C847B847B848C7B7B847B848C7B7B8473
848C7B7B8473848C7B7B84737B8C7B7B84737B8C7B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C9C9C8C9CA5
8C9C9C8C9C9C8C9C9C8C9C9C000000000000000000000000000000000000000000000000
000000000000000000000000000000525273525273525A73525273525A7B525A7B525A7B
525A7B525A7B525A7B5A5A7B525A7B5A5A7B5A5A7B5A63845A5A7B5A63845A6384000000
000000000000000000000000000000000000000000000000000000000000000000848C84
848C84848C84848C7B848C84848C7B848C7B7B8C7B848C7B7B847B848C7B7B847B7B8C7B
7B847B7B847B7B84737B847B7B84737B84737B8473000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000008C9C9C000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A5273525A7B4A5273525A7B525273525A7B525273
5A5A7B525A7B5A5A7B525A7B5A6384525A7B5A6384525A7B5A63845A5A7B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
848C84848C7B848C847B847B848C847B847B848C847B847B848C847B847B848C7B7B8473
848C7B7B84737B8C7B7B8473848C7B7B84737B847B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000525273525A73525A73525A7B525273525A7B
525A7B525A7B525A7B525A7B525A7B5A5A7B525A7B5A5A7B5A5A84000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000848C847B8C7B848C84848C7B848C847B847B7B8C7B7B8C7B848C7B7B847B7B8C7B
7B847B7B8C7B7B84737B84737B84737B847B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000525A7B4A5273525A7B4A5273525A7B525273
525A7B525A735A5A7B525A7B5A637B525A7B5A6384525A7B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007B847B848C847B847B848C847B847B848C7B7B847B848C7B7B8473
7B8C7B7B84737B8C7B7B84737B8C7B7B8473000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000525273525A73525273525A7B525A73525A7B
525A7B525A7B525A7B525A7B525A7B525A7B525A7B5A6384000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B8C7B848C7B7B8C7B7B8C7B7B847B7B8C7B7B847B7B8C7B
7B847B7B847B7B84737B847B7B84737B8473000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A5273525A7B525273525A7B4A5273
5A5A7B525A7B525A7B525A7B5A6384525A7B5A6384000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B847B848C847B847B848C847B847B7B8C7B7B8473
848C7B7B84737B8C7B7384737B8C7B7B8473000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000525273525A73525A7B525273525A7B
525A7B525A7B525A7B525A7B525A7B5A5A7B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007B847B7B8C7B7B8C7B7B8C7B7B847B7B847B
7B847B7B8C7B7B84737B847B7B84737B847B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A5273525A7B4A5273525A7B525273
525A7B5252735A5A7B525A7B5A5A7B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007B847B848C7B7B847B7B8C7B7B8473
7B8C7B7B84737B8C7B7B84737B8C7B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000525273525A73525A73525A7B
525A7B525A7B525A7B525A7B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000007B847B7B8C7B7B847B7B8C7B
7B847B7B847B7B84737B847B7B8473000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000525A7B4A5273525A7B4A5273
525A7B525A73525A7B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007B847B7B8C7B738473
7B8C7B7B84737B8C7B7384737B8C7B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000525273525A7B525273525A7B
525A7B525A7B525A7B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B847B
7B847B7B847B7B84737B84737B8473000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A5273525A7B4A5273
525A7B525273000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7B8C7B7384737B8C7B738473000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000525A73525273525A7B
525A73000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007B847B7B84737B847B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A5273525A7B4A5273
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B8C7B738473000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A5273000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007B8473000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[54]:=",
  ImageSize->{128, 128},
  AnimationDisplayTime->0.169,
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000200000P0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj2LW
9`00?g=cLda<C240<410Y/30`9T0IPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm?ooc020P800082006IVc3<cIP2009VIod:J
Yb8R8_mV<`00P00g??mVc5J0Zk75fBa1EkV>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7De3eDE:b/[:BT
Y>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEeMJ_0eLC4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7i/;2`W9bLP00
o`1o001oO`1oo`3o003oO`3oogl007l0Ogl0ogmo07moOgmoogoo07ooOgooool00?l0Ool0oomo0?mo
Oomooooo0?ooOoooo`0002XZ:UEEEGmoOjZZZ]GEeOoooh0700201`00P0L0080700031gT010L000<7
N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@7
00031gT010L000<7N@041`000`M<0003Wh4002X010L000<7BP000inJX00/00@700031bP000>10000
7P001J2OX9nP02/010L000<7:002PAh000JPVZ2JX9X[00@700031bL000BKPI^170001j2OX9nPWj00
:`041`000`LW00>100>200006@002J2JX9ZPVZ2JX00?0003X00001P010L000<79`001Y^1Vh6KPAL0
00^OX9nPWj2OX9nPVP0<0004^J2iX1X010L000<79`001h62PH:1Vh405@003IZPVZ2JX9ZPVZ2JX9X0
1`0029ZPVZ2JX9ZO6P041`000`LV0009PI^1Vh6KPI^101<000jOX9nPWj2OX9nPWj2JX0@000>KX9/0
0j0019^iX;TK00@700031bH000^2PH:1PX6KPI^1V`0@000MVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2JX9ZP
VZ2JX9X06`041`000`LV000<PI^1Vh6KPI^1Vh6K3P006InPWj2OX9nPWj2OX9nPVj2KX9^PVj2KX9/0
0Z000kVPW`0K00@700031a8000>PQ`004@003H:1PX62PI^1Vh6KPI/030007iZPVZ2JX9ZPVZ2JX9ZP
VZ2JX9ZPVZ2JX9ZPVZ2JX8006`041`000`LC0008^J2iX;VP^J0:000@PX6KPI^1Vh6KPI^1X86KP@P0
00ZPWj2OX9nPWj2O1j001I^PVj2K00>P00>KX:000Z000kV1000K00@700031a<0012JX9ZPVZ2JX9nP
VZ2OX9ZP0P004H62PH:1PX6KPI^1Vh6KPI^100H00261VZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2J
X9ZPVil070041`000`LD000FX;VP^J2iX;VP^J2iX;VP^J2iX;VP^@:P00ZJX9ZPVZ21Vh6K0X430003
PInP00:P01FKX9^PVj2KX9^PVj2KX9^PVj2KX9/01:001;VPX84L00@700031aD01:006iZPVZ2JX9ZP
VZ2JX9ZPVZ2JX9ZPWj2OX9nPW`02X00VVX2JP:21X86PPJ21X86PPJ21X9ZPVZ2JX9^PVj2KX9^PVj2K
X84M00@700031aD003JiX;VP^J2iX;VP^J2iX;VP^J2iX;VP^J2iX;VP^J2iX:JPVX6JVj2KX9^PVj2K
X9^PVj2KX9/@X003W`0001/010L000<75P03X00EVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2O00>P00BOX9nQ
0j0099Z0PH6PPJ21X86PPJ21X9ZPVZ2JX9ZPVj2KX9^PVj2KX9^6P1d010L000<75P00=j2iX;VP^J2i
X;VP^J2iX;VP^J2iX;VP^J2iX;VPYZ2QX:JPVX6JVj2KX9^PVj2KX9^PVj2KX9/03J000h6O000L00@7
00031aL00Z004iZPVZ2JX9ZPVZ2JX9ZPVZ2JX9X02Z009J6PX:2JP9Z1X86PPJ21X9ZPPJ2JX9ZPVZ2K
X9^PVj2KX9^PVh407P041`000`LG000S^J2iX;VP^J2iX;VP^J2iX;VP^J2iX;VP^J2iX:6PYZ2QX:H0
0YX03h6PVj2KX9^PVj2KX9^PV`0>X003Wh4001d010L000<760005:2JX9ZPVZ2JX9ZPVZ2JX9ZPWj2J
3j007X2JPJ21X86PPJ21X9ZPVZ2JX9ZPVj2KX9^PVj2KX0:17P041`000`LH000aX;VP^J2iX;VP^J2i
X;VP^J2iX;VP^J2iX:6PXJ2QX:6PXJ2VPIZKX9^PVj2KX9^PV`0?X003PIn101h010L000<76@001IZP
X:2J00>P00^JX9nPVZ2OX9ZPW`0AX00MNiZ1X86PPJ2JX9ZPVZ2JX9^PVj2KX9^PVj2KQX007`041`00
0`LI000^^J2iX;VP^J2iX;VP^J2iX;VP^J2iX;VPXJ2QX:6PXJ2VX:6PVX6JVj2KX9^PVa6P00>1W`00
7P041`000`LI00ZP00VOX9nPWj2OX9l04:007Z6PVW^PPJ21X86PVZ2JX9^PVj2KX9^PVj2KX866P1l0
10L000<76@00;86P^J2iX;VP^J2iX;VP^J2iX;VP^J2QX:6PXJ2QX:6PXJ2QX:JPVX6PVj2K4Z000in1
QP0O00@700031aT000=kX:000Z000kVPX004X006WkVPX9ni4J006j6PXHNJNj21X9^PVZ2KX9^PVj2K
X9^PVj2KX002PB0010L000<76P00:J2iX;VP^J2iX;VP^J2iX;VP^J2iX:6PXJ2QX:6PXJ2QX:6PYZ2V
PIZ101>P00>1Wh4080041`000`LJ00JP00^iX:2P^J2iX;VP^@0AX00MXHNQQj60VX:PVj2KX9^PVj2K
X9^PVj2KX9^6P84080041`000`LK000Y^J2iX;VP^J2iX;VP^J2iX;VPXJ2QX:6PXJ2QX:6PXJ2QX:JP
YZ2VPIX04Z000h66P@0P00@700031a/00j003KVPX:2iX;VP^J2iX;T00Z000hNPQ`03X003Qj2P00JP
01VQQj67XHNQNiZKX9^PVj2KX9^PVj2KX9^P00:100B0VX2J7P041`000`LL000XX;VP^J2iX;VP^J2i
X;VP^J2QX:6PXJ2QX:6PYZ2VX:JPYZ2QX:JPYP:J3Z0029ZOPIZ1VX6J0X400g/0000H00@700031ad0
4:004hNPQj27X8NPQj27X8NPQj27X8L01Z006YZKX9^PVj2KX9^PVj2KWh2JP9Z0VX2JP9Z06@041`00
0`LM000[^J2iX;VP^J2iX;VP^J2QX:6PXJ2QX:6PXJ2QX:JPXJ2QX:6PXJ2QX:6PVP0;X00<^H6OPIZ1
VX6JPIZ10iXG00@700031a`000>1Nj003:006HNPQj27X8NPQj27X8NPQj27X8NPQj27X8L00j007HNJ
Vj2KX9^PVj2KX866P9Z0VX2JP9Z0VX2JP9Z001D010L000<76P04P@0[X;VP^J2iX;VP^J2iX:6PXJ2Q
X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QP@0:X003Wh6J00:J00N1VX6JPIZ100NJ4`041`000`LJ0005
PG^1Nh402J00@;VPX8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X82PVj2KX9^PVj20VX2JP9Z0
VX2JP9Z0VX2JP9Z0VX0A00@700031aP01h40;;VP^J2iX;VP^J2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:611j000h6OVP0EVPl010L000<75P002X5kPG^1Nh5kPG/2X005^J2PX;T00Z00@hNP
Qj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj20X9^PVj2KQX2JP9Z0VX2JP9Z0VX2JP9Z0VX2J
P9Z0VX003@041`000`LE0003VX6J00F1032JPIZP^J2iX;VP^J2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PXH44X003WiZJ01ZJ2`041`000`LC000DNh5kPG^1Nh5kPG^1Nh6P^J2iX;T2X017
Qj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPVj2KX86JP9Z0VX2JP9Z0VX2JP9Z0VX2J
P9Z0VX2JP9Z0VX002@041`000`LA0007VX6JPIZ1VP03P@0fVX6JPIZ1^J2iX;VPXJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6P89X800@700031`l006AkPG^1Nh5kPG^1Nh5kPG^1
Nh5k^J2iX;V7X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPVin0VX2JP9Z0VX2J
P9Z0VX2JP9Z1VX2JP9Z0VX2JP9Z0VX2J1P041`000`L>0012PIZ1VX6JPIZ1VX6JPIZ1VX6JPIZP^J2i
X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PYX6Q9IX400@700031``001V1Nh5k
PG^1Nh5kPG^1Nh5kPG^1NiYlVZ2i00:P04j7X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NP
Qj27X8NPViZ0VX2JP9Z0VX2PQj27X8NPQj27X8NPQj27X8NPQj27X8JPQZ0400@700031`X003Z1VX6J
PIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1^J2QX:6PXJ2QQj6PXHNQX:6PXJ2QX:6PXJ2QX:6PXJ2Q2iX01J6P
XJ2V00>P00>VX:005Z0018NPQj0500@700031`T003=kPG^1Nh5kPG^1Nh5kPG^JNh5lPG^JO9YkVWbi
Qj27X8NPQj27X8NPQj21VX6JPIZ1VX400iX00h6JVP0<VP0TWj2PQj27X8NPQj27X8NPQj27X8NPQj27
X8NPQj27X8NPQj261`041`000`L8000LPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1X0VJ00nPVZ2J
X9ZPVZ2JX9ZPVZ003iX01inJWj2QX:401Z001hNPX:27X8L04J000hL0000600@700031`H000J1Nh5k
PG/2P@0:PX62PH:1PX62P@:200R7PXN2Qh:7PP:7022iVYZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1
VX6JP@>J00>1VYX00YX00in1X002X00OPJ27X8NPQj27X8NPQj27X8NPQj26X8JPQZ26X8JPQP0900@7
00031`<002N7PXN2Qh:7PXN2Qh:7PXN2Qh:7PZ22Qh:PPZ27X8:PQknP_h6PVY/00iX01I^JVYZK00>J
00FPVY^JX003VP0FX9ZPVZ2JX9ZPVZ2JX9ZPQZ6PXJ2QP@fP00j7X8NPQj27X8NPQj27X0/010L000<7
1@00:H67PHN2Qh:7PXN2Qh:7PXN2Qh:7PXN2Qh:7PXN2^J2iX;V1VX6JPIZ100>J00>1VYX02IX01H6J
PIZ100VJ00>1Wi/01J006XNPQj27X8NPQj27X8NPQj26X8JPQZ26X8JP30041`000`L7000JQh:7PXN2
Qh:7PXN2Qh:7PXN2Qh:PPZ22X882X00X_j2oX;nJX9ZKVY^JX9ZKVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2J
X9ZPQP:P00NQX:6PXJ2Q00BP01B7X8NPQj27X8NPQj27X8NPQj27X0d010L000<7200098N2Qh:7PXN2
Qh:7PXN2Qh:7PXN2Qh:7PXN2_Z2iX;VP^H6JPA>J00>1VYX00YX02YnJVYZOPIn1X9/7X00EQj27X8NP
Qj27X8NPQZ26X8JPQZ2600l010L000<72@00BhN2Qh:7PXN2Qh:7PXN2Qh:7PZ22X8:PP[nP_j2oX;nP
PYZKVZ2JViZPVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2JX8JPQZ6PXJ2QX:6PX@02X00CQj27X8NPQj27X8NP
Qj27X8NPQ`0@00@700031`/00262Qh:7PXN2Qh:7PXN2Qh:7PXN2Qh:7X;VP^J2iX;VPVX404YX049^J
VYZPVYnJX86OPInKX9/7X00CQj27X8NPQj27X8JPQZ26X8JPP@0A00@700031``001>2Qh:7PXN2Qh:7
PXN2Qh:7PZ2200:P00VoX;nP_j2oX;l00Z007IZPVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPQZ2600:P
01bQX:6PXJ2QX:6PXHNPQj27X8NPQj27X8NPQj274P041`000`L>000MQh:7PXN2Qh:7PXN1Qh:7PHN2
^J2nX;VP_Z2iX;h04iX03Y^JX9ZPVZ21X86OPJ2K2j003XNPQj27X8NPQZ21X86P50041`000`L@000?
PXN2Qh:7PXN2Qh:7PXN200:P00VoX;nP_j2oX;l00Z00:YZPVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2JX9ZP
QZ26XJ2QX:6PXJ2QX:6PX@:P00b7X8NPQj27X8NPQj0E00@700031a4001b2Qh:7PXN2Qh67PHN1Qj2o
X;jP_Z2nX;jP^J214IX04:2JX9ZPVZ21X86OVj2KX9/9X00<Qj27X8NPQj21X86P5P041`000`LC000I
Qh:7PXN2Qh:7PXN2_j2oX;nP_j2oX;nP_`02X00IVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2JX8JPQP02X00I
XJ2QX:6PXJ2QX:6PXJ2QQj27X8NPQj27X00G00@700031aD001J1Qh:7PHN1Qh6oX;jP_j2nX;jP^J2n
0Z000aVJVP0>VP0>X9ZPVZ21X86OPZ2KX9/=X007Qj27X8JPP@0I00@700031aH004:2Qh:7PXN2Qj2o
X;nP_j2oX;nP_j2oX;n1X9ZPVZ2JX9ZPVZ2JX9ZPVZ2JX8JPQZ26XJ2QX:6PXJ2QX:6PXJ2QX:42X006
Qj27X8N16@041`000`LH000CQh67PHN1_j2oX;jP_Z2nX;jP_P04X0036IZJ00bJ012PVZ21X86OPJ22
X8:PVj2K2j001XNPQZ26X1X010L000<76@00;862Qh:oX;nP_j2oX;nP_j2oX;nP_j2oPJ2JX9ZPVZ2J
X9ZPVZ2JX8JPQZ260Z005j6PXJ2QX:6PXJ2QX:6PXJ2QX:67X8NP01/010L000<76P004HN1Qj2oX;nP
_j2UX;nP_Z2n00FP00>1VY/00iX01I^JVYZK00>J00jPPJ21X86OPZ22X9^PPPnP00>600006`041`00
0`LL000DX;nP_j2oX;nP_j2oX;nP_j2oX;l2X00YPJ2JX9ZPVZ2JX9ZPQZ26X8JOQj6PXJ2QX:6PXJ2Q
X:6PXJ2QX:6PXJ007P041`000`LL01NP014IX9ZPVZ2JX9ZPPJ21Wh66P@0HX1d010L000<76`00@KnP
_j2oX;nP_j2oX;nP_j2oX;nP_j2oX8:1X9ZPVZ2JX8JPQZ26X8JQX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6P
XJ2Q01d010L000<76`0CX00FYJ2PX:FPPIZPVZ2JX86PPIn1QX66PQBP00FQX:2PX@0L00@700031aX0
04>P_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j22VZ2JX8JPQZ26X8JOX:6PXJ2QX:6PXJ2QX:6PXJ2Q
X:6PXJ2QX:6P01`010L000<76P000j6PX@0?X005YJ2PX:D00j004ZFPPYZPQZ21X866PHJ2X8:PQaBP
00BQX:6P6`041`000`LJ0014X;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_h62QZ26X8JPQZ26XJ2Q
X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:4K00@700031aT000>1XJ000Z000jJPX00=X005YJ2PX:D0
0j003ZDIX8JPPJ21QX6PPZ220j000hNPX00:X005XJ2PX:400j000j6PX@0J00@700031aT004J1X;nP
_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oPJ26X8JPQYnPXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6P
XJ2QX:6P6P041`000`LI0006X:6PXJ2Q3J006ZFPYJ2UX:FPYJ2UX:DIX866PHJ2X8:PQj274:002Z6P
XJ2QX:6PXH4I00@700031aT004NoX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX8:6X8JPQZ6P
XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PP@0I00@700031aT000JPYZ2QX:H3X005YZ2PX:D0
0j003JFPYJ2UX:FPYJ2UX:D00j02P@0<QX6PPZ22X8:PQj271J000hNPX004X00<XJ2PX:6PXJ2QX:6P
6@041`000`LH000RX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_`:P0XH08Z2QX:6PXJ2Q
X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:HI00@700031aP000VVX:JPYZ2QX:H00j000jFPX005X00N
YJ2UX:FPYJ2UX:FPYJ2UX8J2X8:PPZ27X8NPQj270j000hNPX006X00;XJ2QX:6PXJ2QX:4060041`00
0`LG001:_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_h6QX:6PXJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PXJ2QX:6PYZ0H00@700031aL000bPYZ2VX:6PYZ2QX:H3X00EYZ2PX:FPYJ2UX:FP
YJ2UX:FPYJ2U00:P0162X8:PPZ22X8NPQj27X8NPQ`06X00>XJ2QX:6PXJ2QX:6PXJ0G00@700031aL0
04^oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oVZ22X8:Q0:FPYJ2UX:FPYJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PYZ2VX:405`041`000`LF000CYZ2VX:JPYZ2VX:JPYZ2VX:FPY@03X00?YJ2UX9Z2
X8:PPZ22X8:P00<05:004J6PXJ2QX:6PXJ2QX:6PXJ2Q01H010L000<75@003h6P_j2oX;nP_j2oX;nP
_`08X00>PZ22X8:PPZ22X8:PPZ06000DPJFPYJ2UX:FPYJ2UX:FPYJ2UX:D2X004P@0000Z15@041`00
0`LE000;X:JPXJ2VX:JPYZ001`004X67PHN1Qh:7PXN2Qh:7PXN2X0T05:0S00@700031aD000FoX;nP
_`0=000AX8:PPZ22X8:PPZ22X8:PPZ002`002jFPYJ2UX:FPYJ2U00RP8`041`000`LD0003YP000100
00F1Qh67P@02PP09Qh:7PXN2Qh:P00d04Z0S00@700031bP000j2X8:PPZ22X8:PPZ22X0l000VUX:FP
YJ2UX:D01j0T00@700031bP000B7PHN10X801hN2Qh:7PXL04P0>X2@010L000<7:0003H:PPZ22X8:P
PZ22X8804`001Z2UX:FPY@NP90041`000`LY0003PHN100:200J7PXN2Qh8E00bP90041`000`LY000:
X8:PPZ22X8:PPQL000BPYJ2U1j0T00@700031bT000>1Qh4018800hN2000H00VP9@041`000`LZ0007
PZ22X8:PPP0K0003X:FP00FP9@041`000`LZ0006Qh62PXN27@07X2D010L000<7:P001X:PPZ22X1h0
1Z0U00@700031b/000B1PX:28004X2H010L000<7:`000j22X00R00>P9P041`000`L[0003PH8002<0
0Z0V00@700031b/000>1000090000j00000T00@700031gT010L000<7N@041`000`Mi00@700031gT0
10L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700201`00P0L008070000\
\>"],
  ImageRangeCache->{{{0, 127}, {127, 0}} -> {-3.20067, -3.20067, 1.05828, 
  1.05828}},
  CellTags->"Animation"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
%%ImageSize: 128 128 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00744048 0.0238095 0.00744048 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
384 string
128 128 8 [128 0 0 128 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007B635A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007B635A846B63000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000008463637B635A7B635A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000846B637B635A846B637B635A
00000000000000000000000000000063526B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000846B637B635A8463637B6363846B63
7B635A00000000000063526B63526B63526B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000846B637B635A846B637B635A846B637B635A
846B637B635A63527363526B6B5A737B6B527B7352000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000846B637B6363846B63846363846B637B6363846B63
84636363526B63526B63526B63526B8473527B6B52000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000846B637B635A846B637B6363846B637B635A846B6363526B
63527363526B6B5A7363526B6B5A737B6B528473527B6B4A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000846B63846363846B63846B63846B63846363846B6363526B63526B
63526B63526B63526B63527363526B7B73527B73527B7352000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000846B637B6363846B6B846363846B637B6363846B635A526B63526B5A526B
635A7363526B635A7363526B6B5A737B6B527B7352736B4A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000846B63846363846B63846B63846B63846B6363526B63526B63526B63526B63526B
63526B63527363526B6352736352737B73527B6B527B6B527B6B52000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
846B6B7B6363846B6B846363846B6B84636363526B5A526B63526B5A526B63527363526B
635A7363526B6B5A7363526B6B5A737B6B527B7352736B4A7B7352000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000846B6B
846B63846B63846B63846B63846B6363526B5A526B63526B5A526B63526B63526B63526B
63526B63527363526B6352737B73527B73527B6B527B7352736B527B6B52000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B6363
846B6B846B63846B6B5A4A6363526B5A4A6B63526B5A4A6B6352735A526B6352735A526B
635A7363526B635A7363526B84735A7B6B527B7352736B4A7B7352736B4A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000846B63846B63
846B63846B6B5A4A6B5A526B5A526B63526B5A526B63526B63526B63527363526B63526B
63526B63527363526B6352737B73527B73527B6B527B73527B6B527B6B52000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39394A39314A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000846B63846B6B846B63
846B6B5A4A6363526B5A4A6B63526B5A4A6B6352735A526B6352735A526B635A7363526B
635A7363526B6B5A7363526B84735A7B6B527B7352736B527B7352736B4A7B6B52000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39314A39314A39314A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000846B63846B6B846B6B5A526B
5A4A6B5A526B5A526B5A526B5A526B63526B5A526B63526B63526B63526B63526B635273
63526B635273635273635A737B73527B73527B73527B7352736B52736B52736B4A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39394A39314A39394A39314A4A3131000000000000000000000000000000000000000000
000000000000000000000000000000000000000000846B638C736B846B635A526B5A4A63
63526B5A4A6B5A526B5A4A6B6352735A526B6352735A526B6352735A526B6352735A526B
635A7363526B635A7363526B84735A7B6B527B7352736B4A7B7352736B4A736B52736B4A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39314A39314A39314A39314A39314A423952422929000000000000000000000000000000
000000000000000000000000000000000000846B6B846B6B846B6B5A526B5A4A635A526B
5A526B5A526B5A4A6B5A526B5A526B63526B5A526B63526B63526B63527363526B635273
635273635A736352736352737B73527B73527B6B527B7352736B527B6B52736B4A736B52
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39394A39314A39395239314A42395239314A423952422929000000000000000000000000
000000000000000000000000000000846B638C736B524A635A526B524A635A526B5A4A63
5A526B5A4A6B63526B5A4A6B6352735A526B6352735A526B6352735A526B635A7363526B
635A7363526B6B5A736352737B735A736B527B7352736B527B7352736B4A7B7352736B4A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39314A39314A39314A39395239314A3939523939524239524A2929422929000000000000
000000000000000000000000846B6B846B6B5A4A635A526B5A4A6B5A526B5A4A6B5A526B
5A4A6B5A526B5A526B5A526B5A526B63526B5A526B63527363526B635273635273635273
6352736352736352737B735A7B73527B73527B73527B7352736B52736B52736B4A736B52
736B4A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39395239314A39395239314A42395239314A42395239314A4239523931524A3131422929
000000000000000000846B6B6B7B8C524A635A526B524A635A526B5A4A635A526B5A4A63
63526B5A4A6B63526B5A4A6B6352735A526B6352735A526B635A7363526B635A735A526B
635A73635273635A737B735284735A7B73527B7352736B4A7B7352736B4A736B526B634A
736B52000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39314A39394A39314A39314A39314A4239523931523939524239524239523939524A3131
422929000000846B6B524A63524A635A526B524A635A4A6B5A4A6B5A526B5A4A6B5A526B
5A526B5A526B5A526B5A526B5A526B6352735A526B635273635273635273635273635273
635273635A736352737B73527B73527B7352736B527B7352736B52736B52736B4A736B52
736B4A736B4A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39394A39314A42395239314A42395239314A42395239315242395239315242395A393952
5231314229295A526B524A635A526B524A635A526B524A635A526B524A635A526B5A4A6B
5A526B5A4A6B6352735A526B6352735A526B6352735A526B635A735A526B635A7363526B
635A736352736B5A737B73527B735A736B527B7352736B527B7352736B4A7373526B6B4A
736B526B634A0000000000000000000000000000000000004231314A4A31000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39314A39314A39314A39395239314A393952393952423952393952423952634231634231
634231524A63524A635A4A6B524A635A4A6B5A4A6B5A526B5A4A6B5A526B5A526B5A526B
5A526B5A526B5A526B63526B5A526B6352735A526B635273635273635273635273635A73
635273635A736352737B735A7B73527B73527B73527B7352736B52736B52736B4A736B52
6B6B4A736B4A000000000000000000422929392929422929422929000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39395239314A6B4A316342316B4A396342316B4A316342296B4A31634231634231634229
634A316342315A526B524A635A526B524A635A526B524A635A526B5A4A6B5A526B5A4A6B
6352735A526B6352735A4A6B6352735A526B6352735A526B635A735A526B635A735A526B
6B5A73635273635A737B73527B735A7373527B7352736B4A7B7352736B4A736B526B6B4A
736B526B634A736B4A392929423131392929423131392929525239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000734A31734A31734A31734A316B42316B4A31
6B4A316B4A316B42316B42316B42316B4A316B42316B42316342316B4231634231634231
634231634231634231524A6B524A6B5A526B524A6B5A526B5A526B5A526B5A4A6B5A526B
5A526B5A526B5A526B5A526B5A526B6352735A526B635273635273635273635273635273
635273635A736352737B73527B73527B7352736B52737352736B52737352736B4A736B52
736B4A736B4A6B634A4229294229294231314229295252394A4A31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000084737B7B6B73846B737B6B7373849C6B7B94734A396B4231734A396B4231
734A396B4231734A396B42316B4A316B42316B4A316342316B4A316342316B4A31634231
6B4A316342316B4A316342315A526B524A635A526B524A6B5A526B5A4A6B5A52735A4A6B
5A52735A4A6B6352735A526B6352735A526B635A735A526B635A735A526B635A73635273
635A736352736B5A7B7B73527B735A736B527B7352736B527B7352736B4A7373526B6B4A
736B526B634A736B526B634A4231314229295252394A4A31524A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007B6B737B6B737B6B737B6B737B6B73
7B6B737B6B737B6B73846B737B6B73846B736B7B946B7B946B7B94734A31734A31734A31
6B4A31734A316B4A316B4A316B42316B4A316B42316B4A316B42316B42316B42316B4231
6342316B42316342316342316342315A526B5A4A6B5A526B5A4A6B5A526B5A526B5A526B
5A526B5A526B5A526B5A52735A526B635273635273635273635273635273635273635A73
635273635A737B73527B735A7B73527B7352737352737352736B52736B52736B52736B52
6B6B4A6B6B4A6B634A6B6B4A4229294231315252395252394A4A31000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7B6B7B7B63737B6B737363737B6B7B7B6B737B6B737B6373846B7B7B6B73846B737B6B73
84737B7B6B73846B7B7B6B7384737B7B6B73737B9C6B7B9473849C6B7B9473849C6B4231
734A396B4A31734A396B4231734A396B4231734A316B4231734A396B42316B4A31634231
6B4A316342316B4A316342296B4A316342315A526B524A6B5A52735A4A6B5A526B5A4A6B
6352735A526B6352735A526B635A735A526B6352735A526B635A73635273635A73635273
6B5A7B6352737B7B5A7373527B735A7373527B7352736B4A7B7352736B4A736B526B6B4A
736B526B634A6B6B526B634A4A31315252315252394A4A31525239000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007363737B6B73
7B6B737B6B737B6B737B6B737B6B737B6B737B6B737B6B737B6B737B6B737B6B737B6B73
7B6B73846B7B7B6B73846B73846B73846B7B6B7B946B7B946B7B94737B946B7B946B7B94
6B7B94734A39734A31734A31734A31734A316B42316B4A316B4A316B4A316B42316B4A31
6B42316B4A316342316B42316342316B42316342315A526B5A526B5A526B5A526B5A526B
5A526B5A52735A526B5A52735A52736352735A5273635273635273635A73635273635A73
635273635A7B7B73527B73527B73527B7352736B52737352736B52737352736B4A736B52
6B6B4A736B526B634A6B6B4A5A39395A39395A31394A4A31000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000736373
7B6B7B7363737B6B7B7B6B737B6B7B7B6B737B6B7B7B6B73846B7B7B6B7384737B7B6B73
846B7B7B6B7384737B7B6B7384737B7B6B73737B9C6B7B94737B9C6B7B9473849C6B7B94
73849C6B7B9473849C734A31734A396B4231734A396B4231734A396B4231734A396B4231
6B4A316B42316B4A316342316B4A316342316B4A316342315A526B524A6B5A52735A4A6B
6352735A526B6352735A526B6352735A526B635A735A5273635A735A5273635A7B635273
635A7B6352737B7B5A7B73527B735A736B527B735A736B527373526B6B4A7373526B6B4A
736B526B634A736B525A31315A39395A31316339395A31316339395A3131000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7B6B737B6B737B6B737B6B737B6B737B6B737B6B737B6B7B7B6B737B6B737B6B73846B7B
7B6B73846B7B7B6B73846B7B7B6B73846B7B6B7B946B7B946B7B946B7B946B7B946B7B94
6B7B946B7B946B7B947384946B7B94734A31734A31734A316B4A31734A316B4A316B4A31
6B42316B4A316B42316B4A316B42316B4A316B42316B4A315A526B5A526B5A526B5A5273
5A526B5A52735A52736352735A5273635273635273635273635273635A73635273635A73
635273635A737B73527B735A7B73527B7352737352737352736B52736B52736B52736B52
6B6B4A6B6B4A5A39395A39395A39395A39395A39396339395A39396339395A3939633939
5A3939633939000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7B6B7B7B6B737B6B7B736373846B7B7B6B737B6B7B7B6B7384737B7B6B73846B7B7B6B73
84737B7B6B7384737B7B6B7384737B7B6B736B7B946B7394737B9C6B7B946B7B946B7B94
73849C6B7B947384946B7B9473849C6B7B94734A396B4231734A396B4A31734A316B4231
734A396B42316B4A316B42316B4A396B42316B4A316342296B4A315A526B5A52735A4A6B
6352735A526B6352735A526B635A735A5273635A735A5273635A7B635273635A735A5273
6B5A7B6352737B7B5A7373527B735A7373527B7352736B527B7352736B52736B526B6B4A
736B525A31396339395A31316339395A39396339395A31396339395A39396339395A3131
6339396339396339395A3939633939000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007B6B7B7B6B737B6B737B6B737B6B7B7B6B737B6B7B7B6B737B6B7B7B6B737B6B7B
7B6B7B84737B7B6B73846B7B846B7B84737B6B7B946B7B946B7B946B7B946B7B946B7B94
6B7B946B7B946B7B946B7B946B7B947384946B7B946B7B94734A31734A316B4A31734A31
734A31734A316B4A316B4A316B4A316B4A316B42316B4A316B42316B4A315A526B5A5273
5A52735A52735A526B5A52735A52736352735A5273635273635273635A73635273635A73
635A73635A7B7B73527B73527B73527B7352736B52737352736B527373526B6B4A736B52
6B6B4A6339395A39395A39395A39396339395A3939633939633939633939633939633939
633939633939633939633939633939633939633939633939000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B6B7B7B6B737B6B7B7B6B7384737B7B6B7384737B7B6B7384737B7B6B73
84737B7B6B7384737B7B6B7384737B7B6B736B7B946B73946B7B946B7B946B7B946B7B94
6B7B946B7B947384946B7B9473849C6B7B9473849C6B7B9473849C6B7B94734A396B4A31
734A396B4231734A396B4231734A316B4231734A316B42316B4A316342316B4A315A526B
5A52735A526B6352735A526B635A735A5273635A735A5273635A735A5273635A7B635273
635A7B6352737B7B5A7373527B735A736B527B735A736B527373526B6B4A7373526B6B4A
6339395A31396339395A39396339395A39396339395A39396339395A3939633939633939
6339396339396B42426339396B39396339396B42426339396B42426339396B4242000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B6B737B6B7B7B6B737B6B7B7B6B7B7B6B7B7B6B737B6B7B7B6B7B846B7B
7B6B7B84737B846B7B84737B7B6B7B84737B6B7B946B7B946B7B946B7B946B7B946B7B94
6B7B946B7B946B7B946B7B946B7B946B7B946B7B946B84946B7B946B84946B7B94738494
734A31734A31734A31734A316B4A316B4A316B4A316B4A316B4A316B4A316B4A316B4A31
5A526B5A52735A52736352735A5273635273635273635A73635273635A73635273635A7B
635A737B7B5A7B73527B735A7373527B7352737352737352736B52736B52736B52633939
5A3939633939633939633939633939633939633939633939633939633939633939633939
6339396339396339396339396339396339396339396B39396339396B39396B39396B3939
6B39396B39396B6384000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007B6B7384737B7B6B7384737B7B6B7384737B7B6B7384737B7B6B73
84737B7B6B7384737B7B6B7384737B7B6B7B6B7B9463738C6B7B946B7B946B7B9463738C
6B7B946B7B946B7B946B7B9473849C6B7B946B84946B7B9473849C6B7B947384946B7B94
73849C6B7B94734A396B4231734A396B4A31734A316B4231734A396B42316B4A316B4231
6B4A395A526B6352735A526B635A735A5273635A735A5273635A7B635273635A7B635273
6B5A7B7B73527B7B5A7373527B7B5A7373527B7352736B527373526B6B526339395A3139
6339425A39396339395A39396339426339396339396339396B4242633939633939633939
6B42426339396B39396339396B42426339396B42426339396B42426339396B4242633939
6B42426B39396B42426B39397342426B3939000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000007B6B7B7B6B7B7B6B737B6B7B7B6B7B84737B7B6B7B846B7B
7B6B7B84737B7B6B7B84737B84737B84737B6373946B7B946B7B946B7B946B7B946B7B94
6B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B84946B7B946B7B94
6B7B947384946B7B94734A31734A31734A316B4A31734A316B4A31734A316B4A316B4A31
6B4A316B4A315A52735A52735A5273635A735A5273635273635273635A7B635273635A7B
635A7B7B7B5A7B73527B73527B73527B7352737352737352736B526339395A3939633939
633939633939633939633939633939633939633939633939633939633939633939633939
6339396B39396339396B39396B39396B39426339396B39396B39396B42426B39396B3939
6B39396B42426B39396B39396B39396B42426B39396B42396B3939000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000084737B7B6B7384737B7B6B7384737B7B6B7384737B7B6B7B
84737B7B6B7384737B7B6B7B84737B7B6B7B6B7B9463738C6B7B9463738C6B7B94637B8C
6B7B94637B8C6B7B946B7B946B7B946B7B946B84946B7B947384946B7B947384946B7B94
7384946B7B9473849C6B7B9473849C734A31734A396B4A31734A396B4231734A396B4231
734A316B4231734A315A5273635A735A5273635A7B5A5273635A7B5A5273635A7B635273
635A7B7B73527B7B5A7373527B735A736B527B735A736B526339425A3939633942633939
6339426339396339426339396339426339396B42426339396B42426339396B4242633939
6B42426339396B42426339396B42426339396B42426B39396B42426B39396B42426B3939
6B42426B39397342426B39397342426B39397B948C738C847B9484000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B7B7B6B7B7B737B7B6B7B84737B7B6B7B84737B
7B6B7B84737B846B7B84737B84737B84737B637B946B7B94637B946B7B946B7B946B7B94
6B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B8494
6B7B946B84946B7B946B84946B7B947384946B8494734A31734A31734A31734A31734A31
6B4A316B4A316B4A316B4A315A5273635273635273635A73635273635A7B635A73635A7B
635A7B7B7B5A7B73527B735A737352737352737352633939633939633939633939633939
6339396339396339396339396339396339396339396B39396339396B39396339396B3942
6339396B39396B39396B39426B39396B39426B39396B42426B39396B42426B39396B4242
6B39397B8C847B8C847B8C847B8C847B8C847B8C847B8C84000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000084737B7B6B7384737B7B6B7B84737B7B6B73
8473847B6B7B84737B7B6B7B847384846B7B6B7B9463738C6B7B9463738C6B7B9463738C
6B7B94637B8C6B7B9463738C6B7B946B7B946B7B946B7B8C6B84946B7B946B84946B7B8C
7384946B7B947384946B7B9473849C6B7B947384946B7B9473849C734A31734A396B4A31
734A396B4A31734A316B4231635A7B5A5273635A7B5A5273635A7B635273635A7B635273
6B5A7B7B73527B7B5A7373527B735A7373527373525A3939634242633939633942633939
6B42426339396B39426339396B42426339396B42426339396B42426339396B4242633939
6B42426B39396B42426339396B42426B39396B42426B3939734242738C847B8C8C738C84
7B8C8C738C847B8C8C738C847B948C738C847B8C84738C84000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007B6B7B7B737B7B6B7B84737B7B6B7B84737B
84737B84737B846B7B84737B84737B84737B84737B637B946B7B946B7B9463738C6B7B94
6B7B946B7B94637B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B94
6B7B946B84946B7B946B7B946B7B946B84946B7B946B84946B84947384946B8494734A31
734A31734A316B4A31734A316B4A31635A7B5A5273635A73635A73635A7B635273635A7B
635A7B7B7B5A7B73527B735A737352737352633939633939633939633942633939633939
6339396B39426339396339396339396B42426339396B39426B39396B42426B39396B3939
6B39396B42426B39396B39426B39397B8C8C738C847B8C8C7B8C847B8C8C738C847B8C84
738C847B8C84738C847B8C847B8C847B8C84738C84000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007B6B7B84737B7B6B7B84737B7B6B7B
8473847B6B7B8473847B6B7B8473847B6B7B84738463738C6B7B9463738C6B7B9463738C
6B7B9463738C6B7B9463738C6B7B94637B8C6B7B946B7B8C6B7B94637B8C6B84946B7B8C
6B84946B7B947384946B7B947384946B7B947384946B7B947384946B7B947384946B7B94
734A396B4A31734A396B4A31734A396B4A31635A7B5A5273635A7B5A5273635A7B635273
7B7B5A7373527B7B5A7373527B735A6339396B42426339396B42426339396B4242633939
6B42426339396B42426339396B42426339396B42426339396B42426339396B42426B3939
6B4242738C847B8C8C738C847B8C8C7384847B8C8C738C847B8C8C738C847B8C8C738C84
7B8C8C7384847B8C8C738C847B8C84738C848C7B6B8473638C7B6B000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000007B6B7B84737B7B737B84737B
84737B84737B84737B84737B84737B84737B84737B637B9463738C6B7B94637B946B7B94
637B8C6B7B94637B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B94
6B7B946B7B946B7B946B84946B7B946B84946B7B946B84946B7B946B84946B8494738494
6B8494738494734A31734A31734A31734A31734A31635A7B635273635A7B635A7B635A7B
7B735A7B7B5A7B735A7B735A6339396339426339396339426339396B39426339396B3942
6339396B39426B39396B39426B39396B39426B39396B42426B3939738C8C738C8C738C8C
738C8C738C8C738C8C7B8C8C738C84738C8C738C84738C8C738C84738C84738C847B8C84
738C84738C84738C84738C84738C84738C84847B638C7B6B847363847B63847363000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008473847B6B7B8473847B6B7B
8473847B6B7B8473847B6B7B84738484737B84738463738C6B7B9463738C6B7B9463738C
6B7B9463738C6B7B9463738C6B7B94637B8C6B7B9463738C6B7B946B7B8C6B7B94637B8C
6B84946B7B946B84946B7B8C7384946B7B946B84946B7B8C7384946B7B947384946B7B8C
7384946B7B947384946B7B94735239734A31734A316B4A31635A7B635273635A7B635273
7B7B5A7B73527B7B5A6339396B42426339396B42426339396B42426339396B4242633939
6B42426339396B42426339396B424273848C7B8C8C7384847B8C8C738C847B8C8C738484
7B8C8C7384847B8C8C7384847B8C8C7384847B8C8C7384847B8C8C7384847B8C84738484
7B8C8C7384847B8C8473847B7B8C848473638C7B6B8473638C7B6B847363000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000008473847B737B84737B
84737B84738484737B84737B84737B84738484737B63738C63738C6B7B9463738C637B94
637B8C6B7B9463738C6B7B946B7B946B7B94637B8C6B7B946B7B946B7B946B7B8C6B7B94
6B7B946B7B946B7B946B7B946B7B946B84946B7B946B7B946B7B946B84946B7B946B8494
6B84947384946B84946B84946B8494738494734A31734A31734A31635A7B635A7B635A7B
7B7B5A7B7B5A6339396339426339426B42426339396B39426B39426B42426339396B3942
6B3942738C8C738C8C738C8C738C8C738C8C73848C738C8C738C8C738C8C738484738C8C
738C8C738C8C738484738C84738C84738C8C738484738C84738C84738C84738484738C84
738C84738C84738484738C848C7B6B8C7B6B847B63847B63847B63000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008473847B6B7B
8473847B6B7B84738484737B84738484737B84738463738C6B7B9463738C6B7B9463738C
6B7B9463738C6B7B9463738C6B7B9463738C6B7B94637B8C6B7B94637B8C6B7B94637B8C
6B7B94637B8C6B84946B7B8C6B84946B7B8C6B84946B7B8C6B84946B7B8C7384946B7B94
7384946B7B947384946B7B947384946B7B947384946B84947352396B4A31635A7B635A7B
7B7B5A7373526B42426339396B42426339396B42426339396B424273848C7B8C9473848C
738C8C73848C7B8C8C73848C738C8C738484738C8C738484738C8C7384847B8C8C738484
738C8C738484738C8C738484738C8C7384847B8C8C738484738C84738484738C84738484
738C8473847B738C8473847B8C7B6B8473638C7B6B847363000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000084737B847384
84737B84738484737B84738484737B84738484737B63738C63738C63738C63738C637B94
63738C637B94637B8C6B7B94637B8C6B7B94637B8C6B7B94637B8C6B7B946B7B946B7B94
6B7B8C6B7B946B7B946B7B946B7B946B7B946B7B946B84946B7B946B84946B7B946B8494
6B7B946B84946B84946B84946B84946B84946B84947384946B8494738494734A31635A7B
7B735A6B39426339426B42426B3942738C8C738C8C738C8C73848C738C8C73848C738C8C
73848C738C8C73848C738C8C73848C738C8C73848C738C8C738484738C8C738484738C8C
738484738C84738484738C84738484738C84738484738C84738484738C84738484738C84
738484738C847384848C7B6B847B6B8C7B6B847B6B8C7B6B000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B6B7B
84738484737B8473847B737B84738484737B84738463738C6B7B9463738C637B9463738C
6B7B9463738C6B7B9463738C6B7B9463738C6B7B9463738C6B7B94637B8C6B7B9463738C
6B7B94637B8C6B7B94637B8C6B84946B7B8C6B8494637B8C6B84946B7B8C6B84946B7B8C
7384946B7B8C6B84946B7B8C7384946B7B947384946B7B8C7384946B8494738494734A31
738C9473848C738C946B848C738C9473848C738C8C6B848C738C9473848C738C8C6B848C
738C8C73848C738C8C6B8484738C8C738484738C8C6B8484738C8C738484738C8C6B8484
738C8C738484738C8C6B8484738C8C738484738C846B8484738C84738484738C846B847B
738C8473847B738C84847B638C846B847B638C7B6B000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
84738484738484737B84738484738484738484737B63738C63738C637B9463738C63738C
63738C637B9463738C637B8C637B8C6B7B9463738C637B8C637B8C6B7B94637B8C6B7B94
6B7B946B7B94637B8C6B7B946B7B946B7B946B7B8C6B7B946B7B946B84946B7B946B7B94
6B7B946B84946B7B946B84946B84946B84946B84946B84946B849463424A63394A7B846B
735231738C9473848C738C8C73848C738C8C73848C73848C73848C738C8C73848C738C8C
73848C738C8C73848C73848C73848C738C8C73848473848C738484738C8C738484738484
738484738C84738484738484738484738C84738484738484738484738C84738484738484
738484738C848C7B6B8C7B6B8C7B6B8C7B6B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007B737B84738484737B84738484737B847B8463738C637B9463738C637B9463738C
637B9463738C6B7B9463738C6B7B9463738C6B7B9463738C6B7B9463738C6B7B9463738C
6B7B94637B8C6B7B94637B8C6B7B94637B8C6B8494637B8C6B8494637B8C6B84946B7B8C
6B84946B7B8C6B84946B7B8C73849463394A6B424A63394A6B424A63394A7B846B636384
6B6B8C735231738C946B848C738C8C6B848C738C8C6B848C738C8C6B848C738C8C6B8484
738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484
738C8C6B8484738C8C6B8484738C846B8484738C846B8484738C846B847B738C846B847B
738C84847B6B8C846B847B6B8C7B6B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000084738484738484738484738484738484738463738C63738C63738C63738C63738C
63738C63738C63738C637B8C63738C637B8C637B8C637B9463738C637B8C637B8C6B7B94
637B8C6B7B94637B8C6B7B94637B8C6B7B946B7B8C6B7B946B7B8C6B7B946B7B946B8494
6B7B8C6B7B9463394A63424A63394A63424A63394A63424A63394A7B846B7B7B63636384
63638463638C7352317352316B848C73848C73848C73848C73848C73848C73848C73848C
6B848C73848C73848473848C6B848473848C73848473848C6B8484738484738484738484
6B84847384847384847384846B84847384846B84847384846B8484738484738484738484
6B847B8C7B6B8C7B6B8C7B6B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000084738484737B8C7B84847384847B845A738C637B9463738C63738C5A738C
637B9463738C637B8C63738C6B7B9463738C637B9463738C6B7B9463738C6B7B9463738C
6B7B94637B8C6B7B9463738C6B7B94637B8C6B7B94637B8C6B8494637B8C6B84945A3942
63424A63394A63424A63394A6B425263394A63424A63394A7B846B7B7B637B846B636384
6B6B8C6363847B5A397352317B5A396B848C738C8C6B8484738C8C6B8484738C8C6B8484
738C8C6B8484738C8C6B8484738C8C6B848473848C6B8484738C8C6B8484738C846B8484
738C8C6B84847384846B847B738C846B8484738C846B847B738C846B847B7384846B847B
8C84738C7B6B8C846B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000039293184738484738484738484738463738C63738C63738C63738C63738C
63738C637B8C63738C63738C63738C637B8C63738C637B8C637B8C637B8C63738C637B8C
637B8C6B7B94637B8C637B8C637B8C6B7B94637B8C63394A63394A63424A63394A63394A
63394A63424A63394A63424A63424A63424A63394A7B846B7B846B7B846B737B63636384
63638463638C6363847352317352317352317352316B848C6B848C73848C6B84846B848C
6B848C73848C6B84846B84846B848473848C6B84846B84846B84847384846B84846B8484
6B84847384846B84846B84846B84847384846B84846B84846B84847384846B847B8C7B6B
8C7B6B8C846B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000392929423131847384847B848473848C7B845A738C63738C5A738C637B8C5A738C
637B8C5A738C637B9463738C637B8C63738C637B9463738C637B9463738C6B7B9463738C
6B7B9463738C6B7B9463738C63424A5A394263424A5A394A63424A5A394A63424A63394A
63424A63394A63425263394A6B425263394A6B42527B7B6B7B846B737B637B846B636384
63638C6363846B6B8C6363847B5A39735231735A3973523173848C6B848473848C6B8484
73848C6B8484738C8C6B848473848C6B848473848C6B84847384846B84847384846B8484
7384846B847B7384846B847B7384846B847B7384846B847B7384846B847B7384848C7B6B
8C84738C7B6B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39293139293139293139293184738484738484738463738C5A738C63738C63738C63738C
63738C63738C63738C63738C63738C63738C63738C637B8C63738C637B8C637B8C637B8C
5A394A5A394A5A394A63394A5A394A63394A63394A63424A63394A63424A63394A63424A
63394A63424A63394A63424A63394A63424A7B846B7B846B7B7B6B7B846B636384636384
63638463638C636384636B8C63638C735A317352317352317352316B848C6B84846B848C
6B84846B84846B84846B84846B84846B84846B84846B84846B84846B84846B84846B8484
6B84846B84846B84846B84846B84846B84846B84846B84846B847B6B84848C7B6B8C8473
8C7B6B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000392929
4231313929314231313929314231398473848C7B845A6B8463738C5A738C63738C5A7384
637B8C5A738C63738C5A738C637B9463738C637B8C63738C637B945A39425A394A5A3942
63424A5A394A63424A5A394263424A5A394A63424A5A394A63425263394A63424A63394A
6B425263394A63425263394A6B42527B846B7B846B737B637B846B737B63636384636384
6B6B8C6363846B6B8C6363846B6B8C7352317B5A39735231735A397352317352316B7B84
73848C6B84846B848C6B7B8473848C6B84846B84846B7B847384846B84846B84846B7B7B
7384846B847B6B84846B7B7B7384846B847B6B84846B7B7B7384848C7B6B8C84738C7B6B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000392931392931
392931423131392931422931392931847B848473845A738C5A738C63738C5A738C63738C
63738C63738C5A738C63738C63738C5A394A5A394A5A394A5A394A5A394A5A394A5A394A
5A394A63424A5A394A63394A63394A63424A5A394A63394A63394A63424A63394A63424A
63424A63425263394A63424A7B846B7B846B7B846B7B846B7B846B7B846B636384636384
63638463638C63638463638C63638C6B6B8C735231735231735231735231735231735231
6B84846B84846B84846B84846B84846B84846B84846B84846B84846B84846B84846B8484
6B84846B84846B847B6B84846B847B6B84846B847B6B847B6B847B8C84738C7B6B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000393131392929423131392931
4231313929314231393929314231393929318C7B8C5A6B8463738C5A6B8463738C5A738C
63738C5A73845A394A5239425A394A5A39425A424A5A394A5A424A5A394263424A5A394A
63424A5A394A63424A5A394A63424A5A394A6342525A394A63425263394A63425263394A
6B425263394A6B42527B846B7B8473737B6B7B846B737B637B846B737B6363638C636384
63638C6363846B6B8C6363846B6B8C6363846B6B8C7352317B5A39735231735A396B5231
7352316B4A316B848C6B7B846B84846B7B846B84846B7B846B84846B7B7B6B84846B7B7B
6B84846B7B7B6B84846B7B7B6B84846B7B7B6B84846B7B7B9484738C7B6B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000392931392931392931392931392931
3929314231313929314231314229314231314229315A738C5A738C63738C5A738C5A394A
52394A5A394A5A394A5A394A5A394A5A394A5A394A5A394A5A394A5A394A5A394A63394A
5A394A63394A63394A63424A63394A63424A63394A63424A63394A63424A63424A634252
63394A6342527B846B7B846B7B846B7B846B7B846B7B846B737B63737B6B636384636384
63638463638C636384636B8C636384636B8C63638C7B5A31735231735A31735231735231
7352317352316B52316B84846B84846B84846B84846B84846B84846B84846B84846B8484
6B847B6B84846B847B6B84846B847B6B847B6B847B6B847B8C8473000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000393131392929423131392931423131392931
4231393929314231393929314231394229314231395231425A394A5239425A394A523942
5A424A52394A5A394A5239425A424A5A394A5A394A5A394263424A5A394A63424A5A394A
6342525A394A63424A5A394A63425263394A63425263394A6B425263394A63425263394A
6B42527B846B7B8473737B6B7B8473737B6B7B846B737B637B846B5A638463638C5A6384
636B8C636384636B8C6363846B6B8C6363846B6B8C6363847B5A39735231735A39735231
735A396B52317352316B4A31735239637B846B8484637B7B6B8484637B7B6B8484637B7B
6B8484637B7B6B8484637B7B6B8484637B7B6B84848C7B6B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000392931392931392931392931423131392931422931
4229314231314229314231314A637B4A637B4A637B8C849452394A5A394A52394A5A394A
5A394A5A394A5A394A5A394A5A394A5A394A5A394A5A394A5A394A63424A5A394A63394A
63394A63424A5A394A63424A63394A63425263394A63424A63424A63425263394A634252
6342527B84737B846B7B846B7B846B7B846B737B6B73846B737B6B636384636384636384
63638463638C63638463638C63638C636B8C63638C636B8C636B8C7B5A39735231735231
735231735A316B52317352316B52316B5231637B846B84846B84846B8484637B7B6B8484
6B847B6B8484637B7B6B847B6B847B6B847B8C84738C8473000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000393131392931423131392931423131392931423139392931
4231393929314A637B425A734A637B425A734A637B8484945A394A5239425A394A52394A
5A394A52394A5A424A5A394A5A424A5A394A63424A5A394A6342525A394A6342525A394A
6342525A394A6342525A394A63425263394A63425263394A6B425263394A6B425263394A
7B8C7373846B7B847373846B7B846B737B637B846B737B637B846B5A5A8463638C5A6384
63638C636384636B8C6363846B6B8C6363846B6B8C63638C6B6B8C63638C7B5A39735231
7B5A39735231735A396B52317352316B5231735231637B7B6B8484637B7B6B8484637B7B
6B8484637B7B6B8484637B7B6B8484637B7B948473000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003929313929313929313929314231313929314231314229314A637B
425A734A637B4A5A734A637B4A5A734A637B4A637B8C84948C84945A394A5A394A5A394A
5A394A5A394A5A394A5A394A5A394A5A394A5A394A63424A5A394A63424A5A394A63424A
63394A63424A63394A63425263394A63425263424A63425263424A6342526342527B8473
7B846B7B846B7B846B7B846B73846B7B846B737B6B73846B737B63636384636384636384
63638463638C63638463638C63638C636B8C63638C636B8C636B8C6B6B8C7B5A317B5A31
735231735A317352317352316B52317352316B52316B52316B52316B8484637B7B6B8484
637B7B6B847B637B7B6B847B637B7B948473000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000003931313929314231393929314231313929314A637B425A734A5A73425A73
4A637B425A734A637B425A734A637B425A734A637B847B948C8C9C52394A5A394A52394A
5A424A5A394A5A424A5A394A6342525A394A6342525A394A6342525A394A6342525A394A
63425263394A6342525A394A63425263394A63425263394A6B425263394A7B8C7373846B
7B8C7373846B7B846B737B6B7B846B737B6B7B846B737B637B846B5A638463638C5A6384
636B8C636384636B8C6363846B6B8C636384636B8C6363846B6B8C636B8C6B6B8C735231
7B5A39735231735A396B5231735A396B52317352316B52317352316B4A316B8484637B7B
6B8484637B7B6B847B637B7B948C7B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
392931392931392931392931392931423131425A73425A73425A734A5A73425A73425A73
425A734A5A73425A734A5A734A5A734A637B8484948C84948C84948C849C5A394A5A394A
5A394A5A394A5A394A5A394A5A394A6342525A394A63394A5A394A6342525A394A63424A
63424A63425263394A63425263425263425263394A6342526342527B8C737B846B7B8473
7B846B7B847373846B73846B73846B7B846B737B63737B6B737B636363845A6384636384
63638463638C63638463638C63638C636B8C63638C636B8C636B8C6B6B8C636B8C6B6B8C
7B5A317B5A39735231735A31735231735A316B52316B52316B52316B52316B52316B5231
637B7B6B847B637B7B637B7B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000392931
39313139293142313142526B4A5A7342526B4A5A73425A6B4A5A73425A6B4A5A73425A73
4A5A73425A734A637B425A734A637B425A738C849C8484948C849C8484945A42525A394A
5A42525A394A6342525A394A6342525A394A6342525A394A6342525A394A6342525A394A
63425263394A63425263394A63425263394A6B425263394A7B8C737B846B7B8C7373846B
7B8473737B6B7B8473737B6B7B846B737B637B846B737B636363845A5A8463638C5A6384
63638C5A6384636B8C636384636B8C6363846B6B8C63638C6B6B8C63638C6B6B94636B8C
7B5A397352317B5A39735231735A396B5231735A396B52317352316B52316B5231634A31
6B5231637B7B6B847B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000392931392931
42526B425A6B42526B425A7342526B425A73425A6B425A73425A6B425A73425A73425A73
425A73425A73425A734A5A73425A734A5A738484948C849C8C84948C849C8C84945A394A
5A394A5A394A5A394A5A424A5A394A6342525A394A6342525A394A63425263394A634252
63394A6342526342526342526342526342526342527B8C737B84737B84737B84737B8473
73846B7B846B73846B73846B737B6B73846B737B6B737B6B5A6384636384636384636384
63638463638C636384636B8C63638C636B8C63638C636B8C636B8C636B8C636B8C6B6B8C
636B8C7B5A31735A317B5A31735231735A31735231735A316B52316B52316B52316B5231
6B52316B52316352314A3931000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039526B425A6B39526B
425A7342526B425A7339526B425A7342526B425A7342526B4A5A7342526B425A7342526B
4A5A73425A6B4A5A73425A6B4A6373425A738C849C8484948C8C9C8484948C849C5A394A
6342525A394A6342525A394A6342525A394A6342525A394A63425263394A6342525A394A
63425263394A63425263394A6B42526342527B8C737384737B8C7373846B7B847373846B
7B847373846B7B846B737B6B7B846B737B6373846B6B7B6363638C5A638463638C5A6384
636B8C636384636B8C5A6384636B8C63638C636B8C6363846B6B8C63638C6B6B8C63638C
6B6B94636B8C7B5A397352317B5A39735231735A396B5231735A396B52317352316B5231
6B52316B52316B52314A6B6B523931000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039526B42526B
42526B425A6B42526B42526B42526B425A6B42526B425A6B425A6B425A73425A6B425A73
425A73425A73425A6B425A73425A734A5A738484948C849C8C849C8C849C8484948C849C
5A394A6342525A394A6342525A394A6342525A394A63425263425263425263394A634252
6342526342526342526342526342526B42527B84737B84737B84737B847373846B7B846B
73846B7B846B737B6B73846B737B6B73846B737B63737B635A63846363845A6384636384
63638463638C63638463638C63638C636B8C63638C636B8C636B8C6B6B8C636B8C636B8C
636B8C6B6B94636B8C7B5A31735A317B5A39735231735A31735231735A316B52316B5231
6B52316B52316B5231526B6B5239314A3931000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000395263425A6B39526B425A6B39526B425A7342526B425A7342526B425A7342526B
425A7342526B4A5A7342526B4A5A73425A6B8C849C8484948C849C8484948C8C9C848494
6342525A394A6342525A394A6342525A394A6342525A394A63425263394A63425263394A
63425263394A6B42526339526B42527B84737B8C737384737B8C7373846B7B847373846B
7B8473737B6B7B846B737B6B73846B6B7B6373846B6B7B636363845A5A8463638C5A6384
63638C5A6384636B8C636384636B8C636384636B8C63638C6B6B8C63638C6B6B94636B8C
6B6B94636B8C6B7394735A317B5A397352317B5A39735231735A396B5231735A396B5231
7352316B52316B52314A6B6B526B6B4A3929000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000042526B42526B42526B425A6B42526B425A6B42526B425A6B
42526B425A6B425A6B425A73425A6B8C849C84849C8C849C84849C8C849C8C849C8C849C
8C849C6342525A39526342525A394A634252634252634252634252634252634252634252
6342526342526342526B42527B84737B8C737B84737B84737384737B847373846B7B8473
73846B73846B737B6B73846B737B6B737B6B737B635A63845A63846363845A638463638C
5A638463638C63638463638C63638C636B8C63638C636B8C636B8C636B8C636B8C6B6B8C
636B8C6B6B8C636B8C6B6B947B5A317B5A31735A31735A31735A31735A31735231735A31
6B52316B52316B5231526B6B4A6B6B4A6B6B4A3931000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000395263425A6B42526B425A6B39526B
425A7342526B425A7342526B425A738484948C849C8484948C8C9C84849C8C8C9C848494
8C8C9C5A394A6342525A394A63425263394A6342525A394A63425263395263425263394A
6B42526342526B42527B84737B8C7B7B84737B8C7373846B7B8C7373846B7B8473737B6B
7B8473737B6B73846B6B7B6373846B6B7B63737B6B5A5A8463638C5A638463638C5A6384
636B8C5A6384636B8C5A6384636B8C636384636B8C6363846B6B9463638C6B6B8C63638C
6B6B94636B8C6B6B94636B8C6B7394735A317B5A397352317B5A39735231735A316B5231
735A396B52317352314A6363526B6B4A63634A6B6B4A3129000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000042526B42526B
42526B425A6B42526B425A6B425A6B8C849C84849C84849C84849C8C849C84849C8C849C
8C849C6342525A394A634252634252634252633952634252634252634252634252634252
6342526B42527B84737B8C737B84737B8C737384737B84737384737B847373846B73846B
73846B73846B737B6B737B6B737B6B737B6B6B7B635A63845A63846363845A6384636384
63638463638C63638463638C63638C636B8C63638C636B8C636B8C636B8C636B8C636B8C
636B8C6B6B94636B8C6B6B946B6B947B5A397B5A317B5A31735A31735A31735231735A31
735231735A316B52314A6B634A63634A6B634A63634A63634A3931000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000039526B425A6B42526B425A6B8484948C849C8484948C8C9C8484948C8C9C84849C
8C8C9C8484946342525A394A6342526339526342526339526342526339526B4252633952
6B42527B84737B8C7B7B84737B8C7B7384737B8C7373846B7B847373846B7B8473737B6B
73846B737B6B73846B6B7B6373846B6B7B63737B6B5A5A846363845A5A8463638C5A6384
63638C5A6384636B8C5A6384636B8C5A6384636B8C63638C636B8C63638C6B6B94636B8C
6B6B94636B8C6B7394636B8C6B7394636B8C7B6339735A317B5A397352317B5A39735231
735A396B5231735A394A63634A6B6B4A63634A6B634A63634A6B63423129000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000042526B84849C84849C8C849C84849C8C849C84849C8C8C9C
84849C8C8C9C634252634252634252634252634252634252634252634252634252634252
7B8C7B7B8C7B7B8C737B8C737B84737B84737384737B847373847373847373846B73846B
737B6B73846B737B6B737B6B6B7B63737B6B6B7B635A63845A63845A63845A6384636384
5A638463638C63638C63638C63638C636B8C63638C636B8C636B8C636B8C636B8C636B8C
636B8C636B8C636B8C6B6B94636B8C6B6B946B6B947B5A39735A317B5A31735A31735A31
735A31735A317352314A63634A63634A63634A63634A63634A63634A6363000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008C849C84849C8C849C8484948C8C9C84849C8C8C9C848494
8C8C9C84849C8C8C9C5A39526342526339526342526339526B425A6342526B4252633952
7B8C7B7B84737B8C7B7384737B8C7B7384737B847373846B7B847373846B738473737B6B
738473737B6B73846B6B7B6373846B6B7B635A63845A5A8463638C5A638463638C5A6384
636B8C5A6384636B8C5A6384636B8C63638C636B8C5A6384636B9463638C636B8C63638C
6B6B94636B8C6B6B94636B8C6B7394636B8C6B7394636B8C7B6339735A317B5A39735231
7B5A39735231735A3142635A4A6B634263634A6363425A5A4A636342635A4A6363000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000084849C84849C84849C84849C84849C8C849C84849C8C849C
8C849C8C8C9C84849C6342526342526342526342526342526342526B42526342527B8C7B
7B8C7B7B8C7B7B84737B84737B84737B8473738473738473738473738473737B6B73846B
73846B73846B6B7B6B737B6B6B7B6B737B6B5A5A845A63845A63845A63845A63845A6384
5A638463638C5A638463638C63638C636B8C63638C636B8C636B8C636B8C636B8C636B8C
636B8C6B6B94636B8C636B94636B946B7394636B946B6B947B5A317B5A39735A31735A31
735A31735A317352314A63634A63634A636342635A4A63634A635A4A6363425A5A4A635A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008C849C84849C8C8C9C84849C8C8C9C84849C8C8C9C84849C
8C8C9C84849C8C8CA584849C6342526339526B425A6342526B425A6342527B8C7B7B8473
7B8C7B7384737B8C7B7384737B8C737384737B8C7373846B738473737B6B738473737B6B
73846B6B7B6B73846B6B7B63737B6B6B7B635A63845A5A8463638C5A5A8463638C5A6384
63638C5A6384636B8C5A6384636B8C5A6384636B8C63638C636B8C63638C636B94636B8C
6B6B94636B8C6B7394636B8C6B7394636B8C6B7394636B946B73947B5A317B6339735A31
7B5A39735A31735A39425A5A4A6363425A5A4A6363425A5A4A6363425A5A4A6363000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000084849C84849C84849C84849C84849C84849C84849C8C8C9C
84849C8C8C9C8C8C9C8C8C9C6342526342526342526342526342527B8C7B7B8C7B7B8C7B
7B84737B8C737384737B847373847373847373847373847373846B73846B73846B73846B
737B6B737B6B6B7B6B737B6B6B7B636B7B635A63845A63845A63845A63845A638463638C
5A638463638C5A638463638C63638C636B8C63638C636B8C63638C636B8C636B8C636B8C
636B8C636B94636B8C6B6B94636B8C6B6B94636B946B73946B6B946B73947B5A317B5A39
735A31735A31735A314A6363425A5A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008C84A584849C84849C84849C8C8CA584849C8C8C9C84849C
8C8CA584849C8C8C9C84849C8C8CA56342526B425A634252848C7B7B8C7B7B8C7B738473
7B8C7B7384737B8C7B7384737B8C73738473738473737B6B738473737B6B7384736B7B6B
73846B6B7B6B737B6B6B7B6373846B6B7B635A63845A5A8463638C5A638463638C5A6384
636B8C5A6384636B8C5A6384636B8C5A638C636B8C5A6384636B9463638C636B9463638C
6B6B94636B8C6B6B94636B8C6B7394636B8C6B7394636B8C6B7394636B947B6339735A31
7B6339735A317B5A39000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000084849C84849C84849C84849C84849C8C8C9C84849C848C9C
84849C8C8C9C84849C8C8C9C8C8C9C6B425A6342527B8C7B7B8C7B7B8C7B7B8C737B8C7B
7B8C737B8C7B73847373847373847373847373846B73847373846B738473737B6B73846B
6B7B6B737B6B6B7B636B7B6B6B7B635A63845A5A845A63845A63845A63845A63845A6384
5A638463638C5A638463638C63638C636B8C5A638C636B8C636B8C636B8C636B8C636B8C
636B8C636B94636B8C636B94636B946B7394636B946B73946B6B946B73946B6B947B5A31
7B5A317B6339735A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000084849C8484A584849C8C8CA584849C8C8CA584849C8C8CA584849C
8C8CA584849C8C8CA584849C8C8CA584849C8494847B8C7B7B8C7B7384737B8C7B738473
7B8C7B7384737B8C7B7384737B847373846B738473737B6B7384736B7B6B7384736B7B6B
73846B6B7B6373846B6B7B636B7B6B525A845A63845A5A845A63845A5A8463638C5A6384
63638C5A6384636B8C5A6384636B8C5A6384636B8C5A638C636B8C63638C636B94636B8C
636B94636B8C6B6B94636B8C6B7394636B8C6B7394636B946B7394636B946B7394636B94
7B6339735A317B6339000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000084849C84849C84849C84849C84849C84849C84849C84849C8C8CA5
84849C8C8C9C848C9C8C8C9C848C9C8C8C9C8C8C9C8C8CA57B8C7B7B8C7B7B8C7B7B8C7B
7384737B847373847373847373847373847373847373847373846B73846B737B6B73846B
6B7B6B6B7B6B6B7B6B6B7B6B6B7B635A63845A63845A63845A63845A63845A63845A638C
5A63845A638C5A638C63638C5A638C636B8C63638C636B8C636B8C636B8C636B8C636B8C
636B8C636B8C636B8C636B94636B946B6B94636B946B7394636B946B73946B73946B7394
6B73947B63397B5A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007B849C848CA584849C8484A584849C8C8CA584849C8C8CA584849C
8C8CA584849C8C8CA584849C8C8CA5848C9C8C8CA5848C9C8C94A5848C9C8C8CA5738473
7B8C7B7384737B8C737384737B8C7373846B7384736B7B6B7384736B7B6B73846B6B7B6B
73846B6B7B636B7B6B6B73636B7B6B5A5A845A6384525A845A638C5A63845A638C5A6384
636B8C5A638463638C5A6384636B8C5A638C636B8C5A638C636B9463638C636B8C5A638C
6B6B94636B8C636B94636B8C6B7394636B8C6B7394636B8C6B7394636B946B7394636B94
6B739C7B5A317B6339000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000084849C84849C8484A584849C84849C84849C848CA584849C848C9C
848C9C8C8CA584849C8C8C9C848C9C8C8CA5848C9C8C8C9C8C8C9C8C8CA58C8C9C8C8CA5
8C8C9C8C94A573847373847373847373847373846B73846B73846B73846B6B7B6B6B7B6B
6B7B6B6B7B6B6B7B636B7B636B7B635A63845A5A845A63845A63845A63845A63845A6384
5A638463638C5A63845A638C5A638C636B8C5A638C636B8C636B8C636B8C636B8C636B8C
636B8C636B94636B8C636B94636B946B7394636B94636B946B73946B7394636B946B7394
6B73946B739C7B5A317B6331000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000084849C8484A584849C848CA584849C848CA584849C8C8CA584849C
8C8CA584849C8C8CA584849C8C8CA584849C8C8CA5848C9C8C8CA5848C9C8C94A5848C9C
8C94A5848C9C8C94A58C8C9C8C94A58C8C9C7384736B7B6B7384736B7B6B73846B6B7B6B
73846B6B7B636B7B6B6B7B636B7B6B525A845A63845A5A845A638C5A5A845A638C5A6384
63638C5A6384636B8C5A6384636B8C5A6384636B8C5A638C636B8C5A638C636B94636B8C
636B94636B8C6B7394636B8C6B7394636B8C6B7394636B946B7394636B946B739C636B94
6B739C636B946B739C7B5A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000084849C8484A584849C8484A584849C8484A584849C848CA584849C848CA5
84849C848CA5848C9C8C8CA5848C9C8C8CA5848C9C8C8CA5848C9C8C8CA58C8C9C8C8CA5
8C8C9C8C94A58C8C9C8C94A58C8C9C8C94A58C94A58C94A58C949C6B7B6B6B7B6B6B7B6B
6B7B6B6B7B6B6B7B636B7B6B525A845A63845A63845A63845A63845A63845A63845A638C
5A63845A638C5A638C63638C5A638C636B8C5A638C636B8C63638C636B8C636B8C636B8C
636B8C636B94636B8C636B94636B8C636B94636B946B7394636B946B73946B73946B7394
6B73946B73946B73947B6339000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008484A57B849C848CA584849C848CA57B849C8C8CA584849C848CA584849C
8C8CA584849C8C8CA584849C8C8CA5848C9C8C8CA5848C9C8C94A5848C9C8C94A5848C9C
8C94A58C8C9C8C94A5848C9C8C94A58C8C9C8C94A58C8C9C9494A58C949C9494A58C949C
949CA56B7B636B7B6B6373635A6384525A845A6384525A845A638C5A63845A638C5A5A84
636B8C5A6384636B8C5A6384636B8C5A638C636B8C5A638C636B945A638C636B945A638C
949C8C8C94848C9C8C8C9484949C8C8C94848C948C8C948494948C8C94848C94848C8C84
8C948C8C9484000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000084849C8484A58484A58484A584849C8484A58484A5848CA584849C848CA5
848CA5848CA5848C9C848CA5848CA58C8CA5848C9C8C8CA58C8CA58C8CA5848C9C8C8CA5
8C8CA58C94A58C8C9C8C94A58C94A58C94A58C949C8C94A58C94A58C94A58C94A58C94A5
8C94A59494A58C949C8C94A5949C94949C948C9C94949C94949C94949C948C9C8C949C94
8C9C94949C948C9C8C8C9C8C8C9C8C949C8C8C948C8C9C8C8C948C949C8C8C948C8C948C
8C948C8C948C8C94848C948C8C948C8C948C8C94848C94848C94848C94848C94848C9484
8C94848C9484000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008484A57B849C8484A57B849C848CA584849C848CA584849C8C8CA584849C
848CA584849C8C8CA584849C8C8CA5848C9C8C8CA5848C9C8C8CA5848C9C8C94A5848C9C
8C94A5848C9C8C94A58C8C9C8C94A58C8C9C8C94A58C8C9C8C94A58C949C9494A58C949C
425273424A6B425273424A735A6B5A526352949C948C9C94949C948C948C949C948C948C
949C948C948C949C948C948C949C948C948C949C948C948C949C8C8C948C949C8C8C948C
949C8C8C9484949C8C8C94848C9C8C8C94848C9C8C8C94848C948C8C94848C948C8C9484
8C9484848C84000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008484A58484A58484A58484A58484A5848CA58484A5848CA58484A5848CA5
8484A5848CA5848CA5848CA5848CA5848CA5848CA58C8CA5848CA58C8CA5848CA58C8CA5
8C8CA58C94A58C8CA58C94A58C8CA58C94A58C94A58C94A58C94A5424A6B424A6B424A73
424A73424A73424A73526B525263525263525263525263528C9C94949C948C9C94949C94
8C9C8C8C9C948C9C8C949C948C948C8C9C8C8C948C8C9C8C8C948C8C9C8C8C948C8C9C8C
8C948C8C948C8C948C8C948C8C94848C948C8C94848C948C8C94848C94848C94848C9484
8C94848C9484000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008484A57B849C848CA584849C848CA57B849C848CA584849C848CA584849C
8C8CA584849C848CA584849C8C8CA5848C9C8C8CA5848C9C8C94A5848CA58C94A5848C9C
8C94A5848CA58C94A5848C9C8C94A58C8CA5424A6B39426B424A73394A6B424A73394A6B
425273424A73425273526352526B5252634A5263524A5A4A5263524A5A4A949C948C948C
949C948C948C949C948C948C949C948C948C949C8C8C948C949C948C948C8C9C8C8C9484
949C8C8C948C8C9C8C8C94848C9C8C8C94848C948C8494848C9C8C8C94848C948C848C84
8C948C849484000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008484A57B84A58484A58484A5848CA58484A58484A58484A5848CA58484A5848CA5
848CA5848CA5848CA5848CA5848CA58C8CA5848CA5848CA5848CA58C8CA5848CA58C8CA5
8C8CA58C94A58C8CA5394263394263424A6B39426B424A6B424A6B424A6B424A6B424A6B
424A73424A73424A73526B525263525263524A634A4A634A4A5A4A4A5A4A000000000000
8C9C948C9C948C9C8C8C9C8C8C9C8C8C9C8C8C948C8C9C8C8C948C8C9C8C8C948C8C948C
8C948C8C9C8C8C948C8C948C8C948C8C948C8C94848C94848C94848C948C8C94848C9484
8C94848C9484000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007B849C848CA57B849C848CA57B849C848CA57B84A5848CA57B849C848CA58484A5
848CA584849C8C8CA5848CA58C8CA5848C9C8C94A5848CA58C94A5848C9C8C94A5000000
000000394263424A6B394263424A6B394263424A6B39426B424A6B39426B424A73394A6B
424A73394A6B4252735263525263524A5A4A5263524A5A4A4A634A000000000000000000
000000000000949C948C948C949C948C948C949C948C948C8C9C8C8C948C8C9C8C8C948C
8C9C8C8C94848C9C8C8C94848C9C8C8494848C9C8C8494848C948C8494848C948C849484
8C9484848C84000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008484A57B84A58484A57B84A58484A58484A5848CA58484A5848CA5848CA5848CA5
848CA5848CA5848CA5848CA5848CA5848CA5848CA5848CA5000000000000000000000000
000000394263394263394263394263394A6B39426B424A6B394A6B424A6B424A6B424A73
424A6B424A735263525263524A634A5263524A5A4A4A5A4A4A5A4A000000000000000000
0000000000000000000000008C9C8C8C9C8C8C9C8C8C9C8C8C948C8C9C8C8C948C8C9C8C
8C948C8C948C8C948C8C948C8C94848C948C8C94848C948C8C94848C94848C94848C9484
8C9484000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007B84A58484A57B849C848CA57B84A5848CA57B849C848CA58484A5848CA57B84A5
848CA5848CA5848CA584849C8C94A5000000000000000000000000000000000000000000
000000000000394A63394263424A6B394263424A6B394263424A6B394A6B424A7339426B
424A73424A6B526B524A634A5263524A5A4A4A634A4A5A42000000000000000000000000
0000000000000000000000000000000000008C9C948C948C8C9C948C948C8C9C8C849484
8C9C8C8C948C8C9C8C8494848C9C8C8C94848C948C8494848C9C8C8494848C948C848C84
8C948C000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008484A57B84A58484A58484A5848CA57B84A58484A5848CA5848CA58484A5848CA5
848CA5848CA5000000000000000000000000000000000000000000000000000000000000
000000000000394263394263394263424A6B394263394A6B424A6B424A6B394A6B424A6B
424A6B424A735263525263524A634A4A634A4A5A4A000000000000000000000000000000
0000000000000000000000000000000000000000000000008C9C8C8C9C8C8C948C8C9C8C
8C948C8C9C8C8C948C8C948C8C948C8C948C8C94848C948C8C94848C948C8494848C9484
8C9484000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8484A57B84A5848CA57B84A5848CA57B84A5848CA57B84A5848CA57B84A5848CA5000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000394263424A6B394263424A6B394263424A6B39426B424A73394A6B
424A734A634A5263524A5A4A5263524A5A4A4A634A000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C9C8C8C948C
8C9C8C8494848C9C8C8494848C9C8C8494848C9C8C8494848C948C8494848C948C849484
8C948C000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7B84A57B84A57B84A58484A57B84A58484A57B84A5000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000394263394263394263394263424A6B394A6B424A6B424A6B424A6B
424A6B5263524A634A4A63524A5A4A4A5A4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C948C8C948C8C948C8C948C8C948C8C948C8C94848C948C8494848C94848494848C9484
849484000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
848CAD7B84A58484A57B84A5000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000424A6B394263424A6B394263424A6B39426B424A6B39426B
424A734A634A5263524A5A4A526352000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000084948C8C9C8C8494848C9C8C8494848C948C8494848C9C8C8494848C948C849484
8C948C000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7B84A5000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000394263394A6B394263394A6B394A6B424A6B394A6B424A6B
424A6B5263524A5A4A4A5A4A4A5A4A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008C948C8C948C8C948C8494848C948C8494848C948C849484849484
849484000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000394263424A6B394263424A6B394263424A6B39426B
5263524A5A4A5263524A5A4A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008494848C9C8C8494848C948C8494848C948C849484
8C948C000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000394263394263394A6B394263394A6B394A6B424A6B
4A634A4A634A4A5A4A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008C948C84948484948C849484849484
849484000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000394A6B394263424A6B394263424A6B39426B
5263524A5A4A4A5A4A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000008494848C948C849484
8C948C000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000394263394263394263424A6B39426B4A6352
4A5A4A4A5A4A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000849484
849484000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000394263424A6B394263424A6B4A5A4A
4A6352000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000394263394263394A6B394A6B4A6352
4A5A4A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000424A6B394263424A6B425A4A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000394263394A6B4A5A4A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003942634A6352000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000394A63000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[54]:=",
  ImageSize->{128, 128},
  AnimationDisplayTime->0.169,
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000200000P0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj2LW
9`00?g=cLda<C240<410Y/30`9T0IPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm?ooc020P800082006IVc3<cIP2009VIod:J
Yb8R8_mV<`00P00g??mVc5J0Zk75fBa1EkV>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7De3eDE:b/[:BT
Y>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEeMJ_0eLC4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7i/;2`W9bLP00
o`1o001oO`1oo`3o003oO`3oogl007l0Ogl0ogmo07moOgmoogoo07ooOgooool00?l0Ool0oomo0?mo
Oomooooo0?ooOoooo`0002XZ:UEEEGmoOjZZZ]GEeOoooh0700201`00P0L0080700031gT010L000<7
N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7@P000iX0000d00@7
00031d4000>JX000=@041`000`M00004Wj2OX0D000>P0000;@041`000`Lo0006VZ2JX9ZP0P000h6K
P@0_00@700031ch000fOX9nPWj2OX8:PVj2O02h010L000<7?@003YZPVZ2JX9ZPPI^1ViZO;P041`00
0`Ll000>PJ2OX9nPWj22X9^PVj02Wbd010L000<7?0004:2JX9ZPVY^1Vh6KPI^JWiX]00@700031c/0
00nPVZ2OX9nPPZ22X8:PVj000Yl]00@700031cX001>PVZ2JX86KPI^1Vh6KPI^JWiZO02`010L000<7
>@004J2JX9ZPPZ22X8:PPZ2KX9^P00>O;0041`000`Lh000EX9ZPVZ21Vh6KPI^1Vh6KPInJWiZO02`0
10L000<7>0004YZPWj21Vh:KPZ22X8:PPZ2OX0>O00>10000:@041`000`Lg000FVZ2JX86KPI^1Vh6K
PI^1Vh6OVYnJWb`010L000<78P02PA8001JOX9nPPJ22Vh:PPZ22X8:PVj2OX9nP0Yl[00@700031b80
00>16H4040006IZPVY^1Vh6KPI^1Vh6KPI^1Vh6OVYnJWh40:`041`000`LR00F13@005YniWj21X8:P
PZ22X8:PPZ22X8:PWj05WbX010L000<78P001h4IPAV1O8402P0079ZPVY^1Vh6KPI^1Vh6KPI^1Vh6K
PInJWiZOPIlZ00@700031b8028400`1k0004000KPInPPI^1Vh6KPZ22X8:PPZ22X8:PPZ2OX9nP00>O
00>100009`041`000`LR000;PAV16H4IPAV1O8400`008:27PX62PI^1Vh6KPI^1Vh6PPJ21X86PVZ2J
Wh6OPIn1:@041`000`LR00F100F2PH:1PP02P@0N0:21Vh:KPY^2Vh:PPZ22X8:PPZ22X8:PPZ2OX9nP
1IlX00@700031b8000b16H4IPAV16H4IPG/2P@0QPX62PH:1PX6KPI^1Vh6KPJ21X86PPJ2JX86OPIn1
Wh6O00H000=kP@007`041`000`LR00>102N2PH:1PX6JPIZ1Vh:KPY^2Vh:PPZ22X8:PPZ22X8:PPZ22
X9nPWj001Il30004Nh5kPB4010L000<78P00;h4IVX2JP9Z0VX2JP9Z1PX62PH:1Vh6KPI^1Vh6KPJ21
X86PPJ2JX86OPIn1Wh6O00:100EkPG^1P00Q00@700031a`03YX08X6JPIZ1VX:KPY^2X8:KPZ22X8:P
PZ22X8:PPZ22X9nPWj06W`03Nh6100>18@041`000`LI000oVZ2JXHNJP9Z0VX2JP9Z0VX2JP9Z0VX2J
PH:1PX6KPI^1Vh6KPJ21X86PPJ21X9ZPPIn1Wh6OPIn1Wg^1P860024010L000<74P0;X003XJ2J012J
01j1VX:KPY^2X8:PPZ22X8:PPZ22X8:PWj2OX9nPWj05W`:100>OPH408@041`000`L;001=Vj2JX9^P
VZ2KX9ZPVj2JX9ZQQj67XH2JP9Z0VX2JP9Z0VX2JP9Z0VX62PI^1Vh6KPI^1Vh6PPJ21X86PPJ21Wh6O
PIn1Wh6ONh60PH008@041`000`L901BP00JQX:6PXJ0AVP0HPZ22X8:PPZ22X8:PPZ22X8:PWj2OX9nP
1il019Z1VX4R00@700031`X004nPVj2KX9^PVj2KX9^PVj2JX9ZQQj67XHNQQj60VX2JP9Z0VX2JP9Z0
VX2JPH:1Vh6KPI^1X86PPJ21X86PVZ21X86OPIn1Wh6JNiYkVW^J020010L000<72`0BX00:XJ2QX:6P
XJ2QX0jJ01VPPZ22X8:PPZ22X8:PPZ22X9nPWj2OX9nP00>O00bJPIZ1VX6JPIZ1VX4L00@700031`/0
05BKX9^PVj2KX9^PVj2KX9ZPVZ22XHNPQj67X8NQQiZ0VX2JP9Z0VX2JP9Z0Vh6KPI^1Vh6PPJ21X86P
PJ21X86OPIn1Wh6JNiYkVW^JNiYkVW^JNiXJ00@700031``04J003J6PXJ2QX:6PXJ2QX:403IX05J22
X8:PPZ22X8:PPZ22X9nPWj2OX004W`0BPIZ1VX6JPIZ1VX6JPIZ1VX6J5`041`000`L=001HVj2KX9^P
Vj2KX9^PVZ2JX8:PQj27X8NPQj67XHNQPIZ0VX2JP9Z0VX2JP9^1Vh6KPJ21X86PPJ21X86PPJ21Wh6O
NiYkVW^JNiYkVW^JNiZ1VW^JNiYkVQ@010L000<73@0@X00@XJ2QX:6PXJ2QX:6PXJ2QX0bJ02ZPPZ22
X8:PPZ22X8:PWj2OX9nPWj21VX6JPIZ1VX6JPIZ1VX6JPIZ1VX45VP03V`0000h010L000<73P00GZ2K
X9^PVj2KX9^PVZ2KX8:PQj22X8NPQj67X8NQQj27X82JP9Z0VX2JP9Z0Vh6KPJ21X86PPJ2JX86PPIn1
Wh6JO9YkVWbJNiZ1VW^JPIYkVW^JNiYkVW^JNiYkVW/=00@700031`l03J005XNQX:6PXJ2QX:6PXJ2Q
X:6PXJ2QVZ09VP0AVh:PPZ22X8:PPZ2OX9nPWj000YX04h6JPIZ1VX6JPIZ1VX6JPIZ1VX4039X00j2J
000900@700031a0003FPVj2KX9^PVj2KX9^PQj27X8NPQj27X8NPQj27X8NPQj27X8NJP9Z0VX2JP9Z1
Vh6KPZ22X002W`0PVYn1Wg^JNiYkVW^JNiYkVW^JNiYkVW^JNiZ1VX6JPIX7X0/010L000<74@0<X00I
XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PX@06VP0OPJ22X8:PPZ2KX9nPWj2OVX6JPIZ1VX6JPIZ1VX6J
P@03VP03PIZJ00:J00^UX:FPYJ2UX:FPY@0;00@700031a4003BKX9^PVj2KX9^PVj27X8NPQj27X8NP
Qj27X8NPQj27X8NPQj27X8NJP9Z0VX:KPJ22Vh:P0Yl06IZOPIYkVW^JNiYkVW^JNiYkVW^JNiYkVX40
3J0<00@700031a802j007:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ05VP0;PZ22X8:PWj2OX9l0
1IX02h6JPIZ1VX6JPIZ100>J00ZPYZ2VX:FPYJ2U0j000jFPY@03X0X010L000<74`00=9^PVj2KX9^P
Vj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj20VX2JPY^2X8:PWj02VP0KO9YlVW^JNiYkVW^J
NiYkX8NPQj27X8NPQj2700>P00Z7X8NPQj2OX9n120041`000`LC00ZP01nQX:6PXJ2QX:6PXJ2QX:67
XJ2QX:6PXJ2QX:6PXJ2Q00BJ00N2X8:PWj2O00NJ01:1VX6JPIZPYZ2VX:JPYZ2VX:H@X0T010L000<7
5000Fj2KX9^PVj2KX8:PQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQiZ0VX:PPZ2OVWbJPIYl
VW^JO:27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X9nPWj002P041`000`LE00RP02>QX:6PXJ2QX:6P
XJ2QX:67XHNQX:6PXJ2QX:6PXJ2QX:6PX@02VP04PZ2OX0NJ01>VX:JPYZ2VX:JPYZ2VX:JPYZ2V00jP
2`041`000`LE001IVj2KX9^PVj22X8:PQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27VX:P
O9Z1VXNPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPWj2OX9l02`041`000`LF00NP04>QX:67
XJ2QX:6PXJ2QX:67XHNQQj6PXHNQX:6PXJ2QX:6PXJ2QX9ZPXJ2QX:6PXJ2VX:JPYZ2VX:JPYZ2VX:JP
YZ2V00fP30041`000`LG001EVj2KX9^PPZ27X8:PQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NP
PIZOVXNPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X9nPW`0=00@700031aP01J008:67
XHNQQj6PXJ2QX:6PXHNQQj67XHNQQj67XJ2QX:6P1IX3X00KWj2QX:6PXJ2VX:JPYZ2VX:JPYZ2VX:JP
YZ2U00^P3P041`000`LH001BX9^PVj22X8:PPZ22X8NPQj27X8NPQj27X8NPQj27X8NPQj1lVX6JPIZO
X8:PPYZ0X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X9nPW`l010L000<76@04X00SXHNQQj67
XHNQX:67XJ2QQj67XHNQQj67XHNQPIZ1VX6JPIX01Z0019ZOVZ43X00?YZ2VX:JPYZ2QX:JPYJ2V00>P
00>UX:001J000kVP000?00@700031aT004n1X9^PPZ22X8:PQj22X8NPQj22X8NPQj27X8NJO9YlVWbJ
O9Z1VYnPWj22X8:PP9Z0VXNPQj27X8NPQj27X8NPQj27X8NPQj27X8NPWj2O014010L000<760000g^P
X002X00PXHNQQj67XHNQQj67XHNQQj67XHNJPIZ1VX6JPIZ1Vh42VPFP01:2X:2PVYnJWj2QX:6PYZ2Q
X:4?X003^J2i014010L000<76000BX40Vi^PPZ22X8:PPZ22X8:PPZ21X7bJO9YlVWbJO9YlVWbJO9Z1
X8JPQZ22X8:PPZ20VX2JPJ27X8NPQj27X8NPQj27X8NPQj270j000in1000A00@700031aD000>1O840
0X408GbPX:67XHNQQj67XHNJPIZ1VX6JPIZ1VX6JPIZ1VX6JP@02VPFP00f2X8:QPZ62XIZOVYnJ01BP
00>iX0004`041`000`LD001@PG^1Nh40P@2PPZ22X8:PPYYlVWbJO9YlVX6JPIZ1VX6JPIZ1VX6JPJ26
X8JPQZ22X8:PPZ22X82OP9Z0VXNPQj27X8NPQj27X8NPQj27X9lE00@700031a8000ElPGb1O003P@0Q
O85lXHNQQiZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ100FP01:OX8:PPZ62XH:QPYnJWiZOVYlAX1H0
10L000<74P00D85kPG^1Nh40PG^1O9YlVWbJO9YlVWbJO9YlVX6JPIZ1VX6JPIZ1VX6PWj26X8JPPZ22
X8:PPZ22X8:OP9Z0VX2JP:27X8NPQj27X8NPQj2O5`041`000`L@0003O85l00>102AlPH61PXN2^X6J
PIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPI/8X00CPZ22XH:QPZ62XHNOVYnJWiZOVP0<X003^@0001D0
10L000<74000285kPG^1Nh5k0X40AhN1Qj2JO9YlVWbJO9YlVWbJO9Z1VX6JPIZ1VX6JPIZOX9nPQZ26
X8:PPZ22X8:PPZ22X86JVIZ0VX2JP:27X8NPQj27X8NP01P010L000<73P000gb1O005P@0RPXN2Qh:7
P[ZPVX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1V`RP01ROX8:PPZ62XH:QPZ67X9ZOVYnJWiZOVYl8X1T0
10L000<73P00DX5kPG^1PXN1Qh:7PHN1Qj2QO9YlVWbJO9YlVWbJO9Z1VX6JPIZ1VX6JWj2OX9nPQZ26
X8:PPZ22X8:PPZ22X8:JVIZ0VX2JP9Z0VXNPQj27X80I00@700031`/000B1O85l0X409HN2Qh:7PXN2
Qh:7X;ZP^X6JPIZ1VX6KPIZ1VX6JPIZ1Vh6JPI/02:007InPWj22X8:QPZ62XH:QQj6PWiZOVYnJWiZO
VYnJ00>P00>7000060041`000`L:001DP@21Nh:1PX67PHN1Qh67PHN1XJ2QX9YlVWbJO9YlVWbJO9Z1
VX6JPIZ1VYnPWj2OX9nPQZ26X8:PPZ22X8:PPZ22X8:PVIZIVX2JP9Z0VX2JP:276`041`000`L9000[
PGb7PXN2Qh:7PXN2Qh:7PXN2Qj2jX;ZPVX6JPIZ1Vh6KPIZ1VX6KPI^1V`0<X00OPZ22X8:QPZ62XH:Q
Qj6PXIZOVYnJWiZOVYnJWiZOP@0J00@700031`P005MZPH:1PX62PH:1PX62PH:1Qh67PJ6PXJ2QO9Yl
VWbJO9YlVWbJPIZ1VX6JWj2OX9nPWj2OX8JPQZ22X8:PPZ22X8:PPZ22X8:PVIZIVX2JP9Z0VX2JP8L0
6P041`000`L9000ZQh:7PXN2Qh:7PXN2Qh:7PXN2Qj2jX;ZP^X6KPI^1Vh6KPIZ1Vh6KPI^J2j008YnP
PZ22X8:QPZ62XH:QQj67XJ2QVYnJWiZOVYnJWiZOQh4I00@700031``005B2PH:1PX62PH:1PX62PHN1
XJ2QX:6PVWbJO9YlVWbJO9Z1VX6JPJ2OX9nPWj2OX9nPQZ26X8:PPZ22X8:PPZ22X8:PPZ22WiVJVIZ0
VX2JP9Z0Qh4I00@700031`l002:7PXN2Qh:7PXN2Qh:jX;ZP^Z2jX9^1Vh6KPI^1Vh6KPI^J3J0089nP
PZ22XH:QPZ62XH:QQj6PXJ2QX9nJWiZOVYnJWiZO0XL00h40000F00@700031a8001j2PH:1PX62PHNP
XJ2QX:6PXGbJO9YlVWbJO9Z1VX43X00_Wj2OX9nPWj26X8JPPZ22X8:PPZ22X8:PPZ22X8:PPZ22WiVJ
VIZ0VX2JP8N1Qg/05`041`000`LE000=PAV1PXN2^Z2QX;ZP^P02X00@PIZ1Vh6KPI^1ViZUX:FPY@ZP
02BOX8:PPZ22XH:QPZ62XHNQQj67XHNQVYnJWiZOVYnJWh67PHL2PAD010L000<76P004h:PXJ2QX:6P
XJ2JPIZ1VX6JPIX00Z00=InPWj26X8JPQZ26X8JPQZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PPYn0Wh2O
P9n0Qh67PHN101D010L000<76P006:2QX:6PXJ2QX:6PVh6KPI^JVh6UX:FPY@ZP02VOX8:PPZ22X8NP
Qj27XHNPQj62XHNQQj67XHNOVYnJWiZOPHN1Qh67P@0D00@700031aX001:QX:6PXJ2QX:6PXH6JPIZ1
VX43X00gWj2OX8JPQZ26X8JPQZ26X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X9VOP9n0Wh27PHN1Qh67
P@0C00@700031aX0016PXJ2QX:6PXJ2QX:61ViZKVP02X007YJ2UX:FPY@08X00YWj22X8:PPZ22X8NP
Qj27XHNQQj67XHNQQj67XIZOVYnJWh67PHN1Qh4050041`000`LJ000AXJ2QX:6PXJ2QX:6PVX6JPIX0
0Z00<YnPQZ26X8JPQZ26X8JPQZ26X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PVIn0Wh276P041`00
0`LJ000HX:6PXJ2QX:6PXJ2QX9^JVj2UX:FPYJ2U2Z008YnPPZ22X8:PQj27X8NQQj27XH:QQj67XHNQ
Qj67WiZOVYlK00@700031aX000nQX:6PXJ2QX:6PXJ2QPIX00Z00<inPWj26X8JPQZ26X8JPQZ26X8:P
PZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22WiVOP00K00@700031aT001N1X:6PXJ2QX:6PXJ2QX:6P
YJ2UX:FPY@0:X00TWj2OX8:PPZ22X8:PQj27X8NQQj67XHNQQj67XHNQQj67WiZO6`041`000`LI000@
PJ6PXJ2QX:6PXJ2QX:6PX@BP0366X8JPQZ26X8JPQZ26X8JPPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:P
PZ22X8:PPYnI01/010L000<76@006H6PXJ2QX:6PXJ2QX:6PYZ2VX;nPYJ2UX:D02:0099nPWj22X8:P
PZ27X8:PQj67XHNQPZ67XHNQQj67XHNQQj6JWa/010L000<76@00AX6QX:6PXJ2QX:6PXJ2QX:6PXJ2Q
X:66X8JPQZ26X8JPQZ26X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:QVIlJ00@700031aT0
01V1X:6PXJ2QX:6PXJ2QX:6PYZ2oX;nP_j2o00RP02FOX9nPPZ22X8:PPZ27X8NPQj67XHNQQj67XHNQ
Qj67XHNQQj6J01X010L000<76@007:2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:42X00XQZ26X8JP
PZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22WaX010L000<76@00A:6PXJ2QX:6PXJ2QX:6P
YZ2VX;nP_j2oX;nP_j2oX;nP_inPQj22X8:PPZ27X8NPQj67XHNQP[nP_j2oX;nP_j2nX;nP70041`00
0`LI000`X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2VX:FPYJ2UX:FPYJ2U5:0L00@7
00031aT002BQX:6PXJ2QX:6PXJ2QX:JPYZ2oX;nP_j2oX;nP_j2oX8N2Qh42X00N_j2oX;nP_j2oX;nP
_j2oX;nP_j2oX;nP_j2oX;jP70041`000`LI000^X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ22
PH:1PX66PHJ1YJ2UX:FPY@NP00>UX:003:0L00@700031aP001^1XJ2VX:6PYZ2VX:JPYZ2oX:JP_j2o
X;n2X8L02H801X6OPHJ1QP:101VP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP01`010L000<760005:6P
XJ2QX:6PXJ2QX:6PXJ2QX:6P0P004h:1PX62PH:1PX62PH:1QX66PHH01@001jFPYJ2UX:D04:0L00@7
00031aP001:PXJ2QX:JPYZ2VX:JPYZ2VX:H40003PH:100Z200JPPJ21QX47000DX;nP_j2oX;nP_j2n
X;jP_Z2nX;hM00@700031aP000nQX:6PXJ2QX:6PXJ2QX:4020004FZ2PH:1PX62PH:1PX66PHJ100X0
00>UX:D03j0M00@700031aP000bPXJ2VX:6PYZ2VX:H;0003PX6200V200BPPHJ13@004:2oX;nP_j2o
X;jP_Z2UX;hM00@700031aL000ZPXJ2QX:6PXJ2Q3`003h:1PX62PH:1PX66PHJ1QP0>0003PJ2P00bP
7@041`000`LG0006XJ2QX:6P4`000h62P@07PP04PHJ1QQ4000foX;nP_j2nX:FPYJ2U01d010L000<7
5`000j2QX00G000<PH:1PX62PH:1QX664`0<X1d010L000<75`000j40000G00V200>1QX405@002Z2o
X:FPYJ2UX:DM00@700031c8000V2PH:1PX62PHH00X4G00RP7@041`000`Lb0003PH:200B200>6PHH0
6P001Z2UX:FPYAd010L000<7<`002FZ2PH:1PX66P@0L00BP7@041`000`Lc0003PX6200:200>1QX40
7`000j2U000L00@700031c@000J2PH:1QX4o00@700031c@000J1PX:2PHHo00@700031cD000B1PX66
@0041`000`Le00:200>60000?`041`000`Lf0003PP00040010L000<7N@041`000`Mi00@700031gT0
10L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700201`00P0L008070000\
\>"],
  ImageRangeCache->{{{0, 127}, {127, 0}} -> {-3.20067, -3.20067, 1.05828, 
  1.05828}},
  CellTags->"Animation"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
%%ImageSize: 128 128 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00744048 0.0238095 0.00744048 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
384 string
128 128 8 [128 0 0 128 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000735A5A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000736363735A5A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000735A63735A63735A63
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000947B6B000000000000000000736363735A5A7B6363735A5A
7B6363000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000039395A947B63947B638C7B63000000735A63735A63735A63736363
735A63736363000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000039395A947B6B8C7B63947B6B8C73637B6363735A637B6363735A5A
7B6363735A637B6363000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000039395A39395A39395A947B6B8C7B63947B638C7B637B6363735A63736363
7363637B6363735A637B6363000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000042396339395A4239638C7B63947B6B8C7B63947B6B8C73637B6363735A63
7B6363735A637B636B735A637B6363000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000039395A39395A39395A39395A947B6B947B63947B6B8C7B638C7B638C7B638C7B63
735A637B63637B63637B63637B63637B6363000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000039395A42396339395A4242638C7B63947B6B8C7363947B6B8C7B63947B638C7363
947B63735A637B6363735A637B636B7B63637B636B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39395A39395A39395A39395A39395A947B6B8C7B63947B638C7B63947B6B8C7B638C7B63
8C7B638C7B637B63637B63637B63637B636B7B6363000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39395A39395A42396339395A4239638C7B63947B6B8C7B63947B6B8C7B63947B6B8C7363
947B638C73638C7B63735A637B636B7B63637B636B7B6363000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39395A39395A39395A39396339395A947B6B947B63947B6B8C7B63947B638C7B638C7B63
8C7B638C7B638C7B638C7B638C73637B636B7B63637B636B7B6363000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000393152
42396339395A39396339395A4242638C7B63947B6B8C7B6394846B8C7B63947B6B8C7363
947B6B8C73638C7B638C735A8C7B638C73637B636B7B6363846B6B7B6363000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039395A
39395A39395A39395A39395A393963947B6B8C7B63947B6B8C7B63947B6B8C7B638C7B63
8C7B638C7B638C7B638C7B638C7B638C7B638C73637B636B7B636B84636B7B6363000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039395A39395A
39395A39395A42396339395A42426339395A94846B8C7B63947B6B8C7B63947B6B8C7B63
947B6B8C7363947B6B8C73638C7B6384735A8C7B6384735A73848C7B6363846B6B7B6363
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000292929000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039395A39395A
39395A39395A39395A39396339395A393963947B6B947B6B8C7B63947B6B8C7B638C7B6B
8C7B638C7B638C7B638C7B638C7B638C7B638C73638C7B638473638C73637B636B846B6B
7B636B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000635A42635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039315239395A393152
42396339395A39396339395A42426339395A94846B8C7B6394846B8C7B63947B6B8C7B63
947B6B8C7B638C7B638C73638C7B6B8C73638C7B6384735A8C7B6384735A8C7B637B6363
846B6B7B636B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000313131635A426B5A42635242000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039395A39395A39395A
39395A39395A39395A3939633939634239638C7B6B947B6B947B6B947B6B8C7B638C7B6B
8C7B638C7B6B8C7B638C7B638C7B638C7B638C73638C7B638C73638C7B6384735A847363
846B6B846B6B84636B000000000000000000000000000000000000000000000000000000
0000000000003129293129296B5A426B5A42635A42635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039395A31315239395A39395A
39396339395A42396339395A42396339395A94846B8C7B6394846B8C7B6394846B8C7B63
947B6B8C7B63947B6B8C73638C7B6B8C73638C7B638473638C7B6384735A8C7B6384735A
8C7B636B7B84846B6B000000000000000000000000000000000000000000000000000000
0000002929296B634A635A426B634A635A426B5A4A635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039395239395A39395A39395A
39395A39395A39395A39396339395A393963947B6B947B6B8C7B6B947B6B8C7B6B947B6B
8C7B638C7B638C7B638C7B638C7B638C7B638C7B638C7B638473638C7B638473638C7B63
84735A84736384735A846B6B000000000000000000000000000000000000000000312931
3129296B634A6B63426B634A635A426B5A42635A42635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031315239395A39395A39395A39315A
39396339395A39396339395A42426339395A94846B8C7B6394846B8C7B63947B6B8C7B63
94846B8C7B638C7B6B8C73638C7B6B8C7B638C7B638473638C7B638473638C7B6384735A
8C7B6384735A8C7B6384735A8C6B73000000000000000000000000292929313131312929
73634A6B63426B634A635A426B634A635A426B5A42635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039395A39395A39395A39395A39395A
39395A39396339395A393963393963423963947B6B947B6B947B6B94846B8C7B638C7B6B
8C7B6B8C7B6B8C7B638C7B638C7B638C7B638C7B638C7B638C7B638C7B638473638C7B63
8473638C7B6384735A84735A84735A846B6B0000000000003129293131316B634A73634A
6B634A6B634A6B5A426B634A6B5A426B5A4A635A42635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000312942422929
4A313100000000000000000000000000000000000031315239395A31315239395A39395A
39396339395A42396339395A42426339395A94846B8C7B6B94846B8C7B6394846B8C7B63
94846B8C7B63947B6B8C7B638C7B6B8C73638C7B6B8473638C7B638473638C7B6384735A
8C7B6384735A8C7B6384735A847B6384735A73848C3129293131316B634A736B4A6B634A
73634A6B5A426B634A635A426B634A635A426B634A635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000312939
4229294A29294229294A292942292900000039395239395A39395A39395A39395A39395A
39395A39395A39395A39396339396339396339396394846B8C7B6B947B6B8C7B6B947B6B
8C7B638C7B6B8C7B638C7B6B8C7B638C7B638C7B638C7B638C7B638C7B63847B638C7B63
847363847B6384735A84736384735A84735A84735A84735A73634A73634A73634A73634A
6B634A6B634A6B634A6B634A635A426B5A42635A42635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292939
3131424229294A31314229294A313142292939395A31315239395A39395A39395A31315A
42396339395A39396339395A4242633939634242638C7B6B94846B8C7B6B94846B8C7B63
94846B8C7B638C7B6B8C7B6394846B8C7B638C7B6B8473638C7B6B8473638C7B6384735A
8C7B6384735A8C7B6384735A8C7B6384735A8473634229314A31394A29314A3139422931
4A31394A29314A313942426B6B634A635A426B634A635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292939
2929393129424A29294A31294A292939395A31315239395A39395A39395A39395A39395A
39395A39396339395A39396339396339396339396394846B94846B94846B8C7B6B8C7B6B
8C7B6B8C7B6B8C7B638C7B638C7B638C7B6B8C7B638C7B638C7B638C7B638473638C7B63
847B638C7B6384735A84736384735A8473634229314A31314A31314A31314A31314A3131
4A31314A31394A31314A31314A31315231394A3131523139523131523139424A73000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292939
3129422929395231314A292952313131315239395A31315239395A31315239395A39395A
39396339395A42426339395A42426339395A4242638C7B6B94846B8C7B6B94846B8C7B63
94846B8C7B6394846B8C7B638C7B6B8C7B638C7B6B847B638C7B638473638C7B63847363
8C7B6384735A8C7B6384735A847B634229314A31394229314A31394A29314A31394A2931
5231394A29315231394A31315231394A31315231394A31315231394A3131523139523131
5A31395231315A393984948C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000312939
29293931294231294252313131315239395A39395239395A39395A39395A39395A39395A
39395A39395A39395A39396339396339396339396394846B947B6B94846B8C7B6B94846B
8C7B6B8C7B6B8C7B6B8C7B6B8C7B638C7B6B8C7B638C7B638C7B638C7B63847B638C7B63
847B63847B63847363847B634229314A31314A31314A31314A31314A31314A31314A3139
4A31314A31314A31315231394A3131523139523131523139523131523139523131523139
523139849494849494849494947B73947B73947B73947B73000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292939
3131422929393131425A393139395A31315239395A31315239395A39395A39395A31315A
39396339395A39396339395A4242633939634242638C7B6B9484738C7B6B94846B8C7B63
94846B8C7B638C846B8C7B6394846B8C7B638C7B6B8473638C7B6B847B638C7B63847363
8C7B638473638C7B634229314A31394A29314A31394A29315231394A31315231394A2931
5231394A31315231394A29315231395231315231394A31315A3939523131523139848C8C
8C9494848C8C849494848C8C947B73947B73947B738C736B947B73947B6B947B738C736B
947B73947B6B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
63423963423963423163423963423139395A31315239395A39395A39395A39395A39395A
39395A39396339395A39396339396339396339396394846B94846B94846B8C7B6B8C846B
8C7B6B94846B8C7B638C7B6B8C7B6B8C7B6B8C7B638C7B638C7B638C7B63847B638C7B63
847B638C7B634229314A31314A31314A31394A31314A31314A31314A31394A31314A3131
4A31315231394A3131523139523139523139523131523139523139523139848C94849494
84949484949484948C849494947B73947B73947B73947B73947B73947B73947B73947B73
947B73947B73947B73947B73947B73947B73947B6B947B73000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000634239633931634239634231
6342396342316342396342316342395A393139395A31315239395A31315A39395A39395A
39396339395A39396339395A42426339395A4242638C7B6B9484738C7B6B94846B8C7B6B
94846B8C7B6394846B8C7B638C846B8C7B638C846B847B638C7B6B847B638C7B6B847363
8C7B634229314A31394A29314A31394A29315231394A31315231394A29315231394A3131
5231394A31315231394A31315231395231315231395231318494947B8C8C849494848C8C
849494848C8C849494947B73947B73947373947B73947B73947B7394736B947B73947B73
947B73947B6B947B73947B6B947B73947B6B947B73947B6B947B73947B6B000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000634239634239634239634239634239634239634239
63423963423963423963423963423163423963423939395A39395A39395A39395A39395A
39395A39396339395A39396339396339396339396394846B94846B94846B8C846B94846B
8C7B6B8C846B8C7B6B8C7B6B8C7B638C7B6B8C7B638C7B6B8C7B638C7B638C7B638C7B63
847B634A31314A31314A31394A31314A31394A31314A31394A31315231394A3131523139
523131523139523139523139523131523139848C94849494848C94849494848C94849494
848C8C84949484948C947B73947B73947B73947B73947B73947B73947B73947B73947B73
947B73947B73947B73947B73947B73947B73947B73947B73947B73000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000006B42396342316B4A396342316B42396342316B4A396342316B4239634231
6B4A396342316342396342316B4A3963423163423963423139395A31395A39395A31315A
39396339395A39396339395A4242633939634242633939639484738C7B6B94846B8C7B6B
94846B8C7B6B8C846B8C7B6394846B8C7B638C7B6B847B638C846B847B638C7B6B847363
4A31394A31314A31394A29315231394A31315231394A29315231394A31315231394A3131
5231395231315231395231318494947B8C8C8494947B8C8C8494947B8C8C8494947B8C8C
8494947B8C8C849494947B739C7B7B947B73947B73947B739C847B947B73947B73947B73
9C7B73947B73947B73947B6B9C8473947B739C7B73947B6B9C8473000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000634239634239
6342396B42396342396342396342396B42396342396342396342396B4239634239634239
6342396B42396342396342396342396B423963423963423939395A39395A39395A39395A
39395A39396339395A39396339396339396339396339396394846B94846B8C846B94846B
8C846B94846B8C7B6B8C7B6B8C7B6B8C846B8C7B638C7B6B8C7B638C7B6B847B634A3131
4A31314A31394A31314A31394A31315231394A31314A3139523139523139523131523139
5231395231397B8C94848C94848C948494947B8C8C848C94848C948494947B8C8C848C94
848C8C8494947B8C8C947B73947B73947B73947B73947B73947B73947B73947B73947B73
947B73947B73947B73947B73947B73947B73947B73947B73000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006342396B4A396342396B4A39634239
6B4A396342396B4A396342396B4A396342316B4A396342396B4A396342316B4A39634239
6B4A396342316B4A396342316B4A396342316B4A396342316B4A3931315A39395A39395A
39396339395A39426339395A3942633939634242633939639484738C7B6B9484738C7B6B
94846B8C7B6B94846B8C7B638C846B8C7B638C846B8C7B638C7B6B847B634A31394A2931
4A31394A29315231394A31315231394A29315231394A31315231394A3131523139523131
8494947B8C8C8494947B8C8C8494947B8C8C8494947B8C8C8494947B8C8C8494947B8C8C
8494947B8C8C849494947B739C7B7B947B739C847B947B739C847B947B739C847B947B73
9C7B73947B739C8473947B739C8473947B739C8473000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006B42396B42396B42396B42396B4A396B42396B42396B42396B4239
6B42396B42396B42396B42396342396B42396B42396B42396B42396B42396B42396B4239
6342396B42396342396B42396342396B42396B42396B42396342396B423939395A39395A
39395A39396339396339396339396339396339396339426394846B94846B8C846B94846B
8C7B6B8C846B8C7B6B8C846B8C7B6B8C7B6B8C7B6B8C7B6B8C7B634A31314A31314A3139
4A31314A31394A31315231394A31315231394A31395231395231395231397B8C947B8C94
7B8C947B8C947B8C947B8C947B8C8C848C947B8C8C848C947B8C8C848C947B8C8C848C94
7B8C8C84948C7B8C8C947B7B947B73947B73947B73947B73947B73947B73947B739C7B73
947B73947B73947B73947B73947B73947B73947B73000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
6B4A426B42396B4A396342396B4A426B42396B4A396342396B4A426B42396B4A39634239
6B4A426B42396B4A396342396B4A396B42396B4A396342396B4A396342396B4A39634231
6B4A396342396B4A396342316B4A396342396B4A396342316B4A396342396B4A3931315A
39396339395A39396339395A4242633939634242633939639484738C846B9484738C7B6B
9484738C7B6B94846B8C7B6394846B8C7B638C846B847B634A31394A31314A31394A2931
5231394A31315231394A29315231394A31315231394A31318494947B8C947B8C947B8C8C
8494947B8C8C848C947B8C8C8494947B8C8C8494947B8C8C8494947B8C8C8494947B8C8C
8494947B8C8C849494947B739C847B947B739C847B947B739C847B947B739C847B947B73
9C847B947B739C8473947B739C847B947B73000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
5A6B8C63738C5A6B8C6B4A396B42396B4A396B42396B4A396B42396B4A396B42396B4A39
6B42396B4A396B42396B42396B42396B4A396B42396B42396B42396B4A396B42396B4239
6B42396B4A396B42396B42396B42396B4A396B42396B42396B42396B4A396B42396B4239
39395A39396339395A39396339396339396339396339396394846B9484738C846B94846B
8C846B94846B8C7B6B8C846B8C7B6B8C846B8C7B6B4A31314A31314A31394A31314A3139
4A31395231394A31315231395231395231397B8C947B8C947B8C947B8C947B8C8C7B8C94
7B8C947B8C947B8C8C7B8C947B8C8C7B8C947B8C8C7B8C8C7B8C8C7B8C947B8C8C7B8C8C
7B8C8C848C8C7B8C8C947B73947B739C847B947B73947B73947B739C847B947B73947B73
947B739C8473947B73947B73947B73000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A6B8C63738C5A6B8C63738C5A6B8C6373946B42396B4A426B4239734A426B4239
6B4A396B42396B4A426B42396B4A396B42396B4A426B42396B4A396B42396B4A396B4239
6B4A396B42396B4A396B42396B4A396B42396B4A396B42396B4A396B42396B4A396B4239
6B4A3939395A39426339395A39426339396342426B3939639484738C846B9484738C7B6B
9484738C7B6B94846B8C7B6B8C846B8C7B634A31394A29314A31394A29315231394A3131
5231394A31315231394A31317B8C947B8C8C7B8C947B8C8C7B8C947B8C8C7B8C947B8C8C
7B8C947B8C8C7B94947B8C8C7B8C947B8C8C8494947B8C8C8494947B8C8C8494947B8C8C
8494947B8C8C849494947B739C847B947B739C847B947B739C847B947B739C847B947B73
9C847B947B739C847B947B739C847B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A738C5A6B8C5A738C5A6B8C5A738C5A738C63738C5A738C63738C6B4A396B4A39
6B42396B4A396B4A396B4A396B42396B4A396B4A396B4A396B42396B4A396B42396B4A39
6B42396B4A396B42396B4A396B42396B4A396B42396B4A396B42396B4A396B42396B4A39
6B42396B4A3939396339396339396339396339396339426394846B9484738C846B94846B
8C846B8C846B8C846B8C846B8C7B6B4A31394A31314A31394A31314A31394A3139523139
4A31315231395231397B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C94
7B8C8C7B8C947B8C8C7B8C947B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C
7B8C8C7B8C8C7B8C8C9C847B947B739C847B947B739C847B947B739C847B947B739C847B
947B739C8473947B739C8473000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000635A6B5A738C5A6B8463738C5A6B8C63738C5A6B8C6373945A6B8C63738C5A6B8C
6373945A738C734A426B4239734A426B4239734A396B4239734A426B4239734A396B4239
734A426B4239734A396B4239734A426B4239734A396B4239734A426B4239734A396B4239
734A426B423939396339395A424263393963394263393963948C738C846B9484738C7B6B
9484738C7B6B94846B8C7B6B94846B4A31314A31394A29315231394A31315231394A3131
523939738C947B8C9473848C7B8C94738C8C7B8C9473848C7B8C947B8C8C7B8C9473848C
7B8C947B8C8C7B8C9473848C7B94947B8C8C7B8C94738C8C7B94947B8C8C7B8C8C738C8C
8494947B8C8C7B8C8C947B739C847B947B739C847B947B739C847B947B739C847B947B73
9C847B947B739C847B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000635A6B
635A6B6B5A6B5A6B8C5A6B8C5A6B8C63738C5A6B8C5A738C5A738C63738C5A6B8C63738C
63738C63738C5A738C63738C63738C6373946B42396B4A396B4A39734A396B4A39734A39
6B4A39734A396B42396B4A396B4A39734A396B4A396B4A396B4A39734A396B42396B4A39
6B4A39734A396B42393939633939633939633939633939633939639484738C846B94846B
8C846B94846B8C846B8C846B4A31394A31394A31314A31394A31395231394A31317B8C94
7B8C947B8C94738C947B8C947B8C947B8C94738C8C7B8C947B8C947B8C947B8C8C7B8C94
7B8C8C7B8C947B8C8C7B8C8C7B8C8C7B8C947B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C
7B8C8C7B8C8C7B8C8C9C847B9C7B7B9C847B947B739C847B9C847B9C847B947B739C847B
9C7B739C847B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000063526B6B5A73635A6B
6B5A73635A6B6B5A735A6B8463738C5A6B8463738C5A6B8C63738C5A6B8C63738C5A6B8C
63738C5A6B8C6373945A738C6373945A738C6373945A738C637B945A738C734A426B4239
734A426B4239734A426B4239734A426B4239734A426B4239734A396B4239734A426B4239
734A396B4239734A426B423939426339396342426B39396342426B8C846B9484738C846B
9484738C7B6B9484734A31315231394A31315231394A313152313973848C7B8C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C94738C8C
7B8C9473848C7B8C94738C8C7B8C94738C8C7B8C94738C8C7B8C8C738C8C7B8C8C738C8C
7B8C8C738C8C9C847B947B739C847B947B739C847B947B739C847B947B739C847B947B73
9C847B947B73000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000635A6B635A6B
635A6B635A6B635A6B6B5A735A6B8C5A6B8C5A6B8C5A738C5A6B8C5A738C5A738C63738C
5A738C63738C5A738C63738C5A738C63738C63738C63738C63738C63738C63738C637394
63738C637394734A39734A39734A39734A39734A39734A396B4A39734A39734A39734A39
734A39734A39734A39734A396B4A3939396339396339426339396394847394846B948473
8C846B8C846B4A31394A31394A31394A31394A31397B8C94738494738C94738C947B8C94
738C947B8C94738C947B8C94738C8C7B8C94738C8C7B8C94738C8C7B8C94738C8C7B8C94
738C8C7B8C8C738C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C7B8C8C
7B8C8C7B8C8C9C847B9C847B9C7B7B9C847B9C847B9C847B9C847B9C847B9C847B9C847B
9C7B73000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000063526B
6B5A73635A6B6B5A73635A6B63738C5A6B845A738C5A6B8463738C5A6B8C63738C5A6B84
63738C5A6B8C63738C5A6B8C6373945A738C63738C5A738C637B945A738C6373945A738C
637B9463738C637B9463738C637B9463738C734A426B4239734A42734A39734A426B4239
734A426B4A39734A396B4239734A426B4A3939426B39396342426B8C846B9484738C846B
9484734A31314A31394A29315231397384947B8C9473848C7B8C947384947B8C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C7B8C9473848C7B8C94738C8C7B8C8C73848C7B8C94738C8C7B8C8C738484
7B8C8C738C8C9C847B947B739C847B9C7B7B9C847B947B739C847B9C7B739C847B947B73
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000635A6B
635A6B6B5A73635A6B6B5A736B5A6B5A6B8C5A6B845A6B8C5A6B8C5A738C5A6B8C5A738C
5A738C63738C5A6B8C5A738C5A738C63738C5A738C63738C63738C63738C5A738C63738C
63738C63739463738C63738C63738C637B9463738C637B94637B94637B94734A39734A39
734A39734A39734A39734A39734A39734A39734A393939633939639484738C846B948473
4A31394A31394A3139738494738494738C94738494738C94738494738C9473848C738C94
738C94738C9473848C738C94738C8C7B8C9473848C738C8C738C8C7B8C9473848C738C8C
738C8C7B8C8C73848C738C8C738C8C7B8C8C738C8C7B8C8C738C8C7B8C8C738C8C7B8C8C
7B8C8C7B8C8C9C7B7B9C847B9C847B9C847B9C847B9C847B9C847B9C847B9C847B9C847B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
6B5A73635A6B6B6373635A6B6B63735A6B845A738C5A6B845A738C5A6B8463738C5A6B84
63738C5A6B8463738C5A6B8C63738C5A6B8C63738C5A738C63738C5A738C6373945A738C
637B945A738C637B9463738C637B9463738C637B9463738C637B9463738C6B7B9463738C
6B7B94734A39734A42734A39734A42734A39734A42734A3942426B8C846B948C734A3131
523139738494738C94738494738C9473848C7B8C9473848C738C9473848C7B8C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C7B8C9473848C7B8C8C73848C7B8C8C73848C7B8C8C73848C7B8C8C73848C
7B8C8C7384849C847B9C7B7B9C847B947B7B9C847B9C7B7B9C847B947B739C847B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000635A73635A6B6B5A73635A6B6B5A735A6B845A6B8C5A6B845A6B8C5A6B8C5A738C
5A6B8C5A738C5A6B8C5A738C5A738C5A738C5A738C63738C5A738C63738C5A738C63738C
63738C63738C63738C63738C63738C63738C63738C637B9463738C637B94637B8C637B94
637B8C637B94637B94637B94637B94734A39734A39734A393939639484734A3139738494
738494738494738494738494738494738494738494738C94738494738C94738494738C94
73848C738C9473848C738C9473848C738C8C73848C738C8C73848C738C8C73848C738C8C
73848C738C8C738C8C738C8C73848C738C8C738C8C738C8C738C8C738C8C738C8C7B8C8C
738C8C738C8C9C847B9C847B9C847B9C847B9C847B9C847B9C847B9C847B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000006B5A73635A6B6B6373635A6B6B6373526B845A738C5A6B845A738C5A6B84
63738C5A6B8463738C5A6B8463738C5A6B8C63738C5A6B8463738C5A738C63738C5A738C
6373945A738C63738C5A738C637B9463738C637B945A738C637B9463738C637B9463738C
6B7B9463738C637B9463738C6B7B94637B8C6B7B9463738C6B84944A3131738C946B848C
738C94738494738C946B848C738C9473848C738C946B848C738C9473848C738C946B848C
738C9473848C738C9473848C738C9473848C738C9473848C7B8C9473848C738C8C73848C
7B8C9473848C738C8C73848C7B8C8C73848C7B8C8C7384847B8C8C73848C7B8C8C738484
7B8C8C7384849C847B9C7B7B9C84849C847B9C847B9C7B7BA5847B9C847B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006B5A736B5A736B63736B5A735A6B845A6B845A6B8C5A6B845A6B8C
5A6B8C5A738C5A6B845A738C5A738C5A738C5A6B8C5A738C5A738C63738C5A738C63738C
5A738C63738C5A738C63738C63738C63738C63738C63738C63738C637B9463738C637B8C
637B8C637B9463738C637B94637B94637B94637B8C637B94423139948C7B7B5A42845A42
73849473849473849473849473849473849473848C73849473849473849473848C73848C
73848C738C9473848C73848C73848C738C8C73848C73848C73848C738C8C73848C738C8C
73848C738C8C73848C738C8C738C8C738C8C73848C738C8C738C8C738C8C738484738C8C
738C8C738C8C9C847B9C847B9C847B9C847B9C847B9C847B9C847B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006B6373635A6B6B6373635A735A6B8C526B845A738C5A6B84
5A738C5A6B8463738C5A6B845A738C5A6B8463738C5A6B8463738C5A6B8463738C5A738C
63738C5A738C63738C5A738C63738C5A738C637B945A738C637B945A738C637B9463738C
637B9463738C637B9463738C637B9463738C4A3142423139948C848C8C7B424A7B7B5A42
845A4A7B5242845A4A7B5242738C946B848C738C946B848C738C946B848C738C946B848C
738C946B848C738C9473848C738C946B848C738C9473848C738C9473848C738C9473848C
738C8C73848C738C8C73848C738C8C738484738C8C738484738C8C738484738C8C738484
738C8C738484A584849C847B9C84849C847BA584849C847B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000635A736B63736B5A736B63735A6B845A6B845A6B845A6B8C
5A6B845A6B8C5A6B845A738C5A6B845A738C5A6B8C5A738C5A738C5A738C5A738C63738C
5A738C63738C5A738C63738C5A738C63738C63738C63738C63738C63738C63738C637B8C
63738C637B8C63738C637B9463738C423139423139948C7B8C8C7B8C8C7B424A73424A73
7B5A42845A427B5A42845A427B5A427B5A427B5A427384946B848C73849473848C738494
73848C73848C73848C73848C73848C73848C73848C73848C73848C73848C73848C738C8C
73848C73848C73848C738C8C73848C738C8C73848C738C8C73848C738C8C73848C738C8C
738484738C8C9C847B9C847B9C847B9C847B9C847B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000635A736B6373635A736B637B526B845A6B8C526B84
5A738C5A6B845A738C526B8463738C5A6B845A738C5A6B8463738C5A6B8463738C5A6B84
63738C5A738C63738C5A6B8463738C5A738C63738C5A738C637B945A738C637B8C5A738C
637B9463738C637B944229394A31424231394231428C8C7B9494848C8C7B424A73394A73
42527B7B5A42845A4A7B5242845A4A7B5A42845A4A7B5242845A4A7B52427384946B848C
738C946B848C7384946B848C738C946B848C738C8C6B848C738C946B848C738C8C6B8484
738C8C73848C738C8C6B8484738C8C738484738C8C6B8484738C8C738484738C8C6B8484
738C8C7384849C84849C847BA58C849C847BA58484000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B5A736B63736B63735A6B84526B845A6B84
5A6B845A6B8C5A6B845A6B845A6B845A738C5A6B845A738C5A738C5A738C5A6B845A738C
5A738C63738C5A738C5A738C5A738C63738C5A738C63738C63738C63738C5A738C63738C
63738C637B8C4231394231394231394231428C8C7B8C8C7B8C8C7B8C8C7B424A73424A73
424A73424A7B845A42845A42845A42845A4A7B5A42845A427B5A42845A427B5A427B5A42
7B5A4273848C6B848C73848C73848C73848C6B848C73848C73848C73848C6B848C73848C
73848C73848C6B848C73848C73848C738C8C73848473848C73848C738C8C73848473848C
738484738C8C9C847B9C847B9C847BA58484000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000635A736B637B6B5A735A6B8C526B84
5A6B8C526B845A738C526B845A738C5A6B845A738C5A6B845A738C5A6B8463738C5A6B84
63738C5A6B8463738C5A738463738C5A738463738C5A738C63738C5A738C637B8C5A738C
4231424229394231424231394A31428C8C7B9494848C8C7B948C848C8C7B424A73424A73
424A7B424A7384634A7B5A42845A4A7B5A42845A4A7B5A42845A4A7B5A42845A4A7B5A42
845A4A7B5242845A4A7B524273848C6B848C738C8C6B848C73848C6B8484738C8C6B8484
738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484
738C8C9C847BA58C849C847BA58C84523929000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006B63736B63736B63735A6B84
526B845A6B845A6B845A6B845A6B845A6B845A6B845A738C5A6B845A738C5A6B845A738C
5A6B845A738C5A738C5A738C5A738C5A738C5A738C63738C5A738C63738C5A738C423139
4231394231424231394231428C8C7B8C8C848C8C7B8C8C848C8C7B8C8C7B424A73424A73
424A73424A73424A73845A4A845A42845A4A845A42845A4A845A42845A427B5A42845A42
7B5A427B5A427B5A427B5A427B5A427B5A427B5A427B5A426B848C73848C6B848C73848C
6B848C73848C6B848C73848C6B848473848C6B848473848C6B848473848C73848473848C
6B84849C84849C847BA584849C847B5A42315A4229000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006B5A736B637B526384
5A6B8C526B845A6B8C526B845A738C526B845A738C526B845A738C5A6B845A738C5A6B84
63738C5A6B845A738C5A6B8463738C5A738463738C5A6B8463738C392939423142392939
4231424231394231424231399494848C8C7B8C8C848C8C7B9494848C8C7B8C8C7B394A73
424A7B424A73424A7B424A7384634A845A4284634A7B5A4284634A7B5A42845A4A7B5A42
84634A7B5A42845A4A7B5242845A4A7B5A42845A427B5242845A4A7B52427B5A426B8484
738C8C6B848473848C6B8484738C8C6B848473848C6B8484738C8C6B848473848C6B8484
738C8C9C847BA58C849C847B634A315A42295A4231523929000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000006B637B6B63736B637B
5A6B845A6B84526B845A6B845A6B845A6B845A6B845A6B845A6B845A738C5A6B845A7384
5A6B845A738C5A6B845A738C5A738C5A738C5A73845A738C423139423142423139423139
4231394231424231398C8C848C8C848C8C848C8C7B8C8C7B8C8C7B8C8C7B8C8C7B424A73
424A73424A73424A73424A7B424A7B84634A845A42845A4A845A42845A4A845A42845A42
845A42845A4A7B5A42845A427B5A42845A427B5A427B5A427B5A427B5A427B5A427B5A42
7B5A427B5A426B84846B848C6B848C73848C6B84846B84846B848473848C6B84846B8484
6B8484A58C849C847B6342316342316342315A42295A42315A4231000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006B637B6B6373
5A6B84526B845A6B84526B845A6B84526B845A738C526B845A738C526B845A738C5A6B84
5A738C5A6B845A738C5A6B8463738C5A6B84423142392939423142392939423142423139
4231424231399494848C8C7B9494848C8C7B9494848C8C7B8C8C848C8C7B8C8C84394A73
424A73394A7342527B424A7342527B424A738C634A845A4284634A845A4284634A7B5A42
84634A7B5A4284634A7B5A42845A4A7B5A42845A4A7B5A42845A4A7B5A42845A4A7B5A42
7B5A427B52427B5A427B52427B5A426B8484738C8C6B848473848C6B8484738C8C6B8484
738C8C9C847BA58C84634229634A315A4229634A315A42296342315A4229000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000006B637B
6B637B5A6B84526B845A6B84526B845A6B845A6B845A6B845A6B845A6B845A6B845A7384
5A6B845A73845A6B845A738C5A6B84423139423139423142423139423142423139423142
4231398C8C848C8C848C8C848C8C7B8C8C848C8C7B8C8C7B8C8C7B8C8C7B8C8C7B424A73
424A73424A73424A73424A7B424A73424A7B424A7B84634A845A42845A4A845A42845A4A
845A42845A4A845A42845A4A845A42845A427B5A42845A427B5A42845A427B5A427B5A42
7B5A427B5A427B5A427B5A427B5A427B5A427B5A427B5A426B84846B84846B84846B8484
6B8484A58C84634A31634A316342316342316342316342315A42315A42315A4231000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
736B7B6B63735A6B8452637B5A6B84526B845A6B84526B845A738C526B845A7384526B84
5A738C5A6B845A738C392939423142392939423142392939423142423139423142422939
4A31428C8C848C94848C8C7B9494848C8C7B8C8C848C8C7B8C8C848C8C7B8C8C7B39426B
424A7B424A73424A7B394A7342527B424A7342527B424A738C634A845A4284634A845A42
84634A845A4284634A7B5A4284634A7B5A42845A4A7B5A4284634A7B5A42845A427B5A42
84634A7B5A42845A427B5242845A427B5A427B5A427B52397B5A427B52427B5A42735239
738C8C634A316B4A316342296B4A31634231634A315A4229634A315A42296342315A4229
5A4231000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000073637B526B84526B84526B845A6B84526B845A6B845A6B845A6B84526B845A6B84
5A6B845A7384392939423139423139423142423139423139423139423142423139423142
8C8C848C94848C8C848C8C848C8C848C8C848C8C7B8C8C7B8C8C7B8C8C7B8C8C7B424A73
424A73424A73424A73424A73424A7B42527B424A7B424A7B84634A8C634A845A4284634A
845A4A84634A845A42845A4A845A4284634A845A42845A42845A42845A4A7B5A42845A42
7B5A42845A427B5A427B5A427B5A427B5A427B5A427B5A427B5A427B5A427B5A427B5A42
7B5A425A7B7B6B4A316B4A31634A316B4A31634231634A31634231634A315A42295A4231
5A42315A4231000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000736B7B52637B5A6B8452637B5A6B84526B845A6B84526B845A7384526B84
4231393929394231423929394231423929394231424231394231424231394A31428C8C84
9494848C8C7B9494848C8C7B8C94848C8C7B8C94848C8C7B8C8C84848C7B8C8C84848C7B
424A73394A7342527B424A7342527B424A7342527B424A7B42527B845A428C634A845A42
8C634A845A428C634A845A4284634A845A4284634A845A4284634A7B5A4284634A7B5A42
84634A7B5A4284634A7B5A42845A427B5A42845A427B5A42845A427B5A427B5A427B5A42
A594945A73735A7B7B5A73735A7B736342316B4A31634231634A315A4229634A315A4229
634A315A42295A4231000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000526B84526B84526B84526B845A6B84526B845A6B84526B84423139
3931394231393931394231394231394231424231394231424231394231424231428C9484
8C8C848C8C848C8C848C8C848C8C7B8C8C848C8C7B8C8C848C8C7B8C8C7B8C8C7B8C8C7B
424A73424A73424A73424A7B424A73424A7B424A7B42527B424A7B42527B84634A8C634A
845A4A84634A845A4A84634A845A4284634A845A4284634A845A42845A42845A42845A4A
845A42845A42845A42845A427B5A42845A427B5A427B5A427B5A427B5A427B5A427B5A42
9C94945A73735A73735A73735A73735A73735A73735A7373634231634A31634231634A31
5A42315A42315A42315A4231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000006B637B5A6B8452637B5A6B8452637B5A6B84392939423139392939
4231423929394231423929394231424231394231424229394231424231398C94848C8C84
9494848C8C848C9484848C7B8C94848C8C7B8C8C84848C7B8C8C84848C7B8C8C7B84847B
424A73394A73424A73394A7342527B424A7342527B424A734A527B424A7B42527B845A42
8C634A845A428C634A845A428C634A845A4284634A845A428C634A845A4284634A7B5A42
84634A7B5A4284634A7B5A4284634A7B5A4284634A7B5A4284634A7B5A42845A429C948C
A59C944A4A735A7B7352736B5A7B7352736B5A737352736B5A7B7352736B634A315A4229
634A315A4229634A315A4229634A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000006B637B526B84526B7B526B84393139423139392939423139
3931394231423931394231394231394231424231394231424231428C94848C8C848C8C84
8C8C848C94848C8C848C8C848C8C848C8C848C8C7B8C8C7B8C8C7B8C8C7B848C7B8C8C7B
424A73424A73424A73424A73424A73424A7B424A73424A7B424A7B42527B424A7B42527B
8C634A8C634A84634A84634A84634A8C634A845A4284634A84634A84634A845A4284634A
845A4284634A845A42845A42845A4284634A7B5A42845A427B5A42845A427B5A429C9494
9C9494A59C945273735A73735A73735A737352736B5A73735A73735A737352736B5A736B
52736B634A315A42295A42315A42315A4231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000031313152637B423139392939423139392939423142392939
4231423929394231423929394231424231394231424231398C948C8C8C848C948C8C8C84
8C94848C8C848C94848C8C7B8C8C84848C7B8C8C84848C7B8C8C84848C7B8C8C84848C7B
424A73394A73424A7B394A7342527B424A7342527B424A7342527B424A7B4A5284424A7B
4A52848463428C634A8463428C634A845A428C634A845A428C634A845A428C634A845A42
84634A845A4284634A845A4284634A7B5A4284634A7B5A4284634A7B5A42A59C949C948C
A59C949C948C5A737352736B5A7373526B6B5A7373526B6B5A7373526B6B5A7373526B6B
5A736B526B6B5A736B5A4229634A315A4229634A31000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000312931292931393139392939393139393139423139393139423142
3931394231424231394231424231394231424231424231428C8C848C94848C8C848C8C84
8C8C848C8C848C8C848C8C848C8C7B8C8C848C8C7B8C8C84848C7B8C8C7B848C7B8C8C7B
394A73424A73424A73424A73424A73424A7B424A7B42527B424A7B42527B424A7B42527B
42527B4252848C634A8C634A84634A8C634A84634A8C634A84634A84634A84634A84634A
845A4284634A845A4284634A845A4284634A845A4284634A845A42845A429C94949C9494
9C9494A5949452736B5A737352736B52736B52736B52736B52736B52736B52736B52736B
526B6B52736B526B6B52736B526B6B52736B5A42315A42315A4229000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000312931292929425A73392939423139392939423142392939423142392939
4231423931394231423929394231424231394231428C8C8494948C8C8C848C9484848C84
8C94848C8C848C8C84848C7B8C9484848C7B8C8C84848C7B8C8C84848C7B8C8C7B84847B
8C8C7B394A73424A73394A7342527B424A7342527B394A7342527B424A7B42527B424A7B
4A5284424A7B4A52848463428C634A84634A8C634A845A428C634A8463428C634A845A42
8C634A845A4284634A845A428C634A845A4284634A7B5A4284634A845A42A59C949C948C
A59C949C94945A7373526B6B5A7373526B6B5A7373526B6B5A7373526B6B52736B526B63
5A736B526B6B52736B526B6352736B526B6352736B4A6B63634A315A4229000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000312931292931425A73425A6B393139392939393139393139423142393139423139
4231394231424231394231424231424231428C8C848C94848C8C848C94848C8C848C8C84
8C8C848C8C848C8C848C8C848C8C848C8C84848C7B8C8C7B848C7B8C8C7B848C7B848C7B
848C7B424A73394A73424A73424A73424A7B424A73424A7B424A7B42527B424A7B42527B
42527B42528442527B8C634A8C634A8C634A84634A8C634A8C634A8C634A84634284634A
84634A8C634A84634284634A84634284634A845A428463428463429C94949C94949C9494
9C94949C9C94526B6B52736B526B6B52736B526B6B526B6B526B6B52736B526B6B526B6B
526B6B52736B526B63526B6B526B6B526B6B526B63526B63526B63526B6B526B63000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
312931292929425A7339526B425A73392939423142392939423142392939423142392939
4231423929394231424231394231428C8C848C948C8C8C848C948C848C848C948C848C84
8C9484848C7B8C9484848C7B8C8C84848C7B8C8C84848C7B8C8C8484847B8C8C8484847B
8C8C7B39426B424A73394A73424A7B394A7342527B424A7342527B424A7B42527B424A7B
425284424A7B4A528442527B8C6B4A84634A8C6B4A84634A8C634A8463428C634A846342
8C634A845A428C634A845A428C634A845A428C634A845A4284634A9C9494A59C949C9494
A59C949C94944A5273526B6B52736B526B6B52736B526B6B52736B526B6352736B526B63
52736B4A6B6352736B4A6B63526B6B4A6B63526B6B4A6B63000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
292931425A6B42526B425A6B42526B393139393139423142393139423142423139423142
4231394231424231424231424231428C948C8C8C848C948C8C8C848C94848C8C848C8C84
8C8C848C8C848C8C848C8C84848C848C8C84848C7B8C8C84848C7B848C7B848C7B848C7B
848C7B424A73424A73424A73424A73424A73424A73424A7B424A7B42527B424A7B42527B
424A7B42527B42527B4252844252848C634A8C634A8C634A8C634A8C634A8C634A8C634A
84634A8C634A84634A8C634A84634284634A84634284634A9C94949C94949C94949C9C94
9C94949C9C949C9494526B6B526B6B526B6B526B6B526B6B526B6B526B6B526B6B526B6B
526B63526B6B526B63526B6B526B63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292929
425A6B39526B425A6B39526B425A6B392939423142392939423142393139423142392939
42314242313942314242313994948C8C8C848C948C848C848C948C848C848C9484848C84
8C948C848C848C8C84848C7B8C9484848C7B8C8C8484847B8C8C84848C7B8C8C7B84847B
8C8C7B394A73424A7339427342527B394A7342527B394A7342527B424A7B42527B424A7B
4A5284424A7B4A5284424A7B4A52844252848C6B4A8463428C6B4A8C634A8C634A846342
8C6B4A8463428C634A8463428C634A8463428C634A845A429C9C9C9C94949C9C949C9494
A59C9C9C94949C9C944A6B6352736B4A6B63526B6B4A6B63526B6B4A6B63526B6B4A6363
526B6B4A6B63000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000029292939526B
39526B42526B39526B42526B39526B423142392939423139423142423142423139423142
4231424231424231428C948C8C948C8C948C8C8C848C94848C8C848C9484848C848C8C84
8C8C848C8C84848C848C8C84848C848C8C84848C7B848C7B848C7B848C7B848C7B848C7B
848C7B424A73394A73424A73424A73424A7B424A73424A7B424A7B42527B424A7B42527B
42527B42528442527B4252844252844A52844252848C634A8C634A8C6B4A8C634A8C634A
8C634A8C634A84634A8C634A8C634A8C634A8463429C94949C94949C9C949C94949C9494
9C94949C9C949C9494526B6B526B6B526B6B4A6B63526B6B526B63526B6B4A6B63000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000029292942526B395263
42526B395263425A6B395263425A6B392939423142392939423142392939423142423139
4231424231394231428C8C848C948C848C848C948C848C848C948C848C848C948C848C84
8C9484848C848C9484848C7B8C8C84848C7B8C8C84848C7B8C8C8484847B8C8C8484847B
848C7B848473424A73394A73424A7B394A7342527B394A7342527B394A7342527B424A7B
425284424A7B4A5284424A7B4A528442527B4A5284425284946B4A8C634A8C6B4A8C634A
8C6B4A8463428C6B4A8463428C6B4A8463428C6B4A9494949C9C9C9494949C9C9C9C9494
9C9C9C9C94949C9C944A6B63526B6B4A6363526B6B4A6363000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000292929395263395263395263
39526339526339526342526B395263423142393139423142423139423142423142423142
4231424231428C948C8C948C8C8C8C8C948C8C8C848C948C8C8C848C8C84848C848C8C84
848C848C8C84848C848C8C84848C7B848C84848C7B848C84848C7B848C7B848C7B848C7B
848C7B848C7B394A73424A73394A73424A73424A73424A7B424A73424A7B424A7B42527B
424A7B42527B42527B42528442527B4252844252844252844252848C6B4A8C634A8C6B4A
8C634A8C634A8C634A8C634A8C634A8C634A8C634A9C949C9C94949C94949C94949C9C94
9C94949C9C949C9C94526B6B4A6B63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000425263394A63395263394A63
42526B39526342526B394A6342526B393139423142392939423142423139423142392939
4A31428C8C8C8C948C848C848C948C848C848C948C848C848C948C848C848C9484848C84
8C9484848C848C8C84848C7B8C9484848C7B848C8484847B8C8C84848C7B848C7B84847B
848C7B84847B424A7339426B424A7B394A73424A7B394A7342527B424A7342527B394A7B
425284424A7B425284424A7B4A528442527B4A52844252844A5A8C8C634A946B4A8C634A
946B528C634A8C6B4A8C63428C6B4A8C634A9C9C9C9494949C9C9C9494949C9C9C949494
9C9C9C9C94949C9C944A4A6B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000395263395263395263394A63395263
395263395263395263395263395263423142393139423142423142423142423142423142
8C948C8C948C8C8C8C8C948C8C948C8C948C848C848C8C848C8C848C948C848C84848C84
848C848C8C84848C84848C84848C84848C84848C7B848C7B848C7B848C7B848C7B848C7B
848C7B848C7B394A73424A73424A73424A73394A73424A73424A7B42527B424A7B42527B
42527B42527B424A7B42527B4252844252844252844252844252844A52848C634A8C6B4A
8C6B4A8C6B4A8C634A8C6B4A8C634A8C6B4A9494949C94949C94949C9C9C9494949C9C94
9C9C949C9C949C94949C9C94000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000394A5A
395263394A5A425263394A63425263392939423142393139423142393139423142423142
8C948C848C8C8C948C848C848C948C848C848C948C848C848C948C848C848C948C848C84
8C9484848C848C9484848C7B8C8C84848C7B8C8C84848C7B848C8484847B848C8484847B
848C7B7B8473424A73394A73424A73394A73424A7B394A7342527B394A7342527B424A7B
42527B424A7B425284424A7B4A5284424A7B4A52844252844A52844252844A5A8C8C634A
946B528C634A946B528C634A946B4A9494949C9C9C9494949C9C9C9494949C9C9C949494
9C9C9C9494949C9C9C949494000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004231424231424231424231424231424231428C948C
8C948C8C948C8C948C8C948C848C8C8C948C848C848C948C848C848C8C84848C848C8C84
848C84848C84848C84848C84848C84848C84848C7B848C84848C7B848C7B848C7B848C7B
848C7B848C7B394A73424A73394A73424A73424A73424A7B424A73424A7B424A7B42527B
424A7B42527B42527B42528442527B4252844252844252844252844252844252844A528C
8C6B4A946B4A8C6B4A946B4A8C634A94949C9494949C9C9C9494949C9C9C9494949C9C9C
9494949C9C949C9C949C9C94000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003931394231423929394231424231428C9494848C8C
8C9494848C8C8C948C848C848C948C848C848C948C848C848C948C848C848C9484848C84
8C948C848C84848C84848C7B8C8C84848C7B848C847B847B848C84848C7B848C7B7B847B
848C7B7B847B848C7B39426B424A73394A73424A7B394A7342527B394A7342527B394A73
42527B424A7B425284424A7B4A528442527B425284424A7B4A5A844252844A5A8C425284
4A5A8C8C634A946B4A8C634A9C9C9C9494949C9C9C9494949C9C9C9494949C9C9C949494
9C9C9C9494949C9C9C949494000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004231424231424231424231428C94948C948C8C948C
8C948C8C948C848C8C8C948C8C948C8C948C848C84848C8C848C848C9484848C84848C84
848C84848C84848C84848C84848C84848C84848C7B848C7B848C7B848C7B848C7B848C7B
848C7B848C7B7B847B394A73394A73424A73394A73424A73424A73424A7B424A73424A7B
424A7B42527B424A7B42527B42527B42528442527B4252844252844A5284425284425284
4252844A5A8C8C6B4A946B4A94949C9C9C9C949494949C9C94949C9C9C9C949494949C94
949C949C9C9C949C949C9C94000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000393139423142423142423142848C8C8C9494848C8C
8C948C848C8C8C948C848C848C948C848C848C948C848C848C948C848C848C948C848C84
8C9484848C848C9484848C7B848C84848C7B848C847B8C7B848C847B847B848C847B847B
848C7B7B8473848C7B39426B424A73394A73424A7B394A73424A7B394A7342527B394A73
42527B424A7B425284424A7B425284424A7B4A528442527B4A52844252844A5A8C425284
4A5A8C4252844A5A8C8C6B4A9C9C9C9494949C9C9C9494949C9C9C9494949C9C9C949494
9C9C9C9494949C9C9C949C944A5273000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004231424231424231428C948C8C948C8C948C8C948C
848C8C8C948C848C8C8C948C848C8C8C948C848C8484948C848C84848C84848C84848C84
848C84848C84848C84848C84848C84848C84848C7B848C84848C7B848C7B848C7B848C7B
7B847B848C7B7B847B424A73394A73424A73394A73424A73394A73424A7B424A7B42527B
424A7B42527B42527B42527B42527B4252844252844252844252844252844252844A5284
4252844A528C94949C94949C94949C949C9C94949C949C9C949494949C9C949C94949C9C
949C949C9C9C949C949C9C9C4A4A73000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000423142423142848C8C8C9494848C8C8C9494848C8C
8C9494848C8C8C948C848C848C948C848C848C948C848C848C948C848C8484948C848C84
8C948C848C84848C847B8C7B848C84848C7B848C847B847B848C847B847B848C7B7B847B
848C7B7B847B848C7B39426B424A73394A73424A73394A7342527B394A7342527B394A73
42527B424A7B42527B394A7B42528442527B425284424A7B4A52844252844A52848C949C
949CA594949C949C9C9494949C9C9C94949C949C9C9494949C9C9C9494949C9C9C949494
9C9C9C949C949C9C9C9494949C9C9C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004231424231428C94948C948C8C949484948C8C948C
84948C8C948C848C8C84948C848C8C8C948C848C84848C84848C8484948C848C84848C84
848C84848C84848C84848C84848C84848C84848C7B848C7B848C7B848C7B7B8C7B848C7B
7B8C7B848C7B7B847B394A73394A73424A73394A73424A73424A73424A7B394A73424A7B
424A7B42527B424A7B42527B42527B42528442527B94949C94949C94949C94949C94949C
94949C949C9C94949C94949C94949C949C9C94949C949C9C949C9C949C9C949C94949C9C
949C9C949C9C949C94949C94949C94000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004231428C949484948C8C9494848C8C8C9494848C8C
8C948C848C8C8C948C848C848C948C848C848C948C848C8484948C848C8484948C848C84
848C847B8C84848C847B8C7B848C847B8C7B848C847B8C7B848C847B847B848C847B847B
848C7B7B847B848C7B7B8473424A7339426B424A73394A73424A7B394A7342527B394A73
42527B394A7342527B8C949C9494A58C949C9494A58C949C9494A58C949C949CA58C949C
949C9C8C949C949C9C94949C949C9C949494949C9C949494949C9C9494949C9C9C949494
9C9C9C9494949C9C9C949C949C9C9C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000008C94948C949484948C8C948C84948C8C948C
848C8C84948C848C8C84948C848C8C84948C848C84848C8C848C84848C84848C84848C84
848C84848C84848C84848C84848C84848C84848C7B848C847B8C7B848C7B7B8C7B848C7B
7B847B7B8C7B7B847B7B8C7B394A73394A73394A73424A73394A73424A73424A738C94A5
8C94A58C94A58C949C9494A58C949C9494A58C949C9494A58C949C94949C94949C94949C
94949C94949C94949C949C9C94949C949C9C94949C949C9C949C9C949C9C949C9C949C9C
949C94949C9C949C94949C9C949C94000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000073738C6B6B847373946B738C73738C6B738C
737B9473738C737B9473738C7B7B9473738C7B7B9473738C7B7B9473738C7B7B9473738C
7B7B94848C84848C847B8C7B848C847B8C7B848C847B847B848C847B847B848C7B7B847B
848C7B7B847B7B8C7B7B8473424A73394A73424A738C8C9C9494A58C8CA58C94A58C8C9C
9494A58C949C9494A58C8C9C9494A58C949C9494A58C949C949CA58C949C949CA58C949C
949CA58C949C949C9C8C949C949CA594949C949C9C8C9494949C9C94949C949C9C949494
9C9C9C949C94949C9C949C949CA59C4A4A73000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C73738C73738C73738C73738C73738C
73738C73739473738C73738C73738C737B9473738C737B94737B947B7B947373947B7B94
7B7B947B7B94737B947B7B947B7B947B7B947B7B947B7B947B7B947B849C7B7B947B849C
7B7B9C7B849C7B7B947B849C8C8CA58C94A58C8CA58C94A58C94A58C94A58C8CA58C94A5
8C94A58C94A58C949C8C94A58C94A59494A58C949C8C949C8C949C9494A58C949C94949C
94949C949C9C8C949C949C9C949C9C949C9C94949C949C9C949C9C949C9C949C9C949C9C
949C9C949C9C949C94949C9C949C9C4A4A73000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000073738C6B6B8C73738C6B738C7373946B738C
7373946B738C737B9473738C737B9473738C7B7B9473738C7B7B9473738C7B7B94737394
7B7B94737B947B7B9C737B947B7B9C737B947B849C7B7B947B849C7B7B9484849C7B7B94
84849C7B7B9431396331315A737B6B6B7B63737B6B6B73638C94A58C8C9C8C94A58C8C9C
8C94A58C8C9C9494A58C949C9494A58C949C949CA58C949C949CA58C949C949CA58C949C
949CA58C949C949CA58C949C949C9C8C949C949C9C8C949C949C9C8C9494949C9C949C94
949C9C949C94949C9C949C94949C9C949C94000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C73738C6B738C73738C73738C73738C
73738C73738C73738C73739473738C737B94737394737B94737394737B94737B947B7B94
737B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B947B7B9C7B7B947B849C
31315A31395A31395A3139633139636B7B6B6B7B636B7B6B6B73636B7B636373638C94A5
8C94A58C94A58C94A58C94A58C949C8C94A58C949C8C94A58C949C8C94A58C949C949CA5
8C949C949C9C8C949C949C9C8C949C949C9C949C9C949C9C949C9C949C9C949C9C949C9C
949C9C949C9C949C9C949C9C949C94949C9C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000073738C6B6B8C7373946B738C73738C6B738C
737B9473738C737B9473738C7B7B9473738C737B9473738C7B7B947373947B7B9473738C
7B7B94737B947B7B94737B947B849C737B947B7B9C737B947B849C7B7B94313952293152
31395A29315A31395A29315A3139636B7B636B7B6B6373636B7B6B6373636B7B6363735A
6B736363735A8C94A58C949C949CA58C949C8C94A58C949C949CA58C949C949CA58C949C
949CA58C949C949CA58C949C949CA58C949C949C9C8C949C949C9C8C9C9C949C9C8C9C94
949C9C949C9C949C9C8C9C9494A59C949C94000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C73738C73738C73738C6B738C73738C
73738C73739473738C737394737394737B9473738C737B94737B947B7B94737B94737B94
737B947B7B94737B947B7B947B7B947B7B9C7B7B947B7B94000000313152293152313152
31315A31395A31315A31395A31395A6B7B6B6B73636B7B636B73636B7B63637363637363
637363637363636B5A63735A636B5A8C94A58C949C8C94A58C94A58C9CA58C949C8C949C
8C949C949CA58C949C8C9C9C8C9C9C949C9C8C9C9C949C9C949C9C949C9C8C9C9C949C9C
949C9C949C9C949C9C949C9C949C9C94A59C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000073738C6B738C73738C6B738C7373946B738C
737B946B738C737B9473738C737B9473738C7B7B9473738C7B7B9473738C7B7B94737394
7B7B947373947B7B9C737B947B7B9C737B94000000000000000000000000313952293152
31395A29315231395A29315A3139636373636B7B6B6373636B7B6B6373636B7B6363735A
637363636B5A6373635A6B5A63735A5A6B5A0000000000008C9CA58C949C949CA58C949C
949CA58C949C949CA58C949C949CA58C949C949CA58C9C9C949C9C8C9C9C949C9C8C9C9C
949C9C8C9C9494A59C8C9C9494A59C8C9C944A527B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C73738C6B738C73738C73738C73738C
73738C73739473738C737B9473738C737B94737394737B94737394737B94737B947B7B94
737B947B7B94737B947B7B94000000000000000000000000000000000000000000313152
29315231315A31315A31395A31315A6B7B6B6B73636B7B63637363637363637363637363
63735A637363636B5A63735A5A6B5A636B5A0000000000000000000000000000008C9CA5
8C949C8C9CA58C9C9C8C9CA58C9C9C8C9C9C8C9C9C949C9C8C9C9C949C9C8C9C9C949C9C
8C9C9C949C9C949C9C949C9C949C9C949C9C949C9C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000073738C6B6B8C7373946B738C7373946B738C
737B946B738C737B946B738C737B9473738C737B9473738C7B7B947373947B7B9473738C
7B7B9C737B94000000000000000000000000000000000000000000000000000000293152
31395A29315231395A29315231396331315A6B7B6B6373636B7B6B637363637363636B5A
637363636B5A6373635A6B5A63735A000000000000000000000000000000000000000000
0000000000008C9CA58C949C949CA58C9C9C949CA58C949C949CA58C9C9C949C9C8C9C9C
94A59C8C9C9C94A59C8C9C9494A59C8C9C9494A59C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C6B738C6B738C73738C6B738C73738C
73738C73739473738C737394737394737B94737394737B94737B94737B94737B94737B94
000000000000000000000000000000000000000000000000000000000000000000000000
29315231395229315231315A31395A31395A6373636B7363637363637363637363637363
63735A6373635A6B5A636B5A5A6B5A000000000000000000000000000000000000000000
000000000000000000000000000000949CA58C9C9C8C9C9C8C9C9C949C9C8C9C9C8C9C9C
8C9C9C949C9C8C9C9C949C9C949C9C94A59C8C9C9C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000073738C6B738C7373946B738C7373946B738C
7373946B738C737B946B738C737B9473738C737B9473738C737B9473738C000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029315231395A29315231395A29315A6B7B6B6373636B7B6363735A637363636B5A
6373635A6B5A6373635A6B5A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000949CA58C9C9C94A5A58C9C9C
94A59C8C9C9C94A59C8C9C9C94A59C8C9C9494A59C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C73738C6B738C73738C6B738C737394
6B738C73739473738C737394737394737B94737394737B94000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000031315229315231395A31315A31395A6373636B7363637363637363637363637363
636B5A63735A5A6B5A636B5A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C9C9C949C9C8C9C9C94A59C8C9C9C94A59C8C9C9C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000073738C6B6B8C7373946B738C7373946B738C
737B946B738C737B946B738C737B9473738C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000031395229315231395A29315A31395A6373636B7B6363735A6373635A6B5A
6373635A6B5A63735A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008C9C9C94A59C8C9C9C94A59C8C9C94000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C6B738C6B738C7373946B738C73738C
73738C73739473738C737394000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031315231315231395A29315A63736363736363736363735A63735A
636B5A63735A5A6B5A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000008C9C9C94A59C000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007373946B738C7373946B738C7373946B738C
737B946B738C000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000029315231395A29315231395A6373636373635A6B5A6373635A6B5A
6373635A6B5A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C6B738C6B738C6B738C6B738C737394
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000029315231315A31315A63736363736363736363735A637363
5A6B5A5A6B5A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C6B6B8C7373946B738C000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000029315231395A2931526373635A6B5A6373635A6B5A
637363000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000006B738C6B738C000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000031395229315231395A6373636373635A6B5A5A6B5A
5A6B5A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000031395A2931526373635A6B5A6373635A6B5A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000029315231395A5A6B5A6373635A6B5A5A6B5A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000029315231395A5A6B5A637363000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003131526373635A6B5A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031395A5A6B5A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005A6B5A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[54]:=",
  ImageSize->{128, 128},
  AnimationDisplayTime->0.169,
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000200000P0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj2LW
9`00?g=cLda<C240<410Y/30`9T0IPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm?ooc020P800082006IVc3<cIP2009VIod:J
Yb8R8_mV<`00P00g??mVc5J0Zk75fBa1EkV>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7De3eDE:b/[:BT
Y>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEeMJ_0eLC4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7i/;2`W9bLP00
o`1o001oO`1oo`3o003oO`3oogl007l0Ogl0ogmo07moOgmoogoo07ooOgooool00?l0Ool0oomo0?mo
Oomooooo0?ooOoooo`0002XZ:UEEEGmoOjZZZ]GEeOoooh0700201`00P0L0080700031gT010L000<7
N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7=`000j00000o00@7
00031cH00X400j00000n00@700031c4000Vi0000X9ZPVZ00?`041`000`L`000;PYnPW`21X86PPJ00
?P041`000`L`000<PKVO^InPVZ2JX9ZP?@041`000`L_000<6H8IX9nPWj21X86P0X4l00@700031bl0
00j2PH:O^IniWj2JX9ZPVS`010L000<7;P00488IPQVPWj2OX9nPPJ2JX9Xk00@700031bh00161PX62
WkVO^IniWkVJX9ZPV`0j00@700031bd0018IPQV26J2OX9nPWj2OX9ZPVZ0j00@700031bd001>2PH:1
PYniWkVO^IniWkVJX9^P03T010L000<7;@0051V26H8IX9nPWj2OX9nPWj2JX9ZP>0041`000`L/000C
PH:1PX62WkVO^IniWkVO^IniV`03X3L010L000<7;0005h8IPQV26J2OX9nPWj2OX9nPWj2JX9ZP03H0
10L000<7:`02P@0BPX62PH:1^IniWkVO^IniWj2O1J0e00@700031b`001V26H8IPQV2Wj2OX9nPWj2O
X9nPWj2JX9ZP010000>1W`008@041`000`L[000HPX62PH:1PX6iWkVO^IniWkVO^InPWj2O0j0=00:1
00>OP@008@041`000`L[000L6H8IPQV26H:OX9nPWj2OX9nPWj2OX9nPVZ2JX0/00X400in1W`0R00@7
00031bX001j1PX62PH:1PX6iWkVO^IniWkVO^IniWkVO^InPWh4700:100JOPIn1Wh4R00@700031bT0
020IPQV26H8IPX62Wj2OX9nPWj2OX9nPVZ2JX9ZPVZ2JX0D00X401in1Wh6OPIl08P041`000`LY000Q
PX62PH:1PX62PKVP^J2iX;VP^IniWkVO^IniWkVOX9nP008000]kPIZOVYnJWiZOP@0R00@700031b00
00>1Nh401P008aV26H8IPWb2PH:OX9nPWj2OX9nPVZ2JX9ZPVZ2JX9ZPVZ0000:100VOPIn1Wh6OPIl0
8P041`000`LQ000fPG^1Nh5k0862PH:1PX62PH:1PZ2iX;VP^J2iWkVO^IniWkVOX9nPWj2OX9ZOVYnJ
WiZOVYn18P041`000`LR000ePG^1Nh5kPQV26H9lPQV2PH:1X9nPWj2OX9nPWj2JX9ZPVZ2JX9ZPVZ1k
PG^1Nh5kPX6OPIl08P041`000`LQ000ZPAV1Nh5kPX62PH:1PX62PH:1PZ2iX;VP^J2iWkVO^IniWkVO
^InPWj1l38400iZ1PP0O00@700031b8003f1085kPAV26H8IPQV2PH:1PX6PWj2OX9nPWj2OX9nPVZ2J
X9ZPVZ1kPG^1Nh5kPG^1Nh5kPG^1Nh5kPG^P01X010L000<78@00:84IPGbJPH:1PX62PH:1PX62PH:P
^J2iX;VP^IniWkVO^IniWj2OX7`<P@0=VX6JPIZ1VZ2VX;VP^@0G00@700031b8003R1085kPQV26H8I
PQV26H:1PX6PWj2OX9nPWj2OX9ZPWj2JX9ZPNh5kPG^1Nh5kPG^1Nh5kPG^1N`FP00VOX9ZPWj2JX9l0
4@041`000`LR000UPIZ1VX62PH:1PX62PH:1PX62X;VP^J2iX;VP^IniWkVO^IniO00<P@0KVX6JPIZ1
YZ2VX:JP^J2iX;VP^J2iX;VP^J2i00/010L000<77P00>IYkVX2JP9Z0VW^26H8IPQV26H:1PX6PWj2O
X9nPWj2OX9nPWj2JX7^1Nh5kPG^1Nh5kPG^1Nh5kP@08X00AVZ2OX9ZPWj2OX9nPWj2OX9l02@041`00
0`LK000ZVX6JPIZ1VX6JPIZ1VX62PH:1PX62PH:1PZ2iX;VP^J2iWkVO^IniWkVO2X4089Z1VX6JPJJP
YZ2VX:JPYZ2iX;VP^J2iX;VP^J2iX;VP2P041`000`LI000jP9Z0VX2JP9Z0VX2JP9Z0VX226H8IPQV2
PH:1PYnPWj2OX9nPWj2OX9nPVX5kPG^1Nh5kPG^1Nh5kP@bP012OX9niWj2O^InPWkVO^Ini2P041`00
0`LE0003VX6J00:J02V1VX6JPIZ1VX6JPIZ1VX6JPH:1PX62PH:1PX6iX;VP^J2iX;VO^IniW`06P@0U
VX61PIZ1VX6VX:JPYZ2VX:JPYZ2VX;VP^J2iX;VP^J2iX;VP^@0;00@700031a8003fJP9Z0VX2JP9Z0
VX2JP9Z0VX2JP9Z0VX2J6H8IPVZ2JX:1PYnPWj2OX9nPWj2OX9n1Nh5kPG^1Nh5kPG^1012P00jO^Ini
WkVO^IniWkVO^@`010L000<73P0<VP03PIZJ00FJ00F1VX6JP@03VP0GPIZ1PX62PH:1PX6iX;VP^J2i
X;VP^Il01X401YZ1VX6JP@:P01jVX:JPYZ2VX:JPYZ2VX:JP^J2iX;VP^J2iX;VP^J0<00@700031`/0
0461VX2JPIZ0VX6JP9Z1VX2JP9Z0VX2JP9Z0VX2JP9Z0VX226H8IPX62PH:OX9nPWj2OX9nPWh5kPG^1
Nh5kPG^1N`0CX00=WkVO^IniWkVO^IniW`0=00@700031`/000N1Qj61X86P01bJ0161PX62PH:1^J2i
X;VP^J2iVP0:P@:P022VX:JPYZ2VX:JPYZ2VX:JPYZ2VX;VP^J2iX;VP^J2iX0h010L000<73000>j22
X8:PPZ22X82JP9Z0VX2JP9Z0VX2JP9Z0VX2JP9Z0VX2J6H8IPQV2Wj2OX9nPWj1kPG^1Nh5kPG^101NP
00>iWj001Z000inPW`0?00@700031``000f2XHNQQj67XHNQQj6701JJ00j2PH:1PX6nX;VP^J2iX0R1
02BVX:6PYZ2QX:JPYZ2VX:JPYZ2VX:JPYZ2VX;VP^J2iX;VP^J0@00@700031`X0036PPJ22X8:PPZ22
X8:PPZ22X8:JP9Z0VX2JP9Z0VX2JP9Z0VX2JP88IPQV26J2OX9nP00:O00IkPG^1Nh4KX005^J2PX;T0
0j000kVP000@00@700031`P001QlX8:PPZ27XHNQQj67XHNQQj67XHNQQj4AVP0;PX62PH:P^J2iX;T0
1H409j6PXJ2QX:6PXJ2QX:JPYZ2VX:JPYZ2VX:JPYZ2iX;VP^J2iX;VP^@0A00@700031`T003J1X86P
PJ22X8:PPZ22X8:PPZ22X8:PPZ22X8:JP9Z0VX2JP9Z0VX2JP88IPQVPWj2OX7^1Nh42X003Qj2P00VP
00F7X8NPQ`0<X009^J2iX;VP^J2i018010L000<72P007X:PPZ22XHNQQj67XHNQQj67XHNQQj67XHNQ
Qj67X@^J00N2PH:P^J2i00:102YlXJ2QX:6PXJ2QX:6PXJ2QX:JPYZ2VX:JPYZ2VX:JPYZ2iX;VP^J2i
X;TC00@700031`X003jPPJ21X86PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8NPQj21VX2JP9Z0VX226J2O
X7^1Nj27X8NPQj27X8NPQ`>P00n7X8NPQj27X8NPQj27X8L01:002;VP^J2iX;VP4`041`000`L;000S
X8:PPZ27XHNQQj67XHNQQj67XHNQQj67XHNQQj67XHNQQj401iX0<8:P_X6JX:JPXJ2QX:6PXJ2QX:6P
XJ2QX:JPYZ2VX:JPYZ2VX:JPYZ2iX;VP^J2iX1@010L000<73000D:21X86PPZ22X8:PPZ22X8:PPZ22
X8:PPZ22X8:PQj27X8NPQj27VX2J6J1kX8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27
0Z001[VP^J2iX1D010L000<73@00Ej22X8:PQj67XHNQQj67XHNPQj67XHNQQj67XHNQQj67XHNQQj67
XHNQPJJPYZ2VX:6PXJ2QX:6PXJ2QX:6PXJ2VX:JPYZ2VX:JPYZ2VX:JP^J2iX;VP^@0E00@700031`h0
04jPPJ21X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PQj22X8NPQj1kX9ZOQj27X8NPQj27X8NPQj27
X8NPQj27X8NPQj27X8NPQj27X8L2X005^J2iX;T05P041`000`L?000TX8:PPZ67XHNQQj67X8NPQj27
XHNQQj67XHNQQj67XHNQQj670X40;KjPPYZOVYnPXJ2QX:6PXJ2QX:6PXJ2QX:JPYZ2VX:JPYZ2VX:JP
YZ2iX;VP^@0G00@700031`l002F1X86PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PPX5k00>P
02F1PYZOVYnJWhNPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj2700:P00>iX;T060041`000`L@000Q
PZ22X8NQQj67XHNPQj27X8NPQj67X8NQQj67XHNQQj5l00>100>P_Z000X80:8NJWiZOVYnJWj2QX:6P
XJ2QX:6PYZ2VX:JPYZ2VX:JPYZ2VX;VP_Z0H00@700031a4002:1X86PPZ22X8:PPZ22X8:PPZ22X8:P
PZ22X8:PPZ1kPG^11:009H62PH:JWiZOVYnJWiZOQj27X8NPQj27X8NPQj27X8NPQj27X8L00Z000kVP
000H00@700031a8001f2X9^QQj67XHNPQj27X8NPQj27X8NPQj67XHN1O003P@05X;jP_Z0018808inJ
WiZOVYnJWiZOVYnPXJ2QX:JPYZ2VX:JPYZ2VX:JPYZ2n00:P6@041`000`LD000GX867PXN2X8:PPZ22
X8:PPZ22X8:PPZ000X4017^1Nh46X00ZPH:1PX6OVYnJWiZOVYnJWiZOVYnJX86PQj27X8NPQj27X8NP
QkVP^H6J60041`000`LD000FVj2KX8NPQj27X8NQQj27XHNQQj67X@J100RP_Z2nX;jP_PF201BOVYnJ
WiZOVYnJWiZOVYnJWiZOVP:P00>QX:001j001[V1VX2JPAH010L000<75@006X:PPXN2X8:PPZ22X8:P
PZ22X8:10840PGb12J000h:1PP02PP0BPInJWiZOVYnJWiZOVYnJWiZO1IX04:27X8NPQj27^J2JP9Z0
VX0E00@700031aH0016KX8NPQj27X8NPQj27X8NQQ`07P@0:X;jP_Z2nX;jP_PN201ROVYnJWiZOVYnJ
WiZOVYnJWiZOVYnJWiX6X008VX6JPIZ1VX4D00@700031aL001N2X8:PPZ22X8:PPZ22X0210840PG^1
O00:X0F201B1PX6OVYnJWiZOVYnJWiZOVYnJW`>J00>OVYX019X038NJP9Z0VX2JP9Z0VQ8010L000<7
60003I^PQj27X8NPQj27X7`028402kjP_Z2nX;jP_Z2n00F20227PX:J^9ZOVYnJWiZOVYnJWiZOVYnJ
WiZOVYnJWiZOQ`>J00R1VX6JP9Z1VQ4010L000<76@004h:PPZ22X8:PPX40P@21085kPG/03J001H62
PX:100>201ZJWiZOVYnJWiZOVYnJWiZOVYnJWiZOVYnJW`:P00b7X86JP9Z0VX2JP9X@00@700031aT0
00R1Qj27X8NPQ`^100fP_Z2nX;jP_Z2nX;jP00J202^7PXNJ^9ZOVYnJWiZOVYnJWiZOVYnJWiZOVYnJ
_hNPQj27X86JPIZ1VX6J00l010L000<76P004:22X8:P0840P@21085kPG/2P@fP00N1PX62PH:100>2
02RJWiZOVYnJWiZOVYnJWiZOVYnJWiZOV[VPPXJ7PJ21X86JP9Z0VX2J3P041`000`LJ0004Vj27X0:1
00=lPH4028403[jP_Z2nX;jP_Z2nX:FP28806hN2QiZhVYnJ^9ZOVYnJWiZOVYnJWiZOVYnP_`02X00>
Qj27X8NPQj27VX2JPIX=00@700031aX00121PX40P@210840P@21Nh5k3j001h62PX:1PX401H80:9ZO
VYnJWiZOVYnJWiZOVYnJWiZOX;VP^H67PJ21X86PPJ21X82JP9X<00@700031aT000B16H5l3H403j2n
X;jP_Z2nX;jPYJ2UX006PP0MQh:7PXN2QiZhV[RJ^9ZOVYnJWiZOVYnJWiZoX;l00Z004HNPQj27X8NP
Qj27X8NPPIZ000X010L000<76P003hL0P@210840P@21085kP@0AX006PX62PH9Z1h80:IZOVYnJWiZO
VYnJWiZOVYnJWj2iX;V1Qh6PPJ21Qh6PPHN1Qh67PIZ000T010L000<760001Gb2PX5l00^101:P_Z2n
X;jP_Z2nX:FPYJ2UX:D7PP0KQh:7PXN2^9ZhV[RJ^9ZOVYnJWiZOVYnP_j2o00:P01>7X8NPQj27X8NP
QZ27X8JPQZ1Z00P010L000<760003h5ZQ`210840P@21085kP@0CX006PX62JX:128809YnJWiZOVYnJ
WiZOVYnJWiZiX;VP^H67PHN1Qh67PHN1Qh67PHN12`041`000`LG0004PH:7PPb101>oX;nP_Z2nX;jP
YJ2UX:FPYJ2U00N202Z7PXN2Qh:7P[RJ^9ZhVYnJWiZOVYnJ_j2oX;nP_hNPQj27X8NPQZ26X8H>00@7
00031aH00X403X9ZQ`210840P@21Nh5k5:001X:1PVZ2JPZ201jOVYnJWiZOVYnJWiZOX;VP^J2iX8N1
Qh67PHN1Qh4A00@700031aD00X401XN1Qh:1O0R101FoX;nP_Z2nX;jP_Z2nX;jP_Z2nX:D03h8078N2
X9ZhV[RJ^9ZhV[RP_j2oX;nP_hJPQXMZPFXD00@700031a@00122JX9ZPVZ2O85lPGb1O85l5j000fZ2
P@0?PP0BVYnJWiZOVYnJ^Z2iX;jP_Z276P041`000`LC0006Qh67PHN10X88P@0H_j2UX:FPYJ2nX;jP
_Z2nX:FP_Z2UX:FP3X805XN2Qh:hV[VJ^9ZhV[nP_j2oX;nP_h8J00@700031a800122JX9ZPVZ2JX9l
PGb1O85l6J001FZ2PH:100f201:1WiZOVYnJWj2iX;nP_Z2nX;hJ00@700031aH000B1Qh6228406J2V
X:FPYJ2UX;nP_Z2nX;jP_Z2UX:FPYJ00488058N2R9ZiV[VJ^:2oX;nP_j2oX;nP6P041`000`LK0006
O85lPGb16Z000fZ2P@0APP0@VYnJWiZQX;nP_j2oX;jP_QX010L000<76`05P@0LX:JPYZ2UX:FPYJ2n
X;nPYJ2nX:FPYJ2UX:FPY@f201J7PX:2R8:8V[RJ_j2oX;nP_j2oX;nP6P041`000`LK0004O85lPAfP
00B2PH:148803hNJWj2oX:JP_j2UX;nP_P0J00@700031a/018407JJPYZ2VX:FPYJ2UX;nP_Z2nX:FP
YJ2UX:FPYJ2U00n201F7PXR2R8:8WknP_j2oX;nP_j2oX8406@041`000`LK0003O85l01jP00B2PH:1
4H803Z2QX:JPYZ2VX:JP_j2o6P041`000`LK00:101nVX:JPYZ2VX:FPYJ2UX:JP_j2UX:FPYJ2UX:FP
YJ2U00f201N7PX:2Qj2oX;nP_j2oX;nP_j2oX;nPP@0I00@700031a/000=lPJ007Z0018:1PX4;PP0E
X:6PXJ2VX:6PYZ2VX:JPYZ2VX:F101T010L000<76`008H6PYZ2VX:JPYZ2UX:FPYJ2VX:FPYJ2UX:FP
YJ2UX:FPY@0;PP0I_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nPP@0I00@700031a`08J008fZ2PH:1PZ2Q
X:6PXJ2QX:6PXJ2QX:JPYZ2VX:JPYZ2VX:J101T010L000<77003X00MXJ2QX:6PXJ2QX:6PXJ2QX:FP
YJ2UX:FPYJ2UX:D00h808KnP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nPX@0I00@700031a`0
012PPZ22X8:PPZ22X8NPQj220j000hNPX004X00ZXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2VX:JP
YZ2VX:JPYZ2VX:J26@041`000`LL000QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX8:100BP
01noX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2o01T010L000<770004Z22X8:PPZ22X8:PQj27
X8NPQ`RP00bQX:4IPQV2QZ26X8H2X00LXJ2QX:6PXJ2QX:JPYZ2VX:JPYZ2VX:JPYZ2VX1T010L000<7
70006Z2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ210X803FZ2JX9ZX9nPWj26X8H01:005knP_j2oX;nP
_j2oX;nP_j2oX;nP_j2o01T010L000<770004:22X8:PPZ22X8:PPZ22X887X0@001<IPAV16H:1X86P
PJ21X86PPJ21008001FVX:6PXJ2QX:6PYZ2VX:JPYJ2UX88060041`000`LL000EX:6PXJ2QX:6PXJ2Q
X:6PXJ2QX:6P00L001:1PX62PJ2OX8JPQZ26X8JPQZ04000CX;nP_j2oX;nP_j2oX;nP_j2oX00H00@7
00031a`0012PPZ22X8:PPZ22X8:PPZ220j09000APAV16H8IX86PPJ21X86PPJ0020004:2QX:JPXJ2V
X:JPYZ2UX:HH00@700031a`0016PXJ2QX:6PXJ2QX:6PXJ2QX00<000@PX62PH:6X8JPQZ26X8JPQP/0
00fP_j2oX;nP_j2oX;nP01P010L000<770004:22X8:PPZ22X8:PPZ22X1T>000>PAV16J21X86PPJ21
X84?000:X:JPYZ2VX:JPYQP010L000<770003Z2QX:6PXJ2QX:6PXJ2Q40003X62PH:6X8JPQZ26X8JP
4`001[nP_j2oX1P010L000<770003:22X8:PPZ22X8:PPQ<000`IPAV2PJ21X86PPJ0G0004YZ2VX1L0
10L000<770002Z2QX:6PXJ2QX:4F000;PH9ZX8JPQZ26X8H06@000j2o000F00@700031a`000RPPZ22
X8:PPQP000Z16H:1X86PPJ21<`041`000`LL0006X:6PXJ2Q6`002H:1X8JPQZ26X00c00@700031a`0
00BPPZ227@0021V26J21X86P=0041`000`LL0003X:4001h000R2JX:6X8JPQS@010L000<7?P001X8I
X86PPCD010L000<7?P001VZ2QZ26X3D010L000<7?`0011V2PJ0f00@700031d0000>1X8H0=P041`00
0`M000036H4003H010L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT0
10L000<7N@041`000`Mi00@700201`00P0L008070000\
\>"],
  ImageRangeCache->{{{0, 127}, {127, 0}} -> {-3.20067, -3.20067, 1.05828, 
  1.05828}},
  CellTags->"Animation"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
%%ImageSize: 128 128 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00744048 0.0238095 0.00744048 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
384 string
128 128 8 [128 0 0 128 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000063525A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000005A4A5A
635263313952000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000005A4A5A63525A
63525A63525A634A5A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000006352635A4A5A
635263634A5A635263634A5A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000634A5A63525A
63525A63525A63525A63526363525A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000736352736352000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005A4A5A6352635A4A5A
63526363525A635263634A5A6B526363525A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7B6B5273634A7B6352000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000947B73947B73000000000000000000000000
000000000000000000000000000000000000000000000000000000635263634A5A63525A
63525A63526363525A6352636352636B526363525A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B6352
7363527B635273634A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000947B738C7373947B738C736B947B738C736B
0000000000000000000000000000000000000000000000006352635A4A5A635263634A5A
635263634A5A6B526363525A6B526363525A6B526363525A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000007363527B6B52736352
7B6B5273634A7B6B52000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000947B73947B73947B73947B73947B73947B73
947B73947B73947B73000000000000000000000000000000634A5A63526363525A635263
63525A6352636352636352636352636352636352636B52636B52636B5263000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000007B63527B6B527B63527B6B52
7363527B6352736352736352000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000031314A8C7373947B73947B73947B738C736B
947B738C7B73947B738C736B947B738C736B000000000000635263634A5A635263634A5A
6B526363525A63526363525A6B5A636352636B526363525A6B5A636352636B5A63000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007B63527B6B5A7B63527B6B52736352
7B6B527363527B6B5273634A000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000312942947B73947B73947B73947B73947B73
947B73947B73947B73947B73947B73947B738C7B6B947B73947B7363526363525A635263
6352636352636352636352636352636B52636352636B52636B52636B5A636B52636B5263
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007B6B527B6B527B6B527B6B527B63527B6B52
7B63527B6B52736352736352000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000947B73947B738C7B73947B738C7B73
947B738C7B73947B738C7B73947B738C736B947B738C736B947B738C736B947B738C736B
63526363525A6B52636352636B52636352636B5A636352636B5A636352636B5A6B6B5263
6B5A6B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000846B5A7B6B52846B5A7B63527B6B5A7B63527B6B5A736352
7B6B527363527B6B52736352000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031314A947B73947B73947B73947B73
947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73
8C7B6B947B73947B6B947B736352636B52636B52636B52636B52636B5A636B52636B5A63
6B52636B5A63000000000000000000000000000000000000000000000000000000000000
000000000000000000846B5A7B6B52846B5A7B6B527B6B5A7B6B527B6B527B6B527B6B52
7B63527B6B527363527B6B52736352000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031294294847B947B73947B738C7B73
947B7B947B73947B738C736B947B73947B73947B738C736B947B738C7B6B947B738C736B
947B738C7B6B947B738C736B947B738C736B947B736352636B5A6B6B52636B5A6B6B5263
735A6B6B5263735A6B000000000000000000000000000000000000000000000000000000
00000000000084735A7B6B5284735A7B6B52846B5A7B6352846B5A7B6B527B6B5A736352
7B6B5A7363527B6B5273634A7B6B52000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
947B7B8C7373000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031314A947B73947B7B947B73947B73
947B73947B73947B73947B73947B73947B73947B73947B73947B73947B738C7B6B947B73
947B73947B738C7B6B947B73947B6B947B738C7B6B947B6B947B6B947B736B52636B5A63
6B5A636B5A6B6B52636B5A6B394263000000000000000000000000000000000000000000
846B5A84735A846B5A846B5A846B5A846B5A7B6B527B6B5A7B6B527B6B5A7B6B527B6B52
7B6B527B6B527363527B6352736352000000000000000000000000000000000000000000
0000000000000000008C73738C73738C73738C73738C73738C737394737B8C7373947373
947373000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000031294231314A947B7394847B947B73
947B7B8C7B73947B7B8C7B73947B738C7B73947B738C7B73947B738C7B6B947B738C7B6B
947B738C7B6B947B738C7B6B947B738C736B947B738C736B947B738C736B947B738C736B
947B736B5263735A6B6B5263735A6B6B5A63000000000000000000000000000000846B5A
84735A846B5A84735A7B6B5284735A7B6B52846B5A7B6B52846B5A7B6352846B5A7B6352
7B6B5A7363527B6B527363527B6B52846B738C737B846B738C737B846B738C737B8C6B73
8C737B8C6B7394737B8C737394737B8C737394737B8C737394737B8C7373947B7B8C7373
947B7B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031294A947B73947B73947B73
947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73
947B73947B73947B73947B738C7B6B947B73947B6B947B738C7B6B947B73947B6B947B73
8C7B6B947B6B8C7B6B947B6B6B5A6B735A6B735A6B00000000000000000084735A84735A
846B5A84735A846B5A846B5A7B6B5A846B5A7B6B5A846B73846B73846B73846B738C7373
846B738C7373846B738C7373846B738C73738C73738C73738C73738C73738C73738C7373
8C73738C73738C73738C737B8C73738C73738C737394737B8C737394737B8C737394737B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031314A31294A94847B947B73
94847B947B73947B7B8C7B7394847B947B73947B738C7B73948473947B73947B738C7B6B
9484738C7B6B947B738C7B6B947B738C7B6B947B738C736B947B738C7B6B947B738C736B
947B738C7B6B947B738C736B947B738C7B6B947B6B8C736B8C7363846B5A393131312931
393939848C94846B737B6B738C737B846B7384737B846B738C737B846B738C737B846B73
8C737B846B738C737B846B738C737B846B738C737B846B738C737B8C6B738C737B8C6B73
94737B8C73738C737B8C6B7394737B8C737394737B8C7373947B7B8C737394737B8C7373
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031294231314A947B73947B73
947B73947B7B947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73
947B73947B73947B73947B73947B73947B738C7B6B947B73947B73947B738C7B6B947B6B
947B6B947B738C7B6B947B6B8C7B6B947B6B8C7B6B947B6B313131393131313131393131
848C9C84949C846B73846B73846B73846B7B846B73846B73846B738C737B846B73846B73
846B738C737B846B738C73738C73738C737B8C6B738C73738C73738C737B8C73738C7373
8C73738C737B8C73738C737B8C737394737B8C73738C737B8C737394737B8C7373000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031314A31294294847B947B73
94847B947B7394847B947B7394847B947B7394847B947B739484738C7B739484738C7B73
947B738C7B6B9484738C7B6B947B738C7B6B947B738C7B6B947B738C7B6B947B738C7B6B
947B738C7B6B947B738C7B6B947B738C736B947B73312931393131313131393139848C94
84949C848C9484949C846B7384737B846B738C737B846B738C737B846B738C737B846B73
8C737B846B738C737B846B738C737B846B738C737B846B738C737B8C6B738C737B8C7373
8C737B8C737394737B8C737394737B8C7373947B7B8C7373947B7B8C7373947B7B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000031294231314A31314A94847B
947B73947B7B947B73947B7B947B73947B73947B73947B73947B73947B73947B73947B73
947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73
8C7B6B947B73947B6B947B738C7B6B947B6B313131313131313131393131848C9C848C9C
848C94848C9C848C9C846B7B846B73846B7B846B7384737B846B7384737B846B738C737B
846B738C737B8C73738C737B8C73738C737B8C73738C737B8C73738C737B8C73738C737B
8C73738C737B8C737B8C737B8C73738C737B8C737B94737B8C737394737B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000031294231314A947B73
9C847B947B7394847B947B739C847B947B7394847B8C7B7394847B947B739484738C7B73
94847B947B739484738C7B6B9484738C7B73947B738C7B6B9484738C7B6B947B738C7B6B
9484738C7B6B947B738C7B6B94847331293139313131293139313931313184949C7B8C94
84949C848C9484949C848C948C737B846B7384737B846B738C737B846B738C737B846B73
8C737B846B738C737B846B738C737B8C73738C737B846B738C737B8C73738C737B8C7373
947B7B8C737394737B8C7373947B7B8C7373947B7B8C7373947B7B8C7373000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000031314A31294A94847B
947B7B94847B947B73947B73947B7394847B947B73947B73947B73948473947B73947B73
947B73948473947B73947B73947B73947B73947B73947B73947B73947B73947B73947B73
947B73947B738C7B6B947B73313131313131313131393131393131848C9C848C94848C9C
848C9C84949C848C94848C9C846B7384737B846B7384737B84737B8C737B846B738C737B
8C737B8C737B8473738C737B8C737B8C737B8C73738C737B8C737B8C737B8C73738C737B
8C737B8C737B8C737B8C737B8C737B94737B8C737B94737B94737B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000029294231314A31294A
94847B947B7394847B947B7394847B947B7394847B947B7394847B947B7394847B947B73
94847B947B7394847B947B739484738C7B739484738C7B739484738C7B6B9484738C7B6B
947B738C7B6B9484738C7B6B393131312931393139313131848C9C7B8C9484949C7B8C94
84949C7B8C9484949C7B8C9484949C846B738C737B846B738C737B846B738C737B846B73
8C737B846B738C737B846B738C737B8473738C737B8C73738C737B8C73738C737B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B8C7373947B7B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000031314231294A31314A
947B7B94847B947B7B94847B947B7394847B947B7394847B947B73947B73947B73948473
947B73947B73947B73948473947B73947B73947B73947B73947B73947B73947B73947B73
947B73947B73947B73313131313131393131313131848C9C7B8C94848C9C7B8C94848C9C
848C94848C9C848C94848C9C848C9484737B846B7B84737B84737B8C737B84737B8C737B
84737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B
8C737B8C737B8C737B94737B8C737B94737B8C737B947B7B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000029294231314A312942
9C847B947B7394847B947B739C847B947B7394847B947B739C847B947B7394847B947B73
94847B947B739484738C7B7394847B947B739484738C7B73948473947B739484738C7B6B
9484738C7B6B313131312931393139313131848C9C7B8C9484949C7B8C94848C9C7B8C94
84949C7B8C94848C9C7B8C9484949C7B8C948C737B846B738C737B846B738C737B846B73
8C737B84737B8C737B846B738C737B8C737B8C737B8C7373947B7B8C737B8C737B8C7373
947B7B8C737B947B7B8C7373947B7B8C737B947B7B8C737B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000031294231314A
31314A94847B947B7B94847B94847B94847B947B7394847B94847394847B947B73948473
947B7394847B947B73947B73947B73948473947B73947B73947B73948473947B73947B73
947B73313131313131313131313131848C9C7B8C947B8C9C7B8C94848C9C7B8C947B8C94
7B8C94848C9C7B8C94848C94848C94848C9C846B7B84737B84737B8C737B84737B8C737B
8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B
8C737B947B7B8C737B8C737B8C737B947B7B8C737B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000031314A312942
31314A947B739C847B947B7394847B947B739C847B947B7394847B947B7394847B947B73
94847B947B7394847B947B7394847B947B7394847B947B739484738C7B739484738C7B73
948473312931393139313131848C9C7B8494848C9C7B8C94848C9C7B8C94848C9C7B8C94
848C9C7B8C94848C9C7B8C94848C9C7B8C9484949C846B7B8C737B846B738C737B84737B
8C737B84737B8C737B84737B8C737B8C737B8C7B7B8C737B8C7B7B8C737B947B7B8C737B
947B7B8C737B947B7B8C737B947B7B8C737B947B84000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000031294231314A
31294A94847B94847B94847B94847B94847B94847B94847B947B7394847B94847394847B
947B7394847B947B73948473947B73948473947B73948473947B73948473947B73948473
3131313131313131317B8C9C7B8C947B8C9C7B8C947B8C9C7B8C947B8C947B8C947B8C9C
7B8C947B8C947B8C94848C947B8C94848C947B8C948C737B84737B8C737B84737B8C737B
8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C7B7B
8C737B8C7B7B8C737B947B7B8C737B947B7B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000313142292942
31314A31294A94847B947B739C847B947B7B94847B947B739C847B947B7394847B947B73
9C847B947B7394847B947B7394847B947B7394847B8C7B7394847B947B73948473312931
3931393131317B8C9C7B8494848C9C7B8C947B8C9C7B8494848C9C7B8C947B8C9C7B8494
848C9C7B8C94848C9C7B8C9484949C7B8C94848C947B8C948C738484737B8C737B84737B
8C737B84737B8C737B84737B8C7B848C737B8C7B7B8C737B947B848C737B947B7B8C737B
947B848C737B947B7B8C737B947B848C737B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000292942313142
31314A31314A94847B94847B94847B94847B94847B94847B94847B94847B947B7394847B
94847B94847B947B7394847B94847394847B947B73948473948473948473313131313131
3131317B8C9C7B84947B8C947B8C947B8C9C7B8C947B8C947B8C947B8C947B8C947B8C94
7B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C948C737B8C737B84737B8C737B
8C737B8C737B8C737B8C737B8C737B8C7B7B8C737B8C737B8C737B8C7B7B8C737B8C737B
8C737B947B7B8C737B947B7B947B7B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000292942
31314A3129429C847B947B7B9C847B947B7B9C847B947B7B9C847B947B739C847B947B73
94847B947B739C847B947B7394847B947B7394847B947B7394847B312931393139313131
7B8C9C7384947B8C9C7B84947B8C9C7B84947B8C9C7B84947B8C9C7B84947B8C9C7B8494
7B8C9C7B8494848C9C7B8C947B8C947B8C94848C9C7B8C94848C9484737B8C738484737B
8C737B84737B8C7B848C737B8C7B848C737B947B848C737B947B848C737B947B848C737B
947B848C737B947B848C737B947B84000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000523939313142
31294231314A31314A94847B94847B94847B94847B94847B94847B94847B94847B94847B
94847B94847B94847B94847B94847394847B94847394847B9484733131313131317B8C9C
7B84947B8C947B84947B8C947B84947B8C947B8C947B8C947B8C947B8C947B8C947B8C94
7B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C948C737B8C737B8C737B
8C737B8C737B8C737B8C737B8C737B8C737B8C737B8C7B7B8C737B8C7B7B8C737B947B7B
8C737B947B7B8C7B7B947B84000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005A39395239395A3939523131
31314A31294231314A947B7B9C848494847B9C847B947B739C847B94847B94847B947B73
9C847B947B7394847B947B739C847B947B7394847B947B733931393131317B8C9C738494
7B8C9C7384947B8C9C7384947B8C9C7384947B8C947384947B8C9C7B84947B8C94738494
7B8C9C7B8C947B8C947384947B8C9C7B8C947B8C947B8494848C947B8C948C738484737B
8C7B848C737B8C7B848C737B947B848C737B8C7B848C737B947B848C737B947B848C737B
947B848C737B947B848C737B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005239395A39395A39395A39395239395A3939
31294231314A31294A94847B94847B9C847B94847B94847B94847B94847B94847B94847B
94847B94847B94847B94847B94847B94847B9484733131313131317B8C947384947B8494
7B84947B8C947384947B8C947B84947B8C947384947B8C947B8C947B8C947B84947B8C94
7B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C948C737B8C737B
8C737B8C7B848C737B8C737B8C737B8C7B848C737B8C7B7B8C7B7B947B848C737B8C7B84
8C7B7B947B848C7B7B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005239395A39395239395A39395239395A39395239395A3939523939
5A393929294231314A31294A9C848494847B9C848494847B9C847B94847B9C847B94847B
9C847B947B739C847B9484739C847B947B733931313131317B8C9C7384947B8C9C738494
7B8C9C7384947B8C9C7384947B8C947384947B8C9C7384947B8C947384947B8C94738494
7B8C947384947B8C947384947B8C947384947B8C947B8C947B8C947B848C7B8C9484737B
8C7B848C737B8C7B848C737B8C7B848C737B947B848C737B947B848C737B947B848C737B
947B848C737B947B84000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A39395239395A39395A39395A39395A39395A39395A39395A39395A39395A3939
5A393931314A31314A31314A94847B94847B94847B94847B94847B94847B94847B94847B
94847B94847B94847B94847B94847B94847B3131313931317384947384947384947B8494
7384947B84947384947B8C947384947B8C947384947B8C947384947B8C947B8C947B8C94
7B84947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C947B8C948C7384
8C737B8C7B848C737B8C7B848C737B8C7B848C737B8C7B848C7B7B8C7B848C7B7B947B84
8C7B7B947B84000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000523131
5A42395239395A39395239395A42395A39395A39395239395A42425A39395A39395A3939
6342425A393931314A31294A9C8C8494847B9C848494847B9C8C8494847B9C847B94847B
9C847B94847B9C847B9484739C847B3131313931397384947B8C9C7384947B8C94738494
7B8C9C7384947B8C947384947B8C9C7384947B8C947384947B8C947384947B8C9473848C
7B8C947384947B8C9473848C7B8C947384947B8C9473848C7B8C947B8C947B8C9473848C
8C7B848C737B8C7B848C737B947B848C737B947B848C737B947B848C737B947B848C737B
947B848C7B7B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000005A39395239395A3939
5A39395A39395A39395A39395A39395A39395A39395A39395A39395A39395A39395A3939
5A39395A393931294231314A31314A9C848494847B94847B94847B9C847B94847B94847B
94847B9C847B94847B94847B313131393131738494738494738494738494738494738494
7384947384947384947384947384947B8C947384947384947384947B8C94738494738C94
7B8C947B8C947384947B8C947B8C947B8C947384947B8C947B8C947B8C947B8C947B8C94
8C73848C7B848C737B8C7B848C7B848C7B848C737B8C7B848C7B84947B848C7B7B947B84
947B84000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005239395A39395239395A3939523939
5A39395A39395A42425A39395A42395A39395A42425A39396342425A39396342425A3939
6342425A393963424231294231315294847B9C8C8494847B9C848494847B9C8C8494847B
9C847B94847B9C847B31293139313973849473849C7384947384947384947B8C94738494
738C947384947B8C947384947B8C947384947B8C947384947B8C9473848C7B8C94738494
7B8C9473848C7B8C947384947B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C948C737B8C7B848C737B947B848C737B947B848C737B947B848C737B947B848C7B7B
947B84634231634239000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000005A39395A39395A39395A39395A39395A39395A3939
5A39395A39395A39395A39395A39395A39395A39395A39395A39395A39395A3939634239
5A39396339395A393963423931314A31315294847B9C848494847B94847B94847B94847B
94847B94847B313131313131738494738494738494738494738494738494738494738494
738494738494738494738494738494738494738494738C94738494738494738494738C94
738494738C947384947B8C947384947B8C94738C947B8C94738C947B8C947B8C947B8C94
738C8C8C7B848C7B848C7B848C7B848C7B848C7B84947B848C7B84947B848C7B84947B84
634231634231634231634231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005239395A42425A39395A39395239395A42425A39395A42425A3939
6342425A39395A42425A39396342425A39396342425A39396342425A39396342425A3939
6342425A39396342425A393931314A31294A9C8C8494847B9C8C8494847B9C8C8494847B
9C8C8494847B3931316B7B9473849C7384947384946B84947384947384947384946B848C
738C9473849473849473848C7B8C94738494738C9473848C7B8C947384947B8C9473848C
7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C8C7B848C737B947B848C7384947B848C737B947B848C7B84947B848C7B7B
6B4A39634231634A396342316B4A39634231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A39395A39395A39395A39395A39395A39395A39395A39395A39395A39395A3939
5A39395A42425A39395A39395A39396342425A39396339395A39396342425A3939634239
63393963424263393963423963423931314A94847B94847B94847B9C848494847B94847B
94847B3131316B8494738494738494738494738494738494738494738494738494738494
738494738494738494738494738494738494738494738494738494738C94738494738494
738494738C9473848C7384947384947B8C9473848C738C94738C947B8C9473848C738C94
738C947B8C948C7B848C7B848C7B84947B848C7B848C7B848C7B84947B848C7B846B4239
6342316B4A39634231634231634231634239634231634231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000948C8C523939
5A42425239395A42425A39395A42425A39395A42425A39396342425A39396342425A3939
6342425A39396342425A39396342425A39396342425A39396342425A39396342425A3939
6342425A393963424263393963424231294A39315294847B9C8C8494847B9C8C8494847B
3931396B7B947384946B7B947384946B7B947384946B84947384946B848C7384946B8494
7384946B848C738C9473848C73849473848C738C9473848C738C9473848C738C9473848C
738C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C7B8C9473848C
7B8C9473848C7B8C948C7384947B848C7384947B848C7B84947B848C7B84947B84634231
6B4A396342316B4A396342316B4A396342316B4A396342316B4A39000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005A39395A39395A39395A3939
5A39395A39395A39395A39425A39395A39395A39395A39425A39395A39395A3939634242
5A39396342425A39396342425A3939634242633939634242633939634242633939634242
63393963424263423963424263423963424231314A9C8C8494847B9C848494847B313131
6B84947384946B84947384946B84947384947384947384946B8494738494738494738494
738494738494738494738494738494738494738494738494738494738494738494738494
73848C738494738494738C9473848C738C9473848C738C9473848C738C9473848C738C94
73848C738C94738C8C8C7B848C7B848C7B848C7B84947B848C7B84947B846B4A396B4A39
6B42316B4A396B42316B4A396342316B4A396342316B4A39634231634231634231000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000005A39425239395A42425A39395A42425A3939
6342425A39395A42425A39396342425A39396342425A39396342425A39396342425A3939
6342425A39396342425A39396342425A39396342425A3939634242633939634242633939
6342426339396342426339396B42426339396B424294847B9C8C8494847B9C8C846B7B8C
7384946B7B947384946B7B8C7384946B7B947384946B7B8C7384946B84947384946B7B8C
7384946B848C7384946B848C738C9473848C7384946B848C738C9473848C7384946B848C
738C9473848C738C9473848C738C9473848C738C9473848C7B8C9473848C738C9473848C
7B8C9473848C7B8C9473848C947B8C8C7B84947B848C7B8494848C8C7B846B4A396B4231
6B4A396B42316B4A396342316B4A396342316B4A396342316B4A396342316B4A39634231
6B4A39000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000005A39395A39425A39395A39395A39395A42425A39395A3939
5A39395A42425A39395A39395A39396342425A39396342425A39396342425A3939634242
634242634242633939634242634239634242633939634242634242634242634239634242
6342426342426342396342426342426B42426342393131529484849C8C846B7B946B8494
6B84947384946B7B946B84946B84947384946B84946B84946B84947384946B8494738494
7384947384946B848C73849473849473849473848C73849473849473849473848C738494
73849473849473848C73849473848C738C9473848C73848C73848C738C9473848C738C8C
73848C738C9473848C738C8C8C7B84947B848C7B84947B848C7B84734A396B4A316B4A39
6B4A396B4A396B42316B4A396B4A396B4A396B42316B4A396B4A316B4A396342316B4A31
6342316B4A39000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000005263844A637B5263844A637B526B844A637B526B8452637B
526B8452637B526B8452637B526B845263845A6B845263845A6B845A39396342425A3939
6342425A39396342426339396342426339396342426339396342426339396B4242633939
6B42426339396B42426339396B42426339396B42426342399C8C846B7B8C7384946B7B94
7384946B7B8C7384946B7B8C7384946B7B8C7384946B7B8C7384946B7B8C7384946B848C
7384946B7B8C7384946B848C7384946B848C7384946B848C7384946B848C738C946B848C
738C946B848C738C9473848C738C946B848C738C9473848C738C9473848C738C9473848C
738C9473848C738C9473848C947B8C8C7B8494848C8C7B8494848C6B4231734A396B4231
734A396B4231734A396B42316B4A396B42316B4A396B42316B4A396342316B4A39634231
6B4A396342316B4A39634231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000000000052637B52638452637B52638452637B526384
52637B526384526384526B84526384526B84526384526B84526384526B84526B84526B84
526B84526B84526B845A6B84526B845A6B84526B845A6B8C5A6B845A6B8C5A6B8C5A738C
5A6B8C5A738C5A6B8C5A738C5A738C5A738C5A738C5A738C73524A6B84946B7B946B8494
6B7B946B84946B7B946B84946B7B946B84946B84946B84946B84946B84946B8494738494
6B848C7384946B84947384946B848C7384946B84947384946B848C73849473848C738494
73848C73849473848C73849473848C73848C73848C73849473848C73848C73848C738C94
73848C738C8C73848C738C8C73848C947B848C7B84947B84734A39734A396B4A39734A39
6B4A39734A396B4A39734A396B4A316B4A396B4A396B4A396B4A316B4A396B4A316B4A39
6B42316B4A396B42316B4A39634231000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A5A7B5263844A637B5263844A637B
526B8452637B526B844A637B526B8452637B526B8452637B5A6B84526384526B84526384
5A6B8C526B845A6B845263845A6B8C526B845A6B8C526B845A738C526B845A738C526B84
5A738C526B845A738C526B845A738C5A6B8C5A738C5A6B849C948C73524A7B525273524A
7B5A5273524A7B525273524A7B5A5273524A7B525273524A7B5A527B524A7B52526B7B8C
7384946B848C7384946B7B8C7384946B848C7384946B7B8C7384946B848C7384946B848C
738C946B848C7384946B848C738C946B848C738C946B848C738C9473848C738C946B848C
738C9473848C738C946B848C738C948C7B84947B8C8C7B84735239734A39734A396B4A31
734A396B4A39734A396B4231734A396B4A31734A396B4231734A396B4A316B4A396B4231
6B4A396B42316B4A396342316B4A396342316B4A39000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000052637B52637B4A637B52637B
52637B52638452637B526384526384526B8452637B526384526384526B84526384526B84
526B84526B84526B84526B84526B845A6B84526B84526B84526B845A6B8C526B845A6B84
5A6B845A738C5A6B845A738C5A6B8C5A738C5A6B8C94948C948C8C31315A7B524A7B5252
7B524A7B52527B524A7B524A7B524A7B52527B524A7B524A7B524A7B52527B524A7B524A
7B524A7B52527B524A7B524A7B524A7B524A7B524A7B524A7B524A7B5A4A7B524A7B524A
7B524A7384946B848C73848C73848C7384946B848C73848C73848C738C8C73848C73848C
73848C738C8C73848C73848C73848C738C8C8C7B84734A39734A39734A39734A39734A39
734A39734A396B4A39734A39734A39734A396B4A396B4A396B4A39734A396B4A316B4A39
6B4A396B4A396B4A316B4A396B4A316B4A396B42316B4A316B4A31000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A5A7B5263844A637B
5263844A637B526B844A637B526B844A637B526B8452637B526B8452637B526B8452637B
526B845263845A6B845263845A6B845263845A6B8C526B845A6B8C526B845A738C526B84
5A738C526B845A738C526B845A738C526B845A738C948C8C9C948C948C8C3939637B524A
7B5A5273524A7B5A527B524A7B5A527B524A7B5A527B524A7B5A527B524A7B5A527B524A
7B5A527B524A7B5A527B524A7B5A527B524A845A527B524A845A527B524A845A527B524A
845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A
845A4A7B524A738C946B848C738C8C6B848C94848C734A397B5239734A397B5239734A39
735239734A317352396B4A39734A396B4A31734A396B4A31734A396B4A31734A396B4A31
734A396B4231734A396B42316B4A396B42316B4A396B42316B4A39634231000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A637B52637B
4A637B52637B52637B52638452637B52638452637B52638452637B526B84526384526B84
526384526B84526B84526B84526B84526B84526B84526B84526B84526B84526B845A6B84
526B845A6B845A6B845A6B8C5A6B845A738C948C8C94948C948C8C94948C31315A313163
7B524A7B52527B524A7B52527B524A7B52527B524A7B5A527B524A7B52527B524A7B5A52
7B524A7B52527B524A7B5A527B524A7B52527B524A7B5A527B524A7B5A4A7B524A7B5A52
7B524A7B5A4A7B524A7B5A4A7B524A7B5A4A7B524A845A4A7B524A7B5A4A7B524A845A4A
7B524A845A4A7B524A845A4A7B524A845A4A8C84946B848463847B6B848463847B6B8484
637B7B734A39734A39734A39734A39734A39734A39734A39734A39734A396B4A39734A39
6B4A39734A396B4A396B4A396B4A316B4A396B4A316B4A396B4A316B4A396B4A316B4A39
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
5263844A637B52637B4A637B526B844A637B526B844A637B526B8452637B526B844A637B
526B8452637B526B8452637B5A6B84526384526B845263845A6B8C526B845A6B84526B84
5A738C526B845A6B8C526B845A738C3131319C948C948C8C9C9494948C8C94948C31315A
3939637B524A7B5A527B524A845A527B524A7B5A527B524A845A527B524A845A527B524A
845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A
845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A
845A527B524A845A527B524A845A528C8494948494637B7B6B8484637B7B6B8484637B7B
6B8484637B7B6B847B637B7B6B8484637B7B63847B637B736B847B637B7B63847B6B4A31
7352396B4A31734A396B4A31734A396B4A31734A396B4231734A396B4A316B4A396B4231
6B4A396B4A31000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000052637B4A637B52637B52637B5263844A637B52637B52637B52638452637B526384
526384526B8452637B526B84526B84526B84526384526B84526B845A6B84526B84526B84
526B845A6B84526B845A6B845A6B8494948C948C8C94948C948C8C94948C948C8C31315A
31315A3131637B52527B5A527B52527B5A527B524A7B5A527B5252845A527B524A7B5A52
7B5252845A527B524A7B5A527B5A52845A527B524A7B5A527B5A4A845A527B524A7B5A52
7B5A4A845A527B524A845A527B5A4A845A527B524A845A4A845A4A845A527B524A845A4A
845A4A845A527B524A845A4A8C84948C84948C84948C8494637B7B638484637B7B637B7B
637B7B63847B637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B73637B7B
637B73637B7B637B73637B73637B73637B73637B73637B736B4A39734A396B4A316B4A39
6B4A396B4A396B4A31000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000052637B4A637B5263844A637B5263844A637B5263844A637B526B844A637B
526B844A637B526B8452637B526B8452637B526B8452637B5A6B845263845A6B84526B84
5A6B84526B845A6B8C526B84313139948C8C9C9494948C8C94948C948C8C94948C948C8C
39396331315A845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A
845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A
845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A845A527B524A
845A527B524A845A527B524A9484948C8494948C948C8494638484637B7B6B8484637B7B
63847B637B7B63847B637B7B63847B637B7363847B637B73637B7B637B7363847B637B73
637B7B5A7B73637B7B5A7B73637B7B5A7B73637B7B5A7B73637B735A736B637B735A736B
637B735A736B637B735A736B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000052637B4A637B52637B4A637B52637B4A637B52637B52637B52637B
52637B52638452637B526B8452637B526B84526384526B84526B84526B84526B84526B84
526B84526B84526B84526B84948C8C94948C94948C94948C948C8C94948C948C8C94948C
31315A313163313163845A527B5252845A527B5A52845A527B5A52845A527B5A52845A52
7B5252845A527B5A52845A527B5A52845A52845A52845A527B5A4A845A52845A52845A52
845A4A845A52845A52845A52845A4A845A52845A4A845A52845A4A845A52845A4A845A52
845A4A845A52845A4A8C84948C84948C84948C84948C8494637B7B637B7B637B7B637B7B
637B7B637B7B637B7B637B7B637B7B637B7B637B73637B7B637B73637B7B637B73637B7B
637B73637B73637B73637B73637B73637B73637B73637B735A7B73637B735A7B73637B73
5A7B6B637B735A7B6B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A637B52637B4A5A7B526B844A637B5263844A637B
526B844A637B526B844A637B526B8452637B526B8452637B526B8452637B526B8452637B
5A6B84526B845A6B843129319C9494948C8C949494948C8C9C9494948C8C94948C948C8C
94948C31315A31316331315A845A527B5252845A527B524A845A527B5252845A527B524A
845A527B5252845A527B524A845A527B524A845A527B524A845A527B5A4A845A527B524A
845A527B5A4A845A527B524A845A52845A4A845A527B524A845A52845A4A845A527B524A
845A52845A4A8C84948C8494948C948C84948C84948C8494948C9C637B7B637B7B5A7B73
63847B637B73637B7B5A7B73637B7B637B73637B7B5A7373637B7B5A7B73637B7B5A7373
637B7B5A7B73637B735A7373637B735A7373637B735A736B637B735A736B637B735A736B
637B73000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004A637B4A637B4A637B52637B4A637B52637B
52637B52637B4A637B52637B52637B526B8452637B526B84526B84526B8452637B526B84
526B84526B8431293194948C94948C949494948C8C94948C94948C94948C948C8C94948C
948C8C94948C31315A313163313163845A527B5A52845A52845A52845A527B5A52845A52
845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52
845A52845A52845A52845A52845A52845A52845A4A845A52845A52845A52845A4A845A52
845A528C84948C84948C84948C84948C84948C84948C84948C8494637B7B637B7B637B7B
637B7B637B7B637B73637B7B637B73637B7B5A7B73637B73637B73637B735A7B73637B73
637B73637B735A7373637B735A7B73637B735A73735A7B735A7B73637B735A736B000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A5A7B52637B4A637B52637B4A637B
52637B4A637B526B844A637B526B844A637B526B844A637B526B844A637B526B8452637B
526B8452637B949494948C8C949494948C8C949494948C8C949494948C8C94948C948C8C
94948C8C8C8439396331315A393963313163845A527B5A52845A527B5252845A527B5A52
845A527B5252845A527B5A52845A527B524A845A52845A52845A52845A4A845A52845A52
845A52845A4A845A52845A4A845A52845A4A845A52845A4A845A52845A4A845A52845A4A
845A528484948C84948C84948C84948C84948C8C9C8C8494948C9C5A7B73637B7B5A7B73
637B7B5A7373637B7B5A7373637B7B5A7373637B7B5A7373637B7B5A7373637B735A7373
637B735A7373637B735A736B637B735A736B637B735A736B637B73000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A637B4A637B4A637B52637B
4A637B52637B4A637B52637B4A637B52637B52637B526B8452637B526B8452637B526B84
52637B31313194948C949494948C8C94948C94948C94948C948C8C94948C948C8C94948C
948C8C94948C948C8C313163313163313163313163845A52845A52845A52845A52845A52
845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52
845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52
8C84948C84948C84948C84948C84948C84948C84948C84948C84948C8C94637B7B637B7B
5A7B73637B7B5A7B73637B735A7B73637B735A7B73637B735A73735A7B735A73735A7B73
5A73735A7B735A73735A7B735A736B5A73735A736B5A7B73000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A5A7352637B4A5A73
52637B4A637B52637B4A637B526B844A637B526B7B4A637B526B8452637B526B844A637B
526B84948C8C9494948C8C8C949494948C8C9494948C8C8C949494948C8C94948C8C8C8C
949494948C8C94948C31315A39396331315A39396331316384635A845A52845A52845A52
845A5A845A52845A52845A528C635A845A52845A52845A528C6352845A52845A52845A4A
8C6352845A52845A52845A4A8C6352845A52845A52845A4A8C6352845A528C5A52847B94
8C84948484948C84948484948C8C9C8C84948C8C9C8C8494948C9C8C8494637B7B5A7373
637B7B5A7373637B7B5A7373637B7B5A7373637B735A7373637B735A7373637B735A736B
637B735A736B5A7B735A736B637B735A736B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A637B
4A637B52637B4A637B52637B4A637B52637B4A637B52637B52637B526B7B52637B52637B
313131949494948C8C94948C94948C949494948C8C94948C94948C94948C948C8C94948C
94948C94948C948C8C94948C31315A313163313163313163313163845A52845A52845A52
845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52845A52
845A52845A52845A52845A52845A52845A52845A52845A52845A528C5A528484948C8494
8C84948C84948C84948C84948C84948C84948C84948C84948C84948C8C9C8C84945A7B73
5A7B73637B735A73735A7B735A73735A7B735A73735A73735A73735A7B735A73735A7373
5A73735A73735A736B5A736B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
52637B4A5A7352637B4A637B52637B4A637B52637B4A637B526B7B4A637B526B844A637B
9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C
94948C8C8C8C94948C8C8C8C31316331315A39396331316339396B3131638C635A845A52
84635A845A528C635A845A528C635A845A528C635A845A528C6352845A528C635A845A52
8C6352845A528C635A845A528C6352845A528C6352845A528C6352425A428C8494848494
8C84948484948C849C8484948C8C9C8C84948C8C9C8C84948C8C9C8C8494948C9C5A7373
637B7B5A7373637B7B5A73735A7B735A736B637B735A736B5A7B735A736B5A7B735A736B
5A73735A736B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A637B4A637B52637B4A637B4A637B4A637B52637B4A637B52637B52637B313131
94948C94949494948C94949494948C94949494948C949494948C8C94948C948C8C94948C
948C8C94948C948C8C94948C8C8C8C31315A31316331316331316331396B845A52845A5A
845A52845A52845A52845A52845A52845A52845A528C5A52845A52845A52845A528C5A52
845A528C5A52845A528C6352845A528C5A52845A528C5A52845A528C84948484948C8494
8484948C84948C84948C84948C84948C84948C84948C8C9C8C84948C8C948C84945A7373
5A73735A73735A73735A73735A73735A73735A73735A73735A73735A73735A736B5A7373
5A736B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A637B4A5A7352637B4A637B52637B4A6373526B7B4A637B3131318C8C8C
9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C94948C8C8C8C
9494948C8C8C94948C8C8C8494948C31315A31396331315A39396331316339396B845A52
8C635A845A528C635A845A528C635A845A528C635A845A528C635A845A528C635A845A52
8C635A845A528C635A845A528C635A845A528C6352845A528C849C8484948C8494847B94
8C849C8484948C849C8484948C8C9C8C84948C8C9C8484948C8C9C8C84948C8C9C8C8494
5A7B735A73735A737352736B5A7B735A736B5A737352736B5A7B735A736B5A7373000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A637B4A637B4A637B4A637B4A637B4A637B52637B313131949494
94949494949494948C94949494948C9494948C8C8C94948C94948C94948C8C8C8C94948C
94948C94948C8C8C8C94948C94948C94948C31315A31316331316331396331316331396B
845A528C635A845A528C5A52845A528C635A845A528C5A528C5A528C635A845A528C5A52
8C5A528C635A845A528C63528C5A528C635A845A528484948484948C84948484948C8494
8C84948C84948484948C84948C84948C849C8C84948C84948C84948C8C9C8C84948C8C9C
5A73735A73735A73735A73735A73735A73735A736B5A73735A736B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A5A7352637B4A637352637B2929319494948C8C8C
9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C
94948C8C8C8C94948C8C8C8C94948C8C8C8431396331315A39396331316339396B313163
39396B845A528C635A845A528C635A845A528C635A845A528C635A845A528C635A845A52
8C635A845A528C635A845A528C635A845A528C8494847B948C8494847B948C849C848494
8C849C8484948C8C9C8484948C8C9C8484948C8C9C8484948C8C9C8C84948C8C9C8C8494
8C8C9C52736B5A737352736B5A7373526B6B5A7373000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004A63734A637B4A6373313131949494949494
9494949494949494949494948C948C94949494948C9494948C948C94948C8C948C94948C
8C8C8C94948C8C8C8C94948C8C8C8C94948C8C8C8C31316331315A313163313163313963
31316339396B8C5A528C635A8C5A528C635A8C5A528C635A8C5A528C635A8C63528C635A
8C5A528C635A8C63528C635A8C5A528C635A8484948484948484948484948484948C8494
8484948C84948484948C849C8C84948C849C8C84948C8C9C8C84948C8C9C8C849C8C8C9C
8C84945A73735A736B5A737352736B5A736B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A5A733131398C94949494948C8C8C
9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C
9494948C8C8C94948C8C8C8C94948C8C8C8C94948C31315A39396331315A393963313163
39396B31316339396B845A528C635A8C5A528C635A845A528C635A8C5A528C635A845A52
8C635A8C5A528C635A845A528C635A847B94848494847B948C849C8484948C849C848494
8C8C9C8484948C849C8484948C8C9C8484948C8C9C8484948C8C9C8C84948C8C9C848494
948C9C526B6B5A7373526B6B000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003131319494948C9494949494
9494949494948C948C9494949494949494948C948C94948C8C948C9494948C8C8C8C948C
8C948C94948C8C8C8C8C948C8C948C94948C8C8C8C8C948C31315A313163313163313163
31316339396B31316B31396B8C635A8C635A8C63528C635A8C635A8C635A8C5A528C635A
8C635A8C635A8C63528C635A8484948484948484948484948484948C849C848494848494
8484948C849C8484948C84948C84948C8C9C8C84948C849C8C849C8C8C9C8C84948C8C9C
8C8C9C8C8C9C000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000002929319494948C8C949494948C9494
9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C
9494948C8C8C9494948C8C8C94948C8C8C8C94948C8C8C8431396331315A393963313163
39396B31316339396B31316B3939738C5A528C635A8C5A528C635A8C5A528C635A8C5A52
8C635A8C5A5294635A847B94848494847B9484849C847B948C849C8484948C849C848494
8C849C8484948C8C9C8484948C8C9C8484948C8C9C8484948C8C9C8484948C8C9C8C8494
8C8C9C8C8494000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003131318C94949494948C9494949494
8C94949494948C94949494948C948C9494948C948C9494948C948C8C948C8C948C94948C
8C948C8C948C8C948C94948C8C8C8C8C948C8C8C8C8C948C8C8C8C31315A31315A313163
31316331316331316339396B31316B8C635A8C635A8C635A8C635A8C635A8C635A8C635A
8C635A8C635A4A634A84849484849484849484849484849484849484849484849484849C
84849484849C8484948C849C8484948C849C8C849C8C8C9C8C84948C8C9C8C849C8C8C9C
8C8C9C8C8C9C000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003131318C8C9494949C8C94949494948C8C8C
9494948C94949494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C
9494948C8C8C94948C8C8C8C9494948C8C8C94948C8C8C8494948C31315A31316331315A
39396331316339396B31316339396B31316B8C635A8C5A5294635A8C635A94635A8C5A52
94635A8C635A8484947B7B9484849C847B9484849C847B948C849C84849484849C848494
8C8C9C8484948C8C9C8484948C8C9C8484948C8C9C8484948C8C9C8484948C8C9C848494
8C8C9C8C849C8C8C9C000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000003131319494948C94949494948C94948C9494
8C94949494948C94948C94948C94949494948C948C8C94948C948C9494948C8C8C8C948C
8C948C94948C8C8C8C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C31315A313163
31316331396331316331396B31396B39396B31316B8C635A8C635A94635A8C635A8C635A
8C635A848494847B9484849484849484849C84849484849484849484849C848494848494
8484948C849C84849484849C84849C8C8C9C8484948C8C9C8C8C9C8C8C9C8C849C8C8C9C
8C8C9C8C8C9C8C8C9C000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000029293194949C8C94949494948C8C949494948C9494
9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C
9494948C8C8C9494948C8C8C94948C8C8C8C94948C8C8C8C94948C8C8C8431396331315A
31396331316339396B31316339396B31316339397331316B94635A8C635A946B5A8C635A
8484947B7B9484849C7B7B9484849C847B9484849C84849484849C8484948C849C848494
8C849C8484948C8C9C8484948C8C9C8484948C8C9C8484948C8C9C8484948C8C9C84849C
8C8C9C84849C8C8C9C8C8C9C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003131318C94949494948C94948C94948C9494949494
8C94948C94948C94948C94948C94948C94948C948C8C94948C948C8C94948C948C8C9494
8C948C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C8C8C8C8C948C8C8C8C31315A
31315A31316331316331396331316331396B31396B39396B31396B94635A8C635A42634A
7B7B9484849484849484849484849484849484849484849C84849484849C84849484849C
84849484849C84849C8C849C84849C8C849C84849C8C8C9C84849C8C8C9C8C8C9C8C8C9C
8C8C9C8C8C9C8C8C9C8C8C9C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000003131398C94949494948C8C9494949C8C94949494948C8C94
9494948C94949494948C8C8C9494948C948C9494948C8C8C9494948C8C8C8C94948C8C8C
9494948C8C8C8C94948C8C8C9494948C8C8C8C948C8C8C8494948C8C8C8C8C948C31315A
39396331315A31396331316339396B31316339396B31316B39397331316B946B5A7B7B94
84849C7B7B9484849C7B7B9484849C84849484849C7B7B9484849C84849484849C848494
8C8C9C8484948C8C9C8484948C8C9C8484948C8C9C8484948C8C9C84849C8C8C9C848494
8C8C9C8C8C9C8C8C9C84849C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008C94949494948C94948C94948C94949494948C94948C9494
8C94948C94948C94948C94948C94948C94948C948C8C94948C948C8C94948C8C8C8C948C
8C948C8C94948C8C8C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C8C8C8C8C948C
31315A31316331316331316331316339396B31316B31396B31396B3939737B7B94848494
84849484849C7B7B9484849484849484849C84849484849484849C84849C84849484849C
84849C84849C84849484849C84849C8C8C9C84849C8C8C9C848C9C8C8C9C84849C8C8C9C
8C8C9C8C8C9C8C8C9C8C8C9C8C8C9C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000031293194949C8C949494949C8C949494949C8C949494949C8C9494
9494948C8C8C9494948C8C8C9494948C8C8C9494948C8C8C8C94948C8C8C9494948C8C8C
8C94948C8C8C9494948C8C8C8C948C8C8C8C8C948C8C8C8C8C948C8C8C8C8C948C8C8C84
31315A31315A31396331315A39396331316339396B31316B39396B7B7B9484849C7B7B94
84849C7B7B9484849C7B7B9484849C7B7B9484849C7B849484849C848494848C9C848494
848C9C8484948C8C9C8484948C8C9C8484948C8C9C84849C8C8C9C84849C8C8C9C84849C
8C8C9C84849C8C8CA5848C9C8C8C9C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008C949C8C94948C94948C94948C94948C94948C94948C94948C9494
8C94948C94948C94948C94948C94948C94948C94948C94948C948C8C94948C948C8C9494
8C948C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C
8C8C8C31315A31315A31316331316331396331316339396B425A427B84947B7B94848494
7B7B9484849484849484849C84849484849C84849484849C84849484849C84849C84849C
84849C84849C84849C84849C84849C848C9C84849C8C8C9C84849C8C8C9C848C9C8C8C9C
848C9C8C8C9C8C8C9C8C8C9C8C8C9C393963000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000003131398C9494949C9C8C94948C949C8C949494949C8C94948C94948C8C94
9494948C94948C94948C8C8C9494948C948C8C94948C8C8C9494948C8C8C8C94948C8C8C
8C94948C8C8C8C94948C8C8C8C94948C8C8C8C948C848C8C8C948C8C8C8C8C948C848C84
8C948C31315A31316331315A39396331316339396B31316384849C7B7B9484849C7B7B94
84849C7B7B9484849C7B7B9484849C7B849484849C7B7B9484849C84849484849C848494
8C8C9C848494848C9C8484948C8C9C84849C8C8C9C8484948C8C9C84849C8C8C9C84849C
8C8CA5848C9C8C8C9C84849C8C94A58C8C9C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008C94948C949C8C94948C949C8C94948C94948C94948C94948C94948C9494
8C94948C94948C94948C94948C94948C94948C948C8C94948C94948C94948C948C8C9494
8C948C8C94948C8C8C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C8C8C8C8C948C
8C948C8C948C31315A3131633131633139633131637B7B947B7B9484849C7B7B947B8494
7B849484849C7B7B9484849C84849C84849C84849484849C84849C84849C84849484849C
84849C848C9C84849C84849C84849C848C9C84849C848C9C848C9C8C8C9C848C9C8C8C9C
8C8C9C8C8C9C848C9C8C8C9C8C8C9C8C8CA5000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000029293194949C8C94948C949C8C949494949C8C94948C949C8C94948C949C8C9494
8C94948C94948C94948C94948C94948C8C8C8C94948C8C8C8C94948C8C8C8C94948C8C8C
8C94948C8C8C8C94948C8C8C8C9494848C8C8C94948C8C8C8C948C848C8C8C948C848C8C
8C948C848C848C948C31315A31396331315A84849C7B7B9484849C7B7B9484849C7B7B94
84849C7B7B9484849C7B7B9484849C7B7B9484849C7B849484849C7B8494848C9C848494
848C9C848494848C9C84849C848C9C84849C8C8C9C84849C8C8C9C84849C8C8CA584849C
8C8CA584849C8C8CA5848C9C8C8CA5848C9C8C94A5000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C949C8C94948C949C8C94948C949C8C94948C94948C94948C94948C94948C9494
8C94948C94948C94948C94948C94948C94948C94948C94948C94948C94948C948C8C9494
8C948C8C948C8C948C8C948C8C948C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C
8C8C8C8C948C8C8C8C31315A31315A7B7B947B7B947B84947B7B947B84947B7B947B849C
7B7B947B849C7B849484849C7B849484849C84849C84849C84849C84849C84849C84849C
84849C84849C84849C848C9C84849C848C9C84849C848C9C848C9C848C9C848C9C8C8C9C
848C9C8C8C9C848C9C8C8C9C8C8C9C8C8C9C8C8C9C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C94948C949C8C9494949C9C8C94948C949C8C949494949C8C94948C94948C9494
8C949C8C94948C9494848C8C8C94948C94948C9494848C8C8C94948C8C8C8C9494848C8C
8C94948C8C8C8C9494848C8C8C94948C8C8C8C948C848C8C8C94948C8C8C8C948C848C84
8C948C848C8C8C948C848C843939637B7B947B849C7B7B9484849C7B7B947B849C7B7B94
84849C7B7B9484849C7B7B9484849C7B849484849C7B8494848C9C7B849C84849C7B8494
848C9C84849C848C9C8484948C8CA584849C848C9C84849C8C8CA584849C8C8C9C84849C
8C8CA5848C9C8C8CA5848C9C8C94A5848C9C8C8CA5000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C949C8C949C8C94948C949C8C94948C949C8C94948C94948C94948C949C8C94948C9494
8C94948C94948C94948C94948C94948C94948C94948C94948C94948C94948C948C8C9494
8C948C8C94948C948C8C948C8C948C8C948C8C8C8C8C948C8C948C8C948C8C8C8C8C948C
8C948C8C948C848C8C6B738C848C7B7B849C7B7B947B7B947B7B947B849C7B7B947B849C
7B849C84849C7B84947B849C7B849C84849C7B849484849C84849C84849C84849C84849C
84849C848C9C84849C84849C84849C848C9C84849C848C9C848C9C848C9C848C9C848C9C
848C9C8C8CA5848C9C8C8C9C8C8C9C8C8CA5848C9C8C8CA5000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C9C9C8C94948C9C9C8C94948C949C8C94948C949C8C94948C949C8C94948C949C8C9494
8C94948494948C94948494948C9494848C8C8C94948C948C8C9494848C8C8C9494848C8C
8C9494848C8C8C94946363846B6B8C6363846B6B8C636B846B6B8C636B846B738C636B8C
6B738C6B6B8C6B73940000000000007B847B848C7B7B847B848C7B7B847384849C7B7B94
84849C7B7B9484849C7B849484849C7B849484849C7B849484849C7B8494848C9C7B849C
848C9C7B849C848C9C84849C848C9C84849C8C8CA584849C8C8CA584849C8C8CA584849C
8C8CA5848C9C8C8CA5848C9C8C8CA5848C9C8C94A5848C9C000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008C949C
8C949C8C949C8C949C8C949C8C94948C949C8C94948C949C8C94948C94948C94948C9494
8C94948C94948C94948C94945A637B636384636384636384636384636384636384636B84
636384636B84636B84636B8C636B84636B8C636B846B6B8C636B8C6B6B8C6B6B8C6B738C
6B6B8C0000000000000000000000000000007B8C7B7B8C7B7B847B7B847B7B847B7B847B
7B84737B84737B84737B84737B849C84849C7B849C84849C7B849C84849C84849C84849C
84849C84849C84849C848C9C84849C848C9C848C9C848C9C848C9C848C9C848C9C848C9C
848C9C8C8CA5848C9C8C8CA5848C9C8C8CA58C8C9C8C8CA58C8C9C000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000849494
8C9C9C8C94948C949C8494948C9C9C000000000000000000000000000000000000000000
6363845A5A7B6363845A5A7B6363845A637B6363845A637B636B84636384636B845A6384
636B8C636384636B8C6363846B6B8C636B846B6B8C636B846B738C636B8C6B738C636B8C
000000000000000000000000000000000000848C7B7B8473848C7B7B84737B8C7B738473
7B8C7B7B84737B8473737B737B84737384737B8473737B6B7B8473737B6B848C9C7B849C
848CA584849C848C9C7B849C848CA584849C848C9C84849C8C8CA5848C9C8C8CA584849C
8C8CA5848C9C8C8CA5848C9C8C94A5848C9C8C94A5848C9C8C94A5000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
5A637B5A63845A637B5A63845A63846363845A637B636384636384636B84636384636B84
636384636B84636384636B84636B846B6B8C636B846B6B8C636B8C6B6B8C636B8C000000
0000000000000000000000000000000000000000007B8C7B7B847B7B8C7B7B84737B847B
7B84737B847B7384737B84737B84737B8473738473738473738473738473737B6B737B6B
737B6B737B6B737B6B848C9C848C9C848C9C84849C848C9C848C9C848CA5848C9C848C9C
848C9C8C8CA5848C9C848CA5848CA58C8CA5848C9C8C8CA58C8CA5000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
6363845A5A7B6363845A637B6363845A637B636B845A637B636B845A6384636B84636384
636B846363846B6B8C6363846B6B8C6363846B6B8C636B846B6B8C636B84000000000000
0000000000000000000000000000000000000000000000007B8C7B7B84737B8C7B7B8473
7B847B7384737B847B7384737B8473737B6B7B8473737B6B7B8473737B6B7B8473737B6B
7384736B7B6B73846B00000000000000000000000000000000000084849C8C8CA5848C9C
8C8CA5848C9C8C8CA5848C9C8C94A5848C9C8C94A5848C9C8C94A5848C9C000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A637B5A637B5A63845A637B6363845A6384636384636384636384636384636B84
636384636B84636B84636B8C636B84636B8C636B8C6B6B8C636B8C000000000000000000
0000000000000000000000000000000000000000000000000000007B847B7B84737B847B
7B84737B84737B84737B84737384737B8473738473738473737B6B738473737B6B737B6B
737B6B737B6B000000000000000000000000000000000000000000000000000000000000
0000000000000000008C8CA5848CA58C8CA5848CA58C94A5848CA58C94A5000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A5A7B5A63845A5A7B6363845A637B6363845A637B636B845A6384636B845A6384
636B8C636384636B8C6363846B6B8C636B846B6B8C000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007B847B738473
7B8C7B7384737B8473737B737B8473737B737B8473737B6B7B8473737B6B7384736B7B6B
7384736B7B6B000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008C94A5848CA58C94A5000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A63845A637B5A637B5A63846363845A637B636384636384636384636384636384
636B84636B84636384636B84636B84636B8C000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000007B84737B8473
7B84737B8473738473738473738473738473737B6B738473738473738473737B6B737B6B
737B6B737B6B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A5A7B6363845A5A7B6363845A637B636B845A637B636B845A637B636B845A6384
636B8C636384636B8C636384636B8C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000738473
7B847B7384737B847B7384737B8473737B6B7B8473737B6B738473737B6B738473737B6B
73846B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000005A637B5A637B5A63845A637B5A63845A63846363845A6384636384636384636B84
636384636B84636B84636B8C000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
7384737B84737384737B8473738473738473737B73738473737B6B738473737B6B737B6B
737B6B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000005A63845A5A7B6363845A637B6363845A637B636B845A6384636B845A6384
636B8C636384423939000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000007384737B8473737B6B7B8473737B73738473737B6B738473737B6B7384736B7B6B
738473000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000005A637B5A637B5A637B6363845A637B5A63845A6384636B845A6384636384
636384000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000738473738473738473738473737B6B738473737B6B738473737B6B737B6B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000006363845A637B6363845A637B6363845A637B636B845A6384636B845A6384
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000007B8473737B73738473737B6B738473737B6B7384736B7B6B7384736B7B6B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000005A637B5A63845A637B5A63845A63846363845A6384636384636384000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000738473738473738473737B73738473737B6B737B6B737B6B737B6B
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005A5A7B6363845A637B6363845A637B636B845A6384000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000738473737B6B7384736B7B6B7384736B7B6B738473000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005A637B5A63845A63845A637B5A6384000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000738473737B6B737B73737B6B7384736B7B6B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005A5A7B5A63845A637B636384000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007384736B7B6B7384736B7B6B738473000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005A63845A637B5A6384000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000737B6B7384736B7B6B737B6B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005A5A7B636384000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000006B7B6B7384736B7B6B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000006B7B6B737B6B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[54]:=",
  ImageSize->{128, 128},
  AnimationDisplayTime->0.169,
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000200000P0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj2LW
9`00?g=cLda<C240<410Y/30`9T0IPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm?ooc020P800082006IVc3<cIP2009VIod:J
Yb8R8_mV<`00P00g??mVc5J0Zk75fBa1EkV>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7De3eDE:b/[:BT
Y>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEeMJ_0eLC4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7i/;2`W9bLP00
o`1o001oO`1oo`3o003oO`3oogl007l0Ogl0ogmo07moOgmoogoo07ooOgooool00?l0Ool0oomo0?mo
Oomooooo0?ooOoooo`0002XZ:UEEEGmoOjZZZ]GEeOoooh0700201`00P0L0080700031gT010L000<7
N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@7
00031gT010L000<7;0000i/0001:00@700031b`000>1X840BP041`000`L/0004Vh6KPDT010L000<7
:`02P@04X86PPDP010L000<7;0001Y^1Vh6KPAh000>OP@009P041`000`L[0008Vh6PPJ21X84L0003
X9ZP02L010L000<770000j2O000<0009PI^1Vh6KPI^101X000BOPIn19`041`000`LL0006X;VP^J2i
20002h6KPJ21X86PPJ2101P000FPVZ2JX00W00@700031a`000VPWj2OX9nPWj001@003I^1Vh6KPI^1
Vh6KPI/05@001inJWh6OPIl09P041`000`LL000<PKVP^J2iX;VP^J2i0P003X6KPJ21X86PPJ21X86P
4`002:2OX9ZPVZ2J9P041`000`LL000MPInPWj2OX9nPWj2OX9nPPI^1Vh6KPI^1Vh6PPI/04@002InJ
WiZOVYn1W`0V00@700031ad000niX;VP^J2iX;VP^J2iX;T00Z00386PPJ21X86PPJ2JX0h000^OX9nP
Wj2JX9ZPVP0V00@700031ad001hIX9nPWj2OX9nPWj2OX9nPWj2OX86KPI^1X86PPJ0<000=VYnJWiZO
VYnJWh6OP@0U00@700031ad001n1X;VP^J2iX;VP^J2iX;VP^J2iX;VP^H6PVZ2JX9ZP00X000jOX9nP
Wj2OX9ZPVZ2JX18000>P^@0040041`000`LM000P6J2PX9nPWj2OX9nPWj2OX9nPWj2OX9nPWj21X86P
PH87000@X9ZPVZ2JWiZOVYnJWh6OP@X000VPVZ2JX9^PVZ004P041`000`LM000RPGbiX;VP^J2iX;VP
^J2iX;VP^J2iX;VP^J2iX;VJX9^PO0<00160X9nPWj2OX9nPWj2OX9nPW`02X00B^J2iX;VP^J2iX;VP
^J2iX;V14P041`000`LN000WPAVPWj2OX9nPWj2OX9nPWj2OX9nPWj2OX9nPWj2OX9nPVZ2JP@2100:P
01nJX9ZPVZ2JX9^PVj2KX9^PVj2KX9^PVj2JX9ZPVZ2J01<010L000<77P008Wb1X;VP^J2iX;VP^J2i
X;VP^J2iX;VP^J2iX;VP^J2iX;T4P@03X:JP00JP00>iX:000j004[VP^J2iX;VP^J2iX;VP^J2iX1@0
10L000<77P00Ah4IX9nPWj2OX9nPWj2OX9nPWj2OX9nPWj2OX9nPWj2JX021086PXJ2QVZ2JX9ZPVZ2K
X9^PVj2KX9^PVj2KX9^PVZ2JX9ZP01@010L000<77P0087b1PKVP^J2iX;VP^J2iX;VP^J2iX;VP^J2i
X;VP^J2i18401:2VX:H;X00C^J2iX;VP^J2iX;VP^J2iX;VP^@0E00@700031al000dIPIniWj2O^InP
Wj2O00>P01ROX9nPWj2OX9nPWj2OX0210840XJ2QX:42X00AVZ2JX9ZPVj2KX9^PVj2KX9/00j001i^P
VZ2JX9X05@041`000`LO000MPGbiX;VP^J2iX;VP^J2iX;VP^J2iX;VP^J2iX;T01H401jJPYZ2VX:H0
1J000kVP^@03X00@^J2iX;VP^J2iX;VP^J2iX1H010L000<77`003AV16J2OX9nPWj2OX9l01J00:InP
Wj2OX9nPWj2OVP210:6PXJ2QX:6PXIZPVZ2JX9^PVj2KX9^PVj2K00FP00BJX9ZP5P041`000`LO000M
PGb1X;VP^J2iX;VP^J2iX;VP^J2iX;VP^J2iX;T00h402JJPYZ2VX:JPYP05X00E^J2PX;VP^J2iX;VP
^J2iX;VP^J2i01L010L000<77`003AV16KVOX9niWj2O^Il00j006YnPX:2OX9nPWj2OX0210:6PXJ2Q
X:6PXJ2Q0Z002iZPVZ2KX9^PVj2K00>P00>KX:000j000i^PV`0G00@700031b0001YlPH6iX;VP^J2i
X;VP^J2iX;VP^J2iX;VP^@>100fVX:JPYZ2VX:JPYZ2V00>P01BiX;VP^J2iX;VP^J2iX;VP^J2iX1P0
10L000<78000384IPIniWj2O^InPW`VP00NOX9nPWh4000:P01JQX:6PXJ2QX:6PXJ2QVj2JX9^PVj2K
2j000i^P000G00@700031b0001UlPGbiX;VP^J2iX;VP^J2iX;VP^J2iX;VP00>102BPYZ2VX:JPYZ2V
X:JPYZ2iX;VP^J2iX;VP^J2iX;VP^J2iX;TI00@700031b0000N16H4IX9ni00:P00>O^Il02j004Yn1
086PXJ2QX:6PXJ2QX:6PX@BP00FKX9^PV`0=X003V`0001L010L000<780005aV1PH6P^J2iX;VP^J2i
X;VP^J2iX;VP00>102JPXJ2VX:JPYZ2VX:JPYZ2VX:JP^J2iX;VP^J2iX;VP^J2iX;VPPAT010L000<7
8@002aV16KVP^J2iX;VO00ZP00>108400j002j6PXJ2QX:6PXJ2Q00>P00JQX:2KX9/?X003O00001L0
10L000<78002P@0DO861^J2iX;VP^J2iX;VP^J2iX;T2P@0WX:6PXJ2VX:JPYZ2VX:JPYZ2VX:JP^J2i
X;VP^J2iX;VP^J2iX;VP01X010L000<77P003IYkVW^16H6P^J2iWkT02Z000`21X004X005XHNPQj40
0Z001XNQX:27XAJP6P041`000`LI0009N`21PIZ1VX6J00:100ilX:2iX;VP^J2iX;VP^@:P02Z1X:2V
X:JPYZ2VX:JPYZ2VX:JPYZ2VX:JPXJ2iX;VP^J2iX;VP^J2iX;TK00@700031aP000jJNiYkVW^JNiYk
VQV16@fP00T0PHNPQj27X8L04j001i^PVj2KX9/01j0L00@700031aH000j1VX6JPIZ1VX6JPIZ1VP:1
00bP^J2iX;VP^J2iX;T2P@0YX:JPYZ2VX:JPXJ2QX:6PXJ2QX:6PXJ2QX:JP^J2iX;VP^J2iX;VP^J00
70041`000`LD000CVW^JNiYkVW^JNiYkVW^JNh4IP@07X00>^J2P0867X8NPQj27X8L3X005Qj2PX8L0
2j000i^PV`03X003Vj2P00BP7@041`000`LB000BPIZ1VX6JPIZ1VX6JPIZ1VX6J0h40=kVP^J2nX;VP
^Gb1X:6PXJ2VX:JPYZ2QX:6PXJ2QX:6PXJ2QX:6PYZ2VX;VP^J2iX;VP^J2iPIX06`041`000`L@000H
VW^JNiYkVW^JNiYkVW^JNiZ0VW^JP84I1j005`21Qj27X8NPQj27X8NPQj27X8NPQj2700VP00>7X:00
2J00182JP9XJ00@700031`h001R1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIX2P@0hX;VP_Z2nX86PXJ2Q
X:6PXJ2VX:6PXJ2QX:6PXJ2QX:6PXJ2QX:JPYZ2iX;VP^J2iX;VPVX6JPIXI00@700031``001bJNiYk
VW^JNiYkVW^JNiZ1VW^JPIZ0VX6JP9XI1Z006X67X8NPQj27X8NPQj27X8NPQj27X8NPQj270j001HNP
Qj2700ZP00RJP9Z0VX2JP1L010L000<72@007Kn1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6J00>1
03JiX;jPPJ2QX:6PXJ2QX:6PXJ2VX:6PXJ2QX:6PXJ2QX:6PXJ2VX:JPYZ2iX;VP^J2iPIZ1VX44VQH0
10L000<720007iYkVW^JNiYkVW^JNiYkVW^JNiZ1VX6JPIZ1VX6JPIX00X43X00VPHNPQj27X8NPQj27
X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8L7X00<P9Z0VX2JP9Z0VX2J50041`000`L6000OPIZ1VX6J
PIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JP@04VP0`^J2nX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6P
XJ2QX:6PYZ2VX:JP^Z2iX;ZP1IX00h6JVP06VQ8010L000<710009YYlVW^JNiYkVW^JNiYkVW^JNiZ1
VW^JPIZ1VX6JPIZ1VX6JPIXI0j009HNPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8L0
1Z0049Z0VX2JP9Z0VX2JP9Z0VX0A00@700031`@001^2Qh:7PZ22X8:PPZ22X8:PPIZ1VX6JPIZ1VX40
2iX0;Z2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PYZ2VX:JP^Z2jX;XBVPl010L000<7
1P00CXN2Qh:7PXN2Qh:7PXN2Qh:7PXN2Qh:7PZ22X8:PPZ22X8:PPYZ7X8NPQj27X8NPQj27X8NPQj27
X8NPQj27X8NPQj27X8NPQj27X8NPQ`>P01B0VX2JP9Z0VX2JP9Z0VX2JP9Z0VPh010L000<71`008XN2
X8:PPZ22X8:PPZ22X8:QQj27X8NPQj27X8NPQj27XHL2X00<WiZPVZ2JX9ZPVZ2J0Z007j6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PYZ2VX:JPYZ2jX9l05IX<00@700031`P00227PXN2Qh:7PXN2Qh:7PXN2Qh:7PXN2
Qh:7PZ22X8:PPP>P00>1VYX05iX03hNPQj27X8NPQj27X8NPQ`02X00IVX2JP9Z0VX2JP9Z0VX2JP9Z0
VX2JP9Z0VP0:00@700031`P0028IX8:PPZ22X8:PPZ22X8:PPZ27X8NPQj27X8NPQj27X8No0Z00<YZP
VZ2JX9ZPVZ2JX9ZPVZ2JWiZOVYnJWiZOVYnJWiZOVYnJWiZVX:JP^YZOVYnJWiZO4YX900@700031`X0
02B1PXN2Qh:7PXN2Qh:7PXN2Qh:7PXN2Qh:7PZ22X86iX;VPPX4VVP>P01b7X8NPQj21X86PPJ21X82J
P9Z0VX2JP9Z0VP201P041`000`L<001:PXN2X8:7PZ22X8:PQj27X8:PQj27X8NPQj27_j2oX;n1PX6P
VZ2JWiZPVYnJX9ZOVZ2JX9ZPVZ2JX9ZPVZ2JWiZOVYnJWiZjX;X2X00FQj27X8NPQj27X8NPQj27X8NP
Qj27X0JJ1@041`000`L=000LPXN2Qh:7PXN2Qh:7PXN2Qh:PPZ22X8:PPX6P^@BP00<IPYX03YX03inJ
WiZOVYnJWiZOVYnJW`06VPBP01j7X8NPQj27X8JPQZ26X8JPQZ26X8JPQZ21X8JPQZ0400@700031`h0
01N2Qh:7PXN2X8:PPZ22X8:PQj27X8NPQ`02X00__j2oX;nPPX62VZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2J
X9ZPVZ2JX9ZOVYnJWj2jX;X00Z0078NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj0500@700031a00
01N7PXN2Qh:7PXN2Qh:7PXN2Qh:PPZ00^@03X007^J2PX88IPP0=VP0?WiZOVYnJWiZOVYnJWiZO00>J
00>OX:001Z006HNPQj27X8JPPJ26X86PQZ21X86PPJ21X8H01`041`000`LA000nQh:7PXN2X8:7PZ22
X8NPPZ27X7boX;nP_j2oX;nPPX62VZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ2JWiX2X00P^Z2jX;ZP
^XNPQj27X8NPQj27X8NPQj27X8NPQj27X6X800@700031a8001:7PXN2Qh:7PXN2Qh:7PXN2Qh8;X006
6H8IPYZO0iX069nJVYZOVYnJWiZOVYnJWiZOVYnJWiZOVPVP01J7X8NPQj21X86PPJ21X86PPJ21X86P
2P041`000`LC0015Qh:7PXN2Qh:PPZ22X8:PPX6P_j2oX;nP_j2oX;n1PX62VZ2JX9ZPVZ2JX9ZPVZ2J
X9ZPVZ2JX9ZPVZ2J^Z2jX;ZP^Z2j00:P01B7X8NPQj27X8NPQj27X8NPQj27X0/010L000<750003hN2
Qh:7PXN2Qh:7PXN2Q`0=X00NPQV26H:JWiZOVYnJWiZOVYnJWiZOVYnJWiZOVYnJ2j004hNPQj27X86P
PJ21X86PPJ21X8403@041`000`LF001DPXN2Qh:PPZ22X8:PPKnP_j2oX;nP_j2oX;nPPX62PH:JX9ZP
VZ2JX9ZPVZ2JX9ZPVZ2JX9ZPV[VP^Z2jX;ZP^Z2jX;Z7X8NPQj27X8NPQj27X8NP3`041`000`LG000;
PXN2Qh:7PXN2Qh803j007AV26H8IPYZOVYnJWiZPVYnJX9ZOVZ2JWiZOVYn100fP00f7X8NPPJ21X86P
PJ21014010L000<760007H:7PXN2X8:PPX6P_j2oX;nP_j2oX;nP_j2oPH:100:2032PVZ2JX9ZPV[VJ
X9ZiV[VJ^IZiV[VJ^Z2QX;ZP^Z2jX;ZP^XNPQj27X8NPQj27X8LB00@700031aT000R1Qh:7PHN2PA6P
01^26H8IPQVOVYnJWiZPVZ2JX9ZPVZ2JX9ZOVZ403Z002XJPPHN1X86PPJ0D00@700031aT001j1PXN2
Qh:PPKnP_j2oX;nP_j2oX;nP_j2oX8:1PX42PP0QX9ZPV[VJ^IZiV[VJ^IZiV[VJXJ2jX;ZPXJ2jX;ZP
^Z2j00:P00N7X8NPQj2701H010L000<76`001H67PHL001>P01LIPQV26H8IWiZOVZ2JX9ZPVZ2JX9ZP
VP0BX005Qh67PJ0060041`000`LL000MPXN2PJ2oX;nP_j2oX;nP_j2oX;nP_j2oPH:1PX400X809KVJ
^IZiV[VJ^IZiV[VJ^J2QX:6PXJ2QX;ZP^Z2jX;ZP^XNPQj006@041`000`LM00:15J00588IPQV26H8I
WiZPVZ2JX9ZPVZ2J1:000j6PX00=X003Qh6701X010L000<77@02P@0L_j2oX;nP_j2oX;nP_j2oX;nP
_j2oX8:1PX62P@:2022iV[VJ^IZiV[VJ^J2QX:6P^Z2jX;nP_j2oX;nP_j2oX1`010L000<77P0HX00A
PQV26H8IPYZPVZ2JX9ZPVXH01j003j6PXJ2QX:6PXJ2QX:6PX@0L00@700031a`004:1_j2oX;nP_j2o
X;nP_j2oX;nP_j2oX;nP_h62PH:1PX62V[VJ^IZiV[VPXJ2QX:6PXJ2oX;nP_j2oX;nP_j2oX;lK00@7
00031a`000>1X:006:003AV26H8IPQV2VZ2JX9X03:003Z6PXJ2QX:6PXJ2QX:6P6`041`000`LK000R
PJ2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX8:1PX62P@>201nJ^IniX:6PXJ2QX:6P^Z2jX;nP_j2o
X;nP_j2oX;nP01X010L000<7700LX00<PQV26H8IPQV2VZ212Z004J6PXJ2QX:6PXJ2QX:6PXJ2Q01X0
10L000<76P008h6oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_h62PH:100F201fOXJ2QX:6PXJ2Q
X:6P_j2oX;nP_j2oX;nP_j2oX00J00@700031aX07j002AV26H8IPQV26@0:X00DXJ2QX:6PXJ2QX:6P
XJ2QX:6PXH4I00@700031aT002B1X;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX8:1PX44PP0O
XJ2QX:6PXJ2QX:6PYZ2VX;nP_j2oX;nP_j2oX;nPP@0I00@700031aT000>1X:007Z001h8IPQV26HH0
3J004Z6PXJ2QX:6PXJ2QX:6PXJ2QPQT010L000<76@02X00R_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP
_j2oX;nP_h62P@>2022PXJ2QX:6PXJ2QX:6PXJ2oX:JP_j2oX;nP_j2oX;nP_aT010L000<76@0RX004
6H8IPPVP01RQX:2PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ0I00@700031aP004TI_j2oX;nP_j2oX;nP_j2o
X;nP_j2oX;nP_j2oX;nP_j2oX;n1PZ2QX:6PXJ2QX:6PXJ2QX:JPYZ2VX:JP_j2oX;nP_j2oX;nP01P0
10L000<760000j6PX00QX003PZ2P00^P01NQX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PX@0H00@700031aP0
04VP_j2oX;nP_j2oX;nPYZ2oX:JP_j2oX;nP_j2oX;nP_j2oX;n2XJ2QX:6PXJ2QX:6PXJ2QX:6PYZ2V
X;nPYZ2oX;nP_j2oX;nP01P010L000<75`000j2QX00ZX005XJ2PX:400j005Z6PXJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ0G00@700031aL002JoX;nP_j2oX;nP_j2VX:JPYZ2oX;nP_j2oPZ62XHNQQj67XJ2Q00JP
01nQX:6PXJ2QX:6PYZ2VX:JPYZ2oX;nP_j2oX;nP_j2V01L010L000<75P000j6PX@0>X00CPZ22X8:P
PZ22X8:PPZ22X8:PPP0400NP01jOX9nQX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:4F00@700031aH0
1H47000GPZ22X8:PPZ22X8:QPZ62XHNQQj67XH401@00186UX:D=X00DXJ2PX:JPYZ2oX;nP_j2oX;nP
_j0F00@700031b8001FPPZ22X8:PPZ22X8:PPZ22X8:PPZ002@004Z2OX9nPQZ26X8JPQZ26X8JPQPH0
00bQX:6PXJ2QX:6PXJ0E00@700031b<001>PPZ22X8:PPZ27X8NPQj27XJ2Q00/04:0=0006X;nP_j2o
5@041`000`LS000APZ22X8:PPZ22X8:PPZ22X8803P003j26X8JPQZ26X8JPQZ26X00A0003XJ2Q01@0
10L000<78`004:22X8:PPZ22X8:PQj27X8L?00nP:0041`000`LS000?PZ22X8:PPZ22X8:PPZ220140
00f6X8JPQZ26X8JPQZ2602T010L000<78`003Z22X8:PPZ22X8NPQj274`0<X2T010L000<790003:22
X8:PPZ22X8:PNaD000^6X8JPQZ26X8JPQP0Y00@700031b@000Z2X8:PPZ22X8:P6009X2X010L000<7
90002J22X8:PPZ22X00I0009X8JPQZ26X8JP02X010L000<7900028:PPZ22X8:P6`08X2X010L000<7
9@001h:PPZ22X1T070001Z26X8JPQR/010L000<79@001J22X8:P01l01J0[00@700031bD000B2X8:P
8@001:26X8H[00@700031bD000>PPZ008P03X003P@0002T010L000<79@000h:P000S0003QZ0002/0
10L000<7C0000j21000Z00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7N@041`00
0`Mi00@700031gT010L000<7N@041`000`Mi00@700201`00P0L008070000\
\>"],
  ImageRangeCache->{{{0, 127}, {127, 0}} -> {-3.20067, -3.20067, 1.05828, 
  1.05828}},
  CellTags->"Animation"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
%%ImageSize: 128 128 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00744048 0.0238095 0.00744048 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
384 string
128 128 8 [128 0 0 128 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000846B5A7B635A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000846B5A7B635A846B5A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000846B5A7B635A846B637B635A846B5A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000846B5A7B635A846B5A846B5A846B5A7B635A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B635A
846B637B635A846B637B635A846B637B635A846B637B635A000000000000000000000000
000000000000000000000000000000735A737B63737B6373000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000846B5A846B63
846B5A846B5A846B5A846B5A846B5A846B5A846B5A846B5A000000000000000000000000
0000000000000000007B63737B63737B63737B63736B6342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000846B5A846B637B635A
846B63846B5A846B637B635A846B637B6B5A846B637B635A846B63000000000000000000
000000735A737B6373735A737B637B7B63737B6373636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A3952000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000846B63846B63846B5A846B63
846B63846B63846B5A846B5A846B5A846B63846B5A846B5A846B5A000000000000736373
7363737B63737363737B63737B63737B63737B63736B6342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004239524A4252000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000846B5A846B63846B5A846B63846B5A
846B63846B5A846B63846B5A846B637B635A846B637B6B5A7B6373735A6B7B6373735A73
7B6373735A737B63737363737B637B7363737B6B7B636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A42524A39524A42524A4252000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000846B63846B63846B63846B63846B63846B63
846B63846B63846B5A846B63846B5A846B63735A73736373735A737363737363737B6373
7363737B63737B63737B63737B63737B63736B63426B6342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004239524A425A4A39524A42524A3952000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000846B638C7363846B5A8C7363846B63846B63846B5A
8C7363846B5A846B63846B5A736373735A6B736373735A6B7B6373735A737B6373735A73
7B637B7363737B63737363737B6B7B7B63736B6B4A636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004239524A42524A42524A42524A39524A42524A4252000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000846B638C7363846B63846B63846B63846B63846B63846B63
846B63736373735A6B735A73735A73736373735A737363737363737B6373736373736373
7B63737B63737363737B63737B63737B637B636342636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A42524239524A42524A39524A425A4A395252425A4A395252425A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000846B638C7363846B638C7363846B638C7363846B638C73636B5A6B
7363736B5A6B736373735A6B736373735A6B7B6373735A737B6373735A737B637B735A73
7B637B7363737B637B7363737B637B6363426B6B4A636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A39524A42524A39524A42524A42524A425A4A42524A425A4A425A52425A
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000846B638C7363846B638C7363846B638C7363846B63735A736B5A73735A73
6B5A73735A73735A73736373735A737363737363737363737363737363737363737B6373
7363737B63737B63737B637B7B63736B6B4A636342636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A42524239524A425A4A39524A425A4A395252425A4A425252425A4A4252
524A5A4A425A000000000000000000000000000000000000000000000000000000000000
000000846B638C736B846B638C736B846B63735A736B5A6B7363736B5A6B735A736B5A6B
7363736B5A737363736B5A6B736373735A73736373735A737B637B7363737B637B735A73
7B637B7363737B637B7363737B6B7B6363426B634A635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A39524A42524A42524A425A4A42524A425A4A425A52425A4A425A52425A
52425A52425A52425A000000000000000000000000000000000000000000000000000000
8C73638C736B846B638C73636B5A6B6B5A736B5A6B6B5A736B5A73735A736B5A6B735A73
735A73736373735A73736373736373736373735A737363737363737B63737363737B6373
7B63737B637B7363737B637B7B637B6B6B4A636342636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A425A4239524A425A4A395252425A4A425252425A4A425252425A4A4252
524A5A4A425A524A6352425A000000000000000000000000000000000000000000846B63
8C736B846B636B5A736B5A6B6B5A736B5A6B7363736B5A6B7363736B5A6B7363736B5A6B
7363736B5A73736373735A73736373735A737B637B735A737B637B735A737B637B736373
7B637B7363737B6B7B7363736B6B4A6363426B634A635A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A42524A39524A42524A42524A425A4A42524A425A4A42524A425A4A425A52425A
4A425A52425A52425A524A5A52425A524A5A0000000000000000000000008C736B8C736B
6B5A6B6B5A6B6B5A6B6B5A736B5A6B6B5A736B5A736B5A736B5A73735A736B5A73736373
735A73736373735A73736373735A737363737363737363737363737363737363737B637B
7363737B637B7B63737B637B636342636342636342636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004239524A425A4239524A425A4A39524A425A4A395252425A4A425A52425A4A4252
524A5A4A425A524A5A4A425A524A6352425A524A6300000000000063526B6B5A7363526B
6B5A736B5A6B6B5A736B5A6B7363736B5A6B735A736B5A6B7363736B5A737363736B5A6B
7363736B5A737363736B5A7373637B735A7373637B735A737B637B7363737B637B736373
7B6B7B7363737B637B7363736B6B4A6363426363425A5A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008473738473738C737B8473738473738473738C737B8473738C7373
8C73738C737B8473738C73738C73738C737B8C73738C73738C73738C73738C73738C7373
8C73738C73738C73738C73738C73738C73738C73738C73738C7373734A396B42316B4231
6B42316B42316342316342316342316342315A39315A4231635A6B6B5A73635A6B6B5A6B
6B5A6B6B5A736B5A6B6B5A736B5A736B5A736B5A736B5A736B5A737363736B5A73735A73
735A73736373735A737363737363737363737363737363737363737B637B7363737B637B
7B637B7B637B7B63737B637B63634263634A636342636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000846B738C737B846B738C737B846B738C737B846B738C737B847373
8C737B846B738C737B8473738C737B8473738C737B8473738C737B8473738C737B8C7373
8C737B8C73738C7B7B8C73738C737B8C7373947B7B8C7373947B7B7384947B8C9C6B4231
6B4A396342316B4A396342316B42316339316342315A393163423163526B6B5A73635A6B
6B5A73635A6B6B5A736B5A6B6B5A736B5A6B7363736B5A6B7363736B5A737363736B5A73
7363736B5A7373637B735A7373637B735A737B637B7363737B637B7363737B637B736373
7B6B7B7363737B6B7B6363426B6B4A63634263634A5A5A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008473738C737B8473738C737B8473738C737B8C73738C737B
8473738C737B8C73738C737B8C73738C737B8C73738C737B8C73738C73738C73738C737B
8C73738C73738C73738C73738C73738C73738C73738C737373849C7B849C73849C7B849C
6B42316B4A316B42316B42316342316342316342316342315A42315A42316B5A6B6B5A73
6B5A6B6B5A736B5A736B5A736B5A736B5A736B5A73735A736B5A73736373735A73736373
735A7373637373637373637373637373637373637373637B7363737B637B73637B7B637B
73637B7B637B7B637B6B6B4A636342636342636342636342000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000008C737B8473738C737B846B738C737B8473738C737B846B73
8C737B8473738C737B8473738C737B8C73738C737B8473738C7B7B8C73738C737B8C7373
8C7B7B8C73738C737B8C7373947B7B8C7373947B7B8C73737B849C7384947B849C738494
7B8C9C6B42316B4A396342316B4A396342316342315A39296342315A393163423163526B
6B5A736B5A6B6B5A736B5A6B7363736B5A737363736B5A737363736B5A737363736B5A73
73637B735A7373637B735A737B637B73637373637B735A737B6B7B7363737B637B736373
7B6B7B7363737B6B7B6363426B6B4A6363426363425A5A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000008C737B8473738C737B8C737B8C737B8473738C737B
8C737B8C737B8C73738C737B8C73738C737B8C73738C737B8C737B8C737B8C73738C737B
8C73738C737B8C73738C73738C73738C7B7B8C73738C737373849C7B849C73849473849C
7B849C7B849C6B42316B4A316B42316B42316342316342316342316342315A42316B5A73
6B5A736B5A736B5A736B5A736B5A736B63736B5A736B5A736B5A737363736B5A73736373
73637373637B73637373637373637373637B73637373637B73637B7B637B73637B7B637B
7B637B7B6B7B63634263634A6363426363425A5A425A5A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000008C737B8473738C737B8473738C737B847373
8C737B8473738C7B7B8C73738C737B8473738C7B7B8C73738C7B7B8C7373947B7B8C7373
8C7B7B8C7373947B7B8C7373947B7B8C7373947B7B7384947B849C7384947B849C738494
7B849C7384947B8C9C6B4231734A396B42316B4A396342316B42316342316342315A3931
6B5A736B5A6B6B63736B5A737363736B5A737363736B5A737363736B5A7373637B6B5A73
73637B6B5A7373637B735A7373637B735A737B637B7363737B637B7363737B6B7B736373
7B6B7B7363736B6B4A636342636B4A5A63426363425A5A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000008C737B8C737B8C737B8C737B8C737B8C737B
8C737B8C737B8C737B8C737B8C73738C737B8C737B8C737B8C737B8C737B8C737B8C737B
8C73738C737B8C73738C7B7B8C73738C7B7B8C737373849C73849473849C73849473849C
7384947B849C73849C7B849C7384946B4A316B42316B4231634231634231634231634231
5A42316B5A736B5A736B5A736B5A736B5A736B5A737363736B5A73736373735A73736373
73637373637B73637373637B73637373637B73637B73637B73637B7B637B73637B7B637B
7B637B7B637B63634263634A6363426363425A63425A6342422929000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008473738C7B7B8C737B8C737B847373
8C7B7B8C73738C737B8473738C7B7B8C73738C7B7B8C7373947B7B8C73738C7B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C7373947B7B737B947B849C7384947B849C738494
7B8C9C7384947B849C7384947B8C9C738494734A396B42316B4A396342316B4231634231
6342315A42316B63736B5A7373637B6B5A737363736B5A7373637B6B5A7373637B6B5A73
73637B735A7373637B735A737B637B7363737B637B7363737B6B7B73637B7B6B7B736373
7B6B7B7B637B6B6B4A636342636B4A5A5A426363425A5A394A3131000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000008C737B8C737B8C737B8C737B8C737B
8C737B8C737B8C737B8C737B8C737B8C7B7B8C737B8C737B8C737B8C7B7B8C737B8C737B
8C737B947B7B8C737B8C7B7B8C737B947B7B8C737373849C73849473849C73849473849C
73849473849C73849473849C73849C7B849C7384946B4A316B42316B4A31634231634231
6342316342315A42316B5A736B5A737363736B5A736B63736B637373637B6B5A73736373
73637373637B73637373637B73637B73637B73637B73637B73637B7B637B73637B7B637B
7B637B6B6B4A6363426363426363426363425A5A425A5A424A2929000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008C7B7B84737B8C7B7B8C737B
8C7B7B8C73738C7B7B8C737B8C7B7B8C7373947B7B8C737B947B7B8C7373947B7B8C7373
947B7B8C7373947B7B8C7373947B7B8C737373849C737B9473849C737B947B849C738494
7B849C7384947B849C7384947B849C7384947B8C9C738494734A396B42316B4A39634231
6B4A316342316342315A42317363736B5A7373637B6B5A7373637B6B5A7373637B6B5A73
73637B6B5A7373637B73637373637B7363737B6B7B73637B7B6B7B73637B7B6B7B73637B
7B6B7B6363426B6B4A63634263634A5A5A426363425A5A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000008C737B8C737B8C737B8C737B
8C737B8C737B8C737B8C7B7B8C737B8C737B8C737B8C7B7B8C737B8C7B7B8C737B8C7B7B
8C737B8C7B7B8C737B947B7B8C737B947B7B73849473849C73849473849C73849473849C
73849473849C73849473849C73849473849C7384947B849C7384946B4A316B42316B4A31
6342316342316342316342316B5A736B63736B5A7373637B6B5A7373637B73637373637B
73637373637B73637B73637B73637B73637B73637B7B637B73637B7B637B73637B7B6B7B
7B637B6B6B4A63634263634A5A63425A63425A5A425A5A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000008C737B8C7B7B8C737B
947B848C737B8C7B7B8C737B947B7B8C737B8C7B7B8C7373947B7B8C737B947B7B8C7373
947B7B8C737B947B7B8C7373947B7B8C737B73849C6B7B9473849C737B9473849C737B94
7B849C73849473849C7384947B849C7384947B849C7384947B8C9C7384947B8C9C6B4231
6B4A396342316B4231634231634A316B5A7373637B6B5A7373637B6B5A7373637B6B5A73
73637B73637373637B7363737B6B7B73637B7B637B7363737B6B8473637B7B6B7B73637B
7B6B84636342636B4A5A634263634A5A5A425A6342525A39000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C737B8C737B
8C737B8C7B7B8C737B8C7B7B8C737B8C7B7B8C737B8C7B7B8C7B7B947B7B8C737B8C7B7B
8C7B7B947B7B8C737B947B7B947B7B73849C737B9473849473849473849C738494738494
73849473849C73849473849473849473849C7384947384947384947B849C738494738494
6B42316B4A316342316342316342316342316B5A7373637B73637373637B6B637373637B
73637B73637B73637B73637B73637B73637B73637B73637B73637B7B6B7B73637B7B6B7B
6B6B4A6B6B4A6363426363425A63425A63425A5A42633939633939634239000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C7B848C737B
947B848C737B947B848C737B947B7B8C737B947B7B8C737B947B7B8C737B947B848C737B
947B7B8C737B947B7B8C737B947B7B6B7B9473849C6B7B9473849C6B7B9473849C737B94
73849C737B9473849C73849473849C73849473849C7384947B849C7384947B8C9C738494
7B8C9C6B42316B4A396342316B4A316342316342316B5A7373637B6B5A7373637B6B6373
73637B73637373637B73637B7B637B73637B7B6B7B73637B7B6B8473637B7B6B8473637B
6B6B4A636342636B4A5A63426363425A5A425A63426339396342395A3931634239633931
736B7B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008C7B7B
8C737B8C7B7B8C737B8C7B7B8C737B8C7B7B8C737B947B7B8C737B947B7B8C7B7B947B7B
8C7B7B947B7B8C7B7B947B7B947B7B738494737B94738494737B94738494738494738494
73849473849473849473849473849473849473849473849473849473849473849473849C
7384947384946B42316B4A316B42316B42316342316342316B637373637B73637B73637B
73637B73637B73637B73637B73637B73637B73637B7B637B73637B7B6B7B7B637B7B6B84
636B4A636B4A6363426363425A63425A6342633939634239633939634239633939634239
633939634239634239000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008C737B
947B848C737B947B848C737B947B848C737B947B848C737B947B848C737B947B7B8C737B
947B848C737B947B7B8C737B947B846B7B947384946B7B9473849C6B7B947384946B7B94
73849C737B9473849C6B7B9473849C73849473849C7384947B849C73849473849C738494
7B8C9C7384947B8C9C6B42316B4A396342316B4A3163423173637B6B637373637B6B5A73
73637B73637B73637B73637B7B6B8473637B7B6B7B73637B7B6B8473637B7B6B84636342
6B6B4A63634263634A5A5A4263634A6339396342396339316B4239633939634239633931
6B42396339396B42396339316B4239000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C7B7B947B848C737B8C7B7B8C7B7B947B848C737B947B7B8C7B7B947B7B8C7B7B947B7B
947B7B947B7B8C7B7B947B7B6B7B947384946B7B94738494737B947384946B7B94738494
738494738494738494738494738494738494738494738494738494738494738494738494
7384947384947384947384947384946B4A316342316B423163423173637B73637B73637B
73637B73637B73637B73637B73637B7B6B7B73637B73637B73637B7B6B8473637B6B6B4A
636B4A636B4A6363425A63425A63426342396339396342396342396B4239634239634239
6342396B42396342396342396342396B42396342396B4239000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
947B848C737B947B848C737B947B848C737B947B848C737B947B848C737B947B848C737B
947B848C737B947B848C7B7B7384946B7B947384946B7B947384946B7B9473849C6B7B94
7384946B7B9473849C6B7B9473849C6B7B9473849C73849473849C73849473849C738494
73849C7384947B8C9C7384947B8C9C7384946B4A396342316B4A3163423173637B73637B
73637B73637B736B7B73637B7B6B8473637B7B6B8473637B7B6B8473637B7B6B84636342
6B6B4A63634263634A5A5A426B42396339396B42396339396B42396339396B4239633939
6B42396339396B42396339396B42396339396B42396342396B4239633939000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000947B848C7B7B947B848C7B7B947B848C7B7B947B848C7B7B947B7B947B7B947B84
947B7B947B7B947B7B947B846B7B946B7B946B7B947384946B7B947384946B7B94738494
6B7B94738494738494738494738494738494738494738494738494738494738494738494
7384947384947384947384947384947384947384946B4A316B42316B423163423173637B
73637B73637B73637B73637B73637B73637B73637B7B6B8473637B7B6B847B6B7B6B6B4A
63634263634A5A63426B42396342396B42396342396B42396342396B42396B42396B4239
6342396B42396B42396B42396B42396B42396B42396B42396B42396B42396B42396B4239
736B84000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000947B848C737B947B848C737B947B848C737B947B848C7B7B947B848C737B
947B848C7B7B947B846B7B8C7384946B7B947384946B7B947384946B7B947384946B7B94
73849C6B7B947384946B7B9473849C6B7B947384946B7B9473849C7384947384946B8494
73849C7384947384946B84947B8C9C738494738C947384946B4A396342316B4A31634231
736B8473637B736B7B73637B7B6B8473637B7B6B8473637B7B6B8473637B6B6B4A636342
636B4A5A634263634A6339396B42426342396B42426339396B42426342396B4239633939
6B42426B42396B42396339396B42426B42396B42396339396B42426B42396B42396B4239
6B42426B42396B4239000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008C7B7B947B84947B84947B848C7B7B947B84947B84947B84947B7B947B84
947B7B947B84947B7B6B7B946B7B946B7B946B7B946B7B946B7B947384946B7B946B7B94
6B84947384946B7B947384947384947384946B7B94738494738494738494738494738494
7384947384947384947384947384947384947384947384947384946B4A316B42316B4231
63423173637B73637B73637B73637B7B6B8473637B7B6B7B7B6B847B6B84636B4A636B4A
63634A63634A6B42396B42396B42396B42426B42396B42396B42396B42396B42396B4239
6B42396B42396B42396B42396B42396B42396B42396B42396B42396B42396B42396B4239
6B42396B42396B42396B42396B42397B738C000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008C7B7B947B848C737B947B848C7B7B947B848C7B7B947B84947B7B
947B848C7B7B947B846B7B946B84946B7B947384946B7B947384946B7B8C7384946B7B94
7384946B7B947384946B7B947384946B7B947384946B7B947384946B7B947384946B8494
7384946B84947384946B8494738494738494738C94738494738C94738494738C94634231
6B4A3973637B7B6B8473637B7B6B8473637B7B6B8473637B7B6B8473637B6B6B4A636342
636B4A6342396B42426B42396B42426B42396B42426B42396B42426B42396B42426B4239
6B42426B4239734A426B42396B42426B42397342426B42397342426B4239734A426B4239
7342426B4239734A426B4239734A426B4239734A426B4239000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000947B84947B84947B84947B84947B84947B84947B84947B84947B84
947B7B947B84947B846B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B8494
6B7B946B84946B7B947384946B7B947384946B84947384946B8494738494738494738494
738494738494738494738494738494738494738494738494738494738494738494738494
6B42316B4A3173637B736B8473637B736B8473637B7B6B84736B847B6B84636B4A636B4A
6B42396B42396B42396B42426B42396B42396B42396B42426B42396B42396B42396B4239
6B42396B42396B42396B42426B42396B42396B42396B42396B42396B42396B4239734239
6B42396B42396B42397342396B42397342396B42397342396B4239734239000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000947B848C7B7B947B848C7B7B948484947B7B947B848C7B7B
948484947B7B6B7B94637B8C6B84946B7B946B7B946B7B8C7384946B7B946B84946B7B8C
7384946B7B947384946B7B8C7384946B7B947384946B7B8C7384946B7B947384946B7B8C
7384946B84947384946B7B8C7384946B84947384946B848C738C94738494738C94738494
738C946B42316B4A3173637B7B6B8473637B7B6B8473637B7B6B84636B4A6B6B4A5A6342
734A426B42396B42426B4239734A426B42397342426B4239734A426B42397342426B4239
734A426B42397342426B4239734A426B4239734A426B4239734A426B4239734A426B4239
734A426B4239734A426B4239734A426B4239734A426B4239734A426B4239734A426B4239
84738C000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000947B84947B84947B84947B84947B84947B84947B84947B84
947B84947B846B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B94
6B7B946B84946B7B946B7B946B7B947384946B7B946B84946B84947384946B7B94738494
7384947384946B8494738494738494738494738494738494738494738494738494738494
738494738C946B42316B4A317363847B6B8473637B7B6B847B6B846B6B4A636B4A6B4242
6B42397342426B42396B42426B42397342426B42396B42426B42397342426B4239734242
6B42397342426B4239734239734239734A426B4239734239734239734A426B4239734239
734239734A426B4239734239734239734A42734239734239734239734A42734239734239
734239734A42734239000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004229314A31398C7B7B948484947B84947B84947B7B948484947B7B
948484947B7B6B7B94637B8C6B7B94637B8C6B7B946B7B8C6B84946B7B8C6B84946B7B8C
6B84946B7B8C7384946B7B8C7384946B7B8C7384946B7B8C7384946B7B8C7384946B7B8C
7384946B7B8C7384946B84947384946B848C7384946B84947384946B848C738C94738494
738C946B848C738C946B42316B4A3973637B7B6B8473637B7B6B84636B42734A426B4239
734A426B4239734A426B4239734A426B4239734A426B4239734A426B4239734A426B4239
734A426B4239734A426B4239734A426B4239734A426B4239734A426B4239734A426B4239
734A426B4239734A42734239734A426B847B738C846B847B738C846B847B738C846B847B
738C7B6B847B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004229314A31394229314A3139947B84947B84947B84947B84947B84947B84
947B846B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B946B7B94
6B7B946B7B946B7B946B84946B7B946B84946B7B946B84946B7B946B84946B8494738494
6B84947384946B84947384946B8494738494738494738494738494738494738494738494
7384947384947384947384947384946B4A31736B847B6B84636B4A636B4A6B4242734242
6B4239734242734242734A426B4239734242734239734A42734239734242734239734A42
734239734242734239734A42734239734A42734239734A426B8484738484738484738C84
6B84847384846B847B738C846B847B7384846B847B73847B6B847B73847B6B847B73847B
6B847B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
4A31394229314A31394229314A3139422931947B848C7B849C848C947B84948484947B84
9C848463738C6B7B9463738C6B7B94637B8C6B7B94637B8C6B84946B7B8C6B7B94637B8C
6B84946B7B8C6B84946B7B8C7384946B7B8C6B84946B7B8C7384946B7B8C7384946B7B8C
7384946B7B8C7384946B7B8C7384946B84947384946B7B8C7384946B848C7384946B848C
738C94738494738C946B848C738C947384947B6B8473637B6B73526B4239734A426B4239
734A426B4239734A426B4239734A42734239734A426B4239734A42734239734A426B8484
738C8C6B8484738C846B8484738C846B8484738C846B847B738C846B8484738C846B847B
738C846B847B738C846B847B738C846B847B738C846B847B738C846B847B73847B6B847B
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000422931
4A29314A31394229314A31394A31394A31394A2931947B84947B84948484947B84947B84
947B846B7B94637B8C6B7B946B7B946B7B946B7B8C6B7B946B7B946B7B946B7B8C6B7B94
6B7B946B7B946B7B8C6B7B946B7B946B84946B7B8C6B7B946B7B946B84946B7B8C6B8494
6B84947384946B7B8C6B84946B84947384946B84947384946B84947384946B8494738494
7384947384947384947384947384947384947384947B6B84734242734A42734242734242
734242734A426B848473848C73848C738C8C6B8484738484738484738C846B8484738484
738484738C846B84847384846B84847384846B84846B84846B84847384846B847B6B8484
6B847B7384846B847B6B847B6B847B7384846B847B6B847B6B847B73847B6B847B000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A3139422931
4A31394229314A31394A29314A31394A29314A3139947B8494848C947B849C848C947B84
9C848463738C6B7B9463738C6B7B9463738C6B7B94637B8C6B7B94637B8C6B7B94637B8C
6B8494637B8C6B84946B7B8C6B84946B7B8C6B84946B7B8C6B84946B7B8C7384946B7B8C
7384946B7B8C7384946B7B8C7384946B7B8C7384946B848C7384946B848C7384946B848C
7384946B848C738C946B848C738C946B848C734A52734A4A738C8C6B8484738C8C6B8484
738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484738C8C6B8484
738C846B8484738C846B8484738C846B847B738C846B847B738C846B847B738C846B847B
7384846B847B7384846B847B7384846B847B7384846B847B73847B6B847B000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004229314A31394229314A3139
4A29314A31394A31394A31394A29394A31394A31394A3139947B84948484947B84948484
63738C637B94637B8C6B7B94637B8C6B7B946B7B8C6B7B94637B8C6B7B946B7B8C6B7B94
6B7B8C6B7B946B7B8C6B7B946B7B8C6B7B946B7B8C6B84946B7B8C6B84946B7B946B8494
6B7B8C6B84946B84946B84946B84946B84946B84947384946B848C7384946B8494734A52
734A4A734A52734A4A734A52734A4A734A52734A4A7B738C73738C73848C6B848C73848C
6B848473848C6B848473848C6B84847384846B84847384846B84846B84846B84846B8484
6B84846B84846B84846B84846B84846B84846B84846B84846B847B6B84846B847B6B8484
6B847B6B847B6B847B6B847B6B847B6B847B6B847B6B847B6B847B000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004229314A31394229314A3139422931
4A31394A29314A31394A29315231394A29395231394A29319C848C947B849C848C947B84
6B7B9463738C6B7B9463738C6B7B94637B8C6B7B9463738C6B7B94637B8C6B7B94637B8C
6B8494637B8C6B7B94637B8C6B84946B7B8C6B8494637B8C6B84946B7B8C6B84946B7B8C
7384946B7B8C6B84946B7B8C734A526B4A4A734A526B424A734A526B4A4A734A526B424A
734A52734A4A734A52734A4A7B5252734A4A6B735A736B8C7B7394634A3173848C6B8484
738C8C6B848473848C6B8484738C8C6B84847384846B8484738C8C6B84847384846B847B
738C846B84846B84846B847B738C846B847B6B84846B847B7384846B847B6B84846B7B7B
7384846B847B6B847B6B7B7B7384846B847B6B847B6B7B73000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A31394229314A31394A31394A31394A29314A3139
4A31394A31394A29394A31394A31394A31394A31394A31394A313994848C947B84948484
637B8C6B7B9463738C637B8C637B8C6B7B94637B8C6B7B8C637B8C6B7B94637B8C6B7B8C
6B7B8C6B7B94637B8C6B7B946B7B8C6B7B946B7B8C6B7B946B7B8C6B4A526B424A6B4A4A
6B4A4A734A526B4A4A734A4A734A4A734A526B4A4A734A4A734A4A734A52734A4A734A52
734A4A734A52734A4A734A52734A526B7B5A73738C7B738C7B738C7B738C634A316B8484
6B84846B848C6B84846B84846B84846B84846B84846B84846B84846B84846B84846B8484
6B84846B84846B847B6B84846B84846B84846B847B6B84846B847B6B84846B847B6B847B
6B847B6B847B6B847B6B847B6B847B6B847B6B847B000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004A31394229314A31394229314A31394229314A31394A2931
4A31394A29315231394A29395231394A29395231394A31395231394A29399C848C63738C
6B7B9463738C6B7B9463738C6B7B9463738C6B7B9463738C6B7B9463738C6B7B94637B8C
6B7B94637B8C6B4A526B424A6B4A526B424A734A526B424A734A526B424A734A526B424A
734A526B424A734A526B4A4A734A526B424A734A526B4A4A734A526B4A4A734A52734A4A
734A52734A4A7B5252734A4A6B7B5A6373527B739473738C7B739473738C6B52396B8484
73848C6B848473848C6B84846B84846B84846B84846B84846B84846B847B6B84846B847B
6B84846B7B7B6B84846B847B6B84846B7B7B6B84846B7B7B6B84846B7B7B6B84846B7B7B
6B847B6B7B7B6B847B6B7B7B6B847B637B73000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004229314A31394229314A31394A29394A31394A29394A31394A31394A3139
4A31394A31394A31394A31394A31395231394A31395231394A3139523139947B84637B8C
63738C637B8C63738C637B8C63738C637B8C637B8C6B4A4A63424A6B4A4A6B424A6B4A4A
6B424A6B4A4A6B424A6B4A526B424A6B4A4A6B4A4A6B4A526B4A4A6B4A526B4A4A734A52
6B4A4A734A526B4A4A734A52734A4A734A52734A4A734A52734A4A734A52734A52734A52
734A52734A52734A526B7B5A63735A6B735A73738C7B738C73738C7B738C7B738C6B4A31
6B84846B84846B84846B84846B84846B84846B84846B84846B84846B84846B84846B8484
6B847B6B84846B847B6B84846B847B6B847B6B847B6B847B6B847B6B847B6B847B6B847B
6B847B6B847B6B847B6B847B6B7B7B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004229314A31394229314A31394A29314A31394229314A31394A29394A31394A2931
5231394A31395231394A29395231394A31395231394A293952314252313952314263738C
6B4A5263424A6B4A4A63424A6B4A5263424A6B4A5263424A6B4A5263424A6B4A5263424A
6B4A526B424A6B4A526B424A734A526B424A6B4A526B424A734A526B4A4A734A526B424A
734A526B4A4A734A526B424A734A52734A4A734A526B4A4A735252734A4A734A52734A4A
7B5252734A4A6B7B6363735A6B7B5A736B8C7B738C736B8C7B739473738C7B739473738C
6B52396B84846B8484637B846B84846B84846B8484637B7B6B84846B7B7B6B8484637B7B
6B84846B7B7B6B8484637B7B6B8484637B7B6B8484637B7B6B8484637B7B6B847B637B7B
6B847B637B7B6B847B637B73000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A3139
4229394A31394229314A31394A31394A31394A29394A31394A31394A31394A31394A3139
4A31395231394A31395231395231395231394A637B526384526384526B844A638463424A
63424A6B4A4A63424A6B424A6B424A6B4A5263424A6B4A4A6B424A6B4A526B424A6B4A4A
6B4A4A6B4A526B424A6B4A526B4A4A6B4A526B4A4A6B4A526B4A52734A526B4A4A734A52
734A52734A526B4A4A734A52734A52734A52734A52734A52734A52734A52734A52734A52
734A526B7B6363735A6B735A63735A73738C73738C73738C73738C7B739473738C7B738C
634A31634A316B84846B84846B84846B84846B84846B84846B84846B84846B7B7B6B8484
6B847B6B84846B7B7B6B847B6B847B6B847B637B7B6B847B6B847B6B847B637B7B6B847B
6B7B7B6B847B637B7B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A3139422931
4A31394229314A31394229314A31394A29314A31394A29394A31394A29395231394A2939
5231394A637B526B844A637B526B844A637B526B844A637B526B844A637B9C949C9C8C94
6B4A5263424A6B4A5263424A6B4A5263424A6B4A5263424A6B4A5263424A6B4A526B424A
6B4A526B424A734A526B424A734A526B424A734A526B424A734A526B4A4A734A526B4A4A
734A526B4A4A734A526B4A4A734A52734A4A735252734A4A734A52734A4A7B5252734A4A
6B7B6363735A6B7B6363735A73738C736B8C7B739473738C7B739473738C7B7B9473738C
7B7B94634A31634A31637B7B6B8484637B7B6B8484637B7B6B8484637B7B6B8484637B7B
6B8484637B7B6B8484637B7B6B8484637B7B6B847B637B7B6B847B637B7B6B847B637B7B
6B847B637B73000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A31394A29394A3139
4A29394A31394A31394A31394A31394A31394A31394A31394A637B4A637B4A637B4A637B
4A637B4A637B4A637B4A637B4A637B4A637B4A637B5263844A637B5263849C94949C9494
9C94946B4A5263424A6B4A526B424A6B4A526B4A4A6B4A526B424A6B4A526B4A4A6B4A52
6B4A4A6B4A526B4A526B4A526B4A4A6B4A526B4A52734A526B4A52734A52734A52734A52
734A52734A52734A52734A52734A52734A52734A52734A52734A52734A52734A526B7B63
6B735A6B7B5A63735A63735A736B8C73738C73738C73738C73738C7B738C73738C7B7394
73738C6B5231634A31634A31637B846B84846B7B846B8484637B7B6B8484637B7B6B8484
637B7B6B847B637B7B6B847B637B7B6B847B637B7B6B847B637B7B6B847B637B7B6B847B
637B7B000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004229314A31394229314A3139422931
4A31394A29394A637B425A734A637B4A5A734A637B425A734A637B4A5A7B4A637B4A5A73
52637B4A637B4A637B4A5A7B5263844A637B5263844A637B526B844A637B9C9494948C94
9C949C63424A6B4A5263424A6B4A526B424A6B4A5263424A6B4A526B424A6B4A526B424A
734A526B4A4A734A526B424A734A526B4A4A734A526B424A734A526B4A52734A526B4A4A
735252734A52734A52734A4A7B5252734A52734A52734A4A7B5252734A527B525263735A
6B7B6363735A6B735A636B527B7394736B8C737394736B8C7B739473738C7B739473738C
7B7B9473738C6B5239634A316B8484637B7B6B8484637B7B6B8484637B7B6B8484637B7B
6B8484637B7B6B847B637B7B6B8484637B7B6B847B637B736B847B637B736B847B637B73
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004229314A31394A31394A5A73425A734A5A73
4A5A734A5A7B425A734A5A734A5A734A637B4A5A734A637B4A637B4A637B4A5A7B4A637B
4A637B4A637B4A637B4A637B4A637B4A637B4A637B4A637B4A637B9C94949C8C949C9494
9C94949C94946B424A6B4A526B4A526B4A526B424A6B4A526B4A526B4A526B4A526B4A52
6B4A52734A526B4A526B4A526B4A52734A526B4A52734A52734A52734A52734A52734A52
734A52734A52734A52734A52734A52734A52734A52734A52734A527B52526B7B5A6B7B5A
63735A6B735A63735A73738C73738C73738C73738C73738C73738C7B739473738C7B7394
7B73947B7B94634A31634A31634A316B8484637B7B637B7B637B7B6B8484637B7B637B7B
637B7B6B847B637B7B637B7B637B7B6B847B637B7B637B7B637B7B637B7B637B73000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000042526B4A5A734252734A5A73425A734A5A73425A73
4A5A73425A734A637B425A734A637B425A734A637B425A734A637B425A734A637B4A5A73
4A637B4A5A734A637B4A5A734A637B4A5A7B52637B4A637B52637B948C949C949C948C94
9C949C948C949C949C6B424A6B4A526B424A6B4A526B424A6B4A526B424A734A526B4A4A
734A526B424A734A526B4A52734A526B4A52734A526B4A52734A526B4A52735252734A52
735252734A527B5252734A527B5252734A527B525A734A527B525263735A6B7B6363735A
6B7B5A63735A6B735A736B8C737394736B8C7B739473738C7B739473738C7B7B9473738C
7B7B9473738C7B7B94634A316352315A4A316B8484637B7B6B8484637B7B6B8484637B7B
6B847B637B7B6B847B637B7B6B847B637B736B847B637B736B847B637B73000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000425A73425A73425A73425A734A5A73
425A734A5A73425A734A5A73425A734A5A734A5A734A5A734A5A734A637B4A5A734A637B
4A5A734A637B4A637B4A637B4A637B4A637B4A637B4A637B4A637B9C94949C94949C9494
9C94949C94949C94946B4A526B4A526B4A526B4A526B4A526B4A526B4A526B4A52734A52
6B4A52734A526B4A52734A526B4A52734A52734A52734A52734A52734A52734A52734A52
734A52734A52734A52734A52734A52734A52734A527B52526B7B636B7B6363735A6B7B5A
63735A63735A736B8C73738C73738C73738C73738C73738C73738C73738C73738C7B7394
73738C7B73947B73947B7B94634A31634A315A4A31637B7B637B7B637B7B637B7B637B7B
637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B7B637B73000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000042526B4A5A7342526B
4A5A73425A734A5A73425A734A637B425A734A6373425A734A637B425A734A637B425A73
4A637B4A5A734A637B425A734A637B4A5A734A637B4A5A739C949C948C949C949C948C94
9C949C9494949C9494948C94734A526B4A526B4A526B424A734A526B4A52734A526B424A
734A526B4A52734A526B4A527352526B4A52734A526B4A5273525A734A52735252734A52
7B525A734A52735252734A527B525A734A527B525A637B5A6B7B6363735A6B7B6363735A
6B7B6363735A73738C6B6B8C737394736B8C737394736B8C7B739473738C7B739473738C
7B7B9473738C7B7B9473738C6B5239634A31634A315A42316B8484637B7B637B7B637B73
6B847B637B7B637B7B637B73637B7B637B73637B7B5A7B73000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000425A73
425A734A5A73425A73425A73425A734A5A73425A734A5A734A5A734A5A73425A734A5A73
4A5A734A637B4A5A734A637B4A63734A637B4A5A734A637B9C94949C949C9494949C9494
9C94949C94949494949C94949C94946B4A526B4A526B4A526B4A52734A526B4A52734A52
734A52734A526B4A52734A52734A52734A52734A52734A52734A52734A52734A52734A52
734A52735252734A52734A52734A527B525A6B7B636B7B636B7B636B7B6363735A63735A
63735A63735A736B8C73738C73738C73738C73738C73738C73738C73739473738C737394
7373947B7B9473738C7B73947B7B946B5231634A31634A315A4A31637B7B637B7B637B7B
637B7B637B7B637B7B637B7B637B7B637B7B637B73000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000042526B4A5A73425A6B4A5A73425A6B4A5A73425A734A5A73425A734A637B425A73
4A637B425A734A637B425A734A637B425A734A637B4A5A739C949C948C949C949C949494
9C949C948C949C949C9494949C949C949494734A526B4A52734A526B4A52734A526B4A52
734A5A6B4A5273525A6B4A5273525A6B4A5273525A734A5273525A734A527B525A734A52
7B525A734A527B525A734A527B525A6B7B636B7B6363735A6B7B6363735A6B7B6363735A
63735A6B6B8C7373946B6B8C737394736B8C73739473738C7B739473738C7B7B9473738C
7B7B9473738C7B7B9473738C7B7B94634A316B5239634A31634A315A4229637B7B5A7B73
637B7B5A7B73637B7B637B73637B7B5A7373000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000425A73425A73425A73425A734A5A73425A734A5A73425A734A5A73
425A734A5A734A5A734A63734A5A734A63734A5A739C949C9494949C949C9494949C949C
9494949C94949494949C94949494949C94946B4A52734A526B4A52734A526B4A52734A52
734A52734A52734A52734A52734A52734A52734A52734A52734A52734A52734A5273525A
734A52735252734A527B525A6B7B636B7B636B7B636B7B6363735A63735A63735A63735A
63735A73738C73738C73738C73738C73738C73738C73739473738C73739473738C737394
73738C7373947373947B7B947373947B7B946B5231635231634A315A4A31637B7B637B7B
637B7B637B7B637B73637B7B637B73000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000425A6B4A5A7342526B4A5A73425A734A5A73425A6B
4A6373425A734A6373425A734A637B425A734A637B948C949C949C9494949C949C948C94
9C949C9494949C949C9494949C9C9C9494949C94946B4A5273525A6B4A52734A5A6B4A52
73525A734A5273525A6B4A5273525A734A5273525A734A527B525A734A527B525A734A52
7B525A734A527B525A6B7B636B846B637B636B7B6363735A6B7B6363735A63735A5A6B52
7373946B6B8C73738C6B6B8C73739473738C737394736B8C7B739473738C73739473738C
7B7B9473738C7B7B9473738C7B7B947373947B7B94634A316352395A4A31634A315A7373
637B7B5A7373637B7B5A7373000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000425A73425A734A5A73425A73425A73
425A734A5A73425A734A5A734A5A734A6373425A7394949C94949C9C949C94949494949C
9494949C949C9494949C94949494949C94949494949C9494734A52734A5A734A52734A52
734A52734A5A734A52734A52734A5273525A734A52734A5A734A527B525A734A5273525A
734A5A7B525A734A526B7B636B7B636B7B6363735A637B6363735A63735A63735A63735A
6B6B8C73738C6B6B8C73738C73738C73739473738C73738C73738C73739473738C737394
7373947B7B947373947B7B94737B947B7B947373946B5231634A31634A315A4A315A4A31
637B73637B7B5A7373000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000042526B4A5A73425A6B
4A5A73425A6B4A5A73425A734A6373425A739C949C948C949C949C948C949C949C949494
9C949C9494949C949C9494949C949C9494949C9C9C94949473525A6B4A5273525A734A52
73525A734A5273525A734A5273525A734A527B525A734A527B525A734A527B525A734A52
7B525A734A526B846B637B636B7B6363735A6B7B6363735A6B7B6363735A63735A5A6B52
73738C6B6B8C7373946B6B8C7373946B6B8C73739473738C73739473738C7B7B9473738C
7B7B9473738C7B7B9473738C7B7B9473738C7B7B947373946B5239634A316352315A4A31
5A4A315A7373000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
425A6B425A73425A734A5A73425A734A5A7394949C94949C94949C94949C94949C94949C
94949494949C9494949C949C9494949C94949494949C9494949494734A5A734A52734A5A
734A52734A5A734A5273525A734A5273525A734A5A7B525A734A527B525A734A5A7B525A
734A5A6B7B636B7B636B7B63637B636B7B6363735A63736363735A63735A63735A73738C
6B6B8C73738C6B738C73738C73738C73738C73738C73739473738C73739473738C737394
7373947373947373947B7B947373947B7B94737B947B7B946B5231635231634A31634A31
5A4A315A4A31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A5A7342526B4A5A73425A6B94949C948C949C949C9494949C949C949494
9C949C9494949C949C9494949C9C9C9494949C9C9C9494949C9C9C94949473525A734A52
7B525A734A5273525A734A527B525A734A527B525A734A527B525A734A527B525A734A52
6B846B6B7B636B7B6363735A6B7B6363735A6B7B6363735A637B6363735A63735A6B6B8C
7373946B6B8C7373946B6B8C7373946B738C7373946B6B8C737B9473738C73739473738C
7B7B9473738C7B7B9473738C7B7B947373947B7B947373947B7B9C634A316352315A4A31
634A395A4231000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000425A734A5A7394949C94949C94949C94949C94949C94949C
94949C94949C94949494949C9494949C949C9494949494949494949C9C94734A52734A5A
734A5A73525A734A5273525A734A5A7B525A734A5A7B525A7B525A7B525A734A5A6B846B
6B7B636B7B63637B636B7B63637B636B7B6363735A63735A63735A63735A6B6B8C6B6B8C
6B6B8C73738C6B6B8C73738C73738C73739473738C73738C73738C73739473738C737394
737394737B94737394737B94737B947B7B94737B947B7B947B7B947B7B94634A31634A31
634A315A4A31000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000948C9494949C94949494949C9494949C949C949494
94949C9494949C9C9C9494949C9C9C9494949C9C9C9494949C9C9C9494949C9C9C734A52
7B525A734A527B525A734A527B525A734A527B525A734A527B525A734A5273846B6B7B63
6B846B637B636B7B6363735A6B7B6363735A637B6363735A63735A5A6B5273738C6B6B8C
7373946B6B8C7373946B6B8C7373946B6B8C7373946B738C73739473738C737B9473738C
737B9473738C7B7B9473738C7B7B947373947B7B947373947B7B947373946B5239634A31
6352395A4A31634A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000094949C94949C94949C94949C94949C94949C94949C
94949C94949C94949C94949C94949494949C94949494949C949494949494949494949C94
734A5A73525A73525A7B525A734A5A7B525A73525A7B525A7B525A6B846B6B7B6B6B7B6B
6B7B636B7B63637B63637B6363735A63736363735A63735A5A735A63735A6B6B8C6B738C
6B6B8C73738C6B738C73738C6B738C73738C73738C73739473738C737394737394737394
737394737394737394737B947373947B7B94737B947B7B94737B947B7B947B7B946B5231
634A31634A315A4A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000094949C94949C94949494949C94949C94949C949494
9C9C9C94949494949C9494949C9C9C949494949C9C9494949C9C9C9494949C9C9C949494
7B525A734A5A7B525A734A527B525A734A5A7B525A734A5A73846B6B7B636B846B637B63
6B846B637B636B7B6363735A6B7B6363735A63735A5A6B5A63735A6B6B8C6B738C6B6B8C
7373946B6B8C7373946B6B8C7373946B738C7373946B738C737B9473738C73739473738C
7B7B94737394737B9473738C7B7B947373947B7B947373947B7B9C737B947B7B9C634A31
6B5239634A31634A31000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000094949C94949C94949C94949C94949C94949C94949C94949C
94949C94949C94949C94949C94949C949C9C94949494949C949494949C9C949494949C94
9494947B525A734A5A7B525A7B525A7B525A7B525A6B846B6B846B6B846B6B7B636B7B63
6B7B636B7B6363736363736363735A63735A5A735A63735A5A735A6B738C6B6B8C6B6B8C
6B6B8C73738C6B6B8C6B738C73738C7373946B738C73739473739473739473738C737394
737394737B94737394737B94737B947B7B94737B94737B947B7B947B7B94737B947B7B94
6B5231635231634A315A4A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000094949C8C949C94949C94949C94949C8C949C949C9C94949C
949C9C949494949C9C949494949C9C949494949C9C949494949C9C9494949C9C9C949494
949C9C9494947B525A734A5A7B525A734A5A73846B6B7B6B6B846B637B636B846B637B63
6B7B6363735A6B7B6363735A637B635A735A63735A5A6B5A6B738C6B6B8C7373946B6B8C
7373946B6B8C7373946B6B8C7373946B6B8C7373946B738C7373946B738C737B9473738C
737B9473738C7B7B947373947B7B947373947B7B947373947B7B9C737B947B7B9C737B94
7B7B9C634A316352395A4A31000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000094949C94949C94949C94949C94949C94949C94949C94949C
94949C94949C94949C94949C94949C94949C94949C949C9C949494949C9C949494949C9C
949494949C947B525A7B525A7B525A7B525A6B846B6B846B6B7B636B7B6B6B7B636B7B63
637B63637B6363736363736363735A63735A5A735A63735A6B6B8C6B6B8C6B6B8C6B738C
6B6B8C6B738C6B738C73738C6B738C73739473738C73739473738C737394737394737394
737394737B94737394737B94737394737B94737B947B7B94737B947B7B947B7B947B7B94
7B7B946B5231635231635231000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008C949C9494A58C949C94949C8C949C949CA594949C94949C8C949C
949C9C94949C949C9C8C9494949C9C949494949C9C8C9494949C9C949494949C9C949494
9C9C9C949494949C9C734A5A7B52636B846B6B846B637B636B846B637B636B7B6B637363
6B7B6363735A637B635A735A637B635A735A63735A5A6B526B738C6B6B8C6B738C6B6B8C
7373946B6B8C7373946B6B8C7373946B738C7373946B738C737B946B738C737B946B738C
737B94737394737B9473738C7B7B947373947B7B947373947B7B9C737B947B7B9C737B94
7B849C737B946B5239634A31635239000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000094949C94949C94949C8C949C94949C94949C94949C94949C94949C
94949C94949C94949C94949C94949C949C9C94949C949C9C94949C949C9C949494949C9C
949C9C949C9C949494949C946B846B6B846B6B7B6B6B7B6B6B7B636B7B6B637B63637B63
637B63637B6363735A63735A63735A63735A5A6B5A6B6B8C6B6B8C6B738C6B6B8C6B6B8C
6B738C6B73946B6B8C6B738C6B738C7373946B738C737394737394737394737394737394
737394737B94737394737B94737B94737B94737B94737B94737B947B7B94737B947B7B94
7B7B947B7B9C6B5231635231634A31000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000008C949C9494A58C949C949CA58C949C949CA58C949C949CA58C949C
949C9C8C949C949C9C8C949C949C9C8C9494949C9C949494949C9C8C9494949C9C949494
949C9C949494949C9C949494949C9C6B7B636B846B637B636B846B637B636B7B6B637363
637B635A735A637B635A735A63735A5A6B5A63735A636B8C6B738C6B6B8C6B73946B6B8C
6B73946B6B8C7373946B6B8C7373946B6B8C7373946B738C7373946B738C737B946B738C
737B9473738C737B94737394737B947373947B7B9C737B947B7B9C737B947B7B9C737B94
7B849C737B947B849C6352316B5239000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008C949C9494A58C949C94949C8C949C94949C8C949C94949C94949C94949C
8C949C94949C94949C949C9C94949C949C9C94949C949C9C94949C949C9C94949C949C9C
949C94949C9C949C94949C9C949C94949C94949C946B7B6B637B63637B63637B63637B63
63736363736363735A63735A5A735A5A735A6B6B8C6B6B8C6B6B8C6B6B8C6B6B8C6B738C
6B6B8C6B738C6B738C6B73946B738C6B73946B738C7373946B7394737394737394737394
737394737394737394737B94737394737B94737B94737B94737B94737B94737B947B7B9C
737B947B7B9C7B7B947B849C635231635231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000009494A58C949C949CA58C949C9494A58C949C949CA58C949C949C9C8C949C
949CA58C949C949C9C8C949C949C9C8C949C949C9C8C9494949C9C8C9494949C9C8C9494
949C9C949494949C9C8C9494949C9C949C94949C9C8C94946B7B6B637B63637B635A735A
637B635A735A6373635A6B5A63735A5A6B526B6B8C636B8C6B73946B6B8C6B738C6B6B8C
7373946B6B8C6B73946B6B8C7373946B738C7373946B738C737B946B738C737B946B738C
737B94737394737B947373947B7B9C737394737B947373947B7B9C737B947B7B9C737B94
7B849C737B947B849C737B946B5A39635231000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000008C949C8C94A58C949C9494A58C949C94949C8C949C949C9C8C949C94949C
94949C949C9C8C949C949C9C94949C949C9C8C949C949C9C949C9C949C9C94949C949C9C
949C9C949C9C949C94949C9C949C9C949C9C949C94949C94949C94949C9C637363637363
6373636373635A735A5A735A5A735A5A735A636B8C6B6B8C6B6B8C6B738C6B6B8C6B738C
6B738C6B73946B6B8C6B738C6B738C7373946B738C7373946B73947373946B7394737394
737394737B94737394737B94737B94737B94737B94737B94737B947B7B9C737B947B7B94
737B947B7B9C737B947B7B9C7B7B9C6B5239000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C949C949CA58C949C949CA58C949C949CA58C949C949CA58C949C949CA58C949C
949CA58C949C949CA58C949C949C9C8C949C949C9C8C949C949C9C8C9494949C9C8C9494
949C9C8C9494949C9C8C9494949C9C8C9494949C9C8C9C94949C9C8C9C94949C9C949C94
637B635A735A6373635A6B5A5A735A636B8C6B738C636B8C6B738C636B8C6B73946B6B8C
6B73946B6B8C7373946B6B8C7373946B6B8C7373946B738C7373946B738C737B946B738C
737B946B7394737B94737394737B947373947B7B9C7373947B7B9C737B947B7B9C737B94
7B849C737B947B849C737B947B849C6352316B5239000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C94A58C949C8C94A58C949C8C94A58C949C9494A58C949C8C949C8C949C949C9C
8C949C949C9C8C949C949C9C8C949C949C9C8C949C949C9C8C9C9C949C9C949C9C949C9C
8C9C9C949C9C949C9C949C9C949C94949C9C949C94949C9C949C94949C9C949C94949C94
949C945A735A5A735A5A735A5A6B5A6B6B8C636B8C6B6B8C6B6B8C6B6B8C6B6B8C6B738C
6B6B8C6B738C6B738C6B73946B738C6B73946B73947373946B73947373946B7394737394
737394737B94737394737B94737394737B94737B94737B94737B94737B94737B947B7B9C
737B947B7B9C737B947B849C7B7B9C7B849C6B5231000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C949C9494A58C949C949CA58C949C949CA58C949C949CA58C949C949CA58C949C
949CA58C949C949C9C8C949C949CA58C949C949C9C8C949C949C9C8C949C949C9C8C9494
949C9C8C9C9C949C9C8C9494949C9C8C9C94949C9C8C9C9494A59C8C9C94949C9C8C9C94
94A59C8C9C94949C9C5A6B526B738C636B8C6B6B8C636B8C6B7394636B8C6B738C636B8C
6B73946B6B8C6B73946B6B8C7373946B738C7373946B738C737B946B7394737B946B738C
737B946B7394737B946B7394737B9C737394737B9C7373947B7B9C737B947B7B9C737B94
7B849C737B947B849C737B947B849C737B947B849C000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000008C94A58C949C8C94A58C94A58C94A58C949C8C94A58C949C949CA58C949C8C9C9C
8C949C949C9C8C949C8C9C9C8C9C9C949C9C8C949C8C9C9C8C9C9C949C9C8C9C9C949C9C
949C9C949C9C8C9C9C949C9C949C9C949C9C8C9C94949C9C949C9C949C9C8C9C94949C94
949C94949C9C949C94949C948C948C6B6B8C636B8C6B6B8C6B6B8C6B738C6B6B8C6B738C
6B738C6B73946B738C6B73946B738C6B73946B738C6B73946B73947373946B7394737394
737394737B94737394737B94737B94737B94737B94737B94737B94737B9C737B94737B94
737B9C7B849C737B947B7B9C7B7B9C7B849C7B7B9C6B5231000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C9CA58C949C949CA58C949C8C9CA58C949C949CA58C949C949CA58C949C949CA58C949C
949CA58C949C949CA58C949C949CA58C949C949CA58C949C949C9C8C949C949C9C8C949C
949C9C8C9494949C9C8C9C94949C9C8C9C94949C9C8C9C94949C9C8C9C9494A59C4A527B
525A8452527B525A84425A42425A4A3952398C948C848C8C8C948C848C848C948C848C84
8C948C848C848C948C848C848C948C848C848C9484848C848C9484848C7B8C9484848C7B
8C8C84848C7B8C8C84848C7B737B9C737394737B9C737B94737B9C737B947B7B9C737B94
7B849C737B947B849C737B947B849C737B947B849C737B9C000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C94A58C94A58C94A58C94A58C94A58C94A58C949C8C9CA58C949C8C9CA58C949C8C9CA5
8C949C8C9C9C8C949C8C9C9C8C9C9C8C9C9C8C9C9C949C9C8C9C9C8C9C9C8C9C9C949C9C
8C9C9C949C9C8C9C9C949C9C8C9C9C949C9C8C9C9C525A7B4A527B525A7B525A7B525A7B
525A7B525A7B525A7B425A4A4252420000000000008C948C8C8C8C8C948C8C8C8C8C948C
848C848C948C848C848C9484848C84848C84848C84848C84848C84848C84848C84848C84
848C7B848C84848C7B848C84848C7B848C7B848C7B848C7B848C7B848C7B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
8C9CA58C94A58C9CA58C949C949CA58C949C8C9CA58C949C949CA58C949C8C9CA58C949C
949CA58C949C8C9CA58C949C949CA58C949C949C9C8C949C949CA58C9C9C949C9C8C949C
949C9C8C9C9C4A52734A5273525A7B4A5273525A7B4A5273525A7B4A527B525A7B4A527B
525A84525A7B4A5A4A3952424252420000000000000000008C948C848C848C948C848C84
8C948C848C848C948C848C848C948C848C848C9484848C848C9484848C84848C84848C7B
8C9484848C7B848C84848C7B848C84848C7B848C847B847B848C8484847B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008C94A5
8C94A58C94A58C94A58C94A58C94A58C9CA58C949C8C94A58C94A58C9CA58C949C8C9CA5
8C9C9C8C9CA58C949C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C8C9C9C4A4A6B4A5273
4A52734A52734A52734A52734A527352527B4A527352527B4A527B525A7B4A527B525A7B
525A7B525A7B425242425242000000000000000000000000000000000000848C848C948C
848C848C948C848C84848C84848C848C9484848C84848C84848C84848C84848C84848C84
848C84848C84848C7B848C7B848C7B848C84848C7B848C7B848C7B848C7B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008494A5
8C9CA58494A58C9CA58C94A58C9CA58C949C8C9CA58C949C8C9CA58C949C8C9CA58C949C
8C9CA58C949C8C9CA58C949C8C9CA50000000000000000004A5273424A6B4A52734A4A6B
4A52734A4A6B5252734A527352527B4A5273525A7B4A5273525A7B4A5273525A7B4A527B
525A7B425242425A42000000000000000000000000000000000000000000000000848C84
8C948C848C848C948C848C848C948C848C848C948C848C848C9484848C7B8C9484848C7B
848C84848C7B848C84848C7B848C847B847B848C847B847B848C7B7B847B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000008C94A5
8C94A58C94A58C94A58C9CA58C94A58C9CA58C94A58C9CA58C94A58C9CA58C94A58C9CA5
8C949C0000000000000000000000000000000000000000004A4A6B4A526B4A526B4A5273
4A52734A52734A52734A52734A52734A52734A5273525A7B4A5273525A7B52527B525A7B
425A42425242000000000000000000000000000000000000000000000000000000000000
00000084948C848C84849484848C84848C84848C84848C84848C84848C84848C84848C84
848C84848C84848C7B848C84848C7B848C7B848C7B848C7B848C7B848C7B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C9CA58494A5
8C9CAD8494A58C9CA584949C8C9CA58C94A58C9CA584949C000000000000000000000000
000000000000000000000000000000000000000000000000000000424A6B4A5273424A6B
4A52734A52734A52734A5273525A7B4A5273525A7B4A5273525A7B4A5273525A7B4A5273
425A4A395239000000000000000000000000000000000000000000000000000000000000
0000000000008C948C848C848C948C848C84849484848C848C9484848C84848C84848C7B
849484848C7B848C847B847B848C84848C7B848C847B847B848C847B847B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008494A58C94A5
8C94A58C9CA58C94A58C94A5000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A5273424A6B4A5273
4A52734A52734A52734A52734A52734A52734A52734A52734A527B525A7B4A527B425A42
425242000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000848C84849484848C84848C84848C84848C84848C84848C84
848C84848C84848C7B848C7B848C7B848C84848C7B848C7B848C7B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000008C9CAD000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000424A6B4A5273424A6B
4A52734A4A6B4A52734A52734A52734A5273525A7B4A5273525A7B4A5273525A7B395242
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000848C84849484848C84849484848C84849484848C7B
848C847B8C7B848C847B8C7B848C847B847B848C847B847B848C7B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A4A6B4A5273
4A526B4A52734A52734A52734A52734A52734A52734A527B4A5273525A7B425242425242
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000848C84848C84848C84848C84848C84848C84
848C7B848C84848C7B848C84848C7B848C7B848C7B848C7B7B8C7B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A5273424A6B
4A52734A526B4A52734A4A6B525A7B4A52734A527B4A5273525A7B425242425242000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000849484848C84848C847B8C7B
8494847B8C7B848C847B847B848C847B8C7B848C847B847B848C84000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A5273
4A52734A52734A52734A52734A52734A52734A52734A52734A5273425242000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000848C84848C84848C84
848C84848C847B8C7B848C84848C7B848C847B8C7B848C7B7B8C7B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000424A6B
4A5273424A6B4A52734A52734A52734A5273525A7B4A5273425A4A000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B8C7B
848C847B8C7B848C847B8C7B848C847B847B848C847B847B848C7B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A5273
4A4A6B4A52734A52734A52734A52734A52734A5273425A42395242000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
848C84848C84848C7B848C847B8C7B848C7B7B8C7B848C7B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
4A5273424A6B4A5273424A6B4A52734A52734A5273395242000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000848C847B847B848C847B8C7B848C847B847B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
4A526B4A52734A526B4A52734A52734A5273395242000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007B8C7B7B8C7B848C847B8C7B7B8C7B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000424A6B4A527342526B4A52734A5273000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B847B848C847B847B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A52734A52734A52734A5273425242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007B8C7B7B8C7B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000424A6B4A5273424A6B425A4A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007B847B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000424A6B4A5273000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A5273000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[54]:=",
  ImageSize->{128, 128},
  AnimationDisplayTime->0.169,
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000200000P0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj2LW
9`00?g=cLda<C240<410Y/30`9T0IPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm?ooc020P800082006IVc3<cIP2009VIod:J
Yb8R8_mV<`00P00g??mVc5J0Zk75fBa1EkV>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7De3eDE:b/[:BT
Y>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEeMJ_0eLC4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7i/;2`W9bLP00
o`1o001oO`1oo`3o003oO`3oogl007l0Ogl0ogmo07moOgmoogoo07ooOgooool00?l0Ool0oomo0?mo
Oomooooo0?ooOoooo`0002XZ:UEEEGmoOjZZZ]GEeOoooh0700201`00P0L0080700031gT010L000<7
N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@7
00031dP000>OX000;P041`000`M70003VZ2J02l010L000<7AP001InPWj2O02h010L000<7A@001YZP
VZ2JX2h010L000<7@`002J2OX9nPWj2OX0080003Vj2K028010L000<7@P002Z2JX9ZPVZ2JX9X60005
X9ZPVYl08P041`000`M1000<X9nPWj2OX9nPWj2O0`001i^PVj2KX8408P041`000`LU0003O00001P0
01NPVZ2JX9ZPVZ2JX9ZP0:21X86PVZ2JW`0R00@700031bD000>2P@005`006:2OX9nPWj2OX9nPWj2K
X9^PVj2KX9^PPB8010L000<79@04PAD001VPVZ2JX9ZPVZ2JX9ZPPJ21X86PVZ2JX86O028010L000<7
9@001H:1PX6201<001ZPWj2OX9nPWj2OX9^PVj2KX9^PVj2KX9^OPB8010L000<79007PA4001^1VZ2J
X9ZPVZ21X86PPJ21X86PVZ21X9ZPPIl08P041`000`LT0008PH:1PX62PH8@000KWj2OX9nPWj2KX9^P
Vj2KX9^PVj2KX9^PPIn1028010L000<79008P@03PP0000`001bJX9ZPVZ21X86PPJ21X86PPJ21X86P
VZ2JWh6O8P041`000`LT000;PH:1PX62PH:1Vh402`007IniWj2KX9^PVj2KX9^PVj2KX9^PVj2KX9^P
PIn1028010L000<79006P@06PX62PH:12@007YZPVZ21X86PPJ21X86PPJ21X86PPJ2JX86PVin1Wb80
10L000<78`02P@0;PX62PH:1PX62PI/00X45000PPInPVZ2KX9^PVj2KX9^PVj2KX9^PVj2KX9^PVin1
Wh4R00@700031b<000QlPWb2O89lPP>100J2PH:1PX42000QVh6KPJ21X86PPJ21X86PPJ21X86PPJ21
X9^PVj21Wh6O028010L000<73P09X00D^J2iX;VP^J2iX;VP^J2iX;VP^J05VP0WPIZ1VX6PVj2KX9^P
Vj2KX9^PVj2KX9^PVj2KX9^PVj2KX9^OPIn1028010L000<73P00BJ2KX9^PVj2KX9^PVZ2JX9ZPVZ2J
X9ZPVZ2JX9ZPQiZ0VX2JP9Z0VX2KPJ21X86PPJ21X86PPJ21X86PPJ21X86PVj2KWh6OPIl08P041`00
0`L?00JP01RiX:2P^J2iX;VP^J2iX;VP^J2iX;VPXJ05VP0PPIZ1VX6PVj2KX9^PVj2KX9^PVj2KX9^P
Vj2KX9^PVj02W`03PIn1028010L000<73`007I^PVj2KX9^PVj2JX9ZPVZ2JX9ZPVZ2JX9ZPVZ6700:P
02VJP9Z0VX2JP9Z0Vh6PPJ21X86PPJ21X86PPJ21X86PPJ2KX9^OPIn1W`0R00@700031a000j007;VP
X:2iX;VP^J2iX;VP^J2iX;VP^J2iX:6PXJ05VP0KPIZ1VY^PVj2KX9^PVj2KX9^PVj2KX9^PVj2K00:P
00JOPIn1Wh4R00@700031a4001^KX9^PVj2KX9ZPVZ2JX9ZPVZ2JX9ZPVZ27XHL01:009iZ0VX2JP9Z0
VX6PPJ21X86PPJ21X86PPJ21X86PPJ2KX86OPIn1W`0R00@700031a40026iX;VP^J2iX;VP^J2iX;VP
^J2iX;VP^J2QX:6PXJ2QX:4019X0886JPIZKX9^PVj2KX9^PVj2KX9^PVj2KX9^PVin1Wh6O0X4Q00@7
00031a800j005i^PX:2JX9ZPVZ2JX9ZPVZ2JX9ZPQj6700NP02F0VX2JP9Z0VX6PPJ21X86PPJ21X86P
PJ21X8:PVj21Wh6OPImk024010L000<74P008Z2iX;VP^J2iX;VP^J2iX;VP^J2iX;VPXJ2QX:6PXJ2Q
X:44VP0HPIZ1VY^PVj2KX9^PVj2KX9^PVj2KX9^P0Yl0186OPIl2PB4010L000<74`06X00CVj2JX9^P
VZ2JX9ZPVZ2JXHNQQ`09X00RP9Z0VX2JP9Z1X86PPJ21X86PPJ21X8:PPZ2KWh6OPIn1Wb8010L000<7
4`008kVP^J2iX;VP^J2iX;VP^J2iX;VP^J2QX:6PXJ2QX:6PXJ2Q00BJ01>1VX6PVj2KX9^PVj2KX9^P
Vj2K00>P0Yl01H6OPIn1028010L000<75005X00CVj2KX9ZPVj2JX9^PVZ2KXHNQQ`03X003Qj2P00JP
01nJP9Z0VX6PPJ21X86PPJ21X8:PPJ22X9^OPIn1Wh66028010L000<75@009;VP^J2iX;VP^J2iX;VP
^J2iX:6PXJ2QX:6PXJ2QX:6PXJ2QX0>J0161VY^PVj2KX9^PVj2KX9^PV`02X0>O00N1Wh6OPIZ10200
10L000<75@04X00KVj2KX9^PVj2KX9^PVj27XHNQQj27X8NPQj2700JP02:JP9Z0VX6PPJ21X86PPZ22
X8:PPZ21Wh6OPIn1VX2JP9Z27@041`000`LF000UX;VP^J2iX;VP^J2iX;VP^J2QX:6PXJ2QX:6PXJ2Q
X:6PXJ2QX003VP0?PIZKX9^PVj2KX9^PVj2K00:P00nOQYn1Wh6JPIZ1VX6JPIX06`041`000`LF00>P
01^KX9^PVj2KX9^PVj2KX8NPQj67X8NPQj27X8L00j000hNPX002X00SVX2JP:21X86PPZ22X8:PPZ22
Wh6OPIn1VX2JP9Z0VX2JP9X06P041`000`LG000U^J2iX;VP^J2iX;VP^J2iX:6PXJ2QX:6PXJ2QX:6P
XJ2QX:6PYP02X00VPIZ1X9^PVj2KX9^PVj2KX86OPIn1Wh6JPIZ1VX6JPIZ1VX6JNh4E00@700031aP0
3j00@8NPQj27X8NPQj27X8NPQj27X8NPQj27X82JP9Z2X8:PPZ22X9^PVin1Wh6JP9Z0VX2JP9Z0VX2J
P9Z0VX2JP:0B00@700031aT002NiX;VP^J2iX;VP^J2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:40
19X:X00=PIn1Wh6JPIZ1VX6JP@03VP03PIZJ00RJ40041`000`LI00bP04N7X8NPQj27X8NPQj27X8NP
Qj27X8NPQj27X8NPP9Z0VX:PPZ2KX9^PPIn1Wh2JP9Z0VX2JP9Z0VX2JP9Z0VX2JP9Z0VX2JV`0=00@7
00031aT002Z1X;VP^J2iX;VP^J2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ02VPVP00>1Wh40
7IX;00@700031aT000=lX:002J00BhNPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NJP:22X8:P
Vj21Wh2JP9Z0VX2JP9Z0VX2JP9Z0VX2JP9Z0VX2JP9Z0VX2JP00900@700031aX002^P^J2iX;VP^J2i
X:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6P00:J1Z000in1VP0QVP03X@0000D010L000<7
6P0;X01?Qj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27VX2PVj2KX86JP9Z0VX2JP9Z0VX2J
P9Z0VX2JP9Z1VX2JPIZ0VX6JP9Z1VX2JP00500@700031aT00X40;;VP^J2iX;VP^J2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6P0YX4X003WiZJ01^J1Z000jFP000500@700031aT000=kPG/0
1j00ChNPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPP:2KWh6JP9Z1VX2JP9Z0VX2J
P9Z0VX2JPJ27X8NPQj27X8NPQj27X8L01`041`000`LG0003PGb100:102jP^J2iX;VPXJ2QX:6PXJ2Q
Qj6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2Q0Z000inJVP0;VP06X:JPYJ2U4J0800@700031aH0
00N1Nh5kPG^100JP04j7X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPViZ1VX6J
Qj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQh4800@700031aD000B1O85l1840;KVP^J2iX:6P
XJ2QX:6PXHNQQj6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PX@02VP0DX:JPYZ2VX:JPYZ2VX:JPYJ2U
X:D@X0T010L000<74`002g^1Nh5kPG^1Nh5k00FP04b7X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NP
Qj27X86JPIZ1VX6JVj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NPQj27X8NP2P041`000`LB0007
O85lPGb1O005P@0OX;VP^J2QX:6PXJ2QX:67XHNQQj6PXHNQX:6PXJ2QV`0=VP03Wj6P00:P00FVX:6P
YP03X005YZ2PX:D00j000jFPX00<X0/010L000<740003h5kPG^1Nh5kPG^1Nh5kP@04X01:Qj27X8NP
Qj27X8NPQj27X8NPQiZ1VX6JPIZ1VX6JPIZ1VX6JPIZ6X9^PPJ27X8NPQj27X8NPQj27X8NPQj27X8NP
Qj27X8NPQj0<00@700031`l000Z1O85lPGb1O85l1h407:2QX:6PXJ2QX:6PXHNQQi^JViZKVY^JViZK
VY/<VP04Wj2PX@>P00>QX:406Z0=00@700031``00W/0GH5kPG^1Nh5kPG^1Nh5kPG^1Ni^2X86PPJ21
X86JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1X86PPJ22X9^PPJ27X8NPQj27X8NPQj27X8NPQj27
X8NPQj26P@0>00@700031`X04H402YZ2X8:PPYZ1VX42VP03Vh6J00^J00FKVYZJV`03VP0DViZKVY^J
ViZPQZ26XJ2QX:6PVX4?X005Qj2PX8L00j000hL0000>00@700031`T004EkPG^1Nh5kPG^1Nh5kPG^7
PXN2Qh:7P[ZPVX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPIZ1VX6JPJ21X86PPZ22X8800Z0069Z0
X8NPQj27X8NPQj27X8NPQj27X8NPQQ4010L000<72@0:P@0<Qh:7PXN2Qh:7PZ220Z002KnPVh6KVY^J
V`05VP0HViZJVY^JViZKVY^JViZKVY^JViZPWj260Z001j6PXJ2QX:400YX00h6PX00;X007Qj27X8NP
Q`0B00@700031`L004MkPG^1Nh62Qh:7PXN2Qh:7PXN2Qh:7PXN2^J2jPIZ1VX6JPIZ1VX6JPIZ1VX6J
PIZ1VX6JPIZ1VX6JPIZ1X86PPJ22X8:PPP03X00EP9Z7X8NPQj27X8NPQj27X8JPQZ2601<010L000<7
1P00CWb1PHN2Qh:7PXN2Qh:7PXN2Qh:7PXN2Qh:oX;nP_iZKVY^JViZKVY^JViZKVY^JViZKVY^JViZK
VY^JX9nPQZ26XJ2QX:6PXJ2QX9Z1VP:P00F7X:2PQ`03X007Qj27X8NPQ`0D00@700031`D003F1Qh67
PHN1Qh:7PXN2Qh:7PXN2Qh67PXN2Qj2jX;ZP^X6JPIZ1VX6JPIZ1VX6JPIZ1VX6JP@03VP0API^1VX6P
PJ21X8:PPZ22X8801:004YZ0VXNPQj27X8NPQj26X8JPQQD010L000<71`00?XN2Qh:7PXN2Qh:7PXN2
Qh:7PXN2Qh:7P[nP_j2oX9^JViZKVY^JViZKVY^JViZKVY^JViZKVY^JX9nPQZ260Z007:6PXJ2QX:6P
XJ2QX9Z1VXNPQj27X8NPQj27X8LF00@700031`P00X40;X:1Qh67PXN1Qh:7PXN2Qh67PHN1^Z2jX;ZP
^J2JPIZ1VX6JPIZ1VX6JPIZ1VX43VP0CPIZ1VhJPPJ21X86PPZ22X8:PPP05X00?P9Z0VXNPQj27X8NP
QZ2601L010L000<72P00>h:7PXN2Qh:7PXN2Qh:7PXN2Qh:7X;nP_j2oX;nPViZKVY^JViZKVY^JViZK
VY^JViZKVZ2OX9nPQZ2600:P00fQX:6PXJ2QX:6PXJ2Q00:J00^1VXNPQj27X8NPQ`0H00@700031``0
02Z2PHN1Qh67PHN2Qh:7PHN1Qh6jX;ZP^Z2jX;ZPVX6JPIZ1VX6JPIZ1VX47VP0AQZ21X86PPJ22X8:P
PZ22X8801Z0039Z0VX2JQj27X8NPPAT010L000<73P00DX:7PXN2Qh:7PXN2Qh:7P[nP_j2oX;nP_j2o
VY^JViZKVY^JViZKVY^JViZKVY^JX9nPQZ26X8JQX:6PXJ2QX:6PXJ2QX:6PWh6JPJ27X8NPQh4I00@7
00031a0002B7PH:1Qh67PHN1Qh67X;ZP^Z2jX;ZP_j2iPI^1VX6JPIZ1VX48VP0BX8JPPJ21X86PPZ22
X8:PPZ222:002IZ0VX2PQj27X00J00@700031a800362Qh:7PXN2Qh:7P[nP_j2oX;nP_j2oX;nJViZK
VY^JViZKVY^JViZKVY^OX8JPQZ2600:P016QX:6PXJ2QX:6PXJ2QX:6PX@02VP06PIZ1X8NP6`041`00
0`LD000NPX67PHN1Qh6jX;ZP^Z2jX;ZP^Z2oX9^1VX6JPIZ12IX04hJPPJ21X86PPJ22X8:PPZ22X880
2J001X2JP9Z0X1`010L000<75`00AXN2Qh:7X;nP_j2oX;nP_j2oX;nPViZKVY^JViZKVY^JViZKWj26
X8JPQZ26XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2OPIZ1VX4L00@700031aT001:1Qh6QX;ZP^Z2jX;ZP_j2o
X;l4VP03PIZJ00NJ01F6X8JPPJ21X86PPZ22X8:PPZ22X8802Z001IZ0VX2J01`010L000<76P009aVP
X;nP_j2oX;nP_j2oX;nP^IZKVY^JViZKVZ2JX9nPQZ26X8JPQP02X00IXJ2QX:6PXJ2QX:6PXJ2QX:6P
XJ2QX9Z1VP02PA/010L000<76`002Z2QX:6PXJ2QX:47X00/YH6KPI^1Vh6KVZ26X8JPQZ21X86PPJ22
X8:PPZ22X8:PPZ22X8:PPZ27X8L2X004VX2JP1/010L000<76`004KnP_j2oX;nP_j2oX;nP_j2o00:P
032JX9ZPVZ2JX9nPQZ26X8JPQZ26XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXH6OPIXK00@700031aX0
00fQX:6PXJ2QX:6PXJ2Q00FP02ZUX9^1ViZKVZ26X8JPQZ21X86PPJ22X8:PPZ22X8:PPZ22X8:PPZ27
X8L3X006Qj20VX2J6P041`000`LJ000DX;nP_j2oX;nP_j2oX;nP_j2oX;l2X003VZ2J00>P00V6X8JP
QZ26X8H00Z007Z6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXH6OPAX010L000<76P003j6PXJ2QX:6P
XJ2QX:6PX@03X00/YZ2UVY^JVhJPQZ26X8JPPJ21X86PPZ22X8:PPZ22X8:PPZ22X8:PPZ27X8L4X003
VX2J01X010L000<76@00686P_j2oX;nP_j2oX;nP_j2oX;nP_j2oVP>P00^6X8JPQZ26X8JPQP02X00M
XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX9l00X4I00@700031aT00461XJ2QX:6PXJ2QX:6PXJ2Q
X:6PYZ2VX:F6X8JPQZ26X8JPPJ21X8:PPZ22X8:PPZ22X8:PPZ22X8:PQj27X8NPQ`02X003XH2J01X0
10L000<76@00AH6P_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nOX8JPQZ26X8JPQZ26XJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PXJ2QX:6PX@02PAT010L000<76@006Z2QX:6PXJ2QX:6PXJ2QX:6PXJ2VX:JPYZ2U
0Z00:hJPQZ21X86PPJ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PQj27X8NQQj6PXH006@041`000`LI000K
_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2o00:P00N6X8JPQZ2600:P026QX:6PXJ2QX:6PXJ2QX:6P
XJ2QX:6PXJ2QX:6PXJ2QX9l06@041`000`LI0017X:6PXJ2QX:6PXJ2QX:6PXJ2VX:JPYZ2VX:JPYJ2V
PJ21X86PPJ22X8:PPZ22X8:PPZ22X8:PPZ22X8NPQj27XHNPQj67XH406@041`000`LH000PX;nP_j2o
X;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;l2X00TQZ26XJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2Q
X:6PXJ2Q0Z000iX0000F00@700031aP0026QX:6PXJ2QX:6PXJ2QX:6PXJ2VX:JPYZ2VX:JPYZ2VX:D0
0Z009X6PPZ22X8:PPZ22X8:PPZ22X8:PPZ22X8:PQj27X8NQQj67XJ2J60041`000`LH0019X;nP_j2o
X;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oX;nP_j2oQj6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2QX:6PXJ2Q
X:6PXJ2QX00H00@700031aP004ZQX:6PXJ2QX:6PXJ2QX:JPXJ2VX:JPYZ2VX:JPYZ2VX:FPYZ2UPZ22
X8:PPZ22X8:PPZ22X8:PPZ22X8NPQj27XHNPQj67XJ2QP1L010L000<75`00BknP_j2oX;nP_j2oX;nP
_j2oX;nP_j2oX;nP_j2oX;nP_h:PPZ21QZ2VX:FPYJ2UX:FPYJ2UX;jP_Z2nX:6PXJ2QX:6PXJ2QX:6P
X@0G00@700031aL002JPXJ2QX:6PXJ2QX:6PXJ2VX:JPYZ2VX:JPYZ2VX8N2Qh:7PXN2P@806Z0P00@7
00031aH002N1_j2oX;nP_j2oX;nP_j2oX;nP_j2VX:6PQh:7PXN2Qh:7PZ22QX400`02P@0GX;nP_j2U
X;jPYJ2UX;jPYJ2UX:FPYJ0080041`000`LF000AXJ2QX:JPYZ2VX:JPYZ2VX:H00`001X:1PX62P@Z2
0X4701JP80041`000`LF000=X;nP_j2oX;nP_j2oX007000APXN2Qh:7PXN2Qh:7PZ22QX402P005:JP
YJ2UX:FPYJ2UX:FPYJ2UX:FP80041`000`LE0009X:6PXJ2VX:JP00d000F1PX62P@09PP:12`0CX200
10L000<75@001JJP_j2o01400X803HN2Qh:7PXN2Qh:7PHH03P004:2UX:FPYJ2UX:FPYJ2UX:DQ00@7
00031aD000>P00004`001h62PH:1PX401X800h40000>00nP8@041`000`L[000>PH:7PXN2Qh:7PXN2
QX4A000>X:FPYJ2UX:FPYJ2UX:DQ00@700031b`000B2PH:11X82PA@03:0Q00@700031b`000^2Qh:7
PXN2Qh:7P@0F000;YJ2UX:FPYJ2UX:D08@041`000`L/0006PX62PH:10h800h40000F00ZP8@041`00
0`L/00>200N7PXN2Qh6601T000RUX:FPYJ2UX28010L000<7;@001X62PH:1PP:17006X28010L000<7
;@001hN2Qh:7PXH07P001J2UX:FP028010L000<7;P001H:1PX6202400j0R00@700031bh000F2Qh:7
P@0R0003YJ00024010L000<7;P0018:1PX4T0003X000020010L000<7;`02PTP010L000<7N@041`00
0`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT010L000<7N@041`000`Mi00@700031gT0
10L0080700201`00P0L00001\
\>"],
  ImageRangeCache->{{{0, 127}, {127, 0}} -> {-3.20067, -3.20067, 1.05828, 
  1.05828}},
  CellTags->"Animation"]
}, Closed]]
}, Open  ]],

Cell[TextData[{
  "For more complicated motions the simple transformations are not \
sufficient. If one has only a small amount of data the repeated call of ",
  StyleBox["MVShow3D[]",
    FontFamily->"Courier"],
  " with the",
  StyleBox[" MVNewScene\[Rule]True",
    FontFamily->"Courier"],
  " option may be used. Here is an example of the motion of a mass point at \
the end of a spring. The motion is described by the equations"
}], "Text",
  CellTags->"Animation"],

Cell[BoxData[
    RowBox[{
      RowBox[{"deqn", "=", 
        RowBox[{"{", 
          RowBox[{
            RowBox[{
              RowBox[{
              \(-1\), "+", \(\[CapitalOmega]\^2\ Cos[\[Theta][\[Tau]]]\), 
                "+", \(\[Rho][\[Tau]]\), "-", 
                RowBox[{\(\[Rho][\[Tau]]\), " ", 
                  SuperscriptBox[
                    RowBox[{
                      SuperscriptBox["\[Theta]", "\[Prime]",
                        MultilineFunction->None], "[", "\[Tau]", "]"}], 
                    "2"]}], "-", 
                RowBox[{
                \(Sin[\[Theta][\[Tau]]]\^2\), " ", \(\[Rho][\[Tau]]\), " ", 
                  SuperscriptBox[
                    RowBox[{
                      SuperscriptBox["\[Phi]", "\[Prime]",
                        MultilineFunction->None], "[", "\[Tau]", "]"}], 
                    "2"]}], "+", \(\(\(\[Rho]'\)'\)[\[Tau]]\)}], "==", "0"}], 
            ",", 
            RowBox[{
              RowBox[{
              \(Sin[\[Theta][\[Tau]]]\), " ", \(\[Rho][\[Tau]]\), " ", 
                RowBox[{"(", 
                  RowBox[{
                    RowBox[{
                    "2", " ", \(Cos[\[Theta][\[Tau]]]\), " ", 
                      \(\[Rho][\[Tau]]\), " ", 
                      RowBox[{
                        SuperscriptBox["\[Theta]", "\[Prime]",
                          MultilineFunction->None], "[", "\[Tau]", "]"}], 
                      " ", 
                      RowBox[{
                        SuperscriptBox["\[Phi]", "\[Prime]",
                          MultilineFunction->None], "[", "\[Tau]", "]"}]}], 
                    "+", 
                    RowBox[{"2", " ", \(Sin[\[Theta][\[Tau]]]\), " ", 
                      RowBox[{
                        SuperscriptBox["\[Rho]", "\[Prime]",
                          MultilineFunction->None], "[", "\[Tau]", "]"}], 
                      " ", 
                      RowBox[{
                        SuperscriptBox["\[Phi]", "\[Prime]",
                          MultilineFunction->None], "[", "\[Tau]", "]"}]}], 
                    "+", \(Sin[\[Theta][\[Tau]]]\ \[Rho][\[Tau]]\ 
                      \(\(\[Phi]'\)'\)[\[Tau]]\)}], ")"}]}], "==", "0"}], 
            ",", 
            RowBox[{
              RowBox[{\(\[Rho][\[Tau]]\), " ", 
                RowBox[{"(", 
                  RowBox[{
                  \(\(-\[CapitalOmega]\^2\)\ Sin[\[Theta][\[Tau]]]\), "+", 
                    RowBox[{"2", " ", 
                      RowBox[{
                        SuperscriptBox["\[Theta]", "\[Prime]",
                          MultilineFunction->None], "[", "\[Tau]", "]"}], 
                      " ", 
                      RowBox[{
                        SuperscriptBox["\[Rho]", "\[Prime]",
                          MultilineFunction->None], "[", "\[Tau]", "]"}]}], 
                    "-", 
                    RowBox[{
                    \(Cos[\[Theta][\[Tau]]]\), " ", 
                      \(Sin[\[Theta][\[Tau]]]\), " ", \(\[Rho][\[Tau]]\), 
                      " ", 
                      SuperscriptBox[
                        RowBox[{
                          SuperscriptBox["\[Phi]", "\[Prime]",
                            MultilineFunction->None], "[", "\[Tau]", "]"}], 
                        "2"]}], "+", 
                    \(\[Rho][\[Tau]]\ \(\(\[Theta]'\)'\)[\[Tau]]\)}], 
                  ")"}]}], "==", "0"}]}], "}"}]}], ";"}]], "Input",
  CellLabel->"In[18]:=",
  CellTags->"Animation"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Rho][\[Tau]]\ \)\)]],
  " is the distance from the origin, and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Theta][\[Tau]]\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Phi][\[Tau]]\)]],
  " are the angles in spherical coordinates. ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalOmega]\)]],
  " is the ration of the pendel frequency and the frequency of the spring \
oscillation. A numerical solution of the problem is obtained with ",
  StyleBox["NDSolve[]",
    FontFamily->"Courier"],
  "."
}], "Text",
  CellTags->"Animation"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Delta] = 0.25; \n
    init = {\[Rho][0] \[Equal] 1.25, \[Theta][0] \[Equal] \[Pi] - \[Delta], 
        \[Phi][0] \[Equal] 0, \(\[Rho]'\)[0] \[Equal] 0, 
        \(\[Theta]'\)[0] \[Equal] 0, \(\[Phi]'\)[0] \[Equal] 0.4}; \n
    sol = NDSolve[
        Join[init, deqn /. \ \[CapitalOmega] -> 1/4], {\[Rho][\[Tau]], 
          \[Phi][\[Tau]], \[Theta][\[Tau]]}, {\[Tau], 0, 24  \[Pi]}, 
        MaxSteps \[Rule] 6000]\)], "Input",
  CellLabel->"In[20]:=",
  CellTags->"Animation"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{"{", 
        RowBox[{
          RowBox[{\(\[Rho][\[Tau]]\), "\[Rule]", 
            RowBox[{
              
              TagBox[\(InterpolatingFunction[{{0.`, 75.3982236861550347`}}, 
                  "<>"]\),
                False,
                Editable->False], "[", "\[Tau]", "]"}]}], ",", 
          RowBox[{\(\[Phi][\[Tau]]\), "\[Rule]", 
            RowBox[{
              
              TagBox[\(InterpolatingFunction[{{0.`, 75.3982236861550347`}}, 
                  "<>"]\),
                False,
                Editable->False], "[", "\[Tau]", "]"}]}], ",", 
          RowBox[{\(\[Theta][\[Tau]]\), "\[Rule]", 
            RowBox[{
              
              TagBox[\(InterpolatingFunction[{{0.`, 75.3982236861550347`}}, 
                  "<>"]\),
                False,
                Editable->False], "[", "\[Tau]", "]"}]}]}], "}"}], "}"}]], 
  "Output",
  CellLabel->"Out[20]="]
}, Open  ]],

Cell["\<\
Now the graphics for the mass point at the end of the spring is \
generated by a small function:\
\>", "Text",
  CellTags->"Animation"],

Cell[CellGroupData[{

Cell[BoxData[
    \(SpringGraphics[\[IndentingNewLine]len_?NumericQ, 
        \  (*\ the\ distance\ from\ the\ origin\ *) \[IndentingNewLine]th_?
          NumericQ, \ \ \  (*\ \[Theta]\ *) \[IndentingNewLine]phi_?NumericQ, 
        \  (*\ \[Phi]\ *) \[IndentingNewLine]r_?NumericQ, 
        \ \ \ \ \ \  (*\ 
          The\ radius\ of\ the\ sphere\ for\ the\ point\ mass\ *) 
          \[IndentingNewLine]wind_Integer, 
        \ \ \ \  (*\ the\ windings\ of\ the\ spring\ *) 
          \[IndentingNewLine]halfBox_:  1.5] := 
      \[IndentingNewLine]Module[{z, fac = wind/2, rho, mth, mphi, m, plt, x}, 
        \[IndentingNewLine] (*\ 
          Calculate\ the\ rotation\ matrix\ for\ the\ spring\ *) 
          \[IndentingNewLine]mth = {{Cos[th], 0, Sin[th]}, {0, 1, 0}, {
                \(-Sin[th]\), 0, Cos[th]}} // N; 
        \[IndentingNewLine]mphi = {{Cos[phi], Sin[phi], 0}, {\(-Sin[phi]\), 
                Cos[phi], 0}, {0, 0, 1}}\  // N; 
        \[IndentingNewLine]m = mphi . mth; 
        \[IndentingNewLine] (*\ Calculate\ the\ spring\ *) 
          \[IndentingNewLine]\ \ 
        plt = ParametricPlot3D[
            Evaluate[rho = Min[Abs[\((fac^2*z*\((z - len)\))\)^3], r]; 
              Dot[m, {rho*Cos[2*\[Pi]*wind*z/len], 
                  rho*Sin[2  \[Pi]*wind*z/len], z}]], {z, 0, len}, 
            PlotPoints \[Rule] wind*18, DisplayFunction \[Rule] Identity]; 
        \[IndentingNewLine] (*\ The\ point\ mass\ position\ *) 
          \[IndentingNewLine]x = m . {0, 0, len}; 
        \[IndentingNewLine]Graphics3D[{\[IndentingNewLine]plt[\([1]\)], 
            \  (*\ The\ line\ of\ the\ spring\ *) \[IndentingNewLine]{
              PointSize[0.05], RGBColor[1, 0, 0], Point[x]}, 
            \  (*\ The\ point\ for\ the\ point\ mass\ *) \[IndentingNewLine]{
              SurfaceColor[RGBColor[1, 1, 0]], 
              Cuboid[{\(-r\)/2, \(-r\)/2, 0}, {r/2, r/2, 2  r}]}}, 
          \  (*\ The\ origin\ *) \[IndentingNewLine]PlotRange 
            \[Rule] {{\(-halfBox\), halfBox}, {\(-halfBox\), halfBox}, {
                \(-halfBox\), 2  r}}, 
          ViewPoint -> {0.855685, \(-1.57973\), 1.31644}, 
          Boxed \[Rule] False\[IndentingNewLine]]\[IndentingNewLine]]\)], 
  "Input",
  CellLabel->"In[21]:=",
  CellTags->"Animation"],

Cell[BoxData[
    \(General::"spell1" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(mphi\)\" is similar to \
existing symbol \"\!\(phi\)\"."\)], "Message"]
}, Open  ]],

Cell[TextData[{
  "Here is how the initial position of the mass point looks in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Text",
  CellTags->"Animation"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(gr = 
      Show[SpringGraphics[\[Rho][0], \[Theta][0], \[Phi][0], 0.15, 8, 1.5] /. 
          \ \((init\  /. \ Equal\  \[Rule] \ Rule)\)]; \)\)], "Input",
  CellLabel->"In[22]:=",
  CellTags->"Animation"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .89165 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
1.11022e-16 1.07929 -0.0198787 1.07929 [
[ 0 0 0 0 ]
[ 1 .89165 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .89165 L
0 .89165 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.59036 .43455 m
.60175 .4343 L
s
.57928 .43272 m
.59036 .43455 L
s
.60175 .4343 m
.61217 .43179 L
s
.56975 .42919 m
.57928 .43272 L
s
.61217 .43179 m
.62044 .42712 L
s
.56285 .42456 m
.56975 .42919 L
s
.62044 .42712 m
.62558 .4206 L
s
1 0 0 r
.05 w
.59866 .39389 Mdot
0 g
.5 Mabswid
.59718 .39532 m
.59866 .39389 L
s
.55939 .41955 m
.56285 .42456 L
s
.58939 .39443 m
.59718 .39532 L
s
.58208 .4624 m
.59362 .46244 L
s
.571 .46035 m
.58208 .4624 L
s
.57319 .38532 m
.58939 .39443 L
s
.59362 .46244 m
.60432 .46028 L
s
.56162 .45672 m
.571 .46035 L
s
.60432 .46028 m
.61298 .45594 L
s
.62558 .4206 m
.62696 .41282 L
s
.555 .4521 m
.56162 .45672 L
s
.55981 .41494 m
.55939 .41955 L
s
.57776 .38188 m
.57319 .38532 L
s
.61298 .45594 m
.61858 .44973 L
s
.62696 .41282 m
.6244 .40453 L
s
.57363 .49083 m
.5853 .49115 L
s
.56256 .48858 m
.57363 .49083 L
s
.55192 .44724 m
.555 .4521 L
s
.5853 .49115 m
.59628 .48932 L
s
.55334 .48487 m
.56256 .48858 L
s
.56409 .41152 m
.55981 .41494 L
s
.59628 .48932 m
.60532 .48533 L
s
.61858 .44973 m
.62043 .44216 L
s
.58748 .38211 m
.57776 .38188 L
s
.6244 .40453 m
.61819 .39656 L
s
.54702 .4803 m
.55334 .48487 L
s
.55279 .4429 m
.55192 .44724 L
s
.60532 .48533 m
.61138 .47943 L
s
.59841 .38478 m
.58748 .38211 L
s
.61819 .39656 m
.60913 .38975 L
s
.62043 .44216 m
.61829 .43395 L
s
.57172 .40992 m
.56409 .41152 L
s
.60913 .38975 m
.59841 .38478 L
s
.565 .51986 m
.57679 .52045 L
s
.55396 .51744 m
.565 .51986 L
s
.57679 .52045 m
.58804 .51895 L
s
.54434 .47561 m
.54702 .4803 L
s
.54492 .51366 m
.55396 .51744 L
s
.61138 .47943 m
.61371 .47209 L
s
.58804 .51895 m
.59745 .51531 L
s
.55753 .43984 m
.55279 .4429 L
s
.61829 .43395 m
.61241 .42593 L
s
.58174 .41058 m
.57172 .40992 L
s
.5389 .50917 m
.54492 .51366 L
s
.59745 .51531 m
.60396 .50972 L
s
.54567 .47156 m
.54434 .47561 L
s
.61241 .42593 m
.60354 .41892 L
s
.61371 .47209 m
.612 .464 L
s
.59284 .41365 m
.58174 .41058 L
s
.56556 .43866 m
.55753 .43984 L
s
.60354 .41892 m
.59284 .41365 L
s
.56809 .55036 m
.57958 .54918 L
s
.55618 .54951 m
.56809 .55036 L
s
.54519 .54694 m
.55618 .54951 L
s
.60396 .50972 m
.60677 .50264 L
s
.53664 .50468 m
.5389 .50917 L
s
.57958 .54918 m
.58937 .54588 L
s
.53634 .54314 m
.54519 .54694 L
s
.612 .464 m
.60647 .45595 L
s
.55086 .46888 m
.54567 .47156 L
s
.57588 .43976 m
.56556 .43866 L
s
.58937 .54588 m
.59633 .54063 L
s
.53065 .53874 m
.53634 .54314 L
s
.60677 .50264 m
.60552 .49468 L
s
.60647 .45595 m
.59779 .44878 L
s
.53842 .50095 m
.53664 .50468 L
s
.58712 .44321 m
.57588 .43976 L
s
.59779 .44878 m
.58712 .44321 L
s
.55918 .5809 m
.57092 .58003 L
s
.54718 .57982 m
.55918 .5809 L
s
.5593 .46813 m
.55086 .46888 L
s
.53625 .57712 m
.54718 .57982 L
s
.59633 .54063 m
.59963 .53382 L
s
.57092 .58003 m
.58107 .57709 L
s
.52881 .53448 m
.53065 .53874 L
s
.60552 .49468 m
.60034 .48664 L
s
.52761 .57331 m
.53625 .57712 L
s
.54408 .49866 m
.53842 .50095 L
s
.58107 .57709 m
.58848 .57218 L
s
.5699 .46966 m
.5593 .46813 L
s
.59963 .53382 m
.59884 .52603 L
s
.60034 .48664 m
.59189 .47933 L
s
.52226 .56905 m
.52761 .57331 L
s
.58127 .47348 m
.5699 .46966 L
s
.53107 .53109 m
.52881 .53448 L
s
.59189 .47933 m
.58127 .47348 L
s
.55007 .6121 m
.56203 .61153 L
s
.53799 .6108 m
.55007 .6121 L
s
.58848 .57218 m
.59227 .56565 L
s
.52713 .608 m
.53799 .6108 L
s
.55291 .49835 m
.54408 .49866 L
s
.56203 .61153 m
.57254 .60893 L
s
.59884 .52603 m
.59403 .51802 L
s
.52085 .56504 m
.52226 .56905 L
s
.51872 .60422 m
.52713 .608 L
s
.57254 .60893 m
.5804 .60438 L
s
.53718 .52922 m
.53107 .53109 L
s
.59227 .56565 m
.59194 .55805 L
s
.56378 .5003 m
.55291 .49835 L
s
.59403 .51802 m
.58582 .51058 L
s
.51372 .60012 m
.51872 .60422 L
s
.57527 .50449 m
.56378 .5003 L
s
.58582 .51058 m
.57527 .50449 L
s
.5804 .60438 m
.58468 .59815 L
s
.52358 .56201 m
.52085 .56504 L
s
.55293 .6437 m
.56379 .64145 L
s
.54077 .64257 m
.55293 .6437 L
s
.59194 .55805 m
.58753 .55009 L
s
.5464 .52935 m
.53718 .52922 L
s
.51275 .5964 m
.51372 .60012 L
s
.58468 .59815 m
.58483 .59076 L
s
.58753 .55009 m
.57957 .54258 L
s
.53016 .56057 m
.52358 .56201 L
s
.55753 .53172 m
.5464 .52935 L
s
.5637 .6385 m .53256 .64733 L .5154 .6309 L p .573 .806 0 r
F P
0 g
s
.53724 .63421 m
.53729 .63434 L
s
.56439 .70081 m .5637 .6385 L .53256 .64733 L p .455 .201 0 r
F P
0 g
s
.56379 .64145 m
.57209 .63725 L
s
.53256 .64733 m .53236 .70931 L .56439 .70081 L p .455 .201 0 r
F P
0 g
s
.5154 .6309 m .53256 .64733 L .53236 .70931 L p .975 .669 0 r
F P
0 g
s
.57209 .63725 m
.57686 .63134 L
s
.57957 .54258 m
.56911 .53625 L
s
.56911 .53625 m
.55753 .53172 L
s
.58483 .59076 m
.58082 .5829 L
s
.51597 .59374 m
.51275 .5964 L
s
.53976 .56115 m
.53016 .56057 L
s
.57686 .63134 m
.5775 .6242 L
s
.58082 .5829 m
.57314 .57532 L
s
.5154 .6309 m .54692 .62176 L .5637 .6385 L p .573 .806 0 r
F P
0 g
s
.54692 .62176 m .5637 .6385 L .56439 .70081 L p .974 .664 0 r
F P
0 g
s
.55114 .56392 m
.53976 .56115 L
s
.523 .59274 m
.51597 .59374 L
s
.57314 .57532 m
.56279 .56878 L
s
.56279 .56878 m
.55114 .56392 L
s
.5775 .6242 m
.57391 .61644 L
s
.54713 .68467 m .54692 .62176 L .5154 .6309 L p .447 .189 0 r
F P
0 g
s
.53236 .70931 m .5147 .69348 L .5154 .6309 L p .975 .669 0 r
F P
0 g
s
.5154 .6309 m .5147 .69348 L .54713 .68467 L p .447 .189 0 r
F P
0 g
s
.53297 .59377 m
.523 .59274 L
s
.57391 .61644 m
.56652 .60884 L
s
.54459 .59695 m
.53297 .59377 L
s
.56652 .60884 m
.5563 .60212 L
s
.5563 .60212 m
.54459 .59695 L
s
.56439 .70081 m .54713 .68467 L .54692 .62176 L p .974 .664 0 r
F P
0 g
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[22]:=",
  ImageSize->{288, 256.75},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000100P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Ed10YPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm240<;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:L
W<7De3eDE:b/[:BTY>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEe
MJ_0eL30`9T0I_ooc020P800082006IVc3<cIP2009VIod:JYb8R8_mV<l;2`W9bLP00P<C4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7iP0g??mVc5IF
ERXZ:^KViX61PHR8R=;BdP00W000ba0@G0P8U7EeMOknoXV9RA0@GNKViRXZ:YZJVZY3Z]EEeNGUiGin
OQ0@40XHF`0d003oIP3oV@3oc3<003<0<c<0IS<0VC<0c?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00:T71`5`1`00Z0L90Fl7002W1`/1KPL00:D71@090Fd7002T1`0300L100<10P070Fd7
002U1`801@4200D1K@L00:D70P4200D10P030Fd7002V1`<10`040@8000<11`L0J`L00:H730400`07
1`1[1`00YPL<0@0300L706/7002W1`/100<700L0J`L00:<70P031`T10PL00`071`1Z1`00X@L20087
1P050@<700<01`L0JPL00:0700<01`L020L200H700<01`L0JPL00:0700<01`L02PL00`071`031`03
00L706X7002P1`0300L700/700<01`L00PL00`071`1Z1`00X@L00`071`0;1`0400L706d7002R1`<0
2`L206h7002U1`/000<700L0K@L00;4700<01`L0K0L009l7200;1`0300L706/7002N1`0300L700H7
0`081`0300L706/7002N1`0300L700T70P061`0300L706/7002O1`0300L700X700<01`L00PL00`07
1`1/1`00X0L00`071`0:1`0400L706l7002Q1`0300L700X70P1`1`00XPL=0747002_1`0300L706h7
002M1`L02`L00`071`1^1`00W0L00`071`051`802PL00`071`1]1`00W0L00`071`071`8020L00`07
1`1]1`00W0L00`071`091`801@L00`071`1^1`00W0L00`071`0;1`800PL00`071`1_1`00W@L300`7
0P1b1`00X0L300H7101c1`00X`L600@700<01`L0L0L00:d700<01`L0L0L009X72@0:1`0300L70707
002J1`0300L700H70`081`0300L706l7002J1`0300L700T70P061`0300L706l7002J1`0300L700/7
00<01`L00PL00`071`1`1`00V`L200`700@01`L0L`L009d70`0:1`80M0L00:072`1e1`00Z`L00`07
1`1b1`00Z`L00`071`1b1`00V0L900X700<01`L0LPL009L700<01`L01`L200T700<01`L0L@L009P7
00<01`L020L200L700<01`L0L@L009P700<01`L02PL200<70P1d1`00V@L200`70`1f1`00V`L600@7
0`1h1`00X@L400<700<01`L0M@L00:P700<01`L0M@L00:T700<01`L0M0L009H72@0:1`0300L707@7
002E1`0300L700L70P081`0300L707@7002E1`0300L700T700<01`L01@L00`071`1d1`00UPL200X7
0P031`80M`L009P70`091`<0N@L009/72`1j1`00YPL00`071`1g1`00YPL00`071`1g1`00Y`L00`07
1`1f1`00U0L600d700<01`L0MPL009<700<01`L010L400T700<01`L0MPL009<700<01`L020L200L7
00<01`L0MPL009@70P0:1`800`L207T7002F1`<02@L307/7002I1`/0O0L00:<700<01`L0NPL00:@7
00<01`L0N@L009/70P071`0300L707T7002H1`<000>I00001`L00`071`1h1`00U@L30007NiUkV@3C
00061`0300L707P7002C1`801iT0103Cd`051`0300L707P7002C1`0>07^INiUkVG^I0=?>d`041`03
00L707P7002C1`0309VI00JI00<0dm<00]<01007001k1`00T`L03P1kVG^INiUkV@3Cc]?>0P1m1`00
T`L00`2IV@06V@030=?C00;C00<01`L0O0L009<700l0NiUkVG^INiT0dlkCcP00OPL009<700<0VIT0
1YT00`3Cd`02d`0300L707`7002C1`0?07^INiUkVG^IN`3>dlh007h7002C1`0309VI00NI00D0dm?C
001n1`00T`L03`1kVG^INiUkVG/0c]?>001n1`00T`L00`2IV@07V@050=?Cd`00OPL009<700l0NiUk
VG^INiUk0<kCcP00OPL009<700<0VIT01iT01@3Cdm<007h7002C1`0?07^INiUkVG^IN`3>dlh007h7
002C1`0309VI00NI00D0dm?C001n1`00T`L02@1kVG^INiUkV@020004c]?>07h7002C1`0309VI00:I
1002V@040=?C07h7002C1`0307/000<000VINiUkVG/0cP00OPL009<70P03d`0309VI00BI00<0d`00
OPL009@700`0dlkC09UkVG^INiT207h7002E1`8000?C09T01IT307h7002G1`8000>IN`000`211`00
V0L308D7003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47
003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47
003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47
003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47
003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47003o1b47
003o1b47003o1b47003o1b470000\
\>"],
  ImageRangeCache->{{{0, 287}, {255.75, 0}} -> {-0.000282777, 0.0184142, 
  0.00323032, 0.00323032}}]
}, Open  ]],

Cell[TextData[{
  "The image is send to the OpenGL viewer with the",
  StyleBox[" TransferFrame[]",
    FontFamily->"Courier"],
  " function. A optional second argument will save a bitmap image of the \
rendered scene. The option NewScene\[Rule]True causes that the data transfer \
between ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " and the OpenGL viewer is done while the old image is displayed. "
}], "Text",
  CellTags->"Animation"],

Cell[BoxData[
    \(TransferFrame[gr_, savepng_: - 1] := 
      \((\[IndentingNewLine]MVShow3D[gr, \[IndentingNewLine]\ 
          MVLineTubeSize \[Rule] 0.03, 
          \  (*\ make\ a\ solid\ spring\ *) 
            \[IndentingNewLine]MVPointSphereSize \[Rule] 4.5, 
          \  (*\ make\ a\ sphere\ for\ the\ mass\ point\ *) 
            \[IndentingNewLine]MVPolygonShading \[Rule] MVFlat, 
          \  (*\ sharp\ edges\ for\ the\ cuboid\ *) 
            \[IndentingNewLine]MVNewScene \[Rule] True\[IndentingNewLine]]; 
        \[IndentingNewLine]If[
          savepng \[GreaterEqual] 0,  (*\ 
            Then\ save\ the\ frame\ in\ 
              \*"\"\<\!\(pendel  StyleBox[N, FontSlant -> \
\"\<Italic\>\"]\).png\>\""\ *) \[IndentingNewLine]MVWritePNG[
            "\<pendel\>"<>ToString[savepng]<>"\<.png\>"]
            \[IndentingNewLine]]; )\)\)], "Input",
  CellLabel->"In[23]:=",
  CellTags->"Animation"],

Cell["\<\
Here is the initial position of the point mass in the OpenGL viewer\
\
\>", "Text",
  CellTags->"Animation"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(TransferFrame[gr]; \)\), 
    \(\[IndentingNewLine]\), 
    \(\(MVPasteGraphics[]; \)\)}], "Input",
  CellLabel->"In[58]:=",
  CellTags->"Animation"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0031746 0.0238095 0.0031746 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
.02381 .02381 translate
.95238 .95238 scale
900 string
300 300 8 [300 0 0 300 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004200005A0000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000005200005A00005A00005A00005A0000630000630000630000630000
6300006300006300006300005A0000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A00005A00005A00006300006300006B00006300006B00006B00007300006B0000
7300006B00006B00006B00006B0000630000630000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004200004A0000
5200005A00006300006B00006B00006B00006B00007300007300007300007B00007B0000
7B00007B00007300007300007300007300006B00006B00005A0000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A0000520000520000
6308086300006B00006B00007300007300007B00007300007B08087B00007B00007B0000
8400007B00007B00007B00008408087B00007B00007300007300006300005A0000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A00005200005A0000630000
6B00006B00006B00007300007300007B00007B00007B00007B00008400007B0000840000
8400008400007B00008400008400008400007B00007B00007B00007300006B0000630000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000004200004A00005A00005A0000630000630000
6B00006B00007B00007300007B00007B00008400007B00008400007B00008C0000840000
8C00008400008C00008400008400008400008400007B00008400007300007B00006B0000
6B0000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004200004A00005A00005A00006300006300006B0000
6B00007300007300007B00007B00008400008400008400008400008400008400008C0000
8C00008C00008C00008C00008400008C00008400008400008400007B00007B00007B0000
730000630000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004200004200005200005A00006300006300006B00006B0000
7300007300007B00007B00008400008400008400008400008C00008C00008C00008C0000
9400008C00008C00008C00008C00008C00008C00008400008C00008400008400007B0000
7B00007300005A0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004200005200005200006300006300006B00006B0000730000
7300007B00007B00008400008400008C00008400008C00008C00008C00008C0000940000
9400009400008C00009400008C00009400008C00008C00008C00008C0000840000840000
7B00007B0000730000520000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004200005200005200005A00006300006B00006B0000730000730000
7B00007B00008400008400008C00008400009400008C00009400008C0000940000940000
9C00009400009C00009400009400009400009400008C00009400008C00008C0000840000
8400007B00007B00006B0000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A00005200005A00005A00006B00006B00007300007300007B0000
7B00008400008400008C00008C00009400009400009400009400009400009C00009C0000
9C00009C00009C00009C00009400009C00009400009400009400009400008C00008C0000
8400008400007B0000730000630000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004200004A00005A08085A00006300006300007300007300007B0000730000
8408087B00008400008400008C00008C00009C00009400009C08089C00009C00009C0000
A508089C0000A500009C0000A508089C00009C00009400009400008C00009400008C0000
8C0808840000840000730000730000630000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004200005200005A00006300006300006B00007300007300007300007B0000
7B00008400008400008C00008C00009400009400009C00009C0000A50000A50000A50000
A50000A50000A50000A50000A500009C00009C00009C00009400009400008C00008C0000
8C00008C00008400007B00007300006B0000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003900004A00005200005A00005A00006B00006B00007300007300007B00007B0000
8400008400008C00008400009400008C00009C08009C0000A50800A50000AD0808A50000
AD0800A50000AD0800A50000A508009C0000A508009400009C00008C00009400008C0000
8C00008400008C00007B00007B00006B00005A0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004200004A00005A00005A00006300006B00007300007300007B00007B00007B0000
8400008400008400008C00008C00009400009400009C0000A50000AD0000AD0000AD0000
AD0000AD0000AD0000AD0000A50000A500009C00009C0000940000940000940000940000
8C00008C00008400008400007B0000730000630000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
4200004200005200005200006300006300006B00006B00007B00007300007B00007B0000
8400008400008C00008C00009408089400009C00009C0000A50808A50000AD0800AD0000
B50808AD0000AD0800A50000AD0808A50000A500009C00009C08089400009400008C0000
9400008C00008C00007B00008400007300006B0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
3900004A00005200005A00006300006B00006B00007300007300007B00007B0000840000
8400008C00008C00008C00009400009400009C00009C0000A50000AD0800AD0000B50800
B50800B50808AD0000AD0000AD0000A50000A500009C00009C00009C0000940000940000
8C00008C00008400008400007B00007B0000730000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004200005A00005A00006300006300006B00006B00007B00007B00008400007B0000
8400008400008C00008C00009400088C00009C08089C0000A50808A50000AD0808AD0800
B50808AD0000AD0808A50000AD0808A50000A508089C0000A500009400009C0000940000
9400008C00008C00008400008400007300007B00004A0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
3900004A00005200005A00006300006B00006B00007300007300007B00007B0000840000
8400008C00008C00008C00008C00009400089408089C08089C0808A50808A50808AD0808
AD0808AD0808A50808A50808A50808A50808A50000A500009C00009C00009C0000940000
9400008C00008C00008400007B00007B00007300005A0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000310000
4200004A00005200005A00006308086300006B00006B00007B00007B00008400007B0000
8C08088400008C00008C00009400088C00089408109408089C10109C0808A51010A50808
AD1010A50808A510109C0808A510109C0800A508009C00009C08009400009C0000940000
9408088C00008C00008400008400007B00007300005A0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004A00005200005A00006300006B00006B00007300007300007B00007B0000840000
8400008C00008C00008C00008C00089400089400089408109C10109C10109C1010A51010
A51010A510109C08109C10109C0808A508089C00009C00009C00009C0000940000940000
9400009400008C00008400008400007B00007300006B0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004A00005200005A00006300006300007300007300007B00007300008400007B0000
8400008400008C00008C00009400088C00089408109408109C1018941010A51818A51010
A518189C1010A510189C0808A510089C0800A508089C00009C00009400009C0000940000
9400008C00008C00008400008400007B0000730000630000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004A00005200005A00006300006B00006B00007300007300007B00007B0000840000
8400008400008400008C00088C00089400109400109408109408109C10189C1018A51818
9C10189C18189C10109C10109C08089C08009C08009C0800940000940000940000940000
8C00008C00008C00008C00008400007B00006B0000630000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004A00005200005200006300006300006B00006B00007B00087300007B00007B0000
8400088400008C00008400009408108C00109408189400109C10189410189C18219C1018
A518219C10109C18109C10089C10089C08009C08009400009C08089400009400008C0000
9400008C00008C0000840000840000730000730000630000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004A00005200005A00006300006B00006B00006B00007300007B00007B00007B0000
8400008400008400008C00088C00089400108C00109408189408189C10219410189C1821
9C18189C18189C10109C10089C10089C10009408009408009400009400009400008C0000
8C00008C00008400008400008400007B00006B0000630000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004200005200005200006300006300006B00006B00004A39394231314A39394A3939
5A39425A39395239398400088C00108C00109408188C00109408189408109C1821941018
9C18189410109C18109410089C10089408009C10089400009C08009400009400008C0000
8C00008400008C00007B00008400007300006B00005A0000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004A00005200005A00005A00004239394A39425A4252634252634A5A6B525A735263
735A637B5A6373525A735252734A4A734A4A5A42399400108C08109408189410189C1018
9410109410109410089C10089408009408009408009408009400009400008C00008C0000
8C00008400008400008400007B00007300006B0000520000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000390000
4200004200005200003931394A394A524252634A6363526B6B5A736B5A6B736373736373
846B7B7B6B73846B7B8463738C6B73846B6B846B6B7B5A5A7B5A52634239941010941010
9C10109408089410089408009C10089408009408008C00009408008C0000940000840000
8C08088400008400007B00007B00006B00006B0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
3900004200003931394A39525242635A526B635A7363637B6B637B736B84736B847B7384
7B738C84738C847384847B8C8C73848C7B848C737B8C737B8C73738C6B637B5A526B4A42
9408089408089408009408009408009408008C00009400008C00008C00008C00008C0000
8400008400007B00007B00007300007300005A0000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39000031313142425A4A42635A52735A5A736363846363846B6B8C636B8473738C6B6B84
73738C73738C847B947B7B8C8C8494847B8C94849494848C94848C8C7B7B947B7B8C6B63
84635A7352429408008C00009408008C00009408008C00008C00008400008C0000840000
8C00007B00007B0000730000730000630000520000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000039394A424A6B525273525A7B5A63845A63845A6B845A63845A6B84636B84636B84
636B846B738C6B738C737B947B7B948484948484948C8494948C94A594949C8C8C9C8484
947B7394736B8463528C08008C08008C08008C00008C00008400008C0000840000840000
8400007B00007300007300006B0000630000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39394A39395A4A52734A5A7352637B52637B526B7B4A63734A6373425A6B42636B425A6B
4A5A73425A6B526B7B5A6B846B738C6B738C737B947B7B94948C9C948C9CA59C9CAD949C
AD9C94AD8C849C84738463528C08088400008C00008400008C08088400008400007B0000
8400007B00007300006B00006B00005A0000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
393142424A6B42527352637B4A637B4A63734A63635A00006300006B00006B00006B0000
7300007300007300007B00007B00004A5A6B52637B63738C737B8C8484948C8C949C949C
A59C9CBD9C94AD94849C847B8C7363735A428400008400008400008400007B00007B0000
7B00007B00007300006B00005A0000520000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39394A394A6B4A637B4A63735A6B7B5263635A00005A00006300006300006B00006B0000
7300006B00007B00007300007B00007B00008400007B0000526B7B6373847384947B8494
94949C9C9494B59C94AD9484A58C7B8C73637B634A7B00008400007B00007B0000730000
7B00007300006B00005A00005A0000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
39314242526B4A6373526B7B5A6B736B736B5200005A00005A00006300006300006B0000
6B00006B00007300007300007300007B00007B00007B00007B00007B000052637B63738C
737B948C8C9C949494A59C94A59C8CA58C7B947B6B735A427B00007B0000730000730000
7300006B00006300005A0000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000394A634A637B526B736B7B7B4200004A00004A00005A00005A00006300005A0000
6B08086B00007300006B00007300007300007B00007300007B08087B00007B0000394A63
63738C737B8C848C9C8C8C949C9C94A59C8C9C94848C7B637B00007300007300006B0000
7308086300006300004A0000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000394A634A63735A737B6B7B7B7B847B4200004A00005200005A00005A00005A0000
6300006B00006B00006B00006B00007300007300007300007300007B00007300007B0000
7B00005A6B846B7B8C848C948C9494949C949C948C948C73846B5A7300006B00006B0000
6300006300004A0000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000042526B526B736B7B84737B7B4200004200004A00004A00005A0000520000
6300005A00006B00006300006B00006B00007300006B00007300007300007B0000730000
7B000073000052637B6B738C7B8C948C8C9494949494948494846B6B5A426B0000630000
6300004A0000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000004A5A735A737B7384843900003900004200004A00004A0000520000
5A00005A00006300006300006300006B00006B00006B00006B0000730000730000730000
6B00007300006B00004A5A736B7B8C848C948C8C9494948C948C7B8C7B635A00005A0000
4A0000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004239394A31395A42525A424A634A5263424A6B4A526B4A527352526B4A4A
734A4A5A39395A00005A00006300006300006300006300006B00006B00006B00006B0000
6B00006B00006B00006342295A6B846B7B8C848C94848C8C94948C8C846B6B5A4A4A0000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
4239424A394A5242525A4A6363526B6B5A6B6B5A6B735A6B7363737B63737B6373846373
84636B8463637B5A5A7B525273524A5A42425A0000630000630000630000630000630000
63000063000063000063000042526B63738C73848C848C948C8C8C948C847B735A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000042314A
4A425A52425A5A526B6352736B637B6B637B736B8473637B7B73847B6B8484738C84737B
8C7384846B7B8C737B8C6B738C73738463638463526342395A00005A00006300005A0000
6300005A00006300005A00006300004A5A736B7B8C7B848C8C9494848C7B847B63000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000003931424A4263
4A4A6B5252735A5A7B6363846363846B6B846B6B8473738C73738C7B738C7B738C847B8C
847B8C8C7B8C8C7B8C947B8C947B84947B7B9473738C6B637B5A526342425200005A0000
5200005A00005200005200004A0000424A6B5A6B8473848C848C8C848C847B7B63000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039314242395A424A6B
525A7B525A7B5A638452637B636B8C5A6B84636B8C636B846B73946B738C737B9473738C
848494847B948C84948C8494948C9494848C9C848C947B7B9C7B7B8C6B63846352634239
5200004A00004A000042000042080842425A5A6B846B7B84848C947B847B848473000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003939524242634A5273
4A5A7B52637B52637B52637B52637B52637B4A637B526B84526B845A6B845A6B8463738C
6B738C737B947B7B94848494848494948C9C9C9494A594949C8C8C9C8484947B73947363
7B5A4A42000039000000000000000042425A52637B6B7B8C73848C848C847B8473000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039395A425273425273
52637B4A63734A6373425A6B000000000000000000000000000000000000000000394A5A
4A637B526B8463738C63738C7B849484849494949C9C949CADA5A5AD9C94AD9C94A58C7B
947B6B7B5A4A00000000000000000039395A52637B63738C73848C7B84847B8473000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039395A4252734A637B
4A6373526B735A6B5A000000000000000000000000000000000000000000000000000000
0000000000000000004A526B5A6B846B7B8C738494848C9C94949CA5A5A5ADADA5BDA594
AD8C849C8473846352000000393139424A6352637B63738C737B8C7B8484737B6B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039395A425A734A6373
5A6B7B5A6B6B737363735A42000000000000000000000000000000000000000000000000
000000000000000000000000000000424A635A738C63738C7B849C848C9C9C949CA59C94
BDA594AD94849C8473846B5242425A4A526B5A6B8463738473848C737B735A634A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000039425A42526B4A637B
5A737B6B7B73737B6B847B637B634A000000000000000000000000000000000000000000
00000000000000000000000000000000000000000042526B5A73846B7B947B84948C949C
9C9C94ADA594A594849C8C738C735A5A637B5A6B846B7B8C7384846B7B73000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000042526B4A6373
63737B6B7B7B7B847B847B6B8C7B6B846B5A7B5A4A000000000000000000000000000000
0000000000000000000000000000000000000000000000000000004A5A736B7B947B8494
8C949C8C9494A5A5949C9484948C737B634A6B738C6B7B84738484637363000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A5A73
5A6B7B6B7B847B847B848C848C847B8C84738C7B6B8C73637B5A52634A42000000000000
0000000000000000000000000000000000000000000000000000000000004A5A73637B8C
7384948C949C8C94949C9C94948C7B8C846B6B7B8C6B7B8463736B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039395A
526B7B63737B7384847B84848C8C8C8C8C84948C848C847394847B8C736B8C6B637B5A52
7B5A526B4A4A6B4A4A6342426342425A39425A39425A4242634A52634A5263526342526B
637B8C73848C8C949C8C8C8C9C9C8C8C847384735A5A736B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A63735A737B4A39394239398C8C8C94948C9C9C8C9C9C84A594849C8C7B94847B
947B7B8C73738C6B73846B73846B6B8463737B636B7B63737363737363736B637B736B84
424A6B63738C73848C8C94948C948C9494848C846B73634A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A39424A3142
5A394A5A424A6B4A5A6342526B4A526B4A527352526B4A4A7352526B42425A4242A59C94
A5949494848C9C8494947B8C947B8C847384847B847B7384847B8C7373847B738C737384
737B944A5A736B7B8C7B848C8C94948C8C84948C7B847B5A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000003939394A395252425A5A4A63
634A636B526B6B5A6B735A73735A737B63737B6373846B7384636B8463637B5A5A7B5A52
73524A634242948C948C8C948C8494848494847B8C847B947B7B8C7B7B94737B94737B94
6B738C39424A526B846B7B8C7B8C948C8C8C8C8C8484846B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003931424A425A4A42635A526B5A526B
6B637B6B637B736B8473637B7B6B847B6B84847384846B848C738484737B8C737B8C6B73
8C737384636384635A7B524A5239427B8494848494737B8C7384946B7B8C6B7B8C637384
637B84526B6B4252736373847B8494848C8C8C8C8C847B6B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039314242395A4A4A6B5252735252735A5A7B
636384636B846B6B846B6B8C73738C7B738C7B738C7B7B8C847B948C7B948C7B8C8C7B8C
947B84947B84947B7B94736B84635A845A4A637384637B845A737B526B73426363000000
00000000000039425A52637B6B7B8C7B8C94848C84848473000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000042395A4242634A52734A52735A6384525A7B
5A6B8C5A6384636B8C636B8C6B73946B738C737B947373947B7B947B7B948C849C8C8494
948C9C9484949C8C94947B849C847B8C736B8C6B5A7B5242000000000000000000000000
00000000000042425A4A5A7B6B7B8C73848C848C8C7B8473736B52000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000003939393942634252734A52734A5A7B4A637B526384
4A637B4A637B4A637B526B845263845A6B845A6B845A738C63738C737B947B7B9484849C
84849C8C8C9C948C9C9C949C9C848C9C84849C847B947B6B846B5A735242000000000000
0000000000003939524A5A7B63738C73848C7B8C847B847B6B6B52000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000393142394A6B3952734A637B425A734A637B425A63
000000000000000000000000000000000000000000394A5A4A6373526B8463738C6B738C
7B849C7B84948C8C9C94949CA59CA5A59C94A594949C8C849C8C7B8C73637B634A000000
00000000000042395A4A5A7363738C6B7B8C7B8C8C738473000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000393952394A6B425A734A6373526B73526B63736B52
0000000000000000000000000000000000000000000000000000000000000000004A5A73
526B8463738C737B94848C9C8C949C9CA5A5A5AD9CADAD9CADA594A59484947B637B634A
00000042394A424263525A7B6373846B848C738484737B6B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000424A6B425A73526B7B526B736B7B73737363
7B6B52634239000000000000000000000000000000000000000000000000000000000000
0000000000005A6B8463738C7B849C848C9C949CA59CA59CB5AD9CB5A58CAD9484947B6B
846B5239395A525A7352637B6B7B8C6B7B847384845A634A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000394A634A6373526B73637B7B6B7B737B8473
847B6B8473637B634A634239000000000000000000000000000000000000000000000000
000000000000000000424A635A6B846B7B8C7B84948C8C9C94949CADA59CADA58CA59484
94846B846B5252637B5A6B84637B8C6B84846B7B6B000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000039526B526B7B63737B7384847B847B
8C8C7B8C8473948473846B5A846B5A6B4A42523939000000000000000000000000000000
00000000000000000000000000000042526B5A738C6B7B8C8C949C8C94949C9C9CA59C8C
A5948494846B7B6B4A5A73846B7B8C6B7B7B637363000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003939524A5A735A737B6B7B847B8484
848C848C8C84948C7B94847B948473947B6B846B6384635A7B5A5273524A6342425A4242
5242425242424239394A39395A39425A424A5A424A5A6B846B7B94848C9C8C9494949494
9C9C8C9C94848C7B6B6B7B8C637B845A736B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000425A6B6373846B7B84
7B8C8C848C8494948C94948C9C9C8C948C849C8C8494847B947B7B8C73738C737384636B
846B6B7B5A637B636B735A6B7B6373735A6B7363736B5A735A6B84637B8C848C9C848C94
949C9494948C948C7B7B6B525A6B63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004252635A6B7B
6B7B847B8C8C84948C8C9C9494A594A5AD94A5A594A59C949C94949C8C949C848C94848C
947B848C7B848473848473847B73847B73847B6B8473738C6B6B8452637B637B8C7B8C94
8C949494948C9494848C846B6B5A42000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000003931394A39425239396339425A3942
6342425A39395A39424A31398494948C949494949C94949C9C9CAD9C94A59C949C948C94
948C9C8C84948C8494847B8C847B947B738C7B7B9473738C7B7B946B738C52637B637B8C
848C9C848C8C94948C8C84738C7B63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004239424A394A52394A5A4252634A636B52636B5263
73526373526373525A73525A7B525A7B5A5A734A4A634242848C948C8C9C848C94848C94
848494848C948484947B84947B84947B8494737B8C6B7B8C6B7B8C6B7B8C394A6B5A738C
6B848C849494848C8C8C948484846B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000003931394A394A4A395A5A4A635A4A6363527363526B73637B6B5A73
7B637B736373846B7B7B63738C6B7B846B738C6B6B845A63845A5A7B524A6B42426B7B8C
737B8C6B7B8C73848C6B7B8C6B7B8C637384637B8C526B7B526B734263630000004A5A73
6B7B8C73848C849494848C848C8C7B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000003939394239524A395A524A6B5252735A5A7B635A7B6B63846B6384736B8C
736B847B6B8C7B738484738C8473848C73848C73848C737B8C73738C6B738C6B6384635A
845A525A393939525A425A63000000000000000000000000000000000000000000394263
526B846B7B8C7B8C94848C8C848C7B5A5242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000042395242395A4A4A734A5273525A7B525A7B6363845A6384636B8C636B84
73739473738C7B7B947B738C847B94847B948C84948C7B8C94848C947B84948484947B7B
947B7384635A845A4A000000000000000000000000000000000000000000000000393952
4A637B63738C738C947B8484848C7B6B6B4A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000039395A424A6B4252734A5A7B4A5A7B5263845263845A6B8C5263845A6B8C
636B8C6B73946B738C6B7394737B947B7B9C847B948C849C8C849C948C9C9484949C8494
9C84849C847B94736B8C6B5A73524A00000000000000000000000000000000000039395A
4A5A7B637B8C6B848C7B8C8C7B847B6B7352000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000039313939426B394A6B425A7B425A7B4A637B425A7B4A637B425A734A637B425A73
4A637B4A637B526B84526B8463738C63738C73849C7B7B948484A584849C948CA5948C9C
9C949C9C848C9C8C8C9C7B7B947B6B8463526B4A42000000000000000000000000393952
4A5A7B5A73846B848C7384847B847B636B52000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000393942394263425273425A734263734A6373526B5A735A42000000000000000000
000000000000000000000000394A5A4A5A73526B8463738C6B7B947B849C7B849C8C8C9C
94949C9C9CA59C9C9C9C948C9C84849C847B94736384634A00000000000039394A424263
4A5A7B5A738C6B7B8C738C8C738473000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000039426B3952734A637B4A6B735A736B6B7363847B637B634A5A4239000000
0000000000000000000000000000000000000000004A2931526B84526B846B7B94737B94
8C94A58C9C9C9CA5A5A5AD9CADA59CA59C8CA594848C736384634A393142424263424A6B
5263845A73846B848C6B84846B7B6B000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000393952425A734A63735A737B637B7373847384846B8C846B846B5A846352
634242000000000000000000000000000000000000000000000000000000424A635A6B8C
6B7B947B8C9C84949C94A5A5A5B5A5ADBDA5ADA58CA5947B947B6B84634A4A4A6B52637B
526B84637B8C637B846B7B73000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000394A634A637B52737B6B848473847B848C848C8473948C7B8C7B6B
8C736B846352846352634A42523939000000000000000000000000000000000000000000
4A5A735A738C7384947B8C9C8C9CA59C9C9CADAD9CADA58CA59C84947B6B846B52526384
63738C6373846B84845A6B63000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000042526B526B7B637B847384847B8C848C8C84948C84948C84
94847B9C847B947B738C736B846B6384635A7B5A527B525273524A6B4A4A63424A6B424A
63424A634A524A637B637B8C73849484949C949C9CA5A59CA5A58CA5948494846B846B52
637B8C637B84637373000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000394A63526B7B6373847B8C8C7B8C8C8C948C8C948C
9C9C949494849C948C948C849C8C84947B7B947B7B8C73738C737B846B73846B737B636B
846B7373637373637339426363738C6B7B9484949C8C8C949CA59C9C9C8C9C94848C7B6B
637B84526B6B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004A637B5A738473848C7B8C8C849494
8C9C949CA5949CA594A5A59C9C9C9C9C9C9C9C8C949C8C949C849494848C8C7B84847384
84738484738C7B73847B738C3942635A738C6B7B94848C9C8C9494949C9494948C948C7B
84735A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039425A4A6373637B846B7B8C
7B8C94848C948C949C8C949C949CAD9494A59C9CA5948C9C948C9C8C84948C8C9C847B94
8484947B7B8C7B7B9473738C737B943942635A738C6B7B8C84949C848C94949C948C8C84
8C846B6B5239000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004239394A39394A39395239425239425239425239394A5A6B
526B7B637B846B7B8C7384947B8494848C9C848C94848C9C848494848C9C7B849484849C
7B84947B8494737B94737B946B7B8C637B8C637B845A6B846B7B8C7B8C9484949494948C
8C8C73847B5A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000042393952394A52394A5A42525A4252634A5A634A5A73525A6B4A527352526B4A4A
73525263424263424242526B5A73845A7384637B8C6373846B7B8C637B846B7B8C637B8C
6B7B8C637384637B8C5A6B7B52737B4A636B000000394A6B637B946B848C849494848C8C
949484847B63000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000393939
42394252425A52425A5A4A63634A6B6B526B6B5273735A73735A737B63737B63737B636B
7B636B84636B845A637B5A5A7B524A6B4242000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004A637B637B8C738494848C94
848C84848C73000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000042394A
4A42634A42635A52735A5273635A7B635A7B6B63846B63847B6B84736B7B8473847B6B7B
8C7384846B7B8C73848C6B738C6B7384636384635A73524A634242000000000000000000
000000000000000000000000000000000000000000000000425273526B846B84947B8C94
84948C848C7B000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000393952423963
4A426B4A527352527B5A5A7B5A5A846363846363846B6B8C6B6B8C73738C7B738C847B94
84738C8C7B8C8C7B8C8C7B8C8C7B84947B8494737394736B8C6B638C635A6B4A42000000
00000000000000000000000000000000000000000000000039425A4A5A7B637B8C738494
7B8C8C848C7B6B6B4A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000393963394263
4A527B425273525A844A5A7B5A638C5263845A6B8C5A6B8C6B73946B738C73739473738C
7B7B9C7B7B948C849C8C849494849C94848C94848C947B849C8484947B73947363846352
634242000000000000000000000000000000000000000000424A63425A7B637B8C6B7B8C
7B8C8C7B84736B6B52000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000393942393963394A73
4252734A5A7B425A7B425A7B4A637B4A63844A637B4A63845263845A6B8C5A6B8C637394
6B7394737B9C7B7B9C84849C8484A5948CA5948CA5948C9C9C8C9C9C8C8C9C847B947B73
8C6B5A84634A0000000000000000000000000000003939424242634A5A7B5A738C6B7B8C
738C8C73847B636B4A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000039426B395273
425A7B395A73426373425A6B7B7352735A42000000000000000000000000425A6B42637B
526B84526B84637B946B7B947B849C7B849C8C8CAD8C8CAD9C94B59C8C9CA5949C9C8484
9C8C8494736B8C735A735242000000000000423942393952424A6B4A5A7B63738C637B84
738C8C6B7B73000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000394263425273
425A734A6B73526B6B6B7B6B7B7B63847B63846B52735A4A000000000000000000000000
000000000000424A63526B845A738C6B7B9473849C848CA58C8CA59494A59C9CA5A59C9C
9C8C949C8C849C847B947B6B84634A42394A526B7342426B4A52735263845A738C6B7B8C
6B84846B7B63000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000394A6B
4A6B7B526B73637B7B737B73848C7B8484739484738C73638C736384635273524A523939
000000000000000000000000000000425A7B5A738C63738C7B849C848C9C949CA594A59C
A5B5A59CA594A59C8C9C8C7B9C847384634A4A4A6B4A4A73526384526B84637B8C637B84
6B7B73000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000039395A
425A7352737B637B7B7384847B8C848C8C848C8C7B948C7B94847B9484738C736B8C6B63
84635A84635A73524A6B4A4A634A425A4242523939424263526B8C6B7B947B849C8C949C
94A5A5A5B5A5A5B59CADAD949C948494846B84634A5263845A6B8463738C5A7B84637B7B
5A6B5A000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
3942634263735A7384637B847B8C8C848C8494948C8C94849C948C948C849C948494847B
9C84848C7373947373846B6B8C6B6B7B636384636B735A637B5A6B4252736373946B7B94
8494A58C9C9C9CADA5A5AD9CADAD949C947B948473846B4A63738C5A7384637B84526B63
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000425A6B52737B6B7B847B8C8C7B8C8C8C9C948C9C949CA5949CA5949C9C94
9C94949C94949C8C949C8C8C94848C94848C8C7B848C7384846B847B6B847B6B84526B8C
637B8C73849C84949C949C9C9CAD94A5A58C9C9C7B94846B846B4A5A737B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000393952526B7B5A73846B848C73848C849494849494949C9C949CA5
9CADAD9CA5B5ADA5AD9C9CA59C9CA5948C9C948C9C847B948C849C847B94847B947B738C
42526B5A738C73849C7B8C9C8C9C9C949C949CA58C948C7B948C73000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000042526B526B7B637B8473848C7384947B8C9C
84949C8C94A58C949C8C8C9C848C9C848C9C84849C84849C7B84947B8494737B94737B94
737B94424A6B5A738C6B8494848C9C8C949C8C9494949C8C948C7B84735A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000424A6352738452737B
637B8C6B7B8C7384946B7B8C7384947384947B8C9C7384947384946B7B8C6B7B9463738C
6B7B8C5A7384424A6B526B8C6B84947B8C948C949C8C948C94948484846B735A42000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A3939
4A39395239395239395239395239395A3939000000000000000000000000000000000000
0000000000004252634A5A6B4A6373526B7B52737B5A7384526B7B526B734A636B425A63
000000000000000000424A6B52738C6B84947B8C948C94948C948C8C8C737B735A000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000042394242393952394A52394A
6342525A42526B4A5A63425273525A6B4A527352526B4A4A6B4A4A633939000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004A52735A738C7384947B8C948C9494848C7B84846B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000004239424A394A52425A5242635A4A63
634A6B6B52736B526B735A73735A737B63737B5A6B7B5A6B7B5A63845A637B5A5A845A5A
734A4A634242000000000000000000000000000000000000000000000000000000000000
00000000000000000000000039394A526384637B8C738C947B8C8C849484848473000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000042394A4239524A42634A42635A52735A5273
635A7B635A7B6B63846B637B7B6B8473637B846B847B6B7B8C7384846B7B8C737B8C6B73
8C6B6B845A5A845A526B4242000000000000000000000000000000000000000000000000
0000000000000000000000000000004252735A738C637B8C7B94947B8C84849484000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039395242426342426B4A527352527B5A5A84
5A5A8463638463638C6B6B8C736B8C7B738C7B738C84738C84738C8C7B8C8C7B8C947B8C
947B8494737B8C6B6B8C6B6384635A7B524A633942000000000000000000000000000000
000000000000000000000000000000394A634A637B637B8C73848C7B8C8C7B8C7B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000039314239396339426B4A527B42527B4A5A844A5A7B
5A638C5263845A6B8C5A638C6B73946B6B8C737394737394847B9C847B948C849C8C7B94
948494947B8C9C848C947B7B9C7B7B94736B947363845A4A634242000000000000000000
000000000000000000000000423942394A634A63845A738C7384947384847B847B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000039394A393963394A73395273425A7B425A7B425A7B
425A7B4A63844A6384526384526B8C5A6B8C6373946B73946B7B94737B9C7B7B9484849C
8C84A5948CA594849C948C949C848C9C8C8C9C847B947B6B8C6B5A7B524A000000000000
00000000000000000000000039395242526B4A637B5A7B8C6B848C7384846B7B73000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039426B394A73425A7B395A6B426373395A63
847B637B735A846B5273524242525A39526B4A63844A63845A73946373946B7B9C737B9C
848CA58484A59494AD948CAD9C94AD9C8C9CA5949C9C84849C847B8C73638C6B5A634242
00000000000042394A39395242426B424A73526B845A738C6B848C6B847B6B8473000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000039395A395273425A734A6B735A73736B7B73
7B84738C84738C846B8C7B638C6B5A84635273524A634242000000425A734A63845A738C
637B9473849C7B84A5848CAD8C94B59C9CBD9C94ADA5949C9C8C8C9C8C84947B73947363
84634A4A395242425A4A426B4A4A734A5A84526B84637B8C637B846B847B000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000394A6342637B526B7B637B7B6B7B7B
848C84848C7B948C7B8C847B9C8C7B947B73947B6B846B638C6B6384635A7B5A526B524A
4A6384526B8C637B94737B9C848CA5848CA5949CAD9C9CA5A5A5A59C9494A5948C9C847B
9C8473846B525252734A4A735263845263845A738C5A73846B84845A6B5A000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000039526B4A6B7B5A737B6B8484
7B8C8484948C8C948C94948C9494849C948C948C849C8C849C847B947B7B8C73738C7373
846B6B846B6B394263526B845A739473849C7B8C9C8C9CA594A5A5A5B5AD9CA59C9C9C94
9C94849C8473846B52526384526B845A738C5A73845A7B7B5A736B000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003939524A6B7B5A737B
6B848C73848C84949484948C949C94949C949C9C949494949C9C949C9494A59494948C8C
9C8C8C8C7B84947B8C8473848C73844A6B846B7B9473849C849CA58CA59C9CB5A59CB59C
A5AD949C94849C8C73846B525A738C5A73845A7B84527373000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000042526B
4A6B7B637B846B848C7B8C947B94948C9C9C8C9C9C94A5A59CADADA5ADB5A5ADADA5A5AD
9C94A5948C9C8C849C8C8494847B94847B944252735A738C6B84947B949C8C9C9C94ADA5
9CB59CA5AD949C9C84948473846B5252737B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000004A6B7B526B7B6B848C6B848C7B8C9C7B8C948C9CA58C9CA58C9CA5848C9C
8C8CA5848C9C8C8CA584849484849C7B7B947B849C737B94526B8C5A738C738C9C848C9C
94A59C9CAD949CAD9494947B948C73846B52000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A5A6B526B7B5A7384637B8C7384946B8494738494
7384947B8C9C7B849C7B849C73849473849C6B7B946B7B946B7B94424A6B5A738C738494
84949C8C949C94A5949CA58C94947B948C73735A39000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000003942524A5A6B4A6373
527384526B7B5A73845A73845A73845273845A73844A6B73526B7342636B424A6B52738C
6B84947B8C9C8C949C8C9494949C8C8C8C73847B5A000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000424A6B
52738C6B84947B8C9C8494948C948C8C948484846B6B5A42000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000004239424239394A39425231425A42425A39395A39425A3939
6342425A39395A3942000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
424A7352738C6B84947B8C948C94948C94848C8C73736B52000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000039394242394252394A5239525A425A63425A634A5A634A5A6B4A5A
73525A73525A734A52734A526B424A634242000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A5A7B5A738C6B84947B8C9484948C848C7B7B7B63000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000042394A42394A4A425A5242635A4A6B5A4A6B63527363526B735A73735A73
7B637B7B5A738463737B5A638463637B5A5A845A5A735252734A4A633939000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000039426B526B8C5A7B8C7B8C947B8C8C8494847B846B000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000003939423939524239634A426B524A735252735A5A7B635A7B6B5A7B6B637B736384
736B847B6B847B6B84846B84846B7B8C6B7B8C6B738C6B738C63638C635A7B52526B4A42
000000000000000000000000000000000000000000000000000000000000000000000000
00000039394A425A7B5A738C63848C738C8C7B8C847B8473000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000039394A39396B39426B4A4A734A527B525A8452527B63638C6363846B6B8C6B6B8C
737394736B8C7B73947B738C8C7B8C84738C8C7B8C8C7384947B848C73739473738C635A
8C635A734A42000000000000000000000000000000000000000000000000000000000000
0000003939524252734A6384637B946B848C7B8C846B7B63000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000393952393963394A734252734A527B42527B4A5A845263845A638C5A638C636B8C
636B946B73947373947B7B94847B94847B9C847B948C849494848C94848C947B84947B7B
94737394736B8C6B5A7B5A52633939000000000000000000000000000000000000000000
393942424263395273526B845A7B8C6B848C6B847B6B7B63000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000039395242426B394A7342527B395273425A7B395A73426384425A7B526B8C526384
5A6B945A6B8C6B739C6B7394737B9C737B948484A584849C948CA58C84949C8C9494848C
9C8C8C9C847B9C847B94736B947363734A4200000000000000000000000042394239394A
42425A4A637B4A5A7B4A6384637B8C6384846B847B63735A000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000039426B395273395A7342637352736B737B6B7B84738C84738C846B8C7B6B
39526B4263844A63845A6B8C6373946B7B9C6B7B9C7B849C7B84A58C8CAD948CAD9494A5
948C9C9C8C9C9C8C8CA58C8C9C7B739473638C6B526B42424A39424A4252423952424263
42426B4252734A5A7B526B8C5A7B8C638484637B73000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000039526B4A6B7B4A6B73637B7B6B847B848C84848C7B948C848C8C7B
9C8C8494847B94847B8C736B42637B426384526B8C5A73947384A57B84A5848CB58C8CB5
949CBD9494B59C9CB59C949CA594949C847B9C84738C6B5A845A524A4A63524A6B4A4A6B
52527B4A5A84526B84526B84637B8C637B7B000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000425A7352737B5A7B7B6B84847B8C8484948C8C948C94948C
94948C9C948C948C849C8C8C9C8C849C8484947B7B3942524A6B8C637B946B7B9C7B8CA5
848CAD8C94BD949CB5A5A5B5A59CA5A5949C9C8C8C9C847B947B6B8C735A5A3939525A7B
526384526B8C526B845A73845A7B7B526B5A000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000394A634A6B7B5A73846B848C73848C84949484948C
949C9C949C949CA59C949C949CA59C9C9C9CA59C9C9C94949C949494848C4A6B8C5A738C
73849C7384A58494AD8C9CA59CADB59CA5A5A5A5A59C948CA5948494846B94735A5A3939
5A6B8C526B845A7B8C52737B526B63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000042526B4A6B7B5A7B846B848C738C94
7B8C94849C9C8C9CA594ADAD94ADADA5B5AD9CADAD9CA5AD949CAD9494A58C8CA58C84A5
4263845A738C6B849C7B94A5849CA594A5A59CB5AD9CAD9C9C9C949C9C8C9C8C738C735A
5A3939527384527373000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000042637B526B7B
637B8C6B7B8C738C94738C94849CA584949C8494A57B8C9C848CA5848CA5848CA57B849C
84849C7B7B944A6B8C5A738C7B8CA57B949C8CA5A594AD9C9CBDA59CAD949C9C84948473
8C7B5A5A4239000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
00000000000042637B5273845A7384637B8C637B8C6B84946B849473849C73849C73849C
7384946B84946B7B94424A6B52738C6B849C7B8C9C849C9C8CA59C9CB59C9CAD8C9C9C84
948C738C735A000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000004252634A6373527384526B7B
5273844A6B7B526B7B4A63734A6373426384637B946B849C8494A58C9C9C9CAD9C94A58C
9C9C848C846B84735A000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000424A6B5A738C6B84947B949C84949C94A594
949C8C94947B8C846B000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000394A6B5A73946B84947B949C849494
949C948C94848C8C7B7B6B52000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000424A6B5A738C6384947B8C9C
8494948C9C948C8C7B848463000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000424A6B5A738C637B94
7B949C7B8C8C8C948C7B846B736B52000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000846B007B6300846B00000000000000
00000000000000000000000000000000000000000000000000000039394A4A5A7B5A7B8C
6B84947B8C947B8C84848C7B7B7B5A000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000007B6300846B007B6300846B007B6300846B007B6300847B007B7300
000000000000000000000000000000000000000000000000000000000000425273526B84
638494738C8C7B8C8C7B8C7B737B63000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000007B6B00
7B6300846B007B63007B6B007B6B00846B007B6300846B007B6B00846B007B7300847300
8473000000000000000000000000000000000000000000000000003939423942634A6384
5A738C6B8494738C8C738C7B6B6B52000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000007B6300846B007B6300846B007B6300
846B007B6300846B007B6300846B007B6300846B007B6300846B007B6300847B007B7300
847B007B730000000000000000000000000000000000000042394A393952424A6B425A7B
5A738C5A7B8C738C8C6B8473000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007B63007B6B007B63007B6B007B63007B6B00
7B63007B6B007B6300846B007B6300846B007B6B00846B007B6300846B00847300847300
7B73008473007B73000000000000004239394239424239524242523939634252734A6384
52738C637B8C6B84846B8473000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007B6B007B6300846B007B6300846B007B6300
846B087B6300846B007B6300846B007B6300846B007B6300846B087B6300847B007B7300
847B007B73008473007B73005A42524A425A4A425A42425A4A426B42426B425A7B42637B
5A7B8C5A7B846384845A7363000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000007B63007B6B007B63007B6B007B63007B6B00
7B6B00846B007B63007B6B007B6B00846B007B6300846B00846B00846B00847300847300
8473008473007B73007B73007B730052526B4A4A6B4A4A734A52734A5A7B425A84526B84
52738C5A7B845A7B73000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B007B6300846B007B6300846B007B6300846B007B6300846B007B6300847B007B7300
847B007B7300847B007B73008473007B73005A5A84525A7B5263844A6384526B8C527384
5A7B84527373000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B63007B6B007B63007B6B007B63007B6B00
7B63007B6B007B6B00846B007B6300846B007B6B00846B007B6B00846B00847300847B00
7B73008473007B73008473007B73007B73005A638C5A6B8C526B8C526B8C52738C527384
527373000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B63007B6B007B6300846B007B6300846B007B6300
846B007B6300846B007B6300846B007B6300846B007B6300846B007B6300847B007B7300
847B007B73008473007B7300847B007B73005A738C526B845A738C52738452737B000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B63007B6B007B6B007B6B007B63007B6B00
7B6B00846B007B63007B6B007B6B00846B007B6B00846B00846B00846B00847300847300
847300847B007B73007B73007B7300847300527384527384426B73000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B007B6300846B007B6300846B007B6300846B007B6300846B007B6300847B007B7300
847B007B7300847B007B73008473007B7300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B63007B6B007B63007B6B007B6B00846B00
7B63007B6B007B6B00846B007B6B00846B007B6B00846B007B6B00846B00847300847B00
7B73008473007B73008473007B73007B7300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B087B6300846B007B6300846B007B6300846B007B6300846B087B6B00847B007B7300
847B007B7300847B007B7300847B087B7300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B63007B6B007B6B007B6B007B63007B6B00
7B6B00846B007B6B00846B007B6B00846B007B6B00846B00846B00846B00847300847300
847300847B007B73008473007B7300847300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B007B6300846B007B6300846B007B6300846B007B6300846B007B6300847B007B7300
847B007B7300847B007B7300847B007B7300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B6B007B6B007B63007B6B007B6B00846B00
7B6B00846B007B6B00846B007B6B00846B007B6B00846B007B6B00846B00847300847B00
8473008473007B73008473007B7300847300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B007B6300846B007B6300846B007B6B00846B007B6300846B007B6B00847B007B7300
847B007B7300847B007B7300847B007B7300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B63007B6B007B6B00846B007B63007B6B00
7B6B00846B007B6B00846B00846B00846B007B6B00846B00846B00846B00847300847B00
847300847B007B73008473007B7300847300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B007B6300846B007B6300846B007B6300846B007B6B00846B007B6B00847B007B7300
847B007B7300847B007B7300847B007B7300000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B6B007B6B007B6B007B6B007B6B00846B00
7B6B00846B007B6B00846B007B6B00846B00846B00846B007B6B00846B00847300847B00
847300847B00847300847B007B73008473007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B087B6B00846B007B6300846B007B6B00846B007B6300846B087B6B00847B007B7300
847B007B7300847B007B7300847B087B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000846B007B63007B6B007B6B00846B007B6B007B6B00
7B6B00846B007B6B00846B00846B00846B007B6B00846B00846B00846B00847300847B00
847300847B007B7300847300847300847B007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B007B6300846B007B6B00846B007B6300846B007B6B00846B007B6B00847B007B7300
847B007B7300847B007B7300847B007B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B6B00846B007B6B007B6B007B6B00846B00
7B6B00846B007B6B00846B007B6B00846B00846B00846B00846B00846B00847B00847B00
847300847B00847300847B007B73008473007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6300
846B007B6B00846B007B6300846B007B6B00846B007B6B00846B007B6B00847B007B7300
847B007B7300847B007B7300847B007B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000846B007B63007B6B007B6B00846B007B6B007B6B00
7B6B00846B007B6B00846B00846B00846B007B6B00846B00846B00846B00847300847B00
847B00847B007B7300847B00847300847B007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6B00
846B007B6300846B007B6B00846B007B6B00846B007B6B00846B007B6B00847B007B7300
847B007B7300847B007B7300847B007B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B6B00846B007B6B00846B007B6B00846B00
7B6B00846B007B6B00846B007B6B00846B00846B00846B00846B00846B00847B00847B00
847B00847B00847300847B007B7300847B007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6B00846B007B6300
846B087B6B00846B007B6B00846B007B6B00846B007B6B008473087B6B00847B007B7300
847B007B7300847B007B7300847B087B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000846B007B6B007B6B007B6B00846B007B6B007B6B00
846B00846B007B6B00846B00846B00846B007B6B00846B00846B00846B00847B00847B00
847B00847B007B7300847B00847B00847B007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6300846B007B6300846B007B6B00
846B007B6B00846B007B6B00846B007B6B00846B007B6B00846B007B6B00847B00847300
847B007B7300847B007B7300847B007B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B6B00846B007B6B00846B007B6B00846B00
7B6B00846B00846B00846B007B6B00846B00846B00846B00846B00846B00847B00847B00
847B00847B00847B00847B007B7300847B007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B00846B007B6300846B007B6B00846B007B6300
846B007B6B00846B007B6B008473007B6B00846B007B6B00847300846B00847B007B7300
847B00847300847B007B7300847B007B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000846B007B6B007B6B007B6B00846B007B6B00846B00
846B00846B007B6B00846B00846B00846B007B6B00846B00846B00846B00847B00847B00
847B00847B00847300847B00847B00847B007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6300846B007B6B00846B007B6B00846B007B6B00
846B007B6B00846B007B6B00846B007B6B008473007B6B00846B007B6B00847B00847300
847B007B7300847B007B7300847B007B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B007B6B00846B007B6B00846B007B6B00846B00
7B6B00846B00846B00846B00846B005273005273005273004A7300527300527300847B00
847B00847B00847B00847B007B7300847B007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000007B6B00846B007B6300846B007B6B00846B007B6300
527B084A73005273004A7300527B004A73005273004A7300527B084A7300527B004A7300
527B00847300847B007B7300847B087B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000846B007B6B00846B004A73005273004A7300527300
5273005273004A73005273004A73005273004A73005273005273005273004A7300527300
527300527300847B00847B00847B00847B007B7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000004A73005273004A73005273004A73005273004A7300
5273004A73005273004A73005273004A73005273004A73005273004A7300527B004A7300
5273004A7300527B004A7300847B007B7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000005273004A73005273004A7300527300
4A73005273004A73005273004A73005273004A73005273004A73005273004A7300527300
4A73005273004A73005273004A7300847B007B7B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000005273004A73005273004A6B00
5273004A73005273004A6B005273004A73005273004A73005273004A73005273004A7300
527B004A73005273004A7300527B004A7300847B00000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000004A73004A7300
4A73005273004A73004A73004A73005273004A73004A73004A73005273004A73004A7300
4A73005273004A73004A73004A7300000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000000000000000000000000000000000000000000000000000000000000000004A6B00
5273004A6B005273004A6B005273004A6B005273004A73005273004A6B005273004A7300
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
0000004A73004A73005273004A73004A73004A7300000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000000000000000000000
pop
P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[58]:=",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj2LW
9`00?g=cLda<C240<410Y/30`9T0IPX:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm?ooc020P800082006IVc3<cIP2009VIod:J
Yb8R8_mV<`00P00g??mVc5J0Zk75fBa1EkV>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7De3eDE:b/[:BT
Y>c/k61PH>7QhBDU9D1PP<gIiR]0EADO:Rl_;i6ATH]ZJSU>CQ`V9]GQhKW>cUEeMJ_0eLC4a000O`00
_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0
?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo
?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no
?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo
?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0_olo0?lo?oloOolo
_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NWYg@7i/;2`W9bLP00
o`1o001oO`1oo`3o003oO`3oogl007l0Ogl0ogmo07moOgmoogoo07ooOgooool00?l0Ool0oomo0?mo
Oomooooo0?ooOoooo`0002XZ:UEEEGmoOjZZZ]GEeOooool78@L00?l78@L00?l78@L00?l78@L00?l7
8@L00?l78@L00?l78@L00?l78@L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7
00061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001PNR009kKP081`001PNL000=
U7^DNiAkU7^DNiAkU01Y00P700061iX0015kU7^DNiAkU7^DNiAkU7^DN`1W00P700061iP00W/00iAk
U00@U6D020L000H7U`02N`08U7^DNiAkU7/;U003Ni@006<020L000H7UP02NaNDH`081`001PNC00Ak
00JDNiAkU7/AU003Ni@0064020L000H7TP02N`04U7^DN`ZD00^]U:fD[IB]U:fD[@04U003N`0005h0
20L000H7TP03N`04U7^DNaVD00=k0000G@081`001PNA0005NiAkU7/02Y@03JfD[IB]U:fD[IB]U:d0
1YAO00P700061i400W/01YAkU7^DN`ZD00N]U:fD[IB]00RD00=kU000G@081`001PN@009k00BDNiAk
29@04JfD[IB]U:fD[IB]U:fD[IB]00JDG@081`001PN@00=k00JDNiAkU7/8U00;[IB]U:fD[IB]U:d0
29@00g/0001K00P700061hl00g/00iAkU009U004[IB]U0>]00ZD[IB]U:fD[IB]1Y@00g/0001J00P7
00061hl00W/01YAkU7^DN`ZD00^]U:fD[IB]U:fD[@08U003Ni@005/020L000H7SP02N`04U7^DN`RD
00J]U:fD[I@5[@0:U:fD[IB]U:fD[@FDG0081`001PN?009k00JDNiAkU7/8U004[IB]U0>]00RD[IB]
U:fD[@RDG0081`001PN>009k00BDNiAk1Y@02:fD[IB]U:fD1Jd02YB]U:fD[IB]U:d6U5/020L000H7
S`02N`04U7^DN`bD00f]U:fD[IB]U:fD[IB]00RD00=k0000F@081`001PN>009k00>DNi@01i@05jfD
[IB]U:jD/iBcU:jD[IB]U:fD[IB]00JDF`081`001PN?009k00BDNiAk39@03JfD[IB]U:fD[IB]U:d0
29@00g/0001I00P700061hh00W/00iAkU009U00G[IB^U:jD/iBcU;>D[IB]U:fD[IB]U:d019AK00P7
00061hl017/00iAkU00;U009[IB]U:fD[IB]00ZD00=kU7/0F`081`001PN>009k00>DNi@02I@05JfD
[YB^U;>D/iBcU;:D[IB]U:fD[@06U5/020L000H7S`04N`09U7^DNh5kVW^100ND00>]U9@03Y@00g^D
N`1K00P700061hh00W/00iAkU003P@07Vh6PVZ2JX002VP0@PJjD[YBcU;:D[IB]U:fD[@RDF`081`00
1PN?009k01:1O861Vh6PPJ2JX9ZPVZ2JWh0AU003Ni@005/020L000H7SP02N`:100BKPZ221:005J6P
XJ2jX;VP^IZcU:fD[IB]U:fD[@06U003N`0005X020L000H7S`02P@:200ZPPZ22X8:PPZ271Z001;VO
X9X;U003NiAk05d020L000H7SP02P@0:Qh:PPZ22Qh:7PP:70Z001j6PXJ2o^Kl00[T01YZ]U:fD[@RD
G@081`001PN?000<PX67PXN1U7^DNiAk0i@00h67PP02X007XJ2nX;VJW`08U003NiAk05d020L000H7
SP002X62Qh:PPIAkU7/9U008Qj2P_j2o^Kh2W`RD00=k0000G0081`001PN?000>PX67PJ1kU7^DNiAk
U7/6U007X8NQX;jP^@02W`:D00EkU7^DN`1O00P700061hl000B1QhNP0g/00iAkU00:U009O:6PYZ2o
X;jO00FD0W]O00P700061hl000F2PJ27X003N`0:U7^DNiAkU7^DN`BD00FPQj2PY@02X004VYAkU09k
H0081`001PN@00:70Z04N`03U7^D00^D0Z001JJP_j2i00:D0W]Q00P700061i4000B7PZ0017/03YAk
U7^DNiAkU7^DNiAk0XL4X003VW]k068020L000H7T002P@05Vh6KPI/019X01G^DNiAk00ND00R0X:2V
X;nONf<020L000H7S`02P@0HVh6PPZ22X9^PVZ2JX86JNiAkU7^DNiAk0XL3X6@020L000H7S@001H62
PZ2200BP00^QX:6P^Z2jX;VP^@02VP04NiAkU09k00F2X:2VX01T00P700061h`000dIPX67PXN2X8:P
PZ2200JP00>iX;T00YX5N`:20j1T00P700061h`000R2PHN2X8:PPP:700nPQj6PXJ2QX;nP_j2iX9l0
0W/20005PJ2PXJ00I0081`001PN<00066H:1Qh671`0018N2X882X005XJ2oX;T00Yl300:200>PQj00
I0081`001PN<00:20XL00j00000900:700FQX:JP_`02_P:P00L0PH6PQj6P06@020L000H7S0001QV7
PJ21W``000J1X8:QX;l2^@:O00J1PX:PQj1T00P700061h`00X83X0:O3@02Q`08XJ2o^KjP^H83X003
P@00068020L000H7S@000h:1X002X003Wj2J00d000>1X:000Z001kVOWh:PQh40I0081`001PN>00BP
00JiX;VOX84;0008Qj6P_j2nX:42X6D020L000H7S`02Q`03X86P00>P00jiVZ2JX86JPIZ1Vh6PPPFP
00>OX000I@081`001PN=000;PIZ1Vh6KVZ2JX9X00Z000kZP^P07X008XH:QX;nP_YmV00P700061h/0
00f1O8:1Vh6PPJ2JX9ZP00:J00>1X:001j000h:1PP05X6H020L000H7RP02P@:200BPVj2K1:000kVP
^@02X00:VYn1XJ2QX:67X0>700BQX;nPIP081`001PN90004PGb2P@:200JPPZ22X9/7X006VZ2JX8NP
0X420003JXN700>PIP081`001PN9000FPH:2Qh:PPZ67XJ2QX:6P^Z2jX;VP^@:O1@02PP03XJ2V00:P
I@081`001PN9000>PX62PXN2Qh:7PZ22X882X00:XJ2QX;VP^InPP@<000<IQh800j000h40001S00P7
00061hP000N1PXN2Qh:100H000n1QhNQX:6P_j2oX;jP^H000P02PP04XJ2VX6H020L000H7R0001QV2
PHN1Q`/000b7PZ2PXJ2oX;jPW`02P@05Qh:PQj00IP081`001PN90003PXN700>P00>100002@0028NQ
X:JP_kjo0Z00186PPZ42X003P@0006@020L000H7R@001h:1QhNPWj000YX:0003PJ2700:P00>QX;h0
0Yl01H:PQj2606L020L000H7RP02PP:P012UX;jP^IZOP85kP@2108400X801J6P_j2o00:P00B7XHN1
I`081`001PN<0003PJ2700VP00bJX9ZPPJ21X86PQj43X003Wh4006P020L000H7S@003X6PX:JP_jFo
X;nP^Z2i1J002J6PX:2oX;jPW`1Y00P700061h/000N1NiYlVWb100>P00FQX;ZPX@04X005Vj22X880
1J000il0001W00P700061hP00X402IZ1ViZPVZ2JX002VP0BPJJPYZ2VX:6PXJ2QX:62XJ2V0Z1Y00P7
00061hL000n1O9^1Vh6PPJ2KX9^PVZ000YX02X6PQj27X8:PPXL2P@03Qj2P00:PJ@081`001PN5000:
PH:1Vh:PPZ2KX@FP00FiX;VP^@02VP03PHL000D000LIX:2VX:F106P020L000H7Q@002X5lPX67PZ22
X88;X003VYl000H000@IPZ270Z000il0001V00P700061hD001V1PX:7PZ22XH:QX:6PXJ2QX;ZP^Z2i
X;VJ00H00XL01:6PX9mX00P700061hD00122JXN2Qh:7PHN1Qh:7PZ270Z001J6P^Z2i00:P0YX30004
6H:PQ`:P00>O0000IP081`001PN400:100N7PXN7X9n100D001216H61XHNQX;nP_j2nX;V00X42Q`03
XJ2P06T020L000H7Q@02P@08Qh:PQZ2OX8490008PZ2PXJ2oYKh2X006VX:2X8NPJP081`001PN600:2
1:001kVP^IZPPIX01P003h:QX:JP_kjoX;VKXHNPQP1Z00P700061hL00X800j27X003X005^InPVZ00
0YX01H6JPIZ100:70Z000jJP_P02X003Qj2106/020L000H7R0003h:7X:2VX;nP_j2nX;VP^@04X008
Vh:7XJ2oX;l2X003Q`0006X020L000H7RP000h:PQ`02X007_Z2oX;nP^@02X005Vj2KX9/00X85X003
VP0006/020L000H7R`006h67Qj6P_j2oX;nP_j2jX:6PXJ22Qj6PYZ2iVP1/00P700061hL000J1Nh5l
PG`2P@09X8NPX:6PXJ2Q00BP00F7X8NPPP03X003Wj0006/020L000H7Q002P@07VX6KPI^JV`02VP0:
PIZ2X8NQQj67X@:P00^7X8N708:7XJ2nW`1/00P700061h<00X403H:1Vh6KPJ2JX9ZPVZ000YX=00:7
0j1/00P700061h<000N1PX:PPZ2K00JP00FiX;VJX002VPX000F2X:2VX01/00P700061h8000<IPX40
0X802:22X8:PPZ2K1j0019nPP84700:20j000h40001Y00P700061h4001R1PX:7PXN2Qh:7PZ67XJ2Q
X:6P_j2oX;T2W`D00X42Q`03YZ2O06/020L000H7PP001fZ2JXN1Wh40100018N2QhL2X00:XJ2QXKZP
^IZPP@8000N16H:2X8NP06`020L000H7PP02PP:700>PWj000Yl5000@PXNQX:6P_j2oX;jP^IZ1Q`:2
00BPQj2OK0081`001PN30009Qh6PQZ2OX9ZP00:J1@004XN2X:2QX;nP_Z2iVX:1Qh:PQVd020L000H7
P`000h6PQ`02X009_Z2nX;VP^IZO00:J01:1VX6QX:JP_jFoX;VJX8NQQj1]00P700061h<000AZPJ27
1:000kVP^@02X007VZ2JX9ZPVP02PP05XJ2oX;h00Z0019ZPPZ1^00P700061hD00122X:2VX;nP_j2o
X;nP^Z2i0j002j67XJ2oYKnP^InP06l020L000H7QP001AV7PZ2700:P00NVX;nP_j2Q00BP00RKPX:Q
X:JP_P:PL0081`001PN9000HPZ27XJ2VX;nPYZ2QX:6PXJ22Qj6P_j2o0Z1_00P700061h<000Ek07/0
N`05000CPAV1PZ22X8NPQj22PAV1PZ27X@02X003Wh4006d020L000H7P002P@07VX6KVY^JV`04VQ00
00N2XJ2VX:F106h020L000H7O`003h5lPX6KPI^1X9ZPVZ2JX002VPd000B1PZ270Z000h40001/00P7
00061gh00X401X:1X8:PV`RP00BiVZ2J3002Q`04XJ2UPFh020L000H7OP000aV2P@02PP08X8:PPZ2K
X9/5X005VZ2JVW`02@000fZ7Q`02X003P@0006`020L000H7O@006862PXN2Qh:PQj6PXJ2QX:6P^Z2j
X;VP^@:J1P001862PZ42X003P@0006`020L000H7O@00484IPVZ7PXN2Qh:7PZ22X8L2X005XJ2PX;T0
0Z02VPD00X42Q`03X8N106h020L000H7OP02PP05Qh:7QZ000Yl00h67PP02Q`0;XJ2QX;nQ_j2oX;T0
0Yl20006PH:2PXNQ0Z000h40001/00P700061gh000XIPX67PJ2OX9nP0YX01h40PHN2XJ000Z402[nP
^Z2iVYmlPX42PP:700>PQ`00KP081`001PMo0003PHN700:P00FUX;VP^@02X00>Wj2JX8NQX:JP_j2o
X;h2X0>200>PQj000Z1_00P700061h0000=ZQh802:003kVPX9ZPJXN2XJ2VX;nP^@02W`06PXN2X8NP
L0081`001PN2000KPHNQX:JP_j2oX;nP_j2jX;Z7X8NVX;nU_j2i00:P00=ZP@00L0081`001PN5000;
Qh:PQj6PYZ2VX:401Z001X:PX:JP_P:P00>J0000L@081`001PN800:701JPQj6PXJ2QX:6PXJ2QPZ6P
_j2oX;jJLP081`001PN<000=JXN2Qh:PPXN2Qh67PP02X005XJ2UWj00LP081`001PNH0009PHNQX:JP
_Z2O074020L000H7O0000gb1O002P@05VWbJNiX04`02PPFP00>10000K`081`001PMk00>100VKPI^1
ViZPVY/00YXA00:700>QX:H00Z1a00P700061gX001=lPH62PI^1Vh6PVZ2JX9ZPPIYk00d000=ZQhL0
0j000il0001_00P700061gX000Z2PH:2X8:PVj2K1:001kVP^IZiVX402`001H62XHNV00:PL@081`00
1PMi008I00>2PH800h803J22X8:PVj2KX9^PVZ000YX:00076H:2X8NPQP1a00P700061gX00X8068N2
Qh:7PZ67XJ2QX:6P^Z2jX;VP^IZON`@00X401HN2QhNQ00:PL@081`001PMj000JJX9ZQh6PWj2OX8:7
PZ27XJ2QX:6P^Z2iWj02VP0:O85lPX67PZ27X78020L000H7N`000h:7Q`02X008YJ2nX;VP^H82Q`0>
XJ2QXKnQ_j2oX;VOX842PP:700>PQj00L`081`001PMl0003PHN700RP00BiX:1Z0XL04Z6PXJ6oX;ZP
^InPPHN2Qh:PPG<020L000H7O@008H67Qj6PYZ2oX;nP_j2oX;Z7XJ2VX;nP_j2nX9n7X8NPQP1d00P7
00061gl00X806j27X:2VX:JU_j2VX:62X8NQX:JU_j2nWiZ2Q`1f00P700061h800XL05j6PYZ2VX:JP
YZ2QX8R7YZ2oYKnU_YnJ07L020L000H7Q@0038N2X8NPQj27XHNPQ`:200FQX:JP_`02X003VP0007D0
20L000H7R`004H67Qj27X8:7PZ6PYZ2oX;VO07H020L000H7U0001X6PQj6PY@:PMP081`001PNE0006
PZ6PYZ2n0Z1e00P700061iL00XL01J6PX9n107@020L000H7RP000ijIWP0:0007PJ2PYZ2UW`1d00P7
00061hH000RIWYVNVIjIWPX000>2X8L00j1d00P700061h<000>IWYT00ih00iVNWP03WPP000N1QhNQ
X:FO07@020L000H7O`004IjIWYVNVIjIWYVNVIjIWYVN00H000MlPH6PQj2607D020L000H7OP05WP05
VIjIWYT00ih00iVNWP03WP800h401H:1QhNQ00:PM@081`001PMn000IVIjIWYVNVIjIWYVNVIjIWYVN
VIj1PX62P@02PP04X8NPPGD020L000H7OP07WP05VIjNWYT02Ih3PP06Qh:7Qj27MP081`001PMm000B
WYVNVIjIWYVNVIjIWYVNVIjI0ih02IVNPXN2Qh:PP@1g00P700061gd01Yh01IVNWYjI00bN00JQQj67
XHMh00P700061gd001^NVIjIWYVNVIjIWYVNVIjIWYVNVIjIWX:7PX40NP081`001PMm0007VIjIWYVN
V@0@WWh020L000H7O@005ijIWYVNVIjIWYVNVIjIWYVNVIjIWYVN07h020L000H7O@001iVNVIjIWYT0
49in00P700061gd001NNVIjIWYVNVIjIWYVNVIjIWYVNVIjIWP1n00P700061gd000NIWYVNVIjI012N
OP081`001PMm000GWYVNVIjIWYVNVIjIWYVNVIjIWYVNVIh0OP081`001PMm0007VIjIWYVNV@0@WWh0
20L000H7O@005ijIWYVNVIjIWYVNVIjIWYVNVIjIWYVN07h020L000H7O@001iVNVIjIWYT049in00P7
00061gd001RNVIjIWYVNVIjIWYVNVIjIWYVNVIjIWYUm00P700061gd000FIWYVNV@0CWWd020L000H7
O@004YjIWYVNVIjIWYVNVIjIWYVNV@JNO@081`001PMm01RNO@081`001PMm000BWYVNVIjIWYVNVIjI
WYVNVIjI1Yim00P700061gd069im00P700061gd001:NVIjIWYVNVIjIWYVNVIjIWYT6WWd020L000H7
O@0HWWd020L000H7O@004YjIWYVNVIjIWYVNVIjIWYVNV@JNO@081`001PMm01RNO@081`001PMm000B
WYVNVIjIWYVNVIjIWYVNVIjI1Yim00P700061gd069im00P700061gd0012NVIjIWYVNVIjIWYVNVIjI
0ih01IVNWYjI07d020L000H7O@000iVNWP0EWWd020L000H7O@004IjIWYVNVIjIWYVNP8F0QH2500BN
00>IWYT0O@081`001PMm00:N00BIWYjN38D6WWd020L000H7O@004ijIWX25P8F0QH25P8F0QH25P8D0
19h00iT0001k00P700061gd000J5WXFNQIh=Q@03WXFN00:NO@081`001PMo000DQH25P8F0QH25P8F0
QH25P8F0QH02WWd020L000H7P0000ij5WP0=Q@05WXFNQIh0O@081`001PN2000AP8F0QH25P8F0QH25
P8F0QH00O`081`001PN300^500>00000P@081`001PN50005P0200800R0081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P700061ol04`081`001POo01<020L000H7
o`0C00P700061ol04`081`001POo01<020L000H7o`0C00P7003o1b47003o1b47003o1b47003o1b47
003o1b47003o1b470000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-7.50157, -7.50157, 1.09757, 
  1.09757}},
  CellTags->"Animation"]
}, Open  ]],

Cell[TextData[{
  "Now a ",
  Cell[BoxData[
      \(TraditionalForm\`\[Tau]\)]],
  " depend frame of the motion is generated"
}], "Text",
  CellTags->"Animation"],

Cell[BoxData[
    \(\(frame\[Tau] = 
      SpringGraphics[\[Rho][\[Tau]], \[Theta][\[Tau]], \[Phi][\[Tau]], 0.15, 
          8, 1.5]\  /. \ sol; \)\)], "Input",
  CellLabel->"In[24]:=",
  CellTags->"Animation"],

Cell[TextData[{
  "The following command will create the graphics inside ",
  StyleBox["Mathematica, ",
    FontSlant->"Italic"],
  StyleBox[
  "and send the data to the OpenGL viewer. The animation speed depends on the \
speed of the computer, the speed of the OpenGL implementation and of the ",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox["MathLink",
    FontSlant->"Italic",
    FontVariations->{"CompatibilityType"->0}],
  StyleBox[" speed.",
    FontVariations->{"CompatibilityType"->0}]
}], "Text",
  CellTags->"Animation"],

Cell[BoxData[
    \(\(Table[gr = frame\[Tau]; 
      TransferFrame[First[gr]], {\[Tau], 0, 24  \[Pi], \[Pi]/8}]; \)\)], 
  "Input",
  CellLabel->"In[26]:=",
  CellTags->"Animation"],

Cell["\<\
In this example only the two coordinates of the cuboid, the \
coordinate of the point mass and 144 coordinates for the spring are send to \
the OpenGL viewer and it should be fast on the most machines.\
\>", "Text",
  CellTags->"Animation"]
}, Closed]],

Cell[CellGroupData[{

Cell["Polygon Simplification", "Section",
  CellTags->"PolygonReduction"],

Cell["\<\
The OpenGL viewer can reduce the polygons in a scene. This results \
large memory savings in the exported files. It will also increase the render \
speed. Here is an example\
\>", "Text",
  CellTags->"PolygonReduction"],

Cell[BoxData[
    \(Needs["\<Graphics`Shapes`\>"]\)], "Input",
  CellLabel->"In[59]:=",
  CellTags->"PolygonReduction"],

Cell[BoxData[
    \(PolygonNumber[gr_] := Length[Cases[gr, _Polygon, Infinity]]\)], "Input",\

  CellLabel->"In[60]:=",
  CellTags->"PolygonReduction"],

Cell[BoxData[
    \(\(gr = Graphics3D[Sphere[1, 36, 18]]; \)\)], "Input",
  CellLabel->"In[61]:=",
  CellTags->"PolygonReduction"],

Cell[TextData[{
  "The polygon simplification is done by the ",
  StyleBox["MVPolygonReduce[] ",
    FontFamily->"Courier"],
  "command. It returns the number of removed polygones. The MVGetGraphics3D[] \
function is used to return the triangles to ",
  StyleBox["Mathematica. ",
    FontSlant->"Italic"],
  "The tolerance is set to  {0,0.025,0.05,0.075}. The tolerance of zero \
leaves the polygones unchanged."
}], "Text",
  CellTags->"PolygonReduction"],

Cell[BoxData[
    \(\(grs = 
      \(\((MVPolygonReduce[gr, #]; MVGetGraphics3D[])\)\ &\)\ /@
        \ {0, 0.025, 0.05, 0.075}; \)\)], "Input",
  CellLabel->"In[75]:=",
  CellTags->"PolygonReduction"],

Cell[TextData[{
  "The reduction process works as follows. For every vertex the distance of \
the vertex to the averaged plane of the adjacent polygones is calculated. If \
the distance is smaller than the tolerance parameter of ",
  StyleBox["MVPolygonReduce[]",
    FontFamily->"Courier"],
  " the vertex is removed. The outline of the adjacent polygones is \
triangulated and replace the old polygones. This is done all verticies. ",
  StyleBox["MVPolygonReduce[]",
    FontFamily->"Courier"],
  " repeat this search  ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  "-times. The ",
  StyleBox["ReductionLoop",
    FontFamily->"Courier"],
  "-option gives the maximum number of this reduction loops. When the polygon \
number does not change form one iteration to the next the iteration is \
terminated. Here are the results for the sphere."
}], "Text",
  CellTags->"PolygonReduction"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
      GraphicsArray[
        Partition[grs, 2] /. \ 
          Graphics3D[dta_, opt___] :> 
            Graphics3D[dta, 
              Join[{Boxed -> False, RenderAll -> False, Shading -> False, 
                  Lighting -> False, 
                  PlotLabel -> 
                    "\<Polygons :\>"<>ToString[PolygonNumber[dta]]}, {opt}]], 
        GraphicsSpacing \[Rule]  .0]]; \)\)], "Input",
  CellLabel->"In[79]:=",
  CellTags->"PolygonReduction"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.07901 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0256908 0.47619 [
[ 0 0 0 0 ]
[ 1 1.07901 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.07901 L
0 1.07901 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0256908 0.5 0.539506 MathSubStart
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0568365 1.16543 -1.11022e-16 1.16543 [
[.5 1.0982 -41 0 ]
[.5 1.0982 41 9.8125 ]
[ 0 0 0 0 ]
[ 1 1.0857 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
gsave
.5 1.0982 -102 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Polygons) show
117.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(:) show
123.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(392) show
141.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 1.0857 L
0 1.0857 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.19811 .58934 m .20825 .50751 L .19802 .5762 L closepath p 1 g
F P
0 g
s
.19811 .58934 m .22934 .52344 L .20825 .50751 L closepath p 1 g
F P
0 g
s
.21889 .60648 m .22934 .52344 L .19811 .58934 L closepath p 1 g
F P
0 g
s
.20858 .66917 m .21889 .60648 L .19811 .58934 L closepath p 1 g
F P
0 g
s
.20825 .50751 m .26063 .44302 L .21799 .4629 L closepath p 1 g
F P
0 g
s
.22934 .52344 m .26063 .44302 L .20825 .50751 L closepath p 1 g
F P
0 g
s
.20858 .66917 m .28814 .68329 L .21889 .60648 L closepath p 1 g
F P
0 g
s
.23768 .74172 m .28814 .68329 L .20858 .66917 L closepath p 1 g
F P
0 g
s
.21799 .4629 m .30914 .3716 L .25635 .39155 L closepath p 1 g
F P
0 g
s
.26063 .44302 m .30914 .3716 L .21799 .4629 L closepath p 1 g
F P
0 g
s
.21889 .60648 m .26982 .54777 L .22934 .52344 L closepath p 1 g
F P
0 g
s
.21889 .60648 m .34752 .64218 L .26982 .54777 L closepath p 1 g
F P
0 g
s
.28814 .68329 m .34752 .64218 L .21889 .60648 L closepath p 1 g
F P
0 g
s
.22934 .52344 m .30204 .46651 L .26063 .44302 L closepath p 1 g
F P
0 g
s
.26982 .54777 m .30204 .46651 L .22934 .52344 L closepath p 1 g
F P
0 g
s
.3101 .75294 m .28814 .68329 L .23768 .74172 L closepath p 1 g
F P
0 g
s
.25417 .76757 m .3101 .75294 L .23768 .74172 L closepath p 1 g
F P
0 g
s
.28152 .80284 m .3101 .75294 L .25417 .76757 L closepath p 1 g
F P
0 g
s
.25635 .39155 m .42004 .29894 L .30837 .33344 L closepath p 1 g
F P
0 g
s
.30914 .3716 m .42004 .29894 L .25635 .39155 L closepath p 1 g
F P
0 g
s
.30204 .46651 m .47057 .36628 L .26063 .44302 L closepath p 1 g
F P
0 g
s
.26063 .44302 m .47057 .36628 L .30914 .3716 L closepath p 1 g
F P
0 g
s
.26982 .54777 m .35806 .50223 L .30204 .46651 L closepath p 1 g
F P
0 g
s
.32511 .58217 m .35806 .50223 L .26982 .54777 L closepath p 1 g
F P
0 g
s
.26982 .54777 m .34752 .64218 L .32511 .58217 L closepath p 1 g
F P
0 g
s
.28152 .80284 m .34863 .8245 L .3101 .75294 L closepath p 1 g
F P
0 g
s
.31745 .83222 m .34863 .8245 L .28152 .80284 L closepath p 1 g
F P
0 g
s
.34771 .71582 m .34752 .64218 L .28814 .68329 L closepath p 1 g
F P
0 g
s
.3101 .75294 m .34771 .71582 L .28814 .68329 L closepath p 1 g
F P
0 g
s
.47059 .41866 m .47057 .36628 L .30204 .46651 L closepath p 1 g
F P
0 g
s
.35806 .50223 m .47059 .41866 L .30204 .46651 L closepath p 1 g
F P
0 g
s
.43045 .27574 m .44333 .26317 L .30837 .33344 L closepath p 1 g
F P
0 g
s
.42004 .29894 m .43045 .27574 L .30837 .33344 L closepath p 1 g
F P
0 g
s
.30914 .3716 m .47239 .32285 L .42004 .29894 L closepath p 1 g
F P
0 g
s
.30914 .3716 m .47057 .36628 L .47239 .32285 L closepath p 1 g
F P
0 g
s
.36977 .77893 m .34771 .71582 L .3101 .75294 L closepath p 1 g
F P
0 g
s
.34863 .8245 m .36977 .77893 L .3101 .75294 L closepath p 1 g
F P
0 g
s
.31745 .83222 m .36114 .85443 L .34863 .8245 L closepath p 1 g
F P
0 g
s
.31745 .83222 m .3611 .85751 L .36114 .85443 L closepath p 1 g
F P
0 g
s
.32511 .58217 m .53346 .54009 L .35806 .50223 L closepath p 1 g
F P
0 g
s
.53244 .60628 m .53346 .54009 L .32511 .58217 L closepath p 1 g
F P
0 g
s
.34752 .64218 m .53244 .60628 L .32511 .58217 L closepath p 1 g
F P
0 g
s
.53114 .67134 m .53244 .60628 L .34752 .64218 L closepath p 1 g
F P
0 g
s
.34771 .71582 m .53114 .67134 L .34752 .64218 L closepath p 1 g
F P
0 g
s
.436 .74441 m .53114 .67134 L .34771 .71582 L closepath p 1 g
F P
0 g
s
.36977 .77893 m .436 .74441 L .34771 .71582 L closepath p 1 g
F P
0 g
s
.34863 .8245 m .46675 .8524 L .36977 .77893 L closepath p 1 g
F P
0 g
s
.34863 .8245 m .46388 .85753 L .46675 .8524 L closepath p 1 g
F P
0 g
s
.34863 .8245 m .46289 .86281 L .46388 .85753 L closepath p 1 g
F P
0 g
s
.36114 .85443 m .46289 .86281 L .34863 .8245 L closepath p 1 g
F P
0 g
s
.35806 .50223 m .53418 .47577 L .47059 .41866 L closepath p 1 g
F P
0 g
s
.53346 .54009 m .53418 .47577 L .35806 .50223 L closepath p 1 g
F P
0 g
s
.36114 .85443 m .3611 .85751 L .43231 .88599 L closepath p 1 g
F P
0 g
s
.3611 .85751 m .41495 .88871 L .43694 .88784 L p 1 g
F P
0 g
s
.36114 .85443 m .39747 .87054 L p .46289 .86281 L 1 g
F P
0 g
s
.36977 .77893 m .57943 .84904 L .436 .74441 L closepath p 1 g
F P
0 g
s
.46675 .8524 m .57943 .84904 L .36977 .77893 L closepath p 1 g
F P
0 g
s
.39747 .87054 m .43231 .88599 L .57943 .84904 L p 1 g
F P
0 g
s
.44333 .26317 m .42493 .27116 L .40361 .28385 L p 1 g
F P
0 g
s
.41495 .88871 m .4831 .90151 L .48436 .88597 L closepath p 1 g
F P
0 g
s
.42004 .29894 m .59109 .2918 L .43045 .27574 L closepath p 1 g
F P
0 g
s
.47239 .32285 m .59109 .2918 L .42004 .29894 L closepath p 1 g
F P
0 g
s
.43045 .27574 m .58451 .26924 L .44333 .26317 L closepath p 1 g
F P
0 g
s
.59109 .2918 m .58451 .26924 L .43045 .27574 L closepath p 1 g
F P
0 g
s
.43231 .88599 m .48436 .88597 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.43694 .88784 m .48436 .88597 L p .43231 .88599 L 1 g
F P
0 g
s
.57943 .84904 m .53114 .67134 L .436 .74441 L closepath p 1 g
F P
0 g
s
.44333 .26317 m .58451 .26924 L .5317 .25389 L closepath p 1 g
F P
0 g
s
.44333 .26317 m .5317 .25389 L .49369 .25488 L closepath p 1 g
F P
0 g
s
.46289 .86281 m .57943 .84904 L .46388 .85753 L closepath p 1 g
F P
0 g
s
.46388 .85753 m .57943 .84904 L .46675 .8524 L closepath p 1 g
F P
0 g
s
.47057 .36628 m .59816 .36859 L .47239 .32285 L closepath p 1 g
F P
0 g
s
.47059 .41866 m .65857 .43266 L .47057 .36628 L closepath p 1 g
F P
0 g
s
.65857 .43266 m .59816 .36859 L .47057 .36628 L closepath p 1 g
F P
0 g
s
.53418 .47577 m .65857 .43266 L .47059 .41866 L closepath p 1 g
F P
0 g
s
.47239 .32285 m .65429 .3371 L .59109 .2918 L closepath p 1 g
F P
0 g
s
.47239 .32285 m .59816 .36859 L .65429 .3371 L closepath p 1 g
F P
0 g
s
.48436 .88597 m .4831 .90151 L .49293 .88901 L closepath p 1 g
F P
0 g
s
.4831 .90151 m .51909 .90349 L .49293 .88901 L closepath p 1 g
F P
0 g
s
.48436 .88597 m .49293 .88901 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.49293 .88901 m .50242 .89129 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.49293 .88901 m .51909 .90349 L .50242 .89129 L closepath p 1 g
F P
0 g
s
.50242 .89129 m .51254 .89273 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.51909 .90349 m .55521 .90245 L .50242 .89129 L closepath p 1 g
F P
0 g
s
.50242 .89129 m .55521 .90245 L .51254 .89273 L closepath p 1 g
F P
0 g
s
.51254 .89273 m .523 .89331 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.55521 .90245 m .59061 .89842 L .51254 .89273 L closepath p 1 g
F P
0 g
s
.51254 .89273 m .59061 .89842 L .523 .89331 L closepath p 1 g
F P
0 g
s
.523 .89331 m .53351 .89301 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.53351 .89301 m .523 .89331 L .53919 .89302 L p 1 g
F P
0 g
s
.523 .89331 m .59061 .89842 L .6163 .89313 L p 1 g
F P
0 g
s
.6163 .89313 m .55083 .89303 L .523 .89331 L 1 g
F
.53919 .89302 m .523 .89331 L p .55083 .89303 L F P
0 g
s
.53114 .67134 m .63073 .61799 L .53244 .60628 L closepath p 1 g
F P
0 g
s
.68246 .70482 m .63073 .61799 L .53114 .67134 L closepath p 1 g
F P
0 g
s
.57943 .84904 m .68246 .70482 L .53114 .67134 L closepath p 1 g
F P
0 g
s
.53244 .60628 m .63073 .61799 L .53346 .54009 L closepath p 1 g
F P
0 g
s
.53346 .54009 m .64441 .55385 L .53418 .47577 L closepath p 1 g
F P
0 g
s
.63073 .61799 m .64441 .55385 L .53346 .54009 L closepath p 1 g
F P
0 g
s
.53351 .89301 m .55347 .88981 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.53351 .89301 m .62888 .89315 L .55347 .88981 L closepath p 1 g
F P
0 g
s
.64441 .55385 m .65857 .43266 L .53418 .47577 L closepath p 1 g
F P
0 g
s
.55347 .88981 m .56236 .887 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.55347 .88981 m .62888 .89315 L .56236 .887 L closepath p 1 g
F P
0 g
s
.56236 .887 m .72703 .83781 L .57015 .88347 L closepath p 1 g
F P
0 g
s
.68388 .86933 m .72703 .83781 L .56236 .887 L closepath p 1 g
F P
0 g
s
.56236 .887 m .57015 .88347 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.62888 .89315 m .68388 .86933 L .56236 .887 L closepath p 1 g
F P
0 g
s
.72703 .83781 m .7406 .81937 L .57015 .88347 L closepath p 1 g
F P
0 g
s
.57015 .88347 m .7406 .81937 L .57662 .87933 L closepath p 1 g
F P
0 g
s
.57015 .88347 m .57662 .87933 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.7406 .81937 m .74799 .79974 L .57662 .87933 L closepath p 1 g
F P
0 g
s
.57662 .87933 m .74799 .79974 L .58156 .87469 L closepath p 1 g
F P
0 g
s
.57662 .87933 m .58156 .87469 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.57943 .84904 m .70008 .81377 L .68246 .70482 L closepath p 1 g
F P
0 g
s
.58481 .86969 m .70008 .81377 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.58156 .87469 m .58481 .86969 L .57943 .84904 L closepath p 1 g
F P
0 g
s
.58156 .87469 m .74799 .79974 L .74833 .79036 L p 1 g
F P
0 g
s
.62211 .85159 m .58156 .87469 L p .74833 .79036 L 1 g
F P
0 g
s
.58481 .86969 m .58156 .87469 L .62211 .85159 L p 1 g
F P
0 g
s
.59109 .2918 m .69572 .32076 L .58451 .26924 L closepath p 1 g
F P
0 g
s
.58451 .26924 m .69572 .32076 L .71813 .31662 L closepath p 1 g
F P
0 g
s
.65429 .3371 m .69572 .32076 L .59109 .2918 L closepath p 1 g
F P
0 g
s
.59816 .36859 m .76115 .42362 L .65429 .3371 L closepath p 1 g
F P
0 g
s
.65857 .43266 m .76115 .42362 L .59816 .36859 L closepath p 1 g
F P
0 g
s
.71813 .31662 m .66237 .29023 L .59852 .27421 L p 1 g
F P
0 g
s
.65579 .8817 m .68388 .86933 L .62888 .89315 L closepath p 1 g
F P
0 g
s
.68246 .70482 m .64441 .55385 L .63073 .61799 L closepath p 1 g
F P
0 g
s
.64441 .55385 m .70689 .50809 L .65857 .43266 L closepath p 1 g
F P
0 g
s
.68246 .70482 m .70689 .50809 L .64441 .55385 L closepath p 1 g
F P
0 g
s
.65429 .3371 m .77405 .37092 L .69572 .32076 L closepath p 1 g
F P
0 g
s
.65429 .3371 m .76115 .42362 L .77405 .37092 L closepath p 1 g
F P
0 g
s
.65857 .43266 m .78993 .55628 L .76115 .42362 L closepath p 1 g
F P
0 g
s
.70689 .50809 m .78993 .55628 L .65857 .43266 L closepath p 1 g
F P
0 g
s
.74403 .66797 m .70689 .50809 L .68246 .70482 L closepath p 1 g
F P
0 g
s
.68246 .70482 m .70008 .81377 L .74403 .66797 L closepath p 1 g
F P
0 g
s
.73911 .82957 m .72703 .83781 L .68388 .86933 L closepath p 1 g
F P
0 g
s
.69572 .32076 m .77405 .37092 L .71813 .31662 L closepath p 1 g
F P
0 g
s
.70008 .81377 m .74872 .77947 L .74403 .66797 L closepath p 1 g
F P
0 g
s
.74833 .79036 m .7109 .80614 L .70008 .81377 L 1 g
F
.77132 .61225 m .78993 .55628 L .70689 .50809 L closepath p F P
0 g
s
.74403 .66797 m .77132 .61225 L .70689 .50809 L closepath p 1 g
F P
0 g
s
.74833 .79036 m .74872 .77947 L p .7109 .80614 L 1 g
F P
0 g
s
.72703 .83781 m .7878 .76378 L .7406 .81937 L closepath p 1 g
F P
0 g
s
.72703 .83781 m .73911 .82957 L .75057 .80914 L p 1 g
F P
0 g
s
.7406 .81937 m .7878 .76378 L .74799 .79974 L closepath p 1 g
F P
0 g
s
.74403 .66797 m .74872 .77947 L .77132 .61225 L closepath p 1 g
F P
0 g
s
.74799 .79974 m .78932 .73925 L .74872 .77947 L closepath p 1 g
F P
0 g
s
.7878 .76378 m .78932 .73925 L .74799 .79974 L closepath p 1 g
F P
0 g
s
.74872 .77947 m .78932 .73925 L .77132 .61225 L closepath p 1 g
F P
0 g
s
.76115 .42362 m .79951 .45238 L .77405 .37092 L closepath p 1 g
F P
0 g
s
.79934 .50232 m .79951 .45238 L .76115 .42362 L closepath p 1 g
F P
0 g
s
.78993 .55628 m .79934 .50232 L .76115 .42362 L closepath p 1 g
F P
0 g
s
.77132 .61225 m .78932 .73925 L .78993 .55628 L closepath p 1 g
F P
0 g
s
.84272 .50782 m .81671 .43352 L .77405 .37092 L closepath p 1 g
F P
0 g
s
.77405 .37092 m .84544 .51818 L .84272 .50782 L closepath p 1 g
F P
0 g
s
.79951 .45238 m .84544 .51818 L .77405 .37092 L closepath p 1 g
F P
0 g
s
.7878 .76378 m .79275 .75677 L p .82058 .69474 L 1 g
F P
0 g
s
.7878 .76378 m .82058 .69474 L .78932 .73925 L closepath p 1 g
F P
0 g
s
.78932 .73925 m .82058 .69474 L .78993 .55628 L closepath p 1 g
F P
0 g
s
.78993 .55628 m .82058 .69474 L .79934 .50232 L closepath p 1 g
F P
0 g
s
.79275 .75677 m .82682 .70849 L p .82997 .67381 L 1 g
F P
0 g
s
.79934 .50232 m .83466 .6162 L .79951 .45238 L closepath p 1 g
F P
0 g
s
.82058 .69474 m .83466 .6162 L .79934 .50232 L closepath p 1 g
F P
0 g
s
.83466 .6162 m .84525 .56626 L .79951 .45238 L closepath p 1 g
F P
0 g
s
.84525 .56626 m .84544 .51818 L .79951 .45238 L closepath p 1 g
F P
0 g
s
.82058 .69474 m .8417 .64767 L .83466 .6162 L closepath p 1 g
F P
0 g
s
.82682 .70849 m .84919 .63282 L .82997 .67381 L p 1 g
F P
0 g
s
.82997 .67381 m .84919 .63282 L .8417 .64767 L p 1 g
F P
0 g
s
.8417 .64767 m .84525 .56626 L .83466 .6162 L closepath p 1 g
F P
0 g
s
.8417 .64767 m .84495 .64123 L p .84525 .56626 L 1 g
F P
0 g
s
.84495 .64123 m .84919 .63282 L .84544 .51818 L p 1 g
F P
0 g
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.5 0.0256908 0.97619 0.539506 MathSubStart
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0619112 1.17872 0 1.17872 [
[.5 1.1106 -41 0 ]
[.5 1.1106 41 9.8125 ]
[ 0 0 0 0 ]
[ 1 1.0981 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
gsave
.5 1.1106 -102 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Polygons) show
117.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(:) show
123.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(298) show
141.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 1.0981 L
0 1.0981 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.20218 .61938 m .20276 .56596 L .2026 .51726 L closepath p 1 g
F P
0 g
s
.20233 .66521 m .20276 .56596 L .20218 .61938 L closepath p 1 g
F P
0 g
s
.20233 .66521 m .2133 .61683 L .20276 .56596 L closepath p 1 g
F P
0 g
s
.21118 .62618 m .20233 .66521 L p .21595 .63601 L 1 g
F P
0 g
s
.21597 .63617 m .21595 .63601 L .20233 .66521 L 1 g
F
.20233 .66521 m .21443 .69692 L p .21597 .63617 L F P
0 g
s
.29678 .47643 m .34685 .40248 L .2026 .51726 L closepath p 1 g
F P
0 g
s
.2026 .51726 m .34685 .40248 L .23296 .43909 L closepath p 1 g
F P
0 g
s
.20276 .56596 m .29678 .47643 L .2026 .51726 L closepath p 1 g
F P
0 g
s
.20276 .56596 m .35296 .51241 L .29678 .47643 L closepath p 1 g
F P
0 g
s
.2133 .61683 m .35296 .51241 L .20276 .56596 L closepath p 1 g
F P
0 g
s
.2133 .61683 m .21118 .62618 L p .21592 .63585 L 1 g
F P
0 g
s
.21592 .63585 m .21118 .62618 L .21595 .63601 L 1 g
F
.28242 .61508 m .35296 .51241 L .2133 .61683 L closepath p F P
0 g
s
.22432 .69658 m .28242 .61508 L .2133 .61683 L closepath p 1 g
F P
0 g
s
.22432 .69658 m .21597 .63617 L .21443 .69692 L 1 g
F
.22434 .69672 m .22432 .69658 L .21443 .69692 L F
.21443 .69692 m .23037 .73866 L p .22446 .69751 L F P
0 g
s
.22446 .69751 m .22434 .69672 L .21443 .69692 L 1 g
F
.23296 .43909 m .22139 .46543 L .22097 .46997 L p F P
0 g
s
.22432 .69658 m .4431 .68908 L .28242 .61508 L closepath p 1 g
F P
0 g
s
.3049 .764 m .4431 .68908 L .22432 .69658 L closepath p 1 g
F P
0 g
s
.24878 .77868 m .22432 .69658 L .23037 .73866 L closepath p 1 g
F P
0 g
s
.24878 .77868 m .3049 .764 L .22432 .69658 L closepath p 1 g
F P
0 g
s
.23296 .43909 m .4079 .34193 L .31794 .34501 L closepath p 1 g
F P
0 g
s
.34685 .40248 m .4079 .34193 L .23296 .43909 L closepath p 1 g
F P
0 g
s
.27627 .81414 m .3049 .764 L .24878 .77868 L closepath p 1 g
F P
0 g
s
.27627 .81414 m .34363 .83587 L .3049 .764 L closepath p 1 g
F P
0 g
s
.31238 .84367 m .34363 .83587 L .27627 .81414 L closepath p 1 g
F P
0 g
s
.28242 .61508 m .52886 .5505 L .35296 .51241 L closepath p 1 g
F P
0 g
s
.4431 .68908 m .52886 .5505 L .28242 .61508 L closepath p 1 g
F P
0 g
s
.29678 .47643 m .53005 .42607 L .34685 .40248 L closepath p 1 g
F P
0 g
s
.35296 .51241 m .53005 .42607 L .29678 .47643 L closepath p 1 g
F P
0 g
s
.43118 .75545 m .4431 .68908 L .3049 .764 L closepath p 1 g
F P
0 g
s
.34363 .83587 m .43118 .75545 L .3049 .764 L closepath p 1 g
F P
0 g
s
.31794 .34501 m .33778 .31828 L .30743 .35665 L p 1 g
F P
0 g
s
.31238 .84367 m .35625 .86597 L .34363 .83587 L closepath p 1 g
F P
0 g
s
.31238 .84367 m .37986 .88987 L .35625 .86597 L closepath p 1 g
F P
0 g
s
.4079 .34193 m .58719 .30076 L .31794 .34501 L closepath p 1 g
F P
0 g
s
.31794 .34501 m .58719 .30076 L .33778 .31828 L closepath p 1 g
F P
0 g
s
.33778 .31828 m .58072 .27788 L .36244 .30107 L closepath p 1 g
F P
0 g
s
.58719 .30076 m .58072 .27788 L .33778 .31828 L closepath p 1 g
F P
0 g
s
.40021 .84886 m .43118 .75545 L .34363 .83587 L closepath p 1 g
F P
0 g
s
.34363 .83587 m .45837 .87434 L .40021 .84886 L closepath p 1 g
F P
0 g
s
.35625 .86597 m .45837 .87434 L .34363 .83587 L closepath p 1 g
F P
0 g
s
.34685 .40248 m .59171 .33443 L .4079 .34193 L closepath p 1 g
F P
0 g
s
.34685 .40248 m .594 .3782 L .59171 .33443 L closepath p 1 g
F P
0 g
s
.53005 .42607 m .594 .3782 L .34685 .40248 L closepath p 1 g
F P
0 g
s
.52961 .48595 m .53005 .42607 L .35296 .51241 L closepath p 1 g
F P
0 g
s
.52886 .5505 m .52961 .48595 L .35296 .51241 L closepath p 1 g
F P
0 g
s
.35625 .86597 m .37986 .88987 L .40667 .87994 L closepath p 1 g
F P
0 g
s
.40667 .87994 m .45837 .87434 L .35625 .86597 L closepath p 1 g
F P
0 g
s
.36244 .30107 m .52778 .26225 L .4897 .26301 L closepath p 1 g
F P
0 g
s
.36244 .30107 m .58072 .27788 L .52778 .26225 L closepath p 1 g
F P
0 g
s
.40667 .87994 m .37986 .88987 L .47256 .89389 L closepath p 1 g
F P
0 g
s
.47256 .89389 m .37986 .88987 L .44242 .90492 L closepath p 1 g
F P
0 g
s
.4266 .81983 m .43118 .75545 L .40021 .84886 L closepath p 1 g
F P
0 g
s
.40021 .84886 m .46221 .86387 L .4266 .81983 L closepath p 1 g
F P
0 g
s
.40021 .84886 m .45935 .86903 L .46221 .86387 L closepath p 1 g
F P
0 g
s
.40021 .84886 m .45837 .87434 L .45935 .86903 L closepath p 1 g
F P
0 g
s
.40667 .87994 m .40825 .88028 L p .45837 .87434 L 1 g
F P
0 g
s
.59171 .33443 m .58719 .30076 L .4079 .34193 L closepath p 1 g
F P
0 g
s
.40825 .88028 m .47256 .89389 L .57536 .86049 L p 1 g
F P
0 g
s
.4266 .81983 m .49015 .84783 L .43118 .75545 L closepath p 1 g
F P
0 g
s
.4266 .81983 m .46221 .86387 L .46687 .85903 L closepath p 1 g
F P
0 g
s
.4266 .81983 m .46687 .85903 L .47322 .85464 L closepath p 1 g
F P
0 g
s
.4266 .81983 m .47322 .85464 L .48105 .85087 L closepath p 1 g
F P
0 g
s
.4266 .81983 m .48105 .85087 L .49015 .84783 L closepath p 1 g
F P
0 g
s
.43118 .75545 m .52886 .5505 L .4431 .68908 L closepath p 1 g
F P
0 g
s
.50147 .7987 m .52886 .5505 L .43118 .75545 L closepath p 1 g
F P
0 g
s
.43118 .75545 m .49015 .84783 L .50147 .7987 L closepath p 1 g
F P
0 g
s
.43836 .90394 m .44389 .90844 L .45156 .90711 L p 1 g
F P
0 g
s
.57536 .86049 m .44242 .90492 L .48864 .9007 L closepath p 1 g
F P
0 g
s
.47256 .89389 m .44242 .90492 L .47943 .89255 L p 1 g
F P
0 g
s
.45156 .90711 m .48864 .9007 L p .44242 .90492 L 1 g
F P
0 g
s
.45837 .87434 m .57536 .86049 L .45935 .86903 L closepath p 1 g
F P
0 g
s
.45935 .86903 m .57536 .86049 L .46221 .86387 L closepath p 1 g
F P
0 g
s
.46221 .86387 m .57536 .86049 L .46687 .85903 L closepath p 1 g
F P
0 g
s
.46687 .85903 m .57536 .86049 L .47322 .85464 L closepath p 1 g
F P
0 g
s
.48603 .88951 m .47256 .89389 L p .47943 .89255 L 1 g
F P
0 g
s
.47322 .85464 m .57536 .86049 L .48105 .85087 L closepath p 1 g
F P
0 g
s
.47943 .89255 m .57536 .86049 L .48603 .88951 L p 1 g
F P
0 g
s
.48105 .85087 m .57536 .86049 L .49015 .84783 L closepath p 1 g
F P
0 g
s
.48864 .9007 m .49818 .90299 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.50147 .7987 m .49015 .84783 L .50021 .84563 L closepath p 1 g
F P
0 g
s
.49015 .84783 m .57536 .86049 L .50021 .84563 L closepath p 1 g
F P
0 g
s
.49818 .90299 m .50835 .90445 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.50021 .84563 m .51092 .84434 L .50147 .7987 L closepath p 1 g
F P
0 g
s
.50021 .84563 m .57536 .86049 L .51092 .84434 L closepath p 1 g
F P
0 g
s
.5666 .80258 m .52886 .5505 L .50147 .7987 L closepath p 1 g
F P
0 g
s
.52194 .84401 m .5666 .80258 L .50147 .7987 L closepath p 1 g
F P
0 g
s
.51092 .84434 m .52194 .84401 L .50147 .7987 L closepath p 1 g
F P
0 g
s
.50835 .90445 m .51887 .90503 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.51092 .84434 m .57536 .86049 L .52194 .84401 L closepath p 1 g
F P
0 g
s
.51887 .90503 m .52943 .90472 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.57536 .86049 m .67855 .71571 L .52194 .84401 L closepath p 1 g
F P
0 g
s
.52194 .84401 m .67855 .71571 L .5666 .80258 L closepath p 1 g
F P
0 g
s
.58072 .27788 m .65497 .28887 L .52778 .26225 L closepath p 1 g
F P
0 g
s
.52778 .26225 m .65497 .28887 L .63104 .2833 L closepath p 1 g
F P
0 g
s
.67058 .64158 m .75761 .48431 L .52886 .5505 L closepath p 1 g
F P
0 g
s
.52886 .5505 m .75761 .48431 L .52961 .48595 L closepath p 1 g
F P
0 g
s
.5666 .80258 m .67058 .64158 L .52886 .5505 L closepath p 1 g
F P
0 g
s
.52943 .90472 m .54948 .9015 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.52943 .90472 m .68094 .88102 L .54948 .9015 L closepath p 1 g
F P
0 g
s
.52961 .48595 m .75761 .48431 L .53005 .42607 L closepath p 1 g
F P
0 g
s
.53005 .42607 m .75761 .48431 L .594 .3782 L closepath p 1 g
F P
0 g
s
.54948 .9015 m .55839 .89867 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.54948 .9015 m .68094 .88102 L .55839 .89867 L closepath p 1 g
F P
0 g
s
.68094 .88102 m .73654 .84103 L .55839 .89867 L closepath p 1 g
F P
0 g
s
.55839 .89867 m .5662 .89513 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.55839 .89867 m .56304 .89716 L p .5662 .89513 L 1 g
F P
0 g
s
.78507 .77487 m .5662 .89513 L .56304 .89716 L 1 g
F
.56304 .89716 m .73507 .84151 L p .78507 .77487 L F P
0 g
s
.5662 .89513 m .78507 .77487 L .57268 .89096 L closepath p 1 g
F P
0 g
s
.5662 .89513 m .57268 .89096 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.67855 .71571 m .67058 .64158 L .5666 .80258 L closepath p 1 g
F P
0 g
s
.57268 .89096 m .78507 .77487 L .78545 .76791 L p 1 g
F P
0 g
s
.59353 .87723 m .57268 .89096 L p .78545 .76791 L 1 g
F P
0 g
s
.57268 .89096 m .57762 .88629 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.57762 .88629 m .57268 .89096 L .59353 .87723 L p 1 g
F P
0 g
s
.57536 .86049 m .74554 .79064 L .74046 .67871 L closepath p 1 g
F P
0 g
s
.57536 .86049 m .74046 .67871 L .67855 .71571 L closepath p 1 g
F P
0 g
s
.57762 .88629 m .74554 .79064 L .57536 .86049 L closepath p 1 g
F P
0 g
s
.63519 .28595 m .58072 .27788 L p .64075 .30574 L 1 g
F P
0 g
s
.58719 .30076 m .69241 .32972 L .58072 .27788 L closepath p 1 g
F P
0 g
s
.58719 .30076 m .78807 .41748 L .69241 .32972 L closepath p 1 g
F P
0 g
s
.59171 .33443 m .78807 .41748 L .58719 .30076 L closepath p 1 g
F P
0 g
s
.59171 .33443 m .79662 .46203 L .78807 .41748 L closepath p 1 g
F P
0 g
s
.594 .3782 m .79662 .46203 L .59171 .33443 L closepath p 1 g
F P
0 g
s
.75761 .48431 m .79662 .46203 L .594 .3782 L closepath p 1 g
F P
0 g
s
.79778 .41033 m .65497 .28887 L .63519 .28595 L p 1 g
F P
0 g
s
.6812 .32452 m .69241 .32972 L .63519 .28595 L 1 g
F
.64075 .30574 m .6812 .32452 L .63519 .28595 L F
.67058 .64158 m .7867 .56653 L .75761 .48431 L closepath p F P
0 g
s
.67855 .71571 m .7867 .56653 L .67058 .64158 L closepath p 1 g
F P
0 g
s
.74046 .67871 m .7867 .56653 L .67855 .71571 L closepath p 1 g
F P
0 g
s
.78807 .41748 m .79778 .41033 L .69241 .32972 L closepath p 1 g
F P
0 g
s
.73507 .84151 m .73654 .84103 L .81588 .7338 L p 1 g
F P
0 g
s
.78634 .78634 m .81588 .7338 L .73654 .84103 L closepath p 1 g
F P
0 g
s
.74046 .67871 m .74554 .79064 L .7867 .56653 L closepath p 1 g
F P
0 g
s
.78545 .76791 m .75272 .78354 L .74554 .79064 L 1 g
F
.74554 .79064 m .78642 .75023 L .7867 .56653 L closepath p F P
0 g
s
.78545 .76791 m .78642 .75023 L p .75272 .78354 L 1 g
F P
0 g
s
.7867 .56653 m .79662 .46203 L .75761 .48431 L closepath p 1 g
F P
0 g
s
.78507 .77487 m .81796 .70549 L .78642 .75023 L closepath p 1 g
F P
0 g
s
.81588 .7338 m .81796 .70549 L .78507 .77487 L closepath p 1 g
F P
0 g
s
.78642 .75023 m .81796 .70549 L .7867 .56653 L closepath p 1 g
F P
0 g
s
.7867 .56653 m .81796 .70549 L .79662 .46203 L closepath p 1 g
F P
0 g
s
.78807 .41748 m .83932 .65816 L .79778 .41033 L closepath p 1 g
F P
0 g
s
.79662 .46203 m .83932 .65816 L .78807 .41748 L closepath p 1 g
F P
0 g
s
.81796 .70549 m .83932 .65816 L .79662 .46203 L closepath p 1 g
F P
0 g
s
.83932 .65816 m .85051 .56255 L .79778 .41033 L closepath p 1 g
F P
0 g
s
.81588 .7338 m .83932 .65816 L .81796 .70549 L closepath p 1 g
F P
0 g
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.0238095 0.539506 0.5 1.05332 MathSubStart
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0539953 1.16055 0 1.16055 [
[.5 1.09151 -44 0 ]
[.5 1.09151 44 9.8125 ]
[ 0 0 0 0 ]
[ 1 1.07901 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
gsave
.5 1.09151 -105 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Polygons) show
117.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(:) show
123.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1220) show
147.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 1.07901 L
0 1.07901 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.20522 .57511 m .20554 .58807 L .20529 .54148 L closepath p 1 g
F P
0 g
s
.20547 .62076 m .20554 .58807 L .20522 .57511 L closepath p 1 g
F P
0 g
s
.20529 .54148 m .21553 .50771 L .21488 .49743 L closepath p 1 g
F P
0 g
s
.20529 .54148 m .21578 .55525 L .21553 .50771 L closepath p 1 g
F P
0 g
s
.20554 .58807 m .21578 .55525 L .20529 .54148 L closepath p 1 g
F P
0 g
s
.20547 .62076 m .21605 .63574 L .20554 .58807 L closepath p 1 g
F P
0 g
s
.21578 .66648 m .21605 .63574 L .20547 .62076 L closepath p 1 g
F P
0 g
s
.20554 .58807 m .22615 .60493 L .21578 .55525 L closepath p 1 g
F P
0 g
s
.21605 .63574 m .22615 .60493 L .20554 .58807 L closepath p 1 g
F P
0 g
s
.21553 .50771 m .22499 .46385 L .21488 .49743 L closepath p 1 g
F P
0 g
s
.21553 .50771 m .23621 .47498 L .22499 .46385 L closepath p 1 g
F P
0 g
s
.21553 .50771 m .23645 .52344 L .23621 .47498 L closepath p 1 g
F P
0 g
s
.21578 .55525 m .23645 .52344 L .21553 .50771 L closepath p 1 g
F P
0 g
s
.21578 .55525 m .2463 .5751 L .23645 .52344 L closepath p 1 g
F P
0 g
s
.22615 .60493 m .2463 .5751 L .21578 .55525 L closepath p 1 g
F P
0 g
s
.21578 .66648 m .23688 .68286 L .21605 .63574 L closepath p 1 g
F P
0 g
s
.23626 .71073 m .23688 .68286 L .21578 .66648 L closepath p 1 g
F P
0 g
s
.22481 .6962 m .23626 .71073 L .21578 .66648 L closepath p 1 g
F P
0 g
s
.21605 .63574 m .24666 .655 L .22615 .60493 L closepath p 1 g
F P
0 g
s
.23688 .68286 m .24666 .655 L .21605 .63574 L closepath p 1 g
F P
0 g
s
.24434 .73775 m .23626 .71073 L .22481 .6962 L closepath p 1 g
F P
0 g
s
.22499 .46385 m .2452 .43133 L .24341 .42493 L closepath p 1 g
F P
0 g
s
.23621 .47498 m .2452 .43133 L .22499 .46385 L closepath p 1 g
F P
0 g
s
.22615 .60493 m .26576 .62808 L .2463 .5751 L closepath p 1 g
F P
0 g
s
.24666 .655 m .26576 .62808 L .22615 .60493 L closepath p 1 g
F P
0 g
s
.23621 .47498 m .26718 .44451 L .2452 .43133 L closepath p 1 g
F P
0 g
s
.23621 .47498 m .26739 .49383 L .26718 .44451 L closepath p 1 g
F P
0 g
s
.23645 .52344 m .26739 .49383 L .23621 .47498 L closepath p 1 g
F P
0 g
s
.23626 .71073 m .26764 .72769 L .23688 .68286 L closepath p 1 g
F P
0 g
s
.24434 .73775 m .2666 .75193 L .23626 .71073 L closepath p 1 g
F P
0 g
s
.2666 .75193 m .26764 .72769 L .23626 .71073 L closepath p 1 g
F P
0 g
s
.23645 .52344 m .27634 .54737 L .26739 .49383 L closepath p 1 g
F P
0 g
s
.2463 .5751 m .27634 .54737 L .23645 .52344 L closepath p 1 g
F P
0 g
s
.23688 .68286 m .27686 .70355 L .24666 .655 L closepath p 1 g
F P
0 g
s
.26764 .72769 m .27686 .70355 L .23688 .68286 L closepath p 1 g
F P
0 g
s
.2452 .43133 m .26268 .39377 L .24341 .42493 L closepath p 1 g
F P
0 g
s
.24434 .73775 m .27337 .77552 L .2666 .75193 L closepath p 1 g
F P
0 g
s
.26044 .76317 m .27337 .77552 L .24434 .73775 L closepath p 1 g
F P
0 g
s
.2452 .43133 m .27534 .40109 L .26268 .39377 L closepath p 1 g
F P
0 g
s
.26718 .44451 m .27534 .40109 L .2452 .43133 L closepath p 1 g
F P
0 g
s
.2463 .5751 m .29398 .60313 L .27634 .54737 L closepath p 1 g
F P
0 g
s
.26576 .62808 m .29398 .60313 L .2463 .5751 L closepath p 1 g
F P
0 g
s
.24666 .655 m .29432 .68032 L .26576 .62808 L closepath p 1 g
F P
0 g
s
.27686 .70355 m .29432 .68032 L .24666 .655 L closepath p 1 g
F P
0 g
s
.26044 .76317 m .2874 .79779 L .27337 .77552 L closepath p 1 g
F P
0 g
s
.26268 .39377 m .29119 .36486 L .28764 .36323 L closepath p 1 g
F P
0 g
s
.27534 .40109 m .29119 .36486 L .26268 .39377 L closepath p 1 g
F P
0 g
s
.26576 .62808 m .31979 .65885 L .29398 .60313 L closepath p 1 g
F P
0 g
s
.29432 .68032 m .31979 .65885 L .26576 .62808 L closepath p 1 g
F P
0 g
s
.2666 .75193 m .3075 .76849 L .26764 .72769 L closepath p 1 g
F P
0 g
s
.30608 .78849 m .3075 .76849 L .2666 .75193 L closepath p 1 g
F P
0 g
s
.27337 .77552 m .30608 .78849 L .2666 .75193 L closepath p 1 g
F P
0 g
s
.26718 .44451 m .30778 .41759 L .27534 .40109 L closepath p 1 g
F P
0 g
s
.26718 .44451 m .30797 .46766 L .30778 .41759 L closepath p 1 g
F P
0 g
s
.26739 .49383 m .30797 .46766 L .26718 .44451 L closepath p 1 g
F P
0 g
s
.26739 .49383 m .31562 .5229 L .30797 .46766 L closepath p 1 g
F P
0 g
s
.27634 .54737 m .31562 .5229 L .26739 .49383 L closepath p 1 g
F P
0 g
s
.26764 .72769 m .31581 .74867 L .27686 .70355 L closepath p 1 g
F P
0 g
s
.3075 .76849 m .31581 .74867 L .26764 .72769 L closepath p 1 g
F P
0 g
s
.27337 .77552 m .3113 .80805 L .30608 .78849 L closepath p 1 g
F P
0 g
s
.2874 .79779 m .3113 .80805 L .27337 .77552 L closepath p 1 g
F P
0 g
s
.27534 .40109 m .31477 .3744 L .29119 .36486 L closepath p 1 g
F P
0 g
s
.30778 .41759 m .31477 .3744 L .27534 .40109 L closepath p 1 g
F P
0 g
s
.27634 .54737 m .33068 .58115 L .31562 .5229 L closepath p 1 g
F P
0 g
s
.29398 .60313 m .33068 .58115 L .27634 .54737 L closepath p 1 g
F P
0 g
s
.27686 .70355 m .331 .72968 L .29432 .68032 L closepath p 1 g
F P
0 g
s
.31581 .74867 m .331 .72968 L .27686 .70355 L closepath p 1 g
F P
0 g
s
.30798 .8182 m .32277 .82661 L .2874 .79779 L closepath p 1 g
F P
0 g
s
.2874 .79779 m .32277 .82661 L .3113 .80805 L closepath p 1 g
F P
0 g
s
.29119 .36486 m .31377 .33666 L .28764 .36323 L closepath p 1 g
F P
0 g
s
.29119 .36486 m .32831 .33938 L .31377 .33666 L closepath p 1 g
F P
0 g
s
.31477 .3744 m .32831 .33938 L .29119 .36486 L closepath p 1 g
F P
0 g
s
.29398 .60313 m .35268 .64 L .33068 .58115 L closepath p 1 g
F P
0 g
s
.31979 .65885 m .35268 .64 L .29398 .60313 L closepath p 1 g
F P
0 g
s
.29432 .68032 m .35282 .71222 L .31979 .65885 L closepath p 1 g
F P
0 g
s
.331 .72968 m .35282 .71222 L .29432 .68032 L closepath p 1 g
F P
0 g
s
.30608 .78849 m .35516 .80361 L .3075 .76849 L closepath p 1 g
F P
0 g
s
.3113 .80805 m .35354 .81893 L .30608 .78849 L closepath p 1 g
F P
0 g
s
.30608 .78849 m .35354 .81893 L .35516 .80361 L closepath p 1 g
F P
0 g
s
.3075 .76849 m .36214 .78852 L .31581 .74867 L closepath p 1 g
F P
0 g
s
.35516 .80361 m .36214 .78852 L .3075 .76849 L closepath p 1 g
F P
0 g
s
.30778 .41759 m .35687 .39542 L .31477 .3744 L closepath p 1 g
F P
0 g
s
.35701 .44612 m .35687 .39542 L .30778 .41759 L closepath p 1 g
F P
0 g
s
.30797 .46766 m .35701 .44612 L .30778 .41759 L closepath p 1 g
F P
0 g
s
.30797 .46766 m .36302 .50277 L .35701 .44612 L closepath p 1 g
F P
0 g
s
.31562 .5229 m .36302 .50277 L .30797 .46766 L closepath p 1 g
F P
0 g
s
.33991 .84369 m .32277 .82661 L .30798 .8182 L closepath p 1 g
F P
0 g
s
.3113 .80805 m .35715 .834 L .35354 .81893 L closepath p 1 g
F P
0 g
s
.32277 .82661 m .35715 .834 L .3113 .80805 L closepath p 1 g
F P
0 g
s
.31377 .33666 m .34755 .31332 L .34196 .31698 L closepath p 1 g
F P
0 g
s
.32831 .33938 m .34755 .31332 L .31377 .33666 L closepath p 1 g
F P
0 g
s
.31477 .3744 m .36235 .35245 L .32831 .33938 L closepath p 1 g
F P
0 g
s
.35687 .39542 m .36235 .35245 L .31477 .3744 L closepath p 1 g
F P
0 g
s
.31562 .5229 m .37482 .5631 L .36302 .50277 L closepath p 1 g
F P
0 g
s
.33068 .58115 m .37482 .5631 L .31562 .5229 L closepath p 1 g
F P
0 g
s
.31581 .74867 m .37441 .77414 L .331 .72968 L closepath p 1 g
F P
0 g
s
.36214 .78852 m .37441 .77414 L .31581 .74867 L closepath p 1 g
F P
0 g
s
.31979 .65885 m .38073 .69695 L .35268 .64 L closepath p 1 g
F P
0 g
s
.35282 .71222 m .38073 .69695 L .31979 .65885 L closepath p 1 g
F P
0 g
s
.33991 .84369 m .36573 .84837 L .32277 .82661 L closepath p 1 g
F P
0 g
s
.32277 .82661 m .36573 .84837 L .35715 .834 L closepath p 1 g
F P
0 g
s
.32831 .33938 m .37296 .31847 L .34755 .31332 L closepath p 1 g
F P
0 g
s
.36235 .35245 m .37296 .31847 L .32831 .33938 L closepath p 1 g
F P
0 g
s
.33068 .58115 m .39201 .62457 L .37482 .5631 L closepath p 1 g
F P
0 g
s
.35268 .64 m .39201 .62457 L .33068 .58115 L closepath p 1 g
F P
0 g
s
.331 .72968 m .39172 .76099 L .35282 .71222 L closepath p 1 g
F P
0 g
s
.37441 .77414 m .39172 .76099 L .331 .72968 L closepath p 1 g
F P
0 g
s
.37888 .86167 m .36573 .84837 L .33991 .84369 L closepath p 1 g
F P
0 g
s
.33991 .84369 m .36207 .85886 L .37888 .86167 L closepath p 1 g
F P
0 g
s
.34755 .31332 m .37152 .29656 L .34196 .31698 L closepath p 1 g
F P
0 g
s
.37296 .31847 m .38801 .2942 L .34755 .31332 L closepath p 1 g
F P
0 g
s
.34755 .31332 m .38801 .2942 L .37152 .29656 L closepath p 1 g
F P
0 g
s
.35268 .64 m .41388 .68448 L .39201 .62457 L closepath p 1 g
F P
0 g
s
.38073 .69695 m .41388 .68448 L .35268 .64 L closepath p 1 g
F P
0 g
s
.35282 .71222 m .41362 .74955 L .38073 .69695 L closepath p 1 g
F P
0 g
s
.39172 .76099 m .41362 .74955 L .35282 .71222 L closepath p 1 g
F P
0 g
s
.35354 .81893 m .4089 .8316 L .35516 .80361 L closepath p 1 g
F P
0 g
s
.35715 .834 m .40739 .84194 L .35354 .81893 L closepath p 1 g
F P
0 g
s
.35354 .81893 m .40739 .84194 L .4089 .8316 L closepath p 1 g
F P
0 g
s
.35516 .80361 m .41406 .82148 L .36214 .78852 L closepath p 1 g
F P
0 g
s
.35516 .80361 m .4089 .8316 L .41406 .82148 L closepath p 1 g
F P
0 g
s
.35687 .39542 m .41276 .37908 L .36235 .35245 L closepath p 1 g
F P
0 g
s
.41286 .43025 m .41276 .37908 L .35687 .39542 L closepath p 1 g
F P
0 g
s
.35701 .44612 m .41286 .43025 L .35687 .39542 L closepath p 1 g
F P
0 g
s
.41692 .48794 m .41286 .43025 L .35701 .44612 L closepath p 1 g
F P
0 g
s
.36302 .50277 m .41692 .48794 L .35701 .44612 L closepath p 1 g
F P
0 g
s
.35715 .834 m .40949 .85217 L .40739 .84194 L closepath p 1 g
F P
0 g
s
.36573 .84837 m .40949 .85217 L .35715 .834 L closepath p 1 g
F P
0 g
s
.36207 .85886 m .38849 .8718 L .39611 .87355 L closepath p 1 g
F P
0 g
s
.37888 .86167 m .36207 .85886 L .39611 .87355 L closepath p 1 g
F P
0 g
s
.36214 .78852 m .42277 .8119 L .37441 .77414 L closepath p 1 g
F P
0 g
s
.41406 .82148 m .42277 .8119 L .36214 .78852 L closepath p 1 g
F P
0 g
s
.36235 .35245 m .41646 .33628 L .37296 .31847 L closepath p 1 g
F P
0 g
s
.41276 .37908 m .41646 .33628 L .36235 .35245 L closepath p 1 g
F P
0 g
s
.42488 .54983 m .41692 .48794 L .36302 .50277 L closepath p 1 g
F P
0 g
s
.37482 .5631 m .42488 .54983 L .36302 .50277 L closepath p 1 g
F P
0 g
s
.37888 .86167 m .41507 .862 L .36573 .84837 L closepath p 1 g
F P
0 g
s
.36573 .84837 m .41507 .862 L .40949 .85217 L closepath p 1 g
F P
0 g
s
.38801 .2942 m .40671 .27988 L .37152 .29656 L closepath p 1 g
F P
0 g
s
.37296 .31847 m .42363 .30308 L .38801 .2942 L closepath p 1 g
F P
0 g
s
.41646 .33628 m .42363 .30308 L .37296 .31847 L closepath p 1 g
F P
0 g
s
.37441 .77414 m .43483 .8032 L .39172 .76099 L closepath p 1 g
F P
0 g
s
.42277 .8119 m .43483 .8032 L .37441 .77414 L closepath p 1 g
F P
0 g
s
.43647 .61324 m .42488 .54983 L .37482 .5631 L closepath p 1 g
F P
0 g
s
.39201 .62457 m .43647 .61324 L .37482 .5631 L closepath p 1 g
F P
0 g
s
.42388 .87115 m .41507 .862 L .37888 .86167 L closepath p 1 g
F P
0 g
s
.37888 .86167 m .39611 .87355 L .42388 .87115 L closepath p 1 g
F P
0 g
s
.38073 .69695 m .43944 .74025 L .41388 .68448 L closepath p 1 g
F P
0 g
s
.41362 .74955 m .43944 .74025 L .38073 .69695 L closepath p 1 g
F P
0 g
s
.38801 .2942 m .43378 .28016 L .40671 .27988 L closepath p 1 g
F P
0 g
s
.42363 .30308 m .43378 .28016 L .38801 .2942 L closepath p 1 g
F P
0 g
s
.39611 .87355 m .38849 .8718 L .41684 .88372 L closepath p 1 g
F P
0 g
s
.39172 .76099 m .4499 .79566 L .41362 .74955 L closepath p 1 g
F P
0 g
s
.43483 .8032 m .4499 .79566 L .39172 .76099 L closepath p 1 g
F P
0 g
s
.45119 .67536 m .43647 .61324 L .39201 .62457 L closepath p 1 g
F P
0 g
s
.41388 .68448 m .45119 .67536 L .39201 .62457 L closepath p 1 g
F P
0 g
s
.42388 .87115 m .39611 .87355 L .43561 .87937 L closepath p 1 g
F P
0 g
s
.39611 .87355 m .41684 .88372 L .43561 .87937 L closepath p 1 g
F P
0 g
s
.43378 .28016 m .44636 .26766 L .40671 .27988 L closepath p 1 g
F P
0 g
s
.40739 .84194 m .46669 .85131 L .4089 .8316 L closepath p 1 g
F P
0 g
s
.40739 .84194 m .4657 .8565 L .46669 .85131 L closepath p 1 g
F P
0 g
s
.40949 .85217 m .4657 .8565 L .40739 .84194 L closepath p 1 g
F P
0 g
s
.4089 .8316 m .46951 .84626 L .41406 .82148 L closepath p 1 g
F P
0 g
s
.4089 .8316 m .46669 .85131 L .46951 .84626 L closepath p 1 g
F P
0 g
s
.41507 .862 m .46656 .86168 L .40949 .85217 L closepath p 1 g
F P
0 g
s
.40949 .85217 m .46656 .86168 L .4657 .8565 L closepath p 1 g
F P
0 g
s
.41276 .37908 m .47332 .3694 L .41646 .33628 L closepath p 1 g
F P
0 g
s
.47337 .42084 m .47332 .3694 L .41276 .37908 L closepath p 1 g
F P
0 g
s
.41286 .43025 m .47337 .42084 L .41276 .37908 L closepath p 1 g
F P
0 g
s
.47527 .47916 m .47337 .42084 L .41286 .43025 L closepath p 1 g
F P
0 g
s
.41692 .48794 m .47527 .47916 L .41286 .43025 L closepath p 1 g
F P
0 g
s
.41362 .74955 m .46752 .78956 L .43944 .74025 L closepath p 1 g
F P
0 g
s
.41362 .74955 m .4499 .79566 L .46752 .78956 L closepath p 1 g
F P
0 g
s
.41388 .68448 m .46834 .73346 L .45119 .67536 L closepath p 1 g
F P
0 g
s
.43944 .74025 m .46834 .73346 L .41388 .68448 L closepath p 1 g
F P
0 g
s
.41406 .82148 m .47409 .84151 L .42277 .8119 L closepath p 1 g
F P
0 g
s
.41406 .82148 m .46951 .84626 L .47409 .84151 L closepath p 1 g
F P
0 g
s
.41507 .862 m .46922 .86669 L .46656 .86168 L closepath p 1 g
F P
0 g
s
.42388 .87115 m .46922 .86669 L .41507 .862 L closepath p 1 g
F P
0 g
s
.41646 .33628 m .47506 .3267 L .42363 .30308 L closepath p 1 g
F P
0 g
s
.41646 .33628 m .47332 .3694 L .47506 .3267 L closepath p 1 g
F P
0 g
s
.43561 .87937 m .41684 .88372 L .44988 .88644 L closepath p 1 g
F P
0 g
s
.41684 .88372 m .44046 .89195 L .44988 .88644 L closepath p 1 g
F P
0 g
s
.479 .54198 m .47527 .47916 L .41692 .48794 L closepath p 1 g
F P
0 g
s
.42488 .54983 m .479 .54198 L .41692 .48794 L closepath p 1 g
F P
0 g
s
.42277 .8119 m .48031 .83723 L .43483 .8032 L closepath p 1 g
F P
0 g
s
.42277 .8119 m .47409 .84151 L .48031 .83723 L closepath p 1 g
F P
0 g
s
.42363 .30308 m .47841 .29397 L .43378 .28016 L closepath p 1 g
F P
0 g
s
.47506 .3267 m .47841 .29397 L .42363 .30308 L closepath p 1 g
F P
0 g
s
.42388 .87115 m .47357 .87138 L .46922 .86669 L closepath p 1 g
F P
0 g
s
.42388 .87115 m .43561 .87937 L .47357 .87138 L closepath p 1 g
F P
0 g
s
.48443 .60655 m .479 .54198 L .42488 .54983 L closepath p 1 g
F P
0 g
s
.43647 .61324 m .48443 .60655 L .42488 .54983 L closepath p 1 g
F P
0 g
s
.43378 .28016 m .48317 .27186 L .44636 .26766 L closepath p 1 g
F P
0 g
s
.47841 .29397 m .48317 .27186 L .43378 .28016 L closepath p 1 g
F P
0 g
s
.48798 .83354 m .4499 .79566 L .43483 .8032 L closepath p 1 g
F P
0 g
s
.43483 .8032 m .48031 .83723 L .48798 .83354 L closepath p 1 g
F P
0 g
s
.47357 .87138 m .43561 .87937 L .47947 .87562 L closepath p 1 g
F P
0 g
s
.43561 .87937 m .44988 .88644 L .47947 .87562 L closepath p 1 g
F P
0 g
s
.49131 .66998 m .48443 .60655 L .43647 .61324 L closepath p 1 g
F P
0 g
s
.45119 .67536 m .49131 .66998 L .43647 .61324 L closepath p 1 g
F P
0 g
s
.48713 .78512 m .46834 .73346 L .43944 .74025 L closepath p 1 g
F P
0 g
s
.46752 .78956 m .48713 .78512 L .43944 .74025 L closepath p 1 g
F P
0 g
s
.44988 .88644 m .44046 .89195 L .46626 .89218 L closepath p 1 g
F P
0 g
s
.44046 .89195 m .4663 .89806 L .46626 .89218 L closepath p 1 g
F P
0 g
s
.48317 .27186 m .48905 .26045 L .44636 .26766 L closepath p 1 g
F P
0 g
s
.47947 .87562 m .44988 .88644 L .48674 .87929 L closepath p 1 g
F P
0 g
s
.44988 .88644 m .46626 .89218 L .48674 .87929 L closepath p 1 g
F P
0 g
s
.46752 .78956 m .4499 .79566 L .49688 .83056 L closepath p 1 g
F P
0 g
s
.49688 .83056 m .4499 .79566 L .48798 .83354 L closepath p 1 g
F P
0 g
s
.49932 .72947 m .49131 .66998 L .45119 .67536 L closepath p 1 g
F P
0 g
s
.45119 .67536 m .46834 .73346 L .49932 .72947 L closepath p 1 g
F P
0 g
s
.46656 .86168 m .57435 .8385 L .4657 .8565 L closepath p 1 g
F P
0 g
s
.4657 .8565 m .57435 .8385 L .46669 .85131 L closepath p 1 g
F P
0 g
s
.48674 .87929 m .46626 .89218 L .49515 .88229 L closepath p 1 g
F P
0 g
s
.46626 .89218 m .48427 .89646 L .49515 .88229 L closepath p 1 g
F P
0 g
s
.46626 .89218 m .4663 .89806 L .48427 .89646 L closepath p 1 g
F P
0 g
s
.4663 .89806 m .49367 .90193 L .48427 .89646 L closepath p 1 g
F P
0 g
s
.46656 .86168 m .46922 .86669 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.46669 .85131 m .57435 .8385 L .46951 .84626 L closepath p 1 g
F P
0 g
s
.46752 .78956 m .50672 .82841 L .48713 .78512 L closepath p 1 g
F P
0 g
s
.46752 .78956 m .49688 .83056 L .50672 .82841 L closepath p 1 g
F P
0 g
s
.46834 .73346 m .50808 .78252 L .49932 .72947 L closepath p 1 g
F P
0 g
s
.48713 .78512 m .50808 .78252 L .46834 .73346 L closepath p 1 g
F P
0 g
s
.46922 .86669 m .47357 .87138 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.46951 .84626 m .57435 .8385 L .47409 .84151 L closepath p 1 g
F P
0 g
s
.47506 .3267 m .47332 .3694 L .53609 .36688 L closepath p 1 g
F P
0 g
s
.53609 .36688 m .47332 .3694 L .53608 .41839 L closepath p 1 g
F P
0 g
s
.53608 .41839 m .47332 .3694 L .47337 .42084 L closepath p 1 g
F P
0 g
s
.53573 .47687 m .53608 .41839 L .47337 .42084 L closepath p 1 g
F P
0 g
s
.47527 .47916 m .53573 .47687 L .47337 .42084 L closepath p 1 g
F P
0 g
s
.47357 .87138 m .47947 .87562 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.47409 .84151 m .57435 .8385 L .48031 .83723 L closepath p 1 g
F P
0 g
s
.47506 .3267 m .53577 .32421 L .47841 .29397 L closepath p 1 g
F P
0 g
s
.53609 .36688 m .53577 .32421 L .47506 .3267 L closepath p 1 g
F P
0 g
s
.53504 .53994 m .53573 .47687 L .47527 .47916 L closepath p 1 g
F P
0 g
s
.479 .54198 m .53504 .53994 L .47527 .47916 L closepath p 1 g
F P
0 g
s
.47841 .29397 m .53515 .2916 L .48317 .27186 L closepath p 1 g
F P
0 g
s
.53577 .32421 m .53515 .2916 L .47841 .29397 L closepath p 1 g
F P
0 g
s
.53403 .60481 m .53504 .53994 L .479 .54198 L closepath p 1 g
F P
0 g
s
.48443 .60655 m .53403 .60481 L .479 .54198 L closepath p 1 g
F P
0 g
s
.47947 .87562 m .48674 .87929 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.48031 .83723 m .57435 .8385 L .48798 .83354 L closepath p 1 g
F P
0 g
s
.48317 .27186 m .53426 .2697 L .48905 .26045 L closepath p 1 g
F P
0 g
s
.53515 .2916 m .53426 .2697 L .48317 .27186 L closepath p 1 g
F P
0 g
s
.49515 .88229 m .48427 .89646 L .50446 .88453 L closepath p 1 g
F P
0 g
s
.48427 .89646 m .50342 .89918 L .50446 .88453 L closepath p 1 g
F P
0 g
s
.48427 .89646 m .49367 .90193 L .50342 .89918 L closepath p 1 g
F P
0 g
s
.53276 .66859 m .53403 .60481 L .48443 .60655 L closepath p 1 g
F P
0 g
s
.49131 .66998 m .53276 .66859 L .48443 .60655 L closepath p 1 g
F P
0 g
s
.48674 .87929 m .49515 .88229 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.48713 .78512 m .51719 .82715 L .50808 .78252 L closepath p 1 g
F P
0 g
s
.50672 .82841 m .51719 .82715 L .48713 .78512 L closepath p 1 g
F P
0 g
s
.48798 .83354 m .57435 .8385 L .49688 .83056 L closepath p 1 g
F P
0 g
s
.48905 .26045 m .53317 .25858 L .4958 .25939 L closepath p 1 g
F P
0 g
s
.53426 .2697 m .53317 .25858 L .48905 .26045 L closepath p 1 g
F P
0 g
s
.53127 .72843 m .53276 .66859 L .49131 .66998 L closepath p 1 g
F P
0 g
s
.49932 .72947 m .53127 .72843 L .49131 .66998 L closepath p 1 g
F P
0 g
s
.49367 .90193 m .52186 .90347 L .50342 .89918 L closepath p 1 g
F P
0 g
s
.49515 .88229 m .50446 .88453 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.4958 .25939 m .53317 .25858 L .53192 .25785 L closepath p 1 g
F P
0 g
s
.49688 .83056 m .57435 .8385 L .50672 .82841 L closepath p 1 g
F P
0 g
s
.52965 .78184 m .53127 .72843 L .49932 .72947 L closepath p 1 g
F P
0 g
s
.49932 .72947 m .50808 .78252 L .52965 .78184 L closepath p 1 g
F P
0 g
s
.50446 .88453 m .50342 .89918 L .51439 .88595 L closepath p 1 g
F P
0 g
s
.50342 .89918 m .52318 .90026 L .51439 .88595 L closepath p 1 g
F P
0 g
s
.50342 .89918 m .52186 .90347 L .52318 .90026 L closepath p 1 g
F P
0 g
s
.50446 .88453 m .51439 .88595 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.50672 .82841 m .57435 .8385 L .51719 .82715 L closepath p 1 g
F P
0 g
s
.50808 .78252 m .52796 .82682 L .52965 .78184 L closepath p 1 g
F P
0 g
s
.51719 .82715 m .52796 .82682 L .50808 .78252 L closepath p 1 g
F P
0 g
s
.51439 .88595 m .52467 .88652 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.51439 .88595 m .52318 .90026 L .52467 .88652 L closepath p 1 g
F P
0 g
s
.51719 .82715 m .57435 .8385 L .52796 .82682 L closepath p 1 g
F P
0 g
s
.52186 .90347 m .55016 .90266 L .52318 .90026 L closepath p 1 g
F P
0 g
s
.52318 .90026 m .54302 .89969 L .52467 .88652 L closepath p 1 g
F P
0 g
s
.52318 .90026 m .55016 .90266 L .54302 .89969 L closepath p 1 g
F P
0 g
s
.52467 .88652 m .53498 .88622 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.52467 .88652 m .54302 .89969 L .53498 .88622 L closepath p 1 g
F P
0 g
s
.52796 .82682 m .53867 .82744 L .52965 .78184 L closepath p 1 g
F P
0 g
s
.52796 .82682 m .57435 .8385 L .53867 .82744 L closepath p 1 g
F P
0 g
s
.5511 .78312 m .53127 .72843 L .52965 .78184 L closepath p 1 g
F P
0 g
s
.53867 .82744 m .5511 .78312 L .52965 .78184 L closepath p 1 g
F P
0 g
s
.56304 .7304 m .53276 .66859 L .53127 .72843 L closepath p 1 g
F P
0 g
s
.5511 .78312 m .56304 .7304 L .53127 .72843 L closepath p 1 g
F P
0 g
s
.57703 .26214 m .56783 .26077 L .53192 .25785 L closepath p 1 g
F P
0 g
s
.53192 .25785 m .53317 .25858 L .57703 .26214 L closepath p 1 g
F P
0 g
s
.57396 .67124 m .53403 .60481 L .53276 .66859 L closepath p 1 g
F P
0 g
s
.56304 .7304 m .57396 .67124 L .53276 .66859 L closepath p 1 g
F P
0 g
s
.53426 .2697 m .58505 .2738 L .53317 .25858 L closepath p 1 g
F P
0 g
s
.53317 .25858 m .58505 .2738 L .57703 .26214 L closepath p 1 g
F P
0 g
s
.53403 .60481 m .58333 .60811 L .53504 .53994 L closepath p 1 g
F P
0 g
s
.57396 .67124 m .58333 .60811 L .53403 .60481 L closepath p 1 g
F P
0 g
s
.59152 .2961 m .58505 .2738 L .53426 .2697 L closepath p 1 g
F P
0 g
s
.53515 .2916 m .59152 .2961 L .53426 .2697 L closepath p 1 g
F P
0 g
s
.53498 .88622 m .54505 .88506 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.56242 .89748 m .54505 .88506 L .53498 .88622 L closepath p 1 g
F P
0 g
s
.54302 .89969 m .56242 .89748 L .53498 .88622 L closepath p 1 g
F P
0 g
s
.53504 .53994 m .59072 .54381 L .53573 .47687 L closepath p 1 g
F P
0 g
s
.58333 .60811 m .59072 .54381 L .53504 .53994 L closepath p 1 g
F P
0 g
s
.59609 .32894 m .59152 .2961 L .53515 .2916 L closepath p 1 g
F P
0 g
s
.53577 .32421 m .59609 .32894 L .53515 .2916 L closepath p 1 g
F P
0 g
s
.53573 .47687 m .5958 .48121 L .53608 .41839 L closepath p 1 g
F P
0 g
s
.59072 .54381 m .5958 .48121 L .53573 .47687 L closepath p 1 g
F P
0 g
s
.53577 .32421 m .59845 .37166 L .59609 .32894 L closepath p 1 g
F P
0 g
s
.53609 .36688 m .59845 .37166 L .53577 .32421 L closepath p 1 g
F P
0 g
s
.53608 .41839 m .59839 .42304 L .53609 .36688 L closepath p 1 g
F P
0 g
s
.5958 .48121 m .59839 .42304 L .53608 .41839 L closepath p 1 g
F P
0 g
s
.59839 .42304 m .59845 .37166 L .53609 .36688 L closepath p 1 g
F P
0 g
s
.54898 .82899 m .5511 .78312 L .53867 .82744 L closepath p 1 g
F P
0 g
s
.53867 .82744 m .57435 .8385 L .54898 .82899 L closepath p 1 g
F P
0 g
s
.54302 .89969 m .57786 .89951 L .56242 .89748 L closepath p 1 g
F P
0 g
s
.55016 .90266 m .57786 .89951 L .54302 .89969 L closepath p 1 g
F P
0 g
s
.54505 .88506 m .55459 .88307 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.54505 .88506 m .58084 .89368 L .55459 .88307 L closepath p 1 g
F P
0 g
s
.56242 .89748 m .58084 .89368 L .54505 .88506 L closepath p 1 g
F P
0 g
s
.54898 .82899 m .57171 .78631 L .5511 .78312 L closepath p 1 g
F P
0 g
s
.55856 .83141 m .57171 .78631 L .54898 .82899 L closepath p 1 g
F P
0 g
s
.54898 .82899 m .57435 .8385 L .55856 .83141 L closepath p 1 g
F P
0 g
s
.57171 .78631 m .56304 .7304 L .5511 .78312 L closepath p 1 g
F P
0 g
s
.55459 .88307 m .56331 .8803 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.55459 .88307 m .59778 .88837 L .56331 .8803 L closepath p 1 g
F P
0 g
s
.55459 .88307 m .58084 .89368 L .59778 .88837 L closepath p 1 g
F P
0 g
s
.59076 .79129 m .57171 .78631 L .55856 .83141 L closepath p 1 g
F P
0 g
s
.56711 .83461 m .59076 .79129 L .55856 .83141 L closepath p 1 g
F P
0 g
s
.55856 .83141 m .57435 .8385 L .56711 .83461 L closepath p 1 g
F P
0 g
s
.56242 .89748 m .60425 .89409 L .58084 .89368 L closepath p 1 g
F P
0 g
s
.57786 .89951 m .60425 .89409 L .56242 .89748 L closepath p 1 g
F P
0 g
s
.59349 .73529 m .57396 .67124 L .56304 .7304 L closepath p 1 g
F P
0 g
s
.57171 .78631 m .59349 .73529 L .56304 .7304 L closepath p 1 g
F P
0 g
s
.56331 .8803 m .57096 .87683 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.56331 .8803 m .61275 .8817 L .57096 .87683 L closepath p 1 g
F P
0 g
s
.59778 .88837 m .61275 .8817 L .56331 .8803 L closepath p 1 g
F P
0 g
s
.60765 .79786 m .59076 .79129 L .56711 .83461 L closepath p 1 g
F P
0 g
s
.57435 .8385 m .60765 .79786 L .56711 .83461 L closepath p 1 g
F P
0 g
s
.57096 .87683 m .57731 .87276 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.57096 .87683 m .62532 .87382 L .57731 .87276 L closepath p 1 g
F P
0 g
s
.61275 .8817 m .62532 .87382 L .57096 .87683 L closepath p 1 g
F P
0 g
s
.59076 .79129 m .59349 .73529 L .57171 .78631 L closepath p 1 g
F P
0 g
s
.61336 .67782 m .58333 .60811 L .57396 .67124 L closepath p 1 g
F P
0 g
s
.59349 .73529 m .61336 .67782 L .57396 .67124 L closepath p 1 g
F P
0 g
s
.57435 .8385 m .62182 .8058 L .60765 .79786 L closepath p 1 g
F P
0 g
s
.58008 .84295 m .62182 .8058 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.57435 .8385 m .58412 .84781 L .58008 .84295 L closepath p 1 g
F P
0 g
s
.58637 .85292 m .58412 .84781 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.58678 .85812 m .58637 .85292 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.58536 .86327 m .58678 .85812 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.58216 .86819 m .58536 .86327 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.57731 .87276 m .58216 .86819 L .57435 .8385 L closepath p 1 g
F P
0 g
s
.57703 .26214 m .6335 .28394 L .61894 .27095 L closepath p 1 g
F P
0 g
s
.58505 .2738 m .6335 .28394 L .57703 .26214 L closepath p 1 g
F P
0 g
s
.57731 .87276 m .63508 .86494 L .58216 .86819 L closepath p 1 g
F P
0 g
s
.62532 .87382 m .63508 .86494 L .57731 .87276 L closepath p 1 g
F P
0 g
s
.58008 .84295 m .63287 .81481 L .62182 .8058 L closepath p 1 g
F P
0 g
s
.58008 .84295 m .58412 .84781 L .63287 .81481 L closepath p 1 g
F P
0 g
s
.60425 .89409 m .62863 .8865 L .58084 .89368 L closepath p 1 g
F P
0 g
s
.58084 .89368 m .62863 .8865 L .59778 .88837 L closepath p 1 g
F P
0 g
s
.63508 .86494 m .64169 .85529 L .58216 .86819 L closepath p 1 g
F P
0 g
s
.58216 .86819 m .64169 .85529 L .58536 .86327 L closepath p 1 g
F P
0 g
s
.58333 .60811 m .63038 .61629 L .59072 .54381 L closepath p 1 g
F P
0 g
s
.61336 .67782 m .63038 .61629 L .58333 .60811 L closepath p 1 g
F P
0 g
s
.58412 .84781 m .64049 .82459 L .63287 .81481 L closepath p 1 g
F P
0 g
s
.58637 .85292 m .64049 .82459 L .58412 .84781 L closepath p 1 g
F P
0 g
s
.59152 .2961 m .64524 .30723 L .58505 .2738 L closepath p 1 g
F P
0 g
s
.64524 .30723 m .6335 .28394 L .58505 .2738 L closepath p 1 g
F P
0 g
s
.64169 .85529 m .64489 .84515 L .58536 .86327 L closepath p 1 g
F P
0 g
s
.58536 .86327 m .64489 .84515 L .58678 .85812 L closepath p 1 g
F P
0 g
s
.58637 .85292 m .64452 .83481 L .64049 .82459 L closepath p 1 g
F P
0 g
s
.58678 .85812 m .64452 .83481 L .58637 .85292 L closepath p 1 g
F P
0 g
s
.64489 .84515 m .64452 .83481 L .58678 .85812 L closepath p 1 g
F P
0 g
s
.59072 .54381 m .64379 .55341 L .5958 .48121 L closepath p 1 g
F P
0 g
s
.63038 .61629 m .64379 .55341 L .59072 .54381 L closepath p 1 g
F P
0 g
s
.59076 .79129 m .62153 .74289 L .59349 .73529 L closepath p 1 g
F P
0 g
s
.60765 .79786 m .62153 .74289 L .59076 .79129 L closepath p 1 g
F P
0 g
s
.59609 .32894 m .65353 .34064 L .59152 .2961 L closepath p 1 g
F P
0 g
s
.65353 .34064 m .64524 .30723 L .59152 .2961 L closepath p 1 g
F P
0 g
s
.62153 .74289 m .61336 .67782 L .59349 .73529 L closepath p 1 g
F P
0 g
s
.5958 .48121 m .653 .49194 L .59839 .42304 L closepath p 1 g
F P
0 g
s
.64379 .55341 m .653 .49194 L .5958 .48121 L closepath p 1 g
F P
0 g
s
.65781 .38349 m .65353 .34064 L .59609 .32894 L closepath p 1 g
F P
0 g
s
.59609 .32894 m .59845 .37166 L .65781 .38349 L closepath p 1 g
F P
0 g
s
.62863 .8865 m .65034 .87691 L .59778 .88837 L closepath p 1 g
F P
0 g
s
.59778 .88837 m .65034 .87691 L .61275 .8817 L closepath p 1 g
F P
0 g
s
.6577 .43453 m .59845 .37166 L .59839 .42304 L closepath p 1 g
F P
0 g
s
.653 .49194 m .6577 .43453 L .59839 .42304 L closepath p 1 g
F P
0 g
s
.65781 .38349 m .59845 .37166 L .6577 .43453 L closepath p 1 g
F P
0 g
s
.60765 .79786 m .64621 .7529 L .62153 .74289 L closepath p 1 g
F P
0 g
s
.62182 .8058 m .64621 .7529 L .60765 .79786 L closepath p 1 g
F P
0 g
s
.61275 .8817 m .66872 .86554 L .62532 .87382 L closepath p 1 g
F P
0 g
s
.65034 .87691 m .66872 .86554 L .61275 .8817 L closepath p 1 g
F P
0 g
s
.6495 .68803 m .63038 .61629 L .61336 .67782 L closepath p 1 g
F P
0 g
s
.62153 .74289 m .6495 .68803 L .61336 .67782 L closepath p 1 g
F P
0 g
s
.61894 .27095 m .6335 .28394 L .67779 .29965 L closepath p 1 g
F P
0 g
s
.61894 .27095 m .67779 .29965 L .65735 .28463 L closepath p 1 g
F P
0 g
s
.64621 .7529 m .6495 .68803 L .62153 .74289 L closepath p 1 g
F P
0 g
s
.62182 .8058 m .66674 .76493 L .64621 .7529 L closepath p 1 g
F P
0 g
s
.63287 .81481 m .66674 .76493 L .62182 .8058 L closepath p 1 g
F P
0 g
s
.62532 .87382 m .6832 .85265 L .63508 .86494 L closepath p 1 g
F P
0 g
s
.66872 .86554 m .6832 .85265 L .62532 .87382 L closepath p 1 g
F P
0 g
s
.65508 .87533 m .68268 .86315 L .62863 .8865 L closepath p 1 g
F P
0 g
s
.62863 .8865 m .68268 .86315 L .65034 .87691 L closepath p 1 g
F P
0 g
s
.67341 .62896 m .64379 .55341 L .63038 .61629 L closepath p 1 g
F P
0 g
s
.6495 .68803 m .67341 .62896 L .63038 .61629 L closepath p 1 g
F P
0 g
s
.63287 .81481 m .68252 .77851 L .66674 .76493 L closepath p 1 g
F P
0 g
s
.63287 .81481 m .64049 .82459 L .68252 .77851 L closepath p 1 g
F P
0 g
s
.6335 .28394 m .69423 .32443 L .67779 .29965 L closepath p 1 g
F P
0 g
s
.64524 .30723 m .69423 .32443 L .6335 .28394 L closepath p 1 g
F P
0 g
s
.63508 .86494 m .69325 .83857 L .64169 .85529 L closepath p 1 g
F P
0 g
s
.6832 .85265 m .69325 .83857 L .63508 .86494 L closepath p 1 g
F P
0 g
s
.64049 .82459 m .69316 .79316 L .68252 .77851 L closepath p 1 g
F P
0 g
s
.64452 .83481 m .69316 .79316 L .64049 .82459 L closepath p 1 g
F P
0 g
s
.64169 .85529 m .69845 .82367 L .64489 .84515 L closepath p 1 g
F P
0 g
s
.69325 .83857 m .69845 .82367 L .64169 .85529 L closepath p 1 g
F P
0 g
s
.69219 .56825 m .653 .49194 L .64379 .55341 L closepath p 1 g
F P
0 g
s
.67341 .62896 m .69219 .56825 L .64379 .55341 L closepath p 1 g
F P
0 g
s
.64452 .83481 m .69847 .80838 L .69316 .79316 L closepath p 1 g
F P
0 g
s
.64489 .84515 m .69847 .80838 L .64452 .83481 L closepath p 1 g
F P
0 g
s
.69845 .82367 m .69847 .80838 L .64489 .84515 L closepath p 1 g
F P
0 g
s
.70581 .35871 m .69423 .32443 L .64524 .30723 L closepath p 1 g
F P
0 g
s
.65353 .34064 m .70581 .35871 L .64524 .30723 L closepath p 1 g
F P
0 g
s
.64621 .7529 m .68113 .70143 L .6495 .68803 L closepath p 1 g
F P
0 g
s
.66674 .76493 m .68113 .70143 L .64621 .7529 L closepath p 1 g
F P
0 g
s
.68113 .70143 m .67341 .62896 L .6495 .68803 L closepath p 1 g
F P
0 g
s
.68268 .86315 m .70625 .84863 L .65034 .87691 L closepath p 1 g
F P
0 g
s
.65034 .87691 m .70625 .84863 L .66872 .86554 L closepath p 1 g
F P
0 g
s
.653 .49194 m .70508 .50851 L .6577 .43453 L closepath p 1 g
F P
0 g
s
.69219 .56825 m .70508 .50851 L .653 .49194 L closepath p 1 g
F P
0 g
s
.71179 .40174 m .70581 .35871 L .65353 .34064 L closepath p 1 g
F P
0 g
s
.65781 .38349 m .71179 .40174 L .65353 .34064 L closepath p 1 g
F P
0 g
s
.65735 .28463 m .7163 .32018 L .69089 .30256 L closepath p 1 g
F P
0 g
s
.67779 .29965 m .7163 .32018 L .65735 .28463 L closepath p 1 g
F P
0 g
s
.6577 .43453 m .71163 .45227 L .65781 .38349 L closepath p 1 g
F P
0 g
s
.70508 .50851 m .71163 .45227 L .6577 .43453 L closepath p 1 g
F P
0 g
s
.71163 .45227 m .71179 .40174 L .65781 .38349 L closepath p 1 g
F P
0 g
s
.70721 .71745 m .68113 .70143 L .66674 .76493 L closepath p 1 g
F P
0 g
s
.68252 .77851 m .70721 .71745 L .66674 .76493 L closepath p 1 g
F P
0 g
s
.70625 .84863 m .72503 .83211 L .66872 .86554 L closepath p 1 g
F P
0 g
s
.66872 .86554 m .72503 .83211 L .6832 .85265 L closepath p 1 g
F P
0 g
s
.71086 .64554 m .69219 .56825 L .67341 .62896 L closepath p 1 g
F P
0 g
s
.71086 .64554 m .67341 .62896 L .68113 .70143 L closepath p 1 g
F P
0 g
s
.67779 .29965 m .73668 .34688 L .7163 .32018 L closepath p 1 g
F P
0 g
s
.69423 .32443 m .73668 .34688 L .67779 .29965 L closepath p 1 g
F P
0 g
s
.70721 .71745 m .71086 .64554 L .68113 .70143 L closepath p 1 g
F P
0 g
s
.68252 .77851 m .72701 .73546 L .70721 .71745 L closepath p 1 g
F P
0 g
s
.68252 .77851 m .69316 .79316 L .72701 .73546 L closepath p 1 g
F P
0 g
s
.6832 .85265 m .73833 .81396 L .69325 .83857 L closepath p 1 g
F P
0 g
s
.72503 .83211 m .73833 .81396 L .6832 .85265 L closepath p 1 g
F P
0 g
s
.69219 .56825 m .73416 .58761 L .70508 .50851 L closepath p 1 g
F P
0 g
s
.71086 .64554 m .73416 .58761 L .69219 .56825 L closepath p 1 g
F P
0 g
s
.69316 .79316 m .74007 .75478 L .72701 .73546 L closepath p 1 g
F P
0 g
s
.69847 .80838 m .74007 .75478 L .69316 .79316 L closepath p 1 g
F P
0 g
s
.69325 .83857 m .74555 .79466 L .69845 .82367 L closepath p 1 g
F P
0 g
s
.73833 .81396 m .74555 .79466 L .69325 .83857 L closepath p 1 g
F P
0 g
s
.69423 .32443 m .751 .38226 L .73668 .34688 L closepath p 1 g
F P
0 g
s
.70581 .35871 m .751 .38226 L .69423 .32443 L closepath p 1 g
F P
0 g
s
.74555 .79466 m .74623 .77473 L .69845 .82367 L closepath p 1 g
F P
0 g
s
.69845 .82367 m .74623 .77473 L .69847 .80838 L closepath p 1 g
F P
0 g
s
.69847 .80838 m .74623 .77473 L .74007 .75478 L closepath p 1 g
F P
0 g
s
.70508 .50851 m .7501 .5301 L .71163 .45227 L closepath p 1 g
F P
0 g
s
.73416 .58761 m .7501 .5301 L .70508 .50851 L closepath p 1 g
F P
0 g
s
.75839 .42552 m .751 .38226 L .70581 .35871 L closepath p 1 g
F P
0 g
s
.71179 .40174 m .75839 .42552 L .70581 .35871 L closepath p 1 g
F P
0 g
s
.73696 .8241 m .75941 .80437 L .70625 .84863 L closepath p 1 g
F P
0 g
s
.70625 .84863 m .75941 .80437 L .72503 .83211 L closepath p 1 g
F P
0 g
s
.70721 .71745 m .74153 .66529 L .71086 .64554 L closepath p 1 g
F P
0 g
s
.72701 .73546 m .74153 .66529 L .70721 .71745 L closepath p 1 g
F P
0 g
s
.74153 .66529 m .73416 .58761 L .71086 .64554 L closepath p 1 g
F P
0 g
s
.71163 .45227 m .7582 .47537 L .71179 .40174 L closepath p 1 g
F P
0 g
s
.71163 .45227 m .7501 .5301 L .7582 .47537 L closepath p 1 g
F P
0 g
s
.7582 .47537 m .75839 .42552 L .71179 .40174 L closepath p 1 g
F P
0 g
s
.7163 .32018 m .77121 .37355 L .74778 .34464 L closepath p 1 g
F P
0 g
s
.7163 .32018 m .73668 .34688 L .77121 .37355 L closepath p 1 g
F P
0 g
s
.75941 .80437 m .77556 .78262 L .72503 .83211 L closepath p 1 g
F P
0 g
s
.72503 .83211 m .77556 .78262 L .73833 .81396 L closepath p 1 g
F P
0 g
s
.72701 .73546 m .76455 .6874 L .74153 .66529 L closepath p 1 g
F P
0 g
s
.72701 .73546 m .74007 .75478 L .76455 .6874 L closepath p 1 g
F P
0 g
s
.76832 .61063 m .7501 .5301 L .73416 .58761 L closepath p 1 g
F P
0 g
s
.74153 .66529 m .76832 .61063 L .73416 .58761 L closepath p 1 g
F P
0 g
s
.73668 .34688 m .78763 .41019 L .77121 .37355 L closepath p 1 g
F P
0 g
s
.751 .38226 m .78763 .41019 L .73668 .34688 L closepath p 1 g
F P
0 g
s
.73833 .81396 m .78467 .75938 L .74555 .79466 L closepath p 1 g
F P
0 g
s
.77556 .78262 m .78467 .75938 L .73833 .81396 L closepath p 1 g
F P
0 g
s
.74007 .75478 m .77946 .71101 L .76455 .6874 L closepath p 1 g
F P
0 g
s
.74623 .77473 m .77946 .71101 L .74007 .75478 L closepath p 1 g
F P
0 g
s
.76455 .6874 m .76832 .61063 L .74153 .66529 L closepath p 1 g
F P
0 g
s
.74555 .79466 m .7861 .73527 L .74623 .77473 L closepath p 1 g
F P
0 g
s
.78467 .75938 m .7861 .73527 L .74555 .79466 L closepath p 1 g
F P
0 g
s
.74623 .77473 m .7861 .73527 L .77946 .71101 L closepath p 1 g
F P
0 g
s
.79688 .40328 m .77138 .37199 L .74778 .34464 L closepath p 1 g
F P
0 g
s
.77121 .37355 m .79688 .40328 L .74778 .34464 L closepath p 1 g
F P
0 g
s
.7501 .5301 m .7866 .55571 L .7582 .47537 L closepath p 1 g
F P
0 g
s
.76832 .61063 m .7866 .55571 L .7501 .5301 L closepath p 1 g
F P
0 g
s
.75839 .42552 m .7961 .45368 L .751 .38226 L closepath p 1 g
F P
0 g
s
.751 .38226 m .7961 .45368 L .78763 .41019 L closepath p 1 g
F P
0 g
s
.7582 .47537 m .79588 .50274 L .75839 .42552 L closepath p 1 g
F P
0 g
s
.7866 .55571 m .79588 .50274 L .7582 .47537 L closepath p 1 g
F P
0 g
s
.79588 .50274 m .7961 .45368 L .75839 .42552 L closepath p 1 g
F P
0 g
s
.78559 .77067 m .80402 .74587 L .75941 .80437 L closepath p 1 g
F P
0 g
s
.75941 .80437 m .80402 .74587 L .77556 .78262 L closepath p 1 g
F P
0 g
s
.76455 .6874 m .79374 .6363 L .76832 .61063 L closepath p 1 g
F P
0 g
s
.76455 .6874 m .77946 .71101 L .79374 .6363 L closepath p 1 g
F P
0 g
s
.79374 .6363 m .7866 .55571 L .76832 .61063 L closepath p 1 g
F P
0 g
s
.81473 .44127 m .79688 .40328 L .77121 .37355 L closepath p 1 g
F P
0 g
s
.78763 .41019 m .81473 .44127 L .77121 .37355 L closepath p 1 g
F P
0 g
s
.77556 .78262 m .81469 .71927 L .78467 .75938 L closepath p 1 g
F P
0 g
s
.80402 .74587 m .81469 .71927 L .77556 .78262 L closepath p 1 g
F P
0 g
s
.77946 .71101 m .80994 .66362 L .79374 .6363 L closepath p 1 g
F P
0 g
s
.7861 .73527 m .80994 .66362 L .77946 .71101 L closepath p 1 g
F P
0 g
s
.81469 .71927 m .81684 .69158 L .78467 .75938 L closepath p 1 g
F P
0 g
s
.78467 .75938 m .81684 .69158 L .7861 .73527 L closepath p 1 g
F P
0 g
s
.7861 .73527 m .81684 .69158 L .80994 .66362 L closepath p 1 g
F P
0 g
s
.8136 .58422 m .79588 .50274 L .7866 .55571 L closepath p 1 g
F P
0 g
s
.79374 .6363 m .8136 .58422 L .7866 .55571 L closepath p 1 g
F P
0 g
s
.8239 .48499 m .81473 .44127 L .78763 .41019 L closepath p 1 g
F P
0 g
s
.7961 .45368 m .8239 .48499 L .78763 .41019 L closepath p 1 g
F P
0 g
s
.79374 .6363 m .80994 .66362 L .8136 .58422 L closepath p 1 g
F P
0 g
s
.82366 .53317 m .7961 .45368 L .79588 .50274 L closepath p 1 g
F P
0 g
s
.8136 .58422 m .82366 .53317 L .79588 .50274 L closepath p 1 g
F P
0 g
s
.82366 .53317 m .8239 .48499 L .7961 .45368 L closepath p 1 g
F P
0 g
s
.83179 .47425 m .81322 .43491 L .79688 .40328 L closepath p 1 g
F P
0 g
s
.81473 .44127 m .83179 .47425 L .79688 .40328 L closepath p 1 g
F P
0 g
s
.80402 .74587 m .83493 .67588 L .81469 .71927 L closepath p 1 g
F P
0 g
s
.82315 .70512 m .83493 .67588 L .80402 .74587 L closepath p 1 g
F P
0 g
s
.80994 .66362 m .83063 .61447 L .8136 .58422 L closepath p 1 g
F P
0 g
s
.81684 .69158 m .83063 .61447 L .80994 .66362 L closepath p 1 g
F P
0 g
s
.8136 .58422 m .83063 .61447 L .82366 .53317 L closepath p 1 g
F P
0 g
s
.81469 .71927 m .83762 .64537 L .81684 .69158 L closepath p 1 g
F P
0 g
s
.83493 .67588 m .83762 .64537 L .81469 .71927 L closepath p 1 g
F P
0 g
s
.84133 .51817 m .83179 .47425 L .81473 .44127 L closepath p 1 g
F P
0 g
s
.8239 .48499 m .84133 .51817 L .81473 .44127 L closepath p 1 g
F P
0 g
s
.81684 .69158 m .83762 .64537 L .83063 .61447 L closepath p 1 g
F P
0 g
s
.82366 .53317 m .84108 .56542 L .8239 .48499 L closepath p 1 g
F P
0 g
s
.83063 .61447 m .84108 .56542 L .82366 .53317 L closepath p 1 g
F P
0 g
s
.84108 .56542 m .84133 .51817 L .8239 .48499 L closepath p 1 g
F P
0 g
s
.83762 .64537 m .84108 .56542 L .83063 .61447 L closepath p 1 g
F P
0 g
s
.83493 .67588 m .84811 .59831 L .83762 .64537 L closepath p 1 g
F P
0 g
s
.84512 .63076 m .84811 .59831 L .83493 .67588 L closepath p 1 g
F P
0 g
s
.84811 .59831 m .84108 .56542 L .83762 .64537 L closepath p 1 g
F P
0 g
s
.84811 .59831 m .84133 .51817 L .84108 .56542 L closepath p 1 g
F P
0 g
s
.84836 .55202 m .84133 .51817 L .84811 .59831 L closepath p 1 g
F P
0 g
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.5 0.539506 0.97619 1.05332 MathSubStart
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0553515 1.1633 1.11022e-16 1.1633 [
[.5 1.09473 -41 0 ]
[.5 1.09473 41 9.8125 ]
[ 0 0 0 0 ]
[ 1 1.08223 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
gsave
.5 1.09473 -102 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Polygons) show
117.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(:) show
123.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(628) show
141.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 1.08223 L
0 1.08223 L
closepath
clip
newpath
.5 Mabswid
[ ] 0 setdash
.20362 .57448 m .20387 .58747 L .20365 .54052 L closepath p 1 g
F P
0 g
s
.20384 .62047 m .20387 .58747 L .20362 .57448 L closepath p 1 g
F P
0 g
s
.20365 .54052 m .23477 .47335 L .22351 .46223 L closepath p 1 g
F P
0 g
s
.21416 .55433 m .23477 .47335 L .20365 .54052 L closepath p 1 g
F P
0 g
s
.20387 .58747 m .21416 .55433 L .20365 .54052 L closepath p 1 g
F P
0 g
s
.20384 .62047 m .21445 .63552 L .20387 .58747 L closepath p 1 g
F P
0 g
s
.21421 .66655 m .21445 .63552 L .20384 .62047 L closepath p 1 g
F P
0 g
s
.20387 .58747 m .2246 .60442 L .21416 .55433 L closepath p 1 g
F P
0 g
s
.21445 .63552 m .2246 .60442 L .20387 .58747 L closepath p 1 g
F P
0 g
s
.21416 .55433 m .26617 .49227 L .23477 .47335 L closepath p 1 g
F P
0 g
s
.2449 .57428 m .26617 .49227 L .21416 .55433 L closepath p 1 g
F P
0 g
s
.2246 .60442 m .2449 .57428 L .21416 .55433 L closepath p 1 g
F P
0 g
s
.21421 .66655 m .24527 .6549 L .21445 .63552 L closepath p 1 g
F P
0 g
s
.23484 .71116 m .24527 .6549 L .21421 .66655 L closepath p 1 g
F P
0 g
s
.22335 .69654 m .23484 .71116 L .21421 .66655 L closepath p 1 g
F P
0 g
s
.21445 .63552 m .24527 .6549 L .2246 .60442 L closepath p 1 g
F P
0 g
s
.24107 .72468 m .23484 .71116 L .22335 .69654 L closepath p 1 g
F P
0 g
s
.22351 .46223 m .27423 .39882 L .26151 .39156 L closepath p 1 g
F P
0 g
s
.23477 .47335 m .27423 .39882 L .22351 .46223 L closepath p 1 g
F P
0 g
s
.2246 .60442 m .29297 .60251 L .2449 .57428 L closepath p 1 g
F P
0 g
s
.24527 .6549 m .29297 .60251 L .2246 .60442 L closepath p 1 g
F P
0 g
s
.23477 .47335 m .30692 .41534 L .27423 .39882 L closepath p 1 g
F P
0 g
s
.26617 .49227 m .30692 .41534 L .23477 .47335 L closepath p 1 g
F P
0 g
s
.26645 .72825 m .24527 .6549 L .23484 .71116 L closepath p 1 g
F P
0 g
s
.25928 .76407 m .30523 .78958 L .23484 .71116 L closepath p 1 g
F P
0 g
s
.23484 .71116 m .30523 .78958 L .26645 .72825 L closepath p 1 g
F P
0 g
s
.25928 .76407 m .23484 .71116 L .24107 .72468 L closepath p 1 g
F P
0 g
s
.2449 .57428 m .31481 .52153 L .26617 .49227 L closepath p 1 g
F P
0 g
s
.29297 .60251 m .31481 .52153 L .2449 .57428 L closepath p 1 g
F P
0 g
s
.29332 .68042 m .29297 .60251 L .24527 .6549 L closepath p 1 g
F P
0 g
s
.26645 .72825 m .29332 .68042 L .24527 .6549 L closepath p 1 g
F P
0 g
s
.3072 .81954 m .30523 .78958 L .25928 .76407 L closepath p 1 g
F P
0 g
s
.26151 .39156 m .32766 .3366 L .31303 .334 L closepath p 1 g
F P
0 g
s
.27423 .39882 m .32766 .3366 L .26151 .39156 L closepath p 1 g
F P
0 g
s
.26617 .49227 m .35658 .44404 L .30692 .41534 L closepath p 1 g
F P
0 g
s
.31481 .52153 m .35658 .44404 L .26617 .49227 L closepath p 1 g
F P
0 g
s
.26645 .72825 m .35235 .7126 L .29332 .68042 L closepath p 1 g
F P
0 g
s
.30523 .78958 m .35235 .7126 L .26645 .72825 L closepath p 1 g
F P
0 g
s
.27423 .39882 m .36199 .34964 L .32766 .3366 L closepath p 1 g
F P
0 g
s
.30692 .41534 m .36199 .34964 L .27423 .39882 L closepath p 1 g
F P
0 g
s
.29297 .60251 m .37455 .56206 L .31481 .52153 L closepath p 1 g
F P
0 g
s
.3522 .6397 m .37455 .56206 L .29297 .60251 L closepath p 1 g
F P
0 g
s
.29332 .68042 m .3522 .6397 L .29297 .60251 L closepath p 1 g
F P
0 g
s
.35235 .7126 m .3522 .6397 L .29332 .68042 L closepath p 1 g
F P
0 g
s
.35473 .80483 m .35235 .7126 L .30523 .78958 L closepath p 1 g
F P
0 g
s
.3072 .81954 m .32209 .82803 L .30523 .78958 L closepath p 1 g
F P
0 g
s
.32209 .82803 m .40743 .8435 L .30523 .78958 L closepath p 1 g
F P
0 g
s
.30523 .78958 m .40743 .8435 L .35473 .80483 L closepath p 1 g
F P
0 g
s
.30692 .41534 m .41285 .3764 L .36199 .34964 L closepath p 1 g
F P
0 g
s
.35658 .44404 m .41285 .3764 L .30692 .41534 L closepath p 1 g
F P
0 g
s
.3072 .81954 m .33374 .83775 L .32209 .82803 L closepath p 1 g
F P
0 g
s
.32766 .3366 m .38788 .29108 L .31303 .334 L closepath p 1 g
F P
0 g
s
.31303 .334 m .38788 .29108 L .34147 .31424 L closepath p 1 g
F P
0 g
s
.31481 .52153 m .41704 .48618 L .35658 .44404 L closepath p 1 g
F P
0 g
s
.37455 .56206 m .41704 .48618 L .31481 .52153 L closepath p 1 g
F P
0 g
s
.32209 .82803 m .33374 .83775 L .37867 .86339 L closepath p 1 g
F P
0 g
s
.37867 .86339 m .40743 .8435 L .32209 .82803 L closepath p 1 g
F P
0 g
s
.32766 .3366 m .42381 .29989 L .38788 .29108 L closepath p 1 g
F P
0 g
s
.36199 .34964 m .42381 .29989 L .32766 .3366 L closepath p 1 g
F P
0 g
s
.3522 .6397 m .4852 .60589 L .37455 .56206 L closepath p 1 g
F P
0 g
s
.41398 .68459 m .4852 .60589 L .3522 .6397 L closepath p 1 g
F P
0 g
s
.35235 .7126 m .41398 .68459 L .3522 .6397 L closepath p 1 g
F P
0 g
s
.41371 .75027 m .41398 .68459 L .35235 .7126 L closepath p 1 g
F P
0 g
s
.35235 .7126 m .45033 .79681 L .41371 .75027 L closepath p 1 g
F P
0 g
s
.35473 .80483 m .45033 .79681 L .35235 .7126 L closepath p 1 g
F P
0 g
s
.42295 .8132 m .45033 .79681 L .35473 .80483 L closepath p 1 g
F P
0 g
s
.35473 .80483 m .4701 .84786 L .42295 .8132 L closepath p 1 g
F P
0 g
s
.40743 .8435 m .4701 .84786 L .35473 .80483 L closepath p 1 g
F P
0 g
s
.35658 .44404 m .47402 .41847 L .41285 .3764 L closepath p 1 g
F P
0 g
s
.41704 .48618 m .47402 .41847 L .35658 .44404 L closepath p 1 g
F P
0 g
s
.36199 .34964 m .47571 .3236 L .42381 .29989 L closepath p 1 g
F P
0 g
s
.41285 .3764 m .47571 .3236 L .36199 .34964 L closepath p 1 g
F P
0 g
s
.37315 .86024 m .38838 .87358 L .37867 .86339 L p 1 g
F P
0 g
s
.37455 .56206 m .47972 .5407 L .41704 .48618 L closepath p 1 g
F P
0 g
s
.4852 .60589 m .47972 .5407 L .37455 .56206 L closepath p 1 g
F P
0 g
s
.37867 .86339 m .41517 .86374 L .40743 .8435 L closepath p 1 g
F P
0 g
s
.37867 .86339 m .38838 .87358 L .39605 .87536 L closepath p 1 g
F P
0 g
s
.37867 .86339 m .39605 .87536 L .41517 .86374 L closepath p 1 g
F P
0 g
s
.38788 .29108 m .48386 .26849 L .44673 .26439 L closepath p 1 g
F P
0 g
s
.42381 .29989 m .48386 .26849 L .38788 .29108 L closepath p 1 g
F P
0 g
s
.39605 .87536 m .38838 .87358 L .41696 .88561 L closepath p 1 g
F P
0 g
s
.41517 .86374 m .39605 .87536 L .47419 .8732 L closepath p 1 g
F P
0 g
s
.39605 .87536 m .41696 .88561 L .47419 .8732 L closepath p 1 g
F P
0 g
s
.41517 .86374 m .46712 .86342 L .40743 .8435 L closepath p 1 g
F P
0 g
s
.40743 .8435 m .46712 .86342 L .46626 .8582 L closepath p 1 g
F P
0 g
s
.40743 .8435 m .46626 .8582 L .46726 .85296 L closepath p 1 g
F P
0 g
s
.40743 .8435 m .46726 .85296 L .4701 .84786 L closepath p 1 g
F P
0 g
s
.41285 .3764 m .53732 .36407 L .47571 .3236 L closepath p 1 g
F P
0 g
s
.47402 .41847 m .53732 .36407 L .41285 .3764 L closepath p 1 g
F P
0 g
s
.41371 .75027 m .50905 .78354 L .41398 .68459 L closepath p 1 g
F P
0 g
s
.41371 .75027 m .45033 .79681 L .48791 .78617 L closepath p 1 g
F P
0 g
s
.48791 .78617 m .50905 .78354 L .41371 .75027 L closepath p 1 g
F P
0 g
s
.49214 .66994 m .4852 .60589 L .41398 .68459 L closepath p 1 g
F P
0 g
s
.50905 .78354 m .49214 .66994 L .41398 .68459 L closepath p 1 g
F P
0 g
s
.41517 .86374 m .4698 .86847 L .46712 .86342 L closepath p 1 g
F P
0 g
s
.41517 .86374 m .47419 .8732 L .4698 .86847 L closepath p 1 g
F P
0 g
s
.47419 .8732 m .41696 .88561 L .45029 .88837 L closepath p 1 g
F P
0 g
s
.41696 .88561 m .46683 .90007 L .45029 .88837 L closepath p 1 g
F P
0 g
s
.41704 .48618 m .53698 .47499 L .47402 .41847 L closepath p 1 g
F P
0 g
s
.47972 .5407 m .53698 .47499 L .41704 .48618 L closepath p 1 g
F P
0 g
s
.42295 .8132 m .481 .83875 L .45033 .79681 L closepath p 1 g
F P
0 g
s
.42295 .8132 m .4701 .84786 L .47472 .84308 L closepath p 1 g
F P
0 g
s
.42295 .8132 m .47472 .84308 L .481 .83875 L closepath p 1 g
F P
0 g
s
.42381 .29989 m .53632 .28829 L .48386 .26849 L closepath p 1 g
F P
0 g
s
.47571 .3236 m .53632 .28829 L .42381 .29989 L closepath p 1 g
F P
0 g
s
.48386 .26849 m .53427 .25521 L .44673 .26439 L closepath p 1 g
F P
0 g
s
.44673 .26439 m .53427 .25521 L .49657 .25615 L closepath p 1 g
F P
0 g
s
.47419 .8732 m .45029 .88837 L .48014 .87748 L closepath p 1 g
F P
0 g
s
.48014 .87748 m .45029 .88837 L .48747 .88118 L closepath p 1 g
F P
0 g
s
.48747 .88118 m .45029 .88837 L .49596 .8842 L closepath p 1 g
F P
0 g
s
.45029 .88837 m .46683 .90007 L .49596 .8842 L closepath p 1 g
F P
0 g
s
.45033 .79681 m .58841 .85983 L .48791 .78617 L closepath p 1 g
F P
0 g
s
.481 .83875 m .58841 .85983 L .45033 .79681 L closepath p 1 g
F P
0 g
s
.46626 .8582 m .58841 .85983 L .46726 .85296 L closepath p 1 g
F P
0 g
s
.46712 .86342 m .58841 .85983 L .46626 .8582 L closepath p 1 g
F P
0 g
s
.46683 .90007 m .49442 .90396 L .49596 .8842 L closepath p 1 g
F P
0 g
s
.4698 .86847 m .58841 .85983 L .46712 .86342 L closepath p 1 g
F P
0 g
s
.46726 .85296 m .58841 .85983 L .4701 .84786 L closepath p 1 g
F P
0 g
s
.47419 .8732 m .58841 .85983 L .4698 .86847 L closepath p 1 g
F P
0 g
s
.4701 .84786 m .58841 .85983 L .47472 .84308 L closepath p 1 g
F P
0 g
s
.47402 .41847 m .60021 .42069 L .53732 .36407 L closepath p 1 g
F P
0 g
s
.59762 .47938 m .60021 .42069 L .47402 .41847 L closepath p 1 g
F P
0 g
s
.53698 .47499 m .59762 .47938 L .47402 .41847 L closepath p 1 g
F P
0 g
s
.48014 .87748 m .58841 .85983 L .47419 .8732 L closepath p 1 g
F P
0 g
s
.47472 .84308 m .58841 .85983 L .481 .83875 L closepath p 1 g
F P
0 g
s
.47571 .3236 m .59783 .32586 L .53632 .28829 L closepath p 1 g
F P
0 g
s
.53732 .36407 m .59783 .32586 L .47571 .3236 L closepath p 1 g
F P
0 g
s
.47972 .5407 m .5925 .54256 L .53698 .47499 L closepath p 1 g
F P
0 g
s
.53528 .60414 m .5925 .54256 L .47972 .5407 L closepath p 1 g
F P
0 g
s
.4852 .60589 m .53528 .60414 L .47972 .5407 L closepath p 1 g
F P
0 g
s
.48747 .88118 m .58841 .85983 L .48014 .87748 L closepath p 1 g
F P
0 g
s
.48386 .26849 m .58662 .27045 L .53427 .25521 L closepath p 1 g
F P
0 g
s
.53632 .28829 m .58662 .27045 L .48386 .26849 L closepath p 1 g
F P
0 g
s
.53398 .66853 m .53528 .60414 L .4852 .60589 L closepath p 1 g
F P
0 g
s
.49214 .66994 m .53398 .66853 L .4852 .60589 L closepath p 1 g
F P
0 g
s
.49596 .8842 m .58841 .85983 L .48747 .88118 L closepath p 1 g
F P
0 g
s
.48791 .78617 m .57326 .78738 L .50905 .78354 L closepath p 1 g
F P
0 g
s
.48791 .78617 m .58841 .85983 L .57326 .78738 L closepath p 1 g
F P
0 g
s
.50905 .78354 m .53398 .66853 L .49214 .66994 L closepath p 1 g
F P
0 g
s
.49596 .8842 m .49442 .90396 L .50428 .90121 L closepath p 1 g
F P
0 g
s
.49442 .90396 m .55138 .9047 L .50428 .90121 L closepath p 1 g
F P
0 g
s
.50535 .88646 m .58841 .85983 L .49596 .8842 L closepath p 1 g
F P
0 g
s
.49596 .8842 m .50428 .90121 L .50535 .88646 L closepath p 1 g
F P
0 g
s
.50535 .88646 m .50428 .90121 L .51536 .88789 L closepath p 1 g
F P
0 g
s
.51536 .88789 m .50428 .90121 L .52573 .88846 L closepath p 1 g
F P
0 g
s
.50428 .90121 m .55138 .9047 L .52573 .88846 L closepath p 1 g
F P
0 g
s
.51536 .88789 m .58841 .85983 L .50535 .88646 L closepath p 1 g
F P
0 g
s
.57326 .78738 m .53398 .66853 L .50905 .78354 L closepath p 1 g
F P
0 g
s
.52573 .88846 m .58841 .85983 L .51536 .88789 L closepath p 1 g
F P
0 g
s
.56377 .8995 m .58841 .85983 L .52573 .88846 L closepath p 1 g
F P
0 g
s
.52573 .88846 m .54957 .90356 L p .56377 .8995 L 1 g
F P
0 g
s
.53398 .66853 m .61534 .67785 L .53528 .60414 L closepath p 1 g
F P
0 g
s
.59527 .73587 m .61534 .67785 L .53398 .66853 L closepath p 1 g
F P
0 g
s
.57326 .78738 m .59527 .73587 L .53398 .66853 L closepath p 1 g
F P
0 g
s
.53427 .25521 m .58662 .27045 L .62075 .26772 L closepath p 1 g
F P
0 g
s
.53528 .60414 m .63253 .61574 L .5925 .54256 L closepath p 1 g
F P
0 g
s
.61534 .67785 m .63253 .61574 L .53528 .60414 L closepath p 1 g
F P
0 g
s
.53632 .28829 m .64737 .30408 L .58662 .27045 L closepath p 1 g
F P
0 g
s
.59783 .32586 m .64737 .30408 L .53632 .28829 L closepath p 1 g
F P
0 g
s
.5925 .54256 m .59762 .47938 L .53698 .47499 L closepath p 1 g
F P
0 g
s
.53732 .36407 m .66014 .38086 L .59783 .32586 L closepath p 1 g
F P
0 g
s
.60021 .42069 m .66014 .38086 L .53732 .36407 L closepath p 1 g
F P
0 g
s
.56445 .89959 m .56377 .8995 L .54957 .90356 L 1 g
F
.54957 .90356 m .55138 .9047 L p .56445 .89959 L F P
0 g
s
.55138 .9047 m .57931 .90153 L p .56445 .89959 L 1 g
F P
0 g
s
.56377 .8995 m .65243 .87875 L .58841 .85983 L closepath p 1 g
F P
0 g
s
.56377 .8995 m .57931 .90153 L .62948 .88412 L p 1 g
F P
0 g
s
.57326 .78738 m .58841 .85983 L .59527 .73587 L closepath p 1 g
F P
0 g
s
.58662 .27045 m .69676 .32148 L .62075 .26772 L closepath p 1 g
F P
0 g
s
.64737 .30408 m .69676 .32148 L .58662 .27045 L closepath p 1 g
F P
0 g
s
.59527 .73587 m .58841 .85983 L .62356 .74355 L closepath p 1 g
F P
0 g
s
.58841 .85983 m .7011 .80965 L .62356 .74355 L closepath p 1 g
F P
0 g
s
.70104 .82507 m .7011 .80965 L .58841 .85983 L closepath p 1 g
F P
0 g
s
.69577 .84009 m .70104 .82507 L .58841 .85983 L closepath p 1 g
F P
0 g
s
.67099 .86729 m .69577 .84009 L .58841 .85983 L closepath p 1 g
F P
0 g
s
.65243 .87875 m .67099 .86729 L .58841 .85983 L closepath p 1 g
F P
0 g
s
.5925 .54256 m .64607 .55226 L .59762 .47938 L closepath p 1 g
F P
0 g
s
.63253 .61574 m .64607 .55226 L .5925 .54256 L closepath p 1 g
F P
0 g
s
.62356 .74355 m .61534 .67785 L .59527 .73587 L closepath p 1 g
F P
0 g
s
.59762 .47938 m .65535 .49022 L .60021 .42069 L closepath p 1 g
F P
0 g
s
.64607 .55226 m .65535 .49022 L .59762 .47938 L closepath p 1 g
F P
0 g
s
.59783 .32586 m .7085 .35597 L .64737 .30408 L closepath p 1 g
F P
0 g
s
.66014 .38086 m .7085 .35597 L .59783 .32586 L closepath p 1 g
F P
0 g
s
.60021 .42069 m .71447 .45026 L .66014 .38086 L closepath p 1 g
F P
0 g
s
.70789 .50698 m .71447 .45026 L .60021 .42069 L closepath p 1 g
F P
0 g
s
.65535 .49022 m .70789 .50698 L .60021 .42069 L closepath p 1 g
F P
0 g
s
.68371 .7017 m .63253 .61574 L .61534 .67785 L closepath p 1 g
F P
0 g
s
.62356 .74355 m .68371 .7017 L .61534 .67785 L closepath p 1 g
F P
0 g
s
.62075 .26772 m .69676 .32148 L .71893 .31734 L closepath p 1 g
F P
0 g
s
.62356 .74355 m .7011 .80965 L .68371 .7017 L closepath p 1 g
F P
0 g
s
.71372 .64529 m .64607 .55226 L .63253 .61574 L closepath p 1 g
F P
0 g
s
.68371 .7017 m .71372 .64529 L .63253 .61574 L closepath p 1 g
F P
0 g
s
.64607 .55226 m .73722 .58683 L .65535 .49022 L closepath p 1 g
F P
0 g
s
.71372 .64529 m .73722 .58683 L .64607 .55226 L closepath p 1 g
F P
0 g
s
.7085 .35597 m .69676 .32148 L .64737 .30408 L closepath p 1 g
F P
0 g
s
.68499 .86486 m .67099 .86729 L .65243 .87875 L closepath p 1 g
F P
0 g
s
.65535 .49022 m .73722 .58683 L .70789 .50698 L closepath p 1 g
F P
0 g
s
.66014 .38086 m .76157 .42336 L .7085 .35597 L closepath p 1 g
F P
0 g
s
.71447 .45026 m .76157 .42336 L .66014 .38086 L closepath p 1 g
F P
0 g
s
.67099 .86729 m .7412 .81528 L .69577 .84009 L closepath p 1 g
F P
0 g
s
.67099 .86729 m .68167 .86543 L p .7412 .81528 L 1 g
F P
0 g
s
.68167 .86543 m .68499 .86486 L .77871 .78367 L p 1 g
F P
0 g
s
.72995 .73607 m .71372 .64529 L .68371 .7017 L closepath p 1 g
F P
0 g
s
.68371 .7017 m .7011 .80965 L .72995 .73607 L closepath p 1 g
F P
0 g
s
.76238 .8056 m .77871 .78367 L .68499 .86486 L closepath p 1 g
F P
0 g
s
.69577 .84009 m .7412 .81528 L .70104 .82507 L closepath p 1 g
F P
0 g
s
.69676 .32148 m .77432 .37113 L .71893 .31734 L closepath p 1 g
F P
0 g
s
.7085 .35597 m .75405 .37979 L .69676 .32148 L closepath p 1 g
F P
0 g
s
.69676 .32148 m .75405 .37979 L .77432 .37113 L closepath p 1 g
F P
0 g
s
.70104 .82507 m .74926 .7757 L .7011 .80965 L closepath p 1 g
F P
0 g
s
.7412 .81528 m .74926 .7757 L .70104 .82507 L closepath p 1 g
F P
0 g
s
.7011 .80965 m .74926 .7757 L .72995 .73607 L closepath p 1 g
F P
0 g
s
.70789 .50698 m .75329 .52881 L .71447 .45026 L closepath p 1 g
F P
0 g
s
.73722 .58683 m .75329 .52881 L .70789 .50698 L closepath p 1 g
F P
0 g
s
.76157 .42336 m .75405 .37979 L .7085 .35597 L closepath p 1 g
F P
0 g
s
.76782 .68758 m .73722 .58683 L .71372 .64529 L closepath p 1 g
F P
0 g
s
.72995 .73607 m .76782 .68758 L .71372 .64529 L closepath p 1 g
F P
0 g
s
.71447 .45026 m .79938 .50128 L .76157 .42336 L closepath p 1 g
F P
0 g
s
.75329 .52881 m .79938 .50128 L .71447 .45026 L closepath p 1 g
F P
0 g
s
.77432 .37113 m .77439 .36969 L .71893 .31734 L closepath p 1 g
F P
0 g
s
.72995 .73607 m .74926 .7757 L .76782 .68758 L closepath p 1 g
F P
0 g
s
.79724 .63602 m .75329 .52881 L .73722 .58683 L closepath p 1 g
F P
0 g
s
.76782 .68758 m .79724 .63602 L .73722 .58683 L closepath p 1 g
F P
0 g
s
.7412 .81528 m .78945 .73589 L .74926 .7757 L closepath p 1 g
F P
0 g
s
.77871 .78367 m .78945 .73589 L .7412 .81528 L closepath p 1 g
F P
0 g
s
.74926 .7757 m .78945 .73589 L .76782 .68758 L closepath p 1 g
F P
0 g
s
.81725 .5835 m .79938 .50128 L .75329 .52881 L closepath p 1 g
F P
0 g
s
.79724 .63602 m .81725 .5835 L .75329 .52881 L closepath p 1 g
F P
0 g
s
.75405 .37979 m .81822 .43942 L .77432 .37113 L closepath p 1 g
F P
0 g
s
.79955 .45183 m .81822 .43942 L .75405 .37979 L closepath p 1 g
F P
0 g
s
.76157 .42336 m .79955 .45183 L .75405 .37979 L closepath p 1 g
F P
0 g
s
.79938 .50128 m .79955 .45183 L .76157 .42336 L closepath p 1 g
F P
0 g
s
.76238 .8056 m .80736 .74661 L .77871 .78367 L closepath p 1 g
F P
0 g
s
.76782 .68758 m .78945 .73589 L .79724 .63602 L closepath p 1 g
F P
0 g
s
.81822 .43942 m .77439 .36969 L .77432 .37113 L closepath p 1 g
F P
0 g
s
.81822 .43942 m .81657 .43312 L .77439 .36969 L closepath p 1 g
F P
0 g
s
.77871 .78367 m .82041 .69183 L .78945 .73589 L closepath p 1 g
F P
0 g
s
.80736 .74661 m .82041 .69183 L .77871 .78367 L closepath p 1 g
F P
0 g
s
.78945 .73589 m .82041 .69183 L .79724 .63602 L closepath p 1 g
F P
0 g
s
.79724 .63602 m .82041 .69183 L .81725 .5835 L closepath p 1 g
F P
0 g
s
.82735 .53202 m .79955 .45183 L .79938 .50128 L closepath p 1 g
F P
0 g
s
.81725 .5835 m .82735 .53202 L .79938 .50128 L closepath p 1 g
F P
0 g
s
.82735 .53202 m .81822 .43942 L .79955 .45183 L closepath p 1 g
F P
0 g
s
.83853 .67604 m .84133 .64524 L .80736 .74661 L closepath p 1 g
F P
0 g
s
.80736 .74661 m .84133 .64524 L .82041 .69183 L closepath p 1 g
F P
0 g
s
.82659 .70553 m .83853 .67604 L .80736 .74661 L closepath p 1 g
F P
0 g
s
.84234 .50674 m .81657 .43312 L .81822 .43942 L closepath p 1 g
F P
0 g
s
.81725 .5835 m .83435 .61405 L .82735 .53202 L closepath p 1 g
F P
0 g
s
.82041 .69183 m .83435 .61405 L .81725 .5835 L closepath p 1 g
F P
0 g
s
.81822 .43942 m .84503 .51699 L .84234 .50674 L closepath p 1 g
F P
0 g
s
.82735 .53202 m .84503 .51699 L .81822 .43942 L closepath p 1 g
F P
0 g
s
.82041 .69183 m .84133 .64524 L .83435 .61405 L closepath p 1 g
F P
0 g
s
.85185 .59781 m .84503 .51699 L .82735 .53202 L closepath p 1 g
F P
0 g
s
.83435 .61405 m .85185 .59781 L .82735 .53202 L closepath p 1 g
F P
0 g
s
.84133 .64524 m .85185 .59781 L .83435 .61405 L closepath p 1 g
F P
0 g
s
.83853 .67604 m .85185 .59781 L .84133 .64524 L closepath p 1 g
F P
0 g
s
.84875 .63056 m .85185 .59781 L .83853 .67604 L closepath p 1 g
F P
0 g
s
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[79]:=",
  ImageSize->{448.625, 484.062},
  ImageMargins->{{12, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000700001i0P10`40000100000?l0003o0?oo0000
ool0o`3ooooooeEEELIaLG76LHj>>75aaXhhSSR>SZZZZTaaWX20P492@/K6aZFOP=[GbU]GA^SXj000
?g=cLda<C2LW9eJ0Zk75fBa1Eb40<0X:2Q@D51hN7RPX:3<c<cdm?DM7Ae5ADE]KFfIVIW1`L7YjNXB4
Q8j>SYVIVJ>SXjf][KNg]l71`Lc<c=KFe^3Ph>[Zj_Cdm;V>S^;@d6I>C]GEeMCDe5ADE8:2PW:LW<7D
e3eDE:b/[:BTY>c/k61PH>7QhBDU9Bl_;i6ATLC4a000O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1o
O`1o_`1ood1PP<gIiP2o0410YR]0EADO:X]ZJSU>CQ`V9]GQhKW>cUEeMJ_0e@2o?`2oO`2o_`2oo`3o
?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cno
Ocno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmo
Ogmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?klo
Oklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_kooool0?ol0Ool0
_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_gAdM:NW
Yg@7i/;2`W9bL[>c/n;RhYRHV>GUiGinOZj^[YW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI_l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7
`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7
`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7
`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7
`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7
`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7
`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00?l7`@L00747303o1d@7
001/1`/01PL40<H72`1^1`00J0L@00T7102o1`<00PL400@71P1Y1`00IPL300h74@2i1`<000<70000
0P0<1`P0H`L006@74`0?1`0300L00080/`L70147301N1`00HPL400031`070087200C1`0400L71`D0
Z`L700d72@031`0300L000X0F@L006470`021`0300L700X7200<1`0300L700870P000`L000020:@7
1P081``02@L010071`L200<71`1F1`00G`L300@700<01`L04PL800@700<01`L010L200<7102M1`D0
0`L?010700<01`L00PL200P700<01`L0D`L005d70`051`0300L700h750061`@00`L209P72`031`H0
5PL00`071`041`801`L00`071`1B1`00F`L300H74P0;1`@00`L600D70P0010L01`L209@71@051`H0
50L;00H70P061`80DPL005T70`051`@000<700005@L400T700@01`L710031`800PL01007002@1`80
10L600X75P021`L030L00`071`1?1`00E`L300<71P051`803`L400h70P051`@000<700000`L308`7
60081`X01PL00`071`02008710091`0300L704h7001E1`<00PL300870`081`802@L401@700<01`L0
1PL500<70`291`T04@L700031`0000<02PL00`071`021`8010L400031`0700<700<01`L0C@L005<7
0`0010L000041`8030L200<7100I1`0300L700T710031`0300L708D700<01`L04PL600@7100@1`03
00L700<700<01`L01@L400<700<01`L0C0L005471P051`<03`L801T70P081`800PL200<700<01`L0
P`L00`071`0<1`L01@L501@700<01`L010L200T70`000`L01`1<1`00C`L400L70`0>1`<000<700L0
0`021``03`L00`071`031`801PL01@071`L008@700<01`L01@L700P7100@1`/01`L200X700@01`00
B`L004h70P081`@02`L400@700<01`L00PL200`730041`8000<700002@L2008700<01`L0P0L00`07
1`0600/710021a8010L400031`0000801`L200T70P000`L01`191`00C@L00`071`071`<02PL300P7
00<01`L010L301D7200<1`8000<700L0O`LK01<70`051`0300L700<01`L00`071`071`8000<700L0
B0L004`700<01`L01PL300T70`0;1`0300L700L70`0E1`800PL500X700@01`00O@L200l70P0H1`@0
20L010071`L500D70P081`8000<700L0A`L004`700<01`L010L400H7100>1`0300L700X70`0@1`80
1@L2008710071`8000<700L0NPL200h70`0F1`<030L00`071`031`8000<700001@L200L710171`00
B`L00`071`041`0300L000D70`0B1`0300L700d70`0;1`802@L00`071`031`@01@L00`07001i1`80
3@L200031`0701<70`0@1`0300L700@700<01`L00P051`0300L700D700@01`00APL004X700<01`L0
0`L400<70`0E1`0300L701070P071`8030L00`071`061`@00PL307L700<01`L02`L2008700<01`L0
3`L401<700<01`L01@L200870P041`801PL00`07000204@700191`0300L700870P000`L01`040107
5@081`<00PL200l700<01`L02@L607H700<01`L02@L200<700<01`L03@L301L700<01`L01`L01007
1`L300<70P051`0300L00080@`L004P700@01`L70P000`L0000200@740051`@000@7000020L>0147
0P0=1`<0M0L00`071`091`0300L7008700<01`L02PL401X700<01`L020L200@70P031`8010L20003
1`00048700171`0600L71`072P0B1`@00PL00`071`021`804`L4010700<01`L030L2078700<01`L0
20L200@700<01`L020L301X71P0;1`0300L700<70P031`0300L700<700@01`00@@L004L700<01`00
100H1`<000<700L00`L00`071`041`<04@L00`071`0300h700<01`L02PL00`07001a1`0300L700L7
0P051`0300L700D7100C1`X00PL200041`0000X70P051`<000<7000010L0100700101`00APL50003
1`0001H70P021`801@L00`071`071`<03`L00`071`021`<030L00`071`091`0400L006l700<01`L0
1PL200H700<01`L00`L300d72P0;1`0300L700<70`0:1`801PL200031`0000<700@01`L0?`L004D7
0`031`8050L300<700<01`L01@L00`071`0:1`803PL00`071`041`<02PL00`071`081`0400L006h7
00<01`L01@L200L700@01`L70`061`X04`L200L71@091`0300L700D70P000`L000021`0300L70080
?@L004@70P041`804PL300@70P081`0300L700`70`0<1`0300L700H70`081`8020L00`07000206`7
00<01`L010L200P73P0L1`0300L700T700<01`L00P081`801`L200051`071`000PL203`700131`03
00L700<700<01`0040L200D70P0:1`0300L700l70`0:1`0300L700P70`071`0300L700@700@01`L7
0`1Z1`0300L700<70P081`H09@L00`071`0;1`800PL200P700<01`L01PL400031`07008700<01`L0
>@L004<700<01`L00PL00`07000>1`<01PL00`071`0:1`0300L701870`081`0300L700X70`051`03
00L700<700@01`L70`1Y1`0300L700<700<01`L01@L400D72@0J1`8040L010071`L300H70P091`@0
0`L00`071`0h1`00@PL00`071`021`0400L700`70P071`803@L00`071`0E1`801`L00`071`0<1`<0
0`L00`071`021`0600L71`070P1W1`0300L700870P041`@000@700003PL9010700<01`L04@L200@7
0P061`802@L300031`0003X700121`0500L71`000PL00`071`081`<01`L200l700<01`L05`L300D7
00<01`L03PL300031`00008700<01`L00PL200031`0706D700<01`L00P031`<010L00`07000H1`T0
1@L201H700<01`L00`L200H700<01`L020L200031`0003T700111`0500L71`000PL00`071`061`<0
2@L00`071`0>1`P05PL300<700<01`L040L500D700@01`00I@L0100700021`<01PL00`07000R1`H0
60L200D70`041`802@L01007000i1`00@@L010071`031`0300L700@70P0:1`803@L500031`0700@7
300=1`800PL00`071`0>1`H01@L01@0700L006<71`081`0300L002471`000`L0000201H700<01`L0
1@L200@700<01`L01PL010071`0i1`00@0L010071`031`0300L700870`0;1`0300L700X70`021`03
00L7008700<01`L03PL=008710051`X01@L200D700@01`00HPL400X70P021`0300L701X71@021`80
0PL00`071`031`@04`L200L70`021`801`L01007000h1`00?`L010071`031`0300L700<030L200X7
0`041`0300L700@70P0K1`/03`L400<700D01`07001P1`8030L01@071`L001P71@051`8010L00`07
1`071`@04@L200P70P021`0300L700@700@01`00>0L003l700<01`0010L300d70P081`@01@L200T7
00<01`L06PL300031`0001<700@01`L00PL00`071`02060700<01`L02PL01@071`L001D710081`80
1PL00`071`0;1`<040L00`071`071`8000<7000010L01007000h1`00?PL00`0700031`<03PL00`07
1`051`<020L00`071`0:1`0300L701H70`041`<04PL2008700<01`000PL00`07001N1`0300L700X7
00<01`L00PL00`071`0>1`D02PL200P700<01`L03PL400d70P091`D00PL01@0700L003L7000n1`80
0PL600/70P041`@02PL00`071`0<1`0300L701870`071`0400L001470P031`<000@700L0GPL00`07
1`081`8010L00`071`0;1`@03@L200X700<01`L04PL400/700<01`L02@L300061`0000L0=`L003d7
0P0010L000L200<700<01`L01`L010071`L300`70P0@1`0300L700d7100:1`0500L01`0040L00`07
00041`0300L70080G@L200T700<01`L00`L00`071`071`D03`L200`700<01`L05PL400P70P0;1`@0
00<70000=`L003d70`021`0300L000@70P051`8000@700003PL00`071`0A1`802PL300d700L01`L7
00L001070P041`8000@700L0G0L00`071`071`0300L700<700<01`L00`L501870P0>1`0300L701X7
0`071`0300L700X70P0010L01`0f1`00?0L200@70P071`8000<700000`0@1`0300L701@700<01`L0
10L3010700<01`L00PL00`07000?1`0300L000<70P021`80G0L00`071`051`801PL00`07000301D7
0P0@1`0300L701d710041`8020L200061`0700L0=PL003`700<01`L00PL00`0700091`804PL201P7
00<01`L00`0C1`0300L7008700@01`L03PL00`0700031`<000<70000F`L00`071`051`0300L700@7
100G1`804PL00`071`0Q1`@00PL00`071`031`800`L01@0700L003H7000l1`0700L71`0700081`D0
40L00`071`0F1`D05PL00`071`031`0400L700d700@01`L00PL400031`0005X700<01`L010L00`07
1`031`<000<700005@L201@700<01`L09@L500031`0000D70`021`0300L703<7000k1`0500L71`00
0PL00`071`051`0300L000<700<01`L030L00`071`0A1`H00`L301D700<01`L00`L01@071`L000`7
00@01`L00PL400031`0005X700<01`L00`L00`071`021`8000<700L00`L00`071`0@1`8050L`00L7
0`021`0300L703<7000k1`0400L700<700<01`L010L010071`041`802PL200h71P091`0400L701@7
00<01`L010L01@071`L000`700D01`L01`0300041`0705T700<01`L00PL200<70P021`0300L700@7
00<01`L03@L201071P021`0300L702D71P071`8000@700L0=@L003/700<01`000`L00`071`031`80
0`L00`071`041`801`L00`071`081`H03`L01@071`L001<700<01`L01@L01@071`L000/700D01`L0
1`0200031`0700<0F0L01@071`L000<70P031`0300L700H70P0;1`8030L600P700<01`L08PL300@7
00<01`001@L300041`0703D7000k1`0300L000<700<01`L00PL00`071`021`0300L700L700<01`L0
0`L00`071`031`H05@L00`071`021`804@L00`071`051`0300L7008700<01`L020L010071`L20003
1`0000870P1H1`0400L700870P041`0300L700T700<01`L01PL200P71P0>1`0300L701h710061`04
00L700D700D01`0700021`0300L70387000j1`0300L000<700<01`L00PL00`071`031`0300L700P7
0P000`L0000601/700<01`L010L00`071`0=1`0300L700L700<01`L00PL200P700@01`L70P001@L0
1`070080E`L01@071`0700801@L00`071`0;1`0300L700<70P041`H050L00`071`0K1`<02PL01007
1`051`0300L000800PL00`071`0b1`00>PL200@700<01`L00P051`0300L700/70`0Q1`0300L700D7
00<01`L030L00`071`071`0300L700@700<01`L01PL02@071`L01`0700021`80EPL500H700<01`L0
3@L200031`0000H06PL00`071`0G1`@030L00`071`021`0300L7008700<01`000P021`0300L70387
000j1`0300L7008700@01`L01`L00`071`091`<00PL501/700<01`L01`L200`700<01`L020L00`07
1`041`0300L700D700T01`L700L01`000`L00`071`1C1`@01`L00`071`0?1`@080L00`071`0D1`<0
40L00`071`021`0300L7008700<01`000P031`0300L70347000j1`800`L00`0700081`0300L700L7
0P000`L01`071`D05PL00`071`091`0300L700T700<01`L01PL500D700<01`L010L010071`L20004
1`0700870P1D1`801`L200l710031`D06`L00`071`0A1`<04PL00`071`041`0600L71`070P001@L0
1`L003<7000j1`0300L000@020L00`071`061`800`L00`071`0;1`D04@L00`071`0:1`802@L00`07
1`051`0300L7008000<7000010L00`071`031`0300L700<000@700L00`L00`071`1B1`0300L700D7
00<01`L030L300041`0700P71@0F1`0300L700d7100E1`0300L700@700P01`L700L01`800PL00`07
1`0a1`00>PL01@071`L000X700<01`L010L200D700<01`L040L500`700<01`L030L00`071`061`03
00L700@700@01`L70`021`0300L7008700<01`L00PL00`071`0300041`0700<700<01`L0DPL00`07
1`041`0300L700X70`031`0300L000h71@0A1`0300L700X70`0H1`0300L700D700P01`L700L01`80
0`L00`071`0`1`00>@L200<70`081`0300L700870P061`0300L701H71@071`0300L700d700<01`L0
1@L00`071`021`800`L0100700031`800`L00`071`021`0400L01`<00PL010071`1D1`0300L700<7
00<01`L020L300@70P000`L01`0C1`D030L00`071`061`@06`L00`071`061`0700L700L01`0200<7
00<01`L0<0L003T700@01`L70P021`801@L010071`L200P700<01`L06`L5008700<01`L03PL200D7
00D01`L700051`0500L01`0010L2008700P01`L700L01`<00PL010071`1D1`0300L7008700<01`L0
1PL300H700D01`L7000I1`D01`L00`071`031`<07PL00`071`071`0700L700L01`0200<700<01`L0
<0L003T700@01`L70P041`800`L0100700091`0300L702471P0?1`0500L71`000`L00`071`041`03
00L0008700<01`L00`L200081`071`0700L200071`071`07001D1`0500L71`001PL300L70P031`03
00L701d71@021`0300L000<080L00`071`081`0300L0008700@01`0010L00`071`0_1`00>@L01P07
1`0700H71@0;1`0300L701d710021`800PL800P700D01`L01`0200L700H01`L01`071`0800L01`07
00L200071`071`07001D1`0400L700@70`091`0300L7008700<01`L08`L402@700<01`L02@L20087
00<01`L00P031`0300L702l7000i1`0300L700802@L00`071`0;1`0300L701P71@061`0400L000P7
20000`L0000200T700@01`L00PL00`071`051`<000H700L700L2008700<01`00E0L00`0700021`<0
2`L00`071`021`0300L70247200R1`0300L700X70P021`0300L700800`L00`071`0_1`00>@L01@07
00L000P700@01`002PL00`071`0E1`@02`L01@071`L000l7100:1`0400L700<700<01`L01PL30087
00@01`000`L205@71@0<1`801@L00`071`0M1`@010L2008700<01`L07`L00`071`0;1`0600L71`07
0P041`0300L702h7000i1`0500L01`001`L010071`L200T700<01`L040L500l700<01`L00PL00`07
1`0?1`0300L700P700@01`L00`L00`071`081`0500L700L00P031`80E0L200h700<01`L010L00`07
1`0J1`@01`L300<70P0N1`0300L700`700H01`L700L200@700<01`L0;PL003P70`021`0300L700@7
00<01`L00PL0100700071`0300L700/71@0D1`0300L700<700<01`L03PL00`071`081`0500L71`00
0`L00`071`071`0400L700870P031`0300L7058700<01`L02`L200H700<01`L060L300/700@01`00
10L00`071`0K1`0300L700X70`031`0400L000@700<01`L0;PL003P70`021`0300L700<700<01`L0
10L00`071`0200@700<01`L020L401T700<01`L010L200d700<01`0020L00`071`021`0500L71`00
20L2008700<01`L00P031`0300L7058700<01`L02PL00`071`061`0300L701@7100=1`<000<700L0
10L201X700<01`L02PL0100700031`0400L700@700<01`L0;PL003P70P031`0300L7008700<01`L0
1PL01@071`L000<700<01`L00`L501d700<01`L01PL00`071`0:1`0300L000P700<01`L00PL00`07
1`021`0300L700D70`0010L01`L200<700<01`L0DPL200X700<01`L01PL00`071`0A1`@040L20004
1`0700L70P0H1`0300L700P70P000`L00002008700<01`L00P031`0300L702h7000h1`800`L00`07
1`021`0300L700L700@01`L70`021`@08PL00`071`071`0300L700T700<01`0020L00`071`021`03
00L700<700<01`L010L3008700@01`L010L00`071`1A1`8020L200P700<01`L03PL401@70P0010L0
1`091`0300L701@700<01`L020L00`071`05008700<01`L00P031`0300L702h7000h1`0300L70087
00D01`L7000;1`0300L700871@0T1`0300L700P700<01`L020L00`0700081`0300L700<700<01`L0
0PL00`071`031`0400L0008700<01`L00P0010L01`1C1`801`L00`071`081`0300L700X7100G1`80
0PL00`07000:1`804`L00`071`071`800PL00`070003008700<01`L00P021`0300L702l7000h1`05
00L71`000`L00`071`0:1`0300L7008700<01`L00PL801`700<01`L02@L200L700D01`L700071`03
00L700<700<01`L00`L00`071`021`0400L0008700<01`L00P0010L01`1C1`801@L200X700<01`L0
20L301/70P021`0300L000`700<01`L040L00`071`061`0300L7008700<01`000`021`0300L70080
0PL00`071`0_1`00>0L2008700@01`L03PL00`071`021`0300L700T7200D1`0300L700/700<01`L0
10L01@071`L000L700<01`L00`L00`071`041`0400L7008700D01`0700021`8000@700L0D`L200@7
00<01`L02@L00`071`051`@07@L200<700<01`003@L200l700<01`L01@L200@700@01`070`021`05
00L71`000PL00`071`0_1`00>0L01`0700L01`0040L01@071`L001<7200<1`0300L700`700<01`L0
0`L01@071`L000L700<01`L00`L00`071`041`0400L7008700D01`0700031`0500L01`00D`L00`07
00021`0300L700T700<01`L00PL4020700<01`000`L010071`0>1`0300L700`700<01`L00`L200D7
00D01`L01`0200041`0700<700@01`L0<@L003P700<01`L00`0B1`0400L701/720041`0300L700d7
00<01`L00PL01@071`L000L700<01`L010L00`071`041`0300L0008700D01`0700031`0500L01`00
D`L00`07000200`71@0T1`8010L010071`0?1`802`L00`071`031`0300L700@700`01`L01`0700L0
1`031`8000<700L0<0L003P700@01`L70P0B1`0500L71`008PL900`70P000`L01`041`0300L700<7
00<01`L01@L00`071`041`800`L01@0700L000<70P021`0300L7054700<01`003`0W1`0300L000<7
00D01`L7000A1`0300L700P700@01`L70P071`0<00L700L01`0700L00`L200031`070307000i1`03
00L700H03`L010071`0O1`D00PL00`071`021`X010L200D700<01`L00`L00`071`051`0300L700D7
00T01`L700L01`0010L200031`07058700<01`0030L00`071`0202D70P041`0500L71`004PL200L7
00D01`L700081`0;00L71`0700L01`000PL01`071`L01`00<@L003T700H01`L01`031`@030L00`07
000I1`H01`L00`071`0<1`H01@L00`071`031`0300L700D700<01`L01@L01@071`L000870P041`80
00<700L0DPL00`07000<1`0300L700870P0R1`0300L000@700D01`L7000D1`801@L0100700081`05
00L71`000PL00`0700021`0300L000<700<01`00<@L003T700@01`L00PL00`071`041`<02PL00`07
000C1`D03@L00`071`0?1`8000<700L00`L00`071`031`0300L700H700<01`L00`L200<700H01`L0
1`031`8000<700L0DPL300/700<01`L01@L201l700<01`001@L01@071`L001H700L01`L700L000T7
00<01`L00PL01P071`07008700<01`000`L00`07000a1`00>@L00`0700041`0300L700H710071`80
3PL5018700<01`L03@L200<700<01`L00PL00`071`031`0300L700H700<01`L00`L200<700H01`L0
1`031`8000<700L0DPL300X700<01`L020L201d700<01`001@L01@071`L001L70P000`L0000:1`03
00L7008700H01`L01`021`0300L000<70P0b1`00>@L00`0700051`0300L700T710031`8020L601L7
00<01`L02`L200H700<01`L00PL00`071`021`0300L700H700<01`L00PL0100700021`0600L700L0
0PL300031`0705870P0:1`0300L700/70P0J1`0300L000H700<01`L00PL00`071`0F1`0300L700T7
00<01`L00PL010071`021`0600L700L00`L20387000i1`0300L000H700<01`L030L600871@0M1`03
00L700T70P081`0300L7008700<01`L00PL00`071`061`0300L700870P001@L01`L000<70P021`05
00L01`00D`L200X700<01`L03@L201L700@01`L01PL00`071`021`0300L701D70P0:1`0300L700<7
00@01`L00PL01P071`0700@700<01`L0<0L003T700<01`001`L00`071`0?1`D08@L00`071`071`80
2`L01@071`L000@700<01`L01`L020071`0700L00PL010071`L2008700@01`00E0L200T700<01`L0
40L201D700<01`001`L00`071`021`0300L701D70P0:1`0300L7008700D01`L700021`0400L71`80
10L00`071`0`1`00>@L00`0700071`0300L700h70`031`H06`L00`071`061`0300L700`700@01`L0
0`L00`071`081`0800L700L01`021`0400L71`800PL01007001D1`0300L700L700<01`L04`L30147
00@01`L01PL00`071`031`0300L701@700<01`002@L00`071`031`0500L71`000PL010071`L200@7
00<01`L0<0L003X70P081`0300L700`700@01`L02@L601D700<01`L010L200l700D01`L700021`03
00L700P700<01`000PL00`0700021`0400L71`800PL01007001D1`0300L700H700<01`L05`L200l7
00<01`001`L00`071`031`0300L701<700@01`L020L00`071`041`0500L71`000PL010071`L200@7
00<01`L0<0L003X700<01`L020L00`071`091`800`L00`071`0=1`H03`L00`071`021`804PL01007
1`021`0300L700P700<01`000PL00`0700021`0400L71`8000<700L00P1E1`0300L700H700<01`L0
6@L200`700@01`L01`L00`071`031`0300L701<700@01`L01`L00`071`041`0500L71`000`L01007
1`L200<700<01`L0<@L003X700<01`L02@L00`071`071`0300L700<700<01`L04`L600T700<01`L0
0P0D1`0400L7008700<01`L02@L01@071`L0008700<01`0010L0100700L205D700<01`L01@L00`07
1`0L1`802@L010071`081`0300L700<700<01`L04PL01@071`L000L700<01`L010L01@071`L000<7
00@01`L70P031`0300L70347000j1`<02PL00`071`041`801PL00`071`0I1`H00`L301L700<01`00
0PL00`071`091`0500L71`000PL00`0700041`0300L000<0EPL00`071`031`0300L701l70P071`04
00L700P700<01`L00`L00`071`0A1`0300L7008700<01`L010L00`071`041`0300L7008700D01`L7
00041`0300L70080<`L003X700<01`L01@061`0300L7008700<01`L01PL00`071`0O1`H05`L01P07
1`0700T70`031`0600L700L010L505L700H01`L01`0T1`8010L010071`091`0300L700@700<01`L0
40L00`071`031`0300L7008700<01`L01@L00`071`021`0500L71`0010L200031`0003<7000k1`03
00L700@71P000`L01`0200T700<01`L070L700871@0C1`0500L01`001`L200031`00008700L01`L7
00L000@70`000`L01`1F1`0600L700L09PL2008700@01`L02@L00`071`041`0300L700l700<01`L0
10L01@071`L000H70`041`0500L71`0010L200031`0703<7000k1`0300L700X70`0;1`0300L701P7
10031`0500L71`001PL500l70P000`L01`041`800`L2008700L01`L700L000D70P000`L01`1G1`04
00L001@75`021`0300L700P700<01`L010L00`071`0>1`0300L700D700D01`L700041`<000<700L0
0`L01@071`L000@7100d1`00?0L00`071`0:1`@02@L00`071`0D1`@01PL00`071`021`0300L700T7
1@0:1`<010L200D70P021`0700L71`0700051`0300L005T700<01`005@0C1`8000D700L700091`03
00L700D700<01`L03PL00`071`051`0400L700<70P001@L01`L000<700<01`L00PL00`071`031`<0
=0L003`700<01`L02@L01@071`0700801`L00`071`0@1`@020L200D700<01`L03PL500H70P021`80
1PL00`0700021`0300L700870P051`0300L005T700@01`009@L200<700<01`002PL00`071`051`03
00L700d700<01`L01PL00`0700021`800PL01@071`L000<700<01`L00PL00`071`031`0300L003@7
000l1`0300L700T700@01`L00`L00`071`041`0300L700`7100;1`0300L700D700<01`L04`L50003
1`00008020L00`0700021`0300L700870P041`0400L705X71@0Q1`8010L010071`0:1`0300L700D7
00<01`L030L00`071`071`8000<700000`L00`071`021`0500L71`0010L00`071`031`80=@L003d7
00<01`L01`L01@071`L000@70P041`0300L700P7100>1`0300L700L700<01`L05`L300X700<01`00
0PL00`071`021`8010L00`07001K1`<00PL00`071`0L1`801PL00`07000;1`0300L700D700<01`L0
30L00`071`071`<01@L00`071`021`0500L71`0010L00`071`021`<0=@L003d700<01`L01PL00`07
1`031`0300L700<70P021`0300L700@7100@1`802PL00`071`0H1`802PL00`0700021`0300L70087
0P031`0400L705/70`031`0300L701T70P081`0300L000/700<01`L01PL00`071`0:1`0300L700P7
00<01`L010L00`071`031`0500L71`0010L00`071`021`<0=@L003h700<01`L010L00`071`041`03
00L700D70`021`@04`L00`071`0:1`0300L701P70P091`0400L7008700<01`L00PL200870P000`L0
1`1L1`<00`L00`071`0F1`802PL200`700<01`L01PL00`071`0:1`0300L700L700<01`L010L00`07
1`041`0400L700D700@01`L7100e1`00?PL00`071`031`0300L700D700<01`L01`L301H700<01`L0
30L00`071`0F1`0400L700P700H01`L01`051`800PL010071`1M1`<010L00`071`0C1`802`L00`07
000<1`0300L700H700<01`L02@L00`071`061`<01@L00`071`051`0400L700D700<01`L00P000`L0
1`0e1`00?PL00`071`021`0300L700L700<01`L01@L200031`0000804@L200l700<01`L05@L01@07
1`L000P700H01`L01`061`0600L000L0G`L200D70P0A1`803@L200d700<01`L01PL00`071`081`03
00L700H700<01`001@L00`071`061`0400L700D700<01`L00P000`L01`0e1`00?`L01@071`L000T7
00<01`L010L010071`041`<03@L00`071`0?1`0300L701D700D01`L700071`0300L700870P061`80
0PL00`071`1N1`801`L00`071`0<1`803`L200`700<01`L01`L00`071`081`0300L700D700<01`00
1@L00`071`071`0400L700D700H01`001`0g1`00?`L010071`0;1`0300L7008700D01`L700071`@0
20L00`071`0@1`0300L701@700<01`L00PL00`071`051`0300L700870P061`<000<700L0G`L00`07
00071`0300L700T70P0A1`0300L700/700<01`L01`L00`071`071`0300L700@70P021`0300L70087
00<01`L02@L00`0700051`8000@70000=`L004070P0<1`0500L71`001@L00`071`081`@00PL201@7
00<01`L04PL00`071`031`0300L700D700<01`L00PL200H700@01`L0HPL200P700<01`L01PL20187
0P0=1`0300L700L700<01`L01PL00`071`041`800PL00`071`021`0300L700X700<01`001@L30003
1`0703L700101`@02PL010071`061`0300L700`70P0F1`0300L7014700<01`L010L00`071`041`03
00L700<70P051`@0H`L200T70P041`8050L00`071`0<1`0300L700P700<01`L01@L00`071`031`05
00L71`0010L00`071`0;1`801PL00`07000203P700111`0300L700P00`L200L700<01`L02`L20003
1`0001@700<01`L04@L00`071`051`0300L700<700<01`L00`L200@7101U1`802PL01007000:1`d0
3PL00`071`081`0300L700@700<01`L00PL200<700<01`L00PL00`071`0<1`801PL403T700111`03
00L700P71@081`0300L700P70P051`0300L7018700<01`L03`L00`071`061`0300L700<700<01`L0
0`L200<71@1U1`802PL>00/7100<1`0300L700P700<01`L00`L00`071`021`8010L01@071`L000l7
0P051`D0>@L0048700<01`L010L400041`001`801PL00`071`061`8020L00`071`0A1`0300L700h7
00<01`L01`L00`071`021`0300L700@700@01`L70P000`L0001W1`0300L700@710000`L0000E1`@0
00<700002PL00`071`081`0300L700<700@01`L70P041`0500L71`0040L200D7100j1`00@PL00`07
1`0400D700<01`000PL200D700<01`L00`L200/70P0A1`0300L700h700<01`L01`L00`071`021`03
00L700@700<01`L00P021`0300L706H70P000`L0000200@700<01`0050L01P07000700870P081`03
00L700P700<01`L00PL01@071`L000D700<01`L00PL00`071`0>1`801@L200031`0703T700131`80
20L010071`041`<00PL010071`L200l700<01`L03PL00`071`0=1`0300L700P700<01`L00PL00`07
1`041`0700L01`07001Z1`8020L010071`0B1`0300L700800PL01@071`L000H700<01`L02@L01@07
1`L000870P061`0500L71`004@L200031`07008700<01`00>`L004@700<01`L01PL01@071`L000H7
0P0010L0000B1`0300L700h700<01`L02`L00`071`0:1`0400L700L70P0010L01`1/1`0300L700D7
00D01`L7000A1`0600L700L00PL00`071`021`8010L00`071`091`0600L71`070P061`0500L71`00
4PL00`0700021`<000<700L0>PL004D700<01`L010L00`071`021`0300L700H70`0D1`803PL00`07
1`0;1`0300L700X700@01`L01`L00`07000306d700<01`L010L00`071`021`0300L700d700L01`L7
00L000<700<01`L00`L2008700<01`L02PL0100700L200H700D01`L7000C1`8000L7000700L003`7
00161`0300L7008700<01`L010L00`071`051`0300L000<04`L00`071`0;1`0300L700X700<01`L0
2`L010071`051`0300L0008000<700L0KPL00`071`021`0300L700<700<01`L030L00`071`021`80
1@L00`071`041`<030L500L700@01`L04`L4008700<01`00?@L004H700<01`L00PL00`071`041`03
00L700D700<01`L00P0010L0000A1`8030L00`071`081`0300L700`700@01`L010L00`071`020003
1`0706l700<01`L00PL00`071`041`0300L700T70P041`0300L000D700<01`L01PL700H70`081`04
00L7018710041`80?PL004L700@01`L020L00`071`031`0300L700<700@01`L70`0@1`0300L700T7
00<01`L020L00`071`0<1`0300L000<700@01`L70P000`L01`1a1`0400L700L700<01`L020L00`07
1`041`0300L000H700<01`L01@L00`071`041`P020L010071`0@1`D01@L203l700181`0300L000P7
00<01`L00`L00`071`041`0300L700<70`0>1`0300L700P700<01`L01`L00`071`0=1`0300L00087
00@01`L70P000`L01`1c1`0300L000P700<01`L01PL00`071`051`0300L000L700<01`L010L200X7
0P091`0300L000l71@061`80@0L004T70P091`0500L71`0020L00`071`051`<030L200P700<01`L0
1PL00`071`0?1`8000@700L70P0010L0001e1`8020L00`071`051`0300L700D700@01`L01`L00`07
1`031`<02@L200T700<01`003PL600H70P111`00BPL400H700D01`L700091`8020L300/700<01`L0
1PL00`071`051`0300L700l70`000`L000031`80M`L300L700<01`L00`L00`071`061`0400L700P7
00<01`L00PL400L700<01`L020L00`07000<1`<000@7000020L20447001;1`0300L700@00`L00`07
000<1`0300L700P70`091`801PL00`071`041`0300L700l70P000`L000041`<0N0L00`07000200@7
00<01`L00PL00`071`071`0300L7008020L010071`L400D70P0;1`802`L200870`091`80@PL004`7
0P041`@000<700L030L00`071`0:1`<020L00`071`031`0300L700<700<01`L030L300870P051`<0
NPL200870`021`0400L700T700D01`L01`0300H700@01`L70`000`L01`021`802`L200T70`021`<0
2PL204<7001>1`0300L700@70`0>1`803@L300H700<01`L00`L01@071`L000`7100010L000041`@0
00<700L0N`L00`071`021`<000<700L02@L00`0700031`D010L00`071`0200031`0000870P0;1`03
00L000L70P031`<02`L204@7001?1`0300L700<000<700L0100<1`0300L700d70`041`800`L01@07
1`L000P710000`L01`0300@70`031`0300L707`700@01`L71P091`801@L00`07000400870P001@L0
1`07008030L00`0700051`<00`L300d700<01`L0@`L005070P051`0400L71`D02@L00`071`0?1`<0
0`L010071`021`0300L700@70`031`0300L000800`L300@70P201`<00`L00`07000600870P061`H0
00<700000P000`L0000400`70P041`801@L300/70`161`00D@L300<700<01`L010L600H700<01`L0
4@L300061`0700L010L300@700<01`000P021`<01@L208<70P041`0400L71`P01`L0100700L30003
1`0000@000<700000`081`800`L200D70`0:1`D0A`L005<70`021`0300L700H71`031`8050L:00D7
00@01`000PL400<71@261`0500L71`001`L300031`00008010L0100700LA00D70P0010L000051`<0
2@L400031`0704P7001E1`@02PL900031`0700/73@071`@00PL300<710000`L000281`8000<700L0
2@L900031`0701D00PL300H70`081`<010L00`071`181`00E`L700P780061`@000@700000PL400@7
00<01`L0RPL300`78@071`801`L300D70P1;1`00F@L300L71@000`L0000>00<72P041`P000<70000
0P051`80SPL<00H770061`<010L400L700<01`L0B`L005X7100?1`@01`L400031`0000P000<70000
1@000`L0000200L70P2@1`8000<700000`031a0020L:00@70`031`<02PL00`071`1<1`00G0L401D7
0`031aD01`L409@70P041`<05@L=00870`021`<02`L204l7001N1`@03`L400<70`000`L000020003
1`0700X010L809L70P021`800PL300h71@021`H000<70000100010L000091`D0D@L0060710091`@0
1@L300031`0000<00PL900031`0000@010L209/73`051`D00`L6008720041`T0E@L0068710031`@0
1`L300031`0000@00PL00`07000800P70P2O1`@00`L=00@71`021a00F`L006@74@000`L000030003
1`0000X02@L20:D710041`@01PL800872P1R1`00I`L400L700<01`00702[1a`0J0L006/710021`80
0`L800031`0000T0]@L507`7001_1aD0o`Lm1`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00H0L200D70`3;1`801@L30847
001R1`0300L700D700<01`L0bPL00`071`051`0300L707h7001>1`<01@L200<71@021`8010L300<7
0P021`<000@700070`0:1`0300L700<70P031`<00`L408P70`051`800`L500870P041`<00`L20087
0`0010L000L300X700<01`L00PL400870`041`80D@L004l700<01`L00`L010071`041`0500L71`00
0PL010071`021`0400L7008700@01`L00PL00`071`031`0300L700`700@01`L01@L010071`2<1`03
00L700<700@01`L010L01@071`L0008700@01`L00PL010071`021`0400L7008700<01`L00`L00`07
1`0<1`0300L700H700@01`L00PL00`071`1>1`00C`L300<700@01`L010L01@071`L0008700@01`L0
0PL010071`021`0400L7008700@01`L70P0B1`0300L700<700D01`L7002;1`<00`L010071`041`05
00L71`000PL010071`021`0400L7008700@01`L00PL010071`L2010700<01`L01@L010071`021`03
00L704h7001?1`0400L7008700@01`L010L01@071`L0008700@01`L00PL010071`021`0400L70087
00@01`L030L00`071`031`8010L300@700<01`L0R0L010071`021`0400L700@700D01`L700021`04
00L7008700@01`L00PL010071`021`0400L700`700<01`L010L00`071`021`<00`L20547001?1`04
00L700<70P051`0300L7008000<700000`L200041`071`800PL0100700041`<04@L010071`021`03
00L700<700<01`L0Q`L010071`031`801@L00`071`0200031`0000<70P0010L01`L2008700@01`00
10L3014700@01`L00PL010071`021`0300L704h7001>1`@02`L00`071`0[1`0400L7008700@01`L0
0PL010071`281`@02`L00`071`0[1`0400L7008700@01`L00PL010071`1@1`00G0L202h70P041`80
10L209L70P0^1`8010L200@70P1A1`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00o`O11`00
o`O11`00o`O11`00o`M>1`L0K0L007@730341``00`L706H7001_1`T00`L00`0700090;L71P051`<0
2PL:0607001X1`L00`L00`071`021a@0/@L200D7801M1`00I0L400031`0700<72@051`<00PL>0:/7
0P051`D00`L200d70P021`801@L205/7001R1`<00`L00`070004008700<01`000P061`0300L70080
1@L=0:D70P031`@000<700001`L300L70`061`801@L205T7001P1`800`L500L700<01`L00PL200@7
00<01`L00PL200<700<01`L00P031`L0X@L200031`0000800`L200`70P031`802`L200@70`1G1`00
GPL200031`0000800PL00`071`04008700<01`L01@L300031`0700D70P021`0500L71`0010L709d7
1@051`8030L900d70P031`@0E@L005`71@051`0300L700D750031`<010L200<700@01`07102I1`D0
1@L200<72`031`8000@7000720071`800`L01007001C1`00F`L3008710021`0300L000D00PL300T7
0`031`P010L200041`071`800PL309D70`041`X030L200D70P071`P000<700000PL00`071`020547
001I1`800PL00`071`041`@020L00`071`0200L700<01`L00P061``00`L010071`L209870`071`<0
00<700003@L200T700<01`L030L500031`0000870P1?1`00E`L200<700@01`L710000`L01`0200L7
00<01`L00PL200D700<01`L00PL00`071`031`0400L700D710031`8000@70000S`L400D71@041`80
2PL00`071`0:1`8030L200870`000`L01`021`80C@L005H70`031`@01@L010071`L200D700<01`L0
10L200<700<01`L00`L200<700@01`L70P051`D00PL00`07000308/71@041`800PL00`071`061`80
1PL200l700<01`L01`L200D700<01`000P041`80B`L005@70P031`@020L00`071`031`800`L00`07
1`061`8000<700L01PL2008700<01`L00PL00`071`031`800PL500041`0008P70`000`L01`021`<0
0`L00`071`091`800PL201870P051`8020L00`071`0300<70`191`00D`L200<70P000`L01`0200D7
00<01`L01PL200031`0700@730041`<01@L00`071`021`0400L000@70P0010L000261`800PL00`07
1`0200@70P0=1`l02PL200031`0000/700@01`L710031`0300L704H7001B1`0400L01`800PL00`07
1`021`800`L00`071`061`T01PL200@720051`800PL01@071`L000@70`0010L000221`800PL500D7
00<01`L02@L400031`0700802PL>00d700@01`L70`041`0300L704D7001@1`8000@7000010L00`07
1`041`8000@700L71P031`<02@L0100700081`H00`L00`0700041`0300L700870`0010L000201`80
0PL300L700<01`L01@L500@700<01`L00PL00`071`0A1`8000<700000`0:1`800PL300@700<01`L0
A0L004l71@061`0300L700H710091`0400L700P700D01`L700071`0300L000<71`041`0600L71`07
1P1n1`0300L700<01`L200<71@071`801PL200l70P041`0400L71`@020L0100700L300<700<01`L0
@`L004h700<01`L00P000`L000041`0300L70087100010L000091`0400L71`801PL00`071`021`03
00L700@700<01`L00P051`@010L010071`021`0300L000@0N`L01@071`L0008710021`0300L000<0
2`L00`071`081`0300L700X70P071`0300L700@70`061`0600L01`070P021`80@`L004d70P000`L0
00041`8000<700L010051`0400L700P700<01`L00`L00`071`031`0300L700<70P041`0300L70087
00<01`L010L500870`031`0300L700@0NPL010071`071`@03PL00`071`0:1`8020L200X700<01`L0
1PL400<70P021`0400L700870P121`00C0L400031`0700D70`091`0400L71`801@L00`071`051`80
0`L00`071`051`0300L7008700<01`L00PL00`071`031`0300L000871P021`0300L700D0N0L01007
00071`8000<700003@L00`071`0=1`0300L700<70P0=1`0300L700T71P021`8000@700070P111`00
B`L300<700<01`L00PL200041`0000P700<01`L00`L200<700<01`L01`L010071`081`0500L71`00
1@L00`071`021`0400L700D70P031`0300L700D0MPL0100700061`<010L200X700<01`L03`L20003
1`0001070P0=1`<00`L01@071`070080@0L004X700<01`000`L010071`L200<700<01`L00P051`03
00L700H700@01`L02PL300T70P000`L01`051`0500L71`000`L00`071`031`<00PL0100700L407@7
00@01`001@L2008700<01`L010L00`071`051`8040L8014700<01`L020L200870`021`0300L700@0
?`L004T700<01`0010L0100700051`0300L700870P031`0300L700L70`061`d01@L200L70P021`05
00L71`0010L00`07000200031`07008000@70000LPL01@0700L000<70P031`0300L700H70P041`03
00L700P720031`0300L000871P0<1`0300L700D70P051`0600L000L00`L303h700181`0400L01`80
00@700001`L00`071`041`8000<700L010L=00H70P051`T01`L00`0700031`0300L7008700<01`00
0PL400041`071`80L0L01@0700L000<700<01`L00PL00`071`091`8000<700L00PL800X700D01`L7
00071`H01`L00`071`021`801`L010071`L30087100m1`00B0L300<70`071`0300L700L71`061`<0
2PL00`07000:1`8000<7000010021`8010L01@071`L0008700@01`L70`000`L01`0207070P0010L0
1`L200@700<01`L02`L6014700<01`L00`L00`071`0:1`H00PL01007000:1`0300L700870`031`03
00L703/700171`0300L000<70`000`L000051`0300L700@70`000`L0000:1`0400L700T700<01`L0
0P081`0300L000H71@041`0500L71`000PL00`071`021`0500L700L00P1^1`0500L700L00P061`03
00L700T70`000`L0000B1`0300L700D700<01`L03`L500`700<01`L00PL200<70P0l1`00APL01@07
1`0700800PL010071`L200<700<01`L00PL200@700@01`0020L010071`L200L700<01`L00PL00`07
1`051`0400L700P710031`0400L700<700@01`L70P021`@0K0L00`071`0300L700<01`L01`L30003
1`0700<70P0>1`802@L00`071`0?1`8000@700002`L010071`L300<70P0k1`00A@L01P071`0700@7
00<01`L00`L200031`0700<01PL01@071`L000L700<01`L00`L00`071`041`0300L700<700<01`L0
10L01@071`L000L70P021`<000@700L00`L010071`L400031`000080JPL00`071`0200P700<01`L0
1@L300<700<01`L01@L00`071`0:1`0300L700X70P0>1`0300L7008700<01`000P081`0600L71`07
0P031`0300L703T700141`<000<700001@L00`071`051`<02@L00`071`021`0300L700@700<01`L0
10L00`071`031`0300L700@700<01`L00`L00`071`021`0300L700@700<01`0010L400@700@01`L7
0P000`L0000200031`0706L700@01`002@L00`071`041`801PL00`071`061`802@L00`071`0=1`03
00L700T70P051`0300L700870P071`0800L700L01`021`80>PL004@700<01`L00P051`0300L700@7
0P0010L000081`0300L700<70P041`0300L700D70P031`0300L700D700<01`L00PL00`071`031`03
00L700<700@01`L01@L200D700<01`L00P021`@000<700L0IPL00`07000300H700<01`L00PL300L7
00<01`L02@L00`071`051`0300L700l700<01`L01`L00`071`061`0300L700<70`051`0800L700L0
1`021`80>@L004<70P0010L000L200<700@01`L70`031`0400L700L700<01`L01@L01@071`L000T7
00@01`L020L2008700<01`L00`L00`071`031`0500L71`001@L300@700D01`0700021`<000<700L0
I@L0100700031`8010L00`07000200T700<01`L02`L200@700<01`L04@L00`071`041`802PL00`07
1`051`8010L00`0700021`0500L000L00P0h1`00@`L01@0700L000<70P000`L01`0200H700@01`L7
0P051`0300L700H700@01`L02PL300X700<01`001PL00`071`021`0500L71`001@L200031`0000<7
0P021`0400L71`8000<700L0I0L0100700051`8000@700002`L00`071`0>1`8000<700L050L00`07
1`021`0300L700/700<01`L01PL300041`07008700L01`L700L003P700121`<00PL00`071`031`<0
20L00`071`031`800`L00`071`071`<01PL=00D70P071`0500L71`0010L00`071`021`0300L00087
0P001PL01`0700<7101T1`0400L000P70P0<1`0300L700l73`0;1`0400L000h700<01`L02@L400<7
00@01`L70`0g1`00@PL200<700<01`L00PL500L700<01`L01@L010071`081`X01PL200D720071`04
00L700D700D01`L700021`0400L71`@00PL00`071`0506<70P000`L01`061`0400L000X700<01`L0
2`L500031`0700<02@L>010700<01`L02PL200<700<01`L00PL303H700111`800`L00`071`021`04
00L700870P051`0300L700H70`021`H00`L200/700<01`002PL600@700<01`001@L01@071`L00087
00<01`L00`L2008700<01`L01@1R1`8000<700L01@L300<70P081`0300L700H71@051`0600L71`07
0P0C1`H03@L00`071`081`@00`L00`071`021`80=PL0044700<01`L00PL01@071`L000<700<01`L0
0PL00`071`021`0300L700L710091`0300L000X700@01`L02@L00`0700031`@000<700001PL01007
1`031`0300L700<70P021`0400L01`@0H0L01@0700L000D700<01`001PL200D700<01`L00`L400T7
00<01`L00`L00`071`02014700@01`000PL400X700<01`L01PL010071`L3008700<01`L00PL203D7
00101`@00PL0100700041`0300L700<70P021`0300L700<710000`L0000:1`0300L7008020L01@07
1`L000P700<01`001`L400H700<01`0010L00`071`021`<00PL01@0700070080H0L200031`0700@7
00@01`L020L2008700<01`00100=1`0300L700@700@01`L70P0?1`0500L01`001@L400L700<01`L0
0`L200@700H01`L01`051`0300L703<700101`800PL400H700<01`L01@L300031`0000801@L01007
00081`0300L7008700<01`L01@L00`071`021`0300L700D700@01`L020L500@70P041`0300L70087
0P002@L0000700L01`0205h70P021`0400L71`800PL00`071`091`@04@L00`071`061`0300L70087
0P0=1`0600L700L020L400<700<01`L00PL00`071`051`0600L700L00PL403@7000o1`0300L000@7
00<01`L01@L00`071`061`<02@L01@071`L000L700<01`L00`L00`071`041`0300L700<700<01`L0
10L01@071`L000L70P031`H01@L01`071`L01`000PL300031`070080GPL010071`031`0300L70087
00<01`L020L4014700<01`L020L00`071`031`802`L01P071`L01`802PL500031`0000P700@01`L7
0P0010L01`L203@7000o1`0300L000<710051`0300L700@70P0010L000081`0300L7008700<01`L0
10L00`071`041`0300L700<700<01`L00`L00`071`041`0300L7008700<01`L010L00`0700061`80
1PL01P071`0700<70P021`@0G0L2008700@01`L010L00`071`071`8000@700L70P0>1`0300L700X7
00<01`L010L200T700<01`L00PL010071`0=1`802`L00`071`021`0400L71`80=0L003h700<01`00
0`L00`0700021`800`L010071`L300<700@01`L01`L00`071`031`801@L00`071`041`0300L70087
00<01`L010L00`071`021`0300L700<700<01`L010L00`0700071`<010L010071`021`0500L71`00
0PL01007001L1`0400L7008700<01`L00`L00`071`051`800`L00`071`021`0300L700X700<01`L0
30L00`071`051`801`L00`071`021`0500L71`003@L300T700<01`L00PL010071`L303<7000n1`03
00L0008700@01`L010L200031`0700801PL010071`L200H700<01`L010L00`071`021`0300L700D7
0P021`0300L700D700D01`L700061`0300L700<700@01`L01PL200031`0000<700<01`000PL01007
1`L20087101J1`0400L71`<01@L00`071`041`8010L00`071`041`0300L700P700<01`L03PL00`07
1`061`801@L00`071`031`0500L71`0030L200031`0000P700@01`L70P031`80<`L003d700H01`L0
1`031`0300L700@70`081`0300L700<700<01`L00`L00`071`051`0500L71`002@L00`0700081`04
00L700L700D01`L700041`0300L700<70P031`800PL200<700<01`L00`000`L0000205X71@071`03
00L700<700<01`L010L00`071`051`801`L00`071`0@1`0300L700L70P031`0300L700@700@01`L7
0P0:1`0300L0008700<01`L01@L010071`L200<70P0c1`00?@L500@700<01`L00`L500L700<01`L0
10L200<700<01`L01PL200031`0700T70P091`0300L000P700@01`L010L00`071`031`0300L000@7
10031`0300L7008000<700L0101I1`0400L700L700<01`L00PL200H700<01`L020L00`071`041`03
00L7014700<01`L020L200031`0700H700<01`L00PL00`071`071`0300L000<70P061`0300L70080
0`L30387000l1`800PL00`071`031`0300L7008700@01`L00PL00`071`051`0300L700D700@01`L0
2PL200D730051`802@L00`0700051`0300L7008700<01`001PL200@700@01`000PL505P700@01`L7
0`051`0300L7008020L00`071`091`0300L7008700<01`L03@LE00L700<01`L00PL00`071`061`04
00L700@70P041`0600L700L00PL30387000l1`0300L700@00`L01@071`L000@700<01`L00P051`03
00L700H70`051``01PL200@72@071`0300L000D700<01`L00PL010071`061`<00PL01@0700L00087
101H1`0400L700<700H01`L700L200T700<01`L02`L200031`0700<730051`0300L000/71P031`03
00L700<70P051`0500L71`001PL200<700D01`0700021`0300L00387000l1`0500L01`000PL01@07
1`0700801@L00`071`021`800`L00`071`071`L01PL200X700<01`002@L200031`0000@00PL200H7
00@01`L00`L00`071`041`8000@700L70P000`L01`021`<0F0L010071`041`8000<700002`L00`07
1`0<1`L040L01@071`L000X70P041`D01PL00`071`021`0300L7008700<01`L01@L01P071`070087
00@01`070P0a1`00>`L200041`0700<710071`0300L700@700@01`L01@L400031`0000X700<01`00
2@L00`0700091`801`L500H700<01`0010L00`071`021`0300L0008700@01`000PL00`071`0405L7
00@01`001PL00`071`0:1`0300L700X710000`L01`0B1`0300L700<700<01`L01`L00`0700071`<0
1@L01@071`L000@700<01`L01PL200031`00008700@01`070P0a1`00>`L3008700<01`L00`L00`07
1`061`0300L700D70`000`L0000200D700<01`002@L010071`081`0400L700P700<01`002@L01007
00041`0300L000@700<01`L00PL00`0700031`<00PL00`071`0405H700@01`L01PL400X700<01`L0
1`L300<700@01`L04PL00`071`031`0300L700H700@01`L020L600870P000`L01`041`0300L700L7
0`021`0300L0008700<01`L0;`L003/70P031`0300L7008710061`0300L700H70`091`0400L700P7
00D01`L700071`0500L71`001`L010071`081`800PL300870P051`0500L71`000PL00`071`021`04
00L70087101F1`0400L000D700<01`000PL200L700<01`L01@L300D700<01`L00PL200l700<01`L0
1@L00`071`051`0500L71`001`L00`0700031`H01@L00`071`091`0400L71`800PL20307000k1`03
00L7008700L01`L700L000870P041`0300L700@70P000`L0000:1`0400L700L700<01`L00PL00`07
1`041`0300L7008700<01`L010L010071`081`801@L400D700D01`L700021`0300L700870P021`03
00L000<0EPL0100700041`0300L000D700<01`L010L00`071`031`8020L00`071`041`0300L700/7
00<01`L01`L00`071`041`0300L7008700<01`L010L010071`061`<01@L00`071`061`@00PL20087
0P0`1`00>PL300<700@01`L00`L010071`L200<700@01`L70P031`0400L000P700D01`L700061`03
00L700<700<01`L00`L00`071`021`0300L700@700D01`L700061`0400L700H70`051`0400L700<7
00X01`L700L01`070P000`L0001E1`8000<700L00PL010071`061`800`L010071`L300T700<01`L0
1PL00`071`091`0300L700T700<01`L00`L00`071`021`0300L700<700<01`L00PL00`071`041`04
00L000<700<01`L010L2008700L01`0700L0008700<01`L0;PL003X70P001`L000L01`0010L00`07
1`031`8000<700L00P051`0500L71`001`L00`071`021`8010L00`071`041`0300L7008700<01`L0
0`L00`071`021`0300L700<700<01`L00`L010071`071`0400L000<700@01`L00`L010071`L20004
1`001`8000<70000E0L0100700031`0500L71`0020L200031`00008030L00`071`071`0300L700L7
00<01`L02`L00`071`021`0300L700<700<01`L00PL00`071`021`0300L700D700<01`L00P021`04
00L71`801@L0100700L2008700<01`L0;PL003X70P031`<01@L00`071`051`<020L01@071`L000H7
00<01`L010L01@071`L000L700D01`L700061`0500L71`001@L00`071`031`0500L71`001PL20087
0P021`8010L010071`L200031`07008000@700L0E0L200031`07008700D01`L7000:1`803PL00`07
1`091`0300L700D700<01`L03@L01@071`L000D700<01`L00PL00`071`031`0300L700@700<01`L0
0PL300031`0000L700<01`L00`000`L0000`1`00>PL200@700<01`L010L00`071`041`@020L00`07
1`021`8010L00`071`051`0400L700P700@01`L01`L010071`061`0300L7008700D01`L700051`03
00L000@710051`0300L700800PL300041`0705@70P001@L01`L000<700<01`L020L400d700<01`L0
2PL200D700<01`L03@L010071`071`0500L71`001PL00`071`031`0300L700@70P091`0400L71`<0
00<700L0;`L003T70`031`@010L00`071`021`800PL0100700061`0300L700@700D01`L700081`03
00L000T700<01`0020L00`0700071`0500L71`000`L00`071`031`0300L000H70P051`0500L01`00
0`L200041`0705@700<01`000PL00`071`021`0300L700H70`031`0300L700T700<01`L03@L01@07
1`L0014700<01`0020L010071`071`0300L700<700<01`L010L200P700<01`L00PL010071`0`1`00
>@L00`0700021`0300L000870P031`0400L700D700@01`L01@L00`071`051`0500L71`0020L200X7
0P081`0300L000L700D01`L700041`0300L7008700@01`L01PL300<700<01`000PL02@071`L01`07
001C1`8000@700L01@L00`071`051`0400L700@70P091`0300L700h700<01`004`L200P700@01`L0
20L00`071`021`0300L700@70`081`0500L71`000PL00`071`0^1`00>@L00`0700021`0400L700<7
0P0010L000061`0500L71`0010L00`071`061`8000<700L01PLD00@70P081`0300L000D700<01`L0
0PL010071`061`0400L700<70P021`0400L71`8000@70000DPL00`071`0200H700<01`L010L01@07
1`L000H700<01`L01@L00`071`0>1a/01`L200X700<01`L00PL00`071`021`0400L01`801PL01007
1`L2008700<01`L0;@L003T70`000`L01`021`0300L700<70P081`0300L7008700D01`L7000:1`80
0PL500870P0:1`8010L800L70P061`0400L700@700<01`L00PL00`0700021`8000<700000PL01007
1`L200041`00058710061`0300L700<70P031`0300L700H70P051`0300L700T71@021`804`L01007
00L600031`0000/700D01`L700041`0400L7008700<01`L00`L00`071`03008700<01`L0;@L003T7
10041`0300L700<710061`0300L700<700D01`L700091`@01`L00`0700091`802@L200031`0000@0
00<700001PL010071`041`0300L7008700<01`0010L3008700<01`L00P000`L00002058700<01`00
1`L00`071`021`0300L700<700<01`L020L01@071`L000H71P071`0300L0014700@01`L01`L400/7
00@01`L010L010071`031`0300L7008700<01`L00P0010L01`0_1`00>@L00`0700051`0300L70087
00<01`000PL00`071`041`0300L700<70P000`L01`041`@000<7000020L010071`081`0300L000P7
0P071`@01PL00`0700041`0300L7008700<01`001PL01P071`L01`800PL3058700<01`000P051`05
00L71`001PL00`071`091`8000<700L01@0=1`0400L700l700<01`L00PL00`071`061`D02@L01007
1`031`0400L700@700<01`L00PL02007000700L0;`L003T700<01`000P031`0500L71`000`L00`07
1`0200@700<01`L01@L200031`0000801@L00`0700081`0400L700L700@01`L01`L00`0700081`03
00L000800`L00`0700051`0700L71`0700061`800PL0100700021`<0DPL3008700<01`L00PL01007
1`071`0300L700X7100B1`0400L700l700<01`L00PL00`071`061`800`L200P700<01`000PL00`07
1`021`0300L7008700H01`L700L200041`0702l7000i1`<00PL2008700<01`0010L00`071`021`80
0PL00`071`051`<02@L010071`071`0500L71`001PL01@071`L000H700<01`0020L200<70`000`L0
1`051`0400L700<700<01`L00PL200051`0700L00P031`80DPL200051`071`000PL0100700071`03
00L700T70P000`L01`02014700D01`L7000=1`0300L700@700<01`L01@L00`0700041`<01@L00`07
00021`0300L7008700<01`L00`L200031`0700800PL202l7000i1`8000<700L00PL400D700<01`L0
10L00`0700041`<000<700002PL010071`061`0300L7008700<01`L00`L01@071`L000H700@01`L0
1PL00`0700061`<01@L010071`031`0300L700870P001@L01`0700800`L205870P000`L01`021`@0
2@L00`071`071`800PL00`071`021`<03PL00`071`021`0300L700T700<01`L01PL00`071`031`05
00L71`001PL300<700D01`0700051`0300L700@70`0010L01`L202l7000i1`8000<700L010L00`07
1`041`0300L700D70`000`L000051`<020L01@071`L000D700<01`L00PL00`071`031`0300L70087
00<01`L00`L010071`061`0400L700H700@01`0010L00`0700031`0300L700870P021`0500L01`00
0PL305870P000`L01`041`0300L700P700<01`L01@L200@700<01`L01@L200d700<01`L00PL00`07
1`071`0300L700L700<01`L00`L00`071`021`0300L700H70P021`8000<700L010L00`071`051`80
0PL01007000_1`00>@L200031`0700<710051`0300L700H70P071`0400L700L700<01`L00PL200<7
00<01`L00`L00`071`021`0300L700<700D01`L700041`0300L700<700@01`L01PL200870P021`03
00L000<700L01`L700L000<70P021`0300L00080DPL010071`041`0400L000L700<01`L010L200H7
00<01`L01`L200/700<01`L00`L00`071`051`0300L700T700<01`L00PL00`071`021`0300L700P7
1@061`0300L700H700@01`L00PL202h7000i1`0400L700<700H01`L01`041`0300L700@710071`05
00L71`001PL00`071`041`0400L700H700D01`L700051`0500L71`001@L00`071`021`0400L700D7
00<01`0010L200031`0700@700H01`L01`041`0500L700L00`1B1`0400L700<700<01`L00PL200D7
00<01`L00PL200L700<01`L02PL300P700<01`L010L00`071`031`0300L700X700<01`L00PL00`07
1`031`0300L700X70P071`0300L700D700@01`L00PL202h7000i1`0400L700<700@01`L00PL20087
00<01`L00PL200<70`061`0400L71`8010L00`071`051`0400L700H700@01`L01PL010071`051`05
00L71`0010L00`071`021`0300L000H70P051`0300L0008700<01`L00PL01@071`0700<0DPL20003
1`07008700<01`L010L200<700<01`L00P091`0300L700d70P061`0300L700@700<01`L00`L00`07
1`0;1`0500L71`001PL00`071`0:1`801PL00`071`041`<000<700L00PL00`071`0/1`00>@L20087
00<01`000`L00`071`021`0300L000<700<01`L00`L010071`051`0300L700<700D01`L700081`03
00L000L700<01`001`L00`0700061`0400L700@700<01`L00PL00`0700071`<00`L00`0700021`04
00L71`<000@700070P1B1`0700L01`0700091`<000<700002`L00`071`0?1`<00`L00`071`051`05
00L71`003`L010071`071`0300L700P700@01`001PL00`071`031`<000<700L00PL00`071`0/1`00
>@L00`07000300@700<01`L00`L500H700D01`L700041`0300L700@700D01`L700081`8020L200P7
0P061`0400L700D700D01`L700021`0300L700<70P021`8000@700L00`L00`071`0300041`001`80
DPL200041`0000D72@0=1`0300L701870P000`L01`071`0300L0010700<01`0020L00`071`081`04
00L01`8010L00`071`031`8000<700000PL202h7000i1`800PL00`071`041`0300L700@700<01`L0
1`L010071`L200<700<01`L010L200031`0700H74@041`801`L00`0700051`0400L700<700<01`L0
0`L200@70`041`0600L01`070`000`L0001B1`800PL700H7100;1`0300L701D7400>1`802@L00`07
1`061`0400L700<700<01`L00PL020071`L01`070P0010L01`0^1`00>@L200870`041`0300L700<7
10071`0300L700<700@01`L020L200871@021`8020L200<72P031`801PL010071`031`0300L70087
00<01`001PL00`071`021`0600L01`070P0010L01`1B1`800PL00`071`091`0300L0008700<01`L0
20L00`071`0A1`L020L200<72P041`802PL00`071`051`0400L700@700D01`L700031`0300L00087
0P000`L0000^1`00>@L200061`0700L00`L010071`L2008700@01`001@L00`071`041`0300L000L7
1@071`8020L200P700<01`L00PL600H700<01`000`L00`071`021`0300L000D70`031`0500L01`00
0PL00`07000205870P0010L01`0:1`0300L000<70P081`0300L700d710031`0500L01`001`L00`07
000<1`L02PL00`071`041`0500L71`0010L010071`021`0400L7008700@01`L0;`L003T70P0010L0
1`021`8000D700L700041`0500L71`001@L00`071`041`<00PL400870P081`0300L000L700<01`00
1`L200L700<01`000P031`0300L000@700@01`L00PL00`071`031`8000@700L70P021`0500L700L0
0P1C1`8000@700L02@L010071`051`0300L700@700<01`L02PL400H700H01`L01`071`0400L70107
10081`0300L700<700<01`L00PL00`071`031`0600L700L00PL010071`021`0300L702d7000i1`@0
0PL010071`L400H700@01`L70P031`0300L700D710061`0300L000L700@01`L01PL00`0700071`80
1`L200<70`000`L01`041`0400L7008700<01`L00PL00`0700021`0400L0008700<01`00101C1`@0
0PL00`071`051`0500L71`001PL200@700<01`L01PL400X700@01`L00PL00`071`041`0400L71`80
3@L00`0700021`<01PL00`071`021`0300L7008700<01`L010L300051`071`000PL200031`0702h7
000i1`<00`L00`071`031`0300L700D700<01`L00`L200031`0700<70`000`L000081`0300L000H7
00@01`L01PL010071`051`0400L700D700<01`001PL300@700<01`0010L01`071`L01`000`L20087
00<01`L01@1C1`<00`L00`071`041`0300L7008700<01`L01PL01@071`L000@7100=1`0500L71`00
0`L00`071`031`0300L700<700<01`L02PL010071`041`<010L010071`061`0300L700D70P031`07
00L01`07000_1`00>PL200@700H01`L700L200H700<01`L010L300031`0000D70P071`0400L700D7
00D01`L700051`0500L71`0010L010071`051`0300L000L700@01`000PL00`0700041`0400L70087
00<01`L00PL01`071`0700L0101D1`8010L00`071`021`0300L700<700<01`L01`L300031`000080
40L00`071`021`0300L7008700<01`L00PL00`071`041`0300L700T700@01`L01`L200<700<01`00
1PL00`071`061`0800L71`0700L302l7000j1`@00PL020071`L01`070P041`0300L700H70P071`04
00L000D700D01`L700051`0500L71`000`L00`071`021`0300L7008700D01`L700041`0300L000H7
0P031`8000<700L010L010071`021`0400L71`800PL0100700021`80E0L4008700D01`L700061`03
00L700T70P0C1`0300L700@700<01`L00PL01@071`L000L700<01`L01`L00`071`021`0300L700H7
1@071`0300L700D70P031`0300L000870P0_1`00>PL300071`001`0700031`0300L700800`L00`07
1`031`@01`L01@071`L000@700<01`L00PL00`071`021`0300L7008700@01`L01@L01@071`L000<7
00<01`L00PL010071`051`801@L200D700<01`000`L010071`L300031`0700800PL00`071`1C1`80
0PL200051`071`001PL00`071`071`D04@L00`071`051`0300L700<700@01`L020L00`071`061`03
00L700<700<01`L020L200P700<01`L010L3008700<01`000PL00`071`0^1`00>PL300<70P000`L0
1`021`0300L700870P000`L01`021`800`L300H700D01`L700041`0300L7008700D01`L700041`04
00L700H700@01`L00`L01@071`L000<700<01`L00PL00`0700051`@00`L00`0700031`0300L70080
00<700L00P001@L01`L005D70P041`8000<700L01PL00`071`061`0400L7008700<01`L03@L00`07
1`061`0300L700<700@01`L02@L00`071`051`0300L700<700<01`L020L400H700<01`L00`L01007
00021`0400L01`80<0L003/700<01`0010L00`071`031`0300L700<71@051`0400L700D700<01`L0
0PL01@071`L000D700@01`L01@L00`0700061`0400L700@700@01`L00`L00`071`021`0300L000D7
0P021`0300L700800`L010071`L200870`000`L01`0205H700<01`L00`L200P700<01`L010L200<7
00@01`L70P0<1`0300L700L700<01`L010L00`07000:1`0300L700<700<01`L01@L00`071`071`80
0PL00`071`031`0300L700<700@01`000PL200870P0`1`00>`L00`0700041`<010L00`071`041`03
00L700H700@01`L010L00`071`031`0400L700H700<01`001PL200L70P051`0400L700<700<01`L0
0PL00`0700041`0300L000<710031`0600L700L00PL2008700@01`00EPL00`07000300041`0000H7
00<01`L00`L00`071`031`0300L700<700<01`L020L00`071`081`0300L700D70P0;1`800`L00`07
1`051`0300L700H700<01`000`L00`071`021`0300L7008700L01`L01`07008000@70000<0L003/7
00@01`L00PL0100700L2008700<01`L00`L0100700061`0400L71`800`L00`071`031`<01`L200H7
0P071`801PL00`0700041`0400L7008700<01`L00PL00`0700051`0300L7008700H01`L01`031`03
00L700<0E`L200D70P000`L000041`0400L71`801PL00`071`041`0300L700L700<01`L020L00`07
1`021`P02PL010071`081`0300L700D700@01`L00`L200<700@01`L00PL01@071`07008000<70000
<@L003/70P001@L01`L0008700<01`L00P0010L01`L2008700@01`001@L00`071`021`0400L700L7
0P031aL01@L200D700@01`L00`L010071`021`0300L700870`031`0300L0008700<01`L00P021`<0
E`L200@700<01`000`L2008700@01`L020L00`071`051`801PL00`071`071`L010L2008720031`80
2@L00`071`041`0500L71`001@L010071`021`0400L700<71P0a1`00?0L400051`071`0010L20003
1`0700@700@01`L010L00`071`031`0300L000H71P041`801PL00`071`061`0300L000<00PL00`07
1`031`0300L000@700@01`L00PL00`071`021`8000T700L700L01`000`L0100700L305P700<01`L0
0PL00`0700051`D02@L00`071`071`0300L7008700<01`L00`L5008700<01`L01PL300T71P091`03
00L700<700D01`L700061`0500L01`000`L010071`L60347000l1`8000<7000010L00`071`031`03
00L700D700<01`L00P031`0300L700<70`000`L0000200870P061`801PL200L70P041`@010L00`07
00041`0300L000<700L01`L700L0008700@01`000PL00`071`0200031`0000<0F0L200<700<01`00
1`L00`071`091`0300L700P70P000`L01`0500L700<01`L01@L01@0700L000`700<01`000P071`05
00L71`0010L00`071`051`@00`L010071`L300031`000347000l1`800PL00`071`031`0600L71`07
0P051`0300L7008700@01`L01PL300H700<01`001@L00`0700051`801PL00`0700061`0300L000<0
00<700L010L00`0700031`0400L71`8010L3008700H01`0700L405T70P031`0400L700D700@01`00
2@L00`071`091`@03@L00`071`041`0500L01`002`L200@70P051`0500L71`0010L00`071`071`03
00L7008700<01`L00PL40387000m1`8000@700000`L01@071`L000870`041`0300L700870P000`L0
1`021`800PL00`071`051`0300L000D700<01`0010L00`0700051`0300L000D70P051`<010L200@7
00H01`L01`051`0600L71`070P031`<0FPL2008700@01`L01@L010071`L200L700<01`L01`L20003
1`0000h700<01`L010L01P071`0700X70P061`<00`L00`0700051`0300L700H70P041`0300L70087
00@01`00<PL003d70P0010L000L2008700<01`000`L010071`031`0300L700@70P000`L000041`80
1PL010071`041`0300L000@700<01`001@L010071`041`801@L200051`0000L00P041`0600L700L0
10L300870P000`L01`021`<0FPL2008700@01`L010L00`071`041`801@L00`071`041`<00`L00`07
000=1`0300L700@700@01`L00PL00`071`071`0300L000P710000`L01`051`0300L700@700@01`00
0`L010071`L200031`000387000m1`@000@700L70`000`L01`031`0300L700800PL00`071`041`80
1`L200D700D01`L700031`0400L700<700@01`L010L010071`041`801@L200@70P051`0300L00087
00@01`L70P000`L01`02008700<01`L00`1L1`<010L01@071`L000P70P031`0400L71`<01PL01007
1`0<1`0300L700@700@01`L00`L00`071`051`0400L700/70P071`0300L700@70P000`L01`021`04
00L71`@0<`L003h70`021`0400L71`<01@L010071`L200031`0700870P021`801@L00`0700041`05
00L71`000`L01@071`L0008700@01`L00`L01@071`L000<700<01`0010L00`0700041`@00`L00`07
00021`0400L71`800PL300041`071`@0G0L300@700@01`L02`L200031`0700802@L01@071`L000/7
00<01`L00`L00`071`021`0500L71`001PL010071`0:1`@01PL00`071`031`0300L0008700@01`L0
0PL200031`0003<7000n1`0300L0008700<01`L00PL400@700<01`L00PL200031`0000@700<01`00
1@L0100700031`0500L71`000PL01@071`L0008700D01`L700021`0500L71`000`L00`0700041`03
00L000@70P021`0300L700800`L00`071`0200@700D01`L01`0505d70P041`0400L700`7100;1`03
00L7008700<01`L020L00`071`031`0300L7008700<01`L00PL00`071`031`0400L700T700D01`L0
1`0200@700<01`L00PL010071`021`0500L700L01P0c1`00?`L400051`071`000PL0100700031`03
00L700<700<01`L01@L0100700031`0500L71`000PL00`071`021`0300L000@700<01`000`L01007
1`041`0300L000<700@01`L00PL010071`L200@710031`0500L01`000`L200041`071`@0G`L30087
00<01`001`L600031`0000802PL00`071`031`0300L700P700<01`L00PL00`071`021`0300L700<7
00D01`L700031`0300L700H700D01`L700031`0500L71`000`L01@071`L0008700<01`000PL503@7
000o1`<000H700L01`031`0300L7008000<700L00PL200031`0000D700D01`L700031`0700L71`07
00051`0300L000@70P041`0300L000@700<01`000`L010071`021`0600L700L01PL01`071`L01`00
0PL00`071`0300031`07008000<70000G`L500041`001`H01`L200041`0000P700<01`L010L00`07
1`071`0300L7008700<01`L00`L00`071`021`0500L71`000`L00`071`041`801@L01@071`L00087
00@01`L00`L00`071`021`8000@700070`0d1`00@0L0100700L300D700@01`L70`000`L01`031`<0
10L010071`L200031`0700<70P061`8010L200@70P051`0300L000<700<01`000PL01`071`L01`00
1@L3008700D01`0700021`800PL300041`00064700@01`07100<1`0300L000870`071`0300L700D7
00<01`L01PL00`071`021`0300L700<700<01`L00`L010071`031`0300L700<700<01`L01@L00`07
1`021`8000@700L00`L00`071`021`<000<700L00P0e1`00@0L200041`071`8010L00`071`031`80
1@L010071`041`0300L7008700<01`0010L200<74`031`0300L7008700<01`000PL010071`021`03
00L700870P021`@00PL01@0700L000870P000`L01`02064700<01`000`L00`071`0;1`0300L000@7
0`051`0300L700H700<01`L01@L01@071`L000H700<01`L010L200D700D01`L700081`0300L700@7
0`041`0300L700<70P000`L01`0203D700111`0300L0008710031`0300L7008700@01`0010L00`07
1`02008700<01`L00`L200<71P031`8010L00`071`031`0400L71`D00`L200<700H01`L01`041`03
00L000@700H01`L700L200<700<01`L00`1S1`0300L0008700<01`L02PL010071`051`0300L000@7
00<01`L01PL00`071`051`0500L71`001`L00`071`041`8010L0100700091`0300L700D700<01`L0
0`L00`071`031`0400L01`80=PL0044700T01`L01`0700L00`000`L01`021`0300L700<00`L01007
1`L200031`0700@710021`801@L200@70P041`0300L700@700D01`001`0200<700<01`000PL01`07
1`L01`0010L2008700D01`0700021`8000<700000P1S1`0400L7008700<01`L02@L010071`061`04
00L0008700<01`L01`L00`071`041`0500L71`001`L00`071`041`0300L0008000<700002`L00`07
1`051`0300L700<700<01`L00PL200031`070080=PL004870P000`L000021`0400L71`<000<700L0
0`L00`0700031`0300L700870P021`<000<7000010L200@700<01`000`L200<70P051`800`L300<7
0P031`0400L7008700@01`L70P000`L01`0200041`071`<000@70000I0L2008700<01`002PL01@07
1`L000L700<01`L00`0:1`0300L700@700@01`L01`L00`071`031`8010L200`700<01`L010L00`07
00031`0300L7008710021`0300L703D700121`D00`L00`071`021`<01@L0100700021`0300L700<7
0P051`800`L200@700<01`000`L200<700<01`0010L200<70P0010L000L200<700<01`000`L00`07
1`0200870`021`0500L000L0101V1`0700L71`0700091`0500L71`0020L010071`L200X700<01`L0
0`L00`0700091`0300L700870P051`802`L00`071`041`0300L000<700@01`L70`0010L01`0h1`00
@`L200031`0000<700<01`L00PL400<700@01`L70`031`8000<7000010L00`0700031`800`L01007
1`021`0300L0008700<01`000`L00`0700031`0300L700870P031`0300L000<700H01`L01`031`05
00L700L00P021`@0IPL00`071`021`0300L000L700<01`L00PL00`071`071`800PL400P700<01`L0
0PL00`0700091`0700L71`0700041`0400L700X700<01`L00`L01@071`L0008700D01`L01`020004
1`0703P700131`8000<700000P021`0400L7008700@01`000PL00`071`021`8000<700L00`L200@7
00<01`000PL00`0700031`0300L0008700<01`000PL00`0700031`0300L000870P041`@000<70000
10L01@0700L000<70P021`8000@700L70`1X1`0300L700870P061`0300L700@700<01`L020L00`07
1`0500L700L01`L700L000T700L01`L700L000@700@01`L70P091`0300L7008700D01`L700021`06
00L01`070P000`L01`0h1`00A0L300031`0700@000<700L00PL00`071`0200031`0700<70P051`03
00L000<700H01`L01`021`0400L7008700<01`000PL01@0700L0008700@01`L00PL200<70P031`80
10L00`0700021`0300L700<000D7000700021`<0J@L00`071`031`801@L00`071`041`0300L700T7
00<01`000PL500D700H01`L01`0:1`0600L700L010L00`071`031`0300L700H700<01`L00PL00`07
1`021`8000@700L00PL303X700151`<000<700L00PL300@70P021`<000<700000PL200@700@01`00
0PL01P071`07008700<01`000PL00`0700021`0500L01`000PL00`0700021`0300L0008700<01`00
0`L00`0700031`800PL00`071`0200870P0010L01`L306/70P031`0300L000<700<01`L01@L00`07
1`0:1`0300L000<70P0010L000031`0500L01`002PL00`0700021`0500L71`001`L00`071`051`05
00L71`001@L3008700<01`L00P0k1`00APL300031`0000871@031`0400L71`<010L300<700H01`L0
1`031`800PL00`0700031`800PL2008700<01`000PL01@0700L0008700<01`000`L2008710021`03
00L000800`L00`071`0606`70`021`800`L00`071`051`0300L700/700<01`0010L200870P021`03
00L000/700<01`000PL01@071`L000P700<01`L010L01@071`L000H700@01`L00PL303/700171`@0
00@700L00PL0100700021`0500L700L00`031`0300L000<700H01`L01`031`800PL200<70P031`04
00L71`800PL01@0700L0008700<01`000PL200D700@01`L70P000`L01`021`8000@700070`1^1`05
00L000L00P000`L01`071`0300L700`70`051`800PL500`700D01`L700031`0300L700L700<01`L0
0`L010071`051`<000<700L00PL203`700171`<000<700000P031`0300L700@000@700L710021`@0
0PL00`071`0200874@021`0400L71`800PL01@0700L0008700<01`000`L200031`0700800PL01@07
000700<000@700L0L0L00`071`0300031`0700L700<01`L03PL00`071`051a@00`L01@071`L000X7
00<01`L00PL010071`041`8000@700L00`L00`071`0k1`00B0L300870`031`800`L300@700@01`00
0PL00`07000200@70P021`800PL00`071`021`0600L71`0710021`800PL2008700D01`0700031`80
0PL00`0700021`0400L000870P0010L01`1b1`0400L71`<020L00`071`0?1`0300L700@00PL200l7
00D01`L700021`0300L700X700D01`L700021`0300L700<000@700L00PL203h700191`0500L000L0
1@021`0400L71`@00`L00`071`0200041`071`8000D7000700031`0300L700800`L01@071`L000<7
00<01`000`000`L000021`0500L01`000PL300@700D01`L01`0200<700@01`070P1b1`0300L70087
1@051`0300L700`71P081`0300L700`700D01`L700021`0300L700/70P021`800PL00`0700021`04
00L01`<0?PL004X70P000`L000021`0300L000D00`L3008700@01`L70`021`800`L200870P021`04
00L71`8000<700000PL2008700<01`L00P031`800PL01@0700L000<70`001`L0000700L01`1d1`04
00L71`8000<7000010000`L01`031`X00PL200041`0000L700<01`L030L00`0700031`0300L700d7
00P01`001`0700<70P0010L0000o1`00B`L500870P031`<00`L00`07000400<70`021`0400L71`80
0PL200031`0000870P000`L000021`0500L71`000PL00`070003008700D01`0700021`<00PL20087
00D01`001`0407H700<01`L00P031`800PL600/700D01`L700021`801PL200`700<01`000PL00`07
1`0?1`@000<700L00`L300031`0703l7001<1`0300L000800`L010071`L500031`0000870`0010L0
00L200870P021`8000@700070P001@L01`07008000<700000PL00`071`0200<70P021`8000<700L0
0`031`800PL400031`000080N0L0100700051`800PL900@70P051`0400L71`801PL00`071`091`03
00L0008700<01`L03`L2008700<01`L00PL50447001=1`/00`L400031`0700870`031`<000P70007
00L01`<000H7000700L200041`001`D00PL200870P021`<000@700L710000`L0000200870`1j1`80
00<700L01@L00`071`0500<71@071`0300L700870P051`802@L00`0700021`0300L700d70P0010L0
1`031`D0@`L004h70`0010L000L400@700<01`000`000`L01`0300<70P001PL000L01`8000<70000
0P021`0500L000L0100010L01`L200870`000`L000021`0300L000800PL40087101l1`0300L000L7
0P021`0300L700801@L300D700<01`L010L200D700<01`L01PL01@0700L000h70P031`X0A0L004l7
0P031`8000<700L010000`L000021`<00`L300871`000`L0000200061`071`070`001PL000L01`<0
00<700000PL200871P021`8000<700L00P000`L0001n1`<020L200031`0700870`041`<010L00`07
1`051`8010L00`071`061`8000<700L02`L200@70`021`D0A@L005071@0010L000021`L000<700L0
0`021`P000<700L03`021`<000D700001`0400031`0000<000<700001`000`L000211`<020L300D7
0`021`D00PL00`071`061`<00PL200H70P000`L01`091`800`L300031`07008000<70000A`L00547
2P021`800PL400871P001@L0000701400PL600041`0000870`000`L01`0400031`0000D0Q0L800@7
0`061`8000<700001@0;1`800PL00`071`031`8000<700L01PL300870`031`@000<70000B0L005<7
00@01`000PL800031`0700/000<700004`0010L000L400051`0000L01`021`8000<700000P271`03
00L700<720071`@000<70000100:1`8000<700000`L300H70P000`L0000200@70`031`80B@L005@7
1P021`<000<700000`021al000<700001P000`L00002008700<01`000`000`L0000208T70P091`T0
0`L300031`0000P01PL200041`071`<010L500<71@000`L000021`0300L704T7001E1`T000H70007
00L]00031`0700D00PL408d70P071`<000<700006@031`<00PL200872P051`8000<70000C0L005L7
0P021c<00`L500871P2@1`0300L700<700D01`L700031b4000<70000100:1`0300L70080C@L005T7
0`031`D000<7000090000`L0000>09<70P000`L000041`0300L700@79@091`0400L704l7001J1b00
00@7000760021`<0UPL[00870`031``0D0L005`71`000`L01`041`X000D700001`0300031`0000h0
0PL:09T70P021`8010LM008700<01`L00`L400H70`1B1`00GPLC00870`001@L0000700@000<70000
5@2L1b@00`L00`071`081`<00PL205@7001P1`<000<700003P000`L0000200051`0700L02`031`T0
X0L300@71P0010L000L400031`0000P01@L00`071`0:1`<0EPL006<71@000`L000021`D000<70000
0P021`0600L01`07502U1`D010L<008710061`0300L700L7101I1`00IPL700878P2Z1`X000@70007
20071`0300L700<71@1M1`00J0L800<700<01`000`001@L0000700<000<700002@2a1`D000<70000
0`L500870P000`L000031`0300L000@0H@L006`72`021`0300L000/0^`L400<71`021`800`L506@7
001a1a40a0LA06T7003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47
003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47
003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47
003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47
003o1l47003o1l47003o1l47001M1`801@L30<h70P051`<0P@L005l700<01`L01@L00`071`3=1`03
00L700D700<01`L0OPL004/70`051`800`L500870P041`<00`L200870`0010L000L300X700<01`L0
0PL4008710021`@00`L208H70`051`800`L500870P041`<00`L200870`0010L000L300X700<01`L0
0`L200<710031`80D@L004`700<01`L00`L010071`041`0500L71`000PL010071`021`0400L70087
00@01`L00PL00`071`031`0300L700h700D01`L700051`0300L700<700@01`L0QPL00`071`031`04
00L700@700D01`L700021`0400L7008700@01`L00PL010071`021`0300L700<700<01`L030L01007
1`021`0300L700<700@01`L0D0L004`70`031`0400L700@700D01`L700021`0400L7008700@01`L0
0PL010071`021`0400L71`804@L00`071`021`0300L700<700<01`L00PL010071`261`<00`L01007
1`041`0500L71`000PL010071`021`0400L7008700@01`L00PL010071`L200l700@01`L00`L00`07
1`021`0400L70507001<1`0400L7008700@01`L010L01@071`L0008700@01`L00PL010071`021`04
00L7008700@01`L030L00`071`041`0300L700<700<01`L00`L01@071`L0008700<01`L0Q0L01007
1`021`0400L700@700D01`L700021`0400L7008700@01`L00PL010071`021`0400L700`700<01`L0
0PL300D700<01`L00PL20547001<1`0400L700<70P051`0300L7008000<700000`L200041`071`80
0PL0100700041`<040L00`071`041`0300L700<700@01`L00PL00`071`241`0400L700<70P051`03
00L7008000<700000`L200041`071`800PL0100700041`<03PL00`071`061`0400L7008700<01`L0
CPL004/7100;1`0300L702/70`031`0400L7008700@01`L00PL010071`251`@02`L00`071`0[1`03
00L700<700@01`L00PL010071`1@1`00F@L202l700<01`L00PL200@70P041`80U0L202h70`031`80
10L20547003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47003o1l47
0000\
\>"],
  ImageRangeCache->{{{0, 447.625}, {483.062, 0}} -> {-0.0500105, -0.0541278, 
  0.00469148, 0.00469148}, {{10.625, 223.812}, {471.5, 
  241.5}} -> {-0.0238588, -0.049488, 0.00427964, 0.00427964}, {{223.812, 
  436.938}, {471.5, 241.5}} -> {-0.937015, -0.0494825, 0.00427916, 
  0.00427916}, {{10.625, 223.812}, {241.5, 11.5}} -> {-0.0233385, -1.03183, 
  0.00427145, 0.00427145}, {{223.812, 436.938}, {241.5, 
  11.5}} -> {-0.934605, -1.03242, 0.00427393, 0.00427393}},
  CellTags->"PolygonReduction"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Closing the OpenGL Viewer", "Section"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MVClose[]\)], "Input",
  CellLabel->"In[27]:="],

Cell[BoxData[
    \("/export/home/kuska/MathView3D/mathview3d -mathlink"\)], "Output",
  CellLabel->"Out[27]="]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Acknowledgement", "Section"],

Cell["\<\
I would like to thank the people at Wolfram Research for the \
Visiting Scholar Grant for the port of MathGL3d to Macintosh computers. \
\>",
   "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Copyright", "Section"],

Cell[TextData[{
  "The files in the MathGL3d distribution may be freely copied and \
distributed, provided that no changes whatsoever are made. All users are \
asked to help keep the ",
  StyleBox["MathView3D.w",
    FontFamily->"Courier"],
  " and ",
  StyleBox["savepng.w",
    FontFamily->"Courier"],
  " file consistent and \"uncorrupted',' identical everywhere in the world.  \
(The CWEB system has a ``change file'' facility by which users can easily \
make minor alterations without modifying the master source files in any way. \
Everybody is encouraged to use change files instead of changing the files.)\n\
The author has tried his best to produce correct and useful programs, in \
order to help promote computer science research and ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", but no warranty of any kind should be assumed."
}], "Text"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowToolbars->"RulerBar",
WindowSize->{728, 673},
WindowMargins->{{50, Automatic}, {85, Automatic}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PrintingMargins"->{{36, 36}, {72, 72}},
"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PrintCellBrackets"->False,
"PrintRegistrationMarks"->False,
"PrintMultipleHorizontalPages"->False,
"Magnification"->1},
Magnification->1,
StyleDefinitions -> "ArticleClassic.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "WindowCommand"->{
    Cell[8102, 228, 89, 1, 55, "Section",
      CellTags->"WindowCommand"],
    Cell[8194, 231, 365, 11, 27, "Text",
      CellTags->"WindowCommand"],
    Cell[8562, 244, 668, 18, 103, "Text",
      CellTags->"WindowCommand"],
    Cell[9233, 264, 1114, 32, 232, "Text",
      CellTags->"WindowCommand"],
    Cell[10350, 298, 1284, 37, 90, "Text",
      CellTags->"WindowCommand"],
    Cell[11637, 337, 533, 17, 158, "Text",
      CellTags->"WindowCommand"],
    Cell[12173, 356, 826, 18, 74, "Text",
      CellTags->"WindowCommand"],
    Cell[13002, 376, 424, 15, 140, "Text",
      CellTags->"WindowCommand"],
    Cell[13429, 393, 199, 5, 42, "Text",
      CellTags->"WindowCommand"],
    Cell[13631, 400, 300, 10, 86, "Text",
      CellTags->"WindowCommand"],
    Cell[13934, 412, 372, 7, 58, "Text",
      CellTags->"WindowCommand"],
    Cell[14309, 421, 406, 13, 104, "Text",
      CellTags->"WindowCommand"],
    Cell[14718, 436, 316, 12, 26, "Text",
      CellTags->"WindowCommand"],
    Cell[15037, 450, 164, 6, 32, "Text",
      CellTags->"WindowCommand"],
    Cell[15204, 458, 322, 9, 42, "Text",
      CellTags->"WindowCommand"]},
  "MathLinkMode"->{
    Cell[15563, 472, 85, 1, 29, "Section",
      CellTags->"MathLinkMode"],
    Cell[15651, 475, 789, 21, 74, "Text",
      CellTags->"MathLinkMode"],
    Cell[16465, 500, 118, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[16586, 505, 640, 13, 71, "Print",
      CellTags->"MathLinkMode"],
    Cell[17241, 521, 383, 11, 42, "Text",
      CellTags->"MathLinkMode"],
    Cell[17627, 534, 103, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[17733, 539, 4991, 122, 167, "Input",
      CellTags->"MathLinkMode"],
    Cell[22749, 665, 107, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[22859, 670, 3286, 88, 316, "Output",
      CellTags->"MathLinkMode"],
    Cell[26160, 761, 695, 19, 74, "Text",
      CellTags->"MathLinkMode"],
    Cell[26880, 784, 111, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[26994, 789, 1150, 18, 188, "Output",
      CellTags->"MathLinkMode"],
    Cell[28159, 810, 467, 15, 42, "Text",
      CellTags->"MathLinkMode"],
    Cell[28629, 827, 326, 9, 42, "Text",
      CellTags->"MathLinkMode"],
    Cell[28958, 838, 1266, 21, 300, "Input",
      CellTags->"MathLinkMode"],
    Cell[30249, 863, 144, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[30396, 868, 419, 9, 44, "Output",
      CellTags->"MathLinkMode"],
    Cell[30852, 882, 108, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[30963, 887, 584515, 7959, 295, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[615493, 8849, 460, 12, 133, "Input",
      CellTags->"MathLinkMode"],
    Cell[615956, 8863, 130, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[616089, 8868, 381, 13, 42, "Text",
      CellTags->"MathLinkMode"],
    Cell[616495, 8885, 109, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[616607, 8890, 595204, 8091, 295, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[1211826, 16984, 153, 6, 26, "Text",
      CellTags->"MathLinkMode"],
    Cell[1211982, 16992, 117, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[1212124, 16999, 219, 5, 108, "Input",
      CellTags->"MathLinkMode"],
    Cell[1212346, 17006, 579105, 7892, 295, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[1791466, 24901, 244, 6, 41, "Text",
      CellTags->"MathLinkMode"],
    Cell[1791713, 24909, 180, 5, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[1791918, 24918, 182, 5, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[1792103, 24925, 395, 9, 44, "Output",
      CellTags->"MathLinkMode"],
    Cell[1792535, 24939, 109, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[1792647, 24944, 622545, 8428, 295, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[2415207, 33375, 1040, 32, 90, "Text",
      CellTags->"MathLinkMode"],
    Cell[2416272, 33411, 115, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[2416390, 33416, 80582, 2549, 274, 37785, 2015, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[2497009, 35970, 82, 1, 41, "Subsection",
      CellTags->"MathLinkMode"],
    Cell[2497094, 35973, 616, 20, 58, "Text",
      CellTags->"MathLinkMode"],
    Cell[2497735, 35997, 343, 7, 108, "Input",
      CellTags->"MathLinkMode"],
    Cell[2498081, 36006, 24830, 857, 252, 8284, 647, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[2522926, 36866, 119, 4, 26, "Text",
      CellTags->"MathLinkMode"],
    Cell[2523048, 36872, 180, 5, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[2523253, 36881, 84, 2, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[2523340, 36885, 1009910, 14010, 295, 981188, 13650, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[3533265, 50898, 100, 1, 26, "Text",
      CellTags->"MathLinkMode"],
    Cell[3533390, 50903, 362, 7, 76, "Input",
      CellTags->"MathLinkMode"],
    Cell[3533755, 50912, 23571, 743, 257, 6609, 528, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[3557341, 51658, 139, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[3557505, 51665, 109, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[3557617, 51670, 587416, 7995, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[4145082, 59671, 56, 1, 41, "Subsection",
      CellTags->"MathLinkMode"],
    Cell[4145141, 59674, 774, 18, 74, "Text",
      CellTags->"MathLinkMode"],
    Cell[4145940, 59696, 602, 11, 252, "Input",
      CellTags->"MathLinkMode"],
    Cell[4146545, 59709, 238, 5, 37, "Message",
      CellTags->"MathLinkMode"],
    Cell[4146798, 59717, 487, 9, 131, "Input",
      CellTags->"MathLinkMode"],
    Cell[4147288, 59728, 319, 12, 26, "Text",
      CellTags->"MathLinkMode"],
    Cell[4147632, 59744, 393, 9, 76, "Input",
      CellTags->"MathLinkMode"],
    Cell[4148028, 59755, 369339, 19528, 228, 334652, 19094, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[4517382, 79286, 119, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[4517504, 79291, 140, 4, 26, "Text",
      CellTags->"MathLinkMode"],
    Cell[4517669, 79299, 109, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[4517781, 79304, 576356, 7903, 296, 548285, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[5094152, 87210, 260, 9, 26, "Text",
      CellTags->"MathLinkMode"],
    Cell[5094415, 87221, 254, 6, 51, "Input",
      CellTags->"MathLinkMode"],
    Cell[5094672, 87229, 114, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[5094811, 87236, 206, 5, 44, "Input",
      CellTags->"MathLinkMode"],
    Cell[5095020, 87243, 614912, 9747, 207, 579418, 9303, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[5709969, 96995, 488, 10, 76, "Input",
      CellTags->"MathLinkMode"],
    Cell[5710460, 97007, 310189, 15595, 263, 275493, 15161, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[6020652, 112604, 106, 3, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[6020773, 112610, 119, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[6020917, 112617, 109, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[6021029, 112622, 580182, 7905, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[6601260, 120533, 58, 1, 41, "Subsection",
      CellTags->"MathLinkMode"],
    Cell[6601321, 120536, 376, 12, 42, "Text",
      CellTags->"MathLinkMode"],
    Cell[6601700, 120550, 1310, 41, 90, "Text",
      CellTags->"MathLinkMode"],
    Cell[6603035, 120595, 86, 1, 32, "Subsubsection",
      CellTags->"MathLinkMode"],
    Cell[6603124, 120598, 115, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[6603264, 120605, 431, 10, 156, "Input",
      CellTags->"MathLinkMode"],
    Cell[6603698, 120617, 182, 4, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[6603917, 120626, 102, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[6604022, 120631, 594555, 8083, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[7198580, 128716, 182, 5, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[7198777, 128724, 731, 21, 62, "Text",
      CellTags->"MathLinkMode"],
    Cell[7199533, 128749, 471, 11, 188, "Input",
      CellTags->"MathLinkMode"],
    Cell[7200007, 128762, 185, 4, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[7200229, 128771, 102, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[7200334, 128776, 600529, 8157, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[7800866, 136935, 182, 5, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[7801063, 136943, 657, 18, 58, "Text",
      CellTags->"MathLinkMode"],
    Cell[7801745, 136965, 233, 5, 108, "Input",
      CellTags->"MathLinkMode"],
    Cell[7801981, 136972, 183, 4, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[7802201, 136981, 108, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[7802312, 136986, 582267, 7931, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[8384594, 144920, 259, 7, 42, "Text",
      CellTags->"MathLinkMode"],
    Cell[8384856, 144929, 288, 7, 53, "DisplayFormula",
      CellTags->"MathLinkMode"],
    Cell[8385169, 144940, 284, 7, 108, "Input",
      CellTags->"MathLinkMode"],
    Cell[8385456, 144949, 182, 4, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[8385675, 144958, 108, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[8385786, 144963, 582267, 7931, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[8968102, 152900, 154, 6, 32, "Subsubsection",
      CellTags->"MathLinkMode"],
    Cell[8968259, 152908, 1079, 32, 74, "Text",
      CellTags->"MathLinkMode"],
    Cell[8969363, 152944, 349, 8, 60, "Input",
      CellTags->"MathLinkMode"],
    Cell[8969715, 152954, 106, 3, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[8969858, 152962, 231, 6, 44, "Input",
      CellTags->"MathLinkMode"],
    Cell[8970092, 152970, 184, 4, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[8970313, 152979, 109, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[8970425, 152984, 600384, 8155, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[9570824, 161142, 159, 6, 26, "Text",
      CellTags->"MathLinkMode"],
    Cell[9571008, 161152, 268, 7, 60, "Input",
      CellTags->"MathLinkMode"],
    Cell[9571279, 161161, 186, 4, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[9571502, 161170, 109, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[9571614, 161175, 582232, 7931, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[10153861, 169109, 327, 8, 60, "Input",
      CellTags->"MathLinkMode"],
    Cell[10154213, 169121, 514, 10, 92, "Input",
      CellTags->"MathLinkMode"],
    Cell[10154730, 169133, 4944, 137, 144, 1135, 85, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[10159677, 169272, 183, 5, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[10159875, 169280, 1074, 28, 106, "Text",
      CellTags->"MathLinkMode"],
    Cell[10160974, 169312, 215, 6, 44, "Input",
      CellTags->"MathLinkMode"],
    Cell[10161192, 169320, 184, 4, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[10161413, 169329, 103, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[10161519, 169334, 593802, 8073, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[10755324, 177409, 183, 5, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[10755522, 177417, 262, 7, 42, "Text",
      CellTags->"MathLinkMode"],
    Cell[10755787, 177426, 205, 5, 44, "Input",
      CellTags->"MathLinkMode"],
    Cell[10756017, 177435, 296, 7, 44, "Input",
      CellTags->"MathLinkMode"],
    Cell[10756316, 177444, 142663, 6212, 298, 108146, 5780, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"],
    Cell[10898982, 183658, 105, 3, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[10899124, 183666, 486, 11, 76, "Input",
      CellTags->"MathLinkMode"],
    Cell[10899613, 183679, 183, 4, 28, "Output",
      CellTags->"MathLinkMode"],
    Cell[10899833, 183688, 109, 3, 28, "Input",
      CellTags->"MathLinkMode"],
    Cell[10899945, 183693, 624408, 8451, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"MathLinkMode"]},
  "POVRay"->{
    Cell[11524426, 192152, 97, 1, 29, "Section",
      CellTags->"POVRay"],
    Cell[11524526, 192155, 875, 22, 74, "Text",
      CellTags->"POVRay"],
    Cell[11525426, 192181, 222, 5, 44, "Input",
      CellTags->"POVRay"],
    Cell[11525651, 192188, 119, 3, 28, "Output",
      CellTags->"POVRay"],
    Cell[11525785, 192194, 267, 7, 42, "Text",
      CellTags->"POVRay"],
    Cell[11526077, 192205, 865, 18, 236, "Input",
      CellTags->"POVRay"],
    Cell[11526945, 192225, 385107, 17832, 296, 332772, 17180, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"POVRay"],
    Cell[11912067, 210060, 138, 3, 28, "Input",
      CellTags->"POVRay"],
    Cell[11912230, 210067, 103, 3, 28, "Input",
      CellTags->"POVRay"],
    Cell[11912336, 210072, 621648, 8417, 296, 551885, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"POVRay"],
    Cell[12534021, 218494, 129, 3, 28, "Input",
      CellTags->"POVRay"],
    Cell[12534153, 218499, 114, 3, 28, "Output",
      CellTags->"POVRay"],
    Cell[12534282, 218505, 561, 9, 74, "Text",
      CellTags->"POVRay"],
    Cell[12534868, 218518, 118, 3, 28, "Input",
      CellTags->"POVRay"],
    Cell[12534989, 218523, 102, 3, 28, "Output",
      CellTags->"POVRay"]},
  "DXFFiles"->{
    Cell[12535140, 218532, 60, 1, 29, "Section",
      CellTags->"DXFFiles"],
    Cell[12535203, 218535, 461, 12, 58, "Text",
      CellTags->"DXFFiles"],
    Cell[12535689, 218551, 162, 4, 28, "Input",
      CellTags->"DXFFiles"],
    Cell[12535854, 218557, 104, 3, 28, "Output",
      CellTags->"DXFFiles"],
    Cell[12535973, 218563, 638, 17, 74, "Text",
      CellTags->"DXFFiles"],
    Cell[12536648, 218585, 60, 1, 29, "Section",
      CellTags->"DXFFiles"],
    Cell[12536711, 218588, 145, 4, 26, "Text",
      CellTags->"DXFFiles"],
    Cell[12536859, 218594, 123, 3, 28, "Input",
      CellTags->"DXFFiles"],
    Cell[12536985, 218599, 112, 3, 28, "Input",
      CellTags->"DXFFiles"],
    Cell[12537122, 218606, 138, 3, 28, "Input",
      CellTags->"DXFFiles"],
    Cell[12537263, 218611, 1093814, 49150, 69, 1073882, 48898, 
    "GraphicsData", "PostScript", "Graphics",
      CellTags->"DXFFiles"],
    Cell[13631114, 267766, 155, 4, 28, "Input",
      CellTags->"DXFFiles"],
    Cell[13631272, 267772, 125, 3, 28, "Output",
      CellTags->"DXFFiles"],
    Cell[13631434, 267780, 146, 4, 28, "Input",
      CellTags->"DXFFiles"],
    Cell[13631583, 267786, 125, 3, 28, "Output",
      CellTags->"DXFFiles"],
    Cell[13631723, 267792, 218, 8, 26, "Text",
      CellTags->"DXFFiles"],
    Cell[13631944, 267802, 183, 5, 44, "Input",
      CellTags->"DXFFiles"],
    Cell[13632152, 267811, 237, 5, 44, "Input",
      CellTags->"DXFFiles"],
    Cell[13632392, 267818, 342240, 9475, 296, 301971, 8972, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"DXFFiles"],
    Cell[13974647, 277296, 75, 1, 26, "Text",
      CellTags->"DXFFiles"],
    Cell[13974725, 277299, 151, 4, 44, "Input",
      CellTags->"DXFFiles"],
    Cell[13974901, 277307, 229, 5, 44, "Input",
      CellTags->"DXFFiles"],
    Cell[13975133, 277314, 298534, 12413, 250, 264468, 11987, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"DXFFiles"],
    Cell[14273682, 289730, 349, 7, 76, "Input",
      CellTags->"DXFFiles"],
    Cell[14274056, 289741, 287, 6, 44, "Input",
      CellTags->"DXFFiles"],
    Cell[14274346, 289749, 334623, 13411, 317, 283802, 12778, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"DXFFiles"]},
  "QuickDraw3D"->{
    Cell[15671610, 345329, 75, 1, 29, "Section",
      CellTags->"QuickDraw3D"],
    Cell[15671688, 345332, 509, 11, 58, "Text",
      CellTags->"QuickDraw3D"],
    Cell[15672222, 345347, 124, 3, 28, "Input",
      CellTags->"QuickDraw3D"],
    Cell[15672349, 345352, 98, 3, 28, "Output",
      CellTags->"QuickDraw3D"]},
  "VRML"->{
    Cell[15672496, 345361, 61, 1, 29, "Section",
      CellTags->"VRML"],
    Cell[15672560, 345364, 241, 7, 42, "Text",
      CellTags->"VRML"],
    Cell[15672826, 345375, 97, 3, 28, "Input",
      CellTags->"VRML"],
    Cell[15672926, 345380, 176, 4, 28, "Output",
      CellTags->"VRML"],
    Cell[15673117, 345387, 592, 16, 74, "Text",
      CellTags->"VRML"]},
  "PostScript"->{
    Cell[15673746, 345408, 69, 1, 29, "Section",
      CellTags->"PostScript"],
    Cell[15673818, 345411, 698, 16, 74, "Text",
      CellTags->"PostScript"],
    Cell[15674519, 345429, 115, 3, 28, "Input",
      CellTags->"PostScript"],
    Cell[15674659, 345436, 113, 3, 28, "Input",
      CellTags->"PostScript"],
    Cell[15674775, 345441, 98, 3, 28, "Output",
      CellTags->"PostScript"],
    Cell[15674888, 345447, 562, 15, 58, "Text",
      CellTags->"PostScript"]},
  "PNGBitmaps"->{
    Cell[15675487, 345467, 86, 1, 29, "Section",
      CellTags->"PNGBitmaps"],
    Cell[15675576, 345470, 1649, 46, 122, "Text",
      CellTags->"PNGBitmaps"]},
  "Animation"->{
    Cell[15677262, 345521, 54, 1, 29, "Section",
      CellTags->"Animation"],
    Cell[15677319, 345524, 1028, 29, 90, "Text",
      CellTags->"Animation"],
    Cell[15678350, 345555, 115, 3, 28, "Input",
      CellTags->"Animation"],
    Cell[15678468, 345560, 100, 1, 26, "Text",
      CellTags->"Animation"],
    Cell[15678571, 345563, 345, 8, 108, "Input",
      CellTags->"Animation"],
    Cell[15678919, 345573, 97, 1, 26, "Text",
      CellTags->"Animation"],
    Cell[15679019, 345576, 114, 3, 28, "Input",
      CellTags->"Animation"],
    Cell[15679136, 345581, 414, 11, 58, "Text",
      CellTags->"Animation"],
    Cell[15679553, 345594, 345, 6, 156, "Input",
      CellTags->"Animation"],
    Cell[15679923, 345604, 215, 6, 28, "Input",
      CellTags->"Animation"],
    Cell[15680163, 345614, 111968, 1606, 136, 100521, 1459, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"Animation"],
    Cell[15792134, 347222, 111818, 1604, 136, 100521, 1459, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"Animation"],
    Cell[15903955, 348828, 111619, 1601, 136, 100521, 1459, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"Animation"],
    Cell[16015577, 350431, 111862, 1604, 136, 100521, 1459, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"Animation"],
    Cell[16127442, 352037, 111587, 1601, 136, 100521, 1459, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"Animation"],
    Cell[16239032, 353640, 111603, 1601, 136, 100521, 1459, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"Animation"],
    Cell[16350638, 355243, 112053, 1607, 136, 100521, 1459, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"Animation"],
    Cell[16462718, 356854, 470, 11, 56, "Text",
      CellTags->"Animation"],
    Cell[16463191, 356867, 3492, 76, 69, "Input",
      CellTags->"Animation"],
    Cell[16466686, 356945, 623, 19, 42, "Text",
      CellTags->"Animation"],
    Cell[16467334, 356968, 499, 10, 60, "Input",
      CellTags->"Animation"],
    Cell[16468801, 357009, 145, 4, 26, "Text",
      CellTags->"Animation"],
    Cell[16468971, 357017, 2276, 40, 444, "Input",
      CellTags->"Animation"],
    Cell[16471439, 357065, 176, 6, 26, "Text",
      CellTags->"Animation"],
    Cell[16471640, 357075, 227, 5, 28, "Input",
      CellTags->"Animation"],
    Cell[16483085, 357660, 451, 11, 58, "Text",
      CellTags->"Animation"],
    Cell[16483539, 357673, 919, 18, 172, "Input",
      CellTags->"Animation"],
    Cell[16484461, 357693, 118, 4, 26, "Text",
      CellTags->"Animation"],
    Cell[16484604, 357701, 172, 5, 76, "Input",
      CellTags->"Animation"],
    Cell[16484779, 357708, 560173, 7703, 296, 548285, 7550, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"Animation"],
    Cell[17044967, 365414, 162, 6, 26, "Text",
      CellTags->"Animation"],
    Cell[17045132, 365422, 210, 5, 28, "Input",
      CellTags->"Animation"],
    Cell[17045345, 365429, 544, 14, 42, "Text",
      CellTags->"Animation"],
    Cell[17045892, 365445, 181, 5, 28, "Input",
      CellTags->"Animation"],
    Cell[17046076, 365452, 250, 5, 42, "Text",
      CellTags->"Animation"]},
  "PolygonReduction"->{
    Cell[17046363, 365462, 73, 1, 29, "Section",
      CellTags->"PolygonReduction"],
    Cell[17046439, 365465, 229, 5, 42, "Text",
      CellTags->"PolygonReduction"],
    Cell[17046671, 365472, 119, 3, 28, "Input",
      CellTags->"PolygonReduction"],
    Cell[17046793, 365477, 151, 4, 28, "Input",
      CellTags->"PolygonReduction"],
    Cell[17046947, 365483, 130, 3, 28, "Input",
      CellTags->"PolygonReduction"],
    Cell[17047080, 365488, 456, 11, 58, "Text",
      CellTags->"PolygonReduction"],
    Cell[17047539, 365501, 201, 5, 28, "Input",
      CellTags->"PolygonReduction"],
    Cell[17047743, 365508, 892, 20, 90, "Text",
      CellTags->"PolygonReduction"],
    Cell[17048660, 365532, 484, 12, 76, "Input",
      CellTags->"PolygonReduction"],
    Cell[17049147, 365546, 151667, 5619, 493, 90392, 4856, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"PolygonReduction"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"WindowCommand", 17203296, 371250},
  {"MathLinkMode", 17204457, 371281},
  {"POVRay", 17216461, 371572},
  {"DXFFiles", 17217727, 371605},
  {"QuickDraw3D", 17219993, 371662},
  {"VRML", 17220327, 371671},
  {"PostScript", 17220711, 371682},
  {"PNGBitmaps", 17221205, 371695},
  {"Animation", 17221384, 371700},
  {"PolygonReduction", 17224615, 371779}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 124, 4, 112, "Title"],
Cell[1858, 57, 33, 0, 39, "Author"],
Cell[1894, 59, 58, 0, 20, "Address"],
Cell[1955, 61, 565, 13, 90, "Abstract"],
Cell[2523, 76, 642, 18, 58, "Text"],
Cell[3168, 96, 557, 9, 90, "Text"],
Cell[3728, 107, 779, 22, 74, "Text"],
Cell[4510, 131, 1528, 31, 218, "Text"],

Cell[CellGroupData[{
Cell[6063, 166, 46, 0, 41, "Subsection"],
Cell[6112, 168, 87, 2, 26, "Text"],
Cell[6202, 172, 548, 10, 74, "Text"],
Cell[6753, 184, 823, 23, 106, "Text"],
Cell[7579, 209, 176, 4, 42, "Text"],
Cell[7758, 215, 307, 8, 42, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8102, 228, 89, 1, 55, "Section",
  CellTags->"WindowCommand"],
Cell[8194, 231, 365, 11, 27, "Text",
  CellTags->"WindowCommand"],
Cell[8562, 244, 668, 18, 103, "Text",
  CellTags->"WindowCommand"],
Cell[9233, 264, 1114, 32, 232, "Text",
  CellTags->"WindowCommand"],
Cell[10350, 298, 1284, 37, 90, "Text",
  CellTags->"WindowCommand"],
Cell[11637, 337, 533, 17, 158, "Text",
  CellTags->"WindowCommand"],
Cell[12173, 356, 826, 18, 74, "Text",
  CellTags->"WindowCommand"],
Cell[13002, 376, 424, 15, 140, "Text",
  CellTags->"WindowCommand"],
Cell[13429, 393, 199, 5, 42, "Text",
  CellTags->"WindowCommand"],
Cell[13631, 400, 300, 10, 86, "Text",
  CellTags->"WindowCommand"],
Cell[13934, 412, 372, 7, 58, "Text",
  CellTags->"WindowCommand"],
Cell[14309, 421, 406, 13, 104, "Text",
  CellTags->"WindowCommand"],
Cell[14718, 436, 316, 12, 26, "Text",
  CellTags->"WindowCommand"],
Cell[15037, 450, 164, 6, 32, "Text",
  CellTags->"WindowCommand"],
Cell[15204, 458, 322, 9, 42, "Text",
  CellTags->"WindowCommand"]
}, Closed]],

Cell[CellGroupData[{
Cell[15563, 472, 85, 1, 29, "Section",
  CellTags->"MathLinkMode"],
Cell[15651, 475, 789, 21, 74, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[16465, 500, 118, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[16586, 505, 640, 13, 71, "Print",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[17241, 521, 383, 11, 42, "Text",
  CellTags->"MathLinkMode"],
Cell[17627, 534, 103, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[17733, 539, 4991, 122, 167, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[22749, 665, 107, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[22859, 670, 3286, 88, 316, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[26160, 761, 695, 19, 74, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[26880, 784, 111, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[26994, 789, 1150, 18, 188, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[28159, 810, 467, 15, 42, "Text",
  CellTags->"MathLinkMode"],
Cell[28629, 827, 326, 9, 42, "Text",
  CellTags->"MathLinkMode"],
Cell[28958, 838, 1266, 21, 300, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[30249, 863, 144, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[30396, 868, 419, 9, 44, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[30852, 882, 108, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[30963, 887, 584515, 7959, 295, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[615493, 8849, 460, 12, 133, "Input",
  CellTags->"MathLinkMode"],
Cell[615956, 8863, 130, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[616089, 8868, 381, 13, 42, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[616495, 8885, 109, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[616607, 8890, 595204, 8091, 295, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[1211826, 16984, 153, 6, 26, "Text",
  CellTags->"MathLinkMode"],
Cell[1211982, 16992, 117, 3, 28, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[1212124, 16999, 219, 5, 108, "Input",
  CellTags->"MathLinkMode"],
Cell[1212346, 17006, 579105, 7892, 295, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[1791466, 24901, 244, 6, 41, "Text",
  CellTags->"MathLinkMode"],
Cell[1791713, 24909, 180, 5, 28, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[1791918, 24918, 182, 5, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[1792103, 24925, 395, 9, 44, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[1792535, 24939, 109, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[1792647, 24944, 622545, 8428, 295, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[2415207, 33375, 1040, 32, 90, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[2416272, 33411, 115, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[2416390, 33416, 80582, 2549, 274, 37785, 2015, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2497009, 35970, 82, 1, 41, "Subsection",
  CellTags->"MathLinkMode"],
Cell[2497094, 35973, 616, 20, 58, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[2497735, 35997, 343, 7, 108, "Input",
  CellTags->"MathLinkMode"],
Cell[2498081, 36006, 24830, 857, 252, 8284, 647, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[2522926, 36866, 119, 4, 26, "Text",
  CellTags->"MathLinkMode"],
Cell[2523048, 36872, 180, 5, 28, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[2523253, 36881, 84, 2, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[2523340, 36885, 1009910, 14010, 295, 981188, 13650, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[3533265, 50898, 100, 1, 26, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[3533390, 50903, 362, 7, 76, "Input",
  CellTags->"MathLinkMode"],
Cell[3533755, 50912, 23571, 743, 257, 6609, 528, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[3557341, 51658, 139, 3, 28, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[3557505, 51665, 109, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[3557617, 51670, 587416, 7995, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[4145082, 59671, 56, 1, 41, "Subsection",
  CellTags->"MathLinkMode"],
Cell[4145141, 59674, 774, 18, 74, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[4145940, 59696, 602, 11, 252, "Input",
  CellTags->"MathLinkMode"],
Cell[4146545, 59709, 238, 5, 37, "Message",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[4146798, 59717, 487, 9, 131, "Input",
  CellTags->"MathLinkMode"],
Cell[4147288, 59728, 319, 12, 26, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[4147632, 59744, 393, 9, 76, "Input",
  CellTags->"MathLinkMode"],
Cell[4148028, 59755, 369339, 19528, 228, 334652, 19094, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[4517382, 79286, 119, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[4517504, 79291, 140, 4, 26, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[4517669, 79299, 109, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[4517781, 79304, 576356, 7903, 296, 548285, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[5094152, 87210, 260, 9, 26, "Text",
  CellTags->"MathLinkMode"],
Cell[5094415, 87221, 254, 6, 51, "Input",
  CellTags->"MathLinkMode"],
Cell[5094672, 87229, 114, 3, 28, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[5094811, 87236, 206, 5, 44, "Input",
  CellTags->"MathLinkMode"],
Cell[5095020, 87243, 614912, 9747, 207, 579418, 9303, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5709969, 96995, 488, 10, 76, "Input",
  CellTags->"MathLinkMode"],
Cell[5710460, 97007, 310189, 15595, 263, 275493, 15161, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"],
Cell[6020652, 112604, 106, 3, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[6020773, 112610, 119, 3, 28, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[6020917, 112617, 109, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[6021029, 112622, 580182, 7905, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[6601260, 120533, 58, 1, 41, "Subsection",
  CellTags->"MathLinkMode"],
Cell[6601321, 120536, 376, 12, 42, "Text",
  CellTags->"MathLinkMode"],
Cell[6601700, 120550, 1310, 41, 90, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[6603035, 120595, 86, 1, 32, "Subsubsection",
  CellTags->"MathLinkMode"],
Cell[6603124, 120598, 115, 3, 28, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[6603264, 120605, 431, 10, 156, "Input",
  CellTags->"MathLinkMode"],
Cell[6603698, 120617, 182, 4, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6603917, 120626, 102, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[6604022, 120631, 594555, 8083, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"],
Cell[7198580, 128716, 182, 5, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[7198777, 128724, 731, 21, 62, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[7199533, 128749, 471, 11, 188, "Input",
  CellTags->"MathLinkMode"],
Cell[7200007, 128762, 185, 4, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7200229, 128771, 102, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[7200334, 128776, 600529, 8157, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"],
Cell[7800866, 136935, 182, 5, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[7801063, 136943, 657, 18, 58, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[7801745, 136965, 233, 5, 108, "Input",
  CellTags->"MathLinkMode"],
Cell[7801981, 136972, 183, 4, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7802201, 136981, 108, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[7802312, 136986, 582267, 7931, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[8384594, 144920, 259, 7, 42, "Text",
  CellTags->"MathLinkMode"],
Cell[8384856, 144929, 288, 7, 53, "DisplayFormula",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[8385169, 144940, 284, 7, 108, "Input",
  CellTags->"MathLinkMode"],
Cell[8385456, 144949, 182, 4, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8385675, 144958, 108, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[8385786, 144963, 582267, 7931, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[8968102, 152900, 154, 6, 32, "Subsubsection",
  CellTags->"MathLinkMode"],
Cell[8968259, 152908, 1079, 32, 74, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[8969363, 152944, 349, 8, 60, "Input",
  CellTags->"MathLinkMode"],
Cell[8969715, 152954, 106, 3, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8969858, 152962, 231, 6, 44, "Input",
  CellTags->"MathLinkMode"],
Cell[8970092, 152970, 184, 4, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8970313, 152979, 109, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[8970425, 152984, 600384, 8155, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[9570824, 161142, 159, 6, 26, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[9571008, 161152, 268, 7, 60, "Input",
  CellTags->"MathLinkMode"],
Cell[9571279, 161161, 186, 4, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9571502, 161170, 109, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[9571614, 161175, 582232, 7931, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[10153861, 169109, 327, 8, 60, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[10154213, 169121, 514, 10, 92, "Input",
  CellTags->"MathLinkMode"],
Cell[10154730, 169133, 4944, 137, 144, 1135, 85, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"],
Cell[10159677, 169272, 183, 5, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[10159875, 169280, 1074, 28, 106, "Text",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[10160974, 169312, 215, 6, 44, "Input",
  CellTags->"MathLinkMode"],
Cell[10161192, 169320, 184, 4, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10161413, 169329, 103, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[10161519, 169334, 593802, 8073, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"],
Cell[10755324, 177409, 183, 5, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],
Cell[10755522, 177417, 262, 7, 42, "Text",
  CellTags->"MathLinkMode"],
Cell[10755787, 177426, 205, 5, 44, "Input",
  CellTags->"MathLinkMode"],

Cell[CellGroupData[{
Cell[10756017, 177435, 296, 7, 44, "Input",
  CellTags->"MathLinkMode"],
Cell[10756316, 177444, 142663, 6212, 298, 108146, 5780, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"],
Cell[10898982, 183658, 105, 3, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10899124, 183666, 486, 11, 76, "Input",
  CellTags->"MathLinkMode"],
Cell[10899613, 183679, 183, 4, 28, "Output",
  CellTags->"MathLinkMode"]
}, Open  ]],

Cell[CellGroupData[{
Cell[10899833, 183688, 109, 3, 28, "Input",
  CellTags->"MathLinkMode"],
Cell[10899945, 183693, 624408, 8451, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"MathLinkMode"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[11524426, 192152, 97, 1, 29, "Section",
  CellTags->"POVRay"],
Cell[11524526, 192155, 875, 22, 74, "Text",
  CellTags->"POVRay"],

Cell[CellGroupData[{
Cell[11525426, 192181, 222, 5, 44, "Input",
  CellTags->"POVRay"],
Cell[11525651, 192188, 119, 3, 28, "Output",
  CellTags->"POVRay"]
}, Open  ]],
Cell[11525785, 192194, 267, 7, 42, "Text",
  CellTags->"POVRay"],

Cell[CellGroupData[{
Cell[11526077, 192205, 865, 18, 236, "Input",
  CellTags->"POVRay"],
Cell[11526945, 192225, 385107, 17832, 296, 332772, 17180, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"POVRay"]
}, Open  ]],
Cell[11912067, 210060, 138, 3, 28, "Input",
  CellTags->"POVRay"],

Cell[CellGroupData[{
Cell[11912230, 210067, 103, 3, 28, "Input",
  CellTags->"POVRay"],
Cell[11912336, 210072, 621648, 8417, 296, 551885, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"POVRay"]
}, Open  ]],

Cell[CellGroupData[{
Cell[12534021, 218494, 129, 3, 28, "Input",
  CellTags->"POVRay"],
Cell[12534153, 218499, 114, 3, 28, "Output",
  CellTags->"POVRay"]
}, Open  ]],
Cell[12534282, 218505, 561, 9, 74, "Text",
  CellTags->"POVRay"],

Cell[CellGroupData[{
Cell[12534868, 218518, 118, 3, 28, "Input",
  CellTags->"POVRay"],
Cell[12534989, 218523, 102, 3, 28, "Output",
  CellTags->"POVRay"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[12535140, 218532, 60, 1, 29, "Section",
  CellTags->"DXFFiles"],
Cell[12535203, 218535, 461, 12, 58, "Text",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{
Cell[12535689, 218551, 162, 4, 28, "Input",
  CellTags->"DXFFiles"],
Cell[12535854, 218557, 104, 3, 28, "Output",
  CellTags->"DXFFiles"]
}, Open  ]],
Cell[12535973, 218563, 638, 17, 74, "Text",
  CellTags->"DXFFiles"]
}, Closed]],

Cell[CellGroupData[{
Cell[12536648, 218585, 60, 1, 29, "Section",
  CellTags->"DXFFiles"],
Cell[12536711, 218588, 145, 4, 26, "Text",
  CellTags->"DXFFiles"],
Cell[12536859, 218594, 123, 3, 28, "Input",
  CellTags->"DXFFiles"],
Cell[12536985, 218599, 112, 3, 28, "Input",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{
Cell[12537122, 218606, 138, 3, 28, "Input",
  CellTags->"DXFFiles"],
Cell[12537263, 218611, 1093814, 49150, 69, 1073882, 48898, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"DXFFiles"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13631114, 267766, 155, 4, 28, "Input",
  CellTags->"DXFFiles"],
Cell[13631272, 267772, 125, 3, 28, "Output",
  CellTags->"DXFFiles"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13631434, 267780, 146, 4, 28, "Input",
  CellTags->"DXFFiles"],
Cell[13631583, 267786, 125, 3, 28, "Output",
  CellTags->"DXFFiles"]
}, Open  ]],
Cell[13631723, 267792, 218, 8, 26, "Text",
  CellTags->"DXFFiles"],
Cell[13631944, 267802, 183, 5, 44, "Input",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{
Cell[13632152, 267811, 237, 5, 44, "Input",
  CellTags->"DXFFiles"],
Cell[13632392, 267818, 342240, 9475, 296, 301971, 8972, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"DXFFiles"]
}, Open  ]],
Cell[13974647, 277296, 75, 1, 26, "Text",
  CellTags->"DXFFiles"],
Cell[13974725, 277299, 151, 4, 44, "Input",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{
Cell[13974901, 277307, 229, 5, 44, "Input",
  CellTags->"DXFFiles"],
Cell[13975133, 277314, 298534, 12413, 250, 264468, 11987, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"DXFFiles"]
}, Open  ]],
Cell[14273682, 289730, 349, 7, 76, "Input",
  CellTags->"DXFFiles"],

Cell[CellGroupData[{
Cell[14274056, 289741, 287, 6, 44, "Input",
  CellTags->"DXFFiles"],
Cell[14274346, 289749, 334623, 13411, 317, 283802, 12778, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"DXFFiles"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[14609018, 303166, 58, 0, 29, "Section"],
Cell[14609079, 303168, 426, 9, 58, "Text"],
Cell[14609508, 303179, 135, 3, 28, "Input"],

Cell[CellGroupData[{
Cell[14609668, 303186, 93, 2, 28, "Input"],
Cell[14609764, 303190, 103, 2, 28, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14609904, 303197, 236, 6, 44, "Input"],
Cell[14610143, 303205, 135834, 5679, 229, 113523, 5399, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[14745992, 308887, 86, 2, 28, "Input"],
Cell[14746081, 308891, 161, 3, 44, "Input"],

Cell[CellGroupData[{
Cell[14746267, 308898, 132, 3, 28, "Input"],
Cell[14746402, 308903, 315173, 13322, 296, 262246, 12663, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15061612, 322230, 340, 7, 76, "Input"],
Cell[15061955, 322239, 609302, 23071, 296, 584519, 22760, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15671294, 345315, 162, 4, 28, "Input"],
Cell[15671459, 345321, 102, 2, 28, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[15671610, 345329, 75, 1, 29, "Section",
  CellTags->"QuickDraw3D"],
Cell[15671688, 345332, 509, 11, 58, "Text",
  CellTags->"QuickDraw3D"],

Cell[CellGroupData[{
Cell[15672222, 345347, 124, 3, 28, "Input",
  CellTags->"QuickDraw3D"],
Cell[15672349, 345352, 98, 3, 28, "Output",
  CellTags->"QuickDraw3D"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[15672496, 345361, 61, 1, 29, "Section",
  CellTags->"VRML"],
Cell[15672560, 345364, 241, 7, 42, "Text",
  CellTags->"VRML"],

Cell[CellGroupData[{
Cell[15672826, 345375, 97, 3, 28, "Input",
  CellTags->"VRML"],
Cell[15672926, 345380, 176, 4, 28, "Output",
  CellTags->"VRML"]
}, Open  ]],
Cell[15673117, 345387, 592, 16, 74, "Text",
  CellTags->"VRML"]
}, Closed]],

Cell[CellGroupData[{
Cell[15673746, 345408, 69, 1, 29, "Section",
  CellTags->"PostScript"],
Cell[15673818, 345411, 698, 16, 74, "Text",
  CellTags->"PostScript"],
Cell[15674519, 345429, 115, 3, 28, "Input",
  CellTags->"PostScript"],

Cell[CellGroupData[{
Cell[15674659, 345436, 113, 3, 28, "Input",
  CellTags->"PostScript"],
Cell[15674775, 345441, 98, 3, 28, "Output",
  CellTags->"PostScript"]
}, Open  ]],
Cell[15674888, 345447, 562, 15, 58, "Text",
  CellTags->"PostScript"]
}, Closed]],

Cell[CellGroupData[{
Cell[15675487, 345467, 86, 1, 29, "Section",
  CellTags->"PNGBitmaps"],
Cell[15675576, 345470, 1649, 46, 122, "Text",
  CellTags->"PNGBitmaps"]
}, Closed]],

Cell[CellGroupData[{
Cell[15677262, 345521, 54, 1, 29, "Section",
  CellTags->"Animation"],
Cell[15677319, 345524, 1028, 29, 90, "Text",
  CellTags->"Animation"],
Cell[15678350, 345555, 115, 3, 28, "Input",
  CellTags->"Animation"],
Cell[15678468, 345560, 100, 1, 26, "Text",
  CellTags->"Animation"],
Cell[15678571, 345563, 345, 8, 108, "Input",
  CellTags->"Animation"],
Cell[15678919, 345573, 97, 1, 26, "Text",
  CellTags->"Animation"],
Cell[15679019, 345576, 114, 3, 28, "Input",
  CellTags->"Animation"],
Cell[15679136, 345581, 414, 11, 58, "Text",
  CellTags->"Animation"],
Cell[15679553, 345594, 345, 6, 156, "Input",
  CellTags->"Animation"],

Cell[CellGroupData[{
Cell[15679923, 345604, 215, 6, 28, "Input",
  CellTags->"Animation"],

Cell[CellGroupData[{
Cell[15680163, 345614, 111968, 1606, 136, 100521, 1459, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"Animation"],
Cell[15792134, 347222, 111818, 1604, 136, 100521, 1459, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"Animation"],
Cell[15903955, 348828, 111619, 1601, 136, 100521, 1459, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"Animation"],
Cell[16015577, 350431, 111862, 1604, 136, 100521, 1459, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"Animation"],
Cell[16127442, 352037, 111587, 1601, 136, 100521, 1459, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"Animation"],
Cell[16239032, 353640, 111603, 1601, 136, 100521, 1459, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"Animation"],
Cell[16350638, 355243, 112053, 1607, 136, 100521, 1459, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"Animation"]
}, Closed]]
}, Open  ]],
Cell[16462718, 356854, 470, 11, 56, "Text",
  CellTags->"Animation"],
Cell[16463191, 356867, 3492, 76, 69, "Input",
  CellTags->"Animation"],
Cell[16466686, 356945, 623, 19, 42, "Text",
  CellTags->"Animation"],

Cell[CellGroupData[{
Cell[16467334, 356968, 499, 10, 60, "Input",
  CellTags->"Animation"],
Cell[16467836, 356980, 950, 26, 60, "Output"]
}, Open  ]],
Cell[16468801, 357009, 145, 4, 26, "Text",
  CellTags->"Animation"],

Cell[CellGroupData[{
Cell[16468971, 357017, 2276, 40, 444, "Input",
  CellTags->"Animation"],
Cell[16471250, 357059, 174, 3, 22, "Message"]
}, Open  ]],
Cell[16471439, 357065, 176, 6, 26, "Text",
  CellTags->"Animation"],

Cell[CellGroupData[{
Cell[16471640, 357075, 227, 5, 28, "Input",
  CellTags->"Animation"],
Cell[16471870, 357082, 11200, 575, 265, 6116, 507, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[16483085, 357660, 451, 11, 58, "Text",
  CellTags->"Animation"],
Cell[16483539, 357673, 919, 18, 172, "Input",
  CellTags->"Animation"],
Cell[16484461, 357693, 118, 4, 26, "Text",
  CellTags->"Animation"],

Cell[CellGroupData[{
Cell[16484604, 357701, 172, 5, 76, "Input",
  CellTags->"Animation"],
Cell[16484779, 357708, 560173, 7703, 296, 548285, 7550, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"Animation"]
}, Open  ]],
Cell[17044967, 365414, 162, 6, 26, "Text",
  CellTags->"Animation"],
Cell[17045132, 365422, 210, 5, 28, "Input",
  CellTags->"Animation"],
Cell[17045345, 365429, 544, 14, 42, "Text",
  CellTags->"Animation"],
Cell[17045892, 365445, 181, 5, 28, "Input",
  CellTags->"Animation"],
Cell[17046076, 365452, 250, 5, 42, "Text",
  CellTags->"Animation"]
}, Closed]],

Cell[CellGroupData[{
Cell[17046363, 365462, 73, 1, 29, "Section",
  CellTags->"PolygonReduction"],
Cell[17046439, 365465, 229, 5, 42, "Text",
  CellTags->"PolygonReduction"],
Cell[17046671, 365472, 119, 3, 28, "Input",
  CellTags->"PolygonReduction"],
Cell[17046793, 365477, 151, 4, 28, "Input",
  CellTags->"PolygonReduction"],
Cell[17046947, 365483, 130, 3, 28, "Input",
  CellTags->"PolygonReduction"],
Cell[17047080, 365488, 456, 11, 58, "Text",
  CellTags->"PolygonReduction"],
Cell[17047539, 365501, 201, 5, 28, "Input",
  CellTags->"PolygonReduction"],
Cell[17047743, 365508, 892, 20, 90, "Text",
  CellTags->"PolygonReduction"],

Cell[CellGroupData[{
Cell[17048660, 365532, 484, 12, 76, "Input",
  CellTags->"PolygonReduction"],
Cell[17049147, 365546, 151667, 5619, 493, 90392, 4856, "GraphicsData", 
"PostScript", "Graphics",
  CellTags->"PolygonReduction"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[17200863, 371171, 44, 0, 29, "Section"],

Cell[CellGroupData[{
Cell[17200932, 371175, 67, 2, 28, "Input"],
Cell[17201002, 371179, 111, 2, 28, "Output"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[17201162, 371187, 34, 0, 29, "Section"],
Cell[17201199, 371189, 162, 4, 42, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17201398, 371198, 28, 0, 55, "Section"],
Cell[17201429, 371200, 863, 18, 106, "Text"]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

